SystemOrganization addCategory: #'VMMaker-Building'!
SystemOrganization addCategory: #'VMMaker-Interpreter'!
SystemOrganization addCategory: #'VMMaker-InterpreterSimulation'!
SystemOrganization addCategory: #'VMMaker-JIT'!
SystemOrganization addCategory: #'VMMaker-JITSimulation'!
SystemOrganization addCategory: #'VMMaker-Multithreading'!
SystemOrganization addCategory: #'VMMaker-Plugins'!
SystemOrganization addCategory: #'VMMaker-Plugins-FFI'!
SystemOrganization addCategory: #'VMMaker-Plugins-IOS'!
SystemOrganization addCategory: #'VMMaker-PostProcessing'!
SystemOrganization addCategory: #'VMMaker-SmartSyntaxPlugins'!
SystemOrganization addCategory: #'VMMaker-SpurMemoryManager'!
SystemOrganization addCategory: #'VMMaker-SpurMemoryManagerSimulation'!
SystemOrganization addCategory: #'VMMaker-Support'!
SystemOrganization addCategory: #'VMMaker-Tests'!
SystemOrganization addCategory: #'VMMaker-Translation to C'!
SystemOrganization addCategory: #'VMMaker-Utilities'!
SystemOrganization addCategory: #'VMMaker-JIT-Simulation'!

Error subclass: #ResumableVMError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Support'!

!ResumableVMError methodsFor: 'exceptionDescription' stamp: 'eem 3/16/2011 11:32'!
isResumable
	^true! !

Error subclass: #VMMakerException
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Building'!

!VMMakerException commentStamp: 'tpr 5/5/2003 12:28' prior: 0!
An exception class for the use of VMMaker!

!VMMakerException methodsFor: 'priv handling' stamp: 'eem 8/4/2008 15:31'!
isResumable
	^true! !

!PrintfNumberHolderDescriptor methodsFor: '*VMMaker-C code generation' stamp: 'eem 1/7/2021 15:26'!
transformForVMMaker
	^nil! !

InstructionStream subclass: #StackDepthFinder
	instanceVariableNames: 'stackp joins encoderClass deadCode'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Support'!

!StackDepthFinder methodsFor: 'accessing' stamp: 'eem 1/13/2019 15:38'!
allStackPointers
	"Answer a Dictionary from method to Array of stack pointers for all methods ioncluded in the receiver's method."
	"(StackDepthFinder on: StackDepthFinder >> #allStackPointers) allStackPointers"
	| allStackPointers method endPC |
	allStackPointers := Dictionary new.
	allStackPointers at: (method := self method) put: self stackPointers.
	endPC := pc.
	method nestedBlockMethods do:
		[:blockMethod|
		self method: blockMethod pc: blockMethod initialPC.
		allStackPointers at: blockMethod put: self stackPointers].
	self method: method pc: endPC.
	^allStackPointers! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 8/1/2014 07:23'!
blockReturnConstant: value
	"Return Constant From Block bytecode."
	self resetStackAfterBranchOrReturn! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 12:38'!
blockReturnTop
	"Return Top Of Stack bytecode."
	self drop.
	self resetStackAfterBranchOrReturn! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 1/20/2017 17:42'!
branchIfInstanceOf: behaviorOrArrayOfBehavior distance: delta
	self drop.
	self doJump: delta! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'cb 1/19/2017 14:11'!
branchIfNotInstanceOf: behaviorOrArrayOfBehavior distance: delta 
	self drop.
	self doJump: delta! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'cb 1/8/2018 16:06'!
callInlinePrimitive: primitiveIndex
	"The convention for inline primitives is that the argument count is primitiveIndex // 1000 - 1,
	 so receiverless 0-arg prims are from 0 to 999 (does this even make sense?), 0-arg prims
	 are from 1 to 1999, 2-arg prims from 1000 to 1999, and so on.
	 Primitives over 6000 are jumps."
	primitiveIndex >= 6000 ifTrue: [
		| upperByte delta |
		 self drop: (primitiveIndex - 4999) // 1000.
		 upperByte := self method at: pc - 6.
		 upperByte > 127 ifTrue: [upperByte := upperByte - 256].
		 delta := upperByte << 8 + (self method at: pc - 4).
		 delta < 0
			ifTrue:
				[(joins at: pc + delta) ~= stackp ifTrue: [(Notification new tag: #'bad join'; signal)]]
			ifFalse:
				[joins at: pc + delta put: stackp].
		^ self ].
	self drop: (primitiveIndex // 1000 - 1 max: 0)! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'cb 1/8/2018 16:08'!
callMappedInlinePrimitive: primIndex
	self drop: (primIndex // 50 - 1 max: 0)! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 8/8/2014 19:28'!
callPrimitive: primitiveIndex
	"Call Primitive bytecode."
	"Phhh.... stackDeltaForPrimitive:in: is in BytecodeSets.  But since the default response
	 to callPrimitive: is nothing, just handle this with an MNU for now."
	[self drop: (encoderClass stackDeltaForPrimitive: primitiveIndex in: self method) negated]
		on: MessageNotUnderstood
		do: [:ex|
			ex message selector ~~ #stackDeltaForPrimitive:in: ifTrue:
				[ex pass]]! !

!StackDepthFinder methodsFor: 'accessing' stamp: 'eem 11/26/2010 12:18'!
containsMismatchedJoin
	| end |
	end := self method endPC.
	[[self pc <= end]
		whileTrue:
			[((joins at: pc) notNil and: [stackp ~= (joins at: pc)]) ifTrue: [^true].
			 self interpretNextInstructionFor: self]]
		on: Notification
		do: [:ex|
			ex tag == #'bad join' ifTrue: [^true].
			ex pass].
	^false

	"StackDepthFinder methodDict values collect: [:m| (StackDepthFinder on: m) stackPointers]"
	"StackDepthFinder methodDict values select: [:m| (StackDepthFinder on: m) containsMismatchedJoin]"

	"| jumpers |
	 jumpers := CompiledMethod allInstances select:
						[:cm| | is |
						is := InstructionStream on: cm.
						is scanFor: [:ign| is willJump]].
	(jumpers select: [:cm| (StackDepthFinder on: cm) containsMismatchedJoin])
			asSortedCollection: [:cm1 :cm2| cm1 endPC <= cm2 endPC]"

	"| jumpers |
	 jumpers := CompiledMethod allInstances select:
						[:cm| | is |
						is := InstructionStream on: cm.
						is scanFor: [:ign| is willJump]].
	(jumpers reject: [:cm| (StackDepthFinder on: cm) containsMismatchedJoin])
			asSortedCollection: [:cm1 :cm2| cm1 endPC <= cm2 endPC]"

	"| jumpers result |
	 jumpers := CompiledMethod allInstances select:
						[:cm| | is |
						is := InstructionStream on: cm.
						is scanFor: [:ign| is willJump]].
	result := (jumpers select: [:cm| (StackDepthFinder on: cm) containsMismatchedJoin])
				asSortedCollection: [:cm1 :cm2| cm1 endPC <= cm2 endPC].
	 SystemNavigation new browseMessageList: (result collect: [:ea| ea methodReference]) name: 'mismatched joins'.
	 result"! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 4/6/2015 12:11'!
directedSuperSend: selector numArgs: numArgs
	self drop: numArgs + 1! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 10:56'!
doDup
	"Duplicate Top Of Stack bytecode."
	self push! !

!StackDepthFinder methodsFor: 'stack depth' stamp: 'eem 1/14/2017 13:39'!
doJump: delta
	deadCode ifTrue:
		[^self].
	delta < 0
		ifTrue:
			[(joins at: pc + delta) ~= stackp ifTrue: [(Notification new tag: #'bad join'; signal)]]
		ifFalse:
			[joins at: pc + delta put: stackp]! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 4/13/2015 15:25'!
doNop
	"do nothing ;-)"! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 1/14/2017 13:41'!
doPop
	"Remove Top Of Stack bytecode."
	deadCode ifFalse:
		[self drop]! !

!StackDepthFinder methodsFor: 'stack depth' stamp: 'eem 11/26/2010 11:54'!
drop
	stackp := stackp - 1! !

!StackDepthFinder methodsFor: 'stack depth' stamp: 'eem 11/26/2010 11:53'!
drop: n
	stackp := stackp - n! !

!StackDepthFinder methodsFor: 'accessing' stamp: 'eem 8/2/2014 08:57'!
encoderClass

	^encoderClass! !

!StackDepthFinder methodsFor: 'accessing' stamp: 'eem 8/2/2014 08:56'!
encoderClass: anObject

	encoderClass := anObject! !

!StackDepthFinder methodsFor: 'decoding' stamp: 'eem 1/14/2017 13:36'!
interpretNextInstructionFor: client
	(deadCode and: [(joins at: pc) notNil]) ifTrue:
		[deadCode := false].
	joins at: pc put: stackp.
	^encoderClass
		ifNil: [super interpretNextInstructionFor: client]
		ifNotNil: [encoderClass interpretNextInstructionFor: client in: self]! !

!StackDepthFinder methodsFor: 'accessing' stamp: 'eem 11/26/2010 12:48'!
joins
	^joins! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 10:57'!
jump: delta
	"Unconditional Jump bytecode."
	self doJump: delta.
	self resetStackAfterBranchOrReturn! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 10:57'!
jump: delta if: condition 
	"Conditional Jump bytecode."
	self drop.
	self doJump: delta! !

!StackDepthFinder methodsFor: 'private' stamp: 'eem 1/14/2017 13:34'!
method: method pc: startpc
	super method: method pc: startpc.
	"The + 1 is there for full blocks ending with non local return,
	as the following pc is mapped and beyong endPC"
	joins := Array new: method endPC + 1.
	stackp := method numTemps.
	deadCode := false! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 10:58'!
methodReturnConstant: value 
	"Return Constant bytecode."
	self resetStackAfterBranchOrReturn! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 10:58'!
methodReturnReceiver
	"Return Self bytecode."
	self resetStackAfterBranchOrReturn! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 12:38'!
methodReturnTop
	"Return Top Of Stack bytecode."
	self drop.
	self resetStackAfterBranchOrReturn! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 10:59'!
popIntoLiteralVariable: anAssociation 
	"Remove Top Of Stack And Store Into Literal Variable bytecode."
	self drop! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 10:59'!
popIntoReceiverVariable: offset 
	"Remove Top Of Stack And Store Into Instance Variable bytecode."
	self drop! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 10:59'!
popIntoRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	"Remove Top Of Stack And Store Into Offset of Temp Vector bytecode."
	self drop! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 10:59'!
popIntoTemporaryVariable: offset 
	"Remove Top Of Stack And Store Into Temporary Variable bytecode."
	self drop! !

!StackDepthFinder methodsFor: 'stack depth' stamp: 'eem 11/26/2010 11:53'!
push
	stackp := stackp + 1! !

!StackDepthFinder methodsFor: 'stack depth' stamp: 'eem 11/26/2010 11:53'!
push: n
	stackp := stackp + n! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 10:59'!
pushActiveContext
	"Push Active Context On Top Of Its Own Stack bytecode."
	self push! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 12:36'!
pushClosureCopyNumCopiedValues: numCopied numArgs: numArgs blockSize: blockSize
	"Push Closure bytecode."
	numCopied = 0
		ifTrue: [self push]
		ifFalse: [numCopied > 1 ifTrue:
					[self drop: numCopied - 1]].
	self doJump: blockSize.
	stackp := numCopied + numArgs! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 12:19'!
pushConsArrayWithElements: numElements
	"Push Cons Array of size numElements popping numElements items from the stack into the array bytecode."
	self drop: numElements - 1! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 11:01'!
pushConstant: value
	"Push Constant, value, on Top Of Stack bytecode."
	self push! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/15/2012 14:18'!
pushExplicitOuter: level
	"Push receiver for explcit outer, on Top Of Stack bytecode."
	self push! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 3/23/2017 16:57'!
pushFullClosure: literal numCopied: numCopied
	"Push Closure bytecode."
	numCopied = 0
		ifTrue: [self push]
		ifFalse: [numCopied > 1 ifTrue:
					[self drop: numCopied - 1]]! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 3/23/2017 16:59'!
pushFullClosure: literal numCopied: numCopied receiverOnStack: receiverOnStack ignoreOuterContext: ignoreOuterContext
	"Push Closure bytecode."
	| delta |
	delta := 1 - (receiverOnStack ifTrue: [1] ifFalse: [0]) - numCopied.
	delta ~= 0 ifTrue: [self drop: delta negated]! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 11:01'!
pushLiteralVariable: anAssociation
	"Push Contents Of anAssociation On Top Of Stack bytecode."
	self push! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 12:19'!
pushNewArrayOfSize: numElements 
	"Push New Array of size numElements bytecode."
	self push! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 11:01'!
pushReceiver
	"Push Active Context's Receiver on Top Of Stack bytecode."
	self push! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 11:01'!
pushReceiverVariable: offset
	"Push Contents Of the Receiver's Instance Variable Whose Index 
	is the argument, offset, On Top Of Stack bytecode."
	self push! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 11:01'!
pushRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	"Push Contents at Offset in Temp Vector bytecode."
	self push! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 11:02'!
pushTemporaryVariable: offset
	"Push Contents Of Temporary Variable Whose Index Is the 
	argument, offset, On Top Of Stack bytecode."
	self push! !

!StackDepthFinder methodsFor: 'stack depth' stamp: 'eem 1/14/2017 13:34'!
resetStackAfterBranchOrReturn
	pc <= joins size ifTrue:
		[(joins at: pc) ifNil:
			["We would like to raise an error here, but we can't because the bytecode
			  compiler doesn't remove dead code.  e.g. the following doesn't reach ^nil
			  but is legal Smalltalk:
				self foo
					ifTrue: [^self bar]
					ifFalse: [^self baz].
				^nil"
				deadCode := true.
				^self].
		 stackp := joins at: pc]! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 11:02'!
send: selector super: supered numArgs: numArgs
	"Send Message With Selector, selector, bytecode. The argument, 
	supered, indicates whether the receiver of the message is specified with 
	'super' in the source method. The arguments of the message are found in 
	the top numArguments locations on the stack and the receiver just 
	below them."
	self drop: numArgs! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 1/11/2013 16:59'!
sendToAbsentDynamicSuperclass: selector numArgs: numArgs
	"Dynamic Superclass Send Message With Selector, selector, to absent implicit receiver bytecode."
	self drop: numArgs - 1 "e.g. if no args pushes a result"! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 1/11/2013 16:58'!
sendToAbsentImplicitReceiver: selector numArgs: numArgs
	"Send Message With Selector, selector, to absent implicit receiver bytecode."
	self drop: numArgs - 1 "e.g. if no args pushes a result"! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'rmacnak 3/30/2015 20:58'!
sendToAbsentOuter: selector numArgs: numArgs depth: depth
	self drop: numArgs - 1 "e.g. if no args pushes a result"! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'rmacnak 3/30/2015 20:59'!
sendToAbsentSelf: selector numArgs: numArgs
	self drop: numArgs - 1 "e.g. if no args pushes a result"! !

!StackDepthFinder methodsFor: 'accessing' stamp: 'eem 11/26/2010 11:58'!
stackPointerAt: aPC 
	[pc < aPC] whileTrue:
		[self interpretNextInstructionFor: self].
	^pc = aPC ifTrue: [stackp]! !

!StackDepthFinder methodsFor: 'accessing' stamp: 'rsf 9/8/2016 17:40'!
stackPointers
	"Collect the stack depth for each bytecode in the method.
	 For the convenience of clients, write the stack depth for extended bytecodes
	 to the bytecode pc, as well as the bytecocdes' extensions' pcs."
	| method stackPointers end |
	"endPC +1 because of full blocks that can end with NLR effectively mapping the pc right after the end PC"
	stackPointers :=  Array new: (end := (method := self method) endPC) + 1.
	encoderClass ifNil: [encoderClass := method encoderClass].
	[pc <= end] whileTrue:
		[stackPointers at: pc put: stackp.
		 encoderClass extensionsAt: pc in: method into:
			[:extA :extB :delta|
			2 to: delta by: 2 do: [ :dlt |
				stackPointers at: pc + dlt put: stackp]].
		 self interpretNextInstructionFor: self].
	^stackPointers! !

!StackDepthFinder methodsFor: 'accessing' stamp: 'eem 11/26/2010 11:58'!
stackp
	^stackp! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 10:55'!
storeIntoLiteralVariable: anAssociation 
	"Store Top Of Stack Into Literal Variable Of Method bytecode."
! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 10:55'!
storeIntoReceiverVariable: offset 
	"Store Top Of Stack Into Instance Variable Of Method bytecode."
! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 10:55'!
storeIntoRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	"Store Top Of Stack And Store Into Offset of Temp Vector bytecode."! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 11/26/2010 10:55'!
storeIntoTemporaryVariable: offset 
	"Store Top Of Stack Into Temporary Variable Of Method bytecode."
! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'cb 6/9/2015 09:22'!
trap
	! !

!StackDepthFinder methodsFor: 'instruction decoding' stamp: 'eem 8/2/2014 10:16'!
trapIfNotInstanceOf: behaviorOrArrayOfBehavior
	"If the top of stack is not an instance of either the argument, or, if the argument is an Array,
	  any of the elements of the argument, send the class trap message to the current context."
	self drop! !

!Stream methodsFor: '*VMMaker-JITSimulation-testing' stamp: 'eem 9/19/2018 08:42'!
isFakeStdinStream
	^false! !

!Stream methodsFor: '*VMMaker-JITSimulation-testing' stamp: 'eem 9/19/2018 10:21'!
isTranscriptStream
	^false! !

Model subclass: #VMMakerTool
	instanceVariableNames: 'vmMaker allPluginsList allPluginsSelection allPluginsSelectionsArray internalPluginsList internalPluginsSelection internalPluginsSelectionsArray externalPluginsList externalPluginsSelection externalPluginsSelectionsArray logger interpreterClassMorph platformPathMorph platformNameMorph generatedPathMorph'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Building'!

!VMMakerTool commentStamp: '<historical>' prior: 0!
VMMakerTool help information
------------------------------------
If you really get stuck, send mail to the Squeak mailing list, squeak-dev@lists.squeakfoundation.org

	VMMakerTool openInWorld

What this is
--------------
This tool is a simple interactive interface to VMMaker. You can change the directory paths for where the system looks for the platform files (those C files that are handwritten for each platform) and where it will put the assembled sources (the appropriate platform files and generated files) ready for you to compile into a new vm. You can change the platform for which it will generate files. You can choose which plugins are built and whether they are built for internal or external use. 

How to use it
---------------
To build a configuration, drag plugins from the leftmost  'Plugins not built' list to either the 'Internal Plugins' list or the 'External Plugins' list.  Plugins that cannot be built on your machine due to missing files will not be draggable.
Once you have a configuration, you can save it for later retrieval by pressing the 'Save Configuration' button. Unsurprisingly you can reload a saved configuration with the 'Load Configuration' button.

To generate an entire code tree, press the 'Generate All' button. This will process all the vm and plugin files needed for your configuration. To generate only the files for the vm and any internal plugins, use the 'Generate Core VM' button. This will be most useful if you are experimenting with the design of the vm internals or new object memory layouts etc. The 'Generate External Plugins' button will regenerate all the plugins in the External Plugins list. Note that 'excess' directories will be deleted each time you generate the vm in order to reduce potential confusion if you move a plugin from internal to external etc. If you repeatedly generate the vm only the files that appear to be out of date will be recreated; this drastically reduces the time taken if you have only changed a single plugin class for example.

You can also generate internal or external plugins singly, using the menus in the lists but be warned - internal plugins are tightly related to the generated file 'vm/sqNamedPrims.h' and adding or removing an internal plugin without regenerating this (via 'Generate Core VM' or 'Generate All') will cause much grief. The application attempts to prevent this, but there are surely ways to confuse both yourself and the code. In general when writing experimental plugins it is much simpler to build them as external during the development cycle. 

If the default path for the platforms code is not correct for your machine you can use the 'Find Path' button to search for a plausible directory. Note that this could take an arbitrarily long time on a machine with connections to other machines since you may end up searching all their disc space as well.

You can choose from a menu of all known platforms (at least, all those known in the set of files on your machine) by using the 'Find platform' button. This is useful if you want to generate files for some other platform and feel uncertain of the exact spelling. By default the platform will be set to that upon which you are running.

If you feel the need to delete all the generated files you can press the 'Clean out' button - this will recursively delete everything below the path for the generated sources.

Details
-------
You really ought to read the class comment for VMMaker. Really. Go on, do it now.

Errors
-------
A number of errors are possible, mostly relating to the two directory paths and the platform name. As much as possible these are trapped and you will see 'inform' menus to let you know. Inevitably, if you put in the effort, you will be able to confuse the tool and break it.
!

!VMMakerTool class methodsFor: 'instance creation' stamp: 'EstebanLorenzano 5/21/2015 21:39'!
initialize

	Smalltalk at: #TheWorldMenu ifPresent: [ :class |
		((class compiledMethodAt: #registerOpenCommand: ifAbsent: [^self]) sendsSelector: #deprecated:) 
			ifFalse: [class registerOpenCommand: (Array with: 'VMMaker' with: (Array with: self with: #openInWorld))]]
	! !

!VMMakerTool class methodsFor: 'instance creation' stamp: ''!
openInWorld
	"Build a VMMakerTool and open it"
	"VMMakerTool openInWorld"

	^self new buildWindow openInWorld! !

!VMMakerTool class methodsFor: 'unloading' stamp: 'tpr 2/14/2004 21:41'!
unload

	TheWorldMenu unregisterOpenCommandWithReceiver: self! !

!VMMakerTool methodsFor: 'drag and drop' stamp: 'nk 4/5/2005 23:02'!
acceptDroppingMorph: transferMorph event: evt inMorph: aMorph
	"Here we are fetching information from the dropped transferMorph and performing the correct action for this drop.
	As long as the source is part of this tool, move the dragged item from the source list to the destination list"

	^self moveItem: transferMorph passenger from: transferMorph source to: aMorph! !

!VMMakerTool methodsFor: 'window construction' stamp: 'tpr 4/6/2005 21:32'!
addSecondButtonRowToWindow: sysWin startingAt: initialVerticalOffset 
	| verticalOffset box |
	verticalOffset := initialVerticalOffset.
	"add a row of buttons to start up various actions"
	box := AlignmentMorph new vResizing: #shrinkWrap;  layoutInset: 6@3; cellInset: 6@0; wrapCentering: #center.
	box addMorph: (TextMorph new contents: 'Generate:' translated asText allBold) lock.
	box addMorphBack: (SimpleButtonMorph new target: self;
			 label: 'Entire';
			 actionSelector: #generateAll;
			 hResizing: #spaceFill;
			 setBalloonText: 'Generate the sources for the core VM and all chosen internal and external plugins').
	box addMorphBack: (SimpleButtonMorph new target: self;
			 label: 'Core+Internal';
			 actionSelector: #generateCore;
			 hResizing: #spaceFill;
			 setBalloonText: 'Generate the sources for the core vm and any internal plugins').

	box addMorphBack: (SimpleButtonMorph new target: self;
			 label: 'External Plugins';
			 actionSelector: #generateExternal;
			 hResizing: #spaceFill;
			 setBalloonText: 'Generate the sources for all external plugins').
	box addMorphBack: ((AlignmentMorph inARow: {StringMorph new contents: '64 bit VM?'. UpdatingThreePhaseButtonMorph checkBox target: self;
				 actionSelector: #toggle64BitVM;
				 getSelector: #isFor64BitVM}) layoutInset: 3;
			 cellInset: 5;
			 color: Color blue veryMuchLighter;
			 setBalloonText: 'Set to build a 64 bit VM or a 32bit VM' yourself).
	sysWin
		addMorph: box
		fullFrame: (LayoutFrame
				fractions: (0 @ 0 corner: 1 @ 0)
				offsets: (0 @ verticalOffset corner: 0 @ (verticalOffset := verticalOffset + box height - 1))).

	^verticalOffset.! !

!VMMakerTool methodsFor: 'window construction' stamp: 'tpr 4/6/2005 21:32'!
addTopButtonRowToWindow: sysWin
	| verticalOffset box |
	verticalOffset := 0.

	"add a row of buttons to start up various actions"
	box := AlignmentMorph new vResizing: #shrinkWrap;
				 layoutInset: 6 @ 3;
				 cellInset: 6 @ 0;
				 wrapCentering: #center.
	box addMorphBack: (SimpleButtonMorph new target: self;
			 label: 'Help';
			 actionSelector: #helpText;
			 hResizing: #spaceFill;
			 setBalloonText: 'Open the help window').
	box addMorphBack: (TextMorph new contents: 'Configuration File:' translated asText allBold) lock.	
	box addMorphBack: (SimpleButtonMorph new target: self;
			 label: 'Load';
			 actionSelector: #loadConfig;
			 hResizing: #spaceFill;
			 setBalloonText: 'Load a previously saved configuration').
	box addMorphBack: (SimpleButtonMorph new target: self;
			 label: 'Save';
			 actionSelector: #saveConfig;
			 hResizing: #spaceFill;
			 setBalloonText: 'Save the current configuration').
	sysWin
		addMorph: box
		fullFrame: (LayoutFrame
				fractions: (0 @ 0 corner: 1 @ 0)
				offsets: (0 @ verticalOffset corner: 0 @ (verticalOffset := verticalOffset + box height - 1))).
	^ verticalOffset! !

!VMMakerTool methodsFor: 'menus' stamp: 'eem 9/16/2008 15:03'!
availableListMenu: aMenu

	aMenu addTranslatedList: #(
			('make all external'						makeAllPluginsExternal)
			('make all internal'						makeAllPluginsInternal)
			('make all available'					makeAllPluginsAvailable)
			-
			('browse plugin' 						browseSelectedAvailablePlugin)
			('refresh list' 							refreshAvailablePluginsList)).
	^ aMenu! !

!VMMakerTool methodsFor: 'list access' stamp: 'tpr 10/10/2001 12:36'!
availableListSelectionAt: index
"return the boolean to say if the available plugin at index is selected"
	^allPluginsSelectionsArray at: index! !

!VMMakerTool methodsFor: 'list access' stamp: 'tpr 10/10/2001 12:37'!
availableListSelectionAt: index put: bool
"set the boolean to say if the available plugin at index is selected"
	^allPluginsSelectionsArray at: index put: bool! !

!VMMakerTool methodsFor: 'list access' stamp: ''!
availableModules
	^vmMaker availablePlugins! !

!VMMakerTool methodsFor: 'menus' stamp: 'rww 9/25/2001 02:00'!
browseSelectedAvailablePlugin

	| plugin |
	plugin := self availableModules at: self currentAvailableModuleIndex ifAbsent: [^self].
	(Smalltalk classNamed: plugin) browseHierarchy! !

!VMMakerTool methodsFor: 'menus' stamp: 'rww 9/25/2001 02:00'!
browseSelectedExternalPlugin

	| plugin |
	plugin := self externalModules at: self currentExternalModuleIndex ifAbsent: [^self].
	(Smalltalk classNamed: plugin) browseHierarchy! !

!VMMakerTool methodsFor: 'menus' stamp: 'rww 9/25/2001 02:01'!
browseSelectedInternalPlugin

	| plugin |
	plugin := self internalModules at: self currentInternalModuleIndex ifAbsent: [^self].
	(Smalltalk classNamed: plugin) browseHierarchy! !

!VMMakerTool methodsFor: 'window construction' stamp: 'eem 5/17/2015 14:51'!
buildCenterRows
	| rows color1 color2 |
	color1 := Color blue veryMuchLighter.
	color2 := Color green veryMuchLighter.
	rows := Morph new color: Color transparent;
				layoutPolicy: TableLayout new;
				vResizing: #spaceFill;
				extent: 550 @ (TextStyle defaultFont height * 14);
				hResizing: #spaceFill;
				listDirection: #topToBottom;
				borderStyle: (BorderStyle complexAltRaised width: 2);
				wrapDirection: #none;
				wrapCentering: #center;
				yourself.
	rows
		addMorphBack: ((self
				entryRowWithLabel: 'Interpreter class name:'
				balloonText: 'The name of the Interpreter class'
				getFieldText: #interpreterClassName
				setFieldText: #interpreterClassName:
				buttonLabel: nil
				buttonAction: nil
				buttonBalloonText: nil)
				color: color1).
	interpreterClassMorph := rows submorphs last submorphs first.

	rows
		addMorphBack: ((self
				entryRowWithLabel: 'Path to platforms code:'
				balloonText: 'The directory where the platform source tree is found; can be edited in text field to the right. Default of {working directory}/src is strongly recommended'
				getFieldText: #platformsPathText
				setFieldText: #platformsPathText:
				buttonLabel: 'Find Path'
				buttonAction: #findPlatformsPath
				buttonBalloonText: 'Choose the directory where you keep the platform specific code from a file dialogue')
				color: color2).
	platformPathMorph := rows submorphs last submorphs second.

	rows
		addMorphBack: ((self
				entryRowWithLabel: 'Platform name:'
				balloonText: 'The platform name (as returned by Smalltalk platformName - unix, Mac OS, RISCOS, win32 etc); can be edited (in text field to the right) to cross generate'
				getFieldText: #platformNameText
				setFieldText: #platformNameText:
				buttonLabel: 'Find platform'
				buttonAction: #platformsListMenu
				buttonBalloonText: 'Choose from a list of known platforms. The default is this current platform.')
				color: color1).
	platformNameMorph := rows submorphs last submorphs second.

	rows
		addMorphBack: ((self
				entryRowWithLabel: 'Path to generated sources'
				balloonText: 'The directory where the built sources will be placed; can be edited in text field to the right. The default is strongly recommended; makefile alterations may be needed if you use a different path.'
				getFieldText: #sourcePathText
				setFieldText: #sourcePathText:
				buttonLabel: 'Clean out'
				buttonAction: #cleanoutSrcDir
				buttonBalloonText: 'Clean out all the files in the target directory, ready for a clean build')
				color: color2).
	generatedPathMorph := rows submorphs last submorphs second.

	^ rows! !

!VMMakerTool methodsFor: 'window construction' stamp: 'nk 4/5/2005 23:05'!
buildWindow
	"VMMakerTool openInWorld"
	| sysWin box verticalOffset |
	sysWin := (SystemWindow labelled: 'VMMaker')
				model: self.
	verticalOffset := self addTopButtonRowToWindow: sysWin.
	verticalOffset := self addSecondButtonRowToWindow: sysWin startingAt: verticalOffset.
	box := self buildCenterRows.
	sysWin
		addMorph: box
		fullFrame: (LayoutFrame
				fractions: (0 @ 0 corner: 1 @ 0)
				offsets: (0 @ verticalOffset corner: 0 @ (verticalOffset := verticalOffset + box height - 1))).
	"Add the list of plugins that are available to build"
	allPluginsList := (PluggableListMorph
				on: self
				list: #availableModules
				selected: #currentAvailableModuleIndex
				changeSelected: #currentAvailableModuleIndex:
				menu: #availableListMenu:
				keystroke: nil) enableDragNDrop.
	allPluginsList hResizing: #spaceFill;
		 vResizing: #spaceFill;
		 borderWidth: 0.
	box := AlignmentMorph newColumn.
	box addMorphBack: (TextMorph new contents: 'Plugins not built' asText allBold;
			 lock);
		 setBalloonText: 'List of plugins that are available to build but not yet chosen. Drag to either other list or use menu option to move in bulk'.
	box addMorphBack: allPluginsList.
	sysWin
		addMorph: box
		fullFrame: (LayoutFrame
				fractions: (0 @ 0 corner: 1 / 3 @ 1)
				offsets: (0 @ verticalOffset corner: 0 @ -100)).
	"make the list for plugins that will be built for internal linking"
	internalPluginsList := (PluggableListMorph
				on: self
				list: #internalModules
				selected: #currentInternalModuleIndex
				changeSelected: #currentInternalModuleIndex:
				menu: #internalListMenu:
				keystroke: nil) enableDragNDrop.
	internalPluginsList hResizing: #spaceFill;
		 vResizing: #spaceFill;
		 borderWidth: 0.
	box := AlignmentMorph newColumn.
	box addMorphBack: (TextMorph new contents: 'Internal Plugins' asText allBold;
			 lock);
		 setBalloonText: 'List of plugins chosen to be built internally'.
	box addMorphBack: internalPluginsList.
	sysWin
		addMorph: box
		fullFrame: (LayoutFrame
				fractions: (1 / 3 @ 0 corner: 2 / 3 @ 1)
				offsets: (0 @ verticalOffset corner: 0 @ -100)).
	"make the list for plugins to be built externally (ie as DLLs, SO
	or 
	whatever suits the platform"
	externalPluginsList := (PluggableListMorph
				on: self
				list: #externalModules
				selected: #currentExternalModuleIndex
				changeSelected: #currentExternalModuleIndex:
				menu: #externalListMenu:
				keystroke: nil) enableDragNDrop.
	externalPluginsList hResizing: #spaceFill;
		 vResizing: #spaceFill;
		 borderWidth: 0.
	box := AlignmentMorph newColumn.
	box addMorphBack: (TextMorph new contents: 'External Plugins' asText allBold;
			 lock);
		 setBalloonText: 'List of plugins chosen to be built externally'.
	box addMorphBack: externalPluginsList.
	sysWin
		addMorph: box
		fullFrame: (LayoutFrame
				fractions: (2 / 3 @ 0 corner: 1 @ 1)
				offsets: (0 @ verticalOffset corner: 0 @ -100)).
	sysWin
		addMorph: (PluggableTextMorph
				on: logger
				text: nil
				accept: nil
				readSelection: nil
				menu: nil)
		fullFrame: (LayoutFrame
				fractions: (0 @ 1 corner: 1 @ 1)
				offsets: (0 @ -100 corner: 0 @ 0)).
	^ sysWin! !

!VMMakerTool methodsFor: 'settings' stamp: 'nk 4/5/2005 23:31'!
checkOK
	"check as many settings as we can and report true if all seems ok"
	(interpreterClassMorph accept; hasUnacceptedEdits) ifTrue:[^false].
	(platformPathMorph accept; hasUnacceptedEdits) ifTrue:[^false].
	(platformNameMorph accept; hasUnacceptedEdits) ifTrue:[^false].
	(generatedPathMorph accept; hasUnacceptedEdits) ifTrue:[^false].

	[vmMaker platformPluginsDirectory; crossPlatformPluginsDirectory]
		on: VMMakerException
		do: [:ex| self inform: ex messageText.
			^ false].
	^ true! !

!VMMakerTool methodsFor: 'path access' stamp: 'tpr 3/27/2002 15:19'!
cleanoutSrcDir
	"remove the entire generated src tree, ready for a nice clean build"
	vmMaker deleteEntireGeneratedTree! !

!VMMakerTool methodsFor: 'list access' stamp: ''!
currentAvailableModuleIndex
	allPluginsSelection ifNil:[^0].
	^allPluginsSelection! !

!VMMakerTool methodsFor: 'list access' stamp: 'nk 12/16/2002 08:49'!
currentAvailableModuleIndex: anInteger
	allPluginsSelection := anInteger.
	self changed: #currentAvailableModuleIndex! !

!VMMakerTool methodsFor: 'list access' stamp: ''!
currentExternalModuleIndex
	externalPluginsSelection ifNil:[^0].
	^externalPluginsSelection! !

!VMMakerTool methodsFor: 'list access' stamp: 'nk 12/16/2002 08:49'!
currentExternalModuleIndex: anInteger
	externalPluginsSelection := anInteger.
	self changed: #currentExternalModuleIndex! !

!VMMakerTool methodsFor: 'list access' stamp: ''!
currentInternalModuleIndex
	internalPluginsSelection ifNil:[^0].
	^internalPluginsSelection! !

!VMMakerTool methodsFor: 'list access' stamp: 'nk 12/16/2002 08:49'!
currentInternalModuleIndex: anInteger
	internalPluginsSelection := anInteger.
	self changed: #currentInternalModuleIndex! !

!VMMakerTool methodsFor: 'drag and drop' stamp: 'tpr 10/12/2001 15:27'!
dragPassengerFor: item inMorph: dragSource 
	(dragSource isKindOf: PluggableListMorph)
		ifFalse: [^item].
	^item contents! !

!VMMakerTool methodsFor: 'drag and drop' stamp: ''!
dragTransferTypeForMorph: dragSource 
	^(dragSource isKindOf: PluggableListMorph)
		ifTrue: [dragSource getListSelector]! !

!VMMakerTool methodsFor: 'window construction' stamp: 'eem 3/6/2018 14:39'!
entryRowWithLabel: label balloonText: balloonText getFieldText: getTextSelector setFieldText: setTextSelector buttonLabel: buttonLabel buttonAction: buttonAction buttonBalloonText: buttonBalloonText 
	| row lWidth |
	lWidth := TextStyle defaultFont pixelSize * 11.
	row := Morph new color: Color transparent;
				 hResizing: #spaceFill;
				 vResizing: #spaceFill;
				 extent: 550 @ 40;
				 layoutPolicy: ProportionalLayout new;
				 borderWidth: 2;
				 setBalloonText: balloonText translated;
				 yourself.
	row
		addMorph: (TextMorph new contents: label translated asText allBold) lock
		fullFrame: (LayoutFrame
				fractions: (0 @ 0 corner: 0 @ 1)
				offsets: (3 @ 3 corner: lWidth @ -3)).
	row
		addMorph: ((PluggableTextMorph
				on: self
				text: getTextSelector
				accept: setTextSelector) vScrollBarPolicy: #never;
				 acceptOnCR: true)
		fullFrame: (LayoutFrame
				fractions: (0 @ 0 corner: 1 @ 1)
				offsets: (lWidth + 10 @ 0 corner: (lWidth / 1.8 + 10) negated @ 0)).
	buttonAction
		ifNotNil: [row
				addMorph: (SimpleButtonMorph new target: self;
						 label: buttonLabel translated;
						 actionSelector: buttonAction;
						 hResizing: #spaceFill;
						 setBalloonText: buttonBalloonText translated)
				fullFrame: (LayoutFrame
						fractions: (1 @ 0 corner: 1 @ 1)
						offsets: ((lWidth / 1.8 + 5) negated @ 3 corner: -5 @ -3))].
	^ row! !

!VMMakerTool methodsFor: 'menus' stamp: 'nk 4/5/2005 20:25'!
externalListMenu: aMenu

	aMenu addTranslatedList:#(
			('make all external'						makeAllPluginsExternal)
			('make all internal'						makeAllPluginsInternal)
			('make all available'					makeAllPluginsAvailable)
			-
			('browse plugin' 						browseSelectedExternalPlugin)
			-
			('generate plugin'						generateSelectedExternalPlugin)).
	^ aMenu! !

!VMMakerTool methodsFor: 'list access' stamp: 'tpr 10/10/2001 12:36'!
externalListSelectionAt: index
"return the boolean to say if the external plugin at index is selected"
	^externalPluginsSelectionsArray at: index! !

!VMMakerTool methodsFor: 'list access' stamp: 'tpr 10/10/2001 12:37'!
externalListSelectionAt: index put: bool
"set the boolean to say if the external plugin at index is selected"
	^externalPluginsSelectionsArray at: index put: bool! !

!VMMakerTool methodsFor: 'list access' stamp: ''!
externalModules
	^vmMaker externalModules! !

!VMMakerTool methodsFor: 'path access' stamp: 'tpr 9/15/2003 13:58'!
findPlatformsPath
	| dir |
	dir := FileList2 modalFolderSelector.
	dir ifNil: [^nil].
	self platformsPathText: dir pathName! !

!VMMakerTool methodsFor: 'path access' stamp: 'ar 3/10/2002 14:54'!
findPlatformsPathFrom: fd informing: bar
	| dirNames possiblePath |
	bar value: 'Searching in ', fd pathName.
	dirNames := fd directoryNames.
	(dirNames includes: 'platforms') ifTrue:[
		possiblePath := fd pathName, fd pathNameDelimiter asString, 'platforms'.
		(self confirm: 'Found a platforms directory at
', possiblePath,'
Do you want me to use it?') ifTrue:[^possiblePath].
	].
	dirNames do:[:dd|
		possiblePath := self findPlatformsPathFrom: (fd directoryNamed: dd) informing: bar.
		possiblePath ifNotNil:[^possiblePath].
	].
	^nil! !

!VMMakerTool methodsFor: 'generate sources' stamp: 'tpr 10/21/2001 11:00'!
generateAll
	"tell the vmMaker to build all the sources"
	self checkOK
		ifTrue: [[vmMaker generateEntire]
				on: VMMakerException
				do: [:ex| self inform: ex messageText]]! !

!VMMakerTool methodsFor: 'generate sources' stamp: 'tpr 10/21/2001 11:01'!
generateCore
	"tell the vmMaker to build all the core vm sources"
	self checkOK
		ifTrue: [[vmMaker generateMainVM]
		on: VMMakerException
		do: [:ex| self inform: ex messageText]]! !

!VMMakerTool methodsFor: 'generate sources' stamp: 'tpr 10/21/2001 11:02'!
generateExternal
	"tell the vmMaker to build all the externally linked plugin sources"
	self checkOK
		ifTrue: [[vmMaker generateExternalPlugins]
		on: VMMakerException
		do: [:ex | self inform: ex messageText]]! !

!VMMakerTool methodsFor: 'generate sources' stamp: 'eem 6/3/2011 08:55'!
generateSelectedExternalPlugin

	| pluginName |
	pluginName := self externalModules
						at: self currentExternalModuleIndex
						ifAbsent: [^self inform: 'no such plugin'].
	vmMaker interpreterClass
		ifNil: [^self inform: 'please set an interpreter class']
		ifNotNil: [:interpreterClass| interpreterClass initialize].
	self checkOK ifTrue:
		[[(Smalltalk classNamed: pluginName) touch.
		  vmMaker generateExternalPlugin: pluginName; logDateAndTime]
			on: VMMakerException
			do: [:ex| self inform: ex messageText]]! !

!VMMakerTool methodsFor: 'generate sources' stamp: 'tpr 5/29/2013 13:47'!
generateSelectedInternalPlugin

	| pluginName |
	pluginName := self internalModules
						at: self currentInternalModuleIndex
						ifAbsent: [^self inform: 'no such plugin'].
	vmMaker interpreterClass
		ifNil: [^self inform: 'please set an interpreter class']
		ifNotNil: [:interpreterClass| interpreterClass initialize].
	self checkOK ifTrue:
		[[(Smalltalk classNamed: pluginName) touch.
		  vmMaker generateInternalPlugin: pluginName]
			on: VMMakerException
			do: [:ex| self inform: ex messageText]]! !

!VMMakerTool methodsFor: 'menus' stamp: 'tpr 10/14/2001 20:11'!
helpText
	(StringHolder new contents: self class comment)
		openLabel: 'VMMakerTool help'
! !

!VMMakerTool methodsFor: 'initialisation' stamp: 'ar 3/10/2002 15:09'!
initialExtent
 	^600@450! !

!VMMakerTool methodsFor: 'list access' stamp: ''!
initialModules
	^vmMaker availableModules! !

!VMMakerTool methodsFor: 'initialisation' stamp: 'ar 3/10/2002 15:06'!
initialize
	logger := TranscriptStream new.
	vmMaker := VMMaker default.
	vmMaker logger: logger.
	vmMaker addDependent: self.
	allPluginsSelectionsArray := Array new: self availableModules size withAll: false.
	internalPluginsSelectionsArray := Array new.
	externalPluginsSelectionsArray := Array new.! !

!VMMakerTool methodsFor: 'menus' stamp: 'nk 4/5/2005 20:25'!
internalListMenu: aMenu

	aMenu addTranslatedList:#(
			('make all external'						makeAllPluginsExternal)
			('make all internal'						makeAllPluginsInternal)
			('make all available'					makeAllPluginsAvailable)
			-
			('browse plugin' 						browseSelectedInternalPlugin)
			-
			('generate plugin'						generateSelectedInternalPlugin)).
	^ aMenu! !

!VMMakerTool methodsFor: 'list access' stamp: 'tpr 10/10/2001 12:36'!
internalListSelectionAt: index
"return the boolean to say if the internal plugin at index is selected"
	^internalPluginsSelectionsArray at: index! !

!VMMakerTool methodsFor: 'list access' stamp: 'tpr 10/10/2001 12:37'!
internalListSelectionAt: index put: bool
"set the boolean to say if the internal plugin at index is selected"
	^internalPluginsSelectionsArray at: index put: bool! !

!VMMakerTool methodsFor: 'list access' stamp: ''!
internalModules
	^vmMaker internalModules! !

!VMMakerTool methodsFor: 'path access' stamp: 'svp 11/14/2002 21:11'!
interpreterClassName
	"return a Text for the path to the generated sources"
	^[vmMaker interpreterClass name asString] 
		on: VMMakerException 
		do:[:ex| ex return:'<invalid class>'].! !

!VMMakerTool methodsFor: 'path access' stamp: 'svp 11/14/2002 21:27'!
interpreterClassName: aText
	"set the interpreter class name"

	[vmMaker interpreterClassName: aText asString] 
		on: VMMakerException 
		do: [:ex| self inform:'problem with this class name; does this class exist?'. 
			^false].
	^true! !

!VMMakerTool methodsFor: 'generate sources' stamp: 'tpr 4/4/2005 19:11'!
isFor64BitVM
"do I build a 64bit VM or not?"
	^vmMaker isFor64BitVM! !

!VMMakerTool methodsFor: 'list access' stamp: 'tpr 10/12/2001 15:29'!
listForMorph: aMorph
	"work out which list is the one associated with this morph"
	allPluginsList = aMorph ifTrue:[^allPluginsList getListSelector].
	internalPluginsList = aMorph ifTrue:[^internalPluginsList getListSelector].
	externalPluginsList =aMorph ifTrue:[^externalPluginsList getListSelector].
	^nil! !

!VMMakerTool methodsFor: 'list access' stamp: ''!
listMorphs
	^Array with: allPluginsList with: internalPluginsList with: externalPluginsList! !

!VMMakerTool methodsFor: 'configurations' stamp: 'ar 3/10/2002 15:05'!
loadConfig
	| fileResult file |
	fileResult := (StandardFileMenu oldFileMenu: FileDirectory default withPattern: '*.config')
				startUpWithCaption: 'Select VMMaker configuration...'.
	fileResult
		ifNotNil: [file := fileResult directory fullNameFor: fileResult name.
			[vmMaker := VMMaker forConfigurationFile: file.
			vmMaker logger: logger.
			vmMaker platformDirectory]
				on: Error
				do: [self inform: 'Possible problem with path settings or platform name?'].
			self updateAllViews]! !

!VMMakerTool methodsFor: 'menus' stamp: 'tpr 10/16/2001 09:58'!
makeAllPluginsAvailable


	vmMaker makeAllModulesAvailable! !

!VMMakerTool methodsFor: 'menus' stamp: 'tpr 10/16/2001 09:58'!
makeAllPluginsExternal


	vmMaker makeAllModulesExternal! !

!VMMakerTool methodsFor: 'menus' stamp: 'tpr 10/16/2001 09:58'!
makeAllPluginsInternal


	vmMaker makeAllModulesInternal! !

!VMMakerTool methodsFor: 'drag and drop' stamp: 'tpr 10/21/2001 11:19'!
moveItem: transferedMorph from: sourceListMorph to: destListMorph 
	"As part of a drag operation we have to move the item carried by the  
	transfer morph from a source list to a destination list"
	"work out which list is involved and add the item to it"
	| destlist srclist |

	"no need to do anything if we drop on the same place from which we dragged" 
	sourceListMorph = destListMorph
		ifTrue: [^ false].

	(destlist := self listForMorph: destListMorph)
		ifNil: [^ false].
	(srclist := self listForMorph: sourceListMorph)
		ifNil: [^ false].
	vmMaker
		movePlugin: transferedMorph contents
		from: srclist
		to: destlist.
	self changed: sourceListMorph getListSelector.
	self changed: destListMorph getListSelector.
	^ true! !

!VMMakerTool methodsFor: 'menus' stamp: 'rww 9/25/2001 01:02'!
perform: selector orSendTo: otherTarget
	"Selector was just chosen from a menu by a user.  If can respond, then
perform it on myself. If not, send it to otherTarget, presumably the
editPane from which the menu was invoked."

	(self respondsTo: selector)
		ifTrue: [^ self perform: selector]
		ifFalse: [^ otherTarget perform: selector]! !

!VMMakerTool methodsFor: 'path access' stamp: ''!
platformNameText
	"return a Text for the platform name"
	^vmMaker platformName asText! !

!VMMakerTool methodsFor: 'menus' stamp: 'eem 6/18/2008 14:13'!
platformNameText: aText 
	"set the platform name - this will almost certainly mean replacing the  
	vmMaker with one suited to the platform so we do it anyway."
	| prevVMMaker newVMMaker |
	prevVMMaker := vmMaker.
	"make a new vmmaker and ensure I depend on it correctly"
	newVMMaker := VMMaker forPlatform: aText string.
	newVMMaker logger: logger.
	"configure the new vmmaker to match the old one"
	[newVMMaker loadConfiguration: prevVMMaker configurationInfo.
	 newVMMaker platformDirectory]
		on: VMMakerException
		do: [self inform: 'Possible problem with path settings or platform name? Check path, permissions or spellings'.
			^false].
	newVMMaker addDependent: self.
	prevVMMaker removeDependent: self.
	vmMaker := newVMMaker.
	^true! !

!VMMakerTool methodsFor: 'menus' stamp: 'eem 1/14/2010 11:41'!
platformsListMenu
	"create a menu of all known platforms"

	| choice platnames |
	platnames := vmMaker platformRootDirectory directoryNames reject:
					[:dirName|
					dirName first = $. ".svn .git et al"
					or: [dirName ='CVS']].
	choice := (PopUpMenu labelArray: platnames lines: #()) startUp.
	choice ~= 0 ifTrue:
		[self platformNameText: (platnames at: choice) asText]! !

!VMMakerTool methodsFor: 'path access' stamp: 'eem 8/18/2008 16:22'!
platformsPathText
	"return a Text for the path to the platform sources"
	^(vmMaker platformRootDirectoryName ifNil: ['']),
	  ([vmMaker platformRootDirectory. ''] 
		on: VMMakerException 
		do:[:ex| ex return: '<path not valid>']) asText! !

!VMMakerTool methodsFor: 'path access' stamp: 'tpr 3/27/2002 15:47'!
platformsPathText: aText
	"set the path to the platform sources"
	[^vmMaker platformRootDirectoryName: aText asString] on: VMMakerException do:[:ex| self inform:'problem with this directory name; check the path settings, permissions or spelling?'. ex return: false]! !

!VMMakerTool methodsFor: 'menus' stamp: 'eem 9/16/2008 15:07'!
refreshAvailablePluginsList
	vmMaker recomputeProvidedPlugins.
	self changed: #availableModules! !

!VMMakerTool methodsFor: 'menus' stamp: 'nk 4/5/2005 20:27'!
release
	vmMaker ifNotNil: [ vmMaker removeDependent: self ].
	super release.! !

!VMMakerTool methodsFor: 'configurations' stamp: 'rww 9/23/2001 14:17'!
saveConfig

	"write info about the current configuration to a file."
	| fileResult file |
	fileResult := (StandardFileMenu newFileMenu: FileDirectory default withPattern: '*.config')
		startUpWithCaption: 'Save VMMaker configuration...'.
	fileResult ifNotNil: [
		('*.config' match: fileResult name)
			ifFalse: [fileResult name: (fileResult name, '.config')].
		file := fileResult directory fullNameFor: fileResult name.
		vmMaker saveConfigurationTo: file].
! !

!VMMakerTool methodsFor: 'generate sources' stamp: 'tpr 4/4/2005 19:30'!
set64BitVM: boolean
"do I build a 64bit VM or not?"
	boolean ifTrue:[vmMaker for64BitVM] ifFalse:[vmMaker for32BitVM].
	self changed: #sourcePathText! !

!VMMakerTool methodsFor: 'path access' stamp: 'eem 8/18/2008 16:23'!
sourcePathText
	"return a Text for the path to the generated sources"
	^(vmMaker sourceDirectoryName ifNil: ['']),
	  ([vmMaker sourceDirectory. ''] 
		on: VMMakerException 
		do:[:ex| ex return: '<path not valid>']) asText! !

!VMMakerTool methodsFor: 'path access' stamp: ''!
sourcePathText: aText
	"set the path to the generated sources"
	^vmMaker sourceDirectoryName: aText asString! !

!VMMakerTool methodsFor: 'settings' stamp: 'nk 4/5/2005 21:43'!
toggle64BitVM
	self set64BitVM: self isFor64BitVM not! !

!VMMakerTool methodsFor: 'initialisation' stamp: 'tpr 10/12/2001 18:06'!
update: anObject 
	"some related object has changed. Try to handle it"
	anObject == #reinitialize ifTrue: [self updateAllViews]! !

!VMMakerTool methodsFor: 'initialisation' stamp: 'eem 1/15/2010 18:07'!
updateAllViews
	self changed: #interpreterClassName;
		changed: #platformsPathText;
		changed: #platformsPathText;
		changed: #platformNameText;
		changed: #sourcePathText;
		changed: #availableModules;
		changed: #internalModules;
		changed: #externalModules! !

!VMMakerTool methodsFor: 'drag and drop' stamp: 'nk 4/5/2005 23:14'!
wantsDroppedMorph: transferMorph event: anEvent inMorph: destinationLM 
	"We are only interested in TransferMorphs as wrappers for             
	information. If their content is really interesting for us, will             
	determined later in >>acceptDroppingMorph:event:."

	"only want drops on the lists"

	(transferMorph isKindOf: HandleMorph) ifTrue: [ ^false ].

	(transferMorph isKindOf: TransferMorph) ifFalse: [ ^false ].

	transferMorph source model = self ifFalse:[^false].

	^self listMorphs includes: destinationLM! !

!Float class methodsFor: '*VMMaker-plugin generation' stamp: 'eem 10/2/2010 14:12'!
ccg: cg generateCoerceToOopFrom: aNode on: aStream
	"N.B.  The is used both for generation and simulation so answer the result (for interpretation)"
	^cg generateCoerceToFloatObjectFrom: aNode on: aStream! !

!Float class methodsFor: '*VMMaker-plugin generation' stamp: 'eem 10/3/2010 08:09'!
ccg: cg generateCoerceToValueFrom: aNode on: aStream
	"N.B.  The could be used both for generation and simulation so answer the result (for interpretation)"
	^cg generateCoerceToFloatValueFrom: aNode on: aStream! !

!Float class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/18/1999 17:08'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg ccgLoad: aBlock expr: aString asFloatValueFrom: anInteger! !

!Float class methodsFor: '*VMMaker-plugin generation' stamp: 'eem 12/4/2014 11:21'!
ccgCanConvertFrom: anObject

	^anObject isFloat! !

!Float class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/17/1999 01:09'!
ccgDeclareCForVar: aSymbolOrString

	^'double ', aSymbolOrString! !

!Float class methodsFor: '*VMMaker-instance creation' stamp: 'eem 5/26/2017 22:31'!
fromIEEE64BitWord: anInteger
	| value |
	value := self basicNew: 2.
	value
		basicAt: 1 put: (anInteger bitShift: -32);
		basicAt: 2 put: (anInteger bitAnd: 16rFFFFFFFF).
	^value isFinite
		ifTrue: [value * 1.0] "reduce to SmallFloat64 if possible"
		ifFalse: [value]

	"[| r |
	 r := Random new.
	 100000 timesRepeat:
		[| h l f |
		h := (r next * 16r100000000) rounded bitAnd: 16rFFFFFFFF.
		l := (r next * 16r100000000) rounded bitAnd: 16rFFFFFFFF.
		f := Float fromIEEE64BitWord: (h bitShift: 32) + l.
		self assert: h = (f basicAt: 1).
		self assert: l = (f basicAt: 2)]] timeToRun"! !

!Float methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 3/16/2017 16:36'!
asUnsignedInteger
	self assert: self >= 0.
	"C conversion from float/double to integer is by dropping the fractional part"
	^self truncated! !

!Float methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 1/8/2021 20:20'!
coerceTo: cTypeString sim: interpreter

	cTypeString = #float ifTrue:
		[^self].
	cTypeString = #double ifTrue:
		[^self].
	self halt! !

!ClassDescription methodsFor: '*VMMaker-accessing' stamp: 'eem 12/12/2016 10:32'!
>>> aSelector
	"Convenience for creating MethodReferences"
	^MethodReference class: self selector: aSelector! !

!Bitmap class methodsFor: '*VMMaker-debugger' stamp: 'eem 10/1/2015 10:47'!
defaultIntegerBaseInDebugger
	^16! !

!Bitmap methodsFor: '*VMMaker-interpreter simulator' stamp: 'ajh 8/20/2002 01:13'!
coerceTo: cTypeString sim: interpreter

	^ self! !

!Bitmap methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 10/2/2015 12:10'!
long64At: byteAddress
	| lowBits hiWord loWord midWord mask wordIndex result signBit |
	wordIndex := byteAddress - 1 // 4 + 1.
	(lowBits := byteAddress - 1 \\ 4) = 0 ifTrue:
		[hiWord := self at: wordIndex.
		 loWord := self at: wordIndex + 1.
		 ^hiWord = 0
			ifTrue: [loWord]
			ifFalse: [(hiWord signedIntFromLong bitShift: 32) + loWord]].
	"There will always be three accesses; two partial words and a full word in the middle"
	mask := 16rFFFFFFFF bitShift: 4 - lowBits * -8.
	hiWord := (self at: wordIndex) bitAnd: mask.
	midWord := self at: wordIndex + 1.
	loWord := (self at: wordIndex + 2) bitAnd: mask bitInvert32.
	result := loWord bitShift: lowBits * -8.
	midWord ~= 0 ifTrue:
		[result := result + (midWord bitShift: (4 - lowBits * 8))].
	hiWord ~= 0 ifTrue:
		[signBit := 1 << (lowBits * 8 - 1).
		 (signBit anyMask: hiWord) ifTrue:
			[hiWord := hiWord - signBit - signBit].
		 result := result + (hiWord bitShift: (4 - lowBits + 4 * 8))].
	^result! !

!Bitmap methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/1/2017 14:57'!
long64At: byteIndex put: aValue
	| lowBits mask wordIndex |
	(lowBits := byteIndex - 1 \\ 4) = 0 ifTrue:
		[self "N.B. Do the access that can fail first, before altering the receiver"
			longAt: byteIndex put: (aValue bitShift: -32);
			unsignedLongAt: byteIndex + 4 put: (aValue bitAnd: 16rffffffff).
		^aValue].
	"There will always be three accesses; two partial words and a full word in the middle"
	wordIndex := byteIndex - 1 // 4 + 1.
	aValue < 0
		ifTrue:
			[(aValue bitShift: -32) < -2147483648 ifTrue:
				[^self errorImproperStore]]
		ifFalse:
			[16r7FFFFFFF < (aValue bitShift: -32) ifTrue:
				[^self errorImproperStore]].
	mask := 16rFFFFFFFF bitShift: lowBits * -8.
	self at: wordIndex + 2 put: (((self at: wordIndex + 2) bitAnd: mask) bitXor: ((aValue bitShift: lowBits * 8) bitAnd: mask bitInvert32)).
	self at: wordIndex + 1 put: ((aValue bitShift: lowBits * -8) bitAnd: 16rFFFFFFFF).
	self at: wordIndex put: (((self at: wordIndex) bitAnd: mask bitInvert32) bitXor: ((aValue bitShift: 4 - lowBits + 4 * -8) bitAnd: mask)).
	"(wordIndex to: wordIndex + 2) collect: [:i| (self at: i) hex]"
	^aValue! !

!Bitmap methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 10/1/2015 11:50'!
longAt: byteIndex
	"Default bigEndian access"
	| lowBits wordIndex value word0 word1 |
	lowBits := byteIndex - 1 bitAnd: 3.
	wordIndex := byteIndex - 1 // 4 + 1.
	lowBits = 0
		ifTrue:
			[value := self at: wordIndex]
		ifFalse:
			[word0 := self at: wordIndex.
			 word1 := self at: wordIndex + 1.
			 value := 16rFFFFFFFF bitAnd: (word0 bitShift: (lowBits * 8)) + (word1 bitShift: 0 - (4 - lowBits * 8))].
	(16r80000000 bitAnd: value) ~= 0 ifTrue:
		[value := (16r7FFFFFFF bitAnd: value) - 16r80000000].
	^value! !

!Bitmap methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 10/1/2015 11:58'!
longAt: byteIndex bigEndian: bigEndian
	"Compatibility with the ByteArray method of the same name."
	| lowBits wordIndex value word0 word1 |
	lowBits := byteIndex - 1 bitAnd: 3.
	wordIndex := byteIndex - 1 // 4 + 1.
	lowBits = 0
		ifTrue:
			[value := self at: wordIndex]
		ifFalse:
			[word0 := self at: wordIndex.
			 word1 := self at: wordIndex + 1.
			 value := 16rFFFFFFFF bitAnd: (word0 bitShift: (lowBits * 8)) + (word1 bitShift: 4 - lowBits * -8)].
	bigEndian ifFalse:
		[value := ((value bitShift: -24) bitAnd: 16rFF)
				 + ((value bitShift: -8) bitAnd: 16rFF00)
 				 + ((value bitAnd: 16rFF00) bitShift: 8)
				 + ((value bitAnd: 16rFF) bitShift: 24)].
	(16r80000000 bitAnd: value) ~= 0 ifTrue:
		[value := (16r7FFFFFFF bitAnd: value) - 16r80000000].
	^value

	"| bm ba |
	(bm := Bitmap new: 4)
		at: 1 put: 16r01234567;
		at: 2 put: 16r89ABCDEF;
		at: 3 put: 16r89ABCDEF;
		at: 4 put: 16r01234567.
	ba := bm asByteArray.
	(((1 to: 5), (9 to: 13) select:
		[:bi| (bm longAt: bi bigEndian: true) ~= (ba longAt: bi bigEndian: true)]) collect:
			[:i| { i. (bm longAt: i bigEndian: true) hex. (ba longAt: i bigEndian: true) hex}]),
	(((1 to: 5), (9 to: 13) select:
		[:bi| (bm longAt: bi bigEndian: false) ~= (ba longAt: bi bigEndian: false)]) collect:
			[:i| { i. (bm longAt: i bigEndian: false) hex. (ba longAt: i bigEndian: false) hex}])"! !

!Bitmap methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 10/1/2015 11:58'!
longAt: byteIndex put: aValue
	"Default bigEndian access"
	| lowBits wordIndex value mask |
	wordIndex := byteIndex - 1 // 4 + 1.
	aValue < 0
		ifTrue:
			[aValue < -2147483648 ifTrue:
				[^self errorImproperStore].
			 value := 16rFFFFFFFF bitAnd: aValue]
		ifFalse:
			[16r7FFFFFFF < aValue ifTrue:
				[^self errorImproperStore].
			value := aValue].
	(lowBits := byteIndex - 1 bitAnd: 3) = 0 ifTrue:
		[self at: wordIndex put: value.
		 ^aValue].
	mask := 16rFFFFFFFF bitShift: lowBits * -8.
	self at: wordIndex put: (((self at: wordIndex) bitAnd: mask bitInvert) bitXor: ((value bitShift: lowBits * -8) bitAnd: mask)).
	self at: wordIndex + 1 put: (((self at: wordIndex + 1) bitAnd: mask) bitXor: (16rFFFFFFFF bitAnd: ((value bitShift: (4 - lowBits * 8)) bitAnd: mask bitInvert))).
	^aValue! !

!Bitmap methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 10/1/2015 11:57'!
longAt: byteIndex put: aValue bigEndian: bigEndian
	"Compatibility with the ByteArray method of the same name."
	| lowBits wordIndex value mask |
	lowBits := byteIndex - 1 bitAnd: 3.
	wordIndex := byteIndex - 1 // 4 + 1.
	aValue < 0
		ifTrue:
			[value := 16rFFFFFFFF bitAnd: aValue.
			 value = 0 ifTrue:
				[self errorImproperStore]]
		ifFalse:
			[16rFFFFFFFF < aValue ifTrue:
				[self errorImproperStore].
			value := aValue].
	bigEndian ifFalse:
		[value := ((value bitShift: -24) bitAnd: 16rFF)
				 + ((value bitShift: -8) bitAnd: 16rFF00)
 				 + ((value bitAnd: 16rFF00) bitShift: 8)
				 + ((value bitAnd: 16rFF) bitShift: 24)].
	lowBits = 0 ifTrue:
		[self at: wordIndex put: value.
		 ^aValue].
	mask := 16rFFFFFFFF bitShift: lowBits * -8.
	self at: wordIndex put: (((self at: wordIndex) bitAnd: mask bitInvert) bitXor: ((value bitShift: lowBits * -8) bitAnd: mask)).
	self at: wordIndex + 1 put: (((self at: wordIndex + 1) bitAnd: mask) bitXor: (16rFFFFFFFF bitAnd: ((value bitShift: 4 - lowBits * 8) bitAnd: mask bitInvert))).
	^aValue

	"(1 to: 8) collect:
		[:ba| | bm |
		bm := Bitmap new: 4.
		bm at: 1 put: 16r55555555.
		bm at: 2 put: 16rAAAAAAAA.
		bm longAt: ba put: 16r04030201 bigEndian: true.
		{ (bm at: 1) hex. (bm at: 2) hex }]"

	"(1 to: 8) collect:
		[:ba| | bm |
		bm := Bitmap new: 4.
		bm at: 1 put: 16r55555555.
		bm at: 2 put: 16rAAAAAAAA.
		bm longAt: ba put: 16r01020304 bigEndian: false.
		{ (bm at: 1) hex. (bm at: 2) hex }]"! !

!Bitmap methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/28/2008 18:45'!
unsignedByteAt: byteAddress
	"Compatibility with the ByteArray and Alien methods of the same name."
	| lowBits |
	lowBits := byteAddress - 1 bitAnd: 3.
	^((self at: byteAddress - 1 - lowBits // 4 + 1)
		bitShift: (lowBits - 3) * 8)
		bitAnd: 16rFF! !

!Bitmap methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/28/2008 18:46'!
unsignedByteAt: byteAddress put: byte
	"Compatibility with the ByteArray and Alien methods of the same name."
	| longWord shift lowBits longAddr |
	(byte < 0 or:[byte > 255]) ifTrue:[^self errorImproperStore].
	lowBits := byteAddress - 1 bitAnd: 3.
	longWord := self at: (longAddr := (byteAddress - 1 - lowBits) // 4 + 1).
	shift := (3 - lowBits) * 8.
	longWord := longWord - (longWord bitAnd: (16rFF bitShift: shift)) 
		+ (byte bitShift: shift).
	self at: longAddr put: longWord.
	^ byte! !

!Bitmap methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 10/2/2015 12:15'!
unsignedLong64At: byteAddress
	"memory is a Bitmap, a 32-bit indexable array of bits"
	| lowBits hiWord loWord midWord mask wordIndex result |
	wordIndex := byteAddress - 1 // 4 + 1.
	(lowBits := byteAddress - 1 \\ 4) = 0 ifTrue:
		[hiWord := self at: wordIndex.
		 loWord := self at: wordIndex + 1.
		 ^hiWord = 0
			ifTrue: [loWord]
			ifFalse: [(hiWord bitShift: 32) + loWord]].
	"There will always be three accesses; two partial words and a full word in the middle"
	mask := 16rFFFFFFFF bitShift: (4 - lowBits) * -8.
	hiWord := (self at: wordIndex) bitAnd: mask.
	midWord := self at: wordIndex + 1.
	loWord := (self at: wordIndex + 2) bitAnd: mask bitInvert32.
	result := loWord bitShift: lowBits * -8.
	midWord ~= 0 ifTrue:
		[result := result + (midWord bitShift: (4 - lowBits * 8))].
	hiWord ~= 0 ifTrue:
		[result := result + (hiWord bitShift: (4 - lowBits + 4 * 8))].
	^result! !

!Bitmap methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 10/2/2015 12:13'!
unsignedLong64At: byteIndex put: aValue
	| lowBits mask wordIndex |
	(lowBits := byteIndex - 1 \\ 4) = 0 ifTrue:
		[self "N.B. Do the access that can fail first, before altering the receiver"
			unsignedLongAt: byteIndex put: (aValue bitShift: -32);
			unsignedLongAt: byteIndex + 4 put: (aValue bitAnd: 16rffffffff).
		^aValue].
	"There will always be three accesses; two partial words and a full word in the middle"
	wordIndex := byteIndex - 1 // 4 + 1.
	(aValue bitShift: -64) ~= 0 ifTrue:
		[^self errorImproperStore].
	mask := 16rFFFFFFFF bitShift: 4 - lowBits * -8.
	self at: wordIndex + 2 put: (((self at: wordIndex + 2) bitAnd: mask) bitXor: ((aValue bitShift: lowBits * 8) bitAnd: mask bitInvert32)).
	self at: wordIndex + 1 put: ((aValue bitShift: lowBits * -8) bitAnd: 16rFFFFFFFF).
	self at: wordIndex put: (((self at: wordIndex) bitAnd: mask bitInvert32) bitXor: ((aValue bitShift: 4 - lowBits + 4 * -8) bitAnd: mask)).
	"(wordIndex to: wordIndex + 2) collect: [:i| (self at: i) hex]"
	^aValue! !

!Bitmap methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/28/2008 18:42'!
unsignedLongAt: byteIndex
	"Compatiblity with the ByteArray & Alien methods of the same name."
	^(byteIndex - 1 bitAnd: 3) = 0
		ifTrue: [self at: byteIndex - 1 // 4 + 1]
		ifFalse: [self unsignedLongAt: byteIndex bigEndian: true]! !

!Bitmap methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 10/1/2015 11:56'!
unsignedLongAt: byteIndex bigEndian: bigEndian
	"Compatiblity with the ByteArray method of the same name."
	| lowBits wordIndex value word0 word1 |
	lowBits := byteIndex - 1 bitAnd: 3.
	wordIndex := byteIndex - 1 // 4 + 1.
	lowBits = 0
		ifTrue:
			[value := self at: wordIndex]
		ifFalse:
			[word0 := self at: wordIndex.
			 word1 := self at: wordIndex + 1.
			 value := 16rFFFFFFFF bitAnd: (word0 bitShift: lowBits * 8) + (word1 bitShift: 4 - lowBits * -8)].
	bigEndian ifFalse:
		[value := ((value bitShift: -24) bitAnd: 16rFF)
				 + ((value bitShift: -8) bitAnd: 16rFF00)
 				 + ((value bitAnd: 16rFF00) bitShift: 8)
				 + ((value bitAnd: 16rFF) bitShift: 24)].
	^value

	"| bm ba |
	(bm := Bitmap new: 4)
		at: 1 put: 16r01234567;
		at: 2 put: 16r89ABCDEF;
		at: 3 put: 16r89ABCDEF;
		at: 4 put: 16r01234567.
	ba := bm asByteArray.
	(((1 to: 5), (9 to: 13) select:
		[:bi| (bm unsignedLongAt: bi bigEndian: true) ~= (ba unsignedLongAt: bi bigEndian: true)]) collect:
			[:i| { i. (bm unsignedLongAt: i bigEndian: true) hex. (ba unsignedLongAt: i bigEndian: true) hex}]),
	(((1 to: 5), (9 to: 13) select:
		[:bi| (bm unsignedLongAt: bi bigEndian: false) ~= (ba unsignedLongAt: bi bigEndian: false)]) collect:
			[:i| { i. (bm unsignedLongAt: i bigEndian: false) hex. (ba unsignedLongAt: i bigEndian: false) hex}])"! !

!Bitmap methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/1/2017 15:13'!
unsignedLongAt: byteIndex put: aValue
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| lowBits mask wordIndex |
	wordIndex := byteIndex - 1 // 4 + 1.
	(lowBits := byteIndex - 1 bitAnd: 3) = 0 ifTrue:
		[^self at: wordIndex put: aValue].
	mask := 16rFFFFFFFF bitShift: lowBits * -8.
	(aValue bitShift: -32) = 0 ifFalse:
		[^self errorImproperStore].
	self at: wordIndex put: (((self at: wordIndex) bitAnd: mask bitInvert) bitXor: ((aValue bitShift: lowBits * -8) bitAnd: mask)).
	self at: wordIndex + 1 put: (((self at: wordIndex + 1) bitAnd: mask) bitXor: (16rFFFFFFFF bitAnd: ((aValue bitShift: (4 - lowBits * 8)) bitAnd: mask bitInvert))).
	^aValue! !

Bitmap variableWordSubclass: #LittleEndianBitmap
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!LittleEndianBitmap methodsFor: 'converting' stamp: 'eem 12/1/2017 16:21'!
asByteArray
	"Faster way to make a byte array from me.
	copyFromByteArray: makes equal Bitmap."
	| f bytes hack |
	Smalltalk isLittleEndian ifTrue:
		[bytes := self shallowCopy.
		 ByteArray adoptInstance: bytes.
		 ^bytes].
	f := Form extent: 4@self size depth: 8 bits: self.
	bytes := ByteArray new: self size * 4.
	hack := Form new hackBits: bytes.
	hack swapEndianness.
	hack copyBits: f boundingBox
		from: f
		at: 0@0
		clippingBox: hack boundingBox
		rule: Form over
		fillColor: nil
		map: nil.
	"f displayOn: hack."
	^bytes
	"| v | v := LittleEndianBitmap withAll: (1 to: 4).
	 [v asByteArray] bench '19,700,000 per second. 50.8 nanoseconds per run.'"
	"| v | v := Bitmap withAll: (1 to: 4).
	[v asByteArray] bench '977,000 per second. 1.02 microseconds per run.'"! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 1/15/2009 22:17'!
byteAt: byteAddress
	"Extract a byte from a Bitmap (little-endian version)"
	| lowBits |
	lowBits := byteAddress - 1 bitAnd: 3.
	^((self at: byteAddress - 1 - lowBits // 4 + 1)
		bitShift: lowBits * -8)
		bitAnd: 16rFF! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 10/9/2013 11:13'!
byteAt: byteAddress put: byte
	"Insert a byte into a Bitmap (little-endian version).  N.B. Signedness will be lost since byteAt: answers non-negative values.
	 We can add a signedByteAt: if needed."
	| value longWord shift lowBits longAddr |
	(byte < 0 or:[byte > 255]) ifTrue:[^self errorImproperStore].
	value := byte < 0
				ifTrue: [byte < 128 ifTrue:
							[self errorImproperStore].
						16rFF bitAnd: byte]
				ifFalse: [16rFF < byte ifTrue:
							[self errorImproperStore].
						byte].
	lowBits := byteAddress - 1 bitAnd: 3.
	longWord := self at: (longAddr := (byteAddress - 1 - lowBits) // 4 + 1).
	shift := lowBits * 8.
	longWord := longWord
				- (longWord bitAnd: (16rFF bitShift: shift)) 
				+ (value bitShift: shift).
	self at: longAddr put: longWord.
	^byte! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 10/2/2015 11:41'!
long64At: byteAddress
	| lowBits hiWord loWord midWord mask wordIndex result signBit |
	wordIndex := byteAddress - 1 // 4 + 1.
	(lowBits := byteAddress - 1 \\ 4) = 0 ifTrue:
		[loWord := self at: wordIndex.
		 hiWord := self at: wordIndex + 1.
		 ^hiWord = 0
			ifTrue: [loWord]
			ifFalse: [(hiWord signedIntFromLong bitShift: 32) + loWord]].
	"There will always be three accesses; two partial words and a full word in the middle"
	mask := 16rFFFFFFFF bitShift: 4 - lowBits * -8.
	loWord := (self at: wordIndex) bitAnd: mask bitInvert32.
	midWord := self at: wordIndex + 1.
	hiWord := (self at: wordIndex + 2) bitAnd: mask.
	result := loWord bitShift: lowBits * -8.
	midWord ~= 0 ifTrue:
		[result := result + (midWord bitShift: (4 - lowBits * 8))].
	hiWord ~= 0 ifTrue:
		[signBit := 1 << (lowBits * 8 - 1).
		 (signBit anyMask: hiWord) ifTrue:
			[hiWord := hiWord - signBit - signBit].
		 result := result + (hiWord bitShift: (4 - lowBits + 4 * 8))].
	^result! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 10/1/2015 12:42'!
long64At: byteIndex put: aValue
	| lowBits mask wordIndex |
	(lowBits := byteIndex - 1 \\ 4) = 0 ifTrue:
		[self "N.B. Do the access that can fail first, before altering the receiver"
			longAt: byteIndex + 4 put: (aValue bitShift: -32);
			unsignedLongAt: byteIndex put: (aValue bitAnd: 16rffffffff).
		^aValue].
	"There will always be three accesses; two partial words and a full word in the middle"
	wordIndex := byteIndex - 1 // 4 + 1.
	aValue < 0
		ifTrue:
			[(aValue bitShift: -32) < -2147483648 ifTrue:
				[^self errorImproperStore]]
		ifFalse:
			[16r7FFFFFFF < (aValue bitShift: -32) ifTrue:
				[^self errorImproperStore]].
	mask := 16rFFFFFFFF bitShift: 4 - lowBits * -8.
	self at: wordIndex put: (((self at: wordIndex) bitAnd: mask) bitXor: ((aValue bitShift: lowBits * 8) bitAnd: mask bitInvert32)).
	self at: wordIndex + 1 put: ((aValue bitShift: 4 - lowBits * -8) bitAnd: 16rFFFFFFFF).
	self at: wordIndex + 2 put: (((self at: wordIndex + 2) bitAnd: mask bitInvert32) bitXor: ((aValue bitShift: 4 - lowBits + 4 * -8) bitAnd: mask)).
	^aValue! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 10/1/2013 13:20'!
longAt: byteIndex
	^self longAt: byteIndex bigEndian: false! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 10/1/2015 12:14'!
longAt: byteIndex bigEndian: bigEndian
	"Compatibility with the ByteArray method of the same name."
	| lowBits wordIndex value word0 word1 |
	lowBits := byteIndex - 1 bitAnd: 3.
	wordIndex := byteIndex - 1 // 4 + 1.
	lowBits = 0
		ifTrue:
			[value := self at: wordIndex]
		ifFalse:
			[word0 := self at: wordIndex.
			 word1 := self at: wordIndex + 1.
			 value := 16rFFFFFFFF bitAnd: (word0 bitShift: lowBits * -8) + (word1 bitShift: 4 - lowBits * 8)].
	bigEndian ifTrue:
		[value := ((value bitShift: -24) bitAnd: 16rFF)
				 + ((value bitShift: -8) bitAnd: 16rFF00)
 				 + ((value bitAnd: 16rFF00) bitShift: 8)
				 + ((value bitAnd: 16rFF) bitShift: 24)].
	(16r80000000 bitAnd: value) ~= 0 ifTrue:
		[value := (16r7FFFFFFF bitAnd: value) - 16r80000000].
	^value

	"| bm ba |
	(bm := LittleEndianBitmap new: 4)
		at: 1 put: 16r01234567;
		at: 2 put: 16r89ABCDEF;
		at: 3 put: 16r89ABCDEF;
		at: 4 put: 16r01234567.
	ba := bm asByteArray.
	(((1 to: 5), (9 to: 13) select:
		[:bi| (bm longAt: bi bigEndian: true) ~= (ba longAt: bi bigEndian: true)]) collect:
			[:i| { i. (bm longAt: i bigEndian: true) hex. (ba longAt: i bigEndian: true) hex}]),
	(((1 to: 5), (9 to: 13) select:
		[:bi| (bm longAt: bi bigEndian: false) ~= (ba longAt: bi bigEndian: false)]) collect:
			[:i| { i. (bm longAt: i bigEndian: false) hex. (ba longAt: i bigEndian: false) hex}])"! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 12/1/2017 14:53'!
longAt: byteIndex put: aValue
	"Compatibility with the ByteArray method of the same name."
	| lowBits wordIndex mask |
	lowBits := byteIndex - 1 bitAnd: 3.
	wordIndex := byteIndex - 1 // 4 + 1.
	aValue < 0
		ifTrue:
			[aValue < -2147483648 ifTrue:
				[^self errorImproperStore]]
		ifFalse:
			[16r7FFFFFFF < aValue ifTrue:
				[^self errorImproperStore]].
	lowBits = 0 ifTrue:
		[self at: wordIndex put: (16rFFFFFFFF bitAnd: aValue).
		 ^aValue].
	mask := 16rFFFFFFFF bitShift: 4 - lowBits * -8.
	self at: wordIndex put: (((self at: wordIndex) bitAnd: mask) bitXor: ((aValue bitShift: lowBits * 8) bitAnd: mask bitInvert32)).
	self at: wordIndex + 1 put: (((self at: wordIndex + 1) bitAnd: mask bitInvert32) bitXor: ((aValue bitShift: 4 - lowBits * -8) bitAnd: mask)).
	^aValue

	"(1 to: 8) collect:
		[:ba| | bm |
		bm := LittleEndianBitmap new: 4.
		bm at: 1 put: 16r55555555.
		bm at: 2 put: 16rAAAAAAAA.
		bm longAt: ba put: 16r04030201.
		{ (bm at: 1) hex. (bm at: 2) hex }]"! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 10/1/2015 12:23'!
longAt: byteIndex put: aValue bigEndian: bigEndian
	"Compatibility with the ByteArray method of the same name."
	| lowBits wordIndex value mask |
	lowBits := byteIndex - 1 bitAnd: 3.
	wordIndex := byteIndex - 1 // 4 + 1.
	value := aValue < 0
				ifTrue: [16rFFFFFFFF bitAnd: aValue]
				ifFalse: [16rFFFFFFFF < aValue ifTrue:
							[self errorImproperStore].
						aValue].
	bigEndian ifTrue:
		[value := ((value bitShift: -24) bitAnd: 16rFF)
				 + ((value bitShift: -8) bitAnd: 16rFF00)
 				 + ((value bitAnd: 16rFF00) bitShift: 8)
				 + ((value bitAnd: 16rFF) bitShift: 24)].
	lowBits = 0 ifTrue:
		[self at: wordIndex put: value.
		 ^aValue].
	mask := 16rFFFFFFFF bitShift: lowBits * -8.
	self at: wordIndex put: (((self at: wordIndex) bitAnd: mask) bitXor: ((value bitShift: lowBits * 8) bitAnd: mask bitInvert32)).
	self at: wordIndex + 1 put: (((self at: wordIndex + 1) bitAnd: mask bitInvert32) bitXor: ((value bitShift: 4 - lowBits * -8) bitAnd: mask)).
	^aValue

	"(1 to: 8) collect:
		[:ba| | bm |
		bm := LittleEndianBitmap new: 4.
		bm at: 1 put: 16r55555555.
		bm at: 2 put: 16rAAAAAAAA.
		bm longAt: ba put: 16r04030201 bigEndian: false.
		{ (bm at: 1) hex. (bm at: 2) hex }]"

	"(1 to: 8) collect:
		[:ba| | bm |
		bm := LittleEndianBitmap new: 4.
		bm at: 1 put: 16r55555555.
		bm at: 2 put: 16rAAAAAAAA.
		bm longAt: ba put: 16r01020304 bigEndian: true.
		{ (bm at: 1) hex. (bm at: 2) hex }]"! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 1/15/2009 22:16'!
unsignedByteAt: byteAddress
	"Extract a byte from a Bitmap (little-endian version)"
	| lowBits |
	lowBits := byteAddress - 1 bitAnd: 3.
	^((self at: byteAddress - 1 - lowBits // 4 + 1)
		bitShift: lowBits * -8)
		bitAnd: 16rFF! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 12/28/2008 18:44'!
unsignedByteAt: byteAddress put: byte
	"Insert a byte into a Bitmap (little-endian version)"
	| longWord shift lowBits longAddr |
	(byte < 0 or:[byte > 255]) ifTrue:[^self errorImproperStore].
	lowBits := byteAddress - 1 bitAnd: 3.
	longWord := self at: (longAddr := (byteAddress - 1 - lowBits) // 4 + 1).
	shift := lowBits * 8.
	longWord := longWord
				- (longWord bitAnd: (16rFF bitShift: shift)) 
				+ (byte bitShift: shift).
	self at: longAddr put: longWord.
	^byte! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 10/2/2015 11:35'!
unsignedLong64At: byteAddress
	"memory is a Bitmap, a 32-bit indexable array of bits"
	| lowBits hiWord loWord midWord mask wordIndex result |
	wordIndex := byteAddress - 1 // 4 + 1.
	(lowBits := byteAddress - 1 \\ 4) = 0 ifTrue:
		[loWord := self at: wordIndex.
		 hiWord := self at: wordIndex + 1.
		 ^hiWord = 0
			ifTrue: [loWord]
			ifFalse: [(hiWord bitShift: 32) + loWord]].
	"There will always be three accesses; two partial words and a full word in the middle"
	mask := 16rFFFFFFFF bitShift: 4 - lowBits * -8.
	loWord := (self at: wordIndex) bitAnd: mask bitInvert32.
	midWord := self at: wordIndex + 1.
	hiWord := (self at: wordIndex + 2) bitAnd: mask.
	result := loWord bitShift: lowBits * -8.
	midWord ~= 0 ifTrue:
		[result := result + (midWord bitShift: (4 - lowBits * 8))].
	hiWord ~= 0 ifTrue:
		[result := result + (hiWord bitShift: (4 - lowBits + 4 * 8))].
	^result! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 10/1/2015 12:43'!
unsignedLong64At: byteIndex put: aValue
	| lowBits mask wordIndex |
	(lowBits := byteIndex - 1 \\ 4) = 0 ifTrue:
		[self "N.B. Do the access that can fail first, before altering the receiver"
			unsignedLongAt: byteIndex + 4 put: (aValue bitShift: -32);
			unsignedLongAt: byteIndex put: (aValue bitAnd: 16rffffffff).
		^aValue].
	"There will always be three accesses; two partial words and a full word in the middle"
	wordIndex := byteIndex - 1 // 4 + 1.
	(aValue bitShift: -64) ~= 0 ifTrue:
		[^self errorImproperStore].
	mask := 16rFFFFFFFF bitShift: 4 - lowBits * -8.
	self at: wordIndex put: (((self at: wordIndex) bitAnd: mask) bitXor: ((aValue bitShift: lowBits * 8) bitAnd: mask bitInvert32)).
	self at: wordIndex + 1 put: ((aValue bitShift: 4 - lowBits * -8) bitAnd: 16rFFFFFFFF).
	self at: wordIndex + 2 put: (((self at: wordIndex + 2) bitAnd: mask bitInvert32) bitXor: ((aValue bitShift: 4 - lowBits + 4 * -8) bitAnd: mask)).
	^aValue! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 12/28/2008 18:41'!
unsignedLongAt: byteIndex
	"Compatiblity with the ByteArray & Alien methods of the same name."
	^(byteIndex - 1 bitAnd: 3) = 0
		ifTrue: [self at: byteIndex - 1 // 4 + 1]
		ifFalse: [self unsignedLongAt: byteIndex bigEndian: false]! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 10/1/2015 12:27'!
unsignedLongAt: byteIndex bigEndian: bigEndian
	"Compatiblity with the ByteArray method of the same name."
	| lowBits wordIndex value word0 word1 |
	lowBits := byteIndex - 1 bitAnd: 3.
	wordIndex := byteIndex - 1 // 4 + 1.
	lowBits = 0
		ifTrue:
			[value := self at: wordIndex]
		ifFalse:
			[word0 := self at: wordIndex.
			 word1 := self at: wordIndex + 1.
			 value := 16rFFFFFFFF bitAnd: (word0 bitShift: lowBits * -8) + (word1 bitShift: 4 - lowBits * 8)].
	bigEndian ifTrue:
		[value := ((value bitShift: -24) bitAnd: 16rFF)
				 + ((value bitShift: -8) bitAnd: 16rFF00)
 				 + ((value bitAnd: 16rFF00) bitShift: 8)
				 + ((value bitAnd: 16rFF) bitShift: 24)].
	^value

	"| bm ba |
	(bm := LittleEndianBitmap new: 4)
		at: 1 put: 16r01234567;
		at: 2 put: 16r89ABCDEF;
		at: 3 put: 16r89ABCDEF;
		at: 4 put: 16r01234567.
	ba := bm asByteArray.
	(((1 to: 5), (9 to: 13) select:
		[:bi| (bm unsignedLongAt: bi bigEndian: true) ~= (ba unsignedLongAt: bi bigEndian: true)]) collect:
			[:i| { i. (bm unsignedLongAt: i bigEndian: true) hex. (ba unsignedLongAt: i bigEndian: true) hex}]),
	(((1 to: 5), (9 to: 13) select:
		[:bi| (bm unsignedLongAt: bi bigEndian: false) ~= (ba unsignedLongAt: bi bigEndian: false)]) collect:
			[:i| { i. (bm unsignedLongAt: i bigEndian: false) hex. (ba unsignedLongAt: i bigEndian: false) hex}])"! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 12/1/2017 15:57'!
unsignedLongAt: byteIndex put: aValue
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| lowBits mask wordIndex |
	wordIndex := byteIndex - 1 // 4 + 1.
	(lowBits := byteIndex - 1 bitAnd: 3) = 0 ifTrue:
		[^self at: wordIndex put: aValue].
	(aValue bitShift: -32) = 0 ifFalse:
		[^self errorImproperStore].
	mask := 16rFFFFFFFF bitShift: 4 - lowBits * -8.
	self at: wordIndex put: (((self at: wordIndex) bitAnd: mask) bitXor: ((aValue bitShift: lowBits * 8) bitAnd: mask bitInvert32)).
	self at: wordIndex + 1 put: (((self at: wordIndex + 1) bitAnd: mask bitInvert32) bitXor: (16rFFFFFFFF bitAnd: ((aValue bitShift: 4 - lowBits * -8) bitAnd: mask))).
	^aValue! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 1/30/2009 09:58'!
unsignedShortAt: byteIndex
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| zi |
	^((zi := byteIndex - 1) bitAnd: 1) = 0
		ifTrue: [(zi bitAnd: 2) = 0
					ifTrue: [(self at: zi // 4 + 1) bitAnd: 16rFFFF]
					ifFalse: [(self at: zi - 2 // 4 + 1) bitShift: -16]]
		ifFalse: [(self unsignedLongAt: byteIndex bigEndian: false) bitAnd: 16rFFFF]! !

!LittleEndianBitmap methodsFor: 'accessing' stamp: 'eem 1/30/2009 10:08'!
unsignedShortAt: byteIndex put: aValue
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| zi |
	(aValue < 0 or: [aValue > 16rFFFF])  ifTrue:
		[^self errorImproperStore].
	((zi := byteIndex - 1) bitAnd: 1) = 0
		ifTrue: [(zi bitAnd: 2) = 0
					ifTrue: [self at: (zi := zi // 4 + 1) put: (16rFFFF0000 bitAnd: (self at: zi)) + aValue]
					ifFalse: [self at: (zi := zi - 2 // 4 + 1) put: (16rFFFF bitAnd: (self at: zi)) + (aValue bitShift: 16)]]
		ifFalse: [self notYetImplemented].
	^aValue! !

!IntegerArray class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/25/1999 15:00'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg 
		ccgLoad: aBlock 
		expr: aString 
		asIntPtrFrom: anInteger
		andThen: (cg ccgValBlock: 'isWords')! !

!IntegerArray class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/17/1999 01:18'!
ccgDeclareCForVar: aSymbolOrString

	^'int *', aSymbolOrString! !

!FullBlockClosure methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 12/16/2019 23:19'!
shouldStopIfAtPC: address
	<primitive: 207>
	^self value: address! !

RelativeInstructionPrinter subclass: #RelativeDetailedInstructionPrinter
	instanceVariableNames: 'stackHeightComputer saveLabel'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Support'!

!RelativeDetailedInstructionPrinter class methodsFor: 'instance creation' stamp: 'eem 11/26/2010 12:32'!
on: method 
	"Answer an instance of me on the argument, method."

	^self new method: method pc: method initialPC! !

!RelativeDetailedInstructionPrinter methodsFor: 'printing' stamp: 'eem 11/15/2012 20:40'!
emitLabelIfAnyFor: pc on: aStream
	saveLabel := String streamContents: [:s| super emitLabelIfAnyFor: pc on: s]! !

!RelativeDetailedInstructionPrinter methodsFor: 'initialize-release' stamp: 'eem 11/26/2010 12:30'!
method: aMethod
	super method: aMethod.
	stackHeightComputer := StackDepthFinder new method: aMethod pc: nil! !

!RelativeDetailedInstructionPrinter methodsFor: 'initialize-release' stamp: 'eem 11/26/2010 12:32'!
method: aMethod pc: pc
	super method: aMethod.
	stackHeightComputer := StackDepthFinder new method: aMethod pc: pc! !

!RelativeDetailedInstructionPrinter methodsFor: 'printing' stamp: 'eem 11/15/2012 20:37'!
print: instruction
	| current next |
	super print: instruction.
	labelling ifFalse:
		[current := stackHeightComputer stackp.
		 stackHeightComputer interpretNextInstructionFor: stackHeightComputer.
		 next := stackHeightComputer stackp.
		 stream
			skip: -1;
			tab; tab; nextPut: $(;
			print: current.
		 current ~= next ifTrue:
			[stream nextPut: $-; nextPut: $>; print: next].
		 stream
			nextPut: $);
			cr;
			flush.
		 saveLabel isEmpty ifFalse:
			[stream nextPutAll: saveLabel; flush]]! !

!MethodNode methodsFor: '*VMMaker-C translation' stamp: 'eem 8/15/2018 12:49'!
asTranslationMethodOfClass: aClass
 	"Answer a TMethod (or subclass) derived from the receiver."
	^aClass new
		setSelector: selectorOrFalse
		definingClass: encoder associationForClass value
		args: arguments
		locals: encoder tempsAndBlockArgs
		block: block
		primitive: primitive
		properties: properties
		comment: comment! !

!MethodNode methodsFor: '*VMMaker-C translation' stamp: 'eem 9/9/2018 09:57'!
quickMethodReturnLiteral
	^self block statements last expr key! !

!ArrayedCollection class methodsFor: '*VMMaker-plugin generation' stamp: 'eem 10/2/2010 14:12'!
ccg: cg generateCoerceToOopFrom: aNode on: aStream
	"N.B.  The is used both for generation and simulation so answer the result (for interpretation)"
	self instSize > 0 ifTrue: 
		[self error: 'cannot auto-coerce arrays with named instance variables'].
	^cg generateCoerceToObjectFromPtr: aNode on: aStream! !

!ArrayedCollection class methodsFor: '*VMMaker-plugin generation' stamp: 'eem 10/3/2010 08:09'!
ccg: cg generateCoerceToValueFrom: aNode on: aStream
	"N.B.  The could be used both for generation and simulation so answer the result (for interpretation)"
	^cg 
		generateCoerceToPtr: (self ccgDeclareCForVar: '')
		fromObject: aNode on: aStream! !

!ArrayedCollection methodsFor: '*VMMaker-simulated image growing' stamp: 'nk 10/13/2003 22:26'!
copyGrownBy: n
	"Create a new collection containing all the elements from aCollection."

	^ (self class new: self size + n) replaceFrom: 1 to: self size with: self! !

!WordArray class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/20/1999 13:16'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg 
		ccgLoad: aBlock 
		expr: aString 
		asUnsignedPtrFrom: anInteger
		andThen: (cg ccgValBlock: 'isWords')! !

!WordArray class methodsFor: '*VMMaker-plugin generation' stamp: 'dtl 3/8/2013 20:36'!
ccgDeclareCForVar: aSymbolOrString
	"Address of an unsigned 32 bit value, regardless of Smalltalk wordSize"

	^'unsigned *', aSymbolOrString! !

!WordArray class methodsFor: '*VMMaker-inspecting' stamp: 'eem 6/30/2017 12:51'!
defaultIntegerBaseInDebugger
	^16! !

!WordArray methodsFor: '*VMMaker-JITsimulation' stamp: 'nice 12/30/2020 23:20'!
unsignedLongAt: byteIndex bigEndian: bigEndian
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| wordIndex lowBits word hiWord |
	wordIndex := byteIndex - 1 // 4 + 1.
	lowBits := byteIndex - 1 bitAnd: 3.
	word := self at: wordIndex.
	lowBits > 0 ifTrue: "access straddles two words"
		[hiWord := self at: wordIndex + 1.
		 word := (word bitShift: lowBits * -8) + (hiWord bitShift: 4 - lowBits * 8)].
	word := word bitAnd: 16rFFFFFFFF.
	bigEndian
		ifTrue:
			[word := ((word bitShift: -24) bitAnd: 16rFF)
					 + ((word bitShift: -8) bitAnd: 16rFF00)
	 				 + ((word bitAnd: 16rFF00) bitShift: 8)
					 + ((word bitAnd: 16rFF) bitShift: 24)].
	^word! !

!LiteralNode methodsFor: '*VMMaker-C translation' stamp: 'eem 8/15/2018 18:42'!
asTranslatorNodeIn: aTMethod
	"Answer a TParseNode subclass equivalent of me"

	^TConstantNode new setValue: key! !

!ReturnNode methodsFor: '*VMMaker-C translation' stamp: 'eem 8/15/2018 18:43'!
asTranslatorNodeIn: aTMethod
	"Answer a TParseNode subclass equivalent of a return."
	| exprTranslation lastExpr |
	exprTranslation := expr asTranslatorNodeIn: aTMethod.
	(expr isMessage
	 and: [expr receiver isVariableNode
	 and: [expr receiver key = 'super'
	 and: [exprTranslation isStmtList]]]) ifTrue:
		["super expansions containing returns are fine, and (as of 6/25/2012 19:27) the last
		  return is elided from the expansion by TMethod>>superExpansionNodeFor:args:. 
		  So we need to ensure the last expression is a return and simply reuse any other
		  returns in the expansion."
		lastExpr := exprTranslation statements last.
		(lastExpr isReturn
		 or: [lastExpr isReturningIf]) ifFalse:
			[exprTranslation statements
				at: exprTranslation statements size
				put:
					(TReturnNode new 
						setExpression: lastExpr;
						comment: comment;
						yourself)].
		 ^exprTranslation].
	^TReturnNode new 
		setExpression: exprTranslation;
		comment: comment;
		yourself! !

!ReturnNode methodsFor: '*VMMaker-C translation' stamp: 'eem 2/22/2011 16:59'!
isReturn
	^true! !

!CompiledMethod methodsFor: '*VMMaker-support' stamp: 'eem 11/15/2012 20:26'!
abstractDetailedSymbolic
	^String streamContents: [:s| (RelativeDetailedInstructionPrinter on: self) printInstructionsOn: s]! !

!CompiledMethod methodsFor: '*VMMaker-C translation' stamp: 'eem 9/11/2018 18:19'!
asTranslationMethodOfClass: aTMethodClass
 	"Answer a TMethod (or subclass) derived from the receiver."
	^((CompiledMethod includesSelector: #ast)
			ifTrue: [self parseTree] "Pharo Opal Bytecode Compiler"
			ifFalse: [self methodNode]) "Squeak Smalltalk-80 Bytecode Compiler"
		asTranslationMethodOfClass: aTMethodClass! !

!CompiledMethod methodsFor: '*VMMaker-support' stamp: 'eem 12/2/2010 13:24'!
detailedSymbolic
	^String streamContents: [:s| (DetailedInstructionPrinter on: self) printInstructionsOn: s]! !

!Process methodsFor: '*VMMaker-simulation' stamp: 'eem 11/10/2020 11:07'!
asUnsignedInteger
	"Processes are used to model OS threads in the COGMTVM.
	 But native threads are typically simply pointers which can hence be mapped to integers.
	 This is used in locking the VM from an unknown thread on callback.  Hence mimic the
	 ability to map a thread to an integer by answering the receiver's identityHash."
	^self identityHash! !

!SequenceableCollection methodsFor: '*VMMaker-translation support' stamp: 'eem 3/5/2009 15:43'!
copyReplaceAll: oldSubstring with: newSubstring tokenish: tokenishBlock
	"Answer a copy of the receiver in which all occurrences of
	 oldSubstring have been replaced by newSubstring.
	 tokenishBlock specifies what elements are in a token."

	| sequence startSearch currentIndex endIndex |
	sequence := self.
	startSearch := 1.
	[(currentIndex := sequence indexOfSubCollection: oldSubstring startingAt: startSearch)
		> 0] whileTrue: 
		[endIndex := currentIndex + oldSubstring size - 1.
		((currentIndex = 1
				or: [(tokenishBlock value: (sequence at: currentIndex-1)) not])
		  and: [endIndex = sequence size
				or: [(tokenishBlock value: (sequence at: endIndex+1)) not]])
			ifTrue: [sequence := sequence
					copyReplaceFrom: currentIndex
					to: endIndex
					with: newSubstring.
				startSearch := currentIndex + newSubstring size]
			ifFalse: [startSearch := currentIndex + 1]].
	^sequence

"Test case:
	'pid_t (*pid)(pid_t pid)' copyReplaceAll: 'pid' with: 'foo' tokenish: [:ch| ch = $_ or: [ch isAlphaNumeric]]"
! !

!Symbol methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 8/12/2021 10:44'!
& aSymbol
	^(VMClass valueOfSymbol: self) & (VMClass valueOfSymbol: aSymbol)! !

!Symbol methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 7/25/2021 14:43'!
asVoidPointer
	^self! !

!Symbol methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 6/21/2008 11:44'!
coerceTo: cTypeString sim: interpreterSimulator

	^self! !

!Symbol methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 1/7/2016 14:20'!
defined
	"To allow constructs such as self cppIf: #'SA_NOCLDSTOP' defined ifTrue: [...].
	We could go look for a definition but likely there won't be one."
	^(thisContext sender methodClass bindingOf: self)
		ifNil: [false]
		ifNotNil: [:binding| binding value ~~ #undefined]! !

!LargePositiveInteger class methodsFor: '*VMMaker-plugin generation' stamp: 'nice 4/19/2016 23:03'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg ccgLoad: aBlock expr: aString asMemberOfLargePositiveIntegerFrom: anInteger! !

!LargePositiveInteger methodsFor: '*VMMaker-Tests' stamp: 'dtl 7/27/2011 12:22'!
minus64: anInteger 
	"For unit test support only. Subtract two large integers, raising an error on failure."

	<primitive: 22>
	^self primitiveFailed! !

InstructionPrinter subclass: #DetailedInstructionPrinter
	instanceVariableNames: 'stackHeightComputer encoderClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Support'!

!DetailedInstructionPrinter class methodsFor: 'instance creation' stamp: 'eem 11/26/2010 12:32'!
on: method 
	"Answer an instance of me on the argument, method."

	^self new method: method pc: method initialPC! !

!DetailedInstructionPrinter methodsFor: 'accessing' stamp: 'eem 8/2/2014 08:57'!
encoderClass

	^encoderClass! !

!DetailedInstructionPrinter methodsFor: 'accessing' stamp: 'eem 8/2/2014 08:56'!
encoderClass: anObject

	encoderClass := anObject! !

!DetailedInstructionPrinter methodsFor: 'decoding' stamp: 'eem 8/5/2014 11:14'!
interpretNextInstructionFor: client
	^encoderClass
		ifNil: [super interpretNextInstructionFor: client]
		ifNotNil: [encoderClass interpretNextInstructionFor: client in: self]! !

!DetailedInstructionPrinter methodsFor: 'initialize-release' stamp: 'eem 11/26/2010 12:30'!
method: aMethod
	super method: aMethod.
	stackHeightComputer := StackDepthFinder new method: aMethod pc: nil! !

!DetailedInstructionPrinter methodsFor: 'initialize-release' stamp: 'eem 11/26/2010 12:32'!
method: aMethod pc: pc
	super method: aMethod.
	stackHeightComputer := StackDepthFinder new method: aMethod pc: pc! !

!DetailedInstructionPrinter methodsFor: 'printing' stamp: 'eem 11/26/2010 12:33'!
print: instruction
	| current next |
	super print: instruction.
	current := stackHeightComputer stackp.
	stackHeightComputer interpretNextInstructionFor: stackHeightComputer.
	next := stackHeightComputer stackp.
	stream
		skip: -1;
		tab; tab; nextPut: $(;
		print: current.
	current ~= next ifTrue:
		[stream nextPut: $-; nextPut: $>; print: next].
	stream
		nextPut: $);
		cr;
		flush! !

!DetailedInstructionPrinter methodsFor: 'printing' stamp: 'eem 1/4/2019 16:34'!
printInstructionsOn: aStream
	"Append to the stream, aStream, a description of each bytecode in the
	 instruction stream."
	
	| end |
	encoderClass ifNil:
		[^super printInstructionsOn: aStream].
	stackHeightComputer encoderClass: encoderClass.
	stream := aStream.
	scanner := InstructionStream on: method.
	end := method endPC.
	oldPC := scanner pc.
	innerIndents := Array new: end withAll: 0.
	fullBlockRecursionSelector := #printInstructionsOn:.
	[scanner pc <= end] whileTrue:
		[encoderClass interpretNextInstructionFor: self in: scanner]! !

!DetailedInstructionPrinter methodsFor: 'printing' stamp: 'eem 1/4/2019 16:34'!
printInstructionsOn: aStream do: aBlock
	"Append to the stream, aStream, a description of each bytecode in the
	 instruction stream. Evaluate aBlock with the receiver, the scanner and
	 the stream after each instruction."
	
	| end |
	encoderClass ifNil:
		[^super printInstructionsOn: aStream do: aBlock].
	stackHeightComputer encoderClass: encoderClass.
	stream := aStream.
	scanner := InstructionStream on: method.
	end := method endPC.
	oldPC := scanner pc.
	innerIndents := Array new: end withAll: 0.
	[scanner pc <= end] whileTrue:
		[encoderClass interpretNextInstructionFor: self in: scanner.
		 aBlock value: self value: scanner value: stream]! !

!DetailedInstructionPrinter methodsFor: 'accessing' stamp: 'eem 1/11/2013 16:19'!
stackHeightComputer
	"Answer the value of stackHeightComputer"

	^ stackHeightComputer! !

!DetailedInstructionPrinter methodsFor: 'accessing' stamp: 'eem 1/11/2013 16:19'!
stackHeightComputer: anObject
	"Set the value of stackHeightComputer"

	stackHeightComputer := anObject! !

!ParseNode methodsFor: '*VMMaker-C translation' stamp: 'eem 8/15/2018 18:41'!
asTranslatorNodeIn: aTMethod
	"Answer a TParseNode subclass equivalent of me"
	self subclassResponsibility! !

!ParseNode methodsFor: '*VMMaker-C translation' stamp: 'eem 10/22/2013 15:30'!
isAssignmentEqualsEqualsNil
	^false! !

!ParseNode methodsFor: '*VMMaker-C translation' stamp: 'eem 2/22/2011 17:00'!
isReturn
	^false! !

!WarpBlt methodsFor: '*VMMaker-Interpreter-system simulation' stamp: 'nk 11/3/2004 09:26'!
warpBitsSimulated
	"Simulate WarpBlt"

	cellSize < 1 ifTrue: [ ^self error: 'cellSize must be >= 1' ].

	self warpBitsSimulated: cellSize
		sourceMap: (sourceForm colormapIfNeededForDepth: 32).
! !

!WarpBlt methodsFor: '*VMMaker-Interpreter-system simulation' stamp: 'tfel 2/17/2016 17:47'!
warpBitsSimulated: n sourceMap: sourceMap
	"Simulate WarpBlt"
	^ BitBltSimulation warpBitsFrom: self! !

!LargeNegativeInteger class methodsFor: '*VMMaker-plugin generation' stamp: 'nice 4/19/2016 23:02'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg ccgLoad: aBlock expr: aString asMemberOfLargeNegativeIntegerFrom: anInteger! !

!PrintfFormatString methodsFor: '*VMMaker-C code generation' stamp: 'eem 1/7/2021 15:39'!
transformForVMMaker
	[outputStream nextPutAll: string next. 
	 format atEnd] whileFalse:
		[| nextDescriptorOrNil |
		 [nextDescriptorOrNil := format next transformForVMMaker.
		  nextDescriptorOrNil isNil] whileTrue. "skip PrintfNumberHolderDescriptors"
		 outputStream nextPutAll: nextDescriptorOrNil].
	"any remainder is string, if so append to outputStream"
	string atEnd ifFalse: [outputStream nextPutAll: string next].
	^outputStream contents! !

!BlockClosure methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 12/24/2015 12:13'!
shouldStopIfAtPC: address
	<primitive: 202>
	^self value: address! !

!Collection methodsFor: '*VMMaker-printing' stamp: 'eem 10/2/2015 11:19'!
storeStringBase: base
	"This for bit/byte/word collections."
	^String streamContents: [:strm | self storeOn: strm base: base]! !

!BalloonEngine methodsFor: '*VMMaker-InterpreterSimulation' stamp: 'tfel 3/3/2016 15:26'!
simulatePrimitive: aString args: args 
	"simulate primitives in RSqueak"
	| proxy bb |
	proxy := InterpreterProxy new.
	proxy synthesizeStackFor: self with: args.
	bb := BalloonEngineSimulation simulatorClass new.
	bb setInterpreter: proxy.
	bb initialiseModule.
	"rendering state is loaded in the primitive implementations"
	bb perform: aString asSymbol.
	^ proxy stackValue: 0! !

!Class methodsFor: '*VMMaker-testing' stamp: 'eem 7/9/2021 19:31'!
isVMClass
	^false! !

!AssignmentNode methodsFor: '*VMMaker-C translation' stamp: 'eem 6/25/2012 19:23'!
anyReturns: aSequenceOfNodes
	aSequenceOfNodes do:
		[:node|
		node nodesDo:
			[:n|
			n isReturn ifTrue:
				[^true]]].
	^false! !

!AssignmentNode methodsFor: '*VMMaker-C translation' stamp: 'eem 8/15/2018 18:42'!
asTranslatorNodeIn: aTMethod
	"Answer a TParseNode subclass equivalent of me"
	| varNode valueNode |
	varNode := variable asTranslatorNodeIn: aTMethod.
	valueNode := value asTranslatorNodeIn: aTMethod.
	valueNode isStmtList ifFalse:
		[^TAssignmentNode new
			setVariable: varNode
			expression: valueNode;
			comment: comment].
	 "This is a super expansion.  We are in trouble if any statement other than the last is a return."
	(self anyReturns: valueNode statements allButLast) ifTrue:
		[self error: 'haven''t implemented pushing down assignments into other than the last return'].
	"As of 6/25/2012 19:30 superExpansionNodeFor:args: elides the final return."
	self assert: valueNode statements last isReturn not.
	^TStmtListNode new
		setStatements: valueNode statements allButLast,
					{ TAssignmentNode new
						setVariable: varNode
						expression: valueNode statements last;
						comment: comment };
		yourself! !

!PrintfFormatDescriptor methodsFor: '*VMMaker-C code generation' stamp: 'eem 11/13/2017 15:37'!
transformForVMMaker
	^self printString! !

!PrintfNumberFormatDescriptor methodsFor: '*VMMaker-C code generation' stamp: 'eem 11/13/2017 16:01'!
transformForVMMaker
	('duxX' includes: operator) ifFalse:
		[^super transformForVMMaker].
	^'%" PRI', (String with: operator), 'SQINT "'! !

!Boolean class methodsFor: '*VMMaker-plugin generation' stamp: 'eem 10/2/2010 14:12'!
ccg: cg generateCoerceToOopFrom: aNode on: aStream
	"N.B.  The is used both for generation and simulation so answer the result (for interpretation)"
	^cg generateCoerceToBooleanObjectFrom: aNode on: aStream! !

!Boolean class methodsFor: '*VMMaker-plugin generation' stamp: 'eem 10/3/2010 08:09'!
ccg: cg generateCoerceToValueFrom: aNode on: aStream
	"N.B.  The could be used both for generation and simulation so answer the result (for interpretation)"
	^cg generateCoerceToBooleanValueFrom: aNode on: aStream! !

!Boolean class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/18/1999 17:08'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg ccgLoad: aBlock expr: aString asBooleanValueFrom: anInteger! !

!Boolean methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 12/30/2016 14:22'!
asBooleanObj
	^(Notification new tag: #getInterpreter; signal)
		ifNotNil: [:interpreter| interpreter booleanObjectOf: self]
		ifNil: [self]! !

!Boolean methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 13:33'!
isBreakpointFor: address
	^self! !

!Boolean methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 13:08'!
menuPrompt
	^' (CLICK STEPPING!!!!)'! !

!Boolean methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 13:01'!
removeBreakpoint: address
	^nil! !

!Boolean methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 12/24/2015 12:45'!
shouldStopIfAtPC: address
	^self! !

!Boolean methodsFor: '*VMMaker-breakpoints' stamp: 'eem 1/14/2016 15:43'!
singleStepRequiredToTriggerIn: aCogit
	^self! !

TestCase subclass: #BitBltSimulationTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!BitBltSimulationTest commentStamp: 'jmv 10/26/2009 09:03' prior: 0!
These tests require VMMaker and should be considered part of it.!

!BitBltSimulationTest methodsFor: 'running' stamp: 'dtl 6/6/2010 15:19'!
setUp
	(self respondsTo: #timeout: )
		ifTrue: [self perform: #timeout: with: 10]
! !

!BitBltSimulationTest methodsFor: 'tests' stamp: 'jmv 10/26/2009 08:49'!
testAlphaCompositing2Simulated
	"self run: #testAlphaCompositing2Simulated"

	| bb f1 f2 mixColor result eps |
	f1 := Form extent: 1@1 depth: 32.
	f2 := Form extent: 1@1 depth: 32.
	eps := 0.5 / 255.
	0 to: 255 do:[:i|
		f1 colorAt: 0@0 put: Color transparent.
		mixColor := Color red alpha: i / 255.0.
		f2 colorAt: 0@0 put: mixColor.
		mixColor := f2 colorAt: 0@0.
		bb := BitBlt toForm: f1.
		bb sourceForm: f2.
		bb combinationRule: Form blend.
		bb copyBitsSimulated.
		result := f1 colorAt: 0@0.
		self assert: (result red - mixColor alpha) abs < eps.
		self assert: result alpha = mixColor alpha.
	]! !

!BitBltSimulationTest methodsFor: 'tests' stamp: 'jmv 10/26/2009 08:50'!
testAlphaCompositingSimulated
	"self run: #testAlphaCompositingSimulated"
	
	| bb f1 f2 mixColor result eps |

	f1 := Form extent: 1@1 depth: 32.
	f2 := Form extent: 1@1 depth: 32.
	eps := 0.5 / 255.
	0 to: 255 do:[:i|
		f1 colorAt: 0@0 put: Color blue.
		mixColor := Color red alpha: i / 255.0.
		f2 colorAt: 0@0 put: mixColor.
		mixColor := f2 colorAt: 0@0.
		bb := BitBlt toForm: f1.
		bb sourceForm: f2.
		bb combinationRule: Form blend.
		bb copyBitsSimulated.
		result := f1 colorAt: 0@0.
		self assert: (result red - mixColor alpha) abs < eps.
		self assert: (result blue - (1.0 - mixColor alpha)) abs < eps.
		self assert: result alpha = 1.0.
	]! !

!BitBltSimulationTest methodsFor: 'tests' stamp: 'tfel 3/3/2016 16:59'!
testDisplayStringSimulated
	"self run: #testAlphaCompositingSimulated"
	
	| bb f1 font |
	f1 := Form extent: 1@1 depth: 32.
	bb := BitBlt toForm: f1.
	bb combinationRule: 15.
	bb destX: 1.
	bb destY: 1.
	bb sourceX: 100.
	bb width: 100.
	font := TextStyle default fontOfSize: 21.
	bb simulatePrimitive: 'primitiveDisplayString' args: {'hello'. 1. 4. font characterToGlyphMap. font xTable. 1}.
! !

TestCase subclass: #CCodeGeneratorTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!CCodeGeneratorTests methodsFor: 'private' stamp: 'eem 1/2/2013 11:11'!
cg
	^CCodeGenerator new! !

!CCodeGeneratorTests methodsFor: 'tests' stamp: 'nice 12/29/2020 12:19'!
testIntegerGeneration

	"Test the 32-bit integers. They need to be marked as unsigned.
	 Test 16rFFFFFFFF, 16rFFFFFFFE, ... through to ..., 16rC0000000, 16r80000000"
	((0 to: 31) collect: [:shift| 16rFFFFFFFF bitClear: (1 bitShift: shift) - 1]) do:
		[:number| | literal isHex isDec |
		literal := self cg cLiteralFor: number.
		isHex := (literal beginsWith: '0x') and: [((literal allButFirst: 2) allButLast: 1) allSatisfy: [:c| '0123456789CEF' includes: c]].
		isDec := (literal allButLast: 1) allSatisfy: [:c| c isDigit].
		self assert: isHex | isDec.
		self assert: (literal endsWith: 'U').

		literal := self cg cLiteralFor: number name: 'Mask'.
		self assert: (literal beginsWith: '0x').
		self assert: (((literal allButFirst: 2) allButLast: 1) allSatisfy: [:c| '0123456789CEF' includes: c]).
		self assert: (literal endsWith: 'U')].

	"Test the 64-bit integers. They need to be marked as unsigned long longs."
	((32 to: 63) collect: [:shift| 16rFFFFFFFFFFFFFFFF bitClear: (1 bitShift: shift) - 1]) do:
		[:number| | literal isHex isDec |
		literal := self cg cLiteralFor: number.
		isHex := (literal beginsWith: '0x') and: [((literal allButFirst: 2) allButLast: 3) allSatisfy: [:c| '0123456789CEF' includes: c]].
		isDec := (literal allButLast: 3) allSatisfy: [:c| c isDigit].
		self assert: isHex | isDec.
		self assert: (literal endsWith: 'ULL').

		literal := self cg cLiteralFor: number name: 'Mask'.
		self assert: (literal beginsWith: '0x').
		self assert: (((literal allButFirst: 2) allButLast: 3) allSatisfy: [:c| '0123456789CEF' includes: c]).
		self assert: (literal endsWith: 'ULL')]! !

!CCodeGeneratorTests methodsFor: 'tests' stamp: 'nice 12/29/2020 14:30'!
testIntegerGenerationHexOrDec
	| decPrefered hexPrefered |
	hexPrefered := #(
			16r100 16r400 16r10000 "powers of two (more than 8 bits)"
			16r1F 16r3FF 16rFFFF "powers of two minus 1 - with at least 5 bits set"
			16r3E 16r3FF00 16r1FE0  "shifted powers of two minus 1 - wih at least 5 bit set..."
			16r38000 "... or having highBit >= 16"
			16rC00 "... or requiring 1.5 times less different characters than decimal form - here 2 instead of 4"
			16r1F1F 16rCCCC  16r112233 "some regular bit patterns"
			2r111000111000111000111 "not necessarily falling on 4 bits boundaries"
			).
	decPrefered := #(
			0 1 2 3 4 5 6 7 8 9 "single digit"
			10 100 1000 10000 "powers of ten"
			9 99 999 9999 "powers of ten minus 1"
			16r10 16r20 16r40 16r80 "small powers of two"
			16r60 16rC0 16rF0  16r1E00 "shifted powers of two minus 1 - with less than 5 bits set"
			112233 15341 "random patterns when no hex rule apply"
			2r1111100000111110000011111000001111100000 "regular bit pattern if the decimal form appear somehow also regular -  here it uses a pattern of only 4 different digits : 1066193093600").
	hexPrefered do: [:number || literal |
		literal := self cg cLiteralFor: number.
		self assert: (literal beginsWith: '0x')].
	decPrefered do: [:number || literal |
		literal := self cg cLiteralFor: number.
		self deny: (literal beginsWith: '0x')].! !

TestCase subclass: #FloatMathPluginTests
	instanceVariableNames: 'seed'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!FloatMathPluginTests commentStamp: '<historical>' prior: 0!
FloatMathPluginTests buildSuite run.!

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 17:34'!
arcCos: f
	<primitive: 'primitiveArcCos' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 17:34'!
arcCosH: f
	<primitive: 'primitiveArcCosH' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 17:34'!
arcSin: f
	<primitive: 'primitiveArcSin' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 17:34'!
arcSinH: f
	<primitive: 'primitiveArcSinH' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 17:35'!
arcTan2: value with: arg
	<primitive: 'primitiveArcTan2' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 14:29'!
arcTan: value
	<primitive: 'primitiveArcTan' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 17:35'!
arcTanH: value
	<primitive: 'primitiveArcTanH' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 17:35'!
cos: value
	<primitive: 'primitiveCos' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 17:35'!
cosH: value
	<primitive: 'primitiveCosH' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 14:29'!
exp: value
	<primitive: 'primitiveExp' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 14:29'!
fractionPart: value
	<primitive: 'primitiveFractionalPart' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 17:36'!
hypot: x with: y
	<primitive: 'primitiveHypot' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 17:24'!
ln: value
	<primitive: 'primitiveLogN' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 17:41'!
log10: value
	<primitive: 'primitiveLog10' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'running' stamp: 'ar 3/26/2006 16:52'!
makeLargeTestData
	"self basicNew makeLargeTestData"
	self makeTestData: 'sin-large.dat' using:[:f| self sin: f] seed: 432567 rounds: 1000000.
	self makeTestData: 'log-large.dat' using:[:f| self ln: f abs] seed: 432567 rounds: 1000000.
	self makeTestData: 'sqrt-large.dat' using:[:f| self sqrt: f abs] seed: 432567 rounds: 1000000.
	self makeTestData: 'atan-large.dat' using:[:f| self arcTan: f] seed: 432567 rounds: 1000000.
	self makeTestData: 'exp-large.dat' using:[:f| self exp: f] seed: 432567 rounds: 1000000.
! !

!FloatMathPluginTests methodsFor: 'running' stamp: 'ar 3/26/2006 16:52'!
makeSmallTestData
	"self basicNew makeSmallTestData"
	self makeTestData: 'sin-small.dat' using:[:f| self sin: f] seed: 321567 rounds: 10000.
	self makeTestData: 'log-small.dat' using:[:f| self ln: f abs] seed: 321567 rounds: 10000.
	self makeTestData: 'sqrt-small.dat' using:[:f| self sqrt: f abs] seed: 321567 rounds: 10000.
	self makeTestData: 'atan-small.dat' using:[:f| self arcTan: f] seed: 321567 rounds: 10000.
	self makeTestData: 'exp-small.dat' using:[:f| self exp: f] seed: 321567 rounds: 10000.
! !

!FloatMathPluginTests methodsFor: 'running' stamp: 'nice 12/11/2018 22:32'!
makeTestData: fileName using: aBlock seed: seed rounds: rounds
	| bytes out float result |
	bytes := ByteArray new: 8.
	out := FileStream newFileNamed: fileName.
	[
		out binary. 
		out nextNumber: 4 put: rounds.
		out nextNumber: 4 put: seed.
		self seed: seed.
		float := Float basicNew: 2.
		'Creating test data for: ', fileName 
			displayProgressAt: Sensor cursorPoint 
			from: 1 to: rounds during:[:bar|
				1 to: rounds do:[:i|
					i \\ 10000 = 0 ifTrue:[bar value: i].
					[1 to: 8 do:[:j| bytes at: j put: (self nextInt: 256)-1].
					float basicAt: 1 put: (bytes unsignedLongAt: 1 bigEndian: true).
					float basicAt: 2 put: (bytes unsignedLongAt: 5 bigEndian: true).
					float isNaN] whileTrue.
					result := aBlock value: float.
					out nextNumber: 4 put: (result basicAt: 1).
					out nextNumber: 4 put: (result basicAt: 2).
				].
			].
	] ensure:[out close].
! !

!FloatMathPluginTests methodsFor: 'md5' stamp: 'ar 3/26/2006 15:19'!
md5HashMessage: aStringOrByteArray
	^ self md5HashStream: (ReadStream on: aStringOrByteArray asByteArray)
! !

!FloatMathPluginTests methodsFor: 'md5' stamp: 'eem 7/22/2010 13:20'!
md5HashStream: aStream
	| start buffer bytes sz n words hash large |
	hash := WordArray with: 16r67452301 with: 16rEFCDAB89 with: 16r98BADCFE with: 16r10325476.
	words := WordArray new: 16.
	buffer := ByteArray new: 64.
	start := aStream position.
	[aStream atEnd] whileFalse: [
		bytes := aStream nextInto: buffer.
		(bytes size < 64 or:[aStream atEnd]) ifTrue:[
			sz := bytes size.
			buffer replaceFrom: 1 to: sz with: bytes startingAt: 1.
			buffer from: sz+1 to: buffer size put: 0.
			sz < 56 ifTrue:[
				buffer at: sz + 1 put: 128. "trailing bit"
			] ifFalse:[
				"not enough room for the length, so just pad this one, then..."
				sz < 64 ifTrue:[buffer at: sz + 1 put: 128].
				1 to: 16 do:[:i| words at: i put: (buffer unsignedLongAt: i*4-3 bigEndian: false)].
				self md5Transform: words hash: hash.
				"process one additional block of padding ending with the length"
				buffer atAllPut: 0.
				sz = 64 ifTrue: [buffer at: 1 put: 128].
			].
			"Fill in the final 8 bytes with the 64-bit length in bits."
			n := (aStream position - start) * 8.
			7 to: 0 by: -1 do:[:i| buffer at: (buffer size - i) put: ((n bitShift: 7 - i * -8) bitAnd: 255)].
		].
		1 to: 16 do:[:i| words at: i put: (buffer unsignedLongAt: i*4-3 bigEndian: false)].
		self md5Transform: words hash: hash.
	].
	bytes := ByteArray new: 16.
	bytes unsignedLongAt: 1 put: (hash at: 4) bigEndian: true.
	bytes unsignedLongAt: 5 put: (hash at: 3) bigEndian: true.
	bytes unsignedLongAt: 9 put: (hash at: 2) bigEndian: true.
	bytes unsignedLongAt: 13 put: (hash at: 1) bigEndian: true.
	large := LargePositiveInteger new: 16.
	1 to: 16 do:[:i| large digitAt: i put: (bytes at: i)].
	^large normalize! !

!FloatMathPluginTests methodsFor: 'md5' stamp: 'ar 3/26/2006 19:45'!
md5Transform: in hash: hash
	"This adds the incoming words to the existing hash"
	| a b c d |
	<primitive: 'primitiveMD5Transform' module: 'CroquetPlugin'>
	a := hash at: 1.
	b := hash at: 2.
	c := hash at: 3.
	d := hash at: 4.

	a := self step1: a x: b y: c z: d data: (in at:  1) add: 16rD76AA478 shift: 7.
	d := self step1: d x: a y: b z: c data: (in at:  2) add: 16rE8C7B756 shift: 12.
	c := self step1: c x: d y: a z: b data: (in at:  3) add: 16r242070DB shift: 17.
	b := self step1: b x: c y: d z: a data: (in at:  4) add: 16rC1BDCEEE shift: 22.
	a := self step1: a x: b y: c z: d data: (in at:  5) add: 16rF57C0FAF shift:  7.
	d := self step1: d x: a y: b z: c data: (in at:  6) add: 16r4787C62A shift: 12.
	c := self step1: c x: d y: a z: b data: (in at:  7) add: 16rA8304613 shift: 17.
	b := self step1: b x: c y: d z: a data: (in at:  8) add: 16rFD469501 shift: 22.
	a := self step1: a x: b y: c z: d data: (in at:  9) add: 16r698098D8 shift:  7.
	d := self step1: d x: a y: b z: c data: (in at: 10) add: 16r8B44F7AF shift: 12.
	c := self step1: c x: d y: a z: b data: (in at: 11) add: 16rFFFF5BB1 shift: 17.
	b := self step1: b x: c y: d z: a data: (in at: 12) add: 16r895CD7BE shift: 22.
	a := self step1: a x: b y: c z: d data: (in at: 13) add: 16r6B901122 shift:  7.
	d := self step1: d x: a y: b z: c data: (in at: 14) add: 16rFD987193 shift: 12.
	c := self step1: c x: d y: a z: b data: (in at: 15) add: 16rA679438E shift: 17.
	b := self step1: b x: c y: d z: a data: (in at: 16) add: 16r49B40821 shift: 22.

	a := self step2: a x: b y: c z: d data: (in at:  2) add: 16rF61E2562 shift:  5.
	d := self step2: d x: a y: b z: c data: (in at:  7) add: 16rC040B340 shift:  9.
	c := self step2: c x: d y: a z: b data: (in at: 12) add: 16r265E5A51 shift: 14.
	b := self step2: b x: c y: d z: a data: (in at:  1) add: 16rE9B6C7AA shift: 20.
	a := self step2: a x: b y: c z: d data: (in at:  6) add: 16rD62F105D shift:  5.
	d := self step2: d x: a y: b z: c data: (in at: 11) add: 16r02441453 shift:  9.
	c := self step2: c x: d y: a z: b data: (in at: 16) add: 16rD8A1E681 shift: 14.
	b := self step2: b x: c y: d z: a data: (in at:  5) add: 16rE7D3FBC8 shift: 20.
	a := self step2: a x: b y: c z: d data: (in at: 10) add: 16r21E1CDE6 shift:  5.
	d := self step2: d x: a y: b z: c data: (in at: 15) add: 16rC33707D6 shift:  9.
	c := self step2: c x: d y: a z: b data: (in at:  4) add: 16rF4D50D87 shift: 14.
	b := self step2: b x: c y: d z: a data: (in at:  9) add: 16r455A14ED shift: 20.
	a := self step2: a x: b y: c z: d data: (in at: 14) add: 16rA9E3E905 shift:  5.
	d := self step2: d x: a y: b z: c data: (in at:  3) add: 16rFCEFA3F8 shift:  9.
	c := self step2: c x: d y: a z: b data: (in at:  8) add: 16r676F02D9 shift: 14.
	b := self step2: b x: c y: d z: a data: (in at: 13) add: 16r8D2A4C8A shift: 20.

	a := self step3: a x: b y: c z: d data: (in at:  6) add: 16rFFFA3942 shift:  4.
	d := self step3: d x: a y: b z: c data: (in at:  9) add: 16r8771F681 shift: 11.
	c := self step3: c x: d y: a z: b data: (in at: 12) add: 16r6D9D6122 shift: 16.
	b := self step3: b x: c y: d z: a data: (in at: 15) add: 16rFDE5380C shift: 23.
	a := self step3: a x: b y: c z: d data: (in at:  2) add: 16rA4BEEA44 shift:  4.
	d := self step3: d x: a y: b z: c data: (in at:  5) add: 16r4BDECFA9 shift: 11.
	c := self step3: c x: d y: a z: b data: (in at:  8) add: 16rF6BB4B60 shift: 16.
	b := self step3: b x: c y: d z: a data: (in at: 11) add: 16rBEBFBC70 shift: 23.
	a := self step3: a x: b y: c z: d data: (in at: 14) add: 16r289B7EC6 shift:  4.
	d := self step3: d x: a y: b z: c data: (in at:  1) add: 16rEAA127FA shift: 11.
	c := self step3: c x: d y: a z: b data: (in at:  4) add: 16rD4EF3085 shift: 16.
	b := self step3: b x: c y: d z: a data: (in at:  7) add: 16r04881D05 shift: 23.
	a := self step3: a x: b y: c z: d data: (in at: 10) add: 16rD9D4D039 shift:  4.
	d := self step3: d x: a y: b z: c data: (in at: 13) add: 16rE6DB99E5 shift: 11.
	c := self step3: c x: d y: a z: b data: (in at: 16) add: 16r1FA27CF8 shift: 16.
	b := self step3: b x: c y: d z: a data: (in at:  3) add: 16rC4AC5665 shift: 23.

	a := self step4: a x: b y: c z: d data: (in at:  1) add: 16rF4292244 shift:  6.
	d := self step4: d x: a y: b z: c data: (in at:  8) add: 16r432AFF97 shift: 10.
	c := self step4: c x: d y: a z: b data: (in at: 15) add: 16rAB9423A7 shift: 15.
	b := self step4: b x: c y: d z: a data: (in at:  6) add: 16rFC93A039 shift: 21.
	a := self step4: a x: b y: c z: d data: (in at: 13) add: 16r655B59C3 shift:  6.
	d := self step4: d x: a y: b z: c data: (in at:  4) add: 16r8F0CCC92 shift: 10.
	c := self step4: c x: d y: a z: b data: (in at: 11) add: 16rFFEFF47D shift: 15.
	b := self step4: b x: c y: d z: a data: (in at:  2) add: 16r85845DD1 shift: 21.
	a := self step4: a x: b y: c z: d data: (in at:  9) add: 16r6FA87E4F shift:  6.
	d := self step4: d x: a y: b z: c data: (in at: 16) add: 16rFE2CE6E0 shift: 10.
	c := self step4: c x: d y: a z: b data: (in at:  7) add: 16rA3014314 shift: 15.
	b := self step4: b x: c y: d z: a data: (in at: 14) add: 16r4E0811A1 shift: 21.
	a := self step4: a x: b y: c z: d data: (in at:  5) add: 16rF7537E82 shift:  6.
	d := self step4: d x: a y: b z: c data: (in at: 12) add: 16rBD3AF235 shift: 10.
	c := self step4: c x: d y: a z: b data: (in at:  3) add: 16r2AD7D2BB shift: 15.
	b := self step4: b x: c y: d z: a data: (in at: 10) add: 16rEB86D391 shift: 21.

	a := (a + (hash at: 1)) bitAnd: 16rFFFFFFFF. hash at: 1 put: a.
	b := (b + (hash at: 2)) bitAnd: 16rFFFFFFFF. hash at: 2 put: b.
	c := (c + (hash at: 3)) bitAnd: 16rFFFFFFFF. hash at: 3 put: c.
	d := (d + (hash at: 4)) bitAnd: 16rFFFFFFFF. hash at: 4 put: d.

	^hash! !

!FloatMathPluginTests methodsFor: 'rand' stamp: 'nice 12/11/2018 22:32'!
nextInt: anInteger
	"Answer a random integer in the interval [1, anInteger]."

	| a m q r lo hi aLoRHi |
	a := 16r000041A7 asFloat.    " magic constant =      16807 "
	m := 16r7FFFFFFF asFloat.    " magic constant = 2147483647 "
	q := (m quo: a) asFloat.
	r  := (m \\ a) asFloat.
	hi := (seed quo: q) asFloat.
	lo := seed - (hi * q).  " = seed rem: q"  
	aLoRHi := (a * lo) - (r * hi).
	seed := (aLoRHi > 0.0)
		ifTrue:  [aLoRHi]
		ifFalse: [aLoRHi + m].
	^ (seed / m * anInteger) truncated + 1! !

!FloatMathPluginTests methodsFor: 'md5' stamp: 'ar 3/26/2006 15:19'!
rotate: value by: amount
	"Rotate value left by amount"
	| lowMask highMask |
	lowMask := (1 bitShift: 32-amount) - 1.
	highMask := 16rFFFFFFFF - lowMask.
	^((value bitAnd: lowMask) bitShift: amount) + 
		((value bitAnd: highMask) bitShift: amount-32)! !

!FloatMathPluginTests methodsFor: 'running' stamp: 'nice 12/11/2018 22:33'!
runTest: aBlock
	| bytes out float result |
	bytes := ByteArray new: 8.
	out := WriteStream on: ByteArray new.
	float := Float basicNew: 2.
	1 to: 10000 do:[:i|
		[1 to: 8 do:[:j| bytes at: j put: (self nextInt: 256)-1].
		float basicAt: 1 put: (bytes unsignedLongAt: 1 bigEndian: true).
		float basicAt: 2 put: (bytes unsignedLongAt: 5 bigEndian: true).
		float isNaN] whileTrue.
		result := [aBlock value: float] on: Error do:[:ex|
			"we convert all errors into NaNs to have a value for testing"
			ex return: Float nan.
		].
		out nextNumber: 4 put: (result basicAt: 1).
		out nextNumber: 4 put: (result basicAt: 2).
	].
	^self md5HashMessage: out contents.! !

!FloatMathPluginTests methodsFor: 'rand' stamp: 'nice 12/11/2018 22:23'!
seed: anInteger
	seed := anInteger! !

!FloatMathPluginTests methodsFor: 'running' stamp: 'nice 12/11/2018 22:33'!
setUp
	self seed: 253213.! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 14:29'!
sin: value
	<primitive: 'primitiveSin' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 17:37'!
sinH: value
	<primitive: 'primitiveSinH' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 14:29'!
sqrt: value
	<primitive: 'primitiveSqrt' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'md5' stamp: 'ar 3/26/2006 15:19'!
step1: w x: x y: y z: z data: data add: add shift: s
	"First step in MD5 transformation"
	| f result |
	f := z bitXor: (x bitAnd: (y bitXor: z)).
	result := w + f + data + add.
	result := self rotate: result by: s.
	^result + x bitAnd: 16rFFFFFFFF! !

!FloatMathPluginTests methodsFor: 'md5' stamp: 'ar 3/26/2006 15:19'!
step2: w x: x y: y z: z data: data add: add shift: s
	"First step in MD5 transformation"
	| f result |
	f := y bitXor: (z bitAnd: (x bitXor: y)).
	result := w + f + data + add.
	result := self rotate: result by: s.
	^result + x bitAnd: 16rFFFFFFFF! !

!FloatMathPluginTests methodsFor: 'md5' stamp: 'ar 3/26/2006 15:19'!
step3: w x: x y: y z: z data: data add: add shift: s
	"First step in MD5 transformation"
	| f result |
	f := (x bitXor: y) bitXor: z.
	result := w + f + data + add.
	result := self rotate: result by: s.
	^result + x bitAnd: 16rFFFFFFFF! !

!FloatMathPluginTests methodsFor: 'md5' stamp: 'ar 3/26/2006 15:19'!
step4: w x: x y: y z: z data: data add: add shift: s
	"First step in MD5 transformation"
	| f result |
	f := y bitXor: (x bitOr: (z  bitXor: 16rFFFFFFFF)).
	result := w + f + data + add.
	result := self rotate: result by: s.
	^result + x bitAnd: 16rFFFFFFFF! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 17:37'!
tan: value
	<primitive: 'primitiveTan' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 17:37'!
tanH: value
	<primitive: 'primitiveTanH' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 4/15/2007 22:46'!
testArcCos
	| hash |
	hash := self runTest:[:f| self arcCos: f].
	self assert: hash = 320603091210691421897131240956682310429! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 17:44'!
testArcCosH
	| hash |
	hash := self runTest:[:f| self arcCosH: f].
	self assert: hash = 6724426144112251941037505276242428134! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 17:43'!
testArcSin
	| hash |
	hash := self runTest:[:f| self arcSin: f].
	self assert: hash = 27372132577303862731837100895783885417! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 17:42'!
testArcSinH
	| hash |
	hash := self runTest:[:f| self arcSinH: f].
	self assert: hash = 255911863578190171815115260235896145802! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 17:43'!
testArcTan
	| hash |
	hash := self runTest:[:f| self arcTan: f].
	self assert: hash = 17311773710959114634056077345168823659! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 17:51'!
testArcTan2
	| hash |
	hash := self runTest:[:f| self arcTan2: f with: f].
	self assert: hash = 287068347279655848752274030373495709564! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 17:44'!
testArcTanH
	| hash |
	hash := self runTest:[:f| self arcTanH: f].
	self assert: hash = 295711907369004359459882231908879164929! !

!FloatMathPluginTests methodsFor: 'data' stamp: 'ar 3/26/2006 17:56'!
testAtanData
	self verifyTestData: 'atan-small.dat' using:[:f| self arcTan: f].
	self verifyTestData: 'atan-large.dat' using:[:f| self arcTan: f].
! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 17:45'!
testCos
	| hash |
	hash := self runTest:[:f| self cos: f].
	self assert: hash = 110207739557966732640546618158077332978! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 17:42'!
testCosH
	| hash |
	hash := self runTest:[:f| self cosH: f].
	self assert: hash = 139309299067563830037108641802292492276! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 14:27'!
testExp
	| hash |
	hash := self runTest:[:f| self exp: f].
	self assert: hash = 264681209343177480335132131244505189510! !

!FloatMathPluginTests methodsFor: 'data' stamp: 'ar 3/26/2006 17:56'!
testExpData
	self verifyTestData: 'exp-small.dat' using:[:f| self exp: f].
	self verifyTestData: 'exp-large.dat' using:[:f| self exp: f].
! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 14:27'!
testFloatAt
	| hash flt |
	flt := FloatArray new: 1.
	hash := self runTest:[:f| flt at: 1 put: f. flt at: 1].
	self assert: hash = 80498428122197125691266588764018905399! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 14:27'!
testFraction
	| hash |
	hash := self runTest:[:f| self fractionPart: f].
	self assert: hash = 320444785026869345695277323179170692004! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 17:51'!
testHypot
	| hash |
	hash := self runTest:[:f| self hypot: f with: f+1].
	self assert: hash = 217113721886532765853628735806816720346! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 14:27'!
testLog
	| hash |
	hash := self runTest:[:f| self ln: f abs].
	self assert: hash = 24389651894375564945708989023746058645! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 17:42'!
testLog10
	| hash |
	hash := self runTest:[:f| self log10: f abs].
	self assert: hash = 135564553959509933253581837789050718785! !

!FloatMathPluginTests methodsFor: 'data' stamp: 'ar 3/26/2006 17:56'!
testLogData
	self verifyTestData: 'log-small.dat' using:[:f| self ln: f abs].
	self verifyTestData: 'log-large.dat' using:[:f| self ln: f abs].
! !

!FloatMathPluginTests methodsFor: 'md5' stamp: 'ar 3/26/2006 15:20'!
testMD5
	self assert: (self md5HashMessage: 'a') = 16r0CC175B9C0F1B6A831C399E269772661.
	self assert: (self md5HashMessage: 'abc') = 16r900150983CD24FB0D6963F7D28E17F72.
	self assert: (self md5HashMessage: 'message digest') = 16rF96B697D7CB7938D525A2F31AAF161D0.
	self assert: (self md5HashMessage:
		'abcdefghijklmnopqrstuvwxyz') = 16rC3FCD3D76192E4007DFB496CCA67E13B.
	self assert: (self md5HashMessage:
		'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789') =
		16rD174AB98D277D9F5A5611C2C9F419D9F.
	self assert: (self md5HashMessage:
		'12345678901234567890123456789012345678901234567890123456789012345678901234567890') =
 		16r57EDF4A22BE3C955AC49DA2E2107B67A.! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 17:44'!
testSin
	| hash |
	hash := self runTest:[:f| self sin: f].
	self assert: hash = 290162321010315440569513182938961037473! !

!FloatMathPluginTests methodsFor: 'data' stamp: 'ar 3/26/2006 17:56'!
testSinData
	self verifyTestData: 'sin-small.dat' using:[:f| self sin: f].
	self verifyTestData: 'sin-large.dat' using:[:f| self sin: f].
! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 17:44'!
testSinH
	| hash |
	hash := self runTest:[:f| self sinH: f].
	self assert: hash = 146029709156303766079448006055284064911! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 14:28'!
testSqrt
	| hash |
	hash := self runTest:[:f| self sqrt: f abs].
	self assert: hash = 112236588358122834093969606123302196127! !

!FloatMathPluginTests methodsFor: 'data' stamp: 'ar 3/26/2006 17:56'!
testSqrtData
	self verifyTestData: 'sqrt-small.dat' using:[:f| self sqrt: f abs].
	self verifyTestData: 'sqrt-large.dat' using:[:f| self sqrt: f abs].! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 4/15/2007 22:46'!
testTan
	| hash |
	hash := self runTest:[:f| self tan: f].
	self assert: hash = 207143885112027702205238433494882679660! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'ar 3/26/2006 17:43'!
testTanH
	| hash |
	hash := self runTest:[:f| self tanH: f].
	self assert: hash = 15738508136206638425252880299326548123! !

!FloatMathPluginTests methodsFor: 'tests' stamp: 'nice 12/11/2018 22:33'!
testTimesTwoPower
	| hash |
	hash := self runTest:[:f| self timesTwoPower: f with: (self nextInt: 200) - 100].
	self assert: hash = 278837335583284459890979576373223649870.! !

!FloatMathPluginTests methodsFor: 'math' stamp: 'ar 3/26/2006 14:29'!
timesTwoPower: f with: arg
	<primitive: 'primitiveTimesTwoPower' module: 'FloatMathPlugin'>
	^self primitiveFailed! !

!FloatMathPluginTests methodsFor: 'running' stamp: 'nice 12/11/2018 22:33'!
verifyTestData: fileName using: aBlock
	| rounds seed bytes float result in expected count bits |
	in := [FileStream readOnlyFileNamed: fileName] 
			on: FileDoesNotExistException 
			do:[:ex| ex return: nil].
	in ifNil:[^nil].
	count := bits := 0.
	bytes := ByteArray new: 8.
	[
		in binary.
		rounds := in nextNumber: 4.
		seed := in nextNumber: 4.
		self seed: seed.
		float := Float basicNew: 2.
		expected := Float basicNew: 2.
		'Verifying test data from: ', fileName 
			displayProgressAt: Sensor cursorPoint 
			from: 1 to: rounds during:[:bar|
				1 to: rounds do:[:i|
					i \\ 10000 = 0 ifTrue:[bar value: i].
					[1 to: 8 do:[:j| bytes at: j put: (self nextInt: 256)-1].
					float basicAt: 1 put: (bytes unsignedLongAt: 1 bigEndian: true).
					float basicAt: 2 put: (bytes unsignedLongAt: 5 bigEndian: true).
					float isNaN] whileTrue.
					result := aBlock value: float.
					expected basicAt: 1 put: (in nextNumber: 4).
					expected basicAt: 2 put: (in nextNumber: 4).
					((expected isNaN and:[result isNaN]) or:[expected = result]) ifFalse:[
						(expected basicAt: 1) = (result basicAt: 1)
							ifFalse:[self error: 'Verification failure'].
						count := count + 1.
						bits := bits + ((expected basicAt: 2) - (result basicAt: 2)) abs.
					].
				].
			].
	] ensure:[in close].
	self assert: count = 0. "all the same"! !

TestCase subclass: #FullSimulationTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!FullSimulationTest methodsFor: 'as yet unclassified' stamp: 'nice 3/15/2016 22:37'!
testSimulate2MillionBytecodes
	<timeout: 120>
	| vm om |
	vm := StackInterpreterSimulator newWithOptions: (Smalltalk isRunningSpur
		ifTrue: [Smalltalk wordSize = 4
			ifTrue: [#(#ObjectMemory #Spur32BitMemoryManager )]
			ifFalse: [#(#ObjectMemory #Spur64BitMemoryManager )]]
		ifFalse: [#(#ObjectMemory)]).
	om := vm objectMemory.
	vm desiredNumStackPages: 8. "Makes simulation faster by creating fewer stack pages."
	vm openOn: Smalltalk imageName.
	vm instVarNamed: 'assertVEPAES' put: false. "This makes the simulation faster by turning off some expensive asserts"
	[[vm runForNBytes: 1500000]
		on: Halt , ProvideAnswerNotification "This exception handler ignores some halts and confirmers occurring during simulation"
		do: [:ex | 
			ex messageText == #primitiveExecuteMethodArgsArray
				ifTrue: [ex resume].
			ex messageText = 'clear transcript?'
				ifTrue: [ex resume: false].
			ex pass]] ensure: [Display restore].! !

TestCase subclass: #HeapMapTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!HeapMapTest methodsFor: 'tests' stamp: 'eem 2/27/2013 16:21'!
test32BitHeapMap
	| heapMap |
	heapMap := CogCheck32BitHeapMap new.

	"all bits shoud be clear in a new map"
	0 to: 1024 * 1024 - 4 by: 4 do:
		[:addr| self assert: (heapMap heapMapAtWord: addr) = 0].
	(2 raisedTo: 30) to: 1024 * 1024 - 4 + (2 raisedTo: 30) by: 4 do:
		[:addr| self assert: (heapMap heapMapAtWord: addr) = 0].

	"set every 9th bit to 1"
	0 to: 1024 * 1024 - 4 by: 36 do:
		[:addr| heapMap heapMapAtWord: addr Put: 1].	
	(2 raisedTo: 30) to: 1024 * 1024 - 4 + (2 raisedTo: 30) by:36 do:
		[:addr| heapMap heapMapAtWord: addr Put: 1].

	"check every 9th bit is set and all intervening bits are zero"
	0 to: 1024 * 1024 - 4 by: 36 do:
		[:addr|
		 addr - 32 > 0 ifTrue:
			[addr - 32 to: addr - 4 by: 4 do:
				[:zeroAddr| self assert: (heapMap heapMapAtWord: zeroAddr) = 0]].
		self assert: (heapMap heapMapAtWord: addr) = 1].
	
	(2 raisedTo: 30) to: 1024 * 1024 - 4 + (2 raisedTo: 30) by:36 do:
		[:addr|
		 addr - 32 > 0 ifTrue:
			[addr - 32 to: addr - 4 by: 4 do:
				[:zeroAddr| self assert: (heapMap heapMapAtWord: zeroAddr) = 0]].
		self assert: (heapMap heapMapAtWord: addr) = 1].

	"clear the bits"
	0 to: 1024 * 1024 - 4 by: 36 do:
		[:addr| heapMap heapMapAtWord: addr Put: 0].	
	(2 raisedTo: 30) to: 1024 * 1024 - 4 + (2 raisedTo: 30) by:36 do:
		[:addr| heapMap heapMapAtWord: addr Put: 0].

	"check everything is zero again"
	0 to: 1024 * 1024 - 4 by: 4 do:
		[:addr| self assert: (heapMap heapMapAtWord: addr) = 0].
	(2 raisedTo: 30) to: 1024 * 1024 - 4 + (2 raisedTo: 30) by: 4 do:
		[:addr| self assert: (heapMap heapMapAtWord: addr) = 0]! !

TestCase subclass: #IncludedMethodsTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!IncludedMethodsTest commentStamp: 'dtl 11/9/2010 21:03' prior: 0!
Various classes in the image contain methods that are intended to be translated to C and executed as primitives. IncludedMethodsTest provides tests to validate these methods.
!

!IncludedMethodsTest methodsFor: 'primitives' stamp: 'dtl 11/9/2010 19:48'!
compare: string1 with: string2 collated: order
	"Return 1, 2 or 3, if string1 is <, =, or > string2, with the collating order of characters given by the order array."

	<primitive: 'primitiveCompareString' module: 'MiscPrimitivePlugin'>
	self primitiveFailed
! !

!IncludedMethodsTest methodsFor: 'primitives' stamp: 'dtl 11/10/2010 19:42'!
compress: bm toByteArray: ba

	<primitive: 'primitiveCompressToByteArray' module: 'MiscPrimitivePlugin'>
	self primitiveFailed! !

!IncludedMethodsTest methodsFor: 'primitives' stamp: 'dtl 11/10/2010 20:03'!
convert8bitSignedFrom: aByteArray to16Bit: aSoundBuffer
	"Copy the contents of the given array of signed 8-bit samples into the given array of 16-bit signed samples."

	<primitive: 'primitiveConvert8BitSigned' module: 'MiscPrimitivePlugin'>
	self primitiveFailed
! !

!IncludedMethodsTest methodsFor: 'primitives' stamp: 'dtl 11/10/2010 19:42'!
decompress: bm fromByteArray: ba at: index

	<primitive: 'primitiveDecompressFromByteArray' module: 'MiscPrimitivePlugin'>
	self primitiveFailed
! !

!IncludedMethodsTest methodsFor: 'primitives' stamp: 'dtl 11/10/2010 20:03'!
findFirstInString: aString  inSet: inclusionMap  startingAt: start

	<primitive: 'primitiveFindFirstInString' module: 'MiscPrimitivePlugin'>
	self primitiveFailed
! !

!IncludedMethodsTest methodsFor: 'primitives' stamp: 'dtl 11/9/2010 19:38'!
findSubstring: key in: body startingAt: start matchTable: matchTable
	"Answer the index in the string body at which the substring key first occurs, at or beyond start.  The match is determined using matchTable, which can be used to effect, eg, case-insensitive matches.  If no match is found, zero will be returned."

	<primitive: 'primitiveFindSubstring' module: 'MiscPrimitivePlugin'>
	self primitiveFailed
! !

!IncludedMethodsTest methodsFor: 'primitives' stamp: 'dtl 11/10/2010 20:03'!
indexOfAscii: anInteger inString: aString startingAt: start

	<primitive: 'primitiveIndexOfAsciiInString' module: 'MiscPrimitivePlugin'>
	self primitiveFailed
! !

!IncludedMethodsTest methodsFor: 'primitives' stamp: 'dtl 11/10/2010 20:36'!
mixSampleCount: n into: aSoundBuffer startingAt: startIndex leftVol: leftVol rightVol: rightVol
	"Play samples from a wave table by stepping a fixed amount through the table on every sample. The table index and increment are scaled to allow fractional increments for greater pitch accuracy."
	"(FMSound pitch: 440.0 dur: 1.0 loudness: 0.5) play"

	<primitive:'primitiveMixFMSound' module:'SoundGenerationPlugin'>
	self primitiveFailed
! !

!IncludedMethodsTest methodsFor: 'testing - MiscPrimitivePlugin' stamp: 'dtl 11/9/2010 20:00'!
testCompareWithCollated
	"Verify that primitive exists in the VM"

	self assert: 3 = (self compare: 'foo' with: 'bar' collated: ((0 to: 255) as: ByteArray))
! !

!IncludedMethodsTest methodsFor: 'testing - MiscPrimitivePlugin' stamp: 'dtl 11/10/2010 08:03'!
testCompressToByteArray

	| bitmap byteArray |
	bitmap := Bitmap with: 16rFFFFFFFF.
	byteArray := ByteArray new:  4.
	self compress: bitmap toByteArray: byteArray.
	self should: byteArray = #[1 5 255 0]! !

!IncludedMethodsTest methodsFor: 'testing - MiscPrimitivePlugin' stamp: 'dtl 11/10/2010 20:14'!
testConvert8bitSignedFromTo16Bit
	"SampledSound class>>convert8bitSignedFrom:to16Bit:"


	| aByteArray aSoundBuffer |
	aByteArray := #[1 2 3 4 5 6 7 8 9].
	aSoundBuffer := SoundBuffer newMonoSampleCount: aByteArray size.
	self convert8bitSignedFrom: aByteArray to16Bit: aSoundBuffer.
	self assert: aSoundBuffer = ((SoundBuffer new: 10) at: 1 put: 256; at: 2 put: 512;
		at: 3 put: 768; at: 4 put: 1024; at: 5 put: 1280; at: 6 put: 1536; at: 7 put: 1792;
		at: 8 put: 2048; at: 9 put: 2304; at: 10 put: 0; yourself)! !

!IncludedMethodsTest methodsFor: 'testing - MiscPrimitivePlugin' stamp: 'dtl 11/10/2010 20:00'!
testDecompressFromByteArrayAt

	| bitmap byteArray s size |
	byteArray := #(1 5 255  0) asByteArray.
	s := ReadStream on: byteArray.
	size := Bitmap decodeIntFrom: s.
	bitmap := Bitmap new: size.
	self decompress: bitmap fromByteArray: byteArray at: s position + 1.
	self should: bitmap = ((Bitmap new: 1) at: 1 put: 4294967295; yourself)! !

!IncludedMethodsTest methodsFor: 'testing - MiscPrimitivePlugin' stamp: 'dtl 11/9/2010 20:54'!
testFindFirstInStringInSetStartingAt

	| position set |
	set := ((0 to: 255) collect: [:e | (e \\ 2) + $0 asciiValue]) asByteArray.
	position := self findFirstInString: 'abcdef' inSet: set startingAt: 1.
	self assert: position = 1
! !

!IncludedMethodsTest methodsFor: 'testing - MiscPrimitivePlugin' stamp: 'dtl 11/9/2010 20:14'!
testFindSubstring
	"Verify that primitive exists in the VM and that non byte array arguments cause primitive to fail"

	| position |
	position := IncludedMethodsTest new
				findSubstring: 'bc'
				in: 'abcdef'
				startingAt: 1
				matchTable: ((0 to: 255)
						as: ByteArray).
	self assert: position = 2.
	self should: [IncludedMethodsTest new
				findSubstring: 'bc' asWideString
				in: 'abcdef'
				startingAt: 1
				matchTable: ((0 to: 255)
						as: ByteArray)]
					raise: Error.
	self should: [IncludedMethodsTest new
				findSubstring: 'bc'
				in: 'abcdef' asWideString
				startingAt: 1
				matchTable: ((0 to: 255)
						as: ByteArray)]
					raise: Error.
	self should: [IncludedMethodsTest new
				findSubstring: 'bc' asWideString
				in: 'abcdef' asWideString
				startingAt: 1
				matchTable: ((0 to: 255)
						as: ByteArray)]
					raise: Error
! !

!IncludedMethodsTest methodsFor: 'testing - MiscPrimitivePlugin' stamp: 'dtl 11/9/2010 20:57'!
testIindexOfAsciiInStringStartingAt

	| position |
	position := self indexOfAscii: 50 inString: '012345' startingAt: 1.
	self assert: position = 3! !

!IncludedMethodsTest methodsFor: 'testing - SoundGeneratorPlugin' stamp: 'dtl 11/10/2010 20:45'!
testMixSampleCountIntoStartingAtLeftVolRightVol

	"mixSampleCount: n into: aSoundBuffer startingAt: startIndex leftVol: leftVol rightVol: rightVol"! !

!IncludedMethodsTest methodsFor: 'testing - MiscPrimitivePlugin' stamp: 'dtl 11/9/2010 20:33'!
testTranslateFromToTable
	"Verify that primitive exists in the VM"

	| s t |
	s := 'foo' copy. "copy so string is instantiated each time"
	t := ByteArray withAll: ((1 to: 255) as: ByteArray).
	self translate: s from: 1 to: 3 table: t.
	self assert: s = 'gpp'
! !

!IncludedMethodsTest methodsFor: 'testing - ADPCMCodecPlugin' stamp: 'dtl 11/9/2010 21:02'!
todoForADPCMCodecPlugin
	"TODO - write tests for these"

	^#(
		(ADPCMCodec privateDecodeMono:)
		(ADPCMCodec privateDecodeStereo:)
		(ADPCMCodec privateEncodeMono:)
		(ADPCMCodec privateEncodeStereo:)
		(ADPCMCodec indexForDeltaFrom:to:)
		(ADPCMCodec nextBits:)
		(ADPCMCodec nextBits:put:)
		)! !

!IncludedMethodsTest methodsFor: 'testing - SoundGeneratorPlugin' stamp: 'dtl 11/9/2010 21:02'!
todoForSoundGeneratorPlugin
	"TODO - write tests for these"

	^#(
		(FMSound mixSampleCount:into:startingAt:leftVol:rightVol:)
		(PluckedSound mixSampleCount:into:startingAt:leftVol:rightVol:)
		(LoopedSampledSound mixSampleCount:into:startingAt:leftVol:rightVol:)
		(SampledSound mixSampleCount:into:startingAt:leftVol:rightVol:)
		(ReverbSound applyReverbTo:startingAt:count:)
		)! !

!IncludedMethodsTest methodsFor: 'primitives' stamp: 'dtl 11/10/2010 20:03'!
translate: aString from: start  to: stop  table: table
	"translate the characters in the string by the given table, in place"

	<primitive: 'primitiveTranslateStringWithTable' module: 'MiscPrimitivePlugin'>
	self primitiveFailed! !

TestCase subclass: #InterpreterPrimitivesTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!InterpreterPrimitivesTest commentStamp: 'dtl 7/27/2011 11:32' prior: 0!
InterpreterPrimitivesTest provides test coverage for certain potentially problematic
primitives and methods in the interpreter.!

!InterpreterPrimitivesTest methodsFor: 'testing' stamp: 'dtl 9/22/2012 16:06'!
expectedFailures
	^#("testPrimitiveSubtractLargeIntegersParameterBounds")! !

!InterpreterPrimitivesTest methodsFor: 'testing - primitiveSubtractLargeIntegers' stamp: 'dtl 9/22/2012 16:07'!
testPrimitiveSubtractLargeIntegersParameterBounds
	"Prior to VMMaker VMMaker 4.10.1 enhancement by Nicolas Cellier, range was
	restricted and this test would fail."

	| intOne intTwo |
	"LargeNegativeInteger at lower range boundary"
	intOne := -16r8000000000000000. "maximum negative 64 bit value"
	intTwo := -16r7FFFFFFFFFFFFFFF.
	self shouldnt: [intTwo minus64: intOne]
		raise: Error.
! !

!InterpreterPrimitivesTest methodsFor: 'testing - signed64BitValueOf' stamp: 'dtl 9/22/2012 15:53'!
testSigned64BitValueOfLargeNegativeInteger
	"Exercise #signed64BitValueOf: using a LargeNegativeInteger at the maximum of
	its range. Note, IntegerArrayTest provides coverage of 32-bit equivalent."

	| intOne diff intTwo |
	"LargeNegativeInteger at lower range boundary"
	intOne := -16r8000000000000000. "maximum negative 64 bit value"
	intTwo := -16r7FFFFFFFFFFFFFFF.
	self shouldnt: [diff := intOne minus64: intTwo]
		raise: Error.
	self assert: diff = -1.

	"See testPrimitiveSubtractLargeIntegersParameterBounds"
	"self shouldnt: [diff := intTwo minus64: intOne]
		raise: Error.
	self assert: diff = 1."

	self shouldnt: [diff := intOne minus64: -1]
		raise: Error.
	self assert: diff = intTwo.
	
	"Parameters exceeding allowable range"
	"Prior to VMMaker VMMaker 4.10.1 enhancement by Nicolas Cellier, range was
	restricted and the following would raise errors."
	intOne := -16r8000000000000000 - 1. "exceed most negative 64 bit twos complement"
	intTwo := -16r7FFFFFFFFFFFFFFF.
	diff := intOne minus64: intTwo.
	self assert: diff = -2.
	diff := intTwo minus64: intOne.
	self assert: diff = 2.
	"Full 64 bit magnitude range is now available"
	intOne := -16rFFFFFFFFFFFFFFFF. ""
	intTwo := -16rFFFFFFFFFFFFFFFE.
	diff := intOne minus64: intTwo.
	self assert: diff = -1.
	diff := intTwo minus64: intOne.
	self assert: diff = 1.
	intOne := 16rFFFFFFFFFFFFFFFF. ""
	intTwo := 16rFFFFFFFFFFFFFFFE.
	diff := intOne minus64: intTwo.
	self assert: diff = 1.
	diff := intTwo minus64: intOne.
	self assert: diff = -1.
	"Out of range"
	intOne := -16rFFFFFFFFFFFFFFFF - 1. "exceed most negative 64 bit unsigned magnitude"
	intTwo := -16rFFFFFFFFFFFFFFFF.
	self should: [intOne minus64: intTwo]
		raise: Error.
	self should: [intTwo minus64: intOne]
		raise: Error.
! !

!InterpreterPrimitivesTest methodsFor: 'testing - signed64BitValueOf' stamp: 'dtl 9/22/2012 16:04'!
testSigned64BitValueOfLargePositiveInteger
	"Exercise #signed64BitValueOf: using a LargePositiveInteger at the maximum of
	its range. Note, IntegerArrayTest provides coverage of 32-bit equivalent."

	| intOne diff intTwo |
	"LargePositiveInteger at upper range boundary"
	intOne := 16r7FFFFFFFFFFFFFFF. "maximum 64 bit positive"
	intTwo := 16r7FFFFFFFFFFFFFFE.
	self shouldnt: [diff := intOne minus64: intTwo]
		raise: Error.
	self assert: diff = 1.
	self shouldnt: [diff := intTwo minus64: intOne]
		raise: Error.
	self assert: diff = -1.
	self shouldnt: [diff := intOne minus64: 1]
		raise: Error.
	self assert: diff = intTwo.
	
	"Parameters exceeding allowable range"
	"Prior to VMMaker VMMaker 4.10.1 enhancement by Nicolas Cellier, range was
	restricted and the following would raise errors."
	intOne := 16r7FFFFFFFFFFFFFFF + 1.  "exceed maximum 64 bit twos complement positive"
	intTwo := 16r7FFFFFFFFFFFFFFE.
	diff := intOne minus64: intTwo.
	self should: diff = 2.
	diff := intTwo minus64: intOne.
	self should: diff = -2.

	"Full 64 bit magnitude range is now available"
	intOne := 16rFFFFFFFFFFFFFFFF.
	intTwo := 16rFFFFFFFFFFFFFFFE.
	diff := intOne minus64: intTwo.
	self should: diff = 1.
	diff := intTwo minus64: intOne.
	self should: diff = -1.

	intOne := 16rFFFFFFFFFFFFFFFF + 1. "exceed maximum 64 bit unsigned magnitude"
	intTwo := 16rFFFFFFFFFFFFFFFE.
	self should: [intOne minus64: intTwo]
		raise: Error.
	self should: [intTwo minus64: intOne]
		raise: Error.
! !

TestCase subclass: #JPEGReadWriter2PluginTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!JPEGReadWriter2PluginTest commentStamp: 'dtl 10/9/2010 12:22' prior: 0!
Unit tests for JPEGReadWriter2Plugin. Initially this is a single test that documents a bug report, other tests may be added over time.
!

!JPEGReadWriter2PluginTest methodsFor: 'private' stamp: 'dtl 11/2/2010 13:15'!
formBytes
	"A jpeg that was generated from Squeak. When written to and from a
	stream, the same bytes should be obtained."

	^ #(255 216 255 224 0 16 74 70 73 70 0 1 1 0 0 1 0 1 0 0 255 219 0 67 0 8 6 6 7 6 5 8 7 7 7 9 9 8 10 12 20 13 12 11 11 12 25 18 19 15 20 29 26 31 30 29 26 28 28 32 36 46 39 32 34 44 35 28 28 40 55 41 44 48 49 52 52 52 31 39 57 61 56 50 60 46 51 52 50 255 219 0 67 1 9 9 9 12 11 12 24 13 13 24 50 33 28 33 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 50 255 192 0 17 8 0 1 0 1 3 1 34 0 2 17 1 3 17 1 255 196 0 31 0 0 1 5 1 1 1 1 1 1 0 0 0 0 0 0 0 0 1 2 3 4 5 6 7 8 9 10 11 255 196 0 181 16 0 2 1 3 3 2 4 3 5 5 4 4 0 0 1 125 1 2 3 0 4 17 5 18 33 49 65 6 19 81 97 7 34 113 20 50 129 145 161 8 35 66 177 193 21 82 209 240 36 51 98 114 130 9 10 22 23 24 25 26 37 38 39 40 41 42 52 53 54 55 56 57 58 67 68 69 70 71 72 73 74 83 84 85 86 87 88 89 90 99 100 101 102 103 104 105 106 115 116 117 118 119 120 121 122 131 132 133 134 135 136 137 138 146 147 148 149 150 151 152 153 154 162 163 164 165 166 167 168 169 170 178 179 180 181 182 183 184 185 186 194 195 196 197 198 199 200 201 202 210 211 212 213 214 215 216 217 218 225 226 227 228 229 230 231 232 233 234 241 242 243 244 245 246 247 248 249 250 255 196 0 31 1 0 3 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 1 2 3 4 5 6 7 8 9 10 11 255 196 0 181 17 0 2 1 2 4 4 3 4 7 5 4 4 0 1 2 119 0 1 2 3 17 4 5 33 49 6 18 65 81 7 97 113 19 34 50 129 8 20 66 145 161 177 193 9 35 51 82 240 21 98 114 209 10 22 36 52 225 37 241 23 24 25 26 38 39 40 41 42 53 54 55 56 57 58 67 68 69 70 71 72 73 74 83 84 85 86 87 88 89 90 99 100 101 102 103 104 105 106 115 116 117 118 119 120 121 122 130 131 132 133 134 135 136 137 138 146 147 148 149 150 151 152 153 154 162 163 164 165 166 167 168 169 170 178 179 180 181 182 183 184 185 186 194 195 196 197 198 199 200 201 202 210 211 212 213 214 215 216 217 218 226 227 228 229 230 231 232 233 234 242 243 244 245 246 247 248 249 250 255 218 0 12 3 1 0 2 17 3 17 0 63 0 40 162 138 0 255 217) asByteArray
! !

!JPEGReadWriter2PluginTest methodsFor: 'testing' stamp: 'dtl 10/9/2010 12:28'!
testPluginPresent

	self assert: JPEGReadWriter2 new isPluginPresent! !

!JPEGReadWriter2PluginTest methodsFor: 'testing' stamp: 'dtl 10/9/2010 14:13'!
testPrimJPEGReadImage

	| form |
	form := JPEGReadWriter2 formFromStream: self formBytes readStream.
	self should: form bits first = 16rFF808080

"[Vm-dev] JPEGReadWriter2Plugin
Bert Freudenberg bert at freudenbergs.de
Mon Sep 27 15:22:43 UTC 2010

Hi,

have there been changes lately to the jpeg plugin?

On a Mac 4.2.4 VM I see an off-by-one error, all pixels are moved to the right, the first pixel is transparent. On a Mac 5.8.1 Cog VM it works as expected though.

Here's my little test:

(JPEGReadWriter2 formFromStream: #[16rFF 16rD8 16rFF 16rE0 16r00 16r10 16r4A 16r46 16r49 16r46 16r00 16r01 16r01 16r01 16r00 16r48 16r00 16r48 16r00 16r00 16rFF 16rDB 16r00 16r43 16r00 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rFF 16rC2 16r00 16r0B 16r08 16r00 16r01 16r00 16r01 16r01 16r01 16r11 16r00 16rFF 16rC4 16r00 16r14 16r10 16r01 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16r00 16rFF 16rDA 16r00 16r08 16r01 16r01 16r00 16r01 16r3F 16r10] readStream) bits first hex

On 5.8.11 I get 16rFF808080 as expected, on 4.2.4 I get 0.
"! !

!JPEGReadWriter2PluginTest methodsFor: 'testing' stamp: 'dtl 10/9/2010 14:14'!
testPrimJPEGWriteImage

	| form bytes ws |
	form := JPEGReadWriter2 formFromStream: self formBytes readStream.
	ws := WriteStream on: ByteArray new.
	JPEGReadWriter2 putForm: form onStream: ws.
	bytes := ws contents.
	self should: self formBytes = bytes
! !

TestCase subclass: #MemoryTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!MemoryTests class methodsFor: 'debugger' stamp: 'eem 10/2/2015 11:22'!
defaultIntegerBaseInDebugger
	^16! !

!MemoryTests methodsFor: 'test support' stamp: 'eem 11/30/2018 15:38'!
deepStack: n
	"Not tail recursive."
	^0 = n ifTrue: [0] ifFalse: [n + (self deepStack: n - 1) + n]! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 12/5/2017 10:59'!
expectedFailures
	"As yet we don't support Spur on any big endian platforms, let alone 64-bit ones."
	^#(testBitmap64BitLongs)! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 9/30/2015 16:38'!
testBitmap32BitLongs
	"Test that Bitmap provides big-endian access for 32-bit accessors"
	| memory |
	memory := Bitmap new: 64.
	0 to: 30 do:
		[:shift|
		#(-1 1) do:
			[:initial| | value |
			value := initial bitShift: shift.
			memory longAt: 1 put: value.
			self assert: value equals: (memory longAt: 1).
			memory
				longAt: 5 put: 16r00005555;
				longAt: 9 put: 16r55550000.
			self assert: 16r55555555 equals: (memory longAt: 7).
			memory longAt: 7 put: value.
			self assert: (memory longAt: 7) equals: value.
			self assert: (memory at: (shift <= 15 = (initial >= 0) ifTrue: [5] ifFalse: [9])) equals: 0]].
	31 to: 32 do:
		[:shift|
		self should: [memory longAt: 1 put: -1 << shift - 1] raise: Error.
		self should: [memory longAt: 1 put: 1 << shift] raise: Error].
	0 to: 31 do:
		[:shift| | value |
		value := 1 bitShift: shift.
		memory unsignedLongAt: 1 put: value.
		self assert: value equals: (memory unsignedLongAt: 1).
			memory
				longAt: 5 put: 16r00005555;
				longAt: 9 put: 16r55550000.
			self assert: 16r55555555 equals: (memory longAt: 7).
		memory unsignedLongAt: 7 put: value.
		self assert: value equals: (memory unsignedLongAt: 7).
		self assert: 0 equals: (memory at: (shift <= 15 ifTrue: [5] ifFalse: [9]))].
	self should: [memory unsignedLongAt: 1 put: -1] raise: Error.
	32 to: 33 do:
		[:shift|
		self should: [memory unsignedLongAt: 1 put: 1 << shift] raise: Error]! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 10/2/2015 11:53'!
testBitmap64BitLongs
	"Test that Bitmap provides big-endian access for 64-bit accessors"
	| memory |
	memory := Bitmap new: 64.
	0 to: 62 do:
		[:shift|
		#(-1 1) do:
			[:initial| | value |
			memory atAllPut: 0.
			value := initial bitShift: shift.
			memory long64At: 1 put: value.
			self assert: value equals: (memory long64At: 1).
			memory
				long64At: 10 put: 16r0000000000555555;
				long64At: 18 put: 16r5555555555000000.
			self assert: 16r5555555555555555 equals: (memory long64At: 15).
			"(1 to: 7) collect: [:i| (memory at: i) hex]"
			memory long64At: 13 put: value.
			self assert: value equals: (memory long64At: 13).
			self assert: 0 equals: (memory at: (shift <= 15 = (initial >= 0) ifTrue: [9] ifFalse: [17])).
			self assert: 0 equals: (memory at: (shift <= 15 = (initial >= 0) ifTrue: [13] ifFalse: [21]))]].
	63 to: 64 do:
		[:shift|
		self should: [memory long64At: 1 put: -1 << shift - 1] raise: Error.
		self should: [memory long64At: 1 put: 1 << shift] raise: Error].
	0 to: 63 do:
		[:shift| | value |
		value := 1 bitShift: shift.
		memory unsignedLong64At: 1 put: value.
		self assert: value equals: (memory unsignedLong64At: 1).
			memory
				unsignedLong64At: 10 put: 16r0000000000555555;
				unsignedLong64At: 18 put: 16r5555555555000000.
			self assert: 16r5555555555555555 equals: (memory unsignedLong64At: 15).
		memory unsignedLong64At: 7 put: value.
		self assert: value equals: (memory unsignedLong64At: 7).
		self assert: 0 equals: (memory at: (shift <= 31 ifTrue: [9] ifFalse: [17])).
		self assert: 0 equals: (memory at: (shift <= 31  ifTrue: [13] ifFalse: [21]))].
	self should: [memory unsignedLong64At: 1 put: -1] raise: Error.
	64 to: 65 do:
		[:shift|
		self should: [memory unsignedLong64At: 1 put: 1 << shift] raise: Error]! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 12/1/2017 11:02'!
testByteArray16BitShorts
	"Test that ByteArray provides little-endian access for 16-bit accessors"
	self testLittleEndianShortAccessFor: ByteArray! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 12/1/2017 15:48'!
testByteArray32BitLongs
	"Test that ByteArray provides big-endian access for 32-bit accessors"
	self testLittleEndian32BitLongAccessFor: ByteArray! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 12/1/2017 15:41'!
testByteArray64BitLongs
	"Test that ByteArray provides big-endian access for 64-bit accessors"
	self testLittleEndian64BitLongAccessFor: ByteArray! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 12/1/2017 11:02'!
testDoubleWordArray16BitShorts
	"Test that DoubleWordArray provides little-endian access for 16-bit accessors"
	self testLittleEndianShortAccessFor: DoubleWordArray! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 12/1/2017 15:48'!
testDoubleWordArray32BitLongs
	"Test that DoubleWordArray provides little-endian access for 32-bit accessors"
	self testLittleEndian32BitLongAccessFor: DoubleWordArray! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 12/1/2017 15:40'!
testDoubleWordArray64BitLongs
	"Test that DoubleWordArray provides little-endian access for 64-bit accessors"
	self testLittleEndian64BitLongAccessFor: DoubleWordArray! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 11/30/2018 15:39'!
testFrameActivationLeak
	"This test tests if a remarried context hides a reference to an object... Thanks to Ryan Macnak for the test."
	| array object |
	array := WeakArray new: 1.
	object := Object new.

	array at: 1 put: object.
	self assert: (array at: 1) == object.

	Smalltalk garbageCollect.

	self assert: (array at: 1) == object.
	
	"Trigger stack overflow, causing this frame to be flushed to an activation. When control returns here, a new frame will be created for the activation."
	self assert: (self deepStack: 4096) = 16781312.

	"Clears the temporary in the frame."
	object := nil.
	Smalltalk garbageCollect.

	"Check the activation is not retaining a copy of our cleared temporary."
	self assert: (array at: 1) == nil.! !

!MemoryTests methodsFor: 'test support' stamp: 'eem 12/19/2017 09:41'!
testLittleEndian32BitLongAccessFor: aClass
	"Test that ByteArray provides big-endian access for 32-bit accessors"
	"memory copy changeClassTo: ByteArray"
	| memory |
	memory := aClass new: 64 / (aClass new: 0) bytesPerElement.
	"signed access"
	0 to: 30 do:
		[:shift|
		#(-1 1) do:
			[:initial| | value |
			value := initial bitShift: shift.
			memory longAt: 1 put: value.
			self assert: value equals: (memory longAt: 1).
			"aligned store"
			memory
				longAt: 5 put: 16r55550000;
				longAt: 9 put: 16r00005555.
			self assert: 16r55555555 equals: (memory longAt: 7).
			memory longAt: 7 put: value.
			self assert: value equals: (memory longAt: 7).
			"check bytes either side have not been changed"
			5 to: 6 do:
				[:i| self assert: 0 equals: (memory byteAt: i)].
			11 to: 15 do:
				[:i| self assert: 0 equals: (memory byteAt: i)].
			"unaligned store"
			memory
				longAt: 6 put: 16r55550000;
				longAt: 10 put: 16r00005555.
			self assert: 16r55555555 equals: (memory longAt: 8).
			memory longAt: 8 put: value.
			self assert: value equals: (memory longAt: 8).
			"check bytes either side have not been changed"
			5 to: 7 do:
				[:i| self assert: 0 equals: (memory byteAt: i)].
			12 to: 15 do:
				[:i| self assert: 0 equals: (memory byteAt: i)]]].
	memory
		unsignedLongAt: 49 put: 16rAAAAAAAA;
		unsignedLongAt: 53 put: 16rAAAAAAAA.
	49 to: 52 do:
		[:i|
		 self assert: -1431655766 equals: (memory longAt: i)].
	"unsigned access"
	0 to: 31 do:
		[:shift| | value |
		value := 1 bitShift: shift.
		memory unsignedLongAt: 1 put: value.
		self assert: value equals: (memory unsignedLongAt: 1).
		"aligned store"
		memory
			longAt: 5 put: 16r55550000;
			longAt: 9 put: 16r00005555.
		self assert: 16r55555555 equals: (memory unsignedLongAt: 7).
		memory unsignedLongAt: 7 put: value.
		self assert: value equals: (memory unsignedLongAt: 7).
		"check bytes either side have not been changed"
		5 to: 6 do:
			[:i| self assert: 0 equals: (memory byteAt: i)].
		11 to: 15 do:
			[:i| self assert: 0 equals: (memory byteAt: i)].
		"unaligned store"
		memory
			longAt: 6 put: 16r55550000;
			longAt: 10 put: 16r00005555.
		self assert: 16r55555555 equals: (memory unsignedLongAt: 8).
		memory unsignedLongAt: 8 put: value.
		self assert: value equals: (memory unsignedLongAt: 8).
		"check bytes either side have not been changed"
		5 to: 7 do:
			[:i| self assert: 0 equals: (memory byteAt: i)].
		12 to: 15 do:
			[:i| self assert: 0 equals: (memory byteAt: i)]]! !

!MemoryTests methodsFor: 'test support' stamp: 'eem 12/19/2017 09:40'!
testLittleEndian64BitLongAccessFor: aClass
	"Test that the given class provides little-endian access for 64-bit accessors"
	"memory copy changeClassTo: ByteArray"
	| memory |
	memory := aClass new: 64 / (aClass new: 0) bytesPerElement.
	"signed access"
	0 to: 62 do:
		[:shift|
		#(-1 1) do:
			[:initial| | value |
			value := initial bitShift: shift.
			memory long64At: 1 put: value.
			self assert: value equals: (memory long64At: 1).
			"aligned store"
			memory
				long64At: 9 put: 16r5555550000000000;
				long64At: 17 put: 16r0000005555555555.
			self assert: 16r5555555555555555 equals: (memory long64At: 14).
			memory long64At: 14 put: value.
			self assert: value equals: (memory long64At: 14).
			"check bytes either side have not been changed"
			9 to: 13 do:
				[:i| self assert: 0 equals: (memory byteAt: i)].
			22 to: 31 do:
				[:i| self assert: 0 equals: (memory byteAt: i)].
			"unaligned store"
			memory
				long64At: 10 put: 16r5555550000000000;
				long64At: 18 put: 16r0000005555555555.
			self assert: 16r5555555555555555 equals: (memory long64At: 15).
			memory long64At: 15 put: value.
			self assert: value equals: (memory long64At: 15).
			"check bytes either side have not been changed"
			9 to: 14 do:
				[:i| self assert: 0 equals: (memory byteAt: i)].
			23 to: 31 do:
				[:i| self assert: 0 equals: (memory byteAt: i)]]].
	memory
		unsignedLong64At: 49 put: 16rAAAAAAAAAAAAAAAA;
		unsignedLong64At: 57 put: 16rAAAAAAAAAAAAAAAA.
	49 to: 56 do:
		[:i|
		 self assert: -6148914691236517206 equals: (memory long64At: i)].
	"unsigned access"
	0 to: 63 do:
		[:shift| | value |
		value := 1 bitShift: shift.
		memory unsignedLong64At: 1 put: value.
		self assert: value equals: (memory unsignedLong64At: 1).
		"aligned store"
		memory
			unsignedLong64At: 9 put: 16r5555550000000000;
			unsignedLong64At: 17 put: 16r0000005555555555.
		self assert: 16r5555555555555555 equals: (memory unsignedLong64At: 14).
		memory unsignedLong64At: 14 put: value.
		self assert: value equals: (memory unsignedLong64At: 14).
		"check bytes either side have not been changed"
		9 to: 13 do:
			[:i| self assert: 0 equals: (memory byteAt: i)].
		22 to: 31 do:
			[:i| self assert: 0 equals: (memory byteAt: i)].
		"unaligned store"
		memory
			unsignedLong64At: 10 put: 16r5555550000000000;
			unsignedLong64At: 18 put: 16r0000005555555555.
		self assert: 16r5555555555555555 equals: (memory unsignedLong64At: 15).
		memory unsignedLong64At: 15 put: value.
		self assert: value equals: (memory unsignedLong64At: 15).
		"check bytes either side have not been changed"
		9 to: 14 do:
			[:i| self assert: 0 equals: (memory byteAt: i)].
		23 to: 31 do:
			[:i| self assert: 0 equals: (memory byteAt: i)]]! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 12/1/2017 11:02'!
testLittleEndianBitmap16BitShorts
	"Test that LittleEndianBitmap provides little-endian access for 16-bit accessors"
	self testLittleEndianShortAccessFor: LittleEndianBitmap! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 12/1/2017 15:48'!
testLittleEndianBitmap32BitLongs
	"Test that LittleEndianBitmap provides little-endian access for 32-bit accessors"
	self testLittleEndian32BitLongAccessFor: LittleEndianBitmap! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 12/1/2017 15:41'!
testLittleEndianBitmap64BitLongs
	"Test that LittleEndianBitmap provides little-endian access for 64-bit accessors"
	self testLittleEndian64BitLongAccessFor: LittleEndianBitmap! !

!MemoryTests methodsFor: 'test support' stamp: 'eem 12/1/2017 14:00'!
testLittleEndianShortAccessFor: aClass
	| memory |
	1 to: 9 by: 2 do:
		[:i|
		memory := aClass new: 64 / (aClass new: 0) bytesPerElement.
		memory unsignedShortAt: i put: 16r7FFF.
		1 to: i - 1 do:
			[:j| self assert: 0 equals: (memory byteAt: j)].
		self assert: 16rFF equals: (memory byteAt: i).
		self assert: 16r7F equals: (memory byteAt: i + 1).
		(i + 2 to: 8) do:
			[:j| self assert: 0 equals: (memory byteAt: j)]].

	16 to: 32 do:
		[:shift|
		self should: [memory unsignedShortAt: 1 put: -1 << shift - 1] raise: Error.
		self should: [memory unsignedShortAt: 1 put: 1 << shift] raise: Error].
! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 12/1/2017 15:10'!
testSignedOutOfRangeAccess
	{ByteArray. Bitmap. LittleEndianBitmap. DoubleWordArray} do:
		[:class| | bytesPerElement memory |
		bytesPerElement := (class new: 0) bytesPerElement.
		memory := class new: 64 / bytesPerElement.
		1 to: 16 do:
			[:i|
			self should: [memory byteAt: i put: 1 << 8] raise: Error.
			self should: [memory byteAt: i put: -1 << 7 - 1] raise: Error.
			self should: [memory shortAt: i put: 1 << 16] raise: Error.
			self should: [memory shortAt: i put: -1 << 15 - 1] raise: Error.
			self should: [memory longAt: i put: 1 << 32] raise: Error.
			self should: [memory longAt: i put: -1 << 31 - 1] raise: Error.
			self should: [memory long64At: i put: 1 << 64] raise: Error.
			self should: [memory long64At: i put: -1 << 63 - 1] raise: Error].
		1 to: memory size do:
			[:i|
			self assert: 0 equals: (memory at: i)]]! !

!MemoryTests methodsFor: 'tests' stamp: 'eem 12/1/2017 15:11'!
testUnsignedOutOfRangeAccess
	{ByteArray. Bitmap. LittleEndianBitmap. DoubleWordArray} do:
		[:class| | bytesPerElement memory |
		bytesPerElement := (class new: 0) bytesPerElement.
		memory := class new: 64 / bytesPerElement.
		1 to: 16 do:
			[:i|
			self should: [memory unsignedByteAt: i put: 1 << 8] raise: Error.
			self should: [memory unsignedByteAt: i put: -1] raise: Error.
			self should: [memory unsignedShortAt: i put: 1 << 16] raise: Error.
			self should: [memory unsignedShortAt: i put: -1] raise: Error.
			self should: [memory unsignedLongAt: i put: 1 << 32] raise: Error.
			self should: [memory unsignedLongAt: i put: -1] raise: Error.
			self should: [memory unsignedLong64At: i put: 1 << 64] raise: Error.
			self should: [memory unsignedLong64At: i put: -1] raise: Error].
		1 to: memory size do:
			[:i|
			self assert: 0 equals: (memory at: i)]]! !

TestCase subclass: #MiscPrimitivePluginTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!MiscPrimitivePluginTest commentStamp: 'dtl 11/10/2010 22:04' prior: 0!
Various classes in the image contain methods intended for translation to C for execution as primitives in MiscPrimitivePlugin. MiscPrimitivePluginTest provides tests to validate these methods.
!

!MiscPrimitivePluginTest methodsFor: 'primitives' stamp: 'ul 5/20/2018 17:08'!
compare: string1 with: string2 collated: order
	"Return 1, 2 or 3, if string1 is <, =, or > string2, with the collating order of characters given by the order array."

	<primitive: 'primitiveCompareString' module: 'MiscPrimitivePlugin'  error: ec>
	^ec! !

!MiscPrimitivePluginTest methodsFor: 'primitives' stamp: 'ul 5/20/2018 17:08'!
compress: bm toByteArray: ba

	<primitive: 'primitiveCompressToByteArray' module: 'MiscPrimitivePlugin' error: ec>
	^ec! !

!MiscPrimitivePluginTest methodsFor: 'primitives' stamp: 'ul 5/20/2018 17:08'!
convert8bitSignedFrom: aByteArray to16Bit: aSoundBuffer
	"Copy the contents of the given array of signed 8-bit samples into the given array of 16-bit signed samples."

	<primitive: 'primitiveConvert8BitSigned' module: 'MiscPrimitivePlugin' error: ec>
	^ec! !

!MiscPrimitivePluginTest methodsFor: 'primitives' stamp: 'ul 5/20/2018 17:08'!
decompress: bm fromByteArray: ba at: index

	<primitive: 'primitiveDecompressFromByteArray' module: 'MiscPrimitivePlugin' error: ec>
	^ec! !

!MiscPrimitivePluginTest methodsFor: 'testing' stamp: 'dtl 2/1/2011 19:49'!
expectedFailures

	^#(testFindSubstringOldVersionWithMissingTypeChecks
		testFindSubstringInWideString)
! !

!MiscPrimitivePluginTest methodsFor: 'primitives' stamp: 'ul 5/20/2018 17:08'!
findFirstInString: aString  inSet: inclusionMap  startingAt: start

	<primitive: 'primitiveFindFirstInString' module: 'MiscPrimitivePlugin'  error: ec>
	^ec! !

!MiscPrimitivePluginTest methodsFor: 'primitives' stamp: 'ul 5/20/2018 17:09'!
findSubstring: key in: body startingAt: start matchTable: matchTable
	"Answer the index in the string body at which the substring key first occurs, at or beyond start.  The match is determined using matchTable, which can be used to effect, eg, case-insensitive matches.  If no match is found, zero will be returned."

	<primitive: 'primitiveFindSubstring' module: 'MiscPrimitivePlugin' error: ec>
	^ec! !

!MiscPrimitivePluginTest methodsFor: 'primitives' stamp: 'ul 5/20/2018 17:09'!
hashBytes: aByteArray startingWith: speciesHash
	"Answer the hash of a byte-indexed collection,
	using speciesHash as the initial value.
	See SmallInteger>>hashMultiply.

	The primitive should be renamed at a
	suitable point in the future"

	<primitive: 'primitiveStringHash' module: 'MiscPrimitivePlugin' error: ec>
	^ec! !

!MiscPrimitivePluginTest methodsFor: 'primitives' stamp: 'ul 5/20/2018 17:09'!
indexOfAscii: anInteger inString: aString startingAt: start

	<primitive: 'primitiveIndexOfAsciiInString' module: 'MiscPrimitivePlugin' error: ec>
	^ec! !

!MiscPrimitivePluginTest methodsFor: 'testing' stamp: 'ul 5/20/2018 22:25'!
testCompareWithCollated

	| collation |
	collation := (0 to: 255) asByteArray.
	self 
		assert: #'bad argument' 
			equals: (self compare: nil with: '' collated: collation)
			description: 'First argument must be a bytes object';
		assert: #'bad argument'
			equals: (self compare: '' with: nil collated: collation)
			description: 'Second argument must be a bytes object';
		assert: #'bad argument'
			equals: (self compare: '' with: '' collated: nil)
			description: 'Third argument must be a bytes object';
		assert: #'bad argument'
			equals: (self compare: '' with: '' collated: collation allButLast)
			description: 'Size of third argument must be at least 256'.
	self
		assert: 1 equals: (self compare: '' with: 'foo' collated: collation);
		assert: 1 equals: (self compare: 'bar' with: 'foo' collated: collation);
		assert: 1 equals: (self compare: 'bar' with: 'bare' collated: collation);		
		assert: 2 equals: (self compare: '' with: '' collated: collation);
		assert: 2 equals: (self compare: 'bar' with: 'bar' collated: collation);
		assert: 2 equals: (self compare: 'bar' with: 'bar' copy collated: collation);
		assert: 3 equals: (self compare: 'foo' with: '' collated: collation);
		assert: 3 equals: (self compare: 'foo' with: 'bar' collated: collation);
		assert: 3 equals: (self compare: 'bare' with: 'bar' collated: collation).
	collation
		at: $b asInteger + 1 put: $f asInteger;
		at: $a asInteger + 1 put: $o asInteger;
		at: $r asInteger + 1 put: $o asInteger.
	self
		assert: 2 equals: (self compare: 'bar' with: 'foo' collated: collation);
		assert: 2 equals: (self compare: 'foo' with: 'bar' collated: collation)! !

!MiscPrimitivePluginTest methodsFor: 'testing' stamp: 'eem 10/29/2018 16:55'!
testCompressToByteArray

	| bitmap byteArray n |
	bitmap := Bitmap with: 16rFFFFFFFF.
	byteArray := ByteArray new:  12. "Primitive makes an estimate; hence output must have at least 11 bytes"
	n := self compress: bitmap toByteArray: byteArray.
	self assert: n = 3.
	self assert: #[1 5 255] equals: (byteArray copyFrom: 1 to: n)! !

!MiscPrimitivePluginTest methodsFor: 'testing' stamp: 'dtl 11/10/2010 20:14'!
testConvert8bitSignedFromTo16Bit
	"SampledSound class>>convert8bitSignedFrom:to16Bit:"


	| aByteArray aSoundBuffer |
	aByteArray := #[1 2 3 4 5 6 7 8 9].
	aSoundBuffer := SoundBuffer newMonoSampleCount: aByteArray size.
	self convert8bitSignedFrom: aByteArray to16Bit: aSoundBuffer.
	self assert: aSoundBuffer = ((SoundBuffer new: 10) at: 1 put: 256; at: 2 put: 512;
		at: 3 put: 768; at: 4 put: 1024; at: 5 put: 1280; at: 6 put: 1536; at: 7 put: 1792;
		at: 8 put: 2048; at: 9 put: 2304; at: 10 put: 0; yourself)! !

!MiscPrimitivePluginTest methodsFor: 'testing' stamp: 'eem 10/29/2018 16:54'!
testDecompressFromByteArrayAt

	| bitmap byteArray s size |
	byteArray := #[1 5 255].
	s := ReadStream on: byteArray.
	size := Bitmap decodeIntFrom: s.
	bitmap := Bitmap new: size.
	self decompress: bitmap fromByteArray: byteArray at: s position + 1.
	self assert: (Bitmap with: 16rFFFFFFFF) equals: bitmap! !

!MiscPrimitivePluginTest methodsFor: 'testing' stamp: 'dtl 11/9/2010 20:54'!
testFindFirstInStringInSetStartingAt

	| position set |
	set := ((0 to: 255) collect: [:e | (e \\ 2) + $0 asciiValue]) asByteArray.
	position := self findFirstInString: 'abcdef' inSet: set startingAt: 1.
	self assert: position = 1
! !

!MiscPrimitivePluginTest methodsFor: 'testing' stamp: 'ul 5/21/2018 00:56'!
testFindSubstringInByteString
	"Verify that primitive exists in the VM and works as expected for byte
	sized string and key. If key is a WideString, as may be the case if testing
	for a WideString as substring of a byte sized string, then the primitive
	should fail. Earlier version of the primitive would accept the non-bytes
	parameter, leading to incorrect results, as documented in
	testFindSubstringOldVersionWithMissingTypeChecks"

	self
		assert: 2
		equals: (self
			findSubstring: 'bc'
			in: 'abcdef'
			startingAt: 1
			matchTable: (0 to: 255) asByteArray).
	self
		assert: #'bad argument'
		equals: (self
			findSubstring: 'bc' asWideString
			in: 'abcdef'
			startingAt: 1
			matchTable: (0 to: 255) asByteArray)
! !

!MiscPrimitivePluginTest methodsFor: 'testing' stamp: 'dtl 3/17/2011 22:40'!
testFindSubstringInWideString
	"The primitive will fail if invoked on a non-byte receiver. WideString does
	not call the primitive, so this is an expected failure."

	| position |
	position := MiscPrimitivePluginTest new
				findSubstring: 'bc'
				in: 'abcdef' asWideString
				startingAt: 1
				matchTable: ((0 to: 255)
						as: ByteArray).
	self assert: position = 0.
	position := MiscPrimitivePluginTest new
				findSubstring: 'bc' asWideString
				in: 'abcdef' asWideString
				startingAt: 1
				matchTable: ((0 to: 255)
						as: ByteArray).
	self assert: position = 2
! !

!MiscPrimitivePluginTest methodsFor: 'testing' stamp: 'dtl 3/18/2011 08:52'!
testFindSubstringOldVersionWithMissingTypeChecks
	"Verify that primitive exists in the VM. This test documents the behavior of older versions
	of the primitive that did not perform type checking on to fail on WideString parameters.
	Newer versions of the primitive handle this properly, hence this test is an expectedFailure.
	The fix for this issue (by Andreas Raab) is in TMethod>>argConversionExprFor:stackIndex:
	which was added to VMMaker in VMMaker-dtl.202.
	
	If this test passes, it is an indication that the TMethod>>argConversionExprFor:stackIndex:
	update is missing from the VMMaker code generator that produced the VM."
	
	"(MiscPrimitivePluginTest selector: #testArgsAsFlatArrayAndOffsets) run"

	| position |
	"If both receiver and argument are byte size, the primitive performs correctly"
	position := MiscPrimitivePluginTest new
				findSubstring: 'bc'
				in: 'abcdef'
				startingAt: 1
				matchTable: ((0 to: 255)
						as: ByteArray).
	self assert: position = 2.

	"Plugin accepts wide string argument, but should not do so."
	position := MiscPrimitivePluginTest new
				findSubstring: 'bc' asWideString
				in: 'abcdef'
				startingAt: 1
				matchTable: ((0 to: 255)
						as: ByteArray).
	self assert: position = 0.

	position := MiscPrimitivePluginTest new
				findSubstring: 'bc'
				in: 'abcdef' asWideString
				startingAt: 1
				matchTable: ((0 to: 255)
						as: ByteArray).
	self assert: position = 0.

	"Older versions of the plugin produce incorrect results here"
	position := MiscPrimitivePluginTest new
				findSubstring: 'bc' asWideString
				in: 'abcdef' asWideString
				startingAt: 1
				matchTable: ((0 to: 255)
						as: ByteArray).
	self assert: position = 5 "Incorrect!!"
! !

!MiscPrimitivePluginTest methodsFor: 'testing' stamp: 'dtl 11/10/2010 21:48'!
testHashBytesStartingWith

	| ba result |
	ba := #[1 2 3 4 5 6 7 8 9].
	result := self hashBytes: ba startingWith: 12345.
	self assert: result = 170953102
! !

!MiscPrimitivePluginTest methodsFor: 'testing' stamp: 'dtl 11/9/2010 20:57'!
testIindexOfAsciiInStringStartingAt

	| position |
	position := self indexOfAscii: 50 inString: '012345' startingAt: 1.
	self assert: position = 3! !

!MiscPrimitivePluginTest methodsFor: 'testing' stamp: 'ul 5/20/2018 22:18'!
testTranslateFromToTable
	"Verify that primitive exists in the VM"

	| s t |
	s := 'foo' copy. "copy so string is instantiated each time"
	t := (0 to: 255) asByteArray replace: [ :each | each + 1 \\ 256 ].
	self assert: self equals: (self translate: s from: 1 to: 3 table: t).
	self assert: s = 'gpp'
! !

!MiscPrimitivePluginTest methodsFor: 'primitives' stamp: 'ul 5/20/2018 17:10'!
translate: aString from: start  to: stop  table: table
	"translate the characters in the string by the given table, in place"

	<primitive: 'primitiveTranslateStringWithTable' module: 'MiscPrimitivePlugin' error: ec>
	^ec! !

TestCase subclass: #SlangAbstractTestCase
	instanceVariableNames: 'ccg'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!SlangAbstractTestCase methodsFor: 'running' stamp: ''!
setUp
	super setUp.
	ccg := CCodeGenerator new.
! !

SlangAbstractTestCase subclass: #SlangBasicTypeInferenceTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!SlangBasicTypeInferenceTest commentStamp: '' prior: 0!
Basic tests for the type inference used in CCodeGenerator.
Support methods are in SlangTypeInferenceTestClass.
There is only one possible type by node, no ambiguities.

1 Constant
	test constantNode in a method.
	tests are tailor made for the current TConstantNode >> #typeOrNilFrom:in:.

2 Return-constant
	test return a constant node in a method.
	uses the previous categorie's values.

3 Return-temp-assigned-constant
	assign a temporary node with a constant
	then returns the temporary variable in the next statement.
	
4 Return-temp-assigned-message
	assign a temporary node with a message
	then returns the temporary variable in the next statement.
	
5 Return-explicit-temp
	returns a temporary typed by a pragma #var:type:.

6 Return-explicit-arg
	returns an argument typed by a pragma #var:type:.!

!SlangBasicTypeInferenceTest methodsFor: 'running' stamp: ''!
setUp
	super setUp.
	ccg addClass: SlangBasicTypeInferenceTestClass.
	ccg inferTypesForImplicitlyTypedVariablesAndMethods.
! !

!SlangBasicTypeInferenceTest methodsFor: 'constant' stamp: ''!
testABigNegativeIntegerConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #aBigNegativeIntegerConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'long long'! !

!SlangBasicTypeInferenceTest methodsFor: 'constant' stamp: ''!
testAFalseConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #aFalseConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int! !

!SlangBasicTypeInferenceTest methodsFor: 'constant' stamp: ''!
testAFloatConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #aFloatConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #double! !

!SlangBasicTypeInferenceTest methodsFor: 'constant' stamp: ''!
testANilConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #aNilConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int! !

!SlangBasicTypeInferenceTest methodsFor: 'constant' stamp: ''!
testASmallNegativeIntegerConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #aSmallNegativeIntegerConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int! !

!SlangBasicTypeInferenceTest methodsFor: 'constant' stamp: ''!
testAStringConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #aStringConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'char *'! !

!SlangBasicTypeInferenceTest methodsFor: 'constant' stamp: ''!
testATrueConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #aTrueConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int! !

!SlangBasicTypeInferenceTest methodsFor: 'constant' stamp: ''!
testAnIntEqual32ConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #anIntEqual32ConstantNode.
		
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'unsigned int'! !

!SlangBasicTypeInferenceTest methodsFor: 'constant' stamp: ''!
testAnIntEqual64ConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #anIntEqual64ConstantNode.
		
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'unsigned long long'! !

!SlangBasicTypeInferenceTest methodsFor: 'constant' stamp: ''!
testAnIntGreater64ConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #anIntGreater64ConstantNode.
		
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'long long'! !

!SlangBasicTypeInferenceTest methodsFor: 'constant' stamp: ''!
testAnIntLess32ConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #anIntLesser32ConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int.
! !

!SlangBasicTypeInferenceTest methodsFor: 'return-constant' stamp: ''!
testReturnABigNegativeIntegerConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnABigNegativeIntegerConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'long long'.
	self assert: tMethod returnType equals: #sqLong! !

!SlangBasicTypeInferenceTest methodsFor: 'return-message-send' stamp: ''!
testReturnABigNegativeIntegerMessageSend
	| tMethod |
	tMethod := ccg methodNamed: #returnABigNegativeIntegerMessageSend.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqLong.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #sqLong.
	self assert: tMethod returnType equals: #sqLong.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-constant' stamp: ''!
testReturnAFalseConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnAFalseConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int.
	self assert: tMethod returnType equals: #sqInt! !

!SlangBasicTypeInferenceTest methodsFor: 'return-message-send' stamp: ''!
testReturnAFalseMessageSend
	| tMethod |
	tMethod := ccg methodNamed: #returnAFalseMessageSend.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-message-send' stamp: ''!
testReturnAFloatMessageSend
	| tMethod |
	tMethod := ccg methodNamed: #returnAFloatMessageSend.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #double.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #double.
	self assert: tMethod returnType equals: #double.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-constant' stamp: ''!
testReturnANilConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnANilConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int.
	self assert: tMethod returnType equals: #sqInt! !

!SlangBasicTypeInferenceTest methodsFor: 'return-message-send' stamp: ''!
testReturnANilMessageSend
	| tMethod |
	tMethod := ccg methodNamed: #returnANilMessageSend.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-constant' stamp: ''!
testReturnASmallNegativeIntegerConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnASmallNegativeIntegerConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int.
	self assert: tMethod returnType equals: #sqInt! !

!SlangBasicTypeInferenceTest methodsFor: 'return-message-send' stamp: ''!
testReturnASmallNegativeIntegerMessageSend
	| tMethod |
	tMethod := ccg methodNamed: #returnASmallNegativeIntegerMessageSend.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-constant' stamp: ''!
testReturnAStringConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnAStringConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'char *'.
	self assert: tMethod returnType equals: #'char *'! !

!SlangBasicTypeInferenceTest methodsFor: 'return-message-send' stamp: ''!
testReturnAStringMessageSend
	| tMethod |
	tMethod := ccg methodNamed: #returnAStringMessageSend.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'char *'.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #'char *'.
	self assert: tMethod returnType equals: #'char *'.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-constant' stamp: ''!
testReturnATrueConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnATrueConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int.
	self assert: tMethod returnType equals: #sqInt! !

!SlangBasicTypeInferenceTest methodsFor: 'return-message-send' stamp: ''!
testReturnATrueMessageSend
	| tMethod |
	tMethod := ccg methodNamed: #returnATrueMessageSend.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-constant' stamp: ''!
testReturnAnIntEqual32ConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnAnIntEqual32ConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'unsigned int'.
	self assert: tMethod returnType equals: #sqInt
	"returning an integer constants is the same as returning nothing"
	"this therefore return a sqInt"
	"see start of the CCodeGenerator >> harmonizeReturnTypesIn: method"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-message-send' stamp: ''!
testReturnAnIntEqual32MessageSend
	| tMethod |
	tMethod := ccg methodNamed: #returnAnIntEqual32MessageSend.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.
	"the method return type is sqint, so it's coherent, see testReturnAnIntEqual32MessageSend"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-constant' stamp: ''!
testReturnAnIntEqual64ConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnAnIntEqual64ConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'unsigned long long'.
	self assert: tMethod returnType equals: #sqInt
	"returning an integer constants is the same as returning nothing"
	"this therefore return a sqInt"
	"see start of the CCodeGenerator >> harmonizeReturnTypesIn: method"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-message-send' stamp: ''!
testReturnAnIntEqual64MessageSend
	| tMethod |
	tMethod := ccg methodNamed: #returnAnIntEqual64MessageSend.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.
	"the method return type is sqint, so it's coherent, see testReturnAnIntEqual64MessageSend"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-constant' stamp: ''!
testReturnAnIntGreater64ConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnAnIntGreater64ConstantNode.
		
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'long long'.
	self assert: tMethod returnType equals: #sqInt
	"returning an integer constants is the same as returning nothing"
	"this therefore return a sqInt"
	"see start of the CCodeGenerator >> harmonizeReturnTypesIn: method"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-message-send' stamp: ''!
testReturnAnIntGreater64MessageSend
	| tMethod |
	tMethod := ccg methodNamed: #returnAnIntGreater64MessageSend.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.
	"the method return type is sqint, so it's coherent, see testReturnAnIntGreater64MessageSend"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-constant' stamp: ''!
testReturnAnIntLess32ConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnAnIntLesser32ConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int.
	self assert: tMethod returnType equals: #sqInt
	"returning constants is the same as returning nothing"
	"this therefore return a sqInt"
	"see start of the CCodeGenerator >> harmonizeReturnTypesIn: method"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-message-send' stamp: ''!
testReturnAnIntLesser32MessageSend
	| tMethod |
	tMethod := ccg methodNamed: #returnAnIntLesser32MessageSend.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.
	"the method return type is sqint, so it's coherent, see testReturnAnIntLesser64MessageSend"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-arg' stamp: ''!
testReturnExplicitArgDouble
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitArgDouble:.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #double.
	self assert: tMethod returnType equals: #double! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-arg' stamp: ''!
testReturnExplicitArgFloat
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitArgFloat:.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #float.
	self assert: tMethod returnType equals: #float! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-arg' stamp: ''!
testReturnExplicitArgInt
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitArgInt:.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int.
	self assert: tMethod returnType equals: #sqInt! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-arg' stamp: ''!
testReturnExplicitArgLongLong
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitArgLongLong:.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'long long'.
	self assert: tMethod returnType equals: #sqLong! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-arg' stamp: ''!
testReturnExplicitArgRandom
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitArgRandom:.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #completelyRandom.
	self assert: tMethod returnType equals: #completelyRandom! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-arg' stamp: ''!
testReturnExplicitArgString
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitArgString:.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'char *'.
	self assert: tMethod returnType equals: #'char *'! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-arg' stamp: ''!
testReturnExplicitArgUnsignedInt
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitArgUnsignedInt:.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'unsigned int'.
	self assert: tMethod returnType equals: #usqInt! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-arg' stamp: ''!
testReturnExplicitArgUnsignedLongLong
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitArgUnsignedLongLong:.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'unsigned long long'.
	self assert: tMethod returnType equals: #usqLong! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-temp' stamp: ''!
testReturnExplicitTempDouble
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitTempDouble.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #double.
	self assert: tMethod returnType equals: #double! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-temp' stamp: ''!
testReturnExplicitTempFloat
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitTempFloat.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #float.
	self assert: tMethod returnType equals: #float! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-temp' stamp: ''!
testReturnExplicitTempInt
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitTempInt.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int.
	self assert: tMethod returnType equals: #sqInt! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-temp' stamp: ''!
testReturnExplicitTempLongLong
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitTempLongLong.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'long long'.
	self assert: tMethod returnType equals: #sqLong! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-temp' stamp: ''!
testReturnExplicitTempRandom
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitTempRandom.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #completelyRandom.
	self assert: tMethod returnType equals: #completelyRandom! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-temp' stamp: ''!
testReturnExplicitTempString
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitTempString.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'char *'.
	self assert: tMethod returnType equals: #'char *'! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-temp' stamp: ''!
testReturnExplicitTempUnsignedInt
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitTempUnsignedInt.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'unsigned int'.
	self assert: tMethod returnType equals: #usqInt! !

!SlangBasicTypeInferenceTest methodsFor: 'return-explicit-temp' stamp: ''!
testReturnExplicitTempUnsignedLongLong
	| tMethod |
	tMethod := ccg methodNamed: #returnExplicitTempUnsignedLongLong.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'unsigned long long'.
	self assert: tMethod returnType equals: #usqLong! !

!SlangBasicTypeInferenceTest methodsFor: 'return-constant' stamp: ''!
testReturnFloatConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnAFloatConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #double.
	self assert: tMethod returnType equals: #double
	! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-const' stamp: ''!
testReturnTempBigNegativeIntegerConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempBigNegativeIntegerConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'long long'. " assignementNode "
	self assert: (ccg typeFor: tMethod statements first value in: tMethod) equals: #'long long'. " value, constantNode "
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt. " variable, temporaryNode "
	self assert: tMethod returnType equals: #sqInt.
	
	self assert: false "unexpected type conversion, shouldn't it be a sqLong?"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-message' stamp: ''!
testReturnTempBigNegativeIntegerMessageNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempBigNegativeIntegerMessageNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqLong.
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqLong.
	self assert: tMethod returnType equals: #sqLong.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-const' stamp: ''!
testReturnTempFalseConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempFalseConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int. " assignementNode "
	self assert: (ccg typeFor: tMethod statements first value in: tMethod) equals: #int. " value, constantNode "
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt. " variable, temporaryNode "
	self assert: tMethod returnType equals: #sqInt.
	
	self assert: false "unexpected type conversion"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-message' stamp: ''!
testReturnTempFalseMessageNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempFalseMessageNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-const' stamp: ''!
testReturnTempFloatConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempFloatConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #double. " assignementNode "
	self assert: (ccg typeFor: tMethod statements first value in: tMethod) equals: #double. " value, constantNode "
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt. " variable, temporaryNode "
	self assert: tMethod returnType equals: #sqInt.
	
	self assert: false "type lost"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-message' stamp: ''!
testReturnTempFloatMessageNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempFloatMessageNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #double.
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.
	
	self assert: false "wait what"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-const' stamp: ''!
testReturnTempIntEqual32ConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempIntEqual32ConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'unsigned int'. " assignementNode "
	self assert: (ccg typeFor: tMethod statements first value in: tMethod) equals: #'unsigned int'. " value, constantNode "
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt. " variable, temporaryNode "
	self assert: tMethod returnType equals: #sqInt.
	
	self assert: false "unexpected type conversion"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-message' stamp: ''!
testReturnTempIntEqual32MessageNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempIntEqual32MessageNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-const' stamp: ''!
testReturnTempIntEqual64ConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempIntEqual64ConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'unsigned long long'. " assignementNode "
	self assert: (ccg typeFor: tMethod statements first value in: tMethod) equals: #'unsigned long long'. " value, constantNode "
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt. " variable, temporaryNode "
	self assert: tMethod returnType equals: #sqInt.
	
	self assert: false "unexpected type conversion, shouldn't it at least be an sqLong? or better, an uSqLong"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-message' stamp: ''!
testReturnTempIntEqual64MessageNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempIntEqual64MessageNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-const' stamp: ''!
testReturnTempIntGreater64ConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempIntGreater64ConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'long long'. " assignementNode "
	self assert: (ccg typeFor: tMethod statements first value in: tMethod) equals: #'long long'. " value, constantNode "
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt. " variable, temporaryNode "
	self assert: tMethod returnType equals: #sqInt.
	
	self assert: false "unexpected type conversion, shouldn't it be at least an sqLong?"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-message' stamp: ''!
testReturnTempIntGreater64MessageNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempIntGreater64MessageNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-const' stamp: ''!
testReturnTempIntLesser32ConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempIntLesser32ConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int. " assignementNode "
	self assert: (ccg typeFor: tMethod statements first value in: tMethod) equals: #int. " value, constantNode "
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt. " variable, temporaryNode "
	self assert: tMethod returnType equals: #sqInt.
	
	self assert: false "unexpected type conversion"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-message' stamp: ''!
testReturnTempIntLesser32MessageNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempIntLesser32MessageNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-const' stamp: ''!
testReturnTempNilConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempNilConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int. " assignementNode "
	self assert: (ccg typeFor: tMethod statements first value in: tMethod) equals: #int. " value, constantNode "
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt. " variable, temporaryNode "
	self assert: tMethod returnType equals: #sqInt.
	
	self assert: false "unexpected type conversion"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-message' stamp: ''!
testReturnTempNilMessageNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempNilMessageNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-const' stamp: ''!
testReturnTempSmallNegativeIntegerConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempSmallNegativeIntegerConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int. " assignementNode "
	self assert: (ccg typeFor: tMethod statements first value in: tMethod) equals: #int. " value, constantNode "
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt. " variable, temporaryNode "
	self assert: tMethod returnType equals: #sqInt.
	
	self assert: false "unexpected type conversion"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-message' stamp: ''!
testReturnTempSmallNegativeIntegerMessageNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempSmallNegativeIntegerMessageNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-const' stamp: ''!
testReturnTempStringConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempStringConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'char *'. " assignementNode "
	self assert: (ccg typeFor: tMethod statements first value in: tMethod) equals: #'char *'. " value, constantNode "
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt. " variable, temporaryNode "
	self assert: tMethod returnType equals: #sqInt.
	
	self assert: false "type lost"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-message' stamp: ''!
testReturnTempStringMessageNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempStringMessageNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'char *'.
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #'char *'.
	self assert: tMethod returnType equals: #'char *'.! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-const' stamp: ''!
testReturnTempTrueConstantNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempTrueConstantNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int. " assignementNode "
	self assert: (ccg typeFor: tMethod statements first value in: tMethod) equals: #int. " value, constantNode "
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt. " variable, temporaryNode "
	self assert: tMethod returnType equals: #sqInt.
	
	self assert: false "Unexpected conversion"! !

!SlangBasicTypeInferenceTest methodsFor: 'return-temp-assigned-message' stamp: ''!
testReturnTempTrueMessageNode
	| tMethod |
	tMethod := ccg methodNamed: #returnTempTrueMessageNode.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt.
	self assert: (ccg typeFor: tMethod statements first variable in: tMethod) equals: #sqInt.
	self assert: tMethod returnType equals: #sqInt.! !

SlangAbstractTestCase subclass: #SlangTypeForArithmeticTest
	instanceVariableNames: 'visitor propertyName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!SlangTypeForArithmeticTest methodsFor: 'running' stamp: 'hogoww 2/18/2020 08:35'!
setUp
	super setUp.
	ccg addClass: SlangTypeForDereferenceTestClass
! !

!SlangTypeForArithmeticTest methodsFor: 'tests' stamp: 'hogoww 2/18/2020 08:36'!
testAtOnMatrix
	| tMethod |
	tMethod := ccg methodNamed: #aMethodReturningAnAtOnAMatrix.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #sqInt."at: node"
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #sqInt."returnNode"
	self assert: tMethod returnType equals: #sqInt. " euuu, looks weird. Did i do something wrong?"! !

SlangAbstractTestCase subclass: #SlangTypeForDereferenceTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!SlangTypeForDereferenceTest methodsFor: 'running' stamp: ''!
setUp
	super setUp.
	ccg addClass: SlangTypeForDereferenceTestClass.
	ccg inferTypesForImplicitlyTypedVariablesAndMethods.! !

!SlangTypeForDereferenceTest methodsFor: 'tests' stamp: ''!
testAtOnArray
	| tMethod |
	tMethod := ccg methodNamed: #aMethodReturningAnAtOnAnArray.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #int."at: node"
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int."returnNode"
	self assert: tMethod returnType equals: #sqInt.
	
	self assert: false. "Unexpected type conversion"! !

!SlangTypeForDereferenceTest methodsFor: 'tests' stamp: ''!
testAtOnMatrix
	| tMethod |
	tMethod := ccg methodNamed: #aMethodReturningAnAtOnAMatrix.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #'int *'."at: node"
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'int *'."returnNode"
	self assert: tMethod returnType equals: #'int *'.! !

!SlangTypeForDereferenceTest methodsFor: 'tests' stamp: ''!
testReturnPointerTypeExplicitTempVariable
	| tMethod |
	tMethod := ccg methodNamed: #aMethodReturningAPointerType.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #'int *'."variable node"
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'int *'."return node"
	self assert: tMethod returnType equals: #'int *'! !

!SlangTypeForDereferenceTest methodsFor: 'tests' stamp: ''!
testSelfMessageReturningPointerType
	| tMethod |
	tMethod := ccg methodNamed: #aMethodReturningAPointerType.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #'int *'."message node"
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #'int *'."returnNode"
	self assert: tMethod returnType equals: #'int *'! !

!SlangTypeForDereferenceTest methodsFor: 'tests' stamp: ''!
testTwoAtOnMatrix
	| tMethod |
	tMethod := ccg methodNamed: #aMethodReturningTwoAtOnAMatrix.
	
	self assert: tMethod notNil.
	self assert: (ccg typeFor: tMethod statements first expression receiver in: tMethod) equals: #'int *'."inner at:"
	self assert: (ccg typeFor: tMethod statements first expression in: tMethod) equals: #int."outer at:"
	self assert: (ccg typeFor: tMethod statements first in: tMethod) equals: #int."return node"
	self assert: tMethod returnType equals: #sqInt.

	self assert: false. "Unexpected type conversion"! !

SlangAbstractTestCase subclass: #SlangTypePromotionTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!SlangTypePromotionTest methodsFor: 'integer-types' stamp: ''!
testCharAndChar
	"representative of two types smaller than sizeOf(int) with same size"
	| firstType secondType resType |
	firstType := #char.
	secondType := #char.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #int.
	

	"check that the system is in the assumed state."
	self assert: (ccg isIntegralCType: #char).
	self assert: (ccg sizeOfIntegralCType: #char) equals: 1.! !

!SlangTypePromotionTest methodsFor: 'float-types' stamp: ''!
testCharAndFloat
	"representative of two types bigger than sizeOf(int) and with same size"
	| firstType secondType resType |
	firstType := #char.
	secondType := #float.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #float.
	self assert: (ccg isFloatingPointCType: resType).! !

!SlangTypePromotionTest methodsFor: 'integer-types' stamp: ''!
testCharAndLongLong
	"representative of two types bigger than sizeOf(int) and with same size"
	| firstType secondType resType |
	firstType := #char.
	secondType := #'long long'.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #'long long'.
	
	"check that the system is in the assumed state."
	self assert: (ccg isIntegralCType: #char).
	self assert: (ccg sizeOfIntegralCType: #char) equals: 1.
	self assert: (ccg isIntegralCType: #'long long').
	self assert: (ccg sizeOfIntegralCType: #'long long') equals: 8.! !

!SlangTypePromotionTest methodsFor: 'float-types' stamp: ''!
testDoubleAndDouble
	| firstType secondType resType |
	firstType := #double.
	secondType := #double.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #double.
	self assert: (ccg isFloatingPointCType: resType).! !

!SlangTypePromotionTest methodsFor: 'float-types' stamp: ''!
testDoubleAndFloat
	| firstType secondType resType |
	firstType := #double.
	secondType := #float.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #double.
	self assert: (ccg isFloatingPointCType: resType).
	! !

!SlangTypePromotionTest methodsFor: 'float-types' stamp: ''!
testFloatAndDouble
	| firstType secondType resType |
	firstType := #float.
	secondType := #double.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #double.
	self assert: (ccg isFloatingPointCType: resType).
	! !

!SlangTypePromotionTest methodsFor: 'float-types' stamp: ''!
testFloatAndFloat
	| firstType secondType resType |
	firstType := #float.
	secondType := #float.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #float.
	self assert: (ccg isFloatingPointCType: resType).
	! !

!SlangTypePromotionTest methodsFor: 'general' stamp: ''!
testFloatAndNil
	| firstType secondType resType |
	firstType := #float.
	secondType := nil.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #float.
	
	"this feels wrong"
	"inconsistent with integer types too"
	self assert: false.! !

!SlangTypePromotionTest methodsFor: 'float-types' stamp: ''!
testIntAndFloat
	| firstType secondType resType |
	firstType := #int.
	secondType := #float.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #float.
	self assert: (ccg isFloatingPointCType: resType).
	! !

!SlangTypePromotionTest methodsFor: 'integer-types' stamp: ''!
testIntAndInt
	"representative of two types with size = sizeOf(int) of same size"
	| firstType secondType resType |
	firstType := #int.
	secondType := #int.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #int.
	
	"check that the system is in the assumed state."
	self assert: (ccg isIntegralCType: #int).
	self assert: (ccg sizeOfIntegralCType: #int) equals: 4.! !

!SlangTypePromotionTest methodsFor: 'integer-types' stamp: ''!
testIntAndLongLong
	"representative of a different size types"
	| firstType secondType resType |
	firstType := #int .
	secondType := #'long long'.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #'long long'.
	
	"check that the system is in the assumed state."
	self assert: (ccg isIntegralCType: #int).
	self assert: (ccg sizeOfIntegralCType: #int) equals: 4.
	self assert: (ccg isIntegralCType: #'long long').
	self assert: (ccg sizeOfIntegralCType: #'long long') equals: 8.! !

!SlangTypePromotionTest methodsFor: 'general' stamp: ''!
testIntAndNil
	| firstType secondType resType |
	firstType := #int.
	secondType := nil.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: nil.
	! !

!SlangTypePromotionTest methodsFor: 'unsigned-priority' stamp: ''!
testIntAndUnsignedChar
	"representative of an unsigned that is ignored because of integerType promotion"
	| firstType secondType resType |
	firstType := #int.
	secondType := #'unsigned char'.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #int.
	
	"check that the system is in the assumed state."
	self assert: (ccg isIntegralCType: #int).
	self assert: (ccg sizeOfIntegralCType: #int) equals: 4.
	self assert: (ccg isIntegralCType: #'unsigned char').
	self assert: (ccg sizeOfIntegralCType: #'unsigned char') equals: 1.! !

!SlangTypePromotionTest methodsFor: 'unsigned-priority' stamp: ''!
testIntAndUnsignedInt
	"representative of two types bigger than sizeOf(int) and with same size"
	| firstType secondType resType |
	firstType := #int.
	secondType := #'unsigned int'.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #'unsigned int'.
	
	"check that the system is in the assumed state."
	self assert: (ccg isIntegralCType: #int).
	self assert: (ccg sizeOfIntegralCType: #int) equals: 4.
	self assert: (ccg isIntegralCType: #'unsigned int').
	self assert: (ccg sizeOfIntegralCType: #'unsigned int') equals: 4.! !

!SlangTypePromotionTest methodsFor: 'integer-types' stamp: ''!
testLongLongAndChar
	"representative of different sized types"
	| firstType secondType resType |
	firstType := #'long long'.
	secondType := #char.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #'long long'.
	
	"check that the system is in the assumed state."
	self assert: (ccg isIntegralCType: #'long long').
	self assert: (ccg sizeOfIntegralCType: #'long long') equals: 8.
	self assert: (ccg isIntegralCType: #char).
	self assert: (ccg sizeOfIntegralCType: #char) equals: 1.! !

!SlangTypePromotionTest methodsFor: 'float-types' stamp: ''!
testLongLongAndFloat
	| firstType secondType resType |
	firstType := #'long long'.
	secondType := #float.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #float.
	self assert: (ccg isFloatingPointCType: resType).
	! !

!SlangTypePromotionTest methodsFor: 'integer-types' stamp: ''!
testLongLongAndInt
	"representative of different sized types"
	| firstType secondType resType |
	firstType := #'long long'.
	secondType := #int.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #'long long'.

	"check that the system is in the assumed state."
	self assert: (ccg isIntegralCType: #'long long').
	self assert: (ccg sizeOfIntegralCType: #'long long') equals: 8.
	self assert: (ccg isIntegralCType: #int).
	self assert: (ccg sizeOfIntegralCType: #int) equals: 4.! !

!SlangTypePromotionTest methodsFor: 'integer-types' stamp: ''!
testLongLongAndSqLong
	"representative of two types bigger than sizeOf(int)"
	| firstType secondType resType |
	firstType := #'long long'.
	secondType := #sqLong.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #'long long'.
	
	"check that the system is in the assumed state."
	self assert: (ccg isIntegralCType: #'long long').
	self assert: (ccg sizeOfIntegralCType: #'long long') equals: 8.
	self assert: (ccg isIntegralCType: #sqLong).
	self assert: (ccg sizeOfIntegralCType: #sqLong) equals: 8.! !

!SlangTypePromotionTest methodsFor: 'general' stamp: ''!
testNilAndFloat
	| firstType secondType resType |
	firstType := nil.
	secondType := #float.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #float.
	
	"this feels wrong"
	"inconsistent with integer types too"
	self assert: false.! !

!SlangTypePromotionTest methodsFor: 'general' stamp: ''!
testNilAndInt
	| firstType secondType resType |
	firstType := nil.
	secondType := #int.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: nil.
	! !

!SlangTypePromotionTest methodsFor: 'general' stamp: ''!
testNilAndNil
	| firstType secondType resType |
	firstType := nil.
	secondType := nil.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: nil.
	! !

!SlangTypePromotionTest methodsFor: 'unsigned-priority' stamp: 'dtl 7/27/2021 18:45'!
testUint32tAndInt
	"representative of two types bigger than sizeOf(int) and with same size"
	| firstType secondType resType |
	firstType := #'uint32_t'.
	secondType := #'int'.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #'uint32_t'.
	
	"check that the system is in the assumed state."
	self assert: (ccg isIntegralCType: #'uint32_t').
	self assert: (ccg sizeOfIntegralCType: #'uint32_t') equals: 4.
	self assert: (ccg isIntegralCType: #int).
	self assert: (ccg sizeOfIntegralCType: #int) equals: 4.! !

!SlangTypePromotionTest methodsFor: 'unsigned-priority' stamp: 'dtl 7/27/2021 18:40'!
testUnsignedAndInt
	"representative of two types bigger than sizeOf(int) and with same size"
	| firstType secondType resType |
	firstType := #'unsigned'.
	secondType := #'int'.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #'unsigned'.
	
	"check that the system is in the assumed state."
	self assert: (ccg isIntegralCType: #'unsigned').
	self assert: (ccg sizeOfIntegralCType: #'unsigned') equals: 4.
	self assert: (ccg isIntegralCType: #int).
	self assert: (ccg sizeOfIntegralCType: #int) equals: 4.! !

!SlangTypePromotionTest methodsFor: 'unsigned-priority' stamp: ''!
testUnsignedIntAndInt
	"representative of two types bigger than sizeOf(int) and with same size"
	| firstType secondType resType |
	firstType := #'unsigned int'.
	secondType := #'int'.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #'unsigned int'.
	
	"check that the system is in the assumed state."
	self assert: (ccg isIntegralCType: #'unsigned int').
	self assert: (ccg sizeOfIntegralCType: #'unsigned int') equals: 4.
	self assert: (ccg isIntegralCType: #int).
	self assert: (ccg sizeOfIntegralCType: #int) equals: 4.! !

!SlangTypePromotionTest methodsFor: 'unsigned-priority' stamp: ''!
testUnsignedIntAndLongLong
	"representative of an unsigned that is ignored because of integerType promotion"
	| firstType secondType resType |
	firstType := #'unsigned int'.
	secondType := #'long long'.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #'long long'.
	
	"check that the system is in the assumed state."
	self assert: (ccg isIntegralCType: #'unsigned int').
	self assert: (ccg sizeOfIntegralCType: #'unsigned int') equals: 4.
	self assert: (ccg isIntegralCType: #'long long').
	self assert: (ccg sizeOfIntegralCType: #'long long') equals: 8.! !

!SlangTypePromotionTest methodsFor: 'integer-types' stamp: ''!
testsqLongAndLongLong
	"representative of two types bigger than sizeOf(int) and with same size"
	| firstType secondType resType |
	firstType := #sqLong.
	secondType := #'long long'.
	resType := ccg promoteArithmeticTypes: firstType and: secondType.
	
	self assert: resType equals: #sqLong.
	
	"check that the system is in the assumed state."
	self assert: (ccg isIntegralCType: #sqLong).
	self assert: (ccg sizeOfIntegralCType: #sqLong) equals: 8.
	self assert: (ccg isIntegralCType: #'long long').
	self assert: (ccg sizeOfIntegralCType: #'long long') equals: 8.! !

TestCase subclass: #SlangTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!SlangTests methodsFor: 'tests' stamp: 'nice 12/29/2020 12:05'!
testSimpleMethod
	| codeGenerator tMethod code |
	codeGenerator := CCodeGenerator new.
	tMethod := codeGenerator compileToTMethodSelector: #extBBytecode in: StackInterpreter.
	self assert: #(	#'['
					byte #':=' self fetchByte #'.'
					self fetchNextBytecode #'.'
					extB #':=' #(numExtB #= 0 and: #'[' byte #> 127 #']')
									ifTrue: #'[' byte #- 256 #']'
									ifFalse: #'[' #(extB bitShift: 8) #+ byte #']' #'.'
					numExtB #':=' numExtB #+ 1 #'.'
					#'^' self
					#']')
		equals: (Scanner new scanTokens: tMethod parseTree printString).
	code := String streamContents: [:s| tMethod emitCCodeOn: s generator: codeGenerator].
	code := code allButFirst: (code indexOfSubCollection: 'sqInt') - 1.
	self assert:  #('sqInt' 'extBBytecode(void)' '{' 'sqInt' 'byte;'
					'byte' '=' 'fetchByte();'
					'fetchNextBytecode();'
					'extB' '=' '((numExtB' '==' '0)' '&&' '(byte' '>' '0x7F)'
						'?' 'byte' '-' '0x100'
						':' '((((sqInt)((usqInt)(extB)' '<<' '8))))' '+' 'byte);'
					'numExtB' '+=' '1;' 'return' 'self;' '}')
		equals: (code findTokens: Character separators) asArray ! !

!SlangTests methodsFor: 'tests' stamp: 'eem 9/10/2018 04:49'!
testStatementsHaveNoArguments
	"The inliner expects the body of a TMethod to have no arguments."

	| tMethod |
	tMethod := InterpreterStackPages>>#pageIndexFor: asTranslationMethodOfClass: TMethod.
	self assert: (tMethod parseTree args isNil or: [tMethod parseTree args isEmpty])! !

TestCase subclass: #StackInterpreterSimulatorTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!StackInterpreterSimulatorTests methodsFor: 'tests' stamp: 'eem 1/10/2017 11:38'!
testEmptySimulatorCanCloneSimulation
	self shouldnt:
			[(StackInterpreterSimulator newWithOptions: #(ObjectMemory Spur32BitMemoryManager))
				cloneSimulation cloneSimulation]
		raise: Error! !

!StackInterpreterSimulatorTests methodsFor: 'tests' stamp: 'eem 6/29/2017 12:37'!
testEmptySimulatorCloneCanLoadImage
	self shouldnt:
			[(StackInterpreterSimulator newWithOptions: #(ObjectMemory Spur32BitMemoryManager))
				cloneSimulation
					openOn: SpurImageTestResource imageNameForTests extraMemory: 0]
		raise: Error! !

!StackInterpreterSimulatorTests methodsFor: 'tests' stamp: 'eem 12/20/2017 00:10'!
testHackBits
	"Test that the hackBits plumbing, used for display update in the
	 simulated ioShow:D:i:s:p:l:a:y:, works for the underlying memory classes."
	| mem32 mem64 disp32 disp64 hack32 hack64 |
	mem32 := LittleEndianBitmap new: 65536 / 4.
	mem64 := DoubleWordArray new: 65536 / 8.
	1 to: 65536 by: 8 do:
		[:i|
		mem32
			unsignedLong64At: i
			put: ((i // 8) even
					ifTrue: [16r5555555555555555]
					ifFalse: [16rAAAAAAAAAAAAAAAA]).
		mem64
			unsignedLong64At: i
			put: ((i // 8) even
					ifTrue: [16r5555555555555555]
					ifFalse: [16rAAAAAAAAAAAAAAAA])].
	self assert: (mem32 copy changeClassTo: ByteArray) = (mem64 copy changeClassTo: ByteArray).
	hack32 := Form new hackBits: mem32.
	hack64 := Form new hackBits: mem64.
	disp32 := Form new hackBits: (ByteArray new: 65536).
	disp64 := Form new hackBits: (ByteArray new: 65536).
	disp32
		copy: (0 @ 0 extent: 4 @ (65536 / 4))
		from: 0 @ 0
		in: hack32
		rule: Form over.
	disp64
		copy: (0 @ 0 extent: 4 @ (65536 / 4))
		from: 0 @ 0
		in: hack64
		rule: Form over.
	1 to: 65536 by: 8 do:
		[:i|
		self assert: ((i // 8) even
						ifTrue: [16r5555555555555555]
						ifFalse: [16rAAAAAAAAAAAAAAAA])
			equals: (disp32 bits unsignedLong64At: i).
		self assert: ((i // 8) even
						ifTrue: [16r5555555555555555]
						ifFalse: [16rAAAAAAAAAAAAAAAA])
			equals: (disp64 bits unsignedLong64At: i)]
	! !

!StackInterpreterSimulatorTests methodsFor: 'tests' stamp: 'eem 10/26/2018 10:05'!
testPluginSimulators
	"Test that every plugin under InterpreterPlugin that shouldBeTranslated can also instantiate a simulator.
	 Test that every SmartSyntaxInterpreterPlugin is simulated via a SmartSyntaxPluginSimulator."

	| simulatablePlugins objectMemory vm |
	vm := StackInterpreterSimulator newWithOptions: #(ObjectMemory Spur32BitMemoryManager).
	objectMemory := vm objectMemory.


	simulatablePlugins := InterpreterPlugin allSubclasses select:
								[:pc| pc shouldBeTranslated and: [pc simulatorClass notNil]].
	simulatablePlugins do:
		[:pc|
		self shouldnt: [self assert: (pc simulatorForInterpreterInterface: objectMemory) notNil]
			raise: Error].
	
	simulatablePlugins := SmartSyntaxInterpreterPlugin allSubclasses select:
								[:pc| pc shouldBeTranslated and: [pc simulatorClass notNil]].
	simulatablePlugins do:
		[:pc|
		self shouldnt:
				[| simulator |
				 simulator := pc simulatorForInterpreterInterface: objectMemory.
				 self assert: simulator class == SmartSyntaxPluginSimulator.
				 pc simulatorClass ifNotNil:
					[:simulatorClass|
					 simulatorClass == SmartSyntaxPluginSimulator
						ifTrue: [self assert: (simulator actualPlugin isKindOf: pc)]
						ifFalse: [self assert: simulator actualPlugin class == simulatorClass]]]
			raise: Error]
	! !

!StackInterpreterSimulatorTests methodsFor: 'tests' stamp: 'eem 6/29/2017 12:38'!
testSimulatorCanReloadImage
	self shouldnt:
			[(StackInterpreterSimulator newWithOptions: #(ObjectMemory Spur32BitMemoryManager))
				openOn: SpurImageTestResource imageNameForTests extraMemory: 0;
				openOn: SpurImageTestResource imageNameForTests extraMemory: 0]
		raise: Error! !

TestCase subclass: #StackInterpreterTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!StackInterpreterTests methodsFor: 'tests' stamp: 'eem 11/19/2016 09:13'!
testByteMemoryAccess
	"self new testByteMemoryAccess"
	| om |
	om := NewCoObjectMemorySimulator new allocateMemoryOfSize: 16.
	om byteAt: 0 put: 16r11.
	om byteAt: 1 put: 16r22.
	om byteAt: 2 put: 16r33.
	om byteAt: 3 put: 16r44.
	self assert: (om longAt: 0) equals: 16r44332211.
	self assert: (om byteAt: 0) equals: 16r11.
	self assert: (om byteAt: 1) equals: 16r22.
	self assert: (om byteAt: 2) equals: 16r33.
	self assert: (om byteAt: 3) equals: 16r44.
	om := Spur32BitMMLECoSimulator new allocateMemoryOfSize: 16.
	om byteAt: 0 put: 16r11.
	om byteAt: 1 put: 16r22.
	om byteAt: 2 put: 16r33.
	om byteAt: 3 put: 16r44.
	self assert: (om longAt: 0) equals: 16r44332211.
	self assert: (om byteAt: 0) equals: 16r11.
	self assert: (om byteAt: 1) equals: 16r22.
	self assert: (om byteAt: 2) equals: 16r33.
	self assert: (om byteAt: 3) equals: 16r44.
	om := Spur64BitMMLECoSimulator new allocateMemoryOfSize: 32.
	om byteAt: 0 put: 16r11.
	om byteAt: 1 put: 16r22.
	om byteAt: 2 put: 16r33.
	om byteAt: 3 put: 16r44.
	om byteAt: 4 put: 16r55.
	om byteAt: 5 put: 16r66.
	om byteAt: 6 put: 16r77.
	om byteAt: 7 put: 16r88.
	self assert: (om longAt: 0) equals: 16r8877665544332211.
	self assert: (om byteAt: 0) equals: 16r11.
	self assert: (om byteAt: 1) equals: 16r22.
	self assert: (om byteAt: 2) equals: 16r33.
	self assert: (om byteAt: 3) equals: 16r44.
	self assert: (om byteAt: 4) equals: 16r55.
	self assert: (om byteAt: 5) equals: 16r66.
	self assert: (om byteAt: 6) equals: 16r77.
	self assert: (om byteAt: 7) equals: 16r88.
	(Smalltalk classNamed: #DoubleWordArray) ifNotNil:
		[om := Spur64BitMMLECoSimulatorFor64Bits new allocateMemoryOfSize: 32.
		 om byteAt: 0 put: 16r11.
		 om byteAt: 1 put: 16r22.
		 om byteAt: 2 put: 16r33.
		 om byteAt: 3 put: 16r44.
		 om byteAt: 4 put: 16r55.
		 om byteAt: 5 put: 16r66.
		 om byteAt: 6 put: 16r77.
		 om byteAt: 7 put: 16r88.
		 self assert: (om longAt: 0) equals: 16r8877665544332211.
		 self assert: (om byteAt: 0) equals: 16r11.
		 self assert: (om byteAt: 1) equals: 16r22.
		 self assert: (om byteAt: 2) equals: 16r33.
		 self assert: (om byteAt: 3) equals: 16r44.
		 self assert: (om byteAt: 4) equals: 16r55.
		 self assert: (om byteAt: 5) equals: 16r66.
		 self assert: (om byteAt: 6) equals: 16r77.
		 self assert: (om byteAt: 7) equals: 16r88]! !

!StackInterpreterTests methodsFor: 'tests' stamp: 'eem 12/15/2014 15:44'!
testImmediateFloats
	"self new testImmediateFloats"
	| vm smm smallFloatOop |
	vm := StackInterpreterSimulator newWithOptions: #(ObjectMemory Spur64BitMemoryManager).
	smm := vm objectMemory.
	#(0.0 -0.0 1.0e38 -1.0e38 1.0e-38 -1.0e-38 2.0 -2.0) do:
		[:f| | sfo |
		vm initPrimCall.
		self shouldnt: [sfo := smm smallFloatObjectOf: f] raise: Error.
		self assert: (smm smallFloatValueOf: sfo) equals: f.
		self assert: ((smm exponentOfSmallFloat: sfo) between: -126 and: 127)].
	{Float nan. Float infinity. Float negativeInfinity. 1.0e-39. 1.0e39. 5.8774717541114375e-39. 6.80564733841877e38 } do:
		[:f| self deny: (smm isSmallFloatValue: f)].
	vm initPrimCall.
	smallFloatOop := smm smallFloatTag + (1 << (smm smallFloatMantissaBits + smm numTagBits)).
	self shouldnt: [smm floatValueOf: smallFloatOop] raise: Error.
	self deny: vm failed.
	self assert: (smm isSmallFloatValue: (smm floatValueOf: smallFloatOop)).
	self assert: (smm floatObjectOf: (smm floatValueOf: smallFloatOop)) equals: smallFloatOop.
	vm initPrimCall.
	smallFloatOop := ((1 bitShift: 64) - 1 bitClear: (1 bitShift: smm numTagBits + 1) - 1) + smm smallFloatTag.
	self shouldnt: [smm floatValueOf: smallFloatOop] raise: Error.
	self deny: vm failed.
	self assert: (smm isSmallFloatValue: (smm floatValueOf: smallFloatOop)).
	self assert: (smm floatObjectOf: (smm floatValueOf: smallFloatOop)) equals: smallFloatOop! !

!StackInterpreterTests methodsFor: 'tests' stamp: 'nice 3/15/2016 23:41'!
testPointerTaggingDetagging
	"Test the adding/removal of SmallInteger tags to memory adresses used in
	 StackInterpreter and StackInterpreterSimulator for married context management."

	"StackInterpreterTests new testPointerTaggingDetagging"

	{ StackInterpreterSimulatorLSB new. "StackInterpreterSimulatorMSB new" } do: "Need to add MSB soon"
		[:sis|
		({ sis objectMemory minSmallInteger. sis objectMemory minSmallInteger / 2. -1024. -4. -8.
		    sis objectMemory maxSmallInteger - 1. sis objectMemory maxSmallInteger // 2. 4. 8} collect:
			[:n| n bitClear: sis objectMemory tagMask]) do:
				[:n| | tagged untagged |
				self assert: (n bitAnd: sis objectMemory tagMask) = 0.
				self assert: (sis objectMemory isNonIntegerObject: n).
				tagged := sis withSmallIntegerTags: n.
				untagged := sis withoutSmallIntegerTags: tagged.
				self assert: (sis objectMemory isIntegerObject: tagged).
				self assert: untagged = n]]! !

!StackInterpreterTests methodsFor: 'tests' stamp: 'eem 11/19/2016 09:16'!
testShortMemoryAccess
	"self new testShortMemoryAccess"
	| om |
	om := NewCoObjectMemorySimulator new allocateMemoryOfSize: 16.
	om shortAt: 0 put: 16r2211.
	om shortAt: 2 put: 16r4433.
	self assert: (om longAt: 0) equals: 16r44332211.
	self assert: (om shortAt: 0) equals: 16r2211.
	self assert: (om shortAt: 2) equals: 16r4433.
	om := Spur32BitMMLECoSimulator new allocateMemoryOfSize: 16.
	om shortAt: 0 put: 16r2211.
	om shortAt: 2 put: 16r4433.
	self assert: (om longAt: 0) equals: 16r44332211.
	self assert: (om shortAt: 0) equals: 16r2211.
	self assert: (om shortAt: 2) equals: 16r4433.
	om := Spur64BitMMLECoSimulator new allocateMemoryOfSize: 32.
	om shortAt: 0 put: 16r2211.
	om shortAt: 2 put: 16r4433.
	om shortAt: 4 put: 16r6655.
	om shortAt: 6 put: 16r8877.
	self assert: (om longAt: 0) equals: 16r8877665544332211.
	self assert: (om shortAt: 0) equals: 16r2211.
	self assert: (om shortAt: 2) equals: 16r4433.
	self assert: (om shortAt: 4) equals: 16r6655.
	self assert: (om shortAt: 6) equals: 16r8877.
	(Smalltalk classNamed: #DoubleWordArray) ifNotNil:
		[om := Spur64BitMMLECoSimulatorFor64Bits new allocateMemoryOfSize: 32.
		 om shortAt: 0 put: 16r2211.
		 om shortAt: 2 put: 16r4433.
		 om shortAt: 4 put: 16r6655.
		 om shortAt: 6 put: 16r8877.
		 self assert: (om longAt: 0) equals: 16r8877665544332211.
		 self assert: (om shortAt: 0) equals: 16r2211.
		 self assert: (om shortAt: 2) equals: 16r4433.
		 self assert: (om shortAt: 4) equals: 16r6655.
		 self assert: (om shortAt: 6) equals: 16r8877]! !

!StackInterpreterTests methodsFor: 'tests' stamp: 'eem 11/18/2016 17:11'!
testUnalignedMemoryAccess
	"self new testUnalignedMemoryAccess"
	| om |
	om := NewCoObjectMemorySimulator new allocateMemoryOfSize: 16.
	om unalignedLongAt: 1 put: 16r11223344.
	self assert: (om unalignedLongAt: 0) equals: 16r22334400.
	self assert: (om unalignedLongAt: 4) equals: 16r11.
	self assert: (om unalignedLongAt: 1) equals: 16r11223344.
	om longAt: 0 put: 16rAAAAAAAA.
	om longAt: 4 put: 16rAAAAAAAA.
	om unalignedLongAt: 1 put: 16r11223344.
	self assert: (om unalignedLongAt: 0) equals: 16r223344AA.
	self assert: (om unalignedLongAt: 4) equals: 16rAAAAAA11.
	self assert: (om unalignedLongAt: 1) equals: 16r11223344.
	om := Spur32BitMMLECoSimulator new allocateMemoryOfSize: 16.
	om unalignedLongAt: 3 put: 16r11223344.
	self assert: (om unalignedLongAt: 0) equals: 16r44000000.
	self assert: (om unalignedLongAt: 4) equals: 16r112233.
	self assert: (om unalignedLongAt: 3) equals: 16r11223344.
	om longAt: 0 put: 16rAAAAAAAA.
	om longAt: 4 put: 16rAAAAAAAA.
	om unalignedLongAt: 3 put: 16r11223344.
	self assert: (om unalignedLongAt: 0) equals: 16r44AAAAAA.
	self assert: (om unalignedLongAt: 4) equals: 16rAA112233.
	self assert: (om unalignedLongAt: 3) equals: 16r11223344.
	om := Spur64BitMMLECoSimulator new allocateMemoryOfSize: 32.
	om unalignedLongAt: 3 put: 16r1122334455667788.
	self assert: (om unalignedLongAt: 0) equals: 16r4455667788000000.
	self assert: (om unalignedLongAt: 8) equals: 16r112233.
	self assert: (om unalignedLongAt: 3) equals: 16r1122334455667788.
	om longAt: 0 put: 16rAAAAAAAAAAAAAAAA.
	om longAt: 8 put: 16rAAAAAAAAAAAAAAAA.
	om unalignedLongAt: 3 put: 16r1122334455667788.
	self assert: (om unalignedLongAt: 0) equals: 16r4455667788AAAAAA.
	self assert: (om unalignedLongAt: 8) equals: 16rAAAAAAAAAA112233.
	self assert: (om unalignedLongAt: 3) equals: 16r1122334455667788.
	(Smalltalk classNamed: #DoubleWordArray) ifNotNil:
		[om := Spur64BitMMLECoSimulatorFor64Bits new allocateMemoryOfSize: 32.
		 om unalignedLongAt: 3 put: 16r1122334455667788.
		 self assert: (om unalignedLongAt: 0) equals: 16r4455667788000000.
		 self assert: (om unalignedLongAt: 8) equals: 16r112233.
		 self assert: (om unalignedLongAt: 3) equals: 16r1122334455667788.
		 om longAt: 0 put: 16rAAAAAAAAAAAAAAAA.
		 om longAt: 8 put: 16rAAAAAAAAAAAAAAAA.
		 om unalignedLongAt: 3 put: 16r1122334455667788.
		 self assert: (om unalignedLongAt: 0) equals: 16r4455667788AAAAAA.
		 self assert: (om unalignedLongAt: 8) equals: 16rAAAAAAAAAA112233.
		 self assert: (om unalignedLongAt: 3) equals: 16r1122334455667788]! !

TestCase subclass: #VMClassTests
	instanceVariableNames: 'testString vmclass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!VMClassTests methodsFor: 'initialize-release' stamp: 'AlistairGrant 10/14/2018 19:01'!
initialize

	super initialize.
	testString := 'hello world'.! !

!VMClassTests methodsFor: 'running' stamp: 'AlistairGrant 10/14/2018 20:40'!
setUp

	super setUp.
	vmclass := VMClass new.
! !

!VMClassTests methodsFor: 'tests' stamp: 'AlistairGrant 10/14/2018 20:40'!
testStrlen

	| testByteArray |

	"Instances of String must be the correct length"
	self assert: (vmclass strlen: testString) equals: testString size.

	"Instances of ByteArray can optionally have trailing nulls"
	testByteArray := testString asByteArray.
	self assert: (vmclass strlen: testByteArray) equals: testString size.
	testByteArray := testByteArray, (ByteArray new: 3).
	self assert: (vmclass strlen: testByteArray) equals: testString size.
! !

!VMClassTests methodsFor: 'tests' stamp: 'AlistairGrant 10/18/2018 18:07'!
testStrncpy

	| stringA byteArrayA |

	stringA := String new: 5.
	vmclass strncpy: stringA _: testString _: stringA size.
	self assert: stringA equals: 'hello'.

	stringA := String new: testString size + 3.
	vmclass strncpy: stringA _: testString _: stringA size.
	self assert: stringA equals: (testString, (String new: 3)).

	byteArrayA := ByteArray new: 5.
	vmclass strncpy: byteArrayA _: testString _: byteArrayA size.
	self assert: byteArrayA equals: 'hello' asByteArray.

	byteArrayA := ByteArray new: testString size + 3.
	vmclass strncpy: byteArrayA _: testString _: byteArrayA size.
	self assert: byteArrayA equals: (testString, (String new: 3)) asByteArray.

	"SVr4, 4.3BSD, C89, C99 require the destination space after the string be null filled"
	byteArrayA := ByteArray new: testString size + 3.
	byteArrayA atAllPut: 255.
	vmclass strncpy: byteArrayA _: testString _: byteArrayA size.
	self assert: byteArrayA equals: (testString, (String new: 3)) asByteArray.

! !

TestCase subclass: #VMInterfaceConsistencyTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!VMInterfaceConsistencyTests commentStamp: 'eem 12/6/2012 15:03' prior: 0!
VMInterfaceConsistencyTests tests that the interpreter proxy implementations are consistent.!

!VMInterfaceConsistencyTests methodsFor: 'tests' stamp: 'eem 12/6/2012 16:15'!
testInterpreterProxyInterface
	"| fails |
	 fails := Set new.
	 [self new testInterpreterProxyInterface]
		on: AssertionFailure, TestFailure
		do: [:ex|
			fails add: ex messageText.
			ex resume].
	 fails asArray sort do:
		[:m| Transcript cr; nextPutAll: m; flush]"
	(VMClass allSubclasses select: [:ea| (ea name endsWith: 'Interpreter') or: [ea name endsWith: 'InterpreterMT']]) do:
		[:class|
		self testInterpreterProxyInterfaceIsConsistentFor: class]! !

!VMInterfaceConsistencyTests methodsFor: 'private-support' stamp: 'eem 01/19/2021 20:40'!
testInterpreterProxyInterfaceIsConsistentFor: aClass
	| ccg |
	ccg := VMPluginCodeGenerator new.
	InterpreterProxy selectors asArray sort do:
		[:sel| | pm |
		((#(initialize #'initialize-release' private) includes: (InterpreterProxy whichCategoryIncludesSelector: sel))
		  or: [ccg noteUsedPluginFunction: sel]) ifFalse:
			[pm := ccg compileToTMethodSelector: sel in: InterpreterProxy.
			(aClass whichClassIncludesSelector: sel) ifNotNil:
				[:impClass| | im pt it |
				im := ccg compileToTMethodSelector: sel in: impClass.
				"A number of functions are actually void in the interpreter but declared as sqInt in sqVirtualMachine.c, e.g. push"
				self assert: ((pt := pm returnType) = (it := im returnType)
							 or: [pt = #sqInt and: [it = #void]])
					description: 'inconsistent returnType for ', sel, ' in InterpreterProxy vs ', impClass, ' ', pt, ' vs ', it.
				pm args withIndexDo:
					[:pma :i| | ima |
					ima := im args at: i.
					self assert: (pt := pm typeFor: pma in: ccg) = (it := im typeFor: ima in: ccg)
						description: 'inconsistent ', i printString, (#('st' 'nd') at: i ifAbsent: 'th'), ' arg type for ', sel, ' in InterpreterProxy vs ', impClass, ' ', pt, ' vs ', it]]]]! !

TestCase subclass: #VMMakerIntegerTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!VMMakerIntegerTests methodsFor: 'tests' stamp: 'eem 10/16/2015 16:34'!
testSignedIntFromFoo
	self assert: 16r55 signedIntFromChar equals: 16r55.
	self assert: 16r155 signedIntFromChar equals: 16r55.
	self assert: 16rAA signedIntFromChar < 0.
	self assert: (16rAA signedIntFromChar bitAnd: 16rFF) = 16rAA.

	self assert: 16r5555 signedIntFromShort equals: 16r5555.
	self assert: 16r15555 signedIntFromShort equals: 16r5555.
	self assert: 16rAAAA signedIntFromShort < 0.
	self assert: (16rAAAA signedIntFromShort bitAnd: 16rFFFF) = 16rAAAA.

	self assert: 16r55555555 signedIntFromLong equals: 16r55555555.
	self assert: 16r155555555 signedIntFromLong equals: 16r55555555.
	self assert: 16rAAAAAAAA signedIntFromLong< 0.
	self assert: (16rAAAAAAAA signedIntFromLong bitAnd: 16rFFFFFFFF) = 16rAAAAAAAA.

	self assert: 16r5555555555555555 signedIntFromLong64 equals: 16r5555555555555555.
	self assert: 16r15555555555555555 signedIntFromLong64 equals: 16r5555555555555555.
	self assert: 16rAAAAAAAAAAAAAAAA signedIntFromLong64< 0.
	self assert: (16rAAAAAAAAAAAAAAAA signedIntFromLong64 bitAnd: 16rFFFFFFFFFFFFFFFF) = 16rAAAAAAAAAAAAAAAA! !

!VMMakerIntegerTests methodsFor: 'tests' stamp: 'eem 10/9/2015 17:18'!
testSignedIntToFoo
	#(16r55 -16r56 16r5555 -16r5556 16r55555555 -16r55555556 16r5555555555555555 -16r5555555555555556) do:
		[:n|
		n abs digitLength = 1 ifTrue:
			[self assert: n signedIntToChar signedIntFromChar equals: n].
		self assert: (n signedIntToChar signedIntFromChar bitAnd: 16rFF) equals: (n bitAnd: 16rFF).
		n abs digitLength <= 2 ifTrue:
			[self assert: n signedIntToShort signedIntFromShort equals: n].
		self assert: (n signedIntToShort signedIntFromShort bitAnd: 16rFFFF) equals: (n bitAnd: 16rFFFF).
		n abs digitLength <= 4 ifTrue:
			[self assert: n signedIntToLong signedIntFromLong equals: n].
		self assert: (n signedIntToLong signedIntFromLong bitAnd: 16rFFFFFFFF) equals: (n bitAnd: 16rFFFFFFFF).
		n abs digitLength <= 8 ifTrue:
			[self assert: n signedIntToLong64 signedIntFromLong64 equals: n].
		self assert: (n signedIntToLong64 signedIntFromLong64 bitAnd: 16rFFFFFFFFFFFFFFFF) equals: (n bitAnd: 16rFFFFFFFFFFFFFFFF)]! !

!BitBlt methodsFor: '*VMMaker-Interpreter' stamp: 'tfel 2/17/2016 17:39'!
copyBitsSimulated
	^ BitBltSimulation copyBitsFrom: self! !

!BitBlt methodsFor: '*VMMaker-Interpreter' stamp: 'tfel 2/17/2016 17:39'!
copyBitsSimulated: alpha
	^ BitBltSimulation copyBitsFrom: self! !

!BitBlt methodsFor: '*VMMaker-Interpreter' stamp: 'tfel 3/3/2016 15:31'!
simulatePrimitive: aString args: args 
	"simulate primitives in RSqueak"
	| proxy bb |
	proxy := InterpreterProxy new.
	proxy synthesizeStackFor: self with: args.
	bb := BitBltSimulation simulatorClass new.
	bb setInterpreter: proxy.
	bb initialiseModule.
	"rendering state is loaded in the primitive implementations"
	bb perform: aString asSymbol.
	^ proxy stackValue: 0
! !

DecompilerTests subclass: #VMMakerDecompilerTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!VMMakerDecompilerTests methodsFor: 'utilities' stamp: 'eem 6/11/2018 13:41'!
classNames
	"A list of the classes in most of the base packages; excluding EToys arbitrarily for now"
	^(PackageInfo named: 'VMMaker') classes collect: [:ea| ea name]! !

!VMMakerDecompilerTests methodsFor: 'utilities' stamp: 'eem 7/26/2018 10:12'!
decompilerFailures
	"Here is the list of failures: either a syntax error, a hard error or some failure to decompile correctly.
	 Collected via 
		DecompilerTestFailuresCollector new computeFailures."

	"class name, selector, error class name or nil"
	^  #(
		#(BitBltSimulator primitive:parameters:receiver: ResumableTestFailure)  "expr ifTrue: [^foo] ifFalse: [bar] => expr ifTrue: [^foo]. bar"
		)

! !

!VMMakerDecompilerTests methodsFor: 'tests' stamp: 'eem 6/11/2018 13:39'!
testDecompilationOfClasses
	self decompileClassesSelect: [:cn| true]! !

!Debugger methodsFor: '*VMMaker-testing' stamp: 'eem 12/18/2019 10:22'!
isDebugger
	^true! !

!CascadeNode methodsFor: '*VMMaker-C translation' stamp: 'eem 8/15/2018 18:42'!
asTranslatorNodeIn: aTMethod
	"Answer a TParseNode subclass equivalent of me."
	^TStmtListNode new
		setArguments: #()
		statements:
			(Array streamContents:
				[:s| | receiverNode |
				receiverNode := receiver asTranslatorNodeIn: aTMethod.
				"don't expand the receiver if it is a send to get an implicit receiver,
				 e.g self interpreter printHex: oop => printHex(oop), /not/ printHex(cascade0,oop)."
				(receiverNode isSend and: [aTMethod definingClass isNonArgumentImplicitReceiverVariableName: receiverNode selector]) ifTrue:
					[receiverNode := TVariableNode new setName: receiverNode selector].
				receiverNode isLeaf ifFalse:
					[| varNode |
					 varNode := aTMethod newCascadeTempFor: receiverNode.
					 s nextPut: (TAssignmentNode new
								setVariable: varNode
								expression: receiverNode).
					receiverNode := varNode].
				messages do:
					[ :msg | s nextPut: ((msg asTranslatorNodeIn: aTMethod) receiver: receiverNode)]]);
		comment: comment! !

Notification subclass: #VMLongJumpSimulation
	instanceVariableNames: 'returnValue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!VMLongJumpSimulation commentStamp: 'eem 10/30/2020 10:24' prior: 0!
A VMLongJumpSimulation is an exception used to simulate a _setjmp:/_longjmp: pair.  An instance variable which is typed as a jmpbuf in the real VM holds onto a sub-instance of VMLongJumpSimulation. _longjmp:_: then sets its return value and signals it.  It is the responsibility of the holder of the inst var/sender of _setjmp: to provide an exception handler for the exception artound the point where _setjmp: is called.  See for example StackInterpreter>>enterSmalltalkExecutive that provides the exception handler and invoking enterSmalltalkExecutiveImplementation to do the real work. A macro in enterSmalltalkExecutive is used to arrange that in the real VM a mention of enterSmalltalkExecutive is replaced by a use of enterSmalltalkExecutiveImplementation.

Instance Variables
	returnValue:		<Object>

returnValue
	- the value to be "returned from" _setjmp:, sort-of.
!

VMLongJumpSimulation subclass: #ReenterInterpreter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

VMLongJumpSimulation subclass: #ReenterMachineCode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

VMLongJumpSimulation subclass: #ReenterThreadSchedulingLoop
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!VMLongJumpSimulation methodsFor: 'accessing' stamp: 'eem 1/27/2009 17:17'!
returnValue
	"Answer the value of returnValue"

	^ returnValue! !

!VMLongJumpSimulation methodsFor: 'accessing' stamp: 'eem 1/27/2009 17:17'!
returnValue: anObject
	"Set the value of returnValue"

	returnValue := anObject! !

!DoubleByteArray class methodsFor: '*VMMaker-plugin generation' stamp: 'nice 5/1/2020 21:07'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg 
		ccgLoad: aBlock 
		expr: aString 
		asUnsignedShortPtrFrom: anInteger
		andThen: (cg ccgValBlock: 'isShorts')! !

!DoubleByteArray class methodsFor: '*VMMaker-plugin generation' stamp: 'nice 5/1/2020 21:08'!
ccgDeclareCForVar: aSymbolOrString
	"Address of an unsigned 16 bit value, regardless of Smalltalk wordSize"

	^'unsigned short *', aSymbolOrString! !

!MethodReference methodsFor: '*VMMaker-extensions' stamp: 'eem 9/9/2019 09:37'!
sourceStringOrNil
	^(self actualClass sourceCodeAt: self methodSymbol ifAbsent: [^nil]) asString! !

WriteStream subclass: #CrForSpaceWriteStream
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Building'!

!CrForSpaceWriteStream methodsFor: 'character writing' stamp: 'eem 2/25/2009 13:40'!
space
	"Append a cr character instead of a space character to the receiver."

	self cr! !

!WriteStream methodsFor: '*VMMaker-logging' stamp: 'eem 11/10/2017 15:15'!
fflush
	"compatibility to map the stdio fflush(FILE *) routine onto flush"
	self flush! !

!Array class methodsFor: '*VMMaker-plugin generation' stamp: 'eem 1/30/2015 12:27'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg 
		ccgLoad: aBlock 
		expr: aString 
		asOopPtrFrom: anInteger
		andThen: (cg ccgValBlock: 'isIndexable')! !

!Array class methodsFor: '*VMMaker-plugin generation' stamp: 'ikp 3/31/2005 14:20'!
ccgDeclareCForVar: aSymbolOrString

	^'sqInt *', aSymbolOrString! !

!Array methodsFor: '*VMMaker-breakpoints' stamp: 'eem 2/24/2020 21:50'!
addBreakpoint: bkpt
	^self, {bkpt}! !

!Array methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 12/22/2020 16:10'!
coerceTo: cTypeString sim: interpreterSimulator
	^CLiteralArray on: self! !

!Array methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 11:53'!
isActiveBreakpoint
	^self size > 0! !

!Array methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 11:53'!
isBreakpointFor: address
	1 to: self size do:
		[:i| (self at: i) = address ifTrue: [^true]].
	^false! !

!Array methodsFor: '*VMMaker-breakpoints' stamp: 'eem 2/25/2020 11:07'!
menuPrompt
	^String streamContents:
		[:s|
		s space; nextPut: $(.
		self do: [:bkpt| s nextPutAll: bkpt menuPrompt]
			separatedBy: [s space].
		s nextPut: $)]! !

!Array methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 13:00'!
removeBreakpoint: address
	^(self copyWithout: address) ifEmpty: nil! !

!Array methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 12:39'!
singleStepRequiredToTriggerIn: aCogit
	^self anySatisfy: [:address| address between: aCogit cogCodeBase and: aCogit methodZone limitZony]! !

!Integer class methodsFor: '*VMMaker-plugin generation' stamp: 'nice 4/19/2016 23:02'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg ccgLoad: aBlock expr: aString asKindOfIntegerFrom: anInteger! !

!Integer methodsFor: '*VMMaker-bit manipulation' stamp: 'eem 11/21/2014 13:17'!
>>> shiftAmount  "right shift, but causes CCodeGenerator to generate a signed shift"
	shiftAmount < 0 ifTrue: [self error: 'negative arg'].
	^self bitShift: 0 - shiftAmount! !

!Integer methodsFor: '*VMMaker-breakpoints' stamp: 'eem 2/24/2020 21:50'!
addBreakpoint: bkpt
	^{self. bkpt}! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 10/25/2018 16:42'!
asC_int
	^self >= 0
		ifTrue: [self bitAnd: 16rFFFFFFFF]
		ifFalse: [(self bitAnd: 16rFFFFFFFF) - 16r100000000]! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 10/25/2018 16:43'!
asC_unsigned_int
	^self bitAnd: 16rFFFFFFFF! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 12/30/2016 14:22'!
asCharPtr
	^(Notification new tag: #getInterpreter; signal)
		ifNotNil: [:interpreter|
				(interpreter firstIndexableField: self) asInteger
					coerceTo: #'char *'
					sim: interpreter]
		ifNil: [self]! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'nice 7/8/2016 02:33'!
asIntegerPtr
	^self! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 2/20/2013 12:27'!
asLong
	^self! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 3/17/2014 12:29'!
asPositiveIntegerObj
	"This is a SmartSyntaxPlugin convenience, invented with little thought for simulation (grrr).
	  In the VM this is equivalent to
		interpreterProxy methodReturnValue: (interpreterProxy positive32BitIntegerFor: self)
	 but if a plugin is being developed I /think/ it is just
		^self
	 So search the stack to discover what context it is being used in."
	(thisContext findContextSuchThat: [:ctxt| ctxt receiver isInterpreterPlugin]) ifNotNil:
		[:ctxt| | interpreter |
		interpreter := ctxt receiver getInterpreter.
		interpreter methodReturnValue: (interpreter positive32BitIntegerFor: self)].
	^self! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 12/30/2016 14:22'!
asSmallIntegerObj
	^(Notification new tag: #getInterpreter; signal)
		ifNotNil: [:interpreter| interpreter integerObjectOf: self]
		ifNil: [self]! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 7/27/2021 10:51'!
asUnsignedInteger
	"Since the simulator deals with positive integers most of the time we assert that the receiver is greater than
	 zero. One exception is stack pointers in the StackInterpreterSimulator, which are negative. Another is the fast
	 isLargeIntegerInstance: test in SpurMemoryManager. So don't fail in these cases."
	self >= 0 ifFalse:
		[thisContext sender selector == #isLargeIntegerInstance: ifTrue:
			[^self bitAnd: (1 bitShift: thisContext sender receiver wordSize * 8) - 1].
		 self assert: (((thisContext sender methodClass includesBehavior: CoInterpreter)
					   and: [self between: Cogit maxNegativeErrorCode and: -1])
					or: [(thisContext sender methodClass includesBehavior: StackInterpreter)
					   and: [thisContext sender receiver stackPages couldBeFramePointer: self]])].
	
	^self! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'nice 7/8/2016 02:31'!
asUnsignedIntegerPtr
	"Since the simulator deals with positive integers most of the time we assert that the receiver is greater than zero.
	 But one major exception is stack pointers in the StackInterpreterSimulator, which are negative.  So don't fail
	 if the sender is a StackInterpreter and the receiver could be a stack pointer."
	self >= 0 ifFalse:
		[self assert: ((thisContext sender methodClass includesBehavior: StackInterpreter)
					and: [thisContext sender receiver stackPages couldBeFramePointer: self])].
	^self! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 11/11/2013 14:58'!
asUnsignedLong
	self assert: self >= 0.
	^self! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 2/20/2016 23:14'!
asUnsignedLongLong
	self assert: self >= 0.
	^self! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 10/31/2013 08:40'!
asVoidPointer
	^self! !

!Integer methodsFor: '*VMMaker-printing' stamp: 'eem 11/23/2019 11:24'!
binary
	^self printStringRadix: 2! !

!Integer methodsFor: '*VMMaker-bit manipulation' stamp: 'eem 11/19/2014 20:18'!
bitInvert64
	"Answer the 64-bit complement of the receiver."

	^self bitXor: 16rFFFFFFFFFFFFFFFF! !

!Integer methodsFor: '*VMMaker-bit manipulation' stamp: 'nice 4/2/2016 22:34'!
byteSwap32
	"swap the bytes of a 32 bit unsigned integer"
	
	^((self << 24) bitAnd: 16rFF000000) bitOr:
	  (((self << 8) bitAnd: 16rFF0000) bitOr:
	  (((self >> 8) bitAnd: 16rFF00) bitOr:
	  (((self >> 24) bitAnd: 16rFF))))! !

!Integer methodsFor: '*VMMaker-bit manipulation' stamp: 'nice 4/2/2016 22:40'!
byteSwap64
	"swap the bytes of a 64 bit unsigned integer"
	
	^(((self << 56) bitAnd: 16rFF00000000000000) bitOr:
	  (((self << 40) bitAnd: 16rFF000000000000) bitOr:
	  (((self << 24) bitAnd: 16rFF0000000000) bitOr:
	  (((self << 8) bitAnd: 16rFF00000000) bitOr:
	  (((self >> 8) bitAnd: 16rFF000000) bitOr:
	  (((self >> 24) bitAnd: 16rFF0000) bitOr:
	  (((self >> 40) bitAnd: 16rFF00) bitOr:
	  ((self >> 56) bitAnd: 16rFF))))))))! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 10/13/2020 09:19'!
coerceTo: cTypeString sim: interpreter

	| bits unitSize |
	cTypeString last == $* ifTrue:  "C pointer"
		[unitSize := cTypeString caseOf: {
		[#'char *'] -> [1].
		[#'short *'] -> [2].
		[#'int *'] -> [4].
		[#'long long *'] -> [8].
		[#'float *'] -> [^CFloatArray basicNew interpreter: interpreter address: self unitSize: 4; yourself].
		[#'double *'] -> [^CFloatArray basicNew interpreter: interpreter address: self unitSize: 8; yourself].
		[#'unsigned *'] -> [4].
		[#'unsigned int *'] -> [4].
		[#'unsigned char *'] -> [1].
		[#'signed char *'] -> [1].
		[#'unsigned short *'] -> [2].
		[#'unsigned long long *'] -> [8].
		[#'oop *'] -> [interpreter objectMemory bytesPerOop].
		}
		otherwise: [interpreter objectMemory wordSize].
		^CArray basicNew
			interpreter: interpreter address: self unitSize: unitSize;
			yourself].
	cTypeString first == $u ifTrue:
		[bits := cTypeString caseOf: {
		[#usqInt] -> [interpreter objectMemory wordSize * 8].
		[#usqLong] -> [64].
		[#unsigned] -> [32].
		[#'unsigned char'] -> [8].
		[#'unsigned int'] -> [8].
		[#'unsigned long'] -> [48]. "LLP64 on Windows :-("
		[#'unsigned long long'] -> [64].
		[#'unsigned short'] -> [16].
		}
		otherwise: [self error: 'unknown unsigned integer type name'].
		^self bitAnd: 1 << bits - 1].
	bits := cTypeString caseOf: {
		[#'sqIntptr_t'] -> [interpreter objectMemory wordSize * 8].
		[#sqLong] -> [64].
		[#char] -> [^self bitAnd: 255]. "char may be signed, may be unsigned; interpret as unsigned by default"
		[#'signed char'] -> [8].
		[#'short'] -> [16].
		[#int] -> [32].
		[#long] -> [48]. "LLP64 on Windows :-("
		[#'long long'] -> [64].
		}
		otherwise: [self error: 'unknown signed integer type name'].
	^(self bitAnd: (1 bitShift: bits) - 1) - ((self bitAnd: (1 bitShift: bits - 1)) bitShift: 1)! !

!Integer methodsFor: '*VMMaker-printing' stamp: 'eem 5/21/2010 11:09'!
hex8
	"Print the receiver in base 16 with prefixed base, using at least 8 digits.
	 DO NOT CHANGE THIS!!  The Cog VMMaker depends on this.
	 Consider using storeStringBase: 16 length: 11 padded: true instead."
	  "16r3333 hex8"
	| hex |
	hex := self hex.  "16rNNN"
	^hex size < 11
		ifTrue: [hex copyReplaceFrom: 4 to: 3
						 with: ('00000000' copyFrom: 1 to: 11-hex size)]
		ifFalse: [hex]! !

!Integer methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 11:57'!
isActiveBreakpoint
	^true! !

!Integer methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 11:58'!
isBreakpointFor: address
	^self = address! !

!Integer methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 13:07'!
menuPrompt
	^' (', self hex, ')'! !

!Integer methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 13:01'!
removeBreakpoint: address
	^nil! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 3/19/2009 16:15'!
signedBitShift: anInteger
	"For historical reasons Slang generates an unsigned shift from all of the shift operators >>, << & bitShift:.
	 These are too deeply entrenched to try and redefine the semantics.  So instead we provide a signed bitShift:
	 that signals to Slang that its argument should be cast to signed, not to unsigned, when being shifted."
	^self bitShift: anInteger! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 10/16/2015 16:21'!
signedIntFromChar
	"Self is an unsigned 8-bit integer in twos-comp form"

	| shortBits |
	shortBits := self bitAnd: 16rFF.
	^(self bitAnd: 16r80) "sign bit" = 0
		ifTrue: [shortBits]
		ifFalse: [shortBits - 16r100]! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 10/16/2015 16:28'!
signedIntFromLong
	"Self is a signed or unsigned 32-bit integer"

	| bits |
	(self >= -1073741824 and: [self <= 1073741823]) ifTrue: "These are known to be SmallIntegers..."
		[^self].
	bits := self bitAnd: 16rFFFFFFFF.
	(bits digitAt: 4) <= 16r7F ifTrue: [^bits].
	^bits - 16r100000000! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 10/16/2015 16:38'!
signedIntFromLong64
	"Self is a signed or unsigned 64-bit integer"

	| bits |
	"This case is handled by the SmallInteger subclass..."
	"(self >= -1073741824 and: [self <= 1073741823]) ifTrue:
		[^self]."
	bits := self bitAnd: 16rFFFFFFFFFFFFFFFF.
	(bits digitAt: 8) <= 16r7F ifTrue: [^bits].
	^bits - 16r10000000000000000! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 3/2/2009 18:32'!
signedIntFromShort
	"Self is an unsigned 16-bit integer in twos-comp form"

	| shortBits |
	shortBits := self bitAnd: 16rFFFF.
	^(self bitAnd: 16r8000) "sign bit" = 0
		ifTrue: [shortBits]
		ifFalse: [shortBits - 16r10000]! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 10/9/2015 17:13'!
signedIntToChar
	"Produces an 8-bit value in twos-comp form. Truncates if out-of-range as per a C cast"

	^self bitAnd: 16rFF! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 10/9/2015 17:10'!
signedIntToLong
	"Produces a 32-bit value in twos-comp form. Truncates if out-of-range as per a C cast"

	^self bitAnd: 16rFFFFFFFF! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 10/9/2015 17:13'!
signedIntToLong64
	"Produces a 64-bit value in twos-comp form. Truncates if out-of-range as per a C cast"

	^self bitAnd: 16rFFFFFFFFFFFFFFFF! !

!Integer methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 10/9/2015 17:11'!
signedIntToShort
	"Produces a 16-bit value in twos-comp form. Truncates if out-of-range as per a C cast"

	^self bitAnd: 16rFFFF! !

!Integer methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 12:39'!
singleStepRequiredToTriggerIn: aCogit
	^self between: aCogit cogCodeBase and: aCogit methodZone limitZony! !

!MessageNode methodsFor: '*VMMaker-C translation' stamp: 'eem 8/15/2018 18:42'!
asTranslatorNodeIn: aTMethod
	"Answer a TParseNode subclass equivalent of me"
	"selector is sometimes a Symbol, sometimes a SelectorNode!!
	 On top of this, numArgs is needed due to the (truly grody) use of
	 arguments as a place to store the extra expressions needed to generate
	 code for in-line to:by:do:, etc.  see below, where it is used.

	 Expand super nodes in place. Elide sends of halt so that halts can be
	 sprinkled through the simulator but will be eliminated from the generated C."
	| rcvrOrNil sel args ifNotNilBlock |
	rcvrOrNil := receiver ifNotNil: [receiver asTranslatorNodeIn: aTMethod].
	(rcvrOrNil notNil
	and: [rcvrOrNil isVariable
	and: [rcvrOrNil name = 'super']]) ifTrue:
		[^aTMethod superExpansionNodeFor: selector key args: arguments].
	sel := selector isSymbol ifTrue: [selector] ifFalse: [selector key].
	sel == #halt ifTrue: [^rcvrOrNil].
	(sel == #cCode:inSmalltalk: "extracting here rather than in translation allows inlining in the block."
	  or: [sel == #cCode:]) ifTrue:
		[arguments first isBlockNode ifTrue:
			[| block |
			 ^(block := arguments first asTranslatorNodeIn: aTMethod) statements size = 1
				ifTrue: [block statements first]
				ifFalse: [block]].
		 (arguments first isLiteralNode
		 and: [arguments first key isString
		 and: [arguments first key isEmpty]]) ifTrue:
			[^arguments first asTranslatorNodeIn: aTMethod]].
	args := arguments
				select: [:arg| arg notNil]
				thenCollect: [:arg| arg asTranslatorNodeIn: aTMethod].
	(sel = #to:by:do: and: [arguments size = 7 and: [(arguments at: 7) notNil]]) ifTrue:
		["Restore limit expr that got moved by transformToDo:"
		 args := {(arguments at: 7) value asTranslatorNodeIn: aTMethod. 
				  args second.
				  args third. "add the limit var as a hidden extra argument; we may need it later"
				  TVariableNode new setName: arguments first key}].
	(sel == #ifTrue:ifFalse: and: [arguments first isJust: NodeNil]) ifTrue:
		[sel := #ifFalse:. args := {args last}].
	(sel == #ifTrue:ifFalse: and: [arguments last isJust: NodeNil]) ifTrue:
		[sel := #ifTrue:. args := {args first}].
	(sel == #ifFalse:ifTrue: and: [arguments first isJust: NodeNil]) ifTrue:
		[sel := #ifTrue:. args := {args last}].
	(sel == #ifFalse:ifTrue: and: [arguments last isJust: NodeNil]) ifTrue:
		[sel := #ifTrue:. args := {args first}].
	((sel == #ifFalse: or: [sel == #or:])
	 and: [arguments size = 2 and: [(arguments at: 2) notNil]]) ifTrue:
		["Restore argument block that got moved by transformOr: or transformIfFalse:"
		 args := {(arguments at: 2) asTranslatorNodeIn: aTMethod}].
	(args size > sel numArgs and: [sel ~~ #to:by:do:]) ifTrue: "to:by:do: has iLimiT hidden in last arg"
		["prune the extra blocks left by ifTrue:, ifFalse:, and: & or:"
		 self assert: args size - sel numArgs = 1.
		 self assert: (args last isStmtList
					  and: [args last statements size = 1
					  and: [(args last statements first isVariable
							or: [args last statements first isConstant])
					  and: [#('nil' true false) includes: args last statements first nameOrValue]]]).
		 args := args first: sel numArgs].
	"For the benefit of later passes, e.g. value: inlining,
	 transform e ifNotNil: [:v| ...] into  v := e. v ifNotNil: [...],
	 which in fact means transforming (v := e) ifTrue: [:v|...] into v := e. v ifTrue: [...]."
	((sel == #ifTrue: or: [sel == #ifFalse: or: [sel == #ifTrue:ifFalse: or: [sel == #ifFalse:ifTrue:]]])
	 and: [receiver notNil
	 and: [receiver isAssignmentEqualsEqualsNil
	 and: [(ifNotNilBlock := args detect: [:arg| arg isStmtList and: [arg args size = 1]] ifNone: []) notNil]]]) ifTrue:
		[ifNotNilBlock setArguments: #().
		 ^TStmtListNode new
			setArguments: #()
			statements:
				{	receiver receiver asTranslatorNodeIn: aTMethod.
					TSendNode new
						setSelector: sel
						receiver: (TSendNode new
									setSelector: #==
									receiver: (receiver receiver variable asTranslatorNodeIn: aTMethod)
									arguments: {receiver arguments first asTranslatorNodeIn: aTMethod})
						arguments: args }].
	((CCodeGenerator isVarargsSelector: sel)
	 and: [args last isCollection
	 and: [args last isSequenceable]]) ifTrue:
		[args := args allButLast, args last].
	^TSendNode new
		setSelector: sel
		receiver: rcvrOrNil
		arguments: args! !

!MessageNode methodsFor: '*VMMaker-C translation' stamp: 'eem 10/22/2013 15:06'!
isAssignmentEqualsEqualsNil
	"Answer if the receiver if of the form (v := e) == nil, which underlies e ifNotNil: [:v|...]"
	^#== == (selector isSymbol ifTrue: [selector] ifFalse: [selector key])
	  and: [receiver notNil
	  and: [receiver isAssignmentNode
	  and: [(arguments first isJust: NodeNil)
		    or: [arguments first isVariableNode
			    and: [arguments first key = 'nil']]]]]! !

!String class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/19/1999 00:21'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg 
		ccgLoad: aBlock 
		expr: aString 
		asCharPtrFrom: anInteger
		andThen: (cg ccgValBlock: 'isBytes')! !

!String class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/18/1999 17:10'!
ccgDeclareCForVar: aSymbolOrString

	^'char *', aSymbolOrString! !

!String methodsFor: '*VMMaker-translation support' stamp: 'eem 10/24/2013 16:24'!
indexOfWord: aSubString
	^self indexOfWord: aSubString startingAt: 1! !

!String methodsFor: '*VMMaker-translation support' stamp: 'eem 10/24/2013 17:05'!
indexOfWord: aSubString startingAt: startIndex
	| index |
	index := startIndex.
	[index <= self size] whileTrue:
		[index := self indexOfSubCollection: aSubString startingAt: index ifAbsent: [^0].
		 ((index = 1 or: [(self at: index - 1) isLetter not])
		  and: [index + aSubString size > self size
				or: [(self at: index + aSubString size) isAlphaNumeric not]]) ifTrue:
			[^index].
		 index := index + 1].
	^0! !

!SmallInteger class methodsFor: '*VMMaker-plugin generation' stamp: 'eem 10/2/2010 14:12'!
ccg: cg generateCoerceToOopFrom: aNode on: aStream
	"N.B.  The is used both for generation and simulation so answer the result (for interpretation)"
	^cg generateCoerceToSmallIntegerObjectFrom: aNode on: aStream! !

!SmallInteger class methodsFor: '*VMMaker-plugin generation' stamp: 'eem 10/3/2010 08:09'!
ccg: cg generateCoerceToValueFrom: aNode on: aStream
	"N.B.  The could be used both for generation and simulation so answer the result (for interpretation)"
	^cg generateCoerceToSmallIntegerValueFrom: aNode on: aStream! !

!SmallInteger class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/18/1999 17:09'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg ccgLoad: aBlock expr: aString asIntegerValueFrom: anInteger! !

!SmallInteger class methodsFor: '*VMMaker-plugin generation' stamp: 'eem 11/19/2014 22:18'!
ccgCanConvertFrom: anObject

	^anObject isInteger! !

!SmallInteger methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 11:59'!
isBreakpointFor: address
	<primitive: 7>
	^self = address! !

!SmallInteger methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 10/16/2015 16:36'!
signedIntFromLong64
	"Self is a signed or unsigned 64-bit integer.
	 Currently SmallIntegers are either 31-bit (in the 32-bit implementation) or 61-bit
	 (in the 64-bit implementation) so save some time by overriding in the subclass."
	^self! !

TestResource subclass: #SpurImageTestResource
	instanceVariableNames: 'emptyVM vmWithLoadedImage'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!SpurImageTestResource class methodsFor: 'accessing' stamp: 'eem 6/29/2017 16:06'!
baseImageName
	^'core32-preen.image'! !

!SpurImageTestResource class methodsFor: 'accessing' stamp: 'eem 6/29/2017 16:06'!
imageNameForTests
	"self imageNameForTests"
	| baseImageName |
	baseImageName := self baseImageName.
	#('.' 'oscogvm/image' '../oscogvm/image') do:
		[:dirName|
		 ((FileDirectory default directoryExists: dirName)
		  and: [(FileDirectory on: dirName) fileExists: baseImageName]) ifTrue:
			[^dirName, '/', baseImageName]].
	self error: 'cannot find ', baseImageName! !

!SpurImageTestResource class methodsFor: 'preening' stamp: 'eem 6/29/2017 16:05'!
preenImage
	"Assume there's a Pharo bootstrap core32.image in ../oscogvm/image/core32.image.
	 We should find out where the image directory is and write a download script to get it.
	 But for now assume it's there.  See e.g.
		https://bintray.com/pharo-project/pharo/Pharo/201701061402-32bit#files"
	"[self preenImage] timeToRun"
	Spur32BitPreen new
		writeDefaultHeader: true;
		savedWindowSize: 640@480;
		preenImage: self imageNameForTests! !

!SpurImageTestResource methodsFor: 'accessing' stamp: 'eem 1/12/2017 10:12'!
emptyVM
	^emptyVM ifNil:
		[emptyVM := StackInterpreterSimulator newWithOptions: #(ObjectMemory Spur32BitMemoryManager
																	  compactorClass SpurPlanningCompactor)]! !

!SpurImageTestResource methodsFor: 'accessing' stamp: 'eem 6/29/2017 12:31'!
initializedVM
	vmWithLoadedImage ifNil:
		[vmWithLoadedImage := StackInterpreterSimulator newWithOptions: #(ObjectMemory Spur32BitMemoryManager).
		 vmWithLoadedImage
			openOn: self class imageNameForTests extraMemory: 0.
		vmWithLoadedImage objectMemory
			initializeMarkStack; "The Pharo bootstrap has no mark or weakling stacks :-)"
			initializeWeaklingStack].
	^vmWithLoadedImage! !

!SpurImageTestResource methodsFor: 'accessing' stamp: 'eem 6/29/2017 12:31'!
reset
	"self current reset"
	emptyVM := vmWithLoadedImage := nil! !

SpurImageTestResource subclass: #SpurTrunkImageTestResource
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!SpurTrunkImageTestResource class methodsFor: 'accessing' stamp: 'eem 6/29/2017 16:06'!
baseImageName
	^'trunk6-preen.image'! !

!DoubleWordArray class methodsFor: '*VMMaker-plugin generation' stamp: 'nice 5/2/2020 11:11'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg 
		ccgLoad: aBlock 
		expr: aString 
		asUnsignedLong64PtrFrom: anInteger
		andThen: (cg ccgValBlock: 'isLong64s')! !

!DoubleWordArray class methodsFor: '*VMMaker-plugin generation' stamp: 'nice 5/1/2020 21:07'!
ccgDeclareCForVar: aSymbolOrString
	"Address of an unsigned 64 bit value, regardless of Smalltalk wordSize"

	^'unsigned long long *', aSymbolOrString! !

!DoubleWordArray class methodsFor: '*VMMaker-debugger' stamp: 'eem 12/5/2017 10:18'!
defaultIntegerBaseInDebugger
	^16! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/1/2017 14:01'!
byteAt: byteAddress
	"Extract a byte from a DoubleWordArray (little-endian version)"
	| lowBits |
	lowBits := byteAddress - 1 bitAnd: 7.
	^((self at: byteAddress - 1 - lowBits // 8 + 1)
		bitShift: lowBits * -8)
		bitAnd: 16rFF! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/19/2017 09:58'!
long64At: byteIndex
	| lowBits wordIndex value high low |
	wordIndex := byteIndex - 1 // 8 + 1.
	(lowBits := byteIndex - 1 \\ 8) = 0
		ifTrue:
			[value := self at: wordIndex]
		ifFalse:
			[high := ((self at: wordIndex + 1) bitAnd: (1 bitShift: lowBits * 8) - 1) bitShift: 8 - lowBits * 8.
			 low := (self at: wordIndex) bitShift: lowBits * -8.
			 high = 0 ifTrue:
				[^low].
			 value := high + low].
	 ^(value bitShift: -56) <= 127
		ifTrue: [value]
		ifFalse: [value - 16r10000000000000000]! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/1/2017 16:25'!
long64At: byteIndex put: aValue
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| wordIndex lowBits mask allOnes |
	wordIndex := byteIndex - 1 // 8 + 1.
	((aValue bitShift: -63) between: -1 and: 0) ifFalse:
		[self errorImproperStore].
	allOnes := 16rFFFFFFFFFFFFFFFF.
	(lowBits := byteIndex - 1 bitAnd: 7) = 0 ifTrue:
		[^self at: wordIndex put: (aValue >= 0 ifTrue: [aValue] ifFalse: [aValue bitAnd: allOnes])].
	mask := allOnes bitShift: 8 - lowBits * -8.
	self at: wordIndex put: (((self at: wordIndex) bitAnd: mask) bitXor: ((aValue bitShift: lowBits * 8) bitAnd: allOnes - mask)).
	self at: wordIndex + 1 put: (((self at: wordIndex + 1) bitAnd: allOnes - mask) bitXor: (allOnes bitAnd: ((aValue bitShift: 8 - lowBits * -8) bitAnd: mask))).
	^aValue! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/5/2017 09:46'!
longAt: byteIndex
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| wordIndex lowBits word hiWord |
	wordIndex := byteIndex - 1 // 8 + 1.
	lowBits := byteIndex - 1 bitAnd: 7.
	word := (self at: wordIndex) bitShift: lowBits * -8.
	lowBits > 4 ifTrue: "access straddles two words"
		[hiWord := (self at: wordIndex + 1) bitShift: 8 - lowBits * 8.
		 word := word + hiWord].
	word := word bitAnd: 16rFFFFFFFF.
	(word bitShift: -24) > 127 ifTrue:
		[word := word - 16r100000000].
	^word! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/6/2017 12:32'!
longAt: byteIndex bigEndian: bigEndian
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| wordIndex lowBits word hiWord |
	wordIndex := byteIndex - 1 // 8 + 1.
	lowBits := byteIndex - 1 bitAnd: 7.
	word := (self at: wordIndex) bitShift: lowBits * -8.
	lowBits > 4 ifTrue: "access straddles two words"
		[hiWord := (self at: wordIndex + 1) bitShift: 8 - lowBits * 8.
		 word := word + hiWord].
	word := word bitAnd: 16rFFFFFFFF.
	bigEndian ifTrue:
		[word := ((word bitShift: -24) bitAnd: 16rFF)
				 + ((word bitShift: -8) bitAnd: 16rFF00)
 				 + ((word bitAnd: 16rFF00) bitShift: 8)
				 + ((word bitAnd: 16rFF) bitShift: 24)].
	(word bitShift: -24) > 127 ifTrue:
		[word := word - 16r100000000].
	^word! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/5/2017 10:55'!
longAt: byteIndex put: aValue
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| wordIndex lowBits word allOnes loMask hiMask |
	wordIndex := byteIndex - 1 // 8 + 1.
	lowBits := byteIndex - 1 bitAnd: 7.
	((aValue bitShift: -31) between: -1 and: 0) ifFalse:
		[self errorImproperStore].
	lowBits <= 4 ifTrue: "access fits in a single word"
		[| mask |
		 mask := 16rFFFFFFFF bitShift: lowBits * 8.
		 word := self at: wordIndex.
		 self at: wordIndex put: ((word bitOr: mask) bitXor: (((aValue bitShift: lowBits * 8) bitAnd: mask) bitXor: mask)).
		 ^aValue].
	"access straddles two words; make lowMask ones where destination is unchanged to avoid overflow"
	allOnes := 16rFFFFFFFFFFFFFFFF.
	loMask := allOnes bitShift: 8 - lowBits * -8.
	hiMask := 16rFFFFFFFF bitShift: 8 - lowBits * -8.
	word := self at: wordIndex.
	self at: wordIndex put: ((word bitAnd: loMask) bitOr: ((aValue bitAnd: (16rFFFFFFFF bitShift: (lowBits bitAnd: 3) * -8)) bitShift: lowBits * 8)).
	word := self at: wordIndex + 1.
	self at: wordIndex + 1 put: ((word bitOr: hiMask) bitXor: ((((aValue bitShift: 4 - (lowBits bitAnd: 3) * -8)) bitAnd: hiMask) bitXor: hiMask)).
	^aValue! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 11/29/2017 17:50'!
unsignedByteAt: byteAddress
	"Extract a byte from a 64-bit word array (little-endian version)"
	| lowBits |
	lowBits := byteAddress - 1 bitAnd: 7.
	^((self at: byteAddress - 1 - lowBits // 8 + 1)
		bitShift: lowBits * -8)
		bitAnd: 16rFF! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 11/29/2017 17:51'!
unsignedByteAt: byteAddress put: byte
	"Insert a byte into a 64-bit word (little-endian version)"
	| longWord shift lowBits longAddr |
	(byte < 0 or: [byte > 255]) ifTrue:[^self errorImproperStore].
	lowBits := byteAddress - 1 bitAnd: 7.
	longWord := self at: (longAddr := (byteAddress - 1 - lowBits) // 8 + 1).
	shift := lowBits * 8.
	longWord := longWord
				- (longWord bitAnd: (16rFF bitShift: shift)) 
				+ (byte bitShift: shift).
	self at: longAddr put: longWord.
	^byte! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/19/2017 09:58'!
unsignedLong64At: byteIndex
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| wordIndex lowBits high low |
	wordIndex := byteIndex - 1 // 8 + 1.
	(lowBits := byteIndex - 1 bitAnd: 7) = 0 ifTrue:
		[^self at: wordIndex].
	high := ((self at: wordIndex + 1) bitAnd: (1 bitShift: lowBits * 8) - 1) bitShift: 8 - lowBits * 8.
	low := (self at: wordIndex) bitShift: lowBits * -8.
	^high = 0 ifTrue: [low] ifFalse: [high + low]! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/1/2017 15:16'!
unsignedLong64At: byteIndex put: aValue
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| wordIndex lowBits mask allOnes |
	wordIndex := byteIndex - 1 // 8 + 1.
	(lowBits := byteIndex - 1 bitAnd: 7) = 0 ifTrue:
		[^self at: wordIndex put: aValue].
	(aValue bitShift: -64) = 0 ifFalse:
		[self errorImproperStore].
	mask := (allOnes := 16rFFFFFFFFFFFFFFFF) bitShift: 8 - lowBits * -8.
	self at: wordIndex put: (((self at: wordIndex) bitAnd: mask) bitXor: ((aValue bitShift: lowBits * 8) bitAnd: allOnes - mask)).
	self at: wordIndex + 1 put: (((self at: wordIndex + 1) bitAnd: allOnes - mask) bitXor: (allOnes bitAnd: ((aValue bitShift: 8 - lowBits * -8) bitAnd: mask))).
	^aValue! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/5/2017 10:58'!
unsignedLongAt: byteIndex
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| wordIndex lowBits word hiWord |
	wordIndex := byteIndex - 1 // 8 + 1.
	lowBits := byteIndex - 1 bitAnd: 7.
	word := (self at: wordIndex) bitShift: lowBits * -8.
	lowBits > 4 ifTrue: "access straddles two words"
		[hiWord := (self at: wordIndex + 1) bitShift: 8 - lowBits * 8.
		 word := word + hiWord].
	^word bitAnd: 16rFFFFFFFF! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/14/2019 15:12'!
unsignedLongAt: byteIndex bigEndian: bigEndian
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| wordIndex lowBits word hiWord |
	wordIndex := byteIndex - 1 // 8 + 1.
	lowBits := byteIndex - 1 bitAnd: 7.
	word := (self at: wordIndex) bitShift: lowBits * -8.
	lowBits > 4 ifTrue: "access straddles two words"
		[hiWord := (self at: wordIndex + 1) bitShift: 8 - lowBits * 8.
		 word := word + hiWord].
	word := word bitAnd: 16rFFFFFFFF.
	bigEndian
		ifTrue:
			[word := ((word bitShift: -24) bitAnd: 16rFF)
					 + ((word bitShift: -8) bitAnd: 16rFF00)
	 				 + ((word bitAnd: 16rFF00) bitShift: 8)
					 + ((word bitAnd: 16rFF) bitShift: 24)]
		ifFalse:
			[].
	^word! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/5/2017 10:59'!
unsignedLongAt: byteIndex put: aValue
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| wordIndex lowBits word allOnes loMask hiMask |
	wordIndex := byteIndex - 1 // 8 + 1.
	lowBits := byteIndex - 1 bitAnd: 7.
	(aValue bitShift: -32) ~= 0 ifTrue:
		[self errorImproperStore].
	lowBits <= 4 ifTrue: "access fits in a single word"
		[| mask |
		 mask := 16rFFFFFFFF bitShift: lowBits * 8.
		 word := self at: wordIndex.
		 self at: wordIndex put: ((word bitOr: mask) bitXor: (((aValue bitShift: lowBits * 8) bitAnd: mask) bitXor: mask)).
		 ^aValue].
	"access straddles two words; make lowMask ones where destination is unchanged to avoid overflow"
	allOnes := 16rFFFFFFFFFFFFFFFF.
	loMask := allOnes bitShift: 8 - lowBits * -8.
	hiMask := 16rFFFFFFFF bitShift: 8 - lowBits * -8.
	word := self at: wordIndex.
	self at: wordIndex put: ((word bitAnd: loMask) bitOr: ((aValue bitAnd: (16rFFFFFFFF bitShift: (lowBits bitAnd: 3) * -8)) bitShift: lowBits * 8)).
	word := self at: wordIndex + 1.
	self at: wordIndex + 1 put: ((word bitOr: hiMask) bitXor: ((((aValue bitShift: 4 - (lowBits bitAnd: 3) * -8)) bitAnd: hiMask) bitXor: hiMask)).
	^aValue! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/1/2017 10:50'!
unsignedShortAt: byteIndex
	"Compatiblity with the ByteArray & Alien methods of the same name."
	| zi word |
	zi := byteIndex - 1.
	word := self at: zi // 8 + 1.
	(zi bitAnd: 1) ~= 0 ifTrue:
		[self notYetImplemented]. "i.e. odd access implies implementing straddling two words"
	(zi bitAnd: 7) ~= 0 ifTrue:
		[word := word bitShift: (zi bitAnd: 7) * -8].
	^word bitAnd: 16rFFFF! !

!DoubleWordArray methodsFor: '*VMMaker-JITSimulation' stamp: 'eem 12/1/2017 10:50'!
unsignedShortAt: byteAddress put: short
	"Insert a double byte into a 64-bit word (little-endian version)"
	| longWord shift lowBits longAddr |
	(short < 0 or: [short > 65535]) ifTrue:[^self errorImproperStore].
	lowBits := byteAddress - 1 bitAnd: 7.
	(lowBits bitAnd: 1) ~= 0 ifTrue:
		[self notYetImplemented]. "i.e. odd access implies implementing straddling two words"
	longWord := self at: (longAddr := (byteAddress - 1 - lowBits) // 8 + 1).
	shift := lowBits * 8.
	longWord := longWord
				- (longWord bitAnd: (16rFFFF bitShift: shift)) 
				+ (short bitShift: shift).
	self at: longAddr put: longWord.
	^short! !

!BlockNode methodsFor: '*VMMaker-C translation' stamp: 'eem 11/15/2019 14:23'!
asRootTranslatorNodeIn: aTMethod
	"Answer a TParseNode subclass equivalent of me"
	| statementList |
	statementList := OrderedCollection new.
	statements do:
		[:s | | newS |
		 newS := s asTranslatorNodeIn: aTMethod.
		 "inline the statement list returned when a CascadeNode is translated and/or when ifNotNil: is transformed"
		 newS isStmtList
			ifTrue:  [statementList addAll: newS statements]
			ifFalse: [statementList add: newS]].
	^TStmtListNode new
		setArguments: (arguments asArray collect: [:arg | arg key])
		statements: statementList;
		comment: comment! !

!BlockNode methodsFor: '*VMMaker-C translation' stamp: 'eem 11/15/2019 14:04'!
asTranslatorNodeIn: aTMethod
	"Answer a TParseNode subclass equivalent of me"
	| statementList |
	statementList := OrderedCollection new.
	statements do:
		[:s | | newS |
		 newS := s asTranslatorNodeIn: aTMethod.
		 "inline the statement list returned when a CascadeNode is translated and/or when ifNotNil: is transformed"
		 newS isStmtList
			ifTrue:  [statementList addAll: newS statements]
			ifFalse: [statementList add: newS]].
	^TLiteralBlockNode new
		setArguments: (arguments asArray collect: [:arg | arg key])
		statements: statementList;
		comment: comment! !

!BlockNode methodsFor: '*VMMaker-C translation' stamp: 'eem 11/6/2014 12:21'!
isPotentialCCaseLabel: stmt in: aTMethod
	(stmt isVariableNode
	 or: [stmt isLiteralNode
		and: [stmt isConstantNumber or: [stmt literalValue isSymbol]]]) ifTrue:
		[^true].
	stmt isMessageNode ifTrue:
		[| selector implementingClass method |
		 selector := stmt selector key.
		 (#(* + -) includes: selector) ifTrue:
			[^(self isPotentialCCaseLabel: stmt receiver in: aTMethod)
			   and: [self isPotentialCCaseLabel: stmt arguments first in: aTMethod]].

		 (selector = #asSymbol
		  and: [stmt receiver isLiteralNode
		  and: [stmt receiver literalValue isSymbol]]) ifTrue:
			[^true].

		 (stmt arguments isEmpty
		  and: [implementingClass := aTMethod definingClass whichClassIncludesSelector: selector.
			   implementingClass ifNil:
				[implementingClass := aTMethod definingClass objectMemoryClass whichClassIncludesSelector: selector].
			   method := implementingClass >> selector.
			   (method isQuick
				or: [(method literalAt: 1) isInteger
					and: [method numLiterals = 3]])
		   and: [(implementingClass basicNew perform: selector) isInteger]]) ifTrue:
				[^true]].
	^false! !

!BlockNode methodsFor: '*VMMaker-C translation' stamp: 'eem 10/11/2013 15:33'!
isPotentialCCaseLabelIn: aTMethod
	| stmt |
	statements size ~= 1 ifTrue: [^false].
	stmt := statements first.
	^self isPotentialCCaseLabel: stmt in: aTMethod! !

!BraceNode methodsFor: '*VMMaker-C translation' stamp: 'eem 8/21/2021 19:34'!
asTranslatorNodeIn: aTMethod
	"Answer a TParseNode subclass equivalent of me."

	"This is for case statements"
	(elements allSatisfy: [:elem| elem isMessageNode and: [elem selector key = #->]]) ifTrue:
		[self assert: (elements allSatisfy:
			[:elem|
			elem receiver isBlockNode
			and: [elem arguments first isBlockNode
			and: [elem receiver isPotentialCCaseLabelIn: aTMethod]]]).
		^TBraceCaseNode new
			caseLabels: (elements collect: [:elem| elem receiver asTranslatorNodeIn: aTMethod]);
			cases: (elements collect: [:elem| elem arguments first asTranslatorNodeIn: aTMethod]);
			comment: comment].

	"This is for varargs selectors (variants of printf:)"
	^TBraceNode new
		elements: (elements collect: [:elem| elem asTranslatorNodeIn: aTMethod])! !

!Context methodsFor: '*VMMaker-simulation' stamp: 'eem 1/2/2021 21:54'!
getReturnAddress
	"Simulate access to the return address, used in the CoInterpreter in fast interpreter activation.
	 For simulaiton map return addresses to sender selectors."
	^(self findContextSuchThat: [:ctxt| ctxt selector == #interpret]) sender method selector! !

!VariableNode methodsFor: '*VMMaker-C translation' stamp: 'eem 8/15/2018 18:43'!
asTranslatorNodeIn: aTMethod
	"Answer a TParseNode subclass equivalent of me"
	name = 'true' ifTrue: [^ TConstantNode new setValue: true].
	name = 'false' ifTrue: [^ TConstantNode new setValue: false].
	^TVariableNode new setName: name! !

!RawBitsArray methodsFor: '*VMMaker-simulation' stamp: 'nice 12/30/2020 03:23'!
byteAt: anInteger
	"emulate an access to raw (unsigned) bytes, as if the receiver was a ByteArray"
	
	| element p |
	p := self bytesPerBasicElement.
	p = 1 ifTrue: [^self basicAt: 1].
	element := self basicAt: anInteger + p - 1 // p.
	^Smalltalk isLittleEndian
		ifTrue: [element digitAt: anInteger - 1 \\ p + 1]
		ifFalse: [element digitAt: p - (anInteger \\ p)]
	! !

!IdentityDictionary methodsFor: '*VMMaker-accessing' stamp: 'eem 11/30/2017 11:07'!
keyAtNonIdentityValue: value
	"Answer the key that is the external name for the argument, value. If 
	 there is none, answer the result of evaluating exceptionBlock.
	 Use =, not ==, so strings like 'this' can be found."
 
	^self keyAtNonIdentityValue: value ifAbsent: [self errorValueNotFound]! !

!IdentityDictionary methodsFor: '*VMMaker-accessing' stamp: 'eem 11/30/2017 11:06'!
keyAtNonIdentityValue: value ifAbsent: exceptionBlock
	"Answer the key that is the external name for the argument, value. If 
	 there is none, answer the result of evaluating exceptionBlock.
	 Use =, not ==, so strings like 'this' can be found."
 
	self associationsDo: 
		[:association | value = association value ifTrue: [^association key]].
	^exceptionBlock value! !

!Float32Array class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/18/1999 17:07'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg ccgLoad: aBlock expr: aString asWBFloatPtrFrom: anInteger! !

!Float32Array class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/17/1999 01:14'!
ccgDeclareCForVar: aSymbolOrString

	^'float *', aSymbolOrString! !

!UndefinedObject methodsFor: '*VMMaker-breakpoints' stamp: 'eem 2/24/2020 21:50'!
addBreakpoint: bkpt
	^bkpt! !

!UndefinedObject methodsFor: '*VMMaker-interpreter simulator' stamp: 'nice 7/8/2016 02:34'!
asIntegerPtr
	^self! !

!UndefinedObject methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 6/22/2009 14:07'!
asUnsignedInteger
	^self! !

!UndefinedObject methodsFor: '*VMMaker-interpreter simulator' stamp: 'nice 7/8/2016 02:31'!
asUnsignedIntegerPtr
	^self! !

!UndefinedObject methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 10/13/2020 09:07'!
coerceTo: cTypeString sim: interpreterSimulator
	^self! !

!UndefinedObject methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 11:55'!
isActiveBreakpoint
	^false! !

!UndefinedObject methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 11:55'!
isBreakpointFor: address
	^false! !

!UndefinedObject methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 13:07'!
menuPrompt
	^''! !

!UndefinedObject methodsFor: '*VMMaker-interpreter simulator' stamp: 'eem 12/24/2015 12:11'!
shouldStopIfAtPC: address
	^true! !

!UndefinedObject methodsFor: '*VMMaker-breakpoints' stamp: 'eem 12/24/2015 12:03'!
singleStepRequiredToTriggerIn: aCogit
	^false! !

DecompilerTestFailuresCollector subclass: #VMMakerDecompilerTestFailuresCollector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!VMMakerDecompilerTestFailuresCollector methodsFor: 'utilities' stamp: 'eem 6/11/2018 13:41'!
classNames
	"A list of the classes in most of the base packages; excluding EToys arbitrarily for now"
	^(PackageInfo named: 'VMMaker') classes collect: [:ea| ea name]! !

!TranscriptStream methodsFor: '*VMMaker-JITSimulation-testing' stamp: 'eem 9/19/2018 10:21'!
isTranscriptStream
	^true! !

SharedPool subclass: #CogAbstractRegisters
	instanceVariableNames: ''
	classVariableNames: 'ABICalleeSavedRegisterMask ABICallerSavedRegisterMask ABIResultReg ABIResultRegHigh Arg0Reg Arg1Reg CallerSavedRegisterMask ClassReg DPFPReg0 DPFPReg1 DPFPReg10 DPFPReg11 DPFPReg12 DPFPReg13 DPFPReg14 DPFPReg15 DPFPReg2 DPFPReg3 DPFPReg4 DPFPReg5 DPFPReg6 DPFPReg7 DPFPReg8 DPFPReg9 Extra0Reg Extra1Reg Extra2Reg Extra3Reg Extra4Reg Extra5Reg Extra6Reg Extra7Reg Extra8Reg FPReg LinkReg NativeSPReg NoReg NumFloatRegisters NumRegisters PCReg RISCTempReg ReceiverResultReg SPReg SendNumArgsReg TempReg VarBaseReg'
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogAbstractRegisters commentStamp: 'eem 2/4/2020 11:44' prior: 0!
I am a pool for the abstract register set that the Cogit uses to define its model of Smalltalk compiled to machine code.  My values are defined by each processor's CogAbstractInstruction subclass (e.g. CogIA32Compiler for x86).  See each class side initializeAbstractRegisters method.  Each register's value is an integer if it is available on the processor, or is the symbol #undefined if it is not.

NoReg
	a value that cannot be confused with a register.  The registers below are given values from 0 to N, corresponding to their meaning in the processor's ISA.
	The numbers used may overlap, e.g. floating point registers are numbered from 0 to N and overlap the integer registers numbered from 0 to N

NumFloatRegisters
	the number of floating point registers (DPFPReg0 on up) that the processor provides
	
NumRegisters
	the number of integer registers that the processor provides

ABICalleeSavedRegisterMask
	this is the callee saved registers as defined by the platform ABI as a register mask (1 << regM bitOr: 1 << regN etc)
ABICallerSavedRegisterMask
	this is the caller saved registers as defined by the platform ABI as a register mask (1 << regM bitOr: 1 << regN etc)
CallerSavedRegisterMask
	this is the intersection of the ABICallerSavedRegisterMask and the fixed use registers (Arg0Reg, Arg1Reg, ClassReg, ReceiverResultReg, and SendNumArgsReg)
	i.e. these are the registers that must be preserved when calling a run-time routine that returns
ABIResultReg
	the register through which word-sized integral results are retrned according to the platform's ABI; this is the C result register for vanilla calls.  On most platforms it is the same as TempReg.
ABIResultRegHigh
	the register through which the high-word of double-word-sized integral results are retrned according to the platform's ABI 

Fixed use Smalltalk registers:
	Arg0Reg
		the register used to pass the zero'th argument in one or two argument sends (on V3 only one argument sends; see CogObjectRepresentation>>numRegArgs)
	Arg1Reg
		the register used to pass the zero'th argument in one or two argument sends (on V3 unused)
	ClassReg
		the register used to pass either the selector (or selector index) in an unlinked send, or the class tag in a linked send, used for checking inline cache sends
	ReceiverResultReg
		the register holding the receiver on send, the result on return, and typically used to hold self when a method is accessing inst vars
	SendNumArgsReg
		the register that passes the number of arguments if the arith of the send is greater than two (see ceSendNArgs et al)
	TempReg
		a generic temporary, that can usefully be equated with the C result register ABIResultReg.

Dual use Smalltalk registers:
	FPReg
		a register (which is ideally the processor's frame pointer register) used as the frame pointer in frameful methods
	SPReg
		a register (which is ideally the processor's stack pointer register) used as the stack pointer in generated machine code
	NativeSPReg
		the processor's native stack pointer register, which may or may not be equal to SPReg

Allocatable registers
	Extra0Reg through Extra8Reg
		registers allocatable by the code generator. This is work-in-progress.  These should be able to be assigned for temporaries by RegisterAllocatingCogit et al

Code Generator registers
RISCTempReg
	a register reserved by a code generator for a RISC (e.g. CogARMCompiler) that is used to synthesize complex instructions

VarBaseReg
	if enough registers are available then this is a register that holds the base address of the CoInterpreter's variables such as stackPointer, newMethod, etc.
	VarBaseReg allows much more concise code accessing those variables from generated machine code

Floating point registers
	DPFPReg0 through DPFPReg15
		the available floating point registers (note that ARMv8 has 32, and that the code generator only uses DPFPReg0 & DPFPReg1).  The processor defines as many as it has

Processor registers
	FPReg
		the processor's frame pointer register (or a register that is used as the Smalltalk frame pointer)
	SPReg
		the processor's stack pointer register
	LinkReg
		if a RISC then the processor's link register, which is loaded with the return address by a call instruction
	PCReg
		if a RISC with an explicit PC register that is one of its general purpose integer registers (as for example does 32-bit ARM, but /not/ 64-bit ARMv8), then this is the register!

!CogAbstractRegisters class methodsFor: 'class initialization' stamp: 'eem 2/4/2020 11:45'!
initialize
	"Define a fixed set of abstract register names used in code generation for Smalltalk code.
	 These are given concrete values by the currently in-use back end, which is a subclass of
	 CogAbstractInstruction; see implementors of initializeAbstractRegisters & this class's comment.

	 We assume concrete registers defined by the back ends are in the range 0 to N, possibly
	 with integer registers and floating-point registers using overlapping ranges of indices.
	 Here we initialize all registers to #undefined, leaving it to initializeAbstractRegisters to
	 initialize the subset of the abstract registers that a platform actually uses."

	"NoReg encodes no register, e.g. for parameters that supply an optional register.
	 Being negative it is distinct from abstract and concrete registers in the 0 to N range."
	NoReg := -1.

	ABICalleeSavedRegisterMask := ABICallerSavedRegisterMask := CallerSavedRegisterMask := 0.
	ABIResultReg := ABIResultRegHigh := #undefined.

	"The core set of abstract registers that define the Cogit's model of Smalltalk code
	 provide for a register-based calling convention oriented towards inline cacheing and
	 executing a core set of machine code primitives in registers.  The set is composed of
	 8 registers, dictated by the available registers on IA32."
	"Smalltalk machine code executes on stack pages in the stack zone, requiring frame and stack pointers."
	FPReg				:= #undefined.	"A frame pointer is used for Smalltalk frames."
	SPReg				:= #undefined.
	ReceiverResultReg	:= #undefined.	"The receiver at point of send, and return value of a send"
	ClassReg			:= #undefined.	"The inline send cache class tag is in this register, loaded at the send site"
	SendNumArgsReg	:= #undefined.	"Sends > 2 args set the arg count in this reg"
	Arg0Reg			:= #undefined.	"In the StackToRegisterMappingCogit 1 & 2 arg sends marshall into these registers."
	Arg1Reg			:= #undefined.
	TempReg			:= #undefined.

	"A small fixed set of abstract scratch registers for register-rich machines (ARM can use 1, x64 can use 6 or 7)."
	Extra0Reg := #undefined.
	Extra1Reg := #undefined.
	Extra2Reg := #undefined.
	Extra3Reg := #undefined.
	Extra4Reg := #undefined.
	Extra5Reg := #undefined.
	Extra6Reg := #undefined.
	Extra7Reg := #undefined.
	Extra8Reg := #undefined.

	"processor-specific registers"
	NativeSPReg := #undefined.

	"RISC-specific registers"
	LinkReg		:= #undefined.
	RISCTempReg	:= #undefined.		"Used to synthesize CISC instructions from multiple RISC instructions."
	PCReg			:= #undefined.		"If the processor has an assignable pc, e.g. ARM"
	VarBaseReg		:= #undefined.		"If useful, points to base of interpreter variables."

	NumRegisters	:= #undefined.	"Number of basic/integer regsiters (we don't do M68k ;-) )"

	"Up to 16 floating-point registers. e.g. IA32+SSE2 can use 8, x64 can use 16."
	DPFPReg0	:= #undefined.
	DPFPReg1	:= #undefined.
	DPFPReg2	:= #undefined.
	DPFPReg3	:= #undefined.
	DPFPReg4	:= #undefined.
	DPFPReg5	:= #undefined.
	DPFPReg6	:= #undefined.
	DPFPReg7	:= #undefined.
	DPFPReg8	:= #undefined.
	DPFPReg9	:= #undefined.
	DPFPReg10	:= #undefined.
	DPFPReg11	:= #undefined.
	DPFPReg12	:= #undefined.
	DPFPReg13	:= #undefined.
	DPFPReg14	:= #undefined.
	DPFPReg15	:= #undefined.

	NumFloatRegisters	:= #undefined.	"Number of floating-point registers (we don;t do single-precision; we don't do xmm high (yet)."

	classPool keysAndValuesDo:
		[:k :value|
		self assert: value == (#(-1 0 0 0) at: (#(NoReg ABICalleeSavedRegisterMask ABICallerSavedRegisterMask CallerSavedRegisterMask) indexOf: k) ifAbsent: [#undefined])]! !

!CogAbstractRegisters class methodsFor: 'debug printing' stamp: 'eem 12/26/2015 02:32'!
nameForFPRegister: reg "<Integer>"
	^#(DPFPReg0 DPFPReg1 DPFPReg2 DPFPReg3 DPFPReg4 DPFPReg5 DPFPReg6 DPFPReg7
		DPFPReg8 DPFPReg9 DPFPReg10 DPFPReg11 DPFPReg12 DPFPReg13 DPFPReg14 DPFPReg15)
		detect: [:sym| (classPool at: sym) = reg]
		ifNone: ['REG', reg printString, '?']! !

!CogAbstractRegisters class methodsFor: 'debug printing' stamp: 'eem 12/28/2019 23:36'!
nameForRegister: reg "<Integer>"
	^#(Arg0Reg Arg1Reg ClassReg FPReg ReceiverResultReg SPReg SendNumArgsReg TempReg
		LinkReg RISCTempReg VarBaseReg PCReg
		Extra0Reg Extra1Reg Extra2Reg Extra3Reg Extra4Reg Extra5Reg Extra6Reg Extra7Reg Extra8Reg)
			detect: [:sym| (classPool at: sym) = reg]
			ifNone: ['REG', reg printString, '?']! !

TestCase subclass: #RegisterAllocatingCogitTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: 'CogAbstractRegisters'
	category: 'VMMaker-Tests'!

!RegisterAllocatingCogitTests methodsFor: 'tests' stamp: 'eem 2/20/2018 12:08'!
testMergeRegisterSwappingForX64InBlockNodeAnalysis
	"Test the first merge in BlockNode>>#analyseTempsWithin:rootNode:assignmentPools:"
	"self new testMergeRegisterSwappingForX64InBlockNodeAnalysis"
	| cogit fixup start |
	cogit := self x64Cogit.
	cogit
		instVarNamed: 'methodOrBlockNumArgs' put: 3;
		instVarNamed: 'methodOrBlockNumTemps' put: 5;
		initSimStackForFramefulMethod: nil.
	(cogit simStackAt: 0) liveRegister: ReceiverResultReg.
	(cogit simStackAt: 1) liveRegister: Extra4Reg.
	(cogit simStackAt: 5) liveRegister: Extra5Reg.
	cogit ssPushDesc: (cogit simStackAt: 1).
	(fixup := cogit bytecodeFixupClass for: cogit)
		mergeSimStack: cogit copySimStack;
		simStackPtr: (cogit instVarNamed: 'simStackPtr');
		becomeMergeFixup.
	cogit
		ssPop: 1;
		ssPushDesc: (cogit simStackAt: 0).
	(cogit simStackAt: 1) liveRegister: NoReg.
	(cogit simStackAt: 5) liveRegister: Extra5Reg.

	self assert: '
0		(bo FPReg-24 (spilled) (live: ReceiverResultReg))
1		(bo FPReg+32 (spilled))
2		(bo FPReg+24 (spilled))
3		(bo FPReg+16 (spilled))
4		(bo FPReg-32 (spilled))
5		(bo FPReg-40 (spilled) (live: Extra5Reg))
6<-(sb)	(bo FPReg-24 (live: ReceiverResultReg))
'		 equals: cogit simStackPrintString.
	self assert:'
0		(bo FPReg-24 (spilled) (live: ReceiverResultReg))
1		(bo FPReg+32 (spilled) (live: Extra4Reg))
2		(bo FPReg+24 (spilled))
3		(bo FPReg+16 (spilled))
4		(bo FPReg-32 (spilled))
5		(bo FPReg-40 (spilled) (live: Extra5Reg))
6<-		(bo FPReg+32 (live: Extra4Reg))
'		equals: fixup simStackPrintString.

	start := cogit getOpcodeIndex.
	[cogit mergeCurrentSimStackWith: fixup]
		on: AssertionFailure
		do: [:ex|
			self assert: false description: 'assertion failure in cogit'.
			ex resume: nil].

	self assert: '
0		(bo FPReg-24 (spilled) (live: ReceiverResultReg))
1		(bo FPReg+32 (spilled))
2		(bo FPReg+24 (spilled))
3		(bo FPReg+16 (spilled))
4		(bo FPReg-32 (spilled))
5		(bo FPReg-40 (spilled) (live: Extra5Reg))
6<-(sb)	(spill @ FPReg-48 (live: Extra4Reg))
'		 equals: cogit simStackPrintString.

	self assert:  '(MoveRR ReceiverResultReg Extra4Reg)'
		equals: (cogit opcodePrintStringFrom: start to: cogit getOpcodeIndex - 1)! !

!RegisterAllocatingCogitTests methodsFor: 'tests' stamp: 'eem 2/20/2018 12:06'!
testMergeRegisterSwappingForX64InSmallIntegerPrinting
	"self new testMergeRegisterSwappingForX64InSmallIntegerPrinting"
	| cogit fixup start |
	cogit := self x64Cogit.
	cogit
		instVarNamed: 'methodOrBlockNumArgs' put: 4;
		instVarNamed: 'methodOrBlockNumTemps' put: 9;
		initSimStackForFramefulMethod: nil;
		ssPushConstant: (cogit objectMemory integerObjectOf: 0).
	(fixup := cogit bytecodeFixupClass for: cogit)
		mergeSimStack: cogit copySimStack;
		simStackPtr: (cogit instVarNamed: 'simStackPtr');
		becomeMergeFixup.
	(cogit simStackAt: 0) liveRegister: Extra5Reg.
	(cogit simStackAt: 5) liveRegister: Extra4Reg.
	(cogit simStackAt: 7) liveRegister: Extra3Reg.
	(cogit simStack: fixup mergeSimStack at: 10) constant: (cogit objectMemory integerObjectOf: 1).
	(cogit simStack: fixup mergeSimStack at: 5) liveRegister: Extra5Reg.
	(cogit simStack: fixup mergeSimStack at: 7) liveRegister: Extra4Reg.
	(cogit simStack: fixup mergeSimStack at: 10) liveRegister: Extra3Reg.

	self assert: '
0		(bo FPReg-24 (spilled) (live: Extra5Reg))
1		(bo FPReg+40 (spilled))
2		(bo FPReg+32 (spilled))
3		(bo FPReg+24 (spilled))
4		(bo FPReg+16 (spilled))
5		(bo FPReg-32 (spilled) (live: Extra4Reg))
6		(bo FPReg-40 (spilled))
7		(bo FPReg-48 (spilled) (live: Extra3Reg))
8		(bo FPReg-56 (spilled))
9		(bo FPReg-64 (spilled))
10<-(sb)	(const =0 (16r0))
'		 equals: cogit simStackPrintString.
	self assert: '
0		(bo FPReg-24 (spilled))
1		(bo FPReg+40 (spilled))
2		(bo FPReg+32 (spilled))
3		(bo FPReg+24 (spilled))
4		(bo FPReg+16 (spilled))
5		(bo FPReg-32 (spilled) (live: Extra5Reg))
6		(bo FPReg-40 (spilled))
7		(bo FPReg-48 (spilled) (live: Extra4Reg))
8		(bo FPReg-56 (spilled))
9		(bo FPReg-64 (spilled))
10<-		(const =1 (16r1) (live: Extra3Reg))
'		equals: fixup simStackPrintString.

	start := cogit getOpcodeIndex.
	[cogit mergeCurrentSimStackWith: fixup]
		on: AssertionFailure
		do: [:ex|
			self assert: false description: 'assertion failure in cogit'.
			ex resume: nil].

	self assert: '
0		(bo FPReg-24 (spilled))
1		(bo FPReg+40 (spilled))
2		(bo FPReg+32 (spilled))
3		(bo FPReg+24 (spilled))
4		(bo FPReg+16 (spilled))
5		(bo FPReg-32 (spilled) (live: Extra5Reg))
6		(bo FPReg-40 (spilled))
7		(bo FPReg-48 (spilled) (live: Extra4Reg))
8		(bo FPReg-56 (spilled))
9		(bo FPReg-64 (spilled))
10<-(sb)	(reg Extra3Reg)
'		 equals: cogit simStackPrintString.

	self assert:  '(XCHGRR 15 14) (XCHGRR 14 13) (MoveCqR 1 Extra3Reg)'
		equals: (cogit opcodePrintStringFrom: start to: cogit getOpcodeIndex - 1)! !

!RegisterAllocatingCogitTests methodsFor: 'private' stamp: 'eem 2/20/2018 09:17'!
x64Cogit
	^RegisterAllocatingCogit initializedInstanceForTests: #(ObjectMemory Spur64BitCoMemoryManager
															ISA X64
															MULTIPLEBYTECODESETS true)! !

SharedPool subclass: #CogCompilationConstants
	instanceVariableNames: ''
	classVariableNames: 'BadRegisterSet CogCompilerClass InFullBlock InVanillaBlock NumSendTrampolines SSBaseOffset SSConstant SSConstantFloat32 SSConstantFloat64 SSConstantInt32 SSConstantInt64 SSConstantNativePointer SSIllegal SSNativeRegister SSRegister SSRegisterDoubleFloat SSRegisterPair SSRegisterSingleFloat SSSpill SSSpillFloat32 SSSpillFloat64 SSSpillInt64 SSSpillNative UnimplementedOperation'
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

SharedPool subclass: #CogMethodConstants
	instanceVariableNames: ''
	classVariableNames: 'CMBlock CMClosedPIC CMFree CMMaxUsageCount CMMethod CMOpenPIC CompletePrimitive EncounteredUnknownBytecode InsufficientCodeSpace MaxLiteralCountForCompile MaxMethodSize MaxNegativeErrorCode MaxNumArgs MaxStackCheckOffset MethodTooBig NotFullyInitialized PrimCallCollectsProfileSamples PrimCallDoNotJIT PrimCallIsExternalCall PrimCallMayEndureCodeCompaction PrimCallNeedsNewMethod PrimCallNeedsPrimitiveFunction PrimCallOnSmalltalkStack PrimCallOnSmalltalkStackAlign2x PrimCallUseCABI ShouldNotJIT UnfailingPrimitive UnimplementedPrimitive YoungSelectorInPIC'
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

SharedPool subclass: #CogRTLOpcodes
	instanceVariableNames: ''
	classVariableNames: 'AddCqR AddCqRR AddCwR AddRR AddRRR AddRdRd AddRsRs AddcCqR AddcRR AlignmentNops AndCqR AndCqRR AndCwR AndRR ArithmeticShiftRightCqR ArithmeticShiftRightCqRR ArithmeticShiftRightRR Call CallFull CallR ClzRR CmpC32R CmpCqR CmpCwR CmpRR CmpRdRd CmpRsRs ConvertRRd ConvertRRs ConvertRdR ConvertRdRs ConvertRsR ConvertRsRd DivRdRd DivRsRs Fill32 FirstJump FirstShortJump Jump JumpAbove JumpAboveOrEqual JumpBelow JumpBelowOrEqual JumpCarry JumpFPEqual JumpFPGreater JumpFPGreaterOrEqual JumpFPLess JumpFPLessOrEqual JumpFPNotEqual JumpFPOrdered JumpFPUnordered JumpFull JumpGreater JumpGreaterOrEqual JumpLess JumpLessOrEqual JumpLong JumpLongNonZero JumpLongZero JumpMulOverflow JumpNegative JumpNoCarry JumpNoMulOverflow JumpNoOverflow JumpNonNegative JumpNonZero JumpOverflow JumpR JumpZero Label LastJump LastRTLCode Literal LoadEffectiveAddressMwrR LogicalShiftLeftCqR LogicalShiftLeftCqRR LogicalShiftLeftRR LogicalShiftRightCqR LogicalShiftRightCqRR LogicalShiftRightRR MoveA32R MoveAbR MoveAwR MoveC32R MoveCqR MoveCwR MoveM16rR MoveM32rR MoveM32rRs MoveM64rRd MoveM8rR MoveMbrR MoveMs8rR MoveMwrR MovePerfCnt64RL MovePerfCnt64RRL MoveRA32 MoveRAb MoveRAw MoveRM16r MoveRM32r MoveRM8r MoveRMbr MoveRMwr MoveRR MoveRRd MoveRX16rR MoveRX32rR MoveRXbrR MoveRXwrR MoveRdM64r MoveRdR MoveRdRd MoveRsM32r MoveRsRs MoveX16rRR MoveX32rRR MoveXbrRR MoveXwrRR MulRdRd MulRsRs NativePopR NativePushR NativeRetN NegateR Nop NotR OrCqR OrCqRR OrCwR OrRR PopR PrefetchAw PushCq PushCw PushR RetN RotateLeftCqR RotateRightCqR SignExtend16RR SignExtend32RR SignExtend8RR SqrtRd SqrtRs Stop SubCqR SubCwR SubRR SubRRR SubRdRd SubRsRs SubbCqR SubbRR TstCqR XorCqR XorCwR XorRR XorRdRd XorRsRs ZeroExtend16RR ZeroExtend32RR ZeroExtend8RR'
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogRTLOpcodes commentStamp: 'eem 12/26/2015 14:00' prior: 0!
I am a pool for the Register-Transfer-Language to which Cog compiles.  I define unique integer values for all RTL opcodes.  See CogAbstractInstruction for instances of instructions with the opcodes that I define.!

TestCase subclass: #AbstractInstructionTests
	instanceVariableNames: 'processor opcodes'
	classVariableNames: ''
	poolDictionaries: 'CogAbstractRegisters CogRTLOpcodes'
	category: 'VMMaker-Tests'!

!AbstractInstructionTests class methodsFor: 'debugger' stamp: 'eem 9/10/2009 09:36'!
defaultIntegerBaseInDebugger
	^16! !

!AbstractInstructionTests class methodsFor: 'utilities' stamp: 'eem 9/18/2010 13:09'!
defaultTestClass
	"self defaultTestClass"
	| defaultInstructionCompilerClass |
	defaultInstructionCompilerClass := (Cogit classPool at: #ProcessorClass) new abstractInstructionCompilerClass.
	^self subclasses detect: [:sc| sc new concreteCompilerClass inheritsFrom: defaultInstructionCompilerClass]! !

!AbstractInstructionTests class methodsFor: 'utilities' stamp: 'eem 9/18/2010 13:10'!
defaultTester
	^self defaultTestClass new! !

!AbstractInstructionTests class methodsFor: 'testing' stamp: 'eem 1/20/2009 15:00'!
isAbstract
	"Override to true if a TestCase subclass is Abstract and should not have
	TestCase instances built from it"

	^self name = #AbstractInstructionTests! !

!AbstractInstructionTests methodsFor: 'generating machine code' stamp: 'eem 9/9/2009 21:03'!
MoveR: reg1 R: reg2
	^self gen: MoveRR operand: reg1 operand: reg2! !

!AbstractInstructionTests methodsFor: 'generating machine code' stamp: 'eem 9/9/2009 21:03'!
PopR: reg
	^self gen: PopR operand: reg! !

!AbstractInstructionTests methodsFor: 'generating machine code' stamp: 'eem 9/9/2009 21:02'!
PushR: reg
	^self gen: PushR operand: reg! !

!AbstractInstructionTests methodsFor: 'generating machine code' stamp: 'eem 1/21/2009 17:20'!
abstractInstruction: theAbstractInstruction follows: anAbstractInstruction
	<var: #theAbstractInstruction type: #'AbstractInstruction *'>
	<var: #anAbstractInstruction  type: #'AbstractInstruction *'>
	^opcodes notNil
	   and: [(opcodes identityIndexOf: theAbstractInstruction ifAbsent: [0])
			> (opcodes identityIndexOf: anAbstractInstruction ifAbsent: [1])]! !

!AbstractInstructionTests methodsFor: 'cogit compatibility' stamp: 'eem 11/13/2015 11:09'!
addressIsInCurrentCompilation: address
	"Provide a range of addresses that are assumed to be in the current compilation.  Within this range operands
	 can be interpreted as pc-relative addresses, and hence allow testing of pc-relative addresses,"
	^self currentCompilationBase <= address
	  and: [address - self currentCompilationBase < 1024]! !

!AbstractInstructionTests methodsFor: 'generating machine code' stamp: 'eem 1/21/2009 17:24'!
addressIsInFixups: abstractInstruction 
	^false! !

!AbstractInstructionTests methodsFor: 'generating machine code' stamp: 'eem 1/21/2009 17:17'!
addressIsInInstructions: anAbstractOpcode 
	^opcodes notNil and: [(opcodes identityIndexOf: anAbstractOpcode ifAbsent: []) notNil]! !

!AbstractInstructionTests methodsFor: 'accessing' stamp: 'eem 12/3/2015 08:54'!
assert: expected equals: actual
	"Override to use a case-insensitive comparison for strings."
	^self
		assert: (expected isString
				ifTrue: [expected sameAs: actual]
				ifFalse: [expected = actual])
		description: [ self comparingStringBetween: expected and: actual ]
! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'eem 1/20/2009 15:10'!
assertCheckQuickArithOpCodeSize: bytes
	self subclassResponsibility! !

!AbstractInstructionTests methodsFor: 'cogit compatibility' stamp: 'eem 9/18/2015 12:31'!
assertSaneJumpTarget: jumpTarget
	<var: #jumpTarget type: #'AbstractInstruction *'>

	self assert: (self addressIsInInstructions: jumpTarget)! !

!AbstractInstructionTests methodsFor: 'private' stamp: 'eem 12/19/2019 15:01'!
assertValidRunResult: result "{ numSubTests. numPassingSubTests }" ofAtLeast: minimumExpectedNumberOfSubTests
	self assert: result isArray.
	self assert: result size >= 2.
	self assert: (result first isInteger and: [result second isInteger]).
	self assert: result first >= minimumExpectedNumberOfSubTests.
	self assert: result first = result second "All sub-tests pass"! !

!AbstractInstructionTests methodsFor: 'generating machine code' stamp: 'eem 4/7/2009 08:43'!
cCoerceSimple: anAbstractOpcode to: aSymbol
	"hack compatibility with Cogit"
	^anAbstractOpcode! !

!AbstractInstructionTests methodsFor: 'accessing' stamp: 'eem 1/21/2009 17:15'!
coInterpreter
	"hack for testing"
	^nil! !

!AbstractInstructionTests methodsFor: 'accessing' stamp: 'eem 12/19/2019 15:02'!
cogitForTests
	"Answer a new CogitForTests, with initialization performed for the current ISA."
	VMClass initializationOptions at: #ISA put: self isaName.
	^CogitForTests new! !

!AbstractInstructionTests methodsFor: 'accessing' stamp: 'eem 1/20/2009 14:56'!
concreteCompilerClass
	"Answer the concreate subclass of CogAbstractInstruction the concrete test class is testing."
	self subclassResponsibility! !

!AbstractInstructionTests methodsFor: 'cogit compatibility' stamp: 'eem 11/13/2015 11:09'!
currentCompilationBase
	"Provide a range of addresses that are assumed to be in the current compilation.  Within this range operands
	 can be interpreted as pc-relative addresses, and hence allow testing of pc-relative addresses,"
	^16r8000! !

!AbstractInstructionTests methodsFor: 'private' stamp: 'eem 12/19/2019 15:08'!
defaultCodeStart
	^1024! !

!AbstractInstructionTests methodsFor: 'disassembly' stamp: 'eem 9/17/2015 10:02'!
disassembleOpcodesIn: memory to: aStream
	self processor
		disassembleFrom: opcodes first address
		to: opcodes last address + opcodes last machineCodeSize - 1
		in: (self memoryAsBytes: memory)
		on: aStream! !

!AbstractInstructionTests methodsFor: 'generating machine code' stamp: 'eem 2/15/2009 16:15'!
gen: opcode
	| abstractInstruction |
	abstractInstruction := self concreteCompilerClass new.
	abstractInstruction
		opcode: opcode;
		cogit: self.
	^opcodes ifNil: [abstractInstruction] ifNotNil: [opcodes addLast: abstractInstruction]! !

!AbstractInstructionTests methodsFor: 'generating machine code' stamp: 'eem 2/15/2009 16:15'!
gen: opcode operand: operandOne
	| abstractInstruction |
	abstractInstruction := self concreteCompilerClass new.
	abstractInstruction
		opcode: opcode;
		cogit: self.
	abstractInstruction operands at: 0 put: operandOne.
	^opcodes ifNil: [abstractInstruction] ifNotNil: [opcodes addLast: abstractInstruction]! !

!AbstractInstructionTests methodsFor: 'generating machine code' stamp: 'eem 2/15/2009 16:15'!
gen: opcode operand: operandOne operand: operandTwo
	| abstractInstruction |
	abstractInstruction := self concreteCompilerClass new.
	abstractInstruction
		opcode: opcode;
		cogit: self.
	abstractInstruction operands at: 0 put: operandOne.
	abstractInstruction operands at: 1 put: operandTwo.
	^opcodes ifNil: [abstractInstruction] ifNotNil: [opcodes addLast: abstractInstruction]! !

!AbstractInstructionTests methodsFor: 'generating machine code' stamp: 'eem 2/15/2009 16:15'!
gen: opcode operand: operandOne operand: operandTwo operand: operandThree
	| abstractInstruction |
	abstractInstruction := self concreteCompilerClass new.
	abstractInstruction
		opcode: opcode;
		cogit: self.
	abstractInstruction operands at: 0 put: operandOne.
	abstractInstruction operands at: 1 put: operandTwo.
	abstractInstruction operands at: 2 put: operandThree.
	^opcodes ifNil: [abstractInstruction] ifNotNil: [opcodes addLast: abstractInstruction]! !

!AbstractInstructionTests methodsFor: 'generating machine code' stamp: 'rmacnak 11/24/2015 18:20:27'!
genConditionalBranch: opcode operand: operandOne
	^self previousInstruction noteFollowingConditionalBranch: (self gen: opcode operand: operandOne)! !

!AbstractInstructionTests methodsFor: 'generating machine code' stamp: 'nice 12/30/2020 03:45'!
generateInstructions
	"See Cogit>>computeMaximumSizes, generateInstructionsAt: & outputInstructionsAt:.
	 This is a pure Smalltalk (non-Slang) version of that trio of methods."
	| address pcDependentInstructions instructions |
	address := 0.
	pcDependentInstructions := OrderedCollection new.
	opcodes do:
		[:abstractInstruction|
		abstractInstruction
			address: address;
			maxSize: abstractInstruction computeMaximumSize.
		address := address + abstractInstruction maxSize].
	address := 0.
	opcodes do:
		[:abstractInstruction|
		abstractInstruction isPCDependent
			ifTrue:
				[abstractInstruction sizePCDependentInstructionAt: address.
				 pcDependentInstructions addLast: abstractInstruction.
				 address := address + abstractInstruction machineCodeSize]
			ifFalse:
				[address := abstractInstruction concretizeAt: address]].
	pcDependentInstructions do:
		[:abstractInstruction|
		abstractInstruction concretizeAt: abstractInstruction address].
	instructions := ByteArray new: address.
	address := 0.
	opcodes do:
		[:abstractInstruction| | machineCodeBytes |
		self assert: abstractInstruction address = address.
		machineCodeBytes := self memoryAsBytes: abstractInstruction machineCode object.
		1 to: abstractInstruction machineCodeSize do:
			[:j|
			instructions at: address + 1 put: (machineCodeBytes byteAt: j).
			address := address + 1]].
	^instructions! !

!AbstractInstructionTests methodsFor: 'private' stamp: 'eem 12/19/2019 15:09'!
generatedCodeFrom: aCogit
	| instructions memory |
	instructions := aCogit generatedMachineCodeAt: self defaultCodeStart. "Avoid any guard page effects in the simulators, i.e. Gdb's ARM32 requires pc >= 16"
	"self processor disassembleFrom: 1024 to: instructions size - 1 in: instructions"
	(memory := ByteArray new: self defaultCodeStart + 1024) replaceFrom: 1 to: instructions size with: instructions startingAt: 1.
	aCogit stopsFrom: instructions size to: memory size - 1 in: memory.
	^memory! !

!AbstractInstructionTests methodsFor: 'accessing' stamp: 'eem 12/18/2019 22:00'!
isaName
	self subclassResponsibility! !

!AbstractInstructionTests methodsFor: 'private' stamp: 'eem 11/13/2015 11:22'!
memoryAsBytes: instructionMachineCode
	"Answer an AbstractInstruction's machine code as a ByteArray.  Subclasses that represent instruction
	 data using Array should override to convert appropriately."
	instructionMachineCode isCObjectAccessor ifTrue:
		[^self memoryAsBytes: instructionMachineCode object].
	self assert: instructionMachineCode class isBits.
	^instructionMachineCode! !

!AbstractInstructionTests methodsFor: 'cogit compatibility' stamp: 'eem 11/13/2015 11:05'!
methodLabel
	^nil! !

!AbstractInstructionTests methodsFor: 'accessing' stamp: 'eem 10/13/2009 12:42'!
objectMemory
	"hack for testing"
	^nil! !

!AbstractInstructionTests methodsFor: 'private' stamp: 'eem 1/21/2009 17:56'!
pairs: aRange do: aBinaryBlock
	aRange do: [:a| aRange do: [:b| aBinaryBlock value: a value: b]]! !

!AbstractInstructionTests methodsFor: 'accessing' stamp: 'rmacnak 11/24/2015 18:23:08'!
previousInstruction
	<returnTypeC: #'AbstractInstruction *'>
	^opcodes last! !

!AbstractInstructionTests methodsFor: 'accessing' stamp: 'eem 1/20/2009 14:54'!
processor
	self subclassResponsibility! !

!AbstractInstructionTests methodsFor: 'generating machine code' stamp: 'eem 1/21/2009 17:10'!
resetGen
	opcodes := OrderedCollection new! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'eem 11/13/2015 11:27'!
runAddCqR: assertPrintBar
	"self defaultTester runAddCqR: true"
	"self defaultTester runAddCqR: false"
	self concreteCompilerClass dataRegistersWithAccessorsDo:
		[:reg :rgetter :rsetter|
		self pairs: (-2 to: 2)  do:
			[:a :b| | inst len bogus memory |
			inst := self gen: AddCqR operand: a operand: reg.
			len := inst concretizeAt: 0.
			memory := self memoryAsBytes: inst machineCode.
			self processor
				reset;
				perform: rsetter with: (processor convertIntegerToInternal: b).
			[[processor pc < len] whileTrue:
				[self processor singleStepIn: memory]]
				on: Error
				do: [:ex| ].
			"self processor printRegistersOn: Transcript.
			 Transcript show: (self processor disassembleInstructionAt: 0 In: memory); cr"
			assertPrintBar
				ifTrue: [self assert: processor pc equals: inst machineCodeSize.
						self assertCheckQuickArithOpCodeSize: inst machineCodeSize]
				ifFalse: [bogus := processor pc ~= inst machineCodeSize].
			self concreteCompilerClass dataRegistersWithAccessorsDo:
				[:ireg :getter :setter| | expected |
				expected := getter == rgetter ifTrue: [a + b] ifFalse: [0].
				assertPrintBar
					ifTrue: [self assert: (self processor convertInternalToInteger: (self processor perform: getter)) equals: expected]
					ifFalse:
						[(self processor convertInternalToInteger: (self processor perform: getter)) ~= expected ifTrue:
							[bogus := true]]].
				assertPrintBar ifFalse:
					[Transcript
						nextPutAll: rgetter; nextPut: $(; print: b; nextPutAll: ') + '; print: a; nextPutAll: ' = ';
						print: (self processor convertInternalToInteger: (self processor perform: rgetter)); cr; flush.
					 bogus ifTrue:
						[self processor printRegistersOn: Transcript.
						 Transcript show: (self processor disassembleInstructionAt: 0 In: memory); cr]]]]! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'nice 12/30/2020 23:23'!
runAddCwR: assertPrintBar
	"self defaultTester runAddCwR: false"
	self concreteCompilerClass dataRegistersWithAccessorsDo:
		[:reg :rgetter :rsetter|
		self pairs: (-2 to: 2)  do:
			[:a :b| | inst len bogus memory |
			inst := self gen: AddCwR operand: a operand: reg.
			len := inst concretizeAt: 0.
			memory := self memoryAsBytes: inst machineCode.
			self processor
				reset;
				perform: rsetter with: (self processor convertIntegerToInternal: b).
			[[processor pc < len] whileTrue:
				[processor singleStepIn: memory]]
				on: Error
				do: [:ex| ].
			"self processor printRegistersOn: Transcript.
			 Transcript show: (self processor disassembleInstructionAt: 0 In: memory); cr"
			assertPrintBar
				ifTrue: [self assert: processor pc = inst machineCodeSize.
						self assertCheckLongArithOpCodeSize: inst machineCodeSize]
				ifFalse: [bogus := processor pc ~= inst machineCodeSize].
			self concreteCompilerClass dataRegistersWithAccessorsDo:
				[:ireg :getter :setter| | expected |
				(self concreteCompilerClass isRISCTempRegister: ireg) ifFalse:
					[expected := getter == rgetter ifTrue: [b + a] ifFalse: [0].
					assertPrintBar
						ifTrue: [self assert: (self processor convertInternalToInteger: (self processor perform: getter)) equals: expected]
						ifFalse:
							[(self processor convertInternalToInteger: (self processor perform: getter)) ~= expected ifTrue:
								[bogus := true]]].
					assertPrintBar ifFalse:
						[Transcript
							nextPutAll: rgetter; nextPut: $(; print: b; nextPutAll: ') + '; print: a; nextPutAll: ' = ';
							print: (self processor convertInternalToInteger: (self processor perform: rgetter)); cr; flush.
						 bogus ifTrue:
							[self processor printRegistersOn: Transcript.
							 Transcript show: (self processor disassembleInstructionAt: 0 In: memory); cr]]]]]! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'eem 11/13/2015 11:27'!
runAddRR: assertPrintBar
	"self defaultTester runAddRR: false"
	"self defaultTester runAddRR: true"
	self concreteCompilerClass dataRegistersWithAccessorsDo:
		[:sreg :srgetter :srsetter|
		self concreteCompilerClass dataRegistersWithAccessorsDo:
			[:dreg :drgetter :drsetter| | inst len memory |
			inst := self gen: AddRR operand: sreg operand: dreg.
			len := inst concretizeAt: 0.
			memory := self memoryAsBytes: inst machineCode.
			self pairs: (-2 to: 2)  do:
				[:a :b| | bogus |
				self processor
					reset;
					perform: srsetter with: (processor convertIntegerToInternal: a);
					perform: drsetter with: (processor convertIntegerToInternal: b).
				[[processor pc < len] whileTrue:
					[self processor singleStepIn: memory]]
					on: Error
					do: [:ex| ].
				"self processor printRegistersOn: Transcript.
				 Transcript show: (self processor disassembleInstructionAt: 0 In: memory); cr"
				assertPrintBar
					ifTrue: [self assert: processor pc equals: inst machineCodeSize]
					ifFalse: [bogus := processor pc ~= inst machineCodeSize].
				self concreteCompilerClass dataRegistersWithAccessorsDo:
					[:ireg :getter :setter| | expected |
					expected := getter == drgetter
									ifTrue: [srgetter == drgetter
												ifTrue: [b + b]
												ifFalse: [a + b]]
									ifFalse: [getter = srgetter
												ifTrue: [a]
												ifFalse: [0]].
					assertPrintBar
						ifTrue: [self assert: (self processor convertInternalToInteger: (self processor perform: getter)) equals: expected]
						ifFalse:
							[(self processor convertInternalToInteger: (self processor perform: getter)) ~= expected ifTrue:
								[bogus := true]]].
					assertPrintBar ifFalse:
						[Transcript
							nextPutAll: drgetter; nextPut: $(; print: b; nextPutAll: ') + ';
							nextPutAll: srgetter; nextPut: $(; print: a; nextPutAll: ') = ';
							print: (self processor convertInternalToInteger: (self processor perform: drgetter)); cr; flush.
						 bogus ifTrue:
							[self processor printRegistersOn: Transcript.
							 Transcript show: (self processor disassembleInstructionAt: 0 In: memory); cr]]]]]! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'eem 11/13/2015 11:28'!
runArithmeticShiftRightRR: assertPrintBar
	"self defaultTester runArithmeticShiftRightRR: false"
	"self defaultTester runArithmeticShiftRightRR: true"
	self concreteCompilerClass dataRegistersWithAccessorsDo:
		[:sreg :srgetter :srsetter|
		self concreteCompilerClass dataRegistersWithAccessorsDo:
			[:dreg :drgetter :drsetter| | inst len memory |
			inst := self gen: ArithmeticShiftRightRR operand: sreg operand: dreg.
			len := inst concretizeAt: 0.
			memory := self memoryAsBytes: inst machineCode.
			self pairs: (-5 to: 19 by: 6)  do:
				[:a :b| | bogus |
				(a >= 0 and: [sreg ~= dreg or: [b >= 0]]) ifTrue:
					[self processor
						reset;
						perform: srsetter with: (processor convertIntegerToInternal: a);
						perform: drsetter with: (processor convertIntegerToInternal: b).
					[self processor singleStepIn: memory.
					 self processor pc ~= inst machineCodeSize] whileTrue.
					"self processor printRegistersOn: Transcript.
					 self processor disassembleFrom: 0 to: inst machineCodeSize in: memory on: Transcript"
					assertPrintBar
						ifTrue: [self assert: processor pc = inst machineCodeSize]
						ifFalse: [bogus := processor pc ~= inst machineCodeSize].
					self concreteCompilerClass dataRegistersWithAccessorsDo:
						[:ireg :getter :setter| | expected |
						expected := getter == drgetter
										ifTrue: [srgetter == drgetter
													ifTrue: [b >> b]
													ifFalse: [b >> a]]
										ifFalse: [getter = srgetter
													ifTrue: [a]
													ifFalse: [0]].
						assertPrintBar
							ifTrue: [self assert: (self processor convertInternalToInteger: (self processor perform: getter)) equals: expected]
							ifFalse:
								[(self processor convertInternalToInteger: (self processor perform: getter)) ~= expected ifTrue:
								[bogus := true]]].
						assertPrintBar ifFalse:
							[Transcript
								nextPutAll: drgetter; nextPut: $(; print: b; nextPutAll: ') >> ';
								nextPutAll: srgetter; nextPut: $(; print: a; nextPutAll: ') = ';
								print: (self processor convertInternalToInteger: (self processor perform: drgetter)); cr; flush.
							 bogus ifTrue:
								[self processor printRegistersOn: Transcript.
								 self processor disassembleFrom: 0 to: inst machineCodeSize in: memory on: Transcript]]]]]]! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'rmacnak 11/27/2015 12:28:59'!
runBinaryConditionalJumps: assertPrintBar
	"self defaultTester runBinaryConditionalJumps: false"
	| mask count reg1 reg2 reg3 reg1Getter reg2Setter reg3Setter |
	mask := (1 << self processor bitsInWord) - 1.
	count := 0.
	self concreteCompilerClass dataRegistersWithAccessorsDo:
		[:n :get :set|
		count = 0 ifTrue: [reg1 := n. reg1Getter := get].
		count = 1 ifTrue: [reg2 := n. reg2Setter := set].
		count = 2 ifTrue: [reg3 := n. reg3Setter := set].
		count := count + 1].
	self assert: reg1 ~= nil.
	self assert: reg2 ~= nil.
	self assert: reg3 ~= nil.
	#(	(JumpAbove > unsigned)			(JumpBelowOrEqual <= unsigned)
		(JumpBelow < unsigned)			(JumpAboveOrEqual >= unsigned)
		(JumpGreater > signed)			(JumpLessOrEqual <= signed)
		(JumpLess < signed)				(JumpGreaterOrEqual >= signed)
		(JumpZero = signed)				(JumpNonZero ~= signed)) do:
		[:triple|
		[:opName :relation :signednessOrResult| | opcode jumpNotTaken jumpTaken nop memory bogus |
		self resetGen.
		opcode := CogRTLOpcodes classPool at: opName.
		self gen: CmpRR operand: reg3 operand: reg2.
		jumpTaken := self genConditionalBranch: opcode operand: 0.
		self gen: MoveCqR operand: 0 operand: reg1.
		jumpNotTaken := self gen: Jump.
		jumpTaken jmpTarget: (self gen: MoveCqR operand: 1 operand: reg1).
		jumpNotTaken jmpTarget: (nop := self gen: Nop).
		memory := self generateInstructions.
		"self processor disassembleFrom: 0 to: memory size in: memory on: Transcript."
		bogus := false.
		self pairs: (-2 to: 2)  do:
			[:a :b| | taken |
			self processor
				reset;
				perform: reg2Setter with: (processor convertIntegerToInternal: a);
				perform: reg3Setter with: (processor convertIntegerToInternal: b).
			[self processor singleStepIn: memory.
			 self processor pc ~= nop address] whileTrue.
			taken := (self processor perform: reg1Getter) = 1.
			assertPrintBar
				ifTrue:
					[self assert: taken equals: (signednessOrResult == #unsigned
												ifTrue: [(a bitAnd: mask) perform: relation with: (b bitAnd: mask)]
												ifFalse: [a perform: relation with: b])]
				ifFalse:
					[Transcript
						nextPutAll: reg2; nextPut: $(; print: a; nextPutAll: ') '; nextPutAll: relation; space;
						nextPutAll: reg3; nextPut: $(; print: b; nextPutAll: ') = ';
						print: taken; cr; flush.
					 taken = (signednessOrResult == #unsigned
											ifTrue: [(a bitAnd: mask) perform: relation with: (b bitAnd: mask)]
											ifFalse: [a perform: relation with: b]) ifFalse:
						[bogus := true]]].
			 bogus ifTrue:
				[self processor printRegistersOn: Transcript.
				 Transcript show: (self processor disassembleInstructionAt: jumpTaken address In: memory); cr]]
					valueWithArguments: triple]! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'rmacnak 11/24/2015 19:39:05'!
runBinaryConditionalJumpsViaCmpCqR: assertPrintBar
	"self defaultTester runBinaryConditionalJumpsViaCmpCqR: false"
	| mask count reg1 reg2 resultRegNum operandRegNum |
	mask := (1 << self processor bitsInWord) - 1.
	count := 0.
	self concreteCompilerClass dataRegistersWithAccessorsDo:
		[:n :get :set|
		count = 0 ifTrue: [resultRegNum := n. reg1 := get].
		count = 1 ifTrue: [operandRegNum := n. reg2 := set].
		count := count + 1].
	self assert: resultRegNum ~= nil.
	self assert: operandRegNum ~= nil.
	#(	(JumpAbove > unsigned)			(JumpBelowOrEqual <= unsigned)
		(JumpBelow < unsigned)			(JumpAboveOrEqual >= unsigned)
		(JumpGreater > signed)			(JumpLessOrEqual <= signed)
		(JumpLess < signed)				(JumpGreaterOrEqual >= signed)
		(JumpZero = signed)				(JumpNonZero ~= signed)) do:
		[:triple|
		[:opName :relation :signednessOrResult| | opcode |
		opcode := CogRTLOpcodes classPool at: opName.
		(-2 to: 2) do:
			[:b| | jumpNotTaken jumpTaken nop memory bogus |
			self resetGen.
			self gen: CmpCqR operand: b operand: operandRegNum.
			jumpTaken := self genConditionalBranch: opcode operand: 0.
			self gen: MoveCqR operand: 0 operand: resultRegNum.
			jumpNotTaken := self gen: Jump.
			jumpTaken jmpTarget: (self gen: MoveCqR operand: 1 operand: resultRegNum).
			jumpNotTaken jmpTarget: (nop := self gen: Nop).
			memory := self generateInstructions.
			assertPrintBar ifFalse:
				[Transcript print: triple; cr.
				 self disassembleOpcodesIn: memory to: Transcript].
			bogus := false.
			(-2 to: 2) do:
				[:a| | taken expected |
				self processor
					reset;
					perform: reg2 with: (processor convertIntegerToInternal: a).
				[self processor singleStepIn: memory.
				 self processor pc ~= nop address] whileTrue.
				taken := (self processor perform: reg1) = 1.
				expected := signednessOrResult == #unsigned
								ifTrue: [(a bitAnd: mask) perform: relation with: (b bitAnd: mask)]
								ifFalse: [a perform: relation with: b].
				assertPrintBar
					ifTrue:
						[self assert: taken equals: expected]
					ifFalse:
						[Transcript
							nextPutAll: 'CmpCqR '; print: b; space; nextPutAll: reg2; tab; tab;
							nextPutAll: reg2; nextPut: $(; print: a; nextPut: $); space;
							nextPutAll: relation; space; print: b; nextPutAll: ' = ';
							print: taken;  nextPutAll: ' ('; print: expected; nextPut: $).
						 taken ~= expected ifTrue:
							[Transcript nextPutAll: ' !!!!'.
							 bogus := true].
						 Transcript cr; flush]].
				bogus ifTrue:
					[self processor printRegistersOn: Transcript.
					 Transcript nextPutAll: jumpTaken symbolic; tab; show: (self processor disassembleInstructionAt: jumpTaken address In: memory); cr]]]
						valueWithArguments: triple]! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'rmacnak 11/27/2015 12:28:23'!
runBinaryConditionalJumpsViaSubCqR: assertPrintBar
	"self defaultTester runBinaryConditionalJumpsViaSubCqR: false"
	| mask count reg1 reg2 setreg2 resultRegNum operandRegNum |
	mask := (1 << self processor bitsInWord) - 1.
	count := 0.
	self concreteCompilerClass dataRegistersWithAccessorsDo:
		[:n :get :set|
		count = 0 ifTrue: [resultRegNum := n. reg1 := get].
		count = 1 ifTrue: [operandRegNum := n. reg2 := get. setreg2 := set].
		count := count + 1].
	self assert: resultRegNum ~= nil.
	self assert: operandRegNum ~= nil.
	
	#(	(JumpAbove > unsigned)			(JumpBelowOrEqual <= unsigned)
		(JumpBelow < unsigned)			(JumpAboveOrEqual >= unsigned)
		(JumpGreater > signed)			(JumpLessOrEqual <= signed)
		(JumpLess < signed)				(JumpGreaterOrEqual >= signed)
		(JumpZero = signed)				(JumpNonZero ~= signed)) do:
		[:triple|
		[:opName :relation :signednessOrResult| | opcode |
		opcode := CogRTLOpcodes classPool at: opName.
		(-2 to: 2) do:
			[:b| | jumpNotTaken jumpTaken nop memory bogus |
			self resetGen.
			self gen: SubCqR operand: b operand: operandRegNum.
			jumpTaken := self genConditionalBranch: opcode operand: 0.
			self gen: MoveCqR operand: 0 operand: resultRegNum.
			jumpNotTaken := self gen: Jump.
			jumpTaken jmpTarget: (self gen: MoveCqR operand: 1 operand: resultRegNum).
			jumpNotTaken jmpTarget: (nop := self gen: Nop).
			memory := self generateInstructions.
			assertPrintBar ifFalse:
				[Transcript print: triple; cr.
				 self disassembleOpcodesIn: memory to: Transcript].
			bogus := false.
			(-2 to: 2) do:
				[:a| | taken result expected |
				self processor
					reset;
					perform: setreg2 with: (self processor convertIntegerToInternal: a).
				[self processor singleStepIn: memory.
				 self processor pc ~= nop address] whileTrue.
				taken := (self processor perform: reg1) = 1.
				result := signednessOrResult == #unsigned
							ifTrue: [(a bitAnd: mask) - (b bitAnd: mask)]
							ifFalse: [a - b].
				expected := result perform: relation with: 0.
				assertPrintBar
					ifTrue:
						[self assert: (taken = expected
									  and: [(result bitAnd: mask) = (processor perform: reg2)])]
					ifFalse:
						[Transcript
							nextPutAll: 'SubCqR '; print: b; space; nextPutAll: reg2; tab; tab;
							print: b; space; nextPutAll: relation; space;
							nextPutAll: reg2; nextPut: $(; print: a; nextPutAll: ') = ';
							print: taken;  nextPutAll: ' ('; print: expected; nextPut: $).
						 taken ~= expected ifTrue:
							[Transcript nextPutAll: ' !!!!'.
							 bogus := true].
						 Transcript cr; flush]].
				bogus ifTrue:
					[self processor printRegistersOn: Transcript.
					 Transcript nextPutAll: jumpTaken symbolic; tab; show: (self processor disassembleInstructionAt: jumpTaken address In: memory); cr]]]
						valueWithArguments: triple]! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'eem 02/04/2020 13:04'!
runCmpRRJumpCond: assertPrintBar on: aStream
	"Compile and evaluate as many combinations of CmpR:R: JumpCond: as possible, checking that they produce the
	 expected result.  Answer an array of the number of comparisons and the number of them that succeeded."
	"self defaultTester runCmpRRJumpCond: false"
	| cogit nTests nGood |
	cogit := self cogitForTests.
	nTests := nGood := 0.
	self concreteCompilerClass dataRegistersWithAccessorsGiven: self processor do:
		[:sreg :srgetter :srsetter|
		self concreteCompilerClass dataRegistersWithAccessorsGiven: self processor do:
			[:dreg :drgetter :drsetter| 
			(sreg ~= dreg "N.B. We do not expect the backEnd to provide CmpR: SPReg R: reg, only CmpR: reg R: SPReg"
			 and: [drgetter ~= #sp]) ifTrue:
				[#(	= ~=
					> > >= >=
					< < <= <=)
					with: #(JumpZero: JumpNonZero:
							JumpGreater: JumpAbove: JumpGreaterOrEqual: JumpAboveOrEqual:
							JumpLess: JumpBelow: JumpLessOrEqual: JumpBelowOrEqual:)
					do: [:comparison :instruction| | unsigned jmp memory |
						unsigned := (instruction includesSubstring: 'Above')
									or: [instruction includesSubstring: 'Below'].
						"(drgetter = #sp or: [srgetter = #sp]) ifTrue: [self halt]."
						cogit resetGen.
						cogit CmpR: dreg R: sreg.
						jmp := cogit perform: instruction with: 0.
						cogit
							MoveCq: 0 R: ABIResultReg;
							Jump: 1536.
						jmp jmpTarget: (cogit MoveCq: 1 R: ABIResultReg).
						cogit Jump: 1536.
						memory := self generatedCodeFrom: cogit.
						"self processor disassembleFrom: 1024 to: cogit codeSize - 1024 * 2 + 1024 in: memory ''"
						self pairs: { -1 << 30. "-1 << 16." -1 << 8. 0. 1 << 8. "1 << 16." 1 << 30 } do:
							[:a :b| | bogus error expected nInsts |
							nTests := nTests + 1.
							error := false.
							nInsts := 0.
							self processor
								reset;
								pc: self defaultCodeStart;
								perform: srsetter with: (processor convertIntegerToInternal: a);
								perform: drsetter with: (processor convertIntegerToInternal: b).
							[[processor pc ~= 1536 and: [nInsts < 16]] whileTrue:
								[processor singleStepIn: memory. nInsts := nInsts + 1]]
								on: Error
								do: [:ex| error := true].
							nInsts >= 16 ifTrue: [error := true].
							"self processor disassembleInstructionAt: 0 In: memory"
							bogus := (processor pc = 1536 and: [#[0 1] includes: processor cResultRegister]) not.
							assertPrintBar ifTrue: [self deny: bogus].
							expected := unsigned
												ifTrue: [(a bitAnd: 1 << 32 - 1) perform: comparison with: (b bitAnd: 1 << 32 - 1)]
												ifFalse: [a perform: comparison with: b].
							(bogus or: [error or: [self processor cResultRegister = 1 ~= expected]]) ifFalse: [nGood := nGood + 1].
							assertPrintBar
								ifTrue: [self assert: self processor cResultRegister = 1 equals: expected]
								ifFalse:
									[self processor cResultRegister = 1 ~= expected ifTrue:
										[aStream nextPutAll: srgetter; space; nextPutAll: comparison; space; nextPutAll: drgetter.
									 unsigned ifTrue: [aStream nextPutAll: ' (unsigned)'].
									 aStream nextPutAll: (error ifTrue: [' ERRORED'] ifFalse: [' is incorrect']); cr; flush.
									 bogus ifTrue:
										[self processor printRegistersOn: aStream.
										 aStream nextPutAll: (self processor disassembleInstructionAt: 0 In: memory); cr; flush]]]]]]]].
	assertPrintBar ifFalse:
		[aStream print: nTests; nextPutAll: ' tests; '; print: nGood; nextPutAll: ' good'; cr].
	^{nTests. nGood}! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'lw 8/8/2012 14:23'!
runCmpRdRd: assertPrintBar
	"self defaultTester runCmpRdRd: false"
	"self defaultTester runCmpRdRd: true"
	| resultReg result reg1 reg2 setreg1 setreg2 |
	self concreteCompilerClass new hasDoublePrecisionFloatingPointSupport ifFalse: [^true].
	resultReg := 0.
	self concreteCompilerClass dataRegistersWithAccessorsDo:
		[:n :get :set| n = resultReg ifTrue: [result := get]].
	self concreteCompilerClass fp64RegistersWithAccessorsDo:
		[:n :get :set|
		n = DPFPReg0 ifTrue: [reg1 := get. setreg1 := set].
		n = DPFPReg1 ifTrue: [reg2 := set. setreg2 := set]].
	#(	(genJumpFPGreater: > false nan)	(genJumpFPLessOrEqual: <= false nil)
		(genJumpFPLess: < false nil)				(genJumpFPGreaterOrEqual: <= true nan)
		(genJumpFPGreater: < true nan)		(genJumpFPGreaterOrEqual: >= false nan)
		(genJumpFPEqual: = false nan)		(genJumpFPNotEqual: ~= false nan)) do:
		[:tuple|
		[:opName :relation :invert :nan| | them inst jumpNotTaken jumpTaken nop memory |
		them := nan ifNil: [-2.0 to: 2.0] ifNotNil: [(-2.0 to: 2.0), { Float nan }].
		self resetGen.
		inst := invert
				ifTrue: [self gen: CmpRdRd operand: DPFPReg0 operand: DPFPReg1]
				ifFalse: [self gen: CmpRdRd operand: DPFPReg1 operand: DPFPReg0].
		jumpTaken := inst perform: opName with: 0.
		self gen: MoveCqR operand: 0 operand: resultReg.
		jumpNotTaken := self gen: Jump.
		jumpTaken jmpTarget: (self gen: MoveCqR operand: 1 operand: resultReg).
		jumpNotTaken jmpTarget: (nop := self gen: Nop).
		memory := self generateInstructions.
		assertPrintBar ifFalse:
			[Transcript print: tuple; cr.
			 self disassembleOpcodesIn: memory to: Transcript].
		them do:
			[:a|
			them do:
				[:b| | taken expected |
				self processor
					reset;
					perform: setreg1 with: a asIEEE64BitWord;
					perform: setreg2 with: b asIEEE64BitWord.
				[self processor singleStepIn: memory.
				 self processor pc ~= nop address] whileTrue.
				taken := (self processor perform: result) = 1.
				expected := a perform: relation with: b.
				assertPrintBar
					ifTrue:
						[self assert: taken = expected]
					ifFalse:
						[Transcript
							nextPutAll: 'CmpRdRd ', reg1, ' ', reg2; tab; tab;
							nextPutAll: reg1; nextPut: $(; print: a; nextPut: $); space;
							nextPutAll: relation; space;
							nextPutAll: reg2; nextPut: $(; print: b; nextPut: $); space;
							nextPutAll: ' = '; print: taken;  nextPutAll: ' ('; print: expected; nextPut: $).
						 taken ~= expected ifTrue:
							[Transcript nextPutAll: ' !!!!'; cr.
							 self processor printRegistersOn: Transcript.].
						 Transcript cr; flush]]]]
						valueWithArguments: tuple]! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'eem 11/13/2015 11:28'!
runSubCqR: assertPrintBar
	"self defaultTester runSubCqR: false"
	self concreteCompilerClass dataRegistersWithAccessorsDo:
		[:reg :rgetter :rsetter|
		self pairs: (-2 to: 2)  do:
			[:a :b| | inst len bogus memory |
			inst := self gen: SubCqR operand: a operand: reg.
			len := inst concretizeAt: 0.
			memory := self memoryAsBytes: inst machineCode.
			self processor
				reset;
				perform: rsetter with: (processor convertIntegerToInternal: b).
			[[processor pc < len] whileTrue:
				[self processor singleStepIn: memory]]
				on: Error
				do: [:ex| ].
			"self processor printRegistersOn: Transcript.
			 Transcript show: (self processor disassembleInstructionAt: 0 In: memory); cr"
			assertPrintBar
				ifTrue: [self assert: processor pc = inst machineCodeSize.
						self assertCheckQuickArithOpCodeSize: inst machineCodeSize]
				ifFalse: [bogus := processor pc ~= inst machineCodeSize].
			self concreteCompilerClass dataRegistersWithAccessorsDo:
				[:ireg :getter :setter| | expected |
				expected := getter == rgetter ifTrue: [b - a] ifFalse: [0].
				assertPrintBar
					ifTrue: [self assert: (self processor convertInternalToInteger: (self processor perform: getter)) equals: expected]
					ifFalse:
						[(self processor convertInternalToInteger: (self processor perform: getter)) ~= expected ifTrue:
							[bogus := true]]].
				assertPrintBar ifFalse:
					[Transcript
						nextPutAll: rgetter; nextPut: $(; print: b; nextPutAll: ') - '; print: a; nextPutAll: ' = ';
						print: (self processor convertInternalToInteger: (self processor perform: rgetter)); cr; flush.
					 bogus ifTrue:
						[self processor printRegistersOn: Transcript.
						 Transcript show: (self processor disassembleInstructionAt: 0 In: memory); cr]]]]! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'eem 12/25/2015 18:45'!
runSubCwR: assertPrintBar
	"self defaultTester runSubCwR: false"
	self concreteCompilerClass dataRegistersWithAccessorsDo:
		[:reg :rgetter :rsetter|
		self pairs: (-2 to: 2)  do:
			[:a :b| | inst len bogus memory |
			inst := self gen: SubCwR operand: a operand: reg.
			len := inst concretizeAt: 0.
			memory := self memoryAsBytes: inst machineCode.
			self processor
				reset;
				perform: rsetter with: (self processor convertIntegerToInternal: b).
			[[processor pc < len] whileTrue:
				[self processor singleStepIn: memory]]
				on: Error
				do: [:ex| ].
			"self processor printRegistersOn: Transcript.
			 Transcript show: (self processor disassembleInstructionAt: 0 In: memory); cr"
			assertPrintBar
				ifTrue: [self assert: processor pc = inst machineCodeSize.
						self assertCheckLongArithOpCodeSize: inst machineCodeSize]
				ifFalse: [bogus := processor pc ~= inst machineCodeSize].
			self concreteCompilerClass dataRegistersWithAccessorsDo:
				[:ireg :getter :setter| | expected |
				(self concreteCompilerClass isRISCTempRegister: ireg) ifFalse:
					[expected := getter == rgetter ifTrue: [b - a] ifFalse: [0].
					assertPrintBar
						ifTrue: [self assert: (self processor convertInternalToInteger: (self processor perform: getter)) equals: expected]
						ifFalse:
							[(self processor convertInternalToInteger: (self processor perform: getter)) ~= expected ifTrue:
								[bogus := true]]].
					assertPrintBar ifFalse:
						[Transcript
							nextPutAll: rgetter; nextPut: $(; print: b; nextPutAll: ') - '; print: a; nextPutAll: ' = ';
							print: (self processor convertInternalToInteger: (self processor perform: rgetter)); cr; flush.
						 bogus ifTrue:
							[self processor printRegistersOn: Transcript.
							 Transcript show: (self processor disassembleInstructionAt: 0 In: memory); cr]]]]]! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'eem 12/20/2019 19:30'!
runSubRR: assertPrintBar on: aStream
	"Compile and evaluate as many combinations of SubR:R: as possible, checking that they produce the
	 expected result.  Answer an array of the number of comparisons and the number of them that succeeded."
	"self defaultTester runSubRR: false"
	| cogit nTests nGood |
	cogit := self cogitForTests.
	nTests := nGood := 0.
	self concreteCompilerClass dataRegistersWithAccessorsGiven: self processor do:
		[:sreg :srgetter :srsetter|
		self concreteCompilerClass dataRegistersWithAccessorsGiven: self processor do:
			[:dreg :drgetter :drsetter| | memory limit |
			(drgetter == #sp						"Some processors can't do this easily; i.e. ARMv8"
			 or: [drgetter == srgetter]) ifFalse:		"Some processors refuse to do this; i.e. ARMv8"
				[cogit resetGen.
				cogit
					SubR: sreg R: dreg;
					Nop.
				memory := self generatedCodeFrom: cogit.
				limit := cogit codeSize.
				self pairs: (-2 to: 2) do:
					[:a :b| | bogus error nInsts allRegistersGood |
					nTests := nTests + 1.
					error := false.
					nInsts := 0.
					self processor
						reset;
						pc: self defaultCodeStart;
						perform: srsetter with: (processor convertIntegerToInternal: a);
						perform: drsetter with: (processor convertIntegerToInternal: b).
					[[processor pc < limit and: [nInsts < 16]] whileTrue:
						[processor singleStepIn: memory. nInsts := nInsts + 1]]
						on: Error
						do: [:ex| error := true].
					nInsts >= 16 ifTrue: [error := true].
					"self processor disassembleInstructionAt: self defaultCodeStart In: memory"
					bogus := processor pc ~= limit.
					assertPrintBar ifTrue: [self assert: processor pc = limit].
					allRegistersGood := (bogus or: [error]) not.
					self concreteCompilerClass dataRegistersWithAccessorsGiven: self processor do:
						[:ireg :getter :setter| | expected |
						expected := drgetter == srgetter
										ifTrue: [0]
										ifFalse:
											[getter == drgetter
												ifTrue: [b - a]
												ifFalse: [getter = srgetter
															ifTrue: [a]
															ifFalse: [0]]].
						assertPrintBar
							ifTrue: [self assert: (self processor convertInternalToInteger: (self processor perform: getter)) equals: expected]
							ifFalse:
								[(self processor convertInternalToInteger: (self processor perform: getter)) ~= expected ifTrue:
									[bogus := true]].
						(self processor convertInternalToInteger: (self processor perform: getter)) ~= expected ifTrue: [allRegistersGood := false].
						assertPrintBar ifFalse:
							[aStream
								nextPutAll: drgetter; nextPut: $(; print: b; nextPutAll: ') - ';
								nextPutAll: srgetter; nextPut: $(; print: a; nextPutAll: ') = ';
								print: (self processor convertInternalToInteger: (self processor perform: drgetter)); cr; flush]].
					allRegistersGood ifTrue: [nGood := nGood + 1]]]]].
	assertPrintBar ifFalse:
		[aStream print: nTests; nextPutAll: ' tests; '; print: nGood; nextPutAll: ' good'; cr].
	^{nTests. nGood}! !

!AbstractInstructionTests methodsFor: 'private' stamp: 'eem 12/18/2019 23:57'!
sanityCheckRegisterEnumerator
	^String streamContents:
		[:s|
		self concreteCompilerClass dataRegistersWithAccessorsGiven: self processor do:
			[:regNum :getter :setter|
			s print: regNum; nextPutAll: ': '; nextPutAll: getter; space; nextPutAll: setter; cr]]

	"self subclasses collect: [:class| class -> class new sanityCheckRegisterEnumerator]"! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'eem 12/25/2015 19:25'!
setUp
	self concreteCompilerClass initializeAbstractRegisters! !

!AbstractInstructionTests methodsFor: 'tests' stamp: 'eem /9/18/2010 13:14'!
testBinaryConditionalJumps
	"self defaultTester testBinaryConditionalJumps"
	"self defaultTester runBinaryConditionalJumps: false"
	self runBinaryConditionalJumps: true! !

!AbstractInstructionTests methodsFor: 'tests' stamp: 'eem /9/18/2010 13:14'!
testBinaryConditionalJumpsViaCmpCqR
	"self defaultTester testBinaryConditionalJumpsViaCmpCqR"
	"self defaultTester runBinaryConditionalJumpsViaCmpCqR: false"
	self runBinaryConditionalJumpsViaCmpCqR: true! !

!AbstractInstructionTests methodsFor: 'tests' stamp: 'eem 9/18/2010 14:28'!
testBinaryConditionalJumpsViaSubCqR
	"self defaultTester testBinaryConditionalJumpsViaSubCqR"
	"self defaultTester runBinaryConditionalJumpsViaSubCqR: false"
	self runBinaryConditionalJumpsViaSubCqR: true! !

!AbstractInstructionTests methodsFor: 'tests' stamp: 'eem 12/19/2019 15:03'!
testCogitForTests
	self assert: (self cogitForTests backEnd isKindOf: self concreteCompilerClass)! !

!AbstractInstructionTests methodsFor: 'running' stamp: 'nice 12/30/2020 23:36'!
testNegateR
	"self defaultTester testNegateR"
	self concreteCompilerClass dataRegistersWithAccessorsDo:
		[:reg :rgetter :rsetter|
		-2 to: 2 do:
			[:a| | inst len memory |
			inst := self gen: NegateR operand: reg.
			len := inst concretizeAt: 0.
			memory := self memoryAsBytes: inst machineCode.
			self processor
				reset;
				perform: rsetter with: (processor convertIntegerToInternal: a).
			[[processor pc < len] whileTrue:
				[processor singleStepIn: memory]]
				on: Error
				do: [:ex| ].
			"self processor printRegistersOn: Transcript.
			 Transcript show: (self processor disassembleInstructionAt: 0 In: memory); cr"
			self assert: processor pc equals: inst machineCodeSize.
			self concreteCompilerClass dataRegistersWithAccessorsDo:
				[:ireg :getter :setter| | expected |
				expected := getter == rgetter ifTrue: [ a negated ] ifFalse: [0].
				self assert: (processor convertInternalToInteger: (processor perform: getter)) equals: expected]]]! !

!AbstractInstructionTests methodsFor: 'tests' stamp: 'eem /9/18/2010 13:14'!
testRunAddCqR
	"self defaultTester testRunAddCqR"
	self runAddCqR: true! !

!AbstractInstructionTests methodsFor: 'tests' stamp: 'eem /9/18/2010 13:14'!
testRunAddCwR
	"self defaultTester testRunAddCwR"
	self runAddCwR: true! !

!AbstractInstructionTests methodsFor: 'tests' stamp: 'eem /9/18/2010 13:14'!
testRunAddRR
	"self defaultTester testRunAddRR"
	self runAddRR: true! !

!AbstractInstructionTests methodsFor: 'tests' stamp: 'eem /9/18/2010 13:14'!
testRunArithmeticShiftRightRR
	"self defaultTester testRunArithmeticShiftRightRR"
	self runArithmeticShiftRightRR: true! !

!AbstractInstructionTests methodsFor: 'tests' stamp: 'eem 12/19/2019 14:57'!
testRunCmpRR
	"self defaultTester testRunSubRR"
	| result |
	result := self runCmpRRJumpCond: true on: NullStream new.
	"Assume at least 5 registers.  There are ten comparisons. and we compare 5 values"
	self assertValidRunResult: result ofAtLeast: 5 squared * 10 * 5 squared! !

!AbstractInstructionTests methodsFor: 'tests' stamp: 'eem 9/18/2010 16:44'!
testRunCmpRdRd
	"self defaultTester testRunCmpRdRd"
	self runCmpRdRd: true! !

!AbstractInstructionTests methodsFor: 'tests' stamp: 'eem /9/18/2010 13:14'!
testRunSubCqR
	"self defaultTester testRunSubCqR"
	self runSubCqR: true! !

!AbstractInstructionTests methodsFor: 'tests' stamp: 'eem /9/18/2010 13:14'!
testRunSubCwR
	"self defaultTester testRunSubCwR"
	self runSubCwR: true! !

!AbstractInstructionTests methodsFor: 'tests' stamp: 'eem 12/19/2019 15:44'!
testRunSubRR
	"self defaultTester testRunSubRR"
	| result |
	result := self runSubRR: true on: NullStream new.
	"Assume at least 5 registers. We subtract 5 pairs"
	self assertValidRunResult: result ofAtLeast: 5 squared * 5 squared! !

!AbstractInstructionTests methodsFor: 'cogit compatibility' stamp: 'eem 11/13/2015 11:23'!
varBaseAddress
	"Answer a value that should be sufficiently high that var base relative addressing is never generated."
	^1 << 60! !

AbstractInstructionTests subclass: #CogARMCompilerTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!CogARMCompilerTests methodsFor: 'running' stamp: 'lw 7/31/2012 15:02'!
assertCheckLongArithOpCodeSize: bytes
	self assert: bytes > 4! !

!CogARMCompilerTests methodsFor: 'running' stamp: 'lw 7/31/2012 14:25'!
assertCheckQuickArithOpCodeSize: bytes
	"The problem is that there are negative value, which are not quick encodable in ARM"
	self assert: bytes <= 4! !

!CogARMCompilerTests methodsFor: 'accessing' stamp: 'lw 7/16/2012 12:01'!
concreteCompilerClass
	^CogARMCompilerForTests! !

!CogARMCompilerTests methodsFor: 'private' stamp: 'eem 9/17/2015 10:04'!
disassembleInstructionAt: index In: inst into: binaryBlock
	"Manage the fact that in the simulator inst machineCode object is an Array and the disassembler requires a ByteArray or some such."
	^self processor
		disassembleInstructionAt: index
		In: (self memoryAsBytes: inst machineCode object)
		into: binaryBlock! !

!CogARMCompilerTests methodsFor: 'accessing' stamp: 'eem 12/19/2019 00:02'!
isaName
	^#ARMv5! !

!CogARMCompilerTests methodsFor: 'private' stamp: 'eem 11/13/2015 11:22'!
memoryAsBytes: instructionMachineCode
	"Manage the fact that in the simulator inst machineCode object is an Array and the disassembler requires a ByteArray or some such."
	| bytes |
	instructionMachineCode isCObjectAccessor ifTrue:
		[^self memoryAsBytes: instructionMachineCode object].
	instructionMachineCode isArray ifFalse:
		[self assert: instructionMachineCode class isBits.
		 ^instructionMachineCode].
	bytes := ByteArray new: instructionMachineCode size * 4.
	1 to: instructionMachineCode size do:
		[:i|
		(instructionMachineCode at: i) ifNotNil:
			[:word|
			bytes unsignedLongAt: i - 1* 4 + 1 put: word]].
	^bytes! !

!CogARMCompilerTests methodsFor: 'cogit compatibility' stamp: 'eem 11/13/2015 10:06'!
objectMemory
	^self! !

!CogARMCompilerTests methodsFor: 'accessing' stamp: 'lw 7/16/2012 12:02'!
processor
	processor ifNil:
		[processor := GdbARMAlien new].
	^processor! !

!CogARMCompilerTests methodsFor: 'accessing' stamp: 'tpr 4/24/2014 17:24'!
strip: aFancyDisassembledInstruction
	"When an immediate > 15, the disassembler appends '	; 0x\d\d'. That is stripped. Also strip any prepended stuff delimted by $: - perhaps memory addresses etc?"
	^((aFancyDisassembledInstruction 
		allButFirst: (aFancyDisassembledInstruction indexOf: $: ifAbsent:[-1]) + 1)
			copyUpTo: $;) withBlanksTrimmed! !

!CogARMCompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testAdd
	"self new testAdd"
	
	"the forms are valid, "
	"test AddCqR"
	self concreteCompilerClass registersWithNamesDo: [ :reg :regName |
		#(0 16rF 16rFF) do:
			[:n| | inst len |
			inst := self gen: AddCqR operand: n operand: reg.
			len := inst concretizeAt: 0.
			self
				disassembleInstructionAt: 0
				In: inst
				into: [:str :sz| | plainJane herIntended |
					plainJane := self strip: str.
					herIntended := 'adds	', regName, ', ', regName, ', #', n asString.
					self assert: herIntended equals: plainJane]]].
		
	"test AddCwR"
	self concreteCompilerClass registersWithNamesDo: [ :reg :regName |
		#(16rFFFFFFFF 16r88888888 0) do:
			[:n| | inst len |
			inst := self gen: AddCwR operand: n operand: reg.
			len := inst concretizeAt: 0.
			self
				disassembleInstructionAt: 0
				In: inst
				into: [:str :sz| | plainJane herIntended |
					plainJane := self strip: str.
					herIntended := 'mov	ip, #', (n bitAnd: 16rFF << 24) signedIntFromLong asString.
					self assert: herIntended equals: plainJane].
			self
				disassembleInstructionAt: 4
				In: inst
				into: [:str :sz| | plainJane herIntended |
					plainJane := self strip: str.
					herIntended := 'orr	ip, ip, #', (n bitAnd: 16rFF << 16) asString.
					self assert: herIntended equals: plainJane].
			self
				disassembleInstructionAt: 8
				In: inst
				into: [:str :sz| | plainJane herIntended |
					plainJane := self strip: str.
					herIntended := 'orr	ip, ip, #', (n bitAnd: 16rFF << 8) signedIntFromLong asString.
					self assert: herIntended equals: plainJane].
			self
				disassembleInstructionAt: 12
				In: inst
				into: [:str :sz| | plainJane herIntended |
					plainJane := self strip: str.
					herIntended := 'orr	ip, ip, #', (n bitAnd: 16rFF) asString.
					self assert: herIntended equals: plainJane].
			self
				disassembleInstructionAt: 16
				In: inst
				into: [:str :sz| | plainJane herIntended |
					plainJane := self strip: str.
					herIntended := 'adds	', regName, ', ', regName, ', ip'.
					self assert: herIntended equals: plainJane]]]
! !

!CogARMCompilerTests methodsFor: 'tests' stamp: 'tpr 3/23/2015 18:00'!
testGenerateMul
	"self new testGenerateMul"
	"Currently demonstrates that the CogARMCompiler>genMulR:R: algorithm ain't gonna work too well"
	
	|memory compiler |
	compiler := self gen: nil. "grab something usefully like a compiler for later abuse"
	self resetGen. "initialise important stuff"

	"test MulRR"
	self gen: LogicalShiftLeftCqR operand: 1 operand:1. "LSL r1 - to make the right bit pattern for a SmallInt that has had the flag masked out"
	self gen: LogicalShiftLeftCqR operand: 1 operand:0. "LSL r1 - to make the right bit pattern for a SmallInt that has had the flag masked out"
	self gen: MoveRR operand: 0 operand: 2.
	compiler genMulR: 1 R: 0.
	memory := self generateInstructions.
	self disassembleOpcodesIn: memory to: Transcript	.
	{SmallInteger maxVal@2. SmallInteger minVal@2. 0@0. 1@1. -1@0. -1@1. -1@ -1} do:[:val | | x y |
			x := val x.
			y := val y.
			self processor reset.
			processor signedLongAt: 13 put: x;
				signedLongAt: 17 put: y.
			[[processor pc < memory size] whileTrue:
						[self processor singleStepIn: memory]]
						on: Error
						do: [:ex| ].
			Transcript show: x asString, ' * ', y asString; cr.
			processor printRegisterState: processor registerState on: Transcript.
			Transcript flush]
	
! !

!CogARMCompilerTests methodsFor: 'tests' stamp: 'eem 11/13/2015 11:18'!
testMoveCwR
	"self new testMoveCwR"
	
	#(16rFF00FF00 16r00000012 16r12345678) do:
		[:n|
		self concreteCompilerClass dataRegistersWithAccessorsDo: 
			[ :r :rgetter :rset | | inst len memory |
			inst := self gen: MoveCwR operand: n operand: r.
			len := inst concretizeAt: 0.
			memory := self memoryAsBytes: inst machineCode.
			self processor
				reset.
			[[processor pc < inst machineCodeSize] whileTrue:
				[self processor singleStepIn: memory]]
				on: Error
				do: [:ex| ].
			self concreteCompilerClass dataRegistersWithAccessorsDo:
				[:ireg :getter :setter| | expected |
				expected := getter == rgetter ifTrue: [ n ] ifFalse: [0].
				self assert: (self processor perform: getter) = expected].
			self assert: self processor pc = 16]]
! !

!CogARMCompilerTests methodsFor: 'tests' stamp: 'eem 11/13/2015 11:34'!
testPerformAdd
	"self new testPerformAdd"
	
	| memory |
	#(16rFF00FF00 16r00000012 16r12345678) do:
		[:n| | |
		self concreteCompilerClass dataRegistersWithAccessorsDo: 
			[ :r :rgetter :rset |
			self resetGen. "initialise important stuff"
 			self gen: MoveCwR operand: n operand: r.
			self gen: AddCqR operand: 42 operand: r.
			memory := self generateInstructions.
			"self disassembleOpcodesIn: memory to: Transcript."
			self processor
				reset.
			[[processor pc < memory size] whileTrue:
				[self processor singleStepIn: memory]]
				on: Error
				do: [:ex| ].
			self concreteCompilerClass dataRegistersWithAccessorsDo:
				[:ireg :getter :setter| | expected |
				expected := getter == rgetter ifTrue: [ n + 42] ifFalse: [0].
				self assert: (self processor perform: getter) = expected].
			self assert: self processor pc = 20]]
! !

!CogARMCompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testPrefetchAw
	"self new testPrefetchAw"
	
	#(16rFF00FF00 16r00000012) do:
		[:n| | inst len |
		inst := self gen: PrefetchAw operand: n.
		len := inst concretizeAt: 0.
		self
			disassembleInstructionAt: 12
			In: inst
			into: [:str :sz| | plainJane herIntended |
				"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
				plainJane := self strip: str.
				herIntended := 'pld	[sl', ((n bitAnd: 16rFF) = 0 ifTrue: [''] ifFalse: [', #', (n bitAnd: 16rFF) asString]) ,']'.
				self assert: herIntended equals: plainJane.
				self assert: len = 16]].
! !

!CogARMCompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testPushR
	"self new testPushR"
	
	self concreteCompilerClass registersWithNamesDo:
		[:r :name | | inst len |
		inst := self gen: PushR operand: r.
		len := inst concretizeAt: 0.
		self
			disassembleInstructionAt: 0
			In: inst
			into: [:str :sz| | plainJane herIntended |
				"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
				plainJane := self strip: str.
				herIntended := 'push	{', name ,'}'.
				self assert: herIntended equals: plainJane.
				self assert: len = 4]].
! !

!CogARMCompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testRetN
	"self new testRetN"
	
	#(0) do:
		[:n| | inst len |
		inst := self gen: RetN operand: n.
		len := inst concretizeAt: 0.
		self
			disassembleInstructionAt: 0
			In: inst
			into: [:str :sz| | plainJane  |
				"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
				plainJane := self strip: str.
				self assert: (plainJane match: 'mov	pc, lr').
				self assert: len = sz]].
	
	#(1 2 3 4 5 6 7) do:
		[:n| | inst len |
		inst := self gen: RetN operand: n * 4.
		len := inst concretizeAt: 0.
		self
			disassembleInstructionAt: 0
			In: inst
			into: [:str :sz| | plainJane herIntended |
				plainJane := self strip: str.
				herIntended := 'add	sp, sp, #', (n * 4 )asString.
				self assert: herIntended equals: plainJane].
		self
			disassembleInstructionAt: 4
			In: inst
			into: [:str :sz| | plainJane |
				plainJane := self strip: str.
				self assert: (plainJane match: 'mov	pc, lr').		
		self assert: len = 8]]! !

!CogARMCompilerTests methodsFor: 'cogit compatibility' stamp: 'eem 11/13/2015 10:07'!
wordSize
	^4! !

AbstractInstructionTests subclass: #CogARMv8CompilerTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!CogARMv8CompilerTests methodsFor: 'accessing' stamp: 'eem 12/18/2019 21:05'!
concreteCompilerClass
	^CogARMv8CompilerForTests! !

!CogARMv8CompilerTests methodsFor: 'accessing' stamp: 'eem 12/18/2019 22:00'!
isaName
	^#ARMv8! !

!CogARMv8CompilerTests methodsFor: 'accessing' stamp: 'eem 12/18/2019 21:14'!
processor
	processor ifNil:
		[processor := GdbARMv8Alien new].
	^processor! !

AbstractInstructionTests subclass: #CogIA32CompilerTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!CogIA32CompilerTests methodsFor: 'abstract instructions' stamp: 'eem 9/18/2010 18:21'!
Label
	^self gen: Label operand: opcodes size! !

!CogIA32CompilerTests methodsFor: 'running' stamp: 'eem 1/20/2009 15:14'!
assertCheckLongArithOpCodeSize: bytes
	self assert: bytes > 4! !

!CogIA32CompilerTests methodsFor: 'running' stamp: 'eem 1/20/2009 15:10'!
assertCheckQuickArithOpCodeSize: bytes
	self assert: bytes <= 3! !

!CogIA32CompilerTests methodsFor: 'accessing' stamp: 'eem 1/20/2009 14:55'!
concreteCompilerClass
	^CogIA32CompilerForTests! !

!CogIA32CompilerTests methodsFor: 'accessing' stamp: 'eem 12/18/2019 22:57'!
isaName
	^#IA32! !

!CogIA32CompilerTests methodsFor: 'accessing' stamp: 'eem 9/17/2015 10:17'!
methodLabel
	"There is no methodLabel in test code."
	^nil! !

!CogIA32CompilerTests methodsFor: 'accessing' stamp: 'eem 1/20/2009 14:54'!
processor
	processor ifNil:
		[processor := BochsIA32Alien new].
	^processor! !

!CogIA32CompilerTests methodsFor: 'accessing' stamp: 'eem 1/20/2009 14:54'!
strip: aFancyDisassembledInstruction
	"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
	^((((aFancyDisassembledInstruction
		copyReplaceAll: '%ds:' with: '')
			copyReplaceAll: '%ss:' with: '')
				allButFirst: (aFancyDisassembledInstruction indexOf: $:) + 1)
					copyUpTo: $:)
						allButLast: 1! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testAddCqR
	"self new testAddCqR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname|
		#(16r1 16r3 16r7 16r555555 16rAAAAAA) do:
			[:const| | inst len|
			inst := self gen: AddCqR operand: const operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'addl $0x', (const printStringBase: 16 length: 8 padded: true), ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testAndCqR
	"self new testAndCqR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname|
		#(16r1 16r3 16r7 16r555555 16rAAAAAA) do:
			[:const| | inst len|
			inst := self gen: AndCqR operand: const operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'andl $0x', (const printStringBase: 16 length: 8 padded: true), ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 1/20/2009 17:50'!
testArithmeticShiftRightRR
	"CogIA32CompilerTests new testArithmeticShiftRightRR"
	self concreteCompilerClass dataRegistersWithAccessorsDo:
		[:sreg :srgetter :srsetter|
		self concreteCompilerClass dataRegistersWithAccessorsDo:
			[:dreg :drgetter :drsetter| | inst len |
			inst := self gen: ArithmeticShiftRightRR operand: sreg operand: dreg.
			len := inst concretizeAt: 0.
			self assert: len = (srgetter = #ecx
								ifTrue: [2]
								ifFalse:
									[srgetter = #eax
										ifTrue: [4]
										ifFalse: [6]])
			"self processor disassembleFrom: 0 to: inst machineCodeSize in: inst machineCode object on: Transcript"]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testCMPXCHGAwR
	"self new testCMPXCHGAwR"
	| cmpxchgAwR |
	cmpxchgAwR := CogIA32Compiler classPool at: #CMPXCHGAwR.
	CogIA32CompilerForTests registersWithNamesDo:
		[:reg :regname|
		#(16r555555 16rAAAAAA) do:
			[:addr| | inst len |
			inst := self gen: cmpxchgAwR operand: addr operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'cmpxchgl ', regname, ', 0x', (addr hex allButFirst: 3).
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testCMPXCHGMwrR
	"self new testCMPXCHGMwrR"
	| cmpxchgMwrR lock |
	cmpxchgMwrR := CogIA32Compiler classPool at: #CMPXCHGMwrR.
	lock := CogIA32Compiler classPool at: #LOCK.
	CogIA32CompilerForTests registersWithNamesDo:
		[:sreg :srname|
		CogIA32CompilerForTests registersWithNamesDo:
			[:dreg :drname|
			((1 to: 19 by: 3) collect: [:po2| 2 raisedToInteger: po2]) do:
				[:offset|
				#(false true) do:
					[:prefixLock| | memory |
					self resetGen.
					prefixLock ifTrue: [self gen: lock].
					self gen: cmpxchgMwrR operand: offset operand: sreg operand: dreg.
					memory := self generateInstructions.
					self processor
						disassembleInstructionAt: 0
						In: memory
						into: [:str :sz| | plainJane herIntended |
							plainJane := self strip: str.
							herIntended := (prefixLock ifTrue: ['lock '] ifFalse: ['']),
											'cmpxchgl ', drname, ', 0x', (offset hex allButFirst: 3), '(', srname, ')'.
							self assert: herIntended equals: plainJane.
							self assert: memory size = sz]]]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testCmpCqR
	"self new testCmpCqR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname|
		#(16r1 16r3 16r7 16r555555 16rAAAAAA) do:
			[:const| | inst len|
			inst := self gen: CmpCqR operand: const operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'cmpl $0x', (const printStringBase: 16 length: 8 padded: true), ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 10/21/2010 11:32'!
testDivQuoRem
	"| failures ops |
	 failures := Set new.
	 [ops := (CogIA32CompilerTests new testDivQuoRem)]
		on: TestResult failure
		do: [:ex| | c |
			c := ex signalerContext.
			[c tempNames includes: 'op'] whileFalse:
				[c := c sender].
			failures add: (c namedTempAt: (c tempNames indexOf: 'op')).
			ex resume].
	 { ops size. failures size. ops asSortedCollection asArray. failures asSortedCollection asArray}"
	| map compiler memory ops |
	map := Dictionary new.
	compiler := self gen: nil.
	memory := ByteArray new: 4096 * 2.
	ops := Set new.
	CogIA32CompilerForTests dataRegistersWithAccessorsExcept: #(ebp esp) do:
		[:sreg :srget :srset|
		CogIA32CompilerForTests dataRegistersWithAccessorsExcept: { #ebp. #esp. srget } do:
			[:dreg :drget :drset|
			 | instructions op |
			CogIA32CompilerForTests dataRegistersWithAccessorsExcept: #(ebp esp) do:
				[:qreg :qrget :qrset| 
				CogIA32CompilerForTests dataRegistersWithAccessorsExcept: { #ebp. #esp. qrget } do:
					[:rreg :rrget :rrset|
					self resetGen.
					op := qrget, ' := ', drget, ' quo: ', srget, '. ', rrget, ' := ', drget, ' rem: ', srget.
					ops add: op.
					compiler genDivR: sreg R: dreg Quo: qreg Rem: rreg.
					instructions := self generateInstructions.
					memory atAllPut: 0; replaceFrom: 1 to: instructions size with: instructions startingAt: 1; at: instructions size + 1 put: self processor nopOpcode.
					#(-768 -456 -123 123 456 789)
						with: #(987 654 321 -321 -654 -987)
						do: [:dd :dv| "| calc |
							calc := ((dd quo: dv) bitAnd: 16rFFFFFFFF) hex, ' := ', (dd bitAnd: 16rFFFFFFFF) hex, ' quo: ', (dv bitAnd: 16rFFFFFFFF) hex, '. ', ((dd rem: dv) bitAnd: 16rFFFFFFFF) hex, ' := ', (dd bitAnd: 16rFFFFFFFF) hex, ' rem: ', (dv bitAnd: 16rFFFFFFFF) hex.
							calc := calc."
							"Transcript cr; cr; nextPutAll: op; cr; nextPutAll: calc; cr.
							 self processor
								disassembleFrom: 0 to: instructions size in: memory on: Transcript;
								printIntegerRegistersOn: Transcript."
							map
								at: #eax put: (self processor eax: 16rA5A5A5A5);
								at: #ebx put: (self processor ebx: 16rB5B5B5B5);
								at: #ecx put: (self processor ecx: 16rC5C5C5C5);
								at: #edx put: (self processor edx: 16rD5D5D5D5);
								at: #esi put: (self processor esi: 16r51515151);
								at: #edi put: (self processor edi: 16rD1D1D1D1);
								at: srget put: (self processor perform: srset with: dv signedIntToLong);
								at: drget put: (self processor perform: drset with: dd signedIntToLong).
							self processor esp: memory size; eip: 0.
							self shouldnt:
								[[self processor pc < instructions size] whileTrue:
									[self processor singleStepIn: memory readExecuteOnlyBelow: memory size / 2]]
								raise: Error.
							map
								at: qrget put: (dd quo: dv) signedIntToLong;
								at: rrget put: (dd rem: dv) signedIntToLong.
							map keysAndValuesDo:
								[:accessor :value|
								self assert: value = (self processor perform: accessor)]]]]]].
	^ops! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/3/2015 09:22'!
testLogicalShiftRightCqR
	"CogIA32CompilerTests new testLogicalShiftRightCqR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname|
		1 to: 31 do:
			[:shift| | inst len |
			inst := self gen: LogicalShiftRightCqR operand: shift operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'shrl $', (shift > 1
												ifTrue: ['0x', (shift printStringBase: 16 length: 2 padded: true)]
												ifFalse: [shift asInteger]), ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testMoveAbR
	"self new testMoveAbR"
	CogIA32CompilerForTests registersWithNamesDo:
		[:reg :regname|
		#(16r555555 16rAAAAAA) do:
			[:addr| | inst len |
			inst := self gen: MoveAbR operand: addr operand: reg.
			len := inst concretizeAt: 0.
			self assert: inst byteReadsZeroExtend.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'movzbl 0x', (addr hex allButFirst: 3), ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testMoveAwR
	"self new testMoveAwR"
	CogIA32CompilerForTests registersWithNamesDo:
		[:reg :regname|
		#(16r555555 16rAAAAAA) do:
			[:addr| | inst len |
			inst := self gen: MoveAwR operand: addr operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'movl 0x', (addr hex allButFirst: 3), ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 1/12/2016 14:26'!
testMoveM16rR
	"self new testMoveM16rR"
	CogIA32CompilerForTests registersWithNamesDo:
		[:sreg :srname|
		CogIA32CompilerForTests registersWithNamesDo:
			[:dreg :drname|
			((1 to: 19 by: 3) collect: [:po2| 2 raisedToInteger: po2]) do:
				[:offset| | inst len |
				inst := self gen: MoveM16rR operand: offset operand: sreg operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'movzwl 0x', (offset hex allButFirst: 3), '(', srname, '), ', drname.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testMoveM64rRd
	"self new testMoveM64rRd"
	CogIA32CompilerForTests registersWithNamesDo:
		[:sreg :srname|
		CogIA32CompilerForTests xmmRegistersWithNamesDo:
			[:dreg :drname|
			((1 to: 19 by: 3) collect: [:po2| 2 raisedToInteger: po2]) do:
				[:offset| | inst len |
				inst := self gen: MoveM64rRd operand: offset operand: sreg operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'movsd 0x', (offset hex allButFirst: 3), '(', srname, '), ', drname.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testMoveMbrR
	"self new testMoveMbrR"
	CogIA32CompilerForTests registersWithNamesDo:
		[:sreg :srname|
		CogIA32CompilerForTests registersWithNamesDo:
			[:dreg :drname|
			((1 to: 19 by: 3) collect: [:po2| 2 raisedToInteger: po2]) do:
				[:offset| | inst len |
				inst := self gen: MoveMbrR operand: offset operand: sreg operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'movzbl 0x', (offset hex allButFirst: 3), '(', srname, '), ', drname.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testMoveMwrR
	"self new testMoveMwrR"
	self concreteCompilerClass registersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :drname| | inst len |
			#(0 64 65536 -64 -65536) do:
				[:offset|
				inst := self gen: MoveMwrR operand: offset operand: sreg operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended po |
						plainJane := self strip: str.
						po := offset bitAnd: 1 << self processor bitsInWord - 1.
						herIntended := 'movl ', (offset = 0 ifTrue: [''] ifFalse: ['0x', (po printStringBase: 16)]), '(', srname, '), ', drname.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]

	"| failures |
	 failures := Dictionary new.
	 self concreteCompilerClass registersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :drname| | inst len |
			#(0 64 65536 -64 -65536) do:
				[:offset|
				inst := self gen: MoveMwrR operand: offset operand: sreg operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended po |
						plainJane := self strip: str.
						po := offset bitAnd: 1 << self processor bitsInWord - 1.
						herIntended := 'movl ', (offset = 0 ifTrue: [''] ifFalse: ['0x', (po printStringBase: 16)]), '(', srname, '), ', drname.
						((plainJane match: herIntended) and: [len = sz]) ifFalse:
							[failures at: herIntended put: plainJane]]]]].
	 failures"! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testMoveRAb
	"self new testMoveRAb"
	CogIA32CompilerForTests byteRegistersWithNamesDo:
		[:reg :regname|
		#(16r555555 16rAAAAAA) do:
			[:addr| | inst len |
			inst := self gen: MoveRAb operand: reg operand: addr.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					plainJane := self strip: str.
					herIntended := 'movb ', regname, ', 0x', (addr hex allButFirst: 3).
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testMoveRAw
	"self new testMoveRAw"
	CogIA32CompilerForTests registersWithNamesDo:
		[:reg :regname|
		#(16r555555 16rAAAAAA) do:
			[:addr| | inst len |
			inst := self gen: MoveRAw operand: reg operand: addr.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					plainJane := self strip: str.
					herIntended := 'movl ', regname, ', 0x', (addr hex allButFirst: 3).
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 1/12/2016 15:14'!
testMoveRM16r
	"self new testMoveRM16r"
	CogIA32CompilerForTests registersWithNamesDo:
		[:sreg :srname|
		CogIA32CompilerForTests registersWithNamesDo:
			[:dreg :drname|
			((1 to: 19 by: 3) collect: [:po2| 2 raisedToInteger: po2]) do:
				[:offset| | inst len |
				inst := self gen: MoveRM16r operand: sreg operand: offset operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'movw ', (srname copyWithout: $e), ', 0x', (offset hex allButFirst: 3), '(', drname, ')'.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testMoveRMbr
	"self new testMoveRMbr"
	CogIA32CompilerForTests byteRegistersWithNamesDo:
		[:sreg :srname|
		CogIA32CompilerForTests registersWithNamesDo:
			[:dreg :drname|
			((1 to: 19 by: 3) collect: [:po2| 2 raisedToInteger: po2]) do:
				[:offset| | inst len |
				inst := self gen: MoveRMbr operand: sreg operand: offset operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'movb ', srname, ', 0x', (offset hex allButFirst: 3), '(', drname, ')'.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testMoveRMwr
	"self new testMoveRMwr"
	self concreteCompilerClass registersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :drname| | inst len |
			#(0 64 65536 -64 -65536) do:
				[:offset|
				inst := self gen: MoveRMwr operand: sreg operand: offset operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended po |
						"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
						plainJane := self strip: str.
						po := offset bitAnd: 1 << self processor bitsInWord - 1.
						herIntended := 'movl ', srname, ', ', (offset = 0 ifTrue: [''] ifFalse: ['0x', (po printStringBase: 16)]), '(',drname,')'.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]

	"| failures |
	 failures := Dictionary new.
	 self concreteCompilerClass registersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :drname| | inst len |
			#(0 64 65536 -64 -65536) do:
				[:offset|
				inst := self gen: MoveRMwr operand: sreg operand: offset operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended po |
						plainJane := self strip: str.
						po := offset bitAnd: 1 << self processor bitsInWord - 1.
						herIntended := 'movl ', srname, ', ', (offset = 0 ifTrue: [''] ifFalse: ['0x', (po printStringBase: 16)]), '(',drname,')'.
						((plainJane match: herIntended) and: [len = sz]) ifFalse:
							[failures at: herIntended put: plainJane]]]]].
	 failures"! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 1/12/2016 07:49'!
testMoveRXbrR
	"self new testMoveRXbrR"
	self concreteCompilerClass registersWithNamesDo:
		[:idxreg :irname|
		irname ~= '%esp' ifTrue:
			[self concreteCompilerClass registersWithNamesDo:
				[:basereg :brname|
				self concreteCompilerClass byteRegistersWithNamesDo:
					[:sreg :srname| | inst len |
					inst := self gen: MoveRXbrR operand: sreg operand: idxreg operand: basereg.
					len := inst concretizeAt: 0.
					self processor
						disassembleInstructionAt: 0
						In: inst machineCode object
						into: [:str :sz| | plainJane herIntended |
							plainJane := self strip: str.
							herIntended := 'movb ', srname, ', (', brname, ',', irname, ',1)'.
							self assert: herIntended equals: plainJane.
							self assert: len = sz]]]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 1/12/2016 07:51'!
testMoveRXwrR
	"self new testMoveRXwrR"
	self concreteCompilerClass registersWithNamesDo:
		[:idxreg :irname|
		irname ~= '%esp' ifTrue:
			[self concreteCompilerClass registersWithNamesDo:
				[:basereg :brname|
				self concreteCompilerClass registersWithNamesDo:
					[:sreg :srname| | inst len |
					inst := self gen: MoveRXwrR operand: sreg operand: idxreg operand: basereg.
					len := inst concretizeAt: 0.
					self processor
						disassembleInstructionAt: 0
						In: inst machineCode object
						into: [:str :sz| | plainJane herIntended |
							plainJane := self strip: str.
							herIntended := 'movl ', srname, ', (', brname, ',', irname, ',4)'.
							self assert: herIntended equals: plainJane.
							self assert: len = sz]]]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testMoveRdM64r
	"self new testMoveRdM64r"
	CogIA32CompilerForTests xmmRegistersWithNamesDo:
		[:sreg :srname|
		CogIA32CompilerForTests registersWithNamesDo:
			[:dreg :drname|
			((1 to: 19 by: 3) collect: [:po2| 2 raisedToInteger: po2]) do:
				[:offset| | inst len |
				inst := self gen: MoveRdM64r operand: sreg operand: offset operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'movsd ', srname, ', 0x', (offset hex allButFirst: 3), '(', drname, ')'.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 1/12/2016 07:50'!
testMoveXbrRR
	"self new testMoveXbrRR"
	CogIA32CompilerForTests registersWithNamesDo:
		[:idxreg :irname|
		irname ~= '%esp' ifTrue:
			[CogIA32CompilerForTests registersWithNamesDo:
				[:basereg :brname|
				CogIA32CompilerForTests registersWithNamesDo:
					[:dreg :drname| | inst len |
					inst := self gen: MoveXbrRR operand: idxreg operand: basereg operand: dreg.
					len := inst concretizeAt: 0.
					self processor
						disassembleInstructionAt: 0
						In: inst machineCode object
						into: [:str :sz| | plainJane herIntended |
							"Convert e.g. '00000000: movzbl %ds:(%eax,%eax,1), %eax : 0F B6 04 00 ' to  'movzbl (%eax,%eax,1)'"
							plainJane := self strip: str.
							herIntended := 'movzbl (', brname, ',', irname, ',1), ',drname.
							self assert: herIntended equals: plainJane.
							self assert: len = sz]]]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 1/12/2016 07:50'!
testMoveXwrRR
	"self new testMoveXwrRR"
	CogIA32CompilerForTests registersWithNamesDo:
		[:idxreg :irname|
			irname ~= '%esp' ifTrue:
				[CogIA32CompilerForTests registersWithNamesDo:
					[:basereg :brname|
					CogIA32CompilerForTests registersWithNamesDo:
						[:dreg :drname| | inst len |
						inst := self gen: MoveXwrRR operand: idxreg operand: basereg operand: dreg.
						len := inst concretizeAt: 0.
						self processor
							disassembleInstructionAt: 0
							In: inst machineCode object
							into: [:str :sz| | plainJane herIntended |
								"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
								plainJane := self strip: str.
								herIntended := 'movl (', brname, ',', irname, ',4), ',drname.
								self assert: herIntended equals: plainJane.
								self assert: len = sz]]]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 9/12/2009 14:42'!
testMul
	"CogIA32CompilerTests new setUp testMul"
	| map compiler memory |
	map := Dictionary new.
	compiler := self gen: nil.
	memory := ByteArray new: 1024.
	CogIA32CompilerForTests dataRegistersWithAccessorsExcept: #(ebp esp) do:
		[:sreg :srget :srset|
		CogIA32CompilerForTests dataRegistersWithAccessorsExcept: { #ebp. #esp. srget } do:
			[:dreg :drget :drset| | instructions |
			self resetGen.
			compiler genMulR: sreg R: dreg.
			instructions := self generateInstructions.
			memory atAllPut: 0; replaceFrom: 1 to: instructions size with: instructions startingAt: 1.
			#(-768 -456 -123 123 456 789)
				with: #(987 654 321 -321 -654 -987)
				do: [:a :b|
					"self processor
						disassembleFrom: 0 to: instructions size in: memory on: Transcript;
						printIntegerRegistersOn: Transcript."
					map
						at: #eax put: (self processor eax: 16rA5A5A5A5);
						at: #ebx put: (self processor ebx: 16rB5B5B5B5);
						at: #ecx put: (self processor ecx: 16rC5C5C5C5);
						at: #edx put: (self processor edx: 16rD5D5D5D5);
						at: #esi put: (self processor esi: 16r51515151);
						at: #edi put: (self processor edi: 16rD1D1D1D1);
						at: srget put: (self processor perform: srset with: b signedIntToLong);
						at: drget put: (self processor perform: drset with: a signedIntToLong).
					self processor esp: memory size; eip: 0.
					self shouldnt:
						[[self processor pc < instructions size] whileTrue:
							[self processor singleStepIn: memory]]
						raise: Error.
					map at: drget put: (a * b) signedIntToLong.
					map keysAndValuesDo:
						[:accessor :value|
						self assert: value = (self processor perform: accessor)]]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testOrCqR
	"self new testOrCqR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname|
		#(16r1 16r3 16r7 16r555555 16rAAAAAA) do:
			[:const| | inst len|
			inst := self gen: OrCqR operand: const operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'orl $0x', (const printStringBase: 16 length: 8 padded: true), ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testSubCqR
	"self new testSubCqR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname|
		#(16r1 16r3 16r7 16r555555 16rAAAAAA) do:
			[:const| | inst len|
			inst := self gen: SubCqR operand: const operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'subl $0x', (const printStringBase: 16 length: 8 padded: true), ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogIA32CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 08:48'!
testTstCqR
	"self new testTstCqR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :theRegname|
		#(16r1 16r3 16r7 16r555555 16rAAAAAA) do:
			[:const| | op regname inst len constString|
			inst := self gen: TstCqR operand: const operand: reg.
			len := inst concretizeAt: 0.
			((inst isQuick: const) and: [reg < 4])
				ifTrue:
					[op := 'testb'.
					 regname := #('%al' '%cl' '%dl' '%bl') at: reg + 1.
					 constString := const printStringBase: 16 length: 2 padded: true]
				ifFalse:
					[op := 'testl'.
					 regname := theRegname.
					 constString := const printStringBase: 16 length: 8 padded: true].
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := op, ' $0x', constString, ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

AbstractInstructionTests subclass: #CogMIPSELCompilerTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!CogMIPSELCompilerTests methodsFor: 'accessing' stamp: 'rmacnak 11/24/2015 18:58:27'!
assertCheckLongArithOpCodeSize: bytes
	self assert: bytes = 12.! !

!CogMIPSELCompilerTests methodsFor: 'as yet unclassified' stamp: 'rmacnak 11/30/2015 19:44:46'!
assertCheckQuickArithOpCodeSize: bytes
	self assert: bytes <= 12! !

!CogMIPSELCompilerTests methodsFor: 'accessing' stamp: 'rmacnak 11/24/2015 18:13:18'!
concreteCompilerClass
	^CogMIPSELCompilerForTests! !

!CogMIPSELCompilerTests methodsFor: 'accessing' stamp: 'eem 12/19/2019 00:17'!
isaName
	^#MIPSEL! !

!CogMIPSELCompilerTests methodsFor: 'as yet unclassified' stamp: 'rmacnak 11/24/2015 18:24:08'!
memoryAsBytes: instructionMachineCode
	"Manage the fact that in the simulator inst machineCode object is an Array and the disassembler requires a ByteArray or some such."
	| bytes |
	instructionMachineCode isCObjectAccessor ifTrue:
		[^self memoryAsBytes: instructionMachineCode object].
	instructionMachineCode isArray ifFalse:
		[self assert: instructionMachineCode class isBits.
		 ^instructionMachineCode].
	bytes := ByteArray new: instructionMachineCode size * 4.
	1 to: instructionMachineCode size do:
		[:i|
		(instructionMachineCode at: i) ifNotNil:
			[:word|
			bytes unsignedLongAt: i - 1* 4 + 1 put: word]].
	^bytes! !

!CogMIPSELCompilerTests methodsFor: 'accessing' stamp: 'rmacnak 11/24/2015 18:11:54'!
processor
	processor ifNil:
		[processor := MIPSELSimulator new].
	^processor! !

AbstractInstructionTests subclass: #CogX64CompilerTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!CogX64CompilerTests methodsFor: 'abstract instructions' stamp: 'eem 9/22/2015 12:20'!
Label
	^self gen: Label operand: opcodes size! !

!CogX64CompilerTests methodsFor: 'running' stamp: 'eem 9/21/2015 15:16'!
assertCheckLongArithOpCodeSize: bytes
	self assert: bytes > 5! !

!CogX64CompilerTests methodsFor: 'running' stamp: 'eem 9/21/2015 15:16'!
assertCheckQuickArithOpCodeSize: bytes
	self assert: bytes <= 4! !

!CogX64CompilerTests methodsFor: 'accessing' stamp: 'eem 9/17/2015 14:51'!
concreteCompilerClass
	^CogX64CompilerForTests! !

!CogX64CompilerTests methodsFor: 'accessing' stamp: 'eem 12/18/2019 22:01'!
isaName
	^#X64! !

!CogX64CompilerTests methodsFor: 'accessing' stamp: 'eem 9/22/2015 12:21'!
methodLabel
	"There is no methodLabel in test code."
	^nil! !

!CogX64CompilerTests methodsFor: 'accessing' stamp: 'eem 9/17/2015 15:00'!
processor
	processor ifNil:
		[processor := BochsX64Alien new].
	^processor! !

!CogX64CompilerTests methodsFor: 'accessing' stamp: 'eem 9/17/2015 15:51'!
strip: aFancyDisassembledInstruction
	"Convert e.g. '00000000: movl %rax, 0x2(%rax) : 89 40 02' to  'movl %rax, 0x2(%rax)'"
	^((((aFancyDisassembledInstruction
		copyReplaceAll: '%ds:' with: '')
			copyReplaceAll: '%ss:' with: '')
				allButFirst: (aFancyDisassembledInstruction indexOf: $:) + 1)
					copyUpTo: $:)
						allButLast: 1! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testAddCqR
	"self new testAddCqR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname|
		#(16r1 16r3 16r7 16r555555 16rAAAAAA) do:
			[:const| | inst len|
			inst := self gen: AddCqR operand: const operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'addq $0x', (const printStringBase: 16 length: 16 padded: true), ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testAndCqR
	"self new testAndCqR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname|
		#(16r1 16r3 16r7 16r555555 16rAAAAAA) do:
			[:const| | inst len|
			inst := self gen: AndCqR operand: const operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'andq $0x', (const printStringBase: 16 length: 16 padded: true), ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/3/2015 18:24'!
testAndRR
	"CogX64CompilerTests new testAndRR"
	self concreteCompilerClass registersWithNamesDo:
		[:sreg :sregname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :dregname| | inst len |
			inst := self gen: AndRR operand: sreg operand: dreg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'andq ', sregname, ', ', dregname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 9/22/2015 12:07'!
testArithmeticShiftRightRR
	"CogX64CompilerTests new testArithmeticShiftRightRR"
	self concreteCompilerClass dataRegistersWithAccessorsDo:
		[:sreg :srgetter :srsetter|
		self concreteCompilerClass dataRegistersWithAccessorsDo:
			[:dreg :drgetter :drsetter| | inst len |
			inst := self gen: ArithmeticShiftRightRR operand: sreg operand: dreg.
			len := inst concretizeAt: 0.
			self assert: len = (srgetter = #rcx
								ifTrue: [3]
								ifFalse:
									[srgetter = #rax
										ifTrue: [7]
										ifFalse: [9]])
			"self processor disassembleFrom: 0 to: inst machineCodeSize in: inst machineCode object on: Transcript"]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testCMPXCHGMwrR
	"self new testCMPXCHGMwrR"
	| cmpxchgMwrR lock |
	cmpxchgMwrR := CogX64Compiler classPool at: #CMPXCHGMwrR.
	lock := CogX64Compiler classPool at: #LOCK.
	self concreteCompilerClass registersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :drname|
			((1 to: 19 by: 3) collect: [:po2| 2 raisedToInteger: po2]) do:
				[:offset|
				#(false true) do:
					[:prefixLock| | memory |
					self resetGen.
					prefixLock ifTrue: [self gen: lock].
					self gen: cmpxchgMwrR operand: offset operand: sreg operand: dreg.
					memory := self generateInstructions.
					self processor
						disassembleInstructionAt: 0
						In: memory
						into: [:str :sz| | plainJane herIntended |
							plainJane := self strip: str.
							herIntended := (prefixLock ifTrue: ['lock '] ifFalse: ['']),
											'cmpxchgl ', drname, ', 0x', (offset hex allButFirst: 3), '(', srname, ')'.
							self assert: herIntended equals: plainJane.
							self assert: memory size = sz]]]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 5/29/2017 09:23'!
testCallR
	"self new testCallR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname| | inst len|
			inst := self gen: CallR operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'call ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 1/15/2016 10:38'!
testCmpC32R
	"self new testCmpC32R"
	self concreteCompilerClass doubleWordRegistersWithNamesDo:
		[:reg :regname|
		#(16r1 16r3 16r7 16r555555 16rAAAAAA) do:
			[:const| | inst len|
			inst := self gen: CmpC32R operand: const operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'cmpl $0x', (const printStringBase: 16 length: 8 padded: true), ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testCmpCqR
	"self new testCmpCqR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname|
		#(16r1 16r3 16r7 16r555555 16rAAAAAA) do:
			[:const| | inst len|
			inst := self gen: CmpCqR operand: const operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'cmpq $0x', (const printStringBase: 16 length: 16 padded: true), ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 11/4/2015 17:46:45'!
testDivQuoRem
	"| failures ops |
	 failures := Set new.
	 [ops := (CogX64CompilerTests new testDivQuoRem)]
		on: TestResult failure
		do: [:ex| | c |
			c := ex signalerContext.
			[c tempNames includes: 'op'] whileFalse:
				[c := c sender].
			failures add: (c namedTempAt: (c tempNames indexOf: 'op')).
			ex resume].
	 { ops size. failures size. ops asSortedCollection asArray. failures asSortedCollection asArray}"
	| map compiler memory ops |
	map := Dictionary new.
	compiler := self gen: nil.
	memory := ByteArray new: 4096 * 2.
	ops := Set new.
	self concreteCompilerClass dataRegistersWithAccessorsExcept: #(rbp rsp) do:
		[:sreg :srget :srset|
		self concreteCompilerClass dataRegistersWithAccessorsExcept: { #rbp. #rsp. srget } do:
			[:dreg :drget :drset|
			 | instructions op |
			self concreteCompilerClass dataRegistersWithAccessorsExcept: #(rbp rsp) do:
				[:qreg :qrget :qrset| 
				self concreteCompilerClass dataRegistersWithAccessorsExcept: { #rbp. #rsp. qrget } do:
					[:rreg :rrget :rrset|
					self resetGen.
					op := qrget, ' := ', drget, ' quo: ', srget, '. ', rrget, ' := ', drget, ' rem: ', srget.
					ops add: op.
					compiler genDivR: sreg R: dreg Quo: qreg Rem: rreg.
					instructions := self generateInstructions.
					memory atAllPut: 0; replaceFrom: 1 to: instructions size with: instructions startingAt: 1; at: instructions size + 1 put: self processor nopOpcode.
					#(-768 -456 -123 123 456 789)
						with: #(987 654 321 -321 -654 -987)
						do: [:dd :dv| "| calc mask |
							mask := 16rFFFFFFFFFFFFFFFF.
							calc := ((dd quo: dv) bitAnd: mask) hex, ' := ', (dd bitAnd: mask) hex, ' quo: ', (dv bitAnd: mask) hex, '. ', ((dd rem: dv) bitAnd: mask) hex, ' := ', (dd bitAnd: mask) hex, ' rem: ', (dv bitAnd: mask) hex.
							calc := calc."
							"Transcript cr; cr; nextPutAll: op; cr; nextPutAll: calc; cr.
							 self processor
								disassembleFrom: 0 to: instructions size in: memory on: Transcript;
								printIntegerRegistersOn: Transcript."
							map
								at: #rax put: (self processor rax: 16rA5A5A5A5);
								at: #rbx put: (self processor rbx: 16rB5B5B5B5);
								at: #rcx put: (self processor rcx: 16rC5C5C5C5);
								at: #rdx put: (self processor rdx: 16rD5D5D5D5);
								at: #rsi put: (self processor rsi: 16r51515151);
								at: #rdi put: (self processor rdi: 16rD1D1D1D1);
								at: srget put: (self processor perform: srset with: (self processor convertIntegerToInternal: dv));
								at: drget put: (self processor perform: drset with: (self processor convertIntegerToInternal: dd)).
							self processor rsp: memory size; rip: 0.
							self shouldnt:
								[[self processor pc < instructions size] whileTrue:
									[self processor singleStepIn: memory readExecuteOnlyBelow: memory size / 2]]
								raise: Error.
							map
								at: qrget put: (self processor convertIntegerToInternal: (dd quo: dv));
								at: rrget put: (self processor convertIntegerToInternal: (dd rem: dv)).
							map keysAndValuesDo:
								[:accessor :value|
								self assert: value equals: (self processor perform: accessor)]]]]]].
	^ops! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 5/29/2017 09:24'!
testJumpR
	"self new testJumpR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname| | inst len|
			inst := self gen: JumpR operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'jmp ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testMoveAbR
	"self neb testMoveAbR"
	self concreteCompilerClass byteRegistersWithNamesDo:
		[:reg :regname| | reg64name |
		reg64name := self processor registerStateGetters at: reg + 1.
		#(16r555555 16rAAAAAA 16r5A5A5A5A5A5A) do:
			[:addr| | inst len totalsz |
			inst := self gen: MoveAbR operand: addr operand: reg.
			len := inst concretizeAt: 0.
			totalsz := 0.
			regname ~= '%al' ifTrue:
				[self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'xchgq %rax, %', reg64name.
						self assert: herIntended equals: plainJane.
						totalsz := sz]].
			self processor
				disassembleInstructionAt: totalsz
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					plainJane := self strip: str.
					herIntended := 'movb 0x', (addr printStringBase: 16 length: 16 padded: true), ', %al'.
					self assert: herIntended equals: plainJane.
					totalsz := totalsz + sz].
			regname ~= '%al' ifTrue:
				[self processor
					disassembleInstructionAt: totalsz
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'xchgq %rax, %', reg64name.
						self assert: herIntended equals: plainJane.
						totalsz := totalsz + sz]].
			self assert: len = totalsz]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testMoveAwR
	"self new testMoveAwR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname|
		#(16r555555 16rAAAAAA 16r5A5A5A5A5A5A) do:
			[:addr| | inst len totalsz |
			inst := self gen: MoveAwR operand: addr operand: reg.
			len := inst concretizeAt: 0.
			totalsz := 0.
			regname ~= '%rax' ifTrue:
				[self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'xchgq %rax, ', regname.
						self assert: herIntended equals: plainJane.
						totalsz := sz]].
			self processor
				disassembleInstructionAt: totalsz
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					plainJane := self strip: str.
					herIntended := 'movq 0x', (addr printStringBase: 16 length: 16 padded: true), ', %rax'.
					self assert: herIntended equals: plainJane.
					totalsz := totalsz + sz].
			regname ~= '%rax' ifTrue:
				[self processor
					disassembleInstructionAt: totalsz
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'xchgq %rax, ', regname.
						self assert: herIntended equals: plainJane.
						totalsz := totalsz + sz]].
			self assert: len = totalsz]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testMoveC32R
	"self new testMoveMwrR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :rname|
		#(0 64 65536 -64 -65536) do:
			[:offset| | inst len |
			inst := self gen: MoveC32R operand: offset operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended po |
					plainJane := self strip: str.
					po := offset bitAnd: 1 << self processor bitsInWord - 1.
					herIntended := 'movq $0x', (po printStringBase: 16 length: 16 padded: true), ', ', rname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 1/12/2016 15:17'!
testMoveM16rR
	"self new testMoveM16rR"
	self concreteCompilerClass registersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :drname|
			((1 to: 19 by: 3) collect: [:po2| 2 raisedToInteger: po2]) do:
				[:offset| | inst len |
				inst := self gen: MoveM16rR operand: offset operand: sreg operand: dreg.
				self deny: inst byteReadsZeroExtend.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended po |
						po := offset bitAnd: 1 << self processor bitsInWord - 1.
						plainJane := self strip: str.
						herIntended := 'movzwq ', (offset = 0 ifTrue: [''] ifFalse: ['0x', (po printStringBase: 16 length: 16 padded: true)]), '(', srname, '), ', drname.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/2/2015 11:34'!
testMoveM64rRd
	"self new testMoveM64rRd"
	self concreteCompilerClass registersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass xmmRegistersWithNamesDo:
			[:dreg :drname|
			#(0 8 32760) do:
				[:offset| | inst len |
				inst := self gen: MoveM64rRd operand: offset operand: sreg operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'movq ',
										(offset = 0
											ifTrue: ['']
											ifFalse: ['0x', ((offset printStringBase: 16 length: 16 padded: true) asLowercase)]),
										'(', srname, '), ', drname.
						"N.B. bochs 2.3.7 gets many cases wrong, e.g.
							movq %ds:(%rax), %xmm8 : F3 44 0F 7E 00
						 is disassembled as
							rep movq %ds:(%rax), %xmm8 : F3 44 0F 7E 00"
						(plainJane beginsWith: 'rep ') ifTrue:
							[plainJane := plainJane allButFirst: 4].
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]

"| them it |
them := OrderedCollection new.
[(it := CogX64CompilerTests new) testMoveM64rRd]
	on: AssertionFailure, TestResult failure
	do: [:ex| | inst |
		ex class == AssertionFailure
			ifTrue:
				[inst := ex signalerContext receiver.
				it processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| them addLast: (it strip: str)]]
			ifFalse:
				[ResumableTestFailure adoptInstance: ex].
		ex resume].
them size"
"| them it |
them := OrderedCollection new.
[(it := CogX64CompilerTests new) testMoveM64rRd]
	on: TestResult failure
	do: [:ex| | ctxt |
		ctxt := ex signalerContext findContextSuchThat: [:c| c selector == #assert:equals:]..
		them addLast: {ctxt tempAt: 1. ctxt tempAt: 2}.
		ResumableTestFailure adoptInstance: ex.
		ex resume].
them size"! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testMoveMbrR
	"self new testMoveMbrR"
	self concreteCompilerClass registersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass byteRegistersWithNamesDo:
			[:dreg :drname|
			((1 to: 19 by: 3) collect: [:po2| 2 raisedToInteger: po2]) do:
				[:offset| | inst len |
				inst := self gen: MoveMbrR operand: offset operand: sreg operand: dreg.
				self deny: inst byteReadsZeroExtend.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended po |
						po := offset bitAnd: 1 << self processor bitsInWord - 1.
						plainJane := self strip: str.
						herIntended := 'movb ', (offset = 0 ifTrue: [''] ifFalse: ['0x', (po printStringBase: 16 length: 16 padded: true)]), '(', srname, '), ', drname.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testMoveMwrR
	"self new testMoveMwrR"
	self concreteCompilerClass registersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :drname| | inst len |
			#(0 64 65536 -64 -65536) do:
				[:offset|
				inst := self gen: MoveMwrR operand: offset operand: sreg operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended po |
						plainJane := self strip: str.
						po := offset bitAnd: 1 << self processor bitsInWord - 1.
						herIntended := 'movq ', (offset = 0 ifTrue: [''] ifFalse: ['0x', (po printStringBase: 16 length: 16 padded: true)]), '(', srname, '), ', drname.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]

	"| failures |
	 failures := Dictionary new.
	 self concreteCompilerClass registersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :drname| | inst len |
			#(0 64 65536 -64 -65536) do:
				[:offset|
				inst := self gen: MoveMwrR operand: offset operand: sreg operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended po |
						plainJane := self strip: str.
						po := offset bitAnd: 1 << self processor bitsInWord - 1.
						herIntended := 'movq ', (offset = 0 ifTrue: [''] ifFalse: ['0x', (po printStringBase: 16 length: 16 padded: true)]), '(', srname, '), ', drname.
						((plainJane match: herIntended) and: [len = sz]) ifFalse:
							[failures at: herIntended put: plainJane]]]]].
	 failures"! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testMoveRAb
	"self new testMoveRAb"
	self concreteCompilerClass byteRegistersWithNamesDo:
		[:reg :regname| | reg64name |
		reg64name := self processor registerStateGetters at: reg + 1.
		#(16r555555 16rAAAAAA 16r5A5A5A5A5A5A) do:
			[:addr| | inst len totalsz |
			inst := self gen: MoveRAb operand: reg operand: addr.
			len := inst concretizeAt: 0.
			totalsz := 0.
			regname ~= '%al' ifTrue:
				[self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'xchgq %rax, %', reg64name.
						self assert: herIntended equals: plainJane.
						totalsz := sz]].
			self processor
				disassembleInstructionAt: totalsz
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					plainJane := self strip: str.
					herIntended := 'movb %al, 0x', (addr printStringBase: 16 length: 16 padded: true).
					self assert: herIntended equals: plainJane.
					totalsz := totalsz + sz].
			regname ~= '%al' ifTrue:
				[self processor
					disassembleInstructionAt: totalsz
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'xchgq %rax, %', reg64name.
						self assert: herIntended equals: plainJane.
						totalsz := totalsz + sz]].
			self assert: len = totalsz]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testMoveRAw
	"self new testMoveRAw"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname|
		#(16r555555 16rAAAAAA 16r5A5A5A5A5A5A) do:
			[:addr| | inst len totalsz |
			inst := self gen: MoveRAw operand: reg operand: addr.
			len := inst concretizeAt: 0.
			totalsz := 0.
			regname ~= '%rax' ifTrue:
				[self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'xchgq %rax, ', regname.
						self assert: herIntended equals: plainJane.
						totalsz := sz]].
			self processor
				disassembleInstructionAt: totalsz
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					plainJane := self strip: str.
					herIntended := 'movq %rax, 0x', (addr printStringBase: 16 length: 16 padded: true).
					self assert: herIntended equals: plainJane.
					totalsz := totalsz + sz].
			regname ~= '%rax' ifTrue:
				[self processor
					disassembleInstructionAt: totalsz
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'xchgq %rax, ', regname.
						self assert: herIntended equals: plainJane.
						totalsz := totalsz + sz]].
			self assert: len = totalsz]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 11/2/2015 16:06:42'!
testMoveRIPRelativeCwR
	"test rip-relative constant generation"
	"self new testMoveRIPRelativeCwR"
	
	| memory |
	memory := ByteArray new: 16.
	{self currentCompilationBase. self currentCompilationBase + 512} do:
		[:n| | inst len |
		self concreteCompilerClass dataRegistersWithAccessorsDo: 
			[ :r :rgetter :rset |
			inst := self gen: MoveCwR operand: n operand: r.
			len := inst concretizeAt: 0.
			memory replaceFrom: 1 to: len with: inst machineCode object startingAt: 1.
			self processor
				reset.
			[[processor pc < len] whileTrue:
				[self processor singleStepIn: memory]]
				on: Error
				do: [:ex| ].
			self concreteCompilerClass dataRegistersWithAccessorsDo:
				[:ireg :getter :setter| | expected |
				expected := getter == rgetter ifTrue: [n] ifFalse: [0].
				self assert: (self processor perform: getter) = expected].
			self assert: self processor pc = inst machineCodeSize]]

	"processor disassembleFrom: 0 to: inst machineCodeSize in: memory on: Transcript"! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 1/12/2016 16:26'!
testMoveRM16r
	"self new testMoveRM16r"
	self concreteCompilerClass wordRegistersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :drname|
			((1 to: 19 by: 3) collect: [:po2| 2 raisedToInteger: po2]) do:
				[:offset| | inst len |
				inst := self gen: MoveRM16r operand: sreg operand: offset operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended po |
						plainJane := self strip: str.
						po := offset bitAnd: 1 << self processor bitsInWord - 1.
						herIntended := 'movw ', srname, (offset = 0 ifTrue: [', '] ifFalse: [', 0x', (po printStringBase: 16 length: 16 padded: true)]), '(', drname, ')'.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testMoveRMbr
	"self new testMoveRMbr"
	self concreteCompilerClass byteRegistersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :drname|
			((1 to: 19 by: 3) collect: [:po2| 2 raisedToInteger: po2]) do:
				[:offset| | inst len |
				inst := self gen: MoveRMbr operand: sreg operand: offset operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended po |
						plainJane := self strip: str.
						po := offset bitAnd: 1 << self processor bitsInWord - 1.
						herIntended := 'movb ', srname, (offset = 0 ifTrue: [', '] ifFalse: [', 0x', (po printStringBase: 16 length: 16 padded: true)]), '(', drname, ')'.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testMoveRMwr
	"self new testMoveRMwr"
	self concreteCompilerClass registersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :drname| | inst len |
			#(0 64 65536 -64 -65536) do:
				[:offset|
				inst := self gen: MoveRMwr operand: sreg operand: offset operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended po |
						"Convert e.g. '00000000: movq %rax, 0x2(%rax) : 48 89 40 02' to  'movl %rax, 0x2(%rax)'"
						plainJane := self strip: str.
						po := offset bitAnd: 1 << self processor bitsInWord - 1.
						herIntended := 'movq ', srname, ', ', (offset = 0 ifTrue: [''] ifFalse: ['0x', (po printStringBase: 16 length: 16 padded: true)]), '(',drname,')'.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]

	"| failures |
	 failures := Dictionary new.
	 self concreteCompilerClass registersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :drname| | inst len |
			#(0 64 65536 -64 -65536) do:
				[:offset|
				inst := self gen: MoveRMwr operand: sreg operand: offset operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended po |
						plainJane := self strip: str.
						po := offset bitAnd: 1 << self processor bitsInWord - 1.
						herIntended := 'movq ', srname, ', ', (offset = 0 ifTrue: [''] ifFalse: ['0x', (po printStringBase: 16 length: 16 padded: true)]), '(',drname,')'.
						((plainJane match: herIntended) and: [len = sz]) ifFalse:
							[failures at: herIntended put: plainJane]]]]].
	 failures"! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/9/2015 14:34'!
testMoveRRd
	"self new testMoveRRd"
	self concreteCompilerClass registersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass xmmRegistersWithNamesDo:
			[:dreg :drname| | inst len |
			inst := self gen: MoveRRd operand: sreg operand: dreg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					plainJane := self strip: str.
					herIntended := 'movq ', srname, ', ', drname.
					"N.B. bochs 2.3.7 gets many cases wrong, e.g.
						movq %ds:(%rax), %xmm8 : F3 44 0F 7E 00
					 is disassembled as
						rep movq %ds:(%rax), %xmm8 : F3 44 0F 7E 00"
					(plainJane beginsWith: 'rep ') ifTrue:
						[plainJane := plainJane allButFirst: 4].
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]

"| them it |
them := OrderedCollection new.
[(it := CogX64CompilerTests new) testMoveRRd]
	on: AssertionFailure, TestResult failure
	do: [:ex| | inst |
		ex class == AssertionFailure
			ifTrue:
				[inst := ex signalerContext receiver.
				it processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| them addLast: (it strip: str)]]
			ifFalse:
				[ResumableTestFailure adoptInstance: ex].
		ex resume].
them size"
"| them it |
them := OrderedCollection new.
[(it := CogX64CompilerTests new) testMoveRRd]
	on: TestResult failure
	do: [:ex| | ctxt |
		ctxt := ex signalerContext findContextSuchThat: [:c| c selector == #assert:equals:]..
		them addLast: {ctxt tempAt: 1. ctxt tempAt: 2}.
		ResumableTestFailure adoptInstance: ex.
		ex resume].
them size"! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 1/12/2016 07:48'!
testMoveRX32rR
	"self new testMoveRX32rR"
	self concreteCompilerClass registersWithNamesDo:
		[:idxreg :irname|
			irname ~= '%rsp' ifTrue:
				[self concreteCompilerClass registersWithNamesDo:
					[:basereg :brname|
					self concreteCompilerClass doubleWordRegistersWithNamesDo:
						[:sreg :srname| | inst len |
						inst := self gen: MoveRX32rR operand: sreg operand: idxreg operand: basereg.
						len := inst concretizeAt: 0.
						self processor
							disassembleInstructionAt: 0
							In: inst machineCode object
							into: [:str :sz| | plainJane herIntended |
								"Convert e.g. '00000000: movl %rax, 0x2(%rax) : 89 40 02' to  'movl %rax, 0x2(%rax)'"
								plainJane := self strip: str.
								herIntended := 'movl ', srname, ', (', brname, ',', irname, ',4)'.
								self assert: herIntended equals: plainJane.
								self assert: len = sz]]]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 1/12/2016 07:48'!
testMoveRXbrR
	"self new testMoveXbrRR"
	self concreteCompilerClass registersWithNamesDo:
		[:idxreg :irname|
		irname ~= '%rsp' ifTrue:
			[self concreteCompilerClass registersWithNamesDo:
				[:basereg :brname|
				self concreteCompilerClass byteRegistersWithNamesDo:
					[:sreg :srname| | inst len |
					inst := self gen: MoveRXbrR operand: sreg operand: idxreg operand: basereg.
					len := inst concretizeAt: 0.
					self processor
						disassembleInstructionAt: 0
						In: inst machineCode object
						into: [:str :sz| | plainJane herIntended |
							plainJane := self strip: str.
							herIntended := 'movb ', srname, ', (', brname, ',', irname, ',1)'.
							self assert: herIntended equals: plainJane.
							self assert: len = sz]]]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 1/12/2016 07:48'!
testMoveRXwrR
	"self new testMoveXwrRR"
	self concreteCompilerClass registersWithNamesDo:
		[:idxreg :irname|
			irname ~= '%rsp' ifTrue:
				[self concreteCompilerClass registersWithNamesDo:
					[:basereg :brname|
					self concreteCompilerClass registersWithNamesDo:
						[:sreg :srname| | inst len |
						inst := self gen: MoveRXwrR operand: sreg operand: idxreg operand: basereg.
						len := inst concretizeAt: 0.
						self processor
							disassembleInstructionAt: 0
							In: inst machineCode object
							into: [:str :sz| | plainJane herIntended |
								"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
								plainJane := self strip: str.
								herIntended := 'movq ', srname, ', (', brname, ',', irname, ',8)'.
								self assert: herIntended equals: plainJane.
								self assert: len = sz]]]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/6/2015 22:15'!
testMoveRdM64r
	"self new testMoveRdM64r"
	self concreteCompilerClass xmmRegistersWithNamesDo:
		[:sreg :srname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :drname|
			#(0 8 256 32760) do:
				[:offset| | inst len |
				inst := self gen: MoveRdM64r operand: sreg operand: offset operand: dreg.
				len := inst concretizeAt: 0.
				self processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| | plainJane herIntended |
						plainJane := self strip: str.
						herIntended := 'movq ', srname,
											(offset = 0
												ifTrue: [', ']
												ifFalse: [', 0x', ((offset printStringBase: 16 length: 16 padded: true))]),
											'(', drname, ')'.
						self assert: herIntended equals: plainJane.
						self assert: len = sz]]]]

"| them it |
them := OrderedCollection new.
[(it := CogX64CompilerTests new) testMoveRdM64r]
	on: AssertionFailure, TestResult failure
	do: [:ex| | inst |
		ex class == AssertionFailure
			ifTrue:
				[inst := ex signalerContext receiver.
				it processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| them addLast: (it strip: str)]]
			ifFalse:
				[ResumableTestFailure adoptInstance: ex].
		ex resume].
them size"
"| them it |
them := OrderedCollection new.
[(it := CogX64CompilerTests new) testMoveRdM64r]
	on: TestResult failure
	do: [:ex| | ctxt |
		ctxt := ex signalerContext findContextSuchThat: [:c| c selector == #assert:equals:]..
		them addLast: {ctxt tempAt: 1. ctxt tempAt: 2}.
		ResumableTestFailure adoptInstance: ex.
		ex resume].
them size"! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/9/2015 14:38'!
testMoveRdR
	"self new testMoveRdR"
	self concreteCompilerClass registersWithNamesDo:
		[:dreg :drname|
		self concreteCompilerClass xmmRegistersWithNamesDo:
			[:sreg :srname| | inst len |
			inst := self gen: MoveRdR operand: sreg operand: dreg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					plainJane := self strip: str.
					herIntended := 'movq ', srname, ', ', drname.
					"N.B. bochs 2.3.7 gets many cases wrong, e.g.
						movq %ds:(%rax), %xmm8 : F3 44 0F 7E 00
					 is disassembled as
						rep movq %ds:(%rax), %xmm8 : F3 44 0F 7E 00"
					(plainJane beginsWith: 'rep ') ifTrue:
						[plainJane := plainJane allButFirst: 4].
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]

"| them it |
them := OrderedCollection new.
[(it := CogX64CompilerTests new) testMoveRdR]
	on: AssertionFailure, TestResult failure
	do: [:ex| | inst |
		ex class == AssertionFailure
			ifTrue:
				[inst := ex signalerContext receiver.
				it processor
					disassembleInstructionAt: 0
					In: inst machineCode object
					into: [:str :sz| them addLast: (it strip: str)]]
			ifFalse:
				[ResumableTestFailure adoptInstance: ex].
		ex resume].
them size"
"| them it |
them := OrderedCollection new.
[(it := CogX64CompilerTests new) testMoveRdR]
	on: TestResult failure
	do: [:ex| | ctxt |
		ctxt := ex signalerContext findContextSuchThat: [:c| c selector == #assert:equals:]..
		them addLast: {ctxt tempAt: 1. ctxt tempAt: 2}.
		ResumableTestFailure adoptInstance: ex.
		ex resume].
them size"! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 1/12/2016 07:48'!
testMoveX32rRR
	"self new testMoveX32rRR"
	self concreteCompilerClass registersWithNamesDo:
		[:idxreg :irname|
			irname ~= '%rsp' ifTrue:
				[self concreteCompilerClass registersWithNamesDo:
					[:basereg :brname|
					self concreteCompilerClass doubleWordRegistersWithNamesDo:
						[:dreg :drname| | inst len |
						inst := self gen: MoveX32rRR operand: idxreg operand: basereg operand: dreg.
						len := inst concretizeAt: 0.
						self processor
							disassembleInstructionAt: 3
							In: inst machineCode object
							into: [:str :sz| | plainJane herIntended |
								"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
								plainJane := self strip: str.
								herIntended := 'movl (', brname, ',', irname, ',4), ',drname.
								self assert: herIntended equals: plainJane.
								self assert: len - 3 = sz]]]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 1/12/2016 07:49'!
testMoveXbrRR
	"self new testMoveXbrRR"
	self concreteCompilerClass registersWithNamesDo:
		[:idxreg :irname|
		irname ~= '%rsp' ifTrue:
			[self concreteCompilerClass registersWithNamesDo:
				[:basereg :brname|
				self concreteCompilerClass byteRegistersWithNamesDo:
					[:dreg :drname| | inst len |
					inst := self gen: MoveXbrRR operand: idxreg operand: basereg operand: dreg.
					self deny: inst byteReadsZeroExtend.
					len := inst concretizeAt: 0.
					self processor
						disassembleInstructionAt: 0
						In: inst machineCode object
						into: [:str :sz| | plainJane herIntended |
							"Convert e.g. '00000000: movzbq %ds:(%rax,%rax,1), %rax : 48 0F B6 04 00 ' to  'movzbq (%rax,%rax,1), %rax'"
							plainJane := self strip: str.
							herIntended := 'movb (', brname, ',', irname, ',1), ',drname.
							self assert: herIntended equals: plainJane.
							self assert: len = sz]]]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 1/12/2016 07:49'!
testMoveXwrRR
	"self new testMoveXwrRR"
	self concreteCompilerClass registersWithNamesDo:
		[:idxreg :irname|
			irname ~= '%rsp' ifTrue:
				[self concreteCompilerClass registersWithNamesDo:
					[:basereg :brname|
					self concreteCompilerClass registersWithNamesDo:
						[:dreg :drname| | inst len |
						inst := self gen: MoveXwrRR operand: idxreg operand: basereg operand: dreg.
						len := inst concretizeAt: 0.
						self processor
							disassembleInstructionAt: 0
							In: inst machineCode object
							into: [:str :sz| | plainJane herIntended |
								"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
								plainJane := self strip: str.
								herIntended := 'movq (', brname, ',', irname, ',8), ',drname.
								self assert: herIntended equals: plainJane.
								self assert: len = sz]]]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/3/2015 12:28'!
testMul
	"CogX64CompilerTests new setUp testMul"
	| map compiler memory |
	map := Dictionary new.
	compiler := self gen: nil.
	memory := ByteArray new: 1024.
	self concreteCompilerClass dataRegistersWithAccessorsExcept: #(rbp rsp) do:
		[:sreg :srget :srset|
		self concreteCompilerClass dataRegistersWithAccessorsExcept: { #rbp. #rsp. srget } do:
			[:dreg :drget :drset| | instructions |
			self resetGen.
			compiler genMulR: sreg R: dreg.
			instructions := self generateInstructions.
			memory atAllPut: 0; replaceFrom: 1 to: instructions size with: instructions startingAt: 1.
			#(-768 -456 -123 123 456 789)
				with: #(987 654 321 -321 -654 -987)
				do: [:a :b|
					"self processor
						disassembleFrom: 0 to: instructions size in: memory on: Transcript;
						printIntegerRegistersOn: Transcript."
					map
						at: #rax put: (self processor rax: 16rA5A5A5A5);
						at: #rbx put: (self processor rbx: 16rB5B5B5B5);
						at: #rcx put: (self processor rcx: 16rC5C5C5C5);
						at: #rdx put: (self processor rdx: 16rD5D5D5D5);
						at: #rsi put: (self processor rsi: 16r51515151);
						at: #rdi put: (self processor rdi: 16rD1D1D1D1);
								at: srget put: (self processor perform: srset with: (self processor convertIntegerToInternal: b));
								at: drget put: (self processor perform: drset with: (self processor convertIntegerToInternal: a)).
					self processor rsp: memory size; rip: 0.
					self shouldnt:
						[[self processor pc < instructions size] whileTrue:
							[self processor singleStepIn: memory]]
						raise: Error.
					map at: drget put: (self processor convertIntegerToInternal: (a * b)).
					map keysAndValuesDo:
						[:accessor :value|
						self assert: value = (self processor perform: accessor)]]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testOrCqR
	"self new testOrCqR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname|
		#(16r1 16r3 16r7 16r555555 16rAAAAAA) do:
			[:const| | inst len|
			inst := self gen: OrCqR operand: const operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'orq $0x', (const printStringBase: 16 length: 16 padded: true), ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/3/2015 11:36'!
testShiftCqR
	"CogX64CompilerTests new testShiftCqR"
	{LogicalShiftRightCqR. LogicalShiftLeftCqR. ArithmeticShiftRightCqR}
		with: #('shrq' 'shlq' 'sarq')
		do: [:op :opname|
			self concreteCompilerClass registersWithNamesDo:
				[:reg :regname|
				1 to: 63 do:
					[:shift| | inst len |
					inst := self gen: op operand: shift operand: reg.
					len := inst concretizeAt: 0.
					self processor
						disassembleInstructionAt: 0
						In: inst machineCode object
						into: [:str :sz| | plainJane herIntended |
							"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
							plainJane := self strip: str.
							herIntended := opname ,' $', (shift > 1
														ifTrue: ['0x', (shift printStringBase: 16 length: 2 padded: true)]
														ifFalse: [shift asInteger]), ', ', regname.
							self assert: herIntended equals: plainJane.
							self assert: len = sz]]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testSubCqR
	"self new testSubCqR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :regname|
		#(16r1 16r3 16r7 16r555555 16rAAAAAA) do:
			[:const| | inst len|
			inst := self gen: SubCqR operand: const operand: reg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := 'subq $0x', (const printStringBase: 16 length: 16 padded: true), ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 12/03/2015 18:26'!
testTstCqR
	"self new testTstCqR"
	self concreteCompilerClass registersWithNamesDo:
		[:reg :theRegname|
		#(16r1 16r3 16r7 16r555555 16rAAAAAA) do:
			[:const| | op regname inst len constString|
			inst := self gen: TstCqR operand: const operand: reg.
			len := inst concretizeAt: 0.
			(inst isQuick: const)
				ifTrue:
					[op := 'testb'.
					 regname := reg < 8
									ifTrue: [#('%al' '%cl' '%dl' '%bl' '%spl' '%bpl' '%sil' '%dil') at: reg + 1]
									ifFalse: [theRegname, 'b'].
					 constString := const printStringBase: 16 length: 2 padded: true]
				ifFalse:
					[op := 'testq'.
					 regname := theRegname.
					 constString := const printStringBase: 16 length: 16 padded: true].
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := op, ' $0x', constString, ', ', regname.
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogX64CompilerTests methodsFor: 'tests' stamp: 'eem 2/15/2017 17:29'!
testXCHGRR
	"CogX64CompilerTests new testXCHGRR"
	| xchgrr |
	xchgrr := (self concreteCompilerClass bindingOf: #XCHGRR) value.
	self concreteCompilerClass registersWithNamesDo:
		[:sreg :sregname|
		self concreteCompilerClass registersWithNamesDo:
			[:dreg :dregname| | inst len |
			inst := self gen: xchgrr operand: sreg operand: dreg.
			len := inst concretizeAt: 0.
			self processor
				disassembleInstructionAt: 0
				In: inst machineCode object
				into: [:str :sz| | plainJane herIntended |
					"Convert e.g. '00000000: movl %eax, 0x2(%eax) : 89 40 02' to  'movl %eax, 0x2(%eax)'"
					plainJane := self strip: str.
					herIntended := dreg = 0
										ifTrue: [sreg = 0 ifTrue: ['nop '] ifFalse: ['xchgq ', dregname, ', ', sregname]]
										ifFalse: ['xchgq ', sregname, ', ', dregname].
					self assert: herIntended equals: plainJane.
					self assert: len = sz]]]! !

!CogRTLOpcodes class methodsFor: 'class initialization' stamp: 'eem 8/5/2021 07:40'!
initialize
	"Abstract opcodes are a compound of a one word operation specifier and zero or more operand type specifiers.
	 The assembler is in Cogit protocol abstract instructions and uses `at&t' syntax, assigning to the register on the
	 right. e.g. MoveRR is the Move opcode with two register operand specifiers and defines a move register to
	 register instruction from operand 0 to operand 1.  The word and register size is assumed to be either 32-bits
	 on a 32-bit architecture or 64-bits on a 64-bit architecture.  The abstract machine is mostly a 2 address machine
	 with the odd three address instruction added to better exploit RISCs.
			(self initialize)
	The operand specifiers are
		R		- general purpose register
		Rs		- single-precision floating-point register
		Rd		- double-precision floating-point register
		Cq		- a `quick' constant that can be encoded in the minimum space possible.
		Cw		- a constant with word size where word is the default operand size for the Smalltalk VM, 32-bits
				  for a 32-bit VM, 64-bits for a 64-bit VM.  The generated constant must occupy the default number
				  of bits.  This allows e.g. a garbage collector to update the value without invalidating the code.
		C32	- a constant with 32 bit size.  The generated constant must occupy 32 bits.
		C64	- a constant with 64 bit size.  The generated constant must occupy 64 bits.
		Aw		- memory word (32-bits for a 32-bit VM, 64-bits for a 64-bit VM) at an absolute address
				  See note about MoveAwR and MoveRAw in the opcodeNames literal array below!!!!
		Ab		- memory byte at an absolute address
		A32	- memory 32-bit halfword at an absolute address
		L		- a register mask supplying the live registers to be preserved (which need not include SPReg, FPReg, NativeSPReg, LinkReg, & PCReg)
		Mwr	- memory word whose address is at a constant offset from an address in a register
		Mbr		- memory byte whose address is at a constant offset from an address in a register (zero-extended on read)
		M16r	- memory 16-bit halfword whose address is at a constant offset from an address in a register
		M32r	- memory 32-bit halfword whose address is at a constant offset from an address in a register
		M64r	- memory 64-bit doubleword whose address is at a constant offset from an address in a register
		Xbr		- memory byte whose address is r * byte size away from an address in a register
		X16r	- memory 16-bit halfword whose address is r * (2 bytes size) away from an address in a register
		X32r	- memory 32-bit halfword whose address is r * (4 bytes size) away from an address in a register (64-bit ISAs only)
		Xwr		- memory word whose address is r * word size away from an address in a register
		Xowr	- memory word whose address is o + (r * word size) away from an address in a register (scaled indexed)

	An alternative would be to decouple opcodes from operands, e.g.
		Move := 1. Add := 2. Sub := 3...
		RegisterOperand := 1. ConstantQuickOperand := 2. ConstantWordOperand := 3...
	But not all combinations make sense and even fewer are used so we stick with the simple compound approach.

	The assumption is that comparison and arithmetic instructions set condition codes (*) and that move instructions
	leave the condition codes unaffected.  In particular LoadEffectiveAddressMwrR does not set condition codes
	although it can be used to do arithmetic.  On processors such as MIPS this distinction is invalid; there are no
	condition codes.  So the backend is allowed to collapse operation, branch pairs to internal instruction definitions
	(see senders and implementors of noteFollowingConditionalBranch:).
	We also assume that arithmetic sets the carry flag but that Add?R/Sub?R et al ignore the carry flag as a source operand.
	This is because multiple precision arithmetic is implem ented in plugins (LargeIntegerPlugins) and hence not (yet)
	important here.
	(*) the back end is not expected to set the condition codes if arithmetic is done to change the value of the stack pointer.

	Not all of the definitions in opcodeDefinitions below are implemented.  In particular we do not implement the
	XowrR scaled index addressing mode since it requires 4 operands.

	Not all instructions make sense on all architectures.  MoveRRd and MoveRdR are meaningful only on 64-bit machines.

	Note that there are no generic division instructions defined, but a processor may define some.

	Branch/Call ranges.  Jump[Cond] can be generated as short as possible.  Call/Jump[Cond]Long must be generated
	in the same number of bytes irrespective of displacement since their targets may be updated, but they need only
	span 16Mb, the maximum size of the code zone.  This allows e.g. ARM to use single-word call and jump instructions
	for most calls and jumps.  CallFull/JumpFull must also be generated in the same number of bytes irrespective of
	displacement for the same reason, but they must be able to span the full (32-bit or 64-bit) address space because
	they are used to call code in the C runtime, which may be distant from the code zone.  CallFull/JumpFull are allowed
	to use the cResultRegister as a scratch if required (e.g. on x64 where there is no direct 64-bit call or jump).

	Byte reads.  If the concrete compiler class answers true to byteReadsZeroExtend then byte reads must zero-extend
	the byte read into the destination register.  If not, the other bits of the register should be left undisturbed and the
	Cogit will add an instruction to zero the register as required.  Under no circumstances should byte reads sign-extend.

	16-bit (and on 64-bits, 32-bit) reads.  These /are/ expected to always zero-extend.

	Arithmetic operations have as their destination the last operand except Cmp; CmpXR is SubRX with no update of result.
	We only require that CmpR: reg R: SPReg be supported, and do not use CmpR: SPReg R: reg.
	This is in support of ARMv8, which only supports one direction of comparison with the stack pointer.

	By default we use the native stack pointer and in this form PopR PushR and RetN are correspondingly equivalent
	to NativePopR NativePushR and NativeRetN, and SPReg = NativeSPReg.  But if the backEnd answers false to
	usesNativeSPReg then NativePopR, NativePushR, and NativeRetN operate on the NativeSPReg, PopR PushR and
	RetN operate on SPReg, and NativeSPReg ~= SPReg."

	| opcodeNames refs |
	opcodeNames := #("Noops & Pseudo Ops"
						Label
						Literal			"a word-sized literal"
						AlignmentNops
						Fill32			"output four byte's worth of bytes with operand 0"
						Nop

						"Control"
						Call					"call within the code zone"
						CallFull				"call anywhere within the full address space"
						CallR
						RetN
						JumpR				"Not a regular jump, i.e. not pc dependent."
						Stop				"Halt the processor"

						"N.B.  Jumps are contiguous.  Long and Full jumps are contiguous within them.  See FirstJump et al below"
						JumpFull				"Jump anywhere within the address space"
						JumpLong				"Jump anywhere within the 16mb code zone."
						JumpLongZero			"a.k.a. JumpLongEqual"
						JumpLongNonZero		"a.k.a. JumpLongNotEqual"
						Jump					"short jumps; can be encoded in as few bytes as possible; will not be disturbed by GC or relocation."
						JumpZero				"a.k.a. JumpEqual"
						JumpNonZero			"a.k.a. JumpNotEqual"
						JumpNegative
						JumpNonNegative
						JumpOverflow
						JumpNoOverflow
						JumpMulOverflow 		"Only on processors that lack integer multiply overflow detection, e.g. ARMv8"
						JumpNoMulOverflow 	"Defined here so that CogAbstractInstruction>>isJump is a single range comparison."
						JumpCarry
						JumpNoCarry
						JumpLess				"signed"
						JumpGreaterOrEqual
						JumpGreater
						JumpLessOrEqual
						JumpBelow				"unsigned"
						JumpAboveOrEqual
						JumpAbove
						JumpBelowOrEqual

						JumpFPEqual
						JumpFPNotEqual
						JumpFPLess
						JumpFPLessOrEqual
						JumpFPGreater
						JumpFPGreaterOrEqual
						JumpFPOrdered
						JumpFPUnordered

						"Data Movement; destination is always last operand"
						MoveRR
						"N.B. On certain targets (including X64) MoveAwR & MoveRAw may
						 smash TempReg if the register argument is either FPReg or SPReg!!!!"
						MoveAwR MoveA32R
						MoveRAw MoveRA32
						MoveAbR
						MoveRAb
						MoveMwrR MoveRMwr MoveXwrRR MoveRXwrR "MoveXowrR MoveRXowr""Unused"
						MoveM8rR MoveMs8rR MoveRM8r 
						MoveM16rR MoveRM16r MoveX16rRR MoveRX16rR
						MoveM32rR MoveRM32r MoveX32rRR MoveRX32rR
						MoveMbrR MoveRMbr MoveXbrRR MoveRXbrR
						MoveCqR MoveCwR MoveC32R "MoveC64R""Not used"
						MoveRRd MoveRdR MoveRdRd MoveM64rRd MoveRdM64r
						MoveRsRs MoveM32rRs MoveRsM32r
						PopR PushR PushCq PushCw
						NativePopR NativePushR NativeRetN
						PrefetchAw

						"Arithmetic; destination is always last operand except Cmp; CmpXR is SubRX with no update of result"
						"We only require that CmpR: reg R: SPReg be supported, and do not use CmpR: SPReg R: reg.
						 This is in support of ARMv8, which only supports one direction of comparison with the stack pointer."
						LoadEffectiveAddressMwrR "LoadEffectiveAddressXowrR" "Variants of add/multiply"
						NegateR "2's complement negation"
						NotR
						ArithmeticShiftRightCqR ArithmeticShiftRightRR
						LogicalShiftRightCqR LogicalShiftRightRR
						LogicalShiftLeftCqR LogicalShiftLeftRR
						RotateLeftCqR RotateRightCqR

						CmpRR AddRR SubRR AndRR OrRR XorRR
						CmpCqR AddCqR SubCqR AndCqR OrCqR TstCqR XorCqR
						CmpCwR CmpC32R AddCwR SubCwR AndCwR OrCwR XorCwR
						AddcRR AddcCqR SubbRR SubbCqR

						"Three address ops for RISCs; feel free to add and extend"
						AddCqRR AndCqRR OrCqRR AddRRR SubRRR ArithmeticShiftRightCqRR LogicalShiftLeftCqRR LogicalShiftRightCqRR

						CmpRdRd AddRdRd SubRdRd MulRdRd DivRdRd SqrtRd XorRdRd
						CmpRsRs AddRsRs SubRsRs MulRsRs DivRsRs SqrtRs XorRsRs

						"Conversion"
						ConvertRRd ConvertRdR
						ConvertRsRd ConvertRdRs ConvertRsR ConvertRRs

						SignExtend8RR SignExtend16RR SignExtend32RR
						ZeroExtend8RR ZeroExtend16RR ZeroExtend32RR

						"Advanced bit manipulation (aritmetic)"
						ClzRR

						"Access to a 64-bit performance counter on 32 and 64 bit ISAs"
						MovePerfCnt64RRL
						MovePerfCnt64RL

						LastRTLCode).

	"Magic auto declaration. Add to the classPool any new variables and nuke any obsolete ones, and assign values"
	"Find the variables directly referenced by this method"
	refs := (thisContext method literals select: [:l| l isVariableBinding and: [classPool includesKey: l key]]) collect:
				[:ea| ea key].
	"Move to Undeclared any opcodes in classPool not in opcodes or this method."
	(classPool keys reject: [:k| (opcodeNames includes: k) or: [refs includes: k]]) do:
		[:k|
		Undeclared declare: k from: classPool].
	"Declare as class variables and number elements of opcodeArray above"
	opcodeNames withIndexDo:
		[:classVarName :value|
		self classPool
			declare: classVarName from: Undeclared;
			at: classVarName put: value].

	"For CogAbstractInstruction>>isJump etc..."
	FirstJump := JumpFull.
	LastJump := JumpFPUnordered.
	FirstShortJump := Jump.

	"And now initialize the backends; they add their own opcodes and hence these must be reinitialized."
	(Smalltalk classNamed: #CogAbstractInstruction) ifNotNil:
		[:cogAbstractInstruction| cogAbstractInstruction allSubclasses do: [:sc| sc initialize]]! !

!CogRTLOpcodes class methodsFor: 'debug printing' stamp: 'eem 12/26/2015 20:28'!
nameForOpcode: opcode "<Integer>"
	opcode < LastRTLCode ifTrue:
		[classPool associations do:
			[:assoc| | opcodeName |
			 assoc value = opcode ifTrue:
				[(((opcodeName := assoc key) beginsWith: 'First')
				   or: [(opcodeName beginsWith: 'Last')]) ifFalse:
					[^opcodeName]]]].
	^nil! !

!CogRTLOpcodes class methodsFor: 'accessing' stamp: 'eem 1/24/2012 12:55'!
opcodes
	"CogRTLOpcodes opcodes"
	"CogRTLOpcodes opcodes select:
		[:opcode|
		(Cogit organization listAtCategoryNamed: #'abstract instructions') noneSatisfy:
			[:s|
			(Cogit compiledMethodAt: s) literals anySatisfy:
				[:l|
				l isVariableBinding and: [l key = opcode]]]]"
	^(classPool keys reject:
		[:k|
		(#('First' 'Last' 'DPFP') anySatisfy: [:prefix| k beginsWith: prefix])
		or: [#('Reg' 'Min' 'Max') anySatisfy: [:postfix| k endsWith: postfix]]]) asArray sort! !

!CogRTLOpcodes class methodsFor: 'debug printing' stamp: 'eem 12/13/2019 18:31'!
printFormatForOpcodeName: opcodeName
	"Answer a sequence of $r, $f or nil for the operands in the opcode, used for printing, where
	 r => integer register, f => floating point register, and nil => numeric or address operand."
	| printFormat operands |
	printFormat := OrderedCollection new.
	operands := opcodeName.
	[(#(('Ab' nil) ('Aw' nil)
		('Cw' nil) ('Cq' nil) ('C32' nil)
		(Fill32 nil)
		(Label nil) (Literal nil)
		('Mb' nil) ('Mw' nil) ('M16' nil) ('M32' nil) ('M64' nil)
		('Nops' nil)
		('R' $r) ('Rd' $f) ('Rs' $f)
		('Xbr' 'rr') ('Xwr' 'rr') ('X16r' 'rr') ('X32r' 'rr') ('X64r' 'rr') ('X' $r) ('ow' nil)
		('Add' exit) ('And' exit) ('Or' exit)
		('Greater' exit) ('Xor' exit) ('r' $r))
				detect: [:pair| operands endsWith: pair first]
				ifNone: [])
			ifNil: [false]
			ifNotNil:
				[:match|
				match last ~~ #exit
				and: [operands := operands allButLast: match first size.
					 match last isCollection
						ifTrue: [printFormat addAllFirst: match last]
						ifFalse: [printFormat addFirst: match last].
					 true]]]
		whileTrue.
	(printFormat isEmpty and: [(operands beginsWith: 'Jump') or: [operands beginsWith: 'Call']]) ifTrue:
		[printFormat addFirst: nil].
	^printFormat
	
	"classPool keys sort collect: [:k| { k. (self printFormatForOpcodeName: k) asArray}]"
	"Find all initial opcodes ending with r or d; these must have an exit case"
	"classPool keys sort select: [:opcode| | idx | (idx := opcode allButFirst findFirst: [:c| c isUppercase]) > 0 and: ['rd' includes: (opcode at: idx)]]"! !

SharedPool subclass: #MIPSConstants
	instanceVariableNames: ''
	classVariableNames: 'A0 A1 A2 A3 ADDIU ADDU AND ANDI AT BEQ BGEZ BGTZ BLEZ BLTZ BNE BREAK DIV FP GP HintLoad HintStore J JAL JALR JR K0 K1 LB LBU LH LHU LUI LW MFHI MFLO MULT OR ORI OneInstruction PREF R0 R1 R10 R11 R12 R13 R14 R15 R16 R17 R18 R19 R2 R20 R21 R22 R23 R24 R25 R26 R27 R28 R29 R3 R30 R31 R4 R5 R6 R7 R8 R9 RA REGIMM S0 S1 S2 S3 S4 S5 S6 S7 SB SH SLL SLLV SLT SLTI SLTIU SLTU SP SPECIAL SRA SRAV SRL SRLV SUBU SW T0 T1 T2 T3 T4 T5 T6 T7 T8 T9 TwoInstructions V0 V1 XOR XORI ZR'
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!MIPSConstants commentStamp: 'rmacnak 11/11/2015 20:29:59' prior: 0!
MIPS opcodes and register names.!

!MIPSConstants class methodsFor: 'as yet unclassified' stamp: 'rmacnak 11/16/2015 20:29:35'!
initialize
	super initialize.
	
	OneInstruction := 4.
	TwoInstructions := 8.
	
	HintLoad := 0.
	HintStore := 1.
		
	self initializeRegisters.
	self initializeOpcodes.
	self initializeSpecialFunctions.
	self initializeRegImmRts.! !

!MIPSConstants class methodsFor: 'as yet unclassified' stamp: 'rmacnak 11/16/2015 20:30:32'!
initializeOpcodes
	SPECIAL := 0.
	REGIMM := 1.
	J := 2.
	JAL := 3.
	BEQ := 4.
	BNE := 5.
	BLEZ := 6.
	BGTZ := 7.
	ADDIU := 9.
	SLTI := 10.
	SLTIU := 11.
	ANDI := 12.
	ORI := 13.
	XORI := 14.
	LUI := 15.
	LB := 32.
	LH := 33.
	LW := 35.
	LBU := 36.
	LHU := 37.
	SB := 40.
	SH := 41.
	SW := 43.
	PREF := 51.! !

!MIPSConstants class methodsFor: 'as yet unclassified' stamp: 'rmacnak 11/6/2015 20:53:52'!
initializeRegImmRts
	BLTZ := 0.
	BGEZ := 1.! !

!MIPSConstants class methodsFor: 'as yet unclassified' stamp: 'rmacnak 11/10/2015 20:19:12'!
initializeRegisters
	self flag: #OABI.
	R0 := ZR := 0. "Hardwired zero"
	R1 := AT := 1. "Assembler temp - used to expand psuedo instructions"
	R2 := V0 := 2. "ABI: result register"
	R3 := V1 := 3. "ABI: result register"
	R4 := A0 := 4. "ABI: argument register"
	R5 := A1 := 5. "ABI: argument register"
	R6 := A2 := 6. "ABI: argument register"
	R7 := A3 := 7. "ABI: argument register"
	R8 := T0 := 8. "ABI: volatile"
	R9 := T1 := 9. "ABI: volatile"
	R10 := T2 := 10. "ABI: volatile"
	R11 := T3 := 11. "ABI: volatile"
	R12 := T4 := 12. "ABI: volatile"
	R13 := T5 := 13. "ABI: volatile"
	R14 := T6 := 14. "ABI: volatile"
	R15 := T7 := 15. "ABI: volatile"
	R16 := S0 := 16. "ABI: preserved"
	R17 := S1 := 17. "ABI: preserved"
	R18 := S2 := 18. "ABI: preserved"
	R19 := S3 := 19. "ABI: preserved"
	R20 := S4 := 20. "ABI: preserved"
	R21 := S5 := 21. "ABI: preserved"
	R22 := S6 := 22. "ABI: preserved"
	R23 := S7 := 23. "ABI: preserved"
	R24 := T8 := 24. "ABI: volatile"
	R25 := T9 := 25. "Special use in some position-independent code"
	R26 := K0 := 26. "Reserved for OS"
	R27 := K1 := 27. "Reserved for OS"
	R28 := GP := 28. "Special use in some position-independent code"
	R29 := SP := 29. "Stack pointer"
	R30 := FP := 30. "Frame pointer"
	R31 := RA := 31. "Link register"

	! !

!MIPSConstants class methodsFor: 'as yet unclassified' stamp: 'rmacnak 11/21/2015 18:14:45'!
initializeSpecialFunctions
	SLL := 0.
	SRL := 2.
	SRA := 3.
	SLLV := 4.
	SRLV := 6.
	SRAV := 7.
	JR := 8.
	JALR := 9.
	BREAK := 13.
	MFHI := 16.
	MFLO := 18.
	MULT := 24.
	DIV := 26.
	ADDU := 33.
	SUBU := 35.
	AND := 36.
	OR := 37.
	XOR := 38.
	SLT := 42.
	SLTU := 43.! !

!MIPSConstants class methodsFor: 'as yet unclassified' stamp: 'rmacnak 11/10/2015 21:03:56'!
nameForRegister: registerNumber
	^#(zr at v0 v1 a0 a1 a2 a3
		t0 t1 t2 t3 t4 t5 t6 t7
		s0 s1 s2 s3 s4 s5 s6 s7
		t8 t9 k0 k1 gp sp fp ra) at: registerNumber + 1! !

SharedPool subclass: #SpurMemoryManagementConstants
	instanceVariableNames: ''
	classVariableNames: 'DontTenure MarkOnTenure MaxRTRefCount ScavengeInProgress SlidingCompactionInProgress TenureByAge TenureByClass TenureToShrinkRT'
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManager'!

SharedPool subclass: #VMBasicConstants
	instanceVariableNames: ''
	classVariableNames: 'BaseHeaderSize BytecodeSetHasExtensions BytesPerOop BytesPerWord COGMTVM COGVM CloneOnGC CloneOnScavenge DisownVMForFFICall DisownVMForThreading DoAssertionChecks DoExpensiveAssertionChecks FastCPrimitiveAlignForFloatsFlag FastCPrimitiveFlag GCCheckFreeSpace GCCheckImageSegment GCCheckPrimCall GCCheckShorten GCModeBecome GCModeFull GCModeIncremental GCModeNewSpace HashMultiplyConstant HashMultiplyMask IMMUTABILITY LowcodeVM MULTIPLEBYTECODESETS NewspeakVM PharoVM PrimErrBadArgument PrimErrBadIndex PrimErrBadMethod PrimErrBadNumArgs PrimErrBadReceiver PrimErrCallbackError PrimErrFFIException PrimErrGenericFailure PrimErrInappropriate PrimErrInternalError PrimErrLimitExceeded PrimErrNamedInternal PrimErrNeedCompaction PrimErrNoCMemory PrimErrNoMemory PrimErrNoModification PrimErrNotFound PrimErrOSError PrimErrObjectIsPinned PrimErrObjectMayMove PrimErrObjectMoved PrimErrObjectNotPinned PrimErrOperationFailed PrimErrUninitialized PrimErrUnsupported PrimErrWritePastObject PrimNoErr PrimNumberHandlerMarker PrimNumberNoContextSwitchMarker PrimNumberUnwindMarker SPURVM STACKVM SistaVM TempVectReadBarrier VMBIGENDIAN'
	poolDictionaries: ''
	category: 'VMMaker-Interpreter'!

!VMBasicConstants commentStamp: '<historical>' prior: 0!
I am a shared pool for basic constants upon which the VM as a whole depends.

self ensureClassPool.
self classPool declare: #BytesPerWord from: VMSqueakV3ObjectRepresentationConstants classPool.
self classPool declare: #BaseHeaderSize from: VMSqueakV3ObjectRepresentationConstants classPool
(ObjectMemory classPool keys select: [:k| k beginsWith: 'Byte']) do:
	[:k| self classPool declare: k from: ObjectMemory classPool]!

TestCase subclass: #CoInterpreterTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: 'VMBasicConstants'
	category: 'VMMaker-Tests'!

!CoInterpreterTests methodsFor: 'tests' stamp: 'eem 1/24/2015 16:37'!
testMinimumUnusedHeadroom
	"self new testMinimumUnusedHeadroom"
	| ci bpw |
	CoInterpreter initializeWithOptions: Dictionary new.
	ci := CogVMSimulator new.
	bpw := ci objectMemory wordSize.
	ci initStackPagesForTests.
	self assert: ci minimumUnusedHeadroom = ci stackPageByteSize.
	0 to: ci stackPageByteSize - 1 by: bpw do:
		[:p|
		0 to: ci numStackPages - 1 do:
			[:i| | page |
			page := ci stackPages stackPageAt: i.
			ci longAt: page baseAddress - p put: 1].
		self assert: ci minimumUnusedHeadroom = (ci stackPageByteSize - (p + bpw))]! !

ReadStream subclass: #FakeStdinStream
	instanceVariableNames: 'atEnd simulator'
	classVariableNames: ''
	poolDictionaries: 'VMBasicConstants'
	category: 'VMMaker-JITSimulation'!

!FakeStdinStream commentStamp: '<historical>' prior: 0!
Fake Standard input using a dialog to prompt for a line of input at a time.!

!FakeStdinStream class methodsFor: 'instance creation' stamp: 'eem 9/16/2009 12:24'!
for: aCogVMSimulator
	^(self basicNew simulator: aCogVMSimulator)
		on: (String new: 80) from: 1 to: 0

	"self new next"! !

!FakeStdinStream class methodsFor: 'instance creation' stamp: 'eem 9/15/2009 17:25'!
new
	^super on: (String new: 80) from: 1 to: 0

	"self new next"! !

!FakeStdinStream methodsFor: 'testing' stamp: 'eem 9/16/2009 10:06'!
atEnd
	^atEnd ifNil: [atEnd := false]! !

!FakeStdinStream methodsFor: 'accessing' stamp: 'eem 10/1/2014 16:23'!
atEnd: aBoolean
	atEnd := aBoolean! !

!FakeStdinStream methodsFor: 'accessing' stamp: 'eem 9/15/2009 17:36'!
close
	atEnd := true! !

!FakeStdinStream methodsFor: 'testing' stamp: 'eem 9/19/2018 08:42'!
isFakeStdinStream
	^true! !

!FakeStdinStream methodsFor: 'accessing' stamp: 'eem 10/28/2020 19:42'!
next
	"Answer the next object in the Stream represented by the receiver.
	 If there are no more elements in the stream fill up the buffer by prompting for input"
	| sem inputLine next |
	position >= readLimit ifTrue:
		[simulator isThreadedVM
			ifTrue:
				[simulator forceInterruptCheckFromHeartbeat.
				 sem := Semaphore new.
				 WorldState addDeferredUIMessage:
					[inputLine := UIManager default request: 'Input please!!'.
					 sem signal].
				 sem wait]
			ifFalse: "simulate line-oriented input"
				[inputLine := ((Smalltalk classNamed: #FillInTheBlankMorph)
								ifNotNil: "Squeak"
									[:fITBM|
									 fITBM
										request: 'Input please!!'
										initialAnswer: ''
										centerAt: ActiveHand cursorPoint
										inWorld: ActiveWorld
										onCancelReturn: nil
										acceptOnCR: true]
								ifNil: "Pharo; onCancelReturn: nil is the default here"
									[UIManager default 
										request: 'Input please!!' 
										initialAnswer: '']).
				inputLine ifNil: [self atEnd: true. ^nil]].
		 collection size <= inputLine size ifTrue:
			[collection := collection species new: inputLine size + 1].
		 collection
			replaceFrom: 1 to: inputLine size with: inputLine startingAt: 1;
		 	at: (readLimit := inputLine size + 1) put: Character lf.
		 position := 0].
	next := collection at: (position := position + 1).
	^next
	

" This does it with workspaces:
| ws r s |
s := Semaphore new.
ws := Workspace new contents: ''.
ws acceptAction: [:t| r := t asString. s signal].
[ws openLabel: 'Yo!!'; shouldStyle: false.
(ws dependents detect: [:dep | dep isKindOf: PluggableTextMorph] ifNone: [nil]) ifNotNil:
	[:textMorph| textMorph acceptOnCR: true; hasUnacceptedEdits: true]] fork.
Processor activeProcess ==  Project uiProcess
	ifTrue: [[r isNil] whileTrue: [World doOneCycle]]
	ifFalse: [s wait].
ws topView delete.
s wait. s signal.
r"! !

!FakeStdinStream methodsFor: 'initialize-release' stamp: 'eem 9/16/2009 12:24'!
simulator: aCogVMSimulator
	simulator := aCogVMSimulator.
	atEnd := false! !

!FakeStdinStream methodsFor: 'accessing' stamp: 'eem 10/29/2020 17:57'!
sqFile: file Read: count Into: byteArrayIndexArg At: startIndex
	"Simulate the read primitive by answering a line of input"
	| inputLine n |
	position >= readLimit ifTrue:
		[simulator isThreadedVM ifTrue:
			[simulator forceInterruptCheckFromHeartbeat].
		 inputLine := FillInTheBlankMorph
						request: 'Input please!!'
						initialAnswer: ''
						centerAt: ActiveHand cursorPoint
						inWorld: ActiveWorld
						onCancelReturn: nil
						acceptOnCR: true.
		 inputLine ifNil: [self atEnd: true. ^0].
		 collection size <= inputLine size ifTrue:
			[collection := collection species new: inputLine size + 1].
		 collection
			replaceFrom: 1 to: inputLine size with: inputLine startingAt: 1;
		 	at: (readLimit := inputLine size + 1) put: Character lf.
		 position := 0].
	n := count min: readLimit - position.
	simulator
		strncpy: byteArrayIndexArg + startIndex
		_: (collection copyFrom: position + 1 to: position + n)
		_: n.
	position := position + n.
	^n! !

!VMBasicConstants class methodsFor: 'C translation' stamp: 'eem 2/13/2016 17:21'!
defineAtCompileTime: anObject
	^anObject isSymbol
	 and: [self namesDefinedAtCompileTime includes: anObject]! !

!VMBasicConstants class methodsFor: 'C translation' stamp: 'eem 11/14/2014 11:25'!
mostBasicConstantNames
	"c.f. mostBasicConstantSelectors"
	^#(BaseHeaderSize BytesPerWord BytesPerOop ShiftForWord BytesPerWord)! !

!VMBasicConstants class methodsFor: 'C translation' stamp: 'eem 11/19/2014 16:05'!
mostBasicConstantSelectors
	"c.f. mostBasicConstantNames"
	^#(	baseHeaderSize wordSize bytesPerOop shiftForWord
		bytesPerWord "bytesPerWord isn't used, but kept for history"
		minSmallInteger maxSmallInteger)! !

!VMBasicConstants class methodsFor: 'C translation' stamp: 'eem 7/7/2021 09:54'!
namesDefinedAtCompileTime
	"Answer the set of names for variables that should be defined at compile time.
	 Some of these get default values during simulation, and hence get defaulted in
	 the various initializeMiscConstants methods.  But that they have values should
	 /not/ cause the code generator to do dead code elimination based on their
	 default values.  In particular, methods marked with <option: ANameDefinedAtCompileTime>
	 will be emitted within #if defined(ANameDefinedAtCompileTime)...#endif.

	And of course this is backwards.  We'd like to define names that are defined at translation time."
	^#(VMBIGENDIAN
		IMMUTABILITY
		STACKVM COGVM COGMTVM SPURVM
		PharoVM								"Pharo vs Squeak"
		TerfVM VM_TICKER						"Terf vs Squeak & Qwaq/Teleplace/Terf high-priority thread support"
		EnforceAccessControl					"Newspeak"
		CheckRememberedInTrampoline		"IMMUTABILITY"
		BIT_IDENTICAL_FLOATING_POINT PLATFORM_SPECIFIC_FLOATING_POINT	"Alternatives for using fdlibm for floating-point"
		TestingPrimitives
		OBSOLETE_ALIEN_PRIMITIVES			"Ancient crap in the IA32ABI plugin"
		LLDB									"As of lldb-370.0.42 Swift-3.1, passing function parameters to printOopsSuchThat fails with Internal error [IRForTarget]: Couldn't rewrite one of the arguments of a function call.  Turning off link time optimization with -fno-lto has no effect.  hence we define some debugging functions as being <option: LLDB>"

		"processor related"
		__ARM_ARCH__ __arm__ __arm32__ ARM32 __arm64__ ARM64
		_M_I386 _X86_ i386 i486 i586 i686 __i386__ __386__ X86 I386
		x86_64 __amd64 __x86_64 __amd64__ __x86_64__ _M_AMD64 _M_X64
		__mips__ __mips
		__powerpc __powerpc__ __powerpc64__ __POWERPC__
		__ppc__ __ppc64__ __PPC__ __PPC64__
		__sparc__ __sparc __sparc_v8__ __sparc_v9__ __sparcv8 __sparcv9

		"Compiler brand related"
		__ACK__
		__CC_ARM
		__clang__
		__GNUC__
		_MSC_VER
		__ICC
		__SUNPRO_C
		
		"os related"
		ACORN
		_AIX
		__ANDROID__
		__APPLE__
		__BEOS__
		__linux__
		__MACH__
		__MINGW32__
		__FreeBSD__ __NetBSD__ __OpenBSD__
		__osf__
		EPLAN9
		__unix__ __unix UNIX
		WIN32 _WIN32 _WIN32_WCE
		WIN64 _WIN64 _WIN64_WCE)! !

!VMBasicConstants class methodsFor: 'C translation' stamp: 'eem 11/14/2014 11:28'!
valueOfBasicSelector: selector
	"Answer the numberic value of a basic constant accessed via the selector.
	 Answer nil if selector is not a basic constant accessor."
	"self mostBasicConstantSelectors collect: [:s| {s. self valueOfBasicSelector: s}]"
	| index key |
	^((index := self mostBasicConstantSelectors indexOf: selector) > 0
	  and: [(key := self mostBasicConstantNames at: index ifAbsent: []) notNil
	  and: [self includesKey: key]]) ifTrue:
		[classPool at: key]! !

SharedPool subclass: #VMBytecodeConstants
	instanceVariableNames: ''
	classVariableNames: 'BytecodeSetHasDirectedSuperSend CtxtTempFrameStart LargeContextBit LargeContextSize LargeContextSlots NewsqueakV4BytecodeSet PrimNumberDoExternalCall PrimNumberDoPrimitive PrimNumberExternalCall PrimNumberFFICall PrimNumberFlushExternalPrimitives PrimNumberInstVarAt PrimNumberShallowCopy PrimNumberSlotAt PrimNumberUnloadModule SistaV1BytecodeSet SmallContextSize SmallContextSlots SqueakV3PlusClosuresBytecodeSet'
	poolDictionaries: 'VMBasicConstants'
	category: 'VMMaker-Interpreter'!

!VMBytecodeConstants commentStamp: '<historical>' prior: 0!
self ensureClassPool.
#(CtxtTempFrameStart LargeContextBit LargeContextSize SmallContextSize) do:
	[:k|
	self classPool declare: k from: ObjectMemory classPool]!

!VMBytecodeConstants class methodsFor: 'simulator initialization' stamp: 'eem 4/15/2016 11:02'!
falsifyBytecodeSetFlags: initializationOptions
	SqueakV3PlusClosuresBytecodeSet.
	NewsqueakV4BytecodeSet.
	SistaV1BytecodeSet.
	classPool keys do:
		[:k|
		(k endsWith: 'BytecodeSet') ifTrue:
			[classPool at: k put: false.
			 initializationOptions at: k put: false]]

	"classPool keys select: [:k| k endsWith: 'BytecodeSet']"! !

VMBasicConstants subclass: #VMSpurObjectRepresentationConstants
	instanceVariableNames: ''
	classVariableNames: 'BecameActiveClassFlag BecameCompiledMethodFlag BecamePointerObjectFlag OldBecameNewFlag SpurPrimitiveAccessorDepthShift SpurPrimitiveFlagsMask'
	poolDictionaries: ''
	category: 'VMMaker-Interpreter'!

VMBasicConstants subclass: #VMSqueakClassIndices
	instanceVariableNames: ''
	classVariableNames: 'ClassArrayCompactIndex ClassBitmapCompactIndex ClassBlockClosureCompactIndex ClassBlockContextCompactIndex ClassByteArrayCompactIndex ClassByteStringCompactIndex ClassFloatCompactIndex ClassFullBlockClosureCompactIndex ClassLargeNegativeIntegerCompactIndex ClassLargePositiveIntegerCompactIndex ClassMessageCompactIndex ClassMethodContextCompactIndex ClassPointCompactIndex'
	poolDictionaries: ''
	category: 'VMMaker-Interpreter'!

LongTestCase subclass: #SpurImageSegmentTests
	instanceVariableNames: ''
	classVariableNames: 'CheckForLeaks'
	poolDictionaries: 'VMBasicConstants VMSqueakClassIndices'
	category: 'VMMaker-Tests'!

!SpurImageSegmentTests class methodsFor: 'accessing' stamp: 'eem 6/29/2017 16:07'!
resources
	^{SpurTrunkImageTestResource}! !

!SpurImageSegmentTests methodsFor: 'private' stamp: 'eem 6/29/2017 11:04'!
initializedVM
	^self resources anyOne current initializedVM cloneSimulation! !

!SpurImageSegmentTests methodsFor: 'tests' stamp: 'eem 6/21/2021 12:02'!
testSaveHashedCollectionAndAllSubclasses
	SimulatorHarnessForTests new
		withExecutableInterpreter: self initializedVM
		do: [:vm :harness| | error objects |
			CheckForLeaks == true ifTrue: "CheckForLeaks := self confirm: 'Check for leaks?'"
				[vm objectMemory setCheckForLeaks: (vm objectMemory class bindingOf: #GCCheckImageSegment) value].
			error := harness findSymbol: #error.
			self deny: error isNil.
			objects := harness
				interpreter: vm
				object: (harness findClassNamed: 'Compiler')
				perform: (harness findSymbol: #evaluate:)
				withArguments: {vm objectMemory stringForCString:
					'[| seg out roots result |
					 seg := WordArray new: 1024 * 1024.
					 out := Array new: 512.
					 roots := HashedCollection withAllSubclasses asArray.
					 roots := roots, (roots collect: [:ea| ea class]).
					 (thisContext isPrimFailToken: (nil tryPrimitive: 98 withArgs: { roots. seg. out })) ifTrue:
						[^#error].
					 result := { seg. out }.
					 (thisContext isPrimFailToken: (nil tryPrimitive: 99 withArgs: result)) ifTrue:
						[^#error].
					 result]
						on: Error
						do: [:ex| ^#error]'}.
			self deny: objects = error]! !

!SpurImageSegmentTests methodsFor: 'tests' stamp: 'eem 6/21/2021 12:02'!
testSaveHashedCollectionSubclasses
	SimulatorHarnessForTests new
		withExecutableInterpreter: self initializedVM
		do: [:vm :harness| | error objects |
			CheckForLeaks == true ifTrue: "CheckForLeaks := self confirm: 'Check for leaks?'"
				[vm objectMemory setCheckForLeaks: (vm objectMemory class bindingOf: #GCCheckImageSegment) value].
			error := harness findSymbol: #error.
			self deny: error isNil.
			objects := harness
				interpreter: vm
				object: (harness findClassNamed: 'Compiler')
				perform: (harness findSymbol: #evaluate:)
				withArguments: {vm objectMemory stringForCString:
					'[| seg out roots result |
					 seg := WordArray new: 1024 * 1024.
					 out := Array new: 256.
					 roots := HashedCollection subclasses asArray.
					 roots := roots, (roots collect: [:ea| ea class]).
					 (thisContext isPrimFailToken: (nil tryPrimitive: 98 withArgs: { roots. seg. out })) ifTrue:
						[^#error].
					 result := { seg. out }.
					 (thisContext isPrimFailToken: (nil tryPrimitive: 99 withArgs: result)) ifTrue:
						[^#error].
					 result]
						on: Error
						do: [:ex| ^#error]'}.
			self deny: objects = error]! !

LongTestCase subclass: #SpurPlanningCompactorTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: 'VMBasicConstants VMSqueakClassIndices'
	category: 'VMMaker-Tests'!

!SpurPlanningCompactorTests class methodsFor: 'debugger' stamp: 'eem 1/11/2017 10:37'!
defaultIntegerBaseInDebugger
	^VMClass defaultIntegerBaseInDebugger! !

!SpurPlanningCompactorTests class methodsFor: 'utilities' stamp: 'eem 6/29/2017 10:56'!
preenImage
	"Assume there's a Pharo bootstrap core32.image in ../oscogvm/image/core32.image.
	 We should find out where the image directory is and write a download script to get it.
	 But for now assume it's there.  See e.g.
		https://bintray.com/pharo-project/pharo/Pharo/201701061402-32bit#files"
	"[SpurPlanningCompactorTests preenImage] timeToRun"
	Spur32BitPreen new
		writeDefaultHeader: true;
		savedWindowSize: 640@480;
		preenImage: '../oscogvm/image/core32'! !

!SpurPlanningCompactorTests class methodsFor: 'accessing' stamp: 'eem 6/29/2017 12:31'!
resources
	^{SpurImageTestResource}! !

!SpurPlanningCompactorTests methodsFor: 'private' stamp: 'eem 6/21/2021 12:04'!
checkForLeaksIn: om
	om setCheckForLeaks: GCCheckFreeSpace + GCModeFull;
		runLeakCheckerFor: GCModeFull;
		checkFreeSpace: GCModeFull! !

!SpurPlanningCompactorTests methodsFor: 'failures' stamp: 'eem 6/29/2017 12:37'!
expectedFailures
	^(FileDirectory default fileExists: SpurImageTestResource imageNameForTests)
		ifTrue: [#()]
		ifFalse: [self testSelectors]! !

!SpurPlanningCompactorTests methodsFor: 'private' stamp: 'eem 6/29/2017 12:37'!
initializedVM
	| newVM |
	newVM := self resources anyOne current emptyVM cloneSimulation.
	newVM
		openOn: SpurImageTestResource imageNameForTests extraMemory: 0;
		initStackPages.
	newVM objectMemory
		initializeMarkStack; "The Pharo bootstrap has no mark or weakling stacks :-)"
		initializeWeaklingStack.
	^newVM! !

!SpurPlanningCompactorTests methodsFor: 'tests' stamp: 'eem 2/6/2017 10:59'!
testCompactedHeap
	"First test for valid compactibility of an already compacted heap via fullGC"
	| freeSpace om |
	om := self initializedVM objectMemory.
	freeSpace := om bytesLeftInOldSpace.
	om fullGC.
	self assert: freeSpace equals: om bytesLeftInOldSpace.
	self checkForLeaksIn: om.
	om fullGC.
	self assert: freeSpace equals: om bytesLeftInOldSpace.
	self checkForLeaksIn: om! !

!SpurPlanningCompactorTests methodsFor: 'tests' stamp: 'eem 1/11/2017 09:43'!
testIncompactibleHeap
	| errored |
	errored := false.
	"First test for incompactibility via fullGC"
	[self initializedVM objectMemory
		abandonEmptySegmentForTests;
		fullGC]
		on: Error
		do: [:ex|
			errored := true.
			self assert: ex messageText = 'uncompactable heap; no unmarked objects found'].
	self assert: errored.

	"Now check for incompactibility by directly calling compact"
	errored := false.
	"First test for incompactibility via fullGC"
	[| om |
	 om := self initializedVM objectMemory.
	 om abandonEmptySegmentForTests.
	 om allOldSpaceObjectsDo: [:o| om setIsMarkedOf: o to: true].
	 om compactor compact]
		on: Error
		do: [:ex|
			errored := true.
			self assert: ex messageText = 'uncompactable heap; no unmarked objects found'].
	self assert: errored! !

!SpurPlanningCompactorTests methodsFor: 'tests' stamp: 'eem 2/6/2017 11:12'!
testInitializedVM
	self checkForLeaksIn: self initializedVM objectMemory! !

!SpurPlanningCompactorTests methodsFor: 'tests' stamp: 'eem 2/3/2017 10:23'!
testMultiPassRandomAssortments
	"Test that the compactor can handle multi-pass compaction of some
	 number of random assortments of live, pinned, dead, and free chunks."
	<timeout: 60>
	| random |
	random := Random new.
	10 timesRepeat:
		[| theVM |
		 theVM := self initializedVM.
		 theVM objectMemory compactor forceMultiPass.
		 self testRandomAssortment: (random next: 3000) readStream
			with: theVM]! !

!SpurPlanningCompactorTests methodsFor: 'tests' stamp: 'eem 2/3/2017 17:07'!
testMultiPassRandomAssortmentsWithNewSegment
	"Test that the compactor can handle multi-pass compaction of some number of
	 random assortments of live, pinned, dead, and free chunks allocated in a new segment."
	<timeout: 60>
	| random |
	random := Random new.
	10 timesRepeat:
		[| theVM |
		 theVM := self initializedVM.
		 theVM objectMemory compactor forceMultiPass.
		 self testRandomAssortmentWithNewSegment: (random next: 5000) readStream
			with: theVM]! !

!SpurPlanningCompactorTests methodsFor: 'private' stamp: 'eem 06/22/2021 15:31'!
testRandomAssortment: random with: theVM
	"Test that the compactor can handle a random assortment of live, pinned, dead, and free chunks."
	| om lastObj obj expectedFreeSpace liveFill pinFill liveCount pinCount totalLive totalPinned pinned |
	random reset. "random is a read stream on 3000 random numbers; for repeatability"
	om := theVM objectMemory.
	om allOldSpaceObjectsDo: [:o| om setIsMarkedOf: o to: true. lastObj := o].
	pinFill := 16r99999900.
	liveFill := 16r55AA0000.
	liveCount := pinCount := expectedFreeSpace := 0.
	pinned := Set new.
	1000 timesRepeat:
		[| nSlots next newObj |
		 nSlots := (random next * 300) rounded. "Make sure we stray into overflow size field territory."
		 newObj := om allocateSlotsInOldSpace: nSlots format: om firstLongFormat classIndex: ClassByteArrayCompactIndex.
		 (next := random next) > 0.95
			ifTrue: "pinned"
				[om
					fillObj: newObj numSlots: nSlots with: pinFill + (pinCount := pinCount + 1);
					setIsPinnedOf: newObj to: true]
			ifFalse: "mobile"
				[om
					fillObj: newObj numSlots: nSlots with: liveFill + (liveCount := liveCount + 1)].
		 (next := random next) >= 0.333
			ifTrue:
				[om setIsMarkedOf: newObj to: true.
				 (om isPinned: newObj) ifTrue:
					[pinned add: newObj]]
			ifFalse: "dead or free"
				[expectedFreeSpace := expectedFreeSpace + (om bytesInBody: newObj).
				 (om isPinned: newObj) "Must check /before/ setObjectFree: which clears all bits"
					ifTrue: [pinCount := pinCount - 1]
					ifFalse: [liveCount := liveCount - 1].
				 next >= 0.2
					ifTrue: [om setIsMarkedOf: newObj to: false]
					ifFalse: [om setObjectFree: newObj]]].
	totalPinned := pinCount.
	totalLive := liveCount.
	self assert: totalPinned < (totalPinned + totalLive / 10). "should average 5%"

	"useful pre-compaction printing:"
	false ifTrue:
		[liveCount := pinCount := 0.
		 om allOldSpaceEntitiesFrom: (om objectAfter: lastObj) to: (om objectBefore: om endOfMemory) do:
			[:o|
			om coInterpreter print:
				((om isMarked: o)
					ifTrue: [(((om isPinned: o)
									ifTrue: [pinCount := pinCount + 1]
									ifFalse: [liveCount := liveCount + 1])
								printPaddedWith: Character space to: 3 base: 10), ' '] 
					ifFalse: ['     ']).
			 om printEntity: o].
		 om allOldSpaceEntitiesFrom: (om objectAfter: lastObj) to: (om objectBefore: om endOfMemory) do:
			[:o|
			((om isMarked: o) and: [om isPinned: o]) ifTrue:
				[om printEntity: o]]].

	expectedFreeSpace := expectedFreeSpace + om bytesLeftInOldSpace.
	om compactor compact.
	self assert: expectedFreeSpace equals: om bytesLeftInOldSpace.
	self assert: om allObjectsUnmarked.

	"useful post-compaction printing:"
	false ifTrue:
		[liveCount := pinCount := 0.
		 om allOldSpaceEntitiesFrom: (om objectAfter: lastObj) to: (om objectBefore: om endOfMemory) do:
			[:o|
			om coInterpreter print:
				((om isFreeObject: o)
					ifFalse: [(((om isPinned: o)
									ifTrue: [pinCount := pinCount + 1]
									ifFalse: [liveCount := liveCount + 1])
								printPaddedWith: Character space to: 3 base: 10), ' '] 
					ifTrue: ['     ']).
			 om printEntity: o].
		 om allOldSpaceEntitiesFrom: (om objectAfter: lastObj) to: (om objectBefore: om endOfMemory) do:
			[:o|
			(om isPinned: o) ifTrue:
				[om printEntity: o]]].

	"First check and/or count populations..."
	liveCount := pinCount := 0.
	om allOldSpaceObjectsFrom: (om objectAfter: lastObj) do:
		[:o|
		(om isPinned: o)
			ifTrue:
				[pinCount := pinCount + 1.
				 self assert: (pinned includes: o)]
			ifFalse: [liveCount := liveCount + 1]].
	self assert: totalPinned equals: pinCount.
	self assert: totalLive equals: liveCount.

	"Now check fills, which also tests update of first field on move..."
	liveCount := pinCount := 0.
	obj := lastObj.
	1 to: totalLive + totalPinned do:
		[:n| | expectedFill actualFill |
		 [obj := om objectAfter: obj. (om isEnumerableObject: obj) or: [obj >= om endOfMemory]] whileFalse.
		 expectedFill := (om isPinned: obj)
							ifTrue: [pinFill + (pinCount := pinCount + 1)]
							ifFalse: [liveFill + (liveCount := liveCount + 1)].
		 1 to: (om numSlotsOf: obj) do:
			[:i| self assert: expectedFill equals: (actualFill := om fetchPointer: i - 1 ofObject: obj)]].
	"They should be the last objects..."
	self assert: (om isFreeObject: (om objectAfter: obj)).
	self assert: om endOfMemory equals: (om addressAfter: (om objectAfter: obj)).
	self checkForLeaksIn: om! !

!SpurPlanningCompactorTests methodsFor: 'private' stamp: 'eem 06/22/2021 15:31'!
testRandomAssortmentWithNewSegment: random with: theVM
	"Test that the compactor can handle a random assortment of live, pinned, dead, and free chunks,
	 with some allocation in a new segment.  No live pinned objects are created in the new segment
	 to obtain the situation that the last segment is entirely empty after compaction.  This tests shrinkage."
	| om pig lastObj obj expectedFreeSpace liveFill pinFill liveCount pinCount totalLive totalPinned pinned |
	random reset. "random is a read stream on 3000 random numbers; for repeatability"
	om := theVM objectMemory.
	om allOldSpaceObjectsDo: [:o| om setIsMarkedOf: o to: true. lastObj := o].

	pinFill := 16r99999900.
	liveFill := 16r55AA0000.
	liveCount := pinCount := expectedFreeSpace := 0.
	pinned := Set new.

	1000 timesRepeat:
		[| nSlots next newObj |
		 nSlots := (random next * 300) rounded. "Make sure we stray into overflow size field territory."
		 newObj := om allocateSlotsInOldSpace: nSlots format: om firstLongFormat classIndex: ClassByteArrayCompactIndex.
		 (next := random next) > 0.95
			ifTrue: "pinned"
				[om
					fillObj: newObj numSlots: nSlots with: pinFill + (pinCount := pinCount + 1);
					setIsPinnedOf: newObj to: true]
			ifFalse: "mobile"
				[om
					fillObj: newObj numSlots: nSlots with: liveFill + (liveCount := liveCount + 1)].
		 (next := random next) >= 0.333
			ifTrue:
				[om setIsMarkedOf: newObj to: true.
				 (om isPinned: newObj) ifTrue:
					[pinned add: newObj]]
			ifFalse: "dead or free"
				[expectedFreeSpace := expectedFreeSpace + (om bytesInBody: newObj).
				 (om isPinned: newObj) "Must check /before/ setObjectFree: which clears all bits"
					ifTrue: [pinCount := pinCount - 1]
					ifFalse: [liveCount := liveCount - 1].
				 next >= 0.2
					ifTrue: [om setIsMarkedOf: newObj to: false]
					ifFalse: [om setObjectFree: newObj]]].

	 pig := om allocateSlotsInOldSpace: (om numSlotsOfAny: om findLargestFreeChunk) format: om firstLongFormat classIndex: ClassByteArrayCompactIndex.
	 self deny: pig isNil.
	 self assert: 0 equals: om bytesLeftInOldSpace.
	 om growOldSpaceByAtLeast: om growHeadroom // 2.
	 self assert: om growHeadroom equals: om bytesLeftInOldSpace + om bridgeSize.
	 expectedFreeSpace := expectedFreeSpace + (om bytesInBody: pig).

	1000 timesRepeat:
		[| nSlots next newObj |
		 nSlots := (random next * 300) rounded. "Make sure we stray into overflow size field territory."
		 newObj := om allocateSlotsInOldSpace: nSlots format: om firstLongFormat classIndex: ClassByteArrayCompactIndex.
		 "No pinned objects in second segment."
		 om fillObj: newObj numSlots: nSlots with: liveFill + (liveCount := liveCount + 1).
		 (next := random next) >= 0.333
			ifTrue:
				[om setIsMarkedOf: newObj to: true.
				 (om isPinned: newObj) ifTrue:
					[pinned add: newObj]]
			ifFalse: "dead or free"
				[expectedFreeSpace := expectedFreeSpace + (om bytesInBody: newObj).
				 liveCount := liveCount - 1.
				 next >= 0.2
					ifTrue: [om setIsMarkedOf: newObj to: false]
					ifFalse: [om setObjectFree: newObj]]].

	totalPinned := pinCount.
	totalLive := liveCount.
	self assert: totalPinned < (totalPinned + totalLive / 20). "should average 2.5%"

	"useful pre-compaction printing:"
	false ifTrue:
		[liveCount := pinCount := 0.
		 om allOldSpaceEntitiesFrom: (om objectAfter: lastObj) to: (om objectBefore: om endOfMemory) do:
			[:o|
			om coInterpreter print:
				((om isMarked: o)
					ifTrue: [(((om isPinned: o)
									ifTrue: [pinCount := pinCount + 1]
									ifFalse: [liveCount := liveCount + 1])
								printPaddedWith: Character space to: 3 base: 10), ' '] 
					ifFalse: ['     ']).
			 om printEntity: o].
		 om allOldSpaceEntitiesFrom: (om objectAfter: lastObj) to: (om objectBefore: om endOfMemory) do:
			[:o|
			((om isMarked: o) and: [om isPinned: o]) ifTrue:
				[om printEntity: o]]].

	expectedFreeSpace := expectedFreeSpace + om bytesLeftInOldSpace.
	om compactor compact.
	self assert: expectedFreeSpace equals: om bytesLeftInOldSpace.
	self assert: om allObjectsUnmarked.

	"useful post-compaction printing:"
	false ifTrue:
		[liveCount := pinCount := 0.
		 om allOldSpaceEntitiesFrom: (om objectAfter: lastObj) to: (om objectBefore: om endOfMemory) do:
			[:o|
			om coInterpreter print:
				((om isFreeObject: o)
					ifFalse: [(((om isPinned: o)
									ifTrue: [pinCount := pinCount + 1]
									ifFalse: [liveCount := liveCount + 1])
								printPaddedWith: Character space to: 3 base: 10), ' '] 
					ifTrue: ['     ']).
			 om printEntity: o].
		 om allOldSpaceEntitiesFrom: (om objectAfter: lastObj) to: (om objectBefore: om endOfMemory) do:
			[:o|
			(om isPinned: o) ifTrue:
				[om printEntity: o]]].

	"First check and/or count populations..."
	liveCount := pinCount := 0.
	om allOldSpaceObjectsFrom: (om objectAfter: lastObj) do:
		[:o|
		(om isPinned: o)
			ifTrue:
				[pinCount := pinCount + 1.
				 self assert: (pinned includes: o)]
			ifFalse: [liveCount := liveCount + 1]].
	self assert: totalPinned equals: pinCount.
	self assert: totalLive equals: liveCount.

	"Now check fills, which also tests update of first field on move..."
	liveCount := pinCount := 0.
	obj := lastObj.
	1 to: totalLive + totalPinned do:
		[:n| | expectedFill actualFill |
		 [obj := om objectAfter: obj. (om isEnumerableObject: obj) or: [obj >= om endOfMemory]] whileFalse.
		 expectedFill := (om isPinned: obj)
							ifTrue: [pinFill + (pinCount := pinCount + 1)]
							ifFalse: [liveFill + (liveCount := liveCount + 1)].
		 1 to: (om numSlotsOf: obj) do:
			[:i| self assert: expectedFill equals: (actualFill := om fetchPointer: i - 1 ofObject: obj)]].
	"the Last segment should be empty"
	self assert: (om segmentManager isEmptySegment: (om segmentManager segments at: 1)).
	"They should be the last objects, followed by a free object to the end fo the first segment, a bridge, then an empty segment with a single free object in it."
	self assert: (om isFreeObject: (om objectAfter: obj)).
	self assert: (om isSegmentBridge: (om objectAfter: (om objectAfter: obj))).
	self assert: (om isFreeObject: (om objectAfter: (om objectAfter: (om objectAfter: obj)))).
	self assert: om endOfMemory equals: (om addressAfter: (om objectAfter: (om objectAfter: (om objectAfter: obj)))).

	"And the memory should shrink if the shrinkThreshold is low enough"
	om shrinkThreshold: om growHeadroom.
	om attemptToShrink.
	self assert: om segmentManager numSegments = 1.
	self checkForLeaksIn: om! !

!SpurPlanningCompactorTests methodsFor: 'tests' stamp: 'eem 2/3/2017 10:20'!
testRandomAssortments
	"Test that the compactor can handle some number of random assortments of live, pinned, dead, and free chunks."
	<timeout: 60>
	| random |
	random := Random new.
	10 timesRepeat:
		[self testRandomAssortment: (random next: 3000) readStream
			with: self initializedVM]! !

!SpurPlanningCompactorTests methodsFor: 'tests' stamp: 'eem 2/3/2017 10:22'!
testRandomAssortmentsWithNewSegment
	"Test that the compactor can handle some number of random assortments of live, pinned, dead, and free chunks
	 allocated in a new segment."
	<timeout: 60>
	| random |
	random := Random new.
	10 timesRepeat:
		[self testRandomAssortmentWithNewSegment: (random next: 6000) readStream
			with: self initializedVM]! !

!SpurPlanningCompactorTests methodsFor: 'tests' stamp: 'eem 06/22/2021 15:31'!
testRunOfContiguousPinnedObjects
	"Test that the compactor can handle a long run of adjacent pinned objects across which it can and must move some unpinned objects."
	| om expectedFreeSpace firstPinnedObj gapObj obj |
	om := self initializedVM objectMemory.
	om allOldSpaceObjectsDo: [:o| om setIsMarkedOf: o to: true].
	"First create a gap"
	gapObj := om allocateSlotsInOldSpace: 100 format: om firstLongFormat classIndex: ClassArrayCompactIndex.
	om fillObj: gapObj numSlots: 100 with: om falseObject.
	self deny: (om isMarked: gapObj).
	"Now a long run of pinned objects."
	20 timesRepeat:
		[obj := om allocateSlotsInOldSpace: 4 format: om firstLongFormat classIndex: ClassBitmapCompactIndex.
		 om
			fillObj: obj numSlots: 4 with: 16r99999999;
			setIsPinnedOf: obj to: true;
			setIsMarkedOf: obj to: true.
		 firstPinnedObj ifNil:
			[firstPinnedObj := obj]].
	"Now something to move around it."
	obj := om allocateSlotsInOldSpace: 100 format: om firstLongFormat classIndex: ClassBitmapCompactIndex.
	om fillObj: obj numSlots: 100 with: 16r55AA55AA;
		setIsMarkedOf: obj to: true.
	"And something to move to the end of it."
	obj := om allocateSlotsInOldSpace: 100 format: om firstLongFormat classIndex: ClassBitmapCompactIndex.
	om fillObj: obj numSlots: 100 with: 16rAA55AA55;
		setIsMarkedOf: obj to: true.

	"useful debugging:""om printOopsFrom: gapObj to: om endOfMemory"
	expectedFreeSpace := om bytesLeftInOldSpace + (om bytesInBody: gapObj).
	om compactor compact.
	self assert: expectedFreeSpace equals: om bytesLeftInOldSpace.
	self assert: om allObjectsUnmarked.

	"The first mobile object past the pinned objects should have moved."
	self assert: ClassBitmapCompactIndex equals: (om classIndexOf: gapObj).
	self deny: (om isPinned: gapObj). 
	0 to: 99 do: [:i| self assert: 16r55AA55AA equals: (om fetchPointer: i ofObject: gapObj)].
	"The pinned objects should not have moved."
	obj := firstPinnedObj.
	20 timesRepeat:
		[self assert: ClassBitmapCompactIndex equals: (om classIndexOf: obj).
		 0 to: 3 do: [:i| self assert: 16r99999999 equals: (om fetchPointer: i ofObject: obj)].
		 obj := om objectAfter: obj].
	"The last object should have moved down."
	self assert: ClassBitmapCompactIndex equals: (om classIndexOf: obj).
	self deny: (om isPinned: obj). 
	0 to: 99 do: [:i| self assert: 16rAA55AA55 equals: (om fetchPointer: i ofObject: obj)].
	"It should be the last object..."
	self assert: (om isFreeObject: (om objectAfter: obj)).
	self assert: om endOfMemory equals: (om addressAfter: (om objectAfter: obj))! !

!SpurPlanningCompactorTests methodsFor: 'tests' stamp: 'eem 06/22/2021 15:31'!
testRunOfNonContiguousPinnedObjects
	"Test that the compactor can handle a long run of adjacent pinned objects separated by small ammounts of free space, across which it can and must move some unpinned objects."
	| om expectedFreeSpace firstPinnedObj gapObj obj numPins |
	om := self initializedVM objectMemory.
	om allOldSpaceObjectsDo: [:o| om setIsMarkedOf: o to: true].
	"First create a gap"
	gapObj := om allocateSlotsInOldSpace: 100 format: om firstLongFormat classIndex: ClassArrayCompactIndex.
	om fillObj: gapObj numSlots: 100 with: om falseObject.
	self deny: (om isMarked: gapObj).
	"Now a long run of pinned objects."
	(numPins := 20) timesRepeat:
		[obj := om allocateSlotsInOldSpace: 4 format: om firstLongFormat classIndex: ClassBitmapCompactIndex.
		 om
			fillObj: obj numSlots: 4 with: 16r99999999;
			setIsPinnedOf: obj to: true;
			setIsMarkedOf: obj to: true.
		 firstPinnedObj ifNil:
			[firstPinnedObj := obj].
		 om allocateSlotsInOldSpace: 4 format: om firstLongFormat classIndex: ClassArrayCompactIndex].
	"Now something to move around it."
	obj := om allocateSlotsInOldSpace: 100 format: om firstLongFormat classIndex: ClassBitmapCompactIndex.
	om fillObj: obj numSlots: 100 with: 16r55AA55AA;
		setIsMarkedOf: obj to: true.
	"And something to move to the end of it."
	obj := om allocateSlotsInOldSpace: 100 format: om firstLongFormat classIndex: ClassBitmapCompactIndex.
	om fillObj: obj numSlots: 100 with: 16rAA55AA55;
		setIsMarkedOf: obj to: true.

	"useful debugging:""om printOopsFrom: gapObj to: om endOfMemory"
	expectedFreeSpace := om bytesLeftInOldSpace + (om bytesInBody: gapObj) + ((om bytesInBody: firstPinnedObj) * numPins).
	om compactor compact.
	self assert: expectedFreeSpace equals: om bytesLeftInOldSpace.
	self assert: om allObjectsUnmarked.

	"The first mobile object past the pinned objects should have moved."
	self assert: ClassBitmapCompactIndex equals: (om classIndexOf: gapObj).
	self deny: (om isPinned: gapObj). 
	0 to: 99 do: [:i| self assert: 16r55AA55AA equals: (om fetchPointer: i ofObject: gapObj)].
	"The pinned objects should not have moved."
	obj := firstPinnedObj.
	1 to: numPins do:
		[:n|
		 self assert: ClassBitmapCompactIndex equals: (om classIndexOf: obj).
		 0 to: 3 do: [:i| self assert: 16r99999999 equals: (om fetchPointer: i ofObject: obj)].
		 obj := om objectAfter: obj.
		 n < numPins ifTrue:
			[self assert: (om isFreeObject: obj).
			 obj := om objectAfter: obj]].
	"The last object should have moved down."
	self assert: ClassBitmapCompactIndex equals: (om classIndexOf: obj).
	self deny: (om isPinned: obj). 
	0 to: 99 do: [:i| self assert: 16rAA55AA55 equals: (om fetchPointer: i ofObject: obj)].
	"It should be the last object..."
	self assert: (om isFreeObject: (om objectAfter: obj)).
	self assert: om endOfMemory equals: (om addressAfter: (om objectAfter: obj))! !

!SpurPlanningCompactorTests methodsFor: 'tests' stamp: 'eem 06/22/2021 15:31'!
testRunOfNonContiguousPinnedObjectsWithSpaceInBetween
	"Test that the compactor can handle a long run of adjacent pinned objects separated by large ammounts of free space, into which it can and must move some unpinned objects."
	| om expectedFreeSpace firstPinnedObj gapObj obj numPins firstFreeObj |
	om := self initializedVM objectMemory.
	om allOldSpaceObjectsDo: [:o| om setIsMarkedOf: o to: true].
	"First create a gap"
	gapObj := om allocateSlotsInOldSpace: 100 format: om firstLongFormat classIndex: ClassArrayCompactIndex.
	om fillObj: gapObj numSlots: 100 with: om falseObject.
	self deny: (om isMarked: gapObj).
	"Now a long run of pinned objects."
	(numPins := 10) timesRepeat:
		[obj := om allocateSlotsInOldSpace: 4 format: om firstLongFormat classIndex: ClassBitmapCompactIndex.
		 om
			fillObj: obj numSlots: 4 with: 16r99999999;
			setIsPinnedOf: obj to: true;
			setIsMarkedOf: obj to: true.
		 firstPinnedObj ifNil:
			[firstPinnedObj := obj].
		 obj := om allocateSlotsInOldSpace: 104 format: om firstLongFormat classIndex: ClassArrayCompactIndex.
		 firstFreeObj ifNil:
			[firstFreeObj := obj]].
	self deny: (om isMarked: firstFreeObj).
	"Now some objects to move around and into the run of pinned objects."
	numPins timesRepeat:
		[obj := om allocateSlotsInOldSpace: 100 format: om firstLongFormat classIndex: ClassBitmapCompactIndex.
		 om fillObj: obj numSlots: 100 with: 16r55AA55AA;
			setIsMarkedOf: obj to: true.
		 obj := om allocateSlotsInOldSpace: 100 format: om firstLongFormat classIndex: ClassBitmapCompactIndex.
		 om fillObj: obj numSlots: 100 with: 16rAA55AA55;
			setIsMarkedOf: obj to: true].

	"useful debugging:""om printOopsFrom: gapObj to: om endOfMemory"
	expectedFreeSpace := om bytesLeftInOldSpace + (om bytesInBody: gapObj) + ((om bytesInBody: firstFreeObj) * numPins).
	om compactor compact.
	self assert: expectedFreeSpace equals: om bytesLeftInOldSpace.
	self assert: om allObjectsUnmarked.

	"The first mobile object past the pinned objects should have moved. The pinned objects should not have moved.
	 We should see moved obj, pinned obj, (moved obj, free obj, pinned obj) +"
	obj := gapObj.
	1 to: numPins do:
		[:n|
		 self assert: ClassBitmapCompactIndex equals: (om classIndexOf: obj).
		 self deny: (om isPinned: obj). 
		 0 to: 99 do: [:i| self assert: (n odd ifTrue: [16r55AA55AA] ifFalse: [16rAA55AA55]) equals: (om fetchPointer: i ofObject: obj)].
		 obj := om objectAfter: obj.
		 n > 1 ifTrue:
			[self assert: (om isFreeObject: obj).
			 obj := om objectAfter: obj].
		 self assert: ClassBitmapCompactIndex equals: (om classIndexOf: obj).
		 self assert: (om isPinned: obj).
		 0 to: 3 do: [:i| self assert: 16r99999999 equals: (om fetchPointer: i ofObject: obj)].
		 obj := om objectAfter: obj].
	"The last objects should have moved down."
	1 to: numPins do:
		[:n|
		 self assert: ClassBitmapCompactIndex equals: (om classIndexOf: obj).
		 self deny: (om isPinned: obj). 
		 0 to: 99 do: [:i| self assert: (n odd ifTrue: [16r55AA55AA] ifFalse: [16rAA55AA55]) equals: (om fetchPointer: i ofObject: obj)]..
		 obj := om objectAfter: obj].
	"They should be the last objects..."
	self assert: (om isFreeObject: obj).
	self assert: om endOfMemory equals: (om addressAfter: obj)! !

!SpurPlanningCompactorTests methodsFor: 'tests' stamp: 'eem 06/22/2021 15:31'!
testRunOfObjectsWithExtraSegment
	"Test that the compactor can handle compacting more than one segment and shortening the memory."
	| om expectedFreeSpace pig gapObj obj |
	om := self initializedVM objectMemory.
	om allOldSpaceObjectsDo: [:o| om setIsMarkedOf: o to: true].
	"First create a gap"
	gapObj := om allocateSlotsInOldSpace: 100 format: om firstLongFormat classIndex: ClassArrayCompactIndex.
	om fillObj: gapObj numSlots: 100 with: om falseObject.
	self deny: (om isMarked: gapObj).
	expectedFreeSpace := om bytesInBody: gapObj.
	"Now some objects, a gap to a new segment and another run of objects."
	1 to: 2 do:
		[:i|
		10 timesRepeat:
			[obj := om allocateSlotsInOldSpace: 50 format: om firstLongFormat classIndex: ClassBitmapCompactIndex.
			 om fillObj: obj numSlots: 50 with: 16r55AA55AA;
				setIsMarkedOf: obj to: true.
			 obj := om allocateSlotsInOldSpace: 260 format: om firstLongFormat classIndex: ClassBitmapCompactIndex.
			om fillObj: obj numSlots: 260 with: 16rAA55AA55;
				setIsMarkedOf: obj to: true].
		i = 1 ifTrue:
			[pig := om allocateSlotsInOldSpace: (om numSlotsOfAny: om findLargestFreeChunk) format: om firstLongFormat classIndex: ClassBitmapCompactIndex.
			 self deny: pig isNil.
			 self assert: 0 equals: om bytesLeftInOldSpace.
			 om growOldSpaceByAtLeast: om growHeadroom // 2.
			 self assert: om growHeadroom equals: om bytesLeftInOldSpace + om bridgeSize.
			 expectedFreeSpace := expectedFreeSpace + (om bytesInBody: pig)]].

	"useful debugging:""om printOopsFrom: gapObj to: om endOfMemory"
	expectedFreeSpace := expectedFreeSpace + om bytesLeftInOldSpace.
	om compactor compact.
	self assert: expectedFreeSpace equals: om bytesLeftInOldSpace.
	self assert: om allObjectsUnmarked.

	"The first mobile object past the pinned objects should have moved."
	self assert: ClassBitmapCompactIndex equals: (om classIndexOf: gapObj).
	obj := gapObj.
	"The objects have moved."
	20 timesRepeat:
		[self assert: ClassBitmapCompactIndex equals: (om classIndexOf: obj).
		 0 to: (om numSlotsOf: obj) - 1 do: [:i| self assert: 16r55AA55AA equals: (om fetchPointer: i ofObject: obj)].
		 obj := om objectAfter: obj.
		 self assert: ClassBitmapCompactIndex equals: (om classIndexOf: obj).
		 0 to: (om numSlotsOf: obj) - 1 do: [:i| self assert: 16rAA55AA55 equals: (om fetchPointer: i ofObject: obj)].
		 obj := om objectAfter: obj].
	"the Last segment should be empty"
	self assert: (om segmentManager isEmptySegment: (om segmentManager segments at: 1)).
	"They should be the last objects, followed by a free object to the end fo the first segment, a bridge, then an empty segment with a single free object in it."
	self assert: (om isFreeObject: obj).
	self assert: (om isSegmentBridge: (om objectAfter: obj)).
	self assert: (om isFreeObject: (om objectAfter: (om objectAfter: obj))).
	self assert: om endOfMemory equals: (om addressAfter: (om objectAfter: (om objectAfter: obj))).

	"And the memory should shrink if the shrinkThreshold is low enough"
	om shrinkThreshold: om growHeadroom.
	om attemptToShrink.
	self assert: om segmentManager numSegments = 1! !

VMBasicConstants subclass: #VMSqueakV3ObjectRepresentationConstants
	instanceVariableNames: ''
	classVariableNames: 'AllButTypeMask CompactClassMask HashBitsOffset HashMaskUnshifted HeaderTypeShort HeaderTypeSizeAndClass ImmutabilityBit LongSizeMask MarkBit RootBit Size4Bit SizeMask TypeMask'
	poolDictionaries: ''
	category: 'VMMaker-Interpreter'!

!VMSqueakV3ObjectRepresentationConstants commentStamp: '<historical>' prior: 0!
I am a shared pool for the constants that define the Squeak V3 object representation shared between the object memories (e.g. ObjectMemory, NewObjectMemory), the interpreters (e.g. StackInterpreter, CoInterpreter) and the object representations (e.g. ObjectRepresentationForSqueakV3).

self ensureClassPool
self classPool declare: #AllButTypeMask from: VMObjectOffsets classPool
(ObjectMemory classPool keys select: [:k| k includesSubString: 'Compact']) do:
	[:k| self classPool declare: k from: ObjectMemory classPool]!

SharedPool subclass: #VMBytecodeConstants
	instanceVariableNames: ''
	classVariableNames: 'BytecodeSetHasDirectedSuperSend CtxtTempFrameStart LargeContextBit LargeContextSize LargeContextSlots NewsqueakV4BytecodeSet PrimNumberDoExternalCall PrimNumberDoPrimitive PrimNumberExternalCall PrimNumberFFICall PrimNumberFlushExternalPrimitives PrimNumberInstVarAt PrimNumberShallowCopy PrimNumberSlotAt PrimNumberUnloadModule SistaV1BytecodeSet SmallContextSize SmallContextSlots SqueakV3PlusClosuresBytecodeSet'
	poolDictionaries: 'VMBasicConstants'
	category: 'VMMaker-Interpreter'!

!VMBytecodeConstants commentStamp: '<historical>' prior: 0!
self ensureClassPool.
#(CtxtTempFrameStart LargeContextBit LargeContextSize SmallContextSize) do:
	[:k|
	self classPool declare: k from: ObjectMemory classPool]!

SharedPool subclass: #VMMethodCacheConstants
	instanceVariableNames: ''
	classVariableNames: 'LookupRuleDynamicSuper LookupRuleImplicit LookupRuleMNU LookupRuleOrdinary LookupRuleSelf MethodCacheClass MethodCacheEntries MethodCacheEntrySize MethodCacheMask MethodCacheMethod MethodCachePrim MethodCachePrimFunction MethodCacheSelector MethodCacheSize NSMethodCacheActualReceiver NSMethodCacheCallingMethod NSMethodCacheClassTag NSMethodCacheDepthOrLookupRule NSMethodCacheEntries NSMethodCacheEntrySize NSMethodCacheMask NSMethodCachePrimFunction NSMethodCacheSelector NSMethodCacheSize NSMethodCacheTargetMethod'
	poolDictionaries: ''
	category: 'VMMaker-Interpreter'!

!VMMethodCacheConstants commentStamp: '<historical>' prior: 0!
I am a shared pool for the constants that define the first-level method lookup cache.

self ensureClassPool.
#(#MethodCacheClass #MethodCacheEntries #MethodCacheEntrySize #MethodCacheMask #MethodCacheMethod #MethodCachePrimFunction #MethodCacheSelector #MethodCacheSize) do: [:k|
	self classPool declare: k from: StackInterpreter classPool]!

SharedPool subclass: #VMObjectIndices
	instanceVariableNames: ''
	classVariableNames: 'ActiveProcessIndex CharacterTable CharacterValueIndex ClassAlien ClassArray ClassBitmap ClassBlockClosure ClassBlockContext ClassByteArray ClassByteString ClassCharacter ClassDoubleByteArray ClassDoubleWordArray ClassExternalAddress ClassExternalData ClassExternalFunction ClassExternalLibrary ClassExternalStructure ClassFloat ClassFullBlockClosure ClassLargeNegativeInteger ClassLargePositiveInteger ClassMessage ClassMethodContext ClassMutex ClassPoint ClassSemaphore ClassSmallInteger ClassString ClassUnsafeAlien ClassWeakFinalizer ClassWordArray ClosureFirstCopiedValueIndex ClosureIndex ClosureNumArgsIndex ClosureOuterContextIndex ClosureStartPCIndex CompactClasses ConstMinusOne ConstOne ConstTwo ConstZero ExcessSignalsIndex ExternalObjectsArray FalseObject FirstLinkIndex ForeignCallbackProcess FullClosureCompiledBlockIndex FullClosureFirstCopiedValueIndex FullClosureReceiverIndex HeaderIndex InstanceSpecificationIndex InstructionPointerIndex KeyIndex LastLinkIndex LiteralStart LowcodeContextMark LowcodeNativeContextClass MessageArgumentsIndex MessageLookupClassIndex MessageSelectorIndex MethodArrayIndex MethodDictionaryIndex MethodIndex MyListIndex NextLinkIndex NilObject PrimErrTableIndex PriorityIndex ProcessInExternalCodeTag ProcessListsIndex ProcessSignalingLowSpace ReceiverIndex SchedulerAssociation SelectorAboutToReturn SelectorAttemptToAssign SelectorCannotInterpret SelectorCannotReturn SelectorCounterTripped SelectorDoesNotUnderstand SelectorInvokeCallback SelectorMustBeBoolean SelectorRunWithIn SelectorSistaTrap SelectorStart SelectorUnknownBytecode SenderIndex SpecialSelectors StackPointerIndex StreamArrayIndex StreamIndexIndex StreamReadLimitIndex StreamWriteLimitIndex SuperclassIndex SuspendedContextIndex TheDisplay TheFinalizationSemaphore TheInputSemaphore TheInterruptSemaphore TheLowSpaceSemaphore TheTimerSemaphore TrueObject ValueIndex XIndex YIndex'
	poolDictionaries: ''
	category: 'VMMaker-Interpreter'!

!VMObjectIndices commentStamp: '<historical>' prior: 0!
I am a shared pool for the constants that define object layout and well-known objects shared between the object memories (e.g. ObjectMemory, NewObjectMemory), the interpreters (e.g. StackInterpreter, CoInterpreter) and the object representations (e.g. ObjectRepresentationForSqueakV3).

self classPool declare: #Foo from: StackInterpreter classPool

(ObjectMemory classPool keys select: [:k| (k beginsWith: 'Class') and: [(k endsWith: 'Index') not]]) do:
	[:k| self classPool declare: k from: ObjectMemory classPool]!

SharedPool subclass: #VMStackFrameOffsets
	instanceVariableNames: ''
	classVariableNames: 'FoxCallerContext FoxCallerSavedIP FoxFrameFlags FoxIFReceiver FoxIFSavedIP FoxIFrameFlags FoxMFReceiver FoxMethod FoxReceiver FoxSavedFP FoxThisContext FrameSlots IFrameSlots MFMethodFlagHasContextFlag MFMethodFlagIsBlockFlag MFMethodFlagsMask MFMethodMask MFrameSlots'
	poolDictionaries: ''
	category: 'VMMaker-Interpreter'!

!VMStackFrameOffsets commentStamp: '<historical>' prior: 0!
I am a shared pool for the constants that define frame layout shared between the stack interpreters and the cogit.!

SharedPool subclass: #VMThreadingConstants
	instanceVariableNames: ''
	classVariableNames: 'AWOLProcessesIncrement CTMAssignableOrInVM CTMInitializing CTMUnavailable CTMWantingOwnership ThreadIdIndex ThreadIdMask'
	poolDictionaries: ''
	category: 'VMMaker-Multithreading'!

!VMThreadingConstants commentStamp: '<historical>' prior: 0!
VMThreadingConstants ensureClassPool.
CogThreadManager classPool keys do:
	[:k| VMThreadingConstants classPool declare: k from: CogThreadManager classPool].
CoInterpreterMT classPool keys do:
	[:k| VMThreadingConstants classPool declare: k from: CoInterpreterMT classPool].!

Behavior subclass: #Oop
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SmartSyntaxPlugins'!

!Oop commentStamp: '<historical>' prior: 0!
Cooercion specification for Identity coercions:

	x asOop: Oop == x
	y asValue: Oop == y!

!Oop class methodsFor: 'plugin generation' stamp: 'eem 10/2/2010 14:12'!
ccg: cg generateCoerceToOopFrom: aNode on: aStream
	"N.B.  The is used both for generation and simulation so answer the result (for interpretation)"
	^cg emitCExpression: aNode on: aStream! !

!Oop class methodsFor: 'plugin generation' stamp: 'eem 10/3/2010 08:09'!
ccg: cg generateCoerceToValueFrom: aNode on: aStream
	"N.B.  The could be used both for generation and simulation so answer the result (for interpretation)"
	^cg emitCExpression: aNode on: aStream! !

!Oop class methodsFor: 'plugin generation' stamp: 'acg 9/18/1999 15:58'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg ccgLoad: aBlock expr: aString asRawOopFrom: anInteger! !

!Oop class methodsFor: 'plugin generation' stamp: 'eem 12/18/2018 21:56'!
ccgCanConvertFrom: anObject

	^anObject class isImmediateClass not! !

!Oop class methodsFor: 'plugin generation' stamp: 'ikp 3/31/2005 14:20'!
ccgDeclareCForVar: aSymbolOrString

	^'sqInt ', aSymbolOrString! !

Behavior subclass: #Unsigned
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SmartSyntaxPlugins'!

!Unsigned commentStamp: 'tpr 5/5/2003 12:21' prior: 0!
Coercion specification for 32-bit unsigned numbers within plugins.

Note: "x asOop: Unsigned" can cause garbage collection when x is outside SmallInteger range.!

!Unsigned class methodsFor: 'plugin generation' stamp: 'eem 10/2/2010 14:12'!
ccg: cg generateCoerceToOopFrom: aNode on: aStream
	"N.B.  The is used both for generation and simulation so answer the result (for interpretation)"
	^cg generateCoerceToUnsignedObjectFrom: aNode on: aStream! !

!Unsigned class methodsFor: 'plugin generation' stamp: 'eem 10/3/2010 08:09'!
ccg: cg generateCoerceToValueFrom: aNode on: aStream
	"N.B.  The could be used both for generation and simulation so answer the result (for interpretation)"
	^cg generateCoerceToUnsignedValueFrom: aNode on: aStream! !

!Unsigned class methodsFor: 'plugin generation' stamp: 'eem 11/19/2008 16:05'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg ccgLoad: aBlock expr: aString asUnsignedValueFrom: anInteger! !

!Unsigned class methodsFor: 'plugin generation' stamp: 'acg 9/20/1999 11:30'!
ccgCanConvertFrom: anObject

	anObject isInteger ifFalse: 
		[self error: 'Not an Integer object'. ^false].
	anObject >= 0 ifFalse: 
		[self error: 'Object is negative integer'. ^false].
	anObject < (2 raisedToInteger: 32) ifFalse: 
		[self error: 'Object is too large'. ^false].
	^true! !

!Unsigned class methodsFor: 'plugin generation' stamp: 'nice 11/1/2016 21:56'!
ccgDeclareCForVar: aSymbolOrString

	^'usqIntptr_t ', aSymbolOrString! !

Behavior subclass: #WordsOrBytes
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SmartSyntaxPlugins'!

!WordsOrBytes commentStamp: '<historical>' prior: 0!
Coercion specification for bits objects like ByteString, Bitmap, ByteArray, WordArray, LargePositiveInteger et al!

!WordsOrBytes class methodsFor: 'plugin generation' stamp: 'eem 12/11/2008 14:18'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg 
		ccgLoad: aBlock 
		expr: aString 
		asCharPtrFrom: anInteger
		andThen: (cg ccgValBlock: 'isWordsOrBytes')! !

!WordsOrBytes class methodsFor: 'plugin generation' stamp: 'eem 12/11/2008 14:11'!
ccgCanConvertFrom: anObject

	^anObject class isBits! !

!WordsOrBytes class methodsFor: 'plugin generation' stamp: 'eem 12/11/2008 14:14'!
ccgDeclareCForVar: aSymbolOrString

	^'char *', aSymbolOrString! !

Behavior subclass: #WordsOrShorts
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SmartSyntaxPlugins'!

!WordsOrShorts commentStamp: 'eem 9/10/2020 12:45' prior: 0!
Coercion specification for bits objects organized in either 32-bit or 16-bit units like WordArray, DoubleByteArray, etc. Specifically this supports the sound primitives which were built assuming the original 32-bit SoundBuffer.  But now with Spur we can more conveniently use a DoubleByteArray like SoundBuffer with proper signed 16-bit elements.!

!WordsOrShorts class methodsFor: 'plugin generation' stamp: 'eem 12/1/2020 18:04'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg 
		ccgLoad: aBlock 
		expr: aString 
		asShortPtrFrom: anInteger
		andThen: (cg ccgValBlock: 'isWordsOrShorts')! !

!WordsOrShorts class methodsFor: 'plugin generation' stamp: 'eem 9/10/2020 12:46'!
ccgCanConvertFrom: anObject

	^anObject class isLongs or: [anObject class isShorts]! !

!WordsOrShorts class methodsFor: 'plugin generation' stamp: 'eem 9/10/2020 12:45'!
ccgDeclareCForVar: aSymbolOrString

	^'short *', aSymbolOrString! !

Object subclass: #CArray
	instanceVariableNames: 'interpreter arrayBaseAddress ptrOffset unitSize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!CArray commentStamp: '<historical>' prior: 0!
For Simulating the Interpreter only.  Coercing an Integer to a pointer (eg. cCoerce:to: 'char *') returns a CArray.  A CArray responds to at: and at:put: by reading/writing from the memory of the interpreter that sent the cCoerce:to: message in the first place.  A CArray responds to arithmetic by returning the new address.  Since longAt: sends // to its given address this is where the CArray converts back to a Integer for the memory fetch to work.!

CArray subclass: #BalloonArray
	instanceVariableNames: 'simArray simOffset'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!BalloonArray commentStamp: '<historical>' prior: 0!
BalloonArray keeps a shadow copy of its raw memory data in a Smalltalk array.  This allows support for C's inhomogeneous access, returning floats where Floats were stored, and negative ints where they were stored.  This ruse only works, of course where we have control over all the access.!

!BalloonArray methodsFor: 'pointer arithmetic' stamp: 'eem 10/22/2018 12:04'!
+= increment

	super += increment.
	simOffset := simOffset + (increment / unitSize)! !

!BalloonArray methodsFor: 'pointer arithmetic' stamp: 'eem 10/22/2018 12:05'!
-= increment

	super -= increment.
	simOffset := simOffset - (increment / unitSize)! !

!BalloonArray methodsFor: 'memory access' stamp: 'eem 10/22/2018 12:06'!
at: index
	| value |
	value := simArray at: index + simOffset.
	"Debug only..."
	value ifNil:
		[self error: 'attempt to read an uninitialized field'.
		^ super at: index  "Maybe it was set in Squeak.  Return the raw value"].
	(self bitsOf: value) ~= (super at: index) ifTrue:
		[self error: 'inconsistent values: ', (self bitsOf: value) printString, ' vs ', (super at: index) printString].
	^value! !

!BalloonArray methodsFor: 'memory access' stamp: 'eem 10/22/2018 12:07'!
at: index put: value

	super at: index put: (self bitsOf: value).
	^ simArray at: index + simOffset put: value.
	! !

!BalloonArray methodsFor: 'converting' stamp: 'eem 12/1/2020 17:17'!
beSigned! !

!BalloonArray methodsFor: 'converting' stamp: 'eem 12/1/2020 17:17'!
beUnsigned
	self shouldNotImplement! !

!BalloonArray methodsFor: 'memory access' stamp: 'eem 12/15/2019 21:18'!
bitsOf: value
	"Convert pos and neg ints and floats to 32-bit representations expected by C"
	value isInteger ifTrue:
		[^(value >= 0
			ifTrue: [value]
			ifFalse: [value + 16r80000000 + 16r80000000])  bitAnd: 16rFFFFFFFF].
	value isFloat ifTrue:
		[^value asIEEE32BitWord].
	self error: 'unexpected value for 32 bits'.
	^0! !

!BalloonArray methodsFor: 'memory access' stamp: 'di 7/15/2004 16:04'!
floatAt: index
	| value |
	value := self at: index.
	value isFloat ifFalse:
		[value = 0 ifTrue: [^ 0.0].
		self error: 'non-float was stored'.
		^ Float fromIEEE32Bit: value].
	^ value! !

!BalloonArray methodsFor: 'memory access' stamp: 'di 7/15/2004 13:00'!
floatAt: index put: value

	value isFloat
		ifFalse: [self error: 'inconsistent values'].
	^ self at: index put: value! !

!BalloonArray methodsFor: 'memory access' stamp: 'di 7/15/2004 13:02'!
intAt: index
	| value |
	value := self at: index.
	value isInteger
		ifFalse: [self error: 'inconsistent values'].
	^ value! !

!BalloonArray methodsFor: 'memory access' stamp: 'di 7/15/2004 13:01'!
intAt: index put: value

	value isInteger
		ifFalse: [self error: 'inconsistent values'].
	^ self at: index put: value! !

!BalloonArray methodsFor: 'memory access' stamp: 'eem 10/22/2018 12:07'!
setSimArray: anArray
	simArray := anArray.
	"Now sync the contents of the simArray with the actual work buffer.
	 The issue here is that whether an element is an in teger or a float depends
	 on how the BalloonArrayPlugin accesses the workBuffer; hence we look at
	 the existing values in simArray to find that interpretation."
	simOffset ifNil:
		[simOffset := 1].
	0 to: anArray size - 1 do:
		[:i| 
		 anArray
			at: i + simOffset
			put: ((anArray at: i + simOffset) isFloat
					ifTrue: [(Float fromIEEE32Bit: (super at: i))]
					ifFalse: [super at: i])]! !

!CArray methodsFor: 'int arithmetic' stamp: 'eem 10/9/2010 15:09'!
+ n
	^self shallowCopy += (n * unitSize)! !

!CArray methodsFor: 'pointer arithmetic' stamp: 'ajh 8/20/2002 01:34'!
+= increment

	ptrOffset := ptrOffset + increment! !

!CArray methodsFor: 'int arithmetic' stamp: 'eem 10/9/2010 15:09'!
- n
	^self shallowCopy -= (n * unitSize)! !

!CArray methodsFor: 'pointer arithmetic' stamp: 'ajh 8/20/2002 01:35'!
-= decrement

	ptrOffset := ptrOffset - decrement! !

!CArray methodsFor: 'comparing' stamp: 'eem 11/7/2013 12:43'!
= anObject
	^self species = anObject species
	  and: [(1 to: self class instSize) allSatisfy:
			[:i| (self instVarAt: i) = (anObject instVarAt: i)]]! !

!CArray methodsFor: 'converting' stamp: 'ajh 8/20/2002 01:29'!
adaptToNumber: rcvr andSend: selector

	^ rcvr perform: selector with: self asInteger! !

!CArray methodsFor: 'converting' stamp: 'eem 10/18/2018 15:09'!
asByteArray
	"Answer a ByteArray containing the receivers contents"
	| size bytes base |
	size := interpreter byteSizeOf: self cPtrAsOop.
	bytes := ByteArray new: size.
	base := arrayBaseAddress + ptrOffset - 1.
	1 to: size do:
		[:index|
		bytes at: index put: (interpreter byteAt: base + index)].
	^bytes! !

!CArray methodsFor: 'converting' stamp: 'di 7/15/2004 16:55'!
asCArrayAccessor

	^ (CArrayAccessor on: self)
			+= -1   "Defeat the +1 offset in the accessor"! !

!CArray methodsFor: 'converting' stamp: 'eem 10/11/2018 16:16'!
asDoubleFloatAccessor
	| accessor |
	CFloatArray adoptInstance: (accessor := self shallowCopy unitSize: 8).
	^accessor! !

!CArray methodsFor: 'converting' stamp: 'ajh 8/20/2002 01:10'!
asInteger

	^ self ptrAddress! !

!CArray methodsFor: 'converting' stamp: 'eem 10/11/2018 16:16'!
asSingleFloatAccessor
	| accessor |
	CFloatArray adoptInstance: (accessor := self shallowCopy unitSize: 4).
	^accessor! !

!CArray methodsFor: 'converting' stamp: 'eem 8/1/2021 13:46'!
asVoidPointer
	^self! !

!CArray methodsFor: 'accessing' stamp: 'eem 4/7/2016 09:32'!
at: offset
	| address |
	address := unitSize * offset + self ptrAddress.
	^unitSize <= 2
		ifTrue:
			[unitSize = 1
				ifTrue: [interpreter byteAt: address]
				ifFalse: [interpreter shortAt: address]]
		ifFalse:
			[unitSize = 4
				ifTrue: [interpreter long32At: address]
				ifFalse: [interpreter long64At: address]]! !

!CArray methodsFor: 'accessing' stamp: 'eem 4/7/2016 09:34'!
at: offset put: val
	| address |
	address := unitSize * offset + self ptrAddress.
	^unitSize <= 2
		ifTrue:
			[unitSize = 1
				ifTrue: [interpreter byteAt: address put: val]
				ifFalse: [interpreter shortAt: address put: val]]
		ifFalse:
			[unitSize = 4
				ifTrue: [interpreter long32At: address put: val]
				ifFalse: [interpreter long64At: address put: val]]! !

!CArray methodsFor: 'converting' stamp: 'eem 12/1/2020 17:14'!
beSigned
	SignedCArray adoptInstance: self! !

!CArray methodsFor: 'converting' stamp: 'eem 12/1/2020 17:17'!
beUnsigned! !

!CArray methodsFor: 'accessing' stamp: 'eem 12/28/2016 22:27'!
cPtrAsOop
	ptrOffset = 0 ifFalse: [self error: 'offset must be zero'].
	^arrayBaseAddress - interpreter baseHeaderSize! !

!CArray methodsFor: 'converting' stamp: 'eem 8/1/2021 14:51'!
coerceTo: cTypeString sim: interpreterSimulator

	^cTypeString caseOf: {
		['int']				-> [self ptrAddress].
		['float *']			-> [self asSingleFloatAccessor].
		['double *']			-> [self asDoubleFloatAccessor].
		['sqInt *']			-> [self shallowCopy unitSize: interpreter bytesPerOop; beSigned].
		['unsigned int *']	-> [self shallowCopy unitSize: 4; beUnsigned].
		['int *']				-> [self shallowCopy unitSize: 4; beSigned].
		['unsigned short *']	-> [self shallowCopy unitSize: 2; beUnsigned].
		['short *']			-> [self shallowCopy unitSize: 2; beSigned].
		['unsigned char *']	-> [self shallowCopy unitSize: 1; beUnsigned].
		['char *']			-> [self shallowCopy unitSize: 1; beUnsigned]. "C is ambivalent on the issue; sigh... SocketPlugin assumes unsigned"
		['unsigned']		-> [self ptrAddress].
		['sqInt']				-> [self ptrAddress].
		['usqInt']			-> [self ptrAddress].
		['sqIntptr_t']		-> [self] "this is used, in e.g. ffiAddressOf:startingAt:size: to coerce a pointer to an integer; it does not imply a real conversion. it is just to be able to pass the pointer as an integer" }! !

!CArray methodsFor: 'converting' stamp: 'ajh 8/20/2002 01:23'!
doesNotUnderstand: message

	^ self asInteger perform: message selector withArguments: message arguments! !

!CArray methodsFor: 'accessing' stamp: 'di 7/16/2004 12:45'!
floatAt: index
	^ Float fromIEEE32Bit: (self at: index)! !

!CArray methodsFor: 'accessing' stamp: 'di 7/16/2004 12:45'!
floatAt: index put: value
	^ self at: index put: value asIEEE32BitWord! !

!CArray methodsFor: 'comparing' stamp: 'eem 11/8/2013 13:25'!
hash
	^interpreter hash bitXor: arrayBaseAddress + ptrOffset + unitSize! !

!CArray methodsFor: 'accessing' stamp: 'di 7/16/2004 12:45'!
intAt: index
	^ (self at: index) signedIntFromLong! !

!CArray methodsFor: 'accessing' stamp: 'di 7/16/2004 12:45'!
intAt: index put: signedInt
	^ self at: index put: signedInt signedIntToLong! !

!CArray methodsFor: 'private' stamp: 'eem 4/7/2016 09:35'!
interpreter: interpreterSimulator address: arrayAddress unitSize: numBytes

	interpreter := interpreterSimulator.
	arrayBaseAddress := arrayAddress.
	self unitSize: numBytes.
	ptrOffset := 0! !

!CArray methodsFor: 'testing' stamp: 'akg 10/23/2018 21:52'!
isCArray
	"Answer a boolean indicating whether the receiver responds to the CArray interface"

	^true! !

!CArray methodsFor: 'accessing' stamp: 'eem 10/18/2018 15:27'!
overwriteContentsWith: aByteArray
	| base size |
	self assert: (interpreter isWordsOrBytes: self cPtrAsOop).
	size := (interpreter byteSizeOf: self cPtrAsOop) min: aByteArray size.
	base := arrayBaseAddress + ptrOffset - 1.
	1 to: size do:
		[:index|
		interpreter byteAt: base + index put: (aByteArray at: index)]! !

!CArray methodsFor: 'private' stamp: 'ajh 8/20/2002 00:36'!
ptrAddress

	^ arrayBaseAddress + ptrOffset! !

!CArray methodsFor: 'accessing' stamp: 'eem 10/9/2010 16:03'!
unitSize
	^unitSize! !

!CArray methodsFor: 'accessing' stamp: 'eem 4/7/2016 09:29'!
unitSize: n
	(n isPowerOfTwo and: [n <= 8]) ifFalse:
		[self error: 'unitSize must be 1, 2, 4 or 8'].
	unitSize := n! !

!CArray methodsFor: 'accessing' stamp: 'eem 4/6/2016 16:02'!
unsignedIntAt: index
	^self at: index! !

!CArray methodsFor: 'accessing' stamp: 'eem 4/6/2016 16:03'!
unsignedIntAt: index put: unsignedInt
	^ self at: index put: unsignedInt! !

CArray subclass: #CFloatArray
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!CFloatArray commentStamp: 'eem 4/7/2016 09:39' prior: 0!
A CFloatArray is a subclass of CArray that provides access via C float or double values!

!CFloatArray methodsFor: 'converting' stamp: 'eem 10/21/2018 19:32'!
asNonFloatAccessor
	| accessor |
	CArray adoptInstance: (accessor := self shallowCopy unitSize: 8).
	^accessor! !

!CFloatArray methodsFor: 'accessing' stamp: 'eem 4/7/2016 09:41'!
at: offset
	| address |
	address := unitSize * offset + self ptrAddress.
	^unitSize >= 4
		ifTrue:
			[unitSize = 4
				ifTrue: [Float fromIEEE32Bit: (interpreter long32At: address)]
				ifFalse: [Float fromIEEE64BitWord: (interpreter long64At: address)]]
		ifFalse:
			[self error: 'unitSize must be 4 or 8']! !

!CFloatArray methodsFor: 'accessing' stamp: 'eem 10/21/2018 12:57'!
at: offset put: val
	| address |
	address := unitSize * offset + self ptrAddress.
	^unitSize >= 4
		ifTrue:
			[unitSize = 4
				ifTrue: [interpreter long32At: address put: (val isFloat ifTrue: [val asIEEE32BitWord] ifFalse: [val])]
				ifFalse: [interpreter long64At: address put: (val isFloat ifTrue: [val asIEEE64BitWord] ifFalse: [val])]]
		ifFalse:
			[self error: 'unitSize must be 4 or 8']! !

!CFloatArray methodsFor: 'converting' stamp: 'eem 12/1/2020 17:15'!
beSigned! !

!CFloatArray methodsFor: 'converting' stamp: 'eem 12/1/2020 17:15'!
beUnsigned
	self shouldNotImplement! !

!CFloatArray methodsFor: 'converting' stamp: 'eem 8/1/2021 14:51'!
coerceTo: cTypeString sim: interpreterSimulator

	^cTypeString caseOf: {
		['float *']		-> [self shallowCopy unitSize: 4; yourself].
		['double *']		-> [self shallowCopy unitSize: 8; yourself].
		['sqIntptr_t']	-> [self] "this is used, in e.g. ffiAddressOf:startingAt:size: to coerce a pointer to an integer; it does not imply a real conversion. it is just to be able to pass the pointer as an integer" }
		otherwise: [self asNonFloatAccessor coerceTo: cTypeString sim: interpreterSimulator]! !

CArray subclass: #SignedCArray
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!SignedCArray methodsFor: 'accessing' stamp: 'eem 12/1/2020 17:23'!
at: offset
	| value |
	value := super at: offset.
	(value bitShift: unitSize * -8 + 1) = 0 ifTrue:
		[^value].
	^value - (1 bitShift: unitSize * 8)! !

!SignedCArray methodsFor: 'accessing' stamp: 'eem 12/1/2020 17:25'!
at: offset put: value
	super
		at: offset
		put: (value >= 0
				ifTrue: [value]
				ifFalse: [value bitAnd: (1 bitShift: unitSize * 8) - 1]).
	^value! !

!SignedCArray methodsFor: 'converting' stamp: 'eem 12/1/2020 17:15'!
beSigned! !

!SignedCArray methodsFor: 'converting' stamp: 'eem 12/1/2020 17:15'!
beUnsigned
	CArray adoptInstance: self! !

Object subclass: #CCodeGenerator
	instanceVariableNames: 'vmClass structClasses translationDict asArgumentTranslationDict inlineList constants variables variableDeclarations scopeStack methods macros apiMethods apiVariables kernelReturnTypes currentMethod headerFiles globalVariableUsage useSymbolicConstants generateDeadCode requiredSelectors previousCommentMarksInlining previousCommenter logger suppressAsmLabels asmLabelCounts pools selectorTranslations staticallyResolvedPolymorphicReceivers optionsDictionary breakSrcInlineSelectors breakDestInlineSelectors breakOnInline vmMaker accessorDepthCache beganInlining'
	classVariableNames: 'NoRegParmsInAssertVMs'
	poolDictionaries: 'VMBasicConstants'
	category: 'VMMaker-Translation to C'!

!CCodeGenerator commentStamp: 'tpr 5/2/2003 14:30' prior: 0!
This class oversees the translation of a subset of Smalltalk to C, allowing the comforts of Smalltalk during development and the efficiency and portability of C for the resulting interpreter.  
See VMMaker for more useful info!

!CCodeGenerator class methodsFor: 'class initialization' stamp: 'eem 6/27/2014 10:45'!
initialize
	"CCodeGenerator initialize"

	NoRegParmsInAssertVMs := true
		"If NoRegParmsInAssertVMs is true the generator spits out an attribute turning off register parameters for static functions in the Assert and Debug VMs which makes debugging easier, since all functions can be safely called from gdb.  One might hope that -mregparm=0 would work but at least on Mac OS X's gcc 4.2.1 it does not and hence we have to use a per funciton attribute.  Sigh..."! !

!CCodeGenerator class methodsFor: 'testing' stamp: 'eem 11/9/2017 11:40'!
isVarargsSelector: aSymbol
	^aSymbol endsWith: 'printf:'! !

!CCodeGenerator class methodsFor: 'C code generator' stamp: 'eem 2/9/2009 22:44'!
monticelloDescriptionFor: aClass
	"Answer a suitable Monticello package stamp to include in the header."
	| pkgInfo pkg uuid |
	pkgInfo := PackageOrganizer default packageOfClass: aClass.
	pkg := MCWorkingCopy allManagers detect: [:ea| ea packageName = pkgInfo packageName].
	pkg ancestry ancestors isEmpty ifFalse:
		[uuid := pkg ancestry ancestors first id].
	^aClass name, (pkg modified ifTrue: [' * '] ifFalse: [' ']), pkg ancestry ancestorString, ' uuid: ', uuid asString! !

!CCodeGenerator class methodsFor: 'removing from system' stamp: 'jm 5/16/1998 10:26'!
removeCompilerMethods
	"Before removing the C code generator classes from the system, use this method to remove the compiler node methods that support it. This avoids leaving dangling references to C code generator classes in the compiler node classes."

	ParseNode withAllSubclasses do: [ :nodeClass |
		nodeClass removeCategory: 'C translation'.
	].
	Smalltalk at: #AbstractSound ifPresent: [:abstractSound |
		 abstractSound class removeCategory: 'primitive generation'].
! !

!CCodeGenerator class methodsFor: 'C code generator' stamp: 'eem 3/7/2014 17:06'!
shortMonticelloDescriptionForClass: aClass
	"Answer a suitable Monticello package stamp to include in a moduleName."
	| mdesc |
	mdesc := [self monticelloDescriptionFor: aClass]
				on: Error
				do: [:ex| ^' ', Date today asString].
	^mdesc copyFrom: 1 to: (mdesc indexOfSubCollection: ' uuid:') - 1! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 7/23/2014 23:28'!
abortBlock
	^vmMaker ifNotNil: [:vmm| vmm abortBlock]! !

!CCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 2/16/2015 13:24'!
accessorChainsForMethod: method interpreterClass: interpreterClass
	"Answer a set of access paths from arguments through objects, in the method, assuming
	 it is a primitive. This is in support of Spur's lazy become.  A primitive may fail because it
	 may encounter a forwarder.  The primitive failure code needs to know to what depth it
	 must follow arguments to follow forwarders and, if any are found and followed, retry the
	 primitive. This method determines that depth. It starts by collecting references to the
	 stack and then follows these through assignments to variables and use of accessor
	 methods such as fetchPointer:ofObject:. For example
		| obj field  |
		obj := self stackTop.
		field := objectMemory fetchPointer: 1 ofObject: obj.
		self storePointer: 1 ofObject: field withValue: (self stackValue: 1)
	has depth 2, since field is accessed, and field is an element of obj."

	| accessors assignments roots chains extendedChains extended lastPass |
	self accessorsAndAssignmentsForMethod: method
		actuals: (self actualsForMethod: method)
		depth: 0
		interpreterClass: interpreterClass
		into: [:theRoots :theAccessors :theAssignments|
			roots := theRoots.
			accessors := theAccessors.
			assignments := theAssignments].
	"Compute the transitive closure of assignments of accessor sends or variables to variables from the roots.
	 Start from the stack accesses (the roots).
	 On the last pass look only for accessors of the targets of the tip assignments."
	chains := OrderedCollection new.
	roots do: [:root| chains addAll: (assignments
									select: [:assignment| assignment expression = root]
									thenCollect: [:assignment| OrderedCollection with: assignment])].
	lastPass := false.
	[extended := false.
	 extendedChains := OrderedCollection new: chains size * 2.
	 chains do:
		[:chain| | tip refs accessorRefs variableRefs |
		tip := chain last variable.
		refs := accessors select: [:send| send args anySatisfy: [:arg| tip isSameAs: arg]].
		lastPass ifFalse:
			[accessorRefs := refs collect: [:send|
											assignments
												detect: [:assignment|
														assignment expression = send
														and: [(chain includes: assignment) not]]
												ifNone: []]
									thenSelect: [:assignmentOrNil| assignmentOrNil notNil].
			 variableRefs := assignments select:
								[:assignment|
								 (tip isSameAs: assignment expression)
								 and: [(tip isSameAs: assignment variable) not
								 and: [(chain includes: assignment) not]]].
			 refs := (Set withAll: accessorRefs) addAll: variableRefs; yourself].
		refs isEmpty
			ifTrue:
				[extendedChains add: chain]
			ifFalse:
				[lastPass ifFalse: [extended := true].
				 self assert: (refs noneSatisfy: [:assignment| chain includes: assignment]).
				 extendedChains addAll: (refs collect: [:assignment| chain, {assignment}])]].
	 extended or: [lastPass not]] whileTrue:
		[chains := extendedChains.
		 extended ifFalse: [lastPass := true]].
	^chains! !

!CCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 1/3/2014 15:02'!
accessorDepthDeterminationFollowsSelfSends
	^false! !

!CCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 5/17/2015 16:38'!
accessorDepthForChain: chain "OrderedCollection"
	"Answer the actual number of accessors in the access chain, filtering out assignments of variables to variables."
	| accessorDepth |
	accessorDepth := 0.
	chain do:
		[:node|
		 ((node isAssignment and: [node expression isVariable])
		  or: [node isSend and: [SpurMemoryManager isSameLevelObjectAccessor: node selector]]) ifFalse:
			[accessorDepth := accessorDepth + 1]].
	^accessorDepth! !

!CCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 8/2/2021 17:55'!
accessorDepthForMethod: method
	"Compute the depth the method traverses object structure, assuming it is a primitive.
	 This is in support of Spur's lazy become.  A primitive may fail because it may encounter
	 a forwarder.  The primitive failure code needs to know to what depth it must follow
	  arguments to follow forwarders and, if any are found and followed, retry the primitive.
	 This method determines that depth. It starts by collecting references to the stack and
	 then follows these through assignments to variables and use of accessor methods
	 such as fetchPointer:ofObject:. For example
		| obj field  |
		obj := self stackTop.
		field := objectMemory fetchPointer: 1 ofObject: obj.
		self storePointer: 1 ofObject: field withValue: (self stackValue: 1)
	has depth 2, since field is accessed, and field is an element of obj.

	The information is cached since it needs to be computed *before* inlining"
	^accessorDepthCache
		at: method smalltalkSelector
		ifAbsentPut:
			[beganInlining
				ifTrue:
					[(method export
					 or: [vmClass notNil or: [vmClass primitiveTable includes: method smalltalkSelector]])
						ifTrue: [-1]
						ifFalse: [self error: 'it is too late to compute accessor depths!!']]
				ifFalse:
					 [((method definingClass includesSelector: method smalltalkSelector) ifTrue:
							[(method definingClass >> method smalltalkSelector) pragmaAt: #accessorDepth:])
						ifNil:
							[((self
									accessorChainsForMethod: method
									interpreterClass: (vmClass ifNil: [StackInterpreter]))
								inject: 0
								into: [:length :chain| length max: (self accessorDepthForChain: chain)]) - 1]
						ifNotNil: [:pragma| pragma arguments first]]]! !

!CCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 1/3/2014 10:37'!
accessorDepthForSelector: selector
	^(selector = #initialiseModule
	   or: [InterpreterPlugin includesSelector: selector]) ifFalse:
		[(self methodNamed: selector) ifNotNil:
			[:m| self accessorDepthForMethod: m]]! !

!CCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 5/17/2015 17:05'!
accessorsAndAssignmentsForMethod: method actuals: actualParameters depth: depth interpreterClass: interpreterClass into: aTrinaryBlock
	"Evaluate aTrinaryBlock with the root accessor sends, accessor sends and assignments in the method."
	| accessors assignments roots |
	accessors := Set new.
	assignments := Set new.
	roots := Set new.
	actualParameters with: method args do:
		[:actual :argName|
		 (actual isVariable or: [actual isSend]) ifTrue:
			[(actual isSend and: [self isStackAccessor: actual selector given: interpreterClass]) ifTrue:
				[roots add: actual].
			assignments add: (TAssignmentNode new
									setVariable: (TVariableNode new setName: argName)
									expression: actual)]].
	method parseTree nodesDo:
		[:node|
		node isSend ifTrue:
			[(self isStackAccessor: node selector given: interpreterClass) ifTrue:
				[roots add: node].
			 (self isObjectAccessor: node selector given: interpreterClass) ifTrue:
				[accessors add: node].
			 (self accessorDepthDeterminationFollowsSelfSends
			  and: [node receiver isVariable
			  and: [node receiver name = 'self'
			  and: [roots isEmpty
				or: [node args anySatisfy:
					[:arg|
					 (roots includes: arg)
					 or: [(accessors includes: arg)
					 or: [assignments anySatisfy: [:assignment| assignment variable isSameAs: arg]]]]]]]]) ifTrue:
				[self accessorsAndAssignmentsForSubMethodNamed: node selector
					actuals: node args
					depth: depth + 1
					interpreterClass: interpreterClass
					into: [:subRoots :subAccessors :subAssignments|
						(subRoots isEmpty and: [subAccessors isEmpty and: [subAssignments isEmpty]]) ifFalse:
							[roots addAll: subRoots.
							 accessors add: node.
							 accessors addAll: subAccessors.
							 assignments addAll: subAssignments]]]].
		(node isAssignment
		 and: [(node expression isSend and: [SpurMemoryManager isTerminalObjectAccessor: node expression selector]) not
		 and: [(roots includes: node expression)
			or: [(accessors includes: node expression)
			or: [node expression isVariable and: [node expression name ~= 'nil']]]]]) ifTrue:
			[assignments add: node]].
	^aTrinaryBlock
		value: roots
		value: accessors
		value: assignments! !

!CCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 2/16/2015 13:26'!
actualsForMethod: aTMethod
	"Normal primitives have no arguments, but translated primitives do.
	 This class doesn't handle translated primitives and so smply answers an empty array.
	 Subclasses override as required."
	^#()! !

!CCodeGenerator methodsFor: 'public' stamp: 'TPR 3/2/2000 11:22'!
addAllClassVarsFor: aClass
	"Add the class variables for the given class (and its superclasses) to the code base as constants."

	| allClasses |
	allClasses := aClass withAllSuperclasses.
	allClasses do: [:c | self addClassVarsFor: c].
! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 01/19/2021 20:40'!
addClass: aClass
	"Add the variables and methods of the given class to the code base."

	aClass prepareToBeAddedToCodeGenerator: self.
	self checkClassForNameConflicts: aClass.
	self addClassVarsFor: aClass.
	"ikp..."
	self addPoolVarsFor: aClass.
	(aClass inheritsFrom: VMStructType) ifFalse:
		[variables addAll: (self instVarNamesForClass: aClass)].
	self retainMethods: (aClass requiredMethodNames: self options).
	
	'Adding Class ' , aClass name , '...'
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: aClass selectors size
		during:
			[:bar |
			 aClass selectors withIndexDo:
				[:sel :i | | source |
				bar value: i.
				self addMethodFor: aClass selector: sel]].
	aClass declareCVarsIn: self! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 2/19/2012 11:32'!
addClassVarsFor: aClass
	"Add the class variables for the given class to the code base as constants."
	aClass classPool associationsDo:
		[:assoc | self addConstantForBinding: assoc]! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 2/13/2016 16:55'!
addConstantForBinding: variableBinding
	"Add the pool variable to the code base as a constant."
	| node val |
	val := variableBinding value.
	node := (useSymbolicConstants and: [self isCLiteral: val])
				ifTrue:[TDefineNode new
							setName: variableBinding key
							value: variableBinding value]
				ifFalse:[TConstantNode new setValue: variableBinding value].
	constants at: variableBinding key put: node! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 12/14/2012 11:22'!
addHeaderFile: aString
	"Add a header file. As a hack we allow C preprocessor defs such as #ifdef"
	self assert: (('"<#' includes: aString first) or: [(aString last: 2) = '_H']).
	(aString first ~= $#
	 and: [headerFiles includes: aString]) ifTrue:
		[logger nextPutAll: 'warning, attempt to include ', aString, ' a second time'; cr; flush.
		 ^self].
	headerFiles addLast: aString! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 12/14/2012 11:02'!
addHeaderFileFirst: aString
	"Add a header file to the front of the sequence."
	self assert: (('"<' includes: aString first) and: ['">' includes: aString last]).
	self assert: (headerFiles includes: aString) not.
	headerFiles addFirst: aString! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 2/14/2009 09:10'!
addMacro: aString for: selector
	"Add a macro. aString must be the macro arguments and body without the leading #define or name"
	macros at: selector put: aString! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 5/1/2014 08:12'!
addMethod: aTMethod
	"Add the given method to the code base and answer it.
	 Only allow duplicate definitions for struct accessors, since we don't actually
	 generate code for these methods and hence the conflict doesn't matter.
	 Allow subclasses to redefine methods (Smalltalk has inheritance after all)."

	(methods at: aTMethod selector ifAbsent: []) ifNotNil:
		[:conflict |
		aTMethod compiledMethod isSubclassResponsibility ifTrue:
			[^nil].
		(conflict isStructAccessor
		 and: [aTMethod isStructAccessor
		 and: [conflict compiledMethod decompileString = aTMethod compiledMethod decompileString]]) ifTrue:
			[^nil].
		((aTMethod definingClass inheritsFrom: conflict definingClass)
		 or: [(aTMethod compiledMethod pragmaAt: #option:) notNil]) ifFalse:
			[self error: 'Method name conflict: ', aTMethod selector]].
	^methods at: aTMethod selector put: aTMethod! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 9/9/2018 09:57'!
addMethodFor: aClass selector: selector
	"Add the given method to the code base and answer its translation
	 or nil if it shouldn't be translated."

	| method tmethod |
	method := aClass compiledMethodAt: selector.
	(method pragmaAt: #doNotGenerate) ifNotNil:
		["only remove a previous method if this one overrides it, i.e. this is a subclass method.
		 If the existing method is in a different hierarchy this method must be merely a redeirect."
		 (methods at: selector ifAbsent: []) ifNotNil:
			[:tm|
			(aClass includesBehavior: tm definingClass) ifTrue:
				[self removeMethodForSelector: selector]].
		 ^nil].
	method isSubclassResponsibility ifTrue:
		[^nil].
	(self shouldIncludeMethodFor: aClass selector: selector) ifFalse:
		[^nil].
	tmethod := self compileToTMethodSelector: selector in: aClass.
	"Even though we exclude initialize methods, we must consider their
	 global variable usage, otherwise globals may be incorrectly localized."
	selector == #initialize ifTrue:
		[self checkForGlobalUsage: (tmethod allReferencedVariablesUsing: self) in: tmethod.
		 ^nil].
	self addMethod: tmethod.
	"If the method has a macro then add the macro.  But keep the method
	 for analysis purposes (e.g. its variable accesses)."
	(method pragmaAt: #cmacro:) ifNotNil:
		[:pragma|
		self addMacro: (pragma argumentAt: 1) for: selector.
		(inlineList includes: selector) ifTrue:
			[inlineList := inlineList copyWithout: selector]].
	(method pragmaAt: #cmacro) ifNotNil:
		[:pragma| | literal | "Method should be just foo ^const"
		self assert: (method numArgs = 0 and: [method numLiterals = 3 or: [method isQuick]]).
		literal := method isQuick
					ifTrue: [method decompile quickMethodReturnLiteral]
					ifFalse: [method literalAt: 1].
		self addMacro: '() ', (method isReturnField
								ifTrue: [literal]
								ifFalse: [self cLiteralFor: literal value name: method selector]) for: selector.
		(inlineList includes: selector) ifTrue:
			[inlineList := inlineList copyWithout: selector]].
	^tmethod! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 2/16/2015 17:46'!
addMethodsForTranslatedPrimitives: classAndSelectorList 
	| verbose |
	verbose := false.
	classAndSelectorList do:
		[:classAndSelector | | aClass selector meth |
		aClass := Smalltalk at: classAndSelector first.
		selector := classAndSelector last.
		self addAllClassVarsFor: aClass.

		"compile the method source and convert to a suitable translation method.
		 find the method in either the class or the metaclass"
		meth := self
					compileToTMethodSelector: selector
					in: ((aClass includesSelector: selector)
							ifTrue: [aClass]
							ifFalse: [aClass class]).
		meth primitive > 0 ifTrue:
			[meth preparePrimitiveName].
		meth replaceSizeMessages.
		self addMethod: meth].

	self prepareMethods! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 2/13/2016 16:56'!
addPoolVarsFor: aClass 
	"Add the pool variables for the given class to the code base as constants."

	(aClass sharedPools reject: [:pool| pools includes: pool]) do:
		[:pool |
		pools add: pool.
		pool bindingsDo:
			[:binding |
			self addConstantForBinding: binding]]! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 12/11/2010 13:42'!
addSelectorTranslation: aSelector to: aString
	selectorTranslations at: aSelector asSymbol put: aString! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 01/19/2021 20:40'!
addStructClass: aClass
	"Add the non-accessor methods of the given struct class to the code base."

	aClass prepareToBeAddedToCodeGenerator: self.
	self addClassVarsFor: aClass.
	self addPoolVarsFor: aClass.
	self retainMethods: (aClass requiredMethodNames: self options).
	
	'Adding Class ' , aClass name , '...'
		displayProgressAt: Sensor cursorPoint
		from: 0
		to: aClass selectors size
		during:
			[:bar |
			 aClass selectors withIndexDo:
				[:sel :i | | source |
				bar value: i.
				self addStructMethodFor: aClass selector: sel]].
	aClass declareCVarsIn: self! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 4/18/2016 17:18'!
addStructClasses: classes
	"Add the struct classes and save them for emitCTypesOn: later."
	structClasses := classes.
	structClasses do:
		[:structClass|
		 (structClass withAllSuperclasses copyUpTo: VMStructType) do:
			[:structClassOrSuperclass|
			 self addStructClass: structClassOrSuperclass]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 4/18/2016 17:50'!
addStructMethodFor: aClass selector: selector 
	"Add the given struct method to the code base and answer its translation
	 or nil if it shouldn't be translated."
	(self methodNamed: selector) ifNotNil:
		[:tmethod|
		 "If we're repeating an attempt to add the same thing, or
		  if the existing method overrides this one,don't complain."
		 (tmethod definingClass includesBehavior: aClass) ifTrue:
			[^self].
		 "If the methods are both simple accessors, don't complain."
		 ((tmethod definingClass isAccessor: selector)
		 and: [aClass isAccessor: selector]) ifTrue:
			[^self].
		 "If the method is overriding a method in a superclass, don't complain"
		 (aClass inheritsFrom: tmethod definingClass)
			ifTrue: [methods removeKey: selector]
			ifFalse: [self error: 'conflicting implementations for ', selector storeString]].
	^(self addMethodFor: aClass selector: selector) ifNotNil:
		[:tmethod|
		tmethod transformToStructClassMethodFor: self.
		tmethod]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 3/31/2015 11:26'!
addVariablesInVerbatimCIn: aCCodeSendNode to: aCollection
	"If aCCodeSendNode has a string argument, parse it and extract anything
	 that looks like a variable, and add the resulting vars to aCollection."
	| separators tokens |
	(aCCodeSendNode isSend
	 and: [(aCCodeSendNode selector beginsWith: #cCode:)
	 and: [aCCodeSendNode args first isConstant
	 and: [aCCodeSendNode args first value isString]]]) ifFalse:
		[^self].
	separators := (Character space to: 255 asCharacter) reject:
					[:char|
					char isLetter or: [char isDigit or: [char = $_]]].
	tokens := aCCodeSendNode args first value findTokens: separators.
	aCollection addAll: (tokens select: [:token| token first isLetter]) asSet! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 12/2/2013 17:12'!
anyMethodNamed: selector
	"Answer any method in the code base (including api methods) with the given selector."

	^methods
		at: selector
		ifAbsent:
			[apiMethods ifNotNil:
				[apiMethods
					at: selector
					ifAbsent: []]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 2/7/2014 10:35'!
arrayInitializerCalled: varName for: array sizeString: sizeStringOrNil type: cType
	"array is a literal array or a CArray on some array."
	^String streamContents:
		[:s| | sequence lastLine index newLine allIntegers |
		sequence := array isCollection ifTrue: [array] ifFalse: [array object].
		"this is to align -ve and +ve integers nicely in the primitiveAccessorDepthTable"
		allIntegers := sequence allSatisfy: [:element| element isInteger].
		lastLine := index := 0.
		newLine := [sequence size >= 20
						ifTrue: [s cr; nextPutAll: '/*'; print: index; nextPutAll: '*/'; tab]
						ifFalse: [s crtab: 2].
					 lastLine := s position].
		s	nextPutAll: cType;
			space;
			nextPutAll: varName;
			nextPut: $[.
		sizeStringOrNil ifNotNil: [s nextPutAll: sizeStringOrNil].
		s nextPutAll: '] = '.
		sequence isString
			ifTrue: [s nextPutAll: (self cLiteralFor: sequence)]
			ifFalse:
				[s nextPut: ${.
				 newLine value.
				 sequence
					do: [:element|
						(allIntegers
						 and: [element < 0
						 and: [s peekLast = Character space]]) ifTrue:
							[s skip: -1].
						s nextPutAll: (self cLiteralFor: element). index := index + 1]
					separatedBy:
						[s nextPut: $,.
						 ((s position - lastLine) >= 76
						 or: [(index \\ 20) = 0])
							ifTrue: [newLine value]
							ifFalse: [s space]].
				 s crtab; nextPut: $}]]! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 11/14/2019 21:11'!
atMostThreeBytecodeMessagesFrom: aCompiledMethod
	"If aCompiledMethod has two or three btecodes, then answer a sequence of the abstract bytewcode messages
	 for those bytecodes, otherwise answer nil."
	| msgs encoderClass endpc scanner |
	msgs := OrderedCollection new.
	scanner := InstructionStream new method: aCompiledMethod pc: aCompiledMethod initialPC.
	endpc  := aCompiledMethod endPC.
	encoderClass := aCompiledMethod encoderClass.
	[scanner pc <= endpc] whileTrue:
		"i.e. nil will not understand any message and so the exception block will collect all of them."
		[[encoderClass interpretNextInstructionFor: nil in: scanner]
			on: MessageNotUnderstood
			do: [:ex| (msgs addLast: ex message; yourself) size > 3 ifTrue: [^nil]]].
	^msgs! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 3/7/2015 16:25'!
baseTypeForPointerType: aCType
	"Answer the type of the referent of a pointer type."
	self assert: aCType last == $*.
	^self baseTypeForType: aCType allButLast! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 10/29/2018 19:03'!
baseTypeForType: aCType
	"Reduce various declarations to the most basic type we can determine."
	| type fpIndex closeidx openidx |
	type := aCType.
	((openidx := type indexOfSubCollection: 'const ') > 0
	and: [openidx = 1 or: [(type at: openidx) isSeparator]]) ifTrue:
		[type := type copyReplaceFrom: openidx to: openidx + 5 with: ''].
	((type beginsWith: 'unsigned') and: [(type includes: $:) and: [type last isDigit]]) ifTrue:
		[^#usqInt].
	"collapse e.g. void (*foo(int bar))(void) to void (*)(void)"
	(fpIndex := type indexOfSubCollection: '(*') > 0 ifTrue:
		["elide the function arguments after *, if there are any"
		 type := type copyReplaceFrom: (type indexOf: $( startingAt: fpIndex + 1)
					to: (type indexOf: $) startingAt: fpIndex + 1)
					with: ''.
		 "elide the function name after *, if there is one"
		 type := type copyReplaceFrom: fpIndex + 2
					to: (type indexOf: $) startingAt: fpIndex + 1)
					with: ')'].
	"collapse [size] to *"
	openidx := 0.
	[(openidx := type indexOf: $[ startingAt: openidx + 1) > 0
	 and: [(closeidx := type indexOf: $] startingAt: openidx + 1) > 0]] whileTrue:
		[type := type copyReplaceFrom: openidx to: closeidx with: '*'].

	"map foo* to foo *"
	^self conventionalTypeForType: type! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 2/23/2015 16:45'!
breakDestInlineSelector: aSelector
	breakDestInlineSelectors add: aSelector! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 2/16/2015 10:40'!
breakOnInline: aBooleanOrNil
	breakOnInline := aBooleanOrNil! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 2/23/2015 16:44'!
breakSrcInlineSelector: aSelector
	breakSrcInlineSelectors add: aSelector! !

!CCodeGenerator methodsFor: 'utilities' stamp: ''!
cCodeForMethod: selector
	"Answer a string containing the C code for the given method."
	"Example:
		((CCodeGenerator new initialize addClass: TestCClass1; prepareMethods)
			cCodeForMethod: #ifTests)"

	| m s |
	m := self methodNamed: selector.
	m = nil ifTrue: [ self error: 'method not found in code base: ', selector ].

	s := (ReadWriteStream on: '').
	m emitCCodeOn: s generator: self.
	^ s contents! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 10/11/2018 17:18'!
cFunctionNameFor: aSelector
	"Create a C function name from the given selector by finding
	 a specific translation, or if none, simply omitting colons, and
	 any trailing underscores (this supports a varargs convention)."
	^selectorTranslations
		at: aSelector
		ifAbsent:
			[| cSelector |
			 cSelector := aSelector copyWithout: $:.
			 aSelector last = $: ifTrue:
				[[cSelector last = $_] whileTrue:
					[cSelector := cSelector allButLast]].
			 cSelector]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'nice 12/29/2020 11:55'!
cLiteralFor: anObject
	"Return a string representing the C literal value for the given object."
		
	anObject isNumber
		ifTrue:
			[anObject isInteger ifTrue:
				[| hex dec useHexa |
				 hex := anObject printStringBase: 16.
				 dec := anObject printStringBase: 10.
				 useHexa := (anObject > 255
								and: [(hex asSet size * 3) <= (dec asSet size * 2)
									or: [((hex as: RunArray) runs size * 4) < ((dec as: RunArray) runs size * 3)]])
					or: [anObject > 0
								and: [(anObject >> anObject lowBit + 1) isPowerOfTwo
								and: [(anObject highBit = anObject lowBit and: [anObject > 65536])
									  or: [anObject highBit - anObject lowBit >= 4]]]].
				^self cLiteralForInteger: anObject hex: useHexa].
			anObject isFloat ifTrue:
				[^anObject printString]]
		ifFalse:
			[anObject isSymbol ifTrue:
				[^self cFunctionNameFor: anObject].
			anObject isString ifTrue:
				[^'"', (anObject copyReplaceAll: (String with: Character cr) with: '\n') , '"'].
			anObject == nil ifTrue: [^ 'null' ].
			anObject == true ifTrue: [^ '1' ].
			anObject == false ifTrue: [^ '0' ].
			anObject isCharacter ifTrue:
				[^anObject == $'
					ifTrue: ['''\'''''] "i.e. '\''"
					ifFalse: [anObject asString printString]]].
	self error: 'Warning: A Smalltalk literal could not be translated into a C constant: ', anObject printString.
	^'"XXX UNTRANSLATABLE CONSTANT XXX"'! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'nice 4/10/2016 23:12'!
cLiteralFor: anObject name: smalltalkName
	"Return a string representing the C literal value for the given object.
	 This version may use hex for integers that are bit masks."
	anObject isInteger ifTrue:
		[| hex dec useHexa |
		hex := anObject printStringBase: 16.
		dec := anObject printStringBase: 10.
		useHexa := ((smalltalkName endsWith: 'Mask')
				or: [anObject digitLength > 1
					and: [(hex asSet size * 3) <= (dec asSet size * 2)
					and: [(smalltalkName endsWith: 'Size') not]]]).
		^self cLiteralForInteger: anObject hex: useHexa].
	^self cLiteralFor: anObject! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'nice 4/10/2016 23:48'!
cLiteralForInteger: anInteger hex: aBoolean
	"Answer the string for generating a literal integer.
	Use hexadecimal notation as prescribed by aBoolean.
	Use long long suffix (LL) if the integer does not fit on 32 bits.
	Use unsigned suffix (U) if the integer does not fit on a signed integer (resp. long long).
	Correctly generate INT_MIN and LONG_LONG_MIN.
	Indeed -0x8000000 is parsed as - (0x8000000) by C Compiler.
	0x8000000 does not fit on a signed int, it is interpreted as unsigned.
	That makes INT_MIN unsigned which is badly broken..."
	
	| printString |
	printString := aBoolean
		ifTrue: [anInteger positive
			ifTrue: ['0x' , (anInteger printStringBase: 16)]
			ifFalse: ['-0x' , (anInteger negated printStringBase: 16)]]
		ifFalse: [anInteger printString].
	^anInteger positive
		ifTrue: [anInteger > 16r7FFFFFFF "INT_MAX"
			ifTrue: [anInteger > 16rFFFFFFFF "UINT_MAX"
				ifTrue: [anInteger > 16r7FFFFFFFFFFFFFFF "LONG_LONG_MAX"
					ifTrue: [printString , 'ULL']
					ifFalse: [printString , 'LL']]
				ifFalse: [printString , 'U']]
			ifFalse: [printString]]
		ifFalse: [anInteger < -16r8000000
			ifTrue: [anInteger = -16r800000000000000	"LONG_LONG_MIN"
				ifTrue: ['(-0x7FFFFFFFFFFFFFFFLL-1)']
				ifFalse: [printString , 'LL']]
			ifFalse: [anInteger = -16r8000000	"INT_MIN"
				ifTrue: ['(-0x7FFFFFFF-1)']
				ifFalse: [printString]]]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 11/13/2017 15:45'!
cLiteralForPrintfString: aString
	^(('"', (PrintfFormatString new setFormat: aString) transformForVMMaker, '"')
		copyReplaceAll: (String with: Character cr) with: '\n')
			copyReplaceAll: (String with: Character tab) with: '\t'! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'nice 7/19/2016 13:34'!
cLiteralForUnsignedInteger: anInteger hex: aBoolean longlong: llBoolean
	"Answer the string for generating an unsigned literal integer.
	Use hexadecimal notation as prescribed by aBoolean.
	Force long long suffix (LL) if the integer does not fit on 32 bits, or if llBoolean is true."
	
	| printString |
	printString := aBoolean
		ifTrue: [anInteger positive
			ifTrue: ['0x' , (anInteger printStringBase: 16)]
			ifFalse: ['-0x' , (anInteger negated printStringBase: 16)]]
		ifFalse: [anInteger printString].
	^anInteger positive
		ifTrue: [(llBoolean or: [anInteger > 16rFFFFFFFF "UINT_MAX"])
			ifTrue: [printString , 'ULL']
			ifFalse: [printString , 'U']]
		ifFalse: [self error: 'please provide positive integer']! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'nice 7/19/2016 13:37'!
cLiteralForUnsignedInteger: anInteger longlong: llBoolean
	"Answer the string for generating an unsigned literal integer.
	Eventually use hexadecimal.
	Force long long suffix (LL) if the integer does not fit on 32 bits, or if llBoolean is true."
	
	| hex |
	hex := (anInteger > 0
				and: [(anInteger >> anInteger lowBit + 1) isPowerOfTwo
				and: [(anInteger highBit = anInteger lowBit and: [anInteger > 65536])
					  or: [anInteger highBit - anInteger lowBit >= 4]]]).
	^self cLiteralForUnsignedInteger: anInteger hex: hex longlong: llBoolean! !

!CCodeGenerator methodsFor: 'error notification' stamp: 'eem 2/13/2016 16:44'!
checkClassForNameConflicts: aClass
	"Verify that the given class does not have constant, variable, or method names that conflict with
	 those of previously added classes. Raise an error if a conflict is found, otherwise just return."

	"check for constant name collisions in class pools"
	aClass classPool associationsDo:
		[:assoc |
		(constants includesKey: assoc key) ifTrue:
			[self error: 'Constant ', assoc key, ' was defined in a previously added class']].

	"and in shared pools"
	(aClass sharedPools reject: [:pool| pools includes: pool]) do:
		[:pool |
		pool bindingsDo:
			[:assoc |
			(constants includesKey: assoc key) ifTrue:
				[self error: 'Constant ', assoc key, ' was defined in a previously added class']]].

	"check for instance variable name collisions"
	(aClass inheritsFrom: VMStructType) ifFalse:
		[(self instVarNamesForClass: aClass) do:
			[:varName |
			(variables includes: varName) ifTrue:
				[self error: 'Instance variable ', varName, ' was defined in a previously added class']]].

	"check for method name collisions"
	aClass selectors do:
		[:sel | | tmeth meth |
		((self shouldIncludeMethodFor: aClass selector: sel)
		and: [(tmeth := methods at: sel ifAbsent: nil) notNil
		and: [(aClass isStructClass and: [(aClass isAccessor: sel)
				and: [(methods at: sel) isStructAccessor]]) not
		and: [(meth := aClass >> sel) isSubclassResponsibility not
		and: [(aClass includesBehavior: tmeth definingClass) not]]]]) ifTrue:
			[((aClass >>sel) pragmaAt: #option:)
				ifNil: [self error: 'Method ', sel, ' was defined in a previously added class.']
				ifNotNil:
					[logger
						ensureCr;
						show: 'warning, method ', aClass name, '>>', sel storeString,
								' overrides ', tmeth definingClass, '>>', sel storeString;
						cr]]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 2/12/2009 13:44'!
checkDeleteVariable: aName
	"Hook for debugging variable deletion."! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 9/15/2015 11:31'!
checkForGlobalUsage: vars in: aTMethod 
	vars do:
		[:var |
		(variables includes: var) ifTrue: "find the set of method names using this global var"
			[(globalVariableUsage at: var ifAbsentPut: [Set new])
				add: aTMethod selector]].
	aTMethod clearReferencesToGlobalStruct.
	(aTMethod locals select: [:l| self reservedWords includes: l]) do:
		[:l| | em |
		em := aTMethod definingClass name, '>>', aTMethod smalltalkSelector, ' has variable that is a C reserved word: ', l.
		self error: em.
		self logger cr; nextPutAll: em; cr; flush]! !

!CCodeGenerator methodsFor: 'public' stamp: 'ikp 9/26/97 14:48'!
codeString
	"Return a string containing all the C code for the code base. Used for testing."

	| stream |
	stream := ReadWriteStream on: (String new: 1000).
	self emitCCodeOn: stream doInlining: true doAssertions: true.
	^stream contents! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 11/14/2019 12:18'!
collectInlineList: inlineFlagOrSymbol
	"Make a list of methods that should be inlined.
		If inlineFlagOrSymbol == false, inline nothing.
		If inlineFlagOrSymbol == #asSpecified include methods marked with <inline: true>, and <inline: #always>.
		If inlineFlagOrSymbol == #asSpecifiedOrQuick include methods marked with <inline: true>, and <inline: #always>,
			and methods that are quick (^constant, ^inst var).
		If inlineFlagOrSymbol == true include methods marked with <inline: true>, and <inline: #always>,
			and methods with 40 parse nodes or fewer."

	"Details: The method must not include any inline C, since the
	 translator cannot currently map variable names in inlined C code.
	 Methods to be inlined must be small or called from only one place."

	| selectorsOfMethodsNotToInline callsOf |
	self assert: (#(true false asSpecified asSpecifiedOrQuick) includes: inlineFlagOrSymbol).
	selectorsOfMethodsNotToInline := Set new: methods size.
	selectorsOfMethodsNotToInline addAll: macros keys.
	apiMethods ifNotNil:
		[selectorsOfMethodsNotToInline addAll: apiMethods keys].
	methods do:
		[:m|
		m isStructAccessor ifTrue:
			[selectorsOfMethodsNotToInline add: m selector]].

	"build dictionary to record the number of calls to each method"
	callsOf := Dictionary new: methods size * 2.
	methods keysAndValuesDo:
		[:s :m|
		(m isRealMethod
		 and: [self shouldGenerateMethod: m]) ifTrue:
			[callsOf at: s put: 0]].

	"For each method, scan its parse tree once or twice to:
		1. determine if the method contains unrenamable C code or declarations or has a C builtin
		2. determine how many nodes it has
		3. increment the sender counts of the methods it calls"
	inlineList := Set new: methods size * 2.
	(methods reject: [:m| selectorsOfMethodsNotToInline includes: m selector]) do:
		[:m| | inlineIt hasUnrenamableCCode nodeCount |
		((breakSrcInlineSelectors includes: m selector)
		 and: [breakOnInline isNil]) ifTrue:
			[self halt].
		inlineIt := #dontCare.
		(translationDict includesKey: m selector)
			ifTrue: [hasUnrenamableCCode := true]
			ifFalse:
				[hasUnrenamableCCode := m hasUnrenamableCCode.
				 nodeCount := 0.
				 m parseTree nodesDo:
					[:node|
					node isSend ifTrue:
						[callsOf
							at: node selector
							ifPresent:
								[:senderCount| callsOf at: node selector put: senderCount + 1]].
					 nodeCount := nodeCount + 1].
				inlineIt := m extractInlineDirective].  "may be true, false, #always, #never or #dontCare"
		(hasUnrenamableCCode or: [inlineIt == false])
			ifTrue: "don't inline if method has C code or contains negative inline directive"
				[inlineIt == true ifTrue:
					[logger
						ensureCr;
						nextPutAll: 'failed to inline ';
						nextPutAll: m selector;
						nextPutAll: ' as it contains unrenamable C declarations or C code';
						cr; flush].
				selectorsOfMethodsNotToInline add: m selector]
			ifFalse:
				[(inlineFlagOrSymbol caseOf: {
						[#asSpecified]			-> [inlineIt == true].
						[#asSpecifiedOrQuick]	-> [inlineIt == true
													or: [self isQuickCompiledMethod: m compiledMethod]].
						[true]					-> [nodeCount < 40 or: [inlineIt == true]].
						[false]					-> [false]})
					ifTrue: "inline if method has no C code and is either small or contains inline directive"
						[inlineList add: m selector]
					ifFalse:
						[(#(asSpecified asSpecifiedOrQuick) includes: inlineFlagOrSymbol) ifTrue:
							[selectorsOfMethodsNotToInline add: m selector]]]].

	(#(asSpecified asSpecifiedOrQuick) includes: inlineFlagOrSymbol)
		ifTrue:
			[methods do: [:m| m inline ifNil: [m inline: (inlineList includes: m selector)]]]
		ifFalse:
			[callsOf associationsDo:
				[:assoc|
				(assoc value = 1
				 and: [(selectorsOfMethodsNotToInline includes: assoc key) not]) ifTrue:
					[inlineList add: assoc key]]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 7/27/2021 08:30'!
compileToTMethodSelector: selector in: aClass
	"Compile a method to a TMethod"
	(breakSrcInlineSelectors notNil
	 and: [(breakSrcInlineSelectors includes: selector)
	 and: [breakOnInline ~~ true]]) ifTrue:
		[self halt].

	^(aClass >> selector) asTranslationMethodOfClass: self translationMethodClass

	"was:
	| implementingClass |
	implementingClass := aClass.
	^(Compiler new
		parse: ([aClass sourceCodeAt: selector]
					on: KeyNotFound
					do: [:ex| ""Quick hack for simulating Pharo images...""
						(PharoVM and: [aClass == String class and: [selector == #findSubstringViaPrimitive:in:startingAt:matchTable:]]) ifFalse:
							[ex pass].
						(implementingClass := ByteString) sourceCodeAt: #findSubstring:in:startingAt:matchTable:])
		in: implementingClass
		notifying: nil)
			asTranslationMethodOfClass: self translationMethodClass"! !

!CCodeGenerator methodsFor: 'public' stamp: 'rsf 8/9/2016 21:08'!
computeKernelReturnTypes
	| dictionary |
	dictionary := Dictionary newFromPairs:
		#(oopAt: #sqInt oopAt:put: #sqInt
			oopAtPointer: #sqInt oopAtPointer:put: #sqInt
		 byteAt: #sqInt byteAt:put: #sqInt
			byteAtPointer: #sqInt byteAtPointer:put: #sqInt
		 shortAt: #sqInt shortAt:put: #sqInt
			shortAtPointer: #sqInt shortAtPointer:put: #sqInt
		 intAt: #sqInt intAt:put: #sqInt
			intAtPointer: #sqInt intAtPointer:put: #sqInt
		 longAt: #sqInt longAt:put: #sqInt
			longAtPointer: #sqInt longAtPointer:put: #sqInt
				long32At: #int long32At:put: #int
					unalignedLongAt: #sqInt unalignedLongAt:put: #sqInt
						unalignedLong32At: #int unalignedLong32At:put: #int

		 long64At: #sqLong long64At:put: #sqLong
		 long64AtPointer: #sqLong long64AtPointer:put: #sqLong
	
		singleFloatAtPointer: #float singleFloatAtPointerPut: #float		
		floatAtPointer: #double floatAtPointerPut: #double
				
		 fetchFloatAt:into: #void storeFloatAt:from: #void
			fetchFloatAtPointer:into: #void storeFloatAtPointer:from: #void
		 fetchSingleFloatAt:into: #void storeSingleFloatAt:from: #void
			fetchSingleFloatAtPointer:into: #void storeSingleFloatAtPointer:from: #void

		 pointerForOop: #'char *' oopForPointer: #sqInt
		 baseHeaderSize #sqInt wordSize #sqInt bytesPerOop #sqInt).
	BytesPerWord = 8 ifTrue:
		[#(long32At: long32At:put: unalignedLong32At: unalignedLong32At:put:) do:
			[:accessor|
			dictionary at: accessor put: #int]].
	^dictionary! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 2/13/2016 16:44'!
const: constName declareC: declarationString
	"Record the given C declaration for a constant."

	constants
		at: constName
		put: (TDefineNode new
				setName: constName
				value: declarationString)! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 10/29/2018 19:04'!
conventionalTypeForType: aCTypeString
	"The pointer type convention in this version of VMMaker is to have a space between the base type and any *'s.
	 C type comparisons are simple string comparisons; therefore the convention matters.
	 Ensure there is a space between the base type and any trailing *'s. Trim whitespace."
	| type index |
	type := aCTypeString withBlanksTrimmed.
	index := type size.
	[(type at: index) == $*] whileTrue:
		[index := index - 1].
	(index < type size
	 and: [(type at: index) ~~ Character space]) ifTrue:
		[type := (type copyFrom: 1 to: index), ' ', (type copyFrom: index + 1 to: type size)].
	^type! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 7/2/2008 16:01'!
currentMethod
	^currentMethod! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 12/16/2015 10:38'!
currentMethod: aTMethod
	previousCommenter := nil.
	currentMethod := aTMethod! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 2/11/2009 08:52'!
declareVar: varName type: type
	"This both creates a varable and provides its type"
	self var: (variables add: varName asString) type: type! !

!CCodeGenerator methodsFor: 'error notification' stamp: 'eem 7/6/2021 16:37'!
deny: aBooleanOrBlock
	aBooleanOrBlock value ifTrue: [AssertionFailure signal: 'Assertion failed']! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 7/31/2021 19:14'!
doBasicInlining: inlineFlagOrSymbol
	"Inline the bodies of all methods that are suitable for inlining.
	This method does only the basic inlining suitable for both the core VM and plugins - no bytecode inlining etc"

	| pass progress |
	beganInlining := true.
	self collectInlineList: inlineFlagOrSymbol.
	pass := 0.
	progress := true.
	[progress] whileTrue: [
		"repeatedly attempt to inline methods until no further progress is made"
		progress := false.
		('Inlining pass ', (pass := pass + 1) printString, '...')
			displayProgressAt: Sensor cursorPoint
			from: 0 to: methods size
			during: [:bar |
				(self sortMethods: methods) withIndexDo: [:m :i |
					bar value: i.
					currentMethod := m.
					(m tryToInlineMethodsIn: self)
						ifTrue: [progress := true]]]].

! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 7/31/2021 19:13'!
doInlining: inlineFlagOrSymbol
	"Inline the bodies of all methods that are suitable for inlining."
	"Modified slightly for the core VM translator, since the first level of inlining for the interpret loop must be performed in order that the instruction implementations can easily discover their addresses. Remember to inline the bytecode routines as well"

	| removed |
	beganInlining := true.
	inlineFlagOrSymbol isSymbol ifTrue:
		[self inlineDispatchesInMethodNamed: #interpret localizingVars: #().	
		 self doBasicInlining: inlineFlagOrSymbol.
		 self pruneUnreachableMethods.
		 ^self].

	inlineFlagOrSymbol ifFalse:
		[self inlineDispatchesInMethodNamed: #interpret localizingVars: #().	
		self pruneUnreachableMethods.
		^self].

	self doBasicInlining: inlineFlagOrSymbol.

	vmClass ifNil: [^self].

	'Inlining bytecodes'
		displayProgressAt: Sensor cursorPoint
		from: 1 to: 2
		during: [:bar |
			self inlineDispatchesInMethodNamed: #interpret
				localizingVars: vmClass namesOfVariablesToLocalize.
			bar value: 1.
			removed := self removeMethodsReferingToGlobals: vmClass namesOfVariablesToLocalize
							except: #interpret.
			bar value: 2].

	"only prune when generating the interpreter itself"
	self pruneUnreachableMethods.

	self reportShouldNotBeRemoved: removed  varList: vmClass namesOfVariablesToLocalize! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 7/7/2021 10:12'!
emitBuiltinConstructAsArgumentFor: msgNode on: aStream level: level
	"If the given selector is in the translation dictionary, translate it into a target code construct and return true. Otherwise, do nothing and return false."

	(self shouldGenerateAsInterpreterProxySend: msgNode) ifTrue:
		[^false].
	^(asArgumentTranslationDict
			at: msgNode selector
			ifAbsent: [translationDict at: msgNode selector ifAbsent: nil])
		ifNil: [false]
		ifNotNil: [:action|
				self perform: action with: msgNode with: aStream with: level.
				true]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 7/7/2021 10:05'!
emitBuiltinConstructFor: msgNode on: aStream level: level
	"If the given selector is in the translation dictionary, translate it into a target code construct and return true. Otherwise, do nothing and return false."

	(self shouldGenerateAsInterpreterProxySend: msgNode) ifTrue:
		[^false].
	^(translationDict at: msgNode selector ifAbsent: nil)
		ifNil: [false]
		ifNotNil: [:action|
				self perform: action with: msgNode with: aStream with: level.
				true]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 7/5/2021 10:50'!
emitCAPIExportHeaderOn: aStream 
	"Store prototype declarations for all API methods on the given stream."
	| exportedAPIMethods usedConstants |
	exportedAPIMethods := self sortMethods: (methods select: [:m| m isAPIMethod]).
	exportedAPIMethods do:
		[:m|
		m static ifTrue:
			[logger ensureCr; show: m selector, ' excluded from export API because it is static'; cr]].
	exportedAPIMethods := exportedAPIMethods reject: [:m| m static].
	self emitCFunctionPrototypes: exportedAPIMethods on: aStream.
	self emitGlobalCVariablesOn: aStream.
	usedConstants := self emitCMacros: exportedAPIMethods on: aStream.
	self emitCConstants: usedConstants on: aStream! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 2/19/2016 16:51'!
emitCCodeOn: aStream doInlining: inlineFlag doAssertions: assertionFlag
	"Emit C code for all methods in the code base onto the given stream. All inlined method calls should already have been expanded."

	| verbose methodList |
	"method preparation"
	verbose := false.
	self prepareMethods.
	verbose ifTrue: [
		self printUnboundCallWarnings.
		self printUnboundVariableReferenceWarnings.
		logger cr.
	].
	assertionFlag ifFalse: [ self removeAssertions ].
	self doInlining: inlineFlag.

	"code generation"
	"If we're outputting the VM put the main interpreter loop first for two reasons.
	 1, so that the dispdbg.h header included at the bytecode dispatch can define
	 macros that affect all C code in the interpreter,  and 2, so that all primitive
	 functions will come after and have relatively high addresses.  This provides safety
	 in the use of primitiveFunctionPointer as a function pointer and an index by trying
	 to ensure that primitives have addresses much higher than any indices."
	methodList := self sortMethods: methods.
	(methods includesKey: #interpret) ifTrue:
		[methodList := { methods at: #interpret }, (methodList copyWithout: (methods at: #interpret))].
	"clean out no longer valid variable names and then
	 handle any global variable usage in each method"
	methodList do: [:m | self checkForGlobalUsage: (m removeUnusedTempsAndNilIfRequiredIn: self) in: m].
	self localizeGlobalVariables.

	self emitCHeaderOn: aStream.
	self emitCConstantsOn: aStream.
	self emitCTypesOn: aStream. "Emitting types after constants allows use of constant names, e.g. in declaring arrays."
	self emitCFunctionPrototypes: methodList on: aStream.
	self emitCVariablesOn: aStream.
	self emitCMacros: methodList on: aStream.
	self emitCMethods: methodList on: aStream.
	self emitExportsOn: aStream.
! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 3/16/2018 23:42'!
emitCConstants: constList on: aStream
	"Store the global variable declarations on the given stream."
	constList isEmpty ifTrue: [^self].
	aStream cr; nextPutAll: '/*** Constants ***/'; cr.
	(self sortStrings: constList) do:
		[:varName| | node default value |
		node := constants at: varName.
		node name isEmpty ifFalse:
			["If the definition includes a C comment, or looks like a conditional, take it as is, otherwise convert
			  the value from Smalltalk to C.
			  Allow the class to provide an alternative definition, either of just the value or the whole shebang."
			default := (node value isString
						and: [(node value includesSubstring: '/*')
								or: [(node value includesSubstring: ' ? ') and: [node value includesSubstring: ' : ']]])
							ifTrue: [node value]
							ifFalse: [self cLiteralFor: node value name: varName].
			default = #undefined
				ifTrue: [aStream nextPutAll: '#undef '; nextPutAll: node name; cr]
				ifFalse:
					[value := vmClass
								ifNotNil:
									[(vmClass specialValueForConstant: node name default: default)
										ifNotNil: [:specialDef| specialDef]
										ifNil: [default]]
								ifNil: [default].
					((vmClass ifNil: VMBasicConstants) defineAtCompileTime: node name)
						ifTrue: [self putConditionalDefineOf: node name
									as: value
									comment: 'Allow this to be overridden on the compiler command line'
									on: aStream]
						ifFalse: [self putDefineOf: node name as: value on: aStream]]]].
	aStream cr! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 04/27/2017 12:55'!
emitCConstantsOn: aStream 
	"Store the global variable declarations on the given stream."
	| unused |
	unused := constants keys asSet.
	"Don't generate any defines for the externally defined constants,
	 STACKVM, COGVM, COGMTVM et al, unless they're actually used."
	(VMClass class>>#initializeMiscConstants) literalsDo:
		[:lit|
		(lit isVariableBinding and: [lit key isString]) ifTrue:
			[unused add: lit key]].
	methods do:
		[:meth|
		meth declarations keysDo:
			[:v|
			(meth typeFor: v in: self) ifNotNil:
				[:type| unused remove: type ifAbsent: []]].
		unused remove: meth returnType ifAbsent: [].
		meth parseTree nodesDo:
			[:n| n isConstant ifTrue: [unused remove: n name ifAbsent: []]]].
	unused copy do:
		[:const|
		(variableDeclarations anySatisfy: [:value| value includesSubstring: const]) ifTrue:
			[unused remove: const ifAbsent: []]].
	"and VMBasicConstants mostBasicConstantNames *must* be taken from interp.h"
	unused addAll: VMBasicConstants mostBasicConstantNames.
	self emitCConstants: (constants keys reject: [:any| unused includes: any]) on: aStream! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 5/3/2014 22:29'!
emitCExpression: aParseNode on: aStream 
	"Emit C code for the expression described by the given parse node."

	self emitCExpression: aParseNode on: aStream indent: 0! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 11/6/2014 14:35'!
emitCExpression: aParseNode on: aStream indent: level
	"Emit C code for the expression described by the given parse node."

	(aParseNode isLeaf
	 or: [aParseNode isSend
		 and: [#(bytesPerOop bytesPerWord baseHeaderSize wordSize) includes: aParseNode selector]])
		ifTrue: 
			["omit parens"
			 aParseNode emitCCodeAsExpressionOn: aStream level: level generator: self]
		ifFalse: 
			[aStream nextPut: $(.
			 aParseNode emitCCodeAsExpressionOn: aStream level: level generator: self.
			 aStream nextPut: $)]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 2/12/2016 16:52'!
emitCFunctionPrototypes: methodList on: aStream 
	"Store prototype declarations for all non-inlined methods on the given stream.
	 Add a define for a NoDbgRegParms attribute for static functions used for debugging.
	 gcc and other compilers will use non-standard calling conventions for static functions
	 when optimizing.  The optimization can render the functions unusable in gdb.  The sqConfig.h
	 file for the platform should define PlatformNoDbgRegParms suitably for the platform's
	 compiler, if the compiler can be persuaded not to generate such functions.
	 Add a define for a NeverInline attribute that tells the compiler never to inline functions
	 with the attribute.  We mark functions we want to observe in a profiler as NeverInline.
	 The sqConfig.h file for the platform should define NeverInline suitably for the platform's
	 compiler, if the compiler can be persuaded not to inline certain functions."
	aStream cr; nextPutAll: '/*** Function Prototypes ***/'; cr.
	vmClass ifNotNil:
		[NoRegParmsInAssertVMs ifTrue:
			[aStream nextPutAll: '\\#if !!PRODUCTION && defined(PlatformNoDbgRegParms)\# define NoDbgRegParms PlatformNoDbgRegParms\#endif' withCRs.
			 aStream nextPutAll: '\\#if !!defined(NoDbgRegParms)\# define NoDbgRegParms /*empty*/\#endif\\' withCRs].
		 aStream nextPutAll: '\\#if !!defined(NeverInline)\# define NeverInline /*empty*/\#endif\\' withCRs].
	(methodList select: [:m| m isRealMethod and: [self shouldGenerateMethod: m]]) do:
		[:m |
		vmClass ifNotNil:
			[(NoRegParmsInAssertVMs and: [m export not and: [m isStatic and: [m args notEmpty]]]) ifTrue:
				[m addFunctionAttribute: 'NoDbgRegParms'].
			 m inline == #never ifTrue:
				[m addFunctionAttribute: 'NeverInline']].
		m emitCFunctionPrototype: aStream generator: self].
	aStream cr! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'nice 4/23/2021 00:27'!
emitCHeaderOn: aStream
	"Emit the initial part of a source file on aStream, comprising the version stamp,
	 the global struct usage flags, the header files and preamble code."

	| headerClass |
	headerClass := [vmClass coreInterpreterClass]
						on: MessageNotUnderstood
						do: [:ex| vmClass].
	aStream nextPutAll: (self fileHeaderVersionStampForSourceClass: headerClass); cr; cr.
	self emitGlobalStructFlagOn: aStream.

	(vmClass isNil or: [vmClass wantsSqDotH]) ifTrue:
		[self addHeaderFileFirst: '"sq.h"'].
	"Additional header files; use the prescribed order."
	self emitHeaderFiles: headerFiles on: aStream.

	self maybePutPreambleFor: vmClass on: aStream.

	aStream cr! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 04/27/2017 12:55'!
emitCMacros: methodList on: aStream 
	"Store the global variable declarations on the given stream.  Answer any constants used in the macros."
	| usedConstants |
	macros isEmpty ifTrue: [^#()].
	aStream cr; nextPutAll: '/*** Macros ***/'; cr.
	usedConstants := Set new.
	(methodList reject: [:m| m isRealMethod]) do:
		[:m |
		m definedAsMacro ifTrue:
			[aStream
				nextPutAll: '#define ';
				nextPutAll:(self cFunctionNameFor: m selector);
				nextPutAll: (macros at: m selector); cr.
			 m compiledMethod literalsDo:
				[:lit|
				(lit isVariableBinding and: [(macros at: m selector) includesSubstring: lit key]) ifTrue:
					[usedConstants add: lit key]]]].
	aStream cr.
	^usedConstants! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 01/19/2021 20:40'!
emitCMethods: methodList on: aStream
	'Writing Translated Code...'
		displayProgressAt: Sensor cursorPoint
		from: 0 to: methods size
		during:
			[:bar |
			methodList withIndexDo:
				[ :m :i |
				bar value: i.
				(m isRealMethod
				 and: [self shouldGenerateMethod: m]) ifTrue:
					[m emitCCodeOn: aStream generator: self]]]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 6/23/2011 14:26'!
emitCTestBlock: aBlockNode on: aStream indent: level
	"Emit C code for the given block node to be used as a loop test."

	(aBlockNode statements size > 1
			ifTrue: [aBlockNode]
			ifFalse: [aBlockNode statements first])
		emitCCodeAsExpressionOn: aStream level: level generator: self! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 10/12/2013 14:00'!
emitCTypesOn: aStream 
	"Store local type declarations on the given stream."
	structClasses ifNotNil:
		[structClasses do:
			[:structClass|
			(structClass isAbstract not
			 and: [vmClass shouldGenerateTypedefFor: structClass]) ifTrue:
				[structClass printTypedefOn: aStream.
				 aStream cr; cr]]]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 2/3/2020 11:29'!
emitCVariablesOn: aStream 
	"Store the global variable declarations on the given stream."

	aStream cr; nextPutAll: '/*** Variables ***/'; cr.
	(self segregateByGroupingSizeAndVisibility: (self sortStrings: variables)) do:
		[:var | | varString |
		(self variableDeclarationStringsForVariable: (varString := var asString)) do:
			[:declArg| | decl |		
			(decl := declArg) first == $#
				ifTrue:
					[aStream nextPutAll: decl; cr]
				ifFalse:
					[self isGeneratingPluginCode
						ifTrue:
							[varString = 'interpreterProxy'
								ifTrue: "quite special..."
									[self preDeclareInterpreterProxyOn: aStream]
								ifFalse: [(decl beginsWith: 'static') ifFalse:
											[aStream nextPutAll: 'static ']]]
						ifFalse:
							[(vmClass mustBeGlobal: varString)
								ifTrue:
									[(decl beginsWith: 'static ') ifTrue:
										[decl := decl allButFirst: 7]]
								ifFalse:
									[(decl beginsWith: 'static') ifFalse:
										[aStream nextPutAll: 'static ']]].
					aStream
						nextPutAll: decl;
						nextPut: $;;
						cr]]].
	aStream cr! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 7/25/2021 18:08'!
emitExportsNamed: exportsNamePrefix pluginName: pluginName on: aStream
	"Store all the exported primitives in the form used by the internal named prim system."
	| nilVMClass excludeMetadata exportsNeedingMetadata |
	(nilVMClass := vmClass isNil) ifTrue: "We need a vmClass temporarily to compute accessor depths."
		[vmClass := StackInterpreter].
	"Don't include the depth in the vm's named primitives if the vm is non-Spur."
	excludeMetadata := exportsNamePrefix = 'vm'
					  and: [pluginName isEmpty
					  and: [vmClass objectMemoryClass hasSpurMemoryManagerAPI not]].
	aStream cr; cr; nextPutAll: 'static char _m[] = "'; nextPutAll: pluginName; nextPutAll: '";'.
	aStream cr; nextPutAll: 'void* '; nextPutAll: exportsNamePrefix; nextPutAll: '_exports[][3] = {'; cr.
	exportsNeedingMetadata := self exportsNeedingMetadata.
	self sortedExportMethods do:
		[:method|
		self withOptionalConditionalDefineFor: method
			on: aStream
			do: [| primName depth exportFlags |
				 primName := self cFunctionNameFor: method selector.
				 aStream tab; nextPutAll: '{(void*)_m, "'; nextPutAll: primName.
				 (excludeMetadata not and: [exportsNeedingMetadata includes: method]) ifTrue:
					[depth := self accessorDepthForMethod: method.
					 exportFlags := CoInterpreter metadataFlagsForPrimitive: method compiledMethod.
					 (depth notNil or: [exportFlags > 0]) ifTrue:
						"store the metadata in two hidden bytes immediately after the primName"
						[self assert: depth < 128.
						 aStream
							nextPutAll: '\000\';
							nextPutAll: ((depth bitAnd: 255) printStringBase: 8 nDigits: 3);
							nextPut: $\;
							nextPutAll: ((exportFlags bitAnd: 255) printStringBase: 8 nDigits: 3)]].
				 aStream nextPutAll: '", (void*)'; nextPutAll: primName; nextPutAll: '},'; cr]].
	aStream tab; nextPutAll: '{NULL, NULL, NULL}'; cr; nextPutAll: '};'; cr.
	nilVMClass ifTrue:
		[vmClass := nil]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 12/29/2013 15:35'!
emitExportsOn: aStream
	"Store all the exported primitives in the form used by the internal named prim system."
	(vmClass isNil or: [vmClass isInterpreterClass]) ifTrue:
		[self emitExportsNamed: 'vm' pluginName: '' on: aStream]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 8/27/2020 19:18'!
emitGlobalCVariablesOn: aStream
	"Store the global (exported) variable declarations on the given stream."

	aStream cr; nextPutAll: '/*** Global Variables ***/'; cr.
	
	(self sortStrings: (variables select: [:v| vmClass mustBeGlobal: v])) do:
		[:var | | varString |
		(self variableDeclarationStringsForVariable: (varString := var asString)) do:
			[:decl|
			decl first == $#
				ifTrue:
					[aStream nextPutAll: decl; cr]
				ifFalse:
					[(decl includesSubstring: ' private ') ifFalse: "work-around hack to prevent localization of variables only referenced once."
						[(decl beginsWith: 'static') ifFalse:
							[aStream
								nextPutAll: 'extern ';
								nextPutAll:
									((decl includes: $=)
										ifTrue: [decl copyFrom: 1 to: (decl indexOf: $=) - 1]
										ifFalse: [decl]);
								nextPut: $;;
								cr]]]]].
	aStream cr! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'ikp 9/10/2003 05:53'!
emitGlobalStructFlagOn: aStream
	"Default: do nothing.  Overridden in CCGenGlobalStruct."
! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 10/12/2010 14:05'!
emitHeaderFiles: headerFilesSubSequence on: aStream
	headerFilesSubSequence do:
		[:hdr|
		hdr first ~= $# ifTrue:
			[aStream nextPutAll: '#include '].
		aStream nextPutAll: hdr; cr]! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 2/13/2016 12:49'!
exportedPrimitiveNames
	"Return an array of all exported primitives"
	^methods select:[:m| m export] thenCollect:[:m| self cFunctionNameFor: m selector].
! !

!CCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 7/25/2021 19:05'!
exportsNeedingMetadata
	^self sortedExportMethods reject:
		[:m| | selector |
		selector := m smalltalkSelector.
		selector = #initialiseModule
		or: [(InterpreterPlugin includesSelector: selector)
		or: [vmClass notNil and: [(selector beginsWith: 'prim') not]]]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 5/31/2018 11:17'!
extractTypeFor: aVariable fromDeclaration: aVariableDeclaration
	"Eliminate inessentials from aVariableDeclaration to answer a C type without the variable,
	 or initializations etc"
	| decl |
	((aVariableDeclaration beginsWith: 'static ')
	 or: [aVariableDeclaration beginsWith: 'extern ']) ifTrue:
		[^self extractTypeFor: aVariable fromDeclaration: (aVariableDeclaration allButFirst: 7)].
	decl := (aVariableDeclaration indexOf: $= ifAbsent: [])
				ifNotNil: [:index| aVariableDeclaration copyFrom: 1 to: index - 1]
				ifNil: [aVariableDeclaration].
	decl := decl copyReplaceAll: aVariable with: '' tokenish: [:ch| ch = $_ or: [ch isAlphaNumeric]].
	^self baseTypeForType: decl! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 1/3/2013 13:47'!
fileHeaderVersionStampForSourceClass: sourceClass
	"Answer a suitable version stamp to include in the header."
	| exportBuildInfo slangDescription sourceDescription |
	[exportBuildInfo := sourceClass isInterpreterClass
						ifTrue: ['char *__interpBuildInfo = __buildInfo;']
						ifFalse:
							[sourceClass isCogitClass
								ifTrue: ['char *__cogitBuildInfo = __buildInfo;']
								ifFalse:
									[sourceClass isPluginClass ifTrue:
										[sourceClass exportBuildInfoOrNil]]]]
		on: MessageNotUnderstood
		do: [:ex| ex resume: false].
	[slangDescription := self class monticelloDescriptionFor: self class.
	 sourceClass ifNotNil:
		[sourceDescription := [sourceClass monticelloDescription]
								on: MessageNotUnderstood
								do: [:ex| self class monticelloDescriptionFor: sourceClass]]]
		on: Error
		do: [:ex| | now |
			now := Time dateAndTimeNow printString.
			^String streamContents:
				[:s|
				s nextPutAll: '/* Automatically generated from Squeak on '.
				s nextPutAll: now.
				s nextPutAll: ' */'; cr; cr.
				s nextPutAll: 'static char __buildInfo[] = "Generated on '.
				s nextPutAll: now.
				s nextPutAll: '. Compiled on "'.
				s nextPutAll: '__DATE__ ;'; cr.
				exportBuildInfo ifNotNil:
					[s nextPutAll: exportBuildInfo; cr].
				s cr]].
	^String streamContents:
		[:s|
		s nextPutAll: '/* Automatically generated by\	' withCRs.
		s nextPutAll: slangDescription.
		sourceDescription ifNotNil:
			[s nextPutAll: '\   from\	' withCRs; nextPutAll: (sourceDescription copyReplaceAll: '\' withCRs with: '\	' withCRs)].
		s cr; nextPutAll: ' */'; cr.
		sourceDescription ifNotNil:
			[s nextPutAll: 'static char __buildInfo[] = "'.
			 s nextPutAll: (sourceDescription copyReplaceAll: '\' withCRs with: '\n\', '\' withCRs).
			 s nextPutAll: ' " __DATE__ ;'; cr.
			exportBuildInfo ifNotNil:
				[s nextPutAll: exportBuildInfo; cr].
			s cr]]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 1/1/1970 14:18'!
generateAbs: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	| type strippedType absFun |
	type := self typeFor: msgNode receiver in: currentMethod.
	strippedType := (type beginsWith: 'unsigned ')
		ifTrue: [type allButFirst: 'unsigned ' size]
		ifFalse: [(type beginsWith: 'u')
			ifTrue: [type allButFirst: 1]
			ifFalse: [type]].
	absFun :=
		(#(	('long long' llabs) ('__int64' #llabs) ('int64_t' #llabs)
			(sqInt SQABS)
			(sqLong SQLABS)
			(double fabs)
			(float fabsf))
				detect: [:tuple| strippedType beginsWith: tuple first]
				ifNone: [])
			ifNotNil: [:tuple| tuple last]
			ifNil: [#abs].
	
	aStream nextPutAll: absFun.
	aStream nextPut: $(.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/5/2009 13:48'!
generateAddressOf: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPut: $(; nextPut: $&.
	self emitCExpression: msgNode args first on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 1/25/2016 16:05'!
generateAllMask: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '(('.
	self generateBitAnd: msgNode on: aStream indent: level.
	aStream nextPutAll: ') == '.
	self emitCExpression: msgNode args first on: aStream indent: level.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/16/2015 13:37'!
generateAnd: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode receiver on: aStream indent: level.
	aStream nextPutAll: ' && '.
	self emitCExpression: msgNode args first on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 7/7/2021 21:49'!
generateAnyMask: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '(('.
	self generateBitAnd: msgNode on: aStream indent: level.
	aStream nextPutAll: ') !!= 0)'! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 8/4/2014 10:22'!
generateAsAddress: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode args first on: aStream! !

!CCodeGenerator methodsFor: 'C translation support' stamp: 'eem 1/29/2020 15:46'!
generateAsConstantExpression: msgNode on: aStream
	"Attempt to generate the C code for this message onto the given stream.
	 Answer if the attempt succeeded.  This handles integer overflow of int
	 expressions in a 64-bit compiler."

	msgNode constantNumbericValueOrNil ifNotNil:
		[:value|
		 ((vmClass notNil and: [vmClass objectMemoryClass wordSize = 8])
			ifTrue: [value between: -1 << 63 and: 1 << 64 - 1]
			ifFalse: [value between: -1 << 31 and: 1 << 32 - 1]) ifTrue:
				[aStream nextPutAll: (self cLiteralFor: value).
				 ^true]].
	^false! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'ar 10/3/1998 13:45'!
generateAsFloat: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll:'((double) '.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ' )'.! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 2/15/2009 16:38'!
generateAsInteger: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll:'((sqInt)'.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'nice 7/8/2016 02:23'!
generateAsIntegerPtr: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll:'((sqIntptr_t)'.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 2/20/2013 12:28'!
generateAsLong: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll:'((long)'.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 1/3/2009 10:22'!
generateAsSymbol: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream.
	 The receiver is expected to be a TConstantNode."

	aStream nextPutAll: (self cFunctionNameFor: msgNode receiver nameOrValue)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 2/15/2009 16:38'!
generateAsUnsignedInteger: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll:'((usqInt)'.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'nice 7/8/2016 02:24'!
generateAsUnsignedIntegerPtr: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll:'((usqIntptr_t)'.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 11/26/2012 14:23'!
generateAsUnsignedLong: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll:'((unsigned long)'.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 2/20/2016 23:16'!
generateAsUnsignedLongLong: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll:'((unsigned long long)'.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/31/2013 08:44'!
generateAsVoidPointer: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '((void *)'.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/27/2020 21:20'!
generateAt: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(msgNode args first isConstant
	 and: [msgNode args first value isZero
	 and: [msgNode receiver isSend
	 and: [(#(cCoerceSimple:to: cCoerce:to:) includes: msgNode receiver selector)
	 and: [msgNode receiver args last value last == $*]]]])
		ifTrue:
			[aStream nextPutAll: '(*'.
			 msgNode receiver emitCCodeAsExpressionOn: aStream level: 0 generator: self.
			 aStream nextPut: $)]
		ifFalse:
			[self emitCExpression: msgNode receiver on: aStream.
			 aStream nextPut: $[.
			 msgNode args first emitCCodeAsExpressionOn: aStream level: level + 1 generator: self.
			 aStream nextPut: $]]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/27/2020 21:19'!
generateAtPut: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(msgNode args first isConstant
	 and: [msgNode args first value isZero
	 and: [msgNode receiver isSend
	 and: [(#(cCoerceSimple:to: cCoerce:to:) includes: msgNode receiver selector)
	 and: [msgNode receiver args last value last == $*]]]])
		ifTrue:
			[aStream nextPutAll: '(*'.
			 msgNode receiver emitCCodeAsExpressionOn: aStream level: 0 generator: self.
			 aStream nextPutAll: ' = '.
			 msgNode args last emitCCodeAsExpressionOn: aStream level: 0 generator: self.
			 aStream nextPut: $)]
		ifFalse:
			[self emitCExpression: msgNode receiver on: aStream.
			 aStream nextPut: $[.
			 msgNode args first emitCCodeAsExpressionOn: aStream level: level + 1 generator: self.
			 aStream nextPutAll: '] = '.
			 self emitCExpression: msgNode args last on: aStream]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'EstebanLorenzano 9/19/2011 16:33'!
generateBaseHeaderSize: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: 'BaseHeaderSize'
! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 5/31/2020 15:20'!
generateBetweenAnd: msgNode on: aStream indent: level
	"Generate the C code for the between:and: message onto the given stream."

	aStream nextPutAll: '(('.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ' >= '.
	self emitCExpression: msgNode args first on: aStream.
	aStream nextPutAll: ') && ('.
	self emitCExpression: (msgNode receiver isAssignment
							ifTrue: [msgNode receiver variable]
							ifFalse: [msgNode receiver])
		on: aStream.
	aStream nextPutAll: ' <= '.
	self emitCExpression: msgNode args second on: aStream.
	aStream nextPutAll: '))'! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/16/2015 13:37'!
generateBitAnd: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode receiver on: aStream indent: level.
	aStream nextPutAll: ' & '.
	self emitCExpression: msgNode args first on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 2/22/2011 11:42'!
generateBitClear: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream next: 2 put: $(.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ' | '.
	self emitCExpression: msgNode args first on: aStream.
	aStream nextPutAll: ') - '.
	self emitCExpression: msgNode args first on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 8/1/2021 14:54'!
generateBitInvert: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream.
	 If the selector is bitInvert32 then cast to unsigned int to ensure
	 a 32-bit value on 64-bit platforms."

	| castToUnsignedInt castToUnsignedLong castToUnsignedLongLong |
	msgNode selector == #bitInvert ifTrue: "quick hack; can do the full job (of coercing -1 to the right type) later..."
		[aStream nextPutAll: '(-1 - ('.
		 self emitCExpression: msgNode receiver on: aStream.
		 aStream next: 2 put: $).
		 ^self].
	(castToUnsignedInt := msgNode selector = #bitInvert32) ifTrue:
		[aStream nextPutAll: '(unsigned int)'].
	aStream nextPut: $~.
	(castToUnsignedLong := msgNode selector = #bitInvert64 and: [vmClass notNil and: [vmClass objectMemoryClass wordSize = 8]]) ifTrue:
		[aStream nextPutAll: '(usqIntptr_t)'].
	(castToUnsignedLongLong := msgNode selector = #bitInvert64 and: [vmClass isNil or: [vmClass objectMemoryClass wordSize = 4]]) ifTrue:
		[aStream nextPutAll: '(unsigned long long)'].
	self assert: castToUnsignedInt asBit + castToUnsignedLong asBit + castToUnsignedLongLong asBit <= 1. "We should only do a single cast"
	self emitCExpression: msgNode receiver on: aStream! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/16/2015 13:38'!
generateBitOr: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode receiver on: aStream indent: level.
	aStream nextPutAll: ' | '.
	self emitCExpression: msgNode args first on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'nice 3/9/2020 23:49'!
generateBitShift: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	| arg shift rightShift |
	(self isConstantNode: (arg := msgNode args first) valueInto: [:shiftValue| shift := shiftValue])
		ifTrue: "bit shift amount is a constant"
			[aStream nextPut: $(.
			shift < 0
				ifTrue:
					[rightShift := TSendNode new
						setSelector: #>>
						receiver: msgNode receiver
						arguments: {TConstantNode new setValue: shift negated}.
					self generateShiftRight: rightShift on: aStream indent: level]
				ifFalse: [self generateShiftLeft: msgNode on: aStream indent: level].
			aStream nextPut: $)]
		ifFalse: "bit shift amount is an expression"
			[rightShift := TSendNode new
						setSelector: #>>
						receiver: msgNode receiver
						arguments: {TSendNode new
							setSelector: #negated
							receiver: arg
							arguments: #()}.
			aStream nextPutAll: '(('.
			self emitCExpression: arg on: aStream.
			aStream nextPutAll: ' < 0) ? ('.
			self generateShiftRight: rightShift on: aStream indent: level. 
			aStream nextPutAll: ') : ('.
			self generateShiftLeft: msgNode on: aStream indent: level.
			aStream nextPutAll: '))']! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/16/2015 13:38'!
generateBitXor: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode receiver on: aStream indent: level.
	aStream nextPutAll: ' ^ '.
	self emitCExpression: msgNode args first on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'nice 4/2/2016 23:16'!
generateByteSwap32: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: #'SQ_SWAP_4_BYTES'.
	aStream nextPut: $(.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'nice 4/3/2016 09:21'!
generateByteSwap32IfBigEndian: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: #'SQ_SWAP_4_BYTES_IF_BIGENDIAN'.
	aStream nextPut: $(.
	self emitCExpression: msgNode args first on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'nice 4/2/2016 23:16'!
generateByteSwap64: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: #'SQ_SWAP_8_BYTES'.
	aStream nextPut: $(.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'nice 4/3/2016 09:21'!
generateByteSwap64IfBigEndian: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: #'SQ_SWAP_8_BYTES_IF_BIGENDIAN'.
	aStream nextPut: $(.
	self emitCExpression: msgNode args first on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/17/2014 12:15'!
generateBytesPerOop: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream.
	 Would prefer to use OopSize than BytesPerOop (shorter)."

	aStream nextPutAll: 'BytesPerOop'! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 9/10/2013 14:37'!
generateBytesPerWord: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream.
	 Would prefer to use WordSize than BytesPerWord (shorter) but e.g.
	 mpegtype.h has a parameter named WordSize."

	aStream nextPutAll: 'BytesPerWord'! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/22/2018 15:20'!
generateCCoercion: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	| cExpr cType literal |
	cExpr := msgNode args first.
	cType := msgNode args last value.
	(cType = #float
	 and: [cExpr isConstant
	 and: [cExpr value isFloat
	 and: [(literal := self cLiteralFor: cExpr value) allSatisfy: [:c| c == $. or: [c isDigit]]]]]) ifTrue:
		[aStream nextPutAll: literal; nextPut: $f.
		 ^self].
	aStream nextPutAll: '(('.
	aStream nextPutAll: cType.
	aStream nextPutAll: ') '.
	self emitCExpression: cExpr on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 11/14/2019 11:10'!
generateComma: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream.
	 Assume that the arguments are literal strings"
	self emitCExpression: msgNode receiver on: aStream.
	aStream space.
	self emitCExpression: msgNode args first on: aStream! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'ar 7/8/2003 11:26'!
generateDeadCode
	"Answer whether we should generate 'dead code' branches. This can be useful for hacking the VM when used in conjunction with #useSymbolicConstants, e.g., for code like:
		DoAssertionChecks ifTrue:[
			...
		].

	we will generate

		#define DoAssertionChecks 0
		...
		if(DoAssertionChecks) {
			...
		}.

	allowing us to change the #define (or redefine it as a variable) for later use."
	^generateDeadCode! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'ar 7/8/2003 11:26'!
generateDeadCode: aBool
	"Indicate whether we should generate 'dead code' branches."
	generateDeadCode := aBool! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 7/19/2014 11:42'!
generateDeny: denyMsgNode on: aStream indent: level
	"Generate the C code for a deny: expr onto the given stream."

	aStream nextPutAll: 'assert(!!('.
	self emitCExpression: denyMsgNode args first on: aStream.
	aStream nextPutAll: '))'! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 7/5/2014 18:59'!
generateDivide: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	| rcvr arg divisor check |
	rcvr := msgNode receiver.
	arg := msgNode args first.
	check :=	false "If you need to check the validity of divides that had been implemented by signed shifts, change this to true..."
				and: [arg isConstant
				and: [(divisor := arg value) isInteger
				and: [divisor isPowerOfTwo
				and: [divisor > 0
				and: [divisor <= (1 bitShift: 31)]]]]].
	check ifTrue:
		[(headerFiles includes: '"sqAssert.h"') ifFalse:
			[self addHeaderFile: '"sqAssert.h"'].
		aStream nextPut: $(; nextPutAll: 'assert(((sqInt) '.
		 self emitCExpression: rcvr on: aStream.
		 aStream nextPutAll: ' >> ', (divisor log: 2) asInteger printString.
		 aStream nextPutAll: ') == ('.
		 self emitCExpression: rcvr on: aStream.
		aStream nextPutAll: ' / '.
		self emitCExpression: msgNode args first on: aStream.
		aStream nextPutAll: ')),'].
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ' / '.
	self emitCExpression: msgNode args first on: aStream.
	check ifTrue:
		[aStream nextPut: $)]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 5/2/2014 14:29'!
generateDoWhileFalse: msgNode on: aStream indent: level
	"Generate do {stmtList} while(!!(cond))"

	| testStmt receiverWithoutTest |
	testStmt := msgNode receiver statements last.
	receiverWithoutTest := TStmtListNode new setStatements: msgNode receiver statements allButLast.
	aStream nextPutAll: 'do {'; cr.
	receiverWithoutTest emitCCodeOn: aStream level: level + 1 generator: self.
	aStream tab: level.
	aStream nextPutAll: '} while(!!('.
	testStmt emitCCodeAsExpressionOn: aStream level: level + 1 generator: self.
	aStream nextPutAll: '))'! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 5/2/2014 14:29'!
generateDoWhileTrue: msgNode on: aStream indent: level
	"Generate do {stmtList} while(cond)"

	| testStmt receiverWithoutTest |
	testStmt := msgNode receiver statements last.
	receiverWithoutTest := TStmtListNode new setStatements: msgNode receiver statements allButLast.
	aStream nextPutAll: 'do {'; cr.
	receiverWithoutTest emitCCodeOn: aStream level: level + 1 generator: self.
	aStream tab: level.
	aStream nextPutAll: '} while('.
	testStmt emitCCodeAsExpressionOn: aStream level: level + 1 generator: self.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/16/2015 13:39'!
generateEqual: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode receiver on: aStream indent: level.
	aStream nextPutAll: ' == '.
	self emitCExpression: msgNode args first on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 2/25/2020 18:11'!
generateEven: msgNode on: aStream indent: level
	"Generate the C code for even or odd onto the given stream."

	aStream nextPutAll: '(!!('.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ' & 1))'! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/12/2012 11:15'!
generateFlag: msgNode on: aStream indent: level
	"Compoensate for the use of self flag: #aSymbol.  We used to translate  symbols
	 as strings unless they were quoted via #aSymbol asSymbol. But this is too tedious,
	 so we now translate symbols directly.  The only use that this affected was in
		 self flag: #aSymbol,
	 so hard-code it to produce a string value.  Note that this isn't strictly necessary
	 because there's a
		#define flag(foo) 0
	 in C land, but it makes the generated C less dissonant."

	aStream
		nextPutAll: 'flag(';
		nextPutAll: (self cLiteralFor: msgNode args last value asString);
		nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/16/2015 13:39'!
generateGreaterThan: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode receiver on: aStream indent: level.
	aStream nextPutAll: ' > '.
	self emitCExpression: msgNode args first on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/16/2015 13:39'!
generateGreaterThanOrEqual: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode receiver on: aStream indent: level.
	aStream nextPutAll: ' >= '.
	self emitCExpression: msgNode args first on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 8/13/2014 06:18'!
generateIfFalse: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self nilOrBooleanConstantReceiverOf: msgNode receiver)
		ifNil:
			["Eliminate double-negatives"
			 (msgNode receiver isSend and: [msgNode receiver selector == #not]) ifTrue:
				[^self generateIfTrue: (TSendNode new
											setSelector: msgNode selector
											receiver: msgNode receiver receiver
											arguments: msgNode args
											isBuiltInOp: msgNode isBuiltinOperator)
					on: aStream
					indent: level].
			aStream nextPutAll: 'if (!!'.
			msgNode receiver isLeaf ifFalse: [aStream nextPut: $(]. "grrrrr"
			msgNode receiver emitCCodeAsExpressionOn: aStream level: level + 1 generator: self.
			msgNode receiver isLeaf ifFalse: [aStream nextPut: $)]. "grrrrr"
			aStream nextPutAll: ') {'; cr.
			msgNode args last emitCCodeOn: aStream level: level + 1 generator: self.
			aStream tab: level.
			aStream nextPut: $}]
		ifNotNil:
			[:const |
			const ifFalse:
				[msgNode args first emitCCodeOn: aStream level: level generator: self]]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/12/2018 12:07'!
generateIfFalseAsArgument: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self nilOrBooleanConstantReceiverOf: msgNode receiver)
		ifNil:
			["Eliminate double-negatives"
			 (msgNode receiver isSend and: [msgNode receiver selector == #not]) ifTrue:
				[^self generateIfTrueAsArgument: (TSendNode new
														setSelector: msgNode selector
														receiver: msgNode receiver receiver
														arguments: msgNode args
														isBuiltInOp: msgNode isBuiltinOperator)
					on: aStream
					indent: level].
			aStream nextPutAll: '(!!('.
			msgNode receiver emitCCodeAsArgumentOn: aStream level: 0 generator: self.
			aStream nextPut: $); crtab: level + 1; nextPut: $?; space.
			msgNode args last emitCCodeAsArgumentOn: aStream level: level + 2 generator: self.
			aStream crtab: level + 1; nextPutAll: ': 0)']
		ifNotNil:
			[:const|
			const
				ifFalse: [msgNode args first emitCCodeAsArgumentOn: aStream level: level generator: self]
				ifTrue: [aStream nextPutAll: (self cLiteralFor: nil)]]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 11/17/2015 14:47'!
generateIfFalseIfTrue: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self nilOrBooleanConstantReceiverOf: msgNode receiver)
		ifNil:
			[(self tryToCollapseBothArmsOfConditional: msgNode on: aStream indent: level) ifFalse:
				[aStream nextPutAll: 'if ('.
				 msgNode receiver emitCCodeAsExpressionOn: aStream level: level generator: self.
				 aStream nextPutAll: ') {'; cr.
				 msgNode args last emitCCodeOn: aStream level: level + 1 generator: self.
				 aStream tab: level; nextPut: $}; crtab: level; nextPutAll: 'else {'; cr.
				 msgNode args first emitCCodeOn: aStream level: level + 1 generator: self.
				 aStream tab: level; nextPut: $}]]
		ifNotNil:
			[:const |
			 (const ifTrue: [msgNode args last] ifFalse: [msgNode args first])
				emitCCodeOn: aStream level: level generator: self]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 11/17/2015 14:51'!
generateIfFalseIfTrueAsArgument: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self nilOrBooleanConstantReceiverOf: msgNode receiver)
		ifNil:
			[(self tryToCollapseBothArmsOfConditionalExpression: msgNode on: aStream indent: level) ifFalse:
				[aStream nextPut: $(.
				 msgNode receiver emitCCodeAsArgumentOn: aStream level: level generator: self.
				 aStream crtab: level + 1; nextPut: $?; space.
				 msgNode args last emitCCodeAsArgumentOn: aStream level: level + 2 generator: self.
				 aStream crtab: level + 1; nextPut: $:; space.
				 msgNode args first emitCCodeAsArgumentOn: aStream level: level + 2 generator: self.
				 aStream nextPut: $)]]
		ifNotNil:
			[:const|
			(const
				ifTrue: [msgNode args last]
				ifFalse: [msgNode args first])
					emitCCodeAsArgumentOn: aStream level: level generator: self]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 7/16/2020 15:46'!
generateIfNil: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self isNilConstantReceiverOf: msgNode)
		ifFalse:
			[aStream nextPutAll: 'if (!!'.
			 self emitCExpression: msgNode receiver on: aStream indent: level + 1.
			 aStream nextPutAll: ') {'; cr.
			 msgNode args last emitCCodeOn: aStream level: level + 1 generator: self.
			 aStream tab: level; nextPut: $}]
		ifTrue:
			[msgNode args first emitCCodeOn: aStream level: level generator: self]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 9/1/2017 15:11'!
generateIfNilAsArgument: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self isNilConstantReceiverOf: msgNode)
		ifFalse:
			[aStream nextPutAll: '(!!('.
			 msgNode receiver emitCCodeAsArgumentOn: aStream level: 0 generator: self.
			 aStream nextPut: $); crtab: level + 1; nextPut: $?; space.
			 msgNode args last emitCCodeAsArgumentOn: aStream level: level + 2 generator: self.
			 aStream crtab: level + 1; nextPut: $:; space.
			 msgNode receiver emitCCodeAsArgumentOn: aStream level: 0 generator: self.
			 (msgNode receiver isLeaf
			  or: [msgNode receiver isSend
				  and: [(self isStructSend: msgNode receiver)
				  and: [msgNode receiver receiver isLeaf]]]) ifFalse:
				[logger cr; nextPutAll: 'sending ifNil: to non-leaf in '; nextPutAll: currentMethod selector].
			 aStream nextPut: $)]
		ifTrue:
			[msgNode args first emitCCodeAsArgumentOn: aStream level: level generator: self]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/27/2010 12:27'!
generateIfNilIfNotNil: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self isNilConstantReceiverOf: msgNode)
		ifFalse:
			[aStream nextPutAll: 'if ('.
			msgNode receiver emitCCodeAsExpressionOn: aStream level: level generator: self.
			aStream nextPutAll: ') {'; cr.
			msgNode args last emitCCodeOn: aStream level: level + 1 generator: self.
			aStream tab: level; nextPut: $}; cr; tab: level; nextPutAll: 'else {'; cr.
			msgNode args first emitCCodeOn: aStream level: level + 1 generator: self.
			aStream tab: level; nextPut: $}]
		ifTrue:
			[msgNode args first emitCCodeOn: aStream level: level generator: self]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/27/2010 12:31'!
generateIfNilIfNotNilAsArgument: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self isNilConstantReceiverOf: msgNode)
		ifFalse:
			[aStream nextPut: $(.
			msgNode receiver emitCCodeAsArgumentOn: aStream level: level generator: self.
			aStream crtab: level + 1; nextPut: $?; space.
			msgNode args last emitCCodeAsArgumentOn: aStream level: level + 2 generator: self.
			aStream crtab: level + 1; nextPut: $:; space.
			msgNode args first emitCCodeAsArgumentOn: aStream level: level + 2 generator: self.
			aStream nextPut: $)]
		ifTrue:
			[msgNode args first emitCCodeAsArgumentOn: aStream level: level generator: self]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 7/6/2021 19:10'!
generateIfNotNil: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self isNilConstantReceiverOf: msgNode)
		ifFalse:
			[aStream nextPutAll: 'if ('.
			msgNode receiver emitCCodeAsExpressionOn: aStream level: level generator: self.
			aStream nextPutAll: ') {'; cr.
			msgNode args first emitCCodeOn: aStream level: level + 1 generator: self.
			aStream tab: level; nextPut: $}]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/12/2018 12:11'!
generateIfNotNilAsArgument: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self isNilConstantReceiverOf: msgNode)
		ifFalse:
			[aStream nextPut: $(.
			 msgNode receiver emitCCodeAsArgumentOn: aStream level: level generator: self.
			 aStream crtab: level + 1; nextPut: $?; space.
			 msgNode args first emitCCodeAsArgumentOn: aStream level: level + 2 generator: self.
			 aStream crtab: level + 1; nextPutAll: ': 0)']
		ifTrue:
			[aStream nextPutAll: (self cLiteralFor: nil)]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/27/2010 12:28'!
generateIfNotNilIfNil: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self isNilConstantReceiverOf: msgNode)
		ifFalse:
			[aStream nextPutAll: 'if ('.
			msgNode receiver emitCCodeAsExpressionOn: aStream level: level generator: self.
			aStream nextPutAll: ') {'; cr.
			msgNode args first emitCCodeOn: aStream level: level + 1 generator: self.
			aStream tab: level; nextPut: $}; cr; tab: level; nextPutAll: 'else {'; cr.
			msgNode args last emitCCodeOn: aStream level: level + 1 generator: self.
			aStream tab: level; nextPut: $}]
		ifTrue:
			[msgNode args last emitCCodeOn: aStream level: level generator: self]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/27/2010 12:32'!
generateIfNotNilIfNilAsArgument: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self isNilConstantReceiverOf: msgNode)
		ifFalse:
			[aStream nextPut: $(.
			msgNode receiver emitCCodeAsExpressionOn: aStream level: level generator: self.
			aStream crtab: level + 1; nextPut: $?; space.
			msgNode args first emitCCodeAsArgumentOn: aStream level: level + 2 generator: self.
			aStream crtab: level + 1; nextPut: $:; space.
			msgNode args last emitCCodeAsArgumentOn: aStream level: level + 2 generator: self.
			aStream nextPut: $)]
		ifTrue:
			[msgNode args last emitCCodeAsArgumentOn: aStream level: level generator: self]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 11/10/2013 16:01'!
generateIfTrue: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self nilOrBooleanConstantReceiverOf: msgNode receiver)
		ifNil:
			[aStream nextPutAll: 'if ('.
			msgNode receiver emitCCodeAsExpressionOn: aStream level: level generator: self.
			aStream nextPutAll: ') {'; cr.
			msgNode args first emitCCodeOn: aStream level: level + 1 generator: self.
			aStream tab: level.
			aStream nextPut: $}]
		ifNotNil:
			[:const |
			const ifTrue:
				[msgNode args first emitCCodeOn: aStream level: level generator: self]]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/12/2018 12:06'!
generateIfTrueAsArgument: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self nilOrBooleanConstantReceiverOf: msgNode receiver)
		ifNil:
			[aStream nextPut: $(.
			 msgNode receiver emitCCodeAsArgumentOn: aStream level: level generator: self.
			 aStream crtab: level + 1; nextPut: $?; space.
			 msgNode args first emitCCodeAsArgumentOn: aStream level: level + 2 generator: self.
			 aStream crtab: level + 1; nextPutAll: ': 0)']
		ifNotNil:
			[:const|
			const
				ifTrue: [msgNode args first emitCCodeAsArgumentOn: aStream level: level generator: self]
				ifFalse: [aStream nextPutAll: (self cLiteralFor: nil)]]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 11/17/2015 14:47'!
generateIfTrueIfFalse: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self nilOrBooleanConstantReceiverOf: msgNode receiver)
		ifNil:
			[(self tryToCollapseBothArmsOfConditional: msgNode on: aStream indent: level) ifFalse:
				[aStream nextPutAll: 'if ('.
				 msgNode receiver emitCCodeAsExpressionOn: aStream level: level generator: self.
				 aStream nextPutAll: ') {'; cr.
				 msgNode args first emitCCodeOn: aStream level: level + 1 generator: self.
				 aStream tab: level; nextPut: $}; crtab: level; nextPutAll: 'else {'; cr.
				 msgNode args last emitCCodeOn: aStream level: level + 1 generator: self.
				 aStream tab: level; nextPut: $}]]
		ifNotNil:
			[:const |
			(const ifTrue: [msgNode args first] ifFalse: [msgNode args last])
				emitCCodeOn: aStream level: level generator: self]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/12/2018 12:09'!
generateIfTrueIfFalseAsArgument: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self nilOrBooleanConstantReceiverOf: msgNode receiver)
		ifNil:
			[(self tryToCollapseBothArmsOfConditionalExpression: msgNode on: aStream indent: level) ifFalse:
				[aStream nextPut: $(.
				 msgNode receiver emitCCodeAsExpressionOn: aStream level: level generator: self.
				 aStream crtab: level + 1; nextPut: $?; space.
				 msgNode args first emitCCodeAsArgumentOn: aStream level: level + 2 generator: self.
				 aStream crtab: level + 1; nextPut: $:; space.
				 msgNode args last emitCCodeAsArgumentOn: aStream level: level + 2 generator: self.
				 aStream nextPut: $)]]
		ifNotNil:
			[:const|
			(const
				ifTrue: [msgNode args first]
				ifFalse: [msgNode args last])
					emitCCodeAsArgumentOn: aStream level: level generator: self]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/8/2017 13:01'!
generateInlineCCode: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream.
	 There are two forms, self cCode: aString ... and self cCode: aBlock."

	msgNode args first isConstant
		ifTrue:
			[(msgNode args first value at: 1 ifAbsent: nil) = $# ifTrue:
				[aStream cr].
			aStream nextPutAll: msgNode args first value.
			self noteVariableUsageInString: msgNode args first value]
		ifFalse: [msgNode args first
					emitCCodeOn: aStream
					level: level
					generator: self]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/8/2017 13:01'!
generateInlineCCodeAsArgument: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream.
	 There are two forms, self cCode: aString ... and self cCode: aBlock."

	(msgNode args first isConstant
	 and: [msgNode args first value isString])
		ifTrue:
			[(msgNode args first value at: 1 ifAbsent: nil) = $# ifTrue:
				[aStream cr].
			aStream nextPutAll: msgNode args first value.
			self noteVariableUsageInString: msgNode args first value]
		ifFalse: [msgNode args first
					emitCCodeAsArgumentOn: aStream
					level: level
					generator: self]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/8/2017 13:01'!
generateInlineCPreprocessorDirective: msgNode on: aStream indent: level
	"Generate the C preprocessor directive for this message onto the given stream."

	aStream cr; nextPutAll: msgNode args first value.
	self noteVariableUsageInString: msgNode args first value! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 7/6/2021 19:05'!
generateInlineCppIfElse: msgNode asArgument: asArgument on: aStream indent: level
	"Generate the C code for this message onto the given stream."
	| lastCondition allAlternativesWereDeterminable expr pos directive index keywords |
	"Compile-time expansion for constants set in the options dictionary,
	 e.g. to cut down on noise for MULTIPLEBYTECODESETS."

	"First attempt to expand statically..."
	allAlternativesWereDeterminable := true.
	lastCondition := nil.
	msgNode selector keywords "Deal with cppIf:ifTrue:, cppIf:ifTrue:ifFalse:, cppIf:ifTrue:cppIf:ifTrue:ifFalse:, etc"
		with: msgNode args
		do: [:keyword :node| | expansion |
			self assert: (#(cppIf: ifFalse: ifTrue:) includes: keyword).
			keyword = #cppIf:
				ifTrue:
					[(self nilOrBooleanConditionFor: node)
						ifNil: [allAlternativesWereDeterminable := false]
						ifNotNil: [:condition| lastCondition := condition]]
				ifFalse:
					[(allAlternativesWereDeterminable
					 and: [lastCondition == (keyword = #ifTrue:)]) ifTrue:
						[asArgument
							ifTrue: "emitCCodeAsArgumentOn: doesn't indent, the code needs indenting if it takes multiple lines, so post-process."
								[expansion := String streamContents: [:s| node emitCCodeAsArgumentOn: s level: level generator: self].
								  aStream nextPutAll:
									  ((expansion includes: Character cr)
										ifTrue:
											[(String streamContents:
													[:s|
													s position > 0 ifTrue: [s tab: level + 1].
													node emitCCodeAsArgumentOn: s level: level generator: self])
												copyReplaceAll: (String with: Character cr)
												with: (String with: Character cr), (String new: level + 1 withAll: Character tab)]
										ifFalse: [expansion])]
							ifFalse:
								[expansion := String streamContents: [:s| node emitCCodeOn: s level: level generator: self].
								 "Remove tabs from first line to avoid indenting a second time"
								 expansion ifNotEmpty:
									[expansion := expansion allButFirst: (expansion findFirst: [:c| c ~~ Character tab]) - 1].
								 aStream nextPutAll: expansion].
						 ^self]]].
	self deny: allAlternativesWereDeterminable.

	"Expand for compile-time evaluation. Full #if ... #else..."

	directive := 'if'.
	index := 0.
	expr := nil.
	(keywords := msgNode selector keywords) "Deal with cppIf:ifTrue:, cppIf:ifFalse:, cppIf:ifTrue:ifFalse:, cppIf:ifTrue:cppIf:ifTrue:ifFalse:, etc"
		with: msgNode args
		do: [:keyword :node| | expansion |
			self assert: (#(cppIf: ifFalse: ifTrue:) includes: keyword).
			index := index + 1.
			keyword = #cppIf:
				ifTrue:
					[lastCondition := node.
					expansion := String streamContents: [:es| node emitCCodeAsArgumentOn: es level: 0 generator: self].
					(keywords at: index + 1) = #ifFalse: ifTrue:
						[expansion := node isLeaf ifTrue: ['!!', expansion] ifFalse: ['!!(', expansion, ')']].
					[expansion last isSeparator] whileTrue:
						[expansion := expansion allButLast].
					aStream
						ensureCr;
						nextPut: $#; space: level * 2; nextPutAll: directive; space; nextPutAll: expansion; cr.
					directive := 'elif'.
					expr := (expr ifNil: [' // '] ifNotNil: [expr, ' || ']), expansion]
				ifFalse:
					[(keywords at: index - 1) = #cppIf: ifFalse: "this is an else..."
						[aStream
							ensureCr;
							nextPut: $#; next: level * 2 put: Character space; nextPutAll: 'else'.
						 aStream position - pos >= 128 ifTrue:
							[aStream nextPutAll: expr].
						 aStream cr].
					pos := aStream position.
					self with: lastCondition
						ifAppropriateSetTo: keyword = #ifTrue:
						do: [asArgument
								ifTrue: "emitCCodeAsArgumentOn: doesn't indent, the code needs indenting in this case, so post-process."
									[aStream nextPutAll:
										((String streamContents:
												[:s|
												s next: level + 1 put: Character tab.
												node emitCCodeAsArgumentOn: s level: level generator: self])
											copyReplaceAll: (String with: Character cr)
											with: (String with: Character cr), (String new: level + 1 withAll: Character tab))]
								ifFalse:
									[node emitCCodeOn: aStream level: level generator: self]]]].
	aStream
		ensureCr;
		nextPut: $#; next: level * 2 put: Character space; nextPutAll: 'endif'.
	aStream position - pos >= 128 ifTrue:
		[aStream nextPutAll: expr].
	aStream cr.
	asArgument ifTrue:
		[aStream next: level + 1 put: Character tab]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 6/23/2011 10:10'!
generateInlineCppIfElse: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."
	self generateInlineCppIfElse: msgNode asArgument: false on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 6/23/2011 10:10'!
generateInlineCppIfElseAsArgument: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."
	self generateInlineCppIfElse: msgNode asArgument: true on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: ''!
generateInlineDirective: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '/* inline: '.
	aStream nextPutAll: msgNode args first name.
	aStream nextPutAll: ' */'.
! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'nice 8/18/2016 22:49'!
generateIntegerObjectOf: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."
	| expr mustCastToUnsigned type typeIsUnsigned |
	expr := msgNode args first.
	aStream nextPutAll: '(('.
	"Note that the default type of an integer constant in C is int.  Hence we /must/
	 cast expression to long if in the 64-bit world, since e.g. in 64-bits
		(int)(16r1FFFFF << 3) = (int)16rFFFFFFF8 = -8
	 whereas
		(long)(16r1FFFFF << 3) = (long) 16rFFFFFFF8 = 4294967288."
	type := self typeFor: expr in: currentMethod.
	typeIsUnsigned := type first = $u.
	mustCastToUnsigned := typeIsUnsigned not
		or: [(self sizeOfIntegralCType: type) < (self sizeOfIntegralCType: #usqInt)].
	mustCastToUnsigned ifTrue:
		[aStream nextPutAll: '(usqInt)'].
	self emitCExpression: expr on: aStream.
	aStream
		nextPutAll: ' << ';
		print: vmClass objectMemoryClass numSmallIntegerTagBits;
		nextPutAll: ') | 1)'! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 11/18/2014 19:01'!
generateIntegerValueOf: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPut: $(.
	self emitCExpression: msgNode args first on: aStream.
	aStream nextPutAll: ' >> ';
		print: vmClass objectMemoryClass numSmallIntegerTagBits;
		nextPut: $).! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/15/2014 15:02'!
generateIsIntegerObject: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	vmClass objectMemoryClass numSmallIntegerTagBits = 1
		ifTrue:
			[aStream nextPutAll: '('.
			 self emitCExpression: msgNode args first on: aStream.
			 aStream nextPutAll: ' & 1)']
		ifFalse:
			[aStream nextPutAll: '((('.
			 self emitCExpression: msgNode args first on: aStream.
			 aStream
				nextPutAll: ') & ';
				print: 1 << vmClass objectMemoryClass numSmallIntegerTagBits - 1;
				nextPutAll: ') == 1)']! !

!CCodeGenerator methodsFor: 'C translation' stamp: ''!
generateIsNil: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ' == '.
	aStream nextPutAll: (self cLiteralFor: nil).! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/16/2015 13:32'!
generateLessThan: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode receiver on: aStream indent: level.
	aStream nextPutAll: ' < '.
	self emitCExpression: msgNode args first on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/16/2015 13:40'!
generateLessThanOrEqual: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode receiver on: aStream indent: level.
	aStream nextPutAll: ' <= '.
	self emitCExpression: msgNode args first on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: ''!
generateMax: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '(('.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ' < '.
	self emitCExpression: msgNode args first on: aStream.
	aStream nextPutAll: ') ? '.
	self emitCExpression: msgNode args first on: aStream.
	aStream nextPutAll: ' : '.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ')'.! !

!CCodeGenerator methodsFor: 'C translation' stamp: ''!
generateMin: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '(('.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ' < '.
	self emitCExpression: msgNode args first on: aStream.
	aStream nextPutAll: ') ? '.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ' : '.
	self emitCExpression: msgNode args first on: aStream.
	aStream nextPutAll: ')'.! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 4/22/2015 09:27'!
generateMinus: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self generateAsConstantExpression: msgNode on: aStream) ifFalse:
		[msgNode args first constantNumbericValueOrNil ifNotNil:
			[:value|
			 value = 0 ifTrue:
				[^self emitCExpression: msgNode receiver on: aStream]].
		 self emitCExpression: msgNode receiver on: aStream.
		 aStream nextPutAll: ' - '.
		 self emitCExpression: msgNode args first on: aStream]! !

!CCodeGenerator methodsFor: 'C translation' stamp: ''!
generateModulo: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ' % '.
	self emitCExpression: msgNode args first on: aStream.! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 7/8/2008 17:47'!
generateNegated: msgNode on: aStream indent: level
	"Generate the C code for teh negated message onto the given stream."

	aStream nextPut: $-.
	self emitCExpression: msgNode receiver on: aStream! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 7/7/2021 22:10'!
generateNoMask: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '(!!('.
	self generateBitAnd: msgNode on: aStream indent: level.
	aStream nextPutAll: '))'! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 1/9/2017 12:01'!
generateNot: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '!!'.
	self emitCExpression: msgNode receiver on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/16/2015 13:40'!
generateNotEqual: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode receiver on: aStream indent: level.
	aStream nextPutAll: ' !!= '.
	self emitCExpression: msgNode args first on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: ''!
generateNotNil: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ' !!= '.
	aStream nextPutAll: (self cLiteralFor: nil).! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 2/25/2020 17:48'!
generateOdd: msgNode on: aStream indent: level
	"Generate the C code for even or odd onto the given stream."

	aStream nextPut: $(.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ' & 1)'! !

!CCodeGenerator methodsFor: 'C translation' stamp: ''!
generateOr: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ' || '.
	self emitCExpression: msgNode args first on: aStream.! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 1/18/2009 23:34'!
generatePerform: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	self emitCExpression: msgNode args first on: aStream.
	aStream nextPut: $(.
	(msgNode args copyFrom: 2 to: msgNode args size) do:[:arg|
		self emitCExpression: arg on: aStream.
	] separatedBy:[aStream nextPutAll:', '].
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 4/22/2015 09:26'!
generatePlus: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."
	(self generateAsConstantExpression: msgNode on: aStream) ifFalse:
		[msgNode args first constantNumbericValueOrNil ifNotNil:
			[:value|
			 value = 0 ifTrue:
				[^self emitCExpression: msgNode receiver on: aStream]].
		 self emitCExpression: msgNode receiver on: aStream.
		 aStream nextPutAll: ' + '.
		 self emitCExpression: msgNode args first on: aStream]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'ATG 4/8/2004 15:03'!
generatePreDecrement: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	| varNode |
	varNode := msgNode receiver.
	varNode isVariable
		ifFalse: [ self error: 'preDecrement can only be applied to variables' ].
	aStream nextPutAll: '--'.
	aStream nextPutAll: (self returnPrefixFromVariable: varNode name).
! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'ikp 9/11/2003 20:08'!
generatePreIncrement: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	| varNode |
	varNode := msgNode receiver.
	varNode isVariable
		ifFalse: [ self error: 'preIncrement can only be applied to variables' ].
	aStream nextPutAll: '++'.
	aStream nextPutAll: (self returnPrefixFromVariable: varNode name).
! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'ar 2/15/1999 21:43'!
generateRaisedTo: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll:'pow('.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ','.
	self emitCExpression: msgNode args first on: aStream.
	aStream nextPutAll:')'.! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 7/6/2021 19:09'!
generateRepeat: msgNode on: aStream indent: level
	"Generate while(1) { stmtList } "

	aStream nextPutAll: 'while (1) {'; cr.
	msgNode receiver emitCCodeOn: aStream level: level + 1 generator: self.
	aStream tab: level; nextPut: $}! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/14/2013 19:29'!
generateRounded: msgNode on: aStream indent: level
	"Generate the C code for rounded onto the given stream."

	aStream nextPutAll: 'round('.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 9/14/2015 11:33'!
generateSequentialAnd: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self nilOrBooleanConstantReceiverOf: msgNode receiver)
		ifNil:
			[self emitCExpression: msgNode receiver on: aStream indent: level.
			 "Don't eliminate the first expression if the second one is false, because of potential side-effects;
			  leave that to the C compiler. But /do/ eliminate the second expression if it is vaccuously true."
			 (self nilOrBooleanConstantReceiverOf: msgNode args first) ifNotNil:
				[:const| const == true ifTrue: [^self]].
			 aStream crtab: level; nextPutAll: ' && ('.
			 self emitCTestBlock: msgNode args first on: aStream indent: level.
			 aStream nextPut: $)]
		ifNotNil:
			[:const |
			const
				ifTrue: [self emitCTestBlock: msgNode args first on: aStream indent: level]
				ifFalse: [aStream nextPut: $0]]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 9/14/2015 11:34'!
generateSequentialOr: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."
	"Note: PP 2.3 compiler produces two arguments for or:, presumably
	 to help with inlining later. Taking the last agument should do the correct
	 thing even if your compiler is different."

	(self nilOrBooleanConstantReceiverOf: msgNode receiver)
		ifNil:
			[self emitCExpression: msgNode receiver on: aStream indent: level.
			 "Don't eliminate the first expression if the second one is true, because of potential side-effects;
			  leave that to the C compiler. But /do/ eliminate the second expression if it is vaccuously false."
			 (self nilOrBooleanConstantReceiverOf: msgNode args last) ifNotNil:
				[:const| const == false ifTrue: [^self]].
			 aStream crtab: level; nextPutAll: ' || ('.
			 self emitCTestBlock: msgNode args last on: aStream indent: level.
			 aStream nextPut: $)]
		ifNotNil:
			[:const |
			const
				ifTrue: [aStream nextPut: $1]
				ifFalse: [self emitCTestBlock: msgNode args last on: aStream indent: level]]! !

!CCodeGenerator methodsFor: 'C translation' stamp: ''!
generateSharedCodeDirective: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '/* common code: '.
	aStream nextPutAll: msgNode args first value.
	aStream nextPutAll: ' */'.
! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 3/20/2020 12:28'!
generateShiftLeft: msgNode on: aStream indent: level
	"Generate a C bitShift.  If the receiver type is unsigned avoid C99 undefined behaviour of
	 left shifting negative values (what?!!?!!? such quiche eating idiocy to treat this like anything
	 other than a truncated left shift) by casting signed receiver types to unsigned and back.
	 If we can determine the result would overflow the word size, cast to a long integer."
	| rcvr arg castToLong type mustCastBackToSign mustCastToUnsigned canSuffixTheConstant typeIsUnsigned |
	(self generateAsConstantExpression: msgNode on: aStream) ifTrue:
		[^self].
	rcvr := msgNode receiver.
	arg := msgNode args first.
	castToLong := false.
	(rcvr constantNumbericValueIfAtAllPossibleOrNilIn: self) ifNotNil:
		[:rcvrVal |
		 (arg constantNumbericValueIfAtAllPossibleOrNilIn: self)
			ifNil: [castToLong := vmClass notNil and: [vmClass objectMemoryClass wordSize = 8]]
			ifNotNil:
				[:argVal |
				| valueBeyondInt |
				valueBeyondInt := 1 bitShift: 32. "The default type of const << N is int."
				castToLong := rcvrVal < valueBeyondInt
								  and: [(rcvrVal bitShift: argVal) >= valueBeyondInt]]].
	canSuffixTheConstant := rcvr isConstant and: [rcvr name isEmpty and: [rcvr value >= 0]].
	canSuffixTheConstant ifTrue:
		[aStream nextPutAll: (self cLiteralForUnsignedInteger: rcvr value longlong: castToLong).
		aStream nextPutAll: ' << '.
		self emitCExpression: arg on: aStream indent: level.
		^self].
	type := self typeFor: rcvr in: currentMethod.
	castToLong := castToLong and: [(self sizeOfIntegralCType: type) < (self sizeOfIntegralCType: #usqLong)].
	typeIsUnsigned := type first = $u.
	mustCastToUnsigned := typeIsUnsigned not
							or: [castToLong
							or: [(self sizeOfIntegralCType: type) < (self sizeOfIntegralCType: #usqInt)]].
	mustCastBackToSign := typeIsUnsigned not.
	mustCastBackToSign ifTrue:
		[| promotedType |
		promotedType := castToLong
			ifTrue: [#sqLong]
			ifFalse: [(self sizeOfIntegralCType: type) < (self sizeOfIntegralCType: #sqInt)
				ifTrue: [#sqInt]
				ifFalse: [type]].
		aStream nextPutAll: '(('; nextPutAll: promotedType; nextPut: $)].
	mustCastToUnsigned ifTrue:
		[| unsigned |
		unsigned := castToLong
			ifTrue: [#usqLong]
			ifFalse: [(self sizeOfIntegralCType: type) < (self sizeOfIntegralCType: #usqLong)
				ifTrue: [#usqInt]
				ifFalse: [self unsignedTypeForIntegralType: type]].
		aStream nextPutAll: '(('; nextPutAll: unsigned; nextPutAll: ')('].
	self emitCExpression: rcvr on: aStream indent: level.
	mustCastToUnsigned ifTrue: [aStream nextPut: $)].

	aStream nextPutAll: ' << '.
	self emitCExpression: arg on: aStream indent: level.

	mustCastToUnsigned ifTrue: [aStream nextPut: $)].
	mustCastBackToSign ifTrue: [aStream nextPut: $)]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 3/20/2020 12:53'!
generateShiftRight: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream.
	Note that this generates a Logical Shift (unsigned), not an Arithmetic Shift (signed)."

	| type typeIsSigned mustCastToUnsigned unsignedType |
	(self generateAsConstantExpression: msgNode on: aStream) ifTrue:
		[^self].
	type := self typeFor: msgNode receiver in: currentMethod.
	typeIsSigned := type first ~= $u.
	mustCastToUnsigned := typeIsSigned or:
		["cast to usqInt if the int is shorter: we want to avoid UB related to a shift exceeeding bit width"
		(self sizeOfIntegralCType: type) < (self sizeOfIntegralCType: #usqInt)].
	"If not unsigned cast it to unsigned."
	mustCastToUnsigned
		ifTrue:
			["If the variable is a 64-bit type then don't cast it to usqInt (typically a 32-bit type)"
			unsignedType := (self sizeOfIntegralCType: type) < (self sizeOfIntegralCType: #usqLong)
				ifTrue: [#usqInt]
				ifFalse: [self unsignedTypeForIntegralType: type].
			 aStream nextPutAll: '(('; nextPutAll: unsignedType; nextPutAll: ')('.
			 self emitCExpression: msgNode receiver on: aStream indent: level.
			 aStream nextPutAll: '))']
		ifFalse:
			[aStream nextPutAll: '('.
			 self emitCExpression: msgNode receiver on: aStream indent: level.
			 aStream nextPut: $)].
	aStream nextPutAll: ' >> '.
	self emitCExpression: msgNode args first on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'nice 3/9/2020 23:49'!
generateSignedBitShift: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	| arg shift rightShift |
	(self isConstantNode: (arg := msgNode args first) valueInto: [:shiftValue| shift := shiftValue])
		ifTrue: "bit shift amount is a constant"
			[aStream nextPut: $(.
			shift < 0
				ifTrue:
					[rightShift := TSendNode new
						setSelector: #>>
						receiver: msgNode receiver
						arguments: {TConstantNode new setValue: shift negated}.
					self generateSignedShiftRight: rightShift on: aStream indent: level]
				ifFalse: [self generateShiftLeft: msgNode on: aStream indent: level].
			aStream nextPut: $)]
		ifFalse: "bit shift amount is an expression"
			[rightShift := TSendNode new
						setSelector: #>>
						receiver: msgNode receiver
						arguments: {TSendNode new
							setSelector: #negated
							receiver: arg
							arguments: #()}.
			aStream nextPutAll: '(('.
			self emitCExpression: arg on: aStream.
			aStream nextPutAll: ' < 0) ? ('.
			self generateSignedShiftRight: rightShift on: aStream indent: level. 
			aStream nextPutAll: ') : ('.
			self generateShiftLeft: msgNode on: aStream indent: level.
			aStream nextPutAll: '))']! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/14/2015 12:22'!
generateSignedIntFromLong64: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '((sqLong) '.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/14/2015 12:22'!
generateSignedIntFromLong: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '((int) '.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/14/2015 12:22'!
generateSignedIntFromShort: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '((short)'.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/14/2015 19:04'!
generateSignedIntToLong64: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '((usqInt) '.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/15/2015 11:11'!
generateSignedIntToLong: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '((usqInt) '.
	(vmClass notNil
	 and: [vmClass objectMemoryClass wordSize = 8]) ifTrue:
		[aStream nextPutAll: '(int)'].
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/14/2015 19:04'!
generateSignedIntToShort: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: '((usqInt) (short)'.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'nice 3/24/2020 00:35'!
generateSignedShiftRight: msgNode on: aStream indent: level
	"Generate the C code for >>> onto the given stream."

	| type typeIsUnsigned mustCastToSigned signedType |
	type := self typeFor: msgNode receiver in: currentMethod.
	typeIsUnsigned := type first = $u.
	mustCastToSigned := typeIsUnsigned or:
		["cast to sqInt if the int is shorter: we want to avoid UB related to a shift exceeeding bit width"
		(self sizeOfIntegralCType: type) < (self sizeOfIntegralCType: #usqInt)].
	mustCastToSigned
		ifTrue:
			["If the variable is a 64-bit type then don't cast it to usqInt (typically a 32-bit type)"
			signedType := (self sizeOfIntegralCType: type) < (self sizeOfIntegralCType: #usqLong)
				ifTrue: [#sqInt]
				ifFalse: [self signedTypeForIntegralType: type].
			 aStream nextPutAll: '(('; nextPutAll: signedType; nextPutAll: ')('.
			 self emitCExpression: msgNode receiver on: aStream indent: level.
			 aStream nextPutAll: '))']
		ifFalse:
			[aStream nextPutAll: '('.
			 self emitCExpression: msgNode receiver on: aStream indent: level.
			 aStream nextPut: $)].
	aStream nextPutAll: ' >> '.
	self emitCExpression: msgNode args first on: aStream! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 11/19/2014 16:22'!
generateSmallIntegerConstant: aSendNode on: aStream indent: indent
	aSendNode selector = #minSmallInteger ifTrue:
		[^aStream nextPutAll: 'MinSmallInteger'].
	aSendNode selector = #maxSmallInteger ifTrue:
		[^aStream nextPutAll: 'MaxSmallInteger'].
	self error: 'unknown SmallInteger constant'! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 2/26/2009 10:18'!
generateSmalltalkMetaError: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPutAll: 'error("'; nextPutAll: msgNode selector; nextPutAll: '")'! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 4/22/2015 09:26'!
generateTimes: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	(self generateAsConstantExpression: msgNode on: aStream) ifFalse:
		[msgNode args first constantNumbericValueOrNil ifNotNil:
			[:value|
			 value = 0 ifTrue:
				[^self emitCExpression: msgNode args first on: aStream].
			 value = 1 ifTrue:
				[^self emitCExpression: msgNode receiver on: aStream]].
		 self emitCExpression: msgNode receiver on: aStream.
		 aStream nextPutAll: ' * '.
		 self emitCExpression: msgNode args first on: aStream]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/2/2014 11:41'!
generateTimesRepeat: msgNode on: aStream indent: level
	"Generate while(1) { stmtList } "

	msgNode receiver value timesRepeat:
		[msgNode args first emitCCodeOn: aStream level: level generator: self]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 7/25/2020 10:36'!
generateToByDo: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."
	"N.B. MessageNode>>asTranslatorNodeIn: adds the limit var as a hidden fourth argument."
	| blockExpr iterationVar limitExpr mayHaveSideEffects limitVar step |
	blockExpr := msgNode args third.
	blockExpr args size = 1 ifFalse:
		[self error: 'wrong number of block arguments'].
	iterationVar := blockExpr args first.
	limitExpr := msgNode args first.
	aStream nextPutAll: 'for (', iterationVar, ' = '.
	self noteUsedVariableName: iterationVar.
	self emitCExpression: msgNode receiver on: aStream.
	mayHaveSideEffects := msgNode args size = 4. "See TMethod>>prepareMethodIn:"
	mayHaveSideEffects ifTrue:
		[limitVar := msgNode args last.
		 aStream nextPutAll: ', ', limitVar name, ' = '.
		 self emitCExpression: limitExpr on: aStream.
		 limitExpr := limitVar].
	aStream nextPutAll: '; ', iterationVar.
	step := msgNode args at: 2.
	self generateToByDoLimitExpression: limitExpr
		negative: (self stepExpressionIsNegative: step)
		on: aStream.
	aStream nextPutAll: '; ', iterationVar, ' += '.
	self emitCExpression: step on: aStream.
	aStream nextPutAll: ') {'; cr.
	blockExpr emitCCodeOn: aStream level: level + 1 generator: self.
	aStream tab: level; nextPut: $}! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/5/2013 08:42'!
generateToByDoLimitExpression: limitExpr negative: negative on: aStream
	"Generate the limit expression for a to:do: or to:by:do:.
	 If the limitExpr is of the form foo - 1 use < foo or > foo, as appropriate, instead of
	 <= (foo - 1) or >= (foo - 1), so as to avoid underflowing an unsigned value to maxint."
	(limitExpr isSend
	 and: [limitExpr selector = #-
	 and: [limitExpr args first isConstant
	 and: [limitExpr args first value = 1]]])
		ifTrue:
			[aStream nextPutAll: (negative ifTrue: [' > '] ifFalse: [' < ']).
			 self emitCExpression: limitExpr receiver on: aStream]
		ifFalse:
			[aStream nextPutAll: (negative ifTrue: [' >= '] ifFalse: [' <= ']).
			 self emitCExpression: limitExpr on: aStream]! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 7/25/2020 10:36'!
generateToDo: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	| iterationVar |
	(msgNode args last args size = 1) ifFalse: [
		self error: 'wrong number of block arguments'.
	].
	iterationVar := msgNode args last args first.
	aStream nextPutAll: 'for (', iterationVar, ' = '.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: '; ', iterationVar.
	self generateToByDoLimitExpression: msgNode args first negative: false on: aStream.
	aStream nextPutAll: '; ', iterationVar, '++) {'; cr.
	msgNode args last emitCCodeOn: aStream level: level + 1 generator: self.
	aStream tab: level; nextPut: $}! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'tpr 7/26/2003 10:23'!
generateTouch: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream - which is to say absolutely nothing"
! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/26/2016 12:19'!
generateTruncateTo: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	| arg |
	(self isConstantNode: msgNode args first valueInto: [:a| arg := a]) ifFalse:
		[self error: 'can''t find method for inlining truncateTo:'].
	self assert: (arg isInteger and: [arg isPowerOfTwo]).
	aStream nextPut: $(.
	self emitCExpression: msgNode receiver on: aStream.
	aStream nextPutAll: ' & ~'; print: arg - 1; nextPut: $)! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 12/8/2017 15:54'!
generateValue: aTSendNode on: aStream indent: level
	"Reduce [:formal ... :formalN| body ] value: actual ... value: actualN
	 to body with formals substituted for by actuals."
	| substitution substitutionDict newLabels |
	self assert: aTSendNode receiver isStmtList.
	self assert: aTSendNode receiver args size = aTSendNode args size.
	substitution := aTSendNode receiver copy.
	substitution renameLabelsForInliningInto: currentMethod.
	substitutionDict := Dictionary new: aTSendNode args size * 2.
	aTSendNode receiver args with: aTSendNode args do:
		[ :argName :exprNode |
		exprNode isLeaf
			ifTrue: [substitutionDict at: argName put: exprNode]
			ifFalse:
				[aStream nextPutAll: argName; nextPutAll: ' = '.
				 self noteUsedVariableName: argName.
				 exprNode emitCCodeAsExpressionOn: aStream level: level generator: self.
				 aStream nextPut: $; ; crtab: level]].
	substitution
		bindVariablesIn: substitutionDict;
		emitCCodeOn: aStream level: level generator: self.
	newLabels := Set withAll: currentMethod labels.
	substitution nodesDo:
		[:node| node isLabel ifTrue: [node label ifNotNil: [:label| newLabels add: label]]].
	"now add the new labels so that a subsequent inline of
	 the same block will be renamed with different labels."
	currentMethod labels: newLabels! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/22/2013 16:01'!
generateValueAsArgument: aTSendNode on: aStream indent: level
	"Reduce [:formal ... :formalN| body ] value: actual ... value: actualN
	 to body with formals substituted for by actuals."
	| substitution substitutionDict newLabels |
	self assert: aTSendNode receiver isStmtList.
	self assert: aTSendNode receiver args size = aTSendNode args size.
	substitution := aTSendNode receiver copy.
	substitution renameLabelsForInliningInto: currentMethod.
	substitutionDict := Dictionary new: aTSendNode args size * 2.
	aTSendNode receiver args with: aTSendNode args do:
		[ :argName :exprNode |
		substitutionDict at: argName put: exprNode].
	substitution
		bindVariablesIn: substitutionDict;
		emitCCodeAsArgumentOn: aStream level: level generator: self.
	newLabels := Set withAll: currentMethod labels.
	substitution nodesDo:
		[:node| node isLabel ifTrue: [node label ifNotNil: [:label| newLabels add: label]]].
	"now add the new labels so that a subsequent inline of
	 the same block will be renamed with different labels."
	currentMethod labels: newLabels! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 5/4/2014 09:14'!
generateWhile: boolean loop: msgNode on: aStream indent: level
	"Generate while(cond) {stmtList} or while(!!(cond)) {stmtList}."

	aStream nextPutAll: 'while ('.
	boolean ifFalse: [aStream nextPutAll: '!!('].
	self emitCTestBlock: msgNode receiver on: aStream indent: level.
	boolean ifFalse: [aStream nextPut: $)].
	aStream nextPutAll: ') {'; cr.
	msgNode args first isNilStmtListNode ifFalse:
		[msgNode args first emitCCodeOn: aStream level: level + 1 generator: self].
	aStream tab: level.
	aStream nextPut: $}! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 5/3/2014 22:38'!
generateWhileFalse: msgNode on: aStream indent: level
	"Generate C code for a loop in one of the following formats, as appropriate:
		while(!!(cond)) { stmtList }
		do {stmtList} while(!!(cond))
		while(1) {stmtListA; if (cond) break; stmtListB}"

	msgNode receiver statements size <= 1
		ifTrue: [^self generateWhile: false loop: msgNode on: aStream indent: level].
	msgNode args first isNilStmtListNode
		ifTrue: [^self generateDoWhileFalse: msgNode on: aStream indent: level].
	^self generateWhileForeverBreakIf: true loop: msgNode on: aStream indent: level! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 5/2/2014 14:29'!
generateWhileForeverBreakIf: breakBoolean loop: msgNode on: aStream indent: level
	"Generate either of
		while(1) {stmtListA; if(cond) break; stmtListB}
		while(1) {stmtListA; if(!!(cond)) break; stmtListB}."

	| testStmt receiverWithoutTest |
	aStream peekLast ~~ Character tab ifTrue:
		[aStream tab: level - 1].
	aStream nextPutAll: 'while (1) {'; cr.
	testStmt := msgNode receiver statements last.
	receiverWithoutTest := TStmtListNode new setStatements: msgNode receiver statements allButLast.
	receiverWithoutTest emitCCodeOn: aStream level: level + 1 generator: self.
	aStream tab: level + 1; nextPutAll: 'if ('.
	breakBoolean ifFalse: [aStream nextPut: $!!; nextPut: $(].
	testStmt emitCCodeAsExpressionOn: aStream level: level + 1 generator: self.
	breakBoolean ifFalse: [aStream nextPut: $)].
	aStream nextPutAll: ') break;'; cr.
	msgNode args first emitCCodeOn: aStream level: level + 1 generator: self.
	aStream tab: level; nextPut: $}! !

!CCodeGenerator methodsFor: 'C translation' stamp: 'eem 5/3/2014 22:39'!
generateWhileTrue: msgNode on: aStream indent: level
	"Generate C code for a loop in one of the following formats, as appropriate:
		while(cond) { stmtList }
		do {stmtList} while(cond)
		while(1) {stmtListA; if (!!(cond)) break; stmtListB}"

	msgNode receiver statements size <= 1
		ifTrue: [^self generateWhile: true loop: msgNode on: aStream indent: level].
	msgNode args first isNilStmtListNode
		ifTrue: [^self generateDoWhileTrue: msgNode on: aStream indent: level].
	^self generateWhileForeverBreakIf: false loop: msgNode on: aStream indent: level! !

!CCodeGenerator methodsFor: 'initialize-release' stamp: 'eem 9/18/2020 18:23'!
getLogger
	logger := (ProvideAnswerNotification new tag: #logger; signal) ifNil: [Transcript]! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 2/11/2009 09:00'!
globalsAsSet
	"Used by the inliner to avoid name clashes with global variables."

	^variables! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'eem 3/10/2017 17:12'!
harmonizeReturnTypesIn: aSetOfTypes
	"Eliminate signed/unsigned conflicts in aSetOfTypes. Non-negative integers can be either
	 signed or unsigned. Ignore them unless there are no types, in which case default to sqInt."
	| constantIntegers sqs usqs |
	constantIntegers := aSetOfTypes select: [:element| element isInteger].
	aSetOfTypes removeAll: constantIntegers.
	"N.B. Because of LP64 vs LLP64 issues do *not* rename #long to #sqInt or #'unsigned long' to #usqInt"
	#(char short int #'long long' #'unsigned char' #'unsigned short' #'unsigned int' #'unsigned long long')
		with: #(sqInt sqInt sqInt sqLong usqInt usqInt usqInt usqLong)
		do: [:type :replacement|
			(aSetOfTypes includes: type) ifTrue:
				[aSetOfTypes remove: type; add: replacement]].
	sqs := aSetOfTypes select: [:t| t beginsWith: 'sq'].
	usqs := aSetOfTypes select: [:t| t beginsWith: 'usq'].
	^(sqs size + usqs size = aSetOfTypes size
	   and: [sqs notEmpty
	   and: [sqs allSatisfy: [:t| usqs includes: 'u', t]]])
		ifTrue: [sqs]
		ifFalse: [(aSetOfTypes isEmpty and: [constantIntegers notEmpty])
					ifTrue: [Set with: #sqInt]
					ifFalse: [aSetOfTypes]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 2/3/2020 10:15'!
hasVariable: variableName
	^variables includes: variableName! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 6/6/2018 18:33'!
ifStaticallyResolvedPolymorphicReceiverThenUpdateSelectorIn: aSendNode
	"We allow a limited amount of polymorphism; if a class chooses, its selectoers can be
	 prefixed with a given string to disambiguate. This hack allows us to use two different
	 compaction algorithms with the same API at the same time; the selection being done
	 by a class which holds the flag stating which algorithm is in effect at the current time."
	| class |
	staticallyResolvedPolymorphicReceivers ifNil: [^self].
	aSendNode receiver isVariable ifFalse:
		[^self].
	class := staticallyResolvedPolymorphicReceivers
				at: aSendNode receiver name
				ifAbsent: [^self].
	aSendNode selector: (class staticallyResolvePolymorphicSelector: aSendNode selector)! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'eem 10/30/2013 11:16'!
implicitReturnTypeFor: aSelector
	"Answer the default return type for mthods of the defining class of
	 aSelector. Default to #sqInt if there is no method, or if the method's
	 defining class doesn't understand implicitReturnTypeFor:."
	| m |
	m := methods at: aSelector ifAbsent: [^#sqInt].
	^[m definingClass implicitReturnTypeFor: aSelector]
		on: MessageNotUnderstood
		do: [:ex|
			ex message selector ~~ #implicitReturnTypeFor: ifTrue:
				[ex pass].
			#sqInt]! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 11/13/2015 14:24'!
includeAPIFrom: aCCodeGenerator
	apiMethods := aCCodeGenerator selectAPIMethods.
	apiVariables := aCCodeGenerator selectAPIVariables! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'eem 6/8/2016 09:19'!
inferTypesForImplicitlyTypedVariablesAndMethods
	"Infer the return tupe and the types of untyped variables.
	 As far as variables go, for now we try only to infer variables
	 assigned the result of #longLongAt:, but much more could be
	 done here."

	"Iterate over all methods, inferring #void return types, until we reach a fixed point."
	| allMethods |
	allMethods := apiMethods
					ifNil: [methods]
					ifNotNil: [(Set withAll: methods)
								addAll: apiMethods;
								yourself].
	"Make an initial pass to assign the return types of all simple methods that return constants,
	 or those that have explicit return types."						
	allMethods do:
		[:m|
		m removeFinalSelfReturnIn: self. "must precede recordDeclarationsIn: because it may set returnType"
		m recordDeclarationsIn: self.
		(m returnType isNil
		 and: [m isReturnConstant]) ifTrue:
			[m inferReturnTypeIn: self]].

	"now iterate until we reach a fixed point"
	[| changedReturnType |
	 changedReturnType := false.
	 allMethods do:
		[:m|
		 m inferTypesForImplicitlyTypedVariablesIn: self.
		 (m inferReturnTypeIn: self) ifTrue:
			[changedReturnType := true]].
	 changedReturnType] whileTrue.

	"Type all as-yet-untyped methods as the default"
	methods do:
		[:m|
		m returnType ifNil:
			[m returnType: (m returnsExpression
								ifTrue: [#sqInt]
								ifFalse: [self implicitReturnTypeFor: m])].
		m transformReturns].

	"Make a final pass to type anything assigned from the default type"
	allMethods do:
		[:m|
		 m inferTypesForImplicitlyTypedVariablesIn: self]! !

!CCodeGenerator methodsFor: 'initialize-release' stamp: 'eem 7/31/2021 19:14'!
initialize
	translationDict := Dictionary new.
	inlineList := Array new.
	constants := Dictionary new: 100.
	variables := Set new: 100.
	variableDeclarations := Dictionary new: 100.
	methods := Dictionary new: 500.
	kernelReturnTypes := self computeKernelReturnTypes.
	macros := Dictionary new.
	self initializeCTranslationDictionary.
	headerFiles := OrderedCollection new.
	globalVariableUsage := Dictionary new.
	useSymbolicConstants := true.
	generateDeadCode := true.
	scopeStack := OrderedCollection new.
	self getLogger.
	pools := IdentitySet new.
	selectorTranslations := IdentityDictionary new.
	accessorDepthCache := IdentityDictionary new.
	beganInlining := false.
	suppressAsmLabels := false.
	previousCommentMarksInlining := false.
	previousCommenter := nil.
	breakSrcInlineSelectors := IdentitySet new.
	breakDestInlineSelectors := IdentitySet new! !

!CCodeGenerator methodsFor: 'C translation support' stamp: 'eem 8/1/2021 11:38'!
initializeCTranslationDictionary 
	"Initialize the dictionary mapping message names to actions for C code generation."

	| pairs |
	
	translationDict := Dictionary new: 200.
	pairs := #(
	#&				#generateAnd:on:indent:
	#|				#generateOr:on:indent:
	#abs			#generateAbs:on:indent:
	#and:			#generateSequentialAnd:on:indent:
	#or:			#generateSequentialOr:on:indent:
	#not			#generateNot:on:indent:

	#+				#generatePlus:on:indent:
	#-				#generateMinus:on:indent:
	#negated		#generateNegated:on:indent:
	#*				#generateTimes:on:indent:
	#/				#generateDivide:on:indent:
	#//				#generateDivide:on:indent:
	#\\				#generateModulo:on:indent:
	#<<			#generateShiftLeft:on:indent:
	#>>			#generateShiftRight:on:indent:
	#>>>			#generateSignedShiftRight:on:indent:
	#,				#generateComma:on:indent:
	#min:			#generateMin:on:indent:
	#max:			#generateMax:on:indent:
	#between:and:	#generateBetweenAnd:on:indent:

	#bitAnd:			#generateBitAnd:on:indent:
	#bitOr:				#generateBitOr:on:indent:
	#bitXor:			#generateBitXor:on:indent:
	#bitShift:			#generateBitShift:on:indent:
	#signedBitShift:	#generateSignedBitShift:on:indent:
	#bitInvert			#generateBitInvert:on:indent:
	#bitInvert32		#generateBitInvert:on:indent:
	#bitInvert64		#generateBitInvert:on:indent:
	#bitClear:			#generateBitClear:on:indent:
	#truncateTo:		#generateTruncateTo:on:indent:
	#rounded			#generateRounded:on:indent:
	#even				#generateEven:on:indent:
	#odd				#generateOdd:on:indent:

	#byteSwap32		#generateByteSwap32:on:indent:
	#byteSwap64		#generateByteSwap64:on:indent:
	#byteSwapped32IfBigEndian:	generateByteSwap32IfBigEndian:on:indent:
	#byteSwapped64IfBigEndian:	generateByteSwap64IfBigEndian:on:indent:
	
	#<				#generateLessThan:on:indent:
	#<=			#generateLessThanOrEqual:on:indent:
	#=				#generateEqual:on:indent:
	#>				#generateGreaterThan:on:indent:
	#>=			#generateGreaterThanOrEqual:on:indent:
	#~=			#generateNotEqual:on:indent:
	#==			#generateEqual:on:indent:
	#~~			#generateNotEqual:on:indent:
	#isNil			#generateIsNil:on:indent:
	#notNil			#generateNotNil:on:indent:

	#whileTrue: 	#generateWhileTrue:on:indent:
	#whileFalse:	#generateWhileFalse:on:indent:
	#whileTrue 	#generateDoWhileTrue:on:indent:
	#whileFalse		#generateDoWhileFalse:on:indent:
	#to:do:			#generateToDo:on:indent:
	#to:by:do:		#generateToByDo:on:indent:
	#repeat 		#generateRepeat:on:indent:
	#timesRepeat:	#generateTimesRepeat:on:indent:

	#ifTrue:			#generateIfTrue:on:indent:
	#ifFalse:		#generateIfFalse:on:indent:
	#ifTrue:ifFalse:	#generateIfTrueIfFalse:on:indent:
	#ifFalse:ifTrue:	#generateIfFalseIfTrue:on:indent:

	#ifNotNil:		#generateIfNotNil:on:indent:
	#ifNil:			#generateIfNil:on:indent:
	#ifNotNil:ifNil:	#generateIfNotNilIfNil:on:indent:
	#ifNil:ifNotNil:	#generateIfNilIfNotNil:on:indent:

	#at:			#generateAt:on:indent:
	#at:put:		#generateAtPut:on:indent:
	#basicAt:		#generateAt:on:indent:
	#basicAt:put:	#generateAtPut:on:indent:

	#integerValueOf:			#generateIntegerValueOf:on:indent:
	#integerObjectOf:			#generateIntegerObjectOf:on:indent:
	#isIntegerObject: 			#generateIsIntegerObject:on:indent:
	#cCode:					#generateInlineCCode:on:indent:
	#cCode:inSmalltalk:		#generateInlineCCode:on:indent:
	#cPreprocessorDirective:	#generateInlineCPreprocessorDirective:on:indent:
	#cppIf:ifTrue:ifFalse:				#generateInlineCppIfElse:on:indent:
	#cppIf:ifTrue:cppIf:ifTrue:ifFalse:	#generateInlineCppIfElse:on:indent:
	#cppIf:ifTrue:						#generateInlineCppIfElse:on:indent:
	#cppIf:ifFalse:						#generateInlineCppIfElse:on:indent:
	#cCoerce:to:				#generateCCoercion:on:indent:
	#cCoerceSimple:to:			#generateCCoercion:on:indent:
	#addressOf:				#generateAddressOf:on:indent:
	#addressOf:put:			#generateAddressOf:on:indent:
	#asAddress:put:			#generateAsAddress:on:indent:
	#signedIntFromLong64	#generateSignedIntFromLong64:on:indent:
	#signedIntFromLong		#generateSignedIntFromLong:on:indent:
	#signedIntFromShort		#generateSignedIntFromShort:on:indent:
	#signedIntToLong64		#generateSignedIntToLong64:on:indent:
	#signedIntToLong			#generateSignedIntToLong:on:indent:
	#signedIntToShort			#generateSignedIntToShort:on:indent:
	#preIncrement				#generatePreIncrement:on:indent:
	#preDecrement			#generatePreDecrement:on:indent:
	#inline:						#generateInlineDirective:on:indent:
	#asFloat					#generateAsFloat:on:indent:
	#asInteger					#generateAsInteger:on:indent:
	#asIntegerPtr				#generateAsIntegerPtr:on:indent:
	#asUnsignedInteger		#generateAsUnsignedInteger:on:indent:
	#asUnsignedIntegerPtr		#generateAsUnsignedIntegerPtr:on:indent:
	#asLong					#generateAsLong:on:indent:
	#asUnsignedLong			#generateAsUnsignedLong:on:indent:
	#asUnsignedLongLong		#generateAsUnsignedLongLong:on:indent:
	#asVoidPointer				#generateAsVoidPointer:on:indent:
	#asSymbol					#generateAsSymbol:on:indent:
	#flag:						#generateFlag:on:indent:
	#anyMask:					#generateAnyMask:on:indent:
	#allMask:					#generateAllMask:on:indent:
	#noMask:					#generateNoMask:on:indent:
	#raisedTo:					#generateRaisedTo:on:indent:
	#touch:					#generateTouch:on:indent:

	#bytesPerOop 				#generateBytesPerOop:on:indent:
	#bytesPerWord 			#generateBytesPerWord:on:indent:
	#wordSize		 			#generateBytesPerWord:on:indent:
	#baseHeaderSize			#generateBaseHeaderSize:on:indent:
	#minSmallInteger			#generateSmallIntegerConstant:on:indent:
	#maxSmallInteger			#generateSmallIntegerConstant:on:indent:
	
	#sharedCodeNamed:inCase:		#generateSharedCodeDirective:on:indent:

	#perform:							#generatePerform:on:indent:
	#perform:with:						#generatePerform:on:indent:
	#perform:with:with:				#generatePerform:on:indent:
	#perform:with:with:with:			#generatePerform:on:indent:
	#perform:with:with:with:with:		#generatePerform:on:indent:
	#perform:with:with:with:with:with:	#generatePerform:on:indent:

	#value									#generateValue:on:indent:
	#value:									#generateValue:on:indent:
	#value:value:							#generateValue:on:indent:
	#value:value:value:						#generateValue:on:indent:
	#value:value:value:value:				#generateValue:on:indent:
	#value:value:value:value:value:			#generateValue:on:indent:
	#value:value:value:value:value:value:	#generateValue:on:indent:

	#deny:								#generateDeny:on:indent:

	#shouldNotImplement				#generateSmalltalkMetaError:on:indent:
	#shouldBeImplemented			#generateSmalltalkMetaError:on:indent:
	#subclassResponsibility				#generateSmalltalkMetaError:on:indent:
	).

	1 to: pairs size by: 2 do: [:i |
		translationDict at: (pairs at: i) put: (pairs at: i + 1)].

	pairs := #(
	#ifTrue:				#generateIfTrueAsArgument:on:indent:	
	#ifFalse:				#generateIfFalseAsArgument:on:indent:
	#ifTrue:ifFalse:			#generateIfTrueIfFalseAsArgument:on:indent:
	#ifFalse:ifTrue:			#generateIfFalseIfTrueAsArgument:on:indent:
	#ifNotNil:				#generateIfNotNilAsArgument:on:indent:	
	#ifNil:					#generateIfNilAsArgument:on:indent:
	#ifNotNil:ifNil:			#generateIfNotNilIfNilAsArgument:on:indent:
	#ifNil:ifNotNil:			#generateIfNilIfNotNilAsArgument:on:indent:
	#cCode:				#generateInlineCCodeAsArgument:on:indent:
	#cCode:inSmalltalk:	#generateInlineCCodeAsArgument:on:indent:
	#cppIf:ifTrue:ifFalse:	#generateInlineCppIfElseAsArgument:on:indent:
	#cppIf:ifTrue:			#generateInlineCppIfElseAsArgument:on:indent:

	#value					#generateValueAsArgument:on:indent:
	#value:					#generateValueAsArgument:on:indent:
	#value:value:			#generateValueAsArgument:on:indent:
	).

	asArgumentTranslationDict := Dictionary new: 8.
	1 to: pairs size by: 2 do: [:i |
		asArgumentTranslationDict at: (pairs at: i) put: (pairs at: i + 1)].
! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 8/15/2018 18:32'!
initializerForInstVar: varName in: aClass
	| instVarIndex |
	instVarIndex := aClass instVarIndexFor: varName ifAbsent: [^nil].
	aClass selectorsAndMethodsDo:
		[:s :m| | tmeth |
		((s beginsWith: 'initialize')
		 and: [m writesField: instVarIndex]) ifTrue:
			[tmeth := m asTranslationMethodOfClass: TMethod.
			 tmeth parseTree nodesDo:
				[:node| | exprOrAssignment |
				(node isAssignment
				 and: [node variable name = varName]) ifTrue:
					[exprOrAssignment := node.
					 [exprOrAssignment isAssignment] whileTrue:
						[exprOrAssignment := exprOrAssignment expression].
					 ^exprOrAssignment]]]].
	^nil! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 6/7/2018 15:33'!
initializerForInstVar: varName inStartClass: aClass
	| allClasses sizeBefore |
	(aClass inheritsFrom: VMClass) ifFalse:
		[^false].
	allClasses := (aClass withAllSuperclasses copyUpTo: VMClass) asSet.
	[sizeBefore := allClasses size.
	 allClasses copy do:
		[:class|
		class ancilliaryClasses do:
			[:ancilliary|
			allClasses addAll: (ancilliary withAllSuperclasses copyUpTo: VMClass)]].
	 sizeBefore ~= allClasses size] whileTrue.
	allClasses do:
		[:class|
		(self initializerForInstVar: varName in: class) ifNotNil:
			[:initializer| ^initializer]].
	^nil! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 7/26/2011 13:51'!
inlineDispatchesInMethodNamed: selector localizingVars: varsList
	"Inline dispatches (case statements) in the method with the given name."

	| m |
	m := self methodNamed: selector.
	m = nil ifFalse: [
		m inlineCaseStatementBranchesIn: self localizingVars: varsList.
		m parseTree nodesDo: [ :n |
			n isCaseStmt ifTrue: [
				n customizeShortCasesForDispatchVar: 'currentBytecode' in: self method: m.
			].
		].
	].
	varsList do: [ :v | | varString |
		varString := v asString.
		self checkDeleteVariable: varString.
		variables remove: varString ifAbsent: [].
		(variableDeclarations includesKey: varString) ifTrue: [
			m declarationAt: v asString put: (variableDeclarations at: varString).
			variableDeclarations removeKey: varString.
		].
	]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 11/16/2013 13:37'!
instVarNamesForClass: aClass
	^vmClass
		ifNil: [aClass instVarNames]
		ifNotNil: [aClass instVarNames reject:
					[:ivn| vmClass isNonArgumentImplicitReceiverVariableName: ivn]]! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 11/15/2019 21:52'!
interpreterVersion
	^self vmClass interpreterVersion, '[', self vmClass objectMemoryClass memoryManagerName, ']'! !

!CCodeGenerator methodsFor: 'C translation support' stamp: 'eem 7/1/2014 08:21'!
is64BitIntegralVariable: node typeInto: aBlock
	| decl |
	^node isVariable
	  and: [(decl := self typeOfVariable: node name) notNil
	  and: [(#('usqLong' 'sqLong' 'unsigned long long' 'long long' 'unsigned __int64' '__int64')
			anySatisfy: [:type| decl beginsWith: type])
	  and: [aBlock value: ((decl endsWith: node name)
							ifTrue: [decl allButLast: node name size + 1]
							ifFalse: [decl]).
			true]]]! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 2/19/2015 10:41'!
isActualType: actualTypeOrNil compatibleWithFormalType: formalTypeOrNil 
	| actualType formalType |
	actualType := actualTypeOrNil ifNil: [#sqInt].
	formalType := formalTypeOrNil ifNil: [#sqInt].
	((self isIntegralCType: actualType)
	 and: [self isIntegralCType: formalType]) ifFalse:
		[^actualType = formalType
		 or: [formalType = #double and: [actualType = #float]]].
	"For now, insist that the signedness agrees.  If the actual's type is unknown allow inlining.
	 A formal defaults to #sqInt. Allowing inlining an unsigned type within a sqInt formal is wrong;
	 snd for testing, it breaks e.g. the BitBltPlugin."
	^(actualType first = $u) = (formalType first = $u)
	  or: [actualTypeOrNil isNil] ! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 1/9/2015 10:40'!
isAssertSelector: selector
	^#(assert: asserta: assert:l: asserta:l: deny: eassert:) includes: selector! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 7/19/2013 10:32'!
isBuiltinSelector: sel
	"Answer true if the given selector is one of the builtin selectors."

	^(self isKernelSelector: sel) or: [translationDict includesKey: sel]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 9/7/2018 20:29'!
isCLiteral: anObject
	anObject isInteger ifTrue: [^true].
	anObject isString ifTrue: [^true].
	anObject isFloat ifTrue: [^true].
	anObject == nil ifTrue: [^true].
	anObject == true ifTrue: [^true].			"ikp"
	anObject == false ifTrue: [^true].			"ikp"
	anObject isCharacter ifTrue:[^true]. "ar"
	^false! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 6/29/2018 16:33'!
isConstantNode: aNode valueInto: aBlock
	"Answer if aNode evaluates to a constant, and if so, evaluate aBlock with the value of that constant."

	aNode isConstant ifTrue:
		[(aNode isDefine
		  and: [(vmClass ifNil: [VMBasicConstants]) defineAtCompileTime: aNode name]) ifTrue:
			[^false].
		 aBlock value: aNode value.
		 ^true].
	(aNode isVariable
	 and: [aNode name = #nil]) ifTrue:
		[aBlock value: nil.
		 ^true].
	aNode isSend ifTrue:
		[(self anyMethodNamed: aNode selector)
			ifNil:
				[(VMBasicConstants valueOfBasicSelector: aNode selector) ifNotNil:
					[:value|
					 aBlock value: value.
					 ^true].
				 aNode constantNumbericValueOrNil ifNotNil:
					[:value|
					 aBlock value: value.
					 ^true]]
			ifNotNil:
				[:m|
				(m statements size = 1
				 and: [m statements last isReturn]) ifTrue:
					[^self isConstantNode: m statements last expression valueInto: aBlock]]].
	^false! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 3/5/2009 14:18'!
isFloatingPointCType: aCType "<String>"
	^#('float' 'double' 'long double') includes: aCType asString! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'eem 5/31/2018 11:26'!
isFunctionalCType: aType
	"Answer if aType is either a function or a pointer to a function."
	^aType last == $) and: [(aType indexOf: $() > 0]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'ar 10/7/1998 17:53'!
isGeneratingPluginCode
	^false! !

!CCodeGenerator methodsFor: 'public' stamp: 'JMM 11/28/2002 11:52'!
isGlobalStructureBuild
	^false! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'dtl 7/27/2021 18:47'!
isIntegralCType: aCType "<String>"
	^(#('sqLong' 'usqLong' 'sqInt' 'usqInt' 'uint32_t' 'sqIntptr_t' 'usqIntptr_t'
		'long' 'long long' 'int' 'short' 'short int' 'char' 'signed char'
		'size_t' 'pid_t') includes: aCType asString)
	or: [(aCType beginsWith: 'unsigned') "Accept e.g. 'unsigned long' and also 'unsigned  : 8'"
		and: [(aCType includesAnyOf: '[*]') not]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 8/29/2020 10:43'!
isKernelMacroSelector: sel
	"Answer if the given selector is one of the selectors implemented as a macro in platform header fiels."

	^(self isKernelSelector: sel)
	 or: [VMBasicConstants mostBasicConstantSelectors includes: sel]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 11/17/2014 13:51'!
isKernelSelector: sel
	"Answer true if the given selector is one of the kernel selectors that are implemented as macros."

	^(#(error:
		 oopAt: oopAt:put: oopAtPointer: oopAtPointer:put:
		 byteAt: byteAt:put: byteAtPointer: byteAtPointer:put:
		 shortAt: shortAt:put: shortAtPointer: shortAtPointer:put:
		 intAt: intAt:put: intAtPointer: intAtPointer:put:
		 longAt: longAt:put: longAtPointer: longAtPointer:put:
		 long32At: long32At:put: long64At: long64At:put:
		 fetchFloatAt:into: storeFloatAt:from: fetchFloatAtPointer:into: storeFloatAtPointer:from:
		 fetchSingleFloatAt:into: storeSingleFloatAt:from: fetchSingleFloatAtPointer:into: storeSingleFloatAtPointer:from:
		 pointerForOop: oopForPointer:
		 cCoerce:to: cCoerceSimple:to:)
			includes: sel)! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 8/29/2020 12:25'!
isMacroSelector: sel
	"Answer if the given selector is one of the selectors implemented as a macro in platform header fiels."

	^(self isKernelSelector: sel)
	 or: [(VMBasicConstants mostBasicConstantSelectors includes: sel)
	 or: [(self methodNamed: sel)
			ifNil: [false]
			ifNotNil: [:m| m definedAsMacro]]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 12/9/2013 07:15'!
isNilConstantReceiverOf: sendNode
	"Answer true if the receiver of the given message send is the constant nil. Used to suppress conditional code when the condition is a translation-time constant."

	| val |
	generateDeadCode ifTrue: [^false].
	^(self isConstantNode: sendNode receiver valueInto: [:v| val := v])
	  and: [val isNil]! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'eem 12/8/2017 09:18'!
isNode: aNode constantValueWithinRangeOfType: aType
	"Answer if aNode is a constant value within the range of an integral type.
	 If aNode isn't an integral constant or aType isn't an integral type then answer false."
	| value |
	^aNode expression isConstant
	  and: [(self isIntegralCType: aType)
	  and: [(value := aNode expression value) isInteger
	  and: [value positive
	  and: [(aType first = $u
			ifTrue: [value digitLength]
			ifFalse: [value highBit + 7 // 8]) <= (self sizeOfIntegralCType: aType)]]]]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 2/9/2009 20:41'!
isNonArgumentImplicitReceiverVariableName: aString
	^(self typeOfVariable: aString) == #implicit
	    or: [vmClass
			ifNil: [#('interpreterProxy' 'self') includes: aString]
	 		ifNotNil: [vmClass isNonArgumentImplicitReceiverVariableName: aString]]! !

!CCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 5/17/2015 11:43'!
isObjectAccessor: selector given: interpreterClass
	^interpreterClass isObjectAccessor: selector! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'eem 12/16/2015 13:20'!
isPointerCType: aType
	^aType last == $*! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 11/15/2019 10:17'!
isQuickCompiledMethod: aCompiledMethod
	"Answer methods that returen a constant or an instance variable, or assign an instance variable."
	aCompiledMethod isQuick ifTrue:
		[^true].
	(self atMostThreeBytecodeMessagesFrom: aCompiledMethod) ifNotNil:
		[:messages| | s1 sz |
		(#(methodReturnReceiver methodReturnConstant:) includes: (s1 := messages first selector)) ifTrue:
			[^true].
		((sz := messages size) = 2
		 and: [(#(pushConstant: pushLiteralVariable:) includes: s1)
		 and: [messages second selector == #methodReturnTop]]) ifTrue:
			[^true].
		 (aCompiledMethod numArgs = 1
		  and: [sz = 3
		  and: [(s1 = #pushTemporaryVariable: and: [messages first argument = 0])
		  and: [(#(popIntoReceiverVariable: storeIntoReceiverVariable:) includes: messages second selector)
		  and: [#(methodReturnReceiver methodReturnTop) includes: messages third selector]]]]) ifTrue:
			[^true]].
	^false! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 2/23/2020 11:29'!
isSignedIntegralCType: aCType "<String>"
	self assert: (self isIntegralCType: aCType).
	^aCType first ~= $u! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'eem 12/16/2015 13:20'!
isSimpleType: aType
	"For the purposes of the read-before-written initializer, answer if
	 aType is simple, e.g. not a structure, and array or an opaque type."
	^(self isPointerCType: aType)
	  or: [self isIntegralCType: aType]! !

!CCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 5/17/2015 11:43'!
isStackAccessor: selector given: interpreterClass
	^interpreterClass isStackAccessor: selector! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 9/1/2017 15:12'!
isStructSend: aTSendNode
	"Answer if the argument aTSendNode is a send of a structure accessor.
	 This is tricky.  We want
		foo bar => foo->bar
		foo bar => foo.bar
		foo bar: expr => foo->bar = expr
		foo bar: expr => foo.bar = expr
	 depending on whether foo is a struct or a pointer to a struct,
	 but only if both foo is a struct type and bar is a field accessor.
	 The tricky cases are self-sends within struct class methods.  Here we need to
	 distinguish between self-sends of ordinary methods from self sends of accessors."
	^aTSendNode numArgs <= 1
	   and: [(aTSendNode receiver structTargetKindIn: self) notNil
	   and: [(self methodNamed: aTSendNode selector)
				ifNil: [false]
				ifNotNil: [:method| method isStructAccessor]]]! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 3/5/2009 13:15'!
isVoidPointer: aCType "<String>"
	"sigh..."
	^#('void *' 'void*') includes: aCType asString! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 1/31/2017 09:48'!
localizeGlobalVariables
	| candidates elected localized |

	"find all globals used in only one method"
	candidates := globalVariableUsage select: [:e | e size = 1].
	"Don't localize globals; nor those that are only assigned to; they're for debugging..."
	(candidates keys select: [:k| (vmClass mustBeGlobal: k)
								or: [(self methodNamed: (globalVariableUsage at: k) anyOne)
										ifNil: [false]
										ifNotNil: [:m| (m readsVariable: k) not]]]) do:
		[:k| candidates removeKey: k].
	
	elected := Set new.
	localized := Dictionary new. "for an ordered report"
	"move any suitable global to be local to the single method using it"
	candidates keysAndValuesDo:
		[:key :targets |
		targets do:
			[:name |
			(methods at: name ifAbsent: []) ifNotNil:
				[:procedure | | newDeclaration |
				(procedure isRealMethod
				 and: [self shouldGenerateMethod: procedure]) ifTrue:
					[(localized at: name ifAbsentPut: [SortedCollection new]) add: key.
					elected add: (procedure locals add: key).
					newDeclaration := variableDeclarations at: key ifAbsent: ['sqInt ', key].
					(self initializerForInstVar: key inStartClass: procedure definingClass) ifNotNil:
						[:initializerNode|
						newDeclaration := String streamContents:
												[:s|
												 s nextPutAll: newDeclaration; nextPutAll: ' = '.
												 initializerNode emitCCodeOn: s level: 0 generator: self]].
					procedure declarationAt: key put: newDeclaration.
					variableDeclarations removeKey: key ifAbsent: []]]]].
	logger ifNotNil:
		[localized keys asSortedCollection do:
			[:name|
			(localized at: name) do:
				[:var|
				logger ensureCr; show: var, ' localised to ', name; cr]]].
	elected do: [:ea| (variables includes: ea) ifTrue: [self checkDeleteVariable: ea]].
	variables removeAllFoundIn: elected! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 7/28/2008 11:51'!
logger
	^logger! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 7/28/2008 11:51'!
logger: aTranscriptStream
	logger := aTranscriptStream! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 7/30/2013 16:21'!
mayInline: sel
	"Answer if the method with the given selector may be inlined."

	^(self isAssertSelector: sel) not and: [inlineList includes: sel]! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 2/23/2015 16:42'!
maybeBreakForInlineOf: aNode in: aTMethod
	"convenient for debugging..."
	(aNode isSend
	and: [breakSrcInlineSelectors size + breakDestInlineSelectors size > 0
	and: [(breakSrcInlineSelectors isEmpty or: [breakSrcInlineSelectors includes: aNode selector])
	and: [(breakDestInlineSelectors isEmpty or: [(breakDestInlineSelectors includes: aTMethod selector)])
	and: [breakOnInline ~~ false]]]]) ifTrue:
		[aTMethod halt: aTMethod selector, ' ', aNode selector]! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 12/26/2016 14:47'!
maybeBreakForTestOfInliningOf: aNodeOrSelector
	"convenient for debugging..."
	| selector |
	selector := aNodeOrSelector isSymbol
					ifTrue: [aNodeOrSelector]
					ifFalse:
						[aNodeOrSelector isSend
							ifTrue: [aNodeOrSelector selector]
							ifFalse: [^self]].
	((breakSrcInlineSelectors includes: selector)
	 and: [breakDestInlineSelectors isEmpty
	 and: [breakOnInline == true]]) ifTrue:
		[self halt: selector]! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 2/23/2015 18:30'!
maybeBreakForTestToInline: aNodeOrSelector in: aTMethod
	"convenient for debugging..."
	| selector |
	selector := aNodeOrSelector isSymbol
					ifTrue: [aNodeOrSelector]
					ifFalse:
						[aNodeOrSelector isSend
							ifTrue: [aNodeOrSelector selector]
							ifFalse: [^self]].
	(breakSrcInlineSelectors size + breakDestInlineSelectors size > 0
	and: [(breakSrcInlineSelectors isEmpty or: [breakSrcInlineSelectors includes: selector])
	and: [(breakDestInlineSelectors isEmpty or: [(breakDestInlineSelectors includes: aTMethod selector)])
	and: [breakOnInline ~~ true]]]) ifTrue:
		[aTMethod halt: aTMethod selector, ' ', selector]! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 1/29/2020 11:39'!
maybeBreakOnInlineIn: aTMethod
	"convenient for debugging..."
	breakOnInline == true ifTrue:
		[aTMethod halt: aTMethod selector]! !

!CCodeGenerator methodsFor: 'C translation support' stamp: 'eem 4/19/2016 10:21'!
maybeEmitPrimitiveFailureDefineFor: selector on: aStream
	 (vmClass notNil
	 and: [(vmClass inheritsFrom: InterpreterPrimitives)
	 and: [(vmClass primitiveTable includes: selector)]]) ifTrue:
		[aStream nextPutAll: '#else\# define ' withCRs; nextPutAll: selector; nextPutAll: ' (void (*)(void))0\' withCRs]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 7/2/2020 15:32'!
maybePutPreambleFor: aClass on: aStream
	"Only emit preambleCCode if the method is implemented and it answers something non-nil."
	(aClass class whichClassIncludesSelector: #preambleCCode) ifNotNil:
		[:actualMetaclass |
		aClass preambleCCode ifNotNil:
			[:preambleCCode|
			 aStream
				ensureCr; cr;
				nextPutAll: '/* '; print: actualMetaclass; nextPutAll: '>>preambleCCode */'; cr;
				nextPutAll: preambleCCode; cr;
				nextPutAll: '/* end '; print: actualMetaclass; nextPutAll: '>>preambleCCode */'; cr]]! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'eem 12/16/2015 13:25'!
mergeTypeOf: var in: aDictionary with: newType method: tMethod
	"var is a variable that has been assigned an expression of type  newType.
	 Either assign its type, if it is as yet untyped, or merge newType with its existing type.
	 N.B. We refuse to promote a variable that already has integral type to a floating point
	 type.  The existing plugins depend on this; one can always use an explicit type in future."
	| existingType mergedType |
	existingType := self
						extractTypeFor: var
						fromDeclaration: (aDictionary at: var ifAbsentPut: [newType, ' ', var]).
	existingType ~= newType ifTrue:
		[((self isPointerCType: existingType)
		  or: [self isPointerCType: newType])
			ifTrue:
				[existingType = #'void *' ifTrue: [^newType].
				 newType = #'void *' ifTrue: [^existingType].
				 self logger show: 'conflicting types ', existingType, ' ', newType, ' for ', var, ' in ', tMethod selector.
				 ^existingType]
			ifFalse:
				[((self isIntegralCType: existingType)
				  and: [self isFloatingPointCType: newType]) ifFalse:
					[mergedType := self promoteArithmeticTypes: existingType and: newType.
					 aDictionary at: var put: mergedType, ' ', var]]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: ''!
methodNamed: selector
	"Answer the method in the code base with the given selector."

	^ methods at: selector ifAbsent: [ nil ]! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 5/20/2010 16:58'!
methodStatsString
	"Return a string describing the size, # of locals, and # of senders of
	 each method. Note methods that have inline C code or C declarations."

	| methodsWithCCode sizesOf callsOf s |
	methodsWithCCode := Set new: methods size.
	sizesOf := Dictionary new: methods size * 2.  "selector -> nodeCount"
	callsOf := Dictionary new: methods size * 2.  "selector -> senderCount"

	"For each method, scan its parse tree once to:
		1. determine if the method contains C code or declarations
		2. determine how many nodes it has
		3. increment the sender counts of the methods it calls
		4. determine if it includes any C declarations or code"

	methods do: [ :m | | hasCCode nodeCount |
		(translationDict includesKey: m selector) ifTrue: [
			hasCCode := true.
		] ifFalse: [
			hasCCode := m declarations size > 0.
			nodeCount := 0.
			m parseTree nodesDo: [ :node | | selr senderCount |
				node isSend ifTrue: [
					selr := node selector.
					node isNonNullCCode ifTrue: [ hasCCode := true ].
					senderCount := callsOf at: selr ifAbsent: [ 0 ].
					callsOf at: selr put: senderCount + 1.
				].
				nodeCount := nodeCount + 1.
			].
		].
		hasCCode ifTrue: [ methodsWithCCode add: m selector ].
		sizesOf at: m selector put: nodeCount.
	].

	s := WriteStream on: (String new: 5000).
	(self sortStrings: methods keys) do: [ :sel | | m calls registers |
		m := methods at: sel.
		registers := m locals size + m args size.
		calls := callsOf at: sel ifAbsent: [0].
		registers > 11 ifTrue: [
			s nextPutAll: sel; tab.
			s nextPutAll: (sizesOf at: sel) printString; tab.
			s nextPutAll: calls printString; tab.
			s nextPutAll: registers printString; tab.
			(methodsWithCCode includes: sel) ifTrue: [ s nextPutAll: 'CCode' ].
		s cr.
		].
	].
	^ s contents! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 6/6/2018 19:47'!
methodsForDefiningClass: dc
	"Answer a collection of methods that refer to the given global variable."

	^methods select: [:tMethod| tMethod definingClass == dc]! !

!CCodeGenerator methodsFor: 'utilities' stamp: ''!
methodsReferringToGlobal: v
	"Return a collection of methods that refer to the given global variable."

	| out |
	out := OrderedCollection new.
	methods associationsDo: [ :assoc |
		(assoc value freeVariableReferences includes: v) ifTrue: [
			out add: assoc key.
		].
	].
	^ out! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 6/11/2008 16:27'!
methodsThatCanInvoke: aSelectorList
	"Return a set of methods that can invoke one of the given selectors, either directly or via a sequence of intermediate methods."

	| out todo sel |
	out := Set new.
	todo := aSelectorList copy asOrderedCollection.
	[todo isEmpty] whileFalse: [
		sel := todo removeFirst.
		out add: sel.
		methods do: [ :m | | mSelector |
			(m allCalls includes: sel) ifTrue: [
				mSelector := m selector.
				((out includes: mSelector) or:
				 [todo includes: mSelector]) ifFalse: [
					todo add: mSelector.
				].
			].
		].
	].
	^ out
	! !

!CCodeGenerator methodsFor: 'C translation support' stamp: 'eem 7/19/2019 19:17'!
mustBeGlobal: varName
	"Answer if a variable must be global and exported.  Used for inst vars that are accessed from VM support code."

	^vmClass notNil and: [vmClass mustBeGlobal: varName]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 8/7/2014 08:10'!
needToGenerateHeader: headerName file: interpHdrPath contents: newContentsArg
	"Check if we need to regenerate a header file.  We always need to if the contents have changed.
	 But if not we can avoid needless recompilations by not regenerating.  So only regenerate if the
	 package stamp is dirty and the monticello stamp is clean."
	| newContents oldContents dirtyStamp |
	(FileDirectory default fileExists: interpHdrPath) ifFalse:
		[^true].
	newContents := newContentsArg.
	oldContents := (FileDirectory default oldFileNamed: interpHdrPath) contentsOfEntireFile.
	oldContents := oldContents copyReplaceAll: {Character cr. Character lf} with: {Character cr}.
	oldContents replaceAll: Character lf with: Character cr.
	dirtyStamp := (oldContents indexOfSubCollection: '* VMMaker') ~= 0.
	(newContents beginsWith: '/*') = (oldContents beginsWith: '/*') ifFalse:
		[(newContents beginsWith: '/*') ifTrue:
			[newContents := newContents readStream upToAll: '*/'; skipSeparators; upToEnd].
		 (oldContents beginsWith: '/*') ifTrue:
			[oldContents := oldContents readStream upToAll: '*/'; skipSeparators; upToEnd]].
	^oldContents ~= newContents
	 or: [(dirtyStamp and: [[((self class monticelloDescriptionFor: vmClass) includes: $*) not]
							on: Error do: [:ex| false]]) "If no stamp don't worry"
	 or: [self confirm: headerName, ' contents are unchanged.\Writing the file may cause recompilation of support files.\Do you want to write the header file?' withCRs]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 3/13/2019 12:21'!
nilOrBooleanConditionFor: nodeOrNil
	"If nodeOrNil is one of the conditional sends for which we do translation-time dead code elimination
	 (i.e. cppIf:ifTrue: et al or ifTrue: et al) and the conditional does evaluate to a translation-time
	 boolean constant, answer that constant, otherwise answer nil.  Used to prune dead code,
	 either for code generaton or dead variable elimination."
	generateDeadCode ifTrue: [^nil].
	nodeOrNil ifNil:
		[^nil].
	nodeOrNil isSend ifFalse:
		[^nil].
	(#(ifTrue: ifFalse: ifTrue:ifFalse: ifFalse:ifTrue:) includes: nodeOrNil selector) ifTrue:
		[^self nilOrBooleanConstantReceiverOf: nodeOrNil receiver].
	(#(and: or:) includes: nodeOrNil selector) ifTrue:
		[^self nilOrBooleanConstantReceiverOf: nodeOrNil].
	(#(cppIf:ifTrue: cppIf:ifTrue:ifFalse:) includes: nodeOrNil selector) ifTrue:
		[| maybeName value |
		(vmClass notNil
		 and: [nodeOrNil args first isConstant
		 and: [(value := nodeOrNil args first value) isSymbol
		 and: [((vmClass ifNil: [VMBasicConstants]) defineAtCompileTime: value) not
		 and: [(vmClass bindingOf: value) notNil]]]]) ifTrue:
			[self logger
					nextPutAll: 'Warning: cppIf: reference to ';
					store: value;
					nextPutAll: ' when variable of same name exists.'; cr].

		maybeName := nodeOrNil args first isConstant ifTrue:
							[nodeOrNil args first nameOrValue].
		 ^(optionsDictionary notNil
		   and: [nodeOrNil args first isConstant
		   and: [(#(true false) includes: (optionsDictionary at: maybeName ifAbsent: [nil]))
		   and: [((vmClass ifNil: [VMBasicConstants]) defineAtCompileTime: maybeName) not]]]) ifTrue:
			[optionsDictionary at: nodeOrNil args first name]].
	^nil! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 8/9/2017 13:28'!
nilOrBooleanConstantReceiverOf: aNode
	"Answer nil or the boolean constant that is the receiver of the given message send.
	 Used to suppress conditional code when the condition is a translation-time constant."

	| val receiver argument arms |
	generateDeadCode ifTrue:[^nil].
	((self isConstantNode: aNode valueInto: [:v| val := v])
	 and: [#(true false) includes: val]) ifTrue:
		[^val].
	aNode isSend ifTrue:
		[aNode selector == #not ifTrue:
			[(self nilOrBooleanConstantReceiverOf: aNode receiver) ifNotNil:
				[:bool| ^bool not]].
		 ((#(isNil notNil) includes: aNode selector)
		  and: [self isNilConstantReceiverOf: aNode]) ifTrue:
			[^aNode selector == #isNil].
		 ((#(or: and:) includes: aNode selector)
		 and: [aNode args last isStmtList
		 and: [aNode args last statements size = 1]]) ifTrue:
			[(self nilOrBooleanConstantReceiverOf: aNode receiver) ifNotNil:
				[:rcvr|
				((rcvr == false and: [aNode selector == #and:])
				 or: [rcvr == true and: [aNode selector == #or:]]) ifTrue:
					[^rcvr].
				(self nilOrBooleanConstantReceiverOf: aNode args last statements first) ifNotNil:
					[:arg|
					^rcvr perform: aNode selector with: [arg]]].
			 "We can also eliminate expr and: [false], expr or: [true], but only if expr is side-effect free.
			  This is a weak test; we don't traverse calls.  Caveat emptor!!"
			 ((aNode receiver noneSatisfy: [:node| node isAssignment]) "No side-effects in the elided expression"
			  and: [aNode args last statements size = 1]) ifTrue:
				[(self nilOrBooleanConstantReceiverOf: aNode args last statements first) ifNotNil:
					[:arg|
					((arg == false and: [aNode selector == #and:])
					 or: [arg == true and: [aNode selector == #or:]]) ifTrue:
						[^arg]]]].
		"Look for Const ifTrue: [self foo] ifFalse: [false] => false"
		 ((#(ifTrue:ifFalse: ifFalse:ifTrue:) includes: aNode selector)
		  and: [(self isConstantNode: aNode receiver valueInto: [:v| val := v])
		  and: [(#(true false) includes: val)
		  and: [arms := aNode args collect:
							[:altBlock| | bval |
							 (altBlock statements size = 1
							 and: [(self isConstantNode: altBlock statements last valueInto: [:v| bval := v])
							 and: [#(true false) includes: bval]]) ifTrue:
								[bval]].
				arms asArray ~= #(nil nil)]]]) ifTrue:
			[| arm |
			 arm := aNode selector == #ifTrue:ifFalse: == val
						ifTrue: [arms first]
						ifFalse: [arms last].
			 (#(true false) includes: arm) ifTrue:
				[^arm]].
		 ((#(= ~= < > <= >=) includes: aNode selector)
		  and: [(self isConstantNode: aNode receiver valueInto: [:v| receiver := v])
		  and: [receiver isInteger
		  and: [(self isConstantNode: aNode args first valueInto: [:v| argument := v])
		  and: [argument isInteger]]]]) ifTrue:
			[^receiver perform: aNode selector with: argument].
		 "Inlining for e.g. CharacterTable ifNil: [...] ifNotNil: [...]], which compiles to CharacterTable == nil ifTrue: [...] ifFalse: [...]"
		(aNode selector == #==
		 and: [aNode args first isVariable
		 and: [aNode args first name = 'nil'
		 and: [aNode receiver isConstant
		 and: [aNode receiver value == nil]]]]) ifTrue:
			[^true]].
	^nil! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 8/29/2020 17:07'!
node: exprNode typeCompatibleWith: argName inliningInto: targetMethod in: aTMethod
	"Answer either exprNode or, if required, a cast of exprNode to the type of argName.
	 The cast is required if
		- argName is typed and exprNode is untyped
		- argName is untyped and exprNode is an arithmetic type of size > #sqInt
		- both argName and exprNode are typed but they are incompatible"
	| formalType actualType |
	formalType := targetMethod typeFor: argName in: self.
	actualType := self typeFor: exprNode in: aTMethod.
	"First check for inlining a struct method/accessor"
	((argName beginsWith: 'self_in_')
	 and: [formalType last == $*
	 and: [(formalType beginsWith: actualType)
	 and: [(formalType allButFirst: actualType size) withBlanksTrimmed = '*']]]) ifTrue:
		[^TSendNode new
			setSelector: #addressOf:
			receiver: (TVariableNode new setName: 'self')
			arguments: { exprNode }
			isBuiltInOp: false].
	"Second check for arithmetic coercion"
	((exprNode isSend or: [exprNode isVariable])
	   and: [(formalType notNil and: [actualType isNil])
			or: [(formalType isNil and: [actualType notNil and: [(self isIntegralCType: actualType) and: [(self sizeOfIntegralCType: actualType) > (self sizeOfIntegralCType: #sqInt)]]])
			or: [(self variableOfType: formalType acceptsValue: exprNode ofType: actualType) not]]])
		ifTrue:
			[^self nodeToCast: exprNode to: (formalType ifNil: [#sqInt])].
	"Next check for type errors..."
	((exprNode isSend or: [exprNode isVariable])
	  and: [(self
			variableOfType: (targetMethod typeFor: argName in: self)
			acceptsValue: exprNode
			ofType: (self typeFor: exprNode in: aTMethod)) not]) ifTrue:
		[logger
			nextPutAll:
				'type mismatch for formal ', argName, ' and actual "', exprNode asString,
				'" when inlining ', targetMethod selector, ' in ', aTMethod selector, '. Use a cast.';
			cr; flush].
	"No conversion is necessary, or there's a type error..."
	^exprNode! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 7/31/2020 12:44'!
nodeIsDeadCode: aNode withParent: parentNode
	"Answer if aNode would not be generated due to dead code elimination."
	(aNode isLiteralBlock and: [parentNode isSend and: [parentNode selector == #addressOf:put:]]) ifTrue:
		[^true].
	^(self nilOrBooleanConditionFor: parentNode)
		ifNil: [false]
		ifNotNil:
			[:cond| | filter |
			filter := parentNode selector caseOf:
							{   "First element is accessor for filtered (eliminated) node if expression is true.
								Second element is accessor for filtered (eliminated) node if expression is false."
								[#ifFalse:]				-> [#(first nil)].
								[#ifFalse:ifTrue:] 		-> [#(first last)].
								[#ifTrue:]				-> [#(nil first)].
								[#ifTrue:ifFalse:]			-> [#(last first)].
								[#and:]					-> [#(nil first)].
								[#or:]					-> [#(last nil)].
								[#cppIf:ifTrue:]			-> [#(nil #second)].
								[#cppIf:ifTrue:ifFalse:]	-> [#(third #second)] }.
			(cond ifTrue: [filter first] ifFalse: [filter last])
				ifNil: [false]
				ifNotNil: [:accessor| aNode == (parentNode args perform: accessor)]]! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 2/23/2015 10:09'!
nodeToCast: exprNode to: cType
	^TSendNode new
		setSelector: #cCoerceSimple:to:
		receiver: (TVariableNode new setName: 'self')
		arguments: { exprNode. TConstantNode new setValue: cType }
		isBuiltInOp: true! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 9/9/2018 18:36'!
nonStructClassesForTranslationClasses: classes
	"Answer in superclass order (any superclass precedes any subclass)
	 the ancilliaryClasses that are not struct classes for all the given classes."
	| nonStructClasses |
	nonStructClasses := OrderedCollection new.
	classes do:
		[:aTranslationClass|
		([aTranslationClass ancilliaryClasses]
				on: MessageNotUnderstood
				do: [:ex|
					ex message selector == #ancilliaryClasses
						ifTrue: [#()]
						ifFalse: [ex pass]]) do:
			[:class|
			(vmClass isNil or: [vmClass isAcceptableAncilliaryClass: class]) ifTrue:
				[(class isStructClass
				 or: [(nonStructClasses includes: class)
				 or: [classes includes: class]]) ifFalse:
					[nonStructClasses addLast: class]]]].
	^self superclassOrder: nonStructClasses! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 12/8/2017 15:53'!
noteUsedVariableName: variableName
	currentMethod ifNotNil:
		[:m| m noteUsedVariableName: variableName]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 12/8/2017 13:05'!
noteVariableUsageInString: aString
	currentMethod ifNotNil:
		[:m|
		 (Scanner new typedScanTokens: aString) do:
			[:token|
			 (token isString and: [token first isLetter or: [token first == $_]]) ifTrue:
				[m noteUsedVariableName: token]]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 9/5/2021 07:09'!
optionIsFalse: key in: aClass
	"Answer whether a notOption: is false in the context of aClass. The key either a
	 Cogit class name or a class variable name or a variable name in VMBasicConstants."

	"If the option is the name of a subclass of Cogit, include it if it doesn't inherit from the Cogit class."
	(Smalltalk classNamed: key) ifNotNil:
		[:optionClass|
		 vmClass hasCogit
			ifTrue:
				[aClass cogitClass
					ifNotNil:
						[:cogitClass|
						 (optionClass includesBehavior: Cogit) ifTrue:
							[^(cogitClass includesBehavior: optionClass) not]]]
			ifFalse:
				[(optionClass includesBehavior: Cogit) ifTrue:
					[^true]].
		 aClass objectMemoryClass ifNotNil:
			[:objectMemoryClass|
			 ((optionClass includesBehavior: ObjectMemory)
			   or: [optionClass includesBehavior: SpurMemoryManager]) ifTrue:
				[^(objectMemoryClass includesBehavior: optionClass) not]]].
	"Lookup options in options, class variables of the defining class, VMBasicConstants, the interpreterClass and the objectMemoryClass"
	{aClass initializationOptions.
	  aClass.
	  VMBasicConstants.
	  aClass interpreterClass.
	  aClass objectMemoryClass} do:
		[:scopeOrNil|
		 scopeOrNil ifNotNil:
			[:scope|
			 (scope bindingOf: key) ifNotNil:
				[:binding|
				binding value ~~ true ifTrue: [^true]]]].
	^false! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 9/5/2021 07:08'!
optionIsTrue: key in: aClass
	"Answer whether an option: is true in the context of aClass. The key either a
	 Cogit class name or a class variable name or a variable name in VMBasicConstants."

	"If the option is the name of a subclass of Cogit, include it if it inherits from the Cogit class."
	(Smalltalk classNamed: key) ifNotNil:
		[:optionClass|
		 vmClass hasCogit
			ifTrue:
				[aClass cogitClass
					ifNotNil:
						[:cogitClass|
						 (optionClass includesBehavior: Cogit) ifTrue:
							[^cogitClass includesBehavior: optionClass]]]
			ifFalse:
				[(optionClass includesBehavior: Cogit) ifTrue:
					[^false]].
		 aClass objectMemoryClass ifNotNil:
			[:objectMemoryClass|
			 ((optionClass includesBehavior: ObjectMemory)
			   or: [optionClass includesBehavior: SpurMemoryManager]) ifTrue:
				[^objectMemoryClass includesBehavior: optionClass]]].
	"Lookup options in options, class variables of the defining class, VMBasicConstants, the interpreterClass and the objectMemoryClass"
	{aClass initializationOptions.
	  aClass.
	  VMBasicConstants.
	  aClass interpreterClass.
	  aClass objectMemoryClass} do:
		[:scopeOrNil|
		 scopeOrNil ifNotNil:
			[:scope|
			 (scope bindingOf: key) ifNotNil:
				[:binding|
				binding value ~~ false ifTrue: [^true]]]].
	^false! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 8/14/2012 15:07'!
options
	^optionsDictionary! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 11/19/2012 17:39'!
options: aDictionary
	"Set optionsDictionary to the argument and scan it for accessors on the receiver,
	 performing the accessor with the value, allowing one to specify things like
	 generateDeadCode: false in the options."
	optionsDictionary := aDictionary.
	optionsDictionary keysAndValuesDo:
		[:k :v| | accessor |
		((self class instVarNames includes: k)
		and: [(Symbol hasInterned: k, ':' ifTrue: [:s| accessor := s])
		and: [self class canUnderstand: accessor]]) ifTrue:
			[self perform: accessor with: v]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 11/10/2014 13:24'!
outputAsmLabel: selector on: aStream
	| count |
	suppressAsmLabels ifTrue: [^self].
	asmLabelCounts ifNil:
		[asmLabelCounts := Dictionary new].
	count := asmLabelCounts
				at: selector
				put: 1 + (asmLabelCounts at: selector ifAbsent: [-1]).
	 aStream
		nextPutAll: 'VM_LABEL(';
		nextPutAll: (self cFunctionNameFor: selector);
		nextPutAll: (count = 0 ifTrue: [''] ifFalse: [count printString]);
		nextPut: $);
		nextPut: $;! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 12/5/2012 20:36'!
preDeclareInterpreterProxyOn: aStream
	"Put the necessary #defines needed before interpreterProxy.  Basically
	 internal plugins use the VM's interpreterProxy variable and external plugins use
	 their own.  In addition the VMPluginCodeGenerator can choose to keep local copies
	 all functions."
	aStream cr; nextPutAll: '#ifdef SQUEAK_BUILTIN_PLUGIN'.
	aStream cr; nextPutAll: 'extern'.
	aStream cr; nextPutAll: '#endif'; cr! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 5/22/2014 13:52'!
prepareMethods
	| globals |
	globals := Set new: 200.
	globals addAll: variables.
	methods do:
		[:m |
		m locals, m args do:
			[:var |
			(globals includes: var) ifTrue:
				[self error: 'Local variable ''', var, ''' may mask global when inlining ', m selector].
			((methods at: var ifAbsent: [nil]) ifNil: [false] ifNotNil: [:m1| m1 isStructAccessor not]) ifTrue:
				[logger
					ensureCr;
					nextPutAll: 'Local variable name ''', var, ''' in ';
					nextPutAll: m selector;
					nextPutAll: ' may mask method when inlining';
					cr]].
		m bindClassVariablesIn: constants.
		m prepareMethodIn: self]! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 12/16/2015 10:21'!
previousCommentMarksInlining
	^previousCommentMarksInlining! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 12/16/2015 10:21'!
previousCommentMarksInlining: aBoolean
	| previousValue |
	previousValue := previousCommentMarksInlining.
	previousCommentMarksInlining := aBoolean.
	^previousValue! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 12/16/2015 10:26'!
previousCommenter
	^previousCommenter! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 12/16/2015 10:29'!
previousCommenter: aTParseNodeOrNil
	| previousValue |
	previousValue := previousCommenter.
	previousCommenter := aTParseNodeOrNil.
	^previousValue! !

!CCodeGenerator methodsFor: 'error notification' stamp: 'eem 5/20/2010 16:59'!
printUnboundCallWarnings
	"Print a warning message for every unbound method call in the code base."

	| knownSelectors undefinedCalls |
	undefinedCalls := Dictionary new.
	knownSelectors := translationDict keys asSet.
	knownSelectors add: #error:.
	methods do: [ :m | knownSelectors add: m selector ].
	methods do: [ :m |
		m allCalls do: [ :sel |
			(knownSelectors includes: sel) ifFalse: [
				(undefinedCalls includesKey: sel)
					ifTrue: [ (undefinedCalls at: sel) add: m selector ]
					ifFalse: [ undefinedCalls at: sel put: (OrderedCollection with: m selector) ].
			].
		].
	].

	logger cr.
	(self sortStrings: undefinedCalls keys) do: [ :undefined |
		logger show: undefined, ' -- undefined method sent by:'; cr.
		(undefinedCalls at: undefined) do: [ :caller |
			logger tab; show: caller; cr.
		].
	].! !

!CCodeGenerator methodsFor: 'error notification' stamp: 'eem 5/20/2010 16:59'!
printUnboundVariableReferenceWarnings
	"Print a warning message for every unbound variable reference in the code base."

	| undefinedRefs globalVars |
	undefinedRefs := Dictionary new.
	globalVars := Set new: 100.
	globalVars addAll: variables.
	methods do: [ :m | | knownVars |
		knownVars := globalVars copy.
		m args do: [ :var | knownVars add: var ].
		m locals do: [ :var | knownVars add: var ].
		m freeVariableReferences do: [ :varName |
			(knownVars includes: varName) ifFalse: [
				(undefinedRefs includesKey: varName)
					ifTrue: [ (undefinedRefs at: varName) add: m selector ]
					ifFalse: [ undefinedRefs at: varName put: (OrderedCollection with: m selector) ].
			].
		].
	].

	logger cr.
	(self sortStrings: undefinedRefs keys) do: [ :var |
		logger show: var, ' -- undefined variable used in:'; cr.
		(undefinedRefs at: var) do: [ :sel |
			logger tab; show: sel; cr.
		].
	].! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'nice 4/24/2016 10:07'!
promoteArithmeticTypes: firstType and: secondType
	"Answer the return type for an arithmetic send.  This is so that the inliner can still inline
	 simple expressions.  Deal with pointer arithmetic, floating point arithmetic and promotion.
	 It is important to choose deterministically to get stable source generation.
	 Also, the behaviour of inlined and non inlined code should be unchanged."
	((#(#double float) includes: firstType)
	 or: [#(#double float) includes: secondType]) ifTrue:
		[^(firstType = #double or: [secondType = #double])
			ifTrue: [#double]
			ifFalse: [#float] "in C99 6.3.1.8, float+int, int is converted as a float"].
	"deal with unknowns, answering nil."
	(firstType isNil or: [secondType isNil]) ifTrue:
		[^nil].
	"Deal with integer promotion and arithmetic conversion"
	^self promoteIntegerArithmeticTypes: firstType and: secondType! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'eem 7/21/2017 12:16'!
promoteIntegerArithmeticTypes: firstType and: secondType
	"Answer the return type for an arithmetic send.
	Deal with integer promotion rules of C99.
	See section 6.3 Conversions of the standard.
	
	6.3.1.1 ...snip...
		If an int can represent all values of the original type, the value is converted to an int;
		otherwise, it is converted to an unsigned int. These are called the integer promotions.
		All other types are unchanged by the integer promotions

	6.3.1.8 ...snip...
		Otherwise, the integer promotions are performed on both operands
		Then the following rules are applied to the promoted operands:

			If both operands have the same type, then no further conversion is needed.

			Otherwise, if both operands have signed integer types or both have unsigned integer
			types, the operand with the type of lesser integer conversion rank is converted to the
			type of the operand with greater rank.

			Otherwise, if the operand that has unsigned integer type has rank greater or equal to
			the rank of the type of the other operand, then the operand with signed integer type
			is converted to the type of the operand with unsigned integer type.

			Otherwise, if the type of the operand with signed integer type can represent all of the
			values of the type of the operand with unsigned integer type, then the operand with
			unsigned integer type is converted to the type of the operand with signed integer type.

			Otherwise, both operands are converted to the unsigned integer type corresponding to
			the type of the operand with signed integer type.

	This is so that the generated code behaviour is insensitive to inlining."
	| length1 length2 intSize |
	length1 := self sizeOfIntegralCType: firstType.
	length2 := self sizeOfIntegralCType: secondType.
	intSize := self sizeOfIntegralCType: #int.
	(length1 < intSize and: [length2 < intSize]) ifTrue: [^#int].	"Integer promotion"
	length1 > length2 ifTrue: [^firstType].
	length2 > length1 ifTrue: [^secondType].
	firstType first = $u ifTrue: [^firstType].
	secondType first = $u ifTrue: [^secondType].
	^firstType! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 2/11/2009 22:33'!
pruneMethods: selectorList
	"Explicitly prune some methods"
	selectorList do: [:sel| self removeMethodForSelector: sel]! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 12/26/2016 16:08'!
pruneUnreachableMethods
	"Remove any methods that are not reachable. Retain methods needed by the translated classes - see implementors of requiredMethodNames"
 	
	| neededSelectors newMethods previousSize visited |
	"add all the exported methods and all the called methods to the requiredSelectors"
	"keep all the fake methods (macros and struct accessors; these are needed
	 to ensure correct code generation."

	neededSelectors := Set withAll: requiredSelectors.
	methods do: [ :m |
		m export ifTrue:
			[neededSelectors add: m selector].
		m isAPIMethod ifTrue:
			[neededSelectors add: m selector].
		m isRealMethod ifFalse:
			[neededSelectors add: m selector]].

	"Now compute the transitive closure..."
	previousSize := neededSelectors size.
	visited := IdentitySet new: methods size.
	[neededSelectors do:
		[:s|
		(methods at: s ifAbsent: []) ifNotNil:
			[:m|
			(visited includes: m) ifFalse:
				[visited add: m.
				 (m isRealMethod
				  and: [self shouldGenerateMethod: m]) ifTrue:
					[neededSelectors addAll: m allCalls]]]].
	 neededSelectors size > previousSize]
		whileTrue:
			[previousSize := neededSelectors size].

	"build a new dictionary of methods from the collection of all the ones to keep"			
	newMethods := Dictionary new: neededSelectors size.
	neededSelectors do:
		[:sel|
		methods at: sel ifPresent: [:meth| newMethods at: sel put: meth]].
	methods := newMethods! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 7/2/2008 14:21'!
pushScope: variableToType "<Dictionary>" while: aBlock
	scopeStack addLast: variableToType.
	^aBlock ensure: [scopeStack removeLast]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 3/16/2018 19:50'!
putConditionalDefineOf: aConstantNameString as: valueOrValueString comment: commentOrNil on: aStream
	self withConditionalDefineOf: aConstantNameString
		comment: commentOrNil
		on: aStream
		do: [| valueAsString |
			 valueAsString := valueOrValueString asString.
			 valueAsString first ~= $# ifTrue:
				[aStream nextPutAll: '# define '; nextPutAll: aConstantNameString; space].
			 aStream nextPutAll: valueAsString; cr]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 3/16/2018 23:10'!
putDefineOf: aConstantNameString as: valueOrValueString on: aStream
	aStream
		nextPutAll: '#define ';
		nextPutAll: aConstantNameString;
		space;
		nextPutAll: valueOrValueString asString;
		cr! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 01/19/2021 20:40'!
removeAssertions
	"Remove all assertions in method bodies.  This is for the benefit of inlining, which
	fails to recognise and disregard empty method bodies when checking the inlinability
	of sends."

	| newMethods |
	newMethods := Dictionary new.
	'Removing assertions...'
		displayProgressAt: Sensor cursorPoint
		from: 0 to: methods size
		during: [ :bar |
			methods withIndexDo: [ :m :i |
				bar value: i.
				m isAssertion ifFalse: [
					newMethods at: m selector put: m.
					m removeAssertions]]].
	methods := newMethods.! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 8/16/2008 09:56'!
removeConstant: aName
	"Remove the given (class) variable from the code base."

	constants removeKey:  aName ifAbsent: []! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 2/23/2015 16:43'!
removeMethodForSelector: aSelector
	"Remove the given method from the code base"
	((breakSrcInlineSelectors includes: aSelector)
	 or: [(breakDestInlineSelectors includes: aSelector)]) ifTrue:
		[self halt].
	methods removeKey:  aSelector ifAbsent: []! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 2/11/2009 22:33'!
removeMethodsReferingToGlobals: varList except: methodName
	"Remove any methods (presumably inlined) that still contain references to the given
	 global variables. Answer a Dictionary of the removed methods for later reporting."

	| varListAsStrings removed |
	varListAsStrings := varList collect: [ :sym | sym asString ].
	removed := Dictionary new: methods size / 2.
	methods copy keysAndValuesDo:
		[ :sel :meth| | mVars |
		sel ~= methodName ifTrue:
			[mVars := meth freeVariableReferences asSet.
			 (varListAsStrings anySatisfy: [ :v | mVars includes: v]) ifTrue:
				[removed at: sel put: meth.
				self removeMethodForSelector: sel]]].

	^removed! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 4/12/2012 10:10'!
removeVariable: aName
	"Remove the given (instance) variable from the code base."
	self removeVariable: aName
		ifAbsent:
			[(vmClass notNil
			  and: [vmClass isNonArgumentImplicitReceiverVariableName: aName]) ifFalse:
				[logger
					ensureCr;
					show: 'warning, variable '
						, aName
						, ' doesn''t exist or has already been removed';
						cr]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 4/12/2012 10:08'!
removeVariable: aName ifAbsent: ifAbsentBlock
	"Remove the given (instance) variable from the code base."
	self checkDeleteVariable: aName.
	variableDeclarations removeKey: aName ifAbsent: [].
	^variables remove: aName ifAbsent: ifAbsentBlock! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 10/22/2010 11:33'!
reportRecursiveMethods
	"Report in transcript all methods that can call themselves directly or indirectly or via a chain of N intermediate methods."

	methods do: [: m | | visited calls newCalls sel called |
		visited := translationDict keys asSet.
		calls := m allCalls asOrderedCollection.
		5 timesRepeat: [
			newCalls := Set new: 50.
			[calls isEmpty] whileFalse: [
				sel := calls removeFirst.
				sel = m selector ifTrue: [
					logger ensureCr; show: m selector, ' is recursive'; cr.
				] ifFalse: [
					(visited includes: sel) ifFalse: [
						called := self methodNamed: sel.
						called = nil ifFalse: [ newCalls addAll: called allCalls ].
					].
					visited add: sel.
				].
			].
			calls := newCalls asOrderedCollection.
		].
	].! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 6/26/2012 12:54'!
reportShouldNotBeRemoved: removed varList: varList
	"Report whether any of the removed methods are still used."
	| varListAsStrings shouldNotBeRemoved |
	varListAsStrings := varList collect: [ :sym | sym asString ].
	shouldNotBeRemoved := Set new.
	removed do:
		[:m|
		m isAPIMethod ifTrue:
			[shouldNotBeRemoved add: m selector]].
	methods do:
		[:m|
		(m selector = #interpret
		 or: [removed includes: m selector]) ifFalse:
			[m allCalls do:
				[:sel|
				(removed includesKey: sel) ifTrue:
					[shouldNotBeRemoved add: sel]]]].
	vmClass additionalSelectorTables do:
		[:selectorTable|
		 selectorTable do:
			[:selOrInteger|
			selOrInteger isInteger ifFalse:
				[(removed includesKey: selOrInteger) ifTrue:
					[shouldNotBeRemoved add: selOrInteger]]]].
	shouldNotBeRemoved do:
		[:sel| | str |
		str := String streamContents:
				[:strm| | them |
				strm
					nextPutAll: 'Removed ';
					nextPutAll: sel;
					nextPutAll: ' because it refers to the local variable'.
				them := (removed at: sel) freeVariableReferences asSet intersection: varListAsStrings.
				them size > 1 ifTrue:
					[strm nextPut: $s.
					 them := self sortStrings: them].
				them do: [:var| strm space; nextPutAll: var].

				strm
					nextPutAll: ' of interpret.';
					cr;
					nextPutAll: 'But it is either used outside of interpret or exported!!!!';
					cr].
		logger ensureCr; show: str.
		self inform: str]! !

!CCodeGenerator methodsFor: 'C translation support' stamp: 'eem 11/20/2012 11:27'!
reservedWords
	^#(	'auto'
		'break'
		'case' 'char' 'const' 'continue'
		'default' 'do' 'double'
		'else' "'entry' obsolete" 'enum' 'extern'
		'float' 'for'
		'goto'
		'if' 'int'
		'long'
		'register' 'return'
		'short' 'signed' 'sizeof' 'static' 'struct' 'switch'
		'typedef'
		'union' 'unsigned'
		'void' 'volatile'
		'while')! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 6/2/2011 12:27'!
retainMethods: aListOfSelectorsToKeep
	"add aListOfSelectorsToKeep to requiredSelectors so that they will not be pruned"
	requiredSelectors ifNil:[requiredSelectors := Set new:100].
	requiredSelectors addAll: aListOfSelectorsToKeep.
	^aListOfSelectorsToKeep! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'JMM 4/16/2002 22:39'!
returnPrefixFromVariable: aName
	^aName! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'eem 2/23/2020 11:30'!
returnTypeForSend: sendNode in: aTMethod ifNil: typeIfNil
	"Answer the return type for a send.  Unbound sends default to typeIfNil.
	 Methods with types as yet unknown have a type determined either by the
	 kernelReturnTypes or the table below, or, if they are in neither set, then nil.
	 The inferred type should match as closely as possible the C type of
	 generated expessions so that inlining would not change the expression.
	 If there is a method for sel but its return type is as yet unknown it mustn't
	 be defaulted, since on a subsequent pass its type may be computable."
	| sel methodOrNil |
	methodOrNil := self anyMethodNamed: (sel := sendNode selector).
	(methodOrNil notNil and: [methodOrNil returnType notNil]) ifTrue:
		[^self baseTypeForType: methodOrNil returnType].
	^kernelReturnTypes
		at: sel
		ifAbsent:
			[sel
				caseOf: {
				[#integerValueOf:]		->	[#sqInt].
				[#isIntegerObject:]		->	[#int].
				[#negated]				->	[self promoteArithmeticTypes: (sendNode receiver typeFrom: self in: aTMethod) and: #int].
				[#+]					->	[self typeForArithmetic: sendNode in: aTMethod].
				[#-]					->	[self typeForArithmetic: sendNode in: aTMethod].
				[#*]					->	[self typeForArithmetic: sendNode in: aTMethod].
				[#/]						->	[self typeForArithmetic: sendNode in: aTMethod].
				[#//]					->	[self typeForArithmetic: sendNode in: aTMethod].
				[#\\]					->	[self typeForArithmetic: sendNode in: aTMethod].
				[#rem:]				->	[self typeForArithmetic: sendNode in: aTMethod].
				[#quo:]					->	[self typeForArithmetic: sendNode in: aTMethod].
				"C99 Sec Bitwise shift operators ... 3 Semantics ...
				 The integer promotions are performed on each of the operands. The type of the result is that of the promoted left operand..."
				[#>>]					->	[sendNode receiver typeFrom: self in: aTMethod].
				[#<<]					->	[(self isSignedIntegralCType: (sendNode receiver typeFrom: self in: aTMethod))
												ifTrue: [#sqInt]
												ifFalse: [#usqInt]].
				[#addressOf:]			->	[(sendNode args first typeFrom: self in: aTMethod)
												ifNil: [#sqInt]
												ifNotNil: [:type| type, (type last isSeparator ifTrue: ['*'] ifFalse: [' *'])]].
				[#addressOf:put:]		->	[(sendNode args first typeFrom: self in: aTMethod)
												ifNil: [#sqInt]
												ifNotNil: [:type| type, (type last isSeparator ifTrue: ['*'] ifFalse: [' *'])]].
				[#at:]					->	[self typeForDereference: sendNode in: aTMethod].
				[#bitAnd:]				->	[self typeForArithmetic: sendNode in: aTMethod].
				[#bitOr:]				->	[self typeForArithmetic: sendNode in: aTMethod].
				[#bitXor:]				->	[self typeForArithmetic: sendNode in: aTMethod].
				[#bitClear:]				->	[self typeForArithmetic: sendNode in: aTMethod].
				[#bitInvert32]			->	[#'unsigned int'].
				[#bitInvert64]			->	[self promoteArithmeticTypes: (sendNode receiver typeFrom: self in: aTMethod) and: #int].
				[#byteSwap32]			->	[#'unsigned int'].
				[#byteSwap64]			->	[#'unsigned long long'].
				[#byteSwapped32IfBigEndian:]	->	[#'unsigned int'].
				[#byteSwapped64IfBigEndian:]	->	[#'unsigned long long'].
				[#=]					->	[#int].
				[#~=]					->	[#int].
				[#==]					->	[#int].
				[#~~]					->	[#int].
				[#<]					->	[#int].
				[#<=]					->	[#int].
				[#>]					->	[#int].
				[#>=]					->	[#int].
				[#between:and:]		->	[#int].
				[#anyMask:]				->	[#int].
				[#allMask:]				->	[#int].
				[#noMask:]				->	[#int].
				[#isNil]					->	[#int].
				[#notNil]				->	[#int].
				[#&]					->	[#int].
				[#|]						->	[#int].
				[#not]					->	[#int].
				[#asFloat]				->	[#double].
				[#atan]					->	[#double].
				[#exp]					->	[#double].
				[#log]					->	[#double].
				[#sin]					->	[#double].
				[#sqrt]					->	[#double].
				[#asLong]				->	[#long].
				[#asInteger]			->	[#sqInt].
				[#asIntegerPtr]			->	[#'sqIntptr_t'].
				[#asUnsignedInteger]	->	[#usqInt].
				[#asUnsignedIntegerPtr]->	[#'usqIntptr_t'].
				[#asUnsignedLong]		->	[#'unsigned long'].
				[#asUnsignedLongLong]		->	[#'unsigned long long'].
				[#asVoidPointer]		->	[#'void *'].
				[#signedIntToLong]		->	[#usqInt]. "c.f. generateSignedIntToLong:on:indent:"
				[#signedIntToShort]	->	[#usqInt]. "c.f. generateSignedIntToShort:on:indent:"
				[#cCoerce:to:]			->	[self conventionalTypeForType: sendNode args last value].
				[#cCoerceSimple:to:]	->	[self conventionalTypeForType: sendNode args last value].
				[#sizeof:]				->	[#'usqIntptr_t']. "Technically it's a size_t but it matches on target architectures so far..."
				[#ifTrue:ifFalse:]		->	[self typeForConditional: sendNode in: aTMethod].
				[#ifFalse:ifTrue:]		->	[self typeForConditional: sendNode in: aTMethod].
				[#ifTrue:]				->	[self typeForConditional: sendNode in: aTMethod].
				[#ifFalse:]				->	[self typeForConditional: sendNode in: aTMethod].
				[#and:]					->	[#sqInt].
				[#or:]					->	[#sqInt].
				[#caseOf:]				->	[self typeFor: sendNode args first in: aTMethod] }
				otherwise: "If there /is/ a method for sel but its return type is as yet unknown it /mustn't/ be defaulted,
							since on a subsequent pass its type may be computable.  Only default unbound selectors."
					[methodOrNil ifNotNil: [nil] ifNil: [typeIfNil]]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 2/4/2020 12:39'!
segregateByGroupingSizeAndVisibility: variables
	"Sort variables by grouping (clusteredVariables first), by size (pointer & integer
	 vars sorted from defaultWordSize to bytes), and finally by size, public last.  The
	 intent is to group smaller variables together for better locality (because we can)."
	| clusteredVariableNames streams defaultStream defaultWordSize groupedBySize privateStream publicStream |
	defaultWordSize := vmClass ifNotNil: [vmClass objectMemoryClass wordSize] ifNil: 8. "We now live in a 64-bit"
	clusteredVariableNames := ([vmClass clusteredVariableNames select: [:cvn| variables includes: cvn]]
									on: MessageNotUnderstood
									do: [:ex| #()]).
	streams := (1 to: 8) collect: [:i| i isPowerOfTwo ifTrue: [(Array new: variables size // 2) writeStream]].
	streams at: 6 put: (streams at: defaultWordSize). "In plugin land sqInt has size 6..."
	defaultStream := (Array new: variables size // 2) writeStream.
	variables do:
		[:varName| | type |
		(clusteredVariableNames includes: varName) ifFalse:
			[type := variableDeclarations
						at: varName
						ifPresent: [:decl| self extractTypeFor: varName fromDeclaration: decl]
						ifAbsent: [#sqInt].
			 ((self isSimpleType: type)
				ifTrue: [streams at: ((self isPointerCType: type)
										ifTrue: [defaultWordSize]
										ifFalse: [self sizeOfIntegralCType: type])]
				ifFalse: [defaultStream])
					nextPut: varName]].
	groupedBySize := Array new: variables size streamContents:
						[:varStream|
						varStream
							nextPutAll: clusteredVariableNames;
							nextPutAll: (streams at: defaultWordSize) contents;
							nextPutAll: (streams at: (defaultWordSize = 8 ifTrue: [4] ifFalse: [8])) contents;
							nextPutAll: (streams at: 2) contents;
							nextPutAll: (streams at: 1) contents;
							nextPutAll: defaultStream contents].
	publicStream := (Array new: variables size // 2) writeStream.
	privateStream := (Array new: variables size // 2) writeStream.
	groupedBySize do:
		[:varName|
		((self mustBeGlobal: varName) ifTrue: [publicStream] ifFalse: [privateStream]) nextPut: varName].
	^privateStream contents, publicStream contents! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 7/5/2011 19:11'!
selectAPIMethods
	^methods select: [:m| m isAPIMethod]! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 04/27/2017 12:55'!
selectAPIVariables
	^variableDeclarations reject: [:decl| decl includesSubstring: 'static ']! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 7/5/2011 19:09'!
selectorReturnsPointerToStruct: selector "<Symbol>"
	| tMethod |
	^(tMethod := methods
					at: selector
					ifAbsent:
						[apiMethods ifNotNil:
							[apiMethods at: selector ifAbsent: []]]) notNil
	  and: [VMStructType isTypePointerToStruct: tMethod returnType]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 10/14/2013 11:56'!
selectorReturnsStruct: selector "<Symbol>"
	| tMethod |
	^(tMethod := methods
					at: selector
					ifAbsent:
						[apiMethods ifNotNil:
							[apiMethods at: selector ifAbsent: []]]) notNil
	  and: [VMStructType isTypeStruct: tMethod returnType]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 2/10/2009 13:55'!
selectorsAndMethodsDo: aBinaryBlock
	methods keysAndValuesDo: aBinaryBlock! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 3/7/2014 17:07'!
shortMonticelloDescriptionForClass: aClass
	"Answer a suitable Monticello package stamp to include in a moduleName."
	^self class shortMonticelloDescriptionForClass: aClass! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 4/27/2020 14:07'!
shouldGenerateAsInterpreterProxySend: aSendNode
	^false! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 12/14/2016 14:08'!
shouldGenerateMethod: aTMethod
	^(self isBuiltinSelector: aTMethod selector)
		ifTrue: [requiredSelectors includes: aTMethod selector]
		ifFalse: [aTMethod inline ~~ #always]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 7/7/2021 22:28'!
shouldIncludeMethodFor: aClass selector: selector
	"Answer whether a method should be translated.  Process optional methods by
	 interpreting the argument to the option: pragma as either a Cogit class name
	 or a class variable name or a variable name in VMBasicConstants.  Exclude
	 methods with the doNotGenerate pragma."
	| options notOptions |

	(aClass >> selector pragmaAt: #doNotGenerate) ifNotNil:
		[^false].

	"where is pragmasAt: ??"
	options := (aClass >> selector) pragmas select: [:p| p keyword == #option:] thenCollect: [:p| p argumentAt: 1].
	notOptions := (aClass >> selector) pragmas select: [:p| p keyword == #notOption:] thenCollect: [:p| p argumentAt: 1].
	(options notEmpty or: [notOptions notEmpty]) ifTrue:
		["Anything defined at compile tiome must be included."
		((options anySatisfy: [:option| (vmClass ifNil: [VMBasicConstants]) defineAtCompileTime: option])
		 or: [notOptions anySatisfy: [:option| (vmClass ifNil: [VMBasicConstants]) defineAtCompileTime: option]]) ifTrue:
			[^true].
		"We have to include the method if either
			- any one of the options is false (because we want #if option...)
			- any one of the notOptions is true (because we want #if !!option...)
			- all of the options is true and all of the notOptions are false (because they have all been satisfied)"
		^((options anySatisfy: [:option| (self optionIsTrue: option in: aClass) not])
		    and: [notOptions anySatisfy: [:option| (self optionIsFalse: option in: aClass) not]])
		   or: [(options allSatisfy: [:option| self optionIsTrue: option in: aClass])
			and: [notOptions allSatisfy: [:option| self optionIsFalse: option in: aClass]]]].

	^true! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'nice 7/13/2016 18:27'!
signedTypeForIntegralType: aCTypeString
	(aCTypeString beginsWith: 'unsigned ') ifTrue:
		[^aCTypeString allButFirst: 8].
	
	(aCTypeString beginsWith: 'usq') ifTrue:
		[^aCTypeString allButFirst].

	aCTypeString = 'size_t' ifTrue: [^#usqIntptr_t].
	
	self error: 'unknown type'.
	^#long! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'dtl 7/27/2021 18:46'!
sizeOfIntegralCType: anIntegralCType "<String>"
	"N.B. Only works for values for which isIntegralCType: answers true."
	| prunedCType index |
	prunedCType := (anIntegralCType beginsWith: 'unsigned ')
						ifTrue: [(anIntegralCType allButFirst: 9) withBlanksTrimmed]
						ifFalse: [(anIntegralCType beginsWith: 'signed ')
									ifTrue: [(anIntegralCType allButFirst: 7) withBlanksTrimmed]
									ifFalse: [anIntegralCType]].
	
	^prunedCType asString caseOf: {
		['sqLong']	->	[8].
		['usqLong']	->	[8].
		['long long']	->	[8].
		['sqInt']		->	[BytesPerOop].
		['usqInt']	->	[BytesPerOop].
		['sqIntptr_t']	->	[BytesPerWord].
		['usqIntptr_t']	->	[BytesPerWord].
		['int']		->	[4].
		['unsigned']	-> [4].
		['uint32_t']	-> [4].
		['short']		->	[2].
		['short int']	->	[2].
		['char']		->	[1].
		['long']		->	[BytesPerWord]. "It's ambiguous on LLP64 and we'll later remove it"
		['size_t']	->	[BytesPerWord].
		['pid_t']		->	[BytesPerWord].
	}
	otherwise:
		[(anIntegralCType beginsWith: 'register ') ifTrue:
			[^self sizeOfIntegralCType: (anIntegralCType allButFirst: 9)].
		 (anIntegralCType beginsWith: 'static ') ifTrue:
			[^self sizeOfIntegralCType: (anIntegralCType allButFirst: 7)].
		 ((anIntegralCType beginsWith: 'unsigned') "e.g. 'unsigned  : 8'"
		  and: [(anIntegralCType includesAnyOf: '[*]') not
		  and: [(index := anIntegralCType indexOf: $:) > 0]]) ifTrue:
			[^(Integer readFrom: (anIntegralCType copyFrom: index + 1 to: anIntegralCType size) withBlanksTrimmed readStream) + 7 // 8].
		 anIntegralCType first = $# ifTrue:
			[(anIntegralCType subStrings: '\') do:
				[:substring|
				 substring first ~~ $# ifTrue:
					[([self sizeOfIntegralCType: substring withBlanksTrimmed]
						on: Error
						do: [:ex| nil]) ifNotNil: [:size| ^size]]]].
		self error: 'unrecognized integral type']! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 7/3/2014 05:56'!
sortMethods: aTMethodCollection
	"We need to define this since different Squeak versions answer different results
	 for asSortedCollection and if sort order changes, generated code changes too.
	 When generating VM code, use class name as major sort index as this groups
	 some methods by functionality (e.g. SpurGenerationScavenger) and that makes
	 the VMProfiler more useful."
	^aTMethodCollection asSortedCollection:
		(self isGeneratingPluginCode
			ifTrue:
				[[:a :b| a selector caseSensitiveLessOrEqual: b selector]]
			ifFalse:
				[[:a :b|
				  a definingClass = b definingClass
					ifTrue: [a selector caseSensitiveLessOrEqual: b selector]
					ifFalse: [a definingClass name caseSensitiveLessOrEqual: b definingClass name]]])! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 5/20/2010 20:46'!
sortStrings: aStringOrSymbolCollection
	"We need to define this since different Squeak versions answer different results
	 for asSortedCollection and if sort order changes, generated code changes too."
	^aStringOrSymbolCollection asSortedCollection: [:a :b| a caseSensitiveLessOrEqual: b]! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 8/8/2017 12:20'!
sortedExportMethods
	"Answer a suitably-sorted array of all exported TMethods"
	^(methods select: [:m| m export]) asSortedCollection: [:a :b| a selector caseSensitiveLessOrEqual: b selector]! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 6/6/2018 20:28'!
staticallyResolveMethodNamed: selector forClass: aClass to: staticallyResolvedSelector
	"We allow a limited amount of polymorphism; if a class chooses, its selectoers can be
	 prefixed with a given string to disambiguate. This hack allows us to use two different
	 compaction algorithms with the same API at the same time; the selection being done
	 by a class which holds the flag stating which algorithm is in effect at the current time."
	| method |
	method := methods
					removeKey: selector
					ifAbsent:
						[self logger cr; nextPutAll: 'warning: did not find ', selector, ' to be able to map to ', staticallyResolvedSelector.
						^self].
	method selector: staticallyResolvedSelector.
	methods at: staticallyResolvedSelector put: method! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 6/6/2018 18:34'!
staticallyResolvedPolymorphicReceiver: variableName to: aClass
	"We allow a limited amount of polymorphism; if a class chooses, its selectoers can be
	 prefixed with a given string to disambiguate. This hack allows us to use two different
	 compaction algorithms with the same API at the same time; the selection being done
	 by a class which holds the flag stating which algorithm is in effect at the current time."
	(staticallyResolvedPolymorphicReceivers ifNil: [staticallyResolvedPolymorphicReceivers := Dictionary new])
		at: variableName
		put: aClass! !

!CCodeGenerator methodsFor: 'C translation support' stamp: 'eem 11/10/2014 11:15'!
stepExpressionIsNegative: aNode
	"Answer if the step expression (the by: argument in a to:by:do:) is negative."
	self isConstantNode: aNode valueInto: [:stepValue| ^stepValue < 0].
	(aNode isSend and: [aNode selector == #negated]) ifTrue:
		[self isConstantNode: aNode receiver valueInto: [:stepValue| ^stepValue > 0]].
	^false! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 2/13/2009 10:19'!
storeAPIExportHeader: headerName OnFile: fullHeaderPath
	"Store C header code on the given file. Evaluate
	 aBlock with the stream to generate its contents."

	| header |
	header := String streamContents:
				[:s|
				 s nextPutAll: (self fileHeaderVersionStampForSourceClass: nil); cr.
				 self emitCAPIExportHeaderOn: s].
	(self needToGenerateHeader: headerName file: fullHeaderPath contents: header) ifTrue:
		[self storeHeaderOnFile: fullHeaderPath contents: header]! !

!CCodeGenerator methodsFor: 'public' stamp: 'ikp 9/26/97 14:50'!
storeCodeOnFile: fileName doInlining: inlineFlag
	"Store C code for this code base on the given file."

	self storeCodeOnFile: fileName doInlining: inlineFlag doAssertions: true! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 11/25/2020 11:30'!
storeCodeOnFile: fileName doInlining: inlineFlag doAssertions: assertionFlag
	"Store C code for this code base on the given file.
	 Try and reuse the file if it already exists."

	| stream |
	stream := VMMaker fileNamed: fileName.
	stream ifNil: [Error signal: 'Could not open C code file: ', fileName].
	stream truncate.
	self emitCCodeOn: stream doInlining: inlineFlag doAssertions: assertionFlag.
	stream close! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 5/17/2011 12:19'!
storeHeaderOnFile: fileName contents: contents
	"Store C header code on the given file. Evaluate
	 aBlock with the stream to generate its contents."

	| aStream |
	aStream := VMMaker forceNewFileNamed: fileName.
	aStream ifNil: [Error signal: 'Could not open C header file: ', fileName].
	[(contents beginsWith: '/* Automatic') ifFalse:
		[aStream nextPutAll: (self fileHeaderVersionStampForSourceClass: nil); cr].
	 aStream nextPutAll: contents]
		ensure: [aStream close]! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 10/12/2013 14:00'!
structClasses
	^structClasses! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 9/9/2018 18:36'!
structClassesForTranslationClasses: classes
	"Answer in superclass order (any superclass precedes any subclass)
	 the ancilliaryClasses that are struct classes for all the given classes."
	| theStructClasses |
	theStructClasses := OrderedCollection new.
	classes do:
		[:aTranslationClass|
		([aTranslationClass ancilliaryClasses]
				on: MessageNotUnderstood
				do: [:ex|
					ex message selector == #ancilliaryClasses
						ifTrue: [#()]
						ifFalse: [ex pass]]) do:
			[:class|
			(class isStructClass
			 and: [(vmClass isNil or: [vmClass isAcceptableAncilliaryClass: class])
			 and: [(theStructClasses includes: class) not]]) ifTrue:
				[theStructClasses addLast: class]]].
	^self superclassOrder: theStructClasses! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 9/1/2017 14:53'!
structTargetKindForDeclaration: typeName "<String>"
	^VMStructType structTargetKindForDeclaration: typeName! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 9/1/2017 15:14'!
structTargetKindForVariableName: varName "<String>"
	^(self typeOfVariable: varName) ifNotNil:
		[:declaration|
		 self structTargetKindForDeclaration: declaration]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 9/9/2018 18:35'!
superclassOrder: classes
	^[ChangeSet superclassOrder: classes] "Squeak"
		on: MessageNotUnderstood
		do: [ :ex|
			ex message selector == #superclassOrder: ifFalse:
				[ex pass].
			Class superclassOrder: classes] "Pharo"! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 11/18/2012 19:50'!
suppressAsmLabels
	^suppressAsmLabels! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 11/18/2012 19:50'!
suppressAsmLabels: aBoolean
	suppressAsmLabels := aBoolean! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 11/18/2012 19:57'!
suppressAsmLabelsWhile: aBlock
	| oldSuppressAsmLabels |
	oldSuppressAsmLabels := suppressAsmLabels.
	suppressAsmLabels := true.
	^aBlock ensure: [suppressAsmLabels := oldSuppressAsmLabels]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 2/23/2015 16:44'!
testInliningFor: selector as: inlineFlagOrSymbol
	"Test inlining for the method with the given selector.
	 Do all inlining first (cuz that's how the algorithm works).
	 Then try and inline into a copy of the method.  This isn't
	 exactly what happens in the real deal but is close enough."
	| meth |
	((breakDestInlineSelectors includes:  selector)
	 or: [(breakSrcInlineSelectors includes:  selector)]) ifTrue:
		[self halt].
	meth := (self methodNamed: selector) copy.
	self doBasicInlining: inlineFlagOrSymbol.
	self halt.
	meth tryToInlineMethodsIn: self! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'TPR 3/2/2000 11:45'!
translationMethodClass
	"return the class used to produce C translation methods from MethodNodes"
	^TMethod! !

!CCodeGenerator methodsFor: 'C translation support' stamp: 'eem 11/17/2015 15:17'!
tryToCollapseBothArmsOfConditional: msgNode on: aStream indent: level
	"Attempt to generate the code for an ifTrue:ifFalse: if both arms are found to be the same, in which case
	 answer true.  Otherwise output nothing and answer false."
	(msgNode args first isSameAs: msgNode args second) ifFalse:
		[^false].
	msgNode receiver hasSideEffect ifTrue:
		[msgNode receiver emitCCodeAsExpressionOn: aStream level: level generator: self.
		 aStream nextPut: $; ; crtab: level].
	msgNode args first emitCCodeOn: aStream level: level generator: self.
	^true! !

!CCodeGenerator methodsFor: 'C translation support' stamp: 'eem 11/17/2015 15:19'!
tryToCollapseBothArmsOfConditionalExpression: msgNode on: aStream indent: level
	"Attempt to generate the code for an ifTrue:ifFalse: if both arms are found to be the same, in which case
	 answer true.  Otherwise output nothing and answer false."
	(msgNode args first isSameAs: msgNode args second) ifFalse:
		[^false].
	aStream nextPut: $(.
	msgNode receiver hasSideEffect ifTrue:
		[msgNode receiver emitCCodeAsExpressionOn: aStream level: level generator: self.
		 aStream nextPut: $, ; crtab: level]..
	msgNode args first emitCCodeAsArgumentOn: aStream level: level generator: self.
	aStream nextPut: $).
	^true! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'nice 4/6/2016 21:36'!
typeFor: aNode in: aTMethod
	^aNode typeFrom: self in: aTMethod! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'eem 8/31/2017 10:25'!
typeForArithmetic: sendNode in: aTMethod
	"Answer the return type for an arithmetic sendThis is so that the inliner can still
	 inline simple expressions.  Deal with pointer arithmetic, floating point arithmetic
	 and promotion."
	| rcvrType argType arg |
	rcvrType := sendNode receiver typeOrNilFrom: self in: aTMethod.
	argType := (arg := sendNode args first) typeOrNilFrom: self in: aTMethod.
	"deal with pointer arithmetic"
	((rcvrType notNil and: [rcvrType last == $*]) or: [argType notNil and: [argType last == $*]]) ifTrue:
		[(rcvrType isNil or: [argType isNil]) ifTrue:
			[^nil].
		 (rcvrType last == $* and: [argType last == $*]) ifTrue:
			[sendNode selector == #- ifTrue:
				[^#int].
			 self error: 'invalid pointer arithmetic'].
		 ^rcvrType last == $*
			ifTrue: [rcvrType]
			ifFalse: [argType]].
	^(self promoteArithmeticTypes: rcvrType and: argType) ifNotNil:
		[:promotedType|
		 "We have to be very careful with subtraction.  The difference between two unsigned types is signed.
		  But we don't want unsigned - constant to be signed.  We almost always want this to stay unsigned."
		 (sendNode selector == #- and: [promotedType first == $u and: [(arg isConstant and: [arg value isInteger]) not]])
			ifTrue: [promotedType allButFirst: ((promotedType beginsWith: 'unsigned') ifTrue: [9] ifFalse: [1])]
			ifFalse: [promotedType]]! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'eem 2/23/2015 19:29'!
typeForConditional: sendNode in: aTMethod
	"Answer the return type for a conditional, ifTrue:ifFalse: et al"
	| firstType secondType |
	firstType := self typeFor: sendNode args first statements last in: aTMethod.
	sendNode selector numArgs = 1 ifTrue:
		[^firstType].
	secondType := self typeFor: sendNode args second statements last in: aTMethod.
	((firstType notNil and: [(self isIntegralCType: firstType) or: [self isFloatingPointCType: firstType]])
	 and: [secondType notNil and: [(self isIntegralCType: secondType) or: [self isFloatingPointCType: secondType]]]) ifTrue:
		[^self promoteArithmeticTypes: firstType and: secondType].
	^firstType ifNil: [secondType]! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'eem 2/17/2015 10:52'!
typeForDereference: sendNode in: aTMethod
	(self typeFor: sendNode receiver in: aTMethod) ifNotNil:
		[:type|
		 type last = $* ifTrue:
			[^type allButLast withBlanksTrimmed]].
	"It would be nice to warn here, but there are contexts when the type is unknown, for example determining
	 the return types of apiMethods.  inferTypesForImplicitlyTypedVariablesAndMethods could provide a signal
	 handler to suppress the warnings in that case but that's too fancy.  Instead we live with the default."
	"logger
		nextPutAll: 'warning, cannot determine type of at: for ', sendNode receiver asString, ' in ', aTMethod selector;
		cr."
	^#sqInt! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 8/24/2018 20:13'!
typeOfVariable: varName "<String>"
	self assert: varName isString.
	scopeStack reverseDo:
		[:dict|
		(dict includesKey: varName) ifTrue:
			[^self
				extractTypeFor: varName
				fromDeclaration: (dict at: varName)]].
	(variableDeclarations
		at: varName
		ifAbsent:
			[apiVariables
				ifNotNil: [apiVariables at: varName ifAbsent: []]
				ifNil: []]) ifNotNil:
		[:decl| ^self extractTypeFor: varName fromDeclaration: decl].
	^(constants at: varName ifAbsent: []) ifNotNil:
		[:const|
		const value isInteger
			ifTrue: [#sqInt]
			ifFalse:
				[const value isFloat ifTrue:
					[#double]]]! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 12/15/2014 09:09'!
unreachableMethods
	"Return a collection of methods that are never invoked."

	| neededSelectors previousSize visited |
	neededSelectors := Set new.
	"collect the exports"
	methods do:
		[:m|
		 m export ifTrue:
			[neededSelectors add: m selector]].

	"Now compute the transitive closure..."
	previousSize := neededSelectors size.
	visited := IdentitySet new: methods size.
	[neededSelectors do:
		[:s|
		(methods at: s ifAbsent: []) ifNotNil:
			[:m|
			(visited includes: m) ifFalse:
				[visited add: m.
				 neededSelectors
					add: m selector;
					addAll: m allCalls]]].
	 neededSelectors size > previousSize]
		whileTrue:
			[previousSize := neededSelectors size].

	^methods reject: [:m| neededSelectors includes: m selector]! !

!CCodeGenerator methodsFor: 'type inference' stamp: 'nice 3/15/2015 21:54'!
unsignedTypeForIntegralType: aCTypeString
	^aCTypeString first = $u
		ifTrue: [aCTypeString]
		ifFalse:
			[(aCTypeString beginsWith: 'sq')
				ifTrue: ['u' , aCTypeString]
				ifFalse: ['unsigned ' , aCTypeString]]! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'ar 7/8/2003 11:23'!
useSymbolicConstants
	"Answer whether we should generate symbolic constants instead of their actual values"
	^useSymbolicConstants! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'ar 7/8/2003 11:23'!
useSymbolicConstants: aBool
	"Indicate whether we should generate symbolic constants instead of their actual values"
	useSymbolicConstants := aBool! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 04/27/2017 12:55'!
var: varName declareC: declarationString
	"Record the given C declaration for a global variable."

	(declarationString includesSubstring: varName) ifFalse:
		[self error: 'declaration omits variable name.  probably an error.  use e.g. var:type:'].
	variableDeclarations at: varName asString put: declarationString.! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 8/8/2008 12:30'!
var: varName type: type
	"Use this in preference to #var:declareC: when ever possible since it avoids typing
	 the varname twice and thus avoids the potential for a typo. See also #var:type:array:"
	self var: varName declareC: type , ' ' , varName! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 1/6/2014 10:47'!
var: varName type: cType array: array
	"Use this in preference to #var:declareC: when possible. This produces a C
	 statment of the form
		int * fooArray[]={1,2,3}
	 See also #var:type: for simple var decls" 
	self
		var: varName
		declareC: (self arrayInitializerCalled: varName for: array sizeString: nil type: cType)! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 1/6/2014 10:47'!
var: varName type: cType sizeString: sizeString array: array
	"Use this in preference to #var:declareC: when possible. This produces a C
	 statment of the form
		int * fooArray[3 /* Foo */]={1,2,3}
	 See also #var:type: for simple var decls" 
	self
		var: varName
		declareC: (self arrayInitializerCalled: varName for: array sizeString: sizeString type: cType)! !

!CCodeGenerator methodsFor: 'C translation support' stamp: 'eem 4/20/2020 19:33'!
variableDeclarationStringsForVariable: variableNameString
	"We (have to?) abuse declarations for optionality using #if C preprocessor forms.
	 This is ugly, but difficult to avoid.  This routine answers either a single string declaration
	 for a variable declared without one of these hacks, or returns the declaration split up into lines."
	| declString |
	declString := variableDeclarations at: variableNameString ifAbsent: [^{'sqInt ', variableNameString}].
	^((declString includes: $#) and: [declString includes: $\])
		ifTrue: [declString withCRs findTokens: Character cr]
		ifFalse: [{declString}]! !

!CCodeGenerator methodsFor: 'inlining' stamp: 'eem 2/24/2015 12:49'!
variableOfType: variableType acceptsValue: exprNode ofType: expressionType
	"Simple type checking for assignment/return values"

	^variableType isNil "means as yet untyped, hence can adapt"
	  or: [expressionType isNil "means we don't know.  Potentially dangerous to say yes, but Slang is traditionally lax :("
	  or: [variableType = expressionType
	  or: [variableType notNil
		 and: [(variableType includes: $*)
				ifTrue:
					[(exprNode isVariable and: [exprNode name = 'nil'])
					  or: [expressionType last = $*
						   and: [(self isVoidPointer: variableType)
								or: [self isVoidPointer: expressionType]]]]
				ifFalse: "Require a cast if assigning integral with floating-point but not the other way around"
					[((self isIntegralCType: variableType)
					   and: [(self isIntegralCType: expressionType)
					   and: [(self sizeOfIntegralCType: variableType) >= (self sizeOfIntegralCType: expressionType)]])
					or: [(self isFloatingPointCType: variableType)
						and: [(self isFloatingPointCType: expressionType)
							or: [self isIntegralCType: expressionType]]]]]]]]! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 10/12/2013 13:59'!
vmClass
	"Answer the interpreter class if any.  This is nil other than for the core VM."
	^vmClass! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 1/21/2016 16:28'!
vmClass: aClass
	"Set the main translation class if any.  This is nil other than for the core VM.
	 It may be an interpreter or a cogit"
	vmClass := aClass.
	vmClass ifNotNil:
		[generateDeadCode := vmClass shouldGenerateDeadCode]! !

!CCodeGenerator methodsFor: 'public' stamp: 'eem 3/13/2018 15:52'!
vmHeaderContentsWithBytesPerWord: bytesPerWord
	"Store C header code on the given stream."

	^ ByteString
		streamContents: [ :tempStream | vmClass writeVMHeaderTo: tempStream bytesPerWord: bytesPerWord generator: self ]! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 10/30/2013 16:28'!
vmMaker
	^vmMaker! !

!CCodeGenerator methodsFor: 'accessing' stamp: 'eem 10/30/2013 16:28'!
vmMaker: aVMMaker
	vmMaker := aVMMaker! !

!CCodeGenerator methodsFor: 'utilities' stamp: 'eem 11/19/2012 17:25'!
with: aNode ifAppropriateSetTo: aBoolean do: aBlock
	"If aNode is a TDefineNode for a variable, set it to the supplied boolean for the duration
	 of aBlock.  This avoids duplicating both arms of nested cppIf:ifTrue:[ifFalse:]'s."
	| savedOptions |
	(aNode isConstant and: [aNode name notEmpty and: [aNode name first isLetter]]) ifFalse:
		[^aBlock value].
	savedOptions := optionsDictionary.
	optionsDictionary := savedOptions
							ifNil: [Dictionary new]
							ifNotNil: [savedOptions copy].
	optionsDictionary at: aNode name put: aBoolean.
	^aBlock ensure: [optionsDictionary := savedOptions]! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 3/16/2018 19:42'!
withConditionalDefineOf: aConstantNameString comment: commentOrNil on: aStream do: aBlock
	| start |
	aStream nextPutAll: '#if !!defined('; nextPutAll: aConstantNameString; nextPut: $).
	commentOrNil ifNotNil: [aStream nextPutAll: ' /* '; nextPutAll: commentOrNil; nextPutAll: ' */'].
	aStream cr.
	start := aStream position.
	aBlock value.
	aStream nextPutAll: '#endif'.
	aStream position - start > 80 ifTrue:
		[aStream nextPutAll: ' /* !!defined('; nextPutAll: aConstantNameString; nextPutAll: ') */'].
	aStream cr! !

!CCodeGenerator methodsFor: 'C code generator' stamp: 'eem 8/8/2017 12:26'!
withOptionalConditionalDefineFor: aTMethod on: aStream do: aBlock
	"Evaluate aBlock, surrounded by an appropriate conditional define added to the stream if required."
	| compileTimeOptionPragmas |
	(compileTimeOptionPragmas := aTMethod compileTimeOptionPragmas) notEmpty ifTrue:
		[aTMethod outputConditionalDefineFor: compileTimeOptionPragmas on: aStream].
	aBlock value.
	aTMethod terminateConditionalDefineFor: compileTimeOptionPragmas on: aStream! !

CCodeGenerator subclass: #CCodeGeneratorGlobalStructure
	instanceVariableNames: 'localStructDef structDefDefine'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!CCodeGeneratorGlobalStructure commentStamp: 'tpr 5/23/2003 11:17' prior: 0!
This subclass of CCodeGenerator adds support for sticking most global variables into a large global array. This in turn means that suitably minded C compilers can do a better job of accessing those variables; in particular the Mac OS use of PPC and Acorn use of ARM benfits by a substantial margin. 

Only simple globals are currently put in the array. Someday we might try adding pointers to the various arrays etc.!

!CCodeGeneratorGlobalStructure methodsFor: 'C code generator' stamp: 'eem 7/14/2014 12:15'!
buildSortedVariablesCollection
	"Build sorted vars, end result will be sorted collection based on static usage, 
	perhaps cache lines will like this!!"

	| globalNames |
	globalNames := Bag new: globalVariableUsage size.
	globalVariableUsage keysAndValuesDo:
		[:k :v | | count |
		count := 0.
		v do:
			[:methodName|
			(methods at: methodName ifAbsent: []) ifNotNil:
				[:method|
				 method parseTree nodesDo:
					[:n|
					(n isVariable
					 and: [n name hasEqualElements: k]) ifTrue:
						[count := count + 1]]]].
		globalNames
			add: k		"move arrays (e.g. methodCache) to end of struct"
			withOccurrences: (((variableDeclarations at: k ifAbsent: ['']) includes: $[)
								ifTrue: [count]
								ifFalse: [count + 1000])].
	variableDeclarations keysDo:
		[:e | globalNames add: e withOccurrences: 0].
	variables do:
		[:e | globalNames add: e withOccurrences: 0].
	^(globalNames sortedElements asSortedCollection:
		[:a1 :a2| a1 value > a2 value or: [a1 value = a2 value and: [a1 key <= a2 key]]]) collect:
			[:ea| ea key]! !

!CCodeGeneratorGlobalStructure methodsFor: 'utilities' stamp: 'tpr 5/14/2013 16:46'!
checkForGlobalUsage: vars in: aTMethod 
	"override to handle global struct needs"
	super checkForGlobalUsage: vars in: aTMethod.

	vars asSet do:
		[:var |
		"if any var is global and in the global var struct 
		tell the TMethod it may be refering to the  struct, depending upon the #defines"
		 ((variables includes: var)
		  and: [self placeInStructure: var]) ifTrue:
			[aTMethod referencesGlobalStruct]]! !

!CCodeGeneratorGlobalStructure methodsFor: 'C code generator' stamp: 'eem 2/3/2020 11:28'!
emitCVariablesOn: aStream
	"Store the global variable declarations on the given stream.
	 Break logic into vars for structure and vars for non-structure."
	| structure nonstruct |

	structure := WriteStream on: (String new: 32768).
	nonstruct := WriteStream on: (String new: 32768).
	aStream nextPutAll: '/*** Variables ***/'; cr.
	structure
		nextPutAll: '#if SQ_USE_GLOBAL_STRUCT'; cr;
		nextPutAll: '# define _iss /* define in-struct static as void */'; cr;
		nextPutAll: 'static struct foo {'; cr;
		nextPutAll: '#else'; cr;
		nextPutAll: '# define _iss static'; cr;
		nextPutAll: '#endif'; cr.
	(self segregateByGroupingSizeAndVisibility: self buildSortedVariablesCollection) do:
		[ :var | | varString inStruct target |
		target := (inStruct := self placeInStructure: (varString := var asString)) 
					ifTrue: [structure]
					ifFalse: [nonstruct].
		(self variableDeclarationStringsForVariable: varString) do:
			[:decl|
			 decl first == $#
				ifTrue:
					[target nextPutAll: decl; cr]
				ifFalse:
					[self isGeneratingPluginCode
						ifTrue:
							[varString = 'interpreterProxy'
								ifTrue: "quite special..."
									[self preDeclareInterpreterProxyOn: target]
								ifFalse: [target nextPutAll: 'static ']]
						ifFalse:
							[(vmClass mustBeGlobal: varString) ifFalse:
								[target nextPutAll: (inStruct ifTrue: ['_iss '] ifFalse: ['static '])]].
					target nextPutAll: decl; nextPut: $;; cr]]].
	structure
		nextPutAll: '#undef _iss'; cr;
		nextPutAll: '#if SQ_USE_GLOBAL_STRUCT'; cr;
		nextPutAll: ' } fum;'; cr;
		nextPutAll: ' #if SQ_USE_GLOBAL_STRUCT_REG';cr;
		nextPutAll: '# define DECL_MAYBE_SQ_GLOBAL_STRUCT /* using a global reg pointer */'; cr;
		nextPutAll: '# define DECL_MAYBE_VOLATILE_SQ_GLOBAL_STRUCT /* using a global reg pointer */'; cr;
		nextPutAll:'#else';cr;
		nextPutAll: '# define DECL_MAYBE_SQ_GLOBAL_STRUCT register struct foo * foo = &fum;'; cr;
		nextPutAll: '# define DECL_MAYBE_VOLATILE_SQ_GLOBAL_STRUCT volatile register struct foo * foo = &fum;'; cr;
		nextPutAll: '#endif';cr;
		nextPutAll: '# define GIV(interpreterInstVar) (foo->interpreterInstVar)'; cr;
		nextPutAll: '#else'; cr;
		nextPutAll: '# define DECL_MAYBE_SQ_GLOBAL_STRUCT /* oh, no mr bill!! */'; cr;
		nextPutAll: '# define DECL_MAYBE_VOLATILE_SQ_GLOBAL_STRUCT /* oh no, mr bill!! */'; cr;
		nextPutAll: '# define GIV(interpreterInstVar) interpreterInstVar'; cr;
		nextPutAll: '#endif'; cr.

	"if the machine needs the fum structure defining locally, do it now; global register users don't need that, but DO need some batshit insane C macro fudging in order to convert the define of USE_GLOBAL_STRUCT_REG into a simple string to use in the asm clause below. Sigh."
	structure
		nextPutAll: '#if SQ_USE_GLOBAL_STRUCT'; cr;
		nextPutAll: '#if SQ_USE_GLOBAL_STRUCT_REG';cr;
		nextPutAll: '#define fooxstr(s) foostr(s)'; cr;
		nextPutAll: '#define foostr(s)  #s'; cr;
		nextPutAll: 'register struct foo * foo asm(fooxstr(USE_GLOBAL_STRUCT_REG));'; cr;
		nextPutAll: '#else'; cr;
		nextPutAll: 'static struct foo * foo = &fum;'; cr;
		nextPutAll: '#endif'; cr;
		nextPutAll: '#endif'; cr.

	aStream
		nextPutAll: structure contents;
		nextPutAll: nonstruct contents;
		cr! !

!CCodeGeneratorGlobalStructure methodsFor: 'C code generator' stamp: 'tpr 5/14/2013 18:12'!
emitGlobalStructFlagOn: aStream
	"Depending upon the value of structDefDefine (See also #structDefDefine:), define SQ_USE_GLOBAL_STRUCT before including the header. Also derive the flag for using the global register; define USE_GLOBAL_STRUCT_REG to do so"

	aStream
		nextPutAll: '#if ';
		nextPutAll: structDefDefine; cr;
		nextPutAll: '# define SQ_USE_GLOBAL_STRUCT 1'; cr;
		nextPutAll: '#else'; cr;
		nextPutAll: '# define SQ_USE_GLOBAL_STRUCT 0'; cr;
		nextPutAll: '#endif'; cr;
		nextPutAll: '#if USE_GLOBAL_STRUCT_REG '; cr;
		nextPutAll: '# define SQ_USE_GLOBAL_STRUCT_REG 1'; cr;
		nextPutAll: '#else'; cr;
		nextPutAll: '# define SQ_USE_GLOBAL_STRUCT_REG 0'; cr;
		nextPutAll: '#endif'; cr;
		cr! !

!CCodeGeneratorGlobalStructure methodsFor: 'C code generator' stamp: 'tpr 5/14/2013 17:47'!
initialize
	super initialize.
	localStructDef := nil. "ignored ivar - no longer used"
	structDefDefine := '1'! !

!CCodeGeneratorGlobalStructure methodsFor: 'public' stamp: ''!
isGlobalStructureBuild
	^true! !

!CCodeGeneratorGlobalStructure methodsFor: 'utilities' stamp: ''!
localizeGlobalVariables
"TPR - remove all the global vars destined for the structure that are only used once - not worth the space,
actually what will happen is the folding code will fold these variables into the method"

	super localizeGlobalVariables.
	globalVariableUsage := globalVariableUsage select: [:e | e size > 1].
! !

!CCodeGeneratorGlobalStructure methodsFor: 'C code generator' stamp: 'eem 3/29/2020 20:10'!
placeInStructure: var
	"See if we should put this array into a structure.
	The variables listed are hardcoded as C in the interpreter thus they don't get resolved via TVariableNode logic.
	Also let's ignore variables that have special definitions that require initialization, and the function def which has problems."

	| check |
	check := variableDeclarations at: var ifAbsent: [''].
	(check includes: $=) ifTrue: [^false].
	(check includes: $() ifTrue: [^false].
	(check includesSubstring: 'static') ifTrue: [^false].

	^(vmClass mustBeGlobal: var) not! !

!CCodeGeneratorGlobalStructure methodsFor: 'C code generator' stamp: 'eem 2/11/2009 09:03'!
returnPrefixFromVariable: aName
	^((variables includes: aName)
	   and: [self placeInStructure: aName])
		ifTrue: ['GIV(',aName,')']
		ifFalse: [aName]! !

!CCodeGeneratorGlobalStructure methodsFor: 'initialize-release' stamp: 'tpr 5/14/2013 18:08'!
structDefDefine: aString
"set the string that will appear in the C file to define whether or not to use the global struct; reasonable values would be:
'USE_GLOBAL_STRUCT' - which would be defined in a header or makefile
'0' - which would mean never do it
'1' - which would mean always do it"
	structDefDefine := aString! !

CCodeGenerator subclass: #VMPluginCodeGenerator
	instanceVariableNames: 'pluginClass pluginFunctionsUsed inProgressSelectors inliningDone'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!VMPluginCodeGenerator commentStamp: '<historical>' prior: 0!
I generate code that can be loaded dynamically from external libraries (e.g., DSOs on Unix or DLLs on Windows)!

VMPluginCodeGenerator subclass: #SmartSyntaxPluginCodeGenerator
	instanceVariableNames: 'debugFlag'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SmartSyntaxPlugins'!

!SmartSyntaxPluginCodeGenerator commentStamp: 'tpr 5/5/2003 16:03' prior: 0!
Subclass of CCodeGenerator, used in connection with TestInterpreterPlugin to generate named primitives with type coercion specifications. See the plugins implemeted as subclasses of TestInterpreterPlugin!

!SmartSyntaxPluginCodeGenerator methodsFor: 'utilities' stamp: 'eem 12/18/2018 18:13'!
anyMethodNamed: selector
	"Answer any method in the code base (including api methods) with the given selector.
	 Override to find smart syntax methods that get  entered in the dictionary under the
	 name specified in the primitive:parameters: send."

	^(super anyMethodNamed: selector) ifNil:
		[methods
			detect: [:m| m smalltalkSelector == selector]
			ifNone: []]! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'acg 12/17/1999 07:32'!
ccgLoad: aBlock expr: aString asBooleanValueFrom: anInteger
	"Answer codestring for boolean coercion (with validating side-effect) of object, as described in comment to ccgLoad:expr:asRawOopFrom:"

	^aBlock value: (String streamContents: [:aStream | aStream
		nextPutAll: 'interpreterProxy booleanValueOf:';
		crtab: 2;
		nextPutAll: '(interpreterProxy stackValue:';
		nextPutAll: anInteger asString;
		nextPutAll:	')'])! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/1/2020 18:07'!
ccgLoad: aBlock expr: aString asCharPtrFrom: anInteger andThen: valBlock
	"Answer codestring for character pointer to first indexable field of object (without validating side-effect unless specified in valBlock), as described in comment to ccgLoad:expr:asRawOopFrom:"

	^self ccgLoad: aBlock expr: aString asPtrFrom: anInteger andThen: valBlock! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'acg 12/17/1999 07:29'!
ccgLoad: aBlock expr: aString asFloatValueFrom: anInteger
	"Answer codestring for double precision coercion (with validating side-effect) of oop, as described in comment to ccgLoad:expr:asRawOopFrom:"

	^aBlock value: (String streamContents: [:aStream | aStream
		nextPutAll: 'interpreterProxy stackFloatValue: ';
		nextPutAll: anInteger asString])! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/1/2020 18:12'!
ccgLoad: aBlock expr: aString asIntPtrFrom: anInteger andThen: valBlock
	"Answer codestring for integer pointer to first indexable field of object (without validating side-effect unless specified in valBlock), as described in comment to ccgLoad:expr:asRawOopFrom:"

	^self ccgLoad: aBlock expr: aString asPtrFrom: anInteger andThen: valBlock! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'acg 12/17/1999 07:29'!
ccgLoad: aBlock expr: aString asIntegerValueFrom: anInteger
	"Answer codestring for integer coercion (with validating side-effect) of oop, as described in comment to ccgLoad:expr:asRawOopFrom:"

	^aBlock value: (String streamContents: [:aStream | aStream
		nextPutAll: 'interpreterProxy stackIntegerValue: ';
		nextPutAll: anInteger asString])! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 7/27/2021 09:49'!
ccgLoad: aBlock expr: variableName asKindOf: aClass from: stackIndex
	^self
		load: (self ccgLoad: aBlock expr: variableName asRawOopFrom: stackIndex)
		validatingWith: #is:KindOf:
		on: { variableName. aClass asString }
		from: stackIndex! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 7/27/2021 09:43'!
ccgLoad: aBlock expr: variableName asKindOfIntegerFrom: stackIndex
	^self
		load: (self ccgLoad: aBlock expr: variableName asRawOopFrom: stackIndex)
		validatingWith: #isKindOfInteger:
		on: variableName
		from: stackIndex! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 7/27/2021 09:49'!
ccgLoad: aBlock expr: variableName asMemberOf: aClass from: stackIndex
	^self
		load: (self ccgLoad: aBlock expr: variableName asRawOopFrom: stackIndex)
		validatingWith: #is:MemberOf:
		on: { variableName. aClass asString }
		from: stackIndex! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 7/27/2021 09:50'!
ccgLoad: aBlock expr: variableName asMemberOfLargeNegativeIntegerFrom: stackIndex
	^self
		load: (self ccgLoad: aBlock expr: variableName asRawOopFrom: stackIndex)
		validatingWith: #isLargeNegativeIntegerObject:
		on: variableName
		from: stackIndex! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 7/27/2021 09:50'!
ccgLoad: aBlock expr: variableName asMemberOfLargePositiveIntegerFrom: stackIndex
	^self
		load: (self ccgLoad: aBlock expr: variableName asRawOopFrom: stackIndex)
		validatingWith: #isLargePositiveIntegerObject:
		on: variableName
		from: stackIndex! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'acg 12/17/1999 07:29'!
ccgLoad: aBlock expr: aString asNonIntegerValueFrom: anInteger
	"Answer codestring for oop (with validating side effect), as described in comment to ccgLoad:expr:asRawOopFrom:"

	^aBlock value: (String streamContents: [:aStream | aStream
		nextPutAll: 'interpreterProxy stackObjectValue: ';
		nextPutAll: anInteger asString])! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/1/2020 18:12'!
ccgLoad: aBlock expr: aString asOopPtrFrom: anInteger andThen: valBlock
	"Answer codestring for integer pointer to first indexable field of object (without validating side-effect unless specified in valBlock), as described in comment to ccgLoad:expr:asRawOopFrom:"

	^self ccgLoad: aBlock expr: aString asPtrFrom: anInteger andThen: valBlock! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/1/2020 18:07'!
ccgLoad: aBlock expr: aString asPtrFrom: anInteger andThen: valBlock
	"Answer codestring for void pointer to first indexable field of object (without validating side-effect unless specified in valBlock), as described in comment to ccgLoad:expr:asRawOopFrom:"

	^(valBlock value: anInteger),
	  '.',
	  (aBlock value: 'interpreterProxy firstIndexableField:(interpreterProxy stackValue:', anInteger asString, '))')! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 09:32'!
ccgLoad: aBlock expr: aString asRawOopFrom: anInteger
	"Answer a string for a Slang expression that will load an oop (without validation) from stack index anInteger.  Apply aBlock that when passed an expression, will answer a string assigning the expression to the desired identifier, to the string before answering.  aString is a Slang expression that refers to the stack value, once it has been loaded."

	^aBlock value: 'interpreterProxy stackValue: ', anInteger printString! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/1/2020 18:07'!
ccgLoad: aBlock expr: aString asShortPtrFrom: anInteger andThen: valBlock
	^self ccgLoad: aBlock expr: aString asPtrFrom: anInteger andThen: valBlock! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/1/2020 18:12'!
ccgLoad: aBlock expr: aString asUnsignedLong64PtrFrom: anInteger andThen: valBlock
	"Answer a codestring for integer pointer to first indexable field of object (without validating side-effect unless specified in valBlock), as described in comment to ccgLoad:expr:asRawOopFrom:"

	^self ccgLoad: aBlock expr: aString asPtrFrom: anInteger andThen: valBlock! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/1/2020 18:12'!
ccgLoad: aBlock expr: aString asUnsignedPtrFrom: anInteger andThen: valBlock
	"Answer a codestring for integer pointer to first indexable field of object (without validating side-effect unless specified in valBlock), as described in comment to ccgLoad:expr:asRawOopFrom:"

	^self ccgLoad: aBlock expr: aString asPtrFrom: anInteger andThen: valBlock! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/1/2020 18:12'!
ccgLoad: aBlock expr: aString asUnsignedShortPtrFrom: anInteger andThen: valBlock
	"Answer a codestring for integer pointer to first indexable field of object (without validating side-effect unless specified in valBlock), as described in comment to ccgLoad:expr:asRawOopFrom:"

	^self ccgLoad: aBlock expr: aString asPtrFrom: anInteger andThen: valBlock! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/4/2014 11:31'!
ccgLoad: aBlock expr: aString asUnsignedValueFrom: anInteger
	"Answer a codestring for positive integer coercion (with validating side-effect) of oop, as described in comment to ccgLoad:expr:asRawOopFrom:"

	^aBlock value: (String streamContents: [:aStream |
		aStream
			nextPutAll: '(interpreterProxy bytesPerOop = 4';
			crtab: 2;
			nextPutAll: 'ifTrue: [interpreterProxy positive32BitValueOf:';
			crtab: 4;
			nextPutAll: '(interpreterProxy stackValue:';
			print: anInteger;
			nextPutAll: ')]';
			crtab: 2;
			nextPutAll: 'ifFalse: [interpreterProxy positive64BitValueOf:';
			crtab: 4;
			nextPutAll: '(interpreterProxy stackValue:';
			print: anInteger;
			nextPutAll: ')])'])! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'acg 12/17/1999 07:31'!
ccgLoad: aBlock expr: aString asWBCharPtrFrom: anInteger
	"Answer codestring for char pointer to first indexable field of object (with validating side-effect), as described in comment to ccgLoad:expr:asRawOopFrom:"

	^aBlock value: (String streamContents: [:aStream | aStream
		nextPutAll: 'self cCoerce: (interpreterProxy arrayValueOf:';
		crtab: 4;
		nextPutAll: '(interpreterProxy stackValue:';
		nextPutAll: anInteger asString;
		nextPutAll:	'))';
		crtab: 3;
		nextPutAll: 'to: ''char *'''])! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'acg 12/17/1999 07:31'!
ccgLoad: aBlock expr: aString asWBFloatPtrFrom: anInteger
	"Answer codestring for single-precision float pointer to first indexable field of object (with validating side-effect), as described in comment to ccgLoad:expr:asRawOopFrom:"

	^aBlock value: (String streamContents: [:aStream | aStream
		nextPutAll: 'self cCoerce: (interpreterProxy arrayValueOf:';
		crtab: 4;
		nextPutAll: '(interpreterProxy stackValue:';
		nextPutAll: anInteger asString;
		nextPutAll:	'))';
		crtab: 3;
		nextPutAll: 'to: ''float *'''])! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'acg 12/17/1999 07:31'!
ccgLoad: aBlock expr: aString asWBIntPtrFrom: anInteger
	"Answer codestring for integer pointer to first indexable field of object (with validating side-effect), as described in comment to ccgLoad:expr:asRawOopFrom:"

	^aBlock value: (String streamContents: [:aStream | aStream
		nextPutAll: 'self cCoerce: (interpreterProxy arrayValueOf:';
		crtab: 4;
		nextPutAll: '(interpreterProxy stackValue:';
		nextPutAll: anInteger asString;
		nextPutAll:	'))';
		crtab: 3;
		nextPutAll: 'to: ''int *'''])! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'ar 4/4/2006 21:10'!
ccgSetBlock: aString

	^[:expr | aString, ' := ', expr]! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'acg 9/19/1999 13:05'!
ccgTVarBlock: anInteger

	^[:expr | '(thisContext tempAt: 1) tempAt: ', anInteger asString, ' put: (', expr, ')']! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'Igor.Stasenko 2/3/2011 11:35'!
ccgValBlock: valString

	^[:index | String streamContents:
		[:aStream | aStream
			nextPutAll: 'interpreterProxy success: (interpreterProxy ';
			nextPutAll: valString;
			nextPutAll: ': (interpreterProxy stackValue: ';
			nextPutAll: index asString;
			nextPutAll: '))']] ! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 9/10/2020 12:49'!
ccgValBlock: valString or: valorString

	^[:index | String streamContents:
		[:aStream | aStream
			nextPutAll: 'interpreterProxy success: ((interpreterProxy ';
			nextPutAll: valString;
			nextPutAll: ': (interpreterProxy stackValue: ';
			nextPutAll: index asString;
			nextPutAll: ')) || (interpreterProxy ';
			nextPutAll: valorString;
			nextPutAll: ': (interpreterProxy stackValue: ';
			nextPutAll: index asString;
			nextPutAll: ')))']] ! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:41'!
genCallOf: aSelector with: aNode on: aStream
	pluginFunctionsUsed add: aSelector.
	aStream nextPutAll: (self cFunctionNameFor: aSelector).
	aStream nextPut: $(.
	self emitCExpression: aNode on: aStream.
	aStream nextPut: $)! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:47'!
genCallOf: aSelector with: firstArgNode with: secondArgNode on: aStream
	pluginFunctionsUsed add: aSelector.
	aStream nextPutAll: (self cFunctionNameFor: aSelector).
	aStream nextPut: $(.
	self emitCExpression: firstArgNode on: aStream.
	aStream nextPut: $,.
	self emitCExpression: secondArgNode on: aStream.
	aStream nextPut: $)! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:53'!
genCallOf: aSelector with: firstArgNode with: secondArgNode with: thirdArgNode on: aStream
	pluginFunctionsUsed add: aSelector.
	aStream nextPutAll: (self cFunctionNameFor: aSelector).
	aStream nextPut: $(.
	self emitCExpression: firstArgNode on: aStream.
	aStream nextPut: $,.
	self emitCExpression: secondArgNode on: aStream.
	aStream nextPut: $,.
	self emitCExpression: thirdArgNode on: aStream.
	aStream nextPut: $)! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/7/2012 16:30'!
generateAsBooleanObj: aNode on: aStream indent: anInteger
	pluginFunctionsUsed add: #trueObject; add: #falseObject.
	aStream nextPutAll: '(('.
	self emitCExpression: aNode receiver on: aStream.
	aStream nextPutAll: ') ? trueObject() : falseObject())'.! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:41'!
generateAsCBoolean: aNode on: aStream indent: anInteger
	self genCallOf: #booleanValueOf: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:42'!
generateAsCDouble: aNode on: aStream indent: anInteger
	self genCallOf: #floatValueOf: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'nice 4/10/2016 14:58'!
generateAsCInt: aNode on: aStream indent: anInteger
	self genCallOf: #integerValueOf: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:42'!
generateAsCUnsigned: aNode on: aStream indent: anInteger
	self genCallOf: #positive32BitValueOf: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:44'!
generateAsCharPtr: aNode on: aStream indent: anInteger
	self generateCoerceToPtr: 'char *' fromObject: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:42'!
generateAsFloatObj: aNode on: aStream indent: anInteger
	self genCallOf: #floatObjectOf: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/10/2012 10:59'!
generateAsIfVar: aNode on: aStream indent: anInteger

	| cName fName class index |
	cName := String streamContents: 
		[:scStr | self emitCExpression: aNode args first on: scStr].
	(class := Smalltalk classNamed: cName) isBehavior ifFalse: 
		[^self error: 'first arg must identify class'].
	fName := aNode args second value.
	index := class
				instVarIndexFor: fName
				ifAbsent: [^self error: 'second arg must be instVar'].
	pluginFunctionsUsed add: #fetchPointer:ofObject:.
	aStream nextPutAll: 'fetchPointerofObject('; print: index - 1; nextPut: $,.
	self emitCExpression: aNode receiver on: aStream.
	aStream nextPut: $)! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/10/2012 10:58'!
generateAsIfVarAsValue: aNode on: aStream indent: anInteger

	| cName fName class index fetchNode |
	cName := String streamContents: 
				[:scStr | self emitCExpression: aNode args first on: scStr].
	(class := Smalltalk classNamed: cName) isBehavior ifFalse: 
		[^self error: 'first arg must identify class'].
	fName := aNode args second value.
	index := class
				instVarIndexFor: fName
				ifAbsent: [^self error: 'second arg must be instVar'].
	fetchNode := TSendNode new
					setSelector: #fetchPointer:ofObject:
					receiver: (TVariableNode new setName: 'self')
					arguments: { TConstantNode new setValue: index - 1. aNode receiver}.
	cName := aNode args third nameOrValue.
	(class := Smalltalk classNamed: cName) isBehavior ifFalse: 
		[^self error: 'third arg must identify class'].
	class ccg: self generateCoerceToValueFrom: fetchNode on: aStream
! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/10/2012 10:59'!
generateAsIfVarPut: aNode on: aStream indent: anInteger

	| cName fName class index |
	cName := String streamContents: 
		[:scStr | self emitCExpression: aNode args first on: scStr].
	(class := Smalltalk classNamed: cName) isBehavior ifFalse: 
		[^self error: 'first arg must identify class'].
	fName := aNode args second value.
	index := class
				instVarIndexFor: fName
				ifAbsent: [^self error: 'second arg must be instVar'].
	pluginFunctionsUsed add: #storePointer:ofObject:withValue:.
	aStream  nextPutAll: 'storePointerofObjectwithValue('; print: index - 1; nextPut: $,.
	self emitCExpression: aNode receiver on: aStream.
	aStream  nextPut: $,.
	self emitCExpression: aNode args third on: aStream.
	aStream  nextPut: $)! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:44'!
generateAsIntPtr: aNode on: aStream indent: anInteger
	self generateCoerceToPtr: 'int *' fromObject: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'acg 12/17/1999 07:23'!
generateAsOop: aNode on: aStream indent: anInteger

	| cName class |
	cName := aNode args first nameOrValue.
	class := Smalltalk 
		at: (cName asSymbol) 
		ifAbsent: [nil].
	(class isNil not and: [class isBehavior]) ifFalse: 
		[^self error: 'first arg must identify class'].
	class ccg: self generateCoerceToOopFrom: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:42'!
generateAsPositiveIntegerObj: aNode on: aStream indent: anInteger
	self genCallOf: #positive32BitIntegerFor: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:42'!
generateAsSmallIntegerObj: aNode on: aStream indent: anInteger
	self genCallOf: #integerObjectOf: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'acg 12/17/1999 07:22'!
generateAsValue: aNode on: aStream indent: anInteger

	| cName class |
	cName := aNode args first nameOrValue.
	class := Smalltalk 
		at: (cName asSymbol) 
		ifAbsent: [nil].
	(class isNil not and: [class isBehavior]) ifFalse: 
		[^self error: 'first arg must identify class'].
	class ccg: self generateCoerceToValueFrom: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'nice 7/8/2016 03:15'!
generateCPtrAsOop: aNode on: aStream indent: anInteger
	aStream nextPutAll: '((sqInt)(sqIntptr_t)('.
	self emitCExpression: aNode receiver on: aStream.
	aStream nextPutAll: ') - BaseHeaderSize)'! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:42'!
generateClass: aNode on: aStream indent: anInteger
	self genCallOf: #fetchClassOf: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'asOop:/asValue:' stamp: 'eem 12/7/2012 16:30'!
generateCoerceToBooleanObjectFrom: aNode on: aStream
	pluginFunctionsUsed add: #trueObject; add: #falseObject.
	aStream nextPutAll: '(('.
	self emitCExpression: aNode on: aStream.
	aStream nextPutAll: ') ? trueObject() : falseObject())'! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'asOop:/asValue:' stamp: 'eem 12/5/2012 18:42'!
generateCoerceToBooleanValueFrom: aNode on: aStream
	self genCallOf: #booleanValueOf: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'asOop:/asValue:' stamp: 'eem 12/5/2012 18:42'!
generateCoerceToFloatObjectFrom: aNode on: aStream
	self genCallOf: #floatObjectOf: with: aNode on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'asOop:/asValue:' stamp: 'eem 12/5/2012 18:42'!
generateCoerceToFloatValueFrom: aNode on: aStream
	self genCallOf: #floatValueOf: with: aNode on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'asOop:/asValue:' stamp: 'eem 12/4/2014 11:37'!
generateCoerceToObjectFromPtr: aNode on: aStream
	"This code assumes no named instance variables"

	aStream nextPutAll: '((sqInt)(('.
	self emitCExpression: aNode on: aStream.
	aStream nextPutAll: ') - BaseHeaderSize))'! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'asOop:/asValue:' stamp: 'eem 12/5/2012 18:43'!
generateCoerceToPtr: aString fromObject: aNode on: aStream
	"This code assumes no named instance variables"
	aStream 
		nextPutAll: '((';
		nextPutAll: aString;
		nextPut: $).
	self genCallOf: #firstIndexableField: with: aNode on: aStream.
	aStream nextPut: $)! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'asOop:/asValue:' stamp: 'eem 12/5/2012 18:45'!
generateCoerceToSmallIntegerObjectFrom: aNode on: aStream
	self genCallOf: #integerObjectOf: with: aNode on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'asOop:/asValue:' stamp: 'eem 12/5/2012 18:45'!
generateCoerceToSmallIntegerValueFrom: aNode on: aStream
	self genCallOf: #integerValueOf: with: aNode on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'asOop:/asValue:' stamp: 'eem 12/5/2012 18:45'!
generateCoerceToUnsignedObjectFrom: aNode on: aStream
	self genCallOf: #positive32BitIntegerFor: with: aNode on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'asOop:/asValue:' stamp: 'eem 12/5/2012 18:45'!
generateCoerceToUnsignedValueFrom: aNode on: aStream
	self genCallOf: #positive32BitValueOf: with: aNode on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'debug code' stamp: 'sr 4/8/2000 00:52'!
generateDebugCode
	^ debugFlag! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'debug code' stamp: 'sr 4/8/2000 00:52'!
generateDebugCode: aBool 
	debugFlag := aBool! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'sr 4/8/2000 02:38'!
generateDebugCode: aNode on: aStream indent: level 
	"Generate the C debug code for this message onto the given stream, if  
	compiled in debugMode."
	self generateDebugCode
		ifTrue: 
			[aStream nextPutAll: '/* DebugCode... */';
			 cr.
			aNode args first
				emitCCodeOn: aStream
				level: level
				generator: self.
			aStream tab: level.
			aStream nextPutAll: '/* ...DebugCode */']
		ifFalse: [aStream nextPutAll: '/* missing DebugCode */']! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:52'!
generateField: aNode on: aStream indent: anInteger
	self genCallOf: #fetchPointer:ofObject: with: aNode args first with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:55'!
generateFieldPut: aNode on: aStream indent: anInteger
	self genCallOf: #storePointer:ofObject:withValue:
		with: aNode args first
		with: aNode receiver
		with: aNode args second
		on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 6/29/2018 16:18'!
generateFromStack: aNode on: aStream indent: anInteger
	| idList |
	aNode args first isConstant ifFalse: [^self error: 'arg must be constant'].
	pluginFunctionsUsed add: #stackValue:.
	idList := aNode args first value.
	(1 to: idList size)
		do: [:i | 
			aStream 
				nextPutAll: (idList at: i);
				nextPutAll: ' = stackValue(';
				print: idList size - i;
				nextPut: $)]
		separatedBy: [aStream nextPut: $;; crtab: anInteger].
! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'nice 5/2/2020 11:09'!
generateIsBytes: aNode on: aStream indent: anInteger
	self genCallOf: #isBytes: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:58'!
generateIsFloat: aNode on: aStream indent: anInteger
	self genCallOf: #isFloatObject: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:58'!
generateIsIndexable: aNode on: aStream indent: anInteger
	self genCallOf: #isIndexable: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:59'!
generateIsIntegerOop: aNode on: aStream indent: anInteger
	self genCallOf: #isIntegerObject: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 18:59'!
generateIsIntegerValue: aNode on: aStream indent: anInteger
	self genCallOf: #isIntegerValue: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 19:02'!
generateIsKindOf: aNode on: aStream indent: anInteger
	self genCallOf: #is:KindOf: with: aNode receiver with: aNode args first on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'nice 5/1/2020 20:47'!
generateIsLong64s: aNode on: aStream indent: anInteger
	self genCallOf: #isLong64s: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 19:03'!
generateIsMemberOf: aNode on: aStream indent: anInteger
	self genCallOf: #is:MemberOf: with: aNode receiver with: aNode args first on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 19:04'!
generateIsPointers: aNode on: aStream indent: anInteger
	self genCallOf: #isPointers: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'nice 5/2/2020 11:06'!
generateIsShorts: aNode on: aStream indent: anInteger
	self genCallOf: #isShorts: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 19:04'!
generateIsWords: aNode on: aStream indent: anInteger
	self genCallOf: #isWords: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 19:05'!
generateIsWordsOrBytes: aNode on: aStream indent: anInteger
	self genCallOf: #isWordsOrBytes: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'acg 9/19/1999 01:56'!
generateNext: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	| varNode |
	varNode := msgNode receiver.
	varNode isVariable
		ifFalse: [ self error: 'next can only be applied to variables' ].
	aStream nextPutAll: '*'.
	aStream nextPutAll: varNode name.
	aStream nextPutAll: '++'
! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 19:07'!
generateStAt: aNode on: aStream indent: anInteger
	self genCallOf: #stObject:at: with: aNode receiver with: aNode args first on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 19:09'!
generateStAtPut: aNode on: aStream indent: anInteger
	self genCallOf: #stObject:at:put:
		with: aNode receiver
		with: aNode args first
		with: aNode args second
		on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'eem 12/5/2012 19:09'!
generateStSize: aNode on: aStream indent: anInteger
	self genCallOf: #stSizeOf: with: aNode receiver on: aStream! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'initialize' stamp: 'sr 4/8/2000 00:53'!
initialize
	super initialize.
	debugFlag := false! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'translating builtins' stamp: 'nice 5/2/2020 11:10'!
initializeCTranslationDictionary 
	"Initialize the dictionary mapping message names to actions for C code generation."

	| pairs |
	super initializeCTranslationDictionary.
	pairs := #(
		#asCInt						#generateAsCInt:on:indent:
		#asCUnsigned				#generateAsCUnsigned:on:indent:
		#asCBoolean					#generateAsCBoolean:on:indent:
		#asCDouble					#generateAsCDouble:on:indent:

		#asSmallIntegerObj			#generateAsSmallIntegerObj:on:indent:
		#asPositiveIntegerObj		#generateAsPositiveIntegerObj:on:indent:
		#asBooleanObj				#generateAsBooleanObj:on:indent:
		#asFloatObj					#generateAsFloatObj:on:indent:

		#asIf:var:					#generateAsIfVar:on:indent:
		#asIf:var:asValue:			#generateAsIfVarAsValue:on:indent:
		#asIf:var:put:				#generateAsIfVarPut:on:indent:
		#field:						#generateField:on:indent:
		#field:put:					#generateFieldPut:on:indent:
		
		#class						#generateClass:on:indent:

		#stSize						#generateStSize:on:indent:
		#stAt:						#generateStAt:on:indent:
		#stAt:put:					#generateStAtPut:on:indent:

		#asCharPtr					#generateAsCharPtr:on:indent:
		#asIntPtr					#generateAsIntPtr:on:indent:
		#cPtrAsOop					#generateCPtrAsOop:on:indent:
		#next						#generateNext:on:indent:

		#asOop:						#generateAsOop:on:indent:
		#asValue:					#generateAsValue:on:indent:

		#isFloat						#generateIsFloat:on:indent:
		#isIndexable					#generateIsIndexable:on:indent:
		#isIntegerOop				#generateIsIntegerOop:on:indent:
		#isIntegerValue				#generateIsIntegerValue:on:indent:
		"#FloatOop					#generateIsFloatValue:on:indent:"		"unused, never implemented"
		#isBytes					#generateIsBytes:on:indent:
		#isShorts					#generateIsShorts:on:indent:
		#isLong64s					#generateIsLong64s:on:indent:
		#isWords					#generateIsWords:on:indent:
		#isWordsOrBytes				#generateIsWordsOrBytes:on:indent:
		#isPointers					#generateIsPointers:on:indent:
		#isNil						#generateIsNil:on:indent:
		#isMemberOf:				#generateIsMemberOf:on:indent:
		#isKindOf:					#generateIsKindOf:on:indent:

		#fromStack:					#generateFromStack:on:indent:
		"#clone						#generateClone:on:indent:"				"unused, never implemented"
		"#new						#generateNew:on:indent:"				"unused, never implemented"
		"#new:						#generateNewSize:on:indent:"			"unused, never implemented"
		"#superclass					#generateSuperclass:on:indent:"	"unused, never implemented"
		#debugCode:					#generateDebugCode:on:indent:
	).

	1 to: pairs size by: 2 do: [:i |
		translationDict at: (pairs at: i) put: (pairs at: i + 1)].
! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'coercing' stamp: 'eem 7/27/2021 09:48'!
load: loadExpr validatingWith: testSelector on: variableOrVariables from: stackIndex

	^String streamContents:
		[:aStream |
		aStream
			nextPutAll: loadExpr;
			nextPut: $.; cr; 
			nextPutAll: '(interpreterProxy '.
		testSelector numArgs = 1
			ifTrue:
				[aStream nextPutAll: testSelector; nextPutAll: variableOrVariables]
			ifFalse:
				[testSelector keywords with: variableOrVariables do:
					[:keyword :variable|
					aStream nextPutAll: keyword; nextPutAll: variable; space]].
		aStream nextPutAll: ') ifFalse: [^interpreterProxy primitiveFailFor: PrimErrBadArgument]']! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'initialize' stamp: 'tpr 6/9/2003 16:40'!
translationMethodClass
	"return the class used to produce C translation methods from MethodNodes"
	^SmartSyntaxPluginTMethod! !

!SmartSyntaxPluginCodeGenerator methodsFor: 'transforming' stamp: 'ar 3/10/2000 17:59'!
var: varName as: aClass
	"Record the given C declaration for a global variable"

	variableDeclarations at: varName asString put: (aClass ccgDeclareCForVar: varName)! !

!VMPluginCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 1/3/2014 11:27'!
accessorChainsForMethod: method interpreterClass: interpreterClass
	inProgressSelectors := Set new.
	^[super accessorChainsForMethod: method interpreterClass: interpreterClass] ensure:
		[inProgressSelectors := nil]! !

!VMPluginCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 1/3/2014 15:02'!
accessorDepthDeterminationFollowsSelfSends
	^true! !

!VMPluginCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 8/15/2018 18:33'!
accessorsAndAssignmentsForSubMethodNamed: selector actuals: actualParameters depth: depth interpreterClass: interpreterClass into: aTrinaryBlock
	"Evaluate aTrinaryBlock with the root accessor sends, accessor sends and assignments in the sub-method named selector."

	| method map |
	(inProgressSelectors includes: selector) ifTrue:
		[^nil].
	inProgressSelectors add: selector.
	method := self methodNamed: selector.
	"this is unsatisfactory.  a pluggable scheme that asks the relevant plugin the right question would
	 be better but for now the only cross-plugin load is for loadBitBltFrom:warping: and variants."
	(#(loadBitBltFrom: loadWarpBltFrom: loadBitBltFrom:warping:) includes: selector) ifTrue:
		[(method isNil
		  or: [method definingClass ~~ BitBltSimulation]) ifTrue:
			[method := (BitBltSimulation >> selector) asTranslationMethodOfClass: TMethod]].
	method ifNil:
		[^nil].
	map := Dictionary new.
	method args do: [:var| map at: var put: depth asString, var].
	method locals do: [:var| map at: var put: depth asString, var].
	^self accessorsAndAssignmentsForMethod: (method copy renameVariablesUsing: map)
		actuals: actualParameters
		depth: depth + 1
		interpreterClass: interpreterClass
		into: aTrinaryBlock! !

!VMPluginCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 2/16/2015 13:31'!
actualsForMethod: aTMethod
	"Normal primitives have no arguments, but translated primitives do.
	 Override to answer actuals for translated primitives."
	^(aTMethod args size - 1 to: 0 by: -1) collect:
		[:i|
		 TSendNode new
			setSelector: #stackValue:
			receiver: (TVariableNode new setName: 'interpreterProxy')
			arguments: {TConstantNode new setValue: i}]! !

!VMPluginCodeGenerator methodsFor: 'C translation' stamp: 'eem 10/30/2013 14:58'!
atLeastVMProxyMajor: major minor: minor 
	^String streamContents:
		[:s|
		s	nextPutAll: 'VM_PROXY_MAJOR > '; print: major;
			nextPutAll: ' || (VM_PROXY_MAJOR == '; print: major;
			nextPutAll: ' && VM_PROXY_MINOR >= '; print: minor;
			nextPutAll: ')']! !

!VMPluginCodeGenerator methodsFor: 'utilities' stamp: 'eem 1/5/2018 00:05'!
collectAndCheckInterpreterProxyInterfaceFor: selectors verbose: beVerbose
	"self new
		collectAndCheckInterpreterProxyInterfaceFor: (InterpreterProxy selectors reject: [:s| #(initialize private) includes: (InterpreterProxy whichCategoryIncludesSelector: s)])
		verbose: true"
	| interpreterClass objectMemoryClass |
	interpreterClass := self referenceInterpreterClass.
	objectMemoryClass := self referenceObjectMemoryClass.
	^selectors collect:
		[:selector| | reference actual |
		reference := self compileToTMethodSelector: selector
						in: ((interpreterClass whichClassIncludesSelector: selector) ifNil:
							[(objectMemoryClass whichClassIncludesSelector: selector) ifNil:
								[InterpreterProxy]]).
		actual := self compileToTMethodSelector: selector in: InterpreterProxy.
		{ actual. reference } do:
			[:tMethod|
			 tMethod recordDeclarationsIn: self.
			 tMethod returnType ifNil:
				[tMethod inferReturnTypeIn: self.
				 tMethod returnType ifNil:
					[tMethod returnType: #sqInt]]].
		(reference returnType ~= actual returnType
		 or: [(1 to: reference args size) anySatisfy:
				[:i| (reference typeFor: (reference args at: i) in: self)
				  ~= (actual typeFor: (actual args at: i) in: self)]]) ifTrue:
			[self logger
				nextPutAll: 'warning, signature of InterpreterProxy>>';
				nextPutAll: selector;
				nextPutAll: ' does not match reference implementation.';
				cr.
			 beVerbose ifTrue:
				[self logger nextPutAll: 'reference:'; tab.
				 reference emitCFunctionPrototype: self logger generator: self.
				 self logger nextPutAll: 'actual:'; tab; tab.
				 actual emitCFunctionPrototype: self logger generator: self].
			 self logger flush"; halt: selector"].
		actual]! !

!VMPluginCodeGenerator methodsFor: 'public' stamp: 'eem 9/12/2020 13:13'!
computeKernelReturnTypes
	| dictionary |
	dictionary :=super computeKernelReturnTypes.
	InterpreterProxy methodsDo:
		[:method|
		(method pragmaAt: #returnTypeC:) ifNotNil:
			[:pragma|
			 dictionary at: method selector put: pragma arguments first]].
	^dictionary! !

!VMPluginCodeGenerator methodsFor: 'public' stamp: 'eem 1/23/2020 09:23'!
declareModuleName: nameString
	"add the declaration of a module name, version and local/external tag"

	self var: #moduleName declareC: 'const char *moduleName = "', nameString,' " INT_EXT'! !

!VMPluginCodeGenerator methodsFor: 'inlining' stamp: 'eem 2/16/2015 18:48'!
doInlining: inlineFlag
	"do inlining for a plugin; avoid doing it twice and make sure that
	 primitive prolog preparation is done immediately after inlining.
	 Also, since sharing between plugins means that normal pruning
	 can't be done, allow plugins that want to prune specific methods."
	inliningDone ifFalse:
		[self doBasicInlining: inlineFlag.
		 self prepareTranslatedPrimitives.
		 pluginClass methodsToBePruned do:
			[:sel|
			methods removeKey: sel].
		 inliningDone := true]! !

!VMPluginCodeGenerator methodsFor: 'C code generator' stamp: 'tpr 4/9/2013 12:17'!
emitCCodeOn: aStream doInlining: inlineFlag doAssertions: assertionFlag
	"Generate twice; the first time to collect the used functions, the second to output the used functions."
	| savedHeaders |
	savedHeaders := headerFiles copy.
	[super emitCCodeOn: NullStream new doInlining: inlineFlag doAssertions: assertionFlag]
		on: MessageNotUnderstood
		do: [:ex|
				(#(cr crtab: ensureCr peekLast space tab tab:) includes: ex message selector) ifTrue:
					[ex resume: nil].
				ex pass].
	headerFiles := savedHeaders.
	super emitCCodeOn: aStream doInlining: inlineFlag doAssertions: assertionFlag! !

!VMPluginCodeGenerator methodsFor: 'C code generator' stamp: 'eem 2/9/2021 09:00'!
emitCHeaderOn: aStream
	"Write a C file header onto the given stream, adding include files and some basic definitions."
	| standardHeaders |
	aStream nextPutAll: (self fileHeaderVersionStampForSourceClass: pluginClass); cr.

	"config.h should always go first because config.h properly defines flags.
	 One of those is _GNU_SOURCE, as explained in https://www.gnu.org/software/autoconf/manual/autoconf.html#Posix-Variants,
	 where the Autoconf macro AC_USE_SYSTEM_EXTENSIONS makes sure this is defined."
	standardHeaders := #('"config.h"' '<math.h>' '"sqMathShim.h"' '<stdio.h>' '<stdlib.h>' '<string.h>' '<time.h>').
	self emitHeaderFiles: standardHeaders on: aStream.

	headerFiles := headerFiles copyWithoutAll: standardHeaders.
	"Additional header files; include C library ones first."
	self emitHeaderFiles: (headerFiles select: [:hdr| hdr includes: $<]) on: aStream.

	aStream cr; nextPutAll: '/* Do not include the entire sq.h file but just those parts needed. */
#include "sqConfig.h"			/* Configuration options */
#include "sqVirtualMachine.h"	/*  The virtual machine proxy definition */
#include "sqPlatformSpecific.h"	/* Platform specific definitions */'; cr; cr.

	self addHeaderFile: '"sqMemoryAccess.h"'.
	"Additional header files; include squeak VM ones last"
	self emitHeaderFiles: (headerFiles reject: [:hdr| hdr includes: $<]) on: aStream.
	aStream cr; nextPutAll: '#define true 1
#define false 0
#define null 0  /* using ''null'' because nil is predefined in Think C */
#ifdef SQUEAK_BUILTIN_PLUGIN
# undef EXPORT
# define EXPORT(returnType) static returnType
# define INT_EXT "(i)"
#else
# define INT_EXT "(e)"
#endif'; cr.
	self maybePutPreambleFor: pluginClass on: aStream.
	aStream cr! !

!VMPluginCodeGenerator methodsFor: 'C code generator' stamp: 'eem 9/12/2013 14:20'!
emitCTypesOn: aStream
	"Store local type declarations on the given stream."
	(self structClassesForTranslationClasses: { pluginClass }) do:
		[:structClass|
		(structClass isAbstract not
		 and: [pluginClass shouldGenerateTypedefFor: structClass]) ifTrue:
			[structClass printTypedefOn: aStream.
			 aStream cr; cr]]! !

!VMPluginCodeGenerator methodsFor: 'C code generator' stamp: 'eem 8/2/2021 15:29'!
emitExportsOn: aStream
	"Store all the exported primitives in the form used by the internal named prim system."
	aStream cr; cr; nextPutAll:'#ifdef SQUEAK_BUILTIN_PLUGIN'.
	self emitExportsNamed: pluginClass moduleName
		pluginName: pluginClass internalModuleName
		on: aStream.
	aStream cr; nextPutAll: '#else // ifdef SQ_BUILTIN_PLUGIN'; cr; cr.
	self emitPrimitiveMetadataOn: aStream.
	aStream cr; nextPutAll: '#endif // ifdef SQ_BUILTIN_PLUGIN'; cr! !

!VMPluginCodeGenerator methodsFor: 'C translation' stamp: 'eem 1/20/2016 19:03'!
emitIfdefForPluginFunctionOption: anArrayOrSymbol on: aStream
	"See e.g. senders of atLeastVMProxyMajor:minor: or <option: #IMMUTABLITY>
	 in InterpreterProxy"
	aStream nextPutAll: '#if '; nextPutAll: (anArrayOrSymbol isSymbol
											ifTrue: [anArrayOrSymbol]
											ifFalse: [self perform: anArrayOrSymbol first
														withArguments: anArrayOrSymbol allButFirst])! !

!VMPluginCodeGenerator methodsFor: 'C code generator' stamp: 'eem 8/2/2021 15:26'!
emitPrimitiveMetadataOn: aStream 
	"Output accessor depth bytes for all primitives in the plugin.
	 This is for external primitives in Spur."
	| exports |
	(exports := self exportsNeedingMetadata) isEmpty ifTrue: [^self].
	aStream nextPutAll: '#if SPURVM'; cr.
	exports do:
		[:method| | primName depth exportFlags |
		 primName := self cFunctionNameFor: method selector.
		 depth := self accessorDepthForMethod: method.
		 exportFlags := CoInterpreter metadataFlagsForPrimitive: method compiledMethod.
		 (depth >= 0 or: [exportFlags > 0]) ifTrue:
			 ["store the metadata in a short variable; save a little space
			  by omitting depths < 0; support code supplies the default."
			 self assert: depth < 128.
			 self withOptionalConditionalDefineFor: method
				on: aStream
				do: [aStream
						nextPutAll: 'EXPORT(signed short) ';
						nextPutAll: primName;
						nextPutAll: 'Metadata = ';
						nextPutAll: (self cLiteralFor: (depth bitShift: 8) + exportFlags);
						nextPut: $;;
						cr]]].
	aStream nextPutAll: '#endif // SPURVM'; cr! !

!VMPluginCodeGenerator methodsFor: 'C translation' stamp: 'eem 11/14/2016 15:15'!
generateIntegerObjectOf: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."
	| expr mustCastToUnsigned type typeIsUnsigned |
	expr := msgNode args first.
	aStream nextPutAll: '(('.
	"Note that the default type of an integer constant in C is int.  Hence we /must/
	 cast expression to long if in the 64-bit world, since e.g. in 64-bits
		(int)(16r1FFFFF << 3) = (int)16rFFFFFFF8 = -8
	 whereas
		(long)(16r1FFFFF << 3) = (long)16rFFFFFFF8 = 4294967288."
	type := self typeFor: expr in: currentMethod.
	typeIsUnsigned := type first = $u.
	mustCastToUnsigned := typeIsUnsigned not
							  or: [(self sizeOfIntegralCType: type) < (self sizeOfIntegralCType: #usqInt)].
	mustCastToUnsigned ifTrue:
		[aStream nextPutAll: '(usqInt)'].
	self emitCExpression: expr on: aStream.
	aStream nextPutAll: ' << NumSmallIntegerTagBits) | 1)'! !

!VMPluginCodeGenerator methodsFor: 'C translation' stamp: 'eem 11/14/2016 15:12'!
generateIntegerValueOf: msgNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."

	aStream nextPut: $(.
	self emitCExpression: msgNode args first on: aStream.
	aStream nextPutAll: ' >> NumSmallIntegerTagBits)'! !

!VMPluginCodeGenerator methodsFor: 'C translation' stamp: 'eem 11/22/2019 12:41'!
generateInterpreterProxyFunctionDereference: aNode on: aStream indent: anInteger
	| pluginsToClone |
	pluginsToClone := self pluginFunctionsToClone copyWithoutAll: self selectorsThatAreGeneratedAsMacros.
	pluginsToClone isEmpty ifTrue:
		[^self].
	aStream cr.
	self withConditionalDefineOf: 'SQUEAK_BUILTIN_PLUGIN'
		comment: nil
		on: aStream
		do: [pluginsToClone do:
				[:s| | cs |
				self withGuardAgainstDefinitionOf: s on: aStream do:
					[cs := self cFunctionNameFor: s.
						self withOptionalVerbiageFor: s
							on: aStream
							do: [aStream tab: anInteger; nextPutAll: cs; nextPutAll: ' = interpreterProxy->'; nextPutAll: cs; nextPut: $;; cr]
							ifOptionalDo:
								[self
									withConditionalDefineOf: cs
									comment: nil
									on: aStream
									do: [aStream tab: anInteger; nextPutAll: cs; nextPutAll: ' = 0;'; cr]]]]]! !

!VMPluginCodeGenerator methodsFor: 'C translation' stamp: 'eem 7/25/2020 13:41'!
generateRemapOopIn: aNode on: aStream indent: level
	"Generate the C code for this message onto the given stream."
	| arm |
	"Avoid nesting #if SPURVM...#else...#endif within arms of an outer #if SPURVM...#else...#endif."
	(Notification new tag: #inRemapOopInArm; signal) ifNotNil:
		[:inRemapOopInArm|
		 ^inRemapOopInArm
			ifTrue: [self generateSpurRemapOopIn: aNode on: aStream indent: level]
			ifFalse: [self generateV3RemapOopIn: aNode on: aStream indent: level]].
	[aStream cr; nextPutAll: '#if SPURVM'; cr.
	 arm := true.
	 self generateSpurRemapOopIn: aNode on: aStream indent: level.
	 aStream ensureCr; nextPutAll: '#else /* SPURVM */'; cr.
	 arm := false.
	 self generateV3RemapOopIn: aNode on: aStream indent: level.
	 aStream cr; nextPutAll: '#endif /* SPURVM */'; cr]
		on: Notification
		do: [:ex|
			ex tag == #inRemapOopInArm
				ifTrue: [ex resume: arm]
				ifFalse: [ex pass]]! !

!VMPluginCodeGenerator methodsFor: 'C translation' stamp: 'eem 7/2/2014 15:32'!
generateSpurRemapOopIn: aNode on: aStream indent: level
	"Generate just the block argument for this message as Spur does not GC on allocation."

	aNode args second emitCCodeOn: aStream level: level generator: self! !

!VMPluginCodeGenerator methodsFor: 'C translation' stamp: 'eem 7/25/2020 13:39'!
generateV3RemapOopIn: aNode on: aStream indent: level
	"Generate call on remapOop: for the variable oopOrList (or all of the
	 variables in oopOrList) before evaluating aBlock, and restore them after.
	 This keeps the oops valid if, as V3 will, there is a GC on allocation."

	| idList |
	pluginFunctionsUsed add: #pushRemappableOop:; add: #popRemappableOop.
	idList := aNode args first nameOrValue.
	idList class == Array ifFalse: [idList := Array with: idList].
	idList do:
		[:each | 
		 aStream
			tab: level;
			nextPutAll: 'pushRemappableOop(';
			nextPutAll: each asString;
			nextPutAll: ');']
		separatedBy: [aStream cr].
	aStream cr.
	aNode args second emitCCodeOn: aStream level: level generator: self.
	idList reversed do:
		[:each |
		 aStream
			tab: level;
			nextPutAll: each asString;
			nextPutAll: ' = popRemappableOop();']
		separatedBy: [aStream cr]! !

!VMPluginCodeGenerator methodsFor: 'type inference' stamp: 'eem 11/1/2013 23:33'!
harmonizeReturnTypesIn: aSetOfTypes
	"Eliminate signed/unsigned conflicts in aSetOfTypes.  Override to
	 default to void if any one return type is void."
	^(aSetOfTypes includes: #void)
		ifTrue: [Set with: #void]
		ifFalse: [super harmonizeReturnTypesIn: aSetOfTypes]! !

!VMPluginCodeGenerator methodsFor: 'public' stamp: 'eem 2/16/2015 17:16'!
initialize
	super initialize.
	pluginFunctionsUsed := Set new.
	inliningDone := false! !

!VMPluginCodeGenerator methodsFor: 'public' stamp: 'eem 4/22/2016 10:10'!
initializeCTranslationDictionary 
	"Initialize the dictionary mapping message names to actions for C code generation."

	super initializeCTranslationDictionary.
	translationDict
		at: #expandDereferenceInterpreterProxyFunctionTable
			put: #generateInterpreterProxyFunctionDereference:on:indent:;
		at: #remapOop:in:
			put: #generateRemapOopIn:on:indent:! !

!VMPluginCodeGenerator methodsFor: 'testing' stamp: 'ar 10/7/1998 17:54'!
isGeneratingPluginCode
	^true! !

!VMPluginCodeGenerator methodsFor: 'spur primitive compilation' stamp: 'eem 5/17/2015 11:43'!
isObjectAccessor: selector given: interpreterClass
	"Override to include the selectors transformed into accessors for translated primitives.
	 InterpreterPlugin browseTranslatedPrimitives"
	^(#(asciiValue at: at:put: basicAt: size) includes: selector)
	  or: [super isObjectAccessor: selector given: interpreterClass]! !

!VMPluginCodeGenerator methodsFor: 'utilities' stamp: 'eem 8/29/2020 12:26'!
isStructSend: aTSendNode
	"Answer if the argument aTSendNode is a send of a structure accessor.
	 This is tricky.  We want
		foo bar => foo->bar
		foo bar => foo.bar
		foo bar: expr => foo->bar = expr
		foo bar: expr => foo.bar = expr
	 depending on whether foo is a struct or a pointer to a struct,
	 but only if both foo is a struct type and bar is a field accessor.
	 The tricky cases are self-sends within struct class methods.  Here we need to
	 distinguish between self-sends of ordinary methods from self sends of accessors.

	Override to avoid requiring that there be a struct accessor method for the selector."
	^aTSendNode numArgs <= 1
	   and: [(self isMacroSelector: aTSendNode selector) not
	   and: [(aTSendNode receiver structTargetKindIn: self) notNil]]! !

!VMPluginCodeGenerator methodsFor: 'public' stamp: 'tpr 1/10/2003 16:20'!
localizeGlobalVariables
"TPR - we don't do this for plugins"! !

!VMPluginCodeGenerator methodsFor: 'public' stamp: 'nice 4/21/2016 23:43'!
notePluginFunctionsUsedByMacros
	"Declare the plugin functions that are used by macros."
	#(
		isKindOfInteger: #(classLargeNegativeInteger classLargePositiveInteger fetchClassOf: isIntegerObject: )
		isIntegerObject:  #()
		isLargeIntegerObject: #(classLargeNegativeInteger classLargePositiveInteger fetchClassOf: )
		isLargeNegativeIntegerObject: 	#(classLargeNegativeInteger fetchClassOf: )
		isLargePositiveIntegerObject: 	#(classLargePositiveInteger fetchClassOf: ))
			pairsDo: [:macro :funcs |
				(pluginFunctionsUsed includes: macro) ifTrue: [pluginFunctionsUsed addAll: funcs]].! !

!VMPluginCodeGenerator methodsFor: 'public' stamp: 'eem 12/6/2012 12:58'!
noteUsedPluginFunction: aSelector
	"Remember aSelector and answer whether the call should be made through
	 interpreterProxy. If not, it will be made either directly (in an internal plugin)
	 or through a local copy of the function pointer (in an external plugin)."
	pluginFunctionsUsed add: aSelector.
	"These two are static to sqVirtualMachine.c and so
	 they must be called through the interpreterProxy."
	^#(majorVersion minorVersion) includes: aSelector! !

!VMPluginCodeGenerator methodsFor: 'public' stamp: 'eem 2/24/2016 08:26'!
pluginClass: aPluginClass
	"Set the plugin class and name when generating plugins.
	 And for run-time use, answer the name string."
	| packageId |
	pluginClass := aPluginClass.
	packageId := self shortMonticelloDescriptionForClass: pluginClass.
	(packageId beginsWith: pluginClass name) ifTrue:
		[packageId := packageId allButFirst: pluginClass name size].
	(packageId beginsWith: pluginClass moduleName) ifTrue:
		[packageId := packageId allButFirst: pluginClass moduleName size].
	^self declareModuleName: pluginClass moduleNameAndVersion, packageId! !

!VMPluginCodeGenerator methodsFor: 'public' stamp: 'eem 7/19/2013 10:25'!
pluginFunctionsToClone
	"Answer those of the used plugin functions to clone as a sorted collection.
	 Exclude those that are static to sqVirtualMachine.c and hence always need
	 to be called through interpreterProxy."

	^((pluginFunctionsUsed
		reject: [:selector| self noteUsedPluginFunction: selector])
			select: [:selector| InterpreterProxy includesSelector: selector])
				asSortedCollection! !

!VMPluginCodeGenerator methodsFor: 'C code generator' stamp: 'eem 11/22/2019 12:39'!
preDeclareInterpreterProxyOn: aStream
	"Put the necessary #defines needed before interpreterProxy.  Basically
	 internal plugins use the VM's interpreterProxy variable and external plugins
	 use their own.  Override to keep local copies of all functions in external
	 prims, and link directly in internal plugins."
	"| pcc |
	pcc := self new.
	(InterpreterProxy selectors reject: [:s| #(initialize private) includes: (InterpreterProxy whichCategoryIncludesSelector: s)]) do:
		[:s| pcc noteUsedPluginFunction: s].
	pcc preDeclareInterpreterProxyOn: Transcript.
	Transcript flush"
	| pluginFuncs |
	self notePluginFunctionsUsedByMacros.
	(pluginFuncs := self pluginFunctionsToClone) isEmpty ifTrue:
		[^super preDeclareInterpreterProxyOn: aStream].
	(pluginFuncs includesAnyOf: self selectorsThatAreGeneratedAsMacros) ifTrue:
		[self preDeclareMacrosForFastClassCheckingOn: aStream guardWith: #'SQUEAK_BUILTIN_PLUGIN'].
	pluginFuncs := pluginFuncs copyWithoutAll: self selectorsThatAreGeneratedAsMacros.
	pluginFuncs isEmpty ifTrue:
		[^self].
	pluginFuncs := self collectAndCheckInterpreterProxyInterfaceFor: pluginFuncs verbose: false.
	aStream cr; nextPutAll: '#if !!defined(SQUEAK_BUILTIN_PLUGIN)'; cr.
	pluginFuncs do:
		[:tMethod|
		self withGuardAgainstDefinitionOf: tMethod selector on: aStream do:
			[| functionName |
			functionName := self cFunctionNameFor: tMethod selector.
			aStream nextPutAll:
				((String streamContents:
						[:s|
						tMethod
							static: true;
							emitCFunctionPrototype: s generator: self])
					copyReplaceAll: functionName
					with: '(*', functionName, ')'
					tokenish: [:ch| ch = $_ or: [ch isAlphaNumeric]])]].
	aStream nextPutAll: '#else /* !!defined(SQUEAK_BUILTIN_PLUGIN) */'; cr.
	pluginFuncs do:
		[:tMethod|
		self withGuardAgainstDefinitionOf: tMethod selector on: aStream do:
			[self withOptionalVerbiageFor: tMethod selector
				on: aStream
				do: [tMethod static: false; export: false; emitCFunctionPrototype: aStream generator: self]
				ifOptionalDo:
					[aStream nextPutAll: '# define '.
					 (TSendNode new
						setSelector: tMethod selector
							receiver: (TVariableNode new setName: 'interpreterProxy')
								arguments: (tMethod args collect: [:a| TVariableNode new setName: a]))
						emitCCodeAsArgumentOn: aStream
							level: 0
								generator: self.
					 aStream nextPutAll: ' 0'; cr]]].
	aStream nextPutAll: 'extern'; cr; nextPutAll: '#endif'; cr! !

!VMPluginCodeGenerator methodsFor: 'C code generator' stamp: 'nice 2/16/2020 21:07'!
preDeclareMacrosForFastClassCheckingOn: aStream guardWith: guardMacroOrNil
	"These macros can be used to check for various cases of Integer types.
	 Since they can be defined based on existing API, this is a good trade off:
	 - avoid extending the interpreterProxy API unnecessarily
	 - provide fast type checking"
	
	"Speed-up generated code for internal plugins by using macros and fixed class indices to define this well known functionality."
	(guardMacroOrNil ifNotNil: [{'#if defined(', guardMacroOrNil, ')'. #cr}] ifNil: [#()]),
	#(	'# define isIntegerObject(oop) ((oop) & 1)'
		'# define integerObjectOf(value) ((((usqInt) value) << NumSmallIntegerTagBits) | 1)'
		'# define integerValueOf(oop) ((oop) >> NumSmallIntegerTagBits)' cr

		'# if SPURVM'
		'extern sqInt classIndexOf(sqInt);'
	"Compact class indices are hardcoded here because there is no guarantee that the pool values at generation time
	 are that of SPUR.. Make sure they are in sync with SpurMemoryManager class>>initializeCompactClassIndices"
		'#	define LargeNegativeIntegerClassIndex 32'
		'#	define LargePositiveIntegerClassIndex 33'
		'#	if BytesPerOop == 4'
		'#	  define isImmediate(oop) ((oop) & 3)'
		'#	else'
		'#	  define isImmediate(oop) ((oop) & 7)'
		'#	endif'
		'#	define isKindOfInteger(oop) (isImmediate(oop) ? isIntegerObject(oop) : (unsigned)(classIndexOf(oop) - LargeNegativeIntegerClassIndex) <= 1)'
		'#	define isLargeIntegerObject(oop) (!!isImmediate(oop) && (unsigned)(classIndexOf(oop) - LargeNegativeIntegerClassIndex) <= 1)'
		'#	define isLargeNegativeIntegerObject(oop) (!!isImmediate(oop) && classIndexOf(oop) == LargeNegativeIntegerClassIndex)'
		'#	define isLargePositiveIntegerObject(oop) (!!isImmediate(oop) && classIndexOf(oop) == LargePositiveIntegerClassIndex)'
		'# endif /* SPURVM */'),
	(guardMacroOrNil ifNotNil: [{'#endif /* defined(', guardMacroOrNil, ') */'}] ifNil: [#()]),

	"If the functionality has not been defined via macros, define default versions using existing plugin API"
	#(	cr
		'#if !!defined(isKindOfInteger)'
		'# define isLargeNegativeIntegerObject(oop) (fetchClassOf(oop) == classLargeNegativeInteger())'
		'# define isLargePositiveIntegerObject(oop) (fetchClassOf(oop) == classLargePositiveInteger())'
		'# define isLargeIntegerObject(oop) (isLargeNegativeIntegerObject(oop) || isLargePositiveIntegerObject(oop))'
		'# define isKindOfInteger(oop) (isIntegerObject(oop) || isLargeNegativeIntegerObject(oop) || isLargePositiveIntegerObject(oop))'
		'#endif' cr) do:
		[:element|
		aStream cr.
		element ~~ #cr ifTrue: [aStream nextPutAll: element]]! !

!VMPluginCodeGenerator methodsFor: 'C code generator' stamp: 'eem 9/19/2020 08:24'!
prepareTranslatedPrimitives
	"Translated primitives need their prolog and epilog adding and all
	 sends to self that should be sends to interpreterproxy changing."
	methods do:
		[:meth|
		 meth primitive > 0 ifTrue:
			[meth
				preparePrimitivePrologue;
				mapSendsFromSelfToInterpreterProxy: InterpreterProxy selectors, #(maxVal minVal)]]! !

!VMPluginCodeGenerator methodsFor: 'C code generator' stamp: 'eem 1/9/2014 10:08'!
referenceInterpreterClass
	"Define the class from which to take methods to define the interpreter proxy imports."
	^(Smalltalk classNamed: #StackInterpreter) ifNil:
		[(Smalltalk classNamed: #Interpreter) ifNil:
			[InterpreterProxy]]! !

!VMPluginCodeGenerator methodsFor: 'C code generator' stamp: 'eem 1/4/2018 23:55'!
referenceObjectMemoryClass
	"Define the class from which to take methods to define the interpreter proxy imports."
	^(Smalltalk classNamed: #Spur32BitMemoryManager) ifNil:
		[(Smalltalk classNamed: #NewObjectMemory) ifNil:
			[ObjectMemory]]! !

!VMPluginCodeGenerator methodsFor: 'C code generator' stamp: 'eem 8/28/2020 20:22'!
selectorReturnsPointerToStruct: selector "<Symbol>"
	| tMethod |
	^(tMethod := methods
					at: selector
					ifAbsent:
						[apiMethods ifNotNil:
							[apiMethods at: selector ifAbsent: []]]) notNil
	  and: [(VMStructType isTypePointerToStruct: tMethod returnType)
		or: [(pluginClass isStructType: tMethod returnType)
		and: [tMethod returnType last == $*]]]! !

!VMPluginCodeGenerator methodsFor: 'C code generator' stamp: 'eem 8/28/2020 20:23'!
selectorReturnsStruct: selector "<Symbol>"
	| tMethod |
	^(tMethod := methods
					at: selector
					ifAbsent:
						[apiMethods ifNotNil:
							[apiMethods at: selector ifAbsent: []]]) notNil
	  and: [(VMStructType isTypeStruct: tMethod returnType)
		or: [(pluginClass isStructType: tMethod returnType)
		and: [tMethod returnType last ~~ $*]]]! !

!VMPluginCodeGenerator methodsFor: 'public' stamp: 'eem 4/22/2016 09:46'!
selectorsThatAreGeneratedAsMacros
	"Answer a list of selectors that are generated as a C macro rather than as an interpreterProxy function call."
	
	^#(isKindOfInteger: isLargeIntegerObject: isLargeNegativeIntegerObject: isLargePositiveIntegerObject:)! !

!VMPluginCodeGenerator methodsFor: 'public' stamp: 'eem 11/22/2019 12:21'!
selectorsThatMayBeGeneratedAsMacros
	"Answer a list of selectors that maybe generated as a C macro rather than as an interpreterProxy function call."
	
	^self selectorsThatAreGeneratedAsMacros, #(isIntegerObject: integerValueOf: integerObjectOf: isImmediate:)! !

!VMPluginCodeGenerator methodsFor: 'utilities' stamp: 'eem 8/29/2020 12:26'!
shouldGenerateAsInterpreterProxySend: aSendNode
	"Answer if this send should be generated as interpreterProxy->foo or its moral equivalent (*).
	 (*) since we now use function pointers declared in each external plugin we only indirect through
	 interopreterProxy at plugin initialization.  But we still have to find the set of sends a plugin uses."
	(aSendNode receiver isVariable and: ['interpreterProxy' = aSendNode receiver name]) ifFalse:
		[^false].
	(self isMacroSelector: aSendNode selector) ifTrue:
		[^false].
	"Only include genuine InterpreterProxy methods, excluding things not understood
	 by InterpreterProxy and things in its initialize, private and simulation protocols."
	^(#(initialize private #'simulation only') includes: (InterpreterProxy compiledMethodAt: aSendNode selector ifAbsent: [^false]) protocol) not! !

!VMPluginCodeGenerator methodsFor: 'inlining' stamp: 'dtl 7/26/2021 10:22'!
sizeOfIntegralCType: anIntegralCType "<String>"
	"Hack; because the plugin sources are compiled either as 32 or 64 bit
	 size those types which are either 32 or 64 bits in size as 48 bits.
	 This happens to produce sane results for integer promotion."
	"N.B. Only works for values for which isIntegralCType: answers true."
	| prunedCType index |
	(anIntegralCType beginsWith: 'register ') ifTrue:
		[^self sizeOfIntegralCType: (anIntegralCType allButFirst: 9)].
	prunedCType := (anIntegralCType beginsWith: 'unsigned ')
						ifTrue: [(anIntegralCType allButFirst: 9) withBlanksTrimmed]
						ifFalse: [(anIntegralCType beginsWith: 'signed ')
									ifTrue: [(anIntegralCType allButFirst: 7) withBlanksTrimmed]
									ifFalse: [anIntegralCType]].
	
	^prunedCType asString caseOf: {
		['sqLong']		->	[8].
		['usqLong']		->	[8].
		['long long']	->	[8].
		['sqInt']			->	[6].
		['usqInt']		->	[6].
		['sqIntptr_t']	->	[6].
		['usqIntptr_t']	->	[6].
		['int']			->	[4].
		['uint32_t']		->	[4].
		['unsigned']	->	[4].
		['short']			->	[2].
		['short int']		->	[2].
		['char']			->	[1].
		['long']			->	[BytesPerWord]. "It's ambiguous on LLP64 and we'll later remove it"
		['size_t']		->	[6].
		['pid_t']			->	[6].
	}
	otherwise:
		[((anIntegralCType beginsWith: 'unsigned') "e.g. 'unsigned  : 8'"
		  and: [(anIntegralCType includesAnyOf: '[*]') not
		  and: [(index := anIntegralCType indexOf: $:) > 0]])
			ifTrue: [(Integer readFrom: (anIntegralCType copyFrom: index + 1 to: anIntegralCType size) withBlanksTrimmed readStream) + 7 // 8]
			ifFalse: [self error: 'unrecognized integral type']]! !

!VMPluginCodeGenerator methodsFor: 'private' stamp: 'eem 5/20/2010 17:01'!
storeVirtualMachineProxyHeader: categoryList on: fileName
	"Store the interpreter definitions on the given file"
	| stream |
	stream := FileStream newFileNamed: fileName.
	stream nextPutAll:
'#ifndef _SqueakVM_H
#define _SqueakVM_H

/* Increment the following number if you change the order of
   functions listed or if you remove functions */
#define VM_PROXY_MAJOR 1

/* Increment the following number if you add functions at the end */
#define VM_PROXY_MINOR 0

typedef struct VirtualMachine {
	int (*minorVersion) (void);
	int (*majorVersion) (void);
'.

	categoryList do:[:assoc|
		stream cr; crtab; nextPutAll:'/* InterpreterProxy methodsFor: ''',assoc key, ''' */'; cr; crtab.
		(self sortStrings: assoc value) do:[:sel|
			(methods at: sel) emitProxyFunctionPrototype: stream generator: self.
			stream nextPutAll: ';'; crtab]].

	stream nextPutAll:'
} VirtualMachine;

#endif /* _SqueakVM_H */
'.
	stream close.! !

!VMPluginCodeGenerator methodsFor: 'private' stamp: 'eem 2/12/2016 16:51'!
storeVirtualMachineProxyImplementation: categoryList on: fileName
	"Store the interpreter definitions on the given file"
	| stream |
	stream := FileStream newFileNamed: fileName.
	stream nextPutAll:'
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "sqVirtualMachine.h"'; cr;cr.
	stream nextPutAll:'/*** Function prototypes ***/'.

	categoryList do:[:assoc|
		stream cr; cr; nextPutAll:'/* InterpreterProxy methodsFor: ''',assoc key, ''' */'; cr.
		(self sortStrings: assoc value) do:[:sel|
			(methods at: sel) emitCFunctionPrototype: stream generator: self]].

	stream cr; nextPutAll:'struct VirtualMachine *VM = NULL;'; cr.
	stream cr; nextPutAll:
'static int majorVersion(void) {
	return VM_PROXY_MAJOR;
}

static int minorVersion(void) {
	return VM_PROXY_MINOR;
}

struct VirtualMachine* sqGetInterpreterProxy(void)
{
	if(VM) return VM;
	VM = (struct VirtualMachine *) calloc(1, sizeof(VirtualMachine));
	/* Initialize Function pointers */
	VM->majorVersion = majorVersion;
	VM->minorVersion = minorVersion;
'.
	categoryList do:[:assoc|
		stream cr; crtab; nextPutAll:'/* InterpreterProxy methodsFor: ''',assoc key, ''' */'; crtab.
		assoc value asSortedCollection do:[:sel|
		stream nextPutAll:'VM->';
			nextPutAll: (self cFunctionNameFor: sel);
			nextPutAll:' = ';
			nextPutAll: (self cFunctionNameFor: sel);
			nextPutAll:';';
			crtab]].

	stream cr; crtab; nextPutAll:'return VM;'; cr; nextPutAll:'}'; cr.
	stream close.! !

!VMPluginCodeGenerator methodsFor: 'C code generator' stamp: 'eem 8/28/2020 19:41'!
structTargetKindForDeclaration: decl "<String>"
	^(super structTargetKindForDeclaration: decl) ifNil:
		[pluginClass ifNotNil:
			[| isPointer |
			 (pluginClass isStructType: ((isPointer := decl last == $*)
											ifTrue: [decl allButLast withBlanksTrimmed]
											ifFalse: [decl])) ifTrue:
				[isPointer
					ifTrue: [#pointer]
					ifFalse: [#struct]]]]! !

!VMPluginCodeGenerator methodsFor: 'type inference' stamp: 'eem 2/17/2015 14:24'!
typeFor: aNode in: aTMethod
	"Override to provide the type for InterpreterProxy's implicit stack variable."
	aNode isVariable ifTrue:
		[^(aTMethod typeFor: aNode in: self) ifNil:
			[aNode name = 'stack'
				ifTrue: [#'sqInt *']
				ifFalse: [#sqInt]]].
	^super typeFor: aNode in: aTMethod! !

!VMPluginCodeGenerator methodsFor: 'C translation' stamp: 'eem 5/2/2018 18:35'!
withGuardAgainstDefinitionOf: selector on: aStream do: aBlock
	"Evaluate aBlock, surrounded by a define if selector is defined as a macro (i.e. by preDeclareInterpreterProxyOn:"
	(self selectorsThatMayBeGeneratedAsMacros includes: selector) ifFalse:
		[^aBlock value].
	self
		withConditionalDefineOf: (self cFunctionNameFor: selector)
		comment: nil
		on: aStream
		do: aBlock! !

!VMPluginCodeGenerator methodsFor: 'C translation' stamp: 'eem 2/13/2016 17:56'!
withOptionalVerbiageFor: selector on: aStream do: mainBlock ifOptionalDo: optionalBlock
	(InterpreterProxy >> selector pragmaAt: #option:)
		ifNil:
			[mainBlock value]
		ifNotNil:
			[:pragma|
			 self emitIfdefForPluginFunctionOption: pragma arguments first on: aStream.
			 aStream cr.
			 mainBlock value.
			 aStream nextPutAll: '#else'; cr.
			 optionalBlock value.
			 aStream nextPutAll: '#endif'; cr]! !

Object subclass: #CLiteralArray
	instanceVariableNames: 'object'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!CLiteralArray class methodsFor: 'instance creation' stamp: 'eem 6/24/2012 16:14'!
on: anArrayLiteral
	^self new setObject: anArrayLiteral; yourself! !

!CLiteralArray methodsFor: 'accessing' stamp: 'eem 6/24/2012 16:11'!
at: offset
	^object at: offset + 1! !

!CLiteralArray methodsFor: 'accessing' stamp: 'eem 12/22/2020 16:11'!
at: offset put: val
	self error: 'Literals are immutable; proceed to attempt to assign anyway...'.
	^object at: offset + 1 put: val! !

!CLiteralArray methodsFor: 'accessing' stamp: 'eem 6/24/2012 16:14'!
setObject: anArrayLiteral
	object := anArrayLiteral! !

Object subclass: #CObjectAccessor
	instanceVariableNames: 'object offset'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!CObjectAccessor commentStamp: '<historical>' prior: 0!
I am used to simulate the indexed access to any object during plugin simulation.!

CObjectAccessor subclass: #CArrayAccessor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!CArrayAccessor commentStamp: '<historical>' prior: 0!
I am used to simulate the indexed access to arrays during plugin simulation.!

!CArrayAccessor methodsFor: 'comparing' stamp: 'yo 2/9/2001 11:23'!
< other

	^ (object == other object) and: [offset < other offset].! !

!CArrayAccessor methodsFor: 'comparing' stamp: 'yo 2/9/2001 11:24'!
<= other

	^ (object == other object) and: [offset <= other offset].! !

!CArrayAccessor methodsFor: 'comparing' stamp: 'yo 2/9/2001 11:24'!
> other

	^ (object == other object) and: [offset > other offset].! !

!CArrayAccessor methodsFor: 'comparing' stamp: 'yo 2/9/2001 11:24'!
>= other

	^ (object == other object) and: [offset >= other offset].! !

!CArrayAccessor methodsFor: 'accessing' stamp: 'ar 10/9/1998 21:56'!
at: index
	^object at: index + offset + 1! !

!CArrayAccessor methodsFor: 'accessing' stamp: 'cb 9/5/2016 10:42'!
at: index ifAbsent: aBlock
	"This is a simulation-time-only convenience..."
	^object at: index + offset + 1 ifAbsent: aBlock! !

!CArrayAccessor methodsFor: 'accessing' stamp: 'ar 10/9/1998 21:56'!
at: index put: value
	^object at: index + offset + 1 put: value! !

!CArrayAccessor methodsFor: 'accessing' stamp: 'ar 12/31/2001 01:36'!
byteAt: index
	^object byteAt: index + offset + 1! !

!CArrayAccessor methodsFor: 'accessing' stamp: 'acg 9/19/1999 01:50'!
cPtrAsOop

	offset = 0 ifFalse: [self error: 'offset must be zero'].
	^object! !

!CArrayAccessor methodsFor: 'accessing' stamp: 'bak 6/24/2015 16:36'!
floatAt: index 
	^ (self at: index) isFloat
		ifTrue: [self at: index]
		ifFalse: [Float
				fromIEEE32Bit: (self at: index)]
! !

!CArrayAccessor methodsFor: 'accessing' stamp: 'ar 8/2/2007 18:00'!
long32At: index
	| idx |
	idx := (offset + index) // 4 + 1.
	"Note: This is a special hack for BitBlt."
	(idx = (object basicSize + 1)) ifTrue:[^0].
	^object basicAt: idx! !

!CArrayAccessor methodsFor: 'accessing' stamp: 'ar 8/2/2007 18:01'!
long32At: index put: value
	^object basicAt: (offset + index) // 4 + 1 put: value! !

!CArrayAccessor methodsFor: 'accessing' stamp: 'ar 10/10/1998 16:26'!
longAt: index
	| idx |
	idx := (offset + index) // 4 + 1.
	"Note: This is a special hack for BitBlt."
	(idx = (object basicSize + 1)) ifTrue:[^0].
	^object basicAt: idx! !

!CArrayAccessor methodsFor: 'accessing' stamp: 'ar 10/10/1998 16:26'!
longAt: index put: value
	^object basicAt: (offset + index) // 4 + 1 put: value! !

!CArrayAccessor methodsFor: 'accessing' stamp: 'acg 9/19/1999 01:48'!
next

	|val|
	val := self at: 0.
	offset := offset + 1.
	^val! !

!CArrayAccessor methodsFor: 'accessing' stamp: 'acg 9/19/1999 01:46'!
size

	^object size! !

CArrayAccessor subclass: #CPluggableAccessor
	instanceVariableNames: 'readBlock writeBlock objectSize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!CPluggableAccessor commentStamp: 'tpr 5/5/2003 11:49' prior: 0!
This class adds generalized block parameter access to C objects for vm simulation!

!CPluggableAccessor methodsFor: 'coercion' stamp: 'eem 11/19/2013 10:58'!
asInteger
	^self! !

!CPluggableAccessor methodsFor: 'accessing' stamp: 'ar 11/24/1998 20:45'!
at: index
	^readBlock value: object value: index + offset + 1! !

!CPluggableAccessor methodsFor: 'accessing' stamp: 'ar 11/24/1998 20:45'!
at: index put: value
	^writeBlock value: object value: index + offset + 1 value: value! !

!CPluggableAccessor methodsFor: 'initialize' stamp: 'di 7/14/2004 11:55'!
atBlock: rBlock atPutBlock: wBlock
	readBlock := rBlock.
	writeBlock := wBlock! !

!CPluggableAccessor methodsFor: 'initialize' stamp: 'eem 7/14/2015 20:53'!
objectSize: n
	objectSize := n! !

!CPluggableAccessor methodsFor: 'printing' stamp: 'eem 9/18/2018 14:19'!
printOn: aStream
	super printOn: aStream.
	(object isNil and: [readBlock notNil or: [writeBlock notNil]]) ifTrue:
		[[aStream nextPutAll: ' in '; print: (readBlock home tempAt: 2) home]
			on: Error
			do: [:ex| ]]! !

!CPluggableAccessor methodsFor: 'accessing' stamp: 'eem 7/14/2015 20:53'!
size
	^objectSize ifNil: [super size]! !

CObjectAccessor subclass: #CArrayOfLongsAccessor
	instanceVariableNames: 'objectMemory address elementByteSize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CArrayOfLongsAccessor commentStamp: 'eem 10/8/2015 12:49' prior: 0!
A CArrayOfLongsAccessor is a class that wraps an Array stored in the heap.  It maps at:[put:] into a suitably aligned and offset longAt:[put:], for accessing Arrays stored in the heap, such as the primTraceLog.

Instance Variables
	address:			<Integer>
	entryByteSize:		<Integer>
	objectMemory:		<NewCoObjectMemorySimulator|Spur64BitMMLECoSimulator|Spur64BitMMLECoSimulator|Spur64BitMMBECoSimulator|Spur64BitMMBECoSimulator>

address
	- the base address in the heap of the start of the array

entryByteSize
	- the size of an element, in bytes

objectMemory
	- the memory manager whose heap is being accessed
!

!CArrayOfLongsAccessor methodsFor: 'accessing' stamp: 'eem 10/17/2015 11:16'!
address
	^address! !

!CArrayOfLongsAccessor methodsFor: 'accessing' stamp: 'eem 10/8/2015 12:52'!
at: index
	"Map at: into a suitably aligned and offset longAt:, for accessing Arrays stored in the heap, such as the primTraceLog."
	^objectMemory longAt: index * elementByteSize + address! !

!CArrayOfLongsAccessor methodsFor: 'accessing' stamp: 'eem 10/8/2015 12:52'!
at: index put: aValue
	"Map at:put: into a suitably aligned and offset longAt:put:, for accessing Arrays stored in the heap, such as the primTraceLog."
	^objectMemory longAt: index * elementByteSize + address put: aValue! !

!CArrayOfLongsAccessor methodsFor: 'initialize-release' stamp: 'eem 10/17/2015 11:14'!
objectMemory: anObjectMemory at: anAddress
	objectMemory := anObjectMemory.
	object := anObjectMemory memory.
	offset := anAddress / anObjectMemory wordSize.
	elementByteSize := anObjectMemory wordSize.
	address := anAddress! !

CArrayOfLongsAccessor subclass: #CMethodCacheAccessor
	instanceVariableNames: 'methodCacheArray entrySize functionPointerIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CMethodCacheAccessor commentStamp: '<historical>' prior: 0!
I am used to simulate accesses to the methodCache so it can live partly in memory, partly in a Smalltalk Array.  This is necessary because in simulation function pointers are Smalltalk symbols (under simulation primitive dispatch is done via perform:).
!

!CMethodCacheAccessor methodsFor: 'accessing' stamp: 'eem 10/8/2015 12:53'!
at: index
	"The special handling of functionPointerIndex is necessary because in simulation function
	 pointers are Smalltalk symbols (under simulation primitive dispatch is done via perform:)."
	index - 1 \\ entrySize = functionPointerIndex ifTrue:
		[^methodCacheArray at: index].
	^objectMemory longAt: index * elementByteSize + address! !

!CMethodCacheAccessor methodsFor: 'accessing' stamp: 'eem 10/17/2015 07:59'!
at: index put: value
	"The special handling of functionPointerIndex is necessary because in simulation function
	 pointers are Smalltalk symbols (under simulation primitive dispatch is done via perform:)."
	(index = 16r44F and: [value = 16r1D]) ifTrue:
		[self halt].
	index - 1 \\ entrySize = functionPointerIndex ifTrue:
		[objectMemory longAt: index * elementByteSize + address put: (0 = value ifTrue: [value] ifFalse: [value identityHash]).
		 ^methodCacheArray at: index put: value].
	^objectMemory longAt: index * elementByteSize + address put: value! !

!CMethodCacheAccessor methodsFor: 'initialize-release' stamp: 'eem 10/17/2015 11:18'!
objectMemory: anObjectMemory at: anAddress array: cacheArray functionPointerIndex: fpIndex entrySize: wordsPerCacheEntry
	self objectMemory: anObjectMemory
		at: anAddress - anObjectMemory wordSize. "implicit -1 for indices in at:[put:]; the MethodCache is one-relative"
	methodCacheArray := cacheArray.
	functionPointerIndex := fpIndex - 1.
	entrySize := wordsPerCacheEntry! !

!CObjectAccessor class methodsFor: 'accessing' stamp: 'eem 9/2/2011 14:08'!
alignedByteSizeOf: aCObjectAccessor forClient: aVMClass
	"Hack; this only works if the object is actually bytes."
	^aCObjectAccessor object size! !

!CObjectAccessor class methodsFor: 'debugger' stamp: 'eem 10/17/2015 10:38'!
defaultIntegerBaseInDebugger
	^VMClass defaultIntegerBaseInDebugger! !

!CObjectAccessor class methodsFor: 'instance creation' stamp: 'ar 9/16/1998 21:36'!
on: anObject
	^self new setObject: anObject! !

!CObjectAccessor methodsFor: 'pointer arithmetic' stamp: 'dtl 2/14/2014 19:29'!
+ increment
	^self shallowCopy += increment! !

!CObjectAccessor methodsFor: 'pointer arithmetic' stamp: 'ar 10/9/1998 21:57'!
+= increment
	offset := offset + increment! !

!CObjectAccessor methodsFor: 'pointer arithmetic' stamp: 'dtl 2/14/2014 19:30'!
- decrement
	^self shallowCopy -= decrement! !

!CObjectAccessor methodsFor: 'pointer arithmetic' stamp: 'ar 10/9/1998 21:58'!
-= decrement
	offset := offset - decrement! !

!CObjectAccessor methodsFor: 'converting' stamp: 'di 7/14/2004 17:36'!
asFloatAccessor

	^ self asPluggableAccessor
		atBlock: [:obj :index | obj floatAt: index]
		atPutBlock: [:obj :index :value | obj floatAt: index put: value]! !

!CObjectAccessor methodsFor: 'converting' stamp: 'di 7/14/2004 17:36'!
asIntAccessor

	^ self asPluggableAccessor
		atBlock: [:obj :index | obj intAt: index]
		atPutBlock: [:obj :index :value | obj intAt: index put: value]! !

!CObjectAccessor methodsFor: 'converting' stamp: 'tfel 2/22/2016 09:19'!
asInteger

	^ self! !

!CObjectAccessor methodsFor: 'converting' stamp: 'acg 9/20/1999 11:08'!
asOop: aClass

	(aClass ccgCanConvertFrom: object)
		ifFalse: [^self error: 'incompatible object for autocoercion'].
	^object! !

!CObjectAccessor methodsFor: 'converting' stamp: 'di 7/14/2004 11:55'!
asPluggableAccessor
	^ (CPluggableAccessor on: object) += offset! !

!CObjectAccessor methodsFor: 'converting' stamp: 'eem 3/11/2009 10:26'!
asPluggableAccessor: accessorArray
	^((CPluggableAccessor on: object) += offset)
		atBlock: accessorArray first
		atPutBlock: accessorArray last! !

!CObjectAccessor methodsFor: 'converting' stamp: 'eem 7/25/2021 12:47'!
asVoidPointer
	^self! !

!CObjectAccessor methodsFor: 'accessing' stamp: 'ar 10/9/1998 21:56'!
at: index
	^object instVarAt: index + offset + 1! !

!CObjectAccessor methodsFor: 'accessing' stamp: 'ar 10/9/1998 21:56'!
at: index put: value
	^object instVarAt: index + offset + 1 put: value! !

!CObjectAccessor methodsFor: 'converting' stamp: 'eem 2/22/2018 09:58'!
coerceTo: cTypeString sim: interpreterSimulator

	cTypeString caseOf: {
		[#'float *']			-> [^self asFloatAccessor].
		[#'int *']			-> [^self asIntAccessor].
		['sqInputEvent*']	-> [^self] }
		otherwise: [self halt].
	^ self! !

!CObjectAccessor methodsFor: 'accessing' stamp: 'eem 10/11/2013 09:29'!
first
	^self at: 0! !

!CObjectAccessor methodsFor: 'private' stamp: 'ar 11/3/1998 22:37'!
getObject
	^object! !

!CObjectAccessor methodsFor: 'testing' stamp: 'akg 10/23/2018 21:53'!
isCArray
	"Answer a boolean indicating whether the receiver responds to the CArray interface"

	^true! !

!CObjectAccessor methodsFor: 'accessing' stamp: 'di 7/14/2004 12:13'!
isCObjectAccessor

	^ true! !

!CObjectAccessor methodsFor: 'initialize-release' stamp: 'eem 4/26/2010 13:48'!
memory: anObject offset: baseIndex
	object := anObject.
	offset := baseIndex! !

!CObjectAccessor methodsFor: 'accessing' stamp: 'yo 2/9/2001 11:23'!
object

	^ object! !

!CObjectAccessor methodsFor: 'accessing' stamp: 'yo 2/9/2001 11:23'!
offset

	^ offset
! !

!CObjectAccessor methodsFor: 'printing' stamp: 'ar 9/16/1998 21:38'!
printOn: aStream
	super printOn: aStream.
	aStream
		nextPutAll:' on: ';
		print: object.! !

!CObjectAccessor methodsFor: 'private' stamp: 'ar 10/9/1998 21:56'!
setObject: anObject
	object := anObject.
	offset := 0.! !

Object subclass: #CogCheck32BitHeapMap
	instanceVariableNames: 'pages'
	classVariableNames: 'ByteShift NumPages PageMask PageShift PageSize'
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!CogCheck32BitHeapMap commentStamp: 'eem 2/27/2013 16:26' prior: 0!
A CogCheckHeapMap is a simulation of the code in platforms/Cross/vm/sqHeapMap.c.  This is a map for leak checking that allocates 1 bit for every 4 bytes of address space.  It uses an array of pages to keep space overhead low, only allocating a page if that portion of the address space is used.  So the maximum overhead is address space size / (word size * bits per byte), or (2 raisedTo: 32) / (4 * 8) or 134,217,728 bytes.

Instance Variables
	pages:		<Array of: ByteArray>

pages
	- array of pages of bits, 1 bit per word of address space
!

!CogCheck32BitHeapMap class methodsFor: 'class initialization' stamp: 'eem 9/25/2013 17:17'!
initialize
	"self initialize"
	| wordSize bitsPerByte |
	wordSize := 4. "4 bytes per bit in the map"
	bitsPerByte := 8.
	NumPages := 256.
	PageShift := -24. "(32 - (NumPages log: 2)) negated asInteger"
	PageSize := 2 << 32 / wordSize / NumPages / bitsPerByte.
	PageMask := PageSize - 1.
	ByteShift := -5 "1 bit per 4 bytes, 8 bits per byte = 32 bytes of address space per map byte"! !

!CogCheck32BitHeapMap methodsFor: 'accessing' stamp: 'eem 2/27/2013 16:05'!
bitIndex: address
	^(address bitShift: -2) bitAnd: 7! !

!CogCheck32BitHeapMap methodsFor: 'accessing' stamp: 'eem 9/25/2013 17:17'!
byteIndex: address 
	^((address bitShift: ByteShift) bitAnd: PageMask) + 1! !

!CogCheck32BitHeapMap methodsFor: 'accessing' stamp: 'eem 2/27/2013 15:39'!
clearHeapMap
	pages do:
		[:p| p ifNotNil: [p atAllPut: 0]]! !

!CogCheck32BitHeapMap methodsFor: 'accessing' stamp: 'eem 2/27/2013 16:10'!
heapMapAtWord: address
	"answer the bit corresponding to address aBit in the map"
	^(pages at: (self pageIndex: address))
		ifNil: [0]
		ifNotNil:
			[:page|
			((page at: (self byteIndex: address))
				bitShift: 0 - (self bitIndex: address))
					bitAnd: 1]! !

!CogCheck32BitHeapMap methodsFor: 'accessing' stamp: 'eem 2/27/2013 16:08'!
heapMapAtWord: address Put: aBit
	"set the bit corresponding to address in the map to aBit"
	| pageIndex page bitIndex byte byteIndex |
	pageIndex := self pageIndex: address.
	page := pages at: pageIndex.
	page ifNil:
		[page := pages at: pageIndex put: (ByteArray new: PageSize)].
	byteIndex := self byteIndex: address.
	bitIndex := self bitIndex: address.
	byte := page at: byteIndex.
	byte := aBit = 0
				ifTrue: [byte - (byte bitAnd: 1 << bitIndex)]
				ifFalse: [byte bitOr: 1 << bitIndex].
	page at: byteIndex put: byte! !

!CogCheck32BitHeapMap methodsFor: 'instance initialization' stamp: 'eem 2/27/2013 15:39'!
initialize
	pages := Array new: 256! !

!CogCheck32BitHeapMap methodsFor: 'accessing' stamp: 'eem 9/25/2013 21:19'!
pageIndex: address
	^(address bitShift: PageShift) + 1 "32 - (pageSize log: 2)"! !

!CogCheck32BitHeapMap methodsFor: 'accessing' stamp: 'eem 2/27/2013 15:50'!
pageSize
	self shouldBeImplemented! !

Object subclass: #CogStackPageSurrogate
	instanceVariableNames: 'address stackPages memory zoneBase zoneLimit'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CogStackPageSurrogate methodsFor: 'comparing' stamp: 'eem 12/29/2008 14:30'!
= aThing
	^aThing class = self class
	and: [aThing address = address]! !

!CogStackPageSurrogate methodsFor: 'accessing' stamp: 'eem 12/29/2008 14:07'!
address
	^address! !

!CogStackPageSurrogate methodsFor: 'instance initialization' stamp: 'eem 3/6/2009 15:16'!
address: theAddress simulator: aCoInterpreterSimulator zoneBase: base zoneLimit: limit
	address := theAddress.
	stackPages := aCoInterpreterSimulator stackPages.
	memory := aCoInterpreterSimulator memory.
	zoneBase := base.
	zoneLimit := limit! !

!CogStackPageSurrogate methodsFor: 'testing' stamp: 'eem 6/2/2009 14:20'!
addressIsInPage: address
	"For assert-checking"
	^self lastAddress < address and: [address < self baseAddress]! !

!CogStackPageSurrogate methodsFor: 'accessing' stamp: 'eem 1/29/2009 09:38'!
asInteger
	^address! !

!CogStackPageSurrogate methodsFor: 'accessing' stamp: 'nice 7/8/2016 02:33'!
asIntegerPtr
	^address! !

!CogStackPageSurrogate methodsFor: 'accessing' stamp: 'eem 1/15/2011 08:26'!
asUnsignedInteger
	^address! !

!CogStackPageSurrogate methodsFor: 'accessing' stamp: 'nice 7/8/2016 02:31'!
asUnsignedIntegerPtr
	^address! !

!CogStackPageSurrogate methodsFor: 'comparing' stamp: 'eem 12/29/2008 14:30'!
hash
	^address hash! !

!CogStackPageSurrogate methodsFor: 'testing' stamp: 'eem 6/2/2009 14:19'!
isFree
	^self baseFP = 0! !

!CogStackPageSurrogate methodsFor: 'printing' stamp: 'eem 12/29/2008 14:32'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $@; print: address; nextPut: $/.
	address printOn: aStream base: 16! !

CogStackPageSurrogate subclass: #CogStackPageSurrogate32
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CogStackPageSurrogate32 commentStamp: '<historical>' prior: 0!
Here's a doit to generate most of my code:
	| offset crtab |
	offset := 1.
	crtab := String with: Character cr with: Character tab.
	CogStackPage instVarNamesAndTypesForTranslationDo:
		[:name :type| | isByte |
		isByte := type = 'unsigned char'.
		CogStackPageSurrogate32
			compile: name, crtab, '^memory unsigned', (isByte ifTrue: ['ByteAt:'] ifFalse: ['LongAt:']), ' address + ', offset printString
			classified: #accessing.
		CogStackPageSurrogate32
			compile: name, ': aValue', crtab, '^memory unsigned', (isByte ifTrue: ['ByteAt:'] ifFalse: ['LongAt:']), ' address + ', offset printString,
					' put: aValue'
			classified: #accessing.
		offset := offset + (isByte ifTrue: [1] ifFalse: [4])].
	CogStackPageSurrogate32 class compile: 'alignedByteSize', crtab, '^', (offset - 1 + 3 bitAnd: -4) printString classified: #'instance creation'
!

!CogStackPageSurrogate32 class methodsFor: 'accessing' stamp: 'eem 10/9/2015 10:52'!
alignedByteSize
	^40! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 12/28/2008 18:56'!
baseAddress
	^memory unsignedLongAt: address + 17! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 12/28/2008 18:56'!
baseAddress: aValue
	self assert: (address + 16 >= zoneBase and: [address + 19 < zoneLimit]).
	^memory unsignedLongAt: address + 17 put: aValue! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 12/28/2008 18:56'!
baseFP
	^memory unsignedLongAt: address + 13! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 12/28/2008 18:56'!
baseFP: aValue
	self assert: (address + 12 >= zoneBase and: [address + 15 < zoneLimit]).
	^memory unsignedLongAt: address + 13 put: aValue! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 12/28/2008 18:56'!
headFP
	^memory unsignedLongAt: address + 9! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 12/28/2008 18:56'!
headFP: aValue
	self assert: (address + 8 >= zoneBase and: [address + 11 < zoneLimit]).
	^memory unsignedLongAt: address + 9 put: aValue! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 12/28/2008 18:56'!
headSP
	^memory unsignedLongAt: address + 5! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 12/28/2008 18:56'!
headSP: aValue
	self assert: (address + 4 >= zoneBase and: [address + 7 < zoneLimit]).
	^memory unsignedLongAt: address + 5 put: aValue! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 12/28/2008 18:56'!
lastAddress
	^memory unsignedLongAt: address + 25! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 12/28/2008 18:56'!
lastAddress: aValue
	self assert: (address + 24 >= zoneBase and: [address + 27 < zoneLimit]).
	^memory unsignedLongAt: address + 25 put: aValue! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 12/29/2008 14:51'!
nextPage
	^stackPages surrogateAtAddress: (memory unsignedLongAt: address + 33)! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
nextPage: aValue
	self assert: (address + 32 >= zoneBase and: [address + 35 < zoneLimit]).
	memory unsignedLongAt: address + 33 put: aValue asInteger.
	^aValue! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:51'!
padToWord
	^memory longAt: address + 33! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
padToWord: aValue
	self assert: (address + 32 >= zoneBase and: [address + 35 < zoneLimit]).
	^memory longAt: address + 33 put: aValue! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 12/29/2008 14:52'!
prevPage
	^stackPages surrogateAtAddress: (memory unsignedLongAt: address + 37)! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
prevPage: aValue
	self assert: (address + 36 >= zoneBase and: [address + 39 < zoneLimit]).
	memory unsignedLongAt: address + 37 put: aValue asInteger.
	^aValue! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 12/28/2008 18:56'!
realStackLimit
	^memory unsignedLongAt: address + 21! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 12/28/2008 18:56'!
realStackLimit: aValue
	self assert: (address + 20 >= zoneBase and: [address + 23 < zoneLimit]).
	^memory unsignedLongAt: address + 21 put: aValue! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 9/16/2009 17:54'!
stackLimit
	^memory unsignedLongAt: address + 1! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
stackLimit: aValue
	self assert: (address + 0 >= zoneBase and: [address + 3 < zoneLimit]).
	^memory unsignedLongAt: address + 1 put: aValue! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 11/21/2013 12:33'!
trace
	^memory longAt: address + 29! !

!CogStackPageSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 11/21/2013 12:33'!
trace: aValue
	self assert: (address + 28 >= zoneBase and: [address + 31 < zoneLimit]).
	^memory longAt: address + 29 put: aValue! !

CogStackPageSurrogate subclass: #CogStackPageSurrogate64
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CogStackPageSurrogate64 commentStamp: '<historical>' prior: 0!
Here's a doit to generate most of my code:
	| offset crtab |
	offset := 1.
	crtab := String with: Character cr with: Character tab.
	CogStackPage instVarNamesAndTypesForTranslationDo:
		[:name :type| | isByte |
		isByte := type = 'unsigned char'.
		CogStackPageSurrogate64
			compile: name, crtab, '^memory unsigned', (isByte ifTrue: ['ByteAt:'] ifFalse: ['LongLongAt:']), ' address + ', offset printString
			classified: #accessing.
		CogStackPageSurrogate64
			compile: name, ': aValue', crtab, '^memory unsigned', (isByte ifTrue: ['ByteAt:'] ifFalse: ['LongLongAt:']), ' address + ', offset printString,
					' put: aValue'
			classified: #accessing.
		offset := offset + (isByte ifTrue: [1] ifFalse: [8])].
	CogStackPageSurrogate64 class compile: 'alignedByteSize', crtab, '^', (offset - 1 + 3 bitAnd: -4) printString classified: #'instance creation'
!

!CogStackPageSurrogate64 class methodsFor: 'accessing' stamp: 'eem 3/6/2009 09:14'!
alignedByteSize
	^80! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:51'!
baseAddress
	^memory unsignedLong64At: address + 33! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
baseAddress: aValue
	self assert: (address + 32 >= zoneBase and: [address + 39 < zoneLimit]).
	^memory unsignedLong64At: address + 33 put: aValue! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:51'!
baseFP
	^memory unsignedLong64At: address + 25! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
baseFP: aValue
	self assert: (address + 24 >= zoneBase and: [address + 31 < zoneLimit]).
	^memory unsignedLong64At: address + 25 put: aValue! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:51'!
headFP
	^memory unsignedLong64At: address + 17! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
headFP: aValue
	self assert: (address + 16 >= zoneBase and: [address + 23 < zoneLimit]).
	^memory unsignedLong64At: address + 17 put: aValue! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:51'!
headSP
	^memory unsignedLong64At: address + 9! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
headSP: aValue
	self assert: (address + 8 >= zoneBase and: [address + 15 < zoneLimit]).
	^memory unsignedLong64At: address + 9 put: aValue! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:51'!
lastAddress
	^memory unsignedLong64At: address + 49! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
lastAddress: aValue
	self assert: (address + 48 >= zoneBase and: [address + 55 < zoneLimit]).
	^memory unsignedLong64At: address + 49 put: aValue! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
nextPage
	^stackPages surrogateAtAddress: (memory unsignedLong64At: address + 65)! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
nextPage: aValue
	self assert: (address + 64 >= zoneBase and: [address + 71 < zoneLimit]).
	memory unsignedLong64At: address + 65 put: aValue asInteger.
	^aValue! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 11/17/2015 13:34'!
padToWord
	^memory longAt: address + 61! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 11/17/2015 13:34'!
padToWord: aValue
	self assert: (address + 60 >= zoneBase and: [address + 63 < zoneLimit]).
	^memory longAt: address + 61 put: aValue! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
prevPage
	^stackPages surrogateAtAddress: (memory unsignedLong64At: address + 73)! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
prevPage: aValue
	self assert: (address + 72 >= zoneBase and: [address + 79 < zoneLimit]).
	memory unsignedLong64At: address + 73 put: aValue asInteger.
	^aValue! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:51'!
realStackLimit
	^memory unsignedLong64At: address + 41! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
realStackLimit: aValue
	self assert: (address + 40 >= zoneBase and: [address + 47 < zoneLimit]).
	^memory unsignedLong64At: address + 41 put: aValue! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:51'!
stackLimit
	^memory unsignedLong64At: address + 1! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2015 10:52'!
stackLimit: aValue
	self assert: (address + 0 >= zoneBase and: [address + 7 < zoneLimit]).
	^memory unsignedLong64At: address + 1 put: aValue! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 11/17/2015 13:34'!
trace
	^memory longAt: address + 57! !

!CogStackPageSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 11/17/2015 13:34'!
trace: aValue
	self assert: (address + 56 >= zoneBase and: [address + 59 < zoneLimit]).
	^memory longAt: address + 57 put: aValue! !

Object subclass: #CogitFaker
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT-Simulation'!

!CogitFaker methodsFor: 'accessing' stamp: 'eem 1/27/2020 17:59'!
assertValidDualZoneWriteAddress: address! !

!CogitFaker methodsFor: 'accessing' stamp: 'eem 2/25/2020 19:06'!
getCodeToDataDelta
	^0! !

Object subclass: #Gnuifier
	instanceVariableNames: 'directory interpreterFilename'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-PostProcessing'!

!Gnuifier commentStamp: '<historical>' prior: 0!
My instances automate the translation of a Squeak interpreter for use with GCC.  In the specified FileDirectory, I copy 'interp.c' to 'interp.c.old'; translate a gnuified interpreter back into 'interp.c'; and save a working copy of sqGnu.h.

To gnuify an interpreter, try something like one of the following:

	(Gnuifier on: 
		((FileDirectory default 
			directoryNamed: 'src') 
				directoryNamed: 'vm') pathName) gnuify

	(Gnuifier on: 
		'powercow X:Users:werdna:Desktop:squeak:Squeak3.2a-4599 Folder:src:vm') gnuify


Structure:
 directory		FileDirectory -- specifying where I should do my gnuification

I can attempt to undo my damage with #deGnuify.!

!Gnuifier class methodsFor: 'as yet unclassified' stamp: 'eem 12/10/2008 11:49'!
on: aFilePathStringOrDirectory

	^self new setDirectory: (aFilePathStringOrDirectory isString
								ifTrue: [FileDirectory on: aFilePathStringOrDirectory]
								ifFalse: [aFilePathStringOrDirectory])! !

!Gnuifier methodsFor: 'as yet unclassified' stamp: 'acg 12/30/2001 14:16'!
copyFrom: inFileStream to: outFileStream

"convert interp.c to use GNU features"

|  inData |

	Cursor read showWhile:
		[inData := inFileStream upToEnd withSqueakLineEndings].
	Cursor write showWhile:
		[outFileStream nextPutAll: inData].
	outFileStream close! !

!Gnuifier methodsFor: 'as yet unclassified' stamp: 'eem 11/9/2007 14:45'!
deGnuify
	"Gnuifier is canged to great a gnu2-interp.c so no degnuification is necessary."
	self shouldNotImplement! !

!Gnuifier methodsFor: 'as yet unclassified' stamp: 'eem 5/17/2011 12:19'!
gnuify
	"This Gnuifier produces a file compatible with gcc 3.x and gcc 4.x.
	 We label the gnuified VM so one can identify the valid gcc version.
	 The script run on Wndows, platforms/win32/misc/gnuify, produces
	 something compatible with gcc 2.95 (gack)."
	(directory fileExists: 'gcc3x-', interpreterFilename) ifTrue:
		[directory deleteFileNamed: 'gcc3x-', interpreterFilename].
	self
		gnuifyFrom: (directory oldFileNamed: interpreterFilename)
		to: 			(VMMaker forceNewFileNamed: (directory fullNameFor: 'gcc3x-', interpreterFilename))
	
! !

!Gnuifier methodsFor: 'as yet unclassified' stamp: 'eem 12/3/2012 18:23'!
gnuifyFrom: inFileStream to: outFileStream

"convert interp.c to use GNU features"

	| inData beforeInterpret inInterpret inInterpretVars beforePrimitiveResponse inPrimitiveResponse |

	inData := inFileStream upToEnd withSqueakLineEndings.
	inFileStream close.

	"print a header"
	outFileStream
		nextPutAll: '/* This file has been post-processed for GNU C */';
		cr; cr; cr.

	beforeInterpret := true.    "whether we are before the beginning of interpret()"
	inInterpret := false.     "whether we are in the middle of interpret"
	inInterpretVars := false.    "whether we are in the variables of interpret"
	beforePrimitiveResponse := true.  "whether we are before the beginning of primitiveResponse()"
	inPrimitiveResponse := false.   "whether we are inside of primitiveResponse"
	'Gnuifying'
		displayProgressAt: Sensor cursorPoint
		from: 1 to: (inData occurrencesOf: Character cr)
		during:
			[:bar | | lineNumber |
			lineNumber := 0.
			inData linesDo:
				[ :inLine | | outLine extraOutLine |
				bar value: (lineNumber := lineNumber + 1).
				outLine := inLine. 	"print out one line for each input line; by default, print out the line that was input, but some rules modify it"
				extraOutLine := nil.   "occasionally print a second output line..."
				beforeInterpret ifTrue: [
					(inLine = '#include "sq.h"') ifTrue: [
						outLine := '#include "sqGnu.h"'. ].
					(inLine beginsWith: 'interpret(void)') ifTrue: [
						"reached the beginning of interpret"
						beforeInterpret := false.
						inInterpret := true.
						inInterpretVars := true. ] ]
				ifFalse: [
				inInterpretVars ifTrue: [
					(inLine findString: 'register struct foo * foo = &fum;') > 0 ifTrue: [
						outLine := 'register struct foo * foo FOO_REG = &fum;' ].
					(inLine findString: ' localIP;') > 0 ifTrue: [
						outLine := '	register char* localIP IP_REG;' ].
					(inLine findString: ' localFP;') > 0 ifTrue: [
						outLine := '	register char* localFP FP_REG;' ].
					(inLine findString: ' localSP;') > 0 ifTrue: [
						outLine := '	register char* localSP SP_REG;' ].
					(inLine findString: ' currentBytecode;') > 0 ifTrue: [
						outLine := '	register sqInt currentBytecode CB_REG;' ].
					inLine isEmpty ifTrue: [
						"reached end of variables"
						inInterpretVars := false.
						outLine := '    JUMP_TABLE;'.
						extraOutLine := '\#if MULTIPLEBYTECODESETS\	if(!!asserta((sizeof(jumpTable)/sizeof(jumpTable[0])) >= 512))\		error("bytecode jumpTable too small");\#endif\' withCRs]]
				ifFalse: [
				inInterpret ifTrue: [
					"working inside interpret(); translate the switch statement"
					(inLine beginsWith: '		case ') ifTrue: [
						| tokens |
						tokens := inLine findTokens: '	 :'.
						outLine := '		CASE(', tokens second, ')'.
						tokens size > 2 ifTrue:
							[(tokens allButFirst: 2) do:
								[:token| outLine := outLine, ' ', token]]].
					inLine = '			break;' ifTrue: [
						outLine := '			BREAK;' ].
					inLine = '}' ifTrue: [
						"all finished with interpret()"
						inInterpret := false. ] ]
				ifFalse: [
				beforePrimitiveResponse ifTrue: [
					(inLine beginsWith: 'primitiveResponse(') ifTrue: [
						"into primitiveResponse we go"
						beforePrimitiveResponse := false.
						inPrimitiveResponse := true.
						extraOutLine := '    PRIM_TABLE;'.  ] ]
				ifFalse: [
				inPrimitiveResponse ifTrue: [
					(inLine = '	switch (primitiveIndex) {') ifTrue: [
						extraOutLine := outLine.
						outLine := '	PRIM_DISPATCH;' ].
					(inLine = '	switch (GIV(primitiveIndex)) {') ifTrue: [
						extraOutLine := outLine.
						outLine := '	PRIM_DISPATCH;' ].
					(inLine beginsWith: '	case ') ifTrue: [
						| caseLabel |
						caseLabel := (inLine findTokens: '	 :') second.
						outLine := '	CASE(', caseLabel, ')' ].
					inLine = '}' ifTrue: [
						inPrimitiveResponse := false ] ].
				] ] ] ].

				outFileStream nextPutAll: outLine; cr.
				extraOutLine ifNotNil: [
					outFileStream nextPutAll: extraOutLine; cr ]]].

	outFileStream close! !

!Gnuifier methodsFor: 'as yet unclassified' stamp: 'eem 12/10/2008 11:46'!
interpreterFilename: aString
	interpreterFilename := aString! !

!Gnuifier methodsFor: 'as yet unclassified' stamp: 'acg 12/30/2001 14:12'!
setDirectory: aFileDirectory

	directory := aFileDirectory! !

Object subclass: #InterpreterProxy
	instanceVariableNames: 'primFailCode remapBuffer stack method argumentCount bb thisSessionID osErrorCode exceptionPC'
	classVariableNames: ''
	poolDictionaries: 'VMBasicConstants'
	category: 'VMMaker-InterpreterSimulation'!

!InterpreterProxy commentStamp: 'eem 8/24/2018 15:43' prior: 0!
This class should provide the definition of what interpreter support plugins need and can have access to. Note that the proxy related files platforms - Cross - vm - sqVirtualMachine.[ch] are kept under the git tree at http://www.github.com/opensmalltalk/vm.
The main use of the class is to support the vm simulator and the RSqueak VM, which implements as many primitives as possible in Smalltalk.!

!InterpreterProxy class methodsFor: 'translation' stamp: 'eem 7/27/2021 10:17'!
checkingStackLoads
	"Answer the checking stack loads that set the primitive failure flag if validation fails.
	 Used by a SmartSyntaxPluginTMethod to generate minimal prologue code."

	 ^#(stackFloatValue: stackIntegerValue: stackObjectValue: stackPositiveMachineIntegerValue: stackSignedMachineIntegerValue:) ! !

!InterpreterProxy class methodsFor: 'translation' stamp: 'eem 12/6/2012 15:49'!
implicitReturnTypeFor: aSelector
	"Answer the return type for methods that don't have an explicit return."
	^#sqInt! !

!InterpreterProxy class methodsFor: 'private' stamp: 'tpr 12/22/2005 17:13'!
validateProxyImplementation: anInterpreter 
	"InterpreterProxy validateProxyImplementation: Interpreter"

	| proxyClass catList |
	proxyClass := InterpreterProxy.
	catList := proxyClass organization categories copy asOrderedCollection.
	catList remove: 'initialize' ifAbsent:[].
	catList remove: 'private' ifAbsent:[].
	catList do:[:categ|
		(proxyClass organization listAtCategoryNamed: categ) do:[:selector|
			(anInterpreter canUnderstand: selector) 
				ifFalse:
					[self notifyWithLabel: selector, ' is not implemented in ', anInterpreter name]]]! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar 9/16/1998 01:40'!
argumentCountOf: methodPointer
	^methodPointer numArgs! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar (auto pragmas 12/08) 10/11/1998 03:01'!
arrayValueOf: oop
	<returnTypeC: 'void *'>
	self success: (self isWordsOrBytes: oop).
	^CArrayAccessor on: oop.! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 1/5/2018 00:02'!
become: array1 with: array2
	<returnTypeC: #sqInt>
	array1 elementsExchangeIdentityWith: array2! !

!InterpreterProxy methodsFor: 'converting' stamp: 'ar 10/10/1998 21:27'!
booleanValueOf: obj
	obj == true ifTrue:[^true].
	obj == false ifTrue:[^false].
	self primitiveFail.
	^nil! !

!InterpreterProxy methodsFor: 'private' stamp: 'ar 9/18/1998 21:11'!
byteAt: accessor
	^accessor byteAt: 0! !

!InterpreterProxy methodsFor: 'private' stamp: 'ar 9/18/1998 21:12'!
byteAt: accessor put: value
	^accessor byteAt: 0 put: value! !

!InterpreterProxy methodsFor: 'private' stamp: 'tfel 3/3/2016 15:41'!
byteAtPointer: accessor
	^ self byteAt: accessor! !

!InterpreterProxy methodsFor: 'object access' stamp: 'eem 10/18/2018 13:51'!
byteSizeOf: oop
	"Return the size of the receiver in bytes"
	^oop class isBits
		ifTrue: [oop basicSize * oop class elementSize]
		ifFalse: [(self slotSizeOf: oop) * Smalltalk wordSize]! !

!InterpreterProxy methodsFor: 'other' stamp: 'nice 4/5/2016 23:19'!
byteSwapped32IfBigEndian: anInteger
	"Note: for simulation only.
	This message is not generated in C but replaced with a macro"
	<doNotGenerate>
	self vmEndianness = 1 ifTrue: [^anInteger byteSwap32].
	^anInteger! !

!InterpreterProxy methodsFor: 'other' stamp: 'nice 4/5/2016 23:19'!
byteSwapped64IfBigEndian: anInteger
	"Note: for simulation only.
	This message is not generated in C but replaced with a macro"
	<doNotGenerate>
	self vmEndianness = 1 ifTrue: [^anInteger byteSwap64].
	^anInteger! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 9/25/2020 10:32'!
bytesPerElement: oop
	<option: #(atLeastVMProxyMajor:minor: 1 17)>
	^oop bytesPerElement! !

!InterpreterProxy methodsFor: 'simulation only' stamp: 'eem 10/13/2020 09:10'!
cCoerce: value to: cTypeString
	"Type coercion. For translation a cast will be emitted. When running in Smalltalk
	  answer a suitable wrapper for correct indexing."
	^value coerceTo: cTypeString sim: self! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 3/17/2017 11:11'!
cStringOrNullFor: oop
	"Answer either a malloced string with the null-terminated contents of oop if oop is a string,
	 or the null pointer if oop is nil, or fail.  It is the client's responsibility to free the string later."
	<returnTypeC: #'char *'>
	oop isString ifTrue: [^oop] ifFalse: [self primitiveFail. ^0]! !

!InterpreterProxy methodsFor: 'callback support' stamp: 'eem 1/23/2013 15:53'!
callbackEnter: callbackID
	"Re-enter the interpreter for executing a callback"
	<var: #callbackID type: #'sqInt *'>
	^self notYetImplementedError! !

!InterpreterProxy methodsFor: 'callback support' stamp: 'eem 5/19/2015 14:08'!
callbackLeave: cbID
	"Leave from a previous callback"
	^self notYetImplementedError! !

!InterpreterProxy methodsFor: 'object access' stamp: 'eem 8/28/2020 16:22'!
characterObjectOf: characterCode
	<option: #(atLeastVMProxyMajor:minor: 1 13)>
	<var: 'characterCode' type: #int>
	^StackInterpreter objectMemoryClass characterObjectOf: characterCode! !

!InterpreterProxy methodsFor: 'special objects' stamp: 'ar 9/16/1998 21:43'!
characterTable
	^Character characterTable! !

!InterpreterProxy methodsFor: 'object access' stamp: 'eem 10/31/2013 00:27'!
characterValueOf: aCharacter
	<option: #(atLeastVMProxyMajor:minor: 1 13)>
	^aCharacter asInteger! !

!InterpreterProxy methodsFor: 'converting' stamp: 'ar 10/4/1998 15:47'!
checkedIntegerValueOf: intOop
	(self isIntegerObject: intOop)
		ifTrue:[^self integerValueOf: intOop]
		ifFalse:[self primitiveFail. ^0].! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'eem 11/26/2007 17:47'!
classAlien
        ^Smalltalk at: #Alien ifAbsent: [nil]! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'ar 9/16/1998 20:42'!
classArray
	^Array! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'ar 9/16/1998 20:43'!
classBitmap
	^Bitmap! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'ar 9/16/1998 20:43'!
classByteArray
	^ByteArray! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'ar 9/16/1998 20:43'!
classCharacter
	^Character! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'eem 1/16/2020 20:10'!
classDoubleByteArray
	<option: #(atLeastVMProxyMajor:minor: 1 16)>
	^Smalltalk at: #DoubleByteArray ifAbsent: [nil]! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'eem 1/16/2020 20:10'!
classDoubleWordArray
	<option: #(atLeastVMProxyMajor:minor: 1 16)>
	^Smalltalk at: #DoubleWordArray ifAbsent: [nil]! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'ar 11/29/1999 22:03'!
classExternalAddress
	^Smalltalk at: #ExternalAddress ifAbsent:[nil]! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'ar 11/29/1999 22:03'!
classExternalData
	^Smalltalk at: #ExternalData ifAbsent:[nil]! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'ar 11/29/1999 22:03'!
classExternalFunction
	^Smalltalk at: #ExternalFunction ifAbsent:[nil]! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'ar 11/29/1999 22:04'!
classExternalLibrary
	^Smalltalk at: #ExternalLibrary ifAbsent:[nil]! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'ar 11/29/1999 22:04'!
classExternalStructure
	^Smalltalk at: #ExternalStructure ifAbsent:[nil]! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'ar 9/16/1998 20:43'!
classFloat
	^Float! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'eem 1/16/2020 20:15'!
classFloat32Array
	<option: #(atLeastVMProxyMajor:minor: 1 16)>
	^Smalltalk at: #FloatArray ifAbsent: [nil]! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'eem 1/16/2020 20:15'!
classFloat64Array
	<option: #(atLeastVMProxyMajor:minor: 1 16)>
	^Smalltalk at: #Float64Array ifAbsent: [nil]! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'ar 11/19/1999 14:29'!
classLargeNegativeInteger
	^LargeNegativeInteger! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'ar 9/16/1998 20:43'!
classLargePositiveInteger
	^LargePositiveInteger! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'ar 9/16/1998 20:43'!
classPoint
	^Point! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'ar 9/16/1998 20:43'!
classSemaphore
	^Semaphore! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'ar 9/16/1998 21:43'!
classSmallInteger
	^SmallInteger! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'eem 10/7/2010 12:26'!
classString
	^ByteString! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'eem 4/18/2008 15:17'!
classUnsafeAlien
        ^Smalltalk at: #UnsafeAlien ifAbsent: [nil]! !

!InterpreterProxy methodsFor: 'special classes' stamp: 'eem 1/16/2020 20:10'!
classWordArray
	<option: #(atLeastVMProxyMajor:minor: 1 16)>
	^Smalltalk at: #WordArray ifAbsent: [nil]! !

!InterpreterProxy methodsFor: 'instance creation' stamp: 'eem 01/07/2021 10:40'!
cloneObject: oop
	^oop shallowCopy! !

!InterpreterProxy methodsFor: 'BitBlt support' stamp: 'eem 1/4/2018 23:59'!
copyBits
	<returnTypeC: #sqInt>
	bb copyBits! !

!InterpreterProxy methodsFor: 'BitBlt support' stamp: 'eem 1/4/2018 23:59'!
copyBitsFrom: leftX to: rightX at: yValue
	<returnTypeC: #sqInt>
	bb copyBitsFrom: leftX to: rightX at: yValue.! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 7/16/2020 12:27'!
deny: aBooleanOrBlock
	aBooleanOrBlock value ifTrue: [AssertionFailure signal: 'Assertion failed']! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'eem 1/23/2013 15:53'!
disownVM: flags
	^self notYetImplementedError! !

!InterpreterProxy methodsFor: 'special objects' stamp: 'ar 9/16/1998 21:42'!
displayObject
	^Display! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 9/7/2020 13:39'!
error: aString
	<returnTypeC: #void>
	<var: 'aString' type: #'const char *'>
	"In the real VM this prints aString to stderr and then calls exit(-1) or abort()."
	^super error: aString! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 11/1/2007 16:05'!
failed

	^primFailCode ~= 0! !

!InterpreterProxy methodsFor: 'special objects' stamp: 'ar 9/16/1998 20:40'!
falseObject
	^false! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar (auto pragmas 12/08) 10/10/1998 21:22'!
fetchArray: fieldIndex ofObject: objectPointer
	"Fetch the instance variable at the given index of the given object. Return the address of first indexable field of resulting array object, or fail if the instance variable does not contain an indexable bytes or words object."
	"Note: May be called by translated primitive code."

	| arrayOop |
	<returnTypeC: 'void *'>
	arrayOop := self fetchPointer: fieldIndex ofObject: objectPointer.
	^ self arrayValueOf: arrayOop
! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar 9/16/1998 01:07'!
fetchClassOf: oop
	^oop class! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar (auto pragmas 12/08) 10/10/1998 21:21'!
fetchFloat: fieldIndex ofObject: objectPointer
	"Fetch the instance variable at the given index of the given object. Return the C double precision floating point value of that instance variable, or fail if it is not a Float."
	"Note: May be called by translated primitive code."

	| floatOop |
	<returnTypeC: 'double'>
	floatOop := self fetchPointer: fieldIndex ofObject: objectPointer.
	^ self floatValueOf: floatOop! !

!InterpreterProxy methodsFor: 'object access' stamp: 'tpr (auto pragmas 12/08) 3/15/2004 20:20'!
fetchInteger: fieldIndex ofObject: objectPointer
	"Note: May be called by translated primitive code."

	| intOop |
	<inline: false>
	intOop := self fetchPointer: fieldIndex ofObject: objectPointer.
	^self checkedIntegerValueOf: intOop! !

!InterpreterProxy methodsFor: 'private' stamp: 'ar 10/27/1999 14:13'!
fetchIntegerOrTruncFloat: fieldIndex ofObject: objectPointer
	"Support for BitBlt simulation only"
	| intOrFloat |
	intOrFloat := self fetchPointer: fieldIndex ofObject: objectPointer.
	(self isIntegerObject: intOrFloat) ifTrue: [^ self integerValueOf: intOrFloat].
	intOrFloat isFloat ifTrue:[^intOrFloat truncated].
	^self primitiveFail.! !

!InterpreterProxy methodsFor: 'object access' stamp: 'tpr 6/6/2005 19:11'!
fetchLong32: fieldIndex ofObject: oop
"fetchWord:ofObject: is rescinded as of VMMaker 3.8 64bit VM. This is the same code as used therein and may need revision for 64 bit cleanliness"
	^oop instVarAt: fieldIndex+1! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar 10/10/1998 16:04'!
fetchPointer: index ofObject: oop
	^oop instVarAt: index+1! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 11/8/2020 21:12'!
fileTimesInUTC
	<option: #(atLeastVMProxyMajor:minor: 1 17)>
	^(Smalltalk vmParameterAt: 48) anyMask: 128! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar (auto pragmas 12/08) 9/18/1998 20:26'!
firstFixedField: oop
	<returnTypeC:'void *'>
	^CObjectAccessor on: oop! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar (auto pragmas 12/08) 10/10/1998 16:22'!
firstIndexableField: oop
	<returnTypeC:'void *'>
	^CArrayAccessor on: oop! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 3/16/2018 15:32'!
floatArg: offset
	"Like #stackFloatValue: but access method arguments left-to-right"
	"Obsolete; was never used; replaced with methodReturnString: in the VirtualMachine struct."
	| oop |
	<returnTypeC: #double>
	self deprecated.
	oop := self methodArg: offset.
	(self isFloatObject: oop) ifFalse: [self primitiveFail. ^0.0].
	^oop! !

!InterpreterProxy methodsFor: 'converting' stamp: 'eem 3/16/2018 14:13'!
floatObjectOf: aFloat
	<returnTypeC: #sqInt> "...because answering the float argument causes the type inferencer to say this answers a float."
	<var: #aFloat type: #double>
	aFloat isFloat ifFalse: [self error:'Not a float object'].
	^aFloat! !

!InterpreterProxy methodsFor: 'converting' stamp: 'eem 11/14/2019 13:43'!
floatValueOf: oop
	<returnTypeC: #double>
	oop isFloat
		ifTrue:[^oop]
		ifFalse:[self primitiveFail. ^0.0].! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 1/5/2018 00:03'!
fullDisplayUpdate
	<returnTypeC: #sqInt>
	Display display! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 1/5/2021 15:19'!
fullGC
	<returnTypeC: #void>
	Smalltalk garbageCollect! !

!InterpreterProxy methodsFor: 'special objects' stamp: 'eem 3/14/2014 16:45'!
getDeferDisplayUpdates
	^false! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 1/23/2013 15:53'!
getInterruptPending
	^self notYetImplementedError! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 1/23/2013 15:53'!
getStackPointer
	<returnTypeC: #'sqInt *'>
	^self notYetImplementedError! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 1/15/2018 14:04'!
getThisSessionID
	"Answer a session identifier which represents the current instance of Squeak.
	The identifier is expected to be unique among all instances of Squeak on a
	network at any point in time."

	[thisSessionID = 0]
		whileTrue:
			[thisSessionID := (Random new next * (SmallInteger maxVal min: 16rFFFFFFFF)) asInteger].
	^ thisSessionID! !

!InterpreterProxy methodsFor: 'object access' stamp: 'eem 5/7/2020 19:00'!
identityHashOf: anObject
	"N.B. this method will assign an identityHash to an object that does not already have one."
	^anObject identityHash! !

!InterpreterProxy methodsFor: 'testing' stamp: 'ar 11/17/1999 22:04'!
includesBehavior: aClass ThatOf: aSuperclass
	^aClass includesBehavior: aSuperclass! !

!InterpreterProxy methodsFor: 'other' stamp: 'ar 9/16/1998 20:52'!
incrementalGC
	Smalltalk garbageCollectMost.! !

!InterpreterProxy methodsFor: 'initialize' stamp: 'eem 11/1/2007 16:05'!
initialize
	primFailCode := 0.
	remapBuffer := OrderedCollection new.
	stack := OrderedCollection new.! !

!InterpreterProxy methodsFor: 'object access' stamp: 'eem 11/18/2009 15:50'!
instanceSizeOf: classObj
	^classObj instSize! !

!InterpreterProxy methodsFor: 'instance creation' stamp: 'ar 9/16/1998 01:11'!
instantiateClass: classPointer indexableSize: size
	^size = 0 
		ifTrue:[classPointer basicNew]
		ifFalse:[classPointer basicNew: size]! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 3/16/2018 15:32'!
integerArg: offset
	"Like #stackIntegerValue: but access method arguments left-to-right"
	"Obsolete; was never used; replaced with methodReturnInteger: in the VirtualMachine struct."
	| oop |
	self deprecated.
	oop := self methodArg: offset.
	(self isIntegerObject: oop) ifFalse: [self primitiveFail. ^0].
	^oop! !

!InterpreterProxy methodsFor: 'converting' stamp: 'ar 10/10/1998 16:13'!
integerObjectOf: value
	value class == SmallInteger ifFalse:[self error:'Not a SmallInteger object'].
	^value! !

!InterpreterProxy methodsFor: 'converting' stamp: 'ar 10/10/1998 16:10'!
integerValueOf: oop
	oop class == SmallInteger ifFalse:[self error:'Not a SmallInteger'].
	^oop! !

!InterpreterProxy methodsFor: 'converting' stamp: 'eem 12/6/2012 16:01'!
ioFilename: aCharBuffer fromString: aFilenameString ofLength: filenameLength resolveAliases: aBoolean
	<var: #aCharBuffer type: #'char *'>
	<var: #aFilenameString type: #'char *'>
	"the vm has to convert aFilenameString via any canonicalization and char-mapping and put the result in aCharBuffer. This doesn't translate well in Smalltalk since we know how long strings are rather than considering them terminated by a 0 char. Do the best we can.
Note the resolveAliases flag - this is an awful artefact of OSX and Apples demented alias handling. When opening a file, the flag must be  true, when closing or renaming it must be false. Sigh."
	aCharBuffer replaceFrom:1 to: filenameLength with: aFilenameString! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'eem 12/7/2012 14:18'!
ioLoadFunction: functionName From: moduleName
	<returnTypeC: #'void *'>
	<var: #functionName type: #'char *'>
	<var: #moduleName type: #'char *'>
	"Dummy - provided by support code"
	^0! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'eem 7/22/2021 17:54'!
ioLoadFunction: functionName From: moduleName MetadataInto: metadataPtr
	<returnTypeC: #'void *'>
	<var: #functionName type: #'char *'>
	<var: #moduleName type: #'char *'>
	<var: #metadataPtr type: #'sqInt *'>
	"Dummy - provided by support code"
	^0! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'eem 12/7/2012 14:18'!
ioLoadModule: moduleNameIndex OfLength: moduleLength
	<returnTypeC: #'void *'>
	"Dummy - provided by support code"
	^0! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'eem 1/5/2021 15:25'!
ioLoadSymbol: functionNameIndex OfLength: functionLength FromModule: moduleHandle
	<returnTypeC: #'void *'>
	<var: 'moduleHandle' type: #'void *'>
	"Dummy - provided by support code"
	^0! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 1/7/2021 16:57'!
ioMicroMSecs
	<returnTypeC: #'unsigned int'>
	^Time millisecondClockValue! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 9/18/2018 15:43'!
is: oop KindOf: aString
	"InterpreterProxy new is: 42 KindOf: 'Number'"
	<var: #aString type: 'char *'>
	<returnTypeC: #sqInt>
	(Smalltalk at: aString asSymbol ifAbsent:[nil])
		ifNil:[false]
		ifNotNil: [:theClass| oop isKindOf: theClass]! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 10/28/2009 15:12'!
is: oop KindOfClass: aClass
	"InterpreterProxy new is: 42 KindOfClass: Number"
	^oop isKindOf: aClass! !

!InterpreterProxy methodsFor: 'testing' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:35'!
is: oop MemberOf: aString
	"InterpreterProxy new is: 42 MemberOf:'SmallInteger'"
	| theClass |
	<var: #aString type:'char *'>
	theClass := Smalltalk at: aString asSymbol ifAbsent:[nil].
	^theClass isNil
		ifTrue:[false]
		ifFalse:[^oop isMemberOf: theClass]! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 1/30/2014 16:06'!
isArray: oop
	^(self isNonImmediate: oop) and: [oop class instSpec = 2]! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 12/18/2018 17:45'!
isBooleanObject: oop
	<option: #(atLeastVMProxyMajor:minor: 1 15)>
	^oop == true or: [oop == false]! !

!InterpreterProxy methodsFor: 'testing' stamp: 'ar 9/16/1998 01:04'!
isBytes: oop
	^oop class isBytes! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 10/31/2013 00:27'!
isCharacterObject: oop
	<option: #(atLeastVMProxyMajor:minor: 1 13)>
	^oop isCharacter! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 10/31/2013 00:27'!
isCharacterValue: anInteger
	<option: #(atLeastVMProxyMajor:minor: 1 13)>
	^(self isIntegerValue: anInteger) and: [anInteger >= 0]! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 11/20/2014 17:56'!
isFloatObject: oop
	^oop isFloat! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 10/31/2013 00:27'!
isImmediate: anObject
	<option: #(atLeastVMProxyMajor:minor: 1 13)>
	^StackInterpreter objectMemoryClass isImmediate: anObject! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'ar 11/28/1999 19:04'!
isInMemory: address
	"Return true if the given address is in ST object memory"
	^true! !

!InterpreterProxy methodsFor: 'testing' stamp: 'acg 9/19/1999 13:11'!
isIndexable: oop
	^oop class isVariable! !

!InterpreterProxy methodsFor: 'testing' stamp: 'ar 9/16/1998 01:12'!
isIntegerObject: objectPointer
	^objectPointer class == SmallInteger! !

!InterpreterProxy methodsFor: 'testing' stamp: 'ar 9/16/1998 01:13'!
isIntegerValue: intValue
	^intValue class == SmallInteger! !

!InterpreterProxy methodsFor: 'private' stamp: 'ar 10/27/1999 14:21'!
isInterpreterProxy
	"Return true since I am not a real Interpreter simulation"
	^true! !

!InterpreterProxy methodsFor: 'testing' stamp: 'nice 4/18/2016 23:17'!
isKindOfInteger: objectOrientedPointer
	^objectOrientedPointer isInteger! !

!InterpreterProxy methodsFor: 'testing' stamp: 'nice 4/18/2016 23:37'!
isLargeIntegerObject: objectOrientedPointer
	^objectOrientedPointer isKindOf: LargePositiveInteger! !

!InterpreterProxy methodsFor: 'testing' stamp: 'nice 4/19/2016 21:45'!
isLargeNegativeIntegerObject: objectOrientedPointer
	^objectOrientedPointer isMemberOf: LargeNegativeInteger! !

!InterpreterProxy methodsFor: 'testing' stamp: 'nice 4/19/2016 21:44'!
isLargePositiveIntegerObject: objectOrientedPointer
	^objectOrientedPointer isMemberOf: LargePositiveInteger! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 9/10/2020 13:04'!
isLong64s: oop
	<option: #(atLeastVMProxyMajor:minor: 1 17)>
	^oop class isLongs! !

!InterpreterProxy methodsFor: 'testing' stamp: 'tfel 2/11/2016 09:17'!
isNonImmediate: anObject

	^ (self isImmediate: anObject) not! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 5/21/2021 15:16'!
isOopImmutable: oop
	<option: #IMMUTABILITY>
	^oop isReadOnlyObject! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 5/21/2021 15:19'!
isOopMutable: anOop
	<api>
	^anOop isReadOnlyObject not! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 10/31/2013 00:28'!
isPinned: anObject
	<option: #(atLeastVMProxyMajor:minor: 1 13)>
	^self shouldBeImplemented! !

!InterpreterProxy methodsFor: 'testing' stamp: 'ar 9/16/1998 01:04'!
isPointers: oop
	^oop class isPointers! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 12/18/2018 17:40'!
isPositiveMachineIntegerObject: oop
	<option: #(atLeastVMProxyMajor:minor: 1 15)>
	^oop isInteger and: [oop >= 0 and: [oop digitLength <= Smalltalk wordSize]]! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 9/10/2020 12:57'!
isShorts: oop
	<option: #(atLeastVMProxyMajor:minor: 1 17)>
	^oop class isShorts! !

!InterpreterProxy methodsFor: 'testing' stamp: 'ar 9/16/1998 01:16'!
isWeak: oop
	^oop class isWeak! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 9/10/2020 13:05'!
isWords: oop
	^oop class isWords! !

!InterpreterProxy methodsFor: 'testing' stamp: 'ar 9/16/1998 01:05'!
isWordsOrBytes: oop
	^(self isBytes: oop) or:[self isWords: oop]! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 9/10/2020 13:03'!
isWordsOrShorts: oop
	<option: #(atLeastVMProxyMajor:minor: 1 17)>
	^oop class isWords or: [oop class isShorts]! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'eem 1/23/2013 15:54'!
isYoung: anOop
	^self notYetImplementedError! !

!InterpreterProxy methodsFor: 'object access' stamp: 'eem 5/23/2011 17:14'!
literal: offset ofMethod: methodPointer
	^methodPointer literalAt: offset+1! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar 9/16/1998 01:40'!
literalCountOf: methodPointer
	^methodPointer numLiterals! !

!InterpreterProxy methodsFor: 'BitBlt support' stamp: 'eem 1/5/2018 00:04'!
loadBitBltFrom: bbOop
	<returnTypeC: #sqInt>
	bb := bbOop! !

!InterpreterProxy methodsFor: 'initialize' stamp: 'ar 10/3/1998 18:50'!
loadStackFrom: aContext
	self push: aContext receiver.
	method := aContext method.
	argumentCount := method numArgs.
	1 to: argumentCount do:[:i| self push: (aContext at: i) ].! !

!InterpreterProxy methodsFor: 'private' stamp: 'tfel 2/23/2016 18:34'!
long32At: arrayOffset

	^ arrayOffset long32At: 0
! !

!InterpreterProxy methodsFor: 'private' stamp: 'ar 9/18/1998 21:11'!
longAt: accessor
	^accessor longAt: 0! !

!InterpreterProxy methodsFor: 'private' stamp: 'ar 9/18/1998 21:11'!
longAt: accessor put: value
	^accessor longAt: 0 put: value! !

!InterpreterProxy methodsFor: 'other' stamp: 'tfel 2/11/2016 09:03'!
majorVersion
	^ 1! !

!InterpreterProxy methodsFor: 'instance creation' stamp: 'ar 10/10/1998 16:14'!
makePointwithxValue: xValue yValue: yValue
	(xValue class == SmallInteger and:[yValue class == SmallInteger]) 
		ifFalse:[self error:'Not SmallInteger objects'].
	^xValue@yValue! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 3/16/2018 15:35'!
methodArg: offset
	"Like #stackValue: but access method arguments left-to-right"
	"Obsolete; was never used; replaced with methodReturnBool: in the VirtualMachine struct."
	self deprecated.
	offset > argumentCount + 1 ifTrue:[^self primitiveFail].
	^self stackValue: argumentCount - offset! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar 10/7/1998 18:43'!
methodArgumentCount
	^argumentCount! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar 10/7/1998 18:43'!
methodPrimitiveIndex
	^method primitive! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 6/22/2020 22:11'!
methodReturnBool: boolean
	"Sets the return value for a method"
	self deny: self failed.
	self pop: argumentCount+1 thenPushBool: boolean.
	^0! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 6/22/2020 22:12'!
methodReturnFloat: aFloat
	"Sets the return value for a method"
	<var: 'aFloat' type: #double>
	self deny: self failed.
	self pop: argumentCount+1 thenPushFloat: aFloat.
	^0! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 6/22/2020 22:08'!
methodReturnInteger: integer
	"Sets the return value for a method"
	self deny: self failed.
	self pop: argumentCount+1 thenPushInteger: integer.
	^0! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 6/22/2020 22:11'!
methodReturnReceiver
	"Sets the return value for a method"
	self deny: self failed.
	self pop: argumentCount.
	^0! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 6/22/2020 22:10'!
methodReturnString: aCString
	"Attempt to answer a ByteString for a given C string as the result of a primitive."
	<var: 'aCString' type: #'char *'>
	self deny: self failed.
	aCString
		ifNil: [primFailCode := PrimErrOperationFailed]
		ifNotNil:
			[(self stringForCString: aCString)
				ifNil: [primFailCode := PrimErrNoMemory]
				ifNotNil: [:result| self pop: argumentCount+1 thenPush: result]].
	^0! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 6/22/2020 22:11'!
methodReturnValue: oop
	"Sets the return value for a method"
	self deny: self failed.
	self pop: argumentCount+1 thenPush: oop.
	^0! !

!InterpreterProxy methodsFor: 'other' stamp: 'tfel 2/11/2016 09:03'!
minorVersion
	^ 8! !

!InterpreterProxy methodsFor: 'special objects' stamp: 'ar 9/16/1998 20:40'!
nilObject
	^nil! !

!InterpreterProxy methodsFor: 'private' stamp: 'eem 1/23/2013 15:53'!
notYetImplementedError
	^self error: 'not yet implemented in Smalltalk'! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 3/16/2018 15:34'!
objectArg: offset
	"Like #stackObjectValue: but access method arguments left-to-right"
	"Obsolete; was never used; replaced with methodReturnFloat: in the VirtualMachine struct."
	| oop |
	self deprecated.
	oop := self methodArg: offset.
	(self isImmediate: oop) ifTrue: [self primitiveFail. ^ nil].
	^oop! !

!InterpreterProxy methodsFor: 'FFI support' stamp: 'eem 1/23/2013 15:53'!
ownVM: flags
	^self notYetImplementedError! !

!InterpreterProxy methodsFor: 'object access' stamp: 'eem 10/31/2013 00:29'!
pinObject: anObject
	<option: #(atLeastVMProxyMajor:minor: 1 13)>
	^self shouldBeImplemented! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 2/16/2015 13:56'!
pop: nItems
	<returnTypeC: #sqInt>
	1 to: nItems do:[:i| stack removeLast].! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 2/16/2015 13:57'!
pop: nItems thenPush: oop
	<returnTypeC: #sqInt>
	self pop: nItems.
	self push: oop! !

!InterpreterProxy methodsFor: 'private' stamp: 'eem 3/16/2018 14:12'!
pop: nItems thenPushBool: boolean
	<returnTypeC: #sqInt>
	self pop: nItems.
	self push: (boolean ifTrue: [self trueObject] ifFalse: [self falseObject])! !

!InterpreterProxy methodsFor: 'private' stamp: 'eem 3/16/2018 14:12'!
pop: nItems thenPushFloat: float
	<var: #float type: #double>
	<returnTypeC: #sqInt>
	self pop: nItems.
	self push: (self floatObjectOf: float)! !

!InterpreterProxy methodsFor: 'private' stamp: 'eem 3/16/2018 14:14'!
pop: nItems thenPushInteger: integer
	<returnTypeC: #sqInt>
	self pop: nItems.
	self push: (self integerObjectOf: integer)! !

!InterpreterProxy methodsFor: 'instance creation' stamp: 'ar 9/16/1998 01:14'!
popRemappableOop
	^remapBuffer removeLast! !

!InterpreterProxy methodsFor: 'converting' stamp: 'eem 1/3/2018 15:40'!
positive32BitIntegerFor: integerValue
	<returnTypeC: #sqInt>
	<var: 'integerValue' type: #'unsigned int'>
	integerValue isInteger ifFalse:[self error:'Not an Integer object'].
	^integerValue > 0
		ifTrue:[integerValue]
		ifFalse:[ (1 bitShift: 32) + integerValue]! !

!InterpreterProxy methodsFor: 'converting' stamp: 'eem 6/28/2014 19:12'!
positive32BitValueOf: oop
	<returnTypeC: #usqInt>
	oop isInteger ifFalse:[self error:'Not an integer object'].
	oop < 0 
		ifTrue:[self primitiveFail. ^0]
		ifFalse:[^oop]! !

!InterpreterProxy methodsFor: 'converting' stamp: 'nice 3/5/2015 00:05'!
positive64BitIntegerFor: integerValue
	<api>
	<returnTypeC: #sqInt> "...because answering the 64-bit argument causes the type inferencer to say this answers 64-bits."
	<var: 'integerValue' type: #usqLong>
	integerValue isInteger ifFalse:[self error:'Not an Integer object'].
	^integerValue > 0
		ifTrue:[integerValue]
		ifFalse:[ (1 bitShift: 64) + integerValue]! !

!InterpreterProxy methodsFor: 'converting' stamp: 'eem 6/28/2014 19:12'!
positive64BitValueOf: oop
	<returnTypeC: #usqLong>
	oop isInteger ifFalse:[self error:'Not an integer object'].
	oop < 0 
		ifTrue:[self primitiveFail. ^0]
		ifFalse:[^oop]! !

!InterpreterProxy methodsFor: 'converting' stamp: 'nice 7/8/2016 01:56'!
positiveMachineIntegerValueOf: oop
	<returnTypeC: #'usqIntptr_t'>
	oop isInteger ifFalse:[self error: 'Not an integer object'].
	^oop < 0 
		ifTrue: [self primitiveFail. 0]
		ifFalse: [oop]! !

!InterpreterProxy methodsFor: 'special objects' stamp: 'eem 10/29/2009 10:51'!
primitiveErrorTable
	^Smalltalk primitiveErrorTable! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 10/30/2013 14:08'!
primitiveFail
	<returnTypeC: #sqInt>
	"Set general (unspecified) primitive failure."
	self primitiveFailFor: 1! !

!InterpreterProxy methodsFor: 'other' stamp: 'tfel 4/30/2015 07:54'!
primitiveFailFor: reasonCode
	"Set specific primitive failure.
	N.B. primitiveFailFor: PrimNoErr is expected to clear the primFailCode.
	Primitive 255 is called to indicate that we are currently simulating a primitive that should fail and the VM should handle that case appropriately (if supported by the VM)."
	<primitive: 255>
	^ primFailCode := reasonCode! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 12/18/2018 17:42'!
primitiveFailForFFIException: exceptionCode at: pc
	<var: 'exceptionCode' type: #usqLong>
	<var: 'pc' type: #usqInt>
	<option: #(atLeastVMProxyMajor:minor: 1 14)>
	"Set PrimErrFFIException primitive failure and associated exceptionCode (a.k.a. 
	 osErrorCode) and exceptionPC."
	<primitive: 255>
	osErrorCode := exceptionCode.
	exceptionPC := pc.
	^primFailCode := PrimErrFFIException! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 12/18/2018 17:42'!
primitiveFailForOSError: osError
	<var: 'osError' type: #sqLong>
	<option: #(atLeastVMProxyMajor:minor: 1 14)>
	"Set PrimErrOSError primitive failure and associated osErrorCode.
	 Primitive 255 is called to indicate that we are currently simulating a primitive that should fail and the VM should handle that case appropriately (if supported by the VM)."
	<primitive: 255>
	osErrorCode := osError.
	^primFailCode := PrimErrOSError! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 11/25/2009 10:39'!
primitiveFailureCode
	^primFailCode! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar 9/16/1998 01:39'!
primitiveIndexOf: methodPointer
	^methodPointer primitive! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar 11/28/1999 17:43'!
primitiveMethod
	"Return the method an external primitive was defined in"
	^method! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 2/16/2015 14:00'!
push: object
	<returnTypeC: #sqInt>
	stack addLast: object! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 2/16/2015 13:57'!
pushBool: trueOrFalse
	<returnTypeC: #sqInt>
	(trueOrFalse == true or:[trueOrFalse == false]) ifFalse:[self error:'Not a Boolean'].
	self push: trueOrFalse! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 2/16/2015 14:07'!
pushFloat: f
	<returnTypeC: #sqInt>
	<var: #f type: 'double '>
	f isFloat ifFalse:[^self error:'Not a Float'].
	self push: f.! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 2/16/2015 13:57'!
pushInteger: integerValue
	<returnTypeC: #sqInt>
	self push: (self integerObjectOf: integerValue).! !

!InterpreterProxy methodsFor: 'instance creation' stamp: 'eem 2/16/2015 13:57'!
pushRemappableOop: oop
	<returnTypeC: #sqInt>
	remapBuffer addLast: oop! !

!InterpreterProxy methodsFor: 'callback support' stamp: 'eem 1/5/2018 00:01'!
reestablishContextPriorToCallback: callbackContext
	"callbackContext is an activation of invokeCallback:stack:registers:jmpbuf:.  Its sender
	is the interpreter's state prior to the callback.  Reestablish that state."
	<returnTypeC: #sqInt>
	self notYetImplemented! !

!InterpreterProxy methodsFor: 'callback support' stamp: 'eem 2/16/2015 14:03'!
returnAs: returnTypeOop ThroughCallback: vmCallbackContext Context: callbackMethodContext
	<returnTypeC: #sqInt>
	<var: #vmCallbackContext type: #'VMCallbackContext *'>
	"callbackMethodContext is an activation of invokeCallback:[stack:registers:jmpbuf:].
	 Its sender is the VM's state prior to the callback.  Reestablish that state (via longjmp),
	 and mark callbackMethodContext as dead."
	self notYetImplemented! !

!InterpreterProxy methodsFor: 'callback support' stamp: 'eem 1/5/2018 00:01'!
sendInvokeCallback: thunkPtr Stack: stackPtr Registers: regsPtr Jmpbuf: jmpBufPtr
	"Send the 4 argument callback message invokeCallback:stack:registers:jmpbuf:
	 to Alien class with the supplied args.  The arguments are raw C addresses
	 and are converted to integer objects on the way."
	<returnTypeC: #sqInt>
	self notYetImplemented! !

!InterpreterProxy methodsFor: 'callback support' stamp: 'eem 1/5/2018 00:01'!
sendInvokeCallbackContext: vmCallbackContext
	<var: #vmCallbackContext type: #'VMCallbackContext *'>
	"Send the calllback message to Alien class with the supplied arg(s).  Use either the
	 1 arg invokeCallbackContext: or the 4 arg invokeCallback:stack:registers:jmpbuf:
	 message, depending on what selector is installed in the specialObjectsArray.
	 Note that if invoking the legacy invokeCallback:stack:registers:jmpbuf: we pass the
	 vmCallbackContext as the jmpbuf argument (see reestablishContextPriorToCallback:).
	 The arguments are raw C addresses and are converted to integer objects on the way."
	<returnTypeC: #sqInt>
	self notYetImplemented! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 4/10/2020 15:28'!
setInterruptCheckChain: aFunction
	<returnTypeC: #'void *'>
	<var: #aFunction declareC: 'void (*aFunction)()'>
	^self notYetImplementedError! !

!InterpreterProxy methodsFor: 'other' stamp: 'tfel 3/3/2016 17:06'!
showDisplayBits: aForm Left: l Top: t Right: r Bottom: b
	aForm == Display ifTrue:[
		Display class isDeferringUpdates ifTrue: [^ nil].
		Display forceToScreen: (Rectangle left: l right: r top: t bottom: b)].! !

!InterpreterProxy methodsFor: 'callback support' stamp: 'eem 1/23/2013 15:54'!
signalNoResume: aSemaphore
	^self notYetImplementedError! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 2/16/2015 14:02'!
signalSemaphoreWithIndex: semaIndex
	<returnTypeC: #sqInt>
	((Smalltalk externalObjects) at: semaIndex) signal! !

!InterpreterProxy methodsFor: 'converting' stamp: 'ar 11/29/1999 22:01'!
signed32BitIntegerFor: integerValue
	integerValue isInteger ifFalse:[self error:'Not an Integer object'].
	^integerValue! !

!InterpreterProxy methodsFor: 'converting' stamp: 'eem 12/7/2012 11:29'!
signed32BitValueOf: oop
	<returnTypeC: #int>
	oop isInteger ifFalse:[self error:'Not an integer object'].
	^oop! !

!InterpreterProxy methodsFor: 'converting' stamp: 'eem 1/9/2014 14:02'!
signed64BitIntegerFor: integerValue
	<returnTypeC: #sqInt> "...because answering the 64-bit argument causes the type inferencer to say this answers 64-bits."
	<var: 'integerValue' type: #sqLong>
	integerValue isInteger ifFalse:[self error:'Not an Integer object'].
	^integerValue! !

!InterpreterProxy methodsFor: 'converting' stamp: 'eem 12/6/2012 16:04'!
signed64BitValueOf: oop
	<returnTypeC: #sqLong>
	oop isInteger ifFalse:[self error:'Not an integer object'].
	^oop! !

!InterpreterProxy methodsFor: 'converting' stamp: 'nice 7/13/2016 09:44'!
signedMachineIntegerValueOf: oop
	<returnTypeC: #'sqIntptr_t'>
	oop isInteger ifFalse:[self error:'Not an integer object'].
	^oop! !

!InterpreterProxy methodsFor: 'object access' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:35'!
sizeOfSTArrayFromCPrimitive: cPtr
	"Note: Only called by translated primitive code."
	<var: #cPtr type: 'void *'>
	^self shouldNotImplement! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar 10/7/1998 18:24'!
slotSizeOf: oop
	"Returns the number of slots in the receiver.
	If the receiver is a byte object, return the number of bytes.
	Otherwise return the number of words."
	^(oop basicSize) + (oop class instSize)! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar 9/16/1998 01:53'!
stObject: array at: index
	^array at: index! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar 9/16/1998 01:53'!
stObject: array at: index put: value
	^array at: index put: value! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar 10/7/1998 18:26'!
stSizeOf: oop
	"Return the number of indexable fields in the receiver"
	^oop basicSize! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'ar (auto pragmas 12/08) 9/27/1998 15:22'!
stackFloatValue: offset
	| oop |
	<returnTypeC: 'double'>
	oop := self stackValue: offset.
	(self isFloatObject: oop) ifFalse: [self primitiveFail. ^0.0].
	^oop! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'ar 9/16/1998 22:07'!
stackIntegerValue: offset
	| oop |
	oop := self stackValue: offset.
	(self isIntegerObject: oop) ifFalse: [self primitiveFail. ^0].
	^oop! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'eem 1/30/2014 12:37'!
stackObjectValue: offset
	| oop |
	oop := self stackValue: offset.
	(self isImmediate: oop) ifTrue: [self primitiveFail. ^ nil].
	^oop! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'nice 7/13/2016 09:58'!
stackPositiveMachineIntegerValue: offset
	<returnTypeC: #'usqIntptr_t'>
	^self positiveMachineIntegerValueOf: (self stackValue: offset)! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'nice 7/22/2016 00:48'!
stackSignedMachineIntegerValue: offset
	<returnTypeC: #'sqIntptr_t'>
	^self signedMachineIntegerValueOf: (self stackValue: offset)! !

!InterpreterProxy methodsFor: 'stack access' stamp: 'ar 9/16/1998 11:47'!
stackValue: offset
	^stack at: stack size - offset.! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 12/18/2018 17:41'!
statNumGCs
	<option: #(atLeastVMProxyMajor:minor: 1 14)>
	^(Smalltalk vmParameterAt: 7 "statFullGCs") + (Smalltalk vmParameterAt: 9 "statScavenges/statIncrGCs")! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar 10/25/1998 16:16'!
storeInteger: index ofObject: oop withValue: integer
	(self isIntegerValue: integer) 
		ifTrue:[^self storePointer: index ofObject: oop withValue: integer]
		ifFalse:[^self primitiveFail]! !

!InterpreterProxy methodsFor: 'object access' stamp: 'ar 10/10/1998 21:25'!
storePointer: index ofObject: oop withValue: valuePointer
	^oop instVarAt: index+1 put: valuePointer! !

!InterpreterProxy methodsFor: 'testing' stamp: 'eem 7/16/2020 12:28'!
stringForCString: aCString
	"Answer a ByteString object containing the bytes (possibly UTF-8?) in the null-terminated C string aCString."
	<option: #(atLeastVMProxyMajor:minor: 1 14)>
	<returnTypeC: #sqInt>
	<var: #aCString type: #'char *'>
	self assert: aCString isString.
	^aCString! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 2/16/2015 13:59'!
success: aBoolean
	<returnTypeC: #sqInt>
	aBoolean ifTrue: [^self].
	primFailCode = 0 ifTrue: [self primitiveFailFor: 1]! !

!InterpreterProxy methodsFor: 'other' stamp: 'ar 9/16/1998 20:51'!
superclassOf: classPointer
	^classPointer superclass! !

!InterpreterProxy methodsFor: 'initialize' stamp: 'tfel 3/30/2015 16:13'!
synthesizeStackFor: receiver with: args
	<doNotGenerate>
	self push: receiver.
	argumentCount := args size.
	1 to: argumentCount do: [:i | self push: (args at: i)].! !

!InterpreterProxy methodsFor: 'other' stamp: 'eem 7/10/2015 13:51'!
tenuringIncrementalGC
	<returnTypeC: #void>
	Smalltalk forceTenuring; garbageCollectMost! !

!InterpreterProxy methodsFor: 'instance creation' stamp: 'ar 3/6/2009 11:09'!
topRemappableOop
	"Returns the top of the remappable oop. Useful when writing loops."
	^remapBuffer last! !

!InterpreterProxy methodsFor: 'special objects' stamp: 'ar 9/16/1998 20:41'!
trueObject
	^true! !

!InterpreterProxy methodsFor: 'object access' stamp: 'eem 7/28/2014 07:12'!
unpinObject: anObject
	<option: #(atLeastVMProxyMajor:minor: 1 13)>
	^self shouldBeImplemented! !

!InterpreterProxy methodsFor: 'other' stamp: 'tpr 12/22/2005 17:48'!
vmEndianness
	"return 0 for little endian, 1 for big endian"
	^SmalltalkImage current endianness =#big ifTrue:[1] ifFalse:[0]! !

!Object class methodsFor: '*VMMaker-plugin generation' stamp: 'eem 10/2/2010 14:12'!
ccg: cg generateCoerceToOopFrom: aNode on: aStream
	"N.B.  The is used both for generation and simulation so answer the result (for interpretation)"
	^cg emitCExpression: aNode on: aStream! !

!Object class methodsFor: '*VMMaker-plugin generation' stamp: 'eem 10/3/2010 08:09'!
ccg: cg generateCoerceToValueFrom: aNode on: aStream
	"N.B.  The could be used both for generation and simulation so answer the result (for interpretation)"
	^cg emitCExpression: aNode on: aStream! !

!Object class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/18/1999 16:09'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg ccgLoad: aBlock expr: aString asKindOf: self from: anInteger! !

!Object class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/20/1999 11:12'!
ccgCanConvertFrom: anObject

	^anObject isKindOf: self! !

!Object class methodsFor: '*VMMaker-plugin generation' stamp: 'ikp 3/31/2005 14:20'!
ccgDeclareCForVar: aSymbolOrString

	^'sqInt ', aSymbolOrString! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'acg 9/19/1999 20:39'!
asIf: aClass var: aString

	|index|
	index := aClass allInstVarNames 
		indexOf: aString
		ifAbsent: [self error: 'must use instVar name'].
	^self instVarAt: index
! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'acg 10/5/1999 06:35'!
asIf: aClass var: aString asValue: someClass

	^(self asIf: aClass var: aString) asValue: someClass
! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'acg 9/19/1999 20:40'!
asIf: aClass var: aString put: aValue

	|index|
	index := aClass allInstVarNames 
		indexOf: aString
		ifAbsent: [self error: 'must use instVar name'].
	^self instVarAt: index put: aValue
! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'eem 10/2/2010 13:55'!
asOop: aClass

	(self class isVariable and: [self class instSize > 0])
		ifTrue: [self error: 'cannot auto-coerce indexable objects with named instance variables'].
	(aClass ccgCanConvertFrom: self)
		ifFalse: [self error: 'incompatible object for this coercion'].
	^(Notification new tag: #getSimulator; signal)
		ifNotNil: [:simulator| aClass ccg: simulator generateCoerceToOopFrom: self on: nil]
		ifNil: [self]! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'acg 10/5/1999 06:21'!
asValue: aClass

	^self! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'eem 10/2/2010 22:55'!
debugCode: aBlock 
	"Sending this message tells the code generator that there is debug code in 
	aBlock. Debug code will be be generated only, if the correponding flag 
	has been set by TestCodeGenerator>>generateDebugCode:.
	In ST simulation just perform the debug code."
	false ifTrue: [aBlock value]! !

!Object methodsFor: '*VMMaker-testing' stamp: 'akg 10/23/2018 21:52'!
isCArray
	"Answer a boolean indicating whether the receiver responds to the CArray interface"

	^false! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'di 7/14/2004 12:15'!
isCObjectAccessor

	^ false! !

!Object methodsFor: '*VMMaker-testing' stamp: 'eem 12/18/2019 10:21'!
isDebugger
	^false! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'eem 3/17/2014 12:25'!
isInterpreterPlugin
	^false! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'eem 2/11/2009 17:12'!
isTParseNode
	^false! !

!Object methodsFor: '*VMMaker-testing' stamp: 'akg 10/24/2018 15:55'!
isVMSimulationAddress
	"Answer a boolean indicating whether the receiver is a kind of address in the VM simulation"

	^false! !

!Object methodsFor: '*VMMaker-message handling' stamp: 'eem 8/22/2018 15:05'!
perform: aSymbol with: firstObject with: secondObject with: thirdObject with: fourthObject with: fifthObject with: sixthObject
	"Send the selector, aSymbol, to the receiver with the given arguments.
	Fail if the number of arguments expected by the selector is not six.
	Primitive. Optional. See Object documentation whatIsAPrimitive."

	<primitive: 83>
	^ self perform: aSymbol withArguments: { firstObject. secondObject. thirdObject. fourthObject. fifthObject. sixthObject }! !

!Object methodsFor: '*VMMaker-translation support' stamp: ''!
returnTypeC: typeString
	"For translation only; noop when running in Smalltalk."! !

!Object methodsFor: '*VMMaker-abbreviations' stamp: 'eem 1/26/2020 09:44'!
sn
	^self systemNavigation! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'eem 12/30/2016 14:21'!
stAt: index
	"Simulation of generateStAt:on:indent:"
	
	^(Notification new tag: #getInterpreter; signal)
		ifNotNil: [:interpreter| interpreter stObject: self at: index]
		ifNil: [self at: index]! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'eem 12/30/2016 14:20'!
stAt: index put: value
	"Simulation of generateStAtPut:on:indent:"
	
	^(Notification new tag: #getInterpreter; signal)
		ifNotNil: [:interpreter| interpreter stObject: self at: index put: value]
		ifNil: [self at: index put: value]! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'acg 1/1/2000 22:42'!
stSize

	^self size! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'acg 12/18/1999 11:31'!
suppressFailureGuards: failureGuardFlag
	"For translation only; noop when running in Smalltalk."! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'eem 6/28/2012 14:43'!
touch: something
	"For translation only; eliminated by CCodeGenerator.  Used to squash unused varable warnings."! !

!Object methodsFor: '*VMMaker-translation support' stamp: ''!
var: varSymbol declareC: declString
	"For translation only; noop when running in Smalltalk."! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'sma 3/3/2000 12:06'!
var: varSymbol type: typeString
	"For translation only; noop when running in Smalltalk."! !

!Object methodsFor: '*VMMaker-translation support' stamp: 'sma 3/3/2000 12:06'!
var: varSymbol type: typeString array: array
	"For translation only; noop when running in Smalltalk."! !

Object subclass: #ObjectProxyForTranslatedPrimitiveSimulation
	instanceVariableNames: 'interpreter oop unitSize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!ObjectProxyForTranslatedPrimitiveSimulation commentStamp: 'eem 3/11/2015 19:58' prior: 0!
An ObjectProxyForTranslatedPrimitiveSimulation is a wrapper for an object on the simulated heap that allows that "object" to be accessed from within a simulating translated primitive such as Bitmap>>compress:toByteArray:.

Instance Variables
	interpreter:		<ObjectMemory|SpurMemoryManager>
	oop:			<Integer>
	unitSize:		<Integer>

interpreter
	- the object memory class being used as the simulator's interpreterProxy

oop
	- the oop of the "object" being wrapped

unitSize
	- 1, 2, 4 or 8
!

!ObjectProxyForTranslatedPrimitiveSimulation methodsFor: 'accessing' stamp: 'eem 3/7/2015 18:49'!
at: index
	self assert: (index between: 1 and: self size).
	unitSize = 1 ifTrue: [^interpreter fetchByte: index - 1 ofObject: oop].
	unitSize = 2 ifTrue: [^interpreter fetchShort16: index - 1 ofObject: oop].
	unitSize = 4 ifTrue: [^interpreter fetchLong32: index - 1 ofObject: oop].
	unitSize = 8 ifTrue: [^interpreter fetchLong64: index - 1 ofObject: oop].
	self halt: 'Can''t handle unitSize ', unitSize printString! !

!ObjectProxyForTranslatedPrimitiveSimulation methodsFor: 'accessing' stamp: 'eem 3/7/2015 18:49'!
at: index put: val
	self assert: (index between: 1 and: self size).
	unitSize = 1 ifTrue: [^ interpreter storeByte: index - 1 ofObject: oop withValue: val].
	unitSize = 2 ifTrue: [^ interpreter storeShort16: index - 1 ofObject: oop withValue: val].
	unitSize = 4 ifTrue: [^ interpreter storeLong32: index - 1 ofObject: oop withValue: val].
	unitSize = 8 ifTrue: [^ interpreter storeLong64: index - 1 ofObject: oop withValue: val].
	self halt: 'Can''t handle unitSize ', unitSize printString! !

!ObjectProxyForTranslatedPrimitiveSimulation methodsFor: 'accessing' stamp: 'eem 3/7/2015 17:21'!
basicAt: offset
	^self at: offset! !

!ObjectProxyForTranslatedPrimitiveSimulation methodsFor: 'accessing' stamp: 'eem 3/7/2015 17:21'!
basicAt: offset put: val
	^self at: offset put: val! !

!ObjectProxyForTranslatedPrimitiveSimulation methodsFor: 'accessing' stamp: 'eem 3/7/2015 17:20'!
basicSize
	^self size! !

!ObjectProxyForTranslatedPrimitiveSimulation methodsFor: 'accessing' stamp: 'eem 3/7/2015 16:18'!
interpreter

	^ interpreter! !

!ObjectProxyForTranslatedPrimitiveSimulation methodsFor: 'accessing' stamp: 'eem 3/7/2015 16:18'!
interpreter: anObject

	interpreter := anObject! !

!ObjectProxyForTranslatedPrimitiveSimulation methodsFor: 'initialize-release' stamp: 'eem 3/7/2015 16:28'!
interpreter:  anInterpreter oop: objOop unitSize: baseTypeSize
	interpreter := anInterpreter.
	oop := objOop.
	unitSize := baseTypeSize! !

!ObjectProxyForTranslatedPrimitiveSimulation methodsFor: 'accessing' stamp: 'eem 3/7/2015 16:18'!
oop

	^ oop! !

!ObjectProxyForTranslatedPrimitiveSimulation methodsFor: 'accessing' stamp: 'eem 3/7/2015 16:18'!
oop: anObject

	oop := anObject! !

!ObjectProxyForTranslatedPrimitiveSimulation methodsFor: 'accessing' stamp: 'eem 3/7/2015 16:33'!
size
	unitSize = 1 ifTrue: [^interpreter numBytesOf: oop].
	unitSize = 2 ifTrue: [^interpreter num16BitUnitsOf: oop].
	unitSize = 4 ifTrue: [^interpreter num32BitUnitsOf: oop].
	unitSize = 8 ifTrue: [^interpreter num64BitUnitsOf: oop].
	self halt: 'Can''t handle unitSize ', unitSize printString! !

!ObjectProxyForTranslatedPrimitiveSimulation methodsFor: 'accessing' stamp: 'eem 3/7/2015 16:18'!
unitSize

	^ unitSize! !

!ObjectProxyForTranslatedPrimitiveSimulation methodsFor: 'accessing' stamp: 'eem 3/7/2015 16:18'!
unitSize: anObject

	unitSize := anObject! !

Object subclass: #Pragmatizer
	instanceVariableNames: 'package selectors classes stampCutString stampInsertString'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!Pragmatizer commentStamp: 'dtl 9/25/2010 18:26' prior: 0!
Pragmatizer is a utility for converting message send directives (such as 'self inline: true') to their corresponding pragma implementations (<inline: true>) in method source.

The #depragmatize method provides a means for reverting to message send directives. This may be useful in the event of needing to load VMMaker into an image that does not support pragmas.

This is based on an original script provided by Eliot Miranda.!

!Pragmatizer class methodsFor: 'examples' stamp: 'dtl 9/25/2010 15:15'!
convertVMMakerAndPluginsToPragmas
	"Create (unused) packages named 'VMMaker-Interpreter', 'VMMaker-Plugins',
	and 'VMMaker-SmartSyntaxPlugins'. Open a transcript, then evaluate this method.
	Some errors will be found during the conversion. Fix the problem methods
	manually in the notifier dialog, and proceed. The issues are related to declarations
	that are not the first lines of source in the method, so just move the declarations
	to the top of the method to resolve."
	
	"Pragmatizer convertVMMakerAndPluginsToPragmas"

	| selectorsToConvert packages converter |
	selectorsToConvert := #(#export: #static: #inline: #returnTypeC: #var:declareC: #var:type: ).
	packages := { 'VMMaker-Interpreter' . 'VMMaker-Plugins' . 'VMMaker-SmartSyntaxPlugins' }.
	packages do: [:package |
		converter := Pragmatizer on: package selectors: selectorsToConvert.
		converter pragmatize]! !

!Pragmatizer class methodsFor: 'instance creation' stamp: 'dtl 9/23/2010 19:31'!
on: packageName selectors: selectorList
	"Pragmatizer on: 'VMMaker-Interpreter' selectors: #(#export: #static: #inline: #returnTypeC: #var:declareC: #var:type: )"
	"Pragmatizer on: 'VMMaker-Plugins' selectors: #(#export: #static: #inline: #returnTypeC: #var:declareC: #var:type: )"
	"Pragmatizer on: 'VMMaker-SmartSyntaxPlugins' selectors: #(#export: #static: #inline: #returnTypeC: #var:declareC: #var:type: )"

	^ self new package: packageName selectors: selectorList! !

!Pragmatizer class methodsFor: 'examples' stamp: 'dtl 9/26/2010 11:03'!
revertVMMakerAndPluginsToMessageSendDeclarations
	"To revert back to the message send declarations. "
	
	"Pragmatizer revertVMMakerAndPluginsToMessageSendDeclarations"

	| selectorsToConvert packages converter substringToRemoveFromMethodStamp |

	"Optional - Edit the following to match substring to be removed from the method
	stamps to return them to their original form prior to conversion to pragmas:
	Note leading space in the string."
	substringToRemoveFromMethodStamp := ' (auto pragmas dtl 2010-09-26)'.

	selectorsToConvert := #(#export: #static: #inline: #returnTypeC: #var:declareC: #var:type: ).
	packages := { 'VMMaker-Interpreter' . 'VMMaker-Plugins' . 'VMMaker-SmartSyntaxPlugins' }.
	packages do: [:package |
		converter := Pragmatizer on: package selectors: selectorsToConvert.
		converter stampCutString: substringToRemoveFromMethodStamp.
		converter stampInsertString: ''.
		converter depragmatize]! !

!Pragmatizer methodsFor: 'accessing' stamp: 'dtl 9/27/2010 13:41'!
classes: listOfClasses
	"If set, only the classes in this list will be modified. The classes are assumed to be
	classes in the package. This is intended to allow specifying one or more plugins
	in VMMaker-Plugins to be converted individually."

	classes := listOfClasses! !

!Pragmatizer methodsFor: 'convert pragmas to sends' stamp: 'dtl 9/27/2010 11:37'!
depragmatize
	"Convert methods from pragmas to old style message send directives, retaining
	original author initials and time stamps. Answer the methods that were modified."

	| instanceMethods |
	"Only instance side methods are translated, so ignore class side methods in the package"
	instanceMethods := package methods
				reject: [:e | e classIsMeta
						or: [self isExcludedClass: e]].
	^ instanceMethods
		reject: [:mr | self depragmatize: mr]! !

!Pragmatizer methodsFor: 'convert pragmas to sends' stamp: 'dtl 9/25/2010 12:51'!
depragmatize: methodReference
	"Depragmatize the method. Answer true if the method was not modified, and false
	if conversion was performed."

	| method newSource newStamp pragmas stamp |
	pragmas := OrderedCollection new.
	method := methodReference compiledMethod.
	method properties keysAndValuesDo: [:sel :p |
		(selectors includes: sel)
			ifTrue: [pragmas add: p]].
	pragmas isEmpty ifTrue: [^ true].
	newSource := pragmas
		inject: method getSourceFromFile asString
		into: [:src : pragma |
				src copyReplaceAll: pragma asString
					with: 'self ', pragma asString allButFirst allButLast, '.'].
		stamp := method timeStamp.
		newStamp := self newConvertToMessageSendStamp: stamp.
	Transcript
		print: method methodClass;
		nextPutAll: '>>';
		print: method selector;
		nextPut: Character space;
		nextPutAll: newStamp;
		cr; flush.
	method methodClass
		compile: newSource
		classified: methodReference category
		withStamp: newStamp
		notifying: nil.
	^false
! !

!Pragmatizer methodsFor: 'testing' stamp: 'dtl 9/27/2010 13:44'!
isExcludedClass: aMethodReference

	classes notNil
		and: [classes
				detect: [:cls | cls name = aMethodReference classSymbol]
				ifNone: [^ true]].
	^ false! !

!Pragmatizer methodsFor: 'method stamp' stamp: 'dtl 9/25/2010 13:06'!
newConvertToMessageSendStamp: oldStamp
	"Insert a comment in a method stamp to indicate that conversion from pragmas to old
	style message sends was done. If stampCutString has been set, remove this substring
	from the oldStamp prior to adding a new substring. If stampInsertString was set, use
	this as the substring to insert. Otherwise add a substring with current date and author
	initials." 

	| newStamp |
	stampCutString
		ifNil: [newStamp := oldStamp]
		ifNotNil: [newStamp := oldStamp copyReplaceAll: stampCutString with: ''].
	stampInsertString
		ifNil: [newStamp := newStamp
						copyReplaceFrom: (oldStamp indexOf: Character space) + 1
						to: (oldStamp indexOf: Character space)
						with: '(remove pragmas ', Utilities authorInitials, ' ', Date today yyyymmdd, ') ']
		ifNotNil: [newStamp := newStamp
						copyReplaceFrom: (oldStamp indexOf: Character space) + 1
						to: (oldStamp indexOf: Character space)
						with: stampInsertString].
	^ newStamp
! !

!Pragmatizer methodsFor: 'method stamp' stamp: 'dtl 9/25/2010 13:06'!
newConvertToPragmaStamp: oldStamp
	"Insert a comment in a method stamp to indicate that conversion to pragmas was done.
	If stampCutString has been set, remove this substring from the oldStamp prior to adding
	a new substring. If stampInsertString was set, use this as the substring to insert. Otherwise
	add a substring with current date and author initials." 

	| newStamp |
	stampCutString
		ifNil: [newStamp := oldStamp]
		ifNotNil: [newStamp := oldStamp copyReplaceAll: stampCutString with: ''].
	stampInsertString
		ifNil: [newStamp := newStamp
						copyReplaceFrom: (oldStamp indexOf: Character space) + 1
						to: (oldStamp indexOf: Character space)
						with: '(auto pragmas ', Utilities authorInitials, ' ', Date today yyyymmdd, ') ']
		ifNotNil: [newStamp := newStamp
						copyReplaceFrom: (oldStamp indexOf: Character space) + 1
						to: (oldStamp indexOf: Character space)
						with: stampInsertString].
	^ newStamp
! !

!Pragmatizer methodsFor: 'convert sends to pragmas' stamp: 'dtl 9/25/2010 18:05'!
newPragmatizedSourceFrom: source nodes: pragmaNodes nodeRanges: ranges

	^ String streamContents: [:str|
		str nextPutAll: (source copyFrom: 1 to: ranges first - 6).
		pragmaNodes
			do: [:node| | pragma args |
				str nextPut: $<.
				str nextPutAll: (node asString allButFirst: 6) allButLast.
				str nextPut: $>]
			separatedBy: [str crtab: 1].
		str nextPutAll: (source copyFrom: ((ranges last + 1 <= source size
										   and: [(source at: ranges last + 1) = $.])
											ifTrue: [ranges last + 2]
											ifFalse: [ranges last + 1])
							to: source size)]
! !

!Pragmatizer methodsFor: 'initialize-release' stamp: 'dtl 9/20/2010 18:57'!
package: packageName selectors: selectorList

	package := PackageOrganizer default
		packageNamed: packageName
		ifAbsent: [self error: 'package ', packageName, ' not found'].
	selectors := selectorList! !

!Pragmatizer methodsFor: 'convert sends to pragmas' stamp: 'dtl 9/27/2010 11:36'!
pragmatize
	"Convert methods to use pragmas, retaining original author initials and time stamps.
	Answer the methods that were modified."

	| instanceMethods |
	"Only instance side methods are translated, so ignore class side methods in the package"
	instanceMethods := package methods
				reject: [:e | e classIsMeta
						or: [self isExcludedClass: e]].
	^ instanceMethods
		reject: [:mr | self pragmatize: mr]! !

!Pragmatizer methodsFor: 'convert sends to pragmas' stamp: 'dtl 9/25/2010 17:26'!
pragmatize: methodReference
	"Pragmatize the method. Answer true if the method was not modified, and false
	if conversion was performed."

	| method methodNode pragmaNodes sourceRanges ranges source newSource stamp newStamp |
	method := methodReference compiledMethod.
	methodNode := method methodNode.
	pragmaNodes := SortedCollection
						sortBlock: [:a :b | (sourceRanges at: a) first <= (sourceRanges at: b) first].
	(sourceRanges := methodNode rawSourceRanges)
		keysAndValuesDo: [:node :range | (node isMessageNode
					and: [selectors includes: node selector key])
				ifTrue: [pragmaNodes add: node]].
	^ pragmaNodes isEmpty
		ifFalse: [ranges := (sourceRanges at: pragmaNodes first) first
								to: (sourceRanges at: pragmaNodes last) last.

				source := method getSourceFromFile asString.
				"methodBodyStart := method methodClass parserClass new
										parseMethodComment: source setPattern: [:ignored|];
										startOfNextToken.
				tempsStart := source indexOf: $| startingAt: methodBodyStart.
				hasTemps := tempsStart > 0 and: [tempsStart < ranges first]."
				(source copyFrom: ranges first - 5 to: ranges first - 1) ~= 'self ' ifTrue: [self halt].
				(ranges last < source size
					and: [(source indexOf: $. startingAt: ranges last) > (ranges last + 1)]) ifTrue: [self halt].
				newSource := self newPragmatizedSourceFrom: source
							nodes: pragmaNodes
							nodeRanges: ranges.
				stamp := method timeStamp.
				newStamp := self newConvertToPragmaStamp: stamp.
				Transcript
					print: method methodClass;
					nextPutAll: '>>';
					print: method selector;
					nextPut: Character space;
					nextPutAll: newStamp;
					cr; flush.
				method methodClass
					compile: newSource
					classified: methodReference category
					withStamp: newStamp
					notifying: nil.
				false]
! !

!Pragmatizer methodsFor: 'accessing' stamp: 'dtl 9/25/2010 11:19'!
stampCutString: stringToRemoveFromMethodStamp
	"If set, this substring will be removed from method stamps during conversion. This is
	intended to enable conversion to and from pragmas without repeated extension of the
	method stamp."

	stampCutString := stringToRemoveFromMethodStamp! !

!Pragmatizer methodsFor: 'accessing' stamp: 'dtl 9/25/2010 12:49'!
stampInsertString: commentToInsert
	"If set, the commentToInsert will be used for method stamp modification
	rather than the default conversion value. This is intended to permit a
	Pragmatizer to reproduce a previous conversion, e.g. the previously converted
	methods in the Cog VM."

	stampInsertString := commentToInsert
! !

Object subclass: #ReadOnlyArrayWrapper
	instanceVariableNames: 'theArray'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!ReadOnlyArrayWrapper class methodsFor: 'instance creation' stamp: 'eem 7/3/2008 11:05'!
around: anArray
	^self new array: anArray; yourself! !

!ReadOnlyArrayWrapper methodsFor: 'accessing' stamp: 'eem 7/3/2008 11:04'!
array
	^theArray! !

!ReadOnlyArrayWrapper methodsFor: 'accessing' stamp: 'eem 7/3/2008 11:04'!
array: anArray
	theArray := anArray! !

!ReadOnlyArrayWrapper methodsFor: 'accessing' stamp: 'eem 7/3/2008 11:05'!
at: index
	^theArray at: index! !

!ReadOnlyArrayWrapper methodsFor: 'accessing' stamp: 'eem 7/3/2008 11:06'!
at: index put: value
	"Oh where is imutability when you need it??"
	self error: 'attempt to modify memory!!!!'! !

!ReadOnlyArrayWrapper methodsFor: 'accessing' stamp: 'eem 7/3/2008 11:05'!
size
	^theArray size! !

Object subclass: #SimulatorHarness
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: 'VMObjectIndices'
	category: 'VMMaker-Support'!

!SimulatorHarness commentStamp: 'eem 6/29/2017 12:12' prior: 0!
SimulatorHarness provides machinery for executing code within a simulator that is initialized with an image but not up and running.!

!SimulatorHarness class methodsFor: 'debugger' stamp: 'eem 2/9/2020 10:00'!
defaultIntegerBaseInDebugger
	^VMClass defaultIntegerBaseInDebugger! !

!SimulatorHarness methodsFor: 'testing' stamp: 'eem 11/17/2014 19:39'!
deny: aBooleanOrBlock
	aBooleanOrBlock value ifTrue: [AssertionFailure signal: 'Assertion failed']! !

!SimulatorHarness methodsFor: 'execution' stamp: 'eem 9/5/2014 11:44'!
interpreter: sim object: receiver perform: selector withArguments: arguments
	"Interpret an expression in oldHeap using oldInterpreter.
	 Answer the result."
	| fp savedpc savedsp savedStackPages result startByteCount |
	self assert: ({receiver. selector}, arguments allSatisfy:
					[:oop| oop isInteger and: [sim objectMemory addressCouldBeOop: oop]]).
	savedpc := sim localIP.
	savedsp := sim localSP.
	savedStackPages := Set with: sim stackPage.
	sim internalPush: receiver.
	arguments do: [:arg| sim internalPush: arg].
	sim
		argumentCount: arguments size;
		messageSelector: selector.
	fp := sim localFP.
	startByteCount := sim byteCount.
	"sim byteCount = 66849 ifTrue: [self halt]."
	sim normalSend.
	sim incrementByteCount. "otherwise, send is not counted"
	["sim printFrame: sim localFP WithSP: sim localSP"
	 "sim setBreakSelector: #elementsForwardIdentityTo:"
	 "sim byteCount = 66849 ifTrue: [self halt]."
	 "(sim byteCount > 7508930 and: [sim localFP = -16r27894]) ifTrue:
		[self halt]."
	 fp = sim localFP] whileFalse:
		[sim singleStep.
		 (savedStackPages includes: sim stackPage) ifFalse: "If the stack gets deep something has probably gone wrong..."
			[savedStackPages size > 20 ifTrue: [self halt].
			 savedStackPages add: sim stackPage]].
	result := sim internalPopStack.
	self assert: savedsp = sim localSP.
	self assert: sim localIP - 1 = savedpc.
	sim localIP: savedpc.
	^result! !

!SimulatorHarness methodsFor: 'execution' stamp: 'eem 6/29/2017 12:39'!
withExecutableInterpreter: sim do: aBlock
	"With the oldInterpreter ready to execute code, evaluate aBlock,
	 then return the interpreter (and the heap) to the ``just snapshotted'' state."
	| savedpc savedfp initialContext finalContext |
	sim
		initStackPages;
		loadInitialContext;
		internalizeIPandSP.
	savedpc := sim localIP.
	savedfp := sim localFP.
	"sim printHeadFrame."
	aBlock cull: sim cull: self.
	"sim printHeadFrame."
	sim
		internalPush: sim localIP;
		externalizeIPandSP.
	"now undo the execution state"
	self assert: sim localFP = savedfp.
	initialContext := sim frameContext: savedfp.
	finalContext := sim voidVMStateForSnapshotFlushingExternalPrimitivesIf: false.
	self assert: initialContext = finalContext.
	self assert: sim localIP = savedpc.
	sim objectMemory
		storePointer: SuspendedContextIndex
		ofObject: sim activeProcess
		withValue: finalContext! !

SimulatorHarness subclass: #SimulatorHarnessForTests
	instanceVariableNames: 'simulator'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!SimulatorHarnessForTests methodsFor: 'utilities' stamp: 'eem 6/29/2017 16:16'!
findClassNamed: classNameString
	| className classNameIndex om |
	(className := self findSymbol: classNameString) ifNil:
		[^nil].
	classNameIndex := simulator classNameIndex.
	(om := simulator objectMemory) allObjectsDo:
		[:obj|
		((om numSlotsOf: obj) > classNameIndex
		 and: [(simulator objCouldBeClassObj: obj)
		 and: [(om fetchPointer: simulator classNameIndex ofObject: obj) = className]]) ifTrue:
			[^obj]].
	^nil! !

!SimulatorHarnessForTests methodsFor: 'utilities' stamp: 'eem 09/19/2018 17:19'!
findSymbol: aString
	"Find the Symbol equal to aString in oldHeap."
	| om size symbolClassTag |
	symbolClassTag := (om := simulator objectMemory) rawClassTagForClass: self symbolClass.
	size := aString size.
	om allObjectsDo:
		[:obj|
		(symbolClassTag = (om fetchClassTagOfNonImm: obj)
		 and: [(om numBytesOf: obj) = size
		 and: ["(om fetchByte: 0 ofObject: obj) asCharacter == $C ifTrue:
				[simulator printOopShort: obj; halt]."
			(om strncmp: aString _: obj + om baseHeaderSize _: size) = 0]]) ifTrue:
			[^obj]].
	^nil! !

!SimulatorHarnessForTests methodsFor: 'execution' stamp: 'eem 6/29/2017 12:15'!
interpreter: sim object: receiver perform: selector withArguments: arguments
	simulator := sim.
	^super interpreter: sim object: receiver perform: selector withArguments: arguments! !

!SimulatorHarnessForTests methodsFor: 'utilities' stamp: 'eem 6/29/2017 12:24'!
symbolClass
	^simulator objectMemory fetchClassOfNonImm: (simulator objectMemory splObj: SelectorDoesNotUnderstand)! !

!SimulatorHarnessForTests methodsFor: 'execution' stamp: 'eem 6/29/2017 12:15'!
withExecutableInterpreter: sim do: aBlock
	simulator := sim.
	^super withExecutableInterpreter: sim do: aBlock! !

Object subclass: #SmartSyntaxPluginPrologCodeGenerator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SmartSyntaxPlugins'!

!SmartSyntaxPluginPrologCodeGenerator commentStamp: 'eem 12/17/2018 10:08' prior: 0!
SmartSyntaxPluginPrologCodeGenerator is an abstract superclass for two subclasses that generate the validations and assignments at the beginning of SmartSyntaxInterpreterPlugin primitives in response to the primitive:parameters:... sends.!

SmartSyntaxPluginPrologCodeGenerator subclass: #SmartSyntaxPluginAssignmentCodeGenerator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SmartSyntaxPlugins'!

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing support' stamp: 'eem 12/17/2018 21:03'!
assign: variableName coerceTo: cType from: stackIndex
	^String streamContents:
		[:aStream |
		 aStream
			nextPutAll: variableName;
			nextPutAll: ' := self cCoerce: (interpreterProxy firstIndexableField:';
			nextPutAll: (self stackAccessorFor: stackIndex);
			nextPutAll: ') to: ';
			store: cType]! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/18/2018 16:32'!
ccgLoad: aBlock expr: aString asBooleanValueFrom: stackIndex
	^(aString, ' := '), (self loadAs: #booleanValueOf: from: stackIndex)! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 21:02'!
ccgLoad: aBlock expr: aString asCharPtrFrom: stackIndex andThen: valBlock
	^self assign: aString coerceTo: #'char *' from: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 20:42'!
ccgLoad: aBlock expr: aString asFloatValueFrom: anInteger
	^String streamContents:
		[:aStream |
		aStream
			nextPutAll: aString;
			nextPutAll: ' := interpreterProxy stackFloatValue: ';
			print: anInteger]! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 21:02'!
ccgLoad: aBlock expr: aString asIntPtrFrom: stackIndex andThen: valBlock
	^self assign: aString coerceTo: #'int *' from: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 11/22/2019 11:57'!
ccgLoad: aBlock expr: aString asIntegerValueFrom: anInteger
	"Assume the variable has already been assigned by SmartSyntaxPluginCodeGenerator>>ccgLoad:expr:asIntegerValueFrom:,
	 so merely convert it.  This avoids two stack accesses."
	^String streamContents:
		[:aStream |
		aStream
			nextPutAll: aString;
			nextPutAll: ' := interpreterProxy integerValueOf: ';
			nextPutAll: aString]! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 20:44'!
ccgLoad: aBlock expr: aString asKindOf: aClass from: stackIndex
	^self ccgLoad: aBlock expr: aString asRawOopFrom: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 20:45'!
ccgLoad: aBlock expr: aString asKindOfIntegerFrom: stackIndex
	^self ccgLoad: aBlock expr: aString asRawOopFrom: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 20:46'!
ccgLoad: aBlock expr: aString asMemberOf: aClass from: stackIndex
	^self ccgLoad: aBlock expr: aString asRawOopFrom: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 20:46'!
ccgLoad: aBlock expr: aString asMemberOfLargeNegativeIntegerFrom: stackIndex
	^self ccgLoad: aBlock expr: aString asRawOopFrom: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 20:46'!
ccgLoad: aBlock expr: aString asMemberOfLargePositiveIntegerFrom: stackIndex
	^self ccgLoad: aBlock expr: aString asRawOopFrom: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 20:46'!
ccgLoad: aBlock expr: aString asNonIntegerValueFrom: stackIndex
	^self ccgLoad: aBlock expr: aString asRawOopFrom: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 21:03'!
ccgLoad: aBlock expr: aString asOopPtrFrom: stackIndex andThen: valBlock
	^self assign: aString coerceTo: 'sqInt *' from: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 20:44'!
ccgLoad: aBlock expr: aString asRawOopFrom: stackIndex
	^aString, ' := ', (self stackAccessorFor: stackIndex)! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/1/2020 18:04'!
ccgLoad: aBlock expr: aString asShortPtrFrom: stackIndex andThen: valBlock
	^self assign: aString coerceTo: #'short *' from: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'nice 5/2/2020 11:10'!
ccgLoad: aBlock expr: aString asUnsignedLong64PtrFrom: stackIndex andThen: valBlock
	^self assign: aString coerceTo: 'unsigned long long *' from: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 21:03'!
ccgLoad: aBlock expr: aString asUnsignedPtrFrom: stackIndex andThen: valBlock
	^self assign: aString coerceTo: 'unsigned *' from: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'nice 5/1/2020 21:04'!
ccgLoad: aBlock expr: aString asUnsignedShortPtrFrom: stackIndex andThen: valBlock
	^self assign: aString coerceTo: 'unsigned short *' from: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/18/2018 16:30'!
ccgLoad: aBlock expr: aString asUnsignedValueFrom: stackIndex
	^String streamContents:
		[:aStream |
		aStream
			nextPutAll: aString;
			nextPutAll: ' := (interpreterProxy bytesPerOop = 4';
			crtab: 2;
			nextPutAll: 'ifTrue: [interpreterProxy positive32BitValueOf:';
			nextPutAll: (self stackAccessorFor: stackIndex);
			nextPutAll: '] ifFalse: [interpreterProxy positive64BitValueOf:';
			nextPutAll: (self stackAccessorFor: stackIndex);
			nextPutAll: '])']! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 21:03'!
ccgLoad: aBlock expr: aString asWBCharPtrFrom: stackIndex
	^self assign: aString coerceTo: #'char *' from: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 21:03'!
ccgLoad: aBlock expr: aString asWBFloatPtrFrom: stackIndex
	^self assign: aString coerceTo: #'float *' from: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 21:03'!
ccgLoad: aBlock expr: aString asWBIntPtrFrom: stackIndex
	^self assign: aString coerceTo: #'int *' from: stackIndex! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing support' stamp: 'eem 12/17/2018 20:38'!
ccgValBlock: aString
	"ignore"! !

!SmartSyntaxPluginAssignmentCodeGenerator methodsFor: 'coercing support' stamp: 'eem 12/17/2018 20:39'!
loadAs: coercionSelector from: stackIndex

	^String streamContents:
		[:aStream |
		 aStream
			nextPutAll: 'interpreterProxy ';
			nextPutAll: coercionSelector;
			nextPutAll: (self stackAccessorFor: stackIndex)]! !

!SmartSyntaxPluginPrologCodeGenerator methodsFor: 'utilities' stamp: 'eem 12/17/2018 20:59'!
stackAccessorFor: index
	self assert: index isInteger.
	^'(interpreterProxy stackValue: ', (index printString, ')')! !

SmartSyntaxPluginPrologCodeGenerator subclass: #SmartSyntaxPluginValidationCodeGenerator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SmartSyntaxPlugins'!

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:43'!
ccgLoad: aBlock expr: aString asBooleanValueFrom: stackIndex
	^self loadAs: #isBooleanObject: from: stackIndex! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:43'!
ccgLoad: aBlock expr: aString asCharPtrFrom: stackIndex andThen: validationString
	^validationString, (self stackAccessorFor: stackIndex)! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/18/2018 17:20'!
ccgLoad: aBlock expr: aString asFloatValueFrom: stackIndex
	^self loadAs: #isFloatObject: from: stackIndex! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:43'!
ccgLoad: aBlock expr: aString asIntPtrFrom: stackIndex andThen: validationString
	^validationString, (self stackAccessorFor: stackIndex)! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 11/22/2019 11:43'!
ccgLoad: aBlock expr: variableName asIntegerValueFrom: stackIndex
	"Cache the value in the variable to avoid accessing the stack more than once."
	^String streamContents:
		[:aStream |
		 aStream
			nextPutAll: 'interpreterProxy ';
			nextPutAll: #isIntegerObject:;
			nextPutAll: ' ('; nextPutAll: variableName; nextPutAll: ' := ';
			nextPutAll: (self stackAccessorFor: stackIndex);
			nextPut: $)]! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:43'!
ccgLoad: aBlock expr: aString asKindOf: aClass from: stackIndex
	^self loadAs: #is:KindOf: class: aClass from: stackIndex! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:34'!
ccgLoad: aBlock expr: aString asKindOfIntegerFrom: stackIndex
	^self loadAs: #isKindOfInteger: from: stackIndex! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:43'!
ccgLoad: aBlock expr: aString asMemberOf: aClass from: stackIndex
	^self loadAs: #is:MemberOf: class: aClass from: stackIndex! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:35'!
ccgLoad: aBlock expr: aString asMemberOfLargeNegativeIntegerFrom: stackIndex
	^self loadAs: #isLargeNegativeIntegerObject: from: stackIndex! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:35'!
ccgLoad: aBlock expr: aString asMemberOfLargePositiveIntegerFrom: stackIndex
	^self loadAs: #isLargePositiveIntegerObject: from: stackIndex! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:40'!
ccgLoad: aBlock expr: aString asNonIntegerValueFrom: stackIndex
	^self loadAs: #isNonImmediate: from: stackIndex! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/18/2018 11:37'!
ccgLoad: aBlock expr: aString asOopPtrFrom: stackIndex andThen: validationString
	^String streamContents:
		[:s|
		 s	nextPut: $(;
			nextPutAll: (self loadAs: #isPointers: from: stackIndex);
			nextPutAll: ') and: [';
			nextPutAll: validationString;
			nextPutAll: (self stackAccessorFor: stackIndex);
			nextPut: $]]! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:42'!
ccgLoad: aBlock expr: aString asRawOopFrom: anInteger
	^nil! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/1/2020 18:08'!
ccgLoad: aBlock expr: aString asShortPtrFrom: stackIndex andThen: validationString
	^validationString, (self stackAccessorFor: stackIndex)! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'nice 5/2/2020 11:11'!
ccgLoad: aBlock expr: aString asUnsignedLong64PtrFrom: stackIndex andThen: validationString
	^validationString, (self stackAccessorFor: stackIndex)! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:44'!
ccgLoad: aBlock expr: aString asUnsignedPtrFrom: stackIndex andThen: validationString
	^validationString, (self stackAccessorFor: stackIndex)! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'nice 5/1/2020 21:04'!
ccgLoad: aBlock expr: aString asUnsignedShortPtrFrom: stackIndex andThen: validationString
	^validationString, (self stackAccessorFor: stackIndex)! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:54'!
ccgLoad: aBlock expr: aString asUnsignedValueFrom: 	stackIndex
	^self loadAs: #isPositiveMachineIntegerObject: from: stackIndex! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:47'!
ccgLoad: aBlock expr: aString asWBCharPtrFrom: stackIndex
	^self loadAs: #isWordsOrBytes: from: stackIndex! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:48'!
ccgLoad: aBlock expr: aString asWBFloatPtrFrom: stackIndex
	^self loadAs: #isWordsOrBytes: from: stackIndex! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing' stamp: 'eem 12/17/2018 19:48'!
ccgLoad: aBlock expr: aString asWBIntPtrFrom: stackIndex
	^self loadAs: #isWordsOrBytes: from: stackIndex! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing support' stamp: 'eem 12/17/2018 11:48'!
ccgValBlock: aString
	^'interpreterProxy ', (aString, ': ')! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing support' stamp: 'eem 12/18/2018 11:33'!
loadAs: classMembershipSelector class: aClass from: stackIndex

	^String streamContents:
		[:aStream | | keywords |
		 keywords := classMembershipSelector keywords.
		 aStream
			nextPutAll: 'interpreterProxy ';
			nextPutAll: keywords first;
			nextPutAll: (self stackAccessorFor: stackIndex);
			space;
			nextPutAll: keywords last;
			nextPutAll:	' ''';
			nextPutAll:	aClass asString;
			nextPutAll: '''']! !

!SmartSyntaxPluginValidationCodeGenerator methodsFor: 'coercing support' stamp: 'eem 12/17/2018 20:39'!
loadAs: classMembershipSelector from: stackIndex

	^String streamContents:
		[:aStream |
		 aStream
			nextPutAll: 'interpreterProxy ';
			nextPutAll: classMembershipSelector;
			nextPutAll: (self stackAccessorFor: stackIndex)]! !

Object subclass: #TMethod
	instanceVariableNames: 'args comment complete declarations definingClass export extraVariableNumber globalStructureBuildMethodHasFoo inline labels locals parseTree primitive properties returnType selector sharedCase sharedLabel static writtenToGlobalVarsCache functionAttributes usedVariablesCache'
	classVariableNames: 'CaseStatements'
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TMethod commentStamp: 'dtl 9/15/2008 09:06' prior: 0!
A TMethod is a translation method, representing a MethodNode that is to be translated to C source. It has a parseTree of translation nodes that mirrors the parse tree of the corresponding Smalltalk method.!

TMethod subclass: #SmartSyntaxPluginTMethod
	instanceVariableNames: 'isPrimitive fullSelector fullArgs parmSpecs rcvrSpec'
	classVariableNames: 'Them'
	poolDictionaries: ''
	category: 'VMMaker-SmartSyntaxPlugins'!

!SmartSyntaxPluginTMethod commentStamp: 'eem 6/6/2018 14:06' prior: 0!
Variation of TMethod node of the Smalltalk C Code Generator, used in conjunction with SmartSyntaxPluginCodeGenerator and SmartSyntaxInterpreterPlugin to generate named primitives from methods containing type coercion specifications such as
	primitive: functionName parameters: #(Boolean Oop String WordsArray WordsOrBytes)
	primitive: functionName parameters: #(SmallInteger LargeNegativeInteger LargePositiveInteger Integer Unsigned) receiver: #Oop!

!SmartSyntaxPluginTMethod methodsFor: 'accessing' stamp: 'acg 9/19/1999 11:19'!
args: anInteger

	^args := anInteger! !

!SmartSyntaxPluginTMethod methodsFor: 'private' stamp: 'acg 12/15/1999 06:31'!
assign: variable expression: expression

	^TAssignmentNode new
		setVariable: variable
		expression: expression! !

!SmartSyntaxPluginTMethod methodsFor: 'C code generation' stamp: 'eem 11/2/2010 14:20'!
emitCLocalsOn: aStream generator: aCodeGen
	"Emit a C function header for this method onto the given stream."

	aStream cr.
	locals isEmpty ifFalse:
		[(aCodeGen sortStrings: locals) do:
			[ :var |
		aStream 
			tab; 
			nextPutAll: (self declarationAt: var);
			nextPut: $;; 
			cr].
		 aStream cr]! !

!SmartSyntaxPluginTMethod methodsFor: 'testing' stamp: 'eem 1/19/2021 21:15'!
endsWithMethodReturnExpression
	| operativeReturn methodReturns |
	operativeReturn := (parseTree statements last isReturn
						 and: [parseTree statements last expression isLeaf])
							ifTrue: [(parseTree statements last: 2) first]
							ifFalse: [parseTree statements last].
	(operativeReturn isSend
	and: [operativeReturn selector == #ifFalse:
	and: [operativeReturn receiver isSend
	and: [operativeReturn receiver selector == #failed
	and: [operativeReturn receiver receiver isLeaf
	and: [operativeReturn receiver receiver name = 'interpreterProxy'
	and: [operativeReturn args size = 1
	and: [operativeReturn args first statements first isReturn]]]]]]]) ifTrue:
		[operativeReturn := operativeReturn args first statements first expression].
	methodReturns := #(	methodReturnReceiver
							methodReturnFloat:
							methodReturnValue:
							methodReturnInteger:
							methodReturnBool:
							methodReturnString:
							methodReturnStringOrNil:

							asBooleanObj
							asFloatObj
							asPositiveIntegerObj
							asSmallIntegerObj).
	^operativeReturn isSend
	 and: [(methodReturns includes: operativeReturn selector)
		or: [operativeReturn isConditionalSend
			and: [(operativeReturn args collect: [:stmts| stmts statements last]) anySatisfy:
					[:stmt| stmt isSend and: [methodReturns includes: stmt selector]]]]]! !

!SmartSyntaxPluginTMethod methodsFor: 'specifying primitives' stamp: 'eem 1/19/2021 20:59'!
extractPrimitiveDirectives
	"Save selector in fullSelector and args in fullArgs.  Scan top-level statements for a directive of the form:

		self	
			primitive: 	<string>
or
		self
			primitive:	<string>
			parameters: <list of class names>
or
		self
			primitive:	<string>
			parameters: <list of class names>
			receiver: <class name>

or an assignment of that expression to a local, and manipulate the state and parse tree accordingly."

	parseTree setStatements: (Array streamContents:
		[:sStream |
			parseTree statements do:
				[:stmt |
				 (self primitiveDirectiveWasHandled: stmt on: sStream)
					ifFalse: [sStream nextPut: stmt]]]).
	isPrimitive 
		ifTrue:
			[export := true.
			 parseTree statements addAllFirst: self namedPrimitiveProlog.
			 self fixUpReturns.
			 self replaceSizeMessages.
			 ^true]
		ifFalse: [self removeFinalSelfReturnIn: nil].
	^false! !

!SmartSyntaxPluginTMethod methodsFor: 'transforming' stamp: 'eem 11/22/2019 13:18'!
fixUpReturnOneStmt: stmt on: sStream
	| expr exprRetStmts "p t" |
	stmt isReturn ifFalse: [^sStream nextPut: stmt].
	expr := stmt expression.
	(expr isSend
	 and: [self resultSendAlwaysFails: expr]) ifTrue:
		["failure returns"
		 returnType = #sqInt
			ifTrue: [sStream nextPut: (TReturnNode new setExpression: expr)]
			ifFalse: [sStream nextPut: expr; nextPut: self nullReturnExpr].
		 ^nil].
	(expr isVariable and: ['nil' = expr name]) ifTrue:
		["^ nil -- this is never right unless automatically generated"
		 sStream nextPut: stmt.
		 ^nil].
	(expr isVariable and: ['self' = expr name]) ifTrue:
		["^ self"
		 fullArgs isEmpty ifFalse:
			[sStream nextPut: (self statementGuardedWithSuccess: (self popExpr: fullArgs size))].
		 sStream nextPut: self nullReturnExpr.
		 ^nil].
	(expr isVariable or: [expr isConstant]) ifTrue:
		["^ variable or ^ constant without guardchecking"
		sStream
			nextPut: (sStream isEmpty "No statements to cause failure, therefore no need for failure guard."
						ifTrue:
							[self pop: fullArgs size + 1 thenReturnExpr: expr]
						ifFalse:
							[self statementGuardedWithSuccess:
								(self pop: fullArgs size + 1 thenReturnExpr: expr)]);
			nextPut: self nullReturnExpr.
		 ^nil].
	"^ expr with necessary guard checking"
	"p := sStream position."
	exprRetStmts := Array streamContents:
		[:ersStream|
		 (self resultExpressionCanFail: expr)
			ifTrue:
				["t := 1."
				ersStream
					nextPut: (self assign: (self oopVariable: '_return_value') expression: expr);
					nextPut: (self statementGuardedWithSuccess: (self pop: fullArgs size + 1
																		thenReturnExpr: (self oopVariable: '_return_value')))]
			ifFalse:
				["t := 2."
				 ersStream nextPut: (self pop: fullArgs size + 1 thenReturnExpr: expr)]].
		sStream isEmpty "No statements to cause failure, therefore no need for an initial failure guard."
			ifTrue: [sStream nextPutAll: exprRetStmts]
			ifFalse:
				["t := t + 2."
				 sStream nextPut: (self statementGuardedWithSuccess: exprRetStmts)].
	sStream nextPut: self nullReturnExpr.
	"Them := Dictionary new"
	"(Them at: t ifAbsentPut: [Dictionary new])
		at: self selector
		put: (sStream originalContents copyFrom: p + 1 to: sStream position)"! !

!SmartSyntaxPluginTMethod methodsFor: 'transforming' stamp: 'eem 8/24/2020 19:39'!
fixUpReturns
	"Replace each return statement in this method with (a) the given postlog, (b) code to pop the receiver and the given number of arguments, and (c) code to push the integer result and return."
	self endsWithMethodReturnExpression ifTrue:
		[parseTree statements last isSend ifFalse:
			[parseTree setStatements: parseTree statements allButLast]].
	parseTree nodesDo:
		[:node |
		node isStmtList ifTrue:
			[node setStatements: (Array streamContents:
				[:sStream |
				 node statements do: 
					[:stmt | self fixUpReturnOneStmt: stmt on: sStream]])]]! !

!SmartSyntaxPluginTMethod methodsFor: 'initialization' stamp: 'acg 9/19/1999 11:46'!
fromContext: aContext primitive: aString parameters: aClassList receiver: aClass

	fullArgs := args := aContext tempNames
				copyFrom: 1
				to: aContext method numArgs.
	self 
		primitive: aString
		parameters: aClassList
		receiver: aClass! !

!SmartSyntaxPluginTMethod methodsFor: 'specifying primitives' stamp: 'eem 7/27/2021 10:30'!
handlePrimitiveDirective: aStmt on: sStream
	| needsCheckSuccessExpr loadExpressionSource |
	isPrimitive := true.
	fullArgs := args.
	locals addAll: args.
	args := OrderedCollection new.
	fullArgs with: parmSpecs do:
		[:argName :spec |
		self declarationAt: argName
			put: (spec ccgDeclareCForVar: argName)].
	needsCheckSuccessExpr := false.
	aStmt isAssignment ifTrue:
		[self declarationAt: aStmt variable name
			put: (rcvrSpec ccgDeclareCForVar: aStmt variable name).
		 loadExpressionSource := rcvrSpec
									ccg:	SmartSyntaxPluginCodeGenerator new
									prolog:  [:expr | aStmt variable name, ' := ', expr]
									expr: 	aStmt variable name
									index: 	fullArgs size.
		 needsCheckSuccessExpr := (loadExpressionSource includesSubstring: ' success:')
									or: [InterpreterProxy checkingStackLoads anySatisfy:
											[:checkingSelector|
											loadExpressionSource includesSubstring: checkingSelector]].
		 sStream nextPutAll: (self statementsFor: loadExpressionSource varName: '')].

	"only add the failure guard if required"
	needsCheckSuccessExpr ifTrue:
		[sStream nextPutAll: self checkSuccessExpr].
	^true! !

!SmartSyntaxPluginTMethod methodsFor: 'specifying primitives' stamp: 'TPR 2/10/2000 17:43'!
isPrimitiveDirectiveSend: stmt
	
	stmt isSend ifTrue:
		[stmt selector = #primitive: ifTrue:
			[^self primitive: 	stmt args first value
				   parameters:	(Array new: args size withAll: #Oop)
				   receiver:		#Oop].
		 stmt selector = #primitive:parameters: ifTrue:
			[^self primitive: 	stmt args first value
				   parameters: 	stmt args second value
				   receiver:		#Oop].
		 stmt selector = #primitive:parameters:receiver: ifTrue:
			[^self primitive:		stmt args first value
				   parameters:	stmt args second value
				   receiver:		stmt args third value].
		^false].
	^false.
! !

!SmartSyntaxPluginTMethod methodsFor: 'specifying primitives' stamp: 'eem 7/27/2021 09:55'!
namedPrimitiveProlog
	"Generate the code for a primitive:parameters:... send.  This is in two parts.
	 The first is validation; the second is coercing assignment."
	| statements validator validations assigner |
	fullArgs isEmpty ifTrue:
		[^#()].
	validator := SmartSyntaxPluginValidationCodeGenerator new.
	statements := OrderedCollection new.
	validations := fullArgs withIndexCollect:
					[:arg :i|
					(parmSpecs at: i) 
							ccg: 	validator
							prolog:  nil
							expr: arg
							index: (fullArgs size - i)].
	validations := validations reject: [:validation| validation isNil].
	validations isEmpty ifFalse:
		[statements addAllLast: (self statementsFor:
									(String streamContents:
										[:s|
										s nextPut: $(.
										validations
											do: [:validation| s nextPut: $(; nextPutAll: validation; nextPut: $)]
											separatedBy: [s crtab; nextPutAll: 'and: ['].
										s next: validations size - 1 put: $].
										s nextPutAll: ') ifFalse: [^interpreterProxy primitiveFailFor: PrimErrBadArgument]'])
									varName: '')].
	assigner := SmartSyntaxPluginAssignmentCodeGenerator new.
	fullArgs withIndexDo:
		[:arg :i|
		statements addAllLast:
			(self 
				statementsFor: 
					((parmSpecs at: i) 
						ccg: 	assigner
						prolog:  nil
						expr:	arg
						index:	(fullArgs size - i))
				varName: '')].
	^statements! !

!SmartSyntaxPluginTMethod methodsFor: 'private' stamp: 'acg 12/15/1999 05:50'!
nullReturnExpr

	^ TReturnNode new
		setExpression: (TVariableNode new setName: 'null')! !

!SmartSyntaxPluginTMethod methodsFor: 'accessing' stamp: 'acg 9/19/1999 11:18'!
parmSpecs

	^parmSpecs! !

!SmartSyntaxPluginTMethod methodsFor: 'private' stamp: 'eem 9/25/2020 12:32'!
pop: anInteger thenReturnExpr: anExpression
	"Use the methodReturnFoo: protocol; it's simpler, more concise, more reliable (var args)"
	anExpression isSend ifTrue:
		[#(asSmallIntegerObj asBooleanObj asFloatObj)
			with: #(methodReturnInteger: methodReturnBool: methodReturnFloat:)
			do:
				[:abstractMessage :returnMessage|
				anExpression selector == abstractMessage ifTrue:
					[^TSendNode new
							setSelector: returnMessage
							receiver: (TVariableNode new setName: #interpreterProxy)
							arguments: {anExpression receiver}]]].
	^TSendNode new
		setSelector: #methodReturnValue:
		receiver: (TVariableNode new setName: 'interpreterProxy')
		arguments: {anExpression}! !

!SmartSyntaxPluginTMethod methodsFor: 'private' stamp: 'acg 12/15/1999 05:45'!
popExpr: anInteger

	^ TSendNode new
			 setSelector: #pop:
			 receiver: (TVariableNode new setName: 'interpreterProxy')
			 arguments: (Array 
				with: (TConstantNode new 
					setValue: anInteger))! !

!SmartSyntaxPluginTMethod methodsFor: 'specifying primitives' stamp: 'acg 9/20/1999 13:00'!
primitive: aString parameters: anArray receiver: aClassSymbol

	fullSelector := selector.
	selector := aString asSymbol.
	anArray size == args size ifFalse: 
		[^self error: selector, ': incorrect number of parameter specifications'].
	parmSpecs := anArray collect:
		[:each | Smalltalk at: each ifAbsent:
			[^self error: selector, ': parameter spec must be a Behavior']].
	parmSpecs do: [:each | each isBehavior ifFalse:
		[^self error: selector, ': parameter spec must be a Behavior']].
	rcvrSpec := Smalltalk at: aClassSymbol asSymbol ifAbsent:
		[^self error: selector, ': receiver spec must be a Behavior'].
	rcvrSpec isBehavior ifFalse:
		[^self error: selector, ': receiver spec must be a Behavior'].
	^true! !

!SmartSyntaxPluginTMethod methodsFor: 'specifying primitives' stamp: 'acg 9/17/1999 22:41'!
primitiveDirectiveWasHandled: stmt on: sStream

	(self isPrimitiveDirectiveSend: stmt) ifTrue:
		[^self handlePrimitiveDirective: stmt on: sStream].
	(stmt isAssignment and: 
		[self isPrimitiveDirectiveSend: stmt expression]) ifTrue:
			[^self handlePrimitiveDirective: stmt on: sStream].
	^false.
! !

!SmartSyntaxPluginTMethod methodsFor: 'private' stamp: 'eem 12/18/2018 16:08'!
printTempsAndVar: varName on: aStream 
	"add the required temps and the varname to the stream"
	aStream nextPut: $|; space.
	(#('rcvr' 'stackPointer' 'interpreterProxy') reject: [:each | locals includes: each]) do:
		[:each | aStream nextPutAll: each;  space].
	(locals reject: [:each | each first = $_]) do:
		[:each | aStream nextPutAll: each;  space].
"don't add varName twice. Probably a deeper reason for this, but WTH. TPR"
	(locals includes: varName) ifFalse:
		[aStream nextPutAll: varName; space].
	aStream nextPut: $|; cr! !

!SmartSyntaxPluginTMethod methodsFor: 'accessing' stamp: 'acg 9/19/1999 11:50'!
rcvrSpec

	^rcvrSpec! !

!SmartSyntaxPluginTMethod methodsFor: 'transforming' stamp: 'eem 10/29/2018 19:03'!
recordDeclarationsIn: aCCodeGen
	"Record C type declarations of the forms
		<returnTypeC: 'float'>
		<var: #foo declareC: 'float foo'>
		<var: #foo type:'float'>
		<var: #foo as: Class>
	 or the older, obsolete
		self returnTypeC: 'float'.
		self var: #foo declareC: 'float foo'
		self var: #foo type:'float'.
		self var: #foo as: Class
	 and remove the declarations from the method body."

	| newStatements |
	properties pragmas notEmpty ifTrue:
		[properties pragmas do:
			[:pragma|
			pragma keyword == #var:declareC: ifTrue:
				[self checkedDeclarationAt: pragma arguments first asString
					put: pragma arguments last
					in: aCCodeGen].
			pragma keyword == #var:type: ifTrue:
				[| varName varType |
				varName := pragma arguments first asString.
				varType := aCCodeGen conventionalTypeForType: pragma arguments last.
				varType last == $* ifFalse: [varType := varType, ' '].
				self checkedDeclarationAt: varName
					put: varType, varName
					in: aCCodeGen].
			 pragma keyword == #var:as: ifTrue:
				[| theClass |
				 theClass := Smalltalk at: pragma arguments last name asSymbol ifAbsent: [nil].
				 theClass isBehavior ifFalse:
					[^self error: 'declarator must be a Behavior'].
				 self checkedDeclarationAt: pragma arguments first value asString 
					put: (theClass ccgDeclareCForVar: pragma arguments first asString)
					in: aCCodeGen].
			pragma keyword == #returnTypeC: ifTrue:
				[self returnType: pragma arguments last].
			pragma keyword = #doNotGenerate: ifTrue:
				[locals removeKey: pragma arguments last]].
		^self].
	newStatements := OrderedCollection new: parseTree statements size.
	parseTree statements do: 
		[:stmt | | isDeclaration |
		 isDeclaration := false.
		 stmt isSend ifTrue: 
			[stmt selector == #var:declareC: ifTrue:
				[isDeclaration := true.
				self declarationAt: stmt args first value asString put: stmt args last value].
			stmt selector = #var:type: ifTrue:
				[| varName varType |
				isDeclaration := true.
				varName := stmt args first value asString.
				varType := aCCodeGen conventionalTypeForType: stmt args last value.
				varType last == $* ifFalse: [varType := varType, ' '].
				self declarationAt: varName put: varType, varName].
			 stmt selector == #var:as: ifTrue:
				[| theClass |
				 isDeclaration := true.
				 theClass := Smalltalk  at: stmt args last name asSymbol ifAbsent: [nil].
				 theClass isBehavior ifFalse:
					[^self error: 'declarator must be a Behavior'].
				 self declarationAt: stmt args first value asString 
					put: (theClass ccgDeclareCForVar: stmt args first value asString)].
			 stmt selector == #returnTypeC: ifTrue: 
				[isDeclaration := true.
				 returnType := stmt args last value]].
		 isDeclaration ifFalse: [newStatements add: stmt]].
	parseTree setStatements: newStatements asArray! !

!SmartSyntaxPluginTMethod methodsFor: 'transforming' stamp: 'acg 9/20/1999 14:04'!
replaceSizeMessages
	"Replace sends of the message 'size' with calls to sizeOfSTArrayFromCPrimitive."

	parseTree nodesDo: [:n |
		(n isSend and: [n selector = #size]) ifTrue: [
			n
				setSelector: #sizeOfSTArrayFromCPrimitive:
				receiver: (TVariableNode new setName: 'interpreterProxy')
				arguments: (Array with: n receiver)]].
! !

!SmartSyntaxPluginTMethod methodsFor: 'private' stamp: 'eem 1/19/2021 21:31'!
resultExpressionCanFail: aTSendNode
	"Neither asSmallIntegerObj nor asBooleanObj nor asPositiveIntegerObj can fail.
	 asPositiveIntegerObj maps to positive32BitIntegerFor: which uses
	 eeInstantiateSmallClassIndex:format:numSlots: which does not GC and does not fail."
	^(#(asSmallIntegerObj asPositiveIntegerObj asBooleanObj nilObject trueObject falseObject) includes: aTSendNode selector) not! !

!SmartSyntaxPluginTMethod methodsFor: 'private' stamp: 'eem 11/22/2019 13:11'!
resultSendAlwaysFails: aTSendNode
	^(#(primitiveFail primitiveFailFor: primitiveFailForFFIException:at: primitiveFailForOSError:) includes: aTSendNode selector)
		or: [aTSendNode selector == #success:
			 and: [aTSendNode args first isConstant
			 and: [aTSendNode args first value == false]]]! !

!SmartSyntaxPluginTMethod methodsFor: 'initialization' stamp: 'eem 9/18/2020 18:22'!
setSelector: sel definingClass: class args: argList locals: localList block: aBlockNode primitive: aNumber properties: methodProperties comment: aComment
	"Initialize this method using the given information."

	selector := sel.
	definingClass := class.
	returnType := #sqInt. 	 "assume return type is sqInt for now"
	args := argList asOrderedCollection collect: [:arg | arg key].
	locals := (localList collect: [:arg | arg key]) asSet.
	declarations := Dictionary new.
	primitive := aNumber.
	properties := methodProperties.
	comment := aComment.
	parseTree := aBlockNode asRootTranslatorNodeIn: self.
	labels := Set new.
	complete := false.  "set to true when all possible inlining has been done"
	export := self extractExportDirective.
	static := self extractStaticDirective.
	self extractSharedCase.
	isPrimitive := false.  "set to true only if you find a primtive direction."
	self recordDeclarationsIn: CCodeGenerator basicNew getLogger. "Just for conventionalTypeForType:"
	self extractPrimitiveDirectives.
! !

!SmartSyntaxPluginTMethod methodsFor: 'accessing' stamp: 'eem 9/15/2015 11:48'!
smalltalkSelector
	"Answer the selector of the original Smalltalk method, not any mangled one."
	^fullSelector ifNil: [selector]! !

!SmartSyntaxPluginTMethod methodsFor: 'private' stamp: 'eem 12/18/2018 15:33'!
statementGuardedWithSuccess: aTParseNodeOrSequenceThereof
	"Answer a TSendNode for interpreterProxy failed ifFalse: [aTParseNodeOrSequenceThereof]"
	^TSendNode new
		setSelector: #ifFalse:
		receiver: (TSendNode new
					setSelector: #failed
					receiver: (TVariableNode new setName: 'interpreterProxy')
					arguments: #())
		arguments: {(aTParseNodeOrSequenceThereof isTParseNode and: [aTParseNodeOrSequenceThereof isStmtList])
						ifTrue: [aTParseNodeOrSequenceThereof]
						ifFalse: [TStmtListNode new
									setArguments: #()
									statements: (aTParseNodeOrSequenceThereof isCollection
													ifTrue: [aTParseNodeOrSequenceThereof]
													ifFalse: [{aTParseNodeOrSequenceThereof}])]}! !

!SmartSyntaxPluginTMethod methodsFor: 'primitive compilation' stamp: 'TPR 3/2/2000 16:07'!
vmNameString
	"return the string to use as the vm name in code generated for this method"
	^'interpreterProxy'! !

!TMethod class methodsFor: 'class initialization' stamp: 'ar 11/18/1999 20:06'!
initialize
	"TMethod initialize"	
	CaseStatements := IdentitySet new: 10.
	CaseStatements addAll: #(dispatchOn:in: dispatchOn:in:with: dispatchOn:in:with:with:).! !

!TMethod methodsFor: 'accessing' stamp: 'eem 9/5/2021 19:18'!
addFunctionAttribute: aString
	functionAttributes
		ifNil: [functionAttributes := aString]
		ifNotNil:
			[(functionAttributes includesSubstring: aString) ifFalse:
				[functionAttributes := functionAttributes, ' ', aString]]! !

!TMethod methodsFor: 'accessing' stamp: 'eem 7/26/2014 07:34'!
addLabelsTo: aTMethod
	aTMethod labels addAll: labels! !

!TMethod methodsFor: 'initialization' stamp: 'eem 3/3/2009 08:46'!
addTypeForSelf
	"If self should be typed then add a suitable type declaration.
	 Preserve the flagging of an implicit self using the #implicit symbol as the fake type."
	self typeForSelf ifNotNil:
		[:typeForSelf|
		self declarationAt: 'self'
			put: (typeForSelf == #implicit
					ifTrue: [typeForSelf]
					ifFalse: [typeForSelf, ' self'])]! !

!TMethod methodsFor: 'type inference' stamp: 'eem 11/14/2020 20:44'!
addTypesFor: node to: typeSet in: aCodeGen
	"Add the value types for the node to typeSet.
	 Answer if any type was derived from an as-yet-untyped method or variable, which allows us to abort
	 inferReturnTypeFromReturnsIn: if the return type depends on a yet-to-be-typed method or variable."
	| expr |
	expr := node.
	[expr isAssignment or: [expr isStmtList]] whileTrue:
		[expr isAssignment ifTrue:
			[expr := expr variable].
		 expr isStmtList ifTrue:
			[expr := expr statements last]].
	expr isSend ifTrue:
		[(#(ifTrue: ifFalse: ifTrue:ifFalse: ifFalse:ifTrue:) includes: expr selector) ifTrue:
			[^expr args
				inject: false
				into: [:asYetUntyped :block|
					asYetUntyped | (self addTypesFor: block to: typeSet in: aCodeGen)]].
		(aCodeGen returnTypeForSend: expr in: self ifNil: nil)
			ifNil: [^(aCodeGen methodNamed: expr selector) notNil and: [expr selector ~~ selector]]
			ifNotNil:
				[:type |
				typeSet add: type.
				^false]].
	expr isVariable ifTrue:
		[(aCodeGen typeOfVariable: expr name)
			ifNotNil: [:type| typeSet add: type]
			ifNil: [expr name ~= 'nil' ifTrue: "nil could be a pointer or integer value, so it is effectively untyped."
					[(typeSet add: (expr name = 'self' "self definitely means no type, at least in non-struct classes..."
										ifTrue: [#void]
										ifFalse: [#sqInt])) == #sqInt ifTrue:
						[^true]]]].
	expr isConstant ifTrue:
		[(expr value isInteger and: [expr value >= 0]) "cannot determine if signed or unsigned yet..."
			ifTrue: [typeSet add: expr value]
			ifFalse:
				[(expr typeOrNilFrom: aCodeGen in: self) ifNotNil:
					[:type | typeSet add: type]]].
	^false! !

!TMethod methodsFor: 'inlining support' stamp: 'eem 3/30/2013 11:56'!
addVarsDeclarationsAndLabelsOf: methodToBeInlined except: doNotRename
	"Prepare to inline the body of the given method into the receiver by making the args and locals of the argument to the receiver be locals of the receiver. Record any type declarations for these variables. Record labels. Assumes that the variables have already be renamed to avoid name clashes."

	locals
		addAll: (methodToBeInlined args reject: [ :v | doNotRename includes: v]);
		addAll: (methodToBeInlined locals reject: [ :v | doNotRename includes: v]).
	methodToBeInlined declarations keysAndValuesDo:
		[ :v :decl |
		(doNotRename includes: v) ifFalse:
			[self declarationAt: v put: decl]].

	labels addAll: methodToBeInlined labels! !

!TMethod methodsFor: 'utilities' stamp: ''!
allCalls
	"Answer a collection of selectors for the messages sent by this method."

	^parseTree allCalls! !

!TMethod methodsFor: 'accessing' stamp: 'eem 12/11/2010 10:41'!
allLocals
	"The merge of locals and declarations (don't ask)"

	^(declarations keys asSet reject: [:k| (declarations at: k) == #implicit])
		addAll: locals; yourself! !

!TMethod methodsFor: 'accessing' stamp: 'eem 11/14/2020 18:55'!
allReferencedArgumentsUsing: aCodeGen
	"Answer the set of all variables referenced in the receiver."
	| refs |
	refs := Set new.
	"Find all the variable names referenced in this method.
	 Don't descend into conditionals that won't be generated."
	parseTree
		nodesWithParentsDo:
			[:node :parent|
			node isVariable ifTrue: [refs add: node name asString].
			(node isSend
			 and: [node selector beginsWith: #cCode:]) ifTrue:
				[aCodeGen addVariablesInVerbatimCIn: node to: refs]]
		unless:
			[:node :parent|
			parent notNil
			and: [parent isSend
			and: [aCodeGen nodeIsDeadCode: node withParent: parent]]].
	^refs intersection: args! !

!TMethod methodsFor: 'accessing' stamp: 'eem 2/25/2016 16:56'!
allReferencedVariablesUsing: aCodeGen
	"Answer the set of all variables referenced in the receiver."
	| refs |
	refs := Set new.
	"Find all the variable names referenced in this method.
	 Don't descend into conditionals that won't be generated."
	parseTree
		nodesWithParentsDo:
			[:node :parent|
			node isVariable ifTrue: [refs add: node name asString].
			node isStmtList ifTrue: [refs addAll: node args].
			(node isSend
			 and: [node selector beginsWith: #cCode:]) ifTrue:
				[aCodeGen addVariablesInVerbatimCIn: node to: refs]]
		unless:
			[:node :parent|
			parent notNil
			and: [parent isSend
			and: [aCodeGen nodeIsDeadCode: node withParent: parent]]].
	^refs! !

!TMethod methodsFor: 'inlining' stamp: 'eem 11/19/2020 14:10'!
argAssignmentsFor: meth send: aSendNode except: elidedArgs in: aCodeGen
	"Answer a collection of assignment nodes that assign the given argument expressions
	 to the formal parameter variables of the given method.

	 Optimization: If the actual parameters are either constants or local variables in the target
	 method (the receiver), substitute them directly into the body of meth. Note that global
	 variables cannot be substituted because the inlined method might depend on the exact
	 ordering of side effects to the globals.

	 Optimization: Don't generate assignment statements for formal parameters which are unused in
	 the method body, retaining statements for unused parameter expressions that have side-effects."

	| stmtList substitutionDict argList referencedArguments |
	meth args size > (argList := aSendNode args) size ifTrue:
		[self assert: (meth args first beginsWith: 'self_in_').
		 argList := {aSendNode receiver}, aSendNode args].
	stmtList := OrderedCollection new: argList size.
	substitutionDict := Dictionary new: argList size.
	meth args with: argList do:
		[:argName :exprNode |
		(self isNode: exprNode substitutableFor: argName inMethod: meth in: aCodeGen)
			ifTrue:
				[substitutionDict
					at: argName
					put: (aCodeGen
							node: exprNode
							typeCompatibleWith: argName
							inliningInto: meth
							in: self).
				 locals remove: argName ifAbsent: [self assert: (argName beginsWith: 'self_in_')].
				 declarations removeKey: argName ifAbsent: nil]
			ifFalse: "Add an assignment for anything except an unused self_in_foo argument"
				[(elidedArgs includes: argName) ifFalse:
					[self deny: exprNode isLiteralBlock.
					 stmtList addLast:
						(TAssignmentNode new
							setVariable: (TVariableNode new setName: argName)
							expression: (aCodeGen
											node: exprNode copy
											typeCompatibleWith: argName
											inliningInto: meth
											in: self))]]].
	meth parseTree: (meth parseTree bindVariablesIn: substitutionDict).
	referencedArguments := meth allReferencedArgumentsUsing: aCodeGen.
	"Now filter-out assignments for unused formal parameters, while
	 retaining any actual parameter expressions that have side-effects."
	^(stmtList
		collect:
			[:assignment|
			(referencedArguments includes: assignment variable name)
				ifTrue: [assignment]
				ifFalse:
					[assignment expression hasSideEffect ifTrue:
						[assignment expression]]])
		select: [:node| node notNil]! !

!TMethod methodsFor: 'primitive compilation' stamp: 'eem 04/27/2017 12:55'!
argConversionExprFor: varName stackIndex: stackIndex 
	"Return the parse tree for an expression that fetches and converts the 
	primitive argument at the given stack offset."
	| exprList decl type stmtList |
	exprList := OrderedCollection new.
	((decl := declarations at: varName ifAbsent: []) notNil
	 and: ['int' ~= (type := (decl copyReplaceAll: varName with: '') withBlanksTrimmed)])
		ifTrue:
			[(decl includes: $*) ifTrue: "array"
				[(decl includesSubstring: 'char')
					ifTrue:
						[| expr |
						expr := '(interpreterProxy isBytes: (interpreterProxy stackValue: (stackIndex))) ifFalse: [^interpreterProxy primitiveFail]'.
						expr := expr copyReplaceAll: 'interpreterProxy' with: self vmNameString.
						expr := expr copyReplaceAll: 'stackIndex' with: stackIndex printString.
						exprList addLast: expr].
					exprList addLast: varName , ' := ', self vmNameString, ' arrayValueOf: (', self vmNameString, ' stackValue: (' , stackIndex printString , '))'.
					exprList addLast: varName , ' := ' , varName , ' - 1'] "so that varName[1] is the zero'th element"
				ifFalse: "must be a double"
					[type ~= 'double' ifTrue:
						[self error: 'unsupported type declaration in a translated primitive method'].
					 exprList addLast: varName , ' := ', self vmNameString, ' stackFloatValue: ' , stackIndex printString]]
			ifFalse: "undeclared variables are taken to be integer"
				[exprList addLast: varName , ' := ', self vmNameString, ' stackIntegerValue: ' , stackIndex printString].
	stmtList := OrderedCollection new.
	exprList do: [:e | stmtList addAll: (self statementsFor: e varName: varName)].
	^ stmtList! !

!TMethod methodsFor: 'accessing' stamp: ''!
args
	"The arguments of this method."

	^args! !

!TMethod methodsFor: 'transformations' stamp: 'ar 7/6/2003 20:09'!
asInlineNode
	^TInlineNode new method: self! !

!TMethod methodsFor: 'transformations' stamp: ''!
bindClassVariablesIn: constantDictionary
	"Class variables are used as constants. This method replaces all references to class variables in the body of this method with the corresponding constant looked up in the class pool dictionary of the source class. The source class class variables should be initialized before this method is called."

	parseTree := parseTree bindVariablesIn: constantDictionary.! !

!TMethod methodsFor: 'transformations' stamp: 'ar 7/6/2003 20:35'!
bindVariableUsesIn: aDictionary
	parseTree := parseTree bindVariableUsesIn: aDictionary.! !

!TMethod methodsFor: 'transformations' stamp: 'eem 10/29/2012 11:39'!
bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen
	"Answer either the receiver, if it contains no references to the given variables, or a new node with the given variables rebound."
	| newParseTree |
	newParseTree := parseTree bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen.
	^newParseTree = parseTree
		ifTrue: [self]
		ifFalse: [self shallowCopy
					parseTree: newParseTree;
					yourself]! !

!TMethod methodsFor: 'transformations' stamp: 'eem 11/13/2014 15:50'!
buildCaseStmt: aSendNode in: aCodeGen
	"Build a case statement node for the given send of dispatchOn:in:."
	"Note: the first argument is the variable to be dispatched on. The second argument is a constant node holding an array of unary selectors, which will be turned into sends to self."

	| unimplemented errorMessage |
	(aSendNode args size >= 2
	 and: [aSendNode args second isConstant
	 and: [aSendNode args second value isArray]]) ifFalse:
		[self error: 'wrong node structure for a case statement'].

	unimplemented := aSendNode args second value select: [:s| (aCodeGen methodNamed: s) isNil].
	unimplemented isEmpty ifFalse:
		[errorMessage := 'The following selectors in case statement "', (aSendNode printString copyUpTo: $#), '..." are unimplemented: ',
							(String streamContents: [:s| unimplemented do: [:sel| s crtab; store: sel]]).
		 aCodeGen logger nextPutAll: errorMessage; cr; flush.
		 (self confirm: errorMessage
			orCancel: aCodeGen abortBlock) ifFalse:
				[self halt]].

	^TCaseStmtNode new
		setExpression: aSendNode args first
		selectors: aSendNode args second value
		arguments: (aSendNode args copyFrom: 3 to: aSendNode args size)! !

!TMethod methodsFor: 'transformations' stamp: 'eem 10/13/2013 17:46'!
buildSwitchStmt: aSendNode parent: parentNode
	"Build a switch statement node for the given send of caseOf: or caseOf:otherwise:."
	| switch |
	switch := TSwitchStmtNode new
				expression: aSendNode receiver
				cases: aSendNode args first
				otherwiseOrNil: (aSendNode args at: 2 ifAbsent: [nil]).
	(aSendNode receiver isVariable or: [parentNode isStmtList]) ifFalse:
		[switch switchVariable: (locals add: (self extraVariableName: 'switch'))].
	^switch! !

!TMethod methodsFor: 'inlining support' stamp: 'eem 8/29/2020 14:04'!
checkForCompletenessIn: aCodeGen
	"Set the complete flag if the parse tree contains no further candidates for inlining."
	| foundIncompleteSend incompleteSends |
	aCodeGen maybeBreakForTestOfInliningOf: selector.

	foundIncompleteSend := false.
	incompleteSends := IdentitySet new.

	parseTree
		nodesDo:
			[:node|
			 node isSend ifTrue:
				[(self methodIsEffectivelyComplete: node selector in: aCodeGen)
					ifTrue:
						[(self inlineableFunctionCall: node in: aCodeGen) ifTrue:
							[incompleteSends add: node.
							 complete := false.  "more inlining to do"
							 ^self]]
					ifFalse:
						[foundIncompleteSend := true.
						 incompleteSends add: node]]]
		unless:
			[:node|
			node isSend
			and: [node selector == #cCode:inSmalltalk:
				or: [aCodeGen isAssertSelector: node selector]]].

	foundIncompleteSend ifFalse:
		[complete := true]! !

!TMethod methodsFor: 'testing' stamp: 'eem 12/15/2016 09:32'!
checkForRequiredInlinability
	"This is used in methods answering inlinability.
	 Always answer false.  But if the receiver is marked as something that must be inlined (inline == #always) raise an error."
	(inline == #always and: [complete]) ifTrue:
		[self error: 'cannot inline method ', selector, ' marked as <inline: #always>'].
	^false! !

!TMethod methodsFor: 'primitive compilation' stamp: 'eem 2/17/2015 12:11'!
checkSuccessExpr
	"Answer the parse tree for an expression that aborts the primitive if there has been a failure."
	^self
		statementsFor: 'interpreterProxy failed ifTrue: [^nil]'
		varName: ''! !

!TMethod methodsFor: 'accessing' stamp: 'eem 5/31/2018 11:28'!
checkedDeclarationAt: aVariableName put: aDeclaration in: aCCodeGen
	((args includes: aVariableName)
	 or: [(locals includes: aVariableName)
	 or: [(definingClass instVarIndexFor: aVariableName asString ifAbsent: nil) notNil
	 or: [aDeclaration beginsWith: 'extern ']]]) ifFalse:
		[| msg |
		 msg := definingClass name, '>>', selector, ' contains declaration for non-existent variable ', aVariableName.
		 aCCodeGen
			ifNotNil: [aCCodeGen logger show: msg; cr]
			ifNil: [self error: msg]].
	^self declarationAt: aVariableName  "<String>" put: aDeclaration! !

!TMethod methodsFor: 'accessing' stamp: 'tpr 5/14/2013 16:47'!
clearReferencesToGlobalStruct
	globalStructureBuildMethodHasFoo := false! !

!TMethod methodsFor: 'accessing' stamp: 'eem 9/24/2010 09:54'!
comment
	^comment! !

!TMethod methodsFor: 'accessing' stamp: 'hg 8/14/2000 15:57'!
comment: aComment

	comment := aComment ! !

!TMethod methodsFor: 'accessing' stamp: 'eem 2/12/2016 15:12'!
compileTimeOptionPragmas
	"Answer the (possibly empty) sequence of option: or notOption: pragmas
	 whose arguments are values to be defined at compile time."
	^self compiledMethod pragmas select:
		[:pragma|
		 (#option: == pragma keyword or: [#notOption: == pragma keyword])
		 and: [VMBasicConstants defineAtCompileTime: (pragma argumentAt: 1)]]! !

!TMethod methodsFor: 'accessing' stamp: 'eem 7/25/2021 17:50'!
compiledMethod
	^definingClass
		compiledMethodAt: self smalltalkSelector
		ifAbsent: [definingClass compiledMethodAt: properties selector]! !

!TMethod methodsFor: 'inlining support' stamp: 'eem 9/1/2017 15:10'!
computePossibleSideEffectsInto: writtenToVars visited: visitedSelectors in: aCodeGen
	"Add all variables written to by this method and its callees to writtenToVars.
	 Avoid circularity via visitedSelectors"

	(visitedSelectors includes: selector) ifTrue:
		[^self].
	visitedSelectors add: selector.
	writtenToGlobalVarsCache ifNotNil:
		[writtenToVars addAll: writtenToGlobalVarsCache.
		 ^self].
	parseTree nodesDo:
		[ :node |
			(node isAssignment
			 and: [(locals includes: node variable name) not])
				ifTrue:
					[writtenToVars add: node variable name].
			(node isSend
			 and: [node isBuiltinOperator not
			 and: [(aCodeGen isStructSend: node) not]]) ifTrue:
				[(aCodeGen methodNamed: node selector) ifNotNil:
					[:method|
					 method
						computePossibleSideEffectsInto: writtenToVars
						visited: visitedSelectors
						in: aCodeGen]]].
	writtenToGlobalVarsCache := writtenToVars copy! !

!TMethod methodsFor: 'accessing' stamp: 'eem 3/3/2009 08:59'!
declarationAt: aVariableName
	^declarations at: aVariableName ifAbsent: [#sqInt, ' ', aVariableName]! !

!TMethod methodsFor: 'accessing' stamp: 'eem 9/28/2015 19:37'!
declarationAt: aVariableName ifAbsent: absentBlock
	^declarations at: aVariableName ifAbsent: absentBlock! !

!TMethod methodsFor: 'accessing' stamp: 'eem 3/3/2009 08:42'!
declarationAt: aVariableName  "<String>" put: aDeclaration "<String>" "^aDeclaration"
	^declarations at: aVariableName put: aDeclaration! !

!TMethod methodsFor: 'accessing' stamp: ''!
declarations
	"The type declaration dictionary of this method."

	^declarations! !

!TMethod methodsFor: 'testing' stamp: 'eem 12/12/2015 13:51'!
definedAsMacro
	^properties notNil
	  and: [(properties includesKey: #cmacro:)
		or: [properties includesKey: #cmacro]]! !

!TMethod methodsFor: 'accessing' stamp: 'ar 2/3/2001 17:29'!
definingClass
	^definingClass! !

!TMethod methodsFor: 'accessing' stamp: 'ar 2/3/2001 17:29'!
definingClass: aClass
	definingClass := aClass.! !

!TMethod methodsFor: 'error handling' stamp: 'eem 6/23/2020 11:29'!
deny: aBooleanOrBlock
	aBooleanOrBlock value ifTrue: [AssertionFailure signal: 'Assertion failed']! !

!TMethod methodsFor: 'C code generation' stamp: 'eem 6/7/2016 18:47'!
determineTypeFor: aNode in: aCodeGen
	aNode isSend ifTrue:
		[^aCodeGen returnTypeForSend: aNode in: self ifNil: #sqInt].
	aNode isAssignment ifTrue:
		[^self determineTypeFor: aNode expression in: aCodeGen].
	self error: 'don''t know how to extract return type from this kind of node'! !

!TMethod methodsFor: 'transformations' stamp: 'eem 10/30/2013 13:28'!
elideAnyFinalReturn
	"For super expansions we need to eliminate any final return to prevent premature exit.
	 Anything meaningful in the returned expression must be retained."

	| stmtList expr |
	stmtList := parseTree statements asOrderedCollection.
	stmtList last isReturn ifTrue:
		[expr := stmtList last expression.
		 (expr isVariable and: [expr name = 'self'])
			ifTrue: [stmtList := stmtList allButLast]
			ifFalse: [stmtList at: stmtList size put: expr].
		parseTree setStatements: stmtList]! !

!TMethod methodsFor: 'C code generation' stamp: 'eem 12/8/2017 14:02'!
emitCCodeOn: aStream generator: aCodeGen
	"Emit C code for this method onto the given stream.
	 All calls to inlined methods should already have been expanded."
	| bodyStream conditional |

	aCodeGen currentMethod: self.
	usedVariablesCache := Set new.
	self emitCCommentOn: aStream.	"place method comment and method name before function."
	aStream crtab; nextPutAll: '/* '; nextPutAll: self definingClass name; nextPutAll: '>>#'; nextPutAll: self smalltalkSelector; nextPutAll: ' */'.	
	aStream cr. 
	conditional := self emitCFunctionPrototype: aStream generator: aCodeGen isPrototype: false.
	aStream cr; nextPut: ${.
	bodyStream := ReadWriteStream on: (ByteString new: 128).
	aCodeGen "Generation will note used variables in usedVariablesCache"
		pushScope: declarations
		while: [parseTree emitCCodeOn: bodyStream level: 1 generator: aCodeGen].
	self emitCLocalsOn: aStream generator: aCodeGen.
	usedVariablesCache := nil.
	aStream nextPutAll: bodyStream contents.
	(returnType = #void or: [parseTree endsWithReturn]) ifFalse:
		[aStream tab; nextPutAll: 'return 0;'; cr].
	aStream nextPut: $}; cr.
	conditional ifTrue:
		[self terminateConditionalDefineFor: self compileTimeOptionPragmas on: aStream]! !

!TMethod methodsFor: 'C code generation' stamp: 'eem 5/21/2010 17:06'!
emitCCommentOn: aStream
	"Emit the transferred Smalltalk comments as C comments."

	comment ifNotNil: [
		aStream cr;cr.
		1 to: comment size do: [:index | | startPos |
			aStream nextPutAll: '/*'; tab.
			startPos := aStream position.
			self printSingleComment: (comment at: index)
				on: aStream
				indent: 1
				tabWidth: 4
				lineBreak: 78.
			aStream position - startPos > (comment at: index) size ifTrue: "probably a multi-line comment"
				[aStream cr].
			aStream nextPutAll: ' */'; cr]]! !

!TMethod methodsFor: 'C code generation' stamp: 'eem 2/12/2016 16:51'!
emitCFunctionPrototype: aStream generator: aCodeGen
	"Emit a C function header for this method onto the given stream."

	properties ifNotNil:
		[(properties at: #api: ifAbsent: []) ifNotNil:
			[:pragma|
			aStream
				nextPutAll: (pragma argumentAt: 1);
				nextPut: $;;
				cr.
			^self]].
	self emitCFunctionPrototype: aStream generator: aCodeGen isPrototype: true! !

!TMethod methodsFor: 'C code generation' stamp: 'eem 04/27/2017 12:55'!
emitCFunctionPrototype: aStream generator: aCodeGen isPrototype: isPrototype "<Boolean>"
	"Emit a C function header for this method onto the given stream.
	 Answer if the method has any compileTimeOptionPragmas"
	| compileTimeOptionPragmas returnTypeIsFunctionPointer |
	(compileTimeOptionPragmas := self compileTimeOptionPragmas) notEmpty ifTrue:
		[self outputConditionalDefineFor: compileTimeOptionPragmas on: aStream].
	returnTypeIsFunctionPointer := returnType notNil
									and: [returnType last = $)
									and: [returnType includesSubstring: (aCodeGen cFunctionNameFor: selector)]].
	export 
		ifTrue:
			[aStream nextPutAll: 'EXPORT('; nextPutAll: returnType; nextPut: $)]
		ifFalse:
			[self isStatic
				ifTrue: [aStream nextPutAll: 'static ']
				ifFalse:
					[isPrototype ifTrue:
						[aStream nextPutAll: 'extern ']].
			 (isPrototype or: [inline ~~ #always]) ifFalse: [aStream nextPutAll: 'inline '].
			 aStream nextPutAll: (returnType ifNil: [#sqInt])].
	(functionAttributes isNil or: [returnTypeIsFunctionPointer]) ifFalse:
		[aStream space; nextPutAll: functionAttributes].
	isPrototype ifTrue: [aStream space] ifFalse: [aStream cr].
	returnTypeIsFunctionPointer ifFalse:
		[aStream
			nextPutAll: (aCodeGen cFunctionNameFor: selector);
			nextPut: $(.
		args isEmpty
			ifTrue: [aStream nextPutAll: #void]
			ifFalse:
				[args
					do: [:arg| aStream nextPutAll: (self declarationAt: arg)]
					separatedBy: [aStream nextPutAll: ', ']].
		aStream nextPut: $)].
	isPrototype ifTrue:
		[aStream nextPut: $;; cr.
		 compileTimeOptionPragmas isEmpty ifFalse:
			[aCodeGen maybeEmitPrimitiveFailureDefineFor: selector on: aStream.
			 self terminateConditionalDefineFor: compileTimeOptionPragmas on: aStream]].
	^compileTimeOptionPragmas notEmpty! !

!TMethod methodsFor: 'C code generation' stamp: 'eem 5/30/2018 20:39'!
emitCLocalsOn: aStream generator: aCodeGen
	"Emit a C function header for this method onto the given stream."

	| volatileVariables maybeExternFunctions |
	volatileVariables := properties includesKey: #volatile.
	self refersToGlobalStruct ifTrue:
		[aStream
			next: 3 put: Character space; "there's already an opening ${ on this line; see sender"
			nextPutAll: (volatileVariables
						ifTrue: ['DECL_MAYBE_VOLATILE_SQ_GLOBAL_STRUCT']
						ifFalse: ['DECL_MAYBE_SQ_GLOBAL_STRUCT'])].
	aStream cr.
	maybeExternFunctions := (declarations select: [:decl| decl beginsWith: 'extern']) keys.
	(locals isEmpty and: [maybeExternFunctions isEmpty]) ifFalse:
		[(aCodeGen sortStrings: locals, maybeExternFunctions) do:
			[ :var | | decl |
			decl := self declarationAt: var.
			(volatileVariables
			 or: [(decl beginsWith: 'static')
			 or: [(decl beginsWith: 'extern')
			 or: [usedVariablesCache includes: var]]]) ifTrue:
				[aStream next: 4 put: Character space.
				 volatileVariables ifTrue:
					[aStream nextPutAll: #volatile; space].
				 aStream
					nextPutAll: decl;
					nextPut: $;;
					cr]].
		 aStream cr]! !

!TMethod methodsFor: 'C code generation' stamp: 'eem 4/13/2015 17:38'!
emitInlineOn: aStream level: level generator: aCodeGen
	"Emit C code for this method onto the given stream. All calls to inlined methods should already have been expanded."
	self removeUnusedTempsAndNilIfRequiredIn: aCodeGen.
	sharedLabel ifNotNil:
		[aStream crtab: level-1; nextPutAll: sharedLabel; nextPut: $:.
		aStream crtab: level.
		aStream nextPutAll: '/* '; nextPutAll: selector; nextPutAll: ' */'.
		aStream crtab: level].
	aStream nextPut: ${.
	locals isEmpty ifFalse:
		[(aCodeGen sortStrings: locals) do:
			[:var|
			 aStream
				crtab: level+1;
				nextPutAll: (self declarationAt: var);
				nextPut: $;].
			 aStream cr].
	aStream crtab: level+1.
	aCodeGen outputAsmLabel: selector on: aStream.
	aStream crtab: level+1.
	aCodeGen
		pushScope: declarations
		while: [parseTree emitCCodeOn: aStream level: level+1 generator: aCodeGen].
	aStream tab: level; nextPut: $}! !

!TMethod methodsFor: 'C code generation' stamp: 'eem 3/3/2009 08:59'!
emitProxyFunctionPrototype: aStream generator: aCodeGen
	"Emit an indirect C function header for this method onto the given stream."

	aStream
		nextPutAll: returnType;
		space;
		nextPutAll: '(*';
		nextPutAll: (aCodeGen cFunctionNameFor: selector);
		nextPutAll: ')('.
	args isEmpty
		ifTrue: [aStream nextPutAll: #void]
		ifFalse:
			[args
				do: [:arg| aStream nextPutAll: (self declarationAt: arg)]
				separatedBy: [ aStream nextPutAll: ', ' ]].
	aStream nextPut: $)! !

!TMethod methodsFor: 'testing' stamp: 'eem 8/8/2008 12:02'!
endsWithReturn
	"Answer true if the last statement of this method is a return."

	^parseTree endsWithReturn! !

!TMethod methodsFor: 'inlining' stamp: 'eem 3/13/2017 15:48'!
ensureConditionalAssignmentsAreTransformedIn: aCodeGen
	"Make passes transforming
		foo := expr ifTrue: [a] ifFalse: [b]
	 into
		expr ifTrue: [foo := a] ifFalse: [foo := b]
	 until no such instances exist in the tree.  This is needed for correct inlining
	 given the limitations of inlineCodeOrNilForStatement:returningNodes:in:"
	| transformedAssignments |
	[transformedAssignments := Dictionary new.
	 parseTree
		nodesDo:
			[:node|
			(self transformConditionalAssignment: node in: aCodeGen) ifNotNil:
				[:replacement|
				 transformedAssignments at: node put: replacement]]
		unless: "Don't inline the arguments to asserts to keep the asserts readable"
			[:node|
			node isSend
			and: [node selector == #cCode:inSmalltalk:
				or: [aCodeGen isAssertSelector: node selector]]].
	 transformedAssignments notEmpty
	 and: [self replaceNodesIn: transformedAssignments.
		  true]] whileTrue
		! !

!TMethod methodsFor: 'transformations' stamp: 'eem 9/1/2017 15:10'!
ensureToByDoLoopLimitIsSafeAndEfficient: node in: aCodeGen
	"For both safety and efficiency, make sure that to:[by:]do: loops
	 with complex limits have a variable to hold the limit expression.
	 In C the limit expression is evaluated each time round the loop
	 so if the loop has side-effects (which it usually will), the C compiler
	 may not be able to optimize the limit expression itself."
	| limitExpr hasSideEffects |
	 limitExpr := node args first.
	 hasSideEffects := limitExpr anySatisfy:
						[:subNode|
						subNode isSend
						and: [(aCodeGen isBuiltinSelector: subNode selector) not
						and: [(aCodeGen isStructSend: subNode) not]]].
	 node args size = 4
		ifTrue:
			[hasSideEffects
				ifTrue: [locals add: node args last name]
				ifFalse: [node arguments: node args allButLast]]
		ifFalse: "If the expression is complex but as yet there is no limit variable, add it"
			[hasSideEffects ifTrue:
				[| var |
				 var := self unusedNamePrefixedBy: 'toDoLimit' avoiding: locals. "N.B. adds it to locals!!!!"
				 node arguments: node args, {TVariableNode new setName: var; yourself}.
				 declarations
					at: node args third args first
					ifPresent: [:decl| self declarationAt: var put: (self typeFor: node args third args first in: aCodeGen), ' ', var]]]! !

!TMethod methodsFor: 'inlining' stamp: 'eem 4/30/2016 10:54'!
exitVar: exitVar label: exitLabel
	"Replace each return statement in this method with an assignment to the
	 exit variable followed by either a return or a goto to the given label.
	 Answer if a goto was generated."
	"Optimization: If exitVar is nil, the return value of the inlined method is not being used, so don't add the assignment statement."

	| labelUsed map elisions eliminateReturnSelfs |
	labelUsed := false.
	map := Dictionary new.
	elisions := Set new.
	"Conceivably one might ^self from a struct class and mean it.  In most cases though
	 ^self means `get me outta here, fast'.  So unless this method is from a VMStruct class,
	 elide any ^self's"
	eliminateReturnSelfs := ((definingClass inheritsFrom: VMClass) and: [definingClass isStructClass]) not
							  and: [returnType = #void or: [returnType = #sqInt]].
	parseTree nodesDo:
		[:node | | replacement |
		node isReturn ifTrue:
			[self transformReturnSubExpression: node
				toAssignmentOf: exitVar
				andGoto: exitLabel
				unless: eliminateReturnSelfs
				into: [:rep :labelWasUsed|
					replacement := rep.
					labelWasUsed ifTrue: [labelUsed := true]].
			"replaceNodesIn: is strictly top-down, so any replacement for ^expr ifTrue: [...^fu...] ifFalse: [...^bar...]
			 will prevent replacement of either ^fu or ^bar. The corollary is that ^expr ifTrue: [foo] ifFalse: [^bar]
			 must be transformed into expr ifTrue: [^foo] ifFalse: [^bar]"
			(node expression isConditionalSend
			 and: [node expression hasExplicitReturn])
				ifTrue:
					[elisions add: node.
					 (node expression args reject: [:arg| arg endsWithReturn]) do:
						[:nodeNeedingReturn|
						 self transformReturnSubExpression: nodeNeedingReturn statements last
							toAssignmentOf: exitVar
							andGoto: exitLabel
							unless: eliminateReturnSelfs
							into: [:rep :labelWasUsed|
								replacement := rep.
								labelWasUsed ifTrue: [labelUsed := true]].
						 map
							at: nodeNeedingReturn statements last
							put: replacement]]
				ifFalse:
					[map
						at: node
						put: (replacement ifNil:
								[TLabeledCommentNode new setComment: 'return ', node expression printString])]]].
	map isEmpty ifTrue:
		[self deny: labelUsed.
		 ^false].
	"Now do a top-down replacement for all returns that should be mapped to assignments and gotos"
	parseTree replaceNodesIn: map.
	"Now it is safe to eliminate the returning ifs..."
	elisions isEmpty ifFalse:
		[| elisionMap |
		 elisionMap := Dictionary new.
		 elisions do: [:returnNode| elisionMap at: returnNode put: returnNode expression].
		 parseTree replaceNodesIn: elisionMap].
	"Now flatten any new statement lists..."
	parseTree nodesDo:
		[:node| | list |
		(node isStmtList
		 and: [node statements notEmpty
		 and: [node statements last isStmtList]]) ifTrue:
			[list := node statements last statements.
			 node statements removeLast; addAllLast: list]].
	^labelUsed! !

!TMethod methodsFor: 'accessing' stamp: 'jm 11/24/1998 09:03'!
export

	^ export
! !

!TMethod methodsFor: 'accessing' stamp: 'eem 1/9/2014 11:01'!
export: aBoolean
	export := aBoolean! !

!TMethod methodsFor: 'initialization' stamp: 'eem 10/13/2013 17:17'!
extraVariableName: root
	extraVariableNumber := extraVariableNumber
								ifNil: [0]
								ifNotNil: [extraVariableNumber + 1].
	^root, extraVariableNumber printString! !

!TMethod methodsFor: 'transformations' stamp: 'eem 10/13/2013 17:15'!
extraVariableNumber
	^extraVariableNumber! !

!TMethod methodsFor: 'transformations' stamp: 'eem 12/15/2008 09:52'!
extractDirective: theSelector valueBlock: aBlock default: defaultResult
	"Find a pragma of the form:

		<theSelector[args]>

	 Answer the result of evaluating aBock with a TSendNode corresponding
	 to the pragma node, or defaultResult if there is no matching pragma."

	| result found newStatements |
	(properties at: theSelector ifAbsent: []) ifNotNil:
		[:pragma|
		^aBlock value: (TSendNode new
							setSelector: pragma keyword
							receiver: (TVariableNode new setName: 'self')
							arguments: (pragma arguments collect: [:const| TConstantNode new setValue: const]))].
	"Pre-pragma backward compatibility:
	 Scan the top-level statements for a labelling directive of the form:

		self theSelector[args]

	 and remove the directive from the method body if found.
	 Answer the result of evaluating aBock with the send node,
	  or defaultResult if there is no labelling directive."	result := defaultResult.
	found := false.
	newStatements := OrderedCollection new: parseTree statements size.
	parseTree statements do:
		[ :stmt |
		(stmt isSend
		 and: [stmt selector = theSelector])
			ifTrue:
				[found := true.
				 result := aBlock value: stmt]
			ifFalse:
				[newStatements add: stmt]].
	^found
		ifTrue:
			[parseTree setStatements: newStatements asArray.
			 result]
		ifFalse: [defaultResult]! !

!TMethod methodsFor: 'transformations' stamp: 'eem 9/20/2008 18:26'!
extractExpandCaseDirective
	"Scan the top-level statements for an inlining directive of the form:
		self expandCases
	 and remove the directive from the method body. Answer whether
	 there was such a directive."

	^self
		extractDirective: #expandCases
		valueBlock: [:sendNode| true]
		default: false! !

!TMethod methodsFor: 'transformations' stamp: 'eem 7/22/2021 20:54'!
extractExportDirective
	"Scan the top-level statements for an inlining directive of the form:

		self export: <boolean>

	 and remove the directive from the method body. Answer the
	 argument of the directive or false if there is no export directive."

	^self
		extractDirective: ((properties includesKey: #export:flags:) ifTrue: [#export:flags:] ifFalse: [#export:])
		valueBlock: [:sendNode| sendNode args first value ~= false]
		default: false! !

!TMethod methodsFor: 'inlining support' stamp: 'eem 2/12/2015 10:36'!
extractInlineDirective
	"Scan the pragmas (or top-level statements) for an inlining directive of the form:

		<inline: <boolean|#never|#dontCare|#asSpecified|#always>

	 Answer a boolean equivalent to the argument of the directive or #dontCare if there is no inlining directive."

	sharedCase ifNotNil: [^false]. "don't auto-inline shared code; it gets handled specially"
	^self
		extractDirective: #inline:
		valueBlock:
			[:sendNode| #(true always) includes: (inline := sendNode args first value)]
		default: #dontCare! !

!TMethod methodsFor: 'transformations' stamp: 'eem 7/26/2014 06:54'!
extractSharedCase
	"Scan the pragmas for an shared case directive of the form:
		<sharedCodeNamed: 'sharedLabel' inCase: 'sharedCase'.>
		<sharedCodeInCase: 'sharedCase'.>
	or the older top-level statements for the form
		self sharedCodeNamed: 'sharedLabel' inCase: 'sharedCase'.
		self sharedCodeInCase: 'sharedCase'.
	in which case remove the directive from the method body."

	self extractDirective: #sharedCodeNamed:inCase:
		valueBlock: [:sendNode|
			args isEmpty ifFalse:
				[self error: 'Cannot share code sections in methods with arguments'].
			sharedLabel := sendNode args first value.
			sharedCase := sendNode args last value]
		default: nil.
	self extractDirective: #sharedCodeInCase:
		valueBlock: [:sendNode|
			args isEmpty ifFalse:
				[self error: 'Cannot share code sections in methods with arguments'].
			sharedLabel := selector.
			sharedCase := sendNode args last value]
		default: nil! !

!TMethod methodsFor: 'transformations' stamp: 'eem 2/15/2009 09:59'!
extractStaticDirective
	"Scan the top-level statements for an inlining directive of the form:

		self static: <boolean>

	 and remove the directive from the method body. Answer the
	 argument of the directive or true if there is no static directive."

	^self
		extractDirective: #static:
		valueBlock: [:sendNode| sendNode args first value ~= false]
		default: (export or: [(properties includesKey: #api) or: [properties includesKey: #api:]]) not! !

!TMethod methodsFor: 'primitive compilation' stamp: 'ar 4/4/2006 21:15'!
fetchRcvrExpr
	"Return the parse tree for an expression that fetches the receiver from the stack."

	| expr |
	expr := 'rcvr := ', self vmNameString, ' stackValue: (', args size printString, ')'.
	^ self statementsFor: expr varName: ''
! !

!TMethod methodsFor: 'utilities' stamp: 'eem 4/22/2015 17:43'!
findReadBeforeAssignedIn: variables in: aCodeGen
	| readBeforeAssigned |
	readBeforeAssigned := Set new.
	parseTree
		addReadBeforeAssignedIn: variables
		to: readBeforeAssigned
		assignments: Set new
		in: aCodeGen.
	^readBeforeAssigned! !

!TMethod methodsFor: 'primitive compilation' stamp: 'eem 9/19/2020 08:13'!
fixUpReturns: argCount postlog: postlog
	"Replace each return statement in this method with (a) the given postlog, (b) code to pop the receiver and the given number of arguments, and (c) code to push the integer result and return."

	parseTree nodesDo:
		[:node | | newStmts |
		node isStmtList ifTrue:
			[newStmts := OrderedCollection new: node statements size + 10.
			node statements do:
				[:stmt |
				 stmt isReturn
					ifTrue:
						[(stmt expression isSend and: [#primitiveFail = stmt expression selector])
							ifTrue:  "failure return"
								[newStmts
									addLast: stmt expression;
									addLast: (TReturnNode new setExpression:
												(TVariableNode new setName: 'null'))]
							ifFalse:  "normal return"
								[newStmts
									addAll: postlog;
									addLast: (TReturnNode new setExpression:
												((stmt expression isVariable and: [#('nil' 'null' 'self') includes: stmt expression name])
													ifTrue:
														[TSendNode new
															setSelector: #methodReturnReceiver
															receiver: (TVariableNode new setName: self vmNameString)
															arguments: #()]
													ifFalse:
														[TSendNode new
															setSelector: #methodReturnInteger:
															receiver: (TVariableNode new setName: self vmNameString)
															arguments: {stmt expression}]))]]
					ifFalse:
						[newStmts addLast: stmt]].
			node setStatements: newStmts asArray]]! !

!TMethod methodsFor: 'utilities' stamp: 'eem 5/20/2010 17:04'!
freeVariableReferences
	"Answer a collection of variables referenced this method, excluding locals, arguments, and pseudovariables."

	| refs |
	refs := Set new.
	parseTree nodesDo: [ :node |
		node isVariable ifTrue: [ refs add: node name asString ].
	].
	args do: [ :var | refs remove: var asString ifAbsent: [] ].
	locals do: [ :var | refs remove: var asString ifAbsent: [] ].
	#('self' 'nil' 'true' 'false') do: [ :var | refs remove: var ifAbsent: [] ].
	^refs! !

!TMethod methodsFor: 'accessing' stamp: 'eem 9/24/2015 10:18'!
functionAttributes
	^functionAttributes! !

!TMethod methodsFor: 'accessing' stamp: 'eem 9/24/2015 10:18'!
functionAttributes: aString
	functionAttributes := aString! !

!TMethod methodsFor: 'type inference' stamp: 'eem 11/1/2013 14:42'!
harmonizeSignedAndUnsignedTypesIn: aSetOfTypes
	"Eliminate signed/unsigned conflicts in aSetOfTypes"
	| sqs usqs |
	#(char short int #'unsigned char' #'unsigned short' #'unsigned int')
		with: #(sqInt sqInt sqInt #usqInt #usqInt #usqInt)
		do: [:type :replacement|
			(aSetOfTypes includes: type) ifTrue:
				[aSetOfTypes remove: type; add: replacement]].
	sqs := aSetOfTypes select: [:t| t beginsWith: 'sq'].
	usqs := aSetOfTypes select: [:t| t beginsWith: 'usq'].
	^(sqs size + usqs size = aSetOfTypes size
	   and: [sqs notEmpty
	   and: [sqs allSatisfy: [:t| usqs includes: 'u', t]]])
		ifTrue: [sqs]
		ifFalse: [aSetOfTypes]! !

!TMethod methodsFor: 'testing' stamp: 'eem 2/6/2020 14:41'!
hasEffect
	"Answer if this node has an effect on execution state (does something).
	 Statements that don't have any effect can be elided if their value is unused."
	^parseTree hasEffect! !

!TMethod methodsFor: 'utilities' stamp: 'eem 9/20/2008 17:33'!
hasMoreSendsThan: nSends
	"Answer the number of nodes in this method's parseTree (a rough measure of its size)."

	| cnt |
	cnt := 0.
	parseTree nodesDo:
		[:node|
		(node isSend
		 and: [(cnt := cnt + 1) >= nSends]) ifTrue:
			[^true]].
	^false! !

!TMethod methodsFor: 'testing' stamp: 'eem 9/13/2010 12:12'!
hasProperties
	^properties notNil and: [properties notEmpty]! !

!TMethod methodsFor: 'testing' stamp: ''!
hasReturn
	"Answer true if this method contains a return statement."

	parseTree nodesDo: [ :n | n isReturn ifTrue: [ ^ true ]].
	^ false! !

!TMethod methodsFor: 'utilities' stamp: 'eem 7/7/2021 11:10'!
hasUnrenamableCCode
	"Answer if the receiver uses inlined C strings which
	 is not currently renamed properly by the the inliner."

	^parseTree anySatisfy: [:node| node isStringishCCode]! !

!TMethod methodsFor: 'inlining support' stamp: 'eem 12/26/2016 16:55'!
incompleteSendsIn: aCodeGen
	"Debugging support; answer the incomplete and inlineable sends in the receiver."
	| incompleteSends inlineableSends |
	aCodeGen maybeBreakForTestOfInliningOf: selector.

	incompleteSends := IdentitySet new.
	inlineableSends := IdentitySet new.

	parseTree
		nodesDo:
			[:node|
			 node isSend ifTrue:
				[(self methodIsEffectivelyComplete: node selector in: aCodeGen)
					ifTrue:
						[(self inlineableFunctionCall: node in: aCodeGen) ifTrue:
							[inlineableSends add: node]]
					ifFalse:
						[incompleteSends add: node]]]
		unless:
			[:node|
			node isSend
			and: [node selector == #cCode:inSmalltalk:
				or: [aCodeGen isAssertSelector: node selector]]].

	^{incompleteSends. inlineableSends}! !

!TMethod methodsFor: 'type inference' stamp: 'eem 1/29/2020 11:39'!
inferReturnTypeFromReturnsIn: aCodeGen
	"Attempt to infer the return type of the receiver from returns in the parse tree."

	"this for determining which returns have which return types:"
	"aCodeGen
		pushScope: declarations
		while: [parseTree
				nodesSelect: [:n| n isReturn]
				thenCollect: [:n| | s |
					s := Set new.
					self addTypesFor: n expression to: s in: aCodeGen.
					{n. s}]]"
			
	aCodeGen maybeBreakForTestToInline: selector in: self.
	returnType ifNotNil: [^self].
	aCodeGen
		pushScope: declarations
		while:
			[| hasReturn returnTypes |
			 hasReturn := false.
			 returnTypes := Set new.
			 "Debug:
			 (| rettypes |
			  rettypes := Dictionary new.
			  parseTree nodesDo:
				[:node|
				node isReturn ifTrue:
					[| types |
					 self addTypesFor: node expression to: (types := Set new) in: aCodeGen.
					 rettypes at: node expression put: types]].
			  rettypes)"
			 parseTree nodesDo:
				[:node|
				node isReturn ifTrue:
					[hasReturn := true.
					 "If we encounter a send of an as-yet-untyped method then abort,
					  retrying and computing the type when that method is fully typed."
					 (self addTypesFor: node expression to: returnTypes in: aCodeGen) ifTrue:
						[^self]]].
			returnTypes remove: #implicit ifAbsent: [].
			returnTypes := aCodeGen harmonizeReturnTypesIn: returnTypes.
			hasReturn
				ifTrue:
					[returnTypes size > 1 ifTrue:
						[| message |
						 message := String streamContents:
										[:s|
										 s nextPutAll: 'conflicting return types '.
										 returnTypes
											do: [:t| s nextPutAll: t]
											separatedBy: [s nextPutAll: ', '].
										 s nextPutAll: ' in '; nextPutAll: selector; cr].
						 Notification signal: message.
						 aCodeGen logger ensureCr; show: message.
						 aCodeGen maybeBreakOnInlineIn: self].
					 returnTypes size = 1 ifTrue:
						[self returnType: returnTypes anyOne]]
				ifFalse:
					[self returnType: (aCodeGen implicitReturnTypeFor: selector)]]! !

!TMethod methodsFor: 'type inference' stamp: 'eem 6/8/2016 09:20'!
inferReturnTypeIn: aCodeGen
	"Attempt to infer the return type of the receiver and answer if it changed."

	| existingReturnType |
	existingReturnType := returnType.
	self inferReturnTypeFromReturnsIn: aCodeGen.
	^existingReturnType ~= returnType! !

!TMethod methodsFor: 'type inference' stamp: 'eem 9/12/2020 13:07'!
inferTypesForImplicitlyTypedVariablesIn: aCodeGen
	"infer types for untyped variables from assignments and arithmetic uses.
	 For debugging answer a Dictionary from var to the nodes that determined types
	 This for debugging:
		(self copy inferTypesForImplicitlyTypedVariablesIn: aCodeGen)"
	| alreadyExplicitlyTypedOrNotToBeTyped asYetUntyped mustBeSigned newDeclarations effectiveNodes |
	aCodeGen maybeBreakForTestToInline: selector in: self.
	alreadyExplicitlyTypedOrNotToBeTyped := declarations keys asSet.
	asYetUntyped := locals copyWithoutAll: alreadyExplicitlyTypedOrNotToBeTyped.
	mustBeSigned := Set new.
	newDeclarations := Dictionary new.
	effectiveNodes := Dictionary new. "this for debugging"
	parseTree nodesDo:
		[:node| | type var |
		"If there is something of the form i >= 0, then i should be signed, not unsigned."
		(node isSend
		 and: [(locals includes: (var := node receiver variableNameOrNil))
		 and: [(#(<= < >= >) includes: node selector)
		 and: [node args first isConstant
		 and: [node args first value = 0]]]]) ifTrue:
			[mustBeSigned add: var.
			 effectiveNodes at: var put: { #signed. node }, (effectiveNodes at: var ifAbsent: [#()])].
		"if an assignment to an untyped local of a known type, set the local's type to that type.
		 Only observe known sends (methods in the current set) and typed local variables."
		(node isAssignment
		 and: [(locals includes: (var := node variable name))
		 and: [(alreadyExplicitlyTypedOrNotToBeTyped includes: var) not]]) ifTrue: "don't be fooled by previously inferred types"
		 	[type := node expression isSend
						ifTrue: [aCodeGen returnTypeForSend: node expression in: self ifNil: nil]
						ifFalse: [self typeFor: (node expression isAssignment
													ifTrue: [node expression variable]
													ifFalse: [node expression]) in: aCodeGen].
			 type "If untyped, then cannot type the variable yet. A subsequent assignment may assign a subtype of what this type ends up being"
				ifNil: "Further, if the type derives from an as-yet-untyped method, we must defer."
					[node expression isSend ifTrue:
						[alreadyExplicitlyTypedOrNotToBeTyped add: var.
						 (aCodeGen methodNamed: node expression selector) ifNotNil:
							[newDeclarations removeKey: var ifAbsent: nil]]]
				ifNotNil: "Merge simple types (but *don't* merge untyped vars); complex types must be defined by the programmer."
					[((aCodeGen isSimpleType: type) or: [aCodeGen isFloatingPointCType: type]) ifTrue:
						[(asYetUntyped includes: var)
							ifTrue: [newDeclarations at: var put: type, ' ', var. asYetUntyped remove: var]
							ifFalse:
								[aCodeGen mergeTypeOf: var in: newDeclarations with: type method: self].
						 effectiveNodes at: var put: { newDeclarations at: var. node }, (effectiveNodes at: var ifAbsent: [#()])]]]].
	mustBeSigned do:
		[:var|
		 (newDeclarations at: var ifAbsent: nil) ifNotNil:
			[:decl| | type |
			 type := aCodeGen extractTypeFor: var fromDeclaration: decl.
			 type first == $u ifTrue:
				[newDeclarations at: var put: (aCodeGen signedTypeForIntegralType: type), ' ', var]]].
	newDeclarations keysAndValuesDo:
		[:var :decl| declarations at: var put: decl].
	^effectiveNodes! !

!TMethod methodsFor: 'accessing' stamp: 'eem 7/15/2014 16:01'!
inline
	^inline! !

!TMethod methodsFor: 'accessing' stamp: 'eem 12/26/2016 17:20'!
inline: aBoolean
	inline := aBoolean! !

!TMethod methodsFor: 'inlining' stamp: 'eem 12/15/2016 08:56'!
inlineBuiltin: aSendNode in: aCodeGen
	| sel meth inlinedReplacement |
	(aSendNode selector beginsWith: 'perform:') ifTrue:
		[^self inlineFunctionCall: aSendNode asTransformedConstantPerform in: aCodeGen].
	sel := aSendNode receiver selector.
	meth := aCodeGen methodNamed: sel.
	(meth notNil and: [meth inline == true]) ifFalse: [^nil].
	(meth isFunctionalIn: aCodeGen) ifTrue:
		[inlinedReplacement := (aCodeGen methodNamed: aSendNode receiver selector) copy
									inlineFunctionCall: aSendNode receiver
									in: aCodeGen.
		 ^TSendNode new
			setSelector: aSendNode selector
			receiver: inlinedReplacement
			arguments: aSendNode args copy].
	(self isInlineableConditional: aSendNode in: aCodeGen) ifTrue:
		[^self inlineConditional: aSendNode in: aCodeGen].
	^nil! !

!TMethod methodsFor: 'inlining' stamp: 'eem 3/30/2013 11:59'!
inlineCaseStatementBranchesIn: aCodeGen localizingVars: varsList 
	| maxTemp usedVars v exitLabel |
	maxTemp := 0.
	parseTree nodesDo:
		[:n |
		n isCaseStmt ifTrue:
			[n cases do:
				[:stmtNode | | newStatements stmt meth |
				(stmt := stmtNode statements first) isSend ifTrue:
					[(meth := (aCodeGen methodNamed: stmt selector)) isNil ifFalse:
						[(meth hasUnrenamableCCode
						   or: [meth args notEmpty]) ifFalse:
							[meth := meth copy.
							 meth hasReturn
								ifTrue:
									[exitLabel := meth unusedLabelForInliningInto: self.
									 meth exitVar: nil label: exitLabel.
									 labels add: exitLabel]
								ifFalse: [exitLabel := nil].
							meth renameLabelsForInliningInto: self.
							labels addAll: meth labels.
							newStatements := stmtNode statements asOrderedCollection allButFirst.
							exitLabel ifNotNil:
								[newStatements addFirst: (TLabeledCommentNode new
																setLabel: exitLabel
																comment: 'end case')].
							newStatements
								addFirst: meth asInlineNode;
								addFirst: (TLabeledCommentNode new setComment: meth selector).
							stmtNode setStatements: newStatements]]]]]].
	usedVars := (locals , args) asSet.
	1 to: maxTemp do:
		[:i |
		v := 't' , i printString.
		(usedVars includes: v) ifTrue:
			[self error: 'temp variable name conflicts with an existing local or arg'].
		locals addLast: v].
	"make local versions of the given globals"
	locals addAll: (varsList reject: [:var | usedVars includes: var])! !

!TMethod methodsFor: 'inlining' stamp: 'eem 4/28/2016 17:33'!
inlineCodeOrNilForStatement: aNode returningNodes: returningNodes in: aCodeGen
	"If the given statement node can be inlined, answer the statements that replace it. Otherwise, answer nil."

	| stmts |
	(aNode isReturn
	 and: [self inlineableSend: aNode expression in: aCodeGen]) ifTrue:
		[stmts := self inlineSend: aNode expression
						directReturn: true exitVar: nil in: aCodeGen.
		stmts last endsWithReturn ifFalse:
			[stmts at: stmts size put: stmts last asReturnNode].
		^stmts].
	(aNode isAssignment
	 and: [self inlineableSend: aNode expression in: aCodeGen]) ifTrue:
		[^self inlineSend: aNode expression
				directReturn: false exitVar: aNode variable name in: aCodeGen].
	(aNode isSend
	 and: [self inlineableSend: aNode in: aCodeGen]) ifTrue:
		[^self inlineSend: aNode
				directReturn: (returningNodes includes: aNode) exitVar: nil in: aCodeGen].
	^nil! !

!TMethod methodsFor: 'inlining' stamp: 'eem 7/21/2014 05:52'!
inlineConditional: aSendNode in: aCodeGen
	"If possible answer the inlining of a conditional, otherwise answer nil.
	 Currently the only pattern we support is
		aSend ifTrue:/ifFalse: [...]
	 where aSend is marked inline and always answers booleans."
	self assert: (self isInlineableConditional: aSendNode in: aCodeGen).
	self assert: aSendNode args first isStmtList.
	^(aSendNode args first statements size = 1
	  and: [aSendNode args first statements first isReturn])
		ifTrue: [self inlineReturningConditional: aSendNode in: aCodeGen]
		ifFalse: [self inlineGuardingConditional: aSendNode in: aCodeGen]! !

!TMethod methodsFor: 'inlining' stamp: 'eem 6/23/2020 17:13'!
inlineFunctionCall: aSendNode in: aCodeGen
	"Answer the body of the called function, substituting the actual
	 parameters for the formal argument variables in the method body.
	 Assume caller has established that:
		1. the method arguments are all substitutable nodes, and
		2. the method to be inlined contains no additional embedded returns."

	| sel meth doNotRename argsForInlining substitutionDict |
	aCodeGen maybeBreakForInlineOf: aSendNode in: self.
	sel := aSendNode selector.
	meth := (aCodeGen methodNamed: sel) copy.
	meth ifNil:
		[^self inlineBuiltin: aSendNode in: aCodeGen].
	doNotRename := Set withAll: args.
	argsForInlining := aSendNode argumentsForInliningCodeGenerator: aCodeGen.
	meth args with: argsForInlining do:
		[ :argName :exprNode |
		exprNode isLeaf ifTrue:
			[doNotRename add: argName]].
	(meth statements size = 2
	and: [meth statements first isSend
	and: [meth statements first selector == #flag:]]) ifTrue:
		[meth statements removeFirst].
	meth renameVarsForInliningInto: self except: doNotRename in: aCodeGen.
	meth renameLabelsForInliningInto: self.
	self addVarsDeclarationsAndLabelsOf: meth except: doNotRename.
	substitutionDict := Dictionary new: meth args size * 2.
	meth args with: argsForInlining do:
		[ :argName :exprNode |
		(exprNode isVariable and: [exprNode name = argName]) ifFalse:
			[substitutionDict at: argName put: exprNode].
		(doNotRename includes: argName) ifFalse:
			[locals remove: argName]].
	meth parseTree bindVariablesIn: substitutionDict.
	^meth parseTree endsWithReturn
		ifTrue: [meth parseTree copyWithoutReturn]
		ifFalse: [meth parseTree]! !

!TMethod methodsFor: 'inlining' stamp: 'eem 2/16/2015 10:38'!
inlineGuardingConditional: aSendNode in: aCodeGen
	"Inline
		aSend ifTrue:/ifFalse: [statements]
	 where aSend is inlineable and always answers booleans.  We convert
	 the boolean returns in aSend to jumps."
	| evaluateIfTrue replacementTree map lastNode evaluateLabel skipLabel method |
	self assert: self == aCodeGen currentMethod.
	self assert: (self isInlineableConditional: aSendNode in: aCodeGen).
	aCodeGen maybeBreakForInlineOf: aSendNode in: self.
	evaluateIfTrue := aSendNode selector = #ifTrue:.
	method := (aCodeGen methodNamed: aSendNode receiver selector) copy.
	replacementTree := method inlineFunctionCall: aSendNode receiver in: aCodeGen.
	map := Dictionary new.
	(replacementTree statements last isReturn
	 and: [replacementTree statements last expression value = evaluateIfTrue]) ifTrue:
		[lastNode := replacementTree statements last].
	skipLabel := TLabeledCommentNode new setLabel:
					(self unusedLabelForInlining: method).
	replacementTree nodesDo:
		[:node| | expr |
		 node isReturn ifTrue:
			[expr := node expression.
			 self assert: (expr isConstant and: [#(true false) includes: expr value]).
			 map
				at: node
				put: (expr value ~~ evaluateIfTrue
						ifTrue: [TGoToNode new setLabel: skipLabel label]
						ifFalse:
							[node == lastNode
								ifTrue: [TLabeledCommentNode new setComment: 'end ', aSendNode receiver selector, '; fall through']
								ifFalse:
									[evaluateLabel ifNil:
										[evaluateLabel := TLabeledCommentNode new setLabel:
													(self unusedLabelForInlining: method)].
									 TGoToNode new setLabel: evaluateLabel label]])]].
	replacementTree replaceNodesIn: map.
	replacementTree comment: {'inline ', aSendNode receiver selector}.
	self addVarsDeclarationsAndLabelsOf: method except: method args.
	^TStmtListNode new
		setArguments: #()
		statements:
			(evaluateLabel
				ifNil: [replacementTree statements, aSendNode args first statements, {skipLabel}]
				ifNotNil:
					[replacementTree statements, {evaluateLabel}, aSendNode args first statements, {skipLabel}])! !

!TMethod methodsFor: 'inlining' stamp: 'eem 11/8/2018 08:36'!
inlineReturningConditional: aSendNode in: aCodeGen
	"Inline
		aSend ifTrue:/ifFalse: [^expr]
	 where aSend is inlineable and always answers booleans.  We inline ^expr
	 into aSend."
	| returnIfTrue returnNode replacementTree map lastNode label method |
	self assert: self == aCodeGen currentMethod.
	self assert: (self isInlineableConditional: aSendNode in: aCodeGen).
	aCodeGen maybeBreakForInlineOf: aSendNode receiver in: self.
	returnIfTrue := aSendNode selector = #ifTrue:.
	returnNode := aSendNode args first.
	method := (aCodeGen methodNamed: aSendNode receiver selector) copy.
	replacementTree := method inlineFunctionCall: aSendNode receiver in: aCodeGen.
	map := Dictionary new.
	"The last node is either a return or a boolean constant."
	lastNode := replacementTree statements last.
	replacementTree statements last isReturn
		ifTrue:
			[replacementTree statements last expression value == returnIfTrue ifTrue:
				[lastNode := nil "i.e. take the fall-through path and /don't/ return"]]
		ifFalse:
			[self assert: (lastNode isConstant and: [#(true false) includes: lastNode value]).
			 lastNode value == returnIfTrue ifTrue: "i.e. /do/ return"
				[map at: lastNode put: returnNode]].
	replacementTree nodesDo:
		[:node| | expr |
		 node isReturn ifTrue:
			[expr := node expression.
			 self assert: (expr isConstant and: [#(true false) includes: expr value]).
			 map
				at: node
				put: (expr value == returnIfTrue
						ifTrue: [returnNode]
						ifFalse:
							[node == lastNode
								ifTrue: [TLabeledCommentNode new setComment: 'end ', aSendNode receiver selector, '; fall through']
								ifFalse:
									[label ifNil:
										[label := TLabeledCommentNode new setLabel:
													(self unusedLabelForInlining: method)].
									 TGoToNode new setLabel: label label]])]].
	replacementTree replaceNodesIn: map.
	self addVarsDeclarationsAndLabelsOf: method except: method args.
	replacementTree comment: {'inline ', aSendNode receiver selector}.
	^label
		ifNil: [replacementTree]
		ifNotNil:
			[TStmtListNode new
				setArguments: #()
				statements: {replacementTree. label}]! !

!TMethod methodsFor: 'inlining' stamp: 'eem 5/4/2017 13:28'!
inlineSend: aSendNode directReturn: directReturn exitVar: exitVar in: aCodeGen
	"Answer a collection of statements to replace the given send.  directReturn indicates
	 that the send is the expression in a return statement, so returns can be left in the
	 body of the inlined method. If exitVar is nil, the value returned by the send is not
	 used; thus, returns need not assign to the output variable.

	 Types are propagated to as-yet-untyped variables when inlining a send that is assigned,
	 otherwise the assignee variable type must match the return type of the inlinee.  Return
	 types are not propagated."

	| sel meth methArgs exitLabel inlineStmts label exitType elidedArgs |
	sel := aSendNode selector.
	meth := aCodeGen methodNamed: sel.
	methArgs := meth args.
	"convenient for debugging..."
	aCodeGen maybeBreakForInlineOf: aSendNode in: self.
	elidedArgs := #().
	(methArgs notEmpty and: [methArgs first beginsWith: 'self_in_'])
		ifTrue: "If the first arg is not used we can and should elide it."
			[| varNode |
			 varNode := TVariableNode new setName: methArgs first.
			 (meth parseTree noneSatisfy: [:node| varNode isSameAs: node]) ifTrue:
				[elidedArgs := {methArgs first}].
			 methArgs := methArgs allButFirst].
	methArgs size = aSendNode args size ifFalse:
		[^nil].
	meth := meth copy.

	(meth statements size > 1
	 and: [meth statements first isSend
	 and: [meth statements first selector == #flag:]]) ifTrue:
		[meth statements removeFirst].

	"Propagate the return type of an inlined method"
	(directReturn or: [exitVar notNil]) ifTrue:
		[exitType := directReturn 
						ifTrue: [returnType] 
						ifFalse: [(self typeFor: exitVar in: aCodeGen) ifNil: [#sqInt]].
		(exitType = #void or: [exitType = meth returnType]) ifFalse:
			[meth propagateReturnIn: aCodeGen]].

	"Propagate any unusual argument types to untyped argument variables"
	methArgs
		with: aSendNode args
		do: [:formal :actual|
			(meth declarationAt: formal ifAbsent: nil) ifNil:
				[(self typeFor: actual in: aCodeGen) ifNotNil:
					[:type|
					type ~= #sqInt ifTrue:
						[meth declarationAt: formal put: (type last = $* ifTrue: [type, formal] ifFalse: [type, ' ', formal])]]]].

	meth renameVarsForInliningInto: self except: elidedArgs in: aCodeGen.
	meth renameLabelsForInliningInto: self.
	self addVarsDeclarationsAndLabelsOf: meth except: elidedArgs.
	meth hasReturn ifTrue:
		[directReturn ifFalse:
			[exitLabel := self unusedLabelForInliningInto: self.
			 (meth exitVar: exitVar label: exitLabel) "is label used?"
				ifTrue: [ labels add: exitLabel ]
				ifFalse: [ exitLabel := nil ]]].
	(inlineStmts := OrderedCollection new: meth statements size + meth args size + 2)
		add: (label := TLabeledCommentNode new setComment: 'begin ', sel);
		addAll: (self argAssignmentsFor: meth send: aSendNode except: elidedArgs in: aCodeGen);
		addAll: meth statements.  "method body"
	directReturn ifTrue:
		[meth endsWithReturn
			ifTrue:
				[exitVar ifNotNil: "don't remove the returns if being invoked in the context of a return"
					[inlineStmts at: inlineStmts size put: inlineStmts last copyWithoutReturn]]
			ifFalse:
				[inlineStmts add:
					(TReturnNode new setExpression: (TVariableNode new setName: 'nil'))]].
	exitLabel ifNotNil:
		[inlineStmts add:
			(TLabeledCommentNode new setLabel:
				exitLabel comment: 'end ', meth selector)].
	inlineStmts size = 1 ifTrue: "Nuke empty methods; e.g. override of flushAtCache"
		[self assert: inlineStmts first isComment.
		 inlineStmts removeFirst].
	^inlineStmts! !

!TMethod methodsFor: 'inlining' stamp: 'eem 8/29/2020 08:53'!
inlineableFunctionCall: aNode in: aCodeGen
	"Answer if the given send node is a call to a 'functional' method--a method whose body is a single return statement of some expression and whose actual parameters can all be directly substituted."

	aCodeGen maybeBreakForTestToInline: aNode in: self.
	aNode isSend ifFalse:
		[^false].
	((aCodeGen shouldGenerateAsInterpreterProxySend: aNode)
	 or: [aCodeGen isStructSend: aNode]) ifTrue:
		[^false].
	^(aCodeGen methodNamed: aNode selector)
		ifNil:
			[aNode asTransformedConstantPerform
				ifNil: [self isInlineableConditional: aNode in: aCodeGen]
				ifNotNil: [:n| self inlineableFunctionCall: n in: aCodeGen]]
		ifNotNil:
			[:m|
			 (m ~~ self
			  and: [((m isFunctionalIn: aCodeGen) or: [m mustBeInlined and: [m isComplete]])
			  and: [m mayBeInlined
			  and: [(aCodeGen mayInline: m selector)
			  and: [aNode args allSatisfy: [:a| self isSubstitutableNode: a intoMethod: m in: aCodeGen]]]]])
			 or: [m checkForRequiredInlinability]]! !

!TMethod methodsFor: 'inlining' stamp: 'eem 12/26/2016 15:16'!
inlineableSend: aNode in: aCodeGen
	"Answer if the given send node is a call to a method that can be inlined."

	| m |
	aCodeGen maybeBreakForTestToInline: aNode in: self.
	aNode isSend ifFalse: [^false].
	m := aCodeGen methodNamed: aNode selector.  "nil if builtin or external function"
	^m ~= nil
	 and: [m ~~ self
	 and: [m mayBeInlined
	 and: [(m isComplete and: [aCodeGen mayInline: m selector])
		or: [m checkForRequiredInlinability]]]]! !

!TMethod methodsFor: 'primitive compilation' stamp: 'ar 4/4/2006 21:15'!
instVarGetExprFor: varName offset: instIndex
	"Return the parse tree for an expression that fetches and converts the value of the instance variable at the given offset."

	| exprList decl stmtList |
	exprList := OrderedCollection new.
	(declarations includesKey: varName) ifTrue: [
		decl := declarations at: varName.
		(decl includes: $*) ifTrue: [  "array"
			exprList add:
				(varName, ' := ', self vmNameString, ' fetchArray: ', instIndex printString, ' ofObject: rcvr').
			exprList add: (varName, ' := ', varName, ' - 1').
		] ifFalse: [  "must be a double"
			((decl findString: 'double' startingAt: 1) = 0)
				ifTrue: [ self error: 'unsupported type declaration in a primitive method' ].
			exprList add:
				(varName, ' := ', self vmNameString, ' fetchFloat: ', instIndex printString, ' ofObject: rcvr').
		].
	] ifFalse: [  "undeclared variables are taken to be integer"
		exprList add:
			(varName, ' := ', self vmNameString, ' fetchInteger: ', instIndex printString, ' ofObject: rcvr').
	].
	stmtList := OrderedCollection new.
	exprList do: [:e | stmtList addAll: (self statementsFor: e varName: varName)].
	^ stmtList
! !

!TMethod methodsFor: 'primitive compilation' stamp: 'TPR 3/2/2000 16:13'!
instVarPutExprFor: varName offset: instIndex
	"Return the parse tree for an expression that saves the value of the integer instance variable at the given offset."

	| expr |
	(declarations includesKey: varName) ifTrue: [
		self error: 'a primitive method can only modify integer instance variables'.
	].
	expr := '', self vmNameString, ' storeInteger: ', instIndex printString, ' ofObject: rcvr withValue: ', varName.
	^ self statementsFor: expr varName: varName
! !

!TMethod methodsFor: 'testing' stamp: 'eem 7/5/2011 19:04'!
isAPIMethod
	^properties notNil
	  and: [(properties includesKey: #api)
			or: [properties includesKey: #api:]]! !

!TMethod methodsFor: 'testing' stamp: 'ikp 9/26/97 14:50'!
isAssertion
	^(selector beginsWith: 'assert') or: [selector beginsWith: 'verify']! !

!TMethod methodsFor: 'accessing' stamp: ''!
isComplete
	"A method is 'complete' if it does not contain any more inline-able calls."

	^complete! !

!TMethod methodsFor: 'inlining' stamp: 'eem 9/5/2016 15:26'!
isConditionalToBeTransformedForAssignment: aSend in: aCodeGen
	"Answer if a send is of the form
		e1
			ifTrue: [e2 ifTrue: [self m1] ifFalse: [self m2]]
			ifFalse: [self m3]
	 such that at least one of the sends mN may be inlined.."

	^(#(ifTrue:ifFalse: ifFalse:ifTrue:) includes: aSend selector)
	   and: [aSend args anySatisfy:
			[:arg| | stmt |
			self assert: arg isStmtList.
			arg statements size > 1
			or: [(stmt := arg statements first) isSwitch
			or: [stmt isSend
				and: [(aCodeGen mayInline: stmt selector)
					or: [self isConditionalToBeTransformedForAssignment: stmt in: aCodeGen]]]]]]! !

!TMethod methodsFor: 'inlining' stamp: 'eem 11/10/2017 18:29'!
isFunctionalIn: aCodeGen
	"Answer if the receiver is a functional method. That is, if it
	 consists of a single return statement of an expression that
	 contains no other returns, or an assert or flag followed by
	 such a statement.

	 Answer false for methods with return types other than the simple
	 integer types to work around bugs in the inliner."

	parseTree statements size = 1 ifFalse:
		[(parseTree statements size = 2
		  and: [parseTree statements first isSend
		  and: [parseTree statements first selector == #flag:
			or: [(aCodeGen isAssertSelector: parseTree statements first selector)
				and: [parseTree statements first selector ~~ #asserta:]]]]) ifFalse:
			[^false]].
	parseTree statements last isReturn ifFalse:
		[^false].
	parseTree statements last expression nodesDo:
		[ :n | n isReturn ifTrue: [^false]].
	^#(int #'unsigned int' #long #'unsigned long' #'long long' #'unsigned long long'
		sqInt usqInt #'sqIntptr_t' #'usqIntptr_t' sqLong usqLong
		#'int *' #'unsigned int *' #'sqInt *' #'usqInt *' #'sqLong *' #'usqLong *' #'char *'
		#'CogMethod *' #'AbstractInstruction *'
		#'FILE *') includes: returnType! !

!TMethod methodsFor: 'inlining' stamp: 'eem 10/6/2014 17:14'!
isInlineableConditional: aSendNode in: aCodeGen
	"Answer if the given send node is of the form aSend [ifTrue:|ifFalse:] [statements]
	 where the method for aSend is marked as inline and all returns within it answer booleans."
	|method |
	^(#(ifTrue: ifFalse:) includes: aSendNode selector)
	  and: [aSendNode receiver isSend
	  and: [(method := aCodeGen anyMethodNamed: aSendNode receiver selector) notNil
	  and: [method inline == true
	  and: [method parseTree statements last isReturn
	  and: [method parseTree allSatisfy:
			[:node|
			 node isReturn not
			 or: [node expression isConstant
				 and: [#(true false) includes: node expression value]]]]]]]]! !

!TMethod methodsFor: 'inlining' stamp: 'eem 7/6/2021 18:28'!
isNode: aNode substitutableFor: argName inMethod: targetMeth in: aCodeGen
	"Answer if the given parameter node may be substituted directly into the body of
	 the method during inlining, instead of being bound to the actual parameter variable.
	 We allow a constant, a local variable, or a formal parameter, or simple expressions
	 involving only these to to be directly substituted. Note that global variables cannot
	 be subsituted into methods with possible side effects (i.e., methods that may assign
	 to global variables) because the inlined method might depend on having the value of
	 the global variable captured when it is passed in as an argument."

	| theNode madeNonTrivialCall count constantExpression usageCount |
	aNode isConstant ifTrue: [^true].

	theNode := (aNode isSend and: [aNode isCast]) ifTrue: [aNode targetOfCast] ifFalse: [aNode].
	theNode isVariable ifTrue:
		[((locals includes: theNode name)
		 or: [(args includes: theNode name)
		 or: [#('self' 'true' 'false' 'nil') includes: theNode name]]) ifTrue: [^true].
		"We can substitute any variable provided it is only read in the method being inlined,
		 and if it is not read after any non-trivial call (which may update the variable)."
		madeNonTrivialCall := false.
		(targetMeth isComplete
		 and: [targetMeth parseTree
				noneSatisfy:
					[:node|
					 (node isSend
					  and: [(aCodeGen isBuiltinSelector: node selector) not]) ifTrue:
						[madeNonTrivialCall := true].
					 (madeNonTrivialCall and: [node isVariable and: [node name = argName]])
					 or: [node isAssignment
						  and: [node variable name = argName]]]
				unless:
					[:node|
					node isSend and: [aCodeGen isAssertSelector: node selector]]]) ifTrue:
			[^true].
		^targetMeth maySubstituteGlobal: theNode name in: aCodeGen].

	"don't much up asserts with complex expansions"
	(targetMeth usesVariableUninlinably: argName in: aCodeGen) ifTrue:
		[^false].

	"For now allow literal blocks to be substituted.  They better be accessed only
	 with value[:value:*] messages though!!"
	aNode isLiteralBlock ifTrue: [^true].

	"Don't inline expressions unless type-compatible,"
	aNode isSend ifTrue:
		[(aCodeGen
				isActualType: (aCodeGen returnTypeForSend: aNode in: self ifNil: #incompatible)
				compatibleWithFormalType: (self typeFor: argName in: aCodeGen)) ifFalse:
			[^false]].

	count := 0.
	constantExpression := true.
	"scan expression tree; must contain only constants, builtin ops, and inlineable vars"
	aNode nodesDo:
		[:node|
		node isConstant
			ifTrue: [] ifFalse:
		[node isSend
			ifTrue:
				[((VMBasicConstants mostBasicConstantSelectors includes: node selector)
				  or: [node isBuiltinOperator]) ifFalse: [^false].
				 count := count + 1] ifFalse:
		[node isVariable ifTrue:
			[(aCodeGen isNonArgumentImplicitReceiverVariableName: node name) ifFalse:
				[constantExpression := false.
				((locals includes: node name)
				 or: [(args includes: node name)
				 or: [(#('self' 'true' 'false' 'nil') includes: node name)
				 or: [targetMeth maySubstituteGlobal: node name in: aCodeGen]]]) ifFalse: [^false]]] ifFalse:
		[^false]]]].
	"inline constant expressions"
	constantExpression ifNil: [^true].

	"scan target to find usage count"
	usageCount := 0.
	targetMeth parseTree nodesDo:
		[:node|
		(node isVariable and: [node name = argName]) ifTrue:
			[usageCount := usageCount + 1]].
	"(usageCount > 1 and: [count <= usageCount]) ifTrue:
		[[UsageCounts := Dictionary new.
		  self removeClassVarName: #UsageCounts].
		 (UsageCounts at: usageCount ifAbsentPut: [Set new]) add: ({targetMeth. argName. aNode})]."
	"Now only inline expressions if they are used only once or are simple
	 w.r.t. the usage count, and the usage count is not large; a heuristic that seems to work well enough."
	^usageCount = 1 or: [usageCount <= 7 and: [count <= usageCount]]! !

!TMethod methodsFor: 'testing' stamp: 'eem 2/11/2009 21:21'!
isRealMethod
	^(self definedAsMacro or: [self isStructAccessor]) not! !

!TMethod methodsFor: 'testing' stamp: 'eem 2/15/2015 10:48'!
isReturnConstant
	^parseTree statements size = 1
	 and: [parseTree statements last isReturn
	 and: [parseTree statements last expression isLeaf]]! !

!TMethod methodsFor: 'accessing' stamp: 'ar 5/9/2000 12:13'!
isStatic
	^static ifNil:[false].! !

!TMethod methodsFor: 'testing' stamp: 'eem 2/11/2009 21:20'!
isStructAccessor
	^[definingClass isAccessor: selector]
		on: MessageNotUnderstood
		do: [:ex| false]! !

!TMethod methodsFor: 'inlining' stamp: 'eem 10/17/2020 15:38'!
isSubstitutableNode: aNode intoMethod: targetMeth in: aCodeGen
	"Answer true if the given parameter node is either a constant, a local variable, or a formal parameter of the receiver. Such parameter nodes may be substituted directly into the body of the method during inlining. Note that global variables cannot be subsituted into methods with possible side effects (i.e., methods that may assign to global variables) because the inlined method might depend on having the value of the global variable captured when it is passed in as an argument."

	| var |
	aNode isConstant ifTrue: [ ^ true ].

	aNode isVariable ifTrue: [
		var := aNode name.
		((locals includes: var) or: [args includes: var]) ifTrue: [ ^ true ].
		(#(self true false nil) includes: var) ifTrue: [ ^ true ].
		(targetMeth maySubstituteGlobal: var in: aCodeGen) ifTrue: [ ^ true ].
	].

	"For now allow literal blocks to be substituted.  They better be accessed only
	 with value[:value:*] messages though!!"
	aNode isStmtList ifTrue: [^true].

	"Allow inlining of struct accesses and external functions;
	 an external function won't appear in the code gen's set of methods."
	(aNode isSend
	 and: [aNode numArgs = 0
	 and: [(aCodeGen isStructSend: aNode)
		or: [(aCodeGen methodNamed: aNode selector) isNil]]]) ifTrue:
		[^true].

	"scan expression tree; must contain only constants, builtin ops, and inlineable vars"
	aNode nodesDo: [ :node |
		node isSend ifTrue: [
			(node isBuiltinOperator
			 or: [node numArgs = 0
				 and: [aCodeGen isStructSend: node]]) ifFalse: [ ^false ].
		].
		node isVariable ifTrue: [
			var := node name.
			((locals includes: var) or:
			 [(args includes: var) or:
			 [(#(self true false nil) includes: var) or:
			 [targetMeth maySubstituteGlobal: var in: aCodeGen]]]) ifFalse: [ ^ false ].
		].
		(node isConstant or: [node isVariable or: [node isSend]]) ifFalse: [ ^false ].
	].

	^ true! !

!TMethod methodsFor: 'testing' stamp: 'eem 7/26/2014 07:49'!
isTMethod
	^true! !

!TMethod methodsFor: 'accessing' stamp: 'ar 5/9/2000 12:13'!
labels

	^labels! !

!TMethod methodsFor: 'accessing' stamp: 'eem 7/21/2014 06:09'!
labels: aCollection

	labels := aCollection asSet! !

!TMethod methodsFor: 'accessing' stamp: 'ar 5/9/2000 12:13'!
locals
	"The local variables of this method."

	^locals! !

!TMethod methodsFor: 'transformations' stamp: 'eem 9/19/2020 08:41'!
mapSendsFromSelfToInterpreterProxy: selectors
	| interpreterProxyNode replacements |
	interpreterProxyNode := TVariableNode new setName: 'interpreterProxy'.
	replacements := Dictionary new.
	parseTree nodesDo:
		[:node|
		(node isSend
		 and: [node receiver isVariable
		 and: [selectors includes: node selector]]) ifTrue:
			[node receiver name = 'self' ifTrue:
				[node receiver: interpreterProxyNode].
			 node receiver name = 'SmallInteger' ifTrue:
				[replacements
					at: node
					put: (TVariableNode new setName:
							(node selector = #maxVal
									ifTrue: ['MaxSmallInteger']
									ifFalse: ['MinSmallInteger']))]]].
	replacements notEmpty ifTrue:
		[parseTree replaceNodesIn: replacements]! !

!TMethod methodsFor: 'accessing' stamp: 'eem 12/26/2016 15:16'!
mayBeInlined
	^inline == true or: [inline == nil or: [inline == #always]]! !

!TMethod methodsFor: 'inlining support' stamp: 'eem 7/30/2008 15:08'!
maySubstituteGlobal: globalVar in: aCodeGen
	"We can substitute globalVar into this method provided globalVar is only read, not written."

	writtenToGlobalVarsCache = nil ifTrue:
		[self computePossibleSideEffectsInto: (Set new: 50) visited: (Set new: 50) in: aCodeGen].
	^(writtenToGlobalVarsCache includes: globalVar) not! !

!TMethod methodsFor: 'inlining' stamp: 'eem 9/24/2010 09:47'!
mergePropertiesOfSuperMethod: superTMethod
	superTMethod hasProperties ifFalse:
		[^self].
	self hasProperties ifFalse:
		[properties := superTMethod properties.
		 ^self].
	superTMethod properties pragmas do:
		[:aPragma|
		(self shouldIncorporatePragmaFromSuperMethod: aPragma) ifTrue:
			[properties := properties copyWith: aPragma]]! !

!TMethod methodsFor: 'inlining support' stamp: 'eem 8/29/2020 13:39'!
methodIsEffectivelyComplete: selector in: aCodeGen
	"Answer if selector is effectively not inlineable in the receiver.
	 This is tricky because block inlining requires that certain methods must be inlined, which
	 can be at odds wuth the opportunistic strategy the inliner takes.  Since the inliner only
	 inlines complete methods and certain methods may never be marked as complete (e.g.
	 recursive methods) we have to short-cut certain kinds of send.  In particular, short-cut
	 sends that turn into jumps in the interpret routine (sharedCase and sharedLabel below)."
	^(aCodeGen methodNamed: selector)
		ifNil: [true] "builtins or externals are not inlineable"
		ifNotNil:
			[:m|
			 m isComplete
			 "unlinable methods can't be inlined"
			 or: [m mayBeInlined not
			 "Methods which are inlined as jumps don't need inlining"
			 or: [m sharedCase notNil or: [m sharedLabel notNil
			 "Macros and struct accessors don't need inlining"
			 or: [m definedAsMacro or: [m isStructAccessor]]]]]]! !

!TMethod methodsFor: 'accessing' stamp: 'eem 3/20/2019 11:32'!
mustBeInlined
	^inline == #always! !

!TMethod methodsFor: 'initialization' stamp: 'eem 8/21/2021 19:04'!
newCascadeTempFor: aTParseNode
	| varName varNode |
	varName := self extraVariableName: 'cascade'.
	varNode := TVariableNode new setName: varName.
	aTParseNode isLeaf ifFalse:
		[self
			declarationAt: varName
			put: [:tm :cg| | type |
				type := tm determineTypeFor: aTParseNode in: cg.
				(VMStructType structTargetKindForType: type) == #struct ifTrue:
					["can't copy structs into cascade temps; the struct is not updated.
					  must change to a pointer."
					type := type, ' *'.
					parseTree nodesDo:
						[:node|
						(node isAssignment
						 and: [node variable name = varName]) ifTrue:
							[node setExpression: (TSendNode new
													setSelector: #addressOf:
													receiver: (TVariableNode new setName: 'self')
													arguments: {node expression})]]].
				type ifNil: [#sqInt]]].
	^varNode! !

!TMethod methodsFor: 'accessing' stamp: 'eem 3/3/2009 09:06'!
newDeclarations: aDictionary
	declarations := aDictionary! !

!TMethod methodsFor: 'utilities' stamp: ''!
nodeCount
	"Answer the number of nodes in this method's parseTree (a rough measure of its size)."

	| cnt |
	cnt := 0.
	parseTree nodesDo: [ :n | cnt := cnt + 1 ].
	^cnt! !

!TMethod methodsFor: 'utilities' stamp: 'eem 12/8/2017 13:15'!
noteUsedVariableName: token
	usedVariablesCache ifNotNil:
		[usedVariablesCache add: token]! !

!TMethod methodsFor: 'private' stamp: 'eem 3/3/2009 08:44'!
oopVariable: aString

	(locals includes: aString) ifFalse:
		[locals add: aString.
		self declarationAt: aString put: 'sqInt ', aString].
	^TVariableNode new setName: aString! !

!TMethod methodsFor: 'C code generation' stamp: 'eem 2/12/2016 17:02'!
outputConditionalDefineFor: compileTimeOptionPragmas on: aStream
	aStream nextPutAll: '#if '.
	compileTimeOptionPragmas
		do: [:pragma|
			pragma keyword = #notOption: ifTrue:
				[aStream nextPut: $!!].
			aStream nextPutAll: (pragma argumentAt: 1)]
		separatedBy: [aStream nextPutAll: ' && '].
	 aStream cr! !

!TMethod methodsFor: 'accessing' stamp: ''!
parseTree
	"The parse tree of this method."

	^parseTree! !

!TMethod methodsFor: 'accessing' stamp: ''!
parseTree: aNode
	"Set the parse tree of this method."

	parseTree := aNode.! !

!TMethod methodsFor: 'primitive compilation' stamp: 'eem 9/19/2020 08:15'!
popArgsExpr: argCount
	"Answer the parse tree for an expression that pops the given number of arguments from the stack."

	^self statementsFor: '^', self vmNameString, ' ', #methodReturnReceiver varName: ''! !

!TMethod methodsFor: 'copying' stamp: 'eem 12/6/2012 14:09'!
postCopy
	args := args copy.
	locals := locals copy.
	declarations := declarations copy.
	parseTree := parseTree copy.
	labels := labels copy! !

!TMethod methodsFor: 'transformations' stamp: 'eem 7/31/2021 19:29'!
prepareMethodIn: aCodeGen
	"Record sends of builtin operators, map sends of the special selector dispatchOn:in:
	 with case statement nodes, and map sends of caseOf:[otherwise:] to switch statements.
	 Declare limit variables for to:[by:]do: loops with limits that potentially have side-effects.
	 As a hack also update the types of variables introduced to implement cascades correctly.
	 This has to be done at the same time as this is done, so why not piggy back here?"
	extraVariableNumber ifNotNil:
		[declarations keysAndValuesDo:
			[:varName :decl|
			decl isBlock ifTrue:
				[self assert: ((varName beginsWith: 'cascade') and: [varName last isDigit]).
				 locals add: varName.
				 self declarationAt: varName
					put: (decl value: self value: aCodeGen), ' ', varName]]].
	aCodeGen
		pushScope: declarations
		while:"N.B.  nodesWithParentsDo: is bottom-up, hence replacement is destructive and conserved."
			[parseTree nodesWithParentsDo:
				[:node :parent|
				 node isSend ifTrue:
					[aCodeGen ifStaticallyResolvedPolymorphicReceiverThenUpdateSelectorIn: node.
					 (aCodeGen isBuiltinSelector: node selector)
						ifTrue:
							[node isBuiltinOperator: true.
							"If a to:by:do:'s limit has side-effects, declare the limit variable, otherwise delete it from the args"
							 node selector = #to:by:do: ifTrue:
								[self ensureToByDoLoopLimitIsSafeAndEfficient: node in: aCodeGen]]
						ifFalse:
							[(StackInterpreter isStackAccessor: node selector)
								ifTrue: "compute and cache the accessor depth early, before inlining destroys the accessor chains"
									[self export ifTrue:
										[aCodeGen accessorDepthForMethod: self]]
								ifFalse:
									[(CaseStatements includes: node selector) ifTrue:
										[parent replaceNodesIn: (Dictionary newFromPairs: { node. self buildCaseStmt: node in: aCodeGen})].
									 (#(caseOf: #caseOf:otherwise:) includes: node selector) ifTrue:
										[parent replaceNodesIn: (Dictionary newFromPairs: { node. self buildSwitchStmt: node parent: parent })].
									 (#(printf: f:printf:) includes: node selector) ifTrue:
										[self transformPrintf: node]]]]]]! !

!TMethod methodsFor: 'primitive compilation' stamp: 'eem 3/6/2015 13:59'!
preparePrimitiveName
	"Prepare the selector for this method in translation.
	 Remember the original selector in properties."
	| aClass |
	properties := properties copy.
	properties selector: selector.
	aClass := definingClass.
	primitive = 117 
		ifTrue:[selector := ((aClass includesSelector: selector)
					ifTrue: [aClass compiledMethodAt: selector]
					ifFalse: [aClass class compiledMethodAt: selector]) literals first at: 2.
				export := true]
		ifFalse:[selector := 'prim', aClass name, selector].

! !

!TMethod methodsFor: 'primitive compilation' stamp: 'eem 1/20/2021 19:10'!
preparePrimitivePrologue
	"Add a prolog and postlog to a primitive method. The prolog copies any instance variables referenced by this primitive method into local variables. The postlog copies values of assigned-to variables back into the instance. The names of the new locals are added to the local variables list.

The declarations dictionary defines the types of any non-integer variables (locals, arguments, or instance variables). In particular, it may specify the types:

	int *		-- an array of 32-bit values (e.g., a BitMap)
	short *		-- an array of 16-bit values (e.g., a SoundBuffer)
	char *		-- an array of unsigned bytes (e.g., a String)
	double		-- a double precision floating point number (e.g., 3.14159)

Undeclared variables are taken to be integers and will be converted from Smalltalk to C ints."

"Current restrictions:
	o method must not contain message sends
	o method must not allocate objects
	o method must not manipulate raw oops
	o method cannot access class variables
	o method can only return an integer"

	| prolog postlog instVarsUsed varsAssignedTo instVarList primArgCount varName endsWithReturn aClass |
	self assert: selector ~~ #setInterpreter:.
	aClass := definingClass.
	prolog := OrderedCollection new.
	postlog := OrderedCollection new.
	instVarsUsed := self freeVariableReferences asSet.
	varsAssignedTo := self variablesAssignedTo asSet.
	instVarList := aClass allInstVarNames.
	primArgCount := args size.

	"add receiver fetch and arg conversions to prolog"
	prolog addAll: self fetchRcvrExpr.
	1 to: args size do: [:argIndex |
		varName := args at: argIndex.
		prolog addAll:
			(self argConversionExprFor: varName stackIndex: args size - argIndex)].

	"add success check to postlog"
	postlog addAll: self checkSuccessExpr.

	"add instance variable fetches to prolog and instance variable stores to postlog"
	1 to: instVarList size do: [:varIndex |
		varName := instVarList at: varIndex.
		(instVarsUsed includes: varName) ifTrue: [
			locals add: varName.
			prolog addAll: (self instVarGetExprFor: varName offset: varIndex - 1).
			(varsAssignedTo includes: varName) ifTrue: [
				postlog addAll: (self instVarPutExprFor: varName offset: varIndex - 1)]]].
	prolog addAll: self checkSuccessExpr.

	((locals includes: 'rcvr') or: [(locals intersection: args) notEmpty]) ifTrue:
		[self error: 'local name conflicts with instance variable name'].
	locals add: 'rcvr'; addAll: args.
	args := args class new.
	endsWithReturn := self endsWithReturn.
	self fixUpReturns: primArgCount postlog: postlog.

	parseTree statements addAllFirst: prolog.
	endsWithReturn ifFalse:
		[postlog addAll: (self popArgsExpr: primArgCount).
		 parseTree statements addAllLast: postlog].
! !

!TMethod methodsFor: 'accessing' stamp: 'jm 2/12/98 11:56'!
primitive
	"The primitive number of this method; zero if not a primitive."

	^ primitive
! !

!TMethod methodsFor: 'printing' stamp: 'eem 2/18/2011 15:38'!
printOn: aStream

	super printOn: aStream.
	aStream nextPutAll: ' (', definingClass name, '>>', selector, ')'.! !

!TMethod methodsFor: 'C code generation' stamp: 'eem 5/21/2010 17:03'!
printSingleComment: aString on: aStream indent: indent tabWidth: tabWidth lineBreak: lineBreak
	"Shameless hack of ParseNode>>printSingleComment:on:indent: for formatting comments."
	| readStream position wordStream |
	readStream := ReadStream on: aString.
	position := indent * tabWidth.
	wordStream := WriteStream on: (String new: 16).
	[readStream atEnd] whileFalse:
		[| word wordWidth lastChar |
		wordStream reset.
		wordWidth := 0.
		[(readStream peekFor: Character space) 
		 or: [readStream peekFor: Character tab]] whileTrue.
		[readStream atEnd
		 or: [(lastChar := readStream next) = Character cr
			 or: [lastChar = Character space]]] whileFalse:
			[wordWidth := wordWidth + 1.
			 wordStream nextPut: lastChar].
		word := wordStream contents.
		position := position + wordWidth.
		position > lineBreak
			ifTrue:
				[aStream skip: -1; crtab: indent.
				position := indent * tabWidth + wordWidth + 1.
				lastChar = Character cr ifTrue:
					[[readStream peekFor: Character tab] whileTrue].
				word isEmpty ifFalse:
					[aStream nextPutAll: word; space]]
			ifFalse:
				[aStream nextPutAll: word.
				readStream atEnd ifFalse:
					[position := position + 1.
					aStream space].
				lastChar = Character cr ifTrue:
					[aStream skip: -1; crtab: indent.
					position := indent * tabWidth.
					[readStream peekFor: Character tab] whileTrue]]]! !

!TMethod methodsFor: 'private' stamp: 'eem 2/17/2015 12:15'!
printTempsAndVar: varName on: aStream 
	"add the required temps and the varname to the stream"
	aStream nextPutAll: '| rcvr stackPointer interpreterProxy ' , varName , ' |'; cr! !

!TMethod methodsFor: 'inlining support' stamp: 'eem 12/8/2017 09:19'!
propagateReturnIn: aCodeGen
	"Propagate the return type to all return nodes"
	| map coercionType |
	returnType = #void ifTrue:
		[^self].
	"The following is necessary for functions returning functions, which have problematic syntax"
	coercionType := aCodeGen
							extractTypeFor: (aCodeGen cFunctionNameFor: self selector)
							fromDeclaration: returnType.
	map := IdentityDictionary new.
	parseTree nodesDo:[:node|
		(node isReturn
		 and: [(aCodeGen typeFor: node expression in: self) ~= coercionType
		 and: [(aCodeGen isNode: node constantValueWithinRangeOfType: coercionType) not]]) ifTrue:
			[map at: node expression put: (aCodeGen nodeToCast: node expression to: coercionType)]].
	self replaceNodesIn: map! !

!TMethod methodsFor: 'accessing' stamp: 'eem 9/13/2010 12:10'!
properties
	^properties! !

!TMethod methodsFor: 'initialization' stamp: 'eem 12/15/2008 09:15'!
properties: anAdditionalMethodState
	properties := anAdditionalMethodState! !

!TMethod methodsFor: 'accessing' stamp: 'eem 1/31/2017 09:44'!
readsVariable: variableName
	"Answer if the receiver reads the variable (i.e. ignore assignments to the variable)."
	parseTree nodesWithParentsDo:
		[:node :parent|
		 (node isVariable
		  and: [node name = variableName]) ifTrue:
			[(parent notNil
			  and: [parent isAssignment
			  and: [node == parent variable]]) ifFalse:
				[^true]]].
	^false! !

!TMethod methodsFor: 'transformations' stamp: 'eem 10/29/2018 19:01'!
recordDeclarationsIn: aCCodeGen
	"Record C type declarations of the forms
		<returnTypeC: 'float'>
		<var: #foo declareC: 'float foo'>
		<var: #foo type:'float'>
	 or the older, obsolete
		self returnTypeC: 'float'.
		self var: #foo declareC: 'float foo'
		self var: #foo type:'float'.
	 and remove the declarations from the method body."

	| newStatements |
	properties pragmas notEmpty ifTrue:
		[properties pragmas do:
			[:pragma|
			pragma keyword == #var:declareC: ifTrue:
				[self checkedDeclarationAt: pragma arguments first asString
					put: pragma arguments last
					in: aCCodeGen].
			pragma keyword == #var:type: ifTrue:
				[| varName varType |
				varName := pragma arguments first asString.
				varType := aCCodeGen conventionalTypeForType: pragma arguments last.
				varType last == $* ifFalse: [varType := varType, ' '].
				self checkedDeclarationAt: varName
					put: varType, varName
					in: aCCodeGen].
			pragma keyword = #returnTypeC: ifTrue:
				[self returnType: pragma arguments last].
			pragma keyword = #doNotGenerate: ifTrue:
				[locals remove: pragma arguments last]].
		^self].
	newStatements := OrderedCollection new: parseTree statements size.
	parseTree statements do:
		[ :stmt | | isDeclaration |
		isDeclaration := false.
		stmt isSend ifTrue:
			[stmt selector == #var:declareC: ifTrue:
				[isDeclaration := true.
				self declarationAt: stmt args first value asString put: stmt args last value].
			stmt selector == #var:type: ifTrue:
				[| varName varType |
				isDeclaration := true.
				varName := stmt args first value asString.
				varType := aCCodeGen conventionalTypeForType: stmt args last value.
				varType last == $* ifFalse: [varType := varType, ' '].
				self declarationAt: varName put: varType, varName].
			stmt selector = #returnTypeC: ifTrue:
				[isDeclaration := true.
				returnType := stmt args last value]].
		isDeclaration ifFalse:
			[newStatements add: stmt]].
	parseTree setStatements: newStatements asArray.! !

!TMethod methodsFor: 'accessing' stamp: 'tpr 5/14/2013 16:45'!
referencesGlobalStruct
	globalStructureBuildMethodHasFoo := true! !

!TMethod methodsFor: 'accessing' stamp: 'tpr 5/14/2013 16:49'!
refersToGlobalStruct
	^globalStructureBuildMethodHasFoo! !

!TMethod methodsFor: 'transformations' stamp: 'ikp 9/26/97 14:50'!
removeAssertions
	parseTree removeAssertions! !

!TMethod methodsFor: 'transformations' stamp: 'eem 7/9/2021 19:31'!
removeFinalSelfReturnIn: aCodeGenOrNil
	"The Smalltalk parser automatically adds the statement '^self' to the end of methods
	 without explicit returns.  This method removes such statements, since in most VMMaker
	 classes (except struct classes) the generated code has no notion of 'self' anyway.
	 If the statement is removed and no return type has yet been specified and the class
	 specifies a default return type (e.g. #void) for methods that don't return, then set the
	 return type accordingly."

	| removeLastStatement |
	parseTree statements isEmpty ifTrue: [^self].
	removeLastStatement := false.
	(parseTree statements size > 1
	 and: [(parseTree statements at: parseTree statements size - 1) isReturningIf])
		ifTrue: [removeLastStatement := true]
		ifFalse:
			[| nameOfSelf lastStmt  |
			 nameOfSelf := (definingClass isVMClass and: [definingClass isStructClass])
								ifTrue: ['self_in_', ((aCodeGenOrNil ifNil: [CCodeGenerator basicNew]) cFunctionNameFor: selector)]
								ifFalse: ['self'].
			 ((lastStmt := parseTree statements last) isReturn
			  and: [lastStmt expression isVariable
			  and: [nameOfSelf = lastStmt expression name]]) ifTrue:
				[| tokens |
				 tokens := Scanner new scanTokens: (definingClass sourceCodeAt: selector ifAbsent: ['']).
				 (tokens size < 2 or: [tokens last ~= nameOfSelf or: [(tokens at: tokens size - 1) ~= #'^']]) ifTrue:
					[removeLastStatement := true]]].

	removeLastStatement ifTrue:
		[parseTree setStatements: parseTree statements allButLast.
		 (returnType isNil
		  and: [aCodeGenOrNil notNil
		  and: [parseTree noneSatisfy: [:node| node isReturn and: [node expression isVariable not or: [node expression name ~= 'self']]]]]) ifTrue:
			[self returnType: (aCodeGenOrNil implicitReturnTypeFor: selector)]]! !

!TMethod methodsFor: 'utilities' stamp: 'eem 8/23/2018 11:03'!
removeUnusedTempsAndNilIfRequiredIn: aCodeGen
	"Remove all of the unused temps in this method. Answer a set of the references.
	 As a side-effect introduce explicit temp := nil statements for temps that are
	 tested for nil before necessarily being assigned."
	| refs readBeforeAssigned simplyTypedLocals |
	refs := self removeUnusedTempsIn: aCodeGen.
	"reset the locals to be only those still referred to"
	locals := locals select: [:e| refs includes: e].
	(locals notEmpty
	 and: [aCodeGen
			pushScope: declarations
			while: [simplyTypedLocals := locals select:
											[:var|
											 declarations
												at: var
												ifPresent: [:decl| aCodeGen isSimpleType: (aCodeGen extractTypeFor: var fromDeclaration: decl)]
												ifAbsent: [true]].
				(readBeforeAssigned := (self findReadBeforeAssignedIn: simplyTypedLocals in: aCodeGen)) notEmpty]]) ifTrue:
		[readBeforeAssigned := readBeforeAssigned reject:
			[:v| | d | "don't initialize externs, statics, arrays or the explicitly initialized."
			 d := self declarationAt: v.
			 (d beginsWith: 'extern') or: [(d beginsWith: 'static') or: [(d includes: $[) or: [d includes: $=]]]].
		 parseTree statements addAllFirst:
			(readBeforeAssigned asSortedCollection collect:
				[:var| | varNode varType zeroNode |
				 varNode := TVariableNode new setName: var; yourself.
				 varType := aCodeGen typeFor: varNode in: self.
				 zeroNode := TConstantNode new setValue: 0; yourself.
				 TAssignmentNode new
					setVariable: varNode
					expression: (((aCodeGen isIntegralCType: varType)
								    or: [aCodeGen isFloatingPointCType: varType])
									ifTrue: [zeroNode]
									ifFalse: [aCodeGen nodeToCast: zeroNode to: varType])])].
	^refs! !

!TMethod methodsFor: 'utilities' stamp: 'eem 5/30/2018 19:47'!
removeUnusedTempsIn: aCodeGen
	"Remove all of the unused temps in this method. Answer a set of the references."
	"After inlining some variable references are now obsolete, we could fix them there
	 but the code seems a bit complicated, the other choice to to rebuild the locals
	 before extruding. This is done here"
	| usedVariables |
	usedVariables := self allReferencedVariablesUsing: aCodeGen.
	"reset the locals to be only those still referred to"
	locals do:
		[:local|
		 (usedVariables includes: local) ifFalse:
			[(((declarations at: local ifAbsent: ['']) includesSubstring: 'static')
			  or: [(declarations at: local ifAbsent: ['']) includesSubstring: 'extern'])
				ifFalse:
					[locals remove: local.
					 declarations removeKey: local ifAbsent: []]
				ifTrue:
					[usedVariables add: local "In case this is a function declaration, e.g. amInVMThread in ownVM:"]]].
	^usedVariables! !

!TMethod methodsFor: 'inlining support' stamp: 'jm 12/13/1998 10:06'!
renameLabelsForInliningInto: destMethod
	"Rename any labels that would clash with those of the destination method."

	| destLabels usedLabels labelMap newLabelName |
	destLabels := destMethod labels asSet.
	usedLabels := destLabels copy.  "usedLabels keeps track of labels in use"
	usedLabels addAll: labels.
	labelMap := Dictionary new: 100.
	self labels do: [ :l |
		(destLabels includes: l) ifTrue: [
			newLabelName := self unusedNamePrefixedBy: 'l' avoiding: usedLabels.
			labelMap at: l put: newLabelName.
		].
	].
	self renameLabelsUsing: labelMap.! !

!TMethod methodsFor: 'inlining support' stamp: 'eem 1/20/2016 16:35'!
renameLabelsUsing: aDictionary
	"Rename all labels according to the old->new mappings of the given dictionary."

	aDictionary isEmpty ifTrue:
		[^self].
	labels := labels collect: [ :label | aDictionary at: label ifAbsent: [label]].

	parseTree nodesDo:
		[ :node |
		(node isGoTo and: [aDictionary includesKey: node label]) ifTrue:
			[node setLabel: (aDictionary at: node label)].
		(node isLabel and: [aDictionary includesKey: node label]) ifTrue:
			[node setLabel: (aDictionary at: node label)]]! !

!TMethod methodsFor: 'inlining support' stamp: 'eem 1/20/2016 16:37'!
renameVariablesUsing: aDictionary
	"Rename all variables according to old->new mappings of the given dictionary."

	| newDecls newProperties |
	aDictionary isEmpty ifTrue: [^self].

	"map args and locals"
	args := args collect: [ :arg | aDictionary at: arg ifAbsent: [ arg ]].
	locals := locals collect: [ :v | aDictionary at: v ifAbsent: [ v ]].

	"map declarations"
	newDecls := declarations species new.
	declarations keysAndValuesDo:
		[:oldName :decl|
		(aDictionary at: oldName ifAbsent: nil)
			ifNotNil:
				[:newName| | index |
				index := decl indexOfWord: oldName.
				 newDecls
					at: newName
					put: (index ~= 0
							ifTrue: [decl copyReplaceFrom: index to: index + oldName size - 1 with: newName]
							ifFalse: [decl])]
			ifNil: [newDecls at: oldName put: decl]].
	self newDeclarations: newDecls.

	newProperties := properties copy.
	newProperties pragmas do:
		[:pragma| | mappedArgs |
		mappedArgs := pragma arguments collect: [:arg| arg isString ifTrue: [aDictionary at: arg ifAbsent: arg] ifFalse: [arg]].
		mappedArgs ~= pragma arguments ifTrue:
			[pragma setArguments: mappedArgs]].
	self properties: newProperties.

	"map variable names in parse tree"
	parseTree nodesDo:
		[ :node |
		(node isVariable
		and: [aDictionary includesKey: node name]) ifTrue:
			[node setName: (aDictionary at: node name)].
		(node isStmtList and: [node args size > 0]) ifTrue:
			[node setArguments: (node args collect: [ :arg | aDictionary at: arg ifAbsent: [ arg ]])]]! !

!TMethod methodsFor: 'inlining support' stamp: 'eem 3/30/2013 11:57'!
renameVarsForInliningInto: destMethod except: doNotRename in: aCodeGen
	"Rename any variables that would clash with those of the destination method."

	| destVars usedVars varMap newVarName |
	destVars := aCodeGen globalsAsSet copy.
	destVars addAll: destMethod locals.
	destVars addAll: destMethod args.
	usedVars := destVars copy.  "keeps track of names in use"
	usedVars addAll: args; addAll: locals.
	varMap := Dictionary new: 100.
	locals, args do:
		[ :v |
		((doNotRename includes: v) not
		  and: [destVars includes: v]) ifTrue:
			[newVarName := self unusedNamePrefixedBy: v avoiding: usedVars.
			varMap at: v put: newVarName]].
	self renameVariablesUsing: varMap! !

!TMethod methodsFor: 'transformations' stamp: 'ar 7/6/2003 21:31'!
replaceNodesIn: map
	parseTree := parseTree replaceNodesIn: map.! !

!TMethod methodsFor: 'primitive compilation' stamp: 'TPR 3/2/2000 19:08'!
replaceSizeMessages
	"Replace sends of the message 'size' with calls to sizeOfSTArrayFromCPrimitive."

	| argExpr |
	parseTree nodesDo: [:n |
		(n isSend and: [n selector = #size]) ifTrue: [
			argExpr := TSendNode new
				setSelector: #+
				receiver: n receiver
				arguments: (Array with: (TConstantNode new setValue: 1)).
			n
				setSelector: #sizeOfSTArrayFromCPrimitive:
				receiver: (TVariableNode new setName: self vmNameString)
				arguments: (Array with: argExpr)]].
! !

!TMethod methodsFor: 'accessing' stamp: ''!
returnType
	"The type of the values returned by this method. This string will be used in the C declaration of this function."

	^returnType! !

!TMethod methodsFor: 'accessing' stamp: 'eem 6/8/2016 09:02'!
returnType: aString
	"Set the type of the values returned by this method.
	 This string will be used in the C declaration of this function.
	 If the type exists as a symbol, use that."

	returnType := aString isSymbol
					ifTrue: [aString]
					ifFalse: [(Symbol findInterned: aString) ifNil: [aString]]! !

!TMethod methodsFor: 'testing' stamp: 'eem 6/7/2016 17:50'!
returnsExpression
	"Answer true if the last statement of this method is a return of some expression, not merely self or nil."

	^parseTree returnsExpression! !

!TMethod methodsFor: 'accessing' stamp: ''!
selector
	"The Smalltalk selector of this method."

	^selector! !

!TMethod methodsFor: 'accessing' stamp: ''!
selector: newSelector

	selector := newSelector.! !

!TMethod methodsFor: 'initialization' stamp: 'eem 11/15/2019 14:23'!
setSelector: sel definingClass: class args: argList locals: localList block: aBlockNode primitive: aNumber properties: methodProperties comment: aComment
	"Initialize this method using the given information."

	selector := sel.
	definingClass := class.
	args := argList asOrderedCollection collect: [:arg | arg key].
	locals := (localList collect: [:arg | arg key]) asSet.
	declarations := Dictionary new.
	self addTypeForSelf.
	primitive := aNumber.
	properties := methodProperties.
	comment := aComment.
	labels := Set new.
	parseTree := aBlockNode. "hack; allows nodes to find their parent, etc"
	parseTree := aBlockNode asRootTranslatorNodeIn: self.
	complete := false.  "set to true when all possible inlining has been done"
	export := self extractExportDirective.
	static := self extractStaticDirective.
	self extractSharedCase.
	globalStructureBuildMethodHasFoo := false! !

!TMethod methodsFor: 'accessing' stamp: 'ar 7/6/2003 21:18'!
sharedCase
	^sharedCase! !

!TMethod methodsFor: 'accessing' stamp: 'ar 7/6/2003 21:18'!
sharedLabel
	^sharedLabel! !

!TMethod methodsFor: 'inlining' stamp: 'eem 9/24/2010 09:46'!
shouldIncorporatePragmaFromSuperMethod: aPragma
	(properties includesKey: aPragma keyword) ifFalse:
		[^true].
	((aPragma keyword beginsWith: #var:)
	  and: [properties pragmas noneSatisfy:
			[:p|
			(p keyword beginsWith: #var:)
			and: [(p argumentAt: 1) = (aPragma argumentAt: 1)]]]) ifTrue:
		[^true].
	^false! !

!TMethod methodsFor: 'accessing' stamp: 'eem 9/15/2015 11:29'!
smalltalkSelector
	"Answer the selector of the original Smalltalk method, not any mangled one."
	^selector! !

!TMethod methodsFor: 'accessing' stamp: ''!
statements

	parseTree isStmtList
		ifFalse: [ self error: 'expected method parse tree to be a TStmtListNode' ].
	((parseTree args = nil) or: [parseTree args isEmpty])
		ifFalse: [ self error: 'expected method parse tree to have no args' ].

	^parseTree statements! !

!TMethod methodsFor: 'primitive compilation' stamp: 'eem 12/17/2018 21:09'!
statementsFor: sourceText varName: varName
	"Return the parse tree for the given expression. The result is the statements list of the method parsed from the given source text."
	"Details: Various variables are declared as locals to avoid Undeclared warnings from the parser."

	| s |
	s := WriteStream on: String new.
	s nextPutAll: 'temp'; cr; crtab.
	self printTempsAndVar: varName on: s.
	s nextPutAll: sourceText.
	^(([ | compiler |
		compiler := Smalltalk compiler class: VMBasicConstants. "for primitive error codes"
		(compiler parse: s contents)
			compilationContext: compiler compilationContext;
			yourself] "Pharo"
		on: MessageNotUnderstood
		do: [:ex|
			ex message selector == #compiler ifFalse:
				[ex pass].
			Compiler new parse: s contents in: VMBasicConstants notifying: nil]) "Squeak"
		asTranslationMethodOfClass: self class)
			removeFinalSelfReturnIn: nil;
			statements! !

!TMethod methodsFor: 'inlining' stamp: 'eem 7/25/2020 09:57'!
statementsListsForInliningIn: aCodeGen
	"Answer a collection of statement list nodes that are candidates for inlining.
	 Currently, we cannot inline into the argument blocks of and: and or: messages.
	 We do not want to inline code strings within cCode:inSmalltalk: blocks (those with a
	 proper block for the cCode: argument are inlined in MessageNode>>asTranslatorNodeIn:).
	 We do not want to inline code within assert: sends (because we want the assert to read nicely)."

	| stmtLists |
	stmtLists := OrderedCollection new: 10.
	parseTree
		nodesDo:
			[:node|
			node isStmtList ifTrue: [stmtLists add: node]]
		unless:
			[:node|
			node isSend
			and: [node selector == #cCode:inSmalltalk:
				or: [aCodeGen isAssertSelector: node selector]]].
	parseTree nodesDo:
		[:node|
		node isSend ifTrue:
			[node selector = #cCode:inSmalltalk: ifTrue:
				[node nodesDo:
					[:ccisNode| stmtLists remove: ccisNode ifAbsent: []]].
			 (node selector beginsWith: #cppIf:ifTrue:) ifTrue:
				[node args first nodesDo:
					[:inCondNode| stmtLists remove: inCondNode ifAbsent: []]].
			((node selector = #and:) or: [node selector = #or:]) ifTrue:
				"Note: the PP 2.3 compiler produces two arg nodes for these selectors"
				[stmtLists remove: node args first ifAbsent: [].
				stmtLists remove: node args last ifAbsent: []].
			(#(	#ifTrue: #ifFalse: #ifTrue:ifFalse: #ifFalse:ifTrue:
				#ifNil: #ifNotNil: #ifNil:ifNotNil: #ifNotNil:ifNil: ) includes: node selector) ifTrue:
				[stmtLists remove: node receiver ifAbsent: []].
			(#(whileTrue whileTrue: whilefalse whileFalse:) includes: node selector) ifTrue:
				"Allow inlining if it is a [...] whileTrue/whileFalse.
				This is identified by having more than one statement in the 
				receiver block in which case the C code wouldn't work anyways"
				[node receiver statements size = 1 ifTrue:
					[stmtLists remove: node receiver ifAbsent: []]].
			(node selector = #to:do:) ifTrue:
				[stmtLists remove: node receiver ifAbsent: [].
				stmtLists remove: node args first ifAbsent: []].
			(node selector = #to:by:do:) ifTrue:
				[stmtLists remove: node receiver ifAbsent: [].
				stmtLists remove: node args first ifAbsent: [].
				stmtLists remove: node args second ifAbsent: []]].
		node isCaseStmt ifTrue: "don't inline cases"
			[node cases do: [:case| stmtLists remove: case ifAbsent: []]]].
	^stmtLists! !

!TMethod methodsFor: 'accessing' stamp: 'eem 2/10/2009 13:52'!
static
	^static! !

!TMethod methodsFor: 'accessing' stamp: 'eem 2/10/2009 13:53'!
static: aBoolean
	static := aBoolean! !

!TMethod methodsFor: 'inlining' stamp: 'eem 8/15/2018 18:32'!
superExpansionNodeFor: aSelector args: argumentNodes
	"Answer the expansion of a super send.  Merge the super expansion's
	 locals, properties and comment into this method's properties."
	(definingClass superclass lookupSelector: aSelector)
		ifNil: [self error: 'superclass does not define super method']
		ifNotNil:
			[:superMethod| | superTMethod commonVars varMap |
			superTMethod := superMethod asTranslationMethodOfClass: self class.
			((argumentNodes allSatisfy: [:parseNode| parseNode isVariableNode])
			and: [(argumentNodes asOrderedCollection collect: [:parseNode| parseNode key]) = superTMethod args]) ifFalse:
				[self error: definingClass name, '>>',selector, ' args ~= ',
							superTMethod definingClass name, '>>', aSelector,
							(String with: $. with: Character cr),
							'For super expansions to be translated correctly each argument must be a variable with the same name as the corresponding argument in the super method.'].
			(commonVars := superTMethod locals intersection: self locals) notEmpty ifTrue:
				[varMap := Dictionary new.
				 commonVars do:
					[:k| varMap at: k put: (superTMethod unusedNamePrefixedBy: k avoiding: self allLocals)].
				 superTMethod renameVariablesUsing: varMap].
			self mergePropertiesOfSuperMethod: superTMethod.
			self assert: (superTMethod locals allSatisfy: [:var| (self locals includes: var) not]).
			locals addAll: superTMethod locals.
			superTMethod declarations keysAndValuesDo:
				[:var :decl|
				self declarationAt: var put: decl].
			superTMethod comment ifNotNil:
				[:superComment|
				comment := comment
								ifNil: [superComment]
								ifNotNil: [superComment, comment]].
			superTMethod extraVariableNumber ifNotNil:
				[:scvn|
				extraVariableNumber := extraVariableNumber ifNil: [scvn] ifNotNil: [:cvn| cvn + scvn]].
			superTMethod elideAnyFinalReturn.
			^superTMethod parseTree]! !

!TMethod methodsFor: 'C code generation' stamp: 'eem 2/12/2016 16:56'!
terminateConditionalDefineFor: compileTimeOptionPragmas on: aStream
	compileTimeOptionPragmas ifEmpty: [^self].
	aStream nextPutAll: '#endif /* '.
	compileTimeOptionPragmas
		do: [:pragma|
			pragma keyword = #notOption: ifTrue:
				[aStream nextPut: $!!].
			aStream nextPutAll: (pragma argumentAt: 1)]
		separatedBy: [aStream nextPutAll: ' && '].
	 aStream nextPutAll: ' */'; cr! !

!TMethod methodsFor: 'inlining' stamp: 'eem 2/17/2015 13:44'!
transformConditionalAssignment: node in: aCodeGen
	"If possible answer the transformation of code of the form
		var := e1
				ifTrue: [e2 ifTrue: [self m1] ifFalse: [self m2]]
				ifFalse: [self m3]
	 into
		e1
			ifTrue: [e2 ifTrue: [var := self m1] ifFalse: [var := self m2]]
			ifFalse: [var := self m3]
	 to allow inlining of m1, m2, et al.  Otherwise answer nil."

	| expr |
	^(node isAssignment
	   and: [(expr := node expression) isSend
	   and: [(#(ifTrue:ifFalse: ifFalse:ifTrue:) includes: expr selector)
	   and: [self isConditionalToBeTransformedForAssignment: expr  in: aCodeGen]]]) ifTrue:
		[expr copy
			arguments:
				(expr args collect:
					[:stmtList| stmtList copy assignLastExpressionTo: node variable]);
			yourself]! !

!TMethod methodsFor: 'transformations' stamp: 'eem 8/21/2021 20:03'!
transformPrintf: sendNode
	"Handle forms of f:printf: & printf:. f:printf: is either
		logFile f: formatLiteral printf: args
	 or
		formatLiteral f: streamName printf: args
	 printf is
		formatLiteral printf: args"
	| map newArgs |
	newArgs := OrderedCollection new.
	sendNode args do:
		[:arg|
		arg isBrace
			ifTrue: [newArgs addAllLast: arg elements]
			ifFalse: [newArgs addLast: arg]].
	(sendNode receiver isConstant
	 and: [sendNode receiver value isString]) ifTrue:
		[| format |
 		 format := sendNode receiver asPrintfFormatStringNode.
		 sendNode selector first == $f ifTrue: "fprintf et al..."
			[sendNode receiver: (TVariableNode new setName: 'self').
			 sendNode arguments: {sendNode args first. format}, newArgs allButFirst.
			 ^sendNode].
		sendNode receiver: format.
		sendNode arguments: newArgs.
		^sendNode].
	map := Dictionary new.
	sendNode nodesDo:
		[:subNode|
		 (subNode isConstant and: [subNode value isString and: [subNode value includes: $%]]) ifTrue:
			[map at: subNode put: subNode asPrintfFormatStringNode]].
	sendNode arguments: newArgs.
	sendNode replaceNodesIn: map.
	^sendNode! !

!TMethod methodsFor: 'inlining' stamp: 'eem 4/27/2016 19:22'!
transformReturnSubExpression: node toAssignmentOf: exitVar andGoto: exitLabel unless: eliminateReturnSelfs into: aBinaryBlock
	| expr replacement |
	expr := node isReturn ifTrue: [node expression] ifFalse: [node].
	replacement := (expr isVariable "Eliminate ^self's"
					   and: [expr name = 'self'
					   and: [eliminateReturnSelfs]])
						ifTrue: [nil]
						ifFalse:
							[exitVar
								ifNil: [expr]
								ifNotNil: [TAssignmentNode new
											setVariable: (TVariableNode new setName: exitVar)
											expression: expr]].
	 node == parseTree statements last
		ifTrue:
			[aBinaryBlock value: replacement value: false]
		ifFalse:
			[replacement := replacement
								ifNil: [TGoToNode new setLabel: exitLabel; yourself]
								ifNotNil:
									[TStmtListNode new
										setArguments: #()
										statements: {replacement.
													  TGoToNode new setLabel: exitLabel; yourself};
										yourself].
			 aBinaryBlock value: replacement value: true]! !

!TMethod methodsFor: 'type inference' stamp: 'eem 6/8/2016 09:26'!
transformReturns
	"Once the return type has been found or inferred, returns may bneed to be modified.
	 If the return type is #void, any occurrences of ^expr must be replaced with expr. ^self.
	 If the type is #sqInt any any occurrences of ^self are replaced with ^0."
	(returnType == #void or: [returnType == #sqInt]) ifFalse:
		[^self].
	parseTree nodesWithParentsDo:
		[:node :parent|
		node isReturn ifTrue:
			[(node expression isVariable and: [node expression name = 'self'])
				ifTrue:
					[returnType = #sqInt ifTrue:
						[node setExpression: (TConstantNode new setValue: 0)]]
				ifFalse:
					[returnType = #void ifTrue:
						[parent
							replaceChild: node
							with: (TStmtListNode new
									setArguments: #()
									statements: {node expression.
												  TReturnNode new 
													setExpression: (TVariableNode new setName: 'self')
													yourself})]]]]! !

!TMethod methodsFor: 'transformations' stamp: 'eem 3/13/2017 15:42'!
transformToStructClassMethodFor: aCCodeGenerator
	"Transform this method so that it can be used on an instance of a struct class (VMStructType subclass).
	 Convert inst var refs into field dereferences of self.  Add selfSelector as the first argument with the
	 right struct type. As a complete hack to avoid breaking the inlinert don't use 'self' as the name for self
	 as this causes serious type redefinitions ``somewhere'' in the inliner."
	| replacements selfNode typeForSelf |
	self isStructAccessor ifTrue:
		[^self returnType: (definingClass returnTypeForAccessor: selector)].
	replacements := IdentityDictionary new.
	selfNode := TVariableNode new setName: 'self_in_', (aCCodeGenerator cFunctionNameFor: selector).
	args do:
		[:var|
		(definingClass isAccessor: var) ifTrue:
			[self error: 'In ', definingClass name, '>>', selector, ' ', var, ' arg shadows struct field and will break during translation!!']].
	parseTree nodesDo:
		[:node|
		node isVariable ifTrue:
			[node name = 'self' ifTrue:
				[replacements at: node put: selfNode copy].
			 (definingClass isAccessor: node name) ifTrue:
				[replacements
					at: node
					put: (TSendNode new
							setSelector: node name asSymbol
							receiver: selfNode
							arguments: #())]]].
	replacements notEmpty ifTrue:
		[self replaceNodesIn: replacements].
	typeForSelf := self typeForSelf.
	self assert: (typeForSelf notNil and: [typeForSelf ~~ #implicit]).
	self declarationAt: (args addFirst: selfNode name)
		put: (declarations removeKey: 'self'), '_in_', (aCCodeGenerator cFunctionNameFor: selector)! !

!TMethod methodsFor: 'inlining' stamp: 'eem 1/5/2018 00:17'!
tryToInlineMethodExpressionsIn: aCodeGen
	"Expand any (complete) inline methods sent by this method as receivers or parameters.
	 Answer if anything was inlined."

	| sendsToInline |
	sendsToInline := Dictionary new: 100.
	aCodeGen
		pushScope: declarations
		while: [parseTree
					nodesDo:
						[:node|
						(self inlineableFunctionCall: node in: aCodeGen) ifTrue:
							[(self inlineFunctionCall: node in: aCodeGen) ifNotNil:
								[:replacement|
								 (replacement isConstant
								  and: [replacement isDefine not
								  and: [replacement value isNumber
								  and: [replacement comment isNil]]]) ifTrue:
									[replacement comment: node selector].
								 sendsToInline at: node put: replacement]]]
					unless: "Don't inline the arguments to asserts to keep the asserts readable"
						[:node|
						node isSend
						and: [node selector == #cCode:inSmalltalk:
							or: [aCodeGen isAssertSelector: node selector]]]].

	sendsToInline isEmpty ifTrue:
		[^false].
	self replaceNodesIn: sendsToInline.
	^true! !

!TMethod methodsFor: 'inlining' stamp: 'eem 12/15/2016 10:15'!
tryToInlineMethodStatementsIn: aCodeGen statementListsInto: aBlock
	"Expand any (complete) inline methods sent by this method as top-level statements.
	 Answer if anything was inlined."

	| stmtLists didSomething newStatements returningNodes |
	didSomething := false.
	returningNodes := Set new.
	parseTree nodesDo:
		[:node|
		node isReturn ifTrue:
			[returningNodes add: node expression.
			 node expression isConditionalSend ifTrue:
				[returningNodes addAll: (node expression args collect: [:stmtList| stmtList statements last])]]].
	stmtLists := self statementsListsForInliningIn: aCodeGen.
	stmtLists do:
		[:stmtList|
		newStatements := OrderedCollection new: stmtList statements size.
		stmtList statements do:
			[:stmt|
			(self inlineCodeOrNilForStatement: stmt returningNodes: returningNodes in: aCodeGen)
				ifNil: [newStatements addLast: stmt]
				ifNotNil: [:inlinedStmts|
					didSomething := true.
					newStatements addAllLast: inlinedStmts]].
		stmtList setStatements: newStatements asArray].

	"This is a hack; forgive me. The inlining above tends to keep return statements in statement lists.
	 In the case of returning ifs we don't want the returns in case the returning if is generated as an expression."
	returningNodes do:
		[:returningNode|
		 (returningNode isConditionalSend
		  and: [returningNode args anySatisfy: [:alternativeNode| alternativeNode endsWithReturn]]) ifTrue:
			[returningNode args withIndexDo:
				[:alternativeNode :index|
				 alternativeNode endsWithReturn ifTrue:
					[returningNode args at: index put: alternativeNode copyWithoutReturn]]]].

	aBlock value: stmtLists.

	^didSomething! !

!TMethod methodsFor: 'inlining' stamp: 'eem 8/29/2020 14:20'!
tryToInlineMethodsIn: aCodeGen
	"Expand any (complete) inline methods sent by this method.
	 Set the complete flag when all inlining has been done.
	 Answer if something was inlined."

	self definedAsMacro ifTrue:
		[complete ifTrue:
			[^false].
		 ^complete := true].

	^aCodeGen
		pushScope: declarations
		while:
			[| didSomething statementLists |
			self ensureConditionalAssignmentsAreTransformedIn: aCodeGen.
			didSomething := self tryToInlineMethodStatementsIn: aCodeGen statementListsInto: [:stmtLists| statementLists := stmtLists].
			didSomething := (self tryToInlineMethodExpressionsIn: aCodeGen) or: [didSomething].

			didSomething ifTrue:
				[writtenToGlobalVarsCache := nil].

			complete ifFalse:
				[self checkForCompletenessIn: aCodeGen.
				 complete ifTrue: [didSomething := true]].  "marking a method complete is progress"
			didSomething]! !

!TMethod methodsFor: 'utilities' stamp: 'eem 7/16/2020 15:56'!
typeFor: aVariableOrConstantOrVariableNameString in: aCodeGen
	"Answer the type for aVariable, deferring to aCodeGen (which defers to the vmClass)
	 if no type is found and the variable is global (not an arg or a local).  Expect the
	 cCodeGen to answer nil for variables without types. nil for typelessness is required
	 by the type propagation logic in inlineSend:directReturn:exitVar:in:."
	| varName |
	aVariableOrConstantOrVariableNameString isString ifFalse:
		["N.B. Very important *not* to type integers, to allow the client to do the work of merging various integer types."
		 (aVariableOrConstantOrVariableNameString isConstant
		  and: [aVariableOrConstantOrVariableNameString value isInteger not]) ifTrue:
			[^aVariableOrConstantOrVariableNameString typeOrNilFrom: aCodeGen in: self]].
	varName := aVariableOrConstantOrVariableNameString asString.
	^(declarations
			at: varName
			ifAbsent:
				[(args includes: varName) "arg types default to sqInt"
					ifTrue: ['sqInt ', varName]
					ifFalse:
						[(locals includes: varName) ifFalse: "don't provide type for locals"
							[aCodeGen typeOfVariable: varName]]]) ifNotNil:
		[:decl|
		aCodeGen extractTypeFor: varName fromDeclaration: decl]! !

!TMethod methodsFor: 'utilities' stamp: 'eem 2/11/2009 19:00'!
typeForSelf
	self assert: definingClass notNil.
	^[definingClass typeForSelf]
		on: MessageNotUnderstood
		do: [:ex| nil]! !

!TMethod methodsFor: 'inlining' stamp: 'eem 7/21/2014 06:10'!
unusedLabelForInlining: sourceMethod
	^labels add: (self unusedLabelForInliningInto: sourceMethod)! !

!TMethod methodsFor: 'inlining' stamp: 'eem 3/10/2016 13:05'!
unusedLabelForInliningInto: targetMethod
	^self unusedNamePrefixedBy: 'l' avoiding: (targetMethod == self
												ifTrue: [labels]
												ifFalse: [labels copy
															addAll: targetMethod labels;
															yourself])! !

!TMethod methodsFor: 'inlining support' stamp: 'eem 3/10/2016 13:07'!
unusedNamePrefixedBy: aString avoiding: usedNames
	"Choose a unique variable or label name with the given string as a prefix, avoiding
	 the names in the given collection. The selected name is added to usedNames."

	| n newVarName |
	n := 1.
	[newVarName := aString, n printString.
	 usedNames includes: newVarName] whileTrue:
		[n := n + 1].
	^usedNames add: newVarName! !

!TMethod methodsFor: 'inlining' stamp: 'eem 2/23/2015 10:47'!
usesVariableUninlinably: argName in: aCodeGen
	^parseTree anySatisfy:
		[:node|
		node isSend
		and: [(aCodeGen isAssertSelector: node selector)
		and: [node args anySatisfy:
				[:argNode|
				 argNode anySatisfy:
					[:subNode|
					 subNode isVariable and: [subNode name = argName]]]]]]! !

!TMethod methodsFor: 'utilities' stamp: ''!
variablesAssignedTo
	"Answer a collection of variables assigned to by this method."

	| refs |
	refs := Set new.
	parseTree nodesDo: [ :node |
		node isAssignment ifTrue: [ refs add: node variable name ].
	].
	^ refs! !

!TMethod methodsFor: 'primitive compilation' stamp: 'TPR 3/2/2000 16:08'!
vmNameString
	"return the string to use as the vm name in code generated for this method"
	^'self'! !

Object subclass: #TParseNode
	instanceVariableNames: 'comment'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TParseNode commentStamp: 'dtl 9/15/2008 09:05' prior: 0!
A TParseNode is node in the parse tree of a TMethod. Subclasses correspond to different types of nodes in a method parse tree. The tree of translation parse nodes mirrors the parse tree of a Smalltalk method, and is used for translating a Smalltalk method to C source.!

TParseNode subclass: #TAssignmentNode
	instanceVariableNames: 'variable expression'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TAssignmentNode methodsFor: 'transformations' stamp: 'mn 6/30/2000 13:19'!
bindVariableUsesIn: aDictionary
	"Do NOT bind the variable on the left-hand-side of an assignment statement."
	"was bindVariablesIn:"
	expression := expression bindVariableUsesIn: aDictionary.
! !

!TAssignmentNode methodsFor: 'transformations' stamp: 'eem 10/29/2012 11:38'!
bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen
	"Answer either the receiver, if it contains no references to the given variables, or a new node with the given variables rebound."
	"Do NOT bind the variable on the left-hand-side of an assignment statement."
	| boundExpression |
	boundExpression := expression bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen.
	^boundExpression == expression
		ifTrue: [self]
		ifFalse: [self shallowCopy
					setVariable: variable expression: boundExpression;
					yourself]! !

!TAssignmentNode methodsFor: 'transformations' stamp: 'eem 6/23/2020 11:41'!
bindVariablesIn: aDictionary

	self setVar: (variable bindVariablesIn: aDictionary)
		exp: (expression bindVariablesIn: aDictionary)! !

!TAssignmentNode methodsFor: 'C code generation' stamp: 'eem 12/15/2016 11:39'!
emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen
	^self emitCCodeAsExpressionOn: aStream level: level generator: aCodeGen! !

!TAssignmentNode methodsFor: 'C code generation' stamp: 'eem 12/15/2016 11:56'!
emitCCodeAsExpressionOn: aStream level: level generator: aCodeGen
	(expression isStmtList and: [expression statements size > 1]) ifTrue:
		[^self emitStatementListExpansionAsExpression: expression on: aStream level: level generator: aCodeGen].
	aStream nextPut: $(.
	self emitCCodeOn: aStream level: level generator: aCodeGen.
	aStream nextPut: $)! !

!TAssignmentNode methodsFor: 'C code generation' stamp: 'eem 7/19/2019 19:05'!
emitCCodeOn: aStream level: level generator: aCodeGen
	| selfAssignment |
	expression isSwitch ifTrue:
		[^expression emitCCodeOn: aStream addToEndOfCases: self level: level generator: aCodeGen].
	expression isLiteralArrayDeclaration ifTrue:
		[^self emitLiteralArrayDeclarationOn: aStream level: level generator: aCodeGen].
	(expression isSend and: [expression isValueExpansion]) ifTrue:
		[^self emitValueExpansionOn: aStream level: level generator: aCodeGen].
	(expression isStmtList and: [expression statements size > 1]) ifTrue:
		[^self emitStatementListExpansion: expression on: aStream level: level generator: aCodeGen].
	(selfAssignment := (variable isSameAs: expression) and: [(aCodeGen mustBeGlobal: variable name) not]) ifTrue:
		[aStream nextPutAll: '/* '].
	variable emitCCodeOn: aStream level: level generator: aCodeGen.
	self isVariableUpdatingAssignment
		ifTrue:
			[aStream
				space;
				nextPutAll: expression selector;	"+ or -"
				nextPut: $=;
				space.
			expression args first emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen]
		ifFalse:
			[aStream space; nextPut: $=; space.
			 selfAssignment ifTrue: [aStream nextPutAll: '*/ '].
			 expression emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen]! !

!TAssignmentNode methodsFor: 'C code generation' stamp: 'eem 12/22/2020 17:46'!
emitLiteralArrayDeclarationOn: aStream level: level generator: aCCodeGen
	| type literalName |
	type := expression args last value.
	self assert: type last = $*.
	literalName := variable name, 'Initializer'.
	aStream
		crtab: level;
		nextPutAll: '{ static ';
		nextPutAll: (aCCodeGen
						arrayInitializerCalled: literalName
						for: expression args first value
						sizeString: nil
						type: type allButLast) withBlanksTrimmed;
		nextPut: $;;
		crtab: level + 1;
		nextPutAll: variable name;
		nextPutAll: ' = ';
		nextPutAll: literalName;
		nextPut: $; ;
		crtab: level;
		nextPut: $};
		cr! !

!TAssignmentNode methodsFor: 'C code generation' stamp: 'eem 2/13/2015 17:18'!
emitStatementListExpansion: stmtList on: aStream level: level generator: aCodeGen
	stmtList statements last = variable ifTrue:
		[^expression emitCCodeOn: aStream level: level generator: aCodeGen].
	stmtList copy
		assignLastExpressionTo: variable;
		emitCCodeOn: aStream level: level generator: aCodeGen! !

!TAssignmentNode methodsFor: 'C code generation' stamp: 'eem 12/15/2016 11:57'!
emitStatementListExpansionAsExpression: stmtList on: aStream level: level generator: aCodeGen
	stmtList statements last = variable ifTrue:
		[^expression emitCCodeAsExpressionOn: aStream level: level generator: aCodeGen].
	stmtList copy
		assignLastExpressionTo: variable;
		emitCCodeAsExpressionOn: aStream level: level generator: aCodeGen! !

!TAssignmentNode methodsFor: 'C code generation' stamp: 'eem 5/7/2020 19:51'!
emitValueExpansionOn: aStream level: level generator: aCodeGen
	| stmtList lastStmt copiedStatements |
	self assert: (expression isSend and: [expression isValueExpansion]).
	stmtList := expression receiver.
	lastStmt := stmtList statements last.
	(lastStmt = variable or: [lastStmt isReturn]) ifTrue:
		[^expression emitCCodeOn: aStream level: level generator: aCodeGen].
	copiedStatements := stmtList copy.
	copiedStatements statements
		at: stmtList statements size
		put: (TAssignmentNode new
				setVariable: variable
				expression: lastStmt).
	expression copy
		receiver: copiedStatements;
		emitCCodeOn: aStream level: level generator: aCodeGen! !

!TAssignmentNode methodsFor: 'accessing' stamp: ''!
expression

	^expression! !

!TAssignmentNode methodsFor: 'testing' stamp: 'eem 2/6/2020 14:24'!
hasEffect
	"Answer if this node has an effect on execution state (does something).
	 Statements that don't have any effect can be elided if their value is unused."
	^true! !

!TAssignmentNode methodsFor: 'testing' stamp: ''!
isAssignment

	^true! !

!TAssignmentNode methodsFor: 'testing' stamp: 'eem 6/23/2020 11:25'!
isSameAs: aTParseNode
	^self == aTParseNode
	 or: [aTParseNode isAssignment
		 and: [(variable isSameAs: aTParseNode variable)
		 and: [expression isSameAs: aTParseNode expression]]]! !

!TAssignmentNode methodsFor: 'testing' stamp: ''!
isVariableUpdatingAssignment
	"Return true if this assignment statement is of one of the forms:
		var = var + ...
		var = var - ...
	Such assignments statements can exploit the C updating assignment operators. For example, 'x += 4' can be generated instead of 'x = x + 4'. This produces better code under some C compilers, most notably the CodeWarrior 68K compiler."

	| sel |
	(expression isSend and: [expression receiver isVariable]) ifFalse: [^ false].
	sel := expression selector.
	^ (expression receiver name = variable name) and: [(sel = #+) or: [sel = #-]]! !

!TAssignmentNode methodsFor: 'enumerating' stamp: 'eem 11/19/2012 12:12'!
nodesDo: aBlock
	"Apply aBlock to all nodes in the receiver.
	 N.B. This is assumed to be bottom-up, leaves first."
	variable nodesDo: aBlock.
	expression nodesDo: aBlock.
	aBlock value: self! !

!TAssignmentNode methodsFor: 'enumerating' stamp: 'eem 10/13/2013 16:59'!
nodesDo: aBlock parent: parent
	"Apply aBlock to all nodes in the receiver with each node's parent.
	 N.B. This is assumed to be bottom-up, leaves first."
	variable nodesDo: aBlock parent: self.
	expression nodesDo: aBlock parent: self.
	aBlock value: self value: parent! !

!TAssignmentNode methodsFor: 'enumerating' stamp: 'eem 3/31/2015 10:54'!
nodesDo: aBlock parent: parent unless: cautionaryBlock

	(cautionaryBlock value: self value: parent) ifTrue: [^self].
	variable nodesDo: aBlock parent: self unless: cautionaryBlock.
	expression nodesDo: aBlock parent: self unless: cautionaryBlock.
	aBlock value: self value: parent! !

!TAssignmentNode methodsFor: 'enumerating' stamp: 'eem 2/27/2009 16:46'!
nodesDo: aBlock unless: cautionaryBlock

	(cautionaryBlock value: self) ifTrue: [^self].
	variable nodesDo: aBlock unless: cautionaryBlock.
	expression nodesDo: aBlock unless: cautionaryBlock.
	aBlock value: self.! !

!TAssignmentNode methodsFor: 'copying' stamp: 'eem 6/23/2020 11:42'!
postCopy

	self setVar: variable copy exp: expression copy! !

!TAssignmentNode methodsFor: 'C code generation' stamp: 'eem 3/11/2009 13:14'!
prependCCodeOn: aStream level: level generator: aCodeGen
	variable emitCCodeOn: aStream level: level generator: aCodeGen.
	aStream
		space;
		nextPutAll: (self isVariableUpdatingAssignment
						ifTrue: [expression selector "+ or -"]
						ifFalse: ['']);
		nextPut: $=;
		space! !

!TAssignmentNode methodsFor: 'printing' stamp: 'ar 4/4/2006 21:12'!
printOn: aStream level: level

	variable printOn: aStream level: level.
	aStream nextPutAll: ' := '.
	expression printOn: aStream level: level + 2.! !

!TAssignmentNode methodsFor: 'transformations' stamp: 'ikp 9/26/97 14:50'!
removeAssertions

	expression removeAssertions! !

!TAssignmentNode methodsFor: 'transformations' stamp: 'eem 6/23/2020 11:41'!
replaceNodesIn: aDictionary

	^aDictionary
		at: self
		ifAbsent:
			[self setVar: (variable replaceNodesIn: aDictionary)
				exp: (expression replaceNodesIn: aDictionary)]! !

!TAssignmentNode methodsFor: 'accessing' stamp: 'eem 10/25/2013 10:44'!
setExpression: expressionNode

	expression := expressionNode! !

!TAssignmentNode methodsFor: 'private' stamp: 'eem 6/23/2020 11:48'!
setVar: varNode exp: expressionNode
	"This is a private setter, just for breakpointing..."
	variable := varNode.
	expression := expressionNode! !

!TAssignmentNode methodsFor: 'accessing' stamp: 'eem 2/13/2015 17:19'!
setVariable: varNode expression: expressionNode
	self assert: (expressionNode isGoTo or: [expressionNode isLabel]) not.
	variable := varNode.
	expression := expressionNode.! !

!TAssignmentNode methodsFor: 'testing' stamp: 'eem 10/24/2013 13:02'!
structTargetKindIn: aCodeGen
	"Answer if the recever evaluates to a struct or struct pointer
	 and hence can be dereferenced using . or ->.  Answer any of
	 #struct #pointer or nil"
	^variable structTargetKindIn: aCodeGen! !

!TAssignmentNode methodsFor: 'type inference' stamp: 'eem 6/7/2016 18:08'!
typeOrNilFrom: aCodeGenerator in: aTMethod
	"This is the default type in case of doubt"
	^(variable typeOrNilFrom: aCodeGenerator in: aTMethod) ifNil:
		[expression typeOrNilFrom: aCodeGenerator in: aTMethod]! !

!TAssignmentNode methodsFor: 'accessing' stamp: ''!
variable

	^variable! !

!TAssignmentNode methodsFor: 'accessing' stamp: 'eem 11/12/2013 14:22'!
variableNameOrNil
	^variable variableNameOrNil! !

TParseNode subclass: #TBraceCaseNode
	instanceVariableNames: 'caseLabels cases'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TBraceCaseNode methodsFor: 'transformations' stamp: 'eem 2/11/2009 12:13'!
bindVariableUsesIn: aDictionary

	caseLabels := caseLabels collect: [:node| node bindVariableUsesIn: aDictionary].
	cases := cases collect: [:node| node bindVariableUsesIn: aDictionary]! !

!TBraceCaseNode methodsFor: 'transformations' stamp: 'eem 10/29/2012 11:38'!
bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen
	"Answer either the receiver, if it contains no references to the given variables, or a new node with the given variables rebound."
	| newCaseLabels newCases |
	newCaseLabels := caseLabels collect: [:node| node bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen].
	newCases := cases collect: [:node| node bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen].
	^(newCaseLabels = caseLabels
	   and: [newCases = cases])
		ifTrue: [self]
		ifFalse: [self shallowCopy
					caseLabels: newCaseLabels;
					cases: newCases;
					yourself]! !

!TBraceCaseNode methodsFor: 'transformations' stamp: 'eem 2/11/2009 12:12'!
bindVariablesIn: aDictionary

	caseLabels := caseLabels collect: [:node| node bindVariablesIn: aDictionary].
	cases := cases collect: [:node| node bindVariablesIn: aDictionary]! !

!TBraceCaseNode methodsFor: 'accessing' stamp: 'eem 2/9/2009 10:27'!
caseLabels
	"Answer the value of caseLabels"

	^ caseLabels! !

!TBraceCaseNode methodsFor: 'accessing' stamp: 'eem 2/9/2009 10:27'!
caseLabels: anObject
	"Set the value of caseLabels"

	caseLabels := anObject! !

!TBraceCaseNode methodsFor: 'accessing' stamp: 'eem 2/9/2009 10:27'!
cases
	"Answer the value of cases"

	^ cases! !

!TBraceCaseNode methodsFor: 'accessing' stamp: 'eem 2/9/2009 10:27'!
cases: anObject
	"Set the value of cases"

	cases := anObject! !

!TBraceCaseNode methodsFor: 'testing' stamp: 'eem 2/6/2020 14:25'!
hasEffect
	"Answer if this node has an effect on execution state (does something).
	 Statements that don't have any effect can be elided if their value is unused."
	^cases anySatisfy: [:node| node hasEffect]! !

!TBraceCaseNode methodsFor: 'enumerating' stamp: 'eem 11/19/2012 12:12'!
nodesDo: aBlock
	"Apply aBlock to all nodes in the receiver.
	 N.B. This is assumed to be bottom-up, leaves first."
	caseLabels do:
		[:node| node nodesDo: aBlock].
	cases do:
		[:node| node nodesDo: aBlock].
	aBlock value: self! !

!TBraceCaseNode methodsFor: 'enumerating' stamp: 'eem 8/21/2021 19:32'!
nodesDo: aBlock parent: parent
	"Apply aBlock to all nodes in the receiver with each node's parent.
	 N.B. This is assumed to be bottom-up, leaves first."
	caseLabels do:
		[:node| node nodesDo: aBlock parent: self].
	cases do:
		[:node| node nodesDo: aBlock parent: self].
	aBlock value: self value: parent! !

!TBraceCaseNode methodsFor: 'enumerating' stamp: 'eem 3/31/2015 10:56'!
nodesDo: aBlock parent: parent unless: cautionaryBlock

	(cautionaryBlock value: self value: parent) ifTrue: [^self].
	caseLabels do:
		[:node| node nodesDo: aBlock parent: self unless: cautionaryBlock].
	cases do:
		[:node| node nodesDo: aBlock parent: self unless: cautionaryBlock].
	aBlock value: self value: parent! !

!TBraceCaseNode methodsFor: 'enumerating' stamp: 'eem 2/27/2009 16:47'!
nodesDo: aBlock unless: cautionaryBlock

	(cautionaryBlock value: self) ifTrue: [^self].
	caseLabels do:
		[:node| node nodesDo: aBlock unless: cautionaryBlock].
	cases do:
		[:node| node nodesDo: aBlock unless: cautionaryBlock].
	aBlock value: self! !

!TBraceCaseNode methodsFor: 'copying' stamp: 'eem 7/31/2021 22:02'!
postCopy
	caseLabels := caseLabels collect: [:ea| ea copy].
	cases := cases collect: [:ea| ea copy]! !

!TBraceCaseNode methodsFor: 'printing' stamp: 'eem 7/31/2021 21:07'!
printOn: aStream level: level
	aStream crtab: level; nextPut: ${.
	caseLabels with: cases do:
		[:caseLabel :case|
		caseLabel printOn: aStream.
		aStream nextPutAll: ' -> '.
		case printOn: aStream level: level + 1.
		aStream nextPut: $.].
	aStream space; nextPut: $}! !

!TBraceCaseNode methodsFor: 'enumerating' stamp: 'eem 2/11/2009 10:16'!
replaceNodesIn: aDictionary

	^aDictionary at: self ifAbsent: [
		caseLabels := caseLabels collect: [:node| node replaceNodesIn: aDictionary].
		cases := cases collect: [:node| node replaceNodesIn: aDictionary].
		self]! !

!TBraceCaseNode methodsFor: 'type inference' stamp: 'eem 1/24/2017 18:50'!
typeOrNilFrom: aCodeGenerator in: aTMethod
	| types |
	types := Set withAll: (cases collect: [:case| case typeOrNilFrom: aCodeGenerator in: aTMethod]).
	^types size = 1 ifTrue: [types anyOne]! !

TParseNode subclass: #TBraceNode
	instanceVariableNames: 'elements'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TBraceNode commentStamp: 'eem 8/21/2021 19:41' prior: 0!
A TBraceNode is a holder for a naked brace node ({ expr1. ... exprN}).  These are used in printf expressions, and also may pop up in macro methods (see Cogit>>#mapPerMethodProfile).

Instance Variables
	elements:		<OrderedCollection>

elements
	- the expressions of the brace!

!TBraceNode methodsFor: 'transformations' stamp: 'eem 8/21/2021 19:34'!
bindVariableUsesIn: aDictionary

	elements := elements collect: [:node| node bindVariableUsesIn: aDictionary]! !

!TBraceNode methodsFor: 'transformations' stamp: 'eem 8/21/2021 19:36'!
bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen
	"Answer either the receiver, if it contains no references to the given variables, or a new node with the given variables rebound."
	| newElements |
	newElements := elements collect: [:node| node bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen].
	^newElements = elements
		ifTrue: [self]
		ifFalse: [self shallowCopy
					elements: newElements;
					yourself]! !

!TBraceNode methodsFor: 'transformations' stamp: 'eem 8/21/2021 19:36'!
bindVariablesIn: aDictionary

	elements := elements collect: [:node| node bindVariablesIn: aDictionary]! !

!TBraceNode methodsFor: 'accessing' stamp: 'eem 8/21/2021 19:50'!
elements
	^elements! !

!TBraceNode methodsFor: 'accessing' stamp: 'eem 8/21/2021 19:41'!
elements: anObject
	"Set the value of elements"

	elements := anObject! !

!TBraceNode methodsFor: 'testing' stamp: 'eem 8/21/2021 19:37'!
hasEffect
	"Answer if this node has an effect on execution state (does something).
	 Statements that don't have any effect can be elided if their value is unused."
	^elements anySatisfy: [:node| node hasEffect]! !

!TBraceNode methodsFor: 'testing' stamp: 'eem 8/21/2021 19:48'!
isBrace
	^true! !

!TBraceNode methodsFor: 'enumerating' stamp: 'eem 8/21/2021 19:31'!
nodesDo: aBlock
	"Apply aBlock to all nodes in the receiver.
	 N.B. This is assumed to be bottom-up, leaves first."
	elements do:
		[:node| node nodesDo: aBlock].
	aBlock value: self! !

!TBraceNode methodsFor: 'enumerating' stamp: 'eem 8/21/2021 19:32'!
nodesDo: aBlock parent: parent
	"Apply aBlock to all nodes in the receiver with each node's parent.
	 N.B. This is assumed to be bottom-up, leaves first."
	elements do:
		[:node| node nodesDo: aBlock parent: self].
	aBlock value: self value: parent! !

!TBraceNode methodsFor: 'enumerating' stamp: 'eem 8/21/2021 19:33'!
nodesDo: aBlock parent: parent unless: cautionaryBlock

	(cautionaryBlock value: self value: parent) ifTrue: [^self].
	elements do:
		[:node| node nodesDo: aBlock parent: self unless: cautionaryBlock].
	aBlock value: self value: parent! !

!TBraceNode methodsFor: 'enumerating' stamp: 'eem 8/21/2021 19:37'!
nodesDo: aBlock unless: cautionaryBlock

	(cautionaryBlock value: self) ifTrue: [^self].
	elements do:
		[:node| node nodesDo: aBlock unless: cautionaryBlock].
	aBlock value: self! !

!TBraceNode methodsFor: 'copying' stamp: 'eem 8/21/2021 19:37'!
postCopy
	elements := elements collect: [:ea| ea copy]! !

!TBraceNode methodsFor: 'printing' stamp: 'eem 8/21/2021 19:38'!
printOn: aStream level: level
	aStream crtab: level; nextPut: ${.
	elements
		do: [:element| element printOn: aStream]
		separatedBy: [aStream nextPut: $.; space].
	aStream space; nextPut: $}! !

!TBraceNode methodsFor: 'enumerating' stamp: 'eem 8/21/2021 19:39'!
replaceNodesIn: aDictionary

	^aDictionary
		at: self
		ifAbsent:
			[elements := elements collect: [:node| node replaceNodesIn: aDictionary].
			 self]! !

!TBraceNode methodsFor: 'type inference' stamp: 'eem 8/21/2021 19:39'!
typeOrNilFrom: aCodeGenerator in: aTMethod
	^nil! !

TParseNode subclass: #TCaseStmtNode
	instanceVariableNames: 'expression firsts lasts cases'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TCaseStmtNode commentStamp: '<historical>' prior: 0!
I implement the main dispatch case statements for bytecode and primitive dispatch.  See TMethod classPool associationAt: #CaseStatements!

!TCaseStmtNode methodsFor: 'transformations' stamp: 'jm 12/10/1998 18:26'!
bindVariableUsesIn: aDictionary

	expression := expression bindVariableUsesIn: aDictionary.
	cases := cases collect: [ :c | c bindVariableUsesIn: aDictionary ].! !

!TCaseStmtNode methodsFor: 'transformations' stamp: 'eem 10/29/2012 11:38'!
bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen
	"Answer either the receiver, if it contains no references to the given variables, or a new node with the given variables rebound."
	| newExpression newCases |
	newExpression := expression bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen.
	newCases := cases collect: [:c| c bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen].
	^(newExpression = expression
	   and: [newCases = cases])
		ifTrue: [self]
		ifFalse: [self shallowCopy
					setExpression: newExpression cases: newCases;
					yourself]! !

!TCaseStmtNode methodsFor: 'transformations' stamp: ''!
bindVariablesIn: aDictionary

	expression := expression bindVariablesIn: aDictionary.
	cases := cases collect: [ :c | c bindVariablesIn: aDictionary ].! !

!TCaseStmtNode methodsFor: 'accessing' stamp: ''!
cases

	^cases! !

!TCaseStmtNode methodsFor: 'transformations' stamp: 'GuillermoPolito 1/24/2020 16:20'!
customizeCase: caseParseTree forVar: varName from: firstIndex to: lastIndex in: codeGen method: aTMethod expandedCases: seen
	"Return a collection of copies of the given parse tree, each of which has the value of the case index substituted for the given variable."

	| newCases dict newCase |
	newCases := OrderedCollection new.
	firstIndex to: lastIndex do:
		[ :caseIndex |
		dict := Dictionary new.
		dict at: varName put: (TConstantNode new setValue: caseIndex).
		newCase := caseParseTree copy bindVariableUsesIn: dict andConstantFoldIf: true in: codeGen.
		self processSharedCodeBlocks: newCase forCase: caseIndex in: codeGen method: aTMethod expandedCases: seen.
		aTMethod labels addAll: (newCase renameLabelsForInliningInto: aTMethod).
		newCases addLast: newCase].
	^newCases! !

!TCaseStmtNode methodsFor: 'transformations' stamp: 'eem 11/12/2012 14:58'!
customizeShortCasesForDispatchVar: varName in: codeGen method: aTMethod
	"Make customized versions of short bytecode methods, substituting a constant having the case index value for the given variable.
	 This produces better code for short bytecodes such as instance variable pushes that encode the index of the instance variable in the bytecode."

	| newFirsts newLasts newCases seen |
	newFirsts := OrderedCollection new.
	newLasts := OrderedCollection new.
	newCases := OrderedCollection new.
	seen := Set new.	"So that sharedCodeNamed:inCase:'s can be used in multiple bytecode set
						 configurations we need to expand the code only once."
	1 to: cases size do:
		[ :i | | l f case expanded |
		l := lasts at: i.
		f := firsts at: i.
		case := cases at: i.
		expanded := false.
		((l - f) > 1  "case code covers multiple cases"
		 and: [case shouldExpand]) ifTrue:
				[newFirsts addAll: (f to: l) asArray.
				 newLasts addAll: (f to: l) asArray.
				 newCases addAll: (self customizeCase: case forVar: varName from: f to: l in: codeGen method: aTMethod expandedCases: seen).
				 expanded := true].
		expanded ifFalse:
			[self processSharedCodeBlocks: case forCase: f in: codeGen method: aTMethod expandedCases: seen.
			 newFirsts addLast: f.
			 newLasts addLast: l.
			 newCases addLast: case]].
	firsts := newFirsts asArray.
	lasts := newLasts asArray.
	cases := newCases asArray! !

!TCaseStmtNode methodsFor: 'C code generation' stamp: 'eem 7/25/2020 13:30'!
emitCCodeOn: aStream addToEndOfCases: aNodeOrNil level: level generator: aCodeGen

	| indent |
	indent := String new: level withAll: Character tab.
	aStream nextPutAll: 'switch ('.
	expression emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen.
	aStream nextPutAll: ') {'; cr.
	cases withIndexDo:
		[:case :i|
		 (firsts at: i) to: (lasts at: i) do:
			[ :caseIndex |
			aStream nextPutAll: indent; nextPutAll: 'case '; print: caseIndex; nextPut: $:; cr].
		 case emitCCodeOn: aStream prependToEnd: aNodeOrNil level: level + 1 generator: aCodeGen.
		 (aNodeOrNil notNil and: [aNodeOrNil isReturn]) ifFalse:
			[aStream crtab: level + 1; nextPutAll: 'break;'].
		 aStream cr].
	aStream
		crtab: level;
		nextPutAll: 'default:';
		crtab: level + 1;
		nextPutAll: 'error("Case not found");'.
	aNodeOrNil ifNotNil:
		[aStream crtab: level + 1.
		 (aNodeOrNil copy setExpression: (TConstantNode new setValue: -1))
			emitCCodeOn: aStream level: level generator: aCodeGen.
		 aStream nextPut: $;].
	aStream
		crtab: level;
		nextPut: $}! !

!TCaseStmtNode methodsFor: 'C code generation' stamp: 'eem 11/18/2012 19:55'!
emitCCodeOn: aStream level: level generator: aCodeGen
	| printMod expansions duplicates |
	printMod := false.
	(expression isVariable
	 and: [expression name = 'currentBytecode']) ifTrue:
		[printMod := true.
		 aStream nextPutAll: 'bytecodeDispatchDebugHook();'; cr; crtab: level.
		 aCodeGen outputAsmLabel: 'bytecodeDispatch' on: aStream.
		 aStream crtab: level].
	aStream nextPutAll: 'switch ('.
	expression emitCCodeOn: aStream level: level generator: aCodeGen.
	aStream nextPutAll: ') {'; cr.
	expansions := aCodeGen suppressAsmLabelsWhile:
						[cases collect:
							[:case|
							self filterCommentsFrom:
								(String streamContents:
									[:s|
									case emitCCodeOn: s level: 0 generator: aCodeGen])]].
	duplicates := Set new.
	1 to: cases size do:
		[:i|
		(duplicates includes: i) ifFalse:
			[(duplicates addAll: ((i to: cases size) select: [:j| (expansions at: i) = (expansions at: j)])) do:
				[:k|
				(firsts at: k) to: (lasts at: k) do:
					[:caseIndex|
					aStream tab: level; nextPutAll: 'case '; print: caseIndex; nextPut: $:.
					(caseIndex > 255 and: [printMod]) ifTrue:
						[aStream nextPutAll: ' /*'; print: (caseIndex bitAnd: 255); nextPutAll: '*/'].
					aStream cr]].
			(cases at: i) emitCCodeOn: aStream level: level + 1 generator: aCodeGen.
			aStream tab: level + 1; nextPutAll: 'break;'; cr]].
	aStream tab: level; nextPut: $}! !

!TCaseStmtNode methodsFor: 'accessing' stamp: ''!
expression

	^expression! !

!TCaseStmtNode methodsFor: 'private' stamp: 'eem 11/18/2012 19:55'!
filterCommentsFrom: aString
	"elide any /* comment */ occurrences from aString."
	| i |
	i := aString indexOfSubCollection: '/*'startingAt: 1 ifAbsent: [^aString].
	^(aString copyFrom: 1 to: i - 1),
	  (self filterCommentsFrom:
		(aString
			copyFrom: (aString indexOfSubCollection: '*/' startingAt: i + 2) + 2
			to: aString size))! !

!TCaseStmtNode methodsFor: 'testing' stamp: 'eem 2/6/2020 14:26'!
hasEffect
	"Answer if this node has an effect on execution state (does something).
	 Statements that don't have any effect can be elided if their value is unused."
	^expression hasEffect
	  or: [cases anySatisfy: [:node| node hasEffect]]! !

!TCaseStmtNode methodsFor: 'testing' stamp: ''!
isCaseStmt

	^true! !

!TCaseStmtNode methodsFor: 'enumerating' stamp: 'eem 11/19/2012 12:12'!
nodesDo: aBlock
	"Apply aBlock to all nodes in the receiver.
	 N.B. This is assumed to be bottom-up, leaves first."
	expression nodesDo: aBlock.
	cases do: [ :c | c nodesDo: aBlock ].
	aBlock value: self! !

!TCaseStmtNode methodsFor: 'enumerating' stamp: 'eem 10/13/2013 17:00'!
nodesDo: aBlock parent: parent
	"Apply aBlock to all nodes in the receiver with each node's parent.
	 N.B. This is assumed to be bottom-up, leaves first."
	expression nodesDo: aBlock parent: self.
	cases do: [:c| c nodesDo: aBlock parent: self].
	aBlock value: self value: parent! !

!TCaseStmtNode methodsFor: 'enumerating' stamp: 'eem 3/31/2015 10:56'!
nodesDo: aBlock parent: parent unless: cautionaryBlock

	(cautionaryBlock value: self value: parent) ifTrue: [^self].
	expression nodesDo: aBlock parent: self unless: cautionaryBlock.
	cases do: [ :c | c nodesDo: aBlock parent: self unless: cautionaryBlock].
	aBlock value: self value: parent! !

!TCaseStmtNode methodsFor: 'enumerating' stamp: 'eem 2/27/2009 16:47'!
nodesDo: aBlock unless: cautionaryBlock

	(cautionaryBlock value: self) ifTrue: [^self].
	expression nodesDo: aBlock unless: cautionaryBlock.
	cases do: [ :c | c nodesDo: aBlock  unless: cautionaryBlock].
	aBlock value: self! !

!TCaseStmtNode methodsFor: 'copying' stamp: 'eem 2/9/2009 19:24'!
postCopy

	expression := expression copy.
	firsts := firsts copy.
	lasts := lasts copy.
	cases := (cases collect: [ :case | case copy ])! !

!TCaseStmtNode methodsFor: 'printing' stamp: ''!
printOn: aStream level: level

	aStream crtab: level.
	aStream nextPutAll: 'select '.
	expression printOn: aStream level: level.
	aStream nextPutAll: ' in'.
	1 to: cases size do: [ :i |
		(firsts at: i) to: (lasts at: i) do: [ :caseIndex |
			aStream crtab: level.
			aStream nextPutAll: 'case ', caseIndex printString, ':'.
		].
		aStream crtab: level + 1.
		(cases at: i) printOn: aStream level: level + 1.
	].
	aStream crtab: level.
	aStream nextPutAll: 'end select'.! !

!TCaseStmtNode methodsFor: 'transformations' stamp: 'eem 7/26/2014 07:51'!
processSharedCodeBlocks: caseTree forCase: caseIndex in: codeGen method: aTMethod expandedCases: seen
	"Process any shared code blocks in the case parse tree for the given case, either inlining them or making them a 'goto sharedLabel'."
	| caseMethod map meth sharedNode exitLabel |
	exitLabel := nil.
	"caseTree is expected to be a TStmtListNode whose first element is a comment
	 and whose second element is a TInlineNode for a method."
	caseMethod := caseTree statements second method.
	[sharedNode := nil.
	 map := IdentityDictionary new.
	 caseTree nodesDo:
		[:node|
		(sharedNode isNil
		and: [node isSend 
		and:[(meth := codeGen methodNamed: node selector) notNil
		and:[meth sharedCase notNil]]]) ifTrue:
			[(meth sharedCase = (meth sharedCase isSymbol
									ifTrue: [caseMethod selector]
									ifFalse: [caseIndex])
			  and: [(seen includes: meth sharedLabel) not])
				ifTrue:
					["If the bytecode (the caseMethod) ends with a message that has a lastCase (and lastLabel) then
					  that will be converted into a goto and control will continue to that code,  If the bytecode does
					  /not/ end with a message that has a lastCase (and lastLabel) then control should not continue to
					  that shared case.  expandViaFallThrough captures this, true for the former, false for the latter."
					 | expandViaFallThrough |
					 expandViaFallThrough := false.
					 caseMethod statements last isSend ifTrue:
						[(codeGen methodNamed: caseMethod statements last selector) ifNotNil:
							[:m| expandViaFallThrough := m sharedCase notNil]].
					 seen add: meth sharedLabel.
					 map
						at: node
						put: (expandViaFallThrough
								ifTrue: [sharedNode := meth.
										TLabeledCommentNode new setComment: 'goto ', meth sharedLabel]
								ifFalse: ["Still need recursive expansjon to continue but don't want
										  to duplicate the node, so substitue an empty method."
										 sharedNode := TLabeledCommentNode new setComment: 'null '.
										 meth copy
											renameLabelsForInliningInto: aTMethod;
											addLabelsTo: aTMethod;
											asInlineNode])]
				ifFalse:
					[map at: node put: (TGoToNode new setLabel: meth sharedLabel)]]].
	 caseTree replaceNodesIn: map.
	 "recursively expand"
	 sharedNode notNil]
		whileTrue:
			[sharedNode isTMethod ifTrue:
				[meth := sharedNode copy.
				 meth hasReturn ifTrue:
					[exitLabel ifNil:
						[exitLabel := aTMethod unusedLabelForInliningInto: aTMethod.
						 aTMethod labels add: exitLabel].
					meth exitVar: nil label: exitLabel].
				meth
					renameLabelsForInliningInto: aTMethod;
					addLabelsTo: aTMethod.
				caseTree setStatements: (caseTree statements copyWith: meth asInlineNode)]].
	exitLabel ifNotNil:
		[caseTree setStatements: (caseTree statements copyWith:
			(TLabeledCommentNode new setLabel: exitLabel comment: 'end case'))]! !

!TCaseStmtNode methodsFor: 'transformations' stamp: 'ikp 9/26/97 14:50'!
removeAssertions
	expression removeAssertions.
	cases do: [ :case | case removeAssertions ].! !

!TCaseStmtNode methodsFor: 'transformations' stamp: ''!
replaceNodesIn: aDictionary

	^aDictionary at: self ifAbsent: [
		expression := expression replaceNodesIn: aDictionary.
		cases := cases collect: [ :c | c replaceNodesIn: aDictionary ].
		self]! !

!TCaseStmtNode methodsFor: 'private' stamp: 'eem 10/29/2012 11:34'!
setExpression: newExpression cases: newCases
	expression := newExpression.
	cases := newCases! !

!TCaseStmtNode methodsFor: 'accessing' stamp: 'ar 11/18/1999 20:08'!
setExpression: aNode selectors: selectorList arguments: anArray
	"Initialize the node from the given set of selectors."
	"Note: Each case is a statement list with containing one statement, a send to self of a selector from the given selector list. Having statement list nodes makes inlining easier later."

	| selfNode stmt lastSel firstInRun sel |
	expression := aNode.
	selfNode := TVariableNode new setName: 'self'.
	firsts := OrderedCollection new: 400.
	lasts := OrderedCollection new: 400.
	cases := OrderedCollection new: 400.
	lastSel := selectorList first.
	firstInRun := 0.
	1 to: selectorList size do: [ :i |
		sel := selectorList at: i.
		sel ~= lastSel ifTrue: [
			firsts add: firstInRun.
			lasts add: i - 2.
			stmt := TSendNode new setSelector: lastSel receiver: selfNode arguments: anArray.
			cases add: (TStmtListNode new setArguments: #() statements: (Array with: stmt)).
			lastSel := sel.
			firstInRun := i - 1.
		].
	].
	firsts add: firstInRun.
	lasts add: selectorList size - 1.
	stmt := TSendNode new setSelector: lastSel receiver: selfNode arguments: anArray.
	cases add: (TStmtListNode new setArguments: #() statements: (Array with: stmt)).! !

TParseNode subclass: #TConstantNode
	instanceVariableNames: 'value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TConstantNode methodsFor: 'transformations' stamp: 'eem 11/13/2017 15:00'!
asPrintfFormatStringNode
	^TPrintfFormatStringNode new
		setValue: value;
		comment: comment;
		yourself! !

!TConstantNode methodsFor: 'accessing' stamp: 'eem 1/4/2017 17:22'!
constantNumbericValueIfAtAllPossibleOrNilIn: aCCodeGen
	"This is a version of constantNumbericValueOrNil for type checking rather than code generation.
	 It aims to yield a value if at all possible.  This should /not/ be overridden in TDefineNode."
	^value isNumber ifTrue: [value]! !

!TConstantNode methodsFor: 'accessing' stamp: 'eem 12/11/2014 11:38'!
constantNumbericValueOrNil

	^value isNumber ifTrue: [value]! !

!TConstantNode methodsFor: 'C code generation' stamp: 'eem 10/29/2012 13:06'!
emitCCodeOn: aStream level: level generator: aCodeGen
	"Emit a C literal."
	aStream nextPutAll: (aCodeGen cLiteralFor: value).
	comment ifNotNil:
		[aStream
			nextPutAll: ' /* ';
			nextPutAll: comment;
			nextPutAll: ' */']! !

!TConstantNode methodsFor: 'testing' stamp: 'eem 11/17/2015 14:29'!
hasSideEffect
	"Answer if the parse tree rooted at this node has a side-effect or not."
	^false! !

!TConstantNode methodsFor: 'testing' stamp: ''!
isConstant

	^true! !

!TConstantNode methodsFor: 'testing' stamp: 'eem 7/6/2021 17:49'!
isEffectFree
	^true! !

!TConstantNode methodsFor: 'testing' stamp: ''!
isLeaf

	^true! !

!TConstantNode methodsFor: 'comparing' stamp: 'eem 6/23/2020 11:25'!
isSameAs: aTParseNode
	^self == aTParseNode
	 or: [aTParseNode isConstant
		 and: [value class == aTParseNode value class
		 and: [value = aTParseNode value]]]! !

!TConstantNode methodsFor: 'accessing' stamp: 'ar 7/8/2003 11:04'!
name
	^''! !

!TConstantNode methodsFor: 'accessing' stamp: 'acg 12/17/1999 07:19'!
nameOrValue

	^value! !

!TConstantNode methodsFor: 'enumerating' stamp: 'eem 10/13/2013 17:10'!
nodesDo: aBlock parent: parent
	"Apply aBlock to all nodes in the receiver with each node's parent.
	 N.B. This is assumed to be bottom-up, leaves first."
	aBlock value: self value: parent! !

!TConstantNode methodsFor: 'printing' stamp: 'eem 12/12/2012 11:23'!
printOn: aStream level: level
	value isSymbol
		ifTrue: [aStream nextPutAll: (value copyWithout: $:)]
		ifFalse: [value storeOn: aStream]! !

!TConstantNode methodsFor: 'accessing' stamp: ''!
setValue: anObject

	value := anObject.! !

!TConstantNode methodsFor: 'type inference' stamp: 'eem 7/16/2020 16:00'!
typeOrNilFrom: aCodeGenerator in: aTMethod
	"For integers, answer int unless the value does not fit into a 32bits signed int.
	In that case, answer the shortest architecture independant integer type that could hold the constant.
	This method must be consistent with CCodeGenerator>>cLiteralFor:"
	| hb |
	value isInteger
		ifTrue:
			[value positive
				ifTrue:
					[hb := value highBit.
					hb < 32 ifTrue: [^#int].
					hb = 32 ifTrue: [^#'unsigned int'].
					hb = 64 ifTrue: [^#'unsigned long long'].
					^#'long long']
				ifFalse:
					[hb := value bitInvert highBit.
					hb < 32 ifTrue: [^#int].
					^#'long long']].
	value isFloat ifTrue: [^#double].
	(#(nil true false) includes: value) ifTrue: [^#sqInt]. "A machine word sized variable is better on 64-bits than int, we think."
	(value isString and: [value isSymbol not]) ifTrue: [^#'char *'].
	^nil! !

!TConstantNode methodsFor: 'accessing' stamp: ''!
value

	^value! !

TConstantNode subclass: #TDefineNode
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TDefineNode methodsFor: 'accessing' stamp: 'eem 5/17/2015 14:24'!
constantNumbericValueOrNil
	"Override so that named constants don't get elided."
	^nil! !

!TDefineNode methodsFor: 'C code generation' stamp: 'ar 7/8/2003 11:08'!
emitCCodeOn: aStream level: level generator: aCodeGen
	"Emit a C literal."

	aStream nextPutAll: name.! !

!TDefineNode methodsFor: 'testing' stamp: 'eem 1/21/2016 16:12'!
isDefine

	^true! !

!TDefineNode methodsFor: 'comparing' stamp: 'eem 6/23/2020 11:25'!
isSameAs: aTParseNode
	^self == aTParseNode
	 or: [self class == aTParseNode class
		  and: [value class == aTParseNode value class
		  and: [value = aTParseNode value
		  and: [name = aTParseNode nameOrValue]]]]! !

!TDefineNode methodsFor: 'accessing' stamp: 'eem 2/13/2016 16:58'!
name
	^name! !

!TDefineNode methodsFor: 'accessing' stamp: 'ar 7/8/2003 11:05'!
name: aName
	name := aName! !

!TDefineNode methodsFor: 'accessing' stamp: 'ar 7/8/2003 11:08'!
nameOrValue
	^name! !

!TDefineNode methodsFor: 'C code generation' stamp: 'eem 6/23/2011 09:49'!
printOn: aStream level: level

	aStream nextPutAll: name! !

!TDefineNode methodsFor: 'accessing' stamp: 'ar 7/8/2003 11:10'!
setName: aName value: aValue
	self name: aName.
	self setValue: aValue.! !

TConstantNode subclass: #TPrintfFormatStringNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TPrintfFormatStringNode methodsFor: 'C code generation' stamp: 'eem 11/13/2017 15:02'!
emitCCodeOn: aStream level: level generator: aCodeGen
	"Transform the printf string according to the LP64/LLP64 convention in sqMemoryAccess.h."
	aStream nextPutAll: (aCodeGen cLiteralForPrintfString: value).
	comment ifNotNil:
		[aStream
			nextPutAll: ' /* ';
			nextPutAll: comment;
			nextPutAll: ' */']! !

TParseNode subclass: #TGoToNode
	instanceVariableNames: 'label'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TGoToNode methodsFor: 'C code generation' stamp: ''!
emitCCodeOn: aStream level: level generator: aCodeGen
	"Emit a C goto statement."

	aStream nextPutAll: 'goto '.
	aStream nextPutAll: label.! !

!TGoToNode methodsFor: 'C code generation' stamp: 'eem 10/30/2012 15:25'!
emitCCodeOn: aStream prependToEnd: aNodeOrNil level: lev generator: gen
	self emitCCodeOn: aStream level: lev generator: gen! !

!TGoToNode methodsFor: 'testing' stamp: 'eem 2/6/2020 14:27'!
hasEffect
	"Answer if this node has an effect on execution state (does something).
	 Statements that don't have any effect can be elided if their value is unused."
	^true! !

!TGoToNode methodsFor: 'testing' stamp: ''!
isGoTo

	^true! !

!TGoToNode methodsFor: 'accessing' stamp: ''!
label

	^label! !

!TGoToNode methodsFor: 'enumerating' stamp: 'eem 10/25/2013 14:43'!
nodesDo: aBlock parent: parent
	aBlock value: self value: parent! !

!TGoToNode methodsFor: 'printing' stamp: ''!
printOn: aStream level: level

	aStream nextPutAll: 'goto '.
	aStream nextPutAll: label.! !

!TGoToNode methodsFor: 'accessing' stamp: 'eem 4/12/2012 10:51'!
setLabel: aString
	label := aString! !

TParseNode subclass: #TInlineNode
	instanceVariableNames: 'method'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TInlineNode methodsFor: 'transformations' stamp: 'ar 7/6/2003 20:34'!
bindVariableUsesIn: aDictionary
	method := method bindVariableUsesIn: aDictionary.
! !

!TInlineNode methodsFor: 'transformations' stamp: 'eem 10/29/2012 11:39'!
bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen
	"Answer either the receiver, if it contains no references to the given variables, or a new node with the given variables rebound."
	| newMethod |
	newMethod := method bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen.
	^newMethod = method
		ifTrue: [self]
		ifFalse: [self shallowCopy
					method: newMethod;
					yourself]! !

!TInlineNode methodsFor: 'C code generation' stamp: 'ar 7/6/2003 20:19'!
emitCCodeOn: aStream level: level generator: aCodeGen
	method emitInlineOn: aStream level: level generator: aCodeGen.
! !

!TInlineNode methodsFor: 'testing' stamp: 'eem 2/6/2020 14:28'!
hasEffect
	"Answer if this node has an effect on execution state (does something).
	 Statements that don't have any effect can be elided if their value is unused."
	^method hasEffect! !

!TInlineNode methodsFor: 'testing' stamp: 'eem 7/8/2008 19:57'!
isInline
	^true! !

!TInlineNode methodsFor: 'accessing' stamp: 'ar 7/6/2003 20:08'!
method
	^method! !

!TInlineNode methodsFor: 'accessing' stamp: 'ar 7/6/2003 20:08'!
method: aTMethod
	method := aTMethod! !

!TInlineNode methodsFor: 'enumerating' stamp: 'eem 11/19/2012 12:12'!
nodesDo: aBlock
	"Apply aBlock to all nodes in the receiver.
	 N.B. This is assumed to be bottom-up, leaves first."
	method parseTree nodesDo: aBlock.
	aBlock value: self! !

!TInlineNode methodsFor: 'enumerating' stamp: 'eem 1/31/2017 10:03'!
nodesDo: aBlock parent: parent
	"Apply aBlock to all nodes in the receiver with each node's parent.
	 N.B. This is assumed to be bottom-up, leaves first."
	method parseTree nodesDo: aBlock parent: self.
	aBlock value: self value: parent! !

!TInlineNode methodsFor: 'enumerating' stamp: 'eem 3/31/2015 10:56'!
nodesDo: aBlock parent: parent unless: cautionaryBlock

	(cautionaryBlock value: self value: parent) ifTrue: [^self].
	method parseTree nodesDo: aBlock parent: self unless: cautionaryBlock.
	aBlock value: self value: parent! !

!TInlineNode methodsFor: 'enumerating' stamp: 'eem 2/27/2009 16:47'!
nodesDo: aBlock unless: cautionaryBlock

	(cautionaryBlock value: self) ifTrue: [^self].
	method parseTree nodesDo: aBlock unless: cautionaryBlock.
	aBlock value: self.! !

!TInlineNode methodsFor: 'copying' stamp: 'eem 2/9/2009 19:28'!
postCopy

	method := method copy! !

!TInlineNode methodsFor: 'printing' stamp: 'ar 7/6/2003 22:00'!
printOn: aStream level: anInteger
	method parseTree printOn: aStream level: anInteger! !

!TInlineNode methodsFor: 'transformations' stamp: 'ar 7/6/2003 21:30'!
replaceNodesIn: map
	^map at: self ifAbsent: [
		method replaceNodesIn: map.
		self]! !

!TInlineNode methodsFor: 'inlining' stamp: 'eem 7/8/2008 20:03'!
shouldExpand
	^method extractExpandCaseDirective! !

!TInlineNode methodsFor: 'type inference' stamp: 'nice 4/6/2016 23:40'!
typeOrNilFrom: aCodeGenerator in: aTMethod
	^method returnType! !

TParseNode subclass: #TLabeledCommentNode
	instanceVariableNames: 'label asmLabel'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TLabeledCommentNode methodsFor: 'accessing' stamp: 'eem 9/19/2008 14:39'!
asmLabel

	^asmLabel! !

!TLabeledCommentNode methodsFor: 'accessing' stamp: 'eem 9/19/2008 14:39'!
asmLabel: labelString

	asmLabel := labelString! !

!TLabeledCommentNode methodsFor: 'C code generation' stamp: 'eem 12/16/2015 11:31'!
emitCCodeOn: aStream level: level generator: aCodeGen
	"Emit a C comment with optional label."

	self printOptionalLabelOn: aStream.
	comment ifNotNil:
		[(aCodeGen previousCommentMarksInlining: (label isNil and: [asmLabel isNil and: [comment beginsWith: 'begin ']])) ifTrue:
			[^true].
		 aStream nextPutAll: '/* '; nextPutAll: comment; nextPutAll: ' */'.
		 aCodeGen previousCommenter: self].
	(asmLabel notNil "only output labels in the interpret function."
	 and: [aCodeGen currentMethod selector == #interpret]) ifTrue:
		[aStream crtab: level.
		 aCodeGen outputAsmLabel: asmLabel on: aStream]! !

!TLabeledCommentNode methodsFor: 'C code generation' stamp: 'eem 12/16/2015 11:00'!
emitCCommentOn: aStream level: level generator: aCodeGen
	"Override to avoid outputting comments since this is done explicitly in statement lists."! !

!TLabeledCommentNode methodsFor: 'testing' stamp: 'eem 2/6/2020 14:45'!
hasEffect
	"Answer if this node has an effect on execution state (does something).
	 Statements that don't have any effect can be elided if their value is unused."
	^true! !

!TLabeledCommentNode methodsFor: 'testing' stamp: ''!
isComment
	"Answer true if the receiver is just a comment (i.e., it has no label)."

	^label = nil! !

!TLabeledCommentNode methodsFor: 'testing' stamp: ''!
isLabel

	^true! !

!TLabeledCommentNode methodsFor: 'testing' stamp: ''!
isLeaf

	^true! !

!TLabeledCommentNode methodsFor: 'accessing' stamp: ''!
label

	^label! !

!TLabeledCommentNode methodsFor: 'testing' stamp: 'eem 12/13/2012 15:56'!
needsTrailingSemicolon
	"Answer if, when emitted as a statement (in particular in a TStmtList), the
	 receiver needs a trailing semicolon.  Comments do not.  You'd think that
	 labels do not, but we put them at the end of blocks where there needs
	 to be a null statement following the label and before the end of block."
	^self isComment not! !

!TLabeledCommentNode methodsFor: 'enumerating' stamp: 'eem 10/25/2013 14:42'!
nodesDo: aBlock parent: parent
	aBlock value: self value: parent! !

!TLabeledCommentNode methodsFor: 'printing' stamp: 'eem 7/21/2014 05:56'!
printOn: aStream level: level

	self printOptionalLabelOn: aStream.
	comment ifNotNil:
		[aStream nextPut: $".
		 aStream nextPutAll: comment.
		 aStream nextPut: $"]! !

!TLabeledCommentNode methodsFor: 'printing' stamp: 'eem 12/8/2017 13:26'!
printOptionalLabelOn: aStream

	label ifNotNil:
		[self unindentTabs: aStream.
		 aStream
			nextPutAll: label;
			nextPut: $:;
			tab]! !

!TLabeledCommentNode methodsFor: 'accessing' stamp: ''!
setComment: commentString

	label := nil.
	comment := commentString.! !

!TLabeledCommentNode methodsFor: 'accessing' stamp: ''!
setLabel: labelString

	label := labelString.! !

!TLabeledCommentNode methodsFor: 'accessing' stamp: ''!
setLabel: labelString comment: commentString

	label := labelString.
	comment := commentString.! !

!TLabeledCommentNode methodsFor: 'C code generation' stamp: 'eem 12/8/2017 14:08'!
unindentTabs: aStream
	"Remove all but one tab up to the beginning of line from the given stream if possible."

	(aStream isKindOf: ReadWriteStream) ifFalse: [ ^self ].
	[aStream position > 0] whileTrue:
		[aStream position: aStream position - 1.
		 "restore stream position if previous char was not a tab"
		 aStream peek == Character tab ifFalse:
			[^aStream next; tab]]! !

!TParseNode methodsFor: 'utilities' stamp: 'eem 12/12/2012 12:15'!
allCalls
	"Answer a collection of selectors for the messages sent in this parse tree."

	| calls |
	calls := Set new: 32.
	self nodesDo:
		[:node|
		node isSend ifTrue:
			[calls add: node selector].
		(node isConstant and: [node value isSymbol]) ifTrue:
			[calls add: node value]].
	^calls! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 10/30/2013 09:57'!
allSatisfy: aBlock
	self nodesDo: [:n| (aBlock value: n) ifFalse: [^false]].
	^true! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 10/30/2013 09:57'!
allSatisfy: aBlock unless: cautionaryBlock
	self
		nodesDo: [:n| (aBlock value: n) ifFalse: [^false]]
		unless: cautionaryBlock.
	^true! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 3/29/2013 16:58'!
anySatisfy: aBlock
	self nodesDo: [:n| (aBlock value: n) ifTrue: [^true]].
	^false! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 8/1/2013 11:02'!
anySatisfy: aBlock unless: cautionaryBlock
	self
		nodesDo: [:n| (aBlock value: n) ifTrue: [^true]]
		unless: cautionaryBlock.
	^false! !

!TParseNode methodsFor: 'transformations' stamp: 'eem 4/28/2016 17:06'!
asReturnNode
	^TReturnNode new 
		setExpression: self;
		yourself! !

!TParseNode methodsFor: 'transformations' stamp: 'jm 12/10/1998 18:13'!
bindVariableUsesIn: aDictionary
	"Bind uses of all variables in the given dictionary, but do not change variables that appear on the left-hand-side of an assignment statement."

	^ self
! !

!TParseNode methodsFor: 'transformations' stamp: 'eem 10/29/2012 11:40'!
bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen
	"Answer either the receiver, if it contains no references to the given variables, or a new node with the given variables rebound.
	 But do not change variables that appear on the left-hand-side of an assignment statement."
	^self
! !

!TParseNode methodsFor: 'transformations' stamp: ''!
bindVariablesIn: aDictionary

	^self! !

!TParseNode methodsFor: 'accessing' stamp: 'eem 8/14/2012 18:19'!
comment
	^comment! !

!TParseNode methodsFor: 'accessing' stamp: 'hg 8/14/2000 15:32'!
comment: aComment

	comment := aComment ! !

!TParseNode methodsFor: 'accessing' stamp: 'eem 1/4/2017 17:21'!
constantNumbericValueIfAtAllPossibleOrNilIn: aCCodeGen
	"This is a version of constantNumbericValueOrNil for type checking rather than code generation.
	 It aims to yield a value if at all possible."
	^self constantNumbericValueOrNil! !

!TParseNode methodsFor: 'accessing' stamp: 'eem 12/11/2014 11:38'!
constantNumbericValueOrNil

	^nil! !

!TParseNode methodsFor: 'testing' stamp: 'eem 6/23/2020 11:29'!
deny: aBooleanOrBlock
	aBooleanOrBlock value ifTrue: [AssertionFailure signal: 'Assertion failed']! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 8/13/2012 15:35'!
detect: aBlock
	^self detect: aBlock ifNone: [self error: 'Node is not in the tree']! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 8/13/2012 15:34'!
detect: aBlock ifNone: defaultBlock
	self nodesDo: [:n| (aBlock value: n) ifTrue: [^n]].
	^defaultBlock value! !

!TParseNode methodsFor: 'C code generation' stamp: '2/7/08 14:57'!
emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen
	^self emitCCodeOn: aStream level: level generator: aCodeGen! !

!TParseNode methodsFor: 'C code generation' stamp: '2/7/08 14:57'!
emitCCodeAsExpressionOn: aStream level: level generator: aCodeGen
	^self emitCCodeOn: aStream level: level generator: aCodeGen! !

!TParseNode methodsFor: 'C code generation' stamp: 'nk 4/5/2005 21:02'!
emitCCodeOn: aStream level: lev generator: gen
	self subclassResponsibility.! !

!TParseNode methodsFor: 'C code generation' stamp: 'eem 10/30/2012 16:02'!
emitCCodeOn: aStream prependToEnd: aNodeOrNil level: lev generator: gen
	aNodeOrNil ifNotNil:
		[aNodeOrNil prependCCodeOn: aStream level: lev generator: gen].
	self emitCCodeAsExpressionOn: aStream
		level: ((aNodeOrNil notNil and: [self isLeaf]) ifTrue: [0] ifFalse: [lev])
		generator: gen! !

!TParseNode methodsFor: 'C code generation' stamp: 'eem 1/11/2018 14:02'!
emitCCommentOn: aStream level: level generator: aCodeGen
	"Emit the transferred Smalltalk comments as C comments."

	((aCodeGen previousCommenter: self) == self or: [comment isNil]) ifTrue:
		[aCodeGen previousCommentMarksInlining: false.
		 ^self].
	(comment isString ifTrue: [{comment}] ifFalse: [comment]) withIndexDo:
		[:commentString :index |
		(index = 1 and: [commentString beginsWith: 'begin ']) ifTrue:
			[(aCodeGen previousCommentMarksInlining: true) ifTrue: [^true]].
		aStream crtab: level; nextPutAll: '/* '.
		(commentString findTokens: Character cr)
			do: [:line| aStream nextPutAll: line withBlanksTrimmed]
			separatedBy: [aStream crtab: level; next: 3 put: Character space].
		aStream nextPutAll: ' */'].
	 aStream cr! !

!TParseNode methodsFor: 'testing' stamp: 'eem 2/11/2009 15:24'!
endsWithReturn

	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 2/6/2020 14:24'!
hasEffect
	"Answer if this node has an effect on execution state (does something).
	 Statements that don't have any effect can be elided if their value is unused."
	^false! !

!TParseNode methodsFor: 'testing' stamp: ''!
hasExplicitReturn

	self nodesDo: [ :node |
		node isReturn ifTrue: [ ^true ].
	].
	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 11/17/2015 14:27'!
hasSideEffect
	"Answer if the parse tree rooted at this node has a side-effect or not.  By default assume it has.  Nodes that don't override."
	^true! !

!TParseNode methodsFor: 'testing' stamp: 'ikp 9/26/97 14:50'!
isAssertion
	^false! !

!TParseNode methodsFor: 'testing' stamp: 'ikp 9/26/97 14:50'!
isAssignment

	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 8/21/2021 19:48'!
isBrace
	^false! !

!TParseNode methodsFor: 'testing' stamp: ''!
isCaseStmt

	^false! !

!TParseNode methodsFor: 'testing' stamp: ''!
isComment

	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 4/27/2016 18:46'!
isConditionalSend
	"Answer if the receiver is a send of any of the conditionals, ifTrue: ifTrue:ifFalse: et al"
	^false! !

!TParseNode methodsFor: 'testing' stamp: ''!
isConstant

	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 1/21/2016 16:11'!
isDefine

	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 7/6/2021 17:48'!
isEffectFree
	^false! !

!TParseNode methodsFor: 'testing' stamp: ''!
isGoTo

	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 7/8/2008 19:57'!
isInline
	^false! !

!TParseNode methodsFor: 'testing' stamp: ''!
isLabel

	^false! !

!TParseNode methodsFor: 'testing' stamp: ''!
isLeaf
	"Answer true if the receiver is a variable or a constant node."

	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 6/24/2012 16:18'!
isLiteralArrayDeclaration
	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 11/15/2019 14:04'!
isLiteralBlock
	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 10/24/2013 16:29'!
isNonNullCCode
	"overridden in TSendNode"
	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 7/7/2021 11:41'!
isNullCCode
	"overridden in TSendNode"
	^false! !

!TParseNode methodsFor: 'testing' stamp: ''!
isReturn

	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 8/8/2008 11:57'!
isReturningIf
	^false! !

!TParseNode methodsFor: 'comparing' stamp: 'eem 6/23/2020 11:19'!
isSameAs: aTParseNode
	"Answer if the ParseTree rooted at this node is the same as aTParseNode.
	 By default answer false and have subclasses override as appropriate."
	^self == aTParseNode! !

!TParseNode methodsFor: 'testing' stamp: ''!
isSend

	^false! !

!TParseNode methodsFor: 'testing' stamp: ''!
isStmtList

	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 7/7/2021 11:41'!
isStringishCCode
	"overridden in TSendNode"
	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 2/10/2009 07:58'!
isSwitch
	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 7/26/2014 07:49'!
isTMethod
	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 2/11/2009 17:12'!
isTParseNode
	^true! !

!TParseNode methodsFor: 'testing' stamp: ''!
isVariable

	^false! !

!TParseNode methodsFor: 'accessing' stamp: 'acg 12/17/1999 07:21'!
nameOrValue

	self error: 'object is neither variable nor constant'! !

!TParseNode methodsFor: 'testing' stamp: 'eem 12/13/2012 15:14'!
needsTrailingSemicolon
	"Answer if, when emitted as a statement (in particular in a TStmtList), the
	 receiver needs a trailing semicolon.  Subclasses redefine as appropriate."
	^true! !

!TParseNode methodsFor: 'utilities' stamp: ''!
nodeCount
	"Answer the number of nodes in this parseTree (a rough measure of its size)."

	| cnt |
	cnt := 0.
	self nodesDo: [ :n | cnt := cnt + 1 ].
	^cnt! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 11/19/2012 12:12'!
nodesDo: aBlock
	"Apply aBlock to all nodes in the receiver.
	 N.B. This is assumed to be bottom-up, leaves first."
	aBlock value: self! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 10/13/2013 17:01'!
nodesDo: aBlock parent: parent
	"Apply aBlock to all nodes in the receiver with each node's parent.
	 N.B. This is assumed to be bottom-up, leaves first."
	self subclassResponsibility! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 3/31/2015 10:53'!
nodesDo: aBlock parent: parent unless: cautionaryBlock
	"Evaluate aBlock for all nodes in  the tree except those for which cautionaryBlock
	 answers true or are children of those for which cautionaryBlock answers true."
	(cautionaryBlock value: self value: parent) ifFalse:
		[aBlock value: self value: parent]! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 8/1/2013 11:02'!
nodesDo: aBlock unless: cautionaryBlock
	"Evaluate aBlock for all nodes in  the tree except those for which cautionaryBlock
	 answers true or are children of those for which cautionaryBlock answers true."
	(cautionaryBlock value: self) ifFalse:
		[aBlock value: self]! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 11/17/2013 13:00'!
nodesSelect: selectBlock thenCollect: collectBlock
	| result |
	result := OrderedCollection new.
	self nodesDo:
		[:node|
		(selectBlock value: node) ifTrue:
			[result addLast: (collectBlock value: node)]].
	^result! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 10/13/2013 17:01'!
nodesWithParentsDo: aBlock
	"Apply aBlock to all nodes in the receiver with each node's parent.
	 N.B. This is assumed to be bottom-up, leaves first."
	self nodesDo: aBlock parent: nil! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 3/31/2015 10:18'!
nodesWithParentsDo: aBlock unless: cautionaryBlock
	"Evaluate aBlock for all nodes in the tree and their parent except those for which
	 cautionaryBlock answers true or are children of those for which cautionaryBlock
	 answers true. Cautionary block is invoked with node and parent.  parent may be nil."
	self nodesDo: aBlock parent: nil unless: cautionaryBlock! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 3/29/2013 16:58'!
noneSatisfy: aBlock
	self nodesDo: [:n| (aBlock value: n) ifTrue: [^false]].
	^true! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 8/1/2013 10:59'!
noneSatisfy: aBlock unless: cautionaryBlock
	self
		nodesDo: [:n| (aBlock value: n) ifTrue: [^false]]
		unless: cautionaryBlock.
	^true! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 11/11/2014 11:50'!
occurrencesOf: aNode
	| count |
	count := 0.
	self nodesDo:
		[:node|
		node = aNode ifTrue: [count := count + 1]].
	^count! !

!TParseNode methodsFor: 'accessing' stamp: 'eem 11/15/2019 13:38'!
parentIn: aParseTree
	aParseTree nodesWithParentsDo:
		[:node :parent|
		self == node ifTrue: [^parent]].
	^nil! !

!TParseNode methodsFor: 'printing' stamp: ''!
printOn: aStream 
	"Append a description of the receiver onto the given stream."

	self printOn: aStream level: 0.! !

!TParseNode methodsFor: 'printing' stamp: ''!
printOn: aStream level: anInteger 
	"Typically overridden. If control actually gets here, avoid recursion loop by sending to super."

	super printOn: aStream.! !

!TParseNode methodsFor: 'transformations' stamp: 'ikp 9/26/97 14:50'!
removeAssertions
	"default: do nothing"! !

!TParseNode methodsFor: 'transformations' stamp: ''!
replaceNodesIn: aDictionary

	^aDictionary at: self ifAbsent: [self]! !

!TParseNode methodsFor: 'enumerating' stamp: 'eem 4/30/2016 10:28'!
select: aBlock
	| nodes |
	nodes := OrderedCollection new.
	self nodesDo: [:n| (aBlock value: n) ifTrue: [nodes add: n]].
	^nodes! !

!TParseNode methodsFor: 'testing' stamp: 'eem 11/1/2013 17:21'!
stream: aStream endsWithAnyOf: characters
	"Answer if the given stream ends in any of the characters, ignoring whitespace."
	| pos ch popped |
	pos := aStream position.
	aStream class = WriteStream ifTrue: "i.e. nested in StreamContents"
		[[pos > 0] whileTrue:
			[ch := aStream originalContents at: pos.
			 ch isSeparator ifFalse:
				[^characters includes: ch].
			 pos := pos - 1].
		 ^false].

	popped := OrderedCollection new.
	[pos > 0] whileTrue:
		[ch := popped addFirst: (aStream position: pos - 1; peek).
		 ch isSeparator ifFalse:
			[popped notEmpty ifTrue: "it almost always is"
				["*don't* use nextPutAll: to work around a bug in UTF8TextConverter
				  >>nextPut:toStream: which avoids line-end conversion; sigh."
				 popped do:
					[:c| aStream nextPut: c]].
			 ^characters includes: ch].
		 pos := pos - 1].
	"*don't* use nextPutAll: to work around a bug in UTF8TextConverter>>
	 nextPut:toStream: which avoids line-end conversion; sigh."
	popped do:
		[:c| aStream nextPut: c].
	^false! !

!TParseNode methodsFor: 'testing' stamp: 'eem 10/24/2013 13:02'!
structTargetKindIn: aCodeGen
	"Answer if the recever evaluates to a struct or struct pointer
	 and hence can be dereferenced using . or ->.  Answer any of
	 #struct #pointer or nil"
	^nil! !

!TParseNode methodsFor: 'type inference' stamp: 'nice 4/6/2016 23:41'!
typeFrom: aCodeGenerator in: aTMethod
	"This is the default type in case of doubt"
	^(self typeOrNilFrom: aCodeGenerator in: aTMethod) ifNil: [#sqInt]! !

!TParseNode methodsFor: 'type inference' stamp: 'nice 4/6/2016 23:40'!
typeOrNilFrom: aCodeGenerator in: aTMethod
	"When we don't know, it's nil"
	^nil! !

!TParseNode methodsFor: 'accessing' stamp: 'eem 11/12/2013 14:22'!
variableNameOrNil
	"Overridden in TAssignmentNode & TVariableNode to answer their variable name."
	^nil! !

TParseNode subclass: #TReturnNode
	instanceVariableNames: 'expression'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TReturnNode methodsFor: 'transformations' stamp: 'eem 4/28/2016 17:06'!
asReturnNode
	^self! !

!TReturnNode methodsFor: 'transformations' stamp: 'jm 12/10/1998 18:26'!
bindVariableUsesIn: aDictionary

	expression := expression bindVariableUsesIn: aDictionary.! !

!TReturnNode methodsFor: 'transformations' stamp: 'eem 10/29/2012 12:42'!
bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen
	"Answer either the receiver, if it contains no references to the given variables, or a new node with the given variables rebound."
	| newExpression |
	newExpression := expression bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen.
	^newExpression = expression
		ifTrue: [self]
		ifFalse:
			[self shallowCopy
				setExpression: newExpression;
				yourself]! !

!TReturnNode methodsFor: 'transformations' stamp: ''!
bindVariablesIn: aDictionary

	expression := expression bindVariablesIn: aDictionary.! !

!TReturnNode methodsFor: 'transformations' stamp: 'eem 4/28/2016 11:03'!
copyWithoutReturn
	^expression! !

!TReturnNode methodsFor: 'C code generation' stamp: 'eem 12/13/2015 17:05'!
emitCCodeOn: aStream level: level generator: aCodeGen

	(expression isSwitch
	 or: [expression isCaseStmt]) ifTrue:
		[^expression emitCCodeOn: aStream addToEndOfCases: self level: level generator: aCodeGen].

	(expression isSend and: [expression isValueExpansion]) ifTrue:
		[^self emitValueExpansionOn: aStream level: level generator: aCodeGen].

	'void' = aCodeGen currentMethod returnType ifTrue: "If the function is void, don't say 'return x' instead say ' x; return' "
		[expression isLeaf ifFalse:
			[expression emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen.	
			 aStream nextPut: $;; crtab: level].
		 aStream nextPutAll: 'return'.
		 ^self].
	aStream nextPutAll: 'return '.
	expression
		emitCCodeAsArgumentOn: aStream
		level: (expression isStmtList ifTrue: [level + 1] ifFalse: [level])
		generator: aCodeGen! !

!TReturnNode methodsFor: 'C code generation' stamp: 'eem 3/17/2020 18:25'!
emitValueExpansionOn: aStream level: level generator: aCodeGen
	| stmtList lastStmt copiedStatements |
	self assert: (expression isSend and: [expression isValueExpansion]).
	stmtList := expression receiver.
	lastStmt := stmtList statements last.
	lastStmt isReturn ifTrue:
		[^expression emitCCodeOn: aStream level: level generator: aCodeGen].
	copiedStatements := stmtList copy.
	copiedStatements statements
		at: stmtList statements size
		put: (TReturnNode new setExpression: lastStmt).
	expression copy
		receiver: copiedStatements;
		emitCCodeOn: aStream level: level generator: aCodeGen! !

!TReturnNode methodsFor: 'testing' stamp: 'eem 2/11/2009 15:24'!
endsWithReturn

	^true! !

!TReturnNode methodsFor: 'accessing' stamp: ''!
expression

	^expression! !

!TReturnNode methodsFor: 'testing' stamp: 'eem 2/6/2020 14:28'!
hasEffect
	"Answer if this node has an effect on execution state (does something).
	 Statements that don't have any effect can be elided if their value is unused."
	^true! !

!TReturnNode methodsFor: 'testing' stamp: ''!
isReturn

	^true! !

!TReturnNode methodsFor: 'comparing' stamp: 'eem 6/23/2020 11:27'!
isSameAs: aTParseNode
	^self == aTParseNode
	 or: [aTParseNode isReturn
		 and: [expression isSameAs: aTParseNode expression]]! !

!TReturnNode methodsFor: 'enumerating' stamp: 'eem 11/19/2012 12:12'!
nodesDo: aBlock
	"Apply aBlock to all nodes in the receiver.
	 N.B. This is assumed to be bottom-up, leaves first."
	expression nodesDo: aBlock.
	aBlock value: self! !

!TReturnNode methodsFor: 'enumerating' stamp: 'eem 10/13/2013 17:02'!
nodesDo: aBlock parent: parent
	"Apply aBlock to all nodes in the receiver with each node's parent.
	 N.B. This is assumed to be bottom-up, leaves first."
	expression nodesDo: aBlock parent: self.
	aBlock value: self value: parent! !

!TReturnNode methodsFor: 'enumerating' stamp: 'eem 3/31/2015 10:55'!
nodesDo: aBlock parent: parent unless: cautionaryBlock

	(cautionaryBlock value: self value: parent) ifTrue: [^self].
	expression nodesDo: aBlock parent: self unless: cautionaryBlock.
	aBlock value: self value: parent! !

!TReturnNode methodsFor: 'enumerating' stamp: 'eem 2/27/2009 16:47'!
nodesDo: aBlock unless: cautionaryBlock

	(cautionaryBlock value: self) ifTrue: [^self].
	expression nodesDo: aBlock unless: cautionaryBlock.
	aBlock value: self.! !

!TReturnNode methodsFor: 'copying' stamp: 'eem 2/9/2009 19:25'!
postCopy

	expression := expression copy! !

!TReturnNode methodsFor: 'C code generation' stamp: 'eem 2/11/2009 14:45'!
prependCCodeOn: aStream level: level generator: aCCodeGenerator 
	aStream nextPutAll: 'return '! !

!TReturnNode methodsFor: 'printing' stamp: ''!
printOn: aStream level: level

	aStream nextPut: $^.
	expression printOn: aStream level: level.! !

!TReturnNode methodsFor: 'transformations' stamp: 'ikp 9/26/97 14:50'!
removeAssertions
	expression removeAssertions! !

!TReturnNode methodsFor: 'transformations' stamp: ''!
replaceNodesIn: aDictionary

	^aDictionary at: self ifAbsent: [
		expression := expression replaceNodesIn: aDictionary.
		self]! !

!TReturnNode methodsFor: 'accessing' stamp: ''!
setExpression: aNode

	expression := aNode.! !

!TReturnNode methodsFor: 'type inference' stamp: 'nice 4/6/2016 23:54'!
typeOrNilFrom: aCodeGenerator in: aTMethod
	^expression typeOrNilFrom: aCodeGenerator in: aTMethod! !

TParseNode subclass: #TSendNode
	instanceVariableNames: 'selector receiver arguments isBuiltinOperator'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TSendNode methodsFor: 'accessing' stamp: ''!
args

	^arguments! !

!TSendNode methodsFor: 'private' stamp: 'eem 10/29/2012 12:48'!
arguments: aSequence
	arguments := aSequence! !

!TSendNode methodsFor: 'inlining support' stamp: 'eem 2/11/2009 17:03'!
argumentsForInliningCodeGenerator: aCodeGen
	^(self shouldIncludeReceiverAsFirstArgument: aCodeGen)
		ifTrue: [{receiver}, arguments]
		ifFalse: [arguments]! !

!TSendNode methodsFor: 'inlining support' stamp: 'eem 4/13/2014 21:42'!
asTransformedConstantPerform
	"If the receiver is a send of perform: with a constant selector,
	 answer a send node that elides the perform:, otherwise answer nil."
	^(selector isSymbol
	  and: [(selector beginsWith: #perform:)
	  and: [(selector keywords allSatisfy: [:kw| #('perform:' 'with:') includes: kw])
	  and: [arguments first isConstant
	  and: [arguments first value isSymbol
	  and: [arguments first value numArgs + 1 = arguments size]]]]]) ifTrue:
		[TSendNode new
			setSelector: arguments first value
			receiver: receiver
			arguments: arguments allButFirst]! !

!TSendNode methodsFor: 'transformations' stamp: 'jm 12/10/1998 18:26'!
bindVariableUsesIn: aDictionary

	receiver := receiver bindVariableUsesIn: aDictionary.
	arguments := arguments collect: [ :a | a bindVariableUsesIn: aDictionary ].! !

!TSendNode methodsFor: 'transformations' stamp: 'eem 9/11/2018 14:48'!
bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen
	"Answer either the receiver, if it contains no references to the given variables, or a new node with the given variables rebound.
	 Attempt to constant-fold and answer a constant node commented with the original expression.
	 Commenting with the original expression is important because it allows us to detect shared cases.
	 e.g. currentBytecode bitAnd: 15 is the same in case 1 and case 17, but '1 /* 1 bitAnd: 15 */' differs
	 from '1 /* 17 bitAnd: 15 */', whereas '1 /* currentBytecode bitAnd: 15 */' doesn't change."
	| newReceiver newArguments |
	"Constant-fold shiftForWord, but not BytesPerWord"
	((VMBasicConstants mostBasicConstantSelectors includes: selector)
	 and: [(codeGen isBuiltinSelector: selector) not]) ifTrue:
		[codeGen isConstantNode: self valueInto:
			[:val|
			 ^TConstantNode new
				setValue: val;
				yourself]].
	newReceiver := receiver bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen.
	newArguments := arguments collect: [:a| a bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen].
	(newReceiver = receiver
	 and: [newArguments = arguments]) ifTrue:
		[^self].
	(constantFold
	 and: [newReceiver isConstant and: [newReceiver value isInteger
	 and: [(newArguments allSatisfy: [:ea| ea isConstant and: [ea value isInteger]])
	 and: [codeGen isBuiltinSelector: selector]]]]) ifTrue:
		[| value |
		value := [newReceiver value perform: selector withArguments: (newArguments collect: [:ea| ea value])]
					on: Error
					do: [:ea| nil].
		 (value isInteger
		 or: [value == true
		 or: [value == false]]) ifTrue:
			[^TConstantNode new
				setValue: value;
				"We assume Message prints its keywords and arguments interleaved.
				 e.g. that (Message selector: #between:and: arguments: #(0 1)) printString = 'between: 0 and: 1'"
				comment: (receiver isLeaf
								ifTrue: [receiver printString]
								ifFalse: ['(', receiver printString, ')']),
							' ',
							(Message selector: selector arguments: (arguments collect: [:ea| ea value])) printString;
				yourself]].
	^self shallowCopy
		receiver: newReceiver;
		arguments: newArguments;
		yourself
		! !

!TSendNode methodsFor: 'transformations' stamp: ''!
bindVariablesIn: aDictionary

	receiver := receiver bindVariablesIn: aDictionary.
	arguments := arguments collect: [ :a | a bindVariablesIn: aDictionary ].! !

!TSendNode methodsFor: 'accessing' stamp: 'eem 1/4/2017 17:30'!
constantNumbericValueIfAtAllPossibleOrNilIn: aCCodeGen
	"This is a version of constantNumbericValueOrNil for type checking rather than code generation.
	 It aims to yield a value if at all possible."
	(#(* // + - << >> bitAnd: bitOr: bitShift:) includes: selector) ifTrue:
		[(receiver constantNumbericValueIfAtAllPossibleOrNilIn: aCCodeGen) ifNotNil:
			[:rval|
			(arguments first constantNumbericValueIfAtAllPossibleOrNilIn: aCCodeGen) ifNotNil:
				[:aval|
				^rval perform: selector with: aval]]].
	^(aCCodeGen anyMethodNamed: selector) ifNotNil:
		[:m|
		 m isReturnConstant ifTrue:
			[m statements last expression constantNumbericValueIfAtAllPossibleOrNilIn: aCCodeGen]]! !

!TSendNode methodsFor: 'accessing' stamp: 'eem 1/29/2020 15:37'!
constantNumbericValueOrNil
	(#(* // + - << >>) includes: selector) ifTrue:
		[receiver constantNumbericValueOrNil ifNotNil:
			[:rval|
			arguments first constantNumbericValueOrNil ifNotNil:
				[:aval|
				^rval perform: selector with: aval]]].
	^nil! !

!TSendNode methodsFor: 'transformations' stamp: 'eem 4/28/2016 12:57'!
copyWithoutReturn
	self assert: self endsWithReturn.
	^self class new
		setSelector: selector
			receiver: receiver
				arguments: (arguments collect:
								[:arg|
								 arg endsWithReturn
									ifTrue: [arg copyWithoutReturn]
									ifFalse: [arg]]);
		yourself! !

!TSendNode methodsFor: 'C code generation' stamp: 'eem 12/11/2010 12:39'!
emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen
	"Emit the receiver in a form that can be passed as an argument."

	"If the selector is a built-in construct, translate it and return"
	(aCodeGen emitBuiltinConstructAsArgumentFor: self on: aStream level: level) ifFalse:
		["If it is a pointer dereference generate it"
		(self emitCCodeAsFieldReferenceOn: aStream level: level generator: aCodeGen) ifFalse:
			["Otherwise generate the vanilla C function call."
			 self emitCCodeAsFunctionCallOn: aStream level: level generator: aCodeGen]]! !

!TSendNode methodsFor: 'C code generation' stamp: '2/7/08 14:57'!
emitCCodeAsExpressionOn: aStream level: level generator: aCodeGen
	^self emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen! !

!TSendNode methodsFor: 'C code generation' stamp: 'eem 8/28/2020 20:15'!
emitCCodeAsFieldReferenceOn: aStream level: level generator: aCodeGen
	"If appropriate, translate this message send as a pointer dereference"

	| parenCount |
	(aCodeGen isStructSend: self) ifFalse:
		[^false].
	(aCodeGen shouldGenerateAsInterpreterProxySend: self) ifTrue:
		[^false].

	parenCount := receiver isSend ifTrue: [2] ifFalse: [1].
	aStream next: parenCount put: $(.
	receiver  emitCCodeAsExpressionOn: aStream level: 0 generator: aCodeGen.
	parenCount > 1 ifTrue:
		[aStream nextPut: $)].
	(receiver structTargetKindIn: aCodeGen) caseOf: {
		[#pointer] -> [aStream nextPut: $-; nextPut: $>].
		[#struct] -> [aStream nextPut: $.] }.
	aStream nextPutAll: (aCodeGen cFunctionNameFor: selector).
	arguments isEmpty ifFalse:
		[self assert: arguments size = 1.
		 aStream nextPutAll: ' = '.
		 arguments first emitCCodeAsExpressionOn: aStream level: level generator: aCodeGen].
	aStream nextPut: $).
	^true! !

!TSendNode methodsFor: 'C code generation' stamp: 'eem 1/24/2017 18:25'!
emitCCodeAsFunctionCallOn: aStream level: level generator: aCodeGen
	"Translate this message send into a C function call"

	selector == #break ifTrue:
		[aStream nextPutAll: '/* send of break elided */'.
		 ^self].

	"Special case for pluggable modules. Replace messages to interpreterProxy
	 by interpreterProxy->message(..) if the message is not builtin"
	(aCodeGen shouldGenerateAsInterpreterProxySend: self) ifTrue:
		[(aCodeGen noteUsedPluginFunction: selector) ifTrue:
			[aStream nextPutAll: 'interpreterProxy->']].

	aStream nextPutAll: (aCodeGen cFunctionNameFor: selector); nextPut: $(.

	"Only include the receiver as the first argument in certain cases.
	 The receiver is always included if it is an expression.
	 If it is a variable:
		 If the vmClass says it is an implicit variable, don't include it.
		 If the variable is 'self' and the method being called is not in
		 the method set (i.e. it is some external code), don't include it."
	(self shouldExcludeReceiverAsFirstArgument: aCodeGen) ifFalse:
		[(receiver structTargetKindIn: aCodeGen) == #struct ifTrue:
			[aStream nextPut: $&].
		 receiver emitCCodeOn: aStream level: level generator: aCodeGen.
		 arguments isEmpty ifFalse:
			[aStream nextPutAll: ', ']].
	arguments
		do: [ :arg| arg emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen]
		separatedBy: [aStream nextPut: $,; space].
	aStream nextPut: $)! !

!TSendNode methodsFor: 'C code generation' stamp: 'eem 12/11/2010 12:39'!
emitCCodeOn: aStream level: level generator: aCodeGen
	"Emit the receiver as a statement."

	"If the selector is a built-in construct, translate it and return"
	(aCodeGen emitBuiltinConstructFor: self on: aStream level: level) ifFalse:
		["If it is a pointer dereference generate it"
		(self emitCCodeAsFieldReferenceOn: aStream level: level generator: aCodeGen) ifFalse:
			["Otherwise generate the vanilla C function call."
			 self emitCCodeAsFunctionCallOn: aStream level: level generator: aCodeGen]]! !

!TSendNode methodsFor: 'testing' stamp: 'eem 2/11/2009 15:25'!
endsWithReturn
	^self isReturningIf! !

!TSendNode methodsFor: 'testing' stamp: 'eem 7/7/2021 10:52'!
hasEffect
	"Answer if this node has an effect on execution state (does something).
	 Statements that don't have any effect can be elided if their value is unused."
	selector == #not ifTrue:
		[^receiver hasEffect].
	selector == #cCode:inSmalltalk: ifTrue:
		[^(arguments first isConstant and: [arguments first value isString and: [arguments first value notEmpty]])
			or: [arguments first hasEffect]].
	self isBinaryArithmeticOrConditional ifTrue:
		[^receiver hasEffect or: [arguments first hasEffect]].
	self isUnaryCast ifTrue:
		[^receiver hasEffect].
	self isBinaryCast ifTrue:
		[^arguments first hasEffect].
	^true! !

!TSendNode methodsFor: 'testing' stamp: 'eem 7/6/2021 18:07'!
hasSideEffect
	"Answer if the parse tree rooted at this node has a side-effect or not.
	 We assume most sends do, but filter-out arithmetic and certain reads."
	(#(	baseHeaderSize bytesPerOop bytesPerWord logBytesPerOop
		numSmallIntegerBits numSmallIntegerTagBits numTagBits
		shiftForWord tagMask wordSize) includes: selector) ifTrue:
		[^false].
	(#(#+ #- #* #/ #// #\\ #= #== #~= #~~ << >>
		"These two important for Spur is:instanceOf:compactClassIndex: et al"
		longAt: fetchPointer:ofObject: splObj:) includes: selector) ifFalse:
		[^true].
	^receiver hasSideEffect or: [arguments first hasSideEffect]! !

!TSendNode methodsFor: 'testing' stamp: 'ikp 9/26/97 14:50'!
isAssertion
	^(selector beginsWith: 'assert') or: [selector beginsWith: 'verify']! !

!TSendNode methodsFor: 'testing' stamp: 'eem 7/6/2021 18:37'!
isBinaryArithmeticOrConditional
	^arguments size = 2
	 and: [#(+ - * / // \\ mod: quo:
			<< >> bitShift: bitAnd: bitOr: bitXor:
			= == ~= ~~ > >= < <=
			& | and: or:) includes: selector]! !

!TSendNode methodsFor: 'testing' stamp: 'eem 7/6/2021 18:22'!
isBinaryCast
	^#(cCoerce:to: cCoerceSimple:to:) includes: selector! !

!TSendNode methodsFor: 'testing' stamp: ''!
isBuiltinOperator

	^ isBuiltinOperator! !

!TSendNode methodsFor: 'testing' stamp: ''!
isBuiltinOperator: builtinFlag

	isBuiltinOperator := builtinFlag.! !

!TSendNode methodsFor: 'testing' stamp: 'eem 7/6/2021 18:21'!
isCast
	^#(cCoerce:to: cCoerceSimple:to: asUnsignedInteger asVoidPointer) includes: selector! !

!TSendNode methodsFor: 'testing' stamp: 'eem 4/27/2016 18:47'!
isConditionalSend
	"Answer if the receiver is a send of any of the conditionals, ifTrue: ifTrue:ifFalse: et al"
	^#(	ifTrue:ifFalse: ifFalse:ifTrue: ifTrue: ifFalse:
		ifNil:ifNotNil: ifNotNil:ifNil: ifNil: ifNotNil) includes: selector! !

!TSendNode methodsFor: 'testing' stamp: 'eem 7/7/2021 17:57'!
isEffectFree
	^self hasEffect not! !

!TSendNode methodsFor: 'testing' stamp: 'eem 6/24/2012 16:22'!
isLiteralArrayDeclaration
	^selector == #cCoerce:to:
	  and: [arguments first isConstant
	  and: [arguments first value isLiteral
	  and: [arguments first value isCollection]]]! !

!TSendNode methodsFor: 'testing' stamp: 'eem 7/7/2021 11:08'!
isNonNullCCode
	^(#(cCode: cCode:inSmalltalk:) includes: selector)
	   and: [arguments first isConstant
			ifTrue: [arguments first value isString
					and: [arguments first value notEmpty]]
			ifFalse: [arguments first hasEffect]]! !

!TSendNode methodsFor: 'testing' stamp: 'eem 7/7/2021 17:57'!
isNullCCode
	| node |
	(#(cCode: cCode:inSmalltalk:) includes: selector) ifFalse:
		[^false].
	"all of cCode: nil ..., cCode: []..., cCode: '' are null"
	node := arguments first.
	node isConstant ifTrue:
		[^(node value isString
		   and: [node value notEmpty]) not].
	^node hasEffect not! !

!TSendNode methodsFor: 'testing' stamp: 'eem 7/6/2021 16:04'!
isPreprocessorDirective
	^#(cPreprocessorDirective: cppIf:ifTrue: cppIf:ifFalse: cppIf:ifTrue:ifFalse: cppIf:ifTrue:cppIf:ifTrue:ifFalse:) includes: selector! !

!TSendNode methodsFor: 'testing' stamp: 'eem 7/9/2021 18:51'!
isReturningIf
	^((#(ifTrue:ifFalse: ifFalse:ifTrue: ifNil:ifNotNil: ifNotNil:ifNil:) includes: selector)
	   and: [arguments allSatisfy: [:arg| arg endsWithReturn]])
		or: [(#(cppIf:ifTrue:ifFalse: cppIf:ifTrue:cppIf:ifTrue:ifFalse:) includes: selector)
			and: [| keywords |
				keywords := selector keywords.
				keywords first = 'cppIf:'
				and: [(1 to: arguments size) allSatisfy:
							[:i| (keywords at: i) = 'cppIf:' or: [(arguments at: i) endsWithReturn]]]]]! !

!TSendNode methodsFor: 'comparing' stamp: 'eem 6/23/2020 11:26'!
isSameAs: aTParseNode
	self == aTParseNode ifTrue: [^true].
	(aTParseNode isSend
	 and: [selector == aTParseNode selector
	 and: [receiver isSameAs: aTParseNode receiver]]) ifFalse:
		[^false].
	arguments with: aTParseNode args do:
		[:a :b|
		(a isSameAs: b) ifFalse:
			[^false]].
	^true! !

!TSendNode methodsFor: 'C code generation' stamp: 'eem 10/24/2013 12:57'!
isSelfReference: varNode in: aCodeGen
	^(varNode name beginsWith: 'self')
	  and: [varNode name = 'self' or: [varNode name beginsWith: 'self_in_']]! !

!TSendNode methodsFor: 'testing' stamp: ''!
isSend

	^true! !

!TSendNode methodsFor: 'testing' stamp: 'eem 7/7/2021 11:10'!
isStringishCCode
	^(#(cCode: cCode:inSmalltalk:) includes: selector)
	   and: [arguments first isConstant
	   and: [arguments first value isString
	   and: [arguments first value notEmpty]]]! !

!TSendNode methodsFor: 'C code generation' stamp: 'eem 10/24/2013 13:01'!
isStructReference: varNode in: aCodeGen
	^(varNode structTargetKindIn: aCodeGen) notNil! !

!TSendNode methodsFor: 'testing' stamp: 'eem 7/6/2021 18:22'!
isUnaryCast
	^#(asUnsignedInteger asVoidPointer) includes: selector! !

!TSendNode methodsFor: 'testing' stamp: 'eem 6/23/2011 16:48'!
isValueExpansion
	"Answer if the receiver is of the form literalBlock value[: arg]"
	^receiver isStmtList
	 and: [selector keywords allSatisfy: [:k| #('value' 'value:') includes: k]]! !

!TSendNode methodsFor: 'testing' stamp: 'eem 7/25/2020 13:40'!
needsTrailingSemicolon
	"Answer if, when emitted as a statement (in particular in a TStmtList), the
	 receiver needs a trailing semicolon.  Preprocessor directives and special expansions do not."
	^(self isPreprocessorDirective
	   or: [self isValueExpansion
	   or: [self isConditionalSend
	   or: [selector == #expandDereferenceInterpreterProxyFunctionTable
	   or: [selector == #remapOop:in:]]]]) not! !

!TSendNode methodsFor: 'enumerating' stamp: 'eem 11/19/2012 12:12'!
nodesDo: aBlock
	"Apply aBlock to all nodes in the receiver.
	 N.B. This is assumed to be bottom-up, leaves first."
	receiver nodesDo: aBlock.
	arguments do: [ :arg | arg nodesDo: aBlock ].
	aBlock value: self! !

!TSendNode methodsFor: 'enumerating' stamp: 'eem 10/13/2013 17:03'!
nodesDo: aBlock parent: parent
	"Apply aBlock to all nodes in the receiver with each node's parent.
	 N.B. This is assumed to be bottom-up, leaves first."
	receiver nodesDo: aBlock parent: self.
	arguments do: [:arg| arg nodesDo: aBlock parent: self].
	aBlock value: self value: parent! !

!TSendNode methodsFor: 'enumerating' stamp: 'eem 3/31/2015 10:54'!
nodesDo: aBlock parent: parent unless: cautionaryBlock

	(cautionaryBlock value: self value: parent) ifTrue: [^self].
	receiver nodesDo: aBlock parent: self unless: cautionaryBlock.
	arguments do: [ :arg | arg nodesDo: aBlock parent: self unless: cautionaryBlock].
	aBlock value: self value: parent! !

!TSendNode methodsFor: 'enumerating' stamp: 'eem 2/27/2009 21:55'!
nodesDo: aBlock unless: cautionaryBlock

	(cautionaryBlock value: self) ifTrue: [^self].
	receiver nodesDo: aBlock unless: cautionaryBlock.
	arguments do: [ :arg | arg nodesDo: aBlock unless: cautionaryBlock].
	aBlock value: self.! !

!TSendNode methodsFor: 'accessing' stamp: 'eem 3/19/2014 17:08'!
numArgs
	^arguments size! !

!TSendNode methodsFor: 'copying' stamp: 'eem 2/9/2009 19:25'!
postCopy

	receiver := receiver copy.
	arguments := (arguments collect: [ :arg | arg copy ])! !

!TSendNode methodsFor: 'printing' stamp: 'eem 9/9/2018 18:07'!
printOn: aStream level: level
	| possiblyParenthesize |
	possiblyParenthesize :=
		[:node :newLevel|
		node
			ifNil: [aStream print: node]
			ifNotNil: 
				[(node isSend
				  and: [node selector precedence >= 3]) ifTrue:
					[aStream nextPut: $(].
				node printOn: aStream level: newLevel.
				(node isSend
				 and: [node selector precedence >= 3]) ifTrue:
					[aStream nextPut: $)]]].

	possiblyParenthesize value: receiver value: level.
	arguments size = 0 ifTrue:
		[aStream space; nextPutAll: selector.
		^self].
	selector keywords with: (arguments first: selector numArgs) do:
		[:keyword :arg |
		aStream space; nextPutAll: keyword; space.
		possiblyParenthesize value: arg value: level + 1]! !

!TSendNode methodsFor: 'accessing' stamp: ''!
receiver

	^receiver! !

!TSendNode methodsFor: 'private' stamp: ''!
receiver: aNode

	receiver := aNode.! !

!TSendNode methodsFor: 'transformations' stamp: 'ikp 9/26/97 14:50'!
removeAssertions
	receiver removeAssertions.
	arguments do: [:arg | arg removeAssertions].! !

!TSendNode methodsFor: 'transformations' stamp: ''!
replaceNodesIn: aDictionary

	^aDictionary at: self ifAbsent: [
		receiver := receiver replaceNodesIn: aDictionary.
		arguments := arguments collect: [ :a | a replaceNodesIn: aDictionary ].
		self]! !

!TSendNode methodsFor: 'accessing' stamp: ''!
selector

	^selector! !

!TSendNode methodsFor: 'accessing' stamp: 'eem 6/6/2018 19:09'!
selector: aSymbol
	selector := aSymbol! !

!TSendNode methodsFor: 'accessing' stamp: ''!
setSelector: aSymbol receiver: rcvrNode arguments: argList

	selector := aSymbol.
	receiver := rcvrNode.
	arguments := argList asArray.
	isBuiltinOperator := false.! !

!TSendNode methodsFor: 'accessing' stamp: ''!
setSelector: aSymbol receiver: rcvrNode arguments: argList isBuiltInOp: builtinFlag

	selector := aSymbol.
	receiver := rcvrNode.
	arguments := argList asArray.
	isBuiltinOperator := builtinFlag.! !

!TSendNode methodsFor: 'C code generation' stamp: 'eem 12/15/2015 21:31'!
shouldExcludeReceiverAsFirstArgument: aCodeGen
	"Only include the receiver as the first argument in certain cases.
	 The receiver is always included if it is an expression.
	 If it is a variable:
		 If the vmClass says it is an implicit variable, don't include it.
		 If the method's definingClass says it is an implicit variable, don't include it.
		 If the variable is 'self' and the method being called is not in
		 the method set (i.e. it is some external code), don't include it.
		 If it is a struct send of something the vm says is an implicit variable, don't include it."
	| m |
	(aCodeGen isAssertSelector: selector) ifTrue:
		[^true].

	(receiver isSend
	 and: [(receiver shouldExcludeReceiverAsFirstArgument: aCodeGen)
		or: [receiver receiver isVariable
			 and: [(self isSelfReference: receiver receiver in: aCodeGen)
				or: [self isStructReference: receiver receiver in: aCodeGen]]]]) ifTrue:
		[^aCodeGen isNonArgumentImplicitReceiverVariableName: receiver selector].

	^receiver isVariable
	    and: [(aCodeGen isNonArgumentImplicitReceiverVariableName: receiver name)
		    or: [(self isSelfReference: receiver in: aCodeGen)
			    and: [(m := aCodeGen methodNamed: selector) isNil
					or: [#(implicit nil) includes: m typeForSelf]]]]! !

!TSendNode methodsFor: 'C code generation' stamp: 'eem 2/11/2009 17:00'!
shouldIncludeReceiverAsFirstArgument: aCodeGen
	"Only include the receiver as the first argument in certain cases.
	 The receiver is always included if it is an expression.
	 If it is a variable:
		 If the vmClass says it is an implicit variable, don't include it.
		 If the variable is 'self' and the method being called is not in
		 the method set (i.e. it is some external code), don't include it."
	^(self shouldExcludeReceiverAsFirstArgument: aCodeGen) not! !

!TSendNode methodsFor: 'testing' stamp: 'eem 10/24/2013 13:02'!
structTargetKindIn: aCodeGen
	"Answer if the recever evaluates to a struct or struct pointer
	 and hence can be dereferenced using . or ->.  Answer any of
	 #struct #pointer or nil.  Right now we don't need or support
	 structure return so this method answers either #pointer or nil."
	selector == #cCoerceSimple:to: ifTrue:
		[^(VMStructType isTypePointerToStruct: arguments last value) ifTrue:
			[#pointer]].

	selector == #addressOf: ifTrue:
		[^#pointer].

	selector == #at: ifTrue:
		[receiver isVariable ifTrue:
			[(aCodeGen typeOfVariable: receiver name) ifNotNil:
				[:type| | derefType |
				 type last = $* ifFalse:
					[^receiver structTargetKindIn: aCodeGen].
				 (VMStructType isTypeStruct: (aCodeGen
											extractTypeFor: receiver name
											fromDeclaration: type allButLast)) ifTrue:
						[^#struct]]].
		(receiver structTargetKindIn: aCodeGen) ifNotNil:
			[:kind| ^kind]].

	(aCodeGen selectorReturnsPointerToStruct: selector) ifTrue:
		[^#pointer].

	(aCodeGen selectorReturnsStruct: selector) ifTrue:
		[^#struct].

	^nil! !

!TSendNode methodsFor: 'testing' stamp: 'eem 7/6/2021 18:26'!
targetOfCast
	self isUnaryCast ifTrue: [^receiver].
	self isBinaryCast ifTrue: [^arguments first].
	^nil! !

!TSendNode methodsFor: 'type inference' stamp: 'eem 6/7/2016 16:33'!
typeOrNilFrom: aCodeGenerator in: aTMethod
	^aCodeGenerator returnTypeForSend: self in: aTMethod ifNil: nil! !

TParseNode subclass: #TStmtListNode
	instanceVariableNames: 'arguments statements'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

TStmtListNode subclass: #TLiteralBlockNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TLiteralBlockNode methodsFor: 'testing' stamp: 'eem 11/15/2019 14:04'!
isLiteralBlock
	^true! !

!TStmtListNode methodsFor: 'utilities' stamp: 'eem 9/19/2018 17:54'!
addReadBeforeAssignedIn: variables to: readBeforeAssigned assignments: assigned in: aCodeGen
	"Add any variables in variables that are read before written to readBeforeAssigned.
	 Add unconditional assignments to assigned.  For convenience answer assigned."
	self
		nodesWithParentsDo:
			[:node :parent|
			(node isAssignment
			 and: [variables includes: node variable name]) ifTrue:
				[assigned add: node variable name].
			(node isVariable
			 and: [(variables includes: node name)
			 and: [(assigned includes: node name) not
			 and: [(#(nil pointer) includes: (node structTargetKindIn: aCodeGen))
			 and: [(parent notNil and: [parent isAssignment and: [parent variable == node]]) not]]]]) ifTrue:
				[node name = 'theCalloutState' ifTrue:
					[self halt].
				 readBeforeAssigned add: node name]]
		unless:
			[:node :parent| | conditionalAssignments mayHaveSideEffects |
			node isSend
				ifTrue:
					["First deal with implicit assignments..."
					node isValueExpansion ifTrue:
						[assigned addAll: node receiver args].
					(#(#'memcpy:_:_:' #'memmove:_:_:') includes: node selector) ifTrue:
						[assigned add: (node args first detect: [:subnode| subnode isVariable]) name].
					(#(to:do: to:by:do:) includes: node selector) ifTrue:
						[assigned addAll: (node args at: node selector numArgs) args.
						 mayHaveSideEffects := node args size = 4. "See TMethod>>prepareMethodIn:"
						 mayHaveSideEffects ifTrue:
							[assigned add: node args last name]].
					"Then deal with read-before-written in the arms of conditionals..."
					(#(ifTrue: ifFalse: ifNil: ifNotNil:) intersection: node selector keywords) notEmpty
						ifTrue:
							["First find assignments in the expression..."
							 (TStmtListNode new setStatements: {node receiver}; yourself)
								addReadBeforeAssignedIn: variables
								to: readBeforeAssigned
								assignments: assigned
								in: aCodeGen.
							 "Now find read-before-written in each arm, and collect the assignments to spot those assigned in both arms"
							 conditionalAssignments :=
								node args
									collect:
										[:block|
										block isStmtList ifTrue:
											[block
												addReadBeforeAssignedIn: variables
												to: readBeforeAssigned
												assignments: assigned copy
												in: aCodeGen]]
									thenSelect: [:each| each notNil].
							 "add to assigned those variables written to in both arms"
							 conditionalAssignments size = 2 ifTrue:
								[conditionalAssignments := conditionalAssignments collect: [:set| set difference: assigned].
								 assigned addAll: (conditionalAssignments first intersection: conditionalAssignments last)].
							 true]
						ifFalse:
							[false]]
				ifFalse:
					[false]].
	^assigned! !

!TStmtListNode methodsFor: 'accessing' stamp: ''!
args

	^arguments! !

!TStmtListNode methodsFor: 'transformations' stamp: 'eem 4/28/2016 17:09'!
asReturnNode
	self endsWithReturn ifTrue:
		[^self].
	^self class new
		setArguments: arguments
			statements: statements allButLast, {statements last asReturnNode};
		yourself! !

!TStmtListNode methodsFor: 'transformations' stamp: 'eem 2/13/2015 17:16'!
assignLastExpressionTo: variableNode
	"Destructively transform the receiver so that its last expression is assigned to the argument."
	| index |
	index := statements findLast: [:expr| (expr isGoTo or: [expr isLabel]) not].
	statements
		at: index
		put: (TAssignmentNode new
				setVariable: variableNode
				expression: (statements at: index))! !

!TStmtListNode methodsFor: 'transformations' stamp: 'jm 12/10/1998 18:26'!
bindVariableUsesIn: aDictionary

	statements := statements collect: [ :s | s bindVariableUsesIn: aDictionary ].! !

!TStmtListNode methodsFor: 'transformations' stamp: 'eem 10/29/2012 12:52'!
bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen
	"Answer either the receiver, if it contains no references to the given variables, or a new node with the given variables rebound."
	| newStatements |
	newStatements := statements collect: [:s| s bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen].
	^newStatements = statements
		ifTrue: [self]
		ifFalse:
			[self shallowCopy
				setStatements: newStatements;
				yourself]! !

!TStmtListNode methodsFor: 'transformations' stamp: 'eem 6/23/2020 17:14'!
bindVariablesIn: aDictionary

	aDictionary notEmpty ifTrue:
		[statements := statements collect: [:s| s bindVariablesIn: aDictionary]]! !

!TStmtListNode methodsFor: 'transformations' stamp: 'eem 12/15/2016 09:50'!
copyWithoutReturn
	self assert: self endsWithReturn.
	statements size = 1 ifTrue:
		[^statements last expression].
	^self class new
		setArguments: arguments
			statements: statements allButLast, {statements last copyWithoutReturn};
		yourself! !

!TStmtListNode methodsFor: 'C code generation' stamp: 'eem 12/8/2017 09:44'!
emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen
	| size |
	(size := statements size) = 1 ifTrue:
		[^statements first emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen].
	"Get rid of the annoying newlines in ? : if-then-elses"
	(size = 2 and: [statements first isComment and: [statements last isLeaf]]) ifTrue:
		[statements first emitCCodeAsArgumentOn: aStream level: 0 generator: aCodeGen.
		 aStream space.
		 ^statements last emitCCodeAsArgumentOn: aStream level: 0 generator: aCodeGen].
	aStream nextPut: $(. "N.B.  Comma binds weakest of all C operators."
	statements withIndexDo:
		[:s :idx| | p1 p2 |
		p1 := aStream position.
		s emitCCommentOn: aStream level: level generator: aCodeGen.
		(s isLeaf and: [s isLabel not and: [idx < statements size]]) ifFalse:
			[s emitCCodeAsArgumentOn: aStream level: level + 1 generator: aCodeGen].
		p2 := aStream position.
		(idx < size and: [p2 > p1]) ifTrue:
			[((self endsWithCloseBracket: aStream)
			  or: [s isComment]) ifFalse: [aStream nextPut: $,].
			 aStream crtab: level]].
	aStream nextPut: $)! !

!TStmtListNode methodsFor: 'C code generation' stamp: 'eem 8/18/2011 14:47'!
emitCCodeAsExpressionOn: aStream level: level generator: aCodeGen
	^self emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen! !

!TStmtListNode methodsFor: 'C code generation' stamp: 'eem 2/11/2009 14:43'!
emitCCodeOn: aStream level: level generator: aCodeGen

	self emitCCodeOn: aStream prependToEnd: nil level: level generator: aCodeGen
! !

!TStmtListNode methodsFor: 'C code generation' stamp: 'eem 7/6/2021 17:47'!
emitCCodeOn: aStream prependToEnd: aNodeOrNil level: level generator: aCodeGen
	| statementToElide |
	self emitCCommentOn: aStream level: level generator: aCodeGen.
	(aNodeOrNil notNil or: [statements isEmpty or: [statements last hasEffect]]) ifFalse:
		[statementToElide := statements last].
	statements withIndexDo:
		[:s :idx| | position  lastWhilePrepending |
		s emitCCommentOn: aStream level: level generator: aCodeGen.
		(s == statementToElide
		 or: [s isEffectFree and: [aNodeOrNil isNil or: [idx < statements size]]]) ifFalse:
			[((s isSend and: [s isPreprocessorDirective]) or: [aStream peekLast == Character tab]) ifFalse:
				[aStream tab: level].
			position := aStream position.
			(lastWhilePrepending := aNodeOrNil notNil and: [idx = statements size])
				ifTrue: [s emitCCodeOn: aStream prependToEnd: aNodeOrNil level: level generator: aCodeGen]
				ifFalse: [s emitCCodeOn: aStream level: level generator: aCodeGen].
			aStream position > position ifTrue:
				[(self stream: aStream endsWithAnyOf: '};') ifFalse:
					[s needsTrailingSemicolon ifTrue:
						[aStream nextPut: $;]].
					 (lastWhilePrepending or: [aStream peekLast == Character cr]) ifFalse:
						[aStream cr]]]]! !

!TStmtListNode methodsFor: 'testing' stamp: 'eem 7/5/2011 14:29'!
endsWithCloseBracket: aStream
	"Answer true if the given stream ends in a $} character."

	| ch pos |
	ch := (pos := aStream position) > 0 ifTrue:
			[aStream class = WriteStream "i.e. nested in StreamContents"
				ifTrue: [aStream originalContents at: pos]
				ifFalse: [aStream position: pos - 1; next]].
	^ ch = $}
! !

!TStmtListNode methodsFor: 'testing' stamp: 'eem 1/25/2016 18:31'!
endsWithReturn
	"Answer true if the last statement of this lock is a return."

	^statements notEmpty
	  and: [statements last isReturn or: [statements last isReturningIf]]! !

!TStmtListNode methodsFor: 'testing' stamp: 'eem 2/6/2020 14:43'!
hasEffect
	"Answer if this node has an effect on execution state (does something).
	 Statements that don't have any effect can be elided if their value is unused."
	^(arguments anySatisfy: [:node| node hasEffect])
	 or: [statements anySatisfy: [:node| node hasEffect]]! !

!TStmtListNode methodsFor: 'testing' stamp: 'acg 12/21/1999 02:25'!
isNilStmtListNode

	|stmt|
	statements size = 1 ifFalse: [^false].
	stmt := statements at: 1.
	^ stmt isVariable and: [stmt name = 'nil']! !

!TStmtListNode methodsFor: 'testing' stamp: 'eem 6/23/2020 11:26'!
isSameAs: aTParseNode
	self == aTParseNode ifTrue: [^true].
	(aTParseNode isStmtList
	 and: [statements size = aTParseNode statements size]) ifFalse:
		[^false].
	statements with: aTParseNode statements do:
		[:mine :theirs|
		 (mine isSameAs: theirs) ifFalse:
			[^false]].
	^true! !

!TStmtListNode methodsFor: 'testing' stamp: ''!
isStmtList

	^true! !

!TStmtListNode methodsFor: 'enumerating' stamp: 'eem 11/19/2012 12:12'!
nodesDo: aBlock
	"Apply aBlock to all nodes in the receiver.
	 N.B. This is assumed to be bottom-up, leaves first."
	statements do: [ :s | s nodesDo: aBlock ].	
	aBlock value: self! !

!TStmtListNode methodsFor: 'enumerating' stamp: 'eem 10/14/2013 15:02'!
nodesDo: aBlock parent: parent
	"Apply aBlock to all nodes in the receiver with each node's parent.
	 N.B. This is assumed to be bottom-up, leaves first."
	statements do: [:s| s nodesDo: aBlock parent: self].
	aBlock value: self value: parent! !

!TStmtListNode methodsFor: 'enumerating' stamp: 'eem 3/31/2015 10:55'!
nodesDo: aBlock parent: parent unless: cautionaryBlock

	(cautionaryBlock value: self value: parent) ifTrue: [^self].
	statements do: [ :s | s nodesDo: aBlock parent: self unless: cautionaryBlock].
	aBlock value: self value: parent! !

!TStmtListNode methodsFor: 'enumerating' stamp: 'eem 2/27/2009 16:46'!
nodesDo: aBlock unless: cautionaryBlock

	(cautionaryBlock value: self) ifTrue: [^self].
	statements do: [ :s | s nodesDo: aBlock unless: cautionaryBlock ].	
	aBlock value: self! !

!TStmtListNode methodsFor: 'enumerating' stamp: 'eem 10/13/2013 17:03'!
nodesDo: aBlock value: parent
	"Apply aBlock to all nodes in the receiver with each node's parent.
	 N.B. This is assumed to be bottom-up, leaves first."
	statements do: [:s| s nodesDo: aBlock parent: self.].	
	aBlock value: self value: parent! !

!TStmtListNode methodsFor: 'copying' stamp: 'eem 2/9/2009 19:30'!
postCopy

	arguments := arguments copy.
	statements := statements collect: [ :s | s copy ]! !

!TStmtListNode methodsFor: 'printing' stamp: 'eem 4/27/2016 17:26'!
printOn: aStream level: level

	statements size > 1 ifTrue: [ aStream crtab: level + 1 ].
	aStream nextPut: $[.
	(arguments notNil and: [arguments notEmpty]) ifTrue:
		[arguments do: [ :arg | aStream nextPutAll: ' :'; nextPutAll: arg].
		 aStream nextPutAll: ' | '].
	self printStatementsOn: aStream level: level.
	aStream nextPut: $]! !

!TStmtListNode methodsFor: 'printing' stamp: 'eem 4/27/2016 17:30'!
printStatementsOn: aStream level: level

	statements
		do: [:s| s printOn: aStream level: level]
		separatedBy: [aStream nextPut: $.; crtab: level + 1]! !

!TStmtListNode methodsFor: 'transformations' stamp: 'ikp 9/26/97 14:50'!
removeAssertions
	| newStatements |
	newStatements := OrderedCollection new: statements size.
	statements do: [ :stmt |
		stmt isAssertion ifFalse: [
			newStatements add: (stmt removeAssertions; yourself).
		]
	].
	self setStatements: newStatements asArray! !

!TStmtListNode methodsFor: 'inlining support' stamp: 'GuillermoPolito 1/24/2020 16:21'!
renameLabelsForInliningInto: aTMethod
	"TMethod already has a method for this; hijack it..."
	| labels method |
	labels := OrderedCollection new.
	self nodesDo:
		[:node| (node isLabel and: [ node label notNil ])
			ifTrue: [labels add: node label]].
	method := TMethod new
		parseTree: self;
		labels: labels;
		renameLabelsForInliningInto: aTMethod.
	^ method labels! !

!TStmtListNode methodsFor: 'transformations' stamp: 'eem 10/30/2013 10:13'!
replaceChild: aNode with: bNode 
	statements := Array streamContents:
					[:s|
					statements do:
						[:node|
						node == aNode
							ifTrue:
								[bNode isStmtList
									ifTrue: [s nextPutAll: bNode statements]
									ifFalse: [s nextPut: bNode]]
							ifFalse: [s nextPut: node]]]! !

!TStmtListNode methodsFor: 'transformations' stamp: ''!
replaceNodesIn: aDictionary

	^aDictionary at: self ifAbsent: [
		statements := statements collect: [ :s | s replaceNodesIn: aDictionary ].
		self]! !

!TStmtListNode methodsFor: 'testing' stamp: 'eem 6/7/2016 17:53'!
returnsExpression
	"Answer true if the last statement of this block is a return of some expression, not merely self or nil."

	statements isEmpty ifTrue:
		[^false].
	statements last isReturn ifFalse:
		[^false].
	statements last isVariable ifFalse:
		[^true].
	^statements last variable ~= 'self'
	  and: [statements last variable ~= 'nil']! !

!TStmtListNode methodsFor: 'accessing' stamp: ''!
setArguments: argList

	arguments := argList.! !

!TStmtListNode methodsFor: 'accessing' stamp: ''!
setArguments: argList statements: statementList
	"Initialize this method using the given information."

	arguments := argList.
	statements := statementList.! !

!TStmtListNode methodsFor: 'accessing' stamp: ''!
setStatements: stmtList

	statements := stmtList asOrderedCollection.! !

!TStmtListNode methodsFor: 'inlining' stamp: 'eem 7/8/2008 19:58'!
shouldExpand
	^(statements detect: [:stmt| stmt isInline]) shouldExpand! !

!TStmtListNode methodsFor: 'accessing' stamp: ''!
statements

	^statements! !

!TStmtListNode methodsFor: 'type inference' stamp: 'eem 6/7/2016 18:09'!
typeOrNilFrom: aCodeGenerator in: aTMethod
	^statements isEmpty ifFalse:
		[statements last typeOrNilFrom: aCodeGenerator in: aTMethod]! !

TParseNode subclass: #TSwitchStmtNode
	instanceVariableNames: 'expression cases otherwiseOrNil switchVariable'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TSwitchStmtNode commentStamp: '<historical>' prior: 0!
I implement a Smalltalk
	foo caseOf: { [IntegerConstant | GlobalVariable] -> [expr] }
statement converting it into a C switch statement.  I make some effort to discover identical right-hand-side cases.!

!TSwitchStmtNode methodsFor: 'transformations' stamp: 'eem 2/11/2009 12:15'!
bindVariableUsesIn: aDictionary
	expression := expression bindVariableUsesIn: aDictionary.
	cases := (cases collect:
				[:pair| "<Array with: <Array with: ParseTreeNode> with: TStmtListNode"
				{ pair first collect: [:labelNode| labelNode bindVariableUsesIn: aDictionary].
				   pair last bindVariableUsesIn: aDictionary }]).
	otherwiseOrNil ifNotNil:
		[otherwiseOrNil := otherwiseOrNil bindVariableUsesIn: aDictionary]! !

!TSwitchStmtNode methodsFor: 'transformations' stamp: 'eem 10/29/2012 12:55'!
bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen
	"Answer either the receiver, if it contains no references to the given variables, or a new node with the given variables rebound."
	| newExpression newCases newOtherwise |
	newExpression := expression bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen.
	newCases := cases collect:
					[:pair| "<Array with: <Array with: ParseTreeNode> with: TStmtListNode"
					{ pair first collect: [:labelNode| labelNode bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen].
					   pair last bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen}].
	newOtherwise := otherwiseOrNil ifNotNil:
						[otherwiseOrNil bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen].
	^(newExpression = expression
	   and: [newCases = cases
	   and: [newOtherwise = otherwiseOrNil]])
		ifTrue: [self]
		ifFalse:
			[self shallowCopy
				expression: newExpression;
				cases: newCases;
				otherwiseOrNil: newOtherwise;
				yourself]! !

!TSwitchStmtNode methodsFor: 'transformations' stamp: 'eem 2/11/2009 12:14'!
bindVariablesIn: aDictionary
	expression := expression bindVariablesIn: aDictionary.
	cases := (cases collect:
				[:pair| "<Array with: <Array with: ParseTreeNode> with: TStmtListNode"
				{ pair first collect: [:labelNode| labelNode bindVariablesIn: aDictionary].
				   pair last bindVariablesIn: aDictionary }]).
	otherwiseOrNil ifNotNil:
		[otherwiseOrNil := otherwiseOrNil bindVariablesIn: aDictionary]! !

!TSwitchStmtNode methodsFor: 'accessing' stamp: 'eem 2/9/2009 15:56'!
cases
	"Answer the value of cases"

	^ cases! !

!TSwitchStmtNode methodsFor: 'accessing' stamp: 'eem 2/9/2009 15:56'!
cases: anObject
	"Set the value of cases"

	cases := anObject! !

!TSwitchStmtNode methodsFor: 'instance initialization' stamp: 'eem 11/17/2015 15:13'!
createCasesFromBraceNode: aTBraceNode
	"Answer a sequence of tuples of { labels. case } for a TBraceNode, making
	 sure to collect equivalent cases together under a signle sequence of labels."
	| casesToStrings stringsToLabels newCases |
	casesToStrings := Dictionary new.
	stringsToLabels := Dictionary new.
	newCases := OrderedCollection new: aTBraceNode caseLabels size.
	aTBraceNode caseLabels with: aTBraceNode cases do:
		[:label :case| | printString |
		printString := casesToStrings at: case put: case printString.
		(stringsToLabels at: printString ifAbsentPut: [OrderedCollection new]) addLast: label].

	aTBraceNode caseLabels with: aTBraceNode cases do:
		[:label :case| | printString labels |
		printString := casesToStrings at: case.
		label = (labels := (stringsToLabels at: printString) asArray) first ifTrue:
			[newCases addLast: { labels collect: [:ea| ea statements first]. case}]].

	^newCases! !

!TSwitchStmtNode methodsFor: 'C code generation' stamp: 'eem 10/13/2013 18:35'!
emitCCodeAsExpressionOn: aStream level: level generator: aCodeGen
	"Emit the receiver as an if-the-else chain."
	| varName n |
	self assert: (expression isVariable or: [switchVariable notNil]).
	aStream nextPut: $(.
	switchVariable
		ifNil: [varName := String streamContents: [:s| expression emitCCodeOn: s level: 0 generator: aCodeGen].
			aStream nextPutAll: varName]
		ifNotNil:
			[varName := switchVariable.
			 aStream nextPut: $(; nextPutAll: varName; nextPutAll: ' = '.
			 expression emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen.
			 aStream nextPut: $)].
	n := 0.
	cases do:
		[:tuple|
		 [:labels :case|
		  labels do:
			[:label|
			 n > 0 ifTrue:
				[aStream nextPutAll: varName].
			 aStream nextPutAll: ' == '.
			 label emitCCodeAsArgumentOn: aStream level: level + 1 generator: aCodeGen.
			 aStream nextPut: $).
			 aStream crtab: level + n + 1.
			 aStream nextPutAll: '? ('.
			 (TStmtListNode new setArguments: #() statements: case statements)
			 	emitCCodeAsArgumentOn: aStream
				level: level + 2
				generator: aCodeGen.
			 aStream nextPut: $); crtab: level + n + 1; nextPutAll: ': ('.
			 n := n + 1]]
			valueWithArguments: tuple].
	otherwiseOrNil
		ifNotNil: [otherwiseOrNil emitCCodeAsArgumentOn: aStream level: level + 1 generator: aCodeGen]
		ifNil: [aStream nextPutAll: 'error("Case not found and no otherwise clause"), 0'].
	aStream next: n - 1 put: $)! !

!TSwitchStmtNode methodsFor: 'C code generation' stamp: 'eem 7/25/2020 12:53'!
emitCCodeOn: aStream addToEndOfCases: aNodeOrNil level: level generator: aCodeGen

	aStream nextPutAll: 'switch ('.
	expression emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen.
	aStream nextPutAll: ') {'.
	cases do:
		[:tuple|
		 [:labels :case|
		  labels do:
			[:label|
			 aStream
				crtab: level;
				nextPutAll: 'case '.
			label emitCCodeAsArgumentOn: aStream level: level + 1 generator: aCodeGen.
			aStream nextPut: $:].
		  aStream crtab: level + 1.
		  case emitCCodeOn: aStream prependToEnd: aNodeOrNil level: level + 1 generator: aCodeGen]
			valueWithArguments: tuple.
		  (aNodeOrNil notNil and: [aNodeOrNil isReturn])
			ifTrue: [(self stream: aStream endsWithAnyOf: ';}') ifFalse: [aStream nextPut: $;]]
			ifFalse: [aStream crtab: level + 1; nextPutAll: 'break;'].
		  aStream cr].
	aStream
		crtab: level;
		nextPutAll: 'default:';
		crtab: level + 1.
	otherwiseOrNil
		ifNotNil: [otherwiseOrNil emitCCodeOn: aStream prependToEnd: aNodeOrNil level: level + 1 generator: aCodeGen]
		ifNil: [aStream nextPutAll: 'error("Case not found and no otherwise clause")'.
			   aNodeOrNil ifNotNil:
				[| defaultExpr type |
				 aStream nextPut: $;; crtab: level + 1.
				 defaultExpr := TConstantNode new setValue: -1.
				 (aNodeOrNil isAssignment
				  and: [(type := aCodeGen typeFor: aNodeOrNil variable in: aCodeGen currentMethod) notNil
				  and: [aCodeGen isPointerCType: type]]) ifTrue:
					[defaultExpr := TSendNode new
										setSelector: #cCoerceSimple:to:
										receiver: (TVariableNode new setName: 'self')
										arguments: {defaultExpr. TConstantNode new setValue: type}].
				 (aNodeOrNil copy setExpression: defaultExpr)
					emitCCodeOn: aStream level: level generator: aCodeGen]].
	aStream
		nextPut: $;;
		crtab: level;
		nextPut: $}! !

!TSwitchStmtNode methodsFor: 'C code generation' stamp: 'eem 7/25/2020 12:48'!
emitCCodeOn: aStream level: level generator: aCodeGen

	aStream nextPutAll: 'switch ('.
	expression emitCCodeAsArgumentOn: aStream level: level generator: aCodeGen.
	aStream nextPutAll: ') {'.
	cases do:
		[:tuple|
		 [:labels :case|
		  labels do:
			[:label|
			 aStream
				crtab: level;
				nextPutAll: 'case '.
			label emitCCodeAsArgumentOn: aStream level: level + 1 generator: aCodeGen.
			aStream nextPut: $:].
		  aStream crtab: level + 1.
		  case emitCCodeOn: aStream level: level + 1 generator: aCodeGen.
		  case endsWithReturn ifFalse:
			[aStream tab: level + 1; nextPutAll: 'break;']]
			valueWithArguments: tuple].
	aStream
		crtab: level;
		nextPutAll: 'default:';
		crtab: level + 1.
	otherwiseOrNil
		ifNotNil:
			[otherwiseOrNil emitCCodeOn: aStream level: level + 1 generator: aCodeGen.
			(self stream: aStream endsWithAnyOf: '};') ifFalse:
				[aStream nextPut: $;]]
		ifNil: [aStream nextPutAll: 'error("Case not found and no otherwise clause");'].
	aStream
		crtab: level;
		nextPut: $}! !

!TSwitchStmtNode methodsFor: 'accessing' stamp: 'eem 2/9/2009 15:56'!
expression
	"Answer the value of expression"

	^ expression! !

!TSwitchStmtNode methodsFor: 'accessing' stamp: 'eem 2/9/2009 15:56'!
expression: anObject
	"Set the value of expression"

	expression := anObject! !

!TSwitchStmtNode methodsFor: 'instance initialization' stamp: 'eem 2/9/2009 16:30'!
expression: expr cases: aTBraceNode otherwiseOrNil: otherwiseOrNilNode

	self expression: expr.
	self cases: (self createCasesFromBraceNode: aTBraceNode).
	self otherwiseOrNil: otherwiseOrNilNode! !

!TSwitchStmtNode methodsFor: 'testing' stamp: 'eem 7/7/2021 10:44'!
hasEffect
	"Answer if this node has an effect on execution state (does something).
	 Statements that don't have any effect can be elided if their value is unused."
	^expression hasEffect
	  or: [(otherwiseOrNil notNil and: [otherwiseOrNil hasEffect])
	  or: [cases anySatisfy:
			[:tuple|
			(tuple first anySatisfy: [:array| array anySatisfy: [:node| node hasEffect]])
			or: [tuple second hasEffect]]]]! !

!TSwitchStmtNode methodsFor: 'testing' stamp: 'eem 2/10/2009 07:58'!
isSwitch
	^true! !

!TSwitchStmtNode methodsFor: 'enumerating' stamp: 'eem 11/19/2012 12:12'!
nodesDo: aBlock
	"Apply aBlock to all nodes in the receiver.
	 N.B. This is assumed to be bottom-up, leaves first."
	expression nodesDo: aBlock.
	cases do:
		[:pair|
		pair first do: [:node| node nodesDo: aBlock].
		pair last nodesDo: aBlock].
	otherwiseOrNil ifNotNil:
		[otherwiseOrNil nodesDo: aBlock].
	aBlock value: self! !

!TSwitchStmtNode methodsFor: 'enumerating' stamp: 'eem 8/21/2021 19:32'!
nodesDo: aBlock parent: parent
	"Apply aBlock to all nodes in the receiver with each node's parent.
	 N.B. This is assumed to be bottom-up, leaves first."
	expression nodesDo: aBlock parent: self.
	cases do:
		[:pair|
		pair first do: [:node| node nodesDo: aBlock parent: self].
		pair last nodesDo: aBlock parent: self].
	otherwiseOrNil ifNotNil:
		[otherwiseOrNil nodesDo: aBlock parent: self].
	aBlock value: self value: parent! !

!TSwitchStmtNode methodsFor: 'enumerating' stamp: 'eem 03/31/2015 10:12'!
nodesDo: aBlock parent: parent unless: cautionaryBlock

	(cautionaryBlock value: self value: parent) ifTrue: [^self].
	expression nodesDo: aBlock parent: self unless: cautionaryBlock.
	cases do:
		[:pair|
		pair first do: [:node| node nodesDo: aBlock parent: self unless: cautionaryBlock].
		pair last nodesDo: aBlock parent: self unless: cautionaryBlock].
	otherwiseOrNil ifNotNil:
		[otherwiseOrNil nodesDo: aBlock parent: self unless: cautionaryBlock]! !

!TSwitchStmtNode methodsFor: 'enumerating' stamp: 'eem 2/27/2009 16:48'!
nodesDo: aBlock unless: cautionaryBlock

	(cautionaryBlock value: self) ifTrue: [^self].
	expression nodesDo: aBlock unless: cautionaryBlock.
	cases do:
		[:pair|
		pair first do: [:node| node nodesDo: aBlock unless: cautionaryBlock].
		pair last nodesDo: aBlock unless: cautionaryBlock].
	otherwiseOrNil ifNotNil:
		[otherwiseOrNil nodesDo: aBlock unless: cautionaryBlock]! !

!TSwitchStmtNode methodsFor: 'accessing' stamp: 'eem 2/9/2009 15:56'!
otherwiseOrNil
	"Answer the value of otherwiseOrNil"

	^ otherwiseOrNil! !

!TSwitchStmtNode methodsFor: 'accessing' stamp: 'eem 2/9/2009 15:56'!
otherwiseOrNil: anObject
	"Set the value of otherwiseOrNil"

	otherwiseOrNil := anObject! !

!TSwitchStmtNode methodsFor: 'copying' stamp: 'eem 2/9/2009 19:26'!
postCopy
	expression := expression copy.
	cases := (cases collect:
				[:pair| "<Array with: <Array with: ParseTreeNode> with: TStmtListNode"
				{ pair first collect: [:labelNode| labelNode copy].
				   pair last copy }]).
	otherwiseOrNil := otherwiseOrNil copy! !

!TSwitchStmtNode methodsFor: 'printing' stamp: 'eem 7/25/2020 12:48'!
printOn: aStream level: level

	aStream nextPutAll: 'switch ('.
	expression printOn: aStream level: level.
	aStream nextPutAll: ') {'.
	cases do:
		[:tuple|
		 [:labels :case|
		  labels do:
			[:label|
			 aStream
				crtab: level;
				nextPutAll: 'case '.
			label printOn: aStream level: level + 1.
			aStream nextPut: $:].
		  aStream crtab: level + 1.
		  case printOn: aStream level: level + 1.
		  aStream crtab: level + 1; nextPutAll: 'break;']
			valueWithArguments: tuple].
	otherwiseOrNil ifNotNil:
		[aStream
			crtab: level;
			nextPutAll: 'default:';
			crtab: level + 1.
		 otherwiseOrNil printOn: aStream level: level + 1].
	aStream
		crtab: level;
		nextPut: $}! !

!TSwitchStmtNode methodsFor: 'transformations' stamp: 'eem 2/11/2009 12:18'!
removeAssertions
	expression removeAssertions.
	cases do:
		[:pair| "<Array with: <Array with: ParseTreeNode> with: TStmtListNode"
		pair first do: [:labelNode| labelNode removeAssertions].
		pair last removeAssertions].
	otherwiseOrNil ifNotNil:
		[otherwiseOrNil removeAssertions]! !

!TSwitchStmtNode methodsFor: 'transformations' stamp: 'eem 2/11/2009 17:16'!
replaceNodesIn: aDictionary
	^aDictionary
		at: self
		ifAbsent:
			[expression := expression replaceNodesIn: aDictionary.
			 cases := (cases collect:
						[:pair| "<Array with: <Array with: ParseTreeNode> with: TStmtListNode"
						{ pair first collect: [:labelNode| labelNode replaceNodesIn: aDictionary].
						   pair last replaceNodesIn: aDictionary }]).
			 otherwiseOrNil ifNotNil:
				[otherwiseOrNil := otherwiseOrNil replaceNodesIn: aDictionary].
			 self]! !

!TSwitchStmtNode methodsFor: 'accessing' stamp: 'eem 10/13/2013 17:20'!
switchVariable
	"Answer the value of switchVariable"

	^ switchVariable! !

!TSwitchStmtNode methodsFor: 'accessing' stamp: 'eem 10/13/2013 17:20'!
switchVariable: anObject
	"Set the value of switchVariable"

	switchVariable := anObject! !

TParseNode subclass: #TVariableNode
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Translation to C'!

!TVariableNode methodsFor: 'transformations' stamp: 'eem 2/9/2009 19:22'!
bindVariableUsesIn: aDictionary

	^ (aDictionary at: name ifAbsent: [^ self]) copy! !

!TVariableNode methodsFor: 'transformations' stamp: 'eem 10/29/2012 12:57'!
bindVariableUsesIn: aDictionary andConstantFoldIf: constantFold in: codeGen
	"Answer either the receiver, if it is not a reference to one of the given variables, or the replacement if it is."
	^(aDictionary at: name ifAbsent: [^self]) copy! !

!TVariableNode methodsFor: 'transformations' stamp: 'eem 2/9/2009 19:22'!
bindVariablesIn: aDictionary 
	^ (aDictionary at: name ifAbsent: [^ self]) copy! !

!TVariableNode methodsFor: 'C code generation' stamp: 'eem 12/8/2017 13:07'!
emitCCodeOn: aStream level: level generator: aCodeGen

	aStream nextPutAll: (name = 'nil'
						ifTrue: [aCodeGen cLiteralFor: nil]
						ifFalse: [aCodeGen returnPrefixFromVariable: name]).
	aCodeGen currentMethod ifNotNil:
		[:m| m noteUsedVariableName: name]! !

!TVariableNode methodsFor: 'testing' stamp: 'eem 11/17/2015 14:28'!
hasSideEffect
	"Answer if the parse tree rooted at this node has a side-effect or not."
	^false! !

!TVariableNode methodsFor: 'testing' stamp: 'eem 7/6/2021 17:48'!
isEffectFree
	^true! !

!TVariableNode methodsFor: 'testing' stamp: ''!
isLeaf

	^true! !

!TVariableNode methodsFor: 'comparing' stamp: 'eem 6/23/2020 11:26'!
isSameAs: aTParseNode
	^self == aTParseNode
	 or: [aTParseNode isVariable
		 and: [name = aTParseNode name]]! !

!TVariableNode methodsFor: 'testing' stamp: ''!
isVariable

	^true! !

!TVariableNode methodsFor: 'accessing' stamp: 'eem 2/15/2009 09:09'!
name

	^name asString! !

!TVariableNode methodsFor: 'accessing' stamp: 'acg 12/17/1999 07:18'!
nameOrValue

	^name! !

!TVariableNode methodsFor: 'enumerating' stamp: 'eem 10/13/2013 17:10'!
nodesDo: aBlock parent: parent
	"Apply aBlock to all nodes in the receiver with each node's parent.
	 N.B. This is assumed to be bottom-up, leaves first."
	aBlock value: self value: parent! !

!TVariableNode methodsFor: 'printing' stamp: ''!
printOn: aStream level: level

	aStream nextPutAll: name.! !

!TVariableNode methodsFor: 'accessing' stamp: ''!
setName: aString

	name := aString.! !

!TVariableNode methodsFor: 'testing' stamp: 'eem 10/24/2013 13:02'!
structTargetKindIn: aCodeGen
	"Answer if the recever evaluates to a struct or struct pointer
	 and hence can be dereferenced using . or ->.  Answer any of
	 #struct #pointer or nil"
	^aCodeGen structTargetKindForVariableName: name! !

!TVariableNode methodsFor: 'type inference' stamp: 'nice 4/6/2016 23:40'!
typeOrNilFrom: aCodeGenerator in: aTMethod
	^aTMethod typeFor: self in: aCodeGenerator! !

!TVariableNode methodsFor: 'accessing' stamp: 'eem 11/12/2013 14:22'!
variableNameOrNil
	^name! !

Object subclass: #VMClass
	instanceVariableNames: ''
	classVariableNames: 'DefaultBase ExpensiveAsserts InitializationOptions'
	poolDictionaries: 'VMBasicConstants VMObjectIndices'
	category: 'VMMaker-Support'!

VMClass class
	instanceVariableNames: 'timeStamp'!

!VMClass commentStamp: 'eem 6/7/2018 15:16' prior: 0!
I am an abstract superclass for all classes in the VM that want to maintain a source timeStamp.  I am also the holder of the InitializationOptions class variable which holds options such as which JIT or which memory manager to use when creating or generating a VM.!

VMClass class
	instanceVariableNames: 'timeStamp'!

VMClass subclass: #CogClass
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogClass commentStamp: '<historical>' prior: 0!
I am a place to hang various pieces of translation support.!

!CogClass class methodsFor: 'system primitives' stamp: 'eem 1/23/2013 16:19'!
minimumUnusedStackHeadroom
	"self minimumUnusedStackHeadroom"
	<primitive: 'primitiveMinimumUnusedHeadroom'>
	self primitiveFailed! !

!CogClass methodsFor: 'translation support' stamp: 'eem 11/11/2020 18:35'!
cCoerceSimple: value to: cTypeString
	<doNotGenerate>
	"Type coercion for translation and simulation.
	 For simulation answer a suitable surrogate for the struct types"
	^cTypeString caseOf:
	   {	[#'unsigned long']							->	[value].
		[#'unsigned int']							->	[value].
		[#'unsigned short']							->	[value].
		[#sqInt]									->	[value].
		[#'sqIntptr_t']								->	[value].
		[#'usqIntptr_t']								->	[value].
		[#usqInt]									->	[value].
		[#sqLong]									->	[value].
		[#usqLong]								->	[value].
		[#'AbstractInstruction *']					->	[value].
		[#'SpurSegmentInfo *']						->	[value].
		[#'BytecodeFixup *']						->	[value].
		[#'CogMethod *']							->	[value].
		[#'char *']									->	[value].
		[#'sqInt *']									->	[value].
		[#'void *']									->	[value].
		[#void]										->	[value].
		[#'void (*)()']								->	[value].
		[#'void (*)(void)']							->	[value].
		[#'usqIntptr_t (*)(void)']					->	[value].
		[#'void (*)(usqIntptr_t,usqIntptr_t)']		->	[value].
		[#'usqIntptr_t (*)(usqIntptr_t)']			->	[value].
		[#'usqIntptr_t (*)(void)']					->	[value] }! !

CogClass subclass: #CogMethodZone
	instanceVariableNames: 'youngReferrers unpairedMethodList methodCount openPICList mzFreeStart baseAddress limitAddress methodBytesFreedSinceLastCompaction compactionInProgress coInterpreter objectRepresentation cogit objectMemory'
	classVariableNames: ''
	poolDictionaries: 'CogMethodConstants VMBasicConstants'
	category: 'VMMaker-JIT'!

!CogMethodZone commentStamp: 'eem 9/24/2014 15:59' prior: 0!
I am a simple allocator/deallocator for the native code zone.  I also manage the youngReferers list, which contains methods that may refer to one or more young objects, and the openPICList which is a linked list of all open PICs in the zone.

Instance Variables
	baseAddress:								<Integer address>
	coInterpreter:								<CoInterpreter>
	cogit:										<Cogit>
	limitAddress:								<Integer address>
	methodBytesFreedSinceLastCompaction:	<Integer>
	methodCount:								<Integer>
	mzFreeStart:								<Integer address>
	objectMemory:								<NewCoObjectMemory|SpurCoMemoryManager>
	objectRepresentation:						<CogObjectRepresentation:>
	openPICList:								<CogMethod *|nil>
	unpairedMethodList:						<CogMethod *|nil>
	youngReferrers:							<Integer address>

baseAddress
	- the lowest address in the native method zone

coInterpreter
	- simulation-only

cogit
	- simulation-only

limitAddress
	- the address immediately following the native method zone

methodBytesFreedSinceLastCompaction
	- a count of the bytes in methods freed since the last compaction of the native method zone, used to answer the used bytes in the zone

methodCount
	- a count of the number of methods in the native method zone

mzFreeStart
	- the start of free space in the zone

objectMemory
	- simulation-only

objectRepresentation
	- simulation-only

openPICList
	- the head of the list of open PICs

unpairedMethodList
	- the head of the list of Cog methods with no associated CompiledMethod object (Newspeak only)

youngReferrers
	- the pointer to the start of an array of pointers to CogMethods that refer to young objects.  May contain false positives.  Occupies the top of the zone from youngReferrers up to limitAddress
!

!CogMethodZone class methodsFor: 'translation' stamp: 'eem 7/9/2014 19:43'!
declareCVarsIn: aCCodeGenerator
	aCCodeGenerator
		removeVariable: 'coInterpreter';
		removeVariable: 'objectRepresentation'.
	
	self declareC: #(youngReferrers mzFreeStart baseAddress limitAddress unpairedMethodList)
		as: #usqInt
		in: aCCodeGenerator! !

!CogMethodZone class methodsFor: 'translation' stamp: 'eem 3/12/2009 11:14'!
implicitReturnTypeFor: aSelector
	"Answer the return type for methods that don't have an explicit return."
	^#void! !

!CogMethodZone class methodsFor: 'translation' stamp: 'eem 10/19/2009 12:19'!
isNonArgumentImplicitReceiverVariableName: aString
	^#('self' 'coInterpreter' 'objectRepresentation') includes: aString! !

!CogMethodZone class methodsFor: 'translation' stamp: 'eem 2/11/2009 15:48'!
typeForSelf
	^#implicit! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 2/25/2020 18:58'!
addToOpenPICList: anOpenPIC
	<var: #anOpenPIC type: #'CogMethod *'>
	self assert: anOpenPIC cmType = CMOpenPIC.
	self assert: (openPICList == nil
				or: [openPICList cmType = CMOpenPIC]).
	cogit assertValidDualZoneWriteAddress: anOpenPIC.
	anOpenPIC nextOpenPIC: openPICList asUnsignedInteger.
	openPICList := cogit cCoerceSimple: anOpenPIC asUnsignedInteger - cogit getCodeToDataDelta to: #'CogMethod *'.
	self cCode: '' inSmalltalk: [self deny: openPICList isInteger]! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 2/25/2020 19:01'!
addToUnpairedMethodList: aCogMethod
	<option: #NewspeakVM>
	<var: #aCogMethod type: #'CogMethod *'>
	self assert: aCogMethod cmType = CMMethod.
	self assert: (cogit noAssertMethodClassAssociationOf: aCogMethod methodObject) = objectMemory nilObject.
	self assert: (unpairedMethodList == nil
				or: [(self cCoerceSimple: unpairedMethodList to: #'CogMethod *') cmType = CMMethod]).
	cogit assertValidDualZoneWriteAddress: aCogMethod.
	aCogMethod nextMethodOrIRCs: unpairedMethodList.
	unpairedMethodList := aCogMethod asUnsignedInteger - cogit getCodeToDataDelta! !

!CogMethodZone methodsFor: 'young referers' stamp: 'eem 2/25/2020 18:58'!
addToYoungReferrers: cogMethod
	<var: #cogMethod type: #'CogMethod *'>
	cogit assertValidDualZoneWriteAddress: cogMethod.
	self assert: (self occurrencesInYoungReferrers: cogMethod) = 0.
	self assert: cogMethod cmRefersToYoung.
	self assert: (youngReferrers <= limitAddress
				and: [youngReferrers >= (limitAddress - (methodCount * objectMemory wordSize))]).
	(self asserta: limitAddress - (methodCount * objectMemory wordSize) >= mzFreeStart) ifFalse:
		[self error: 'no room on youngReferrers list'].
	youngReferrers := youngReferrers - objectMemory wordSize.
	cogit
		codeLongAt: youngReferrers
		put: cogMethod asUnsignedInteger - cogit getCodeToDataDelta! !

!CogMethodZone methodsFor: 'testing' stamp: 'eem 12/27/2019 18:08'!
addressIsLikelyCogMethod: address
	<doNotGenerate>
	(address anyMask: (self roundUpLength: 1) - 1) ifTrue:
		[^false].
	(self oop: address isGreaterThanOrEqualTo: cogit minCogMethodAddress andLessThan: mzFreeStart) ifFalse:
		[^false].
	^(objectMemory baseHeader: address) = objectMemory nullHeaderForMachineCodeMethod! !

!CogMethodZone methodsFor: 'allocating' stamp: 'eem 3/16/2017 16:10'!
allocate: numBytes
	| roundedBytes allocation |
	roundedBytes := numBytes + 7 bitAnd: -8.
	mzFreeStart + roundedBytes >= self allocationLimit ifTrue:
		[^0].
	allocation := mzFreeStart.
	mzFreeStart := mzFreeStart + roundedBytes.
	methodCount := methodCount + 1.
	self cCode: '' inSmalltalk:
		[(cogit breakPC isInteger
		   and: [cogit breakPC between: allocation and: mzFreeStart]) ifTrue:
			[cogit singleStep: true]].
	^allocation! !

!CogMethodZone methodsFor: 'allocating' stamp: 'eem 3/16/2017 16:10'!
allocationLimit
	<inline: true>
	^limitAddress - (methodCount * objectMemory wordSize)! !

!CogMethodZone methodsFor: 'jit - api' stamp: 'eem 7/22/2021 15:03'!
clearCogCompiledCode
	"Free all methods. Do it the slow way to keep methodBytesFreedSinceLastCompaction accurate."
	<inline: true>
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	NewspeakVM ifTrue: [unpairedMethodList := nil].
	cogMethod := coInterpreter cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod asUnsignedInteger < mzFreeStart] whileTrue:
		[cogMethod cmType = CMMethod ifTrue:
			[self freeMethod: cogMethod].
		 cogMethod := self methodAfter: cogMethod].
	self manageFrom: baseAddress to: limitAddress! !

!CogMethodZone methodsFor: 'compaction' stamp: 'eem 1/29/2020 14:56'!
clearSavedPICUsageCount: cogMethod
	"For Sista, where we want PICs to last so they can be observed, we need to keep PICs unless
	 they are definitely unused.  So we need to identify unused PICs.  So in planCompact, zero the
	 usage counts of all PICs, saving the actual usage count in blockEntryOffset.  Then in
	 relocateMethodsPreCompaction (actually in relocateIfCallOrMethodReference:mcpc:delta:)
	 restore the usage counts of used PICs.  Finally in compactCompiledCode, clear the blockEntryOffset
	 of the unused PICs; they will then have a zero count and be reclaimed in the next code compaction."
	<inline: #always>
	(SistaVM
	 and: [cogMethod cmType = CMClosedPIC]) ifTrue:
		[cogMethod blockEntryOffset: 0]! !

!CogMethodZone methodsFor: 'jit - api' stamp: 'eem 7/25/2021 14:36'!
cogMethodContaining: mcpc
	"Answer the method containing mcpc for the purposes of code zone compaction,
	 where mcpc is actually the value of instructionPointer at the time of a compaction."
	<var: 'mcpc' type: #usqInt>
	<api>
	| cogMethod prevMethod |
	mcpc > limitAddress ifTrue:
		[^nil].
	mcpc < baseAddress ifTrue:
		[cogit assertMcpcIsPrimReturn: mcpc.
		 ^nil].
	self assert: mcpc < self freeStart.
	cogMethod := coInterpreter cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod asUnsignedInteger < mcpc] whileTrue:
		[prevMethod := cogMethod.
		 cogMethod := self methodAfter: cogMethod].

	"Since mcpc is actually instructionPointer we expect that it is either at the stack check
	 (normal code zone reclamation invoked through checkForEventsMayContextSwitch:)
	 or is in a primitive, immediately following the call of the C primitive routine."
	self assert: (prevMethod notNil
				and: [mcpc = (prevMethod asUnsignedInteger + prevMethod stackCheckOffset)
					or: [(self mcpc: mcpc isAtStackCheckOfBlockMethodIn: prevMethod)
					or: [(coInterpreter
							primitiveIndexOfMethod: prevMethod methodObject
							header: prevMethod methodHeader) > 0
					or: [(cogit backEnd isCallPrecedingReturnPC: mcpc)
						and: [(cogit backEnd callTargetFromReturnAddress: mcpc) = cogit ceCheckForInterruptTrampoline]]]]]).
	 ^prevMethod! !

!CogMethodZone methodsFor: 'printing' stamp: 'eem 6/22/2009 09:38'!
cogMethodsSelect: aBlock
	<doNotGenerate>
	| cogMethod cogMethods |
	cogMethods := OrderedCollection new: 256.
	cogMethod := coInterpreter cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod < self limitZony] whileTrue:
		[(aBlock value: cogMethod) ifTrue:
			[cogMethods addLast: cogMethod].
		 cogMethod := self methodAfter: cogMethod].
	^cogMethods! !

!CogMethodZone methodsFor: 'simulation only' stamp: 'eem 8/13/2012 10:18'!
cogit
	"This is for the sizeof: CogMethod hook that allows different cogit classes to use differet CogMethod variants."
	<doNotGenerate>
	^cogit! !

!CogMethodZone methodsFor: 'compaction' stamp: 'eem 2/3/2020 20:50'!
compactCompiledCode
	| objectHeaderValue source dest writableVersion bytes |
	<var: #source type: #'CogMethod *'>
	<var: #dest type: #'CogMethod *'>
	compactionInProgress := true.
	methodCount := 0.
	objectHeaderValue := objectMemory nullHeaderForMachineCodeMethod.
	source := coInterpreter cCoerceSimple: baseAddress to: #'CogMethod *'.
	self voidOpenPICList. "The(se) list(s) will be rebuilt with the current live set"
	self voidUnpairedMethodList.
	[source < self limitZony
	 and: [source cmType ~= CMFree]] whileTrue:
		[self assert: (cogit cogMethodDoesntLookKosher: source) = 0.
		 writableVersion := cogit writableMethodFor: source. 
		 writableVersion objectHeader: objectHeaderValue.
		 source cmUsageCount > 0 ifTrue:
			[writableVersion cmUsageCount: source cmUsageCount // 2].
		 self maybeLinkOnUnpairedMethodList: source.
		 self clearSavedPICUsageCount: writableVersion.
		 source cmType = CMOpenPIC ifTrue:
			[self addToOpenPICList: writableVersion].
		 methodCount := methodCount + 1.
		 source := self methodAfter: source].
	source >= self limitZony ifTrue:
		[^self halt: 'no free methods; cannot compact.'].
	dest := source.
	[source < self limitZony] whileTrue:
		[self assert: (cogit maybeFreeCogMethodDoesntLookKosher: source) = 0.
		 bytes := source blockSize.
		 source cmType ~= CMFree ifTrue:
			[methodCount := methodCount + 1.
			 cogit
				codeMemmove: dest _: source _: bytes;
				maybeFlushWritableZoneFrom: dest asUnsignedInteger to: dest asUnsignedInteger + bytes.
			 (writableVersion := cogit writableMethodFor: dest) objectHeader: objectHeaderValue.
			 dest cmType = CMMethod
				ifTrue:
					["For non-Newspeak there should be a one-to-one mapping between bytecoded and
					  cog methods.  For Newspeak not necessarily, but only for anonymous accessors."
					"Only update the original method's header if it is referring to this CogMethod."
					 (coInterpreter rawHeaderOf: dest methodObject) asInteger = source asInteger
						ifTrue:
							[coInterpreter rawHeaderOf: dest methodObject put: dest asInteger]
						ifFalse:
							[self assert: (cogit noAssertMethodClassAssociationOf: dest methodObject) = objectMemory nilObject.
							 self linkOnUnpairedMethodList: dest]]
				ifFalse:
					[self clearSavedPICUsageCount: writableVersion.
					 dest cmType = CMOpenPIC ifTrue:
						[self addToOpenPICList: writableVersion]].
			 dest cmUsageCount > 0 ifTrue:
				[writableVersion cmUsageCount: dest cmUsageCount // 2].
			 cogit maybeFlushWritableZoneFrom: dest asUnsignedInteger to: (dest + 1) asUnsignedInteger.
			 dest := coInterpreter cCoerceSimple: dest asUnsignedInteger + bytes to: #'CogMethod *'].
		 source := coInterpreter cCoerceSimple: source asUnsignedInteger + bytes to: #'CogMethod *'].
	mzFreeStart := dest asUnsignedInteger.
	methodBytesFreedSinceLastCompaction := 0.
	compactionInProgress := false! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 2/17/2017 15:30'!
compactionInProgress
	^compactionInProgress! !

!CogMethodZone methodsFor: 'allocating' stamp: 'eem 3/17/2017 15:21'!
effectiveLimit
	<inline: true>
	^limitAddress! !

!CogMethodZone methodsFor: 'young referers' stamp: 'eem 7/20/2021 17:10'!
ensureInYoungReferrers: cogMethod
	<var: #cogMethod type: #'CogMethod *'>
	cogit assertValidDualZoneReadAddress: cogMethod.
	cogMethod cmRefersToYoung ifFalse:
		[| writableMethod |
		 self assert: (self occurrencesInYoungReferrers: cogMethod) = 0.
		 cogit ensureWritableCodeZone.
		 (writableMethod := cogit writableMethodFor: cogMethod) cmRefersToYoung: true.
		 self addToYoungReferrers: writableMethod]! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 7/9/2014 19:58'!
findPreviouslyCompiledVersionOf: aMethodObj with: aSelectorOop
	"Newspeak uses a set of methods to implement accessors, a setter and a getter for
	 each inst var offset (e.g. 0 to 255).  These accessors are installed under the relevant
	 selectors in different method dictionaries as required.  These methods effectively
	 have multiple selectors.  The current inline cache design stores the selector of a
	 linked send in the header of the target method.  So this requires a one-to-many
	 mapping of bytecoded method to cog method, with the bytecoded method referring
	 directly to only one cog method, which will have a specific selector, not necessarily
	 the right one.  It is therefore worth-while searching for a cog method on this bytecoded
	 method that has the right selector.  To speed up the search we maintain all such unpaired
	 methods on the unpairedMethodList, which is linked through nextMethodOrIRCs."
	<returnTypeC: #'CogMethod *'>
	<option: #NewspeakVM>
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	((coInterpreter methodHasCogMethod: aMethodObj)
	 and: [(coInterpreter methodClassAssociationOf: aMethodObj) = objectMemory nilObject]) ifTrue:
		[cogMethod := self cCoerceSimple: unpairedMethodList to: #'CogMethod *'.
		[cogMethod notNil] whileTrue:
			[self assert: cogMethod cmType = CMMethod.
			 (cogMethod selector = aSelectorOop
			  and: [cogMethod methodObject = aMethodObj]) ifTrue:
				[^cogMethod].
			 cogMethod := self cCoerceSimple: cogMethod nextMethodOrIRCs to: #'CogMethod *']].
	^nil! !

!CogMethodZone methodsFor: 'young referers' stamp: 'eem 9/9/2019 11:56'!
firstBogusYoungReferrer
	"Answer that all entries in youngReferrers are in-use and have the cmRefersToYoung flag set.
	 Used to check that the youngreferrers pruning routines work correctly."
	| pointer cogMethod |
	<doNotGenerate>
	(youngReferrers > limitAddress
	 or: [youngReferrers < mzFreeStart]) ifTrue:
		[^#invalidListPointers].
	pointer := youngReferrers.
	[pointer < limitAddress] whileTrue:
		[cogMethod := coInterpreter cCoerceSimple: (objectMemory longAt: pointer) to: #'CogMethod *'.
		 cogMethod cmType ~= CMFree ifTrue:
			[cogMethod cmRefersToYoung ifFalse:
				[^cogMethod].
			 (self occurrencesInYoungReferrers: cogMethod) ~= 1 ifTrue:
				[^cogMethod]].
		 pointer := pointer + objectMemory wordSize].
	cogMethod := cogit cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod < mzFreeStart] whileTrue:
		[cogMethod cmType ~= CMFree ifTrue:
			[(self occurrencesInYoungReferrers: cogMethod) ~= (cogMethod cmRefersToYoung ifTrue: [1] ifFalse: [0]) ifTrue:
				[^cogMethod]].
		 cogMethod := self methodAfter: cogMethod]..
	^nil! !

!CogMethodZone methodsFor: 'jit - api' stamp: 'eem 9/9/2019 11:05'!
followForwardedLiteralsInOpenPICList
	<option: #SpurObjectMemory>
	| openPIC |
	<var: #openPIC type: #'CogMethod *'>
	openPIC := openPICList.
	[openPIC notNil] whileTrue:
		[cogit followForwardedLiteralsIn: openPIC.
		 openPIC := self cCoerceSimple: openPIC nextOpenPIC to: #'CogMethod *'].
	self pruneYoungReferrers! !

!CogMethodZone methodsFor: 'compaction' stamp: 'eem 7/9/2021 15:13'!
freeMethod: cogMethod
	<var: #cogMethod type: #'CogMethod *'>
	<inline: false>
	| writableMethod |
	self assert: cogMethod cmType ~= CMFree.
	self assert: (cogit cogMethodDoesntLookKosher: cogMethod) = 0.
	cogit ensureWritableCodeZone.
	cogMethod cmType = CMMethod ifTrue:
		["For non-Newspeak there should be a one-to-one mapping between bytecoded and
		  cog methods.  For Newspeak not necessarily, but only for anonymous accessors."
		"Only reset the original method's header if it is referring to this CogMethod."
		 (coInterpreter rawHeaderOf: cogMethod methodObject) asInteger = cogMethod asInteger
			ifTrue:
				[coInterpreter rawHeaderOf: cogMethod methodObject put: cogMethod methodHeader.
				 NewspeakVM ifTrue:
					[(objectRepresentation canPinObjects and: [cogMethod nextMethodOrIRCs > self zoneEnd]) ifTrue:
						[objectRepresentation freeIRCs: cogMethod nextMethodOrIRCs]]]
			ifFalse:
				[self cCode: [self assert: (cogit noAssertMethodClassAssociationOf: cogMethod methodObject) = objectMemory nilObject]
					inSmalltalk: [self assert: ((cogit noAssertMethodClassAssociationOf: cogMethod methodObject) = objectMemory nilObject
											or: [coInterpreter isKindOf: CurrentImageCoInterpreterFacade])].
				 NewspeakVM ifTrue:
					[self removeFromUnpairedMethodList: cogMethod]].
		 cogit maybeFreeCountersOf: cogMethod].
	cogMethod cmType = CMOpenPIC ifTrue:
		[self removeFromOpenPICList: cogMethod].
	writableMethod := cogit writableMethodFor: cogMethod.
	writableMethod cmRefersToYoung: false.
	writableMethod cmType: CMFree.
	methodBytesFreedSinceLastCompaction := methodBytesFreedSinceLastCompaction
												+ cogMethod blockSize! !

!CogMethodZone methodsFor: 'compaction' stamp: 'eem 2/21/2018 18:34'!
freeOlderMethodsForCompaction
	"Free methods, preferring older methods for compaction, up to some fraction, currently a quarter."
	| zoneSize amountToFree initialFreeSpace freedSoFar freeableUsage cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	zoneSize := self effectiveLimit - baseAddress.
	initialFreeSpace := self effectiveLimit - mzFreeStart + methodBytesFreedSinceLastCompaction.
	freedSoFar := initialFreeSpace.
	amountToFree := zoneSize // 4. "4 needs to be e.g. a start-up parameter"
	freeableUsage := 0.
	[self cCode: ''
		inSmalltalk: [coInterpreter transcript nextPutAll: 'freeing methods with usage '; print: freeableUsage; cr; flush].
	 cogMethod := cogit cCoerceSimple: baseAddress to: #'CogMethod *'.
	 [cogMethod asUnsignedInteger < mzFreeStart
	  and: [freedSoFar < amountToFree]] whileTrue:
		[(self shouldFreeMethod: cogMethod given: freeableUsage) ifTrue:
			[self freeMethod: cogMethod.
			 freedSoFar := freedSoFar + cogMethod blockSize].
		 cogMethod := self methodAfter: cogMethod].
	 freedSoFar < amountToFree
	 and: [(freeableUsage := freeableUsage + 1) < CMMaxUsageCount]] whileTrue.
	self cCode: ''
		inSmalltalk: [coInterpreter transcript
						nextPutAll: 'Code Compaction freeing '; print: freedSoFar;
						nextPutAll: ' of '; print: zoneSize;
						nextPutAll: ' (target: '; print: amountToFree;
						nextPutAll: ' (newly freed: '; print: freedSoFar - initialFreeSpace;
						cr; flush]! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 12/3/2013 13:41'!
freeStart
	"declared as a macro so as not to conflict with the memory manager's freeStart."
	<cmacro: '() mzFreeStart'>
	^mzFreeStart! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 10/23/2015 14:25:46'!
freeStart: zoneLimit
	<doNotGenerate>
	^baseAddress := mzFreeStart := zoneLimit! !

!CogMethodZone methodsFor: 'initialization' stamp: 'eem 2/17/2017 15:25'!
initialize
	"Make youngReferrers arithmetic for addressIsInCodeZone:"
	youngReferrers := methodCount := methodBytesFreedSinceLastCompaction := 0.
	compactionInProgress := false! !

!CogMethodZone methodsFor: 'young referers' stamp: 'eem 7/5/2021 17:23'!
kosherYoungReferrers
	"Answer that all entries in youngReferrers are in-use and have the cmRefersToYoung flag set.
	 Used to check that the youngreferrers pruning routines work correctly."
	<api>
	| pointer cogMethod prevMethod |
	<var: #pointer type: #usqInt>
	(youngReferrers > limitAddress
	 or: [youngReferrers < mzFreeStart]) ifTrue:
		[^false].
	pointer := youngReferrers.
	[pointer < limitAddress] whileTrue:
		[cogMethod := coInterpreter cCoerceSimple: (objectMemory longAt: pointer) to: #'CogMethod *'.
		 cogMethod cmType ~= CMFree ifTrue:
			[cogMethod cmRefersToYoung ifFalse:
				[^false].
			 (self occurrencesInYoungReferrers: cogMethod) ~= 1 ifTrue:
				[^false]].
		 pointer := pointer + objectMemory wordSize].
	cogMethod := cogit cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod < self limitZony] whileTrue:
		[prevMethod := cogMethod.
		 cogMethod cmType ~= CMFree ifTrue:
			[(self occurrencesInYoungReferrers: cogMethod) ~= (cogMethod cmRefersToYoung ifTrue: [1] ifFalse: [0]) ifTrue:
				[^false]].
		 cogMethod := self methodAfter: cogMethod.
		 cogMethod = prevMethod ifTrue:
			[^false]].
	^true! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 6/7/2015 11:37'!
limitZony "Let us salute Andrei Tarkovsky and his movie Stalker"
	<cmacro: '() ((CogMethod *)mzFreeStart)'>
	^mzFreeStart! !

!CogMethodZone methodsFor: 'compaction' stamp: 'eem 1/29/2020 15:03'!
linkOnUnpairedMethodList: cogMethod
	<inline: #always>
	NewspeakVM ifTrue:
		[cogMethod nextMethodOrIRCs: unpairedMethodList.
		 unpairedMethodList := cogMethod asUnsignedInteger]! !

!CogMethodZone methodsFor: 'initialization' stamp: 'eem 10/9/2019 14:36'!
manageFrom: theStartAddress to: theLimitAddress
	<returnTypeC: #void>
	<inline: true>
	mzFreeStart := baseAddress := theStartAddress.
	youngReferrers := limitAddress := theLimitAddress.
	openPICList := nil.
	NewspeakVM ifTrue: [unpairedMethodList := nil].
	methodBytesFreedSinceLastCompaction := 0.
	methodCount := 0! !

!CogMethodZone methodsFor: 'compaction' stamp: 'eem 1/29/2020 14:55'!
maybeLinkOnUnpairedMethodList: cogMethod
	<inline: #always>
	NewspeakVM ifTrue:
		[(cogMethod cmType = CMMethod
		  and: [(coInterpreter rawHeaderOf: cogMethod methodObject) asInteger ~= cogMethod asInteger]) ifTrue:
			[(cogit writableMethodFor: cogMethod) nextMethodOrIRCs: unpairedMethodList.
			 unpairedMethodList := cogMethod asUnsignedInteger]]! !

!CogMethodZone methodsFor: 'testing' stamp: 'eem 6/18/2021 20:16'!
mcpc: mcpc isAtStackCheckOfBlockMethodIn: cogMethod
	"For assert checking..."
	<var: 'cogMethod' type: #'CogMethod *'>
	cogMethod blockEntryOffset = 0 ifTrue:
		[^false].
	^(cogit blockDispatchTargetsFor: cogMethod perform: #stackCheckOffsetOfBlockAt:isMcpc: arg: mcpc) ~= 0! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 3/12/2014 13:01'!
methodAfter: cogMethod
	<var: #cogMethod type: #'CogMethod *'>
	<inline: true>
	^coInterpreter
		cCoerceSimple: (self roundUpLength: cogMethod asInteger + cogMethod blockSize)
		to: #'CogMethod *'! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 1/21/2014 15:06'!
methodBytesFreedSinceLastCompaction
	<cmacro: '() methodBytesFreedSinceLastCompaction'>
	^methodBytesFreedSinceLastCompaction! !

!CogMethodZone methodsFor: 'jit - api' stamp: 'eem 2/24/2020 08:20'!
methodFor: address
	<api>
	<returnTypeC: #'CogMethod *'>
	<var: #address type: #'void *'>
	| cogMethod nextMethod |
	cogMethod := coInterpreter cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod < self limitZony
	 and: [cogMethod asUnsignedInteger <= address asUnsignedInteger]] whileTrue:
		[nextMethod := self methodAfter: cogMethod.
		 nextMethod = cogMethod ifTrue:
			[^nil].
		 (address asUnsignedInteger >= cogMethod asUnsignedInteger
		  and: [address asUnsignedInteger < nextMethod asUnsignedInteger]) ifTrue:
			[^cogMethod].
		 cogMethod := nextMethod].
	^nil! !

!CogMethodZone methodsFor: 'simulation only' stamp: 'eem 12/11/2010 00:14'!
methods
	<doNotGenerate>
	| methods |
	methods := OrderedCollection new.
	self methodsDo:
		[:m| methods addLast: m].
	^methods! !

!CogMethodZone methodsFor: 'method zone introspection' stamp: 'cb 4/26/2016 09:17'!
methodsCompiledToMachineCodeInto: arrayObj 
	<api>
	<var: #cogMethod type: #'CogMethod *'>
	| cogMethod methodIndex |
	methodIndex := 0.
	cogMethod := cogit cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod < self limitZony] whileTrue:
		[cogMethod cmType = CMMethod ifTrue:
			[objectMemory 
				storePointerUnchecked: methodIndex 
				ofObject: arrayObj 
				withValue: cogMethod methodObject.
			 methodIndex := methodIndex + 1].
		 cogMethod := self methodAfter: cogMethod].
	^ methodIndex
	! !

!CogMethodZone methodsFor: 'simulation only' stamp: 'eem 12/10/2009 10:28'!
methodsDo: aBlock
	<doNotGenerate>
	| cogMethod |
	cogMethod := cogit cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod < self limitZony] whileTrue:
		[cogMethod cmType ~= CMFree ifTrue:
			[aBlock value: cogMethod].
		 cogMethod := self methodAfter: cogMethod]
	"<api>
	<returnTypeC: #void>
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := cogit cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod < self limitZony] whileTrue:
		[cogMethod cmType ~= CMFree ifTrue:
			[aBlock value: cogMethod].
		 cogMethod := self methodAfter: cogMethod]"! !

!CogMethodZone methodsFor: 'accessing' stamp: 'cb 4/25/2016 17:24'!
numMethods
	<api>
	^methodCount! !

!CogMethodZone methodsFor: 'jit - api' stamp: 'eem 7/19/2010 09:39'!
numMethodsOfType: cogMethodType
	<api>
	| n cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	n := 0.
	cogMethod := cogit cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod < self limitZony] whileTrue:
		[cogMethod cmType = cogMethodType ifTrue:
			[n := n + 1].
		 cogMethod := self methodAfter: cogMethod].
	^n! !

!CogMethodZone methodsFor: 'young referers' stamp: 'eem (const to msg rewrite) 11/6/2014 11:51'!
occurrencesInYoungReferrers: cogMethod
	<var: #cogMethod type: #'CogMethod *'>
	| count pointer |
	<var: #pointer type: #usqInt>
	self assert: youngReferrers <= limitAddress.
	count := 0.
	pointer := youngReferrers.
	[pointer < limitAddress] whileTrue:
		[cogMethod asInteger = (objectMemory longAt: pointer) ifTrue:
			[count := count + 1].
		 pointer := pointer + objectMemory wordSize].
	^count! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 7/9/2009 21:07'!
openPICWithSelector: aSelector
	| openPIC |
	<var: #openPIC type: #'CogMethod *'>
	<returnTypeC: #'CogMethod *'>
	openPIC := openPICList.
	[(openPIC == nil
	  or: [openPIC selector = aSelector]) ifTrue:
		[^openPIC].
	 "N.B. Use self rather than coInterpreter to avoid attempting to cast nil.
	  Conversion to CogMethod done in the nextOpenPIC accessor."
	 openPIC := self cCoerceSimple: openPIC nextOpenPIC to: #'CogMethod *'.
	 true] whileTrue! !

!CogMethodZone methodsFor: 'compaction' stamp: 'eem 1/29/2020 09:39'!
planCompaction
	"Some methods have been freed.  Compute how much each survivor needs to
	 move during the ensuing compaction and record it in the objectHeader field.

	 For Sista, where we want PICs to last so they can be observed, we need to keep PICs unless
	 they are definitely unused.  So we need to identify unused PICs.  So in planCompact, zero the
	 usage counts of all PICs, saving the actual usage count in blockEntryOffset.  Then in
	 relocateMethodsPreCompaction (actually in relocateIfCallOrMethodReference:mcpc:delta:)
	 restore the usage counts of used PICs.  Finally in compactCompiledCode, clear the blockEntryOffset
	 of the unused PICs; they will then have a zero count and be reclaimed in the next code compaction."
	| delta cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	delta := 0.
	cogMethod := cogit cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod asUnsignedInteger < mzFreeStart] whileTrue:
		[cogMethod cmType = CMFree
			ifTrue: [delta := delta - cogMethod blockSize]
			ifFalse:
				[self assert: (cogit cogMethodDoesntLookKosher: cogMethod) = 0.
				 (cogit writableMethodFor: cogMethod) objectHeader: delta.
				 SistaVM ifTrue:
					[self savePICUsageCount: cogMethod]].
		 cogMethod := self methodAfter: cogMethod]! !

!CogMethodZone methodsFor: 'printing' stamp: 'eem 9/5/2019 10:37'!
printCogMethods
	<api>
	<returnTypeC: #void>
	self printCogMethodsSummarizing: false! !

!CogMethodZone methodsFor: 'printing' stamp: 'eem 6/22/2009 09:13'!
printCogMethodsOfType: cmType
	<api>
	<returnTypeC: #void>
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := coInterpreter cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod < self limitZony] whileTrue:
		[cogMethod cmType = cmType ifTrue:
			[coInterpreter printCogMethod: cogMethod].
		 cogMethod := self methodAfter: cogMethod]! !

!CogMethodZone methodsFor: 'printing' stamp: 'eem 9/5/2019 13:43'!
printCogMethodsSummarizing: summarize
	<inline: true>
	| cogMethod nm nc no nf nu |
	<var: #cogMethod type: #'CogMethod *'>
	nm := nc := no := nf := nu := 0.
	cogMethod := coInterpreter cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod < self limitZony] whileTrue:
		[summarize ifFalse:
			[coInterpreter printCogMethod: cogMethod].
		 cogMethod cmType
			caseOf: {
			[CMFree]		->	[nf := nf + 1].
			[CMMethod]	->	[nm := nm + 1].
			[CMClosedPIC]	->	[nc := nc + 1].
			[CMOpenPIC]	->	[no:= no+ 1] }
			otherwise: [nu := nu + 1].
		 cogMethod := self methodAfter: cogMethod].
	coInterpreter print: 'CMMethod '; printNum: nm;  print: ' CMClosedPIC '; printNum: nc;  print: ' CMOpenPIC '; printNum: no;  print: ' CMFree '; printNum: nf.
	nu > 0 ifTrue:
		[coInterpreter print: ' UNKNOWN '; printNum: nu].
	coInterpreter print: ' total '; printNum: nm+nc+no+nf+nu; cr! !

!CogMethodZone methodsFor: 'printing' stamp: 'eem 8/6/2011 16:16'!
printCogMethodsWithMethod: methodOop
	<api>
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := coInterpreter cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod < self limitZony] whileTrue:
		[(cogMethod cmType ~= CMFree
		  and: [cogMethod methodObject = methodOop]) ifTrue:
			[coInterpreter printCogMethod: cogMethod].
		 cogMethod := self methodAfter: cogMethod]! !

!CogMethodZone methodsFor: 'printing' stamp: 'eem 2/14/2017 17:08'!
printCogMethodsWithPrimitive: primIdx
	<api>
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := coInterpreter cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod < self limitZony] whileTrue:
		[(cogMethod cmType = CMMethod
		  and: [primIdx = (coInterpreter primitiveIndexOfMethod: cogMethod methodObject
							header: cogMethod methodHeader)]) ifTrue:
			[coInterpreter printCogMethod: cogMethod].
		 cogMethod := self methodAfter: cogMethod]! !

!CogMethodZone methodsFor: 'printing' stamp: 'eem 9/24/2009 10:41'!
printCogMethodsWithSelector: selectorOop
	<api>
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := coInterpreter cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod < self limitZony] whileTrue:
		[(cogMethod cmType ~= CMFree
		  and: [cogMethod selector = selectorOop]) ifTrue:
			[coInterpreter printCogMethod: cogMethod].
		 cogMethod := self methodAfter: cogMethod]! !

!CogMethodZone methodsFor: 'printing' stamp: 'eem (const to msg rewrite) 11/6/2014 11:51'!
printCogYoungReferrers
	<api>
	<returnTypeC: #void>
	| pointer cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	pointer := youngReferrers.
	[pointer < limitAddress] whileTrue:
		[cogMethod := coInterpreter cCoerceSimple: (objectMemory longAt: pointer) to: #'CogMethod *'.
		 cogMethod cmRefersToYoung ifFalse:
			[coInterpreter print: '*'].
		 cogMethod cmType = CMFree ifTrue:
			[coInterpreter print: '!!'].
		 (cogMethod cmRefersToYoung and: [cogMethod cmType ~= CMFree]) ifFalse:
			[coInterpreter print: ' '].
		 coInterpreter printCogMethod: cogMethod.
		 pointer := pointer + objectMemory wordSize]! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 9/5/2019 10:40'!
printOpenPICList
	<api>
	^self printOpenPICListSummarizing: false! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 9/5/2019 13:42'!
printOpenPICListSummarizing: summarize
	<inline: true>
	| n openPIC |
	<var: #openPIC type: #'CogMethod *'>
	n := 0.
	openPIC := openPICList.
	[openPIC == nil] whileFalse:
		[n := n + 1.
		 summarize ifFalse:
			[coInterpreter printCogMethod: openPIC].
		 openPIC := self cCoerceSimple: openPIC nextOpenPIC to: #'CogMethod *'].
	summarize ifTrue:
		[coInterpreter print: 'num open PICs '; printNum: n; cr].
	^n! !

!CogMethodZone methodsFor: 'young referers' stamp: 'eem 7/9/2021 15:43'!
pruneYoungReferrers
	| source dest next |
	<api>
	<var: #source type: #usqInt>
	<var: #dest type: #usqInt>
	<var: #next type: #usqInt>
	<inline: false>

	self assert: youngReferrers <= limitAddress.
	youngReferrers = limitAddress ifTrue:
		[^nil].
	dest := limitAddress.
	[next := dest - objectMemory wordSize.
	 next >= youngReferrers
	 and: [(coInterpreter cCoerceSimple: (objectMemory longAt: next) to: #'CogMethod *') cmRefersToYoung]] whileTrue:
		[dest := next].
	self assert: dest >= youngReferrers.
	source := dest - objectMemory wordSize.
	[source >= youngReferrers] whileTrue:
		[(coInterpreter cCoerceSimple: (objectMemory longAt: source) to: #'CogMethod *') cmRefersToYoung ifTrue:
			[self assert: source < (dest - objectMemory wordSize).
			 next ifNotNil: "convenient first-time flag"
				[next := nil.
				 cogit ensureWritableCodeZone].
			 cogit codeLongAt: (dest := dest - objectMemory wordSize) put: (objectMemory longAt: source)].
		 source := source - objectMemory wordSize].
	youngReferrers := dest.
	self assert: self kosherYoungReferrers! !

!CogMethodZone methodsFor: 'jit - api' stamp: 'eem 1/25/2018 11:12'!
quickClearCogCompiledCode
	"Quick, for in-image testing"
	<doNotGenerate>
	mzFreeStart := baseAddress.
	self clearCogCompiledCode.
	cogit backEnd stopsFrom: mzFreeStart to: limitAddress - 1! !

!CogMethodZone methodsFor: 'young referers' stamp: 'eem 1/29/2020 10:24'!
relocateAndPruneYoungReferrers
	| source dest next cogMethod |
	<var: #source type: #usqInt>
	<var: #dest type: #usqInt>
	<var: #next type: #usqInt>
	<var: #cogMethod type: #'CogMethod *'>
	<inline: false>

	self assert: youngReferrers <= limitAddress.
	youngReferrers = limitAddress ifTrue:
		[^nil].
	dest := limitAddress.
	[next := dest - objectMemory wordSize.
	 next >= youngReferrers
	 and: [(cogMethod := coInterpreter cCoerceSimple: (objectMemory longAt: next) to: #'CogMethod *') cmType ~= CMFree
	 and: [cogMethod cmRefersToYoung]]] whileTrue:
		[cogMethod objectHeader ~= 0 ifTrue:
			[cogit codeLongAt: next put: cogMethod asInteger + cogMethod objectHeader].
		 dest := next].
	self assert: dest >= youngReferrers.
	source := dest - objectMemory wordSize.
	[source >= youngReferrers] whileTrue:
		[cogMethod := coInterpreter cCoerceSimple: (objectMemory longAt: source) to: #'CogMethod *'.
		 (cogMethod cmType ~= CMFree
		  and: [cogMethod cmRefersToYoung]) ifTrue:
			[self assert: source < (dest - objectMemory wordSize).
			 cogMethod objectHeader ~= 0 ifTrue:
				[cogMethod := coInterpreter
									cCoerceSimple: cogMethod asInteger + cogMethod objectHeader asInteger
									to: #'CogMethod *'].
			 cogit codeLongAt: (dest := dest - objectMemory wordSize) put: cogMethod asInteger].
		 source := source - objectMemory wordSize].
	youngReferrers := dest.
	"this assert must be deferred until after compaction.  See the end of compactCogCompiledCode"
	"self assert: self kosherYoungReferrers"! !

!CogMethodZone methodsFor: 'compaction' stamp: 'eem 6/19/2015 08:38'!
relocateMethodsPreCompaction
	"All surviving methods have had the amount they are going to relocate by
	 stored in their objectHeader fields.  Relocate all relative calls so that after
	 the compaction of both the method containing each call and the call target
	 the calls invoke the same target."
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := cogit cCoerceSimple: baseAddress to: #'CogMethod *'.
	[cogMethod asUnsignedInteger < mzFreeStart] whileTrue:
		[cogMethod cmType ~= CMFree ifTrue:
			[cogMethod cmType = CMClosedPIC
				ifTrue: [cogit relocateCallsInClosedPIC: cogMethod]
				ifFalse: [cogit relocateCallsAndSelfReferencesInMethod: cogMethod]].
		 cogMethod := self methodAfter: cogMethod].
	self relocateAndPruneYoungReferrers.
	^true! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 1/29/2020 09:55'!
removeFromOpenPICList: anOpenPIC
	<var: #anOpenPIC type: #'CogMethod *'>
	| prevPIC |
	<var: #prevPIC type: #'CogMethod *'>
	self assert: anOpenPIC cmType = CMOpenPIC.
	openPICList ifNil: [^nil]. "As it is when compacting or unlinking all sends"
	self assert: (openPICList cmType = CMOpenPIC
				and: [openPICList nextOpenPIC isNil
					or: [(self cCoerceSimple: openPICList nextOpenPIC to: #'CogMethod *') cmType = CMOpenPIC]]).
	anOpenPIC = openPICList ifTrue:
		["N.B. Use self rather than coInterpreter to avoid attempting to cast nil.
		  Conversion to CogMethod done in the nextOpenPIC accessor."
		 openPICList := self cCoerceSimple: anOpenPIC nextOpenPIC to: #'CogMethod *'.
		 ^nil].
	prevPIC := openPICList.
	[self assert: (prevPIC ~~ nil
				and: [prevPIC cmType = CMOpenPIC]).
	 prevPIC nextOpenPIC = anOpenPIC asUnsignedInteger ifTrue:
		[(cogit writableMethodFor: prevPIC) nextOpenPIC: anOpenPIC nextOpenPIC.
		 ^nil].
	  prevPIC := self cCoerceSimple: prevPIC nextOpenPIC to: #'CogMethod *'.
	  true] whileTrue! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 7/9/2014 19:45'!
removeFromUnpairedMethodList: aCogMethod
	<option: #NewspeakVM>
	<var: #aCogMethod type: #'CogMethod *'>
	| prevMethod |
	<var: #prevMethod type: #'CogMethod *'>
	self assert: aCogMethod cmType = CMMethod.
	aCogMethod asUnsignedInteger = unpairedMethodList ifTrue:
		[unpairedMethodList := aCogMethod nextMethodOrIRCs.
		 ^nil].
	prevMethod := self cCoerceSimple: unpairedMethodList to: #'CogMethod *'.
	[prevMethod notNil] whileTrue:
		[self assert: (prevMethod ~~ nil and: [prevMethod cmType = CMMethod]).
		 prevMethod nextMethodOrIRCs = aCogMethod asUnsignedInteger ifTrue:
			[prevMethod nextMethodOrIRCs: aCogMethod nextMethodOrIRCs.
			 ^nil].
		  prevMethod := self cCoerceSimple: prevMethod nextMethodOrIRCs to: #'CogMethod *']! !

!CogMethodZone methodsFor: 'compaction' stamp: 'eem 6/3/2015 19:16'!
restorePICUsageCount: cogMethod
	"For Sista, where we want PICs to last so they can be observed, we need to keep PICs unless
	 they are definitely unused.  So we need to identify unused PICs.  So in planCompact, zero the
	 usage counts of all PICs, saving the actual usage count in blockEntryOffset.  Then in
	 relocateMethodsPreCompaction (actually in relocateIfCallOrMethodReference:mcpc:delta:)
	 restore the usage counts of used PICs.  Finally in compactCompiledCode, clear the blockEntryOffset
	 of the unused PICs; they will then have a zero count and be reclaimed in the next code compaction."
	<var: #cogMethod type: #'CogMethod *'>
	<option: #SistaVM>
	(cogMethod cmType = CMClosedPIC
	 and: [cogMethod blockEntryOffset ~= 0]) ifTrue:
		[cogMethod
			cmUsageCount: cogMethod blockEntryOffset;
			blockEntryOffset: 0]! !

!CogMethodZone methodsFor: 'accessing' stamp: 'nice 7/8/2016 03:04'!
roundUpAddress: address
	<returnTypeC: #'void *'>
	<var: #address type: #'void *'>
	^self cCoerce: ((self cCoerce: address to: #'usqIntptr_t') + 7 bitAnd: -8) to: #'void *'! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 12/13/2019 15:11'!
roundUpLength: numBytes
	"Determine the default alignment for the start of a CogMethod, which in turn
	 determines the size of the mask used to distinguish the checked and unchecked
	 entry-points, used to distinguish normal and super sends on method unlinking.
	 This is passed onto the backEnd to allow processors with coarse instructions
	 (ARM) to increase the alignment if required."
	<inline: true>
	^cogit backEnd roundUpToMethodAlignment: numBytes! !

!CogMethodZone methodsFor: 'compaction' stamp: 'eem 6/3/2015 19:16'!
savePICUsageCount: cogMethod
	"For Sista, where we want PICs to last so they can be observed, we need to keep PICs unless
	 they are definitely unused.  So we need to identify unused PICs.  So in planCompact, zero the
	 usage counts of all PICs, saving the actual usage count in blockEntryOffset.  Then in
	 relocateMethodsPreCompaction (actually in relocateIfCallOrMethodReference:mcpc:delta:)
	 restore the usage counts of used PICs.  Finally in compactCompiledCode, clear the blockEntryOffset
	 of the unused PICs; they will then have a zero count and be reclaimed in the next code compaction."
	<var: #cogMethod type: #'CogMethod *'>
	<option: #SistaVM>
	cogMethod cmType = CMClosedPIC ifTrue:
		[cogMethod
			blockEntryOffset: cogMethod cmUsageCount;
			cmUsageCount: 0]! !

!CogMethodZone methodsFor: 'initialization' stamp: 'eem 7/5/2011 16:22'!
setInterpreter: aCoInterpreter objectRepresentation: anObjectRepresentation cogit: aCogit
	<doNotGenerate>
	coInterpreter := aCoInterpreter.
	objectRepresentation := anObjectRepresentation.
	cogit := aCogit.
	objectMemory := aCoInterpreter objectMemory! !

!CogMethodZone methodsFor: 'compaction' stamp: 'eem 6/3/2015 19:17'!
shouldFreeMethod: cogMethod given: freeableUsage
	"Answer if cogMethod should be freed in the current pass of freeOlderMethodsForCompaction.

	 For Sista, where we want PICs to last so they can be observed, we need to keep PICs unless
	 they are definitely unused.  So we need to identify unused PICs.  So in planCompact, zero the
	 usage counts of all PICs, saving the actual usage count in blockEntryOffset.  Then in
	 relocateMethodsPreCompaction (actually in relocateIfCallOrMethodReference:mcpc:delta:)
	 restore the usage counts of used PICs.  Finally in compactCompiledCode, clear the blockEntryOffset
	 of the unused PICs; they will then have a zero count and be reclaimed in the next code compaction."
	<var: #cogMethod type: #'CogMethod *'>
	<inline: true>
	^SistaVM
		ifTrue:
			[cogMethod cmType = CMMethod
			 	ifTrue: [cogMethod cmUsageCount <= freeableUsage]
				ifFalse: [cogMethod cmType ~= CMFree
						  and: [cogMethod cmUsageCount = 0]]]
		ifFalse:
			[cogMethod cmType ~= CMFree
			  and: [cogMethod cmUsageCount <= freeableUsage]]! !

!CogMethodZone methodsFor: 'printing' stamp: 'eem 9/5/2019 10:43'!
summarizeZone
	<doNotGenerate>
	| pointer cogMethod ny nf nn |
	<var: #cogMethod type: #'CogMethod *'>
	self printCogMethodsSummarizing: true.
	self printOpenPICListSummarizing: true.
	ny := nf := nn := 0.
	pointer := youngReferrers.
	[pointer < limitAddress] whileTrue:
		[cogMethod := coInterpreter cCoerceSimple: (objectMemory longAt: pointer) to: #'CogMethod *'.
		 cogMethod cmRefersToYoung ifTrue:
			[ny := ny + 1].
		 cogMethod cmType = CMFree ifTrue:
			[nf := nf + 1].
		 (cogMethod cmRefersToYoung or: [cogMethod cmType = CMFree]) ifFalse:
			[nn := nn + 1].
		 pointer := pointer + objectMemory wordSize].
	coInterpreter print: 'num free '; printNum: nf; print: ' num young '; printNum: ny; print: ' num neither '; printNum: nn; cr! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 1/18/2018 15:04'!
voidOpenPICList
	openPICList := nil! !

!CogMethodZone methodsFor: 'compaction' stamp: 'eem 1/26/2020 09:52'!
voidUnpairedMethodList
	NewspeakVM ifTrue: [unpairedMethodList := nil]! !

!CogMethodZone methodsFor: 'jit - api' stamp: 'eem (const to msg rewrite) 11/6/2014 11:51'!
voidYoungReferrersPostTenureAll
	<var: #cogMethod type: #'CogMethod *'>
	| pointer cogMethod |
	<var: #pointer type: #usqInt>
	<var: #cogMethod type: #'CogMethod *'>
	self assert: youngReferrers <= limitAddress.
	pointer := youngReferrers.
	[pointer < limitAddress] whileTrue:
		[cogMethod := coInterpreter cCoerceSimple: (objectMemory longAt: pointer) to: #'CogMethod *'.
		 cogMethod cmType ~= CMFree ifTrue:
			[cogMethod cmRefersToYoung: false].
		 pointer := pointer + objectMemory wordSize].
	youngReferrers := limitAddress! !

!CogMethodZone methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:43'!
whereIsMaybeCodeThing: anOop
	<export: true> "useful for VM debugging; use export: so it will be accessible on win32"
	<api>
	<returnTypeC: #'char *'>
	(self oop: anOop isGreaterThanOrEqualTo: cogit cogCodeBase andLessThan: limitAddress) ifTrue:
		[(self oop: anOop isLessThan: cogit minCogMethodAddress) ifTrue:
			[^' is in generated runtime'].
		 (self oop: anOop isLessThan: mzFreeStart) ifTrue:
			[^' is in generated methods'].
		 (self oop: anOop isLessThan: youngReferrers) ifTrue:
			[^' is in code zone'].
		 ^' is in young referrers'].
	^nil! !

!CogMethodZone methodsFor: 'young referers' stamp: 'eem 6/26/2009 12:47'!
youngReferrers
	<cmacro: '() youngReferrers'>
	^youngReferrers! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 3/10/2017 17:16'!
zoneAlignment
	^8! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 01/10/2016 10:01'!
zoneEnd
	^limitAddress! !

!CogMethodZone methodsFor: 'simulation only' stamp: 'eem 11/22/2013 11:57'!
zoneEnd: zoneEnd
	<doNotGenerate> 
	limitAddress := zoneEnd! !

!CogMethodZone methodsFor: 'accessing' stamp: 'eem 01/10/2016 10:01'!
zoneFree
	^mzFreeStart! !

CogMethodZone subclass: #SistaMethodZone
	instanceVariableNames: 'allocationThreshold thresholdRatio'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!SistaMethodZone class methodsFor: 'translation' stamp: 'eem 3/16/2017 15:56'!
declareCVarsIn: aCCodeGenerator
	aCCodeGenerator
		declareVar: 'allocationThreshold' type: #usqInt;
		var: 'thresholdRatio' declareC: (BytesPerWord = 8 ifTrue: ['double thresholdRatio = 0.5'] ifFalse: ['float thresholdRatio = 0.5f'])! !

!SistaMethodZone methodsFor: 'allocating' stamp: 'eem 3/16/2017 16:11'!
allocationLimit
	<inline: true>
	^super allocationLimit min: allocationThreshold! !

!SistaMethodZone methodsFor: 'initialization' stamp: 'eem 3/16/2017 16:36'!
computeAllocationThreshold
	<inline: true>
	allocationThreshold := ((limitAddress - baseAddress * thresholdRatio) asUnsignedInteger + (self zoneAlignment - 1) truncateTo: self zoneAlignment) + baseAddress! !

!SistaMethodZone methodsFor: 'allocating' stamp: 'eem 3/17/2017 15:22'!
effectiveLimit
	<inline: true>
	^super effectiveLimit min: allocationThreshold! !

!SistaMethodZone methodsFor: 'accessing' stamp: 'eem 3/16/2017 16:03'!
getCogCodeZoneThreshold
	<api>
	<returnTypeC: #double>
	^thresholdRatio! !

!SistaMethodZone methodsFor: 'initialization' stamp: 'eem 3/17/2017 15:10'!
initialize
	super initialize.
	thresholdRatio := 0.5! !

!SistaMethodZone methodsFor: 'initialization' stamp: 'eem 3/16/2017 16:06'!
manageFrom: theStartAddress to: theLimitAddress
	super manageFrom: theStartAddress to: theLimitAddress.
	self computeAllocationThreshold! !

!SistaMethodZone methodsFor: 'accessing' stamp: 'eem 3/16/2017 16:09'!
setCogCodeZoneThreshold: ratio
	<api>
	<var: #ratio type: #double>
	(ratio >= 0.1 and: [ratio <= 1.0]) ifFalse:
		[^PrimErrBadArgument].
	thresholdRatio := ratio.
	self computeAllocationThreshold.
	^0! !

CogClass subclass: #CogObjectRepresentation
	instanceVariableNames: 'cogit methodZone objectMemory coInterpreter ceStoreCheckTrampoline ceByteSizeOfTrampoline ceInstantiateClassTrampoline ceInstantiateClassIndexableSizeTrampoline ceFloatObjectOfTrampoline ceFloatValueOfTrampoline ceSigned64BitIntegerTrampoline ceSigned64BitValueOfTrampoline cePositive64BitIntegerTrampoline cePositive64BitValueOfTrampoline'
	classVariableNames: ''
	poolDictionaries: 'CogAbstractRegisters CogCompilationConstants CogMethodConstants CogRTLOpcodes VMBasicConstants VMObjectIndices VMSqueakClassIndices VMStackFrameOffsets'
	category: 'VMMaker-JIT'!

!CogObjectRepresentation commentStamp: '<historical>' prior: 0!
I am an abstract superclass for object representations whose job it is to generate abstract instructions for accessing objects.  It is hoped that this level of indirection between the Cogit code generator and object access makes it easier to adapt the code generator to different garbage collectors, object representations and languages.!

!CogObjectRepresentation class methodsFor: 'instance creation' stamp: 'eem 11/25/2013 13:39'!
forCogit: aCogit methodZone: methodZone
	^self new setCogit: aCogit methodZone: methodZone! !

!CogObjectRepresentation class methodsFor: 'translation' stamp: 'eem 2/28/2009 07:30'!
implicitReturnTypeFor: aSelector
	"Answer the return type for methods that don't have an explicit return."
	^#void! !

!CogObjectRepresentation class methodsFor: 'initialization' stamp: 'eem 1/1/1970 12:43'!
initializeMiscConstants
	"Override to avoid inheriting (and hence repeating) VMClass class>>initializeMiscConstants.
	 Subclasses that have misc constants to initialize will override further."! !

!CogObjectRepresentation class methodsFor: 'trampoline support' stamp: 'rsf 8/9/2016 21:27'!
numTrampolines
	^1 "ceStoreCheckTrampoline" + (LowcodeVM ifTrue: [ 9 ] ifFalse: [ 0 ])! !

!CogObjectRepresentation class methodsFor: 'translation' stamp: 'eem 2/11/2009 12:47'!
typeForSelf
	^#implicit! !

!CogObjectRepresentation class methodsFor: 'accessing' stamp: 'eem 12/23/2015 05:35'!
wordSize
	^4! !

!CogObjectRepresentation methodsFor: 'garbage collection' stamp: 'eem 11/18/2013 10:58'!
allYoungObjectsAgeInFullGC
	^self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'sista support' stamp: 'eem 7/5/2014 17:55'!
allocateCounters: nCounters
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'testing' stamp: 'eem 8/3/2021 18:53'!
canPinObjects
	"Answer if the memory manager supports pinned objects."
	<inline: #always>
	^false! !

!CogObjectRepresentation methodsFor: 'garbage collection' stamp: 'eem 5/9/2014 19:35'!
checkValidObjectReference: anOop
	^(objectMemory isImmediate: anOop) not
	   and: [(objectMemory heapMapAtWord: (self pointerForOop: anOop)) ~= 0]! !

!CogObjectRepresentation methodsFor: 'in-line cacheing' stamp: 'eem 3/14/2017 12:05'!
classForInlineCacheTag: classIndex
	^self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'testing' stamp: 'eem 8/3/2021 18:53'!
createsArraysInline
	"Answer if the object representation allocates arrays inline.  By
	 default answer false. Better code can be generated when creating
	 arrays inline if values are /not/ flushed to the stack."
	<inline: #always>
	^false! !

!CogObjectRepresentation methodsFor: 'testing' stamp: 'eem 8/3/2021 18:53'!
createsClosuresInline
	"Answer if the object representation allocates closures inline.  By
	 default answer false. Better code can be generated when creating
	 closures inline if copied values are /not/ flushed to the stack."
	<inline: #always>
	^false! !

!CogObjectRepresentation methodsFor: 'compilation' stamp: 'eem 6/2/2014 19:48'!
ensureNoForwardedLiteralsIn: aMethodObj
	"Ensure there are no forwarded literals in the argument.
	 By default this is a noop.  Subclasses redefine as necessary."
	<inline: true>! !

!CogObjectRepresentation methodsFor: 'sista support' stamp: 'eem 7/5/2014 17:58'!
freeCounters: theCounters
	<var: #theCounters type: #usqInt>
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 12/6/2015 20:38'!
genClearAndSetSmallIntegerTagsIn: scratchReg
	"Set the SmallInteger tag bits when the tag bits may be filled with garbage."
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 5/29/2015 19:53'!
genCmpClassFloatCompactIndexR: reg
	^cogit CmpCq: ClassFloatCompactIndex R: reg! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 5/29/2015 19:52'!
genCmpClassMethodContextCompactIndexR: reg
	^cogit CmpCq: ClassMethodContextCompactIndex R: reg! !

!CogObjectRepresentation methodsFor: 'sista support' stamp: 'cb 11/23/2017 14:53'!
genConvertSmallFloatToSmallFloatHashAsIntegerInReg: reg scratch: scratch
	^EncounteredUnknownBytecode! !

!CogObjectRepresentation methodsFor: 'bytecode generator support' stamp: 'eem 5/31/2014 23:12'!
genCreateClosureAt: bcpc numArgs: numArgs numCopied: numCopied contextNumArgs: ctxtNumArgs large: isLargeCtxt inBlock: isInBlock
	"Create a closure with the given startpc, numArgs and numCopied
	 within a context with ctxtNumArgs, large if isLargeCtxt that is in a
	 block if isInBlock.  If numCopied > 0 pop those values off the stack."
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'bytecode generator support' stamp: 'cb 4/8/2016 15:40'!
genCreateFullClosure: compiledBlock numArgs: numArgs numCopied: numCopied ignoreContext: ignoreContext contextNumArgs: contextNumArgs large: contextIsLarge inBlock: contextIsBlock
	"Create a full closure"
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 13:42'!
genDoubleArithmetic: arithmeticOperator preOpCheck: preOpCheckOrNil
	<option: #DPFPReg0>
	<var: #preOpCheckOrNil declareC: 'AbstractInstruction *(*preOpCheckOrNil)(int rcvrReg, int argReg)'>
	| jumpFailClass jumpFailAlloc jumpFailCheck jumpImmediate jumpNonInt doOp |
	<var: #jumpFailClass type: #'AbstractInstruction *'>
	<var: #jumpFailAlloc type: #'AbstractInstruction *'>
	<var: #jumpImmediate type: #'AbstractInstruction *'>
	<var: #jumpNonInt type: #'AbstractInstruction *'>
	<var: #jumpFailCheck type: #'AbstractInstruction *'>
	<var: #doOp type: #'AbstractInstruction *'>
	cogit processorHasDoublePrecisionFloatingPointSupport ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	self genGetDoubleValueOf: ReceiverResultReg into: DPFPReg0.
	cogit MoveR: Arg0Reg R: ClassReg.
	jumpImmediate := self genJumpImmediate: Arg0Reg.
	self genGetCompactClassIndexNonImmOf: Arg0Reg into: SendNumArgsReg.
	self genCmpClassFloatCompactIndexR: SendNumArgsReg.
	jumpFailClass := cogit JumpNonZero: 0.
	self genGetDoubleValueOf: Arg0Reg into: DPFPReg1.
	doOp := cogit Label.
	preOpCheckOrNil ifNotNil:
		[jumpFailCheck := cogit perform: preOpCheckOrNil with: DPFPReg0 with: DPFPReg1].
	cogit gen: arithmeticOperator operand: DPFPReg1 operand: DPFPReg0.
	jumpFailAlloc := self
						genAllocFloatValue: DPFPReg0
						into: SendNumArgsReg
						scratchReg: ClassReg
						scratchReg: TempReg.
	cogit MoveR: SendNumArgsReg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpImmediate jmpTarget: cogit Label.
	self maybeGenConvertIfSmallFloatIn: Arg0Reg scratchReg: TempReg into: DPFPReg1 andJumpTo: doOp.
	self smallIntegerIsOnlyImmediateType ifFalse:
		[jumpNonInt := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg].
	self genConvertSmallIntegerToIntegerInReg: ClassReg.
	cogit ConvertR: ClassReg Rd: DPFPReg1.
	cogit Jump: doOp.
	jumpFailAlloc jmpTarget:  (jumpFailClass jmpTarget: cogit Label).
	self smallIntegerIsOnlyImmediateType ifFalse:
		[jumpNonInt jmpTarget: jumpFailClass getJmpTarget].
	preOpCheckOrNil ifNotNil:
		[jumpFailCheck jmpTarget: jumpFailClass getJmpTarget].
	^0! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'rsf 1/4/2017 20:27'!
genDoubleComparison: jumpOpcodeGenerator invert: invertComparison
	<option: #DPFPReg0>
	<var: #jumpOpcodeGenerator declareC: 'AbstractInstruction * NoDbgRegParms (*jumpOpcodeGenerator)(void *)'>
	| jumpFail jumpImmediate jumpNonInt jumpCond compare |
	<var: #jumpImmediate type: #'AbstractInstruction *'>
	<var: #jumpNonInt type: #'AbstractInstruction *'>
	<var: #jumpCond type: #'AbstractInstruction *'>
	<var: #compare type: #'AbstractInstruction *'>
	<var: #jumpFail type: #'AbstractInstruction *'>
	cogit processorHasDoublePrecisionFloatingPointSupport ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	self genGetDoubleValueOf: ReceiverResultReg into: DPFPReg0.
	jumpImmediate := self genJumpImmediate: Arg0Reg.
	self genGetCompactClassIndexNonImmOf: Arg0Reg into: SendNumArgsReg.
	self genCmpClassFloatCompactIndexR: SendNumArgsReg.
	jumpFail := cogit JumpNonZero: 0.
	self genGetDoubleValueOf: Arg0Reg into: DPFPReg1.
	invertComparison "May need to invert for NaNs"
		ifTrue: [compare := cogit CmpRd: DPFPReg0 Rd: DPFPReg1]
		ifFalse: [compare := cogit CmpRd: DPFPReg1 Rd: DPFPReg0].
	jumpCond := cogit perform: jumpOpcodeGenerator with: 0. "FP jumps are a little weird"
	cogit genMoveFalseR: ReceiverResultReg.
	cogit genPrimReturn.
	jumpCond jmpTarget: (cogit genMoveTrueR: ReceiverResultReg).
	cogit genPrimReturn.
	jumpImmediate jmpTarget: cogit Label.
	self maybeGenConvertIfSmallFloatIn: Arg0Reg scratchReg: TempReg into: DPFPReg1 andJumpTo: compare.
	self smallIntegerIsOnlyImmediateType ifFalse:
		[jumpNonInt := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg].
	self genConvertSmallIntegerToIntegerInReg: Arg0Reg.
	cogit ConvertR: Arg0Reg Rd: DPFPReg1.
	cogit Jump: compare.
	jumpFail jmpTarget: cogit Label.
	self smallIntegerIsOnlyImmediateType ifFalse:
		[jumpNonInt jmpTarget: jumpFail getJmpTarget].
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 5/14/2015 18:45'!
genEnsureObjInRegNotForwarded: reg scratchReg: scratch
	"Make sure that the object in reg is not forwarded.  By default there is
	 nothing to do.  Subclasses for memory managers that forward will override."
	<inline: true>
	^0! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 5/14/2015 18:45'!
genEnsureOopInRegNotForwarded: reg scratchReg: scratch
	"Make sure that the oop in reg is not forwarded.  By default there is
	 nothing to do.  Subclasses for memory managers that forward will override."
	<inline: true>
	^0! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 5/14/2015 18:46'!
genEnsureOopInRegNotForwarded: reg scratchReg: scratch ifForwarder: fwdJumpTarget ifNotForwarder: nonFwdJumpTargetOrZero
	"Make sure that the oop in reg is not forwarded.  By default there is
	 nothing to do.  Subclasses for memory managers that forward will override."
	<inline: true>
	^0! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 5/14/2015 18:43'!
genEnsureOopInRegNotForwarded: reg scratchReg: scratch jumpBackTo: instruction
	"Make sure that the oop in reg is not forwarded.  By default there is
	 nothing to do.  Subclasses for memory managers that forward will override."
	<var: #instruction type: #'AbstractInstruction *'>
	<inline: true>
	^0! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 6/3/2015 16:17'!
genEnsureOopInRegNotForwarded: reg scratchReg: scratch updatingMw: offset r: baseReg
	"Make sure that the oop in reg is not forwarded, updating the field at offset from baseReg.
	 By default there is nothing to do.  Subclasses for memory managers that forward will override."
	<inline: true>
	^0! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 6/4/2015 11:30'!
genEnsureOopInRegNotForwarded: reg scratchReg: scratch updatingSlot: index in: objReg
	"Make sure that the oop in reg is not forwarded, updating the slot in objReg with the value."
	<inline: true>
	self flag: 'not safe unless there''s also a store check!!!!'.
	self assert: false.
	^self genEnsureOopInRegNotForwarded: reg
		scratchReg: scratch
		updatingMw: index * objectMemory wordSize + objectMemory baseHeaderSize
		r: objReg! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 13:54'!
genFastPrimFail
	<doNotGenerate>
	^cogit genFastPrimFail! !

!CogObjectRepresentation methodsFor: 'bytecode generator support' stamp: 'eem 5/31/2014 21:46'!
genGetActiveContextNumArgs: numArgs large: isLargeContext inBlock: isInBlock
	"Get the active context into ReceiverResultReg, creating it if necessary."
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 5/30/2015 09:38'!
genGetFormatOf: srcReg into: destReg
	"Get the format field of the object in srcReg into destReg.
	 srcReg may equal destReg."
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 7/7/2014 11:53'!
genGetInlineCacheClassTagFrom: sourceReg into: destReg forEntry: forEntry
	"Extract the inline cache tag for the object in sourceReg into destReg. The inline cache tag
	 for a given object is the value loaded in inline caches to distinguish objects of different
	 classes. If forEntry is true answer the entry label at which control is to enter (cmEntryOffset).
	 If forEntry is false, control enters at the start."
	<returnTypeC: #'AbstractInstruction *'>
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 1/23/2016 11:44'!
genGetMethodHeaderOf: methodReg into: headerReg scratch: scratchReg
	"Get the method header (first word) of a CompiledMethod into headerReg.
	 Deal with the method possibly being cogged."
	| jumpNotCogged |
	<var: #jumpNotCogged type: #'AbstractInstruction *'>
	cogit MoveMw: objectMemory baseHeaderSize r: methodReg R: headerReg.
	jumpNotCogged := self genJumpSmallInteger: headerReg scratchReg: scratchReg.
	cogit MoveMw: (cogit offset: CogMethod of: #methodHeader) r: headerReg R: headerReg.
	jumpNotCogged jmpTarget: cogit Label.
	^0! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 8/26/2014 12:10'!
genGetNumSlotsOf: srcReg into: destReg
	"Get the size in word-sized slots of the object in srcReg into destReg.
	 srcReg may equal destReg."
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 5/30/2015 11:16'!
genJumpImmediate: aRegister
	"Generate a compare and branch to test if aRegister contains an immediate.
	 Answer the jump."
	<returnTypeC: #'AbstractInstruction *'>
	self subclassResponsibility
	! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 12/5/2015 11:04'!
genJumpIsSmallIntegerValue: aRegister scratch: scratchReg
	"Generate a test for aRegister containing an integer value in the SmallInteger range, and a jump if so, answering the jump."
	<returnTypeC: #'AbstractInstruction *'>
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 5/30/2015 11:21'!
genJumpNotImmediate: aRegister
	"Generate a compare and branch to test if aRegister contains a non-immediate.
	 Answer the jump."
	<returnTypeC: #'AbstractInstruction *'>
	self subclassResponsibility
	! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 5/30/2015 11:22'!
genJumpNotSmallInteger: aRegister
	"Generate a compare and branch to test if aRegister contains other than a SmallInteger.
	 Answer the jump, or UnimplementedOperation if this cannot be done with a single register."
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	^cogit cCoerceSimple: UnimplementedOperation to: #'AbstractInstruction *'! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 12/13/2015 16:53'!
genJumpNotSmallInteger: aRegister scratchReg: scratch
	"Generate a compare and branch to test if aRegister contains other than a SmallInteger.
	 Answer the jump.  Use scratch if required.  Subclasses will override if scratch is needed."
	<inline: true>
	^self genJumpNotSmallInteger: aRegister! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 12/13/2015 13:52'!
genJumpNotSmallIntegerInScratchReg: aRegister
	"Generate a test for aRegister containing an integer value outside the SmallInteger range, and a jump if so, answering the jump."
	<returnTypeC: #'AbstractInstruction *'>
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 12/9/2015 11:10'!
genJumpNotSmallIntegerValue: aRegister scratch: scratchReg
	"Generate a test for aRegister containing an integer value outside the SmallInteger range, and a jump if so, answering the jump."
	<returnTypeC: #'AbstractInstruction *'>
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 2/24/2017 16:58'!
genJumpNotSmallIntegersIn: aRegister and: bRegister scratch: scratchRegister
	"Generate a compare and branch to test if aRegister and bRegister contains other than SmallIntegers,
	 i.e. don't branch if both aRegister and bRegister contain SmallIntegers.
	 Answer the jump.  Destroy scratchRegister if required."
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	cogit
		MoveR: aRegister R: scratchRegister;
		AndR: bRegister R: scratchRegister.
	^self genJumpNotSmallIntegerInScratchReg: scratchRegister! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 5/30/2015 11:22'!
genJumpSmallInteger: aRegister
	"Generate a compare and branch to test if aRegister contains a SmallInteger.
	 Answer the jump, or UnimplementedOperation if this cannot be done with
	 a single register."
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	^cogit cCoerceSimple: UnimplementedOperation to: #'AbstractInstruction *'! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 3/10/2016 10:39'!
genJumpSmallInteger: aRegister scratchReg: scratch
	"Generate a compare and branch to test if aRegister contains a SmallInteger.
	 Answer the jump.  Use scratch if required.  Subclasses will override if scratch is needed."
	<inline: true>
	^self genJumpSmallInteger: aRegister! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 1/4/2017 18:40'!
genLcByteSizeOf: oop to: resultRegister
	<option: #LowcodeVM>
	"TODO: Optimize this one avoiding the trampoline"
	oop ~= Arg0Reg ifTrue: [cogit MoveR: oop R: Arg0Reg ].
	cogit CallRT: ceByteSizeOfTrampoline.

	cogit MoveR: TempReg R: resultRegister.
	cogit ssPushNativeRegister: resultRegister.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 3/26/2017 17:14'!
genLcFloat32: value toOop: object
	<option: #LowcodeVM>
	cogit ConvertRs: value Rd: DPFPReg0.
	cogit CallRT: ceFloatObjectOfTrampoline.

	cogit MoveR: TempReg R: object.
	cogit ssPushRegister: object.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 3/26/2017 17:14'!
genLcFloat64: value toOop: object
	<option: #LowcodeVM>
	value ~= DPFPReg0 ifTrue: [ cogit MoveRd: value Rd: DPFPReg0].
	cogit CallRT: ceFloatObjectOfTrampoline.

	cogit MoveR: TempReg R: object.
	cogit ssPushRegister: object.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:30'!
genLcInstantiateOop: classOop
	<option: #LowcodeVM>
	classOop ~= ReceiverResultReg ifTrue: [cogit MoveR: classOop R: ReceiverResultReg].
	cogit CallRT: ceInstantiateClassTrampoline.

	cogit MoveR: TempReg R: classOop.
	cogit ssPushRegister: classOop.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 1/4/2017 18:37'!
genLcInstantiateOop: classOop constantIndexableSize: indexableSize
	<option: #LowcodeVM>
	classOop ~= Arg0Reg ifTrue: [cogit MoveR: classOop R: Arg0Reg].
	cogit MoveCq: indexableSize R: Arg1Reg.
	cogit CallRT: ceInstantiateClassIndexableSizeTrampoline.

	cogit MoveR: TempReg R: classOop.
	cogit ssPushRegister: classOop.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 1/4/2017 18:38'!
genLcInstantiateOop: classOop indexableSize: indexableSize
	<option: #LowcodeVM>
	classOop ~= Arg0Reg ifTrue: [
		indexableSize = Arg0Reg ifTrue: [
			cogit MoveR: indexableSize R: TempReg.
		].
		cogit MoveR: classOop R: Arg0Reg
	].

	indexableSize ~= Arg1Reg ifTrue: [
		indexableSize = Arg0Reg ifTrue: [
			cogit MoveR: TempReg R: Arg1Reg.
		] ifFalse: [
			cogit MoveR: indexableSize R: Arg1Reg.
		]
	].

	cogit CallRT: ceInstantiateClassIndexableSizeTrampoline.

	cogit MoveR: TempReg R: classOop.
	cogit ssPushRegister: classOop.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:30'!
genLcInt64ToOop: value
	<option: #LowcodeVM>
	value ~= ReceiverResultReg ifTrue: [ cogit MoveR: value R: ReceiverResultReg.].
	cogit CallRT: ceSigned64BitIntegerTrampoline.

	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit ssPushRegister: ReceiverResultReg.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:30'!
genLcInt64ToOop: valueLow highPart: valueHigh
	<option: #LowcodeVM>
	"Put the arguments in the correct registers"
	valueLow ~= ReceiverResultReg ifTrue: [
		valueHigh = ReceiverResultReg ifTrue: [
			cogit MoveR: valueHigh R: TempReg.
		].
		cogit MoveR: valueLow R: ReceiverResultReg.
	].

	valueHigh ~= Arg0Reg ifTrue: [
		valueHigh = ReceiverResultReg ifTrue: [
			cogit MoveR: TempReg R: Arg0Reg.
		] ifFalse: [
			cogit MoveR: valueHigh R: Arg0Reg.
		]
	].

	"Call the trampoline"
	cogit CallRT: ceSigned64BitIntegerTrampoline.

	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit ssPushRegister: ReceiverResultReg.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:30'!
genLcOop: object toFloat32: value
	<option: #LowcodeVM>
	object ~= ReceiverResultReg ifTrue: [ cogit MoveR: object R: ReceiverResultReg].
	cogit CallRT: ceFloatValueOfTrampoline.

	cogit ConvertRd: DPFPReg0 Rs: value.
	cogit ssPushNativeRegisterSingleFloat: value.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:30'!
genLcOop: object toFloat64: value
	<option: #LowcodeVM>
	object ~= ReceiverResultReg ifTrue: [ cogit MoveR: object R: ReceiverResultReg].
	cogit CallRT: ceFloatValueOfTrampoline.

	DPFPReg0 ~= value ifTrue: [cogit MoveRd: DPFPReg0 Rd: value.].
	cogit ssPushNativeRegisterDoubleFloat: value.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:30'!
genLcOop: object toInt64: valueLow highPart: valueHigh
	<option: #LowcodeVM>
	object ~= ReceiverResultReg ifTrue: [ cogit MoveR: object R: ReceiverResultReg.].
	cogit CallRT: ceSigned64BitValueOfTrampoline.

	Arg0Reg ~= valueHigh ifTrue: [cogit MoveR: Arg0Reg R: valueHigh].
	cogit MoveR: TempReg R: valueLow.

	cogit ssPushNativeRegister: valueLow secondRegister: valueHigh.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:30'!
genLcOop: object toUInt64: valueLow highPart: valueHigh
	<option: #LowcodeVM>
	object ~= ReceiverResultReg ifTrue: [ cogit MoveR: object R: ReceiverResultReg.].
	cogit CallRT: cePositive64BitValueOfTrampoline.

	Arg0Reg ~= valueHigh ifTrue: [cogit MoveR: Arg0Reg R: valueHigh].
	cogit MoveR: TempReg R: valueLow.

	cogit ssPushNativeRegister: valueLow secondRegister: valueHigh.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:30'!
genLcOopToInt64: value
	<option: #LowcodeVM>
	value ~= ReceiverResultReg ifTrue: [ cogit MoveR: value R: ReceiverResultReg.].
	cogit CallRT: ceSigned64BitValueOfTrampoline.

	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit ssPushNativeRegister: ReceiverResultReg.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:30'!
genLcOopToPointer: object
	<option: #LowcodeVM>
	"Assume this is always correct"
	cogit MoveMw: BaseHeaderSize r: object R: object.
	cogit ssPushNativeRegister: object.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:30'!
genLcOopToUInt64: value
	<option: #LowcodeVM>
	value ~= ReceiverResultReg ifTrue: [ cogit MoveR: value R: ReceiverResultReg.].
	cogit CallRT: cePositive64BitValueOfTrampoline.

	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit ssPushNativeRegister: ReceiverResultReg.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'eem 2/11/2020 10:17'!
genLcPointerToOop: pointer class: pointerClass
	<option: #LowcodeVM>
	cogit PushR: pointer.
	cogit annotate: (cogit MoveCw: pointerClass R: Arg0Reg) objRef: pointerClass.
	cogit MoveCq: objectMemory bytesPerOop R: Arg1Reg.
	cogit CallRT: ceInstantiateClassIndexableSizeTrampoline.

	cogit PopR: pointer.
	cogit MoveR: pointer Mw: BaseHeaderSize r: TempReg.

	cogit MoveR: TempReg R: pointer.
	cogit ssPushRegister: pointer.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:30'!
genLcUInt64ToOop: value
	<option: #LowcodeVM>
	value ~= ReceiverResultReg ifTrue: [ cogit MoveR: value R: ReceiverResultReg.].
	cogit CallRT: cePositive64BitIntegerTrampoline.

	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit ssPushRegister: ReceiverResultReg.! !

!CogObjectRepresentation methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:30'!
genLcUInt64ToOop: valueLow highPart: valueHigh
	<option: #LowcodeVM>
	"Put the arguments in the correct registers"
	valueLow ~= ReceiverResultReg ifTrue: [
		valueHigh = ReceiverResultReg ifTrue: [
			cogit MoveR: valueHigh R: TempReg.
		].
		cogit MoveR: valueLow R: ReceiverResultReg.
	].

	valueHigh ~= Arg0Reg ifTrue: [
		valueHigh = ReceiverResultReg ifTrue: [
			cogit MoveR: TempReg R: Arg0Reg.
		] ifFalse: [
			cogit MoveR: valueHigh R: Arg0Reg.
		]
	].

	"Call the trampoline"
	cogit CallRT: cePositive64BitIntegerTrampoline.

	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit ssPushRegister: ReceiverResultReg.! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 9/30/2013 22:32'!
genLoadSlot: index sourceReg: sourceReg destReg: destReg
	cogit
		MoveMw: index * objectMemory wordSize + objectMemory baseHeaderSize
		r: sourceReg
		R: destReg.
	^0! !

!CogObjectRepresentation methodsFor: 'bytecode generator support' stamp: 'eem 5/31/2014 08:52'!
genNewArrayOfSize: size initialized: initialized
	"Generate a call to code that allocates a new Array of size.
	 The Array should be initialized with nils iff initialized is true.
	 The size arg is passed in SendNumArgsReg, the result
	 must come back in ReceiverResultReg."
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 13:37'!
genPrimitiveAdd
	| jumpNotSI jumpOvfl |
	<var: #jumpNotSI type: #'AbstractInstruction *'>
	<var: #jumpOvfl type: #'AbstractInstruction *'>
	cogit mclassIsSmallInteger ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	cogit MoveR: Arg0Reg R: ClassReg.
	jumpNotSI := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	self genRemoveSmallIntegerTagsInScratchReg: ClassReg.
	cogit AddR: ReceiverResultReg R: ClassReg.
	jumpOvfl := cogit JumpOverflow: 0.
	cogit MoveR: ClassReg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpOvfl jmpTarget: (jumpNotSI jmpTarget: cogit Label).
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:23'!
genPrimitiveAsCharacter
	"subclasses override if they can"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 13:42'!
genPrimitiveAsFloat
	<option: #DPFPReg0>
	| jumpFailAlloc |
	<var: #jumpFailAlloc type: #'AbstractInstruction *'>
	cogit processorHasDoublePrecisionFloatingPointSupport ifFalse:
		[^UnimplementedPrimitive].
	cogit MoveR: ReceiverResultReg R: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	cogit ConvertR: TempReg Rd: DPFPReg0.
	jumpFailAlloc := self
						genAllocFloatValue: DPFPReg0
						into: SendNumArgsReg
						scratchReg: ClassReg
						scratchReg: TempReg.
	cogit MoveR: SendNumArgsReg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpFailAlloc jmpTarget: cogit Label.
	^0! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 09:54'!
genPrimitiveAt
	"subclasses override if they can"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 09:58'!
genPrimitiveAtPut
	"subclasses override if they can"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 13:52'!
genPrimitiveBitAnd
	| jumpNotSI |
	<var: #jumpNotSI type: #'AbstractInstruction *'>
	cogit mclassIsSmallInteger ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	jumpNotSI := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	"Whether the SmallInteger tags are zero or non-zero, anding them together will preserve them."
	cogit AndR: Arg0Reg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpNotSI jmpTarget: cogit Label.
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 13:53'!
genPrimitiveBitOr
	| jumpNotSI |
	<var: #jumpNotSI type: #'AbstractInstruction *'>
	cogit mclassIsSmallInteger ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	jumpNotSI := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	"Whether the SmallInteger tags are zero or non-zero, oring them together will preserve them."
	cogit OrR: Arg0Reg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpNotSI jmpTarget: cogit Label.
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 14:26'!
genPrimitiveBitShift
	"rTemp := rArg0
	rClass := tTemp
	rTemp := rTemp & 1
	jz nonInt
	rClass >>= 1
	cmp 0,rClass
	jge neg
	cmp 31,rClass // numSmallIntegerBits, jge for sign
	jge tooBig
	rTemp := rReceiver
	rTemp <<= rClass
	rTemp >>= rClass (arithmetic)
	cmp rTemp,rReceiver
	jnz ovfl
	rReceiver := rReceiver - 1
	rReceiver := rReceiver <<= rClass
	rReceiver := rReceiver + 1
	ret
neg:
	rClass := 0 - rClass
	cmp 31,rClass // numSmallIntegerBits
	jge inRange
	rClass := 31
inRange
	rReceiver := rReceiver >>= rClass.
	rReceiver := rReceiver | smallIntegerTags.
	ret
ovfl
tooBig
nonInt:
	fail"
	| jumpNotSI jumpOvfl jumpNegative jumpTooBig jumpInRange |
	<var: #jumpNotSI type: #'AbstractInstruction *'>
	<var: #jumpOvfl type: #'AbstractInstruction *'>
	<var: #jumpNegative type: #'AbstractInstruction *'>
	<var: #jumpTooBig type: #'AbstractInstruction *'>
	<var: #jumpInRange type: #'AbstractInstruction *'>
	cogit mclassIsSmallInteger ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	cogit MoveR: Arg0Reg R: ClassReg.
	jumpNotSI := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	self genConvertSmallIntegerToIntegerInReg: ClassReg.
	(cogit lastOpcode setsConditionCodesFor: JumpNegative) ifFalse:
		[cogit CmpCq: 0 R: ClassReg]. "N.B. FLAGS := ClassReg - 0"
	jumpNegative := cogit JumpNegative: 0.
	cogit CmpCq: self numSmallIntegerBits R: ClassReg. "N.B. FLAGS := ClassReg - numSmallIntegerBits"
	jumpTooBig := cogit JumpGreaterOrEqual: 0.
	cogit MoveR: ReceiverResultReg R: TempReg.
	cogit LogicalShiftLeftR: ClassReg R: TempReg.
	cogit ArithmeticShiftRightR: ClassReg R: TempReg.
	cogit CmpR: TempReg R: ReceiverResultReg. "N.B. FLAGS := RRReg - TempReg"
	jumpOvfl := cogit JumpNonZero: 0.
	self genRemoveSmallIntegerTagsInScratchReg: ReceiverResultReg.
	cogit LogicalShiftLeftR: ClassReg R: ReceiverResultReg.
	self genAddSmallIntegerTagsTo: ReceiverResultReg.
	cogit genPrimReturn.
	jumpNegative jmpTarget: (cogit NegateR: ClassReg).
	cogit CmpCq: self numSmallIntegerBits R: ClassReg. "N.B. FLAGS := ClassReg - numSmallIntegerBits"
	jumpInRange := cogit JumpLessOrEqual: 0.
	cogit MoveCq: self numSmallIntegerBits R: ClassReg.
	jumpInRange jmpTarget: (cogit ArithmeticShiftRightR: ClassReg R: ReceiverResultReg).
	self genClearAndSetSmallIntegerTagsIn: ReceiverResultReg.
	cogit genPrimReturn.
	jumpNotSI jmpTarget: (jumpTooBig jmpTarget: (jumpOvfl jmpTarget: cogit Label)).
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 13:38'!
genPrimitiveBitXor
	| jumpNotSI |
	<var: #jumpNotSI type: #'AbstractInstruction *'>
	cogit mclassIsSmallInteger ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	jumpNotSI := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	"Clear one or the other tag so that xoring will preserve them."
	self genRemoveSmallIntegerTagsInScratchReg: Arg0Reg.
	cogit XorR: Arg0Reg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpNotSI jmpTarget: cogit Label.
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'cb 9/11/2019 07:49'!
genPrimitiveClass
	| reg |
	reg := ReceiverResultReg.
	cogit methodNumArgs > 0 ifTrue:
		[cogit methodNumArgs > 1 ifTrue:
			[^UnimplementedPrimitive].
		 cogit genLoadArgAtDepth: 0 into: (reg := Arg0Reg)].
	(self
			genGetClassObjectOf: reg
			into: ReceiverResultReg
			scratchReg: TempReg
			mayBeAForwarder: reg ~= ReceiverResultReg) = BadRegisterSet ifTrue:
		[self
			genGetClassObjectOf: reg
			into: ClassReg
			scratchReg: TempReg
			mayBeAForwarder: reg ~= ReceiverResultReg.
		 cogit MoveR: ClassReg R: ReceiverResultReg].
	cogit genPrimReturn.
	^UnfailingPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:44'!
genPrimitiveClosureValue
	"Defer to the cogit for this one.  This is for two reasons:
	 a)	Cogit's version speeds up cull: et al by not calling the interpreter primitive if numArgs is wrong.
		Our convention for compiling the interpreter primitive call doesn't agree with this usage.
	b)	Cogit refuses to compile primitiveClosureValueNoContextSwitch until the blockNoContextSwitchOffset
		is known, and that isn't information the CogObjectRepresentation is privvy to."
	<doNotGenerate>
	^cogit genPrimitiveClosureValue! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 12/26/2019 21:50'!
genPrimitiveDiv
	| jumpNotSI jumpIsSI jumpZero jumpExact jumpSameSign convert |
	<var: #convert type: #'AbstractInstruction *'>
	<var: #jumpIsSI type: #'AbstractInstruction *'>
	<var: #jumpZero type: #'AbstractInstruction *'>
	<var: #jumpNotSI type: #'AbstractInstruction *'>
	<var: #jumpExact type: #'AbstractInstruction *'>
	<var: #jumpSameSign type: #'AbstractInstruction *'>
	cogit processorHasDivQuoRemAndMClassIsSmallInteger ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	cogit MoveR: Arg0Reg R: ClassReg.
	cogit MoveR: Arg0Reg R: Arg1Reg.
	jumpNotSI := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	"We must shift away the tags, not just subtract them, so that the
	 overflow case doesn't actually overflow the machine instruction."
	self genShiftAwaySmallIntegerTagsInScratchReg: ClassReg.
	(cogit lastOpcode setsConditionCodesFor: JumpZero) ifFalse:
		[cogit CmpCq: 0 R: ClassReg].
	jumpZero := cogit JumpZero: 0.
	cogit MoveR: ReceiverResultReg R: TempReg.
	self genShiftAwaySmallIntegerTagsInScratchReg: TempReg.
	cogit DivR: ClassReg R: TempReg Quo: TempReg Rem: ClassReg.
	"If remainder is zero we must check for overflow."
	cogit CmpCq: 0 R: ClassReg.
	jumpExact := cogit JumpZero: 0.
	"If arg and remainder signs are different we must round down."
	cogit XorR: ClassReg R: Arg1Reg.
	(cogit lastOpcode setsConditionCodesFor: JumpZero) ifFalse:
		[cogit CmpCq: 0 R: Arg1Reg].
	jumpSameSign := cogit JumpGreaterOrEqual: 0.
	cogit SubCq: 1 R: TempReg.
	jumpSameSign jmpTarget: (convert := cogit Label).
	self genConvertIntegerInReg: TempReg toSmallIntegerInReg: ReceiverResultReg.
	cogit genPrimReturn.
	"test for overflow; the only case is SmallInteger minVal // -1"
	jumpExact jmpTarget: cogit Label.
	jumpIsSI := self genJumpIsSmallIntegerValue: TempReg scratch: Arg1Reg.
	jumpIsSI jmpTarget: convert.
	jumpZero jmpTarget: (jumpNotSI jmpTarget: cogit Label).
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 12/26/2019 21:50'!
genPrimitiveDivide
	| jumpNotSI jumpZero jumpInexact jumpOverflow |
	<var: #jumpNotSI type: #'AbstractInstruction *'>
	<var: #jumpZero type: #'AbstractInstruction *'>
	<var: #jumpInexact type: #'AbstractInstruction *'>
	<var: #jumpOverflow type: #'AbstractInstruction *'>
	cogit processorHasDivQuoRemAndMClassIsSmallInteger ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	cogit MoveR: Arg0Reg R: ClassReg.
	jumpNotSI := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	"We must shift away the tags, not just subtract them, so that the
	 overflow case doesn't actually overflow the machine instruction."
	self genShiftAwaySmallIntegerTagsInScratchReg: ClassReg.
	jumpZero := cogit JumpZero: 0.
	cogit MoveR: ReceiverResultReg R: TempReg.
	self genShiftAwaySmallIntegerTagsInScratchReg: TempReg.
	cogit DivR: ClassReg R: TempReg Quo: TempReg Rem: ClassReg.
	"If remainder is non-zero fail."
	cogit CmpCq: 0 R: ClassReg.
	jumpInexact := cogit JumpNonZero: 0.
	"test for overflow; the only case is SmallInteger minVal / -1"
	jumpOverflow := self genJumpNotSmallIntegerValue: TempReg scratch: Arg1Reg.
	self genConvertIntegerInReg: TempReg toSmallIntegerInReg: ReceiverResultReg.
	cogit genPrimReturn.
	jumpOverflow jmpTarget: (jumpInexact jmpTarget: (jumpZero jmpTarget: (jumpNotSI jmpTarget: cogit Label))).
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:29'!
genPrimitiveEqual
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self 
				genSmallIntegerComparison: JumpZero
				orDoubleComparison: #JumpFPEqual:
				invert: false]
		ifFalse: [self genSmallIntegerComparison: JumpZero]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/5/2019 21:38'!
genPrimitiveFloatAdd
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleArithmetic: AddRdRd preOpCheck: nil]
		ifFalse: [self genPureDoubleArithmetic: AddRdRd preOpCheck: nil]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/5/2019 21:38'!
genPrimitiveFloatDivide
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleArithmetic: DivRdRd preOpCheck: #genDoubleFailIfZeroArgRcvr:arg:]
		ifFalse: [self genPureDoubleArithmetic: DivRdRd preOpCheck: #genDoubleFailIfZeroArgRcvr:arg:]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:29'!
genPrimitiveFloatEqual
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleComparison: #JumpFPEqual: invert: false]
		ifFalse: [self genPureDoubleComparison: #JumpFPEqual: invert: false]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:30'!
genPrimitiveFloatGreaterOrEqual
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleComparison: #JumpFPGreaterOrEqual: invert: false]
		ifFalse: [self genPureDoubleComparison: #JumpFPGreaterOrEqual: invert: false]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:30'!
genPrimitiveFloatGreaterThan
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleComparison: #JumpFPGreater: invert: false]
		ifFalse: [self genPureDoubleComparison: #JumpFPGreater: invert: false]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:32'!
genPrimitiveFloatLessOrEqual
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleComparison: #JumpFPGreaterOrEqual: invert: true]
		ifFalse: [self genPureDoubleComparison: #JumpFPGreaterOrEqual: invert: true]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:31'!
genPrimitiveFloatLessThan
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleComparison: #JumpFPGreater: invert: true]
		ifFalse: [self genPureDoubleComparison: #JumpFPGreater: invert: true]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/5/2019 21:39'!
genPrimitiveFloatMultiply
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleArithmetic: MulRdRd preOpCheck: nil]
		ifFalse: [self genPureDoubleArithmetic: MulRdRd preOpCheck: nil]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:31'!
genPrimitiveFloatNotEqual
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleComparison: #JumpFPNotEqual: invert: false]
		ifFalse: [self genPureDoubleComparison: #JumpFPNotEqual: invert: false]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 13:54'!
genPrimitiveFloatSquareRoot
	<option: #DPFPReg0>
	| jumpFailAlloc |
	<var: #jumpFailAlloc type: #'AbstractInstruction *'>
	cogit processorHasDoublePrecisionFloatingPointSupport ifFalse:
		[^UnimplementedPrimitive].
	self genGetDoubleValueOf: ReceiverResultReg into: DPFPReg0.
	cogit SqrtRd: DPFPReg0.
	jumpFailAlloc := self
						genAllocFloatValue: DPFPReg0
						into: SendNumArgsReg
						scratchReg: ClassReg
						scratchReg: TempReg.
	cogit MoveR: SendNumArgsReg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpFailAlloc jmpTarget: cogit Label.
	^0! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/5/2019 21:39'!
genPrimitiveFloatSubtract
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleArithmetic: SubRdRd preOpCheck: nil]
		ifFalse: [self genPureDoubleArithmetic: SubRdRd preOpCheck: nil]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'cb 4/11/2016 16:16'!
genPrimitiveFullClosureValue
	"Defer to the cogit for this one, to match the split for genPrimitiveClosureValue."
	<doNotGenerate>
	<option: #SistaV1BytecodeSet>
	^cogit genPrimitiveFullClosureValue! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:31'!
genPrimitiveGreaterOrEqual
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self
				genSmallIntegerComparison: JumpGreaterOrEqual
				orDoubleComparison: #JumpFPGreaterOrEqual:
				invert: false]
		ifFalse: [self genSmallIntegerComparison: JumpGreaterOrEqual]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:31'!
genPrimitiveGreaterThan
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self
				genSmallIntegerComparison: JumpGreater
				orDoubleComparison: #JumpFPGreater:
				invert: false]
		ifFalse: [self genSmallIntegerComparison: JumpGreater]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 4/17/2017 15:24'!
genPrimitiveHashMultiply
	<doNotGenerate>
	^cogit genPrimitiveHashMultiply! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/23/2020 16:27'!
genPrimitiveHighBit
	"Implementation notes: there are two reasons to use TempReg
	-1) if primitive fails, ReceiverResultReg must remain unchanged (we CompletePrimitive)
	-2) CLZ/BSR only work on 64bits for registers R0-R7 on Intel X64. But Win64 uses R9
	Normally, this should be backEnd dependent, but for now we have a single 64bits target..."
	
	| jumpNegativeReceiver |
	<var: #jumpNegativeReceiver type: #'AbstractInstruction *'>
	"remove excess tag bits from the receiver oop"
	self numSmallIntegerTagBits > 1
		ifTrue:
			[cogit OrCq: 1 << self numSmallIntegerTagBits - 1 R: ReceiverResultReg R: TempReg.
			 cogit ArithmeticShiftRightCq: self numSmallIntegerTagBits - 1 R: TempReg.
			 jumpNegativeReceiver := cogit backEnd genHighBitIn: TempReg ofSmallIntegerOopWithSingleTagBit: TempReg]
		ifFalse:
			["and use the abstract cogit facility for case of single tag-bit"
			jumpNegativeReceiver := cogit backEnd genHighBitIn: TempReg ofSmallIntegerOopWithSingleTagBit: ReceiverResultReg].
	
	"Jump is NULL if above operation is not implemented, else return the result"
	jumpNegativeReceiver = 0 ifTrue:
		[^UnimplementedPrimitive].

	cogit MoveR: TempReg R: ReceiverResultReg.
	self genConvertIntegerToSmallIntegerInReg: ReceiverResultReg.
	cogit genPrimReturn.
	jumpNegativeReceiver jmpTarget: cogit Label.
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:32'!
genPrimitiveIdentical
	^self genPrimitiveIdenticalOrNotIf: false! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 2/28/2016 10:47'!
genPrimitiveImmediateAsInteger
	"subclasses override if they can"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 6/20/2018 13:03'!
genPrimitiveIntegerAt
	"subclasses override if they can"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 2/21/2019 09:16'!
genPrimitiveIntegerAtPut
	"subclasses override if they can"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:31'!
genPrimitiveLessOrEqual
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self
				genSmallIntegerComparison: JumpLessOrEqual
				orDoubleComparison: #JumpFPGreaterOrEqual:
				invert: true]
		ifFalse: [self genSmallIntegerComparison: JumpLessOrEqual]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:31'!
genPrimitiveLessThan
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self
				genSmallIntegerComparison: JumpLess
				orDoubleComparison: #JumpFPGreater:
				invert: true]
		ifFalse: [self genSmallIntegerComparison: JumpLess]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/22/2021 20:39'!
genPrimitiveMakePoint
	"subclasses override if they so desire"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 13:38'!
genPrimitiveMod
	| jumpNotSI jumpZero jumpExact jumpSameSign |
	<var: #jumpNotSI type: #'AbstractInstruction *'>
	<var: #jumpZero type: #'AbstractInstruction *'>
	<var: #jumpExact type: #'AbstractInstruction *'>
	<var: #jumpSameSign type: #'AbstractInstruction *'>
	cogit processorHasDivQuoRemAndMClassIsSmallInteger ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	cogit MoveR: Arg0Reg R: ClassReg.
	jumpNotSI := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	self genRemoveSmallIntegerTagsInScratchReg: ClassReg.
	jumpZero := cogit JumpZero: 0.
	cogit MoveR: ClassReg R: Arg1Reg.
	cogit MoveR: ReceiverResultReg R: TempReg.
	self genRemoveSmallIntegerTagsInScratchReg: TempReg.
	cogit DivR: ClassReg R: TempReg Quo: TempReg Rem: ClassReg.
	"If remainder is zero we're done."
	cogit CmpCq: 0 R: ClassReg.
	jumpExact := cogit JumpZero: 0.
	"If arg and remainder signs are different we must reflect around zero."
	cogit XorR: ClassReg R: Arg1Reg.
	(cogit lastOpcode setsConditionCodesFor: JumpZero) ifFalse:
		[cogit CmpCq: 0 R: Arg1Reg].
	jumpSameSign := cogit JumpGreaterOrEqual: 0.
	cogit XorR: ClassReg R: Arg1Reg.
	cogit AddR: Arg1Reg R: ClassReg.
	jumpSameSign jmpTarget: (jumpExact jmpTarget: cogit Label).
	self genSetSmallIntegerTagsIn: ClassReg.
	cogit MoveR: ClassReg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpZero jmpTarget: (jumpNotSI jmpTarget: cogit Label).
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 7/4/2020 11:35'!
genPrimitiveMultiply
	| jumpNotSI jumpOvfl |
	<var: #jumpNotSI type: #'AbstractInstruction *'>
	<var: #jumpOvfl type: #'AbstractInstruction *'>
	(cogit processorHasMultiplyAndMClassIsSmallInteger) ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	cogit MoveR: Arg0Reg R: ClassReg.
	cogit MoveR: ReceiverResultReg R: Arg1Reg.
	jumpNotSI := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	self genShiftAwaySmallIntegerTagsInScratchReg: ClassReg.
	self genRemoveSmallIntegerTagsInScratchReg: Arg1Reg.
	cogit MulOverflowR: Arg1Reg R: ClassReg.
	jumpOvfl := cogit JumpOverflow: 0.
	self genSetSmallIntegerTagsIn: ClassReg.
	cogit MoveR: ClassReg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpOvfl jmpTarget: (jumpNotSI jmpTarget: cogit Label).
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:09'!
genPrimitiveNew
	"subclasses override if they can"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 11:01'!
genPrimitiveNewMethod
	"subclasses override if they can"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:20'!
genPrimitiveNewWithArg
	"subclasses override if they can"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:31'!
genPrimitiveNotEqual
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self 
				genSmallIntegerComparison: JumpNonZero
				orDoubleComparison: #JumpFPNotEqual:
				invert: false]
		ifFalse: [self genSmallIntegerComparison: JumpNonZero]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:32'!
genPrimitiveNotIdentical
	^self genPrimitiveIdenticalOrNotIf: true! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:08'!
genPrimitiveObjectAt
	"subclasses override if they can"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:48'!
genPrimitivePerform
	"Defer to the cogit for this one.  All the code is method-cache related."
	<doNotGenerate>
	^cogit genPrimitivePerform! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 12/26/2019 21:53'!
genPrimitiveQuo
	| convert jumpNotSI jumpZero jumpIsSI jumpExact |
	<var: #convert type: #'AbstractInstruction *'>
	<var: #jumpIsSI type: #'AbstractInstruction *'>
	<var: #jumpZero type: #'AbstractInstruction *'>
	<var: #jumpNotSI type: #'AbstractInstruction *'>
	<var: #jumpExact type: #'AbstractInstruction *'>
	cogit processorHasDivQuoRemAndMClassIsSmallInteger ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	cogit MoveR: Arg0Reg R: ClassReg.
	jumpNotSI := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	"We must shift away the tags, not just subtract them, so that the
	 overflow case doesn't actually overflow the machine instruction."
	self genShiftAwaySmallIntegerTagsInScratchReg: ClassReg.
	(cogit lastOpcode setsConditionCodesFor: JumpZero) ifFalse:
		[cogit CmpCq: 0 R: ClassReg].
	jumpZero := cogit JumpZero: 0.
	cogit MoveR: ReceiverResultReg R: TempReg.
	self genShiftAwaySmallIntegerTagsInScratchReg: TempReg.
	cogit DivR: ClassReg R: TempReg Quo: TempReg Rem: ClassReg.
	"If remainder is zero we must check for overflow."
	cogit CmpCq: 0 R: ClassReg.
	jumpExact := cogit JumpZero: 0.
	convert := cogit Label.
	self genConvertIntegerInReg: TempReg toSmallIntegerInReg: ReceiverResultReg.
	cogit genPrimReturn.
	jumpExact jmpTarget: cogit Label.
	jumpIsSI := self genJumpIsSmallIntegerValue: TempReg scratch: Arg1Reg.
	jumpIsSI jmpTarget: convert.
	jumpZero jmpTarget: (jumpNotSI jmpTarget: cogit Label).
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 8/5/2016 09:56'!
genPrimitiveShallowCopy
	"subclasses override if they can"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/5/2019 21:39'!
genPrimitiveSmallFloatAdd
	<option: #Spur64BitMemoryManager>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genSmallFloatArithmetic: AddRdRd preOpCheck: nil]
		ifFalse: [self genPureSmallFloatArithmetic: AddRdRd preOpCheck: nil]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/5/2019 21:38'!
genPrimitiveSmallFloatDivide
	<option: #Spur64BitMemoryManager>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genSmallFloatArithmetic: DivRdRd preOpCheck: #genDoubleFailIfZeroArgRcvr:arg:]
		ifFalse: [self genPureSmallFloatArithmetic: DivRdRd preOpCheck: #genDoubleFailIfZeroArgRcvr:arg:]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:31'!
genPrimitiveSmallFloatEqual
	<option: #Spur64BitMemoryManager>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genSmallFloatComparison: #JumpFPEqual: orIntegerComparison: JumpZero invert: false]
		ifFalse: [self genPureSmallFloatComparison: #JumpFPEqual: invert: false]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:31'!
genPrimitiveSmallFloatGreaterOrEqual
	<option: #Spur64BitMemoryManager>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genSmallFloatComparison: #JumpFPGreaterOrEqual: orIntegerComparison: JumpGreaterOrEqual invert: false]
		ifFalse: [self genPureSmallFloatComparison: #JumpFPGreaterOrEqual: invert: false]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:31'!
genPrimitiveSmallFloatGreaterThan
	<option: #Spur64BitMemoryManager>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genSmallFloatComparison: #JumpFPGreater: orIntegerComparison: JumpGreater invert: false]
		ifFalse: [self genPureSmallFloatComparison: #JumpFPGreater: invert: false]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:30'!
genPrimitiveSmallFloatLessOrEqual
	<option: #Spur64BitMemoryManager>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genSmallFloatComparison: #JumpFPGreaterOrEqual: orIntegerComparison: JumpLessOrEqual invert: true]
		ifFalse: [self genPureSmallFloatComparison: #JumpFPGreaterOrEqual: invert: true]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:30'!
genPrimitiveSmallFloatLessThan
	<option: #Spur64BitMemoryManager>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genSmallFloatComparison: #JumpFPGreater: orIntegerComparison: JumpLess invert: true]
		ifFalse: [self genPureSmallFloatComparison: #JumpFPGreater: invert: true]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/5/2019 21:39'!
genPrimitiveSmallFloatMultiply
	<option: #Spur64BitMemoryManager>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genSmallFloatArithmetic: MulRdRd preOpCheck: nil]
		ifFalse: [self genPureSmallFloatArithmetic: MulRdRd preOpCheck: nil]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:30'!
genPrimitiveSmallFloatNotEqual
	<option: #Spur64BitMemoryManager>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genSmallFloatComparison: #JumpFPNotEqual: orIntegerComparison: JumpNonZero invert: false]
		ifFalse: [self genPureSmallFloatComparison: #JumpFPNotEqual: invert: false]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 4/23/2020 12:50'!
genPrimitiveSmallFloatSquareRoot
	<option: #Spur64BitMemoryManager>
	<option: #DPFPReg0>
	| jumpFailAlloc jumpNegative |
	<var: #jumpFailAlloc type: #'AbstractInstruction *'>
	<var: #jumpNegative type: #'AbstractInstruction *'>
	self genGetSmallFloatValueOf: ReceiverResultReg scratch: SendNumArgsReg into: DPFPReg0.
	cogit
		XorRd: DPFPReg1 Rd: DPFPReg1; "+0.0 is all zeros"
		CmpRd: DPFPReg0 Rd: DPFPReg1.
	jumpNegative := cogit JumpFPGreater: 0.
	cogit SqrtRd: DPFPReg0.
	jumpFailAlloc := self
						genAllocFloatValue: DPFPReg0
						into: SendNumArgsReg
						scratchReg: ClassReg
						scratchReg: TempReg.
	cogit MoveR: SendNumArgsReg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpNegative jmpTarget: (jumpFailAlloc jmpTarget: cogit Label).
	^0! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 9/5/2019 21:40'!
genPrimitiveSmallFloatSubtract
	<option: #Spur64BitMemoryManager>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genSmallFloatArithmetic: SubRdRd preOpCheck: nil]
		ifFalse: [self genPureSmallFloatArithmetic: SubRdRd preOpCheck: nil]! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:07'!
genPrimitiveStringAtPut
	"subclasses override if they can"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'sk 3/11/2018 21:49'!
genPrimitiveStringCompareWith
	"subclasses override if they can"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'cb 10/20/2017 15:38'!
genPrimitiveStringReplace
	"subclasses override if they can"
	^UnimplementedPrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 13:38'!
genPrimitiveSubtract
	| jumpNotSI jumpOvfl |
	<var: #jumpNotSI type: #'AbstractInstruction *'>
	<var: #jumpOvfl type: #'AbstractInstruction *'>
	cogit mclassIsSmallInteger ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	jumpNotSI := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	cogit MoveR: ReceiverResultReg R: TempReg.
	cogit SubR: Arg0Reg R: TempReg.
	jumpOvfl := cogit JumpOverflow: 0.
	self genAddSmallIntegerTagsTo: TempReg.
	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpOvfl jmpTarget: (jumpNotSI jmpTarget: cogit Label).
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 8/28/2019 17:35'!
genPureDoubleArithmetic: arithmeticOperator preOpCheck: preOpCheckOrNil
	"In the Pure version, mixed arithmetic with SmallInteger is forbidden"
	<option: #DPFPReg0>
	<var: #preOpCheckOrNil declareC: 'AbstractInstruction *(*preOpCheckOrNil)(int rcvrReg, int argReg)'>
	| jumpFailClass jumpFailAlloc jumpFailCheck jumpImmediate doOp |
	<var: #jumpFailClass type: #'AbstractInstruction *'>
	<var: #jumpFailAlloc type: #'AbstractInstruction *'>
	<var: #jumpImmediate type: #'AbstractInstruction *'>
	<var: #jumpFailCheck type: #'AbstractInstruction *'>
	<var: #doOp type: #'AbstractInstruction *'>
	cogit processorHasDoublePrecisionFloatingPointSupport ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	self genGetDoubleValueOf: ReceiverResultReg into: DPFPReg0.
	cogit MoveR: Arg0Reg R: ClassReg.
	jumpImmediate := self genJumpImmediate: Arg0Reg.
	self genGetCompactClassIndexNonImmOf: Arg0Reg into: SendNumArgsReg.
	self genCmpClassFloatCompactIndexR: SendNumArgsReg.
	jumpFailClass := cogit JumpNonZero: 0.
	self genGetDoubleValueOf: Arg0Reg into: DPFPReg1.
	doOp := cogit Label.
	preOpCheckOrNil ifNotNil:
		[jumpFailCheck := cogit perform: preOpCheckOrNil with: DPFPReg0 with: DPFPReg1].
	cogit gen: arithmeticOperator operand: DPFPReg1 operand: DPFPReg0.
	jumpFailAlloc := self
						genAllocFloatValue: DPFPReg0
						into: SendNumArgsReg
						scratchReg: ClassReg
						scratchReg: TempReg.
	cogit MoveR: SendNumArgsReg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpImmediate jmpTarget: cogit Label.
	self maybeGenConvertIfSmallFloatIn: Arg0Reg scratchReg: TempReg into: DPFPReg1 andJumpTo: doOp.
	jumpFailAlloc jmpTarget:  (jumpFailClass jmpTarget: cogit Label).
	preOpCheckOrNil ifNotNil:
		[jumpFailCheck jmpTarget: jumpFailClass getJmpTarget].
	^0! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'nice 8/28/2019 17:52'!
genPureDoubleComparison: jumpOpcodeGenerator invert: invertComparison
	"In the Pure version, mixed arithmetic with SmallInteger is forbidden"
	<option: #DPFPReg0>
	<var: #jumpOpcodeGenerator declareC: 'AbstractInstruction * NoDbgRegParms (*jumpOpcodeGenerator)(void *)'>
	| jumpFail jumpImmediate jumpCond compare |
	<var: #jumpImmediate type: #'AbstractInstruction *'>
	<var: #jumpCond type: #'AbstractInstruction *'>
	<var: #compare type: #'AbstractInstruction *'>
	<var: #jumpFail type: #'AbstractInstruction *'>
	cogit processorHasDoublePrecisionFloatingPointSupport ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	self genGetDoubleValueOf: ReceiverResultReg into: DPFPReg0.
	jumpImmediate := self genJumpImmediate: Arg0Reg.
	self genGetCompactClassIndexNonImmOf: Arg0Reg into: SendNumArgsReg.
	self genCmpClassFloatCompactIndexR: SendNumArgsReg.
	jumpFail := cogit JumpNonZero: 0.
	self genGetDoubleValueOf: Arg0Reg into: DPFPReg1.
	invertComparison "May need to invert for NaNs"
		ifTrue: [compare := cogit CmpRd: DPFPReg0 Rd: DPFPReg1]
		ifFalse: [compare := cogit CmpRd: DPFPReg1 Rd: DPFPReg0].
	jumpCond := cogit perform: jumpOpcodeGenerator with: 0. "FP jumps are a little weird"
	cogit genMoveFalseR: ReceiverResultReg.
	cogit genPrimReturn.
	jumpCond jmpTarget: (cogit genMoveTrueR: ReceiverResultReg).
	cogit genPrimReturn.
	jumpImmediate jmpTarget: cogit Label.
	self maybeGenConvertIfSmallFloatIn: Arg0Reg scratchReg: TempReg into: DPFPReg1 andJumpTo: compare.
	jumpFail jmpTarget: cogit Label.
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 12:45'!
genQuickReturnConst
	<doNotGenerate>
	^cogit genQuickReturnConst! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 12:46'!
genQuickReturnInstVar
	<doNotGenerate>
	^cogit genQuickReturnInstVar! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 12:46'!
genQuickReturnSelf
	<doNotGenerate>
	^cogit genQuickReturnSelf! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'eem 12/6/2015 20:34'!
genSetSmallIntegerTagsIn: scratchReg
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 13:38'!
genSmallIntegerComparison: jumpOpcode
	| jumpFail jumpTrue |
	<var: #jumpFail type: #'AbstractInstruction *'>
	<var: #jumpTrue type: #'AbstractInstruction *'>
	cogit mclassIsSmallInteger ifFalse:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	jumpFail := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	cogit CmpR: Arg0Reg R: ReceiverResultReg. "N.B. FLAGS := RRReg - Arg0Reg"
	jumpTrue := cogit genConditionalBranch: jumpOpcode operand: 0.
	cogit genMoveFalseR: ReceiverResultReg.
	cogit genPrimReturn.
	jumpTrue jmpTarget: (cogit genMoveTrueR: ReceiverResultReg).
	cogit genPrimReturn.
	jumpFail jmpTarget: cogit Label.
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 1/29/2020 13:51'!
genSmallIntegerComparison: jumpOpcode orDoubleComparison: jumpFPOpcodeGenerator invert: invertComparison
	"Stack looks like
		return address"
	| jumpNonInt jumpFail jumpCond r |
	<var: 'jumpFPOpcodeGenerator' declareC: 'AbstractInstruction * NoDbgRegParms (*jumpFPOpcodeGenerator)(void *)'>
	<var: 'jumpCond' type: #'AbstractInstruction *'>
	r := self genSmallIntegerComparison: jumpOpcode.
	r < 0 ifTrue:
		[^r].
	self cppIf: #DPFPReg0 defined ifTrue:
	"Fall through on non-SmallInteger argument.  Argument may be a Float : let us check or fail"
	[self smallIntegerIsOnlyImmediateType ifFalse:
		[jumpNonInt := self genJumpImmediate: Arg0Reg].
	self genGetCompactClassIndexNonImmOf: Arg0Reg into: SendNumArgsReg.
	self genCmpClassFloatCompactIndexR: SendNumArgsReg.
	jumpFail := cogit JumpNonZero: 0.

	"It was a Float, so convert the receiver to double and perform the operation"
	self genConvertSmallIntegerToIntegerInReg: ReceiverResultReg.
	cogit ConvertR: ReceiverResultReg Rd: DPFPReg0.
	self genGetDoubleValueOf: Arg0Reg into: DPFPReg1.
	invertComparison "May need to invert for NaNs"
		ifTrue: [cogit CmpRd: DPFPReg0 Rd: DPFPReg1]
		ifFalse: [cogit CmpRd: DPFPReg1 Rd: DPFPReg0].
	jumpCond := cogit perform: jumpFPOpcodeGenerator with: 0. "FP jumps are a little weird"
	cogit genMoveFalseR: ReceiverResultReg.
	cogit genPrimReturn.
	jumpCond jmpTarget: (cogit genMoveTrueR: ReceiverResultReg).
	cogit genPrimReturn.

	self smallIntegerIsOnlyImmediateType
		ifTrue: [jumpFail jmpTarget: cogit Label]
		ifFalse: [jumpNonInt jmpTarget: (jumpFail jmpTarget: cogit Label)]].
	^CompletePrimitive! !

!CogObjectRepresentation methodsFor: 'compile abstract instructions' stamp: 'cb 3/30/2016 15:17'!
genStoreSourceReg: sourceReg slotIndex: index destReg: destReg scratchReg: scratchReg inFrame: inFrame
	<inline: true>
	self 
		cppIf: IMMUTABILITY
		ifTrue:
			[ ^ self 
				genStoreWithImmutabilityCheckSourceReg: sourceReg 
				slotIndex: index 
				destReg: destReg 
				scratchReg: scratchReg 
				needsStoreCheck: true 
				needRestoreRcvr: false "RcvrResultReg doesn't need to be live across the instructions" ]
		ifFalse: 
			[ ^ self 
				genStoreSourceReg: sourceReg 
				slotIndex: index 
				destReg: destReg 
				scratchReg: scratchReg 
				inFrame: inFrame 
				needsStoreCheck: true ]! !

!CogObjectRepresentation methodsFor: 'initialization' stamp: 'eem 2/11/2020 10:18'!
generateLowcodeObjectTrampolines
	<option: #LowcodeVM>
	ceFloatObjectOfTrampoline := cogit genTrampolineFor: #floatObjectOf:
												called: 'ceFloatObjectOfTrampoline'
												floatArg: DPFPReg0
												result: TempReg.
	ceFloatValueOfTrampoline := cogit genTrampolineFor: #floatValueOf:
												called: 'ceFloatValueOfTrampoline'
												arg: ReceiverResultReg
												floatResult: DPFPReg0.
	ceInstantiateClassIndexableSizeTrampoline := cogit genTrampolineFor: #instantiateClass:indexableSize:
												called: 'ceInstantiateClassIndexableSizeTrampoline'
												arg: Arg0Reg
												arg: Arg1Reg
												result: TempReg.
	ceInstantiateClassTrampoline := cogit genTrampolineFor: #instantiateClass:indexableSize:
												called: 'ceInstantiateClassTrampoline'
												arg: ReceiverResultReg
												arg: 0
												result: TempReg.
	ceByteSizeOfTrampoline := cogit genTrampolineFor: #byteSizeOf:
												called: 'ceByteSizeOfTrampoline'
												arg: Arg0Reg
												result: TempReg.
	objectMemory bytesPerOop = 4
		ifTrue:
			[cePositive64BitIntegerTrampoline := cogit genTrampolineFor: #positive64BitIntegerFor:
														called: 'cePositive64BitIntegerTrampoline'
														arg: ReceiverResultReg
														arg: Arg0Reg
														result: TempReg.
			cePositive64BitValueOfTrampoline := cogit genTrampolineFor: #positive64BitValueOf:
														called: 'cePositive64BitValueOfTrampoline'
														arg: ReceiverResultReg
														result: TempReg
														result: Arg0Reg.
			ceSigned64BitIntegerTrampoline := cogit genTrampolineFor: #signed64BitIntegerFor:
														called: 'ceSigned64BitIntegerTrampoline'
														arg: ReceiverResultReg
														arg: Arg0Reg
														result: TempReg.
			ceSigned64BitValueOfTrampoline := cogit genTrampolineFor: #signed64BitValueOf:
														called: 'ceSigned64BitValueOfTrampoline'
														arg: ReceiverResultReg
														result: TempReg
														result: Arg0Reg]
		ifFalse:
			[cePositive64BitIntegerTrampoline := cogit genTrampolineFor: #positive64BitIntegerFor:
														called: 'cePositive64BitIntegerTrampoline'
														arg: ReceiverResultReg
														result: TempReg.
			cePositive64BitValueOfTrampoline := cogit genTrampolineFor: #positive64BitValueOf:
														called: 'cePositive64BitValueOfTrampoline'
														arg: ReceiverResultReg
														result: TempReg.
			ceSigned64BitIntegerTrampoline := cogit genTrampolineFor: #signed64BitIntegerFor:
														called: 'ceSigned64BitIntegerTrampoline'
														arg: ReceiverResultReg
														result: TempReg.
			ceSigned64BitValueOfTrampoline := cogit genTrampolineFor: #signed64BitValueOf:
														called: 'ceSigned64BitValueOfTrampoline'
														arg: ReceiverResultReg
														result: TempReg]! !

!CogObjectRepresentation methodsFor: 'initialization' stamp: 'eem 7/14/2014 17:02'!
generateObjectRepresentationTrampolines
	"Do the store check.  Answer the argument for the benefit of the code generator;
	 ReceiverResultReg may be caller-saved and hence smashed by this call.  Answering
	 it allows the code generator to reload ReceiverResultReg cheaply."
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'bytecode generator support' stamp: 'eem 5/31/2014 16:49'!
getActiveContextAllocatesInMachineCode
	"subclasses override if they want, and will have SendNumArgsReg and ClassReg
	 available in addition to ReceiverResultReg and TempReg if they do."
	^false! !

!CogObjectRepresentation methodsFor: 'testing' stamp: 'eem 8/3/2021 18:51'!
hasSpurMemoryManagerAPI
	<inline: #always>
	^false! !

!CogObjectRepresentation methodsFor: 'in-line cacheing' stamp: 'rmacnak 4/4/2015 16:08'!
illegalClassTag
	"Answer a class tag that will fail for every object."
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'in-line cacheing' stamp: 'eem 7/7/2014 09:49'!
inlineCacheTagForClass: classOop
	"Answer the relevant inline cache tag for a class.
	 c.f. inlineCacheTagForInstance:"
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'in-line cacheing' stamp: 'eem 7/7/2014 09:56'!
inlineCacheTagForInstance: oop
	"Answer the relevant inline cache tag for an instance.
	 c.f. getInlineCacheClassTagFrom:into: & inlineCacheTagForClass:"
	^self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'in-line cacheing' stamp: 'eem 11/25/2013 17:03'!
inlineCacheTagIsYoung: cacheTag
	^self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'in-line cacheing' stamp: 'eem 5/27/2015 11:24'!
inlineCacheTagsMayBeObjects
	<inline: true>
	^self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'testing' stamp: 'eem 10/25/2013 15:03'!
isImmediate: oop
	<doNotGenerate>
	^objectMemory isImmediate: oop! !

!CogObjectRepresentation methodsFor: 'testing' stamp: 'eem 10/2/2013 17:09'!
isSmallIntegerTagNonZero
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'testing' stamp: 'cb 9/13/2019 11:51'!
isUnannotatableConstant: simStackEntry
	<inline: true>
	<var: 'simStackEntry' type: #'CogSimStackEntry *'>
	^simStackEntry type = SSConstant 
	  and: [(objectMemory isImmediate: simStackEntry constant)
		or: [(self shouldAnnotateObjectReference: simStackEntry constant) not]]! !

!CogObjectRepresentation methodsFor: 'garbage collection' stamp: 'eem 09/28/2015 18:48'!
markAndTraceCacheTagLiteral: literal in: cogMethodOrNil atpc: address
	"Mark and trace a literal in an inline cache preceding address in cogMethodOrNil.
	 Answer if code was modified."
	<var: #cogMethodOrNil type: #'CogMethod *'>
	<var: #address type: #usqInt>
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'garbage collection' stamp: 'eem 11/18/2013 15:39'!
markAndTraceLiteral: literal
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'garbage collection' stamp: 'eem 8/4/2014 08:20'!
markAndTraceLiteral: literal in: cogMethod at: address
	"Mark and trace a literal in a sqInt variable of cogMethod."
	<var: #cogMethod type: #'CogMethod *'>
	<var: #address type: #'sqInt *'>
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'garbage collection' stamp: 'eem 09/28/2015 18:48'!
markAndTraceLiteral: literal in: cogMethodOrNil atpc: address
	"Mark and trace a literal in a machine code instruction preceding address in cogMethodOrNil.
	 Answer if code was modified."
	<var: #cogMethodOrNil type: #'CogMethod *'>
	<var: #address type: #usqInt>
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'garbage collection' stamp: 'eem 11/18/2013 15:44'!
markAndTraceLiteralIfYoung: literal
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 10/31/2014 09:32'!
maybeCompileRetry: retryInst onPrimitiveFail: primIndex primPropertyFlags: flags
	<var: #retryInst type: #'AbstractInstruction *'>
	"Object representations with lazy forwarding will want to check for
	 forwarding pointers on primitive failure and retry the primitive if found.
	 By default do nothing."! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 6/3/2016 13:45'!
maybeCompileRetryOnPrimitiveFail: primIndex
	"Object representations with lazy forwarding will want to check for
	 forwarding pointers on primitive failure and retry the primitive if found.
	 By default do nothing."
	<inline: true>! !

!CogObjectRepresentation methodsFor: 'primitive generators' stamp: 'eem 12/7/2015 22:20'!
maybeGenConvertIfSmallFloatIn: oopReg scratchReg: scratch into: dpReg andJumpTo: targetInst
	"If the receiver supports immediate floats then generate a test for a smallFloat in  oopReg,
	 converting it to the float value in dpReg and jumping to targetInst.  Otherwise do nothing."
	<var: 'targetInst' type: #'AbstractInstruction *'>
	^0! !

!CogObjectRepresentation methodsFor: 'initialization' stamp: 'eem 01/10/2016 10:01'!
maybeGenerateSelectorIndexDereferenceRoutine
	"If required, generate the routine that converts selector indices into selector objects.
	 If it exists, it is called from the send trampolines.
	 If the selector index is negative, convert it into a positive index into the
	 special selectors array and index that.  Otherwise, index the current method.
	 This is only required on 64-bit platforms."! !

!CogObjectRepresentation methodsFor: 'sista support' stamp: 'eem 10/21/2014 18:52'!
maybeMarkCounters: theCounters
	"In SIsta Spur counters are held on the heap in pinned objects which must be marked
	 to avoid them being garbage collected.  This is the hook through which that happens.
	 By default, do nothing."
	<var: #theCounters type: #usqInt>
	<inline: true>! !

!CogObjectRepresentation methodsFor: 'compilation' stamp: 'eem 6/4/2015 09:31'!
maybeNoteDescriptor: descriptor blockStart: blockStart
	"A hook for the object representation to set state in a blockStart during scanBlock:.
	 By default do nothing.  Subclasses that need to do something special will override."
	<var: #blockStart type: #'BlockStart *'>
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<inline: true>! !

!CogObjectRepresentation methodsFor: 'method cacheing' stamp: 'eem 8/4/2016 11:47'!
maybeShiftClassTagRegisterForMethodCacheProbe: classTagReg
	"If required, generate a shift of the register containing the class tag in a method cache probe.
	 By default this is a no-op.  Subclasses redefine as required."

	^0! !

!CogObjectRepresentation methodsFor: 'sista support' stamp: 'eem 7/5/2014 17:54'!
numCountersFor: theCounters
	<var: #counters type: #usqInt>
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'calling convention' stamp: 'eem 5/1/2014 07:20'!
numRegArgs
	"Define how many register arguments a StackToRegisterMappingCogit can and should use
	 with the receiver.  The value must be 0, 1 or 2.  Note that a SimpleStackBasedCogit always
	 has 0 register args (although the receiver is passed in a register).  The method must
	 be inlined in CoInterpreter, and dead code eliminated so that the register-popping
	 enilopmarts such as enterRegisterArgCogMethod:at:receiver: do not have to be
	 implemented in SimpleStackBasedCogit."
	<api>
	<option: #StackToRegisterMappingCogit>
	<inline: true>
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'sista support' stamp: 'eem 7/4/2014 12:04'!
resetCountersIn: cogMethod
	<var: #cogMethod type: #'CogMethod *'>
	<api>
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'calling convention' stamp: 'eem 01/10/2016 10:01'!
selectorIndexDereferenceRoutine
	"If the object representation is 64-bits, selectors must be accessed via indices in the inline caches.
	 Answer the routine that will map from indices to selector objects if so, otherwise nil."
	^nil! !

!CogObjectRepresentation methodsFor: 'initialization' stamp: 'eem 6/30/2014 06:50'!
setCogit: aCogit methodZone: aMethodZone
	<doNotGenerate>
	cogit := aCogit.
	methodZone := aMethodZone.
	coInterpreter := aCogit coInterpreter.
	objectMemory := (coInterpreter isKindOf: StackInterpreter)
						ifTrue: [coInterpreter objectMemory]
						ifFalse: [coInterpreter]! !

!CogObjectRepresentation methodsFor: 'testing' stamp: 'eem 10/3/2013 10:17'!
smallIntegerIsOnlyImmediateType
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'debug support' stamp: 'eem 12/5/2014 18:44'!
validInlineCacheTag: classIndexOrTagPattern
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'sista support' stamp: 'eem 4/6/2015 09:27'!
valueOfAssociation: association
	self subclassResponsibility! !

!CogObjectRepresentation methodsFor: 'accessing' stamp: 'eem 12/2/2016 09:15'!
wordSize
	<doNotGenerate>
	^objectMemory wordSize! !

CogObjectRepresentation subclass: #CogObjectRepresentationForSpur
	instanceVariableNames: 'ceScheduleScavengeTrampoline ceSmallActiveContextInMethodTrampoline ceSmallActiveContextInBlockTrampoline ceSmallActiveContextInFullBlockTrampoline ceLargeActiveContextInMethodTrampoline ceLargeActiveContextInBlockTrampoline ceLargeActiveContextInFullBlockTrampoline ceStoreCheckContextReceiverTrampoline ceStoreTrampolines ceNewHashTrampoline ceInlineNewHashTrampoline'
	classVariableNames: 'CheckRememberedInTrampoline NumStoreTrampolines'
	poolDictionaries: 'VMBytecodeConstants VMSqueakClassIndices'
	category: 'VMMaker-JIT'!

CogObjectRepresentationForSpur subclass: #CogObjectRepresentationFor32BitSpur
	instanceVariableNames: 'ceSigned32BitIntegerTrampoline ceSigned32BitValueOfTrampoline cePositive32BitValueOfTrampoline cePositive32BitIntegerTrampoline'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogObjectRepresentationFor32BitSpur class methodsFor: 'accessing class hierarchy' stamp: 'eem 1/22/2014 15:53'!
defaultObjectMemoryClass
	"For in-image compilation."
	^Spur32BitCoMemoryManager! !

!CogObjectRepresentationFor32BitSpur class methodsFor: 'trampoline support' stamp: 'rsf 1/4/2017 22:01'!
numTrampolines
	^ super numTrampolines + (LowcodeVM ifTrue: [4] ifFalse: [0])! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'sista support' stamp: 'cb 12/11/2016 16:45'!
allImmediate: immediateMask branchIf: reg instanceOfBehaviors: arrayObj target: targetFixUp
	< inline: true>	
	<var: #targetFixUp type: #'AbstractInstruction *'>
	self assert: immediateMask = objectMemory tagMask.
	(self genJumpImmediate: reg) jmpTarget: targetFixUp.
	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'sista support' stamp: 'cb 12/11/2016 16:44'!
allImmediate: immediateMask branchIf: reg notInstanceOfBehaviors: arrayObj target: targetFixUp
	< inline: true>	
	<var: #targetFixUp type: #'AbstractInstruction *'>
	self assert: immediateMask = objectMemory tagMask.
	(self genJumpNotImmediate: reg) jmpTarget: targetFixUp.
	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'sista support' stamp: 'eem 7/9/2014 21:20'!
allocateCounters: nCounters
	"On Spur allocate the counters on the heap as pinned objects.
	 Each counter occupies 32-bits.  The number of counters can
	 be derived from the number of slots in the obj."
	<inline: true>
	| objOop |
	objOop := objectMemory allocatePinnedSlots: nCounters.
	^objOop
		ifNil: [0]
		ifNotNil: [objOop + objectMemory baseHeaderSize]! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'sista support' stamp: 'cb 12/11/2016 16:52'!
branch2CasesIf: reg instanceOfBehaviors: arrayObj target: targetFixUp
	<inline: true>
	self Jump: targetFixUp! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'sista support' stamp: 'cb 4/16/2015 12:05'!
branch2CasesIf: reg notInstanceOfBehaviors: arrayObj target: targetFixUp
	"nothing to do, 2 immediates so all immediates are allowed"
	<inline: true>! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'sista support' stamp: 'eem 1/21/2017 08:07'!
branchIf: reg hasImmediateTag: classIndex target: targetFixUp
	<var: #targetFixUp type: #'AbstractInstruction *'>
	| jmpImmediate|
	<inline: true>
	classIndex = objectMemory smallIntegerTag ifTrue:
		[jmpImmediate := self genJumpSmallInteger: reg].
	classIndex = objectMemory characterTag ifTrue:
		[cogit MoveR: reg R: TempReg.
		 jmpImmediate := self genJumpCharacterInScratchReg: reg].
	jmpImmediate jmpTarget: targetFixUp! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'sista support' stamp: 'cb 5/10/2017 13:53'!
branchIf: reg hasNotImmediateTag: classIndex target: targetFixUp
	<var: #targetFixUp type: #'AbstractInstruction *'>
	| jmpImmediate |
	<inline: true>
	classIndex = objectMemory smallIntegerTag ifTrue:
		[jmpImmediate := self genJumpNotSmallInteger: reg].
	classIndex = objectMemory characterTag ifTrue:
		["Character test destroy register value in Spur"
		 cogit MoveR: reg R: TempReg.
		 jmpImmediate := self genJumpNotCharacterInScratchReg: TempReg].
	jmpImmediate jmpTarget: targetFixUp! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'in-line cacheing' stamp: 'eem 3/14/2017 12:08'!
classForInlineCacheTag: inlineCacheTag
	"Character gets mapped to zero.  See inlineCacheTagForInstance:."
	^objectMemory classOrNilAtIndex: (inlineCacheTag = 0
											ifTrue: [objectMemory characterTag]
											ifFalse: [inlineCacheTag])! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/2/2013 15:08'!
genAddSmallIntegerTagsTo: aRegister
	cogit AddCq: 1 R: aRegister.
	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/6/2015 20:40'!
genClearAndSetSmallIntegerTagsIn: scratchReg
	"Set the SmallInteger tag bits when the tag bits may be filled with garbage."
	^self genSetSmallIntegerTagsIn: scratchReg! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 1/11/2016 18:05'!
genConvertCharacterToSmallIntegerInReg: reg
	"Convert the Character in reg to a SmallInteger, assuming
	 the Character's value is a valid character."
	"self assume: objectMemory smallIntegerTag = 1"
	self assert: (objectMemory characterTag = 2
				 and: [self numCharacterBits + 1 = self numSmallIntegerBits]).
	cogit LogicalShiftRightCq: 1 R: reg! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/26/2019 21:49'!
genConvertIntegerInReg: srcReg toSmallIntegerInReg: destReg
	cogit LogicalShiftLeftCq: 1 R: srcReg R: destReg.
	cogit AddCq: 1 R: destReg.
	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/6/2013 14:50'!
genConvertIntegerToSmallIntegerInReg: reg
	cogit LogicalShiftLeftCq: 1 R: reg.
	cogit AddCq: 1 R: reg.
	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/24/2013 18:52'!
genConvertSmallIntegerToCharacterInReg: reg
	"Convert the SmallInteger in reg to a Character, assuming
	 the SmallInteger's value is a valid character."
	"self assume: objectMemory smallIntegerTag = 1"
	self assert: (objectMemory characterTag = 2
				 and: [self numCharacterBits + 1 = self numSmallIntegerBits]).
	cogit LogicalShiftLeftCq: 1 R: reg! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/6/2013 14:51'!
genConvertSmallIntegerToIntegerInReg: reg
	cogit ArithmeticShiftRightCq: 1 R: reg.
	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'cb 4/17/2015 15:23'!
genFetchIndexRegister: indexReg from: tableObj into: destReg
	"indexReg contains the 1-relative index of an element in tableObj.
	 Since BaseHeaderSize > BytesPerOop we must adjust it to use
	 it as a zero-relative index from the beginning of the object."
	self assert: indexReg ~= destReg.
	cogit AddCq: objectMemory baseHeaderSize / objectMemory wordSize - 1 R: indexReg.
	cogit genMoveConstant: tableObj R: destReg.
	cogit MoveXwr: indexReg R: destReg R: destReg.
	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/6/2013 14:50'!
genGetHashFieldNonImmOf: instReg asSmallIntegerInto: destReg
	"Fetch the instance's identity hash into destReg, encoded as a SmallInteger."
	"Get header word in scratchReg"
	cogit MoveMw: 4 r: instReg R: destReg.
	"Shift and mask the field leaving room for the SmallInteger tag."
	cogit AndCq: objectMemory identityHashHalfWordMask R: destReg.
	self genConvertIntegerToSmallIntegerInReg: destReg.
	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/3/2013 14:48'!
genGetHashFieldNonImmOf: instReg into: destReg
	"Fetch the instance's identity hash into destReg, unencoded."
	cogit MoveMw: 4 r: instReg R: destReg.
	cogit AndCq: objectMemory identityHashHalfWordMask R: destReg.
	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 5/23/2015 16:56'!
genGetInlineCacheClassTagFrom: sourceReg into: destReg forEntry: forEntry
	"Extract the inline cache tag for the object in sourceReg into destReg. The inline
	 cache tag for a given object is the value loaded in inline caches to distinguish
	 objects of different classes.  In Spur this is either the tags for immediates, (with
	 1 & 3 collapsed to 1 for SmallIntegers, and 2 collapsed to 0 for Characters), or
	 the receiver's classIndex.
	 If forEntry is true answer the entry label at which control is to enter (cmEntryOffset).
	 If forEntry is false, control enters at the start.
	 If forEntry is true, generate something like this:
		Limm:
			andl $0x1, rDest
			j Lcmp
		Lentry:
			movl rSource, rDest
			andl $0x3, rDest
			jnz Limm
			movl 0(%edx), rDest
			andl $0x3fffff, rDest
		Lcmp:
	 If forEntry is false, generate something like the following.
	 At least on a 2.2GHz Intel Core i7 the following is slightly faster than the above,
	 136m sends/sec vs 130m sends/sec for nfib in tinyBenchmarks
		Lentry:
			movl rSource, rDest
			andl $0x3, rDest
			jz LnotImm
			andl $1, rDest
			j Lcmp
		LnotImm:
			movl 0(%edx), rDest
			andl $0x3fffff, rDest
		Lcmp:
	 But we expect most SmallInteger arithmetic to be performed in-line and so prefer the
	 version that is faster for non-immediates (because it branches for immediates only)."
	| immLabel jumpNotImm entryLabel jumpCompare |
	<var: #immLabel type: #'AbstractInstruction *'>
	<var: #jumpNotImm type: #'AbstractInstruction *'>
	<var: #entryLabel type: #'AbstractInstruction *'>
	<var: #jumpCompare type: #'AbstractInstruction *'>
	forEntry
		ifFalse:
			[entryLabel := cogit Label.
			 cogit AndCq: objectMemory tagMask R: sourceReg R: destReg.
			 jumpNotImm := cogit JumpZero: 0.
			 cogit AndCq: 1 R: destReg.
			 jumpCompare := cogit Jump: 0.
			 "Get least significant half of header word in destReg"
			 self flag: #endianness.
			 jumpNotImm jmpTarget:
				(cogit MoveMw: 0 r: sourceReg R: destReg).
			 jumpCompare jmpTarget:
				(cogit AndCq: objectMemory classIndexMask R: destReg)]
		ifTrue:
			[cogit AlignmentNops: objectMemory wordSize.
			 immLabel := cogit Label.
			 cogit AndCq: 1 R: destReg.
			 jumpCompare := cogit Jump: 0.
			 cogit AlignmentNops: objectMemory wordSize.
			 entryLabel := cogit Label.
			 cogit AndCq: objectMemory tagMask R: sourceReg R: destReg.
			 cogit JumpNonZero: immLabel.
			 self flag: #endianness.
			 "Get least significant half of header word in destReg"
			 cogit MoveMw: 0 r: sourceReg R: destReg.
			 cogit AndCq: objectMemory classIndexMask R: destReg.
			 jumpCompare jmpTarget: cogit Label].
	^entryLabel! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/10/2015 08:47'!
genGetNumBytesOf: srcReg into: destReg
	"Get the size in byte-sized slots of the object in srcReg into destReg.
	 srcReg may equal destReg.
	 destReg <- numSlots << self shiftForWord - (fmt bitAnd: 3).
	 Assumes the object in srcReg has a byte format, i.e. 16 to 23 or 24 to 31 "
	<var: #jmp type: #'AbstractInstruction *'>
	| jmp |
	self genGetRawSlotSizeOfNonImm: srcReg into: destReg.
	cogit CmpCq: objectMemory numSlotsMask R: destReg.
	jmp := cogit JumpLess: 0.
	self genGetOverflowSlotsOf: srcReg into: destReg.
	jmp jmpTarget: (cogit LogicalShiftLeftCq: objectMemory shiftForWord R: destReg). 
	self genGetBits: 3 ofFormatByteOf: srcReg into: TempReg.
	"Now: fmt bitAnd: 3 in TempReg"
	cogit SubR: TempReg R: destReg.
	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 4/10/2015 15:54'!
genGetOverflowSlotsOf: srcReg into: destReg
	cogit MoveMw: objectMemory baseHeaderSize negated r: srcReg R: destReg.
	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'primitive generators' stamp: 'eem 10/11/2018 22:25'!
genGetSizeOf: sourceReg into: destReg formatReg: formatReg scratchReg: scratchReg abortJumpsInto: aBinaryBlock
	"Get the size of the non-immediate object in sourceReg into destReg using formatReg
	 and scratchReg as temps.  None of these registers can overlap.  Supply the jumps
	 taken if the object in sourceReg is not indexable, or if the object in sourceReg is a
	 context.. Hack: If the object has a pointer format other than 2 leave the number of
	 fixed fields in formatReg.  Used by primitiveSize, primitiveAt, and primitiveAtPut"
	<returnTypeC: #'AbstractInstruction *'>
	| jumpNotIndexable
	  jumpBytesDone jumpShortsDone jumpArrayDone jump32BitLongsDone jump64BitLongsDone
	  jumpIsBytes jumpHasFixedFields jumpIsShorts jumpIs64BitLongs jumpIsContext  |
	<inline: true>
	"c.f. StackInterpreter>>stSizeOf: SpurMemoryManager>>lengthOf:format: fixedFieldsOf:format:length:"

	"formatReg := self formatOf: sourceReg"
	self genGetFormatOf: sourceReg
		into: formatReg
		leastSignificantHalfOfBaseHeaderIntoScratch: scratchReg.

	self genGetNumSlotsOf: sourceReg into: destReg.

	"dispatch on format in a combination of highest dynamic frequency order first and convenience.
		  0 = 0 sized objects (UndefinedObject True False et al)
		  1 = non-indexable objects with inst vars (Point et al)
		  2 = indexable objects with no inst vars (Array et al)
		  3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
		  4 = weak indexable objects with inst vars (WeakArray et al)
		  5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
		  6 unused, reserved for exotic pointer objects?
		  7 Forwarded Object, 1st field is pointer, rest of fields are ignored
		  8 unused, reserved for exotic non-pointer objects?
		  9 64-bit indexable
		10 - 11 32-bit indexable
		12 - 15 16-bit indexable
		16 - 23 byte indexable
		24 - 31 compiled method"
	cogit CmpCq: objectMemory firstByteFormat R: formatReg.
	jumpIsBytes := cogit JumpGreaterOrEqual: 0.
					cogit CmpCq: objectMemory arrayFormat R: formatReg.
	jumpArrayDone := cogit JumpZero: 0.
	jumpNotIndexable := cogit JumpLess: 0.
					cogit CmpCq: objectMemory weakArrayFormat R: formatReg.
	jumpHasFixedFields := cogit JumpLessOrEqual: 0.
					cogit CmpCq: objectMemory firstShortFormat R: formatReg.
	jumpIsShorts := cogit JumpGreaterOrEqual: 0.
					cogit CmpCq: objectMemory firstLongFormat R: formatReg.
	jump32BitLongsDone := cogit JumpGreaterOrEqual: 0.
					cogit CmpCq: objectMemory sixtyFourBitIndexableFormat R: formatReg.
	jumpIs64BitLongs := cogit JumpZero: 0.
	"For now ignore 64-bit indexability."
	jumpNotIndexable jmpTarget: cogit Label.
	jumpNotIndexable := cogit Jump: 0.

	jumpIsBytes jmpTarget:
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord R: destReg).
		cogit AndCq: objectMemory wordSize - 1 R: formatReg.
		cogit SubR: formatReg R: destReg.
	jumpBytesDone := cogit Jump: 0.

	jumpIsShorts jmpTarget:
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord - 1 R: destReg).
		cogit AndCq: 1 R: formatReg.
		cogit SubR: formatReg R: destReg.
	jumpShortsDone := cogit Jump: 0.

	jumpIs64BitLongs jmpTarget:
		(cogit LogicalShiftRightCq: 1 R: destReg).
	jump64BitLongsDone := cogit Jump: 0.

	"formatReg contains fmt, now up for grabs.
	 destReg contains numSlots, precious.
	 sourceReg must be preserved"
	jumpHasFixedFields jmpTarget:
		(cogit AndCq: objectMemory classIndexMask R: scratchReg).
	cogit MoveR: scratchReg R: formatReg.
	cogit CmpCq: ClassMethodContextCompactIndex R: scratchReg.
	jumpIsContext := cogit JumpZero: 0.
	cogit PushR: destReg.
	self genGetClassObjectOfClassIndex: formatReg into: destReg scratchReg: scratchReg.
	self genLoadSlot: InstanceSpecificationIndex sourceReg: destReg destReg: formatReg.
	self genConvertSmallIntegerToIntegerInReg: formatReg.
	cogit
		PopR: destReg;
		AndCq: objectMemory fixedFieldsOfClassFormatMask R: formatReg;
		SubR: formatReg R: destReg.

	jumpArrayDone jmpTarget:
	(jump64BitLongsDone jmpTarget:
	(jump32BitLongsDone jmpTarget:
	(jumpShortsDone jmpTarget:
	(jumpBytesDone jmpTarget:
		cogit Label)))).
	aBinaryBlock value: jumpNotIndexable value: jumpIsContext! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/26/2019 16:51'!
genJumpIsSmallIntegerValue: aRegister scratch: scratchReg
	"Generate a test for aRegister containing an integer value in the SmallInteger range, and a jump if so, answering the jump.
	 c.f. Spur32BitMemoryManager>>isIntegerValue:"
	<returnTypeC: #'AbstractInstruction *'>
	^cogit
		LogicalShiftLeftCq: 1 R: aRegister R: scratchReg;
		XorR: aRegister R: scratchReg;
		JumpGreaterOrEqual: 0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 5/30/2015 11:07'!
genJumpNotSmallInteger: aRegister
	<returnTypeC: #'AbstractInstruction *'>
	cogit TstCq: 1 R: aRegister.
	^cogit JumpZero: 0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/1/2013 06:52'!
genJumpNotSmallIntegerInScratchReg: aRegister
	<returnTypeC: #'AbstractInstruction *'>
	cogit AndCq: 1 R: aRegister.
	^cogit JumpZero: 0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/26/2019 16:54'!
genJumpNotSmallIntegerValue: aRegister scratch: scratchReg
	"Generate a test for aRegister containing an integer value outside the SmallInteger range, and a jump if so, answering the jump.
	 c.f. Spur32BitMemoryManager>>isIntegerValue:"
	<returnTypeC: #'AbstractInstruction *'>
	^cogit
		ArithmeticShiftRightCq: 1 R: aRegister R: scratchReg;
		XorR: aRegister R: scratchReg;
		JumpLess: 0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 5/29/2015 18:14'!
genJumpSmallInteger: aRegister
	<returnTypeC: #'AbstractInstruction *'>
	cogit TstCq: 1 R: aRegister.
	^cogit JumpNonZero: 0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/1/2013 06:52'!
genJumpSmallIntegerInScratchReg: aRegister
	<returnTypeC: #'AbstractInstruction *'>
	cogit AndCq: 1 R: aRegister.
	^cogit JumpNonZero: 0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:29'!
genLcInt32ToOop: value
	<option: #LowcodeVM>
	value ~= ReceiverResultReg ifTrue: [ cogit MoveR: value R: ReceiverResultReg.].
	cogit CallRT: ceSigned32BitIntegerTrampoline.

	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit ssPushRegister: ReceiverResultReg.! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:29'!
genLcOopToInt32: value
	<option: #LowcodeVM>
	value ~= ReceiverResultReg ifTrue: [ cogit MoveR: value R: ReceiverResultReg.].
	cogit CallRT: ceSigned32BitValueOfTrampoline.

	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit ssPushNativeRegister: ReceiverResultReg.! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:29'!
genLcOopToUInt32: value
	<option: #LowcodeVM>
	value ~= ReceiverResultReg ifTrue: [ cogit MoveR: value R: ReceiverResultReg.].
	cogit CallRT: cePositive32BitValueOfTrampoline.

	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit ssPushNativeRegister: ReceiverResultReg.! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:29'!
genLcUInt32ToOop: value
	<option: #LowcodeVM>
	value ~= ReceiverResultReg ifTrue: [ cogit MoveR: value R: ReceiverResultReg.].
	cogit CallRT: cePositive32BitIntegerTrampoline.

	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit ssPushRegister: ReceiverResultReg.! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/10/2017 15:58'!
genPrimitiveAtPut
	| formatReg nSlotsOrBytesReg methodInBounds
	  jumpImmediate jumpBadIndex jumpImmutable jumpNotIndexablePointers jumpNotIndexableBits
	  jumpIsContext jumpIsCompiledMethod jumpIsBytes jumpIsShorts jumpHasFixedFields
	  jumpArrayOutOfBounds jumpFixedFieldsOutOfBounds
	  jumpWordsOutOfBounds jumpShortsOutOfBounds jumpBytesOutOfBounds
	  jumpShortsOutOfRange jumpWordsOutOfRange jumpBytesOutOfRange
	  jumpNonSmallIntegerValue jumpNotPointers
	  |
	"c.f. StackInterpreter>>stSizeOf: SpurMemoryManager>>lengthOf:format: fixedFieldsOf:format:length:"
	<var: #jumpIsBytes type: #'AbstractInstruction *'>
	<var: #jumpIsShorts type: #'AbstractInstruction *'>
	<var: #jumpBadIndex type: #'AbstractInstruction *'>
	<var: #jumpIsContext type: #'AbstractInstruction *'>
	<var: #jumpImmediate type: #'AbstractInstruction *'>
	<var: #methodInBounds type: #'AbstractInstruction *'>
	<var: #jumpHasFixedFields type: #'AbstractInstruction *'>
	<var: #jumpNotIndexableBits type: #'AbstractInstruction *'>
	<var: #jumpBytesOutOfRange type: #'AbstractInstruction *'>
	<var: #jumpWordsOutOfRange type: #'AbstractInstruction *'>
	<var: #jumpShortsOutOfRange type: #'AbstractInstruction *'>
	<var: #jumpArrayOutOfBounds type: #'AbstractInstruction *'>
	<var: #jumpBytesOutOfBounds type: #'AbstractInstruction *'>
	<var: #jumpWordsOutOfBounds type: #'AbstractInstruction *'>
	<var: #jumpShortsOutOfBounds type: #'AbstractInstruction *'>
	<var: #jumpNotIndexablePointers type: #'AbstractInstruction *'>

	nSlotsOrBytesReg := ClassReg.

	cogit genLoadArgAtDepth: 1 into: Arg0Reg.
	cogit genLoadArgAtDepth: 0 into: Arg1Reg.
	jumpImmediate := self genJumpImmediate: ReceiverResultReg.
	jumpBadIndex := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	self genConvertSmallIntegerToIntegerInReg: Arg0Reg.
	cogit SubCq: 1 R: Arg0Reg. "1-rel => 0-rel"

	"formatReg := self formatOf: ReceiverResultReg"
	self cppIf: IMMUTABILITY
		ifTrue:
		[ self genGetFormatOf: ReceiverResultReg
			into: (formatReg := SendNumArgsReg)
			leastSignificantHalfOfBaseHeaderIntoScratch: TempReg.
		jumpImmutable := self genJumpBaseHeaderImmutable: TempReg ]
		ifFalse: 
		[ self genGetFormatOf: ReceiverResultReg
			into: (formatReg := SendNumArgsReg)
			leastSignificantHalfOfBaseHeaderIntoScratch: NoReg ].

	self genGetNumSlotsOf: ReceiverResultReg into: nSlotsOrBytesReg.

	"dispatch on format in a combination of highest dynamic frequency order first and convenience.
		  0 = 0 sized objects (UndefinedObject True False et al)
		  1 = non-indexable objects with inst vars (Point et al)
		  2 = indexable objects with no inst vars (Array et al)
		  3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
		  4 = weak indexable objects with inst vars (WeakArray et al)
		  5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
		  6 unused, reserved for exotic pointer objects?
		  7 Forwarded Object, 1st field is pointer, rest of fields are ignored
		  8 unused, reserved for exotic non-pointer objects?
		  9 64-bit indexable
		10 - 11 32-bit indexable
		12 - 15 16-bit indexable
		16 - 23 byte indexable
		24 - 31 compiled method"
	cogit CmpCq: objectMemory weakArrayFormat R: formatReg.
	jumpNotPointers := cogit JumpAbove: 0.
	"optimistic store check; assume index in range (almost always is)."
	self genStoreCheckReceiverReg: ReceiverResultReg
		valueReg: Arg1Reg
		scratchReg: TempReg
		inFrame: false.

	cogit CmpCq: objectMemory arrayFormat R: formatReg.
	jumpNotIndexablePointers := cogit JumpBelow: 0.
	jumpHasFixedFields := cogit JumpNonZero: 0.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpArrayOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg0Reg.
	cogit MoveR: Arg1Reg Xwr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpHasFixedFields jmpTarget: cogit Label.
	self genGetClassIndexOfNonImm: ReceiverResultReg into: formatReg.
	cogit CmpCq: ClassMethodContextCompactIndex R: formatReg.
	jumpIsContext := cogit JumpZero: 0.
	"get # fixed fields in formatReg"
	cogit PushR: nSlotsOrBytesReg.
	self genGetClassObjectOfClassIndex: formatReg into: nSlotsOrBytesReg scratchReg: TempReg.
	self genLoadSlot: InstanceSpecificationIndex sourceReg: nSlotsOrBytesReg destReg: formatReg.
	cogit PopR: nSlotsOrBytesReg.
	self genConvertSmallIntegerToIntegerInReg: formatReg.
	cogit AndCq: objectMemory fixedFieldsOfClassFormatMask R: formatReg.
	cogit SubR: formatReg R: nSlotsOrBytesReg.
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: formatReg.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpFixedFieldsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddR: formatReg R: Arg0Reg.
	cogit MoveR: Arg1Reg Xwr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpNotPointers jmpTarget: cogit Label.
	jumpNonSmallIntegerValue := self genJumpNotSmallInteger: Arg1Reg scratchReg: TempReg.
					cogit CmpCq: objectMemory firstByteFormat R: formatReg.
	jumpIsBytes := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory firstShortFormat R: formatReg.
	jumpIsShorts := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory firstLongFormat R: formatReg.
	"For now ignore 64-bit indexability."
	jumpNotIndexableBits := cogit JumpBelow: 0.

	"fall through to double words"
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpWordsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	(cogit lastOpcode setsConditionCodesFor: JumpLess) ifFalse:
		[cogit CmpCq: 0 R: TempReg]. "N.B. FLAGS := TempReg - 0"
	jumpWordsOutOfRange := cogit JumpLess: 0.
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg0Reg.
	cogit MoveR: TempReg Xwr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpIsBytes jmpTarget:
		(cogit CmpCq: (objectMemory integerObjectOf: 255) R: Arg1Reg).
	jumpBytesOutOfRange := cogit JumpAbove: 0.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord R: nSlotsOrBytesReg.
	cogit AndCq: objectMemory wordSize - 1 R: formatReg R: TempReg.
	cogit SubR: TempReg R: nSlotsOrBytesReg.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpBytesOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit CmpCq: objectMemory firstCompiledMethodFormat R: formatReg.
	jumpIsCompiledMethod := cogit JumpAboveOrEqual: 0.
	methodInBounds :=
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	cogit AddCq: objectMemory baseHeaderSize R: Arg0Reg.
	cogit MoveR: TempReg Xbr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpIsShorts jmpTarget:
		(cogit CmpCq: (objectMemory integerObjectOf: 65535) R: Arg1Reg).
	jumpShortsOutOfRange := cogit JumpAbove: 0.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord - 1 R: nSlotsOrBytesReg.
	cogit AndCq: objectMemory wordSize / 2 - 1 R: formatReg.
	cogit SubR: formatReg R: nSlotsOrBytesReg.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpShortsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	cogit AddR: Arg0Reg R: ReceiverResultReg.
	cogit AddR: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: TempReg M16: objectMemory baseHeaderSize r: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	"Now check that the index is beyond the method's literals..."
	jumpIsCompiledMethod jmpTarget: cogit Label.
	self getLiteralCountOf: ReceiverResultReg plusOne: true inBytes: true into: nSlotsOrBytesReg scratch: TempReg.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	cogit JumpBelow: methodInBounds.

	jumpIsContext jmpTarget: 
	(jumpNotIndexableBits jmpTarget:
	(jumpBytesOutOfRange jmpTarget:
	(jumpWordsOutOfRange jmpTarget:
	(jumpShortsOutOfRange jmpTarget:
	(jumpIsCompiledMethod jmpTarget:
	(jumpArrayOutOfBounds jmpTarget:
	(jumpBytesOutOfBounds jmpTarget:
	(jumpShortsOutOfBounds jmpTarget:
	(jumpWordsOutOfBounds jmpTarget:
	(jumpNotIndexablePointers jmpTarget:
	(jumpNonSmallIntegerValue jmpTarget:
	(jumpFixedFieldsOutOfBounds jmpTarget: cogit Label)))))))))))).
	
	self cppIf: IMMUTABILITY
		ifTrue: [jumpImmutable jmpTarget: jumpIsContext getJmpTarget].

	cogit AddCq: 1 R: Arg0Reg. "0-rel => 1-rel"
	self genConvertIntegerToSmallIntegerInReg: Arg0Reg.

	jumpBadIndex jmpTarget: (jumpImmediate jmpTarget: cogit Label).

	^0 "Can't be complete because of contexts."! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/1/2021 20:00'!
genPrimitiveAtPutSigned: signedVersion
	"Generate the code for primitives 61 & 165, at:put:/basicAt:put: & integerAt:put:.  If signedVersion is true
	 then generate signed accesses to the bits classes (a la 164 & 165).  If signedVersion is false,
	 generate unsigned accesses (a la 60, 61, 63 & 64)."
	| formatReg nSlotsOrBytesReg methodInBounds
	  jumpImmediate jumpBadIndex jumpImmutable jumpNotIndexablePointers jumpNotIndexableBits
	  jumpIsContext jumpIsCompiledMethod jumpIsBytes jumpIsShorts jumpHasFixedFields
	  jumpArrayOutOfBounds jumpFixedFieldsOutOfBounds
	  jumpWordsOutOfBounds jumpShortsOutOfBounds jumpBytesOutOfBounds
	  jumpShortsOutOfRange jumpWordsOutOfRange jumpBytesOutOfRange
	  jumpNonSmallIntegerValue jumpNotPointers
	  |
	"c.f. StackInterpreter>>stSizeOf: SpurMemoryManager>>lengthOf:format: fixedFieldsOf:format:length:"

	nSlotsOrBytesReg := ClassReg.

	cogit genLoadArgAtDepth: 1 into: Arg0Reg.
	cogit genLoadArgAtDepth: 0 into: Arg1Reg.
	jumpImmediate := self genJumpImmediate: ReceiverResultReg.
	jumpBadIndex := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	self genConvertSmallIntegerToIntegerInReg: Arg0Reg.
	cogit SubCq: 1 R: Arg0Reg. "1-rel => 0-rel"

	"formatReg := self formatOf: ReceiverResultReg"
	self cppIf: IMMUTABILITY
		ifTrue:
		[ self genGetFormatOf: ReceiverResultReg
			into: (formatReg := SendNumArgsReg)
			leastSignificantHalfOfBaseHeaderIntoScratch: TempReg.
		jumpImmutable := self genJumpBaseHeaderImmutable: TempReg ]
		ifFalse: 
		[ self genGetFormatOf: ReceiverResultReg
			into: (formatReg := SendNumArgsReg)
			leastSignificantHalfOfBaseHeaderIntoScratch: NoReg ].

	self genGetNumSlotsOf: ReceiverResultReg into: nSlotsOrBytesReg.

	"dispatch on format in a combination of highest dynamic frequency order first and convenience.
		  0 = 0 sized objects (UndefinedObject True False et al)
		  1 = non-indexable objects with inst vars (Point et al)
		  2 = indexable objects with no inst vars (Array et al)
		  3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
		  4 = weak indexable objects with inst vars (WeakArray et al)
		  5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
		  6 unused, reserved for exotic pointer objects?
		  7 Forwarded Object, 1st field is pointer, rest of fields are ignored
		  8 unused, reserved for exotic non-pointer objects?
		  9 64-bit indexable
		10 - 11 32-bit indexable
		12 - 15 16-bit indexable
		16 - 23 byte indexable
		24 - 31 compiled method"
	cogit CmpCq: objectMemory weakArrayFormat R: formatReg.
	jumpNotPointers := cogit JumpAbove: 0.
	"optimistic store check; assume index in range (almost always is)."
	self genStoreCheckReceiverReg: ReceiverResultReg
		valueReg: Arg1Reg
		scratchReg: TempReg
		inFrame: false.

	cogit CmpCq: objectMemory arrayFormat R: formatReg.
	jumpNotIndexablePointers := cogit JumpBelow: 0.
	jumpHasFixedFields := cogit JumpNonZero: 0.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpArrayOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg0Reg.
	cogit MoveR: Arg1Reg Xwr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpHasFixedFields jmpTarget: cogit Label.
	self genGetClassIndexOfNonImm: ReceiverResultReg into: formatReg.
	cogit CmpCq: ClassMethodContextCompactIndex R: formatReg.
	jumpIsContext := cogit JumpZero: 0.
	"get # fixed fields in formatReg"
	cogit PushR: nSlotsOrBytesReg.
	self genGetClassObjectOfClassIndex: formatReg into: nSlotsOrBytesReg scratchReg: TempReg.
	self genLoadSlot: InstanceSpecificationIndex sourceReg: nSlotsOrBytesReg destReg: formatReg.
	cogit PopR: nSlotsOrBytesReg.
	self genConvertSmallIntegerToIntegerInReg: formatReg.
	cogit AndCq: objectMemory fixedFieldsOfClassFormatMask R: formatReg.
	cogit SubR: formatReg R: nSlotsOrBytesReg.
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: formatReg.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpFixedFieldsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddR: formatReg R: Arg0Reg.
	cogit MoveR: Arg1Reg Xwr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpNotPointers jmpTarget: cogit Label.
	jumpNonSmallIntegerValue := self genJumpNotSmallInteger: Arg1Reg scratchReg: TempReg.
					cogit CmpCq: objectMemory firstByteFormat R: formatReg.
	jumpIsBytes := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory firstShortFormat R: formatReg.
	jumpIsShorts := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory firstLongFormat R: formatReg.
	"For now ignore 64-bit indexability."
	jumpNotIndexableBits := cogit JumpBelow: 0.

	"fall through to words"
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpWordsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	signedVersion ifFalse:
		[(cogit lastOpcode setsConditionCodesFor: JumpLess) ifFalse:
			[cogit CmpCq: 0 R: TempReg]. "N.B. FLAGS := TempReg - 0"
		jumpWordsOutOfRange := cogit JumpLess: 0].
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg0Reg.
	cogit MoveR: TempReg Xwr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	signedVersion
		ifTrue:
			[jumpIsBytes jmpTarget:
			 (cogit ArithmeticShiftRightCq: 7 + objectMemory numSmallIntegerTagBits R: Arg1Reg R: TempReg). "Maps in range to -1,0".
			 cogit AddCq: 1 R: TempReg. "Maps in range to 0,1"
			 cogit CmpCq: 1 R: TempReg]
		ifFalse:
			[jumpIsBytes jmpTarget:
			 (cogit CmpCq: (objectMemory integerObjectOf: 255) R: Arg1Reg)].
	jumpBytesOutOfRange := cogit JumpAbove: 0.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord R: nSlotsOrBytesReg.
	cogit AndCq: objectMemory wordSize - 1 R: formatReg R: TempReg.
	cogit SubR: TempReg R: nSlotsOrBytesReg.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpBytesOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit CmpCq: objectMemory firstCompiledMethodFormat R: formatReg.
	jumpIsCompiledMethod := cogit JumpAboveOrEqual: 0.
	methodInBounds :=
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	cogit AddCq: objectMemory baseHeaderSize R: Arg0Reg.
	cogit MoveR: TempReg Xbr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	signedVersion
		ifTrue:
			[jumpIsShorts jmpTarget:
			 (cogit ArithmeticShiftRightCq: 15 + objectMemory numSmallIntegerTagBits R: Arg1Reg R: TempReg). "Maps in range to -1,0".
			 cogit AddCq: 1 R: TempReg. "Maps in range to 0,1"
			 cogit CmpCq: 1 R: TempReg]
		ifFalse:
			[jumpIsShorts jmpTarget:
			 (cogit CmpCq: (objectMemory integerObjectOf: 65535) R: Arg1Reg)].
	jumpShortsOutOfRange := cogit JumpAbove: 0.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord - 1 R: nSlotsOrBytesReg.
	cogit AndCq: objectMemory wordSize / 2 - 1 R: formatReg.
	cogit SubR: formatReg R: nSlotsOrBytesReg.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpShortsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	cogit AddR: Arg0Reg R: ReceiverResultReg.
	cogit AddR: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: TempReg M16: objectMemory baseHeaderSize r: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	"Now check that the index is beyond the method's literals..."
	jumpIsCompiledMethod jmpTarget: cogit Label.
	self getLiteralCountOf: ReceiverResultReg plusOne: true inBytes: true into: nSlotsOrBytesReg scratch: TempReg.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	cogit JumpBelow: methodInBounds.

	jumpIsContext jmpTarget: 
	(jumpNotIndexableBits jmpTarget:
	(jumpBytesOutOfRange jmpTarget:
	(jumpShortsOutOfRange jmpTarget:
	(jumpIsCompiledMethod jmpTarget:
	(jumpArrayOutOfBounds jmpTarget:
	(jumpBytesOutOfBounds jmpTarget:
	(jumpShortsOutOfBounds jmpTarget:
	(jumpWordsOutOfBounds jmpTarget:
	(jumpNotIndexablePointers jmpTarget:
	(jumpNonSmallIntegerValue jmpTarget:
	(jumpFixedFieldsOutOfBounds jmpTarget: cogit Label))))))))))).
	
	signedVersion ifFalse:
		[jumpWordsOutOfRange jmpTarget: jumpIsContext getJmpTarget].
	self cppIf: IMMUTABILITY
		ifTrue: [jumpImmutable jmpTarget: jumpIsContext getJmpTarget].

	cogit AddCq: 1 R: Arg0Reg. "0-rel => 1-rel"
	self genConvertIntegerToSmallIntegerInReg: Arg0Reg.

	jumpBadIndex jmpTarget: (jumpImmediate jmpTarget: cogit Label).

	^0 "Can't be complete because of contexts."! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'primitive generators' stamp: 'eem 10/11/2018 22:55'!
genPrimitiveAtSigned: signedVersion
	"Generate the code for primitives 60 & 164, at:/basicAt: & integerAt:.  If signedVersion is true
	 then generate signed accesses to the bits classes (a la 164 & 165).  If signedVersion is false,
	 generate unsigned accesses (a la 60, 61, 63 & 64)."
	| formatReg nSlotsOrBytesReg convertToIntAndReturn methodInBounds
	  jumpNotIndexable jumpImmediate jumpBadIndex
	  jumpIsBytes jumpIsShorts jumpIsWords jumpIsMethod
	  jumpWordTooBig jumpIsArray jumpHasFixedFields jumpIsContext
	  jumpBytesOutOfBounds jumpShortsOutOfBounds jumpWordsOutOfBounds
	  jumpArrayOutOfBounds jumpFixedFieldsOutOfBounds |
	"c.f. StackInterpreter>>stSizeOf: SpurMemoryManager>>lengthOf:format: fixedFieldsOf:format:length:"

	nSlotsOrBytesReg := ClassReg.

	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	jumpImmediate := self genJumpImmediate: ReceiverResultReg.
	cogit MoveR: Arg0Reg R: Arg1Reg.
	jumpBadIndex := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	self genConvertSmallIntegerToIntegerInReg: Arg1Reg.
	cogit SubCq: 1 R: Arg1Reg. "1-rel => 0-rel"

	"formatReg := self formatOf: ReceiverResultReg"
	self genGetFormatOf: ReceiverResultReg
		into: (formatReg := SendNumArgsReg)
		leastSignificantHalfOfBaseHeaderIntoScratch: TempReg.

	self genGetNumSlotsOf: ReceiverResultReg into: nSlotsOrBytesReg.

	"dispatch on format in a combination of highest dynamic frequency order first and convenience.
		  0 = 0 sized objects (UndefinedObject True False et al)
		  1 = non-indexable objects with inst vars (Point et al)
		  2 = indexable objects with no inst vars (Array et al)
		  3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
		  4 = weak indexable objects with inst vars (WeakArray et al)
		  5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
		  6 unused, reserved for exotic pointer objects?
		  7 Forwarded Object, 1st field is pointer, rest of fields are ignored
		  8 unused, reserved for exotic non-pointer objects?
		  9 64-bit indexable
		10 - 11 32-bit indexable
		12 - 15 16-bit indexable
		16 - 23 byte indexable
		24 - 31 compiled method"
	cogit CmpCq: objectMemory firstByteFormat R: formatReg.
	jumpIsBytes := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory arrayFormat R: formatReg.
	jumpIsArray := cogit JumpZero: 0.
	jumpNotIndexable := cogit JumpBelow: 0.
					cogit CmpCq: objectMemory weakArrayFormat R: formatReg.
	jumpHasFixedFields := cogit JumpBelowOrEqual: 0.
					cogit CmpCq: objectMemory firstShortFormat R: formatReg.
	jumpIsShorts := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory firstLongFormat R: formatReg.
	jumpIsWords := cogit JumpAboveOrEqual: 0.
	"For now ignore 64-bit indexability."
	jumpNotIndexable jmpTarget: cogit Label.
	jumpNotIndexable := cogit Jump: 0.

	jumpIsArray jmpTarget:
		(cogit CmpR: Arg1Reg R: nSlotsOrBytesReg).
	jumpArrayOutOfBounds := cogit JumpBelowOrEqual: 0.	
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg1Reg.
	cogit MoveXwr: Arg1Reg R: ReceiverResultReg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpIsBytes jmpTarget:
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord R: nSlotsOrBytesReg).
		cogit AndCq: objectMemory wordSize - 1 R: formatReg R: TempReg.
		cogit SubR: TempReg R: nSlotsOrBytesReg;
		CmpR: Arg1Reg R: nSlotsOrBytesReg.
	jumpBytesOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit CmpCq: objectMemory firstCompiledMethodFormat R: formatReg.
	jumpIsMethod := cogit JumpAboveOrEqual: 0.
	methodInBounds :=
	cogit AddCq: objectMemory baseHeaderSize R: Arg1Reg.
	(cogit backEnd byteReadsZeroExtend
	 or: [signedVersion])
		ifTrue:
			[cogit MoveXbr: Arg1Reg R: ReceiverResultReg R: ReceiverResultReg]
		ifFalse:
			[cogit "formatReg already contains a value <= 16r1f, so no need to zero it"
				MoveXbr: Arg1Reg R: ReceiverResultReg R: formatReg;
				MoveR: formatReg R: ReceiverResultReg].
	signedVersion ifTrue:
		[cogit SignExtend8R: ReceiverResultReg R: ReceiverResultReg].
	convertToIntAndReturn := cogit Label.
	self genConvertIntegerToSmallIntegerInReg: ReceiverResultReg.
	cogit genPrimReturn.

	jumpIsShorts jmpTarget:
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord - 1 R: nSlotsOrBytesReg).
		cogit AndCq: 1 R: formatReg.
		cogit SubR: formatReg R: nSlotsOrBytesReg;
		CmpR: Arg1Reg R: nSlotsOrBytesReg.
	jumpShortsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddR: Arg1Reg R: ReceiverResultReg.
	cogit AddR: Arg1Reg R: ReceiverResultReg.
	cogit MoveM16: objectMemory baseHeaderSize r: ReceiverResultReg R: ReceiverResultReg.
	signedVersion ifTrue:
		[cogit SignExtend16R: ReceiverResultReg R: ReceiverResultReg].
	cogit Jump: convertToIntAndReturn.

	jumpIsWords jmpTarget:
		(cogit CmpR: Arg1Reg R: nSlotsOrBytesReg).
	jumpWordsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg1Reg.
	cogit MoveXwr: Arg1Reg R: ReceiverResultReg R: TempReg.
	jumpWordTooBig := self jumpNotSmallIntegerUnsignedValueInRegister: TempReg.
	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit Jump: convertToIntAndReturn.

	jumpHasFixedFields jmpTarget:
		(cogit AndCq: objectMemory classIndexMask R: TempReg).
	cogit MoveR: TempReg R: formatReg.
	cogit CmpCq: ClassMethodContextCompactIndex R: TempReg.
	jumpIsContext := cogit JumpZero: 0.
	cogit PushR: nSlotsOrBytesReg.
	self genGetClassObjectOfClassIndex: formatReg into: nSlotsOrBytesReg scratchReg: TempReg.
	self genLoadSlot: InstanceSpecificationIndex sourceReg: nSlotsOrBytesReg destReg: formatReg.
	cogit PopR: nSlotsOrBytesReg.
	self genConvertSmallIntegerToIntegerInReg: formatReg.
	cogit
		AndCq: objectMemory fixedFieldsOfClassFormatMask R: formatReg;
		SubR: formatReg R: nSlotsOrBytesReg;
		CmpR: Arg1Reg R: nSlotsOrBytesReg.
	jumpFixedFieldsOutOfBounds := cogit JumpBelowOrEqual: 0.
	"index is (formatReg (fixed fields) + Arg1Reg (0-rel index)) * wordSize + baseHeaderSize"
	cogit AddR: formatReg R: Arg1Reg.
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg1Reg.
	cogit MoveXwr: Arg1Reg R: ReceiverResultReg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpIsMethod jmpTarget: cogit Label.
	"Now check that the index is beyond the method's literals..."
	self getLiteralCountOf: ReceiverResultReg plusOne: true inBytes: true into: nSlotsOrBytesReg scratch: TempReg.
	cogit CmpR: Arg1Reg R: nSlotsOrBytesReg.
	cogit JumpBelow: methodInBounds.

	jumpWordTooBig jmpTarget:
	(jumpFixedFieldsOutOfBounds jmpTarget:
	(jumpArrayOutOfBounds jmpTarget:
	(jumpBytesOutOfBounds jmpTarget:
	(jumpShortsOutOfBounds jmpTarget:
	(jumpWordsOutOfBounds jmpTarget:
	(jumpNotIndexable jmpTarget:
	(jumpIsContext jmpTarget:
	(jumpBadIndex jmpTarget:
	(jumpImmediate jmpTarget: cogit Label))))))))).

	^0 "Can't be complete because of contexts."! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'primitive generators' stamp: 'eem 5/4/2017 11:46'!
genPrimitiveIdentityHash
	"Arguably we should fail for immediates, but so far no one has complained, so..."
	| jumpImm jumpSI jumpNotSet ret |
	<var: #jumpSI type: #'AbstractInstruction *'>
	<var: #jumpImm type: #'AbstractInstruction *'>
	<var: #jumpNotSet type: #'AbstractInstruction *'>
	jumpImm := self genJumpImmediate: ReceiverResultReg.
	self genGetHashFieldNonImmOf: ReceiverResultReg asSmallIntegerInto: TempReg.
	cogit CmpCq: ConstZero R: TempReg.
	jumpNotSet := cogit JumpZero: 0.
	cogit MoveR: TempReg R: ReceiverResultReg.
	ret := cogit genPrimReturn.
	jumpImm jmpTarget: cogit Label.
	jumpSI := self genJumpSmallInteger: ReceiverResultReg.
	jumpSI jmpTarget: ret.
	self genConvertCharacterToSmallIntegerInReg: ReceiverResultReg.
	cogit Jump: ret.
	jumpNotSet jmpTarget: cogit Label.
	"For non-behaviors we can call ceNewHash to assign a fresh hash.
	 For behaviors we must fail into the interpreter primitive to enter the receiver into the class table."
	cogit getPrimitiveIndex = 75 ifFalse:
		[^0].
	cogit backEnd saveAndRestoreLinkRegAround:
		[cogit CallRT: ceNewHashTrampoline].
	cogit genPrimReturn.
	^UnfailingPrimitive! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'primitive generators' stamp: 'eem 2/28/2016 11:35'!
genPrimitiveImmediateAsInteger
	":Assume the receiuver is never a SmallInteger.  One would use ^self for that."
	self genConvertCharacterToSmallIntegerInReg: ReceiverResultReg.
	cogit genPrimReturn.
	^UnfailingPrimitive! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 11:40'!
genPrimitiveMirrorNew
	"Implement 1-arg (instantiateFixedClass:) primitiveNew for convenient cases:
	- the class argument has a hash
	- the class argument is fixed size (excluding ephemerons to save instructions & miniscule time)
	- single word header/num slots < numSlotsMask
	- the result fits in eden (actually below scavengeThreshold)"

	<option: #NewspeakVM>
	| halfHeaderReg fillReg instSpecReg byteSizeReg
	  jumpImmediate jumpUnhashed jumpNotFixedPointers jumpTooSmall jumpBadFormat
	  jumpNoSpace jumpTooBig jumpHasSlots jumpVariableOrEphemeron
	  fillLoop skip |
	<var: 'skip' type: #'AbstractInstruction *'>
	<var: 'fillLoop' type: #'AbstractInstruction *'>
	<var: 'jumpTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpHasSlots' type: #'AbstractInstruction *'>
	<var: 'jumpNoSpace' type: #'AbstractInstruction *'>
	<var: 'jumpTooSmall' type: #'AbstractInstruction *'>
	<var: 'jumpUnhashed' type: #'AbstractInstruction *'>
	<var: 'jumpImmediate' type: #'AbstractInstruction *'>
	<var: 'jumpBadFormat' type: #'AbstractInstruction *'>
	<var: 'jumpNotFixedPointers' type: #'AbstractInstruction *'>
	<var: 'jumpVariableOrEphemeron' type: #'AbstractInstruction *'>

	self assert: cogit methodNumArgs = 1.
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.

	"half header will contain 1st half of header (classIndex/class's hash & format),
	 then 2nd half of header (numSlots/fixed size) and finally fill value (nilObject)."
	halfHeaderReg := fillReg := SendNumArgsReg.
	"inst spec will hold class's instance specification, then byte size and finally end of new object."
	instSpecReg := byteSizeReg := ClassReg.

	"get freeStart as early as possible so as not to wait later..."
	cogit MoveAw: objectMemory freeStartAddress R: Arg1Reg.

	"validate class arg; sigh, this mirror crap hobbles unfairly; there is a better way with selector namespaces..."
	jumpImmediate := self genJumpImmediate: Arg0Reg.

	"Is the class arg pointers with at least 3 fields?"
	self genGetFormatOf: Arg0Reg
		into: TempReg
		leastSignificantHalfOfBaseHeaderIntoScratch: NoReg.
	cogit CmpCq: objectMemory nonIndexablePointerFormat R: TempReg.
	jumpNotFixedPointers := cogit JumpNonZero: 0.
	
	self genGetRawSlotSizeOfNonImm: Arg0Reg into: TempReg.
	cogit CmpCq: InstanceSpecificationIndex + 1 R: TempReg.
	jumpTooSmall := cogit JumpLess: 0.

	"get class's hash & fail if 0"
	self genGetHashFieldNonImmOf: Arg0Reg into: halfHeaderReg.
	jumpUnhashed := cogit JumpZero: 0.

	"get class's format inst var for both inst spec (format field) and num fixed fields"
	self genLoadSlot: InstanceSpecificationIndex sourceReg: Arg0Reg destReg: instSpecReg.
	jumpBadFormat := self genJumpNotSmallInteger: instSpecReg scratchReg: TempReg.
	self genConvertSmallIntegerToIntegerInReg: instSpecReg.
	cogit MoveR: instSpecReg R: TempReg.
	cogit LogicalShiftRightCq: objectMemory fixedFieldsFieldWidth R: TempReg.
	cogit AndCq: objectMemory formatMask R: TempReg.
	cogit AndCq: objectMemory fixedFieldsOfClassFormatMask R: instSpecReg.
	"fail if not fixed or if ephemeron (rare beasts so save the cycles)"
	cogit CmpCq: objectMemory nonIndexablePointerFormat R: TempReg.
	jumpVariableOrEphemeron := cogit JumpAbove: 0.
	cogit CmpCq: objectMemory numSlotsMask R: instSpecReg.
	jumpTooBig := cogit JumpAboveOrEqual: 0.
	"Add format to classIndex/format half header; other word contains numSlots"
	cogit LogicalShiftLeftCq: objectMemory formatShift R: TempReg.
	cogit AddR: TempReg R: halfHeaderReg.
	"write half header now; it frees halfHeaderReg"
	cogit MoveR: halfHeaderReg Mw: 0 r: Arg1Reg.
	"save unrounded numSlots for header"
	cogit MoveR: instSpecReg R: halfHeaderReg.
	"compute byte size; remember 0-sized objects still need 1 slot & allocation is
	 rounded up to 8 bytes."
	cogit CmpCq: 0 R: byteSizeReg. "a.k.a. instSpecReg"
	jumpHasSlots := cogit JumpNonZero: 0.
	cogit MoveCq: objectMemory baseHeaderSize * 2 R: byteSizeReg.
	skip := cogit Jump: 0.
	"round up to allocationUnit"
	jumpHasSlots jmpTarget:
	(cogit MoveR: byteSizeReg R: TempReg).
	cogit AndCq: 1 R: TempReg.
	cogit AddR: TempReg R: byteSizeReg.
	cogit AddCq: objectMemory baseHeaderSize / objectMemory wordSize R: byteSizeReg.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord R: byteSizeReg.
	skip jmpTarget:
	"shift halfHeaderReg to put numSlots in correct place"
	(cogit LogicalShiftLeftCq: objectMemory numSlotsHalfShift R: halfHeaderReg).
	"check if allocation fits (freeSize + byteSize < scavengeThreshold); scavengeThreshold is constant."
	cogit AddR: Arg1Reg R: byteSizeReg.
	cogit CmpCq: objectMemory getScavengeThreshold R: byteSizeReg.
	jumpNoSpace := cogit JumpAboveOrEqual: 0.
	"write back new freeStart; get result. byteSizeReg holds new freeStart, the limit of the object"
	cogit MoveR: byteSizeReg Aw: objectMemory freeStartAddress.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	"write other half of header (numSlots/identityHash)"
	cogit MoveR: halfHeaderReg Mw: 4 r: Arg1Reg.
	"now fill"
	cogit LoadEffectiveAddressMw: objectMemory baseHeaderSize r: ReceiverResultReg R: Arg1Reg.
	cogit MoveCq: objectMemory nilObject R: fillReg.
	"at least two words; so can make this a [fill 2 words. reached limit?] whileFalse"
	fillLoop := 
	cogit MoveR: fillReg Mw: 0 r: Arg1Reg.
	cogit MoveR: fillReg Mw: 4 r: Arg1Reg.
	cogit AddCq: 8 R: Arg1Reg.
	cogit CmpR: Arg1Reg R: byteSizeReg.
	cogit JumpAbove: fillLoop.
	cogit genPrimReturn.

	jumpNotFixedPointers jmpTarget:
	(jumpBadFormat jmpTarget:
	(jumpTooSmall jmpTarget:
	(jumpImmediate jmpTarget:
	(jumpUnhashed jmpTarget:
	(jumpVariableOrEphemeron jmpTarget:
	(jumpTooBig jmpTarget:
	(jumpNoSpace jmpTarget: cogit Label))))))).

	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/11/2016 14:03'!
genPrimitiveMirrorNewWithArg
	"Implement instantiateVariableClass:withSize: for convenient cases:
	- the class argument has a hash
	- the class argument is variable and not compiled method
	- single word header/num slots < numSlotsMask
	- the result fits in eden
	See superclass method for dynamic frequencies of formats.
	For the moment we implement only arrayFormat, firstByteFormat & firstLongFormat"

	<option: #NewspeakVM>
	| halfHeaderReg fillReg instSpecReg byteSizeReg maxSlots
	  jumpArrayTooBig jumpByteTooBig jumpLongTooBig
	  jumpArrayFormat jumpByteFormat jumpBytePrepDone jumpLongPrepDone
	  jumpUnhashed jumpTooSmall jumpImmediate jumpNotFixedPointers
	  jumpNElementsNonInt jumpFailCuzFixed jumpNoSpace jumpHasSlots fillLoop skip |
	<var: 'skip' type: #'AbstractInstruction *'>
	<var: 'fillLoop' type: #'AbstractInstruction *'>	
	<var: 'jumpHasSlots' type: #'AbstractInstruction *'>
	<var: 'jumpNoSpace' type: #'AbstractInstruction *'>
	<var: 'jumpTooSmall' type: #'AbstractInstruction *'>
	<var: 'jumpUnhashed' type: #'AbstractInstruction *'>
	<var: 'jumpImmediate' type: #'AbstractInstruction *'>
	<var: 'jumpByteFormat' type: #'AbstractInstruction *'>
	<var: 'jumpByteTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpLongTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpArrayFormat' type: #'AbstractInstruction *'>
	<var: 'jumpArrayTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpFailCuzFixed' type: #'AbstractInstruction *'>
	<var: 'jumpBytePrepDone' type: #'AbstractInstruction *'>
	<var: 'jumpLongPrepDone' type: #'AbstractInstruction *'>
	<var: 'jumpNElementsNonInt' type: #'AbstractInstruction *'>
	<var: 'jumpNotFixedPointers' type: #'AbstractInstruction *'>

	self assert: cogit methodNumArgs = 2.
	cogit genLoadArgAtDepth: 1 into: Arg0Reg. "class arg"
	cogit genLoadArgAtDepth: 0 into: Arg1Reg. "num indexable slots arg"

	"half header will contain 1st half of header (classIndex/class's hash & format),
	 then 2nd half of header (numSlots) and finally fill value (nilObject)."
	halfHeaderReg := fillReg := SendNumArgsReg.
	"inst spec will hold class's instance specification and then byte size and finally numSlots half of header"
	instSpecReg := byteSizeReg := ClassReg.
	"The max slots we'll allocate here are those for a single header"
	maxSlots := objectMemory numSlotsMask - 1.

	"check size and fail if not a +ve integer"
	jumpNElementsNonInt := self genJumpNotSmallInteger: Arg1Reg scratchReg: TempReg.

	"Is the class arg pointers with at least 3 fields?"
	jumpImmediate := self genJumpImmediate: Arg0Reg.
	self genGetRawSlotSizeOfNonImm: Arg0Reg into: TempReg.
	cogit CmpCq: InstanceSpecificationIndex + 1 R: TempReg.
	jumpTooSmall := cogit JumpLess: 0.

	"fail if not indexable class"
	self genGetFormatOf: Arg0Reg
		into: TempReg
		leastSignificantHalfOfBaseHeaderIntoScratch: NoReg.
	cogit CmpCq: objectMemory nonIndexablePointerFormat R: TempReg.
	jumpNotFixedPointers := cogit JumpNonZero: 0.
	
	"get class's hash & fail if 0"
	self genGetHashFieldNonImmOf: Arg0Reg into: halfHeaderReg.
	jumpUnhashed := cogit JumpZero: 0.

	"The basicNew: code below (copied from genPrimitiveNewWithArg) expects class
	 in ReceiverResultReg and size in Arg0Reg.  Shuffle args to match, undoing on failure."
	cogit
		PushR: ReceiverResultReg;
		MoveR: Arg0Reg R: ReceiverResultReg;
		MoveR: Arg1Reg R: Arg0Reg.

	"get freeStart as early as possible so as not to wait later..."
	cogit MoveAw: objectMemory freeStartAddress R: Arg1Reg.
	"get class's format inst var for inst spec (format field)"
	self genLoadSlot: InstanceSpecificationIndex sourceReg: ReceiverResultReg destReg: instSpecReg.
	cogit LogicalShiftRightCq: objectMemory fixedFieldsFieldWidth + self numSmallIntegerTagBits R: instSpecReg.
	cogit AndCq: objectMemory formatMask R: instSpecReg.
	"Add format to classIndex/format half header now"
	cogit MoveR: instSpecReg R: TempReg.
	cogit LogicalShiftLeftCq: objectMemory formatShift R: TempReg.
	cogit AddR: TempReg R: halfHeaderReg.
	"get integer value of num fields in TempReg now"
	cogit MoveR: Arg0Reg R: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	"dispatch on format, failing if not variable or if compiled method"
	cogit CmpCq: objectMemory arrayFormat R: instSpecReg.
	jumpArrayFormat := cogit JumpZero: 0.
	cogit CmpCq: objectMemory firstByteFormat R: instSpecReg.
	jumpByteFormat := cogit JumpZero: 0.
	cogit CmpCq: objectMemory firstLongFormat R: instSpecReg.
	jumpFailCuzFixed := cogit JumpNonZero: 0.

	cogit CmpCq: (objectMemory integerObjectOf: maxSlots) R: Arg0Reg.
	jumpLongTooBig := cogit JumpAbove: 0.
	"save num elements/slot size to instSpecReg"
	cogit MoveR: TempReg R: instSpecReg.
	"push fill value"
	cogit PushCq: 0.
	jumpLongPrepDone := cogit Jump: 0. "go allocate"

	jumpByteFormat jmpTarget:
	(cogit CmpCq: (objectMemory integerObjectOf: maxSlots * objectMemory wordSize) R: Arg0Reg).
	jumpByteTooBig := cogit JumpAbove: 0.
	"save num elements to instSpecReg"
	cogit MoveR: TempReg R: instSpecReg.
	"compute odd bits and add into halfHeaderReg; oddBits := 4 - nElements bitAnd: 3"
	cogit MoveCq: objectMemory wordSize R: TempReg.
	cogit SubR: instSpecReg R: TempReg.
	cogit AndCq: objectMemory wordSize - 1 R: TempReg.
	cogit LogicalShiftLeftCq: objectMemory formatShift R: TempReg.
	cogit AddR: TempReg R: halfHeaderReg.
	"round up num elements to numSlots in instSpecReg"
	cogit AddCq: objectMemory wordSize - 1 R: instSpecReg.
	cogit LogicalShiftRightCq: objectMemory shiftForWord R: instSpecReg.
	"push fill value"
	cogit PushCq: 0.
	jumpBytePrepDone := cogit Jump: 0. "go allocate"

	jumpArrayFormat jmpTarget:
		(cogit CmpCq: (objectMemory integerObjectOf: maxSlots) R: Arg0Reg).
	jumpArrayTooBig := cogit JumpAbove: 0.
	"save num elements/slot size to instSpecReg"
	cogit MoveR: TempReg R: instSpecReg.
	"push fill value"
	cogit PushCw: objectMemory nilObject.
	"fall through to allocate"

	jumpBytePrepDone jmpTarget:
	(jumpLongPrepDone jmpTarget: cogit Label).

	"write half header now; it frees halfHeaderReg"
	cogit MoveR: halfHeaderReg Mw: 0 r: Arg1Reg.
	"save numSlots to halfHeaderReg"
	cogit MoveR: instSpecReg R: halfHeaderReg.
	"compute byte size; remember 0-sized objects still need 1 slot & allocation is
	 rounded up to 8 bytes."
	cogit CmpCq: 0 R: byteSizeReg. "a.k.a. instSpecReg"
	jumpHasSlots := cogit JumpNonZero: 0.
	cogit MoveCq: objectMemory baseHeaderSize * 2 R: byteSizeReg.
	skip := cogit Jump: 0.
	"round up to allocationUnit"
	jumpHasSlots jmpTarget:
	(cogit MoveR: byteSizeReg R: TempReg).
	cogit AndCq: 1 R: TempReg.
	cogit AddR: TempReg R: byteSizeReg.
	cogit AddCq: objectMemory baseHeaderSize / objectMemory wordSize R: byteSizeReg.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord R: byteSizeReg.
	skip jmpTarget:
	"shift halfHeaderReg to put numSlots in correct place"
	(cogit LogicalShiftLeftCq: objectMemory numSlotsHalfShift R: halfHeaderReg).
	"check if allocation fits"
	cogit AddR: Arg1Reg R: byteSizeReg.
	cogit CmpCq: objectMemory getScavengeThreshold R: byteSizeReg.
	jumpNoSpace := cogit JumpAboveOrEqual: 0.
	"get result, increment freeStart and write it back. Arg1Reg holds new freeStart, the limit of the object"
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit MoveR: byteSizeReg Aw: objectMemory freeStartAddress.
	"write other half of header (numSlots/0 identityHash)"
	cogit MoveR: halfHeaderReg Mw: 4 r: ReceiverResultReg.
	"now fill"
	cogit PopR: fillReg.
	cogit PopR: TempReg. "discard pushed receiver"
	cogit LoadEffectiveAddressMw: objectMemory baseHeaderSize r: ReceiverResultReg R: Arg1Reg.
	"at least two words; so can make this a [fill 2 words. reached limit?] whileFalse"
	fillLoop := 
	cogit MoveR: fillReg Mw: 0 r: Arg1Reg.
	cogit MoveR: fillReg Mw: 4 r: Arg1Reg.
	cogit AddCq: 8 R: Arg1Reg.
	cogit CmpR: Arg1Reg R: byteSizeReg.
	cogit JumpAbove: fillLoop.
	cogit genPrimReturn.

	"pop discarded fill value & fall through to failure"
	jumpNoSpace jmpTarget: (cogit PopR: TempReg).

	jumpFailCuzFixed jmpTarget:
	(jumpArrayTooBig jmpTarget:
	(jumpByteTooBig jmpTarget:
	(jumpLongTooBig jmpTarget: cogit Label))).

	"unshuffle arguments"
	cogit
		MoveR: Arg0Reg R: Arg1Reg;
		MoveR: ReceiverResultReg R: Arg0Reg;
		PopR: ReceiverResultReg.

	jumpUnhashed jmpTarget:
	(jumpImmediate jmpTarget:
	(jumpNotFixedPointers jmpTarget:
	(jumpTooSmall jmpTarget:
	(jumpNElementsNonInt jmpTarget: cogit Label)))).

	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:53'!
genPrimitiveNew
	"Implement primitiveNew for convenient cases:
	- the receiver has a hash
	- the receiver is fixed size (excluding ephemerons to save instructions & miniscule time)
	- single word header/num slots < numSlotsMask
	- the result fits in eden (actually below scavengeThreshold)"

	| halfHeaderReg fillReg instSpecReg byteSizeReg
	  jumpUnhashed jumpVariableOrEphemeron jumpNoSpace jumpTooBig jumpHasSlots
	  fillLoop skip |
	<var: 'skip' type: #'AbstractInstruction *'>
	<var: 'fillLoop' type: #'AbstractInstruction *'>
	<var: 'jumpTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpHasSlots' type: #'AbstractInstruction *'>
	<var: 'jumpNoSpace' type: #'AbstractInstruction *'>
	<var: 'jumpUnhashed' type: #'AbstractInstruction *'>
	<var: 'jumpVariableOrEphemeron' type: #'AbstractInstruction *'>

	NewspeakVM ifTrue:
		[cogit methodNumArgs = 1 ifTrue:
			[^self genPrimitiveMirrorNew]].
	cogit methodNumArgs ~= 0 ifTrue:
		[^UnimplementedPrimitive].

	"half header will contain 1st half of header (classIndex/class's hash & format),
	 then 2nd half of header (numSlots/fixed size) and finally fill value (nilObject)."
	halfHeaderReg := fillReg := SendNumArgsReg.
	"inst spec will hold class's instance specification, then byte size and finally end of new object."
	instSpecReg := byteSizeReg := ClassReg.

	"get freeStart as early as possible so as not to wait later..."
	cogit MoveAw: objectMemory freeStartAddress R: Arg1Reg.
	"get class's hash & fail if 0"
	self genGetHashFieldNonImmOf: ReceiverResultReg into: halfHeaderReg.
	jumpUnhashed := cogit JumpZero: 0.
	"get class's format inst var for both inst spec (format field) and num fixed fields"
	self genLoadSlot: InstanceSpecificationIndex sourceReg: ReceiverResultReg destReg: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	cogit MoveR: TempReg R: instSpecReg.
	cogit LogicalShiftRightCq: objectMemory fixedFieldsFieldWidth R: TempReg.
	cogit AndCq: objectMemory formatMask R: TempReg.
	cogit AndCq: objectMemory fixedFieldsOfClassFormatMask R: instSpecReg.
	"fail if not fixed or if ephemeron (rare beasts so save the cycles)"
	cogit CmpCq: objectMemory nonIndexablePointerFormat R: TempReg.
	jumpVariableOrEphemeron := cogit JumpAbove: 0.
	cogit CmpCq: objectMemory numSlotsMask R: instSpecReg.
	jumpTooBig := cogit JumpAboveOrEqual: 0.
	"Add format to classIndex/format half header; other word contains numSlots"
	cogit LogicalShiftLeftCq: objectMemory formatShift R: TempReg.
	cogit AddR: TempReg R: halfHeaderReg.
	"write half header now; it frees halfHeaderReg"
	cogit MoveR: halfHeaderReg Mw: 0 r: Arg1Reg.
	"save unrounded numSlots for header"
	cogit MoveR: instSpecReg R: halfHeaderReg.
	"compute byte size; remember 0-sized objects still need 1 slot & allocation is
	 rounded up to 8 bytes."
	cogit CmpCq: 0 R: byteSizeReg. "a.k.a. instSpecReg"
	jumpHasSlots := cogit JumpNonZero: 0.
	cogit MoveCq: objectMemory baseHeaderSize * 2 R: byteSizeReg.
	skip := cogit Jump: 0.
	"round up to allocationUnit"
	jumpHasSlots jmpTarget:
	(cogit MoveR: byteSizeReg R: TempReg).
	cogit AndCq: 1 R: TempReg.
	cogit AddR: TempReg R: byteSizeReg.
	cogit AddCq: objectMemory baseHeaderSize / objectMemory wordSize R: byteSizeReg.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord R: byteSizeReg.
	skip jmpTarget:
	"shift halfHeaderReg to put numSlots in correct place"
	(cogit LogicalShiftLeftCq: objectMemory numSlotsHalfShift R: halfHeaderReg).
	"check if allocation fits (freeSize + byteSize < scavengeThreshold); scavengeThreshold is constant."
	cogit AddR: Arg1Reg R: byteSizeReg.
	cogit CmpCq: objectMemory getScavengeThreshold R: byteSizeReg.
	jumpNoSpace := cogit JumpAboveOrEqual: 0.
	"write back new freeStart; get result. byteSizeReg holds new freeStart, the limit of the object"
	cogit MoveR: byteSizeReg Aw: objectMemory freeStartAddress.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	"write other half of header (numSlots/identityHash)"
	cogit MoveR: halfHeaderReg Mw: 4 r: Arg1Reg.
	"now fill"
	cogit LoadEffectiveAddressMw: objectMemory baseHeaderSize r: ReceiverResultReg R: Arg1Reg.
	cogit MoveCq: objectMemory nilObject R: fillReg.
	"at least two words; so can make this a [fill 2 words. reached limit?] whileFalse"
	fillLoop := 
	cogit MoveR: fillReg Mw: 0 r: Arg1Reg.
	cogit MoveR: fillReg Mw: 4 r: Arg1Reg.
	cogit AddCq: 8 R: Arg1Reg.
	cogit CmpR: Arg1Reg R: byteSizeReg.
	cogit JumpAbove: fillLoop.
	cogit genPrimReturn.

	jumpUnhashed jmpTarget:
	(jumpVariableOrEphemeron jmpTarget:
	(jumpTooBig jmpTarget:
	(jumpNoSpace jmpTarget: cogit Label))).

	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:57'!
genPrimitiveNewWithArg
	"Implement primitiveNewWithArg for convenient cases:
	- the receiver has a hash
	- the receiver is variable and not compiled method
	- single word header/num slots < numSlotsMask
	- the result fits in eden
	See superclass method for dynamic frequencies of formats.
	For the moment we implement only arrayFormat, firstByteFormat & firstLongFormat"

	| halfHeaderReg fillReg instSpecReg byteSizeReg maxSlots
	  jumpArrayTooBig jumpByteTooBig jumpLongTooBig
	  jumpArrayFormat jumpByteFormat jumpBytePrepDone jumpLongPrepDone
	  jumpUnhashed jumpNElementsNonInt jumpFailCuzFixed jumpNoSpace jumpHasSlots fillLoop skip |
	<var: 'skip' type: #'AbstractInstruction *'>
	<var: 'fillLoop' type: #'AbstractInstruction *'>	
	<var: 'jumpHasSlots' type: #'AbstractInstruction *'>
	<var: 'jumpNoSpace' type: #'AbstractInstruction *'>
	<var: 'jumpUnhashed' type: #'AbstractInstruction *'>
	<var: 'jumpByteFormat' type: #'AbstractInstruction *'>
	<var: 'jumpByteTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpLongTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpArrayFormat' type: #'AbstractInstruction *'>
	<var: 'jumpArrayTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpFailCuzFixed' type: #'AbstractInstruction *'>
	<var: 'jumpBytePrepDone' type: #'AbstractInstruction *'>
	<var: 'jumpLongPrepDone' type: #'AbstractInstruction *'>
	<var: 'jumpNElementsNonInt' type: #'AbstractInstruction *'>

	NewspeakVM ifTrue:
		[cogit methodNumArgs = 2 ifTrue:
			[^self genPrimitiveMirrorNewWithArg]].
	cogit methodNumArgs ~= 1 ifTrue:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.

	"half header will contain 1st half of header (classIndex/class's hash & format),
	 then 2nd half of header (numSlots) and finally fill value (nilObject)."
	halfHeaderReg := fillReg := SendNumArgsReg.
	"inst spec will hold class's instance specification and then byte size and finally numSlots half of header"
	instSpecReg := byteSizeReg := ClassReg.
	"The max slots we'll allocate here are those for a single header"
	maxSlots := objectMemory numSlotsMask - 1.

	"get freeStart as early as possible so as not to wait later..."
	cogit MoveAw: objectMemory freeStartAddress R: Arg1Reg.
	"get class's hash & fail if 0"
	self genGetHashFieldNonImmOf: ReceiverResultReg into: halfHeaderReg.
	jumpUnhashed := cogit JumpZero: 0.
	"get index and fail if not a +ve integer"
	jumpNElementsNonInt := self genJumpNotSmallInteger: Arg0Reg.
	"get class's format inst var for inst spec (format field)"
	self genLoadSlot: InstanceSpecificationIndex sourceReg: ReceiverResultReg destReg: instSpecReg.
	cogit LogicalShiftRightCq: objectMemory fixedFieldsFieldWidth + self numSmallIntegerTagBits R: instSpecReg.
	cogit AndCq: objectMemory formatMask R: instSpecReg.
	"Add format to classIndex/format half header now"
	cogit MoveR: instSpecReg R: TempReg.
	cogit LogicalShiftLeftCq: objectMemory formatShift R: TempReg.
	cogit AddR: TempReg R: halfHeaderReg.
	"get integer value of num fields in TempReg now"
	cogit MoveR: Arg0Reg R: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	"dispatch on format, failing if not variable or if compiled method"
	cogit CmpCq: objectMemory arrayFormat R: instSpecReg.
	jumpArrayFormat := cogit JumpZero: 0.
	cogit CmpCq: objectMemory firstByteFormat R: instSpecReg.
	jumpByteFormat := cogit JumpZero: 0.
	cogit CmpCq: objectMemory firstLongFormat R: instSpecReg.
	jumpFailCuzFixed := cogit JumpNonZero: 0.

	cogit CmpCq: (objectMemory integerObjectOf: maxSlots) R: Arg0Reg.
	jumpLongTooBig := cogit JumpAbove: 0.
	"save num elements/slot size to instSpecReg"
	cogit MoveR: TempReg R: instSpecReg.
	"push fill value"
	cogit PushCq: 0.
	jumpLongPrepDone := cogit Jump: 0. "go allocate"

	jumpByteFormat jmpTarget:
	(cogit CmpCq: (objectMemory integerObjectOf: maxSlots * objectMemory wordSize) R: Arg0Reg).
	jumpByteTooBig := cogit JumpAbove: 0.
	"save num elements to instSpecReg"
	cogit MoveR: TempReg R: instSpecReg.
	"compute odd bits and add into halfHeaderReg; oddBits := 4 - nElements bitAnd: 3"
	cogit MoveCq: objectMemory wordSize R: TempReg.
	cogit SubR: instSpecReg R: TempReg.
	cogit AndCq: objectMemory wordSize - 1 R: TempReg.
	cogit LogicalShiftLeftCq: objectMemory formatShift R: TempReg.
	cogit AddR: TempReg R: halfHeaderReg.
	"round up num elements to numSlots in instSpecReg"
	cogit AddCq: objectMemory wordSize - 1 R: instSpecReg.
	cogit LogicalShiftRightCq: objectMemory shiftForWord R: instSpecReg.
	"push fill value"
	cogit PushCq: 0.
	jumpBytePrepDone := cogit Jump: 0. "go allocate"

	jumpArrayFormat jmpTarget:
		(cogit CmpCq: (objectMemory integerObjectOf: maxSlots) R: Arg0Reg).
	jumpArrayTooBig := cogit JumpAbove: 0.
	"save num elements/slot size to instSpecReg"
	cogit MoveR: TempReg R: instSpecReg.
	"push fill value"
	cogit PushCw: objectMemory nilObject.
	"fall through to allocate"

	jumpBytePrepDone jmpTarget:
	(jumpLongPrepDone jmpTarget: cogit Label).

	"write half header now; it frees halfHeaderReg"
	cogit MoveR: halfHeaderReg Mw: 0 r: Arg1Reg.
	"save numSlots to halfHeaderReg"
	cogit MoveR: instSpecReg R: halfHeaderReg.
	"compute byte size; remember 0-sized objects still need 1 slot & allocation is
	 rounded up to 8 bytes."
	cogit CmpCq: 0 R: byteSizeReg. "a.k.a. instSpecReg"
	jumpHasSlots := cogit JumpNonZero: 0.
	cogit MoveCq: objectMemory baseHeaderSize * 2 R: byteSizeReg.
	skip := cogit Jump: 0.
	"round up to allocationUnit"
	jumpHasSlots jmpTarget:
	(cogit MoveR: byteSizeReg R: TempReg).
	cogit AndCq: 1 R: TempReg.
	cogit AddR: TempReg R: byteSizeReg.
	cogit AddCq: objectMemory baseHeaderSize / objectMemory wordSize R: byteSizeReg.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord R: byteSizeReg.
	skip jmpTarget:
	"shift halfHeaderReg to put numSlots in correct place"
	(cogit LogicalShiftLeftCq: objectMemory numSlotsHalfShift R: halfHeaderReg).
	"check if allocation fits"
	cogit AddR: Arg1Reg R: byteSizeReg.
	cogit CmpCq: objectMemory getScavengeThreshold R: byteSizeReg.
	jumpNoSpace := cogit JumpAboveOrEqual: 0.
	"get result, increment freeStart and write it back. Arg1Reg holds new freeStart, the limit of the object"
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit MoveR: byteSizeReg Aw: objectMemory freeStartAddress.
	"write other half of header (numSlots/0 identityHash)"
	cogit MoveR: halfHeaderReg Mw: 4 r: ReceiverResultReg.
	"now fill"
	cogit PopR: fillReg.
	cogit LoadEffectiveAddressMw: objectMemory baseHeaderSize r: ReceiverResultReg R: Arg1Reg.
	"at least two words; so can make this a [fill 2 words. reached limit?] whileFalse"
	fillLoop := 
	cogit MoveR: fillReg Mw: 0 r: Arg1Reg.
	cogit MoveR: fillReg Mw: 4 r: Arg1Reg.
	cogit AddCq: 8 R: Arg1Reg.
	cogit CmpR: Arg1Reg R: byteSizeReg.
	cogit JumpAbove: fillLoop.
	cogit genPrimReturn.

	"pop discarded fill value & fall through to failure"
	jumpNoSpace jmpTarget: (cogit PopR: TempReg).

	jumpUnhashed jmpTarget:
	(jumpFailCuzFixed jmpTarget:
	(jumpArrayTooBig jmpTarget:
	(jumpByteTooBig jmpTarget:
	(jumpLongTooBig jmpTarget:
	(jumpNElementsNonInt jmpTarget: cogit Label))))).

	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'primitive generators' stamp: 'eem 9/7/2016 11:48'!
genPrimitiveShallowCopy
	"Implement primitiveShallowCopy/primitiveClone for convenient cases:
	- the receiver is not a context
	- the receiver is not a compiled method
	- the result fits in eden (actually below scavengeThreshold)"

	| formatReg resultReg slotsReg ptrReg
	  jumpImmediate jumpIsMethod jumpVariable jumpTooBig jumpEmpty jumpNoSpace
	  continuance copyLoop |
	<var: #copyLoop type: #'AbstractInstruction *'>
	<var: #jumpTooBig type: #'AbstractInstruction *'>
	<var: #continuance type: #'AbstractInstruction *'>
	<var: #jumpVariable type: #'AbstractInstruction *'>
	<var: #jumpNoSpace type: #'AbstractInstruction *'>
	<var: #jumpIsMethod type: #'AbstractInstruction *'>
	<var: #jumpImmediate type: #'AbstractInstruction *'>

	jumpImmediate := self genJumpImmediate: ReceiverResultReg.
	resultReg := Arg0Reg.
	slotsReg := Arg1Reg.
	"get freeStart as early as possible so as not to wait later..."
	cogit MoveAw: objectMemory freeStartAddress R: resultReg.

	"formatReg := self formatOf: ReceiverResultReg"
	self genGetFormatOf: ReceiverResultReg
		into: (ptrReg := formatReg := SendNumArgsReg)
		leastSignificantHalfOfBaseHeaderIntoScratch: NoReg.

	cogit CmpCq: objectMemory firstCompiledMethodFormat R: formatReg.
	jumpIsMethod := cogit JumpAboveOrEqual: 0.
	cogit CmpCq: objectMemory indexablePointersFormat R: formatReg.
	jumpVariable := cogit JumpZero: 0.
	continuance := cogit Label.

	self genGetRawSlotSizeOfNonImm: ReceiverResultReg into: slotsReg.
	cogit CmpCq: objectMemory numSlotsMask R: slotsReg.
	jumpTooBig := cogit JumpZero: 0.

	cogit CmpCq: 0 R: slotsReg.
	jumpEmpty := cogit JumpZero: 0.

	"round up to allocationUnit"
	cogit
		MoveR: slotsReg R: TempReg;
		AndCq: 1 R: TempReg;
		AddR: TempReg R: slotsReg;
		AddCq: objectMemory baseHeaderSize / objectMemory wordSize R: slotsReg;
		LogicalShiftLeftCq: objectMemory shiftForWord R: slotsReg;
	"check if allocation fits (freeSize + byteSize < scavengeThreshold); scavengeThreshold is constant."
		AddR: resultReg R: slotsReg;
		CmpCq: objectMemory getScavengeThreshold R: slotsReg.
	jumpNoSpace := cogit JumpAboveOrEqual: 0.
	cogit
		MoveR: resultReg R: ptrReg;
	"write back new freeStart; get result. slotsReg holds new freeStart, the limit of the object"
		MoveR: slotsReg Aw: objectMemory freeStartAddress;
	"set up loop bounds"
		SubCq: objectMemory wordSize * 2 R: slotsReg;
	"copy header, masking off irrelevant bits"
		MoveMw: 0 r: ReceiverResultReg R: TempReg;
		AndCq: objectMemory formatMask << objectMemory formatShift + objectMemory classIndexMask R: TempReg;
		MoveR: TempReg Mw: 0 r: resultReg;
		MoveMw: objectMemory wordSize r: ReceiverResultReg R: TempReg;
		AndCq: objectMemory numSlotsMask << objectMemory numSlotsHalfShift R: TempReg;
		MoveR: TempReg Mw: objectMemory wordSize r: resultReg.
	"copy two fields at a time..."
	copyLoop := cogit Label.
	cogit
		AddCq: objectMemory wordSize * 2 R: ReceiverResultReg;
		AddCq: objectMemory wordSize * 2 R: ptrReg;
		MoveMw: 0 r: ReceiverResultReg R: TempReg;
		MoveR: TempReg Mw: 0 r: ptrReg;
		MoveMw: objectMemory wordSize r: ReceiverResultReg R: TempReg;
		MoveR: TempReg Mw: objectMemory wordSize r: ptrReg;
		CmpR: ptrReg R: slotsReg;
		JumpAbove: copyLoop;
		MoveR: resultReg R: ReceiverResultReg;
		genPrimReturn.

	"If the receiver is variable pointers, fail if its a context, otherwise continue"
	jumpVariable jmpTarget: cogit Label.
	self genGetClassIndexOfNonImm: ReceiverResultReg into: ClassReg.
	cogit
		CmpCq: ClassMethodContextCompactIndex R: ClassReg;
		JumpNonZero: continuance.

	jumpImmediate jmpTarget:
	(jumpNoSpace jmpTarget:
	(jumpIsMethod jmpTarget:
	(jumpTooBig jmpTarget:
	(jumpEmpty jmpTarget: cogit Label)))).

	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/10/2017 15:58'!
genPrimitiveStringAt
	| formatReg jumpNotIndexable jumpBadIndex done
	  jumpIsBytes jumpIsShorts jumpWordTooBig jumpWordsDone
	  jumpBytesOutOfBounds jumpShortsOutOfBounds jumpWordsOutOfBounds |
	"c.f. StackInterpreter>>stSizeOf: SpurMemoryManager>>lengthOf:format: fixedFieldsOf:format:length:"
	<var: #done type: #'AbstractInstruction *'>
	<var: #jumpIsBytes type: #'AbstractInstruction *'>
	<var: #jumpIsShorts type: #'AbstractInstruction *'>
	<var: #jumpBadIndex type: #'AbstractInstruction *'>
	<var: #jumpWordsDone type: #'AbstractInstruction *'>
	<var: #jumpWordTooBig type: #'AbstractInstruction *'>
	<var: #jumpNotIndexable type: #'AbstractInstruction *'>
	<var: #jumpBytesOutOfBounds type: #'AbstractInstruction *'>
	<var: #jumpShortsOutOfBounds type: #'AbstractInstruction *'>
	<var: #jumpWordsOutOfBounds type: #'AbstractInstruction *'>

	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	cogit MoveR: Arg0Reg R: Arg1Reg.
	jumpBadIndex := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	self genConvertSmallIntegerToIntegerInReg: Arg1Reg.
	cogit SubCq: 1 R: Arg1Reg. "1-rel => 0-rel"

	self genGetFormatOf: ReceiverResultReg
		into: (formatReg := SendNumArgsReg)
		leastSignificantHalfOfBaseHeaderIntoScratch: NoReg.

	self genGetNumSlotsOf: ReceiverResultReg into: ClassReg.

	"dispatch on format in a combination of highest dynamic frequency order first and convenience.
		  0 = 0 sized objects (UndefinedObject True False et al)
		  1 = non-indexable objects with inst vars (Point et al)
		  2 = indexable objects with no inst vars (Array et al)
		  3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
		  4 = weak indexable objects with inst vars (WeakArray et al)
		  5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
		  6 unused, reserved for exotic pointer objects?
		  7 Forwarded Object, 1st field is pointer, rest of fields are ignored
		  8 unused, reserved for exotic non-pointer objects?
		  9 64-bit indexable
		10 - 11 32-bit indexable
		12 - 15 16-bit indexable
		16 - 23 byte indexable
		24 - 31 compiled method"
	cogit CmpCq: objectMemory firstByteFormat R: formatReg.
	jumpIsBytes := cogit JumpGreaterOrEqual: 0.
					cogit CmpCq: objectMemory firstShortFormat R: formatReg.
	jumpIsShorts := cogit JumpGreaterOrEqual: 0.
					cogit CmpCq: objectMemory firstLongFormat R: formatReg.
	jumpNotIndexable := cogit JumpLess: 0.

	"fall through to double words"
	cogit CmpR: Arg1Reg R: ClassReg.
	jumpWordsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg1Reg.
	cogit MoveXwr: Arg1Reg R: ReceiverResultReg R: TempReg.
	jumpWordTooBig := self jumpNotCharacterUnsignedValueInRegister: TempReg.
	cogit MoveR: TempReg R: ReceiverResultReg.
	jumpWordsDone := cogit Jump: 0.

	jumpIsBytes jmpTarget:
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord R: ClassReg).
	cogit AndCq: objectMemory wordSize - 1 R: formatReg.
	cogit SubR: formatReg R: ClassReg.
	cogit CmpR: Arg1Reg R: ClassReg.
	jumpBytesOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddCq: objectMemory baseHeaderSize R: Arg1Reg.
	cogit MoveXbr: Arg1Reg R: ReceiverResultReg R: ReceiverResultReg.
	cogit backEnd byteReadsZeroExtend ifFalse:
			[cogit AndCq: 255 R: ReceiverResultReg].
	jumpWordsDone jmpTarget: (done := cogit Label).
	self genConvertIntegerToCharacterInReg: ReceiverResultReg.
	cogit genPrimReturn.

	jumpIsShorts jmpTarget:
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord - 1 R: ClassReg).
	cogit AndCq: objectMemory wordSize / 1 - 1 R: formatReg.
	cogit SubR: formatReg R: ClassReg.
	cogit CmpR: Arg1Reg R: ClassReg.
	jumpShortsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddR: Arg1Reg R: ReceiverResultReg.
	cogit MoveM16: objectMemory baseHeaderSize r: ReceiverResultReg R: ReceiverResultReg.
	cogit Jump: done.


	jumpWordTooBig jmpTarget:
	(jumpBytesOutOfBounds jmpTarget:
	(jumpShortsOutOfBounds jmpTarget:
	(jumpWordsOutOfBounds jmpTarget:
	(jumpNotIndexable jmpTarget: 
	(jumpBadIndex jmpTarget: cogit Label))))).

	^CompletePrimitive! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/10/2017 15:58'!
genPrimitiveStringAtPut
	| formatReg jumpBadIndex jumpBadArg jumpImmutable
	  jumpIsBytes jumpIsShorts jumpNotString jumpIsCompiledMethod
	  jumpBytesOutOfRange jumpShortsOutOfRange jumpWordsOutOfRange
	  jumpBytesOutOfBounds jumpWordsOutOfBounds jumpShortsOutOfBounds |
	"c.f. StackInterpreter>>stSizeOf: SpurMemoryManager>>lengthOf:format: fixedFieldsOf:format:length:"
	<var: #jumpBadArg type: #'AbstractInstruction *'>
	<var: #jumpIsBytes type: #'AbstractInstruction *'>
	<var: #jumpIsShorts type: #'AbstractInstruction *'>
	<var: #jumpBadIndex type: #'AbstractInstruction *'>
	<var: #jumpBytesOutOfRange type: #'AbstractInstruction *'>
	<var: #jumpShortsOutOfRange type: #'AbstractInstruction *'>
	<var: #jumpWordsOutOfRange type: #'AbstractInstruction *'>
	<var: #jumpBytesOutOfBounds type: #'AbstractInstruction *'>
	<var: #jumpShortsOutOfBounds type: #'AbstractInstruction *'>
	<var: #jumpWordsOutOfBounds type: #'AbstractInstruction *'>

	cogit genLoadArgAtDepth: 1 into: Arg0Reg.
	cogit genLoadArgAtDepth: 0 into: Arg1Reg.

	jumpBadIndex := self genJumpNotSmallInteger: Arg0Reg.
	cogit MoveR: Arg1Reg R: TempReg.
	jumpBadArg := self genJumpNotCharacterInScratchReg: TempReg.
	self genConvertSmallIntegerToIntegerInReg: Arg0Reg.
	cogit SubCq: 1 R: Arg0Reg. "1-rel => 0-rel"

	"formatReg := self formatOf: ReceiverResultReg"
	self cppIf: IMMUTABILITY
		ifTrue:
		[ self genGetFormatOf: ReceiverResultReg
			into: (formatReg := SendNumArgsReg)
			leastSignificantHalfOfBaseHeaderIntoScratch: TempReg.
		jumpImmutable := self genJumpBaseHeaderImmutable: TempReg ]
		ifFalse: 
		[ self genGetFormatOf: ReceiverResultReg
			into: (formatReg := SendNumArgsReg)
			leastSignificantHalfOfBaseHeaderIntoScratch: NoReg ].

	self genGetNumSlotsOf: ReceiverResultReg into: ClassReg.

	"dispatch on format; words and/or bytes.
		  0 to 8 = pointer objects, forwarders, reserved.
		  9 64-bit indexable
		10 - 11 32-bit indexable
		12 - 15 16-bit indexable
		16 - 23 byte indexable
		24 - 31 compiled method"
	cogit CmpCq: objectMemory firstLongFormat R: formatReg.
	jumpNotString := cogit JumpBelow: 0.
					cogit CmpCq: objectMemory firstCompiledMethodFormat R: formatReg.
	jumpIsCompiledMethod := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory firstByteFormat R: formatReg.
	jumpIsBytes := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory firstShortFormat R: formatReg.
	jumpIsShorts := cogit JumpAboveOrEqual: 0.

	"fall through to double words"
	cogit CmpCq: 0 R: Arg1Reg.
	jumpWordsOutOfRange := cogit JumpLess: 0.
	cogit CmpR: Arg0Reg R: ClassReg.
	jumpWordsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertCharacterToCodeInReg: TempReg.
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg0Reg.
	cogit MoveR: TempReg Xwr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpIsShorts jmpTarget:
		(cogit CmpCq: (objectMemory characterObjectOf: 65535) R: Arg1Reg).
	jumpShortsOutOfRange := cogit JumpAbove: 0.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord - 1 R: ClassReg.
	cogit AndCq: objectMemory wordSize / 2 - 1 R: formatReg.
	cogit SubR: formatReg R: ClassReg.
	cogit CmpR: Arg0Reg R: ClassReg.
	jumpShortsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertCharacterToCodeInReg: TempReg.
	cogit AddR: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: TempReg M16: objectMemory baseHeaderSize r: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpIsBytes jmpTarget:
		(cogit CmpCq: (objectMemory characterObjectOf: 255) R: Arg1Reg).
	jumpBytesOutOfRange := cogit JumpAbove: 0.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord R: ClassReg.
	cogit AndCq: objectMemory wordSize - 1 R: formatReg.
	cogit SubR: formatReg R: ClassReg.
	cogit CmpR: Arg0Reg R: ClassReg.
	jumpBytesOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertCharacterToCodeInReg: TempReg.
	cogit AddCq: objectMemory baseHeaderSize R: Arg0Reg.
	cogit MoveR: TempReg Xbr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpNotString jmpTarget:
	(jumpBytesOutOfRange jmpTarget:
	(jumpShortsOutOfRange jmpTarget:
	(jumpWordsOutOfRange jmpTarget:
	(jumpIsCompiledMethod jmpTarget:
	(jumpBytesOutOfBounds jmpTarget:
	(jumpShortsOutOfBounds jmpTarget:
	(jumpWordsOutOfBounds jmpTarget: cogit Label))))))).

	self cppIf: IMMUTABILITY
		ifTrue: [jumpImmutable jmpTarget: jumpNotString getJmpTarget].

	cogit AddCq: 1 R: Arg0Reg. "0-rel => 1-rel"
	self genConvertIntegerToSmallIntegerInReg: Arg0Reg.

	jumpBadArg jmpTarget:
	(jumpBadIndex jmpTarget: cogit Label).

	^CompletePrimitive! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/1/2013 17:49'!
genRemoveSmallIntegerTagsInScratchReg: scratchReg
	cogit SubCq: 1 R: scratchReg.
	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/2/2013 22:12'!
genShiftAwaySmallIntegerTagsInScratchReg: scratchReg
	cogit ArithmeticShiftRightCq: 1 R: scratchReg.
	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'initialization' stamp: 'eem 12/7/2015 16:12'!
genStoreHeader: header intoNewInstance: rcvrReg using: scratchReg
	"Generate the instructions to move the constant header into a new instance pointed to by rcvrReg."
	<inline: true>
	self flag: #endianness.
	cogit
		MoveCq: (self low32BitsOf: header) R: scratchReg;
		MoveR: scratchReg Mw: 0 r: rcvrReg;
		MoveCq: header >> 32 R: scratchReg;
		MoveR: scratchReg Mw: 4 r: rcvrReg! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'initialization' stamp: 'eem 10/1/2016 16:06'!
generateLowcodeObjectTrampolines
	<option: #LowcodeVM>
	super generateLowcodeObjectTrampolines.
	cePositive32BitIntegerTrampoline := cogit genTrampolineFor: #positive32BitIntegerFor:
												called: 'cePositive32BitIntegerTrampoline'
												arg: ReceiverResultReg
												result: TempReg.
	cePositive32BitValueOfTrampoline := cogit genTrampolineFor: #positive32BitValueOf:
												called: 'cePositive32BitValueOfTrampoline'
												arg: ReceiverResultReg
												result: TempReg.
	ceSigned32BitIntegerTrampoline := cogit genTrampolineFor: #signed32BitIntegerFor:
												called: 'ceSigned32BitIntegerTrampoline'
												arg: ReceiverResultReg
												result: TempReg.
	ceSigned32BitValueOfTrampoline := cogit genTrampolineFor: #signed32BitValueOf:
												called: 'ceSigned32BitValueOfTrampoline'
												arg: ReceiverResultReg
												result: TempReg.! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 1/23/2016 12:26'!
getLiteralCountOf: methodReg plusOne: plusOne inBytes: inBytes into: litCountReg scratch: scratchReg
	"Get the literal count of a CompiledMethod into headerReg, plus one if requested.
	  If inBytes is true, scale the count by the word size.  Deal with the possibility of
	 the method being cogged."
	self genGetMethodHeaderOf: methodReg into: litCountReg scratch: scratchReg.
	inBytes
		ifTrue:
			[cogit
				AndCq: coInterpreter alternateHeaderNumLiteralsMask << 1 R: litCountReg;
				LogicalShiftLeftCq: 1 R: litCountReg]
		ifFalse:
			[cogit
				LogicalShiftRightCq: 1 R: litCountReg;
				AndCq: coInterpreter alternateHeaderNumLiteralsMask R: litCountReg].
	plusOne ifTrue:
		[cogit AddCq: (inBytes ifTrue: [objectMemory wordSize] ifFalse: [1]) R: litCountReg].
	^0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'in-line cacheing' stamp: 'eem 01/10/2016 10:01'!
illegalClassTag
	^2! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'in-line cacheing' stamp: 'eem 8/7/2014 04:48'!
inlineCacheTagForClass: classOop
	"Answer the relevant inline cache tag for a class.
	 c.f. inlineCacheTagForInstance:"
	| hash |
	hash := objectMemory classTagForClass: classOop.
	^hash <= objectMemory tagMask
		ifTrue: [hash bitAnd: 1]
		ifFalse: [hash]! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'in-line cacheing' stamp: 'eem 7/7/2014 09:57'!
inlineCacheTagForInstance: oop
	"Answer the relevant inline cache tag for an instance.
	 c.f. getInlineCacheClassTagFrom:into: & inlineCacheTagForClass:"
	^(objectMemory isImmediate: oop)
		ifTrue: [oop bitAnd: 1] "SmallInteger => 1, Character => 0"
		ifFalse: [objectMemory classIndexOf: oop]! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'primitive generators' stamp: 'eem 10/2/2013 14:41'!
jumpNotSmallIntegerUnsignedValueInRegister: reg
	cogit CmpCq: 16r3FFFFFFF R: reg.
	^cogit JumpAbove: 0! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'initialization' stamp: 'eem 5/31/2014 17:43'!
log2BytesPerWord
	^2! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'private' stamp: 'eem 4/7/2015 16:51'!
low32BitsOf: a64BitHeader
	<inline: true>
	^self cCode: [self cCoerceSimple: a64BitHeader to: #usqInt] inSmalltalk: [a64BitHeader bitAnd: 16rffffffff]! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'garbage collection' stamp: 'eem 7/9/2021 14:48'!
markAndTraceCacheTagLiteral: literal in: cogMethodOrNil atpc: address
	"Mark and trace a literal in an inline cache preceding address in cogMethodOrNil.
	 Answer if code was modified."
	<var: #cogMethodOrNil type: #'CogMethod *'>
	<var: #address type: #usqInt>
	| objOop |
	(self couldBeObject: literal) ifFalse:
		[^false].
	self assert: (objectMemory addressCouldBeObj: literal).
	(objectMemory isForwarded: literal) ifFalse:
		[objectMemory markAndTrace: literal.
		 ^false].
	objOop := objectMemory followForwarded: literal.
	self setCodeModified.
	cogit backEnd rewriteInlineCacheTag: objOop at: address.
	self markAndTraceUpdatedLiteral: objOop in: cogMethodOrNil.
	^true! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'sista support' stamp: 'eem 7/5/2014 17:54'!
numCountersFor: theCounters
	<var: #theCounters type: #usqInt>
	| objOop |
	theCounters = 0 ifTrue:
		[^0].
	objOop := theCounters - objectMemory baseHeaderSize.
	^objectMemory numSlotsOf: objOop! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/2/2013 15:08'!
numSmallIntegerBits
	^31! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'object representation' stamp: 'eem 10/4/2013 16:48'!
numSmallIntegerTagBits
	^1! !

!CogObjectRepresentationFor32BitSpur methodsFor: 'debug support' stamp: 'eem 12/5/2014 18:52'!
validInlineCacheTag: classIndexOrTagPattern
	"The two valid tag patterns are 0 (Character) and 1 (SmallInteger)"
	<var: 'classIndexOrTagPattern' type: #usqInt>
	^classIndexOrTagPattern <= 1
	  or: [(objectMemory classAtIndex: classIndexOrTagPattern) notNil]! !

CogObjectRepresentationForSpur subclass: #CogObjectRepresentationFor64BitSpur
	instanceVariableNames: 'ceDereferenceSelectorIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogObjectRepresentationFor64BitSpur class methodsFor: 'accessing class hierarchy' stamp: 'eem 1/22/2014 15:53'!
defaultObjectMemoryClass
	"For in-image compilation.  Spur64BitCoMemoryManager is as yet undefined."
	^Smalltalk classNamed: #Spur64BitCoMemoryManager! !

!CogObjectRepresentationFor64BitSpur class methodsFor: 'accessing' stamp: 'eem 12/23/2015 05:35'!
wordSize
	^8! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'sista support' stamp: 'cb 12/11/2016 16:47'!
allImmediate: immediateMask branchIf: reg instanceOfBehaviors: arrayObj target: targetFixUp
	| tag1 tag2 |
	<var: #targetFixUp type: #'AbstractInstruction *'>
	immediateMask = objectMemory tagMask ifTrue: [(self genJumpImmediate: reg) jmpTarget: targetFixUp. ^0].
	"TempReg holds the rcvr tag"
	"In this case one immediate tag out of the three is not present in arrayObj."
	tag1 := objectMemory classTagForClass: (objectMemory fetchPointer: 0 ofObject: arrayObj).
	tag2 := objectMemory classTagForClass: (objectMemory fetchPointer: 1 ofObject: arrayObj).
	(tag1 = objectMemory smallIntegerTag or: [tag2 = objectMemory smallIntegerTag]) ifTrue:
		[(self genJumpSmallIntegerInScratchReg: TempReg) jmpTarget: targetFixUp].
	(tag1 = objectMemory characterTag or: [tag2 = objectMemory characterTag]) ifTrue:
		[(self genJumpCharacterInScratchReg: TempReg) jmpTarget: targetFixUp].
	(tag1 = objectMemory smallFloatTag or: [tag2 = objectMemory smallFloatTag]) ifTrue:
		[(self genJumpSmallFloatInScratchReg: TempReg) jmpTarget: targetFixUp].
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'sista support' stamp: 'eem 12/17/2015 09:10'!
allImmediate: immediateMask branchIf: reg notInstanceOfBehaviors: arrayObj target: targetFixUp
	| incorrectTag tag1 tag2 |
	<var: #targetFixUp type: #'AbstractInstruction *'>
	(self genJumpNotImmediate: reg) jmpTarget: targetFixUp. 
	immediateMask = objectMemory tagMask ifFalse: 
		[ "TempReg holds the rcvr tag"
		"In this case one immediate tag out of the three is not present in arrayObj. 
		We look for it, and generate a jump to the fixup if the rcvr tag matches"
		tag1 := objectMemory classTagForClass: (objectMemory fetchPointer: 0 ofObject: arrayObj).
		tag2 := objectMemory classTagForClass: (objectMemory fetchPointer: 1 ofObject: arrayObj).
		incorrectTag :=  self fetchImmediateTagOtherThanTag1: tag1 tag2: tag2.
		cogit CmpCq: incorrectTag R: TempReg.
		cogit JumpZero: targetFixUp].
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'sista support' stamp: 'eem 7/9/2014 21:21'!
allocateCounters: nCounters
	"On Spur allocate the counters on the heap as pinned objects.
	 Each counter occupies 32-bits.  The number of counters can
	 be derived from the number of slots in the obj."
	<inline: true>
	| objOop |
	objOop := objectMemory allocatePinnedSlots: nCounters + 1 // 2.
	^objOop
		ifNil: [0]
		ifNotNil: [objOop + objectMemory baseHeaderSize]! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'in-line cacheing' stamp: 'eem 1/27/2020 17:59'!
bitAndByteOffsetOfIsFullBlockBitInto: aBlock
	<inline: true>
	"This supplies the bitmask for the isFullBlock bit, and the offset of the byte containing
	 that bit in a CogMethod header to aBlock.  We don't have named variables holding this
	 offset.  The following assert tests whether the values are correct by creating a surrogate
	 on an empty ByteArray, setting the bit, and checking that the expected values are set
	 in the ByteArray."
	self cCode: [] inSmalltalk:
		[| m |
		m := ByteArray new: 16.
		CogBlockMethodSurrogate64 new
			at: 0 memory: m headerSize: 8 cogit: CogitFaker new;
			cpicHasMNUCaseOrCMIsFullBlock: true.
		self assert: m = #[0 0 0 0 0 0 0 0 0 16 0 0 0 0 0 0].
		self assert: (m at: objectMemory baseHeaderSize + 2) = 16].
	aBlock value: 16 value: objectMemory baseHeaderSize + 1 "zero-relative"! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'sista support' stamp: 'cb 12/11/2016 16:54'!
branch2CasesIf: reg instanceOfBehaviors: arrayObj target: targetFixUp
	| classObj tag1 tag2 |
	< inline: true>		
	
	"look for the 2 tags requiring jmps"
	0 to: (objectMemory numSlotsOf: arrayObj) - 1 do:
		[:i|
		 classObj := objectMemory fetchPointer: i ofObject: arrayObj.
		 (objectMemory isImmediateClass: classObj)
			ifTrue: [ tag1
				ifNil: [tag1 := objectMemory classTagForClass: classObj ]
				ifNotNil: [tag2 := objectMemory classTagForClass: classObj ] ] ].
		
	(tag1 = objectMemory smallIntegerTag or: [tag2 = objectMemory smallIntegerTag]) ifTrue:
		[(self genJumpSmallIntegerInScratchReg: TempReg) jmpTarget: targetFixUp].
	(tag1 = objectMemory characterTag or: [tag2 = objectMemory characterTag]) ifTrue:
		[(self genJumpCharacterInScratchReg: TempReg) jmpTarget: targetFixUp].
	(tag1 = objectMemory smallFloatTag or: [tag2 = objectMemory smallFloatTag]) ifTrue:
		[(self genJumpSmallFloatInScratchReg: TempReg) jmpTarget: targetFixUp].
	^ 0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'sista support' stamp: 'cb 4/16/2015 12:19'!
branch2CasesIf: reg notInstanceOfBehaviors: arrayObj target: targetFixUp
	"Only 2 immediate type allowed out of the three. Look for the third and jump to target fixup if it's the third.
	TempReg currently holds the rcvr tag and the receiver is immediate."
	
	| incorrectTag classObj tag1 tag2 |
	< inline: true>		
	
	"look for the 2 allowed tags"
	0 to: (objectMemory numSlotsOf: arrayObj) - 1 do:
		[:i|
		 classObj := objectMemory fetchPointer: i ofObject: arrayObj.
		 (objectMemory isImmediateClass: classObj)
			ifTrue: [ tag1
				ifNil: [tag1 := objectMemory classTagForClass: classObj ]
				ifNotNil: [tag2 := objectMemory classTagForClass: classObj ] ] ].
	
	incorrectTag := self fetchImmediateTagOtherThanTag1: tag1 tag2: tag2.
	cogit CmpCq: incorrectTag R: TempReg.
	cogit JumpZero: targetFixUp.
	! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'sista support' stamp: 'eem 1/20/2017 17:24'!
branchIf: reg hasImmediateTag: classIndex target: targetFixUp
	<var: #targetFixUp type: #'AbstractInstruction *'>
	| jmpImmediate |
	<inline: true>
	classIndex = objectMemory smallIntegerTag ifTrue:
		[jmpImmediate := self genJumpSmallInteger: reg].
	classIndex = objectMemory characterTag ifTrue:
		[jmpImmediate := self genJumpCharacter: reg].
	classIndex = objectMemory smallFloatTag ifTrue:
		[jmpImmediate := self genJumpSmallFloat: reg].
	jmpImmediate jmpTarget: targetFixUp! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'sista support' stamp: 'eem 1/20/2017 17:24'!
branchIf: reg hasNotImmediateTag: classIndex target: targetFixUp
	<var: #targetFixUp type: #'AbstractInstruction *'>
	| jmpImmediate |
	<inline: true>
	classIndex = objectMemory smallIntegerTag ifTrue:
		[jmpImmediate := self genJumpNotSmallInteger: reg].
	classIndex = objectMemory characterTag ifTrue:
		[jmpImmediate := self genJumpNotCharacter: reg].
	classIndex = objectMemory smallFloatTag ifTrue:
		[jmpImmediate := self genJumpNotSmallFloat: reg].
	jmpImmediate jmpTarget: targetFixUp! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'in-line cacheing' stamp: 'eem 3/14/2017 12:05'!
classForInlineCacheTag: classIndex
	^objectMemory classOrNilAtIndex: classIndex! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'sista support' stamp: 'cb 4/16/2015 12:00'!
fetchImmediateTagOtherThanTag1: tag1 tag2: tag2
	"Answers the immediate tag which is not tag1 nor tag2, usually Character."
	<inline: true>
	(tag1 = objectMemory characterTag or: [ tag2 = objectMemory characterTag ]) ifFalse: 
		[ ^ objectMemory characterTag ].
	(tag1 = objectMemory smallIntegerTag or: [ tag2 = objectMemory smallIntegerTag ]) ifFalse: 
		[ ^ objectMemory smallIntegerTag ].
	^ objectMemory smallFloatTag! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/4/2015 10:36'!
genAddSmallIntegerTagsTo: aRegister
	cogit AddCq: 1 R: aRegister.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 6/24/2018 09:31'!
genAlloc64BitPositiveIntegerValue: valueReg into: resultReg scratchReg: scratch1 scratchReg: scratch2
	<returnTypeC: #'AbstractInstruction *'>
	| allocSize newLPIHeader jumpFail |
	allocSize := objectMemory baseHeaderSize + objectMemory wordSize.
	newLPIHeader := objectMemory
							headerForSlots: 1
							format: objectMemory firstByteFormat
							classIndex: ClassLargePositiveIntegerCompactIndex.
	cogit MoveAw: objectMemory freeStartAddress R: resultReg.
	cogit LoadEffectiveAddressMw: allocSize r: resultReg R: scratch1.
	cogit CmpCq: objectMemory getScavengeThreshold R: scratch1.
	jumpFail := cogit JumpAboveOrEqual: 0.
	cogit MoveR: scratch1 Aw: objectMemory freeStartAddress.
	self genStoreHeader: newLPIHeader intoNewInstance: resultReg using: scratch1.
	cogit MoveR: valueReg Mw: objectMemory baseHeaderSize r: resultReg.
	^jumpFail! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 6/25/2018 14:44'!
genAlloc64BitSignedIntegerValue: valueReg into: resultReg scratchReg: scratch1 scratchReg: scratch2
	<returnTypeC: #'AbstractInstruction *'>
	| allocSize newLNIHeader jumpFail jumpNeg |
	allocSize := objectMemory baseHeaderSize + objectMemory wordSize.
	newLNIHeader := objectMemory
							headerForSlots: 1
							format: objectMemory firstByteFormat
							classIndex: ClassLargeNegativeIntegerCompactIndex.
	cogit MoveAw: objectMemory freeStartAddress R: resultReg.
	cogit LoadEffectiveAddressMw: allocSize r: resultReg R: scratch1.
	cogit CmpCq: objectMemory getScavengeThreshold R: scratch1.
	jumpFail := cogit JumpAboveOrEqual: 0.
	cogit MoveR: scratch1 Aw: objectMemory freeStartAddress.
	cogit MoveCq: newLNIHeader R: scratch1.
	cogit CmpCq: 0 R: valueReg.
	jumpNeg := cogit JumpLess: 0.
	"We can avoid duplicating the large constant and a jump choosing between
	 the alternatives by incrementing the single constant if positive.  The assert
	 checks that the hack works. Compact code is to be preferred because it is
	 an uncommon case; usually the value will fit in a SmallInteger."
	self assert: (objectMemory headerForSlots: 0 format: 0 classIndex: 1) = 1.
	cogit AddCq: ClassLargePositiveIntegerCompactIndex - ClassLargeNegativeIntegerCompactIndex R: scratch1.
	jumpNeg jmpTarget:
	(cogit MoveR: scratch1 Mw: 0 r: resultReg).
	"The old less compact code was
	cogit CmpCq: 0 R: valueReg.
	jumpNeg := cogit JumpLess: 0.
	cogit MoveCq: newLPIHeader R: scratch1.
	jumpJoin := cogit Jump: 0.
	jumpNeg jmpTarget:
	(cogit MoveCq: newLNIHeader R: scratch1).
	jumpJoin jmpTarget:
	(cogit MoveR: scratch1 Mw: 0 r: resultReg)."
	cogit MoveR: valueReg Mw: objectMemory baseHeaderSize r: resultReg.
	^jumpFail! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 12/9/2015 17:54'!
genAllocFloatValue: dpreg into: resultReg scratchReg: scratch1 scratchReg: scratch2
	"Override to answer a SmallFloat64 if possible."
	<returnTypeC: #'AbstractInstruction *'>
	| jumpFail jumpNotSF jumpMerge |
	<var: #jumpFail type: #'AbstractInstruction *'>
	<var: #jumpNotSF type: #'AbstractInstruction *'>
	<var: #jumpMerge type: #'AbstractInstruction *'>
	cogit MoveRd: dpreg R: resultReg.
	jumpNotSF := self genJumpNotSmallFloatValueBits: resultReg scratch: scratch1.
	self genConvertBitsToSmallFloatIn: resultReg scratch: scratch1.
	jumpMerge := cogit Jump: 0.
	jumpNotSF jmpTarget: cogit Label.
	jumpFail := super genAllocFloatValue: dpreg into: resultReg scratchReg: scratch1 scratchReg: scratch2.
	jumpMerge jmpTarget: cogit Label.
	^jumpFail! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/12/2015 11:35'!
genClearAndSetSmallIntegerTagsIn: scratchReg
	"Set the SmallInteger tag bits when the tag bits may be filled with garbage."
	cogit
		AndCq: -1 - objectMemory tagMask R: scratchReg;
		OrCq: 1 R: scratchReg.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/12/2015 11:38'!
genConvertBitsToSmallFloatIn: reg scratch: scratch
	"Convert the in-SmallFloat64-range floating point value in integer register into a tagged SmallFloat64 oop.
	 c.f. Spur64BitMemoryManager>>smallFloatObjectOf:"
	| jumpZero |
	<var: #jumpZero type: #'AbstractInstruction *'>
	cogit
		RotateLeftCq: 1 R: reg;
		CmpCq: 1 R: reg.
	jumpZero :=
	cogit JumpBelowOrEqual: 0.
	cogit
		SubCq: objectMemory smallFloatExponentOffset << (objectMemory smallFloatMantissaBits + 1) R: reg.
	jumpZero jmpTarget:
	(cogit LogicalShiftLeftCq: objectMemory numTagBits R: reg).
	cogit AddCq: objectMemory smallFloatTag R: reg.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 1/11/2016 18:04'!
genConvertCharacterToSmallIntegerInReg: reg
	"Convert the Character in reg to a SmallInteger, assuming
	 the Character's value is a valid character."
	cogit SubCq: objectMemory characterTag - objectMemory smallIntegerTag R: reg! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/26/2019 21:50'!
genConvertIntegerInReg: srcReg toSmallIntegerInReg: destReg
	cogit LogicalShiftLeftCq: objectMemory numTagBits R: srcReg R: destReg.
	cogit AddCq: 1 R: destReg.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/12/2015 11:38'!
genConvertIntegerToSmallIntegerInReg: reg
	cogit LogicalShiftLeftCq: objectMemory numTagBits R: reg.
	cogit AddCq: 1 R: reg.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 3/9/2016 18:10'!
genConvertSmallFloatToSmallFloatHashAsIntegerInReg: reg scratch: scratch
	"Convert the SmallFloat in reg to its identityHash as a SmallInteger.
	 Rotate the sign bit from bit 3 (zero-relative) to the sign bit.  
	 c.f. Spur64BitMemoryManager>>rotatedFloatBitsOf:"
	self assert: objectMemory smallFloatTag >> 1 - objectMemory smallIntegerTag = objectMemory smallIntegerTag.
	cogit
		LogicalShiftRightCq: 1 R: reg;
		AndCq: 1 << (objectMemory numTagBits - 1) R: reg R: scratch;
		SubR: scratch R: reg;
		SubCq: objectMemory smallFloatTag >> 1 - objectMemory smallIntegerTag R: reg;
		LogicalShiftLeftCq: 63 - (objectMemory numTagBits - 1) R: scratch;
		OrR: scratch R: reg.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 1/11/2016 18:03'!
genConvertSmallIntegerToCharacterInReg: reg
	"Convert the SmallInteger in reg to a Character, assuming
	 the SmallInteger's value is a valid character."
	cogit AddCq: objectMemory characterTag - objectMemory smallIntegerTag R: reg! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/12/2015 11:38'!
genConvertSmallIntegerToIntegerInReg: reg
	cogit ArithmeticShiftRightCq: objectMemory numTagBits R: reg.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/16/2016 20:28'!
genDoubleArithmetic: arithmeticOperator preOpCheck: preOpCheckOrNil
	<inline: true>
	^self genFloatArithmetic: arithmeticOperator preOpCheck: preOpCheckOrNil boxed: true! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 8/21/2019 01:08'!
genDoubleComparison: jumpOpcodeGenerator orIntegerComparison: jumpOpcode invert: invertComparison
	<inline: true>
	^self genFloatComparison: jumpOpcodeGenerator orIntegerComparison: jumpOpcode invert: invertComparison boxed: true! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'cb 4/17/2015 15:23'!
genFetchIndexRegister: indexReg from: tableObj into: destReg
	"indexReg contains the 1-relative index of an element in tableObj.
	 Since BaseHeaderSize = BytesPerOop we can use it as a
	 zero-relative index from the beginning of the object."
	self assert: indexReg ~= destReg.
	cogit genMoveConstant: tableObj R: destReg.
	cogit MoveXwr: indexReg R: destReg R: destReg.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 3/10/2016 10:47'!
genFloatArithmetic: arithmeticOperator preOpCheck: preOpCheckOrNil boxed: rcvrBoxed
	<var: #preOpCheckOrNil declareC: 'AbstractInstruction *(*preOpCheckOrNil)(int rcvrReg, int argReg)'>
	<inline: false>
	| jumpFailAlloc jumpFailCheck jumpImmediate jumpNotSmallFloat jumpNotSmallInteger jumpNotBoxedFloat doOp |
	<var: #jumpNotSmallInteger type: #'AbstractInstruction *'>
	<var: #jumpNotBoxedFloat type: #'AbstractInstruction *'>
	<var: #jumpNotSmallFloat type: #'AbstractInstruction *'>
	<var: #jumpImmediate type: #'AbstractInstruction *'>
	<var: #jumpFailCheck type: #'AbstractInstruction *'>
	<var: #jumpFailAlloc type: #'AbstractInstruction *'>
	<var: #doOp type: #'AbstractInstruction *'>
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	rcvrBoxed
		ifTrue: [self genGetDoubleValueOf: ReceiverResultReg into: DPFPReg0]
		ifFalse: [self genGetSmallFloatValueOf: ReceiverResultReg scratch: TempReg into: DPFPReg0].
	jumpNotSmallFloat := self genJumpNotSmallFloat: Arg0Reg.
	self genGetSmallFloatValueOf: Arg0Reg scratch: TempReg into: DPFPReg1.
	doOp := cogit Label.
	preOpCheckOrNil ifNotNil:
		[jumpFailCheck := cogit perform: preOpCheckOrNil with: DPFPReg0 with: DPFPReg1].
	cogit gen: arithmeticOperator operand: DPFPReg1 operand: DPFPReg0.
	jumpFailAlloc := self genAllocFloatValue: DPFPReg0 into: SendNumArgsReg scratchReg: ClassReg scratchReg: TempReg.
	cogit MoveR: SendNumArgsReg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpNotSmallFloat jmpTarget: cogit Label.
	jumpNotSmallInteger := self genJumpNotSmallInteger: Arg0Reg.
	cogit MoveR: Arg0Reg R: Arg1Reg.
	self genConvertSmallIntegerToIntegerInReg: Arg1Reg.
	cogit ConvertR: Arg1Reg Rd: DPFPReg1.
	cogit Jump: doOp.
	jumpNotSmallInteger jmpTarget: cogit Label.
	jumpImmediate := self genJumpImmediate: Arg0Reg.
	self genGetCompactClassIndexNonImmOf: Arg0Reg into: SendNumArgsReg.
	self genCmpClassFloatCompactIndexR: SendNumArgsReg.
	jumpNotBoxedFloat := cogit JumpNonZero: 0.
	self genGetDoubleValueOf: Arg0Reg into: DPFPReg1.
	cogit Jump: doOp.
	jumpImmediate jmpTarget:
	(jumpNotBoxedFloat jmpTarget:
	(jumpNotSmallInteger jmpTarget:
	(jumpFailAlloc jmpTarget: cogit Label))).
	preOpCheckOrNil ifNotNil:
		[jumpFailCheck jmpTarget: jumpFailAlloc getJmpTarget].
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 8/21/2019 14:07'!
genFloatComparison: jumpFPOpcodeGenerator orIntegerComparison: jumpOpcode invert: invertComparison boxed: rcvrBoxed
	<var: #jumpFPOpcodeGenerator declareC: 'AbstractInstruction *(*jumpFPOpcodeGenerator)(void *)'>
	<inline: false>
	| jumpImmediate jumpNotSmallFloat jumpNotSmallInteger jumpNotBoxedFloat jumpCond compareFloat jumpAmbiguous jumpTrue returnTrue |
	<var: #jumpNotSmallInteger type: #'AbstractInstruction *'>
	<var: #jumpNotBoxedFloat type: #'AbstractInstruction *'>
	<var: #jumpNotSmallFloat type: #'AbstractInstruction *'>
	<var: #jumpImmediate type: #'AbstractInstruction *'>
	<var: #jumpAmbiguous type: #'AbstractInstruction *'>
	<var: #jumpCond type: #'AbstractInstruction *'>
	<var: #jumpTrue type: #'AbstractInstruction *'>
	<var: #returnTrue type: #'AbstractInstruction *'>
	<var: #compareFloat type: #'AbstractInstruction *'>
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	rcvrBoxed
		ifTrue: [self genGetDoubleValueOf: ReceiverResultReg into: DPFPReg0]
		ifFalse: [self genGetSmallFloatValueOf: ReceiverResultReg scratch: TempReg into: DPFPReg0].
	jumpNotSmallFloat := self genJumpNotSmallFloat: Arg0Reg.
	self genGetSmallFloatValueOf: Arg0Reg scratch: TempReg into: DPFPReg1.
	compareFloat := invertComparison "May need to invert for NaNs"
					ifTrue: [cogit CmpRd: DPFPReg0 Rd: DPFPReg1]
					ifFalse: [cogit CmpRd: DPFPReg1 Rd: DPFPReg0].
	jumpCond := cogit perform: jumpFPOpcodeGenerator with: 0. "FP jumps are a little weird"
	cogit genMoveFalseR: ReceiverResultReg.
	cogit genPrimReturn.
	jumpCond jmpTarget: (returnTrue := cogit genMoveTrueR: ReceiverResultReg).
	cogit genPrimReturn.
	
	jumpNotSmallFloat jmpTarget: cogit Label.
	jumpNotSmallInteger := self genJumpNotSmallInteger: Arg0Reg.
	"Test for ambiguity, that is when floatRcvr == (double) intArg"
	self genConvertSmallIntegerToIntegerInReg: Arg0Reg.
	cogit ConvertR: Arg0Reg Rd: DPFPReg1.
	cogit CmpRd: DPFPReg0 Rd: DPFPReg1.
	"If floatRcvr == (double) intArg then use compareInt(intArg,(int64) floatRcvr)"
	"else use compareFloat(floatRcvr,(double) intArg)"
	jumpAmbiguous := cogit perform: #JumpFPEqual: with: 0.
	"Case of non ambiguity, use compareFloat(floatRcvr,(double) intArg)"
	cogit Jump: compareFloat.
	"Case of ambiguity, use compareInt((int64) floatRcvr, intArg)"
	jumpAmbiguous jmpTarget: (cogit ConvertRd: DPFPReg0 R: ReceiverResultReg).
	cogit CmpR: Arg0Reg R: ReceiverResultReg. "N.B. FLAGS := RRReg - Arg0Reg"
	jumpTrue := cogit genConditionalBranch: jumpOpcode operand: 0.
	cogit genMoveFalseR: ReceiverResultReg.
	cogit genPrimReturn.
	jumpTrue jmpTarget: returnTrue.

	"not a Small Float, nor Small Integer, check for Boxed Float argument"
	jumpNotSmallInteger jmpTarget: cogit Label.
	jumpImmediate := self genJumpImmediate: Arg0Reg.
	self genGetCompactClassIndexNonImmOf: Arg0Reg into: SendNumArgsReg.
	self genCmpClassFloatCompactIndexR: SendNumArgsReg.
	jumpNotBoxedFloat := cogit JumpNonZero: 0.
	self genGetDoubleValueOf: Arg0Reg into: DPFPReg1.
	cogit Jump: compareFloat.
	
	jumpImmediate jmpTarget:
	(jumpNotBoxedFloat jmpTarget: cogit Label).
	^CompletePrimitive! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/12/2015 11:38'!
genGetHashFieldNonImmOf: instReg asSmallIntegerInto: destReg
	"Fetch the instance's identity hash into destReg, encoded as a SmallInteger."
	cogit
		MoveMw: 0 r: instReg R: destReg;
		"Shift and mask the field leaving room for the SmallInteger tag."
		LogicalShiftRightCq: objectMemory identityHashFullWordShift - objectMemory numTagBits R: destReg;
		AndCq: objectMemory identityHashHalfWordMask << objectMemory numTagBits R: destReg;
		AddCq: objectMemory smallIntegerTag R: destReg.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/22/2015 16:53:59'!
genGetHashFieldNonImmOf: instReg into: destReg
	"Fetch the instance's identity hash into destReg, unencoded."
	cogit
		MoveMw: 0 r: instReg R: destReg;
		LogicalShiftRightCq: 32 R: destReg;
		AndCq: objectMemory identityHashHalfWordMask R: destReg.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 9/28/2015 12:40'!
genGetInlineCacheClassTagFrom: sourceReg into: destReg forEntry: forEntry
	"Extract the inline cache tag for the object in sourceReg into destReg. The inline
	 cache tag for a given object is the value loaded in inline caches to distinguish
	 objects of different classes.  In Spur this is either the tags for immediates, or
	 the receiver's classIndex.  Answer the label for the start of the sequence."
	| jumpImm entryLabel |
	<var: #jumpImm type: #'AbstractInstruction *'>
	<var: #entryLabel type: #'AbstractInstruction *'>
	forEntry ifTrue:
		[cogit AlignmentNops: objectMemory wordSize].
	entryLabel := cogit Label.
	cogit AndCq: objectMemory tagMask R: sourceReg R: destReg.
	jumpImm := cogit JumpNonZero: 0.
	"Get least significant half of header word in destReg"
	self flag: #endianness.
	cogit MoveMw: 0 r: sourceReg R: destReg.
	cogit AndCq: objectMemory classIndexMask R: destReg.
	jumpImm jmpTarget: cogit Label.
	^entryLabel! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/26/2016 11:35'!
genGetNumBytesOf: srcReg into: destReg
	"Get the size in byte-sized slots of the object in srcReg into destReg.
	 srcReg may equal destReg.
	 destReg <- numSlots << self shiftForWord - (fmt bitAnd: 7).
	 Assumes the object in srcReg has a byte format, i.e. 16 to 23 or 24 to 31 "
	<var: #jmp type: #'AbstractInstruction *'>
	| jmp |
	self genGetRawSlotSizeOfNonImm: srcReg into: destReg.
	cogit CmpCq: objectMemory numSlotsMask R: destReg.
	jmp := cogit JumpLess: 0.
	self genGetOverflowSlotsOf: srcReg into: destReg.
	jmp jmpTarget: (cogit LogicalShiftLeftCq: objectMemory shiftForWord R: destReg). 
	self genGetBits: 7 ofFormatByteOf: srcReg into: TempReg.
	"Now: fmt bitAnd: 3 in TempReg"
	cogit SubR: TempReg R: destReg.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 11/4/2015 12:26:50'!
genGetOverflowSlotsOf: srcReg into: destReg
	cogit
		MoveMw: objectMemory baseHeaderSize negated r: srcReg R: destReg;
		LogicalShiftLeftCq: 8 R: destReg;
		LogicalShiftRightCq: 8 R: destReg.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 10/11/2018 22:31'!
genGetSizeOf: sourceReg into: destReg formatReg: formatReg scratchReg: scratchReg abortJumpsInto: aBinaryBlock
	"Get the size of the non-immediate object in sourceReg into destReg using formatReg
	 and scratchReg as temps.  None of these registers can overlap.  Supply the jumps
	 taken if the object in sourceReg is not indexable, or if the object in sourceReg is a
	 context.. Hack: If the object has a pointer format other than 2 leave the number of
	 fixed fields in formatReg.  Used by primitiveSize, primitiveAt, and primitiveAtPut"
	<returnTypeC: #'AbstractInstruction *'>
	| jumpNotIndexable
	  jumpBytesDone jumpShortsDone jumpArrayDone jump32BitLongsDone jump64BitLongsDone
	  jumpIsBytes jumpHasFixedFields jumpIsShorts jumpIs32BitLongs jumpIsContext  |
	<inline: true>
	"c.f. StackInterpreter>>stSizeOf: SpurMemoryManager>>lengthOf:format: fixedFieldsOf:format:length:"

	"formatReg := self formatOf: sourceReg"
	self genGetFormatOf: sourceReg
		into: formatReg
		leastSignificantHalfOfBaseHeaderIntoScratch: scratchReg.

	self genGetNumSlotsOf: sourceReg into: destReg.

	"dispatch on format in a combination of highest dynamic frequency order first and convenience.
		  0 = 0 sized objects (UndefinedObject True False et al)
		  1 = non-indexable objects with inst vars (Point et al)
		  2 = indexable objects with no inst vars (Array et al)
		  3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
		  4 = weak indexable objects with inst vars (WeakArray et al)
		  5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
		  6 unused, reserved for exotic pointer objects?
		  7 Forwarded Object, 1st field is pointer, rest of fields are ignored
		  8 unused, reserved for exotic non-pointer objects?
		  9 64-bit indexable
		10 - 11 32-bit indexable
		12 - 15 16-bit indexable
		16 - 23 byte indexable
		24 - 31 compiled method"
	cogit CmpCq: objectMemory firstByteFormat R: formatReg.
	jumpIsBytes := cogit JumpGreaterOrEqual: 0.
					cogit CmpCq: objectMemory arrayFormat R: formatReg.
	jumpArrayDone := cogit JumpZero: 0.
	jumpNotIndexable := cogit JumpLess: 0.
					cogit CmpCq: objectMemory weakArrayFormat R: formatReg.
	jumpHasFixedFields := cogit JumpLessOrEqual: 0.
					cogit CmpCq: objectMemory firstShortFormat R: formatReg.
	jumpIsShorts := cogit JumpGreaterOrEqual: 0.
					cogit CmpCq: objectMemory firstLongFormat R: formatReg.
	jumpIs32BitLongs := cogit JumpGreaterOrEqual: 0.
					cogit CmpCq: objectMemory sixtyFourBitIndexableFormat R: formatReg.
	jump64BitLongsDone := cogit JumpZero: 0.
	jumpNotIndexable jmpTarget: cogit Label.
	jumpNotIndexable := cogit Jump: 0.

	jumpIsBytes jmpTarget:
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord R: destReg).
		cogit AndCq: objectMemory wordSize - 1 R: formatReg.
		cogit SubR: formatReg R: destReg.
	jumpBytesDone := cogit Jump: 0.

	jumpIsShorts jmpTarget:
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord - 1 R: destReg).
		cogit AndCq: objectMemory wordSize >> 1 - 1 R: formatReg.
		cogit SubR: formatReg R: destReg.
	jumpShortsDone := cogit Jump: 0.

	jumpIs32BitLongs jmpTarget:
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord - 2 R: destReg).
		cogit AndCq: objectMemory wordSize >> 2 - 1 R: formatReg.
		cogit SubR: formatReg R: destReg.
	jump32BitLongsDone := cogit Jump: 0.

	"formatReg contains fmt, now up for grabs.
	 destReg contains numSlots, precious.
	 sourceReg must be preserved"
	jumpHasFixedFields jmpTarget:
		(cogit AndCq: objectMemory classIndexMask R: scratchReg).
	cogit MoveR: scratchReg R: formatReg.
	cogit CmpCq: ClassMethodContextCompactIndex R: scratchReg.
	jumpIsContext := cogit JumpZero: 0.
	self genGetClassObjectOfClassIndex: formatReg into: Extra0Reg scratchReg: scratchReg.
	self genLoadSlot: InstanceSpecificationIndex sourceReg: Extra0Reg destReg: formatReg.
	self genConvertSmallIntegerToIntegerInReg: formatReg.
	cogit
		AndCq: objectMemory fixedFieldsOfClassFormatMask R: formatReg;
		SubR: formatReg R: destReg.

	jumpArrayDone jmpTarget:
	(jump64BitLongsDone jmpTarget:
	(jump32BitLongsDone jmpTarget:
	(jumpShortsDone jmpTarget:
	(jumpBytesDone jmpTarget:
		cogit Label)))).
	aBinaryBlock value: jumpNotIndexable value: jumpIsContext! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/24/2015 13:39'!
genGetSmallFloatValueOf: oopReg scratch: scratch into: dpReg
	"Convert the SmallFloat oop in ooppReg into the corresponding float value in dpReg.
	 c.f. Spur64BitMemoryManager>>smallFloatBitsOf:"
	| jumpSFZero |
	<var: 'jumpSFZero' type: #'AbstractInstruction *'>
	cogit
		MoveR: oopReg R: scratch;
		LogicalShiftRightCq: objectMemory numTagBits R: scratch;
		CmpCq: 1 R: scratch.
	jumpSFZero := cogit JumpLessOrEqual: 0.
	cogit AddCq: objectMemory smallFloatExponentOffset << (objectMemory smallFloatMantissaBits + 1) R: scratch.
	jumpSFZero jmpTarget: (cogit RotateRightCq: 1 R: scratch).
	cogit MoveR: scratch Rd: dpReg.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 3/10/2016 10:53'!
genJumpCharacter: reg
	"Generate a compare and branch to test if aRegister contains a Character."
	^cogit
		TstCq: objectMemory characterTag R: reg;
		JumpNonZero: 0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 3/10/2016 10:52'!
genJumpCharacterInScratchReg: aRegister
	^self genJumpCharacter: aRegister! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/26/2019 17:19'!
genJumpIsSmallIntegerValue: aRegister scratch: scratchReg
	"Generate a test for aRegister containing an integer value in the SmallInteger range, and a jump if so, answering the jump.
	 c.f. Spur64BitMemoryManager>>isIntegerValue:"
	<returnTypeC: #'AbstractInstruction *'>
	^cogit
		ArithmeticShiftRightCq: 63 - objectMemory numTagBits R: aRegister R: scratchReg;
		AddCq: 1 R: scratchReg;
		AndCq: 1 << (objectMemory numTagBits + 1) - 1 R: scratchReg; "sign and top numTags bits must be the same"
		CmpCq: 1 R: scratchReg;
		JumpLessOrEqual: 0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 3/10/2016 10:42'!
genJumpNotCharacter: reg
	"Generate a compare and branch to test if aRegister contains other than a Character."
	^cogit
		TstCq: objectMemory characterTag R: reg;
		JumpZero: 0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 3/10/2016 10:40'!
genJumpNotCharacterInScratchReg: reg 
	<inline: true>
	^self genJumpNotCharacter: reg! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 3/10/2016 10:43'!
genJumpNotSmallFloat: reg
	"Generate a compare and branch to test if aRegister contains other than a SmallFloat.
	 Answer the jump."
	^cogit
		TstCq: objectMemory smallFloatTag R: reg;
		JumpZero: 0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 3/10/2016 10:37'!
genJumpNotSmallFloatInScratchReg: aRegister
	<inline: true>
	^self genJumpNotSmallFloat: aRegister! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/26/2019 16:53'!
genJumpNotSmallFloatValueBits: reg scratch: exponent
	"Generate a test to check that the integer register contains a floating point value within the SmallFloat64 range,
	 and answer the jump.  c.f. Spur64BitMemoryManager>>isSmallFloatValue:"
	| jumpFail jumpTest jumpMinExponent jumpMaxExponent jumpZeroMantissa |
	<var: #jumpFail type: #'AbstractInstruction *'>
	<var: #jumpTest type: #'AbstractInstruction *'>
	<var: #jumpMinExponent type: #'AbstractInstruction *'>
	<var: #jumpMaxExponent type: #'AbstractInstruction *'>
	<var: #jumpZeroMantissa type: #'AbstractInstruction *'>
	self flag: 'if we combine the exponent range test with the conversion to tagged representation we test for a zero exponent only once. further, if we extract tags once into a scratch on the input side we test for immediates, SmallInteger and SmallFloat using the same intermediate result.  so to do is to move fp arithmetic into the object representations'.
	true
		ifTrue: [cogit
				LogicalShiftLeftCq: 1 R: reg R: exponent; "drop sign"
				LogicalShiftRightCq: objectMemory smallFloatMantissaBits + 1 R: exponent] "shift exponent down"
		ifFalse: [cogit
				LogicalShiftRightCq: objectMemory smallFloatMantissaBits R: reg R: exponent;
				AndCq: 16r7FF R: exponent].  "ieee double precision mantissa is 11 bits"
	cogit CmpCq: objectMemory smallFloatExponentOffset R: exponent.
	jumpMinExponent := cogit JumpLessOrEqual: 0.
	cogit CmpCq: 255 + objectMemory smallFloatExponentOffset R: exponent. "SmallFloat exponent is 8 bits"
	jumpMaxExponent := cogit JumpLessOrEqual: 0.
	jumpFail :=
	cogit Jump: 0.
	jumpMinExponent jmpTarget:
	(cogit TstCq: 1 << objectMemory smallFloatMantissaBits - 1 R: reg). "test mantissa bits"
	jumpZeroMantissa := cogit JumpZero: 0.
	cogit CmpCq: objectMemory smallFloatExponentOffset R: exponent.
	jumpTest :=
	cogit Jump: 0.
	jumpZeroMantissa jmpTarget:
	(cogit CmpCq: 0 R: exponent).
	jumpTest jmpTarget:
	(cogit JumpNonZero: jumpFail).
	jumpMaxExponent jmpTarget: cogit Label.
	^jumpFail! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 3/10/2016 10:43'!
genJumpNotSmallInteger: reg
	"Generate a compare and branch to test if aRegister contains other than a SmallInteger."
	^cogit
		TstCq: objectMemory smallIntegerTag R: reg;
		JumpZero: 0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 3/10/2016 10:29'!
genJumpNotSmallIntegerInScratchReg: aRegister
	<inline: true>
	^self genJumpNotSmallInteger: aRegister! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/26/2019 17:18'!
genJumpNotSmallIntegerValue: aRegister scratch: scratchReg
	"Generate a test for aRegister containing an integer value outside the SmallInteger range, and a jump if so, answering the jump.
	 c.f. Spur64BitMemoryManager>>isIntegerValue:"
	<returnTypeC: #'AbstractInstruction *'>
	^cogit
		ArithmeticShiftRightCq: 63 - objectMemory numTagBits R: aRegister R: scratchReg;
		AddCq: 1 R: scratchReg;
		AndCq: 1 << (objectMemory numTagBits + 1) - 1 R: scratchReg; "sign and top numTags bits must be the same"
		CmpCq: 1 R: scratchReg;
		JumpGreater: 0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 3/10/2016 10:43'!
genJumpSmallFloat: aRegister
	"Generate a compare and branch to test if aRegister contains a SmallFloat.
	 Answer the jump, or UnimplementedOperation if this cannot be done with
	 a single register."
	^cogit
		TstCq: objectMemory smallFloatTag R: aRegister;
		JumpNonZero: 0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 3/10/2016 10:37'!
genJumpSmallFloatInScratchReg: aRegister
	^self genJumpSmallFloat: aRegister! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 3/10/2016 10:44'!
genJumpSmallInteger: aRegister
	"Generate a compare and branch to test if aRegister contains a SmallInteger.
	 Answer the jump, or UnimplementedOperation if this cannot be done with
	 a single register."
	^cogit
		TstCq: objectMemory smallIntegerTag R: aRegister;
		JumpNonZero: 0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 3/10/2016 10:38'!
genJumpSmallIntegerInScratchReg: aRegister
	^self genJumpSmallInteger: aRegister! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'inline primitive support' stamp: 'rsf 1/17/2017 04:47'!
genLcInt32ToOop: value
	<option: #LowcodeVM>
	cogit SignExtend32R: value R: value.
	self genConvertIntegerToSmallIntegerInReg: value.
	cogit ssPushRegister: value.
	^ 0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'inline primitive support' stamp: 'rsf 1/17/2017 04:02'!
genLcOopToInt32: value
	<option: #LowcodeVM>
	self genConvertSmallIntegerToIntegerInReg: value.
	cogit ssPushNativeRegister: value.
	^ 0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'inline primitive support' stamp: 'rsf 1/17/2017 04:02'!
genLcOopToUInt32: value
	<option: #LowcodeVM>
	self genConvertSmallIntegerToIntegerInReg: value.
	cogit ssPushNativeRegister: value.
	^ 0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'inline primitive support' stamp: 'rsf 1/17/2017 04:47'!
genLcUInt32ToOop: value
	<option: #LowcodeVM>
	cogit ZeroExtend32R: value R: value.
	self genConvertIntegerToSmallIntegerInReg: value.
	cogit ssPushRegister: value.
	^ 0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/1/2021 19:52'!
genPrimitiveAtPutSigned: signedVersion
	"Generate the code for primitives 61 & 165, at:put:/basicAt:put: & integerAt:put:.  If signedVersion is true
	 then generate signed accesses to the bits classes (a la 164 & 165).  If signedVersion is false,
	 generate unsigned accesses (a la 60, 61, 63 & 64)."
	| formatReg nSlotsOrBytesReg methodInBounds
	  jumpImmediate jumpBadIndex jumpImmutable jumpNotIndexablePointers jumpNotIndexableBits
	  jumpIsContext jumpIsCompiledMethod jumpIsBytes jumpIsShorts jumpIsWords jumpHasFixedFields
	  jumpArrayOutOfBounds jumpFixedFieldsOutOfBounds
	  jumpWordsOutOfBounds jumpShortsOutOfBounds jumpBytesOutOfBounds
	  jumpShortsOutOfRange jumpWordsOutOfRange jumpBytesOutOfRange jumpDoubleWordsOutOfRange
	  jumpNonSmallIntegerValue jumpNotPointers
	  rejoin jumpNegative jump64BitsOutOfBounds jumpNot64BitIndexable jump64BitArgIsImmediate jumpNot8ByteInteger
	  |
	"c.f. StackInterpreter>>stSizeOf: SpurMemoryManager>>lengthOf:format: fixedFieldsOf:format:length:"
	nSlotsOrBytesReg := ClassReg.

	cogit genLoadArgAtDepth: 1 into: Arg0Reg.
	cogit genLoadArgAtDepth: 0 into: Arg1Reg.
	jumpImmediate := self genJumpImmediate: ReceiverResultReg.
	jumpBadIndex := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	self genConvertSmallIntegerToIntegerInReg: Arg0Reg.
	cogit SubCq: 1 R: Arg0Reg. "1-rel => 0-rel"

	"formatReg := self formatOf: ReceiverResultReg"
	self cppIf: IMMUTABILITY
		ifTrue:
		[ self genGetFormatOf: ReceiverResultReg
			into: (formatReg := SendNumArgsReg)
			leastSignificantHalfOfBaseHeaderIntoScratch: TempReg.
		jumpImmutable := self genJumpBaseHeaderImmutable: TempReg ]
		ifFalse: 
		[ self genGetFormatOf: ReceiverResultReg
			into: (formatReg := SendNumArgsReg)
			leastSignificantHalfOfBaseHeaderIntoScratch: NoReg ].

	self genGetNumSlotsOf: ReceiverResultReg into: nSlotsOrBytesReg.

	"dispatch on format in a combination of highest dynamic frequency order first and convenience.
		  0 = 0 sized objects (UndefinedObject True False et al)
		  1 = non-indexable objects with inst vars (Point et al)
		  2 = indexable objects with no inst vars (Array et al)
		  3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
		  4 = weak indexable objects with inst vars (WeakArray et al)
		  5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
		  6 unused, reserved for exotic pointer objects?
		  7 Forwarded Object, 1st field is pointer, rest of fields are ignored
		  8 unused, reserved for exotic non-pointer objects?
		  9 64-bit indexable
		10 - 11 32-bit indexable
		12 - 15 16-bit indexable
		16 - 23 byte indexable
		24 - 31 compiled method"
	cogit CmpCq: objectMemory weakArrayFormat R: formatReg.
	jumpNotPointers := cogit JumpAbove: 0.
	"optimistic store check; assume index in range (almost always is)."
	self genStoreCheckReceiverReg: ReceiverResultReg
		valueReg: Arg1Reg
		scratchReg: TempReg
		inFrame: false.

	cogit CmpCq: objectMemory arrayFormat R: formatReg.
	jumpNotIndexablePointers := cogit JumpBelow: 0.
	jumpHasFixedFields := cogit JumpNonZero: 0.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpArrayOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg0Reg.
	cogit MoveR: Arg1Reg Xwr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpHasFixedFields jmpTarget: cogit Label.
	self genGetClassIndexOfNonImm: ReceiverResultReg into: formatReg.
	cogit CmpCq: ClassMethodContextCompactIndex R: formatReg.
	jumpIsContext := cogit JumpZero: 0.
	"get # fixed fields in formatReg"
	self genGetClassObjectOfClassIndex: formatReg into: Extra0Reg scratchReg: TempReg.
	self genLoadSlot: InstanceSpecificationIndex sourceReg: Extra0Reg destReg: formatReg.
	self genConvertSmallIntegerToIntegerInReg: formatReg.
	cogit AndCq: objectMemory fixedFieldsOfClassFormatMask R: formatReg.
	cogit SubR: formatReg R: nSlotsOrBytesReg.
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: formatReg.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpFixedFieldsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddR: formatReg R: Arg0Reg.
	cogit MoveR: Arg1Reg Xwr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpNotPointers jmpTarget: cogit Label.
	jumpNonSmallIntegerValue := self genJumpNotSmallInteger: Arg1Reg scratchReg: TempReg.
					cogit CmpCq: objectMemory firstByteFormat R: formatReg.
	jumpIsBytes := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory firstShortFormat R: formatReg.
	jumpIsShorts := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory firstLongFormat R: formatReg.
	jumpIsWords := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory sixtyFourBitIndexableFormat R: formatReg.
	jumpNotIndexableBits := cogit JumpNonZero: 0.
	"fall through to 64-bit words"
	cogit MoveR: Arg1Reg R: SendNumArgsReg.
	self genConvertSmallIntegerToIntegerInReg: SendNumArgsReg.
	signedVersion ifFalse:
		[cogit CmpCq: 0 R: SendNumArgsReg.
		 jumpNegative := cogit JumpLess: 0].
	rejoin :=
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jump64BitsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg0Reg.
	cogit MoveR: SendNumArgsReg Xwr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	"Handle LargePositiveInteger (unsigned) or LargePositiveInteger and LargeNegativeInteger (signed)
	 for sixtyFourBitIndexableFormat"
	jumpNonSmallIntegerValue jmpTarget:
		(cogit CmpCq: objectMemory sixtyFourBitIndexableFormat R: formatReg).
	jumpNot64BitIndexable := cogit JumpNonZero: 0.
	jump64BitArgIsImmediate := self genJumpImmediate: Arg1Reg.
	signedVersion ifTrue: "Test top bit of 64-bit word in large integer for range check."
		[cogit MoveMw: objectMemory baseHeaderSize r: Arg1Reg R: TempReg.
		 cogit CmpCq: 0 R: TempReg.
		 jumpDoubleWordsOutOfRange := cogit JumpLess: 0].
	"Now check if the header is that of an 8 byte LargePositiveInteger"
	cogit MoveMw: 0 r: Arg1Reg R: SendNumArgsReg.
	cogit AndCq: (objectMemory
						headerForSlots: objectMemory numSlotsMask
						format: objectMemory formatMask
						classIndex: objectMemory classIndexMask)
		R: SendNumArgsReg.
	cogit CmpCq: (objectMemory
						headerForSlots: 1
						format: objectMemory firstByteFormat
						classIndex: ClassLargePositiveIntegerCompactIndex)
		R: SendNumArgsReg.
	jumpNot8ByteInteger := cogit JumpNonZero: 0.
	cogit MoveMw: objectMemory baseHeaderSize r: Arg1Reg R: SendNumArgsReg.
	cogit Jump: rejoin.
	signedVersion ifTrue:
		["Now check if the header is that of an 8 byte LargeNegativeInteger"
		 jumpNot8ByteInteger jmpTarget:
		 (cogit CmpCq: (objectMemory
							headerForSlots: 1
							format: objectMemory firstByteFormat
							classIndex: ClassLargeNegativeIntegerCompactIndex)
				R: SendNumArgsReg).
		 jumpNot8ByteInteger := cogit JumpNonZero: 0.
		 cogit MoveMw: objectMemory baseHeaderSize r: Arg1Reg R: TempReg.
		 cogit MoveCq: 0 R: SendNumArgsReg.
		 cogit SubR: TempReg R: SendNumArgsReg.
		 cogit Jump: rejoin].

	signedVersion
		ifTrue:
			[jumpIsWords jmpTarget:
			 (cogit ArithmeticShiftRightCq: 31 + objectMemory numTagBits R: Arg1Reg R: TempReg). "Maps in range to -1,0".
			 cogit AddCq: 1 R: TempReg. "Maps in range to 0,1"
			 cogit CmpCq: 1 R: TempReg]
		ifFalse:
			[jumpIsWords jmpTarget:
			 (cogit CmpCq: (objectMemory integerObjectOf: 16rFFFFFFFF) R: Arg1Reg)].
	jumpWordsOutOfRange := cogit JumpAbove: 0.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord - 2 R: nSlotsOrBytesReg.
	cogit AndCq: objectMemory wordSize / 4 - 1 R: formatReg.
	cogit SubR: formatReg R: nSlotsOrBytesReg.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpWordsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	cogit AddCq: objectMemory baseHeaderSize >> (objectMemory shiftForWord - 1) R: Arg0Reg.
	cogit MoveR: TempReg X32r: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	signedVersion
		ifTrue:
			[jumpIsBytes jmpTarget:
			 (cogit ArithmeticShiftRightCq: 7 + objectMemory numTagBits R: Arg1Reg R: TempReg). "Maps in range to -1,0".
			 cogit AddCq: 1 R: TempReg. "Maps in range to 0,1"
			 cogit CmpCq: 1 R: TempReg]
		ifFalse:
			[jumpIsBytes jmpTarget:
			 (cogit CmpCq: (objectMemory integerObjectOf: 255) R: Arg1Reg)].
	jumpBytesOutOfRange := cogit JumpAbove: 0.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord R: nSlotsOrBytesReg.
	cogit AndCq: objectMemory wordSize - 1 R: formatReg R: TempReg.
	cogit SubR: TempReg R: nSlotsOrBytesReg.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpBytesOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit CmpCq: objectMemory firstCompiledMethodFormat R: formatReg.
	jumpIsCompiledMethod := cogit JumpAboveOrEqual: 0.
	methodInBounds :=
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	cogit AddCq: objectMemory baseHeaderSize R: Arg0Reg.
	cogit MoveR: TempReg Xbr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	signedVersion
		ifTrue:
			[jumpIsShorts jmpTarget:
			 (cogit ArithmeticShiftRightCq: 15 + objectMemory numTagBits R: Arg1Reg R: TempReg). "Maps in range to -1,0".
			 cogit AddCq: 1 R: TempReg. "Maps in range to 0,1"
			 cogit CmpCq: 1 R: TempReg]
		ifFalse:
			[jumpIsShorts jmpTarget:
			 (cogit CmpCq: (objectMemory integerObjectOf: 65535) R: Arg1Reg)].
	jumpShortsOutOfRange := cogit JumpAbove: 0.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord - 1 R: nSlotsOrBytesReg.
	cogit AndCq: objectMemory wordSize / 2 - 1 R: formatReg.
	cogit SubR: formatReg R: nSlotsOrBytesReg.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	jumpShortsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	cogit AddR: Arg0Reg R: ReceiverResultReg.
	cogit AddR: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: TempReg M16: objectMemory baseHeaderSize r: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	"Now check that the index is beyond the method's literals..."
	jumpIsCompiledMethod jmpTarget: cogit Label.
	self getLiteralCountOf: ReceiverResultReg plusOne: true inBytes: true into: nSlotsOrBytesReg scratch: TempReg.
	cogit CmpR: Arg0Reg R: nSlotsOrBytesReg.
	cogit JumpBelow: methodInBounds.

	
	jumpNot8ByteInteger jmpTarget:
	(jump64BitArgIsImmediate jmpTarget:
	(jumpNot64BitIndexable jmpTarget:
	(jumpIsContext jmpTarget:
	(jumpNotIndexableBits jmpTarget:
	(jumpBytesOutOfRange jmpTarget:
	(jumpShortsOutOfRange jmpTarget:
	(jumpWordsOutOfRange jmpTarget:
	(jumpIsCompiledMethod jmpTarget:
	(jumpArrayOutOfBounds jmpTarget:
	(jumpBytesOutOfBounds jmpTarget:
	(jumpShortsOutOfBounds jmpTarget:
	(jumpWordsOutOfBounds jmpTarget:
	(jump64BitsOutOfBounds jmpTarget:
	(jumpNotIndexablePointers jmpTarget:
	(jumpFixedFieldsOutOfBounds jmpTarget: cogit Label))))))))))))))).
	signedVersion
		ifTrue: [jumpDoubleWordsOutOfRange jmpTarget: jumpIsContext getJmpTarget]
		ifFalse: [jumpNegative jmpTarget: jumpIsContext getJmpTarget].
	self cppIf: IMMUTABILITY
		ifTrue: [jumpImmutable jmpTarget: jumpIsContext getJmpTarget].

	cogit AddCq: 1 R: Arg0Reg. "0-rel => 1-rel"
	self genConvertIntegerToSmallIntegerInReg: Arg0Reg.

	jumpBadIndex jmpTarget: (jumpImmediate jmpTarget: cogit Label).

	^0 "Can't be complete because of contexts."! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 6/29/2018 18:27'!
genPrimitiveAtSigned: signedVersion
	"Generate the code for primitives 60 & 164, at:/basicAt: & integerAt:.  If signedVersion is true
	 then generate signed accesses to the bits classes (a la 164 & 165).  If signedVersion is false,
	 generate unsigned accesses (a la 60, 61, 63 & 64)."
	| formatReg nSlotsOrElementsReg convertToIntAndReturn methodInBounds
	  jumpNotIndexable jumpImmediate jumpBadIndex
	  jumpIsBytes jumpIsShorts jumpIsWords jumpIsLongs jumpIsMethod jumpIsArray jumpIsContext
	  jumpHasFixedFields jumpArrayOutOfBounds jumpFixedFieldsOutOfBounds
	  jumpBytesOutOfBounds jumpShortsOutOfBounds jumpWordsOutOfBounds jumpLongsOutOfBounds
	  jumpFailAlloc jumpNotSmallInteger |
	"c.f. StackInterpreter>>stSizeOf: SpurMemoryManager>>lengthOf:format: fixedFieldsOf:format:length:"

	nSlotsOrElementsReg := ClassReg.

	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	jumpImmediate := self genJumpImmediate: ReceiverResultReg.
	cogit MoveR: Arg0Reg R: Arg1Reg.
	jumpBadIndex := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	self genConvertSmallIntegerToIntegerInReg: Arg1Reg.
	cogit SubCq: 1 R: Arg1Reg. "1-rel => 0-rel"

	"formatReg := self formatOf: ReceiverResultReg"
	self genGetFormatOf: ReceiverResultReg
		into: (formatReg := SendNumArgsReg)
		leastSignificantHalfOfBaseHeaderIntoScratch: TempReg.

	self genGetNumSlotsOf: ReceiverResultReg into: nSlotsOrElementsReg.

	"dispatch on format in a combination of highest dynamic frequency order first and convenience.
		  0 = 0 sized objects (UndefinedObject True False et al)
		  1 = non-indexable objects with inst vars (Point et al)
		  2 = indexable objects with no inst vars (Array et al)
		  3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
		  4 = weak indexable objects with inst vars (WeakArray et al)
		  5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
		  6 unused, reserved for exotic pointer objects?
		  7 Forwarded Object, 1st field is pointer, rest of fields are ignored
		  8 unused, reserved for exotic non-pointer objects?
		  9 64-bit indexable
		10 - 11 32-bit indexable
		12 - 15 16-bit indexable
		16 - 23 byte indexable
		24 - 31 compiled method"
	cogit CmpCq: objectMemory firstByteFormat R: formatReg.
	jumpIsBytes := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory arrayFormat R: formatReg.
	jumpIsArray := cogit JumpZero: 0.
	jumpNotIndexable := cogit JumpBelow: 0.
					cogit CmpCq: objectMemory weakArrayFormat R: formatReg.
	jumpHasFixedFields := cogit JumpBelowOrEqual: 0.
					cogit CmpCq: objectMemory firstShortFormat R: formatReg.
	jumpIsShorts := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory firstLongFormat R: formatReg.
	jumpIsWords := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory sixtyFourBitIndexableFormat R: formatReg.
	jumpIsLongs := cogit JumpZero: 0.
	jumpNotIndexable jmpTarget: cogit Label.
	jumpNotIndexable := cogit Jump: 0.

	jumpIsArray jmpTarget:
		(cogit CmpR: Arg1Reg R: nSlotsOrElementsReg).
	jumpArrayOutOfBounds := cogit JumpBelowOrEqual: 0.	
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg1Reg.
	cogit MoveXwr: Arg1Reg R: ReceiverResultReg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpIsBytes jmpTarget:
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord R: nSlotsOrElementsReg).
		cogit AndCq: 7 R: formatReg R: TempReg.
		cogit SubR: TempReg R: nSlotsOrElementsReg;
		CmpR: Arg1Reg R: nSlotsOrElementsReg.
	jumpBytesOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit CmpCq: objectMemory firstCompiledMethodFormat R: formatReg.
	jumpIsMethod := cogit JumpAboveOrEqual: 0.
	methodInBounds :=
	cogit AddCq: objectMemory baseHeaderSize R: Arg1Reg.
	(cogit backEnd byteReadsZeroExtend
	 or: [signedVersion])
		ifTrue:
			[cogit MoveXbr: Arg1Reg R: ReceiverResultReg R: ReceiverResultReg]
		ifFalse:
			[cogit "formatReg already contains a value <= 16r1f, so no need to zero it"
				MoveXbr: Arg1Reg R: ReceiverResultReg R: formatReg;
				MoveR: formatReg R: ReceiverResultReg].
	signedVersion ifTrue:
		[cogit SignExtend8R: ReceiverResultReg R: ReceiverResultReg].
	convertToIntAndReturn := cogit Label.
	self genConvertIntegerToSmallIntegerInReg: ReceiverResultReg.
	cogit genPrimReturn.

	jumpIsShorts jmpTarget:
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord - 1 R: nSlotsOrElementsReg).
		cogit AndCq: 3 R: formatReg.
		cogit SubR: formatReg R: nSlotsOrElementsReg;
		CmpR: Arg1Reg R: nSlotsOrElementsReg.
	jumpShortsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddR: Arg1Reg R: ReceiverResultReg.
	cogit AddR: Arg1Reg R: ReceiverResultReg.
	cogit MoveM16: objectMemory baseHeaderSize r: ReceiverResultReg R: ReceiverResultReg.
	signedVersion ifTrue:
		[cogit SignExtend16R: ReceiverResultReg R: ReceiverResultReg].
	cogit Jump: convertToIntAndReturn.

	jumpIsWords jmpTarget:
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord - 2 R: nSlotsOrElementsReg).
		cogit AndCq: 1 R: formatReg.
		cogit SubR: formatReg R: nSlotsOrElementsReg;
		CmpR: Arg1Reg R: nSlotsOrElementsReg.
	jumpWordsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddCq: objectMemory baseHeaderSize >>  (objectMemory shiftForWord - 1) R: Arg1Reg.
	cogit MoveX32r: Arg1Reg R: ReceiverResultReg R: TempReg.
	cogit MoveR: TempReg R: ReceiverResultReg.
	signedVersion ifTrue:
		[cogit SignExtend32R: ReceiverResultReg R: ReceiverResultReg].
	cogit Jump: convertToIntAndReturn.

	jumpHasFixedFields jmpTarget:
		(cogit AndCq: objectMemory classIndexMask R: TempReg).
	cogit MoveR: TempReg R: formatReg.
	cogit CmpCq: ClassMethodContextCompactIndex R: TempReg.
	jumpIsContext := cogit JumpZero: 0.
	self genGetClassObjectOfClassIndex: formatReg into: Extra0Reg scratchReg: TempReg.
	self genLoadSlot: InstanceSpecificationIndex sourceReg: Extra0Reg destReg: formatReg.
	self genConvertSmallIntegerToIntegerInReg: formatReg.
	cogit
		AndCq: objectMemory fixedFieldsOfClassFormatMask R: formatReg;
		SubR: formatReg R: nSlotsOrElementsReg;
		CmpR: Arg1Reg R: nSlotsOrElementsReg.
	jumpFixedFieldsOutOfBounds := cogit JumpBelowOrEqual: 0.
	"index is (formatReg (fixed fields) + Arg1Reg (0-rel index)) * wordSize + baseHeaderSize"
	cogit AddR: formatReg R: Arg1Reg.
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg1Reg.
	cogit MoveXwr: Arg1Reg R: ReceiverResultReg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpIsLongs jmpTarget:
		(cogit CmpR: Arg1Reg R: nSlotsOrElementsReg).
	jumpLongsOutOfBounds := cogit JumpBelowOrEqual: 0.	
	cogit AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg1Reg.
	cogit MoveXwr: Arg1Reg R: ReceiverResultReg R: ClassReg.
	cogit MoveR: ClassReg R: TempReg.
	signedVersion
		ifTrue: "c.f. Spur64BitMemoryManager>>#isIntegerValue:"
			[cogit ArithmeticShiftRightCq: self numSmallIntegerBits R: TempReg. "If in range this is now 0 or -1"
			 cogit AndCq: 16rF R: TempReg.
			 cogit CmpCq: 1 R: TempReg.
			 jumpNotSmallInteger := cogit JumpAbove: 0.
			 cogit MoveR: ClassReg R: ReceiverResultReg.
			 cogit Jump: convertToIntAndReturn.
			 jumpNotSmallInteger jmpTarget: cogit Label.
			 jumpFailAlloc := self genAlloc64BitSignedIntegerValue: ClassReg into: SendNumArgsReg scratchReg: Extra0Reg scratchReg: TempReg.
			 cogit MoveR: SendNumArgsReg R: ReceiverResultReg.
			 cogit genPrimReturn]
		ifFalse:
			[cogit LogicalShiftRightCq: self numSmallIntegerBits - 1 R: TempReg. "If in range this is now 0"
			 (cogit lastOpcode setsConditionCodesFor: JumpZero) ifFalse:
				[cogit CmpCq: 0 R: TempReg]. "N.B. FLAGS := ClassReg - 0"
			 jumpNotSmallInteger := cogit JumpNonZero: 0.
			 cogit MoveR: ClassReg R: ReceiverResultReg.
			 cogit Jump: convertToIntAndReturn.
			 jumpNotSmallInteger jmpTarget: cogit Label.
			 jumpFailAlloc := self genAlloc64BitPositiveIntegerValue: ClassReg into: SendNumArgsReg scratchReg: Extra0Reg scratchReg: TempReg.
			 cogit MoveR: SendNumArgsReg R: ReceiverResultReg.
			 cogit genPrimReturn].

	jumpIsMethod jmpTarget: cogit Label.
	"Now check that the index is beyond the method's literals..."
	self getLiteralCountOf: ReceiverResultReg plusOne: true inBytes: true into: nSlotsOrElementsReg scratch: TempReg.
	cogit CmpR: Arg1Reg R: nSlotsOrElementsReg.
	cogit JumpBelow: methodInBounds.

	jumpFailAlloc jmpTarget:
	(jumpLongsOutOfBounds jmpTarget:
	(jumpFixedFieldsOutOfBounds jmpTarget:
	(jumpArrayOutOfBounds jmpTarget:
	(jumpBytesOutOfBounds jmpTarget:
	(jumpShortsOutOfBounds jmpTarget:
	(jumpWordsOutOfBounds jmpTarget:
	(jumpNotIndexable jmpTarget:
	(jumpIsContext jmpTarget:
	(jumpBadIndex jmpTarget:
	(jumpImmediate jmpTarget: cogit Label)))))))))).

	^0 "Can't be complete because of contexts."! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:30'!
genPrimitiveFloatEqual
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleComparison: #JumpFPEqual: orIntegerComparison: JumpZero invert: false]
		ifFalse: [self genPureDoubleComparison: #JumpFPEqual: invert: false]! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:30'!
genPrimitiveFloatGreaterOrEqual
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleComparison: #JumpFPGreaterOrEqual: orIntegerComparison: JumpGreaterOrEqual invert: false]
		ifFalse: [self genPureDoubleComparison: #JumpFPGreaterOrEqual: invert: false]! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:30'!
genPrimitiveFloatGreaterThan
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleComparison: #JumpFPGreater: orIntegerComparison: JumpGreater invert: false]
		ifFalse: [self genPureDoubleComparison: #JumpFPGreater: invert: false]! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:30'!
genPrimitiveFloatLessOrEqual
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleComparison: #JumpFPGreaterOrEqual: orIntegerComparison: JumpLessOrEqual invert: true]
		ifFalse: [self genPureDoubleComparison: #JumpFPGreaterOrEqual: invert: true]! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:30'!
genPrimitiveFloatLessThan
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleComparison: #JumpFPGreater: orIntegerComparison: JumpLess invert: true]
		ifFalse: [self genPureDoubleComparison: #JumpFPGreater: invert: true]! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 9/6/2019 01:30'!
genPrimitiveFloatNotEqual
	<option: #DPFPReg0>
	^coInterpreter primitiveDoMixedArithmetic
		ifTrue: [self genDoubleComparison: #JumpFPNotEqual: orIntegerComparison: JumpNonZero invert: false]
		ifFalse: [self genPureDoubleComparison: #JumpFPNotEqual: invert: false]! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 5/4/2017 11:47'!
genPrimitiveIdentityHash
	"Arguably we should fail for immediates, but so far no one has complained, so..."
	| jumpImm jumpNotSet jumpNotCharacter ret |
	<var: #jumpImm type: #'AbstractInstruction *'>
	<var: #jumpNotSet type: #'AbstractInstruction *'>
	<var: #jumpNotCharacter type: #'AbstractInstruction *'>
	jumpImm := self genJumpImmediate: ReceiverResultReg. "uses TstCqR"
	self genGetHashFieldNonImmOf: ReceiverResultReg asSmallIntegerInto: TempReg.
	cogit CmpCq: ConstZero R: TempReg.
	jumpNotSet := cogit JumpZero: 0.
	cogit MoveR: TempReg R: ReceiverResultReg.
	ret := cogit genPrimReturn.
	jumpImm jmpTarget: (cogit AndCq: objectMemory tagMask R: ReceiverResultReg R: TempReg).
	cogit CmpCq: objectMemory characterTag R: TempReg.
	jumpNotCharacter := cogit JumpNonZero: 0.
	self genConvertCharacterToSmallIntegerInReg: ReceiverResultReg.
	ret := cogit genPrimReturn.
	jumpNotCharacter jmpTarget: (cogit CmpCq: objectMemory smallFloatTag R: TempReg).
	cogit JumpNonZero: ret.
	self genConvertSmallFloatToSmallFloatHashAsIntegerInReg: ReceiverResultReg scratch: TempReg.
	cogit genPrimReturn.
	jumpNotSet jmpTarget: cogit Label.
	"For non-behaviors we can call ceNewHash to assign a fresh hash.
	 For behaviors we must fail into the interpreter primitive to enter the receiver into the class table."
	cogit getPrimitiveIndex = 75 ifFalse:
		[^0].
	cogit backEnd saveAndRestoreLinkRegAround:
		[cogit CallRT: ceNewHashTrampoline].
	cogit genPrimReturn.
	^UnfailingPrimitive! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 2/28/2016 11:27'!
genPrimitiveImmediateAsInteger
	| jumpNotCharacter ret |
	<var: #jumpNotCharacter type: #'AbstractInstruction *'>
	jumpNotCharacter := cogit
		AndCq: objectMemory tagMask R: ReceiverResultReg R: TempReg;
		CmpCq: objectMemory characterTag R: TempReg;
		JumpNonZero: 0.
	self genConvertCharacterToSmallIntegerInReg: ReceiverResultReg.
	ret := cogit genPrimReturn.
	jumpNotCharacter jmpTarget:
		(cogit CmpCq: objectMemory smallFloatTag R: TempReg).
	cogit JumpNonZero: ret.
	self genConvertSmallFloatToSmallFloatHashAsIntegerInReg: ReceiverResultReg scratch: TempReg.
	cogit genPrimReturn.
	^UnfailingPrimitive! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/11/2016 13:15'!
genPrimitiveMirrorNew
	"Implement 1-arg (instantiateFixedClass:) primitiveNew for convenient cases:
	- the class argument has a hash
	- the class argument is fixed size (excluding ephemerons to save instructions & miniscule time)
	- single word header/num slots < numSlotsMask
	- the result fits in eden (actually below scavengeThreshold)"

	<option: #NewspeakVM>
	| headerReg fillReg instSpecReg byteSizeReg
	  jumpImmediate jumpUnhashed jumpNotFixedPointers jumpTooSmall jumpBadFormat
	  jumpNoSpace jumpTooBig jumpHasSlots jumpVariableOrEphemeron
	  fillLoop skip |
	<var: 'skip' type: #'AbstractInstruction *'>
	<var: 'fillLoop' type: #'AbstractInstruction *'>
	<var: 'jumpTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpHasSlots' type: #'AbstractInstruction *'>
	<var: 'jumpNoSpace' type: #'AbstractInstruction *'>
	<var: 'jumpTooSmall' type: #'AbstractInstruction *'>
	<var: 'jumpUnhashed' type: #'AbstractInstruction *'>
	<var: 'jumpImmediate' type: #'AbstractInstruction *'>
	<var: 'jumpBadFormat' type: #'AbstractInstruction *'>
	<var: 'jumpNotFixedPointers' type: #'AbstractInstruction *'>
	<var: 'jumpVariableOrEphemeron' type: #'AbstractInstruction *'>

	self assert: cogit methodNumArgs = 1.
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.

	"header will contain classIndex/class's hash & format & numSlots/fixed size and finally fill value (nilObject)."
	headerReg := fillReg := SendNumArgsReg.
	"inst spec will hold class's instance specification, then byte size and finally end of new object."
	instSpecReg := byteSizeReg := ClassReg.

	"get freeStart as early as possible so as not to wait later..."
	cogit MoveAw: objectMemory freeStartAddress R: Arg1Reg.

	"validate class arg; sigh, this mirror crap hobbles unfairly; there is a better way with selector namespaces..."
	jumpImmediate := self genJumpImmediate: Arg0Reg.

	"Is the class arg pointers with at least 3 fields?"
	self genGetFormatOf: Arg0Reg
		into: TempReg
		leastSignificantHalfOfBaseHeaderIntoScratch: NoReg.
	cogit CmpCq: objectMemory nonIndexablePointerFormat R: TempReg.
	jumpNotFixedPointers := cogit JumpNonZero: 0.
	
	self genGetRawSlotSizeOfNonImm: Arg0Reg into: TempReg.
	cogit CmpCq: InstanceSpecificationIndex + 1 R: TempReg.
	jumpTooSmall := cogit JumpLess: 0.

	"get class's hash & fail if 0"
	self genGetHashFieldNonImmOf: Arg0Reg into: headerReg.
	jumpUnhashed := cogit JumpZero: 0.

	"get class's format inst var for both inst spec (format field) and num fixed fields"
	self genLoadSlot: InstanceSpecificationIndex sourceReg: Arg0Reg destReg: instSpecReg.
	jumpBadFormat := self genJumpNotSmallInteger: instSpecReg scratchReg: TempReg.
	self genConvertSmallIntegerToIntegerInReg: instSpecReg.
	cogit MoveR: instSpecReg R: TempReg.
	cogit LogicalShiftRightCq: objectMemory fixedFieldsFieldWidth R: TempReg.
	cogit AndCq: objectMemory formatMask R: TempReg.
	cogit AndCq: objectMemory fixedFieldsOfClassFormatMask R: instSpecReg.
	"fail if not fixed or if ephemeron (rare beasts so save the cycles)"
	cogit CmpCq: objectMemory nonIndexablePointerFormat R: TempReg.
	jumpVariableOrEphemeron := cogit JumpAbove: 0.
	cogit CmpCq: objectMemory numSlotsMask R: instSpecReg.
	jumpTooBig := cogit JumpAboveOrEqual: 0.
	"Add format to classIndex/format in header; then add in numSlots"
	cogit LogicalShiftLeftCq: objectMemory formatShift R: TempReg.
	cogit AddR: TempReg R: headerReg.
	cogit MoveR: instSpecReg R: TempReg.
	cogit LogicalShiftLeftCq: objectMemory numSlotsFullShift R: TempReg.
	cogit AddR: TempReg R: headerReg.
	"compute byte size; remember 0-sized objects still need 1 slot."
	cogit CmpCq: 0 R: byteSizeReg. "a.k.a. instSpecReg"
	jumpHasSlots := cogit JumpNonZero: 0.
	cogit MoveCq: objectMemory baseHeaderSize * 2 R: byteSizeReg.
	skip := cogit Jump: 0.
	"round up to allocationUnit"
	jumpHasSlots jmpTarget:
	(cogit AddCq: objectMemory baseHeaderSize / objectMemory wordSize R: byteSizeReg).
	cogit LogicalShiftLeftCq: objectMemory shiftForWord R: byteSizeReg.
	skip jmpTarget:
	"check if allocation fits (freeSize + byteSize < scavengeThreshold); scavengeThreshold is constant."
	(cogit AddR: Arg1Reg R: byteSizeReg).
	cogit CmpCq: objectMemory getScavengeThreshold R: byteSizeReg.
	jumpNoSpace := cogit JumpAboveOrEqual: 0.
	"write back new freeStart; get result. byteSizeReg holds new freeStart, the limit of the object"
	cogit MoveR: byteSizeReg Aw: objectMemory freeStartAddress.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	"write header"
	cogit MoveR: headerReg Mw: 0 r: Arg1Reg.
	"now fill"
	cogit LoadEffectiveAddressMw: objectMemory baseHeaderSize r: ReceiverResultReg R: Arg1Reg.
	cogit MoveCq: objectMemory nilObject R: fillReg.
	fillLoop := 
	cogit MoveR: fillReg Mw: 0 r: Arg1Reg.
	cogit AddCq: 8 R: Arg1Reg.
	cogit CmpR: Arg1Reg R: byteSizeReg.
	cogit JumpAbove: fillLoop.
	cogit genPrimReturn.

	jumpNotFixedPointers jmpTarget:
	(jumpBadFormat jmpTarget:
	(jumpTooSmall jmpTarget:
	(jumpImmediate jmpTarget:
	(jumpUnhashed jmpTarget:
	(jumpVariableOrEphemeron jmpTarget:
	(jumpTooBig jmpTarget:
	(jumpNoSpace jmpTarget: cogit Label))))))).

	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 04/16/2016 15:12'!
genPrimitiveMirrorNewWithArg
	"Implement instantiateVariableClass:withSize: for convenient cases:
	- the class argument has a hash
	- the class argument is variable and not compiled method
	- single word header/num slots < numSlotsMask
	- the result fits in eden
	See superclass method for dynamic frequencies of formats.
	For the moment we implement only arrayFormat, firstByteFormat & firstLongFormat"

	<option: #NewspeakVM>
	| headerReg fillReg instSpecReg byteSizeReg maxSlots
	  jumpArrayTooBig jumpByteTooBig jumpLongTooBig
	  jumpArrayFormat jumpByteFormat jumpBytePrepDone jumpLongPrepDone
	  jumpUnhashed jumpTooSmall jumpImmediate
	  jumpNElementsNonInt jumpFailCuzFixed jumpNoSpace jumpHasSlots fillLoop skip |
	<var: 'skip' type: #'AbstractInstruction *'>
	<var: 'fillLoop' type: #'AbstractInstruction *'>	
	<var: 'jumpHasSlots' type: #'AbstractInstruction *'>
	<var: 'jumpNoSpace' type: #'AbstractInstruction *'>
	<var: 'jumpTooSmall' type: #'AbstractInstruction *'>
	<var: 'jumpUnhashed' type: #'AbstractInstruction *'>
	<var: 'jumpImmediate' type: #'AbstractInstruction *'>
	<var: 'jumpByteFormat' type: #'AbstractInstruction *'>
	<var: 'jumpByteTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpLongTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpArrayFormat' type: #'AbstractInstruction *'>
	<var: 'jumpArrayTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpFailCuzFixed' type: #'AbstractInstruction *'>
	<var: 'jumpBytePrepDone' type: #'AbstractInstruction *'>
	<var: 'jumpLongPrepDone' type: #'AbstractInstruction *'>
	<var: 'jumpNElementsNonInt' type: #'AbstractInstruction *'>

	self assert: cogit methodNumArgs = 2.
	cogit genLoadArgAtDepth: 1 into: Arg0Reg. "class arg"
	cogit genLoadArgAtDepth: 0 into: Arg1Reg. "num indexable slots arg"
	"header will contain classIndex/class's hash & format & numSlots/fixed size"
	headerReg := SendNumArgsReg.
	"Assume there's an available scratch register on 64-bit machines.  This holds the saved numFixedFields and then the value to fill with"
	fillReg := Extra0Reg.
	self assert: fillReg > 0.
	"inst spec will hold class's instance specification and then byte size"
	instSpecReg := byteSizeReg := ClassReg.
	"The max slots we'll allocate here are those for a single header"
	maxSlots := objectMemory numSlotsMask - 1.

	"check size and fail if not a +ve integer"
	jumpNElementsNonInt := self genJumpNotSmallInteger: Arg1Reg scratchReg: TempReg.

	"Is the class arg pointers with at least 3 fields?"
	jumpImmediate := self genJumpImmediate: Arg0Reg.
	self genGetRawSlotSizeOfNonImm: Arg0Reg into: TempReg.
	cogit CmpCq: InstanceSpecificationIndex + 1 R: TempReg.
	jumpTooSmall := cogit JumpLess: 0.

	"get freeStart as early as possible so as not to wait later..."
	cogit MoveAw: objectMemory freeStartAddress R: Arg1Reg.
	"get class's hash & fail if 0"
	self genGetHashFieldNonImmOf: ReceiverResultReg into: headerReg.
	jumpUnhashed := cogit JumpZero: 0.

	"The basicNew: code below (copied from genPrimitiveNewWithArg) expects class
	 in ReceiverResultReg and size in Arg0Reg.  Shuffle args to match, undoing on failure."
	cogit
		PushR: ReceiverResultReg;
		MoveR: Arg0Reg R: ReceiverResultReg;
		MoveR: Arg1Reg R: Arg0Reg.

	"get class's format inst var for inst spec (format field)"
	self genLoadSlot: InstanceSpecificationIndex sourceReg: ReceiverResultReg destReg: instSpecReg.
	cogit LogicalShiftRightCq: objectMemory fixedFieldsFieldWidth + self numSmallIntegerTagBits R: instSpecReg.
	cogit AndCq: objectMemory formatMask R: instSpecReg.
	"Add format to classIndex/format header now"
	cogit MoveR: instSpecReg R: TempReg.
	cogit LogicalShiftLeftCq: objectMemory formatShift R: TempReg.
	cogit AddR: TempReg R: headerReg.
	"get integer value of num fields in fillReg now"
	cogit MoveR: Arg0Reg R: fillReg.
	self genConvertSmallIntegerToIntegerInReg: fillReg.
	"dispatch on format, failing if not variable or if compiled method"
	cogit CmpCq: objectMemory arrayFormat R: instSpecReg.
	jumpArrayFormat := cogit JumpZero: 0.
	cogit CmpCq: objectMemory firstByteFormat R: instSpecReg.
	jumpByteFormat := cogit JumpZero: 0.
	cogit CmpCq: objectMemory firstLongFormat R: instSpecReg.
	jumpFailCuzFixed := cogit JumpNonZero: 0.

	cogit CmpCq: (objectMemory integerObjectOf: maxSlots) R: Arg0Reg.
	jumpLongTooBig := cogit JumpAbove: 0.
	"save num elements/slot size to instSpecReg"
	cogit MoveR: fillReg R: instSpecReg.
	"compute odd bits and add into headerReg; oddBits := 2 - nElements bitAnd: 1"
	cogit MoveCq: objectMemory wordSize / 4 R: TempReg.
	cogit SubR: instSpecReg R: TempReg.
	cogit AndCq: objectMemory wordSize / 4 - 1 R: TempReg.
	cogit LogicalShiftLeftCq: objectMemory formatShift R: TempReg.
	cogit AddR: TempReg R: headerReg.
	"round up num elements to numSlots in instSpecReg"
	cogit AddCq: objectMemory wordSize / 4 - 1 R: instSpecReg.
	cogit LogicalShiftRightCq: objectMemory shiftForWord - 2 R: instSpecReg.
	cogit MoveCq: 0 R: fillReg.
	jumpLongPrepDone := cogit Jump: 0. "go allocate"

	jumpByteFormat jmpTarget:
	(cogit CmpCq: (objectMemory integerObjectOf: maxSlots * objectMemory wordSize) R: Arg0Reg).
	jumpByteTooBig := cogit JumpAbove: 0.
	"save num elements to instSpecReg"
	cogit MoveR: fillReg R: instSpecReg.
	"compute odd bits and add into headerReg; oddBits := 4 - nElements bitAnd: 3"
	cogit MoveCq: objectMemory wordSize R: TempReg.
	cogit SubR: instSpecReg R: TempReg.
	cogit AndCq: objectMemory wordSize - 1 R: TempReg.
	cogit LogicalShiftLeftCq: objectMemory formatShift R: TempReg.
	cogit AddR: TempReg R: headerReg.
	"round up num elements to numSlots in instSpecReg"
	cogit AddCq: objectMemory wordSize - 1 R: instSpecReg.
	cogit LogicalShiftRightCq: objectMemory shiftForWord R: instSpecReg.
	cogit MoveCq: 0 R: fillReg.
	jumpBytePrepDone := cogit Jump: 0. "go allocate"

	jumpArrayFormat jmpTarget:
		(cogit CmpCq: (objectMemory integerObjectOf: maxSlots) R: Arg0Reg).
	jumpArrayTooBig := cogit JumpAbove: 0.
	"save num elements/slot size to instSpecReg"
	cogit MoveR: fillReg R: instSpecReg.
	cogit MoveCq: objectMemory nilObject R: fillReg.
	"fall through to allocate"

	jumpBytePrepDone jmpTarget:
	(jumpLongPrepDone jmpTarget: cogit Label).

	"store numSlots to headerReg"
	cogit MoveR: instSpecReg R: TempReg.
	cogit LogicalShiftLeftCq: objectMemory numSlotsFullShift R: TempReg.
	cogit AddR: TempReg R: headerReg.
	"compute byte size; remember 0-sized objects still need 1 slot."
	cogit CmpCq: 0 R: byteSizeReg. "a.k.a. instSpecReg"
	jumpHasSlots := cogit JumpNonZero: 0.
	cogit MoveCq: objectMemory baseHeaderSize * 2 R: byteSizeReg.
	skip := cogit Jump: 0.
	jumpHasSlots jmpTarget:
	(cogit AddCq: objectMemory baseHeaderSize / objectMemory wordSize R: byteSizeReg).
	cogit LogicalShiftLeftCq: objectMemory shiftForWord R: byteSizeReg.
	skip jmpTarget:
	"check if allocation fits"
	(cogit AddR: Arg1Reg R: byteSizeReg).
	cogit CmpCq: objectMemory getScavengeThreshold R: byteSizeReg.
	jumpNoSpace := cogit JumpAboveOrEqual: 0.
	"get result, increment freeStart and write it back. Arg1Reg holds new freeStart, the limit of the object"
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit MoveR: byteSizeReg Aw: objectMemory freeStartAddress.
	"write other half of header (numSlots/0 identityHash)"
	cogit MoveR: headerReg Mw: 0 r: ReceiverResultReg.
	"now fill"
	cogit PopR: TempReg. "discard pushed receiver"
	cogit LoadEffectiveAddressMw: objectMemory baseHeaderSize r: ReceiverResultReg R: Arg1Reg.
	fillLoop := 
	cogit MoveR: fillReg Mw: 0 r: Arg1Reg.
	cogit AddCq: 8 R: Arg1Reg.
	cogit CmpR: Arg1Reg R: byteSizeReg.
	cogit JumpAbove: fillLoop.
	cogit genPrimReturn.

	jumpFailCuzFixed jmpTarget:
	(jumpArrayTooBig jmpTarget:
	(jumpByteTooBig jmpTarget:
	(jumpLongTooBig jmpTarget:
	(jumpNoSpace jmpTarget:  cogit Label)))).

	"unshuffle arguments"
	cogit
		MoveR: Arg0Reg R: Arg1Reg;
		MoveR: ReceiverResultReg R: Arg0Reg;
		PopR: ReceiverResultReg.

	jumpUnhashed jmpTarget:
	(jumpImmediate jmpTarget:
	(jumpTooSmall jmpTarget:
	(jumpNElementsNonInt jmpTarget: cogit Label))).

	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:53'!
genPrimitiveNew
	"Implement primitiveNew for convenient cases:
	- the receiver has a hash
	- the receiver is fixed size (excluding ephemerons to save instructions & miniscule time)
	- single word header/num slots < numSlotsMask
	- the result fits in eden (actually below scavengeThreshold)"

	| headerReg fillReg instSpecReg byteSizeReg
	  jumpUnhashed jumpVariableOrEphemeron jumpNoSpace jumpTooBig jumpHasSlots
	  fillLoop skip |
	<var: 'skip' type: #'AbstractInstruction *'>
	<var: 'fillLoop' type: #'AbstractInstruction *'>
	<var: 'jumpTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpHasSlots' type: #'AbstractInstruction *'>
	<var: 'jumpNoSpace' type: #'AbstractInstruction *'>
	<var: 'jumpUnhashed' type: #'AbstractInstruction *'>
	<var: 'jumpVariableOrEphemeron' type: #'AbstractInstruction *'>

	NewspeakVM ifTrue:
		[cogit methodNumArgs = 1 ifTrue:
			[^self genPrimitiveMirrorNew]].
	cogit methodNumArgs ~= 0 ifTrue:
		[^UnimplementedPrimitive].

	"header will contain classIndex/class's hash & format & numSlots/fixed size and finally fill value (nilObject)."
	headerReg := fillReg := SendNumArgsReg.
	"inst spec will hold class's instance specification, then byte size and finally end of new object."
	instSpecReg := byteSizeReg := ClassReg.

	"get freeStart as early as possible so as not to wait later..."
	cogit MoveAw: objectMemory freeStartAddress R: Arg1Reg.
	"get class's hash & fail if 0"
	self genGetHashFieldNonImmOf: ReceiverResultReg into: headerReg.
	jumpUnhashed := cogit JumpZero: 0.
	"get class's format inst var for both inst spec (format field) and num fixed fields"
	self genLoadSlot: InstanceSpecificationIndex sourceReg: ReceiverResultReg destReg: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	cogit MoveR: TempReg R: instSpecReg.
	cogit LogicalShiftRightCq: objectMemory fixedFieldsFieldWidth R: TempReg.
	cogit AndCq: objectMemory formatMask R: TempReg.
	cogit AndCq: objectMemory fixedFieldsOfClassFormatMask R: instSpecReg.
	"fail if not fixed or if ephemeron (rare beasts so save the cycles)"
	cogit CmpCq: objectMemory nonIndexablePointerFormat R: TempReg.
	jumpVariableOrEphemeron := cogit JumpAbove: 0.
	cogit CmpCq: objectMemory numSlotsMask R: instSpecReg.
	jumpTooBig := cogit JumpAboveOrEqual: 0.
	"Add format to classIndex/format in header; the add in numSlots"
	cogit LogicalShiftLeftCq: objectMemory formatShift R: TempReg.
	cogit AddR: TempReg R: headerReg.
	cogit MoveR: instSpecReg R: TempReg.
	cogit LogicalShiftLeftCq: objectMemory numSlotsFullShift R: TempReg.
	cogit AddR: TempReg R: headerReg.
	"compute byte size; remember 0-sized objects still need 1 slot."
	cogit CmpCq: 0 R: byteSizeReg. "a.k.a. instSpecReg"
	jumpHasSlots := cogit JumpNonZero: 0.
	cogit MoveCq: objectMemory baseHeaderSize * 2 R: byteSizeReg.
	skip := cogit Jump: 0.
	"round up to allocationUnit"
	jumpHasSlots jmpTarget:
	(cogit AddCq: objectMemory baseHeaderSize / objectMemory wordSize R: byteSizeReg).
	cogit LogicalShiftLeftCq: objectMemory shiftForWord R: byteSizeReg.
	skip jmpTarget:
	"check if allocation fits (freeSize + byteSize < scavengeThreshold); scavengeThreshold is constant."
	(cogit AddR: Arg1Reg R: byteSizeReg).
	cogit CmpCq: objectMemory getScavengeThreshold R: byteSizeReg.
	jumpNoSpace := cogit JumpAboveOrEqual: 0.
	"write back new freeStart; get result. byteSizeReg holds new freeStart, the limit of the object"
	cogit MoveR: byteSizeReg Aw: objectMemory freeStartAddress.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	"write header"
	cogit MoveR: headerReg Mw: 0 r: Arg1Reg.
	"now fill"
	cogit LoadEffectiveAddressMw: objectMemory baseHeaderSize r: ReceiverResultReg R: Arg1Reg.
	cogit MoveCq: objectMemory nilObject R: fillReg.
	fillLoop := 
	cogit MoveR: fillReg Mw: 0 r: Arg1Reg.
	cogit AddCq: 8 R: Arg1Reg.
	cogit CmpR: Arg1Reg R: byteSizeReg.
	cogit JumpAbove: fillLoop.
	cogit genPrimReturn.

	jumpUnhashed jmpTarget:
	(jumpVariableOrEphemeron jmpTarget:
	(jumpTooBig jmpTarget:
	(jumpNoSpace jmpTarget: cogit Label))).

	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 04/16/2016 15:12'!
genPrimitiveNewWithArg
	"Implement primitiveNewWithArg for convenient cases:
	- the receiver has a hash
	- the receiver is variable and not compiled method
	- single word header/num slots < numSlotsMask
	- the result fits in eden
	See superclass method for dynamic frequencies of formats.
	For the moment we implement only arrayFormat, firstByteFormat & firstLongFormat"

	| headerReg fillReg instSpecReg byteSizeReg maxSlots
	  jumpArrayTooBig jumpByteTooBig jumpLongTooBig
	  jumpArrayFormat jumpByteFormat jumpBytePrepDone jumpLongPrepDone
	  jumpUnhashed jumpNElementsNonInt jumpFailCuzFixed jumpNoSpace jumpHasSlots fillLoop skip |
	<var: 'skip' type: #'AbstractInstruction *'>
	<var: 'fillLoop' type: #'AbstractInstruction *'>	
	<var: 'jumpHasSlots' type: #'AbstractInstruction *'>
	<var: 'jumpNoSpace' type: #'AbstractInstruction *'>
	<var: 'jumpUnhashed' type: #'AbstractInstruction *'>
	<var: 'jumpByteFormat' type: #'AbstractInstruction *'>
	<var: 'jumpByteTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpLongTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpArrayFormat' type: #'AbstractInstruction *'>
	<var: 'jumpArrayTooBig' type: #'AbstractInstruction *'>
	<var: 'jumpFailCuzFixed' type: #'AbstractInstruction *'>
	<var: 'jumpBytePrepDone' type: #'AbstractInstruction *'>
	<var: 'jumpLongPrepDone' type: #'AbstractInstruction *'>
	<var: 'jumpNElementsNonInt' type: #'AbstractInstruction *'>

	NewspeakVM ifTrue:
		[cogit methodNumArgs = 2 ifTrue:
			[^self genPrimitiveMirrorNewWithArg]].
	cogit methodNumArgs ~= 1 ifTrue:
		[^UnimplementedPrimitive].
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.

	"header will contain classIndex/class's hash & format & numSlots/fixed size"
	headerReg := SendNumArgsReg.
	"Assume there's an available scratch register on 64-bit machines.  This holds the saved numFixedFields and then the value to fill with"
	fillReg := Extra0Reg.
	self assert: fillReg > 0.
	"inst spec will hold class's instance specification and then byte size"
	instSpecReg := byteSizeReg := ClassReg.
	"The max slots we'll allocate here are those for a single header"
	maxSlots := objectMemory numSlotsMask - 1.

	"get freeStart as early as possible so as not to wait later..."
	cogit MoveAw: objectMemory freeStartAddress R: Arg1Reg.
	"get class's hash & fail if 0"
	self genGetHashFieldNonImmOf: ReceiverResultReg into: headerReg.
	jumpUnhashed := cogit JumpZero: 0.
	"get index and fail if not a +ve integer"
	jumpNElementsNonInt := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	"get class's format inst var for inst spec (format field)"
	self genLoadSlot: InstanceSpecificationIndex sourceReg: ReceiverResultReg destReg: instSpecReg.
	cogit LogicalShiftRightCq: objectMemory fixedFieldsFieldWidth + self numSmallIntegerTagBits R: instSpecReg.
	cogit AndCq: objectMemory formatMask R: instSpecReg.
	"Add format to classIndex/format header now"
	cogit MoveR: instSpecReg R: TempReg.
	cogit LogicalShiftLeftCq: objectMemory formatShift R: TempReg.
	cogit AddR: TempReg R: headerReg.
	"get integer value of num fields in fillReg now"
	cogit MoveR: Arg0Reg R: fillReg.
	self genConvertSmallIntegerToIntegerInReg: fillReg.
	"dispatch on format, failing if not variable or if compiled method"
	cogit CmpCq: objectMemory arrayFormat R: instSpecReg.
	jumpArrayFormat := cogit JumpZero: 0.
	cogit CmpCq: objectMemory firstByteFormat R: instSpecReg.
	jumpByteFormat := cogit JumpZero: 0.
	cogit CmpCq: objectMemory firstLongFormat R: instSpecReg.
	jumpFailCuzFixed := cogit JumpNonZero: 0.

	cogit CmpCq: (objectMemory integerObjectOf: maxSlots * 2) R: Arg0Reg.
	jumpLongTooBig := cogit JumpAbove: 0.
	"save num elements/slot size to instSpecReg"
	cogit MoveR: fillReg R: instSpecReg.
	"compute odd bits and add into headerReg; oddBits := 2 - nElements bitAnd: 1"
	cogit MoveCq: objectMemory wordSize / 4 R: TempReg.
	cogit SubR: instSpecReg R: TempReg.
	cogit AndCq: objectMemory wordSize / 4 - 1 R: TempReg.
	cogit LogicalShiftLeftCq: objectMemory formatShift R: TempReg.
	cogit AddR: TempReg R: headerReg.
	"round up num elements to numSlots in instSpecReg"
	cogit AddCq: objectMemory wordSize / 4 - 1 R: instSpecReg.
	cogit LogicalShiftRightCq: objectMemory shiftForWord - 2 R: instSpecReg.
	cogit MoveCq: 0 R: fillReg.
	jumpLongPrepDone := cogit Jump: 0. "go allocate"

	jumpByteFormat jmpTarget:
	(cogit CmpCq: (objectMemory integerObjectOf: maxSlots * objectMemory wordSize) R: Arg0Reg).
	jumpByteTooBig := cogit JumpAbove: 0.
	"save num elements to instSpecReg"
	cogit MoveR: fillReg R: instSpecReg.
	"compute odd bits and add into headerReg; oddBits := 4 - nElements bitAnd: 3"
	cogit MoveCq: objectMemory wordSize R: TempReg.
	cogit SubR: instSpecReg R: TempReg.
	cogit AndCq: objectMemory wordSize - 1 R: TempReg.
	cogit LogicalShiftLeftCq: objectMemory formatShift R: TempReg.
	cogit AddR: TempReg R: headerReg.
	"round up num elements to numSlots in instSpecReg"
	cogit AddCq: objectMemory wordSize - 1 R: instSpecReg.
	cogit LogicalShiftRightCq: objectMemory shiftForWord R: instSpecReg.
	cogit MoveCq: 0 R: fillReg.
	jumpBytePrepDone := cogit Jump: 0. "go allocate"

	jumpArrayFormat jmpTarget:
		(cogit CmpCq: (objectMemory integerObjectOf: maxSlots) R: Arg0Reg).
	jumpArrayTooBig := cogit JumpAbove: 0.
	"save num elements/slot size to instSpecReg"
	cogit MoveR: fillReg R: instSpecReg.
	cogit MoveCq: objectMemory nilObject R: fillReg.
	"fall through to allocate"

	jumpBytePrepDone jmpTarget:
	(jumpLongPrepDone jmpTarget: cogit Label).

	"store numSlots to headerReg"
	cogit MoveR: instSpecReg R: TempReg.
	cogit LogicalShiftLeftCq: objectMemory numSlotsFullShift R: TempReg.
	cogit AddR: TempReg R: headerReg.
	"compute byte size; remember 0-sized objects still need 1 slot."
	cogit CmpCq: 0 R: byteSizeReg. "a.k.a. instSpecReg"
	jumpHasSlots := cogit JumpNonZero: 0.
	cogit MoveCq: objectMemory baseHeaderSize * 2 R: byteSizeReg.
	skip := cogit Jump: 0.
	jumpHasSlots jmpTarget:
	(cogit AddCq: objectMemory baseHeaderSize / objectMemory wordSize R: byteSizeReg).
	cogit LogicalShiftLeftCq: objectMemory shiftForWord R: byteSizeReg.
	skip jmpTarget:
	"check if allocation fits"
	(cogit AddR: Arg1Reg R: byteSizeReg).
	cogit CmpCq: objectMemory getScavengeThreshold R: byteSizeReg.
	jumpNoSpace := cogit JumpAboveOrEqual: 0.
	"get result, increment freeStart and write it back. Arg1Reg holds new freeStart, the limit of the object"
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit MoveR: byteSizeReg Aw: objectMemory freeStartAddress.
	"write other half of header (numSlots/0 identityHash)"
	cogit MoveR: headerReg Mw: 0 r: ReceiverResultReg.
	"now fill"
	cogit LoadEffectiveAddressMw: objectMemory baseHeaderSize r: ReceiverResultReg R: Arg1Reg.
	fillLoop := 
	cogit MoveR: fillReg Mw: 0 r: Arg1Reg.
	cogit AddCq: 8 R: Arg1Reg.
	cogit CmpR: Arg1Reg R: byteSizeReg.
	cogit JumpAbove: fillLoop.
	cogit genPrimReturn.
	
	jumpNoSpace jmpTarget:
	(jumpUnhashed jmpTarget:
	(jumpFailCuzFixed jmpTarget:
	(jumpArrayTooBig jmpTarget:
	(jumpByteTooBig jmpTarget:
	(jumpLongTooBig jmpTarget:
	(jumpNElementsNonInt jmpTarget: cogit Label)))))).

	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 9/7/2016 11:48'!
genPrimitiveShallowCopy
	"Implement primitiveShallowCopy/primitiveClone for convenient cases:
	- the receiver is not a context
	- the receiver is not a compiled method
	- the result fits in eden (actually below scavengeThreshold)"

	| formatReg resultReg slotsReg ptrReg
	  jumpImmediate jumpIsMethod jumpVariable jumpTooBig jumpEmpty jumpNoSpace
	  continuance copyLoop |
	<var: #copyLoop type: #'AbstractInstruction *'>
	<var: #jumpTooBig type: #'AbstractInstruction *'>
	<var: #continuance type: #'AbstractInstruction *'>
	<var: #jumpVariable type: #'AbstractInstruction *'>
	<var: #jumpNoSpace type: #'AbstractInstruction *'>
	<var: #jumpIsMethod type: #'AbstractInstruction *'>
	<var: #jumpImmediate type: #'AbstractInstruction *'>

	jumpImmediate := self genJumpImmediate: ReceiverResultReg.
	resultReg := Arg0Reg.
	slotsReg := Arg1Reg.
	"get freeStart as early as possible so as not to wait later..."
	cogit MoveAw: objectMemory freeStartAddress R: resultReg.

	"formatReg := self formatOf: ReceiverResultReg"
	self genGetFormatOf: ReceiverResultReg
		into: (ptrReg := formatReg := SendNumArgsReg)
		leastSignificantHalfOfBaseHeaderIntoScratch: NoReg.

	cogit CmpCq: objectMemory firstCompiledMethodFormat R: formatReg.
	jumpIsMethod := cogit JumpAboveOrEqual: 0.
	cogit CmpCq: objectMemory indexablePointersFormat R: formatReg.
	jumpVariable := cogit JumpZero: 0.
	continuance := cogit Label.

	self genGetRawSlotSizeOfNonImm: ReceiverResultReg into: slotsReg.
	cogit CmpCq: objectMemory numSlotsMask R: slotsReg.
	jumpTooBig := cogit JumpZero: 0.

	cogit CmpCq: 0 R: slotsReg.
	jumpEmpty := cogit JumpZero: 0.

	"compute byte size for slots"
	cogit
		AddCq: objectMemory baseHeaderSize / objectMemory wordSize R: slotsReg;
		LogicalShiftLeftCq: objectMemory shiftForWord R: slotsReg;
	"check if allocation fits (freeSize + byteSize < scavengeThreshold); scavengeThreshold is constant."
		AddR: resultReg R: slotsReg;
		CmpCq: objectMemory getScavengeThreshold R: slotsReg.
	jumpNoSpace := cogit JumpAboveOrEqual: 0.
	cogit
		MoveR: resultReg R: ptrReg;
	"write back new freeStart; get result. slotsReg holds new freeStart, the limit of the object"
		MoveR: slotsReg Aw: objectMemory freeStartAddress;
	"set up loop bounds"
		SubCq: objectMemory wordSize * 2 R: slotsReg;
	"copy header, masking off irrelevant bits"
		MoveMw: 0 r: ReceiverResultReg R: TempReg;
		AndCq: (objectMemory
					headerForSlots: objectMemory numSlotsMask
					format: objectMemory formatMask
					classIndex: objectMemory classIndexMask) R: TempReg;
		MoveR: TempReg Mw: 0 r: resultReg.
	copyLoop := cogit Label.
	cogit
		AddCq: objectMemory wordSize R: ReceiverResultReg;
		AddCq: objectMemory wordSize R: ptrReg;
		MoveMw: 0 r: ReceiverResultReg R: TempReg;
		MoveR: TempReg Mw: 0 r: ptrReg;
		CmpR: ptrReg R: slotsReg;
		JumpAboveOrEqual: copyLoop;
		MoveR: resultReg R: ReceiverResultReg;
		genPrimReturn.

	"If the receiver is variable pointers, fail if its a context, otherwise continue"
	jumpVariable jmpTarget: cogit Label.
	self genGetClassIndexOfNonImm: ReceiverResultReg into: ClassReg.
	cogit
		CmpCq: ClassMethodContextCompactIndex R: ClassReg;
		JumpNonZero: continuance.

	jumpImmediate jmpTarget:
	(jumpNoSpace jmpTarget:
	(jumpIsMethod jmpTarget:
	(jumpTooBig jmpTarget:
	(jumpEmpty jmpTarget: cogit Label)))).

	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/10/2017 15:57'!
genPrimitiveStringAt
	| formatReg jumpNotIndexable jumpBadIndex done
	  jumpIsBytes jumpIsShorts jumpWordTooBig jumpWordsDone
	  jumpBytesOutOfBounds jumpShortsOutOfBounds jumpWordsOutOfBounds |
	"c.f. StackInterpreter>>stSizeOf: SpurMemoryManager>>lengthOf:format: fixedFieldsOf:format:length:"
	<var: #done type: #'AbstractInstruction *'>
	<var: #jumpIsBytes type: #'AbstractInstruction *'>
	<var: #jumpIsShorts type: #'AbstractInstruction *'>
	<var: #jumpBadIndex type: #'AbstractInstruction *'>
	<var: #jumpWordsDone type: #'AbstractInstruction *'>
	<var: #jumpWordTooBig type: #'AbstractInstruction *'>
	<var: #jumpNotIndexable type: #'AbstractInstruction *'>
	<var: #jumpBytesOutOfBounds type: #'AbstractInstruction *'>
	<var: #jumpShortsOutOfBounds type: #'AbstractInstruction *'>
	<var: #jumpWordsOutOfBounds type: #'AbstractInstruction *'>

	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	cogit MoveR: Arg0Reg R: Arg1Reg.
	jumpBadIndex := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	self genConvertSmallIntegerToIntegerInReg: Arg1Reg.
	cogit SubCq: 1 R: Arg1Reg. "1-rel => 0-rel"

	self genGetFormatOf: ReceiverResultReg
		into: (formatReg := SendNumArgsReg)
		leastSignificantHalfOfBaseHeaderIntoScratch: NoReg.

	self genGetNumSlotsOf: ReceiverResultReg into: ClassReg.

	"dispatch on format in a combination of highest dynamic frequency order first and convenience.
		  0 = 0 sized objects (UndefinedObject True False et al)
		  1 = non-indexable objects with inst vars (Point et al)
		  2 = indexable objects with no inst vars (Array et al)
		  3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
		  4 = weak indexable objects with inst vars (WeakArray et al)
		  5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
		  6 unused, reserved for exotic pointer objects?
		  7 Forwarded Object, 1st field is pointer, rest of fields are ignored
		  8 unused, reserved for exotic non-pointer objects?
		  9 64-bit indexable
		10 - 11 32-bit indexable
		12 - 15 16-bit indexable
		16 - 23 byte indexable
		24 - 31 compiled method"
	cogit CmpCq: objectMemory firstByteFormat R: formatReg.
	jumpIsBytes := cogit JumpGreaterOrEqual: 0.
					cogit CmpCq: objectMemory firstShortFormat R: formatReg.
	jumpIsShorts := cogit JumpGreaterOrEqual: 0.
					cogit CmpCq: objectMemory firstLongFormat R: formatReg.
	jumpNotIndexable := cogit JumpLess: 0.

	"fall through to double words"
	cogit LogicalShiftLeftCq: objectMemory shiftForWord - 2 R: ClassReg.
	cogit AndCq: objectMemory wordSize / 4 - 1 R: formatReg.
	cogit SubR: formatReg R: ClassReg.
	cogit CmpR: Arg1Reg R: ClassReg.
	jumpWordsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddCq: objectMemory baseHeaderSize >> (objectMemory shiftForWord - 1) R: Arg1Reg.
	cogit MoveX32r: Arg1Reg R: ReceiverResultReg R: TempReg.
	jumpWordTooBig := self jumpNotCharacterUnsignedValueInRegister: TempReg.
	cogit MoveR: TempReg R: ReceiverResultReg.
	jumpWordsDone := cogit Jump: 0.

	jumpIsBytes jmpTarget:
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord R: ClassReg).
	cogit AndCq: objectMemory wordSize - 1 R: formatReg.
	cogit SubR: formatReg R: ClassReg.
	cogit CmpR: Arg1Reg R: ClassReg.
	jumpBytesOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddCq: objectMemory baseHeaderSize R: Arg1Reg.
	cogit MoveXbr: Arg1Reg R: ReceiverResultReg R: ReceiverResultReg.
	cogit backEnd byteReadsZeroExtend ifFalse:
			[cogit AndCq: 255 R: ReceiverResultReg].
	jumpWordsDone jmpTarget: (done := cogit Label).
	self genConvertIntegerToCharacterInReg: ReceiverResultReg.
	cogit genPrimReturn.

	jumpIsShorts jmpTarget:
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord - 1 R: ClassReg).
	cogit AndCq: objectMemory wordSize / 2 - 1 R: formatReg.
	cogit SubR: formatReg R: ClassReg.
	cogit CmpR: Arg1Reg R: ClassReg.
	jumpShortsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit AddR: Arg1Reg R: ReceiverResultReg.
	cogit MoveM16: objectMemory baseHeaderSize r: ReceiverResultReg R: ReceiverResultReg.
	cogit Jump: done.

	jumpWordTooBig jmpTarget:
	(jumpBytesOutOfBounds jmpTarget:
	(jumpShortsOutOfBounds jmpTarget:
	(jumpWordsOutOfBounds jmpTarget:
	(jumpNotIndexable jmpTarget: 
	(jumpBadIndex jmpTarget: cogit Label))))).

	^CompletePrimitive! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/10/2017 15:57'!
genPrimitiveStringAtPut
	| formatReg jumpBadIndex jumpBadArg jumpImmutable
	  jumpIsBytes jumpIsShorts jumpNotString jumpIsCompiledMethod
	  jumpBytesOutOfRange jumpShortsOutOfRange jumpWordsOutOfRange
	  jumpBytesOutOfBounds jumpWordsOutOfBounds jumpShortsOutOfBounds |
	"c.f. StackInterpreter>>stSizeOf: SpurMemoryManager>>lengthOf:format: fixedFieldsOf:format:length:"
	<var: #jumpBadArg type: #'AbstractInstruction *'>
	<var: #jumpIsBytes type: #'AbstractInstruction *'>
	<var: #jumpIsShorts type: #'AbstractInstruction *'>
	<var: #jumpBadIndex type: #'AbstractInstruction *'>
	<var: #jumpBytesOutOfRange type: #'AbstractInstruction *'>
	<var: #jumpShortsOutOfRange type: #'AbstractInstruction *'>
	<var: #jumpWordsOutOfRange type: #'AbstractInstruction *'>
	<var: #jumpBytesOutOfBounds type: #'AbstractInstruction *'>
	<var: #jumpShortsOutOfBounds type: #'AbstractInstruction *'>
	<var: #jumpWordsOutOfBounds type: #'AbstractInstruction *'>

	cogit genLoadArgAtDepth: 1 into: Arg0Reg.
	cogit genLoadArgAtDepth: 0 into: Arg1Reg.

	jumpBadIndex := self genJumpNotSmallInteger: Arg0Reg.
	jumpBadArg := self genJumpNotCharacter: Arg1Reg.
	self genConvertSmallIntegerToIntegerInReg: Arg0Reg.
	cogit SubCq: 1 R: Arg0Reg. "1-rel => 0-rel"

	"formatReg := self formatOf: ReceiverResultReg"
	self cppIf: IMMUTABILITY
		ifTrue:
		[ self genGetFormatOf: ReceiverResultReg
			into: (formatReg := SendNumArgsReg)
			leastSignificantHalfOfBaseHeaderIntoScratch: TempReg.
		jumpImmutable := self genJumpBaseHeaderImmutable: TempReg ]
		ifFalse: 
		[ self genGetFormatOf: ReceiverResultReg
			into: (formatReg := SendNumArgsReg)
			leastSignificantHalfOfBaseHeaderIntoScratch: NoReg ].

	self genGetNumSlotsOf: ReceiverResultReg into: ClassReg.

	"dispatch on format; words and/or bytes.
		  0 to 8 = pointer objects, forwarders, reserved.
		  9 64-bit indexable
		10 - 11 32-bit indexable
		12 - 15 16-bit indexable
		16 - 23 byte indexable
		24 - 31 compiled method"
	cogit CmpCq: objectMemory firstLongFormat R: formatReg.
	jumpNotString := cogit JumpBelow: 0.
					cogit CmpCq: objectMemory firstCompiledMethodFormat R: formatReg.
	jumpIsCompiledMethod := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory firstByteFormat R: formatReg.
	jumpIsBytes := cogit JumpAboveOrEqual: 0.
					cogit CmpCq: objectMemory firstShortFormat R: formatReg.
	jumpIsShorts := cogit JumpAboveOrEqual: 0.

	"fall through to double words"
	cogit CmpCq: (objectMemory characterObjectOf: 1 << self numCharacterBits - 1) R: Arg1Reg.
	jumpWordsOutOfRange := cogit JumpAbove: 0.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord - 2 R: ClassReg.
	cogit AndCq: objectMemory wordSize / 4 - 1 R: formatReg.
	cogit SubR: formatReg R: ClassReg.
	cogit CmpR: Arg0Reg R: ClassReg.
	jumpWordsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertCharacterToCodeInReg: TempReg.
	cogit AddCq: objectMemory baseHeaderSize >> (objectMemory shiftForWord - 1) R: Arg0Reg.
	cogit MoveR: TempReg X32r: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpIsShorts jmpTarget:
		(cogit CmpCq: (objectMemory characterObjectOf: 65535) R: Arg1Reg).
	jumpShortsOutOfRange := cogit JumpAbove: 0.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord - 1 R: ClassReg.
	cogit AndCq: objectMemory wordSize / 2 - 1 R: formatReg.
	cogit SubR: formatReg R: ClassReg.
	cogit CmpR: Arg0Reg R: ClassReg.
	jumpShortsOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertCharacterToCodeInReg: TempReg.
	cogit AddR: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: TempReg M16: objectMemory baseHeaderSize r: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpIsBytes jmpTarget:
		(cogit CmpCq: (objectMemory characterObjectOf: 255) R: Arg1Reg).
	jumpBytesOutOfRange := cogit JumpAbove: 0.
	cogit LogicalShiftLeftCq: objectMemory shiftForWord R: ClassReg.
	cogit AndCq: objectMemory wordSize - 1 R: formatReg.
	cogit SubR: formatReg R: ClassReg.
	cogit CmpR: Arg0Reg R: ClassReg.
	jumpBytesOutOfBounds := cogit JumpBelowOrEqual: 0.
	cogit MoveR: Arg1Reg R: TempReg.
	self genConvertCharacterToCodeInReg: TempReg.
	cogit AddCq: objectMemory baseHeaderSize R: Arg0Reg.
	cogit MoveR: TempReg Xbr: Arg0Reg R: ReceiverResultReg.
	cogit MoveR: Arg1Reg R: ReceiverResultReg.
	cogit genPrimReturn.

	jumpNotString jmpTarget:
	(jumpBytesOutOfRange jmpTarget:
	(jumpShortsOutOfRange jmpTarget:
	(jumpWordsOutOfRange jmpTarget:
	(jumpIsCompiledMethod jmpTarget:
	(jumpBytesOutOfBounds jmpTarget:
	(jumpShortsOutOfBounds jmpTarget:
	(jumpWordsOutOfBounds jmpTarget: cogit Label))))))).

	self cppIf: IMMUTABILITY
		ifTrue: [jumpImmutable jmpTarget: jumpNotString getJmpTarget].

	cogit AddCq: 1 R: Arg0Reg. "0-rel => 1-rel"
	self genConvertIntegerToSmallIntegerInReg: Arg0Reg.

	jumpBadArg jmpTarget:
	(jumpBadIndex jmpTarget: cogit Label).

	^CompletePrimitive! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 8/28/2019 17:37'!
genPureDoubleArithmetic: arithmeticOperator preOpCheck: preOpCheckOrNil
	"In the Pure version, mixed arithmetic with SmallInteger is forbidden"
	<inline: true>
	^self genPureFloatArithmetic: arithmeticOperator preOpCheck: preOpCheckOrNil boxed: true! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 8/28/2019 18:07'!
genPureDoubleComparison: jumpOpcodeGenerator invert: invertComparison
	"In the Pure version, mixed arithmetic with SmallInteger is forbidden"
	<inline: true>
	^self genPureFloatComparison: jumpOpcodeGenerator invert: invertComparison boxed: true! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 8/28/2019 17:37'!
genPureFloatArithmetic: arithmeticOperator preOpCheck: preOpCheckOrNil boxed: rcvrBoxed
	"In the Pure version, mixed arithmetic with SmallInteger is forbidden"
	<var: #preOpCheckOrNil declareC: 'AbstractInstruction *(*preOpCheckOrNil)(int rcvrReg, int argReg)'>
	<inline: false>
	| jumpFailAlloc jumpFailCheck jumpImmediate jumpNotSmallFloat jumpNotBoxedFloat doOp |
	<var: #jumpNotBoxedFloat type: #'AbstractInstruction *'>
	<var: #jumpNotSmallFloat type: #'AbstractInstruction *'>
	<var: #jumpImmediate type: #'AbstractInstruction *'>
	<var: #jumpFailCheck type: #'AbstractInstruction *'>
	<var: #jumpFailAlloc type: #'AbstractInstruction *'>
	<var: #doOp type: #'AbstractInstruction *'>
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	rcvrBoxed
		ifTrue: [self genGetDoubleValueOf: ReceiverResultReg into: DPFPReg0]
		ifFalse: [self genGetSmallFloatValueOf: ReceiverResultReg scratch: TempReg into: DPFPReg0].
	jumpNotSmallFloat := self genJumpNotSmallFloat: Arg0Reg.
	self genGetSmallFloatValueOf: Arg0Reg scratch: TempReg into: DPFPReg1.
	doOp := cogit Label.
	preOpCheckOrNil ifNotNil:
		[jumpFailCheck := cogit perform: preOpCheckOrNil with: DPFPReg0 with: DPFPReg1].
	cogit gen: arithmeticOperator operand: DPFPReg1 operand: DPFPReg0.
	jumpFailAlloc := self genAllocFloatValue: DPFPReg0 into: SendNumArgsReg scratchReg: ClassReg scratchReg: TempReg.
	cogit MoveR: SendNumArgsReg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpNotSmallFloat jmpTarget: cogit Label.
	jumpImmediate := self genJumpImmediate: Arg0Reg.
	self genGetCompactClassIndexNonImmOf: Arg0Reg into: SendNumArgsReg.
	self genCmpClassFloatCompactIndexR: SendNumArgsReg.
	jumpNotBoxedFloat := cogit JumpNonZero: 0.
	self genGetDoubleValueOf: Arg0Reg into: DPFPReg1.
	cogit Jump: doOp.
	jumpImmediate jmpTarget:
	(jumpNotBoxedFloat jmpTarget:
	(jumpFailAlloc jmpTarget: cogit Label)).
	preOpCheckOrNil ifNotNil:
		[jumpFailCheck jmpTarget: jumpFailAlloc getJmpTarget].
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 8/28/2019 18:06'!
genPureFloatComparison: jumpFPOpcodeGenerator invert: invertComparison boxed: rcvrBoxed
	"In the Pure version, mixed arithmetic with SmallInteger is forbidden"
	<var: #jumpFPOpcodeGenerator declareC: 'AbstractInstruction *(*jumpFPOpcodeGenerator)(void *)'>
	<inline: false>
	| jumpImmediate jumpNotSmallFloat jumpNotBoxedFloat jumpCond compareFloat |
	<var: #jumpNotBoxedFloat type: #'AbstractInstruction *'>
	<var: #jumpNotSmallFloat type: #'AbstractInstruction *'>
	<var: #jumpImmediate type: #'AbstractInstruction *'>
	<var: #jumpCond type: #'AbstractInstruction *'>
	<var: #compareFloat type: #'AbstractInstruction *'>
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	rcvrBoxed
		ifTrue: [self genGetDoubleValueOf: ReceiverResultReg into: DPFPReg0]
		ifFalse: [self genGetSmallFloatValueOf: ReceiverResultReg scratch: TempReg into: DPFPReg0].
	jumpNotSmallFloat := self genJumpNotSmallFloat: Arg0Reg.
	self genGetSmallFloatValueOf: Arg0Reg scratch: TempReg into: DPFPReg1.
	compareFloat := invertComparison "May need to invert for NaNs"
					ifTrue: [cogit CmpRd: DPFPReg0 Rd: DPFPReg1]
					ifFalse: [cogit CmpRd: DPFPReg1 Rd: DPFPReg0].
	jumpCond := cogit perform: jumpFPOpcodeGenerator with: 0. "FP jumps are a little weird"
	cogit genMoveFalseR: ReceiverResultReg.
	cogit genPrimReturn.
	jumpCond jmpTarget: (cogit genMoveTrueR: ReceiverResultReg).
	cogit genPrimReturn.
	
	"not a Small Float, check for Boxed Float argument"
	jumpNotSmallFloat jmpTarget: cogit Label.
	jumpImmediate := self genJumpImmediate: Arg0Reg.
	self genGetCompactClassIndexNonImmOf: Arg0Reg into: SendNumArgsReg.
	self genCmpClassFloatCompactIndexR: SendNumArgsReg.
	jumpNotBoxedFloat := cogit JumpNonZero: 0.
	self genGetDoubleValueOf: Arg0Reg into: DPFPReg1.
	cogit Jump: compareFloat.
	
	jumpImmediate jmpTarget:
	(jumpNotBoxedFloat jmpTarget: cogit Label).
	^CompletePrimitive! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 8/28/2019 17:37'!
genPureSmallFloatArithmetic: arithmeticOperator preOpCheck: preOpCheckOrNil
	"In the Pure version, mixed arithmetic with SmallInteger is forbidden"
	<inline: true>
	^self genPureFloatArithmetic: arithmeticOperator preOpCheck: preOpCheckOrNil boxed: false! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 8/28/2019 18:20'!
genPureSmallFloatComparison: jumpOpcodeGenerator invert: invertComparison
	"In the Pure version, mixed arithmetic with SmallInteger is forbidden"
	<inline: true>
	^self genPureFloatComparison: jumpOpcodeGenerator invert: invertComparison boxed: false! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/23/2015 19:35:09'!
genRemoveSmallIntegerTagsInScratchReg: scratchReg
	cogit SubCq: 1 R: scratchReg.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/12/2015 11:38'!
genShiftAwaySmallIntegerTagsInScratchReg: scratchReg
	cogit ArithmeticShiftRightCq: objectMemory numTagBits R: scratchReg.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/16/2016 20:28'!
genSmallFloatArithmetic: arithmeticOperator preOpCheck: preOpCheckOrNil
	<inline: true>
	^self genFloatArithmetic: arithmeticOperator preOpCheck: preOpCheckOrNil boxed: false! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'nice 8/21/2019 01:08'!
genSmallFloatComparison: jumpOpcodeGenerator orIntegerComparison: jumpOpcode invert: invertComparison
	<inline: true>
	^self genFloatComparison: jumpOpcodeGenerator orIntegerComparison: jumpOpcode invert: invertComparison boxed: false! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 1/29/2020 15:01'!
genSmallIntegerComparison: jumpOpcode orDoubleComparison: jumpFPOpcodeGenerator invert: invertComparison
	"Stack looks like
		return address"
	| jumpCond r compareIntFloat jumpAmbiguous jumpNotBoxedFloat jumpNotFloatAtAll jumpNotSmallFloat jumpTrue returnTrue |
	<var: #jumpFPOpcodeGenerator declareC: 'AbstractInstruction * NoDbgRegParms (*jumpFPOpcodeGenerator)(void *)'>
	<var: 'jumpCond' type: #'AbstractInstruction *'>
	r := self genSmallIntegerComparison: jumpOpcode.
	r < 0 ifTrue:
		[^r].
	self cppIf: #DPFPReg0 defined ifTrue:
	"Fall through on non-SmallInteger argument.  Argument may be a Float : let us check or fail"
	[
	"check for Small Float argument"
	jumpNotSmallFloat := self genJumpNotSmallFloat: Arg0Reg.
	self genGetSmallFloatValueOf: Arg0Reg scratch: TempReg into: DPFPReg1.
	
	"Case of (int compare: float). Test for ambiguity, that is when (double) intRcvr == floatArg"
	compareIntFloat := cogit Label.
	self genConvertSmallIntegerToIntegerInReg: ReceiverResultReg.
	cogit ConvertR: ReceiverResultReg Rd: DPFPReg0.
	cogit CmpRd: DPFPReg0 Rd: DPFPReg1.
	jumpAmbiguous := cogit JumpFPEqual: 0.
	"Case of non ambiguity, use compareFloat((double) intRcvr,floatArg)"
	invertComparison "May need to invert for NaNs"
					ifTrue: [cogit CmpRd: DPFPReg0 Rd: DPFPReg1]
					ifFalse: [cogit CmpRd: DPFPReg1 Rd: DPFPReg0].
	jumpCond := cogit perform: jumpFPOpcodeGenerator with: 0. "FP jumps are a little weird"
	cogit genMoveFalseR: ReceiverResultReg.
	cogit genPrimReturn.
	jumpCond jmpTarget: (returnTrue := cogit genMoveTrueR: ReceiverResultReg).
	cogit genPrimReturn.
	"Case of ambiguity, use compareInt(intRcvr , (int64) floatArg)"
	jumpAmbiguous jmpTarget: (cogit ConvertRd: DPFPReg1 R: Arg0Reg).
	cogit CmpR: Arg0Reg R: ReceiverResultReg. "N.B. FLAGS := RRReg - Arg0Reg"
	jumpTrue := cogit genConditionalBranch: jumpOpcode operand: 0.
	cogit genMoveFalseR: ReceiverResultReg.
	cogit genPrimReturn.
	jumpTrue jmpTarget: returnTrue.
					
	"not a Small Float, check for Boxed Float argument"
	jumpNotSmallFloat jmpTarget:cogit Label.
	jumpNotFloatAtAll := self genJumpImmediate: Arg0Reg.
	self genGetCompactClassIndexNonImmOf: Arg0Reg into: SendNumArgsReg.
	self genCmpClassFloatCompactIndexR: SendNumArgsReg.
	jumpNotBoxedFloat := cogit JumpNonZero: 0.
	"It was a Boxed Float, so convert the receiver to double and perform the (int compare: float) operation"
	self genGetDoubleValueOf: Arg0Reg into: DPFPReg1.
	cogit Jump: compareIntFloat.
	
	"not a Float, just let the primitive fall thru failure"
	jumpNotBoxedFloat jmpTarget: (jumpNotFloatAtAll jmpTarget: cogit Label)].
	^CompletePrimitive! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'initialization' stamp: 'eem 12/7/2015 16:14'!
genStoreHeader: header intoNewInstance: rcvrReg using: scratchReg
	"Generate the instructions to move the constant header into a new instance pointed to by rcvrReg."
	<inline: true>
	cogit
		MoveCq: header R: scratchReg;
		MoveR: scratchReg Mw: 0 r: rcvrReg! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 1/23/2016 12:28'!
getLiteralCountOf: methodReg plusOne: plusOne inBytes: inBytes into: litCountReg scratch: scratchReg
	"Get the literal count of a CompiledMethod into headerReg, plus one if requested.
	  If inBytes is true, scale the count by the word size.  Deal with the possibility of
	 the method being cogged."
	self genGetMethodHeaderOf: methodReg into: litCountReg scratch: scratchReg.
	self assert: 1 << objectMemory numTagBits = objectMemory wordSize.
	inBytes
		ifTrue:
			[cogit
				AndCq: coInterpreter alternateHeaderNumLiteralsMask << objectMemory numTagBits
				R: litCountReg]
		ifFalse:
			[cogit
				LogicalShiftRightCq: objectMemory numTagBits R: litCountReg;
				AndCq: coInterpreter alternateHeaderNumLiteralsMask R: litCountReg].
	plusOne ifTrue:
		[cogit
			AddCq: (inBytes
						ifTrue: [LiteralStart * objectMemory wordSize]
						ifFalse: [LiteralStart])
			R: litCountReg].
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'in-line cacheing' stamp: 'eem 01/10/2016 10:01'!
illegalClassTag
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'in-line cacheing' stamp: 'eem 8/7/2014 04:48'!
inlineCacheTagForClass: classOop
	"Answer the relevant inline cache tag for a class.
	 c.f. inlineCacheTagForInstance:"
	^objectMemory classTagForClass: classOop! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'in-line cacheing' stamp: 'eem 7/7/2014 09:56'!
inlineCacheTagForInstance: oop
	"Answer the relevant inline cache tag for an instance.
	 c.f. getInlineCacheClassTagFrom:into: & inlineCacheTagForClass:"
	^(objectMemory isImmediate: oop)
		ifTrue: [oop bitAnd: objectMemory tagMask]
		ifFalse: [objectMemory classIndexOf: oop]! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'initialization' stamp: 'eem 5/31/2014 17:43'!
log2BytesPerWord
	^3! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'garbage collection' stamp: 'eem 11/21/2015 11:39'!
markAndTraceCacheTagLiteral: literal in: cogMethodOrNil atpc: address
	"Mark and trace a literal in an inline cache preceding address in cogMethodOrNil.
	 Answer if code was modified.  In 64-bit Spur, cache tags are either selector
	 indices or class indices and so this is a noop."
	<var: #cogMethodOrNil type: #'CogMethod *'>
	<var: #address type: #usqInt>
	<inline: true>
	^false! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'primitive generators' stamp: 'eem 3/10/2016 10:30'!
maybeGenConvertIfSmallFloatIn: oopReg scratchReg: scratch into: dpReg andJumpTo: targetInst
	"Generate a test for a smallFloat in oopReg, converting it to the float value in dpReg
	 and jumping to targetInst. If oopReg does not contain a SmallFloat, fall through."
	<var: 'targetInst' type: #'AbstractInstruction *'>
	| jumpNotSF |
	<var: 'jumpNotSF' type: #'AbstractInstruction *'>
	jumpNotSF := self genJumpNotSmallFloat: oopReg.
	self genGetSmallFloatValueOf: oopReg scratch: scratch into: dpReg.
	cogit Jump: targetInst.
	jumpNotSF jmpTarget: cogit Label.
	^0! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'initialization' stamp: 'eem 3/11/2020 11:30'!
maybeGenerateSelectorIndexDereferenceRoutine
	"Generate the routine that converts selector indices into selector objects.
	 It is called from the send trampolines.
	 If the selector index is negative, convert it into a positive index into the
	 special selectors array and index that.  Otherwise, index the current method.
	 The routine uses Extra0Reg & Extra1Reg, which are available, since they
	 are not live at point of send."
	| jumpNegative jumpNotBlock jumpFullBlock |
	<var: 'jumpNegative' type: #'AbstractInstruction *'>
	<var: 'jumpNotBlock' type: #'AbstractInstruction *'>
	<var: 'jumpFullBlock' type: #'AbstractInstruction *'>
	cogit zeroOpcodeIndex.
	cogit CmpCq: 0 R: ClassReg.
	jumpNegative := cogit JumpLess: 0.
	cogit
		MoveMw: FoxMethod r: FPReg R: Extra0Reg;
		AddCq: 2 R: ClassReg; "Change selector index to 1-relative, skipping the method header"
		TstCq: MFMethodFlagIsBlockFlag R: Extra0Reg.
	jumpNotBlock := cogit JumpZero: 0.
	"If in a block, need to find the home method...  If using full blocks, need to test the cpicHasMNUCaseOrCMIsFullBlock bit"
	cogit AndCq: methodZone zoneAlignment negated R: Extra0Reg.
	SistaV1BytecodeSet ifTrue:
		[self bitAndByteOffsetOfIsFullBlockBitInto:
			[:bitmask :byteOffset|
			jumpFullBlock := cogit
				MoveMb: byteOffset r: Extra0Reg R: Extra1Reg;
				TstCq: bitmask R: Extra1Reg;
				JumpNonZero: 0]].
	cogit 
		MoveM16: 0 r: Extra0Reg R: Extra1Reg;
		SubR: Extra1Reg R: Extra0Reg.
	jumpNotBlock jmpTarget: cogit Label.
	SistaV1BytecodeSet ifTrue:
		[jumpFullBlock jmpTarget: jumpNotBlock getJmpTarget].
	cogit "Now fetch the method object and index with the literal index to retrieve the selector"
		AndCq: methodZone zoneAlignment negated R: Extra0Reg;
		MoveMw: (cogit offset: CogMethod of: #methodObject) r: Extra0Reg R: Extra1Reg;
		MoveXwr: ClassReg R: Extra1Reg R: ClassReg;
		RetN: 0.
	jumpNegative jmpTarget: cogit Label.
	cogit
		NegateR: ClassReg;
		LogicalShiftLeftCq: 1 R: ClassReg;
		MoveAw: objectMemory specialObjectsOopAddress R: Extra0Reg;
		SubCq: 1 R: ClassReg;
		MoveMw: SpecialSelectors + 1 * objectMemory wordSize r: Extra0Reg R: Extra1Reg; "Index, including header size"
		MoveXwr: ClassReg R: Extra1Reg R: ClassReg;
		RetN: 0.
	ceDereferenceSelectorIndex := cogit methodZoneBase.
	cogit
		outputInstructionsForGeneratedRuntimeAt: ceDereferenceSelectorIndex;
		recordGeneratedRunTime: 'ceDereferenceSelectorIndex' address: ceDereferenceSelectorIndex;
		recordRunTimeObjectReferences! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'sista support' stamp: 'eem 7/5/2014 17:54'!
numCountersFor: theCounters
	<var: #theCounters type: #usqInt>
	| objOop |
	theCounters = 0 ifTrue:
		[^0].
	objOop := theCounters - objectMemory baseHeaderSize.
	^2 * (objectMemory numSlotsOf: objOop)! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'compile abstract instructions' stamp: 'eem 9/27/2015 10:36'!
numSmallIntegerBits
	^61! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'object representation' stamp: 'eem 10/22/2015 16:59:03'!
numSmallIntegerTagBits
	^3! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'calling convention' stamp: 'eem 01/10/2016 10:01'!
selectorIndexDereferenceRoutine
	"If the object representation is 64-bits, selectors must be accessed via indices in the inline caches.
	 Answer the routine that will map from indices to selector objects if so, otherwise nil."
	^ceDereferenceSelectorIndex! !

!CogObjectRepresentationFor64BitSpur methodsFor: 'debug support' stamp: 'eem 12/5/2014 18:53'!
validInlineCacheTag: classIndexOrTagPattern
	"The three valid tag patterns are 1 (SmallInteger), 2 (Character) and 3 (SmallFloat64)."
	^(classIndexOrTagPattern >= 1
	  and: [classIndexOrTagPattern <= 3])
	  or: [(objectMemory classAtIndex: classIndexOrTagPattern) notNil]! !

!CogObjectRepresentationForSpur class methodsFor: 'translation' stamp: 'eem 11/10/2020 18:52'!
declareCVarsIn: aCodeGen
	aCodeGen
		var: #ceStoreTrampolines
			declareC: '#if IMMUTABILITY\sqInt ceStoreTrampolines[', NumStoreTrampolines printString, '];\#endif'! !

!CogObjectRepresentationForSpur class methodsFor: 'class initialization' stamp: 'eem 1/1/1970 11:49'!
initialize
	"Number of trampolines for the combined IMMUTABILITY/store check instance variable store.
	 Trampolines from 0 to NumStoreTrampolines - 2 are dedicated to an inst var index, the last one is generic."
	NumStoreTrampolines := 5! !

!CogObjectRepresentationForSpur class methodsFor: 'initialization' stamp: 'eem 6/7/2018 15:16'!
initializeMiscConstants
	CheckRememberedInTrampoline := InitializationOptions at: #CheckRememberedInTrampoline ifAbsent: [false]! !

!CogObjectRepresentationForSpur class methodsFor: 'trampoline support' stamp: 'eem 11/10/2020 19:29'!
numTrampolines
	^super numTrampolines
		 + (SistaV1BytecodeSet
			ifTrue: [9] "(small,large)x(method,block,fullBlock) context creation,
						 ceNewHashTrampoline, ceStoreCheckContextReceiverTrampoline and ceScheduleScavengeTrampoline"
			ifFalse: [7] "(small,large)x(method,block) context creation, 
						 ceNewHashTrampoline, ceStoreCheckContextReceiverTrampoline and ceScheduleScavengeTrampoline")
		 + (IMMUTABILITY
			ifTrue: [NumStoreTrampolines]
			ifFalse: [0])
		 + (SistaVM
			ifTrue: [1] "inline newHash"
			ifFalse: [0])! !

!CogObjectRepresentationForSpur methodsFor: 'garbage collection' stamp: 'eem 11/18/2013 10:57'!
allYoungObjectsAgeInFullGC
	^false! !

!CogObjectRepresentationForSpur methodsFor: 'newspeak support' stamp: 'eem 7/9/2014 21:23'!
allocateNPinnedSlots: nSlots
	"On Newspeak allocate the implicit receiver caches on the heap as pinned objects."
	<inline: true>
	<option: #NewspeakVM>
	| objOop |
	objOop := objectMemory allocatePinnedSlots: nSlots.
	^objOop
		ifNil: [0]
		ifNotNil: [objOop + objectMemory baseHeaderSize]! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'cb 1/23/2017 10:11'!
branchIf: reg instanceOfBehavior: classObj target: targetFixUp
	"Generate a branch if reg is an instance of classObj, otherwise fall-
	 through. Cannot change the value of reg (may be used afterwards)."
	| classIndex jmp |
	<inline: true>
	<var: #targetFixUp type: #'AbstractInstruction *'>
	<var: #jmp type: #'AbstractInstruction *'>
	classIndex := objectMemory classTagForClass: classObj.
	classIndex = (objectMemory fetchClassTagOf: objectMemory falseObject)
		ifTrue: [ self branchIf: reg isOop: objectMemory falseObject target: targetFixUp ].
	classIndex = (objectMemory fetchClassTagOf: objectMemory trueObject)
		ifTrue: [ self branchIf: reg isOop: objectMemory trueObject target: targetFixUp ].
	classIndex = (objectMemory fetchClassTagOf: objectMemory nilObject)
		ifTrue: [ self branchIf: reg isOop: objectMemory nilObject target: targetFixUp ].
	(objectMemory isImmediateClass: classObj)
		ifTrue:
			[self branchIf: reg hasImmediateTag: classIndex target: targetFixUp ]
		ifFalse:
			[jmp := (self genJumpImmediate: reg) .
			 self genGetClassIndexOfNonImm: reg into: TempReg.
			 self genCmpClassIndex: classIndex R: TempReg.
			 cogit JumpZero: targetFixUp.
			jmp jmpTarget: cogit Label ].
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'cb 12/11/2016 16:24'!
branchIf: reg instanceOfBehaviors: arrayObj target: targetFixUp
	"Generate a branch if reg is an instance of any of the classes in arrayObj,
	 otherwise fall-through. reg should not be edited."
	
	| allImmediate noneImmediate immediateMask numNonImmediates classObj |
	<var: #targetFixUp type: #'AbstractInstruction *'>
	
	"let me tell you all about it, let me falsify"
	allImmediate := true. noneImmediate := true. immediateMask := 0. numNonImmediates := 0.
	0 to: (objectMemory numSlotsOf: arrayObj) - 1 do:
		[:i|
		 classObj := objectMemory fetchPointer: i ofObject: arrayObj.
		 (objectMemory isImmediateClass: classObj)
			ifTrue:
				[noneImmediate := false.
				 immediateMask := immediateMask + (objectMemory classTagForClass: classObj)]
			ifFalse:
				[allImmediate := false.
				 numNonImmediates := numNonImmediates + 1]].

	noneImmediate ifTrue: [ ^ self noneImmediateBranchIf: reg instanceOfBehaviors: arrayObj target: targetFixUp ].

	allImmediate ifTrue: [ ^ self allImmediate: immediateMask branchIf: reg instanceOfBehaviors: arrayObj target: targetFixUp ].

	^ self mixed: numNonImmediates branchIf: reg instanceOfBehaviors: arrayObj target: targetFixUp! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'eem 2/5/2017 13:22'!
branchIf: reg isNotOop:  oop target: targetFixUp
	<var: #targetFixUp type: #'AbstractInstruction *'>
	<inline: true>
	cogit CmpCq: oop R: reg.
	cogit JumpNonZero: targetFixUp.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'eem 2/5/2017 13:22'!
branchIf: reg isOop:  oop target: targetFixUp
	<var: #targetFixUp type: #'AbstractInstruction *'>
	<inline: true>
	cogit CmpCq: oop R: reg.
	cogit JumpZero: targetFixUp.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'cb 1/23/2017 10:12'!
branchIf: reg notInstanceOfBehavior: classObj target: targetFixUp
	"Generate a branch if reg is an instance of classObj, otherwise fall-
	 through. Cannot change the value of reg (may be used afterwards)."
	| classIndex |
	<inline: true>
	<var: #targetFixUp type: #'AbstractInstruction *'>
	classIndex := objectMemory classTagForClass: classObj.
	classIndex = (objectMemory fetchClassTagOf: objectMemory falseObject)
		ifTrue: [ self branchIf: reg isNotOop: objectMemory falseObject target: targetFixUp ].
	classIndex = (objectMemory fetchClassTagOf: objectMemory trueObject)
		ifTrue: [ self branchIf: reg isNotOop:  objectMemory trueObject target: targetFixUp ].
	classIndex = (objectMemory fetchClassTagOf: objectMemory nilObject)
		ifTrue: [ self branchIf: reg isNotOop:  objectMemory nilObject target: targetFixUp ].
	(objectMemory isImmediateClass: classObj)
		ifTrue:
			[self branchIf: reg hasNotImmediateTag: classIndex target: targetFixUp ]
		ifFalse:
			[(self genJumpImmediate: reg) jmpTarget: targetFixUp.
			 self genGetClassIndexOfNonImm: reg into: TempReg.
			 self genCmpClassIndex: classIndex R: TempReg.
			 cogit JumpNonZero: targetFixUp ].
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'cb 4/28/2015 09:57'!
branchIf: reg notInstanceOfBehaviors: arrayObj target: targetFixUp
	"Generate a branch if reg is an instance of any of the classes in arrayObj,
	 otherwise fall-through. reg should not be edited."
	
	| allImmediate noneImmediate immediateMask numNonImmediates classObj |
	<var: #targetFixUp type: #'AbstractInstruction *'>
	
	"let me tell you all about it, let me falsify"
	allImmediate := true. noneImmediate := true. immediateMask := 0. numNonImmediates := 0.
	0 to: (objectMemory numSlotsOf: arrayObj) - 1 do:
		[:i|
		 classObj := objectMemory fetchPointer: i ofObject: arrayObj.
		 (objectMemory isImmediateClass: classObj)
			ifTrue:
				[noneImmediate := false.
				 immediateMask := immediateMask + (objectMemory classTagForClass: classObj)]
			ifFalse:
				[allImmediate := false.
				 numNonImmediates := numNonImmediates + 1]].

	noneImmediate ifTrue: [ ^ self noneImmediateBranchIf: reg notInstanceOfBehaviors: arrayObj target: targetFixUp ].

	allImmediate ifTrue: [ ^ self allImmediate: immediateMask branchIf: reg notInstanceOfBehaviors: arrayObj target: targetFixUp ].

	^ self mixed: numNonImmediates branchIf: reg notInstanceOfBehaviors: arrayObj target: targetFixUp! !

!CogObjectRepresentationForSpur methodsFor: 'garbage collection' stamp: 'eem 11/18/2013 13:30'!
cacheTagIsMarked: cacheTag
	"Answer if the cacheTag is not unmarked, i.e. answer true for compact class
	 indices and immediates; only answer false for unmarked objects.  In Spur
	 linked send cache tags are class indices so effectively they're always marked."
	^true! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'cb 4/10/2017 15:47'!
callStoreCheckTrampoline
	cogit CallRT: ceStoreCheckTrampoline! !

!CogObjectRepresentationForSpur methodsFor: 'testing' stamp: 'eem 8/3/2021 18:54'!
canPinObjects
	"Answer if the memory manager supports pinned objects."
	<inline: #always>
	^true! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'cb 4/10/2017 15:44'!
checkRememberedInTrampoline
	<inline: true>
	^ CheckRememberedInTrampoline! !

!CogObjectRepresentationForSpur methodsFor: 'debug support' stamp: 'eem 1/26/2017 11:05'!
checkValidDerivedObjectReference: bodyAddress
	^(objectMemory heapMapAtWord: (self pointerForOop: bodyAddress - objectMemory baseHeaderSize)) ~= 0! !

!CogObjectRepresentationForSpur methodsFor: 'debug support' stamp: 'eem 5/9/2014 09:59'!
checkValidOopReference: anOop
	^(objectMemory isImmediate: anOop)
	   or: [(objectMemory heapMapAtWord: (self pointerForOop: anOop)) ~= 0]! !

!CogObjectRepresentationForSpur methodsFor: 'garbage collection' stamp: 'eem 1/25/2017 16:23'!
couldBeDerivedObject: bodyAddress
	^self oop: bodyAddress - objectMemory baseHeaderSize isGreaterThanOrEqualTo: objectMemory startOfMemory! !

!CogObjectRepresentationForSpur methodsFor: 'garbage collection' stamp: 'eem 12/6/2013 18:41'!
couldBeObject: literal
	^(objectMemory isNonImmediate: literal)
	  and: [self oop: literal isGreaterThanOrEqualTo: objectMemory startOfMemory]! !

!CogObjectRepresentationForSpur methodsFor: 'bytecode generator support' stamp: 'eem 8/3/2021 18:53'!
createsArraysInline
	"Answer if the object representation allocates arrays inline.  By
	 default answer false. Better code can be generated when creating
	 arrays inline if values are /not/ flushed to the stack."
	<inline: #always>
	^true! !

!CogObjectRepresentationForSpur methodsFor: 'bytecode generator support' stamp: 'eem 8/3/2021 18:53'!
createsClosuresInline
	"Answer if the object representation allocates closures inline.  By
	 default answer false. Better code can be generated when creating
	 closures inline if copied values are /not/ flushed to the stack."
	<inline: #always>
	^true! !

!CogObjectRepresentationForSpur methodsFor: 'initialization' stamp: 'eem 7/10/2014 17:09'!
ensureNoForwardedLiteralsIn: aMethodObj
	"Ensure there are no forwarded literals in the argument."
	<doNotGenerate>
	objectMemory ensureNoForwardedLiteralsIn: aMethodObj! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'eem 4/19/2016 14:23'!
freeCounters: theCounters
	<var: #theCounters type: #usqInt>
	<inline: true>
	<option: #SistaCogit>
	theCounters ~= 0 ifTrue:
		[objectMemory freeObject: theCounters - objectMemory baseHeaderSize]! !

!CogObjectRepresentationForSpur methodsFor: 'newspeak support' stamp: 'eem 7/9/2014 22:39'!
freeIRCs: maybeIRCs
	<var: #maybeIRCs type: #usqInt>
	<option: #NewspeakVM>
	<inline: true>
	(self oop: maybeIRCs isGreaterThan: objectMemory nilObject) ifTrue:
		[objectMemory freeObject: maybeIRCs - objectMemory baseHeaderSize]! !

!CogObjectRepresentationForSpur methodsFor: 'initialization' stamp: 'eem 6/7/2014 14:57'!
genActiveContextTrampolineLarge: isLarge inBlock: isInBlock called: aString
	<var: #aString type: #'char *'>
	"Create a trampoline to answer the active context that will
	 answer it if a frame is already married, and create it otherwise.
	 Assume numArgs is in SendNumArgsReg and ClassReg is free."
	| startAddress |
	startAddress := cogit methodZoneBase.
	cogit zeroOpcodeIndex.
	self genGetActiveContextLarge: isLarge inBlock: isInBlock.
	cogit outputInstructionsForGeneratedRuntimeAt: startAddress.
	cogit recordGeneratedRunTime: aString address: startAddress.
	cogit recordRunTimeObjectReferences.
	^startAddress! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'eem 12/7/2015 16:12'!
genAllocFloatValue: dpreg into: resultReg scratchReg: scratch1 scratchReg: scratch2
	<returnTypeC: #'AbstractInstruction *'>
	| allocSize newFloatHeader jumpFail |
	<var: #jumpFail type: #'AbstractInstruction *'>
	allocSize := objectMemory baseHeaderSize + (objectMemory sizeof: #double).
	newFloatHeader := objectMemory
							headerForSlots: (self sizeof: #double) / objectMemory wordSize
							format: objectMemory firstLongFormat
							classIndex: ClassFloatCompactIndex.
	cogit MoveAw: objectMemory freeStartAddress R: resultReg.
	cogit LoadEffectiveAddressMw: allocSize r: resultReg R: scratch1.
	cogit CmpCq: objectMemory getScavengeThreshold R: scratch1.
	jumpFail := cogit JumpAboveOrEqual: 0.
	cogit MoveR: scratch1 Aw: objectMemory freeStartAddress.
	self genStoreHeader: newFloatHeader intoNewInstance: resultReg using: scratch1.
	cogit MoveRd: dpreg M64: objectMemory baseHeaderSize r: resultReg.
	^jumpFail! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 1/1/1970 12:53'!
genCheckRememberedBitOf: objReg scratch: scratchReg
	"Check the remembered bit of the object in objReg; answer the jump taken if the bit is already set.
	 Only need to fetch the byte containing it, which reduces the size of the mask constant."
	| rememberedBitByteOffset mask |
	rememberedBitByteOffset := cogit backEnd isBigEndian
									ifTrue: [objectMemory baseHeaderSize - 1 - (objectMemory rememberedBitShift // 8)]
									ifFalse:[objectMemory rememberedBitShift // 8].
	mask := 1 << (objectMemory rememberedBitShift \\ 8).
	cogit MoveMb: rememberedBitByteOffset r: objReg R: scratchReg.
	cogit TstCq: mask R: scratchReg.
	^cogit JumpNonZero: 0! !

!CogObjectRepresentationForSpur methodsFor: 'in-line cacheing' stamp: 'eem 10/26/2016 11:47'!
genCmpClassIndex: classIndex R: reg
	"It is safe to use a short comparison for the known classes; these will not
	 change with become, etc... But it's probably not safe to assume the hash of
	 some other class won't change over time, so to be sure of generating the
	 same size code over time, use a long comparison for unknown classes."
	classIndex < objectMemory classTablePageSize
		ifTrue: [cogit CmpCq: classIndex R: TempReg]
		ifFalse: [cogit CmpCw: classIndex R: TempReg]! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/9/2013 18:35'!
genConvertCharacterToCodeInReg: reg 
	cogit LogicalShiftRightCq: objectMemory numTagBits R: reg.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/6/2013 14:49'!
genConvertIntegerToCharacterInReg: reg
	cogit
		LogicalShiftLeftCq: objectMemory numTagBits R: reg;
		AddCq: objectMemory characterTag R: reg.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/24/2013 18:31'!
genConvertSmallIntegerToCharacterInReg: reg
	self subclassResponsibility! !

!CogObjectRepresentationForSpur methodsFor: 'bytecode generator support' stamp: 'eem (const to msg rewrite) 11/6/2014 10:37'!
genCreateClosureAt: bcpc numArgs: numArgs numCopied: numCopied contextNumArgs: ctxtNumArgs large: isLargeCtxt inBlock: isInBlock
	"Create a closure with the given startpc, numArgs and numCopied
	 within a context with ctxtNumArgs, large if isLargeCtxt that is in a
	 block if isInBlock.  If numCopied > 0 pop those values off the stack."
	self genNoPopCreateClosureAt: bcpc
		numArgs: numArgs
		numCopied: numCopied
		contextNumArgs: ctxtNumArgs
		large: isLargeCtxt
		inBlock: isInBlock.
	1 to: numCopied do:
		[:i|
		cogit
			PopR: TempReg;
			MoveR: TempReg
				Mw: numCopied - i + ClosureFirstCopiedValueIndex * objectMemory bytesPerOop + objectMemory baseHeaderSize
					r: ReceiverResultReg].
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'bytecode generator support' stamp: 'cb 4/8/2016 15:40'!
genCreateFullClosure: compiledBlock numArgs: numArgs numCopied: numCopied ignoreContext: ignoreContext contextNumArgs: contextNumArgs large: contextIsLarge inBlock: contextIsBlock
	"Create a full closure with the given values."
	| numSlots byteSize header skip |
	<var: #skip type: #'AbstractInstruction *'>

	"First get thisContext into ReceiverResultReg and thence in ClassReg."
	ignoreContext
		ifTrue: [ cogit genMoveNilR: ClassReg ]
		ifFalse: 
			[self genGetActiveContextNumArgs: contextNumArgs large: contextIsLarge inBlock: contextIsBlock.
			cogit MoveR: ReceiverResultReg R: ClassReg ].
	
	numSlots := FullClosureFirstCopiedValueIndex + numCopied.
	byteSize := objectMemory smallObjectBytesForSlots: numSlots.
	self assert: ClassFullBlockClosureCompactIndex ~= 0.
	header := objectMemory
					headerForSlots: numSlots
					format: objectMemory indexablePointersFormat
					classIndex: ClassFullBlockClosureCompactIndex.
	cogit MoveAw: objectMemory freeStartAddress R: ReceiverResultReg.
	self genStoreHeader: header intoNewInstance: ReceiverResultReg using: TempReg.
	cogit
		LoadEffectiveAddressMw: byteSize r: ReceiverResultReg R: TempReg;
		MoveR: TempReg Aw: objectMemory freeStartAddress;
		CmpCq: objectMemory getScavengeThreshold R: TempReg.
	skip := cogit JumpBelow: 0.
	cogit CallRT: ceScheduleScavengeTrampoline.
	skip jmpTarget: cogit Label.

	cogit
		MoveR: ClassReg Mw: ClosureOuterContextIndex * objectMemory bytesPerOop + objectMemory baseHeaderSize r: ReceiverResultReg;
		genMoveConstant: compiledBlock R: TempReg;
		MoveR: TempReg Mw: ClosureStartPCIndex * objectMemory bytesPerOop + objectMemory baseHeaderSize r: ReceiverResultReg;
		MoveCq: (objectMemory integerObjectOf: numArgs) R: TempReg;
		MoveR: TempReg Mw: ClosureNumArgsIndex * objectMemory bytesPerOop + objectMemory baseHeaderSize r: ReceiverResultReg.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 6/4/2014 16:48'!
genEnsureObjInRegNotForwarded: reg scratchReg: scratch
	"Make sure that the object in reg is not forwarded.  This routine assumes the object will
	 never be forwarded to an immediate, as it is used to unforward  literal variables (associations). 
	 Use the fact that isForwardedObjectClassIndexPun is a power of two to save an instruction."
	| loop ok |
	<var: #ok type: #'AbstractInstruction *'>
	<var: #loop type: #'AbstractInstruction *'>
	self assert: reg ~= scratch.
	loop := cogit Label.
	"notionally
		self genGetClassIndexOfNonImm: reg into: scratch.
		cogit CmpCq: objectMemory isForwardedObjectClassIndexPun R: TempReg.
	 but the following is an instruction shorter:"
	cogit MoveMw: 0 r: reg R: scratch.
	cogit
		AndCq: objectMemory classIndexMask - objectMemory isForwardedObjectClassIndexPun
		R: scratch.
	ok := cogit JumpNonZero:  0.
	self genLoadSlot: 0 sourceReg: reg destReg: reg.
	cogit Jump: loop.
	ok jmpTarget: cogit Label.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'cb 4/17/2015 13:32'!
genEnsureOopInRegNotForwarded: reg scratchReg: scratch
	^ self 
		genEnsureOopInRegNotForwarded: reg 
		scratchReg: scratch 
		jumpBackTo: cogit Label "label is just before the forwarder check"! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 1/29/2020 15:12'!
genEnsureOopInRegNotForwarded: reg scratchReg: scratch ifForwarder: fwdJumpTarget ifNotForwarder: nonFwdJumpTargetOrZero
	"Make sure that the oop in reg is not forwarded.  
	 Use the fact that isForwardedObjectClassIndexPun is a power of two to save an instruction."
	<var: 'fwdJumpTarget' type: #'void *'> "maybe a fixup or an instruction"
	<var: 'nonFwdJumpTargetOrZero' type: #'void *'> "maybe a fixup or an instruction"
	| imm ok finished |
	self assert: reg ~= scratch.
	imm := self genJumpImmediate: reg.
	"notionally
		self genGetClassIndexOfNonImm: reg into: scratch.
		cogit CmpCq: objectMemory isForwardedObjectClassIndexPun R: TempReg.
	 but the following is an instruction shorter:"
	cogit MoveMw: 0 r: reg R: scratch.
	cogit
		AndCq: objectMemory classIndexMask - objectMemory isForwardedObjectClassIndexPun
		R: scratch.
	ok := cogit JumpNonZero: 0.
	self genLoadSlot: 0 sourceReg: reg destReg: reg.
	cogit Jump: fwdJumpTarget.
	finished := nonFwdJumpTargetOrZero asUnsignedInteger = 0
					ifTrue: [cogit Label]
					ifFalse: [nonFwdJumpTargetOrZero].
	imm jmpTarget: (ok jmpTarget: finished).
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 2/11/2017 09:20'!
genEnsureOopInRegNotForwarded: reg scratchReg: scratch jumpBackTo: instruction
	<var: #instruction type: #'AbstractInstruction *'>
	<inline: true>
	^ self 
		genEnsureOopInRegNotForwarded: reg 
		scratchReg: scratch 
		ifForwarder: instruction
		ifNotForwarder: 0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 6/4/2015 09:38'!
genEnsureOopInRegNotForwarded: reg scratchReg: scratch updatingMw: offset r: baseReg
	"Make sure that the oop in reg is not forwarded, and the field reg[offset] is updated
	 if the object in reg is forwarded.  Use the fact that isForwardedObjectClassIndexPun is
	 a power of two to save an instruction."
	| loop imm ok |
	<var: #ok type: #'AbstractInstruction *'>
	<var: #imm type: #'AbstractInstruction *'>
	<var: #loop type: #'AbstractInstruction *'>
	self assert: (reg ~= scratch and: [baseReg ~= scratch]).
	loop := cogit Label.
	imm := self genJumpImmediate: reg.
	"notionally
		self genGetClassIndexOfNonImm: reg into: scratch.
		cogit CmpCq: objectMemory isForwardedObjectClassIndexPun R: TempReg.
	 but the following is an instruction shorter:"
	cogit MoveMw: 0 r: reg R: scratch.
	cogit
		AndCq: objectMemory classIndexMask - objectMemory isForwardedObjectClassIndexPun
		R: scratch.
	ok := cogit JumpNonZero:  0.
	self genLoadSlot: 0 sourceReg: reg destReg: reg.
	cogit MoveR: reg Mw: offset r: baseReg.
	cogit Jump: loop.
	ok jmpTarget: (imm jmpTarget: cogit Label).
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 6/4/2015 12:38'!
genEnsureOopInRegNotForwarded: reg scratchReg: scratch updatingSlot: index in: objReg
	"Make sure that the oop in reg is not forwarded, updating the slot in objReg with the value."
	| loop imm ok |
	<var: #ok type: #'AbstractInstruction *'>
	<var: #imm type: #'AbstractInstruction *'>
	<var: #loop type: #'AbstractInstruction *'>
	"Open-code
		self genEnsureOopInRegNotForwarded: reg
			scratchReg: scratch
			updatingMw: index * objectMemory wordSize + objectMemory baseHeaderSize
			r: objReg.
	 to avoid calling the store check unless the receiver is forwarded."
	self assert: (reg ~= scratch and: [objReg ~= scratch]).
	loop := cogit Label.
	imm := self genJumpImmediate: reg.
	"notionally
		self genGetClassIndexOfNonImm: reg into: scratch.
		cogit CmpCq: objectMemory isForwardedObjectClassIndexPun R: TempReg.
	 but the following is an instruction shorter:"
	cogit MoveMw: 0 r: reg R: scratch.
	cogit
		AndCq: objectMemory classIndexMask - objectMemory isForwardedObjectClassIndexPun
		R: scratch.
	ok := cogit JumpNonZero:  0.
	self genLoadSlot: 0 sourceReg: reg destReg: reg.
	cogit MoveR: reg Mw: index * objectMemory wordSize + objectMemory baseHeaderSize r: objReg.

	"Check that we're meeting the contract of ceStoreCheckContextReceiverTrampoline."
	self assert: (reg = Arg0Reg and: [scratch = TempReg and: [objReg = ReceiverResultReg]]).
	cogit CallRT: ceStoreCheckContextReceiverTrampoline.

	cogit Jump: loop.
	ok jmpTarget: (imm jmpTarget: cogit Label).
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/3/2013 10:07'!
genFetchIndexRegister: indexReg from: tableObj into: destReg
	"indexReg contains the 1-relative index of an element in tableObj."
	self subclassResponsibility! !

!CogObjectRepresentationForSpur methodsFor: 'initialization' stamp: 'eem 1/9/2020 16:38'!
genGetActiveContextLarge: isLarge inBlock: isInBlock
	"Create a trampoline to answer the active context that will
	 answer it if a frame is already married, and create it otherwise.
	 Assume numArgs is in SendNumArgsReg and ClassReg is free."
	| header slotSize jumpSingle loopHead jumpNeedScavenge continuation exit |
	cogit
		MoveMw: FoxMethod r: FPReg R: ClassReg;
		TstCq: MFMethodFlagHasContextFlag R: ClassReg.
	jumpSingle := cogit JumpZero: 0. "jump if flag bit not set"
	cogit "since the flag bit was set, get the context in the receiver reg and return"
		MoveMw: FoxThisContext r: FPReg R: ReceiverResultReg;
		RetN: 0.
	jumpSingle jmpTarget: cogit Label.

	"OK, it doesn't exist; instantiate and initialize it"
	"set the hasContext flag; See CoInterpreter class>>initializeFrameIndices"
	cogit
		OrCq: MFMethodFlagHasContextFlag R: ClassReg;
		MoveR: ClassReg Mw: FoxMethod r: FPReg.
	"now get the home CogMethod into ClassReg and save for post-instantiation."
	isInBlock caseOf: {
	[InFullBlock]	-> [cogit SubCq: 3 R: ClassReg]. "-3 is -(hasContext+isBlock) flags"
	[InVanillaBlock]	-> [cogit
							SubCq: 3 R: ClassReg; "-3 is -(hasContext+isBlock) flags"
							MoveM16: 0 r: ClassReg R: TempReg;
							SubR: TempReg R: ClassReg].
	[0]				-> [cogit SubCq: 1 R: ClassReg]. "-1 is hasContext flag" }.

	"instantiate the context..."
	slotSize := isLarge ifTrue: [LargeContextSlots] ifFalse: [SmallContextSlots].
	header := objectMemory
					headerForSlots: slotSize
					format: objectMemory indexablePointersFormat
					classIndex: ClassMethodContextCompactIndex.
	self flag: #endianness.
	cogit MoveAw: objectMemory freeStartAddress R: ReceiverResultReg.
	self genStoreHeader: header intoNewInstance: ReceiverResultReg using: TempReg.
	cogit
		LoadEffectiveAddressMw: (objectMemory smallObjectBytesForSlots: slotSize) r: ReceiverResultReg R: TempReg;
		MoveR: TempReg Aw: objectMemory freeStartAddress;
		CmpCq: objectMemory getScavengeThreshold R: TempReg.
	jumpNeedScavenge := cogit JumpAboveOrEqual: 0.

	"Now initialize the fields of the context.  See CoInterpreter>>marryFrame:SP:copyTemps:"
	"sender gets frame pointer as a SmallInteger"
	continuation :=
	self genSetSmallIntegerTagsIn: FPReg into: TempReg.
	cogit MoveR: TempReg Mw: objectMemory baseHeaderSize + (SenderIndex * objectMemory bytesPerOop) r: ReceiverResultReg.

	"pc gets frame caller as a SmallInteger"
	cogit MoveMw: FoxSavedFP r: FPReg R: TempReg.
	self genSetSmallIntegerTagsIn: TempReg.
	cogit MoveR: TempReg Mw: objectMemory baseHeaderSize + (InstructionPointerIndex * objectMemory bytesPerOop) r: ReceiverResultReg.

	"Set the method field, freeing up ClassReg again, and frame's context field,"
	cogit
		MoveMw: (cogit offset: CogMethod of: #methodObject) r: ClassReg R: TempReg;
		MoveR: TempReg Mw: objectMemory baseHeaderSize + (MethodIndex * objectMemory wordSize) r: ReceiverResultReg;
		MoveR: ReceiverResultReg Mw: FoxThisContext r: FPReg.

	"Now compute stack pointer; this is stackPointer (- 1 for return pc if a CISC) - framePointer - wordSize (1 each for saved pc, method, context, receiver) + 1 (1-relative) + numArgs"
	"TPR note - the code here is actually doing
	context stackPointer := ((((fp - sp) / wordSize) - [3|4]) + num args) asSmallInteger"
	cogit
		SubR: SPReg R: FPReg R: TempReg; "TempReg := FPReg - SPReg"
		LogicalShiftRightCq: self log2BytesPerWord R: TempReg;
		SubCq: (cogit backEnd hasLinkRegister ifTrue: [3] ifFalse: [4]) R: TempReg;
		AddR: SendNumArgsReg R: TempReg.
	self genConvertIntegerToSmallIntegerInReg: TempReg.
	cogit MoveR: TempReg Mw: objectMemory baseHeaderSize + (StackPointerIndex * objectMemory bytesPerOop) r: ReceiverResultReg.

	"Set closureOrNil to either the stacked receiver or nil"
	isInBlock > 0
		ifTrue:
			[cogit
				LoadEffectiveAddressMw: 2 r: SendNumArgsReg R: TempReg; "+2 for saved fp and saved pc"
				MoveXwr: TempReg R: FPReg R: TempReg]
		ifFalse:
			[cogit genMoveNilR: TempReg].
	cogit MoveR: TempReg Mw: objectMemory baseHeaderSize + (ClosureIndex * objectMemory bytesPerOop) r: ReceiverResultReg.

	"Set the receiver"
	cogit
		MoveMw: FoxMFReceiver r: FPReg R: TempReg;
		MoveR: TempReg Mw: objectMemory baseHeaderSize + (ReceiverIndex * objectMemory bytesPerOop) r: ReceiverResultReg.

	"Now copy the arguments.  This is tricky because of the shortage of registers,.  ClassReg ranges
	 from 1 to numArgs (SendNumArgsReg), and from ReceiverIndex + 1 to ReceiverIndex + numArgs.
	 1 to: numArgs do:
		[:i|
		temp := longAt(FPReg + ((SendNumArgs - i + 2) * wordSize)). +2 for saved pc and savedfp
		longAtput(FPReg + FoxMFReceiver + (i * wordSize), temp)]"
	"TPR note: this is a prime candidate for passing off to the backend to do at least faintly optimal code"
	cogit MoveCq: 1 R: ClassReg.
	loopHead := cogit CmpR: SendNumArgsReg R: ClassReg.
	exit := cogit JumpGreater: 0.
	cogit
		MoveR: SendNumArgsReg R: TempReg;
		SubR: ClassReg R: TempReg;
		AddCq: 2 R: TempReg; "+2 for saved fp and saved pc"
		MoveXwr: TempReg R: FPReg R: TempReg;
		AddCq: ReceiverIndex + (objectMemory baseHeaderSize / objectMemory wordSize) R: ClassReg; "Now convert ClassReg from frame index to context index"
		MoveR: TempReg Xwr: ClassReg R: ReceiverResultReg;
		SubCq: ReceiverIndex + (objectMemory baseHeaderSize / objectMemory wordSize) - 1 R: ClassReg; "convert back adding 1 ;-)"
		Jump: loopHead.
	exit jmpTarget: cogit Label.

	"Finally nil or copy the non-argument temps.
	 ClassReg := FPReg + FoxMFReceiver.
	 SendNumArgsReg := SendNumArgsReg+ReceiverIndex.
	 [ClassReg := ClassReg - wordSize.
	  backEnd hasLinkRegister
			ifTrue: [SPReg <= ClassReg]
			ifFalse: [SPReg < ClassReg]] whileTrue:
		[receiver[SendNumArgsReg] := *ClassReg.
		 SendNumArgsReg := SendNumArgsReg + 1]]"
	coInterpreter marryFrameCopiesTemps ifFalse:
		[cogit MoveCq: objectMemory nilObject R: TempReg].
	cogit LoadEffectiveAddressMw: FoxMFReceiver r: FPReg R: ClassReg.
	cogit AddCq: ReceiverIndex + 1 + (objectMemory baseHeaderSize / objectMemory wordSize) R: SendNumArgsReg.
	"The receiver has already been set.
	 If on a CISC there's a retpc for the trampoline call on top of stack; if on a RISC there isn't."
	loopHead := cogit SubCq: objectMemory wordSize R: ClassReg.
	cogit CmpR: ClassReg R: SPReg.
	exit := cogit backEnd hasLinkRegister
			ifTrue: [cogit JumpAbove: 0]
			ifFalse: [cogit JumpAboveOrEqual: 0].

	coInterpreter marryFrameCopiesTemps ifTrue:
		[cogit MoveMw: 0 r: ClassReg R: TempReg].
	cogit
		MoveR: TempReg Xwr: SendNumArgsReg R: ReceiverResultReg;
		AddCq: 1 R: SendNumArgsReg;
		Jump: loopHead.
	exit jmpTarget: cogit Label.

	cogit RetN: 0.
	
	jumpNeedScavenge jmpTarget: cogit Label.
	cogit backEnd saveAndRestoreLinkRegAround:
		[cogit
			CallRT: ceScheduleScavengeTrampoline
			registersToBeSavedMask: (cogit registerMaskFor: ReceiverResultReg and: SendNumArgsReg and: ClassReg)].
	cogit Jump: continuation.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'bytecode generator support' stamp: 'eem 9/5/2016 14:26'!
genGetActiveContextNumArgs: numArgs large: isLargeContext inBlock: isInBlock
	"Get the active context into ReceiverResultReg, creating it if necessary."
	| routine |
	routine := isLargeContext
				ifFalse: [isInBlock caseOf: {
						[0]				-> [ceSmallActiveContextInMethodTrampoline].
						[InVanillaBlock]	-> [ceSmallActiveContextInBlockTrampoline].
						[InFullBlock]	-> [ceSmallActiveContextInFullBlockTrampoline] }]
				ifTrue: [isInBlock caseOf: {
						[0]				-> [ceLargeActiveContextInMethodTrampoline].
						[InVanillaBlock]	-> [ceLargeActiveContextInBlockTrampoline].
						[InFullBlock]	-> [ceLargeActiveContextInFullBlockTrampoline] }].
	cogit
		MoveCq: numArgs R: SendNumArgsReg;
		CallRT: routine.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/10/2015 08:47'!
genGetBits: mask ofFormatByteOf: sourceReg into: destReg
	self flag: #endianness.
	cogit MoveMb: 3 r: sourceReg R: destReg.
	cogit AndCq: mask R: destReg.	"formatReg := self formatOfHeader: destReg"
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/3/2013 10:15'!
genGetClassIndexOfNonImm: sourceReg into: destReg
	"Fetch the instance's class index into destReg."

	cogit MoveMw: 0 r: sourceReg R: destReg.
	cogit AndCq: objectMemory classIndexMask R: destReg.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/10/2019 11:56'!
genGetClassObjectOf: instReg into: destReg scratchReg: scratchReg mayBeAForwarder: mayBeForwarder
	"Fetch the instance's class into destReg.  If the instance is not the receiver and is forwarded, follow forwarding."
	| jumpIsImm jumpNotForwarded loop |
	<var: #jumpIsImm type: #'AbstractInstruction *'>
	<var: #jumpNotForwarded type: #'AbstractInstruction *'>
	<var: #loop type: #'AbstractInstruction *'>
	(instReg = destReg or: [instReg = scratchReg or: [destReg = scratchReg]]) ifTrue:
		[^BadRegisterSet].
	loop := cogit MoveR: instReg R: scratchReg.
	cogit AndCq: objectMemory tagMask R: scratchReg.
	jumpIsImm := cogit JumpNonZero: 0.
	self flag: #endianness.
	"Get least significant half of header word in destReg"
	cogit MoveMw: 0 r: instReg R: scratchReg.
	"mask off class index"
	cogit AndCq: objectMemory classIndexMask R: scratchReg.
	mayBeForwarder ifTrue:
		["if it is forwarded..."
		cogit CmpCq: objectMemory isForwardedObjectClassIndexPun R: scratchReg.
		jumpNotForwarded := cogit JumpNonZero: 0.
		"...follow the forwarding pointer and loop to fetch its classIndex"
		cogit MoveMw: objectMemory baseHeaderSize r: instReg R: instReg.
		cogit Jump: loop.
		jumpNotForwarded jmpTarget: cogit Label].
	jumpIsImm jmpTarget:
	(cogit MoveR: scratchReg R: destReg).
	scratchReg = TempReg
		ifTrue:
			[cogit PushR: instReg.
			 self genGetClassObjectOfClassIndex: destReg into: instReg scratchReg: TempReg.
			 cogit MoveR: instReg R: destReg.
			 cogit PopR: instReg]
		ifFalse:
			[self genGetClassObjectOfClassIndex: destReg into: scratchReg scratchReg: TempReg.
			 cogit MoveR: scratchReg R: destReg].
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/15/2015 21:03'!
genGetClassObjectOfClassIndex: instReg into: destReg scratchReg: scratchReg
	"Fetch the class object whose index is in instReg into destReg.
	 It is non-obvious, but the Cogit assumes loading a class does not involve
	 a runtime call, so do not call classAtIndex:"
	self assert: instReg ~= destReg.
	self assert: instReg ~= scratchReg.
	self assert: destReg ~= scratchReg.
	cogit
		MoveR: instReg R: scratchReg;
		LogicalShiftRightCq: objectMemory classTableMajorIndexShift R: scratchReg;
		LogicalShiftLeftCq: objectMemory shiftForWord R: scratchReg.
	self assert: (self shouldAnnotateObjectReference: objectMemory classTableRootObj) not.
	(cogit backEnd isWithinMwOffsetRange: objectMemory classTableRootObj + objectMemory baseHeaderSize)
		ifTrue: [cogit MoveMw: objectMemory classTableRootObj + objectMemory baseHeaderSize r: scratchReg R: destReg]
		ifFalse: [cogit
					AddCq: objectMemory classTableRootObj R: scratchReg;
					MoveMw: objectMemory baseHeaderSize r: scratchReg R: destReg].
	cogit
		MoveR: instReg R: scratchReg;
		AndCq: objectMemory classTableMinorIndexMask R: scratchReg;
		AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: scratchReg;
		MoveXwr: scratchReg R: destReg R: destReg.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 7/7/2014 11:57'!
genGetClassTagOf: instReg into: destReg scratchReg: scratchReg
	^self genGetInlineCacheClassTagFrom: instReg into: destReg forEntry: true! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/3/2013 10:15'!
genGetCompactClassIndexNonImmOf: instReg into: destReg
	"Fetch the instance's class index into destReg."
	^self genGetClassIndexOfNonImm: instReg into: destReg! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/3/2013 10:11'!
genGetDoubleValueOf: srcReg into: destFPReg 
	cogit MoveM64: objectMemory baseHeaderSize r: srcReg Rd: destFPReg.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/10/2015 08:47'!
genGetFormatOf: srcReg into: destReg
	"Get the format field of the object in srcReg into destReg.
	 srcReg may equal destReg."
	^self genGetBits: objectMemory formatMask ofFormatByteOf: srcReg into: destReg! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/25/2019 15:30'!
genGetFormatOf: sourceReg into: destReg leastSignificantHalfOfBaseHeaderIntoScratch: scratchRegOrNone
	"Get the format of the object in sourceReg into destReg.  If scratchRegOrNone
	 is not NoReg, load at least the least significant 32-bits (64-bits in 64-bits) of the
	 header word, which contains the format, into scratchRegOrNone."
	scratchRegOrNone = NoReg
		ifTrue:
			[self flag: #endianness.
			 cogit MoveMb: 3 r: sourceReg R: destReg]
		ifFalse:
			[cogit MoveMw: 0 r: sourceReg R: scratchRegOrNone.
			 cogit LogicalShiftRightCq: objectMemory formatShift R: scratchRegOrNone R: destReg].
	cogit AndCq: objectMemory formatMask R: destReg.	"formatReg := self formatOfHeader: destReg"
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'eem 4/26/2017 14:20'!
genGetIdentityHash: rcvrReg resultReg: resultReg
	<var: #jumpSet type: #'AbstractInstruction *'>
	| jumpSet |
	"ReceiverResultReg is required for the trampoline. We force the allocation,
	 and we have two path to avoid conflicts in ReceiverResultReg."
	cogit voidReceiverResultRegContainsSelf. 
	resultReg = ReceiverResultReg 
		ifTrue: 
			[cogit ssTop popToReg: rcvrReg.
			 self genGetHashFieldNonImmOf: rcvrReg asSmallIntegerInto: resultReg.
			 cogit CmpCq: ConstZero R: resultReg.
			 jumpSet := cogit JumpNonZero: 0.
			 cogit MoveR: rcvrReg R: resultReg.
			 cogit CallRT: ceInlineNewHashTrampoline]
		ifFalse: 
			[cogit ssTop popToReg: ReceiverResultReg.
		 	 self genGetHashFieldNonImmOf: ReceiverResultReg asSmallIntegerInto: resultReg.
			 cogit CmpCq: ConstZero R: resultReg.
			 jumpSet := cogit JumpNonZero: 0.
			 cogit CallRT: ceInlineNewHashTrampoline.
			 cogit MoveR: ReceiverResultReg R: resultReg].
	jumpSet jmpTarget: cogit Label! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'cb 12/13/2017 11:43'!
genGetInstanceOfByteClass: classObj into: destReg initializingIf: initializeInstance numBytes: numBytes
	"Create an instance of classObj and assign it to destReg, initializing the instance
	 if initializeInstance is true with 0 This is for inline primitives.
	 Assume there is sufficient space in new space to complete the operation."
	| classIndex numSlots byteFormat |
	classIndex := objectMemory rawHashBitsOf: classObj.
	self flag: #duplication. "Duplicated byteFormatForNumBytes: and numSlotsForBytes:, might be worth adding api"
	numSlots := numBytes + (objectMemory wordSize - 1) // objectMemory wordSize.
	byteFormat := objectMemory firstByteFormat + (8 - numBytes bitAnd: objectMemory wordSize - 1).
	self assert: classIndex ~= 0.
	self genGetUninitializedInstanceWithClassIndex: classIndex numSlots: numSlots format: byteFormat into: destReg.
	(initializeInstance and: [numBytes > 0]) ifTrue: 
		[self genStoreValue: 0 instance: destReg numSlots: numSlots].
	^0
	! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'cb 12/14/2017 20:04'!
genGetInstanceOfPointerClass: classObj into: destReg initializingIf: initializeInstance numVariableSlots: varSlots
	"Create an instance of classObj and assign it to destReg, initializing the instance
	 if initializeInstance is true with nil This is for inline primitives.
	 Assume there is sufficient space in new space to complete the operation."
	| classIndex classFormat totalNumSlots fixedSlots |
	classIndex := objectMemory rawHashBitsOf: classObj.
	classFormat := objectMemory formatOfClass: classObj.
	fixedSlots := objectMemory fixedFieldsOfClassFormat: classFormat.
	totalNumSlots := varSlots + fixedSlots.
	self assert: classIndex ~= 0.
	self genGetUninitializedInstanceWithClassIndex: classIndex numSlots: totalNumSlots format: (objectMemory instSpecOfClassFormat: classFormat) into: destReg.
	(initializeInstance and: [totalNumSlots > 0]) ifTrue: 
		[self genStoreValue: objectMemory nilObject instance: destReg numSlots: totalNumSlots].
	^0
	! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/26/2016 11:34'!
genGetNumBytesOf: srcReg into: destReg
	"Get the size in byte-sized slots of the object in srcReg into destReg.
	 srcReg may equal destReg.
	 destReg <- numSlots << self shiftForWord - (fmt bitAnd: 3).
	 Assumes the object in srcReg has a byte format, i.e. 16 to 23 or 24 to 31 "
	self subclassResponsibility! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 4/10/2015 15:56'!
genGetNumSlotsOf: srcReg into: destReg
	"Get the size in word-sized slots of the object in srcReg into destReg.
	 srcReg may equal destReg."
	<var: #jmp type: #'AbstractInstruction *'>
	| jmp |
	self assert: srcReg ~= destReg.
	self genGetRawSlotSizeOfNonImm: srcReg into: destReg.
	cogit CmpCq: objectMemory numSlotsMask R: destReg.
	jmp := cogit JumpLess: 0.
	self genGetOverflowSlotsOf: srcReg into: destReg.
	jmp jmpTarget: cogit Label.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 4/10/2015 15:54'!
genGetOverflowSlotsOf: srcReg into: destReg
	self subclassResponsibility! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 11/4/2015 12:24:12'!
genGetRawSlotSizeOfNonImm: sourceReg into: destReg
	"The raw numSlots field is the most significant byte of the 64-bit header word.
	 MoveMbrR zero-extends."
	cogit backEnd byteReadsZeroExtend ifFalse:
		[cogit MoveCq: 0 R: destReg].
	cogit MoveMb: 7 r: sourceReg R: destReg.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'cb 4/19/2018 14:13'!
genGetUninitializedInstanceWithClassIndex: classIndex numSlots: numSlots format: format into: destReg
	"Write in destReg the pointer to the object (adjusted based on header size)
	Can deal with large header but not with with old space allocation (max allocation size)"
	| overflowHeader header lowNumSlots |
	<var: #overflowHeader type: #usqLong>
	numSlots >= objectMemory fixedFieldsOfClassFormatMask ifTrue: [^UnimplementedOperation].
	cogit MoveAw: objectMemory freeStartAddress R: destReg.
	numSlots >= objectMemory numSlotsMask
		ifTrue: 
			[overflowHeader := numSlots + (objectMemory numSlotsMask << objectMemory numSlotsFullShift).
			 self genStoreHeader: overflowHeader intoNewInstance: destReg using: TempReg.
			 cogit AddCq: objectMemory baseHeaderSize R: destReg. "Allow to store base header at correct place and to use smallObjectBytes in the rest of the method"
			 lowNumSlots := objectMemory numSlotsMask]
		ifFalse: 
			[lowNumSlots := numSlots].	
	header := objectMemory
					headerForSlots: lowNumSlots
					format: format
					classIndex: classIndex.
	self genStoreHeader: header intoNewInstance: destReg using: TempReg.
	cogit
		LoadEffectiveAddressMw: (objectMemory objectBytesForSlots: numSlots) r: destReg R: TempReg;
		MoveR: TempReg Aw: objectMemory freeStartAddress.

	! !

!CogObjectRepresentationForSpur methodsFor: 'mapped inlined primitive support' stamp: 'cb 11/6/2017 11:32'!
genIfRequiredCheckRememberedBitOf: rr scratch: scratchReg
	CheckRememberedInTrampoline ifFalse: 
		[^self genCheckRememberedBitOf: rr scratch: scratchReg]! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'cb 11/24/2016 13:21'!
genInlinedIdenticalOrNotIfGuts: orNot
	<inline: true>
	^ cogit genForwardersInlinedIdenticalOrNotIf: orNot! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'cb 3/30/2016 15:32'!
genJumpBaseHeaderImmutable: baseHeaderReg
	"baseHeader holds at least the least significant 32 bits of the object"
	<returnTypeC: #'AbstractInstruction *'>
	<option: #IMMUTABILITY>
	<inline: true>
	cogit TstCq: objectMemory immutableBitMask R: baseHeaderReg.
	^ cogit JumpNonZero: 0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'cb 3/30/2016 15:32'!
genJumpBaseHeaderMutable: baseHeaderReg
	"baseHeader holds at least the least significant 32 bits of the object"
	<returnTypeC: #'AbstractInstruction *'>
	<option: #IMMUTABILITY>
	<inline: true>
	cogit TstCq: objectMemory immutableBitMask R: baseHeaderReg.
	^ cogit JumpZero: 0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 8/7/2014 04:54'!
genJumpCharacterInScratchReg: reg 
	cogit AndCq: objectMemory tagMask R: reg.
	cogit CmpCq: objectMemory characterTag R: reg.
	^cogit JumpZero: 0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 5/29/2015 18:13'!
genJumpImmediate: aRegister
	<returnTypeC: #'AbstractInstruction *'>
	cogit TstCq: objectMemory tagMask R: aRegister.
	^cogit JumpNonZero: 0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'cb 3/30/2016 15:32'!
genJumpImmutable: sourceReg scratchReg: scratchReg
	<returnTypeC: #'AbstractInstruction *'>
	<option: #IMMUTABILITY>
	cogit MoveMw: 0 r: sourceReg R: scratchReg. 
	^ self genJumpBaseHeaderImmutable: scratchReg! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 6/1/2016 16:33'!
genJumpInOldSpace: reg
	"Jump if reg is old."
	<inline: true>
	^cogit
		CmpCq: objectMemory storeCheckBoundary R: reg; "N.B. FLAGS := destReg - scratchReg"
		JumpAboveOrEqual: 0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'cb 12/29/2015 10:20'!
genJumpMutable: sourceReg scratchReg: scratchReg
	<returnTypeC: #'AbstractInstruction *'>
	<option: #IMMUTABILITY>
	cogit MoveMw: 0 r: sourceReg R: scratchReg. 
	^ self genJumpBaseHeaderMutable: scratchReg! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/9/2013 18:36'!
genJumpNotCharacterInScratchReg: reg 
	cogit AndCq: objectMemory tagMask R: reg.
	cogit CmpCq: objectMemory characterTag R: reg.
	^cogit JumpNonZero: 0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 5/29/2015 18:15'!
genJumpNotImmediate: aRegister
	<returnTypeC: #'AbstractInstruction *'>
	cogit TstCq: objectMemory tagMask R: aRegister.
	^cogit JumpZero: 0! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcFirstFieldPointer: objectReg
	<option: #LowcodeVM>
	cogit AddCq: 8 R: objectReg.
	cogit ssPushNativeRegister: objectReg.! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcFirstIndexableFieldPointer: objectReg
	<option: #LowcodeVM>
	"TODO: Retrieve the number of fixed fields."
	cogit AddCq: 8 R: objectReg.
	cogit ssPushNativeRegister: objectReg.! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'rsf 9/2/2016 14:56'!
genLcIsBytes: objectReg to: valueReg
	<option: #LowcodeVM>
	
	"Check for integer"
	| isImmediate isNotBytes isCompiledMethod cont falseTarget |
	cogit MoveR: objectReg R: valueReg.
	cogit AndCq: objectMemory tagMask R: valueReg.
	isImmediate := cogit JumpNonZero: 0.
	
	"Get the format"
	self genGetFormatOf: objectReg into: valueReg.
	
	cogit CmpCq: objectMemory firstByteFormat R: valueReg.
	isNotBytes := cogit JumpLess: 0.
	cogit CmpCq: objectMemory firstCompiledMethodFormat R: valueReg.
	isCompiledMethod := cogit JumpGreaterOrEqual: 0.
	
	"True result"
	cogit MoveCq: 1 R: valueReg.
	cont := cogit Jump: 0.
	
	"False result"
	falseTarget := cogit Label.
	isImmediate jmpTarget: falseTarget.
	isNotBytes jmpTarget: falseTarget.
	isCompiledMethod jmpTarget: falseTarget.
	cogit MoveCq: 0 R: valueReg.

	"Return"
	cont jmpTarget: cogit Label.
	cogit ssPushNativeRegister: valueReg.! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcIsFloatObject: objectReg to: valueReg
	<option: #LowcodeVM>
	"TODO: Implement this one"
	cogit MoveCq: 1 R: valueReg.
	cogit ssPushNativeRegister: valueReg.! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcIsIndexable: objectReg to: valueReg
	<option: #LowcodeVM>
	"TODO: Implement this one"
	cogit MoveCq: 1 R: valueReg.
	cogit ssPushNativeRegister: valueReg.! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'rsf 9/2/2016 15:09'!
genLcIsIntegerObject: objectReg to: valueReg
	<option: #LowcodeVM>
	
	"Check for the immediate case"
	| isImmediate isLargePositiveInteger isLargeNegativeInteger falseResult trueResult cont |
	cogit MoveR: objectReg R: valueReg.
	cogit AndCq: objectMemory smallIntegerTag R: valueReg.
	isImmediate := cogit JumpNonZero: 0.
	
	"Check the non-immediate case"
	self genGetClassIndexOfNonImm: objectReg into: TempReg.
	cogit CmpCq: ClassLargePositiveInteger R: TempReg.
	isLargePositiveInteger := cogit JumpNonZero: 0.
	cogit CmpCq: ClassLargeNegativeInteger R: TempReg.
	isLargeNegativeInteger := cogit JumpNonZero: 0.
	cogit XorR: valueReg R: valueReg.
	falseResult := cogit Jump: 0.
	
	"True result"
	trueResult := cogit Label.
	isLargePositiveInteger jmpTarget: trueResult.
	isLargeNegativeInteger jmpTarget: trueResult.
	cogit MoveCq: 1 R: valueReg.
	
	"Return"
	cont := cogit Label.
	falseResult jmpTarget: cont.
	isImmediate jmpTarget: cont.
	cogit ssPushNativeRegister: valueReg.! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'rsf 9/2/2016 14:55'!
genLcIsPointers: objectReg to: valueReg
	<option: #LowcodeVM>
	
	"Check for immediate"
	| isImmediate isNotPointers cont falseTarget |
	cogit MoveR: objectReg R: valueReg.
	cogit AndCq: objectMemory tagMask R: valueReg.
	isImmediate := cogit JumpNonZero: 0.
	
	"Get the format"
	self genGetFormatOf: objectReg into: valueReg.
	
	cogit CmpCq: 9 R: valueReg.
	isNotPointers := cogit JumpGreaterOrEqual: 0.
	
	"True result"
	cogit MoveCq: 1 R: valueReg.
	cont := cogit Jump: 0.
	
	"False result"
	falseTarget := cogit Label.
	isImmediate jmpTarget: falseTarget.
	isNotPointers jmpTarget: falseTarget.
	cogit MoveCq: 0 R: valueReg.

	"Return"
	cont jmpTarget: cogit Label.
	cogit ssPushNativeRegister: valueReg.! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'rsf 1/13/2017 04:00'!
genLcIsWords: objectReg to: valueReg
	<option: #LowcodeVM>
	
	"Check for immediate"
	| isImmediate isNotWords cont falseTarget isBelowRange isAboveRange |
	cogit MoveR: objectReg R: valueReg.
	cogit AndCq: objectMemory tagMask R: valueReg.
	isImmediate := cogit JumpNonZero: 0.
	
	"Get the format"
	self genGetFormatOf: objectReg into: valueReg.
	
	"Hack: consider a compiled method as bytes too."
	BytesPerWord = 8 ifTrue: [
		cogit CmpCq: objectMemory firstLongFormat R: valueReg.
		isBelowRange := cogit JumpLess: 0.
		cogit CmpCq: objectMemory firstShortFormat R: valueReg.
		isAboveRange := cogit JumpGreaterOrEqual: 0.
	] ifFalse: [ 
		cogit CmpCq: objectMemory firstLongFormat R: valueReg.
		isNotWords := cogit JumpNonZero: 0.
	].
	
	"True result"
	cogit MoveCq: 1 R: valueReg.
	cont := cogit Jump: 0.
	
	"False result"
	falseTarget := cogit Label.
	isImmediate jmpTarget: falseTarget.
	BytesPerWord = 4 ifTrue: [
		isNotWords jmpTarget: falseTarget.
	] ifFalse: [ 
		isBelowRange jmpTarget: falseTarget.
		isAboveRange jmpTarget: falseTarget.
	].
	cogit MoveCq: 0 R: valueReg.

	"Return"
	cont jmpTarget: cogit Label.
	cogit ssPushNativeRegister: valueReg.! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'rsf 9/2/2016 14:55'!
genLcIsWordsOrBytes: objectReg to: valueReg
	<option: #LowcodeVM>
	
	"Check for immediate"
	| isImmediate isNotBits isCompiledMethod cont falseTarget |
	cogit MoveR: objectReg R: valueReg.
	cogit AndCq: objectMemory tagMask R: valueReg.
	isImmediate := cogit JumpNonZero: 0.
	
	"Get the format"
	self genGetFormatOf: objectReg into: valueReg.
	
	cogit CmpCq: 9 R: valueReg.
	isNotBits := cogit JumpLess: 0.
	cogit CmpCq: 24 R: valueReg.
	isCompiledMethod := cogit JumpGreaterOrEqual: 0.
	
	"True result"
	cogit MoveCq: 1 R: valueReg.
	cont := cogit Jump: 0.
	
	"False result"
	falseTarget := cogit Label.
	isImmediate jmpTarget: falseTarget.
	isNotBits jmpTarget: falseTarget.
	isCompiledMethod jmpTarget: falseTarget.
	cogit MoveCq: 0 R: valueReg.

	"Return"
	cont jmpTarget: cogit Label.
	cogit ssPushNativeRegister: valueReg.! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:29'!
genLcLoadObject: object at: fieldIndex
	<option: #LowcodeVM>
	cogit AddCq: 8 R: object.
	cogit MoveXwr: fieldIndex R: object R: object.
	cogit ssPushRegister: object! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'eem 2/11/2020 10:19'!
genLcLoadObject: object field: fieldIndex
	<option: #LowcodeVM>
	cogit MoveMw: objectMemory baseHeaderSize + (objectMemory bytesPerOop * fieldIndex) r: object R: object.
	cogit ssPushRegister: object! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:29'!
genLcStore: value object: object at: fieldIndex
	<option: #LowcodeVM>
	cogit AddCq: 8 R: object.
	cogit MoveR: value Xwr: fieldIndex R: object.! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'eem 2/11/2020 10:19'!
genLcStore: value object: object field: fieldIndex
	<option: #LowcodeVM>
	cogit MoveR: value Mw: objectMemory baseHeaderSize + (fieldIndex * objectMemory bytesPerOop) r: object.! !

!CogObjectRepresentationForSpur methodsFor: 'bytecode generator support' stamp: 'eem 7/6/2021 17:07'!
genNewArrayOfSize: size initialized: initialize
	"Generate a call to code that allocates a new Array of size.
	 The Array should be initialized with nils iff initialize is true.
	 The size arg is passed in SendNumArgsReg, the result
	 must come back in ReceiverResultReg."
	| header skip |
	<var: #skip type: #'AbstractInstruction *'>
	self assert: size < objectMemory numSlotsMask.
	header := objectMemory
					headerForSlots: size
					format: objectMemory arrayFormat
					classIndex: ClassArrayCompactIndex.
	cogit MoveAw: objectMemory freeStartAddress R: ReceiverResultReg.
	self genStoreHeader: header intoNewInstance: ReceiverResultReg using: TempReg.
	(initialize and: [size > 0]) ifTrue:
		[cogit genMoveConstant: objectMemory nilObject R: TempReg.
		 0 to: size - 1 do:
			[:i| cogit MoveR: TempReg
					Mw: i * objectMemory wordSize + objectMemory baseHeaderSize
					r: ReceiverResultReg]].
	cogit
		LoadEffectiveAddressMw: (objectMemory smallObjectBytesForSlots: size) r: ReceiverResultReg R: TempReg;
		MoveR: TempReg Aw: objectMemory freeStartAddress;
		CmpCq: objectMemory getScavengeThreshold R: TempReg.
	skip := cogit JumpBelow: 0.
	cogit CallRT: ceScheduleScavengeTrampoline.
	skip jmpTarget: cogit Label.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'initialization' stamp: 'eem 3/16/2017 17:39'!
genNewHashTrampoline: saveRegs called: name
	"In non sista VM this is used only from the identityHash primitive, hence only the result of the trampoline, the hash, should be in ReceiverResultReg, other registers can just be ignored.
	In the sista VM, the inlined hash operation requires registers to be saved"
	<inline: true>
	^cogit
		genTrampolineFor: #ceNewHashOf:
		called: name
		numArgs: 1
		arg: ReceiverResultReg
		arg: nil
		arg: nil
		arg: nil
		regsToSave: (saveRegs
						ifTrue: [CallerSavedRegisterMask bitClear: (cogit registerMaskFor: ReceiverResultReg)]
						ifFalse: [cogit emptyRegisterMask])
		pushLinkReg: true
		resultReg: ReceiverResultReg
		appendOpcodes: false! !

!CogObjectRepresentationForSpur methodsFor: 'bytecode generator support' stamp: 'eem 12/7/2015 16:12'!
genNoPopCreateClosureAt: bcpc numArgs: numArgs numCopied: numCopied contextNumArgs: ctxtNumArgs large: isLargeCtxt inBlock: isInBlock
	"Create a closure with the given startpc, numArgs and numCopied
	 within a context with ctxtNumArgs, large if isLargeCtxt that is in a
	 block if isInBlock.  Do /not/ initialize the copied values."
	| numSlots byteSize header skip |
	<var: #skip type: #'AbstractInstruction *'>

	"First get thisContext into ReceiverResultRega and thence in ClassReg."
	self genGetActiveContextNumArgs: ctxtNumArgs large: isLargeCtxt inBlock: isInBlock.
	cogit MoveR: ReceiverResultReg R: ClassReg.

	numSlots := ClosureFirstCopiedValueIndex + numCopied.
	byteSize := objectMemory smallObjectBytesForSlots: numSlots.
	header := objectMemory
					headerForSlots: numSlots
					format: objectMemory indexablePointersFormat
					classIndex: ClassBlockClosureCompactIndex.
	cogit MoveAw: objectMemory freeStartAddress R: ReceiverResultReg.
	self genStoreHeader: header intoNewInstance: ReceiverResultReg using: TempReg.
	cogit
		LoadEffectiveAddressMw: byteSize r: ReceiverResultReg R: TempReg;
		MoveR: TempReg Aw: objectMemory freeStartAddress;
		CmpCq: objectMemory getScavengeThreshold R: TempReg.
	skip := cogit JumpBelow: 0.
	cogit CallRT: ceScheduleScavengeTrampoline.
	skip jmpTarget: cogit Label.

	cogit
		MoveR: ClassReg Mw: ClosureOuterContextIndex * objectMemory bytesPerOop + objectMemory baseHeaderSize r: ReceiverResultReg;
		MoveCq: (objectMemory integerObjectOf: bcpc) R: TempReg;
		MoveR: TempReg Mw: ClosureStartPCIndex * objectMemory bytesPerOop + objectMemory baseHeaderSize r: ReceiverResultReg;
		MoveCq: (objectMemory integerObjectOf: numArgs) R: TempReg;
		MoveR: TempReg Mw: ClosureNumArgsIndex * objectMemory bytesPerOop + objectMemory baseHeaderSize r: ReceiverResultReg.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'eem 1/11/2016 18:01'!
genPrimitiveAsCharacter
	| reg jumpNotInt jumpOutOfRange |
	<var: 'jumpNotInt' type: #'AbstractInstruction *'>
	<var: 'jumpOutOfRange' type: #'AbstractInstruction *'>
	cogit methodNumArgs = 0
		ifTrue: [reg := ReceiverResultReg]
		ifFalse:
			[cogit methodNumArgs > 1 ifTrue:
				[^UnimplementedPrimitive].
			 reg := Arg0Reg.
			 cogit genLoadArgAtDepth: 0 into: Arg0Reg.
			 jumpNotInt := self genJumpNotSmallInteger: reg scratchReg: TempReg].
	cogit MoveR: reg R: TempReg.
	self genConvertSmallIntegerToIntegerInReg: TempReg.
	jumpOutOfRange := self jumpNotCharacterUnsignedValueInRegister: TempReg.
	self genConvertSmallIntegerToCharacterInReg: reg.
	reg ~= ReceiverResultReg ifTrue:
		[cogit MoveR: reg R: ReceiverResultReg].
	cogit genPrimReturn.
	jumpOutOfRange jmpTarget: cogit Label.
	reg ~= ReceiverResultReg ifTrue:
		[jumpNotInt jmpTarget: jumpOutOfRange getJmpTarget].
	^CompletePrimitive! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'eem 6/21/2018 16:42'!
genPrimitiveAt
	"Generate primitive 60, at: with unsigned access for pure bits classes."
	^self genPrimitiveAtSigned: false! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'eem 6/30/2018 16:22'!
genPrimitiveAtPut
	"Generate primitive 61, at:put: with unsigned access for pure bits classes."
	^self genPrimitiveAtPutSigned: false! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'cb 3/24/2016 10:19'!
genPrimitiveIdenticalOrNotIf: orNot
	| jumpCmp comp |
	<var: #jumpCmp type: #'AbstractInstruction *'>
	<var: #comp type: #'AbstractInstruction *'>
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	comp := cogit CmpR: Arg0Reg R: ReceiverResultReg.
	orNot
		ifTrue: 
			[ jumpCmp := cogit JumpZero: 0.
			self genEnsureOopInRegNotForwarded: Arg0Reg scratchReg: TempReg jumpBackTo: comp ]
		ifFalse: 
			[ jumpCmp := cogit JumpNonZero: 0 ].
	cogit genMoveTrueR: ReceiverResultReg.
	cogit genPrimReturn.
	jumpCmp jmpTarget: cogit Label.
	orNot ifFalse: 
		[ self genEnsureOopInRegNotForwarded: Arg0Reg scratchReg: TempReg jumpBackTo: comp ].
	cogit genMoveFalseR: ReceiverResultReg.
	cogit genPrimReturn.
	^UnfailingPrimitive
! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:31'!
genPrimitiveIdentityHash
	self subclassResponsibility! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'eem 2/28/2016 10:52'!
genPrimitiveImmediateAsInteger
	self subclassResponsibility! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'eem 6/21/2018 16:42'!
genPrimitiveIntegerAt
	"Generate primitive 164, at: with signed access for pure bits classes."
	^self genPrimitiveAtSigned: true! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'eem 6/30/2018 07:23'!
genPrimitiveIntegerAtPut
	"Generate primitive 165, at:put: with signed access for pure bits classes."
	^self genPrimitiveAtPutSigned: true! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'eem 1/22/2021 21:18'!
genPrimitiveMakePoint
	"<returnTypeC: #'AbstractInstruction *'>"
	| allocSize newPointHeader jumpFail resultReg scratchReg |
	resultReg := ClassReg.
	scratchReg := SendNumArgsReg.
	"<var: #jumpFail type: #'AbstractInstruction *'>"
	allocSize := objectMemory baseHeaderSize + (objectMemory wordSize * 2).
	newPointHeader := objectMemory
							headerForSlots: 2
							format: objectMemory nonIndexablePointerFormat
							classIndex: ClassPointCompactIndex.
	cogit
		MoveAw: objectMemory freeStartAddress R: resultReg;
		LoadEffectiveAddressMw: allocSize r: resultReg R: scratchReg;
		CmpCq: objectMemory getScavengeThreshold R: scratchReg.
	jumpFail := cogit JumpAboveOrEqual: 0.
	cogit MoveR: scratchReg Aw: objectMemory freeStartAddress.
	self genStoreHeader: newPointHeader intoNewInstance: resultReg using: scratchReg.
	cogit
		MoveR: ReceiverResultReg Mw: objectMemory baseHeaderSize r: resultReg;
		MoveR: Arg0Reg Mw: objectMemory baseHeaderSize + objectMemory wordSize r: resultReg;
		MoveR: resultReg R: ReceiverResultReg;
		genPrimReturn.
	jumpFail jmpTarget: cogit Label.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'eem 1/23/2016 11:46'!
genPrimitiveObjectAt
	| headerReg jumpBadIndex jumpBounds jumpNotHeaderIndex |
	<var: #jumpBounds type: #'AbstractInstruction *'>
	<var: #jumpBadIndex type: #'AbstractInstruction *'>
	<var: #jumpNotHeaderIndex type: #'AbstractInstruction *'>
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	jumpBadIndex := self genJumpNotSmallInteger: Arg0Reg scratchReg: TempReg.
	self genGetMethodHeaderOf: ReceiverResultReg into: (headerReg := Arg1Reg) scratch: TempReg.
	cogit CmpCq: (objectMemory integerObjectOf: 1) R: Arg0Reg.
	jumpNotHeaderIndex := cogit JumpNonZero: 0.
	cogit
		MoveR: headerReg R: ReceiverResultReg;
		genPrimReturn.
	jumpNotHeaderIndex jmpTarget: (cogit
		AndCq: (objectMemory integerObjectOf: coInterpreter alternateHeaderNumLiteralsMask) R: headerReg).
	cogit
		SubCq: (objectMemory integerObjectOf: 1) - objectMemory smallIntegerTag R: Arg0Reg;
		CmpR: headerReg R: Arg0Reg.
	jumpBounds := cogit JumpAbove: 0.

	self genConvertSmallIntegerToIntegerInReg: Arg0Reg.
	cogit
		AddCq: objectMemory baseHeaderSize >> objectMemory shiftForWord R: Arg0Reg;
		MoveXwr: Arg0Reg R: ReceiverResultReg R: ReceiverResultReg;
		genPrimReturn.

	jumpBounds jmpTarget: (cogit
		AddCq: (objectMemory integerObjectOf: 1) - objectMemory smallIntegerTag R: Arg0Reg).
	jumpBadIndex jmpTarget: cogit Label.
	^CompletePrimitive! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'eem 12/26/2019 21:59'!
genPrimitiveSize
	| jumpImm jumpNotIndexable jumpIsContext |
	"c.f. StackInterpreter>>stSizeOf: lengthOf:baseHeader:format: fixedFieldsOf:format:length:"
	<var: #jumpImm type: #'AbstractInstruction *'>
	<var: #jumpIsContext type: #'AbstractInstruction *'>
	<var: #jumpNotIndexable type: #'AbstractInstruction *'>
	jumpImm := self genJumpImmediate: ReceiverResultReg.
	self
		genGetSizeOf: ReceiverResultReg
		into: ClassReg
		formatReg: SendNumArgsReg
		scratchReg: TempReg
		abortJumpsInto: [:jnx :jic| jumpNotIndexable := jnx. jumpIsContext := jic].
	self genConvertIntegerInReg: ClassReg toSmallIntegerInReg: ReceiverResultReg.
	cogit genPrimReturn.
	jumpImm jmpTarget: (jumpNotIndexable jmpTarget: (jumpIsContext jmpTarget: cogit Label)).
	^CompletePrimitive! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'sk 4/19/2018 11:59'!
genPrimitiveStringCompareWith
	"primitiveCompareWith:"
	
	| instr jump jumpAbove jumpIncorrectFormat1 jumpIncorrectFormat2 jumpIncorrectFormat3 jumpIncorrectFormat4 jumpMidFailure jumpSuccess minSizeReg string1CharOrByteSizeReg string2CharOrByteSizeReg string1Reg string2Reg |
	
	<var: #jumpIncorrectFormat1 type: #'AbstractInstruction *'>
	<var: #jumpIncorrectFormat2 type: #'AbstractInstruction *'>
	<var: #jumpIncorrectFormat3 type: #'AbstractInstruction *'>
	<var: #jumpIncorrectFormat4 type: #'AbstractInstruction *'>
	<var: #jumpAbove type: #'AbstractInstruction *'>
	<var: #jumpSuccess type: #'AbstractInstruction *'>
	<var: #jump type: #'AbstractInstruction *'>
	<var: #jumpMidFailure type: #'AbstractInstruction *'>
	
	"I redefine those name to ease program comprehension"
	string1Reg := ReceiverResultReg.
	string2Reg := Arg0Reg.
	string1CharOrByteSizeReg := Arg1Reg.
	string2CharOrByteSizeReg := ClassReg.
	minSizeReg := SendNumArgsReg.
	
	"Load arguments in reg"
	cogit genLoadArgAtDepth: 0 into: string2Reg.
	
	"checks if string1 is a byteobject and get its size in bytes"
	self genGetFormatOf: string1Reg into: TempReg.
	cogit CmpCq: objectMemory firstByteFormat R: TempReg.
	jumpIncorrectFormat1 := cogit JumpLess: 0.
	cogit CmpCq: objectMemory firstCompiledMethodFormat R: TempReg.
	jumpIncorrectFormat2 := cogit JumpAboveOrEqual: 0.
		
	self genGetNumSlotsOf: string1Reg into: string1CharOrByteSizeReg.
	(cogit LogicalShiftLeftCq: objectMemory shiftForWord R: string1CharOrByteSizeReg). 
	cogit AndCq: objectMemory wordSize - 1 R: TempReg R: TempReg. 
	cogit SubR: TempReg R: string1CharOrByteSizeReg. 
	
	"checks if string2 is a byteobject and get its size in bytes"
	self genGetFormatOf: string2Reg into: TempReg.
	cogit CmpCq: objectMemory firstByteFormat R: TempReg.
	jumpIncorrectFormat3 := cogit JumpLess: 0.
	cogit CmpCq: objectMemory firstCompiledMethodFormat R: TempReg.
	jumpIncorrectFormat4 := cogit JumpAboveOrEqual: 0.
		
	self genGetNumSlotsOf: string2Reg into: string2CharOrByteSizeReg.
	(cogit LogicalShiftLeftCq: objectMemory shiftForWord R: string2CharOrByteSizeReg).
	cogit AndCq: objectMemory wordSize - 1 R: TempReg R: TempReg.
	cogit SubR: TempReg R: string2CharOrByteSizeReg.
	
	"Type and number of arguments are correct"
	"Compute the min"	 
	cogit CmpR: string1CharOrByteSizeReg R: string2CharOrByteSizeReg.
	jumpAbove := cogit JumpBelow: 0. 
	cogit MoveR: string1CharOrByteSizeReg R: minSizeReg. 
	jump := cogit Jump: 0. 
	jumpAbove jmpTarget: (cogit MoveR: string2CharOrByteSizeReg R: minSizeReg). 
	jump jmpTarget: (cogit CmpCq: 0 R: minSizeReg). 
	jumpSuccess := cogit JumpZero: 0. "if one of the string is empty, no need to go through the comparing loop"
	
	"Compare the bytes"
	cogit MoveCq: objectMemory baseHeaderSize  R: TempReg.
	cogit AddCq: objectMemory baseHeaderSize R: minSizeReg.
	
	instr := cogit MoveXbr: TempReg R: string1Reg R: string1CharOrByteSizeReg.
	cogit MoveXbr: TempReg R: string2Reg R: string2CharOrByteSizeReg.
	cogit SubR: string2CharOrByteSizeReg R: string1CharOrByteSizeReg. 
	jumpMidFailure := cogit JumpNonZero: 0. "the 2 compared characters are different, exit the loop"
	cogit AddCq: 1 R: TempReg.
	cogit CmpR: TempReg R: minSizeReg. 
	cogit JumpNonZero: instr.
	
	"all bytes from 1 to minSize are equal"
	self genGetNumBytesOf: string1Reg into: string1CharOrByteSizeReg.
	self genGetNumBytesOf: string2Reg into: string2CharOrByteSizeReg.
	jumpSuccess jmpTarget: (cogit SubR: string2CharOrByteSizeReg R: string1CharOrByteSizeReg).
	jumpMidFailure  jmpTarget: (cogit MoveR: string1CharOrByteSizeReg R: ReceiverResultReg).	
	self genConvertIntegerToSmallIntegerInReg: ReceiverResultReg.
	cogit genPrimReturn.
	
	jumpIncorrectFormat4 
		jmpTarget: (jumpIncorrectFormat3 
			jmpTarget: (jumpIncorrectFormat2 
				jmpTarget: (jumpIncorrectFormat1 jmpTarget: cogit Label))).
	
	^ CompletePrimitive! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'cb 3/15/2018 11:18'!
genPrimitiveStringReplace
	"replaceFrom: start to: stop with: replacement startingAt: repStart. 
	
	The primitive in the JIT tries to deal with two pathological cases, copy of arrays and byteStrings,
	which often copies only a dozen of fields and where switching to the C runtime cost a lot.
	
	Based on heuristics on the method class, I generate a quick array path (typically for Array),
	a quick byteString path (typically for ByteString, ByteArray and LargeInteger) or no quick 
	path at all (Typically for Bitmap).
	
	The many tests to ensure that the primitive won't fail are not super optimised (multiple reloading
	or stack arguments in registers) but this is still good enough and worth it since we're avoiding 
	the Smalltalk to C stack switch. The tight copying loops are optimised. 
	
	It is possible to build a bigger version with the 2 different paths but I (Clement) believe this 
	is too big machine code wise to be worth it."
	
	|arrayReg startReg stopReg replReg repStartReg jumpNotSmi1 jumpNotSmi2 jumpNotSmi3 jumpImm jumpEmpty jumpImmutable jumpOutOfBounds1 jumpOutOfBounds2 jumpOutOfBounds3 jumpOutOfBounds4 jumpIncorrectFormat1 jumpIncorrectFormat2 jumpIncorrectFormat3 jumpIncorrectFormat4 result jmpDestYoung jmpAlreadyRemembered instr adjust|
	
	<var: #jumpImm type: #'AbstractInstruction *'>
	<var: #jumpEmpty type: #'AbstractInstruction *'>
	<var: #jumpNotSmi1 type: #'AbstractInstruction *'>
	<var: #jumpNotSmi2 type: #'AbstractInstruction *'>
	<var: #jumpNotSmi3 type: #'AbstractInstruction *'>
	<var: #jmpDestYoung type: #'AbstractInstruction *'>
	<var: #jumpImmutable type: #'AbstractInstruction *'>
	<var: #jumpOutOfBounds1 type: #'AbstractInstruction *'>
	<var: #jumpOutOfBounds2 type: #'AbstractInstruction *'>
	<var: #jumpOutOfBounds3 type: #'AbstractInstruction *'>
	<var: #jumpOutOfBounds4 type: #'AbstractInstruction *'>
	<var: #jumpIncorrectFormat1 type: #'AbstractInstruction *'>
	<var: #jumpIncorrectFormat2 type: #'AbstractInstruction *'>
	<var: #jumpIncorrectFormat3 type: #'AbstractInstruction *'>
	<var: #jumpIncorrectFormat4 type: #'AbstractInstruction *'>
	<var: #jmpAlreadyRemembered type: #'AbstractInstruction *'>

	"Can I generate a quick path for this method ?"
	((cogit seemsToBeInstantiating: objectMemory arrayFormat)
		or: [cogit seemsToBeInstantiating: objectMemory firstByteFormat]) ifFalse: [^UnimplementedPrimitive].

	"I redefine those name to ease program comprehension"
	arrayReg := ReceiverResultReg.
	startReg := Arg0Reg.
	stopReg := Arg1Reg.
	replReg := ClassReg.
	repStartReg := SendNumArgsReg.
	
	"Load arguments in reg"
	cogit genStackArgAt: 0 into: repStartReg.
	cogit genStackArgAt: 1 into: replReg.
	cogit genStackArgAt: 2 into: stopReg.
	cogit genStackArgAt: 3 into: startReg.

	"start,stop,repStart Smis or fail the primitive"
	jumpNotSmi1 := self genJumpNotSmallInteger: repStartReg scratchReg: TempReg.
	jumpNotSmi2 := self genJumpNotSmallInteger: stopReg scratchReg: TempReg.
	jumpNotSmi3 := self genJumpNotSmallInteger: startReg scratchReg: TempReg.
	"repl non immediate or fail the primitive"
	jumpImm := self genJumpImmediate: replReg.
	
	"if start>stop primitive success"
	cogit CmpR: startReg R: stopReg.
	jumpEmpty := cogit JumpLess: 0.
	
	"If receiver immutable fail the primitive "
	self
		cppIf: IMMUTABILITY
		ifTrue: [jumpImmutable := self genJumpImmutable: ReceiverResultReg scratchReg: TempReg].
		
	"0 >= start, fail"
	cogit CmpCq: (objectMemory integerObjectOf: 0) R: startReg.
	jumpOutOfBounds1 := cogit JumpLessOrEqual: 0.
	
	"0 >= replStart, fail"
	cogit CmpCq: (objectMemory integerObjectOf: 0) R: repStartReg.
	jumpOutOfBounds2 := cogit JumpLessOrEqual: 0.

	"--- Pointer object version ---"
	(cogit seemsToBeInstantiating: objectMemory arrayFormat) ifTrue:
		["Are they both array format ?"
		self genGetFormatOf: arrayReg into: TempReg.
		self genGetFormatOf: replReg into: startReg.
		cogit CmpCq: objectMemory arrayFormat R: startReg.
		jumpIncorrectFormat1 := cogit JumpNonZero: 0.
		cogit CmpCq: objectMemory arrayFormat R: TempReg.
		jumpIncorrectFormat2 := cogit JumpNonZero: 0.
	
		"Both objects are arrays,"
		self genGetNumSlotsOf: arrayReg into: TempReg.
		self genConvertSmallIntegerToIntegerInReg: stopReg.
	
		"arr size < stop"
		cogit CmpR: TempReg R: stopReg.
		jumpOutOfBounds3 := cogit JumpGreater: 0.
	
		"rep size < repStart - start + stop"
		self genGetNumSlotsOf: replReg into: TempReg.
		cogit genStackArgAt: 3 into: startReg.
		self genConvertSmallIntegerToIntegerInReg: startReg.
		self genConvertSmallIntegerToIntegerInReg: repStartReg.
		cogit SubR: startReg R: stopReg.
		cogit AddR: repStartReg R: stopReg.
		"stopReg: stop - start + repStart"
		cogit CmpR: TempReg R: stopReg.
		jumpOutOfBounds4 := cogit JumpGreater: 0.
	
		"Everything in bounds"
		"PossibleRemembered object"
		cogit MoveCw: objectMemory storeCheckBoundary R: TempReg.
		cogit CmpR: TempReg R: arrayReg.
		jmpDestYoung := cogit JumpBelow: 0.
		self checkRememberedInTrampoline ifFalse: 
			[jmpAlreadyRemembered := self genCheckRememberedBitOf: arrayReg scratch: TempReg].
		cogit backEnd saveAndRestoreLinkRegAround:
			[self callStoreCheckTrampoline].
		jmpDestYoung jmpTarget: cogit Label.
		self checkRememberedInTrampoline ifFalse: 
			[jmpAlreadyRemembered jmpTarget: cogit Label].
	
		"Copy the array"
		cogit genStackArgAt: 2 into: stopReg.
		self genConvertSmallIntegerToIntegerInReg: stopReg.
		"Shift replReg to read it with startReg offset"
		cogit SubR: startReg R: repStartReg. 
		cogit LogicalShiftLeftCq: objectMemory shiftForWord R: repStartReg.
		cogit AddR: repStartReg R: replReg. 
		"shift by baseHeaderSize and then move from 1 relative to zero relative"
		adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1. 
		adjust ~= 0 ifTrue: 
			[ cogit AddCq: adjust R: startReg. 
			  cogit AddCq: adjust R: stopReg ].
	
		"Tight copying loop"
		instr := cogit MoveXwr: startReg R: replReg R: TempReg.
		cogit MoveR: TempReg Xwr: startReg R: arrayReg.
		cogit AddCq: 1 R: startReg.
		cogit CmpR: startReg R: stopReg.
		cogit JumpAboveOrEqual: instr.
		
		jumpEmpty jmpTarget: cogit genPrimReturn.
	
		"CANNOT REACH by falling though"
	
		jumpIncorrectFormat1 jmpTarget: (jumpIncorrectFormat2 jmpTarget: cogit Label)].
	
	"--- Byte object version ---"
	(cogit seemsToBeInstantiating: objectMemory firstByteFormat) ifTrue:
		["Are they both byte array format ? CompiledMethod excluded"
		self genGetFormatOf: arrayReg into: TempReg.
		self genGetFormatOf: replReg into: repStartReg.
		cogit CmpCq: objectMemory firstByteFormat R: repStartReg.
		jumpIncorrectFormat1 := cogit JumpLess: 0.
		cogit CmpCq: objectMemory firstCompiledMethodFormat R: repStartReg.
		jumpIncorrectFormat2 := cogit JumpGreaterOrEqual: 0.
		cogit CmpCq: objectMemory firstByteFormat R: TempReg.
		jumpIncorrectFormat3 := cogit JumpLess: 0.
		cogit CmpCq: objectMemory firstCompiledMethodFormat R: TempReg.
		jumpIncorrectFormat4 := cogit JumpGreaterOrEqual: 0.
		
		"Both objects are byte arrays"
		self genGetNumSlotsOf: arrayReg into: startReg.
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord R: startReg).
		cogit AndCq: objectMemory wordSize - 1 R: TempReg R: TempReg.
		cogit SubR: TempReg R: startReg.
		self genConvertSmallIntegerToIntegerInReg: stopReg.
	
		"arr size < stop"
		cogit CmpR: startReg R: stopReg.
		jumpOutOfBounds3 := cogit JumpGreater: 0.
	
		"rep size < repStart - start + stop"
		cogit MoveR: repStartReg R: TempReg. "TempReg holds repl format"
		cogit genStackArgAt: 0 into: repStartReg.
		cogit genStackArgAt: 3 into: startReg.
		self genConvertSmallIntegerToIntegerInReg: startReg.
		self genConvertSmallIntegerToIntegerInReg: repStartReg.
		cogit SubR: startReg R: stopReg.
		cogit AddR: repStartReg R: stopReg.
		
		self genGetNumSlotsOf: replReg into: startReg.
		(cogit LogicalShiftLeftCq: objectMemory shiftForWord R: startReg).
		cogit AndCq: objectMemory wordSize - 1 R: TempReg R: TempReg.
		cogit SubR: TempReg R: startReg.
		
		"stopReg: stop - start + repStart"
		cogit CmpR: startReg R: stopReg.
		jumpOutOfBounds4 := cogit JumpGreater: 0.
	
		"Everything in bounds"
		"Copy the bytes"
		cogit genStackArgAt: 3 into: startReg.
		self genConvertSmallIntegerToIntegerInReg: startReg.
		cogit genStackArgAt: 2 into: stopReg.
		self genConvertSmallIntegerToIntegerInReg: stopReg.
		"Shift replReg to read it with startReg offset"
		cogit SubR: startReg R: repStartReg. 
		cogit AddR: repStartReg R: replReg. 
		"shift by baseHeaderSize and then move from 1 relative to zero relative"
		adjust := objectMemory baseHeaderSize - 1.
		adjust ~= 0 ifTrue: 
			[ cogit AddCq: adjust R: startReg. 
			  cogit AddCq: adjust R: stopReg.  ].
	
		"Tight copying loop"
		instr := cogit MoveXbr: startReg R: replReg R: TempReg.
		cogit MoveR: TempReg Xbr: startReg R: arrayReg.
		cogit AddCq: 1 R: startReg.
		cogit CmpR: startReg R: stopReg.
		cogit JumpAboveOrEqual: instr.
		
		jumpEmpty jmpTarget: cogit genPrimReturn.
	
		"CANNOT REACH by falling though"
	
		jumpIncorrectFormat4 
			jmpTarget: (jumpIncorrectFormat3 
				jmpTarget: (jumpIncorrectFormat2 
					jmpTarget: (jumpIncorrectFormat1 jmpTarget: cogit Label)))].

	(result := cogit compileInterpreterPrimitive) < 0 ifTrue: [^result].
	
	jumpImm 
		jmpTarget: (jumpNotSmi1
				jmpTarget: (jumpNotSmi2
						jmpTarget: (jumpNotSmi3 jmpTarget: cogit Label))).
	jumpOutOfBounds1 
		jmpTarget: (jumpOutOfBounds2 
			jmpTarget: (jumpOutOfBounds3 
				jmpTarget: (jumpOutOfBounds4 
					jmpTarget: jumpImm getJmpTarget))).
	self
		cppIf: IMMUTABILITY
		ifTrue: [jumpImmutable jmpTarget: jumpImm getJmpTarget].

	^ CompletePrimitive! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/24/2013 16:53'!
genRemoveCharacterTagsInScratchReg: scratchReg
	cogit SubCq: objectMemory characterTag R: scratchReg! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/1/2013 17:49'!
genRemoveSmallIntegerTagsInScratchReg: scratchReg
	self subclassResponsibility! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 12/24/2013 18:38'!
genSetCharacterTagsIn: reg
	cogit OrCq: objectMemory characterTag R: reg! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'eem 6/23/2017 12:40'!
genSetGCNeeded
	<inline: true>
	cogit
		MoveCq: 1 R: TempReg;
		MoveR: TempReg Aw: objectMemory needGCFlagAddress! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/1/2013 17:49'!
genSetSmallIntegerTagsIn: scratchReg
	cogit OrCq: 1 R: scratchReg.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 1/9/2020 16:38'!
genSetSmallIntegerTagsIn: sourceReg into: destReg
	<inline: true>
	^cogit LoadEffectiveAddressMw: 1 r: sourceReg R: destReg! !

!CogObjectRepresentationForSpur methodsFor: 'initialization' stamp: 'eem 6/4/2015 13:10'!
genStoreCheckContextReceiverTrampoline
	"Create a trampoline to store-check the update of the receiver in a closure's
	 outerContext in compileBlockFrameBuild:."
	| startAddress |
	startAddress := cogit methodZoneBase.
	cogit zeroOpcodeIndex.
	self genStoreCheckReceiverReg: ReceiverResultReg valueReg: Arg0Reg scratchReg: TempReg inFrame: false.
	cogit RetN: 0.
	cogit outputInstructionsForGeneratedRuntimeAt: startAddress.
	cogit recordGeneratedRunTime: 'ceStoreCheckContextReceiver' address: startAddress.
	cogit recordRunTimeObjectReferences.
	^startAddress! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 1/1/1970 13:06'!
genStoreCheckReceiverReg: destReg valueReg: valueReg scratchReg: scratchReg inFrame: inFrame
	"Generate the code for a store check of valueReg into destReg."
	| jmpImmediate jmpDestYoung jmpSourceOld jmpAlreadyRemembered |
	<var: #jmpImmediate type: #'AbstractInstruction *'>
	<var: #jmpDestYoung type: #'AbstractInstruction *'>
	<var: #jmpSourceOld type: #'AbstractInstruction *'>
	<var: #jmpAlreadyRemembered type: #'AbstractInstruction *'>
	"Is value stored an immediate?  If so we're done"
	jmpImmediate := self genJumpImmediate: valueReg.
	"Get the old/new boundary in scratchReg"
	cogit MoveCw: objectMemory storeCheckBoundary R: scratchReg.
	"Is target young?  If so we're done"
	cogit CmpR: scratchReg R: destReg. "N.B. FLAGS := destReg - scratchReg"
	jmpDestYoung := cogit JumpBelow: 0.
	"Is value stored old?  If so we're done."
	cogit CmpR: scratchReg R: valueReg. "N.B. FLAGS := valueReg - scratchReg"
	jmpSourceOld := cogit JumpAboveOrEqual: 0.
	"value is young and target is old.
	 Need to remember this only if the remembered bit is not already set."
	CheckRememberedInTrampoline ifFalse:
		[jmpAlreadyRemembered := self genCheckRememberedBitOf: destReg scratch: scratchReg].
	"Remembered bit is not set.  Call store check to insert dest into remembered table."
	self assert: destReg = ReceiverResultReg.
	cogit 
		evaluateTrampolineCallBlock: [cogit CallRT: ceStoreCheckTrampoline]
		protectLinkRegIfNot: inFrame.
	jmpImmediate jmpTarget:
	(jmpDestYoung jmpTarget:
	(jmpSourceOld jmpTarget:
		cogit Label)).
	CheckRememberedInTrampoline ifFalse:
		[jmpAlreadyRemembered jmpTarget: jmpSourceOld getJmpTarget].
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'initialization' stamp: 'eem 02/04/2020 13:04'!
genStoreCheckTrampoline
	| jumpSC |
	<var: #jumpSC type: #'AbstractInstruction *'>
	<inline: true>
	CheckRememberedInTrampoline ifTrue:
		[cogit zeroOpcodeIndex.
		 jumpSC := self genCheckRememberedBitOf: ReceiverResultReg scratch: ABIResultReg.
		 self assert: jumpSC opcode = JumpNonZero.
		 jumpSC opcode: JumpZero.
		 cogit RetN: 0.
		 jumpSC jmpTarget: cogit Label].
	^cogit
		genTrampolineFor: #remember:
		called: 'ceStoreCheckTrampoline'
		numArgs: 1
		arg: ReceiverResultReg
		arg: nil
		arg: nil
		arg: nil
		regsToSave: (CallerSavedRegisterMask bitClear: (cogit registerMaskFor: ReceiverResultReg))
		pushLinkReg: true
		resultReg: cogit returnRegForStoreCheck
		appendOpcodes: CheckRememberedInTrampoline! !

!CogObjectRepresentationForSpur methodsFor: 'initialization' stamp: 'eem 12/7/2015 16:12'!
genStoreHeader: header intoNewInstance: rcvrReg using: scratchReg
	"Generate the instructions to move the constant header into a new instance pointed to by rcvrReg."
	<inline: true>
	self subclassResponsibility! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'cb 3/30/2016 15:06'!
genStoreSourceReg: sourceReg slotIndex: index destReg: destReg scratchReg: scratchReg inFrame: inFrame needsStoreCheck: needsStoreCheck
	
	cogit genTraceStores.
	"do the store"
	cogit MoveR: sourceReg
		   Mw: index * objectMemory wordSize + objectMemory baseHeaderSize
		   r: destReg.
	"now the check. needStoreCheck is false if the JIT has figured out that the value stored does not need the check (immediate, nil, true, false)"
	needsStoreCheck ifTrue: 
		[ ^ self 
			genStoreCheckReceiverReg: destReg 
			valueReg: sourceReg 
			scratchReg: scratchReg 
			inFrame: inFrame ].
	^ 0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'cb 11/23/2015 11:22'!
genStoreSourceReg: sourceReg slotIndex: index intoNewObjectInDestReg: destReg
	"This method is used for unchecked stores in objects after their creation (typically, inlined creation of Array, closures and some temp vectors). 
	Currently there is no need to do the immutability check here"
	cogit MoveR: sourceReg
		   Mw: index * objectMemory wordSize + objectMemory baseHeaderSize
		   r: destReg.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 3/10/2018 10:16'!
genStoreTrampolineCall: instVarIndex
	<inline: true>
	self assert: IMMUTABILITY.
	instVarIndex >= (NumStoreTrampolines - 1)
		ifTrue: 
			[ cogit MoveCq: instVarIndex R: TempReg.
			  cogit CallRT: (ceStoreTrampolines at: NumStoreTrampolines - 1) ]
		ifFalse: 
			[ cogit CallRT: (ceStoreTrampolines at: instVarIndex) ].
	 cogit annotateBytecode: cogit Label! !

!CogObjectRepresentationForSpur methodsFor: 'initialization' stamp: 'eem 1/29/2020 15:28'!
genStoreTrampolineCalled: trampolineName instVarIndex: instVarIndex
	"Convention:
	- RcvrResultReg holds the object mutated.
	If immutability failure:
	- TempReg holds the instance variable index mutated 
		if instVarIndex > numDedicatedStoreTrampoline
	- ClassReg holds the value to store
	Registers are not lived across this trampoline as the 
	immutability failure may need new stack frames."
	
	<option: #IMMUTABILITY>
	<var: 'trampolineName' type: #'char *'>
	<inline: false>
	| startAddress |
	startAddress := cogit methodZoneBase.
	cogit zeroOpcodeIndex.
	CheckRememberedInTrampoline
		ifTrue:
			[self genStoreTrampolineCheckingRememberedCalled: trampolineName instVarIndex: instVarIndex]
		ifFalse:
			[self genStoreTrampolineNotCheckingRememberedCalled: trampolineName instVarIndex: instVarIndex].
		
	cogit outputInstructionsForGeneratedRuntimeAt: startAddress.
	cogit recordGeneratedRunTime: trampolineName address: startAddress.
	cogit recordRunTimeObjectReferences.
	^startAddress! !

!CogObjectRepresentationForSpur methodsFor: 'initialization' stamp: 'eem 1/8/2020 13:09'!
genStoreTrampolineCheckingRememberedCalled: trampolineName instVarIndex: instVarIndex
	"Convention:
	- RcvrResultReg holds the object to be mutated.
	If immutability failure:
	- TempReg holds the instance variable index mutated 
		if instVarIndex > numDedicatedStoreTrampoline
	- ClassReg holds the value to store
	Registers are not live across this trampoline as the 
	immutability failure may need new stack frames."
	
	| jumpImmutable jumpRC |
	<option: #IMMUTABILITY>
	<var: #trampolineName type: #'char *'>
	<inline: true>
	"SendNumArgsReg is mutated but we don't care as registers are not live across the trampoline.
	 There is no reason why registers cannot be saved over the remember: call, but since the
	 immutability check is a suspension point, registers cannot remain live."
	
	"Jump in the uncommon case"
	jumpImmutable := self genJumpImmutable: ReceiverResultReg scratchReg: SendNumArgsReg.
	"Store check"
	"If on 64-bits and doing the remembered bit test here, we can combine the tests to fetch the header once."
	objectMemory wordSize = 8
		ifTrue:
			[cogit TstCq: 1 << objectMemory rememberedBitShift R: SendNumArgsReg.
			 jumpRC := cogit JumpZero: 0]
		ifFalse:
			[jumpRC := self genCheckRememberedBitOf: ReceiverResultReg scratch: SendNumArgsReg.
			 self assert: jumpRC opcode = JumpNonZero.
			 jumpRC opcode: JumpZero].
	cogit RetN: 0.
	jumpRC jmpTarget: cogit Label.
	cogit
		compileTrampolineFor: #remember:
		numArgs: 1
		arg: ReceiverResultReg
		arg: nil
		arg: nil
		arg: nil
		regsToSave: cogit emptyRegisterMask
		pushLinkReg: true
		resultReg: cogit returnRegForStoreCheck.
	jumpImmutable jmpTarget: cogit Label.
	cogit
		compileTrampolineFor: #ceCannotAssignTo:withIndex:valueToAssign:
		numArgs: 3
		arg: ReceiverResultReg
		arg: (instVarIndex < (NumStoreTrampolines - 1)
				ifTrue: [cogit trampolineArgConstant: instVarIndex]
				ifFalse: [TempReg])
		arg: ClassReg
		arg: nil
		regsToSave: cogit emptyRegisterMask
		pushLinkReg: true
		resultReg: NoReg.
	^0
! !

!CogObjectRepresentationForSpur methodsFor: 'initialization' stamp: 'eem 2/1/2020 18:49'!
genStoreTrampolineNotCheckingRememberedCalled: trampolineName instVarIndex: instVarIndex
	"Convention:
	- RcvrResultReg holds the object to be mutated.
	If immutability failure:
	- TempReg holds the instance variable index mutated 
		if instVarIndex > numDedicatedStoreTrampoline
	- ClassReg holds the value to store
	Registers are not live across this trampoline as the 
	immutability failure may need new stack frames."
	
	| jumpImmutable pushLinkReg |
	<option: #IMMUTABILITY>
	<var: #trampolineName type: #'char *'>
	<inline: true>
	"We are  going to call something, either remember: in the common case,
	 or much more rarely ceCannotAssignTo:withIndex:valueToAssign:.  So
	 share the stack switch between the two."
	cogit genSmalltalkToCStackSwitch: (pushLinkReg := cogit backEnd hasLinkRegister).
	"SendNumArgsReg is mutated but we don't care as registers are not live across the trampoline.
	 There is no reason why registers cannot be saved over the remember: call, but since the
	 immutability check is a suspension point, registers cannot remain live."
	jumpImmutable := self genJumpImmutable: ReceiverResultReg scratchReg: SendNumArgsReg.
	"Store check"
	cogit
		compileCallFor: #remember:
		numArgs: 1
		arg: ReceiverResultReg
		arg: nil
		arg: nil
		arg: nil
		resultReg: cogit returnRegForStoreCheck
		regsToSave: cogit emptyRegisterMask.
	cogit backEnd genLoadStackPointers.
	cogit genTrampolineReturn: pushLinkReg.
	jumpImmutable jmpTarget: cogit Label.
	cogit
		compileCallFor: #ceCannotAssignTo:withIndex:valueToAssign:
		numArgs: 3
		arg: ReceiverResultReg
		arg: (instVarIndex < (NumStoreTrampolines - 1)
				ifTrue: [cogit trampolineArgConstant: instVarIndex]
				ifFalse: [TempReg])
		arg: ClassReg
		arg: nil
		resultReg: NoReg
		regsToSave: cogit emptyRegisterMask.
	cogit backEnd genLoadStackPointers.
	cogit genTrampolineReturn: pushLinkReg.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'inline primitive support' stamp: 'cb 3/6/2018 12:03'!
genStoreValue: value instance: destReg numSlots: numSlots 
	"Store the value in the instance from field 1 to numSlots
	 Typically used for for inlined allocations, initializing objects with 0 or nil.
	 destReg is referencing the object (oop) and is restored at the end of this code.
	 If less than 8 fields, use a full unrolled initialization (up to 8 stores)
	 If more than 8 fields, use a duff device to initialize with a 8-vectorized loop."

	| slotsPerIteration constReg delta branch loopCount loop inst |
	numSlots <= (slotsPerIteration := 8) ifTrue: "slotsPerIteration must be even; see cogit SubCq: objectMemory bytesPerOop R: TempReg below"
		[cogit genMoveConstant: value R: TempReg.
		 0 to: numSlots - 1 do:
			[:i| cogit MoveR: TempReg
					Mw: i * objectMemory wordSize + objectMemory baseHeaderSize
					r: destReg].
		^0].
	
	constReg := cogit allocateRegNotConflictingWith: (cogit registerMaskFor: destReg).
	cogit genMoveConstant: value R: constReg.
	
	numSlots \\ slotsPerIteration ~= 0
		ifTrue: "delta maps the offset at the loop entryPoint onto destReg + objectMemory baseHeaderSize"
			[delta := (slotsPerIteration - (numSlots \\ slotsPerIteration) * objectMemory bytesPerOop) - objectMemory baseHeaderSize.
			 delta > 0 ifTrue: [cogit SubCq: delta R: destReg].
			 delta < 0 ifTrue: [cogit AddCq: delta negated R: destReg].
			 "now delta maps (loopCount * slotsPerIteration * objectMemory bytesPerOop) + objectMemory baseHeaderSize - delta to the start of the object"
			 delta := delta + objectMemory baseHeaderSize.
			 (objectMemory bytesPerOop < objectMemory baseHeaderSize
			  and: [numSlots \\ 2 = 1]) ifTrue: "if end of loop is not at start of next object, adjust loop limit in TempReg to point to last field filled."
				[cogit SubCq: objectMemory bytesPerOop R: TempReg].
			 branch := cogit Jump: 0]
		ifFalse:
			[delta := 0.
			 cogit AddCq: objectMemory baseHeaderSize R: destReg].
	"loopCount is number of times through the increment of destReg."
	loopCount := numSlots + slotsPerIteration - 1 // slotsPerIteration.
	self assert: loopCount > 1.
	loop := cogit Label.
	0 to: 7 do:
		[:i|
		inst := cogit MoveR: constReg Mw: i * objectMemory bytesPerOop r: destReg.
		slotsPerIteration - (numSlots \\ slotsPerIteration) = i ifTrue:
			[branch jmpTarget: inst]].
	cogit
		AddCq: slotsPerIteration * objectMemory bytesPerOop R: destReg;
		CmpR: TempReg R: destReg;
		JumpBelow: loop;
		SubCq: (loopCount * slotsPerIteration * objectMemory bytesPerOop) + objectMemory baseHeaderSize - delta R: destReg.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 7/24/2018 20:29'!
genStoreWithImmutabilityAndStoreCheckSourceReg: sourceReg slotIndex: index destReg: destReg scratchReg: scratchReg needRestoreRcvr: needRestoreRcvr 
	"Store check code is duplicated to use a single trampoline"
	<option: #IMMUTABILITY>
	| immutableJump jmpImmediate jmpDestYoung jmpSourceOld jmpAlreadyRemembered |
	<var: #immutableJump type: #'AbstractInstruction *'>
	<var: #jmpImmediate type: #'AbstractInstruction *'>
	<var: #jmpDestYoung type: #'AbstractInstruction *'>
	<var: #jmpSourceOld type: #'AbstractInstruction *'>
	<var: #jmpAlreadyRemembered type: #'AbstractInstruction *'>

	immutableJump := self genJumpImmutable: destReg scratchReg: scratchReg.
	
	cogit genTraceStores.
	
	"do the store"
	cogit MoveR: sourceReg
		   Mw: index * objectMemory wordSize + objectMemory baseHeaderSize
		   r: destReg.
	
	"store check"
	jmpImmediate := self genJumpImmediate: sourceReg.
	"Get the old/new boundary in scratchReg"
	cogit MoveCw: objectMemory storeCheckBoundary R: scratchReg.
	"Is target young?  If so we're done"
	cogit CmpR: scratchReg R: destReg. "N.B. FLAGS := destReg - scratchReg"
	jmpDestYoung := cogit JumpBelow: 0.
	"Is value stored old?  If so we're done."
	cogit CmpR: scratchReg R: sourceReg. "N.B. FLAGS := valueReg - scratchReg"
	jmpSourceOld := cogit JumpAboveOrEqual: 0.
	"value is young and target is old.
	 Need to remember this only if the remembered bit is not already set."
	CheckRememberedInTrampoline ifFalse:
		[jmpAlreadyRemembered := self genCheckRememberedBitOf: destReg scratch: scratchReg].
	"Set the inst var index for the benefit of the immutability check. The trampoline will
	 repeat the check to choose between the immutbality violation and the store check."
	immutableJump jmpTarget: cogit Label.
	self genStoreTrampolineCall: index.
	cogit voidReceiverOptStatus.
	needRestoreRcvr ifTrue:
		[cogit putSelfInReceiverResultReg].

	jmpImmediate jmpTarget:
	(jmpDestYoung jmpTarget:
	(jmpSourceOld jmpTarget:
		cogit Label)).
	CheckRememberedInTrampoline ifFalse:
		[jmpAlreadyRemembered jmpTarget: jmpSourceOld getJmpTarget].
	^ 0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 7/24/2018 20:29'!
genStoreWithImmutabilityButNoStoreCheckSourceReg: sourceReg slotIndex: index destReg: destReg scratchReg: scratchReg needRestoreRcvr: needRestoreRcvr
	"Gen an immutability check with no store check (e.g. assigning an immediate literal)"
	<option: #IMMUTABILITY>
	<var: #mutableJump type: #'AbstractInstruction *'>
	<var: #immutabilityFailure type: #'AbstractInstruction *'>
	| immutabilityFailure mutableJump |
	
	"imm check has its own trampoline"
	mutableJump := self genJumpMutable: destReg scratchReg: scratchReg.
	self genStoreTrampolineCall: index.
	cogit voidReceiverOptStatus.
	needRestoreRcvr ifTrue:
		[cogit putSelfInReceiverResultReg].
	immutabilityFailure := cogit Jump: 0.
	mutableJump jmpTarget: cogit Label.

	cogit genTraceStores.
	
	"do the store"
	cogit MoveR: sourceReg
		   Mw: index * objectMemory wordSize + objectMemory baseHeaderSize
		   r: destReg.
		
	immutabilityFailure jmpTarget: cogit Label.

	^ 0! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'cb 7/25/2016 11:12'!
genStoreWithImmutabilityCheckSourceReg: sourceReg slotIndex: index destReg: destReg scratchReg: scratchReg needsStoreCheck: needsStoreCheck needRestoreRcvr: needRestoreRcvr
	<option: #IMMUTABILITY>
	"We know there is a frame as immutability check requires a frame"
	"needRestoreRcvr has to be true to keep RcvrResultReg live with the receiver in it across the trampoline"
	
	"Trampoline convention..."
	self assert: destReg == ReceiverResultReg.
	self assert: scratchReg == TempReg.
  	self assert: sourceReg == ClassReg.
	
	"The two paths are very different as the trampoline is shared in 1 case so we split here"
	needsStoreCheck
		ifTrue: 
			[ self 
				genStoreWithImmutabilityAndStoreCheckSourceReg: sourceReg 
				slotIndex: index 
				destReg: destReg 
				scratchReg: scratchReg 
				needRestoreRcvr: needRestoreRcvr ]
		ifFalse: 
			[ self 
				genStoreWithImmutabilityButNoStoreCheckSourceReg: sourceReg 
				slotIndex: index 
				destReg: destReg 
				scratchReg: scratchReg 
				needRestoreRcvr: needRestoreRcvr ].
	^ 0! !

!CogObjectRepresentationForSpur methodsFor: 'mapped inlined primitive support' stamp: 'cb 12/11/2017 15:53'!
genVarIndexCallStoreTrampoline	
	self assert: IMMUTABILITY.
	self cppIf: #IMMUTABILITY
		ifTrue: [cogit CallRT: (ceStoreTrampolines at: NumStoreTrampolines - 1).
				cogit annotateBytecode: cogit Label ].! !

!CogObjectRepresentationForSpur methodsFor: 'initialization' stamp: 'eem 3/16/2017 17:39'!
generateObjectRepresentationTrampolines
	"Do the store check.  Answer the argument for the benefit of the code generator;
	 ReceiverResultReg may be caller-saved and hence smashed by this call.  Answering
	 it allows the code generator to reload ReceiverResultReg cheaply.
	 In Spur the only thing we leave to the run-time is adding the receiver to the
	 remembered set and setting its isRemembered bit."
	self 
		cppIf: IMMUTABILITY
		ifTrue: 
			[self cCode: [] inSmalltalk:
				[ceStoreTrampolines := CArrayAccessor on: (Array new: NumStoreTrampolines)].
			 0 to: NumStoreTrampolines - 1 do:
				[:instVarIndex |
				 ceStoreTrampolines
					at: instVarIndex
					put: (self 
							genStoreTrampolineCalled: (cogit 
															trampolineName: 'ceStoreTrampoline' 
															numArgs: instVarIndex 
															limit: NumStoreTrampolines - 2) 
							instVarIndex: instVarIndex)]].
	ceNewHashTrampoline := self genNewHashTrampoline: false called: 'ceNewHash'.
	SistaVM ifTrue: [ceInlineNewHashTrampoline := self genNewHashTrampoline: true  called: 'ceInlineNewHash'].
	ceStoreCheckTrampoline := self genStoreCheckTrampoline.
	ceStoreCheckContextReceiverTrampoline := self genStoreCheckContextReceiverTrampoline.
	ceScheduleScavengeTrampoline := cogit
											genTrampolineFor: #ceScheduleScavenge
											called: 'ceScheduleScavengeTrampoline'
											regsToSave: CallerSavedRegisterMask.
	ceSmallActiveContextInMethodTrampoline := self genActiveContextTrampolineLarge: false inBlock: 0 called: 'ceSmallMethodContext'.
	ceSmallActiveContextInBlockTrampoline := self genActiveContextTrampolineLarge: false inBlock: InVanillaBlock called: 'ceSmallBlockContext'.
	SistaV1BytecodeSet ifTrue:
		[ceSmallActiveContextInFullBlockTrampoline := self genActiveContextTrampolineLarge: false inBlock: InFullBlock called: 'ceSmallFullBlockContext'].
	ceLargeActiveContextInMethodTrampoline := self genActiveContextTrampolineLarge: true inBlock: 0 called: 'ceLargeMethodContext'.
	ceLargeActiveContextInBlockTrampoline := self genActiveContextTrampolineLarge: true inBlock: InVanillaBlock called: 'ceLargeBlockContext'.
	SistaV1BytecodeSet ifTrue:
		[ceLargeActiveContextInFullBlockTrampoline := self genActiveContextTrampolineLarge: true inBlock: InFullBlock called: 'ceLargeFullBlockContext'].
		
	LowcodeVM ifTrue: [ self generateLowcodeObjectTrampolines ]! !

!CogObjectRepresentationForSpur methodsFor: 'bytecode generator support' stamp: 'eem 5/31/2014 16:49'!
getActiveContextAllocatesInMachineCode
	"Make sure SendNumArgsReg and ClassReg are available in addition to
	 ReceiverResultReg and TempReg in genGetActiveContextNumArgs:large:inBlock:."
	^true! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 1/23/2016 12:23'!
getLiteralCountOf: methodReg plusOne: plusOne inBytes: inBytes into: litCountReg scratch: scratchReg
	"Get the literal count of a CompiledMethod into headerReg, plus one if requested.
	  If inBytes is true, scale the count by the word size.  Deal with the possibility of
	 the method being cogged."
	self subclassResponsibility! !

!CogObjectRepresentationForSpur methodsFor: 'testing' stamp: 'eem 8/3/2021 18:51'!
hasSpurMemoryManagerAPI
	<inline: #always>
	^true! !

!CogObjectRepresentationForSpur methodsFor: 'garbage collection' stamp: 'eem 10/1/2013 17:16'!
hasValidHeaderPostGC: cogMethod
	<var: #cogMethod type: #'CogMethod *'>
	<inline: true>
	^cogMethod objectHeader = objectMemory nullHeaderForMachineCodeMethod! !

!CogObjectRepresentationForSpur methodsFor: 'garbage collection' stamp: 'eem 11/18/2013 15:17'!
inlineCacheTagIsYoung: cacheTag
	"Since all cache tags in Spur are class indices none of
	 them are young or have to be updated in a scavenge."
	^false! !

!CogObjectRepresentationForSpur methodsFor: 'in-line cacheing' stamp: 'eem 01/10/2016 10:01'!
inlineCacheTagsMayBeObjects
	^false! !

!CogObjectRepresentationForSpur methodsFor: 'object representation' stamp: 'eem 8/3/2021 18:54'!
isSmallIntegerTagNonZero
	<inline: #always>
	^true! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 1/11/2016 17:58'!
jumpNotCharacterUnsignedValueInRegister: reg
	cogit CmpCq: 1 << self numCharacterBits - 1 R: reg.
	^cogit JumpAbove: 0! !

!CogObjectRepresentationForSpur methodsFor: 'garbage collection' stamp: 'eem 11/18/2013 15:39'!
markAndTraceLiteral: literal
	(self couldBeObject: literal) ifTrue:
		[self assert: (objectMemory addressCouldBeObj: literal).
		 objectMemory markAndTrace: literal]! !

!CogObjectRepresentationForSpur methodsFor: 'garbage collection' stamp: 'eem 8/4/2014 10:52'!
markAndTraceLiteral: literal in: cogMethod at: address
	"Mark and trace a literal in a sqInt variable of cogMethod."
	<var: #cogMethod type: #'CogMethod *'>
	<var: #address type: #'sqInt *'>
	| objOop |
	(self couldBeObject: literal) ifFalse:
		[^self].
	self assert: (objectMemory addressCouldBeObj: literal).
	(objectMemory isForwarded: literal) ifFalse:
		[objectMemory markAndTrace: literal.
		 ^self].
	objOop := objectMemory followForwarded: literal.
	address at: 0 put: objOop.
	self markAndTraceUpdatedLiteral: objOop in: cogMethod! !

!CogObjectRepresentationForSpur methodsFor: 'garbage collection' stamp: 'eem 7/9/2021 14:10'!
markAndTraceLiteral: literal in: cogMethodOrNil atpc: address
	"Mark and trace a literal in a machine code instruction preceding address in cogMethodOrNil.
	 Answer if code was modified."
	<var: #cogMethodOrNil type: #'CogMethod *'>
	<var: #address type: #usqInt>
	| objOop |
	(self couldBeObject: literal) ifFalse:
		[^false].
	self assert: (objectMemory addressCouldBeObj: literal).
	(objectMemory isForwarded: literal) ifFalse:
		[objectMemory markAndTrace: literal.
		 ^false].
	cogit setCodeModified.
	objOop := objectMemory followForwarded: literal.
	cogit backEnd storeLiteral: objOop beforeFollowingAddress: address.
	self markAndTraceUpdatedLiteral: objOop in: cogMethodOrNil.
	^true! !

!CogObjectRepresentationForSpur methodsFor: 'garbage collection' stamp: 'eem 3/12/2014 17:08'!
markAndTraceLiteralIfYoung: literal
	((self couldBeObject: literal)
	 and: [objectMemory isYoungObject: literal]) ifTrue:
		[self assert: (objectMemory addressCouldBeObj: literal).
		 objectMemory markAndTrace: literal]! !

!CogObjectRepresentationForSpur methodsFor: 'garbage collection' stamp: 'eem 8/4/2014 17:54'!
markAndTraceUpdatedLiteral: objOop in: cogMethodOrNil
	"Common code to mark a literal in cogMethod and add
	 the cogMethod to youngReferrers if the literal is young."
	<var: #cogMethodOrNil type: #'CogMethod *'>
	(objectMemory isNonImmediate: objOop) ifTrue:
		[(cogMethodOrNil notNil
		  and: [objectMemory isYoungObject: objOop]) ifTrue:
			[methodZone ensureInYoungReferrers: cogMethodOrNil].
		 objectMemory markAndTrace: objOop]! !

!CogObjectRepresentationForSpur methodsFor: 'newspeak support' stamp: 'eem 1/25/2017 16:55'!
markIfIRC: maybeIRCs
	"If maybeIRCs (from some cogMehtod's nextMethodOrIRCs) is in old space it is
	 a pointer to the first field of a pinned object in old space holding the implicit
	 receiver caches for a method.  If so, map it back to an oop and mark it."
	<var: #maybeIRCs type: #usqInt>
	<option: #NewspeakVM>
	<inline: true>
	(self oop: maybeIRCs isGreaterThan: objectMemory nilObject) ifTrue:
		[objectMemory markAndTrace: maybeIRCs - objectMemory baseHeaderSize]! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'eem 6/6/2016 18:38'!
maybeCompileRetryOnPrimitiveFail: primIndex
	"If primIndex has an accessorDepth and fails, or it is external and fails with PrimErrNoMemory,
	 call ceCheckAndMaybeRetryPrimitive if so  If ceCheck.... answers true, retry the primitive."
	| jmp |
	<var: #jmp type: #'AbstractInstruction *'>
	(coInterpreter accessorDepthForPrimitiveIndex: primIndex) >= 0
		ifTrue:
			[jmp := cogit
				MoveAw: coInterpreter primFailCodeAddress R: TempReg;
				CmpCq: 0 R: TempReg;
				JumpZero: 0]
		ifFalse:
			[coInterpreter primNumberExternalCall ~= primIndex ifTrue:
				[^0].
			 jmp := cogit
				MoveAw: coInterpreter primFailCodeAddress R: TempReg;
				CmpCq: PrimErrNoMemory R: TempReg;
				JumpNonZero: 0].
	cogit
		compileCallFor: #ceCheckAndMaybeRetryPrimitive:
		numArgs: 1
		arg: (cogit trampolineArgConstant: primIndex)
		arg: nil
		arg: nil
		arg: nil
		resultReg: TempReg
		regsToSave: cogit emptyRegisterMask.
	jmp jmpTarget: cogit Label.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'eem 4/19/2016 14:23'!
maybeMarkCounters: theCounters
	"In SIsta Spur counters are held on the heap in pinned objects which must be marked
	 to avoid them being garbage collected.  This is the hook through which that happens."
	<var: #theCounters type: #usqInt>
	<option: #SistaCogit>
	<inline: true>
	theCounters ~= 0 ifTrue:
		[objectMemory markAndTrace: theCounters - objectMemory baseHeaderSize]! !

!CogObjectRepresentationForSpur methodsFor: 'compilation' stamp: 'eem 6/4/2015 09:33'!
maybeNoteDescriptor: descriptor blockStart: blockStart
	"Override to note inst var refs in blocks.  Used to avoid checking
	 for forwarded receivers in blocks that don't refer to inst vars."
	<var: #blockStart type: #'BlockStart *'>
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<inline: true>
	descriptor isInstVarRef ifTrue:
		[blockStart hasInstVarRef: true]! !

!CogObjectRepresentationForSpur methodsFor: 'method cacheing' stamp: 'eem 8/4/2016 11:49'!
maybeShiftClassTagRegisterForMethodCacheProbe: classTagReg
	"Generate a shift of the register containing the class tag in a method cache probe.
	 c.f. SpurMemoryManager>>methodCacheHashOf:with:"

	cogit LogicalShiftLeftCq: 2 R: classTagReg.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'eem 12/14/2016 13:46'!
mixed: numNonImmediates branchIf: reg instanceOfBehaviors: arrayObj target: targetFixUp
	| jmpImmediate numCases classObj index jmp |
	<var: #jmpImmediate type: #'AbstractInstruction *'>
	<var: #targetFixUp type: #'AbstractInstruction *'>
	<var: #jmp type: #'AbstractInstruction *'>
	numCases := objectMemory numSlotsOf: arrayObj.
	jmpImmediate := self genJumpImmediate: reg.
	
	"Rcvr is non immediate"
	self genGetClassIndexOfNonImm: reg into: TempReg.
	index := 0.
	0 to: numCases - 1 do:
		[:i|
			classObj := objectMemory fetchPointer: i ofObject: arrayObj.
			(objectMemory isImmediateClass: classObj) ifFalse: [
				self genCmpClassIndex: (objectMemory classTagForClass: classObj) R: TempReg.
				cogit JumpZero: targetFixUp.
				index := index + 1 ] ].
	jmp := cogit Jump: 0.
	
	"Rcvr is immediate"
	jmpImmediate jmpTarget: cogit Label.
	numCases - numNonImmediates "num Immediates allowed"
		caseOf:
		{[ 1 ] -> [ "1 immediate needs to jump. Find it and jump."
			0 to: numCases - 1 do:
				[ :j |
				classObj := objectMemory fetchPointer: j ofObject: arrayObj.
				(objectMemory isImmediateClass: classObj) ifTrue: [
					self branchIf: reg hasImmediateTag: (objectMemory classTagForClass: classObj) target: targetFixUp ] ] ] .
		[ 2 ] -> [ "2 immediates needs to jump"
				self branch2CasesIf: reg instanceOfBehaviors: arrayObj target: targetFixUp ] .
		[ 3 ] -> [ "all 3 needs to jump" self Jump: targetFixUp ] }.
	
	jmp jmpTarget: self Label.
	
	^ 0
		! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'eem 6/3/2015 09:13'!
mixed: numNonImmediates branchIf: reg notInstanceOfBehaviors: arrayObj target: targetFixUp
	| jmpImmediate jumps label numCases classObj index |
	<var: #label type: #'AbstractInstruction *'>
	<var: #jmpImmediate type: #'AbstractInstruction *'>
	<var: #jumps type: #'AbstractInstruction **'>
	<var: #targetFixUp type: #'AbstractInstruction *'>
	numCases := objectMemory numSlotsOf: arrayObj.
	jmpImmediate := self genJumpImmediate: reg.
	
	"Rcvr is non immediate"
	jumps := self alloca: numNonImmediates type: (self cCode: [#'AbstractInstruction *'] inSmalltalk: [cogit backEnd class]).
	self genGetClassIndexOfNonImm: reg into: TempReg.
	index := 0.
	0 to: numCases - 1 do:
		[:i|
			classObj := objectMemory fetchPointer: i ofObject: arrayObj.
			(objectMemory isImmediateClass: classObj) ifFalse: [
				self genCmpClassIndex: (objectMemory classTagForClass: classObj) R: TempReg.
				jumps at: index put: (cogit JumpZero: 0).
				index := index + 1 ] ].
	cogit Jump: targetFixUp.
	
	"Rcvr is immediate"
	jmpImmediate jmpTarget: cogit Label.
	numCases - numNonImmediates "num Immediates allowed"
		caseOf:
		{[ 1 ] -> [ "1 immediate allowed. jump to targetFixUp if the rcvr is not this immediate"
			0 to: numCases - 1 do:
				[ :j |
				classObj := objectMemory fetchPointer: j ofObject: arrayObj.
				(objectMemory isImmediateClass: classObj) ifTrue: [
					self branchIf: reg hasNotImmediateTag: (objectMemory classTagForClass: classObj) target: targetFixUp ] ] ] .
		[ 2 ] -> [ "2 immediates allowed. On 32 bits nothing to do, all immediate are allowed, on 64 bits generates the jump to fixup for the third tag"
				self branch2CasesIf: reg notInstanceOfBehaviors: arrayObj target: targetFixUp ] .
		[ 3 ] -> [ "nothing to do, all immediates are allowed." ] }.
	
	label := self Label.
	0 to: numNonImmediates - 1 do: [:i |
		(jumps at: i) jmpTarget: label ].
	
	^ 0
		! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'eem 12/14/2016 13:46'!
noneImmediateBranchIf: reg instanceOfBehaviors: arrayObj target: targetFixUp
	"All classes in arrayObj are not immediate"
	| classObj jmp |
	<var: #targetFixUp type: #'AbstractInstruction *'>
	<var: #jmp type: #'AbstractInstruction *'>
	jmp := self genJumpImmediate: reg.
	self genGetClassIndexOfNonImm: reg into: TempReg.
	0 to: (objectMemory numSlotsOf: arrayObj) - 1 do:
		[:i|
		 classObj := objectMemory fetchPointer: i ofObject: arrayObj.
		 self genCmpClassIndex: (objectMemory classTagForClass: classObj) R: TempReg.
		cogit JumpZero: targetFixUp ].
	jmp jmpTarget: self Label.
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'eem 1/24/2017 13:51'!
noneImmediateBranchIf: reg notInstanceOfBehaviors: arrayObj target: targetFixUp
	"All classes in arrayObj are not immediate"
	| label numJumps jumps classObj |
	<var: #label type: #'AbstractInstruction *'>
	<var: #jumps type: #'AbstractInstruction **'>
	<var: #targetFixUp type: #'AbstractInstruction *'>
	jumps := self alloca: (objectMemory numSlotsOf: arrayObj) type: (self cCode: [#'AbstractInstruction *'] inSmalltalk: [cogit backEnd class]).
	(self genJumpImmediate: reg) jmpTarget: targetFixUp.
	self genGetClassIndexOfNonImm: reg into: TempReg.
	0 to: (numJumps := objectMemory numSlotsOf: arrayObj) - 1 do:
		[:i|
		 classObj := objectMemory fetchPointer: i ofObject: arrayObj.
		 self genCmpClassIndex: (objectMemory classTagForClass: classObj) R: TempReg.
		jumps at: i put: (cogit JumpZero: 0) ].
	cogit Jump: targetFixUp.
	label := cogit Label.
	0 to: numJumps - 1 do: [:i |
		(jumps at: i) jmpTarget: label ].
	^0! !

!CogObjectRepresentationForSpur methodsFor: 'primitive generators' stamp: 'eem 1/11/2016 17:52'!
numCharacterBits
	^30! !

!CogObjectRepresentationForSpur methodsFor: 'calling convention' stamp: 'eem 9/5/2021 07:55'!
numRegArgs
	"Define how many register arguments a StackToRegisterMappingCogit can
	 and should use with the receiver.  The value must be 0, 1 or 2.  Note that a
	 SimpleStackBasedCogit always has 0 register args (although the receiver is
	 passed in a register).  The Spur object representation is simple enough that
	 implementing at:put: is straight-forward and hence 2 register args are worth
	 while.  The method must be inlined in CoInterpreter, and dead code eliminated
	 so that the register-popping enilopmarts such as enterRegisterArgCogMethod:-
	 at:receiver: do not have to be implemented in SimpleStackBasedCogit."
	<api>
	<option: #StackToRegisterMappingCogit>
	<cmacro: '() 2'> "to allow inlining optimization across the CoInterpreter/Cogit api..."
	^2! !

!CogObjectRepresentationForSpur methodsFor: 'garbage collection' stamp: 'eem 2/20/2016 18:11'!
remapObject: objOop
	self assert: (objectMemory addressCouldBeObj: objOop).
	^(objectMemory shouldRemapObj: objOop)
		ifTrue: [objectMemory remapObj: objOop]
		ifFalse: [objOop]! !

!CogObjectRepresentationForSpur methodsFor: 'garbage collection' stamp: 'eem 11/18/2013 18:45'!
remapOop: objOop
	^(objectMemory shouldRemapOop: objOop)
		ifTrue: [objectMemory remapObj: objOop]
		ifFalse: [objOop]! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'eem 4/19/2016 14:24'!
resetCountersIn: cogMethod
	<var: #cogMethod type: #'CogMethod *'>
	<api>
	<option: #SistaCogit>
	cogit fillInCounters: (self numCountersFor: cogMethod counters) atStartAddress: cogMethod counters! !

!CogObjectRepresentationForSpur methodsFor: 'mapped inlined primitive support' stamp: 'cb 11/23/2017 14:27'!
setIfRequiredTargetOf: jumpRemembered toTargetOf: jmpImmediate
	<inline: true> "To avoid typing..."
	CheckRememberedInTrampoline ifFalse:
		[jumpRemembered jmpTarget: jmpImmediate getJmpTarget].! !

!CogObjectRepresentationForSpur methodsFor: 'garbage collection' stamp: 'eem 10/2/2013 22:38'!
shouldAnnotateObjectReference: anOop
	"Objects in newSpace or oldSpace except nil, true, false & classTableRootObj need to be annotated."
	^(objectMemory isNonImmediate: anOop)
	  and: [(self oop: anOop isGreaterThan: objectMemory classTableRootObj)
		or: [self oop: anOop isLessThan: objectMemory nilObject]]! !

!CogObjectRepresentationForSpur methodsFor: 'compile abstract instructions' stamp: 'eem 10/1/2013 17:32'!
slotOffsetOfInstVarIndex: index
	^index * objectMemory wordSize + objectMemory baseHeaderSize! !

!CogObjectRepresentationForSpur methodsFor: 'testing' stamp: 'eem 8/3/2021 18:54'!
smallIntegerIsOnlyImmediateType
	<inline: #always>
	^false! !

!CogObjectRepresentationForSpur methodsFor: 'accessing' stamp: 'eem 8/31/2021 13:38'!
storeTrampolines
	<doNotGenerate>
	^ceStoreTrampolines! !

!CogObjectRepresentationForSpur methodsFor: 'sista support' stamp: 'eem 4/6/2015 09:31'!
valueOfAssociation: associationOop
	| association |
	association := associationOop.
	(objectMemory isForwarded: association) ifTrue:
		[association := objectMemory followForwarded: association].
	^objectMemory fetchPointer: ValueIndex ofObject: association! !

CogObjectRepresentation subclass: #CogObjectRepresentationForSqueakV3
	instanceVariableNames: 'ceCreateNewArrayTrampoline cePositive32BitIntegerTrampoline ceActiveContextTrampoline ceClosureCopyTrampoline ceSigned32BitIntegerTrampoline ceSigned32BitValueOfTrampoline cePositive32BitValueOfTrampoline'
	classVariableNames: 'RootBitDigitLength'
	poolDictionaries: 'VMSqueakClassIndices VMSqueakV3ObjectRepresentationConstants'
	category: 'VMMaker-JIT'!

!CogObjectRepresentationForSqueakV3 commentStamp: '<historical>' prior: 0!
Read my superclass' class comment.  I am a CogObjectRepresentation for the Squeak V3 object representation.!

!CogObjectRepresentationForSqueakV3 class methodsFor: 'accessing class hierarchy' stamp: 'eem 1/22/2014 15:53'!
defaultObjectMemoryClass
	"For in-image compilation."
	^NewCoObjectMemory! !

!CogObjectRepresentationForSqueakV3 class methodsFor: 'class initialization' stamp: 'eem 9/27/2009 10:21'!
initialize
	RootBit ifNil: [ObjectMemory initializeObjectHeaderConstants].
	RootBitDigitLength := RootBit digitLength! !

!CogObjectRepresentationForSqueakV3 class methodsFor: 'trampoline support' stamp: 'rsf 8/9/2016 21:30'!
numTrampolines
	^super numTrampolines + 4 + (LowcodeVM ifTrue: [ 3 ] ifFalse: [ 0 ])! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 3/29/2009 14:12'!
allYoungObjectsAgeInFullGC
	^true! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'sista support' stamp: 'eem 7/5/2014 17:56'!
allocateCounters: nCounters
	"Malloc the counter space.  Fill the first slot with the number of counters for resetCountersIn:"
	| theCounters |
	<var: #theCounters type: #usqInt>
	theCounters := self malloc: nCounters + 1 * (self sizeof: #sqInt).
	theCounters = 0 ifTrue:
		[^theCounters].
	self longAt: theCounters put: nCounters.
	^theCounters + (self sizeof: #sqInt)! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 10/9/2009 03:08'!
assertIsValidOop: anOop
	<doNotGenerate>
	"Method map assertion chekcing only (development time only).
	 Note we never put SmallIntegers in the map."
	^objectMemory addressCouldBeObj: anOop! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 10/9/2009 03:10'!
cacheTagIsMarked: cacheTag
	"Answer if the cacheTag is not unmarked, i.e. answer true for compact
	 class indices and immediates; only answer false for unmarked objects"
	(self couldBeObject: cacheTag) ifFalse: [^true].
	self assert: (objectMemory addressCouldBeObj: cacheTag).
	^objectMemory isMarked: cacheTag! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 5/9/2014 19:51'!
checkValidOopReference: anOop
	^(objectMemory isIntegerObject: anOop)
	   or: [(objectMemory heapMapAtWord: (self pointerForOop: anOop)) ~= 0]! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'in-line cacheing' stamp: 'eem 7/7/2014 08:43'!
classForInlineCacheTag: inlineCacheTag
	(objectMemory isIntegerObject: inlineCacheTag) ifTrue:
		[^objectMemory classSmallInteger].
	(self couldBeObject: inlineCacheTag) ifTrue:
		[^inlineCacheTag].
	^objectMemory compactClassAt: inlineCacheTag >> objectMemory compactClassFieldLSB! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'object representation' stamp: 'eem 8/19/2011 10:34'!
compactClassFieldMask
	"This is the mask for the field when shifted into the least significant bits"
	<inline: true>
	^(1 << objectMemory compactClassFieldWidth) - 1! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 11/18/2013 15:41'!
couldBeObject: oop
	"Note this version filters-out compact class indices via the >= nilObj clause"
	^(objectMemory isNonIntegerObject: oop)
	  and: [self oop: oop isGreaterThanOrEqualTo: objectMemory nilObject]! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'sista support' stamp: 'eem 7/5/2014 19:04'!
freeCounters: theCounters
	<var: #theCounters type: #usqInt>
	theCounters ~= 0 ifTrue:
		[self free: theCounters - (self sizeof: #sqInt)]! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'initialization' stamp: 'eem 03/03/2016 12:23'!
genActiveContextTrampoline
	"Short-circuit the interpreter call if a frame is already married."
	| jumpSingle |
	<var: #jumpSingle type: #'AbstractInstruction *'>
	cogit
		zeroOpcodeIndex;
		MoveMw: FoxMethod r: FPReg R: TempReg;
		AndCq: MFMethodFlagHasContextFlag R: TempReg.
	jumpSingle := cogit JumpZero: 0.
	cogit
		MoveMw: FoxThisContext r: FPReg R: ReceiverResultReg;
		RetN: 0.
	jumpSingle jmpTarget: cogit Label.
	^cogit genTrampolineFor: #ceActiveContext
		called: 'ceActiveContextTrampoline'
		numArgs: 0
		arg: nil
		arg: nil
		arg: nil
		arg: nil
		regsToSave: cogit emptyRegisterMask
		pushLinkReg: true
		resultReg: ReceiverResultReg
		appendOpcodes: true! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 2/11/2009 13:13'!
genAddSmallIntegerTagsTo: aRegister
	cogit AddCq: 1 R: aRegister.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 1/9/2020 06:58'!
genAllocFloatValue: dpreg into: resultReg scratchReg: scratch1 scratchReg: scratch2
	<returnTypeC: #'AbstractInstruction *'>
	| allocSize newFloatHeaderSansHash jumpFail |
	<var: #jumpFail type: #'AbstractInstruction *'>
	allocSize := objectMemory baseHeaderSize + (objectMemory sizeof: #double).
	newFloatHeaderSansHash := ((ClassFloatCompactIndex << objectMemory compactClassFieldLSB
									bitOr: (objectMemory formatOfClass: objectMemory classFloat))
									bitOr: allocSize)
									 bitOr: HeaderTypeShort.
	cogit MoveAw: objectMemory freeStartAddress R: resultReg.
	cogit LoadEffectiveAddressMw: allocSize r: resultReg R: scratch1.
	cogit MoveAw: objectMemory scavengeThresholdAddress R: scratch2.
	cogit CmpR: scratch2 R: scratch1.
	jumpFail := cogit JumpAboveOrEqual: 0.
	cogit MoveR: resultReg R: scratch2.
	self flag: #newObjectHash.
	cogit AndCq: HashMaskUnshifted << objectMemory wordSize R: scratch2.
	cogit LogicalShiftLeftCq: HashBitsOffset - objectMemory wordSize R: scratch2.
	cogit OrCq: newFloatHeaderSansHash R: scratch2.
	cogit MoveR: scratch2 Mw: 0 r: resultReg.
	cogit MoveRd: dpreg M64: objectMemory baseHeaderSize r: resultReg.
	cogit MoveR: scratch1 Aw: objectMemory freeStartAddress.
	^jumpFail! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 12/6/2015 20:41'!
genClearAndSetSmallIntegerTagsIn: scratchReg
	"Set the SmallInteger tag bits when the tag bits may be filled with garbage."
	^self genSetSmallIntegerTagsIn: scratchReg! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 12/26/2019 21:49'!
genConvertIntegerInReg: srcReg toSmallIntegerInReg: destReg
	cogit LogicalShiftLeftCq: 1 R: srcReg R: destReg.
	cogit AddCq: 1 R: destReg.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 12/6/2013 14:50'!
genConvertIntegerToSmallIntegerInReg: reg
	cogit LogicalShiftLeftCq: 1 R: reg.
	cogit AddCq: 1 R: reg.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 12/6/2013 14:51'!
genConvertSmallIntegerToIntegerInReg: reg
	cogit ArithmeticShiftRightCq: 1 R: reg.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'bytecode generator support' stamp: 'eem (const to msg rewrite) 11/6/2014 10:37'!
genCreateClosureAt: bcpc numArgs: numArgs numCopied: numCopied contextNumArgs: ctxtNumArgs large: isLargeCtxt inBlock: isInBlock
	"Create a closure with the given startpc, numArgs and numCopied
	 within a context with ctxtNumArgs, large if isLargeCtxt that is in a
	 block if isInBlock.  If numCopied > 0 pop those values off the stack."
	
	"see ceClosureCopyDescriptor:"
	cogit MoveCq: numArgs + (numCopied << 6) + (bcpc << 12) R: SendNumArgsReg.
	cogit CallRT: ceClosureCopyTrampoline.
	numCopied > 0 ifTrue:
		[cogit AddCq: numCopied * objectMemory wordSize R: SPReg].
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'cb 4/17/2015 15:24'!
genFetchIndexRegister: indexReg from: tableObj into: destReg
	"indexReg contains the 1-relative index of an element in tableObj.
	 Since BaseHeaderSize = BytesPerOop we can use it as a
	 zero-relative index from the beginning of the object."
	self assert: indexReg ~= destReg.
	cogit genMoveConstant: tableObj R: destReg.
	cogit MoveXwr: indexReg R: destReg R: destReg.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'bytecode generator support' stamp: 'eem 5/31/2014 21:47'!
genGetActiveContextNumArgs: numArgs large: isLargeContext inBlock: isInBlock
	"Get the active context into ReceiverResultReg, creating it if necessary."
	cogit CallRT: ceActiveContextTrampoline! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem (const to msg rewrite) 11/13/2014 14:04'!
genGetClassFormatOfNonInt: instReg into: destReg scratchReg: scratchReg
	"Fetch the instance's class format into destReg, assuming the object is non-int."
	| jumpCompact jumpGotClass |
	<var: #jumpCompact type: #'AbstractInstruction *'>
	<var: #jumpGotClass type: #'AbstractInstruction *'>
	"Get header word in destReg"
	cogit MoveMw: 0 r: instReg R: destReg.
	"Form the byte index of the compact class field"
	cogit LogicalShiftRightCq: (objectMemory compactClassFieldLSB - objectMemory shiftForWord) R: destReg.
	cogit AndCq: self compactClassFieldMask << objectMemory shiftForWord R: destReg.
	jumpCompact := cogit JumpNonZero: 0.
	cogit MoveMw: objectMemory classFieldOffset r: instReg R: scratchReg.
	"The use of signedIntFromLong is a hack to get round short addressing mode computations.
	 Much easier if offsets are signed and the arithmetic machinery we have makes it difficult to
	 mix signed and unsigned offsets."
	cogit AndCq: AllButTypeMask signedIntFromLong R: scratchReg.
	jumpGotClass := cogit Jump: 0.
	"Don't have to subtract one from the destReg compactClassArray index because of the header word."
	self assert: objectMemory baseHeaderSize = objectMemory wordSize.
	jumpCompact jmpTarget:
		(cogit annotate: (cogit MoveMw: (objectMemory splObj: CompactClasses) r: destReg R: scratchReg)
			objRef: (objectMemory splObj: CompactClasses)).
	jumpGotClass jmpTarget:
		(cogit MoveMw: InstanceSpecificationIndex << objectMemory shiftForWord + objectMemory wordSize r: scratchReg R: destReg).
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'cb 9/11/2019 07:47'!
genGetClassObjectOf: instReg into: destReg scratchReg: scratchReg mayBeAForwarder: mayBeAForwarder
	"Fetch the instance's class into destReg.  This is almost identical
	 to genGetClassFormatOfNonInt:into:scratchReg: but because we
	 put the fetch of SmallInteger between the then and the else for 
	 compact class/non-compact class we cannot easily share code.
	 instRegIsReceiver is ignored.  It is for Spur compatibility where
	 objects may be forwarded."
	| jumpIsInt jumpCompact jumpGotClass jumpGotClass2 |
	<var: #jumpIsInt type: #'AbstractInstruction *'>
	<var: #jumpCompact type: #'AbstractInstruction *'>
	<var: #jumpGotClass type: #'AbstractInstruction *'>
	<var: #jumpGotClass2 type: #'AbstractInstruction *'>
	cogit MoveR: instReg R: scratchReg.
	cogit AndCq: 1 R: scratchReg.
	jumpIsInt := cogit JumpNonZero: 0.
	"Get header word in scratchReg"
	cogit MoveMw: 0 r: instReg R: scratchReg.
	"Form the byte index of the compact class field"
	cogit LogicalShiftRightCq: (objectMemory compactClassFieldLSB - objectMemory shiftForWord) R: scratchReg.
	cogit AndCq: self compactClassFieldMask << objectMemory shiftForWord R: scratchReg.
	jumpCompact := cogit JumpNonZero: 0.
	cogit MoveMw: objectMemory classFieldOffset r: instReg R: destReg.
	cogit AndCq: AllButTypeMask signedIntFromLong R: destReg.
	jumpGotClass := cogit Jump: 0.
	jumpIsInt jmpTarget: (cogit genMoveConstant: objectMemory classSmallInteger R: destReg).
	jumpGotClass2 := cogit Jump: 0.
	"Don't have to subtract one from the destReg compactClassArray index because of the header word."
	self assert: objectMemory baseHeaderSize = objectMemory wordSize.
	jumpCompact jmpTarget:
		(cogit annotate: (cogit MoveMw: (objectMemory splObj: CompactClasses) r: scratchReg R: destReg)
			objRef: (objectMemory splObj: CompactClasses)).
	jumpGotClass jmpTarget:
	(jumpGotClass2 jmpTarget: cogit Label).
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 7/31/2014 06:01'!
genGetClassObjectOfNonCompact: instReg into: destReg
	cogit
		MoveMw: objectMemory classFieldOffset r: instReg R: destReg;
		AndCq: AllButTypeMask signedIntFromLong R: destReg.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'cb 9/11/2019 07:49'!
genGetClassTagOf: instReg into: destReg scratchReg: scratchReg
	"Compatibility with SpurObjectRepresentation/SpurMemoryManager."
	| entryLabel |
	<var: #entryLabel type: #'AbstractInstruction *'>
	cogit AlignmentNops: (objectMemory wordSize max: 8).
	entryLabel := cogit Label.
	(self genGetClassObjectOf: instReg into: destReg scratchReg: scratchReg mayBeAForwarder: nil) ~= 0 ifTrue:
		[self error: 'internal error'].
	^entryLabel! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 10/3/2013 10:12'!
genGetCompactClassIndexNonImmOf: instReg into: destReg
	"Fetch the instance's compact class index into destReg."
	"Get header word in scratchReg"
	cogit MoveMw: 0 r: instReg R: destReg.
	"Form the byte index of the compact class field"
	cogit LogicalShiftRightCq: objectMemory compactClassFieldLSB R: destReg.
	cogit AndCq: self compactClassFieldMask R: destReg.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem (const to msg rewrite) 11/6/2014 10:37'!
genGetDoubleValueOf: srcReg into: destFPReg 
	cogit MoveM64: objectMemory baseHeaderSize r: srcReg Rd: destFPReg.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 2/3/2009 17:08'!
genGetFixedFieldsOfPointerNonInt: instReg into: destReg scratchReg: scratchReg
	"Fetch the instance's class format into destReg, assuming the object is pointers and non-int"
	self genGetClassFormatOfNonInt: instReg into: destReg scratchReg: scratchReg.
	"compute  (classFormat >> 11 bitAnd: 16rC0) + (classFormat >> 2 bitAnd: 16r3F) - 1"
	cogit
		MoveR: destReg R: scratchReg;
		LogicalShiftRightCq: 2 R: destReg;
		LogicalShiftRightCq: 11 R: scratchReg;
		AndCq: 16r3F R: destReg;
		AndCq: 16rC0 R: scratchReg;
		SubCq: 1 R: destReg;
		AddR: scratchReg R: destReg.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'primitive generators' stamp: 'eem 5/30/2015 09:41'!
genGetFormatOf: srcReg into: destReg
	"Get the format field of the object in srcReg into destReg.
	 srcReg may equal destReg."
	<inline: true>
	cogit
		MoveMw: 0 r: srcReg R: destReg;
		LogicalShiftRightCq: objectMemory instFormatFieldLSB R: destReg;
		AndCq: self instFormatFieldMask R: destReg.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 10/2/2013 17:03'!
genGetHashFieldNonImmOf: instReg asSmallIntegerInto: destReg
	"Fetch the instance's identity hash into destReg, encoded as a SmallInteger."
	"Get header word in scratchReg"
	cogit MoveMw: 0 r: instReg R: destReg.
	"Shift and mask the field leaving room for the SmallInteger tag."
	cogit LogicalShiftRightCq: HashBitsOffset - 1 R: destReg.
	cogit AndCq: HashMaskUnshifted << 1 R: destReg.
	"Set the SmallInteger tag."
	cogit AddCq: 1 R: destReg.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem (const to msg rewrite) 11/6/2014 10:37'!
genGetInlineCacheClassTagFrom: sourceReg into: destReg forEntry: forEntry
	"Extract the inline cache tag for the object in sourceReg into destReg. The inline cache tag
	 for a given object is the value loaded in inline caches to distinguish objects of different
	 classes.  In Spur this is either the tags for immediates, or the receiver's classIndex.
	 The inline cache tag for a given object is the value loaded in inline caches to distinguish
	 objects of different classes.  In SqueakV3 the tag is the integer tag bit for SmallIntegers (1),
	 the compact class index shifted by log: 2 word size for objects with compact classes
	 (1 * 4 to: 31 * 4 by: 4), or the class.  These ranges cannot overlap because the heap
	 (and hence the lowest class object) is beyond the machine code zone.
	 If forEntry is true answer the entry label at which control is to enter (cmEntryOffset).
	 If forEntry is false, control enters at the start."
	| entryLabel jumpIsInt jumpCompact |
	<var: #entryLabel type: #'AbstractInstruction *'>
	<var: #jumpIsInt type: #'AbstractInstruction *'>
	<var: #jumpCompact type: #'AbstractInstruction *'>
	cogit AlignmentNops: (objectMemory wordSize max: 8).
	entryLabel := cogit Label.
	cogit MoveR: sourceReg R: destReg.
	cogit AndCq: 1 R: destReg.
	jumpIsInt := cogit JumpNonZero: 0.
	"Get header word in destReg"
	cogit MoveMw: 0 r: sourceReg R: destReg.
	"Extract the compact class field, and if non-zero use it as the tag.."
	self assert: self compactClassFieldMask << objectMemory compactClassFieldLSB < objectMemory nilObject asUnsignedInteger.
	cogit AndCq: self compactClassFieldMask << objectMemory compactClassFieldLSB R: destReg.
	jumpCompact := cogit JumpNonZero: 0.
	cogit MoveMw: objectMemory classFieldOffset r: sourceReg R: destReg.
	"The use of signedIntFromLong is a hack to get round short addressing mode computations.
	 Much easier if offsets are signed and the arithmetic machinery we have makes it difficult to
	 mix signed and unsigned offsets."
	cogit AndCq: AllButTypeMask signedIntFromLong R: destReg.
	jumpCompact jmpTarget: (jumpIsInt jmpTarget: cogit Label).
	^entryLabel! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'RonieSalgado 5/1/2016 01:01'!
genGetNumSlotsOf: srcRegister into: destRegister! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'primitive generators' stamp: 'eem (const to msg rewrite) 11/13/2014 14:04'!
genGetSizeOf: sourceReg into: destReg formatReg: formatReg scratchReg: scratchReg abortJumpsInto: aBinaryBlock
	"Get the size of the non-immediate object in sourceReg into destReg using formatReg
	 and scratchReg as temps.  None of these registers can overlap.  Supply the jumps
	 taken if the object in sourceReg is not indexable, or if the object in sourceReg is a
	 context.. Hack: If the object has a pointer format other than 2 leave the number of
	 fixed fields in formatReg.  Used by primitiveSize, primitiveAt, and primitiveAtPut"
	<returnTypeC: #'AbstractInstruction *'>
	| jumpNotIndexable jumpIsContext jumpShortHeader jumpSkip
	  jumpFmtLeWeakArray jumpFmtIsArray jmpFmtGeFirstByte jumpGotByteSize jumpGotWordSize |
	<inline: true>
	"c.f. StackInterpreter>>stSizeOf: lengthOf:baseHeader:format: fixedFieldsOf:format:length:"
	"and one wonders why Squeak V1 through V3 are slow..."
	<var: #jumpNotIndexable type: #'AbstractInstruction *'>
	<var: #jumpIsContext type: #'AbstractInstruction *'>
	<var: #jumpShortHeader type: #'AbstractInstruction *'>
	<var: #jumpSkip type: #'AbstractInstruction *'>
	<var: #jumpFmtLeWeakArray type: #'AbstractInstruction *'>
	<var: #jumpFmtIsArray type: #'AbstractInstruction *'>
	<var: #jmpFmtGeFirstByte type: #'AbstractInstruction *'>
	<var: #jumpGotByteSize type: #'AbstractInstruction *'>
	<var: #jumpGotWordSize type: #'AbstractInstruction *'>
	cogit
		MoveMw: 0 r: sourceReg R: formatReg;				"destReg := self baseHeader: receiver"
		MoveR: formatReg R: destReg;
		LogicalShiftRightCq: objectMemory instFormatFieldLSB R: formatReg;
		AndCq: self instFormatFieldMask R: formatReg;		"formatReg := self formatOfHeader: destReg"
		CmpCq: objectMemory arrayFormat R: formatReg.
	jumpNotIndexable := cogit JumpLess: 0.
	cogit
		MoveR: destReg R: scratchReg;
		LogicalShiftRightCq: objectMemory compactClassFieldLSB R: scratchReg;
		AndCq: self compactClassFieldMask R: scratchReg;	"scratchReg := self compactClassIndexOfHeader: destReg"
		CmpCq: ClassMethodContextCompactIndex R: scratchReg.
	jumpIsContext := cogit JumpZero: 0.
	cogit
		MoveR: destReg R: scratchReg;
		AndCq: TypeMask R: scratchReg;
		CmpCq: HeaderTypeSizeAndClass R: scratchReg.	"(hdr bitAnd: TypeMask) = HeaderTypeSizeAndClass"
	jumpShortHeader := cogit JumpNonZero: 0.
	self assert: Size4Bit = 0.
	cogit
		MoveMw: 0 - (2 * objectMemory wordSize) r: sourceReg R: destReg; "(self sizeHeader: oop) bitAnd: LongSizeMask"
		AndCq: LongSizeMask signedIntFromLong R: destReg.
	jumpSkip :=  cogit Jump: 0.
	jumpShortHeader jmpTarget: (cogit AndCq: SizeMask R: destReg).	"hdr bitAnd: SizeMask"
	jumpSkip jmpTarget: (cogit SubCq: objectMemory baseHeaderSize R: destReg). "sz - BaseheaderSize for all three arms"
	cogit CmpCq: objectMemory weakArrayFormat R: formatReg.	"fmt <= 4"
	jumpFmtLeWeakArray := cogit JumpLessOrEqual: 0.
	cogit CmpCq: objectMemory firstByteFormat R: formatReg.
	jmpFmtGeFirstByte := cogit JumpLess: 0.
	cogit
		AndCq: 3 R: formatReg;	"(sz - BaseHeaderSize) - (fmt bitAnd: 3) bytes"
		SubR: formatReg R: destReg.
	jumpGotByteSize := cogit Jump: 0.
	jmpFmtGeFirstByte jmpTarget: (cogit LogicalShiftRightCq: 2 R: destReg). "(sz - BaseHeaderSize) >> 2 32-bit longs"
	jumpGotWordSize := cogit Jump: 0.
	jumpFmtLeWeakArray jmpTarget: cogit Label.
	cogit
		LogicalShiftRightCq: objectMemory shiftForWord R: destReg; "(sz - BaseHeaderSize) >> ShiftForWord words"
		CmpCq: objectMemory arrayFormat R: formatReg.
	jumpFmtIsArray := cogit JumpZero: 0.
	self genGetFixedFieldsOfPointerNonInt: sourceReg into: formatReg scratchReg: scratchReg.
	cogit SubR: formatReg R: destReg.
	jumpFmtIsArray jmpTarget:
	(jumpGotWordSize jmpTarget:
	(jumpGotByteSize jmpTarget:
		cogit Label)).
	aBinaryBlock value: jumpNotIndexable value: jumpIsContext! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 12/7/2016 15:18'!
genInlinedIdenticalOrNotIfGuts: orNot
	<inline: true>
	^ cogit genVanillaInlinedIdenticalOrNotIf: orNot! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 5/29/2015 18:14'!
genJumpImmediate: aRegister
	^self genJumpSmallInteger: aRegister! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 6/1/2016 16:09'!
genJumpInOldSpace: reg
	"Jump if reg is old."
	<inline: true>
	^cogit
		MoveAw: objectMemory youngStartAddress R: TempReg;
		CmpR: TempReg R: reg;	"N.B. FLAGS := destReg - scratchReg"
		JumpBelow: 0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 12/26/2019 16:52'!
genJumpIsSmallIntegerValue: aRegister scratch: scratchReg
	"Generate a test for aRegister containing an integer value in the SmallInteger range, and a jump if so, answering the jump.
	 c.f. ObjectMemory>>isIntegerValue:"
	<returnTypeC: #'AbstractInstruction *'>
	^cogit
		LogicalShiftLeftCq: 1 R: aRegister R: scratchReg;
		XorR: aRegister R: scratchReg;
		JumpGreaterOrEqual: 0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 5/30/2015 11:22'!
genJumpNotImmediate: aRegister
	<inline: true>
	^self genJumpNotSmallInteger: aRegister! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 5/30/2015 11:07'!
genJumpNotSmallInteger: aRegister
	<returnTypeC: #'AbstractInstruction *'>
	cogit TstCq: 1 R: aRegister.
	^cogit JumpZero: 0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 2/15/2009 10:20'!
genJumpNotSmallIntegerInScratchReg: aRegister
	<returnTypeC: #'AbstractInstruction *'>
	cogit AndCq: 1 R: aRegister.
	^cogit JumpZero: 0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 12/26/2019 16:54'!
genJumpNotSmallIntegerValue: aRegister scratch: scratchReg
	"Generate a test for aRegister containing an integer value outside the SmallInteger range, and a jump if so, answering the jump.
	 c.f. ObjectMemory>>isIntegerValue:"
	<returnTypeC: #'AbstractInstruction *'>
	^cogit
		ArithmeticShiftRightCq: 1 R: aRegister R: scratchReg;
		XorR: aRegister R: scratchReg;
		JumpLess: 0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 5/29/2015 18:14'!
genJumpSmallInteger: aRegister
	<returnTypeC: #'AbstractInstruction *'>
	cogit TstCq: 1 R: aRegister.
	^cogit JumpNonZero: 0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 2/15/2009 10:19'!
genJumpSmallIntegerInScratchReg: aRegister
	<returnTypeC: #'AbstractInstruction *'>
	cogit AndCq: 1 R: aRegister.
	^cogit JumpNonZero: 0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:27'!
genLcFirstFieldPointer: objectReg
	<option: #LowcodeVM>
	cogit AddCq: 4 R: objectReg.
	cogit ssPushNativeRegister: objectReg.! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:27'!
genLcFirstIndexableFieldPointer: objectReg
	<option: #LowcodeVM>
	"TODO: Fixme for fixed fields"
	cogit AddCq: 4 R: objectReg.
	cogit ssPushNativeRegister: objectReg.! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:27'!
genLcInt32ToOop: value
	<option: #LowcodeVM>
	value ~= ReceiverResultReg ifTrue: [ cogit MoveR: value R: ReceiverResultReg.].
	cogit CallRT: ceSigned32BitIntegerTrampoline.

	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit ssPushRegister: ReceiverResultReg.! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:27'!
genLcIsBytes: objectReg to: valueReg
	<option: #LowcodeVM>
	| isInteger isNotBytes falseTarget cont |
	
	"Check for integer"
	cogit MoveR: objectReg R: TempReg.
	cogit AndCq: 1 R: TempReg.
	isInteger := cogit JumpNonZero: 0.

	"Get the class format"
	self genGetClassFormatOfNonInt: objectReg into: valueReg scratchReg: TempReg.

	"Hack: consider a compiled method as bytes too."
	cogit CmpCq: 8 R: valueReg.
	isNotBytes := cogit JumpLess: 0.

	"True result"
	cogit MoveCq: 1 R: valueReg.
	cont := cogit Jump: 0.

	"False result"
	falseTarget := cogit Label.
	isInteger jmpTarget: falseTarget.
	isNotBytes jmpTarget: falseTarget.
	cogit MoveCq: 0 R: valueReg.

	"Return"
	cont jmpTarget: cogit Label.
	cogit ssPushNativeRegister: valueReg.! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:27'!
genLcIsFloatObject: objectReg to: valueReg
	<option: #LowcodeVM>
	"TODO: Implement this one"
	cogit MoveCq: 1 R: valueReg.! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcIsIndexable: objectReg to: valueReg
	<option: #LowcodeVM>
	| isInteger isNotIndexable falseTarget cont |
	
	"Check for integer"
	cogit MoveR: objectReg R: TempReg.
	cogit AndCq: 1 R: TempReg.
	isInteger := cogit JumpNonZero: 0.

	"Get the class format"
	self genGetClassFormatOfNonInt: objectReg into: valueReg scratchReg: TempReg.

	"Hack: consider a compiled method as bytes too."
	cogit CmpCq: 2 R: valueReg.
	isNotIndexable := cogit JumpLess: 0.

	"True result"
	cogit MoveCq: 1 R: valueReg.
	cont := cogit Jump: 0.

	"False result"
	falseTarget := cogit Label.
	isInteger jmpTarget: falseTarget.
	isNotIndexable jmpTarget: falseTarget.
	cogit MoveCq: 0 R: valueReg.

	"Return"
	cont jmpTarget: cogit Label.
	cogit ssPushNativeRegister: valueReg.! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcIsIntegerObject: objectReg to: valueReg
	<option: #LowcodeVM>
	"TODO: Support larger integers"
	cogit AndCq: 1 R: objectReg.
	cogit ssPushNativeRegister: objectReg.! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcIsPointers: objectReg to: valueReg
	<option: #LowcodeVM>
	| isInteger isNotIndexable falseTarget cont |
	
	"Check for integer"
	cogit MoveR: objectReg R: TempReg.
	cogit AndCq: 1 R: TempReg.
	isInteger := cogit JumpNonZero: 0.

	"Get the class format"
	self genGetClassFormatOfNonInt: objectReg into: valueReg scratchReg: TempReg.

	"Hack: consider a compiled method as bytes too."
	cogit CmpCq: 3 R: valueReg.
	isNotIndexable := cogit JumpGreater: 0.

	"True result"
	cogit MoveCq: 1 R: valueReg.
	cont := cogit Jump: 0.

	"False result"
	falseTarget := cogit Label.
	isInteger jmpTarget: falseTarget.
	isNotIndexable jmpTarget: falseTarget.
	cogit MoveCq: 0 R: valueReg.

	"Return"
	cont jmpTarget: cogit Label.
	cogit ssPushNativeRegister: valueReg.! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcIsWords: objectReg to: valueReg
	<option: #LowcodeVM>
	| isInteger isNotWords falseTarget cont |
	
	"Check for integer"
	cogit MoveR: objectReg R: TempReg.
	cogit AndCq: 1 R: TempReg.
	isInteger := cogit JumpNonZero: 0.

	"Get the class format"
	self genGetClassFormatOfNonInt: objectReg into: valueReg scratchReg: TempReg.

	"Check for the words format"
	cogit CmpCq: 6 R: valueReg.
	isNotWords := cogit JumpNonZero: 0.

	"True result"
	cogit MoveCq: 1 R: valueReg.
	cont := cogit Jump: 0.

	"False result"
	falseTarget := cogit Label.
	isInteger jmpTarget: falseTarget.
	isNotWords jmpTarget: falseTarget.
	cogit MoveCq: 0 R: valueReg.

	"Return"
	cont jmpTarget: cogit Label.
	cogit ssPushNativeRegister: valueReg.! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcIsWordsOrBytes: objectReg to: valueReg
	<option: #LowcodeVM>
	| isInteger isNotBytes falseTarget cont |
	
	"Check for integer"
	cogit MoveR: objectReg R: TempReg.
	cogit AndCq: 1 R: TempReg.
	isInteger := cogit JumpNonZero: 0.

	"Get the class format"
	self genGetClassFormatOfNonInt: objectReg into: valueReg scratchReg: TempReg.

	"Hack: consider a compiled method as bytes too."
	cogit CmpCq: 6 R: valueReg.
	isNotBytes := cogit JumpLess: 0.

	"True result"
	cogit MoveCq: 1 R: valueReg.
	cont := cogit Jump: 0.

	"False result"
	falseTarget := cogit Label.
	isInteger jmpTarget: falseTarget.
	isNotBytes jmpTarget: falseTarget.
	cogit MoveCq: 0 R: valueReg.

	"Return"
	cont jmpTarget: cogit Label.
	cogit ssPushNativeRegister: valueReg.! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcLoadObject: object at: fieldIndex
	<option: #LowcodeVM>
	cogit AddCq: 1 R: fieldIndex.
	cogit MoveXwr: fieldIndex R: object R: object.
	cogit ssPushRegister: object! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcLoadObject: object field: fieldIndex
	<option: #LowcodeVM>
	cogit MoveMw: (fieldIndex+ 1) * BytesPerWord r: object R: object.
	cogit ssPushRegister: object! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcOopToInt32: value
	<option: #LowcodeVM>
	value ~= ReceiverResultReg ifTrue: [ cogit MoveR: value R: ReceiverResultReg.].
	cogit CallRT: ceSigned32BitValueOfTrampoline.

	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit ssPushNativeRegister: ReceiverResultReg.! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcOopToUInt32: value
	<option: #LowcodeVM>
	value ~= ReceiverResultReg ifTrue: [ cogit MoveR: value R: ReceiverResultReg.].
	cogit CallRT: cePositive32BitValueOfTrampoline.

	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit ssPushNativeRegister: ReceiverResultReg.! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcStore: value object: object at: fieldIndex
	<option: #LowcodeVM>
	cogit AddCq: 1 R: fieldIndex.
	cogit MoveR: value Xwr: fieldIndex R: object! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcStore: value object: object field: fieldIndex
	<option: #LowcodeVM>
	cogit MoveR: value Mw: (fieldIndex+ 1) * BytesPerWord r: object! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'inline primitive support' stamp: 'rsf 8/10/2016 00:28'!
genLcUInt32ToOop: value
	<option: #LowcodeVM>
	value ~= ReceiverResultReg ifTrue: [ cogit MoveR: value R: ReceiverResultReg.].
	cogit CallRT: cePositive32BitIntegerTrampoline.

	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit ssPushRegister: ReceiverResultReg.! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'bytecode generator support' stamp: 'eem 7/6/2021 17:07'!
genNewArrayOfSize: size initialized: initialize
	"Generate a call to code that allocates a new Array of size.
	 The Array should be initialized with nils iff initialized is true.
	 The size arg is passed in SendNumArgsReg, the result
	 must come back in ReceiverResultReg."
	cogit
		MoveCq: size R: SendNumArgsReg;
		CallRT: ceCreateNewArrayTrampoline! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 09:56'!
genPrimitiveAt
	| jumpSI jumpNotSI jumpNotIndexable jumpIsContext jumpBounds jumpFmtGt4 jumpFmtEq2 jumpFmtLt8 jumpFmtGt11 jumpLarge |
	"c.f. StackInterpreter>>stSizeOf: lengthOf:baseHeader:format: fixedFieldsOf:format:length:"
	<var: #jumpSI type: #'AbstractInstruction *'>
	<var: #jumpNotSI type: #'AbstractInstruction *'>
	<var: #jumpNotIndexable type: #'AbstractInstruction *'>
	<var: #jumpIsContext type: #'AbstractInstruction *'>
	<var: #jumpBounds type: #'AbstractInstruction *'>
	<var: #jumpFmtGt4 type: #'AbstractInstruction *'>
	<var: #jumpFmtEq2 type: #'AbstractInstruction *'>
	<var: #jumpFmtLt8 type: #'AbstractInstruction *'>
	<var: #jumpFmtGt11 type: #'AbstractInstruction *'>
	<var: #jumpLarge type: #'AbstractInstruction *'>
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	jumpSI := self genJumpSmallInteger: ReceiverResultReg.
	cogit MoveR: Arg0Reg R: Arg1Reg.
	jumpNotSI := self genJumpNotSmallInteger: Arg0Reg.
	self
		genGetSizeOf: ReceiverResultReg
		into: ClassReg
		formatReg: SendNumArgsReg
		scratchReg: TempReg
		abortJumpsInto: [:jnx :jic| jumpNotIndexable := jnx. jumpIsContext := jic].
	self genConvertSmallIntegerToIntegerInReg: Arg1Reg.
	cogit SubCq: 1 R: Arg1Reg.
	cogit CmpR: ClassReg R: Arg1Reg.
	jumpBounds := cogit JumpAboveOrEqual: 0.
	"This is tedious.  Because of register pressure on x86 (and the baroque
	 complexity of the size computation) we have to recompute the format
	 because it may have been smashed computing the fixed fields.  But at
	 least we have the fixed fields, if any, in formatReg and recomputing
	 these is more expensive than recomputing format.  In any case this
	 should still be faster than the interpreter and we hope this object
	 representation's days are numbered."
	cogit
		MoveMw: 0 r: ReceiverResultReg R: ClassReg;	"self baseHeader: receiver"
		LogicalShiftRightCq: objectMemory instFormatFieldLSB R: ClassReg;
		AndCq: self instFormatFieldMask R: ClassReg;	"self formatOfHeader: ClassReg"
		CmpCq: 4 R: ClassReg.
	jumpFmtGt4 := cogit JumpGreater: 0.
	cogit CmpCq: 2 R: ClassReg.	"Common case, e.g. Array, has format = 2"
	jumpFmtEq2 := cogit JumpZero: 0.
	cogit AddR: SendNumArgsReg R: Arg1Reg. "Add fixed fields to index"
	jumpFmtEq2 jmpTarget: cogit Label.
	cogit "Too lazy [knackered, more like. ed.] to define index with displacement addressing right now"
		AddCq: objectMemory baseHeaderSize / objectMemory wordSize R: Arg1Reg;
		MoveXwr: Arg1Reg R: ReceiverResultReg R: ReceiverResultReg;
		genPrimReturn.
	jumpFmtGt4 jmpTarget: cogit Label.
	"Byte objects have formats 8 through 15, Compiled methods being 12 through 15;
	 fail for CompiledMethod allowing the CoInterpeter to impose stricter bounds checks."
	cogit CmpCq: 8 R: ClassReg.
	jumpFmtLt8 := cogit JumpLess: 0.
	cogit CmpCq: 11 R: ClassReg.
	jumpFmtGt11 := cogit JumpGreater: 0.
	cogit
		AddCq: objectMemory baseHeaderSize R: Arg1Reg;
		MoveXbr: Arg1Reg R: ReceiverResultReg R: ReceiverResultReg.
	self genConvertIntegerToSmallIntegerInReg: ReceiverResultReg.
	cogit genPrimReturn.
	jumpFmtLt8 jmpTarget: cogit Label.
	self assert: objectMemory wordSize = 4. "documenting my laziness"
	cogit "Too lazy [knackered, more like. ed.] to define index with displacement addressing right now"
		AddCq: objectMemory baseHeaderSize / objectMemory wordSize R: Arg1Reg;
		MoveXwr: Arg1Reg R: ReceiverResultReg R: ReceiverResultReg;
		CmpCq: 16r3FFFFFFF R: ReceiverResultReg.
	jumpLarge := cogit JumpAbove: 0.
	self genConvertIntegerToSmallIntegerInReg: ReceiverResultReg.
	cogit genPrimReturn.
	jumpLarge jmpTarget: (cogit CallRT: cePositive32BitIntegerTrampoline).
	cogit
		MoveR: TempReg R: ReceiverResultReg;
		genPrimReturn.
	jumpSI jmpTarget:
	(jumpNotSI jmpTarget:
	(jumpNotIndexable jmpTarget:
	(jumpIsContext jmpTarget:
	(jumpBounds jmpTarget:
	(jumpFmtGt11 jmpTarget:
		cogit Label))))).
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'primitive generators' stamp: 'nice 4/29/2016 22:49'!
genPrimitiveIdenticalOrNotIf: orNot
	| jumpCmp |
	<var: #jumpCmp type: #'AbstractInstruction *'>
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	cogit CmpR: Arg0Reg R: ReceiverResultReg.
	jumpCmp := orNot
					ifTrue: [cogit JumpZero: 0]
					ifFalse: [cogit JumpNonZero: 0].
	cogit genMoveTrueR: ReceiverResultReg.
	cogit genPrimReturn.
	jumpCmp jmpTarget: (cogit genMoveFalseR: ReceiverResultReg).
	cogit genPrimReturn.
	^UnfailingPrimitive! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:37'!
genPrimitiveIdentityHash
	| jumpSI |
	<var: #jumpSI type: #'AbstractInstruction *'>
	jumpSI := self genJumpSmallInteger: ReceiverResultReg.
	self genGetHashFieldNonImmOf: ReceiverResultReg asSmallIntegerInto: TempReg.
	cogit MoveR: TempReg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpSI jmpTarget: cogit Label.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'primitive generators' stamp: 'eem 12/26/2019 21:59'!
genPrimitiveSize
	| jumpSI jumpNotIndexable jumpIsContext |
	"c.f. StackInterpreter>>stSizeOf: lengthOf:baseHeader:format: fixedFieldsOf:format:length:"
	<var: #jumpSI type: #'AbstractInstruction *'>
	<var: #jumpIsContext type: #'AbstractInstruction *'>
	<var: #jumpNotIndexable type: #'AbstractInstruction *'>
	jumpSI := self genJumpSmallInteger: ReceiverResultReg.
	self
		genGetSizeOf: ReceiverResultReg
		into: ClassReg
		formatReg: SendNumArgsReg
		scratchReg: TempReg
		abortJumpsInto: [:jnx :jic| jumpNotIndexable := jnx. jumpIsContext := jic].
	self genConvertIntegerInReg: ClassReg toSmallIntegerInReg: ReceiverResultReg.
	cogit genPrimReturn.
	jumpSI jmpTarget: (jumpNotIndexable jmpTarget: (jumpIsContext jmpTarget: cogit Label)).
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:35'!
genPrimitiveStringAt
	| jumpNotSI jumpNotByteIndexable jumpBounds jumpShortHeader jumpSkip |
	"c.f. StackInterpreter>>stSizeOf: lengthOf:baseHeader:format: fixedFieldsOf:format:length:"
	<var: #jumpNotSI type: #'AbstractInstruction *'>
	<var: #jumpNotByteIndexable type: #'AbstractInstruction *'>
	<var: #jumpBounds type: #'AbstractInstruction *'>
	<var: #jumpShortHeader type: #'AbstractInstruction *'>
	<var: #jumpSkip type: #'AbstractInstruction *'>
	cogit genLoadArgAtDepth: 0 into: Arg0Reg.
	cogit MoveR: Arg0Reg R: Arg1Reg.
	jumpNotSI := self genJumpNotSmallInteger: Arg0Reg.
	cogit
		MoveMw: 0 r: ReceiverResultReg R: TempReg;	"self baseHeader: receiver"
		MoveR: TempReg R: ClassReg;					"copy header word; we'll need it later"
		LogicalShiftRightCq: objectMemory instFormatFieldLSB R: TempReg;
		AndCq: self instFormatFieldMask R: TempReg;	"self formatOfHeader: destReg"
		MoveR: TempReg R: SendNumArgsReg;
		AndCq: 3 R: SendNumArgsReg;					"get odd bytes from format (if it turns out to be bytes)"
		SubR: SendNumArgsReg R: TempReg;
		CmpCq: 8 R: TempReg.							"check format is 8"
	jumpNotByteIndexable := cogit JumpNonZero: 0.
	cogit
		MoveR: ClassReg R: TempReg;
		AndCq: TypeMask R: TempReg;
		CmpCq: HeaderTypeSizeAndClass R: TempReg.	"(hdr bitAnd: TypeMask) = HeaderTypeSizeAndClass"
	jumpShortHeader := cogit JumpNonZero: 0.
	self assert: Size4Bit = 0.
	cogit
		MoveMw: 0 - (2 * objectMemory wordSize) r: ReceiverResultReg R: ClassReg; "(self sizeHeader: oop) bitAnd: LongSizeMask"
		AndCq: LongSizeMask signedIntFromLong R: ClassReg.
	jumpSkip :=  cogit Jump: 0.
	jumpShortHeader jmpTarget: (cogit AndCq: SizeMask R: ClassReg).	"hdr bitAnd: SizeMask"
	jumpSkip jmpTarget: (cogit SubCq: objectMemory baseHeaderSize R: ClassReg). "sz - BaseHeaderSize"
	cogit SubR: SendNumArgsReg R: ClassReg. "sz - (fmt bitAnd: 3)"
	self genConvertSmallIntegerToIntegerInReg: Arg1Reg.
	cogit SubCq: 1 R: Arg1Reg.
	cogit CmpR: ClassReg R: Arg1Reg.
	jumpBounds := cogit JumpAboveOrEqual: 0.
	cogit AddCq: objectMemory baseHeaderSize R: Arg1Reg.
	cogit genMoveConstant: objectMemory characterTable R: Arg0Reg.
	cogit MoveXbr: Arg1Reg R: ReceiverResultReg R: ReceiverResultReg;
		AddCq: objectMemory baseHeaderSize / objectMemory wordSize R: ReceiverResultReg.
	cogit MoveXwr: ReceiverResultReg R: Arg0Reg R: ReceiverResultReg.
	cogit genPrimReturn.
	jumpNotSI jmpTarget:
	(jumpNotByteIndexable jmpTarget:
	(jumpBounds jmpTarget:
		cogit Label)).
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 1/22/2009 12:04'!
genRemoveSmallIntegerTagsInScratchReg: scratchReg 
	cogit SubCq: 1 R: scratchReg.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 1/22/2009 12:05'!
genSetSmallIntegerTagsIn: scratchReg 
	cogit OrCq: 1 R: scratchReg.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem 9/9/2009 12:06'!
genShiftAwaySmallIntegerTagsInScratchReg: scratchReg
	cogit ArithmeticShiftRightCq: 1 R: scratchReg.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'cb 3/30/2016 15:06'!
genStoreSourceReg: sourceReg slotIndex: index destReg: destReg scratchReg: scratchReg inFrame: inFrame needsStoreCheck: needsStoreCheck
	| jmpImmediate jmpDestYoung jmpSourceOld jmpAlreadyRoot mask rootBitByteOffset |
	<var: #jmpImmediate type: #'AbstractInstruction *'>
	<var: #jmpDestYoung type: #'AbstractInstruction *'>
	<var: #jmpSourceOld type: #'AbstractInstruction *'>
	<var: #jmpAlreadyRoot type: #'AbstractInstruction *'>
	
	cogit genTraceStores.
	"do the store"
	cogit MoveR: sourceReg Mw: index * objectMemory wordSize + objectMemory baseHeaderSize r: destReg.
	"if no need for the store check then returns"
	needsStoreCheck ifFalse: [ ^ 0 ].
	"now the check.  Is value stored an integer?  If so we're done"
	jmpImmediate := self genJumpImmediate: sourceReg.
	"Get the old/new boundary in scratchReg"
	cogit MoveAw: objectMemory youngStartAddress R: scratchReg.
	"Is target young?  If so we're done"
	cogit CmpR: scratchReg R: destReg. "N.B. FLAGS := destReg - scratchReg"
	jmpDestYoung := cogit JumpAboveOrEqual: 0.
	"Is value stored old?  If so we're done."
	cogit CmpR: scratchReg R: sourceReg. "N.B. FLAGS := sourceReg - scratchReg"
	jmpSourceOld := cogit JumpBelow: 0.
	"value is young and target is old.
	 Need to make this a root if the root bit is not already set.
	 Test the root bit.  Only need to fetch the byte containing it,
	 which reduces the size of the mask constant."
	rootBitByteOffset := jmpSourceOld isBigEndian
							ifTrue: [objectMemory wordSize - RootBitDigitLength]
							ifFalse:[RootBitDigitLength - 1].
	mask := RootBitDigitLength > 1
				ifTrue: [RootBit >> (RootBitDigitLength - 1 * 8)]
				ifFalse: [RootBit].
	cogit MoveMb: rootBitByteOffset r: destReg R: scratchReg.
	cogit AndCq: mask R: scratchReg.
	jmpAlreadyRoot := cogit JumpNonZero: 0.
	"Root bit is not set.  Call store check to insert dest into root table."
	self assert: destReg == ReceiverResultReg.
	cogit 
		evaluateTrampolineCallBlock: [cogit CallRT: ceStoreCheckTrampoline]
		protectLinkRegIfNot: inFrame.
	jmpImmediate jmpTarget:
	(jmpDestYoung jmpTarget:
	(jmpSourceOld jmpTarget:
	(jmpAlreadyRoot jmpTarget:
		cogit Label))).
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem (const to msg rewrite) 11/6/2014 10:37'!
genStoreSourceReg: sourceReg slotIndex: index intoNewObjectInDestReg: destReg
	cogit MoveR: sourceReg Mw: index * objectMemory wordSize + objectMemory baseHeaderSize r: destReg.
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'initialization' stamp: 'nice 10/26/2016 23:39'!
generateObjectRepresentationTrampolines
	"Do the store check.  Answer the argument for the benefit of the code generator;
	 ReceiverResultReg may be caller-saved and hence smashed by this call.  Answering
	 it allows the code generator to reload ReceiverResultReg cheaply."
	ceStoreCheckTrampoline := cogit
									genTrampolineFor: #ceStoreCheck:
									called: 'ceStoreCheckTrampoline'
									arg: ReceiverResultReg
									regsToSave: (CallerSavedRegisterMask bitClear: (cogit registerMaskFor: ReceiverResultReg))
									result: cogit returnRegForStoreCheck.
	ceCreateNewArrayTrampoline := cogit genTrampolineFor: #ceNewArraySlotSize:
											called: 'ceCreateNewArrayTrampoline'
											arg: SendNumArgsReg
											regsToSave: (CallerSavedRegisterMask bitClear: (cogit registerMaskFor: ReceiverResultReg))
											result: ReceiverResultReg.
	cePositive32BitIntegerTrampoline := cogit genTrampolineFor: #positive32BitIntegerFor:
												called: 'cePositive32BitIntegerTrampoline'
												arg: ReceiverResultReg
												regsToSave: (CallerSavedRegisterMask bitClear: (cogit registerMaskFor: ReceiverResultReg))
												result: TempReg.
	ceActiveContextTrampoline := self genActiveContextTrampoline.
	ceClosureCopyTrampoline := cogit genTrampolineFor: #ceClosureCopyDescriptor:
										called: 'ceClosureCopyTrampoline'
										arg: SendNumArgsReg
										regsToSave: (CallerSavedRegisterMask bitClear: (cogit registerMaskFor: ReceiverResultReg))
										result: ReceiverResultReg.
										
	LowcodeVM ifTrue:
		[cePositive32BitValueOfTrampoline := cogit genTrampolineFor: #positive32BitValueOf:
													called: 'cePositive32BitValueOfTrampoline'
													arg: ReceiverResultReg
													result: TempReg.
		ceSigned32BitIntegerTrampoline := cogit genTrampolineFor: #signed32BitIntegerFor:
													called: 'ceSigned32BitIntegerTrampoline'
													arg: ReceiverResultReg
													result: TempReg.
		ceSigned32BitValueOfTrampoline := cogit genTrampolineFor: #signed32BitValueOf:
													called: 'ceSigned32BitValueOfTrampoline'
													arg: ReceiverResultReg
													result: TempReg.

		self generateLowcodeObjectTrampolines]! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 8/17/2011 11:30'!
hasValidHeaderPostGC: cogMethod
	<var: #cogMethod type: #'CogMethod *'>
	<inline: true>
	^cogMethod objectHeader = objectMemory nullHeaderForMachineCodeMethod! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'in-line cacheing' stamp: 'eem 01/10/2016 10:01'!
illegalClassTag
	^0! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'in-line cacheing' stamp: 'eem 7/7/2014 09:53'!
inlineCacheTagForClass: classOop
	"Answer the relevant inline cache tag for a class.
	 c.f. inlineCacheTagForInstance:"
	| cci |
	classOop = objectMemory classSmallInteger ifTrue:
		[^objectMemory integerObjectOf: 0]. "the SmallInteger tag"
	(cci :=  objectMemory noShiftCompactClassIndexOfHeader: (objectMemory formatOfClass: classOop)) > 0 ifTrue:
		[^cci].
	^classOop! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'in-line cacheing' stamp: 'eem 7/7/2014 09:57'!
inlineCacheTagForInstance: oop
	"Answer the relevant inline cache tag for an instance.
	 c.f. getInlineCacheClassTagFrom:into: & inlineCacheTagForClass:"
	| cci |
	(objectMemory isIntegerObject: oop) ifTrue:
		[^objectMemory integerObjectOf: 0]. "the SmallInteger tag"
	(cci :=  objectMemory noShiftCompactClassIndexOf: oop) > 0 ifTrue:
		[^cci].
	^(objectMemory classHeader: oop) bitAnd: AllButTypeMask! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'in-line cacheing' stamp: 'eem 10/9/2009 03:09'!
inlineCacheTagIsYoung: cacheTag
	^objectMemory isYoung: cacheTag! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'in-line cacheing' stamp: 'eem 01/10/2016 10:01'!
inlineCacheTagsMayBeObjects
	^true! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'object representation' stamp: 'eem 8/19/2011 10:34'!
instFormatFieldMask
	"This is the mask for the field when shifted into the least significant bits"
	<inline: true>
	^(1 << objectMemory instFormatFieldWidth) - 1! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 12/7/2013 10:48'!
isOopForwarded: oop
	"Compatibility wth SpurMemoryManager.  In ObjectMemory, no forwarding pointers
	 are visible to the VM."
	<inline: true>
	^false! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'object representation' stamp: 'eem 8/3/2021 18:54'!
isSmallIntegerTagNonZero
	<inline: #always>
	^true! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 09/28/2015 18:48'!
markAndTraceCacheTagLiteral: literal in: cogMethodOrNil atpc: address
	"Mark and trace a literal in an inline cache preceding address in cogMethodOrNil.
	 Answer if code was modified."
	<var: #cogMethodOrNil type: #'CogMethod *'>
	<var: #address type: #usqInt>
	<inline: true>
	self markAndTraceLiteral: literal.
	^false! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 10/9/2009 03:09'!
markAndTraceLiteral: literal
	(self couldBeObject: literal) ifTrue:
		[self assert: (objectMemory addressCouldBeObj: literal).
		 objectMemory markAndTrace: literal]! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 8/4/2014 08:24'!
markAndTraceLiteral: literal in: cogMethod at: address
	"Mark and trace a literal in a sqInt variable of cogMethod."
	<var: #cogMethod type: #'CogMethod *'>
	<var: #address type: #'sqInt *'>
	<inline: true>
	self markAndTraceLiteral: literal! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 09/28/2015 18:48'!
markAndTraceLiteral: literal in: cogMethodOrNil atpc: address
	"Mark and trace a literal in a machine code instruction preceding address in cogMethodOrNil.
	 Answer if code was modified."
	<var: #cogMethodOrNil type: #'CogMethod *'>
	<var: #address type: #usqInt>
	<inline: true>
	self markAndTraceLiteral: literal.
	^false! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 11/18/2013 15:18'!
markAndTraceLiteralIfYoung: literal
	((self couldBeObject: literal)
	 and: [objectMemory isYoungObject: literal]) ifTrue:
		[self assert: (objectMemory addressCouldBeObj: literal).
		 objectMemory markAndTrace: literal]! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'sista support' stamp: 'eem 7/5/2014 17:54'!
numCountersFor: theCounters
	<var: #theCounters type: #usqInt>
	^objectMemory longAt: theCounters - (self sizeof: #sqInt)! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'calling convention' stamp: 'eem 9/5/2021 07:55'!
numRegArgs
	"Define how many register arguments a StackToRegisterMappingCogit can
	 and should use with the receiver.  The value must be 0, 1 or 2.  Note that a
	 SimpleStackBasedCogit always has 0 register args (although the receiver is
	 passed in a register).  CogObjectRepresentationForSqueakV3 only implements
	 at most 1-arg primitives, because the complexity of the object representation
	 makes it difficult to implement at:put:, the most performance-critical 2-argument
	 primitive..  The method must be inlined in CoInterpreter, and dead code eliminated
	 so that the register-popping enilopmarts such as enterRegisterArgCogMethod:-
	 at:receiver: do not have to be implemented in SimpleStackBasedCogit."
	<api>
	<option: #StackToRegisterMappingCogit>
	<cmacro: '() 1'> "to allow inlining optimization across the CoInterpreter/Cogit api..."
	^1! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'object representation' stamp: 'eem 1/20/2009 15:40'!
numSmallIntegerBits
	^31! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'object representation' stamp: 'eem 9/9/2009 12:28'!
numSmallIntegerTagBits
	^1! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 2/20/2016 18:12'!
remapObject: objOop
	self assert: (objectMemory addressCouldBeObj: objOop).
	^objectMemory remap: objOop! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 10/9/2009 03:10'!
remapOop: oop
	^(objectMemory isIntegerObject: oop)
		ifTrue: [oop]
		ifFalse: [objectMemory remap: oop]! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'sista support' stamp: 'eem 4/19/2016 14:24'!
resetCountersIn: cogMethod
	<var: #cogMethod type: #'CogMethod *'>
	<api>
	<option: #SistaCogit>
	cogit
		fillInCounters: (self numCountersFor: cogMethod counters)
		atStartAddress: cogMethod counters! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 4/3/2013 15:07'!
shouldAnnotateObjectReference: anOop
	"self assert: ((objectMemory isIntegerObject: anOop)
				or: [objectMemory addressCouldBeObj: anOop])."
	^(objectMemory isNonIntegerObject: anOop)
	  and: [self oop: anOop isGreaterThan: objectMemory trueObject]! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'compile abstract instructions' stamp: 'eem (const to msg rewrite) 11/6/2014 10:37'!
slotOffsetOfInstVarIndex: index
	^index * objectMemory wordSize + objectMemory baseHeaderSize! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'testing' stamp: 'eem 8/3/2021 18:54'!
smallIntegerIsOnlyImmediateType
	<inline: #always>
	^true! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'garbage collection' stamp: 'eem 12/5/2014 18:43'!
validInlineCacheTag: cacheTag
	^cacheTag = ConstZero
	  or: [((cacheTag bitAnd: 1 << objectMemory shiftForWord - 1) = 0
		   and: [cacheTag
				between: 1 << objectMemory compactClassFieldLSB
				and: (objectMemory compactClassIndexOfHeader: -1) << objectMemory compactClassFieldLSB])
		 or: [self checkValidObjectReference: cacheTag]]! !

!CogObjectRepresentationForSqueakV3 methodsFor: 'sista support' stamp: 'eem 4/6/2015 09:28'!
valueOfAssociation: association
	^objectMemory fetchPointer: ValueIndex ofObject: association! !

CogClass subclass: #CogStackPages
	instanceVariableNames: 'coInterpreter objectMemory pages mostRecentlyUsedPage overflowLimit bytesPerPage statNumMaps statPageCountWhenMappingSum statMaxPageCountWhenMapping'
	classVariableNames: ''
	poolDictionaries: 'VMBasicConstants'
	category: 'VMMaker-Interpreter'!

!CogStackPages commentStamp: 'eem 8/14/2015 16:43' prior: 0!
I am a class that helps organize the StackInterpreter's collection of stack pages.  I hold the set of stack pages represented by InterpreterStackPage instances/StackPage structs.  The pages are held in a doubly-linked list that notionally has two heads:

mostRecentlyUsedPage-->used page<->used page<->used page<->used page<--leastRecentlyUsedPage
                                       ^                        <-next-prev->                         ^
                                        |                                                                       |
                                        v                        <-prev-next->                         v
                                        free page<->free page<->free page<->free page

In fact we don't need the least-recently-used page, and so it is only present conceptually.  The point is that there is a possibly empty but contiguous sequence of free pages starting at mostRecentlyUsedPage nextPage.  New pages are allocated preferentially from the free page next to the MRUP.
If there are no free pages then (effectively) the LRUP's frames are flushed to contexts and it is used instead.

I have two concrete classes, one for the StackInterpreter and one for the CoInterpreter.

Instance Variables
	bytesPerPage:						<Integer>
	coInterpreter:						<StackInterpreter>
	mostRecentlyUsedPage:			<CogStackPage>
	objectMemory:						<ObjectMemory|SpurMemoryManager>
	overflowLimit:						<Integer>
	pages:								<Array of: CogStackPage>
	statNumMaps:						<Integer>
	statPageCountWhenMappingSum:		<Integer>

bytesPerPage
	- the size of a page in bytes

coInterpreter
	- the interpreter the receiver is holding pages for

mostRecentlyUsedPage
	- the most recently used stack page

objectMemory
	- the objectMemory of the interpreter

overflowLimit
	- the length in bytes of the portion of teh stack that can be used for frames before the page is judged to have overflowed

pages
	- the collection of stack pages the receiver is managing

statNumMaps
	- the number of mapStackPages calls

statPageCountWhenMappingSum:
	- the sum of the number of in use pages at each mapStackPages, used to estimate the average number of in use pages at scavenge, which heavily influences scavenger performance
!

CogStackPages subclass: #CoInterpreterStackPages
	instanceVariableNames: 'stackBasePlus1 pageMap minStackAddress maxStackAddress'
	classVariableNames: ''
	poolDictionaries: 'VMBasicConstants'
	category: 'VMMaker-JIT'!

!CoInterpreterStackPages commentStamp: 'eem 8/14/2015 16:10' prior: 0!
I am a class that helps organize the CoInterpreter's collection of stack pages.  I hold the set of stack pages represented by CogStackPageSurrogate instances/StackPage structs.  The pages are held in a doubly-linked list that notionally has two heads:

mostRecentlyUsedPage-->used page<->used page<->used page<->used page<--leastRecentlyUsedPage
                                       ^                        <-next-prev->                         ^
                                        |                                                                       |
                                        v                        <-prev-next->                         v
                                        free page<->free page<->free page<->free page

In fact we don't need the least-recently-used page, and so it is only present conceptually.  The point is that there is a possibly empty but contiguous sequence of free pages starting at mostRecentlyUsedPage nextPage.  New pages are allocated preferentially from the free page next to the MRUP.
If there are no free pages then (effectively) the LRUP's frames are flushed to contexts and it is used instead.

Instance Variables
	maxStackAddress:		<Integer>
	minStackAddress:		<Integer>
	pageMap:				<Dictionary>
	stackBasePlus1:		<Integer>

maxStackAddress
	- the maximum valid byte address in the stack zone

minStackAddress
	- the minimum valid byte address in the stack zone

pageMap
	- a map from address to the CogStackPageSurrogate for that address

stackBasePlus1
	- the address of the 2nd byte in the stack memory, used for mapping stack addresses to page indices!

!CoInterpreterStackPages class methodsFor: 'translation' stamp: 'eem 8/14/2015 17:54'!
declareCVarsIn: aCCodeGenerator
	aCCodeGenerator
		var: #stackBasePlus1 type: #'char *';
		removeVariable: 'pageMap'; "These are simulation/debugging things only"
		removeVariable: 'maxStackAddress'; "These are simulation/debugging things only"
		removeVariable: 'minStackAddress' "These are simulation/debugging things only"! !

!CoInterpreterStackPages methodsFor: 'assertions' stamp: 'eem (const to msg rewrite) 11/6/2014 09:35'!
couldBeFramePointer: pointer
	"Answer if the argument is a properly aligned pointer into the stack zone."
	<var: #pointer type: #'void *'>
	^(pointer asUnsignedInteger bitAnd: objectMemory wordSize - 1) = 0
	   and: [pointer asUnsignedInteger
				between: (stackBasePlus1 - 1) asUnsignedInteger
				and: (self cCode: [pages asUnsignedInteger]
							inSmalltalk: [(self stackPageAt: 0) asUnsignedInteger])]! !

!CoInterpreterStackPages methodsFor: 'initialization' stamp: 'eem 1/24/2020 21:05'!
initializeStack: theStackPages numSlots: stackSlots pageSize: slotsPerPage
	"Initialize the stack pages.  In the C VM theStackPages will be alloca'ed memory to hold the
	 stack pages on the C stack.  In the simulator they are housed in the memory between the
	 cogMethodZone and the heap."

	<var: #theStackPages type: #'char *'>
	<returnTypeC: #void>
	| numPages page structStackPageSize pageStructBase count |
	<var: #page type: #'StackPage *'>
	<var: #pageStructBase type: #'char *'>
	self cCode: []
		inSmalltalk:
			[self assert: objectMemory startOfMemory - coInterpreter effectiveCogCodeSize - Cogit guardPageSize - coInterpreter methodCacheSize - coInterpreter primTraceLogSize - coInterpreter rumpCStackSize
					= (stackSlots * objectMemory wordSize roundUpTo: objectMemory allocationUnit)].
	structStackPageSize := coInterpreter sizeof: CogStackPage.
	bytesPerPage := slotsPerPage * objectMemory wordSize.
	numPages := coInterpreter numStkPages.

	"Because stack pages grow down baseAddress is at the top of a stack page and so to avoid
	 subtracting BytesPerWord from baseAddress and lastAddress in the init loop below we simply
	 push the stackPage array up one word to avoid the overlap.  This word is extraStackBytes."
	pageStructBase := theStackPages + (numPages * bytesPerPage) + objectMemory wordSize.
	pages := self cCode: [self cCoerceSimple: pageStructBase to: #'StackPage *']
				  inSmalltalk:
					[pageMap := Dictionary new.
					 ((0 to: numPages - 1) collect:
						[:i|
						 CogStackPage surrogateClass new
							address: pageStructBase + (i * structStackPageSize)
							simulator: coInterpreter
							zoneBase: coInterpreter stackZoneBase
							zoneLimit: objectMemory startOfMemory])
						do: [:pageSurrogate|
							pageMap at: pageSurrogate address put: pageSurrogate];
						yourself].
	"make sure there's enough headroom"
	self assert: coInterpreter stackPageByteSize - coInterpreter stackLimitBytes - coInterpreter stackLimitOffset
				>= coInterpreter stackPageHeadroom.
	0 to: numPages - 1 do:
		[:index|
		 page := self stackPageAt: index.
		 page
			lastAddress: theStackPages + (index * bytesPerPage);
			baseAddress: page lastAddress + bytesPerPage;
			stackLimit: page baseAddress - coInterpreter stackLimitBytes;
			realStackLimit: page stackLimit;
			baseFP: 0;
			nextPage: (self stackPageAt: (index = (numPages - 1) ifTrue: [0] ifFalse: [index + 1]));
			prevPage: (self stackPageAt: (index = 0 ifTrue: [numPages - 1] ifFalse: [index - 1]))].

	"Now compute stackBasePlus1 so that the pageIndexFor: call maps all addresses from
	 aPage baseAddress to aBase limitAddress + 1 to the same index (stacks grow down)"
	stackBasePlus1 := (self cCoerceSimple: theStackPages to: #'char *') + 1.
	self cCode: []
		inSmalltalk:
			[minStackAddress := theStackPages.
			 maxStackAddress := theStackPages + (numPages * bytesPerPage) + objectMemory wordSize - 1].

	"The overflow limit is the amount of stack to retain when moving frames from an overflowing
	 stack to reduce thrashing.  See stackOverflowOrEvent:mayContextSwitch:"
	page := self stackPageAt: 0.
	overflowLimit := page baseAddress - page realStackLimit * 3 // 5.
	0 to: numPages - 1 do:
		[:index|
		 page := self stackPageAt: index.
		 self assert: (self pageIndexFor: page baseAddress) == index.
		 self assert: (self pageIndexFor: page baseAddress - (slotsPerPage - 1 * objectMemory wordSize)) == index.
		 self assert: (self stackPageFor: page baseAddress) == page.
		 self assert: (self stackPageFor: page stackLimit) == page.
		 self cCode: []
			inSmalltalk:
				[| memIndex |
				 memIndex := index * slotsPerPage + 1. "this is memIndex in the block above"
				 self assert: (self memIndexFor: (self oopForPointer: page baseAddress))
							== (memIndex + slotsPerPage - 1).
				 index < (numPages - 1) ifTrue:
					[self assert: (self stackPageFor: page baseAddress + objectMemory wordSize) == (self stackPageAt: index + 1)]].
		coInterpreter initializePageTraceToInvalid: page].

	mostRecentlyUsedPage := self stackPageAt: 0.
	page := mostRecentlyUsedPage.
	count := 0.
	[| theIndex |
	 count := count + 1.
	 theIndex := self pageIndexFor: page baseAddress.
	 self assert: (self stackPageAt: theIndex) == page.
	 self assert: (self pageIndexFor: page baseAddress) == theIndex.
	 self assert: (self pageIndexFor: page stackLimit) == theIndex.
	 self assert: (self pageIndexFor: page lastAddress + 1) == theIndex.
	 (page := page nextPage) ~= mostRecentlyUsedPage] whileTrue.
	self assert: count == numPages.
	self assert: self pageListIsWellFormed! !

!CoInterpreterStackPages methodsFor: 'initialization' stamp: 'eem 8/14/2015 15:48'!
initializeWithByteSize: byteSize "<Integer>" for: anInterpreter "<CoInterpreter>" "^<Array of: <Integer>"
	"Initialize the stackPages memory for simulation.  To keep access monitoring
	 in one place we defer to the coInterpreter for accessing memory.  Answer the
	 base address of th ememory."
	<doNotGenerate>
	coInterpreter := anInterpreter.
	objectMemory := coInterpreter objectMemory.
	^anInterpreter stackZoneBase! !

!CoInterpreterStackPages methodsFor: 'memory access' stamp: 'eem 11/10/2014 15:45'!
longAt: byteAddress
	<doNotGenerate>
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	self assert: (byteAddress >= minStackAddress and: [byteAddress < maxStackAddress]).
	^objectMemory longAt: byteAddress! !

!CoInterpreterStackPages methodsFor: 'memory access' stamp: 'eem 10/8/2015 16:26'!
longAt: byteAddress put: a32Or64BitValue
	<doNotGenerate>
	self assert: (byteAddress >= minStackAddress and: [byteAddress < maxStackAddress]).
	^objectMemory longAt: byteAddress put: a32Or64BitValue! !

!CoInterpreterStackPages methodsFor: 'page access' stamp: 'eem 8/14/2015 16:45'!
memIndexFor: byteAddress
	"Map an address into the stack zone into a word index into the slots in the stack zone."
	<doNotGenerate>
	^(self oopForPointer: byteAddress) - coInterpreter stackZoneBase - 1 // objectMemory wordSize + 1! !

!CoInterpreterStackPages methodsFor: 'page access' stamp: 'eem 12/14/2009 17:38'!
pageIndexFor: pointer "<Integer>"
	"Answer the page index for a pointer into stack memory, i.e. the index
	 for the page the address is in.  N.B.  This is a zero-relative index."
	<var: #pointer type: #'void *'>
	<inline: true>
	self assert: ((self cCoerceSimple: pointer to: #'char *')
					between: stackBasePlus1 - 1
					and: (self cCode: [self cCoerceSimple: pages to: #'char *']
							inSmalltalk: [(self stackPageAt: 0) asInteger])).
	^self pageIndexFor: pointer stackBasePlus1: stackBasePlus1 bytesPerPage: bytesPerPage! !

!CoInterpreterStackPages methodsFor: 'page access' stamp: 'eem 2/28/2009 23:23'!
pageIndexFor: pointer "<Integer>" stackBasePlus1: stkBasePlus1 "<Integer>" bytesPerPage: pageByteSize "<Integer>"
	"Answer the page index for a pointer into stack memory, i.e. the index
	 for the page the address is in.  N.B.  This is a zero-relative index."
	<cmacro: '(pointer,stkBasePlus1,pageByteSize) (((char *)(pointer) - (stkBasePlus1)) / (pageByteSize))'>
	^pointer - stkBasePlus1 // pageByteSize! !

!CoInterpreterStackPages methodsFor: 'accessing' stamp: 'eem 11/6/2014 09:56'!
surrogateAtAddress: anAddress
	<doNotGenerate>
	^pageMap at: anAddress! !

!CoInterpreterStackPages methodsFor: 'debug printing' stamp: 'eem 12/22/2019 19:03'!
whereIsMaybeStackThing: anOop
	"If anOop is an address within the stack zone answer a string stating that, otherwise answer nil."
	<returnTypeC: 'char *'>
	self cCode: '' inSmalltalk: [stackBasePlus1 isNil ifTrue: [^nil]].
	(self oop: anOop
		isGreaterThanOrEqualTo: stackBasePlus1 - 1
		andLessThan: (self cCode: [pages]
							inSmalltalk: [(self stackPageAt: 0) asUnsignedInteger])) ifTrue:
		[^' is in the stack zone'].
	^nil! !

CoInterpreterStackPages subclass: #CoInterpreterStackPagesLSB
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CoInterpreterStackPagesLSB methodsFor: 'memory access' stamp: 'eem 10/8/2015 16:29'!
byteAt: byteAddress
	| lowBits long |
	lowBits := byteAddress bitAnd: objectMemory wordSize - 1.
	long := self longAt: byteAddress - lowBits.
	lowBits > 0 ifTrue:
		[long := long bitShift: lowBits * -8].
	^long bitAnd: 16rFF! !

!CoInterpreterStackPagesLSB methodsFor: 'memory access' stamp: 'eem 10/8/2015 16:32'!
byteAt: byteAddress put: byte
	| lowBits long longAddress |
	self assert: (byte between: 0 and: 16rFF).
	lowBits := byteAddress bitAnd: objectMemory wordSize - 1.
	longAddress := byteAddress - lowBits.
	long := self longAt: longAddress.
	long := (long bitOr: (16rFF bitShift: lowBits * 8)) bitXor: (byte bitXor: 16rFF).
	self longAt: longAddress put: long.
	^byte! !

CoInterpreterStackPages subclass: #CoInterpreterStackPagesMSB
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CoInterpreterStackPagesMSB methodsFor: 'memory access' stamp: 'eem 9/10/2013 14:33'!
byteAt: byteAddress
	| lowBits bpwMinus1 |
	bpwMinus1 := objectMemory wordSize - 1.
	lowBits := byteAddress bitAnd: bpwMinus1.
	^((self longAt: byteAddress - lowBits)
		bitShift: (lowBits - bpwMinus1) * 8)
		bitAnd: 16rFF! !

!CogStackPages class methodsFor: 'translation' stamp: 'eem 8/14/2015 16:46'!
declareCVarsIn: aCCodeGenerator
	aCCodeGenerator
		var: #mostRecentlyUsedPage type: #'StackPage *';
		var: #pages type: #'StackPage *'.
	aCCodeGenerator
		removeVariable: 'coInterpreter'; "These are simulation/debugging things only"
		removeVariable: 'objectMemory' "These are simulation/debugging things only"! !

!CogStackPages methodsFor: 'assertions' stamp: 'bgf 3/15/2009 01:20'!
allPagesFree
	<doNotGenerate>
	^pages allSatisfy: [:page| (self isFree: page)]! !

!CogStackPages methodsFor: 'memory access' stamp: 'eem 6/20/2008 11:07'!
byteAt: byteAddress "<Integer>" 
	self subclassResponsibility! !

!CogStackPages methodsFor: 'initialization' stamp: 'eem (const to msg rewrite) 11/6/2014 09:35'!
extraStackBytes
	"See initializeStack:numSlots:pageSize:stackLimitOffset:stackPageHeadroom:
	``Because stack pages grow down...''"
	^objectMemory wordSize! !

!CogStackPages methodsFor: 'page access' stamp: 'eem 7/30/2013 11:26'!
freeStackPage: aPage "<InterpreterStackPage>"
	"MRUP-->used page<->used page<->used page<->used page<--LRUP
	               ^                        <-next-prev->                         ^
	                |                                                                       |
	                v                        <-prev-next->                         v
	                free page<->free page<->free page<->free page"
	<var: #aPage type: #'StackPage *'>
	<inline: false>
	self freeStackPageNoAssert: aPage.
	self assert: self pageListIsWellFormed! !

!CogStackPages methodsFor: 'page access' stamp: 'eem 6/25/2009 11:41'!
freeStackPageNoAssert: aPage "<InterpreterStackPage>"
	"MRUP-->used page<->used page<->used page<->used page<--LRUP
	               ^                        <-next-prev->                         ^
	                |                                                                       |
	                v                        <-prev-next->                         v
	                free page<->free page<->free page<->free page"
	<var: #aPage type: #'StackPage *'>
	<returnTypeC: #void>
	| prev |
	<var: #prev type: #'StackPage *'>
	aPage baseFP: 0.
	aPage == mostRecentlyUsedPage ifTrue:
		[mostRecentlyUsedPage := mostRecentlyUsedPage prevPage.
		 ^nil].
	(prev := aPage prevPage) isFree ifTrue:
		[^nil].
	prev nextPage: aPage nextPage.
	aPage nextPage prevPage: prev.
	aPage nextPage: mostRecentlyUsedPage nextPage.
	mostRecentlyUsedPage nextPage prevPage: aPage.
	aPage prevPage: mostRecentlyUsedPage.
	mostRecentlyUsedPage nextPage: aPage! !

!CogStackPages methodsFor: 'initialization' stamp: 'eem 8/15/2015 01:15'!
initialize
	"Here we can initialize the variables C initializes to zero.  #initialize methods do /not/ get translated."
	statNumMaps := statPageCountWhenMappingSum := statMaxPageCountWhenMapping := 0! !

!CogStackPages methodsFor: 'initialization' stamp: 'eem 8/14/2015 15:57'!
initializeStack: theStackPages numSlots: stackSlots pageSize: slotsPerPage
	"Initialize the stack pages.  In the C VM theStackPages will be alloca'ed memory to hold the
	 stack pages on the C stack.  In the simulator they are housed in the memory between the
	 cogMethodZone and the heap."
	<returnTypeC: #void>
	self subclassResponsibility! !

!CogStackPages methodsFor: 'initialization' stamp: 'eem 8/14/2015 15:47'!
initializeWithByteSize: byteSize "<Integer>" for: anInterpreter "<CoInterpreter>" "^<Array of: <Integer>"
	"Initialize the stackPages memory for simulation. Answer the base address of the memory."
	self subclassResponsibility! !

!CogStackPages methodsFor: 'page access' stamp: 'eem 6/24/2009 13:31'!
isFree: thePage
	"This is an anachronism.  Previously Slang couldn't generate the method correctly
	 from e.g. CogStackPageSurrogate>>isFree since Slang didn't do substitution on self.
	 Now it does, but there are still callers of isFree: so we keep this for simulation."
	<doNotGenerate>
	^thePage baseFP = 0! !

!CogStackPages methodsFor: 'page access' stamp: 'eem 3/8/2013 15:25'!
markStackPageLeastMostRecentlyUsed: page "<InterpreterStackPage>"
	"This method is used to move a page to the end of the used pages.
	 This is to keep asserts checking pageListIsWellFormed happy."

	"MRUP-->used page<->used page<->used page<->used page<--LRUP
	               ^                        <-next-prev->                         ^
	                |                                                                       |
	                v                        <-prev-next->                         v
	                free page<->free page<->free page<->free page"

	<var: #page type: #'StackPage *'>
	<returnTypeC: #void>
	| lastUsedPage |
	<var: #lastUsedPage type: #'StackPage *'>
	self assert: page = mostRecentlyUsedPage nextPage.
	lastUsedPage := page nextPage.
	[lastUsedPage isFree] whileTrue:
		[lastUsedPage := lastUsedPage nextPage].
	lastUsedPage nextPage = page ifTrue:
		[^nil].
	page prevPage nextPage: page nextPage.
	page nextPage prevPage: page prevPage.
	lastUsedPage prevPage nextPage: page.
	page prevPage: lastUsedPage prevPage.
	page nextPage: lastUsedPage.
	lastUsedPage prevPage: page.
	self assert: self pageListIsWellFormed! !

!CogStackPages methodsFor: 'page access' stamp: 'eem 11/10/2014 12:49'!
markStackPageMostRecentlyUsed: page "<InterpreterStackPage>"
	"MRUP-->used page<->used page<->used page<->used page<--LRUP
	               ^                        <-next-prev->                         ^
	                |                                                                       |
	                v                        <-prev-next->                         v
	                free page<->free page<->free page<->free page"
	<var: #page type: #'StackPage *'>
	<returnTypeC: #void>
	page == mostRecentlyUsedPage ifTrue:
		[^nil].
	"Common case; making new page most recently used."
	page prevPage == mostRecentlyUsedPage ifTrue:
		[mostRecentlyUsedPage := page.
		 self assert: self pageListIsWellFormed.
		 ^nil].
	page prevPage nextPage: page nextPage.
	page nextPage prevPage: page prevPage.
	mostRecentlyUsedPage nextPage prevPage: page.
	page prevPage: mostRecentlyUsedPage.
	page nextPage: mostRecentlyUsedPage nextPage.
	mostRecentlyUsedPage nextPage: page.
	mostRecentlyUsedPage := page.
	self assert: self pageListIsWellFormed! !

!CogStackPages methodsFor: 'page access' stamp: 'eem 3/8/2013 14:57'!
markStackPageNextMostRecentlyUsed: page "<InterpreterStackPage>"
	"This method is used to move a page to a position in the list such that it cannot
	 be deallocated when a new page is allocated, without changing the most recently
	 used page.  There must be at least 3 pages in the system.  So making the page
	 the MRU's prevPage is sufficient to ensure it won't be deallocated."

	"MRUP-->used page<->used page<->used page<->used page<--LRUP
	               ^                        <-next-prev->                         ^
	                |                                                                       |
	                v                        <-prev-next->                         v
	                free page<->free page<->free page<->free page"

	<var: #page type: #'StackPage *'>
	<returnTypeC: #void>
	self assert: page ~~ mostRecentlyUsedPage.
	page nextPage == mostRecentlyUsedPage ifTrue:
		[^nil].
	page prevPage nextPage: page nextPage.
	page nextPage prevPage: page prevPage.
	mostRecentlyUsedPage prevPage nextPage: page.
	page prevPage: mostRecentlyUsedPage prevPage.
	page nextPage: mostRecentlyUsedPage.
	mostRecentlyUsedPage prevPage: page.
	self assert: self pageListIsWellFormed! !

!CogStackPages methodsFor: 'page access' stamp: 'eem 8/14/2015 16:44'!
memIndexFor: byteAddress
	"Map an address into the stack zone into a word index into the slots in the stack zone."
	<doNotGenerate>
	self subclassResponsibility! !

!CogStackPages methodsFor: 'page access' stamp: 'eem 3/3/2009 09:49'!
mostRecentlyUsedPage
	<cmacro: '() GIV(mostRecentlyUsedPage)'>
	<returnTypeC: #'StackPage *'> "this is to guide Slang's inliner"
	^mostRecentlyUsedPage! !

!CogStackPages methodsFor: 'page access' stamp: 'eem 8/14/2015 12:14'!
newStackPage
	"MRUP-->used page<->used page<->used page<->used page<--LRUP
	               ^                        <-next-prev->                         ^
	                |                                                                       |
	                v                        <-prev-next->                         v
	                free page<->free page<->free page<->free page"
	<returnTypeC: #'StackPage *'>
	| lruOrFree |
	<var: #lruOrFree type: #'StackPage *'>
	lruOrFree := mostRecentlyUsedPage nextPage.
	lruOrFree isFree ifTrue:
		[^lruOrFree].
	coInterpreter divorceFramesIn: lruOrFree.
	^lruOrFree! !

!CogStackPages methodsFor: 'page access' stamp: 'eem 10/29/2008 03:05'!
overflowLimit
	^overflowLimit! !

!CogStackPages methodsFor: 'page access' stamp: 'eem 8/14/2015 16:05'!
pageIndexFor: pointer "<Integer>"
	"Answer the page index for a pointer into stack memory, i.e. the index
	 for the page the address is in.  N.B.  This is a zero-relative index."
	<var: #pointer type: #'void *'>
	<inline: true>
	self subclassResponsibility! !

!CogStackPages methodsFor: 'assertions' stamp: 'eem 1/16/2015 10:39'!
pageListIsWellFormed
	"Answer if the stack page list is well-formed.
	 MRUP-->used page<->used page<->used page<->used page<--LRUP
	               ^                        <-next-prev->                         ^
	                |                                                                       |
	                v                        <-prev-next->                         v
	                free page<->free page<->free page<->free page"
	| ok page count limit |
	<inline: false>
	<var: #page type: #'StackPage *'>
	ok := true.
	page := mostRecentlyUsedPage nextPage.
	count := 1.
	limit := coInterpreter numStkPages * 2.
	[page isFree
	 and: [page ~= mostRecentlyUsedPage
	 and: [count <= limit]]] whileTrue:
		[(self asserta: page nextPage prevPage == page) ifFalse:
			[ok := false].
		 page := page nextPage.
		 count := count + 1].
	[page ~= mostRecentlyUsedPage
	 and: [count <= limit]] whileTrue:
		[(self asserta: page nextPage prevPage == page) ifFalse:
			[ok := false].
		 (self asserta: page isFree not)
			ifTrue:
				[(self asserta: ((page addressIsInPage: page baseFP)
								and: [page addressIsInPage: page headSP])) ifFalse:
					[ok := false]]
			ifFalse:
				[ok := false].
		 page := page nextPage.
		 count := count + 1].
	(self asserta: count = coInterpreter numStkPages) ifFalse:
		[ok := false].
	^ok! !

!CogStackPages methodsFor: 'accessing' stamp: 'eem 11/6/2014 09:55'!
pages
	<doNotGenerate>
	^pages! !

!CogStackPages methodsFor: 'statistics' stamp: 'eem 8/15/2015 01:09'!
recordLivePagesOnMapping: numLivePages
	<inline: true>
	statNumMaps := statNumMaps + 1.
	statPageCountWhenMappingSum := statPageCountWhenMappingSum + numLivePages.
	statMaxPageCountWhenMapping := statMaxPageCountWhenMapping max: numLivePages! !

!CogStackPages methodsFor: 'initialization' stamp: 'eem 11/6/2014 09:56'!
setInterpreter: anInterpreter
	"Initialize the stackPages memory for simulation.  To keep access monitoring
	 in one place we defer to the coInterpreter for accessing memory."
	<doNotGenerate>
	coInterpreter := anInterpreter.
	objectMemory := coInterpreter objectMemory! !

!CogStackPages methodsFor: 'assertions' stamp: 'bgf 3/15/2009 01:20'!
somePageHasHeadFrameFP: theFP
	<doNotGenerate>
	^pages anySatisfy: [:page| page headFP = theFP]
! !

!CogStackPages methodsFor: 'page access' stamp: 'eem 8/14/2015 17:42'!
stackPageAt: index
	"Answer the page for a page index.
	 N.B.  This is a zero-relative index."
	<returnTypeC: #'StackPage *'>
	<inline: true>
	^self stackPageAt: index pages: pages! !

!CogStackPages methodsFor: 'page access' stamp: 'eem 3/5/2009 12:03'!
stackPageAt: index pages: thePages
	"Answer the page for a page index.
	 N.B.  This is a zero-relative index."
	<cmacro: '(index,pages) ((pages) + (index))'>
	<returnTypeC: #'StackPage *'> "for Slang..."
	^thePages at: index + 1! !

!CogStackPages methodsFor: 'page access' stamp: 'eem 12/14/2009 18:01'!
stackPageFor: pointer "<Integer>"
	<inline: true>
	<var: #pointer type: #'void *'>
	<returnTypeC: #'StackPage *'>
	^self stackPageAt: (self pageIndexFor: pointer)! !

!CogStackPages methodsFor: 'statistics' stamp: 'eem 6/15/2017 15:38'!
statAverageLivePagesWhenMapping
	<returnTypeC: #double>
	^statNumMaps = 0
		ifTrue: [0.0]
		ifFalse: [statPageCountWhenMappingSum asFloat / statNumMaps]! !

!CogStackPages methodsFor: 'statistics' stamp: 'eem 6/15/2017 15:38'!
statAverageLivePagesWhenMapping: aFloat
	<var: #aFloat type: #double>
	aFloat == 0.0
		ifTrue: [statPageCountWhenMappingSum := statNumMaps := 0]
		ifFalse: [coInterpreter primitiveFailFor: PrimErrBadArgument]! !

!CogStackPages methodsFor: 'statistics' stamp: 'eem 8/15/2015 01:14'!
statMaxPageCountWhenMapping
	<cmacro: '() GIV(statMaxPageCountWhenMapping)'>
	^statMaxPageCountWhenMapping! !

!CogStackPages methodsFor: 'statistics' stamp: 'eem 6/15/2017 14:17'!
statMaxPageCountWhenMapping: num
	statMaxPageCountWhenMapping := num! !

!CogStackPages methodsFor: 'debug printing' stamp: 'eem 8/14/2015 16:04'!
whereIsMaybeStackThing: anOop
	"If anOop is an address within the stack zone answer a string stating that, otherwise answer nil."
	<returnTypeC: 'char *'>
	self subclassResponsibility! !

CogStackPages subclass: #InterpreterStackPages
	instanceVariableNames: 'stackMemory indexOffset pageSizeInSlots'
	classVariableNames: ''
	poolDictionaries: 'VMBasicConstants'
	category: 'VMMaker-Interpreter'!

!InterpreterStackPages commentStamp: '<historical>' prior: 0!
I am a class that helps organize the StackInterpreter's collection of stack pages.  I hold the set of stack pages represented by InterpreterStackPage instances/StackPage structs.  The pages are held in a doubly-linked list that notionally has two heads:

mostRecentlyUsedPage-->used page<->used page<->used page<->used page<--leastRecentlyUsedPage
                                       ^                        <-next-prev->                         ^
                                        |                                                                       |
                                        v                        <-prev-next->                         v
                                        free page<->free page<->free page<->free page

In fact we don't need the least-recently-used page, and so it is only present conceptually.  The point is that there is a possibly empty but contiguous sequence of free pages starting at mostRecentlyUsedPage nextPage.  New pages are allocated preferentially from the free page next to the MRUP.
If there are no free pages then (effectively) the LRUP's frames are flushed to contexts and it is used instead.!

!InterpreterStackPages class methodsFor: 'translation' stamp: 'eem 8/14/2015 16:50'!
declareCVarsIn: aCCodeGenerator
	aCCodeGenerator
		var: #stackMemory type: 'char *';
		removeVariable: 'indexOffset'; "These are simulation/debugging things only"
		removeVariable: 'pageSizeInSlots' "These are simulation/debugging things only"! !

!InterpreterStackPages methodsFor: 'assertions' stamp: 'eem (const to msg rewrite) 11/6/2014 09:35'!
couldBeFramePointer: pointer
	"Answer if the argument is a properly aligned pointer into the stack zone."
	<var: #pointer type: #'void *'>
	^self
		cCode:
			[(pointer asUnsignedInteger bitAnd: objectMemory wordSize - 1) = 0
			   and: [pointer asUnsignedInteger
						between: stackMemory asUnsignedInteger
						and: pages asUnsignedInteger]]
		inSmalltalk:
			[(pointer  bitAnd: objectMemory wordSize - 1) = 0
			 and: [(self memIndexFor: pointer)
					between: 1 and: stackMemory size]]! !

!InterpreterStackPages methodsFor: 'initialization' stamp: 'eem 8/14/2015 11:39'!
extraStackBytes
	"See initializeStack:numSlots:pageSize:stackLimitOffset:stackPageHeadroom:
	``Because stack pages grow down...''"
	^self cCode: [objectMemory wordSize] inSmalltalk: [0]! !

!InterpreterStackPages methodsFor: 'initialization' stamp: 'eem 8/14/2015 15:56'!
initializeStack: theStackPages numSlots: stackSlots pageSize: slotsPerPage
	"Initialize the stack pages.  For testing I want stack addresses to be disjoint from
	 normal memory addresses so stack addresses are negative.  The first address is
	 -pageSize bytes.  So for example if there are 1024 bytes per page and 3 pages
	 then the pages are organized as

		byte address: -1024 <-> -2047 | -2048 <-> -3071 | -3072 <-> -4096 |
							page 3			page 2			page 1
		mem index:        769 <-> 513  |     512 <->  257  |   256 <->        1 |

	 The byte address is the external address corresponding to a real address in the VM.
	 mem index is the index in the memory Array holding the stack, an index internal to
	 the stack pages.  The first stack page allocated will be the last page in the array of pages
	 at the highest effective address.  Its base address be -1024  and grow down towards -2047."

	"The lFoo's are to get around the foo->variable scheme in the C call to allocStackPages below."
	<var: #theStackPages type: #'char *'>
	| numPages page structStackPageSize pageStructBase count |
	<var: #page type: #'StackPage *'>
	<var: #pageStructBase type: #'char *'>
	self cCode: ''
		inSmalltalk:
			[self assert: stackMemory size = stackSlots.
			 self assert: stackMemory == theStackPages].
	stackMemory := theStackPages. "For initialization in the C code."
	self cCode: '' inSmalltalk: [pageSizeInSlots := slotsPerPage].
	structStackPageSize := coInterpreter sizeof: CogStackPage.
	bytesPerPage := slotsPerPage * objectMemory wordSize.
	numPages := coInterpreter numStkPages.

	"Because stack pages grow down baseAddress is at the top of a stack page and so to avoid
	 subtracting BytesPerWord from baseAddress and lastAddress in the init loop below we simply
	 push the stackPage array up one word to avoid the overlap.  This word is extraStackBytes."
	pageStructBase := theStackPages + (numPages * bytesPerPage) + objectMemory wordSize.
	pages := self cCode: '(StackPage *)pageStructBase'
				  inSmalltalk:
						[pageStructBase class.
						 (1 to: numPages) collect: [:i| CogStackPage new]].

	"Simulation only.  Since addresses are negative the offset is positive.  To make all
	 stack addresses negative we make the offset a page more than it needs to be so the
	 address of the last slot in memory (the highest address in the stack, or its start) is
		- pageByteSize
	 and the address of the first slot (the lowest address, or its end) is
		- pageByteSize * (numPages + 1)"
	self cCode: '' inSmalltalk: [indexOffset := (numPages + 1) * slotsPerPage].
	"make sure there's enough headroom"
	self assert: coInterpreter stackPageByteSize - coInterpreter stackLimitBytes - coInterpreter stackLimitOffset
				>= coInterpreter stackPageHeadroom.
	0 to: numPages - 1 do:
		[:index|
		 page := self stackPageAt: index.
		 page
			lastAddress: (self cCode: '(char *)theStackPages + (index * GIV(bytesPerPage))'
							inSmalltalk: [(index * slotsPerPage - indexOffset) * objectMemory wordSize]);
			baseAddress: (page lastAddress + bytesPerPage);
			stackLimit: page baseAddress - coInterpreter stackLimitBytes;
			realStackLimit: page stackLimit;
			baseFP: 0;
			nextPage: (self stackPageAt: (index = (numPages - 1) ifTrue: [0] ifFalse: [index + 1]));
			prevPage: (self stackPageAt: (index = 0 ifTrue: [numPages - 1] ifFalse: [index - 1]))].
	self cCode: ''
		inSmalltalk:
			[| lowestAddress highestAddress |
			lowestAddress := (pages at: 1) lastAddress + objectMemory wordSize.
			highestAddress := (pages at: numPages) baseAddress.
			"see InterpreterStackPages>>longAt:"
			self assert: lowestAddress // objectMemory wordSize + indexOffset = 1.
			self assert: highestAddress // objectMemory wordSize + indexOffset = (numPages * slotsPerPage)].

	"The overflow limit is the amount of stack to retain when moving frames from an overflowing
	 stack to reduce thrashing.  See stackOverflowOrEvent:mayContextSwitch:"
	page := self stackPageAt: 0.
	overflowLimit := page baseAddress - page realStackLimit * 3 // 5.

	0 to: numPages - 1 do:
		[:index|
		 page := self stackPageAt: index.
		 self assert: (self pageIndexFor: page baseAddress) == index.
		 self assert: (self pageIndexFor: page baseAddress - (slotsPerPage - 1 * objectMemory wordSize)) == index.
		 self assert: (self stackPageFor: page baseAddress) == page.
		 self assert: (self stackPageFor: page stackLimit) == page.
		 self cCode: ''
			inSmalltalk:
				[| memIndex |
				 memIndex := index * slotsPerPage + 1. "this is memIndex in the block above"
				 self assert: (self memIndexFor: (self oopForPointer: page baseAddress))
							== (memIndex + slotsPerPage - 1).
				 index < (numPages - 1) ifTrue:
					[self assert: (self stackPageFor: page baseAddress + objectMemory wordSize) == (self stackPageAt: index + 1)]].
		coInterpreter initializePageTraceToInvalid: page].

	mostRecentlyUsedPage := self stackPageAt: 0.
	page := mostRecentlyUsedPage.
	count := 0.
	[| theIndex |
	 count := count + 1.
	 theIndex := self pageIndexFor: page baseAddress.
	 self assert: (self stackPageAt: theIndex) == page.
	 self assert: (self pageIndexFor: page baseAddress) == theIndex.
	 self assert: (self pageIndexFor: page stackLimit) == theIndex.
	 self assert: (self pageIndexFor: page lastAddress + objectMemory wordSize) == theIndex.
	 (page := page nextPage) ~= mostRecentlyUsedPage] whileTrue.
	self assert: count == numPages.
	self assert: self pageListIsWellFormed! !

!InterpreterStackPages methodsFor: 'initialization' stamp: 'eem 8/14/2015 15:48'!
initializeWithByteSize: byteSize "<Integer>" for: anInterpreter "<StackInterpreter>" "^<Array of: <Integer>"
	"Initialize the stackPages memory for simulation. Answer the base address of the memory."
	<doNotGenerate>
	coInterpreter := anInterpreter.
	objectMemory := anInterpreter objectMemory.
	^stackMemory := Array new: byteSize / objectMemory wordSize withAll: 0! !

!InterpreterStackPages methodsFor: 'memory access' stamp: 'eem 11/10/2014 15:45'!
longAt: byteAddress
	<doNotGenerate>
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	self assert: (byteAddress bitAnd: objectMemory wordSize - 1) == 0.
	^stackMemory at: byteAddress // objectMemory wordSize + indexOffset! !

!InterpreterStackPages methodsFor: 'memory access' stamp: 'eem 3/7/2015 17:43'!
longAt: byteAddress put: a32BitValue
	<doNotGenerate>
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	self assert: a32BitValue isInteger.
	self assert: (byteAddress bitAnd: objectMemory wordSize - 1) == 0.
	^stackMemory at: byteAddress // objectMemory wordSize + indexOffset put: a32BitValue! !

!InterpreterStackPages methodsFor: 'page access' stamp: 'eem 8/14/2015 16:45'!
memIndexFor: byteAddress
	"Map an address into the stack zone into a word index into the slots in the stack zone."
	<doNotGenerate>
	^(self oopForPointer: byteAddress) // objectMemory wordSize + indexOffset! !

!InterpreterStackPages methodsFor: 'page access' stamp: 'eem 3/8/2013 15:31'!
pageIndexFor: pointer "<Integer>"
	"Answer the page index for a pointer into stack memory, i.e. the index
	 for the page the address is in.  N.B.  This is a zero-relative index."
	<var: #pointer type: #'void *'>
	^self pageIndexFor: pointer stackMemory: stackMemory bytesPerPage: bytesPerPage! !

!InterpreterStackPages methodsFor: 'page access' stamp: 'eem 3/4/2009 09:01'!
pageIndexFor: pointer "<Integer>" stackMemory: stackmem "<Integer>" bytesPerPage: pageByteSize "<Integer>"
	"Answer the page index for a pointer into stack memory, i.e. the index
	 for the page the address is in.  N.B.  This is a zero-relative index."
	| memIndex |
	<cmacro: '(pointer,stackmem,pageByteSize) (((char *)(pointer) - (stackmem) - 1) / (pageByteSize))'>
	memIndex := self memIndexFor: pointer.
	self assert: (memIndex between: 1 and: stackMemory size).
	^memIndex - 1 // pageSizeInSlots! !

!InterpreterStackPages methodsFor: 'debug printing' stamp: 'eem 8/14/2015 16:04'!
whereIsMaybeStackThing: anOop
	"If anOop is an address within the stack zone answer a string stating that, otherwise answer nil."
	<returnTypeC: 'char *'>
	(self cCode:
			[self oop: anOop isGreaterThanOrEqualTo: stackMemory andLessThan: pages]
		 inSmalltalk:
			[(self memIndexFor: anOop) between: 1 and: stackMemory size]) ifTrue:
		[^' is in the stack zone'].
	^nil! !

!InterpreterStackPages methodsFor: 'memory access' stamp: 'bgf 3/15/2009 00:25'!
writeEnableMemory
	<doNotGenerate>
	stackMemory := stackMemory array! !

!InterpreterStackPages methodsFor: 'memory access' stamp: 'bgf 3/15/2009 00:24'!
writeProtectMemory
	<doNotGenerate>
	stackMemory := ReadOnlyArrayWrapper around: stackMemory! !

InterpreterStackPages subclass: #InterpreterStackPagesLSB
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!InterpreterStackPagesLSB commentStamp: '<historical>' prior: 0!
I am a class that helps simulate the StackInterpreter's stack organization on little-endian platforms.!

!InterpreterStackPagesLSB methodsFor: 'memory access' stamp: 'eem 6/20/2008 11:09'!
byteAt: byteAddress
	| lowBits long |
	lowBits := byteAddress bitAnd: 3.
	long := self longAt: byteAddress - lowBits.
	^(lowBits caseOf: {
		[0] -> [ long ].
		[1] -> [ long bitShift: -8  ].
		[2] -> [ long bitShift: -16 ].
		[3] -> [ long bitShift: -24 ]
	}) bitAnd: 16rFF! !

!InterpreterStackPagesLSB methodsFor: 'memory access' stamp: 'eem 2/27/2009 11:11'!
byteAt: byteAddress put: byte
	| lowBits long longAddress |
	lowBits := byteAddress bitAnd: 3.
	longAddress := byteAddress - lowBits.
	long := self longAt: longAddress.
	long := lowBits caseOf: {
		[0] -> [ (long bitAnd: 16rFFFFFF00) bitOr: byte ].
		[1] -> [ (long bitAnd: 16rFFFF00FF) bitOr: (byte bitShift: 8) ].
		[2] -> [ (long bitAnd: 16rFF00FFFF) bitOr: (byte bitShift: 16)  ].
		[3] -> [ (long bitAnd: 16r00FFFFFF) bitOr: (byte bitShift: 24)  ]
	}.

	self longAt: longAddress put: long.
	^byte! !

InterpreterStackPages subclass: #InterpreterStackPagesMSB
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!InterpreterStackPagesMSB commentStamp: '<historical>' prior: 0!
I am a class that helps simulate the StackInterpreter's stack organization on big-endian platforms.!

!InterpreterStackPagesMSB methodsFor: 'memory access' stamp: 'eem 9/10/2013 14:39'!
byteAt: byteAddress
	| lowBits bpwMinus1 |
	bpwMinus1 := StackInterpreter wordSize - 1.
	lowBits := byteAddress bitAnd: bpwMinus1.
	^((self longAt: byteAddress - lowBits)
		bitShift: (lowBits - bpwMinus1) * 8)
		bitAnd: 16rFF! !

CogClass subclass: #CogThreadManager
	instanceVariableNames: 'coInterpreter cogit threadLocalStorage processorOwner vmOwner threads numThreads numThreadsIncrement memoryIsScarce vmOSThread registerStates'
	classVariableNames: ''
	poolDictionaries: 'VMThreadingConstants'
	category: 'VMMaker-Multithreading'!

!CogThreadManager commentStamp: 'eem 10/28/2020 23:29' prior: 0!
The sole instance of this class manages native threads that cooperatively schedule the VM.  Although there may be any number of threads attempting to use the VM there is only one thread that is actually running VM code at any one time.  That thread owns the VM and other threads queue to acquire the VM so that they may execute VM code.  Each thread that can own and run the VM has a "vmThread" thread control block which contains the necessary state to orchestrate this shared execution.

Acquiring ownership

VM ownership is indicated by a single variable "vmOwner" holding the index of the thread that currently owns the thread, or 0 if the VM is unowned.  The owning thread may disown the VM at any time by setting vmOwner to zero, but running threads must contend to acquire the vmOwner by setting vmOwner to their thread index. vmOwner expected to be is set by compare-and-swap.

Ensuring Threads Exist

The VM has a heartbeat thread that is used for a number of things including periodically interrupting the VM to check for I/O, expired timers and so on.  This heartbeat thread is also essental to this threading system, where it is used as a watchdog.  When it wakes up it checks vmOwner and if zero then it either wakes up a thread in the queue, or if the queue is empty, spawns a new thread.  This thread will contend to become vmOwner, queueing itself if it fails.

Fast Call-out and Return

External calls can be made very quickly and simply using the watchdog to ensure that eventually there will be another thread to run the VM if a blocking call is made.  Immediately before the owning thread calls-out it remembers its index, which by definition is the value of vmOwner, in a local variable and sets vmOwner to zero.  If the call continues for long enough the watchdog will notice vmOwner is zero and wake up a thread to run the VM.  When the calling thread returns it attempts to reacquire the VM, using vmOwnerLock.  When it acquires vmOwnerLock it tests vmOwner and if zero simply reassigns its index to vmOwner and releases vmOwnerLock.  If vmOwner is non-zero then the thread adds itself to the queue of waiting threads, and will complete the call when the VM next becomess available.  To make a call-out that prevents other threads from running until after it returns the call-out machinery simply neglects to clear vmOwner.  A bit in the ExternalFunction's flags instance variable enables threading, set if the FFI pragma begins with the keyword threaded, e.g. <cdecl: threaded char* 'ffiPrintString' (char *)>.

If not already "bound", a call-out binds a Process to the thread making a call-out on its behalf for the dynamic extent of the call.  This means that subsequent call-outs from the process that are nested within the dynamc extent of the outer call occur on the same thread.  See the section below on Thread/Process Affinity and Binding.  Note that the process cannot be bound to a different thread.  If it were it would be running on that thread.  A thread may run any unbound process, but a bound process can only run on the thread to which it is bound.

A bound process that is having a call-out evaluated on its behalf is no longer strictly runnable.  The thread is running an external call but the process cannot be scheduled until the call-out has returned or called-back, and therefore it must not be left on the runnable process queues.  But if it is removed from the runnable process queues it must have a strong reference from somewhere otherwise it will be garbage collected.  Hence for the duration of the call-out->call-back or call-out->return execution during which the thread is execution externally a process is referenced strongly from the awolProcesses stack of the vmThread. And to be identifieable as running external code the process will have its myList field set to the ProcessInExternalCodeTag (a LinkedList), but for simplicity won't actually be on the list to avoid having the VM know how to remove a process from an arbitrary position in the list.  The awolProcesses set is a stack to allow a thread to make nested callouts on behalf of any number of unbound threads.

Call-back

A call-back must do two things; first it must acquire ownership, temporarily setting vmOwner to some illegal non-zero value such as -1, and then it must then search for its index.  Since it may be a thread as-yet unknown to the threading system it may not find its index and must allocate and initialize a vmThread and assign an index for itself.  Finally it can set vmOwner to its actual thread index, and proceed with the call-back.  The call-back machinery must note the state of vmOwner on entry and restore it on return to C to keep vmOwner consistent with its value at the start of a call-out.

Thread/Process Affinity and Binding

There must be some correspondence between Smalltalk processes and threads.  Some APIs, in particular the OpenGL and Windows debug APIs, require that all calls to the API are made from a single thread.  Any call into Smalltalk can reasonably expect to be called-back on the same thread.  The system therefore allows Smalltalk processes to be bound to specific threads and if a process is unbound when it makes a call-out, binds that Process to the thread making a call-out on its behalf for the dynamic extent of the call.  The system maintains a many-to-one mapping of processes to threads. A process indicates the thread to which it is bound using the fourth slot of Process which is threadId.

One can control the binding of processes to threads in advance of a call-out using thread affinity.  An affinity is simply the binding of a process to a specific thread that is created before any call-out and that will not be undone when any call-out returns.

A Cog VM flag (the least significant bit of Smalltalk vmParameterAt: 48) records whether the current image's Process class defines the fourth slot.  If the bit is not set the thread system is inoperative.  If set, the VM knows it can use the fourth slot of a Process for the process's threadId.  If a Process's threadId is nil then any thread can execute that process.  If the threadId is an integer then its least significant bit is zero for a temporary binding but 1 for a permanent affinity, and the value shifted down 1 position to remove the lsb is the threadId of the thread to which it is bound or affined.

Scheduling

In effect a bound thread is free to run an unbound or unaffined process until it attempts a call-out.  An unbound thread is free to run any unbound process, but a bound or affined process can only run on the thread to which it is bound or affined.  A primitive is provided to set or clear (but not change) a Process's affinity.  We use a primitive because it can be atomic [Open to question: , and can maintain the invariant that a bound process is in the boundProcesses set of its thread].

When the current thread makes a process switch to another Smalltalk process it checks the threadId of the highest-priority runnable process.  If the threadId is nil or matches the current thread then that thread continues, running the process.  If not, it attempts to wake the process's thread and adds itself to the queue of waiting threads.  If the thread is already executing external code then the VM blocks; i.e. the VM will not preempt a runnable process because that process is bound to an unavailable thread.

To avoid the situation where the thread to be awoken is unavailable because it is running on behalf of some other thread, when the current thread considers making a process switch to another Smalltalk process and that process is bound, the thread first checks if it is bound to the current process and if so simply disowns the VM and adds itself to the queue of waiting threads.  The watchdog will notice the VM is unowned and make sure some other thread is available to run the VM.  The watchdog therefore has to ensure that there exists at least one unbound thread in the queue.

Implementation

The effect of vmLock is to ensure that the only manipulation of the core thread data structures, the set of vmThreads et al, is made by the thread that owns the VM.  In particular the watchdog temporarily locks the VM before waking a thread on the queue and so no thread touches the queue unless it owns the VM.  Since the core data structures are only manipulated by the owning thread their use is entirely lock-free (apart from vmLock itself) and we can simply use a realloc(3)'ed array of pointers to vmThreads, which in the simulator are instances of CogVMThread and in the actual VM CogVMThread structs derived there-from.  We use an array of pointers so that realloc'ing the array doesn't affect individual vmThread structs, reducing the need for locking.  This also allows us to represent the awolProcesses stack as slots on the end of a CogVMThread struct, reallocing the struct as necessary when pushing processes on the stack.

To save time locating the vmThread for a thread the VM stores a thread's index in thread-local storage.  However, vmLock holds the index of the owning thread, so the thread-local storage mechanism need be used only for other than the owning thread.

vmOwnerLock is locked and unlocked in a processor-specific manner.  The Cogit generates two functions ceTryLockVMOwner and ceUnlockVMOwner that lock and unlock vmOwnerLock using processor-specific mechanisms, which on x86 is the XCHG locked-swap instruction.

There are abstractions for the OS threading system defined in platforms/<plat>/vm/sqPlatformSpecific.h and a small API of helper functions defined in platforms/Cross/vm/sq.h to insulate the VM from platform specifics.  Look for the COGMTVM define.

sqOSThread is the type of the platform's thread handle, e.g. pthread_t on pthreads and HANDLE on win32.
sqOSSemaphore is the type of the platform's queueing semaphore upon which a thread can wait, used to implement threads waiting in the queue.

void ioInitThreadLocalThreadIndices(void)
	initializes the thread-local storage mechanism, enabling the following two functions.
void ioSetThreadLocalThreadIndex(long)
	sets the calling thread's index to the argument.
long ioGetThreadLocalThreadIndex(void)
	answers the calling thread's index, answering zero in a thread that has not yet set its index.

int ioNewOSThread(void (*startFunction)(int), int)
	spawns a new thread that starts by evaluating func with the supplied argument (which
	will be its thread index).  Answers 0 on success and an error code on failure.
void ioExitOSThread(sqOSThread thread)
	terminates the thread argument and releases any resources associated with it maintained by this subsystem.
void ioReleaseOSThreadState(sqOSThread thread)
	releases any resources associated with the thread maintained by this subsystem.
int ioNewOSSemaphore(sqOSSemaphore *)
	Assigns through the argument a new OS semaphore on which a sqOSThread
	may wait.  Answers 0 on success and an error code on failure.  The semaphore
	functions take pointers because in e.g. the pthreads implementation an OSSemaphore
	is a structure, not a pointer, and teh structure cannot be copied and still function correctly.
void ioSignalOSSemaphore(sqOSSemaphore *)
	signals the argument's referent, allowing the thread waiting on it to proceed.
void ioWaitOnOSSemaphore(sqOSSemaphore *)
	blocks the calling thread until the argument's referent is signalled.
sqOSThread ioCurrentOSThread(void)
	answers the current thread.
int ioOSThreadsEqual(sqOSThread,sqOSThread)
	answers non-zero if the two arguments are the same thread.
int ioOSThreadIsAlive(sqOSThread)
	answers non-zero if the argument is still running (has not crashed or terminated).


The above API is used to implement the following functions used by the VM to share itself amongst threads.

void startThreadSubsystem(void)
	initializes the threading subsystem, aborting if there is an error.
sqInt disownVM(void)
	disownVM clears vmOwner, answering its previous value which is the index of the disowning thread, marked with a flag indicating if the activeProcess was newly affined to the thread in disownVM.  Sets the activeProcess's affinity and sets its list to the ProcessInExternalCodeTag.  Pushes the activeProcess on the vmThread's awolProcesses stack.
void ownVM(sqInt)
	ownVM gains ownership of the VM, blocking until possible.  The argument is the index of the owning thread, or zero if not known, combined with the newly-affined flag.  Pops the activeProcess from the vmThread's awolProcesses stack, clears its thread affinity if appropriate, and puts the old activeProcess to sleep.

To initialize a thread, foreign or otherwise, to use the VM the thread must first evaluate
	startVMThread(CogVMThread *)
This will either be evaluated as the startFunction of ioNewOSThread or by a thread making a callback that finds it is unknown to the VM (its ioGetThreadLocalThreadIndex() is zero).


Simulation only variables:
	coInterpreter		CoInterpreterMT
	cogit				the jit
	threadLocalStorage	Dictionary of Process -> Array
	processorOwner		Process
Implemetation variables:
	vmOwner			Integer the index of the owning VM thread.  If 0, the VM is unowned
	threads				Array of pointer to CogVMThread the set of thread control blocks
	numThreads			size of threads
	numThreadsIncrement	increment by which to increase threads
	memoryIsScarce		boolean
	vmOSThread		sqOSThread the OS thread that most recently owned the VM.  This is for signalSemaphoreWithIndex!

!CogThreadManager class methodsFor: 'accessing' stamp: 'eem 8/14/2009 14:58'!
awolProcessesIncrement
	^AWOLProcessesIncrement! !

!CogThreadManager class methodsFor: 'translation' stamp: 'eem 11/19/2020 19:46'!
declareCVarsIn: cCodeGen
	cCodeGen
		removeVariable: 'coInterpreter';
		removeVariable: 'cogit';
		removeVariable: 'threadLocalStorage';
		removeVariable: 'processorOwner';
		removeVariable: 'registerStates'.
	cCodeGen
		var: #threads type: #'CogVMThread **';
		var: #vmOSThread type: #sqOSThread! !

!CogThreadManager class methodsFor: 'class initialization' stamp: 'eem 8/14/2009 14:57'!
initialize
	"CogThreadManager initialize"
	CTMInitializing := 1.
	CTMUnavailable := 2.	"off doing its own thing and not available to run the VM, e.g. calling-out."
	CTMAssignableOrInVM := 3.	"either owning the VM or blocked on its osSemaphore available for VM work"
	CTMWantingOwnership := 4.	"with something specific to do in the VM (e.g. return a result from a call-out, make a call-back)"

	"Define the size of the stack of processes at time of disown/own."
	AWOLProcessesIncrement := 4! !

!CogThreadManager methodsFor: 'public api' stamp: 'eem 11/24/2020 23:00'!
acquireVMFor: threadIndex
	"Attempt to acquire the VM, eventually blocking until it becomes available.
	 Spin until the maxWaitingPriority has been updated if it is lower than this thread's priority."
	<returnTypeC: #'CogVMThread *'>
	| vmThread |
	<var: #vmThread type: #'CogVMThread *'>
	self assert: threadIndex = self ioGetThreadLocalThreadIndex.
	vmThread := self vmThreadAt: threadIndex.
	self assert: (vmThread state = CTMUnavailable
				or: [vmThread state = CTMWantingOwnership]).
	(self tryLockVMOwnerTo: threadIndex) ifFalse:
		[vmThread state: CTMWantingOwnership.
		 [(self vmOwnerIs: threadIndex)
		  or: [self tryLockVMOwnerTo: threadIndex]] whileFalse:
			[vmThread priority ifNotNil:
				[coInterpreter waitingPriorityIsAtLeast: vmThread priority].
			 (self vmOwnerIs: threadIndex) ifFalse:
				[self ioWaitOnOSSemaphore: (self addressOf: vmThread osSemaphore)]]].
	coInterpreter assertProcessorStackPointersBelongToCurrentThread.
	vmOSThread := vmThread osThread.
	vmThread state: CTMAssignableOrInVM.
	^vmThread! !

!CogThreadManager methodsFor: 'public api' stamp: 'eem 10/30/2020 16:14'!
currentVMThread
	^self vmThreadAt: self getVMOwner! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 11/22/2020 21:38'!
ensureInitializedProcessor: aProcessor forThreadIndex: threadIndex
	"Ensure aProcessor has stack pointers within its defined range, initializing it if not."
	<doNotGenerate>
	| range |
	range := coInterpreter cStackRangeForThreadIndex: threadIndex.
	self assert: (range includes: aProcessor fp) = (range includes: aProcessor sp).
	((range includes: aProcessor fp) and: [range includes: aProcessor sp]) ifFalse:
		[aProcessor smashCallerSavedRegistersWithValuesFrom: 16r90000000 by: coInterpreter objectMemory wordSize / 2.
		 cogit initializeProcessorStack: range last]! !

!CogThreadManager methodsFor: 'public api' stamp: 'eem 11/19/2020 19:41'!
ensureRunningVMThread: vmIsRelinquishing
	"Called from checkVMOwnershipFromHeartbeat if the VM is unowned.
	 Hence we are in the heartbeat thread.  The race is against that thread
	 owning the VM and against foreign callbacks."
	<returnTypeC: #void>
	<var: #vmThread type: #'CogVMThread *'>
	self willingVMThread ifNotNil:
		[:vmThread|
		 "If the VM is relinquishing the processor then only schedule a thread if it has work to do."
		 (vmIsRelinquishing
		  and: [vmThread state ~= CTMWantingOwnership]) ifTrue:
			[^self].
		 (self tryLockVMOwnerTo: vmThread index) ifFalse: "someone beat us to it..."
			[^self].
		 vmOSThread := vmThread osThread.
		 "release the thread from its blocking loop"
		 self ioSignalOSSemaphore: (self addressOf: vmThread osSemaphore).
		 self ioTransferTimeslice.
		 "self cCode: [coInterpreter print: 'ERVT signalled '; printNum: vmThread index; cr]."
		 ^self].

	"If the VM is relinquishing the processor then only schedule a thread if it has work to do
	 (willingVMThread not nil above).
	 If we have failed to allocate thread storage before there is no point continuing to
	 try to do so.  By this time we should have quite a few threads in the pool."
	(vmIsRelinquishing or: [memoryIsScarce]) ifTrue:
		[^self].
	self unusedThreadInfo ifNotNil:
		[:vmThread|
		 (self tryLockVMOwnerTo: vmThread index) ifTrue:
			[(self startThreadForThreadInfo: vmThread) ifFalse:
				[self releaseVM]]]! !

!CogThreadManager methodsFor: 'public api' stamp: 'eem 8/14/2009 16:36'!
getNumThreads
	<cmacro: '() GIV(numThreads)'>
	^numThreads! !

!CogThreadManager methodsFor: 'public api-vm platform code' stamp: 'eem 11/4/2010 17:45'!
getVMOSThread
	"This is for signalSemaphoreWithIndex in Cross/vm/sqExternalSemaphores.c"
	<api>
	<returnTypeC: #sqOSThread>
	^vmOSThread! !

!CogThreadManager methodsFor: 'public api' stamp: 'eem 7/27/2010 20:07'!
getVMOwner
	self sqLowLevelMFence.
	^vmOwner! !

!CogThreadManager methodsFor: 'thread set' stamp: 'eem 8/26/2010 18:15'!
growThreadInfosToAtLeast: index
	"Grow the thread infos to at least index in numThreadsIncrement quanta."
	| newThreads newNumThreads |
	<var: #newThreads type: #'CogVMThread **'>
	<inline: false>
	memoryIsScarce ifTrue:
		[^false].
	newNumThreads := index + numThreadsIncrement - 1 // numThreadsIncrement * numThreadsIncrement.
	newNumThreads >= self maxNumThreads ifTrue:
		[^false].
	"Since 0 is not a valid index we allocate one extra CogVMThread and use 1-relative indices."
	newThreads := self cCode: 'realloc(GIV(threads), (newNumThreads + 1) * sizeof(CogVMThread *))'
						inSmalltalk: [(Array new: newNumThreads)
										replaceFrom: 1 to: numThreads
										with: threads startingAt: 1].
	(newThreads notNil
	 and: [self populate: newThreads from: numThreads + 1 to: newNumThreads]) ifFalse:
		[self free: newThreads.
		 memoryIsScarce := true.
		 ^false].
	threads := newThreads.
	numThreads := newNumThreads.
	^true! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 10/24/2020 18:50'!
guiProcess
	"Simulation only; answer the simulator's first process."
	<doNotGenerate>
	^threads ifNotNil: [threads first osThread]! !

!CogThreadManager methodsFor: 'public api' stamp: 'eem 8/27/2010 14:51'!
highestPriorityThreadIfHigherThan: activePriority expectedMax: maxPriority
	"Answer the first vmThread waiting to acquire the VM that is of higher priority
	 than activePriority, or answer nil if none. If there is a higher priority thread
	 then set the coInterpreter's maxWaitingPriority to either the priority of the
	 next highest priority vmThread, or to 0 if none is waiting."
	<returnTypeC: #'CogVMThread *'>
	| vmThread highest nextHighest |
	<var: #vmThread type: #'CogVMThread *'>
	<var: #highest type: #'CogVMThread *'>
	<var: #nextHighest type: #'CogVMThread *'>
	highest := nextHighest := nil.
	"To make this fair we could remember the last index at which we
	 found the highest and start the search at the following index."
	1 to: numThreads do:
		[:i|
		vmThread := threads at: i.
		vmThread state = CTMWantingOwnership ifTrue:
			[self assert: vmThread priority <= maxPriority.
			 highest isNil
				ifTrue: [highest := vmThread]
				ifFalse:
					[vmThread priority > highest priority
						ifTrue:
							[nextHighest := highest.
							highest := vmThread]
						ifFalse:
							[nextHighest isNil
								ifTrue: [nextHighest := vmThread]
								ifFalse: [vmThread priority > nextHighest priority ifTrue:
											[nextHighest := vmThread]]]]]].
	highest isNil ifTrue: [^nil].

	highest priority <= activePriority ifTrue:
		[^nil].
	coInterpreter setMaxWaitingPriorityTo: (nextHighest isNil
											ifTrue: [0]
											ifFalse: [nextHighest priority]).
	^highest! !

!CogThreadManager methodsFor: 'initialize-release' stamp: 'eem 11/19/2020 19:46'!
initialize
	vmOwner := numThreads := numThreadsIncrement := 0.
	memoryIsScarce := false.
	"N.B.  Do not initialize threadLocalStorage; leave this to ioInitThreadLocalThreadIndices".
	registerStates := IdentityDictionary new! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 7/16/2009 14:45'!
ioCurrentOSThread
	<doNotGenerate>
	"See platforms/<plat>/vm/sqPlatformSpecific.h for the real definition."
	^Processor activeProcess! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 8/26/2010 18:21'!
ioDestroyOSSemaphore: semaphorePointer "<BlockClosure>" 
	<doNotGenerate>
	"See platforms/Cross/vm/sq.h for the real definition."
	^self! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 10/30/2020 09:49'!
ioExitOSThread: anOSThread
	<doNotGenerate>
	"See platforms/Cross/vm/sq.h for the real definition."
	| active |
	active := Processor activeProcess == anOSThread.
	self halt.
	anOSThread terminate.
	self ioReleaseOSThreadState: anOSThread! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 10/19/2010 15:11'!
ioGetThreadLocalThreadIndex
	<doNotGenerate>
	"Answer the thread index of the current thread/process or 0 if it is not yet defined.
	 See platforms/<plat>/vm/sqPlatformSpecific.h for the actual definition."
	^threadLocalStorage
		ifNil: [1] "pre initialization"
		ifNotNil: [threadLocalStorage at: Processor activeProcess]! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 8/27/2010 08:59'!
ioMilliSleep: milliseconds
	"Pause the calling thread/process for milliseconds.
	 See platforms/<plat>/vm/sqPlatformSpecific.h for the actual definition."
	<doNotGenerate>
	(Delay forMilliseconds: milliseconds) wait! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 9/18/2009 15:54'!
ioNewOS: function Thread: vmThread
	"Fork a new thread to evaluate function with the supplied argument.
	 Return 0 on success and non-zero on error.  The only errors equate
	 to running out of resources, ie. memory.
	 See platforms/Cross/vm/sq.h for the real definition."
	 
	<doNotGenerate>
	[self perform: function with: vmThread] fork.
	^0! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 11/10/2020 11:02'!
ioNewOSSemaphore: semaphorePointer "<BlockClosure>" 
	<doNotGenerate>
	"See platforms/Cross/vm/sq.h for the real definition."
	semaphorePointer at: 0 put: Semaphore new.
	^0! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 7/17/2009 11:33'!
ioNumProcessors
	<doNotGenerate>
	"See platforms/<plat>/vm/sq<Plat>Threads.c for the real definition."
	^3 + 4 // 2! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 7/15/2009 15:26'!
ioOSThreads: aThread Equal: bThread
	<doNotGenerate>
	"See platforms/<plat>/vm/sqPlatformSpecific.h for the real definition."
	^aThread = bThread! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 7/17/2009 10:35'!
ioReleaseThreadState: anOSThread
	<doNotGenerate>
	"See platforms/Cross/vm/sq.h for the real definition."
	threadLocalStorage removeKey: anOSThread ifAbsent: []! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 10/19/2010 15:19'!
ioSetThreadLocalThreadIndex: index
	<doNotGenerate>
	"See platforms/<plat>/vm/sqPlatformSpecific.h for the real definition."
	threadLocalStorage ifNil:
		[threadLocalStorage := WeakIdentityKeyDictionary new].
	threadLocalStorage at: Processor activeProcess put: index! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 7/17/2009 18:57'!
ioSignalOSSemaphore: aSemaphorePtr
	<var: #anOSSemaphore type: #'sqOSSemaphore *'>
	<returnTypeC: #void>
	<doNotGenerate>
	"See platforms/Cross/vm/sq.h for the real definition."
	aSemaphorePtr value signal! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 7/17/2009 10:35'!
ioTransferTimeslice
	<doNotGenerate>
	"See platforms/Cross/vm/sq.h for the real definition."
	Processor yield! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 11/28/2020 21:54'!
ioWaitOnOSSemaphore: aSemaphorePtr
	<var: #anOSSemaphore type: #'sqOSSemaphore *'>
	<returnTypeC: #void>
	<doNotGenerate>
	"See platforms/Cross/vm/sq.h for the real definition."
	"Simulate the VM's heartbeat by calling checkVMOwnershipFromHeartbeat
	 if the wait times-out."
	[aSemaphorePtr value waitTimeoutMSecs: 1000] whileTrue:
		[coInterpreter checkVMOwnershipFromHeartbeat].
	self deny: vmOwner = 0.
	cogit processor setRegisterState: (registerStates at: vmOwner
										ifAbsentPut:
											[self ensureInitializedProcessor: cogit processor forThreadIndex: vmOwner.
											 cogit processor registerState])! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 7/27/2010 10:02'!
maxNumThreads
	^self cCode: [ThreadIdMask] inSmalltalk: [16]! !

!CogThreadManager methodsFor: 'public api' stamp: 'eem 8/14/2009 15:34'!
popAWOLProcess: vmThread
	<var: #vmThread type: #'CogVMThread *'>
	self assert: vmThread awolProcIndex > 0.
	^vmThread awolProcesses at: (vmThread awolProcIndex: vmThread awolProcIndex - 1)! !

!CogThreadManager methodsFor: 'thread set' stamp: 'eem 11/10/2020 10:58'!
populate: vmThreadPointers from: start to: finish
	"Populate vmThreadPointers with vmThreads over the given range."
	<var: #vmThreadPointers type: #'CogVMThread **'>
	| nThreads vmThreads |
	<var: #vmThreads type: #'CogVMThread *'>
	<var: #vmThread type: #'CogVMThread *'>
	<inline: true>
	nThreads := finish - start + 1.
	vmThreads := self cCode: [self calloc: nThreads _: (self sizeof: CogVMThread)]
						inSmalltalk: [CArrayAccessor on: ((1 to: nThreads) collect: [:ign| CogVMThread new])].
	vmThreads ifNil:
		[^false].
	"Since 0 is not a valid index, in C we allocate one extra CogVMThread and use 1-relative indices."
	self cCode: [start = 1 ifTrue: [vmThreadPointers at: 0 put: nil]]
		inSmalltalk: [].
	start to: finish do:
		[:i| | vmThread |
		vmThread := self addressOf: (vmThreads at: i - start).
		(self ioNewOSSemaphore: (self addressOf: vmThread osSemaphore put: [:sem| vmThread osSemaphore: sem])) ~= 0 ifTrue:
			[start to: i - 1 do:
				[:j|
				vmThread := self addressOf: (vmThreads at: j - start).
				self ioDestroyOSSemaphore: (self addressOf: vmThread osSemaphore)].
			self free: vmThreads.
			^false].
		vmThreadPointers at: i put: vmThread.
		vmThread awolProcLength: AWOLProcessesIncrement.
		vmThread index: i].
	^true! !

!CogThreadManager methodsFor: 'public api' stamp: 'eem 8/18/2009 10:27'!
pushAWOLProcess: awolProcess on: vmThread
	<var: #vmThread type: #'CogVMThread *'>
	<returnTypeC: #'CogVMThread *'>
	| cvt |
	<var: #cvt type: #'CogVMThread *'>
	cvt := vmThread.
	self assert: (cvt awolProcIndex between: 0 and: cvt awolProcLength).
	cvt awolProcIndex >= cvt awolProcLength ifTrue:
		["The realloc doesn't look like it grows but it does so by AWOLProcessesIncrement
		  entries because sizeof(CogVMThread) includes room for that many entries."
		 cvt := self cCode: 'realloc(cvt,sizeof(CogVMThread) + (sizeof(sqInt) * cvt->awolProcLength))'
					inSmalltalk: [cvt growAWOLProcesses].
		 threads at: vmThread index put: cvt.
		 cvt awolProcLength: cvt awolProcLength + AWOLProcessesIncrement].
	cvt awolProcesses at: cvt awolProcIndex put: awolProcess.
	cvt awolProcIndex: cvt awolProcIndex + 1.
	^cvt! !

!CogThreadManager methodsFor: 'scheduling' stamp: 'eem 8/30/2010 09:53'!
registerVMThread: vmThread
	"Register the current thread in the set of threads, initializing the supplied
	 vmThread and setting the thread's thread block index there-to."
	<var: #vmThread type: #'CogVMThread *'>
	<returnTypeC: #'CogVMThread *'>
	self assert: (vmThread state = CTMInitializing
				or: [vmThread state = CTMWantingOwnership]).
	vmThread osThread: self ioCurrentOSThread.
	self ioSetThreadLocalThreadIndex: vmThread index.
	self assert: self ioGetThreadLocalThreadIndex = vmThread index.
	^vmThread! !

!CogThreadManager methodsFor: 'public api' stamp: 'eem 11/22/2020 21:26'!
releaseVM
	"In the simulation this is where register state is saved; it is switched in tryLockVMOwnerTo:."
	self cCode: [] inSmalltalk:
		[vmOwner ~= 0 ifTrue:
			[registerStates at: vmOwner put: cogit processor registerState]].
	self setVMOwner: 0! !

!CogThreadManager methodsFor: 'initialize-release' stamp: 'eem 7/10/2009 11:48'!
setInterpreter: aCoInterpreter cogit: aCogit
	coInterpreter := aCoInterpreter.
	cogit := aCogit! !

!CogThreadManager methodsFor: 'public api' stamp: 'eem 11/25/2020 19:34'!
setVMOwner: indexOrZero
	"An ugly accessor used in only three cases:
	 1.	by ownVMFromUnidentifiedThread when the VM is first locked to the thread id
		of the unidentified thread, and then, once identified, to the thread's index.
	 2.	by wakeVMThreadFor: used by the two-level scheduler to switch threads when
		a Smalltalk process switch occurs to a process affined to another thread.
	 3. to release the VM (set the owner to zero)"
	<inline: #always>
	vmOwner := indexOrZero.
	self sqLowLevelMFence! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 11/20/2020 07:30'!
shutdownModule
	<doNotGenerate>
	| guiProcess activeProcess |
	threads ifNil: [^self].
	(guiProcess := self guiProcess) ~= (activeProcess := Processor activeProcess) ifTrue:
		[guiProcess
			signalException:
				(Notification new tag: #evaluateQuit; yourself).
		Project current spawnNewProcessIfThisIsUI: activeProcess.
		activeProcess terminate].
	threads do:
		[:ea|
		ea osThread ifNotNil:
			[:aProcess|
			(aProcess ~~ activeProcess and: [aProcess ~~ guiProcess]) ifTrue:
				[aProcess terminate]]].
	threads := nil! !

!CogThreadManager methodsFor: 'scheduling' stamp: 'eem 7/23/2010 14:44'!
startThreadForThreadIndex: index
	index > numThreads ifTrue:
		[(self growThreadInfosToAtLeast: index) ifFalse:
			[^false]].
	^self startThreadForThreadInfo: (self vmThreadAt: index)! !

!CogThreadManager methodsFor: 'scheduling' stamp: 'eem 12/12/2012 11:36'!
startThreadForThreadInfo: vmThread
	<var: #vmThread type: #'CogVMThread *'>
	<inline: false>
	self assert: vmThread state isNil.
	vmThread state: CTMInitializing.
	"self cCode: ''
		inSmalltalk: [coInterpreter transcript
						cr;
						nextPutAll: 'starting VM thread ';
						print: vmThread index;
						flush.
					 (thisContext home stackOfSize: 10) do:
						[:ctxt|
						coInterpreter transcript cr; print: ctxt; flush]]."
	(self ioNewOS: #startVMThread: Thread: vmThread) = 0 ifTrue:
		[self ioTransferTimeslice.
		^true].
	memoryIsScarce := true.
	"self cCode: [coInterpreter print: 'ERVT failed to spawn so memory is scarce'; cr]"
	^false! !

!CogThreadManager methodsFor: 'public api' stamp: 'eem 11/25/2020 19:03'!
startThreadSubsystem
	"Initialize the threading subsystem, aborting if there is an error."
	| vmThread |
	<inline: false>
	self assert: threads = nil.
	vmOSThread := self ioCurrentOSThread.
	numThreadsIncrement := (self ioNumProcessors max: 2) min: 16.
	(self growThreadInfosToAtLeast: numThreadsIncrement * 2) ifFalse:
		[self error: 'no memory to start thread system'].
	vmThread := threads at: (vmOwner := 1).
	vmThread state: CTMInitializing.
	self registerVMThread: vmThread.
	vmThread state: CTMAssignableOrInVM! !

!CogThreadManager methodsFor: 'scheduling' stamp: 'eem 7/22/2010 20:38'!
startVMThread: vmThread
	"Start a VM thread that will attempt to acquire the VM and proceed
	 to run the VM, taking processes from the runnable process queue."
	<returnTypeC: #void>
	<var: #vmThread type: #'CogVMThread *'>
	self registerVMThread: vmThread.
	vmThread state: CTMAssignableOrInVM.
	coInterpreter threadSchedulingLoop: vmThread! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 11/20/2020 13:44'!
threadsDo: aBlock
	1 to: numThreads do:
		[:i|
		(threads at: i) ifNotNil:
			[:thread| aBlock value: thread]]! !

!CogThreadManager methodsFor: 'simulation' stamp: 'eem 1/3/2021 13:08'!
tryLockVMOwnerTo: threadIndex
	"In the real VM this is a direct call of Cogit>>#tryLockVMOwnerTo:/ceTryLockVMOwner.
	 In the simulation this is where register state is saved and switched, simulaitng a thread switch.
	 releaseVM also saves register state.  The code here and in registerState allow us to avoid the
	 expensive and complex MultiProcessor hack."
	<doNotGenerate>
	| result |
	self deny: threadIndex = 0.
	cogit withProcessorHaltedDo:
		[| prior processor  |
		processor := cogit processor.
		prior := processor registerState.
		"A thread switch would (have) occur(ed) if it were that the VM were owned other than by threadIndex"
		vmOwner ~= threadIndex ifTrue:
			[vmOwner ~= 0 ifTrue:
				[registerStates at: vmOwner put: prior].
			 processor setRegisterState: (registerStates
											at: threadIndex
											ifAbsentPut:
												[self ensureInitializedProcessor: processor forThreadIndex: threadIndex.
												 processor registerState])].
		result := cogit tryLockVMOwnerTo: threadIndex.
		self assert: result = (threadIndex = vmOwner).
		registerStates at: threadIndex put: processor registerState.
		threadIndex ~= vmOwner ifTrue: "the lock attempt failed; undo the (processor) thread switch."
			[processor setRegisterState: prior]].
	^result! !

!CogThreadManager methodsFor: 'scheduling' stamp: 'eem 8/30/2010 10:01'!
unregisterVMThread: vmThread
	"Unegister the current thread in the set of threads."
	<var: #vmThread type: #'CogVMThread *'>
	self assert: (vmThread state ~= nil
				and: [vmThread state ~= CTMInitializing]).
	vmThread
		state: nil;
		osThread: nil.
	self ioSetThreadLocalThreadIndex: 0! !

!CogThreadManager methodsFor: 'thread set' stamp: 'eem 7/23/2010 14:22'!
unusedThreadInfo
	"Answer a pointer to an unused CogVMThread, growing the sequence if necessary."
	<returnTypeC: #'CogVMThread *'>
	| vmThread index |
	<var: #vmThread type: #'CogVMThread *'>
	1 to: numThreads do:
		[:i|
		vmThread := threads at: i.
		self assert: vmThread index = i.
		vmThread state isNil ifTrue:
			[^vmThread]].
	index := numThreads + 1.
	(self growThreadInfosToAtLeast: numThreads + numThreadsIncrement) ifFalse:
		[^nil].
	^threads at: index! !

!CogThreadManager methodsFor: 'public api-testing' stamp: 'eem 11/9/2020 18:38'!
vmIsOwned
	"Answer if the vm is owned"
	<inline: true>
	self sqLowLevelMFence.
	^vmOwner > 0! !

!CogThreadManager methodsFor: 'public api-testing' stamp: 'eem 10/30/2020 13:52'!
vmIsUnowned
	^self vmOwnerIs: 0! !

!CogThreadManager methodsFor: 'Cogit lock implementation' stamp: 'eem 10/28/2020 23:27'!
vmOwnerAddress
	<api> "NB. For the JIT only, so it can generate the lock & unlock functions."
	<returnTypeC: #usqInt>
	^self
		cCode: [(self addressOf: vmOwner) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #vmOwnerFromMachineCode in: self]! !

!CogThreadManager methodsFor: 'Cogit lock implementation' stamp: 'eem 10/29/2020 17:26'!
vmOwnerFromMachineCode
	<doNotGenerate>
	^vmOwner! !

!CogThreadManager methodsFor: 'Cogit lock implementation' stamp: 'eem 10/29/2020 17:26'!
vmOwnerFromMachineCode: aValue
	<doNotGenerate>
	self assert: (aValue between: 0 and: numThreads).
	vmOwner := aValue! !

!CogThreadManager methodsFor: 'public api-testing' stamp: 'eem 11/9/2020 18:39'!
vmOwnerIs: index
	"Test if the vmOwner is index."
	<inline: true>
	self sqLowLevelMFence.
	^vmOwner = index! !

!CogThreadManager methodsFor: 'public api' stamp: 'eem 10/30/2020 16:15'!
vmThreadAt: index
	self assert: (index between: 0 and: numThreads).
	^index > 0 ifTrue: [threads at: index]! !

!CogThreadManager methodsFor: 'public api' stamp: 'eem 10/30/2020 15:55'!
waitForWork: vmThread
	"Wait for work."
	<var: #vmThread type: #'CogVMThread *'>
	<returnTypeC: #void>
	self assert: vmThread state = CTMAssignableOrInVM.
	self deny: (self vmOwnerIs: vmThread index).
	self ioWaitOnOSSemaphore: (self addressOf: vmThread osSemaphore)! !

!CogThreadManager methodsFor: 'public api' stamp: 'eem 10/30/2020 16:16'!
wakeVMThreadFor: index
	"Transfer the VM to the thread with index.  Called from a thread that finds the
	 highest priority runnable process is bound to the thread with index index."
	<returnTypeC: #void>
	| vmThread |
	self assert: (self vmIsOwned and: [(self vmOwnerIs: index) not]).
	self assert: (index between: 1 and: numThreads).
	self setVMOwner: index.
	vmThread := threads at: index.
	vmThread state
		ifNil: [self startThreadForThreadInfo: vmThread]
		ifNotNil:
			[self assert: vmThread state = CTMWantingOwnership.
			 self ioSignalOSSemaphore: (self addressOf: vmThread osSemaphore)].
	self ioTransferTimeslice! !

!CogThreadManager methodsFor: 'thread set' stamp: 'eem 10/30/2020 14:48'!
willingVMThread
	"Answer a pointer to a live CogVMThread in any of the ``will do VM work''
	 states (other than the current owner if the VM is owned), or nil if none.
	 Preferentially answer threads wanting ownership."
	<returnTypeC: #'CogVMThread *'>
	| thread threadWantingVM threadWilling |
	<inline: false>
	threadWantingVM := threadWilling := nil.
	1 to: numThreads do:
		[:i|
		 (self vmOwnerIs: i) ifFalse:
			[thread := threads at: i.
			 thread state =  CTMWantingOwnership ifTrue:
				[(threadWantingVM isNil
				  or: [threadWantingVM priority < thread priority]) ifTrue:
					[threadWantingVM := thread]].
			 thread state =  CTMAssignableOrInVM ifTrue:
				[(threadWilling isNil
				  or: [threadWilling priority < thread priority]) ifTrue:
					[threadWilling := thread]]]].
	^threadWantingVM ifNil:
		[threadWilling]! !

CogClass subclass: #Cogit
	instanceVariableNames: 'coInterpreter objectMemory objectRepresentation processor processorLock methodZone methodZoneBase codeBase minValidCallAddress lastNInstructions simulatedAddresses simulatedTrampolines simulatedVariableGetters simulatedVariableSetters printRegisters printInstructions compilationTrace clickConfirm breakPC breakBlock singleStep guardPageSize traceFlags traceStores breakMethod methodObj enumeratingCogMethod methodHeader initialPC endPC methodOrBlockNumArgs inBlock needsFrame hasYoungReferent hasMovableLiteral primitiveIndex backEnd literalsManager methodLabel stackCheckLabel blockEntryLabel blockEntryNoContextSwitch blockNoContextSwitchOffset stackOverflowCall sendMiss missOffset entryPointMask checkedEntryAlignment uncheckedEntryAlignment cmEntryOffset entry cmNoCheckEntryOffset noCheckEntry fullBlockEntry cbEntryOffset fullBlockNoContextSwitchEntry cbNoSwitchEntryOffset picMNUAbort picInterpretAbort endCPICCase0 endCPICCase1 firstCPICCaseOffset cPICCaseSize cPICEndSize closedPICSize openPICSize fixups abstractOpcodes generatorTable byte0 byte1 byte2 byte3 bytecodePC bytecodeSetOffset opcodeIndex numAbstractOpcodes blockStarts blockCount labelCounter cStackAlignment expectedSPAlignment expectedFPAlignment codeModified maxLitIndex ceMethodAbortTrampoline cePICAbortTrampoline ceCheckForInterruptTrampoline ceCPICMissTrampoline ceReturnToInterpreterTrampoline ceBaseFrameReturnTrampoline ceReapAndResetErrorCodeTrampoline ceSendMustBeBooleanAddTrueTrampoline ceSendMustBeBooleanAddFalseTrampoline ceCannotResumeTrampoline ceEnterCogCodePopReceiverReg ceCallCogCodePopReceiverReg ceCallCogCodePopReceiverAndClassRegs cePrimReturnEnterCogCode cePrimReturnEnterCogCodeProfiling ceNonLocalReturnTrampoline ceFetchContextInstVarTrampoline ceStoreContextInstVarTrampoline ceEnclosingObjectTrampoline ceFlushDCache ceFlushICache ceTraceLinkedSendTrampoline ceTraceBlockActivationTrampoline ceTraceStoreTrampoline ceGetFP ceGetSP ceCaptureCStackPointers ceInvokeInterpret ordinarySendTrampolines superSendTrampolines directedSuperSendTrampolines directedSuperBindingSendTrampolines dynamicSuperSendTrampolines outerSendTrampolines selfSendTrampolines firstSend lastSend realCEEnterCogCodePopReceiverReg realCECallCogCodePopReceiverReg realCECallCogCodePopReceiverAndClassRegs trampolineTableIndex trampolineAddresses objectReferencesInRuntime runtimeObjectRefIndex cFramePointerInUse ceTryLockVMOwner extA extB numExtB tempOop numIRCs indexOfIRC theIRCs receiverTags implicitReceiverSendTrampolines cogMethodSurrogateClass cogBlockMethodSurrogateClass nsSendCacheSurrogateClass cPICPrototype cPICEndOfCodeOffset cPICEndOfCodeLabel ceMallocTrampoline ceFreeTrampoline ceFFICalloutTrampoline debugBytecodePointers debugOpcodeIndices disassemblingMethod cogConstituentIndex directedSendUsesBinding processorFrameValid codeToDataDelta varBaseAddress codeZoneIsExecutableNotWritable debugAPISelector shortCutTrampolineBlocks ioHighResClock perMethodProfile instructionProfile'
	classVariableNames: 'AltBlockCreationBytecodeSize AltFirstSpecialSelector AltNSSendIsPCAnnotated AltNumSpecialSelectors AnnotationConstantNames AnnotationShift AnnotationsWithBytecodePCs BlockCreationBytecodeSize Debug DisplacementMask DisplacementX2N EagerInstructionDecoration FirstAnnotation FirstSpecialSelector HasBytecodePC IsAbsPCReference IsAnnotationExtension IsDirectedSuperBindingSend IsDirectedSuperSend IsDisplacementX2N IsNSDynamicSuperSend IsNSImplicitReceiverSend IsNSSelfSend IsNSSendCall IsObjectReference IsRelativeCall IsSendCall IsSuperSend MapEnd MaxCPICCases MaxCompiledPrimitiveIndex MaxStackAllocSize MaxX2NDisplacement NSCClassTagIndex NSCEnclosingObjectIndex NSCNumArgsIndex NSCSelectorIndex NSCTargetIndex NSSendIsPCAnnotated NumObjRefsInRuntime NumOopsPerNSC NumSpecialSelectors NumTrampolines ProcessorClass RRRName'
	poolDictionaries: 'CogAbstractRegisters CogCompilationConstants CogMethodConstants CogRTLOpcodes VMBasicConstants VMBytecodeConstants VMObjectIndices VMStackFrameOffsets'
	category: 'VMMaker-JIT'!

Cogit class
	instanceVariableNames: 'generatorTable primitiveTable'!

!Cogit commentStamp: 'eem 1/28/2020 17:27' prior: 0!
I am the code generator for the Cog VM.  My job is to produce machine code versions of methods for faster execution and to manage inline caches for faster send performance.  I can be tested in the current image using my class-side in-image compilation facilities.  e.g. try

	StackToRegisterMappingCogit genAndDis: (Integer >> #benchFib)

I have concrete subclasses that implement different levels of optimization:
	SimpleStackBasedCogit is the simplest code generator.

	StackToRegisterMappingCogit is the current production code generator  It defers pushing operands
	to the stack until necessary and implements a register-based calling convention for low-arity sends.

	SistaCogit is an experimental code generator with support for counting
	conditional branches, intended to support adaptive optimization.

	RegisterAllocatingCogit is an experimental code generator with support for allocating temporary variables
	to registers. It is inended to serve as the superclass to SistaCogit once it is working.

	SistaRegisterAllocatingCogit and SistaCogitClone are temporary classes that allow testing a clone of
	SistaCogit that inherits from RegisterAllocatingCogit.  Once things work these will be merged and
	will replace SistaCogit.

coInterpreter <CoInterpreterSimulator>
	the VM's interpreter with which I cooperate
methodZoneManager <CogMethodZoneManager>
	the manager of the machine code zone
objectRepresentation <CogObjectRepresentation>
	the object used to generate object accesses
processor <BochsIA32Alien|?>
	the simulator that executes the IA32/x86 machine code I generate when simulating execution in Smalltalk
simulatedTrampolines <Dictionary of Integer -> MessageSend>
	the dictionary mapping trap jump addresses to run-time routines used to warp from simulated machine code in to the Smalltalk run-time.
simulatedVariableGetters <Dictionary of Integer -> MessageSend>
	the dictionary mapping trap read addresses to variables in run-time objects used to allow simulated machine code to read variables in the Smalltalk run-time.
simulatedVariableSetters <Dictionary of Integer -> MessageSend>
	the dictionary mapping trap write addresses to variables in run-time objects used to allow simulated machine code to write variables in the Smalltalk run-time.
printRegisters printInstructions clickConfirm <Boolean>
	flags controlling debug printing and code simulation
codeToDataDelta <Integer>
	if non-zero this is the delta between the read/execute method zone and the read/write mapping of the method zone.  On operating systems where it is entirely disallowed to execute code in a writable region this split is necessary to be able to modify code.  In this regime all writes are to the read/write mapped zone.
breakPC <Integer|UndefinedObject|Boolean|Array>
	machine code pc breakpoint
selectorOop <sqInt>
	the oop of the methodObj being compiled
methodObj <sqInt>
	the bytecode method being compiled
initialPC endPC <Integer>
	the start and end pcs of the methodObj being compiled
methodOrBlockNumArgs <Integer>
	argument count of current method or block being compiled
needsFrame <Boolean>
	whether methodObj or block needs a frame to execute
primitiveIndex <Integer>
	primitive index of current method being compiled
methodLabel <CogAbstractOpcode>
	label for the method header
blockEntryLabel <CogAbstractOpcode>
	label for the start of the block dispatch code
stackOverflowCall <CogAbstractOpcode>
	label for the call of ceStackOverflow in the method prolog
sendMissCall <CogAbstractOpcode>
	label for the call of ceSICMiss in the method prolog
entryOffset <Integer>
	offset of method entry code from start (header) of method
entry <CogAbstractOpcode>
	label for the first instruction of the method entry code
noCheckEntryOffset <Integer>
	offset of the start of a method proper (after the method entry code) from start (header) of method
noCheckEntry <CogAbstractOpcode>
	label for the first instruction of start of a method proper
fixups <Array of <AbstractOpcode Label | nil>>
	the labels for forward jumps that will be fixed up when reaching the relevant bytecode.  fixups has one element per byte in methodObj's bytecode; initialPC maps to fixups[0].
abstractOpcodes <Array of <AbstractOpcode>>
	the code generated when compiling methodObj
byte0 byte1 byte2 byte3 <Integer>
	individual bytes of current bytecode being compiled in methodObj
bytecodePointer <Integer>
	bytecode pc (same as Smalltalk) of the current bytecode being compiled
opcodeIndex <Integer>
	the index of the next free entry in abstractOpcodes (this code is translated into C where OrderedCollection et al do not exist)
numAbstractOpcodes <Integer>
	the number of elements in abstractOpcocdes
blockStarts <Array of <BlockStart>>
	the starts of blocks in the current method
blockCount
	the index into blockStarts as they are being noted, and hence eventually the total number of blocks in the current method
labelCounter <Integer>
	a nicety for numbering labels not needed in the production system but probably not expensive enough to worry about
ceStackOverflowTrampoline <Integer>
ceSend0ArgsTrampoline <Integer>
ceSend1ArgsTrampoline <Integer>
ceSend2ArgsTrampoline <Integer>
ceSendNArgsTrampoline <Integer>
ceSendSuper0ArgsTrampoline <Integer>
ceSendSuper1ArgsTrampoline <Integer>
ceSendSuper2ArgsTrampoline <Integer>
ceSendSuperNArgsTrampoline <Integer>
ceSICMissTrampoline <Integer>
ceCPICMissTrampoline <Integer>
ceStoreCheckTrampoline <Integer>
ceReturnToInterpreterTrampoline <Integer>
ceBaseFrameReturnTrampoline <Integer>
ceSendMustBeBooleanTrampoline <Integer>
ceClosureCopyTrampoline <Integer>
	the various trampolines (system-call-like jumps from machine code to the run-time).
	See Cogit>>generateTrampolines for the mapping from trampoline to run-time
	routine and then read the run-time routine for a funcitonal description.
ceEnterCogCodePopReceiverReg <Integer>
	the enilopmart (jump from run-time to machine-code)
methodZoneBase <Integer>
!

Cogit class
	instanceVariableNames: 'generatorTable primitiveTable'!

!Cogit class methodsFor: 'translation' stamp: 'eem 6/7/2018 15:17'!
activeCompilerClass
	^InitializationOptions
		at: #CogCompilerClass
		ifPresent: [:compilerClassName| Smalltalk classNamed: compilerClassName]
		ifAbsent:
			[(CogAbstractInstruction subclasses detect:
				[:compilerClass|
				 compilerClass ISA == (InitializationOptions at: #ISA ifAbsent: [#IA32])]) defaultCompilerClass]! !

!Cogit class methodsFor: 'translation' stamp: 'eem 3/31/2015 09:04'!
additionalHeadersDo: aBinaryBlock
	"Evaluate aBinaryBlock with the names and contents of
	 any additional header files that need to be generated."

	NewspeakVM ifTrue:
		[aBinaryBlock
			value: 'cogmethod.h'
			value: NewspeakCogMethod cogMethodHeader.
		 aBinaryBlock
			value: 'nssendcache.h'
			value: NSSendCache typedef.
		 ^self].

	aBinaryBlock
		value: 'cogmethod.h'
		value: CogMethod cogMethodHeader! !

!Cogit class methodsFor: 'translation' stamp: 'eem 4/16/2009 12:50'!
additionalSelectorTables
	^self tableFunctions! !

!Cogit class methodsFor: 'translation' stamp: 'eem 1/23/2020 14:46'!
ancilliaryClasses
	"Answer any extra classes to be included in the translation."
	ProcessorClass ifNil:
		[Cogit initializeMiscConstants].
	^(self activeCompilerClass withAllSuperclasses copyUpThrough: CogAbstractInstruction),
	  {	CogMethodZone.
		CogBlockStart.
		CogBytecodeDescriptor.
		CogBytecodeFixup.
		CogPrimitiveDescriptor.
		CogBlockMethod.
		CogMethod},
	self activeCompilerClass literalsManagerClass translationClasses,
	((InitializationOptions at: #NewspeakVM ifAbsent: [false])
		ifTrue: [{NewspeakCogMethod. NSSendCache}]
		ifFalse: [#()]),
	((InitializationOptions at: #SistaVM ifAbsent: [false])
		ifTrue: [{SistaMethodZone}]
		ifFalse: [#()])! !

!Cogit class methodsFor: 'class initialization' stamp: 'eem 2/27/2009 10:34'!
annotationConstantNames
	"for printing..."
	^AnnotationConstantNames! !

!Cogit class methodsFor: 'translation' stamp: 'eem 2/10/2009 16:34'!
apiExportHeaderName
	^'cogit.h'! !

!Cogit class methodsFor: 'in-image compilation support' stamp: 'eem 10/31/2012 16:10'!
asOptionsDictionary: optionsDictionaryOrArray
	optionsDictionaryOrArray isArray ifTrue:
		[| dict |
		dict := Dictionary new.
		self assert: optionsDictionaryOrArray size even.
		1 to: optionsDictionaryOrArray size by: 2 do:
			[:i| | key |
			key := optionsDictionaryOrArray at: i.
			self assert: key isSymbol.
			dict at: key put: (optionsDictionaryOrArray at: i + 1)].
		^dict].
	^optionsDictionaryOrArray! !

!Cogit class methodsFor: 'documentation' stamp: 'eem 5/6/2009 11:15'!
assumptions
	"Byte architecture; no 9-bit bit bytes etc.

	 Stacks grow down.  No support for HPPA.

	 No delay slots.  No support for SPARC.

	 Relative jumps.

	 All long conditional branches are of the same size.

	 All long unconditional branches are of the same size.

	 All MoveCwRs are of the same size.

	 A long jump and a call have the same displacement layout so both can be labelled with IsRelativeCall.

	 A PushCw and a MoveCwR have the same constant layout so both can be labelled with IsObjectReference or IsMethodReference.

	 64k of code per method more than enough"! !

!Cogit class methodsFor: 'in-image compilation support' stamp: 'eem 7/21/2021 08:01'!
attemptToComputeInstVarNamesFor: aCompiledMethod
	aCompiledMethod methodClass instSize > 0 ifTrue:
		[InitializationOptions
			at: #instVarNames
			put: aCompiledMethod methodClass allInstVarNames]! !

!Cogit class methodsFor: 'in-image compilation support' stamp: 'eem 9/17/2018 17:50'!
attemptToComputeTempNamesFor: aCompiledMethod
	(aCompiledMethod respondsTo: #tempNames) ifTrue:
		[| schematicTemps blocks |
		 schematicTemps := aCompiledMethod methodNode schematicTempNamesString.
		 blocks := aCompiledMethod embeddedBlockClosures.
		 InitializationOptions
			at: #tempNames
			put: (Dictionary newFrom: {aCompiledMethod initialPC -> (self decomposeSchematicTemps: (schematicTemps copyUpTo: $[))},
				(blocks
					ifEmpty: [#()]
					ifNotEmpty:
						[aCompiledMethod embeddedBlockClosures
							with: (schematicTemps first = $[
									ifTrue: [schematicTemps piecesCutWhere: [:a :b| b = $[]]
									ifFalse: [(schematicTemps piecesCutWhere: [:a :b| b = $[]) allButFirst])
							collect: [:c :s| c startpc -> (self decomposeSchematicTemps: (s copyWithoutAll: '[]'))]]))]! !

!Cogit class methodsFor: 'translation' stamp: 'eem 11/11/2020 18:14'!
bindingForNumTrampolines
	"The number of trampolines depends on some translation-time constants and some compile-time constants.
	  This answers a binding the computes the right value for the compile-time defintions."
	| mtValues roValues |
	self savingCurrentValuesOf: #(COGMTVM IMMUTABILITY)
		do: [mtValues := self numTrampolineValuesFor: #COGMTVM.
			 roValues := self numTrampolineValuesFor: #IMMUTABILITY].
	^#NumTrampolines -> ('(', (mtValues min min: roValues min) printString,
							' + (COGMTVM ? ', (mtValues max - mtValues min) printString, ' : 0)',
							' + (IMMUTABILITY ? ', (roValues max - roValues min) printString, ' : 0))')

	"self bindingForNumTrampolines"! !

!Cogit class methodsFor: 'accessing class hierarchy' stamp: 'eem 2/4/2020 11:57'!
choose32BitISA
	^[:choices| choices at: (UIManager default chooseFrom: choices) ifAbsent: [^nil]]
		value: (CogAbstractInstruction allSubclasses select: [:c| c wordSize = 4] thenCollect: [:ea| ea ISA]) asSet asArray sorted! !

!Cogit class methodsFor: 'accessing class hierarchy' stamp: 'eem 2/4/2020 11:58'!
choose64BitISA
	^[:choices| choices at: (UIManager default chooseFrom: choices) ifAbsent: [^nil]]
		value: (CogAbstractInstruction allSubclasses select: [:c| c wordSize = 8] thenCollect: [:ea| ea ISA]) asSet asArray sorted! !

!Cogit class methodsFor: 'accessing class hierarchy' stamp: 'eem 1/20/2011 21:18'!
chooseCogitClass
	^Smalltalk at: ([:choices| choices at: (UIManager default chooseFrom: choices) ifAbsent: [^nil]]
						value: (Cogit allSubclasses collect: [:ea| ea  name]) sorted)! !

!Cogit class methodsFor: 'in-image compilation' stamp: 'eem 4/4/2014 12:11'!
cog: aCompiledMethod
	^self cog: aCompiledMethod options: #()! !

!Cogit class methodsFor: 'in-image compilation' stamp: 'cb 4/11/2016 15:38'!
cog: aCompiledMethod options: optionsArray
	^self cog: aCompiledMethod selectorOrNumCopied: aCompiledMethod selector options: optionsArray! !

!Cogit class methodsFor: 'in-image compilation' stamp: 'cb 4/11/2016 15:39'!
cog: aCompiledMethod selector: aSelector
	^self cog: aCompiledMethod selectorOrNumCopied: aSelector options: #()! !

!Cogit class methodsFor: 'in-image compilation' stamp: 'eem 1/24/2018 14:14'!
cog: aCompiledMethod selectorOrNumCopied: selectorOrNumCopied options: optionsDictionaryOrArray
	"StackToRegisterMappingCogit cog: (Integer >> #benchFib) selector: #benchFib options: #(COGMTVM false)"
	| cogit coInterpreter |
	cogit := self initializedInstanceForTests: optionsDictionaryOrArray.
	coInterpreter := cogit coInterpreter.
	self attemptToComputeTempNamesFor: aCompiledMethod.
	self attemptToComputeInstVarNamesFor: aCompiledMethod.
	^{ coInterpreter.
		cogit.
		selectorOrNumCopied isInteger
			ifTrue: [ cogit cogFullBlockMethod: (coInterpreter oopForObject: aCompiledMethod) numCopied: selectorOrNumCopied ]
			ifFalse: [ cogit cog: (coInterpreter oopForObject: aCompiledMethod) selector: (coInterpreter oopForObject: selectorOrNumCopied) ] }! !

!Cogit class methodsFor: 'documentation' stamp: 'eem 7/5/2011 16:59'!
cogToBytecodeMethodMapping
	"When first written the VM maintained a one-to-one mapping between bytecoded methods and cog methods.
	The map is implemented with a direct reference form the methodObject field of a CogMethod to the bytecoded
	method and by setting the header word of the bytecoded method to the CogMethod.  Newspeak forced a change
	in this since Newspeak shares anonymous (nil methodClassAssociation) accessors (setters and getters) amongst
	classes and mixins throughout the system (i.e. these accessors are in dictionaries under selector keys that don't
	match the accessor methods' selectors).  But the inline cacheing relinking/un;linking machinery depends on
	CogMethods having the correct selector, since to unlink/relink the selector is fetched from the inline cache's
	target method.  So the VM has been modified to not insist on a one-to-one mapping.  Instead, when a bytecoded
	method is to be used with a different selector it is attached to the new CogMethod.

	This choice means that CogMethods do refer to a valid bytecoded method, but the bytecoded method might
	not necessarily refer back to a CogMethod referring to it.  So when CogMethods are freed care must be taken
	to reset the bytecoded method's header if it is referring to the freed CogMethod."! !

!Cogit class methodsFor: 'in-image compilation support' stamp: 'eem 10/8/2020 17:34'!
compileMethodSource: source options: optionsArray
	"The key thing here is to work out what bytecode set to use for the options..."
	| encoderClass encoderClassName methodNode |
	StackInterpreter initializeWithOptions: (self asOptionsDictionary: optionsArray).
	encoderClassName := StackInterpreter classPool at: #BytecodeEncoderClassName.
	(encoderClass := Smalltalk classNamed: encoderClassName) ifNil:
		[self error: 'no available compiler backend for this bytecode set'].
	methodNode := Parser new
						encoderClass: encoderClass;
						parse: '| temp | temp class' class: nil class noPattern: true notifying: nil ifFail: [self error: 'bad doit'].
	^methodNode generate! !

!Cogit class methodsFor: 'translation' stamp: 'eem 8/20/2021 23:19'!
declareCVarsIn: aCCodeGenerator
	| backEnd |
	backEnd := CogCompilerClass basicNew.
	#(	'coInterpreter' 'objectMemory' 'methodZone' 'objectRepresentation'
		'cogBlockMethodSurrogateClass' 'cogMethodSurrogateClass' 'nsSendCacheSurrogateClass'
		'threadManager' 'processor' 'lastNInstructions' 'simulatedAddresses' 'ioHighResClock'
		'simulatedTrampolines' 'simulatedVariableGetters' 'simulatedVariableSetters'
		'processorFrameValid' 'printRegisters' 'printInstructions' 'clickConfirm' 'singleStep'
		'codeZoneIsExecutableNotWritable' 'debugAPISelector' 'shortCutTrampolineBlocks'
		'perMethodProfile' 'instructionProfile') do:
			[:simulationVariableUnusedByRealVM|
			aCCodeGenerator removeVariable: simulationVariableUnusedByRealVM].
	NewspeakVM ifFalse:
		[#(	'selfSendTrampolines' 'dynamicSuperSendTrampolines'
			'implicitReceiverSendTrampolines' 'outerSendTrampolines'
			'ceEnclosingObjectTrampoline' 'numIRCs' 'indexOfIRC' 'theIRCs') do:
				[:variableNotNeededInNormalVM|
				aCCodeGenerator removeVariable: variableNotNeededInNormalVM]].
	aCCodeGenerator removeConstant: #COGMTVM. "this should be defined at compile time"
	"N.B. We *do not* include sq.h; it pulls in conflicting definitions now that sqVirtualMachine.h
	 declares cointerp's functions, and declares some of them inaccurately for histrical reasons.
	 We pull in CoInterpreter's api via cointerp.h which is accurate."
	aCCodeGenerator
		addHeaderFile:'"sqConfig.h"'; "config.h must be first on linux"
		addHeaderFile:'<stddef.h>'; "for e.g. offsetof"
		addHeaderFile:'<stdio.h>';
		addHeaderFile:'<stdlib.h>';
		addHeaderFile:'<string.h>';
		addHeaderFile:'"sqPlatformSpecific.h"'; "e.g. solaris overrides things for sqCogStackAlignment.h"
		addHeaderFile:'"sqMemoryAccess.h"';
		addHeaderFile:'"sqCogStackAlignment.h"';
		addHeaderFile:'"dispdbg.h"'; "must precede cointerp.h & cogit.h otherwise NoDbgRegParms gets screwed up"
		addHeaderFile:'"cogmethod.h"'.
	NewspeakVM ifTrue:
		[aCCodeGenerator addHeaderFile:'"nssendcache.h"'].
	aCCodeGenerator
		addHeaderFile:'#if COGMTVM';
		addHeaderFile:'"cointerpmt.h"';
		addHeaderFile:'#else';
		addHeaderFile:'"cointerp.h"';
		addHeaderFile:'#endif';
		addHeaderFile:'"cogit.h"'.
	aCCodeGenerator
		var: #ceGetFP
			declareC: 'usqIntptr_t (*ceGetFP)(void)';
		var: #ceGetSP
			declareC: 'usqIntptr_t (*ceGetSP)(void)';
		var: #ceCaptureCStackPointers
			declareC: 'void (*ceCaptureCStackPointers)(void)';
		var: #ceInvokeInterpret
			declareC: 'void (*ceInvokeInterpret)(void)';
		var: #ceEnterCogCodePopReceiverReg
			declareC: 'void (*ceEnterCogCodePopReceiverReg)(void)';
		var: #realCEEnterCogCodePopReceiverReg
			declareC: 'void (*realCEEnterCogCodePopReceiverReg)(void)';
		var: #ceCallCogCodePopReceiverReg
			declareC: 'void (*ceCallCogCodePopReceiverReg)(void)';
		var: #realCECallCogCodePopReceiverReg
			declareC: 'void (*realCECallCogCodePopReceiverReg)(void)';
		var: #ceCallCogCodePopReceiverAndClassRegs
			declareC: 'void (*ceCallCogCodePopReceiverAndClassRegs)(void)';
		var: #realCECallCogCodePopReceiverAndClassRegs
			declareC: 'void (*realCECallCogCodePopReceiverAndClassRegs)(void)';
		var: #postCompileHook
			declareC: 'void (*postCompileHook)(CogMethod *)';
		var: #openPICList declareC: 'CogMethod *openPICList = 0';
		var: #maxMethodBefore type: #'CogBlockMethod *';
		var: 'enumeratingCogMethod' type: #'CogMethod *'.
	
	aCCodeGenerator
		var: #ceTryLockVMOwner
		declareC: '#if COGMTVM\usqIntptr_t (*ceTryLockVMOwner)(usqIntptr_t)\#endif'.

	backEnd numICacheFlushOpcodes > 0 ifTrue:
		[aCCodeGenerator
			var: #ceFlushICache
				declareC: 'static void (*ceFlushICache)(usqIntptr_t from, usqIntptr_t to)'].
	aCCodeGenerator
		var: #ceFlushDCache
			declareC: '#if DUAL_MAPPED_CODE_ZONE\static void (*ceFlushDCache)(usqIntptr_t from, usqIntptr_t to)\#endif';
		var: #codeToDataDelta
			declareC: '#if DUAL_MAPPED_CODE_ZONE\static sqInt codeToDataDelta\#else\# define codeToDataDelta 0\#endif'.

	aCCodeGenerator
		declareVar: 'aMethodLabel' type: #'AbstractInstruction'; "Has to come lexicographically before backEnd & methodLabel"
		var: #backEnd declareC: 'AbstractInstruction * const backEnd = &aMethodLabel';
		var: #methodLabel declareC: 'AbstractInstruction * const methodLabel = &aMethodLabel'.
	self declareC: #(abstractOpcodes stackCheckLabel
					blockEntryLabel blockEntryNoContextSwitch
					stackOverflowCall sendMiss
					entry noCheckEntry selfSendEntry dynSuperEntry
					fullBlockNoContextSwitchEntry fullBlockEntry
					picMNUAbort picInterpretAbort  endCPICCase0 endCPICCase1 cPICEndOfCodeLabel)
			as: #'AbstractInstruction *'
				in: aCCodeGenerator.
	aCCodeGenerator
		declareVar: #cPICPrototype type: #'CogMethod *';
		declareVar: #blockStarts type: #'BlockStart *';
		declareVar: #fixups type: #'BytecodeFixup *';
		declareVar: #methodZoneBase type: #usqInt.
	aCCodeGenerator
		var: #ordinarySendTrampolines
			declareC: 'sqInt ordinarySendTrampolines[NumSendTrampolines]';
		var: #superSendTrampolines
			declareC: 'sqInt superSendTrampolines[NumSendTrampolines]'.
	BytecodeSetHasDirectedSuperSend ifTrue:
		[aCCodeGenerator
			var: #directedSuperSendTrampolines
				declareC: 'sqInt directedSuperSendTrampolines[NumSendTrampolines]';
			var: #directedSuperBindingSendTrampolines
				declareC: 'sqInt directedSuperBindingSendTrampolines[NumSendTrampolines]'].
	NewspeakVM ifTrue:
		[aCCodeGenerator
			var: #selfSendTrampolines
				declareC: 'sqInt selfSendTrampolines[NumSendTrampolines]';
			var: #dynamicSuperSendTrampolines
				declareC: 'sqInt dynamicSuperSendTrampolines[NumSendTrampolines]';
			var: #implicitReceiverSendTrampolines
				declareC: 'sqInt implicitReceiverSendTrampolines[NumSendTrampolines]';
			var: #outerSendTrampolines
				declareC: 'sqInt outerSendTrampolines[NumSendTrampolines]'].
	aCCodeGenerator
		addConstantForBinding: self bindingForNumTrampolines;
		var: #trampolineAddresses
			declareC: 'static char *trampolineAddresses[NumTrampolines*2]';
		var: #objectReferencesInRuntime
			declareC: 'static usqInt objectReferencesInRuntime[NumObjRefsInRuntime+1]';
		var: #labelCounter
			type: #int;
		var: #traceFlags
			declareC: 'int traceFlags = 8 /* prim trace log on by default */';
		var: #cStackAlignment
			declareC: 'const int cStackAlignment = STACK_ALIGN_BYTES'.
	aCCodeGenerator
		declareVar: #minValidCallAddress type: #'usqIntptr_t'.
	aCCodeGenerator vmClass generatorTable ifNotNil:
		[:bytecodeGenTable|
		aCCodeGenerator
			var: #generatorTable
				declareC: 'static BytecodeDescriptor generatorTable[', bytecodeGenTable size printString, ']',
							(self tableInitializerFor: bytecodeGenTable
								in: aCCodeGenerator)].
	"In C the abstract opcode names clash with the Smalltalk generator syntactic sugar.
	 Most of the syntactic sugar is inlined, but alas some remains.  Rename the syntactic
	 sugar to avoid the clash."
	(self organization listAtCategoryNamed: #'abstract instructions') do:
		[:s|
		aCCodeGenerator addSelectorTranslation: s to: 'g', (aCCodeGenerator cFunctionNameFor: s)].
	aCCodeGenerator addSelectorTranslation: #halt: to: 'haltmsg'.
	self declareFlagVarsAsByteIn: aCCodeGenerator! !

!Cogit class methodsFor: 'translation' stamp: 'eem 2/3/2020 20:18'!
declareFlagVarsAsByteIn: aCCodeGenerator
	CogCompilerClass basicNew byteReadsZeroExtend ifTrue:
		[self declareC: #(codeModified deadCode directedSendUsesBinding
						hasMovableLiteral hasNativeFrame hasYoungReferent
						inBlock needsFrame regArgsHaveBeenPushed useTwoPaths)
				as: #'unsigned char'
					ifPresentIn: aCCodeGenerator]! !

!Cogit class methodsFor: 'in-image compilation support' stamp: 'eem 3/9/2018 12:11'!
decomposeSchematicTemps: aString
	^Array streamContents:
		[:ws| | rs |
		rs := aString readStream.
		[rs atEnd] whileFalse:
			[ws nextPut: (rs peek = $(
						ifTrue: [[rs upThrough: $)]
									on: MessageNotUnderstood
									do: [:ex| (rs upTo: $)), ')']]
						ifFalse: [rs upTo: Character space])]]! !

!Cogit class methodsFor: 'accessing class hierarchy' stamp: 'eem 10/9/2013 17:50'!
defaultObjectMemoryClass
	^NewCoObjectMemory! !

!Cogit class methodsFor: 'C translation' stamp: 'eem 4/28/2016 21:21'!
defineAtCompileTime: anObject
	"Override to define at translation time those variables that need to
	 be defined at compile time only in plugins, but not in the main VM,
	 because the VM generated is specific to these varables."
	anObject isSymbol ifFalse:
		[^false].
	(#(STACKVM COGVM COGMTVM SPURVM) includes: anObject) ifTrue:
		[^false].
	^VMBasicConstants namesDefinedAtCompileTime includes: anObject! !

!Cogit class methodsFor: 'translation' stamp: 'eem 1/10/2016 08:50'!
doInlining
	"inline only those methods that are marked with <inline: true>"
	^#asSpecifiedOrQuick! !

!Cogit class methodsFor: 'in-image compilation' stamp: 'eem 10/31/2012 16:13'!
genAndDis: methodOrDoitString
	^self genAndDis: methodOrDoitString options: #()! !

!Cogit class methodsFor: 'in-image compilation' stamp: 'cb 4/11/2016 15:40'!
genAndDis: compiledBlock numCopied: numCopied options: optionsDictionaryOrArray
	| tuple |
	tuple := self cog: compiledBlock selectorOrNumCopied: numCopied options: optionsDictionaryOrArray.
	tuple second disassembleMethod: tuple last.
	^tuple! !

!Cogit class methodsFor: 'in-image compilation' stamp: 'eem 10/8/2020 17:35'!
genAndDis: methodOrDoitString options: optionsDictionaryOrArray
	| tuple |
	methodOrDoitString isCompiledCode ifFalse:
		[^self
			genAndDis: (self compileMethodSource: methodOrDoitString options: optionsDictionaryOrArray)
			 options: optionsDictionaryOrArray].
	tuple := self cog: methodOrDoitString
				selectorOrNumCopied: (methodOrDoitString isCompiledBlock
											ifTrue: [methodOrDoitString numCopiedValues]
											ifFalse: [methodOrDoitString selector])
				options: optionsDictionaryOrArray.
	tuple second disassembleMethod: tuple last.
	^tuple! !

!Cogit class methodsFor: 'in-image compilation' stamp: 'eem 1/24/2018 14:15'!
genAndDisPICoptions: optionsDictionaryOrArray
	"StackToRegisterMappingCogit genAndDisPICoptions: #(ISA ARMv5 CogCompilerClass CogOutOfLineLiteralsARMCompiler)"
	| cogit |
	cogit := self initializedInstanceForTests: optionsDictionaryOrArray.
	cogit disassembleFrom: cogit cPICPrototype + (cogit sizeof: CogMethod) to: cogit cPICPrototype + cogit closedPICSize! !

!Cogit class methodsFor: 'translation' stamp: 'eem 10/28/2020 17:13'!
generateCodeStringForCogitDotC
	"Generate a skeletal cogit.c that includes the relevant cogitFOO.c
	 for the appropriate subclasses of CogAbstractInstruction.
	 self generateCodeStringForCogitDotC"
	 
	| string insertPosition abis defaultDef |
	abis := OrderedCollection new.
	string := String streamContents:
		[:s|
		 s nextPutAll: '/* Automatically generated by\	' withCRs.
		 s nextPutAll: (CCodeGenerator monticelloDescriptionFor: self).
		 s cr; nextPutAll: ' */'.
		 s cr; cr; nextPut: $#.
		 insertPosition := s position.
		 self translateableInstructionSubclassesAndInstalledOptionsDo:
			[:class | | abi |
			 s nextPutAll: 'if '.
			 (abi := InitializationOptions at: #ABI ifAbsent: []) ifNotNil:
				[s nextPutAll: (abis addLast: abi); nextPutAll: ' && ('].
			 class identifyingPredefinedMacros
				do: [:predefinedMacro|
					(predefinedMacro includes: Character space) "not a macr name but an expression"
						ifTrue: [s nextPut: $(; nextPutAll: predefinedMacro]
						ifFalse: [s nextPutAll: 'defined('; nextPutAll: predefinedMacro].
					s nextPut: $)]
				separatedBy: [s nextPutAll: ' || '].
			 abi ifNotNil: [s nextPut: $)].
			 s cr; cr; nextPutAll: '#	include "'; nextPutAll: class moduleName; nextPutAll: '.c"'.
			 s cr; cr; nextPutAll: '#el'].
		 s nextPutAll: 'se'.
		 #(	'As yet no Cogit implementation appears to exist for your platform.'
			'Consider implementing it, starting by adding a subclass of CogAbstractInstruction.') do:
			[:msg| s cr; nextPutAll: '#	error '; nextPutAll: msg].
		 s cr; nextPutAll: '#endif'; cr].
	abis isEmpty ifTrue:
		[^string].
	defaultDef := String streamContents:
		[:s|
		s nextPutAll: '#if !!'.
		abis do: [:abi| s nextPutAll: abi] separatedBy: [s nextPutAll: ' && !!'].
		s cr; nextPutAll: '# define '; nextPutAll: abis first; nextPutAll: ' 1'; cr.
		s nextPutAll: '#endif'; cr; cr].
	^string copyReplaceFrom: insertPosition to: insertPosition - 1 with: defaultDef! !

!Cogit class methodsFor: 'accessing' stamp: 'eem 11/20/2010 09:17'!
generatorTable
	^generatorTable! !

!Cogit class methodsFor: 'class initialization' stamp: 'cb 1/3/2018 15:10'!
generatorTableFrom: anArray
	| blockCreationBytecodeSize |
	generatorTable := CArrayAccessor on: (Array new: 256).
	anArray do:
		[:tuple| | descriptor |
		(descriptor := CogBytecodeDescriptor new)
						numBytes: tuple first;
						generator: tuple fourth;
						isReturn: (tuple includes: #return);
						hasUnsafeJump: (tuple includes: #hasUnsafeJump);
						isMapped: ((tuple includes: #isMappedIfImmutability)
										ifTrue: [self bindingOf: #IMMUTABILITY]
										ifFalse: [tuple includes: #isMapped]);
						isMappedInBlock: (tuple includes: #isMappedInBlock);
						isBlockCreation: (tuple includes: #block);
						spanFunction: (((tuple includes: #block) or: [(tuple includes: #branch)]) ifTrue:
										[tuple detect: [:thing| thing isSymbol and: [thing numArgs = 4]]]);
						isBranchTrue: (tuple includes: #isBranchTrue);
						isBranchFalse: (tuple includes: #isBranchFalse);
						isExtension: (tuple includes: #extension);
						isInstVarRef: (tuple includes: #isInstVarRef);	"for Spur"
						is1ByteInstVarStore: (tuple includes: #is1ByteInstVarStore);	"for Spur"
						hasIRC: (tuple includes: #hasIRC);			"for Newspeak"
						yourself.
		"As a hack to cut down on descriptor flags, use opcode to tag unusedBytecode for scanning.
		 Currently descriptors are exactly 16 bytes with all 8 flag bits used (in Newspeak at least 17 bytes,
		 9 flag bits).  As another hack to eliminate a test in scanMethod mark unknowns as extensions."
		descriptor generator == #unknownBytecode ifTrue:
			[descriptor opcode: Nop; isExtension: true].
		descriptor isBlockCreation ifTrue:
			[blockCreationBytecodeSize
				ifNil: [blockCreationBytecodeSize := descriptor numBytes]
				ifNotNil: [self assert: blockCreationBytecodeSize = descriptor numBytes]].
		tuple do:
			[:thing|
			thing isSymbol ifTrue:
				[(thing beginsWith: #needsFrame) ifTrue:
					[descriptor needsFrameFunction: thing].
				 (CogRTLOpcodes classPool at: thing ifAbsent: []) ifNotNil:
					[:opcode| descriptor opcode: opcode]]].
		tuple last isInteger
			ifTrue: [descriptor stackDelta: tuple last]
			ifFalse:
				[descriptor needsFrameFunction ifNotNil:
					[self error: 'frameless block bytecodes must specify a stack delta']].
		tuple second to: tuple third do:
			[:index|
			generatorTable at: index put: descriptor]].
	BlockCreationBytecodeSize := blockCreationBytecodeSize.
	^generatorTable! !

!Cogit class methodsFor: 'accessing' stamp: 'eem 11/22/2013 14:30'!
guardPageSize
	^1024! !

!Cogit class methodsFor: 'accessing class hierarchy' stamp: 'eem 9/5/2021 07:34'!
hasCogit
	^true! !

!Cogit class methodsFor: 'translation' stamp: 'eem 2/15/2009 09:12'!
implicitReturnTypeFor: aSelector
	"Answer the return type for methods that don't have an explicit return."
	^#void! !

!Cogit class methodsFor: 'class initialization' stamp: 'eem 10/24/2020 10:24'!
initializeAnnotationConstants
	"These form the method map for a cog method.  The map defines which addresses
	 in a machine code method are ones with important functions, such as being a send
	 site or being a reference to a heap object.  Each annotated instruction has a byte
	 in the map, and each byte in the map has two parts.  In the least signficant bits is
	 a distance.  In the most signficant bits is the type of annotation at the point reached.
	 The first map entry's distance is that from the  the cmNoCheckEntryOffset, and
	 subsequent distances are from the preceding map entry  A null byte ends the map.

	 The map occurs at the end of a method (*), in reverse, so that its start is found by
	 adding the method's block size.  If the distance between two mapped instructions
	 will not fit in the displacement field then one or more displacement entries are placed
	 in the map to bridge the gap.  The displacements are in codeGranularity units so that
	 processors like e.g. ARM, with 4-byte instructions, do not have overly large maps.  In
	 practice, maps are very compact, but they should be as quick to navigate as possible,
	 hence they have a simple regular structure.  Each element of a map is a byte with the
	 three bit type in the most significant bits, and the five bit displacement in the least
	 significant. One annotation, IsAnnotationExtension, implicitly has a zero displacement
	 and reinterprets the displacement field as an extended annotation type. One annotation,
	 IsDisplacementX2N, interprets the displacement field as a power of two number of
	 codeGranularity units, for spanning large displacements compactly.

	 The map is the structure that allows methods to be dynamically linked, unlinked and
	 relocated.  There are two categories of call that may need map entries, calls to (non-
	 send) run-time and/or primitive routines, and send calls. Run-time calls may be
	 changed during relocation. Send calls can be rewritten when linked or unlinked and
	 during relocation. Absolute run- time calls do not need a map entry because they don't
	 change on relocation. Hence the annotaion for non-send run-time/primitive calls is
	 IsRelativeCall. There are several kinds of send call, sends, super sends, and exotic
	 Newspeak sends (self sends, directed super sends, etc).  These send calls need different
	 treatment at different times.  For example, when the send cache is flushed or the
	 method zone is shrunk some sends must be unlinked and some sends must be relocated.
	
	 The map is also used to map machine code pcs to bytecode pcs and vice verse.  This is
	 done by walking the machine code method using its map at the same time as walking
	 the bytecoded method using its bytecodes, processing the pair of pcs at relevant points.
	 In all cases, sends and run-time calls, the address in the map is the address following the
	 call, which is also the return address for the call.

	 Sends are in two states, linked, in which case they call other machine code methods, or
	 unlinked, in which case they call a particular unlinked send run-time routine that will initiate
	 linking. Linked can be distinguished from unlinked sends based on address; only linked
	 sends have their target between methodZoneBase and methodZone freeStart.
	 We used to distinguish normal sends from super sends based on alignment of entry-point,
	 because normal sends link to the checked entry-point, whereas super sends link to the
	 unchecked entry-point, and both entry points have different alignments. But with the
	 arrival of exotic Newspeak sends came the need for supporting more varieties than merely
	 send and super send. So now we use the IsAnnotationExtension to label sends other than
	 normal sends. For these ``exotic'' sends there is both an IsAnnotationExtension annotation
	 and an IsSendCall annotation.

	 While run-time calls can be distinguished from direct primitive calls on the basis
	 of address there is no need to do so.  They are merely calls to locations that don't
	 move during method zone compaction.

	 Absolute PC references are used for method references and counter references.
	 These are references from within a particular method to absolute pcs in that same
	 method that must be relocated when the method moves."
	"self initializeAnnotationConstants"

	AnnotationShift := 5.
	IsDisplacementX2N := 0.	"N.B. A 0 byte ends the map"
	IsAnnotationExtension := 1.	"Used to extend IsSendCall with different codes for exotic send types."
	IsObjectReference := 2.
	IsAbsPCReference := 3.
	IsRelativeCall := 4.
	HasBytecodePC := 5.
	IsNSSendCall := NewspeakVM ifTrue: [6].
	IsSendCall := 7.
	"These are formed by combining IsSendCall and IsAnnotationExtension annotations."
	IsSuperSend := 8.
	IsDirectedSuperSend := BytecodeSetHasDirectedSuperSend ifTrue: [9].
	IsDirectedSuperBindingSend := BytecodeSetHasDirectedSuperSend ifTrue: [10].
	IsNSSelfSend := NewspeakVM ifTrue: [11].
	IsNSDynamicSuperSend := NewspeakVM ifTrue: [12].
	IsNSImplicitReceiverSend := NewspeakVM ifTrue: [13].

	DisplacementMask := (1 << AnnotationShift) - 1.
	DisplacementX2N := IsDisplacementX2N << AnnotationShift.
	FirstAnnotation := IsObjectReference << AnnotationShift.
	MaxX2NDisplacement := DisplacementMask << AnnotationShift.

	MapEnd := 0.

	"These two tables are used for printing annotations during disassembly"
	AnnotationConstantNames := #(	IsDisplacementX2N
										IsAnnotationExtension
										IsObjectReference
										IsAbsPCReference
										IsRelativeCall
										HasBytecodePC
										IsNSSendCall
										IsSendCall
										IsSuperSend
										IsDirectedSuperSend
										IsDirectedSuperBindingSend
										IsNSSelfSend
										IsNSDynamicSuperSend
										IsNSImplicitReceiverSend).
	AnnotationsWithBytecodePCs := #(HasBytecodePC
										IsNSSendCall
										IsSendCall
										IsSuperSend
										IsDirectedSuperSend
										IsDirectedSuperBindingSend
										IsNSSelfSend
										IsNSDynamicSuperSend
										IsNSImplicitReceiverSend),
										{'IsRelativeCall:\HasBytecodePC' withCRs}! !

!Cogit class methodsFor: 'class initialization' stamp: 'eem 6/7/2018 15:18'!
initializeBytecodeTable
	"SimpleStackBasedCogit initializeBytecodeTableWith: Dictionary new"
	"StackToRegisterMappingCogit initializeBytecodeTableWith: Dictionary new"

	| initializer |
	BytecodeSetHasDirectedSuperSend := BytecodeSetHasExtensions := false.
	initializer := InitializationOptions
					at: #bytecodeTableInitializer
					ifAbsent:
						[NewspeakVM
							ifTrue:
								[MULTIPLEBYTECODESETS
									ifTrue: [#initializeBytecodeTableForSqueakV3PlusClosuresNewspeakV4Hybrid]
									ifFalse: [#initializeBytecodeTableForNewspeakV4]]
							ifFalse:
								[MULTIPLEBYTECODESETS
									ifTrue: [#initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid]
									ifFalse: [#initializeBytecodeTableForSqueakV3PlusClosures]]].
	"Now make sure all classes in the hierarchy have initialized to the same bytecode table."
	(self withAllSuperclasses copyUpTo: Cogit) reverseDo: "i.e. exclude Cogit"
		[:cogitClass|
		 cogitClass perform: initializer]! !

!Cogit class methodsFor: 'class initialization' stamp: 'eem 1/19/2017 14:49'!
initializeBytecodeTableForSqueakV3PlusClosuresNewspeakV4Hybrid
	"SimpleStackBasedCogit initializeBytecodeTableForSqueakV3PlusClosuresNewspeakV4Hybrid"
	"StackToRegisterMappingCogit initializeBytecodeTableForSqueakV3PlusClosuresNewspeakV4Hybrid"

	| v3Table v4Table |
	"N.B. Must do it backwards to evaluate AltBlockCreationBytecodeSize & BlockCreationBytecodeSize et al correctly."
	self initializeBytecodeTableForNewspeakV4.
	v4Table := generatorTable.
	AltBlockCreationBytecodeSize := BlockCreationBytecodeSize.
	AltNSSendIsPCAnnotated := NSSendIsPCAnnotated.
	AltFirstSpecialSelector := FirstSpecialSelector.
	AltNumSpecialSelectors := NumSpecialSelectors.
	self initializeBytecodeTableForSqueakV3PlusClosures.
	BytecodeSetHasExtensions := true.
	v3Table := generatorTable.
	generatorTable := CArrayAccessor on: v3Table object, v4Table object! !

!Cogit class methodsFor: 'class initialization' stamp: 'eem 1/19/2017 14:49'!
initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid
	"SimpleStackBasedCogit initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid"
	"StackToRegisterMappingCogit initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid"

	| v3Table v1Table |
	"N.B. Must do it backwards to evaluate AltBlockCreationBytecodeSize & BlockCreationBytecodeSize et al correctly."
	self initializeBytecodeTableForSistaV1.
	v1Table := generatorTable.
	AltBlockCreationBytecodeSize := BlockCreationBytecodeSize.
	AltFirstSpecialSelector := FirstSpecialSelector.
	AltNumSpecialSelectors := NumSpecialSelectors.
	self initializeBytecodeTableForSqueakV3PlusClosures.
	BytecodeSetHasExtensions := true.
	v3Table := generatorTable.
	generatorTable := CArrayAccessor on: v3Table object, v1Table object! !

!Cogit class methodsFor: 'class initialization' stamp: 'eem 1/21/2014 14:55'!
initializeCogMethodConstants
	CMOpenPIC := 1 + (CMClosedPIC := 1 + (CMBlock := 1 + (CMMethod := 1 + (CMFree := 1))))! !

!Cogit class methodsFor: 'class initialization' stamp: 'eem 12/5/2015 12:28'!
initializeErrorCodes
	"External errors, returned to or from cog:selector:"
	NotFullyInitialized := -1.
	InsufficientCodeSpace := -2.
	MethodTooBig := -4.
	YoungSelectorInPIC := -5.
	EncounteredUnknownBytecode := -6.
	UnimplementedPrimitive := -7.
	ShouldNotJIT := -8.
	MaxNegativeErrorCode := ShouldNotJIT.
	"Internal errors returned by generator routines to other generator routines"
	BadRegisterSet := 1.
	UnimplementedOperation := 2.
	"Internal successes answered by CogObjectRepresentation to JIT, etc"
	UnfailingPrimitive := 3. "Answered by a primitive generator for a primitive that will never fail"
	CompletePrimitive := 4 "Answered by a primitive generator that does not bneed to fall back on the interpreter primitive except for an error code."! !

!Cogit class methodsFor: 'class initialization' stamp: 'eem 1/30/2020 20:22'!
initializeMiscConstants
	super initializeMiscConstants.
	Debug := InitializationOptions at: #Debug ifAbsent: [false].
	(InitializationOptions includesKey: #EagerInstructionDecoration)
		ifTrue:
			[EagerInstructionDecoration := InitializationOptions at: #EagerInstructionDecoration]
		ifFalse:
			[EagerInstructionDecoration ifNil:
				[EagerInstructionDecoration := false]]. "speeds up single stepping but could lose fidelity"

	ProcessorClass := (InitializationOptions at: #ISA ifAbsentPut: [self objectMemoryClass defaultISA]) caseOf: {
							[#X64] 	->	[BochsX64Alien].
							[#IA32] 	->	[BochsIA32Alien].
							[#ARMv5]	->	[GdbARMAlien].
							[#ARMv8]	->	[GdbARMv8Alien].
							[#MIPSEL]	->	[MIPSELSimulator] }.
	CogCompilerClass := self activeCompilerClass.
	CogAbstractInstruction initializeAbstractRegisters.

	self objectMemoryClass objectRepresentationClass initializeMiscConstants.
	"Our criterion for which methods to JIT is literal count.  The default value is 60 literals or less."
	MaxLiteralCountForCompile := InitializationOptions at: #MaxLiteralCountForCompile ifAbsent: [60].
	"we special-case 0, 1 & 2 argument sends, N is numArgs >= 3"
	NumSendTrampolines := 4.
	"Currently not even the ceImplicitReceiverTrampoline contains object references."
	NumObjRefsInRuntime := 0.
	"6 is a fine number for the max number of PCI entries.  8 is too large."
	MaxCPICCases := 6.

	"One variable defines whether in a block and whether in a vanilla or full block."
	InVanillaBlock := 1.
	InFullBlock := 2.

	NSCSelectorIndex := (NSSendCache instVarNames indexOf: #selector) - 1.
	NSCNumArgsIndex := (NSSendCache instVarNames indexOf: #numArgs) - 1.
	NSCClassTagIndex := (NSSendCache instVarNames indexOf: #classTag) - 1.
	NSCEnclosingObjectIndex := (NSSendCache instVarNames indexOf: #enclosingObject) - 1.
	NSCTargetIndex := (NSSendCache instVarNames indexOf: #target) - 1.
	NumOopsPerNSC := NSSendCache instVarNames size.

	"Max size to alloca when compiling.
	 Mac OS X 10.6.8 segfaults approaching 8Mb.
	 Linux 2.6.9 segfaults above 11Mb.
	 WIndows XP segfaults approaching 2Mb."
	MaxStackAllocSize := 1024 * 1024 * 3 / 2 ! !

!Cogit class methodsFor: 'class initialization' stamp: 'eem 1/11/2018 19:02'!
initializeNumTrampolines
	NumTrampolines := self numTrampolines
						+ self objectRepresentationClass numTrampolines
						+ (NewspeakVM ifTrue: [NumSendTrampolines * 4 + 2] ifFalse: [0])
						+ (BytecodeSetHasDirectedSuperSend ifTrue: [NumSendTrampolines * 2] ifFalse: [0])! !

!Cogit class methodsFor: 'class initialization' stamp: 'eem 5/15/2014 13:24'!
initializePrimitiveTable
	NewspeakVM
		ifTrue: [self initializePrimitiveTableForNewsqueak]
		ifFalse: [self initializePrimitiveTableForSqueak]! !

!Cogit class methodsFor: 'class initialization' stamp: 'eem 6/7/2018 15:05'!
initializeWithOptions: optionsDictionary
	super initializeWithOptions: optionsDictionary.
	self initializeMiscConstants. "must precede other initialization."
	self initializeErrorCodes.
	self initializeCogMethodConstants.
	self initializeAnnotationConstants.
	self initializeBytecodeTable.
	self initializeNumTrampolines.
	self initializePrimitiveTable! !

!Cogit class methodsFor: 'in-image compilation support' stamp: 'eem 2/4/2020 10:10'!
initializedInstanceForTests: optionsDictionaryOrArray
	"Answer an instance of a Cogit suitable for running tests that has initialized
	 its method zone (generated trampolines etc)"
	| cogit coInterpreter |
	cogit := self instanceForTests: optionsDictionaryOrArray.
	coInterpreter := CurrentImageCoInterpreterFacade forCogit: cogit.
	[| rumpStackSize |
	 rumpStackSize := 1024 * 4.
	 cogit
		setInterpreter: coInterpreter;
		singleStep: true;
		initializeCodeZoneFrom: 1024 upTo: coInterpreter memory byteSize - rumpStackSize.
	 cogit methodZone freeStart: (cogit methodZone freeStart roundUpTo: 1024)]
		on: Notification
		do: [:ex|
			(ex messageText beginsWith: 'cannot find receiver for') ifTrue:
				[ex resume: coInterpreter].
			ex pass].
	^cogit! !

!Cogit class methodsFor: 'in-image compilation support' stamp: 'eem 3/1/2016 17:57'!
instanceForTests: optionsDictionaryOrArray
	"Initialize all the relevant classes from the options and answer a new instance of me."
	| initOptions |
	initOptions := self asOptionsDictionary: optionsDictionaryOrArray.
	CoInterpreter initializeWithOptions: initOptions.
	CoInterpreter objectMemoryClass initializeWithOptions: initOptions.
	self initializeWithOptions: initOptions.
	^self new! !

!Cogit class methodsFor: 'translation' stamp: 'eem 10/30/2013 15:47'!
isAcceptableAncilliaryClass: aClass
	^aClass ~~ CogSSBytecodeFixup! !

!Cogit class methodsFor: 'translation' stamp: 'eem 5/29/2009 09:56'!
isCogitClass
	^true! !

!Cogit class methodsFor: 'translation' stamp: 'eem 6/11/2015 13:22'!
isNonArgumentImplicitReceiverVariableName: aString
	^#('cogit' 'coInterpreter'
		'methodZone' 'literalsManager'
		'objectMemory' 'objectRepresentation' 'manager') includes: aString! !

!Cogit class methodsFor: 'in-image compilation support' stamp: 'eem 3/22/2017 17:14'!
isSistaMessage: aMessageOrSelector unimplementedIn: aClass
	| selector |
	selector := aMessageOrSelector selector.
	^(InstructionClient includesSelector: selector)
	   and: [(aClass includesSelector: selector) not]! !

!Cogit class methodsFor: 'accessing' stamp: 'cb 8/30/2016 11:02'!
maxNegativeErrorCode
	^MaxNegativeErrorCode! !

!Cogit class methodsFor: 'accessing class hierarchy' stamp: 'eem 3/16/2017 16:52'!
methodZoneClass
	^CogMethodZone! !

!Cogit class methodsFor: 'translation' stamp: 'eem 7/2/2020 15:55'!
monticelloDescription
	"Answer the Monticello version of the packlage containing the receiver.
	 Override to include the package of the activeCompilerClass if it differs."
	| description accDescription |
	description := CCodeGenerator monticelloDescriptionFor: self.
	accDescription := CCodeGenerator monticelloDescriptionFor: self activeCompilerClass.
	^(description allButFirst: (description indexOf: Character space))
		= (accDescription allButFirst: (accDescription indexOf: Character space))
		ifTrue: [description]
		ifFalse: [description, String cr, accDescription]! !

!Cogit class methodsFor: 'translation' stamp: 'eem 7/1/2021 11:24'!
mustBeGlobal: var
	"Answer if a variable must be global and exported.  Used for inst vars that are
	 accessed from VM support code, or that need visibility at the gdb/lldb level."
	^#('ceBaseFrameReturnTrampoline' ceCaptureCStackPointers 'ceCheckForInterruptTrampoline'
		ceEnterCogCodePopReceiverReg realCEEnterCogCodePopReceiverReg
		ceCallCogCodePopReceiverReg realCECallCogCodePopReceiverReg
		ceCallCogCodePopReceiverAndClassRegs realCECallCogCodePopReceiverAndClassRegs
		ceInvokeInterpret 'ceReturnToInterpreterTrampoline' 'ceCannotResumeTrampoline'
		ceTryLockVMOwner
		'cmEntryOffset' 'cmNoCheckEntryOffset' 'cmDynSuperEntryOffset' 'cmSelfSendEntryOffset'
		'missOffset' 'cbEntryOffset' 'cbNoSwitchEntryOffset' 'blockNoContextSwitchOffset' breakPC
		ceGetFP ceGetSP cFramePointerInUse
		methodZoneBase
		traceFlags traceStores)
			includes: var! !

!Cogit class methodsFor: 'translation' stamp: 'eem 2/9/2009 18:55'!
namesOfVariablesToLocalize
	^#()! !

!Cogit class methodsFor: 'documentation' stamp: 'eem 10/16/2010 18:42'!
notesAndQueries
	"Mitigating cost of bytecode to machinecode pc mapping.
		HPS's scheme is to allow contexts to contain machine-code pcs in the form of negative pc values.
		This is reified at the image level and relevant accessors must check before accessing the pc.

		We hide a similar scheme entirely within the VM.  When a spouse context of a machine-code
		frame is divorced we set its pc to the negative machine-code pc.  All accesses to the instruction
		pointer of a stable context are checked (in instVar:ofContext: and externalInstVar:ofContext:).
		If the value is negative we look for, and if not found, compile the context's method and map the
		negative pc.

		The downside of this scheme is on snapshot having to trawl through the entire image looking for
		single contexts with negative pcs and mapping them to proper bytecode pcs.  But this is a minor
		inconvenience since we scan anyway to divorce or widow married contexts."! !

!Cogit class methodsFor: 'translation' stamp: 'eem 11/10/2020 18:55'!
numTrampolineValuesFor: option
	| current values |
	current := VMBasicConstants classPool at: option.
	values := #(true false) collect:
				[:bool|
				 VMBasicConstants classPool at: option put: bool.
				 self cogitClass initializeNumTrampolines.
				 Cogit classPool at: #NumTrampolines].
	VMBasicConstants classPool at: option put: current.
	^values! !

!Cogit class methodsFor: 'trampoline support' stamp: 'eem 11/10/2020 19:24'!
numTrampolines
	^37 "29 + 4 each for self and super sends"
	+ (COGMTVM ifTrue: [1] ifFalse: [0]) "try lock routine"
	+ (LowcodeVM ifTrue: [3] ifFalse: [0])
	+ CogCompilerClass numTrampolines

	"self withAllSubclasses collect: [:c| {c. (c instVarNames select: [:ea| ea beginsWith: 'ce']) size}]"! !

!Cogit class methodsFor: 'translation' stamp: 'eem 12/26/2016 20:41'!
preGenerationHook: aCCodeGenerator
	"Perform any last-minute changes to the code generator immediately
	 before it performs code analysis and generation.  In this case, make
	 all non-exported methods private."
	aCCodeGenerator selectorsAndMethodsDo:
		[:s :m|
		m isAPIMethod
			ifTrue: [m static: false]
			ifFalse:
				[m export ifFalse:
					[m static: true]]]! !

!Cogit class methodsFor: 'translation' stamp: 'eem 7/2/2020 15:37'!
preambleCCode
	^(self activeCompilerClass class whichClassIncludesSelector: #preambleCCode) ifNotNil:
		[self activeCompilerClass preambleCCode]! !

!Cogit class methodsFor: 'translation' stamp: 'eem 1/29/2020 14:37'!
prepareToBeAddedToCodeGenerator: aCodeGen
	"It is either this or scan cmacro methods for selectors."
	BytesPerWord = 8 ifTrue:
		[aCodeGen retainMethods: #(indexForSelector:in:)]! !

!Cogit class methodsFor: 'accessing' stamp: 'eem 11/20/2010 09:18'!
primitiveTable
	^primitiveTable! !

!Cogit class methodsFor: 'translation' stamp: 'eem 4/21/2015 23:08'!
processorSpecificSourceFileName
	^self activeCompilerClass moduleName, '.c'! !

!Cogit class methodsFor: 'translation' stamp: 'eem 12/26/2016 20:46'!
requiredMethodNames: options
	"self requiredMethodNames"
	^self tableFunctions! !

!Cogit class methodsFor: 'documentation' stamp: 'eem 9/25/2015 17:11'!
runtime
	"Generated machine code makes use of a number of run-time routines for support, for executing certain primitives,
	 and for event handling.  These run-time entry points all begin with ce, for ``code entry''.  They are called from
	 trampolines whose job it is to a) switch from the Smalltalk stack to the C stack (because run-time routines are C
	 code in the CoInterpreter or Cogit and run on the C stack, whereas machine code runs on the Smalltalk stack), and
	 b) to marshall the register parameters that trampol,ines take into what ever the platform's ABI requires.
	 See the method trampolines for more info on trampoilines.

	 Here's a doit to collect the signatures of the current run-time routines
 
		(((CoInterpreter selectors select: [:ea| (ea beginsWith: 'ce') and: [ea third isUppercase]]) ,
		(SistaStackToRegisterMappingCogit allSelectors select: [:ea| (ea beginsWith: 'ce') and: [ea third isUppercase]]) asArray) sort collect:
			[:s| | m |
			m := CoInterpreter compiledMethodAt: s ifAbsent: [(SistaStackToRegisterMappingCogit whichClassIncludesSelector: s) >> s].
			s numArgs = 0
				ifTrue: [s asString]
				ifFalse:
					[(String streamContents:
						[:str|
						s keywords with: (CoInterpreter newParser parseParameterNames: m getSource) do:
							[:k :p| str nextPutAll: k; space; nextPutAll: p; nextPutAll: ' (Reg) ']]) allButLast]])


	Run-time:
		ceActivateFailingPrimitiveMethod: aPrimitiveMethod (SendNumArgsReg)
		ceActiveContext => ReceiverResultReg
		ceBaseFrameReturn: returnValue (ReceiverResultReg)
		ceBaseFrameReturnPC
		ceCPICMiss: cPIC (ClassReg) receiver: receiver (ReceiverResultReg)
		ceCall0ArgsPIC
		ceCall1ArgsPIC
		ceCall2ArgsPIC
		ceCannotResume
		ceCannotResumePC
		ceCaptureCStackPointers
		ceCheckFeatures
		ceCheckForAndFollowForwardedPrimitiveState
		ceCheckForInterrupts
		ceCheckProfileTick
		ceClosureCopyDescriptor: descriptor (SendNumArgsReg) => ReceiverResultReg
		ceContext: maybeContext (ReceiverResultReg) instVar: slotIndex (SendNumArgsReg) => ReceiverResultReg
		ceContext: maybeMarriedContext (ReceiverResultReg) instVar: slotIndex (SendNumArgsReg) value: anOop (ClassReg) => ReceiverResultReg
		ceCounterTripped: condition (TempReg)
		ceDynamicSuperSend: cacheAddress (SendNumArgsReg) receiver: methodReceiver (ReceiverResultReg)
		ceEnclosingObjectAt: level (SendNumArgsReg) => ReceiverResultReg
		ceImplicitReceiverSend: cacheAddress (SendNumArgsReg) receiver: methodReceiver (ReceiverResultReg)
		ceInterpretMethodFromPIC: aMethodObj (SendNumArgsReg) receiver: rcvr (ReceiverResultReg)
		ceMNUFromPICMNUMethod: aMethodObj (SendNumArgsReg) receiver: rcvr (ReceiverResultReg)
		ceNewArraySlotSize: slotSize (SendNumArgsReg) => ReceiverResultReg
		ceNonLocalReturn: returnValue (ReceiverResultReg)
		ceOuterSend: cacheAddress (SendNumArgsReg) receiver: methodReceiver (ReceiverResultReg)
		ceReturnToInterpreter: anOop (ReceiverResultReg)
		ceReturnToInterpreterPC
		ceSICMiss: receiver (ReceiverResultReg)
		ceScheduleScavenge
		ceSelfSend: cacheAddress (SendNumArgsReg) receiver: methodReceiver (ReceiverResultReg)
		ceSend: selector (ClassReg) above: startAssociationArg (TempReg) to: rcvr (ReceiverResultReg) numArgs: numArgs (n)
		ceSend: selector (ClassReg) super: superNormalBar (n) to: rcvr (ReceiverResultReg) numArgs: numArgs (n)
		ceSendFromInLineCacheMiss: cogMethodOrPIC (SendNumArgsReg)
		ceSendMustBeBoolean: anObject (TempReg)
		ceSistaTrap
		ceStackOverflow: contextSwitchIfNotNil (SendNumArgsReg)
		ceTraceBlockActivation
		ceTraceLinkedSend: theReceiver (ReceiverResultReg)
		ceTraceStoreOf: aValue (ClassReg) into: anObject (ReceiverResultReg)
		positive32BitIntegerFor: integerValue (ReceiverResultReg) => TempReg

	Enilopmarts:
		ceCallCogCodePopReceiverAndClassRegs
		ceCallCogCodePopReceiverArg0Regs
		ceCallCogCodePopReceiverArg1Arg0Regs
		ceCallCogCodePopReceiverReg
		ceEnterCogCodePopReceiverReg

	Simulation only
		ceShortCutTraceBlockActivation: aProcessorSimulationTrap
		ceShortCutTraceLinkedSend: aProcessorSimulationTrap
		ceShortCutTraceStore: aProcessorSimulationTrap

	Misnamed:
		ceSendAbort:to:numArgs: not a code entry.  Used by ceCannotResume, ceSendMustBeBoolean et al to execute their sends"! !

!Cogit class methodsFor: 'translation' stamp: 'eem 11/11/2020 18:13'!
savingCurrentValuesOf: anArrayOfOptions do: aBlock
	| current |
	current := anArrayOfOptions collect: [:option| VMBasicConstants classPool at: option].
	anArrayOfOptions do: [:option| VMBasicConstants classPool at: option put: false].
	^aBlock ensure:
		[anArrayOfOptions with: current do:
			[:option :value| VMBasicConstants classPool at: option put: value]]! !

!Cogit class methodsFor: 'translation' stamp: 'rmacnak 3/27/2015 22:35'!
shouldGenerateTypedefFor: aStructClass
	"Hack to work-around mutliple definitions.  Sometimes a type has been defined in an include."
	^({ CogBlockMethod. CogMethod. SistaCogMethod. NewspeakCogMethod. NSSendCache } includes: aStructClass) not! !

!Cogit class methodsFor: 'translation' stamp: 'eem 2/9/2009 14:09'!
sourceFileName
	^'cogit.c'! !

!Cogit class methodsFor: 'translation' stamp: 'eem 7/6/2010 15:32'!
specialValueForConstant: constantName default: defaultValue
	constantName = 'Debug' ifTrue: [^'DEBUGVM'].
	^CoInterpreter specialValueForConstant: constantName default: defaultValue! !

!Cogit class methodsFor: 'tests' stamp: 'eem 6/4/2015 08:52'!
structureOfACogMethod
	"A CogMethod is the machine code for executable code in the Cog VM, and in the simulator these are
	 instances of CogMethod.  In actuality they are structures in memory in the CogMethodZone..  There
	 are four real kinds, defined by the cmType field, free space: CMFree, methods: CMMethod, closed
	 PICs: CMClosedPIC (finite polymorphic inline caches with up to 6 entries), and open PICs: CMOpenPIC
	 (infinite megamorphicinline caches that probe the first-level method lookup cache).  There is a fifth
	 kind of method, which is merely a header, for blocks: CMBlock, one which exists only within CMMethods,
	 and exist only to allow block activations to refer to something that looks like a CogMethod.

	 The blockSize field in a CogMethod is the size in bytes of the entire method, including the header.
	 Methods are aligned to an 8 byte boundary in the CogMethodZone.  The size is used to iterate over
	 the methods in the zone.

	 Follwing the header is the abort and entry code.  Starting immediately after the header is the call to
	 the abort routine called when either a send fails or a stack limit check fails.  Following that is the
	 checked entry point that checks the receiver is of the right class, and this code ends in the unchecked
	 entry point.  Following this is either primitive code, followed by frame building code, or frame-building
	 code, or, for frameless methods, the code for the frameless method.  Following that is the code for the
	 method.  If the method contains blocks  then followng the method code will be a CMBlock header,
	 followed by code for the block, for each block, and following this will be the block dispatch code,
	 which is indexed by the blockEntryOffset field in the CogMethod.

	 Following either the return in the method, or the block dispatch, is the method map, the meta data which
	 identifies interesting points in the machine code.  The map starts at the end of the structure and is read
	 backwards towards the last instruftion of the method, and is terminated by a null byte.  So the blockSize
	 is used to find the start of the map.  The map reveals where object references, sends and pc-mapping
	 points exist in the machine code.  The map is parsed when garbage collecting to find and update object
	 references, and when unlinking sends for method cache flushing, and to convert between bytecode and
	 machine code pcs by scanning both bytecode and machine code looking for matching points in the map."! !

!Cogit class methodsFor: 'translation' stamp: 'eem 1/8/2016 15:57'!
tableFunctions
	"self tableFunctions"
	self cogitClass ~= self ifTrue: [^Set new].
	generatorTable ifNil:
		[^Set new].
	^Set new
		addAll: (generatorTable object
					collect: [:ea| ea generator]);
		addAll: (generatorTable object
					select: [:ea| ea spanFunction notNil]
					thenCollect: [:ea| ea spanFunction]);
		addAll: (generatorTable object
					select: [:ea| ea needsFrameFunction notNil]
					thenCollect: [:ea| ea needsFrameFunction]);
		addAll: (primitiveTable object
					select: [:ea| ea notNil and: [ea primitiveGenerator notNil]]
					thenCollect: [:ea| ea primitiveGenerator]);
		asSortedCollection: [:a :b| a caseInsensitiveLessOrEqual: b]! !

!Cogit class methodsFor: 'translation' stamp: 'eem 4/20/2020 17:48'!
tableInitializerFor: aTable in: aCCodeGenerator
	^String streamContents:
		[:s|
		s nextPutAll: ' = {'.
		aTable object
			do: [:gt|
				s crtab.
				gt printCInitializerOn: s in: aCCodeGenerator]
			separatedBy: [s peekLast == $} ifTrue: [s nextPut: $,]].
		s cr; nextPut: $}]! !

!Cogit class methodsFor: 'tests' stamp: 'eem 3/18/2013 11:12'!
testPCMappingFor: aCompiledMethod
	^self testPCMappingFor: aCompiledMethod options: #()! !

!Cogit class methodsFor: 'tests' stamp: 'cb 4/11/2016 15:38'!
testPCMappingFor: aCompiledMethod options: optionsDictionaryOrArray
	| tuple |
	tuple := self cog: aCompiledMethod selectorOrNumCopied: aCompiledMethod selector options: optionsDictionaryOrArray.
	tuple second testPCMappingForCompiledMethod: aCompiledMethod cogMethod: tuple last! !

!Cogit class methodsFor: 'tests' stamp: 'eem 3/18/2013 11:13'!
testPCMappingSelect: aBlock
	"Test pc mapping both ways using the methods in the current image"
	self testPCMappingSelect: aBlock options: #()! !

!Cogit class methodsFor: 'tests' stamp: 'eem 9/18/2019 15:02'!
testPCMappingSelect: aBlock options: optionsDictionaryOrArray
	"Test pc mapping both ways using a selection of the methods in the current image."
	| cogit coInterpreter n |
	cogit := self initializedInstanceForTests: optionsDictionaryOrArray.
	coInterpreter := cogit coInterpreter.
	n := -1.
	SystemNavigation new allSelect:
		[:m| | retrym cm |
		 (m isQuick not
		  and: [aBlock value: m]) ifTrue:
			[(n := n + 1) \\ 10 = 0 ifTrue: [Transcript nextPut: $.; flush].
			 retrym := true.
			 [(cm := cogit
						cog: (coInterpreter oopForObject: m)
						selector: (coInterpreter oopForObject: m selector)) isNil and: [retrym]] whileTrue:
				[retrym := false.
				 cogit methodZone clearCogCompiledCode.
				 coInterpreter initializeObjectMap].
			  cm
				ifNil: [Transcript halt show: 'After 1 Cog compiled code compaction, still not able to generate the cog method...']
				ifNotNil: [cogit testPCMappingForCompiledMethod: m cogMethod: cm].
			 m encoderClass supportsFullBlocks ifTrue:
				[m nestedBlockMethods do:
					[:bm| | retryb cbm |			
					 (n := n + 1) \\ 10 = 0 ifTrue: [Transcript nextPut: $,; flush].
					 retryb := true.
					 [(cbm := cogit
								cogFullBlockMethod: (coInterpreter oopForObject: bm)
								numCopied: bm numCopiedValues) isNil and: [retryb]] whileTrue:
						[retryb := false.
						 cogit methodZone clearCogCompiledCode.
						 coInterpreter initializeObjectMap].
					  cbm
						ifNil: [Transcript show: 'After 1 Cog compiled code compaction, still not able to generate the cog block...']
						ifNotNil: [cogit testPCMappingForCompiledMethod: bm cogMethod: cbm]]]].
		 false]! !

!Cogit class methodsFor: 'documentation' stamp: 'nice 11/16/2016 00:49'!
trampolines
	"Trampolines are called from machine-code for a number of tasks (e.g. doing an unlinked send,
	 creating a closure, doing a non-local return, etc, etc.
	 Some trampolines save their return pc on the Smalltalk stack, and some pop it into instructionPointer,
	 depending on whether the trampoline always returns or whether control might transfer elsewhere,
	 or depending on whether the callee run-time routine expects to see a normal Smalltalk stack.  This
	 documents which trampolines do what and why.

ceSend0Args .. ceSendNArgs, ceSuperSend0Args..ceSuperSendNArgs
	return pc left on stack because it is the return pc of the send and must be there

ceMethodAbort, cePICAbort
	return pc left on stack because it is used to locate the failing send target method/PIC of a linked send

ceClosureCopyTrampoline
	return pc left on stack because ceClosureCopy:... always returns directly

cePushActiveContextTrampoline
	return pc left on stack because cePushActiveContext always returns directly

ceNonLocalReturnTrampoline
	pops return pc into instructionPointer

ceBaseFrameReturnTrampoline
	this is returned-to and never called; i.e. it is the return pc of a base machine-code frame

ceCreateNewArrayTrampoline
	return pc left on stack because ceNewArraySlotSize always returns directly

ceSendMustBeBooleanTrampoline
	return pc left on stack because it is the return pc of the send and must be there

ceCheckForInterruptsTrampoline
	pops return pc into instructionPointer

ceCPICMissTrampoline
	return pc left on stack because it is used to locate the failing send target PIC of a linked send

ceSendFromInLineCacheMissTrampoline
	not used, but its address is used as a key in the simulator.  An open PIC calls ceSendFromInLineCacheMiss
	after it has switched from the Smalltalk to the C stack (an inline trampoline) hence there is no return pc
	left on the Smalltalk stack (except that of the send that invoked the open PIC).

ceStoreCheckTrampoline
	return pc left on stack because ceStoreCheck always returns directly

ceFetchContextInstVarTrampoline
	return pc left on stack because ceContextInstVar always returns directly.
	ceContextInstVar manages popping the pc into instructionPointer if necessary.

ceStoreContextInstVarTrampoline
	return pc left on stack because ceContextInstVarvalue always returns directly.
	ceContextInstVarvalue manages popping the pc into instructionPointer if necessary.

cePositive32BitIntegerTrampoline
	return pc left on stack because cePositive32BitInteger always returns directly
	
ceReturnToInterpreterTrampoline
	this is returned-to and never called; i.e. it is the return pc of a machine-code frame with an interpreted callee

ceResendCannotReturnTrampoline
	this is returned-to and never called; it is the pc for a machine-code frame which has been returned from

ceEnterCogCodePopReceiverReg
	this is an enilopmart and not called from machine code

cePrimReturnEnterCogCode
	this is an enilopmart and not called from machine code

ceTraceLinkedSendTrampoline
	return pc left on stack because ceTraceLinkedSend always returns directly

ceTraceBlockActivationTrampoline
	return pc left on stack because ceTraceBlockActivation always returns directly

ceTraceStoreTrampoline
	return pc left on stack because ceTraceStore always returns directly"! !

!Cogit class methodsFor: 'translation' stamp: 'eem 6/7/2018 15:18'!
translateableInstructionSubclassesAndInstalledOptionsDo: aBlock
	"Evaluate aBlock with the translateable subclass and its options installed, being careful to clean-up afterwards."
	CogAbstractInstruction translateableSubclassesAndOptions do:
		[:pair|
		[:class :options| | toRemove |
		 toRemove := Set new.
		 options pairsDo:
			[:key :value|
			 (InitializationOptions includesKey: key) ifFalse:
				[toRemove add: key].
			 InitializationOptions at: key put: value].
		 aBlock value: class.
		 toRemove do: [:key| InitializationOptions removeKey: key]]
			valueWithArguments: pair]! !

!Cogit class methodsFor: 'translation' stamp: 'eem 2/9/2009 20:42'!
typeForSelf
	^#implicit! !

!Cogit class methodsFor: 'translation' stamp: 'eem 1/5/2021 15:52'!
wantsSqDotH
	^false! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:11'!
AddCq: quickConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: AddCqR quickConstant: quickConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/21/2020 11:02'!
AddCq: quickConstant R: srcReg R: destReg
	"N.B. if the condition codes don't require setting and three address arithmetic
	 is unavailable, then LoadEffectiveAddressMw:r:R: can be used instead."
	<inline: false>
	<returnTypeC: #'AbstractInstruction *'>
	| first |
	backEnd hasThreeAddressArithmetic ifTrue:
		[^self gen: AddCqRR quickConstant: quickConstant operand: srcReg operand: destReg].
	srcReg = destReg ifTrue:
		[^self gen: AddCqR quickConstant: quickConstant operand: destReg].
	first := self gen: MoveRR operand: srcReg operand: destReg.
	self gen: AddCqR quickConstant: quickConstant operand: destReg.
	^first! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:06'!
AddCw: wordConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: AddCwR literal: wordConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
AddR: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: AddRR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/9/2020 07:18'!
AddR: addendReg R: badendReg R: destReg
	"destReg := addendReg + badendReg"
	<inline: false>
	<returnTypeC: #'AbstractInstruction *'>
	| first |
	backEnd hasThreeAddressArithmetic ifTrue:
		[^self gen: AddRRR operand: addendReg operand: badendReg operand: destReg].
	self assert: badendReg ~= destReg.
	first := self gen: MoveRR operand: addendReg operand: destReg.
	self gen: AddRR operand: badendReg operand: destReg.
	^first! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/16/2009 09:03'!
AddRd: dpreg1 Rd: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: AddRdRd operand: dpreg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:08'!
AddRs: dpreg1 Rs: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: AddRsRs operand: dpreg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/30/2016 17:13'!
AddcCq: quickConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: AddcCqR quickConstant: quickConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:08'!
AddcR: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: AddcRR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
AlignmentNops: alignment
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: AlignmentNops operand: alignment! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:11'!
AndCq: quickConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: AndCqR quickConstant: quickConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/9/2020 07:17'!
AndCq: quickConstant R: srcReg R: destReg
	<inline: false>
	<returnTypeC: #'AbstractInstruction *'>
	| first |
	backEnd hasThreeAddressArithmetic ifTrue:
		[^self gen: AndCqRR quickConstant: quickConstant operand: srcReg operand: destReg].
	srcReg = destReg ifTrue:
		[^self gen: AndCqR quickConstant: quickConstant operand: destReg].
	first := self gen: MoveRR operand: srcReg operand: destReg.
	self gen: AndCqR quickConstant: quickConstant operand: destReg.
	^first! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:07'!
AndCw: wordConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: AndCwR literal: wordConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
AndR: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: AndRR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
ArithmeticShiftRightCq: quickConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: ArithmeticShiftRightCqR operand: quickConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/9/2020 07:18'!
ArithmeticShiftRightCq: quickConstant R: srcReg R: destReg
	"destReg := (signed)srcReg >> quickConstant"
	<inline: false>
	<returnTypeC: #'AbstractInstruction *'>
	| first |
	backEnd hasThreeAddressArithmetic ifTrue:
		[^self gen: ArithmeticShiftRightCqRR operand: quickConstant operand: srcReg operand: destReg].
	first := self gen: MoveRR operand: srcReg operand: destReg.
	self gen: ArithmeticShiftRightCqR operand: quickConstant operand: destReg.
	^first! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
ArithmeticShiftRightR: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: ArithmeticShiftRightRR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/11/2009 20:26'!
Call: callTarget 
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: Call operand: callTarget! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:05'!
CallFull: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: CallFull literal: jumpTarget! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/20/2015 18:23'!
CallFullRT: callTarget
	"Big assumption here that calls and jumps look the same as regards their displacement.
	 This works on at least x86, ARM and x86_64.
	 CallFull is intended to be for calls anywhere in our address space.
	 See also Call which calls within our *code* space"
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^backEnd fullCallsAreRelative
		ifTrue: [self annotateCall: (self CallFull: callTarget)]
		ifFalse: [self CallFull: callTarget]! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'nice 5/19/2017 14:33'!
CallFullRT: callTarget registersToBeSavedMask: registersToBeSaved
	<returnTypeC: #'AbstractInstruction *'>
	| callerSavedRegsToBeSaved lastInst reg registersToBePushed |
	<var: 'lastInst' type: #'AbstractInstruction *'>
	callerSavedRegsToBeSaved := CallerSavedRegisterMask bitAnd: registersToBeSaved.

	backEnd canPushPopMultipleRegisters
		ifTrue: [backEnd genPushRegisterMask: callerSavedRegsToBeSaved]
		ifFalse:
			[registersToBePushed := callerSavedRegsToBeSaved.
			 reg := 0.
			 [registersToBePushed ~= 0] whileTrue:
				[(registersToBePushed anyMask: 1) ifTrue:
					[self PushR: reg].
				 reg := reg + 1.
				 registersToBePushed := registersToBePushed >>> 1]].

	lastInst := self CallFullRT: callTarget.

	backEnd canPushPopMultipleRegisters
		ifTrue: [^backEnd genPopRegisterMask: callerSavedRegsToBeSaved]
		ifFalse:
			[[reg > 0] whileTrue:
				[reg := reg - 1.
				(callerSavedRegsToBeSaved anyMask: 1 << reg) ifTrue:
					[lastInst := self PopR: reg]].

			 ^lastInst]! !

!Cogit methodsFor: 'method map' stamp: 'eem 6/23/2011 09:42'!
CallNewspeakSend: callTarget
	<returnTypeC: #'AbstractInstruction *'>
	^self annotateNewspeakSend: (self Call: callTarget)! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:08'!
CallR: callTarget 
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: CallR operand: callTarget! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/19/2015 11:23'!
CallRT: callTarget
	"Big assumption here that calls and jumps look the same as regards their displacement.
	 This works on at least x86, ARM and x86_64."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self annotateCall: (self Call: callTarget)! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'nice 5/19/2017 14:33'!
CallRT: callTarget registersToBeSavedMask: registersToBeSaved
	<returnTypeC: #'AbstractInstruction *'>
	| callerSavedRegsToBeSaved lastInst reg registersToBePushed |
	<var: 'lastInst' type: #'AbstractInstruction *'>
	callerSavedRegsToBeSaved := CallerSavedRegisterMask bitAnd: registersToBeSaved.

	backEnd canPushPopMultipleRegisters
		ifTrue: [backEnd genPushRegisterMask: callerSavedRegsToBeSaved]
		ifFalse:
			[registersToBePushed := callerSavedRegsToBeSaved.
			 reg := 0.
			 [registersToBePushed ~= 0] whileTrue:
				[(registersToBePushed anyMask: 1) ifTrue:
					[self PushR: reg].
				 reg := reg + 1.
				 registersToBePushed := registersToBePushed >>> 1]].

	lastInst := self CallRT: callTarget.

	backEnd canPushPopMultipleRegisters
		ifTrue: [^backEnd genPopRegisterMask: callerSavedRegsToBeSaved]
		ifFalse:
			[[reg > 0] whileTrue:
				[reg := reg - 1.
				(callerSavedRegsToBeSaved anyMask: 1 << reg) ifTrue:
					[lastInst := self PopR: reg]].

			 ^lastInst]! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/23/2020 15:57'!
ClzR: reg1 R: reg2
	"reg2 := reg1 countLeadingZeros"
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: ClzRR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 12/17/2019 12:54'!
CmpC32: wordConstant R: reg
	"Generate a CmpC32R instruction to compare a 32-bit constant with a
	 register.  If this is a 32-bit platform, simply generate a CmpCwR instruction,
	 to avoid needless duplication in the 32-bit code generators.."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self
		gen: (objectMemory wordSize = 8
				ifTrue: [CmpC32R]
				ifFalse: [CmpCwR])
		literal32: wordConstant
		operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:11'!
CmpCq: quickConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: CmpCqR quickConstant: quickConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:07'!
CmpCw: wordConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: CmpCwR literal: wordConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 12/19/2019 13:35'!
CmpR: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	self deny: reg1 = SPReg. "See comment in CogRTLOpcodes class initialize"
	^self gen: CmpRR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/16/2009 09:03'!
CmpRd: dpreg1 Rd: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: CmpRdRd operand: dpreg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:09'!
CmpRs: dpreg1 Rs: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: CmpRsRs operand: dpreg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/16/2009 15:25'!
ConvertR: reg1 Rd: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: ConvertRRd operand: reg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:09'!
ConvertR: reg1 Rs: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: ConvertRRs operand: reg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:09'!
ConvertRd: reg1 R: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: ConvertRdR operand: reg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:09'!
ConvertRd: reg1 Rs: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: ConvertRdRs operand: reg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:09'!
ConvertRs: reg1 R: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: ConvertRsR operand: reg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:09'!
ConvertRs: reg1 Rd: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: ConvertRsRd operand: reg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 9/10/2009 16:43'!
DivR: rDivisor R: rDividend Quo: rQuotient Rem: rRemainder
	"Division is a little weird on some processors.  Defer to the backEnd
	 to allow it to generate any special code it may need to."
	<returnTypeC: #'AbstractInstruction *'>
	<inline: false>
	backEnd genDivR: rDivisor R: rDividend Quo: rQuotient Rem: rRemainder.
	^self abstractInstructionAt: opcodeIndex - 1! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/16/2009 09:03'!
DivRd: dpreg1 Rd: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: DivRdRd operand: dpreg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:09'!
DivRs: dpreg1 Rs: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: DivRsRs operand: dpreg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/8/2011 20:46'!
Fill32: value
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: Fill32 operand: value! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/11/2009 20:23'!
Jump: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self gen: Jump operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:20:56'!
JumpAbove: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpAbove operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:20:50'!
JumpAboveOrEqual: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpAboveOrEqual operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:20:45'!
JumpBelow: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpBelow operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:20:41'!
JumpBelowOrEqual: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpBelowOrEqual operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:20:36'!
JumpCarry: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpCarry operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/17/2009 15:45'!
JumpFPEqual: jumpTarget
	"Floating-point jumps are a little weird on some processors.  Defer to
	 the backEnd to allow it to generate any special code it may need to."
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	<inline: false>
	^backEnd genJumpFPEqual: jumpTarget! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/17/2009 15:45'!
JumpFPGreater: jumpTarget
	"Floating-point jumps are a little weird on some processors.  Defer to
	 the backEnd to allow it to generate any special code it may need to."
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	<inline: false>
	^backEnd genJumpFPGreater: jumpTarget! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/17/2009 15:45'!
JumpFPGreaterOrEqual: jumpTarget
	"Floating-point jumps are a little weird on some processors.  Defer to
	 the backEnd to allow it to generate any special code it may need to."
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	<inline: false>
	^backEnd genJumpFPGreaterOrEqual: jumpTarget! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/17/2009 15:46'!
JumpFPLess: jumpTarget
	"Floating-point jumps are a little weird on some processors.  Defer to
	 the backEnd to allow it to generate any special code it may need to."
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	<inline: false>
	^backEnd genJumpFPLess: jumpTarget! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/17/2009 15:45'!
JumpFPLessOrEqual: jumpTarget
	"Floating-point jumps are a little weird on some processors.  Defer to
	 the backEnd to allow it to generate any special code it may need to."
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	<inline: false>
	^backEnd genJumpFPLessOrEqual: jumpTarget! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/17/2009 15:46'!
JumpFPNotEqual: jumpTarget
	"Floating-point jumps are a little weird on some processors.  Defer to
	 the backEnd to allow it to generate any special code it may need to."
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	<inline: false>
	^backEnd genJumpFPNotEqual: jumpTarget! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 8/11/2015 11:43'!
JumpFull: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: JumpFull literal: jumpTarget asInteger! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/20/2015 18:23'!
JumpFullRT: callTarget
	"Big assumption here that calls and jumps look the same as regards their displacement.
	 This works on at least x86, ARM and x86_64.
	 JumpFull is intended to be for jumps anywhere in our address space.
	 See also JumpLong et al. which jump within our *code* space"
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^backEnd fullCallsAreRelative
		ifTrue: [self annotateCall: (self JumpFull: callTarget)]
		ifFalse: [self JumpFull: callTarget]! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:20:26'!
JumpGreater: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpGreater operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:20:20'!
JumpGreaterOrEqual: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpGreaterOrEqual operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:20:14'!
JumpLess: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpLess operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:20:09'!
JumpLessOrEqual: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpLessOrEqual operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 8/9/2011 16:51'!
JumpLong: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: JumpLong operand: jumpTarget! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:20:02'!
JumpLongNonZero: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self genConditionalBranch: JumpLongNonZero operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:19:55'!
JumpLongZero: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self genConditionalBranch: JumpLongZero operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:19:51'!
JumpNegative: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpNegative operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:18:10'!
JumpNoCarry: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpNoCarry operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:18:26'!
JumpNoOverflow: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpNoOverflow operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:19:43'!
JumpNonNegative: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpNonNegative operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:19:39'!
JumpNonZero: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpNonZero operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:19:33'!
JumpOverflow: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpOverflow operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/14/2009 17:56'!
JumpR: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: JumpR operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rmacnak 11/23/2015 20:19:15'!
JumpZero: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genConditionalBranch: JumpZero operand: jumpTarget asInteger! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 11/2/2012 10:46'!
Label
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: Label operand: (labelCounter := labelCounter + 1) operand: bytecodePC! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 17:48'!
LoadEffectiveAddressMw: offset r: baseReg R: destReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: LoadEffectiveAddressMwrR quickConstant: offset operand: baseReg operand: destReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
LogicalShiftLeftCq: quickConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: LogicalShiftLeftCqR operand: quickConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/9/2020 07:20'!
LogicalShiftLeftCq: quickConstant R: srcReg R: destReg
	"destReg := srcReg << quickConstant"
	<inline: false>
	<returnTypeC: #'AbstractInstruction *'>
	| first |
	backEnd hasThreeAddressArithmetic ifTrue:
		[^self gen: LogicalShiftLeftCqRR operand: quickConstant operand: srcReg operand: destReg].
	first := self gen: MoveRR operand: srcReg operand: destReg.
	self gen: LogicalShiftLeftCqR operand: quickConstant operand: destReg.
	^first! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
LogicalShiftLeftR: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: LogicalShiftLeftRR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
LogicalShiftRightCq: quickConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: LogicalShiftRightCqR operand: quickConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/9/2020 07:20'!
LogicalShiftRightCq: quickConstant R: srcReg R: destReg
	"destReg := (unsigned)srcReg >> quickConstant"
	<inline: false>
	<returnTypeC: #'AbstractInstruction *'>
	| first |
	backEnd hasThreeAddressArithmetic ifTrue:
		[^self gen: LogicalShiftRightCqRR operand: quickConstant operand: srcReg operand: destReg].
	first := self gen: MoveRR operand: srcReg operand: destReg.
	self gen: LogicalShiftRightCqR operand: quickConstant operand: destReg.
	^first! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
LogicalShiftRightR: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: LogicalShiftRightRR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 9/7/2016 15:00'!
MoveA32: address R: reg 
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: (objectMemory wordSize = 8
					ifTrue: [MoveA32R]
					ifFalse: [MoveAwR])
		literal: address
		operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 10/31/2015 15:56:52'!
MoveAb: address R: reg 
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveAbR literal: address operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 10/9/2019 14:37'!
MoveAw: address R: reg
	"N.B. On certain targets (including X64) this instruction may smash TempReg if the target reg is either FPReg or SPReg."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveAwR literal: address operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 12/17/2019 12:55'!
MoveC32: wordConstant R: reg
	"Generate a MoveC32R instruction to move a 32-bit constant into a register.
	 If this is a 32-bit platform, simply generate a MoveCwR instruction, to avoid
	 needless duplication in the 32-bit code generators.."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self
		gen: (objectMemory wordSize = 8
				ifTrue: [MoveC32R]
				ifFalse: [MoveCwR])
		literal32: wordConstant
		operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 5/2/2016 15:49'!
MoveCf32: constantFloat32 Rs: register
	<inline: true>
	<var: #constantFloat32 type: #float>
	<returnTypeC: #'AbstractInstruction *'>
	^ backEnd genMoveCf32: constantFloat32 Rs: register! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 5/2/2016 15:49'!
MoveCf64: constantFloat64 Rd: register
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #constantFloat64 type: #double>
	^ backEnd genMoveCf64: constantFloat64 Rd: register! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:12'!
MoveCq: quickConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveCqR quickConstant: quickConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:07'!
MoveCw: wordConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveCwR literal: wordConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:11'!
MoveM16: offset r: baseReg R: destReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveM16rR quickConstant: offset operand: baseReg operand: destReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/30/2016 17:15'!
MoveM32: offset r: baseReg R: destReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveM32rR quickConstant: offset operand: baseReg operand: destReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/30/2016 17:15'!
MoveM32: offset r: baseReg Rs: destDPReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveM32rRs quickConstant: offset operand: baseReg operand: destDPReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rsf 1/13/2017 04:06'!
MoveM64: offset r: baseReg R: destReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	self assert: BytesPerWord = 8.
	^self MoveMw: offset r: baseReg R: destReg
! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:11'!
MoveM64: offset r: baseReg Rd: destDPReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveM64rRd quickConstant: offset operand: baseReg operand: destDPReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/30/2016 17:15'!
MoveM8: offset r: baseReg R: destReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveM8rR quickConstant: offset operand: baseReg operand: destReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:17'!
MoveMb: offset r: baseReg R: destReg
	"N.B.  This instruction is guaranteed to zero-extend the byte into destReg."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveMbrR quickConstant: offset operand: baseReg operand: destReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:16'!
MoveMw: offset r: baseReg R: destReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveMwrR quickConstant: offset operand: baseReg operand: destReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 8/8/2021 10:45'!
MovePerfCnt64R: destReg L: liveRegisterMask
	<returnTypeC: #'AbstractInstruction *'>
	self assert: (backEnd has64BitPerformanceCounter and: [objectMemory wordSize = 8]).
	^self gen: MovePerfCnt64RL operand: destReg operand: liveRegisterMask! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 8/8/2021 10:46'!
MovePerfCnt64R: destRegLo R: destRegHi L: liveRegisterMask
	<returnTypeC: #'AbstractInstruction *'>
	self assert: (backEnd has64BitPerformanceCounter and: [objectMemory wordSize = 4]).
	^self gen: MovePerfCnt64RRL operand: destRegLo operand: destRegHi operand: liveRegisterMask! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 9/7/2016 15:01'!
MoveR: reg A32: address
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: (objectMemory wordSize = 8
					ifTrue: [MoveRA32]
					ifFalse: [MoveRAw])
		operand: reg
		literal: address! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 10/31/2015 15:56:41'!
MoveR: reg Ab: address
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRAb operand: reg literal: address! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 10/9/2019 14:37'!
MoveR: reg Aw: address
	"N.B. On certain targets (including X64) this instruction may smash TempReg if the source reg is either FPReg or SPReg."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRAw operand: reg literal: address! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/11/2016 17:26'!
MoveR: srcReg M16: offset r: baseReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRM16r operand: srcReg quickConstant: offset operand: baseReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/30/2016 17:16'!
MoveR: sourceReg M32: offset r: baseReg 
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRM32r operand: sourceReg quickConstant: offset operand: baseReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'rsf 1/13/2017 04:06'!
MoveR: sourceReg M64: offset r: baseReg 
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	self assert: BytesPerWord = 8.
	^self MoveR: sourceReg Mw: offset r: baseReg ! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/30/2016 17:16'!
MoveR: sourceReg M8: offset r: baseReg 
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRM8r operand: sourceReg quickConstant: offset operand: baseReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:15'!
MoveR: sourceReg Mb: offset r: baseReg 
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRMbr operand: sourceReg quickConstant: offset operand: baseReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:15'!
MoveR: sourceReg Mw: offset r: baseReg 
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRMwr operand: sourceReg quickConstant: offset operand: baseReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
MoveR: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 12/1/2015 14:38'!
MoveR: srcReg Rd: destDPReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	self assert: objectMemory wordSize = 8.
	^self gen: MoveRRd operand: srcReg operand: destDPReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/11/2016 14:23'!
MoveR: sourceReg X32r: indexReg R: baseReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRX32rR operand: sourceReg operand: indexReg operand: baseReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 12/9/2013 15:26'!
MoveR: srcReg Xbr: indexReg R: baseReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRXbrR operand: srcReg operand: indexReg operand: baseReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/26/2010 12:17'!
MoveR: sourceReg Xwr: indexReg R: baseReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRXwrR operand: sourceReg operand: indexReg operand: baseReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:16'!
MoveRd: sourceDPReg M64: offset r: baseReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRdM64r operand: sourceDPReg quickConstant: offset operand: baseReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 12/1/2015 14:38'!
MoveRd: srcDPReg R: destReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	self assert: objectMemory wordSize = 8.
	^self gen: MoveRdR operand: srcDPReg operand: destReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/16/2009 09:10'!
MoveRd: dpreg1 Rd: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRdRd operand: dpreg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/30/2016 17:16'!
MoveRs: sourceDPReg M32: offset r: baseReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRsM32r operand: sourceDPReg quickConstant: offset operand: baseReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:10'!
MoveRs: dpreg1 Rs: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveRsRs operand: dpreg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 12/17/2019 13:46'!
MoveUniqueC32: wordConstant R: reg
	"Generate a MoveC32R instruction to move a 32-bit constant into a register.
	 If the backEnd is using out-of-line literals then those for inline caches cannot be shared,
	 and this method ensures the instruction has its own unique label.  If the backEnd is using
	 in-line literals then the literal is unique anyway and this is equivalent to MoveC32:R:.
	 If this is a 32-bit platform, simply generate a MoveCwR instruction, to avoid
	 needless duplication in the 32-bit code generators.."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self
		gen: (objectMemory wordSize = 8
				ifTrue: [MoveC32R]
				ifFalse: [MoveCwR])
		uniqueLiteral32: wordConstant
		operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/10/2015 19:36'!
MoveUniqueCw: wordConstant R: reg
	"If the backEnd is using out-of-line literals then those for inline caches cannot be shared, and
	 this method ensures the instruction has its own unique label.  If the backEnd is using
	 in-line literals then the literal is unique anyway and this is equivalent to MoveCw:R:"
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveCwR uniqueLiteral: wordConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 11/4/2015 11:53:19'!
MoveX32r: indexReg R: baseReg R: destReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveX32rRR operand: indexReg operand: baseReg operand: destReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
MoveXbr: indexReg R: baseReg R: destReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveXbrRR operand: indexReg operand: baseReg operand: destReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
MoveXwr: indexReg R: baseReg R: destReg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MoveXwrRR operand: indexReg operand: baseReg operand: destReg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 7/4/2020 13:35'!
MulOverflowR: reg1 R: reg2
	"Multiplication is a little weird on some processors.  Defer to the backEnd
	 to allow it to generate any special code it may need to."
	<inline: true>
	^backEnd canMulRRDetectOverflow
		ifTrue: [backEnd genMulR: reg1 R: reg2]
		ifFalse: [backEnd genMulOverflowR: reg1 R: reg2]! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 12/16/2015 09:44'!
MulR: reg1 R: reg2
	"Multiplication is a little weird on some processors.  Defer to the backEnd
	 to allow it to generate any special code it may need to."
	<inline: true>
	^backEnd genMulR: reg1 R: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/16/2009 09:03'!
MulRd: dpreg1 Rd: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MulRdRd operand: dpreg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:10'!
MulRs: dpreg1 Rs: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: MulRsRs operand: dpreg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/2/2020 18:49'!
NativePopR: reg
	^self gen: (NativeSPReg = SPReg ifTrue: [PopR] ifFalse: [NativePopR]) operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/2/2020 18:48'!
NativePushR: reg
	^self gen: (NativeSPReg = SPReg ifTrue: [PushR] ifFalse: [NativePushR]) operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/1/2020 13:09'!
NativeRetN: offset
	^self gen: (NativeSPReg = SPReg ifTrue: [RetN] ifFalse: [NativeRetN]) operand: offset! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
NegateR: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: NegateR operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 12/24/2010 09:43'!
Nop
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: Nop! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:11'!
NotR: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: NotR operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:12'!
OrCq: quickConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: OrCqR quickConstant: quickConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/23/2020 16:14'!
OrCq: quickConstant R: srcReg R: destReg
	<inline: false>
	<returnTypeC: #'AbstractInstruction *'>
	| first |
	backEnd hasThreeAddressArithmetic ifTrue:
		[^self gen: OrCqRR quickConstant: quickConstant operand: srcReg operand: destReg].
	srcReg = destReg ifTrue:
		[^self gen: OrCqR quickConstant: quickConstant operand: destReg].
	first := self gen: MoveRR operand: srcReg operand: destReg.
	self gen: OrCqR quickConstant: quickConstant operand: destReg.
	^first! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:07'!
OrCw: wordConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: OrCwR literal: wordConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
OrR: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: OrRR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 5/6/2009 10:32'!
PopR: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: PopR operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 5/2/2016 13:03'!
PopRd: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^ backEnd genPopRd: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 5/2/2016 13:03'!
PopRs: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^ backEnd genPopRs: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 16:46'!
PrefetchAw: address
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: PrefetchAw literal: address! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 5/2/2016 16:08'!
PushC64: constant64Bits
	<inline: true>
	<var: #constant64Bits type: #sqLong>
	<returnTypeC: #'AbstractInstruction *'>
	^ backEnd genPushC64: constant64Bits! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:13'!
PushCq: wordConstant
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: PushCq quickConstant: wordConstant! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:05'!
PushCw: wordConstant
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: PushCw literal: wordConstant! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 5/6/2009 10:37'!
PushR: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: PushR operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:11'!
PushRd: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^ backEnd genPushRd: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:11'!
PushRs: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^ backEnd genPushRs: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
RetN: offset
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: RetN operand: offset! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 12/9/2015 14:02'!
RotateLeftCq: quickConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: RotateLeftCqR operand: quickConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 12/7/2015 22:35'!
RotateRightCq: quickConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: RotateRightCqR operand: quickConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/21/2018 12:27'!
SignExtend16R: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	backEnd canSignExtend
		ifTrue: [^self gen: SignExtend16RR operand: reg1 operand: reg2]
		ifFalse:
			[| first |
			 reg1 = reg2
				ifTrue:
					[first := self LogicalShiftLeftCq: BytesPerWord * 8 - 16 R: reg1]
				ifFalse:
					[first := self MoveR: reg1 R: reg2.
					 self LogicalShiftLeftCq: BytesPerWord * 8 - 16 R: reg1].
			self ArithmeticShiftRightCq: BytesPerWord * 8 - 16 R: reg2.
			^first]! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/21/2018 12:26'!
SignExtend32R: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	backEnd canSignExtend
		ifTrue: [^self gen: SignExtend32RR operand: reg1 operand: reg2]
		ifFalse:
			[| first |
			 reg1 = reg2
				ifTrue:
					[first := self LogicalShiftLeftCq: 32 R: reg1]
				ifFalse:
					[first := self MoveR: reg1 R: reg2.
					 self LogicalShiftLeftCq: 32 R: reg1].
			self ArithmeticShiftRightCq: 32 R: reg2.
			^first]! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/21/2018 12:27'!
SignExtend8R: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	backEnd canSignExtend
		ifTrue: [^self gen: SignExtend8RR operand: reg1 operand: reg2]
		ifFalse:
			[| first |
			 reg1 = reg2
				ifTrue:
					[first := self LogicalShiftLeftCq: BytesPerWord * 8 - 8 R: reg1]
				ifFalse:
					[first := self MoveR: reg1 R: reg2.
					 self LogicalShiftLeftCq: BytesPerWord * 8 - 8 R: reg1].
			self ArithmeticShiftRightCq: BytesPerWord * 8 - 8 R: reg2.
			^first]! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/21/2009 10:44'!
SqrtRd: dpreg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: SqrtRd operand: dpreg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:11'!
SqrtRs: dpreg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: SqrtRs operand: dpreg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/4/2015 14:48'!
Stop
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: Stop! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:12'!
SubCq: quickConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: SubCqR quickConstant: quickConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:07'!
SubCw: wordConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: SubCwR literal: wordConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
SubR: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: SubRR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/9/2020 07:21'!
SubR: subReg R: fromReg R: destReg
	"destReg := fromReg - subReg"
	<inline: false>
	<returnTypeC: #'AbstractInstruction *'>
	| first |
	backEnd hasThreeAddressArithmetic ifTrue:
		[^self gen: SubRRR operand: subReg operand: fromReg operand: destReg].
	self assert: subReg ~= destReg.
	first := self gen: MoveRR operand: fromReg operand: destReg.
	self gen: SubRR operand: subReg operand: destReg.
	^first! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/16/2009 09:03'!
SubRd: dpreg1 Rd: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: SubRdRd operand: dpreg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:11'!
SubRs: dpreg1 Rs: dpreg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: SubRsRs operand: dpreg1 operand: dpreg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 4/20/2017 13:47'!
SubbCq: quickConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: SubbCqR quickConstant: quickConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 17:11'!
SubbR: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: SubbRR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/15/2017 13:29'!
SwapR: regA R: regB Scratch: regTmp
	<returnTypeC: #'AbstractInstruction *'>
	^backEnd genSwapR: regA R: regB Scratch: regTmp! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:12'!
TstCq: quickConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: TstCqR quickConstant: quickConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 6/8/2015 12:07'!
XorCw: wordConstant R: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: XorCwR literal: wordConstant operand: reg! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 2/10/2009 17:58'!
XorR: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: XorRR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 5/8/2016 00:59'!
XorRd: reg1 Rd: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: XorRdRd operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'RonieSalgado 5/8/2016 00:59'!
XorRs: reg1 Rs: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^self gen: XorRsRs operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/9/2020 07:22'!
ZeroExtend16R: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	backEnd canZeroExtend ifFalse:
		[| first |
		 reg1 = reg2
			ifTrue:
				[first := self LogicalShiftLeftCq: BytesPerWord * 8 - 16 R: reg1]
			ifFalse:
				[first := self MoveR: reg1 R: reg2.
				 self LogicalShiftLeftCq: BytesPerWord * 8 - 16 R: reg1].
		self LogicalShiftRightCq: BytesPerWord * 8 - 16 R: reg2.
		^first].
	^self gen: ZeroExtend16RR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/9/2020 07:23'!
ZeroExtend32R: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	backEnd canZeroExtend ifFalse:
		[| first |
		 reg1 = reg2
			ifTrue:
				[first := self LogicalShiftLeftCq: 32 R: reg1]
			ifFalse:
				[first := self MoveR: reg1 R: reg2.
				 self LogicalShiftLeftCq: 32 R: reg1].
		self LogicalShiftRightCq: 32 R: reg2.
		^first].
	^self gen: ZeroExtend32RR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'abstract instructions' stamp: 'eem 1/9/2020 07:23'!
ZeroExtend8R: reg1 R: reg2
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	backEnd canZeroExtend ifFalse:
		[| first |
		 reg1 = reg2
			ifTrue:
				[first := self LogicalShiftLeftCq: BytesPerWord * 8 - 8 R: reg1]
			ifFalse:
				[first := self MoveR: reg1 R: reg2.
				 self LogicalShiftLeftCq: BytesPerWord * 8 - 8 R: reg1].
		self LogicalShiftRightCq: BytesPerWord * 8 - 8 R: reg2.
		^first].
	^self gen: ZeroExtend8RR operand: reg1 operand: reg2! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 4/28/2009 11:36'!
abortOffset
	<api>
	^missOffset! !

!Cogit methodsFor: 'testing' stamp: 'eem 2/10/2009 11:38'!
abstractInstruction: theAbstractInstruction follows: anAbstractInstruction
	<var: #theAbstractInstruction type: #'AbstractInstruction *'>
	<var: #anAbstractInstruction  type: #'AbstractInstruction *'>
	^theAbstractInstruction > anAbstractInstruction! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 1/6/2020 12:24'!
abstractInstructionAt: index
	<cmacro: '(index) (&abstractOpcodes[index])'>
	<returnTypeC: #'AbstractInstruction *'>
	((debugOpcodeIndices includes: index)
	 and: [breakMethod isNil or: [methodObj = breakMethod]]) ifTrue:
		[self halt: 'Accessing opcode index ', index printString].
	^abstractOpcodes at: index! !

!Cogit methodsFor: 'accessing' stamp: 'eem 2/10/2009 11:38'!
abstractOpcodes
	"For CogAbstractInstruction>>#>, which in turn is for Cogit>>abstractInstruction:follows:"
	<doNotGenerate>
	^abstractOpcodes! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/4/2015 09:53'!
addBlockStartAt: bcpc numArgs: numArgs numCopied: numCopied span: span
	"Add a blockStart for an embedded block.  For a binary tree walk block dispatch
	 blocks must be compiled in pc/depth-first order but are scanned in breadth-first
	 order, so do an insertion sort (which of course is really a bubble sort because we
	 have to move everything higher to make room)."
	<returnTypeC: #'BlockStart *'>
	| i blockStart |
	<var: #blockStart type: #'BlockStart *'>
	blockCount > 0
		ifTrue:
			[i := blockCount - 1.
			 [blockStart := self addressOf: (blockStarts at: i).
			   blockStart startpc > bcpc
			   and: [i > 0]] whileTrue:
				[i := i - 1].
			 blockCount to: i + 1 by: -1 do:
				[:j|
				blockStarts at: j put: (blockStarts at: j - 1)].
			blockStart := self cCode: [self addressOf: (blockStarts at: i + 1)]
								inSmalltalk: [blockStarts at: i + 1 put: CogBlockStart new]]
		ifFalse:
			[blockStart := self cCode: [self addressOf: (blockStarts at: blockCount)]
								inSmalltalk: [blockStarts at: blockCount put: CogBlockStart new]].
	blockCount := blockCount + 1.
	blockStart
		startpc: bcpc;
		numArgs: numArgs;
		numCopied: numCopied;
		stackCheckLabel: nil;
		hasInstVarRef: false;
		span: span.
	^blockStart! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 8/8/2014 15:08'!
addCleanBlockStarts
	1 to: (objectMemory literalCountOf: methodObj) do:
		[:i| | lit |
		lit := coInterpreter fetchPointer: i ofObject: methodObj.
		(coInterpreter startPCOrNilOfLiteral: lit in: methodObj) ifNotNil:
			[:startPCOrNil|
			 maxLitIndex := maxLitIndex max: i.
			 self addBlockStartAt: startPCOrNil - 1 "1-rel => 0-rel"
				numArgs: (coInterpreter argumentCountOfClosure: lit)
				numCopied: (coInterpreter copiedValueCountOfClosure: lit)
				span: (self spanForCleanBlockStartingAt: startPCOrNil - 1)]]! !

!Cogit methodsFor: 'debugging' stamp: 'eem 10/1/2013 16:59'!
addCogMethodsToHeapMap
	<api>
	"Perform an integrity/leak check using the heapMap.
	 Set a bit at each cog method's header."	
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType = CMMethod ifTrue:
			[coInterpreter heapMapAtWord: cogMethod Put: 1].
		cogMethod := methodZone methodAfter: cogMethod]! !

!Cogit methodsFor: 'method map' stamp: 'eem 1/26/2020 17:04'!
addToMap: annotation instruction: instruction byte: byte at: address for: mcpc
	<inline: true>
	self codeByteAt: address put: byte.
	self cCode: [] inSmalltalk:
		[| s bytecode |
		(compilationTrace anyMask: 64) ifTrue:
			[(s := coInterpreter transcript)
				ensureCr;
				print: annotation; nextPut: $/; nextPutAll: byte hex; space;
				nextPutAll: address hex; space; nextPutAll: mcpc hex; space;
				nextPutAll: (AnnotationConstantNames detect: [:name| (Cogit classPool at: name ifAbsent: []) = annotation]); cr; flush.
			 (instruction notNil
			  and: [instruction bcpc isInteger]) ifTrue:
				[s tab; print: instruction bcpc; nextPut: $/.
				 instruction bcpc printOn: s base: 16.
				 s space.
				 instruction printStateOn: s.
				 s space.
				 bytecode := objectMemory fetchByte: instruction bcpc ofObject: methodObj.
				 bytecode := bytecode + (self bytecodeSetOffsetForHeader: methodHeader).
				 (self generatorAt: bytecode) printStateOn: s.
				 s cr; flush]]]! !

!Cogit methodsFor: 'testing' stamp: 'eem 11/10/2020 15:19'!
addressIsInCodeZone: address
	<inline: true>
	^address asUnsignedInteger >= codeBase
	  and: [address asUnsignedInteger < methodZone zoneEnd]! !

!Cogit methodsFor: 'testing' stamp: 'eem 1/27/2020 09:03'!
addressIsInCurrentCompilation: address
	<inline: true>
	self cCode: '' inSmalltalk: [address < 0 ifTrue: [^false]].
	^address asUnsignedInteger >= methodLabel address
	  and: [address asUnsignedInteger < (methodZone youngReferrers min: methodLabel address + MaxMethodSize)]! !

!Cogit methodsFor: 'testing' stamp: 'eem 2/17/2017 15:38'!
addressIsInFixups: address
	<var: #address type: #'BytecodeFixup *'>
	^self cCode: '(BytecodeFixup *)address >= fixups && (BytecodeFixup *)address < (fixups + numAbstractOpcodes)'
		inSmalltalk:
			[fixups notNil
			 and: [(fixups object identityIndexOf: address) between: 1 and: numAbstractOpcodes]]! !

!Cogit methodsFor: 'testing' stamp: 'eem 1/29/2020 15:25'!
addressIsInInstructions: address
	<cmacro: '(address) (!!((usqInt)(address) & (BytesPerWord-1)) \
							&& (address) >= &abstractOpcodes[0] \
							&& (address) < &abstractOpcodes[opcodeIndex])'>
	^(abstractOpcodes object identityIndexOf: address) between: 1 and: opcodeIndex! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 09/01/2016 14:00'!
addressOfEndOfCase: n inCPIC: cPIC 
	"calculate the end of the n'th case statement - which is complicated because we have case 1 right at the top of our CPIC and then build up from the last one. Yes I know this sounds strange, but trust me - I'm an Engineer, we do things backwards all the emit"

	<var: #cPIC type: #'CogMethod *'>
	self assert: (n >= 1and: [n <= MaxCPICCases]).
	^n = 1
		ifTrue: [cPIC asInteger + firstCPICCaseOffset]
		ifFalse: [cPIC asInteger + firstCPICCaseOffset + (MaxCPICCases + 1 - n * cPICCaseSize)]! !

!Cogit methodsFor: 'accessing' stamp: 'eem (const to msg rewrite) 11/6/2014 11:21'!
addressSpaceMask
	<doNotGenerate>
	"Quad-byte-align, because the ARM requires 4-byte aligned jump & call targets."
	^((1 << (8 * objectMemory wordSize)) - 1) bitAnd: -4! !

!Cogit methodsFor: 'initialization' stamp: 'eem 12/17/2019 23:28'!
alignMethodZoneBase
	"Align methodZoneBase to that for the start of a method."
	| oldBase |
	oldBase := methodZoneBase.
	methodZoneBase := methodZone roundUpLength: methodZoneBase.
	backEnd stopsFrom: oldBase to: methodZoneBase - 1! !

!Cogit methodsFor: 'generate machine code' stamp: 'eem 6/23/2012 18:51'!
alignUptoRoutineBoundary: anAddress 
	^anAddress + 7 bitClear: 7! !

!Cogit methodsFor: 'accessing' stamp: 'eem (const to msg rewrite) 11/6/2014 11:21'!
allButTopBitOfAddressSpaceMask
	<doNotGenerate>
	"Quad-byte-align, because the ARM requires 4-byte aligned jump & call targets."
	^((1 << (8 * objectMemory wordSize - 1)) - 1) bitAnd: -4! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 10/8/2014 09:48'!
allCogMethodsFor: cogMethod
	<doNotGenerate>
	| blockEntry end methods pc |
	cogMethod isInteger ifTrue: [^self allCogMethodsFor: (self cogMethodSurrogateAt: cogMethod)].
	cogMethod cmType = CMBlock ifTrue:
		[^self allCogMethodsFor: cogMethod cmHomeMethod].
	(cogMethod cmType ~= CMMethod
	 or: [cogMethod blockEntryOffset = 0]) ifTrue:
		[^{cogMethod}].

	methods := OrderedCollection with: cogMethod.
	pc := blockEntry := cogMethod blockEntryOffset + cogMethod asInteger.
	end := (self mapEndFor: cogMethod) - 1.
	[pc < end] whileTrue:
		[| targetpc |
		 targetpc := blockEntry.
		 (backEnd isJumpAt: pc) ifTrue:
			[targetpc := backEnd jumpTargetPCAt: pc.
			 targetpc < blockEntry ifTrue:
				[methods add: (self cCoerceSimple: targetpc - (self sizeof: CogBlockMethod) to: #'CogBlockMethod *')]].
		 pc := pc + (backEnd instructionSizeAt: pc)].
	^methods sort! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/25/2021 10:34'!
allMachineCodeObjectReferencesValid
	"Check that all methods have valid selectors, and that all linked sends are to valid targets and have valid cache tags"
	| ok cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	ok := true.
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType ~= CMFree ifTrue:
			[(self asserta: (objectRepresentation checkValidOopReference: cogMethod selector)) ifFalse:
				[ok := false].
			 (self asserta: (self cogMethodDoesntLookKosher: cogMethod) = 0) ifFalse:
				[ok := false]].
		(cogMethod cmType = CMMethod
		 or: [cogMethod cmType = CMOpenPIC]) ifTrue:
			[(self asserta: ((self mapFor: cogMethod
								 performUntil: #checkIfValidOopRefAndTarget:pc:cogMethod:
								 arg: cogMethod) = 0)) ifFalse:
				[ok := false]].
		(cogMethod cmType = CMMethod
		 and: [(NewspeakVM or: [SistaVM])
		 and: [objectRepresentation canPinObjects]]) ifTrue:
			[(SistaVM and: [cogMethod counters ~= 0]) ifTrue:
				[(self asserta: (objectRepresentation checkValidDerivedObjectReference: cogMethod counters)) ifFalse:
					[ok := false]].
			 (NewspeakVM and: [cogMethod nextMethodOrIRCs ~= 0]) ifTrue:
				[(cogMethod nextMethodOrIRCs > methodZone zoneEnd) ifTrue:
					[(self asserta: (objectRepresentation checkValidDerivedObjectReference: cogMethod nextMethodOrIRCs)) ifFalse:
						[ok := false]]]].
		cogMethod cmType = CMClosedPIC ifTrue:
			[(self asserta: (self noTargetsFreeInClosedPIC: cogMethod)) ifFalse:
				[ok := false]].
		cogMethod := methodZone methodAfter: cogMethod].
	^ok! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 4/5/2009 11:59'!
allMethodsHaveCorrectHeader
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType = CMMethod ifTrue:
			[(objectRepresentation hasValidHeaderPostGC: cogMethod) ifFalse:
				[^false]].
		 cogMethod := methodZone methodAfter: cogMethod].
	^true! !

!Cogit methodsFor: 'initialization' stamp: 'eem 8/14/2013 14:34'!
allocateBlockStarts: numBlocks
	"Allocate the structures used to manage block compilation.  This
	 needs to be a macro since the structures are alloca'ed (stack
	 allocated) to ensure their being freed when compilation is done."
	<cmacro: '(numBlocks) do { \
		blockStarts = (numBlocks) ? alloca(sizeof(BlockStart) * (numBlocks)) : 0; \
} while (0)'>
	blockStarts := numBlocks > 0 ifTrue:
					[CArrayAccessor on:
						((1 to: numBlocks) collect:
							[:ign| CogBlockStart new])]! !

!Cogit methodsFor: 'initialization' stamp: 'eem 10/1/2016 14:26'!
allocateOpcodes: numberOfAbstractOpcodes bytecodes: numberOfBytecodes
	"Allocate the various arrays needed to compile abstract instructions.
	 Notionally we only need as many fixups as there are bytecodes.  But we
	 reuse fixups to record pc-dependent instructions in generateInstructionsAt:
	 and so need at least as many as there are abstract opcodes.

	 This *must* be inlined since the arrays are alloca'ed (stack allocated)
	 so that they are freed when compilation is done.

	 N.B. We do one single alloca to save embarrassing C optimizers that
	 generate incorrect code as both gcc and the intel compiler do on x86."
	<inline: true>
	numAbstractOpcodes := numberOfAbstractOpcodes.
	self
		cCode:
			[| opcodeSize fixupSize|
			 opcodeSize := (self sizeof: CogAbstractInstruction) * numAbstractOpcodes.
			 fixupSize := (self sizeof: CogBytecodeFixup) * numAbstractOpcodes.
			 abstractOpcodes := self alloca: opcodeSize + fixupSize.
			 self b: abstractOpcodes zero: opcodeSize + fixupSize.
			 fixups := (abstractOpcodes asUnsignedInteger + opcodeSize) asVoidPointer]
		inSmalltalk:
			[abstractOpcodes := CArrayAccessor on:
									 ((1 to: numAbstractOpcodes) collect: [:ign| CogCompilerClass for: self]).
			 fixups := CArrayAccessor on:
						((1 to: numAbstractOpcodes) collect: [:ign| self bytecodeFixupClass for: self])].
	self zeroOpcodeIndexForNewOpcodes.
	labelCounter := 0! !

!Cogit methodsFor: 'initialization' stamp: 'eem 3/14/2017 09:59'!
allocateOpcodes: numberOfAbstractOpcodes bytecodes: numberOfBytecodes ifFail: failBlock
	"Allocate the various arrays needed to compile abstract instructions, failing if the size
	 needed is considered too high.  Notionally we only need as many fixups as there are
	 bytecodes.  But we reuse fixups to record pc-dependent instructions in
	 generateInstructionsAt: and so need at least as many as there are abstract opcodes.

	 This *must* be inlined since the arrays are alloca'ed (stack allocated)
	 so that they are freed when compilation is done.

	 N.B. We do one single alloca to save embarrassing C optimizers that
	 generate incorrect code as both gcc and the intel compiler do on x86."
	<inline: true>
	| opcodeBytes fixupBytes allocBytes |
	numAbstractOpcodes := numberOfAbstractOpcodes.
	opcodeBytes := (self sizeof: CogAbstractInstruction) * numAbstractOpcodes.
	fixupBytes := (self sizeof: CogBytecodeFixup) * numAbstractOpcodes.
	allocBytes := opcodeBytes + fixupBytes.
	"Document the fact that the MaxStackAllocSize ensures that the number of abstract
	 opcodes fits in a 16 bit integer (e.g. CogBytecodeFixup's instructionIndex)."
	self assert: (self sizeof: CogAbstractInstruction) + (self sizeof: CogBytecodeFixup) * 49152 > MaxStackAllocSize.
	allocBytes > MaxStackAllocSize ifTrue:
		[^failBlock value].
	self
		cCode:
			[abstractOpcodes := self alloca: allocBytes.
			 self b: abstractOpcodes zero: allocBytes.
			 fixups := (abstractOpcodes asUnsignedInteger + opcodeBytes) asVoidPointer]
		inSmalltalk:
			[abstractOpcodes := CArrayAccessor on:
									 ((1 to: numAbstractOpcodes) collect: [:ign| CogCompilerClass for: self]).
			 fixups := CArrayAccessor on:
						((1 to: numAbstractOpcodes) collect: [:ign| self bytecodeFixupClass for: self])].
	self zeroOpcodeIndexForNewOpcodes.
	labelCounter := 0! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'cb 12/1/2016 10:15'!
allowEarlyOpenPICPromotion
	<inline: true>
	^ true! !

!Cogit methodsFor: 'method map' stamp: 'eem 9/9/2019 09:51'!
annotate: abstractInstruction objRef: anOop
	<var: #abstractInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	(objectRepresentation shouldAnnotateObjectReference: anOop) ifTrue:
		[self setHasMovableLiteral: true.
		 (objectMemory isYoungObject: anOop) ifTrue:
			[self setHasYoungReferent: true].
		 abstractInstruction annotation: IsObjectReference].
	^abstractInstruction! !

!Cogit methodsFor: 'method map' stamp: 'eem 6/10/2015 17:09'!
annotateAbsolutePCRef: abstractInstruction
	<var: #abstractInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	abstractInstruction annotation: IsAbsPCReference.
	^abstractInstruction! !

!Cogit methodsFor: 'method map' stamp: 'eem 6/10/2015 17:10'!
annotateBytecode: abstractInstruction
	<var: #abstractInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	abstractInstruction annotation: HasBytecodePC.
	^abstractInstruction! !

!Cogit methodsFor: 'method map' stamp: 'eem 6/10/2015 17:10'!
annotateCall: abstractInstruction
	<var: #abstractInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	abstractInstruction annotation: IsRelativeCall.
	^abstractInstruction! !

!Cogit methodsFor: 'method map' stamp: 'eem 6/10/2015 17:10'!
annotateNewspeakSend: abstractInstruction
	<var: #abstractInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	abstractInstruction annotation: IsNSSendCall.
	^abstractInstruction! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/6/2015 08:57'!
annotationForMcpc: mcpc in: cogHomeMethod
	"Answer the annotation for mcpc in cogHomeMethod's map, or 0 if no entry exists."
	<var: #cogHomeMethod type: #'CogMethod *'>
	| mapLocation mapByte annotation |
	mapLocation := self findMapLocationForMcpc: mcpc inMethod: cogHomeMethod.
	mapLocation = 0 ifTrue:
		[^0].
	mapByte := objectMemory byteAt: mapLocation.
	annotation := mapByte >> AnnotationShift.
	annotation = IsSendCall ifTrue:
		[mapByte := objectMemory byteAt: mapLocation - 1.
		 mapByte >> AnnotationShift = IsAnnotationExtension ifTrue:
			[annotation := annotation + (mapByte bitAnd: DisplacementMask)]].
	^annotation! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 1/12/2018 10:10'!
annotationIsForUncheckedEntryPoint: annotation
	<inline: true>
	^annotation = IsSuperSend
	  or: [BytecodeSetHasDirectedSuperSend
		  and: [annotation
					between: IsDirectedSuperSend
					and: IsDirectedSuperBindingSend]]! !

!Cogit methodsFor: 'debugging' stamp: 'eem 1/27/2020 09:40'!
armDisassembleDualZoneAnomalies
	<doNotGenerate>
	| badRangeStart |
	codeToDataDelta > 0 ifTrue:
		[codeBase to: methodZone zoneEnd - 4 by: 4 do:
			[:address|
			(objectMemory long32At: address) = (objectMemory long32At: address + codeToDataDelta)
				ifTrue:
					[badRangeStart ifNotNil:
						[self disassembleFrom: badRangeStart to: address - 4.
						 self disassembleFrom: badRangeStart + codeToDataDelta  to: address + codeToDataDelta - 4].
					 badRangeStart := nil]
				ifFalse:
					[badRangeStart ifNil: [badRangeStart := address]]].
		badRangeStart ifNotNil:
			[self disassembleFrom: badRangeStart to: methodZone zoneEnd - 4.
			 self disassembleFrom: badRangeStart + codeToDataDelta  to: methodZone zoneEnd + codeToDataDelta - 4]].
	^nil! !

!Cogit methodsFor: 'debugging' stamp: 'eem 2/24/2020 08:27'!
armPrintDualZoneAnomalies
	<doNotGenerate>
	| badRangeStart out |
	codeToDataDelta > 0 ifTrue:
		[out := coInterpreter transcript.
		 codeBase to: methodZone zoneEnd - 4 by: 4 do:
			[:address|
			 (objectMemory long32At: address) = (objectMemory long32At: address + codeToDataDelta)
				ifTrue:
					[badRangeStart ifNotNil:
						[out
							ensureCr;
							nextPutAll: 'anomaly '; nextPutAll: badRangeStart hex; nextPutAll: ' to: '; nextPutAll: (address - 4) hex;
							nextPutAll: ' vs ';
							nextPutAll: (badRangeStart + codeToDataDelta) hex; nextPutAll: ' to: '; nextPutAll: (address + codeToDataDelta - 4) hex.
						 (methodZone methodFor: badRangeStart) ifNotNil:
							[:cogMethod| out nextPutAll: ' (method '; nextPutAll: cogMethod address hex; nextPut: $)].
						 out cr; flush].
					 badRangeStart := nil]
				ifFalse:
					[badRangeStart ifNil: [badRangeStart := address]]].
		badRangeStart ifNotNil:
			[out
				ensureCr;
				nextPutAll: 'anomaly '; nextPutAll: badRangeStart hex; nextPutAll: ' to: '; nextPutAll: (methodZone zoneEnd - 4) hex;
				nextPutAll: ' vs ';
				nextPutAll: (badRangeStart + codeToDataDelta) hex; nextPutAll: ' to: '; nextPutAll: (methodZone zoneEnd + codeToDataDelta - 4) hex.
			 (methodZone methodFor: badRangeStart) ifNotNil:
				[:cogMethod| out nextPutAll: ' (method '; nextPutAll: cogMethod address hex; nextPut: $)].
			 out cr; flush]].
	^nil! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 3/30/2009 11:44'!
assertCStackWellAligned
	"Check alignment of the C stack.  This is a simulation only facsimilie.
	 See platforms/Cross/vm/sqCogStackAlignment.h for the real code."
	<doNotGenerate>
	self assert: processor sp \\ cStackAlignment = expectedSPAlignment.
	self assert: processor fp \\ cStackAlignment = expectedFPAlignment! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'rsf 1/5/2017 15:47'!
assertExtsAreConsumed: descriptor
	 "extended bytecodes must consume their extensions"
	 <inline: true>
	 descriptor isExtension ifFalse:
		[self assert: (extA = 0 and: [extB = 0 and: [numExtB = 0]])].! !

!Cogit methodsFor: 'debugging' stamp: 'eem 11/14/2020 20:15'!
assertMcpcIsPrimReturn: mcpc
	<inline: #always>
	^self assert: (mcpc = cePrimReturnEnterCogCode
					or: [mcpc = cePrimReturnEnterCogCodeProfiling])! !

!Cogit methodsFor: 'debugging' stamp: 'eem 6/21/2013 14:44'!
assertSaneJumpTarget: jumpTarget
	<var: #jumpTarget type: #'AbstractInstruction *'>

	self assert: (closedPICSize isNil "don't whinge when producing the PIC prototypes"
			or: [openPICSize isNil
			or: [(self addressIsInInstructions: jumpTarget)
			or: [(jumpTarget asUnsignedInteger
					between: codeBase
					and: methodZone limitZony asInteger + (closedPICSize max: openPICSize))]]])! !

!Cogit methodsFor: 'debugging' stamp: 'eem 1/27/2020 17:44'!
assertValidDualZone
	"{self firstInvalidDualZoneAddress. self firstInvalidDualZoneAddress + codeToDataDelta }"
	"{self firstInvalidDualZoneAddress hex. (self firstInvalidDualZoneAddress + codeToDataDelta) hex }"
	"{(objectMemory longAt: self firstInvalidDualZoneAddress) hex. (objectMemory longAt: self firstInvalidDualZoneAddress + codeToDataDelta) hex }"
	"self armDisassembleDualZoneAnomalies"
	"self armPrintDualZoneAnomalies"
	self cCode: ''
		inSmalltalk: [self assert: self firstInvalidDualZoneAddress isNil]! !

!Cogit methodsFor: 'debugging' stamp: 'eem 2/4/2020 19:16'!
assertValidDualZoneFrom: startAddress to: endAddress
	"{(self firstInvalidDualZoneAddressFrom: startAddress to: endAddress) hex. ((self firstInvalidDualZoneAddressFrom: startAddress to: endAddress) + codeToDataDelta) hex }"
	"{self firstInvalidDualZoneAddress. self firstInvalidDualZoneAddress + codeToDataDelta }"
	"{self firstInvalidDualZoneAddress hex. (self firstInvalidDualZoneAddress + codeToDataDelta) hex }"
	"{(objectMemory longAt: self firstInvalidDualZoneAddress) hex. (objectMemory longAt: self firstInvalidDualZoneAddress + codeToDataDelta) hex }"
	"self armDisassembleDualZoneAnomalies"
	"self armPrintDualZoneAnomalies"
	<inline: #always>
	self cCode:
			[self cppIf: #DUAL_MAPPED_CODE_ZONE
				ifTrue:
					[backEnd assertCoherentCodeAt: startAddress + cmNoCheckEntryOffset delta: codeToDataDelta]]
		inSmalltalk:
			[self assert: (self firstInvalidDualZoneAddressFrom: startAddress to: endAddress) isNil]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 1/29/2020 10:11'!
assertValidDualZoneReadAddress: address
	"Make sure that a surrogate is trying to read from the read/executable part of the code zone(s)"
	<cmacro: '(address) 0'>
	self assert: (address asInteger between: methodZoneBase and: methodZone zoneEnd)! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 1/29/2020 10:12'!
assertValidDualZoneWriteAddress: address
	"Make sure that a surrogate is trying to write to the writable part of the code zone(s)"
	<cmacro: '(address) 0'>
	self assert: (address asInteger - codeToDataDelta between: methodZoneBase and: methodZone zoneEnd)! !

!Cogit methodsFor: 'register management' stamp: 'eem 2/1/2020 15:16'!
availableRegisterOrNoneIn: liveRegsMask
	"Answer an unused abstract register in the registerMask, or NoReg if none."
	liveRegsMask ~= 0 ifTrue:
		[0 to: 31 do:
			[:reg| (liveRegsMask anyMask: 1 << reg) ifTrue: [^reg]]].
	^NoReg! !

!Cogit methodsFor: 'trampoline support' stamp: 'eem 5/31/2014 16:37'!
backEnd
	<cmacro: '() backEnd'>
	^backEnd! !

!Cogit methodsFor: 'tests-method map' stamp: 'eem 1/21/2016 15:59'!
bcpcsAndDescriptorsFor: aMethod bsOffset: bsOffset do: quaternaryBlock
	"Evaluate quaternaryBlock with the pc, byte, descriptor and numExtensions for
	 all the bytecodes in aMethod.  Evaluate with byte, descriptor and numExtensions
	 nil for the initialPC of the mehtod and any blocks within it."
	<doNotGenerate>
	| nExts byte descriptor endpc latestContinuation pc primIdx |
	((primIdx := coInterpreter primitiveIndexOf: aMethod) > 0
	and: [coInterpreter isQuickPrimitiveIndex: primIdx]) ifTrue:
		[^self].
	latestContinuation := pc := coInterpreter startPCOfMethod: aMethod.
	quaternaryBlock value: pc value: nil value: nil value: 0. "stackCheck/entry pc"
	primIdx > 0 ifTrue:
		[pc := pc + (self deltaToSkipPrimAndErrorStoreIn: aMethod
							header: (coInterpreter methodHeaderOf: aMethod))].
	nExts := 0.
	endpc := objectMemory numBytesOf: aMethod.
	[pc <= endpc] whileTrue:
		[byte := objectMemory fetchByte: pc ofObject: aMethod.
		descriptor := self generatorAt: byte + bsOffset.
		descriptor isExtension ifFalse:
			[quaternaryBlock value: pc value: byte value: descriptor value: nExts].
		(descriptor isReturn
		 and: [pc >= latestContinuation]) ifTrue:
			[endpc := pc].
		(descriptor isBranch
		 or: [descriptor isBlockCreation]) ifTrue:
			[| targetPC |
			 descriptor isBlockCreation ifTrue:
				[quaternaryBlock value: pc + descriptor numBytes value: nil value: nil value: 0]. "stackCheck/entry pc"
			 targetPC := self latestContinuationPCFor: descriptor at: pc exts: nExts in: aMethod.
			 self assert: targetPC < endpc.
			 latestContinuation := latestContinuation max: targetPC].
		pc := pc + descriptor numBytes.
		nExts := descriptor isExtension ifTrue: [nExts + 1] ifFalse: [0]]! !

!Cogit methodsFor: 'tests-method map' stamp: 'eem 7/25/2018 17:40'!
bcpcsDescriptorsAndStartpcsFor: aMethod bsOffset: bsOffset do: quinternaryBlock
	"Evaluate quinternaryBlock with the pc, byte, descriptor and numExtensions for
	 all the bytecodes in aMethod.  Evaluate with byte, descriptor and numExtensions
	 nil for the initialPC of the mehtod and any blocks within it."
	<doNotGenerate>
	| nExts byte descriptor endpc latestContinuation pc primIdx blockEndPCs startpcs |
	((primIdx := coInterpreter primitiveIndexOf: aMethod) > 0
	and: [coInterpreter isQuickPrimitiveIndex: primIdx]) ifTrue:
		[^self].
	latestContinuation := pc := coInterpreter startPCOfMethod: aMethod.
	startpcs := OrderedCollection with: pc.
	blockEndPCs := OrderedCollection with: (objectMemory numBytesOf: aMethod).
	quinternaryBlock value: pc value: nil value: nil value: 0 value: pc. "stackCheck/entry pc"
	primIdx > 0 ifTrue:
		[pc := pc + (self deltaToSkipPrimAndErrorStoreIn: aMethod
							header: (objectMemory methodHeaderOf: aMethod))].
	nExts := 0.
	endpc := objectMemory numBytesOf: aMethod.
	[pc <= endpc] whileTrue:
		[byte := objectMemory fetchByte: pc ofObject: aMethod.
		descriptor := self generatorAt: byte + bsOffset.
		descriptor isExtension ifFalse:
			[quinternaryBlock value: pc value: byte value: descriptor value: nExts value: startpcs last].
		descriptor isReturn ifTrue:
			[pc >= latestContinuation ifTrue:
				[endpc := pc]].
		(descriptor isBranch
		 or: [descriptor isBlockCreation]) ifTrue:
			[| targetPC |
			 targetPC := self latestContinuationPCFor: descriptor at: pc exts: nExts in: aMethod.
			 descriptor isBlockCreation ifTrue:
				[quinternaryBlock value: (startpcs addLast: pc + descriptor numBytes) value: nil value: nil value: 0 value: startpcs last.
				 blockEndPCs addLast: targetPC]. "stackCheck/entry pc"
			 self assert: targetPC < endpc.
			 latestContinuation := latestContinuation max: targetPC].
		descriptor isReturn ifTrue:
			[pc + descriptor numBytes >= blockEndPCs last ifTrue:
				[blockEndPCs removeLast.
				 startpcs removeLast]].
		pc := pc + descriptor numBytes.
		nExts := descriptor isExtension ifTrue: [nExts + 1] ifFalse: [0]]! !

!Cogit methodsFor: 'accessing' stamp: 'eem 12/19/2019 17:49'!
blockAlignment
	"Block method headers must be aligned on the correct boundary, just like non-block method headers.
	 This is because the CoInterpreter encodes flags in the least significant three bits of the method field."
	<api>
	<cmacro: '() 8'>
	self assert: (methodZone roundUpLength: 1) >= 8.
	^8! !

!Cogit methodsFor: 'method map' stamp: 'eem 6/7/2016 14:44'!
blockCreationBytecodeSizeForHeader: aMethodHeader
	<inline: true>
	^self cppIf: MULTIPLEBYTECODESETS
		ifTrue:
			[(objectMemory headerIndicatesAlternateBytecodeSet: aMethodHeader)
				ifTrue: [AltBlockCreationBytecodeSize]
				ifFalse: [BlockCreationBytecodeSize]]
		ifFalse: [BlockCreationBytecodeSize]! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 11/21/2012 17:00'!
blockDispatchFor: cogMethod perform: quaternaryFunction arg: arg
	"Evaluate quaternaryFunction with the block start mcpc, prev pc in block
	 dispatch, current pc in block dispatch and the supplied arg for each entry
	 in the block dispatch.  If the function answers non-zero answer the value
	 it answered. Used for disassembling blockDispatch."
	<doNotGenerate>
	| pc prevpc blockEntry end targetpc result |
	cogMethod blockEntryOffset = 0 ifTrue:
		[^nil].
	blockEntry := cogMethod blockEntryOffset + cogMethod asInteger.
	prevpc := pc := blockEntry.
	end := (self mapEndFor: cogMethod) - 1.
	[pc < end] whileTrue:
		[(backEnd isJumpAt: pc)
			ifTrue:
				[targetpc := backEnd jumpTargetPCAt: pc.
				 pc := pc + (backEnd instructionSizeAt: pc).
				 targetpc < blockEntry ifTrue:
					[result := self perform: quaternaryFunction
								with: targetpc
								with: prevpc
								with: pc
								with: arg.
					 result ~= 0 ifTrue:
						[^result].
					 prevpc := pc]]
			ifFalse:
				[pc := pc + (backEnd instructionSizeAt: pc)]].
	^0! !

!Cogit methodsFor: 'method map' stamp: 'eem 11/21/2012 17:00'!
blockDispatchTargetsFor: cogMethod perform: binaryFunction arg: arg
	"Evaluate binaryFunction with the block start mcpc and supplied arg for each
	 entry in the block dispatch.  If the function answers non-zero answer the value
	 it answered.  Used to update back-references to the home method in compaction."
	<var: #cogMethod type: #'CogMethod *'>
	<var: #binaryFunction declareC: 'usqInt (*binaryFunction)(sqInt mcpc, sqInt arg)'>
	| pc blockEntry end targetpc result |
	cogMethod blockEntryOffset = 0 ifTrue:
		[^nil].
	blockEntry := cogMethod blockEntryOffset + cogMethod asInteger.
	pc := blockEntry.
	end := (self mapEndFor: cogMethod) - 1.
	[pc < end] whileTrue:
		[(backEnd isJumpAt: pc) ifTrue:
			[targetpc := backEnd jumpTargetPCAt: pc.
			 targetpc < blockEntry ifTrue:
				[result := self perform: binaryFunction
							with: targetpc
							with: arg.
				 result ~= 0 ifTrue:
					[^result]]].
		pc := pc + (backEnd instructionSizeAt: pc)].
	^0! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 2/19/2017 14:32'!
blockStartAt: index
	<cmacro: '(index) (&blockStarts[index])'>
	"hack set startpc for printSimStack:toDepth:spillBase:on:"
	self maybeNoteStartpcFor: (blockStarts at: index).
	^blockStarts at: index! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 07/28/2014 08:07'!
blockStartPcsIn: aMethod
	"Answer the start bytecopde pcs in a method in compilation order, i.e. depth-first.
	 Blocks must occur in pc/depth-first order for binary tree block dispatch to work."
	| startpcs pc latestContinuation end descriptor byte bsOffset nExts |
	<doNotGenerate>
	startpcs := OrderedCollection new.
	startpcs add: (pc := latestContinuation := coInterpreter startPCOfMethod: aMethod).
	end := objectMemory numBytesOf: aMethod.
	bsOffset := self bytecodeSetOffsetFor: aMethod.
	nExts := 0.
	[pc <= end] whileTrue:
		[byte := objectMemory fetchByte: pc ofObject: aMethod.
		 descriptor := self generatorAt: byte + bsOffset.
		 (descriptor isReturn
		  and: [pc >= latestContinuation]) ifTrue:
			[end := pc].
		 (descriptor isBranch
		  or: [descriptor isBlockCreation]) ifTrue:
			[| targetPC |
			 targetPC := self latestContinuationPCFor: descriptor at: pc exts: nExts in: aMethod.
			 latestContinuation := latestContinuation max: targetPC].
		 pc := pc + descriptor numBytes.
		 descriptor isBlockCreation ifTrue:
			[startpcs add: pc].
		 nExts := descriptor isExtension ifTrue: [nExts + 1] ifFalse: [0]].
	^startpcs! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 12/24/2015 12:20'!
breakAt: address
	((breakPC isBreakpointFor: address)
	 and: [breakBlock shouldStopIfAtPC: address]) ifTrue:
		[coInterpreter changed: #byteCountText.
		 self halt: 'machine code breakpoint at ', address]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 4/8/2014 14:27'!
breakBlock
	<doNotGenerate>
	^breakBlock! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 12/28/2019 19:34'!
breakBlock: aBlock
	<doNotGenerate>
	(breakBlock := aBlock) ifNil:
		[breakPC == true ifTrue: [breakPC := nil].
		 ^self].
	breakPC ifNil: [breakPC := true].
	singleStep := singleStep or: [breakPC singleStepRequiredToTriggerIn: self]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 10/16/2009 14:21'!
breakMethod
	<doNotGenerate>
	^breakMethod! !

!Cogit methodsFor: 'debugging' stamp: 'eem 11/27/2012 14:59'!
breakOnImplicitReceiver
	<api>
	<cmacro: '() (traceFlags & 64)'>
	^(traceFlags bitAnd: 64) ~= 0! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 1/8/2009 10:39'!
breakPC
	<doNotGenerate>
	^breakPC! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 12/24/2015 13:12'!
breakPC: anAddressArrayOrNil
	<doNotGenerate>
	breakPC := anAddressArrayOrNil.
	singleStep := singleStep or: [anAddressArrayOrNil singleStepRequiredToTriggerIn: self]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 11/22/2010 11:08'!
bytecodeFixupClass
	self subclassResponsibility! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/11/2014 19:09'!
bytecodePCFor: mcpc startBcpc: startbcpc in: cogMethod
	"Answer the zero-relative bytecode pc matching the machine code pc argument in
	 cogMethod, given the start of the bytecodes for cogMethod's block or method object."
	<api>
	<var: #cogMethod type: #'CogBlockMethod *'>
	^self
		mapFor: cogMethod
		bcpc: startbcpc
		performUntil: #find:IsBackwardBranch:Mcpc:Bcpc:MatchingMcpc:
		arg: mcpc asVoidPointer! !

!Cogit methodsFor: 'initialization' stamp: 'eem 9/10/2019 18:54'!
bytecodeSetOffsetFor: aMethodObj
	<inline: true>
	^self
		cppIf: MULTIPLEBYTECODESETS
		ifTrue:
			[(coInterpreter methodUsesAlternateBytecodeSet: aMethodObj)
				ifTrue: [256]
				ifFalse: [0]]
		ifFalse:
			[self deny: (coInterpreter methodUsesAlternateBytecodeSet: aMethodObj).
			 0]! !

!Cogit methodsFor: 'initialization' stamp: 'eem 6/7/2016 14:45'!
bytecodeSetOffsetForHeader: aMethodHeader
	<inline: true>
	^self
		cppIf: MULTIPLEBYTECODESETS
		ifTrue:
			[(objectMemory headerIndicatesAlternateBytecodeSet: aMethodHeader)
				ifTrue: [256]
				ifFalse: [0]]
		ifFalse: [0]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 10/13/2020 09:10'!
cCoerce: value to: cTypeString
	"Type coercion. For translation a cast will be emmitted. When running in Smalltalk
	  answer a suitable wrapper for correct indexing."
	<doNotGenerate>
	^value coerceTo: cTypeString sim: objectMemory! !

!Cogit methodsFor: 'translation support' stamp: 'eem 10/8/2015 14:23'!
cCoerceSimple: value to: cTypeString
	<doNotGenerate>
	cTypeString last == $* ifTrue:
		[cTypeString == #'CogMethod *' ifTrue:
			[^(value isInteger and: [value < 0])
				ifTrue: [value] "it's an error code; leave it be"
				ifFalse: [self cogMethodSurrogateAt: value asUnsignedInteger]].
		cTypeString == #'CogBlockMethod *' ifTrue:
			[^self cogBlockMethodSurrogateAt: value asUnsignedInteger].
		cTypeString == #'NSSendCache *' ifTrue:
			[^self nsSendCacheSurrogateAt: value asUnsignedInteger].
		(cTypeString == #'AbstractInstruction *'
		 and: [value isBehavior]) ifTrue:
			[^CogCompilerClass].
		cTypeString == #'StackPage *' ifTrue:
			[^coInterpreter stackPages surrogateAtAddress: value]].
	^super cCoerceSimple: value to: cTypeString! !

!Cogit methodsFor: 'trampoline support' stamp: 'eem 12/15/2019 22:14'!
cFramePointerAddress
	<doNotGenerate>
	^coInterpreter cFramePointerAddress! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 09/01/2016 14:00'!
cPIC: cPIC HasTarget: targetMethod
	"Are any of the jumps from this CPIC to targetMethod?"
	<var: #cPIC type: #'CogMethod *'>
	<var: #targetMethod type: #'CogMethod *'>
	| pc target |
	target := targetMethod asUnsignedInteger + cmNoCheckEntryOffset.
	pc := cPIC asInteger + firstCPICCaseOffset.
	"Since this is a fast test doing simple compares we don't need to care that some
	cases have nonsense addresses in there. Just zip on through."
	"First jump is unconditional; subsequent ones are conditional"
	target = (backEnd jumpLongTargetBeforeFollowingAddress: pc) ifTrue:
		[^true].
	2 to: MaxCPICCases do:
		[:i|
		pc := pc + cPICCaseSize.
		target = (backEnd jumpLongConditionalTargetBeforeFollowingAddress: pc) ifTrue:
			[^true]].
	^false! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 1/28/2020 17:32'!
cPICCompactAndIsNowEmpty: cPIC
	"Scan the CPIC for target methods that have been freed and eliminate them.
	 Since the first entry cannot be eliminated, answer that the PIC should be
	 freed if the first entry is to a free target.  Answer if the PIC is now empty or should be freed."
	<var: #cPIC type: #'CogMethod *'>
	| pc entryPoint targetMethod targets tags methods used |
	<var: #targetMethod	type: #'CogMethod *'>
	<var: #tags			declareC: 'int tags[MaxCPICCases]'>
	<var: #targets			declareC: 'sqInt targets[MaxCPICCases]'>
	<var: #methods		declareC: 'sqInt methods[MaxCPICCases]'>
	self cCode: [] inSmalltalk:
		[tags := CArrayAccessor on: (Array new: MaxCPICCases).
		 targets := CArrayAccessor on: (Array new: MaxCPICCases).
		 methods := CArrayAccessor on: (Array new: MaxCPICCases)].
	used := 0.
	1 to: cPIC cPICNumCases do:
		[:i| | valid |
		 pc := self addressOfEndOfCase: i inCPIC: cPIC.
		 entryPoint := i = 1
						ifTrue: [backEnd jumpLongTargetBeforeFollowingAddress: pc]
						ifFalse: [backEnd jumpLongConditionalTargetBeforeFollowingAddress: pc].
		 valid := true.
		 "Collect all target triples except for triples whose entry-point is a freed method"
		 (cPIC containsAddress: entryPoint) ifFalse:
			[targetMethod := self cCoerceSimple: entryPoint - cmNoCheckEntryOffset to: #'CogMethod *'.
			 self assert: (targetMethod cmType = CMMethod or: [targetMethod cmType = CMFree]).
			 targetMethod cmType = CMFree ifTrue:
				[i = 1 ifTrue: [^true]. "cannot filter out the first entry cuz classTag is at point of send."
				 valid := false]].
		 valid ifTrue:
			[tags at: used put: (i > 1 ifTrue: [backEnd literal32BeforeFollowingAddress: pc - backEnd jumpLongConditionalByteSize]).
			 targets at: used put: entryPoint.
			 methods at: used put: (backEnd literalBeforeFollowingAddress: pc - (i = 1
																				ifTrue: [backEnd jumpLongByteSize]
																				ifFalse: [backEnd jumpLongConditionalByteSize + backEnd cmpC32RTempByteSize])).
			 used := used + 1]].
	used = cPIC cPICNumCases ifTrue:
		[^false].
	used = 0 ifTrue:
		[^true].

	(self writableMethodFor: cPIC) cPICNumCases: used.
	used = 1 ifTrue:
		[pc := self addressOfEndOfCase: 2 inCPIC: cPIC.
		 self rewriteCPIC: cPIC caseJumpTo: pc.
		 ^false].
	"the first entry cannot change..."
	1 to: used - 1 do:
		[:i|
		 pc := self addressOfEndOfCase: i + 1 inCPIC: cPIC.
		 self rewriteCPICCaseAt: pc tag: (tags at: i) objRef: (methods at: i) target: (targets at: i)].

	"finally, rewrite the jump 3 instr before firstCPICCaseOffset to jump to the beginning of this new case"
	self rewriteCPIC: cPIC caseJumpTo: pc - cPICCaseSize.
	^false! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'rmacnak 12/10/2015 18:49:23'!
cPICHasForwardedClass: cPIC 
	"The first case in a CPIC doesn't have a class reference so we need only step over actually usd subsequent cases."
	| pc |
	<var: #cPIC type: #'CogMethod *'>
	"start by finding the address of the topmost case, the cPICNumCases'th one"
	pc := (self addressOfEndOfCase: cPIC cPICNumCases inCPIC: cPIC)
				- backEnd jumpLongConditionalByteSize.
	2 to: cPIC cPICNumCases do: 
			[:i |  | classIndex |
			classIndex := backEnd literal32BeforeFollowingAddress: pc.
			(objectMemory isForwardedClassIndex: classIndex)
				ifTrue: [^ true].
			"since we started at the top, we can just add the case size each time to move on to the next case"
			pc := pc + cPICCaseSize].
	^ false! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 11/17/2015 14:23'!
cPICHasFreedTargets: cPIC
	"scan the CPIC for target methods that have been freed. "
	<var: #cPIC type: #'CogMethod *'>
	| pc entryPoint targetMethod |
	<var: #targetMethod type: #'CogMethod *'>

	1 to: cPIC cPICNumCases do:
		[:i|
		pc := self addressOfEndOfCase: i inCPIC: cPIC.
		entryPoint := i = 1
						ifTrue: [backEnd jumpLongTargetBeforeFollowingAddress: pc]
						ifFalse: [backEnd jumpLongConditionalTargetBeforeFollowingAddress: pc].
		"Find target from jump.  Ignore jumps to the interpret and MNU calls within this PIC"
		(cPIC containsAddress: entryPoint) ifFalse:
			[targetMethod := self cCoerceSimple: entryPoint - cmNoCheckEntryOffset to: #'CogMethod *'.
			 self assert: (targetMethod cmType = CMMethod or: [targetMethod cmType = CMFree]).
			 targetMethod cmType = CMFree ifTrue:
				[^true]]].
	^false! !

!Cogit methodsFor: 'accessing' stamp: 'eem 1/29/2020 14:50'!
cPICPrototype
	"For Cogit clas>>#genAndDisPICoptions:"
	<doNotGenerate>
	^cPICPrototype address! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 3/11/2017 10:20'!
cPICPrototypeCaseOffset
	"Whimsey; we want 16rCA5E10 + cPICPrototypeCaseOffset to be somewhere in the middle of the zone."
	<inline: true>
	^methodZoneBase + methodZone youngReferrers / 2 - 16rCA5E10! !

!Cogit methodsFor: 'trampoline support' stamp: 'eem 12/15/2019 22:14'!
cStackPointerAddress
	<doNotGenerate>
	^coInterpreter cStackPointerAddress! !

!Cogit methodsFor: 'debugging' stamp: 'eem 7/3/2014 16:54'!
callCogCodePopReceiver
	"This is a static version of ceCallCogCodePopReceiverReg
	 for break-pointing when debugging in C."
	<api>
	<inline: false>
	"This exists only for break-pointing."
	self cCode: [self realCECallCogCodePopReceiverReg]
		inSmalltalk: [self ceCallCogCodePopReceiverReg].
	"(and this exists only to reference Debug)"
	Debug ifFalse: [self error: 'what??']! !

!Cogit methodsFor: 'debugging' stamp: 'eem 7/3/2014 16:54'!
callCogCodePopReceiverAndClassRegs
	"This is a static version of ceCallCogCodePopReceiverAndClassRegs
	 for break-pointing when debugging in C."
	<api>
	<inline: false>
	"This exists only for break-pointing."
	self cCode: [self realCECallCogCodePopReceiverAndClassRegs]
		inSmalltalk: [self ceCallCogCodePopReceiverAndClassRegs]! !

!Cogit methodsFor: 'accessing' stamp: 'eem 2/12/2009 17:00'!
ceBaseFrameReturnPC
	<api>
	<cmacro: '() ceBaseFrameReturnTrampoline'>
	^ceBaseFrameReturnTrampoline! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 7/5/2021 16:28'!
ceCPICMiss: cPIC receiver: receiver
	"Code entry closed PIC miss.  A send has fallen
	 through a closed (finite) polymorphic inline cache.
	 Either extend it or patch the send site to an open PIC.
	 The stack looks like:
			receiver
			args
	  sp=>	sender return address"
	<var: #cPIC type: #'CogMethod *'>
	<api>
	<static: true>
	| outerReturn newTargetMethodOrNil errorSelectorOrNil cacheTag result |
	self cCode: ''
		inSmalltalk:
			[cPIC isInteger ifTrue:
				[^self ceCPICMiss: (self cogMethodSurrogateAt: cPIC) receiver: receiver]].
	(objectMemory isOopForwarded: receiver) ifTrue:
		[^coInterpreter ceSendFromInLineCacheMiss: cPIC].
	outerReturn := coInterpreter stackTop.
	self deny: (backEnd inlineCacheTagAt: outerReturn) = self picAbortDiscriminatorValue. 
	cPIC cPICNumCases < MaxCPICCases
		ifTrue:
			[self lookup: cPIC selector
				for: receiver
				methodAndErrorSelectorInto:
					[:method :errsel|
					newTargetMethodOrNil := method.
					errorSelectorOrNil := errsel]]
		ifFalse: [newTargetMethodOrNil := errorSelectorOrNil := nil].
	"We assume lookupAndCog:for: will *not* reclaim the method zone"
	self assert: outerReturn = coInterpreter stackTop.
	self ensureWritableCodeZone.
	cacheTag := objectRepresentation inlineCacheTagForInstance: receiver.
	(cPIC cPICNumCases >= MaxCPICCases
	 or: [(errorSelectorOrNil notNil and: [errorSelectorOrNil ~= SelectorDoesNotUnderstand])
	 or: [(objectRepresentation inlineCacheTagIsYoung: cacheTag)
	 or: [newTargetMethodOrNil isNil
	 or: [objectMemory isYoung: newTargetMethodOrNil]]]]) ifTrue:
		[result := self patchToOpenPICFor: cPIC selector
					numArgs: cPIC cmNumArgs
					receiver: receiver.
		 self assert: result not. "If patchToOpenPICFor:.. returns we're out of code memory"
		 self ensureExecutableCodeZone.
		 ^coInterpreter ceSendFromInLineCacheMiss: cPIC].
	"Now extend the PIC with the new case."
	self cogExtendPIC: cPIC
		CaseNMethod: newTargetMethodOrNil
		tag: cacheTag
		isMNUCase: errorSelectorOrNil = SelectorDoesNotUnderstand.
	self ensureExecutableCodeZone.
	"Jump back into the pic at its entry in case this is an MNU."
	coInterpreter
		executeCogPIC: cPIC
		fromLinkedSendWithReceiver: receiver
		andCacheTag: (backEnd inlineCacheTagAt: outerReturn).
	"NOTREACHED"
	^nil! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 7/3/2014 16:43'!
ceCallCogCodePopReceiverAndClassRegs
	<api: 'extern void (*ceCallCogCodePopReceiverAndClassRegs)()'>
	<doNotGenerate>
	self simulateEnilopmart: ceCallCogCodePopReceiverAndClassRegs numArgs: 2! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 7/3/2014 16:44'!
ceCallCogCodePopReceiverReg
	<api: 'extern void (*ceCallCogCodePopReceiverReg)()'>
	<doNotGenerate>
	self simulateEnilopmart: ceCallCogCodePopReceiverReg numArgs: 1! !

!Cogit methodsFor: 'accessing' stamp: 'eem 6/23/2009 18:45'!
ceCannotResumePC
	<api>
	<cmacro: '() ((usqInt)ceCannotResumeTrampoline)'>
	<returnTypeC: #usqInt> "for Slang"
	^ceCannotResumeTrampoline! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 11/22/2020 21:04'!
ceCaptureCStackPointers
	<api: 'extern void (*ceCaptureCStackPointers)()'>
	<doNotGenerate>
	coInterpreter isCurrentImageFacade ifTrue:
		[^self].
	self simulateLeafCallOf: ceCaptureCStackPointers.
	thisContext sender selector == #generateStackPointerCapture ifTrue:
		[^self].
	coInterpreter isThreadedVM ifTrue:
		[coInterpreter assertProcessorStackPointersBelongToCurrentThread]! !

!Cogit methodsFor: 'accessing' stamp: 'eem 12/1/2020 17:53'!
ceCheckForInterruptTrampoline
	<cmacro: '() ceCheckForInterruptTrampoline'>
	^ceCheckForInterruptTrampoline! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 10/22/2009 18:15'!
ceEnterCogCodePopReceiverReg
	<api: 'extern void (*ceEnterCogCodePopReceiverReg)()'>
	<doNotGenerate>
	self simulateEnilopmart: ceEnterCogCodePopReceiverReg numArgs: 1! !

!Cogit methodsFor: 'trampoline support' stamp: 'RonieSalgado 4/29/2016 17:12'!
ceFree: pointer
	<api>
	<var: #pointer type: #'void*'>
	self free: pointer! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 4/1/2020 13:41'!
ceInvokeInterpret
	<api: 'extern void (*ceInvokeInterpret)()'>
	<doNotGenerate>
	self simulateEnilopmart: ceInvokeInterpret numArgs: 0! !

!Cogit methodsFor: 'trampoline support' stamp: 'eem 7/5/2021 10:44'!
ceMalloc: size
	<api>
	<static: true>
	<var: #size type: #'size_t'>
	<returnTypeC: #'void*'>
	^ self malloc: size! !

!Cogit methodsFor: 'accessing' stamp: 'eem 6/23/2009 17:56'!
ceReturnToInterpreterPC
	<api>
	<cmacro: '() ((usqInt)ceReturnToInterpreterTrampoline)'>
	<returnTypeC: #usqInt> "for Slang"
	^ceReturnToInterpreterTrampoline! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 7/5/2021 16:30'!
ceSICMiss: receiver
	"An in-line cache check in a method has failed.  The failing entry check has jumped
	 to the ceMethodAbort abort call at the start of the method which has called this routine.
	 If possible allocate a closed PIC for the current and existing classes.
	 The stack looks like:
			receiver
			args
			sender return address
	  sp=>	ceMethodAbort call return address
	 So we can find the method that did the failing entry check at
		ceMethodAbort call return address - missOffset
	 and we can find the send site from the outer return address."
	<api>
	<static: true>
	| pic innerReturn outerReturn entryPoint targetMethod newTargetMethodOrNil errorSelectorOrNil cacheTag extent result |
	<var: #pic type: #'CogMethod *'>
	<var: #targetMethod type: #'CogMethod *'>
	"Whether we can relink to a PIC or not we need to pop off the inner return and identify the target method."
	innerReturn := coInterpreter popStack asUnsignedInteger.
	targetMethod := self cCoerceSimple: innerReturn - missOffset to: #'CogMethod *'.
	(objectMemory isOopForwarded: receiver) ifTrue:
		[^coInterpreter ceSendFromInLineCacheMiss: targetMethod].
	outerReturn := coInterpreter stackTop asUnsignedInteger.
	self assert: (outerReturn between: methodZoneBase and: methodZone freeStart).
	entryPoint := backEnd callTargetFromReturnAddress: outerReturn.

	self assert: targetMethod selector ~= objectMemory nilObject.
	self assert: targetMethod asInteger + cmEntryOffset = entryPoint.

	self lookup: targetMethod selector
		for: receiver
		methodAndErrorSelectorInto:
			[:method :errsel|
			newTargetMethodOrNil := method.
			errorSelectorOrNil := errsel].
	"We assume lookupAndCog:for: will *not* reclaim the method zone"
	self assert: outerReturn = coInterpreter stackTop.
	cacheTag := objectRepresentation inlineCacheTagForInstance: receiver.
	((errorSelectorOrNil notNil and: [errorSelectorOrNil ~= SelectorDoesNotUnderstand])
	 or: [(objectRepresentation inlineCacheTagIsYoung: cacheTag)
	 or: [(backEnd inlineCacheTagAt: outerReturn) = self picAbortDiscriminatorValue
	 or: [newTargetMethodOrNil isNil
	 or: [objectMemory isYoung: newTargetMethodOrNil]]]]) ifTrue:
		[result := self patchToOpenPICFor: targetMethod selector
					numArgs: targetMethod cmNumArgs
					receiver: receiver.
		 self assert: result not. "If patchToOpenPICFor:.. returns we're out of code memory"
		 ^coInterpreter ceSendFromInLineCacheMiss: targetMethod].
	self ensureWritableCodeZone.
	"See if an Open PIC is already available."
	pic := methodZone openPICWithSelector: targetMethod selector.
	(pic isNil or: [self allowEarlyOpenPICPromotion not]) ifTrue:
		["otherwise attempt to create a closed PIC for the two cases."
		 pic := self cogPICSelector: targetMethod selector
					numArgs: targetMethod cmNumArgs
					Case0Method: targetMethod
					Case1Method: newTargetMethodOrNil
					tag: cacheTag
					isMNUCase: errorSelectorOrNil = SelectorDoesNotUnderstand.
		 (pic asInteger between: MaxNegativeErrorCode and: -1) ifTrue:
			["For some reason the PIC couldn't be generated, most likely a lack of code memory.
			  Continue as if this is an unlinked send."
			 pic asInteger = InsufficientCodeSpace ifTrue:
				[coInterpreter callForCogCompiledCodeCompaction].
			 self ensureExecutableCodeZone.
			^coInterpreter ceSendFromInLineCacheMiss: targetMethod]].
	"Relink the send site to the pic.  If to an open PIC then reset the cache tag to the selector,
	 for the benefit of the cacheTag assert check in checkIfValidOopRef:pc:cogMethod: et al."
	extent := pic cmType = CMOpenPIC
				ifTrue:
					[backEnd
						rewriteInlineCacheAt: outerReturn
						tag: (backEnd
								inlineCacheValueForSelector: targetMethod selector
								in: coInterpreter mframeHomeMethodExport)
						target: pic asInteger + cmEntryOffset]
				ifFalse:
					[backEnd
						rewriteCallAt: outerReturn
						target: pic asInteger + cmEntryOffset].

	self assertValidDualZoneFrom: pic asUnsignedInteger to: pic asUnsignedInteger + closedPICSize.
	"These also implicitly flush the read/write mapped dual zone to the read/execute zone."
	backEnd flushICacheFrom: pic asUnsignedInteger to: pic asUnsignedInteger + closedPICSize.
	backEnd flushICacheFrom: outerReturn asUnsignedInteger - extent to: outerReturn asUnsignedInteger.
	"Jump back into the pic at its entry in case this is an MNU (newTargetMethodOrNil is nil)"
	coInterpreter
		executeCogPIC: pic
		fromLinkedSendWithReceiver: receiver
		andCacheTag: (backEnd inlineCacheTagAt: outerReturn).
	"NOTREACHED"
	^nil! !

!Cogit methodsFor: 'debugging' stamp: 'eem 8/22/2011 10:23'!
checkEnoughOpcodes
	<inline: true>
	opcodeIndex > numAbstractOpcodes ifTrue:
		[self error: 'Cog JIT internal error. Too many abstract opcodes.  Num opcodes heuristic is too optimistic.']! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 09/05/2021 19:12'!
checkIfValidOopRef: annotation pc: mcpc cogMethod: cogMethod
	"Check for a valid object reference, if any, at a map entry.  Answer a code unique to each error for debugging."
	<var: #mcpc type: #'char *'>
	<var: #cogMethod type: #'CogMethod *'>

	<var: #nsSendCache type: #'NSSendCache *'>
	annotation = IsObjectReference ifTrue:
		[| literal |
		 literal := literalsManager fetchLiteralAtAnnotatedAddress: mcpc asUnsignedInteger using: backEnd.
		 (objectRepresentation checkValidOopReference: literal) ifFalse:
			[coInterpreter print: 'object ref leak in CM '; printHex: cogMethod asInteger; print: ' @ '; printHex: mcpc asInteger; eekcr.
			^1]].

	NewspeakVM ifTrue:
		[annotation = IsNSSendCall ifTrue:
			[| nsSendCache enclosingObject |
			nsSendCache := self nsSendCacheFromReturnAddress: mcpc asInteger.
			[(objectRepresentation checkValidOopReference: nsSendCache selector) ifFalse:
				[coInterpreter print: 'selector leak in CM '; printHex: cogMethod asInteger; print: ' @ '; printHex: mcpc asInteger; eekcr.
				^1]].
			(enclosingObject := nsSendCache enclosingObject) ~= 0 ifTrue:
				[[(objectRepresentation checkValidOopReference: enclosingObject) ifFalse:
					[coInterpreter print: 'enclosing object leak in CM '; printHex: cogMethod asInteger; print: ' @ '; printHex: mcpc asInteger; eekcr.
					^1]]]]].

	(self isPureSendAnnotation: annotation) ifTrue:
		[| entryPoint selectorOrCacheTag offset |
		 entryPoint := backEnd callTargetFromReturnAddress: mcpc asInteger.
		 entryPoint <= methodZoneBase
			ifTrue:
				[offset := entryPoint]
			ifFalse:
				[self
					offsetAndSendTableFor: entryPoint
					annotation: annotation
					into: [:off :table| offset := off]].
		 selectorOrCacheTag := backEnd inlineCacheTagAt: mcpc asInteger.
		 (entryPoint > methodZoneBase
		  and: [offset ~= cmNoCheckEntryOffset
		  and: [(self cCoerceSimple: entryPoint - offset to: #'CogMethod *') cmType ~= CMOpenPIC]])
			ifTrue: "linked non-super send, cacheTag is a cacheTag"
				[(objectRepresentation validInlineCacheTag: selectorOrCacheTag) ifFalse:
					[coInterpreter print: 'cache tag leak in CM '; printHex: cogMethod asInteger; print: ' @ '; printHex: mcpc asInteger; eekcr.
					^1]]
			ifFalse: "unlinked send or super send; cacheTag is a selector unless 64-bit, in which case it is an index."
				[(self inlineCacheTagsAreIndexes
				  or: [objectRepresentation checkValidOopReference: selectorOrCacheTag]) ifFalse:
					[coInterpreter print: 'selector leak in CM '; printHex: cogMethod asInteger; print: ' @ '; printHex: mcpc asInteger; eekcr.
					^1]]].
	^0 "keep scanning"! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/27/2021 13:46'!
checkIfValidOopRefAndTarget: annotation pc: mcpc cogMethod: cogMethod
	"Check for a valid object reference, if any, at a map entry.  Answer a code unique to each error for debugging."
	<var: #mcpc type: #'char *'>
	<var: #cogMethod type: #'CogMethod *'>

	<var: #nsSendCache type: #'NSSendCache *'>
	| literal entryPoint |
	annotation = IsObjectReference ifTrue:
		[literal := literalsManager fetchLiteralAtAnnotatedAddress: mcpc asUnsignedInteger using: backEnd.
		 (self asserta: (objectRepresentation checkValidOopReference: literal)) ifFalse:
			[^1].
		((objectRepresentation couldBeObject: literal)
		 and: [objectMemory isReallyYoungObject: literal]) ifTrue:
			[(self asserta: (self cCoerceSimple: cogMethod to: #'CogMethod *') cmRefersToYoung) ifFalse:
				[^2]]].

	NewspeakVM ifTrue:
		[annotation = IsNSSendCall ifTrue:
			[| nsSendCache classTag enclosingObject nsTargetMethod |
			nsSendCache := self nsSendCacheFromReturnAddress: mcpc asInteger.
			(self asserta: (objectRepresentation checkValidOopReference: nsSendCache selector)) ifFalse:
				[^9].
			classTag := nsSendCache classTag.
			(self asserta: (classTag = 0 or: [objectRepresentation validInlineCacheTag: classTag])) ifFalse:
				[^10].
			enclosingObject := nsSendCache enclosingObject.
			(self asserta: (enclosingObject = 0 or: [objectRepresentation checkValidOopReference: enclosingObject])) ifFalse:
				[^11].
			entryPoint := nsSendCache target.
			entryPoint ~= 0 ifTrue: [
				nsTargetMethod := self cCoerceSimple: entryPoint - cmNoCheckEntryOffset to: #'CogMethod *'.
				(self asserta: (nsTargetMethod cmType = CMMethod)) ifFalse:
					[^12]]]].

	(self isPureSendAnnotation: annotation) ifTrue:
		[(self asserta: (self cCoerceSimple: cogMethod to: #'CogMethod *') cmType = CMMethod) ifFalse:
			[^3].
		 self entryCacheTagAndCouldBeObjectAt: mcpc annotation: annotation into:
			[:entryPt :cacheTag :tagCouldBeObject|
			entryPoint := entryPt.
			tagCouldBeObject
				ifTrue:
					[(objectRepresentation couldBeObject: cacheTag)
						ifTrue:
							[(self asserta: (objectRepresentation checkValidOopReference: cacheTag)) ifFalse:
								[^4]]
						ifFalse:
							[(self asserta: (objectRepresentation validInlineCacheTag: cacheTag)) ifFalse:
								[^5]].
					((objectRepresentation couldBeObject: cacheTag)
					 and: [objectMemory isReallyYoungObject: cacheTag]) ifTrue:
						[(self asserta: (self cCoerceSimple: cogMethod to: #'CogMethod *') cmRefersToYoung) ifFalse:
							[^6]]]
				ifFalse:
					[(self inlineCacheTagsAreIndexes
					  and: [self entryPointTagIsSelector: entryPoint])
						ifTrue:
							[cacheTag signedIntFromLong < 0
								ifTrue:
									[cacheTag signedIntFromLong negated > NumSpecialSelectors ifTrue:
										[^7]]
								ifFalse:
									[cacheTag >= (objectMemory literalCountOf: enumeratingCogMethod methodObject) ifTrue:
										[^8]]]
						ifFalse:
							[(self asserta: (objectRepresentation validInlineCacheTag: cacheTag)) ifFalse:
								[^9]]]].
		entryPoint > methodZoneBase ifTrue:
			["It's a linked send; find which kind."
			 self targetMethodAndSendTableFor: entryPoint annotation: annotation into:
					[:targetMethod :sendTable|
					 (self asserta: (targetMethod cmType = CMMethod
								   or: [targetMethod cmType = CMClosedPIC
								   or: [targetMethod cmType = CMOpenPIC]])) ifFalse:
						[^10]]]].
	^0 "keep scanning"! !

!Cogit methodsFor: 'debugging' stamp: 'eem 09/05/2021 19:12'!
checkIntegrityOfObjectReferencesInCode: gcModes
	<api>
	"Answer if all references to objects in machine-code are valid."	
	| cogMethod ok count |
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	ok := true.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType ~= CMFree ifTrue:
			[cogMethod cmRefersToYoung ifTrue:
				[(count := methodZone occurrencesInYoungReferrers: cogMethod) ~= 1 ifTrue:
					[coInterpreter print: 'young referrer CM '; printHex: cogMethod asInteger.
					 count = 0
						ifTrue: [coInterpreter print: ' is not in youngReferrers'; eekcr]
						ifFalse: [coInterpreter print: ' is in youngReferrers '; printNum: count; print: ' times!!'; eekcr].
					 ok := false]].
			 (objectRepresentation checkValidOopReference: cogMethod selector) ifFalse:
				[coInterpreter print: 'object leak in CM '; printHex: cogMethod asInteger; print: ' selector'; eekcr.
				 ok := false].
			 cogMethod cmType = CMMethod
				ifTrue:
					[self assert: cogMethod objectHeader = objectMemory nullHeaderForMachineCodeMethod.
					 (objectRepresentation checkValidObjectReference: cogMethod methodObject) ifFalse:
						[coInterpreter print: 'object leak in CM '; printHex: cogMethod asInteger; print: ' methodObject'; eekcr.
						 ok := false].
					 (objectMemory isOopCompiledMethod: cogMethod methodObject) ifFalse:
						[coInterpreter print: 'non-method in CM '; printHex: cogMethod asInteger; print: ' methodObject'; eekcr.
						 ok := false].
					 (self mapFor: cogMethod
						 performUntil: #checkIfValidOopRef:pc:cogMethod:
						 arg: cogMethod) ~= 0
							ifTrue: [ok := false].
					 (objectRepresentation hasSpurMemoryManagerAPI
					  or: [gcModes anyMask: GCModeNewSpace]) ifTrue:
						[(((objectMemory isYoungObject: cogMethod methodObject)
						    or: [objectMemory isYoung: cogMethod selector])
						   and: [cogMethod cmRefersToYoung not]) ifTrue:
							[coInterpreter print: 'CM '; printHex: cogMethod asInteger; print: ' refers to young but not marked as such'; eekcr.
							 ok := false]]]
				ifFalse:
					[cogMethod cmType = CMClosedPIC
						ifTrue:
							[(self checkValidObjectReferencesInClosedPIC: cogMethod) ifFalse:
								[ok := false]]
						ifFalse:
							[cogMethod cmType = CMOpenPIC
								ifTrue:
									[(self mapFor: cogMethod
										performUntil: #checkIfValidOopRef:pc:cogMethod:
										arg: cogMethod) ~= 0
											ifTrue: [ok := false]]]]].
		cogMethod := methodZone methodAfter: cogMethod].
	^ok! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 6/16/2015 09:08'!
checkMaybeObjRefInClosedPIC: maybeObject
	maybeObject = 0 ifTrue:
		[^true].
	(objectRepresentation couldBeObject: maybeObject) ifFalse:
		[^true].
	^objectRepresentation checkValidObjectReference: maybeObject! !

!Cogit methodsFor: 'debugging' stamp: 'eem 3/17/2014 18:16'!
checkStackDepthOnSend
	<doNotGenerate>
	^(traceFlags bitAnd: 128) ~= 0! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 2/20/2016 19:00'!
checkValidObjectReferencesInClosedPIC: cPIC
	<var: #cPIC type: #'CogMethod *'>
	| ok pc |
	ok := true.
	pc := cPIC asInteger + firstCPICCaseOffset.
	
	"first we check the obj ref at the beginning of the CPIC"
	(self checkMaybeObjRefInClosedPIC: (backEnd literalBeforeFollowingAddress: pc - backEnd jumpLongByteSize)) ifFalse:
		[self print: 'object leak in CPIC '; printHex: cPIC asInteger;
			print: ' @ '; printHex: pc - backEnd jumpLongByteSize; cr.
		 ok := false].
	
	"Next we step over each case that is in use. We find the end address of the cPICNumCases'th case and can then just step forward by the case size thereafter"
	pc := self addressOfEndOfCase: cPIC cPICNumCases inCPIC: cPIC.
	
	"For each case we check any object reference at the end address - sizeof(conditional instruction) and then increment the end address by case size"
	2 to: cPIC cPICNumCases do:
		[:i|
		(self inlineCacheTagsAreIndexes not
		 and: [objectRepresentation inlineCacheTagsMayBeObjects]) ifTrue:
			[(self checkMaybeObjRefInClosedPIC: (backEnd literal32BeforeFollowingAddress: pc - backEnd jumpLongConditionalByteSize)) ifFalse:
				[self print: 'object leak in CPIC '; printHex: cPIC asInteger;
					print: ' @ '; printHex: pc - backEnd jumpLongConditionalByteSize - backEnd loadLiteralByteSize; cr.
				 ok := false]].
		(self checkMaybeObjRefInClosedPIC: (backEnd literalBeforeFollowingAddress: pc - backEnd jumpLongConditionalByteSize - backEnd cmpC32RTempByteSize)) ifFalse:
			[self print: 'object leak in CPIC '; printHex: cPIC asInteger;
				print: ' @ '; printHex: pc - backEnd jumpLongConditionalByteSize; cr.
			 ok := false].
		pc := pc + cPICCaseSize].
	^ok! !

!Cogit methodsFor: 'profiling primitives' stamp: 'eem 6/1/2018 12:11'!
cleanUpFailingCogCodeConstituents: cogMethodArg
	<var: #cogMethodArg type: #'CogMethod *'>
	<inline: #never> "i.e. this should never be called, so keep it out of the main path."
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := cogMethodArg.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType = CMClosedPIC ifTrue:
			[cogMethod methodObject: 0].
		cogMethod := methodZone methodAfter: cogMethod].
	"would like to assert this, but it requires the leak checked be run :-(
		self assert: self allMachineCodeObjectReferencesValid."
	coInterpreter popRemappableOop.
	^nil! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 5/11/2016 11:23'!
closedPICRefersToUnmarkedObject: cPIC
	"Answer if the ClosedPIC refers to any unmarked objects or freed/freeable target methods,
	 applying markAndTraceOrFreeCogMethod:firstVisit: to those targets to determine if freed/freeable."
	<var: #cPIC type: #'CogMethod *'>
	| pc object |
	((objectMemory isImmediate: cPIC selector)
	or: [objectMemory isMarked: cPIC selector]) ifFalse:
		[^true].

	"First jump is unconditional; subsequent ones are conditional."
	"Check the potential method oop for the first case only.
	 Inline cache tags for the 1st case are at the send site."
	pc := self addressOfEndOfCase: 1 inCPIC: cPIC.
	(objectRepresentation couldBeObject: (object := backEnd literalBeforeFollowingAddress: pc - backEnd jumpLongByteSize)) ifTrue:
		[(objectMemory isMarked: object) ifFalse:
			[^true]].

	"Check the first target"
	(self markAndTraceOrFreePICTarget: (backEnd jumpLongTargetBeforeFollowingAddress: pc) in: cPIC) ifTrue:
		[^true].

	2 to: cPIC cPICNumCases do:
		[:i| 
		pc := self addressOfEndOfCase: i inCPIC: cPIC.
		(self inlineCacheTagsAreIndexes not
		 and: [objectRepresentation inlineCacheTagsMayBeObjects
		 and: [objectRepresentation couldBeObject: (object := backEnd literal32BeforeFollowingAddress: pc - backEnd jumpLongConditionalByteSize)]]) ifTrue:
			[(objectMemory isMarked: object) ifFalse:
				[^true]].
		"Check the potential method oop for subsequent cases."
		(objectRepresentation couldBeObject: (object := backEnd literalBeforeFollowingAddress: pc - backEnd jumpLongConditionalByteSize - backEnd cmpC32RTempByteSize)) ifTrue:
			[(objectMemory isMarked: object) ifFalse:
				[^true]].
		"Check subsequent targets"
		(self markAndTraceOrFreePICTarget: (backEnd jumpLongConditionalTargetBeforeFollowingAddress: pc) in: cPIC) ifTrue:
			[^true]].

	^false! !

!Cogit methodsFor: 'accessing' stamp: 'eem 11/17/2015 17:28'!
closedPICSize
	"For Cogit clas>>#genAndDisPICoptions:"
	<doNotGenerate>
	^closedPICSize! !

!Cogit methodsFor: 'accessing' stamp: 'eem 12/21/2008 19:29'!
coInterpreter
	<doNotGenerate>
	^coInterpreter! !

!Cogit methodsFor: 'generate machine code - dual mapped zone support' stamp: 'eem 1/29/2020 11:47'!
codeByteAt: address put: aByte
	"production uses the macro..."
	<cmacro: '(address,value) byteAtput((address) + codeToDataDelta, value)'>
	self codeWriteBreakpoint: address.
	"simulation writes twice if simulating dual mapping..."
	codeToDataDelta ~= 0 ifTrue:
		[objectMemory byteAt: address + codeToDataDelta put: aByte].
	^objectMemory byteAt: address put: aByte! !

!Cogit methodsFor: 'debugging' stamp: 'eem 3/30/2009 13:51'!
codeEntryFor: address
	<api>
	<returnTypeC: #'char *'>
	<var: #address type: #'char *'>
	0 to: trampolineTableIndex - 3 by: 2 do:
		[:i|
		(address between: (trampolineAddresses at: i + 1)
				and: (trampolineAddresses at: i + 3) - 1) ifTrue:
			[^trampolineAddresses at: i + 1]].
	^nil! !

!Cogit methodsFor: 'debugging' stamp: 'eem 3/30/2009 13:50'!
codeEntryNameFor: address
	<api>
	<returnTypeC: #'char *'>
	<var: #address type: #'char *'>
	0 to: trampolineTableIndex - 3 by: 2 do:
		[:i|
		(address between: (trampolineAddresses at: i + 1)
				and: (trampolineAddresses at: i + 3) - 1) ifTrue:
			[^trampolineAddresses at: i]].
	^nil! !

!Cogit methodsFor: 'generate machine code - dual mapped zone support' stamp: 'eem 1/27/2020 08:52'!
codeLong32At: address put: anInt
	"production uses the macro..."
	<cmacro: '(address,value) long32Atput((address) + codeToDataDelta, value)'>
	self codeWriteBreakpoint: address.
	"simulation writes twice if simulating dual mapping..."
	codeToDataDelta ~= 0 ifTrue:
		[objectMemory long32At: address + codeToDataDelta put: anInt].
	^objectMemory long32At: address put: anInt! !

!Cogit methodsFor: 'generate machine code - dual mapped zone support' stamp: 'eem 1/27/2020 08:52'!
codeLong64At: address put: anInt
	"production uses the macro..."
	<cmacro: '(address,value) long64Atput((address) + codeToDataDelta, value)'>
	self codeWriteBreakpoint: address.
	"simulation writes twice if simulating dual mapping..."
	codeToDataDelta ~= 0 ifTrue:
		[objectMemory long64At: address + codeToDataDelta put: anInt].
	^objectMemory long64At: address put: anInt! !

!Cogit methodsFor: 'generate machine code - dual mapped zone support' stamp: 'eem 1/29/2020 11:47'!
codeLongAt: address put: aLong
	"production uses the macro..."
	<cmacro: '(address,value) longAtput((address) + codeToDataDelta, value)'>
	self codeWriteBreakpoint: address.
	"simulation writes twice if simulating dual mapping..."
	codeToDataDelta ~= 0 ifTrue:
		[objectMemory longAt: address + codeToDataDelta put: aLong].
	^objectMemory longAt: address put: aLong! !

!Cogit methodsFor: 'generate machine code - dual mapped zone support' stamp: 'eem 1/29/2020 09:46'!
codeMemcpy: dest _: src _: bytes
	"Copy memory applying the codeToDataDelta.  This is used for creating PICs and so
	 works on writable addresses (no delta required)."
	"production uses the macro..."
	<cmacro: '(dest,src,bytes) memcpy(dest,src,bytes)'>
	self codeWriteBreakpoint: dest.
	"simulation writes twice if simulating dual mapping..."
	codeToDataDelta ~= 0 ifTrue:
		[objectMemory memcpy: dest asInteger - codeToDataDelta _: src _: bytes].
	objectMemory memcpy: dest _: src _: bytes! !

!Cogit methodsFor: 'generate machine code - dual mapped zone support' stamp: 'eem 1/1/2021 20:39'!
codeMemmove: dest _: src _: bytes
	"Move memory (copy allowing for overlap), applying the codeToDataDelta.
	 This is used for code compaction and so works on readable addresses (delta is required)."
	"production uses the macro..."
	<cmacro: '(dest,src,bytes) memmove((char *)(dest)+codeToDataDelta,src,bytes)'>
	self codeWriteBreakpoint: dest.
	"simulation writes twice if simulating dual mapping..."
	codeToDataDelta ~= 0 ifTrue:
		[objectMemory memmove: dest asUnsignedInteger + codeToDataDelta _: src _: bytes].
	objectMemory memmove: dest _: src _: bytes.
	"and as a convenience, relocate the breakPC(s) if in this method..."
	breakPC ifNotNil:
		[breakPC isArray
			ifTrue:
				[breakPC withIndexDo:
					[:pc :idx|
					(pc between: src and: src + bytes - 1) ifTrue:
						[breakPC at: idx put: pc - (src asInteger - dest asInteger)]]]
			ifFalse:
				[(breakPC between: src and: src + bytes - 1) ifTrue:
					[breakPC := breakPC - (src asInteger - dest asInteger)]]]! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 12/28/2019 01:07'!
codeRangesFor: cogMethod
	"Answer a sequence of ranges of code for the main method and all of the blocks in a CogMethod.
	 N.B.  These are in order of block dispatch, _not_ necessarily address order in the method."
	<doNotGenerate>
	| pc end blockEntry starts |
	cogMethod cmType = CMClosedPIC ifTrue:
		[end := false
					ifTrue: [cogMethod asInteger + cPICEndOfCodeOffset - backEnd jumpLongByteSize]
					ifFalse: [cogMethod asInteger + closedPICSize - 1].
		 ^{ CogCodeRange
				from: cogMethod asInteger + (self sizeof: CogMethod)
				to: end
				cogMethod: cogMethod
				startpc: nil }].
	end := (self mapEndFor: cogMethod) - 1.
	cogMethod blockEntryOffset = 0 ifTrue:
		[^{ CogCodeRange
				from: cogMethod asInteger + (self sizeof: CogMethod)
				to: end
				cogMethod: cogMethod
				startpc: (cogMethod cmType ~= CMOpenPIC ifTrue:
							[coInterpreter startPCOfMethodHeader: cogMethod methodHeader]) }].
	pc := blockEntry := cogMethod blockEntryOffset + cogMethod asInteger.
	starts := OrderedCollection with: cogMethod.
	[pc < end] whileTrue:
		[| targetpc |
		 targetpc := blockEntry.
		 (backEnd isJumpAt: pc) ifTrue:
			[targetpc := backEnd jumpTargetPCAt: pc.
			 targetpc < blockEntry ifTrue:
				[starts add: (self cCoerceSimple: targetpc - (self sizeof: CogBlockMethod) to: #'CogBlockMethod *')]].
		 pc := pc + (backEnd instructionSizeAt: pc)].
	starts := starts asSortedCollection.
	^(1 to: starts size + 1) collect:
		[:i| | cogSubMethod nextpc |
		i <= starts size
			ifTrue:
				[cogSubMethod := starts at: i.
				 nextpc := i < starts size ifTrue: [(starts at: i + 1) address] ifFalse: [blockEntry].
				 CogCodeRange
					from: cogSubMethod address + (self sizeof: cogSubMethod)
					to: nextpc - 1
					cogMethod: cogSubMethod
					startpc: (i = 1
								ifTrue: [coInterpreter startPCOfMethodHeader: cogMethod methodHeader]
								ifFalse: [cogSubMethod startpc])]
			ifFalse:
				[CogCodeRange
					from: blockEntry
					to: end]]! !

!Cogit methodsFor: 'generate machine code - dual mapped zone support' stamp: 'eem 7/5/2021 16:15'!
codeWriteBreakpoint: address
	<doNotGenerate>
	codeZoneIsExecutableNotWritable ifTrue: [self deny: codeZoneIsExecutableNotWritable].
	"(address = 16r520) ifTrue:
		[self halt]"
	"(address between: 16r2398 and: 16r23B0) ifTrue:
		[self halt]"! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 8/31/2021 14:31'!
cog: aMethodObj selector: aSelectorOop
	"Attempt to produce a machine code method for the bytecode method
	 object aMethodObj.  N.B. If there is no code memory available do *NOT*
	 attempt to reclaim the method zone.  Certain clients (e.g. ceSICMiss:)
	 depend on the zone remaining constant across method generation."
	<api>
	<returnTypeC: #'CogMethod *'>
	| selector cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	(self exclude: aMethodObj selector: aSelectorOop) ifTrue:
		[^nil].
	"In Newspeak we support anonymous accessors and hence tolerate the same
	 method being cogged multiple times.  But only if the method class association is nil."
	NewspeakVM
		ifTrue:
			[(coInterpreter methodHasCogMethod: aMethodObj) ifTrue:
				[cogMethod := coInterpreter cogMethodOf: aMethodObj.
				 self deny: cogMethod selector = aSelectorOop.
				 cogMethod selector = aSelectorOop ifTrue:
					[^cogMethod].
				 (coInterpreter methodClassAssociationOf: aMethodObj) ~= objectMemory nilObject ifTrue:
					[self cCode: 'extern void *firstIndexableField(sqInt)'. "Slang, au natural"
					 self warnMultiple: cogMethod selectors: aSelectorOop.
					^nil]]]
		ifFalse: [self deny: (coInterpreter methodHasCogMethod: aMethodObj)].
	self deny: (objectMemory isOopCompiledMethod: (coInterpreter ultimateLiteralOf: aMethodObj)).
	selector := aSelectorOop = objectMemory nilObject
					ifTrue: [coInterpreter maybeSelectorOfMethod: aMethodObj]
					ifFalse: [aSelectorOop].
	"coInterpreter stringOf: selector"
	selector ifNotNil:
		[coInterpreter compilationBreakpoint: selector isMNUCase: false].
	aMethodObj = breakMethod ifTrue: [self halt: 'Compilation of breakMethod'].
	NewspeakVM ifTrue:
		[cogMethod := methodZone findPreviouslyCompiledVersionOf: aMethodObj with: aSelectorOop.
		 cogMethod ifNotNil:
			[(coInterpreter methodHasCogMethod: aMethodObj) not ifTrue:
				[self assert: (coInterpreter rawHeaderOf: aMethodObj) = cogMethod methodHeader.
				 cogMethod methodObject: aMethodObj.
				 coInterpreter rawHeaderOf: aMethodObj put: cogMethod asInteger].
			^cogMethod]].
	"If the generators for the alternate bytecode set are missing then interpret."
	(coInterpreter methodUsesAlternateBytecodeSet: aMethodObj)
		ifTrue:
			[(self numElementsIn: generatorTable) <= 256 ifTrue:
				[^nil].
			 bytecodeSetOffset := 256]
		ifFalse:
			[bytecodeSetOffset := 0].
	objectRepresentation ensureNoForwardedLiteralsIn: aMethodObj.
	methodObj := aMethodObj.
	methodHeader := objectMemory methodHeaderOf: aMethodObj.
	receiverTags := -1. "lazy initialization"
	cogMethod := self compileCogMethod: aSelectorOop.
	(cogMethod asInteger between: MaxNegativeErrorCode and: -1) ifTrue:
		[cogMethod asInteger = InsufficientCodeSpace ifTrue:
			[coInterpreter callForCogCompiledCodeCompaction].
		 self maybeFreeCounters.
		 "Right now no errors should be reported, so nothing more to do."
		 "self reportError: (self cCoerceSimple: cogMethod to: #sqInt)."
		 ^nil].
	"self cCode: ''
		inSmalltalk:
			[coInterpreter printCogMethod: cogMethod.
			 ""coInterpreter symbolicMethod: aMethodObj.""
			 self assertValidMethodMap: cogMethod."
			 "self disassembleMethod: cogMethod."
			 "printInstructions := clickConfirm := true""]."
	^cogMethod! !

!Cogit methodsFor: 'simulation only' stamp: 'eem (const to msg rewrite) 11/6/2014 11:21'!
cogBlockMethodSurrogateAt: address
	<doNotGenerate>
	self assert: (address bitAnd: objectMemory wordSize - 1) = 0.
	^cogBlockMethodSurrogateClass new
		at: address
		objectMemory: objectMemory
		cogit: self! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 2/15/2011 10:21'!
cogBlockMethodSurrogateClass
	<doNotGenerate>
	^cogBlockMethodSurrogateClass! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 1/10/2010 16:17'!
cogCodeBase
	<api>
	^codeBase! !

!Cogit methodsFor: 'profiling primitives' stamp: 'eem 6/1/2018 12:22'!
cogCodeConstituents: withDetails
	"Answer the contents of the code zone as an array of pair-wise element, address in ascending address order.
	 Answer a string for a runtime routine or abstract label (beginning, end, etc), a CompiledMethod for a CMMethod,
	 or a selector (presumably a Symbol) for a PIC.
	 If withDetails is true
		- answer machine-code to bytecode pc mapping information for methods
		- answer class, target pair information for closed PIC
	 N.B. Since the class tag for the first case of a closed PIC is stored at the send site, it must be collected
		  by scanning methods (see collectCogConstituentFor:Annotation:Mcpc:Bcpc:Method:).  Since closed PICs
		  are never shared they always come after the method that references them, so we don't need an extra pass
		  to collect the first case class tags, which are (temporarily) assigned to each closed PIC's methodObject field.
		  But we do need to reset the methodObject fields to zero.  This is done in createPICData:, unless memory
		  runs out, in which case it is done by cleanUpFailingCogCodeConstituents:."
	<api>
	| count cogMethod constituents label value |
	<var: #cogMethod type: #'CogMethod *'>
	count := trampolineTableIndex / 2 + 3. "+ 3 for start, freeStart and end"
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType ~= CMFree ifTrue:
			[count := count + 1].
		cogMethod := methodZone methodAfter: cogMethod].
	constituents := coInterpreter instantiateClass: coInterpreter classArray indexableSize: count * 2.
	constituents ifNil:
		[^constituents].
	coInterpreter pushRemappableOop: constituents.
	((label := objectMemory stringForCString: 'CogCode') isNil
	 or: [(value := self positiveMachineIntegerFor: codeBase) isNil]) ifTrue:
		[coInterpreter popRemappableOop.
		 ^nil].
	coInterpreter
		storePointerUnchecked: 0 ofObject: (self maybeTopRemapped: constituents) withValue: label;
		storePointerUnchecked: 1 ofObject: (self maybeTopRemapped: constituents) withValue: value.
	0 to: trampolineTableIndex - 1 by: 2 do:
		[:i|
		((label := objectMemory stringForCString: (trampolineAddresses at: i)) isNil
		 or: [(value := self positiveMachineIntegerFor: (trampolineAddresses at: i + 1) asUnsignedInteger) isNil]) ifTrue:
			[coInterpreter popRemappableOop.
			 ^nil].
		coInterpreter
			storePointerUnchecked: 2 + i ofObject: (self maybeTopRemapped: constituents) withValue: label;
			storePointerUnchecked: 3 + i ofObject: (self maybeTopRemapped: constituents) withValue: value].
	count := trampolineTableIndex + 2.
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType ~= CMFree ifTrue:
			[| profileData |
			 profileData := self profileDataFor: cogMethod withDetails: withDetails.
			 profileData ifNil: [^self cleanUpFailingCogCodeConstituents: cogMethod].
			 coInterpreter
				storePointerUnchecked: count
				ofObject: (self maybeTopRemapped: constituents)
				withValue: profileData.
			value := withDetails
						ifTrue: [self collectCogMethodConstituent: cogMethod]
						ifFalse: [self positiveMachineIntegerFor: cogMethod asUnsignedInteger].
			value ifNil: [^self cleanUpFailingCogCodeConstituents: cogMethod].
			coInterpreter
						storePointerUnchecked: count + 1
						ofObject: (self maybeTopRemapped: constituents)
						withValue: value.
			 count := count + 2].
		cogMethod := methodZone methodAfter: cogMethod].
	((label := objectMemory stringForCString: 'CCFree') isNil
	 or: [(value := self positiveMachineIntegerFor: methodZone zoneFree) isNil]) ifTrue:
		[coInterpreter popRemappableOop.
		 ^nil].
	coInterpreter
		storePointerUnchecked: count ofObject: (self maybeTopRemapped: constituents) withValue: label;
		storePointerUnchecked: count + 1 ofObject: (self maybeTopRemapped: constituents) withValue: value.
	((label := objectMemory stringForCString: 'CCEnd') isNil
	 or: [(value := self positiveMachineIntegerFor: methodZone zoneEnd) isNil]) ifTrue:
		[coInterpreter popRemappableOop.
		 ^nil].
	coInterpreter
		storePointerUnchecked: count + 2 ofObject: (self maybeTopRemapped: constituents) withValue: label;
		storePointerUnchecked: count + 3 ofObject: (self maybeTopRemapped: constituents) withValue: value.
	constituents := coInterpreter popRemappableOop.
	coInterpreter beRootIfOld: constituents.
	"would like to assert this, but it requires the leak checked be run :-(
		self assert: self allMachineCodeObjectReferencesValid."
	^constituents! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 7/5/2021 15:45'!
cogExtendPIC: cPIC CaseNMethod: caseNMethod tag: caseNTag isMNUCase: isMNUCase
	"Extend the cPIC with the supplied case.  If caseNMethod is cogged dispatch direct to
	 its unchecked entry-point.  If caseNMethod is not cogged, jump to the fast interpreter
	 dispatch, and if isMNUCase then dispatch to fast MNU invocation and mark the cPIC as
	 having the MNU case for cache flushing."
 	<var: #cPIC type: #'CogMethod *'>
	| operand target address |

	coInterpreter compilationBreakpoint: cPIC selector classTag: caseNTag isMNUCase: isMNUCase.

	self assert: (objectRepresentation inlineCacheTagIsYoung: caseNTag) not.
	"Caller patches to open pic if caseNMethod is young."
	self assert: (caseNMethod notNil and: [(objectMemory isYoung: caseNMethod) not]).
	(isMNUCase not and: [coInterpreter methodHasCogMethod: caseNMethod])
		ifTrue: "this isn't an MNU and we have an already cogged method to jump to"
			[operand := 0.
			 target := (coInterpreter cogMethodOf: caseNMethod) asInteger + cmNoCheckEntryOffset]
		ifFalse: 
			[operand := caseNMethod.
			 isMNUCase
				ifTrue: "this is an MNU so tag the CPIC header and setup a jump to the MNUAbort"
					[(self writableMethodFor: cPIC) cpicHasMNUCase: true.
					 target := cPIC asInteger + (self sizeof: CogMethod)]
				ifFalse: "setup a jump to the interpretAborth so we can cog the target method"
					[target := cPIC asInteger + self picInterpretAbortOffset]].

	"find the end address of the new case"
	address := self addressOfEndOfCase: cPIC cPICNumCases +1 inCPIC: cPIC.
	
	self rewriteCPICCaseAt: address tag: caseNTag objRef: operand target: target.

	"finally, rewrite the jump 3 instr  before firstCPICCaseOffset to jump to the beginning of this new case"
	self rewriteCPIC: cPIC caseJumpTo: address - cPICCaseSize.

	"and update the header flag for the number of cases"
	(self writableMethodFor: cPIC) cPICNumCases: cPIC cPICNumCases + 1.

	"This also implicitly flushes the read/write mapped dual zone to the read/execute zone."
	backEnd flushICacheFrom: cPIC asUnsignedInteger to: cPIC asUnsignedInteger + closedPICSize.
	self assertValidDualZoneFrom: cPIC asUnsignedInteger to: cPIC asUnsignedInteger + closedPICSize! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 8/31/2021 14:31'!
cogFullBlockMethod: aMethodObj numCopied: numCopied
	"Attempt to produce a machine code method for the bytecode method
	 object aMethodObj.  N.B. If there is no code memory available do *NOT*
	 attempt to reclaim the method zone.  Certain clients (e.g. ceSICMiss:)
	 depend on the zone remaining constant across method generation."
	<api>
	<option: #SistaV1BytecodeSet>
	<returnTypeC: #'CogMethod *'>
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	(self exclude: aMethodObj) ifTrue:
		[^nil].
	self deny: (coInterpreter methodHasCogMethod: aMethodObj).
	self assert: (objectMemory isOopCompiledMethod: (coInterpreter ultimateLiteralOf: aMethodObj)).
	aMethodObj = breakMethod ifTrue: [self halt: 'Compilation of breakMethod'].
	"If the generators for the alternate bytecode set are missing then interpret."
	(coInterpreter methodUsesAlternateBytecodeSet: aMethodObj)
		ifTrue:
			[(self numElementsIn: generatorTable) <= 256 ifTrue:
				[^nil].
			 bytecodeSetOffset := 256]
		ifFalse:
			[bytecodeSetOffset := 0].
	objectRepresentation ensureNoForwardedLiteralsIn: aMethodObj.
	methodObj := aMethodObj.
	methodHeader := objectMemory methodHeaderOf: aMethodObj.
	receiverTags := -1. "lazy initialization"
	cogMethod := self compileCogFullBlockMethod: numCopied.
	(cogMethod asInteger between: MaxNegativeErrorCode and: -1) ifTrue:
		[cogMethod asInteger = InsufficientCodeSpace ifTrue:
			[coInterpreter callForCogCompiledCodeCompaction].
		 self maybeFreeCounters.
		 "Right now no errors should be reported, so nothing more to do."
		 "self reportError: (self cCoerceSimple: cogMethod to: #sqInt)."
		 ^nil].
	"self cCode: ''
		inSmalltalk:
			[coInterpreter printCogMethod: cogMethod.
			 ""coInterpreter symbolicMethod: aMethodObj.""
			 self assertValidMethodMap: cogMethod."
			 "self disassembleMethod: cogMethod."
			 "printInstructions := clickConfirm := true""]."
	^cogMethod! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 7/5/2021 17:20'!
cogMNUPICSelector: selector receiver: rcvr methodOperand: methodOperand numArgs: numArgs
	<api>
	"Attempt to create a one-case PIC for an MNU.
	 The tag for the case is at the send site and so doesn't need to be generated."
	<returnTypeC: #'CogMethod *'>
	| startAddress writablePIC actualPIC |
	((objectMemory isYoung: selector)
	 or: [(objectRepresentation inlineCacheTagForInstance: rcvr) = self picAbortDiscriminatorValue]) ifTrue:
		[^0].
	coInterpreter compilationBreakpoint: selector classTag: (objectMemory fetchClassTagOf: rcvr) isMNUCase: true.
	self assert: endCPICCase0 notNil.

	"get memory in the code zone for the CPIC; if that fails we return an error code for the sender to use to work out how to blow up"
	startAddress := methodZone allocate: closedPICSize.
	startAddress = 0 ifTrue:
		[coInterpreter callForCogCompiledCodeCompaction.
		 ^0].
	self maybeBreakGeneratingFrom: startAddress to: startAddress + closedPICSize.

	self ensureWritableCodeZone.
	writablePIC := self writableMethodFor: startAddress.
	"memcpy the prototype across to our allocated space; because anything else would be silly"
	self codeMemcpy: writablePIC _: cPICPrototype _: closedPICSize.

	self
		fillInCPICHeader: writablePIC
		numArgs: numArgs
		numCases: 1
		hasMNUCase: true
		selector: selector.

	self configureMNUCPIC: (actualPIC := self cCoerceSimple: startAddress to: #'CogMethod *')
		methodOperand: methodOperand
		numArgs: numArgs
		delta: startAddress - cPICPrototype asUnsignedInteger.

	"This also implicitly flushes the read/write mapped dual zone to the read/execute zone."
	backEnd flushICacheFrom: startAddress to: startAddress + closedPICSize.

	self assert: (backEnd callTargetFromReturnAddress: startAddress + missOffset) = (self picAbortTrampolineFor: numArgs).
	self assertValidDualZoneFrom: startAddress to: startAddress + closedPICSize.

	^actualPIC! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 11/14/2020 18:06'!
cogMethodContaining: mcpc
	<doNotGenerate>
	^methodZone cogMethodContaining: mcpc! !

!Cogit methodsFor: 'debugging' stamp: 'eem 1/29/2020 13:43'!
cogMethodDoesntLookKosher: cogMethod
	"Check that the header fields onf a non-free method are consistent with
	 the type. Answer 0 if it is ok, otherwise answer a code for the error."
	<api>
	<inline: false>
	<var: #cogMethod type: #'CogMethod *'>
	((cogMethod blockSize bitAnd: objectMemory wordSize - 1) ~= 0
	 or: [cogMethod blockSize < (self sizeof: CogMethod)
	 or: [cogMethod blockSize >= 32768]]) ifTrue:
		[^1].

	cogMethod cmType = CMFree ifTrue: [^2].

	cogMethod cmType = CMMethod ifTrue:
		[(objectMemory isIntegerObject: cogMethod methodHeader) ifFalse:
			[^11].
		 (objectRepresentation couldBeObject: cogMethod methodObject) ifFalse:
			[^12].
		 (cogMethod stackCheckOffset > 0
		  and: [cogMethod stackCheckOffset < cmNoCheckEntryOffset]) ifTrue:
			[^13].
		 (SistaVM
		  and: [objectRepresentation canPinObjects
		  and: [cogMethod counters ~= 0]]) ifTrue:
			[(objectRepresentation couldBeDerivedObject: cogMethod counters) ifFalse:
				[^14]].
		 (NewspeakVM
		  and: [objectRepresentation canPinObjects
		  and: [cogMethod nextMethodOrIRCs ~= 0]]) ifTrue:
			[(cogMethod nextMethodOrIRCs < methodZone zoneEnd)
				ifTrue: "check the nextMethod (unpairedMethodList) unless we're compacting."
					[(methodZone compactionInProgress
					  or: [cogMethod nextMethodOrIRCs = (methodZone methodFor: cogMethod nextMethodOrIRCs asVoidPointer) asUnsignedInteger]) ifFalse:
						[^15]]
				ifFalse:
					[(objectRepresentation couldBeDerivedObject: cogMethod nextMethodOrIRCs) ifFalse:
						[^16]]].
		 ^0].

	cogMethod cmType = CMOpenPIC ifTrue:
		[cogMethod blockSize ~= openPICSize ifTrue:
			[^21].
		 cogMethod methodHeader ~= 0 ifTrue:
			[^22].
		 "Check the nextOpenPIC link unless we're compacting"
		 cogMethod objectHeader >= 0 ifTrue:
			[(cogMethod methodObject = 0
			  or: [methodZone compactionInProgress
			  or: [cogMethod methodObject = (methodZone methodFor: cogMethod methodObject asVoidPointer) asUnsignedInteger]]) ifFalse:
				[^23]].
		 cogMethod stackCheckOffset ~= 0 ifTrue:
			[^24].
		 ^0].

	cogMethod cmType = CMClosedPIC ifTrue:
		[cogMethod blockSize ~= closedPICSize ifTrue:
			[^31].
		 (cogMethod cPICNumCases between: 1 and: MaxCPICCases) ifFalse:
			[^32].
		 cogMethod methodHeader ~= 0 ifTrue:
			[^33].
		 cogMethod methodObject ~= 0 ifTrue:
			[^34].
		 ^0].

	^9! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 2/15/2011 10:20'!
cogMethodOrBlockSurrogateAt: address
	<doNotGenerate>
	| surrogate |
	surrogate := self cogMethodSurrogateAt: address.
	^surrogate cmType = CMBlock
		ifTrue: [self cogBlockMethodSurrogateAt: address]
		ifFalse: [surrogate]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem (const to msg rewrite) 11/6/2014 11:21'!
cogMethodSurrogateAt: address
	<doNotGenerate>
	self assert: (address < 0 or: [(address bitAnd: objectMemory wordSize - 1) = 0]).
	^cogMethodSurrogateClass new
		at: address
		objectMemory: objectMemory
		cogit: self! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 2/15/2011 10:21'!
cogMethodSurrogateClass
	<doNotGenerate>
	^cogMethodSurrogateClass! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 7/9/2021 15:10'!
cogOpenPICSelector: selector numArgs: numArgs
	"Create an Open PIC.  Temporarily create a direct call of ceSendFromOpenPIC:.
	 Should become a probe of the first-level method lookup cache followed by a
	 call of ceSendFromOpenPIC: if the probe fails."
	<returnTypeC: #'CogMethod *'>
	| startAddress codeSize mapSize end |
	coInterpreter compilationBreakpoint: selector isMNUCase: false.
	startAddress := methodZone allocate: openPICSize.
	startAddress = 0 ifTrue:
		[^self cCoerceSimple: InsufficientCodeSpace to: #'CogMethod *'].
	methodLabel
		address: startAddress;
		dependent: nil.
	"stack allocate the various collections so that they
	 are effectively garbage collected on return."
	self allocateOpcodes: 100 bytecodes: 0.
	self compileOpenPIC: selector numArgs: numArgs.
	self computeMaximumSizes.
	methodLabel concretizeAt: startAddress.
	codeSize := self generateInstructionsAt: startAddress + (self sizeof: CogMethod).
	self ensureWritableCodeZone. "other clients first use generateMapAt: for sizing; this is an exception"
	mapSize := self generateMapAt: startAddress + openPICSize - 1 start: startAddress + cmNoCheckEntryOffset.
	self assert: entry address - startAddress = cmEntryOffset.
	self assert: (methodZone roundUpLength: (self sizeof: CogMethod) + codeSize) + (methodZone roundUpLength: mapSize) <= openPICSize.
	end := self outputInstructionsAt: startAddress + (self sizeof: CogMethod).
	self
		fillInOPICHeader: (self writableMethodFor: startAddress)
		numArgs: numArgs
		selector: selector.
	^self cCoerceSimple: startAddress to: #'CogMethod *'! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 7/5/2021 16:24'!
cogPICSelector: selector numArgs: numArgs Case0Method: case0CogMethod Case1Method: case1MethodOrNil tag: case1Tag isMNUCase: isMNUCase
	"Attempt to create a two-case PIC for case0CogMethod and  case1Method,case1Tag.
	 The tag for case0CogMethod is at the send site and so doesn't need to be generated.
	 case1Method may be any of
		- a Cog method; link to its unchecked entry-point
		- a CompiledMethod; link to ceInterpretMethodFromPIC:
		- a CompiledMethod; link to ceMNUFromPICMNUMethod:receiver:"
	<var: #case0CogMethod type: #'CogMethod *'>
	<returnTypeC: #'CogMethod *'>
	| startAddress writablePIC actualPIC |
	(objectMemory isYoung: selector) ifTrue:
		[^self cCoerceSimple: YoungSelectorInPIC to: #'CogMethod *'].
	coInterpreter compilationBreakpoint: selector classTag: case1Tag isMNUCase: isMNUCase.
	
	"get memory in the code zone for the CPIC; if that fails we return an error code for the sender to use to work out how to blow up"
	startAddress := methodZone allocate: closedPICSize.
	startAddress = 0 ifTrue:
		[^self cCoerceSimple: InsufficientCodeSpace to: #'CogMethod *'].
	self maybeBreakGeneratingFrom: startAddress to: startAddress + closedPICSize.

	writablePIC := self writableMethodFor: startAddress.
	"memcpy the prototype across to our allocated space; because anything else would be silly"
	self codeMemcpy: writablePIC _: cPICPrototype _: closedPICSize.

	self
		fillInCPICHeader: writablePIC
		numArgs: numArgs
		numCases: 2
		hasMNUCase: isMNUCase
		selector: selector.

	self configureCPIC: (actualPIC := self cCoerceSimple: startAddress to: #'CogMethod *')
		Case0: case0CogMethod
		Case1Method: case1MethodOrNil
		tag: case1Tag
		isMNUCase: isMNUCase
		numArgs: numArgs
		delta: startAddress - cPICPrototype asUnsignedInteger.

	self assert: (backEnd callTargetFromReturnAddress: startAddress + missOffset) = (self picAbortTrampolineFor: numArgs).

	^actualPIC! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 2/18/2011 15:39'!
cogit
	"This is for the sizeof: CogMethod hook that allows different cogit classes to use differet CogMethod variants."
	<doNotGenerate>
	^self! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 7/5/2021 16:00'!
cogitPostGCAction: gcMode
	<api>
	(gcMode = GCModeFull
	 and: [objectRepresentation allYoungObjectsAgeInFullGC]) ifTrue:
		[methodZone voidYoungReferrersPostTenureAll].
	self cppIf: SPURVM ifTrue:
		[gcMode = GCModeBecome ifTrue:
			[methodZone followForwardedLiteralsInOpenPICList]].
	"Post-GC update every full method's objectHeader to whatever it needs to be"
	self assert: self allMethodsHaveCorrectHeader.
	"The youngReferrers should be correct after a GC since that is the point at which it is
	 pruned.  But at other times false positives or free methods on the list are acceptable."
	self assert: ((gcMode noMask: GCModeFull+GCModeNewSpace) or: [methodZone kosherYoungReferrers]).
	self ensureExecutableCodeZone! !

!Cogit methodsFor: 'profiling primitives' stamp: 'eem 1/29/2020 13:41'!
collectCogConstituentFor: descriptor Annotation: isBackwardBranchAndAnnotation Mcpc: mcpc Bcpc: bcpc Method: cogMethodArg
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<var: #mcpc type: #'char *'>
	<var: #cogMethodArg type: #'void *'>
	<var: #targetMethod type: #'CogMethod *'>
	| address entryPoint |
	descriptor ifNil: [^0].
	descriptor isMapped ifFalse: [^0].
	address := self positiveMachineIntegerFor: mcpc asUnsignedInteger.
	address ifNil: [^PrimErrNoMemory]. "This cannot trigger a GC but fails if not enough space in Eden,"
	"Assumes we write the values into topRemappableOop"
	coInterpreter
		storePointerUnchecked: cogConstituentIndex
		ofObject: coInterpreter topRemappableOop
		withValue: address.
	coInterpreter
		storePointerUnchecked: cogConstituentIndex + 1
		ofObject: coInterpreter topRemappableOop
		withValue: (objectMemory integerObjectOf: bcpc).
	cogConstituentIndex := cogConstituentIndex + 2.

	"Collect any first case classTags for closed PICs."
	((isBackwardBranchAndAnnotation noMask: 1)
	 and: [self isSendAnnotation: isBackwardBranchAndAnnotation >> 1]) ifTrue:
		[entryPoint := backEnd callTargetFromReturnAddress: mcpc asInteger.
		 entryPoint > methodZoneBase ifTrue: "send is linked"
			[self targetMethodAndSendTableFor: entryPoint annotation: isBackwardBranchAndAnnotation >> 1 into:
				[:targetMethod :sendTable|
				  targetMethod cmType = CMClosedPIC ifTrue:
					[targetMethod methodObject: (objectRepresentation classForInlineCacheTag: (backEnd inlineCacheTagAt: mcpc))]]]].
	^0! !

!Cogit methodsFor: 'profiling primitives' stamp: 'cb 5/22/2017 16:55'!
collectCogMethodConstituent: cogMethod
	"Answer a description of the mapping between machine code pointers and bytecode pointers for the Cog Method.
	 First value is the address of the cog method.
	 Following values are pairs of machine code pc and bytecode pc"
	<var: #cogMethod type: #'CogMethod *'>
	<var: #cogBlockMethod type: #'CogBlockMethod *'>
	| cm nSlots errCode cogBlockMethod address data |
	(cogMethod cmType = CMMethod) 
		ifFalse: [^self positiveMachineIntegerFor: cogMethod asUnsignedInteger ].
	cogBlockMethod := self cCoerceSimple: cogMethod to: #'CogBlockMethod *'.
	cogBlockMethod stackCheckOffset = 0 "isFrameless ?"
		ifTrue: [^self positiveMachineIntegerFor: cogMethod asUnsignedInteger].
	cm := cogMethod methodObject.
	nSlots := ((objectMemory byteSizeOf: cm) - (coInterpreter startPCOfMethod: cm)) * 2 + objectMemory minSlotsForShortening + 1."+1 for first address"
	data := objectMemory instantiateClass: (objectMemory splObj: ClassArray) indexableSize: nSlots.
	data ifNil: [^nil].
	coInterpreter pushRemappableOop: data.
	"The iteration assumes the object is the top remappable oop"
	address := (self positiveMachineIntegerFor: cogMethod asUnsignedInteger).
	address ifNil: [coInterpreter popRemappableOop. ^nil].
	coInterpreter
		storePointerUnchecked: 0
		ofObject: coInterpreter topRemappableOop
		withValue: address.
	cogConstituentIndex := 1.
	errCode := self
		mapFor: cogBlockMethod
		bcpc: (coInterpreter startPCOfMethod: cogMethod methodObject)
		performUntil: #collectCogConstituentFor:Annotation:Mcpc:Bcpc:Method:
		arg: cogMethod asVoidPointer.
	errCode ~= 0 ifTrue: [coInterpreter popRemappableOop. ^nil].
	cogConstituentIndex < nSlots ifTrue:
		[objectMemory shorten: coInterpreter topRemappableOop toIndexableSize: cogConstituentIndex].
	^coInterpreter popRemappableOop.! !

!Cogit methodsFor: 'disassembly' stamp: 'rsf 8/10/2016 00:54'!
collectMapEntry: annotation address: mcpc into: aDictionary
	<doNotGenerate>
	aDictionary
		at: mcpc
		ifPresent:
			[:extant|
			aDictionary
				at: mcpc
				put: extant, ':\' withCRs, (self class annotationConstantNames at: annotation + 1)]
		ifAbsent: [ aDictionary at: mcpc put: (self class annotationConstantNames at: annotation + 1)].
	^0! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 8/20/2021 15:49'!
compactCogCompiledCode
	<api>
	self assertValidDualZone.
	self assert: self noCogMethodsMaximallyMarked.

	self moveProfileToMethods.

	self ensureWritableCodeZone.
	coInterpreter markActiveMethodsAndReferents.
	methodZone freeOlderMethodsForCompaction.
	self compactPICsWithFreedTargets.
	methodZone planCompaction.
	coInterpreter updateStackZoneReferencesToCompiledCodePreCompaction.
	methodZone relocateMethodsPreCompaction.
	self assertValidDualZone.
	methodZone compactCompiledCode.

	backEnd
		stopsFrom: methodZone freeStart to: methodZone youngReferrers - 1;
		flushICacheFrom: methodZoneBase asUnsignedInteger
			to: methodZone youngReferrers asUnsignedInteger.

	self assert: self allMethodsHaveCorrectHeader.
	self assert: methodZone kosherYoungReferrers.
	self assertValidDualZone! !

!Cogit methodsFor: 'compaction' stamp: 'eem 1/29/2020 09:37'!
compactPICsWithFreedTargets
	| cogMethod count |
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	count := 0.
	[cogMethod < methodZone limitZony] whileTrue:
		[(cogMethod cmType = CMClosedPIC
		  and: [self cPICCompactAndIsNowEmpty: cogMethod]) ifTrue:
			[(self writableMethodFor: cogMethod) cmType: CMFree].
		 cogMethod := methodZone methodAfter: cogMethod.
		 count := count + 1].
	self assert: count = methodZone numMethods! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 11/25/2010 10:04'!
compilationTrace
	^compilationTrace! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 1/18/2017 08:16'!
compilationTrace: anInteger
	"  1 = method/block compilation
	   2 = bytecode descriptor.
	   4 = simStack & optStatus
	   8 = spill
	 16 = merge
	 32 = fixup
	 64 = map"
	compilationTrace := anInteger! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 12/6/2014 11:06'!
compileAbort
	"The start of a CogMethod has a call to a run-time abort routine that either
	 handles an in-line cache failure or a stack overflow.  The routine selects the
	 path depending on ReceiverResultReg; if zero it takes the stack overflow
	 path; if nonzero the in-line cache miss path.  Neither of these paths returns.
	 The abort routine must be called;  In the callee the method is located by
	 adding the relevant offset to the return address of the call.

	 N.B. This code must match that in compilePICAbort: so that the offset of the
	 return address of the call is the same in methods and closed PICs."
	<returnTypeC: #'AbstractInstruction *'>
	stackOverflowCall := self MoveCq: 0 R: ReceiverResultReg.
	backEnd hasLinkRegister
		ifTrue:
			["If there is a link register it must be saved (pushed onto the stack) before it
			  is smashed by the abort call, and hence needs to be manually handled here"
			 sendMiss := self PushR: LinkReg.
			 ^self Call: (self methodAbortTrampolineFor: methodOrBlockNumArgs)]
		ifFalse:
			[^sendMiss := self Call: (self methodAbortTrampolineFor: methodOrBlockNumArgs)]! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 02/09/2017 09:39'!
compileAbstractInstructionsFrom: start through: end
	"Loop over bytecodes, dispatching to the generator for each bytecode, handling fixups in due course."
	| nextOpcodeIndex descriptor fixup result nExts |
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<var: #fixup type: #'BytecodeFixup *'>
	bytecodePC := start.
	nExts := result := 0.
	descriptor := nil.
	[self maybeHaltIfDebugPC.
	 descriptor := self loadBytesAndGetDescriptor.
	 nextOpcodeIndex := opcodeIndex.
	 result := self perform: descriptor generator.
	 self assertExtsAreConsumed: descriptor.
	 fixup := self fixupAt: bytecodePC.
	 self patchFixupTargetIfNeeded: fixup nextOpcodeIndex: nextOpcodeIndex.
	 self maybeDumpLiterals: descriptor.
	 bytecodePC := self nextBytecodePCFor: descriptor exts: nExts.
	 result = 0 and: [bytecodePC <= end]]
		whileTrue:
			[nExts := descriptor isExtension ifTrue: [nExts + 1] ifFalse: [0]].
	self checkEnoughOpcodes.
	^result! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 10/13/2009 13:54'!
compileBlockDispatchFrom: lowBlockStartIndex to: highBlockStartIndex
	<var: #blockStart type: #'BlockStart *'>
	<var: #jmp type: #'AbstractInstruction *'>
	| blockStart halfWay jmp |
	lowBlockStartIndex = highBlockStartIndex ifTrue:
		[blockStart := self blockStartAt: lowBlockStartIndex.
		 self Jump: blockStart entryLabel.
		^nil].
	halfWay := highBlockStartIndex + lowBlockStartIndex // 2.
	self assert: (halfWay between: lowBlockStartIndex and: highBlockStartIndex).
	blockStart := self blockStartAt: halfWay.
	"N.B. FLAGS := TempReg - startpc"
	self CmpCq: (objectMemory integerObjectOf: blockStart startpc + 1) R: TempReg.
	lowBlockStartIndex = halfWay ifTrue:
		[self JumpLessOrEqual: blockStart entryLabel.
		 self compileBlockDispatchFrom: halfWay + 1 to: highBlockStartIndex.
		 ^nil].
	halfWay + 1 = highBlockStartIndex ifTrue:
		[blockStart := self blockStartAt: highBlockStartIndex.
		 self JumpGreater: blockStart entryLabel.
		 ^self compileBlockDispatchFrom: lowBlockStartIndex to: halfWay].
	jmp := self JumpGreater: 0.
	self compileBlockDispatchFrom: lowBlockStartIndex to: halfWay.
	halfWay = highBlockStartIndex
		ifTrue:
			[blockStart := self blockStartAt: highBlockStartIndex.
			 jmp jmpTarget: blockStart entryLabel]
		ifFalse:
			[jmp jmpTarget: self Label.
			 self compileBlockDispatchFrom: halfWay + 1 to: highBlockStartIndex]! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/3/2015 17:22:36'!
compileBlockEntry: blockStart
	"Compile a block's entry.  This looks like a dummy CogBlockMethod header (for frame parsing)
	 followed by either a frame build, if a frame is required, or nothing.  The CogMethodHeader's
	 objectHeader field is a back pointer to the method, but this can't be filled in until code generation."
	<var: #blockStart type: #'BlockStart *'>
	self AlignmentNops: self blockAlignment.
	blockStart fakeHeader: self Label.
	(self sizeof: CogBlockMethod) caseOf:
		{ [8]					"ObjectMemory"
			->	[self Fill32: 0.		"gets filled in later with the homeOffset and startpc"
				 self Fill32: 0].		"gets filled in later with numArgs et al"
		   [12]					"Spur 32-bit"
			->	[self Fill32: 0.		"gets filled in later with the homeOffset and startpc"
				 self Fill32: 0.		"is left fallow"
				 self Fill32: 0].		"gets filled in later with numArgs et al"
		   [16]					"Spur 64-bit"
			->	[self Fill32: 0.		"gets filled in later with the homeOffset and startpc"
				 self Fill32: 0.		"is left fallow"
				 self Fill32: 0.		"gets filled in later with numArgs et al"
				 self Fill32: 0].
		}.
	blockStart entryLabel: self Label.
	needsFrame
		ifTrue:
			[self compileBlockFrameBuild: blockStart.
			 self recordBlockTrace ifTrue:
				[self CallRT: ceTraceBlockActivationTrampoline]]
		ifFalse:
			[self compileBlockFramelessEntry: blockStart]! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 4/25/2009 11:05'!
compileBlockFramelessEntry: blockStart
	self subclassResponsibility! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 4/4/2015 13:59'!
compileCPICEntry
	<returnTypeC: #'AbstractInstruction *'>
	"Compile the cache tag computation and the first comparison.  Answer the address of that comparison."
	entry := objectRepresentation genGetInlineCacheClassTagFrom: ReceiverResultReg into: TempReg forEntry: true.
	self CmpR: ClassReg R: TempReg.
	^self JumpNonZero: 0! !

!Cogit methodsFor: 'initialization' stamp: 'eem 12/8/2019 23:18'!
compileCallFor: aRoutine numArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3 floatResultReg: resultRegOrNone regsToSave: regMask
	"Generate a call to aRoutine with up to 4 arguments.  If resultRegOrNone is not
	 NoReg assign the C result to resultRegOrNone.  If saveRegs, save all registers.
	 Hack: a negative arg value indicates an abstract register, a non-negative value
	 indicates a constant."
	<var: #aRoutine type: #'void *'>
	<inline: false>
	| regsToSave |
	regsToSave := resultRegOrNone = NoReg
						ifTrue: [regMask]
						ifFalse: [regMask bitClear: (self registerMaskFor: resultRegOrNone)].
	cStackAlignment > objectMemory wordSize ifTrue:
		[backEnd
			genAlignCStackSavingRegisters: regsToSave
			numArgs: numArgs
			wordAlignment: cStackAlignment / objectMemory wordSize].
	backEnd
		genSaveRegs: regsToSave;
		genMarshallNArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3.
	self CallFullRT: (self cCode: [aRoutine asUnsignedInteger]
						inSmalltalk: [self simulatedTrampolineFor: aRoutine]).
	resultRegOrNone ~= NoReg ifTrue:
		[backEnd cFloatResultToRd: resultRegOrNone].
	backEnd
		genRemoveNArgsFromStack: numArgs;
		genRestoreRegs: regsToSave! !

!Cogit methodsFor: 'initialization' stamp: 'eem 1/8/2020 13:18'!
compileCallFor: aRoutine numArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3 resultReg: resultRegOrNone regsToSave: regMask
	"Generate a call to aRoutine with up to 4 arguments.  If resultRegOrNone is not
	 NoReg assign the C result to resultRegOrNone.  If saveRegs, save all registers.
	 Hack: a negative arg value indicates an abstract register, a non-negative value
	 indicates a constant."
	<var: #aRoutine type: #'void *'>
	<inline: false>
	| regsToSave |
	regsToSave := resultRegOrNone = NoReg
						ifTrue: [regMask]
						ifFalse: [regMask bitClear: (self registerMaskFor: resultRegOrNone)].
	cStackAlignment > objectMemory wordSize ifTrue:
		[backEnd
			genAlignCStackSavingRegisters: regsToSave
			numArgs: numArgs
			wordAlignment: cStackAlignment / objectMemory wordSize].
	backEnd
		genSaveRegs: regsToSave;
		genMarshallNArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3.
	self CallFullRT: (self cCode: [aRoutine asUnsignedInteger]
						inSmalltalk: [self simulatedTrampolineFor: aRoutine]).
	backEnd
		genWriteCResultIntoReg: resultRegOrNone;
		genRemoveNArgsFromStack: numArgs;
		genRestoreRegs: regsToSave! !

!Cogit methodsFor: 'initialization' stamp: 'eem 1/8/2020 13:17'!
compileCallFor: aRoutine numArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3 resultReg: resultRegOrNone resultReg: resultReg2OrNone regsToSave: regMask
	"Generate a call to aRoutine with up to 4 arguments.  If resultRegOrNone is not
	 NoReg assign the C result to resultRegOrNone.  If saveRegs, save all registers.
	 Hack: a negative arg value indicates an abstract register, a non-negative value
	 indicates a constant."
	<var: #aRoutine type: #'void *'>
	<inline: false>
	| regsToSave |
	regsToSave := resultRegOrNone = NoReg
						ifTrue: [regMask]
						ifFalse: [regMask bitClear: (self registerMaskFor: resultRegOrNone)].
	cStackAlignment > objectMemory wordSize ifTrue:
		[backEnd
			genAlignCStackSavingRegisters: regsToSave
			numArgs: numArgs
			wordAlignment: cStackAlignment / objectMemory wordSize].
	backEnd
		genSaveRegs: regsToSave;
		genMarshallNArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3.
	self CallFullRT: (self cCode: [aRoutine asUnsignedInteger]
						inSmalltalk: [self simulatedTrampolineFor: aRoutine]).
	backEnd
		genWriteCResultIntoReg: resultRegOrNone;
		genWriteCSecondResultIntoReg: resultReg2OrNone;
		genRemoveNArgsFromStack: numArgs;
		genRestoreRegs: regsToSave! !

!Cogit methodsFor: 'initialization' stamp: 'eem 1/8/2020 13:17'!
compileCallFor: aRoutine numArgs: numArgs floatArg: regOrConst0 floatArg: regOrConst1 floatArg: regOrConst2 floatArg: regOrConst3 resultReg: resultRegOrNone regsToSave: regMask
	"Generate a call to aRoutine with up to 4 arguments.  If resultRegOrNone is not
	 NoReg assign the C result to resultRegOrNone.  If saveRegs, save all registers.
	 Hack: a negative arg value indicates an abstract register, a non-negative value
	 indicates a constant."
	<option: #LowcodeVM>
	<var: #aRoutine type: #'void *'>
	<inline: false>
	| regsToSave |
	regsToSave := resultRegOrNone = NoReg
						ifTrue: [regMask]
						ifFalse: [regMask bitClear: (self registerMaskFor: resultRegOrNone)].
	cStackAlignment > objectMemory wordSize ifTrue:
		[backEnd
			genAlignCStackSavingRegisters: regsToSave
			numArgs: numArgs
			wordAlignment: cStackAlignment / objectMemory wordSize].
	backEnd
		genSaveRegs: regsToSave;
		genMarshallNArgs: numArgs floatArg: regOrConst0 floatArg: regOrConst1 floatArg: regOrConst2 floatArg: regOrConst3.
	self CallFullRT: (self cCode: [aRoutine asUnsignedInteger]
						inSmalltalk: [self simulatedTrampolineFor: aRoutine]).
	backEnd
		genWriteCResultIntoReg: resultRegOrNone;
		genRemoveNFloatArgsFromStack: numArgs;
		genRestoreRegs: regsToSave! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 12/7/2016 14:40'!
compileClosedPICPrototype
	"Compile the abstract instructions for a full closed PIC, used to generate the chunk of code
	 which is copied to form each closed PIC.  A Closed Polymorphic Inline Cache is a small jump
	 table used to optimize sends with a limited degree of polymorphism (currently up to 6 cases).
	 We call it closed because it deals only with a finite number of cases, as opposed to an Open PIC.
	 When a monomorphic linked send (a send with a single case, linking direct to the checked entry
	 point of a CogMethod) fails a class check, the Cogit attempts to create a two-entry PIC that will
	 handle jumping to the original target for the original class and the relevant target for the new
	 class.  This jump table will be extended on subsequent failures up to a limit (6).

	 We avoid extending CPICs to Open PICs by linking the send site to an Open PIC if one already
	 exists with the send's selector, a good policy since measurements show that sends of mega-
	 morphic selectors usually become megamorphic at all send sites.  Hence the Open PIC list.

	 A CPIC also optimizes MNUs and interpret-only methods.  Each case can load SendNumArgs with
	 the oop of a method, or will load SendNumArgs with 0 if not.  MNUs are optimized by jumping to
	 the mnuAbort in the CPIC, which calls code that creates the Message, thereby avoiding looking up
	 the original message which will not be found, and either looks up doesNotUnderstand: or directly
	 activates the method loaded into SendNumArgs, hence avoiding looking up doesNotUnderstand:.
	 Interpret-only methods are handled by jumping to the picInterpretAbort, which enters the
	 interpreter activating the method loaded in SendNumArgs.

	 CPICs look like the following, where rClass is set at the original send site for the 1st case, and #Foo
	 is some constant, either an oop, a class tag or an instruction address.

		rTemp := (rRecever bitAnd: TagMask) = 0 ifTrue: [rReceiver class] ifFalse: [rRecever bitAnd: TagMask].
		rTemp = rClass ifFalse:
			[self goto: #Label].
		rSendNumArgs := #MethodForCase1Or0.
		self goto: #TargetForCase1.
	 #Label
		rTemp = #ClassTagForCase6 ifTrue:
			[rSendNumArgs := #MethodForCase6Or0.
			 self goto: #TargetForCase6].
		...cases 5, 4 & 3
		rTemp = #ClassTagForCase2 ifTrue:
			[rSendNumArgs := #MethodForCase2Or0.
			 self goto: #TargetForCase2].
		self goto: #CPICMissTrampoline
		literals (if out-of-line literals)

	 where we short-cut as many cases as needed by making the self goto: #Label skip as many cases
	 as needed."
	<inline: true>
	| numArgs jumpNext |
	<var: #jumpNext type: #'AbstractInstruction *'>
	self compilePICAbort: (numArgs := 0). "Will get rewritten to appropriate arity when configuring."
	jumpNext := self compileCPICEntry.
	"At the end of the entry code we need to jump to the first case code, which is actually the last chunk.
	 On each entension we must update this jump to move back one case."
	self MoveUniqueCw: self firstPrototypeMethodOop R: SendNumArgsReg.
	self JumpLong: self cPICPrototypeCaseOffset + 16rCA5E10.
	endCPICCase0 := self Label.
	1 to: MaxCPICCases - 1 do:
		[:h|
		h = (MaxCPICCases - 1) ifTrue:
			[jumpNext jmpTarget: self Label]. "this is where we jump to for the first case"
		self MoveUniqueCw: self subsequentPrototypeMethodOop + h R: SendNumArgsReg.
		"16rBABE1F15+h is the class tag for the Nth case"
		self CmpC32: 16rBABE1F15+h R: TempReg.
		self JumpLongZero: self cPICPrototypeCaseOffset + 16rCA5E10 + (h * 16).
		h = 1 ifTrue:
			[endCPICCase1 := self Label]].
	self MoveCw: methodLabel address R: ClassReg.
	self JumpLong: (self cPICMissTrampolineFor: numArgs).	"Will get rewritten to appropriate arity when configuring."
	cPICEndOfCodeLabel := self Label.
	literalsManager dumpLiterals: false.
	^0! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 7/25/2021 11:20'!
compileCogFullBlockMethod: numCopied
	<returnTypeC: #'CogMethod *'>
	<option: #SistaV1BytecodeSet>
	| numBytecodes numBlocks numCleanBlocks result |
	self setHasMovableLiteral: false.
	self setHasYoungReferent: (objectMemory isYoungObject: methodObj).
	methodOrBlockNumArgs := coInterpreter argumentCountOf: methodObj.
	inBlock := InFullBlock.
	maxLitIndex := -1.
	self assert: (coInterpreter primitiveIndexOf: methodObj) = 0.
	initialPC := coInterpreter startPCOfMethod: methodObj.
	"initial estimate.  Actual endPC is determined in scanMethod."
	endPC := objectMemory numBytesOf: methodObj.
	numBytecodes := endPC - initialPC + 1.
	primitiveIndex := 0.
	self allocateOpcodes: (numBytecodes + 10) * self estimateOfAbstractOpcodesPerBytecodes
		bytecodes: numBytecodes
		ifFail: [^coInterpreter cCoerceSimple: MethodTooBig to: #'CogMethod *'].
	self flag: #TODO. "currently copiedValue access implies frameful method, this is suboptimal"
	(numBlocks := self scanMethod) < 0 ifTrue:
		[^coInterpreter cCoerceSimple: numBlocks to: #'CogMethod *'].
	self assert: numBlocks = 0. "blocks in full blocks are full blocks, they are not inlined."
	numCleanBlocks := self scanForCleanBlocks.
	self assert: numCleanBlocks = 0. "blocks in full blocks are full blocks, they are not inlined."
	self allocateBlockStarts: numBlocks + numCleanBlocks.
	blockCount := 0.
	numCleanBlocks > 0 ifTrue:
		[self addCleanBlockStarts].
	(self maybeAllocAndInitCounters
	 and: [self maybeAllocAndInitIRCs]) ifFalse: "Inaccurate error code, but it'll do.  This will likely never fail."
		[^coInterpreter cCoerceSimple: InsufficientCodeSpace to: #'CogMethod *'].

	blockEntryLabel := nil.
	methodLabel dependent: nil.
	(result := self compileEntireFullBlockMethod: numCopied) < 0 ifTrue:
		[^coInterpreter cCoerceSimple: result to: #'CogMethod *'].
	^self generateCogFullBlock! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 7/25/2021 11:20'!
compileCogMethod: selector
	<returnTypeC: #'CogMethod *'>
	| numBytecodes numBlocks numCleanBlocks result extra |
	self setHasMovableLiteral: false.
	self setHasYoungReferent: ((objectMemory isYoungObject: methodObj)
								  or: [objectMemory isYoung: selector]).
	methodOrBlockNumArgs := coInterpreter argumentCountOf: methodObj.
	inBlock := 0.
	maxLitIndex := -1.
	extra := ((primitiveIndex := coInterpreter primitiveIndexOf: methodObj) > 0
			and: [(coInterpreter isQuickPrimitiveIndex: primitiveIndex) not])
				ifTrue: [30]
				ifFalse: [10].
	initialPC := coInterpreter startPCOfMethod: methodObj.
	"initial estimate.  Actual endPC is determined in scanMethod."
	endPC := (coInterpreter isQuickPrimitiveIndex: primitiveIndex)
					ifTrue: [initialPC - 1]
					ifFalse: [objectMemory numBytesOf: methodObj].
	numBytecodes := endPC - initialPC + 1.
	self allocateOpcodes: (numBytecodes + extra) * self estimateOfAbstractOpcodesPerBytecodes
		bytecodes: numBytecodes
		ifFail: [^coInterpreter cCoerceSimple: MethodTooBig to: #'CogMethod *'].
	(numBlocks := self scanMethod) < 0 ifTrue:
		[^coInterpreter cCoerceSimple: numBlocks to: #'CogMethod *'].
	numCleanBlocks := self scanForCleanBlocks.
	self methodFoundInvalidPostScan ifTrue:
		[^coInterpreter cCoerceSimple: ShouldNotJIT to: #'CogMethod *'].
	self allocateBlockStarts: numBlocks + numCleanBlocks.
	blockCount := 0.
	numCleanBlocks > 0 ifTrue:
		[self addCleanBlockStarts].
	(self maybeAllocAndInitCounters
	 and: [self maybeAllocAndInitIRCs]) ifFalse: "Inaccurate error code, but it'll do.  This will likely never fail."
		[^coInterpreter cCoerceSimple: InsufficientCodeSpace to: #'CogMethod *'].
	
	blockEntryLabel := nil.
	methodLabel dependent: nil.
	(result := self compileEntireMethod) < 0 ifTrue:
		[^coInterpreter cCoerceSimple: result to: #'CogMethod *'].
	^self generateCogMethod: selector! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'cb 9/2/2016 14:29'!
compileEntireFullBlockMethod: numCopied
	"Compile the abstract instructions for the entire full block method."
	<option: #SistaV1BytecodeSet>
	| result |
	self preenMethodLabel.
	self compileFullBlockEntry.

	"Frame build"
	self compileFullBlockMethodFrameBuild: numCopied.
	"Method body"
	(result := self compileMethodBody) < 0 ifTrue:
		[^result].
	self assert: blockCount = 0.
	^0! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 4/14/2016 09:45'!
compileEntireMethod
	"Compile the abstract instructions for the entire method, including blocks."
	| result |
	self preenMethodLabel.
	self compileAbort.
	self compileEntry.
	(result := self compilePrimitive) < 0 ifTrue:
		[^result].
	self compileFrameBuild.
	(result := self compileMethodBody) < 0 ifTrue:
		[^result].
	blockCount = 0 ifTrue:
		[^0].
	(result := self compileBlockBodies) < 0 ifTrue:
		[^result].
	^self compileBlockDispatch! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 7/8/2021 13:54'!
compileEntry
	"The entry code to a method checks that the class of the current receiver matches
	 that in the inline cache.  Other non-obvious elements are that its alignment must be
	 different from the alignment of the noCheckEntry so that the method map machinery
	 can distinguish normal and super sends (super sends bind to the noCheckEntry)."

	entry := objectRepresentation genGetInlineCacheClassTagFrom: ReceiverResultReg into: TempReg forEntry: true.
	self CmpR: ClassReg R: TempReg.
	self JumpNonZero: sendMiss.
	noCheckEntry := self Label.
	self compileSendTrace ifTrue:
		[backEnd saveAndRestoreLinkRegAround:
			[self CallFullRT: ceTraceLinkedSendTrampoline]]! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 4/25/2009 11:05'!
compileFrameBuild
	self subclassResponsibility! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'cb 4/15/2016 11:10'!
compileFullBlockEntry
	<option: #SistaV1BytecodeSet>
	"Compile the abstract instructions for the entire method, including blocks."
	| jumpNoContextSwitch |

	"Abort for stack overflow on full block activation (no inline cache miss possible).
	 The flag is SendNumArgsReg."
	stackOverflowCall := self MoveCq: 0 R: ReceiverResultReg.
	backEnd hasLinkRegister ifTrue: [self PushR: LinkReg].
	"Since the only case in which this is called is the
	 stack overflow case we can reuse the trampoline."
	self Call: (self methodAbortTrampolineFor: methodOrBlockNumArgs).

	"Entries"
	"No context switch entry"
	fullBlockNoContextSwitchEntry := self MoveCq: 0 R: SendNumArgsReg.
	jumpNoContextSwitch := self Jump: 0.

	self AlignmentNops: (objectMemory wordSize max: 8).
	"Context switch entry (use ReceiverResultReg as a non-zero value; it's shorter)."
	fullBlockEntry := self MoveR: ReceiverResultReg R: SendNumArgsReg.
	jumpNoContextSwitch jmpTarget: self Label.

	^0! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 1/21/2016 18:29'!
compileMethodBody
	"Compile the top-level method body."
	<inline: true>
	endPC < initialPC ifTrue: [^0]. "quick primitives"
	"When compiling, skip any initial CallPrimitive and optional StorePrimErrCode bytecodes.
	 These are dealt with in compileFrameBuild."
	^self compileAbstractInstructionsFrom: initialPC
										+ (self deltaToSkipPrimAndErrorStoreIn: methodObj
												header: methodHeader)
		through: endPC! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 12/4/2010 21:18'!
compileOpenPIC: selector numArgs: numArgs
	"Compile the code for an open PIC.  Perform a probe of the first-level method
	 lookup cache followed by a call of ceSendFromOpenPIC: if the probe fails.
	 Since open PICs replicate the CoInterpreter's first-level method cache lookup
	 this is a subclass responsibility."
	self subclassResponsibility! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 12/6/2014 11:03'!
compilePICAbort: numArgs
	"The start of a PIC has a call to a run-time abort routine that either handles a dispatch to an
	 interpreted method or a dispatch of an MNU case.  The routine selects the path by testing
	 ClassReg, which holds the inline cache tag; if equal to the picAbortDiscriminatorValue (zero)
	 it takes the MNU path; if nonzero the dispatch to interpreter path.  Neither of these paths
	 returns. The abort routine must be called;  In the callee the PIC is located by adding the
	 relevant offset to the return address of the call.

	 N.B. This code must match that in compileAbort so that the offset of the return address of
	 the call is the same in methods and closed PICs."
	picMNUAbort := self MoveCq: self picAbortDiscriminatorValue R: ClassReg.
	backEnd hasLinkRegister
		ifTrue:
			["If there is a link register it must be saved (pushed onto the stack) before it
			  is smashed by the abort call, and hence needs to be manually handled here".
			 picInterpretAbort := self PushR: LinkReg.
			 self Call: (self picAbortTrampolineFor: numArgs)]
		ifFalse:
			[picInterpretAbort := self Call: (self picAbortTrampolineFor: numArgs)].
	^0! !

!Cogit methodsFor: 'debugging' stamp: 'eem 7/8/2021 17:30'!
compileSendTrace
	"2 is trace sends; 256 is traceLinkedSends"
	<inline: #always>
	^traceFlags allMask: 256! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 12/19/2019 13:27'!
compileStackOverflowCheck: canContextSwitch
	"Compile the compare of stackLimit against the stack pointer, jumping to the stackOverflowCall if
	 the stack pointer is below the limit.  Answer a bytecode annotated label that follows the sequence.

	 The stack check functions both as a genuine stack limit check to prevent calls overflowing stack pages,
	 and as an event/context-switch break out.  To cause an event check (including a check for a required
	 context switch), stackLimit is set to the highest possible value, and hence all stack limit checks will
	 fail.  A path in the stack overflow abort then arranges to call event checking if it has been requested.

	 Certain block activations (e.g. valueNoContextSwitch:) must not context switch, and in that
	 case, SendNumArgs is set to zero to communicate to the stack overflow abort that it should
	 not perform event/context-switch (yet)."

	| jumpSkip label |
	<inline: false>
	self MoveAw: coInterpreter stackLimitAddress R: TempReg.
	"N.B. We do not expect the backEnd to provide CmpR: SPReg R: reg, only CmpR: reg R: SPReg"
	self CmpR: TempReg R: SPReg. "N.B. FLAGS := SPReg - TempReg"
	"If we can't context switch for this method, use a slightly
	 slower overflow check that clears SendNumArgsReg."
	canContextSwitch
		ifTrue:
			[self JumpBelow: stackOverflowCall.
			 label := self Label]
		ifFalse:
			[jumpSkip := self JumpAboveOrEqual: 0.
			 self MoveCq: 0 R: SendNumArgsReg.
			 self Jump: stackOverflowCall.
			 jumpSkip jmpTarget: (label := self Label)].
	self annotateBytecode: label.
	^label! !

!Cogit methodsFor: 'initialization' stamp: 'eem 2/1/2020 18:49'!
compileTrampolineFor: aRoutine numArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3 regsToSave: regMask pushLinkReg: pushLinkReg floatResultReg: resultRegOrNone
	"Generate a trampoline with up to four arguments.  Generate either a call or a jump to aRoutine
	 as requested by callJumpBar.  If generating a call and resultRegOrNone is not NoReg pass the C
	 result back in resultRegOrNone.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<inline: false>
	self genSmalltalkToCStackSwitch: pushLinkReg.
	self
		compileCallFor: aRoutine
		numArgs: numArgs
		arg: regOrConst0
		arg: regOrConst1
		arg: regOrConst2
		arg: regOrConst3
		floatResultReg: resultRegOrNone
		regsToSave: regMask.
	backEnd genLoadStackPointers.
	self genTrampolineReturn: pushLinkReg! !

!Cogit methodsFor: 'initialization' stamp: 'eem 2/1/2020 18:49'!
compileTrampolineFor: aRoutine numArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3 regsToSave: regMask pushLinkReg: pushLinkReg resultReg: resultRegOrNone
	"Generate a trampoline with up to four arguments.  Generate either a call or a jump to aRoutine
	 as requested by callJumpBar.  If generating a call and resultRegOrNone is not NoReg pass the C
	 result back in resultRegOrNone.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<inline: false>
	self genSmalltalkToCStackSwitch: pushLinkReg.
	self
		compileCallFor: aRoutine
		numArgs: numArgs
		arg: regOrConst0
		arg: regOrConst1
		arg: regOrConst2
		arg: regOrConst3
		resultReg: resultRegOrNone
		regsToSave: regMask.
	backEnd genLoadStackPointers.
	self genTrampolineReturn: pushLinkReg! !

!Cogit methodsFor: 'initialization' stamp: 'eem 2/1/2020 18:50'!
compileTrampolineFor: aRoutine numArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3 regsToSave: regMask pushLinkReg: pushLinkReg resultReg: resultRegOrNone resultReg: resultReg2OrNone
	"Generate a trampoline with up to four arguments.  Generate either a call or a jump to aRoutine
	 as requested by callJumpBar.  If generating a call and resultRegOrNone is not NoReg pass the C
	 result back in resultRegOrNone.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<inline: false>
	self genSmalltalkToCStackSwitch: pushLinkReg.
	self
		compileCallFor: aRoutine
		numArgs: numArgs
		arg: regOrConst0
		arg: regOrConst1
		arg: regOrConst2
		arg: regOrConst3
		resultReg: resultRegOrNone
		resultReg: resultReg2OrNone
		regsToSave: regMask.
	backEnd genLoadStackPointers.
	self genTrampolineReturn: pushLinkReg! !

!Cogit methodsFor: 'initialization' stamp: 'eem 2/1/2020 18:50'!
compileTrampolineFor: aRoutine numArgs: numArgs floatArg: regOrConst0 floatArg: regOrConst1 floatArg: regOrConst2 floatArg: regOrConst3 regsToSave: regMask pushLinkReg: pushLinkReg resultReg: resultRegOrNone
	"Generate a trampoline with up to four arguments.  Generate either a call or a jump to aRoutine
	 as requested by callJumpBar.  If generating a call and resultRegOrNone is not NoReg pass the C
	 result back in resultRegOrNone.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<option: #LowcodeVM>
	<var: #aRoutine type: #'void *'>
	<inline: false>
	self genSmalltalkToCStackSwitch: pushLinkReg.
	self
		compileCallFor: aRoutine
		numArgs: numArgs
		floatArg: regOrConst0
		floatArg: regOrConst1
		floatArg: regOrConst2
		floatArg: regOrConst3
		resultReg: resultRegOrNone
		regsToSave: regMask.
	backEnd genLoadStackPointers.
	self genTrampolineReturn: pushLinkReg! !

!Cogit methodsFor: 'initialization' stamp: 'eem 4/4/2015 14:00'!
computeEntryOffsets
	"Generate the entry code for a method to determine cmEntryOffset and cmNoCheckEntryOffset.  We
	 need cmNoCheckEntryOffset up front to be able to generate the map starting from cmNoCheckEntryOffset"
	"stack allocate the various collections so that they
	 are effectively garbage collected on return."
	| sendMissCall |
	<var: 'sendMissCall' type: #'AbstractInstruction *'>
	self allocateOpcodes: 24 bytecodes: 0.
	methodOrBlockNumArgs := 0.
	sendMissCall := self compileAbort.
	self compileEntry.
	self computeMaximumSizes.
	self generateInstructionsAt: methodZoneBase + (self sizeof: CogMethod).
	cmEntryOffset := entry address - methodZoneBase.
	cmNoCheckEntryOffset := noCheckEntry address - methodZoneBase.
	missOffset := sendMissCall address + sendMissCall machineCodeSize - methodZoneBase.
	entryPointMask := objectMemory wordSize - 1.
	[(cmEntryOffset bitAnd: entryPointMask) = (cmNoCheckEntryOffset bitAnd: entryPointMask)] whileTrue:
		[entryPointMask := entryPointMask + entryPointMask + 1].
	entryPointMask >= (methodZone roundUpLength: 1) ifTrue:
		[self error: 'cannot differentiate checked and unchecked entry-points with current cog method alignment'].
	checkedEntryAlignment := cmEntryOffset bitAnd: entryPointMask.
	uncheckedEntryAlignment := cmNoCheckEntryOffset bitAnd: entryPointMask.
	self assert: checkedEntryAlignment ~= uncheckedEntryAlignment! !

!Cogit methodsFor: 'initialization' stamp: 'cb 4/15/2016 18:00'!
computeFullBlockEntryOffsets
	"Generate the entry code for a method to determine cmEntryOffset and cmNoCheckEntryOffset.  We
	 need cmNoCheckEntryOffset up front to be able to generate the map starting from cmNoCheckEntryOffset"
	"stack allocate the various collections so that they
	 are effectively garbage collected on return."
	SistaV1BytecodeSet ifTrue:
		[self allocateOpcodes: 24 bytecodes: 0.
		 methodOrBlockNumArgs := 0.
		 self compileFullBlockEntry.
		 self computeMaximumSizes.
		 self generateInstructionsAt: methodZoneBase + (self sizeof: CogMethod).
		 cbEntryOffset := fullBlockEntry address - methodZoneBase.
		 cbNoSwitchEntryOffset := fullBlockNoContextSwitchEntry address - methodZoneBase]! !

!Cogit methodsFor: 'initialization' stamp: 'eem 10/29/2020 14:03'!
computeGoodVarBaseAddress
	"While we order variables in the CoInterpreter in order of dynamic frequency, and hence
	 expect that stackPointer will be output first, C optimizers and linkers may get their own
	 ideas and ``improve upon'' this ordering.  So we cannot depend on stackPointer being
	 at the lowest address of the variables we want to access through VarBaseReg.  Here we
	 choose the minimum amongst a set to try to choose a varBaseAddress that is just less
	 than but iwht in range of all variables we want to access through it."
	| minAddress |
	minAddress := coInterpreter stackLimitAddress. "stackLimit is e.g. lowest using the clang toolchain on MacOS X"
	coInterpreter stackPointerAddress < minAddress ifTrue:
		[minAddress := coInterpreter stackPointerAddress].
	coInterpreter framePointerAddress < minAddress ifTrue:
		[minAddress := coInterpreter framePointerAddress].
	coInterpreter instructionPointerAddress < minAddress ifTrue:
		[minAddress := coInterpreter instructionPointerAddress].
	coInterpreter argumentCountAddress < minAddress ifTrue:
		[minAddress := coInterpreter argumentCountAddress].
	coInterpreter primFailCodeAddress < minAddress ifTrue:
		[minAddress := coInterpreter primFailCodeAddress].
	^minAddress! !

!Cogit methodsFor: 'generate machine code' stamp: 'eem 7/21/2021 09:03'!
computeMaximumSizes
	"This pass assigns maximum sizes to all abstract instructions and eliminates jump fixups.
	 It hence assigns the maximum address an instruction will occur at which allows the next
	 pass to conservatively size jumps."
	<inline: false>
	<var: #abstractInstruction type: #'AbstractInstruction *'>
	| relativeAddress |
	literalsManager dumpLiterals: false.
	relativeAddress := 0.
	0 to: opcodeIndex - 1 do:
		[:i| | abstractInstruction |
		self maybeBreakGeneratingInstructionWithIndex: i.
		abstractInstruction := self abstractInstructionAt: i.
		abstractInstruction
			address: relativeAddress;
			maxSize: abstractInstruction computeMaximumSize.
		relativeAddress := relativeAddress + abstractInstruction maxSize]! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 12/13/2015 13:41'!
configureCPIC: cPIC Case0: case0CogMethod Case1Method: case1Method tag: case1Tag isMNUCase: isMNUCase numArgs: numArgs delta: addrDelta
	"Configure a copy of the prototype CPIC for a two-case PIC for 
	case0CogMethod and
	case1Method
	case1Tag.
	 The tag for case0CogMethod is at the send site and so doesn't need to be generated.
	 case1Method may be any of
		- a Cog method; jump to its unchecked entry-point
		- a CompiledMethod; jump to the ceInterpretFromPIC trampoline
		- nil; call ceMNUFromPIC
	addDelta is the address change from the prototype to the new CPIC location, needed
	because the loading of the CPIC label at the end may use a literal instead of a pc relative load."
	"self disassembleFrom: cPIC asInteger + (self sizeof: CogMethod) to: cPIC asInteger + closedPICSize"
	<var: #cPIC type: #'CogMethod *'>
	<var: #case0CogMethod type: #'CogMethod *'>
	| operand targetEntry caseEndAddress |
	self assert: case1Method notNil.

	"adjust the call at missOffset, the ceAbortXArgs"
	backEnd rewriteCallAt: cPIC asInteger + missOffset target: (self picAbortTrampolineFor: numArgs).
	
	self assert: (objectRepresentation inlineCacheTagIsYoung: case1Tag) not.
	(isMNUCase not
	 and: [coInterpreter methodHasCogMethod: case1Method])
		ifTrue:
			[operand := 0.
			 targetEntry := (coInterpreter cogMethodOf: case1Method) asInteger + cmNoCheckEntryOffset]
		ifFalse: "We do not scavenge PICs, hence we cannot cache the MNU method if it is in new space."
			[operand := (case1Method isNil or: [objectMemory isYoungObject: case1Method])
							ifTrue: [0]
							ifFalse: [case1Method].
			 targetEntry := case1Method ifNil: [cPIC asInteger + (self sizeof: CogMethod)] ifNotNil: [cPIC asInteger + self picInterpretAbortOffset]].

	"set the jump to the case0 method"
	backEnd rewriteJumpLongAt: cPIC asInteger + firstCPICCaseOffset target: case0CogMethod asInteger + cmNoCheckEntryOffset.

	caseEndAddress := self addressOfEndOfCase: 2 inCPIC: cPIC.

	"update the cpic case"
	self
		rewriteCPICCaseAt: caseEndAddress
		tag: case1Tag
		objRef: operand
		target: (isMNUCase ifTrue: [cPIC asInteger + (self sizeof: CogMethod)] ifFalse: [targetEntry]) asInteger.

	"update the loading of the CPIC address"
	backEnd relocateMethodReferenceBeforeAddress: cPIC asInteger + cPICEndOfCodeOffset - backEnd jumpLongByteSize by: addrDelta.

	"write the final desperate jump to cePICMissXArgs"
	backEnd rewriteJumpLongAt: cPIC asInteger + cPICEndOfCodeOffset target: (self cPICMissTrampolineFor: numArgs).
	^0
	"self disassembleFrom: cPIC + (self sizeof: CogMethod) to: cPIC + closedPICSize - 1."! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'rmacnak 11/23/2015 23:32:56'!
configureMNUCPIC: cPIC methodOperand: methodOperand numArgs: numArgs delta: addrDelta
	"Configure a copy of the prototype CPIC for a one-case MNU CPIC that calls ceMNUFromPIC for
	 case0Tag The tag for case0 is at the send site and so doesn't need to be generated.
	 addDelta is the address change from the prototype to the new CPIC location, needed
	 because the loading of the CPIC label at the end may be a literal instead of a pc-relative load."
	<var: #cPIC type: #'CogMethod *'>
	| operand |

	"adjust the jump at missOffset, the ceAbortXArgs"
	backEnd rewriteCallAt: cPIC asInteger + missOffset target: (self picAbortTrampolineFor: numArgs).
	
	"We do not scavenge PICs, hence we cannot cache the MNU method if it is in new space."
	operand := (methodOperand isNil or: [objectMemory isYoungObject: methodOperand])
					ifTrue: [0]
					ifFalse: [methodOperand].
	"set the jump to the case0 method"
	backEnd rewriteJumpLongAt: cPIC asInteger + firstCPICCaseOffset target: cPIC asInteger + (self sizeof: CogMethod) .

	backEnd storeLiteral: operand beforeFollowingAddress: cPIC asInteger + firstCPICCaseOffset - backEnd jumpLongByteSize.

	"rewrite the final desperate jump to cePICMissXArgs"
	backEnd rewriteJumpLongAt: cPIC asInteger + cPICEndOfCodeOffset target: (self cPICMissTrampolineFor: numArgs).	

	"update the loading of the CPIC label address"
	backEnd relocateMethodReferenceBeforeAddress: cPIC asInteger + cPICEndOfCodeOffset - backEnd jumpLongByteSize by: addrDelta.

	"finally, rewrite the jump 3 instr before firstCPICCaseOffset to jump to the end of case 2, missing the actual case"
	self rewriteCPIC: cPIC caseJumpTo: (self addressOfEndOfCase: 2 inCPIC: cPIC). 

	^0! !

!Cogit methodsFor: 'printing' stamp: 'eem 10/12/2009 01:07'!
cr
	<cmacro: '() putchar(''\n'')'>
	coInterpreter transcript cr; flush! !

!Cogit methodsFor: 'profiling primitives' stamp: 'eem 6/1/2018 13:06'!
createCPICData: cPIC
	"Answer an Array of the PIC's selector, followed by class and targetMethod/doesNotUnderstand: for each entry in the PIC."
	<var: #cPIC type: #'CogMethod *'>
	| picData |
	<var: #targetMethod type: #'CogMethod *'>
	self assert: (cPIC methodObject = 0 or: [objectMemory addressCouldBeOop: cPIC methodObject]).
	picData := objectMemory instantiateClass: objectMemory classArray indexableSize: cPIC cPICNumCases * 2 + 1.
	picData ifNil: [^picData].
      objectMemory storePointerUnchecked: 0 ofObject: picData withValue: cPIC selector.
	1 to: cPIC cPICNumCases do:
		[:i| | pc entryPoint target targetMethod class |
		pc := self addressOfEndOfCase: i inCPIC: cPIC.
		i = 1
			ifTrue:
				[class := cPIC methodObject. "first case may have been collected and stored here by collectCogConstituentFor:Annotation:Mcpc:Bcpc:Method:"
				 class = 0 ifTrue: [class := objectMemory nilObject]. "cPIC is unreferenced; likely evolved to OpenPIC"
				 entryPoint := backEnd jumpLongTargetBeforeFollowingAddress: pc]
			ifFalse:
				[class := objectRepresentation classForInlineCacheTag:
							(backEnd literal32BeforeFollowingAddress: pc - backEnd jumpLongConditionalByteSize).
				 entryPoint := backEnd jumpLongConditionalTargetBeforeFollowingAddress: pc].
		"Find target from jump.  A jump to the MNU entry-point should collect #doesNotUnderstand:"
		(cPIC containsAddress: entryPoint)
			ifTrue:
				[target := objectMemory splObj: SelectorDoesNotUnderstand]
			ifFalse:
				[targetMethod := self cCoerceSimple: entryPoint - cmNoCheckEntryOffset to: #'CogMethod *'.
				 self assert: targetMethod cmType = CMMethod.
				 target := targetMethod methodObject].
		objectMemory
			storePointerUnchecked: i * 2 - 1 ofObject: picData withValue: class;
			storePointerUnchecked: i * 2 ofObject: picData withValue: target].
	objectMemory beRootIfOld: picData.
	cPIC methodObject: 0. "restore invariant."
	^picData! !

!Cogit methodsFor: 'debugging' stamp: 'eem 7/20/2021 16:59'!
debugAPISelector
	"Answer the selector the CoInterpreter or machine code called in on. Simulation only.
	 Used to help the codeZoneIsExecutableNotWritable assert work."
	<doNotGenerate>
	^(thisContext findContextSuchThat:
		[:ctxt|
		((ctxt selector beginsWith: 'ce') and: [ctxt selector third isUppercase])
		or: [ctxt sender notNil and: [ctxt sender receiver == coInterpreter]]]) ifNotNil: [:ctxt| ctxt selector]! !

!Cogit methodsFor: 'accessing' stamp: 'eem 6/3/2015 22:39'!
defaultCogCodeSize
	"Return the default number of bytes to allocate for native code at startup.
	 The actual value can be set via vmParameterAt: and/or a preference in the ini file."
	<api>
	^backEnd getDefaultCogCodeSize! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 1/21/2016 18:32'!
deltaToSkipPrimAndErrorStoreIn: aMethodObj header: aMethodHeader
	"Answer the number of bytecodes to skip to get to the first bytecode
	 past the primitive call and any store of the error code."
	^(self methodUsesPrimitiveErrorCode: aMethodObj header: aMethodHeader)
		ifTrue: [(coInterpreter sizeOfCallPrimitiveBytecode: aMethodHeader)
			  + (coInterpreter sizeOfLongStoreTempBytecode: aMethodHeader)]
		ifFalse: [0]! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 1/9/2011 11:12'!
disassemble: targetmcpc from: startpc to: endpc arg: aStream
	<doNotGenerate>
	| startbcpc |
	self disassembleFrom: startpc to: endpc - 1 labels: Dictionary new on: aStream.
	startbcpc := (self cCoerceSimple: targetmcpc - (self sizeof: CogBlockMethod) to: #'CogBlockMethod *') startpc.
	aStream nextPutAll: 'block startpc: '.
	startbcpc printOn: aStream base: 16.
	aStream nextPut: $/.
	(objectMemory integerObjectOf: startbcpc) printOn: aStream base: 16.
	aStream cr; flush.
	^0! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 1/11/2013 14:18'!
disassembleCachedOop: address
	<doNotGenerate>
	| oopOrZero |
	"Answer a disassembly string for an implicit receiver cache entry in a Newspeak send."
	oopOrZero := objectMemory unalignedLongAt: address.
	^(objectMemory addressCouldBeObj: oopOrZero)
		ifTrue: [oopOrZero hex, ': ', (coInterpreter shortPrint: oopOrZero)]
		ifFalse: [oopOrZero = 0 ifTrue: [oopOrZero printString] ifFalse: [oopOrZero hex]]! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 1/11/2013 13:07'!
disassembleCodeAt: pc
	<doNotGenerate>
	(pc between: (trampolineAddresses at: 1) and: methodZoneBase - 1) ifTrue:
		[^self disassembleTrampolineFor: pc].
	self disassembleMethodFor: pc! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 12/4/2010 08:50'!
disassembleFrom: startAddress to: endAddress
	<doNotGenerate>
	self disassembleFrom: startAddress to: endAddress labels: Dictionary new on: coInterpreter transcript! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 1/22/2018 11:03'!
disassembleFrom: startAddress to: endAddress labels: labelDictionary on: aStream
	<doNotGenerate>
	| previousTVP |
	aStream ensureCr.
	previousTVP := processor class printTempNames.
	processor class
		printTempNames: (self class initializationOptions includesKey: #tempNames);
		setReceiverResultReg: ((self class initializationOptions includesKey: #instVarNames) ifTrue:
									[ReceiverResultReg]).
	[processor
		disassembleFrom: startAddress
		to: endAddress
		in: coInterpreter memory
		for: self
		labels: labelDictionary
		on: aStream] ensure:
			[processor class
				printTempNames: previousTVP;
				setReceiverResultReg: nil].
	aStream flush! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 10/12/2009 01:07'!
disassembleMethod: surrogateOrAddress
	<doNotGenerate>
	self disassembleMethod: surrogateOrAddress on: coInterpreter transcript! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 2/23/2020 15:12'!
disassembleMethod: surrogateOrAddress on: aStream
	<doNotGenerate>
	| cogMethod mapEntries codeRanges |
	cogMethod := surrogateOrAddress isInteger
								ifTrue: [self cogMethodSurrogateAt: surrogateOrAddress]
								ifFalse: [surrogateOrAddress].
	cogMethod cmType = CMBlock ifTrue:
		[^self disassembleMethod: cogMethod cmHomeMethod on: aStream].
	(disassemblingMethod isNil
	 and: [self class initializationOptions at: #relativeAddressDisassembly ifAbsent: [false]]) ifTrue:
		[^[disassemblingMethod := cogMethod.
		    self disassembleMethod: surrogateOrAddress on: aStream] ensure:
			[disassemblingMethod := nil]].
	self printMethodHeader: cogMethod on: aStream.

	mapEntries := Dictionary new.
	(cogMethod cmType = CMMethod and: [cogMethod cmIsFullBlock]) ifFalse:
		[mapEntries at: cogMethod asInteger + cmEntryOffset put: 'entry'].
	
	cogMethod cmType = CMMethod ifTrue:
		[cogMethod cmIsFullBlock
			ifTrue: [mapEntries at: cogMethod asInteger + cbNoSwitchEntryOffset put: 'noSwitchEntry']
			ifFalse: [mapEntries at: cogMethod asInteger + cmNoCheckEntryOffset put: 'noCheckEntry']].

	cogMethod cmType = CMClosedPIC
		ifTrue: "Since Tim R's lovely work on Closed PICs, PIC cases go backwards..."
			[mapEntries at: cogMethod asInteger + firstCPICCaseOffset put: 'ClosedPICCase', MaxCPICCases printString.
			 1 to: MaxCPICCases - 1 do:
				[:i|
				mapEntries
					at: cogMethod asInteger + firstCPICCaseOffset + (i * cPICCaseSize)
					put: 'ClosedPICCase', (MaxCPICCases - i) printString]]
		ifFalse:
			[self mapFor: cogMethod
				performUntil: #collectMapEntry:address:into:
				arg: mapEntries].

	NewspeakVM ifTrue:
		[objectRepresentation canPinObjects ifFalse:
			[mapEntries keys do:
				[:a|
				(mapEntries at: a) = #IsNSSendCall ifTrue:
					[mapEntries
						at: a + backEnd jumpShortByteSize
							put: {'Class'. #disassembleCachedOop:. (objectMemory wordSize)};
						at: a + backEnd jumpShortByteSize + objectMemory bytesPerOop
							put: {'ImplicitReceiver'. #disassembleCachedOop:. (objectMemory wordSize)}]]]].

	"This would all be far more elegant and simple if we used blocks.
	 But there are no blocks in C and the basic enumerators here need
	 to be used in the real VM.  Apologies."
	(codeRanges := self codeRangesFor: cogMethod) do:
		[:range|
		(cogMethod cmType = CMMethod) ifTrue:
			[mapEntries keysAndValuesDo:
				[:mcpc :label| | bcpc selectorOrNone |
				(((range includes: mcpc) or: [range last + 1 = mcpc])
				 and: [(AnnotationsWithBytecodePCs includes: label)
				 and: [range cogMethod stackCheckOffset > 0]]) ifTrue:
					[bcpc := self bytecodePCFor: mcpc startBcpc: range startpc in: range cogMethod.
					 bcpc ~= 0 ifTrue:
						[label = #IsSendCall
							ifTrue:
								[selectorOrNone := (self selectorForSendAt: mcpc annotation: IsSendCall in: cogMethod methodObject).
								 (selectorOrNone isInteger and: [objectMemory addressCouldBeOop: selectorOrNone]) ifTrue:
									[selectorOrNone := objectMemory stringOf: selectorOrNone].
								selectorOrNone := ' ', selectorOrNone]
							ifFalse: [selectorOrNone := ''].
						 mapEntries
							at: mcpc
							put: label, selectorOrNone, ' bc ', bcpc printString, '/', (bcpc + 1) printString]]]].
		(cogMethod blockEntryOffset ~= 0
		 and: [range first = (cogMethod blockEntryOffset + cogMethod asInteger)])
			ifTrue:
				[aStream nextPutAll: 'blockEntry:'; cr.
				 self blockDispatchFor: cogMethod
					perform: #disassemble:from:to:arg:
					arg: aStream]
			ifFalse:
				[range first > (cogMethod address + cmNoCheckEntryOffset) ifTrue:
					[self printMethodHeader: range cogMethod
						on: aStream].
				self maybeNoteStartpcFor: range.
				self disassembleFrom: range first to: range last labels: mapEntries on: aStream]].
	aStream nextPutAll: 'startpc: '; print: codeRanges first startpc; cr.
	(cogMethod cmType = CMMethod
	 or: [cogMethod cmType = CMOpenPIC]) ifTrue:
		[[self mapFor: cogMethod
			performUntil: #printMapEntry:mcpc:args:
			arg: { aStream. codeRanges. cogMethod }]
			on: AssertionFailure
			do: [:ex|
				ex primitiveChangeClassTo: ResumableVMError basicNew. ":) :) :)"
				ex resume: nil]].
	^cogMethod! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 6/10/2010 17:15'!
disassembleMethodFor: pc
	<doNotGenerate>
	| method |
	method := methodZone methodFor: pc.
	(method isNil or: [method isZero]) ifTrue:
		[self error: 'not a method'].
	self disassembleMethod: method! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 11/24/2020 23:18'!
disassembleTrampolineFor: pc
	<doNotGenerate>
	(self trampolineRangeFor: pc) ifNotNil:
		[:range| self disassembleFrom: range first to: range last]! !

!Cogit methodsFor: 'debugging' stamp: 'eem 12/4/2010 08:46'!
disassembleTrampolineTable
	<doNotGenerate>
	0 to: trampolineTableIndex - 1 by: 2 do:
		[:i|
		self disassembleTrampolineFor: (trampolineAddresses at: i + 1) asInteger.
		coInterpreter transcript cr]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 12/4/2010 19:25'!
doesNotUnderstand: aMessage
	(aMessage selector beginsWith: 'print') ifTrue:
		[(coInterpreter respondsTo: aMessage selector) ifTrue:
			[^aMessage lookupClass: nil; sentTo: coInterpreter].
		(methodZone respondsTo: aMessage selector) ifTrue:
			[^aMessage lookupClass: nil; sentTo: methodZone].
		(objectMemory respondsTo: aMessage selector) ifTrue:
			[^aMessage lookupClass: nil; sentTo: objectMemory]].
	^super doesNotUnderstand: aMessage! !

!Cogit methodsFor: 'register management' stamp: 'eem 3/3/2016 14:34'!
emptyRegisterMask
	<inline: true>
	^0! !

!Cogit methodsFor: 'analysis' stamp: 'eem 8/20/2021 23:19'!
enablePerMethodProfiling
	<doNotGenerate>
	perMethodProfile := Dictionary new.
	instructionProfile := DoubleWordArray new: methodZone zoneEnd.
	singleStep := true! !

!Cogit methodsFor: 'compiled methods' stamp: 'cb 3/1/2018 15:40'!
endPCOf: aMethod
	<var: #descriptor type: #'BytecodeDescriptor *'>
	| pc end latestContinuation descriptor prim distance targetPC byte bsOffset nExts |
	pc := latestContinuation := coInterpreter startPCOfMethod: aMethod.
	(prim := coInterpreter primitiveIndexOf: aMethod) > 0 ifTrue:
		[(coInterpreter isQuickPrimitiveIndex: prim) ifTrue:
			[^pc - 1]].
	bsOffset := self bytecodeSetOffsetFor: aMethod.
	nExts := 0.
	end := objectMemory numBytesOf: aMethod.
	[pc <= end] whileTrue:
		[byte := objectMemory fetchByte: pc ofObject: aMethod.
		descriptor := self generatorAt: byte + bsOffset.
		(descriptor isReturn
		 and: [pc >= latestContinuation]) ifTrue:
			[end := pc].
		(descriptor isBranch or: [descriptor isBlockCreation]) 
			ifTrue:
				[distance := self spanFor: descriptor at: pc exts: nExts in: aMethod.
			 	targetPC := pc + descriptor numBytes + distance.
			 	latestContinuation := latestContinuation max: targetPC.
			 	descriptor isBlockCreation ifTrue:
					[pc := pc + distance]]
			ifFalse: 
				[latestContinuation := self maybeUnsafeJumpContinuation: latestContinuation at: pc for: descriptor in: aMethod].
		nExts := descriptor isExtension ifTrue: [nExts + 1] ifFalse: [0].
		pc := pc + descriptor numBytes].
	^end! !

!Cogit methodsFor: 'memory access' stamp: 'eem 7/20/2021 17:16'!
ensureExecutableCodeZone
	"On some platforms run-time calls may be required to enable execution and disable
	 write-protect of the code zone. This is sequenced by ensuring that the code zone is
	 executable most of the time.  Note that any code space modification requires an
	 icache flush (on processors with such an icache). Hence the least invasive time to
	 ensure code is executable is post icache flush.  Making sure code is writable can be
	 done either before any bulk edit (e.g. code zone reclamation) or as part of any fine-
	 grained code modification (e.g. setting an anonymous method's selector)."
	<inline: #always>
	
	self cppIf: #DUAL_MAPPED_CODE_ZONE
		ifFalse:
			[backEnd needsCodeZoneExecuteWriteSwitch ifTrue:
				[self cCode: nil inSmalltalk: [| currentAPISelector |
											"What's all this crap?  We're trying to catch cases where ensureExecutableCodeZone
											 is called without first calling ensureWritableCodeZone, and vice verse.  But there are
											 lots of exceptions where code is not modified but executability is turned on unnecessarily.
											 The list of exceptions follows."
											 currentAPISelector := self debugAPISelector.
											self assert: (codeZoneIsExecutableNotWritable not
														or: [currentAPISelector == #mapObjectReferencesInMachineCode:
															or: [currentAPISelector == #cogitPostGCAction:
															or: [currentAPISelector == #ceSICMiss:
															or: [currentAPISelector == #ceCPICMiss:receiver:
															or: [currentAPISelector == #unlinkSendsOf:isMNUSelector:
															or: [currentAPISelector == #unlinkSendsTo:andFreeIf:
															or: [currentAPISelector == #followMovableLiteralsAndUpdateYoungReferrers]]]]]]])].
				 backEnd makeCodeZoneExecutable.
				 self cCode: nil inSmalltalk: [codeZoneIsExecutableNotWritable := true. debugAPISelector := self debugAPISelector]]]! !

!Cogit methodsFor: 'memory access' stamp: 'eem 7/5/2021 14:49'!
ensureExecutableCodeZoneWithin: aBlock
	"On some platforms run-time calls may be required to enable execution and disable
	 write-protect of the code zone. See the comment in ensureExecutableCodeZone."
	<inline: #always>
	self ensureExecutableCodeZone.
	aBlock value.
	self ensureWritableCodeZone! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 02/09/2017 09:39'!
ensureFixupAt: targetPC
	"Make sure there's a flagged fixup at the target pc in fixups.
	 Initially a fixup's target is just a flag.  Later on it is replaced with a proper instruction."
	<returnTypeC: #'BytecodeFixup *'>
	| fixup |
	<var: #fixup type: #'BytecodeFixup *'>
	fixup := self fixupAt: targetPC.
	fixup notAFixup ifTrue:
		[fixup becomeFixup].
	fixup recordBcpc: bytecodePC.
	^fixup! !

!Cogit methodsFor: 'memory access' stamp: 'eem 7/7/2021 15:53'!
ensureWritableCodeZone
	"On some platforms run-time calls may be required to enable execution and disable
	 write-protect of the code zone. See the comment in ensureExecutableCodeZone."
	<inline: #always>
	self cppIf: #DUAL_MAPPED_CODE_ZONE
		ifFalse:
			[backEnd needsCodeZoneExecuteWriteSwitch ifTrue:
				[self cCode: nil inSmalltalk: [debugAPISelector ifNil: [debugAPISelector := self debugAPISelector].
											self assert: (codeZoneIsExecutableNotWritable or: [debugAPISelector == self debugAPISelector])].
				 backEnd makeCodeZoneWritable.
				 self cCode: nil inSmalltalk: [codeZoneIsExecutableNotWritable := false. debugAPISelector := self debugAPISelector]]]! !

!Cogit methodsFor: 'debugging' stamp: 'eem 11/1/2013 17:25'!
enterCogCodePopReceiver
	"This is a static version of ceEnterCogCodePopReceiverReg
	 for break-pointing when debugging in C."
	<api>
	<inline: false>
	"This exists only for break-pointing."
	self cCode: [self realCEEnterCogCodePopReceiverReg]
		inSmalltalk: [self ceEnterCogCodePopReceiverReg].
	"(and this exists only to reference Debug)"
	Debug ifFalse: [self error: 'what??']! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 1/28/2017 15:23'!
entryCacheTagAndCouldBeObjectAt: mcpc annotation: annotation into: trinaryBlock
	"Evaluate trinaryBlock with the entryPoint, inline cache tag and whether the cache
	 tag could be an object, for the send at mcpc with annotation annotation."
	<inline: true>
	| cacheTag entryPoint tagCouldBeObj |
	cacheTag := backEnd inlineCacheTagAt: mcpc asInteger.
	entryPoint := backEnd callTargetFromReturnAddress: mcpc asInteger.
	"in-line cache tags are the selectors of sends if sends are unlinked,
	 the selectors of super sends (entry offset = cmNoCheckEntryOffset),
	 the selectors of open PIC sends (entry offset = cmEntryOffset, target is an Open PIC)
	 or in-line cache tags (classes, class indices, immediate bit patterns, etc).
	 Note that selectors can be immediate so there is no guarantee that they
	 are markable/remappable objects."
	tagCouldBeObj := self inlineCacheTagsAreIndexes not
						and: [objectRepresentation inlineCacheTagsMayBeObjects
							or: [self entryPointTagIsSelector: entryPoint]].
	trinaryBlock
		value: entryPoint
		value: cacheTag
		value: tagCouldBeObj! !

!Cogit methodsFor: 'accessing' stamp: 'eem 9/6/2016 19:24'!
entryOffset
	<api>
	<cmacro>
	^cmEntryOffset! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 1/28/2017 15:22'!
entryPointTagIsSelector: entryPoint
	"Answer if the entryPoint's tag is expected to be a selector reference, as opposed to a class tag."
	^entryPoint < methodZoneBase
	 or: [(entryPoint bitAnd: entryPointMask) = uncheckedEntryAlignment
	 or: [(entryPoint bitAnd: entryPointMask) = checkedEntryAlignment
		and: [(self cCoerceSimple: entryPoint - cmEntryOffset to: #'CogMethod *') cmType = CMOpenPIC]]]! !

!Cogit methodsFor: 'simulation processor access' stamp: 'eem 12/18/2019 19:42'!
errorProcessingSimulationTrap: aProcessorSimulationTrap in: aDictionary
	self provideContextForCurrentInstruction.
	^aDictionary errorKeyNotFound: aProcessorSimulationTrap address! !

!Cogit methodsFor: 'accessing' stamp: 'cb 5/6/2015 09:20'!
estimateOfAbstractOpcodesPerBytecodes
	<inline: true>
	^ 10! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 11/23/2016 14:39'!
exclude: aMethodObj
	"For debugging, allow excluding methods based on selector or methodClass.  Answer if the mehtod should be excluded."
	<inline: true>
	self cCode: [] inSmalltalk: "for debugging, allow excluding methods based on selector or methodClass"
		[self class initializationOptions
			at: #DoNotJIT
			ifPresent:
				[:excluded| 
				(excluded anySatisfy: [:exclude| aMethodObj = exclude]) ifTrue:
					[coInterpreter transcript
						ensureCr; nextPutAll: 'EXCLUDING ';
						nextPutAll: aMethodObj; nextPutAll: ' (compiled block)';
						cr; flush.
					 ^true]].
		 (compilationTrace anyMask: 1) ifTrue:
			[| methodClass |
			 methodClass := coInterpreter nameOfClass: (coInterpreter methodClassOf: aMethodObj).
			 coInterpreter transcript
				ensureCr;
				nextPutAll: 'compiling compiled block in ';
				nextPutAll: methodClass;
				cr; flush]].
	^false! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 11/23/2016 14:40'!
exclude: aMethodObj selector: aSelectorOop
	"For debugging, allow excluding methods based on selector or methodClass.  Answer if the mehtod should be excluded."
	<inline: true>
	self cCode: [] inSmalltalk:
		[| methodClass selector |
		 self class initializationOptions
			at: #DoNotJIT
			ifPresent:
				[:excluded|
				methodClass := coInterpreter nameOfClass: (coInterpreter methodClassOf: aMethodObj).
				selector := coInterpreter stringOf: aSelectorOop.
				(excluded anySatisfy: [:exclude| selector = exclude or: [methodClass = exclude]]) ifTrue:
					[coInterpreter transcript
						ensureCr; nextPutAll: 'EXCLUDING ';
						nextPutAll: methodClass; nextPutAll: '>>#'; nextPutAll: selector;
						cr; flush.
					 ^true]].
		 (compilationTrace anyMask: 1) ifTrue:
			[methodClass := coInterpreter nameOfClass: (coInterpreter methodClassOf: aMethodObj).
			 selector := coInterpreter stringOf: aSelectorOop.
			 selector isEmpty ifTrue:
				[selector := coInterpreter stringOf: (coInterpreter maybeSelectorOfMethod: aMethodObj)].
			 coInterpreter transcript
				ensureCr; nextPutAll: 'compiling ';
				nextPutAll: methodClass; nextPutAll: '>>#'; nextPutAll: selector;
				cr; flush]].
	^false! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 12/11/2019 15:52'!
expectedClosedPICPrototype: cPIC
	"Use asserts to check if the ClosedPICPrototype is as expected from compileClosedPICPrototype,
	 and can be updated as required via rewriteCPICCaseAt:tag:objRef:target:.  If all asserts pass, answer
	 0, otherwise answer a bit mask identifying all the errors."
	"self disassembleFrom: methodZoneBase + (self sizeof: CogMethod) to: methodZoneBase + closedPICSize"
	<var: #cPIC type: #'CogMethod *'>
	| pc errors object classTag entryPoint |
	errors := 0.
	pc := cPIC asUnsignedInteger + firstCPICCaseOffset.
	"First jump is unconditional; subsequent ones are conditional"
	object := backEnd literalBeforeFollowingAddress: pc - backEnd jumpLongByteSize.
	(self asserta: object = self firstPrototypeMethodOop) ifFalse:
		[errors := 1].

	entryPoint := backEnd jumpLongTargetBeforeFollowingAddress: pc.
	(self asserta: entryPoint = (self cPICPrototypeCaseOffset + 16rCA5E10)) ifFalse:
		[errors := errors + 2].

	1 to: MaxCPICCases - 1 do:
		[:i | | methodObjPC classTagPC |
		pc := pc + cPICCaseSize.

		"verify information in case is as expected."
		methodObjPC := pc - backEnd jumpLongConditionalByteSize - backEnd cmpC32RTempByteSize.
		object := backEnd literalBeforeFollowingAddress: methodObjPC.
		(self asserta: object = (self subsequentPrototypeMethodOop+ i)) ifFalse:
			[errors := errors bitOr: 4].

		classTagPC := pc - backEnd jumpLongConditionalByteSize.
		classTag := backEnd literal32BeforeFollowingAddress: classTagPC.
		(self asserta: classTag = (16rBABE1F15 + i)) ifFalse:
			[errors := errors bitOr: 8].

		entryPoint := backEnd jumpLongConditionalTargetBeforeFollowingAddress: pc.
		(self asserta: entryPoint = (self cPICPrototypeCaseOffset + 16rCA5E10 + (i * 16))) ifFalse:
			[errors := errors bitOr: 16].

		"change case via rewriteCPICCaseAt:tag:objRef:target:"
		self rewriteCPICCaseAt: pc
			tag: (classTag bitXor: 16r5A5A5A5A)
			objRef: (object bitXor: 16rA5A5A5A5)
			target: (entryPoint bitXor: 16r55AA50). "don't xor least 4 bits to leave instruction alignment undisturbed"

		"verify information in case is as expected post update."
		object := backEnd literalBeforeFollowingAddress: methodObjPC.
		(self asserta: object = (self subsequentPrototypeMethodOop + i bitXor: 16rA5A5A5A5)) ifFalse:
			[errors := errors bitOr: 32].
		classTag := backEnd literal32BeforeFollowingAddress: classTagPC.
		(self asserta: classTag = (16rBABE1F15 + i bitXor: 16r5A5A5A5A)) ifFalse:
			[errors := errors bitOr: 64].
		entryPoint := backEnd jumpLongConditionalTargetBeforeFollowingAddress: pc.
		(self asserta: entryPoint = ((self cPICPrototypeCaseOffset + 16rCA5E10 + (i * 16)) bitXor: 16r55AA50)) ifFalse:
			[errors := errors bitOr: 128].

		"finally restore case to the original state"
		self rewriteCPICCaseAt: pc
			tag: (classTag bitXor: 16r5A5A5A5A)
			objRef: (object bitXor: 16rA5A5A5A5)
			target: (entryPoint bitXor: 16r55AA50)].

	entryPoint := backEnd jumpLongTargetBeforeFollowingAddress: cPIC asUnsignedInteger + cPICEndOfCodeOffset.
	(self asserta: entryPoint = (self cPICMissTrampolineFor: 0)) ifFalse:
		[errors := errors + 256].
	
	^errors! !

!Cogit methodsFor: 'bytecode generators' stamp: 'eem 11/2/2012 11:15'!
extABytecode
	"224		11100000	aaaaaaaa	Extend A (Ext A = Ext A prev * 256 + Ext A)"
	extA := (extA bitShift: 8) + byte1.
	^0! !

!Cogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 15:47'!
extBBytecode
	"225		11100001	sbbbbbbb	Extend B (Ext B = Ext B prev * 256 + Ext B)"
	extB := (numExtB = 0 and: [byte1 > 127])
				ifTrue: [byte1 - 256]
				ifFalse: [(extB bitShift: 8) + byte1].
	numExtB := numExtB + 1.
	^0! !

!Cogit methodsFor: 'accessing' stamp: 'eem 2/2/2020 13:18'!
fakeVarBaseAddress
	<doNotGenerate>
	^simulatedAddresses at: #stackPointer! !

!Cogit methodsFor: 'generate machine code' stamp: 'eem 1/28/2020 17:37'!
fillInBlockHeadersAt: startAddress
	"Fill in the block headers now we know the exact layout of the code."
	| blockStart blockHeader |
	<var: #blockStart type: #'BlockStart *'>
	<var: #blockHeader type: #'CogBlockMethod *'>

	(needsFrame and: [blockCount > 0]) ifFalse:
		[^nil].
	blockNoContextSwitchOffset = nil
		ifTrue: [blockNoContextSwitchOffset := blockEntryLabel address - blockEntryNoContextSwitch address]
		ifFalse: [self assert: blockNoContextSwitchOffset = (blockEntryLabel address - blockEntryNoContextSwitch address)].
	0 to: blockCount - 1 do:
		[:i|
		blockStart := self blockStartAt: i.
		blockHeader := self writableBlockMethodFor: blockStart fakeHeader address.
		blockHeader
			homeOffset: (blockStart fakeHeader address - startAddress);
			startpc: blockStart startpc;
			cmType: CMBlock;
			cmNumArgs: blockStart numArgs;
			cbUsesInstVars: blockStart hasInstVarRef;
			stackCheckOffset: (blockStart stackCheckLabel = nil
								ifTrue: [0]
								ifFalse: [blockStart stackCheckLabel address - blockStart fakeHeader address])]! !

!Cogit methodsFor: 'generate machine code' stamp: 'eem 1/27/2020 17:14'!
fillInCPICHeader: pic numArgs: numArgs numCases: numCases hasMNUCase: hasMNUCase selector: selector
	"Fill in the header for the ClosedPIC pic.  This may be located at the writable mapping."
	<var: #pic type: #'CogMethod *'>
	<inline: true>
	self assert: (objectMemory isYoung: selector) not.
	pic cmType: CMClosedPIC.
	pic objectHeader: 0.
	pic blockSize: closedPICSize.
	pic methodObject: 0.
	pic methodHeader: 0.
	pic selector: selector.
	pic cmNumArgs: numArgs.
	pic cmHasMovableLiteral: false.
	pic cmRefersToYoung: false.
	pic cmUsageCount: self initialClosedPICUsageCount.
	pic cpicHasMNUCase: hasMNUCase.
	pic cPICNumCases: numCases.
	pic blockEntryOffset: 0.
	self assert: pic cmType = CMClosedPIC.
	self assert: pic selector = selector.
	self assert: pic cmNumArgs = numArgs.
	self assert: pic cPICNumCases = numCases.
	self assert: closedPICSize = (methodZone roundUpLength: closedPICSize).
	self maybeEnableSingleStep
	"No simulateDualCodeZoneWriteFor:; we do all the simulated copying in the sender..."! !

!Cogit methodsFor: 'generate machine code' stamp: 'eem 7/5/2021 15:25'!
fillInMethodHeader: method size: size selector: selector
	"Fill in the header for theCogMehtod method.  This may be located at the writable mapping."
	<var: #method type: #'CogMethod *'>
	| originalMethod rawHeader actualMethodLocation |
	<var: #originalMethod type: #'CogMethod *'>
	actualMethodLocation := method asUnsignedInteger - codeToDataDelta.
	method cmType: CMMethod.
	method objectHeader: objectMemory nullHeaderForMachineCodeMethod.
	method blockSize: size.
	method methodObject: methodObj.
	rawHeader := coInterpreter rawHeaderOf: methodObj.
	"If the method has already been cogged (e.g. Newspeak accessors) then
	 leave the original method attached to its cog method, but get the right header."
	(coInterpreter isCogMethodReference: rawHeader)
		ifTrue:
			[originalMethod := self cCoerceSimple: rawHeader to: #'CogMethod *'.
			self assert: originalMethod blockSize = size.
			self assert: methodHeader = originalMethod methodHeader.
			NewspeakVM ifTrue:
				[methodZone addToUnpairedMethodList: method]]
		ifFalse:
			[coInterpreter rawHeaderOf: methodObj put: actualMethodLocation.
			 NewspeakVM ifTrue:
				[method nextMethodOrIRCs: theIRCs]].
	method methodHeader: methodHeader.
	method selector: selector.
	method cmNumArgs: (coInterpreter argumentCountOfMethodHeader: methodHeader).
	method cmHasMovableLiteral: hasMovableLiteral.
	(method cmRefersToYoung: hasYoungReferent) ifTrue:
		[methodZone addToYoungReferrers: method].
	method cmUsageCount: self initialMethodUsageCount.
	method cpicHasMNUCase: false.
	method cmUsesPenultimateLit: maxLitIndex >= ((objectMemory literalCountOfMethodHeader: methodHeader) - 2).
	method blockEntryOffset: (blockEntryLabel notNil
								ifTrue: [blockEntryLabel address - actualMethodLocation]
								ifFalse: [0]).
	"This can be an error check since a large stackCheckOffset is caused by compiling
	 a machine-code primitive, and hence depends on the Cogit, not the input method."
	needsFrame ifTrue:
		[stackCheckLabel address - actualMethodLocation <= MaxStackCheckOffset ifFalse:
			[self error: 'too much code for stack check offset']].
	method stackCheckOffset: (needsFrame
								ifTrue: [stackCheckLabel address - actualMethodLocation]
								ifFalse: [0]).

	self assert: (backEnd callTargetFromReturnAddress: actualMethodLocation + missOffset)
				= (self methodAbortTrampolineFor: method cmNumArgs).
	self assert: size = (methodZone roundUpLength: size).
	self assertValidDualZoneFrom: actualMethodLocation to: actualMethodLocation + size.
	self maybeEnableSingleStep! !

!Cogit methodsFor: 'generate machine code' stamp: 'eem 2/3/2020 19:28'!
fillInOPICHeader: pic numArgs: numArgs selector: selector
	"Fill in the header for the OpenPIC pic.  This may be located at the writable mapping."
	<var: #pic type: #'CogMethod *'>
	<inline: true>
	pic cmType: CMOpenPIC.
	pic objectHeader: 0.
	pic blockSize: openPICSize.
	"pic methodObject: 0.""This is also the nextOpenPIC link so don't initialize it"
	methodZone addToOpenPICList: pic.
	pic methodHeader: 0.
	pic selector: selector.
	pic cmNumArgs: numArgs.
	pic cmHasMovableLiteral: (objectMemory isNonImmediate: selector).
	(pic cmRefersToYoung: (objectMemory isYoung: selector)) ifTrue:
		[methodZone addToYoungReferrers: pic].
	pic cmUsageCount: self initialOpenPICUsageCount.
	pic cpicHasMNUCase: false.
	pic cPICNumCases: 0.
	pic blockEntryOffset: 0.

	"This also implicitly flushes the read/write mapped dual zone to the read/execute zone."
	backEnd flushICacheFrom: pic asUnsignedInteger - codeToDataDelta to: pic asUnsignedInteger - codeToDataDelta + openPICSize.

	self assert: pic cmType = CMOpenPIC.
	self assert: pic selector = selector.
	self assert: pic cmNumArgs = numArgs.
	self assert: (backEnd callTargetFromReturnAddress: pic asInteger - codeToDataDelta + missOffset) = (self picAbortTrampolineFor: numArgs).
	self assert: openPICSize = (methodZone roundUpLength: openPICSize).
	self assertValidDualZoneFrom: pic asUnsignedInteger - codeToDataDelta to: pic asUnsignedInteger - codeToDataDelta + openPICSize.
	self maybeEnableSingleStep! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/19/2016 15:04'!
find: descriptor IsBackwardBranch: isBackwardBranchAndAnnotation Mcpc: mcpc Bcpc: bcpc MatchingBcpc: targetBcpc
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<var: #mcpc type: #'char *'>
	<var: #targetBcpc type: #'void *'>
	<inline: true>
	^targetBcpc asInteger = ((descriptor isNil or: [isBackwardBranchAndAnnotation anyMask: 1])
									ifTrue: [bcpc]
									ifFalse: [bcpc + descriptor numBytes])
		ifTrue: [mcpc asInteger]
		ifFalse: [0]! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/19/2016 15:05'!
find: descriptor IsBackwardBranch: isBackwardBranchAndAnnotation Mcpc: mcpc Bcpc: bcpc MatchingMcpc: targetMcpc
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<var: #mcpc type: #'char *'>
	<var: #targetMcpc type: #'void *'>
	"Machine code addresses map to the following bytecode for all bytecodes
	 except backward branches, where they map to the backward branch itself.
	 This is so that loops continue, rather than terminate prematurely."
	^targetMcpc = mcpc
		ifTrue: [(descriptor isNil or: [isBackwardBranchAndAnnotation anyMask: 1])
					ifTrue: [bcpc]
					ifFalse: [bcpc + descriptor numBytes]]
		ifFalse: [0]! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/19/2016 15:03'!
findBackwardBranch: descriptor IsBackwardBranch: isBackwardBranchAndAnnotation Mcpc: mcpc Bcpc: bcpc MatchingBcpc: targetBcpc
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<var: #mcpc type: #'char *'>
	<var: #targetBcpc type: #'void *'>
	<inline: true>
	^((isBackwardBranchAndAnnotation anyMask: 1) and: [targetBcpc asInteger = bcpc])
		ifTrue: [mcpc asInteger]
		ifFalse: [0]! !

!Cogit methodsFor: 'method map' stamp: 'eem 1/9/2011 12:04'!
findBlockMethodWithEntry: blockEntryMcpc startBcpc: startBcpc
	<returnTypeC: #usqInt>
	| cogBlockMethod |
	<var: #cogBlockMethod type: #'CogBlockMethod *'>
	cogBlockMethod := self cCoerceSimple: blockEntryMcpc - (self sizeof: CogBlockMethod)
							  to: #'CogBlockMethod *'.
	cogBlockMethod startpc = startBcpc ifTrue:
		[^cogBlockMethod asUnsignedInteger].
	^0 "keep scanning..."! !

!Cogit methodsFor: 'method map' stamp: 'eem 3/11/2017 10:15'!
findMapLocationForMcpc: targetMcpc inMethod: cogMethod
	<var: #targetMcpc type: #usqInt>
	<var: #cogMethod type: #'CogMethod *'>
	| mcpc map mapByte annotation |
	mcpc := self firstMappedPCFor: cogMethod.
	map := self mapStartFor: cogMethod.
	mcpc = targetMcpc ifTrue: [^map].
	[(mapByte := objectMemory byteAt: map) ~= MapEnd] whileTrue:
		[annotation := mapByte >> AnnotationShift.
		 annotation ~= IsAnnotationExtension ifTrue:
			[mcpc := mcpc + (backEnd codeGranularity
								* (annotation = IsDisplacementX2N
									ifTrue: [mapByte - DisplacementX2N << AnnotationShift]
									ifFalse: [mapByte bitAnd: DisplacementMask]))].
		 mcpc >= targetMcpc ifTrue:
			[self assert: mcpc = targetMcpc.
			 annotation = IsDisplacementX2N ifTrue:
				[map := map - 1.
				 mapByte := objectMemory byteAt: map.
				 annotation := mapByte >> AnnotationShift.
				 self assert: annotation > IsAnnotationExtension].
			 ^map].
		 map := map - 1].
	^0! !

!Cogit methodsFor: 'method map' stamp: 'eem 12/12/2012 11:39'!
findMethodForStartBcpc: startbcpc inHomeMethod: cogMethod
	<api>
	<var: #cogMethod type: #'CogMethod *'>
	<returnTypeC: #'CogBlockMethod *'>
	"Find the CMMethod or CMBlock that has zero-relative startbcpc as its first bytecode pc.
	 As this is for cannot resume processing and/or conversion to machine-code on backward
	 branch, it doesn't have to be fast.  Enumerate block returns and map to bytecode pcs."
	self assert: cogMethod cmType = CMMethod.
	startbcpc = (coInterpreter startPCOfMethodHeader: cogMethod methodHeader) ifTrue:
		[^self cCoerceSimple: cogMethod to: #'CogBlockMethod *'].
	self assert: cogMethod blockEntryOffset ~= 0.
	^self cCoerceSimple: (self blockDispatchTargetsFor: cogMethod
								perform: #findBlockMethodWithEntry:startBcpc:
								arg: startbcpc)
		to: #'CogBlockMethod *'! !

!Cogit methodsFor: 'debugging' stamp: 'eem 1/26/2020 18:48'!
firstInvalidDualZoneAddress
	codeToDataDelta > 0 ifTrue:
		[codeBase to: methodZone zoneEnd - 8 by: objectMemory wordSize do:
			[:address|
			(objectMemory longAt: address) ~= (objectMemory longAt: address + codeToDataDelta) ifTrue:
				[^address]]].
	^nil! !

!Cogit methodsFor: 'debugging' stamp: 'eem 1/27/2020 19:42'!
firstInvalidDualZoneAddressFrom: startAddress to: endAddress
	codeToDataDelta > 0 ifTrue:
		[startAddress to: endAddress - 1 by: objectMemory wordSize do:
			[:address|
			(objectMemory longAt: address) ~= (objectMemory longAt: address + codeToDataDelta) ifTrue:
				[^address]]].
	^nil! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/15/2016 10:04'!
firstMappedPCFor: cogMethod
	<var: #cogMethod type: #'CogMethod *'>
	<inline: true>
	^cogMethod cmIsFullBlock
		ifTrue: [cogMethod asUnsignedInteger + cbNoSwitchEntryOffset]
		ifFalse: [cogMethod asUnsignedInteger + cmNoCheckEntryOffset]! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 12/7/2016 14:35'!
firstPrototypeMethodOop
	"Answer a fake value for the first method oop in the PIC prototype.
	 Since we use MoveUniqueCw:R: it must not be confused with a method-relative address."
	<inline: false>
	^(self addressIsInCurrentCompilation: 16r5EAF00D)
		ifTrue: [16rCA7F00D]
		ifFalse: [16r5EAF00D]! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 2/9/2017 09:33'!
fixupAt: fixupPC
	<inline: true>
	^self fixupAtIndex: fixupPC - initialPC! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 2/9/2017 09:32'!
fixupAtIndex: index
	"The fixups Array maps to bytecode pcs such that initialPC maps to index 0.
	 fixupAt: does the conversion."
	<cmacro: '(index) (&fixups[index])'>
	<returnTypeC: #'BytecodeFixup *'>
	^self addressOf: (fixups at: index)! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/25/2021 10:46'!
followForwardedLiteralsIn: cogMethod
	<api>
	<option: #SpurObjectMemory>
	<var: #cogMethod type: #'CogMethod *'>
	| writableCogMethod hasYoungObj hasYoungObjPtr |
	self assert: (cogMethod cmType ~= CMMethod or: [(objectMemory isForwarded: cogMethod methodObject) not]).
	writableCogMethod := self writableMethodFor: cogMethod.
	self ensureWritableCodeZone.
	hasYoungObj := objectMemory isYoung: cogMethod methodObject.
	(objectMemory shouldRemapOop: cogMethod selector) ifTrue:
		[writableCogMethod selector: (objectMemory remapObj: cogMethod selector).
		 (objectMemory isYoung: cogMethod selector) ifTrue:
			[hasYoungObj := true]].
	hasYoungObjPtr := (self addressOf: hasYoungObj put: [:val| hasYoungObj := val]) asInteger.
	self mapFor: cogMethod
		performUntil: #remapIfObjectRef:pc:hasYoung:
		arg: hasYoungObjPtr asVoidPointer.
	hasYoungObj
		ifTrue: [methodZone ensureInYoungReferrers: cogMethod]
		ifFalse: [writableCogMethod cmRefersToYoung: false].
	self ensureExecutableCodeZone! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/9/2021 14:01'!
followForwardedMethods
	<api>
	<option: #SpurObjectMemory>
	| cogMethod freedPIC |
	self ensureWritableCodeZone.
	freedPIC := false.
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType = CMMethod ifTrue:
			[(objectMemory isForwarded: cogMethod methodObject) ifTrue:
				[cogMethod methodObject: (objectMemory followForwarded: cogMethod methodObject).
				 (objectMemory isYoungObject: cogMethod methodObject) ifTrue:
					[methodZone ensureInYoungReferrers: cogMethod]]].
		 cogMethod cmType = CMClosedPIC ifTrue:
			[(self followMethodReferencesInClosedPIC: cogMethod) ifTrue:
				[freedPIC := true.
				 methodZone freeMethod: cogMethod]].
		 cogMethod := methodZone methodAfter: cogMethod].
	freedPIC ifTrue:
		[self unlinkSendsToFree].
	self ensureExecutableCodeZone! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/9/2021 13:59'!
followMaybeObjRefInClosedPICAt: mcpc
	"Follow a potential object reference from a closed PIC.
	 This may be a method reference or null.
	 Answer if the followed literal is young.
	'mcpc' refers to the jump/branch instruction at the end of
	each cpic case"
	| object subject |
	object := backEnd literalBeforeFollowingAddress: mcpc.
	(objectRepresentation couldBeObject: object) ifFalse:
		[^false].
	(objectMemory isForwarded: object) ifFalse:
		[^objectMemory isYoungObject: object].
	subject := objectMemory followForwarded: object.
	self setCodeModified.
	backEnd storeLiteral: subject beforeFollowingAddress: mcpc.
	^objectMemory isYoungObject: subject! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 11/20/2015 17:44'!
followMethodReferencesInClosedPIC: cPIC
	"Remap all object references in the closed PIC.  Answer if any references are young.
	Set codeModified if any modifications are made."
	<var: #cPIC type: #'CogMethod *'>
	| pc refersToYoung |
	pc := self addressOfEndOfCase: 1 inCPIC: cPIC.

	"first we check the potential method oop load at the beginning of the CPIC"
	refersToYoung := self followMaybeObjRefInClosedPICAt: pc - backEnd jumpLongByteSize.

	"We find the end address of the cPICNumCases'th case and can then just step forward by the case size thereafter"
	pc := self addressOfEndOfCase:  cPIC cPICNumCases inCPIC: cPIC.
	
	"Next we check the potential potential method oop load for each case."
	2 to: cPIC cPICNumCases do:
		[:i|
		(self followMaybeObjRefInClosedPICAt: pc - backEnd jumpLongConditionalByteSize - backEnd cmpC32RTempByteSize) ifTrue:
			[refersToYoung := true].
		pc := pc + cPICCaseSize].
	^refersToYoung! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/9/2021 16:19'!
followMovableLiteralsAndUpdateYoungReferrers
	"To avoid runtime checks on literal variable and literal accesses in == and ~~, 
	 we follow literals in methods having movable literals in the postBecome action.
	 To avoid scanning every method, we annotate cogMethods with the 
	 cmHasMovableLiteral flag."
	<option: #SpurObjectMemory>
	<api>
	<returnTypeC: #void>
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	self assert: methodZone kosherYoungReferrers.
	"methodZone firstBogusYoungReferrer"
	"methodZone occurrencesInYoungReferrers: methodZone firstBogusYoungReferrer"
	codeModified := false.
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType ~= CMFree ifTrue:
			[cogMethod cmHasMovableLiteral ifTrue:
				[self followForwardedLiteralsIn: cogMethod]].
		 cogMethod := methodZone methodAfter: cogMethod]..
	methodZone pruneYoungReferrers.
	codeModified ifTrue: "After updating oops in inline caches we need to flush the icache."
		[backEnd flushICacheFrom: codeBase asUnsignedInteger to: methodZone freeStart].
	"And ensure code zone is executable.  May have pruned young referrers..."
	self ensureExecutableCodeZone! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 5/15/2014 09:51'!
framePointerAddress
	"redirect for the backEnd's genSaveStackPointers"
	<doNotGenerate>
	^coInterpreter framePointerAddress! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 8/20/2021 15:35'!
freeCogMethod: cogMethod
	<api>
	<var: #cogMethod type: #'CogMethod *'>
	self moveProfileToMethods.
	methodZone freeMethod: cogMethod.
	self ensureExecutableCodeZone! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 2/24/2015 11:03'!
freeMethod: cogMethod
	<doNotGenerate>
	methodZone freeMethod: cogMethod! !

!Cogit methodsFor: 'compaction' stamp: 'eem 8/13/2012 13:13'!
freePICsWithFreedTargets
	| cogMethod count |
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	count := 0.
	[cogMethod < methodZone limitZony] whileTrue:
		[(cogMethod cmType = CMClosedPIC
		 and: [self cPICHasFreedTargets: cogMethod]) ifTrue:
			[cogMethod cmType: CMFree].
		 cogMethod := methodZone methodAfter: cogMethod.
		 count := count + 1].
	self assert: count = methodZone numMethods! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 8/20/2021 15:50'!
freeUnmarkedMachineCode
	"Free machine-code methods whose compiled methods are unmarked
	 and open PICs whose selectors are not marked, and closed PICs that
	 refer to unmarked objects."
	<api>
	<option: #SpurObjectMemory>
	| cogMethod freedMethod |
	self moveProfileToMethods. "simulation only..."

	freedMethod := false.
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[(cogMethod cmType = CMMethod
		  and: [(objectMemory isMarked: cogMethod methodObject) not]) ifTrue:
			[freedMethod := true.
			 methodZone freeMethod: cogMethod].
		 (cogMethod cmType = CMOpenPIC
		  and: [(objectMemory isImmediate: cogMethod selector) not
		  and: [(objectMemory isMarked: cogMethod selector) not]]) ifTrue:
			[freedMethod := true.
			 methodZone freeMethod: cogMethod].
		 (cogMethod cmType = CMClosedPIC
		  and: [self closedPICRefersToUnmarkedObject: cogMethod]) ifTrue:
			[freedMethod := true.
			 methodZone freeMethod: cogMethod].
		 cogMethod := methodZone methodAfter: cogMethod].
	freedMethod ifTrue:
		[self unlinkSendsToFree]! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 7/24/2021 16:19'!
gen: opcode "<Integer>"
	| abstractInstruction |
	<inline: false>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #abstractInstruction type: #'AbstractInstruction *'>
	self assert: opcodeIndex < numAbstractOpcodes.
	abstractInstruction := self abstractInstructionAt: opcodeIndex.
	opcodeIndex := opcodeIndex + 1.
	abstractInstruction opcode: opcode.
	self cCode: '' inSmalltalk:
		[abstractInstruction bcpc: bytecodePC.
		 self maybeBreakGeneratingInstructionWithIndex: opcodeIndex - 1].
	^abstractInstruction! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 12/17/2019 12:54'!
gen: opcode "<Integer>" literal32: operandOne "<Integer>" operand: operandTwo "<Integer|CogAbstractInstruction>"
	"Literals are constants that either represent objects on the heap that may get updated by
	 the garbage collector, or pc-relative spans that may get changed by code compaction, and
	 must hence always be encoded in a form that allows updating to refer to a different value."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^literalsManager
		checkLiteral32: operandOne
		forInstruction: (self gen: opcode operand: operandOne operand: operandTwo)! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/11/2015 14:18'!
gen: opcode "<Integer>" literal: operand "<Integer|CogAbstractInstruction>"
	"Literals are constants that either represent objects on the heap that may get updated by
	 the garbage collector, or pc-relative spans that may get changed by code compaction, and
	 must hence always be encoded in a form that allows updating to refer to a different value."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^literalsManager
		checkLiteral: operand
		forInstruction: (self gen: opcode operand: operand)! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/11/2015 14:18'!
gen: opcode "<Integer>" literal: operandOne "<Integer>" operand: operandTwo "<Integer|CogAbstractInstruction>"
	"Literals are constants that either represent objects on the heap that may get updated by
	 the garbage collector, or pc-relative spans that may get changed by code compaction, and
	 must hence always be encoded in a form that allows updating to refer to a different value."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^literalsManager
		checkLiteral: operandOne
		forInstruction: (self gen: opcode operand: operandOne operand: operandTwo)! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 7/24/2021 16:20'!
gen: opcode "<Integer>" operand: operand "<Integer|CogAbstractInstruction>"
	| abstractInstruction |
	<inline: false>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #abstractInstruction type: #'AbstractInstruction *'>
	self assert: opcodeIndex < numAbstractOpcodes.
	abstractInstruction := self abstractInstructionAt: opcodeIndex.
	opcodeIndex := opcodeIndex + 1.
	abstractInstruction opcode: opcode.
	abstractInstruction operands at: 0 put: operand.
	self cCode: '' inSmalltalk:
		[abstractInstruction bcpc: bytecodePC.
		 self maybeBreakGeneratingInstructionWithIndex: opcodeIndex - 1].
	^abstractInstruction! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/11/2015 14:18'!
gen: opcode "<Integer>" operand: operandOne "<Integer|CogAbstractInstruction>" literal: operandTwo "<Integer>"
	"Literals are constants that either represent objects on the heap that may get updated by
	 the garbage collector, or pc-relative spans that may get changed by code compaction, and
	 must hence always be encoded in a form that allows updating to refer to a different value."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^literalsManager
		checkLiteral: operandTwo
		forInstruction: (self gen: opcode operand: operandOne operand: operandTwo)! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 7/24/2021 16:20'!
gen: opcode "<Integer>" operand: operandOne "<Integer|CogAbstractInstruction>" operand: operandTwo "<Integer|CogAbstractInstruction>"
	| abstractInstruction |
	<inline: false>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #abstractInstruction type: #'AbstractInstruction *'>
	self assert: opcodeIndex < numAbstractOpcodes.
	abstractInstruction := self abstractInstructionAt: opcodeIndex.
	opcodeIndex := opcodeIndex + 1.
	abstractInstruction opcode: opcode.
	abstractInstruction operands at: 0 put: operandOne.
	abstractInstruction operands at: 1 put: operandTwo.
	self cCode: '' inSmalltalk:
		[abstractInstruction bcpc: bytecodePC.
		 self maybeBreakGeneratingInstructionWithIndex: opcodeIndex - 1].
	^abstractInstruction! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 7/24/2021 16:20'!
gen: opcode "<Integer>" operand: operandOne "<Integer|CogAbstractInstruction>" operand: operandTwo "<Integer|CogAbstractInstruction>" operand: operandThree "<Integer|CogAbstractInstruction>"
	| abstractInstruction |
	<inline: false>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #abstractInstruction type: #'AbstractInstruction *'>
	self assert: opcodeIndex < numAbstractOpcodes.
	abstractInstruction := self abstractInstructionAt: opcodeIndex.
	opcodeIndex := opcodeIndex + 1.
	abstractInstruction opcode: opcode.
	abstractInstruction operands at: 0 put: operandOne.
	abstractInstruction operands at: 1 put: operandTwo.
	abstractInstruction operands at: 2 put: operandThree.
	self cCode: '' inSmalltalk:
		[abstractInstruction bcpc: bytecodePC.
		 self maybeBreakGeneratingInstructionWithIndex: opcodeIndex - 1].
	^abstractInstruction! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/11/2015 14:18'!
gen: opcode "<Integer>" operand: operandOne "<Integer|CogAbstractInstruction>" quickConstant: operandTwo "Integer>"
	"Quick constants are those the back end is free to encode as compactly as possible."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^literalsManager
		checkQuickConstant: operandTwo
		forInstruction: (self gen: opcode operand: operandOne operand: operandTwo)! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/11/2015 14:19'!
gen: opcode "<Integer>" operand: operandOne "<Integer|CogAbstractInstruction>" quickConstant: operandTwo "<Integer>" operand: operandThree "<Integer|CogAbstractInstruction>"
	"Quick constants are those the back end is free to encode as compactly as possible.""<Integer|CogAbstractInstruction>"
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^literalsManager
		checkQuickConstant: operandTwo
		forInstruction: (self gen: opcode operand: operandOne operand: operandTwo operand: operandThree)! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/11/2015 14:19'!
gen: opcode "<Integer>" quickConstant: operand "<Integer>"
	"Quick constants are those the back end is free to encode as compactly as possible."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^literalsManager
		checkQuickConstant: operand
		forInstruction: (self gen: opcode operand: operand)! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/11/2015 14:19'!
gen: opcode "<Integer>" quickConstant: operandOne "<Integer>" operand: operandTwo "<Integer|CogAbstractInstruction>"
	"Quick constants are those the back end is free to encode as compactly as possible."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^literalsManager
		checkQuickConstant: operandOne
		forInstruction: (self gen: opcode operand: operandOne operand: operandTwo)! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/11/2015 14:19'!
gen: opcode "<Integer>" quickConstant: operandOne "<Integer>" operand: operandTwo "<Integer|CogAbstractInstruction>" operand: operandThree "<Integer|CogAbstractInstruction>"
	"Quick constants are those the back end is free to encode as compactly as possible.""<Integer|CogAbstractInstruction>"
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^literalsManager
		checkQuickConstant: operandOne
		forInstruction: (self gen: opcode operand: operandOne operand: operandTwo operand: operandThree)! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 12/17/2019 13:49'!
gen: opcode "<Integer>" uniqueLiteral32: operandOne "<Integer>" operand: operandTwo "<Integer|CogAbstractInstruction>"
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^literalsManager
		uniqueLiteral32: operandOne
		forInstruction: (self gen: opcode operand: operandOne operand: operandTwo)! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/11/2015 14:19'!
gen: opcode "<Integer>" uniqueLiteral: operandOne "<Integer>" operand: operandTwo "<Integer|CogAbstractInstruction>"
	"Literals are constants that either represent objects on the heap that may get updated by
	 the garbage collector, or pc-relative spans that may get changed by code compaction, and
	 must hence always be encoded in a form that allows updating to refer to a different value."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^literalsManager
		uniqueLiteral: operandOne
		forInstruction: (self gen: opcode operand: operandOne operand: operandTwo)! !

!Cogit methodsFor: 'trampoline support' stamp: 'eem 7/24/2018 15:48'!
genCallMustBeBooleanFor: boolean
	"Call ceSendMustBeBooleanTo: via the relevant trampoline."
	^self CallRT: (boolean = objectMemory falseObject
					ifTrue: [ceSendMustBeBooleanAddFalseTrampoline]
					ifFalse: [ceSendMustBeBooleanAddTrueTrampoline])! !

!Cogit methodsFor: 'initialization' stamp: 'eem 12/1/2020 17:56'!
genCheckForInterruptsTrampoline
	<inline: true>
	self zeroOpcodeIndex.
	"if we have a link register we will assume that it does not get automatically pushed onto the stack
	and thus there is no need to pop it before saving to instructionPointerAddress"
	backEnd hasLinkRegister
		ifTrue:
			[self MoveR: LinkReg Aw: coInterpreter instructionPointerAddress]
		ifFalse:
			[self PopR: TempReg. "instruction pointer"
			 self MoveR: TempReg Aw: coInterpreter instructionPointerAddress].
	^self genTrampolineFor: #ceCheckForInterrupt
		called: 'ceCheckForInterruptTrampoline'
		numArgs: 0
		arg: nil
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: false
		resultReg: NoReg
		appendOpcodes: true! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'rmacnak 11/27/2015 13:16:50'!
genConditionalBranch: opcode operand: operandOne
	^self previousInstruction noteFollowingConditionalBranch: (self gen: opcode operand: operandOne)! !

!Cogit methodsFor: 'initialization' stamp: 'eem 2/1/2020 18:50'!
genEnilopmartFor: regArg1 and: regArg2OrNone and: regArg3OrNone forCall: forCall called: trampolineName
	"An enilopmart (the reverse of a trampoline) is a piece of code that makes
	 the system-call-like transition from the C runtime into generated machine
	 code.  The desired arguments and entry-point are pushed on a stackPage's
	 stack.  The enilopmart pops off the values to be loaded into registers and
	 then executes a return instruction to pop off the entry-point and jump to it.

						BEFORE				AFTER			(stacks grow down)
						whatever			stackPointer ->	whatever
						target address =>	reg1 = reg1val, etc
						reg1val				pc = target address
						reg2val
		stackPointer ->	reg3val"

	<var: #trampolineName type: #'char *'>
	<returnTypeC: #'void (*genEnilopmartForandandforCallcalled(sqInt regArg1, sqInt regArg2OrNone, sqInt regArg3OrNone, sqInt forCall, char *trampolineName))(void)'>

	| size endAddress enilopmart |
	self zeroOpcodeIndex.
	backEnd hasVarBaseRegister ifTrue:
		[self MoveCq: self varBaseAddress R: VarBaseReg]. "Must happen first; value may be used in genLoadStackPointers"
	backEnd genLoadStackPointers.
	regArg3OrNone ~= NoReg ifTrue: [self PopR: regArg3OrNone].
	regArg2OrNone ~= NoReg ifTrue: [self PopR: regArg2OrNone].
	self PopR: regArg1.
	self genEnilopmartReturn: forCall.
	self computeMaximumSizes.
	size := self generateInstructionsAt: methodZoneBase.
	endAddress := self outputInstructionsAt: methodZoneBase.
	self assert: methodZoneBase + size = endAddress.
	enilopmart := methodZoneBase.
	methodZoneBase := self alignUptoRoutineBoundary: endAddress.
	backEnd stopsFrom: endAddress to: methodZoneBase - 1.
	self recordGeneratedRunTime: trampolineName address: enilopmart.
	^self cCoerceSimple: enilopmart to: #'void (*)(void)'! !

!Cogit methodsFor: 'initialization' stamp: 'eem 12/10/2015 09:06'!
genEnilopmartFor: regArg1 and: regArg2 forCall: forCall called: trampolineName
	<inline: true>
	^self genEnilopmartFor: regArg1 and: regArg2 and: NoReg forCall: forCall called: trampolineName! !

!Cogit methodsFor: 'initialization' stamp: 'eem 12/10/2015 09:06'!
genEnilopmartFor: regArg1 forCall: forCall called: trampolineName
	<inline: true>
	^self genEnilopmartFor: regArg1 and: NoReg and: NoReg forCall: forCall called: trampolineName! !

!Cogit methodsFor: 'initialization' stamp: 'eem 3/18/2015 18:13'!
genEnilopmartReturn: forCall
	"An enilopmart (the reverse of a trampoline) is a piece of code that makes
	 the system-call-like transition from the C runtime into generated machine
	 code.  At the point the enilopmart enters machine code via a return instruction,
	 any argument registers have been loaded with their values and the stack, if
	 for call, looks like
							ret pc
			stackPointer ->	target address

	 and if not for call, looks like
							whatever
			stackPointer ->	target address

	 If forCall and running on a CISC, ret pc must be left on the stack.  If forCall and
	 running on a RISC, ret pc must be popped into LinkReg.  In either case, target
	 address must be removed from the stack and jumped/returned to."

	backEnd hasLinkRegister
		ifTrue:
			[forCall
				ifTrue:
					[self PopR: RISCTempReg.
					 self PopR: LinkReg.
					 self JumpR: RISCTempReg]
				ifFalse:
					[backEnd hasPCRegister
						ifTrue: [self PopR: PCReg]
						ifFalse:
							[self PopR: RISCTempReg.
							 self JumpR: RISCTempReg]]]
		ifFalse:
			[self RetN: 0]! !

!Cogit methodsFor: 'trampoline support' stamp: 'eem 3/18/2020 12:33'!
genExternalizePointersForPrimitiveCall
	backEnd hasLinkRegister
		ifTrue:
			[self MoveR: LinkReg Aw: coInterpreter instructionPointerAddress]
		ifFalse: "Set coInterpreter stackPointer to the topmost argument, skipping the return address."
			[self PopR: TempReg. "get retpc"
			 self MoveR: TempReg Aw: coInterpreter instructionPointerAddress].
	^backEnd genSaveStackPointers! !

!Cogit methodsFor: 'trampoline support' stamp: 'eem 7/25/2021 15:43'!
genExternalizeStackPointerForFastPrimitiveCall
	backEnd hasLinkRegister
		ifTrue: [self MoveR: SPReg Aw: coInterpreter stackPointerAddress]
		ifFalse:
			[self AddCq: objectRepresentation wordSize R: SPReg R: TempReg.
			 self MoveR: TempReg Aw: coInterpreter stackPointerAddress]! !

!Cogit methodsFor: 'initialization' stamp: 'eem 1/6/2020 12:10'!
genFFICalloutTrampoline
	<option: #LowcodeVM>
	| startAddress |
	self zeroOpcodeIndexForNewOpcodes.
	backEnd hasLinkRegister ifTrue: [
		self MoveR: LinkReg Aw: coInterpreter instructionPointerAddress
	] ifFalse: [
		self PopR: backEnd registerToSaveIP.
		self MoveR: backEnd registerToSaveIP Aw: coInterpreter instructionPointerAddress
	].

	self CallR: TempReg.

	backEnd hasLinkRegister ifTrue: [
		self MoveAw: coInterpreter instructionPointerAddress R: LinkReg 
	] ifFalse: [
		self MoveAw: coInterpreter instructionPointerAddress R: backEnd registerToSaveIP.
		self PushR: backEnd registerToSaveIP.
	].

	self RetN: 0.

	startAddress := methodZoneBase.
	self outputInstructionsForGeneratedRuntimeAt: startAddress.
	self recordGeneratedRunTime: 'ceFFICalloutTrampoline' address: startAddress.
	self recordRunTimeObjectReferences.
	^ startAddress
! !

!Cogit methodsFor: 'initialization' stamp: 'eem 2/4/2020 12:05'!
genGetLeafCallStackPointers
	"Generate a pair of routines that answer the frame pointer, and the stack pointer immediately
	 after a leaf call, used for checking stack pointer alignment, frame pointer usage, etc. N.B.
	 these are exported to the CoInterpreter et al via Cogit class>>mustBeGlobal:."
	| startAddress |
	<inline: false>
	self allocateOpcodes: 4 bytecodes: 0.
	startAddress := methodZoneBase.
	self
		MoveR: FPReg R: ABIResultReg;
		RetN: 0.
	self outputInstructionsForGeneratedRuntimeAt: startAddress.
	self recordGeneratedRunTime: 'ceGetFP' address: startAddress.
	ceGetFP := self cCoerceSimple: startAddress to: #'usqIntptr_t (*)(void)'.
	startAddress := methodZoneBase.
	self zeroOpcodeIndex.
	self MoveR: NativeSPReg R: ABIResultReg.
	backEnd leafCallStackPointerDelta ~= 0 ifTrue:
		[self AddCq: backEnd leafCallStackPointerDelta R: ABIResultReg].
	self RetN: 0.
	self outputInstructionsForGeneratedRuntimeAt: startAddress.
	self recordGeneratedRunTime: 'ceGetSP' address: startAddress.
	ceGetSP := self cCoerceSimple: startAddress to: #'usqIntptr_t (*)(void)'! !

!Cogit methodsFor: 'initialization' stamp: 'eem 03/03/2016 12:23'!
genInnerPICAbortTrampoline: name
	"Generate the abort for a PIC.  This abort performs either a call of
	 ceInterpretMethodFromPIC:receiver: to handle invoking an uncogged target
	 or a call of ceMNUFromPICMNUMethod:receiver: to handle an MNU dispatch
	 in a closed PIC.  It distinguishes the two by testing ClassReg.  If the register
	 is zero then this is an MNU.
	
	 This poses a problem in 32-bit Spur, where zero is the cache tag for immediate
	 characters (tag pattern 2r10) because SmallIntegers have tag patterns 2r11
	 and 2r01, so anding with 1 reduces these to 0 & 1.  We solve the ambiguity by
	 patching send sites with a 0 cache tag to open PICs instead of closed PICs."
	<var: #name type: #'char *'>
	| jumpMNUCase |
	<var: #jumpMNUCase type: #'AbstractInstruction *'>
	self CmpCq: self picAbortDiscriminatorValue R: ClassReg.
	jumpMNUCase := self JumpZero: 0.
	self compileTrampolineFor: #ceInterpretMethodFromPIC:receiver:
		numArgs: 2
		arg: SendNumArgsReg
		arg: ReceiverResultReg
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: false
		resultReg: NoReg.
	jumpMNUCase jmpTarget: self Label.
	^self genTrampolineFor: #ceMNUFromPICMNUMethod:receiver:
		called: name
		numArgs: 2
		arg: SendNumArgsReg
		arg: ReceiverResultReg
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: false
		resultReg: NoReg
		appendOpcodes: true! !

!Cogit methodsFor: 'initialization' stamp: 'eem 7/16/2020 16:49'!
genInvokeInterpretTrampoline
	"Switch to the C stack (do *not* save the Smalltalk stack pointers;
	 this is the caller's responsibility), and invoke interpret PDQ."
	<returnTypeC: 'void (*genInvokeInterpretTrampoline(void))(void)'>
	| startAddress |
	<inline: false>
	startAddress := methodZoneBase.
	self zeroOpcodeIndex.
	backEnd hasVarBaseRegister ifTrue:
		[self MoveCq: self varBaseAddress R: VarBaseReg]. "Must happen first; value may be used in genLoadStackPointers"
	cFramePointerInUse
		ifTrue: [backEnd genLoadCStackPointers]
		ifFalse: [backEnd genLoadCStackPointer].
	"Sideways call interpret so that the stack looks correct, for exception handling etc"
	backEnd genMarshallNArgs: 0 arg: nil arg: nil arg: nil arg: nil.
	backEnd hasLinkRegister
		ifTrue:
			[self MoveAw: coInterpreter cReturnAddressAddress R: LinkReg]
		ifFalse:
			[self MoveAw: coInterpreter cReturnAddressAddress R: ABIResultReg.
			 backEnd genSubstituteReturnAddressR: ABIResultReg].
	self JumpFullRT: (self
						cCode: [#interpret asUnsignedInteger]
						inSmalltalk: [self simulatedTrampolineFor: #interpret]).
	self outputInstructionsForGeneratedRuntimeAt: startAddress.
	self recordGeneratedRunTime: 'ceInvokeInterpret' address: startAddress.
	^self cCoerceSimple: startAddress to: #'void (*)(void)'! !

!Cogit methodsFor: 'trampoline support' stamp: 'eem 3/18/2020 12:36'!
genLoadCStackPointersForPrimCall
	<inline: #always>
	cFramePointerInUse
		ifTrue: [backEnd genLoadCStackPointers]
		ifFalse: [backEnd genLoadCStackPointer]! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 9/9/2019 09:54'!
genLoadInlineCacheWithSelector: selectorIndex
	"The in-line cache for a send is implemented as a constant load into ClassReg.
	 We always use a 32-bit load, even in 64-bits.

	 In the initial (unlinked) state the in-line cache is notionally loaded with the selector.
	 But since in 64-bits an arbitrary selector oop won't fit in a 32-bit constant load, we
	 instead load the cache with the selector's index, either into the literal frame of the
	 current method, or into the special selector array.  Negative values are 1-relative
	 indices into the special selector array.

	 When a send is linked, the load of the selector, or selector index, is overwritten with a
	 load of the receiver's class, or class tag.  Hence, the 64-bit VM is currently constrained
	 to use class indices as cache tags.  If out-of-line literals are used, distinct caches /must
	 not/ share acche locations, for if they do, send cacheing will be confused by the sharing.
	 Hence we use the MoveUniqueC32:R: instruction that will not share literal locations."

	| cacheValue |
	self assert: (selectorIndex < 0
					ifTrue: [selectorIndex negated between: 1 and: self numSpecialSelectors]
					ifFalse: [selectorIndex between: 0 and: (objectMemory literalCountOf: methodObj) - 1]).

	self inlineCacheTagsAreIndexes
		ifTrue:
			[cacheValue := selectorIndex]
		ifFalse:
			[| selector |
			 selector := selectorIndex < 0
							ifTrue: [(coInterpreter specialSelector: -1 - selectorIndex)]
							ifFalse: [self getLiteral: selectorIndex].
			 self assert: (objectMemory addressCouldBeOop: selector).
			(objectMemory isNonImmediate: selector) ifTrue:
				[self setHasMovableLiteral: true].
			 (objectMemory isYoung: selector) ifTrue:
				[self setHasYoungReferent: true].
			 cacheValue := selector].

	self MoveUniqueC32: cacheValue R: ClassReg! !

!Cogit methodsFor: 'initialization' stamp: 'eem 03/03/2016 12:23'!
genMethodAbortTrampoline
	"Generate the abort for a method.  This abort performs either a call of ceSICMiss:
	 to handle a single-in-line cache miss or a call of ceStackOverflow: to handle a
	 stack overflow.  It distinguishes the two by testing ResultReceiverReg.  If the
	 register is zero then this is a stack-overflow because a) the receiver has already
	 been pushed and so can be set to zero before calling the abort, and b) the
	 receiver must always contain an object (and hence be non-zero) on SIC miss."
	| jumpSICMiss |
	<var: #jumpSICMiss type: #'AbstractInstruction *'>
	self zeroOpcodeIndex.
	self CmpCq: 0 R: ReceiverResultReg.
	jumpSICMiss := self JumpNonZero: 0.

	"The abort sequencer has pushed the LinkReg a second time.
	 Overwrite it with the right one."
	backEnd hasLinkRegister ifTrue:
		[self MoveR: LinkReg Mw: 0 r: SPReg].
	self compileTrampolineFor: #ceStackOverflow:
		numArgs: 1
		arg: SendNumArgsReg
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: false "The LinkReg has already been set above."
		resultReg: NoReg.
	jumpSICMiss jmpTarget: self Label.
	^self genTrampolineFor: #ceSICMiss:
		called: 'ceMethodAbort'
		numArgs: 1
		arg: ReceiverResultReg
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true "Push the LinkReg for the ceMethodAbort call."
		resultReg: NoReg
		appendOpcodes: true! !

!Cogit methodsFor: 'initialization' stamp: 'eem 03/03/2016 12:23'!
genNSSendTrampolineFor: aRoutine numArgs: numArgs enclosingObjectCheck: eoCheckFlag called: aString
	"ReceiverResultReg: method receiver
	SendNumArgsReg: the NSSendCache cache"
	<option: #NewspeakVM>
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	| jumpMiss jumpItsTheReceiverStupid |
	<var: #jumpMiss type: #'AbstractInstruction *'>
	<var: #jumpItsTheReceiverStupid type: #'AbstractInstruction *'>
	self zeroOpcodeIndex.
	objectRepresentation
		genGetInlineCacheClassTagFrom: ReceiverResultReg
		into: ClassReg
		forEntry: false.
	self MoveMw: NSCClassTagIndex * objectMemory wordSize r: SendNumArgsReg R: TempReg.
	self CmpR: ClassReg R: TempReg.
	jumpMiss := self JumpNonZero: 0.

	eoCheckFlag ifTrue:
		[self MoveMw: NSCEnclosingObjectIndex * objectMemory wordSize r: SendNumArgsReg R: TempReg.
		self CmpCq: 0 R: TempReg.
		jumpItsTheReceiverStupid := self JumpZero: 0.
		self MoveR: TempReg R: ReceiverResultReg.
		"Now set the stacked receiver, if needed.  If there are reg args this is
		 not required; see genPushRegisterArgsForNumArgs:numArgs: below."
		(self numRegArgs = 0 or: [numArgs > self numRegArgs]) ifTrue:
			[numArgs >= (NumSendTrampolines - 1)
				ifTrue: "arbitrary argument count"
					[self MoveMw: NSCNumArgsIndex * objectMemory wordSize r: SendNumArgsReg R: TempReg.
					 backEnd hasLinkRegister ifFalse:
						[self AddCq: 1 R: TempReg]..
					 self MoveR: ReceiverResultReg Xwr: TempReg R: SPReg]
				ifFalse: "Known argument count"
					[self MoveR: TempReg Mw: (backEnd hasLinkRegister ifTrue: [0] ifFalse: [1]) + numArgs * objectMemory wordSize r: SPReg]].
		jumpItsTheReceiverStupid jmpTarget: self Label].

	self MoveMw: NSCTargetIndex * objectMemory wordSize r: SendNumArgsReg R: TempReg.
	self JumpR: TempReg.

	jumpMiss jmpTarget: self Label.
	objectRepresentation
		genEnsureOopInRegNotForwarded: ReceiverResultReg
		scratchReg: TempReg
		updatingMw: FoxMFReceiver
		r: FPReg.
	self numRegArgs > 0 ifTrue:
		[backEnd genPushRegisterArgsForNumArgs: numArgs scratchReg: TempReg].
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 2
		arg: SendNumArgsReg "The NSSendCache"
		arg: ReceiverResultReg
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: true! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:00'!
genNonLocalReturnTrampoline
	<inline: true>
	self zeroOpcodeIndex.
	"write the return address to the coInterpreter instructionPointerAddress;
	 following the CallRT to this CISCs will have pushed it on the stack, so pop it first; RISCs will have it in
	 their link register so just write it directly."
	backEnd hasLinkRegister
		ifTrue:
			[self MoveR: LinkReg Aw: coInterpreter instructionPointerAddress]
		ifFalse:
			[self PopR: TempReg. "instruction pointer"
			 self MoveR: TempReg Aw: coInterpreter instructionPointerAddress].
	^self genTrampolineFor: #ceNonLocalReturn:
		called: 'ceNonLocalReturnTrampoline'
		numArgs: 1
		arg: ReceiverResultReg
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: false
		resultReg: NoReg
		appendOpcodes: true! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:00'!
genPICAbortTrampoline
	"Generate the abort for a PIC.  This abort performs either a call of
	 ceInterpretMethodFromPIC:receiver: to handle invoking an uncogged
	 target or a call of ceMNUFromPICMNUMethod:receiver: to handle an
	 MNU dispatch in a closed PIC.  It distinguishes the two by testing
	 ClassReg.  If the register is zero then this is an MNU."
	<inline: true>
	self zeroOpcodeIndex.
	backEnd hasLinkRegister ifTrue:
		[self PushR: LinkReg].
	^self genInnerPICAbortTrampoline: 'cePICAbort'! !

!Cogit methodsFor: 'initialization' stamp: 'eem 10/29/2020 12:52'!
genReturnToInterpreterTrampoline
	| startAddress |
	<inline: false>
	startAddress := methodZoneBase.
	self zeroOpcodeIndex.
	"Push the result, set the instruction pointer to the interpreter frame's saved ip,
	 set the method and the bytecode set offset, then call interpret."
	self PushR: ReceiverResultReg. "The result"
	"Assign the iframeSavedIP to instructionPointer"
	self MoveMw: FoxIFSavedIP r: FPReg R: TempReg.
	self MoveR: TempReg Aw: coInterpreter instructionPointerAddress.
	self genSmalltalkToCStackSwitch: false "pushLinkReg".
	"Sideways call interpret so that the stack looks correct, for exception handling etc"
	backEnd genMarshallNArgs: 0 arg: nil arg: nil arg: nil arg: nil.
	backEnd hasLinkRegister
		ifTrue:
			[self MoveAw: coInterpreter cReturnAddressAddress R: LinkReg]
		ifFalse:
			[self MoveAw: coInterpreter cReturnAddressAddress R: ABIResultReg.
			 backEnd genSubstituteReturnAddressR: ABIResultReg].
	self JumpFullRT: (self
						cCode: [#interpret asUnsignedInteger]
						inSmalltalk: [self simulatedTrampolineFor: #interpret]).
	self outputInstructionsForGeneratedRuntimeAt: startAddress.
	self recordGeneratedRunTime: 'ceReturnToInterpreterTrampoline' address: startAddress.
	^startAddress! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:01'!
genReturnTrampolineFor: aRoutine  called: aString arg: regOrConst0
	"Generate a trampoline for a routine used as a return address, that has one argument.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	<inline: #always>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 1
		arg: regOrConst0
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: false "Since the routine is reached by a return instruction it should /not/ push the link register."
		resultReg: NoReg
		appendOpcodes: false! !

!Cogit methodsFor: 'trampoline support' stamp: 'eem 5/1/2015 16:14'!
genSmalltalkToCStackSwitch: pushLinkReg
	"If the client requires, then on an ARM-like RISC processor, the return address needs to
	 be pushed to the stack so that the interpreter sees the same stack layout as on CISC."
	(backEnd hasLinkRegister and: [pushLinkReg]) ifTrue:
		[self PushR: LinkReg].
	backEnd genSaveStackPointers.
	cFramePointerInUse
		ifTrue: [backEnd genLoadCStackPointers]
		ifFalse: [backEnd genLoadCStackPointer].
	^0! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:01'!
genTrampolineFor: aRoutine called: aString
	"Generate a trampoline with no arguments"
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	<inline: #always>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 0
		arg: nil
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:01'!
genTrampolineFor: aRoutine  called: aString arg: regOrConst0
	"Generate a trampoline with one argument.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	<inline: #always>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 1
		arg: regOrConst0
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:01'!
genTrampolineFor: aRoutine called: aString arg: regOrConst0 arg: regOrConst1
	"Generate a trampoline with two arguments.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	<inline: #always>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 2
		arg: regOrConst0
		arg: regOrConst1
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:01'!
genTrampolineFor: aRoutine called: aString arg: regOrConst0 arg: regOrConst1 arg: regOrConst2
	"Generate a trampoline with three arguments.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	<inline: #always>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 3
		arg: regOrConst0
		arg: regOrConst1
		arg: regOrConst2
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:01'!
genTrampolineFor: aRoutine called: aString arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3
	"Generate a trampoline with four arguments.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	<inline: #always>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 4
		arg: regOrConst0
		arg: regOrConst1
		arg: regOrConst2
		arg: regOrConst3
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:01'!
genTrampolineFor: aRoutine called: aString arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 result: resultReg
	"Generate a trampoline with two arguments that answers a result.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	<inline: #always>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 3
		arg: regOrConst0
		arg: regOrConst1
		arg: regOrConst2
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: resultReg
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:01'!
genTrampolineFor: aRoutine called: aString arg: regOrConst0 arg: regOrConst1 regsToSave: regMask
	"Generate a trampoline with two arguments.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	<inline: #always>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 2
		arg: regOrConst0
		arg: regOrConst1
		arg: nil
		arg: nil
		regsToSave: regMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:01'!
genTrampolineFor: aRoutine called: aString arg: regOrConst0 arg: regOrConst1 result: resultReg
	"Generate a trampoline with two arguments that answers a result.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	<inline: #always>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 2
		arg: regOrConst0
		arg: regOrConst1
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: resultReg
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'RonieSalgado 5/2/2016 15:12'!
genTrampolineFor: aRoutine called: aString arg: regOrConst0 floatResult: resultReg
	"Generate a trampoline with one argument that answers a result.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 1
		arg: regOrConst0
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		floatResultReg: resultReg
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:01'!
genTrampolineFor: aRoutine  called: aString arg: regOrConst0 regsToSave: regMask
	"Generate a trampoline with one argument.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	<inline: #always>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 1
		arg: regOrConst0
		arg: nil
		arg: nil
		arg: nil
		regsToSave: regMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:01'!
genTrampolineFor: aRoutine called: aString arg: regOrConst0 regsToSave: regMask result: resultReg
	"Generate a trampoline with one argument that answers a result.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	<inline: #always>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 1
		arg: regOrConst0
		arg: nil
		arg: nil
		arg: nil
		regsToSave: regMask
		pushLinkReg: true
		resultReg: resultReg
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:01'!
genTrampolineFor: aRoutine called: aString arg: regOrConst0 result: resultReg
	"Generate a trampoline with one argument that answers a result.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	<inline: #always>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 1
		arg: regOrConst0
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: resultReg
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'eem 10/1/2016 16:16'!
genTrampolineFor: aRoutine called: aString arg: regOrConst0 result: resultReg result: resultReg2
	"Generate a trampoline with one argument that answers a result.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<option: #LowcodeVM>
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 1
		arg: regOrConst0
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: resultReg
		resultReg: resultReg2
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'RonieSalgado 5/2/2016 14:23'!
genTrampolineFor: aRoutine called: aString floatArg: regOrConst0 result: resultReg
	"Generate a trampoline with one argument that answers a result.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 1
		floatArg: regOrConst0
		floatArg: nil
		floatArg: nil
		floatArg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: resultReg
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'RonieSalgado 5/2/2016 14:15'!
genTrampolineFor: aRoutine called: trampolineName numArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3 regsToSave: regMask pushLinkReg: pushLinkReg floatResultReg: resultRegOrNone appendOpcodes: appendBoolean
	"Generate a trampoline with up to four arguments.  Generate either a call or a jump to aRoutineOrNil
	 as requested by callJumpBar.  If generating a call and resultRegOrNone is not NoReg pass the C result
	 back in resultRegOrNone.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #trampolineName type: #'char *'>
	| startAddress |
	<inline: false>
	startAddress := methodZoneBase.
	appendBoolean ifFalse:
		[self zeroOpcodeIndex].
	self compileTrampolineFor: aRoutine
		numArgs: numArgs
		arg: regOrConst0
		arg: regOrConst1
		arg: regOrConst2
		arg: regOrConst3
		regsToSave: regMask
		pushLinkReg: pushLinkReg
		floatResultReg: resultRegOrNone.
	self outputInstructionsForGeneratedRuntimeAt: startAddress.
	self recordGeneratedRunTime: trampolineName address: startAddress.
	self recordRunTimeObjectReferences.
	^startAddress! !

!Cogit methodsFor: 'initialization' stamp: 'eem 1/27/2020 19:41'!
genTrampolineFor: aRoutine called: trampolineName numArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3 regsToSave: regMask pushLinkReg: pushLinkReg resultReg: resultRegOrNone appendOpcodes: appendBoolean
	"Generate a trampoline with up to four arguments.  Generate either a call or a jump to aRoutineOrNil
	 as requested by callJumpBar.  If generating a call and resultRegOrNone is not NoReg pass the C result
	 back in resultRegOrNone.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #trampolineName type: #'char *'>
	| startAddress |
	<inline: false>
	startAddress := methodZoneBase.
	appendBoolean ifFalse:
		[self zeroOpcodeIndex].
	self compileTrampolineFor: aRoutine
		numArgs: numArgs
		arg: regOrConst0
		arg: regOrConst1
		arg: regOrConst2
		arg: regOrConst3
		regsToSave: regMask
		pushLinkReg: pushLinkReg
		resultReg: resultRegOrNone.
	self outputInstructionsForGeneratedRuntimeAt: startAddress.
	self recordGeneratedRunTime: trampolineName address: startAddress.
	self recordRunTimeObjectReferences.
	self assertValidDualZoneFrom: codeBase to: methodZoneBase.
	^startAddress! !

!Cogit methodsFor: 'initialization' stamp: 'RonieSalgado 5/1/2016 18:28'!
genTrampolineFor: aRoutine called: trampolineName numArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3 regsToSave: regMask pushLinkReg: pushLinkReg resultReg: resultRegOrNone resultReg: resultReg2OrNone appendOpcodes: appendBoolean
	"Generate a trampoline with up to four arguments.  Generate either a call or a jump to aRoutineOrNil
	 as requested by callJumpBar.  If generating a call and resultRegOrNone is not NoReg pass the C result
	 back in resultRegOrNone.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #trampolineName type: #'char *'>
	| startAddress |
	<inline: false>
	startAddress := methodZoneBase.
	appendBoolean ifFalse:
		[self zeroOpcodeIndex].
	self compileTrampolineFor: aRoutine
		numArgs: numArgs
		arg: regOrConst0
		arg: regOrConst1
		arg: regOrConst2
		arg: regOrConst3
		regsToSave: regMask
		pushLinkReg: pushLinkReg
		resultReg: resultRegOrNone
		resultReg: resultReg2OrNone.
	self outputInstructionsForGeneratedRuntimeAt: startAddress.
	self recordGeneratedRunTime: trampolineName address: startAddress.
	self recordRunTimeObjectReferences.
	^startAddress! !

!Cogit methodsFor: 'initialization' stamp: 'eem 10/1/2016 15:56'!
genTrampolineFor: aRoutine called: trampolineName numArgs: numArgs floatArg: regOrConst0 floatArg: regOrConst1 floatArg: regOrConst2 floatArg: regOrConst3 regsToSave: regMask pushLinkReg: pushLinkReg resultReg: resultRegOrNone appendOpcodes: appendBoolean
	"Generate a trampoline with up to four arguments.  Generate either a call or a jump to aRoutineOrNil
	 as requested by callJumpBar.  If generating a call and resultRegOrNone is not NoReg pass the C result
	 back in resultRegOrNone.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<option: #LowcodeVM>
	<var: #aRoutine type: #'void *'>
	<var: #trampolineName type: #'char *'>
	| startAddress |
	<inline: false>
	startAddress := methodZoneBase.
	appendBoolean ifFalse:
		[self zeroOpcodeIndex].
	self compileTrampolineFor: aRoutine
		numArgs: numArgs
		floatArg: regOrConst0
		floatArg: regOrConst1
		floatArg: regOrConst2
		floatArg: regOrConst3
		regsToSave: regMask
		pushLinkReg: pushLinkReg
		resultReg: resultRegOrNone.
	self outputInstructionsForGeneratedRuntimeAt: startAddress.
	self recordGeneratedRunTime: trampolineName address: startAddress.
	self recordRunTimeObjectReferences.
	^startAddress! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:01'!
genTrampolineFor: aRoutine called: aString regsToSave: regMask
	"Generate a trampoline with no arguments"
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	<inline: #always>
	^self
		genTrampolineFor: aRoutine
		called: aString
		numArgs: 0
		arg: nil
		arg: nil
		arg: nil
		arg: nil
		regsToSave: regMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: false! !

!Cogit methodsFor: 'initialization' stamp: 'eem 1/7/2020 10:57'!
genTrampolineReturn: lnkRegWasPushed
	"To return from a trampoline call we have to take the return address off the stack,
	 iof it has been saved"

	(lnkRegWasPushed and: [backEnd hasLinkRegister])
		ifTrue:
			[backEnd hasPCRegister
				ifTrue: [self PopR: PCReg]
				ifFalse: [self PopR: LinkReg. 
						self RetN: 0]]
		ifFalse: [self RetN: 0]! !

!Cogit methodsFor: 'initialization' stamp: 'eem 2/3/2020 19:41'!
generateCaptureCStackPointers: captureFramePointer
	"Generate the routine that writes the current values of the C frame and stack pointers into
	 variables.  These are used to establish the C stack in trampolines back into the C run-time.
	 This routine assumes the system's frame pointer is the same as that used in generated code."
	| startAddress callerSavedReg pushedVarBaseReg |
	<inline: #never>
	self allocateOpcodes: 32 bytecodes: 0.
	startAddress := methodZoneBase.
	 "Must happen first; value may be used in accessing any of the following addresses"
	callerSavedReg := 0.
	pushedVarBaseReg := false.
	backEnd hasVarBaseRegister ifTrue:
		[(self isCallerSavedReg: VarBaseReg) ifFalse:
			["VarBaseReg is not caller-saved; must save and restore it, either by using an available caller-saved reg or push/pop."
			 callerSavedReg := self availableRegisterOrNoneIn: (ABICallerSavedRegisterMask bitClear: 1 << TempReg). "TempReg used below"
			 callerSavedReg = NoReg
				ifTrue: [self NativePushR: VarBaseReg. pushedVarBaseReg := true]
				ifFalse: [self MoveR: VarBaseReg R: callerSavedReg]].
		 self MoveCq: self varBaseAddress R: VarBaseReg].
	captureFramePointer ifTrue:
		[self MoveR: FPReg Aw: self cFramePointerAddress].
	"Capture the stack pointer prior to the call.  If we've pushed VarBaseReg take that into account."
	(backEnd leafCallStackPointerDelta ~= 0 or: [pushedVarBaseReg])
		ifTrue:
			[self LoadEffectiveAddressMw:
					(pushedVarBaseReg
						ifTrue: [backEnd leafCallStackPointerDelta + objectMemory wordSize]
						ifFalse: [backEnd leafCallStackPointerDelta])
				r: NativeSPReg
				R: TempReg.
			 self MoveR: TempReg Aw: self cStackPointerAddress]
		ifFalse:
			[self MoveR: NativeSPReg Aw: self cStackPointerAddress].
	backEnd hasVarBaseRegister ifTrue:
		[(self isCallerSavedReg: VarBaseReg) ifFalse:
			[pushedVarBaseReg
				ifTrue: [self NativePopR: VarBaseReg]
				ifFalse: [self MoveR: callerSavedReg R: VarBaseReg]]].
	self NativeRetN: 0.
	self outputInstructionsForGeneratedRuntimeAt: startAddress.
	"This also implicitly flushes the read/write mapped dual zone to the read/execute zone."
	backEnd flushICacheFrom: startAddress asUnsignedInteger to: methodZoneBase asUnsignedInteger.
	self recordGeneratedRunTime: 'ceCaptureCStackPointers' address: startAddress.
	ceCaptureCStackPointers := self cCoerceSimple: startAddress to: #'void (*)(void)'! !

!Cogit methodsFor: 'initialization' stamp: 'eem 7/6/2021 14:22'!
generateClosedPICPrototype
	"Generate the prototype ClosedPIC to determine how much space a full closed PIC takes.
	 When we first allocate a closed PIC it only has one or two cases and we want to grow it.
	 So we have to determine how big a full one is before hand."
	| cPIC endAddress |
	"stack allocate the various collections so that they
	 are effectively garbage collected on return."
	self allocateOpcodes: MaxCPICCases * 9 bytecodes: 0.
	methodLabel address: methodZoneBase; dependent: nil. "for pc-relative MoveCw: cPIC R: ClassReg"
	self compileClosedPICPrototype.
	self computeMaximumSizes.
	cPIC := (self cCoerceSimple: methodZoneBase to: #'CogMethod *').
	closedPICSize := (self sizeof: CogMethod) + (self generateInstructionsAt: methodZoneBase + (self sizeof: CogMethod)).
	endAddress := self outputInstructionsAt: methodZoneBase + (self sizeof: CogMethod).
	self assert: methodZoneBase + closedPICSize = endAddress.
	firstCPICCaseOffset := endCPICCase0 address - methodZoneBase.
	cPICEndOfCodeOffset := cPICEndOfCodeLabel address - methodZoneBase.
	cPICCaseSize := endCPICCase1 address - endCPICCase0 address.
	cPICEndSize := closedPICSize - (MaxCPICCases - 1 * cPICCaseSize + firstCPICCaseOffset).
	closedPICSize := methodZone roundUpLength: closedPICSize.
	self assert: picInterpretAbort address = (methodLabel address + self picInterpretAbortOffset).
	self assert: (self expectedClosedPICPrototype: cPIC) = 0.
	
	"tpr this is a little tiresome but after any assert checking we need to 0 out the case0 objRef rather than leaving 16r5EAF00D lying around"

	backEnd storeLiteral: 0 beforeFollowingAddress: endCPICCase0 address - backEnd jumpLongByteSize.
	
	"update the methodZoneBase so we keep the prototype around for later use"
	methodZoneBase := self alignUptoRoutineBoundary: endAddress.
	cPICPrototype := cPIC.
	"self cCode: ''
		inSmalltalk:
			[self disassembleFrom: cPIC + (self sizeof: CogMethod) to: cPIC + closedPICSize - 1.
			 self halt]"! !

!Cogit methodsFor: 'generate machine code' stamp: 'eem 7/25/2021 11:20'!
generateCogFullBlock
	"We handle jump sizing simply.  First we make a pass that asks each
	 instruction to compute its maximum size.  Then we make a pass that
	 sizes jumps based on the maxmimum sizes.  Then we make a pass
	 that fixes up jumps.  When fixing up a jump the jump is not allowed to
	 choose a smaller offset but must stick to the size set in the second pass."
	<returnTypeC: #'CogMethod *'>
	<option: #SistaV1BytecodeSet>
	| codeSize headerSize mapSize totalSize startAddress result method |
	<var: #method type: #'CogMethod *'>
	headerSize := self sizeof: CogMethod.
	methodLabel address: methodZone freeStart.
	self computeMaximumSizes.
	methodLabel concretizeAt: methodZone freeStart.
	codeSize := self generateInstructionsAt: methodLabel address + headerSize.
	mapSize := self generateMapAt: nil start: methodLabel address + cbNoSwitchEntryOffset.
.
	totalSize := methodZone roundUpLength: headerSize + codeSize + mapSize.
	totalSize > MaxMethodSize ifTrue:
		[^self cCoerceSimple: MethodTooBig to: #'CogMethod *'].
	startAddress := methodZone allocate: totalSize.
	startAddress = 0 ifTrue:
		[^self cCoerceSimple: InsufficientCodeSpace to: #'CogMethod *'].
	self assert: startAddress + cbEntryOffset = fullBlockEntry address.
	self assert: startAddress + cbNoSwitchEntryOffset = fullBlockNoContextSwitchEntry address.
	result := self outputInstructionsAt: startAddress + headerSize.
	self assert: startAddress + headerSize + codeSize = result.
	backEnd padIfPossibleWithStopsFrom: result to: startAddress + totalSize - mapSize - 1.
	self generateMapAt: startAddress + totalSize - 1 start: startAddress + cbNoSwitchEntryOffset.
	self flag: #TOCHECK. "It's not clear we want the same header than regular methods. 
	It could be of the same size, but maybe the cmType could be different and the selector could be ignored." 
	method := self writableMethodFor: startAddress.
	self fillInMethodHeader: method size: totalSize selector: objectMemory nilObject.
	method cpicHasMNUCaseOrCMIsFullBlock: true.	
	"This also implicitly flushes the read/write mapped dual zone to the read/execute zone."
	backEnd flushICacheFrom: startAddress to: startAddress + totalSize.
	^self cCoerceSimple: startAddress to: #'CogMethod *'! !

!Cogit methodsFor: 'generate machine code' stamp: 'eem 8/1/2021 23:34'!
generateCogMethod: selector
	"We handle jump sizing simply.  First we make a pass that asks each
	 instruction to compute its maximum size.  Then we make a pass that
	 sizes jumps based on the maxmimum sizes.  Then we make a pass
	 that fixes up jumps.  When fixing up a jump the jump is not allowed to
	 choose a smaller offset but must stick to the size set in the second pass."
	<returnTypeC: #'CogMethod *'>
	| codeSize headerSize mapSize totalSize startAddress result |
	headerSize := self sizeof: CogMethod.
	methodLabel address: methodZone freeStart.
	self computeMaximumSizes.
	methodLabel concretizeAt: methodZone freeStart.
	codeSize := self generateInstructionsAt: methodLabel address + headerSize.
	mapSize := self generateMapAt: nil start: methodLabel address + cmNoCheckEntryOffset.
	totalSize := methodZone roundUpLength: headerSize + codeSize + mapSize.
	totalSize > MaxMethodSize ifTrue:
		[^self cCoerceSimple: MethodTooBig to: #'CogMethod *'].
	startAddress := methodZone allocate: totalSize.
	startAddress = 0 ifTrue:
		[^self cCoerceSimple: InsufficientCodeSpace to: #'CogMethod *'].
	self assert: startAddress + cmEntryOffset = entry address.
	self assert: startAddress + cmNoCheckEntryOffset = noCheckEntry address.
	result := self outputInstructionsAt: startAddress + headerSize.
	self assert: startAddress + headerSize + codeSize = result.
	backEnd padIfPossibleWithStopsFrom: result to: startAddress + totalSize - mapSize - 1.
	self generateMapAt: startAddress + totalSize - 1 start: startAddress + cmNoCheckEntryOffset.
	self fillInBlockHeadersAt: startAddress.
	self fillInMethodHeader: (self writableMethodFor: startAddress)
		size: totalSize
		selector: selector.
	"This also implicitly flushes the read/write mapped dual zone to the read/execute zone."
	backEnd flushICacheFrom: startAddress to: startAddress + totalSize.
	^self cCoerceSimple: startAddress to: #'CogMethod *'! !

!Cogit methodsFor: 'initialization' stamp: 'eem 2/20/2020 19:23'!
generateEnilopmarts
	"Enilopmarts transfer control from C into machine code (backwards trampolines)."
	self cppIf: Debug
		ifTrue:
			[realCEEnterCogCodePopReceiverReg :=
				self genEnilopmartFor: ReceiverResultReg
					forCall: false
					called: 'realCEEnterCogCodePopReceiverReg'.
			 ceEnterCogCodePopReceiverReg := #enterCogCodePopReceiver.
			 realCECallCogCodePopReceiverReg :=
				self genEnilopmartFor: ReceiverResultReg
					forCall: true
					called: 'realCECallCogCodePopReceiverReg'.
			 ceCallCogCodePopReceiverReg := #callCogCodePopReceiver.
			 realCECallCogCodePopReceiverAndClassRegs :=
				self genEnilopmartFor: ReceiverResultReg
					and: ClassReg
					forCall: true
					called: 'realCECallCogCodePopReceiverAndClassRegs'.
			 ceCallCogCodePopReceiverAndClassRegs := #callCogCodePopReceiverAndClassRegs]
		ifFalse:
			[ceEnterCogCodePopReceiverReg := self genEnilopmartFor: ReceiverResultReg
														forCall: false
														called: 'ceEnterCogCodePopReceiverReg'.
			 ceCallCogCodePopReceiverReg := self genEnilopmartFor: ReceiverResultReg
													forCall: true
													called: 'ceCallCogCodePopReceiverReg'.
			 ceCallCogCodePopReceiverAndClassRegs :=
				self genEnilopmartFor: ReceiverResultReg
					and: ClassReg
					forCall: true
					called: 'ceCallCogCodePopReceiverAndClassRegs'].

	self genPrimReturnEnterCogCodeEnilopmart: false.
	cePrimReturnEnterCogCode := methodZoneBase.
	self outputInstructionsForGeneratedRuntimeAt: cePrimReturnEnterCogCode.
	self recordGeneratedRunTime: 'cePrimReturnEnterCogCode' address: cePrimReturnEnterCogCode.

	self genPrimReturnEnterCogCodeEnilopmart: true.
	cePrimReturnEnterCogCodeProfiling := methodZoneBase.
	self outputInstructionsForGeneratedRuntimeAt: cePrimReturnEnterCogCodeProfiling.
	self recordGeneratedRunTime: 'cePrimReturnEnterCogCodeProfiling' address: cePrimReturnEnterCogCodeProfiling! !

!Cogit methodsFor: 'generate machine code' stamp: 'eem 7/21/2021 09:04'!
generateInstructionsAt: eventualAbsoluteAddress
	"Size pc-dependent instructions and assign eventual addresses to all instructions.
	 Answer the size of the code.
	 Compute forward branches based on virtual address (abstract code starts at 0),
	 assuming that any branches branched over are long.
	 Compute backward branches based on actual address.
	 Reuse the fixups array to record the pc-dependent instructions that need to have
	 their code generation postponed until after the others."
	| absoluteAddress pcDependentIndex abstractInstruction fixup |
	<var: #abstractInstruction type: #'AbstractInstruction *'>
	<var: #fixup type: #'BytecodeFixup *'>
	absoluteAddress := eventualAbsoluteAddress.
	pcDependentIndex := 0.
	0 to: opcodeIndex - 1 do:
		[:i|
		self maybeBreakGeneratingInstructionWithIndex: i.
		abstractInstruction := self abstractInstructionAt: i.
		"N.B. if you want to break in resizing, break here, note the instruction index, back up to the
		 sender, restart, and step into computeMaximumSizes, breaking at this instruction's index."
		self maybeBreakGeneratingFrom: absoluteAddress to: absoluteAddress + abstractInstruction maxSize.
		abstractInstruction isPCDependent
			ifTrue:
				[abstractInstruction sizePCDependentInstructionAt: absoluteAddress.
				 fixup := self fixupAtIndex: pcDependentIndex.
				 pcDependentIndex := pcDependentIndex + 1.
				 fixup instructionIndex: i.
				 absoluteAddress := absoluteAddress + abstractInstruction machineCodeSize]
			ifFalse:
				[absoluteAddress := abstractInstruction concretizeAt: absoluteAddress]].
	0 to: pcDependentIndex - 1 do:
		[:j|
		fixup := self fixupAtIndex: j.
		self maybeBreakGeneratingInstructionWithIndex: fixup instructionIndex.
		abstractInstruction := self abstractInstructionAt: fixup instructionIndex.
		"N.B. if you want to break in resizing, break here, note the instruction index, back up to the
		 sender, restart, and step into computeMaximumSizes, breaking at this instruction's index."
		self maybeBreakGeneratingFrom: abstractInstruction address to: abstractInstruction address + abstractInstruction maxSize - 1.
		abstractInstruction concretizeAt: abstractInstruction address].
	^absoluteAddress - eventualAbsoluteAddress! !

!Cogit methodsFor: 'method map' stamp: 'eem 3/10/2017 17:58'!
generateMapAt: addressOrNull start: startAddress
	"Generate the method map at addressrNull (or compute it if addressOrNull is null).
	 Answer the length of the map in byes.  Each entry in the map is in two parts.  In the
	 least signficant bits are a displacement of how far from the start or previous entry,
	 unless it is an IsAnnotationExtension byte, in which case those bits are the extension.
	 In the most signficant bits are the type of annotation at the point reached.  A null
	 byte ends the map."
	<var: 'addressOrNull' type: #usqInt>
	<var: 'startAddress' type: #usqInt>
	| length location |
	<var: #instruction type: #'AbstractInstruction *'>
	length := 0.
	location := startAddress.
	0 to: opcodeIndex - 1 do:
		[:i| | instruction mcpc delta maxDelta mapEntry |
		instruction := self abstractInstructionAt: i.
		instruction annotation ifNotNil:
			[:annotation|
			 literalsManager assertValidAnnotation: annotation for: instruction.
			 mcpc := instruction mapEntryAddress.
			 [(delta := mcpc - location / backEnd codeGranularity) > DisplacementMask] whileTrue:
				[maxDelta := (delta min: MaxX2NDisplacement) bitClear: DisplacementMask.
				 self assert: maxDelta >> AnnotationShift <= DisplacementMask.
				 addressOrNull ifNotNil:
					[self addToMap: IsDisplacementX2N
						instruction: instruction
						byte: maxDelta >> AnnotationShift + DisplacementX2N
						at: addressOrNull - length
						for: mcpc].
				 location := location + (maxDelta * backEnd codeGranularity).
				 length := length + 1].
			 addressOrNull ifNotNil:
				[mapEntry := delta + ((annotation min: IsSendCall) << AnnotationShift).
				 self addToMap: annotation instruction: instruction byte: mapEntry at: addressOrNull - length for: mcpc].
			 location := location + (delta * backEnd codeGranularity).
			 length := length + 1.
			 annotation > IsSendCall ifTrue: "Add the necessary IsAnnotationExtension"
				[addressOrNull ifNotNil:
					[mapEntry := IsAnnotationExtension << AnnotationShift + (annotation - IsSendCall).
					 self addToMap: annotation instruction: instruction byte: mapEntry at: addressOrNull - length for: mcpc].
				 length := length + 1]]].
	addressOrNull ifNotNil:
		[self addToMap: MapEnd instruction: nil byte: MapEnd at: addressOrNull - length for: 0].
	^length + 1! !

!Cogit methodsFor: 'initialization' stamp: 'eem 10/4/2013 14:18'!
generateMissAbortTrampolines
	"Generate the run-time entries for the various method and PIC entry misses and aborts.
	 Read the class-side method trampolines for documentation on the various trampolines"

	self subclassResponsibility! !

!Cogit methodsFor: 'initialization' stamp: 'rmacnak 3/28/2015 16:25'!
generateNewspeakRuntime
	<option: #NewspeakVM>
	"Generate the non-send runtime support for Newspeak: push enclosing object.
	The dynamic frequency is so low we merely call an interpreter routine."
	ceEnclosingObjectTrampoline := self genTrampolineFor: #ceEnclosingObjectAt:
										called: 'ceEnclosingObjectTrampoline'
										arg: SendNumArgsReg
										result: ReceiverResultReg.! !

!Cogit methodsFor: 'initialization' stamp: 'rmacnak 7/13/2015 19:35'!
generateNewspeakSendTrampolines
	"Self send, dynamic super send, implicit receiver send, and outer send"
	<option: #NewspeakVM>
	0 to: NumSendTrampolines - 1 do:
		[:numArgs|
		selfSendTrampolines
			at: numArgs
			put: (self 
				genNSSendTrampolineFor: #ceSelfSend:receiver:
				numArgs: numArgs enclosingObjectCheck: false
				called: (self trampolineName: 'ceSelfSend' numArgs: numArgs))].

	0 to: NumSendTrampolines - 1 do:
		[:numArgs|
		dynamicSuperSendTrampolines
			at: numArgs
			put: (self 
				genNSSendTrampolineFor: #ceDynamicSuperSend:receiver:
				numArgs: numArgs enclosingObjectCheck: false
				called: (self trampolineName: 'ceDynamicSuperSend' numArgs: numArgs))].

	0 to: NumSendTrampolines - 1 do:
		[:numArgs|
		implicitReceiverSendTrampolines
			at: numArgs
			put: (self 
				genNSSendTrampolineFor: #ceImplicitReceiverSend:receiver:
				numArgs: numArgs enclosingObjectCheck: true
				called: (self trampolineName: 'ceImplicitReceiverSend' numArgs: numArgs))].

	0 to: NumSendTrampolines - 1 do:
		[:numArgs|
		outerSendTrampolines
			at: numArgs
			put: (self 
				genNSSendTrampolineFor: #ceOuterSend:receiver:
				numArgs: numArgs enclosingObjectCheck: true
				called: (self trampolineName: 'ceOuterSend' numArgs: numArgs))].

! !

!Cogit methodsFor: 'initialization' stamp: 'eem 7/6/2021 14:21'!
generateOpenPICPrototype
	"Generate the prototype OpenPIC to determine how much space an open PIC takes."
	| codeSize mapSize |
	"stack allocate the various collections so that they
	 are effectively garbage collected on return."
	self allocateOpcodes: 100 bytecodes: 0.
	methodLabel
		address: methodZoneBase;
		dependent: nil.
	"Need a real selector here so that the map accomodates the annotations for the selector.
	 Use self numRegArgs to generate the longest possible code sequence due to
	 genPushRegisterArgsForNumArgs:"
	self compileOpenPIC: (coInterpreter specialSelector: 0) numArgs: self numRegArgs.
	self computeMaximumSizes.
	methodLabel concretizeAt: methodZoneBase.
	codeSize := self generateInstructionsAt: methodZoneBase + (self sizeof: CogMethod).
	mapSize := self generateMapAt: nil start: methodZoneBase + cmNoCheckEntryOffset.
	openPICSize := (methodZone roundUpLength: (self sizeof: CogMethod) + codeSize) + (methodZone roundUpLength: mapSize).
	"self cCode: ''
		inSmalltalk:
			[| end |
			 end := self outputInstructionsAt: methodZoneBase + headerSize.
			 self disassembleFrom: methodZoneBase + (self sizeof: CogMethod) to: end - 1.
			 self halt]"! !

!Cogit methodsFor: 'initialization' stamp: 'eem 3/29/2020 18:39'!
generateRunTimeTrampolines
	"Generate the run-time entries at the base of the native code zone and update the base."
	
	ceSendMustBeBooleanAddFalseTrampoline := self genMustBeBooleanTrampolineFor: objectMemory falseObject
														called: 'ceSendMustBeBooleanAddFalseTrampoline'.
	ceSendMustBeBooleanAddTrueTrampoline := self genMustBeBooleanTrampolineFor: objectMemory trueObject
														called: 'ceSendMustBeBooleanAddTrueTrampoline'.
	ceNonLocalReturnTrampoline := self genNonLocalReturnTrampoline.
	ceCheckForInterruptTrampoline := self genCheckForInterruptsTrampoline.
	"Neither of the context inst var access trampolines save registers.  Their operation could cause
	 arbitrary update of stack frames, so the assumption is that callers flush the stack before calling
	 the context inst var access trampolines, and that everything except the result is dead afterwards."
	ceFetchContextInstVarTrampoline := self genTrampolineFor: #ceContext:instVar:
											called: 'ceFetchContextInstVarTrampoline'
											arg: ReceiverResultReg
											arg: SendNumArgsReg
											result: SendNumArgsReg.
	ceStoreContextInstVarTrampoline := self genTrampolineFor: #ceContext:instVar:value:
											called: 'ceStoreContextInstVarTrampoline'
											arg: ReceiverResultReg
											arg: SendNumArgsReg
											arg: ClassReg
											result: ReceiverResultReg. "to keep ReceiverResultReg live.".
	ceCannotResumeTrampoline := self genTrampolineFor: #ceCannotResume
											called: 'ceCannotResumeTrampoline'.
	"ceInvokeInterpreter is an optimization and a work-around. Historically we used setjmp/longjmp to reenter the
	 interpreter at the current C stack base.  The C stack base is set at start-up and on each callback enter and
	 callback return. The interpreter must be invoked whenever a non-machine-code method must be run.  That might
	 be when invoking an interpreter method from one of the send linking routines (ceSend:...), or on continuing from
	 an evaluation primitive such as primitiveExecuteMethod.  The problem here is that such primitives could have
	 been invoked by the interpreter or by machine code.  So some form of non-local jump is required. But at least as
	 early as MSVC Community 2017, the Microshaft longjmp performs stack unwinding which gets hoplessly confused
	 (bless its little heart) by any stack switch between machine code and C stack, and raises a spurious
		Stack cookie instrumentation code detected a stack-based buffer overrun
	 error from the bowels of gs_report.c _GSHandlerCheck.
	 Since the CoInterpreter maintains the base of the C stack in CFramePointer & CStackPointer, it is straight-forward
	 for us to simply call interpret after doing the switch to the C stack, avoiding the stack unwind issue altogether."
	ceInvokeInterpret := self genInvokeInterpretTrampoline.
	"These two are unusual; they are reached by return instructions."
	ceReturnToInterpreterTrampoline := self genReturnToInterpreterTrampoline.
	ceBaseFrameReturnTrampoline := self genReturnTrampolineFor: #ceBaseFrameReturn:
											called: 'ceBaseFrameReturnTrampoline'
											arg: ReceiverResultReg.
	LowcodeVM ifTrue:
		[ceFFICalloutTrampoline := self genFFICalloutTrampoline.
		 ceMallocTrampoline := self genTrampolineFor: #ceMalloc:
									called: 'ceMallocTrampoline'
									arg: ReceiverResultReg
									result: TempReg.
		 ceFreeTrampoline := self genTrampolineFor: #ceFree:
									called: 'ceFreeTrampoline'
									arg: ReceiverResultReg]! !

!Cogit methodsFor: 'initialization' stamp: 'eem 1/11/2018 18:57'!
generateSendTrampolines
	0 to: NumSendTrampolines - 1 do:
		[:numArgs|
		ordinarySendTrampolines
			at: numArgs
			put: (self genTrampolineFor: #ceSend:super:to:numArgs:
					  called: (self trampolineName: 'ceSend' numArgs: numArgs)
					  arg: ClassReg
					  arg: (self trampolineArgConstant: false)
					  arg: ReceiverResultReg
					  arg: (self numArgsOrSendNumArgsReg: numArgs))].

	"Generate these in the middle so they are within [firstSend, lastSend]."
	NewspeakVM ifTrue:
		[self generateNewspeakSendTrampolines].
	BytecodeSetHasDirectedSuperSend ifTrue:
		[0 to: NumSendTrampolines - 1 do:
			[:numArgs|
			directedSuperSendTrampolines
				at: numArgs
				put: (self genTrampolineFor: #ceSend:above:to:numArgs:
						  called: (self trampolineName: 'ceDirectedSuperSend' numArgs: numArgs)
						  arg: ClassReg
						  arg: TempReg
						  arg: ReceiverResultReg
						  arg: (self numArgsOrSendNumArgsReg: numArgs)).
			directedSuperBindingSendTrampolines
				at: numArgs
				put: (self genTrampolineFor: #ceSend:aboveClassBinding:to:numArgs:
						  called: (self trampolineName: 'ceDirectedSuperBindingSend' numArgs: numArgs)
						  arg: ClassReg
						  arg: TempReg
						  arg: ReceiverResultReg
						  arg: (self numArgsOrSendNumArgsReg: numArgs))]].

	0 to: NumSendTrampolines - 1 do:
		[:numArgs|
		superSendTrampolines
			at: numArgs
			put: (self genTrampolineFor: #ceSend:super:to:numArgs:
					  called: (self trampolineName: 'ceSuperSend' numArgs: numArgs)
					  arg: ClassReg
					  arg: (self trampolineArgConstant: true)
					  arg: ReceiverResultReg
					  arg: (self numArgsOrSendNumArgsReg: numArgs))].
	firstSend := ordinarySendTrampolines at: 0.
	lastSend := superSendTrampolines at: NumSendTrampolines - 1! !

!Cogit methodsFor: 'initialization' stamp: 'eem 7/8/2021 11:27'!
generateStackPointerCapture
	"Generate a routine ceCaptureCStackPointers that will capture the C stack pointer,
	 and, if it is in use, the C frame pointer.  These are used in trampolines to call
	 run-time routines in the interpreter from machine-code."

	| oldMethodZoneBase oldTrampolineTableIndex |
	cFramePointerInUse := false. "For the benefit of the following assert, assume the minimum at first."
	self assertCStackWellAligned.
	oldMethodZoneBase := methodZoneBase.
	oldTrampolineTableIndex := trampolineTableIndex.
	self generateCaptureCStackPointers: true.
	self perform: #ceCaptureCStackPointers.
	(cFramePointerInUse := coInterpreter checkIfCFramePointerInUse) ifFalse:
		[methodZoneBase := oldMethodZoneBase.
		 trampolineTableIndex := oldTrampolineTableIndex.
		 self generateCaptureCStackPointers: false].
	self ensureWritableCodeZone.
	self assertCStackWellAligned! !

!Cogit methodsFor: 'initialization' stamp: 'eem 2/3/2020 19:51'!
generateTrampolines
	"Generate the run-time entries and exits at the base of the native code zone and update the base.
	 Read the class-side method trampolines for documentation on the various trampolines"
	| methodZoneStart |
	methodZoneStart := methodZoneBase.
	methodLabel address: methodZoneStart.
	self allocateOpcodes: 80 bytecodes: 0.
	self setHasYoungReferent: false.
	objectRepresentation maybeGenerateSelectorIndexDereferenceRoutine.
	self generateSendTrampolines.
	self generateMissAbortTrampolines.
	objectRepresentation generateObjectRepresentationTrampolines.
	self generateRunTimeTrampolines.
	NewspeakVM ifTrue: [self generateNewspeakRuntime].
	SistaVM ifTrue: [self generateSistaRuntime].
	self generateEnilopmarts.
	self generateTracingTrampolines.

	"finish up"
	self recordGeneratedRunTime: 'methodZoneBase' address: methodZoneBase! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/11/2020 18:36'!
generateVMOwnerLockFunctions
	<inline: true>
	self cppIf: COGMTVM ifTrue:
		[| startAddress |
		self allocateOpcodes: backEnd numLowLevelLockOpcodes bytecodes: 0.
		self zeroOpcodeIndex.
		startAddress := methodZoneBase.
		backEnd generateLowLevelTryLock: coInterpreter vmOwnerAddress.
		self outputInstructionsForGeneratedRuntimeAt: startAddress.
		self recordGeneratedRunTime: 'ceTryLockVMOwner' address: startAddress.
		ceTryLockVMOwner := self cCoerceSimple: startAddress to: #'usqIntptr_t (*)(usqIntptr_t)']! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/17/2010 21:26'!
generatorAt: index
	<cmacro: '(index) (&generatorTable[index])'>
	<returnTypeC: #'BytecodeDescriptor *'>
	^generatorTable at: index! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 1/21/2018 16:58'!
generatorForPC: pc
	<returnTypeC: #'BytecodeDescriptor *'>
	<inline: true>
	^self generatorAt: bytecodeSetOffset + (objectMemory fetchByte: pc ofObject: methodObj)! !

!Cogit methodsFor: 'accessing' stamp: 'eem 10/8/2014 09:25'!
generatorTable
	<doNotGenerate>
	^generatorTable! !

!Cogit methodsFor: 'generate machine code - dual mapped zone support' stamp: 'eem 2/25/2020 19:09'!
getCodeToDataDelta
	"If non-zero this is the delta between the read/execute method zone and the
	 read/write mapping of the method zone.  On operating systems where it is
	 entirely disallowed to execute code in a writable region this split is necessary
	 to be able to modify code.  In this regime all writes must be made to the
	 read/write mapped zone."
	<cmacro: '() codeToDataDelta'>
	^codeToDataDelta! !

!Cogit methodsFor: 'method map' stamp: 'eem 12/12/2015 14:11'!
getIsObjectReference
	<cmacro>
	^IsObjectReference! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 3/1/2013 16:32'!
getLiteral: litIndex
	maxLitIndex < litIndex ifTrue:
		[maxLitIndex := litIndex].
	^coInterpreter literal: litIndex ofMethod: methodObj! !

!Cogit methodsFor: 'accessing' stamp: 'eem 01/10/2016 10:01'!
getOpcodeIndex
	"Access for the literal manager."
	^opcodeIndex! !

!Cogit methodsFor: 'accessing' stamp: 'eem 3/18/2017 21:52'!
getPrimitiveIndex
	"Access for the object representation primitive generation routines."
	^primitiveIndex! !

!Cogit methodsFor: 'translation support' stamp: 'eem 10/22/2010 10:49'!
halt
	<cmacro: '() warning("halt")'>
	Halt signal! !

!Cogit methodsFor: 'translation support' stamp: 'eem 9/2/2011 17:10'!
halt: aString
	<cmacro: '(msg) warning("halt: " msg)'>
	Halt new signal: aString! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 7/8/2021 14:37'!
handleABICallOrJumpSimulationTrap: aProcessorSimulationTrap evaluable: aMessage
	<doNotGenerate>

	self assert: aProcessorSimulationTrap type = #call.
	(shortCutTrampolineBlocks at: aMessage selector ifAbsent: nil) ifNotNil:
		[:block| ^self shortcutTrampoline: aProcessorSimulationTrap to: block].

	"more conventional abi call..."
	processor
		simulateLeafCallOf: aProcessorSimulationTrap address
		nextpc: aProcessorSimulationTrap nextpc
		memory: coInterpreter memory.
	self recordInstruction: {'(simulated call of '. aProcessorSimulationTrap address. '/'. aMessage selector. ')'}.
	clickConfirm ifTrue:
 		[(self confirm: 'skip run-time call?') ifFalse:
			[clickConfirm := false. self halt]].
	aMessage valueWithArguments: (processor
										postCallArgumentsNumArgs: aMessage numArgs
										in: coInterpreter memory).
	self recordInstruction: {'(simulated return to '. processor retpcIn: coInterpreter memory. ')'}.
	processor
		smashABICallerSavedRegistersWithValuesFrom: 16r80000000 by: objectMemory wordSize;
		simulateLeafReturnIn: coInterpreter memory! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 8/20/2021 23:30'!
handleCallOrJumpSimulationTrap: aProcessorSimulationTrap
	<doNotGenerate>
	| evaluable function memory result savedFramePointer savedStackPointer savedArgumentCount retpc invalidStackPointersExpected index |
	"Execution of a single instruction must be within the processorLock critical section to ensure
	 simulation traps are executed atomically.  However, at this point control is leaving machine
	 code and entering the run-time and hence the lock must be released."
	processorLock primitiveExitCriticalSection.
	"This is a hack fix before we revise the simulators.  When a jump call is made, the next
	 pc is effectively the return address on the stack, not the instruction following the jump."
	aProcessorSimulationTrap type == #jump ifTrue:
		[processor hackFixNextPCOfJumpFor: aProcessorSimulationTrap using: objectMemory].

	evaluable := simulatedTrampolines
					at: aProcessorSimulationTrap address
					ifAbsent: [self errorProcessingSimulationTrap: aProcessorSimulationTrap
								in: simulatedTrampolines].
	function := evaluable isBlock
					ifTrue: ['aBlock; probably some plugin primitive']
					ifFalse:
						[evaluable receiver == backEnd ifTrue: "this is for invoking ARMv5 floating-point intrinsics, and for the short-cut tracing trampolines"
							[^self handleABICallOrJumpSimulationTrap: aProcessorSimulationTrap evaluable: evaluable].
						 evaluable selector].
	memory := coInterpreter memory.
	function == #interpret ifTrue: "i.e. we're here via ceInvokeInterpret/ceReturnToInterpreterTrampoline and should discard all state back to enterSmalltalkExecutiveImplementation"
		[self recordInstruction: {'(simulated jump call of '. aProcessorSimulationTrap address. '/'. function. ')'}.
		 "self halt: evaluable selector."
	   	 clickConfirm ifTrue:
		 	[(self confirm: 'skip jump to interpret?') ifFalse:
				[clickConfirm := false. self halt]].
		 processor simulateJumpCallOf: aProcessorSimulationTrap address memory: memory.
		 coInterpreter reenterInterpreter.
		 "NOTREACHED"
		 self halt].
	(function == #ceBaseFrameReturn:
	or: [function == #ceTakeProfileSample:
	or: [function == #primitiveObjectAtPut]])
		ifTrue: [invalidStackPointersExpected := true]
		ifFalse:
			[invalidStackPointersExpected := false.
			 evaluable isBlock
				ifTrue: "external primitives..."
					["The only acceptable exception to the rule are fast C primitive calls..."
					 (methodZone cogMethodContaining: (self mostLikelyPrimInvocationPC: processor pc or: (processor leafRetpcIn: memory)))
						ifNil: [self assertf: 'call to block evaluable from non-external method']
						ifNotNil: [:cogMethod|
								self assert: (self cogMethodHasExternalPrim: cogMethod).
								(coInterpreter hasFastCLinkage: cogMethod methodObject)
									ifTrue: [invalidStackPointersExpected := true. coInterpreter nilLocalFP]
									ifFalse: [coInterpreter assertValidExternalStackPointers]]]
				ifFalse:
					[coInterpreter assertValidExternalStackPointers]].
	processor
		simulateCallOf: aProcessorSimulationTrap address
		nextpc: aProcessorSimulationTrap nextpc
		memory: memory.
	retpc := processor retpcIn: memory.
	self recordInstruction: {'(simulated call of '. aProcessorSimulationTrap address. '/'. function. ')'}.
	savedFramePointer := coInterpreter framePointer.
	savedStackPointer := coInterpreter stackPointer.
	savedArgumentCount := coInterpreter argumentCount.
	result := ["self halt: evaluable selector."
		   	   clickConfirm ifTrue:
			 	[(self confirm: 'skip run-time call?') ifFalse:
					[clickConfirm := false. self halt]].
			   evaluable valueWithArguments: (processor
												postCallArgumentsNumArgs: evaluable numArgs
												in: memory)]
				on: ReenterMachineCode
				do: [:ex| ex return: #continueNoReturn].
			
	invalidStackPointersExpected ifFalse:
		[coInterpreter assertValidExternalStackPointers].
	"Verify the stack layout assumption compileInterpreterPrimitive: makes, provided we've
	 not called something that has built a frame, such as closure value or evaluate method, or
	 switched frames, such as primitiveSignal, primitiveWait, primitiveResume, primitiveSuspend et al."
	(function beginsWith: 'primitive') ifTrue:
		[coInterpreter primFailCode = 0
			ifTrue: [(CogVMSimulator stackAlteringPrimitives includes: function) ifFalse:
						["This is a rare case (e.g. in Scorch where a married context's sender is set to nil on trapTrpped and hence the stack layout is altered."
						 (function == #primitiveSlotAtPut and: [objectMemory isContext: (coInterpreter frameReceiver: coInterpreter framePointer)]) ifFalse:
							[self assert: savedFramePointer = coInterpreter framePointer.
							 self assert: savedStackPointer + (savedArgumentCount * objectMemory wordSize)
									= coInterpreter stackPointer]]]
			ifFalse:
				[self assert: savedFramePointer = coInterpreter framePointer.
				 self assert: savedStackPointer = coInterpreter stackPointer]].
	result ~~ #continueNoReturn ifTrue:
		[self recordInstruction: {'(simulated return to '. processor retpcIn: memory. ')'}.
		 processor simulateReturnIn: memory.
		 self assert: processor pc = retpc.
		 processor smashCallerSavedRegistersWithValuesFrom: 16r80000000 by: objectMemory wordSize].
	self assert: (result isInteger "an oop result"
			or: [result == coInterpreter
			or: [result == objectMemory
			or: [(index := #(nil true false continueNoReturn) indexOf: result) > 0
				and: [result := #(0 1 0 16rF00BA4) at: index. true]]]]).
	processor cResultRegister: (result
								ifNil: [0]
								ifNotNil: [result isInteger
											ifTrue: [result]
											ifFalse: [16rF00BA222]])! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 1/7/2021 13:05'!
handleCompareAndSwapSimulationTrap: aCompareAndSwapSimulationTrap
	| variableValue |
	variableValue := (simulatedVariableGetters
						at: aCompareAndSwapSimulationTrap address
						ifAbsent: [self errorProcessingSimulationTrap: aCompareAndSwapSimulationTrap
									in: simulatedVariableGetters])
							value asInteger.
	variableValue = aCompareAndSwapSimulationTrap expectedValue
		ifTrue:
			[(simulatedVariableSetters
				at: aCompareAndSwapSimulationTrap address
				ifAbsent: [self errorProcessingSimulationTrap: aCompareAndSwapSimulationTrap
							in: simulatedVariableSetters]) value: aCompareAndSwapSimulationTrap storedValue.
			processor
				setFlagsForCompareAndSwap: true;
				perform: aCompareAndSwapSimulationTrap registerAccessor
				with: (processor convertIntegerToInternal: variableValue)]
		ifFalse:
			[processor
				setFlagsForCompareAndSwap: false;
				perform: aCompareAndSwapSimulationTrap failedComparisonRegisterAccessor
				with: (processor convertIntegerToInternal: variableValue)].
	processor pc: aCompareAndSwapSimulationTrap nextpc.
	aCompareAndSwapSimulationTrap resume: processor! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 10/28/2020 22:59'!
handleReadSimulationTrap: aProcessorSimulationTrap
	<doNotGenerate>
	| variableValue accessor |
	variableValue := (simulatedVariableGetters
						at: aProcessorSimulationTrap address
						ifAbsent: [self errorProcessingSimulationTrap: aProcessorSimulationTrap
									in: simulatedVariableGetters])
							value asInteger.
	accessor := aProcessorSimulationTrap registerAccessor.
	processor
		perform: accessor
		with: (processor convertIntegerToInternal: variableValue).
	accessor ~~ #pc: ifTrue:
		[processor pc: aProcessorSimulationTrap nextpc.
		 "In an enilopmart stackPointer is assigned to sp before framePointer.
		  In a trampoline fp and sp are written to the interpreter variables immediately before
		  assigning sp with CStackPointer and immediately there-after fp with CFramePointer.
		  So set processorFrameValid appropriately when assigning fp.  This is for CogHeadFrameInspector"
		 (processor accessorIsFramePointerSetter: accessor) ifTrue:
			[processorFrameValid := aProcessorSimulationTrap address ~= (simulatedAddresses at: #getCFramePointer)]].
	aProcessorSimulationTrap resume: processor! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 10/28/2020 22:38'!
handleReturnSimulationTrap: aProcessorSimulationTrap 
	<doNotGenerate>
	| retpc |
	retpc := processor leafRetpcIn: coInterpreter memory.
	processor simulateLeafReturnIn: coInterpreter memory.
	self recordInstruction: {'(simulated return to '. retpc. ')'}! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 10/28/2020 23:07'!
handleWriteSimulationTrap: aProcessorSimulationTrap 
	<doNotGenerate>
	| variableValue |
	(self addressIsInCodeZone: aProcessorSimulationTrap address) ifTrue:
		[self error: 'attempt to write to code space'].
	variableValue := processor perform: aProcessorSimulationTrap registerAccessor.
	(simulatedVariableSetters
			at: aProcessorSimulationTrap address
			ifAbsent: [self errorProcessingSimulationTrap: aProcessorSimulationTrap
						in: simulatedVariableSetters])
		value: variableValue.
	processor pc: aProcessorSimulationTrap nextpc.
	^aProcessorSimulationTrap resume: processor! !

!Cogit methodsFor: 'accessing' stamp: 'cb 9/11/2019 07:41'!
inBlock
	<doNotGenerate>
	^inBlock! !

!Cogit methodsFor: 'compaction' stamp: 'eem 1/29/2020 09:13'!
incrementUsageOfTargetIfLinkedSend: annotation mcpc: mcpc ignored: superfluity
	<var: #mcpc type: #'char *'>
	<var: #nsSendCache type: #'NSSendCache *'>
	| entryPoint |

	NewspeakVM ifTrue:
		[annotation = IsNSSendCall ifTrue:
			[| nsSendCache |
			 nsSendCache := self nsSendCacheFromReturnAddress: mcpc asInteger.
			 nsSendCache classTag ~= objectRepresentation illegalClassTag ifTrue: "send is linked"
				[ | targetMethod |
				entryPoint := nsSendCache target.
				targetMethod := self writableMethodFor: entryPoint - cmNoCheckEntryOffset.
				self assertValidDualZoneWriteAddress: targetMethod.
				targetMethod cmUsageCount < (CMMaxUsageCount // 2) ifTrue:
					[targetMethod cmUsageCount: targetMethod cmUsageCount + 1]]]].

	(self isPureSendAnnotation: annotation) ifTrue:
		[self assert: annotation ~= IsNSSendCall.
		 entryPoint := backEnd callTargetFromReturnAddress: mcpc asInteger.
		 entryPoint > methodZoneBase ifTrue: "It's a linked send."
			[self targetMethodAndSendTableFor: entryPoint annotation: annotation into:
				[:targetMethod :sendTable|
				 targetMethod cmUsageCount < (CMMaxUsageCount // 2) ifTrue:
					[(self writableMethodFor: targetMethod) cmUsageCount: targetMethod cmUsageCount + 1]]]].

	^0 "keep scanning"! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 1/29/2020 14:32'!
indexForSelector: selector in: cogMethod
	"Answer the value to put in an inline-cache that is being loaded with the selector.
	 Usually this is simply the selector, but in 64-bits the cache is only 32-bits wide
	 and so the cache is loaded with the index of the selector."
	<var: #cogMethod type: #'CogMethod *'>
	<inline: false>
	| methodOop |
	"First search the special selectors; there are only 32 of them so this shouldn't take too long.
	 We could short-circuit this by keeping a hint bit in the target method, or by maintaining the
	 maximum range of selector oops in specialSelectors since they're likely to cluster."
	0 to: NumSpecialSelectors - 1 do:
		[:i|
		selector = (coInterpreter specialSelector: i) ifTrue:
			[^-1 - i]].
	methodOop := cogMethod methodObject.
	"Then search the method's literal frame... open code fetchPointer:ofObject: for speed..."
	LiteralStart to: (objectMemory literalCountOfMethodHeader: cogMethod methodHeader) do:
		[:i|
		(objectMemory longAt: i * objectMemory bytesPerOop + objectMemory baseHeaderSize + methodOop) = selector ifTrue:
			[self assert: selector = (coInterpreter literal: i - 1 ofMethod: methodOop).
			 ^i - 1]].

	self error: 'could not find selector in method when unlinking send site'.
	^0! !

!Cogit methodsFor: 'generate machine code' stamp: 'eem 4/14/2015 18:49'!
initialClosedPICUsageCount
	"Answer a usage count that reflects likely long-term usage."
	^CMMaxUsageCount // 2! !

!Cogit methodsFor: 'generate machine code' stamp: 'eem 12/10/2015 08:32'!
initialMethodUsageCount
	"Answer a usage count that reflects likely long-term usage.
	 Answer 1 for non-primitives or quick primitives (inst var accessors),
	 2 for methods with interpreter primitives, and 3 for compiled primitives."
	(primitiveIndex = 1
	 or: [coInterpreter isQuickPrimitiveIndex: primitiveIndex]) ifTrue:
		[^1].
	self primitiveGeneratorOrNil ifNil:
		[^2].
	^3! !

!Cogit methodsFor: 'generate machine code' stamp: 'eem 4/14/2015 18:49'!
initialOpenPICUsageCount
	"Answer a usage count that reflects likely long-term usage."
	^CMMaxUsageCount - 1! !

!Cogit methodsFor: 'initialization' stamp: 'eem 7/5/2021 11:23'!
initialize
	"Here we can initialize the variables C initialized to zero.  #initialize methods do /not/ get translated."
	| wordSize |
	initialPC := 0.
	codeZoneIsExecutableNotWritable := processorFrameValid := false.
	codeToDataDelta := 0.
	wordSize := self class objectMemoryClass wordSize.
	cogMethodSurrogateClass := NewspeakVM
									ifTrue:
										[wordSize = 4
											ifTrue: [CogNewspeakMethodSurrogate32]
											ifFalse: [CogNewspeakMethodSurrogate64]]
									ifFalse:
										[wordSize = 4
											ifTrue: [CogMethodSurrogate32]
											ifFalse: [CogMethodSurrogate64]].
	cogBlockMethodSurrogateClass := wordSize = 4
											ifTrue: [CogBlockMethodSurrogate32]
											ifFalse: [CogBlockMethodSurrogate64].
	nsSendCacheSurrogateClass := wordSize = 4
											ifTrue: [NSSendCacheSurrogate32]
											ifFalse: [NSSendCacheSurrogate64]! !

!Cogit methodsFor: 'initialization' stamp: 'eem 7/5/2021 17:48'!
initializeBackend
	methodLabel machineCodeSize: 0.
	methodLabel opcode: Label.
	methodLabel operands at: 0 put: 0.
	methodLabel operands at: 1 put: 0. "label offset"
	backEnd hasVarBaseRegister ifTrue:
		[self assert: ((self registerMaskFor: VarBaseReg) noMask: CallerSavedRegisterMask).
		 varBaseAddress := self computeGoodVarBaseAddress.
		 self assert: coInterpreter stackLimitAddress >= varBaseAddress.
		 self assert: coInterpreter cStackPointerAddress >= varBaseAddress.
		 self assert: coInterpreter cFramePointerAddress >= varBaseAddress.
		 self assert: coInterpreter cReturnAddressAddress >= varBaseAddress.
		 self assert: coInterpreter nextProfileTickAddress >= varBaseAddress].
	literalsManager allocateLiterals: 4; resetLiterals.
	"If the platform requires flipping between executable and writable
	 states, the system expects to start in the writable state."
	backEnd needsCodeZoneExecuteWriteSwitch ifTrue:
		[backEnd makeCodeZoneWritable]! !

!Cogit methodsFor: 'initialization' stamp: 'eem 8/6/2021 12:32'!
initializeCodeZoneFrom: startAddress upTo: endAddress
	<api>
	self initializeBackend.
	self sqMakeMemoryExecutableFrom: startAddress
		To: endAddress
		CodeToDataDelta: (self cppIf: #DUAL_MAPPED_CODE_ZONE
								ifTrue: [self addressOf: codeToDataDelta put: [:v| codeToDataDelta := v]]
								ifFalse: [nil]).
	codeBase := methodZoneBase := startAddress.
	backEnd stopsFrom: startAddress to: endAddress - 1.
	self cCode: '' inSmalltalk:
		[self initializeProcessor.
		 backEnd stopsFrom: 0 to: guardPageSize - 1.
		 backEnd has64BitPerformanceCounter ifTrue:
			[self initializeSimulationIOHighResClockForProfiling]].
	minValidCallAddress := (codeBase min: coInterpreter interpretAddress)
								min: coInterpreter primitiveFailAddress.
	methodZone manageFrom: methodZoneBase to: endAddress.
	self assertValidDualZone.
	backEnd detectFeatures.
	self maybeGenerateCacheFlush.
	self generateVMOwnerLockFunctions.
	self genGetLeafCallStackPointers.
	self generateStackPointerCapture.
	self generateTrampolines.
	self computeEntryOffsets.
	self computeFullBlockEntryOffsets.
	self generateClosedPICPrototype.
	self alignMethodZoneBase.

	"None of the above is executed beyond ceCheckFeatures, so a bulk flush now is the leanest thing to do."
	backEnd flushICacheFrom: startAddress to: methodZoneBase asUnsignedInteger.
	self maybeFlushWritableZoneFrom: startAddress to: methodZoneBase asUnsignedInteger.
	"Repeat so that now the methodZone ignores the generated run-time."
	methodZone manageFrom: methodZoneBase to: endAddress.
	"N.B. this is assumed to be the last thing done in initialization; see Cogit>>initialized.
	 This is done only to compute openPICSize; the generated code is discarded."
	self generateOpenPICPrototype! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 3/6/2018 17:51'!
initializeFixupAt: targetPC
	"Make sure there's a flagged fixup at the targetPC in fixups.
	 Initially a fixup's target is just a flag.  Later on it is replaced with a proper instruction."
	<returnTypeC: #'BytecodeFixup *'>
	<inline: true>
	(self fixupAt: targetPC) becomeFixup! !

!Cogit methodsFor: 'newspeak support' stamp: 'eem 4/4/2015 18:26'!
initializeNSSendCache: cacheAddress selector: selector numArgs: numArgs depth: depth
	<inline: true>
	| nsSendCache |
	nsSendCache := self cCoerceSimple: cacheAddress to: #'NSSendCache *'.
	nsSendCache selector: selector.
	nsSendCache numArgs: numArgs.
	nsSendCache depth: depth.
	nsSendCache classTag: objectRepresentation illegalClassTag.
! !

!Cogit methodsFor: 'initialization' stamp: 'eem 1/3/2021 13:26'!
initializeProcessor
	"Initialize the simulation processor, arranging that its initial stack is somewhere on the rump C stack."
	<doNotGenerate>
	guardPageSize := self class guardPageSize.
	lastNInstructions := OrderedCollection new.
	processor initializeStackFor: self.
	self initializeProcessorStack: coInterpreter rumpCStackAddress.
	coInterpreter setCFramePointer: processor fp setCStackPointer: processor sp.
	(InitializationOptions at: #UseMultiProcessor ifAbsent: [false]) ifTrue:
		[processor := MultiProcessor for: processor coInterpreter: coInterpreter].
	processorLock := Mutex new! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/22/2020 20:39'!
initializeProcessorStack: rumpCStackAddress
	"Initialize the simulation processor's stack pointers, arranging that they are somewhere on the rump C stack."
	<doNotGenerate>
	| stackPad cFramePointer cStackPointer |
	stackPad := 64 max: cStackAlignment.
	cStackPointer := rumpCStackAddress - stackPad + expectedSPAlignment.
	cFramePointer := rumpCStackAddress - stackPad + cStackAlignment + expectedFPAlignment.
	self assert: cStackPointer \\ cStackAlignment = expectedSPAlignment.
	self assert: cFramePointer \\ cStackAlignment = expectedFPAlignment.
	processor setFramePointer: cFramePointer stackPointer: cStackPointer! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 8/6/2021 14:26'!
initializeSimulationIOHighResClockForProfiling
	"For simulating timing, hack in ioHighResClock in the guard page"
	self allocateOpcodes: 4 bytecodes: 0.
	objectRepresentation wordSize = 8
		ifTrue:
			[self MovePerfCnt64R: ABIResultReg L: ABICalleeSavedRegisterMask]
		ifFalse:
			[self MovePerfCnt64R: ABIResultReg R: ABIResultRegHigh L: ABICalleeSavedRegisterMask].
	self RetN: 0.
	ioHighResClock := self outputInstructionsForGeneratedRuntimeAt: codeBase - (backEnd machineCodeBytes * 2).
	self resetMethodZoneBase: codeBase.
	coInterpreter ifTestProfilingAdvanceNextProfileTick

	"self disassembleFrom: ioHighResClock to: self class guardPageSize"! !

!Cogit methodsFor: 'testing' stamp: 'eem 10/7/2020 18:07'!
initialized
	"Answerr if the code generator has been initialized (all trampolines and prtotypes
	 have been generated and the code generator is ready to start compiling methods.
	 Since the last thing is the generation of the open PIC prototype, its size is the marker."
	<inline: #always>
	^self cCode: [openPICSize] inSmalltalk: [openPICSize isInteger]! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 2/20/2016 18:54'!
inlineCacheTagsAreIndexes
	"The Cogit always generates 32-bit inline caches.  This implies that in 64-bits there is no room
	 in an unlinked inline cache for a selector oop.  Instead it contains a signed 32-bit index, positive
	 for selectors in a method's literal frame and negative for selectors in the specialSelectorsOop.
	 And it implies that linked inline cache entries contain class indices, not class oops."
	<inline: true>
	^objectMemory wordSize = 8! !

!Cogit methodsFor: 'tests-method map' stamp: 'eem 1/8/2011 11:48'!
innermostSubMethodFor: bcpc in: subMethods startingAt: index
	<doNotGenerate>
	| subMethod |
	^index <= subMethods size ifTrue:
		[subMethod := subMethods at: index.
		 (bcpc between: subMethod startpc and: subMethod endPC)
			ifTrue:
				[(self innermostSubMethodFor: bcpc in: subMethods startingAt: index + 1)
					ifNil: [subMethod]
					ifNotNil: [:innerSubMethod| innerSubMethod]]
			ifFalse:
				[self innermostSubMethodFor: bcpc in: subMethods startingAt: index + 1]]! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 7/2/2014 21:50'!
interpretOffset
	<api>
	<cmacro: '() missOffset'>
	^missOffset! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/18/2010 07:13'!
inverseBranchFor: opcode
	opcode caseOf: {
		[JumpLongZero]		->	[^JumpLongNonZero].
		[JumpLongNonZero]	->	[^JumpLongZero].
		[JumpZero]				->	[^JumpNonZero].
		[JumpNonZero]			->	[^JumpZero].
		[JumpNegative]			->	[^JumpNonNegative].
		[JumpNonNegative]		->	[^JumpNegative].
		[JumpOverflow]			->	[^JumpNoOverflow].
		[JumpNoOverflow]		->	[^JumpOverflow].
		[JumpCarry]			->	[^JumpNoCarry].
		[JumpNoCarry]			->	[^JumpCarry].
		[JumpLess]				->	[^JumpGreaterOrEqual].
		[JumpGreaterOrEqual]	->	[^JumpLess].
		[JumpGreater]			->	[^JumpLessOrEqual].
		[JumpLessOrEqual]		->	[^JumpGreater].
		[JumpBelow]			->	[^JumpAboveOrEqual].
		[JumpAboveOrEqual]	->	[^JumpBelow].
		[JumpAbove]			->	[^JumpBelowOrEqual].
		[JumpBelowOrEqual]	->	[^JumpAbove] }.
	self error: 'invalid opcode for inverse'.
	^0! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 8/6/2021 14:46'!
ioHighResClock
	<doNotGenerate>
	| end memory savedsp |
	end := self class guardPageSize.
	memory := coInterpreter memory.
	(coInterpreter isOnRumpCStack: (savedsp := processor sp)) ifFalse:
		[processor sp: (coInterpreter heapBase bitAnd: (cStackAlignment - 1) bitInvert)].
	processor simulateLeafCallOf: ioHighResClock nextpc: end memory: memory.
	[processor pc ~= end] whileTrue:
		[processor
			singleStepIn: memory
			minimumAddress: ioHighResClock
			readOnlyBelow: end].
	processor sp: savedsp.
	backEnd class wordSize = 8 ifTrue:
		[^processor cResultRegister].
	^(processor cResultRegisterHigh bitShift: 32) + processor cResultRegister

	"self disassembleFrom: ioHighResClock to: self class guardPageSize"! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 1/11/2013 18:02'!
isAbsentReceiverSendAt: mcpc in: cogHomeMethod
	| prev this |
	self mapFor: cogHomeMethod
		do: [:a :m|
			m < mcpc
				ifTrue: [prev := a]
				ifFalse: [m = mcpc ifTrue: [this := a]].
			false].
	^this = IsSendCall and: [prev = IsNSSendCall]! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/15/2012 15:47'!
isBackwardBranch: descriptor at: pc exts: nExts in: aMethodObj
	"Answer if the branch bytecode with the given descriptor is a backward branch."
	<inline: true>
	<var: #descriptor type: #'BytecodeDescriptor *'>
	self assert: descriptor spanFunction notNil.
	^(self
		perform: descriptor spanFunction
		with: descriptor
		with: pc
		with: nExts
		with: aMethodObj) < 0! !

!Cogit methodsFor: 'bytecode generator support' stamp: 'eem 4/6/2015 12:31'!
isDirectedSuper: descriptor extA: exta extB: extb
	"235	(1)	11101011	iiiiijjj	ExtendB < 64
										ifTrue: [Send To Superclass
													Literal Selector #iiiii (+ Extend A * 32)
													with jjj (+ Extend B * 8) Arguments]
										ifFalse: [Send To Superclass of Stacked Class
													Literal Selector #iiiii (+ Extend A * 32)
													with jjj (+ (Extend B bitAnd: 63) * 8) Arguments]"
	<inline: true>
	^descriptor notNil
	  and: [descriptor generator == #genExtSendSuperBytecode
	  and: [extb >= 64]]! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 1/24/2017 13:34'!
isNonLocalReturnPC: retpc
	<doNotGenerate>
	"Answer if the instruction preceding retpc is a call instruction."
	^(backEnd isCallPrecedingReturnPC: retpc)
	 and: [(backEnd callTargetFromReturnAddress: retpc) = ceNonLocalReturnTrampoline]! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/26/2016 08:34'!
isPCMappedAnnotation: annotation
	"See Cogit class>>initializeAnnotationConstants"
	<inline: true>
	^annotation >= HasBytecodePC! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 4/7/2015 10:32'!
isPCWithinMethodZone: address
	<api>
	<var: #address type: #'void *'>
	^address asUnsignedInteger
		between: methodZoneBase
		and: methodZone freeStart! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/3/2015 14:42'!
isPureSendAnnotation: annotation
	<inline: true>
	^annotation >= IsSendCall! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/3/2015 14:55'!
isSendAnnotation: annotation
	<inline: true>
	^annotation >= IsSendCall
	  or: [NewspeakVM and: [annotation = IsNSSendCall]]! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 9/28/2015 19:04'!
isSendReturnPC: retpc
	<api>
	"Answer if the instruction preceding retpc is a call instruction."
	| target |
	(backEnd isCallPrecedingReturnPC: retpc) ifFalse:
		[^false].
	target := backEnd callTargetFromReturnAddress: retpc.
	^(target between: firstSend and: lastSend)
	   or: [target between: methodZoneBase and: methodZone freeStart]! !

!Cogit methodsFor: 'initialization' stamp: 'eem 12/18/2015 13:56'!
isTrampolineArgConstant: n
	"Test for true and false and 0 to N encoded via trampolineArgConstant:"
	<inline: true>
	^n < NoReg! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 9/10/2019 18:59'!
kosherYoungReferrers
	<doNotGenerate>
	^methodZone kosherYoungReferrers! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 2/15/2009 17:13'!
labelForPC: address
	<doNotGenerate>
	^address < methodZoneBase ifTrue:
		[self lookupAddress: address]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 8/3/2021 16:33'!
labelForSimulationAccessor: blockOrMessageSendOrSelector
	<doNotGenerate>
	^'&', (blockOrMessageSendOrSelector isBlock
			ifTrue: [(coInterpreter mappedPluginEntries detect: [:tuple| tuple third == blockOrMessageSendOrSelector] ifNone: [])
						ifNil: ['block in ', blockOrMessageSendOrSelector method selector]
						ifNotNil: [:tuple| '[=>', tuple second, ' (prim 117)]']]
			ifFalse: [blockOrMessageSendOrSelector isMessageSend
						ifTrue: [blockOrMessageSendOrSelector selector]
						ifFalse: [blockOrMessageSendOrSelector]])! !

!Cogit methodsFor: 'method map' stamp: 'eem 12/2/2015 12:05'!
lastBytecodePCForBlockAt: startbcpc in: aMethodObj
	"Answer the 0-relative pc of the last bytecode in the block starting at the 0-relative startbcpc in aMethodObj."
	| aMethodHeader bcpc bsOffset byte descriptor |
	<var: #descriptor type: #'BytecodeDescriptor *'>
	aMethodHeader := objectMemory methodHeaderOf: aMethodObj.
	bcpc := startbcpc - (self blockCreationBytecodeSizeForHeader: aMethodHeader).
	bsOffset := self bytecodeSetOffsetForHeader: aMethodHeader.
	byte := (objectMemory fetchByte: bcpc ofObject: aMethodObj) + bsOffset.
	descriptor := self generatorAt: byte.
	^(self nextBytecodePCFor: descriptor at: bcpc exts: -1 in: aMethodObj) - 1! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 2/10/2009 19:26'!
lastOpcode
	<returnTypeC: #'AbstractInstruction *'>
	self assert: opcodeIndex > 0.
	^self abstractInstructionAt: opcodeIndex - 1! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/1/2012 14:36'!
latestContinuationPCFor: descriptor at: pc exts: nExts in: aMethodObj
	"Assuming the descriptor is that for a branch or block,
	 answer the furthest reachable pc for this bytecode."
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<inline: true>
	| distance |
	distance := self spanFor: descriptor at: pc exts: nExts in: aMethodObj.
	^pc + descriptor numBytes + (distance max: 0)! !

!Cogit methodsFor: 'newspeak support' stamp: 'rmacnak 3/30/2015 21:38'!
linkNSSendCache: nsSendCache classTag: classTag enclosingObject: enclosingObject target: targetMethod caller: callingMethod
	<api>
	<option: #NewspeakVM>
	<var: #nsSendCache type: #'NSSendCache *'>
	<var: #targetMethod type: #'CogMethod *'>
	<var: #callingMethod type: #'CogMethod *'>
	nsSendCache classTag: classTag.
	nsSendCache enclosingObject: enclosingObject.
	nsSendCache target: targetMethod asInteger + cmNoCheckEntryOffset.
	callingMethod cmRefersToYoung ifFalse:
		[(enclosingObject ~= 0 and: [objectMemory isYoung: enclosingObject]) ifTrue:
			[methodZone ensureInYoungReferrers: callingMethod]].
! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 7/5/2021 16:53'!
linkSendAt: callSiteReturnAddress in: sendingMethod to: targetMethod offset: theEntryOffset receiver: receiver
	<api>
	<var: #sendingMethod type: #'CogMethod *'>
	<var: #targetMethod type: #'CogMethod *'>
	| inlineCacheTag extent |
	self assert: (theEntryOffset = cmEntryOffset or: [theEntryOffset = cmNoCheckEntryOffset]).
	self assert: (callSiteReturnAddress between: methodZoneBase and: methodZone freeStart).
	self ensureWritableCodeZone.
	theEntryOffset = cmNoCheckEntryOffset
		ifTrue: "no need to change selector cache tag"
			[extent := backEnd rewriteCallAt: callSiteReturnAddress target: targetMethod asInteger + cmNoCheckEntryOffset]
		ifFalse:
			[inlineCacheTag := objectRepresentation inlineCacheTagForInstance: receiver.
			 (self inlineCacheTagsAreIndexes not and: [objectRepresentation inlineCacheTagIsYoung: inlineCacheTag]) ifTrue:
				[methodZone ensureInYoungReferrers: sendingMethod].
			 extent := backEnd
						rewriteInlineCacheAt: callSiteReturnAddress
						tag: inlineCacheTag
						target: targetMethod asInteger + theEntryOffset].
	backEnd
		flushICacheFrom: callSiteReturnAddress asUnsignedInteger  - extent
		to: callSiteReturnAddress asUnsignedInteger! !

!Cogit methodsFor: 'analysis' stamp: 'eem 11/18/2013 12:51'!
linkedSuperSendCacheTags
	"An example; answer the cache tags for linked super sends.  They should all be
	 selectors because super sends don't have their cache tag rewritten when linked."
	<doNotGenerate>
	| cacheTags |
	cacheTags := Set new.
	methodZone methodsDo:
		[:m|
		 m cmType = CMMethod ifTrue:
			[self sendSitesIn: m do:
				[:a :mcpc| | entryPoint |
				 entryPoint := backEnd callTargetFromReturnAddress: mcpc.
				 entryPoint > methodZoneBase ifTrue:
					[self offsetAndSendTableFor: entryPoint
						annotation: a
						into:
							[:off :table|
							 off = cmNoCheckEntryOffset ifTrue:
								[cacheTags add: (backEnd inlineCacheTagAt: mcpc)]]]]]].
	^cacheTags! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'cb 4/20/2016 12:33'!
loadBytesAndGetDescriptor
	<var: #descriptor type: #'BytecodeDescriptor *'>
	| descriptor |
	byte0 := (objectMemory fetchByte: bytecodePC ofObject: methodObj)  + bytecodeSetOffset.
	descriptor := self generatorAt: byte0.
	self loadSubsequentBytesForDescriptor: descriptor at: bytecodePC.
	^ descriptor
	
	! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/1/2014 17:25'!
loadSubsequentBytesForDescriptor: descriptor at: pc
	<var: #descriptor type: #'BytecodeDescriptor *'>
	descriptor numBytes > 1 ifTrue:
		[byte1 := objectMemory fetchByte: pc + 1 ofObject: methodObj.
		 descriptor numBytes > 2 ifTrue:
			[byte2 := objectMemory fetchByte: pc + 2 ofObject: methodObj.
			 descriptor numBytes > 3 ifTrue:
				[byte3 := objectMemory fetchByte: pc + 3 ofObject: methodObj.
				 descriptor numBytes > 4 ifTrue:
					[self notYetImplemented]]]]! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 12/28/2019 02:27'!
lookup32BitWordConstant: word
	<doNotGenerate>
	^objectMemory lookup32BitWordConstant: word! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'rmacnak 6/24/2015 21:09'!
lookup: selector for: receiver methodAndErrorSelectorInto: binaryBlock
	"Lookup selector in the class of receiver.  If found, evaluate binaryBlock with the
	 method, cogged if appropriate..  If not found, due to MNU, lookup the DNU selector
	 and evaluate binaryBlock with the MNU method, cogged if appropriate..  If not found
	 due to cannot interpret, evaluate binaryBlock with a nil method and the error selector."
	| methodOrSelectorIndex |
	<inline: true>
	methodOrSelectorIndex := coInterpreter
									lookupOrdinary: selector
									receiver: receiver.
	methodOrSelectorIndex asUnsignedInteger > coInterpreter maxLookupNoMNUErrorCode ifTrue:
		[(objectMemory isOopCompiledMethod: methodOrSelectorIndex) ifFalse:
			[^binaryBlock value: methodOrSelectorIndex value: SelectorCannotInterpret].
		 ((coInterpreter methodHasCogMethod: methodOrSelectorIndex) not
		  and: [coInterpreter methodShouldBeCogged: methodOrSelectorIndex]) ifTrue:
			["We assume cog:selector: will *not* reclaim the method zone"
			 self cog: methodOrSelectorIndex selector: selector].
		^binaryBlock value: methodOrSelectorIndex value: nil].
	methodOrSelectorIndex = SelectorDoesNotUnderstand ifTrue:
		[methodOrSelectorIndex := coInterpreter
										lookupMNU: (objectMemory splObj: SelectorDoesNotUnderstand)
										receiver: receiver.
		 methodOrSelectorIndex asUnsignedInteger > coInterpreter maxLookupNoMNUErrorCode ifTrue:
			[self assert: (objectMemory isOopCompiledMethod: methodOrSelectorIndex).
			 ((coInterpreter methodHasCogMethod: methodOrSelectorIndex) not
			  and: [coInterpreter methodShouldBeCogged: methodOrSelectorIndex]) ifTrue:
				["We assume cog:selector: will *not* reclaim the method zone"
				 self cog: methodOrSelectorIndex selector: (objectMemory splObj: SelectorDoesNotUnderstand)].
			^binaryBlock value: methodOrSelectorIndex value: SelectorDoesNotUnderstand].
		^binaryBlock value: nil value: SelectorDoesNotUnderstand].
	^binaryBlock value: nil value: methodOrSelectorIndex! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 7/7/2021 22:48'!
lookupAddress: address
	<doNotGenerate>
	address < methodZone freeStart ifTrue:
		[^address >= methodZoneBase
			ifTrue:
				[(methodZone methodFor: address) ifNotNil:
					[:cogMethod|
					 ((cogMethod selector ~= objectMemory nilObject
					    and: [objectRepresentation couldBeObject: cogMethod selector])
						ifTrue: [coInterpreter stringOf: cogMethod selector]
						ifFalse: [cogMethod asInteger hex]),
					   '@', ((address - cogMethod asInteger) hex allButFirst: 3)]]
			ifFalse:
				[(self trampolineRangeFor: address) ifNotNil:
					[:range|
					 (self codeEntryNameFor: range first) ifNotNil:
						[:name| name, (address = range first ifTrue: [''] ifFalse: [' + ', (address - range first) hex])]]]].
	(simulatedTrampolines includesKey: address) ifTrue:
		[^self labelForSimulationAccessor: (simulatedTrampolines at: address)].
	(simulatedVariableGetters includesKey: address) ifTrue:
		[^self labelForSimulationAccessor: (simulatedVariableGetters at: address)].
	^(coInterpreter lookupAddress: address) ifNil:
		[address = self cStackPointerAddress
			ifTrue: [#CStackPointer]
			ifFalse:
				[address = self cFramePointerAddress ifTrue:
					[#CFramePointer]]]! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 12/11/2015 10:50'!
lookupCHexString: aCHexString 
	<doNotGenerate>
	| pastLastZero shortened address |
	(aCHexString beginsWith: '0x') ifFalse:
		[^aCHexString].
	pastLastZero := aCHexString findFirst: [:c| c ~= $0 and: [c ~= $x]].
	shortened := pastLastZero = 0
					ifTrue: ['0x0']
					ifFalse:
						[(aCHexString size >= 16 and: [pastLastZero >= 4])
							ifTrue: [aCHexString copyReplaceFrom: 3 to: pastLastZero - 1 with: '']
							ifFalse: [aCHexString]].
	address := Number readFrom: (ReadStream on: shortened from: 3 to: shortened size) base: 16.
	(disassemblingMethod notNil
	 and: [address > disassemblingMethod
	 and: [address < (disassemblingMethod asInteger + disassemblingMethod blockSize)]]) ifTrue:
		[shortened := '.+', (address - disassemblingMethod asInteger printStringBase: 16 length: 4 padded: true)].
	^(self lookupAddress: (Number
								readFrom: (ReadStream on: shortened from: 3 to: shortened size)
								base: 16))
		ifNotNil: [:string| shortened, '=', string]
		ifNil: [shortened]! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 3/10/2017 17:14'!
lookupFrameOffset: anInteger
	<doNotGenerate>
	(self class initializationOptions at: #tempNames ifAbsent: nil) ifNotNil:
		[:dict|
		 (self class initializationOptions at: #startpc ifAbsent: nil) ifNotNil:
			[:startpc|
			 (dict at: startpc + 1 ifAbsent: nil) ifNotNil:
				[:tempNames| | numArgs |
				 anInteger = FoxMFReceiver ifTrue:
					[^'self'].
				numArgs := self class initializationOptions at: #numArgs.
				1 to: tempNames size do:
					[:i|
					anInteger = (self frameOffsetOfTemporary: i - 1 numArgs: numArgs) ifTrue:
						[^tempNames at: i]]]]].
	^nil! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 1/22/2018 11:10'!
lookupInstVarOffset: offset
	^offset \\ objectMemory bytesPerOop = 0 ifTrue:
		[(self class initializationOptions at: #instVarNames ifAbsent: nil) ifNotNil:
			[:array|
			 array
				at: offset - objectMemory baseHeaderSize / objectMemory bytesPerOop + 1
				ifAbsent: nil]]! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 11/3/2020 09:59'!
lookupRegisterNumber: n
	true ifTrue:
		[Arg0Reg = n ifTrue: [^'Arg0'].
		 Arg1Reg = n ifTrue: [^'Arg1'].
		 ClassReg = n ifTrue: [^'Class'].
		 ReceiverResultReg = n ifTrue: [^'ReceiverResult'].
		 (SPReg = n and: [n ~= NativeSPReg]) ifTrue: [^'SP'].
		 FPReg = n ifTrue: [^'FP'].
		 SendNumArgsReg = n ifTrue: [^'SendNumArgs'].
		 VarBaseReg = n ifTrue: [^'VarBase'].
		 RISCTempReg = n ifTrue: [^'RISCTemp'].
		 ^nil].

	Arg0Reg = n ifTrue: [^'a0'].
	Arg1Reg = n ifTrue: [^'a1'].
	ClassReg = n ifTrue: [^'cr'].
	ReceiverResultReg = n ifTrue: [^'rr'].
	(SPReg = n and: [n ~= NativeSPReg]) ifTrue: [^'sp'].
	FPReg = n ifTrue: [^'fp'"#FPReg"].
	SendNumArgsReg = n ifTrue: [^'sna'].
	VarBaseReg = n ifTrue: [^'vb'].
	RISCTempReg = n ifTrue: [^'rt'].
	
	Arg0Reg = n ifTrue: [^#Arg0Reg].
	Arg1Reg = n ifTrue: [^#Arg1Reg].
	ClassReg = n ifTrue: [^#ClassReg].
	ReceiverResultReg = n ifTrue: [^#ReceiverResultReg].
	(SPReg = n and: [n ~= NativeSPReg]) ifTrue: [^#SPReg].
	FPReg = n ifTrue: [^#FPReg].
	SendNumArgsReg = n ifTrue: [^#SendNumArgsReg].
	VarBaseReg = n ifTrue: [^#VarBaseReg].
	RISCTempReg = n ifTrue: [^#RISCTempReg].
	^nil! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/15/2016 10:05'!
mapEndFor: cogMethod
	"Answer the address of the null byte at the end of the method map."
	<var: #cogMethod type: #'CogMethod *'>
	<inline: true>
	| end |
	end := self mapStartFor: cogMethod.
	[(objectMemory byteAt: end) ~= MapEnd] whileTrue:
		[end := end - 1.
		 self assert: end > (self firstMappedPCFor: cogMethod)].
	^end! !

!Cogit methodsFor: 'method map' stamp: 'cb 3/1/2018 15:40'!
mapFor: cogMethod bcpc: startbcpc performUntil: functionSymbol arg: arg
	"Machine-code <-> bytecode pc mapping support.  Evaluate functionSymbol
	 for each mcpc, bcpc pair in the map until the function returns non-zero,
	 answering that result, or 0 if it fails to.  To cut down on number of arguments.
	 and to be usable for both pc-mapping and method introspection, we encode
	 the annotation and the isBackwardBranch flag in the same parameter.
	 Guilty as charged."
	<var: #cogMethod type: #'CogBlockMethod *'>
	<var: #functionSymbol declareC: 'sqInt (*functionSymbol)(BytecodeDescriptor *desc, sqInt annotationAndIsBackwardBranch, char *mcpc, sqInt bcpc, void *arg)'>
	<var: #arg type: #'void *'>
	<inline: true>
	| isInBlock mcpc bcpc endbcpc map mapByte homeMethod aMethodObj result
	  latestContinuation byte descriptor bsOffset nExts annotation |
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<var: #homeMethod type: #'CogMethod *'>

	self assert: cogMethod stackCheckOffset > 0.
	mcpc := cogMethod asUnsignedInteger + cogMethod stackCheckOffset.
	"The stack check maps to the start of the first bytecode,
	 the first bytecode being effectively after frame build."
	result := self perform: functionSymbol
					with: nil
					with: 0 + (HasBytecodePC << 1)
					with: (self cCoerceSimple: mcpc to: #'char *')
					with: startbcpc
					with: arg.
	result ~= 0 ifTrue:
		[^result].
	bcpc := startbcpc.
	"In both CMMethod and CMBlock cases find the start of the map and
	 skip forward to the bytecode pc map entry for the stack check."
	cogMethod cmType = CMMethod
		ifTrue:
			[isInBlock := cogMethod cmIsFullBlock.
			 homeMethod := self cCoerceSimple: cogMethod to: #'CogMethod *'.
			 self assert: startbcpc = (coInterpreter startPCOfMethodHeader: homeMethod methodHeader).
			 map := self mapStartFor: homeMethod.
			 annotation := (objectMemory byteAt: map) >> AnnotationShift.
			 self assert: (annotation = IsAbsPCReference
						 or: [annotation = IsObjectReference
						 or: [annotation = IsRelativeCall
						 or: [annotation = IsDisplacementX2N]]]).
			 latestContinuation := startbcpc.
			 aMethodObj := homeMethod methodObject.
			 endbcpc := (objectMemory numBytesOf: aMethodObj) - 1.
			 bsOffset := self bytecodeSetOffsetForHeader: homeMethod methodHeader.
			"If the method has a primitive, skip it and the error code store, if any;
			 Logically. these come before the stack check and so must be ignored."
			 bcpc := bcpc + (self deltaToSkipPrimAndErrorStoreIn: aMethodObj
									header: homeMethod methodHeader)]
		ifFalse:
			[isInBlock := true.
			 self assert: bcpc = cogMethod startpc.
			 homeMethod := cogMethod cmHomeMethod.
			 map := self findMapLocationForMcpc: cogMethod asUnsignedInteger + (self sizeof: CogBlockMethod)
						inMethod: homeMethod.
			 self assert: map ~= 0.
			 annotation := (objectMemory byteAt: map) >> AnnotationShift.
			 self assert: (annotation >> AnnotationShift = HasBytecodePC "fiducial"
						 or: [annotation >> AnnotationShift = IsDisplacementX2N]).
			 [(annotation := (objectMemory byteAt: map) >> AnnotationShift) ~= HasBytecodePC] whileTrue:
				[map := map - 1].
			 map := map - 1. "skip fiducial; i.e. the map entry for the pc immediately following the method header."
			 aMethodObj := homeMethod methodObject.
			 bcpc := startbcpc - (self blockCreationBytecodeSizeForHeader: homeMethod methodHeader).
			 bsOffset := self bytecodeSetOffsetForHeader: homeMethod methodHeader.
			 byte := (objectMemory fetchByte: bcpc ofObject: aMethodObj) + bsOffset.
			 descriptor := self generatorAt: byte.
			 endbcpc := self nextBytecodePCFor: descriptor at: bcpc exts: -1 in: aMethodObj.
			 bcpc := startbcpc].
	nExts := 0.
	self inlineCacheTagsAreIndexes ifTrue:
		[enumeratingCogMethod := homeMethod].
	"Now skip up through the bytecode pc map entry for the stack check." 
	[(objectMemory byteAt: map) >> AnnotationShift ~= HasBytecodePC] whileTrue:
		[map := map - 1].
	map := map - 1.
	[(mapByte := objectMemory byteAt: map) ~= MapEnd] whileTrue: "defensive; we exit on bcpc"
		[mapByte >= FirstAnnotation
			ifTrue:
				[| nextBcpc isBackwardBranch |
				annotation := mapByte >> AnnotationShift.
				mcpc := mcpc + ((mapByte bitAnd: DisplacementMask) * backEnd codeGranularity).
				(self isPCMappedAnnotation: annotation) ifTrue:
					[(annotation = IsSendCall
					  and: [(mapByte := objectMemory byteAt: map - 1) >> AnnotationShift = IsAnnotationExtension]) ifTrue:
						[annotation := annotation + (mapByte bitAnd: DisplacementMask).
						 map := map - 1].
					 [byte := (objectMemory fetchByte: bcpc ofObject: aMethodObj) + bsOffset.
					  descriptor := self generatorAt: byte.
					  isInBlock
						ifTrue: [bcpc >= endbcpc ifTrue: [^0]]
						ifFalse:
							[(descriptor isReturn and: [bcpc >= latestContinuation]) ifTrue: [^0].
							 (descriptor isBranch or: [descriptor isBlockCreation]) ifTrue:
								[| targetPC |
								 targetPC := self latestContinuationPCFor: descriptor at: bcpc exts: nExts in: aMethodObj.
								 latestContinuation := latestContinuation max: targetPC].
							 latestContinuation := self maybeUnsafeJumpContinuation: latestContinuation at: bcpc for: descriptor in: aMethodObj].
					  nextBcpc := self nextBytecodePCFor: descriptor at: bcpc exts: nExts in: aMethodObj.
					  descriptor isMapped
					  or: [isInBlock and: [descriptor isMappedInBlock]]] whileFalse:
						[bcpc := nextBcpc.
						 nExts := descriptor isExtension ifTrue: [nExts + 1] ifFalse: [0]].
					 isBackwardBranch := descriptor isBranch
										   and: [self isBackwardBranch: descriptor at: bcpc exts: nExts in: aMethodObj].
					 result := self perform: functionSymbol
									with: descriptor
									with: (isBackwardBranch ifTrue: [annotation << 1 + 1] ifFalse: [annotation << 1])
									with: (self cCoerceSimple: mcpc to: #'char *')
									with: (isBackwardBranch ifTrue: [bcpc - (2 * nExts)] ifFalse: [bcpc])
									with: arg.
					 result ~= 0 ifTrue:
						[^result].
					 bcpc := nextBcpc.
					 nExts := descriptor isExtension ifTrue: [nExts + 1] ifFalse: [0]]]
			ifFalse:
				[self assert: (mapByte >> AnnotationShift = IsDisplacementX2N
							or: [mapByte >> AnnotationShift = IsAnnotationExtension]).
				 mapByte < (IsAnnotationExtension << AnnotationShift) ifTrue:
					[mcpc := mcpc + ((mapByte - DisplacementX2N << AnnotationShift) * backEnd codeGranularity)]].
		 map := map - 1].
	^0! !

!Cogit methodsFor: 'method map' stamp: 'eem 10/17/2010 13:16'!
mapFor: cogMethod do: aBlock
	<doNotGenerate>
	self mapFor: cogMethod performUntil: #withAnnotation:pc:evaluate: arg: aBlock! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/15/2016 10:09'!
mapFor: cogMethod performAllMapEntriesUntil: functionSymbol arg: arg
	"Analysis support"
	<doNotGenerate>
	| mcpc map mapByte result |
	mcpc := self firstMappedPCFor: cogMethod.
	map := self mapStartFor: cogMethod.
	[(mapByte := objectMemory byteAt: map) ~= MapEnd] whileTrue:
		[mapByte >= FirstAnnotation
			ifTrue:
				[mcpc := mcpc + ((mapByte bitAnd: DisplacementMask) * backEnd codeGranularity)]
			ifFalse:
				[mapByte < (IsAnnotationExtension << AnnotationShift) ifTrue:
					[mcpc := mcpc + ((mapByte - DisplacementX2N << AnnotationShift) * backEnd codeGranularity)]].
		 result := self perform: functionSymbol
					   with: mapByte >> AnnotationShift
					   with: (self cCoerceSimple: mcpc to: #'char *')
					   with: arg.
		 result ~= 0 ifTrue:
			[^result].
		 map := map - 1].
	^0! !

!Cogit methodsFor: 'method map' stamp: 'eem 7/25/2021 10:40'!
mapFor: cogMethod performUntil: functionSymbol arg: arg
	"Unlinking/GC/Disassembly support"
	<var: #cogMethod type: #'CogMethod *'>
	<var: #arg type: #'CogMethod *'> "most of the time arg is a CogMethod..."
	<var: #functionSymbol declareC: 'sqInt (*functionSymbol)(sqInt annotation, char *mcpc, CogMethod *arg)'>
	<inline: true>
	| mcpc map mapByte annotation result |
	mcpc := self firstMappedPCFor: cogMethod.
	map := self mapStartFor: cogMethod.
	self inlineCacheTagsAreIndexes ifTrue:
		[enumeratingCogMethod := cogMethod].
	[(mapByte := objectMemory byteAt: map) ~= MapEnd] whileTrue:
		[mapByte >= FirstAnnotation
			ifTrue:
				[mcpc := mcpc + ((mapByte bitAnd: DisplacementMask) * backEnd codeGranularity).
				 "If this is an IsSendCall annotation, peek ahead for an IsAnnotationExtension, and consume it."
				 ((annotation := mapByte >> AnnotationShift) = IsSendCall
				  and: [(mapByte := objectMemory byteAt: map - 1) >> AnnotationShift = IsAnnotationExtension]) ifTrue:
					[annotation := annotation + (mapByte bitAnd: DisplacementMask).
					 map := map - 1].
				 result := self perform: functionSymbol
							   with: annotation
							   with: (self cCoerceSimple: mcpc to: #'char *')
							   with: arg.
				 result ~= 0 ifTrue:
					[^result]]
			ifFalse:
				[mapByte < (IsAnnotationExtension << AnnotationShift) ifTrue:
					[mcpc := mcpc + ((mapByte - DisplacementX2N << AnnotationShift) * backEnd codeGranularity)]].
		 map := map - 1].
	^0! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 2/20/2016 19:01'!
mapObjectReferencesInClosedPIC: cPIC
	"Remap all object references in the closed PIC.  Answer if any references are young.
	Set codeModified if any modifications are made."
	<var: #cPIC type: #'CogMethod *'>
	| pc refersToYoung |
	pc := self addressOfEndOfCase:1 inCPIC:cPIC.

	"first we check the potential method oop load at the beginning of the CPIC"
	refersToYoung := self remapMaybeObjRefInClosedPICAt: pc - backEnd jumpLongByteSize.

	"We find the end address of the cPICNumCases'th case and can then just step forward by the case size thereafter"
	pc := self addressOfEndOfCase: cPIC cPICNumCases inCPIC: cPIC.
	
	"Next we check the potential class ref in the compare instruction, and the potential method oop load for each case."
	2 to: cPIC cPICNumCases do:
		[:i|
		(self inlineCacheTagsAreIndexes not
		 and: [objectRepresentation inlineCacheTagsMayBeObjects]) ifTrue:
			[(self remapMaybeObjRefInClosedPICAt: pc - backEnd jumpLongConditionalByteSize) ifTrue:
				[refersToYoung := true]].
		(self remapMaybeObjRefInClosedPICAt: pc - backEnd jumpLongConditionalByteSize - backEnd cmpC32RTempByteSize) ifTrue:
			[refersToYoung := true].
		pc := pc + cPICCaseSize].
	^refersToYoung! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/9/2021 14:09'!
mapObjectReferencesInGeneratedRuntime
	"Update all references to objects in the generated runtime."
	0 to: runtimeObjectRefIndex - 1 do:
		[:i| | mcpc literal mappedLiteral |
		 mcpc := objectReferencesInRuntime at: i.
		 literal := literalsManager fetchLiteralAtAnnotatedAddress: mcpc using: backEnd.
		 mappedLiteral := objectRepresentation remapObject: literal.
		 mappedLiteral ~= literal ifTrue:
			[self setCodeModified.
			 literalsManager storeLiteral: mappedLiteral atAnnotatedAddress: mcpc using: backEnd]]! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 8/20/2021 15:21'!
mapObjectReferencesInMachineCode: gcMode
	<api>
	"Update all references to objects in machine code."
	gcMode caseOf: {
		[GCModeNewSpace]	-> ["N.B. do *not* ensureWritableCodeZone for every scavenge."
									self mapObjectReferencesInMachineCodeForYoungGC].
		[GCModeFull]			-> [self ensureWritableCodeZone.
									self mapObjectReferencesInMachineCodeForFullGC].
		[GCModeBecome]		-> [self ensureWritableCodeZone.
									self mapObjectReferencesInMachineCodeForBecome] }.

	self mapPerMethodProfile. "simulation only..."

	(self asserta: methodZone freeStart <= methodZone youngReferrers) ifFalse:
		[self error: 'youngReferrers list overflowed']! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/7/2021 11:48'!
mapObjectReferencesInMachineCodeForBecome
	"Update all references to objects in machine code for a become.
	 Unlike incrementalGC or fullGC a method that does not refer to young may
	 refer to young as a result of the become operation.  Unlike incrementalGC
	 or fullGC the reference from a Cog method to its methodObject *must not*
	 change since the two are two halves of the same object."
	| cogMethod writableCogMethod hasYoungObj hasYoungObjPtr freedPIC |
	hasYoungObj := false.
	hasYoungObjPtr := (self addressOf: hasYoungObj put: [:val| hasYoungObj := val]) asInteger.
	codeModified := freedPIC := false.
	self mapObjectReferencesInGeneratedRuntime.
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[self assert: hasYoungObj not.
		 cogMethod cmType ~= CMFree ifTrue:
			[self assert: (self cogMethodDoesntLookKosher: cogMethod) = 0.
			 writableCogMethod := self writableMethodFor: cogMethod.
			 writableCogMethod selector: (objectRepresentation remapOop: cogMethod selector).
			 cogMethod cmType = CMClosedPIC
				ifTrue:
					[((objectMemory isYoung: cogMethod selector)
					   or: [self mapObjectReferencesInClosedPIC: cogMethod]) ifTrue:
						[freedPIC := true.
						 methodZone freeMethod: cogMethod]]
				ifFalse:
					[(objectMemory isYoung: cogMethod selector) ifTrue:
						[hasYoungObj := true].
					 cogMethod cmType = CMMethod ifTrue:
						[| remappedMethod |
						 self assert: cogMethod objectHeader = objectMemory nullHeaderForMachineCodeMethod.
						 remappedMethod := objectRepresentation remapOop: cogMethod methodObject.
						 remappedMethod ~= cogMethod methodObject ifTrue:
							[(coInterpreter methodHasCogMethod: remappedMethod) ifTrue:
								[self error: 'attempt to become two cogged methods'].
							 (objectMemory
									withoutForwardingOn: cogMethod methodObject
									and: remappedMethod
									with: cogMethod cmUsesPenultimateLit
									sendToCogit: #method:hasSameCodeAs:checkPenultimate:) ifFalse:
								[self error: 'attempt to become cogged method into different method'].
							 "For non-Newspeak there should ne a one-to-one mapping between bytecoded and
							  cog methods.  For Newspeak not necessarily, but only for anonymous accessors."
							 "Only reset the method object's header if it is referring to this CogMethod."
							 (coInterpreter rawHeaderOf: cogMethod methodObject) = cogMethod asInteger
								ifTrue:
									[coInterpreter
										rawHeaderOf: cogMethod methodObject
										put: cogMethod methodHeader.
									 writableCogMethod
										methodHeader: (coInterpreter rawHeaderOf: remappedMethod);
										methodObject: remappedMethod.
									 coInterpreter
										rawHeaderOf: remappedMethod
										put: cogMethod asInteger]
								ifFalse:
									[self assert: (self noAssertMethodClassAssociationOf: cogMethod methodObject)
													= objectMemory nilObject.
									 writableCogMethod
										methodHeader: (coInterpreter rawHeaderOf: remappedMethod);
										methodObject: remappedMethod]].
						 (objectMemory isYoung: cogMethod methodObject) ifTrue:
							[hasYoungObj := true]].
					 self mapFor: cogMethod
						 performUntil: #remapIfObjectRef:pc:hasYoung:
						 arg: hasYoungObjPtr.
					 hasYoungObj
						ifTrue:
							[methodZone ensureInYoungReferrers: cogMethod.
							hasYoungObj := false]
						ifFalse:
							[cogMethod cmRefersToYoung: false]]].
		cogMethod := methodZone methodAfter: cogMethod].
	"we /must/ prune youngReferrers here because a) the [cogMethod cmRefersToYoung: false]
	 block could have removed a method and subsequently it could be added back, and b) we
	 can not tolerate duplicates in the youngReferrers list."  
	methodZone pruneYoungReferrers.
	freedPIC ifTrue:
		[self unlinkSendsToFree].
	codeModified ifTrue: "After updating oops in inline caches we need to flush the icache."
		[backEnd flushICacheFrom: codeBase asUnsignedInteger to: methodZone freeStart]! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/7/2021 11:48'!
mapObjectReferencesInMachineCodeForFullGC
	"Update all references to objects in machine code for a full gc.  Since
	 the current (New)ObjectMemory GC makes everything old in a full GC
	 a method not referring to young will not refer to young afterwards"
	| cogMethod writableCogMethod |
	codeModified := false.
	self mapObjectReferencesInGeneratedRuntime.
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType ~= CMFree ifTrue:
			[self assert: (self cogMethodDoesntLookKosher: cogMethod) = 0.
			 writableCogMethod := self writableMethodFor: cogMethod.
			 writableCogMethod selector: (objectRepresentation remapOop: cogMethod selector).
			 cogMethod cmType = CMClosedPIC
				ifTrue:
					[self assert: cogMethod cmRefersToYoung not.
					 self mapObjectReferencesInClosedPIC: cogMethod]
				ifFalse:
					[cogMethod cmType = CMMethod ifTrue:
						[self assert: cogMethod objectHeader = objectMemory nullHeaderForMachineCodeMethod.
						 writableCogMethod methodObject: (objectRepresentation remapOop: cogMethod methodObject)].
					 self mapFor: cogMethod
						 performUntil: #remapIfObjectRef:pc:hasYoung:
						 arg: 0.
					 (cogMethod cmRefersToYoung
					  and: [objectRepresentation allYoungObjectsAgeInFullGC]) ifTrue:
						[writableCogMethod cmRefersToYoung: false]]].
		cogMethod := methodZone methodAfter: cogMethod].
	methodZone pruneYoungReferrers.
	codeModified ifTrue: "After updating oops in inline caches we need to flush the icache."
		[backEnd flushICacheFrom: codeBase asUnsignedInteger to: methodZone freeStart]! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/25/2021 10:48'!
mapObjectReferencesInMachineCodeForYoungGC
	"Update all references to objects in machine code for either a Spur scavenging gc
	 or a Squeak V3 incremental GC.  Avoid scanning all code by using the youngReferrers
	 list.  In a young gc a method referring to young may no longer refer to young, but a
	 method not referring to young cannot and will not refer to young afterwards."
	| pointer cogMethod hasYoungObj hasYoungObjPtr zoneIsWritable |
	codeModified := zoneIsWritable := hasYoungObj := false.
	hasYoungObjPtr := (self addressOf: hasYoungObj put: [:val| hasYoungObj := val]) asInteger.
	pointer := methodZone youngReferrers.
	[pointer < methodZone zoneEnd] whileTrue:
		[self assert: hasYoungObj not.
		 cogMethod := coInterpreter cCoerceSimple: (objectMemory longAt: pointer) to: #'CogMethod *'.
		 cogMethod cmType = CMFree
			ifTrue: [self assert: cogMethod cmRefersToYoung not]
			ifFalse:
				[self assert: (self cogMethodDoesntLookKosher: cogMethod) = 0.
				 cogMethod cmRefersToYoung ifTrue:
					[| writableCogMethod |
					 self assert: (cogMethod cmType = CMMethod
								or: [cogMethod cmType = CMOpenPIC]).
					 zoneIsWritable ifFalse:
						[self ensureWritableCodeZone.
						 zoneIsWritable := true].
					 writableCogMethod := self writableMethodFor: cogMethod.
					 writableCogMethod selector: (objectRepresentation remapOop: cogMethod selector).
					 (objectMemory isYoung: cogMethod selector) ifTrue:
						[hasYoungObj := true].
					 cogMethod cmType = CMMethod ifTrue:
						[self assert: cogMethod objectHeader = objectMemory nullHeaderForMachineCodeMethod.
						 writableCogMethod methodObject: (objectRepresentation remapOop: cogMethod methodObject).
						 (objectMemory isYoung: cogMethod methodObject) ifTrue:
							[hasYoungObj := true]].
					 self mapFor: cogMethod
						 performUntil: #remapIfObjectRef:pc:hasYoung:
						 arg: hasYoungObjPtr asVoidPointer.
					 hasYoungObj
						ifTrue: [hasYoungObj := false]
						ifFalse: [writableCogMethod cmRefersToYoung: false]]].
		 pointer := pointer + objectMemory wordSize].
	methodZone pruneYoungReferrers.
	codeModified ifTrue: "After updating oops in inline caches we need to flush the icache."
		[backEnd flushICacheFrom: methodZoneBase asUnsignedInteger to: methodZone freeStart]! !

!Cogit methodsFor: 'analysis' stamp: 'eem 8/20/2021 15:48'!
mapPerMethodProfile
	"Simulation only counting of instructions per method/pic/trampoline..."
	<cmacro: '() 0'>
	perMethodProfile ifNotNil:
		[:pmp| | map |
		 "The tricky thing here is that a method may get remapped to another method already in the profile, etc..."
		 map := Dictionary new.
		 pmp keysAndValuesDo:
			[:methodOopOrCodeThang :count|
			(methodOopOrCodeThang >= objectMemory startOfMemory
			 and: [objectMemory shouldRemapOop: methodOopOrCodeThang]) ifTrue:
				[map at: methodOopOrCodeThang put: {objectMemory remapObj: methodOopOrCodeThang. count}]].
		 map isEmpty ifFalse:
			[map keysAndValuesDo:
				[:newOop :tuple|
				 [:oldOop :count|
				  (map includesKey: oldOop) ifFalse:
					[pmp removeKey: oldOop].
				  pmp at: newOop put: count] valueWithArguments: tuple]]]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 7/23/2013 17:07'!
mapPrimitive: primitiveRoutine "<Symbol>" withIndexToUniqueAddress: primitiveIndex "<SmallInteger>"
	| uniqueAddress |
	<doNotGenerate>
	self assert: (primitiveRoutine isSymbol or: [primitiveRoutine isBlock]).
	uniqueAddress := -1 - methodZoneBase - (primitiveIndex * 4) - 16r1000 bitAnd: self allButTopBitOfAddressSpaceMask.
	simulatedTrampolines
		at: uniqueAddress
		ifAbsentPut:
			[primitiveRoutine isSymbol
				ifTrue: [MessageSend receiver: coInterpreter selector: primitiveRoutine]
				ifFalse: [primitiveRoutine]].
	^uniqueAddress! !

!Cogit methodsFor: 'method map' stamp: 'eem 2/11/2011 14:09'!
mapStartFor: cogMethod
	"Answer the address of the first byte of the method map."
	<var: #cogMethod type: #'CogMethod *'>
	<inline: true>
	^cogMethod asUnsignedInteger + cogMethod blockSize - 1! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/25/2021 10:35'!
markAndTraceLiteralsIn: cogMethod
	<option: #SpurObjectMemory>
	"Unlink sends that have unmarked classes in inline caches or freed/freeable targets.
	 Nil-out inline caches linked to open PICs.
	 Assert that any selectors are marked.  We can do this since
	 this is only run on marked methods and thus any selectors they
	 reference should already be marked."
	<var: #cogMethod type: #'CogMethod *'>
	<inline: true>
	self assert: ((cogMethod cmType = CMMethod
				 and: [objectMemory isMarked: cogMethod methodObject])
				 or: [cogMethod cmType = CMOpenPIC
				 and: [(objectMemory isImmediate: cogMethod selector)
					or: [objectMemory isMarked: cogMethod selector]]]).
	objectRepresentation
		markAndTraceLiteral: cogMethod selector
		in: cogMethod
		at: (self addressOf: cogMethod selector put: [:val| cogMethod selector: val]).
	self maybeMarkCountersIn: cogMethod.
	self maybeMarkIRCsIn: cogMethod.
	self mapFor: cogMethod
		 performUntil: #markLiterals:pc:method:
		 arg: cogMethod! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 7/7/2021 11:48'!
markAndTraceMachineCodeForNewSpaceGC
	"Free any methods that refer to unmarked objects, unlinking sends to freed methods."
	| pointer cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	objectMemory leakCheckNewSpaceGC ifTrue:
		[self asserta: self allMachineCodeObjectReferencesValid].
	codeModified := false.
	pointer := methodZone youngReferrers.
	[pointer < methodZone zoneEnd] whileTrue:
		[cogMethod := coInterpreter cCoerceSimple: (objectMemory longAt: pointer) to: #'CogMethod *'.
		 cogMethod cmRefersToYoung ifTrue:
			[self assert: (self cogMethodDoesntLookKosher: cogMethod) = 0.
			 self assert: (cogMethod cmType = CMMethod
						or: [cogMethod cmType = CMOpenPIC]).
			 (objectMemory isYoung: cogMethod selector) ifTrue:
				[objectMemory markAndTrace: cogMethod selector].
			 cogMethod cmType = CMMethod ifTrue:
				[(objectMemory isYoung: cogMethod methodObject) ifTrue:
					[objectMemory markAndTrace: cogMethod methodObject].
				self markYoungObjectsIn: cogMethod]].
		 pointer := pointer + objectMemory wordSize].
	objectMemory leakCheckNewSpaceGC ifTrue:
		[self asserta: self allMachineCodeObjectReferencesValid].
	codeModified ifTrue: "After updating oops in inline caches we need to flush the icache."
		[backEnd flushICacheFrom: methodZoneBase asUnsignedInteger to: methodZone freeStart]! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 7/7/2021 11:49'!
markAndTraceMachineCodeOfMarkedMethods
	"Mark objects in machine-code of marked methods (or open PICs with marked selectors)."
	<api>
	<option: #SpurObjectMemory>
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	objectMemory leakCheckFullGC ifTrue:
		[self asserta: self allMachineCodeObjectReferencesValid].
	codeModified := false.
	self markAndTraceObjectReferencesInGeneratedRuntime.
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[(cogMethod cmType = CMMethod
		  and: [objectMemory isMarked: cogMethod methodObject]) ifTrue:
			[self markAndTraceLiteralsIn: cogMethod].
		 (cogMethod cmType = CMOpenPIC
		  and: [(objectMemory isImmediate: cogMethod selector)
				or: [objectMemory isMarked: cogMethod selector]]) ifTrue:
			[self markAndTraceLiteralsIn: cogMethod].
		 cogMethod := methodZone methodAfter: cogMethod].
	objectMemory leakCheckFullGC ifTrue:
		[self asserta: self allMachineCodeObjectReferencesValid].
	codeModified ifTrue: "After updating oops in inline caches we need to flush the icache."
		[backEnd flushICacheFrom: methodZoneBase asUnsignedInteger to: methodZone freeStart]! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 6/16/2015 12:23'!
markAndTraceObjectReferencesInGeneratedRuntime
	"Mark and trace any object references in the generated run-time."
	0 to: runtimeObjectRefIndex - 1 do:
		[:i| | mcpc literal |
		 mcpc := objectReferencesInRuntime at: i.
		 literal := literalsManager fetchLiteralAtAnnotatedAddress: mcpc using: backEnd.
		 objectRepresentation
			markAndTraceLiteral: literal
			in: (self cCoerceSimple: nil to: #'CogMethod *')
			atpc: mcpc asUnsignedInteger]! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 3/13/2014 17:29'!
markAndTraceObjectsOrFreeMachineCode: inFullGC
	<api>
	<option: #SqueakV3ObjectMemory>
	inFullGC
		ifTrue: [self markAndTraceOrFreeMachineCodeForFullGC]
		ifFalse: [self markAndTraceMachineCodeForNewSpaceGC]! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/9/2021 14:13'!
markAndTraceOrFreeCogMethod: cogMethod firstVisit: firstVisit
	"Mark and trace objects in the argument and free if it is appropriate.
	 Answer if the method has been freed.  firstVisit is a hint used to avoid
	 scanning methods we've already seen.  False positives are fine.
	 For a CMMethod this
			frees if the bytecode method isnt marked,
			marks and traces object literals and selectors,
			unlinks sends to targets that should be freed.
	 For a CMClosedPIC this
			frees if it refers to anything that should be freed or isn't marked.
	 For a CMOpenPIC this
			frees if the selector isn't marked."
	<var: #cogMethod type: #'CogMethod *'>
	<inline: false> "this recurses at most one level down"
	cogMethod cmType = CMFree ifTrue:
		[^true].
	self assert: (self cogMethodDoesntLookKosher: cogMethod) = 0.
	cogMethod cmType = CMMethod ifTrue:
		[(objectMemory isMarked: cogMethod methodObject) ifFalse:
			[self ensureWritableCodeZone.
			 methodZone freeMethod: cogMethod.
			 ^true].
		 firstVisit ifTrue:
			[self markLiteralsAndUnlinkUnmarkedSendsIn: cogMethod].
		^false].
	cogMethod cmType = CMClosedPIC ifTrue:
		[(self closedPICRefersToUnmarkedObject: cogMethod) ifFalse:
			[^false].
		 self ensureWritableCodeZone.
		 methodZone freeMethod: cogMethod.
		 ^true].
	cogMethod cmType = CMOpenPIC ifTrue:
		[(objectMemory isMarked: cogMethod selector) ifTrue:
			[^false].
		 self ensureWritableCodeZone.
		 methodZone freeMethod: cogMethod.
		 ^true].
	self assert: (cogMethod cmType = CMMethod
				or: [cogMethod cmType = CMClosedPIC
				or: [cogMethod cmType = CMOpenPIC]]).
	^false! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 7/7/2021 11:49'!
markAndTraceOrFreeMachineCodeForFullGC
	"Free any methods that refer to unmarked objects, unlinking sends to freed methods."
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	objectMemory leakCheckFullGC ifTrue:
		[self asserta: self allMachineCodeObjectReferencesValid].
	codeModified := false.
	self markAndTraceObjectReferencesInGeneratedRuntime.
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[self markAndTraceOrFreeCogMethod: cogMethod firstVisit: true.
		 cogMethod := methodZone methodAfter: cogMethod].
	objectMemory leakCheckFullGC ifTrue:
		[self asserta: self allMachineCodeObjectReferencesValid].
	codeModified ifTrue: "After updating oops in inline caches we need to flush the icache."
		[backEnd flushICacheFrom: methodZoneBase asUnsignedInteger to: methodZone freeStart]! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 11/21/2015 12:08'!
markAndTraceOrFreePICTarget: entryPoint in: cPIC
	"If entryPoint is that of some method, then mark and trace objects in it and free if it is appropriate.
	 Answer if the method has been freed."
	<var: #cPIC type: #'CogMethod *'>
	| targetMethod |
	<var: #targetMethod type: #'CogMethod *'>
	self assert: (entryPoint > methodZoneBase and: [entryPoint < methodZone freeStart]).
	(cPIC containsAddress: entryPoint) ifTrue:
		[^false].
	targetMethod := self cCoerceSimple: entryPoint - cmNoCheckEntryOffset to: #'CogMethod *'.
	self assert: (targetMethod cmType = CMMethod or: [targetMethod cmType = CMFree]).
	^self markAndTraceOrFreeCogMethod: targetMethod
		  firstVisit: targetMethod asUnsignedInteger > cPIC asUnsignedInteger! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/25/2021 10:43'!
markLiterals: annotation pc: mcpc method: cogMethod
	"Mark and trace literals.
	 Additionally in Newspeak, void push implicits that have unmarked classes."
	<var: #mcpc type: #'char *'>
	<var: #cogMethod type: #'CogMethod *'>
	<var: #nsSendCache type: #'NSSendCache *'>
	| literal |
	annotation = IsObjectReference ifTrue:
		[literal := literalsManager fetchLiteralAtAnnotatedAddress: mcpc asUnsignedInteger using: backEnd.
		 (objectRepresentation markAndTraceLiteral: literal in: cogMethod atpc: mcpc asUnsignedInteger) ifTrue:
			[codeModified := true]].

	NewspeakVM ifTrue:
		[annotation = IsNSSendCall ifTrue:
			[| nsSendCache sel eo |
			nsSendCache := self nsSendCacheFromReturnAddress: mcpc asInteger.
			sel := nsSendCache selector.
				(objectMemory isForwarded: sel)
					ifFalse: [objectMemory markAndTrace: sel]
					ifTrue: [sel := objectMemory followForwarded: literal.
							nsSendCache selector: sel.
							self markAndTraceUpdatedLiteral: sel in: (self cCoerceSimple: cogMethod to: #'CogMethod *')].
			eo := nsSendCache enclosingObject.
			eo ~= 0 ifTrue:
				[(objectMemory isForwarded: eo)
					ifFalse: [objectMemory markAndTrace: eo]
					ifTrue: [eo := objectMemory followForwarded: literal.
							nsSendCache enclosingObject: eo.
							self markAndTraceUpdatedLiteral: eo in: (self cCoerceSimple: cogMethod to: #'CogMethod *')]]]].

	(self isPureSendAnnotation: annotation) ifTrue:
		[self entryCacheTagAndCouldBeObjectAt: mcpc annotation: annotation into:
			[:entryPoint :cacheTag :tagCouldBeObj |
			 (tagCouldBeObj
			  and: [objectRepresentation markAndTraceCacheTagLiteral: cacheTag in: cogMethod atpc: mcpc asUnsignedInteger]) ifTrue:
				["cacheTag is selector" codeModified := true]]].

	^0 "keep scanning"! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 1/28/2017 14:52'!
markLiteralsAndUnlinkIfUnmarkedSend: annotation pc: mcpc method: cogMethod
	"Mark and trace literals.  Unlink sends that have unmarked cache tags or targets."
	<var: #mcpc type: #'char *'>
	<var: #nsSendCache type: #'NSSendCache *'>
	| literal |
	annotation = IsObjectReference ifTrue:
		[literal := literalsManager fetchLiteralAtAnnotatedAddress: mcpc asUnsignedInteger using: backEnd.
		 (objectRepresentation
				markAndTraceLiteral: literal
				in: (self cCoerceSimple: cogMethod to: #'CogMethod *')
				atpc: mcpc asUnsignedInteger) ifTrue:
			[codeModified := true]].

	NewspeakVM ifTrue:
		[annotation = IsNSSendCall ifTrue:
			[| nsSendCache entryPoint targetMethod sel eo |
			nsSendCache := self nsSendCacheFromReturnAddress: mcpc asInteger.
			entryPoint := nsSendCache target.
			entryPoint ~= 0 ifTrue: "Send is linked"
				[targetMethod := self cCoerceSimple: entryPoint - cmNoCheckEntryOffset to: #'CogMethod *'.
				 (self markAndTraceOrFreeCogMethod: targetMethod
					firstVisit: targetMethod asUnsignedInteger > mcpc asUnsignedInteger) ifTrue:	
						[self voidNSSendCache: nsSendCache]].
			sel := nsSendCache selector.
			(objectMemory isForwarded: sel)
				ifFalse: [objectMemory markAndTrace: sel]
				ifTrue: [sel := objectMemory followForwarded: literal.
						nsSendCache selector: sel.
						self markAndTraceUpdatedLiteral: sel in: (self cCoerceSimple: cogMethod to: #'CogMethod *')].
			eo := nsSendCache enclosingObject.
			eo ~= 0 ifTrue:
				[(objectMemory isForwarded: eo)
					ifFalse: [objectMemory markAndTrace: eo]
					ifTrue: [eo := objectMemory followForwarded: literal.
							nsSendCache enclosingObject: eo.
							self markAndTraceUpdatedLiteral: eo in: (self cCoerceSimple: cogMethod to: #'CogMethod *')]]]].

	(self isPureSendAnnotation: annotation) ifTrue:
		[self entryCacheTagAndCouldBeObjectAt: mcpc annotation: annotation into:
			[:entryPoint :cacheTag :tagCouldBeObj | | cacheTagMarked |
			 cacheTagMarked := tagCouldBeObj and: [objectRepresentation cacheTagIsMarked: cacheTag].
			 entryPoint > methodZoneBase
				ifTrue: "It's a linked send."
					[self targetMethodAndSendTableFor: entryPoint annotation: annotation into:
						[:targetMethod :sendTable| 
						 (cacheTagMarked not
						  or: [self markAndTraceOrFreeCogMethod: targetMethod
								firstVisit: targetMethod asUnsignedInteger > mcpc asUnsignedInteger]) ifTrue:
							["Either the cacheTag is unmarked (e.g. new class) or the target
							  has been freed (because it is unmarked), so unlink the send."
							 self unlinkSendAt: mcpc targetMethod: targetMethod sendTable: sendTable.
							 objectRepresentation
								markAndTraceLiteral: targetMethod selector
								in: targetMethod
								at: (self addressOf: targetMethod selector put: [:val| targetMethod selector: val])]]]
				ifFalse:  "cacheTag is selector"
					[(objectRepresentation
							markAndTraceCacheTagLiteral: cacheTag
							in: (self cCoerceSimple: cogMethod to: #'CogMethod *')
							atpc: mcpc asUnsignedInteger) ifTrue:
						[codeModified := true]]]].

	^0 "keep scanning"! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/25/2021 10:35'!
markLiteralsAndUnlinkUnmarkedSendsIn: cogMethod
	"Unlink sends that have unmarked classes in inline caches or freed/freeable targets.
	 Nil-out inline caches linked to open PICs.
	 Assert that any selectors are marked.  We can do this since
	 this is only run on marked methods and thus any selectors they
	 reference should already be marked."
	<var: #cogMethod type: #'CogMethod *'>
	<inline: true>
	self assert: cogMethod cmType = CMMethod.
	self assert: (objectMemory isMarked: cogMethod methodObject).
	objectRepresentation
		markAndTraceLiteral: cogMethod selector
		in: cogMethod
		at: (self addressOf: cogMethod selector put: [:val| cogMethod selector: val]).
	self maybeMarkCountersIn: cogMethod.
	self maybeMarkIRCsIn: cogMethod.
	self mapFor: cogMethod
		 performUntil: #markLiteralsAndUnlinkIfUnmarkedSend:pc:method:
		 arg: cogMethod! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 1/28/2020 17:40'!
markMethodAndReferents: aCogMethod
	<api>
	<var: #aCogMethod type: #'CogBlockMethod *'>
	| cogMethod writableMethod |
	self assert: (aCogMethod cmType = CMMethod
				or: [aCogMethod cmType = CMBlock]).
	cogMethod := aCogMethod cmType = CMMethod
					ifTrue: [self cCoerceSimple: aCogMethod to: #'CogMethod *']
					ifFalse: [aCogMethod cmHomeMethod].
	writableMethod := self writableMethodFor: cogMethod.
	writableMethod cmUsageCount: CMMaxUsageCount.
	self mapFor: cogMethod
		performUntil: #incrementUsageOfTargetIfLinkedSend:mcpc:ignored:
		arg: 0! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 1/28/2017 14:52'!
markYoungObjects: annotation pc: mcpc method: cogMethod
	"Mark and trace young literals."
	<var: #mcpc type: #'char *'>
	<var: #nsSendCache type: #'NSSendCache *'>
	| literal |
	annotation = IsObjectReference ifTrue:
		[literal := literalsManager fetchLiteralAtAnnotatedAddress: mcpc asUnsignedInteger using: backEnd.
		 objectRepresentation markAndTraceLiteralIfYoung: literal].

	NewspeakVM ifTrue: [annotation = IsNSSendCall ifTrue:
		[| nsSendCache |
		 nsSendCache := self nsSendCacheFromReturnAddress: mcpc asInteger.
		 objectRepresentation markAndTraceLiteralIfYoung: nsSendCache selector.
		 nsSendCache enclosingObject ~= 0 ifTrue:
			[objectRepresentation markAndTraceLiteralIfYoung: nsSendCache enclosingObject]]].

	(self isPureSendAnnotation: annotation) ifTrue:
		[self entryCacheTagAndCouldBeObjectAt: mcpc annotation: annotation into:
			[:entryPoint :cacheTag :tagCouldBeObj |
			 tagCouldBeObj ifTrue:
				[objectRepresentation markAndTraceLiteralIfYoung: cacheTag]]].

	^0 "keep scanning"! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/25/2021 10:35'!
markYoungObjectsIn: cogMethod
	"Mark young literals in the method."
	<var: #cogMethod type: #'CogMethod *'>
	<inline: true>
	self assert: (cogMethod cmType = CMMethod
				or: [cogMethod cmType = CMOpenPIC]).
	 (objectMemory isYoung: cogMethod selector) ifTrue:
		[objectMemory markAndTrace: cogMethod selector].
	(cogMethod cmType = CMMethod
	 and: [objectMemory isYoung: cogMethod methodObject]) ifTrue:
		[objectMemory markAndTrace: cogMethod methodObject].
	self mapFor: cogMethod
		 performUntil: #markYoungObjects:pc:method:
		 arg: cogMethod! !

!Cogit methodsFor: 'accessing' stamp: 'eem 3/16/2017 14:56'!
maxCogCodeSize
	"We restrict the maximum size of the code zone to 16Mb to allow inter-method
	 calls and jumps to use small offset call and jump instructions if appropriate."
	<api>
	<cmacro: '() (16*1024*1024)'>
	^16*1024*1024! !

!Cogit methodsFor: 'debugging' stamp: 'eem 11/25/2009 11:13'!
maxCogMethodAddress
	<api>
	<returnTypeC: #usqInt>
	^methodZone limitZony asUnsignedInteger! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 01/10/2016 10:01'!
maybeAllocAndInitCounters
	"No-op in the non-Sista Cogits..."
	^true! !

!Cogit methodsFor: 'newspeak support' stamp: 'eem 2/12/2016 13:59'!
maybeAllocAndInitIRCs
	"If this is the Newspeak VM and the objectRepresentation supports pinning
	 then allocate space for the implicit receiver caches on the heap."
	NewspeakVM
		ifTrue:
			[indexOfIRC := theIRCs := 0.
			 (objectRepresentation canPinObjects and: [numIRCs > 0]) ifTrue:
				[self assert: (self noAssertMethodClassAssociationOf: methodObj) ~= objectMemory nilObject.
				 theIRCs := objectRepresentation allocateNPinnedSlots: numIRCs * NumOopsPerNSC.
				 ^theIRCs ~= 0].
			 ^true]
		ifFalse:
			[^true]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 12/27/2015 12:36'!
maybeBreakAt: address
	<doNotGenerate>
	((breakPC isBreakpointFor: address)
	 and: [breakBlock shouldStopIfAtPC: address]) ifTrue:
		[coInterpreter changed: #byteCountText.
		 self halt: 'machine code breakpoint at ', address hex, ' in ', thisContext sender selector]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 9/15/2020 13:28'!
maybeBreakGeneratingFrom: address to: end
	"Variation on maybeBreakAt: that only works for integer breakPCs,
	 so we can have break blocks that stop at any pc, except when generating."
	<cmacro: '(address,end) 0'> "Simulation only; void in C"
	(breakPC isInteger
	 and: [breakPC >= address
	 and: [breakPC < end
	 and: [breakBlock shouldStopIfAtPC: address]]]) ifTrue:
		[coInterpreter changed: #byteCountText.
		 self halt: 'machine code generation at ', address hex, ' in ', thisContext sender selector]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 7/21/2021 09:23'!
maybeBreakGeneratingInstructionWithIndex: index
	"Variation on maybeBreakAt: that only works for integer abstract instruction indexes,
	 so we can have break blocks that stop at any pc, except when generating."
	<cmacro: '(i) 0'> "Simulation only; void in C"
	(InitializationOptions at: #instructionIndex ifAbsent: nil) ifNotNil: [:breakIndex| index = breakIndex ifTrue: [self halt]]! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'cb 1/3/2018 15:39'!
maybeDealWithUnsafeJumpForDescriptor: descriptor pc: pc latestContinuation: latestContinuation
	<inline: true>
	^ latestContinuation! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/11/2015 14:12'!
maybeDumpLiterals: descriptor
	<inline: true>
	<var: #descriptor type: #'BytecodeDescriptor *'>
	((literalsManager mustDumpLiterals: opcodeIndex)
	  or: [descriptor isUnconditionalBranch
	  or: [descriptor isReturn]]) ifTrue:
		[literalsManager dumpLiterals: (descriptor isUnconditionalBranch
										or: [descriptor isReturn]) not]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 12/24/2015 12:38'!
maybeEnableSingleStep
	<inline: true>
	self cCode: '' inSmalltalk:
		[singleStep ifFalse: [singleStep := breakPC singleStepRequiredToTriggerIn: self]]! !

!Cogit methodsFor: 'generate machine code - dual mapped zone support' stamp: 'eem 7/9/2020 09:43'!
maybeFlushWritableZoneFrom: startAddress to: endAddress
	"If there is a dual mapped code zone (the normal zone but marked with read/execute,
	 and a read/write zone codeToDataDelta bytes away) then the data cache for the read/write
	 zone must be flushed, and the data cache for the read/execute zone must be invalidated,
	 for the Cogit to see the same values in both zones after a write to the read/write zone."
	<var: 'startAddress' type: #usqInt>
	<var: 'endAddress' type: #usqInt>
	<inline: #always>
	self cppIf: #DUAL_MAPPED_CODE_ZONE
		ifTrue:
			[codeToDataDelta > 0 ifTrue:
				[backEnd flushDCacheFrom: startAddress to: endAddress]]! !

!Cogit methodsFor: 'debugging' stamp: 'eem 7/10/2009 13:57'!
maybeFreeCogMethodDoesntLookKosher: cogMethod
	"Check that the header fields are consistent with the type.
	 Answer 0 if it is ok, otherwise answer a code for the error."
	<var: #cogMethod type: #'CogMethod *'>
	| result |
	result := self cogMethodDoesntLookKosher: cogMethod.
	^result = 2 ifTrue: [0] ifFalse: [result]! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 01/10/2016 10:01'!
maybeFreeCounters
	"No-op in the non-Sista Cogits..."! !

!Cogit methodsFor: 'compaction' stamp: 'eem 7/3/2014 14:51'!
maybeFreeCountersOf: aCogMethod
	"Sista allocates counters out-of-line that need to be freed later on.
	 This is the hook Sista uses.  By default do nothing."
	<inline: true>! !

!Cogit methodsFor: 'initialization' stamp: 'eem 7/9/2021 18:38'!
maybeGenerateCacheFlush
	| startAddress |
	<inline: true>
	backEnd numICacheFlushOpcodes > 0 ifTrue:
		[self allocateOpcodes: backEnd numICacheFlushOpcodes bytecodes: 0.
		 startAddress := methodZoneBase.
		 backEnd generateICacheFlush.
		 self outputInstructionsForGeneratedRuntimeAt: startAddress.
		 self recordGeneratedRunTime: 'ceFlushICache' address: startAddress.
		 ceFlushICache := self cCoerceSimple: startAddress to: #'void (*)(usqIntptr_t,usqIntptr_t)'.
		 backEnd initialFlushICacheFrom: startAddress to: methodZoneBase].
	backEnd numDCacheFlushOpcodes > 0 ifTrue:
		[self allocateOpcodes: backEnd numDCacheFlushOpcodes bytecodes: 0.
		 startAddress := methodZoneBase.
		 backEnd generateDCacheFlush.
		 self outputInstructionsForGeneratedRuntimeAt: startAddress.
		 self recordGeneratedRunTime: 'ceFlushDCache' address: startAddress.
		 ceFlushDCache := self cCoerceSimple: startAddress to: #'void (*)(usqIntptr_t,usqIntptr_t)'.
		 backEnd initialFlushICacheFrom: startAddress to: methodZoneBase]! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 2/3/2018 15:47'!
maybeHaltIfDebugPC 
	<cmacro: '() 0'> "Simulation only; void in C"
	((debugBytecodePointers includes: bytecodePC)
	 and: [breakMethod isNil or: [methodObj = breakMethod]]) ifTrue:
		[self halt: ' at bcpc ', bytecodePC printString, '/', (bytecodePC + 1) printString]! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/5/2014 20:07'!
maybeMarkCountersIn: cogMethod
	"In SIsta Spur counters are held on the heap in pinned objects which must be marked
	 to avoid them being garbage collected.  This is the hook through which that happens."
	<var: #cogMethod type: #'CogMethod *'>! !

!Cogit methodsFor: 'newspeak support' stamp: 'eem 2/12/2016 13:59'!
maybeMarkIRCsIn: cogMethod
	<inline: true>
	NewspeakVM ifTrue:
		[objectRepresentation canPinObjects ifTrue:
			[objectRepresentation markIfIRC: cogMethod nextMethodOrIRCs]]! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 2/20/2017 10:19'!
maybeNoteStartpcFor: thing "<CogBlockStart|CogCodeRange>"
	<doNotGenerate>
	(self class initializationOptions at: #tempNames ifAbsent: nil) ifNotNil:
		[self class initializationOptions
			at: #startpc put: thing startpc;
			at: #numArgs put: ([thing cogMethod ifNil: [-1] ifNotNil: [:cm| cm cmNumArgs]] "CogCodeRange"
									on: MessageNotUnderstood
									do: [:ex| thing numArgs])] "CogBlockStart"! !

!Cogit methodsFor: 'profiling primitives' stamp: 'eem 6/1/2018 12:16'!
maybeTopRemapped: anOop
	<inline: true>
	^SPURVM ifTrue: [anOop] ifFalse: [objectMemory topRemappableOop]! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'cb 1/8/2018 16:27'!
maybeUnsafeJumpContinuation: latestContinuation at: bcpc for: descriptor in: aMethodObj
	<inline: true>
	^ latestContinuation! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/14/2014 11:28'!
mcPCFor: bcpc startBcpc: startbcpc in: cogMethod
	"Answer the absolute machine code pc matching the zero-relative bytecode pc argument
	 in cogMethod, given the start of the bytecodes for cogMethod's block or method object."
	<var: #cogMethod type: #'CogBlockMethod *'>
	<returnTypeC: #usqInt>
	^self
		mapFor: cogMethod
		bcpc: startbcpc
		performUntil: #find:IsBackwardBranch:Mcpc:Bcpc:MatchingBcpc:
		arg: bcpc asVoidPointer! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/14/2014 11:27'!
mcPCForBackwardBranch: bcpc startBcpc: startbcpc in: cogMethod
	"Answer the absolute machine code pc matching the zero-relative
	 bytecode pc of a backward branch in cogMethod, given the start
	 of the bytecodes for cogMethod's block or method object."
	<api>
	<var: #cogMethod type: #'CogBlockMethod *'>
	<returnTypeC: #usqInt>
	^self
		mapFor: cogMethod
		bcpc: startbcpc
		performUntil: #findBackwardBranch:IsBackwardBranch:Mcpc:Bcpc:MatchingBcpc:
		arg: bcpc asVoidPointer! !

!Cogit methodsFor: 'initialization' stamp: 'eem 8/31/2021 14:30'!
mclassCouldBeContext
	receiverTags < 0 ifTrue:
		[receiverTags := objectMemory receiverTagBitsForMethod: methodObj].
	^receiverTags anyMask: (1 bitShift: objectMemory numTagBits)! !

!Cogit methodsFor: 'initialization' stamp: 'eem 8/31/2021 14:29'!
mclassIsSmallInteger
	receiverTags < 0 ifTrue:
		[receiverTags := objectMemory receiverTagBitsForMethod: methodObj].
	^objectMemory isIntegerObject: receiverTags! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 08/08/2014 10:42'!
method: methodA hasSameCodeAs: methodB
	"For the purposes of become: see if the two methods are similar, i.e. can be safely becommed.
	 This is pretty strict.  All literals and bytecodes must be identical.  Only trailer bytes and header
	  flags can differ."
	<inline: false>
	| headerA headerB numLitsA endPCA |
	headerA := objectMemory methodHeaderOf: methodA.
	headerB := objectMemory methodHeaderOf: methodB.
	numLitsA := objectMemory literalCountOfMethodHeader: headerA.
	endPCA := self endPCOf: methodA.
	((coInterpreter argumentCountOfMethodHeader: headerA) ~= (coInterpreter argumentCountOfMethodHeader: headerB)
	 or: [(coInterpreter temporaryCountOfMethodHeader: headerA) ~= (coInterpreter temporaryCountOfMethodHeader: headerB)
	 or: [(coInterpreter primitiveIndexOfMethod: methodA header: headerA) ~= (coInterpreter primitiveIndexOfMethod: methodB header: headerB)
	 or: [numLitsA ~= (objectMemory literalCountOfMethodHeader: headerB)
	 or: [endPCA > (objectMemory numBytesOf: methodB)]]]]) ifTrue:
		[^false].
	 1 to: numLitsA - 1 do:
		[:li|
		(objectMemory fetchPointer: li ofObject: methodA) ~= (objectMemory fetchPointer: li ofObject: methodB) ifTrue:
			[^false]].
	(coInterpreter startPCOfMethod: methodA) to: endPCA do:
		[:bi|
		(objectMemory fetchByte: bi ofObject: methodA) ~= (objectMemory fetchByte: bi ofObject: methodB) ifTrue:
			[^false]].
	^true! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 08/08/2014 10:42'!
method: methodA hasSameCodeAs: methodB checkPenultimate: comparePenultimateLiteral
	"For the purposes of become: see if the two methods are similar, i.e. can be safely becommed.
	 This is pretty strict.  All literals and bytecodes must be identical.  Only trailer bytes and header
	  flags can differ."
	<inline: false>
	| headerA headerB numLitsA endPCA |
	headerA := objectMemory methodHeaderOf: methodA.
	headerB := objectMemory methodHeaderOf: methodB.
	numLitsA := objectMemory literalCountOfMethodHeader: headerA.
	endPCA := self endPCOf: methodA.
	((coInterpreter argumentCountOfMethodHeader: headerA) ~= (coInterpreter argumentCountOfMethodHeader: headerB)
	 or: [(coInterpreter temporaryCountOfMethodHeader: headerA) ~= (coInterpreter temporaryCountOfMethodHeader: headerB)
	 or: [(coInterpreter primitiveIndexOfMethod: methodA header: headerA) ~= (coInterpreter primitiveIndexOfMethod: methodB header: headerB)
	 or: [numLitsA ~= (objectMemory literalCountOfMethodHeader: headerB)
	 or: [endPCA > (objectMemory numBytesOf: methodB)]]]]) ifTrue:
		[^false].
	 1 to: numLitsA - 1 do:
		[:li|
		(objectMemory fetchPointer: li ofObject: methodA) ~= (objectMemory fetchPointer: li ofObject: methodB) ifTrue:
			[(li < (numLitsA - 1) "If the method doesn't use the penultimate literal then don't fail the comparison."
			  or: [comparePenultimateLiteral]) ifTrue:
				[^false]]].
	(coInterpreter startPCOfMethod: methodA) to: endPCA do:
		[:bi|
		(objectMemory fetchByte: bi ofObject: methodA) ~= (objectMemory fetchByte: bi ofObject: methodB) ifTrue:
			[^false]].
	^true! !

!Cogit methodsFor: 'testing' stamp: 'eem 2/19/2018 12:49'!
methodFoundInvalidPostScan
	"This is a hook for subclasses to filter out methods they can't deal with."
	<inline: true>
	^false! !

!Cogit methodsFor: 'accessing' stamp: 'eem 5/20/2015 19:31'!
methodLabel
	<cmacro: '() methodLabel'>
	^methodLabel! !

!Cogit methodsFor: 'accessing' stamp: 'eem 1/7/2016 10:16'!
methodNumArgs
	^methodOrBlockNumArgs! !

!Cogit methodsFor: 'accessing' stamp: 'eem 12/15/2015 21:24'!
methodZone
	^methodZone! !

!Cogit methodsFor: 'trampoline support' stamp: 'eem 5/31/2014 16:32'!
methodZoneBase
	<cmacro: '() methodZoneBase'>
	^methodZoneBase! !

!Cogit methodsFor: 'accessing' stamp: 'eem 9/20/2011 17:21'!
minCallAddress
	<cmacro: '() minValidCallAddress'>
	^minValidCallAddress! !

!Cogit methodsFor: 'debugging' stamp: 'eem 7/1/2021 11:23'!
minCogMethodAddress
	<api>
	<cmacro: '() methodZoneBase'>
	^methodZoneBase! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 4/13/2009 20:23'!
mnuOffset
	<api>
	^missOffset! !

!Cogit methodsFor: 'simulation support' stamp: 'eem 7/25/2021 14:25'!
mostLikelyPrimInvocationPC: branchPC or: retPC
	(branchPC > methodZoneBase and: [branchPC < methodZone freeStart]) ifTrue:
		[^branchPC].
	^retPC! !

!Cogit methodsFor: 'analysis' stamp: 'eem 8/21/2021 00:11'!
moveProfileToMethods
	"Simulation only counting of instructions per method/pic/trampoline..."
	<cmacro: '() 0'>
	"Whenever a change in the code zone is about to occur (e.g. compact code zone, free a method)
	 move all counts to the corresponding Smalltalk objects."
	perMethodProfile ifNotNil:
		[:pmp|
		0 to: trampolineTableIndex - 3 by: 2 do:
			[:i| | trampoline total |
			total := 0.
			trampoline := trampolineAddresses at: i + 1.
			trampoline to: (trampolineAddresses at: i + 3) - 1 do:
				[:pc| total := total + (instructionProfile at: pc)].
			total > 0 ifTrue:
				[perMethodProfile at: trampoline put: (perMethodProfile at: trampoline ifAbsent: 0) + total]].
		methodZone methodsDo:
			[:cogMethod| | total |
			total := 0.
			cogMethod + 1 to: cogMethod address + cogMethod blockSize do:
				[:pc| total := total + (instructionProfile at: pc)].
			total > 0 ifTrue:
				[(cogMethod cmType = CMMethod
					ifTrue: [cogMethod methodObject]
					ifFalse:
						[(cogMethod cmType = CMClosedPIC
						  or: [cogMethod cmType = CMOpenPIC]) ifTrue:
							[cogMethod selector]]) ifNotNil:
					[:thang|
					perMethodProfile at: thang put: (perMethodProfile at: thang ifAbsent: 0) + total]]].
		instructionProfile atAllPut: 0]! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'cb 1/7/2016 14:44'!
needsFrameIfImmutability: stackDelta
	^ IMMUTABILITY! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 9/5/2016 14:11'!
needsFrameIfInBlock: stackDelta
	^inBlock > 0! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 8/17/2013 18:47'!
needsFrameNever: stackDelta
	^false! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/1/2012 16:01'!
nextBytecodePCFor: descriptor at: pc exts: nExts in: aMethodObj
	"Compute the distance to the logically subsequent bytecode, i.e. skip over blocks."
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<inline: true>
	^pc
		+ descriptor numBytes
		+ (descriptor isBlockCreation
			ifTrue: [self spanFor: descriptor at: pc exts: nExts in: aMethodObj]
			ifFalse: [0])! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'cb 4/20/2016 12:32'!
nextBytecodePCFor: descriptor exts: nExts
	<inline: true>
	<var: #descriptor type: #'BytecodeDescriptor *'>
	 ^ self nextBytecodePCFor: descriptor at: bytecodePC exts: nExts in: methodObj! !

!Cogit methodsFor: 'bytecode generator support' stamp: 'eem 1/24/2017 11:01'!
nextDescriptorExtensionsAndNextPCInto: aQuaternaryBlock
	"Peek ahead and deliver the next descriptor, extension bytes and next pc."
	<inline: true>
	| savedB0 savedB1 savedB2 savedB3 savedEA savedEB savedNEB descriptor bcpc |
	<var: #descriptor type: #'BytecodeDescriptor *'>
	descriptor := self generatorAt: byte0.
	savedB0 := byte0. savedB1 := byte1. savedB2 := byte2. savedB3 := byte3.
	savedEA := extA. savedEB := extB. savedNEB := numExtB.
	bcpc := bytecodePC + descriptor numBytes.
	[bcpc > endPC ifTrue:
		[^aQuaternaryBlock value: nil value: 0 value: 0 value: 0].
	 byte0 := (objectMemory fetchByte: bcpc ofObject: methodObj)  + bytecodeSetOffset.
	 descriptor := self generatorAt: byte0.
	 self loadSubsequentBytesForDescriptor: descriptor at: bcpc.
	 descriptor isExtension ifFalse:
		[| eA eB |
		 eA := extA. eB := extB.
		 extA := savedEA. extB := savedEB. numExtB := savedNEB.
		 byte0 := savedB0. byte1 := savedB1. byte2 := savedB2. byte3 := savedB3.
	 	 ^aQuaternaryBlock value: descriptor value: eA value: eB value: bcpc].
	 self perform: descriptor generator.
	 bcpc := bcpc + descriptor numBytes.
	 true] whileTrue! !

!Cogit methodsFor: 'debugging' stamp: 'eem 8/8/2014 10:19'!
noAssertMethodClassAssociationOf: methodPointer
	^coInterpreter
		literal: (objectMemory literalCountOfMethodHeader: (coInterpreter noAssertHeaderOf: methodPointer)) - 1
		ofMethod: methodPointer! !

!Cogit methodsFor: 'accessing' stamp: 'eem 9/6/2016 19:23'!
noCheckEntryOffset
	<api>
	<cmacro>
	^cmNoCheckEntryOffset! !

!Cogit methodsFor: 'compaction' stamp: 'eem 4/14/2015 18:48'!
noCogMethodsMaximallyMarked
	"Check that no method is maximally marked.  A maximal mark is an indication the
	 method has been scanned to increase the usage count of its referent methods."
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[(cogMethod cmType ~= CMFree
		  and: [cogMethod cmUsageCount = CMMaxUsageCount]) ifTrue:
			[^false].
		 cogMethod := methodZone methodAfter: cogMethod].
	^true! !

!Cogit methodsFor: 'accessing' stamp: 'eem 1/9/2011 10:40'!
noContextSwitchBlockEntryOffset
	<api>
	<cmacro: '() blockNoContextSwitchOffset'>
	self assert: blockNoContextSwitchOffset notNil.
	^blockNoContextSwitchOffset! !

!Cogit methodsFor: 'compaction' stamp: 'eem 5/26/2016 16:55'!
noTargetsFreeInClosedPIC: cPIC
	"Answer if all targets in the PIC are in-use methods."
	<var: #cPIC type: #'CogMethod *'>
	^(self cPICHasFreedTargets: cPIC) not! !

!Cogit methodsFor: 'translation support' stamp: 'eem 2/12/2009 15:53'!
notYetImplemented
	<cmacro: '() warning("not yet implemented")'>
	super notYetImplemented! !

!Cogit methodsFor: 'newspeak support' stamp: 'rmacnak 12/12/2015 11:45:48'!
nsSendCacheFromReturnAddress: mcpc
	"Caller looks like
			mov LcacheAddress, SendNumArgsReg
			call ceImplicitReceiver"
	<option: #NewspeakVM>
	<var: #mcpc type: #'char *'>
	<inline: true>
	| entryPoint cacheAddress |
	entryPoint := backEnd callTargetFromReturnAddress: mcpc asInteger.
	self assert: entryPoint < methodZoneBase.
	cacheAddress := (backEnd nsSendCacheAt: mcpc) asUnsignedInteger.
	self assert: (objectMemory isInOldSpace: cacheAddress).
	^self cCoerceSimple: cacheAddress to: #'NSSendCache *'
! !

!Cogit methodsFor: 'simulation only' stamp: 'rmacnak 3/27/2015 19:46'!
nsSendCacheSurrogateAt: address
	<doNotGenerate>
	self assert: (address < 0 or: [(address bitAnd: objectMemory wordSize - 1) = 0]).
	^nsSendCacheSurrogateClass new
		at: address
		objectMemory: objectMemory
		cogit: self! !

!Cogit methodsFor: 'translation' stamp: 'eem 1/5/2021 16:03'!
null
	"This is needed since Cogits do not include sq.h."
	<cmacro: ' 0'>
	^nil! !

!Cogit methodsFor: 'initialization' stamp: 'eem 12/18/2015 14:01'!
numArgsOrSendNumArgsReg: numArgs
	"The send trampolines have different versions for different arg counts, with special
	 cases for 0 through NumSendTrampolines - 2, and a general case for more, passing
	 the arg count in SendNumArgsReg.  This computes the relevant argument."
	<inline: true>
	^numArgs <= (NumSendTrampolines - 2)
		ifTrue: [self trampolineArgConstant: numArgs]
		ifFalse: [SendNumArgsReg]! !

!Cogit methodsFor: 'accessing' stamp: 'eem 10/13/2009 12:40'!
objectMemory
	<doNotGenerate>
	^objectMemory! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 11/29/2010 14:33'!
objectRepresentation
	<doNotGenerate>
	^objectRepresentation! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 2/15/2011 10:26'!
offset: aClass of: fieldSymbol
	"This is implemented by stddef's offsetof macro."
	<doNotGenerate>
	^aClass caseOf:
		{ [CogMethod] -> [cogMethodSurrogateClass offsetOf: fieldSymbol] }! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 6/1/2018 11:08'!
offsetAndSendTableFoAnnotation: annotation into: binaryBlock
	"Find the relevant sendTable for a linked-send with the given annotation.
	 c.f. annotationForSendTable:"
	<inline: true>
	| offset sendTable |
	<var: #sendTable type: #'sqInt *'>
	annotation = IsSendCall ifTrue:
		[offset := cmEntryOffset.
		 sendTable := ordinarySendTrampolines] ifFalse:
	[(BytecodeSetHasDirectedSuperSend and: [annotation = IsDirectedSuperSend]) ifTrue:
		[offset := cmNoCheckEntryOffset.
		 sendTable := directedSuperSendTrampolines] ifFalse:
	[(BytecodeSetHasDirectedSuperSend and: [annotation = IsDirectedSuperBindingSend]) ifTrue:
		[offset := cmNoCheckEntryOffset.
		 sendTable := directedSuperBindingSendTrampolines] ifFalse:
	[(NewspeakVM and: [annotation = IsNSSelfSend]) ifTrue:
		[offset := cmEntryOffset.
		 sendTable := selfSendTrampolines] ifFalse:
	[(NewspeakVM and: [annotation = IsNSDynamicSuperSend]) ifTrue:
		[offset := cmEntryOffset.
		 sendTable := dynamicSuperSendTrampolines] ifFalse:
	[self assert: annotation = IsSuperSend.
	 offset := cmNoCheckEntryOffset.
	 sendTable := superSendTrampolines]]]]].

	binaryBlock
		value: offset
		value: sendTable! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 1/12/2018 10:12'!
offsetAndSendTableFor: entryPoint annotation: annotation into: binaryBlock
	"Find the relevant sendTable for a linked-send to entryPoint.  Do this based on the
	 annotation.  c.f. annotationForSendTable:"
	<inline: true>
	| offset sendTable |
	<var: #sendTable type: #'sqInt *'>
	annotation = IsSendCall ifTrue:
		[offset := cmEntryOffset.
		 sendTable := ordinarySendTrampolines] ifFalse:
	[(BytecodeSetHasDirectedSuperSend and: [annotation = IsDirectedSuperSend]) ifTrue:
		[offset := cmNoCheckEntryOffset.
		 sendTable := directedSuperSendTrampolines] ifFalse:
	[(BytecodeSetHasDirectedSuperSend and: [annotation = IsDirectedSuperBindingSend]) ifTrue:
		[offset := cmNoCheckEntryOffset.
		 sendTable := directedSuperBindingSendTrampolines] ifFalse:
	[(NewspeakVM and: [annotation = IsNSSelfSend]) ifTrue:
		[offset := cmEntryOffset.
		 sendTable := selfSendTrampolines] ifFalse:
	[(NewspeakVM and: [annotation = IsNSDynamicSuperSend]) ifTrue:
		[offset := cmEntryOffset.
		 sendTable := dynamicSuperSendTrampolines] ifFalse:
	[self assert: annotation = IsSuperSend.
	 offset := cmNoCheckEntryOffset.
	 sendTable := superSendTrampolines]]]]].

	binaryBlock
		value: offset
		value: sendTable! !

!Cogit methodsFor: 'debug printing' stamp: 'eem 2/20/2018 12:05'!
opcodePrintStringFrom: startIndex to: stopIndex
	^(String streamContents:
		[:s|
		startIndex to: stopIndex do:
			[:i|
			(abstractOpcodes at: i) printStateOn: s]]) allButFirst! !

!Cogit methodsFor: 'generate machine code' stamp: 'eem 7/21/2021 09:09'!
outputInstructionsAt: startAddress
	"Store the generated machine code, answering the last address"
	| absoluteAddress |
	self ensureWritableCodeZone.
	absoluteAddress := startAddress.
	0 to: opcodeIndex - 1 do:
		[:i| | abstractInstruction |
		self maybeBreakGeneratingInstructionWithIndex: i.
		abstractInstruction := self abstractInstructionAt: i.
		self assert: abstractInstruction address = absoluteAddress.
		abstractInstruction outputMachineCodeAt: absoluteAddress.
		absoluteAddress := absoluteAddress + abstractInstruction machineCodeSize].
	^absoluteAddress

	"self abstractInstructionAt: i - 1" "it's usually the preceding instruction that's at fault..."! !

!Cogit methodsFor: 'initialization' stamp: 'rmacnak 12/3/2015 18:59:24'!
outputInstructionsForGeneratedRuntimeAt: startAddress
	"Output instructions generated for one of the generated run-time routines, a trampoline, etc"
	| size endAddress |
	<inline: false>
	self computeMaximumSizes.
	methodLabel address: startAddress. "for addressIsInCurrentCompilation:"
	size := self generateInstructionsAt: startAddress.
	endAddress := self outputInstructionsAt: startAddress.
	self assert: startAddress + size = endAddress.
	methodZoneBase := self alignUptoRoutineBoundary: endAddress.
	backEnd stopsFrom: endAddress to: methodZoneBase - 1.
	self cCode: '' inSmalltalk: [methodZone freeStart: methodZoneBase].
	^startAddress! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'cb 4/20/2016 12:42'!
patchFixupTargetIfNeeded: fixup nextOpcodeIndex: nextOpcodeIndex
	<var: #fixup type: #'BytecodeFixup *'>
	<inline: true>
	 fixup needsFixup ifTrue:
		["There is a fixup for this bytecode.  It must point to the first generated
		   instruction for this bytecode.  If there isn't one we need to add a label."
		 opcodeIndex = nextOpcodeIndex ifTrue: [self Label].
		 fixup targetInstruction: (self abstractInstructionAt: nextOpcodeIndex)].! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 7/5/2021 16:29'!
patchToOpenPICFor: selector numArgs: numArgs receiver: receiver
	"Code entry closed PIC full or miss to an instance of a young class or to a young target method.
	 Attempt to patch the send site to an open PIC.  Answer if the attempt succeeded; in fact it will
	 only return if the attempt failed.
	 The stack looks like:
			receiver
			args
	 sp=>	sender return address"
	<api>
	| oPIC outerReturn extent |
	<var: #oPIC type: #'CogMethod *'>
	outerReturn := coInterpreter stackTop.
	"See if an Open PIC is already available."
	oPIC := methodZone openPICWithSelector: selector.
	oPIC ifNil:
		["otherwise attempt to create an Open PIC."
		oPIC := self cogOpenPICSelector: selector numArgs: numArgs.
		(oPIC asInteger between: MaxNegativeErrorCode and: -1) ifTrue:
			["For some reason the PIC couldn't be generated, most likely a lack of code memory."
			oPIC asInteger = InsufficientCodeSpace ifTrue:
				[coInterpreter callForCogCompiledCodeCompaction].
			^false]].
	"Relink the send site to the pic.  Reset the cache tag to the selector, for the
	 benefit of the cacheTag assert check in checkIfValidOopRef:pc:cogMethod: et al."
	self ensureWritableCodeZone.
	extent := backEnd
				rewriteInlineCacheAt: outerReturn
				tag: (backEnd
						inlineCacheValueForSelector: selector
						in: coInterpreter mframeHomeMethodExport)
				target: oPIC asInteger + cmEntryOffset.
	backEnd
		flushICacheFrom: outerReturn asUnsignedInteger - extent to: outerReturn asUnsignedInteger;
		flushICacheFrom: oPIC asUnsignedInteger to: oPIC asUnsignedInteger + openPICSize.
	"Jump into the oPIC at its entry"
	coInterpreter executeCogMethod: oPIC fromLinkedSendWithReceiver: receiver.
	"NOTREACHED"
	^true! !

!Cogit methodsFor: 'accessing' stamp: 'eem 01/10/2016 10:01'!
picAbortDiscriminatorValue
	"This value is used to decide between MNU processing
	 or interpretation in the closed PIC aborts."
	^0! !

!Cogit methodsFor: 'debugging' stamp: 'eem 5/27/2015 09:03'!
picInterpretAbortOffset
	"Answer the start of the abort sequence for invoking the interpreter in a closed PIC."
	^self interpretOffset
	 - (backEnd hasLinkRegister
		ifTrue: [backEnd pushLinkRegisterByteSize + backEnd callInstructionByteSize]
		ifFalse: [backEnd callInstructionByteSize])! !

!Cogit methodsFor: 'profiling primitives' stamp: 'nice 7/8/2016 02:18'!
positiveMachineIntegerFor: value
	<var: #value type: #'usqIntptr_t'>
	<inline: true>
	^objectMemory wordSize = 8
		ifTrue: [coInterpreter positive64BitIntegerFor: value]
		ifFalse: [coInterpreter positive32BitIntegerFor: value]! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 4/23/2016 11:25'!
preenMethodLabel
	"The methodLabel serves as the reference to the start of the current code object
	 being produced (CMMethod, CMClosedPIC etc), but it also carries type flags for
	 the frame method field, set via the labelOffset.  So we must clean the flags on each
	 compilation to avoid stale lags being left behind from previous compilations."
	<inline: true>
	methodLabel setLabelOffset: 0! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/18/2015 10:56'!
previousInstruction
	<returnTypeC: #'AbstractInstruction *'>
	self assert: opcodeIndex > 0.
	^self abstractInstructionAt: opcodeIndex - 1! !

!Cogit methodsFor: 'printing' stamp: 'HolgerHansPeterFreyther 2/20/2017 21:07'!
print: aString
	<cmacro: '(aString) printf("%s", aString)'>
	coInterpreter transcript print: aString! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/14/2014 09:56'!
print: descriptor IsBackwardBranch: isBackwardBranch Mcpc: mcpc Bcpc: bcpc on: aStream
	<doNotGenerate>
	aStream ensureCr.
	mcpc printOn: aStream base: 16.
	aStream
		space; tab;
		print: (isBackwardBranch ifTrue: [bcpc] ifFalse: [bcpc + descriptor numBytes]);
		cr; flush.
	^0! !

!Cogit methodsFor: 'printing' stamp: 'eem 2/24/2020 08:21'!
printCogMethodFor: address
	<api>
	<var: #address type: #'void *'>
	(methodZone methodFor: address)
		ifNil: [(self codeEntryFor: address)
					ifNil: [coInterpreter print: 'not a method'; cr]
					ifNotNil: [coInterpreter print: 'trampoline '; print: (self codeEntryNameFor: address); cr]]
		ifNotNil: [:cogMethod| coInterpreter printCogMethod: cogMethod]! !

!Cogit methodsFor: 'printing' stamp: 'eem 2/24/2020 09:44'!
printCogMethodHeaderFor: address
	<doNotGenerate>
	(methodZone methodFor: address)
		ifNil: [coInterpreter print: 'not a method'; cr]
		ifNotNil: [:cogMethod| self printMethodHeader: cogMethod on: coInterpreter transcript]! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 9/27/2010 14:44'!
printInstructions
	<doNotGenerate>
	^printInstructions! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 3/30/2009 12:31'!
printInstructions: aBoolean
	<doNotGenerate>
	printInstructions := aBoolean.
	singleStep := singleStep or: [aBoolean]! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 9/6/2016 12:05'!
printMapEntry: annotation mcpc: mcpc args: tupleOfStreamCodeRangesAndMethod
	"Print the Map entry's mcpc, its annotation and the corresponding bytecode pc, if any."
	<doNotGenerate>
	| printHex |
	printHex := disassemblingMethod
					ifNil: [[:pc| pc hex]]
					ifNotNil: [[:pc| '.+', (pc - disassemblingMethod asInteger printStringBase: 16 length: 4 padded: true)]].
	[:aStream :codeRanges :cogMethod|
	self startMcpcAndCogMethodForMcpc: mcpc in: cogMethod do:
		[:startmcpc :subMethod| | name codeRange |
		"Find the start of the block by searching the code ranges."
		codeRange := codeRanges detect: [:range| range includes: mcpc] ifNone: [codeRanges detect: [:range| range last + 1 = mcpc]].
		codeRange first = mcpc ifTrue:
			[aStream nextPutAll: 'startpc: '; print: codeRange startpc; cr].
		aStream
			next: 2 put: Character space;
			nextPutAll: (printHex value: mcpc);  space;
			nextPutAll: (name := self class annotationConstantNames at: annotation + 1);
			next: 20 - name size put: Character space;
			nextPut: $(;
			nextPutAll: (printHex value: (self findMapLocationForMcpc: mcpc inMethod: cogMethod)).
		(self isPCMappedAnnotation: annotation) ifTrue:
			[aStream
				nextPutAll: ', bc: ';
				print: (self bytecodePCFor: mcpc startBcpc: codeRange startpc in: subMethod)].
		(self isSendAnnotation: annotation) ifTrue:
			[| sel |
			sel := self selectorForSendAt: mcpc annotation: annotation in: cogMethod methodObject.
			sel isInteger ifTrue:
				[sel := self lookupAddress: sel].
			sel isString ifTrue:
				[aStream space; nextPutAll: sel]].
		aStream
			nextPut: $);
			cr; flush]]
		valueWithArguments: tupleOfStreamCodeRangesAndMethod.
	^0! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 9/9/2019 15:35'!
printMethodHeader: cogMethod on: aStream
	<doNotGenerate>
	self cCode: ''
		inSmalltalk:
			[cogMethod isInteger ifTrue:
				[^self printMethodHeader: (self cogMethodOrBlockSurrogateAt: cogMethod) on: aStream]].
	aStream ensureCr.
	cogMethod asInteger printOn: aStream base: 16.
	cogMethod cmType = CMMethod ifTrue:
		[aStream crtab; nextPutAll: 'objhdr: '.
		cogMethod objectHeader printOn: aStream base: 16].
	cogMethod cmType = CMBlock ifTrue:
		[aStream crtab; nextPutAll: 'homemth: '.
		cogMethod cmHomeMethod asUnsignedInteger printOn: aStream base: 16.
		aStream
			nextPutAll: ' (offset '; print: cogMethod homeOffset; nextPut: $);
			crtab; nextPutAll: 'startpc: '; print: cogMethod startpc].
	aStream
		crtab; nextPutAll: 'nArgs: ';	print: cogMethod cmNumArgs;
		tab;    nextPutAll: 'type: ';	print: cogMethod cmType.
	(cogMethod cmType ~= 0 and: [cogMethod cmType ~= CMBlock]) ifTrue:
		[aStream crtab; nextPutAll: 'blksiz: '.
		cogMethod blockSize printOn: aStream base: 16.
		cogMethod cmType = CMMethod ifTrue:
			[aStream crtab; nextPutAll: 'method: '.
			 cogMethod methodObject printOn: aStream base: 16.
			 aStream crtab; nextPutAll: 'mthhdr: '.
			 cogMethod methodHeader printOn: aStream base: 16].
		aStream crtab; nextPutAll: 'selctr: '.
		cogMethod selector printOn: aStream base: 16.
		(coInterpreter lookupAddress: cogMethod selector) ifNotNil:
			[:string| aStream nextPut: $=; nextPutAll: string].
		cogMethod selector = objectMemory nilObject ifTrue:
			[aStream space; nextPut: $(; nextPutAll: (coInterpreter stringOf: (coInterpreter maybeSelectorOfMethod: cogMethod methodObject)); nextPut: $)]. 
		cogMethod cmType = CMMethod ifTrue:
			[aStream crtab; nextPutAll: 'blkentry: '.
			 cogMethod blockEntryOffset printOn: aStream base: 16.
			 cogMethod blockEntryOffset ~= 0 ifTrue:
				[aStream nextPutAll: ' => '.
				 cogMethod asInteger + cogMethod blockEntryOffset printOn: aStream base: 16]]].
	cogMethod cmType = CMClosedPIC
		ifTrue:
			[aStream crtab; nextPutAll: 'cPICNumCases: '.
			 cogMethod cPICNumCases printOn: aStream base: 16.
			 aStream tab; nextPutAll: 'cpicHasMNUCase: ';
			 nextPutAll: (cogMethod cpicHasMNUCase ifTrue: ['yes'] ifFalse: ['no'])]
		ifFalse:
			[aStream crtab; nextPutAll: 'stackCheckOffset: '.
			 cogMethod stackCheckOffset printOn: aStream base: 16.
			 cogMethod stackCheckOffset > 0 ifTrue:
				[aStream nextPut: $/.
				 cogMethod asInteger + cogMethod stackCheckOffset printOn: aStream base: 16].
			cogMethod cmType = CMBlock
				ifTrue:
					[aStream
						crtab;
						nextPutAll: 'cbUsesInstVars ';
						nextPutAll: (cogMethod cbUsesInstVars ifTrue: ['yes'] ifFalse: ['no'])]
				ifFalse:
					[aStream
						crtab;
						nextPutAll: 'cmRefersToYoung: ';
						nextPutAll: (cogMethod cmRefersToYoung ifTrue: ['yes'] ifFalse: ['no']);
						tab;
						nextPutAll: 'cmHasMovableLiteral: ';
						nextPutAll: (cogMethod cmHasMovableLiteral ifTrue: ['yes'] ifFalse: ['no']);
						tab;
						nextPutAll: 'cmIsFullBlock: ';
						nextPutAll: (cogMethod cmIsFullBlock ifTrue: ['yes'] ifFalse: ['no'])].
			cogMethod cmType = CMMethod ifTrue:
				[([cogMethod nextMethodOrIRCs] on: MessageNotUnderstood do: [:ex| nil]) ifNotNil:
					[:nmoircs| aStream crtab; nextPutAll: 'nextMethodOrIRCs: '.
						nmoircs = 0 ifTrue: [aStream print: nmoircs] ifFalse: [coInterpreter printHex: nmoircs]].
				 ([cogMethod counters] on: MessageNotUnderstood do: [:ex| nil]) ifNotNil:
					[:cntrs| aStream crtab; nextPutAll: 'counters: '.
						cntrs = 0 ifTrue: [aStream print: cntrs] ifFalse: [coInterpreter printHex: cntrs]]]].
	aStream cr; flush! !

!Cogit methodsFor: 'printing' stamp: 'nice 12/27/2018 15:34'!
printNum: n
	<cmacro: '(n) printf("%" PRIdSQINT, (sqInt) (n))'>
	coInterpreter transcript printNum: n! !

!Cogit methodsFor: 'debugging' stamp: 'eem 11/15/2012 10:15'!
printOnTrace
	<api>
	<cmacro: '() (traceFlags & 1)'>
	^(traceFlags bitAnd: 1) ~= 0! !

!Cogit methodsFor: 'method map' stamp: 'eem 1/12/2018 10:22'!
printPCMapPairsFor: cogMethod
	<doNotGenerate>
	"<api>
	<var: 'cogMethod' type: #'CogMethod *'>
	<var: 'mapByte' type: #'unsigned char'>"
	| mcpc map mapByte annotation value |
	mcpc := self firstMappedPCFor: cogMethod.
	map := self mapStartFor: cogMethod.
	[(mapByte := objectMemory byteAt: map) ~= MapEnd] whileTrue:
		[annotation := mapByte >> AnnotationShift.
		 annotation = IsAnnotationExtension
			ifTrue:
				[value := (mapByte bitAnd: DisplacementMask) + IsSendCall]
			ifFalse:
				[value := annotation.
				 mcpc := mcpc + (backEnd codeGranularity
									* (annotation = IsDisplacementX2N
										ifTrue: [mapByte - DisplacementX2N << AnnotationShift]
										ifFalse: [mapByte bitAnd: DisplacementMask]))].
		 coInterpreter
			printHexnp: map;
		 	print: ': '.
		 self
			cCode: [self print: '%02x' f: mapByte]
			inSmalltalk:
				[mapByte < 16 ifTrue:
					[coInterpreter putchar: $0].
				 coInterpreter printHexnp: mapByte].
		 coInterpreter
		 	printChar: $ ;
			printNum: annotation;
			print: ' ('.
		 (NewspeakVM
		  and: [value = IsNSSendCall
			     or: [value between: IsNSSendCall and: IsNSImplicitReceiverSend]]) ifTrue:
			[value
				caseOf: {
					[IsNSSendCall]				->	[coInterpreter print: 'NSSendCall'].
					[IsNSSelfSend]				->	[coInterpreter print: 'NSSelfSend'].
					[IsNSDynamicSuperSend]	->	[coInterpreter print: 'NSDynamicSuperSend'].
					[IsNSImplicitReceiverSend]	->	[coInterpreter print: 'NSImplicitReceiverSend'] }] ifFalse:
		 [(BytecodeSetHasDirectedSuperSend
		    and: [value between: IsDirectedSuperSend and: IsDirectedSuperBindingSend]) ifTrue:
			[value
				caseOf: {
					[IsDirectedSuperSend]			->	[coInterpreter print: 'DirectedSuperSend'].
					[IsDirectedSuperBindingSend]	->	[coInterpreter print: 'DirectedSuperBindingSend'] }] ifFalse:
		 [value
			caseOf: {
				[IsDisplacementX2N]		->	[coInterpreter print: 'DisplacementX2N'].
				[IsAnnotationExtension]	->	[coInterpreter print: 'AnnotationExtension'].
				[IsObjectReference]			->	[coInterpreter print: 'ObjectReference'].
				[IsAbsPCReference]			->	[coInterpreter print: 'AbsPCReference'].
				[HasBytecodePC]			->	[coInterpreter print: 'HasBytecodePC'].
				[IsRelativeCall]				->	[coInterpreter print: 'RelativeCall'].
				[IsSendCall]				->	[coInterpreter print: 'SendCall'].
				[IsSuperSend]				->	[coInterpreter print: 'SuperSend'] }
			otherwise: [coInterpreter print: '??? '; printHexnp: value]]].
		 coInterpreter
			print: ') ';
			printHexnp: (mapByte bitAnd: DisplacementMask);
			printChar: $ ;
			putchar: $@;
		 printHex: mcpc;
		 cr;
		 flush.
		 map := map - 1]! !

!Cogit methodsFor: 'method map' stamp: 'eem 4/15/2016 10:09'!
printPCMapPairsFor: cogMethod on: aStream
	<doNotGenerate>
	<inline: true>
	| mcpc map mapByte annotation |
	mcpc := self firstMappedPCFor: cogMethod.
	map := self mapStartFor: cogMethod.
	[(mapByte := objectMemory byteAt: map) ~= MapEnd] whileTrue:
		[annotation := mapByte >> AnnotationShift.
		 annotation ~= IsAnnotationExtension ifTrue:
			[mcpc := mcpc + (backEnd codeGranularity
								* (annotation = IsDisplacementX2N
									ifTrue: [mapByte - DisplacementX2N << AnnotationShift]
									ifFalse: [mapByte bitAnd: DisplacementMask]))].
		 aStream ensureCr.
		 map printOn: aStream base: 16.
		 aStream nextPutAll: ': '.
		 mapByte printOn: aStream base: 16 length: 2 padded: true.
		 aStream space.
		 annotation printOn: aStream base: 16.
		 aStream nextPutAll: ' ('; print: (AnnotationConstantNames at: annotation + 1); nextPutAll: ') '.
		 (mapByte bitAnd: DisplacementMask) printOn: aStream base: 16.
		 aStream space.
		 aStream nextPut: $@.
		 mcpc printOn: aStream base: 16.
		 aStream flush.
		 map := map - 1]! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 5/31/2016 08:57'!
printRegisterMapOn: aStream
	<doNotGenerate>
	| map n |
	map := backEnd generalPurposeRegisterMap.
	n := 0.
	(map keys asSortedCollection: [:a :b| (map at: a) < (map at: b)])
		do:	[:regName| | abstractName |
			abstractName := backEnd nameForRegister: (map at: regName).
			aStream nextPutAll: abstractName; nextPutAll: ' => '; nextPutAll: regName]
		separatedBy: [(n := n + 1) \\ 4 = 0 ifTrue: [aStream cr] ifFalse: [aStream tab]].
	aStream cr; flush! !

!Cogit methodsFor: 'debugging' stamp: 'eem 9/27/2010 14:44'!
printRegisters
	<doNotGenerate>
	^printRegisters! !

!Cogit methodsFor: 'debugging' stamp: 'eem 1/17/2009 12:21'!
printRegisters: aBoolean
	<doNotGenerate>
	printRegisters := aBoolean! !

!Cogit methodsFor: 'debugging' stamp: 'eem 6/7/2014 14:58'!
printTrampolineTable
	<api>
	0 to: trampolineTableIndex - 1 by: 2 do:
		[:i|
		coInterpreter
			printHex: (trampolineAddresses at: i + 1) asInteger;
			print: ': ';
			print: (self cCoerceSimple: (trampolineAddresses at: i) to: #'char *');
			cr]! !

!Cogit methodsFor: 'accessing' stamp: 'eem 2/11/2009 09:07'!
processor
	<doNotGenerate>
	^processor! !

!Cogit methodsFor: 'initialization' stamp: 'eem 1/7/2016 13:36'!
processorHasDivQuoRemAndMClassIsSmallInteger
	^backEnd canDivQuoRem and: [self mclassIsSmallInteger]! !

!Cogit methodsFor: 'initialization' stamp: 'eem 1/7/2016 13:41'!
processorHasDoublePrecisionFloatingPointSupport
	<option: #DPFPReg0>
	<inline: true>
	^backEnd hasDoublePrecisionFloatingPointSupport! !

!Cogit methodsFor: 'initialization' stamp: 'eem 1/7/2016 13:37'!
processorHasMultiplyAndMClassIsSmallInteger
	^backEnd canMulRR and: [self mclassIsSmallInteger]! !

!Cogit methodsFor: 'profiling primitives' stamp: 'eem 5/31/2018 18:16'!
profileDataFor: cogMethod withDetails: withDetails
	"Answers characteristic data for the type of the cogMethod, answering
		a CompiledMethod for a compiled method,
		a selector for an open PIC
		if withDetails then an array containing a selector followed by pairs of class and target method for a closed PIC, otherwise simply a selector."
	<inline: true>
	<var: #cogMethod type: #'CogMethod *'>
	^cogMethod cmType = CMMethod 
		ifTrue: [cogMethod methodObject]
		ifFalse: [(withDetails and: [cogMethod cmType = CMClosedPIC])
					ifTrue: [self createCPICData: cogMethod]
					ifFalse: [cogMethod selector]]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 2/24/2020 21:41'!
promptForBreakPC
	<doNotGenerate>
	| s first bkpt |
	s := UIManager default request: 'Break pc (hex, + to add, - to remove)'.
	s := s withBlanksTrimmed.
	s isEmpty ifTrue: [^self].
	('+-' includes: s first) ifTrue: [first := s first. s := s allButFirst].
	(s isEmpty and: [first = $-]) ifTrue:
		[^self breakPC: nil].
	bkpt := (s includes: $r)
			ifTrue:
				[Number readFrom: s readStream]
			ifFalse:
				[(#('0x' '-0x') detect: [:prefix| s beginsWith: prefix] ifNone: []) ifNotNil:
					[:prefix|
					s := s allButFirst: prefix size.
					prefix first = $- ifTrue: [s := '-', s]].
				Integer readFrom: s readStream base: 16].
	((methodZone addressIsLikelyCogMethod: bkpt)
	 and: [UIManager confirm: 'pc is method; break anywhere within method?']) ifTrue:
		[bkpt := methodZone methodFor: bkpt].
	first = $+ ifTrue:
		[^self breakPC: (breakPC addBreakpoint: bkpt)].
	first = $- ifTrue:
		[^self breakPC: (breakPC removeBreakpoint: bkpt)].
	self breakPC: bkpt! !

!Cogit methodsFor: 'simulation processor access' stamp: 'eem 7/4/2020 12:05'!
provideContextForCurrentInstruction
	printRegisters ifTrue:
		[processor printRegistersOn: coInterpreter transcript].
	(printInstructions
	 and: [self addressIsInCodeZone: processor pc]) ifTrue:
		[self disassembleFrom: processor pc to: processor pc]! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 9/9/2019 15:55'!
pruneYoungReferrers
	<doNotGenerate>
	methodZone pruneYoungReferrers! !

!Cogit methodsFor: 'debugging' stamp: 'eem 11/15/2012 10:15'!
recordBlockTrace
	<api>
	<cmacro: '() (traceFlags & 4)'>
	^(traceFlags bitAnd: 4) ~= 0! !

!Cogit methodsFor: 'debugging' stamp: 'eem 11/15/2012 10:15'!
recordEventTrace
	<api>
	<cmacro: '() (traceFlags & 16)'>
	^(traceFlags bitAnd: 16) ~= 0! !

!Cogit methodsFor: 'debugging' stamp: 'eem 8/3/2021 18:29'!
recordFastCCallPrimTrace
	<api>
	<cmacro: '() (traceFlags & 512)'>
	^(traceFlags bitAnd: 512) ~= 0! !

!Cogit methodsFor: 'initialization' stamp: 'eem 11/10/2020 19:24'!
recordGeneratedRunTime: aString address: address
	<var: #aString type: #'char *'>
	self assert: trampolineTableIndex + 2 <= (NumTrampolines * 2).
	trampolineAddresses
		at: trampolineTableIndex put: aString;
		at: trampolineTableIndex + 1 put: (self cCoerceSimple: address to: #'char *').
	trampolineTableIndex := trampolineTableIndex + 2

	"self printTrampolineTable"! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 2/24/2020 15:53'!
recordInstruction: thing
	<doNotGenerate>
	lastNInstructions addLast: thing.
	[lastNInstructions size > 320"160""80"] whileTrue:
		[lastNInstructions removeFirst.
		 lastNInstructions size * 2 > lastNInstructions capacity ifTrue:
			[lastNInstructions makeRoomAtLast]].
	^thing! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 5/17/2015 10:21'!
recordLastInstruction
	<doNotGenerate>
	| inst pc |
	(EagerInstructionDecoration or: [printInstructions])
		ifTrue:
			[inst := processor
						disassembleNextInstructionIn: coInterpreter memory
						for: self.
			 printInstructions ifTrue:
				[pc := Integer readFrom: (ReadStream on: inst from: 1 to: (inst indexOf: $:) - 1) base: 16.
				 (self relativeLabelForPC: pc) ifNotNil:
					[:label| inst := inst, ' ', label]]]
		ifFalse:
			[inst := processor
						disassembleNextInstructionIn: coInterpreter memory
						for: nil].
	^self recordInstruction: inst! !

!Cogit methodsFor: 'debugging' stamp: 'eem 11/27/2012 14:59'!
recordOverflowTrace
	<api>
	<cmacro: '() (traceFlags & 32)'>
	^(traceFlags bitAnd: 32) ~= 0! !

!Cogit methodsFor: 'debugging' stamp: 'eem 11/15/2012 10:15'!
recordPrimTrace
	<api>
	<cmacro: '() (traceFlags & 8)'>
	^(traceFlags bitAnd: 8) ~= 0! !

!Cogit methodsFor: 'debugging' stamp: 'eem 4/28/2010 14:08'!
recordPrimTraceFunc
	"This one for C support code."
	<api>
	^self recordPrimTrace! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 8/20/2021 23:03'!
recordProcessing
	| pc inst |
	instructionProfile ifNotNil:
		[:ip| ip at: (pc := processor pc) put: (ip at: pc) + 1].
	self recordRegisters.
	inst := self recordLastInstruction.
	"Set RRRName ito the selector that accesses ReceiverResultReg (RRR) to alter instruction printing to add the value of RRR as a suffix
		(RRRName := #rdx)
		(RRRName := #edx)
		(RRRName := nil)"
	printRegisters ifTrue:
		[RRRName ifNil: [processor printRegistersOn: coInterpreter transcript].
		 printInstructions ifFalse:
			[coInterpreter transcript cr]].
	printInstructions ifTrue:
		[printRegisters ifTrue:
			[coInterpreter transcript cr].
		 coInterpreter transcript nextPutAll: inst.
		 RRRName ifNotNil:
			[coInterpreter transcript space; nextPutAll: RRRName; space.
			 (processor perform: RRRName) printOn: coInterpreter transcript base: 16 length: 8 padded: false].
		 coInterpreter transcript cr; flush]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 9/30/2009 19:40'!
recordRegisters
	<doNotGenerate>
	self recordInstruction: processor integerRegisterState
	"self recordInstruction: processor registerState"! !

!Cogit methodsFor: 'initialization' stamp: 'eem 6/16/2015 12:18'!
recordRunTimeObjectReferences
	<var: #instruction type: #'AbstractInstruction *'>
	0 to: opcodeIndex - 1 do:
		[:i| | instruction |
		instruction := self abstractInstructionAt: i.
		instruction annotation = IsObjectReference ifTrue:
			[self assert: runtimeObjectRefIndex < NumObjRefsInRuntime.
			 self assert: hasYoungReferent not.
			 hasYoungReferent ifTrue:
				[self error: 'attempt to generate run-time routine containing young object reference.  Cannot initialize Cogit run-time.'].
			 objectReferencesInRuntime
				at: runtimeObjectRefIndex
				put: instruction mapEntryAddress asUnsignedInteger.
			 runtimeObjectRefIndex := runtimeObjectRefIndex + 1]]! !

!Cogit methodsFor: 'debugging' stamp: 'eem 11/15/2012 10:15'!
recordSendTrace
	<api>
	<cmacro: '() (traceFlags & 2)'>
	^(traceFlags bitAnd: 2) ~= 0! !

!Cogit methodsFor: 'register management' stamp: 'eem 8/5/2021 07:18'!
registerMaskFor: reg
	"N.B. (self registerMaskFor: NoReg) = 0"
	<inline: true>
	^1 bitShift: reg! !

!Cogit methodsFor: 'register management' stamp: 'eem 3/3/2016 12:32'!
registerMaskFor: reg1 and: reg2
	<inline: true>
	^1 << reg1 bitOr: 1 << reg2! !

!Cogit methodsFor: 'register management' stamp: 'eem 3/3/2016 12:32'!
registerMaskFor: reg1 and: reg2 and: reg3
	<inline: true>
	^(1 << reg1 bitOr: 1 << reg2) bitOr: 1 << reg3! !

!Cogit methodsFor: 'register management' stamp: 'eem 3/3/2016 12:32'!
registerMaskFor: reg1 and: reg2 and: reg3 and: reg4
	<inline: true>
	^((1 << reg1 bitOr: 1 << reg2) bitOr: 1 << reg3) bitOr: 1 << reg4! !

!Cogit methodsFor: 'register management' stamp: 'eem 3/3/2016 12:32'!
registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5
	<inline: true>
	^(((1 << reg1 bitOr: 1 << reg2) bitOr: 1 << reg3) bitOr: 1 << reg4) bitOr: 1 << reg5! !

!Cogit methodsFor: 'register management' stamp: 'eem 3/3/2016 12:32'!
registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6
	<inline: true>
	^((((1 << reg1 bitOr: 1 << reg2) bitOr: 1 << reg3) bitOr: 1 << reg4) bitOr: 1 << reg5) bitOr: 1 << reg6! !

!Cogit methodsFor: 'register management' stamp: 'eem 3/3/2016 12:32'!
registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6 and: reg7
	<inline: true>
	^(((((1 << reg1 bitOr: 1 << reg2) bitOr: 1 << reg3) bitOr: 1 << reg4) bitOr: 1 << reg5) bitOr: 1 << reg6) bitOr: 1 << reg7! !

!Cogit methodsFor: 'register management' stamp: 'eem 3/3/2016 12:32'!
registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6 and: reg7 and: reg8
	<inline: true>
	^((((((1 << reg1 bitOr: 1 << reg2) bitOr: 1 << reg3) bitOr: 1 << reg4) bitOr: 1 << reg5) bitOr: 1 << reg6) bitOr: 1 << reg7) bitOr: 1 << reg8! !

!Cogit methodsFor: 'register management' stamp: 'eem 3/3/2016 12:33'!
registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6 and: reg7 and: reg8 and: reg9
	<inline: true>
	^(((((((1 << reg1 bitOr: 1 << reg2) bitOr: 1 << reg3) bitOr: 1 << reg4) bitOr: 1 << reg5) bitOr: 1 << reg6) bitOr: 1 << reg7) bitOr: 1 << reg8) bitOr: 1 << reg9! !

!Cogit methodsFor: 'register management' stamp: 'eem 3/3/2016 12:33'!
registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6 and: reg7 and: reg8 and: reg9 and: reg10
	<inline: true>
	^((((((((1 << reg1 bitOr: 1 << reg2) bitOr: 1 << reg3) bitOr: 1 << reg4) bitOr: 1 << reg5) bitOr: 1 << reg6) bitOr: 1 << reg7) bitOr: 1 << reg8) bitOr: 1 << reg9) bitOr: 1 << reg10! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 12/11/2015 10:15'!
relativeBaseForDisassemblyInto: aBlock
	<doNotGenerate>
	disassemblingMethod ifNotNil:
		[aBlock value: disassemblingMethod asInteger value: '.']! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 12/5/2010 11:58'!
relativeLabelForPC: pc 
	<doNotGenerate>
	pc < methodZone limitZony ifFalse:
		[^nil].
	pc < methodZoneBase ifTrue:
		[^(self codeEntryNameFor: pc) ifNotNil:
			[:name| name, '+', (pc - (self codeEntryFor: pc)) printString]].
	^self lookupAddress: pc! !

!Cogit methodsFor: 'compaction' stamp: 'eem 7/25/2021 10:48'!
relocateCallsAndSelfReferencesInMethod: cogMethod
	<var: #cogMethod type: #'CogMethod *'>
	| refDelta callDelta |
	refDelta := cogMethod objectHeader.
	callDelta := backEnd zoneCallsAreRelative ifTrue: [refDelta] ifFalse: [0].
	
	self assert: (cogMethod cmType = CMMethod or: [cogMethod cmType = CMOpenPIC]).
	self assert: (backEnd callTargetFromReturnAddress: cogMethod asInteger + missOffset)
				= (cogMethod cmType = CMMethod
					ifTrue: [self methodAbortTrampolineFor: cogMethod cmNumArgs]
					ifFalse: [self picAbortTrampolineFor: cogMethod cmNumArgs]).
	backEnd relocateCallBeforeReturnPC: cogMethod asInteger + missOffset by: callDelta negated.
	self mapFor: cogMethod
		performUntil: #relocateIfCallOrMethodReference:mcpc:delta:
		arg: refDelta asVoidPointer! !

!Cogit methodsFor: 'compaction' stamp: 'rmacnak 12/12/2015 11:42:50'!
relocateCallsInClosedPIC: cPIC
	<var: #cPIC type: #'CogMethod *'>
	| refDelta callDelta pc entryPoint targetMethod |
	<var: #targetMethod type: #'CogMethod *'>
	refDelta := cPIC objectHeader.
	callDelta := backEnd zoneCallsAreRelative ifTrue: [refDelta] ifFalse: [0].
	
	self assert: (backEnd callTargetFromReturnAddress: cPIC asInteger + missOffset)
					= (self picAbortTrampolineFor: cPIC cmNumArgs).
	backEnd relocateCallBeforeReturnPC: cPIC asInteger + missOffset by: callDelta negated.

	pc := cPIC asInteger + firstCPICCaseOffset.
	1 to: cPIC cPICNumCases do:
		[:i|
		pc := self addressOfEndOfCase: i inCPIC: cPIC.
		entryPoint := i = 1
						ifTrue: [backEnd jumpLongTargetBeforeFollowingAddress: pc]
						ifFalse: [backEnd jumpLongConditionalTargetBeforeFollowingAddress: pc].
		(cPIC containsAddress: entryPoint) 
			ifTrue: 
			["Interpret/MNU"
			backEnd zoneCallsAreRelative ifFalse: [
				i = 1 ifTrue:
					[backEnd
						relocateJumpLongBeforeFollowingAddress: pc
						by: refDelta]
					ifFalse:
					[backEnd
						relocateJumpLongConditionalBeforeFollowingAddress: pc
						by: refDelta]]]
			ifFalse:
			[targetMethod := self cCoerceSimple: entryPoint - cmNoCheckEntryOffset to: #'CogMethod *'.
			 self assert: targetMethod cmType = CMMethod.
			 i = 1 ifTrue:
				[backEnd
					relocateJumpLongBeforeFollowingAddress: pc
					by: (callDelta - targetMethod objectHeader) negated]
				ifFalse:
				[backEnd
					relocateJumpLongConditionalBeforeFollowingAddress: pc
					by: (callDelta - targetMethod objectHeader) negated]]].
	self assert: cPIC cPICNumCases > 0.

	"Finally relocate the load of the PIC and the jump to the overflow routine ceCPICMiss:receiver:"
	backEnd relocateMethodReferenceBeforeAddress: (self addressOfEndOfCase: 2 inCPIC: cPIC)+ backEnd loadPICLiteralByteSize by: refDelta.
	backEnd relocateJumpLongBeforeFollowingAddress: cPIC asInteger + cPICEndOfCodeOffset by: callDelta negated! !

!Cogit methodsFor: 'compaction' stamp: 'eem 7/25/2021 10:50'!
relocateIfCallOrMethodReference: annotation mcpc: mcpc delta: refDeltaArg
	<var: #mcpc type: #'char *'>
	<var: #refDeltaArg type: #'CogMethod *'> "To placate the C static type system..."
	| refDelta callDelta entryPoint targetMethod unlinkedRoutine |
	<var: #targetMethod type: #'CogMethod *'>
	<var: #nsSendCache type: #'NSSendCache *'>

	refDelta := self cCoerceSimple: refDeltaArg to: #sqInt.
	callDelta := backEnd zoneCallsAreRelative ifTrue: [refDelta] ifFalse: [0].
	
	NewspeakVM ifTrue:
		[| nsSendCache |
		 annotation = IsNSSendCall ifTrue:
			["Retrieve the send cache before relocating the stub call. Fetching the send
			  cache asserts the stub call points below all the cogged methods, but
			  until this method is actually moved, the adjusted stub call may appear to
			  point to somewhere in the method zone."
			nsSendCache := self nsSendCacheFromReturnAddress: mcpc asInteger.

			"Fix call to trampoline. This method is moving [delta] bytes, and calls are
			 relative, so adjust the call by -[delta] bytes"
			backEnd relocateCallBeforeReturnPC: mcpc asInteger by: callDelta negated.

			nsSendCache target ~= 0 ifTrue: "Send is linked"
				[entryPoint := nsSendCache target.
				targetMethod := self cCoerceSimple: entryPoint - cmNoCheckEntryOffset to: #'CogMethod *'.
				targetMethod cmType = CMMethod
					ifTrue: "send target not freed; just relocate. The cache has an absolute
							target, so only adjust by the target method's displacement."
						[nsSendCache target: entryPoint + targetMethod objectHeader]
					ifFalse: "send target was freed, unlink"
						[self voidNSSendCache: nsSendCache]].
			^0]].

	(self isPureSendAnnotation: annotation) ifTrue:
		[entryPoint := backEnd callTargetFromReturnAddress: mcpc asInteger.
		entryPoint <= methodZoneBase ifTrue: "send is not linked; just relocate"
			[backEnd relocateCallBeforeReturnPC: mcpc asInteger by: callDelta negated.
			 ^0].
		"It's a linked send; find which kind."
		self
			offsetAndSendTableFor: entryPoint
			annotation: annotation
			into: [:offset :sendTable|
				 targetMethod := self cCoerceSimple: entryPoint - offset to: #'CogMethod *'.
				 targetMethod cmType ~= CMFree ifTrue: "send target not freed; just relocate."
					[backEnd
						relocateCallBeforeReturnPC: mcpc asInteger
						by: (callDelta - targetMethod objectHeader) negated.
					 SistaVM ifTrue: "See comment in planCompaction"
						[methodZone restorePICUsageCount: targetMethod].
					 ^0].
				 "Target was freed; map back to an unlinked send; but include this method's reocation"
				 unlinkedRoutine := sendTable at: (targetMethod cmNumArgs min: NumSendTrampolines - 1).
				 unlinkedRoutine := unlinkedRoutine - callDelta.
				 backEnd
					rewriteInlineCacheAt: mcpc asInteger
					tag: (backEnd inlineCacheValueForSelector: targetMethod selector in: enumeratingCogMethod)
					target: unlinkedRoutine.
				 ^0]].

	annotation = IsRelativeCall ifTrue:
		[backEnd relocateCallBeforeReturnPC: mcpc asInteger by: callDelta negated.
		 ^0].

	annotation = IsAbsPCReference ifTrue:
		[backEnd relocateMethodReferenceBeforeAddress: mcpc asInteger by: refDelta].

	^0 "keep scanning"! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/25/2021 10:47'!
remapIfObjectRef: annotation pc: mcpc hasYoung: hasYoungPtr
	<var: #mcpc type: #'char *'>
	<var: #hasYoungPtr type: #'CogMethod *'> "to placate the C static type system..."

	<var: #targetMethod type: #'CogMethod *'>
	<var: #nsSendCache type: #'NSSendCache *'>
	annotation = IsObjectReference ifTrue:
		[| literal mappedLiteral |
		 literal := literalsManager fetchLiteralAtAnnotatedAddress: mcpc asUnsignedInteger using: backEnd.
		 (objectRepresentation couldBeObject: literal) ifTrue:
			[mappedLiteral := objectRepresentation remapObject: literal.
			 literal ~= mappedLiteral ifTrue:
				[self setCodeModified.
				 literalsManager storeLiteral: mappedLiteral atAnnotatedAddress: mcpc asUnsignedInteger using: backEnd].
			 (hasYoungPtr ~= 0
			  and: [objectMemory isYoung: mappedLiteral]) ifTrue:
				[(self cCoerceSimple: hasYoungPtr to: #'sqInt *') at: 0 put: true]]].

	NewspeakVM ifTrue: [annotation = IsNSSendCall ifTrue:
		[| nsSendCache oop mappedOop |
		nsSendCache := self nsSendCacheFromReturnAddress: mcpc.
		oop := nsSendCache selector.	
		mappedOop := objectRepresentation remapObject: oop.
		oop ~= mappedOop ifTrue:
			[nsSendCache selector: mappedOop.
			(hasYoungPtr ~= 0 and: [objectMemory isYoung: mappedOop]) ifTrue:
				[(self cCoerceSimple: hasYoungPtr to: #'sqInt *') at: 0 put: true]].
		oop := nsSendCache enclosingObject.	
		oop ~= 0 ifTrue: [
			mappedOop := objectRepresentation remapObject: oop.
			oop ~= mappedOop ifTrue:
				[nsSendCache enclosingObject: mappedOop.
				(hasYoungPtr ~= 0 and: [objectMemory isYoung: mappedOop]) ifTrue:
					[(self cCoerceSimple: hasYoungPtr to: #'sqInt *') at: 0 put: true]]].
		^0 "keep scanning"]].

	(self isPureSendAnnotation: annotation) ifTrue:
		[self entryCacheTagAndCouldBeObjectAt: mcpc annotation: annotation into:
			[:entryPoint :cacheTag :tagCouldBeObj | | mappedCacheTag |
			 (tagCouldBeObj
			  and: [objectRepresentation couldBeObject: cacheTag]) ifTrue:
				[mappedCacheTag := objectRepresentation remapObject: cacheTag.
				 cacheTag ~= mappedCacheTag ifTrue:
					[self setCodeModified.
					 backEnd rewriteInlineCacheTag: mappedCacheTag at: mcpc asUnsignedInteger].
				 (hasYoungPtr ~= 0
				  and: [objectMemory isYoung: mappedCacheTag]) ifTrue:
					[(self cCoerceSimple: hasYoungPtr to: #'sqInt *') at: 0 put: true]].
			hasYoungPtr ~= 0 ifTrue:
				["Since the unlinking routines may rewrite the cacheTag to the send's selector, and
				  since they don't have the cogMethod to hand and can't add it to youngReferrers,
				  the method must remain in youngReferrers if the targetMethod's selector is young."
				 entryPoint > methodZoneBase ifTrue: "It's a linked send."
					[self targetMethodAndSendTableFor: entryPoint annotation: annotation into:
						[:targetMethod :ignored|
						 (objectMemory isYoung: targetMethod selector) ifTrue:
							[(self cCoerceSimple: hasYoungPtr to: #'sqInt *') at: 0 put: true]]]]]].
	^0 "keep scanning"! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 7/9/2021 13:57'!
remapMaybeObjRefInClosedPICAt: mcpc
	"Remap a potential object reference from a closed PIC.
	 This may be an object reference, an inline cache tag or null.
	 Answer if the updated literal is young.
	 mcpc is the address of the next instruction following either
	 the load of the method literal or the compare of the class tag."
	| object subject |
	object := backEnd literalBeforeFollowingAddress: mcpc.
	(objectRepresentation couldBeObject: object) ifFalse:
		[^false].
	subject := objectRepresentation remapOop: object.
	object ~= subject ifTrue:
		[self setCodeModified.
		 backEnd storeLiteral: subject beforeFollowingAddress: mcpc].
	^objectMemory isYoungObject: subject! !

!Cogit methodsFor: 'translation support' stamp: 'eem 2/12/2009 22:10'!
reportError: anInteger
	<cmacro: '(n) warning("compilation error")'>
	self halt: 'Compilation error ', anInteger printString! !

!Cogit methodsFor: 'debugging' stamp: 'eem 1/4/2021 10:04'!
reportInstructions: instructions
	<doNotGenerate>
	| skipNext printInst |
	skipNext := false.
	printInst := [:inst|
				coInterpreter transcript nextPutAll:
					(EagerInstructionDecoration
						ifTrue: [inst]
						ifFalse: [processor
									decorateDisassembly: inst
									for: self
									fromAddress: ((inst at: 3) = $r
													ifTrue: [Integer readFrom: inst readStream]
													ifFalse: [Integer readFrom: inst readStream base: 16])]); cr].
	instructions withIndexDo:
		[:thing :idx|
		skipNext
			ifTrue: [skipNext := false]
			ifFalse:
				[thing isString
					ifFalse:
						[thing first isString "i.e. { '(simulated return to '. processor retpcIn: coInterpreter memory. ')'}"
							ifTrue:
								[thing do:
									[:stringOrNumber|
									coInterpreter transcript nextPutAll: (stringOrNumber isString
															ifTrue: [stringOrNumber]
															ifFalse: [stringOrNumber hex])].
									coInterpreter transcript cr]
							ifFalse: "if possible, add the label to the instruction line to condense the output"
								[coInterpreter transcript cr.
								 (thing at: processor registerStatePCIndex ifAbsent: []) ifNotNil:
									[:pc| | next label |
									 label := self relativeLabelForPC: pc.
									 ((next := instructions at: idx + 1 ifAbsent: []) notNil
									  and: [next isString
									  and: [(Integer readFrom: next readStream radix: 16) = pc]])
										ifTrue: "Decorate instruction and eliminate pc line"
											[skipNext := true.
											 processor printRegisterStateExceptPC: thing on: coInterpreter transcript.
											 label ifNotNil: [coInterpreter transcript nextPutAll: label; space].
											 printInst value: next]
										ifFalse:
											[label ifNotNil: [coInterpreter transcript nextPutAll: label; nextPut: $:; cr].
											 processor printRegisterState: thing on: coInterpreter transcript]]]]
					ifTrue:
						[printInst value: thing]]].
	coInterpreter transcript flush! !

!Cogit methodsFor: 'debugging' stamp: 'eem 1/4/2021 10:04'!
reportLastNInstructions
	<doNotGenerate>
	self reportInstructions: lastNInstructions! !

!Cogit methodsFor: 'analysis' stamp: 'eem 8/23/2021 10:46'!
reportPerMethodProfilingOn: aStream
	<doNotGenerate>
	| mTotal pTotal tTotal |
	perMethodProfile ifNil:
		[aStream nextPutAll: 'not collecting profile'.
		 ^self].
	self moveProfileToMethods.
	aStream nextPutAll: 'instructions dispatched per trampoline'; cr.
	tTotal := 0.
	((perMethodProfile associations select: [:assoc| assoc key < methodZoneBase]) sorted: [:a1 :a2| a1 value >= a2 value]) do:
		[:assoc|
		tTotal := tTotal + assoc value.
		aStream tab; print: assoc value; nextPut: $:; tab; nextPutAll: (self codeEntryNameFor: assoc key); cr].
	aStream nextPutAll: 'instructions dispatched per method'; cr.
	mTotal := 0.
	((perMethodProfile associations
			select: [:assoc| assoc key >= objectMemory startOfMemory and: [objectMemory isCompiledMethod: assoc key]])
			sorted: [:a1 :a2| a1 value >= a2 value]) do:
		[:assoc|
		mTotal := mTotal + assoc value.
		coInterpreter reportSingleMethodProfile: assoc on: aStream].
	aStream nextPutAll: 'instructions dispatched per pic'; cr.
	pTotal := 0.
	((perMethodProfile associations
			select: [:assoc| assoc key >= objectMemory startOfMemory and: [(objectMemory isCompiledMethod: assoc key) not]])
			sorted: [:a1 :a2| a1 value >= a2 value]) do:
		[:assoc|
		pTotal := pTotal + assoc value.
		aStream tab; print: assoc value; nextPut: $:; tab.
		coInterpreter printStringOf: assoc key on: aStream.
		aStream cr].
	aStream flush.
	^{tTotal. mTotal. pTotal }! !

!Cogit methodsFor: 'analysis' stamp: 'eem 8/23/2021 11:17'!
reportPerMethodProfilingOn: aStream cutoff: cutoffOrNil
	<doNotGenerate>
	| mTotal pTotal tTotal |
	perMethodProfile ifNil:
		[aStream nextPutAll: 'not collecting profile'.
		 ^self].
	self moveProfileToMethods.
	aStream nextPutAll: 'instructions dispatched per trampoline'; cr.
	tTotal := 0.
	((perMethodProfile associations select: [:assoc| assoc key < methodZoneBase]) sorted: [:a1 :a2| a1 value >= a2 value]) do:
		[:assoc|
		tTotal := tTotal + assoc value.
		(cutoffOrNil isNil or: [assoc value >= cutoffOrNil]) ifTrue:
			[aStream tab; print: assoc value; nextPut: $:; tab; nextPutAll: (self codeEntryNameFor: assoc key); cr]].
	aStream nextPutAll: 'instructions dispatched per method'; cr.
	mTotal := 0.
	((perMethodProfile associations
			select: [:assoc| assoc key >= objectMemory startOfMemory and: [objectMemory isCompiledMethod: assoc key]])
			sorted: [:a1 :a2| a1 value >= a2 value]) do:
		[:assoc|
		mTotal := mTotal + assoc value.
		(cutoffOrNil isNil or: [assoc value >= cutoffOrNil]) ifTrue:
			[coInterpreter reportSingleMethodProfile: assoc on: aStream]].
	aStream nextPutAll: 'instructions dispatched per pic'; cr.
	pTotal := 0.
	((perMethodProfile associations
			select: [:assoc| assoc key >= objectMemory startOfMemory and: [(objectMemory isCompiledMethod: assoc key) not]])
			sorted: [:a1 :a2| a1 value >= a2 value]) do:
		[:assoc|
		pTotal := pTotal + assoc value.
		(cutoffOrNil isNil or: [assoc value >= cutoffOrNil]) ifTrue:
			[aStream tab; print: assoc value; nextPut: $:; tab.
			coInterpreter printStringOf: assoc key on: aStream.
			aStream cr]].
	aStream flush.
	^{tTotal. mTotal. pTotal }! !

!Cogit methodsFor: 'initialization' stamp: 'eem 10/29/2020 22:47'!
resetMethodZoneBase: address
	"For throw-away/one-off compilations"
	<inline: #always>
	methodZoneBase := address! !

!Cogit methodsFor: 'tests-method map' stamp: 'eem 7/25/2018 17:42'!
resolveToOopOrNil: proxyOrSpecialSelector
	"Answer the oop of a proxy or a special selector.  When testing pc mapping on simulation
	 methods (as opposed to CurrentImageFacades) selectorForSendBefore:in: answers either
	 VMObjectProxy instances (normal sends) or symbols (special selector sends)."
	<doNotGenerate>
	| index |
	proxyOrSpecialSelector isSymbol ifFalse:
		[^proxyOrSpecialSelector oop].
	index := Smalltalk specialSelectors indexOf: proxyOrSpecialSelector.
	^index > 0 ifTrue:
		[coInterpreter specialSelector: index - 1 // 2]! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 1/15/2016 17:07'!
rewriteCPIC: cPIC caseJumpTo: target 
	"adding a new CPIC case, or making an MNU CPIC, requires altering the jump that takes us to the first case to be used"
	<inline: true>
	backEnd rewriteCPICJumpAt: cPIC asInteger + firstCPICCaseOffset - backEnd jumpLongByteSize - backEnd loadLiteralByteSize target: target! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'rmacnak 11/23/2015 22:12:44'!
rewriteCPICCaseAt: followingAddress tag: newTag objRef: newObjRef target: newTarget
	"Rewrite the three values involved in a CPIC case.  Used by the initialize & extend CPICs.
	 c.f. expectedClosedPICPrototype:"

	"write the obj ref/operand via the second ldr"
	| classTagPC methodObjPC |
	methodObjPC := followingAddress - backEnd jumpLongConditionalByteSize - backEnd cmpC32RTempByteSize.
	backEnd storeLiteral: newObjRef beforeFollowingAddress: methodObjPC.

	classTagPC := followingAddress - backEnd jumpLongConditionalByteSize.
	"rewite the tag via the first ldr"	
	backEnd storeLiteral32: newTag beforeFollowingAddress: classTagPC.

	"write the jump address for the new target address"
	backEnd rewriteConditionalJumpLongAt: followingAddress target: newTarget! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'rsf 1/5/2017 15:50'!
scanBlock: blockStart
	"Scan the block to determine if the block needs a frame or not"
	| descriptor pc end framelessStackDelta nExts |
	<var: #blockStart type: #'BlockStart *'>
	<var: #descriptor type: #'BytecodeDescriptor *'>
	needsFrame := false.
	methodOrBlockNumArgs := blockStart numArgs.
	inBlock := InVanillaBlock.
	pc := blockStart startpc.
	end := blockStart startpc + blockStart span.
	framelessStackDelta := nExts := extA := numExtB := extB := 0.
	[pc < end] whileTrue:
		[byte0 := (objectMemory fetchByte: pc ofObject: methodObj) + bytecodeSetOffset.
		 descriptor := self generatorAt: byte0.
		 descriptor isExtension ifTrue:
			[self loadSubsequentBytesForDescriptor: descriptor at: pc.
			 self perform: descriptor generator].
		 needsFrame ifFalse:
			[(descriptor needsFrameFunction isNil
			  or: [self perform: descriptor needsFrameFunction with: framelessStackDelta])
				ifTrue: [needsFrame := true]
				ifFalse: [framelessStackDelta := framelessStackDelta + descriptor stackDelta]].
		 objectRepresentation maybeNoteDescriptor: descriptor blockStart: blockStart.
		 pc := self nextBytecodePCFor: descriptor at: pc exts: nExts in: methodObj.
		 descriptor isExtension
			ifTrue: [nExts := nExts + 1]
			ifFalse: [nExts := extA := numExtB := 0. extB := 0]].
	needsFrame ifFalse:
		[framelessStackDelta < 0 ifTrue:
			[self error: 'negative stack delta in block; block contains bogus code or internal error'].
		 [framelessStackDelta > 0] whileTrue:
			[descriptor := self generatorAt: (objectMemory fetchByte: blockStart startpc ofObject: methodObj) + bytecodeSetOffset.
			 descriptor generator ~~ #genPushConstantNilBytecode ifTrue:
				[self error: 'frameless block doesn''t start with enough pushNils'].
			 blockStart
				startpc: blockStart startpc + descriptor numBytes;
				span: blockStart span - descriptor numBytes.
			 framelessStackDelta := framelessStackDelta - 1]].
	^0! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 8/8/2014 15:08'!
scanForCleanBlocks
	"Answer the number of clean blocks found in the literal frame"
	| numCleanBlocks |
	numCleanBlocks := 0.
	1 to: (objectMemory literalCountOf: methodObj) do:
		[:i| | lit |
		lit := objectMemory fetchPointer: i ofObject: methodObj.
		(coInterpreter startPCOrNilOfLiteral: lit in: methodObj) ifNotNil:
			[:startPCOrNil| numCleanBlocks := numCleanBlocks + 1]].
	^numCleanBlocks! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'cb 12/21/2017 11:56'!
scanMethod
	"Scan the method (and all embedded blocks) to determine
		- what the last bytecode is; extra bytes at the end of a method are used to encode things like source pointers or temp names
		- if the method needs a frame or not
		- what are the targets of any backward branches.
		- how many blocks it creates
		- if it contans an unknown bytecode
	 Answer the block count or on error a negative error code"
	| latestContinuation nExts descriptor pc numBlocks distance targetPC framelessStackDelta |
	<var: #descriptor type: #'BytecodeDescriptor *'>
	needsFrame := false.
	NewspeakVM ifTrue:
		[numIRCs := 0].
	(primitiveIndex > 0
	 and: [coInterpreter isQuickPrimitiveIndex: primitiveIndex]) ifTrue:
		[^0].
	pc := latestContinuation := initialPC.
	numBlocks := framelessStackDelta := nExts := extA := numExtB := extB := 0.
	[pc <= endPC] whileTrue:
		[byte0 := (objectMemory fetchByte: pc ofObject: methodObj) + bytecodeSetOffset.
		 descriptor := self generatorAt: byte0.
		 descriptor isExtension ifTrue:
			[descriptor opcode = Nop ifTrue: "unknown bytecode tag; see Cogit class>>#generatorTableFrom:"
				[^EncounteredUnknownBytecode].
			 self loadSubsequentBytesForDescriptor: descriptor at: pc.
			 self perform: descriptor generator].
		 (descriptor isReturn
		  and: [pc >= latestContinuation]) ifTrue:
			[endPC := pc].
		 needsFrame ifFalse:
			[(descriptor needsFrameFunction isNil
			  or: [self perform: descriptor needsFrameFunction with: framelessStackDelta])
				ifTrue: [needsFrame := true]
				ifFalse: [framelessStackDelta := framelessStackDelta + descriptor stackDelta]].
		 descriptor isBranch ifTrue:
			[distance := self spanFor: descriptor at: pc exts: nExts in: methodObj.
			 targetPC := pc + descriptor numBytes + distance.
			 (self isBackwardBranch: descriptor at: pc exts: nExts in: methodObj)
				ifTrue: [self initializeFixupAt: targetPC]
				ifFalse: [latestContinuation := latestContinuation max: targetPC]].
		 descriptor isBlockCreation ifTrue:
			[numBlocks := numBlocks + 1.
			 distance := self spanFor: descriptor at: pc exts: nExts in: methodObj.
			 targetPC := pc + descriptor numBytes + distance.
			 latestContinuation := latestContinuation max: targetPC].
		 NewspeakVM 
			ifTrue:
				[descriptor hasIRC ifTrue:
					[numIRCs := numIRCs + 1]]
			ifFalse: 
				[descriptor hasIRC ifTrue:
					[ (objectMemory fetchByte: pc + 1 ofObject: methodObj) 
						<< 8 
						+ (objectMemory fetchByte: pc + 2 ofObject: methodObj)
						= ((1 << 15) + 6000) ifTrue: [self initializeFixupAt: targetPC]	 ] ].
		 pc := pc + descriptor numBytes.
		 descriptor isExtension
			ifTrue: [nExts := nExts + 1]
			ifFalse: [nExts := extA := numExtB := extB := 0]].
	^numBlocks! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 2/20/2016 18:54'!
selectorForSendAt: mcpc annotation: annotation in: aCompiledMethod
	<doNotGenerate>
	| entryPoint offset targetMethod selector |
	entryPoint := backEnd callTargetFromReturnAddress: mcpc asInteger.
	selector := entryPoint > methodZoneBase
					ifTrue: "It's a linked send."
						[self
							offsetAndSendTableFor: entryPoint
							annotation: annotation
							into: [:off :table| offset := off].
						targetMethod := self cCoerceSimple: entryPoint - offset to: #'CogMethod *'.
						targetMethod selector]
					ifFalse:
						[self inlineCacheTagsAreIndexes
							ifTrue: [self selectorFromSelectorIndex: (backEnd inlineCacheTagAt: mcpc) signedIntFromLong
										in: aCompiledMethod]
							ifFalse: [backEnd inlineCacheTagAt: mcpc]].
	^(annotation ~= IsNSSendCall and: [coInterpreter isCurrentImageFacade])
		ifTrue: [coInterpreter objectForOop: selector]
		ifFalse: [selector]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 11/4/2012 08:12'!
selectorForSendBefore: bcpc in: aCompiledMethod
	<doNotGenerate>
	| is sel |
	"sends map to the following pc.  need to find the selector for the previous pc"
	is := InstructionStream on: aCompiledMethod.
	is pc: (aCompiledMethod pcPreviousTo: bcpc + 1). "bcpc is 0-rel"
	^(sel := is selectorToSendOrSelf) ~~ is ifTrue: [sel]! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 10/6/2015 18:21'!
selectorFromSelectorIndex: selectorIndex in: aCompiledMethod
	<inline: true>
	^selectorIndex < 0
		ifTrue: [(coInterpreter specialSelector: -1 - selectorIndex)]
		ifFalse: [coInterpreter literal: selectorIndex ofMethod: aCompiledMethod]! !

!Cogit methodsFor: 'analysis' stamp: 'eem 11/18/2013 12:28'!
sendSitesIn: cogMethod do: binaryBlock
	"Evaluate binaryBlock with the annotation and mcpc for each send site in cogMethod"
	<doNotGenerate>
	self mapFor: cogMethod
		 performUntil: (#withAnnotation:mcpc:evaluate:)
		 arg: [:ann :mcpc | (self isSendAnnotation: ann) ifTrue: [binaryBlock value: ann value: mcpc]]! !

!Cogit methodsFor: 'debugging' stamp: 'eem 11/15/2012 10:16'!
sendTrace
	<doNotGenerate>
	^traceFlags! !

!Cogit methodsFor: 'debugging' stamp: 'eem 8/3/2021 18:28'!
sendTrace: aBooleanOrInteger
	<doNotGenerate>
	"traceFlags is a set of flags.
	 1 => print trace (if something below is selected)
	 2 => trace sends
	 4 => trace block activations
	 8 => trace interpreter primitives (on by default; see Cogit class>>#declareCVarsIn:)
	 16 => trace events (context switches, GCs, etc)
	 32 => trace stack overflow
	 64 => send breakpoint on implicit receiver (Newspeak VM only)
	128 => check stack depth on send (simulation only)
	256 => trace linked sends
	512 => trace fast C call interpreter primitives (off by default)"
	traceFlags := aBooleanOrInteger isInteger
						ifTrue: [aBooleanOrInteger]
						ifFalse: [aBooleanOrInteger ifTrue: [traceFlags bitOr: 6] ifFalse: [traceFlags bitClear: 6]]! !

!Cogit methodsFor: 'debugging' stamp: 'eem 2/21/2009 11:08'!
setBreakMethod: anObj
	<api>
	breakMethod := anObj! !

!Cogit methodsFor: 'memory access' stamp: 'eem 7/20/2021 17:06'!
setCodeModified
	"Three cases to avoid reading codeModified unless absolutely necessary.  On some platforms run-time calls
	 may be required to enable execution and disable write-protect of the code zone. See the comment in ensureExecutableCodeZone."
	<inline: #always>
	self cppIf: #DUAL_MAPPED_CODE_ZONE
		ifTrue: [codeModified := true]
		ifFalse:
			[backEnd needsCodeZoneExecuteWriteSwitch
				ifTrue:
					[codeModified ifFalse:
						[codeModified := true.
						 backEnd makeCodeZoneWritable.
						 self cCode: nil inSmalltalk: [codeZoneIsExecutableNotWritable := false]]]
				ifFalse: [codeModified := true]]! !

!Cogit methodsFor: 'accessing' stamp: 'eem 9/9/2019 16:23'!
setHasMovableLiteral: boolean
	"Written this way to allow break-pointing in the simulator."
	<cmacro: '(b) (hasMovableLiteral = (b))'>
	hasMovableLiteral := boolean! !

!Cogit methodsFor: 'accessing' stamp: 'eem 9/9/2019 16:24'!
setHasYoungReferent: boolean
	"Written this way to allow break-pointing in the simulator."
	<cmacro: '(b) (hasYoungReferent = (b))'>
	"boolean ifTrue:
		[self halt]."
	"(hasYoungReferent == false and: [boolean == true]) ifTrue:
		[self halt]."
	hasYoungReferent := boolean! !

!Cogit methodsFor: 'initialization' stamp: 'eem 7/8/2021 17:31'!
setInterpreter: aCoInterpreter
	"Initialization of the code generator in the simulator.
	 These objects already exist in the generated C VM
	 or are used only in the simulation."
	<doNotGenerate>
	coInterpreter := aCoInterpreter.
	objectMemory := aCoInterpreter objectMemory.
	methodZone := self class methodZoneClass new.
	objectRepresentation := objectMemory objectRepresentationClass
								forCogit: self methodZone: methodZone.
	methodZone setInterpreter: aCoInterpreter
				objectRepresentation: objectRepresentation
				cogit: self.
	generatorTable := self class generatorTable.
	processor := ProcessorClass new.
	simulatedAddresses := Dictionary new.
	coInterpreter class clusteredVariableNames do:
		[:cvn| self simulatedAddressFor: (cvn first = $C ifTrue: ['get', cvn] ifFalse: [cvn]) asSymbol].
	simulatedTrampolines := Dictionary new.
	simulatedVariableGetters := Dictionary new.
	simulatedVariableSetters := Dictionary new.
	traceStores := 0.
	traceFlags := (InitializationOptions at: #linkedSendTrace ifAbsent: [false])
					ifTrue: [257 "compileSendTrace + print"]
					ifFalse:
						[(InitializationOptions at: #recordPrimTrace ifAbsent: [true])
							ifTrue: [8] "record prim trace on by default (see Cogit class>>decareCVarsIn:)"
							ifFalse: [0]].
	singleStep := printRegisters := printInstructions := clickConfirm := false.
	backEnd := CogCompilerClass for: self.
	methodLabel := CogCompilerClass for: self.
	(literalsManager := backEnd class literalsManagerClass new) cogit: self.
	ordinarySendTrampolines := CArrayAccessor on: (Array new: NumSendTrampolines).
	superSendTrampolines := CArrayAccessor on: (Array new: NumSendTrampolines).
	BytecodeSetHasDirectedSuperSend ifTrue:
		[directedSuperSendTrampolines := CArrayAccessor on: (Array new: NumSendTrampolines).
		 directedSuperBindingSendTrampolines := CArrayAccessor on: (Array new: NumSendTrampolines).
		 directedSendUsesBinding := false].
	NewspeakVM ifTrue:
		[selfSendTrampolines := CArrayAccessor on: (Array new: NumSendTrampolines).
		dynamicSuperSendTrampolines := CArrayAccessor on: (Array new: NumSendTrampolines).
		implicitReceiverSendTrampolines := CArrayAccessor on: (Array new: NumSendTrampolines).
		outerSendTrampolines := CArrayAccessor on: (Array new: NumSendTrampolines)].
	"debug metadata"
	objectReferencesInRuntime := CArrayAccessor on: (Array new: NumObjRefsInRuntime).
	runtimeObjectRefIndex := 0.
	"debug metadata"
	trampolineAddresses := CArrayAccessor on: (Array new: NumTrampolines * 2).
	trampolineTableIndex := 0.

	extA := numExtB := extB := 0.

	compilationTrace ifNil: [compilationTrace := self class initializationOptions at: #compilationTrace ifAbsent: [0]].
	debugOpcodeIndices := self class initializationOptions at: #debugOpcodeIndices ifAbsent: [Set new].
	debugBytecodePointers := self class initializationOptions at: #debugBytecodePointers ifAbsent: [Set new].
	self class initializationOptions at: #breakPC ifPresent: [:pc| breakPC := pc]! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 7/5/2021 16:54'!
setSelectorOf: cogMethod to: aSelectorOop
	<api>
	"If a method is compiled to machine code via a block entry it won't have a selector.
	 A subsequent send can find the method and hence fill in the selector."
	<var: #cogMethod type: #'CogMethod *'>
	"self disassembleMethod: cogMethod"
	coInterpreter compilationBreakpoint: aSelectorOop isMNUCase: false.
	self assert: cogMethod cmType = CMMethod.
	self ensureWritableCodeZone.
	(self writableMethodFor: cogMethod) selector: aSelectorOop.
	(objectMemory isYoung: aSelectorOop) ifTrue:
		[methodZone ensureInYoungReferrers: cogMethod].
	self ensureExecutableCodeZone! !

!Cogit methodsFor: 'initialization' stamp: 'eem 12/3/2015 13:41'!
setStackAlignment: stackAlignment expectedSPOffset: spOffset expectedFPOffset: fpOffset
	"Spcific platform ABIs mandate specific stack frame alignments.  We capture
	 these constraints in the variables here and test they are adhered to via
	 assertCStackWellAligned whenever transitioning to code that will run in C."
	self assert: stackAlignment isPowerOfTwo.
	expectedSPAlignment := spOffset \\ stackAlignment.
	expectedFPAlignment := fpOffset \\ stackAlignment.
	cStackAlignment := stackAlignment
! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 7/8/2021 17:19'!
shortcutTrampoline: aProcessorSimulationTrap to: aBlock
	<doNotGenerate>
	"As a simulation performance hack for debugging trampolines such as ceTraceLinkedSend: allow the entire
	 trampoline to be implemented by this method instead of simulating all of the code for the trampoline."
	| stsp |
	stsp := processor registerAt: SPReg put: (processor registerAt: SPReg) - backEnd class wordSize.
	"when the processor simulators are fixed to call the illegal address, and push the retpc on the stack, this will change to only push if there is a link reg."
	objectMemory
		longAt: stsp
		put: (backEnd hasLinkRegister ifTrue: [processor lr] ifFalse: [aProcessorSimulationTrap nextpc]).
	coInterpreter
		stackPointer: (processor registerAt: SPReg);
		framePointer: processor fp.
	aBlock value.
	processor
		registerAt: SPReg put: (processor registerAt: SPReg) + backEnd class wordSize;
		pc: aProcessorSimulationTrap nextpc! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 10/29/2020 22:30'!
simulateCeFlushDCacheFrom: start to: finish
	<doNotGenerate>
	processor abiMarshalArg0: start arg1: finish in: objectMemory memory.
	self simulateLeafCallOf: ceFlushDCache.
	processor abiUnmarshal: 2! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 7/13/2021 11:47'!
simulateCeFlushICacheFrom: start to: finish
	<doNotGenerate>
	ceFlushICache ifNotNil:
		[processor abiMarshalArg0: start arg1: finish in: objectMemory memory.
		 self simulateLeafCallOf: ceFlushICache.
		 processor abiUnmarshal: 2]! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 7/8/2021 13:49'!
simulateCogCodeAt: address "<Integer>"
	<doNotGenerate>
	| stackZoneBase |
	backEnd needsCodeZoneExecuteWriteSwitch ifTrue:
		[self assert: codeZoneIsExecutableNotWritable].
	stackZoneBase := coInterpreter stackZoneBase.
	processor pc: address.
	[[[singleStep
		ifTrue:
			[[processor sp < stackZoneBase ifTrue: [self halt].
			  self recordProcessing.
			  self maybeBreakAt: processor pc] value. "So that the Debugger's Over steps over all this"
			  processorLock critical:
				[processor
					singleStepIn: coInterpreter memory
					minimumAddress: guardPageSize
					readOnlyBelow: methodZone zoneEnd]]
		ifFalse:
			[processorLock critical:
				[processor
					runInMemory: coInterpreter memory
					minimumAddress: guardPageSize
					readOnlyBelow: methodZone zoneEnd]].
	   "((printRegisters or: [printInstructions]) and: [clickConfirm]) ifTrue:
	 	[(self confirm: 'continue?') ifFalse:
			[clickConfirm := false. self halt]]."
	   true] whileTrue]
		on: ProcessorSimulationTrap
		do: [:ex| ex applyTo: self].
	 true] whileTrue! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 4/25/2020 17:42'!
simulateEnilopmart: enilopmartAddress numArgs: n
	<doNotGenerate>
	"Enter Cog code, popping the class reg and receiver from the stack
	 and then returning to the address beneath them.
	 In the actual VM the enilopmart is a function pointer and so senders
	 of this method end up calling the enilopmart to enter machine code.
	 In simulation we either need to start simulating execution (if we're in
	 the interpreter) or return to the simulation (if we're in the run-time
	 called from machine code. We should also smash the register state
	 since, being an abnormal entry, no saved registers will be restored."
	self assert: (coInterpreter isOnRumpCStack: processor sp).
	self assert: (n = 0 or: [(coInterpreter stackValue: n) between: guardPageSize and: methodZone freeStart - 1]).
	(printInstructions or: [printRegisters]) ifTrue:
		[coInterpreter printExternalHeadFrame].
	processor
		smashRegistersWithValuesFrom: 16r80000000 by: objectMemory wordSize;
		simulateLeafCallOf: enilopmartAddress
		nextpc: 16rBADF00D
		memory: coInterpreter memory.
	"If we're already simulating in the context of machine code then
	 this will take us back to handleCallSimulationTrap:.  Otherwise
	 start executing machine code in the simulator."
	ReenterMachineCode new signal.
	self simulateCogCodeAt: enilopmartAddress.
	"We should either longjmp back to the interpreter or
	 stay in machine code so control should not reach here."
	self assert: false! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 7/8/2021 13:48'!
simulateLeafCallOf: someFunction
	"Simulate execution of machine code that leaf-calls someFunction,
	 answering the result returned by someFunction."
	"CogProcessorAlienInspector openFor: coInterpreter"
	<doNotGenerate>
	| priorSP priorPC priorLR spOnEntry bogusRetPC |
	backEnd needsCodeZoneExecuteWriteSwitch ifTrue:
		[self assert: codeZoneIsExecutableNotWritable].
	self recordRegisters.
	priorSP := processor sp.
	priorPC := processor pc.
	priorLR := backEnd hasLinkRegister ifTrue: [processor lr].
	processor
		simulateLeafCallOf: someFunction
		nextpc: (bogusRetPC := 16rBADF00D5 roundTo: backEnd codeGranularity)
		memory: coInterpreter memory.
	spOnEntry := processor sp.
	self recordInstruction: {'(simulated call of '. someFunction. ')'}.
	^[[[processor pc between: self class guardPageSize and: methodZone zoneEnd] whileTrue:
		[singleStep
			ifTrue: [self recordProcessing.
					self maybeBreakAt: processor pc.
					processorLock critical:
						[processor
							singleStepIn: coInterpreter memory
							minimumAddress: guardPageSize
							readOnlyBelow: methodZone zoneEnd]]
			ifFalse: [processorLock critical:
						[processor
							runInMemory: coInterpreter memory
							minimumAddress: guardPageSize
							readOnlyBelow: methodZone zoneEnd]]]]
			on: ProcessorSimulationTrap, Error
			do: [:ex|
				"Again this is a hack for the processor simulators not properly simulating returns to bogus addresses.
				 In this case BochsX64Alien doesn't do the right thing."
				processor pc = bogusRetPC ifTrue:
					[self recordInstruction: {'(simulated (real) return to '. processor pc. ')'}.
					 ^processor cResultRegister].
				ex isProcessorSimulationTrap ifFalse:
					[ex pass].
				ex applyTo: self.
				ex type == #return ifTrue:
					[^processor cResultRegister]].
	processor pc = bogusRetPC ifTrue:
		[self recordInstruction: {'(simulated (real) return to '. processor pc. ')'}].
	processor cResultRegister]
		ensure:
			[processor sp: priorSP.
			 processor pc: priorPC.
			 priorLR ifNotNil: [:lr| processor lr: lr]]! !

!Cogit methodsFor: 'simulation processor access' stamp: 'eem 10/18/2010 13:57'!
simulateLeafReturnIn: aMemory
	<doNotGenerate>
	self assertCorrectProcessorOwnership.
	^processor simulateLeafReturnIn: aMemory! !

!Cogit methodsFor: 'initialization' stamp: 'eem 7/8/2021 15:11'!
simulatedAddressFor: anObject
	"Answer a simulated address for a block or a symbol.  This is an address
	 that can be called, read or written by generated machine code, and will
	 be mapped into a Smalltalk message send or block evaluation. These
	 addresses are at the top end of the bottom half of the address space to
	 avoid having the sign bit set and so will not look like negative numbers."
	<doNotGenerate>
	^simulatedAddresses
		at: anObject
		ifAbsentPut: [simulatedAddresses size * objectMemory wordSize + self allButTopBitOfAddressSpaceMask]! !

!Cogit methodsFor: 'initialization' stamp: 'eem 10/19/2015 09:56:58'!
simulatedReadWriteVariableAddress: getter in: receiver
	"Answer a simulated variable.  This is a variable whose value can be read
	 and written by generated machine code."
	<doNotGenerate>
	| address |
	address := self simulatedVariableAddress: getter in: receiver.
	simulatedVariableSetters
		at: address
		ifAbsentPut:
			[| setter |
			setter := (((getter beginsWith: 'get')
						ifTrue: ['s', getter allButFirst]
						ifFalse: [getter]), ':') asSymbol.
			[:value| receiver perform: setter with: value]].
	^address! !

!Cogit methodsFor: 'initialization' stamp: 'tpr 7/9/2015 16:19'!
simulatedTrampolineFor: selectorOrAddress
	"Set a simulated trampoline.  This is a method in the cogit, coInterpreter
	 or objectMemory that is called from a machine code trampoline."
	<doNotGenerate>
	| address |
	selectorOrAddress isInteger ifTrue:
		[self assert: (simulatedTrampolines includesKey: selectorOrAddress).
		 ^selectorOrAddress].
	self assert: selectorOrAddress isSymbol.
	address := self simulatedAddressFor: selectorOrAddress.
	simulatedTrampolines
		at: address
		ifAbsentPut:
			[MessageSend
				receiver: ((self respondsTo: selectorOrAddress)
							ifTrue: [self]
							ifFalse: [(coInterpreter respondsTo: selectorOrAddress)
										ifTrue: [coInterpreter]
										ifFalse: [(objectMemory respondsTo: selectorOrAddress)
											ifTrue: [objectMemory]
											ifFalse: [(backEnd respondsTo: selectorOrAddress)
												ifTrue:[backEnd]
												ifFalse:[self notify: 'cannot find receiver for ', selectorOrAddress]]]])
				selector: selectorOrAddress
				arguments: (1 to: selectorOrAddress numArgs) asArray].
	^address! !

!Cogit methodsFor: 'initialization' stamp: 'eem 7/2/2014 17:14'!
simulatedTrampolines
	<doNotGenerate>
	^simulatedTrampolines! !

!Cogit methodsFor: 'initialization' stamp: 'eem 10/16/2010 18:14'!
simulatedVariableAddress: getter in: receiver
	"Answer a simulated variable.  This is a variable whose value can be read
	 by generated machine code."
	<doNotGenerate>
	| address |
	address := self simulatedAddressFor: getter.
	simulatedVariableGetters
		at: address
		ifAbsentPut: [MessageSend receiver: receiver selector: getter].
	^address! !

!Cogit methodsFor: 'initialization' stamp: 'tpr 5/29/2014 21:59'!
simulatedVariableAt: address
	"Answer a simulated variable's value for handling the push/pop tracer."
	<doNotGenerate>
	^(simulatedVariableGetters
		at: address
		ifAbsent:[0]) value! !

!Cogit methodsFor: 'debugging' stamp: 'eem 9/29/2009 22:24'!
singleStep
	<doNotGenerate>
	^singleStep! !

!Cogit methodsFor: 'debugging' stamp: 'eem 1/17/2009 12:21'!
singleStep: aBoolean
	<doNotGenerate>
	singleStep := aBoolean! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/2/2012 13:13'!
spanFor: descriptor at: pc exts: nExts in: aMethodObj
	"Compute the span (jump distance) for a particular bytecode, e.g. a backward branch or a block."
	<inline: true>
	<var: #descriptor type: #'BytecodeDescriptor *'>
	^self
		perform: descriptor spanFunction
		with: descriptor
		with: pc
		with: nExts
		with: aMethodObj! !

!Cogit methodsFor: 'compile abstract instructions' stamp: 'eem 1/21/2018 16:10'!
spanForCleanBlockStartingAt: startPC
	<var: #descriptor type: #'BytecodeDescriptor *'>
	| pc end descriptor |
	pc := startPC.
	end := objectMemory numBytesOf: methodObj.
	[pc <= end] whileTrue:
		[descriptor := self generatorForPC: pc.
		 pc := pc + descriptor numBytes.
		 descriptor isReturn ifTrue:
			[^pc - startPC]].
	self error: 'couldn''t locate end of clean block'.
	^0! !

!Cogit methodsFor: 'initialization' stamp: 'eem 2/3/2020 18:29'!
sqMakeMemoryExecutableFrom: startAddress To: endAddress CodeToDataDelta: codeToDataDeltaPtr
	<doNotGenerate>
	"Simulate setting executable permissions on the code zone.  In production this will apply execute permission
	 to startAddress throguh endAddress - 1.  If starting up in the DUAL_MAPPED_CODE_ZONE regime then it
	 will also create a writable mapping for the code zone and assign the distance from executable zone to the
	 writable zone throguh codeToDataDeltaPtr.  If in this regime when simulating, the CogVMSimulator will
	 have allocated twice as much code memory as asked for (see CogVMSimulator openOn:extraMemory:) and
	 so simply set the delta to the code size."
	(InitializationOptions at: #DUAL_MAPPED_CODE_ZONE ifAbsent: [false]) ifTrue:
		[codeToDataDeltaPtr at: 0 put: coInterpreter cogCodeSize]! !

!Cogit methodsFor: 'testing' stamp: 'eem 6/18/2021 20:17'!
stackCheckOffsetOfBlockAt: blockEntryMcpc isMcpc: mcpc
	<returnTypeC: #usqInt>
	| cogBlockMethod |
	<var: #cogBlockMethod type: #'CogBlockMethod *'>
	cogBlockMethod := self cCoerceSimple: blockEntryMcpc - (self sizeof: CogBlockMethod)
							  to: #'CogBlockMethod *'.
	cogBlockMethod asInteger + cogBlockMethod stackCheckOffset = mcpc ifTrue:
		[^cogBlockMethod asUnsignedInteger].
	^0 "keep scanning..."! !

!Cogit methodsFor: 'simulation only' stamp: 'eem 5/15/2014 09:51'!
stackPointerAddress
	"redirect for the backEnd's genSaveStackPointers"
	<doNotGenerate>
	^coInterpreter stackPointerAddress! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 9/6/2016 12:04'!
startMcpcAndCogMethodForMcpc: mcpc in: cogMethod do: aBinaryBlock
	"Evaluate aBinaryBlock with the startmcpc and method containing mcpc in cogMethod."
	<doNotGenerate>
	| startMcpc |
	startMcpc := ((self codeRangesFor: cogMethod)
					detect: [:range| range includes: mcpc]
					ifNone:
						[(self codeRangesFor: cogMethod)
							detect: [:range| range last + 1 = mcpc]
							ifNone: [^nil]]) first.
	^aBinaryBlock
		value: startMcpc
		value: (startMcpc = (cogMethod asInteger + (self sizeof: CogMethod))
					ifTrue: [cogMethod]
					ifFalse: [self cCoerceSimple: startMcpc - (self sizeof: CogBlockMethod)
								to: #'CogBlockMethod *'])! !

!Cogit methodsFor: 'method map' stamp: 'eem 2/15/2011 10:22'!
subMethodsAsRangesFor: surrogateOrAddress
	<doNotGenerate>
	| cogMethod codeRanges |
	cogMethod := surrogateOrAddress isInteger
								ifTrue: [self cogMethodSurrogateAt: surrogateOrAddress]
								ifFalse: [surrogateOrAddress].
	^cogMethod cmType = CMMethod ifTrue:
		[codeRanges := self codeRangesFor: cogMethod.
		 ^codeRanges size > 1 "omit the block dispatch range"
			ifTrue: [codeRanges allButLast]
			ifFalse: [codeRanges]]! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 12/7/2016 14:39'!
subsequentPrototypeMethodOop
	"Answer a fake value for the method oop in other than the first case in the PIC prototype.
	 Since we use MoveUniqueCw:R: it must not be confused with a method-relative address."
	<inline: false>
	^(self addressIsInCurrentCompilation: 16rBADA550)
		ifTrue: [16rDEADEAD]
		ifFalse: [16rBADA550]! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 1/12/2018 10:26'!
targetMethodAndSendTableFor: entryPoint annotation: annotation into: binaryBlock
	"Evaluate binaryBlock with the targetMethod and relevant send table for a linked-send
	 to entryPoint.  Do so based on the alignment of entryPoint.  N.B.  For Newspeak sends
	 we don't need to distinguish between ceImplicitReceiver and the other sends since
	 ceImplicitReceiver will never appear to be linked, so only three cases here."
	<inline: true>
	<var: #targetMethod type: #'CogMethod *'>

	self offsetAndSendTableFor: entryPoint
		annotation: annotation
		into: [:offset :sendTable| | targetMethod |
			targetMethod := self cCoerceSimple: entryPoint - offset to: #'CogMethod *'.
			binaryBlock
				value: targetMethod
				value: sendTable]! !

!Cogit methodsFor: 'testing' stamp: 'eem 4/26/2017 14:43'!
testBcToMcPcMappingForCogMethod: cogMethod
	<doNotGenerate>
	"self disassembleMethod: cogMethod"
	"self printPCMapPairsFor: cogMethod on: Transcript"
	| aMethodObj subMethods bsOffset |
	aMethodObj := cogMethod methodObject.
	subMethods := self subMethodsAsRangesFor: cogMethod.
	subMethods first endPC: (self endPCOf: aMethodObj).
	bsOffset := self bytecodeSetOffsetFor: aMethodObj.
	self bcpcsDescriptorsAndStartpcsFor: aMethodObj bsOffset: bsOffset do:
		[:bcpc :byte :desc :nExts :startpc|
		(desc notNil and: [desc isBlockCreation]) ifTrue:
			["dead code removal may result in blocks not being generated ;-)"
			 (subMethods detect: [:sm| sm startpc = (bcpc + desc numBytes)] ifNone: [nil]) ifNotNil:
				[:subMethod|
				 subMethod endPC: bcpc + desc numBytes + (self spanFor: desc at: bcpc exts: -1 in: aMethodObj) - 1]]].
	subMethods allButFirst do:
		[:blockSubMethod| | cogBlockMethod |
		cogBlockMethod := self
								findMethodForStartBcpc: blockSubMethod startpc
								inHomeMethod: cogMethod.
		self assert: cogBlockMethod address = (blockSubMethod first - (self sizeof: CogBlockMethod))].
	self bcpcsDescriptorsAndStartpcsFor: aMethodObj bsOffset: bsOffset do:
		[:bcpc :byte :desc :nExts :startpc| | startBcpc currentSubMethod subCogMethod absMcpc mappedBcpc |
		currentSubMethod := self innermostSubMethodFor: bcpc in: subMethods startingAt: 1.
		startpc = currentSubMethod startpc ifTrue:
			[subCogMethod := currentSubMethod cogMethod.
			(subCogMethod stackCheckOffset > 0
			 and: [desc isNil or: [desc isMapped 
			 or: [inBlock = InFullBlock and: [desc isMappedInBlock]]]]) ifTrue:
				[startBcpc := subCogMethod = cogMethod
								ifTrue: [coInterpreter startPCOfMethod: aMethodObj]
								ifFalse: [currentSubMethod startpc].
				 "The first bytecode and backward branch bytecodes are mapped to their pc.
				  Other bytecodes map to their following pc."
				 absMcpc := (desc notNil
							   and: [desc isBranch
							   and: [self isBackwardBranch: desc at: bcpc exts: nExts in: aMethodObj]])
								ifTrue: "Backward branches have a special mapper"
									[mappedBcpc := bcpc - (2 * nExts).
									 self
										mcPCForBackwardBranch: mappedBcpc
										startBcpc: startBcpc
										in: subCogMethod]
								ifFalse: "All others use the generic mapper"
									[mappedBcpc := desc ifNil: [bcpc] ifNotNil: [bcpc + desc numBytes].
									 self
										mcPCFor: mappedBcpc
										startBcpc: startBcpc
										in: subCogMethod].
				 self assert: absMcpc >= (subCogMethod asInteger + subCogMethod stackCheckOffset).
				 self assert: (self bytecodePCFor: absMcpc startBcpc: startBcpc in: subCogMethod) = mappedBcpc]]]! !

!Cogit methodsFor: 'tests-method map' stamp: 'eem 7/25/2018 17:38'!
testMcToBcPcMappingForCompiledMethod: aCompiledMethod cogMethod: cogMethod
	<doNotGenerate>
	| bcMethod subMethods prevMcpc |
	"self disassembleMethod: cogMethod"
	"coInterpreter symbolicMethod: cogMethod methodObject"
	"coInterpreter printOop: cogMethod methodObject"
	"self printPCMapPairsFor: cogMethod on: Transcript"
	cogMethod stackCheckOffset = 0 ifTrue: "frameless"
		[^self].
	bcMethod := coInterpreter isCurrentImageFacade
					ifTrue: [coInterpreter objectForOop: cogMethod methodObject]
					ifFalse: [VMCompiledMethodProxy new
								for: cogMethod methodObject
								coInterpreter: coInterpreter
								objectMemory: objectMemory].
	subMethods := self subMethodsAsRangesFor: cogMethod.
	self mapFor: cogMethod do:
		[:annotation :mcpc| | subMethod subCogMethod bcpc mappedpc |
		(self isPCMappedAnnotation: annotation) ifTrue:
			[subMethod := subMethods
								detect: [:range| range includes: mcpc]
								ifNone: ["a trailing call ceNonLocalReturnTrampoline's following
										 pc is the start of a following block or the end of the map"
										subMethods detect: [:range| range includes: mcpc - 1]].
			mcpc > subMethod first ifTrue:
				[bcpc := self
							bytecodePCFor: mcpc
							startBcpc: subMethod startpc
							in: (subCogMethod := subMethod cogMethod).
				self assert: bcpc ~= 0.
				mappedpc := self mcPCFor: bcpc startBcpc: subMethod startpc in: subCogMethod.
				subCogMethod stackCheckOffset = 0
					ifTrue: [self assert: mappedpc > (subCogMethod address + self noCheckEntryOffset)]
					ifFalse: [self assert: mappedpc >= (subCogMethod address + subCogMethod stackCheckOffset)].
				"mcpc = mappedpc is obviously what we want and expect.  prevMcpc = mappedpc hacks
				 around frame building accessors where the first bytecode is mapped twice, once for the
				 stack check and once for the context inst var access.  The bytecode pc can only map
				 back to a single mcpc, the first, so the second map entry will fail without this hack."
				self assert: (mcpc = mappedpc or: [prevMcpc = mappedpc]).
				(self isSendAnnotation: annotation) ifTrue:
					[| mcSelector bcSelector |
					mcSelector := self selectorForSendAt: mcpc annotation: annotation in: aCompiledMethod.
					"sends map to the following pc.  need to find the selector for the previous pc"
					bcSelector := self selectorForSendBefore: bcpc in: bcMethod.
					self assert: (mcSelector = bcSelector
								or: [mcSelector = (self resolveToOopOrNil: bcSelector)])]].
			 prevMcpc := mcpc].
		 false "keep scanning"]! !

!Cogit methodsFor: 'tests-method map' stamp: 'eem 7/25/2018 17:39'!
testPCMappingForCompiledMethod: aCompiledMethod cogMethod: cm
	<doNotGenerate>
	methodObj := methodHeader := nil.
	cm isInteger ifTrue: "If testing in a simulation..."
		[^self testPCMappingForCompiledMethod: aCompiledMethod cogMethod: (self cogMethodSurrogateAt: cm)].
	self
		testMcToBcPcMappingForCompiledMethod: aCompiledMethod cogMethod: cm;
		testBcToMcPcMappingForCogMethod: cm! !

!Cogit methodsFor: 'debugging' stamp: 'eem 7/8/2021 14:47'!
traceLinkedSendOffset
	<api>
	^cmNoCheckEntryOffset
	 + backEnd callFullInstructionByteSize
	 + (backEnd hasLinkRegister
		ifTrue: [backEnd pushLinkRegisterByteSize]
		ifFalse: [0])! !

!Cogit methodsFor: 'debugging' stamp: 'eem 3/15/2009 06:44'!
traceStores: aBooleanOrInteger
	<doNotGenerate>
	traceStores := aBooleanOrInteger isInteger
							ifTrue: [aBooleanOrInteger]
							ifFalse: [aBooleanOrInteger ifTrue: [1] ifFalse: [0]]! !

!Cogit methodsFor: 'debugging' stamp: 'eem 2/3/2020 15:09'!
trampolineAddressFor: trampolineName
	<doNotGenerate>
	0 to: trampolineTableIndex - 3 by: 2 do:
		[:i|
		(trampolineAddresses at: i) = trampolineName ifTrue:
			[^trampolineAddresses at: i + 1]].
	^nil! !

!Cogit methodsFor: 'initialization' stamp: 'eem 12/18/2015 14:01'!
trampolineArgConstant: booleanOrInteger
	"Encode true and false and 0 to N such that they can't be confused for register numbers (including NoReg)
	 and can be tested for by isTrampolineArgConstant: and decoded by trampolineArgValue:"
	<inline: true>
	self cCode: []
		inSmalltalk: [booleanOrInteger isInteger ifFalse: [^self trampolineArgConstant: (booleanOrInteger ifTrue: [1] ifFalse: [0])]].
	self assert: booleanOrInteger >= 0.
	^-2 - booleanOrInteger "0...N => -2...-(N+2)"! !

!Cogit methodsFor: 'initialization' stamp: 'eem 12/18/2015 13:56'!
trampolineArgValue: n
	"Decode true and false and 0 to N to their C equivalents from the encoding by trampolineArgConstant:"
	<inline: true>
	^-2 - n! !

!Cogit methodsFor: 'initialization' stamp: 'tpr 6/5/2015 10:45'!
trampolineName: routinePrefix numArgs: numArgs
	<returnTypeC: #'char *'>
	<var: #routinePrefix type: #'char *'>
	^self trampolineName: routinePrefix numArgs: numArgs limit: NumSendTrampolines - 2! !

!Cogit methodsFor: 'initialization' stamp: 'nice 3/13/2016 23:20'!
trampolineName: routinePrefix numArgs: numArgs limit: argsLimit
	"Malloc a string with the contents for the trampoline table"
	<inline: true>
	<returnTypeC: #'char *'>
	<var: #routinePrefix type: #'char *'>
	<var: #numArgs type: #int>
	| theString |
	<var: #theString type: #'char *'>
	self cCode: '' inSmalltalk:
		[^routinePrefix, (numArgs <= argsLimit ifTrue: [numArgs printString] ifFalse: ['N']), 'Args'].
	theString := self malloc: (self strlen: routinePrefix) + 6.
	self s: theString pr: '%s%cArgs' in: routinePrefix tf: (numArgs <= argsLimit ifTrue: [$0 + numArgs] ifFalse: [$N]).
	^theString! !

!Cogit methodsFor: 'initialization' stamp: 'tpr 6/5/2015 09:16'!
trampolineName: routinePrefix numRegArgs: numArgs
	<returnTypeC: #'char *'>
	<var: #routinePrefix type: #'char *'>
	^self trampolineName: routinePrefix numArgs: numArgs limit: self numRegArgs! !

!Cogit methodsFor: 'disassembly' stamp: 'eem 11/24/2020 23:21'!
trampolineRangeFor: pc
	<doNotGenerate>
	| limit |
	limit := methodZoneBase - 1.
	pc > methodZoneBase ifTrue: [^nil].
	trampolineTableIndex - 1 to: 0 by: -2 do:
		[:i| | addr |
		pc >= (addr := (trampolineAddresses at: i) asInteger) ifTrue:
			[^addr to: limit].
		limit := addr - 1].
	^nil! !

!Cogit methodsFor: 'multi-threading' stamp: 'eem 1/4/2021 10:07'!
tryLockVMOwnerTo: value
	<api>
	"ceTryLockVMOwner does an atomic compare-and-swap of the vmOwner
	 variable with zero and the argument, setting vmOwner to value if it was
	 zero. It answers if the lock was zero and hence was acquired.

	 See CogThreadManager>>#tryLockVMOwnerTo: for the simulation of
	 processor thread switching which surrounds this method."
	<cmacro: '(value) ceTryLockVMOwner(value)'>
	| breakPCWasTrue |
	"(thisContext findContextSuchThat: [:ctxt| ctxt selector == #primitiveRelinquishProcessor]) ifNil:
		[self halt]."
	(breakPCWasTrue := breakPC == true) ifTrue:
		[breakPC := nil].
	processor abiMarshalArg0: value in: objectMemory memory.
	^[	| result instructions |
		lastNInstructions removeAll.
		result := self simulateLeafCallOf: ceTryLockVMOwner.
		instructions := lastNInstructions copy.
		self assert: (result ~= 0) = (coInterpreter threadManager getVMOwner = value).
		result ~= 0] ensure:
			[processor abiUnmarshal: 1.
			 breakPCWasTrue ifTrue:
				[breakPC := true]]! !

!Cogit methodsFor: 'bytecode generators' stamp: 'eem 2/10/2009 18:53'!
unknownBytecode
	^EncounteredUnknownBytecode! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 7/7/2021 11:49'!
unlinkAllSends
	<api>
	"Unlink all sends in cog methods."
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	methodZoneBase ifNil: [^self].
	self ensureWritableCodeZone.
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	methodZone voidOpenPICList.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType = CMMethod
			ifTrue:
				[self mapFor: cogMethod
					 performUntil: #unlinkIfLinkedSend:pc:ignored:
					 arg: 0]
			ifFalse:
				[cogMethod cmType ~= CMFree ifTrue:
					[methodZone freeMethod: cogMethod]].
		cogMethod := methodZone methodAfter: cogMethod].
	"After updating inline caches we need to flush the icache."
	backEnd flushICacheFrom: methodZoneBase asUnsignedInteger to: methodZone freeStart! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 7/25/2021 10:53'!
unlinkIfFreeOrLinkedSend: annotation pc: mcpc of: theSelector
	<var: #mcpc type: #'char *'>
	<var: #theSelector type: #'CogMethod *'> "To placate the C static type system..."

	<var: #nsSendCache type: #'NSSendCache *'>
	| entryPoint |

	NewspeakVM ifTrue:
		[| nsSendCache |
		 annotation = IsNSSendCall ifTrue:
			[nsSendCache := self nsSendCacheFromReturnAddress: mcpc asInteger.
			 (entryPoint := nsSendCache target) ~= 0 ifTrue:
				[ | targetMethod |
				targetMethod := self cCoerceSimple: entryPoint - cmNoCheckEntryOffset to: #'CogMethod *'.
				(targetMethod cmType = CMFree
				 or: [nsSendCache selector = (self cCoerceSimple: theSelector to: #sqInt)]) ifTrue:
					[self voidNSSendCache: nsSendCache]].
			^0 "keep scanning"]].

	(self isPureSendAnnotation: annotation) ifTrue:
		[entryPoint := backEnd callTargetFromReturnAddress: mcpc asInteger.
		 entryPoint > methodZoneBase
			ifTrue: "It's a linked send."
				[self targetMethodAndSendTableFor: entryPoint annotation: annotation into:
					[:targetMethod :sendTable| 
					 (targetMethod cmType = CMFree
					  or: [targetMethod selector = (self cCoerceSimple: theSelector to: #sqInt)]) ifTrue:
						[self unlinkSendAt: mcpc targetMethod: targetMethod sendTable: sendTable]]]].

	^0 "keep scanning"! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 09/27/2016 11:27'!
unlinkIfInvalidClassSend: annotation pc: mcpc ignored: superfluity
	<var: #mcpc type: #'char *'>
	<var: #nsSendCache type: #'NSSendCache *'>
	| entryPoint |

	NewspeakVM ifTrue:
		[| nsSendCache |
		 annotation = IsNSSendCall ifTrue:
			[nsSendCache := self nsSendCacheFromReturnAddress: mcpc asInteger.
			 (nsSendCache classTag ~= objectRepresentation illegalClassTag
			  and: [objectMemory isForwardedClassIndex: nsSendCache classTag]) ifTrue:
				[self voidNSSendCache: nsSendCache]].
			"Should we check if the enclosing object's class is forwarded as well?"
			^0 "keep scanning"].

	(self isPureSendAnnotation: annotation) ifTrue:
		[entryPoint := backEnd callTargetFromReturnAddress: mcpc asInteger.
		 entryPoint > methodZoneBase ifTrue: "It's a linked send, but maybe a super send or linked to an OpenPIC, in which case the cache tag will be a selector...."
			[self targetMethodAndSendTableFor: entryPoint annotation: annotation into:
				[:targetMethod :sendTable|
				 ((self annotationIsForUncheckedEntryPoint: annotation)
				  or: [targetMethod cmType = CMOpenPIC]) ifFalse:
					[(objectMemory isValidClassTag: (backEnd inlineCacheTagAt: mcpc asInteger)) ifFalse:
						[self unlinkSendAt: mcpc targetMethod: targetMethod sendTable: sendTable]]]]].

	^0 "keep scanning"! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 7/25/2021 10:59'!
unlinkIfLinkedSend: annotation pc: mcpc if: criterionArg
	<var: #mcpc type: #'char *'>
	<var: #criterionArg type: #'CogMethod *'> "To placate the C static type system..."

	<var: 'criterion' declareC: 'sqInt (*criterion)(CogMethod *)'>
	<var: #nsSendCache type: #'NSSendCache *'>
	| criterion entryPoint |
	criterion := criterionArg asVoidPointer.
	NewspeakVM ifTrue:
		[| nsSendCache |
		 annotation = IsNSSendCall ifTrue:
			[nsSendCache := self nsSendCacheFromReturnAddress: mcpc asInteger.
			((entryPoint := nsSendCache target) ~= 0
			 and: [self perform: criterion
						with: (self cCoerce: entryPoint - cmNoCheckEntryOffset to: #'CogMethod *')]) ifTrue:
				[self voidNSSendCache: nsSendCache].
			^0 "keep scanning"]].

	(self isPureSendAnnotation: annotation) ifTrue:
		[entryPoint := backEnd callTargetFromReturnAddress: mcpc asInteger.
		 entryPoint > methodZoneBase ifTrue: "It's a linked send."
			[self targetMethodAndSendTableFor: entryPoint annotation: annotation into:
				[:targetMethod :sendTable| 
				 (self perform: criterion with: targetMethod) ifTrue:
					[self unlinkSendAt: mcpc targetMethod: targetMethod sendTable: sendTable]]]].

	^0 "keep scanning"! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 09/27/2016 11:27'!
unlinkIfLinkedSend: annotation pc: mcpc ignored: superfluity
	<var: #mcpc type: #'char *'>
	<var: #nsSendCache type: #'NSSendCache *'>
	| entryPoint |

	NewspeakVM ifTrue:
		[| nsSendCache |
		 annotation = IsNSSendCall ifTrue:
			[nsSendCache := self nsSendCacheFromReturnAddress: mcpc asInteger.
			nsSendCache classTag ~= objectRepresentation illegalClassTag ifTrue: "Send is linked"
				[self voidNSSendCache: nsSendCache].
			^0 "keep scanning"]].

	(self isPureSendAnnotation: annotation) ifTrue:
		[entryPoint := backEnd callTargetFromReturnAddress: mcpc asInteger.
		 entryPoint > methodZoneBase
			ifTrue: "It's a linked send."
				[self targetMethodAndSendTableFor: entryPoint annotation: annotation into:
					[:targetMethod :sendTable| 
					 self unlinkSendAt: mcpc targetMethod: targetMethod sendTable: sendTable]]].

	^0 "keep scanning"! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 09/27/2016 11:27'!
unlinkIfLinkedSend: annotation pc: mcpc of: theSelector
	<var: #mcpc type: #'char *'>
	<var: #nsSendCache type: #'NSSendCache *'>
	| entryPoint |

	NewspeakVM ifTrue:
		[| nsSendCache |
		 annotation = IsNSSendCall ifTrue:
			[nsSendCache := self nsSendCacheFromReturnAddress: mcpc asInteger.
			nsSendCache selector = theSelector ifTrue:
				[self voidNSSendCache: nsSendCache].
			^0 "keep scanning"]].

	(self isPureSendAnnotation: annotation) ifTrue:
		[entryPoint := backEnd callTargetFromReturnAddress: mcpc asInteger.
		 entryPoint > methodZoneBase
			ifTrue: "It's a linked send."
				[self targetMethodAndSendTableFor: entryPoint annotation: annotation into:
					[:targetMethod :sendTable| 
					 targetMethod selector = theSelector ifTrue:
						[self unlinkSendAt: mcpc targetMethod: targetMethod sendTable: sendTable]]]].

	^0 "keep scanning"! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 09/27/2016 11:27'!
unlinkIfLinkedSend: annotation pc: mcpc to: theCogMethod
	<var: #mcpc type: #'char *'>
	<var: #nsSendCache type: #'NSSendCache *'>
	| entryPoint |

	NewspeakVM ifTrue:
		[| nsSendCache |
		 annotation = IsNSSendCall ifTrue:
			[nsSendCache := self nsSendCacheFromReturnAddress: mcpc asInteger.
			(entryPoint := nsSendCache target) ~= 0 ifTrue:
				[ | targetMethod |
				targetMethod := entryPoint - cmNoCheckEntryOffset.
				targetMethod = theCogMethod ifTrue:
					[self voidNSSendCache: nsSendCache]].
			^0 "keep scanning"]].

	(self isPureSendAnnotation: annotation) ifTrue:
		[entryPoint := backEnd callTargetFromReturnAddress: mcpc asInteger.
		 entryPoint > methodZoneBase
			ifTrue: "It's a linked send."
				[self targetMethodAndSendTableFor: entryPoint annotation: annotation into:
					[:targetMethod :sendTable| 
					 targetMethod asInteger = theCogMethod ifTrue:
						[self unlinkSendAt: mcpc targetMethod: targetMethod sendTable: sendTable]]]].

	^0 "keep scanning"! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 09/27/2016 11:27'!
unlinkIfLinkedSendToFree: annotation pc: mcpc ignored: superfluity
	<var: #mcpc type: #'char *'>
	<var: #nsSendCache type: #'NSSendCache *'>
	<var: #nsTargetMethod type: #'CogMethod *'>
	| entryPoint |

	NewspeakVM ifTrue:
		[| nsSendCache nsTargetMethod |
		 annotation = IsNSSendCall ifTrue:
			[nsSendCache := self nsSendCacheFromReturnAddress: mcpc asInteger.
			(entryPoint := nsSendCache target) ~= 0 ifTrue: "It's a linked send."
				[nsTargetMethod := self cCoerceSimple: entryPoint - cmNoCheckEntryOffset to: #'CogMethod *'.
				nsTargetMethod cmType = CMFree ifTrue:
					[self voidNSSendCache: nsSendCache]].
			^0 "keep scanning"]].

	(self isPureSendAnnotation: annotation) ifTrue:
		[entryPoint := backEnd callTargetFromReturnAddress: mcpc asInteger.
		 entryPoint > methodZoneBase ifTrue: "It's a linked send."
			[self targetMethodAndSendTableFor: entryPoint annotation: annotation into:
				[:targetMethod :sendTable| 
				 targetMethod cmType = CMFree ifTrue:
					[self unlinkSendAt: mcpc targetMethod: targetMethod sendTable: sendTable]]]].

	^0 "keep scanning"! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 7/9/2021 13:56'!
unlinkSendAt: mcpc targetMethod: targetMethod sendTable: sendTable
	<inline: true>
	| unlinkedRoutine |
	unlinkedRoutine := sendTable at: (targetMethod cmNumArgs min: NumSendTrampolines - 1).
	self setCodeModified.
	backEnd
		rewriteInlineCacheAt: mcpc asInteger
		tag: (backEnd inlineCacheValueForSelector: targetMethod selector in: enumeratingCogMethod)
		target: unlinkedRoutine! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 7/7/2021 11:49'!
unlinkSendsLinkedForInvalidClasses
	<api>
	<option: #SpurObjectMemory>
	"Unlink all sends in cog methods whose class tag is that of a forwarded class."
	| cogMethod freedPIC |
	<var: #cogMethod type: #'CogMethod *'>
	methodZoneBase ifNil: [^self].
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	codeModified := freedPIC := false.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType = CMMethod
			ifTrue:
				[self mapFor: cogMethod
					 performUntil: #unlinkIfInvalidClassSend:pc:ignored:
					 arg: 0]
			ifFalse:
				[(cogMethod cmType = CMClosedPIC
				  and: [self cPICHasForwardedClass: cogMethod]) ifTrue:
					[methodZone freeMethod: cogMethod.
					 freedPIC := true]].
		cogMethod := methodZone methodAfter: cogMethod].
	freedPIC
		ifTrue: [self unlinkSendsToFree]
		ifFalse:
			[codeModified ifTrue: "After possibly updating inline caches we need to flush the icache."
				[backEnd flushICacheFrom: methodZoneBase asUnsignedInteger to: methodZone freeStart]]! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 7/25/2021 10:51'!
unlinkSendsOf: selector isMNUSelector: isMNUSelector
	<api>
	"Unlink all sends in cog methods. Free all Closed PICs with the selector,
	 or with an MNU case if isMNUSelector.  First check if any method actually
	 has the selector; if not there can't be any linked send to it.  This routine
	 (including descendents) is performance critical.  It contributes perhaps
	 30% of entire execution time in Compiler recompileAll."
	| cogMethod mustScanAndUnlink |
	<var: #cogMethod type: #'CogMethod *'>
	methodZoneBase ifNil: [^self].
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	mustScanAndUnlink := false.
	isMNUSelector
		ifTrue:
			[[cogMethod < methodZone limitZony] whileTrue:
				[cogMethod cmType ~= CMFree ifTrue:
					[cogMethod cpicHasMNUCase
						ifTrue:
							[self assert: cogMethod cmType = CMClosedPIC.
							 methodZone freeMethod: cogMethod.
							 mustScanAndUnlink := true]
						ifFalse:
							[cogMethod selector = selector ifTrue:
								[mustScanAndUnlink := true.
								 cogMethod cmType = CMClosedPIC ifTrue:
									[methodZone freeMethod: cogMethod]]]].
				 cogMethod := methodZone methodAfter: cogMethod]]
		ifFalse:
			[[cogMethod < methodZone limitZony] whileTrue:
				[(cogMethod cmType ~= CMFree
				  and: [cogMethod selector = selector]) ifTrue:
					[mustScanAndUnlink := true.
					 cogMethod cmType = CMClosedPIC ifTrue:
						[methodZone freeMethod: cogMethod]].
				 cogMethod := methodZone methodAfter: cogMethod]].
	mustScanAndUnlink ifFalse:
		[^self].
	codeModified := false.
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType = CMMethod ifTrue:
			[self mapFor: cogMethod
				 performUntil: #unlinkIfFreeOrLinkedSend:pc:of:
				 arg: (self cCoerceSimple: selector to: #'CogMethod *')].
		cogMethod := methodZone methodAfter: cogMethod].
	codeModified ifTrue: "After possibly updating inline caches we need to flush the icache."
		[backEnd flushICacheFrom: methodZoneBase asUnsignedInteger to: methodZone freeStart].
	"And ensure code zone is executable.  May merely have freed methods..."
	self ensureExecutableCodeZone! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 7/25/2021 10:35'!
unlinkSendsTo: targetMethodObject andFreeIf: freeIfTrue
	<api>
	"Unlink all sends in cog methods to a particular target method.
	 If targetMethodObject isn't actually a method (perhaps being
	 used via invokeAsMethod) then there's nothing to do."
	| cogMethod targetMethod freedPIC |
	<var: #cogMethod type: #'CogMethod *'>
	<var: #targetMethod type: #'CogMethod *'>
	((objectMemory isOopCompiledMethod: targetMethodObject)
	and: [coInterpreter methodHasCogMethod: targetMethodObject]) ifFalse:
		[^self].
	targetMethod := coInterpreter cogMethodOf: targetMethodObject.
	methodZoneBase ifNil: [^self].
	self ensureWritableCodeZone.
	codeModified := freedPIC := false.
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType = CMMethod
			ifTrue:
				[self mapFor: cogMethod
					 performUntil: #unlinkIfLinkedSend:pc:to:
					 arg: targetMethod]
			ifFalse:
				[(cogMethod cmType = CMClosedPIC
				  and: [self cPIC: cogMethod HasTarget: targetMethod]) ifTrue:
					[methodZone freeMethod: cogMethod.
					 freedPIC := true]].
		cogMethod := methodZone methodAfter: cogMethod].
	freeIfTrue ifTrue: [self freeMethod: targetMethod].
	freedPIC
		ifTrue: [self unlinkSendsToFree]
		ifFalse:
			[codeModified ifTrue: "After possibly updating inline caches we need to flush the icache."
				[backEnd flushICacheFrom: methodZoneBase asUnsignedInteger to: methodZone freeStart]].
	self ensureExecutableCodeZone! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 7/7/2021 11:49'!
unlinkSendsToFree
	<api>
	"Unlink all sends in cog methods to free methods and/or pics."
	| cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	methodZoneBase ifNil: [^self].
	codeModified := false.
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType = CMMethod
			ifTrue:
				[self mapFor: cogMethod
					 performUntil: #unlinkIfLinkedSendToFree:pc:ignored:
					 arg: 0]
			ifFalse:
				[cogMethod cmType = CMClosedPIC ifTrue:
					[self assert: (self noTargetsFreeInClosedPIC: cogMethod)]].
		cogMethod := methodZone methodAfter: cogMethod].
	codeModified ifTrue: "After possibly updating inline caches we need to flush the icache."
		[backEnd flushICacheFrom: methodZoneBase asUnsignedInteger to: methodZone freeStart]! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 7/25/2021 16:53'!
unlinkSendsToMethodsSuchThat: criterion AndFreeIf: freeIfTrue
	<api>
	"Unlink all sends in cog methods to methods with a machine code
	 primitive, and free machine code primitive methods if freeIfTrue.
	 To avoid having to scan PICs, free any and all PICs"
	<var: 'criterion' declareC: 'sqInt (*criterion)(CogMethod *)'>
	| cogMethod freedSomething |
	<var: #cogMethod type: #'CogMethod *'>
	methodZoneBase ifNil: [^self].
	self cCode: nil inSmalltalk: [debugAPISelector := nil].
	codeModified := freedSomething := false.
	cogMethod := self cCoerceSimple: methodZoneBase to: #'CogMethod *'.
	[cogMethod < methodZone limitZony] whileTrue:
		[cogMethod cmType = CMMethod
			ifTrue:
				[(freeIfTrue
				  and: [self perform: criterion with: cogMethod])
					ifTrue:
						[methodZone freeMethod: cogMethod.
						 freedSomething := true]
					ifFalse:
						[self mapFor: cogMethod
							 performUntil: #unlinkIfLinkedSend:pc:if:
							 arg: criterion]]
			ifFalse:
				[cogMethod cmType = CMClosedPIC ifTrue:
					[methodZone freeMethod: cogMethod.
					 freedSomething := true]].
		cogMethod := methodZone methodAfter: cogMethod].
	freedSomething
		ifTrue: [self unlinkSendsToFree]
		ifFalse:
			[codeModified ifTrue: "After possibly updating inline caches we need to flush the icache."
				[backEnd flushICacheFrom: methodZoneBase asUnsignedInteger to: methodZone freeStart]]! !

!Cogit methodsFor: 'accessing' stamp: 'eem 10/29/2020 13:38'!
varBaseAddress
	<cmacro: '() varBaseAddress'>
	^varBaseAddress! !

!Cogit methodsFor: 'jit - api' stamp: 'eem 8/3/2021 21:04'!
voidCogCompiledCode
	<api>
	methodZone clearCogCompiledCode.
	"code must still be executable, e.g. for ceInvokeInterpret"
	self ensureExecutableCodeZone! !

!Cogit methodsFor: 'in-line cacheing' stamp: 'eem 4/4/2015 18:24'!
voidNSSendCache: nsSendCache
	<inline: true>
	nsSendCache classTag: objectRepresentation illegalClassTag; enclosingObject: 0; target: 0! !

!Cogit methodsFor: 'debug printing' stamp: 'eem 1/7/2021 15:17'!
warnMultiple: cogMethod selectors: aSelectorOop
	<inline: true>
	<var: 'cogMethod' type: #'CogMethod *'>
	 'Warning, attempt to use method with selector %.*s and selector %.*s\n'
		f: #stderr
		printf: {self cCoerceSimple: (objectMemory numBytesOf: cogMethod selector) to: #int.
				self cCoerceSimple: (objectMemory firstIndexableField: cogMethod selector) to: #'char *'.
				self cCoerceSimple: (objectMemory numBytesOf: aSelectorOop) to: #int.
				self cCoerceSimple: (objectMemory firstIndexableField: aSelectorOop) to: #'char *' }! !

!Cogit methodsFor: 'debug printing' stamp: 'eem 12/22/2019 18:22'!
whereIsMaybeCodeThing: aThing
	<doNotGenerate>
	aThing = self varBaseAddress ifTrue:
		[^' is variables base'].
	(methodZone whereIsMaybeCodeThing: aThing) ifNotNil:
		[:where| ^where].
	^((simulatedTrampolines includesKey: aThing)
	or: [(simulatedVariableGetters includesKey: aThing)
	or: [simulatedVariableSetters includesKey: aThing]]) ifTrue:
		[simulatedAddresses keyAtValue: aThing ifAbsent: ['variable lookup failure']]! !

!Cogit methodsFor: 'analysis' stamp: 'eem 11/18/2013 12:26'!
withAnnotation: annotation mcpc: mcpc evaluate: binaryBlock
	<doNotGenerate>
	binaryBlock value: annotation value: mcpc.
	^0 "keep scanning"! !

!Cogit methodsFor: 'garbage collection' stamp: 'eem 10/17/2010 13:17'!
withAnnotation: annotation pc: mcpc evaluate: aBinaryBlock
	<doNotGenerate>
	^(aBinaryBlock value: annotation value: mcpc) ifTrue: [1] ifFalse: [0]! !

!Cogit methodsFor: 'simulation processor access' stamp: 'eem 1/3/2021 13:25'!
withProcessorHaltedDo: aBlock
	^processorLock critical:
		[| oldBreakPC oldSingleStep |
		 oldBreakPC := breakPC.
		 oldSingleStep := singleStep.
		 breakPC := singleStep := true.
		 aBlock ensure:
			[singleStep := oldSingleStep.
			 breakPC := oldBreakPC]]! !

!Cogit methodsFor: 'generate machine code - dual mapped zone support' stamp: 'eem 1/28/2020 17:35'!
writableBlockMethodFor: aCogMethodOrInteger
	<inline: #always>
	^self cCoerceSimple: aCogMethodOrInteger asUnsignedInteger + codeToDataDelta to: #'CogBlockMethod *'! !

!Cogit methodsFor: 'generate machine code - dual mapped zone support' stamp: 'eem 1/28/2020 17:34'!
writableMethodFor: aCogMethodOrInteger
	<inline: #always>
	^self cCoerceSimple: aCogMethodOrInteger asUnsignedInteger + codeToDataDelta to: #'CogMethod *'! !

!Cogit methodsFor: 'accessing' stamp: 'eem 10/1/2016 14:59'!
zeroOpcodeIndex
	"Access for the object representations when they need to prepend code to trampolines."
	"Eliminate stale dependent info."
	0 to: opcodeIndex - 1 do:
		[:i| (abstractOpcodes at: i) dependent: nil].
	self zeroOpcodeIndexForNewOpcodes! !

!Cogit methodsFor: 'accessing' stamp: 'eem 10/1/2016 14:24'!
zeroOpcodeIndexForNewOpcodes
	"Access for the object representations when they need to prepend code to trampolines."
	opcodeIndex := 0.
	literalsManager resetLiterals! !

Cogit subclass: #CogitForTests
	instanceVariableNames: 'codeBytes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!CogitForTests methodsFor: 'accessing-memory' stamp: 'eem 12/18/2019 22:07'!
byteAt: zeroRelativeIndex put: aByte
	^codeBytes at: zeroRelativeIndex + 1 put: aByte! !

!CogitForTests methodsFor: 'accessing' stamp: 'eem 12/18/2019 21:20'!
bytecodeFixupClass
	^CogBytecodeFixup! !

!CogitForTests methodsFor: 'accessing' stamp: 'eem 12/18/2019 21:52'!
coInterpreter
	^self! !

!CogitForTests methodsFor: 'generate machine code' stamp: 'eem 12/19/2019 15:12'!
codeSize
	"Answer the size of code generated, which includes the offset at which the code was generated."
	^codeBytes size! !

!CogitForTests methodsFor: 'generate machine code' stamp: 'eem 12/19/2019 11:08'!
generatedMachineCodeAt: startAddress
	| codeSize |
	self computeMaximumSizes.
	codeSize := self generateInstructionsAt: startAddress.
	codeBytes := ByteArray new: startAddress + codeSize.
	backEnd stopsFrom: 0 to: startAddress - 1.
	self outputInstructionsAt: startAddress.
	^codeBytes! !

!CogitForTests methodsFor: 'initialization' stamp: 'eem 12/18/2019 22:06'!
initialize
	"Ugh, this needs fixing!!!!"
	CogCompilerClass := nil.
	self class initializeMiscConstants.
	objectMemory := self.
	backEnd := CogCompilerClass for: self.
	methodLabel := CogCompilerClass for: self.
	(literalsManager := backEnd class literalsManagerClass new) cogit: self.
	debugOpcodeIndices := #().
	self initializeBackend.
	self allocateOpcodes: 64 bytecodes: 64.
	self resetGen! !

!CogitForTests methodsFor: 'accessing-memory' stamp: 'eem 12/18/2019 21:37'!
long32At: zeroRelativeIndex put: word
	^codeBytes unsignedLongAt: zeroRelativeIndex + 1 put: word! !

!CogitForTests methodsFor: 'accessing-memory' stamp: 'eem 12/18/2019 22:12'!
long64At: zeroRelativeIndex put: aByte
	^codeBytes unsignedLong64At: zeroRelativeIndex + 1 put: aByte! !

!CogitForTests methodsFor: 'accessing-memory' stamp: 'eem 12/19/2019 00:03'!
longAt: zeroRelativeByteIndex put: aWord
	^codeBytes integerAt: zeroRelativeByteIndex + 1 put: aWord size: backEnd class wordSize signed: false! !

!CogitForTests methodsFor: 'accessing' stamp: 'eem 12/18/2019 21:53'!
lookupAddress: address
	^nil! !

!CogitForTests methodsFor: 'testing' stamp: 'eem 12/18/2019 21:16'!
resetGen
	opcodeIndex := 0.
	literalsManager resetLiterals! !

!CogitForTests methodsFor: 'accessing-memory' stamp: 'eem 12/18/2019 21:37'!
stopsFrom: start to: stop
	backEnd stopsFrom: start to: stop! !

!CogitForTests methodsFor: 'accessing-memory' stamp: 'eem 12/18/2019 21:41'!
stopsFrom: start to: stop in: aByteArray
	| save |
	save := codeBytes.
	codeBytes := aByteArray.
	[backEnd stopsFrom: start to: stop]
		ensure: [codeBytes := save]! !

!CogitForTests methodsFor: 'accessing' stamp: 'eem 12/18/2019 21:53'!
wordSize
	^backEnd class wordSize! !

Cogit subclass: #SimpleStackBasedCogit
	instanceVariableNames: 'primitiveGeneratorTable introspectionDataIndex introspectionData'
	classVariableNames: ''
	poolDictionaries: 'VMMethodCacheConstants VMObjectIndices VMSqueakClassIndices'
	category: 'VMMaker-JIT'!

!SimpleStackBasedCogit commentStamp: '<historical>' prior: 0!
I am the stage one JIT for Cog that does not attempt to eliminate the stack via deferred code generation.!

!SimpleStackBasedCogit class methodsFor: 'translation' stamp: 'eem 6/7/2018 15:35'!
ancilliaryClasses
	"Answer any extra classes to be included in the translation."
	^super ancilliaryClasses, (self objectRepresentationClass withAllSuperclasses copyUpThrough: CogObjectRepresentation) reverse! !

!SimpleStackBasedCogit class methodsFor: 'documentation' stamp: 'eem 3/21/2013 15:35'!
callingConvention
	"The Smalltalk-to-Smalltalk calling convention for SimpleStackBasedCogit is
	 designed to be congruent with the interpreter and convenient for inline cacheing.
	 For inline cacheing it is convenient if the receiver is in a register.

	 Hence the calling convention is:
	
		On call ReceiverResultReg (edx on x86) contains the receiver, and the receiver
		and arguments are all on the stack, receiver furthest from top-of-stack.
	
		If the number of arguments is 3 or greater then the argument count is passed in
		SendNumArgsReg (this is for the linking run-time routine; it is ignored in linked sends).

		On return result is in ReceiverResultReg.  The callee removes arguments from the stack.
		The caller pushes the result if the result is used."! !

!SimpleStackBasedCogit class methodsFor: 'translation' stamp: 'eem 7/25/2021 11:19'!
declareCVarsIn: aCCodeGenerator

	aCCodeGenerator vmClass primitiveTable ifNotNil:
		[:bytecodeGenTable|
		aCCodeGenerator
			var: #primitiveGeneratorTable
				declareC: 'static PrimitiveDescriptor primitiveGeneratorTable[MaxCompiledPrimitiveIndex+1]',
							(self tableInitializerFor: aCCodeGenerator vmClass primitiveTable
								in: aCCodeGenerator)].
	aCCodeGenerator
		var: #primSetFunctionLabel type: #'AbstractInstruction *';
		var: #primInvokeInstruction type: #'AbstractInstruction *'! !

!SimpleStackBasedCogit class methodsFor: 'class initialization' stamp: 'nice 11/9/2019 10:19'!
initializeBytecodeTableForNewspeakV4
	"SimpleStackBasedCogit initializeBytecodeTableForNewspeakV4"

	BytecodeSetHasDirectedSuperSend := true.
	BytecodeSetHasExtensions := true.
	NSSendIsPCAnnotated := false. "IsNSSendCall used by SendAbsentImplicit"
	FirstSpecialSelector := 80.
	NumSpecialSelectors := 32.
	self flag:
'Special selector send class must be inlined to agree with the interpreter, which
 inlines class.  If class is sent to e.g. a general instance of ProtoObject then unless
 class is inlined there will be an MNU.  It must be that the Cointerpreter and Cogit
 have identical semantics.  We get away with not hardwiring the other special
 selectors either because in the Cointerpreter they are not inlined or because they
 are inlined only to instances of classes for which there will always be a method.'.
	self generatorTableFrom: #(
		"1 byte bytecodes"
		(1    0   15 genPushReceiverVariableBytecode isInstVarRef)
		(1  16   31 genPushLiteralVariable16CasesBytecode needsFrameNever: 1)
		(1  32   63 genPushLiteralConstantBytecode needsFrameNever: 1)
		(1  64   75 genPushTemporaryVariableBytecode)
		(1  76   76 genPushReceiverBytecode)
		(1  77   77 genExtPushPseudoVariableOrOuterBytecode)
		(1  78   78 genPushConstantZeroBytecode)
		(1  79   79 genPushConstantOneBytecode)

		(1   80 101 genSpecialSelectorSend isMapped) "#+ #- #< #> #<= #>= #= #~= #* #/ #\\ #@ #bitShift: #// #bitAnd: #bitOr: #at: #at:put: #size #next #nextPut: #atEnd"
		(1 102 102 genSpecialSelectorEqualsEquals needsFrameNever: notMapped -1) "not mapped because it is directly inlined (for now)"
		(1 103 103 genSpecialSelectorClass needsFrameNever: notMapped 0) "not mapped because it is directly inlined (for now)"
		(1 104 104 genSpecialSelectorNotEqualsEquals needsFrameNever: notMapped -1) "not mapped because it is directly inlined (for now)"
		(1 105 111 genSpecialSelectorSend isMapped) "#value #value: #do: #new #new: #x #y"

		(1 112 127 genSendLiteralSelector0ArgsBytecode isMapped)
		(1 128 143 genSendLiteralSelector1ArgBytecode isMapped)
		(1 144 159 genSendLiteralSelector2ArgsBytecode isMapped)
		(1 160 175	genSendAbsentImplicit0ArgsBytecode isMapped hasIRC)

		(1 176 183 genStoreAndPopReceiverVariableBytecode isInstVarRef isMappedIfImmutability needsFrameIfImmutability: -1)

		(1 184 191 genStoreAndPopTemporaryVariableBytecode)

		(1 192 199 genShortUnconditionalJump	branch v3:ShortForward:Branch:Distance:)
		(1 200 207 genShortJumpIfTrue			branch isBranchTrue isMapped "because of mustBeBoolean"
													v3:ShortForward:Branch:Distance:)
		(1 208 215 genShortJumpIfFalse			branch isBranchFalse isMapped "because of mustBeBoolean"
													v3:ShortForward:Branch:Distance:)

		(1 216 216 genReturnReceiver				return needsFrameIfInBlock: isMappedInBlock 0)
		(1 217 217 genReturnTopFromMethod		return needsFrameIfInBlock: isMappedInBlock -1)
		(1 218 218 genExtReturnTopFromBlock	return needsFrameNever: -1)

		(1 219 219 duplicateTopBytecode			needsFrameNever: 1)
		(1 220 220 genPopStackBytecode			needsFrameNever: -1)
		(1 221 221 genExtNopBytecode			needsFrameNever: 0)
		(1 222 223	unknownBytecode)

		"2 byte bytecodes"
		(2 224 224 extABytecode extension)
		(2 225 225 extBBytecode extension)
		(2 226 226 genExtPushReceiverVariableBytecode isInstVarRef)
		(2 227 227 genExtPushLiteralVariableBytecode		needsFrameNever: 1)
		(2 228 228 genExtPushLiteralBytecode					needsFrameNever: 1)
		(2 229 229 genExtPushIntegerBytecode				needsFrameNever: 1)
		(2 230 230 genLongPushTemporaryVariableBytecode)
		(2 231 231 genPushNewArrayBytecode)
		(2 232 232 genExtStoreReceiverVariableBytecode isInstVarRef isMappedIfImmutability)
		(2 233 233 genExtStoreLiteralVariableBytecode isMappedIfImmutability)
		(2 234 234 genLongStoreTemporaryVariableBytecode)
		(2 235 235 genExtStoreAndPopReceiverVariableBytecode isInstVarRef isMappedIfImmutability)
		(2 236 236 genExtStoreAndPopLiteralVariableBytecode isMappedIfImmutability)
		(2 237 237 genLongStoreAndPopTemporaryVariableBytecode)

		(2 238 238 genExtSendBytecode isMapped)
		(2 239 239 genExtSendSuperBytecode isMapped)
		(2 240 240 genExtSendAbsentImplicitBytecode isMapped hasIRC)
		(2 241 241 genExtSendAbsentDynamicSuperBytecode isMapped hasIRC)

		(2 242 242 genExtUnconditionalJump	branch isMapped "because of interrupt check" v4:Long:Branch:Distance:)
		(2 243 243 genExtJumpIfTrue			branch isBranchTrue isMapped "because of mustBeBoolean" v4:Long:Branch:Distance:)
		(2 244 244 genExtJumpIfFalse			branch isBranchFalse isMapped "because of mustBeBoolean" v4:Long:Branch:Distance:)

		(2 245 245 genExtSendAbsentSelfBytecode isMapped hasIRC)

		(2 246 248	unknownBytecode)

		"3 byte bytecodes"
		(3 249 249 genCallPrimitiveBytecode)
		(3 250 250 genPushRemoteTempLongBytecode)
		(3 251 251 genStoreRemoteTempLongBytecode)
		(3 252 252 genStoreAndPopRemoteTempLongBytecode)
		(3 253 253 genExtPushClosureBytecode block v4:Block:Code:Size:)
		(3 254 254 genExtSendAbsentOuterBytecode isMapped hasIRC)

		(3 255 255	unknownBytecode))! !

!SimpleStackBasedCogit class methodsFor: 'class initialization' stamp: 'eem 3/10/2018 12:38'!
initializeBytecodeTableForSistaV1
	"SimpleStackBasedCogit initializeBytecodeTableForSistaV1"

	BytecodeSetHasDirectedSuperSend := true.
	FirstSpecialSelector := 96.
	NumSpecialSelectors := 32.
	self flag:
'Special selector send class must be inlined to agree with the interpreter, which
 inlines class.  If class is sent to e.g. a general instance of ProtoObject then unless
 class is inlined there will be an MNU.  It must be that the Cointerpreter and Cogit
 have identical semantics.  We get away with not hardwiring the other special
 selectors either because in the Cointerpreter they are not inlined or because they
 are inlined only to instances of classes for which there will always be a method.'.
	self generatorTableFrom: #(
		"1 byte bytecodes"
		"pushes"
		(1    0   15 genPushReceiverVariableBytecode isInstVarRef)
		(1  16   31 genPushLiteralVariable16CasesBytecode	needsFrameNever: 1)
		(1  32   63 genPushLiteralConstantBytecode			needsFrameNever: 1)
		(1  64   75 genPushTemporaryVariableBytecode)
		(1  76   76 genPushReceiverBytecode)
		(1  77   77 genPushConstantTrueBytecode				needsFrameNever: 1)
		(1  78   78 genPushConstantFalseBytecode			needsFrameNever: 1)
		(1  79   79 genPushConstantNilBytecode				needsFrameNever: 1)
		(1  80   80 genPushConstantZeroBytecode				needsFrameNever: 1)
		(1  81   81 genPushConstantOneBytecode				needsFrameNever: 1)
		(1  82   82 genExtPushPseudoVariable)
		(1  83   83 duplicateTopBytecode						needsFrameNever: 1)

		(1  84   87 unknownBytecode)

		"returns"
		(1  88   88 genReturnReceiver				return needsFrameIfInBlock: isMappedInBlock 0)
		(1  89   89 genReturnTrue					return needsFrameIfInBlock: isMappedInBlock 0)
		(1  90   90 genReturnFalse					return needsFrameIfInBlock: isMappedInBlock 0)
		(1  91   91 genReturnNil					return needsFrameIfInBlock: isMappedInBlock 0)
		(1  92   92 genReturnTopFromMethod		return needsFrameIfInBlock: isMappedInBlock -1)
		(1  93   93 genReturnNilFromBlock			return needsFrameNever: -1)
		(1  94   94 genReturnTopFromBlock		return needsFrameNever: -1)
		(1  95   95 genExtNopBytecode			needsFrameNever: 0)

		"sends"
		(1   96 117 genSpecialSelectorSend isMapped) "#+ #- #< #> #<= #>= #= #~= #* #/ #\\ #@ #bitShift: #// #bitAnd: #bitOr: #at: #at:put: #size #next #nextPut: #atEnd"
		(1 118 118 genSpecialSelectorEqualsEquals needsFrameNever: notMapped -1) "not mapped because it is directly inlined (for now)"
		(1 119 119 genSpecialSelectorClass needsFrameNever: notMapped 0) "not mapped because it is directly inlined (for now)"
		(1 120 120 genSpecialSelectorNotEqualsEquals needsFrameNever: notMapped -1) "not mapped because it is directly inlined (for now)"
		(1 121 127 genSpecialSelectorSend isMapped) "#value #value: #do: #new #new: #x #y"

		(1 128 143 genSendLiteralSelector0ArgsBytecode isMapped)
		(1 144 159 genSendLiteralSelector1ArgBytecode isMapped)
		(1 160 175 genSendLiteralSelector2ArgsBytecode isMapped)

		"jumps"
		(1 176 183 genShortUnconditionalJump	branch v3:ShortForward:Branch:Distance:)
		(1 184 191 genShortJumpIfTrue			branch isBranchTrue isMapped "because of mustBeBoolean"
													v3:ShortForward:Branch:Distance:)
		(1 192 199 genShortJumpIfFalse			branch isBranchFalse isMapped "because of mustBeBoolean"
													v3:ShortForward:Branch:Distance:)

		"stores"
		(1 200 207 genStoreAndPopReceiverVariableBytecode isInstVarRef isMappedIfImmutability needsFrameIfImmutability: -1)
		(1 208 215 genStoreAndPopTemporaryVariableBytecode)

		(1 216 216 genPopStackBytecode needsFrameNever: -1)

		(1 217 217 genUnconditionalTrapBytecode isMapped)

		(1 218 223 unknownBytecode)

		"2 byte bytecodes"
		(2 224 224 extABytecode extension)
		(2 225 225 extBBytecode extension)

		"pushes"
		(2 226 226 genExtPushReceiverVariableBytecode isInstVarRef)		"Needs a frame for context inst var access"
		(2 227 227 genExtPushLiteralVariableBytecode		needsFrameNever: 1)
		(2 228 228 genExtPushLiteralBytecode					needsFrameNever: 1)
		(2 229 229 genLongPushTemporaryVariableBytecode)
		(2 230 230 unknownBytecode)
		(2 231 231 genPushNewArrayBytecode)
		(2 232 232 genExtPushIntegerBytecode				needsFrameNever: 1)
		(2 233 233 genExtPushCharacterBytecode				needsFrameNever: 1)

		"returns"
		"sends"
		(2 234 234 genExtSendBytecode isMapped)
		(2 235 235 genExtSendSuperBytecode isMapped)

		"sista bytecodes"
		(2 236 236 unknownBytecode)

		"jumps"
		(2 237 237 genExtUnconditionalJump	branch isMapped "because of interrupt check" v4:Long:Branch:Distance:)
		(2 238 238 genExtJumpIfTrue			branch isBranchTrue isMapped "because of mustBeBoolean" v4:Long:Branch:Distance:)
		(2 239 239 genExtJumpIfFalse			branch isBranchFalse isMapped "because of mustBeBoolean" v4:Long:Branch:Distance:)

		"stores"
		(2 240 240 genExtStoreAndPopReceiverVariableBytecode isInstVarRef isMappedIfImmutability)
		(2 241 241 genExtStoreAndPopLiteralVariableBytecode isMappedIfImmutability)
		(2 242 242 genLongStoreAndPopTemporaryVariableBytecode)
		(2 243 243 genExtStoreReceiverVariableBytecode isInstVarRef isMappedIfImmutability)
		(2 244 244 genExtStoreLiteralVariableBytecode isMappedIfImmutability)
		(2 245 245 genLongStoreTemporaryVariableBytecode)

		(2 246 247	unknownBytecode)

		"3 byte bytecodes"
		(3 248 248 genCallPrimitiveBytecode)
		(3 249 249 genExtPushFullClosureBytecode) 
		(3 250 250 genExtPushClosureBytecode block v4:Block:Code:Size:)
		(3 251 251 genPushRemoteTempLongBytecode)
		(3 252 252 genStoreRemoteTempLongBytecode)
		(3 253 253 genStoreAndPopRemoteTempLongBytecode)

		(3 254 255	unknownBytecode))! !

!SimpleStackBasedCogit class methodsFor: 'class initialization' stamp: 'eem 6/7/2018 15:20'!
initializeBytecodeTableForSqueakV3PlusClosures
	"SimpleStackBasedCogit initializeBytecodeTableForSqueakV3PlusClosures"

	FirstSpecialSelector := 176.
	NumSpecialSelectors := 32.
	self flag:
'Special selector send class must be inlined to agree with the interpreter, which
 inlines class.  If class is sent to e.g. a general instance of ProtoObject then unless
 class is inlined there will be an MNU.  It must be that the Cointerpreter and Cogit
 have identical semantics.  We get away with not hardwiring the other special
 selectors either because in the Cointerpreter they are not inlined or because they
 are inlined only to instances of classes for which there will always be a method.'.
	self generatorTableFrom: #(
		(1    0   15 genPushReceiverVariableBytecode isInstVarRef)
		(1  16   31 genPushTemporaryVariableBytecode)
		(1  32   63 genPushLiteralConstantBytecode needsFrameNever: 1)
		(1  64   95 genPushLiteralVariableBytecode needsFrameNever: 1)
		(1  96 103 genStoreAndPopReceiverVariableBytecode isInstVarRef isMappedIfImmutability needsFrameIfImmutability: -1)
		(1 104 111 genStoreAndPopTemporaryVariableBytecode)
		(1 112 112 genPushReceiverBytecode)
		(1 113 113 genPushConstantTrueBytecode needsFrameNever: 1)
		(1 114 114 genPushConstantFalseBytecode needsFrameNever: 1)
		(1 115 115 genPushConstantNilBytecode needsFrameNever: 1)
		(1 116 119 genPushQuickIntegerConstantBytecode needsFrameNever: 1)
		"method returns in blocks need a frame because of nonlocalReturn:through:"
		(1 120 120 genReturnReceiver				return needsFrameIfInBlock: isMappedInBlock 0)
		(1 121 121 genReturnTrue					return needsFrameIfInBlock: isMappedInBlock 0)
		(1 122 122 genReturnFalse					return needsFrameIfInBlock: isMappedInBlock 0)
		(1 123 123 genReturnNil					return needsFrameIfInBlock: isMappedInBlock 0)
		(1 124 124 genReturnTopFromMethod		return needsFrameIfInBlock: isMappedInBlock -1)
		(1 125 125 genReturnTopFromBlock		return needsFrameNever: -1)

		(1 126 127 unknownBytecode)

		(2 128 128 extendedPushBytecode isInstVarRef) "well, maybe inst var ref"
		(2 129 129 extendedStoreBytecode isInstVarRef isMappedIfImmutability) "well, maybe inst var ref"
		(2 130 130 extendedStoreAndPopBytecode isInstVarRef isMappedIfImmutability) "well, maybe inst var ref"
		(2 131 131 genExtendedSendBytecode isMapped)
		(3 132 132 doubleExtendedDoAnythingBytecode isInstVarRef isMapped) "well, maybe inst var ref"
		(2 133 133 genExtendedSuperBytecode isMapped)
		(2 134 134 genSecondExtendedSendBytecode isMapped)
		(1 135 135 genPopStackBytecode needsFrameNever: -1)
		(1 136 136 duplicateTopBytecode needsFrameNever: 1)

		(1 137 137 genPushActiveContextBytecode)
		(2 138 138 genPushNewArrayBytecode)),

		((InitializationOptions at: #SpurObjectMemory ifAbsent: [false])
			ifTrue: [#((3 139 139 genCallPrimitiveBytecode))]
			ifFalse: [#((1 139 139 unknownBytecode))]),

	  #(
		(3 140 140 genPushRemoteTempLongBytecode)
		(3 141 141 genStoreRemoteTempLongBytecode)
		(3 142 142 genStoreAndPopRemoteTempLongBytecode)
		(4 143 143 genPushClosureCopyCopiedValuesBytecode block v3:Block:Code:Size:)

		(1 144 151 genShortUnconditionalJump			branch v3:ShortForward:Branch:Distance:)
		(1 152 159 genShortJumpIfFalse					branch isBranchFalse isMapped "because of mustBeBoolean"
															v3:ShortForward:Branch:Distance:)
		(2 160 163 genLongUnconditionalBackwardJump	branch isMapped "because of interrupt check"
															v3:Long:Branch:Distance:)
		(2 164 167 genLongUnconditionalForwardJump		branch v3:Long:Branch:Distance:)
		(2 168 171 genLongJumpIfTrue					branch isBranchTrue isMapped "because of mustBeBoolean"
															v3:LongForward:Branch:Distance:)
		(2 172 175 genLongJumpIfFalse					branch isBranchFalse isMapped "because of mustBeBoolean"
															v3:LongForward:Branch:Distance:)

		(1 176 197 genSpecialSelectorSend isMapped)
		(1 198 198 genSpecialSelectorEqualsEquals needsFrameNever: notMapped -1) "not mapped because it is directly inlined (for now)"
		(1 199 199 genSpecialSelectorClass needsFrameNever: notMapped 0) "not mapped because it is directly inlined (for now)"
		(1 200 200 genSpecialSelectorNotEqualsEquals needsFrameNever: notMapped -1) "not mapped because it is directly inlined (for now)"
		(1 201 207 genSpecialSelectorSend isMapped)
		(1 208 223 genSendLiteralSelector0ArgsBytecode isMapped)
		(1 224 239 genSendLiteralSelector1ArgBytecode isMapped)
		(1 240 255 genSendLiteralSelector2ArgsBytecode isMapped))! !

!SimpleStackBasedCogit class methodsFor: 'class initialization' stamp: 'eem 7/13/2021 11:42'!
initializePrimitiveTableForNewsqueak
	"Initialize the table of primitive generators.  This does not include normal primitives implemented in the coInterpreter.
	 N.B. primitives that don't have an explicit arg count (the integer following the generator) may be variadic."
	"SimpleStackBasedCogit initializePrimitiveTableForNewsqueak"
	MaxCompiledPrimitiveIndex := self objectRepresentationClass wordSize = 8
										ifTrue: [575]
										ifFalse: [575].
	primitiveTable := CArrayAccessor on: (Array new: MaxCompiledPrimitiveIndex + 1).
	self table: primitiveTable from: 
	#(	"Integer Primitives (0-19)"
		(1 genPrimitiveAdd				1)
		(2 genPrimitiveSubtract			1)
		(3 genPrimitiveLessThan		1)
		(4 genPrimitiveGreaterThan		1)
		(5 genPrimitiveLessOrEqual		1)
		(6 genPrimitiveGreaterOrEqual	1)
		(7 genPrimitiveEqual			1)
		(8 genPrimitiveNotEqual		1)
		(9 genPrimitiveMultiply			1)
		(10 genPrimitiveDivide			1)
		(11 genPrimitiveMod			1)
		(12 genPrimitiveDiv				1)
		(13 genPrimitiveQuo			1)
		(14 genPrimitiveBitAnd			1)
		(15 genPrimitiveBitOr			1)
		(16 genPrimitiveBitXor			1)
		(17 genPrimitiveBitShift			1)
		"(18 primitiveMakePoint)"
		"(19 primitiveFail)"					"Guard primitive for simulation -- *must* fail"

		"LargeInteger Primitives (20-39)"
		"(20 primitiveFail)"
		"(21 primitiveAddLargeIntegers)"
		"(22 primitiveSubtractLargeIntegers)"
		"(23 primitiveLessThanLargeIntegers)"
		"(24 primitiveGreaterThanLargeIntegers)"
		"(25 primitiveLessOrEqualLargeIntegers)"
		"(26 primitiveGreaterOrEqualLargeIntegers)"
		"(27 primitiveEqualLargeIntegers)"
		"(28 primitiveNotEqualLargeIntegers)"
		"(29 primitiveMultiplyLargeIntegers)"
		"(30 primitiveDivideLargeIntegers)"
		"(31 primitiveModLargeIntegers)"
		"(32 primitiveDivLargeIntegers)"
		"(33 primitiveQuoLargeIntegers)"
		"(34 primitiveBitAndLargeIntegers)"
		"(35 primitiveBitOrLargeIntegers)"
		"(36 primitiveBitXorLargeIntegers)"
		"(37 primitiveBitShiftLargeIntegers)"

		"Float Primitives (38-59)"
		"(38 genPrimitiveFloatAt)"
		"(39 genPrimitiveFloatAtPut)"
		(40 genPrimitiveAsFloat					0)
		(41 genPrimitiveFloatAdd				1)
		(42 genPrimitiveFloatSubtract			1)
		(43 genPrimitiveFloatLessThan			1)
		(44 genPrimitiveFloatGreaterThan		1)
		(45 genPrimitiveFloatLessOrEqual		1)
		(46 genPrimitiveFloatGreaterOrEqual	1)
		(47 genPrimitiveFloatEqual				1)
		(48 genPrimitiveFloatNotEqual			1)
		(49 genPrimitiveFloatMultiply			1)
		(50 genPrimitiveFloatDivide				1)
		"(51 genPrimitiveTruncated)"
		"(52 genPrimitiveFractionalPart)"
		"(53 genPrimitiveExponent)"
		"(54 genPrimitiveTimesTwoPower)"
		(55 genPrimitiveFloatSquareRoot		0)
		"(56 genPrimitiveSine)"
		"(57 genPrimitiveArctan)"
		"(58 genPrimitiveLogN)"
		"(59 genPrimitiveExp)"

		"Subscript and Stream Primitives (60-67)"
		(60 genPrimitiveAt				1)
		(61 genPrimitiveAtPut			2)
		(62 genPrimitiveSize			0)
		(63 genPrimitiveStringAt		1)
		(64 genPrimitiveStringAtPut		2)
		"The stream primitives no longer pay their way; normal Smalltalk code is faster."
		(65 genFastPrimFail)"was primitiveNext"
		(66 genFastPrimFail) "was primitiveNextPut"
		(67 genFastPrimFail) "was primitiveAtEnd"

		"StorageManagement Primitives (68-79)"
		(68 genPrimitiveObjectAt			1)	"Good for debugger/InstructionStream performance"
		"(69 primitiveObjectAtPut)"
		(70 genPrimitiveNew)				"For VMMirror support 1 argument instantiateFixedClass: as well as baiscNew"
		(71 genPrimitiveNewWithArg)		"For VMMirror support 2 argument instantiateVariableClass:withSize: as well as baiscNew:"
		"(72 primitiveArrayBecomeOneWay)"		"Blue Book: primitiveBecome"
		"(73 primitiveInstVarAt)"
		"(74 primitiveInstVarAtPut)"
		(75 genPrimitiveIdentityHash	0)
		"(76 primitiveStoreStackp)"					"Blue Book: primitiveAsObject"
		"(77 primitiveSomeInstance)"
		"(78 primitiveNextInstance)"
		(79 genPrimitiveNewMethod	2)

		"Control Primitives (80-89)"
		"(80 primitiveFail)"							"Blue Book: primitiveBlockCopy"
		"(81 primitiveFail)"							"Blue Book: primitiveValue"
		"(82 primitiveFail)"							"Blue Book: primitiveValueWithArgs"
		(83 genPrimitivePerform)
		"(84 primitivePerformWithArgs)"
		"(85 primitiveSignal)"
		"(86 primitiveWait)"
		"(87 primitiveResume)"
		"(88 primitiveSuspend)"
		"(89 primitiveFlushCache)"

		"System Primitives (110-119)"
		(110 genPrimitiveIdentical 1)
		(111 genPrimitiveClass)			"For objectClass: and VMMirror support 1 argument classOf: as well as class"
		"(112 primitiveBytesLeft)"
		"(113 primitiveQuit)"
		"(114 primitiveExitToDebugger)"
		"(115 primitiveChangeClass)"					"Blue Book: primitiveOopsLeft"
		"(116 primitiveFlushCacheByMethod)"
		"(117 primitiveExternalCall)"
		"(118 primitiveDoPrimitiveWithArgs)"
		"(119 primitiveFlushCacheSelective)"

		(148 genPrimitiveShallowCopy 0)			"a.k.a. clone"

		(165 genPrimitiveIntegerAt			1)	"Signed version of genPrimitiveAt"
		(166 genPrimitiveIntegerAtPut		2)	"Signed version of genPrimitiveAtPut"

		(169 genPrimitiveNotIdentical 1)

		(170 genPrimitiveAsCharacter)				"SmallInteger>>asCharacter, Character class>>value:"
		(171 genPrimitiveImmediateAsInteger 0)	"Character>>asInteger/hash/identityHash, SmallFloat64>>identityHash"
			
		"(173 primitiveSlotAt 1)"
		"(174 primitiveSlotAtPut 2)"
		(175 genPrimitiveIdentityHash	0)		"Behavior>>identityHash"

		"Old closure primitives"
		"(186 primitiveFail)" "was primitiveClosureValue"
		"(187 primitiveFail)" "was primitiveClosureValueWithArgs"

		"Perform method directly"
		"(188 primitiveExecuteMethodArgsArray)"
		"(189 primitiveExecuteMethod)"

		"Unwind primitives"
		"(195 primitiveFindNextUnwindContext)"
		"(196 primitiveTerminateTo)"
		"(197 primitiveFindHandlerContext)"
		(198 genFastPrimFail "primitiveMarkUnwindMethod")
		(199 genFastPrimFail "primitiveMarkHandlerMethod")

		"new closure primitives"
		"(200 primitiveClosureCopyWithCopiedValues)"
		(201 genPrimitiveClosureValue	0) "value"
		(202 genPrimitiveClosureValue	1) "value:"
		(203 genPrimitiveClosureValue	2) "value:value:"
		(204 genPrimitiveClosureValue	3) "value:value:value:"
		(205 genPrimitiveClosureValue	4) "value:value:value:value:"
		"(206 genPrimitiveClosureValueWithArgs)" "valueWithArguments:"

		"(210 primitiveContextAt)"
		"(211 primitiveContextAtPut)"
		"(212 primitiveContextSize)"

		"(218 primitiveDoNamedPrimitiveWithArgs)"
		"(219 primitiveFail)"	"reserved for Cog primitives"

		"(220 primitiveFail)"		"reserved for Cog primitives"

		(221 genPrimitiveClosureValue	0) "valueNoContextSwitch"
		(222 genPrimitiveClosureValue	1) "valueNoContextSwitch:"

		"SmallFloat primitives (540-559)"
		(541 genPrimitiveSmallFloatAdd				1)
		(542 genPrimitiveSmallFloatSubtract			1)
		(543 genPrimitiveSmallFloatLessThan			1)
		(544 genPrimitiveSmallFloatGreaterThan		1)
		(545 genPrimitiveSmallFloatLessOrEqual		1)
		(546 genPrimitiveSmallFloatGreaterOrEqual		1)
		(547 genPrimitiveSmallFloatEqual				1)
		(548 genPrimitiveSmallFloatNotEqual			1)
		(549 genPrimitiveSmallFloatMultiply				1)
		(550 genPrimitiveSmallFloatDivide				1)
		"(551 genPrimitiveSmallFloatTruncated			0)"
		"(552 genPrimitiveSmallFloatFractionalPart		0)"
		"(553 genPrimitiveSmallFloatExponent			0)"
		"(554 genPrimitiveSmallFloatTimesTwoPower	1)"
		(555 genPrimitiveSmallFloatSquareRoot			0)
		"(556 genPrimitiveSmallFloatSine				0)"
		"(557 genPrimitiveSmallFloatArctan				0)"
		"(558 genPrimitiveSmallFloatLogN				0)"
		"(559 genPrimitiveSmallFloatExp				0)"
		(575 genPrimitiveHighBit			0)
	)! !

!SimpleStackBasedCogit class methodsFor: 'class initialization' stamp: 'eem 7/13/2021 11:42'!
initializePrimitiveTableForSqueak
	"Initialize the table of primitive generators.  This does not include normal primitives implemented in the coInterpreter.
	 N.B. primitives that don't have an explicit arg count (the integer following the generator) may be variadic."
	"SimpleStackBasedCogit initializePrimitiveTableForSqueak"
	MaxCompiledPrimitiveIndex := self objectRepresentationClass wordSize = 8
										ifTrue: [575]
										ifFalse: [575].
	primitiveTable := CArrayAccessor on: (Array new: MaxCompiledPrimitiveIndex + 1).
	self table: primitiveTable from: 
	#(	"Integer Primitives (0-19)"
		(1 genPrimitiveAdd				1)
		(2 genPrimitiveSubtract			1)
		(3 genPrimitiveLessThan		1)
		(4 genPrimitiveGreaterThan		1)
		(5 genPrimitiveLessOrEqual		1)
		(6 genPrimitiveGreaterOrEqual	1)
		(7 genPrimitiveEqual			1)
		(8 genPrimitiveNotEqual		1)
		(9 genPrimitiveMultiply			1)
		(10 genPrimitiveDivide			1)
		(11 genPrimitiveMod			1)
		(12 genPrimitiveDiv				1)
		(13 genPrimitiveQuo			1)
		(14 genPrimitiveBitAnd			1)
		(15 genPrimitiveBitOr			1)
		(16 genPrimitiveBitXor			1)
		(17 genPrimitiveBitShift			1)
		(18 genPrimitiveMakePoint		1)	"this is here mainly to remove noise from printPrimTraceLog()"
		"(19 primitiveFail)"					"Guard primitive for simulation -- *must* fail"

		"LargeInteger Primitives (20-39)"
		"(20 primitiveFail)"
		"(21 primitiveAddLargeIntegers)"
		"(22 primitiveSubtractLargeIntegers)"
		"(23 primitiveLessThanLargeIntegers)"
		"(24 primitiveGreaterThanLargeIntegers)"
		"(25 primitiveLessOrEqualLargeIntegers)"
		"(26 primitiveGreaterOrEqualLargeIntegers)"
		"(27 primitiveEqualLargeIntegers)"
		"(28 primitiveNotEqualLargeIntegers)"
		"(29 primitiveMultiplyLargeIntegers)"
		"(30 primitiveDivideLargeIntegers)"
		"(31 primitiveModLargeIntegers)"
		"(32 primitiveDivLargeIntegers)"
		"(33 primitiveQuoLargeIntegers)"
		"(34 primitiveBitAndLargeIntegers)"
		"(35 primitiveBitOrLargeIntegers)"
		"(36 primitiveBitXorLargeIntegers)"
		"(37 primitiveBitShiftLargeIntegers)"

		"Float Primitives (38-59)"
		"(38 genPrimitiveFloatAt)"
		"(39 genPrimitiveFloatAtPut)"
		(40 genPrimitiveAsFloat					0)
		(41 genPrimitiveFloatAdd				1)
		(42 genPrimitiveFloatSubtract			1)
		(43 genPrimitiveFloatLessThan			1)
		(44 genPrimitiveFloatGreaterThan		1)
		(45 genPrimitiveFloatLessOrEqual		1)
		(46 genPrimitiveFloatGreaterOrEqual	1)
		(47 genPrimitiveFloatEqual				1)
		(48 genPrimitiveFloatNotEqual			1)
		(49 genPrimitiveFloatMultiply			1)
		(50 genPrimitiveFloatDivide				1)
		"(51 genPrimitiveTruncated)"
		"(52 genPrimitiveFractionalPart)"
		"(53 genPrimitiveExponent)"
		"(54 genPrimitiveTimesTwoPower)"
		(55 genPrimitiveFloatSquareRoot		0)
		"(56 genPrimitiveSine)"
		"(57 genPrimitiveArctan)"
		"(58 genPrimitiveLogN)"
		"(59 genPrimitiveExp)"

		"Subscript and Stream Primitives (60-67)"
		(60 genPrimitiveAt				1)
		(61 genPrimitiveAtPut			2)
		(62 genPrimitiveSize			0)
		(63 genPrimitiveStringAt		1)
		(64 genPrimitiveStringAtPut		2)
		"The stream primitives no longer pay their way; normal Smalltalk code is faster."
		(65 genFastPrimFail)"was primitiveNext"
		(66 genFastPrimFail) "was primitiveNextPut"
		(67 genFastPrimFail) "was primitiveAtEnd"

		"StorageManagement Primitives (68-79)"
		(68 genPrimitiveObjectAt			1)	"Good for debugger/InstructionStream performance"
		"(69 primitiveObjectAtPut)"
		(70 genPrimitiveNew			0)
		(71 genPrimitiveNewWithArg	1)
		"(72 primitiveArrayBecomeOneWay)"		"Blue Book: primitiveBecome"
		"(73 primitiveInstVarAt)"
		"(74 primitiveInstVarAtPut)"
		(75 genPrimitiveIdentityHash	0)
		"(76 primitiveStoreStackp)"					"Blue Book: primitiveAsObject"
		"(77 primitiveSomeInstance)"
		"(78 primitiveNextInstance)"
		(79 genPrimitiveNewMethod	2)

		"Control Primitives (80-89)"
		"(80 primitiveFail)"							"Blue Book: primitiveBlockCopy"
		"(81 primitiveFail)"							"Blue Book: primitiveValue"
		"(82 primitiveFail)"							"Blue Book: primitiveValueWithArgs"
		(83 genPrimitivePerform)
		"(84 primitivePerformWithArgs)"
		"(85 primitiveSignal)"
		"(86 primitiveWait)"
		"(87 primitiveResume)"
		"(88 primitiveSuspend)"
		"(89 primitiveFlushCache)"

		"(90 primitiveMousePoint)"
		"(91 primitiveTestDisplayDepth)"			"Blue Book: primitiveCursorLocPut"
		"(92 primitiveSetDisplayMode)"				"Blue Book: primitiveCursorLink"
		"(93 primitiveInputSemaphore)"
		"(94 primitiveGetNextEvent)"				"Blue Book: primitiveSampleInterval"
		"(95 primitiveInputWord)"
		"(96 primitiveFail)"	"primitiveCopyBits"
		"(97 primitiveSnapshot)"
		"(98 primitiveStoreImageSegment)"
		"(99 primitiveLoadImageSegment)"
		"(100 primitivePerformInSuperclass)"		"Blue Book: primitiveSignalAtTick"
		"(101 primitiveBeCursor)"
		"(102 primitiveBeDisplay)"
		"(103 primitiveScanCharacters)"
		"(104 primitiveFail)"	"primitiveDrawLoop"
		(105 genPrimitiveStringReplace)
		"(106 primitiveScreenSize)"
		"(107 primitiveMouseButtons)"
		"(108 primitiveKbdNext)"
		"(109 primitiveKbdPeek)"


		"System Primitives (110-119)"
		(110 genPrimitiveIdentical 1)
		(111 genPrimitiveClass)				"Support both class and Context>>objectClass:"
		"(112 primitiveBytesLeft)"
		"(113 primitiveQuit)"
		"(114 primitiveExitToDebugger)"
		"(115 primitiveChangeClass)"					"Blue Book: primitiveOopsLeft"
		"(116 primitiveFlushCacheByMethod)"
		"(117 primitiveExternalCall)"
		"(118 primitiveDoPrimitiveWithArgs)"
		"(119 primitiveFlushCacheSelective)"

		(148 genPrimitiveShallowCopy 0)			"a.k.a. clone"

		(158 genPrimitiveStringCompareWith 1)
		(159 genPrimitiveHashMultiply 0)

		(165 genPrimitiveIntegerAt			1)	"Signed version of genPrimitiveAt"
		(166 genPrimitiveIntegerAtPut		2)	"Signed version of genPrimitiveAtPut"

		(169 genPrimitiveNotIdentical 1)

		(170 genPrimitiveAsCharacter)				"SmallInteger>>asCharacter, Character class>>value:"
		(171 genPrimitiveImmediateAsInteger 0)	"Character>>asInteger/hash/identityHash, SmallFloat64>>identityHash"
			
		"(173 primitiveSlotAt 1)"
		"(174 primitiveSlotAtPut 2)"
		(175 genPrimitiveIdentityHash	0)		"Behavior>>identityHash"

		"Old closure primitives"
		"(186 primitiveFail)" "was primitiveClosureValue"
		"(187 primitiveFail)" "was primitiveClosureValueWithArgs"

		"Perform method directly"
		"(188 primitiveExecuteMethodArgsArray)"
		"(189 primitiveExecuteMethod)"

		"Unwind primitives"
		"(195 primitiveFindNextUnwindContext)"
		"(196 primitiveTerminateTo)"
		"(197 primitiveFindHandlerContext)"
		(198 genFastPrimFail "primitiveMarkUnwindMethod")
		(199 genFastPrimFail "primitiveMarkHandlerMethod")

		"new closure primitives"
		"(200 primitiveClosureCopyWithCopiedValues)"
		(201 genPrimitiveClosureValue	0) "value"
		(202 genPrimitiveClosureValue	1) "value:"
		(203 genPrimitiveClosureValue	2) "value:value:"
		(204 genPrimitiveClosureValue	3) "value:value:value:"
		(205 genPrimitiveClosureValue	4) "value:value:value:value:"
		"(206 genPrimitiveClosureValueWithArgs)" "valueWithArguments:"

		(207 genPrimitiveFullClosureValue) "value[:value:value:value:] et al"
		"(208 genPrimitiveFullClosureValueWithArgs)" "valueWithArguments:"
		(209 genPrimitiveFullClosureValue) "valueNoContextSwitch[:value:] et al"

		"(210 primitiveContextAt)"
		"(211 primitiveContextAtPut)"
		"(212 primitiveContextSize)"

		"(218 primitiveDoNamedPrimitiveWithArgs)"
		"(219 primitiveFail)"	"reserved for Cog primitives"

		"(220 primitiveFail)"		"reserved for Cog primitives"

		(221 genPrimitiveClosureValue	0) "valueNoContextSwitch"
		(222 genPrimitiveClosureValue	1) "valueNoContextSwitch:"

		"SmallFloat primitives (540-559)"
		(541 genPrimitiveSmallFloatAdd				1)
		(542 genPrimitiveSmallFloatSubtract			1)
		(543 genPrimitiveSmallFloatLessThan			1)
		(544 genPrimitiveSmallFloatGreaterThan		1)
		(545 genPrimitiveSmallFloatLessOrEqual		1)
		(546 genPrimitiveSmallFloatGreaterOrEqual		1)
		(547 genPrimitiveSmallFloatEqual				1)
		(548 genPrimitiveSmallFloatNotEqual			1)
		(549 genPrimitiveSmallFloatMultiply				1)
		(550 genPrimitiveSmallFloatDivide				1)
		"(551 genPrimitiveSmallFloatTruncated			0)"
		"(552 genPrimitiveSmallFloatFractionalPart		0)"
		"(553 genPrimitiveSmallFloatExponent			0)"
		"(554 genPrimitiveSmallFloatTimesTwoPower	1)"
		(555 genPrimitiveSmallFloatSquareRoot			0)
		"(556 genPrimitiveSmallFloatSine				0)"
		"(557 genPrimitiveSmallFloatArctan				0)"
		"(558 genPrimitiveSmallFloatLogN				0)"
		"(559 genPrimitiveSmallFloatExp				0)"
		(575 genPrimitiveHighBit			0)
	)! !

!SimpleStackBasedCogit class methodsFor: 'translation' stamp: 'eem 4/20/2020 17:24'!
preGenerationHook: aCCodeGenerator
	"Define the primitiveTable initializer once all methods have been added."
	aCCodeGenerator vmClass primitiveTable ifNotNil:
		[:bytecodeGenTable|
		aCCodeGenerator
			var: #primitiveGeneratorTable
				declareC: 'static PrimitiveDescriptor primitiveGeneratorTable[MaxCompiledPrimitiveIndex+1]',
							(self tableInitializerFor: aCCodeGenerator vmClass primitiveTable
								in: aCCodeGenerator)]! !

!SimpleStackBasedCogit class methodsFor: 'class initialization' stamp: 'eem 1/9/2016 10:42'!
table: primArray from: specArray 
	"Fill in the specified entries in the primitive table."
	specArray do:
		[:spec | 
		 (spec first <= primArray size
		  and: [spec second == #genFastPrimFail
			  or: [self objectRepresentationClass shouldIncludeMethodForSelector: spec second]]) ifTrue:
			[(primArray at: spec first put: CogPrimitiveDescriptor new)
				primitiveGenerator: spec second;
				primNumArgs: (spec at: 3 ifAbsent: -1)]].
	primArray object withIndexDo:
		[:generator :i|
		generator ifNil:
			[(primArray object at: i put: CogPrimitiveDescriptor new)
				primNumArgs: -1]]! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 2/23/2016 16:48'!
adjustArgumentsForPerform: numArgs
	"e.g.	Receiver				Receiver	or	Receiver				Receiver	(RISC)
			Selector/Arg0	=>		Arg1			Selector/Arg0	=>		Arg1
			Arg1					Arg2			Arg1					Arg2
			Arg2					Arg3			Arg2			sp->	Arg3
	 		Arg3			sp->	retpc	sp->	Arg3
	 sp->	retpc"

	backEnd hasLinkRegister
		ifTrue:
			[numArgs - 2 to: 0 by: -1 do:
				[:index|
				self MoveMw: index * objectMemory wordSize r: SPReg R: TempReg.
				self MoveR: TempReg Mw: index + 1 * objectMemory wordSize r: SPReg].
			self AddCq: objectMemory wordSize R: SPReg]
		ifFalse:
			[numArgs - 1 to: 1 by: -1 do:
				[:index|
				self MoveMw: index * objectMemory wordSize r: SPReg R: TempReg.
				self MoveR: TempReg Mw: index + 1 * objectMemory wordSize r: SPReg].
			self PopR: TempReg.
			self MoveR: TempReg Mw: 0 r: SPReg]! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/12/2018 10:31'!
annotationForSendTable: sendTable
	"c.f. offsetAndSendTableFor:annotation:into:"
	<inline: true>
	<var: #sendTable type: #'sqInt *'>
	sendTable == ordinarySendTrampolines ifTrue:
		[^IsSendCall].
	BytecodeSetHasDirectedSuperSend ifTrue:
		[sendTable == directedSuperSendTrampolines ifTrue:
			[^IsDirectedSuperSend].
		 sendTable == directedSuperBindingSendTrampolines ifTrue:
			[^IsDirectedSuperBindingSend]].
	NewspeakVM ifTrue:
		[sendTable == implicitReceiverSendTrampolines ifTrue:
			[self error: 'Unexpected implicit receiver send using dirty send machinery'].
		 sendTable == outerSendTrampolines ifTrue:
			[self error: 'Unexpected outer send using dirty send machinery'].
		sendTable == selfSendTrampolines ifTrue:
			[^IsNSSelfSend].
		sendTable == dynamicSuperSendTrampolines ifTrue:
			[^IsNSDynamicSuperSend]].
	self assert: sendTable == superSendTrampolines.
	^IsSuperSend! !

!SimpleStackBasedCogit methodsFor: 'simulation only' stamp: 'eem 11/22/2010 11:09'!
bytecodeFixupClass
	<doNotGenerate>
	^CogBytecodeFixup! !

!SimpleStackBasedCogit methodsFor: 'trampolines' stamp: 'eem 12/3/2010 17:05'!
cPICMissTrampolineFor: numArgs
	^ceCPICMissTrampoline! !

!SimpleStackBasedCogit methodsFor: 'trampolines' stamp: 'eem 11/1/2012 17:56'!
ceClosureCopyDescriptor: descriptor "<Integer>" 
	^coInterpreter
		createClosureNumArgs: (descriptor bitAnd: 16r3F)
		numCopied: ((descriptor >> 6) bitAnd: 16r3F)
		startpc: descriptor >> 12! !

!SimpleStackBasedCogit methodsFor: 'in-line cacheing' stamp: 'eem 7/25/2021 11:58'!
cogMethodHasExternalPrim: aCogMethod
	<api>
	<var: 'aCogMethod' type: #'CogMethod *'>
	| primIndex |
	primIndex := coInterpreter primitiveIndexOfMethod: aCogMethod methodObject header: aCogMethod methodHeader.
	^primIndex = PrimNumberExternalCall
	 or: [primIndex = PrimNumberFFICall]! !

!SimpleStackBasedCogit methodsFor: 'in-line cacheing' stamp: 'eem 8/15/2021 17:31'!
cogMethodHasMachineCodePrim: aCogMethod
	<api>
	<var: 'aCogMethod' type: #'CogMethod *'>
	| primIndex |
	primIndex := coInterpreter primitiveIndexOfMethod: aCogMethod methodObject header: aCogMethod methodHeader.
	^(primIndex between: 1 and: MaxCompiledPrimitiveIndex)
	  and: [(primitiveGeneratorTable at: primIndex) primitiveGenerator notNil]! !

!SimpleStackBasedCogit methodsFor: 'compile abstract instructions' stamp: 'eem 9/5/2016 14:10'!
compileBlockBodies
	<inline: false>
	| result compiledBlocksCount blockStart savedNeedsFrame savedNumArgs |
	<var: #blockStart type: #'BlockStart *'>
	self assert: blockCount > 0.
	"scanBlock: in compileBlockEntry: sets both of these appropriately for each block."
	savedNeedsFrame := needsFrame.
	savedNumArgs := methodOrBlockNumArgs.
	inBlock := InVanillaBlock.
	compiledBlocksCount := 0.
	[compiledBlocksCount < blockCount] whileTrue:
		[blockStart := self blockStartAt: compiledBlocksCount.
		 self scanBlock: blockStart.
		 self compileBlockEntry: blockStart.
		 (result := self compileAbstractInstructionsFrom: blockStart startpc
						through: blockStart startpc + blockStart span - 1) < 0 ifTrue:
			[^result].
		 compiledBlocksCount := compiledBlocksCount + 1].
	needsFrame := savedNeedsFrame.
	methodOrBlockNumArgs := savedNumArgs.
	^0! !

!SimpleStackBasedCogit methodsFor: 'compile abstract instructions' stamp: 'eem 1/18/2011 09:59'!
compileBlockDispatch
	"Compile the jump instruction(s) at the end of the method that dispatch to each block body."
	<inline: false>
	| jumpSkip |
	<var: #jumpSkip type: #'AbstractInstruction *'>
	self assert: blockCount > 0.
	"Generate the special block entry for primitiveClosureValueNoContextSwitch
	 that signals a context-switch is not possible by zeroing SendNumArgsReg.
	 Even though conceptually this isn't needed if no blocks build a frame it has
	 to be present so that primitiveClosureValueNoContextSwitch can call it."
	blockEntryNoContextSwitch := self MoveCq: 0 R: SendNumArgsReg.
	"Now generate a binary search through start pcs to jump to blocks. 1 block is a special case."
	jumpSkip := self Jump: 0.
	"Set OK to context switch flag to non-zero."
	blockEntryLabel := self MoveR: ReceiverResultReg R: SendNumArgsReg.
	jumpSkip jmpTarget: self Label.
	blockCount > 1 ifTrue:
		[objectRepresentation
			genLoadSlot: ClosureStartPCIndex
			sourceReg: ReceiverResultReg
			destReg: TempReg].
	self compileBlockDispatchFrom: 0 to: blockCount - 1.
	^0! !

!SimpleStackBasedCogit methodsFor: 'compile abstract instructions' stamp: 'eem 12/15/2019 20:55'!
compileBlockFrameBuild: blockStart
	"Build a frame for a block activation.  See CoInterpreter class>>initializeFrameIndices.
	 		closure (in ReceiverResultReg)
			arg0
			...
			argN
			caller's saved ip/this stackPage (for a base frame)
	fp->	saved fp
			method
			context (uninitialized?)
			receiver
			first temp
			...
	sp->	Nth temp
	Avoid use of SendNumArgsReg which is the flag determining whether
	context switch is allowed on stack-overflow."
	<var: #blockStart type: #'BlockStart *'>
	<inline: false>
	self annotateBytecode: self Label.
	backEnd hasLinkRegister ifTrue:
		[self PushR: LinkReg].
	self PushR: FPReg.
	self MoveR: SPReg R: FPReg.
	"Think of ClassReg as ClosureReg"
	self MoveR: ReceiverResultReg R: ClassReg.
	"The block method field must have its MFMethodFlagIsBlockFlag bit set.
	 We arrange this using a labelOffset.  A hack, but it works."
	blockStart fakeHeader
		addDependent: (self annotateAbsolutePCRef:
			(self PushCw: blockStart fakeHeader asInteger)); "method"
		setLabelOffset: MFMethodFlagIsBlockFlag.
	self annotate: (self PushCw: objectMemory nilObject) "context"
		objRef: objectMemory nilObject.
	"Fetch home receiver from outer context. closure is on stack and initially in ReceiverResultReg.
	 It is safe to use Arg0Reg because reg args are pushed by the value primitives if there are any.".
	blockStart hasInstVarRef
		ifTrue: "Use ReceiverResultReg for Context to agree with store check trampoline"
			[objectRepresentation
				genLoadSlot: ClosureOuterContextIndex
					sourceReg: ClassReg
						destReg: ReceiverResultReg;
				genLoadSlot: ReceiverIndex
					sourceReg: ReceiverResultReg
						destReg: Arg0Reg.
			objectRepresentation
				genEnsureOopInRegNotForwarded: Arg0Reg scratchReg: TempReg updatingSlot: ReceiverIndex in: ReceiverResultReg.
			self MoveR: Arg0Reg R: ReceiverResultReg]
		ifFalse:
			[objectRepresentation
				genLoadSlot: ClosureOuterContextIndex
					sourceReg: ClassReg
						destReg: Arg0Reg;
				genLoadSlot: ReceiverIndex
					sourceReg: Arg0Reg
						destReg: ReceiverResultReg].
	self PushR: ReceiverResultReg. "home receiver"
	"Push copied values; bytecode initializes temporaries"
	0 to: blockStart numCopied - 1 do:
		[:i|
		objectRepresentation
			genLoadSlot: i + ClosureFirstCopiedValueIndex
			sourceReg: ClassReg
			destReg: TempReg.
		self PushR: TempReg].
	blockStart stackCheckLabel:
		(self compileStackOverflowCheck: true)! !

!SimpleStackBasedCogit methodsFor: 'compile abstract instructions' stamp: 'cb 9/5/2016 15:36'!
compileBlockFramelessEntry: blockStart
	"Make sure ReceiverResultReg holds the receiver, loaded from the closure,
	 which is what is initially in ReceiverResultReg.  We must annotate the first
	 instruction in vanilla blocks so that findMethodForStartBcpc:inHomeMethod:
	 can function. We need two annotations because the first is a fiducial."
	<var: #blockStart type: #'BlockStart *'>
	blockStart entryLabel ifNotNil:
		[self annotateBytecode: blockStart entryLabel.
		 self annotateBytecode: blockStart entryLabel].
	blockStart hasInstVarRef
		ifTrue: "Use ReceiverResultReg for Context to agree with store check trampoline"
			[objectRepresentation
				genLoadSlot: ClosureOuterContextIndex
					sourceReg: ReceiverResultReg
						destReg: ReceiverResultReg;
				genLoadSlot: ReceiverIndex
					sourceReg: ReceiverResultReg
						destReg: Arg0Reg.
			objectRepresentation
				genEnsureOopInRegNotForwarded: Arg0Reg scratchReg: TempReg updatingSlot: ReceiverIndex in: ReceiverResultReg.
			self MoveR: Arg0Reg R: ReceiverResultReg]
		ifFalse:
			[objectRepresentation
				genLoadSlot: ClosureOuterContextIndex
					sourceReg: ReceiverResultReg
						destReg: TempReg;
				genLoadSlot: ReceiverIndex
					sourceReg: TempReg
						destReg: ReceiverResultReg].! !

!SimpleStackBasedCogit methodsFor: 'compile abstract instructions' stamp: 'eem 12/15/2019 20:56'!
compileFrameBuild
	"Build a frame for a CogMethod activation.  See CoInterpreter class>>initializeFrameIndices.
	 		receiver (in ReceiverResultReg)
			arg0
			...
			argN
			caller's saved ip/this stackPage (for a base frame)
	fp->	saved fp
			method
			context (uninitialized?)
			receiver
			first temp
			...
	sp->	Nth temp
	If there is a primitive and an error code the Nth temp is the error code.
	Ensure SendNumArgsReg is set early on (incidentally to nilObj) because
	it is the flag determining whether context switch is allowed on stack-overflow."
	<inline: false>
	needsFrame ifFalse: [^self].
	backEnd hasLinkRegister ifTrue: [self PushR: LinkReg].
	self PushR: FPReg.
	self MoveR: SPReg R: FPReg.
	methodLabel addDependent: (self annotateAbsolutePCRef:
		(self PushCw: methodLabel asInteger)). "method"
	self genMoveNilR: SendNumArgsReg.
	self PushR: SendNumArgsReg. "context"
	self PushR: ReceiverResultReg.
	methodOrBlockNumArgs + 1 to: (coInterpreter temporaryCountOfMethodHeader: methodHeader) do:
		[:i|
		self PushR: SendNumArgsReg].
	(self methodUsesPrimitiveErrorCode: methodObj header: methodHeader) ifTrue:
		[self compileGetErrorCode].
	stackCheckLabel := self compileStackOverflowCheck: (coInterpreter
															canContextSwitchIfActivating: methodObj
															header: methodHeader).
	NewspeakVM ifTrue:
		[numIRCs > 0 ifTrue:
		 	[self PrefetchAw: theIRCs]]! !

!SimpleStackBasedCogit methodsFor: 'compile abstract instructions' stamp: 'cb 5/28/2018 18:09'!
compileFullBlockFramelessEntry: numCopied 
	"Make sure ReceiverResultReg holds the receiver, loaded from the closure,
	 which is what is initially in ReceiverResultReg.  "
	self flag: #TODO. "we could follow the receiver only if the block has inst var ref. Currently we use scanMethod for fullBlock 
	and that scanner does not provide this information. We could extend it based on the scanBlock: method"
	"Use ReceiverResultReg for the closure to agree with store check trampoline"
	objectRepresentation
		genLoadSlot: FullClosureReceiverIndex
			sourceReg: ReceiverResultReg
				destReg: Arg0Reg.
	objectRepresentation
		genEnsureOopInRegNotForwarded: Arg0Reg scratchReg: TempReg updatingSlot: FullClosureReceiverIndex in: ReceiverResultReg.
	self MoveR: Arg0Reg R: ReceiverResultReg.! !

!SimpleStackBasedCogit methodsFor: 'compile abstract instructions' stamp: 'eem 12/15/2019 20:51'!
compileFullBlockMethodFrameBuild: numCopied
	"Build a frame for a block activation.  See CoInterpreter class>>initializeFrameIndices.
	 		closure (in ReceiverResultReg)
			arg0
			...
			argN
			caller's saved ip/this stackPage (for a base frame)
	fp->	saved fp
			method
			context (uninitialized?)
			receiver
			first temp
			...
	sp->	Nth temp
	Avoid use of SendNumArgsReg which is the flag determining whether
	context switch is allowed on stack-overflow."
	<inline: false>
	needsFrame ifFalse: [^self].
	backEnd hasLinkRegister ifTrue: [self PushR: LinkReg].
	self PushR: FPReg.
	self MoveR: SPReg R: FPReg.
	"Think of ClassReg as ClosureReg"
	self MoveR: ReceiverResultReg R: ClassReg.
	"The block method field must have its MFMethodFlagIsBlockFlag bit set.
	 We arrange this using a labelOffset.  A hack, but it works."
	methodLabel addDependent: (self annotateAbsolutePCRef:
			(self PushCw: methodLabel asInteger));
			setLabelOffset: MFMethodFlagIsBlockFlag.. "method"
	self genMoveNilR: SendNumArgsReg.
	self PushR: SendNumArgsReg. "context"
	"Closure is on stack and initially in ReceiverResultReg.
	 It is safe to use Arg0Reg because reg args are pushed by the value primitives if there are any.".

	self flag: #TODO. "we could follow the receiver only if the block has inst var ref. Currently we use scanMethod for fullBlock 
	and that scanner does not provide this information. We could extend it based on the scanBlock: method"
	"Use ReceiverResultReg for the closure to agree with store check trampoline"
	objectRepresentation
		genLoadSlot: FullClosureReceiverIndex
			sourceReg: ClassReg
				destReg: Arg0Reg.
	objectRepresentation
		genEnsureOopInRegNotForwarded: Arg0Reg scratchReg: TempReg updatingSlot: FullClosureReceiverIndex in: ReceiverResultReg.
	self MoveR: Arg0Reg R: ReceiverResultReg.

	self PushR: ReceiverResultReg. "closure receiver"
	"Push copied values"
	0 to: numCopied - 1 do:
		[:i|
		objectRepresentation
			genLoadSlot: i + FullClosureFirstCopiedValueIndex
			sourceReg: ClassReg
			destReg: TempReg.
		self PushR: TempReg].
	"Push slots for temps"
	methodOrBlockNumArgs + numCopied + 1 to: (coInterpreter temporaryCountOfMethodHeader: methodHeader) do:
		[:i|
		self PushR: SendNumArgsReg].

	stackCheckLabel := self compileStackOverflowCheck: true! !

!SimpleStackBasedCogit methodsFor: 'compile abstract instructions' stamp: 'eem 12/7/2017 17:06'!
compileGetErrorCode
	"After pushing the temporaries but before the stack limit check a primitive method
	 needs to fetch the error code, if any.  If the primitive has failed, call the trampoline
	 that will assign it to the last temp."
	<inline: false>
	| jmpNoError |
	<var: #jmpNoError type: #'AbstractInstruction *'>
	self MoveAw: coInterpreter primFailCodeAddress R: TempReg.
	self flag: 'ask concrete code gen if move sets condition codes?'.
	self CmpCq: 0 R: TempReg.
	jmpNoError := self JumpZero: 0.
	methodLabel addDependent:
		(self annotateAbsolutePCRef:
			(self MoveCw: methodLabel asInteger R: ClassReg)).
	self CallRT: ceReapAndResetErrorCodeTrampoline.
	jmpNoError jmpTarget: self Label! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 7/24/2021 17:20'!
compileInterpreterPrimitive
	<inline: true>
	| primitiveRoutine flags |
	<var: #primitiveRoutine declareC: 'void (*primitiveRoutine)(void)'>
	primitiveRoutine := coInterpreter
							functionPointerForCompiledMethod: methodObj
							primitiveIndex: primitiveIndex
							primitivePropertyFlagsInto: (self addressOf: flags put: [:val| flags := val]).
	self deny: (flags anyMask: PrimCallOnSmalltalkStack).
	^self compileInterpreterPrimitive: primitiveRoutine flags: flags! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 8/6/2021 13:01'!
compileInterpreterPrimitive: primitiveRoutine flags: flags
	"Compile a call to an interpreter primitive.  Call the C routine with the
	 usual stack-switching dance, test the primFailCode and then either
	 return on success or continue to the method body."
	<var: #primitiveRoutine declareC: 'void (*primitiveRoutine)(void)'>
	| jmp continueAfterProfileSample jumpToTakeSample |
	self deny: (backEnd hasVarBaseRegister
				and: [self register: VarBaseReg isInMask: ABICallerSavedRegisterMask]).

	"Save processor fp, sp and return pc in the interpreter's frame stack and instruction pointers"
	self genExternalizePointersForPrimitiveCall.
	"Switch to the C stack."
	self genLoadCStackPointersForPrimCall.

	"Old old full prim trace is in VMMaker-eem.550 and prior.
	 Old simpler full prim trace is in VMMaker-eem.2969 and prior."
	(coInterpreter recordPrimTraceForMethod: methodObj) ifTrue:
		[self genFastPrimTraceUsing: ClassReg and: SendNumArgsReg].

	"Clear the primFailCode and set argumentCount"
	self MoveCq: 0 R: TempReg.
	self MoveR: TempReg Aw: coInterpreter primFailCodeAddress.
	methodOrBlockNumArgs ~= 0 ifTrue:
		[self AddCq: methodOrBlockNumArgs R: TempReg]. "As small or smaller than move on most archs"
	self MoveR: TempReg Aw: coInterpreter argumentCountAddress.

	"If required, set primitiveFunctionPointer and newMethod"
	(flags anyMask: PrimCallNeedsPrimitiveFunction) ifTrue:
		[self MoveCw: primitiveRoutine asInteger R: TempReg.
		 self MoveR: TempReg Aw: coInterpreter primitiveFunctionPointerAddress].
	(flags anyMask: PrimCallNeedsNewMethod+PrimCallMayEndureCodeCompaction) ifTrue:
		["The ceActivateFailingPrimitiveMethod: machinery can't handle framelessness."
		 (flags anyMask: PrimCallMayEndureCodeCompaction) ifTrue:
			[needsFrame := true].
		 methodLabel addDependent:
			(self annotateAbsolutePCRef:
				(self MoveCw: methodLabel asInteger R: ClassReg)).
		 self MoveMw: (self offset: CogMethod of: #methodObject) r: ClassReg R: TempReg.
		 self MoveR: TempReg Aw: coInterpreter newMethodAddress].

	"Invoke the primitive.  If the primitive (potentially) contains a call-back then its code
	 may disappear and consequently we cannot return here, sicne here may evaporate.
	 Instead sideways-call the routine, substituting cePrimReturnEnterCogCode[Profiling]
	 as the return address, so the call always returns there."
	self PrefetchAw: coInterpreter primFailCodeAddress.
	(flags anyMask: PrimCallMayEndureCodeCompaction) ifTrue:
		["On Spur ceActivateFailingPrimitiveMethod: would like to retry if forwarders
		  are found. So insist on PrimCallNeedsPrimitiveFunction being set too."
		 objectMemory hasSpurMemoryManagerAPI ifTrue:
			[self assert: (flags anyMask: PrimCallNeedsPrimitiveFunction)].
		 backEnd
			genMarshallNArgs: 0 arg: nil arg: nil arg: nil arg: nil;
			genSubstituteReturnAddress:
				((flags anyMask: PrimCallCollectsProfileSamples)
					ifTrue: [cePrimReturnEnterCogCodeProfiling]
					ifFalse: [cePrimReturnEnterCogCode]).
		 self JumpFullRT: primitiveRoutine asInteger.
		 ^0].

	"Call the C primitive routine."
	backEnd genMarshallNArgs: 0 arg: 0 arg: 0 arg: 0 arg: 0.
	self CallFullRT: primitiveRoutine asInteger.
	backEnd genRemoveNArgsFromStack: 0.
	objectRepresentation maybeCompileRetryOnPrimitiveFail: primitiveIndex.
	"Switch back to the Smalltalk stack.  Stack better be in either of these two states:
		success:	stackPointer ->	result (was receiver)
									arg1
									...
									argN
									return pc
		failure:						receiver
									arg1
									...
					stackPointer ->	argN
									return pc"
	backEnd genLoadStackPointersForPrimCall: ClassReg.
	"genLoadStackPointersForPrimCall: leaves the stack in these states:
			NoLinkRegister 												LinkRegister
		success:					result (was receiver)		stackPointer ->	result (was receiver)
					stackPointer ->	arg1										arg1
									...											...
									argN										argN
									return pc

		failure:						receiver									receiver
									arg1										arg1
									...											...
									argN						stackPointer ->	argN
					stackPointer ->	return pc
	which corresponds to the stack on entry after pushRegisterArgs.
	 In either case we can write the instructionPointer to top of stack or load it into the LinkRegister to reestablish the return pc."
	backEnd hasLinkRegister
		ifTrue:
			[self MoveAw: coInterpreter instructionPointerAddress R: LinkReg]
		ifFalse:
			[self MoveAw: coInterpreter instructionPointerAddress R: ClassReg.
			 self MoveR: ClassReg Mw: 0 r: SPReg].
	"Test primitive failure"
	self MoveAw: coInterpreter primFailCodeAddress R: TempReg.
	self flag: 'ask concrete code gen if move sets condition codes?'.
	self CmpCq: 0 R: TempReg.
	jmp := self JumpNonZero: 0.
	"placing the test here attributes the tick to the primitive plus any checkForAndFollowForwardedPrimitiveState
	 scanning, but attributes all of a failing primitive to the current method (in ceStackOverflow: on frame build)."
	(backEnd has64BitPerformanceCounter
	 and: [flags anyMask: PrimCallCollectsProfileSamples]) ifTrue:
		[jumpToTakeSample := self genCheckForProfileTimerTick: (self registerMaskFor: NoReg)].
	"Fetch result from stack"
	continueAfterProfileSample :=
	self MoveMw: (backEnd hasLinkRegister ifTrue: [0] ifFalse: [objectMemory wordSize])
		r: SPReg
		R: ReceiverResultReg.
	self RetN: objectMemory wordSize.	"return to caller, popping receiver"
	(backEnd has64BitPerformanceCounter
	 and: [flags anyMask: PrimCallCollectsProfileSamples]) ifTrue:
		[jumpToTakeSample jmpTarget: self Label.
		 self genTakeProfileSample.
		 backEnd genLoadStackPointerForPrimCall: ClassReg.
		 backEnd hasLinkRegister
			ifTrue:
				[self MoveAw: coInterpreter instructionPointerAddress R: LinkReg]
			ifFalse:
				[self MoveAw: coInterpreter instructionPointerAddress R: ClassReg.
				 self MoveR: ClassReg Mw: 0 r: SPReg].
		 self Jump: continueAfterProfileSample].

	"Jump to restore of receiver reg and proceed to frame build for failure."
	 jmp jmpTarget: self Label.
	 "Restore receiver reg from stack.  If on RISCs ret pc is in LinkReg, if on CISCs ret pc is on stack."
	 self MoveMw: objectMemory wordSize * (methodOrBlockNumArgs + (backEnd hasLinkRegister ifTrue: [0] ifFalse: [1]))
		r: SPReg
		R: ReceiverResultReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 8/19/2021 16:23'!
compileOnStackExternalPrimitive: primitiveRoutine flags: flags
	"Compile a fast call of a C primitive using the current stack page, avoiding the stack switch except on failure.
	 This convention still uses stackPointer and argumentCount to access operands.  Push all operands to the stack,
	 assign stackPointer, argumentCount, and zero primFailCode.  Make the call (saving a LinkReg if required).
	 Test for failure and return.  On failure on Spur, if there is an accessor depth, assign framePointer and newMethod,
	 do the stack switch, call checkForAndFollowForwardedPrimitiveState, and loop back if forwarders are found.
	 Fall through to frame build."
	<option: #SpurObjectMemory>
	<var: #primitiveRoutine declareC: 'void (*primitiveRoutine)(void)'>
	| calleeSavedRegisterMask linkRegSaveRegister spRegSaveRegister jmp retry continueAfterProfileSample jumpToTakeSample |
	self assert: (objectRepresentation hasSpurMemoryManagerAPI and: [flags anyMask: PrimCallOnSmalltalkStack]).
	self deny: (backEnd hasVarBaseRegister
				and: [self register: VarBaseReg isInMask: ABICallerSavedRegisterMask]).

	(coInterpreter recordFastCCallPrimTraceForMethod: methodObj) ifTrue:
		[self genFastPrimTraceUsing: ClassReg and: SendNumArgsReg].

	"Clear the primFailCode and set argumentCount"
	self MoveCq: 0 R: TempReg.
	self MoveR: TempReg Aw: coInterpreter primFailCodeAddress.
	methodOrBlockNumArgs ~= 0 ifTrue:
		[self AddCq: methodOrBlockNumArgs R: TempReg]. "As small or smaller than move on most archs"
	self MoveR: TempReg Aw: coInterpreter argumentCountAddress.
	self genExternalizeStackPointerForFastPrimitiveCall.
	"We may need to save LinkReg and/or SPReg, and given the stack machinations
	  it is much easier to save them in callee saved registers than on the stack itself."
	calleeSavedRegisterMask := ABICalleeSavedRegisterMask bitClear: (self registerMaskFor: ClassReg).
	backEnd hasLinkRegister ifTrue:
		[linkRegSaveRegister := self availableRegisterOrNoneIn: calleeSavedRegisterMask.
		 self deny: linkRegSaveRegister = NoReg.
		 self MoveR: LinkReg R: linkRegSaveRegister.
		 calleeSavedRegisterMask := calleeSavedRegisterMask bitClear: (self registerMaskFor: linkRegSaveRegister)].
	spRegSaveRegister := NoReg.
	(SPReg ~= NativeSPReg
	 and: [(self isCalleeSavedReg: SPReg) not]) ifTrue:
		[spRegSaveRegister := self availableRegisterOrNoneIn: calleeSavedRegisterMask.
		 self deny: spRegSaveRegister = NoReg.
		 self MoveR: SPReg R: spRegSaveRegister].
	retry := self Label.
	(flags anyMask: PrimCallOnSmalltalkStackAlign2x)
		ifTrue: [self AndCq: (objectMemory wordSize * 2 - 1) bitInvert R: SPReg R: NativeSPReg]
		ifFalse:
			[SPReg ~= NativeSPReg ifTrue:
				[backEnd genLoadNativeSPRegWithAlignedSPReg]].
	backEnd genMarshallNArgs: 0 arg: 0 arg: 0 arg: 0 arg: 0.
	self CallFullRT: primitiveRoutine asInteger.
	backEnd genRemoveNArgsFromStack: 0.
	self MoveAw: coInterpreter primFailCodeAddress R: TempReg.
	spRegSaveRegister ~= NoReg ifTrue:
		[self MoveR: spRegSaveRegister R: SPReg].
	self CmpCq: 0 R: TempReg.
	jmp := self JumpNonZero: 0.
	"Remember to restore the native stack pointer to point to the C stack,
	 otherwise the Smalltalk frames will get overwritten on an interrupt."
	SPReg ~= NativeSPReg ifTrue:
		[backEnd genLoadCStackPointer].
	"placing the test here attributes the tick to the primitive plus any checkForAndFollowForwardedPrimitiveState
	 scanning, but attributes all of a failing primitive to the current method (in ceStackOverflow: on frame build)."
	(backEnd has64BitPerformanceCounter
	 and: [flags anyMask: PrimCallCollectsProfileSamples]) ifTrue:
		[jumpToTakeSample := self genCheckForProfileTimerTick: (self registerMaskFor: NoReg)].
	"At this point the primitive has cut back stackPointer to point to the result."
	continueAfterProfileSample :=
	self MoveAw: coInterpreter stackPointerAddress R: TempReg.
	"get result and restore retpc"
	backEnd hasLinkRegister
		ifTrue:
			[self MoveMw: 0 r: TempReg R: ReceiverResultReg;
				AddCq: objectMemory wordSize R: TempReg R: SPReg;
				MoveR: linkRegSaveRegister R: LinkReg]
		ifFalse:
			[| retpcOffset |
			"The original retpc is (argumentCount + 1) words below stackPointer."
			 retpcOffset := (methodOrBlockNumArgs + 1 * objectMemory wordSize) negated.
			 self MoveMw: retpcOffset r: TempReg R: ClassReg; "get retpc"
				MoveR: TempReg R: SPReg;
			 	MoveMw: 0 r: TempReg R: ReceiverResultReg;
				MoveR: ClassReg Mw: 0 r: TempReg "put it back on stack for the return..."].
	self RetN: 0.

	(backEnd has64BitPerformanceCounter
	 and: [flags anyMask: PrimCallCollectsProfileSamples]) ifTrue:
		[jumpToTakeSample jmpTarget: self Label.
		 self genTakeProfileSample.
		 self Jump: continueAfterProfileSample].

	jmp jmpTarget: self Label.
	(coInterpreter accessorDepthForPrimitiveMethod: methodObj) >= 0
		ifTrue:
			[| skip |
			 "Given that following primitive state to the accessor depth is recursive, we're asking for
			  trouble if we run the fixup on the Smalltalk stack page.  Run it on the full C stack instead.
			 This won't be a performance issue since primitive failure should be very rare."
			self MoveR: FPReg Aw: coInterpreter framePointerAddress.
			self MoveCw: primitiveRoutine asInteger R: TempReg.
			self MoveR: TempReg Aw: coInterpreter primitiveFunctionPointerAddress.
			methodLabel addDependent:
				(self annotateAbsolutePCRef:
					(self MoveCw: methodLabel asInteger R: ClassReg)).
			self MoveMw: (self offset: CogMethod of: #methodObject) r: ClassReg R: TempReg.
			self MoveR: TempReg Aw: coInterpreter newMethodAddress.
			self genLoadCStackPointersForPrimCall.
			backEnd genMarshallNArgs: 0 arg: 0 arg: 0 arg: 0 arg: 0.
			self CallFullRT: (self cCode: [#checkForAndFollowForwardedPrimitiveState asUnsignedIntegerPtr]
								   inSmalltalk: [self simulatedTrampolineFor: #checkForAndFollowForwardedPrimitiveState]).
			backEnd genLoadStackPointersForPrimCall: ClassReg.
			self CmpCq: 0 R: ABIResultReg.
			skip := self JumpZero: 0.
			self MoveCq: 0 R: TempReg.
			self MoveR: TempReg Aw: coInterpreter primFailCodeAddress.
			self Jump: retry.
			skip jmpTarget: self Label]
		ifFalse: "must reload SPReg to undo any alignment change,"
			[(flags anyMask: PrimCallOnSmalltalkStackAlign2x) ifTrue:
				[backEnd genLoadStackPointersForPrimCall: ClassReg]].
	"Remember to restore the native stack pointer to point to the C stack,
	 otherwise the Smalltalk frames will get overwritten on an interrupt."
	SPReg ~= NativeSPReg ifTrue:
		[backEnd genLoadCStackPointer].
	"The LinkRegister now contains the return address either of the primitive call or of checkForAndFollowForwardedPrimitiveState.
	 It must be restored to the return address of the send invoking this primtiive method."
	backEnd hasLinkRegister ifTrue:
		[self MoveR: linkRegSaveRegister R: LinkReg].
	"Finally remember to reload ReceiverResultReg if required.  Even if
	 arguments have been pushed, the prolog sequence assumes it is live."
	(self register: ReceiverResultReg isInMask: ABICallerSavedRegisterMask) ifTrue:
		[self MoveMw: (methodOrBlockNumArgs + (backEnd hasLinkRegister ifTrue: [0] ifFalse: [1])) * objectMemory wordSize
			r: SPReg
			R: ReceiverResultReg].
	"continue to frame build..."
	^0! !

!SimpleStackBasedCogit methodsFor: 'in-line cacheing' stamp: 'eem 8/3/2016 20:15'!
compileOpenPIC: selector numArgs: numArgs
	"Compile the code for an open PIC.  Perform a probe of the first-level method
	 lookup cache followed by a call of ceSendFromInLineCacheMiss: if the probe fails."
	| cacheBaseReg jumpSelectorMiss jumpClassMiss itsAHit jumpBCMethod |
	<var: #jumpSelectorMiss type: #'AbstractInstruction *'>
	<var: #jumpClassMiss type: #'AbstractInstruction *'>
	<var: #itsAHit type: #'AbstractInstruction *'>
	<var: #jumpBCMethod type: #'AbstractInstruction *'>
	self preenMethodLabel.
	self compilePICAbort: numArgs.
	entry := objectRepresentation genGetClassTagOf: ReceiverResultReg into: SendNumArgsReg scratchReg: TempReg.

	self flag: #lookupInMethodCacheSel:classTag:. "so this method shows up as a sender of lookupInMethodCacheSel:class:"

	cacheBaseReg := NoReg.
	(backEnd isWithinMwOffsetRange: coInterpreter methodCacheAddress) ifFalse:
		[self MoveCq: coInterpreter methodCacheAddress R: (cacheBaseReg := Extra0Reg)].

	"Do first of three probes.  See CoInterpreter>>lookupInMethodCacheSel:classTag:"
	jumpSelectorMiss := self compileOpenPICMethodCacheProbeFor: selector withShift: 0 baseRegOrNone: cacheBaseReg.
	jumpClassMiss := self JumpNonZero: 0.

	"Fetch the method.  The interpret trampoline requires the bytecoded method in SendNumArgsReg"
	itsAHit := self MoveMw: (cacheBaseReg = NoReg
								ifTrue: [coInterpreter methodCacheAddress asUnsignedInteger + (MethodCacheMethod << objectMemory shiftForWord)]
								ifFalse: [MethodCacheMethod << objectMemory shiftForWord])
					r: ClassReg
					R: SendNumArgsReg.
			.
	"If the method is compiled jump to its unchecked entry-point, otherwise interpret it."
	objectRepresentation genLoadSlot: HeaderIndex sourceReg: SendNumArgsReg destReg: ClassReg.
	jumpBCMethod := objectRepresentation genJumpImmediate: ClassReg.
	jumpBCMethod jmpTarget: picInterpretAbort.
	self AddCq: cmNoCheckEntryOffset R: ClassReg.
	self JumpR: ClassReg.

	"First probe missed.  Do second of three probes.  Shift hash right one and retry."
	jumpSelectorMiss jmpTarget: (jumpClassMiss jmpTarget: self Label).
	jumpSelectorMiss := self compileOpenPICMethodCacheProbeFor: selector withShift: 1 baseRegOrNone: cacheBaseReg.
	self JumpZero: itsAHit.

	"Second probe missed.  Do last probe.  Shift hash right two and retry."
	jumpSelectorMiss jmpTarget: self Label.
	jumpSelectorMiss := self compileOpenPICMethodCacheProbeFor: selector withShift: 2 baseRegOrNone: cacheBaseReg.
	self JumpZero: itsAHit.

	"Last probe missed.  Call ceSendFromInLineCacheMiss: to do the full lookup."
	jumpSelectorMiss jmpTarget: self Label.
	self numRegArgs > 0 ifTrue:
		[backEnd genPushRegisterArgsForNumArgs: numArgs scratchReg: SendNumArgsReg].
	self genSmalltalkToCStackSwitch: true.
	methodLabel addDependent: (self annotateAbsolutePCRef: (self MoveCw: methodLabel asInteger R: SendNumArgsReg)).
	self 
		compileCallFor: #ceSendFromInLineCacheMiss:
		numArgs: 1
		arg: SendNumArgsReg
		arg: nil
		arg: nil
		arg: nil
		resultReg: NoReg
		regsToSave: self emptyRegisterMask
	"Note that this call does not return."! !

!SimpleStackBasedCogit methodsFor: 'in-line cacheing' stamp: 'eem 8/4/2016 14:32'!
compileOpenPICMethodCacheProbeFor: selector withShift: shift baseRegOrNone: baseRegOrNone
	"Compile one method cache probe in an OpenPIC's lookup of selector.
	 Answer the jump taken if the selector probe fails.
	 The class tag of the receiver must be in SendNumArgsReg.  ClassReg and TempReg are used as scratch registers.
	 On a hit, the offset of the entry is in ClassReg."
	<returnTypeC: #'AbstractInstruction *'>
	<inline: false>
	| jumpSelectorMiss |
	<var: 'jumpSelectorMiss' type: #'AbstractInstruction *'>
	self MoveR: SendNumArgsReg R: ClassReg.
	objectRepresentation maybeShiftClassTagRegisterForMethodCacheProbe: ClassReg.
	self annotate: (self XorCw: selector R: ClassReg) objRef: selector.
	self assert: shift <= objectMemory shiftForWord.
	"Need to shift the hash right by shift to form the probe, and then shift the probe left by shiftForWord to form the index.
	 So shift left by shiftForWord - shift and and with the shifted mask."
	shift < objectMemory shiftForWord ifTrue:
		[self LogicalShiftLeftCq: objectMemory shiftForWord - shift R: ClassReg].
	self AndCq: MethodCacheMask << objectMemory shiftForWord R: ClassReg.
	baseRegOrNone = NoReg
		ifTrue:
			[self MoveMw: coInterpreter methodCacheAddress asUnsignedInteger + (MethodCacheSelector << objectMemory shiftForWord)
				r: ClassReg
				R: TempReg]
		ifFalse:
			[self AddR: baseRegOrNone R: ClassReg;
				MoveMw: MethodCacheSelector << objectMemory shiftForWord r: ClassReg R: TempReg].
	self annotate: (self CmpCw: selector R: TempReg) objRef: selector.
	jumpSelectorMiss := self JumpNonZero: 0.
	baseRegOrNone = NoReg
		ifTrue:
			[self MoveMw: coInterpreter methodCacheAddress asUnsignedInteger + (MethodCacheClass << objectMemory shiftForWord)
				r: ClassReg
				R: TempReg]
		ifFalse:
			[self MoveMw: MethodCacheClass << objectMemory shiftForWord r: ClassReg R: TempReg].
	self CmpR: SendNumArgsReg R: TempReg.
	^jumpSelectorMiss! !

!SimpleStackBasedCogit methodsFor: 'in-line cacheing' stamp: 'eem 8/4/2016 14:32'!
compilePerformMethodCacheProbeFor: selectorReg withShift: shift baseRegOrNone: baseRegOrNone
	"Compile one method cache probe in a perform: primitive's lookup of selector.
	 Answer the jump taken if the selector probe fails."
	<returnTypeC: #'AbstractInstruction *'>
	<inline: false>
	| jumpSelectorMiss |
	<var: 'jumpSelectorMiss' type: #'AbstractInstruction *'>
	self MoveR: SendNumArgsReg R: ClassReg.
	objectRepresentation maybeShiftClassTagRegisterForMethodCacheProbe: ClassReg.
	self XorR: selectorReg R: ClassReg.
	self assert: shift <= objectMemory shiftForWord.
	"Need to shift the hash right by shift to form the probe, and then shift the probe left by shiftForWord to form the index.
	 So shift left by shiftForWord - shift and and with the shifted mask."
	shift < objectMemory shiftForWord ifTrue:
		[self LogicalShiftLeftCq: objectMemory shiftForWord - shift R: ClassReg].
	self AndCq: MethodCacheMask << objectMemory shiftForWord R: ClassReg.
	baseRegOrNone = NoReg
		ifTrue:
			[self MoveMw: coInterpreter methodCacheAddress asUnsignedInteger + (MethodCacheSelector << objectMemory shiftForWord)
				r: ClassReg
				R: TempReg]
		ifFalse:
			[self AddR: baseRegOrNone R: ClassReg;
				MoveMw: MethodCacheSelector << objectMemory shiftForWord r: ClassReg R: TempReg].
	self CmpR: selectorReg R: TempReg.
	jumpSelectorMiss := self JumpNonZero: 0.
	baseRegOrNone = NoReg
		ifTrue:
			[self MoveMw: coInterpreter methodCacheAddress asUnsignedInteger + (MethodCacheClass << objectMemory shiftForWord)
				r: ClassReg
				R: TempReg]
		ifFalse:
			[self MoveMw: MethodCacheClass << objectMemory shiftForWord r: ClassReg R: TempReg].
	self CmpR: SendNumArgsReg R: TempReg.
	^jumpSelectorMiss! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 8/19/2021 16:12'!
compilePrimitive
	"Compile a primitive.  If possible, performance-critical primitives will
	 be generated by their own routines (primitiveGenerator).  Otherwise,
	 if there is a primitive at all, we call the C routine with the usual
	 stack-switching dance, test the primFailCode and then either return
	 on success or continue to the method body."
	<inline: false>
	| primitiveDescriptor primitiveRoutine flags |
	<var: #primitiveDescriptor type: #'PrimitiveDescriptor *'>
	<var: #primitiveRoutine declareC: 'void (*primitiveRoutine)(void)'>
	primitiveIndex = 0 ifTrue: [^0].
	"If a descriptor specifies an argument count (by numArgs >= 0) then it must match
	 for the generated code to be correct.  For example for speed many primitives use
	 ResultReceiverReg instead of accessing the stack, so the receiver better be at
	 numArgs down the stack.  Use the interpreter version if not."
	((primitiveDescriptor := self primitiveGeneratorOrNil) notNil
	 and: [primitiveDescriptor primitiveGenerator notNil
	 and: [(primitiveDescriptor primNumArgs < 0 "means generator doesn't care"
		   or: [primitiveDescriptor primNumArgs = (coInterpreter argumentCountOf: methodObj)])]]) ifTrue:
		[| opcodeIndexAtPrimitive code |
		"Note opcodeIndex so that any arg load instructions
		 for unimplemented primitives can be discarded."
		 opcodeIndexAtPrimitive := opcodeIndex.
		 code := objectRepresentation perform: primitiveDescriptor primitiveGenerator.

		(code < 0 and: [code ~= UnimplementedPrimitive]) ifTrue: "Generator failed, so no point continuing..."
			[^code].
		"If the primitive can never fail then there is nothing more that needs to be done."
		code = UnfailingPrimitive ifTrue:
			[^0].
		"If the machine code version handles all cases the only reason to call the interpreter
		 primitive is to reap the primitive error code.  Don't bother if it isn't used."
		(code = CompletePrimitive
		 and: [(self methodUsesPrimitiveErrorCode: methodObj header: methodHeader) not]) ifTrue:
			[^0].
		"Discard any arg load code generated by the primitive generator."
		code = UnimplementedPrimitive ifTrue:
			[opcodeIndex := opcodeIndexAtPrimitive]].

	primitiveRoutine := coInterpreter
							functionPointerForCompiledMethod: methodObj
							primitiveIndex: primitiveIndex
							primitivePropertyFlagsInto: (self addressOf: flags put: [:val| flags := val]).
	(flags anyMask: PrimCallDoNotJIT) ifTrue:
		[^ShouldNotJIT].

	(primitiveRoutine = 0 "no primitive"
	or: [primitiveRoutine = (self cCoerceSimple: #primitiveFail to: 'void (*)(void)')]) ifTrue:
		[^self genFastPrimFail].

	(objectRepresentation hasSpurMemoryManagerAPI
	 and: [flags anyMask: PrimCallOnSmalltalkStack]) ifTrue:
		[^self compileOnStackExternalPrimitive: primitiveRoutine flags: flags].
	minValidCallAddress := minValidCallAddress min: primitiveRoutine asUnsignedInteger.
	^self compileInterpreterPrimitive: primitiveRoutine flags: flags! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 1/21/2016 10:04'!
doubleExtendedDoAnythingBytecode
	"Replaces the Blue Book double-extended send [132], in which the first byte was wasted on 8 bits of argument count. 
	Here we use 3 bits for the operation sub-type (opType),  and the remaining 5 bits for argument count where needed. 
	The last byte give access to 256 instVars or literals. 
	See also secondExtendedSendBytecode"
	| opType |
	opType := byte1 >> 5.
	opType = 0 ifTrue:
		[^self genSend: byte2 numArgs: (byte1 bitAnd: 31)].
	opType = 1 ifTrue:
		[^self genSendSuper: byte2 numArgs: (byte1 bitAnd: 31)].
	"We need a map entry for this bytecode for correct parsing.
	 The sends will get an IsSend entry anyway.  The other cases need a fake one."
	opType caseOf: {
			[2]	->	[(coInterpreter isReadMediatedContextInstVarIndex: byte2)
						ifTrue: [self genPushMaybeContextReceiverVariable: byte2]
						ifFalse: [self genPushReceiverVariable: byte2]].
			[3]	->	[self genPushLiteralIndex: byte2].
			[4]	->	[self genPushLiteralVariable: byte2].
			[7]	->	[self genStorePop: false LiteralVariable: byte2.
					 self cppIf: IMMUTABILITY ifTrue: ["genStorePop:LiteralVariable: annotates; don't annotate twice" ^0]] }
		otherwise: "5 & 6"
			[(coInterpreter isWriteMediatedContextInstVarIndex: byte2)
				ifTrue: [self genStorePop: opType = 6 MaybeContextReceiverVariable: byte2]
				ifFalse: [self genStorePop: opType = 6 ReceiverVariable: byte2].
			 self cppIf: IMMUTABILITY ifTrue: ["genStorePop:...ReceiverVariable: annotate; don't annotate twice" ^0]].
	"We need a map entry for this bytecode for correct parsing (if the method builds a frame)."
	needsFrame ifTrue:
		[self annotateBytecode: self Label].
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 5/6/2009 10:20'!
duplicateTopBytecode
	self MoveMw: 0 r: SPReg R: TempReg.
	self PushR: TempReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'trampoline support' stamp: 'cb 11/23/2015 11:46'!
evaluateTrampolineCallBlock: block protectLinkRegIfNot: inFrame
	<inline: true>
	inFrame 
		ifFalse: 
			[ backEnd saveAndRestoreLinkRegAround: [ block value ] ]
		ifTrue:
			[ block value ].! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/20/2010 22:16'!
extendedPushBytecode
	| variableType variableIndex |
	variableType := (byte1 >> 6) bitAnd: 16r3.
	variableIndex := byte1 bitAnd: 16r3F.
	variableType = 0 ifTrue:
		[^self genPushReceiverVariable: variableIndex].
	variableType = 1 ifTrue:
		[^self genPushTemporaryVariable: variableIndex].
	variableType = 2 ifTrue:
		[^self genPushLiteralIndex: variableIndex].
	^self genPushLiteralVariable: variableIndex! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 01/21/2016 10:06'!
extendedStoreAndPopBytecode
	| variableType variableIndex |
	variableType := byte1 >> 6 bitAnd: 3.
	variableIndex := byte1 bitAnd: 63.
	variableType = 0 ifTrue:
		[^self genStorePop: true ReceiverVariable: variableIndex].
	variableType = 1 ifTrue:
		[self genStorePop: true TemporaryVariable: variableIndex.
		"needs a fake map entry if Immutability is ON..."
		self cppIf: IMMUTABILITY ifTrue: [ self annotateBytecode: self Label. ].
		^ 0].
	variableType = 3 ifTrue:
		[^self genStorePop: true LiteralVariable: variableIndex].
	^EncounteredUnknownBytecode! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 01/21/2016 10:06'!
extendedStoreBytecode
	| variableType variableIndex |
	variableType := byte1 >> 6 bitAnd: 3.
	variableIndex := byte1 bitAnd: 63.
	variableType = 0 ifTrue:
		[^self genStorePop: false ReceiverVariable: variableIndex].
	variableType = 1 ifTrue:
		[self genStorePop: false TemporaryVariable: variableIndex.
		"needs a fake map entry if Immutability is ON..."
		self cppIf: IMMUTABILITY ifTrue: [ self annotateBytecode: self Label. ].
		^ 0].
	variableType = 3 ifTrue:
		[^self genStorePop: false LiteralVariable: variableIndex].
	^EncounteredUnknownBytecode! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 8/8/2014 15:00'!
firstSpecialSelectorBytecodeOffset
	<inline: true>
	^self cppIf: MULTIPLEBYTECODESETS
		ifTrue: [bytecodeSetOffset = 256 ifTrue: [AltFirstSpecialSelector + 256] ifFalse: [FirstSpecialSelector]]
		ifFalse: [FirstSpecialSelector]! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'rsf 8/10/2016 00:25'!
frameOffsetOfNativeFrameMark
	<option: #LowcodeVM>
	^ FoxMFReceiver - BytesPerWord! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'rsf 8/10/2016 00:25'!
frameOffsetOfNativeFramePointer
	<option: #LowcodeVM>
	^ FoxMFReceiver - (BytesPerWord*3)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'rsf 8/10/2016 00:25'!
frameOffsetOfNativeStackPointer
	<option: #LowcodeVM>
	^ FoxMFReceiver - (BytesPerWord*4)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'rsf 8/10/2016 00:25'!
frameOffsetOfPreviousNativeStackPointer
	<option: #LowcodeVM>
	^ FoxMFReceiver - (BytesPerWord*2)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/20/2017 10:20'!
frameOffsetOfTemporary: index
	^self frameOffsetOfTemporary: index numArgs: methodOrBlockNumArgs! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/20/2017 10:20'!
frameOffsetOfTemporary: index numArgs: numArgs
	<inline: true>
	^index < numArgs
		ifTrue: [FoxCallerSavedIP + ((numArgs - index) * objectMemory wordSize)]
		ifFalse: [FoxMFReceiver - objectMemory wordSize + ((numArgs - index) * objectMemory wordSize)]! !

!SimpleStackBasedCogit methodsFor: 'accessing' stamp: 'eem 10/12/2016 19:05'!
fullBlockEntryOffset
	<api>
	<cmacro>
	^cbEntryOffset! !

!SimpleStackBasedCogit methodsFor: 'accessing' stamp: 'eem 10/12/2016 19:06'!
fullBlockNoContextSwitchEntryOffset
	<api>
	<cmacro>
	^cbNoSwitchEntryOffset! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 4/19/2015 17:57'!
genBlockReturn
	"Return from block, assuming result already loaded into ReceiverResultReg."
	needsFrame ifTrue:
		[self MoveR: FPReg R: SPReg.
		 self PopR: FPReg.
		 backEnd hasLinkRegister ifTrue:
			[self PopR: LinkReg]].
	self RetN: methodOrBlockNumArgs + 1 * objectMemory wordSize.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 10/27/2017 15:15'!
genCallMappedInlinedPrimitive
	"Implemented with SistaCogit only"
	^EncounteredUnknownBytecode! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 8/30/2016 10:58'!
genCallPrimitiveBytecode
	"V3PlusClosures:	139 10001011	iiiiiiii   jjjjjjjj  Call Primitive #iiiiiiii + (jjjjjjjj * 256)
	 NewsqueakV4:		249 11111001	iiiiiiii   jjjjjjjj  Call Primitive #iiiiiiii + (jjjjjjjj * 256)
	 SistaV1:			248 11111000 iiiiiiii mjjjjjjj  Call Primitive #iiiiiiii + ( jjjjjjj * 256)
							m=1 means inlined primitive, no hard return after execution."
	(bytecodePC = initialPC and: [byte2 < 128]) ifFalse: [^EncounteredUnknownBytecode].
	^0! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 8/9/2021 11:15'!
genCheckForProfileTimerTick: liveRegisterMask
	"Compare the performance counter against the nextProfileTick, if nextProfileTick
	 and answer the jump taken if the performance counter has past nextProfileTick.
	 So implement (nextProfileTick > 0 and: [self ioHighResClock > nextProfileTick]) ifTrue:"
	<inline: #always>
	objectMemory wordSize = 8
		ifTrue:
			[| reg skip jump |
			 reg := backEnd preferredRegisterForMovePerfCnt64RL = NoReg
						ifTrue: [Arg0Reg]
						ifFalse: [backEnd preferredRegisterForMovePerfCnt64RL].
			 self MoveAw: coInterpreter nextProfileTickAddress R: Arg1Reg.
			 self CmpCq: 0 R: Arg1Reg.
			 skip := self JumpZero: 0.
			 self MovePerfCnt64R: reg L: liveRegisterMask.
			 self CmpR: Arg1Reg R: reg.
			 jump := self JumpAboveOrEqual: 0.
			 skip jmpTarget: self Label.
			 ^jump]
		ifFalse:
			[| effectiveLiveRegisters regLo regHi skip jump |
			 self flag: #endianness.
			 self deny: ((self registerMaskFor: ClassReg and: SendNumArgsReg) anyMask: liveRegisterMask).
			 effectiveLiveRegisters := liveRegisterMask bitOr: (self registerMaskFor: ClassReg and: SendNumArgsReg).
			 regLo := Arg0Reg. regHi := Arg1Reg.
			 backEnd preferredRegisterPairForMovePerfCnt64RRLInto:
				[:prefRegLo :prefRegHi|
				(self register: prefRegLo isInMask: liveRegisterMask) ifFalse:
					[regLo := prefRegLo].
				(self register: prefRegHi isInMask: liveRegisterMask) ifFalse:
					[regHi := prefRegHi]].
			 self MoveAw: coInterpreter nextProfileTickAddress R: ClassReg.
			 self MoveAw: coInterpreter nextProfileTickAddress + 4 R: SendNumArgsReg.
			 self AddR: ClassReg R: SendNumArgsReg R: regLo.
			 skip := self JumpZero: 0.
			 self MovePerfCnt64R: regLo R: regHi L: liveRegisterMask.
			 self SubR: ClassReg R: regLo.
			 self SubbR: SendNumArgsReg R: regHi.
			 jump := self JumpAboveOrEqual: 0.
			 skip jmpTarget: self Label.
			 ^jump]! !

!SimpleStackBasedCogit methodsFor: 'constant support' stamp: 'cb 4/8/2016 11:56'!
genCmpConstant: constant R: register
	"If the objectMemory allows it, generates a quick constant cmp, else generates a word constant cmp"
	<inline: true>
	^ (objectRepresentation shouldAnnotateObjectReference: constant)
		ifTrue: [ self annotate: (self CmpCw: constant R: register) objRef: constant ]
		ifFalse: [ self CmpCq: constant R: register ]
	! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'nice 7/26/2016 00:14'!
genDoubleFailIfZeroArgRcvr: rcvrReg arg: argReg
	<option: #DPFPReg0>
	<var: 'rcvrReg' type: #int>
	<var: 'argReg' type: #int>
	<returnTypeC: #'AbstractInstruction *'>
	self MoveCq: 0 R: TempReg.
	self ConvertR: TempReg Rd: DPFPReg2.
	self CmpRd: DPFPReg2 Rd: argReg.
	^self JumpFPEqual: 0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:01'!
genExtJumpIfFalse
	"244		11110100	i i i i i i i i	Pop and Jump 0n False i i i i i i i i (+ Extend B * 256, where Extend B >= 0)"
	| distance target |
	distance := byte1 + (extB << 8).
	self assert: distance = (self v4: (self generatorAt: byte0)
								LongForward: bytecodePC
								Branch: (extA ~= 0 ifTrue: [1] ifFalse: [0]) + (extB ~= 0 ifTrue: [1] ifFalse: [0])
								Distance: methodObj).
	extB := 0.
	numExtB := 0.
	target := distance + 2 + bytecodePC.
	^self genJumpIf: objectMemory falseObject to: target! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:01'!
genExtJumpIfTrue
	"243		11110011	i i i i i i i i	Pop and Jump 0n True i i i i i i i i (+ Extend B * 256, where Extend B >= 0)"
	| distance target |
	distance := byte1 + (extB << 8).
	self assert: distance = (self v4: (self generatorAt: byte0)
								LongForward: bytecodePC
								Branch: (extA ~= 0 ifTrue: [1] ifFalse: [0]) + (extB ~= 0 ifTrue: [1] ifFalse: [0])
								Distance: methodObj).
	extB := 0.
	numExtB := 0.
	target := distance + 2 + bytecodePC.
	^self genJumpIf: objectMemory trueObject to: target! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:01'!
genExtNopBytecode
	"NewspeakV4: 221		11011101		Nop"
	"SistaV1:		 91		01011011'		Nop"
	extA := numExtB := extB := 0.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:02'!
genExtPushCharacterBytecode
	"SistaV1:		233		11101001	iiiiiiii		Push Character #iiiiiiii (+ Extend B * 256)"
	| value |
	value := byte1 + (extB << 8).
	extB := 0.
	numExtB := 0.
	^self genPushLiteral: (objectMemory characterObjectOf: value)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 2/10/2017 14:33'!
genExtPushClosureBytecode
	"Block compilation.  At this point in the method create the block.  Note its start
	 and defer generating code for it until after the method and any other preceding
	 blocks.  The block's actual code will be compiled later."
	"253		11111101 eei i i kkk	jjjjjjjj		Push Closure Num Copied iii (+ Ext A // 16 * 8) Num Args kkk (+ Ext A \\ 16 * 8) BlockSize jjjjjjjj (+ Ext B * 256). ee = num extensions"
	| startpc numArgs numCopied |
	self assert: needsFrame.
	startpc := bytecodePC + (self generatorAt: byte0) numBytes.
	self addBlockStartAt: startpc "0 relative"
		numArgs: (numArgs := (byte1 bitAnd: 16r7) + (extA \\ 16 * 8))
		numCopied: (numCopied := ((byte1 >> 3) bitAnd: 7) + (extA // 16 * 8))
		span: byte2 + (extB << 8).
	extA := numExtB := extB := 0.
	objectRepresentation
		genCreateClosureAt: startpc + 1 "1 relative"
		numArgs: numArgs
		numCopied: numCopied
		contextNumArgs: methodOrBlockNumArgs
		large: (coInterpreter methodNeedsLargeContext: methodObj)
		inBlock: inBlock.
	self PushR: ReceiverResultReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 4/8/2016 15:41'!
genExtPushFullClosureBytecode
	"Full Block creation compilation. The block's actual code will be compiled separatedly."
	"*	255		11111111	xxxxxxxx	siyyyyyy	push Closure Compiled block literal index xxxxxxxx (+ Extend A * 256) numCopied yyyyyy receiverOnStack: s = 1 ignoreOuterContext: i = 1"
	| numCopied ignoreContext receiverIsOnStack compiledBlock |
	self assert: needsFrame.
	compiledBlock := self getLiteral: byte1 + (extA << 8).
	extA := 0.
	numCopied := byte2 bitAnd: 1<< 6 - 1.
	receiverIsOnStack := byte2 anyMask: 1 << 7.
	ignoreContext := byte2 anyMask: 1 << 6.
	objectRepresentation
		genCreateFullClosure: compiledBlock
		numArgs: (coInterpreter argumentCountOf: compiledBlock)
		numCopied: numCopied
		ignoreContext: ignoreContext
		contextNumArgs: methodOrBlockNumArgs
		large: (coInterpreter methodNeedsLargeContext: methodObj)
		inBlock: inBlock.
	"Closure in ReceiverResultReg"
	1 to: numCopied do:
		[:i|
		self
			PopR: TempReg;
			MoveR: TempReg
				Mw: numCopied - i + FullClosureFirstCopiedValueIndex * objectMemory bytesPerOop + objectMemory baseHeaderSize
					r: ReceiverResultReg].
	receiverIsOnStack
		ifTrue: [self PopR: TempReg]
		ifFalse: [ self MoveMw: FoxMFReceiver r: FPReg R: TempReg].
	self
		MoveR: TempReg
		Mw: FullClosureReceiverIndex * objectMemory bytesPerOop + objectMemory baseHeaderSize
		r: ReceiverResultReg.
	self PushR: ReceiverResultReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:02'!
genExtPushIntegerBytecode
	"NewsqueakV4:	229		11100101	iiiiiiii	Push Integer #iiiiiiii (+ Extend B * 256, where bbbbbbbb = sddddddd, e.g. -32768 = i=0, a=0, s=1)
	SistaV1:		232		11101000	iiiiiiii	Push Integer #iiiiiiii (+ Extend B * 256, where bbbbbbbb = sddddddd, e.g. -32768 = i=0, a=0, s=1)"
	| value |
	value := byte1 + (extB << 8).
	extB := 0.
	numExtB := 0.
	^self genPushLiteral: (objectMemory integerObjectOf: value)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/5/2012 16:12'!
genExtPushLiteralBytecode
	"228		11100100	i i i i i i i i	Push Literal #iiiiiiii (+ Extend A * 256)"
	| index |
	index := byte1 + (extA << 8).
	extA := 0.
	^self genPushLiteralIndex: index! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/5/2012 16:05'!
genExtPushLiteralVariableBytecode
	"227		11100011	i i i i i i i i	Push Literal Variable #iiiiiiii (+ Extend A * 256)"
	| index |
	index := byte1 + (extA << 8).
	extA := 0.
	^self genPushLiteralVariable: index! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:02'!
genExtPushPseudoVariable
	"SistaV1: *	82			01010010			Push thisContext, (then Extend B = 1 => push thisProcess)"
	| ext |
	ext := extB.
	extB := 0.
	numExtB := 0.
	ext caseOf: {
		[0]	->	[^self genPushActiveContextBytecode].
		}
		otherwise:
			[^self unknownBytecode].
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:02'!
genExtPushPseudoVariableOrOuterBytecode
	"77			01001101		Push false [* 1:true, 2:nil, 3:thisContext, ..., -N: pushEnclosingObjectAt: N, N = Extend B]"
	| ext |
	ext := extB.
	extB := 0.
	numExtB := 0.
	ext caseOf: {
		[0]	->	[^self genPushLiteral: objectMemory falseObject].
		[1]	->	[^self genPushLiteral: objectMemory trueObject].
		[2]	->	[^self genPushLiteral: objectMemory nilObject].
		[3]	->	[^self genPushActiveContextBytecode]
		}
		otherwise:
			[ext < 0 ifTrue:
				[^self genPushEnclosingObjectAt: 0 - ext].
			 self warning: 'undefined extension for extPushPseudoVariableOrOuter'.
			 ^self unknownBytecode].
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 8/31/2021 14:37'!
genExtPushReceiverVariableBytecode
	"226		11100010	i i i i i i i i	Push Receiver Variable #iiiiiiii (+ Extend A * 256)"
	| index |
	index := byte1 + (extA << 8).
	extA := 0.
	^(self mclassCouldBeContext
	  and: [coInterpreter isReadMediatedContextInstVarIndex: index])
		ifTrue: [self genPushMaybeContextReceiverVariable: index]
		ifFalse: [self genPushReceiverVariable: index]! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:03'!
genExtPushRemoteTempOrInstVarLongBytecode
	| index |
	^ (byte2 noMask: coInterpreter remoteIsInstVarAccess)
		ifTrue: [ self genPushRemoteTempLongBytecode ]
		ifFalse: 
			[ index := byte1 + (extA << 8).
			extA := 0.
			extB := 0. "don't use flags in the simple cogit"
			numExtB := 0.
			(coInterpreter isReadMediatedContextInstVarIndex: index)
				ifTrue: [self 
							genPushMaybeContextRemoteInstVar: index 
							inObjectAt: byte2 - coInterpreter remoteIsInstVarAccess]
				ifFalse: [self 
							genPushRemoteInstVar: index 
							inObjectAt: byte2 - coInterpreter remoteIsInstVarAccess]]! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/15/2012 14:06'!
genExtReturnTopFromBlock
	"218		11011010		Return Stack Top From Block [* return from enclosing block N, N = Extend A]
	 If extA is zero, return to the caller of the current block activation.
	 If extA is non-zero return to the caller of the Nth enclosing block activation."
	extA = 0 ifTrue:
		[^self genReturnTopFromBlock].
	self shouldBeImplemented.
	extA := 0.
	^EncounteredUnknownBytecode! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:03'!
genExtSendAbsentDynamicSuperBytecode
	"241		11110001	i i i i i j j j	Send To Absent Dynamic Superclass Literal Selector #iiiii (+ Extend A * 32) with jjj (+ Extend B * 8) Arguments"
	| litIndex nArgs |
	litIndex := (byte1 >> 3) + (extA << 5).
	extA := 0.
	nArgs := (byte1 bitAnd: 7) + (extB << 3).
	extB := 0.
	numExtB := 0.
	^self genSendAbsentDynamicSuper: litIndex numArgs: nArgs! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:03'!
genExtSendAbsentImplicitBytecode
	"240		11110000	i i i i i j j j	Send To Absent Implicit Receiver Literal Selector #iiiii (+ Extend A * 32) with jjj (+ Extend B * 8) Arguments"
	| litIndex nArgs |
	litIndex := (byte1 >> 3) + (extA << 5).
	extA := 0.
	nArgs := (byte1 bitAnd: 7) + (extB << 3).
	extB := 0.
	numExtB := 0.
	^self genSendAbsentImplicit: litIndex numArgs: nArgs! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:03'!
genExtSendAbsentOuterBytecode
	"254		 11111110	i i i i i j j j kkkkkkkk	Send To Absent Outer Literal Selector #iiiii (+ Extend A * 32) with jjj (+ Extend B * 8) Arguments at Depth kkkkkkkk "
	| litIndex nArgs depth |
	litIndex := (byte1 >> 3) + (extA << 5).
	extA := 0.
	nArgs := (byte1 bitAnd: 7) + (extB << 3).
	extB := 0.
	numExtB := 0.
	depth := byte2.
	^self genSendAbsentOuter: litIndex numArgs: nArgs depth: depth
! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:03'!
genExtSendAbsentSelfBytecode
	"245		11110101	i i i i i j j j	Send To Absent Self Literal Selector #iiiii (+ Extend A * 32) with jjj (+ Extend B * 8) Arguments"
	| litIndex nArgs |
	litIndex := (byte1 >> 3) + (extA << 5).
	extA := 0.
	nArgs := (byte1 bitAnd: 7) + (extB << 3).
	extB := 0.
	numExtB := 0.
	^self genSendAbsentSelf: litIndex numArgs: nArgs! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:03'!
genExtSendBytecode
	"238		11101110	i i i i i j j j	Send Literal Selector #iiiii (+ Extend A * 32) with jjj (+ Extend B * 8) Arguments"
	| litIndex nArgs |
	litIndex := (byte1 >> 3) + (extA << 5).
	extA := 0.
	nArgs := (byte1 bitAnd: 7) + (extB << 3).
	extB := 0.
	numExtB := 0.
	^self genSend: litIndex numArgs: nArgs! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 1/9/2020 19:01'!
genExtSendSuperBytecode
	"239		11101111	i i i i i j j j	Send To Superclass Literal Selector #iiiii (+ Extend A * 32) with jjj (+ Extend B * 8) Arguments"
	| isDirected litIndex nArgs |
	(isDirected := extB >= 64) ifTrue:
		[extB := extB bitAnd: 63].
	litIndex := (byte1 >> 3) + (extA << 5).
	extA := 0.
	nArgs := (byte1 bitAnd: 7) + (extB << 3).
	extB := 0.
	numExtB := 0.
	BytecodeSetHasDirectedSuperSend
		ifTrue:
			[^isDirected
				ifTrue: [self genSendDirectedSuper: litIndex numArgs: nArgs]
				ifFalse: [self genSendSuper: litIndex numArgs: nArgs]]
		ifFalse:
			[self deny: isDirected.
			 ^self genSendSuper: litIndex numArgs: nArgs]! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/5/2012 16:12'!
genExtStoreAndPopLiteralVariableBytecode
	"236		11101100	i i i i i i i i	Pop and Store Literal Variable #iiiiiiii (+ Extend A * 256)"
	| index |
	index := byte1 + (extA << 8).
	extA := 0.
	^self genStorePop: true LiteralVariable: index! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 8/31/2021 14:33'!
genExtStoreAndPopReceiverVariableBytecode
	"235		11101011	i i i i i i i i	Pop and Store Receiver Variable #iiiiiii (+ Extend A * 256)"
	| index |
	index := byte1 + (extA << 8).
	extA := 0.
	^(self mclassCouldBeContext
	   and: [coInterpreter isWriteMediatedContextInstVarIndex: index])
		ifTrue: [self genStorePop: true MaybeContextReceiverVariable: index]
		ifFalse: [self genStorePop: true ReceiverVariable: index]! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 4/21/2016 16:17'!
genExtStoreAndPopRemoteTempOrInstVarLongBytecode
	^ self genExtStorePopRemoteTempOrInstVarLongBytecodePopBoolean: true! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/5/2012 16:12'!
genExtStoreLiteralVariableBytecode
	"233		11101001	i i i i i i i i	Store Literal Variable #iiiiiiii (+ Extend A * 256)"
	| index |
	index := byte1 + (extA << 8).
	extA := 0.
	^self genStorePop: false LiteralVariable: index! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:03'!
genExtStorePopRemoteTempOrInstVarLongBytecodePopBoolean: popBoolean
	| index |
	extB := 0. "simple cogit don't use the extra flag"
	numExtB := 0.
	(byte2 noMask: coInterpreter remoteIsInstVarAccess)
		ifTrue: 
			[ self genStorePop: popBoolean RemoteTemp: byte1 At: byte2.
			self cppIf: IMMUTABILITY ifTrue: [ self annotateBytecode: self Label ] ]
		ifFalse: 
			[ index := byte1 + (extA << 8).
			extA := 0.
			(coInterpreter isWriteMediatedContextInstVarIndex: index)
				ifTrue: [ self 
						genStorePop: popBoolean 
						MaybeContextRemoteInstVar: index 
						ofObjectAt: byte2 - coInterpreter remoteIsInstVarAccess ]
				ifFalse: [ self 
						genStorePop: popBoolean 
						RemoteInstVar: index 
						ofObjectAt: byte2 - coInterpreter remoteIsInstVarAccess  ] ].
	^ 0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 8/31/2021 14:34'!
genExtStoreReceiverVariableBytecode
	"232		11101000	i i i i i i i i	Store Receiver Variable #iiiiiii (+ Extend A * 256)"
	| index |
	index := byte1 + (extA << 8).
	extA := 0.
	^(self mclassCouldBeContext
	   and: [coInterpreter isWriteMediatedContextInstVarIndex: index])
		ifTrue: [self genStorePop: false MaybeContextReceiverVariable: index]
		ifFalse: [self genStorePop: false ReceiverVariable: index]! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 4/21/2016 16:16'!
genExtStoreRemoteTempOrInstVarLongBytecode
	^ self genExtStorePopRemoteTempOrInstVarLongBytecodePopBoolean: false! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 8/7/2014 06:19'!
genExtTrapIfNotInstanceOfBehaviorsBytecode
	"SistaV1: *	236		11101100	iiiiiiii		Trap If Not Instance Of Behavior/Array Of Behavior #iiiiiiii (+ Extend A * 256, where Extend A >= 0)"
	"This really only makes sense for the optimizing JITs"
	^EncounteredUnknownBytecode! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:04'!
genExtUnconditionalJump
	"242		11110010	i i i i i i i i	Jump i i i i i i i i (+ Extend B * 256, where bbbbbbbb = sddddddd, e.g. -32768 = i=0, a=0, s=1)"
	| distance target |
	distance := byte1 + (extB << 8).
	self assert: distance = (self v4: (self generatorAt: byte0)
								Long: bytecodePC
								Branch: (extA ~= 0 ifTrue: [1] ifFalse: [0]) + (extB ~= 0 ifTrue: [1] ifFalse: [0])
								Distance: methodObj).
	extB := 0.
	numExtB := 0.
	target := distance + 2 + bytecodePC.
	distance < 0 ifTrue:
		[^self genJumpBackTo: target].
	self genJumpTo: target.
	"The bytecode must be mapped since it can be either forward or backward, and
	  backwards branches must be mapped. So if forward, we need to map."
	self annotateBytecode: self lastOpcode.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 10/6/2015 15:01'!
genExtendedSendBytecode
	"Can use any of the first 32 literals for the selector and pass up to 7 arguments."

	^self genSend: (byte1 bitAnd: 16r1F) numArgs: byte1 >> 5! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 10/6/2015 15:01'!
genExtendedSuperBytecode
	^self genSendSuper: (byte1 bitAnd: 16r1F) numArgs: byte1 >> 5! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 14:01'!
genFastPrimFail
	primitiveIndex := 0.
	^UnfailingPrimitive! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 1/9/2020 07:00'!
genFastPrimTraceUsing: r1 and: r2
	"Suport for compileInterpreterPrimitive.  Generate inline code so as to record the primitive
	 trace as fast as possible."
	backEnd byteReadsZeroExtend ifFalse:
		[self MoveCq: 0 R: r2].
	self MoveAb: coInterpreter primTraceLogIndexAddress R: r2.
	self LoadEffectiveAddressMw: 1 r: r2 R: r1.
	self MoveR: r1 Ab: coInterpreter primTraceLogIndexAddress.
	methodLabel addDependent:
		(self annotateAbsolutePCRef:
			(self MoveCw: methodLabel asInteger R: r1)).
	self MoveMw: (self offset: CogMethod of: #selector) r: r1 R: TempReg.
	self MoveCw: coInterpreter primTraceLogAddress asInteger R: r1.
	self MoveR: TempReg Xwr: r2 R: r1! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'cb 11/24/2016 10:44'!
genInlinedIdenticalOrNotIf: orNot
	| jumpNotEqual jumpPush |
	<var: #jumpNotEqual type: #'AbstractInstruction *'>
	<var: #jumpPush type: #'AbstractInstruction *'>
	self PopR: Arg0Reg.
	objectRepresentation
		genEnsureOopInRegNotForwarded: Arg0Reg
		scratchReg: TempReg.
	self MoveMw: 0 r: SPReg R: ClassReg.
	objectRepresentation
		genEnsureOopInRegNotForwarded: ClassReg
		scratchReg: TempReg.
	self CmpR: Arg0Reg R: ClassReg.
	jumpNotEqual := self genConditionalBranch: (orNot ifTrue: [JumpZero] ifFalse: [JumpZero]) operand: 0.
	self annotate: (self genMoveTrueR: Arg0Reg)
		objRef: objectMemory trueObject.
	jumpPush := self Jump: 0.
	jumpNotEqual jmpTarget: (self genMoveFalseR: Arg0Reg).
	jumpPush jmpTarget: (self MoveR: Arg0Reg Mw: 0 r: SPReg).
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 02/09/2017 09:39'!
genJumpBackTo: targetBytecodePC
	self MoveAw: coInterpreter stackLimitAddress R: TempReg.
	self CmpR: TempReg R: SPReg. "N.B. FLAGS := SPReg - TempReg"
	self JumpAboveOrEqual: (self fixupAt: targetBytecodePC).
	self CallRT: ceCheckForInterruptTrampoline.
	self annotateBytecode: self Label.
	self Jump: (self fixupAt: targetBytecodePC).
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/21/2018 13:06'!
genJumpIf: boolean to: targetBytecodePC
	<inline: false>
	"Cunning trick by LPD.  If true and false are contiguous subtract the smaller.
	 Correct result is either 0 or the distance between them.  If result is not 0 or
	 their distance send mustBeBoolean."
	| ok |
	<var: #ok type: #'AbstractInstruction *'>
	extA := 0.
	self assert: (objectMemory objectAfter: objectMemory falseObject) = objectMemory trueObject.
	self PopR: TempReg.
	self genSubConstant: boolean R: TempReg.
	self JumpZero: (self ensureFixupAt: targetBytecodePC).
	self CmpCq: (boolean = objectMemory falseObject
					ifTrue: [objectMemory trueObject - objectMemory falseObject]
					ifFalse: [objectMemory falseObject - objectMemory trueObject])
		R: TempReg.
	ok := self JumpZero: 0.
	self genCallMustBeBooleanFor: boolean.
	ok jmpTarget: (self annotateBytecode: self Label).
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 02/07/2017 15:59'!
genJumpTo: targetBytecodePC
	self Jump: (self ensureFixupAt: targetBytecodePC).
	^0! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 4/30/2015 11:39'!
genLoadArgAtDepth: n into: reg
	"Load an argument at depth from top-of-stack (0 relative) into a register.
	 The actual offset depends on whether there's a link register or not."
	self MoveMw: (backEnd hasLinkRegister ifTrue: [n] ifFalse: [n + 1]) * objectMemory wordSize
		r: SPReg
		R: reg! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/2/2012 10:57'!
genLongJumpIfFalse
	| distance target |
	distance := self v3: (self generatorAt: byte0) LongForward: bytecodePC Branch: 0 Distance: methodObj.
	target := distance + 2 + bytecodePC.
	^self genJumpIf: objectMemory falseObject to: target! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/2/2012 10:57'!
genLongJumpIfTrue
	| distance target |
	distance := self v3: (self generatorAt: byte0) LongForward: bytecodePC Branch: 0 Distance: methodObj.
	target := distance + 2 + bytecodePC.
	^self genJumpIf: objectMemory trueObject to: target! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/4/2012 13:16'!
genLongPushTemporaryVariableBytecode
	"230		11100110	i i i i i i i i	Push Temporary Variable #iiiiiiii"
	^self genPushTemporaryVariable: byte1! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/4/2012 13:16'!
genLongStoreAndPopTemporaryVariableBytecode
	"237		11101101	i i i i i i i i	Pop and Store Temporary Variable #iiiiiiii"
	^self genStorePop: true TemporaryVariable: byte1! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/4/2012 13:16'!
genLongStoreTemporaryVariableBytecode
	"234		11101010	i i i i i i i i	Store Temporary Variable #iiiiiiii"
	^self genStorePop: false TemporaryVariable: byte1! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 2/22/2017 18:09'!
genLongUnconditionalBackwardJump
	| distance |
	distance := self v3: (self generatorAt: byte0)
					Long: bytecodePC
					Branch: 0
					Distance: methodObj.
	self assert: distance < 0.
	^self genJumpBackTo: distance + 2 + bytecodePC! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/2/2012 10:57'!
genLongUnconditionalForwardJump
	| distance targetpc |
	distance := self v3: (self generatorAt: byte0)
					Long: bytecodePC
					Branch: 0
					Distance: methodObj.
	self assert: distance >= 0.
	targetpc := distance + 2 + bytecodePC.
	^self genJumpTo: targetpc! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 05/07/2020 16:13'!
genLookupForPerformNumArgs: numArgs
	"Compile the code for a probe of the first-level method cache for a perform primitive.
	 The selector is assumed to be in Arg0Reg.  Defer to adjustArgumentsForPerform: to
	 adjust the arguments before the jump to the method."
	| jumpSelectorMiss jumpClassMiss jumpInterpret itsAHit cacheBaseReg |
	<var: #jumpSelectorMiss type: #'AbstractInstruction *'>
	<var: #jumpClassMiss type: #'AbstractInstruction *'>
	<var: #jumpInterpret type: #'AbstractInstruction *'>
	<var: #itsAHit type: #'AbstractInstruction *'>

	"N.B.  Can't assume TempReg already contains the tag because a method can
	 of course be invoked via the unchecked entry-point, e.g. as does perform:."
	objectRepresentation genGetInlineCacheClassTagFrom: ReceiverResultReg into: SendNumArgsReg forEntry: false.

	self flag: #lookupInMethodCacheSel:classTag:. "so this method shows up as a sender of lookupInMethodCacheSel:class:"

	cacheBaseReg := NoReg.
	(backEnd isWithinMwOffsetRange: coInterpreter methodCacheAddress) ifFalse:
		[self MoveCq: coInterpreter methodCacheAddress R: (cacheBaseReg := Extra0Reg)].

	"Do first of three probes.  See CoInterpreter>>lookupInMethodCacheSel:classTag:"
	jumpSelectorMiss := self compilePerformMethodCacheProbeFor: Arg0Reg withShift: 0 baseRegOrNone: cacheBaseReg.
	jumpClassMiss := self JumpNonZero: 0.

	"Fetch the method, and check if it is cogged."
	itsAHit := self MoveMw: (cacheBaseReg = NoReg
								ifTrue: [coInterpreter methodCacheAddress asUnsignedInteger + (MethodCacheMethod << objectMemory shiftForWord)]
								ifFalse: [MethodCacheMethod << objectMemory shiftForWord])
					r: ClassReg
					R: SendNumArgsReg.
	"If the method is not compiled fall back on the interpreter primitive."
	objectRepresentation genLoadSlot: HeaderIndex sourceReg: SendNumArgsReg destReg: ClassReg.
	jumpInterpret := objectRepresentation genJumpImmediate: ClassReg.
	"Adjust arguments and jump to the method's unchecked entry-point."
	self AddCq: cmNoCheckEntryOffset R: ClassReg.
	self adjustArgumentsForPerform: numArgs.
	self JumpR: ClassReg.

	"First probe missed.  Do second of three probes.  Shift hash right one and retry."
	jumpSelectorMiss jmpTarget: (jumpClassMiss jmpTarget: self Label).
	jumpSelectorMiss := self compilePerformMethodCacheProbeFor: Arg0Reg withShift: 1 baseRegOrNone: cacheBaseReg.
	self JumpZero: itsAHit.

	"Second probe missed.  Do last probe.  Shift hash right two and retry."
	jumpSelectorMiss jmpTarget: self Label.
	jumpSelectorMiss := self compilePerformMethodCacheProbeFor: Arg0Reg withShift: 2 baseRegOrNone: cacheBaseReg.
	self JumpZero: itsAHit.

	"Last probe missed.  Caller will generate the call to fall back on the interpreter primitive."
	jumpSelectorMiss jmpTarget:
	(jumpInterpret jmpTarget: self Label).
	^0! !

!SimpleStackBasedCogit methodsFor: 'constant support' stamp: 'cb 4/8/2016 12:10'!
genMoveConstant: constant R: reg
	"If the objectMemory allows it, generates a quick constant move, else generates a word constant move"
	<inline: true>
	^ (objectRepresentation shouldAnnotateObjectReference: constant)
		ifTrue: [ self annotate: (self MoveCw: constant R: reg) objRef: constant ]
		ifFalse: [ self MoveCq: constant R: reg ]
	
	! !

!SimpleStackBasedCogit methodsFor: 'constant support' stamp: 'cb 4/25/2015 15:04'!
genMoveFalseR: reg
	<inline: true>
	^ self genMoveConstant: objectMemory falseObject R: reg
	! !

!SimpleStackBasedCogit methodsFor: 'constant support' stamp: 'cb 4/8/2016 12:07'!
genMoveNilR: reg
	<inline: true>
	^ self genMoveConstant: objectMemory nilObject R: reg
	! !

!SimpleStackBasedCogit methodsFor: 'constant support' stamp: 'cb 4/25/2015 15:04'!
genMoveTrueR: reg
	<inline: true>
	^ self genMoveConstant: objectMemory trueObject R: reg
	! !

!SimpleStackBasedCogit methodsFor: 'initialization' stamp: 'eem 03/03/2016 12:23'!
genMustBeBooleanTrampolineFor: boolean called: trampolineName
	<var: #trampolineName type: #'char *'>
	<inline: false>
	self zeroOpcodeIndex.
	"If the objectRepresentation does want true & false to be mobile then we need to record these addresses."
	self assert: (objectRepresentation shouldAnnotateObjectReference: boolean) not.
	self AddCq: boolean R: TempReg.
	^self genTrampolineFor: #ceSendMustBeBoolean:
		called: trampolineName
		numArgs: 1
		arg: TempReg
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: true! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 9/9/2019 09:54'!
genNSSend: selectorIndex numArgs: numArgs depth: depth sendTable: sendTable
	<var: #sendTable type: #'sqInt *'>
	| selector nsSendCache |
	self assert: (selectorIndex between: 0 and: (objectMemory literalCountOf: methodObj) - 1).
	selector := self getLiteral: selectorIndex.
	self assert: (objectMemory addressCouldBeOop: selector).	
	(objectMemory isNonImmediate: selector) ifTrue:
		[self setHasMovableLiteral: true].
	(objectMemory isYoung: selector) ifTrue:
		[self setHasYoungReferent: true].

	nsSendCache := theIRCs + (NumOopsPerNSC * objectMemory bytesPerOop * indexOfIRC).
	indexOfIRC := indexOfIRC + 1.
	self assert: (objectMemory isInOldSpace: nsSendCache).
	self initializeNSSendCache: nsSendCache selector: selector numArgs: numArgs depth: depth.

	"This leaves the method receiver on the stack, which might not be the implicit receiver.
	 But the lookup trampoline will establish the on-stack receiver once it locates it."
	self marshallAbsentReceiverSendArguments: numArgs.

	"Load the cache last so it is a fixed distance from the call."
	self MoveUniqueCw: nsSendCache R: SendNumArgsReg.
	self CallNewspeakSend: (sendTable at: (numArgs min: NumSendTrampolines - 1)).

	self PushR: ReceiverResultReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem (const to msg rewrite) 11/6/2014 11:21'!
genPopStackBytecode
	self AddCq: objectMemory wordSize R: SPReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 13:24'!
genPrimReturn
	"Generate a return that cuts back the stack to remove the receiver
	 and arguments after an invocation of a primitive with nargs arguments.
	 This is similar to a Pascal calling convention."
	<inline: true>
	^self RetN: methodOrBlockNumArgs + 1 * objectMemory wordSize! !

!SimpleStackBasedCogit methodsFor: 'initialization' stamp: 'eem 8/8/2021 10:42'!
genPrimReturnEnterCogCodeEnilopmart: profiling
	"Generate the substitute return code for an external or FFI primitive call.
	 On success simply return, extracting numArgs from newMethod.
	 On primitive failure call ceActivateFailingPrimitiveMethod: newMethod."
	| jmpSample continuePostSample jmpFail |
	<var: #jmpSample type: #'AbstractInstruction *'>
	<var: #continuePostSample type: #'AbstractInstruction *'>
	<var: #jmpFail type: #'AbstractInstruction *'>
	self zeroOpcodeIndex.
	backEnd hasVarBaseRegister ifTrue:
		[self MoveCq: self varBaseAddress R: VarBaseReg]. "Must happen sometime"

	"Test primitive failure"
	self MoveAw: coInterpreter primFailCodeAddress R: TempReg.
	self flag: 'ask concrete code gen if move sets condition codes?'.
	self CmpCq: 0 R: TempReg.
	jmpFail := self JumpNonZero: 0.

	(profiling and: [backEnd has64BitPerformanceCounter]) ifTrue:
		[jmpSample := self genCheckForProfileTimerTick: (self registerMaskFor: NoReg).
		continuePostSample := self Label].

	"Switch back to the Smalltalk stack.  Stack better be in either of these two states:
		success:	stackPointer	->	result (was receiver)
										arg1
										...
										argN
										return pc
		failure:							receiver
										arg1
										...
					stackPointer	->	argN
										return pc
	We push the instructionPointer to reestablish the return pc in the success case,
	but leave it to ceActivateFailingPrimitiveMethod: to do so in the failure case."

	backEnd hasLinkRegister
		ifTrue:
			[backEnd genLoadStackPointers.											"Switch back to Smalltalk stack."
			 backEnd hasPCRegister
				ifTrue:
					[self PopR: ReceiverResultReg.										"Pop result from stack"
					 self MoveAw: coInterpreter instructionPointerAddress R: PCReg]	"Return"
				ifFalse:
					[self MoveMw: 0 r: SPReg R: ReceiverResultReg.						"Fetch result from stack"
					 self MoveAw: coInterpreter instructionPointerAddress R: LinkReg.	"Get ret pc"
					 self RetN: objectMemory wordSize]]								"Return, popping result from stack"
		ifFalse:
			[self MoveAw: coInterpreter instructionPointerAddress R: ClassReg.	"Get return pc"
			 backEnd genLoadStackPointers.									"Switch back to Smalltalk stack."
			 self MoveMw: 0 r: SPReg R: ReceiverResultReg.						"Fetch result from stack"
			 self MoveR: ClassReg Mw: 0 r: SPReg.								"Restore return pc"
			 self RetN: 0].														"Return, popping result from stack"

	"Primitive failed.  Invoke C code to build the frame and continue."
	jmpFail jmpTarget: (self MoveAw: coInterpreter newMethodAddress R: SendNumArgsReg).
	"Reload sp with CStackPointer; easier than popping args of checkProfileTick."
	self MoveAw: self cStackPointerAddress R: SPReg.
	self 
		compileCallFor: #ceActivateFailingPrimitiveMethod:
		numArgs: 1
		arg: SendNumArgsReg
		arg: nil
		arg: nil
		arg: nil
		resultReg: NoReg
		regsToSave: self emptyRegisterMask.

	"On Spur ceActivateFailingPrimitiveMethod: may retry the primitive and return if successful.
	 So continue by returning to the caller.
	 Switch back to the Smalltalk stack.  Stack should be in this state:
				success:	stackPointer ->	result (was receiver)
											arg1
											...
											argN
											return pc
	 We can push the instructionPointer or load it into the LinkRegister to reestablish the return pc"
	self MoveAw: coInterpreter instructionPointerAddress
		R: (backEnd hasLinkRegister ifTrue: [LinkReg] ifFalse: [ClassReg]).
	backEnd genLoadStackPointers.
	backEnd hasLinkRegister
		ifTrue:
			[self MoveMw: 0 r: SPReg R: ReceiverResultReg]		"Fetch result from stack"
		ifFalse:
			[self MoveMw: objectMemory wordSize r: SPReg R: ReceiverResultReg.	"Fetch result from stack"
			 self PushR: ClassReg].													"Restore return pc on CISCs"
	self RetN: objectMemory wordSize.	"return to caller, popping receiver"

	(profiling and: [backEnd has64BitPerformanceCounter]) ifTrue:
		["Call ceTakeProfileSample: to record sample and then continue.  newMethod
		 should be up-to-date.  Need to save and restore the link reg around this call."
		 jmpSample jmpTarget: self Label.
		 backEnd genMarshallNArgs: 1 arg: 0 arg: 0 arg: 0 arg: 0.
		 self CallFullRT: (self cCode: [#ceTakeProfileSample: asUnsignedInteger]
							inSmalltalk: [self simulatedTrampolineFor: #ceTakeProfileSample:]).
		 backEnd genRemoveNArgsFromStack: 1.
		 self Jump: continuePostSample]! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 7/24/2021 16:57'!
genPrimitiveClosureValue
	"Check the argument count.  Fail if wrong.
	 Get the method from the outerContext and see if it is cogged.  If so, jump to the
	 block entry or the no-context-switch entry, as appropriate, and we're done.  If not,
	 invoke the interpreter primitive."
	| jumpFailNArgs jumpFail1 jumpFail2 jumpFail3 jumpFail4 jumpBCMethod primitiveRoutine result |
	<var: #jumpFail1 type: #'AbstractInstruction *'>
	<var: #jumpFail2 type: #'AbstractInstruction *'>
	<var: #jumpFail3 type: #'AbstractInstruction *'>
	<var: #jumpFail4 type: #'AbstractInstruction *'>
	<var: #jumpFailNArgs type: #'AbstractInstruction *'>
	<var: #jumpBCMethod type: #'AbstractInstruction *'>
	<var: #primitiveRoutine declareC: 'void (*primitiveRoutine)(void)'>
	objectRepresentation genLoadSlot: ClosureNumArgsIndex sourceReg: ReceiverResultReg destReg: TempReg.
	self CmpCq: (objectMemory integerObjectOf: methodOrBlockNumArgs) R: TempReg.
	jumpFailNArgs := self JumpNonZero: 0.
	objectRepresentation genLoadSlot: ClosureOuterContextIndex sourceReg: ReceiverResultReg destReg: ClassReg.
	jumpFail1 := objectRepresentation genJumpImmediate: ClassReg.
	objectRepresentation genGetCompactClassIndexNonImmOf: ClassReg into: TempReg.
	objectRepresentation genCmpClassMethodContextCompactIndexR: TempReg.
	jumpFail2 := self JumpNonZero: 0.
	"We defer unforwarding the receiver to the prologue; scanning blocks
	 for inst var refs and only unforwarding if the block refers to inst vars."
	(false
	 and: [objectRepresentation hasSpurMemoryManagerAPI]) ifTrue:
		[objectRepresentation
			genLoadSlot: ReceiverIndex sourceReg: ClassReg destReg: SendNumArgsReg;
			genEnsureOopInRegNotForwarded: SendNumArgsReg
			scratchReg: TempReg
			updatingSlot: ReceiverIndex
			in: ClassReg].
	objectRepresentation genLoadSlot: MethodIndex sourceReg: ClassReg destReg: SendNumArgsReg.
	jumpFail3 := objectRepresentation genJumpImmediate: SendNumArgsReg.
	objectRepresentation genGetFormatOf: SendNumArgsReg into: TempReg.
	self CmpCq: objectMemory firstCompiledMethodFormat R: TempReg.
	jumpFail4 := self JumpLess: 0.
	objectRepresentation genLoadSlot: HeaderIndex sourceReg: SendNumArgsReg destReg: ClassReg.
	jumpBCMethod := objectRepresentation genJumpImmediate: ClassReg.
	self MoveM16: (self offset: CogMethod of: #blockEntryOffset) r: ClassReg R: TempReg.
	self AddR: ClassReg R: TempReg.
	primitiveRoutine := coInterpreter
							functionPointerForCompiledMethod: methodObj
							primitiveIndex: primitiveIndex
							primitivePropertyFlagsInto: nil.
	primitiveRoutine = #primitiveClosureValueNoContextSwitch ifTrue:
		[blockNoContextSwitchOffset = nil ifTrue:
			[^NotFullyInitialized].
		 self SubCq: blockNoContextSwitchOffset R: TempReg].
	self JumpR: TempReg.
	jumpBCMethod jmpTarget: (jumpFail1 jmpTarget: (jumpFail2 jmpTarget: (jumpFail3 jmpTarget: (jumpFail4 jmpTarget: self Label)))).
	(result := self
				compileInterpreterPrimitive: primitiveRoutine
				flags: (coInterpreter primitivePropertyFlags: primitiveIndex)) < 0 ifTrue:
		[^result].
	jumpFailNArgs jmpTarget: self Label.
	^CompletePrimitive! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 7/24/2021 16:45'!
genPrimitiveFullClosureValue
	"Check the argument count.  Fail if wrong.
	 Get the method from the outerContext and see if it is cogged.  If so, jump to the
	 block entry or the no-context-switch entry, as appropriate, and we're done.  If not,
	 invoke the interpreter primitive."
	| jumpFailNArgs jumpFailImmediateMethod jumpFail4 jumpBCMethod primitiveRoutine result |
	<option: #SistaV1BytecodeSet>
	<var: #jumpFailImmediateMethod type: #'AbstractInstruction *'>
	<var: #jumpFail4 type: #'AbstractInstruction *'>
	<var: #jumpFailNArgs type: #'AbstractInstruction *'>
	<var: #jumpBCMethod type: #'AbstractInstruction *'>
	<var: #primitiveRoutine declareC: 'void (*primitiveRoutine)(void)'>
	objectRepresentation genLoadSlot: ClosureNumArgsIndex sourceReg: ReceiverResultReg destReg: TempReg.
	self CmpCq: (objectMemory integerObjectOf: methodOrBlockNumArgs) R: TempReg.
	jumpFailNArgs := self JumpNonZero: 0.

	"We defer unforwarding the receiver to the prologue; scanning blocks
	 for inst var refs and only unforwarding if the block refers to inst vars."
	(false
	 and: [objectRepresentation hasSpurMemoryManagerAPI]) ifTrue:
		[objectRepresentation
			genLoadSlot: FullClosureReceiverIndex sourceReg: ReceiverResultReg destReg: SendNumArgsReg;
			genEnsureOopInRegNotForwarded: SendNumArgsReg
			scratchReg: TempReg
			updatingSlot: FullClosureReceiverIndex
			in: ReceiverResultReg].
	objectRepresentation genLoadSlot: FullClosureCompiledBlockIndex sourceReg: ReceiverResultReg destReg: SendNumArgsReg.
	jumpFailImmediateMethod := objectRepresentation genJumpImmediate: SendNumArgsReg.
	objectRepresentation genGetFormatOf: SendNumArgsReg into: TempReg.
	self CmpCq: objectMemory firstCompiledMethodFormat R: TempReg.
	jumpFail4 := self JumpLess: 0.
	objectRepresentation genLoadSlot: HeaderIndex sourceReg: SendNumArgsReg destReg: ClassReg.
	jumpBCMethod := objectRepresentation genJumpImmediate: ClassReg.

	primitiveRoutine := coInterpreter
							functionPointerForCompiledMethod: methodObj
							primitiveIndex: primitiveIndex
							primitivePropertyFlagsInto: nil.
	self AddCq: (primitiveRoutine = #primitiveFullClosureValueNoContextSwitch
					ifTrue: [self fullBlockNoContextSwitchEntryOffset]
					ifFalse: [self fullBlockEntryOffset])
		 R: ClassReg.
	self JumpR: ClassReg.
	jumpBCMethod jmpTarget: (jumpFailImmediateMethod jmpTarget: (jumpFail4 jmpTarget: self Label)).
	(result := self
				compileInterpreterPrimitive: primitiveRoutine
				 flags: (coInterpreter primitivePropertyFlags: primitiveIndex)) < 0 ifTrue:
		[^result].
	jumpFailNArgs jmpTarget: self Label.
	^CompletePrimitive! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 12/21/2019 19:27'!
genPrimitiveHashMultiply
	"Implement 28-bit hashMultiply for SmallInteger and LargePositiveInteger receivers."
	| jmpFailImm jmpFailNotPositiveLargeInt |
	backEnd canMulRR ifFalse:
		[^UnimplementedPrimitive].

	self mclassIsSmallInteger ifTrue:
		[objectRepresentation genConvertSmallIntegerToIntegerInReg: ReceiverResultReg.
		 self MoveCq: HashMultiplyConstant R: TempReg.
		 self MulR: TempReg R: ReceiverResultReg.
		 self AndCq: HashMultiplyMask R: ReceiverResultReg.
		 objectRepresentation genConvertIntegerToSmallIntegerInReg: ReceiverResultReg.
		 self RetN: 0.
		 ^CompletePrimitive].

	"This check is necessary.  LargeNegativeInteger is a subclass of LargePositiveInteger in Squeak, Pharo, Cuis, et al."
	jmpFailImm := objectRepresentation genJumpImmediate: ReceiverResultReg.
	objectRepresentation genGetCompactClassIndexNonImmOf: ReceiverResultReg into: ClassReg.
	self CmpCq: ClassLargePositiveIntegerCompactIndex R: ClassReg.
	jmpFailNotPositiveLargeInt := self JumpNonZero: 0.
	objectRepresentation genLoadSlot: 0 sourceReg: ReceiverResultReg destReg: ReceiverResultReg.
	self MoveCq: HashMultiplyConstant R: TempReg.
	self MulR: TempReg R: ReceiverResultReg.
	self AndCq: HashMultiplyMask R: ReceiverResultReg.
	objectRepresentation genConvertIntegerToSmallIntegerInReg: ReceiverResultReg.
	self RetN: 0.

	jmpFailImm jmpTarget: (jmpFailNotPositiveLargeInt jmpTarget: self Label).
	^CompletePrimitive! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:47'!
genPrimitivePerform
	"Generate an in-line perform primitive.  The lookup code requires the selector to be in Arg0Reg.
	 adjustArgumentsForPerform: adjusts the arguments once genLookupForPerformNumArgs:
	 has generated the code for the lookup."
	self MoveMw: (backEnd hasLinkRegister
					ifTrue: [methodOrBlockNumArgs - 1]
					ifFalse: [methodOrBlockNumArgs]) * objectMemory wordSize
		r: SPReg
		R: Arg0Reg.
	^self genLookupForPerformNumArgs: methodOrBlockNumArgs! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 5/31/2014 12:57'!
genPushActiveContextBytecode
	self assert: needsFrame.
	objectRepresentation
		genGetActiveContextNumArgs: methodOrBlockNumArgs
		large: (coInterpreter methodNeedsLargeContext: methodObj)
		inBlock: inBlock.
	self PushR: ReceiverResultReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 2/10/2017 14:33'!
genPushClosureCopyCopiedValuesBytecode
	"Block compilation.  At this point in the method create the block.  Note its start
	 and defer generating code for it until after the method and any other preceding
	 blocks.  The block's actual code will be compiled later."
	"143   10001111 llllkkkk jjjjjjjj iiiiiiii	Push Closure Num Copied llll Num Args kkkk BlockSize jjjjjjjjiiiiiiii"
	| startpc numArgs numCopied |
	self assert: needsFrame.
	startpc := bytecodePC + (self generatorAt: byte0) numBytes.
	self addBlockStartAt: startpc "0 relative"
		numArgs: (numArgs := byte1 bitAnd: 16rF)
		numCopied: (numCopied := byte1 >> 4)
		span: (byte2 << 8) + byte3.
	objectRepresentation
		genCreateClosureAt: startpc + 1 "1 relative"
		numArgs: numArgs
		numCopied: numCopied
		contextNumArgs: methodOrBlockNumArgs
		large: (coInterpreter methodNeedsLargeContext: methodObj)
		inBlock: inBlock.
	self PushR: ReceiverResultReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'constant support' stamp: 'cb 4/8/2016 12:00'!
genPushConstant: constant
	"If the objectMemory allows it, generates a quick constant push, else generates a word constant push"
	<inline: true>
	^ (objectRepresentation shouldAnnotateObjectReference: constant)
		ifTrue: [ self annotate: (self PushCw: constant) objRef: constant ]
		ifFalse: [ self PushCq: constant ]
	! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/22/2010 18:07'!
genPushConstantFalseBytecode
	^self genPushLiteral: objectMemory falseObject! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/22/2010 18:07'!
genPushConstantNilBytecode
	^self genPushLiteral: objectMemory nilObject! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/2/2012 11:29'!
genPushConstantOneBytecode
	"79			01001111		Push 1"
	^self genPushLiteral: (objectMemory integerObjectOf: 1)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/22/2010 18:07'!
genPushConstantTrueBytecode
	^self genPushLiteral: objectMemory trueObject! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/2/2012 11:29'!
genPushConstantZeroBytecode
	"78			01001110		Push 0"
	^self genPushLiteral: (objectMemory integerObjectOf: 0)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'rmacnak 11/1/2014 22:53'!
genPushEnclosingObjectAt: level
	"Uncached push enclosing object"
	self MoveCq: level R: SendNumArgsReg.
	self CallRT: ceEnclosingObjectTrampoline.
	self PushR: ReceiverResultReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'cb 4/8/2016 12:01'!
genPushLiteral: literal
	self genPushConstant: literal.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/20/2010 22:25'!
genPushLiteralConstantBytecode
	^self genPushLiteralIndex: (byte0 bitAnd: 31)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/12/2018 09:55'!
genPushLiteralIndex: literalIndex "<SmallInteger>"
	<inline: false>
	| literal |
	literal := self getLiteral: literalIndex.
	BytecodeSetHasDirectedSuperSend ifTrue:
		[self nextDescriptorExtensionsAndNextPCInto:
			[:descriptor :exta :extb :followingPC|
			(self isDirectedSuper: descriptor extA: exta extB: extb) ifTrue:
				[tempOop := literal.
				 ^0]]].
	^self genPushLiteral: literal! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/2/2012 14:12'!
genPushLiteralVariable16CasesBytecode
	"16-31		0001 i i i i		Push Literal Variable #iiii"
	^self genPushLiteralVariable: (byte0 bitAnd: 15)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 9/9/2019 10:13'!
genPushLiteralVariable: literalIndex
	<inline: false>
	| association |
	association := self getLiteral: literalIndex.
	"If followed by a directed super send bytecode, avoid generating any code yet.
	 The association will be passed to the directed send trampoline in a register
	 and fully dereferenced only when first linked.  It will be ignored in later sends."
	BytecodeSetHasDirectedSuperSend ifTrue:
		[self deny: directedSendUsesBinding.
		 self nextDescriptorExtensionsAndNextPCInto:
			[:descriptor :exta :extb :followingPC|
			(self isDirectedSuper: descriptor extA: exta extB: extb) ifTrue:
				[tempOop := association.
				 directedSendUsesBinding := true.
				 ^0]]].
	"N.B. Do _not_ use ReceiverResultReg to avoid overwriting receiver in assignment in frameless methods."
	self genMoveConstant: association R: ClassReg.
	objectRepresentation
		genLoadSlot: ValueIndex
		sourceReg: ClassReg
		destReg: TempReg.
	self PushR: TempReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 1/14/2009 19:57'!
genPushLiteralVariableBytecode
	^self genPushLiteralVariable: (byte0 bitAnd: 31)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/20/2016 14:22'!
genPushMaybeContextReceiverVariable: slotIndex 
	<inline: false>
	| jmpSingle jmpDone |
	<var: #jmpSingle type: #'AbstractInstruction *'>
	<var: #jmpDone type: #'AbstractInstruction *'>
	self assert: needsFrame.
	"See CoInterpreter>>contextInstructionPointer:frame: for an explanation
	 of the instruction pointer slot handling."
	slotIndex = InstructionPointerIndex ifTrue:
		[self putSelfInReceiverResultReg.
		 self MoveCq: slotIndex R: SendNumArgsReg.
		 self CallRT: ceFetchContextInstVarTrampoline.
		 self PushR: SendNumArgsReg.
		 ^0].
	self MoveMw: FoxMFReceiver r: FPReg R: ReceiverResultReg.
	objectRepresentation
		genLoadSlot: SenderIndex
		sourceReg: ReceiverResultReg
		destReg: TempReg.
	jmpSingle := objectRepresentation genJumpNotSmallIntegerInScratchReg: TempReg.
	self MoveCq: slotIndex R: SendNumArgsReg.
	self CallRT: ceFetchContextInstVarTrampoline.
	jmpDone := self Jump: 0.
	jmpSingle jmpTarget: self Label.
	objectRepresentation
		genLoadSlot: slotIndex
		sourceReg: ReceiverResultReg
		destReg: SendNumArgsReg.
	jmpDone jmpTarget: (self PushR: SendNumArgsReg).
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'cb 4/8/2016 08:58'!
genPushMaybeContextRemoteInstVar: slotIndex inObjectAt: objectIndex
	| jmpSingle jmpDone |
	<var: #jmpSingle type: #'AbstractInstruction *'>
	<var: #jmpDone type: #'AbstractInstruction *'>
	self assert: needsFrame.
	"See CoInterpreter>>contextInstructionPointer:frame: for an explanation
	 of the instruction pointer slot handling."
	self MoveMw: (self frameOffsetOfTemporary: objectIndex) r: FPReg R: ReceiverResultReg.
	objectRepresentation 
		genEnsureOopInRegNotForwarded: ReceiverResultReg 
		scratchReg: TempReg.
	slotIndex = InstructionPointerIndex ifTrue:
		[self MoveCq: slotIndex R: SendNumArgsReg.
		 self CallRT: ceFetchContextInstVarTrampoline.
		 self PushR: SendNumArgsReg.
		 ^0].
	objectRepresentation
		genLoadSlot: SenderIndex
		sourceReg: ReceiverResultReg
		destReg: TempReg.
	jmpSingle := objectRepresentation genJumpNotSmallIntegerInScratchReg: TempReg.
	self MoveCq: slotIndex R: SendNumArgsReg.
	self CallRT: ceFetchContextInstVarTrampoline.
	jmpDone := self Jump: 0.
	jmpSingle jmpTarget: self Label.
	objectRepresentation
		genLoadSlot: slotIndex
		sourceReg: ReceiverResultReg
		destReg: SendNumArgsReg.
	jmpDone jmpTarget: (self PushR: SendNumArgsReg).
	^0
! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 5/31/2014 08:52'!
genPushNewArrayBytecode
	| size popValues |
	self assert: needsFrame.
	popValues := byte1 > 127.
	size := byte1 bitAnd: 127.
	objectRepresentation genNewArrayOfSize: size initialized: popValues not.
	popValues ifTrue:
		[size - 1 to: 0 by: -1 do:
			[:i|
			self PopR: TempReg.
			objectRepresentation
				genStoreSourceReg: TempReg
				slotIndex: i
				intoNewObjectInDestReg: ReceiverResultReg]].
	self PushR: ReceiverResultReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/20/2010 22:18'!
genPushQuickIntegerConstantBytecode
	^self genPushLiteral: (objectMemory integerObjectOf: byte0 - 117)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 5/6/2009 10:20'!
genPushReceiverBytecode
	needsFrame
		ifTrue:
			[self MoveMw: FoxMFReceiver r: FPReg R: TempReg.
			 self PushR: TempReg]
		ifFalse:
			[self PushR: ReceiverResultReg].
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/20/2016 14:22'!
genPushReceiverVariable: index
	<inline: false>
	| maybeErr |
	needsFrame ifTrue:
		[self putSelfInReceiverResultReg].
	maybeErr := objectRepresentation genLoadSlot: index sourceReg: ReceiverResultReg destReg: TempReg.
	maybeErr < 0 ifTrue:
		[^maybeErr].
	self PushR: TempReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 1/7/2009 21:47'!
genPushReceiverVariableBytecode
	^self genPushReceiverVariable: (byte0 bitAnd: 15)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'cb 4/8/2016 08:58'!
genPushRemoteInstVar: index inObjectAt: objectIndex
	self MoveMw: (self frameOffsetOfTemporary: objectIndex) r: FPReg R: ClassReg.
	objectRepresentation 
		genEnsureOopInRegNotForwarded: ClassReg 
		scratchReg: TempReg.
	objectRepresentation
		genLoadSlot: index
		sourceReg: ClassReg
		destReg: TempReg.
	self PushR: TempReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 5/29/2018 08:43'!
genPushRemoteTempLongBytecode
	self MoveMw: (self frameOffsetOfTemporary: byte2) r: FPReg R: ClassReg.
	TempVectReadBarrier
		ifTrue: [objectRepresentation
				genEnsureObjInRegNotForwarded: ClassReg
				scratchReg: TempReg].
	objectRepresentation
		genLoadSlot: byte1
		sourceReg: ClassReg
		destReg: TempReg.
	self PushR: TempReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 5/6/2009 10:20'!
genPushTemporaryVariable: index
	self MoveMw: (self frameOffsetOfTemporary: index) r: FPReg R: TempReg.
	self PushR: TempReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 1/2/2009 14:18'!
genPushTemporaryVariableBytecode
	^self genPushTemporaryVariable: (byte0 bitAnd: 15)! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'cb 4/8/2016 12:11'!
genQuickReturnConst
	<api> "because selected by CoInterpreter>>quickPrimitiveGeneratorFor:"
	| constant |
	constant := coInterpreter quickPrimitiveConstantFor: primitiveIndex.
	self genMoveConstant: constant R: ReceiverResultReg.
	self genUpArrowReturn.
	^UnfailingPrimitive! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 09:44'!
genQuickReturnInstVar
	<api> "because selected by CoInterpreter>>quickPrimitiveGeneratorFor:"
	| index |
	index := coInterpreter quickPrimitiveInstVarIndexFor: primitiveIndex.
	objectRepresentation genLoadSlot: index sourceReg: ReceiverResultReg destReg: ReceiverResultReg.
	self genUpArrowReturn.
	^UnfailingPrimitive! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 09:43'!
genQuickReturnSelf
	<api> "because selected by CoInterpreter>>quickPrimitiveGeneratorFor:"
	self genUpArrowReturn.
	^UnfailingPrimitive! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 4/25/2015 15:08'!
genReturnFalse
	self genMoveFalseR: ReceiverResultReg.
	^self genUpArrowReturn! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 4/8/2016 12:08'!
genReturnNil
	self genMoveNilR: ReceiverResultReg.
	^self genUpArrowReturn! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 9/5/2016 14:12'!
genReturnNilFromBlock
	self assert: inBlock > 0.
	self genMoveNilR: ReceiverResultReg.
	^self genBlockReturn! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 1/20/2016 14:22'!
genReturnReceiver
	"Frameless method activation looks like
				receiver
				args
		sp->	ret pc.
	 Return pops receiver and arguments off the stack.  Callee pushes the result."
	needsFrame ifTrue:
		[self putSelfInReceiverResultReg].
	^self genUpArrowReturn! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 9/5/2016 14:12'!
genReturnTopFromBlock
	self assert: inBlock > 0.
	self PopR: ReceiverResultReg.
	^self genBlockReturn! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 4/30/2015 12:06'!
genReturnTopFromMethod
	"Return pops receiver and arguments off the stack.  Callee pushes the result."
	self PopR: ReceiverResultReg.
	^self genUpArrowReturn! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 4/25/2015 15:10'!
genReturnTrue
	self genMoveTrueR: ReceiverResultReg.
	^self genUpArrowReturn! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 10/6/2015 15:02'!
genSecondExtendedSendBytecode
	"Can use any of the first 64 literals for the selector and pass up to 3 arguments."

	^self genSend: (byte1 bitAnd: 16r3F) numArgs: byte1 >> 6! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 10/6/2015 14:53'!
genSend: selectorIndex numArgs: numArgs
	<inline: true>
	^self genSend: selectorIndex numArgs: numArgs sendTable: ordinarySendTrampolines! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/12/2018 10:32'!
genSend: selectorIndex numArgs: numArgs sendTable: sendTable
	<inline: false>
	<var: #sendTable type: #'sqInt *'>
	| annotation |
	self assert: needsFrame.
	annotation := self annotationForSendTable: sendTable.
	self assert: (numArgs between: 0 and: 255). "say"
	self MoveMw: numArgs * objectMemory wordSize r: SPReg R: ReceiverResultReg.
	"Deal with stale super sends; see SpurMemoryManager's class comment."
	(self annotationIsForUncheckedEntryPoint: annotation) ifTrue:
		[objectRepresentation genEnsureOopInRegNotForwarded: ReceiverResultReg scratchReg: TempReg].
	"0 through (NumSendTrampolines - 2) numArgs sends have the arg count implciti in the trampoline.
	 The last send trampoline (NumSendTrampolines - 1) passes numArgs in SendNumArgsReg."
	numArgs >= (NumSendTrampolines - 1) ifTrue:
		[self MoveCq: numArgs R: SendNumArgsReg].
	(BytecodeSetHasDirectedSuperSend
	 and: [annotation
			between: IsDirectedSuperSend
			and: IsDirectedSuperBindingSend]) ifTrue:
		[self genMoveConstant: tempOop R: TempReg].
	self genLoadInlineCacheWithSelector: selectorIndex.
	(self Call: (sendTable at: (numArgs min: NumSendTrampolines - 1))) annotation: annotation.
	self PushR: ReceiverResultReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 10/6/2015 14:59'!
genSendAbsentDynamicSuper: selectorIndex numArgs: numArgs
	<inline: false>
	^self
		genNSSend: selectorIndex
		numArgs: numArgs
		depth: LookupRuleDynamicSuper
		sendTable: dynamicSuperSendTrampolines! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 10/6/2015 15:02'!
genSendAbsentImplicit0ArgsBytecode
	"160-175	1010 i i i i		Send To Absent Implicit Receiver Literal Selector #iiii With 0 Arguments."
	^self genSendAbsentImplicit: (byte0 bitAnd: 15) numArgs: 0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 10/6/2015 14:59'!
genSendAbsentImplicit: selectorIndex numArgs: numArgs
	<inline: false>
	^self
		genNSSend: selectorIndex
		numArgs: numArgs
		depth: LookupRuleImplicit
		sendTable: implicitReceiverSendTrampolines! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 10/6/2015 14:59'!
genSendAbsentOuter: selectorIndex numArgs: numArgs depth: depth
	<inline: false>
	^self
		genNSSend: selectorIndex
		numArgs: numArgs
		depth: depth
		sendTable: outerSendTrampolines! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 10/6/2015 14:59'!
genSendAbsentSelf: selectorIndex numArgs: numArgs
	<inline: false>
	^self
		genNSSend: selectorIndex
		numArgs: numArgs
		depth: LookupRuleSelf
		sendTable: selfSendTrampolines! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'nice 11/8/2019 10:31'!
genSendDirectedSuper: selectorIndex numArgs: numArgs
	<inline: false>
	<option: #BytecodeSetHasDirectedSuperSend>
	"N.B. genPushLiteralVariableGivenDirectedSuper: has already loaded tempOop with the association."
	
	| result |
	result := self
				genSend: selectorIndex
				numArgs: numArgs
				sendTable: (directedSendUsesBinding
								ifTrue: [directedSuperBindingSendTrampolines]
								ifFalse: [directedSuperSendTrampolines]).
	directedSendUsesBinding := false.
	^result! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 10/6/2015 15:02'!
genSendLiteralSelector0ArgsBytecode
	^self genSend: (byte0 bitAnd: 15) numArgs: 0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 10/6/2015 15:02'!
genSendLiteralSelector1ArgBytecode
	^self genSend: (byte0 bitAnd: 15) numArgs: 1! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 10/6/2015 15:02'!
genSendLiteralSelector2ArgsBytecode
	^self genSend: (byte0 bitAnd: 15) numArgs: 2! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 10/23/2015 19:53:21'!
genSendSuper: selectorIndex numArgs: numArgs
	<inline: false>
	^self genSend: selectorIndex numArgs: numArgs sendTable: superSendTrampolines! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/2/2012 10:57'!
genShortJumpIfFalse
	| distance target |
	distance := self v3: (self generatorAt: byte0)
					ShortForward: bytecodePC
					Branch: 0
					Distance: methodObj.
	target := distance + 1 + bytecodePC.
	^self genJumpIf: objectMemory falseObject to: target! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/2/2012 10:57'!
genShortJumpIfTrue
	| distance target |
	distance := self v3: (self generatorAt: byte0)
					ShortForward: bytecodePC
					Branch: 0
					Distance: methodObj.
	target := distance + 1 + bytecodePC.
	^self genJumpIf: objectMemory trueObject to: target! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 11/2/2012 10:57'!
genShortUnconditionalJump
	| distance target |
	distance := self v3: (self generatorAt: byte0)
					ShortForward: bytecodePC
					Branch: 0
					Distance: methodObj.
	target := distance + 1 + bytecodePC.
	^self genJumpTo: target! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 4/11/2016 11:10'!
genSistaExtStoreAndPopLiteralVariableBytecode
	^ self genSistaExtStoreLiteralVariableBytecodePopBoolean: true! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 4/11/2016 11:09'!
genSistaExtStoreAndPopReceiverVariableBytecode
	^ self genSistaExtStoreAndPopReceiverVariableBytecodePopBoolean: true! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 8/31/2021 14:34'!
genSistaExtStoreAndPopReceiverVariableBytecodePopBoolean: boolean
	| index |
	extB := 0. "Simple cogit don't use the extra flags"
	numExtB := 0.
	index := byte1 + (extA << 8).
	extA := 0.
	^(self mclassCouldBeContext
	   and: [coInterpreter isWriteMediatedContextInstVarIndex: index])
		ifTrue: [self genStorePop: boolean MaybeContextReceiverVariable: index ]
		ifFalse: [self genStorePop: boolean ReceiverVariable: index ]! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 4/11/2016 11:10'!
genSistaExtStoreLiteralVariableBytecode
	^ self genSistaExtStoreLiteralVariableBytecodePopBoolean: false! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:04'!
genSistaExtStoreLiteralVariableBytecodePopBoolean: boolean
	| index |
	extB := 0. "SimpleCogit don't use the extra flags"
	numExtB := 0.
	index := byte1 + (extA << 8).
	extA := 0.
	^ self genStorePop: boolean LiteralVariable: index! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 4/11/2016 11:09'!
genSistaExtStoreReceiverVariableBytecode
	^ self genSistaExtStoreAndPopReceiverVariableBytecodePopBoolean: false! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 9/11/2019 07:50'!
genSpecialSelectorClass
	self MoveMw: 0 r: SPReg R: SendNumArgsReg.
	objectRepresentation
		genGetClassObjectOf: SendNumArgsReg
		into: ClassReg
		scratchReg: TempReg
		mayBeAForwarder: true.
	self MoveR: ClassReg Mw: 0 r: SPReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 11/24/2016 10:44'!
genSpecialSelectorEqualsEquals
	^ self genInlinedIdenticalOrNotIf: false! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'cb 11/24/2016 10:45'!
genSpecialSelectorNotEqualsEquals
	^ self genInlinedIdenticalOrNotIf: true! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 10/6/2015 15:09'!
genSpecialSelectorSend
	| index numArgs |
	index := byte0 - self firstSpecialSelectorBytecodeOffset.
	numArgs := coInterpreter specialSelectorNumArgs: index.
	^self genSend: index negated - 1 numArgs: numArgs! !

!SimpleStackBasedCogit methodsFor: 'simulation stack' stamp: 'cb 10/20/2017 16:36'!
genStackArgAt: n into: reg 
	<inline: true>
	self
		MoveMw: (self backEnd hasLinkRegister
				ifTrue: [n]
				ifFalse: [n + 1])
				* objectMemory wordSize
		r: SPReg
		R: reg.
	^ 0
! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 1/9/2009 11:44'!
genStoreAndPopReceiverVariableBytecode
	^self genStorePop: true ReceiverVariable: (byte0 bitAnd: 7)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 1/16/2009 18:09'!
genStoreAndPopRemoteTempLongBytecode
	^self genStorePop: true RemoteTemp: byte1 At: byte2! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 1/8/2009 14:20'!
genStoreAndPopTemporaryVariableBytecode
	^self genStorePop: true TemporaryVariable: (byte0 bitAnd: 7)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 9/9/2019 10:14'!
genStorePop: popBoolean LiteralVariable: litVarIndex
	<inline: false>
	| association |
	"The only reason we assert needsFrame here is that in a frameless method
	 ReceiverResultReg must and does contain only self, but the ceStoreCheck
	 trampoline expects the target of the store to be in ReceiverResultReg.  So
	 in a frameless method we would have a conflict between the receiver and
	 the literal store, unless we we smart enough to realise that ReceiverResultReg
	 was unused after the literal variable store, unlikely given that methods
	 return self by default."
	self assert: needsFrame.
	association := self getLiteral: litVarIndex.
	self genMoveConstant: association R: ReceiverResultReg.
	popBoolean
		ifTrue: [self PopR: ClassReg]
		ifFalse: [self MoveMw: 0 r: SPReg R: ClassReg].
	self
		genStoreSourceReg: ClassReg 
		slotIndex: ValueIndex 
		destReg: ReceiverResultReg 
		scratchReg: TempReg 
		inFrame: needsFrame.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'cb 3/30/2016 15:16'!
genStorePop: popBoolean MaybeContextReceiverVariable: slotIndex
	<inline: false>
	| jmpSingle jmpDone |
	<var: #jmpSingle type: #'AbstractInstruction *'>
	<var: #jmpDone type: #'AbstractInstruction *'>
	"The reason we need a frame here is that assigning to an inst var of a context may
	 involve wholesale reorganization of stack pages, and the only way to preserve the
	 execution state of an activation in that case is if it has a frame."
	self assert: needsFrame.
	self putSelfInReceiverResultReg.
	objectRepresentation
		genLoadSlot: SenderIndex
		sourceReg: ReceiverResultReg
		destReg: TempReg.
	self MoveMw: 0 r: SPReg R: ClassReg.
	jmpSingle := objectRepresentation genJumpNotSmallIntegerInScratchReg: TempReg.
	self MoveCq: slotIndex R: SendNumArgsReg.
	self CallRT: ceStoreContextInstVarTrampoline.
	jmpDone := self Jump: 0.
	jmpSingle jmpTarget: self Label.
	popBoolean ifTrue:
		[self AddCq: objectMemory wordSize R: SPReg].
	self
		genStoreSourceReg: ClassReg 
		slotIndex: slotIndex 
		destReg: ReceiverResultReg 
		scratchReg: TempReg 
		inFrame: needsFrame.
	jmpDone jmpTarget: self Label.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'cb 4/8/2016 09:02'!
genStorePop: popBoolean MaybeContextRemoteInstVar: slotIndex ofObjectAt: objectIndex
	<inline: false>
	| jmpSingle jmpDone |
	<var: #jmpSingle type: #'AbstractInstruction *'>
	<var: #jmpDone type: #'AbstractInstruction *'>
	"The reason we need a frame here is that assigning to an inst var of a context may
	 involve wholesale reorganization of stack pages, and the only way to preserve the
	 execution state of an activation in that case is if it has a frame."
	self assert: needsFrame.
	self MoveMw: (self frameOffsetOfTemporary: objectIndex) r: FPReg R: ReceiverResultReg.
	objectRepresentation 
		genEnsureOopInRegNotForwarded: ReceiverResultReg 
		scratchReg: TempReg.
	objectRepresentation
		genLoadSlot: SenderIndex
		sourceReg: ReceiverResultReg
		destReg: TempReg.
	self MoveMw: 0 r: SPReg R: ClassReg.
	jmpSingle := objectRepresentation genJumpNotSmallIntegerInScratchReg: TempReg.
	self MoveCq: slotIndex R: SendNumArgsReg.
	self CallRT: ceStoreContextInstVarTrampoline.
	jmpDone := self Jump: 0.
	jmpSingle jmpTarget: self Label.
	popBoolean ifTrue:
		[self AddCq: objectMemory wordSize R: SPReg].
	self
		genStoreSourceReg: ClassReg 
		slotIndex: slotIndex 
		destReg: ReceiverResultReg 
		scratchReg: TempReg 
		inFrame: needsFrame.
	jmpDone jmpTarget: self Label.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'cb 3/30/2016 15:16'!
genStorePop: popBoolean ReceiverVariable: slotIndex
	<inline: false>
	needsFrame ifTrue:
		[self putSelfInReceiverResultReg].
	popBoolean
		ifTrue: [self PopR: ClassReg]
		ifFalse: [self MoveMw: 0 r: SPReg R: ClassReg].
	self
		genStoreSourceReg: ClassReg 
		slotIndex: slotIndex 
		destReg: ReceiverResultReg 
		scratchReg: TempReg 
		inFrame: needsFrame.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'cb 4/8/2016 08:57'!
genStorePop: popBoolean RemoteInstVar: slotIndex ofObjectAt: objectIndex
	<inline: false>
	"The only reason we assert needsFrame here is that in a frameless method
	 ReceiverResultReg must and does contain only self, but the ceStoreCheck
	 trampoline expects the target of the store to be in ReceiverResultReg.  So
	 in a frameless method we would have a conflict between the receiver and
	 the temote temp store, unless we we smart enough to realise that
	 ReceiverResultReg was unused after the literal variable store, unlikely given
	 that methods return self by default."
	self assert: needsFrame.
	popBoolean
		ifTrue: [self PopR: ClassReg]
		ifFalse: [self MoveMw: 0 r: SPReg R: ClassReg].
	self MoveMw: (self frameOffsetOfTemporary: objectIndex) r: FPReg R: ReceiverResultReg.
	objectRepresentation 
		genEnsureOopInRegNotForwarded: ReceiverResultReg 
		scratchReg: TempReg.
	^objectRepresentation
		genStoreSourceReg: ClassReg
		slotIndex: slotIndex
		destReg: ReceiverResultReg
		scratchReg: TempReg
		inFrame: needsFrame! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'cb 5/29/2018 08:43'!
genStorePop: popBoolean RemoteTemp: slotIndex At: remoteTempIndex
	<inline: false>
	"The only reason we assert needsFrame here is that in a frameless method
	 ReceiverResultReg must and does contain only self, but the ceStoreCheck
	 trampoline expects the target of the store to be in ReceiverResultReg.  So
	 in a frameless method we would have a conflict between the receiver and
	 the temote temp store, unless we we smart enough to realise that
	 ReceiverResultReg was unused after the literal variable store, unlikely given
	 that methods return self by default."
	self assert: needsFrame.
	popBoolean
		ifTrue: [self PopR: ClassReg]
		ifFalse: [self MoveMw: 0 r: SPReg R: ClassReg].
	self MoveMw: (self frameOffsetOfTemporary: remoteTempIndex) r: FPReg R: ReceiverResultReg.
	TempVectReadBarrier
		ifTrue: [objectRepresentation
				genEnsureObjInRegNotForwarded: ReceiverResultReg
				scratchReg: TempReg].
	^objectRepresentation
		genStoreSourceReg: ClassReg
		slotIndex: slotIndex
		destReg: ReceiverResultReg
		scratchReg: TempReg
		inFrame: needsFrame! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 5/6/2009 10:20'!
genStorePop: popBoolean TemporaryVariable: tempIndex
	<inline: false>
	popBoolean
		ifTrue: [self PopR: TempReg]
		ifFalse: [self MoveMw: 0 r: SPReg R: TempReg].
	self MoveR: TempReg
		Mw: (self frameOffsetOfTemporary: tempIndex)
		r: FPReg.
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 1/16/2009 18:09'!
genStoreRemoteTempLongBytecode
	^self genStorePop: false RemoteTemp: byte1 At: byte2! !

!SimpleStackBasedCogit methodsFor: 'constant support' stamp: 'eem 12/6/2016 14:03'!
genSubConstant: constant R: reg
	"If the objectMemory allows it, generates a quick constant sub, else generates a word constant sub"
	<inline: true>
	^ (objectRepresentation shouldAnnotateObjectReference: constant)
		ifTrue: [ self annotate: (self SubCw: constant R: reg) objRef: reg. ]
		ifFalse: [ self SubCq: constant R: reg ]! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 8/5/2021 15:57'!
genTakeProfileSample
	methodLabel addDependent:
		(self annotateAbsolutePCRef:
			(self MoveCw: methodLabel asInteger R: ClassReg)).
	backEnd genMarshallNArgs: 1 arg: ClassReg arg: nil arg: nil arg: nil.
	SPReg ~= NativeSPReg ifTrue:
				[backEnd genLoadNativeSPRegWithAlignedSPReg].
	self CallFullRT: (self cCode: [#ceTakeProfileSample: asUnsignedIntegerPtr]
						   inSmalltalk: [self simulatedTrampolineFor: #ceTakeProfileSample:]).
	backEnd genRemoveNArgsFromStack: 1.
	"Remember to restore the native stack pointer to point to the C stack,
	 otherwise the Smalltalk frames will get overwritten on an interrupt."
	SPReg ~= NativeSPReg ifTrue:
		[backEnd genLoadCStackPointer]! !

!SimpleStackBasedCogit methodsFor: 'initialization' stamp: 'eem 5/30/2016 13:57'!
genTraceStoreTrampoline
	ceTraceStoreTrampoline := self genTrampolineFor: #ceTraceStoreOf:into:
										called: 'ceTraceStoreTrampoline'
										arg: ClassReg
										arg: ReceiverResultReg
										regsToSave: CallerSavedRegisterMask! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'cb 3/30/2016 15:07'!
genTraceStores
	<inline: true>
	traceStores > 0 ifTrue: [ self CallRT: ceTraceStoreTrampoline ].! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 4/7/2015 10:42'!
genUnconditionalTrapBytecode
	"SistaV1: *	217		Trap"
	^EncounteredUnknownBytecode! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 9/5/2016 14:12'!
genUpArrowReturn
	"Generate a method return from within a method or a block.
	 Frameless method activation looks like
				receiver
				args
		sp->	ret pc.
	 Return pops receiver and arguments off the stack.  Callee pushes the result."
	inBlock > 0 ifTrue:
		[self assert: needsFrame.
		 self CallRT: ceNonLocalReturnTrampoline.
		 self annotateBytecode: self Label.
		 ^0].
	needsFrame ifTrue:
		[self MoveR: FPReg R: SPReg.
		 self PopR: FPReg.
		 backEnd hasLinkRegister ifTrue:
			[self PopR: LinkReg]].
	self RetN: methodOrBlockNumArgs + 1 * objectMemory wordSize.
	^0! !

!SimpleStackBasedCogit methodsFor: 'initialization' stamp: 'eem 12/7/2017 17:01'!
generateMissAbortTrampolines
	"Generate the run-time entries for the various method and PIC entry misses and aborts.
	 Read the class-side method trampolines for documentation on the various trampolines"

	ceMethodAbortTrampoline := self genMethodAbortTrampoline.
	cePICAbortTrampoline := self genPICAbortTrampoline.
	ceCPICMissTrampoline := self genTrampolineFor: #ceCPICMiss:receiver:
								called: 'ceCPICMissTrampoline'
								arg: ClassReg
								arg: ReceiverResultReg.
	ceReapAndResetErrorCodeTrampoline := self genTrampolineFor: #ceReapAndResetErrorCodeFor:
												called: 'ceReapAndResetErrorCodeTrampoline'
												arg: ClassReg! !

!SimpleStackBasedCogit methodsFor: 'initialization' stamp: 'eem 4/28/2020 13:36'!
generateTracingTrampolines
	"Generate trampolines for tracing.  In the simulator we can save a lot of time
	 and avoid noise instructions in the lastNInstructions log by short-cutting these
	 trampolines, but we need them in the real vm."
	ceTraceLinkedSendTrampoline :=
		self genTrampolineFor: #ceTraceLinkedSend:
			called: 'ceTraceLinkedSendTrampoline'
			arg: ReceiverResultReg
			regsToSave: CallerSavedRegisterMask.
	ceTraceBlockActivationTrampoline :=
		self genTrampolineFor: #ceTraceBlockActivation
			called: 'ceTraceBlockActivationTrampoline'
			regsToSave: CallerSavedRegisterMask.
	ceTraceStoreTrampoline :=
		self genTrampolineFor: #ceTraceStoreOf:into:
			called: 'ceTraceStoreTrampoline'
			arg: ClassReg
			arg: ReceiverResultReg
			regsToSave: CallerSavedRegisterMask.
	self cCode: [] inSmalltalk:
		[ceTraceLinkedSendTrampoline := self simulatedTrampolineFor: #ceShortCutTraceLinkedSend:.
		 ceTraceBlockActivationTrampoline := self simulatedTrampolineFor: #ceShortCutTraceBlockActivation:.
		 ceTraceStoreTrampoline := self simulatedTrampolineFor: #ceShortCutTraceStore:]! !

!SimpleStackBasedCogit methodsFor: 'register management' stamp: 'eem 7/24/2021 20:28'!
isCalleeSavedReg: reg
	<inline: true>
	^self register: reg isInMask: ABICalleeSavedRegisterMask! !

!SimpleStackBasedCogit methodsFor: 'register management' stamp: 'eem 5/30/2016 13:56'!
isCallerSavedReg: reg
	<inline: true>
	^self register: reg isInMask: CallerSavedRegisterMask! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'rsf 9/5/2016 16:19'!
loadNativeArgumentAddress: baseOffset to: reg
	<option: #LowcodeVM>
	self MoveMw: self frameOffsetOfPreviousNativeStackPointer r: FPReg R: reg.
	self AddCq: baseOffset - 1 R: reg.! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'rsf 8/10/2016 00:24'!
loadNativeFramePointerInto: reg
	<option: #LowcodeVM>
	self MoveMw: self frameOffsetOfNativeFramePointer r: FPReg R: reg! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'rsf 8/10/2016 00:24'!
loadNativeLocalAddress: baseOffset to: reg
	<option: #LowcodeVM>
	self MoveMw: self frameOffsetOfNativeFramePointer r: FPReg R: reg.
	self AddCq: baseOffset - 1 R: reg.! !

!SimpleStackBasedCogit methodsFor: 'method introspection' stamp: 'eem 4/19/2016 17:08'!
mapPCDataFor: cogMethod into: arrayObj
	"Collect the branch and send data for cogMethod, storing it into arrayObj."
	<api>
	<var: #cogMethod type: #'CogMethod *'>
	| errCode |
	introspectionDataIndex := 0.
	introspectionData := arrayObj.
	cogMethod stackCheckOffset = 0 ifTrue:
		[self assert: introspectionDataIndex = 0.
		 cogMethod cmIsFullBlock
			ifTrue:
				[objectMemory
					storePointerUnchecked: 0 ofObject: introspectionData withValue: objectMemory nilObject;
					storePointerUnchecked: 1 ofObject: introspectionData withValue: (objectMemory integerObjectOf: cbNoSwitchEntryOffset);
					storePointerUnchecked: 2 ofObject: introspectionData withValue: objectMemory nilObject;
					storePointerUnchecked: 3 ofObject: introspectionData withValue: (objectMemory integerObjectOf: cbEntryOffset)]
			ifFalse:
				[objectMemory
					storePointerUnchecked: 0 ofObject: introspectionData withValue: objectMemory nilObject;
					storePointerUnchecked: 1 ofObject: introspectionData withValue: (objectMemory integerObjectOf: cmEntryOffset);
					storePointerUnchecked: 2 ofObject: introspectionData withValue: objectMemory nilObject;
					storePointerUnchecked: 3 ofObject: introspectionData withValue: (objectMemory integerObjectOf: cmNoCheckEntryOffset)].
		 ^4].
	errCode := self
					mapFor: (self cCoerceSimple: cogMethod to: #'CogBlockMethod *')
					bcpc: (coInterpreter startPCOfMethod: cogMethod methodObject)
					performUntil: #pcDataFor:Annotation:Mcpc:Bcpc:Method:
					arg: cogMethod asVoidPointer.
	errCode ~= 0 ifTrue:
		[self assert: errCode = PrimErrNoMemory.
		 ^-1].
	cogMethod blockEntryOffset ~= 0 ifTrue:
		[errCode := self blockDispatchTargetsFor: cogMethod
						perform: #pcDataForBlockEntry:Method:
						arg: cogMethod asInteger.
		 errCode ~= 0 ifTrue:
			[self assert: errCode = PrimErrNoMemory.
			 ^-1]].
	^introspectionDataIndex! !

!SimpleStackBasedCogit methodsFor: 'bytecode generators' stamp: 'eem 1/20/2016 14:23'!
marshallAbsentReceiverSendArguments: numArgs
	self assert: needsFrame.
	self putSelfInReceiverResultReg.

	"Shuffle arguments if necessary and push receiver."
	numArgs = 0
		ifTrue:
			[self PushR: ReceiverResultReg]
		ifFalse:
			[self MoveMw: 0 r: SPReg R: TempReg.
			self PushR: TempReg.
			2 to: numArgs do:
				[:index|
				self MoveMw: index * objectMemory wordSize r: SPReg R: TempReg.
				self MoveR: TempReg Mw: index - 1 * BytesPerWord r: SPReg].
			self MoveR: ReceiverResultReg Mw: numArgs * BytesPerWord r: SPReg].! !

!SimpleStackBasedCogit methodsFor: 'trampolines' stamp: 'eem 12/3/2010 16:52'!
methodAbortTrampolineFor: numArgs
	^ceMethodAbortTrampoline! !

!SimpleStackBasedCogit methodsFor: 'compile abstract instructions' stamp: 'eem 7/25/2018 15:58'!
methodUsesPrimitiveErrorCode: aMethodObj header: aMethodHeader
	"Answer if aMethodObj contains a primitive and uses the primitive error code."
	<inline: true>
	^(coInterpreter primitiveIndexOfMethod: aMethodObj header: aMethodHeader) > 0
	  and: [(coInterpreter longStoreBytecodeForHeader: aMethodHeader)
			= (objectMemory
				fetchByte: (coInterpreter startPCOfMethod: aMethodObj) + (coInterpreter sizeOfCallPrimitiveBytecode: aMethodHeader)
				ofObject: aMethodObj)]! !

!SimpleStackBasedCogit methodsFor: 'testing' stamp: 'eem 9/5/2021 07:55'!
numRegArgs
	<api>
	<cmacro: '() 0'> "to allow inlining optimization across the CoInterpreter/Cogit api..."
	^0! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 10/6/2015 15:52'!
numSpecialSelectors
	<inline: true>
	^self cppIf: MULTIPLEBYTECODESETS
		ifTrue: [bytecodeSetOffset = 256 ifTrue: [AltNumSpecialSelectors] ifFalse: [NumSpecialSelectors]]
		ifFalse: [NumSpecialSelectors]! !

!SimpleStackBasedCogit methodsFor: 'span functions' stamp: 'eem 11/13/2012 14:46'!
parseV4Exts: nExts priorTo: bcpc in: aMethodObj into: aBinaryBlock
	"224		11100000	aaaaaaaa	Extend A (Ext A = Ext A prev * 256 + Ext A)
	 225		11100001	sbbbbbbb	Extend B (Ext B = Ext B prev * 256 + Ext B)"
	| extAValue extBValue pc byte extByte |
	<inline: true>
	extAValue := extBValue := 0.
	pc := bcpc - nExts - nExts.
	[pc < bcpc] whileTrue:
		[byte := objectMemory fetchByte: pc ofObject: aMethodObj.
		 pc := pc + 1.
		 extByte := objectMemory fetchByte: pc ofObject: aMethodObj.
		 pc := pc + 1.
		 self assert: (byte = 224 or: [byte = 225]).
		 byte = 224
			ifTrue:
				[extAValue := (extAValue bitShift: 8) + extByte]
			ifFalse:
				[extBValue := (extBValue = 0 and: [extByte > 127])
							ifTrue: [extByte - 256]
							ifFalse: [(extBValue bitShift: 8) + extByte]]].
	aBinaryBlock value: extAValue value: extBValue! !

!SimpleStackBasedCogit methodsFor: 'method introspection' stamp: 'eem 4/19/2016 17:06'!
pcDataFor: descriptor Annotation: isBackwardBranchAndAnnotation Mcpc: mcpc Bcpc: bcpc Method: cogMethodArg
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<var: #mcpc type: #'char *'>
	<var: #cogMethodArg type: #'void *'>

	descriptor ifNil: "this is the stackCheck offset"
		[self assert: introspectionDataIndex = 0.
		 (self cCoerceSimple: cogMethodArg to: #'CogMethod *') cmIsFullBlock
			ifTrue:
				[objectMemory
					storePointerUnchecked: introspectionDataIndex + 0 ofObject: introspectionData withValue: objectMemory nilObject;
					storePointerUnchecked: introspectionDataIndex + 1 ofObject: introspectionData withValue: (objectMemory integerObjectOf: cbNoSwitchEntryOffset);
					storePointerUnchecked: introspectionDataIndex + 2 ofObject: introspectionData withValue: objectMemory nilObject;
					storePointerUnchecked: introspectionDataIndex + 3 ofObject: introspectionData withValue: (objectMemory integerObjectOf: cbEntryOffset)]
			ifFalse:
				[objectMemory
					storePointerUnchecked: introspectionDataIndex + 0 ofObject: introspectionData withValue: objectMemory nilObject;
					storePointerUnchecked: introspectionDataIndex + 1 ofObject: introspectionData withValue: (objectMemory integerObjectOf: cmEntryOffset);
					storePointerUnchecked: introspectionDataIndex + 2 ofObject: introspectionData withValue: objectMemory nilObject;
					storePointerUnchecked: introspectionDataIndex + 3 ofObject: introspectionData withValue: (objectMemory integerObjectOf: cmNoCheckEntryOffset)].
		 objectMemory
			storePointerUnchecked: introspectionDataIndex + 4 ofObject: introspectionData withValue: (objectMemory integerObjectOf: bcpc + 1);
			storePointerUnchecked: introspectionDataIndex + 5 ofObject: introspectionData withValue: (objectMemory integerObjectOf: (self cCoerceSimple: cogMethodArg to: #'CogMethod *') stackCheckOffset).
		 introspectionDataIndex := introspectionDataIndex + 6.
		 ^0].

	(self isPCMappedAnnotation: isBackwardBranchAndAnnotation >> 1) ifTrue:
		[| actualBcpc actualMcpc |
		 actualBcpc := (isBackwardBranchAndAnnotation anyMask: 1)
							ifTrue: [bcpc + 1]
							ifFalse: [bcpc + descriptor numBytes + 1].
		 actualMcpc := mcpc asUnsignedInteger - cogMethodArg asUnsignedInteger.
		 objectMemory
			storePointerUnchecked: introspectionDataIndex + 0 ofObject: introspectionData withValue: (objectMemory integerObjectOf: actualBcpc);
			storePointerUnchecked: introspectionDataIndex + 1 ofObject: introspectionData withValue: (objectMemory integerObjectOf: actualMcpc).
		 introspectionDataIndex := introspectionDataIndex + 2].

	^0! !

!SimpleStackBasedCogit methodsFor: 'method introspection' stamp: 'eem 4/19/2016 16:19'!
pcDataForBlockEntry: blockEntryMcpc Method: cogMethod
	"Collect the branch and send data for the block method starting at blockEntryMcpc, storing it into picData."
	<returnTypeC: #usqInt>
	objectMemory
		storePointerUnchecked: introspectionDataIndex + 0 ofObject: introspectionData withValue: objectMemory nilObject;
		storePointerUnchecked: introspectionDataIndex + 1 ofObject: introspectionData withValue: (objectMemory integerObjectOf: blockEntryMcpc - blockNoContextSwitchOffset);
		storePointerUnchecked: introspectionDataIndex + 2 ofObject: introspectionData withValue: objectMemory nilObject;
		storePointerUnchecked: introspectionDataIndex + 3 ofObject: introspectionData withValue: (objectMemory integerObjectOf: blockEntryMcpc).
	introspectionDataIndex := introspectionDataIndex + 4.
	^0! !

!SimpleStackBasedCogit methodsFor: 'trampolines' stamp: 'eem 12/3/2010 17:02'!
picAbortTrampolineFor: numArgs
	^cePICAbortTrampoline! !

!SimpleStackBasedCogit methodsFor: 'simulation only' stamp: 'eem 8/23/2021 10:55'!
primitiveGeneratorForPrimitiveIndex: index
	"If there is a generator for the primitive index then answer it;
	 otherwise answer nil."
	<doNotGenerate>
	^(index between: 1 and: MaxCompiledPrimitiveIndex) ifTrue:
		[self class initializationOptions at: #DoNotJIT ifPresent:
			[:excluded| (excluded includes: index) ifTrue: [^nil]].
		 primitiveGeneratorTable at: index]! !

!SimpleStackBasedCogit methodsFor: 'primitive generators' stamp: 'eem 1/13/2016 17:10'!
primitiveGeneratorOrNil
	"If there is a generator for the current primitive then answer it;
	 otherwise answer nil."
	<returnTypeC: #'PrimitiveDescriptor *'>
	| primitiveDescriptor |
	<var: #primitiveDescriptor type: #'PrimitiveDescriptor *'>
	(coInterpreter isQuickPrimitiveIndex: primitiveIndex) ifTrue:
		[primitiveDescriptor := self addressOf: (primitiveGeneratorTable at: 0). "an unused one"
		 primitiveDescriptor primitiveGenerator: (coInterpreter quickPrimitiveGeneratorFor: primitiveIndex).
		 ^primitiveDescriptor].
	(primitiveIndex between: 1 and: MaxCompiledPrimitiveIndex) ifTrue:
		[self cCode: [] inSmalltalk: "for debugging, allow excluding specific primitives"
			[self class initializationOptions at: #DoNotJIT ifPresent:
				[:excluded|
				((excluded includes: primitiveIndex)
				 and: [(primitiveGeneratorTable at: primitiveIndex) primitiveGenerator notNil]) ifTrue:
					[coInterpreter transcript nextPutAll: 'EXCLUDING primitive #'; print: primitiveIndex; space; nextPutAll: (primitiveGeneratorTable at: primitiveIndex) primitiveGenerator; cr; flush.
				 ^nil]]].
		 ^self addressOf: (primitiveGeneratorTable at: primitiveIndex)].
	^nil! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/20/2016 14:20'!
putSelfInReceiverResultReg
	<inline: true>
	self MoveMw: FoxMFReceiver r: FPReg R: ReceiverResultReg! !

!SimpleStackBasedCogit methodsFor: 'testing' stamp: 'eem 2/20/2018 07:29'!
register: reg1 and: reg2 isNotInMask: mask
	<inline: true>
	^mask noMask: (self registerMaskFor: reg1 and: reg2)! !

!SimpleStackBasedCogit methodsFor: 'testing' stamp: 'cb 4/13/2015 12:55'!
register: reg isInMask: mask
	<inline: true>
	^ mask anyMask: (self registerMaskFor: reg)! !

!SimpleStackBasedCogit methodsFor: 'testing' stamp: 'eem 2/1/2018 15:20'!
register: reg isNotInMask: mask
	<inline: true>
	^ mask noMask: (self registerMaskFor: reg)! !

!SimpleStackBasedCogit methodsFor: 'trampolines' stamp: 'eem 02/04/2020 11:38'!
returnRegForStoreCheck
	"We must ensure the ReceiverResultReg is live across the store check so that
	 receiver inst vars can be stored into in a frameless method since self exists only
	 in ReceiverResultReg in a frameless method.  So if ReceiverResultReg is caller-saved
	 we use the fact that ceStoreCheck:/remember: answer their argument to reload
	 ReceiverResultReg cheaply.  Otherwise we don't care about the result and use the
	 cResultRegister, effectively a no-op (see compileTrampoline...)."
	<inline: true>
	^(self isCallerSavedReg: ReceiverResultReg)
		ifTrue: [ReceiverResultReg]
		ifFalse: [ABIResultReg]! !

!SimpleStackBasedCogit methodsFor: 'testing' stamp: 'cb 10/23/2017 11:16'!
seemsToBeInstantiating: format
	"Answers if the code is installed in a class instantiating objects with the format. Used in primitive 
	 generation to make a quick path based on where the method is installed. This method cannot
	 be used as a guarantee as there can be false positive, it's just a heuristic.
	 Tries to interpret the last literal of the method as a behavior (more than 3 fields, 3rd field a Smi).
	 If it can be interpreted as a behavior, answers if instSpec matches the format, else answers false."
	<inline: true>
	^ objectMemory maybeMethodClassOf: methodObj seemsToBeInstantiating: format! !

!SimpleStackBasedCogit methodsFor: 'initialization' stamp: 'eem 7/25/2021 11:17'!
setInterpreter: aCoInterpreter
	"Initialization of the code generator in the simulator.
	 These objects already exist in the generated C VM
	 or are used only in the simulation."
	<doNotGenerate>
	super setInterpreter: aCoInterpreter.
	primitiveGeneratorTable := self class primitiveTable.
	(shortCutTrampolineBlocks := IdentityDictionary new)
		at: #ceShortCutTraceBlockActivation:
			put: [coInterpreter ceTraceBlockActivation];
		at: #ceShortCutTraceLinkedSend:
			put: [coInterpreter ceTraceLinkedSend: (processor registerAt: ReceiverResultReg)];
		at: #ceShortCutTraceStore:
			put: [coInterpreter
					ceTraceStoreOf: (processor registerAt: ClassReg)
					into: (processor registerAt: ReceiverResultReg)]! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/21/2017 15:10'!
temporaryIndexOfFrameOffset: offset
	^self temporaryIndexOfFrameOffset: offset numArgs: methodOrBlockNumArgs! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/21/2017 15:38'!
temporaryIndexOfFrameOffset: offset numArgs: numArgs
	"For assert checking..."
	^numArgs - ((offset < 0 "args are > 0"
					ifTrue: [offset - FoxMFReceiver + objectMemory wordSize]
					ifFalse: [offset - FoxCallerSavedIP]) / objectMemory wordSize)! !

!SimpleStackBasedCogit methodsFor: 'span functions' stamp: 'eem 11/1/2012 17:47'!
v3: descriptor Block: pc Code: nExts Size: aMethodObj
	<var: #descriptor type: #'BytecodeDescriptor *'>
	self assert: nExts <= 0.
	^((objectMemory fetchByte: pc + 2 ofObject: aMethodObj) << 8)
	+ (objectMemory fetchByte: pc + 3 ofObject: aMethodObj)! !

!SimpleStackBasedCogit methodsFor: 'span functions' stamp: 'eem 11/1/2012 13:02'!
v3: descriptor Long: pc Branch: nExts Distance: aMethodObj
	"Answer the distance of a two byte forward long jump."
	<var: #descriptor type: #'BytecodeDescriptor *'>
	self assert: nExts = 0.
	^(((objectMemory fetchByte: pc ofObject: aMethodObj) bitAnd: 7) - 4 << 8)
	 + (objectMemory fetchByte: pc + 1 ofObject: aMethodObj)! !

!SimpleStackBasedCogit methodsFor: 'span functions' stamp: 'eem 11/1/2012 13:01'!
v3: descriptor LongForward: pc Branch: nExts Distance: aMethodObj
	"Answer the distance of a two byte forward long jump."
	<var: #descriptor type: #'BytecodeDescriptor *'>
	self assert: nExts = 0.
	^(((objectMemory fetchByte: pc ofObject: aMethodObj) bitAnd: 3) << 8)
	 + (objectMemory fetchByte: pc + 1 ofObject: aMethodObj)! !

!SimpleStackBasedCogit methodsFor: 'span functions' stamp: 'eem 11/2/2012 10:25'!
v3: descriptor ShortForward: pc Branch: nExts Distance: aMethodObj
	"N.B.  This serves for both BlueBook/V3 and V4 short jumps."
	<var: #descriptor type: #'BytecodeDescriptor *'>
	self assert: nExts = 0.
	^((objectMemory fetchByte: pc ofObject: aMethodObj) bitAnd: 7) + 1! !

!SimpleStackBasedCogit methodsFor: 'span functions' stamp: 'eem 11/13/2012 14:43'!
v4: descriptor Block: pc Code: nExts Size: aMethodObj
	<var: #descriptor type: #'BytecodeDescriptor *'>
	"253		11111101 eei i i kkk	jjjjjjjj		Push Closure Num Copied iii (+ Ext A // 16 * 8) Num Args kkk (+ Ext A \\ 16 * 8) BlockSize jjjjjjjj (+ Ext B * 256). ee = num extensions"
	| byteOne extBValue |
	byteOne := objectMemory fetchByte: pc + 1 ofObject: aMethodObj.
	"If nExts < 0 it isn't known and we rely on the number of extensions encoded in the eeiiikkk byte."
	self assert: (nExts < 0 or: [nExts = (byteOne >> 6)]).
	self parseV4Exts: byteOne >> 6 priorTo: pc in: aMethodObj into: [:ea :eb| extBValue := eb].
	^(objectMemory fetchByte: pc + 2 ofObject: aMethodObj)
	+ (extBValue << 8)! !

!SimpleStackBasedCogit methodsFor: 'span functions' stamp: 'eem 11/5/2012 16:25'!
v4: descriptor Long: pc Branch: nExts Distance: aMethodObj
	"242		11110010	i i i i i i i i	Jump i i i i i i i i (+ Extend B * 256, where bbbbbbbb = sddddddd, e.g. -32768 = i=0, a=0, s=1)"
	<var: #descriptor type: #'BytecodeDescriptor *'>
	| extBValue |
	self assert: nExts >= 0.
	self parseV4Exts: nExts priorTo: pc in: aMethodObj into: [:ea :eb| extBValue := eb].
	^(objectMemory fetchByte: pc + 1 ofObject: aMethodObj)
	+ (extBValue << 8)! !

!SimpleStackBasedCogit methodsFor: 'span functions' stamp: 'cb 1/19/2017 11:54'!
v4: descriptor Long: pc BranchIfNotInstanceOf: nExts Distance: aMethodObj
	"**	254		11111110	kkkkkkkk	jjjjjjjj		branch If Not Instance Of Behavior/Array Of Behavior kkkkkkkk (+ Extend A * 256, where Extend A >= 0) distance jjjjjjjj (+ Extend B * 256, where Extend B >= 0)"
	<var: #descriptor type: #'BytecodeDescriptor *'>
	| extBValue |
	self assert: nExts >= 0.
	self parseV4Exts: nExts priorTo: pc in: aMethodObj into: [:ea :eb| extBValue := eb].
	extBValue < 0 ifTrue: [extBValue := extBValue + 128].
	^(objectMemory fetchByte: pc + 2 ofObject: aMethodObj) + (extBValue << 8)! !

!SimpleStackBasedCogit methodsFor: 'span functions' stamp: 'eem 11/5/2012 16:25'!
v4: descriptor LongForward: pc Branch: nExts Distance: aMethodObj
	"242		11110010	i i i i i i i i	Jump i i i i i i i i (+ Extend B * 256, where bbbbbbbb = sddddddd, e.g. -32768 = i=0, a=0, s=1)"
	"243		11110011	i i i i i i i i	Pop and Jump 0n True i i i i i i i i (+ Extend A * 256)"
	"244		11110100	i i i i i i i i	Pop and Jump 0n False i i i i i i i i (+ Extend A * 256)"
	<var: #descriptor type: #'BytecodeDescriptor *'>
	| extBValue |
	self assert: nExts >= 0.
	self parseV4Exts: nExts priorTo: pc in: aMethodObj into: [:ea :eb| extBValue := eb].
	^(objectMemory fetchByte: pc + 1 ofObject: aMethodObj)
	+ (extBValue << 8)! !

!SimpleStackBasedCogit methodsFor: 'bytecode generator support' stamp: 'eem 7/24/2018 20:21'!
voidReceiverOptStatus
	"No op in this cogit. Provided for compatibility with the cleverer subclasses."
	<inline: true>! !

SimpleStackBasedCogit subclass: #StackToRegisterMappingCogit
	instanceVariableNames: 'prevBCDescriptor numPushNilsFunction pushNilSizeFunction methodOrBlockNumTemps regArgsHaveBeenPushed simStack simStackPtr simSpillBase ceCallCogCodePopReceiverArg0Regs ceCallCogCodePopReceiverArg1Arg0Regs methodAbortTrampolines picAbortTrampolines picMissTrampolines ceCall0ArgsPIC ceCall1ArgsPIC ceCall2ArgsPIC debugStackPointers debugFixupBreaks realCECallCogCodePopReceiverArg0Regs realCECallCogCodePopReceiverArg1Arg0Regs deadCode useTwoPaths currentCallCleanUpSize simNativeStack simNativeStackPtr simNativeSpillBase simNativeStackSize hasNativeFrame compilationPass'
	classVariableNames: 'NeedsMergeFixupFlag NeedsNonMergeFixupFlag'
	poolDictionaries: 'CogCompilationConstants VMMethodCacheConstants VMObjectIndices VMStackFrameOffsets'
	category: 'VMMaker-JIT'!

StackToRegisterMappingCogit class
	instanceVariableNames: 'numPushNilsFunction pushNilSizeFunction'!

!StackToRegisterMappingCogit commentStamp: 'eem 1/26/2018 11:21' prior: 0!
StackToRegisterMappingCogit is an optimizing code generator that eliminates a lot of stack operations and inlines some special selector arithmetic.  It does so by a simple stack-to-register mapping scheme based on deferring the generation of code to produce operands until operand-consuming operations.  The operations that consume operands are sends, stores and returns.  This scheme was first conceived by L. Peter Deutch in the HPS Smalltalk VM (the VisualWorks VM).  Thank you, Peter.

See methods in the class-side documentation protocol for more detail.

Instance Variables
	compilationPass:								<Integer>
	currentCallCleanUpSize:						<Integer>
	ceCall0ArgsPIC:									<Integer>
	ceCall1ArgsPIC:									<Integer>
	ceCall2ArgsPIC:									<Integer>
	ceCallCogCodePopReceiverArg0Regs:			<Integer>
	ceCallCogCodePopReceiverArg1Arg0Regs:		<Integer>
	deadCode										<Boolean>
	debugBytecodePointers:						<Set of Integer>
	debugFixupBreaks:								<Set of Integer>
	debugStackPointers:							<CArrayAccessor of (Integer|nil)>
	hasNativeFrame								<Boolean>
	methodAbortTrampolines:						<CArrayAccessor of Integer>
	methodOrBlockNumTemps:						<Integer>
	numPushNilsFunction:							<Symbol>
	picAbortTrampolines:							<CArrayAccessor of Integer>
	picMissTrampolines:							<CArrayAccessor of Integer>
	pushNilSizeFunction:							<Symbol>
	realCECallCogCodePopReceiverArg0Regs:		<Integer>
	realCECallCogCodePopReceiverArg1Arg0Regs:	<Integer>
	regArgsHaveBeenPushed:						<Boolean>
	simNativeSpillBase:								<Integer>
	simNativeStack:								<CArrayAccessor of CogSimStackNativeEntry>
	simNativeStackPtr:								<Integer>
	simNativeStackSize:							<Integer>
	simSpillBase:									<Integer>
	simStack:										<CArrayAccessor of CogSimStackEntry>
	simStackPtr:									<Integer>
	traceSimStack:									<Integer>
	useTwoPaths									<Boolean>

compilationPass
	- counter indicating whether on the first pass through bytecodes in a V3-style embedded block or not.  The V3 closure implementation uses pushNil to initialize temporary variables and this makes an initial pushNil ambiguous.  With the V3 bytecode set, the JIT must compile to the end of the block to discover if a pushNil is for initializing a temp or to produce an operand.

currentCallCleanUpSize
	- the number of bytes to remove from the stack in a Lowcode call.

ceCall0ArgsPIC ceCall1ArgsPIC ceCall2ArgsPIC
	- the trampoline for entering an N-arg PIC

ceCallCogCodePopReceiverArg0Regs ceCallCogCodePopReceiverArg1Arg0Regs
	- the trampoline for invokinging a method with N register args
	
debugBytecodePointers
	- a Set of bytecode pcs for setting breakpoints (simulation only)

deadCode
	- set to true to indicate that the next bytecode (up to the next fixup) is not reachable.  Used to avoid generating dead code.

debugFixupBreaks
	- a Set of fixup indices for setting breakpoints (simulation only)

debugStackPointers
	- an Array of stack depths for each bytecode for code verification (simulation only)

hasNativeFrame
	- set to true when Lowcode creates a native stack frame for Lowcode callouts.

methodAbortTrampolines
	- a CArrayAccessor of abort trampolines for 0, 1, 2 and N args

methodOrBlockNumTemps
	- the number of method or block temps (including args) in the current compilation unit (method or block)

numPushNilsFunction
	- the function used to determine the number of push nils at the beginning of a block.  This abstracts away from the specific bytecode set(s).

picAbortTrampolines
	- a CArrayAccessor of abort trampolines for 0, 1, 2 and N args

picMissTrampolines
	- a CArrayAccessor of abort trampolines for 0, 1, 2 and N args

pushNilSizeFunction
	- the function used to determine the number of bytes in the push nils bytecode(s) at the beginning of a block.  This abstracts away from the specific bytecode set(s).

realCECallCogCodePopReceiverArg0Regs realCECallCogCodePopReceiverArg1Arg0Regs
	- the real trampolines for invoking machine code with N reg args when in the Debug regime

regArgsHaveBeenPushed
	- whether the register args have been pushed before frame build (e.g. when an interpreter primitive is called)

simNativeSpillBase
	- the variable tracking how much of the Lowcode simulation stack has been spilled to the real stack

simNativeStack
	- the Lowcode simulation stack itself

simNativeStackPtr
	- the pointer to the top of the Lowcode simulation stack

simNativeStackSize
	- the size of the Lowcode stack so far

simSpillBase
	- the variable tracking how much of the simulation stack has been spilled to the real stack

simStack
	- the simulation stack itself, comprising the receiver, arguments, temporaries, and volatile stack contents.  The receiver is the 0'th entry, and the 1st is that of the first argument, etc.

simStackPtr
	- the pointer to the top of the simulation stack

useTwoPaths
	- a variable controlling whether to create two paths through a method based on the existence of inst var stores.  With immutability this causes a frameless path to be generated if an otherwise frameless method is frameful simply because of inst var stores.  In this case the test to take the first frameless path is if the receiver is not immutable.  Without immutability, if a frameless method contains two or more inst var stores, the first path will be code with no store check, chosen by a single check for the receiver being in new space.
!

StackToRegisterMappingCogit class
	instanceVariableNames: 'numPushNilsFunction pushNilSizeFunction'!

StackToRegisterMappingCogit subclass: #RegisterAllocatingCogit
	instanceVariableNames: 'numFixups mergeSimStacksBase nextFixup scratchSimStack scratchSpillBase ceSendMustBeBooleanAddTrueLongTrampoline ceSendMustBeBooleanAddFalseLongTrampoline recompileForLoopRegisterAssignments scratchBytecodePC'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!RegisterAllocatingCogit commentStamp: 'eem 2/9/2017 10:40' prior: 0!
RegisterAllocatingCogit is an optimizing code generator that is specialized for register allocation.

On the contrary to StackToRegisterMappingCogit, RegisterAllocatingCogit keeps at each control flow merge point the state of the simulated stack to merge into and not only an integer fixup. Each branch and jump record the current state of the simulated stack, and each fixup is responsible for merging this state into the saved simulated stack.

Instance Variables
	ceSendMustBeBooleanAddFalseLongTrampoline:		<Integer>
	ceSendMustBeBooleanAddTrueLongTrampoline:		<Integer>
	mergeSimStacksBase:									<Integer>
	nextFixup:												<Integer>
	numFixups:												<Integer>
	scratchOptStatus:										<CogSSOptStatus>
	scratchSimStack:										<Array of CogRegisterAllocatingSimStackEntry>
	scratchSpillBase:										<Integer>

ceSendMustBeBooleanAddFalseLongTrampoline
	- the must-be-boolean trampoline for long jump false bytecodes (the existing ceSendMustBeBooleanAddFalseTrampoline is used for short branches)

ceSendMustBeBooleanAddTrueLongTrampoline
	- the must-be-boolean trampoline for long jump true bytecodes (the existing ceSendMustBeBooleanAddTrueTrampoline is used for short branches)

mergeSimStacksBase
	- the base address of the alloca'ed memory for merge fixups

nextFixup
	- the index into mergeSimStacksBase from which the next needed mergeSimStack will be allocated

numFixups
	- a conservative (over) estimate of the number of merge fixups needed in a method

scratchOptStatus
	- a scratch variable to hold the state of optStatus while merge code is generated

scratchSimStack
	- a scratch variable to hold the state of simStack while merge code is generated

scratchSpillBase
	- a scratch variable to hold the state of spillBase while merge code is generated!

!RegisterAllocatingCogit class methodsFor: 'C translation' stamp: 'eem 4/21/2017 16:47'!
declareCVarsIn: aCodeGen
	aCodeGen
		var: #scratchSimStack
			type: #'SimStackEntry *';
		var: #scratchSimSelf
			type: #SimStackEntry! !

!RegisterAllocatingCogit class methodsFor: 'trampoline support' stamp: 'eem 12/7/2016 13:06'!
numTrampolines
	^super numTrampolines + 2 "includes long sendMustBeBoolean trampolines"

	"Cogit withAllSubclasses, CogObjectRepresentation withAllSubclasses collect:
		[:c| {c. (c instVarNames select: [:ea| ea beginsWith: 'ce']) size}]"
	"self allInstVarNames select: [:ea| ea beginsWith: 'ce']"
	"self instVarNames select: [:ea| ea beginsWith: 'ce']"! !

!RegisterAllocatingCogit methodsFor: 'compile abstract instructions' stamp: 'cb 4/16/2016 09:24'!
allocateMergeFixups
	"Allocate the various arrays needed to allocate the merge fixups, failing if the size
	 needed is considered too high.

	 This *must* be inlined since the arrays are alloca'ed (stack allocated)
	 so that they are freed when compilation is done.

	 N.B. We do one single alloca to save embarrassing C optimizers that
	 generate incorrect code as both gcc and the intel compiler do on x86."
	<inline: true>
	| mergeSimStackBytes |
	mergeSimStackBytes := numFixups * self simStackSlots * (self sizeof: CogSimStackEntry).
	nextFixup := 0.
	self cCode:
		[mergeSimStacksBase := self alloca: mergeSimStackBytes.
		 self b: mergeSimStacksBase zero: mergeSimStackBytes]! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 12/4/2016 20:12'!
allocateRegForStackEntryAt: index
	"If the stack entry is already in a register, answers it,
	else allocate a new register for it"
	<inline: true>
	| reg |
	(reg := (self ssValue: index) registerOrNone) ~= NoReg ifTrue:
		[^reg].
	^self allocateRegForStackEntryAt: index notConflictingWith: (self liveRegisters bitOr: (self registerMaskFor: FPReg and: SPReg and: TempReg))! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 4/5/2017 12:03'!
allocatedRegisters
	| regsSet |
	self assert: needsFrame.
	regsSet := 0.
	0 to: methodOrBlockNumTemps do:
		[:i|
		regsSet := regsSet bitOr: (self simStackAt: i) registerMask].
	^regsSet! !

!RegisterAllocatingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 7/24/2018 17:29'!
assertCorrectSimStackPtr
	<inline: true> "generates nothing anyway" "self simStackPrintString"
	 self cCode: '' inSmalltalk:
		[deadCode ifFalse:
			[self assert: simStackPtr + (needsFrame ifTrue: [0] ifFalse: [1])
						= (self debugStackPointerFor: bytecodePC).
			 self assert: (simSpillBase >= methodOrBlockNumTemps
						or: [self maybeCompilingFirstPassOfBlockWithInitialPushNil and: [simSpillBase > methodOrBlockNumArgs]]).
			 (needsFrame and: [simSpillBase > 0]) ifTrue:
				[self assert: (self simStackAt: simSpillBase - 1) spilled == true.
				 self assert: (simSpillBase >= simStackPtr or: [(self simStackAt: simSpillBase) spilled == false])]].
		 self deny: self duplicateRegisterAssignmentsInTemporaries]! !

!RegisterAllocatingCogit methodsFor: 'simulation only' stamp: 'cb 4/15/2016 16:27'!
bytecodeFixupClass
	<doNotGenerate>
	^CogRASSBytecodeFixup! !

!RegisterAllocatingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 2/7/2018 12:32'!
compileAbstractInstructionsFrom: start through: end
	"Loop over bytecodes, dispatching to the generator for each bytecode, handling fixups in due course.
	 Override to recompile after a loop requiring a merge is detected."
	scratchBytecodePC := nil.
	^super compileAbstractInstructionsFrom: start through: end! !

!RegisterAllocatingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 2/7/2018 14:18'!
compileBlockBodies
	<inline: false>
	| result compiledBlocksCount blockStart savedNeedsFrame savedNumArgs savedNumTemps
	  initialStackPtr initialOpcodeIndex initialIndexOfIRC initialCounterIndex |
	<var: #blockStart type: #'BlockStart *'>
	self assert: blockCount > 0.
	"scanBlock: in compileBlockEntry: sets both of these appropriately for each block."
	savedNeedsFrame := needsFrame.
	savedNumArgs := methodOrBlockNumArgs.
	savedNumTemps := methodOrBlockNumTemps.
	inBlock := InVanillaBlock.
	compiledBlocksCount := 0.
	[compiledBlocksCount < blockCount] whileTrue:
		[blockStart := self blockStartAt: compiledBlocksCount.
		 (result := self scanBlock: blockStart) < 0 ifTrue: [^result].
		 compilationPass := 1.
		 scratchBytecodePC := nil.
		 initialOpcodeIndex := opcodeIndex.
		 initialCounterIndex := self maybeCounterIndex."for SistaCogit"
		 literalsManager saveForRecompile.
		 NewspeakVM ifTrue:
			[initialIndexOfIRC := indexOfIRC].
		 [recompileForLoopRegisterAssignments := false.
		  self compileBlockEntry: blockStart.
		  initialStackPtr := simStackPtr.
		  (result := self compileAbstractInstructionsFrom: blockStart startpc + (self pushNilSize: methodObj numInitialNils: blockStart numInitialNils)
						through: blockStart startpc + blockStart span - 1) < 0 ifTrue:
			[^result].
		  "If the final simStackPtr is less than the initial simStackPtr then scanBlock: over-
		   estimated the number of initial nils (because it assumed one or more pushNils to
		   produce an operand were pushNils to initialize temps.  This is very rare, so
		   compensate by checking, adjusting numInitialNils and recompiling the block body.
		   N.B.  No need to reinitialize the literalsManager because it answers existing literals."
		  initialStackPtr ~= simStackPtr or: [recompileForLoopRegisterAssignments]]
			whileTrue:
				[self assert: (initialStackPtr > simStackPtr or: [deadCode]).
				 self assert: compilationPass = 1.
				 compilationPass := compilationPass + 1. "for asserts"
				 blockStart numInitialNils: blockStart numInitialNils + simStackPtr - initialStackPtr.
				 blockStart fakeHeader dependent: nil.
				 self reinitializeAllButBackwardFixupsFrom: blockStart startpc + blockStart numInitialNils through: blockStart startpc + blockStart span - 1.
				 nextFixup := 0.
				 self resetSimStack: blockStart startpc.
				 self reinitializeOpcodesFrom: initialOpcodeIndex to: opcodeIndex - 1.
				 opcodeIndex := initialOpcodeIndex.
				 self maybeSetCounterIndex: initialCounterIndex. "For SistaCogit"
				 literalsManager resetForRecompile.
				 NewspeakVM ifTrue:
					[indexOfIRC := initialIndexOfIRC]].
		compiledBlocksCount := compiledBlocksCount + 1].
	needsFrame := savedNeedsFrame.
	methodOrBlockNumArgs := savedNumArgs.
	methodOrBlockNumTemps := savedNumTemps.
	^0! !

!RegisterAllocatingCogit methodsFor: 'compile abstract instructions' stamp: 'cb 4/15/2016 16:19'!
compileEntireFullBlockMethod: numCopied
	"Compile the abstract instructions for the entire full block method."
	self allocateMergeFixups.
	^super compileEntireFullBlockMethod: numCopied! !

!RegisterAllocatingCogit methodsFor: 'compile abstract instructions' stamp: 'cb 4/15/2016 16:14'!
compileEntireMethod
	"Compile the abstract instructions for the entire method, including blocks."
	self allocateMergeFixups.
	^super compileEntireMethod! !

!RegisterAllocatingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 2/7/2018 12:34'!
compileMethodBody
	"Compile the top-level method body."
	<inline: true>
	| result initialOpcodeIndex initialCounterIndex initialIndexOfIRC start |
	endPC < initialPC ifTrue: [^0]. "quick primitives"
	"When compiling, skip any initial CallPrimitive and optional StorePrimErrCode bytecodes.
	 These are dealt with in compileFrameBuild."
	compilationPass := 1.
	scratchBytecodePC := nil.
	initialOpcodeIndex := opcodeIndex.
	initialCounterIndex := self maybeCounterIndex."for SistaCogit"
	literalsManager saveForRecompile.
	NewspeakVM ifTrue:
		[initialIndexOfIRC := indexOfIRC].
	[recompileForLoopRegisterAssignments := false.
	 start := initialPC + (self deltaToSkipPrimAndErrorStoreIn: methodObj header: methodHeader).
	 result := self compileAbstractInstructionsFrom: start through: endPC.
	 result = 0
	 and: [recompileForLoopRegisterAssignments]]
		whileTrue:
			[self assert: compilationPass <= 2.
			 self reinitializeAllButBackwardFixupsFrom: start through: endPC.
			 self resetSimStack: start.
			 self reinitializeOpcodesFrom: initialOpcodeIndex to: opcodeIndex - 1.
			 compilationPass := compilationPass + 1.
			 nextFixup := 0.
			 opcodeIndex := initialOpcodeIndex.
			 self maybeSetCounterIndex: initialCounterIndex. "For SistaCogit"
			 literalsManager resetForRecompile.
			 NewspeakVM ifTrue:
				[indexOfIRC := initialIndexOfIRC]].
	^result! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 3/4/2017 14:17'!
conflictingRegistersBetweenSimStackAnd: mergeSimStack
	<var: #mergeSimStack type: #'SimStackEntry *'>
	| currentRegsMask mergeRegsMask potentialConflictRegMask |
	<var: #currentEntry type: #'SimStackEntry *'>
	<var: #targetEntry type: #'SimStackEntry *'>
	currentRegsMask := mergeRegsMask := potentialConflictRegMask := 0.
	0 to: simStackPtr do:
		[:i| | currentEntry targetEntry currentRegMask mergeRegMask |
		 currentRegMask := (currentEntry := self simStack: simStack at: i) registerMaskOrNone.
		 mergeRegMask := (targetEntry := self simStack: mergeSimStack at: i) registerMaskOrNone.
		 (currentRegMask ~= mergeRegMask
		  and: [currentRegMask ~= 0 or: [mergeRegMask ~= 0]]) ifTrue:
			[potentialConflictRegMask := potentialConflictRegMask bitOr: (currentRegMask bitOr: mergeRegMask)].
		 currentRegsMask := currentRegsMask bitOr: currentRegMask.
		 mergeRegsMask := mergeRegsMask bitOr: mergeRegMask].
	^potentialConflictRegMask bitAnd: (currentRegsMask bitAnd: mergeRegsMask)! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/4/2018 11:21'!
conflictsResolvedBetweenSimStackAnd: mergeSimStack 
	"There are no register conflicts between simStack and mergeSimStack if
	 traversing both stacks from hot end (simStackPtr) to cold end (0) no register
	 exists in simStack that has previously existed in mergeSimStack.  This is because
	 the resolution assigns values from simStack to registers in mergeSimStack and so
	 must not assign to a register yet to be read."
	 | regsWrittenToMask |
	regsWrittenToMask := 0.
	simStackPtr to: 0 by: -1 do:
		[:i| | mergeMask currentMask |
		mergeMask := (self simStack: mergeSimStack at: i) registerMaskOrNone.
		currentMask := (self simStack: simStack at: i) registerMaskOrNone.
		"Ignore targets with no registers.  Current's register can be deassigned and so there's no conflict."
		(mergeMask ~= 0 and: [mergeMask ~= currentMask]) ifTrue:
			[(currentMask anyMask: regsWrittenToMask) ifTrue:
				[^false]].
		regsWrittenToMask := regsWrittenToMask bitOr: mergeMask].
	^true! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 11/26/2016 10:36'!
copyLiveRegisterToCopiesOf: simStackEntry
	"Copy the liveRegister in simStackEntry into all corresponding stack entries."
	<var: #simStackEntry type: #'SimStackEntry *'>
	simStackPtr to: 0 by: -1 do:
		[:i|
		(self simStackAt: i) copyLiveRegisterIfSameAs: simStackEntry]! !

!RegisterAllocatingCogit methodsFor: 'simulation only' stamp: 'cb 4/16/2016 09:31'!
copySimStack
	<doNotGenerate>
	^CArrayAccessor on: (simStack object collect: [:stackEntry| stackEntry copy])! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 9/19/2018 17:31'!
copySimStackToFixup: fixup
	<var: #fixup type: #'BytecodeFixup *'>
	<inline: true>
	self cCode: [self memcpy: fixup mergeSimStack _: simStack _: simStackPtr + 1 * (self sizeof: CogSimStackEntry)]
		inSmalltalk: [0 to: simStackPtr do:
						[:i|
						fixup mergeSimStack at: i put: (simStack at: i) copy]]! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 9/19/2018 17:31'!
copySimStackToScratch: spillBase
	<inline: true>
	self assert: (spillBase > methodOrBlockNumTemps
				or: [self maybeCompilingFirstPassOfBlockWithInitialPushNil and: [spillBase > methodOrBlockNumArgs]]).
	scratchBytecodePC = bytecodePC ifTrue:
		[^self].
	scratchBytecodePC := bytecodePC.
	self cCode: [self memcpy: scratchSimStack _: simStack _: self simStackSlots * (self sizeof: CogSimStackEntry)]
		inSmalltalk: [0 to: simStackPtr do:
						[:i|
						scratchSimStack at: i put: (simStack at: i) copy]].
	scratchSpillBase := spillBase! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/6/2018 10:34'!
deassignRegister: register in: aSimStack
	"When a merge has assigned a register to a volatile stack entry
	 deassign that register from any temporaries."
	self assert: register ~= NoReg.
	0 to: methodOrBlockNumTemps do:
		[:i|
		(self simStack: aSimStack at: i) liveRegister = register ifTrue:
			[(self simStack: aSimStack at: i) liveRegister: NoReg]]! !

!RegisterAllocatingCogit methodsFor: 'debugging' stamp: 'eem 2/6/2018 17:58'!
duplicateRegisterAssignmentsInTemporaries
	| liveRegisters |
	liveRegisters := 0.
	0 to: (methodOrBlockNumTemps min: (simStackPtr min: simSpillBase - 1)) do: "The min:s are for the initial push nil hack"
		[:i| | current liveRegister |
		liveRegister := (current := self simStackAt: i) liveRegister.
		liveRegister ~= NoReg ifTrue:
			[(self register: liveRegister isInMask: liveRegisters) ifTrue:
				["Filter out pushed temps in first-pass vanilla blocks"
				 (self maybeCompilingFirstPassOfBlockWithInitialPushNil
				  and: [current type = SSBaseOffset
				  and: [current offset ~= (self frameOffsetOfTemporary: i - 1)]]) ifFalse:
					[^true]].
			 liveRegisters := liveRegisters bitOr: (self registerMaskFor: liveRegister)]].
	^false! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/8/2018 13:55'!
ensureFixupAt: targetPC
	"Make sure there's a flagged fixup at the target pc in fixups.
	 Initially a fixup's target is just a flag.  Later on it is replaced with a proper instruction.
	 Override to generate stack merging code if required."
	| fixup |	
	<var: #fixup type: #'BytecodeFixup *'>
	self assert: targetPC > bytecodePC.
	self cCode: '' inSmalltalk:
		[self assert: simStackPtr + (needsFrame ifTrue: [0] ifFalse: [1])
					= (self debugStackPointerFor: targetPC)].
	fixup := self fixupAt: targetPC.
	"If a non-merge fixup has already been defined then where-ever that was done didn't
	 realise there needed to be a merge and forgot to save the stack state for that merge."
	self deny: fixup isNonMergeFixup.
	fixup needsFixup 
		ifTrue:
			[fixup mergeSimStack
				ifNil: [self setMergeSimStackOf: fixup]
				ifNotNil:
					[self copySimStackToScratch: simSpillBase.
					 self mergeCurrentSimStackWith: fixup.
					 self restoreSimStackFromScratch]]
		ifFalse: 
			[self assert: (fixup mergeSimStack isNil or: [compilationPass = 2]).
			 fixup mergeSimStack
				ifNil: [self setMergeSimStackOf: fixup]
				ifNotNil:
					[self assert: (self simStack: simStack isIdenticalTo: fixup mergeSimStack)]].
	^super ensureFixupAt: targetPC! !

!RegisterAllocatingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 2/8/2018 13:58'!
ensureNonMergeFixupAt: targetPC
	"RegisterAllocatingCogit insists on merging."
	self shouldNotImplement! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 7/24/2018 19:56'!
ensureReceiverResultRegContainsSelf
	"First ensure that ReceiverResultReg is allocated to self,
	 which may cause spills, etc.  Then copy the register into
	 any instances of simSelf on the stack."

	| freeReg live ssEntry |
	needsFrame
		ifTrue:
			[self receiverIsInReceiverResultReg ifFalse: "Avoid spilling if possible"
				[live := self liveRegisters.
				 ssEntry := self ssEntrySuchThat: [:e| e registerOrNone = ReceiverResultReg].
				 (ssEntry notNil
				  and: [(freeReg := backEnd availableRegisterOrNoneFor: live) ~= NoReg])
					ifTrue: "ReceiverResultReg is being used, but we have a free register.  So avoid spill by assigining."
						[ssEntry storeToRegNoAssign: freeReg.
						 self ssEntriesDo:
							[:e|
							 e registerOrNone = ReceiverResultReg ifTrue:
								[e type = SSRegister
									ifTrue: [e register: freeReg]
									ifFalse: [e liveRegister: freeReg]]]]
					ifFalse:
						[self ssAllocateRequiredReg: ReceiverResultReg].
				 self putSelfInReceiverResultReg.
				 self simSelf liveRegister: ReceiverResultReg]]
		ifFalse:
			[self assert: (self simSelf type = SSRegister
						  and: [self simSelf register = ReceiverResultReg
						  and: [self receiverIsInReceiverResultReg]])].
	"the storeToReg: in putSelfInReceiverResultReg in
	 ensureReceiverResultRegContainsSelf copies the register to all copies.
	 So simply check that the stack agrees with this."
	self assert: self receiverResultRegIsAssignedToSelfAndNothingElse! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/3/2018 16:02'!
ensureRegisterAssignmentsAreAtHeadOfLoop: target
	"Compiling a loop body will compute a set of live registers.  The backward branch must merge
	 with the head of the loop.  So it is preferrable to make the register assignments at the end of
	 the loop available at the head.  To do this, simply copy the register assignments to the loop
	 head's fixup in the first compilation pass and schedule a second compilation pass.  On the
	 second pass the merge will occur when encountering the fixup for the loop head, using
	 exactly the same code as for a merge at the end of an if."
	| conflictingRegsMask |
	compilationPass > 1 ifTrue:
		[self assert: (target mergeSimStack notNil).
		 ^self].
	(self mergeRequiredToTarget: target mergeSimStack) ifFalse:
		[^self].
	"Schedule a recompile and merge the end-of-loop assignments into the head of the loop,
	 replacing any and all register assignments with the state as of the back jump.  Because
	 typically the back jump will be taken much more often than the loop entered, favouring
	 the assignments here is more efficient than trying to merge."
	recompileForLoopRegisterAssignments := true.
	conflictingRegsMask := self conflictingRegistersBetweenSimStackAnd: target mergeSimStack.
	self deny: (self register: FPReg isInMask: conflictingRegsMask).
	0 to: simStackPtr do:
		[:i| | currentEntry targetEntry |
		 currentEntry := self simStack: simStack at: i.
		 targetEntry := self simStack: target mergeSimStack at: i.
		 targetEntry liveRegister: currentEntry liveRegister]! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/21/2018 16:55'!
existsInstVarRefBeforeSendOrReturn
	"Answer if the current bytecode is followed by an inst var ref before the next full send."
	| pc nExts descriptor |
	pc := bytecodePC.
	nExts := 0.
	[pc <= endPC] whileTrue:
		[descriptor := self generatorForPC: pc.
		 (descriptor isMapped
		  or: [descriptor isConditionalBranch
		  or: [descriptor spanFunction notNil]]) ifTrue:
			[^false].
		 descriptor isInstVarRef ifTrue:
			[^true].
		 nExts := descriptor isExtension ifTrue: [nExts + 1] ifFalse: [0].
		 pc := self nextBytecodePCFor: descriptor at: pc exts: nExts in: methodObj].
	^false! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/25/2018 14:10'!
flushLiveRegistersForCRunTimeCall
	"Flush any live registers for a C call, i.e. don't flush caller-saved registers.
	 Answer if any registers were flushed."
	<inline: true>
	| flushed reg |
	flushed := false.
	0 to: simStackPtr do:
		[:i|
		 self assert: (self simStackAt: i) type = (i <= methodOrBlockNumTemps
													ifTrue: [SSBaseOffset]
													ifFalse: [SSSpill]).
		 reg := (self simStackAt: i) liveRegister.
		 (reg ~= NoReg and: [(self isCallerSavedReg: reg)]) ifTrue:
			[(self simStackAt: i) liveRegister: NoReg.
			 flushed := true]].
	^flushed! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/25/2018 14:24'!
flushLiveRegistersForSend
	<inline: true>
	0 to: simStackPtr do:
		[:i|
		 self assert: ((self simStackAt: i) spilled
					 and: [(self simStackAt: i) type = SSConstant
						or: [((self simStackAt: i) type = SSBaseOffset
							or: [i > methodOrBlockNumTemps
								and: [(self simStackAt: i) type = SSSpill]])
							 and: [(self simStackAt: i) register = FPReg
							 and: [i = 0
								or: [(self simStackAt: i) offset = (self frameOffsetOfTemporary: i - 1)]]]]]).
		 (self simStackAt: i) liveRegister: NoReg]! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/25/2018 13:58'!
flushLiveRegistersForSuspensionPoint
	"Flush any live registers for a C call at a suspension/resumption point, i.e.flush all registers.
	 Answer if any registers were flushed."
	<inline: true>
	| flushed |
	flushed := false.
	0 to: simStackPtr do:
		[:i|
		 self assert: (i <= methodOrBlockNumTemps
						ifTrue: [(self simStackAt: i) type = SSBaseOffset]
						ifFalse: [(self simStackAt: i)  spilled]).
		 (self simStackAt: i) liveRegister ~= NoReg ifTrue:
			[(self simStackAt: i) liveRegister: NoReg.
			 flushed := true]].
	^flushed! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 2/12/2017 16:33'!
flushRegistersOnlyLiveOnFallThrough: fixup
	"Forward jumps won't generate merge code if the source has a register that is live but the destination does not.
	 For example in
			| v | v := expr1. self expr2 ifTrue: [v := expr3]. ^v
	 v will be assigned to a register in v := expr1 and [v := expr3], but the send of expr2 will flush it along the jumpFalse across [v := expr3].
	 So v will not be in a register if reached from the jump.  Hence at the join at the end of [v := expr3] v must be marked as not being in a register."
	| targetSimStack |
	targetSimStack := fixup mergeSimStack.
	0 to: simStackPtr do:
		[:i| | fallThrough target |
		 fallThrough := self simStack: simStack at: i.
		 target := self simStack: targetSimStack at: i.
		 self assert: (fallThrough liveRegister = target liveRegister or: [target liveRegister = NoReg or: [fallThrough liveRegister = NoReg]]).
		 target liveRegister = NoReg ifTrue:
			[fallThrough liveRegister: NoReg]]! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 1/25/2018 18:21'!
freeAnyRegNotConflictingWith: regMask
	"Spill the closest register on stack not conflicting with regMask. 
	 Override to unassign assigned temp regs if necessary."
	<var: #desc type: #'CogSimStackEntry *'>
	| index desc |
	self assert: needsFrame.
	self assert: simSpillBase >= 0.
	index := simSpillBase.
	[index < simStackPtr] whileTrue: 
		[desc := self simStackAt: index.
		 desc type = SSRegister ifTrue:
			[(regMask anyMask: (self registerMaskFor: desc register)) ifFalse: 
				[self ssAllocateRequiredReg: desc register.
				 ^desc register]].
		 index := index + 1].
	1 to: methodOrBlockNumTemps do:
		[:i|
		 desc := self simStackAt: i.
		 (desc liveRegister = NoReg
		  or: [self register: desc liveRegister isInMask: regMask]) ifFalse:
			[self ssAllocateRequiredReg: desc liveRegister.
			 ^desc liveRegister]].
	(self simSelf liveRegister = NoReg
	 or: [self register: self simSelf liveRegister isInMask: regMask]) ifFalse:
		[self ssAllocateRequiredReg: self simSelf liveRegister.
		 ^self simSelf liveRegister].
	^NoReg! !

!RegisterAllocatingCogit methodsFor: 'trampoline support' stamp: 'eem 7/24/2018 15:48'!
genCallMustBeBooleanFor: boolean
	"Call ceSendMustBeBooleanTo:interpretingAtDelta: via the relevant trampoline."
	self assert: ((self generatorAt: byte0) numBytes between: 1 and: 2).
	^self CallRT: ((self generatorAt: byte0) numBytes = 1
					ifTrue:
						[boolean = objectMemory falseObject
							ifTrue: [ceSendMustBeBooleanAddFalseTrampoline]
							ifFalse: [ceSendMustBeBooleanAddTrueTrampoline]]
					ifFalse:
						[boolean = objectMemory falseObject
							ifTrue: [ceSendMustBeBooleanAddFalseLongTrampoline]
							ifFalse: [ceSendMustBeBooleanAddTrueLongTrampoline]])! !

!RegisterAllocatingCogit methodsFor: 'bytecode generators' stamp: 'cb 9/13/2019 12:04'!
genForwardersInlinedIdenticalOrNotIf: orNot
	| nextPC branchDescriptor unforwardRcvr argReg targetPC
	  unforwardArg  rcvrReg postBranchPC retry fixup
	  comparison rcvrConstant argConstant
	  needMergeToTarget needMergeToContinue |
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>
	<var: #toContinueLabel type: #'AbstractInstruction *'>
	<var: #toTargetLabel type: #'AbstractInstruction *'>
	<var: #comparison type: #'AbstractInstruction *'>
	<var: #retry type: #'AbstractInstruction *'>
	
	self extractMaybeBranchDescriptorInto: [ :descr :next :postBranch :target | 
		branchDescriptor := descr. nextPC := next. postBranchPC := postBranch. targetPC := target ].

	"If an operand is an annotable constant, it may be forwarded, so we need to store it into a 
	register so the forwarder check can jump back to the comparison after unforwarding the constant.
	However, if one of the operand is an unnanotable constant, does not allocate a register for it 
	(machine code will use operations on constants) and does not generate forwarder checks."
	unforwardRcvr := (self ssValue: 1) mayBeAForwarder.
	unforwardArg := self ssTop mayBeAForwarder.
	(unforwardRcvr not and: [unforwardArg not])
		ifTrue: [^self genVanillaInlinedIdenticalOrNotIf: orNot].
	self assert: (unforwardArg or: [unforwardRcvr]).
	"We use reg for non annotable constants to avoid duplicating objRef."
	rcvrConstant := objectRepresentation isUnannotatableConstant: (self ssValue: 1).
	argConstant := objectRepresentation isUnannotatableConstant: self ssTop.

	self 
		allocateEqualsEqualsRegistersArgNeedsReg: argConstant not 
		rcvrNeedsReg: rcvrConstant not 
		into: [ :rcvr :arg | rcvrReg:= rcvr. argReg := arg ].

	"If not followed by a branch, resolve to true or false."
	(branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse]) ifFalse:
		[^self 
			genIdenticalNoBranchArgIsConstant: argConstant
			rcvrIsConstant: rcvrConstant
			argReg: argReg 
			rcvrReg: rcvrReg 
			orNotIf: orNot].
	
	self assert: (unforwardArg or: [unforwardRcvr]).
	self ssPop: 2. "If we had moveAllButTop: 2 volatileSimStackEntriesToRegistersPreserving: we could avoid the extra ssPop:s"
	self moveVolatileSimStackEntriesToRegistersPreserving:
		(self allocatedRegisters bitOr: (argReg = NoReg
										ifTrue: [self registerMaskFor: rcvrReg]
										ifFalse:
											[rcvrReg = NoReg
												ifTrue: [self registerMaskFor: argReg]
												ifFalse: [self registerMaskFor: rcvrReg and: argReg]])).
	retry := self Label.
	self ssPop: -2.
	self genCmpArgIsConstant: argConstant rcvrIsConstant: rcvrConstant argReg: argReg rcvrReg: rcvrReg.
	self ssPop: 2.

	(self fixupAt: nextPC) notAFixup "The next instruction is dead.  we can skip it."
		ifTrue:  [deadCode := true]
		ifFalse: [self deny: deadCode]. "push dummy value below"

	"self printSimStack; printSimStack: (self fixupAt: postBranchPC) mergeSimStack"
	"If there are merges to be performed on the forward branches we have to execute
	 the merge code only along the path requiring that merge, and exactly once."
	needMergeToTarget := self mergeRequiredForJumpTo: targetPC.
	needMergeToContinue := self mergeRequiredForJumpTo: postBranchPC.
	orNot == branchDescriptor isBranchTrue
		ifFalse: "a == b ifTrue: ... or a ~~ b ifFalse: ... jump on equal to target pc"
			[fixup := needMergeToContinue
						ifTrue: [0] "jumps will fall-through to to-continue merge code"
						ifFalse: [self ensureFixupAt: postBranchPC].
			 comparison := self JumpZero: (needMergeToTarget
												ifTrue: [0] "comparison will be fixed up to to-target merge code"
												ifFalse: [self ensureFixupAt: targetPC])]
		ifTrue: "a == b ifFalse: ... or a ~~ b ifTrue: ... jump on equal to post-branch pc"
			[fixup := needMergeToTarget
						ifTrue: [0] "jumps will fall-through to to-target merge code"
						ifFalse: [self ensureFixupAt: targetPC].
			 comparison := self JumpZero: (needMergeToContinue
												ifTrue: [0] "comparison will be fixed up to to-continue merge code"
												ifFalse: [self ensureFixupAt: postBranchPC])].

	"The forwarders check(s) need(s) to jump back to the comparison (retry) if a forwarder is found,
	 else jump forward either to the next forwarder check or to the postBranch or branch target (fixup).
	 But if there is merge code along a path, the jump must be to the merge code."
	(unforwardArg and: [unforwardRcvr]) ifTrue:
		[objectRepresentation genEnsureOopInRegNotForwarded: argReg scratchReg: TempReg jumpBackTo: retry].
	objectRepresentation 
		genEnsureOopInRegNotForwarded: (unforwardRcvr ifTrue: [rcvrReg] ifFalse: [argReg]) 
		scratchReg: TempReg 
		ifForwarder: retry
		ifNotForwarder: fixup.
	"If fixup is zero then the ifNotForwarder path falls through to a Label which is interpreted
	 as either to-continue or to-target, depending on orNot == branchDescriptor isBranchTrue."
	orNot == branchDescriptor isBranchTrue
		ifFalse: "a == b ifTrue: ... or a ~~ b ifFalse: ... jump on equal to target pc"
			[needMergeToContinue ifTrue: "fall-through to to-continue merge code"
				[self Jump: (self ensureFixupAt: postBranchPC)].
			 needMergeToTarget ifTrue: "fixup comparison to to-target merge code"
				[comparison jmpTarget: self Label.
				 self Jump: (self ensureFixupAt: targetPC)]]
		ifTrue: "a == b ifFalse: ... or a ~~ b ifTrue: ... jump on equal to post-branch pc"
			[needMergeToTarget ifTrue: "fall-through to to-target merge code"
				[self Jump: (self ensureFixupAt: targetPC)].
			 needMergeToContinue ifTrue: "fixup comparison to to-continue merge code"
				[comparison jmpTarget: self Label.
				 self Jump: (self ensureFixupAt: postBranchPC)]].

	deadCode ifFalse: "duplicate the merge fixup's top of stack so as to avoid a false confict."
		[self ssPushDesc: ((self fixupAt: nextPC) mergeSimStack at: simStackPtr + 1)].
	^0! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/3/2018 15:41'!
genJumpBackTo: targetPC
	| target |
	"On first pass install register allocations (if any) as of the end of the loop and back up to recompile.
	 One the second pass generate
				(any merge other than self elided because register assignments copied to loop head in first pass)
				cmp stackLimit
				maybe reload self
				jumpAboveOrEqual target
				flush
				checkForInterrupts
				merge from flushed (N.B. If stack was flushed before loop we could conceivably jump to the pre-loop merge code)
				jmp target
	 self printSimStack; printSimStack: target mergeSimStack"
	self assert: targetPC < bytecodePC.
	target := self fixupAt: targetPC.
	self ensureRegisterAssignmentsAreAtHeadOfLoop: target.
	self copySimStackToScratch: simSpillBase.
	self mergeCurrentSimStackWith: target.
	self MoveAw: coInterpreter stackLimitAddress R: TempReg.
	self CmpR: TempReg R: SPReg. "N.B. FLAGS := SPReg - TempReg"
	self JumpAboveOrEqual: target.

	self ssFlushTo: simStackPtr.
	self CallRT: ceCheckForInterruptTrampoline.
	self annotateBytecode: self Label.
	self restoreSimStackFromScratch.
	self flushLiveRegistersForSuspensionPoint.
	self mergeCurrentSimStackWith: target.
	self Jump: target.
	deadCode := true. "can't fall through"
	^0! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 7/24/2018 09:02'!
genJumpIf: boolean to: targetBytecodePC
	<inline: false>
	| eventualTarget desc reg fixup ok mbb |
	<var: #fixup type: #'BytecodeFixup *'>
	<var: #ok type: #'AbstractInstruction *'>
	<var: #desc type: #'CogSimStackEntry *'>
	<var: #mbb type: #'AbstractInstruction *'>
	eventualTarget := self eventualTargetOf: targetBytecodePC.
	desc := self ssTop.
	self ssPop: 1.

	extA := 0.

	self moveVolatileSimStackEntriesToRegisters.

	(self stackEntryIsBoolean: desc) ifTrue:
		["Must annotate the bytecode for correct pc mapping."
		 desc constant = boolean
			ifTrue:
				[deadCode := true. "Can't fall through."
				 fixup := self ensureFixupAt: eventualTarget.
				 self annotateBytecode: (self Jump: fixup)]
		 	ifFalse:
				[self annotateBytecode: (self prevInstIsPCAnnotated
												ifTrue: [self Nop]
												ifFalse: [self Label])].
		 ^0].

	"try and use the top entry's register if any, but only if it can be destroyed."
	reg := (desc type ~= SSRegister
			or: [(self anyReferencesToRegister: desc register inAllButTopNItems: 0)
			or: [(desc register = ReceiverResultReg and: [self receiverIsInReceiverResultReg])]])
				ifTrue: [TempReg]
				ifFalse: [desc register].
	desc popToReg: reg.
	"Cunning trick by LPD.  If true and false are contiguous subtract the smaller.
	 Correct result is either 0 or the distance between them.  If result is not 0 or
	 their distance send mustBeBoolean."
	self assert: (objectMemory objectAfter: objectMemory falseObject) = objectMemory trueObject.

	"Copy simStack before any branching.  If no branch then the stack is as it is here, not after the merge for eventualTarget."
	self copySimStackToScratch: simSpillBase.
	"Merge required; must not generate merge code along untaken branch, so flip the order."
	(self mergeRequiredForJumpTo: eventualTarget)
		ifTrue:
			[self genSubConstant: (boolean = objectMemory trueObject
										ifTrue: [objectMemory falseObject]
										ifFalse: [objectMemory trueObject])
				R: reg.
			 ok := self JumpZero: 0.
			 self CmpCq: (boolean = objectMemory trueObject
							ifTrue: [objectMemory trueObject - objectMemory falseObject]
							ifFalse: [objectMemory falseObject - objectMemory trueObject])
				R: reg.
			 mbb := self JumpNonZero: 0.
			 self Jump: (self ensureFixupAt: eventualTarget). "generates merge code"
			 mbb jmpTarget: self Label]
		ifFalse:
			[self genSubConstant: boolean R: reg.
			 self JumpZero: (self ensureFixupAt: eventualTarget).
			 self CmpCq: (boolean = objectMemory falseObject
							ifTrue: [objectMemory trueObject - objectMemory falseObject]
							ifFalse: [objectMemory falseObject - objectMemory trueObject])
				R: reg.
			 ok := self JumpZero: 0].
	"Restore simStack after the merges in ensureFixupAt: above."
	self restoreSimStackFromScratch.
	reg ~= TempReg ifTrue:
		[self MoveR: reg R: TempReg].
	self ssFlushTo: simStackPtr.
	"In the RegisterAllocatingCogit we map to an interpreter frame in ceSendMustBeBooleanTo:interpretingAtDelta:"
	self genCallMustBeBooleanFor: boolean.
	"NOTREACHED"
	ok jmpTarget: (self annotateBytecode: self Label).
	self restoreSimStackFromScratch.
	^0! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/8/2018 13:43'!
genJumpTo: targetBytecodePC
	"Overriden to avoid the flush because in this cogit stack state is merged at merge point."
	| eventualTarget generator fixup |
	self moveVolatileSimStackEntriesToRegisters.
	eventualTarget := self eventualTargetOf: targetBytecodePC.
	(eventualTarget > bytecodePC
	 and: [self stackTopIsBoolean
	 and: [(generator := self generatorForPC: eventualTarget) isConditionalBranch]])
		ifTrue:
			[eventualTarget := eventualTarget
							  + generator numBytes
							  + (generator isBranchTrue == (self ssTop constant = objectMemory trueObject)
									ifTrue: [self spanFor: generator at: eventualTarget exts: 0 in: methodObj]
									ifFalse: [0]).
			self ssPop: 1.
			fixup := self ensureFixupAt: eventualTarget.
			self ssPop: -1]
		ifFalse:
			[fixup := self ensureFixupAt: eventualTarget].
	deadCode := true. "can't fall through"
	self Jump: fixup.
	^0! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'cb 4/16/2016 11:58'!
genMarshalledSend: selectorIndex numArgs: numArgs sendTable: sendTable
	self flushLiveRegistersForSend.
	^super genMarshalledSend: selectorIndex numArgs: numArgs sendTable: sendTable! !

!RegisterAllocatingCogit methodsFor: 'initialization' stamp: 'eem 12/7/2016 14:46'!
genMustBeBooleanLongTrampolineFor: boolean called: trampolineName
	<inline: true>
	^self genMustBeBooleanTrampolineFor: boolean branchBytes: 2 called: trampolineName! !

!RegisterAllocatingCogit methodsFor: 'initialization' stamp: 'eem 12/7/2016 13:02'!
genMustBeBooleanTrampolineFor: boolean branchBytes: branchBytes called: trampolineName
	<var: #trampolineName type: #'char *'>
	"For RegisterAllocatingCogit we want the address following a conditional branch not to be reachable, so we
	 don't have to generate code to reload registers.  Instead simply convert to an interpreter frame."
	<inline: false>
	self zeroOpcodeIndex.
	"If the objectRepresentation does want true & false to be mobile then we need to record these addresses."
	self assert: (objectRepresentation shouldAnnotateObjectReference: boolean) not.
	self AddCq: boolean R: TempReg.
	^self genTrampolineFor: #ceSendMustBeBooleanTo:interpretingAtDelta:
		called: trampolineName
		numArgs: 2
		arg: TempReg
		arg: (self trampolineArgConstant: branchBytes)
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: true! !

!RegisterAllocatingCogit methodsFor: 'initialization' stamp: 'eem 12/7/2016 14:45'!
genMustBeBooleanTrampolineFor: boolean called: trampolineName
	<inline: true>
	^self genMustBeBooleanTrampolineFor: boolean branchBytes: 1 called: trampolineName! !

!RegisterAllocatingCogit methodsFor: 'bytecode generators' stamp: 'eem 3/6/2018 17:38'!
genPushReceiverBytecode
	^self ssPushDesc: self ssSelfDescriptor! !

!RegisterAllocatingCogit methodsFor: 'bytecode generators' stamp: 'eem 1/25/2018 17:09'!
genSpecialSelectorArithmetic
	| primDescriptor rcvrIsConst argIsConst rcvrIsInt argIsInt rcvrInt argInt destReg
	 jumpNotSmallInts jumpContinue jumpOverflow index rcvrReg argReg regMask |
	<var: #jumpOverflow type: #'AbstractInstruction *'>
	<var: #jumpContinue type: #'AbstractInstruction *'>
	<var: #primDescriptor type: #'BytecodeDescriptor *'>
	<var: #jumpNotSmallInts type: #'AbstractInstruction *'>
	primDescriptor := self generatorAt: byte0.
	argIsInt := (argIsConst := self ssTop type = SSConstant)
				 and: [objectMemory isIntegerObject: (argInt := self ssTop constant)].
	rcvrIsInt := ((rcvrIsConst := (self ssValue: 1) type = SSConstant)
				  and: [objectMemory isIntegerObject: (rcvrInt := (self ssValue: 1) constant)])
				or: [self mclassIsSmallInteger and: [(self ssValue: 1) isSameEntryAs: self simSelf]].

	(argIsInt and: [rcvrIsInt and: [rcvrIsConst]]) ifTrue:
		[| result |
		 rcvrInt := objectMemory integerValueOf: rcvrInt.
		 argInt := objectMemory integerValueOf: argInt.
		 primDescriptor opcode caseOf: {
			[AddRR]	-> [result := rcvrInt + argInt].
			[SubRR]	-> [result := rcvrInt - argInt].
			[AndRR]	-> [result := rcvrInt bitAnd: argInt].
			[OrRR]		-> [result := rcvrInt bitOr: argInt] }.
		(objectMemory isIntegerValue: result) ifTrue:
			["Must annotate the bytecode for correct pc mapping."
			^self ssPop: 2; ssPushAnnotatedConstant: (objectMemory integerObjectOf: result)].
		^self genSpecialSelectorSend].

	"If there's any constant involved other than a SmallInteger don't attempt to inline."
	((rcvrIsConst and: [rcvrIsInt not])
	 or: [argIsConst and: [argIsInt not]]) ifTrue:
		[^self genSpecialSelectorSend].

	"If we know nothing about the types then better not to inline as the inline cache and
	 primitive code is not terribly slow so wasting time on duplicating tag tests is pointless."
	(argIsInt or: [rcvrIsInt]) ifFalse:
		[^self genSpecialSelectorSend].

	"Since one or other of the arguments is an integer we can very likely profit from inlining.
	 But if the other type is not SmallInteger or if the operation overflows then we will need
	 to do a send.  Since we're allocating values in registers we would like to keep those
	 registers live on the inlined path and reload registers along the non-inlined send path.
	 See reconcileRegisterStateForJoinAfterSpecialSelectorSend below."
	argIsInt
		ifTrue:
			[rcvrReg := self allocateRegForStackEntryAt: 1.
			 (self ssValue: 1) popToReg: rcvrReg.
			 regMask := self registerMaskFor: rcvrReg]
		ifFalse:
			[self allocateRegForStackTopTwoEntriesInto: [:rTop :rNext| argReg := rTop. rcvrReg := rNext].
			 self ssTop popToReg: argReg.
			 (self ssValue: 1) popToReg: rcvrReg.
			 regMask := self registerMaskFor: rcvrReg and: argReg].

	"rcvrReg can be reused for the result iff the receiver is a constant or is an SSRegister that is not used elsewhere."
	destReg := ((rcvrIsInt and: [rcvrIsConst])
				 or: [(self ssValue: 1) type = SSRegister
					 and: [(self anyReferencesToRegister: rcvrReg inAllButTopNItems: 2) not]])
					ifTrue: [rcvrReg]
					ifFalse: [self allocateRegNotConflictingWith: regMask].
	self ssPop: 2.
	jumpNotSmallInts := (rcvrIsInt and: [argIsInt]) ifFalse:
							[argIsInt
								ifTrue: [objectRepresentation genJumpNotSmallInteger: rcvrReg]
								ifFalse:
									[rcvrIsInt
										ifTrue: [objectRepresentation genJumpNotSmallInteger: argReg]
										ifFalse: [objectRepresentation genJumpNotSmallIntegersIn: rcvrReg and: argReg scratch: TempReg]]].
	rcvrReg ~= destReg ifTrue:
		[self MoveR: rcvrReg R: destReg].
	primDescriptor opcode caseOf: {
		[AddRR] -> [argIsInt
						ifTrue:
							[self AddCq: argInt - ConstZero R: destReg.
							 jumpContinue := self JumpNoOverflow: 0.
							 "overflow; must undo the damage before doing send"
							 rcvrReg = destReg ifTrue:
								[self SubbCq: argInt - ConstZero R: rcvrReg]]
						ifFalse:
							[objectRepresentation genRemoveSmallIntegerTagsInScratchReg: destReg.
							 self AddR: argReg R: destReg.
							 jumpContinue := self JumpNoOverflow: 0.
							"overflow; must undo the damage before doing send"
							 destReg = rcvrReg ifTrue:
								[(rcvrIsInt and: [rcvrIsConst])
									ifTrue: [self MoveCq: rcvrInt R: rcvrReg]
									ifFalse:
										[self SubbR: argReg R: rcvrReg.
										 objectRepresentation genSetSmallIntegerTagsIn: rcvrReg]]]].
		[SubRR] -> [argIsInt
						ifTrue:
							[self SubCq: argInt - ConstZero R: destReg.
							 jumpContinue := self JumpNoOverflow: 0.
							 "overflow; must undo the damage before doing send"
							 rcvrReg = destReg ifTrue:
								[self AddcCq: argInt - ConstZero R: rcvrReg]]
						ifFalse:
							[(self anyReferencesToRegister: argReg inAllButTopNItems: 0)
								ifTrue: "argReg is live; cannot strip tags and continue on no overflow without restoring tags"
									[objectRepresentation genRemoveSmallIntegerTagsInScratchReg: argReg.
									 self SubR: argReg R: destReg.
									 jumpOverflow := self JumpOverflow: 0.
									 "no overflow; must undo the damage before continuing"
									 objectRepresentation genSetSmallIntegerTagsIn: argReg.
									 jumpContinue := self Jump: 0.
									 jumpOverflow jmpTarget: self Label.
									 "overflow; must undo the damage before doing send"
									 ((rcvrIsInt and: [rcvrIsConst]) or: [destReg ~= rcvrReg]) ifFalse:
										[self AddcR: argReg R: destReg].
									 objectRepresentation genSetSmallIntegerTagsIn: argReg]
								ifFalse:
									[objectRepresentation genRemoveSmallIntegerTagsInScratchReg: argReg.
									 self SubR: argReg R: destReg.
									 jumpContinue := self JumpNoOverflow: 0.
									 "overflow; must undo the damage before doing send"
									 ((rcvrIsInt and: [rcvrIsConst]) or: [destReg ~= rcvrReg]) ifFalse:
										[self AddcR: argReg R: rcvrReg].
									 objectRepresentation genSetSmallIntegerTagsIn: argReg]]].
		[AndRR] -> [argIsInt
						ifTrue: [self AndCq: argInt R: destReg]
						ifFalse: [self AndR: argReg R: destReg].
					jumpContinue := jumpNotSmallInts ifNotNil: [self Jump: 0]].
		[OrRR]	-> [argIsInt
						ifTrue: [self OrCq: argInt R: destReg]
						ifFalse: [self OrR: argReg R: destReg].
					jumpContinue := jumpNotSmallInts ifNotNil: [self Jump: 0]] }.
	jumpNotSmallInts
		ifNil: [jumpContinue ifNil: "overflow cannot happen"
				[self annotateInstructionForBytecode.
				 self ssPushRegister: destReg.
				 ^0]]
		ifNotNil:
			[jumpNotSmallInts jmpTarget: self Label].
	self ssPushRegister: destReg.
	self copySimStackToScratch: simSpillBase.
	self ssPop: 1.
	self ssFlushTo: simStackPtr.
	rcvrReg = Arg0Reg
		ifTrue:
			[argReg = ReceiverResultReg
				ifTrue: [self SwapR: Arg0Reg R: Arg0Reg Scratch: TempReg. argReg := Arg0Reg]
				ifFalse: [self MoveR: rcvrReg R: ReceiverResultReg].
			 rcvrReg := ReceiverResultReg].
	argIsInt
		ifTrue: [self MoveCq: argInt R: Arg0Reg]
		ifFalse: [argReg ~= Arg0Reg ifTrue: [self MoveR: argReg R: Arg0Reg]].
	rcvrReg ~= ReceiverResultReg ifTrue: [self MoveR: rcvrReg R: ReceiverResultReg].
	index := byte0 - self firstSpecialSelectorBytecodeOffset.
	self genMarshalledSend: index negated - 1 numArgs: 1 sendTable: ordinarySendTrampolines.
	self reconcileRegisterStateForJoinAfterSpecialSelectorSend.
	jumpContinue jmpTarget: self Label.
	^0! !

!RegisterAllocatingCogit methodsFor: 'bytecode generators' stamp: 'cb 9/11/2019 07:50'!
genSpecialSelectorClass
	| topReg destReg scratchReg |
	topReg := self allocateRegForStackEntryAt: 0.
	destReg := self allocateRegNotConflictingWith: (self registerMaskFor: topReg).
	scratchReg := self allocateRegNotConflictingWith: (self registerMaskFor: topReg and: destReg).
	self ssTop popToReg: topReg.
	self asserta: (objectRepresentation
					genGetClassObjectOf: topReg
					into: destReg
					scratchReg: scratchReg
					mayBeAForwarder: self ssTop mayBeAForwarder) ~= BadRegisterSet.
	self ssPop: 1; ssPushRegister: destReg.
	^0! !

!RegisterAllocatingCogit methodsFor: 'bytecode generators' stamp: 'eem 2/8/2018 13:48'!
genSpecialSelectorComparison
	| nextPC postBranchPC targetPC primDescriptor branchDescriptor
	  rcvrIsInt rcvrIsConst argIsIntConst argInt jumpNotSmallInts inlineCAB index rcvrReg argReg branchToTarget needMergeToContinue needMergeToTarget |
	<var: #primDescriptor type: #'BytecodeDescriptor *'>
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>
	<var: #jumpNotSmallInts type: #'AbstractInstruction *'>
	primDescriptor := self generatorAt: byte0.
	argIsIntConst := self ssTop type = SSConstant
				 and: [objectMemory isIntegerObject: (argInt := self ssTop constant)].
	rcvrIsInt := ((rcvrIsConst := (self ssValue: 1) type = SSConstant)
				  and: [objectMemory isIntegerObject: (self ssValue: 1) constant])
				or: [self mclassIsSmallInteger and: [(self ssValue: 1) isSameEntryAs: self simSelf]].

	(argIsIntConst and: [rcvrIsInt and: [rcvrIsConst]]) ifTrue:
		[^self genStaticallyResolvedSpecialSelectorComparison].

	self extractMaybeBranchDescriptorInto: [ :descr :next :postBranch :target | 
		branchDescriptor := descr. nextPC := next. postBranchPC := postBranch. targetPC := target ].

	"Only interested in inlining if followed by a conditional branch."
	inlineCAB := branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse].
	"Further, only interested in inlining = and ~= if there's a SmallInteger constant involved.
	 The relational operators successfully statically predict SmallIntegers; the equality operators do not."
	(inlineCAB and: [primDescriptor opcode = JumpZero or: [primDescriptor opcode = JumpNonZero]]) ifTrue:
		[inlineCAB := argIsIntConst or: [rcvrIsInt]].
	inlineCAB ifFalse:
		[^self genSpecialSelectorSend].

	"In-line the comparison and the jump, but if the types are not SmallInteger then we will need
	 to do a send and fall through to the following conditional branch.  Since we're allocating values
	 in registers we would like to keep those registers live on the inlined path and reload registers
	 along the non-inlined send path.  The merge logic at the branch destinations handles this."
	argIsIntConst
		ifTrue:
			[rcvrReg := self allocateRegForStackEntryAt: 1.
			 (self ssValue: 1) popToReg: rcvrReg.
			 argReg := NoReg]
		ifFalse:
			[self allocateRegForStackTopTwoEntriesInto: [:rTop :rNext| argReg := rTop. rcvrReg := rNext].
			 self ssTop popToReg: argReg.
			 (self ssValue: 1) popToReg: rcvrReg].
	self ssPop: 2.
	self moveVolatileSimStackEntriesToRegistersPreserving:
		(self allocatedRegisters bitOr: (argReg = NoReg
										ifTrue: [self registerMaskFor: rcvrReg]
										ifFalse: [self registerMaskFor: rcvrReg and: argReg])).

	jumpNotSmallInts := (rcvrIsInt and: [argIsIntConst]) ifFalse:
							[argIsIntConst
								ifTrue: [objectRepresentation genJumpNotSmallInteger: rcvrReg]
								ifFalse:
									[rcvrIsInt
										ifTrue: [objectRepresentation genJumpNotSmallInteger: argReg]
										ifFalse: [objectRepresentation genJumpNotSmallIntegersIn: rcvrReg and: argReg scratch: TempReg]]].
	argIsIntConst
		ifTrue: [self CmpCq: argInt R: rcvrReg]
		ifFalse: [self CmpR: argReg R: rcvrReg].

	"self printSimStack; printSimStack: (self fixupAt: postBranchPC) mergeSimStack; printSimStack: (self fixupAt: targetPC) mergeSimStack"
	"If there are merges to be performed on the forward branches we have to execute
	 the merge code only along the path requiring that merge, and exactly once."
	needMergeToTarget := self mergeRequiredForJumpTo: targetPC.
	needMergeToContinue := self mergeRequiredForJumpTo: postBranchPC.
	"Cmp is weird/backwards so invert the comparison."
	(needMergeToTarget and: [needMergeToContinue]) ifTrue:
		[branchToTarget := self genConditionalBranch: (branchDescriptor isBranchTrue
										ifTrue: [primDescriptor opcode]
										ifFalse: [self inverseBranchFor: primDescriptor opcode])
								operand: 0.
		 self Jump: (self ensureFixupAt: postBranchPC).
		 branchToTarget jmpTarget: self Label.
		 self Jump: (self ensureFixupAt: targetPC)].
	(needMergeToTarget and: [needMergeToContinue not]) ifTrue:
		[self genConditionalBranch: (branchDescriptor isBranchFalse
										ifTrue: [primDescriptor opcode]
										ifFalse: [self inverseBranchFor: primDescriptor opcode])
			operand: (self ensureFixupAt: postBranchPC) asUnsignedInteger.
		 self Jump: (self ensureFixupAt: targetPC)].
	(needMergeToTarget not and: [needMergeToContinue]) ifTrue:
		[self genConditionalBranch: (branchDescriptor isBranchTrue
										ifTrue: [primDescriptor opcode]
										ifFalse: [self inverseBranchFor: primDescriptor opcode])
			operand: (self ensureFixupAt: targetPC) asUnsignedInteger.
		 self Jump: (self ensureFixupAt: postBranchPC)].
	(needMergeToTarget or: [needMergeToContinue]) ifFalse:
		[self genConditionalBranch: (branchDescriptor isBranchTrue
										ifTrue: [primDescriptor opcode]
										ifFalse: [self inverseBranchFor: primDescriptor opcode])
			operand: (self ensureFixupAt: targetPC) asUnsignedInteger.
		 self Jump: (self ensureFixupAt: postBranchPC)].
	jumpNotSmallInts ifNil:
		[self annotateInstructionForBytecode.
		 deadCode := true.
		 ^0].
	jumpNotSmallInts jmpTarget: self Label.
	self ssFlushTo: simStackPtr.
	rcvrReg = Arg0Reg
		ifTrue:
			[argReg = ReceiverResultReg
				ifTrue: [self SwapR: Arg0Reg R: Arg0Reg Scratch: TempReg. argReg := Arg0Reg]
				ifFalse: [self MoveR: rcvrReg R: ReceiverResultReg].
			 rcvrReg := ReceiverResultReg].
	argIsIntConst
		ifTrue: [self MoveCq: argInt R: Arg0Reg]
		ifFalse: [argReg ~= Arg0Reg ifTrue: [self MoveR: argReg R: Arg0Reg]].
	rcvrReg ~= ReceiverResultReg ifTrue: [self MoveR: rcvrReg R: ReceiverResultReg].
	index := byte0 - self firstSpecialSelectorBytecodeOffset.
	^self genMarshalledSend: index negated - 1 numArgs: 1 sendTable: ordinarySendTrampolines! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/19/2018 14:41'!
genStorePop: popBoolean TemporaryVariable: tempIndex
	<inline: false>
	| top targetTemp srcRegOrNone destReg |
	self deny: self duplicateRegisterAssignmentsInTemporaries.
	self ssFlushUpThroughTemporaryVariable: tempIndex.
	"To avoid a stall writing through destReg, remember srcReg before the potential ssPop: 1 in ssStorePop:toReg:"
	top := self ssTop.
	targetTemp := self simStackAt: tempIndex + 1.
	srcRegOrNone := top registerOrNone.
	"Take care to avoid duplicating registers in different temps and/or self."
	destReg := targetTemp liveRegister.
	destReg = NoReg ifTrue:
		[destReg := self availableRegOrNoneNotConflictingWith: (self registerMaskUndesirableForTempVars bitOr: self liveRegisters)].
	destReg = NoReg ifTrue:
		[destReg := TempReg].
	self deny: srcRegOrNone = destReg.
	srcRegOrNone = NoReg
		ifTrue: "Avoid duplicating destReg into top; this would only be ok if top
				is the same entry as the target, which is not worth optimizing."
			[self ssStorePopNoAssign: popBoolean toReg: destReg]
		ifFalse:
			[self ssStorePop: popBoolean toReg: destReg].
	(destReg ~= TempReg and: [targetTemp liveRegister = NoReg]) ifTrue:
		[targetTemp liveRegister: destReg.
		 self copyLiveRegisterToCopiesOf: targetTemp].
	self MoveR: (srcRegOrNone ~= NoReg ifTrue: [srcRegOrNone] ifFalse: [destReg])
		Mw: (self frameOffsetOfTemporary: tempIndex)
		r: FPReg.
	targetTemp bcptr: bytecodePC. "for debugging"
	self deny: self duplicateRegisterAssignmentsInTemporaries.
	^0! !

!RegisterAllocatingCogit methodsFor: 'bytecode generators' stamp: 'eem 2/8/2018 13:49'!
genVanillaInlinedIdenticalOrNotIf: orNot
	| nextPC postBranchPC targetBytecodePC branchDescriptor
	  rcvrReg argReg argIsConstant rcvrIsConstant  |
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>
	
	self extractMaybeBranchDescriptorInto: [ :descr :next :postBranch :target | 
		branchDescriptor := descr. nextPC := next. postBranchPC := postBranch. targetBytecodePC := target ].
	
	argIsConstant := self ssTop type = SSConstant.
	"They can't be both constants to use correct machine opcodes.
	 However annotable constants can't be resolved statically, hence we need to careful."
	rcvrIsConstant := argIsConstant not and: [(self ssValue: 1) type = SSConstant].
	
	self 
		allocateEqualsEqualsRegistersArgNeedsReg: argIsConstant not 
		rcvrNeedsReg: rcvrIsConstant not 
		into: [ :rcvr :arg | rcvrReg:= rcvr. argReg := arg ].
	
	"If not followed by a branch, resolve to true or false."
	(branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse]) ifFalse:
		[^ self 
			genIdenticalNoBranchArgIsConstant: argIsConstant 
			rcvrIsConstant: rcvrIsConstant 
			argReg: argReg 
			rcvrReg: rcvrReg 
			orNotIf: orNot].
	
	self ssPop: 2.
	self moveVolatileSimStackEntriesToRegistersPreserving:
		(self allocatedRegisters bitOr: (argReg = NoReg
										ifTrue: [self registerMaskFor: rcvrReg]
										ifFalse: [self registerMaskFor: rcvrReg and: argReg])).

	self genCmpArgIsConstant: argIsConstant rcvrIsConstant: rcvrIsConstant argReg: argReg rcvrReg: rcvrReg.

	"For now just deny we're in the situation we have yet to implement ;-)"
	self deny: (self mergeRequiredForJumpTo: targetBytecodePC).
	self deny: (self mergeRequiredForJumpTo: postBranchPC).

	"Further since there is a following conditional jump bytecode, define
	 non-merge fixups and leave the cond bytecode to set the mergeness."
	(self fixupAt: nextPC) notAFixup
		ifTrue: "The next instruction is dead.  we can skip it."
			[deadCode := true.
		 	 self ensureFixupAt: targetBytecodePC.
			 self ensureFixupAt: postBranchPC]
		ifFalse:
			[self deny: deadCode]. "push dummy value below"
		
	self genConditionalBranch: (orNot == branchDescriptor isBranchTrue ifTrue: [JumpNonZero] ifFalse: [JumpZero])
		operand: (self ensureNonMergeFixupAt: targetBytecodePC) asUnsignedInteger.

	"If the branch is dead, then we can just fall through postBranchPC (only a nop in-between), else 
	we need to jump over the code of the branch"
	deadCode ifFalse:
		[self Jump: (self ensureNonMergeFixupAt: postBranchPC).
		 "duplicate the merge fixup's top of stack so as to avoid a false confict."
		 self ssPushDesc: ((self fixupAt: nextPC) mergeSimStack at: simStackPtr + 1)].

	^0! !

!RegisterAllocatingCogit methodsFor: 'initialization' stamp: 'eem 12/7/2016 14:43'!
generateRunTimeTrampolines
	"Generate the run-time entries at the base of the native code zone and update the base."
	
	ceSendMustBeBooleanAddFalseLongTrampoline := self genMustBeBooleanLongTrampolineFor: objectMemory falseObject
														called: 'ceSendMustBeBooleanAddFalseTrampoline'.
	ceSendMustBeBooleanAddTrueLongTrampoline := self genMustBeBooleanLongTrampolineFor: objectMemory trueObject
														called: 'ceSendMustBeBooleanAddTrueTrampoline'.
	super generateRunTimeTrampolines! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 1/25/2018 14:00'!
initSimStackForFramefulMethod: startpc
	super initSimStackForFramefulMethod: startpc.
	0 to: simStackPtr do:
		[:i| (self simStackAt: i) liveRegister: NoReg]! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 1/25/2018 14:00'!
initSimStackForFramelessBlock: startpc
	super initSimStackForFramelessBlock: startpc.
	self simSelf liveRegister: ReceiverResultReg.
	1 to: simStackPtr do:
		[:i| (self simStackAt: i) liveRegister: NoReg]! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 1/25/2018 14:01'!
initSimStackForFramelessMethod: startpc
	super initSimStackForFramelessMethod: startpc.
	0 to: simStackPtr do:
		[:i| | desc |
		desc := self simStackAt: i.
		desc liveRegister: (desc type = SSRegister ifTrue: [desc register] ifFalse: [NoReg])]! !

!RegisterAllocatingCogit methodsFor: 'initialization' stamp: 'eem 11/22/2016 10:25'!
initializeCodeZoneFrom: startAddress upTo: endAddress
	scratchSimStack := self cCode: [self malloc: self simStackSlots * (self sizeof: CogSimStackEntry)]
							inSmalltalk: [CArrayAccessor on: ((1 to: self simStackSlots) collect: [:ign| CogRegisterAllocatingSimStackEntry new])].
	super initializeCodeZoneFrom: startAddress upTo: endAddress! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 12/1/2016 08:49'!
isAPowerOfTwo: anInteger 
	<inline: true>
	^(anInteger bitAnd: anInteger - 1) = 0! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 4/10/2017 16:51'!
liveRegisters
	| regsSet |
	needsFrame
		ifTrue: [regsSet := 0]
		ifFalse:
			[regsSet := (methodOrBlockNumArgs > self numRegArgs
						  or: [methodOrBlockNumArgs = 0])
							ifTrue:
								[self registerMaskFor: ReceiverResultReg]
							ifFalse:
								[(self numRegArgs > 1 and: [methodOrBlockNumArgs > 1])
									ifFalse: [self registerMaskFor: ReceiverResultReg and: Arg0Reg]
									ifTrue: [self registerMaskFor: ReceiverResultReg and: Arg0Reg and: Arg1Reg]]].
	0 to: simStackPtr do:
		[:i|
		regsSet := regsSet bitOr: (self simStackAt: i) registerMask].
	LowcodeVM ifTrue:
		[(simNativeSpillBase max: 0) to: simNativeStackPtr do:
			[:i|
			regsSet := regsSet bitOr: (self simNativeStackAt: i) nativeRegisterMask]].
	^regsSet! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 1/28/2018 19:45'!
liveRegistersExceptingTopNItems: n in: aSimStack
	<var: 'aSimStack' type: #'SimStackEntry *'>
	^self liveRegistersFrom: 0 to: simStackPtr - n in: aSimStack! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 1/28/2018 16:59'!
liveRegistersFrom: low to: high in: aSimStack
	<var: 'aSimStack' type: #'SimStackEntry *'>
	<inline: true>
	| regsSet |
	regsSet := 0.
	low to: high do:
		[:i|
		regsSet := regsSet bitOr: (self simStack: aSimStack at: i) registerMask].
	^regsSet! !

!RegisterAllocatingCogit methodsFor: 'debugging' stamp: 'eem 2/6/2018 10:37'!
liveRegistersInSelfAndTemps
	^self liveRegistersFrom: 0 to: methodOrBlockNumTemps in: simStack! !

!RegisterAllocatingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/29/2016 13:43'!
maybeCountFixup: descriptor
	"Count needed fixups; descriptor is known to be a branch or a block creation."
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<inline: true>
	numFixups := numFixups + ((descriptor isBranchTrue or: [descriptor isBranchFalse])
									ifTrue: [2]
									ifFalse: [1])! !

!RegisterAllocatingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/1/2016 15:11'!
maybeInitNumFixups
	<inline: true>
	numFixups := 0! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/19/2018 17:08'!
mergeCurrentSimStackWith: fixup
	"At a merge point the cogit expects the stack to be in the same state as fixup's mergeSimStack.
	 mergeSimStack is the state as of some jump forward or backward to this point.  So make
	 simStack agree with mergeSimStack (it is, um, problematic to plant code at the jump). Values
	 may have to be assigned to registers.  Registers may have to be swapped.
	 Generate code to merge the current simStack with that of the target fixup, the goal being to
	 keep as many registers live as possible."
	"self printSimStack; printSimStack: fixup mergeSimStack"
	"self simStackPrintString-> fixup simStackPrintString"
	"abstractOpcodes object copyFrom: startIndex to: opcodeIndex"
	<var: #fixup type: #'BytecodeFixup *'>
	| currentRegisters targetRegisters mergeSimStack current target spillOffset them |
	(mergeSimStack := fixup mergeSimStack) ifNil: [^self].
	self cCode: '' inSmalltalk: [them := {self simStackPrintString. fixup simStackPrintString}].
	self assert: simStackPtr = fixup simStackPtr.
	currentRegisters := self liveRegistersFrom: 0 to: simStackPtr in: simStack.
	targetRegisters := self liveRegistersFrom: 0 to: simStackPtr in: mergeSimStack.
	self resolveConflicts: (currentRegisters bitAnd: targetRegisters) with: fixup mergeSimStack to: fixup simStackPtr.
	self assert: (self conflictsResolvedBetweenSimStackAnd: mergeSimStack).
	(self pushForMergeWith: mergeSimStack)
		ifTrue:
			[0 to: simStackPtr do:
				[:i|
				 spillOffset := i > methodOrBlockNumTemps
									ifTrue: [self frameOffsetOfTemporary: i - 1]
									ifFalse: [0].
				 ((current := self simStack: simStack at: i)
					reconcileWith: (target := self simStack: mergeSimStack at: i)
					spillOffset: spillOffset
					onSpillOrUnspill:
						[:targetReg|
						 self deny: current spilled.
						 self assert: spillOffset ~= 0.
						 current ensureSpilledAt: spillOffset from: FPReg.
						 simSpillBase <= i ifTrue:
							[simSpillBase := i + 1]]) ifTrue:
					[| targetReg |
					 (i > methodOrBlockNumTemps and: [(targetReg := target registerOrNone) ~= NoReg]) ifTrue:
						[self deassignRegister: targetReg in: simStack.
						 self deassignRegister: targetReg in: mergeSimStack.
						 self deny: (self register: targetReg isInMask: self liveRegistersInSelfAndTemps)]]]]
		ifFalse:
			[simStackPtr to: 0 by: -1 do:
				[:i|
				 spillOffset := i > methodOrBlockNumTemps
									ifTrue: [self frameOffsetOfTemporary: i - 1]
									ifFalse: [0].
				 ((current := self simStack: simStack at: i)
					reconcileWith: (target := self simStack: mergeSimStack at: i)
					spillOffset: spillOffset
					onSpillOrUnspill:
						[:targetReg|
						 self assert: current spilled.
						 self assert: spillOffset ~= 0.
						 targetReg  ~= NoReg
							ifTrue: [self PopR: targetReg]
							ifFalse: [self AddCq: objectRepresentation wordSize R: SPReg].
						 current spilled: false.
						 simSpillBase > i ifTrue:
							[simSpillBase := i]]) ifTrue:
					[| targetReg |
					 (i > methodOrBlockNumTemps and: [(targetReg := target registerOrNone) ~= NoReg]) ifTrue:
						[self deassignRegister: targetReg in: simStack.
						 self deassignRegister: targetReg in: mergeSimStack.
						 self deny: (self register: targetReg isInMask: self liveRegistersInSelfAndTemps)]]]].
	self updateSimSpillBase! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 4/20/2017 14:44'!
mergeRequiredForJumpTo: targetPC
	"While this is a multi-pass compiler, no intermediate control-flow graph is built from bytecode and
	 there is a monotonically increasing one-to-one relationship between bytecode pcs and machine
	 code pcs that map to one another.  Therefore, when jumping forward, any required code to merge
	 the state of the current simStack with that at the target must be generated before the jump
	 (because at the target the simStack state will be whatever falls through). If only one forward jump
	 to the target exists then that jump can simply install its simStack as the required simStack at the
	 target and the merge code wil be generated just before the target as control falls through.  But if
	 there are two or more forward jumps to the target, a situation that occurs given that the
	 StackToRegisterMappingCogit follows jump chains, then jumps other than the first must generate
	 merge code before jumping.  This poses a problem for conditional branches.  The merge code must
	 only be generated along the path that takes the jump  Therefore this must *not* be generated:

			... merge code ...
			jump cond Ltarget

	 which incorrectly executes the merge code along both the taken and untaken paths.  Instead
	 this must be generated so that the merge code is only executed if the branch is taken.

			jump not cond Lcontinue
			... merge code ...
			jump Ltarget
		Lcontinue:

	 Note that no merge code is required for code such as self at: (expr ifTrue: [1] ifFalse: [2])
		17 <70> self
		18 <71> pushConstant: true
		19 <99> jumpFalse: 22
		20 <76> pushConstant: 1
		21 <90> jumpTo: 23
		22 <77> pushConstant: 2
		23 <C0> send: at:
	 provided that 1 and 2 are assigned to the same target register."
	| fixup |
	(fixup := self fixupAt: targetPC) hasMergeSimStack ifFalse:
		[^false].
	self assert: (simStackPtr = fixup simStackPtr or: [fixup isBackwardBranchFixup]).
	^self mergeRequiredToTarget: fixup mergeSimStack! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/11/2017 12:45'!
mergeRequiredToTarget: targetSimStack
	<var: 'targetSimStack' type: #'SimStackEntry *'>
	<inline: true>
	<var: 'here' type: #'SimStackEntry *'>
	<var: 'there' type: #'SimStackEntry *'>
	simStackPtr to: 0 by: -1 do:
		[:i| | here there |
		 here := self simStack: simStack at: i.
		 there := self simStack: targetSimStack at: i.
		 (here isMergedWithTargetEntry: there) ifFalse:
			[^true]].
	^false! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 2/9/2018 19:35'!
mergeWithFixupIfRequired: fixup
	"If this bytecode has a fixup, some kind of merge needs to be done. There are 4 cases:
		1) the bytecode has no fixup (fixup isNotAFixup)
			do nothing
		2) the bytecode has a non merge fixup
			the fixup has needsNonMergeFixup.
			The code generating non merge fixup (currently only special selector code) is responsible
				for the merge so no need to do it.
			We set deadCode to false as the instruction can be reached from jumps.
		3) the bytecode has a merge fixup, but execution flow *cannot* fall through to the merge point.
			the fixup has needsMergeFixup and deadCode = true.
			ignores the current simStack as it does not mean anything 
			restores the simStack to the state the jumps to the merge point expects it to be.
		4) the bytecode has a merge fixup and execution flow *can* fall through to the merge point.
			the fixup has needsMergeFixup and deadCode = false.
			Merge the state into the fixup's state via mergeCurrentSimStackWith:forwards:.
			
	In addition, if this is a backjump merge point, we patch the fixup to hold the current simStackPtr 
	for later assertions. self printSimStack: fixup mergeSimStack"

	<var: #fixup type: #'BytecodeFixup *'>
	deadCode ifFalse:
		[self assertCorrectSimStackPtr].

	"case 1"
	fixup notAFixup ifTrue:
		[^0].

	"case 2"
	fixup isNonMergeFixup ifTrue:
		[deadCode
			ifTrue:
				[self deny: fixup simStackPtr isNil.
				 simStackPtr := fixup simStackPtr.
				 self restoreSimStackAtMergePoint: fixup.
				 deadCode := false.
				 self assertCorrectSimStackPtr]
			ifFalse:
				[self flushRegistersOnlyLiveOnFallThrough: fixup].
		 ^0].

	"cases 3 and 4"
	self assert: fixup isMergeFixup.
	self traceMerge: fixup.
	deadCode 
		ifTrue: "case 3"
			[(fixup isBackwardBranchFixup and: [fixup mergeSimStack isNil]) "e.g. a loop within false ifTrue: []"
				ifTrue: [self assert: fixup simStackPtr isNil]
				ifFalse:
					[simStackPtr := fixup simStackPtr.
					 self restoreSimStackAtMergePoint: fixup.
					 deadCode := false]]
		ifFalse: "case 4"
			[(fixup isBackwardBranchFixup and: [compilationPass > 1])
				ifTrue:
					[fixup simStackPtr: simStackPtr.
					 self mergeCurrentSimStackWith: fixup.
					 self copySimStackToFixup: fixup]
				ifFalse:
					[self mergeCurrentSimStackWith: fixup]].
	"cases 3 and 4"
	fixup isBackwardBranchFixup ifTrue:
		[fixup mergeSimStack ifNil:
			[self assert: compilationPass = 1.
			 self setMergeSimStackOf: fixup]].
	fixup targetInstruction: self Label.
	self assertCorrectSimStackPtr.
	self assert: (self simStackMergeCompatibleWith: fixup).
	"self simStackPrintString, fixup simStackPrintString"
	^0! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 7/24/2018 17:54'!
moveVolatileSimStackEntriesToRegisters
	<inline: true>
	self moveVolatileSimStackEntriesToRegistersPreserving: self allocatedRegisters! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/8/2018 13:45'!
moveVolatileSimStackEntriesToRegistersPreserving: registerSet
	"When jumping forward to a merge point the stack must be reconcilable with the state that falls through to the merge point.
	 We cannot easily arrange that later we add code to the branch, e.g. to spill values.  Instead, any volatile contents must be
	 moved to registers.
		[In fact, that's not exactly true, consider these two code sequences:
							self at: (expr ifTrue: [1] ifFalse: [2]) put: a
							self at: 1 put: (expr ifTrue: [a] ifFalse: [b])
						 The first one needs 1 saving to a register to reconcile with 2.
						 The second one has 1 on both paths, but we're not clever enough to spot this case yet.
		 First of all, if the constant requires an annotation then it is difficult to deal with.  But if the constant
		 does not require an annotation one way would be for a SimStackEntry for an SSConstant to refer to
		 the loading instruction and then at the merge simply change the loading instruction to a Label if the
		 constant is the same on both branches].
	 Volatile contents are anything not spilled to the stack, because as yet we can only merge registers."
	<inline: true>
	| allocatedRegs |
	<var: #desc type: #'SimStackEntry *'>
	allocatedRegs := registerSet.
	self assert: simSpillBase >= 0.
	simSpillBase to: simStackPtr do: 
		[:i| | desc reg |
		 desc := self simStackAt: i.
		 desc spilled
			ifTrue: [simSpillBase := i]
			ifFalse:
				[desc registerOrNone = NoReg ifTrue:
					[reg := self allocateRegNotConflictingWith: allocatedRegs.
					 reg = NoReg
						ifTrue: [self halt] "have to spill"
						ifFalse:
							[desc storeToReg: reg.
							 allocatedRegs := allocatedRegs bitOr: (self registerMaskFor: reg)]]]].
	self deny: self duplicateRegisterAssignmentsInTemporaries! !

!RegisterAllocatingCogit methodsFor: 'accessing' stamp: 'eem 12/2/2016 16:46'!
needsFrame
	"for asserts"
	<cmacro: '() needsFrame'>
	^needsFrame! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/25/2018 14:19'!
pushForMergeWith: mergeSimStack
	"Answer if values must be pushed from simStack to merge with mergeSimStack, otherwise < 0 (the default)."
	<var: #mergeSimStack type: #'SimStackEntry *'>
	<inline: true>
	simStackPtr to: methodOrBlockNumTemps + 1 by: -1 do:
		[:i|
		 (self simStack: mergeSimStack at: i) spilled ~= (self simStack: simStack at: i) spilled ifTrue:
			[^(self simStack: mergeSimStack at: i) spilled]].
	^false! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/9/2018 18:15'!
putSelfInReceiverResultReg
	"Override to force copying the register to duplicates on stack."
	<inline: true>
	| simSelfHasRegister |
	simSelfHasRegister := self simSelf liveRegister ~= NoReg.
	super putSelfInReceiverResultReg.
	simSelfHasRegister ifTrue:
		[self simSelf liveRegister: ReceiverResultReg.
		 self copyLiveRegisterToCopiesOf: self simSelf]! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/25/2018 18:22'!
receiverRefOnScratchSimStack

	self assert: scratchSpillBase >= 0.
	simStackPtr to: scratchSpillBase by: -1 do:
		[:i|
		 ((self simStack: scratchSimStack at: i) register = ReceiverResultReg
		  and: [(self simStack: scratchSimStack at: i) type = SSBaseOffset]) ifTrue:
			[^true]].
	^false! !

!RegisterAllocatingCogit methodsFor: 'testing' stamp: 'eem 1/25/2018 17:53'!
receiverResultRegIsAssignedToSelfAndNothingElse
	| culprit |
	0 to: simStackPtr do:
		[:i|
		 (self simSelf isSameEntryAs: (self simStackAt: i))
		 ~= ((self simStackAt: i) liveRegister = ReceiverResultReg) ifTrue:
			[culprit := i.
			 ^false]].
	^true! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/6/2018 11:30'!
reconcileRegisterStateForJoinAfterSpecialSelectorSend
	"When the control flow from the inlined special selector code (e.g. add or comparison)
	 joins the control flow from the send, taken when the inlined code fails, we should decide
	 whether to reload any registers known to contain useful values or mark them as dead."
	 
	"Restore the simStack to that in scratchSimStack,
	 popping any spilled state back into allocated registers."
	| current spillOffset target |
	simSpillBase := scratchSpillBase.
	simStackPtr to: 0 by: -1 do:
		[:i|
		 spillOffset := i > methodOrBlockNumTemps
							ifTrue: [self frameOffsetOfTemporary: i - 1]
							ifFalse: [0].
		 (current := self simStack: simStack at: i)
			reconcileWith: (target := self simStack: scratchSimStack at: i)
			spillOffset: spillOffset
			onSpillOrUnspill:
				[:targetReg|
				 self assert: current spilled.
				 self assert: spillOffset ~= 0.
				 targetReg  ~= NoReg
					ifTrue: [self PopR: targetReg]
					ifFalse: [self AddCq: objectRepresentation wordSize R: SPReg]].
		 simStack
			at: i
			put: (self
					cCode: [scratchSimStack at: i]
					inSmalltalk: [(scratchSimStack at: i) copy])]! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 11/28/2016 09:45'!
registerMaskUndesirableForTempVars
	"Answer the mask containing registers to avoid for temporary variables."
	<inline: true>
	^self registerMaskFor: ReceiverResultReg and: ClassReg and: SendNumArgsReg and: TempReg! !

!RegisterAllocatingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 2/7/2018 11:04'!
reinitializeAllButBackwardFixupsFrom: start through: end
	"When a method must be recompiled due to moving a loop's register
	 assignments to the head of a loop, backward fixups must be marked
	 as such, and all but backward fixups must be reinitialized."
	| descriptor nExts pc distance targetPC |
	<var: #descriptor type: #'BytecodeDescriptor *'>
	pc := start.
	nExts := 0.
	[pc <= end] whileTrue:
		[byte0 := (objectMemory fetchByte: pc ofObject: methodObj) + bytecodeSetOffset.
		 descriptor := self generatorAt: byte0.
		 (descriptor isBranch
		  and: [self isBackwardBranch: descriptor at: pc exts: nExts in: methodObj]) ifTrue:
			[distance := self spanFor: descriptor at: pc exts: nExts in: methodObj.
			 targetPC := pc + descriptor numBytes + distance.
			 self initializeFixupAt: targetPC].
		 descriptor isBlockCreation
			ifTrue:
				[distance := self spanFor: descriptor at: pc exts: nExts in: methodObj.
				 pc := pc + descriptor numBytes + distance]
			ifFalse: [pc := pc + descriptor numBytes].
		 nExts := descriptor isExtension ifTrue: [nExts + 1] ifFalse: [0]].
	start to: end do:
		[:i| | fixup |
		 fixup := self fixupAt: i.
		 (fixup notAFixup or: [fixup isBackwardBranchFixup]) ifFalse:
			[fixup reinitialize]]! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 3/8/2017 15:41'!
reinitializeOpcodesFrom: start to: end
	<inline: true>
	start to: end do:
		[:i|
		(self abstractInstructionAt: i) reinitializeOpcode]! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/6/2018 14:11'!
resetSimStack: startPC
	<inline: true>
	simSpillBase := methodOrBlockNumTemps + 1.
	simStackPtr := methodOrBlockNumTemps.
	0 to: simStackPtr do:
		[:i|
		(self simStackAt: i) liveRegister: NoReg.
		self cCode: '' inSmalltalk: [(self simStackAt: i) bcptr: startPC]]! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/20/2018 12:08'!
resolveConflicts: registersInCommon with: mergeSimStack to: simStackPtr
	"registersInCommon is the register mask of registers in use in both the current
	 simStack and the target mergeSimStack. Swap any and all conflicting register uses
	 in registersInCommon, until register uses in simStack agree with mergeSimStack."
	| registerExchanges registerLocations agreements visited initialIndex initialStack |
	"registerLocations records where a register has moved to
	 during an exchange. This allows a single pass of the stack
	 to rename registers, instead of 1/2 N^2; max stack ~ 56"
	<var: 'registerExchanges' declareC: 'int registerExchanges[NumRegisters]'>
	<var: 'registerLocations' declareC: 'int registerLocations[NumRegisters]'>
	self deny: self duplicateRegisterAssignmentsInTemporaries.
	registersInCommon = (self registerMaskFor: FPReg) ifTrue:
		[self assert: (self conflictsResolvedBetweenSimStackAnd: mergeSimStack).
		 ^self].
	self cCode: '' inSmalltalk:
		[registerExchanges := CArrayAccessor on: (Array new: NumRegisters).
		 registerLocations := CArrayAccessor on: (Array new: NumRegisters)].
	0 to: NumRegisters - 1 do:
		[:i| registerExchanges at: i put: i].
	self cCode: '' inSmalltalk:
		[initialIndex := opcodeIndex. initialStack := self simStackPrintString]. "for debugging"
	agreements := visited := 0.
	0 to: simStackPtr do:
		[:i| | currentReg targetReg |
		currentReg := (self simStackAt: i) registerOrNone.
		targetReg := (self simStack: mergeSimStack at: i) registerOrNone.
		(currentReg ~= NoReg
		 and: [targetReg ~= NoReg]) ifTrue:
			[currentReg := registerExchanges at: currentReg.
			 currentReg = targetReg
				ifTrue:
					[(self register: currentReg isInMask: visited) ifFalse:
						[visited := visited bitOr: (self registerMaskFor: currentReg).
						 agreements := agreements bitOr: (self registerMaskFor: currentReg)]]
				ifFalse:
					[((self register: currentReg isInMask: registersInCommon)
					  and: [self register: currentReg isNotInMask: (visited bitOr: agreements)]) ifTrue:
						[| this that |
						 visited := visited bitOr: (self registerMaskFor: currentReg and: targetReg).
						 self SwapR: targetReg R: currentReg Scratch: RISCTempReg.
						 this := registerExchanges at: currentReg.
						 that := registerExchanges at: targetReg.
						 registerExchanges
							at: currentReg put: that;
							at: targetReg put: this]]]].
	(visited := visited bitClear: agreements) = 0 ifTrue:
		[self assert: (self conflictsResolvedBetweenSimStackAnd: mergeSimStack).
		 ^self].
	0 to: NumRegisters - 1 do:
		[:i| registerLocations at: (registerExchanges at: i) put: i].
	0 to: simStackPtr do:
		[:i| | ssEntry reg |
		ssEntry := self simStackAt: i.
		reg := ssEntry registerOrNone.
		(reg ~= NoReg
		 and: [(self register: reg isInMask: registersInCommon)
		 and: [reg ~= (self simStack: mergeSimStack at: i) registerOrNone]]) ifTrue:
			[ssEntry type = SSRegister
				ifTrue: [ssEntry register: (registerLocations at: reg)]
				ifFalse: [ssEntry liveRegister: (registerLocations at: reg)]]].
	self deny: self duplicateRegisterAssignmentsInTemporaries.
	self assert: (self conflictsResolvedBetweenSimStackAnd: mergeSimStack)
	"(initialIndex to: opcodeIndex - 1) collect: [:x| abstractOpcodes at: x]"! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 2/6/2018 14:07'!
restoreSimStackAtMergePoint: fixup
	<inline: true>
	"All the execution paths reaching a merge point expect everything to be spilled
	 on stack and the optStatus is unknown.  If the merge point follows a return, it
	 isn't a merge, but a skip past a return.  If it is a real merge point then throw
	 away all simStack optimization state."

	fixup mergeSimStack ifNotNil:
		[simSpillBase := methodOrBlockNumTemps + 1.
		 0 to: simStackPtr do:
			[:i|
			self cCode: [simStack at: i put: (fixup mergeSimStack at: i)]
				inSmalltalk: [(simStack at: i) copyFrom: (fixup mergeSimStack at: i)]].
		 self updateSimSpillBase].
	^0! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 9/19/2018 17:31'!
restoreSimStackFromScratch
	<inline: true>
	self cCode: [self memcpy: simStack _: scratchSimStack _: self simStackSlots * (self sizeof: CogSimStackEntry)]
		inSmalltalk: [0 to: simStackPtr do:
						[:i|
						simStack at: i put: (scratchSimStack at: i) copy]].
	simSpillBase := scratchSpillBase! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 9/19/2018 17:31'!
setMergeSimStackOf: fixup
	<var: #fixup type: #'BytecodeFixup *'>
	fixup mergeSimStack
		ifNil:
			[self assert: nextFixup <= numFixups.
			 self cCode: [fixup mergeSimStack: mergeSimStacksBase + (nextFixup * self simStackSlots * (self sizeof: CogSimStackEntry))].
			 nextFixup := nextFixup + 1]
		ifNotNil:
			[self assert: fixup simStackPtr = simStackPtr.
			 0 to: simStackPtr do:
				[:i|
				self assert: ((self simStackAt: i) isSameEntryAs: (self addressOf: (fixup mergeSimStack at: i))).
				(self simStackAt: i) liveRegister ~= (self addressOf: (fixup mergeSimStack at: i)) liveRegister ifTrue:
					[(self simStackAt: i) liveRegister: NoReg]]].
	fixup simStackPtr: simStackPtr.
	self cCode: [self memcpy: fixup mergeSimStack _: simStack _: self simStackSlots * (self sizeof: CogSimStackEntry)]
		inSmalltalk: [fixup mergeSimStack: self copySimStack]! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/25/2018 14:03'!
simSelfOnStackInReceiverResultReg
	"For assert checking only."
	methodOrBlockNumTemps + 1 to: simStackPtr do:
		[:i|
		 ((self simSelf isSameEntryAs: (self simStackAt: i))
		  and: [(self simStackAt: i) registerOrNone = ReceiverResultReg]) ifTrue:
			[^true]].
	^false! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 11/22/2016 10:12'!
simStack: stack at: index
	<cmacro: '(stack,index) ((stack) + (index))'>
	<returnTypeC: #'SimStackEntry *'>
	^self addressOf: (stack at: index)! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 2/11/2017 10:41'!
simStack: aSimStack isIdenticalTo: bSimStack
	<var: 'aSimStack' type: #'SimStackEntry *'>
	<var: 'bSimStack' type: #'SimStackEntry *'>
	0 to: simStackPtr do:
		[:i|
		((self simStack: aSimStack at: i) isIdenticalEntryAs: (self simStack: bSimStack at: i)) ifFalse:
			[^false]].
	^true! !

!RegisterAllocatingCogit methodsFor: 'initialization' stamp: 'cb 4/16/2016 10:55'!
simStackEntryClass
	<doNotGenerate>
	^CogRegisterAllocatingSimStackEntry! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/9/2018 13:27'!
simStackMergeCompatibleWith: fixup
	simStackPtr = fixup simStackPtr ifFalse:
		[^false].
	simStackPtr to: 0 by: -1 do:
		[:i| | target current |
		target := self simStack: fixup mergeSimStack at: i.
		current := self simStack: simStack at: i.
		(target isSameEntryAs: current) ifFalse:
			[^current type caseOf: {
				[SSBaseOffset]	-> [(current offset = (self frameOffsetOfTemporary: i - 1)
									and: [current register = FPReg])
										ifTrue: [true] "current has been spilled virtually"
										ifFalse:
											[target type caseOf: {
												[SSBaseOffset]	-> [false].
												[SSSpill]		-> [false].
												[SSConstant]	-> [false].
												[SSRegister]	-> [false] }]].
				[SSSpill]		-> [target type caseOf: {
										[SSBaseOffset]	-> [true].
										[SSSpill]		-> [true].
										[SSConstant]	-> [true].
										[SSRegister]	-> [true] }].
				[SSConstant]	-> [target type caseOf: {
										[SSBaseOffset]	-> [false].
										[SSSpill]		-> [false].
										[SSConstant]	-> [current constant = target constant].
										[SSRegister]	-> [false] }].
				[SSRegister]	-> [target type caseOf: {
										[SSBaseOffset]	-> [current register = target liveRegister].
										[SSSpill]		-> [current register = target liveRegister].
										[SSConstant]	-> [current register = target liveRegister].
										[SSRegister]	-> [current register = target register] }] }]].
	^true! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 2/1/2018 13:55'!
simStackSameAs: fixup 
	"There are no register conflicts between simStack and mergeSimStack if
	 traversing both stacks from hot end (simStackPtr) to cold end (0) no register
	 exists in simStack that has previously existed in mergeSimStack.  This is because
	 the resolution assigns values from simStack to registers in mergeSimStack and so
	 must not assign to a register yet to be read."
	 | |
	simStackPtr = fixup simStackPtr ifFalse:
		[^false].
	simStackPtr to: 0 by: -1 do:
		[:i| | target |
		target := self simStack: fixup mergeSimStack at: i.
		(target isSameEntryAs: (self simStack: simStack at: i)) ifFalse:
			[^false]].
	^true! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 1/25/2018 14:20'!
ssAllocateRequiredRegMask: requiredRegsMask upThrough: stackPtr upThroughNative: nativeStackPtr
	"Override to void any required registers in temp vars."
	0 to: methodOrBlockNumTemps do:
		[:i|
		((self simStackAt: i) registerMask anyMask: requiredRegsMask) ifTrue:
			[(self simStackAt: i) liveRegister: NoReg]].
	super ssAllocateRequiredRegMask: requiredRegsMask upThrough: stackPtr upThroughNative: nativeStackPtr! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 7/24/2018 19:49'!
ssEntriesDo: aBlock
	"Evaluate aBlock with all simStackEntries"
	<inline: true>
	0 to: simStackPtr do:
		[:i| aBlock value: (self simStackAt: i)]! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 7/24/2018 19:34'!
ssEntrySuchThat: aBlock
	"Answer the SimStackEntry for which aBlock answers true, or nil if none."
	<inline: true>
	0 to: simStackPtr do:
		[:i|
		(aBlock value: (self simStackAt: i)) ifTrue:
			[^self simStackAt: i]].
	^nil! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 1/25/2018 18:23'!
ssFlushFrom: start upThrough: unaryBlock
	"Any occurrences on the stack of the value being stored (which is the top of stack)
	 must be flushed, and hence any values colder than them stack."
	<inline: true>
	self assert: simSpillBase >= 0.
	start to: simSpillBase by: -1 do:
		[ :index |
		(unaryBlock value: (self simStackAt: index)) ifTrue: [ ^ self ssFlushTo: index ] ]! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 11/22/2016 10:12'!
ssFlushFrom: start upThroughRegister: reg
	"Any occurrences on the stack of the register must be
	 flushed, and hence any values colder than them stack."
	<var: #desc type: #'SimStackEntry *'>
	self ssFlushFrom: start upThrough: [ :desc | desc type = SSRegister and: [ desc register = reg ] ]! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'cb 4/17/2016 17:13'!
ssPushAnnotatedConstant: literal
	super ssPushAnnotatedConstant: literal.
	self ssTop liveRegister: NoReg.
	^0! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'cb 4/17/2016 17:12'!
ssPushBase: reg offset: offset
	super ssPushBase: reg offset: offset.
	self ssTop liveRegister: NoReg.
	^0! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'cb 4/17/2016 17:10'!
ssPushConstant: literal
	super ssPushConstant: literal.
	self ssTop liveRegister: NoReg.
	^0! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'cb 4/17/2016 17:12'!
ssPushRegister: reg
	super ssPushRegister: reg.
	self ssTop liveRegister: NoReg.
	^0! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 11/30/2016 12:16'!
ssStorePop: popBoolean toPreferredReg: preferredReg
	"Store or pop the top simulated stack entry to a register.
	 Use preferredReg if the entry is not itself a register.
	 Answer the actual register the result ends up in."
	| actualReg liveRegisters |
	actualReg := preferredReg.
	self ssTop type = SSRegister ifTrue: 
		[self assert: (self ssTop liveRegister = NoReg
					  or: [self ssTop liveRegister = self ssTop register]).
		self assert: self ssTop spilled not.
		actualReg := self ssTop register].
	self ssTop liveRegister ~= NoReg ifTrue:
		[actualReg := self ssTop liveRegister].
	liveRegisters := self liveRegistersExceptingTopNItems: 1 in: simStack.
	(self register: actualReg isInMask: liveRegisters) ifTrue:
		[actualReg := self allocateRegNotConflictingWith: (self registerMaskFor: preferredReg).
		 actualReg = NoReg ifTrue:
			[actualReg := preferredReg]].
	self deny: (self register: actualReg isInMask: liveRegisters).
	self ssStorePop: popBoolean toReg: actualReg. "generates nothing if ssTop is already in actualReg"
	^actualReg! !

!RegisterAllocatingCogit methodsFor: 'simulation stack' stamp: 'eem 2/19/2018 13:58'!
ssStorePopNoAssign: popBoolean toReg: reg
	"Store or pop the top simulated stack entry to a register.
	 Do /not/ assign reg to ssTop.
	N.B.: popToReg: and storeToReg: does not generate anything if 
	it moves a register to the same register."	
	popBoolean
		ifTrue: [self ssTop popToRegNoAssign: reg.
				self ssPop: 1]
		ifFalse: [self ssTop storeToRegNoAssign: reg].! !

!RegisterAllocatingCogit methodsFor: 'simulation only' stamp: 'eem 1/25/2018 14:27'!
traceSimStack
	<cmacro: '() 0'>
	(compilationTrace anyMask: 4) ifTrue:
		[self printSimStack]! !

!RegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/25/2018 14:05'!
voidReceiverResultRegContainsSelf
	"Used when ReceiverResultReg is allocated for other than simSelf, and
	 there may be references to ReceiverResultReg which need to be spilled."
	self receiverIsInReceiverResultReg ifFalse:
		[self deny: self simSelfOnStackInReceiverResultReg.
		 ^self].
	1 to: methodOrBlockNumTemps do:
		[:i|
		self deny: (self simStackAt: i) liveRegister = ReceiverResultReg].
	methodOrBlockNumTemps + 1 to: simStackPtr do:
		[:i|
		(self simSelf isSameEntryAs: (self simStackAt: i)) ifTrue:
			[(self simStackAt: i) liveRegister: NoReg]].
	self simSelf liveRegister: NoReg! !

RegisterAllocatingCogit subclass: #SistaRegisterAllocatingCogit
	instanceVariableNames: 'numCounters counters counterIndex initialCounterValue ceTrapTrampoline branchReachedOnlyForCounterTrip'
	classVariableNames: 'CounterBytes MaxCounterValue'
	poolDictionaries: 'VMSqueakClassIndices'
	category: 'VMMaker-JIT'!

!SistaRegisterAllocatingCogit commentStamp: 'eem 12/12/2016 09:18' prior: 0!
SistaRegisterAllocatingCogit is a port of SistaCogit under RegisterAllocatingCogit.  Its subclass SistaCogitClone holds those methods that are identical to SistaCogit's.  This class holds the methods that are different.  SistaRegisterAllocatingCogit's initialize method keeps things up-to-date and arranges that no method implemented in SistaRegisterAllocatingCogit is implemented in SistaCogitClone so that super sends in SistaCogitClone activate code in RegisterAllocatingCogit as intended.

Instance Variables
	branchReachedOnlyForCounterTrip:		<Object>
	ceTrapTrampoline:		<Object>
	counterIndex:		<Object>
	counters:		<Object>
	initialCounterValue:		<Object>
	numCounters:		<Object>

branchReachedOnlyForCounterTrip
	- xxxxx

ceTrapTrampoline
	- xxxxx

counterIndex
	- xxxxx

counters
	- xxxxx

initialCounterValue
	- xxxxx

numCounters
	- xxxxx
!

SistaRegisterAllocatingCogit subclass: #SistaCogitClone
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!SistaCogitClone class methodsFor: 'translation' stamp: 'eem 2/21/2011 17:32'!
additionalHeadersDo: aBinaryBlock
	"Evaluate aBinaryBlock with the names and contents of
	 any additional header files that need to be generated."
	aBinaryBlock
		value: 'cogmethod.h'
		value: SistaCogMethod cogMethodHeader! !

!SistaCogitClone class methodsFor: 'translation' stamp: 'eem 6/7/2018 15:35'!
ancilliaryClasses
	"Answer any extra classes to be included in the translation."
	^super ancilliaryClasses copyWith: SistaCogMethod! !

!SistaCogitClone class methodsFor: 'translation' stamp: 'eem 7/5/2014 09:53'!
declareCVarsIn: aCodeGen
	aCodeGen var: 'counters' type: #usqInt! !

!SistaCogitClone class methodsFor: 'class initialization' stamp: 'eem 4/6/2012 14:55'!
initializeWithOptions: optionsDictionary

	super initializeWithOptions: optionsDictionary.
	CounterBytes := 4.
	MaxCounterValue := (1 << 16) - 1! !

!SistaCogitClone class methodsFor: 'accessing class hierarchy' stamp: 'eem 3/16/2017 16:52'!
methodZoneClass
	^SistaMethodZone! !

!SistaCogitClone class methodsFor: 'trampoline support' stamp: 'eem 6/4/2015 12:08'!
numTrampolines
	^super numTrampolines + 1

	"Cogit withAllSubclasses collect: [:c| {c. (c instVarNames select: [:ea| ea beginsWith: 'ce']) size}]"
	"self instVarNames select: [:ea| ea beginsWith: 'ce']"! !

!SistaCogitClone methodsFor: 'in-line cacheing' stamp: 'cb 12/1/2016 10:15'!
allowEarlyOpenPICPromotion
	<inline: true>
	^ false! !

!SistaCogitClone methodsFor: 'compile abstract instructions' stamp: 'cb 4/15/2016 11:09'!
compileCogFullBlockMethod: numCopied
	<option: #SistaV1BytecodeSet>
	counters := 0.
	^super compileCogFullBlockMethod: numCopied! !

!SistaCogitClone methodsFor: 'compile abstract instructions' stamp: 'eem 7/5/2014 17:21'!
compileCogMethod: selector
	counters := 0.
	^super compileCogMethod: selector! !

!SistaCogitClone methodsFor: 'compile abstract instructions' stamp: 'eem 7/7/2014 07:05'!
compileFrameBuild
	"Override to prefetch counters, if any."
	super compileFrameBuild.
	counters ~= 0 ifTrue:
		[self PrefetchAw: counters]! !

!SistaCogitClone methodsFor: 'compile abstract instructions' stamp: 'cb 4/7/2016 18:32'!
compileFullBlockMethodFrameBuild: numCopied
	"Override to prefetch counters if any"
	super compileFullBlockMethodFrameBuild: numCopied.
	counters ~= 0 ifTrue:
		[self PrefetchAw: counters]! !

!SistaCogitClone methodsFor: 'compile abstract instructions' stamp: 'cb 1/8/2018 16:30'!
decodePushIntegerLongBefore: bcpc in: aMethodObj
	<inline: true>
	| distance upperByte | 
	distance := objectMemory fetchByte: bcpc - 1 ofObject: aMethodObj.
	upperByte := objectMemory fetchByte: bcpc - 3 ofObject: aMethodObj.
	upperByte > 127 ifTrue: [upperByte := upperByte - 256].
	^ (upperByte << 8) + distance.
		 ! !

!SistaCogitClone methodsFor: 'accessing' stamp: 'eem 3/16/2017 16:57'!
defaultCogCodeSize
	"Return the default number of bytes to allocate for native code at startup.
	 The actual value can be set via vmParameterAt: and/or a preference in the ini file."
	<api>
	^2 * backEnd getDefaultCogCodeSize! !

!SistaCogitClone methodsFor: 'disassembly' stamp: 'eem 1/23/2017 13:23'!
disassembleMethod: surrogateOrAddress on: aStream
	<doNotGenerate>
	| cogMethod |
	cogMethod := super disassembleMethod: surrogateOrAddress on: aStream.
	(cogMethod cmType = CMMethod
	 and: [cogMethod counters ~= 0]) ifTrue:
		[aStream nextPutAll: 'counters:'; cr.
		 0 to: (objectRepresentation numCountersFor: cogMethod counters) - 1 do:
			[:i| | addr |
			 addr := i * CounterBytes + counters.
			 addr printOn: aStream base: 16.
			 aStream nextPut: $:; space.
			 (objectMemory long32At: addr) printOn: aStream base: 16.
			 aStream cr].
		 aStream flush]! !

!SistaCogitClone methodsFor: 'accessing' stamp: 'cb 6/22/2015 16:19'!
estimateOfAbstractOpcodesPerBytecodes
	"Due to the counter logic, the estimation is higher"
	<inline: true>
	^ 11! !

!SistaCogitClone methodsFor: 'generate machine code' stamp: 'eem 1/27/2020 13:10'!
fillInCPICHeader: pic numArgs: numArgs numCases: numCases hasMNUCase: hasMNUCase selector: selector
	pic counters: 0.
	super fillInCPICHeader: pic numArgs: numArgs numCases: numCases hasMNUCase: hasMNUCase selector: selector! !

!SistaCogitClone methodsFor: 'generate machine code' stamp: 'eem 9/23/2013 07:23'!
fillInCounters: nCounters atEndAddress: endAddress
	endAddress - (nCounters * CounterBytes)
		to: endAddress - CounterBytes
		by: CounterBytes
		do: [:address|
			objectMemory
				long32At: address
				put: (initialCounterValue << 16 + initialCounterValue)]! !

!SistaCogitClone methodsFor: 'generate machine code' stamp: 'eem 7/4/2014 12:08'!
fillInCounters: nCounters atStartAddress: startAddress
	startAddress
		to: startAddress + (nCounters - 1 * CounterBytes)
		by: CounterBytes
		do: [:address|
			objectMemory
				long32At: address
				put: (initialCounterValue << 16 + initialCounterValue)]! !

!SistaCogitClone methodsFor: 'generate machine code' stamp: 'eem 1/27/2020 13:11'!
fillInMethodHeader: method size: size selector: selector
	super fillInMethodHeader: method size: size selector: selector.
	self fillInCounters: numCounters atStartAddress: counters.
	method counters: counters! !

!SistaCogitClone methodsFor: 'generate machine code' stamp: 'eem 1/27/2020 13:12'!
fillInOPICHeader: pic numArgs: numArgs selector: selector
	pic counters: 0.
	super fillInOPICHeader: pic numArgs: numArgs selector: selector! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 12/18/2017 17:18'!
genAtPutInlinePrimitive: prim
	"The store check requires rr to be ReceiverResultReg"
	prim caseOf: {
		"3000	pointerAt:put:
		 Mutable pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
		[0] ->	[self genPointerAtPutMaybeContext: false storeCheck: false immutabilityCheck: false].
		"3001	storeCheckPointerAt:put:
		 Mutable pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
		[1] ->	[self genPointerAtPutMaybeContext: false storeCheck: true immutabilityCheck: false].
		"3002	maybeContextPointerAt:put:
		 Mutable pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
		[2] ->	[self genPointerAtPutMaybeContext: true storeCheck: false immutabilityCheck: false].
		"3003	maybeContextStoreCheckPointerAt:put:
		 Mutable pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
		[3] ->	[self genPointerAtPutMaybeContext: true storeCheck: true immutabilityCheck: false].
		"3004	byteAt:put:
		 Mutable byte object, Smi, 8 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)"
		[4] ->	[self genByteAtPut].
	}
	otherwise: [^EncounteredUnknownBytecode].
	^0! !

!SistaCogitClone methodsFor: 'inline primitive binary generators' stamp: 'cb 3/15/2018 11:19'!
genBinaryAtConstInlinePrimitive: primIndex
	"2064	pointerAt:
	Pointer object (Fixed sized or not) and not a context, Smi =>  (1-based, optimised if arg1 is a constant)
	2065	maybeContextPointerAt:
	Pointer object (Fixed sized or not), Smi =>  (1-based, optimised if arg1 is a constant)
	2066	byteAt:
	byte object, Smi => 8 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2067	shortAt:
	short object, Smi => 16 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2068	wordAt:
	word object, Smi => 32 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2069	doubleWordAt:
	double word object, Smi => 64 bits unsigned Smi or LargePositiveInteger (1-based, optimised if arg1 is a constant)"
	| rr val zeroBasedIndex |
	val := self ssTop constant.
	rr := primIndex = 65 
		ifFalse: [self allocateRegForStackEntryAt: 1]
		ifTrue: [self ssAllocateRequiredReg: ReceiverResultReg.
				self voidReceiverResultRegContainsSelf.
				ReceiverResultReg].
	(self ssValue: 1) popToReg: rr.
	self ssPop: 2.
	zeroBasedIndex := (objectMemory integerValueOf: val) - 1.
	primIndex caseOf: {
		[64] ->	[objectRepresentation genLoadSlot: zeroBasedIndex sourceReg: rr destReg: rr].
		[65] ->	[self ssAllocateRequiredReg: SendNumArgsReg.
				 ^self genPushMaybeContextSlotIndex: zeroBasedIndex].
		[66] ->	[self MoveMb: zeroBasedIndex + objectMemory baseHeaderSize r: rr R: rr.
				 objectRepresentation genConvertIntegerToSmallIntegerInReg: rr]
	}
	otherwise: [^EncounteredUnknownBytecode].
	^self ssPushRegister: rr! !

!SistaCogitClone methodsFor: 'inline primitive binary generators' stamp: 'cb 12/14/2017 16:59'!
genBinaryAtInlinePrimitive: primIndex
	"2064	pointerAt:
	Pointer object (Fixed sized or not) and not a context, Smi =>  (1-based, optimised if arg1 is a constant)
	2065	maybeContextPointerAt:
	Pointer object (Fixed sized or not), Smi =>  (1-based, optimised if arg1 is a constant)
	2066	byteAt:
	byte object, Smi => 8 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2067	shortAt:
	short object, Smi => 16 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2068	wordAt:
	word object, Smi => 32 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2069	doubleWordAt:
	double word object, Smi => 64 bits unsigned Smi or LargePositiveInteger (1-based, optimised if arg1 is a constant)"
	| ra rr adjust |
	self allocateRegForStackTopTwoEntriesInto: [:rTop :rNext | ra := rTop. rr := rNext ].
	self ssTop popToReg: ra.
	self ssPop: 1.
	self ssTop popToReg: rr.
	self ssPop: 1.
	primIndex caseOf: {
		[64] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1.
				adjust ~= 0 ifTrue: [ self AddCq: adjust R: ra. ]. 
				self MoveXwr: ra R: rr R: rr ].
		[66] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				adjust := objectMemory baseHeaderSize - 1. 
				self AddCq: adjust R: ra.
				self MoveXbr: ra R: rr R: rr.
				objectRepresentation genConvertIntegerToSmallIntegerInReg: rr].
	}
	otherwise: [^EncounteredUnknownBytecode].
	^self ssPushRegister: rr! !

!SistaCogitClone methodsFor: 'inline primitive binary generators' stamp: 'cb 3/1/2018 15:42'!
genBinaryCompInlinePrimitive: primIndex
	"2032	>
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2033	<
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2034	>=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2035	<=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2036	=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2037	~=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2038	rawEqualsEquals:
	not a forwarder, not a forwarder => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2039	rawNotEqualsEquals:
	not a forwarder, not a forwarder => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)"
	|resultReg otherReg opTrue opFalse invertedOpTrue invertedOpFalse|
	self assert: (primIndex between: 32 and: 39).
	primIndex caseOf: {
		[32]	->	[opTrue := JumpGreater. opFalse := JumpLessOrEqual. invertedOpTrue := JumpLess. invertedOpFalse := JumpGreaterOrEqual].
		[33]	->	[opTrue := JumpLess. opFalse := JumpGreaterOrEqual. invertedOpTrue := JumpGreater. invertedOpFalse := JumpLessOrEqual].
		[34]	->	[opTrue := JumpGreaterOrEqual. opFalse := JumpLess. invertedOpTrue := JumpLessOrEqual. invertedOpFalse := JumpGreater].
		[35]	->	[opTrue := JumpLessOrEqual. opFalse := JumpGreater. invertedOpTrue := JumpGreaterOrEqual. invertedOpFalse := JumpLess].
		[36]	->	[opTrue := JumpZero. opFalse := JumpNonZero. invertedOpTrue := JumpZero. invertedOpFalse := JumpNonZero].
		[37]	->	[opTrue := JumpNonZero. opFalse := JumpZero. invertedOpTrue := JumpNonZero. invertedOpFalse := JumpZero].
		[38]	->	[opTrue := JumpZero. opFalse := JumpNonZero. invertedOpTrue := JumpZero. invertedOpFalse := JumpNonZero].
		[39]	->	[opTrue := JumpNonZero. opFalse := JumpZero. invertedOpTrue := JumpNonZero. invertedOpFalse := JumpZero].
	}.
	"Gen comparison"
	self ssTop type = SSConstant 
		ifTrue: [resultReg := self allocateRegForStackEntryAt: 1.
				(self ssValue: 1) popToReg: resultReg.
				self genCmpConstant: self ssTop constant R: resultReg.
				self ssPop: 2.
				self genBinaryInlineComparison: opTrue opFalse: opFalse destReg: resultReg.
				^self ssPushRegister: resultReg].
	(self ssValue: 1) type = SSConstant 
		ifTrue: [resultReg := self allocateRegForStackEntryAt: 0.
				self ssTop popToReg: resultReg.
				self genCmpConstant: (self ssValue: 1) constant R: resultReg.
				self ssPop: 2.
				self genBinaryInlineComparison: invertedOpTrue opFalse: invertedOpFalse destReg: resultReg.
				^self ssPushRegister: resultReg].
	otherReg := self allocateRegForStackEntryAt: 0.	
	self ssTop popToReg: otherReg.
	resultReg := self allocateRegForStackEntryAt: 1 notConflictingWith: (self registerMaskFor: otherReg).	
	(self ssValue: 1) popToReg: resultReg.
	self CmpR: otherReg R: resultReg.
	self ssPop: 2.
	self genBinaryInlineComparison: opTrue opFalse: opFalse destReg: resultReg.
	^self ssPushRegister: resultReg
	
	! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'eem 4/11/2017 15:08'!
genBinaryConstOpVarInlinePrimitive: prim
	"Const op var version of binary inline primitives."
	"SistaV1: 248		11111000 	iiiiiiii		mjjjjjjj		Call Primitive #iiiiiiii + (jjjjjjj * 256) m=1 means inlined primitive, no hard return after execution.
	 See EncoderForSistaV1's class comment and StackInterpreter>>#binaryInlinePrimitive:"
	<option: #SistaVM>
	| ra val untaggedVal adjust |
	ra := self allocateRegForStackEntryAt: 0.
	self ssTop popToReg: ra.
	self ssPop: 1.
	val := self ssTop constant.
	self ssPop: 1.
	untaggedVal := val - objectMemory smallIntegerTag.
	prim caseOf: {
		"0 through 6, +, -, *, /, //, \\, quo:, SmallInteger op SmallInteger => SmallInteger, no overflow"
		[0]	->	[self AddCq: untaggedVal R: ra].
		[1]	->	[self MoveCq: val R: TempReg.
				 self SubR: ra R: TempReg.
				 objectRepresentation genAddSmallIntegerTagsTo: TempReg.
				 self MoveR: TempReg R: ra].
		[2]	->	[objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: ra.
				 self MoveCq: untaggedVal R: TempReg.
				 self MulR: TempReg R: ra.
				 objectRepresentation genSetSmallIntegerTagsIn: ra].

		"2011 through 2015	Variable-sized pointers new, byte new, 16-bit new, 32-bit new, 64-bit new"

		"2016 through 2020, bitAnd:, bitOr:, bitXor, bitShiftLeft:, bitShiftRight:, SmallInteger op SmallInteger => SmallInteger, no overflow"
		[16] -> [ self AndCq: val R: ra ].
		[17] -> [ self OrCq: val R: ra ].
		[18] -> [ self XorCw: untaggedVal R: ra. ].
		[19] -> [ objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				 self MoveCq: untaggedVal R: TempReg.
				 self LogicalShiftLeftR: ra R: TempReg.
				 objectRepresentation genAddSmallIntegerTagsTo: TempReg.
				 self MoveR: TempReg R: ra].
		[20] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				 self MoveCq: untaggedVal R: TempReg.
				 self ArithmeticShiftRightR: ra R: TempReg.
				 objectRepresentation genClearAndSetSmallIntegerTagsIn: TempReg.
				 self MoveR: TempReg R: ra].

		"2032	through 2037, >, <, >=, <=. =, ~=, SmallInteger op SmallInteger => Boolean (flags?? then in jump bytecodes if ssTop is a flags value, just generate the instruction!!!!)"
		"CmpCqR is SubRCq so everything is reversed, but because no CmpRCq things are reversed again and we invert the sense of the jumps."
		[32] -> [ self CmpCq: val R: ra.
				self genBinaryInlineComparison: JumpLess opFalse: JumpGreaterOrEqual destReg: ra ].
		[33] -> [ self CmpCq: val R: ra.
				self genBinaryInlineComparison: JumpGreater opFalse: JumpLessOrEqual destReg: ra ].
		[34] -> [ self CmpCq: val R: ra.
				self genBinaryInlineComparison: JumpLessOrEqual opFalse: JumpGreater destReg: ra ].
		[35] -> [ self CmpCq: val R: ra.
				self genBinaryInlineComparison: JumpGreaterOrEqual opFalse: JumpLess destReg: ra ].
		[36] -> [ self CmpCq: val R: ra.
				self genBinaryInlineComparison: JumpZero opFalse: JumpNonZero destReg: ra ].
		[37] -> [ self CmpCq: val R: ra.
				self genBinaryInlineComparison: JumpNonZero opFalse: JumpZero destReg: ra ].

		"2064	through 2068, Pointer Object>>at:, Byte Object>>at:, Short16 Word Object>>at: LongWord32 Object>>at: Quad64Word Object>>at:. obj op 0-rel SmallInteger => oop"
		[64] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
				adjust ~= 0 ifTrue: [ self AddCq: adjust R: ra. ]. 
				self genMoveConstant: val R: TempReg.
				self MoveXwr: ra R: TempReg R: ra].
		[65] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				adjust := objectMemory baseHeaderSize - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
				self AddCq: adjust R: ra.
				self genMoveConstant: val R: TempReg.
				self MoveXbr: ra R: TempReg R: ra.
				objectRepresentation genConvertIntegerToSmallIntegerInReg: ra]
	}
	otherwise: [^EncounteredUnknownBytecode].
	self ssPushRegister: ra.
	^0! !

!SistaCogitClone methodsFor: 'inline primitive binary generators' stamp: 'cb 12/14/2017 14:27'!
genBinaryConstOpVarSmiInlinePrimitive: primIndex
	| ra val untaggedVal |
	ra := self allocateRegForStackEntryAt: 0.
	self ssTop popToReg: ra.
	self ssPop: 1.
	val := self ssTop constant.
	self ssPop: 1.
	untaggedVal := val - objectMemory smallIntegerTag.
	primIndex caseOf: {
		"2000	+
		Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[0]	->	[self AddCq: untaggedVal R: ra].
		"2001	-
		Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[1]	->	[self MoveCq: val R: TempReg.
				 self SubR: ra R: TempReg.
				 objectRepresentation genAddSmallIntegerTagsTo: TempReg.
				 self MoveR: TempReg R: ra].
		"2002	*
		Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[2]	->	[objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: ra.
				 self MoveCq: untaggedVal R: TempReg.
				 self MulR: TempReg R: ra.
				 objectRepresentation genSetSmallIntegerTagsIn: ra].
		
		"2016	bitAnd:
		 Smi, Smi => Smi (optimised if one operand is a constant)"
		[16] -> [ self AndCq: val R: ra ].
		"2017	bitOr:
		Smi, Smi => Smi (optimised if one operand is a constant)"
		[17] -> [ self OrCq: val R: ra ].
		"2018	bitXor:
		Smi, Smi => Smi (optimised if one operand is a constant)"
		[18] -> [ self XorCw: untaggedVal R: ra. ].
	}
	otherwise: [^EncounteredUnknownBytecode].
	self ssPushRegister: ra.
	^0! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'eem 02/09/2017 09:39'!
genBinaryInlineComparison: opTrue opFalse: opFalse destReg: destReg
	"Inlined comparison. opTrue = jump for true and opFalse = jump for false"
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>
	| nextPC branchDescriptor targetBytecodePC postBranchPC |	
		
	self extractMaybeBranchDescriptorInto: [ :descr :next :postBranch :target | 
		branchDescriptor := descr. nextPC := next. postBranchPC := postBranch. targetBytecodePC := target ].

	(branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse])
		ifTrue: "This is the path where the inlined comparison is followed immediately by a branch"
			[ (self fixupAt: nextPC) notAFixup
				ifTrue: "The next instruction is dead.  we can skip it."
					[deadCode := true.
				 	 self ensureFixupAt: targetBytecodePC.
					 self ensureFixupAt: postBranchPC ]
				ifFalse:
					[self ssPushConstant: objectMemory trueObject]. "dummy value"
			self genConditionalBranch: (branchDescriptor isBranchTrue ifTrue: [opTrue] ifFalse: [opFalse])
				operand: (self ensureNonMergeFixupAt: targetBytecodePC) asUnsignedInteger.
			"We can only elide the jump if the pc after nextPC is the same as postBranchPC.
			 Branch following means it may not be."
			self nextDescriptorExtensionsAndNextPCInto:
				[:iguana1 :iguana2 :iguana3 :followingPC| nextPC := followingPC].
			(deadCode and: [nextPC = postBranchPC]) ifFalse:
				[ self Jump: (self ensureNonMergeFixupAt: postBranchPC) ] ]
		ifFalse: "This is the path where the inlined comparison is *not* followed immediately by a branch"
			[| condJump jump |
			condJump := self genConditionalBranch: opTrue operand: 0.
			self genMoveFalseR: destReg.
	 		jump := self Jump: 0.
			condJump jmpTarget: (self genMoveTrueR: destReg).
			jump jmpTarget: self Label].
	^ 0! !

!SistaCogitClone methodsFor: 'inline primitive binary generators' stamp: 'cb 12/14/2017 14:26'!
genBinaryInlinePrimitive: primIndex
	"Bulk comments: each sub-method has its own comment with the specific case.
	2000	+
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2001	-
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2002	*
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2003	/
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2004	//
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2005	\
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2006	quo:
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2016	bitAnd:
	Smi, Smi => Smi (optimised if one operand is a constant)
	2017	bitOr:
	Smi, Smi => Smi (optimised if one operand is a constant)
	2018	bitXor:
	Smi, Smi => Smi (optimised if one operand is a constant)
	2019	bitShiftLeft:
	Smi greater or equal to 0, Smi greater or equal to 0 => Smi (no overflow, optimised if arg1 is a constant)
	2020	bitShiftRight:
	Smi, Smi greater or equal to 0 => Smi (optimised if arg1 is a constant)
	2032	>
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2033	<
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2034	>=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2035	<=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2036	=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2037	~=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2038	rawEqualsEquals:
	not a forwarder, not a forwarder => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2039	rawNotEqualsEquals:
	not a forwarder, not a forwarder => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2048	rawNew:
	literal which is a fixed-sized behavior, Smi => instance of receiver, fields nilled out (optimised if arg1 is a constant)
	2049	rawNewNoInit:
	literal which is a fixed-sized behavior, Smi => instance of receiver (Fields of returned value contain undefined data, optimised if arg1 is a constant)
	2064	pointerAt:
	Pointer object (Fixed sized or not) and not a context, Smi =>  (1-based, optimised if arg1 is a constant)
	2065	maybeContextPointerAt:
	Pointer object (Fixed sized or not), Smi =>  (1-based, optimised if arg1 is a constant)
	2066	byteAt:
	byte object, Smi => 8 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2067	shortAt:
	short object, Smi => 16 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2068	wordAt:
	word object, Smi => 32 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2069	doubleWordAt:
	double word object, Smi => 64 bits unsigned Smi or LargePositiveInteger (1-based, optimised if arg1 is a constant)"
	(primIndex <= 18 and: [primIndex <= 2 or: [primIndex > 6]])  ifTrue: 
		[self ssTop type = SSConstant 
			ifTrue: [^self genBinaryVarOpConstSmiInlinePrimitive: primIndex].
		(self ssValue: 1) type = SSConstant 
			ifTrue: [^self genBinaryConstOpVarSmiInlinePrimitive: primIndex].
		^self genBinaryVarOpVarSmiInlinePrimitive: primIndex].
	primIndex <= 6 ifTrue: [^self genDivInlinePrimitive: primIndex].
	primIndex = 19 ifTrue: [^self genBinarySmiBitShiftLeftInlinePrimitive].
	primIndex = 20 ifTrue: [^self genBinarySmiBitShiftRightInlinePrimitive].
	primIndex < 32 ifTrue: [^EncounteredUnknownBytecode].
	primIndex <= 39 ifTrue: [^self genBinaryCompInlinePrimitive: primIndex].
	primIndex < 48 ifTrue: [^EncounteredUnknownBytecode].
	primIndex <= 49 ifTrue: [^self genBinaryNewInlinePrimitive: primIndex].
	primIndex < 64 ifTrue: [^EncounteredUnknownBytecode].
	primIndex <= 69 ifTrue: 
		[self ssTop type = SSConstant
			ifTrue: [^self genBinaryAtConstInlinePrimitive: primIndex].
		^self genBinaryAtInlinePrimitive: primIndex].
	^EncounteredUnknownBytecode! !

!SistaCogitClone methodsFor: 'inline primitive binary generators' stamp: 'cb 3/1/2018 17:53'!
genBinaryNewInlinePrimitive: primIndex
	"2048	rawNew:
	literal which is a variable-sized behavior, Smi => instance of receiver, fields nilled/zeroed out (optimised if arg1 is a constant)
	2049	rawNewNoInit:
	literal which is a variable-sized behavior, Smi => instance of receiver (Fields of returned value contain undefined data, optimised if arg1 is a constant)"
	| resultReg classObj classFormat argInt |

	"Assertion"
	(self ssValue: 1) type = SSConstant ifFalse: [^UnimplementedOperation].
	
	"It would be nice to deal with variable-sized allocation on non-constant.
	We need to see, however, than inlined allocation cannot deal with large allocations (i.e.
	allocation directly in old space, > 65k currently. So Scorch still needs to perform range analysis
	or generate 2 paths at the bytecode level in this case."
	self ssTop type = SSConstant ifFalse: [^UnimplementedOperation].
	
	classObj := (self ssValue: 1) constant.
	self assert: (objectMemory isNonImmediate: classObj).
	self assert: (coInterpreter objCouldBeClassObj: classObj).
	self deny: (objectMemory isFixedSizePointerFormat: (objectMemory instSpecOfClassFormat: (objectMemory formatOfClass: classObj))).
	objectMemory classTagForClass: classObj. "Ensure Behavior hash"
	
	"Actual code"
	resultReg := self allocateRegNotConflictingWith: 0.
	classFormat := objectMemory instSpecOfClassFormat: (objectMemory formatOfClass: classObj).
	argInt := objectMemory integerValueOf: self ssTop constant.
	self ssPop: 2.
	self ssPushRegister: resultReg.
	(classFormat = objectMemory arrayFormat or: [classFormat = objectMemory indexablePointersFormat]) ifTrue: 
		[^objectRepresentation genGetInstanceOfPointerClass: classObj into: resultReg initializingIf: primIndex = 48 numVariableSlots: argInt].
	(classFormat = objectMemory firstByteFormat) ifTrue: 
		[^objectRepresentation genGetInstanceOfByteClass: classObj into: resultReg initializingIf: primIndex = 48 numBytes: argInt].
	^UnimplementedOperation! !

!SistaCogitClone methodsFor: 'inline primitive binary generators' stamp: 'cb 12/14/2017 16:50'!
genBinarySmiBitShiftLeftInlinePrimitive
	"2019	bitShiftLeft:
	Smi greater or equal to 0, Smi greater or equal to 0 => Smi (no overflow, optimised if arg1 is a constant)"
	| rr ra |
	rr := self allocateRegForStackEntryAt: 1.
	(self ssValue: 1) popToReg: rr.
	objectRepresentation genRemoveSmallIntegerTagsInScratchReg: rr.
	self ssTop type = SSConstant 
		ifTrue: [self LogicalShiftLeftCq: (objectMemory integerValueOf: self ssTop constant) R: rr]
		ifFalse: [ra := self allocateRegForStackEntryAt: 0 notConflictingWith: (self registerMaskFor: rr).
				self ssTop popToReg: ra.
				objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				self LogicalShiftLeftR: ra R: rr.].
	objectRepresentation genAddSmallIntegerTagsTo: rr.
	self ssPop: 2.
	^self ssPushRegister: rr
! !

!SistaCogitClone methodsFor: 'inline primitive binary generators' stamp: 'cb 12/14/2017 16:50'!
genBinarySmiBitShiftRightInlinePrimitive
	"2019	bitShiftLeft:
	Smi greater or equal to 0, Smi greater or equal to 0 => Smi (no overflow, optimised if arg1 is a constant)"
	| rr ra |
	rr := self allocateRegForStackEntryAt: 1.
	(self ssValue: 1) popToReg: rr.
	self ssTop type = SSConstant 
		ifTrue: [self ArithmeticShiftRightCq: (objectMemory integerValueOf: self ssTop constant) R: rr]
		ifFalse: [ra := self allocateRegForStackEntryAt: 0 notConflictingWith: (self registerMaskFor: rr).
				self ssTop popToReg: ra.
				objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				 self ArithmeticShiftRightR: ra R: rr].
	objectRepresentation genClearAndSetSmallIntegerTagsIn: rr.
	self ssPop: 2.
	^self ssPushRegister: rr! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 1/13/2017 11:53'!
genBinaryVarOpConstInlinePrimitive: prim
	"Var op const version of inline binary inline primitives."
	"SistaV1: 248		11111000 	iiiiiiii		mjjjjjjj		Call Primitive #iiiiiiii + (jjjjjjj * 256) m=1 means inlined primitive, no hard return after execution.
	 See EncoderForSistaV1's class comment and StackInterpreter>>#binaryInlinePrimitive:"
	<option: #SistaVM>
	| rr val untaggedVal |
	val := self ssTop constant.
	self ssPop: 1.
	rr := self allocateRegForStackEntryAt: 0.
	self ssTop popToReg: rr.
	self ssPop: 1.
	untaggedVal := val - objectMemory smallIntegerTag.
	prim caseOf: {
		"0 through 6, +, -, *, /, //, \\, quo:, SmallInteger op SmallInteger => SmallInteger, no overflow"
		[0]	->	[self AddCq: untaggedVal R: rr].
		[1]	->	[self SubCq: untaggedVal R: rr ].
		[2]	->	[self flag: 'could use MulCq:R'.
				 objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: rr.
				 self MoveCq: untaggedVal R: TempReg.
				 self MulR: TempReg R: rr.
				 objectRepresentation genSetSmallIntegerTagsIn: rr].

		"2016 through 2020, bitAnd:, bitOr:, bitXor, bitShiftLeft:, bitShiftRight:, SmallInteger op SmallInteger => SmallInteger, no overflow"
		[16] -> [ self AndCq: val R: rr ].
		[17] -> [ self OrCq: val R: rr ].
		[18] -> [ self flag: 'could use XorCq:'.
				self XorCw: untaggedVal R: rr. ].
		[19] -> [ objectRepresentation genRemoveSmallIntegerTagsInScratchReg: rr.
				 self LogicalShiftLeftCq: (objectMemory integerValueOf: val) R: rr.
				 objectRepresentation genAddSmallIntegerTagsTo: rr ].
		[20] ->	[self ArithmeticShiftRightCq: (objectMemory integerValueOf: val) R: rr.
				 objectRepresentation genClearAndSetSmallIntegerTagsIn: rr].

		"2032	through 2037, >, <, >=, <=. =, ~=, SmallInteger op SmallInteger => Boolean (flags?? then in jump bytecodes if ssTop is a flags value, just generate the instruction!!!!)"
		"CmpCqR is SubRCq so everything is reversed."
		[32] -> [ self CmpCq: val R: rr.
				self genBinaryInlineComparison: JumpGreater opFalse: JumpLessOrEqual destReg: rr ].
		[33] -> [ self CmpCq: val R: rr.
				self genBinaryInlineComparison: JumpLess opFalse: JumpGreaterOrEqual destReg: rr ].
		[34] -> [ self CmpCq: val R: rr.
				self genBinaryInlineComparison: JumpGreaterOrEqual opFalse: JumpLess destReg: rr ].
		[35] -> [ self CmpCq: val R: rr.
				self genBinaryInlineComparison: JumpLessOrEqual opFalse: JumpGreater destReg: rr ].
		[36] -> [ self CmpCq: val R: rr.
				self genBinaryInlineComparison: JumpZero opFalse: JumpNonZero destReg: rr ].
		[37] -> [ self CmpCq: val R: rr.
				self genBinaryInlineComparison: JumpNonZero opFalse: JumpZero destReg: rr ].

		"2064	through 2068, Pointer Object>>at:, Byte Object>>at:, Short16 Word Object>>at: LongWord32 Object>>at: Quad64Word Object>>at:. obj op 0-rel SmallInteger => oop"
		[64] ->	[objectRepresentation genLoadSlot: (objectMemory integerValueOf: val) - 1 sourceReg: rr destReg: rr].
		[65] ->	[self MoveCq: (objectMemory integerValueOf: val) + objectMemory baseHeaderSize - 1 R: TempReg.
				self MoveXbr: TempReg R: rr R: rr.
				objectRepresentation genConvertIntegerToSmallIntegerInReg: rr]

	}
	otherwise: [^EncounteredUnknownBytecode].
	self ssPushRegister: rr.
	^0! !

!SistaCogitClone methodsFor: 'inline primitive binary generators' stamp: 'cb 12/14/2017 14:27'!
genBinaryVarOpConstSmiInlinePrimitive: primIndex
	| rr val untaggedVal |
	self assert: primIndex <= 18.
	val := self ssTop constant.
	self ssPop: 1.
	rr := self allocateRegForStackEntryAt: 0.
	self ssTop popToReg: rr.
	self ssPop: 1.
	untaggedVal := val - objectMemory smallIntegerTag.
	primIndex caseOf: {
		"2000	+
		 Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[0]	->	[self AddCq: untaggedVal R: rr].
		"2001	-
		 Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[1]	->	[self SubCq: untaggedVal R: rr ].
		"2002	*
		 Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[2]	->	[self flag: 'could use MulCq:R'.
				 objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: rr.
				 self MoveCq: untaggedVal R: TempReg.
				 self MulR: TempReg R: rr.
				 objectRepresentation genSetSmallIntegerTagsIn: rr].
		
		"2016	bitAnd:
		 Smi, Smi => Smi (optimised if one operand is a constant)"
		[16] -> [ self AndCq: val R: rr ].
		"2017	bitOr:
		 Smi, Smi => Smi (optimised if one operand is a constant)"
		[17] -> [ self OrCq: val R: rr ].
		"2018	bitXor:
		 Smi, Smi => Smi (optimised if one operand is a constant)"
		[18] -> [ self flag: 'could use XorCq:'.
				self XorCw: untaggedVal R: rr. ]

	}
	otherwise: [^EncounteredUnknownBytecode].
	self ssPushRegister: rr.
	^0! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 3/22/2017 14:47'!
genBinaryVarOpVarInlinePrimitive: prim
	"Var op var version of binary inline primitives."
	"SistaV1: 248		11111000 	iiiiiiii		mjjjjjjj		Call Primitive #iiiiiiii + (jjjjjjj * 256) m=1 means inlined primitive, no hard return after execution.
	 See EncoderForSistaV1's class comment and StackInterpreter>>#binaryInlinePrimitive:"
	<option: #SistaVM>
	| ra rr adjust |
	self allocateRegForStackTopTwoEntriesInto: [:rTop :rNext | ra := rTop. rr := rNext ].
	self ssTop popToReg: ra.
	self ssPop: 1.
	self ssTop popToReg: rr.
	self ssPop: 1.
	prim caseOf: {
		"0 through 6, +, -, *, /, //, \\, quo:, SmallInteger op SmallInteger => SmallInteger, no overflow"
		[0]	->	[objectRepresentation genRemoveSmallIntegerTagsInScratchReg: ra.
				 self AddR: ra R: rr].
		[1]	->	[self SubR: ra R: rr.
				 objectRepresentation genAddSmallIntegerTagsTo: rr].
		[2]	->	[self genShiftAwaySmallIntegerTagsInScratchReg: rr.
				 self genRemoveSmallIntegerTagsInScratchReg: ra.
				 self MulR: ra R: rr.
				 self genSetSmallIntegerTagsIn: rr].
		"[4]	->	[].
		[5]	->	[].
		[6]	->	[]."

		"2016 through 2020, bitAnd:, bitOr:, bitXor, bitShiftLeft:, bitShiftRight:, SmallInteger op SmallInteger => SmallInteger, no overflow"
		[16] -> [ self AndR: ra R: rr ].
		[17] -> [ self OrR: ra R: rr ].
		[18] -> [objectRepresentation genRemoveSmallIntegerTagsInScratchReg: ra. 
				self XorR: ra R: rr. ].
		[19] -> [ objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				 objectRepresentation genRemoveSmallIntegerTagsInScratchReg: rr.
				 self LogicalShiftLeftR: ra R: rr.
				 objectRepresentation genAddSmallIntegerTagsTo: rr].
		[20] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				 self ArithmeticShiftRightR: ra R: rr.
				 objectRepresentation genClearAndSetSmallIntegerTagsIn: rr.].


		"2032	through 2037, >, <, >=, <=. =, ~=, SmallInteger op SmallInteger => Boolean (flags?? then in jump bytecodes if ssTop is a flags value, just generate the instruction!!!!)"
		"CmpCqR is SubRCq so everything is reversed."
		[32] -> [ self CmpR: ra R: rr.
				self genBinaryInlineComparison: JumpGreater opFalse: JumpLessOrEqual destReg: rr ].
		[33] -> [ self CmpR: ra R: rr.
				self genBinaryInlineComparison: JumpLess opFalse: JumpGreaterOrEqual destReg: rr ].
		[34] -> [ self CmpR: ra R: rr.
				self genBinaryInlineComparison: JumpGreaterOrEqual opFalse: JumpLess destReg: rr ].
		[35] -> [ self CmpR: ra R: rr.
				self genBinaryInlineComparison: JumpLessOrEqual opFalse: JumpGreater destReg: rr ].
		[36] -> [ self CmpR: ra R: rr.
				self genBinaryInlineComparison: JumpZero opFalse: JumpNonZero destReg: rr ].
		[37] -> [ self CmpR: ra R: rr.
				self genBinaryInlineComparison: JumpNonZero opFalse: JumpZero destReg: rr ].

		"2064	through 2068, Pointer Object>>at:, Byte Object>>at:, Short16 Word Object>>at: LongWord32 Object>>at: Quad64Word Object>>at:. obj op 0-rel SmallInteger => oop"
		[64] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
				adjust ~= 0 ifTrue: [ self AddCq: adjust R: ra. ]. 
				self MoveXwr: ra R: rr R: rr ].
		[65] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				adjust := objectMemory baseHeaderSize - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
				self AddCq: adjust R: ra.
				self MoveXbr: ra R: rr R: rr.
				objectRepresentation genConvertIntegerToSmallIntegerInReg: rr]

	}
	otherwise: [^EncounteredUnknownBytecode].
	self ssPushRegister: rr.
	^0! !

!SistaCogitClone methodsFor: 'inline primitive binary generators' stamp: 'cb 12/14/2017 14:27'!
genBinaryVarOpVarSmiInlinePrimitive: primIndex
	| ra rr |
	self allocateRegForStackTopTwoEntriesInto: [:rTop :rNext | ra := rTop. rr := rNext ].
	self ssTop popToReg: ra.
	self ssPop: 1.
	self ssTop popToReg: rr.
	self ssPop: 1.
	primIndex caseOf: {
		"2000	+
		Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[0]	->	[objectRepresentation genRemoveSmallIntegerTagsInScratchReg: ra.
				 self AddR: ra R: rr].
		"2001	-
		Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[1]	->	[self SubR: ra R: rr.
				 objectRepresentation genAddSmallIntegerTagsTo: rr].
		"2002	*
		Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[2]	->	[self genShiftAwaySmallIntegerTagsInScratchReg: rr.
				 self genRemoveSmallIntegerTagsInScratchReg: ra.
				 self MulR: ra R: rr.
				 self genSetSmallIntegerTagsIn: rr].
			
		"2016	bitAnd:
		 Smi, Smi => Smi (optimised if one operand is a constant)"
		[16] -> [ self AndR: ra R: rr ].
		"2017	bitOr:
		Smi, Smi => Smi (optimised if one operand is a constant)"
		[17] -> [ self OrR: ra R: rr ].
		"2018	bitXor:
		Smi, Smi => Smi (optimised if one operand is a constant)"
		[18] -> [objectRepresentation genRemoveSmallIntegerTagsInScratchReg: ra. 
				self XorR: ra R: rr. ]
	}
	otherwise: [^EncounteredUnknownBytecode].
	self ssPushRegister: rr.
	^0! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 12/14/2017 17:00'!
genByteAtPut
	| ra1 ra2 rr adjust |
	self 
		allocateRegForStackTopThreeEntriesInto: [:rTop :rNext :rThird | ra2 := rTop. ra1 := rNext. rr := rThird ] 
		thirdIsReceiver: false.
	self assert: (rr ~= ra1 and: [rr ~= ra2 and: [ra1 ~= ra2]]).
	self ssTop popToReg: ra2.
	self ssPop: 1.
	self ssTop popToReg: ra1.
	self ssPop: 1.
	self ssTop popToReg: rr.
	self ssPop: 1.
	adjust := objectMemory baseHeaderSize - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
	objectRepresentation genConvertSmallIntegerToIntegerInReg: ra1.
	objectRepresentation genConvertSmallIntegerToIntegerInReg: ra2.
	self AddCq: adjust R: ra1.
	self MoveR: ra2 Xbr: ra1 R: rr.
	objectRepresentation genConvertIntegerToSmallIntegerInReg: ra2.
	^self ssPushRegister: ra2! !

!SistaCogitClone methodsFor: 'mapped inline primitive generators' stamp: 'cb 12/14/2017 17:08'!
genByteAtPutImmutabilityCheck
	| ra1 ra2 rr adjust mutableJump immutabilityFailure indexIsCst |
	<var: #mutableJump type: #'AbstractInstruction *'>
	<var: #immutabilityFailure type: #'AbstractInstruction *'>
	"Assumes rr is not a context and no store check is needed"
	indexIsCst := (self ssValue: 1) type = SSConstant.
	rr := ReceiverResultReg.
	ra1 := TempReg.
	ra2 := ClassReg.
	self voidReceiverResultRegContainsSelf.
	self ssFlushTo: simStackPtr - 1. "we use top value immediately after, so it does not require flush"
	self ssTop popToReg: ra2.
	adjust := objectMemory baseHeaderSize - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
	indexIsCst
		ifFalse:
			[(self ssValue: 1) popToReg: ra1.
			objectRepresentation genConvertSmallIntegerToIntegerInReg: ra1.
			self AddCq: adjust R: ra1]
		ifTrue: [self MoveCq: (objectMemory integerValueOf: (self ssValue: 1) constant) + adjust R: ra1].
	(self ssValue: 2) popToReg: rr.
	self ssPop: 3.
	self ssPushRegister: ra2.

	"Everything is flushed except returned value. All regs are unused except rr ra1 ra2"
	mutableJump := self genJumpMutable: rr scratchReg: Arg0Reg.
	self PushR: ra2. "flush sim stack top"
	self SubCq: 1 + adjust R: ra1. "index 0-relative for trampoline, ra1 unused afterwards"
	objectRepresentation genVarIndexCallStoreTrampoline.
	self PopR: ra2. "restore sim stack top"
	immutabilityFailure := self Jump: 0.
	
	mutableJump jmpTarget: (self MoveR: ra2 R: Arg0Reg).
	objectRepresentation genConvertSmallIntegerToIntegerInReg: Arg0Reg.
	self MoveR: Arg0Reg Xbr: ra1 R: rr.
	
	immutabilityFailure jmpTarget: self Label.
	^0! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 4/10/2017 18:29'!
genByteEqualsInlinePrimitive: prim

	"3021	Byte Object >> equals:length:	
	The receiver and the arguments are both byte objects and have both the same size (length in bytes). 
	The length argument is a smallinteger. 
	Answers true if all fields are equal, false if not. 
	Comparison is bulked to word comparison."
	
	"Overview: 
	 1.	The primitive is called like that: [byteObj1 equals: byteObj2 length: length].
	  	In the worst case we use 5 registers including TempReg 
		and we produce a loop bulk comparing words.
	 2.	The common case is a comparison against a cst: [byteString = 'foo'].
		which produces in Scorch [byteString equals: 'foo' length: 3].
		We try to generate fast code for this case with 3 heuristics:
		- specific fast code if len is a constant
		- unroll the loop if len < 2 * wordSize
		- compile-time reads if str1 or str2 is a constant and loop is unrolled.
		We use 3 registers including TempReg in the common case. 
		We could use 1 less reg if the loop is unrolled, the instr is followed by a branch
		AND one operand is a constant, but this is complicated enough.
	3.	We ignore the case where all operands are constants 
		(We assume Scorch simplifies it, it works but it is not optimised)"
		
	| str1Reg str2Reg lenReg extraReg jmp jmp2 needjmpZeroSize needLoop unroll jmpZeroSize instr lenCst mask |
	<var: #jmp type: #'AbstractInstruction *'>
	<var: #instr type: #'AbstractInstruction *'>
	<var: #jmp2 type: #'AbstractInstruction *'>
	<var: #jmpZeroSize type: #'AbstractInstruction *'>

	"--- quick path for empty string---"
	"This path does not allocate registers and right shift on negative int later in the code.
	 Normally this is resolved by Scorch but we keep it for correctness and consistency"
	self ssTop type = SSConstant ifTrue: 
		[ lenCst := objectMemory integerValueOf: self ssTop constant.
		  lenCst = 0 ifTrue: [ self ssPop: 3. self ssPushConstant: objectMemory trueObject. ^ 0 ] ].

	"--- Allocating & loading registers --- "
	needLoop := (self ssTop type = SSConstant and: [ lenCst <= (objectMemory wordSize * 2) ]) not.
	unroll := needLoop not and: [lenCst > objectMemory wordSize ].
	needLoop 
		ifTrue: 
			[ self assert: self ssTop spilled not.
			  str1Reg := self allocateRegForStackEntryAt: 1 notConflictingWith: self emptyRegisterMask.
			  str2Reg := self allocateRegForStackEntryAt: 2 notConflictingWith: (self registerMaskFor: str1Reg).
			  lenReg := self allocateRegForStackEntryAt: 0 notConflictingWith: (self registerMaskFor: str1Reg and: str2Reg).
			  (self ssValue: 1) popToReg: str1Reg.
			  (self ssValue: 2) popToReg: str2Reg.
			  extraReg := self allocateRegNotConflictingWith: (self registerMaskFor: str1Reg and: str2Reg and: lenReg)]
		ifFalse: 
			[ mask := self emptyRegisterMask.
			  (self ssValue: 1) type = SSConstant ifFalse: 
				[ str1Reg := self allocateRegForStackEntryAt: 1 notConflictingWith: mask.
				  (self ssValue: 1) popToReg: str1Reg.
				  mask := mask bitOr: (self registerMaskFor: str1Reg) ].
			  (self ssValue: 2) type = SSConstant ifFalse: 
				[ str2Reg := self allocateRegForStackEntryAt: 2 notConflictingWith: mask.
				  (self ssValue: 2) popToReg: str2Reg.
				  mask := mask bitOr: (self registerMaskFor: str2Reg) ].
			  extraReg := self allocateRegNotConflictingWith: mask].
	
	"--- Loading LenReg (or statically resolving it) --- "
	"LenReg is loaded with (lenInBytes + objectMemory baseHeaderSize - 1 >> shiftForWord)
	 LenReg is the index for the last word to compare with MoveXwr:r:R:.
	 The loop iterates from LenReg to first word of ByteObj"
	self ssTop type = SSConstant 
		ifTrue: "common case, str = 'foo'. We can precompute lenReg."
			[ lenCst := lenCst + objectMemory baseHeaderSize - 1 >> objectMemory shiftForWord.
			  needLoop ifTrue: [self MoveCq: lenCst R: lenReg ].
			  needjmpZeroSize := false] 
		ifFalse: "uncommon case, str = str2. lenReg in word computed at runtime."
			[ self ssTop popToReg: lenReg.
			  objectRepresentation genConvertSmallIntegerToIntegerInReg: lenReg.
			  self CmpCq: 0 R: lenReg.
			  jmpZeroSize := self JumpZero: 0.
			  needjmpZeroSize := true.
			  self AddCq: objectMemory baseHeaderSize - 1 R: lenReg.
			  self ArithmeticShiftRightCq: objectMemory shiftForWord R: lenReg ].
	
	"--- Comparing the strings --- "
	"LenReg has the index of the last word to read (unless no loop). 
	 We decrement it to adjust -1 (0 in 64 bits) while comparing"
	needLoop 
		ifTrue:
			[instr := self MoveXwr: lenReg R: str1Reg R: extraReg.
			self MoveXwr: lenReg R: str2Reg R: TempReg.
			self CmpR: extraReg R: TempReg.
			jmp := self JumpNonZero: 0. "then string are not equal (jmp target)"
			self AddCq: -1 R: lenReg.
			self CmpCq: (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1 R: lenReg. "first word of ByteObj, stop looping."
			self JumpNonZero: instr]
		ifFalse: "Common case, only 1 or 2 word to check: no lenReg allocation, cst micro optimisations"
			[self genByteEqualsInlinePrimitiveCmp: str1Reg with: str2Reg scratch1: extraReg scratch2: TempReg field: 0.
			jmp := self JumpNonZero: 0. "then string are not equal (jmp target)"
			unroll ifTrue: "unrolling more than twice generate more instructions than the loop so we don't do it"
				[self genByteEqualsInlinePrimitiveCmp: str1Reg with: str2Reg scratch1: extraReg scratch2: TempReg field: 1.
				jmp2 := self JumpNonZero: 0. "then string are not equal (jmp target)"]].
	needjmpZeroSize ifTrue: [ jmpZeroSize jmpTarget: self Label ].
	"fall through, strings are equal"
	
	"--- Pushing the result or pipelining a branch --- "	
	self ssPop: 3.
	self genByteEqualsInlinePrimitiveResult: jmp returnReg: extraReg.
	unroll ifTrue: [jmp2 jmpTarget: jmp getJmpTarget].
	^0! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 3/23/2017 14:33'!
genByteEqualsInlinePrimitiveCmp: str1Reg with: str2Reg scratch1: scratch1Reg scratch2: scratch2Reg field: index
	| shift |
	<inline: true>
	shift := objectMemory baseHeaderSize + (index * objectMemory wordSize).
	(self ssValue: 1) type = SSConstant 
		ifTrue: [self MoveCq: (objectMemory fetchPointer: index ofObject: (self ssValue: 1) constant) R: scratch1Reg]
		ifFalse: [self MoveMw: shift r: str1Reg R: scratch1Reg].
	 (self ssValue: 2) type = SSConstant 
		ifTrue: [self MoveCq: (objectMemory fetchPointer: index ofObject: (self ssValue: 2) constant) R: scratch2Reg]
		ifFalse: [self MoveMw: shift r: str2Reg R: scratch2Reg].
	self CmpR: scratch1Reg R: scratch2Reg.! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 3/23/2017 14:56'!
genByteEqualsInlinePrimitiveResult: jmp returnReg: reg
	"Byte equal is falling through if the result is true, or jumping using jmp if the result is false.
	 The method is required to set the jump target of jmp.
	 We look ahead for a branch and pipeline the jumps if possible..
	 ReturnReg is used only if not followed immediately by a branch."
	| branchDescriptor nextPC postBranchPC targetBytecodePC localJump canElide |
	<var: #localJump type: #'AbstractInstruction *'>
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>
	self extractMaybeBranchDescriptorInto: [ :descr :next :postBranch :target | 
		branchDescriptor := descr. nextPC := next. postBranchPC := postBranch. targetBytecodePC := target ].
	
	"Case 1 - not followed by a branch"
	(branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse])
		ifFalse: 
			[self genMoveTrueR: reg.
			 localJump := self Jump: 0.
			 jmp jmpTarget: (self genMoveFalseR: reg).
			 localJump jmpTarget: self Label.
			 self ssPushRegister: reg.
			^ 0].

	"Case 2 - followed by a branch"
	(self fixupAt: nextPC) notAFixup
		ifTrue: "The next instruction is dead.  we can skip it."
			[deadCode := true.
		 	 self ensureFixupAt: targetBytecodePC.
			 self ensureFixupAt: postBranchPC ]
		ifFalse:
			[self ssPushConstant: objectMemory trueObject]. "dummy value"
	"We can only elide the jump if the pc after nextPC is the same as postBranchPC.
	 Branch following means it may not be."
	self nextDescriptorExtensionsAndNextPCInto:
		[:iguana1 :iguana2 :iguana3 :followingPC| nextPC := followingPC].
	canElide := deadCode and: [nextPC = postBranchPC].
	 branchDescriptor isBranchTrue
		ifTrue: 
			[ self Jump: (self ensureNonMergeFixupAt: targetBytecodePC).
			  canElide 
					ifFalse: [ jmp jmpTarget: (self ensureNonMergeFixupAt: postBranchPC) ]
					ifTrue: [ jmp jmpTarget: self Label ] ]
		ifFalse: [ canElide ifFalse: [ self Jump: (self ensureNonMergeFixupAt: postBranchPC).
				 jmp jmpTarget: (self ensureNonMergeFixupAt: targetBytecodePC) ] ].
	^0! !

!SistaCogitClone methodsFor: 'mapped inline primitive generators' stamp: 'cb 1/8/2018 14:51'!
genCallMappedInlinedPrimitive
	"SistaV1:	236		11101100	iiiiiiii		callMappedInlinedPrimitive"
	^ self genMappedInlinePrimitive: byte1! !

!SistaCogitClone methodsFor: 'bytecode generators' stamp: 'cb 11/6/2017 16:51'!
genCallPrimitiveBytecode
	"SistaV1: 248		11111000 	iiiiiiii		mjjjjjjj		Call Primitive #iiiiiiii + (jjjjjjj * 256) m=1 means inlined primitive, no hard return after execution.
	 See EncoderForSistaV1's class comment and StackInterpreter>>#inlinePrimitiveBytecode:"
	| prim primSet |
	byte2 < 128 ifTrue:
		[^bytecodePC = initialPC
			ifTrue: [0]
			ifFalse: [EncounteredUnknownBytecode]].
	prim := byte2 - 128 << 8 + byte1.
	primSet := prim >> 13 bitAnd: 3.
	prim := prim bitAnd: 8191.
	LowcodeVM
		ifTrue:
			[
			primSet = 1 ifTrue: [
				prim < 1000 ifTrue:
					[^self genLowcodeNullaryInlinePrimitive: prim].

				prim < 2000 ifTrue:
					[^self genLowcodeUnaryInlinePrimitive: prim - 1000].
				
				prim < 3000 ifTrue:
					[^ self genLowcodeBinaryInlinePrimitive: prim - 2000].

				prim < 4000 ifTrue:
					[^self genLowcodeTrinaryInlinePrimitive: prim - 3000].
			]
		].
	
	self assert: primSet = 0.
	
	^ self genSistaInlinePrimitive: prim! !

!SistaCogitClone methodsFor: 'mapped inline primitive generators' stamp: 'cb 11/23/2017 14:43'!
genDirectCall
	"250	directCall
	literal index of the method to call on top of stack =>  (variable number of parameters)"
	| newMethod newMethodHeader newMethodArgCount sendTable annotation litIndex |
	self assert: (self ssTop type = SSConstant and: [objectMemory isCompiledMethod: self ssTop constant]).
	litIndex := objectMemory integerValueOf: self ssTop constant.
	newMethod := self getLiteral: litIndex.
	self ssPop: 1.
	self flag: #TODO. "One needs to finish the implementation with the annotation and the sendTable"
	sendTable := 1. "directCallSendTable"
	annotation := 1. "directCallAnnotation"
	newMethodHeader := self rawHeaderOf: newMethod.
	newMethodArgCount := self argumentCountOfMethodHeader: methodHeader.
	"The receiver cannot be a forwader"
	"numArgs >= (NumSendTrampolines - 1) ifTrue:
		[self MoveCq: newMethodArgCount R: SendNumArgsReg]."
	"Load inline cache with method index"
	"self MoveUniqueC32: litIndex R: ClassReg.
	(self Call: (sendTable at: (newMethodArgCount min: NumSendTrampolines - 1))) annotation: annotation.
	self voidReceiverOptStatus.
	self ssPushRegister: ReceiverResultReg."
	^ EncounteredUnknownBytecode "Need to finish implementation"! !

!SistaCogitClone methodsFor: 'inline primitive binary generators' stamp: 'cb 3/1/2018 20:34'!
genDivInlinePrimitive: primIndex
	"2003	/
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant) "
	"2004	//
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant) "
	"2005	\\
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant) "
	"2006	quo:
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant) "
	
	"We don't deal with constants here. Too complex and does nto bring much"
	| ra rr jumpExact jumpSameSign |
	<var: #jumpExact type: #'AbstractInstruction *'>
	<var: #jumpSameSign type: #'AbstractInstruction *'>
	self allocateRegForStackTopTwoEntriesInto: [:rTop :rNext | ra := rTop. rr := rNext ].
	self ssTop popToReg: ra.
	self ssPop: 1.
	self ssTop popToReg: rr.
	self ssPop: 1.
	self assert: backEnd canDivQuoRem.
	"arg non zero, no overflow check. / remains (unused in optimised code)"
	primIndex caseOf: {
		[4] ->	[objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: ra.
				objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: rr.
				self DivR: ra R: rr Quo: rr Rem: TempReg.
				"If remainder is zero, success"
				self CmpCq: 0 R: TempReg.
				jumpExact := self JumpZero: 0.
				"If arg and remainder signs are different we must round down."
				objectRepresentation genConvertIntegerToSmallIntegerInReg: ra.
				self XorR: TempReg R: ra.
				(self lastOpcode setsConditionCodesFor: JumpZero) ifFalse:
					[self CmpCq: 0 R: ra].
				jumpSameSign := self JumpGreaterOrEqual: 0.
				self SubCq: 1 R: rr.
				jumpSameSign jmpTarget: (jumpExact jmpTarget: self Label).
				objectRepresentation genConvertIntegerToSmallIntegerInReg: rr ].
		[5] ->	[objectRepresentation genRemoveSmallIntegerTagsInScratchReg: ra.
				objectRepresentation genRemoveSmallIntegerTagsInScratchReg: rr.
			 	self DivR: ra R: rr Quo: TempReg Rem: rr.
				"If remainder is zero we're done."
				self CmpCq: 0 R: rr.
				jumpExact := self JumpZero: 0.
				"If arg and remainder signs are different we must reflect around zero."
				self XorR: rr R: ra.
				(self lastOpcode setsConditionCodesFor: JumpZero) ifFalse:
					[self CmpCq: 0 R: ra].
				jumpSameSign := self JumpGreaterOrEqual: 0.
				self XorR: rr R: ra.
				self AddR: ra R: rr.
				jumpSameSign jmpTarget: (jumpExact jmpTarget: self Label).
				objectRepresentation genSetSmallIntegerTagsIn: rr ].
		[6] ->	[objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: ra.
				objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: rr.
				self DivR: ra R: rr Quo: rr Rem: ra.
				objectRepresentation genConvertIntegerToSmallIntegerInReg: rr].
	} otherwise: [^EncounteredUnknownBytecode].
	self ssPushRegister: rr.
	^0! !

!SistaCogitClone methodsFor: 'mapped inline primitive generators' stamp: 'cb 11/23/2017 14:31'!
genEnsureEnoughSlots
	"50	EnsureEnoughWords
	 literal which is a Smi => ret value is receiver"
	|slots skip|
	<var: #skip type: #'AbstractInstruction *'>
	self assert: (self ssTop type = SSConstant and: [objectMemory isIntegerObject: self ssTop constant]).
	slots := objectMemory integerValueOf: self ssTop constant.
	self ssFlushTo: simStackPtr - 1.
	self
		MoveAw: objectMemory freeStartAddress R: TempReg;
		CmpCq: objectMemory getScavengeThreshold - (objectMemory bytesPerOop * slots) R: TempReg.
	skip := self JumpBelow: 0.
	objectRepresentation genSetGCNeeded.
	self CallRT: ceCheckForInterruptTrampoline.
	skip jmpTarget: self Label.
	self annotateBytecode: skip getJmpTarget.
	^0! !

!SistaCogitClone methodsFor: 'bytecode generator support' stamp: 'eem 12/12/2016 11:22'!
genExecutionCountLogicInto: binaryBlock counterReg: counterReg
	<var: #countTripped type: #'AbstractInstruction *'>
	<inline: true>
	| counterAddress countTripped |
	counterAddress := counters + (CounterBytes * counterIndex).
	self MoveA32: counterAddress R: counterReg.
	self SubCq: 16r10000 R: counterReg. "Count executed"
	"If counter trips simply abort the comparison continuing to the following
	 branch *without* writing back.  A double decrement will not trip the second time."
	countTripped := self JumpCarry: 0.
	self MoveR: counterReg A32: counterAddress. "write back"
	binaryBlock value: counterAddress value: countTripped! !

!SistaCogitClone methodsFor: 'bytecode generator support' stamp: 'eem 9/7/2016 15:03'!
genFallsThroughCountLogicCounterReg: counterReg counterAddress: counterAddress
	<inline: true>
	"Gen this when the branch has not been taken and forwarders have been followed."
	self SubCq: 1 R: counterReg. "Count untaken"
	self MoveR: counterReg A32: counterAddress. "write back"! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 1/3/2018 16:02'!
genJumpBinaryInlinePrimitive: primIndex
	| target testReg |
	self assert: self ssTop type = SSConstant.
	target := self eventualTargetOf: ((objectMemory integerValueOf: self ssTop constant) + 3 + bytecodePC).
	testReg := self allocateRegForStackEntryAt: 1.
	(self ssValue: 1) popToReg: testReg.
	self ssPop: 2.
	primIndex caseOf: {
		"7016	jumpIfWritable:
		 Not a forwarder, literal which is a Smi"
		[16]	->	[self cppIf: #IMMUTABILITY
						ifTrue: [(objectRepresentation genJumpMutable: testReg scratchReg: TempReg) jmpTarget: (self ensureFixupAt: target).]
						ifFalse: [self Jump: (self ensureFixupAt: target)].
					^0].
		"7017	jumpIfReadOnly:
		 Not a forwarder, literal which is a Smi"
		[17]	->	[self cppIf: #IMMUTABILITY
						ifTrue: [(objectRepresentation genJumpImmutable: testReg scratchReg: TempReg) jmpTarget: (self ensureFixupAt: target)]
						ifFalse: [0"Do nothing - fall through"].
					^0].
		"7018	jumpIfYoung:
		 Not a forwarder, literal which is a Smi"
		[18]	->	[self CmpCw: objectMemory storeCheckBoundary R: testReg.
	 				 self JumpBelow: (self ensureFixupAt: target).
					 ^0].
		"7019	jumpIfOld:
		 Not a forwarder, literal which is a Smi"
		[19]	->	[self CmpCw: objectMemory storeCheckBoundary R: testReg.
	 				 self JumpAboveOrEqual: (self ensureFixupAt: target).
					^0].
	}. 
	^ EncounteredUnknownBytecode
	! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 1/8/2018 14:41'!
genJumpTrinaryInlinePrimitive: primIndex
	"Note: those tests work with forwarders (wrong class index)"
	| testReg target behavior targetFixUp |
	<var: #targetFixUp type: #'BytecodeFixup *'>
	self assert: self ssTop type = SSConstant.
	self assert: (self ssValue: 1) type = SSConstant.
	testReg := self allocateRegForStackEntryAt: 2.
	(self ssValue: 2) popToReg: testReg.
	behavior := (self ssValue: 1) constant.
	target := self eventualTargetOf: ((objectMemory integerValueOf: self ssTop constant) + 3 + bytecodePC).
	self ssPop: 3.
	targetFixUp := self cCoerceSimple: (self ensureFixupAt: target) to: #'AbstractInstruction *'.
	primIndex caseOf: {
	[0] -> 	["8000	jumpIfInstanceOf:distance:
			 Anything, literal which is a Behavior, literal which is a Smi"
			objectRepresentation branchIf: testReg instanceOfBehavior: behavior target: targetFixUp.
			^0].
	[1] -> 	["8001	jumpIfNotInstanceOf:distance:
			 Anything, literal which is a Behavior, literal which is a Smi"
			 objectRepresentation branchIf: testReg notInstanceOfBehavior: behavior target: targetFixUp.
			^0].
	[2] -> 	["8002	jumpIfInstanceOfOneOf:distance:
			 Anything, Array of behaviors, literal which is a Smi"
			 objectRepresentation branchIf: testReg instanceOfBehaviors: behavior target: targetFixUp.
			^0].
	[3] -> 	["8003	jumpIfNotInstanceOfOneOf:distance:
			  Anything, Array of behaviors, literal which is a Smi"
			 objectRepresentation branchIf: testReg notInstanceOfBehaviors: behavior target: targetFixUp.
			^0].
	}.
	^ EncounteredUnknownBytecode
	
	! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 1/3/2018 16:01'!
genJumpUnaryInlinePrimitive: primIndex
	"6000	backjumpNoInterrupt
	 literal which is a Smi"
	primIndex = 0 ifTrue: 
		[|targetBytecodePC|
		 self assert: self ssTop type = SSConstant.
		 targetBytecodePC := (objectMemory integerValueOf: self ssTop constant) + 3 + bytecodePC.
		 self ssPop: 1.
		 self ssFlushTo: simStackPtr.
		 deadCode := true. "can't fall through"
		 self Jump: (self fixupAt: targetBytecodePC).
		 ^0].
	^EncounteredUnknownBytecode! !

!SistaCogitClone methodsFor: 'mapped inline primitive generators' stamp: 'cb 1/8/2018 14:52'!
genMappedInlinePrimitive: primIndex
	"SistaV1:	236		11101100	iiiiiiii		callMappedInlinedPrimitive"
	"Number of arguments:
	 0-49 nullary
	 50-99 unary
	 100-149  binary
	 150-199 trinary
	 200-255 variable"
	"Specification:
	50	EnsureEnoughWords
	literal which is a Smi => ret value is receiver
	150	immCheckPointerAt:put:
	pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	151	immCheckStoreCheckPointerAt:put:
	pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	152	immCheckMaybeContextPointerAt:put:
	pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	153	immCheckMaybeContextStoreCheckPointerAt:put:
	pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	154	immCheckByteAt:put:
	byte object, Smi, 8 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	155	immCheckShortAt:put:
	short object, Smi, 16 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	156	immCheckWordAt:put:
	word object, Smi, 32 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	157	immCheckDoubleWordAt:put:
	double word object, Smi, 64 bits unsigned Smi or LargePositiveInteger => arg2 (1-based, optimised if arg1 is a constant)
	250	directCall
	method to call on top of stack =>  (variable number of parameters)"
	primIndex caseOf: 
	{ 
		[50] ->	 [^self genEnsureEnoughSlots].
		[150] -> [^self genPointerAtPutMaybeContext: false storeCheck: false immutabilityCheck: true].
		[151] -> [^self genPointerAtPutMaybeContext: false storeCheck: true immutabilityCheck: true].
		[152] -> [^self genPointerAtPutMaybeContext: true storeCheck: false immutabilityCheck: true].
		[153] -> [^self genPointerAtPutMaybeContext: true storeCheck: true immutabilityCheck: true].
		[154] -> [self cppIf: #IMMUTABILITY
					ifTrue: [^self genByteAtPutImmutabilityCheck]
					ifFalse: [self genByteAtPut.
							self annotateBytecode: self Label.
							 ^0]].
		[155] -> [^EncounteredUnknownBytecode "not implemented, missing short instruction in Cog RTL"].
		[156] -> [^EncounteredUnknownBytecode "not implemented, need to deal with LargePositiveInteger in 32 bits"].
		[157] -> [^EncounteredUnknownBytecode "not implemented, need to deal with LargePositiveInteger in 64 bits"].
		[250] -> [^self genDirectCall].
	} otherwise: [^EncounteredUnknownBytecode].
	! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 4/10/2017 18:20'!
genMoveAndAdjustSSEntry: ssEntry into: reg
	<var: #ssEntry type: #'SimStackEntry *'>
	<inline: true>
	| adjust |
	adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
	ssEntry type = SSConstant 
		ifTrue: [self MoveCq: (objectMemory integerValueOf: ssEntry constant) + adjust R: reg.
			      ssEntry spilled ifTrue: [self AddCq: objectMemory wordSize R: SPReg]]
		ifFalse: 
			[ssEntry popToReg: reg.
			 objectRepresentation genConvertSmallIntegerToIntegerInReg: reg.
			 adjust ~= 0 ifTrue: [ self AddCq: adjust R: reg. ]].	! !

!SistaCogitClone methodsFor: 'initialization' stamp: 'eem 2/1/2020 18:51'!
genMustBeBooleanTrampolineFor: boolean called: trampolineName
	"This can be entered in one of two states, depending on SendNumArgsReg. See
	 e.g. genJumpIf:to:.  If SendNumArgsReg is non-zero then this has been entered via
	 the initial test of the counter in the jump executed count (i.e. the counter has
	 tripped).  In this case TempReg contains the boolean to be tested and should not
	 be offset, and ceCounterTripped should be invoked with the unoffset TempReg.
	 If SendNumArgsReg is zero then this has been entered for must-be-boolean
	 processing. TempReg has been offset by boolean and must be corrected and
	 ceSendMustBeBoolean: invoked with the corrected value."
	<var: #trampolineName type: #'char *'>
	| jumpMBB |
	<var: #jumpMBB type: #'AbstractInstruction *'>
	<inline: false>
	self zeroOpcodeIndex.
	self CmpCq: 0 R: SendNumArgsReg.
	jumpMBB := self JumpZero: 0.
	"Open-code self compileTrampolineFor: #ceCounterTripped: numArgs: 1 arg: TempReg ...
	 so we can restore ResultReceiverReg."
	self genSmalltalkToCStackSwitch: true.
	self
		compileCallFor: #ceCounterTripped:
		numArgs: 1
		arg: TempReg
		arg: nil
		arg: nil
		arg: nil
		resultReg: TempReg "(*)"
		regsToSave: self emptyRegisterMask.
	"(*) For the case where the ceCounterTripped: call returns (e.g. because there's no callback selector
	 installed), the call to the ceSendMustBeBooleanAddTrue/FalseTrampoline is followed by a jump
	 back to the start of the counter/condition test sequence.  For this case copy the C result to
	 TempReg (the register that is tested), to reload it with the boolean to be tested."
	backEnd genLoadStackPointers.
	backEnd hasLinkRegister ifTrue:
		[self PopR: LinkReg].
	"To keep ResultReceiverReg live if optStatus thought it was, simply reload it
	 from the frame pointer.  This avoids having to reload it in the common case
	 (counter does not trip) if it was live.  Note we can't use putSelfInReceiverResultReg
	 when generating trampolines because simSelf has not yet been initialized."
	self MoveMw: FoxMFReceiver r: FPReg R: ReceiverResultReg.
	self RetN: 0.
	"If the objectRepresentation does want true & false to be mobile then we need to record these addresses."
	self assert: (objectRepresentation shouldAnnotateObjectReference: boolean) not.
	jumpMBB jmpTarget: (self AddCq: boolean R: TempReg).
	^self genTrampolineFor: #ceSendMustBeBoolean:
		called: trampolineName
		numArgs: 1
		arg: TempReg
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: true! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 11/6/2017 16:53'!
genNullaryInlinePrimitive: prim
	<option: #SistaVM>
	^EncounteredUnknownBytecode! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 3/1/2018 17:55'!
genPointerAtPutConstantMaybeContext: maybeContext storeCheck: needsStoreCheck immutabilityCheck: needsImmCheck 
	| rcvrReg valReg indexCst |
	indexCst := (objectMemory integerValueOf: (self ssValue: 1) constant) - 1.
	"we want to have on top of stack the value to write descriptor."
	(maybeContext or: [needsImmCheck]) 
		ifTrue: [ valReg := ClassReg.
				 self voidReceiverResultRegContainsSelf.
				 rcvrReg := ReceiverResultReg ]
		ifFalse: [ valReg := self allocateRegForStackEntryAt: 0 notConflictingWith: 0.
				  needsStoreCheck 
					ifFalse: [rcvrReg := self allocateRegForStackEntryAt: 2 notConflictingWith: (self registerMaskFor: valReg)]
					ifTrue: [self voidReceiverResultRegContainsSelf.
							rcvrReg := ReceiverResultReg]].
	"If rcvr is valReg, we flush it. Could generate a move instead but not that common"
	self ssAllocateRequiredReg: valReg upThrough: simStackPtr - 2.
	self ssAllocateRequiredReg: rcvrReg upThrough: simStackPtr - 3.
	self ssTop popToReg: valReg.
	self ssPop: 1.
	self ssTop spilled ifTrue: [self AddCq: objectRepresentation wordSize R: SPReg].
	self ssPop: 1.
	self ssTop popToReg: rcvrReg.
	self ssPop: 1.
	self ssPushRegister: valReg.
			
	maybeContext
		ifFalse: 
			[^self 
				genGenericStorePop: false 
				slotIndex: indexCst
				destReg: rcvrReg 
				needsStoreCheck: needsStoreCheck 
				needsRestoreRcvr: false 
				needsImmutabilityCheck: needsImmCheck]
		ifTrue:
			[^self
				genGenericStorePop: false 
				MaybeContextSlotIndex: indexCst
				needsStoreCheck: needsStoreCheck 
				needsRestoreRcvr: false 
				needsImmutabilityCheck: needsImmCheck]! !

!SistaCogitClone methodsFor: 'mapped inline primitive generators' stamp: 'cb 12/11/2017 15:39'!
genPointerAtPutImmCheckAndStoreCheck
	| ra1 ra2 rr adjust scratchReg immutableJump jmpImmediate jmpDestYoung jmpSourceOld jumpRemembered indexIsCst |
	<var: #jmpSourceOld type: #'AbstractInstruction *'>
	<var: #jmpDestYoung type: #'AbstractInstruction *'>
	<var: #jmpImmediate type: #'AbstractInstruction *'>
	<var: #immutableJump type: #'AbstractInstruction *'>
	<var: #jumpRemembered type: #'AbstractInstruction *'>
	"Assumes rr is not a context and no store check is needed"
	indexIsCst := (self ssValue: 1) type = SSConstant.
	self ssFlushTo: simStackPtr - 3.
	rr := ReceiverResultReg.
	ra1 := TempReg.
	ra2 := ClassReg.
	scratchReg := Arg0Reg.
	adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
	(self ssValue: 2) popToReg: rr.
	indexIsCst
		ifFalse:
			[(self ssValue: 1) popToReg: ra1.
			objectRepresentation genConvertSmallIntegerToIntegerInReg: ra1.
			adjust ~= 0 ifTrue: [self AddCq: adjust R: ra1]]
		ifTrue: [self MoveCq: (objectMemory integerValueOf: (self ssValue: 1) constant) + adjust R: ra1].
	self ssTop popToReg: ra2.
	self ssPop: 3.
	self ssPushRegister: ra2.
	self voidReceiverResultRegContainsSelf.

	immutableJump := self genJumpImmutable: rr scratchReg: scratchReg.
	self MoveR: ra2 Xwr: ra1 R: rr.
	
	"store check"
	jmpImmediate := self genJumpImmediate: ra2.
	"Get the old/new boundary in scratchReg"
	self MoveCw: objectMemory storeCheckBoundary R: scratchReg.
	"Is target young?  If so we're done"
	self CmpR: scratchReg R: rr. "N.B. FLAGS := destReg - scratchReg"
	jmpDestYoung := self JumpBelow: 0.
	"Is value stored old?  If so we're done."
	self CmpR: scratchReg R: ra2. "N.B. FLAGS := valueReg - scratchReg"
	jmpSourceOld := self JumpAboveOrEqual: 0.
	"value is young and target is old.
	 Need to remember this only if the remembered bit is not already set."
	jumpRemembered := objectRepresentation genIfRequiredCheckRememberedBitOf: rr scratch: scratchReg.
	"Set the inst var index for the benefit of the immutability check. The trampoline will
	 repeat the check to choose between the immutbality violation and the store check."
	immutableJump jmpTarget: self Label.
	self PushR: ra2.
	self SubCq: 1 + adjust R: ra1. "index 0-relative for trampoline, ra1 unused afterwards"
	objectRepresentation genVarIndexCallStoreTrampoline.
	self PopR: ra2.
	jmpImmediate jmpTarget:
	(jmpDestYoung jmpTarget:
	(jmpSourceOld jmpTarget:
		self Label)).
	self setIfRequiredTargetOf: jumpRemembered toTargetOf: jmpImmediate.
	
	^0! !

!SistaCogitClone methodsFor: 'mapped inline primitive generators' stamp: 'cb 11/23/2017 14:57'!
genPointerAtPutImmCheckButNoStoreCheck
	| ra1 ra2 rr adjust mutableJump immutabilityFailure indexIsCst |
	<var: #mutableJump type: #'AbstractInstruction *'>
	<var: #immutabilityFailure type: #'AbstractInstruction *'>
	"Assumes rr is not a context and no store check is needed"
	indexIsCst := (self ssValue: 1) type = SSConstant.
	self ssFlushTo: simStackPtr - 3.
	rr := ReceiverResultReg.
	ra1 := TempReg.
	ra2 := ClassReg.
	(self ssValue: 2) popToReg: rr.
	adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
	indexIsCst
		ifFalse:
			[(self ssValue: 1) popToReg: ra1.
			objectRepresentation genConvertSmallIntegerToIntegerInReg: ra1.
			adjust ~= 0 ifTrue: [self AddCq: adjust R: ra1]]
		ifTrue: [self MoveCq: (objectMemory integerValueOf: (self ssValue: 1) constant) + adjust R: ra1].
	self ssTop popToReg: ra2.
	self ssPop: 3.
	self ssPushRegister: ra2.
	self voidReceiverResultRegContainsSelf.

	mutableJump := self genJumpMutable: rr scratchReg: Arg0Reg.
	"simStack is flushed, but result is not"
	self PushR: ra2.
	self SubCq: 1 + adjust R: ra1. "index 0-relative for trampoline, ra1 unused afterwards"
	objectRepresentation genVarIndexCallStoreTrampoline.
	self PopR: ra2.
	immutabilityFailure := self Jump: 0.
	
	mutableJump jmpTarget: self Label.
	self MoveR: ra2 Xwr: ra1 R: rr.
	immutabilityFailure jmpTarget: self Label.
	^0! !

!SistaCogitClone methodsFor: 'mapped inline primitive generators' stamp: 'cb 1/8/2018 14:52'!
genPointerAtPutMaybeContext: maybeContext storeCheck: needsStoreCheck immutabilityCheck: needsImmCheck
	"150	immCheckPointerAt:put:
	pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	151	immCheckStoreCheckPointerAt:put:
	pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	152	immCheckMaybeContextPointerAt:put:
	pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	153	immCheckMaybeContextStoreCheckPointerAt:put:
	pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
	| index |
	index := self ssValue: 1.
	"Instruction always annoted - uncommon, IMMUTABILITY disabled. We need a Nop since previous instr can be annotated."
	self cppIf: #IMMUTABILITY ifTrue: [] ifFalse: [ needsImmCheck ifTrue: [ self annotateBytecode: self Nop ] ].
	"Optimised case if arg1 is constant"
	index type = SSConstant 
		ifTrue:  [^self genPointerAtPutConstantMaybeContext: maybeContext storeCheck: needsStoreCheck immutabilityCheck: needsImmCheck].
	maybeContext ifTrue: [^EncounteredUnknownBytecode "not implemented, not used right now, for tempAt: optimisation on non constant"].
	self cppIf: #IMMUTABILITY ifTrue:
		[needsImmCheck
			ifTrue: [needsStoreCheck
				ifTrue: [^self genPointerAtPutImmCheckAndStoreCheck]
				ifFalse: [^self genPointerAtPutImmCheckButNoStoreCheck] ] ].
	^self genPointerAtPutStoreCheck: needsStoreCheck! !

!SistaCogitClone methodsFor: 'mapped inline primitive generators' stamp: 'cb 11/6/2017 10:02'!
genPointerAtPutStoreCheck: needsStoreCheck 
	| ra1 ra2 rr adjust |
	"Assumes rr is not a context and no immutability check is needed"
	"The store check requires rr to be ReceiverResultReg"
	self 
		allocateRegForStackTopThreeEntriesInto: [:rTop :rNext :rThird | ra2 := rTop. ra1 := rNext. rr := rThird ] 
		thirdIsReceiver: needsStoreCheck.
	self assert: (rr ~= ra1 and: [rr ~= ra2 and: [ra1 ~= ra2]]).
	self ssTop popToReg: ra2.
	self ssPop: 1.
	self ssTop popToReg: ra1.
	self ssPop: 1.
	self ssTop popToReg: rr.
	self ssPop: 1.
	objectRepresentation genConvertSmallIntegerToIntegerInReg: ra1.
	adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
	adjust ~= 0 ifTrue: [ self AddCq: adjust R: ra1. ]. 
	self MoveR: ra2 Xwr: ra1 R: rr.
	needsStoreCheck ifTrue: 
		[ self assert: needsFrame. 
		objectRepresentation genStoreCheckReceiverReg: rr valueReg: ra2 scratchReg: TempReg inFrame: true].
	self ssPushRegister: ra2.
	^0! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 11/10/2017 11:06'!
genQuaternaryInlinePrimitive: prim
	^ EncounteredUnknownBytecode! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 11/10/2017 11:06'!
genQuinaryInlinePrimitive: prim
	^ EncounteredUnknownBytecode! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 3/6/2018 09:55'!
genSistaInlinePrimitive: prim
	"SistaV1:	248		11111000 	iiiiiiii		mjjjjjjj		Call Primitive #iiiiiiii + (jjjjjjj * 256) m=1 means inlined primitive, no hard return after execution. This is the dispatch for unmapped sista inlined primitives."
	
	prim < 1000 ifTrue:
		[^self genNullaryInlinePrimitive: prim].

	prim < 2000 ifTrue:
		[^self genUnaryInlinePrimitive: prim - 1000].
	
	prim < 3000 ifTrue:
		[^self genBinaryInlinePrimitive: prim - 2000].

	prim < 4000 ifTrue:
		[^self genTrinaryInlinePrimitive: prim - 3000].
	
	prim < 5000 ifTrue: 
		[^self genQuaternaryInlinePrimitive: prim - 4000].
	
	prim < 6000 ifTrue: 
		[^self genQuinaryInlinePrimitive: prim - 5000].
		
	 prim < 7000 ifTrue:
		[^self genJumpUnaryInlinePrimitive: prim - 6000].
	
 	prim < 8000 ifTrue:
		[^self genJumpBinaryInlinePrimitive: prim - 7000].
	
	^ self genJumpTrinaryInlinePrimitive: prim - 8000.! !

!SistaCogitClone methodsFor: 'bytecode generators' stamp: 'cb 4/29/2015 14:39'!
genSpecialSelectorComparisonWithoutCounters
	"This method is there because if I put directly the super send in genSpecialSelectorComparison Slang does not correctly translte the code to C, it does not correctly type one of the branchDescriptor to BytecodeDescriptor"
	^ super genSpecialSelectorComparison! !

!SistaCogitClone methodsFor: 'inline primitive generators' stamp: 'cb 11/10/2017 15:17'!
genTrinaryInlinePrimitive: prim
	"Bulk comment, each sub method has its own comment
	3000	pointerAt:put:
	Mutable pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	3001	storeCheckPointerAt:put:
	Mutable pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	3002	maybeContextPointerAt:put:
	Mutable pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	3003	maybeContextStoreCheckPointerAt:put:
	Mutable pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	3004	byteAt:put:
	Mutable byte object, Smi, 8 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	3005	shortAt:put:
	Mutable short object, Smi, 16 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	3006	wordAt:put:
	Mutable word object, Smi, 32 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	3007	doubleWordAt:put:
	Mutable double word object, Smi, 64 bits unsigned Smi or LargePositiveInteger => arg2 (1-based, optimised if arg1 is a constant)
	3021 is deprecated."
	prim <= 7 ifTrue: [^self genAtPutInlinePrimitive: prim].
	prim = 21 ifTrue: [^self genByteEqualsInlinePrimitive: prim].
	^ EncounteredUnknownBytecode! !

!SistaCogitClone methodsFor: 'inline primitive unary generators' stamp: 'cb 11/23/2017 11:41'!
genUnaryClassPrimitive 
	"1000	rawClass
	not a forwarder => Behavior (Same as class special send, but receiver is not a forwarder)"
	
	"Important performance note:
	In Scorch, typically a value is known as not being a forwarder if there is a trap.
	If the trap is due to a monomorphic send, the #class send leads to:
		trapIf: X notInstanceOf: C
		X rawClass
	therefore X rawClass is simplified in Scorch to the cst:C
	The rawClass is therefore used only for PICs. 
		trapIf: X notInstanceOf: C, C', C''
		X rawClass
	This unsafe operation is important to avoid register flush, but the performance 
	difference in differencing rawClass for immediate and rawClass for non immediate 
	classes is not that relevant"
	
	| topReg jumpIsImm destReg |
	<var: #jumpIsImm type: #'AbstractInstruction *'>
	topReg := self allocateRegForStackEntryAt: 0 notConflictingWith: 0.
	destReg := self allocateRegNotConflictingWith: (self registerMaskFor: topReg).
	self ssTop popToReg: topReg.
	
	"1. Read the class index"
	self MoveR: topReg R: TempReg.
	self AndCq: objectMemory tagMask R: TempReg.
	jumpIsImm := self JumpNonZero: 0.
	self flag: #endianness.
	"Get least significant half of header word in destReg"
	self MoveMw: 0 r: topReg R: TempReg.
	"mask off class index"
	self AndCq: objectMemory classIndexMask R: TempReg.
	
	"2. Read the class from class index"
	jumpIsImm jmpTarget: self Label.
	objectRepresentation genGetClassObjectOfClassIndex: TempReg into: destReg scratchReg: topReg.
	self ssPop: 1.
	^self ssPushRegister: destReg! !

!SistaCogitClone methodsFor: 'inline primitive unary generators' stamp: 'cb 11/23/2017 14:20'!
genUnaryConvertInlinePrimitive: primIndex
	| resultReg |
	self assert: (primIndex between: 30 and: 32).
	primIndex caseOf: {
		"1030	characterAsInteger
		 Character => 22 bits strictly positive Smi (Unicode)"
		[30] -> [resultReg := self allocateRegForStackEntryAt: 0. 
				self ssTop popToReg: resultReg.
				objectRepresentation genConvertCharacterToSmallIntegerInReg: resultReg].
		"1031	smallFloatAsInteger
		 SmallFloat => Smi"
		[31] -> [^EncounteredUnknownBytecode "to implement"].
		"1032	smiAsFloat
		 Smi => SmallFloat"
		[32] -> [resultReg := self allocateRegForStackEntryAt: 0. 
				self ssTop popToReg: resultReg.
				self assert: self processorHasDoublePrecisionFloatingPointSupport.
				self MoveR: resultReg R: TempReg.
				self genConvertSmallIntegerToIntegerInReg: TempReg.
				self ConvertR: TempReg Rd: DPFPReg0.
				self flag: #TODO. "Should never fail"
				self
					genAllocFloatValue: DPFPReg0
					into: resultReg
					scratchReg: TempReg
					scratchReg: NoReg. "scratch2 for V3 only"] 
	}.
	self ssPop: 1.
	^self ssPushRegister: resultReg! !

!SistaCogitClone methodsFor: 'inline primitive unary generators' stamp: 'cb 11/23/2017 14:54'!
genUnaryHashInlinePrimitive: primIndex
	| rcvrReg resultReg |
	self assert: (primIndex between: 20 and: 23).
	primIndex caseOf: {
		"1020	objectIdentityHash
		 non-immediate and non-behavior => 22 bits strictly positive Smi"
		[20] ->	[rcvrReg := self allocateRegForStackEntryAt: 0.
				 resultReg := self allocateRegNotConflictingWith: (self registerMaskFor: rcvrReg).
				 self ssTop popToReg: rcvrReg.
				 objectRepresentation genGetIdentityHash: rcvrReg resultReg: resultReg].
		"1021	smiIdentityHash
		 Smi => Smi"
		[21] -> [resultReg := self allocateRegForStackEntryAt: 0. 
				self ssTop popToReg: resultReg].		
		"1022	charIdentityHash
		 Character => 22 bits strictly positive Smi"
		[22] -> [resultReg := self allocateRegForStackEntryAt: 0. 
				self ssTop popToReg: resultReg.
				objectRepresentation genConvertCharacterToSmallIntegerInReg: resultReg].
		"1023	smallfloatIdentityHash
		 SmallFloat => Smi"
		[23] -> [resultReg := self allocateRegForStackEntryAt: 0. 
				objectRepresentation genConvertSmallFloatToSmallFloatHashAsIntegerInReg: resultReg scratch: TempReg].
		"1024	behaviorIdentityHash
		 Behavior => 22 bits strictly positive Smi"
		[24] -> [^EncounteredUnknownBytecode "not implemented"].	
		}.
		self ssPop: 1.
		^self ssPushRegister: resultReg! !

!SistaCogitClone methodsFor: 'inline primitive unary generators' stamp: 'cb 12/14/2017 14:56'!
genUnaryInlinePrimitive: primIndex
	"Bulk comments: each sub-method has its own comment with the specific case.
	1000	rawClass
	not a forwarder => Behavior (Same as class special send, but receiver is not a forwarder)
	1001	numSlots
	pointer object => Smi between 0 and SmallInteger maxVal // 4 - 1 (Answers total size in pointer-sized slots)
	1002	numBytes
	byte object => Smi between 0 and SmallInteger maxVal - 9 (Includes compiled code)
	1003	numShorts
	short object => Smi between 0 and SmallInteger maxVal - 9
	1004	numWords
	word object => Smi between 0 and SmallInteger maxVal - 9
	1005	numDoubleWords
	double word object => Smi between 0 and SmallInteger maxVal - 9
	1011	RawNew
	literal which is a fixed-sized behavior => instance of the receiver with fields nilled out
	1012	RawNewNoInit
	literal which is a fixed-sized behavior => instance of the receiver (Fields of returned value contain undefined data)
	1020	objectIdentityHash
	non-immediate and non-behavior => 22 bits strictly positive Smi
	1021	smiIdentityHash
	Smi => Smi
	1022	charIdentityHash
	Character => 22 bits strictly positive Smi
	1023	smallfloatIdentityHash
	SmallFloat => Smi
	1024	behaviorIdentityHash
	Behavior => 22 bits strictly positive Smi
	1030	characterAsInteger
	Character => 22 bits strictly positive Smi (Unicode)
	1031	smallFloatAsInteger
	SmallFloat => Smi
	1032	smiAsFloat
	Smi => SmallFloat
	1040	unforward
	Anything => Not a forwarder
	1041	possibleRoot
	non-immediate, not a forwarder => receiver is returned (should be effect-only) (If old, becomes gray and remembered to allow many unchecked stores in a row afterwards)"
	primIndex = 0 ifTrue: [^self genUnaryClassPrimitive].
	primIndex <= 6 ifTrue: [^self genUnarySizeInlinePrimitive: primIndex].
	primIndex < 11 ifTrue: [^EncounteredUnknownBytecode].
	primIndex <= 12 ifTrue: [^self genUnaryNewInlinePrimitive: primIndex].
	primIndex < 20 ifTrue: [^EncounteredUnknownBytecode].
	primIndex <= 24 ifTrue: [^self genUnaryHashInlinePrimitive: primIndex].
	primIndex < 30 ifTrue: [^EncounteredUnknownBytecode].
	primIndex <= 32 ifTrue: [^self genUnaryConvertInlinePrimitive: primIndex].
	primIndex = 39 ifTrue: [^self genUnaryUnforwardNonImmediateInlinePrimitive].
	primIndex = 40 ifTrue: [^self genUnaryUnforwardInlinePrimitive].
	primIndex = 41 ifTrue: [^self genUnaryPossibleRootInlinePrimitive].
	^EncounteredUnknownBytecode! !

!SistaCogitClone methodsFor: 'inline primitive unary generators' stamp: 'cb 3/1/2018 20:58'!
genUnaryNewInlinePrimitive: primIndex 
	"1011	RawNew
	 literal which is a fixed-sized behavior => instance of the receiver with fields nilled out
	1012	RawNewNoInit
	literal which is a fixed-sized behavior => instance of the receiver (Fields of returned value contain undefined data)"
	| resultReg classObj |
	self assert: (self ssTop type = SSConstant).
	classObj := self ssTop constant.
	self assert: (objectMemory isNonImmediate: classObj).
	self assert: (coInterpreter objCouldBeClassObj: classObj).
	self assert: (objectMemory isFixedSizePointerFormat: (objectMemory instSpecOfClassFormat: (objectMemory formatOfClass: classObj))).
	objectMemory classTagForClass: classObj. "Ensure Behavior hash"
	resultReg := self allocateRegNotConflictingWith: 0.
	objectRepresentation genGetInstanceOfPointerClass: classObj into: resultReg initializingIf: primIndex = 11 numVariableSlots: 0.
	self ssPop: 1.
	^self ssPushRegister: resultReg! !

!SistaCogitClone methodsFor: 'inline primitive unary generators' stamp: 'cb 12/21/2017 11:36'!
genUnaryPossibleRootInlinePrimitive
	"1041	possibleRoot
	 non-immediate, not a forwarder => receiver is returned (should be effect-only) (If old, becomes gray and remembered to allow many unchecked stores in a row afterwards)"
	| topReg jmpDestYoung jmpAlreadyRemembered|
	<var: #jmpDestYoung type: #'AbstractInstruction *'>
	<var: #jmpAlreadyRemembered type: #'AbstractInstruction *'>
	topReg := self allocateRegForStackEntryAt: 0 notConflictingWith: 0.
	self ssTop popToReg: topReg.
	self MoveCw: objectMemory storeCheckBoundary R: TempReg.
	 self CmpR: TempReg R: topReg.
	 jmpDestYoung := self JumpBelow: 0.
	 objectRepresentation checkRememberedInTrampoline ifFalse: 
		[jmpAlreadyRemembered := objectRepresentation genCheckRememberedBitOf: topReg scratch: TempReg].
	 objectRepresentation callStoreCheckTrampoline.
	 jmpDestYoung jmpTarget: self Label.
	 objectRepresentation checkRememberedInTrampoline ifFalse: 
		[jmpAlreadyRemembered jmpTarget: self Label].
	self ssPop: 1.
	^self ssPushRegister: topReg! !

!SistaCogitClone methodsFor: 'inline primitive unary generators' stamp: 'cb 11/23/2017 14:20'!
genUnarySizeInlinePrimitive: primIndex 
	| rcvrReg resultReg |
	self assert: (primIndex between: 1 and: 6).
	rcvrReg := self allocateRegForStackEntryAt: 0 notConflictingWith: 0.
	resultReg := self allocateRegNotConflictingWith: (self registerMaskFor: rcvrReg).
	self ssTop popToReg: rcvrReg.
	self ssPop: 1.
	self ssPushRegister: resultReg.
	primIndex caseOf: {
		"1001	numSlots
		 pointer object => Smi between 0 and SmallInteger maxVal // 4 - 1 (Answers total size in pointer-sized slots)"
		[1]	->	[objectRepresentation
					genGetNumSlotsOf: rcvrReg into: resultReg;
					genConvertIntegerToSmallIntegerInReg: resultReg].
		"1002	numBytes
		 byte object => Smi between 0 and SmallInteger maxVal - 9 (Includes compiled code)"
		[2]	->	[objectRepresentation
					genGetNumBytesOf: rcvrReg into: resultReg;
					genConvertIntegerToSmallIntegerInReg: resultReg].
		"1003	numShorts
		 short object => Smi between 0 and SmallInteger maxVal - 9"
		[3]	->	[^EncounteredUnknownBytecode "not implemented"].
		"1004	numWords
		 word object => Smi between 0 and SmallInteger maxVal - 9"
		[5]	->	[^EncounteredUnknownBytecode "not implemented"].
		"1005	numDoubleWords 
		 double word object => Smi between 0 and SmallInteger maxVal - 9"
		[6]	->	[^EncounteredUnknownBytecode "not implemented"].
	}..
	^ 0! !

!SistaCogitClone methodsFor: 'inline primitive unary generators' stamp: 'cb 11/23/2017 11:43'!
genUnaryUnforwardInlinePrimitive
	"1040	unforward
	 Anything => Not a forwarder"
	| topReg |
	topReg := self allocateRegForStackEntryAt: 0 notConflictingWith: 0.
	self ssTop popToReg: topReg.
	objectRepresentation genEnsureOopInRegNotForwarded: topReg scratchReg: TempReg.
	self ssPop: 1.
	^self ssPushRegister: topReg! !

!SistaCogitClone methodsFor: 'inline primitive unary generators' stamp: 'eem 9/9/2019 10:15'!
genUnaryUnforwardNonImmediateInlinePrimitive
	"1039	unforwardNonImmediate
	 non immediate => Not a forwarder"
	"unforwardNonImmediate was used exclusively for literal variables,
	 which are now never forwarders because we scan the zone in postBecome,
	 so this is effectively a noop."
	| topReg |
	topReg := self allocateRegForStackEntryAt: 0 notConflictingWith: 0.
	self ssTop popToReg: topReg.
	"objectRepresentation genEnsureObjInRegNotForwarded: topReg scratchReg: TempReg."
	self ssPop: 1.
	^self ssPushRegister: topReg! !

!SistaCogitClone methodsFor: 'bytecode generators' stamp: 'cb 1/19/2017 12:11'!
genUnconditionalTrapBytecode
	"SistaV1: *	217		Trap"
	self ssFlushTo: simStackPtr.
	self CallRT: ceTrapTrampoline.
	self annotateBytecode: self Label.
	deadCode := true.
	^0! !

!SistaCogitClone methodsFor: 'bytecode generators' stamp: 'cb 4/28/2015 12:53'!
genUnoptimizedSpecialSelectorComparison
	"This method is there because if I put directly the super send in genSpecialSelectorComparison Slang does not correctly translte the code to C, it does not correctly type one of the branchDescriptor to BytecodeDescriptor"
	^ super genSpecialSelectorComparison! !

!SistaCogitClone methodsFor: 'initialization' stamp: 'cb 4/16/2015 16:16'!
generateSistaRuntime
	"Trap sends Sista trap message to context with top of stack, so we don't need any arguments..."
	ceTrapTrampoline := self genTrampolineFor: #ceSistaTrap called: 'ceSistaTrapTrampoline'! !

!SistaCogitClone methodsFor: 'accessing' stamp: 'eem 4/12/2017 09:15'!
getCogCodeZoneThreshold
	<doNotGenerate>
	^methodZone getCogCodeZoneThreshold! !

!SistaCogitClone methodsFor: 'method introspection' stamp: 'cb 4/28/2015 16:19'!
getJumpTargetPCAt: pc
	<api>
	^backEnd jumpTargetPCAt: pc! !

!SistaCogitClone methodsFor: 'initialization' stamp: 'cb 5/5/2015 13:12'!
initialize
	super initialize.
	branchReachedOnlyForCounterTrip := false.
	cogMethodSurrogateClass := (objectMemory ifNil: [self class objectMemoryClass]) wordSize = 4
										ifTrue: [CogSistaMethodSurrogate32]
										ifFalse: [CogSistaMethodSurrogate64]! !

!SistaCogitClone methodsFor: 'initialization' stamp: 'eem 7/4/2014 17:37'!
initializeCodeZoneFrom: startAddress upTo: endAddress
	initialCounterValue := MaxCounterValue.
	super initializeCodeZoneFrom: startAddress upTo: endAddress! !

!SistaCogitClone methodsFor: 'simulation only' stamp: 'eem 1/24/2017 13:34'!
isTrapAt: retpc
	"For stack depth checking."
	<doNotGenerate>
	^(backEnd isCallPrecedingReturnPC: retpc)
	 and: [(backEnd callTargetFromReturnAddress: retpc) = ceTrapTrampoline]! !

!SistaCogitClone methodsFor: 'compile abstract instructions' stamp: 'eem 7/5/2014 17:22'!
maybeAllocAndInitCounters
	<inline: true>
	self assert: counters = 0.
	counterIndex := 0.
	numCounters = 0 ifTrue:
		[^true].
	counters := objectRepresentation allocateCounters: numCounters.
	^counters ~= 0! !

!SistaCogitClone methodsFor: 'compile abstract instructions' stamp: 'eem 6/1/2016 15:20'!
maybeCountCounter
	<inline: true>
	numCounters := numCounters + 1! !

!SistaCogitClone methodsFor: 'compile abstract instructions' stamp: 'eem 11/25/2016 17:58'!
maybeCounterIndex
	<inline: true>
	^counterIndex! !

!SistaCogitClone methodsFor: 'compile abstract instructions' stamp: 'cb 1/8/2018 16:30'!
maybeDealWithUnsafeJumpForDescriptor: descriptor pc: pc latestContinuation: latestContinuation
	"Mapped: 250	backjumpAlwaysInterrupt
	 Unmapped: 6000	backjumpNoInterrupt 
				 7016-7020 jumpWritable/Young
				 8000-8003 type branches
	In all cases the distance is an integer pushed on stack just before with pushIntegerLong:"
	<var: #descriptor type: #'BytecodeDescriptor *'>
	| distance targetPC byte01 byte02 newContinuation |
	newContinuation := latestContinuation.
	descriptor hasUnsafeJump ifTrue:
		[ byte01 := objectMemory fetchByte: pc + 1 ofObject: methodObj.
		  byte02 := objectMemory fetchByte: pc + 2 ofObject: methodObj.
		  distance := self decodePushIntegerLongBefore: pc in: methodObj.
		  targetPC := pc + descriptor numBytes + distance.
		  descriptor isMapped 
			ifTrue: [ byte01 = 250 ifTrue: 
				[ "mapped always interrupt backjump" 
				 self maybeCountFixup: descriptor.
				 self initializeFixupAt: targetPC ] ]
			ifFalse: 
				[ byte02 >> 5 = 2r100 ifTrue: 
					[ "inlined sista primitive" 
					| prim |
					prim := (byte02 bitAnd: 16r1F) << 8 + byte01.
					prim >= 7000
						ifTrue: 
							["branch forward"
							 self maybeCountFixup: descriptor.
							 newContinuation := latestContinuation max: targetPC ]
						ifFalse: 
							[prim >= 6000 ifTrue: 
								["no interrupt back jump"
								self maybeCountFixup: descriptor.
								self initializeFixupAt: targetPC]]]]].
	^newContinuation! !

!SistaCogitClone methodsFor: 'compile abstract instructions' stamp: 'eem 4/6/2015 19:54'!
maybeFreeCounters
	<inline: true>
	counters ~= 0 ifTrue:
		[objectRepresentation freeCounters: counters]! !

!SistaCogitClone methodsFor: 'compaction' stamp: 'eem 7/4/2014 12:12'!
maybeFreeCountersOf: aCogMethod
	"Free any counters in the method."
	<inline: true>
	objectRepresentation freeCounters: aCogMethod counters! !

!SistaCogitClone methodsFor: 'compile abstract instructions' stamp: 'eem 6/1/2016 15:20'!
maybeInitNumCounters
	<inline: true>
	numCounters := 0! !

!SistaCogitClone methodsFor: 'garbage collection' stamp: 'eem 7/5/2014 20:09'!
maybeMarkCountersIn: cogMethod
	"In SIsta Spur counters are held on the heap in pinned objects which must be marked
	 to avoid them being garbage collected.  This is the hook through which that happens."
	<var: #cogMethod type: #'CogMethod *'>
	<inline: true>
	objectRepresentation maybeMarkCounters: cogMethod counters! !

!SistaCogitClone methodsFor: 'compile abstract instructions' stamp: 'eem 11/25/2016 17:58'!
maybeSetCounterIndex: value
	<inline: true>
	counterIndex := value! !

!SistaCogitClone methodsFor: 'compile abstract instructions' stamp: 'cb 1/8/2018 16:31'!
maybeUnsafeJumpContinuation: latestContinuation at: bcpc for: descriptor in: aMethodObj
	<inline: true>
	"Note: ignore backward jumps."
	<var: #descriptor type: #'BytecodeDescriptor *'>
	| distance targetPC byte01 byte02 newContinuation | 
	newContinuation := latestContinuation.
	descriptor hasUnsafeJump ifTrue:
		[ byte01 := objectMemory fetchByte: bcpc + 1 ofObject: aMethodObj.
		  byte02 := objectMemory fetchByte: bcpc + 2 ofObject: aMethodObj.
		  "pushIntegerLong"
		  distance := self decodePushIntegerLongBefore: bcpc in: methodObj.
		  targetPC := bcpc + descriptor numBytes + distance.
		  descriptor isMapped 
			ifFalse: 
				[ byte02 >> 5 = 2r100 ifTrue: 
					[ "inlined sista primitive" 
					| prim |
					prim := (byte02 bitAnd: 16r1F) << 8 + byte01.
					prim >= 7000
						ifTrue: 
							["branch forward"
							 newContinuation := latestContinuation max: targetPC ]]]].
	^newContinuation! !

!SistaCogitClone methodsFor: 'simulation only' stamp: 'cb 4/26/2016 09:27'!
methodsCompiledToMachineCodeInto: arrayObject
	<doNotGenerate>
	^methodZone methodsCompiledToMachineCodeInto: arrayObject! !

!SistaCogitClone methodsFor: 'simulation only' stamp: 'cb 4/26/2016 09:27'!
numMethods
	<doNotGenerate>
	^methodZone numMethods! !

!SistaCogitClone methodsFor: 'method introspection' stamp: 'eem 11/7/2016 08:32'!
picDataFor: descriptor Annotation: isBackwardBranchAndAnnotation Mcpc: mcpc Bcpc: bcpc Method: cogMethodArg
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<var: #mcpc type: #'char *'>
	<var: #cogMethodArg type: #'void *'>
	| annotation entryPoint tuple counter |
	"N.B. Counters are always 32-bits, having two 16-bit halves for the reached and taken counts."
	<var: #counter type: #'unsigned int'>

	descriptor ifNil:
		[^0].
	descriptor isBranch ifTrue:
		["it's a branch; conditional?"
		 (descriptor isBranchTrue or: [descriptor isBranchFalse]) ifTrue:
			[counter := (self
							cCoerce: ((self
											cCoerceSimple: cogMethodArg
											to: #'CogMethod *') counters)
							to: #'usqInt *')
								at: counterIndex.
			 tuple := self picDataForCounter: counter at: bcpc + 1.
			 tuple = 0 ifTrue: [^PrimErrNoMemory].
			 objectMemory storePointer: introspectionDataIndex ofObject: introspectionData withValue: tuple.
			 introspectionDataIndex := introspectionDataIndex + 1.
			 counterIndex := counterIndex + 1].
		 ^0].
	annotation := isBackwardBranchAndAnnotation >> 1.
	((self isPureSendAnnotation: annotation)
	 and: [entryPoint := backEnd callTargetFromReturnAddress: mcpc asUnsignedInteger.
		 entryPoint > methodZoneBase]) ifFalse: "send is not linked, or is not a send"
		[^0].
	self targetMethodAndSendTableFor: entryPoint "It's a linked send; find which kind."
		annotation: annotation
		into: [:targetMethod :sendTable| | methodClassIfSuper association |
			methodClassIfSuper := nil.
			sendTable = superSendTrampolines ifTrue:
				[methodClassIfSuper := coInterpreter methodClassOf: (self cCoerceSimple: cogMethodArg to: #'CogMethod *') methodObject].
			sendTable = directedSuperSendTrampolines ifTrue:
				[association := backEnd literalBeforeInlineCacheTagAt: mcpc asUnsignedInteger.
				 methodClassIfSuper := objectRepresentation valueOfAssociation: association].
			tuple := self picDataForSendTo: targetMethod
						methodClassIfSuper: methodClassIfSuper
						at: mcpc
						bcpc: bcpc + 1].
	tuple = 0 ifTrue: [^PrimErrNoMemory].
	objectMemory storePointer: introspectionDataIndex ofObject: introspectionData withValue: tuple.
	introspectionDataIndex := introspectionDataIndex + 1.
	^0! !

!SistaCogitClone methodsFor: 'method introspection' stamp: 'eem 4/19/2016 15:20'!
picDataFor: cogMethod into: arrayObj
	"Collect the branch and send data for cogMethod, storing it into arrayObj."
	<api>
	<var: #cogMethod type: #'CogMethod *'>
	| errCode |
	cogMethod stackCheckOffset = 0 ifTrue:
		[^0].
	introspectionDataIndex := counterIndex := 0.
	introspectionData := arrayObj.
	errCode := self
					mapFor: (self cCoerceSimple: cogMethod to: #'CogBlockMethod *')
					bcpc: (coInterpreter startPCOfMethod: cogMethod methodObject)
					performUntil: #picDataFor:Annotation:Mcpc:Bcpc:Method:
					arg: cogMethod asVoidPointer.
	errCode ~= 0 ifTrue:
		[self assert: errCode = PrimErrNoMemory.
		 ^-1].
	cogMethod blockEntryOffset ~= 0 ifTrue:
		[errCode := self blockDispatchTargetsFor: cogMethod
						perform: #picDataForBlockEntry:Method:
						arg: cogMethod asInteger.
		 errCode ~= 0 ifTrue:
			[self assert: errCode = PrimErrNoMemory.
			 ^-1]].
	^introspectionDataIndex! !

!SistaCogitClone methodsFor: 'method introspection' stamp: 'eem 4/19/2016 15:07'!
picDataForBlockEntry: blockEntryMcpc Method: cogMethod
	"Collect the branch and send data for the block method starting at blockEntryMcpc, storing it into picData."
	<returnTypeC: #usqInt>
	| cogBlockMethod |
	<var: #cogBlockMethod type: #'CogBlockMethod *'>
	cogBlockMethod := self cCoerceSimple: blockEntryMcpc - (self sizeof: CogBlockMethod)
							  to: #'CogBlockMethod *'.
	cogBlockMethod stackCheckOffset = 0 ifTrue:
		[^0].
	^self
		mapFor: cogBlockMethod
		bcpc: cogBlockMethod startpc
		performUntil: #picDataFor:Annotation:Mcpc:Bcpc:Method:
		arg: cogMethod asVoidPointer! !

!SistaCogitClone methodsFor: 'method introspection' stamp: 'eem 11/7/2016 08:32'!
picDataForCounter: counter at: bcpc
	| executedCount tuple untakenCount |
	"N.B. Counters are always 32-bits, having two 16-bit halves for the reached and taken counts."
	<var: #counter type: #'unsigned int'>
	tuple := objectMemory
				eeInstantiateClassIndex: ClassArrayCompactIndex
				format: objectMemory arrayFormat
				numSlots: 3.
	tuple = 0 ifTrue:
		[^0].
	self assert: CounterBytes = 4.
	executedCount := initialCounterValue - (counter >> 16).
	untakenCount := initialCounterValue - (counter bitAnd: 16rFFFF).
	objectMemory
		storePointerUnchecked: 0 ofObject: tuple withValue: (objectMemory integerObjectOf: bcpc);
		storePointerUnchecked: 1 ofObject: tuple withValue: (objectMemory integerObjectOf: executedCount);
		storePointerUnchecked: 2 ofObject: tuple withValue: (objectMemory integerObjectOf: untakenCount).
	^tuple! !

!SistaCogitClone methodsFor: 'method introspection' stamp: 'eem 4/3/2015 16:08'!
picDataForSendTo: cogMethod methodClassIfSuper: methodClassOrNil at: sendMcpc bcpc: sendBcpc
	"Answer a tuple with the send data for a linked send to cogMethod.
	 If the target is a CogMethod (monomorphic send) answer
		{ bytecode pc, inline cache class, target method }
	 If the target is an open PIC (megamorphic send) answer
		{ bytecode pc, nil, send selector }
	If the target is a closed PIC (polymorphic send) answer
		{ bytecode pc, first class, target method, second class, second target method, ... }"
	<var: #cogMethod type: #'CogMethod *'>
	<var: #sendMcpc type: #'char *'>
	| tuple class |
	tuple := objectMemory
					eeInstantiateClassIndex: ClassArrayCompactIndex
					format: objectMemory arrayFormat
					numSlots: (cogMethod cmType = CMClosedPIC
								ifTrue: [2 * cogMethod cPICNumCases + 1]
								ifFalse: [3]).
	tuple = 0 ifTrue:
		[^0].
	objectMemory storePointerUnchecked: 0 ofObject: tuple withValue: (objectMemory integerObjectOf: sendBcpc).
	cogMethod cmType = CMMethod ifTrue:
		[class := methodClassOrNil ifNil:
					[objectRepresentation classForInlineCacheTag: (backEnd inlineCacheTagAt: sendMcpc asUnsignedInteger)].
		 objectMemory
			storePointer: 1 ofObject: tuple withValue: class;
			storePointer: 2 ofObject: tuple withValue: cogMethod methodObject.
		^tuple].
	cogMethod cmType = CMClosedPIC ifTrue:
		[self populate: tuple withPICInfoFor: cogMethod firstCacheTag: (backEnd inlineCacheTagAt: sendMcpc asUnsignedInteger).
		^tuple].
	cogMethod cmType = CMOpenPIC ifTrue:
		[objectMemory
			storePointerUnchecked: 1 ofObject: tuple withValue: objectMemory nilObject;
			storePointer: 2 ofObject: tuple withValue: cogMethod selector.
		^tuple].
	self error: 'invalid method type'.
	^0 "to get Slang to type this method as answering sqInt"! !

!SistaCogitClone methodsFor: 'method introspection' stamp: 'eem 3/14/2016 11:38'!
populate: tuple withPICInfoFor: cPIC firstCacheTag: firstCacheTag
	"Populate tuple (which must be large enough) with the ClosedPIC's target method class pairs.
	 The first entry in tuple contains the bytecode pc for the send, so skip the tuple's first field."
	<var: #cPIC type: #'CogMethod *'>
	| pc cacheTag classOop entryPoint targetMethod value |
	<var: #targetMethod type: #'CogMethod *'>

	1 to: cPIC cPICNumCases do:
		[:i|
		pc := self addressOfEndOfCase: i inCPIC: cPIC.
		cacheTag := i = 1
						ifTrue: [firstCacheTag]
						ifFalse: [backEnd literalBeforeFollowingAddress: pc - backEnd jumpLongConditionalByteSize].
		classOop := objectRepresentation classForInlineCacheTag: cacheTag.
		objectMemory storePointer: i * 2 - 1 ofObject: tuple withValue: classOop.
		entryPoint := i = 1
						ifTrue: [backEnd jumpLongTargetBeforeFollowingAddress: pc]
						ifFalse: [backEnd jumpLongConditionalTargetBeforeFollowingAddress: pc].
		"Find target from jump.  A jump to the MNU entry-point should collect #doesNotUnderstand:"
		(cPIC containsAddress: entryPoint)
			ifTrue:
				[value := objectMemory splObj: SelectorDoesNotUnderstand]
			ifFalse:
				[targetMethod := self cCoerceSimple: entryPoint - cmNoCheckEntryOffset to: #'CogMethod *'.
				 self assert: targetMethod cmType = CMMethod.
				 value := targetMethod methodObject].
		objectMemory storePointer: i * 2 ofObject: tuple withValue: value]! !

!SistaCogitClone methodsFor: 'tests' stamp: 'eem 2/14/2011 15:35'!
printPICDataForMethods
	<doNotGenerate>
	methodZone methodsDo:
		[:cogMethod|
		cogMethod cmType = CMMethod ifTrue:
			[(coInterpreter picDataFor: cogMethod) ifNotNil:
				[:thePicData|
				coInterpreter printOop: thePicData]]]! !

!SistaCogitClone methodsFor: 'sista callbacks' stamp: 'eem 7/4/2014 12:23'!
resetCountersIn: cogMethod
	<doNotGenerate>
	objectRepresentation resetCountersIn: cogMethod! !

!SistaCogitClone methodsFor: 'accessing' stamp: 'eem 3/16/2017 16:49'!
setCogCodeZoneThreshold: threshold
	<doNotGenerate>
	^methodZone setCogCodeZoneThreshold: threshold! !

!SistaRegisterAllocatingCogit class methodsFor: 'clone maintennance' stamp: 'eem 12/12/2016 11:18'!
dateAndTimeFrom: timeStampString
	^(timeStampString copyFrom: (timeStampString findFirst: [:c| c isDigit]) to: timeStampString size) asDateAndTime! !

!SistaRegisterAllocatingCogit class methodsFor: 'clone maintennance' stamp: 'eem 12/12/2016 10:36'!
implementation: aMethodReference isDifferentThan: bMethodReference
	^aMethodReference isValid
	 and: [bMethodReference isValid
			ifTrue: [aMethodReference sourceString ~= bMethodReference sourceString]
			ifFalse: [true]]! !

!SistaRegisterAllocatingCogit class methodsFor: 'clone maintennance' stamp: 'eem 12/12/2016 11:51'!
implementation: aMethodReference isNewerThan: bMethodReference
	^(self dateAndTimeFrom: aMethodReference timeStamp) > (self dateAndTimeFrom: bMethodReference timeStamp)! !

!SistaRegisterAllocatingCogit class methodsFor: 'class initialization' stamp: 'eem 12/12/2016 09:21'!
initializeWithOptions: optionsDictionary

	self syncCodeWithSistaCogit ifTrue: "The subclass's #initializeWithOptions: method changed; resend."
		[^self initializeWithOptions: optionsDictionary].

	"Things are up-to-date; continue"
	^super initializeWithOptions: optionsDictionary! !

!SistaRegisterAllocatingCogit class methodsFor: 'class initialization' stamp: 'eem 12/12/2016 11:14'!
syncCodeIn: sourceClass with: destClass leavingUnchanged: selectorOrNil
	"Make sure that the SistaRegisterAllocatingCogit/SistaCogitClone combination is up-to-date w.r.t. SistaCogit.
	 - 	SistaCogitClone should have all methods in SistaCogit except those implemented in SistaRegisterAllocatingCogit.
		This means that super sends in SistaCogitClone will be understood in SistaRegisterAllocatingCogit, not
		SistaRegisterAllocatingCogit.
	 -	newer methods in SistaCogitClone that are implemented in SistaRegisterAllocatingCogit should be moved up to
		SistaRegisterAllocatingCogit becaude it means that we probably changed them in SistaCogitClone by mistake
		and forgot to copy them up.
	 -	the same goes for the class side, except for the intializeWithOptions: method which /should/ exist in both
		SistaRegisterAllocatingCogit and SistaCogitClone, because it runs this initialization."

	| methodTimeStamp |
	methodTimeStamp := selectorOrNil ifNotNil:
					[(destClass superclass >>> selectorOrNil) isValid ifTrue:
						[(destClass >>> selectorOrNil) isValid
							ifTrue: [(destClass >>> selectorOrNil) timeStamp]
							ifFalse: ['ancient 01/01/1901 00:00']]].

	destClass selectorsDo:
		[:s|
		(s ~~ selectorOrNil
		 and: [(destClass superclass >>> s) isValid
		 and: [self implementation: destClass >>> s isNewerThan: destClass superclass >>> s]]) ifTrue:
			[destClass superclass recompile: s from: destClass.
			 (destClass superclass whichCategoryIncludesSelector: s) ~= (destClass whichCategoryIncludesSelector: s) ifTrue:
				[destClass superclass organization classify: s under: (destClass whichCategoryIncludesSelector: s)]]].

	sourceClass selectorsDo:
		[:s|
		(self implementation: sourceClass >>> s isDifferentThan: destClass >>> s) ifTrue:
			[destClass recompile: s from: sourceClass.
			 (destClass whichCategoryIncludesSelector: s) ~= (sourceClass whichCategoryIncludesSelector: s) ifTrue:
				[destClass organization classify: s under: (sourceClass whichCategoryIncludesSelector: s)]]].


	destClass superclass selectorsDo:
		[:s|
		(s ~~ selectorOrNil
		 and: [(destClass >>> s) isValid]) ifTrue:
			[destClass removeSelector: s]].

	^methodTimeStamp notNil
	  and: [(destClass >>> selectorOrNil) isValid
	  and: [(methodTimeStamp beginsWith: 'ancient')
		or: [(self dateAndTimeFrom: (destClass >>> selectorOrNil) timeStamp) ~= (self dateAndTimeFrom: methodTimeStamp)]]]! !

!SistaRegisterAllocatingCogit class methodsFor: 'class initialization' stamp: 'eem 2/20/2018 16:28'!
syncCodeWithSistaCogit
	"Make sure that the SistaRegisterAllocatingCogit/SistaCogitClone combination is up-to-date w.r.t. SistaCogit.
	 - 	SistaCogitClone should have all methods in SistaCogit except those implemented in SistaRegisterAllocatingCogit.
		This means that super sends in SistaCogitClone will be understood in SistaRegisterAllocatingCogit, not
		RegisterAllocatingCogit.
	 -	newer methods in SistaCogitClone that are implemented in SistaRegisterAllocatingCogit should be moved up to
		SistaRegisterAllocatingCogit because it means that we probably changed them in SistaCogitClone by mistake
		and forgot to copy them up.
	 -	the same goes for the class side, except for the intializeWithOptions: method which /should/ exist in both
		SistaRegisterAllocatingCogit and SistaCogitClone, because it runs this initialization."
	| syncAction |
	syncAction :=
		[self syncCodeIn: SistaCogit with: SistaCogitClone leavingUnchanged: nil.
		 self syncCodeIn: SistaCogit class with: SistaCogitClone class leavingUnchanged: #initializeWithOptions:].
	^(Smalltalk classNamed: #CurrentReadOnlySourceFiles)
		ifNil: syncAction
		ifNotNil: [:crosf| crosf cacheDuring: syncAction]! !

!SistaRegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 02/07/2017 15:59'!
genCounterTripOnlyJumpIf: boolean to: targetBytecodePC 
	"Specific version if the branch is only reached while falling through if the counter trips after an inlined #== branch. We do not regenerate the counter logic in this case to avoid 24 bytes instructions."
	
	<var: #ok type: #'AbstractInstruction *'>
	<var: #mustBeBooleanTrampoline type: #'AbstractInstruction *'>

	| ok mustBeBooleanTrampoline |

	extA := 0.
	
	self ssTop popToReg: TempReg.
	self ssPop: 1.

	counterIndex := counterIndex + 1. "counters are increased / decreased in the inlined branch"

	"We need SendNumArgsReg because of the mustBeBooleanTrampoline"
	self ssAllocateRequiredReg: SendNumArgsReg.
	self MoveCq: 1 R: SendNumArgsReg.
	
	"The first time this is reached, it calls necessarily the counter trip for the trampoline because SendNumArgsReg is non zero"
	mustBeBooleanTrampoline := self genCallMustBeBooleanFor: boolean.

	self annotateBytecode: self Label.

	"Cunning trick by LPD.  If true and false are contiguous subtract the smaller.
	 Correct result is either 0 or the distance between them.  If result is not 0 or
	 their distance send mustBeBoolean."
	self assert: (objectMemory objectAfter: objectMemory falseObject) = objectMemory trueObject.
	self genSubConstant: boolean R: TempReg.
	self JumpZero: (self ensureFixupAt: targetBytecodePC).

	self CmpCq: (boolean = objectMemory falseObject
					ifTrue: [objectMemory trueObject - objectMemory falseObject]
					ifFalse: [objectMemory falseObject - objectMemory trueObject])
		R: TempReg.
		
	ok := self JumpZero: 0.
	self MoveCq: 0 R: SendNumArgsReg. "if counterReg is 0 this is a mustBeBoolean, not a counter trip."		

	self Jump: mustBeBooleanTrampoline.
	
	ok jmpTarget: self Label.
	^0! !

!SistaRegisterAllocatingCogit methodsFor: 'bytecode generators' stamp: 'eem 3/17/2020 18:20'!
genForwardersInlinedIdenticalOrNotIf: orNot
	"Override to count inlined branches if followed by a conditional branch.
	 We borrow the following conditional branch's counter and when about to
	 inline the comparison we decrement the counter (without writing it back)
	 and if it trips simply abort the inlining, falling back to the normal send which
	 will then continue to the conditional branch which will trip and enter the abort."
	| nextPC postBranchPC targetBytecodePC branchDescriptor counterReg fixup jumpEqual jumpNotEqual
	  counterAddress countTripped unforwardArg unforwardRcvr argReg rcvrReg regMask 
	  rcvrConstant argConstant |
	<var: #fixup type: #'BytecodeFixup *'>
	<var: #countTripped type: #'AbstractInstruction *'>
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>
	<var: #jumpEqual type: #'AbstractInstruction *'>
	<var: #jumpNotEqual type: #'AbstractInstruction *'>

	((coInterpreter isOptimizedMethod: methodObj) or: [needsFrame not]) ifTrue:
		[^super genForwardersInlinedIdenticalOrNotIf: orNot].

	regMask := 0.
	
	self extractMaybeBranchDescriptorInto: [ :descr :next :postBranch :target | 
		branchDescriptor := descr. nextPC := next. postBranchPC := postBranch. targetBytecodePC := target ].
	
	unforwardRcvr := (self ssValue: 1) mayBeAForwarder.
	unforwardArg := self ssTop mayBeAForwarder.
	(unforwardRcvr not and: [unforwardArg not])
		ifTrue: [unforwardRcvr := true.
				"TODO: use genVanilla with profiling counters (not implemented).
				^self genVanillaInlinedIdenticalOrNotIf: orNot"].
	self assert: (unforwardArg or: [unforwardRcvr]).
	"We use reg for non annotable constants to avoid duplicating objRef."
	rcvrConstant := objectRepresentation isUnannotatableConstant: (self ssValue: 1).
	argConstant := objectRepresentation isUnannotatableConstant: self ssTop.
	
	"If an operand is an annotable constant, it may be forwarded, so we need to store it into a 
	register so the forwarder check can jump back to the comparison after unforwarding the constant.
	However, if one of the operand is an unnanotable constant, does not allocate a register for it 
	(machine code will use operations on constants)."
	rcvrReg:= argReg := NoReg.
	self 
		allocateEqualsEqualsRegistersArgNeedsReg: argConstant not 
		rcvrNeedsReg: rcvrConstant not 
		into: [ :rcvr :arg | rcvrReg:= rcvr. argReg := arg ].
		
	argReg ~= NoReg ifTrue: [ regMask := self registerMaskFor: argReg ].
	rcvrReg ~= NoReg ifTrue: [ regMask := regMask bitOr: (self registerMaskFor: rcvrReg) ].
	
	"Only interested in inlining if followed by a conditional branch."
	(branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse]) ifFalse:
		[^ self 
			genIdenticalNoBranchArgIsConstant: argConstant
			rcvrIsConstant: rcvrConstant
			argReg: argReg 
			rcvrReg: rcvrReg 
			orNotIf: orNot].
	
	unforwardArg ifTrue: [ objectRepresentation genEnsureOopInRegNotForwarded: argReg scratchReg: TempReg ].
	unforwardRcvr ifTrue: [ objectRepresentation genEnsureOopInRegNotForwarded: rcvrReg scratchReg: TempReg ].
	
	counterReg := self allocateRegNotConflictingWith: regMask.
	self 
		genExecutionCountLogicInto: [ :cAddress :countTripBranch | 
			counterAddress := cAddress. 
			countTripped := countTripBranch ] 
		counterReg: counterReg.
	
	self assert: (unforwardArg or: [ unforwardRcvr ]).
	self genCmpArgIsConstant: argConstant rcvrIsConstant: rcvrConstant argReg: argReg rcvrReg: rcvrReg.
	self ssPop: 2.
	
	orNot == branchDescriptor isBranchTrue "orNot is true for ~~"
		ifFalse:
			[ fixup := (self ensureNonMergeFixupAt: postBranchPC) asUnsignedInteger.
			self JumpZero:  (self ensureNonMergeFixupAt: targetBytecodePC) asUnsignedInteger ]
		ifTrue:
			[ fixup := (self ensureNonMergeFixupAt: targetBytecodePC) asUnsignedInteger.
			self JumpZero: (self ensureNonMergeFixupAt: postBranchPC) asUnsignedInteger ].
	
	self genFallsThroughCountLogicCounterReg: counterReg counterAddress: counterAddress.
	self Jump: fixup.
	
	countTripped jmpTarget: self Label.
	
	"inlined version of #== ignoring the branchDescriptor if the counter trips to have normal state for the optimizer"
	self ssPop: -2. 
	self genCmpArgIsConstant: unforwardArg not rcvrIsConstant: unforwardRcvr not argReg: argReg rcvrReg: rcvrReg.
	self ssPop: 2. 
	
	"This code necessarily directly falls through the jumpIf: code which pops the top of the stack into TempReg. 
	We therefore directly assign the result to TempReg to save one move instruction"
	jumpEqual := orNot ifFalse: [self JumpZero: 0] ifTrue: [self JumpNonZero: 0].
	self genMoveFalseR: TempReg.
	jumpNotEqual := self Jump: 0.
	jumpEqual jmpTarget: (self genMoveTrueR: TempReg).
	jumpNotEqual jmpTarget: self Label.
	self ssPushRegister: TempReg.
	
	(self fixupAt: nextPC) notAFixup ifTrue: [ branchReachedOnlyForCounterTrip := true ].
	
	^ 0! !

!SistaRegisterAllocatingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/21/2018 16:17'!
genJumpIf: boolean to: targetBytecodePC
	"The heart of performance counting in Sista.  Conditional branches are 6 times less
	 frequent than sends and can provide basic block frequencies (send counters can't).
	 Each conditional has a 32-bit counter split into an upper 16 bits counting executions
	 and a lower half counting untaken executions of the branch.  Executing the branch
	 decrements the upper half, tripping if the count goes negative.  Not taking the branch
	 decrements the lower half.  N.B. We *do not* eliminate dead branches (true ifTrue:/true ifFalse:)
	 so that scanning for send and branch data is simplified and that branch data is correct."
	<inline: false>
	| ok counterAddress countTripped retry nextPC nextDescriptor desc eventualTarget reg |
	<var: #ok type: #'AbstractInstruction *'>
	<var: #desc type: #'CogSimStackEntry *'>
	<var: #retry type: #'AbstractInstruction *'>
	<var: #countTripped type: #'AbstractInstruction *'>
	<var: #nextDescriptor type: #'BytecodeDescriptor *'>

	"In optimized code we don't generate counters to improve performance"
	(coInterpreter isOptimizedMethod: methodObj) ifTrue:
		[^super genJumpIf: boolean to: targetBytecodePC].
	
	"If the branch is reached only for the counter trip trampoline 
	(typically, var1 == var2 ifTrue: falls through to the branch only for the trampoline)
	we generate a specific path to drastically reduce the number of machine instructions"
	branchReachedOnlyForCounterTrip ifTrue: 
		[branchReachedOnlyForCounterTrip := false.
		 ^self genCounterTripOnlyJumpIf: boolean to: targetBytecodePC].

	"We detect and: / or:, if found, we don't generate the counters to avoid pathological counter slow down"
	boolean = objectMemory falseObject ifTrue:
		[ nextPC := bytecodePC + (self generatorAt: byte0) numBytes.
		  nextDescriptor := self generatorForPC: nextPC.
		  nextDescriptor generator ==  #genPushConstantTrueBytecode ifTrue: [ ^ super genJumpIf: boolean to: targetBytecodePC ].
		  nextDescriptor := self generatorForPC: targetBytecodePC.
		  nextDescriptor generator == #genPushConstantFalseBytecode ifTrue: [ ^ super genJumpIf: boolean to: targetBytecodePC ]. ].

	extA := 0. "We ignore the noMustBeBoolean flag. It should not be present in methods with counters, and if it is we don't care."

	"We don't generate counters on branches on true/false, the basicblock usage can be inferred"
	desc := self ssTop.
	(self stackEntryIsBoolean: desc) ifTrue:
		[ ^ super genJumpIf: boolean to: targetBytecodePC ].

	eventualTarget := self eventualTargetOf: targetBytecodePC.

	self flag: 'Because of the restriction on x64 that absolute loads must target %rax, it would perhaps be a better choice to use TempReg (%rax) for the counter reg and SendNumArgsReg for the boolean.'.
	"try and use the top entry's register if any, but only if it can be destroyed."
	reg := (desc type ~= SSRegister
			or: [(self anyReferencesToRegister: desc register inAllButTopNItems: 0)
			or: [(desc register = ReceiverResultReg and: [self receiverIsInReceiverResultReg])]])
				ifTrue: [TempReg]
				ifFalse: [desc register].
	desc popToReg: reg.
	self ssPop: 1.

	"We need SendNumArgsReg because of the mustBeBooleanTrampoline"
	self ssAllocateRequiredReg: SendNumArgsReg.

	retry := self Label.
	self 
		genExecutionCountLogicInto: [ :cAddress :countTripBranch | 
			counterAddress := cAddress. 
			countTripped := countTripBranch ] 
		counterReg: SendNumArgsReg.
	counterIndex := counterIndex + 1.

	"Cunning trick by LPD.  If true and false are contiguous subtract the smaller.
	 Correct result is either 0 or the distance between them.  If result is not 0 or
	 their distance send mustBeBoolean."
	self assert: (objectMemory objectAfter: objectMemory falseObject) = objectMemory trueObject.
	self genSubConstant: boolean R: reg.
	self JumpZero: (self ensureFixupAt: eventualTarget).

	self genFallsThroughCountLogicCounterReg: SendNumArgsReg counterAddress: counterAddress.

	self CmpCq: (boolean = objectMemory falseObject
					ifTrue: [objectMemory trueObject - objectMemory falseObject]
					ifFalse: [objectMemory falseObject - objectMemory trueObject])
		R: reg.
	ok := self JumpZero: 0.
	self MoveCq: 0 R: SendNumArgsReg. "if counterReg is 0 this is a mustBeBoolean, not a counter trip."
	reg ~= TempReg ifTrue:
		[self MoveR: reg R: TempReg].
	countTripped jmpTarget: self Label.
	self copySimStackToScratch: simSpillBase.
	self ssFlushTo: simStackPtr.
	self genCallMustBeBooleanFor: boolean.
						
	"If we're in an image which hasn't got the Sista code loaded then the ceCounterTripped: trampoline
	 will return directly to machine code, returning the boolean.  So the code should jump back to the
	 retry point. The trampoline preserves register state when taking the ceCounterTripped: path."
	"Clément: For some reason if I write self annotateBytecode: (self Jump: retry) the annotation is not at the correct place."
	"Eliot: Annotations apply the the address following an instruction, and the annotation must be for the return address
	 of the call (since this is the address the run-time sees), so it must be on a label before the jump, not after the jump."
	self annotateBytecode: self Label.
	simSpillBase ~= scratchSpillBase ifTrue:
		[self assert: simSpillBase > scratchSpillBase.
		 self AddCq: simSpillBase - scratchSpillBase * objectMemory wordSize R: SPReg].
	self Jump: retry.

	ok jmpTarget: self Label.
	self restoreSimStackFromScratch.
	^0! !

!SistaRegisterAllocatingCogit methodsFor: 'bytecode generators' stamp: 'eem 1/25/2018 14:05'!
genSpecialSelectorComparison
	"Override to count inlined branches if followed by a conditional branch.
	 We borrow the following conditional branch's counter and when about to
	 inline the comparison we decrement the counter (without writing it back)
	 and if it trips simply abort the inlining, falling back to the normal send which
	 will then continue to the conditional branch which will trip and enter the abort."
	| nextPC postBranchPC targetPC primDescriptor branchDescriptor
	  rcvrIsInt rcvrIsConst argIsIntConst argInt jumpNotSmallInts inlineCAB
	  counterAddress countTripped counterReg index rcvrReg argReg
	   branchToTarget needMergeToContinue needMergeToTarget |
	<var: #countTripped type: #'AbstractInstruction *'>
	<var: #primDescriptor type: #'BytecodeDescriptor *'>
	<var: #jumpNotSmallInts type: #'AbstractInstruction *'>
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>

	(coInterpreter isOptimizedMethod: methodObj) ifTrue:
		[^self genSpecialSelectorComparisonWithoutCounters].

	primDescriptor := self generatorAt: byte0.
	argIsIntConst := self ssTop type = SSConstant
				 and: [objectMemory isIntegerObject: (argInt := self ssTop constant)].
	rcvrIsInt := ((rcvrIsConst := (self ssValue: 1) type = SSConstant)
				  and: [objectMemory isIntegerObject: (self ssValue: 1) constant])
				or: [self mclassIsSmallInteger and: [(self ssValue: 1) isSameEntryAs: self simSelf]].

	"short-cut the jump if operands are SmallInteger constants."
	(argIsIntConst and: [rcvrIsInt and: [rcvrIsConst]]) ifTrue:
		[^ self genStaticallyResolvedSpecialSelectorComparison].

	self extractMaybeBranchDescriptorInto: [ :descr :next :postBranch :target | 
		branchDescriptor := descr. nextPC := next. postBranchPC := postBranch. targetPC := target ].

	"Only interested in inlining if followed by a conditional branch."
	inlineCAB := branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse].
	"Further, only interested in inlining = and ~= if there's a SmallInteger constant involved.
	 The relational operators successfully statically predict SmallIntegers; the equality operators do not."
	(inlineCAB and: [primDescriptor opcode = JumpZero or: [primDescriptor opcode = JumpNonZero]]) ifTrue:
		[inlineCAB := argIsIntConst or: [rcvrIsInt]].
	inlineCAB ifFalse:
		[^self genSpecialSelectorSend].

	"In-line the comparison and the jump, but if the types are not SmallInteger then we will need
	 to do a send and fall through to the following conditional branch.  Since we're allocating values
	 in registers we would like to keep those registers live on the inlined path and reload registers
	 along the non-inlined send path.  The merge logic at the branch destinations handles this."
	argIsIntConst
		ifTrue:
			[rcvrReg := self allocateRegForStackEntryAt: 1.
			 (self ssValue: 1) popToReg: rcvrReg.
			 counterReg := self allocateRegNotConflictingWith: (self registerMaskFor: rcvrReg).
			 argReg := NoReg]
		ifFalse:
			[self allocateRegForStackTopTwoEntriesInto: [:rTop :rNext| argReg := rTop. rcvrReg := rNext].
			 self ssTop popToReg: argReg.
			 (self ssValue: 1) popToReg: rcvrReg.
			 counterReg := self allocateRegNotConflictingWith: (self registerMaskFor: rcvrReg and: argReg)].
	self ssPop: 2.
	jumpNotSmallInts := (rcvrIsInt and: [argIsIntConst]) ifFalse:
							[argIsIntConst
								ifTrue: [objectRepresentation genJumpNotSmallInteger: rcvrReg]
								ifFalse:
									[rcvrIsInt
										ifTrue: [objectRepresentation genJumpNotSmallInteger: argReg]
										ifFalse: [objectRepresentation genJumpNotSmallIntegersIn: rcvrReg and: argReg scratch: TempReg]]].

	self
		genExecutionCountLogicInto: [ :cAddress :countTripBranch | 
			counterAddress := cAddress. 
			countTripped := countTripBranch ] 
		counterReg: counterReg.

	argIsIntConst
		ifTrue: [self CmpCq: argInt R: rcvrReg]
		ifFalse: [self CmpR: argReg R: rcvrReg].
	"self printSimStack; printSimStack: (self fixupAt: postBranchPC) mergeSimStack; printSimStack: (self fixupAt: targetPC) mergeSimStack"
	"If there are merges to be performed on the forward branches we have to execute
	 the merge code only along the path requiring that merge, and exactly once."
	needMergeToTarget := self mergeRequiredForJumpTo: targetPC.
	needMergeToContinue := self mergeRequiredForJumpTo: postBranchPC.
	"Cmp is weird/backwards so invert the comparison."
	(needMergeToTarget and: [needMergeToContinue]) ifTrue:
		[branchToTarget := self genConditionalBranch: (branchDescriptor isBranchTrue
										ifTrue: [primDescriptor opcode]
										ifFalse: [self inverseBranchFor: primDescriptor opcode])
								operand: 0.
		 self Jump: (self ensureFixupAt: postBranchPC).
		 branchToTarget jmpTarget: self Label.
		 self Jump: (self ensureFixupAt: targetPC)].
	(needMergeToTarget and: [needMergeToContinue not]) ifTrue:
		[self genConditionalBranch: (branchDescriptor isBranchFalse
										ifTrue: [primDescriptor opcode]
										ifFalse: [self inverseBranchFor: primDescriptor opcode])
			operand: (self ensureFixupAt: postBranchPC) asUnsignedInteger.
		 self Jump: (self ensureFixupAt: targetPC)].
	(needMergeToTarget not and: [needMergeToContinue]) ifTrue:
		[self genConditionalBranch: (branchDescriptor isBranchTrue
										ifTrue: [primDescriptor opcode]
										ifFalse: [self inverseBranchFor: primDescriptor opcode])
			operand: (self ensureFixupAt: targetPC) asUnsignedInteger.
		 self Jump: (self ensureFixupAt: postBranchPC)].
	(needMergeToTarget or: [needMergeToContinue]) ifFalse:
		[self genConditionalBranch: (branchDescriptor isBranchTrue
										ifTrue: [primDescriptor opcode]
										ifFalse: [self inverseBranchFor: primDescriptor opcode])
			operand: (self ensureFixupAt: targetPC) asUnsignedInteger.
		 self Jump: (self ensureFixupAt: postBranchPC)].

	countTripped jmpTarget: self Label.
	jumpNotSmallInts ifNil:
		[self annotateInstructionForBytecode.
		 deadCode := true.
		 ^0].
	jumpNotSmallInts jmpTarget: countTripped getJmpTarget.
	
	self ssFlushTo: simStackPtr.
	rcvrReg = Arg0Reg
		ifTrue:
			[argReg = ReceiverResultReg
				ifTrue: [self SwapR: Arg0Reg R: Arg0Reg Scratch: TempReg. argReg := Arg0Reg]
				ifFalse: [self MoveR: rcvrReg R: ReceiverResultReg].
			 rcvrReg := ReceiverResultReg].
	argIsIntConst
		ifTrue: [self MoveCq: argInt R: Arg0Reg]
		ifFalse: [argReg ~= Arg0Reg ifTrue: [self MoveR: argReg R: Arg0Reg]].
	rcvrReg ~= ReceiverResultReg ifTrue: [self MoveR: rcvrReg R: ReceiverResultReg].
	index := byte0 - self firstSpecialSelectorBytecodeOffset.
	^self genMarshalledSend: index negated - 1 numArgs: 1 sendTable: ordinarySendTrampolines! !

StackToRegisterMappingCogit subclass: #SistaCogit
	instanceVariableNames: 'numCounters counters counterIndex initialCounterValue ceTrapTrampoline branchReachedOnlyForCounterTrip'
	classVariableNames: 'CounterBytes MaxCounterValue'
	poolDictionaries: 'VMSqueakClassIndices'
	category: 'VMMaker-JIT'!

!SistaCogit commentStamp: 'eem 4/19/2016 14:22' prior: 0!
A SistaCogit is a refinement of RegisterAllocatingCogit that generates code suitable for dynamic optimization by Sista, the Speculative Inlining Smalltalk Architecture, a project by Clément Bera and Eliot Miranda.  Sista is an optimizer that exists in the Smalltalk image, /not/ in the VM,  and optimizes by substituting normal bytecoded methods by optimized bytecoded methods that may use special bytecodes for which the Cogit can generate faster code.  These bytecodes eliminate overheads such as bounds checks or polymorphic code (indexing Array, ByteArray, String etc).  But the bulk of the optimization performed is in inlining blocks and sends for the common path.

The basic scheme is that SistaCogit generates code containing performance counters.  When these counters trip, a callback into the image is performed, at which point Sista analyses some portion of the stack, looking at performance data for the methods on the stack, and optimises based on the stack and performance data.  Execution then resumes in the optimized code.

SistaCogit adds counters to conditional branches.  Each branch has an executed and a taken count, implemented at the two 16-bit halves of a single 32-bit word.  Each counter pair is initialized with initialCounterValue.  On entry to the branch the executed count is decremented and if the count goes below zero the ceMustBeBooleanAdd[True|False] trampoline called.  The trampoline distinguishes between true mustBeBoolean and counter trips because in the former the register temporarily holding the counter value will contain zero.  Then the condition is tested, and if the branch is taken the taken count is decremented.  The two counter values allow an optimizer to collect basic block execution paths and to know what are the "hot" paths through execution that are worth agressively optimizing.  Since conditional branches are about 1/6 as frequent as sends, and since they can be used to determine the hot path through code, they are a better choice to count than, for example, method or block entry.

SistaCogit implements picDataFor:into: that fills an Array with the state of the counters in a method and the state of each linked send in a method.  This is used to implement a primitive used by the optimizer to answer the branch and send data for a method as an Array.

Instance Variables
	counterIndex:			<Integer>
	counterMethodCache:	<CogMethod>
	counters:				<Array of AbstractInstruction>
	initialCounterValue:		<Integer>
	numCounters:			<Integer>
	picData:				<Integer Oop>
	picDataIndex:			<Integer>
	prevMapAbsPCMcpc:	<Integer>

counterIndex
	- xxxxx

counterMethodCache
	- xxxxx

counters
	- xxxxx

initialCounterValue
	- xxxxx

numCounters
	- xxxxx

picData
	- xxxxx

picDataIndex
	- xxxxx

prevMapAbsPCMcpc
	- xxxxx
!

!SistaCogit class methodsFor: 'translation' stamp: 'eem 2/21/2011 17:32'!
additionalHeadersDo: aBinaryBlock
	"Evaluate aBinaryBlock with the names and contents of
	 any additional header files that need to be generated."
	aBinaryBlock
		value: 'cogmethod.h'
		value: SistaCogMethod cogMethodHeader! !

!SistaCogit class methodsFor: 'translation' stamp: 'eem 6/7/2018 15:35'!
ancilliaryClasses
	"Answer any extra classes to be included in the translation."
	^super ancilliaryClasses copyWith: SistaCogMethod! !

!SistaCogit class methodsFor: 'translation' stamp: 'eem 7/5/2014 09:53'!
declareCVarsIn: aCodeGen
	aCodeGen var: 'counters' type: #usqInt! !

!SistaCogit class methodsFor: 'class initialization' stamp: 'eem 4/6/2012 14:55'!
initializeWithOptions: optionsDictionary

	super initializeWithOptions: optionsDictionary.
	CounterBytes := 4.
	MaxCounterValue := (1 << 16) - 1! !

!SistaCogit class methodsFor: 'accessing class hierarchy' stamp: 'eem 3/16/2017 16:52'!
methodZoneClass
	^SistaMethodZone! !

!SistaCogit class methodsFor: 'trampoline support' stamp: 'eem 6/4/2015 12:08'!
numTrampolines
	^super numTrampolines + 1

	"Cogit withAllSubclasses collect: [:c| {c. (c instVarNames select: [:ea| ea beginsWith: 'ce']) size}]"
	"self instVarNames select: [:ea| ea beginsWith: 'ce']"! !

!SistaCogit methodsFor: 'in-line cacheing' stamp: 'cb 12/1/2016 10:15'!
allowEarlyOpenPICPromotion
	<inline: true>
	^ false! !

!SistaCogit methodsFor: 'compile abstract instructions' stamp: 'cb 4/15/2016 11:09'!
compileCogFullBlockMethod: numCopied
	<option: #SistaV1BytecodeSet>
	counters := 0.
	^super compileCogFullBlockMethod: numCopied! !

!SistaCogit methodsFor: 'compile abstract instructions' stamp: 'eem 7/5/2014 17:21'!
compileCogMethod: selector
	counters := 0.
	^super compileCogMethod: selector! !

!SistaCogit methodsFor: 'compile abstract instructions' stamp: 'eem 7/7/2014 07:05'!
compileFrameBuild
	"Override to prefetch counters, if any."
	super compileFrameBuild.
	counters ~= 0 ifTrue:
		[self PrefetchAw: counters]! !

!SistaCogit methodsFor: 'compile abstract instructions' stamp: 'cb 4/7/2016 18:32'!
compileFullBlockMethodFrameBuild: numCopied
	"Override to prefetch counters if any"
	super compileFullBlockMethodFrameBuild: numCopied.
	counters ~= 0 ifTrue:
		[self PrefetchAw: counters]! !

!SistaCogit methodsFor: 'compile abstract instructions' stamp: 'cb 1/8/2018 16:30'!
decodePushIntegerLongBefore: bcpc in: aMethodObj
	<inline: true>
	| distance upperByte | 
	distance := objectMemory fetchByte: bcpc - 1 ofObject: aMethodObj.
	upperByte := objectMemory fetchByte: bcpc - 3 ofObject: aMethodObj.
	upperByte > 127 ifTrue: [upperByte := upperByte - 256].
	^ (upperByte << 8) + distance.
		 ! !

!SistaCogit methodsFor: 'accessing' stamp: 'eem 3/16/2017 16:57'!
defaultCogCodeSize
	"Return the default number of bytes to allocate for native code at startup.
	 The actual value can be set via vmParameterAt: and/or a preference in the ini file."
	<api>
	^2 * backEnd getDefaultCogCodeSize! !

!SistaCogit methodsFor: 'disassembly' stamp: 'eem 1/23/2017 13:23'!
disassembleMethod: surrogateOrAddress on: aStream
	<doNotGenerate>
	| cogMethod |
	cogMethod := super disassembleMethod: surrogateOrAddress on: aStream.
	(cogMethod cmType = CMMethod
	 and: [cogMethod counters ~= 0]) ifTrue:
		[aStream nextPutAll: 'counters:'; cr.
		 0 to: (objectRepresentation numCountersFor: cogMethod counters) - 1 do:
			[:i| | addr |
			 addr := i * CounterBytes + counters.
			 addr printOn: aStream base: 16.
			 aStream nextPut: $:; space.
			 (objectMemory long32At: addr) printOn: aStream base: 16.
			 aStream cr].
		 aStream flush]! !

!SistaCogit methodsFor: 'accessing' stamp: 'cb 6/22/2015 16:19'!
estimateOfAbstractOpcodesPerBytecodes
	"Due to the counter logic, the estimation is higher"
	<inline: true>
	^ 11! !

!SistaCogit methodsFor: 'generate machine code' stamp: 'eem 1/27/2020 13:10'!
fillInCPICHeader: pic numArgs: numArgs numCases: numCases hasMNUCase: hasMNUCase selector: selector
	pic counters: 0.
	super fillInCPICHeader: pic numArgs: numArgs numCases: numCases hasMNUCase: hasMNUCase selector: selector! !

!SistaCogit methodsFor: 'generate machine code' stamp: 'eem 9/23/2013 07:23'!
fillInCounters: nCounters atEndAddress: endAddress
	endAddress - (nCounters * CounterBytes)
		to: endAddress - CounterBytes
		by: CounterBytes
		do: [:address|
			objectMemory
				long32At: address
				put: (initialCounterValue << 16 + initialCounterValue)]! !

!SistaCogit methodsFor: 'generate machine code' stamp: 'eem 7/4/2014 12:08'!
fillInCounters: nCounters atStartAddress: startAddress
	startAddress
		to: startAddress + (nCounters - 1 * CounterBytes)
		by: CounterBytes
		do: [:address|
			objectMemory
				long32At: address
				put: (initialCounterValue << 16 + initialCounterValue)]! !

!SistaCogit methodsFor: 'generate machine code' stamp: 'eem 1/27/2020 13:11'!
fillInMethodHeader: method size: size selector: selector
	super fillInMethodHeader: method size: size selector: selector.
	self fillInCounters: numCounters atStartAddress: counters.
	method counters: counters! !

!SistaCogit methodsFor: 'generate machine code' stamp: 'eem 1/27/2020 13:12'!
fillInOPICHeader: pic numArgs: numArgs selector: selector
	pic counters: 0.
	super fillInOPICHeader: pic numArgs: numArgs selector: selector! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 12/18/2017 17:18'!
genAtPutInlinePrimitive: prim
	"The store check requires rr to be ReceiverResultReg"
	prim caseOf: {
		"3000	pointerAt:put:
		 Mutable pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
		[0] ->	[self genPointerAtPutMaybeContext: false storeCheck: false immutabilityCheck: false].
		"3001	storeCheckPointerAt:put:
		 Mutable pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
		[1] ->	[self genPointerAtPutMaybeContext: false storeCheck: true immutabilityCheck: false].
		"3002	maybeContextPointerAt:put:
		 Mutable pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
		[2] ->	[self genPointerAtPutMaybeContext: true storeCheck: false immutabilityCheck: false].
		"3003	maybeContextStoreCheckPointerAt:put:
		 Mutable pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
		[3] ->	[self genPointerAtPutMaybeContext: true storeCheck: true immutabilityCheck: false].
		"3004	byteAt:put:
		 Mutable byte object, Smi, 8 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)"
		[4] ->	[self genByteAtPut].
	}
	otherwise: [^EncounteredUnknownBytecode].
	^0! !

!SistaCogit methodsFor: 'inline primitive binary generators' stamp: 'cb 3/15/2018 11:19'!
genBinaryAtConstInlinePrimitive: primIndex
	"2064	pointerAt:
	Pointer object (Fixed sized or not) and not a context, Smi =>  (1-based, optimised if arg1 is a constant)
	2065	maybeContextPointerAt:
	Pointer object (Fixed sized or not), Smi =>  (1-based, optimised if arg1 is a constant)
	2066	byteAt:
	byte object, Smi => 8 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2067	shortAt:
	short object, Smi => 16 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2068	wordAt:
	word object, Smi => 32 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2069	doubleWordAt:
	double word object, Smi => 64 bits unsigned Smi or LargePositiveInteger (1-based, optimised if arg1 is a constant)"
	| rr val zeroBasedIndex |
	val := self ssTop constant.
	rr := primIndex = 65 
		ifFalse: [self allocateRegForStackEntryAt: 1]
		ifTrue: [self ssAllocateRequiredReg: ReceiverResultReg.
				self voidReceiverResultRegContainsSelf.
				ReceiverResultReg].
	(self ssValue: 1) popToReg: rr.
	self ssPop: 2.
	zeroBasedIndex := (objectMemory integerValueOf: val) - 1.
	primIndex caseOf: {
		[64] ->	[objectRepresentation genLoadSlot: zeroBasedIndex sourceReg: rr destReg: rr].
		[65] ->	[self ssAllocateRequiredReg: SendNumArgsReg.
				 ^self genPushMaybeContextSlotIndex: zeroBasedIndex].
		[66] ->	[self MoveMb: zeroBasedIndex + objectMemory baseHeaderSize r: rr R: rr.
				 objectRepresentation genConvertIntegerToSmallIntegerInReg: rr]
	}
	otherwise: [^EncounteredUnknownBytecode].
	^self ssPushRegister: rr! !

!SistaCogit methodsFor: 'inline primitive binary generators' stamp: 'cb 12/14/2017 16:59'!
genBinaryAtInlinePrimitive: primIndex
	"2064	pointerAt:
	Pointer object (Fixed sized or not) and not a context, Smi =>  (1-based, optimised if arg1 is a constant)
	2065	maybeContextPointerAt:
	Pointer object (Fixed sized or not), Smi =>  (1-based, optimised if arg1 is a constant)
	2066	byteAt:
	byte object, Smi => 8 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2067	shortAt:
	short object, Smi => 16 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2068	wordAt:
	word object, Smi => 32 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2069	doubleWordAt:
	double word object, Smi => 64 bits unsigned Smi or LargePositiveInteger (1-based, optimised if arg1 is a constant)"
	| ra rr adjust |
	self allocateRegForStackTopTwoEntriesInto: [:rTop :rNext | ra := rTop. rr := rNext ].
	self ssTop popToReg: ra.
	self ssPop: 1.
	self ssTop popToReg: rr.
	self ssPop: 1.
	primIndex caseOf: {
		[64] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1.
				adjust ~= 0 ifTrue: [ self AddCq: adjust R: ra. ]. 
				self MoveXwr: ra R: rr R: rr ].
		[66] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				adjust := objectMemory baseHeaderSize - 1. 
				self AddCq: adjust R: ra.
				self MoveXbr: ra R: rr R: rr.
				objectRepresentation genConvertIntegerToSmallIntegerInReg: rr].
	}
	otherwise: [^EncounteredUnknownBytecode].
	^self ssPushRegister: rr! !

!SistaCogit methodsFor: 'inline primitive binary generators' stamp: 'cb 3/1/2018 15:42'!
genBinaryCompInlinePrimitive: primIndex
	"2032	>
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2033	<
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2034	>=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2035	<=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2036	=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2037	~=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2038	rawEqualsEquals:
	not a forwarder, not a forwarder => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2039	rawNotEqualsEquals:
	not a forwarder, not a forwarder => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)"
	|resultReg otherReg opTrue opFalse invertedOpTrue invertedOpFalse|
	self assert: (primIndex between: 32 and: 39).
	primIndex caseOf: {
		[32]	->	[opTrue := JumpGreater. opFalse := JumpLessOrEqual. invertedOpTrue := JumpLess. invertedOpFalse := JumpGreaterOrEqual].
		[33]	->	[opTrue := JumpLess. opFalse := JumpGreaterOrEqual. invertedOpTrue := JumpGreater. invertedOpFalse := JumpLessOrEqual].
		[34]	->	[opTrue := JumpGreaterOrEqual. opFalse := JumpLess. invertedOpTrue := JumpLessOrEqual. invertedOpFalse := JumpGreater].
		[35]	->	[opTrue := JumpLessOrEqual. opFalse := JumpGreater. invertedOpTrue := JumpGreaterOrEqual. invertedOpFalse := JumpLess].
		[36]	->	[opTrue := JumpZero. opFalse := JumpNonZero. invertedOpTrue := JumpZero. invertedOpFalse := JumpNonZero].
		[37]	->	[opTrue := JumpNonZero. opFalse := JumpZero. invertedOpTrue := JumpNonZero. invertedOpFalse := JumpZero].
		[38]	->	[opTrue := JumpZero. opFalse := JumpNonZero. invertedOpTrue := JumpZero. invertedOpFalse := JumpNonZero].
		[39]	->	[opTrue := JumpNonZero. opFalse := JumpZero. invertedOpTrue := JumpNonZero. invertedOpFalse := JumpZero].
	}.
	"Gen comparison"
	self ssTop type = SSConstant 
		ifTrue: [resultReg := self allocateRegForStackEntryAt: 1.
				(self ssValue: 1) popToReg: resultReg.
				self genCmpConstant: self ssTop constant R: resultReg.
				self ssPop: 2.
				self genBinaryInlineComparison: opTrue opFalse: opFalse destReg: resultReg.
				^self ssPushRegister: resultReg].
	(self ssValue: 1) type = SSConstant 
		ifTrue: [resultReg := self allocateRegForStackEntryAt: 0.
				self ssTop popToReg: resultReg.
				self genCmpConstant: (self ssValue: 1) constant R: resultReg.
				self ssPop: 2.
				self genBinaryInlineComparison: invertedOpTrue opFalse: invertedOpFalse destReg: resultReg.
				^self ssPushRegister: resultReg].
	otherReg := self allocateRegForStackEntryAt: 0.	
	self ssTop popToReg: otherReg.
	resultReg := self allocateRegForStackEntryAt: 1 notConflictingWith: (self registerMaskFor: otherReg).	
	(self ssValue: 1) popToReg: resultReg.
	self CmpR: otherReg R: resultReg.
	self ssPop: 2.
	self genBinaryInlineComparison: opTrue opFalse: opFalse destReg: resultReg.
	^self ssPushRegister: resultReg
	
	! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'eem 4/11/2017 15:08'!
genBinaryConstOpVarInlinePrimitive: prim
	"Const op var version of binary inline primitives."
	"SistaV1: 248		11111000 	iiiiiiii		mjjjjjjj		Call Primitive #iiiiiiii + (jjjjjjj * 256) m=1 means inlined primitive, no hard return after execution.
	 See EncoderForSistaV1's class comment and StackInterpreter>>#binaryInlinePrimitive:"
	<option: #SistaVM>
	| ra val untaggedVal adjust |
	ra := self allocateRegForStackEntryAt: 0.
	self ssTop popToReg: ra.
	self ssPop: 1.
	val := self ssTop constant.
	self ssPop: 1.
	untaggedVal := val - objectMemory smallIntegerTag.
	prim caseOf: {
		"0 through 6, +, -, *, /, //, \\, quo:, SmallInteger op SmallInteger => SmallInteger, no overflow"
		[0]	->	[self AddCq: untaggedVal R: ra].
		[1]	->	[self MoveCq: val R: TempReg.
				 self SubR: ra R: TempReg.
				 objectRepresentation genAddSmallIntegerTagsTo: TempReg.
				 self MoveR: TempReg R: ra].
		[2]	->	[objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: ra.
				 self MoveCq: untaggedVal R: TempReg.
				 self MulR: TempReg R: ra.
				 objectRepresentation genSetSmallIntegerTagsIn: ra].

		"2011 through 2015	Variable-sized pointers new, byte new, 16-bit new, 32-bit new, 64-bit new"

		"2016 through 2020, bitAnd:, bitOr:, bitXor, bitShiftLeft:, bitShiftRight:, SmallInteger op SmallInteger => SmallInteger, no overflow"
		[16] -> [ self AndCq: val R: ra ].
		[17] -> [ self OrCq: val R: ra ].
		[18] -> [ self XorCw: untaggedVal R: ra. ].
		[19] -> [ objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				 self MoveCq: untaggedVal R: TempReg.
				 self LogicalShiftLeftR: ra R: TempReg.
				 objectRepresentation genAddSmallIntegerTagsTo: TempReg.
				 self MoveR: TempReg R: ra].
		[20] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				 self MoveCq: untaggedVal R: TempReg.
				 self ArithmeticShiftRightR: ra R: TempReg.
				 objectRepresentation genClearAndSetSmallIntegerTagsIn: TempReg.
				 self MoveR: TempReg R: ra].

		"2032	through 2037, >, <, >=, <=. =, ~=, SmallInteger op SmallInteger => Boolean (flags?? then in jump bytecodes if ssTop is a flags value, just generate the instruction!!!!)"
		"CmpCqR is SubRCq so everything is reversed, but because no CmpRCq things are reversed again and we invert the sense of the jumps."
		[32] -> [ self CmpCq: val R: ra.
				self genBinaryInlineComparison: JumpLess opFalse: JumpGreaterOrEqual destReg: ra ].
		[33] -> [ self CmpCq: val R: ra.
				self genBinaryInlineComparison: JumpGreater opFalse: JumpLessOrEqual destReg: ra ].
		[34] -> [ self CmpCq: val R: ra.
				self genBinaryInlineComparison: JumpLessOrEqual opFalse: JumpGreater destReg: ra ].
		[35] -> [ self CmpCq: val R: ra.
				self genBinaryInlineComparison: JumpGreaterOrEqual opFalse: JumpLess destReg: ra ].
		[36] -> [ self CmpCq: val R: ra.
				self genBinaryInlineComparison: JumpZero opFalse: JumpNonZero destReg: ra ].
		[37] -> [ self CmpCq: val R: ra.
				self genBinaryInlineComparison: JumpNonZero opFalse: JumpZero destReg: ra ].

		"2064	through 2068, Pointer Object>>at:, Byte Object>>at:, Short16 Word Object>>at: LongWord32 Object>>at: Quad64Word Object>>at:. obj op 0-rel SmallInteger => oop"
		[64] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
				adjust ~= 0 ifTrue: [ self AddCq: adjust R: ra. ]. 
				self genMoveConstant: val R: TempReg.
				self MoveXwr: ra R: TempReg R: ra].
		[65] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				adjust := objectMemory baseHeaderSize - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
				self AddCq: adjust R: ra.
				self genMoveConstant: val R: TempReg.
				self MoveXbr: ra R: TempReg R: ra.
				objectRepresentation genConvertIntegerToSmallIntegerInReg: ra]
	}
	otherwise: [^EncounteredUnknownBytecode].
	self ssPushRegister: ra.
	^0! !

!SistaCogit methodsFor: 'inline primitive binary generators' stamp: 'cb 12/14/2017 14:27'!
genBinaryConstOpVarSmiInlinePrimitive: primIndex
	| ra val untaggedVal |
	ra := self allocateRegForStackEntryAt: 0.
	self ssTop popToReg: ra.
	self ssPop: 1.
	val := self ssTop constant.
	self ssPop: 1.
	untaggedVal := val - objectMemory smallIntegerTag.
	primIndex caseOf: {
		"2000	+
		Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[0]	->	[self AddCq: untaggedVal R: ra].
		"2001	-
		Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[1]	->	[self MoveCq: val R: TempReg.
				 self SubR: ra R: TempReg.
				 objectRepresentation genAddSmallIntegerTagsTo: TempReg.
				 self MoveR: TempReg R: ra].
		"2002	*
		Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[2]	->	[objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: ra.
				 self MoveCq: untaggedVal R: TempReg.
				 self MulR: TempReg R: ra.
				 objectRepresentation genSetSmallIntegerTagsIn: ra].
		
		"2016	bitAnd:
		 Smi, Smi => Smi (optimised if one operand is a constant)"
		[16] -> [ self AndCq: val R: ra ].
		"2017	bitOr:
		Smi, Smi => Smi (optimised if one operand is a constant)"
		[17] -> [ self OrCq: val R: ra ].
		"2018	bitXor:
		Smi, Smi => Smi (optimised if one operand is a constant)"
		[18] -> [ self XorCw: untaggedVal R: ra. ].
	}
	otherwise: [^EncounteredUnknownBytecode].
	self ssPushRegister: ra.
	^0! !

!SistaCogit methodsFor: 'inline primitive binary generators' stamp: 'eem 02/09/2017 09:39'!
genBinaryInlineComparison: opTrue opFalse: opFalse destReg: destReg
	"Inlined comparison. opTrue = jump for true and opFalse = jump for false"
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>
	| nextPC branchDescriptor targetBytecodePC postBranchPC |	
		
	self extractMaybeBranchDescriptorInto: [ :descr :next :postBranch :target | 
		branchDescriptor := descr. nextPC := next. postBranchPC := postBranch. targetBytecodePC := target ].

	(branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse])
		ifTrue: "This is the path where the inlined comparison is followed immediately by a branch"
			[ (self fixupAt: nextPC) notAFixup
				ifTrue: "The next instruction is dead.  we can skip it."
					[deadCode := true.
				 	 self ensureFixupAt: targetBytecodePC.
					 self ensureFixupAt: postBranchPC ]
				ifFalse:
					[self ssPushConstant: objectMemory trueObject]. "dummy value"
			self genConditionalBranch: (branchDescriptor isBranchTrue ifTrue: [opTrue] ifFalse: [opFalse])
				operand: (self ensureNonMergeFixupAt: targetBytecodePC) asUnsignedInteger.
			"We can only elide the jump if the pc after nextPC is the same as postBranchPC.
			 Branch following means it may not be."
			self nextDescriptorExtensionsAndNextPCInto:
				[:iguana1 :iguana2 :iguana3 :followingPC| nextPC := followingPC].
			(deadCode and: [nextPC = postBranchPC]) ifFalse:
				[ self Jump: (self ensureNonMergeFixupAt: postBranchPC) ] ]
		ifFalse: "This is the path where the inlined comparison is *not* followed immediately by a branch"
			[| condJump jump |
			condJump := self genConditionalBranch: opTrue operand: 0.
			self genMoveFalseR: destReg.
	 		jump := self Jump: 0.
			condJump jmpTarget: (self genMoveTrueR: destReg).
			jump jmpTarget: self Label].
	^ 0! !

!SistaCogit methodsFor: 'inline primitive binary generators' stamp: 'cb 12/14/2017 14:26'!
genBinaryInlinePrimitive: primIndex
	"Bulk comments: each sub-method has its own comment with the specific case.
	2000	+
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2001	-
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2002	*
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2003	/
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2004	//
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2005	\
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2006	quo:
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2016	bitAnd:
	Smi, Smi => Smi (optimised if one operand is a constant)
	2017	bitOr:
	Smi, Smi => Smi (optimised if one operand is a constant)
	2018	bitXor:
	Smi, Smi => Smi (optimised if one operand is a constant)
	2019	bitShiftLeft:
	Smi greater or equal to 0, Smi greater or equal to 0 => Smi (no overflow, optimised if arg1 is a constant)
	2020	bitShiftRight:
	Smi, Smi greater or equal to 0 => Smi (optimised if arg1 is a constant)
	2032	>
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2033	<
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2034	>=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2035	<=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2036	=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2037	~=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2038	rawEqualsEquals:
	not a forwarder, not a forwarder => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2039	rawNotEqualsEquals:
	not a forwarder, not a forwarder => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2048	rawNew:
	literal which is a fixed-sized behavior, Smi => instance of receiver, fields nilled out (optimised if arg1 is a constant)
	2049	rawNewNoInit:
	literal which is a fixed-sized behavior, Smi => instance of receiver (Fields of returned value contain undefined data, optimised if arg1 is a constant)
	2064	pointerAt:
	Pointer object (Fixed sized or not) and not a context, Smi =>  (1-based, optimised if arg1 is a constant)
	2065	maybeContextPointerAt:
	Pointer object (Fixed sized or not), Smi =>  (1-based, optimised if arg1 is a constant)
	2066	byteAt:
	byte object, Smi => 8 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2067	shortAt:
	short object, Smi => 16 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2068	wordAt:
	word object, Smi => 32 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2069	doubleWordAt:
	double word object, Smi => 64 bits unsigned Smi or LargePositiveInteger (1-based, optimised if arg1 is a constant)"
	(primIndex <= 18 and: [primIndex <= 2 or: [primIndex > 6]])  ifTrue: 
		[self ssTop type = SSConstant 
			ifTrue: [^self genBinaryVarOpConstSmiInlinePrimitive: primIndex].
		(self ssValue: 1) type = SSConstant 
			ifTrue: [^self genBinaryConstOpVarSmiInlinePrimitive: primIndex].
		^self genBinaryVarOpVarSmiInlinePrimitive: primIndex].
	primIndex <= 6 ifTrue: [^self genDivInlinePrimitive: primIndex].
	primIndex = 19 ifTrue: [^self genBinarySmiBitShiftLeftInlinePrimitive].
	primIndex = 20 ifTrue: [^self genBinarySmiBitShiftRightInlinePrimitive].
	primIndex < 32 ifTrue: [^EncounteredUnknownBytecode].
	primIndex <= 39 ifTrue: [^self genBinaryCompInlinePrimitive: primIndex].
	primIndex < 48 ifTrue: [^EncounteredUnknownBytecode].
	primIndex <= 49 ifTrue: [^self genBinaryNewInlinePrimitive: primIndex].
	primIndex < 64 ifTrue: [^EncounteredUnknownBytecode].
	primIndex <= 69 ifTrue: 
		[self ssTop type = SSConstant
			ifTrue: [^self genBinaryAtConstInlinePrimitive: primIndex].
		^self genBinaryAtInlinePrimitive: primIndex].
	^EncounteredUnknownBytecode! !

!SistaCogit methodsFor: 'inline primitive binary generators' stamp: 'cb 3/1/2018 17:53'!
genBinaryNewInlinePrimitive: primIndex
	"2048	rawNew:
	literal which is a variable-sized behavior, Smi => instance of receiver, fields nilled/zeroed out (optimised if arg1 is a constant)
	2049	rawNewNoInit:
	literal which is a variable-sized behavior, Smi => instance of receiver (Fields of returned value contain undefined data, optimised if arg1 is a constant)"
	| resultReg classObj classFormat argInt |

	"Assertion"
	(self ssValue: 1) type = SSConstant ifFalse: [^UnimplementedOperation].
	
	"It would be nice to deal with variable-sized allocation on non-constant.
	We need to see, however, than inlined allocation cannot deal with large allocations (i.e.
	allocation directly in old space, > 65k currently. So Scorch still needs to perform range analysis
	or generate 2 paths at the bytecode level in this case."
	self ssTop type = SSConstant ifFalse: [^UnimplementedOperation].
	
	classObj := (self ssValue: 1) constant.
	self assert: (objectMemory isNonImmediate: classObj).
	self assert: (coInterpreter objCouldBeClassObj: classObj).
	self deny: (objectMemory isFixedSizePointerFormat: (objectMemory instSpecOfClassFormat: (objectMemory formatOfClass: classObj))).
	objectMemory classTagForClass: classObj. "Ensure Behavior hash"
	
	"Actual code"
	resultReg := self allocateRegNotConflictingWith: 0.
	classFormat := objectMemory instSpecOfClassFormat: (objectMemory formatOfClass: classObj).
	argInt := objectMemory integerValueOf: self ssTop constant.
	self ssPop: 2.
	self ssPushRegister: resultReg.
	(classFormat = objectMemory arrayFormat or: [classFormat = objectMemory indexablePointersFormat]) ifTrue: 
		[^objectRepresentation genGetInstanceOfPointerClass: classObj into: resultReg initializingIf: primIndex = 48 numVariableSlots: argInt].
	(classFormat = objectMemory firstByteFormat) ifTrue: 
		[^objectRepresentation genGetInstanceOfByteClass: classObj into: resultReg initializingIf: primIndex = 48 numBytes: argInt].
	^UnimplementedOperation! !

!SistaCogit methodsFor: 'inline primitive binary generators' stamp: 'cb 12/14/2017 16:50'!
genBinarySmiBitShiftLeftInlinePrimitive
	"2019	bitShiftLeft:
	Smi greater or equal to 0, Smi greater or equal to 0 => Smi (no overflow, optimised if arg1 is a constant)"
	| rr ra |
	rr := self allocateRegForStackEntryAt: 1.
	(self ssValue: 1) popToReg: rr.
	objectRepresentation genRemoveSmallIntegerTagsInScratchReg: rr.
	self ssTop type = SSConstant 
		ifTrue: [self LogicalShiftLeftCq: (objectMemory integerValueOf: self ssTop constant) R: rr]
		ifFalse: [ra := self allocateRegForStackEntryAt: 0 notConflictingWith: (self registerMaskFor: rr).
				self ssTop popToReg: ra.
				objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				self LogicalShiftLeftR: ra R: rr.].
	objectRepresentation genAddSmallIntegerTagsTo: rr.
	self ssPop: 2.
	^self ssPushRegister: rr
! !

!SistaCogit methodsFor: 'inline primitive binary generators' stamp: 'cb 12/14/2017 16:50'!
genBinarySmiBitShiftRightInlinePrimitive
	"2019	bitShiftLeft:
	Smi greater or equal to 0, Smi greater or equal to 0 => Smi (no overflow, optimised if arg1 is a constant)"
	| rr ra |
	rr := self allocateRegForStackEntryAt: 1.
	(self ssValue: 1) popToReg: rr.
	self ssTop type = SSConstant 
		ifTrue: [self ArithmeticShiftRightCq: (objectMemory integerValueOf: self ssTop constant) R: rr]
		ifFalse: [ra := self allocateRegForStackEntryAt: 0 notConflictingWith: (self registerMaskFor: rr).
				self ssTop popToReg: ra.
				objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				 self ArithmeticShiftRightR: ra R: rr].
	objectRepresentation genClearAndSetSmallIntegerTagsIn: rr.
	self ssPop: 2.
	^self ssPushRegister: rr! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 1/13/2017 11:53'!
genBinaryVarOpConstInlinePrimitive: prim
	"Var op const version of inline binary inline primitives."
	"SistaV1: 248		11111000 	iiiiiiii		mjjjjjjj		Call Primitive #iiiiiiii + (jjjjjjj * 256) m=1 means inlined primitive, no hard return after execution.
	 See EncoderForSistaV1's class comment and StackInterpreter>>#binaryInlinePrimitive:"
	<option: #SistaVM>
	| rr val untaggedVal |
	val := self ssTop constant.
	self ssPop: 1.
	rr := self allocateRegForStackEntryAt: 0.
	self ssTop popToReg: rr.
	self ssPop: 1.
	untaggedVal := val - objectMemory smallIntegerTag.
	prim caseOf: {
		"0 through 6, +, -, *, /, //, \\, quo:, SmallInteger op SmallInteger => SmallInteger, no overflow"
		[0]	->	[self AddCq: untaggedVal R: rr].
		[1]	->	[self SubCq: untaggedVal R: rr ].
		[2]	->	[self flag: 'could use MulCq:R'.
				 objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: rr.
				 self MoveCq: untaggedVal R: TempReg.
				 self MulR: TempReg R: rr.
				 objectRepresentation genSetSmallIntegerTagsIn: rr].

		"2016 through 2020, bitAnd:, bitOr:, bitXor, bitShiftLeft:, bitShiftRight:, SmallInteger op SmallInteger => SmallInteger, no overflow"
		[16] -> [ self AndCq: val R: rr ].
		[17] -> [ self OrCq: val R: rr ].
		[18] -> [ self flag: 'could use XorCq:'.
				self XorCw: untaggedVal R: rr. ].
		[19] -> [ objectRepresentation genRemoveSmallIntegerTagsInScratchReg: rr.
				 self LogicalShiftLeftCq: (objectMemory integerValueOf: val) R: rr.
				 objectRepresentation genAddSmallIntegerTagsTo: rr ].
		[20] ->	[self ArithmeticShiftRightCq: (objectMemory integerValueOf: val) R: rr.
				 objectRepresentation genClearAndSetSmallIntegerTagsIn: rr].

		"2032	through 2037, >, <, >=, <=. =, ~=, SmallInteger op SmallInteger => Boolean (flags?? then in jump bytecodes if ssTop is a flags value, just generate the instruction!!!!)"
		"CmpCqR is SubRCq so everything is reversed."
		[32] -> [ self CmpCq: val R: rr.
				self genBinaryInlineComparison: JumpGreater opFalse: JumpLessOrEqual destReg: rr ].
		[33] -> [ self CmpCq: val R: rr.
				self genBinaryInlineComparison: JumpLess opFalse: JumpGreaterOrEqual destReg: rr ].
		[34] -> [ self CmpCq: val R: rr.
				self genBinaryInlineComparison: JumpGreaterOrEqual opFalse: JumpLess destReg: rr ].
		[35] -> [ self CmpCq: val R: rr.
				self genBinaryInlineComparison: JumpLessOrEqual opFalse: JumpGreater destReg: rr ].
		[36] -> [ self CmpCq: val R: rr.
				self genBinaryInlineComparison: JumpZero opFalse: JumpNonZero destReg: rr ].
		[37] -> [ self CmpCq: val R: rr.
				self genBinaryInlineComparison: JumpNonZero opFalse: JumpZero destReg: rr ].

		"2064	through 2068, Pointer Object>>at:, Byte Object>>at:, Short16 Word Object>>at: LongWord32 Object>>at: Quad64Word Object>>at:. obj op 0-rel SmallInteger => oop"
		[64] ->	[objectRepresentation genLoadSlot: (objectMemory integerValueOf: val) - 1 sourceReg: rr destReg: rr].
		[65] ->	[self MoveCq: (objectMemory integerValueOf: val) + objectMemory baseHeaderSize - 1 R: TempReg.
				self MoveXbr: TempReg R: rr R: rr.
				objectRepresentation genConvertIntegerToSmallIntegerInReg: rr]

	}
	otherwise: [^EncounteredUnknownBytecode].
	self ssPushRegister: rr.
	^0! !

!SistaCogit methodsFor: 'inline primitive binary generators' stamp: 'cb 12/14/2017 14:27'!
genBinaryVarOpConstSmiInlinePrimitive: primIndex
	| rr val untaggedVal |
	self assert: primIndex <= 18.
	val := self ssTop constant.
	self ssPop: 1.
	rr := self allocateRegForStackEntryAt: 0.
	self ssTop popToReg: rr.
	self ssPop: 1.
	untaggedVal := val - objectMemory smallIntegerTag.
	primIndex caseOf: {
		"2000	+
		 Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[0]	->	[self AddCq: untaggedVal R: rr].
		"2001	-
		 Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[1]	->	[self SubCq: untaggedVal R: rr ].
		"2002	*
		 Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[2]	->	[self flag: 'could use MulCq:R'.
				 objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: rr.
				 self MoveCq: untaggedVal R: TempReg.
				 self MulR: TempReg R: rr.
				 objectRepresentation genSetSmallIntegerTagsIn: rr].
		
		"2016	bitAnd:
		 Smi, Smi => Smi (optimised if one operand is a constant)"
		[16] -> [ self AndCq: val R: rr ].
		"2017	bitOr:
		 Smi, Smi => Smi (optimised if one operand is a constant)"
		[17] -> [ self OrCq: val R: rr ].
		"2018	bitXor:
		 Smi, Smi => Smi (optimised if one operand is a constant)"
		[18] -> [ self flag: 'could use XorCq:'.
				self XorCw: untaggedVal R: rr. ]

	}
	otherwise: [^EncounteredUnknownBytecode].
	self ssPushRegister: rr.
	^0! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 3/22/2017 14:47'!
genBinaryVarOpVarInlinePrimitive: prim
	"Var op var version of binary inline primitives."
	"SistaV1: 248		11111000 	iiiiiiii		mjjjjjjj		Call Primitive #iiiiiiii + (jjjjjjj * 256) m=1 means inlined primitive, no hard return after execution.
	 See EncoderForSistaV1's class comment and StackInterpreter>>#binaryInlinePrimitive:"
	<option: #SistaVM>
	| ra rr adjust |
	self allocateRegForStackTopTwoEntriesInto: [:rTop :rNext | ra := rTop. rr := rNext ].
	self ssTop popToReg: ra.
	self ssPop: 1.
	self ssTop popToReg: rr.
	self ssPop: 1.
	prim caseOf: {
		"0 through 6, +, -, *, /, //, \\, quo:, SmallInteger op SmallInteger => SmallInteger, no overflow"
		[0]	->	[objectRepresentation genRemoveSmallIntegerTagsInScratchReg: ra.
				 self AddR: ra R: rr].
		[1]	->	[self SubR: ra R: rr.
				 objectRepresentation genAddSmallIntegerTagsTo: rr].
		[2]	->	[self genShiftAwaySmallIntegerTagsInScratchReg: rr.
				 self genRemoveSmallIntegerTagsInScratchReg: ra.
				 self MulR: ra R: rr.
				 self genSetSmallIntegerTagsIn: rr].
		"[4]	->	[].
		[5]	->	[].
		[6]	->	[]."

		"2016 through 2020, bitAnd:, bitOr:, bitXor, bitShiftLeft:, bitShiftRight:, SmallInteger op SmallInteger => SmallInteger, no overflow"
		[16] -> [ self AndR: ra R: rr ].
		[17] -> [ self OrR: ra R: rr ].
		[18] -> [objectRepresentation genRemoveSmallIntegerTagsInScratchReg: ra. 
				self XorR: ra R: rr. ].
		[19] -> [ objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				 objectRepresentation genRemoveSmallIntegerTagsInScratchReg: rr.
				 self LogicalShiftLeftR: ra R: rr.
				 objectRepresentation genAddSmallIntegerTagsTo: rr].
		[20] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				 self ArithmeticShiftRightR: ra R: rr.
				 objectRepresentation genClearAndSetSmallIntegerTagsIn: rr.].


		"2032	through 2037, >, <, >=, <=. =, ~=, SmallInteger op SmallInteger => Boolean (flags?? then in jump bytecodes if ssTop is a flags value, just generate the instruction!!!!)"
		"CmpCqR is SubRCq so everything is reversed."
		[32] -> [ self CmpR: ra R: rr.
				self genBinaryInlineComparison: JumpGreater opFalse: JumpLessOrEqual destReg: rr ].
		[33] -> [ self CmpR: ra R: rr.
				self genBinaryInlineComparison: JumpLess opFalse: JumpGreaterOrEqual destReg: rr ].
		[34] -> [ self CmpR: ra R: rr.
				self genBinaryInlineComparison: JumpGreaterOrEqual opFalse: JumpLess destReg: rr ].
		[35] -> [ self CmpR: ra R: rr.
				self genBinaryInlineComparison: JumpLessOrEqual opFalse: JumpGreater destReg: rr ].
		[36] -> [ self CmpR: ra R: rr.
				self genBinaryInlineComparison: JumpZero opFalse: JumpNonZero destReg: rr ].
		[37] -> [ self CmpR: ra R: rr.
				self genBinaryInlineComparison: JumpNonZero opFalse: JumpZero destReg: rr ].

		"2064	through 2068, Pointer Object>>at:, Byte Object>>at:, Short16 Word Object>>at: LongWord32 Object>>at: Quad64Word Object>>at:. obj op 0-rel SmallInteger => oop"
		[64] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
				adjust ~= 0 ifTrue: [ self AddCq: adjust R: ra. ]. 
				self MoveXwr: ra R: rr R: rr ].
		[65] ->	[objectRepresentation genConvertSmallIntegerToIntegerInReg: ra.
				adjust := objectMemory baseHeaderSize - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
				self AddCq: adjust R: ra.
				self MoveXbr: ra R: rr R: rr.
				objectRepresentation genConvertIntegerToSmallIntegerInReg: rr]

	}
	otherwise: [^EncounteredUnknownBytecode].
	self ssPushRegister: rr.
	^0! !

!SistaCogit methodsFor: 'inline primitive binary generators' stamp: 'cb 12/14/2017 14:27'!
genBinaryVarOpVarSmiInlinePrimitive: primIndex
	| ra rr |
	self allocateRegForStackTopTwoEntriesInto: [:rTop :rNext | ra := rTop. rr := rNext ].
	self ssTop popToReg: ra.
	self ssPop: 1.
	self ssTop popToReg: rr.
	self ssPop: 1.
	primIndex caseOf: {
		"2000	+
		Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[0]	->	[objectRepresentation genRemoveSmallIntegerTagsInScratchReg: ra.
				 self AddR: ra R: rr].
		"2001	-
		Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[1]	->	[self SubR: ra R: rr.
				 objectRepresentation genAddSmallIntegerTagsTo: rr].
		"2002	*
		Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[2]	->	[self genShiftAwaySmallIntegerTagsInScratchReg: rr.
				 self genRemoveSmallIntegerTagsInScratchReg: ra.
				 self MulR: ra R: rr.
				 self genSetSmallIntegerTagsIn: rr].
			
		"2016	bitAnd:
		 Smi, Smi => Smi (optimised if one operand is a constant)"
		[16] -> [ self AndR: ra R: rr ].
		"2017	bitOr:
		Smi, Smi => Smi (optimised if one operand is a constant)"
		[17] -> [ self OrR: ra R: rr ].
		"2018	bitXor:
		Smi, Smi => Smi (optimised if one operand is a constant)"
		[18] -> [objectRepresentation genRemoveSmallIntegerTagsInScratchReg: ra. 
				self XorR: ra R: rr. ]
	}
	otherwise: [^EncounteredUnknownBytecode].
	self ssPushRegister: rr.
	^0! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 12/14/2017 17:00'!
genByteAtPut
	| ra1 ra2 rr adjust |
	self 
		allocateRegForStackTopThreeEntriesInto: [:rTop :rNext :rThird | ra2 := rTop. ra1 := rNext. rr := rThird ] 
		thirdIsReceiver: false.
	self assert: (rr ~= ra1 and: [rr ~= ra2 and: [ra1 ~= ra2]]).
	self ssTop popToReg: ra2.
	self ssPop: 1.
	self ssTop popToReg: ra1.
	self ssPop: 1.
	self ssTop popToReg: rr.
	self ssPop: 1.
	adjust := objectMemory baseHeaderSize - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
	objectRepresentation genConvertSmallIntegerToIntegerInReg: ra1.
	objectRepresentation genConvertSmallIntegerToIntegerInReg: ra2.
	self AddCq: adjust R: ra1.
	self MoveR: ra2 Xbr: ra1 R: rr.
	objectRepresentation genConvertIntegerToSmallIntegerInReg: ra2.
	^self ssPushRegister: ra2! !

!SistaCogit methodsFor: 'mapped inline primitive generators' stamp: 'cb 12/14/2017 17:08'!
genByteAtPutImmutabilityCheck
	| ra1 ra2 rr adjust mutableJump immutabilityFailure indexIsCst |
	<var: #mutableJump type: #'AbstractInstruction *'>
	<var: #immutabilityFailure type: #'AbstractInstruction *'>
	"Assumes rr is not a context and no store check is needed"
	indexIsCst := (self ssValue: 1) type = SSConstant.
	rr := ReceiverResultReg.
	ra1 := TempReg.
	ra2 := ClassReg.
	self voidReceiverResultRegContainsSelf.
	self ssFlushTo: simStackPtr - 1. "we use top value immediately after, so it does not require flush"
	self ssTop popToReg: ra2.
	adjust := objectMemory baseHeaderSize - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
	indexIsCst
		ifFalse:
			[(self ssValue: 1) popToReg: ra1.
			objectRepresentation genConvertSmallIntegerToIntegerInReg: ra1.
			self AddCq: adjust R: ra1]
		ifTrue: [self MoveCq: (objectMemory integerValueOf: (self ssValue: 1) constant) + adjust R: ra1].
	(self ssValue: 2) popToReg: rr.
	self ssPop: 3.
	self ssPushRegister: ra2.

	"Everything is flushed except returned value. All regs are unused except rr ra1 ra2"
	mutableJump := self genJumpMutable: rr scratchReg: Arg0Reg.
	self PushR: ra2. "flush sim stack top"
	self SubCq: 1 + adjust R: ra1. "index 0-relative for trampoline, ra1 unused afterwards"
	objectRepresentation genVarIndexCallStoreTrampoline.
	self PopR: ra2. "restore sim stack top"
	immutabilityFailure := self Jump: 0.
	
	mutableJump jmpTarget: (self MoveR: ra2 R: Arg0Reg).
	objectRepresentation genConvertSmallIntegerToIntegerInReg: Arg0Reg.
	self MoveR: Arg0Reg Xbr: ra1 R: rr.
	
	immutabilityFailure jmpTarget: self Label.
	^0! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 4/10/2017 18:29'!
genByteEqualsInlinePrimitive: prim

	"3021	Byte Object >> equals:length:	
	The receiver and the arguments are both byte objects and have both the same size (length in bytes). 
	The length argument is a smallinteger. 
	Answers true if all fields are equal, false if not. 
	Comparison is bulked to word comparison."
	
	"Overview: 
	 1.	The primitive is called like that: [byteObj1 equals: byteObj2 length: length].
	  	In the worst case we use 5 registers including TempReg 
		and we produce a loop bulk comparing words.
	 2.	The common case is a comparison against a cst: [byteString = 'foo'].
		which produces in Scorch [byteString equals: 'foo' length: 3].
		We try to generate fast code for this case with 3 heuristics:
		- specific fast code if len is a constant
		- unroll the loop if len < 2 * wordSize
		- compile-time reads if str1 or str2 is a constant and loop is unrolled.
		We use 3 registers including TempReg in the common case. 
		We could use 1 less reg if the loop is unrolled, the instr is followed by a branch
		AND one operand is a constant, but this is complicated enough.
	3.	We ignore the case where all operands are constants 
		(We assume Scorch simplifies it, it works but it is not optimised)"
		
	| str1Reg str2Reg lenReg extraReg jmp jmp2 needjmpZeroSize needLoop unroll jmpZeroSize instr lenCst mask |
	<var: #jmp type: #'AbstractInstruction *'>
	<var: #instr type: #'AbstractInstruction *'>
	<var: #jmp2 type: #'AbstractInstruction *'>
	<var: #jmpZeroSize type: #'AbstractInstruction *'>

	"--- quick path for empty string---"
	"This path does not allocate registers and right shift on negative int later in the code.
	 Normally this is resolved by Scorch but we keep it for correctness and consistency"
	self ssTop type = SSConstant ifTrue: 
		[ lenCst := objectMemory integerValueOf: self ssTop constant.
		  lenCst = 0 ifTrue: [ self ssPop: 3. self ssPushConstant: objectMemory trueObject. ^ 0 ] ].

	"--- Allocating & loading registers --- "
	needLoop := (self ssTop type = SSConstant and: [ lenCst <= (objectMemory wordSize * 2) ]) not.
	unroll := needLoop not and: [lenCst > objectMemory wordSize ].
	needLoop 
		ifTrue: 
			[ self assert: self ssTop spilled not.
			  str1Reg := self allocateRegForStackEntryAt: 1 notConflictingWith: self emptyRegisterMask.
			  str2Reg := self allocateRegForStackEntryAt: 2 notConflictingWith: (self registerMaskFor: str1Reg).
			  lenReg := self allocateRegForStackEntryAt: 0 notConflictingWith: (self registerMaskFor: str1Reg and: str2Reg).
			  (self ssValue: 1) popToReg: str1Reg.
			  (self ssValue: 2) popToReg: str2Reg.
			  extraReg := self allocateRegNotConflictingWith: (self registerMaskFor: str1Reg and: str2Reg and: lenReg)]
		ifFalse: 
			[ mask := self emptyRegisterMask.
			  (self ssValue: 1) type = SSConstant ifFalse: 
				[ str1Reg := self allocateRegForStackEntryAt: 1 notConflictingWith: mask.
				  (self ssValue: 1) popToReg: str1Reg.
				  mask := mask bitOr: (self registerMaskFor: str1Reg) ].
			  (self ssValue: 2) type = SSConstant ifFalse: 
				[ str2Reg := self allocateRegForStackEntryAt: 2 notConflictingWith: mask.
				  (self ssValue: 2) popToReg: str2Reg.
				  mask := mask bitOr: (self registerMaskFor: str2Reg) ].
			  extraReg := self allocateRegNotConflictingWith: mask].
	
	"--- Loading LenReg (or statically resolving it) --- "
	"LenReg is loaded with (lenInBytes + objectMemory baseHeaderSize - 1 >> shiftForWord)
	 LenReg is the index for the last word to compare with MoveXwr:r:R:.
	 The loop iterates from LenReg to first word of ByteObj"
	self ssTop type = SSConstant 
		ifTrue: "common case, str = 'foo'. We can precompute lenReg."
			[ lenCst := lenCst + objectMemory baseHeaderSize - 1 >> objectMemory shiftForWord.
			  needLoop ifTrue: [self MoveCq: lenCst R: lenReg ].
			  needjmpZeroSize := false] 
		ifFalse: "uncommon case, str = str2. lenReg in word computed at runtime."
			[ self ssTop popToReg: lenReg.
			  objectRepresentation genConvertSmallIntegerToIntegerInReg: lenReg.
			  self CmpCq: 0 R: lenReg.
			  jmpZeroSize := self JumpZero: 0.
			  needjmpZeroSize := true.
			  self AddCq: objectMemory baseHeaderSize - 1 R: lenReg.
			  self ArithmeticShiftRightCq: objectMemory shiftForWord R: lenReg ].
	
	"--- Comparing the strings --- "
	"LenReg has the index of the last word to read (unless no loop). 
	 We decrement it to adjust -1 (0 in 64 bits) while comparing"
	needLoop 
		ifTrue:
			[instr := self MoveXwr: lenReg R: str1Reg R: extraReg.
			self MoveXwr: lenReg R: str2Reg R: TempReg.
			self CmpR: extraReg R: TempReg.
			jmp := self JumpNonZero: 0. "then string are not equal (jmp target)"
			self AddCq: -1 R: lenReg.
			self CmpCq: (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1 R: lenReg. "first word of ByteObj, stop looping."
			self JumpNonZero: instr]
		ifFalse: "Common case, only 1 or 2 word to check: no lenReg allocation, cst micro optimisations"
			[self genByteEqualsInlinePrimitiveCmp: str1Reg with: str2Reg scratch1: extraReg scratch2: TempReg field: 0.
			jmp := self JumpNonZero: 0. "then string are not equal (jmp target)"
			unroll ifTrue: "unrolling more than twice generate more instructions than the loop so we don't do it"
				[self genByteEqualsInlinePrimitiveCmp: str1Reg with: str2Reg scratch1: extraReg scratch2: TempReg field: 1.
				jmp2 := self JumpNonZero: 0. "then string are not equal (jmp target)"]].
	needjmpZeroSize ifTrue: [ jmpZeroSize jmpTarget: self Label ].
	"fall through, strings are equal"
	
	"--- Pushing the result or pipelining a branch --- "	
	self ssPop: 3.
	self genByteEqualsInlinePrimitiveResult: jmp returnReg: extraReg.
	unroll ifTrue: [jmp2 jmpTarget: jmp getJmpTarget].
	^0! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 3/23/2017 14:33'!
genByteEqualsInlinePrimitiveCmp: str1Reg with: str2Reg scratch1: scratch1Reg scratch2: scratch2Reg field: index
	| shift |
	<inline: true>
	shift := objectMemory baseHeaderSize + (index * objectMemory wordSize).
	(self ssValue: 1) type = SSConstant 
		ifTrue: [self MoveCq: (objectMemory fetchPointer: index ofObject: (self ssValue: 1) constant) R: scratch1Reg]
		ifFalse: [self MoveMw: shift r: str1Reg R: scratch1Reg].
	 (self ssValue: 2) type = SSConstant 
		ifTrue: [self MoveCq: (objectMemory fetchPointer: index ofObject: (self ssValue: 2) constant) R: scratch2Reg]
		ifFalse: [self MoveMw: shift r: str2Reg R: scratch2Reg].
	self CmpR: scratch1Reg R: scratch2Reg.! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 3/23/2017 14:56'!
genByteEqualsInlinePrimitiveResult: jmp returnReg: reg
	"Byte equal is falling through if the result is true, or jumping using jmp if the result is false.
	 The method is required to set the jump target of jmp.
	 We look ahead for a branch and pipeline the jumps if possible..
	 ReturnReg is used only if not followed immediately by a branch."
	| branchDescriptor nextPC postBranchPC targetBytecodePC localJump canElide |
	<var: #localJump type: #'AbstractInstruction *'>
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>
	self extractMaybeBranchDescriptorInto: [ :descr :next :postBranch :target | 
		branchDescriptor := descr. nextPC := next. postBranchPC := postBranch. targetBytecodePC := target ].
	
	"Case 1 - not followed by a branch"
	(branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse])
		ifFalse: 
			[self genMoveTrueR: reg.
			 localJump := self Jump: 0.
			 jmp jmpTarget: (self genMoveFalseR: reg).
			 localJump jmpTarget: self Label.
			 self ssPushRegister: reg.
			^ 0].

	"Case 2 - followed by a branch"
	(self fixupAt: nextPC) notAFixup
		ifTrue: "The next instruction is dead.  we can skip it."
			[deadCode := true.
		 	 self ensureFixupAt: targetBytecodePC.
			 self ensureFixupAt: postBranchPC ]
		ifFalse:
			[self ssPushConstant: objectMemory trueObject]. "dummy value"
	"We can only elide the jump if the pc after nextPC is the same as postBranchPC.
	 Branch following means it may not be."
	self nextDescriptorExtensionsAndNextPCInto:
		[:iguana1 :iguana2 :iguana3 :followingPC| nextPC := followingPC].
	canElide := deadCode and: [nextPC = postBranchPC].
	 branchDescriptor isBranchTrue
		ifTrue: 
			[ self Jump: (self ensureNonMergeFixupAt: targetBytecodePC).
			  canElide 
					ifFalse: [ jmp jmpTarget: (self ensureNonMergeFixupAt: postBranchPC) ]
					ifTrue: [ jmp jmpTarget: self Label ] ]
		ifFalse: [ canElide ifFalse: [ self Jump: (self ensureNonMergeFixupAt: postBranchPC).
				 jmp jmpTarget: (self ensureNonMergeFixupAt: targetBytecodePC) ] ].
	^0! !

!SistaCogit methodsFor: 'mapped inline primitive generators' stamp: 'cb 1/8/2018 14:51'!
genCallMappedInlinedPrimitive
	"SistaV1:	236		11101100	iiiiiiii		callMappedInlinedPrimitive"
	^ self genMappedInlinePrimitive: byte1! !

!SistaCogit methodsFor: 'bytecode generators' stamp: 'cb 11/6/2017 16:51'!
genCallPrimitiveBytecode
	"SistaV1: 248		11111000 	iiiiiiii		mjjjjjjj		Call Primitive #iiiiiiii + (jjjjjjj * 256) m=1 means inlined primitive, no hard return after execution.
	 See EncoderForSistaV1's class comment and StackInterpreter>>#inlinePrimitiveBytecode:"
	| prim primSet |
	byte2 < 128 ifTrue:
		[^bytecodePC = initialPC
			ifTrue: [0]
			ifFalse: [EncounteredUnknownBytecode]].
	prim := byte2 - 128 << 8 + byte1.
	primSet := prim >> 13 bitAnd: 3.
	prim := prim bitAnd: 8191.
	LowcodeVM
		ifTrue:
			[
			primSet = 1 ifTrue: [
				prim < 1000 ifTrue:
					[^self genLowcodeNullaryInlinePrimitive: prim].

				prim < 2000 ifTrue:
					[^self genLowcodeUnaryInlinePrimitive: prim - 1000].
				
				prim < 3000 ifTrue:
					[^ self genLowcodeBinaryInlinePrimitive: prim - 2000].

				prim < 4000 ifTrue:
					[^self genLowcodeTrinaryInlinePrimitive: prim - 3000].
			]
		].
	
	self assert: primSet = 0.
	
	^ self genSistaInlinePrimitive: prim! !

!SistaCogit methodsFor: 'bytecode generator support' stamp: 'eem 02/07/2017 15:59'!
genCounterTripOnlyJumpIf: boolean to: targetBytecodePC 
	"Specific version if the branch is only reached while falling through if the counter trips after an inlined #== branch. We do not regenerate the counter logic in this case to avoid 24 bytes instructions."
	
	<var: #ok type: #'AbstractInstruction *'>
	<var: #mustBeBooleanTrampoline type: #'AbstractInstruction *'>

	| ok mustBeBooleanTrampoline |

	extA := 0.

	self ssFlushTo: simStackPtr - 1.
	
	self ssTop popToReg: TempReg.
	
	self ssPop: 1.

	counterIndex := counterIndex + 1. "counters are increased / decreased in the inlined branch"

	"We need SendNumArgsReg because of the mustBeBooleanTrampoline"
	self ssAllocateRequiredReg: SendNumArgsReg.
	self MoveCq: 1 R: SendNumArgsReg.
	
	"The first time this is reached, it calls necessarily the counter trip for the trampoline because SendNumArgsReg is non zero"
	mustBeBooleanTrampoline := self genCallMustBeBooleanFor: boolean.
	self annotateBytecode: self Label.

	"Cunning trick by LPD.  If true and false are contiguous subtract the smaller.
	 Correct result is either 0 or the distance between them.  If result is not 0 or
	 their distance send mustBeBoolean."
	self assert: (objectMemory objectAfter: objectMemory falseObject) = objectMemory trueObject.
	self genSubConstant: boolean R: TempReg.
	self JumpZero: (self ensureFixupAt: targetBytecodePC).

	self CmpCq: (boolean = objectMemory falseObject
					ifTrue: [objectMemory trueObject - objectMemory falseObject]
					ifFalse: [objectMemory falseObject - objectMemory trueObject])
		R: TempReg.
		
	ok := self JumpZero: 0.
	self MoveCq: 0 R: SendNumArgsReg. "if counterReg is 0 this is a mustBeBoolean, not a counter trip."		

	self Jump: mustBeBooleanTrampoline.
	
	ok jmpTarget: self Label.
	^0! !

!SistaCogit methodsFor: 'mapped inline primitive generators' stamp: 'cb 11/23/2017 14:43'!
genDirectCall
	"250	directCall
	literal index of the method to call on top of stack =>  (variable number of parameters)"
	| newMethod newMethodHeader newMethodArgCount sendTable annotation litIndex |
	self assert: (self ssTop type = SSConstant and: [objectMemory isCompiledMethod: self ssTop constant]).
	litIndex := objectMemory integerValueOf: self ssTop constant.
	newMethod := self getLiteral: litIndex.
	self ssPop: 1.
	self flag: #TODO. "One needs to finish the implementation with the annotation and the sendTable"
	sendTable := 1. "directCallSendTable"
	annotation := 1. "directCallAnnotation"
	newMethodHeader := self rawHeaderOf: newMethod.
	newMethodArgCount := self argumentCountOfMethodHeader: methodHeader.
	"The receiver cannot be a forwader"
	"numArgs >= (NumSendTrampolines - 1) ifTrue:
		[self MoveCq: newMethodArgCount R: SendNumArgsReg]."
	"Load inline cache with method index"
	"self MoveUniqueC32: litIndex R: ClassReg.
	(self Call: (sendTable at: (newMethodArgCount min: NumSendTrampolines - 1))) annotation: annotation.
	self voidReceiverOptStatus.
	self ssPushRegister: ReceiverResultReg."
	^ EncounteredUnknownBytecode "Need to finish implementation"! !

!SistaCogit methodsFor: 'inline primitive binary generators' stamp: 'cb 3/1/2018 20:34'!
genDivInlinePrimitive: primIndex
	"2003	/
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant) "
	"2004	//
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant) "
	"2005	\\
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant) "
	"2006	quo:
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant) "
	
	"We don't deal with constants here. Too complex and does nto bring much"
	| ra rr jumpExact jumpSameSign |
	<var: #jumpExact type: #'AbstractInstruction *'>
	<var: #jumpSameSign type: #'AbstractInstruction *'>
	self allocateRegForStackTopTwoEntriesInto: [:rTop :rNext | ra := rTop. rr := rNext ].
	self ssTop popToReg: ra.
	self ssPop: 1.
	self ssTop popToReg: rr.
	self ssPop: 1.
	self assert: backEnd canDivQuoRem.
	"arg non zero, no overflow check. / remains (unused in optimised code)"
	primIndex caseOf: {
		[4] ->	[objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: ra.
				objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: rr.
				self DivR: ra R: rr Quo: rr Rem: TempReg.
				"If remainder is zero, success"
				self CmpCq: 0 R: TempReg.
				jumpExact := self JumpZero: 0.
				"If arg and remainder signs are different we must round down."
				objectRepresentation genConvertIntegerToSmallIntegerInReg: ra.
				self XorR: TempReg R: ra.
				(self lastOpcode setsConditionCodesFor: JumpZero) ifFalse:
					[self CmpCq: 0 R: ra].
				jumpSameSign := self JumpGreaterOrEqual: 0.
				self SubCq: 1 R: rr.
				jumpSameSign jmpTarget: (jumpExact jmpTarget: self Label).
				objectRepresentation genConvertIntegerToSmallIntegerInReg: rr ].
		[5] ->	[objectRepresentation genRemoveSmallIntegerTagsInScratchReg: ra.
				objectRepresentation genRemoveSmallIntegerTagsInScratchReg: rr.
			 	self DivR: ra R: rr Quo: TempReg Rem: rr.
				"If remainder is zero we're done."
				self CmpCq: 0 R: rr.
				jumpExact := self JumpZero: 0.
				"If arg and remainder signs are different we must reflect around zero."
				self XorR: rr R: ra.
				(self lastOpcode setsConditionCodesFor: JumpZero) ifFalse:
					[self CmpCq: 0 R: ra].
				jumpSameSign := self JumpGreaterOrEqual: 0.
				self XorR: rr R: ra.
				self AddR: ra R: rr.
				jumpSameSign jmpTarget: (jumpExact jmpTarget: self Label).
				objectRepresentation genSetSmallIntegerTagsIn: rr ].
		[6] ->	[objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: ra.
				objectRepresentation genShiftAwaySmallIntegerTagsInScratchReg: rr.
				self DivR: ra R: rr Quo: rr Rem: ra.
				objectRepresentation genConvertIntegerToSmallIntegerInReg: rr].
	} otherwise: [^EncounteredUnknownBytecode].
	self ssPushRegister: rr.
	^0! !

!SistaCogit methodsFor: 'mapped inline primitive generators' stamp: 'cb 11/23/2017 14:31'!
genEnsureEnoughSlots
	"50	EnsureEnoughWords
	 literal which is a Smi => ret value is receiver"
	|slots skip|
	<var: #skip type: #'AbstractInstruction *'>
	self assert: (self ssTop type = SSConstant and: [objectMemory isIntegerObject: self ssTop constant]).
	slots := objectMemory integerValueOf: self ssTop constant.
	self ssFlushTo: simStackPtr - 1.
	self
		MoveAw: objectMemory freeStartAddress R: TempReg;
		CmpCq: objectMemory getScavengeThreshold - (objectMemory bytesPerOop * slots) R: TempReg.
	skip := self JumpBelow: 0.
	objectRepresentation genSetGCNeeded.
	self CallRT: ceCheckForInterruptTrampoline.
	skip jmpTarget: self Label.
	self annotateBytecode: skip getJmpTarget.
	^0! !

!SistaCogit methodsFor: 'bytecode generator support' stamp: 'eem 12/12/2016 11:22'!
genExecutionCountLogicInto: binaryBlock counterReg: counterReg
	<var: #countTripped type: #'AbstractInstruction *'>
	<inline: true>
	| counterAddress countTripped |
	counterAddress := counters + (CounterBytes * counterIndex).
	self MoveA32: counterAddress R: counterReg.
	self SubCq: 16r10000 R: counterReg. "Count executed"
	"If counter trips simply abort the comparison continuing to the following
	 branch *without* writing back.  A double decrement will not trip the second time."
	countTripped := self JumpCarry: 0.
	self MoveR: counterReg A32: counterAddress. "write back"
	binaryBlock value: counterAddress value: countTripped! !

!SistaCogit methodsFor: 'bytecode generator support' stamp: 'eem 9/7/2016 15:03'!
genFallsThroughCountLogicCounterReg: counterReg counterAddress: counterAddress
	<inline: true>
	"Gen this when the branch has not been taken and forwarders have been followed."
	self SubCq: 1 R: counterReg. "Count untaken"
	self MoveR: counterReg A32: counterAddress. "write back"! !

!SistaCogit methodsFor: 'bytecode generators' stamp: 'cb 9/13/2019 12:06'!
genForwardersInlinedIdenticalOrNotIf: orNot
	"Override to count inlined branches if followed by a conditional branch.
	 We borrow the following conditional branch's counter and when about to
	 inline the comparison we decrement the counter (without writing it back)
	 and if it trips simply abort the inlining, falling back to the normal send which
	 will then continue to the conditional branch which will trip and enter the abort."
	| nextPC postBranchPC targetBytecodePC branchDescriptor counterReg fixup jumpEqual jumpNotEqual
	  counterAddress countTripped unforwardArg unforwardRcvr argReg rcvrReg regMask
	  rcvrConstant argConstant |
	<var: #fixup type: #'BytecodeFixup *'>
	<var: #countTripped type: #'AbstractInstruction *'>
	<var: #label type: #'AbstractInstruction *'>
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>
	<var: #jumpEqual type: #'AbstractInstruction *'>
	<var: #jumpNotEqual type: #'AbstractInstruction *'>

	((coInterpreter isOptimizedMethod: methodObj) or: [needsFrame not]) ifTrue:
		[^super genForwardersInlinedIdenticalOrNotIf: orNot].

	regMask := 0.
	
	unforwardRcvr := (self ssValue: 1) mayBeAForwarder.
	unforwardArg := self ssTop mayBeAForwarder.
	(unforwardRcvr not and: [unforwardArg not])
		ifTrue: [unforwardRcvr := true.
				"TODO: use genVanilla with profiling counters (not implemented).
				^self genVanillaInlinedIdenticalOrNotIf: orNot"].
	self assert: (unforwardArg or: [unforwardRcvr]).
	"We use reg for non annotable constants to avoid duplicating objRef."
	rcvrConstant := objectRepresentation isUnannotatableConstant: (self ssValue: 1).
	argConstant := objectRepresentation isUnannotatableConstant: self ssTop.
	
	self extractMaybeBranchDescriptorInto: [ :descr :next :postBranch :target | 
		branchDescriptor := descr. nextPC := next. postBranchPC := postBranch. targetBytecodePC := target ].
	
	"If an operand is an annotable constant, it may be forwarded, so we need to store it into a 
	register so the forwarder check can jump back to the comparison after unforwarding the constant.
	However, if one of the operand is an unnanotable constant, does not allocate a register for it 
	(machine code will use operations on constants)."
	self 
		allocateEqualsEqualsRegistersArgNeedsReg: argConstant not 
		rcvrNeedsReg: rcvrConstant not 
		into: [ :rcvr :arg | rcvrReg:= rcvr. argReg := arg ].

	"Only interested in inlining if followed by a conditional branch."
	(branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse]) ifFalse:
		[^ self 
			genIdenticalNoBranchArgIsConstant: argConstant
			rcvrIsConstant: rcvrConstant
			argReg: argReg 
			rcvrReg: rcvrReg 
			orNotIf: orNot].
	
	"If branching the stack must be flushed for the merge"
	self ssFlushTo: simStackPtr - 2.
	
	unforwardArg ifTrue: [ objectRepresentation genEnsureOopInRegNotForwarded: argReg scratchReg: TempReg ].
	unforwardRcvr ifTrue: [ objectRepresentation genEnsureOopInRegNotForwarded: rcvrReg scratchReg: TempReg ].
	
	regMask := argReg = NoReg
					ifTrue: [self registerMaskFor: rcvrReg]
					ifFalse:
						[rcvrReg = NoReg
							ifTrue: [self registerMaskFor: argReg]
							ifFalse: [self registerMaskFor: rcvrReg and: argReg]].
	counterReg := self allocateRegNotConflictingWith: regMask.
	self 
		genExecutionCountLogicInto: [ :cAddress :countTripBranch | 
			counterAddress := cAddress. 
			countTripped := countTripBranch ] 
		counterReg: counterReg.

	self genCmpArgIsConstant: argConstant rcvrIsConstant: rcvrConstant argReg: argReg rcvrReg: rcvrReg.
	self ssPop: 2.
	
	orNot == branchDescriptor isBranchTrue "orNot is true for ~~"
		ifFalse:
			[ fixup := (self ensureNonMergeFixupAt: postBranchPC) asUnsignedInteger.
			self JumpZero:  (self ensureNonMergeFixupAt: targetBytecodePC) asUnsignedInteger ]
		ifTrue:
			[ fixup := (self ensureNonMergeFixupAt: targetBytecodePC) asUnsignedInteger.
			self JumpZero: (self ensureNonMergeFixupAt: postBranchPC) asUnsignedInteger ].
	
	self genFallsThroughCountLogicCounterReg: counterReg counterAddress: counterAddress.
	self Jump: fixup.
	
	countTripped jmpTarget: self Label.
	
	"inlined version of #== ignoring the branchDescriptor if the counter trips to have normal state for the optimizer"
	self ssPop: -2. 
	self genCmpArgIsConstant: unforwardArg not rcvrIsConstant: unforwardRcvr not argReg: argReg rcvrReg: rcvrReg.
	self ssPop: 2. 
	
	"This code necessarily directly falls through the jumpIf: code which pops the top of the stack into TempReg. 
	We therefore directly assign the result to TempReg to save one move instruction"
	jumpEqual := orNot ifFalse: [self JumpZero: 0] ifTrue: [self JumpNonZero: 0].
	self genMoveFalseR: TempReg.
	jumpNotEqual := self Jump: 0.
	jumpEqual jmpTarget: (self genMoveTrueR: TempReg).
	jumpNotEqual jmpTarget: self Label.
	self ssPushRegister: TempReg.
	
	(self fixupAt: nextPC) notAFixup ifTrue: [ branchReachedOnlyForCounterTrip := true ].
	
	^ 0! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 1/3/2018 16:02'!
genJumpBinaryInlinePrimitive: primIndex
	| target testReg |
	self assert: self ssTop type = SSConstant.
	target := self eventualTargetOf: ((objectMemory integerValueOf: self ssTop constant) + 3 + bytecodePC).
	testReg := self allocateRegForStackEntryAt: 1.
	(self ssValue: 1) popToReg: testReg.
	self ssPop: 2.
	primIndex caseOf: {
		"7016	jumpIfWritable:
		 Not a forwarder, literal which is a Smi"
		[16]	->	[self cppIf: #IMMUTABILITY
						ifTrue: [(objectRepresentation genJumpMutable: testReg scratchReg: TempReg) jmpTarget: (self ensureFixupAt: target).]
						ifFalse: [self Jump: (self ensureFixupAt: target)].
					^0].
		"7017	jumpIfReadOnly:
		 Not a forwarder, literal which is a Smi"
		[17]	->	[self cppIf: #IMMUTABILITY
						ifTrue: [(objectRepresentation genJumpImmutable: testReg scratchReg: TempReg) jmpTarget: (self ensureFixupAt: target)]
						ifFalse: [0"Do nothing - fall through"].
					^0].
		"7018	jumpIfYoung:
		 Not a forwarder, literal which is a Smi"
		[18]	->	[self CmpCw: objectMemory storeCheckBoundary R: testReg.
	 				 self JumpBelow: (self ensureFixupAt: target).
					 ^0].
		"7019	jumpIfOld:
		 Not a forwarder, literal which is a Smi"
		[19]	->	[self CmpCw: objectMemory storeCheckBoundary R: testReg.
	 				 self JumpAboveOrEqual: (self ensureFixupAt: target).
					^0].
	}. 
	^ EncounteredUnknownBytecode
	! !

!SistaCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/21/2018 16:20'!
genJumpIf: boolean to: targetBytecodePC
	"The heart of performance counting in Sista.  Conditional branches are 6 times less
	 frequent than sends and can provide basic block frequencies (send counters can't).
	 Each conditional has a 32-bit counter split into an upper 16 bits counting executions
	 and a lower half counting untaken executions of the branch.  Executing the branch
	 decrements the upper half, tripping if the count goes negative.  Not taking the branch
	 decrements the lower half.  N.B. We *do not* eliminate dead branches (true ifTrue:/true ifFalse:)
	 so that scanning for send and branch data is simplified and that branch data is correct."
	<inline: false>
	| ok counterAddress countTripped retry nextPC nextDescriptor desc eventualTarget |
	<var: #ok type: #'AbstractInstruction *'>
	<var: #desc type: #'CogSimStackEntry *'>
	<var: #retry type: #'AbstractInstruction *'>
	<var: #countTripped type: #'AbstractInstruction *'>
	<var: #nextDescriptor type: #'BytecodeDescriptor *'>

	"In optimized code we don't generate counters to improve performance"
	(coInterpreter isOptimizedMethod: methodObj) ifTrue: [ ^ super genJumpIf: boolean to: targetBytecodePC ].
	
	"If the branch is reached only for the counter trip trampoline 
	(typically, var1 == var2 ifTrue: falls through to the branch only for the trampoline)
	we generate a specific path to drastically reduce the number of machine instructions"
	branchReachedOnlyForCounterTrip ifTrue: 
		[ branchReachedOnlyForCounterTrip := false.
		^ self genCounterTripOnlyJumpIf: boolean to: targetBytecodePC ].
	
	"We detect and: / or:, if found, we don't generate the counters to avoid pathological counter slow down"
	boolean = objectMemory falseObject ifTrue:
		[ nextPC := bytecodePC + (self generatorAt: byte0) numBytes.
		  nextDescriptor := self generatorForPC: nextPC.
		  nextDescriptor generator ==  #genPushConstantTrueBytecode ifTrue: [ ^ super genJumpIf: boolean to: targetBytecodePC ].
		  nextDescriptor := self generatorForPC: targetBytecodePC.
		  nextDescriptor generator == #genPushConstantFalseBytecode ifTrue: [ ^ super genJumpIf: boolean to: targetBytecodePC ].  ].

	extA := 0. "We ignore the noMustBeBoolean flag. It should not be present in methods with counters, and if it is we don't care."

	"We don't generate counters on branches on true/false, the basicblock usage can be inferred"
	desc := self ssTop.
	(self stackEntryIsBoolean: desc) ifTrue:
		[ ^ super genJumpIf: boolean to: targetBytecodePC ].

	eventualTarget := self eventualTargetOf: targetBytecodePC.

	self ssFlushTo: simStackPtr - 1.
	desc popToReg: TempReg.
	self ssPop: 1.

	"We need SendNumArgsReg because of the mustBeBooleanTrampoline"
	self ssAllocateRequiredReg: SendNumArgsReg.

	retry := self Label.
	self 
		genExecutionCountLogicInto: [ :cAddress :countTripBranch | 
			counterAddress := cAddress. 
			countTripped := countTripBranch ] 
		counterReg: SendNumArgsReg.
	counterIndex := counterIndex + 1.

	"Cunning trick by LPD.  If true and false are contiguous subtract the smaller.
	 Correct result is either 0 or the distance between them.  If result is not 0 or
	 their distance send mustBeBoolean."
	self assert: (objectMemory objectAfter: objectMemory falseObject) = objectMemory trueObject.
	self genSubConstant: boolean R: TempReg.
	self JumpZero: (self ensureFixupAt: eventualTarget).

	self genFallsThroughCountLogicCounterReg: SendNumArgsReg counterAddress: counterAddress.

	self CmpCq: (boolean = objectMemory falseObject
					ifTrue: [objectMemory trueObject - objectMemory falseObject]
					ifFalse: [objectMemory falseObject - objectMemory trueObject])
		R: TempReg.
	ok := self JumpZero: 0.
	self MoveCq: 0 R: SendNumArgsReg. "if counterReg is 0 this is a mustBeBoolean, not a counter trip."
	
	countTripped jmpTarget: (self genCallMustBeBooleanFor: boolean).
						
	"If we're in an image which hasn't got the Sista code loaded then the ceCounterTripped:
	 trampoline will return directly to machine code, returning the boolean.  So the code should
	 jump back to the retry point. The trampoline makes sure that TempReg has been reloaded."
	
	"Clément: For some reason if I write self annotateBytecode: (self Jump: retry) the annotation is not at the correct place."
	"Eliot: Annotations apply the the address following an instruction, and the annotation must be for the return address
	 of the call (since this is the address the run-time sees), so it must be on a label before the jump, not after the jump."
	self annotateBytecode: self Label.
	self Jump: retry.
	
	ok jmpTarget: self Label.
	^0! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 1/8/2018 14:41'!
genJumpTrinaryInlinePrimitive: primIndex
	"Note: those tests work with forwarders (wrong class index)"
	| testReg target behavior targetFixUp |
	<var: #targetFixUp type: #'BytecodeFixup *'>
	self assert: self ssTop type = SSConstant.
	self assert: (self ssValue: 1) type = SSConstant.
	testReg := self allocateRegForStackEntryAt: 2.
	(self ssValue: 2) popToReg: testReg.
	behavior := (self ssValue: 1) constant.
	target := self eventualTargetOf: ((objectMemory integerValueOf: self ssTop constant) + 3 + bytecodePC).
	self ssPop: 3.
	targetFixUp := self cCoerceSimple: (self ensureFixupAt: target) to: #'AbstractInstruction *'.
	primIndex caseOf: {
	[0] -> 	["8000	jumpIfInstanceOf:distance:
			 Anything, literal which is a Behavior, literal which is a Smi"
			objectRepresentation branchIf: testReg instanceOfBehavior: behavior target: targetFixUp.
			^0].
	[1] -> 	["8001	jumpIfNotInstanceOf:distance:
			 Anything, literal which is a Behavior, literal which is a Smi"
			 objectRepresentation branchIf: testReg notInstanceOfBehavior: behavior target: targetFixUp.
			^0].
	[2] -> 	["8002	jumpIfInstanceOfOneOf:distance:
			 Anything, Array of behaviors, literal which is a Smi"
			 objectRepresentation branchIf: testReg instanceOfBehaviors: behavior target: targetFixUp.
			^0].
	[3] -> 	["8003	jumpIfNotInstanceOfOneOf:distance:
			  Anything, Array of behaviors, literal which is a Smi"
			 objectRepresentation branchIf: testReg notInstanceOfBehaviors: behavior target: targetFixUp.
			^0].
	}.
	^ EncounteredUnknownBytecode
	
	! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 1/3/2018 16:01'!
genJumpUnaryInlinePrimitive: primIndex
	"6000	backjumpNoInterrupt
	 literal which is a Smi"
	primIndex = 0 ifTrue: 
		[|targetBytecodePC|
		 self assert: self ssTop type = SSConstant.
		 targetBytecodePC := (objectMemory integerValueOf: self ssTop constant) + 3 + bytecodePC.
		 self ssPop: 1.
		 self ssFlushTo: simStackPtr.
		 deadCode := true. "can't fall through"
		 self Jump: (self fixupAt: targetBytecodePC).
		 ^0].
	^EncounteredUnknownBytecode! !

!SistaCogit methodsFor: 'mapped inline primitive generators' stamp: 'cb 1/8/2018 14:52'!
genMappedInlinePrimitive: primIndex
	"SistaV1:	236		11101100	iiiiiiii		callMappedInlinedPrimitive"
	"Number of arguments:
	 0-49 nullary
	 50-99 unary
	 100-149  binary
	 150-199 trinary
	 200-255 variable"
	"Specification:
	50	EnsureEnoughWords
	literal which is a Smi => ret value is receiver
	150	immCheckPointerAt:put:
	pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	151	immCheckStoreCheckPointerAt:put:
	pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	152	immCheckMaybeContextPointerAt:put:
	pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	153	immCheckMaybeContextStoreCheckPointerAt:put:
	pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	154	immCheckByteAt:put:
	byte object, Smi, 8 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	155	immCheckShortAt:put:
	short object, Smi, 16 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	156	immCheckWordAt:put:
	word object, Smi, 32 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	157	immCheckDoubleWordAt:put:
	double word object, Smi, 64 bits unsigned Smi or LargePositiveInteger => arg2 (1-based, optimised if arg1 is a constant)
	250	directCall
	method to call on top of stack =>  (variable number of parameters)"
	primIndex caseOf: 
	{ 
		[50] ->	 [^self genEnsureEnoughSlots].
		[150] -> [^self genPointerAtPutMaybeContext: false storeCheck: false immutabilityCheck: true].
		[151] -> [^self genPointerAtPutMaybeContext: false storeCheck: true immutabilityCheck: true].
		[152] -> [^self genPointerAtPutMaybeContext: true storeCheck: false immutabilityCheck: true].
		[153] -> [^self genPointerAtPutMaybeContext: true storeCheck: true immutabilityCheck: true].
		[154] -> [self cppIf: #IMMUTABILITY
					ifTrue: [^self genByteAtPutImmutabilityCheck]
					ifFalse: [self genByteAtPut.
							self annotateBytecode: self Label.
							 ^0]].
		[155] -> [^EncounteredUnknownBytecode "not implemented, missing short instruction in Cog RTL"].
		[156] -> [^EncounteredUnknownBytecode "not implemented, need to deal with LargePositiveInteger in 32 bits"].
		[157] -> [^EncounteredUnknownBytecode "not implemented, need to deal with LargePositiveInteger in 64 bits"].
		[250] -> [^self genDirectCall].
	} otherwise: [^EncounteredUnknownBytecode].
	! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 4/10/2017 18:20'!
genMoveAndAdjustSSEntry: ssEntry into: reg
	<var: #ssEntry type: #'SimStackEntry *'>
	<inline: true>
	| adjust |
	adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
	ssEntry type = SSConstant 
		ifTrue: [self MoveCq: (objectMemory integerValueOf: ssEntry constant) + adjust R: reg.
			      ssEntry spilled ifTrue: [self AddCq: objectMemory wordSize R: SPReg]]
		ifFalse: 
			[ssEntry popToReg: reg.
			 objectRepresentation genConvertSmallIntegerToIntegerInReg: reg.
			 adjust ~= 0 ifTrue: [ self AddCq: adjust R: reg. ]].	! !

!SistaCogit methodsFor: 'initialization' stamp: 'eem 2/1/2020 18:51'!
genMustBeBooleanTrampolineFor: boolean called: trampolineName
	"This can be entered in one of two states, depending on SendNumArgsReg. See
	 e.g. genJumpIf:to:.  If SendNumArgsReg is non-zero then this has been entered via
	 the initial test of the counter in the jump executed count (i.e. the counter has
	 tripped).  In this case TempReg contains the boolean to be tested and should not
	 be offset, and ceCounterTripped should be invoked with the unoffset TempReg.
	 If SendNumArgsReg is zero then this has been entered for must-be-boolean
	 processing. TempReg has been offset by boolean and must be corrected and
	 ceSendMustBeBoolean: invoked with the corrected value."
	<var: #trampolineName type: #'char *'>
	| jumpMBB |
	<var: #jumpMBB type: #'AbstractInstruction *'>
	<inline: false>
	self zeroOpcodeIndex.
	self CmpCq: 0 R: SendNumArgsReg.
	jumpMBB := self JumpZero: 0.
	"Open-code self compileTrampolineFor: #ceCounterTripped: numArgs: 1 arg: TempReg ...
	 so we can restore ResultReceiverReg."
	self genSmalltalkToCStackSwitch: true.
	self
		compileCallFor: #ceCounterTripped:
		numArgs: 1
		arg: TempReg
		arg: nil
		arg: nil
		arg: nil
		resultReg: TempReg "(*)"
		regsToSave: self emptyRegisterMask.
	"(*) For the case where the ceCounterTripped: call returns (e.g. because there's no callback selector
	 installed), the call to the ceSendMustBeBooleanAddTrue/FalseTrampoline is followed by a jump
	 back to the start of the counter/condition test sequence.  For this case copy the C result to
	 TempReg (the register that is tested), to reload it with the boolean to be tested."
	backEnd genLoadStackPointers.
	backEnd hasLinkRegister ifTrue:
		[self PopR: LinkReg].
	"To keep ResultReceiverReg live if optStatus thought it was, simply reload it
	 from the frame pointer.  This avoids having to reload it in the common case
	 (counter does not trip) if it was live.  Note we can't use putSelfInReceiverResultReg
	 when generating trampolines because simSelf has not yet been initialized."
	self MoveMw: FoxMFReceiver r: FPReg R: ReceiverResultReg.
	self RetN: 0.
	"If the objectRepresentation does want true & false to be mobile then we need to record these addresses."
	self assert: (objectRepresentation shouldAnnotateObjectReference: boolean) not.
	jumpMBB jmpTarget: (self AddCq: boolean R: TempReg).
	^self genTrampolineFor: #ceSendMustBeBoolean:
		called: trampolineName
		numArgs: 1
		arg: TempReg
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: true! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 11/6/2017 16:53'!
genNullaryInlinePrimitive: prim
	<option: #SistaVM>
	^EncounteredUnknownBytecode! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 3/1/2018 17:55'!
genPointerAtPutConstantMaybeContext: maybeContext storeCheck: needsStoreCheck immutabilityCheck: needsImmCheck 
	| rcvrReg valReg indexCst |
	indexCst := (objectMemory integerValueOf: (self ssValue: 1) constant) - 1.
	"we want to have on top of stack the value to write descriptor."
	(maybeContext or: [needsImmCheck]) 
		ifTrue: [ valReg := ClassReg.
				 self voidReceiverResultRegContainsSelf.
				 rcvrReg := ReceiverResultReg ]
		ifFalse: [ valReg := self allocateRegForStackEntryAt: 0 notConflictingWith: 0.
				  needsStoreCheck 
					ifFalse: [rcvrReg := self allocateRegForStackEntryAt: 2 notConflictingWith: (self registerMaskFor: valReg)]
					ifTrue: [self voidReceiverResultRegContainsSelf.
							rcvrReg := ReceiverResultReg]].
	"If rcvr is valReg, we flush it. Could generate a move instead but not that common"
	self ssAllocateRequiredReg: valReg upThrough: simStackPtr - 2.
	self ssAllocateRequiredReg: rcvrReg upThrough: simStackPtr - 3.
	self ssTop popToReg: valReg.
	self ssPop: 1.
	self ssTop spilled ifTrue: [self AddCq: objectRepresentation wordSize R: SPReg].
	self ssPop: 1.
	self ssTop popToReg: rcvrReg.
	self ssPop: 1.
	self ssPushRegister: valReg.
			
	maybeContext
		ifFalse: 
			[^self 
				genGenericStorePop: false 
				slotIndex: indexCst
				destReg: rcvrReg 
				needsStoreCheck: needsStoreCheck 
				needsRestoreRcvr: false 
				needsImmutabilityCheck: needsImmCheck]
		ifTrue:
			[^self
				genGenericStorePop: false 
				MaybeContextSlotIndex: indexCst
				needsStoreCheck: needsStoreCheck 
				needsRestoreRcvr: false 
				needsImmutabilityCheck: needsImmCheck]! !

!SistaCogit methodsFor: 'mapped inline primitive generators' stamp: 'cb 12/11/2017 15:39'!
genPointerAtPutImmCheckAndStoreCheck
	| ra1 ra2 rr adjust scratchReg immutableJump jmpImmediate jmpDestYoung jmpSourceOld jumpRemembered indexIsCst |
	<var: #jmpSourceOld type: #'AbstractInstruction *'>
	<var: #jmpDestYoung type: #'AbstractInstruction *'>
	<var: #jmpImmediate type: #'AbstractInstruction *'>
	<var: #immutableJump type: #'AbstractInstruction *'>
	<var: #jumpRemembered type: #'AbstractInstruction *'>
	"Assumes rr is not a context and no store check is needed"
	indexIsCst := (self ssValue: 1) type = SSConstant.
	self ssFlushTo: simStackPtr - 3.
	rr := ReceiverResultReg.
	ra1 := TempReg.
	ra2 := ClassReg.
	scratchReg := Arg0Reg.
	adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
	(self ssValue: 2) popToReg: rr.
	indexIsCst
		ifFalse:
			[(self ssValue: 1) popToReg: ra1.
			objectRepresentation genConvertSmallIntegerToIntegerInReg: ra1.
			adjust ~= 0 ifTrue: [self AddCq: adjust R: ra1]]
		ifTrue: [self MoveCq: (objectMemory integerValueOf: (self ssValue: 1) constant) + adjust R: ra1].
	self ssTop popToReg: ra2.
	self ssPop: 3.
	self ssPushRegister: ra2.
	self voidReceiverResultRegContainsSelf.

	immutableJump := self genJumpImmutable: rr scratchReg: scratchReg.
	self MoveR: ra2 Xwr: ra1 R: rr.
	
	"store check"
	jmpImmediate := self genJumpImmediate: ra2.
	"Get the old/new boundary in scratchReg"
	self MoveCw: objectMemory storeCheckBoundary R: scratchReg.
	"Is target young?  If so we're done"
	self CmpR: scratchReg R: rr. "N.B. FLAGS := destReg - scratchReg"
	jmpDestYoung := self JumpBelow: 0.
	"Is value stored old?  If so we're done."
	self CmpR: scratchReg R: ra2. "N.B. FLAGS := valueReg - scratchReg"
	jmpSourceOld := self JumpAboveOrEqual: 0.
	"value is young and target is old.
	 Need to remember this only if the remembered bit is not already set."
	jumpRemembered := objectRepresentation genIfRequiredCheckRememberedBitOf: rr scratch: scratchReg.
	"Set the inst var index for the benefit of the immutability check. The trampoline will
	 repeat the check to choose between the immutbality violation and the store check."
	immutableJump jmpTarget: self Label.
	self PushR: ra2.
	self SubCq: 1 + adjust R: ra1. "index 0-relative for trampoline, ra1 unused afterwards"
	objectRepresentation genVarIndexCallStoreTrampoline.
	self PopR: ra2.
	jmpImmediate jmpTarget:
	(jmpDestYoung jmpTarget:
	(jmpSourceOld jmpTarget:
		self Label)).
	self setIfRequiredTargetOf: jumpRemembered toTargetOf: jmpImmediate.
	
	^0! !

!SistaCogit methodsFor: 'mapped inline primitive generators' stamp: 'cb 11/23/2017 14:57'!
genPointerAtPutImmCheckButNoStoreCheck
	| ra1 ra2 rr adjust mutableJump immutabilityFailure indexIsCst |
	<var: #mutableJump type: #'AbstractInstruction *'>
	<var: #immutabilityFailure type: #'AbstractInstruction *'>
	"Assumes rr is not a context and no store check is needed"
	indexIsCst := (self ssValue: 1) type = SSConstant.
	self ssFlushTo: simStackPtr - 3.
	rr := ReceiverResultReg.
	ra1 := TempReg.
	ra2 := ClassReg.
	(self ssValue: 2) popToReg: rr.
	adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
	indexIsCst
		ifFalse:
			[(self ssValue: 1) popToReg: ra1.
			objectRepresentation genConvertSmallIntegerToIntegerInReg: ra1.
			adjust ~= 0 ifTrue: [self AddCq: adjust R: ra1]]
		ifTrue: [self MoveCq: (objectMemory integerValueOf: (self ssValue: 1) constant) + adjust R: ra1].
	self ssTop popToReg: ra2.
	self ssPop: 3.
	self ssPushRegister: ra2.
	self voidReceiverResultRegContainsSelf.

	mutableJump := self genJumpMutable: rr scratchReg: Arg0Reg.
	"simStack is flushed, but result is not"
	self PushR: ra2.
	self SubCq: 1 + adjust R: ra1. "index 0-relative for trampoline, ra1 unused afterwards"
	objectRepresentation genVarIndexCallStoreTrampoline.
	self PopR: ra2.
	immutabilityFailure := self Jump: 0.
	
	mutableJump jmpTarget: self Label.
	self MoveR: ra2 Xwr: ra1 R: rr.
	immutabilityFailure jmpTarget: self Label.
	^0! !

!SistaCogit methodsFor: 'mapped inline primitive generators' stamp: 'cb 1/8/2018 14:52'!
genPointerAtPutMaybeContext: maybeContext storeCheck: needsStoreCheck immutabilityCheck: needsImmCheck
	"150	immCheckPointerAt:put:
	pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	151	immCheckStoreCheckPointerAt:put:
	pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	152	immCheckMaybeContextPointerAt:put:
	pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	153	immCheckMaybeContextStoreCheckPointerAt:put:
	pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
	| index |
	index := self ssValue: 1.
	"Instruction always annoted - uncommon, IMMUTABILITY disabled. We need a Nop since previous instr can be annotated."
	self cppIf: #IMMUTABILITY ifTrue: [] ifFalse: [ needsImmCheck ifTrue: [ self annotateBytecode: self Nop ] ].
	"Optimised case if arg1 is constant"
	index type = SSConstant 
		ifTrue:  [^self genPointerAtPutConstantMaybeContext: maybeContext storeCheck: needsStoreCheck immutabilityCheck: needsImmCheck].
	maybeContext ifTrue: [^EncounteredUnknownBytecode "not implemented, not used right now, for tempAt: optimisation on non constant"].
	self cppIf: #IMMUTABILITY ifTrue:
		[needsImmCheck
			ifTrue: [needsStoreCheck
				ifTrue: [^self genPointerAtPutImmCheckAndStoreCheck]
				ifFalse: [^self genPointerAtPutImmCheckButNoStoreCheck] ] ].
	^self genPointerAtPutStoreCheck: needsStoreCheck! !

!SistaCogit methodsFor: 'mapped inline primitive generators' stamp: 'cb 11/6/2017 10:02'!
genPointerAtPutStoreCheck: needsStoreCheck 
	| ra1 ra2 rr adjust |
	"Assumes rr is not a context and no immutability check is needed"
	"The store check requires rr to be ReceiverResultReg"
	self 
		allocateRegForStackTopThreeEntriesInto: [:rTop :rNext :rThird | ra2 := rTop. ra1 := rNext. rr := rThird ] 
		thirdIsReceiver: needsStoreCheck.
	self assert: (rr ~= ra1 and: [rr ~= ra2 and: [ra1 ~= ra2]]).
	self ssTop popToReg: ra2.
	self ssPop: 1.
	self ssTop popToReg: ra1.
	self ssPop: 1.
	self ssTop popToReg: rr.
	self ssPop: 1.
	objectRepresentation genConvertSmallIntegerToIntegerInReg: ra1.
	adjust := (objectMemory baseHeaderSize >> objectMemory shiftForWord) - 1. "shift by baseHeaderSize and then move from 1 relative to zero relative"
	adjust ~= 0 ifTrue: [ self AddCq: adjust R: ra1. ]. 
	self MoveR: ra2 Xwr: ra1 R: rr.
	needsStoreCheck ifTrue: 
		[ self assert: needsFrame. 
		objectRepresentation genStoreCheckReceiverReg: rr valueReg: ra2 scratchReg: TempReg inFrame: true].
	self ssPushRegister: ra2.
	^0! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 11/10/2017 11:06'!
genQuaternaryInlinePrimitive: prim
	^ EncounteredUnknownBytecode! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 11/10/2017 11:06'!
genQuinaryInlinePrimitive: prim
	^ EncounteredUnknownBytecode! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 3/6/2018 09:55'!
genSistaInlinePrimitive: prim
	"SistaV1:	248		11111000 	iiiiiiii		mjjjjjjj		Call Primitive #iiiiiiii + (jjjjjjj * 256) m=1 means inlined primitive, no hard return after execution. This is the dispatch for unmapped sista inlined primitives."
	
	prim < 1000 ifTrue:
		[^self genNullaryInlinePrimitive: prim].

	prim < 2000 ifTrue:
		[^self genUnaryInlinePrimitive: prim - 1000].
	
	prim < 3000 ifTrue:
		[^self genBinaryInlinePrimitive: prim - 2000].

	prim < 4000 ifTrue:
		[^self genTrinaryInlinePrimitive: prim - 3000].
	
	prim < 5000 ifTrue: 
		[^self genQuaternaryInlinePrimitive: prim - 4000].
	
	prim < 6000 ifTrue: 
		[^self genQuinaryInlinePrimitive: prim - 5000].
		
	 prim < 7000 ifTrue:
		[^self genJumpUnaryInlinePrimitive: prim - 6000].
	
 	prim < 8000 ifTrue:
		[^self genJumpBinaryInlinePrimitive: prim - 7000].
	
	^ self genJumpTrinaryInlinePrimitive: prim - 8000.! !

!SistaCogit methodsFor: 'bytecode generators' stamp: 'eem 1/26/2018 11:03'!
genSpecialSelectorComparison
	"Override to count inlined branches if followed by a conditional branch.
	 We borrow the following conditional branch's counter and when about to
	 inline the comparison we decrement the counter (without writing it back)
	 and if it trips simply abort the inlining, falling back to the normal send which
	 will then continue to the conditional branch which will trip and enter the abort."
	| nextPC postBranchPC targetPC primDescriptor branchDescriptor
	  rcvrIsInt rcvrIsConst argIsIntConst argInt jumpNotSmallInts inlineCAB
	  counterAddress countTripped counterReg index |
	<var: #countTripped type: #'AbstractInstruction *'>
	<var: #primDescriptor type: #'BytecodeDescriptor *'>
	<var: #jumpNotSmallInts type: #'AbstractInstruction *'>
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>

	(coInterpreter isOptimizedMethod: methodObj) ifTrue: [ ^ self genSpecialSelectorComparisonWithoutCounters ].

	self ssFlushTo: simStackPtr - 2.
	primDescriptor := self generatorAt: byte0.
	argIsIntConst := self ssTop type = SSConstant
				 and: [objectMemory isIntegerObject: (argInt := self ssTop constant)].
	rcvrIsInt := ((rcvrIsConst := (self ssValue: 1) type = SSConstant)
				  and: [objectMemory isIntegerObject:(self ssValue: 1) constant])
				or: [self mclassIsSmallInteger and: [(self ssValue: 1) isSameEntryAs: self simSelf]].

	"short-cut the jump if operands are SmallInteger constants."
	(argIsIntConst and: [rcvrIsInt and: [rcvrIsConst]]) ifTrue:
		[^ self genStaticallyResolvedSpecialSelectorComparison].

	self extractMaybeBranchDescriptorInto: [ :descr :next :postBranch :target | 
		branchDescriptor := descr. nextPC := next. postBranchPC := postBranch. targetPC := target ].
	
	"Only interested in inlining if followed by a conditional branch."
	inlineCAB := branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse].
	"Further, only interested in inlining = and ~= if there's a SmallInteger constant involved.
	 The relational operators successfully statically predict SmallIntegers; the equality operators do not."
	(inlineCAB and: [primDescriptor opcode = JumpZero or: [primDescriptor opcode = JumpNonZero]]) ifTrue:
		[inlineCAB := argIsIntConst or: [rcvrIsInt]].
	inlineCAB ifFalse:
		[^self genSpecialSelectorSend].

	argIsIntConst
		ifTrue:
			[(self ssValue: 1) popToReg: ReceiverResultReg.
			 self ssPop: 2]
		ifFalse:
			[self marshallSendArguments: 1].
	jumpNotSmallInts := (rcvrIsInt and: [argIsIntConst]) ifFalse:
							[argIsIntConst
								ifTrue: [objectRepresentation genJumpNotSmallInteger: ReceiverResultReg]
								ifFalse:
									[rcvrIsInt
										ifTrue: [objectRepresentation genJumpNotSmallInteger: Arg0Reg]
										ifFalse: [objectRepresentation genJumpNotSmallIntegersIn: ReceiverResultReg and: Arg0Reg scratch: TempReg]]].

	counterReg := self allocateRegNotConflictingWith: (self registerMaskFor: ReceiverResultReg and: Arg0Reg).
	self 
		genExecutionCountLogicInto: [ :cAddress :countTripBranch | 
			counterAddress := cAddress. 
			countTripped := countTripBranch ] 
		counterReg: counterReg.

	argIsIntConst
		ifTrue: [self CmpCq: argInt R: ReceiverResultReg]
		ifFalse: [self CmpR: Arg0Reg R: ReceiverResultReg].
	"Cmp is weird/backwards so invert the comparison.  Further since there is a following conditional
	 jump bytecode define non-merge fixups and leave the cond bytecode to set the mergeness."
	self genConditionalBranch: (branchDescriptor isBranchTrue
				ifTrue: [primDescriptor opcode]
				ifFalse: [self inverseBranchFor: primDescriptor opcode])
		operand: (self ensureNonMergeFixupAt: targetPC) asUnsignedInteger.
		
	self genFallsThroughCountLogicCounterReg: counterReg counterAddress: counterAddress.
	
	self Jump: (self ensureNonMergeFixupAt: postBranchPC).
	countTripped jmpTarget: self Label.
	jumpNotSmallInts
		ifNil: [(self fixupAt: nextPC) notAFixup ifTrue:
				[branchReachedOnlyForCounterTrip := true]]
		ifNotNil: [jumpNotSmallInts jmpTarget: countTripped getJmpTarget].
	
	argIsIntConst ifTrue:
		[self MoveCq: argInt R: Arg0Reg].
	index := byte0 - self firstSpecialSelectorBytecodeOffset.
	^self genMarshalledSend: index negated - 1 numArgs: 1 sendTable: ordinarySendTrampolines! !

!SistaCogit methodsFor: 'bytecode generators' stamp: 'cb 4/29/2015 14:39'!
genSpecialSelectorComparisonWithoutCounters
	"This method is there because if I put directly the super send in genSpecialSelectorComparison Slang does not correctly translte the code to C, it does not correctly type one of the branchDescriptor to BytecodeDescriptor"
	^ super genSpecialSelectorComparison! !

!SistaCogit methodsFor: 'inline primitive generators' stamp: 'cb 11/10/2017 15:17'!
genTrinaryInlinePrimitive: prim
	"Bulk comment, each sub method has its own comment
	3000	pointerAt:put:
	Mutable pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	3001	storeCheckPointerAt:put:
	Mutable pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	3002	maybeContextPointerAt:put:
	Mutable pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	3003	maybeContextStoreCheckPointerAt:put:
	Mutable pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	3004	byteAt:put:
	Mutable byte object, Smi, 8 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	3005	shortAt:put:
	Mutable short object, Smi, 16 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	3006	wordAt:put:
	Mutable word object, Smi, 32 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	3007	doubleWordAt:put:
	Mutable double word object, Smi, 64 bits unsigned Smi or LargePositiveInteger => arg2 (1-based, optimised if arg1 is a constant)
	3021 is deprecated."
	prim <= 7 ifTrue: [^self genAtPutInlinePrimitive: prim].
	prim = 21 ifTrue: [^self genByteEqualsInlinePrimitive: prim].
	^ EncounteredUnknownBytecode! !

!SistaCogit methodsFor: 'inline primitive unary generators' stamp: 'cb 11/23/2017 11:41'!
genUnaryClassPrimitive 
	"1000	rawClass
	not a forwarder => Behavior (Same as class special send, but receiver is not a forwarder)"
	
	"Important performance note:
	In Scorch, typically a value is known as not being a forwarder if there is a trap.
	If the trap is due to a monomorphic send, the #class send leads to:
		trapIf: X notInstanceOf: C
		X rawClass
	therefore X rawClass is simplified in Scorch to the cst:C
	The rawClass is therefore used only for PICs. 
		trapIf: X notInstanceOf: C, C', C''
		X rawClass
	This unsafe operation is important to avoid register flush, but the performance 
	difference in differencing rawClass for immediate and rawClass for non immediate 
	classes is not that relevant"
	
	| topReg jumpIsImm destReg |
	<var: #jumpIsImm type: #'AbstractInstruction *'>
	topReg := self allocateRegForStackEntryAt: 0 notConflictingWith: 0.
	destReg := self allocateRegNotConflictingWith: (self registerMaskFor: topReg).
	self ssTop popToReg: topReg.
	
	"1. Read the class index"
	self MoveR: topReg R: TempReg.
	self AndCq: objectMemory tagMask R: TempReg.
	jumpIsImm := self JumpNonZero: 0.
	self flag: #endianness.
	"Get least significant half of header word in destReg"
	self MoveMw: 0 r: topReg R: TempReg.
	"mask off class index"
	self AndCq: objectMemory classIndexMask R: TempReg.
	
	"2. Read the class from class index"
	jumpIsImm jmpTarget: self Label.
	objectRepresentation genGetClassObjectOfClassIndex: TempReg into: destReg scratchReg: topReg.
	self ssPop: 1.
	^self ssPushRegister: destReg! !

!SistaCogit methodsFor: 'inline primitive unary generators' stamp: 'cb 11/23/2017 14:20'!
genUnaryConvertInlinePrimitive: primIndex
	| resultReg |
	self assert: (primIndex between: 30 and: 32).
	primIndex caseOf: {
		"1030	characterAsInteger
		 Character => 22 bits strictly positive Smi (Unicode)"
		[30] -> [resultReg := self allocateRegForStackEntryAt: 0. 
				self ssTop popToReg: resultReg.
				objectRepresentation genConvertCharacterToSmallIntegerInReg: resultReg].
		"1031	smallFloatAsInteger
		 SmallFloat => Smi"
		[31] -> [^EncounteredUnknownBytecode "to implement"].
		"1032	smiAsFloat
		 Smi => SmallFloat"
		[32] -> [resultReg := self allocateRegForStackEntryAt: 0. 
				self ssTop popToReg: resultReg.
				self assert: self processorHasDoublePrecisionFloatingPointSupport.
				self MoveR: resultReg R: TempReg.
				self genConvertSmallIntegerToIntegerInReg: TempReg.
				self ConvertR: TempReg Rd: DPFPReg0.
				self flag: #TODO. "Should never fail"
				self
					genAllocFloatValue: DPFPReg0
					into: resultReg
					scratchReg: TempReg
					scratchReg: NoReg. "scratch2 for V3 only"] 
	}.
	self ssPop: 1.
	^self ssPushRegister: resultReg! !

!SistaCogit methodsFor: 'inline primitive unary generators' stamp: 'cb 11/23/2017 14:54'!
genUnaryHashInlinePrimitive: primIndex
	| rcvrReg resultReg |
	self assert: (primIndex between: 20 and: 23).
	primIndex caseOf: {
		"1020	objectIdentityHash
		 non-immediate and non-behavior => 22 bits strictly positive Smi"
		[20] ->	[rcvrReg := self allocateRegForStackEntryAt: 0.
				 resultReg := self allocateRegNotConflictingWith: (self registerMaskFor: rcvrReg).
				 self ssTop popToReg: rcvrReg.
				 objectRepresentation genGetIdentityHash: rcvrReg resultReg: resultReg].
		"1021	smiIdentityHash
		 Smi => Smi"
		[21] -> [resultReg := self allocateRegForStackEntryAt: 0. 
				self ssTop popToReg: resultReg].		
		"1022	charIdentityHash
		 Character => 22 bits strictly positive Smi"
		[22] -> [resultReg := self allocateRegForStackEntryAt: 0. 
				self ssTop popToReg: resultReg.
				objectRepresentation genConvertCharacterToSmallIntegerInReg: resultReg].
		"1023	smallfloatIdentityHash
		 SmallFloat => Smi"
		[23] -> [resultReg := self allocateRegForStackEntryAt: 0. 
				objectRepresentation genConvertSmallFloatToSmallFloatHashAsIntegerInReg: resultReg scratch: TempReg].
		"1024	behaviorIdentityHash
		 Behavior => 22 bits strictly positive Smi"
		[24] -> [^EncounteredUnknownBytecode "not implemented"].	
		}.
		self ssPop: 1.
		^self ssPushRegister: resultReg! !

!SistaCogit methodsFor: 'inline primitive unary generators' stamp: 'cb 12/14/2017 14:56'!
genUnaryInlinePrimitive: primIndex
	"Bulk comments: each sub-method has its own comment with the specific case.
	1000	rawClass
	not a forwarder => Behavior (Same as class special send, but receiver is not a forwarder)
	1001	numSlots
	pointer object => Smi between 0 and SmallInteger maxVal // 4 - 1 (Answers total size in pointer-sized slots)
	1002	numBytes
	byte object => Smi between 0 and SmallInteger maxVal - 9 (Includes compiled code)
	1003	numShorts
	short object => Smi between 0 and SmallInteger maxVal - 9
	1004	numWords
	word object => Smi between 0 and SmallInteger maxVal - 9
	1005	numDoubleWords
	double word object => Smi between 0 and SmallInteger maxVal - 9
	1011	RawNew
	literal which is a fixed-sized behavior => instance of the receiver with fields nilled out
	1012	RawNewNoInit
	literal which is a fixed-sized behavior => instance of the receiver (Fields of returned value contain undefined data)
	1020	objectIdentityHash
	non-immediate and non-behavior => 22 bits strictly positive Smi
	1021	smiIdentityHash
	Smi => Smi
	1022	charIdentityHash
	Character => 22 bits strictly positive Smi
	1023	smallfloatIdentityHash
	SmallFloat => Smi
	1024	behaviorIdentityHash
	Behavior => 22 bits strictly positive Smi
	1030	characterAsInteger
	Character => 22 bits strictly positive Smi (Unicode)
	1031	smallFloatAsInteger
	SmallFloat => Smi
	1032	smiAsFloat
	Smi => SmallFloat
	1040	unforward
	Anything => Not a forwarder
	1041	possibleRoot
	non-immediate, not a forwarder => receiver is returned (should be effect-only) (If old, becomes gray and remembered to allow many unchecked stores in a row afterwards)"
	primIndex = 0 ifTrue: [^self genUnaryClassPrimitive].
	primIndex <= 6 ifTrue: [^self genUnarySizeInlinePrimitive: primIndex].
	primIndex < 11 ifTrue: [^EncounteredUnknownBytecode].
	primIndex <= 12 ifTrue: [^self genUnaryNewInlinePrimitive: primIndex].
	primIndex < 20 ifTrue: [^EncounteredUnknownBytecode].
	primIndex <= 24 ifTrue: [^self genUnaryHashInlinePrimitive: primIndex].
	primIndex < 30 ifTrue: [^EncounteredUnknownBytecode].
	primIndex <= 32 ifTrue: [^self genUnaryConvertInlinePrimitive: primIndex].
	primIndex = 39 ifTrue: [^self genUnaryUnforwardNonImmediateInlinePrimitive].
	primIndex = 40 ifTrue: [^self genUnaryUnforwardInlinePrimitive].
	primIndex = 41 ifTrue: [^self genUnaryPossibleRootInlinePrimitive].
	^EncounteredUnknownBytecode! !

!SistaCogit methodsFor: 'inline primitive unary generators' stamp: 'cb 3/1/2018 20:58'!
genUnaryNewInlinePrimitive: primIndex 
	"1011	RawNew
	 literal which is a fixed-sized behavior => instance of the receiver with fields nilled out
	1012	RawNewNoInit
	literal which is a fixed-sized behavior => instance of the receiver (Fields of returned value contain undefined data)"
	| resultReg classObj |
	self assert: (self ssTop type = SSConstant).
	classObj := self ssTop constant.
	self assert: (objectMemory isNonImmediate: classObj).
	self assert: (coInterpreter objCouldBeClassObj: classObj).
	self assert: (objectMemory isFixedSizePointerFormat: (objectMemory instSpecOfClassFormat: (objectMemory formatOfClass: classObj))).
	objectMemory classTagForClass: classObj. "Ensure Behavior hash"
	resultReg := self allocateRegNotConflictingWith: 0.
	objectRepresentation genGetInstanceOfPointerClass: classObj into: resultReg initializingIf: primIndex = 11 numVariableSlots: 0.
	self ssPop: 1.
	^self ssPushRegister: resultReg! !

!SistaCogit methodsFor: 'inline primitive unary generators' stamp: 'cb 12/21/2017 11:36'!
genUnaryPossibleRootInlinePrimitive
	"1041	possibleRoot
	 non-immediate, not a forwarder => receiver is returned (should be effect-only) (If old, becomes gray and remembered to allow many unchecked stores in a row afterwards)"
	| topReg jmpDestYoung jmpAlreadyRemembered|
	<var: #jmpDestYoung type: #'AbstractInstruction *'>
	<var: #jmpAlreadyRemembered type: #'AbstractInstruction *'>
	topReg := self allocateRegForStackEntryAt: 0 notConflictingWith: 0.
	self ssTop popToReg: topReg.
	self MoveCw: objectMemory storeCheckBoundary R: TempReg.
	 self CmpR: TempReg R: topReg.
	 jmpDestYoung := self JumpBelow: 0.
	 objectRepresentation checkRememberedInTrampoline ifFalse: 
		[jmpAlreadyRemembered := objectRepresentation genCheckRememberedBitOf: topReg scratch: TempReg].
	 objectRepresentation callStoreCheckTrampoline.
	 jmpDestYoung jmpTarget: self Label.
	 objectRepresentation checkRememberedInTrampoline ifFalse: 
		[jmpAlreadyRemembered jmpTarget: self Label].
	self ssPop: 1.
	^self ssPushRegister: topReg! !

!SistaCogit methodsFor: 'inline primitive unary generators' stamp: 'cb 11/23/2017 14:20'!
genUnarySizeInlinePrimitive: primIndex 
	| rcvrReg resultReg |
	self assert: (primIndex between: 1 and: 6).
	rcvrReg := self allocateRegForStackEntryAt: 0 notConflictingWith: 0.
	resultReg := self allocateRegNotConflictingWith: (self registerMaskFor: rcvrReg).
	self ssTop popToReg: rcvrReg.
	self ssPop: 1.
	self ssPushRegister: resultReg.
	primIndex caseOf: {
		"1001	numSlots
		 pointer object => Smi between 0 and SmallInteger maxVal // 4 - 1 (Answers total size in pointer-sized slots)"
		[1]	->	[objectRepresentation
					genGetNumSlotsOf: rcvrReg into: resultReg;
					genConvertIntegerToSmallIntegerInReg: resultReg].
		"1002	numBytes
		 byte object => Smi between 0 and SmallInteger maxVal - 9 (Includes compiled code)"
		[2]	->	[objectRepresentation
					genGetNumBytesOf: rcvrReg into: resultReg;
					genConvertIntegerToSmallIntegerInReg: resultReg].
		"1003	numShorts
		 short object => Smi between 0 and SmallInteger maxVal - 9"
		[3]	->	[^EncounteredUnknownBytecode "not implemented"].
		"1004	numWords
		 word object => Smi between 0 and SmallInteger maxVal - 9"
		[5]	->	[^EncounteredUnknownBytecode "not implemented"].
		"1005	numDoubleWords 
		 double word object => Smi between 0 and SmallInteger maxVal - 9"
		[6]	->	[^EncounteredUnknownBytecode "not implemented"].
	}..
	^ 0! !

!SistaCogit methodsFor: 'inline primitive unary generators' stamp: 'cb 11/23/2017 11:43'!
genUnaryUnforwardInlinePrimitive
	"1040	unforward
	 Anything => Not a forwarder"
	| topReg |
	topReg := self allocateRegForStackEntryAt: 0 notConflictingWith: 0.
	self ssTop popToReg: topReg.
	objectRepresentation genEnsureOopInRegNotForwarded: topReg scratchReg: TempReg.
	self ssPop: 1.
	^self ssPushRegister: topReg! !

!SistaCogit methodsFor: 'inline primitive unary generators' stamp: 'eem 9/9/2019 10:15'!
genUnaryUnforwardNonImmediateInlinePrimitive
	"1039	unforwardNonImmediate
	 non immediate => Not a forwarder"
	"unforwardNonImmediate was used exclusively for literal variables,
	 which are now never forwarders because we scan the zone in postBecome,
	 so this is effectively a noop."
	| topReg |
	topReg := self allocateRegForStackEntryAt: 0 notConflictingWith: 0.
	self ssTop popToReg: topReg.
	"objectRepresentation genEnsureObjInRegNotForwarded: topReg scratchReg: TempReg."
	self ssPop: 1.
	^self ssPushRegister: topReg! !

!SistaCogit methodsFor: 'bytecode generators' stamp: 'cb 1/19/2017 12:11'!
genUnconditionalTrapBytecode
	"SistaV1: *	217		Trap"
	self ssFlushTo: simStackPtr.
	self CallRT: ceTrapTrampoline.
	self annotateBytecode: self Label.
	deadCode := true.
	^0! !

!SistaCogit methodsFor: 'bytecode generators' stamp: 'cb 4/28/2015 12:53'!
genUnoptimizedSpecialSelectorComparison
	"This method is there because if I put directly the super send in genSpecialSelectorComparison Slang does not correctly translte the code to C, it does not correctly type one of the branchDescriptor to BytecodeDescriptor"
	^ super genSpecialSelectorComparison! !

!SistaCogit methodsFor: 'initialization' stamp: 'cb 4/16/2015 16:16'!
generateSistaRuntime
	"Trap sends Sista trap message to context with top of stack, so we don't need any arguments..."
	ceTrapTrampoline := self genTrampolineFor: #ceSistaTrap called: 'ceSistaTrapTrampoline'! !

!SistaCogit methodsFor: 'accessing' stamp: 'eem 4/12/2017 09:15'!
getCogCodeZoneThreshold
	<doNotGenerate>
	^methodZone getCogCodeZoneThreshold! !

!SistaCogit methodsFor: 'method introspection' stamp: 'cb 4/28/2015 16:19'!
getJumpTargetPCAt: pc
	<api>
	^backEnd jumpTargetPCAt: pc! !

!SistaCogit methodsFor: 'initialization' stamp: 'cb 5/5/2015 13:12'!
initialize
	super initialize.
	branchReachedOnlyForCounterTrip := false.
	cogMethodSurrogateClass := (objectMemory ifNil: [self class objectMemoryClass]) wordSize = 4
										ifTrue: [CogSistaMethodSurrogate32]
										ifFalse: [CogSistaMethodSurrogate64]! !

!SistaCogit methodsFor: 'initialization' stamp: 'eem 7/4/2014 17:37'!
initializeCodeZoneFrom: startAddress upTo: endAddress
	initialCounterValue := MaxCounterValue.
	super initializeCodeZoneFrom: startAddress upTo: endAddress! !

!SistaCogit methodsFor: 'simulation only' stamp: 'eem 1/24/2017 13:34'!
isTrapAt: retpc
	"For stack depth checking."
	<doNotGenerate>
	^(backEnd isCallPrecedingReturnPC: retpc)
	 and: [(backEnd callTargetFromReturnAddress: retpc) = ceTrapTrampoline]! !

!SistaCogit methodsFor: 'compile abstract instructions' stamp: 'eem 7/5/2014 17:22'!
maybeAllocAndInitCounters
	<inline: true>
	self assert: counters = 0.
	counterIndex := 0.
	numCounters = 0 ifTrue:
		[^true].
	counters := objectRepresentation allocateCounters: numCounters.
	^counters ~= 0! !

!SistaCogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/1/2016 15:20'!
maybeCountCounter
	<inline: true>
	numCounters := numCounters + 1! !

!SistaCogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/25/2016 17:58'!
maybeCounterIndex
	<inline: true>
	^counterIndex! !

!SistaCogit methodsFor: 'compile abstract instructions' stamp: 'cb 1/8/2018 16:30'!
maybeDealWithUnsafeJumpForDescriptor: descriptor pc: pc latestContinuation: latestContinuation
	"Mapped: 250	backjumpAlwaysInterrupt
	 Unmapped: 6000	backjumpNoInterrupt 
				 7016-7020 jumpWritable/Young
				 8000-8003 type branches
	In all cases the distance is an integer pushed on stack just before with pushIntegerLong:"
	<var: #descriptor type: #'BytecodeDescriptor *'>
	| distance targetPC byte01 byte02 newContinuation |
	newContinuation := latestContinuation.
	descriptor hasUnsafeJump ifTrue:
		[ byte01 := objectMemory fetchByte: pc + 1 ofObject: methodObj.
		  byte02 := objectMemory fetchByte: pc + 2 ofObject: methodObj.
		  distance := self decodePushIntegerLongBefore: pc in: methodObj.
		  targetPC := pc + descriptor numBytes + distance.
		  descriptor isMapped 
			ifTrue: [ byte01 = 250 ifTrue: 
				[ "mapped always interrupt backjump" 
				 self maybeCountFixup: descriptor.
				 self initializeFixupAt: targetPC ] ]
			ifFalse: 
				[ byte02 >> 5 = 2r100 ifTrue: 
					[ "inlined sista primitive" 
					| prim |
					prim := (byte02 bitAnd: 16r1F) << 8 + byte01.
					prim >= 7000
						ifTrue: 
							["branch forward"
							 self maybeCountFixup: descriptor.
							 newContinuation := latestContinuation max: targetPC ]
						ifFalse: 
							[prim >= 6000 ifTrue: 
								["no interrupt back jump"
								self maybeCountFixup: descriptor.
								self initializeFixupAt: targetPC]]]]].
	^newContinuation! !

!SistaCogit methodsFor: 'compile abstract instructions' stamp: 'eem 4/6/2015 19:54'!
maybeFreeCounters
	<inline: true>
	counters ~= 0 ifTrue:
		[objectRepresentation freeCounters: counters]! !

!SistaCogit methodsFor: 'compaction' stamp: 'eem 7/4/2014 12:12'!
maybeFreeCountersOf: aCogMethod
	"Free any counters in the method."
	<inline: true>
	objectRepresentation freeCounters: aCogMethod counters! !

!SistaCogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/1/2016 15:20'!
maybeInitNumCounters
	<inline: true>
	numCounters := 0! !

!SistaCogit methodsFor: 'garbage collection' stamp: 'eem 7/5/2014 20:09'!
maybeMarkCountersIn: cogMethod
	"In SIsta Spur counters are held on the heap in pinned objects which must be marked
	 to avoid them being garbage collected.  This is the hook through which that happens."
	<var: #cogMethod type: #'CogMethod *'>
	<inline: true>
	objectRepresentation maybeMarkCounters: cogMethod counters! !

!SistaCogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/25/2016 17:58'!
maybeSetCounterIndex: value
	<inline: true>
	counterIndex := value! !

!SistaCogit methodsFor: 'compile abstract instructions' stamp: 'cb 1/8/2018 16:31'!
maybeUnsafeJumpContinuation: latestContinuation at: bcpc for: descriptor in: aMethodObj
	<inline: true>
	"Note: ignore backward jumps."
	<var: #descriptor type: #'BytecodeDescriptor *'>
	| distance targetPC byte01 byte02 newContinuation | 
	newContinuation := latestContinuation.
	descriptor hasUnsafeJump ifTrue:
		[ byte01 := objectMemory fetchByte: bcpc + 1 ofObject: aMethodObj.
		  byte02 := objectMemory fetchByte: bcpc + 2 ofObject: aMethodObj.
		  "pushIntegerLong"
		  distance := self decodePushIntegerLongBefore: bcpc in: methodObj.
		  targetPC := bcpc + descriptor numBytes + distance.
		  descriptor isMapped 
			ifFalse: 
				[ byte02 >> 5 = 2r100 ifTrue: 
					[ "inlined sista primitive" 
					| prim |
					prim := (byte02 bitAnd: 16r1F) << 8 + byte01.
					prim >= 7000
						ifTrue: 
							["branch forward"
							 newContinuation := latestContinuation max: targetPC ]]]].
	^newContinuation! !

!SistaCogit methodsFor: 'simulation only' stamp: 'cb 4/26/2016 09:27'!
methodsCompiledToMachineCodeInto: arrayObject
	<doNotGenerate>
	^methodZone methodsCompiledToMachineCodeInto: arrayObject! !

!SistaCogit methodsFor: 'simulation only' stamp: 'cb 4/26/2016 09:27'!
numMethods
	<doNotGenerate>
	^methodZone numMethods! !

!SistaCogit methodsFor: 'method introspection' stamp: 'eem 11/7/2016 08:32'!
picDataFor: descriptor Annotation: isBackwardBranchAndAnnotation Mcpc: mcpc Bcpc: bcpc Method: cogMethodArg
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<var: #mcpc type: #'char *'>
	<var: #cogMethodArg type: #'void *'>
	| annotation entryPoint tuple counter |
	"N.B. Counters are always 32-bits, having two 16-bit halves for the reached and taken counts."
	<var: #counter type: #'unsigned int'>

	descriptor ifNil:
		[^0].
	descriptor isBranch ifTrue:
		["it's a branch; conditional?"
		 (descriptor isBranchTrue or: [descriptor isBranchFalse]) ifTrue:
			[counter := (self
							cCoerce: ((self
											cCoerceSimple: cogMethodArg
											to: #'CogMethod *') counters)
							to: #'usqInt *')
								at: counterIndex.
			 tuple := self picDataForCounter: counter at: bcpc + 1.
			 tuple = 0 ifTrue: [^PrimErrNoMemory].
			 objectMemory storePointer: introspectionDataIndex ofObject: introspectionData withValue: tuple.
			 introspectionDataIndex := introspectionDataIndex + 1.
			 counterIndex := counterIndex + 1].
		 ^0].
	annotation := isBackwardBranchAndAnnotation >> 1.
	((self isPureSendAnnotation: annotation)
	 and: [entryPoint := backEnd callTargetFromReturnAddress: mcpc asUnsignedInteger.
		 entryPoint > methodZoneBase]) ifFalse: "send is not linked, or is not a send"
		[^0].
	self targetMethodAndSendTableFor: entryPoint "It's a linked send; find which kind."
		annotation: annotation
		into: [:targetMethod :sendTable| | methodClassIfSuper association |
			methodClassIfSuper := nil.
			sendTable = superSendTrampolines ifTrue:
				[methodClassIfSuper := coInterpreter methodClassOf: (self cCoerceSimple: cogMethodArg to: #'CogMethod *') methodObject].
			sendTable = directedSuperSendTrampolines ifTrue:
				[association := backEnd literalBeforeInlineCacheTagAt: mcpc asUnsignedInteger.
				 methodClassIfSuper := objectRepresentation valueOfAssociation: association].
			tuple := self picDataForSendTo: targetMethod
						methodClassIfSuper: methodClassIfSuper
						at: mcpc
						bcpc: bcpc + 1].
	tuple = 0 ifTrue: [^PrimErrNoMemory].
	objectMemory storePointer: introspectionDataIndex ofObject: introspectionData withValue: tuple.
	introspectionDataIndex := introspectionDataIndex + 1.
	^0! !

!SistaCogit methodsFor: 'method introspection' stamp: 'eem 4/19/2016 15:20'!
picDataFor: cogMethod into: arrayObj
	"Collect the branch and send data for cogMethod, storing it into arrayObj."
	<api>
	<var: #cogMethod type: #'CogMethod *'>
	| errCode |
	cogMethod stackCheckOffset = 0 ifTrue:
		[^0].
	introspectionDataIndex := counterIndex := 0.
	introspectionData := arrayObj.
	errCode := self
					mapFor: (self cCoerceSimple: cogMethod to: #'CogBlockMethod *')
					bcpc: (coInterpreter startPCOfMethod: cogMethod methodObject)
					performUntil: #picDataFor:Annotation:Mcpc:Bcpc:Method:
					arg: cogMethod asVoidPointer.
	errCode ~= 0 ifTrue:
		[self assert: errCode = PrimErrNoMemory.
		 ^-1].
	cogMethod blockEntryOffset ~= 0 ifTrue:
		[errCode := self blockDispatchTargetsFor: cogMethod
						perform: #picDataForBlockEntry:Method:
						arg: cogMethod asInteger.
		 errCode ~= 0 ifTrue:
			[self assert: errCode = PrimErrNoMemory.
			 ^-1]].
	^introspectionDataIndex! !

!SistaCogit methodsFor: 'method introspection' stamp: 'eem 4/19/2016 15:07'!
picDataForBlockEntry: blockEntryMcpc Method: cogMethod
	"Collect the branch and send data for the block method starting at blockEntryMcpc, storing it into picData."
	<returnTypeC: #usqInt>
	| cogBlockMethod |
	<var: #cogBlockMethod type: #'CogBlockMethod *'>
	cogBlockMethod := self cCoerceSimple: blockEntryMcpc - (self sizeof: CogBlockMethod)
							  to: #'CogBlockMethod *'.
	cogBlockMethod stackCheckOffset = 0 ifTrue:
		[^0].
	^self
		mapFor: cogBlockMethod
		bcpc: cogBlockMethod startpc
		performUntil: #picDataFor:Annotation:Mcpc:Bcpc:Method:
		arg: cogMethod asVoidPointer! !

!SistaCogit methodsFor: 'method introspection' stamp: 'eem 11/7/2016 08:32'!
picDataForCounter: counter at: bcpc
	| executedCount tuple untakenCount |
	"N.B. Counters are always 32-bits, having two 16-bit halves for the reached and taken counts."
	<var: #counter type: #'unsigned int'>
	tuple := objectMemory
				eeInstantiateClassIndex: ClassArrayCompactIndex
				format: objectMemory arrayFormat
				numSlots: 3.
	tuple = 0 ifTrue:
		[^0].
	self assert: CounterBytes = 4.
	executedCount := initialCounterValue - (counter >> 16).
	untakenCount := initialCounterValue - (counter bitAnd: 16rFFFF).
	objectMemory
		storePointerUnchecked: 0 ofObject: tuple withValue: (objectMemory integerObjectOf: bcpc);
		storePointerUnchecked: 1 ofObject: tuple withValue: (objectMemory integerObjectOf: executedCount);
		storePointerUnchecked: 2 ofObject: tuple withValue: (objectMemory integerObjectOf: untakenCount).
	^tuple! !

!SistaCogit methodsFor: 'method introspection' stamp: 'eem 4/3/2015 16:08'!
picDataForSendTo: cogMethod methodClassIfSuper: methodClassOrNil at: sendMcpc bcpc: sendBcpc
	"Answer a tuple with the send data for a linked send to cogMethod.
	 If the target is a CogMethod (monomorphic send) answer
		{ bytecode pc, inline cache class, target method }
	 If the target is an open PIC (megamorphic send) answer
		{ bytecode pc, nil, send selector }
	If the target is a closed PIC (polymorphic send) answer
		{ bytecode pc, first class, target method, second class, second target method, ... }"
	<var: #cogMethod type: #'CogMethod *'>
	<var: #sendMcpc type: #'char *'>
	| tuple class |
	tuple := objectMemory
					eeInstantiateClassIndex: ClassArrayCompactIndex
					format: objectMemory arrayFormat
					numSlots: (cogMethod cmType = CMClosedPIC
								ifTrue: [2 * cogMethod cPICNumCases + 1]
								ifFalse: [3]).
	tuple = 0 ifTrue:
		[^0].
	objectMemory storePointerUnchecked: 0 ofObject: tuple withValue: (objectMemory integerObjectOf: sendBcpc).
	cogMethod cmType = CMMethod ifTrue:
		[class := methodClassOrNil ifNil:
					[objectRepresentation classForInlineCacheTag: (backEnd inlineCacheTagAt: sendMcpc asUnsignedInteger)].
		 objectMemory
			storePointer: 1 ofObject: tuple withValue: class;
			storePointer: 2 ofObject: tuple withValue: cogMethod methodObject.
		^tuple].
	cogMethod cmType = CMClosedPIC ifTrue:
		[self populate: tuple withPICInfoFor: cogMethod firstCacheTag: (backEnd inlineCacheTagAt: sendMcpc asUnsignedInteger).
		^tuple].
	cogMethod cmType = CMOpenPIC ifTrue:
		[objectMemory
			storePointerUnchecked: 1 ofObject: tuple withValue: objectMemory nilObject;
			storePointer: 2 ofObject: tuple withValue: cogMethod selector.
		^tuple].
	self error: 'invalid method type'.
	^0 "to get Slang to type this method as answering sqInt"! !

!SistaCogit methodsFor: 'method introspection' stamp: 'eem 3/14/2016 11:38'!
populate: tuple withPICInfoFor: cPIC firstCacheTag: firstCacheTag
	"Populate tuple (which must be large enough) with the ClosedPIC's target method class pairs.
	 The first entry in tuple contains the bytecode pc for the send, so skip the tuple's first field."
	<var: #cPIC type: #'CogMethod *'>
	| pc cacheTag classOop entryPoint targetMethod value |
	<var: #targetMethod type: #'CogMethod *'>

	1 to: cPIC cPICNumCases do:
		[:i|
		pc := self addressOfEndOfCase: i inCPIC: cPIC.
		cacheTag := i = 1
						ifTrue: [firstCacheTag]
						ifFalse: [backEnd literalBeforeFollowingAddress: pc - backEnd jumpLongConditionalByteSize].
		classOop := objectRepresentation classForInlineCacheTag: cacheTag.
		objectMemory storePointer: i * 2 - 1 ofObject: tuple withValue: classOop.
		entryPoint := i = 1
						ifTrue: [backEnd jumpLongTargetBeforeFollowingAddress: pc]
						ifFalse: [backEnd jumpLongConditionalTargetBeforeFollowingAddress: pc].
		"Find target from jump.  A jump to the MNU entry-point should collect #doesNotUnderstand:"
		(cPIC containsAddress: entryPoint)
			ifTrue:
				[value := objectMemory splObj: SelectorDoesNotUnderstand]
			ifFalse:
				[targetMethod := self cCoerceSimple: entryPoint - cmNoCheckEntryOffset to: #'CogMethod *'.
				 self assert: targetMethod cmType = CMMethod.
				 value := targetMethod methodObject].
		objectMemory storePointer: i * 2 ofObject: tuple withValue: value]! !

!SistaCogit methodsFor: 'tests' stamp: 'eem 2/14/2011 15:35'!
printPICDataForMethods
	<doNotGenerate>
	methodZone methodsDo:
		[:cogMethod|
		cogMethod cmType = CMMethod ifTrue:
			[(coInterpreter picDataFor: cogMethod) ifNotNil:
				[:thePicData|
				coInterpreter printOop: thePicData]]]! !

!SistaCogit methodsFor: 'sista callbacks' stamp: 'eem 7/4/2014 12:23'!
resetCountersIn: cogMethod
	<doNotGenerate>
	objectRepresentation resetCountersIn: cogMethod! !

!SistaCogit methodsFor: 'accessing' stamp: 'eem 3/16/2017 16:49'!
setCogCodeZoneThreshold: threshold
	<doNotGenerate>
	^methodZone setCogCodeZoneThreshold: threshold! !

!StackToRegisterMappingCogit class methodsFor: 'translation' stamp: 'eem 06/07/2018 15:29'!
ancilliaryClasses
	^super ancilliaryClasses,
	  { self basicNew simStackEntryClass. self basicNew bytecodeFixupClass. CogSSOptStatus } ,
	(LowcodeVM ifTrue: [ { self basicNew simStackNativeEntryClass } ] ifFalse: [ #() ])! !

!StackToRegisterMappingCogit class methodsFor: 'documentation' stamp: 'eem 1/7/2016 14:51'!
callingConvention
	"The Smalltalk-to-Smalltalk calling convention aims to trade simplicity of compilation against
	 effectiveness of optimization.  Most Smalltalk methods, and certainly most performance-
	 critical primitives have two or less arguments.  So arranging that the receiver and up to two
	 args are in registers means that performance-critical primitives can access their arguments
	 in registers.  So if the argument count is <= numRegArgs nothing is passed on the stack and
	 everything is passed in ReceiverResultReg, Arg0Reg et al.  Above numRegArgs everything is
	 passed on the stack.

	 To save the CoInterpreter from change we shuffle the retpc and push the register args in
	 the prolog so that the frame format is unchanged by register args.  Also, the trampolines for
	 unlinked sends do the same, as does the code preceding an interpreter primitive.  It turns
	 out that this protocol is faster than always pushing arguments.  Comparing benchFib with the
	 shuffling protocol against an always-push protocol on a 2.66 GHz Core i7 (MacBook Pro) , the
	 shuffling protocol is 6.3% faster than the always push protocol.

	 Not shuffling the stack and pushing register arguments after frame build is faster yet again,
	 5.8% faster that the stack shuffle.  So it might be worth-while to change the CoInterpreter's
	 frame management to allow numArgs <= numRegArgs frames to push receiver and arguments
	 after saving the return pc.  This implies changes in stack-to-context mapping, GC,
	 interpreter-to-machine code frame conversion and no doubt else where.

	 Hence the calling convention is

		- if the number of arguments is less than or equal to numRegArgs then the receiver and
		  arguments are passed in registers.  numRegArgs is 1 for V3, and 2 for Spur.  The receiver
		  is passed in ReceiverResultReg, the first argument in Arg0Reg (esi on x86) and the second
		  argument (if numRegArgs = 2) in Arg1Reg (edi on x86).

		- if the number of arguments is greater than numRegArgs then the calling convention is as
		  for SimpleStackBasedCogIt; ReceiverResultReg contains the receiver, and the receiver and
		  arguments are all on the stack, receiver furthest from top-of-stack.  If the argument count
		  is > 2 then argument count is passed in SendNumArgsReg (for the benefit of the run-time
		  linking routines; it is ignored in linked sends).

		On return the result is in ReceiverResultReg.  The callee removes arguments from the stack
		(Pascal convention).

		Note that if a machine code method contains a call to an interpreter primitive it will push any
		register arguments (and if on a RISC, the return pc from the LinReg) on the stack before calling
		the primitive so that to the primitive the stack looks the same as it does in the interpreter.

		Within all machine code primitives except genPrimitiveClosureValue and genPrimitivePerform all
		arguments are taken from registers since no machine code primitive has more than numRegArgs
		arguments.  genPrimitiveClosureValue pushes its register arguments immedately only for laziness
		to be able to reuse SimpleStackBasedCogit's code.  genPrimitivePerform adjusts its arguments
		as required by special-purpose code.

		Within machine code methods with interpreter primitives the register arguments are pushed
		before calling the interpreter primitive.  In normal methods and if not already done so in primitive
		code, the register arguments are pushed during frame build.  If a method is compiled frameless
		it will access its arguments in registers."! !

!StackToRegisterMappingCogit class methodsFor: 'translation' stamp: 'eem 2/3/2020 12:41'!
declareCVarsIn: aCodeGen
	aCodeGen
		var: #methodAbortTrampolines
			declareC: 'sqInt methodAbortTrampolines[4]';
		var: #picAbortTrampolines
			declareC: 'sqInt picAbortTrampolines[4]';
		var: #picMissTrampolines
			declareC: 'sqInt picMissTrampolines[4]';
		var: 'ceCall0ArgsPIC'
			declareC: 'void (*ceCall0ArgsPIC)(void)';
		var: 'ceCall1ArgsPIC'
			declareC: 'void (*ceCall1ArgsPIC)(void)';
		var: 'ceCall2ArgsPIC'
			declareC: 'void (*ceCall2ArgsPIC)(void)';
		var: #ceCallCogCodePopReceiverArg0Regs
			declareC: 'void (*ceCallCogCodePopReceiverArg0Regs)(void)';
		var: #realCECallCogCodePopReceiverArg0Regs
			declareC: 'void (*realCECallCogCodePopReceiverArg0Regs)(void)';
		var: #ceCallCogCodePopReceiverArg1Arg0Regs
			declareC: 'void (*ceCallCogCodePopReceiverArg1Arg0Regs)(void)';
		var: #realCECallCogCodePopReceiverArg1Arg0Regs
			declareC: 'void (*realCECallCogCodePopReceiverArg1Arg0Regs)(void)';
		var: 'simStack'
			declareC: 'SimStackEntry simStack[', self simStackSlots asString, ']';
		var: 'simSelf'
			type: #CogSimStackEntry;
		var: #optStatus
			type: #CogSSOptStatus;
		var: 'prevBCDescriptor'
			type: #'BytecodeDescriptor *'.
	LowcodeVM
		ifTrue:
			[aCodeGen
				var: 'simNativeStack'
				declareC: 'CogSimStackNativeEntry simNativeStack[', self simNativeStackSlots asString, ']']
		ifFalse:
			[#(simNativeSpillBase simNativeStack "simNativeStackPtr" simNativeStackSize hasNativeFrame) do:
				[:lowcodeVar| aCodeGen removeVariable: lowcodeVar ifAbsent: []]].

	self numPushNilsFunction ifNotNil:
		[aCodeGen
			var: 'numPushNilsFunction'
				declareC: 'sqInt (* const numPushNilsFunction)(struct _BytecodeDescriptor *,sqInt,sqInt,sqInt) = ', (aCodeGen cFunctionNameFor: self numPushNilsFunction);
			var: 'pushNilSizeFunction'
				declareC: 'sqInt (* const pushNilSizeFunction)(sqInt,sqInt) = ', (aCodeGen cFunctionNameFor: self pushNilSizeFunction)].

	aCodeGen
		addSelectorTranslation: #register to: (aCodeGen cFunctionNameFor: 'registerr');
		addSelectorTranslation: #register: to: (aCodeGen cFunctionNameFor: 'registerr:').

	self declareFlagVarsAsByteIn: aCodeGen! !

!StackToRegisterMappingCogit class methodsFor: 'class initialization' stamp: 'nice 11/9/2019 10:19'!
initializeBytecodeTableForNewspeakV4
	"StackToRegisterMappingCogit initializeBytecodeTableForNewspeakV4"

	BytecodeSetHasDirectedSuperSend := true.
	BytecodeSetHasExtensions := true.
	numPushNilsFunction := #v4:Num:Push:Nils:.
	pushNilSizeFunction := #v4PushNilSize:numInitialNils:.
	NSSendIsPCAnnotated := true. "IsNSSendCall used by SendAbsentImplicit"
	FirstSpecialSelector := 80.
	NumSpecialSelectors := 32.
	self flag:
'Special selector send class must be inlined to agree with the interpreter, which
 inlines class.  If class is sent to e.g. a general instance of ProtoObject then unless
 class is inlined there will be an MNU.  It must be that the Cointerpreter and Cogit
 have identical semantics.  We get away with not hardwiring the other special
 selectors either because in the Cointerpreter they are not inlined or because they
 are inlined only to instances of classes for which there will always be a method.'.
	self generatorTableFrom: #(
		"1 byte bytecodes"
		(1    0   15 genPushReceiverVariableBytecode isInstVarRef needsFrameNever: 1)
		(1  16   31 genPushLiteralVariable16CasesBytecode needsFrameNever: 1)
		(1  32   63 genPushLiteralConstantBytecode needsFrameNever: 1)
		(1  64   75 genPushTemporaryVariableBytecode needsFrameIfMod16GENumArgs: 1)
		(1  76   76 genPushReceiverBytecode needsFrameNever: 1)
		(1  77   77 genExtPushPseudoVariableOrOuterBytecode needsFrameIfExtBGT2: 1)
		(1  78   78 genPushConstantZeroBytecode needsFrameNever: 1)
		(1  79   79 genPushConstantOneBytecode needsFrameNever: 1)

		(1   80   80 genSpecialSelectorArithmetic isMapped AddRR)
		(1   81   81 genSpecialSelectorArithmetic isMapped SubRR)
		(1   82   82 genSpecialSelectorComparison isMapped JumpLess)
		(1   83   83 genSpecialSelectorComparison isMapped JumpGreater)
		(1   84   84 genSpecialSelectorComparison isMapped JumpLessOrEqual)
		(1   85   85 genSpecialSelectorComparison isMapped JumpGreaterOrEqual)
		(1   86   86 genSpecialSelectorComparison isMapped JumpZero)
		(1   87   87 genSpecialSelectorComparison isMapped JumpNonZero)
		(1   88   93 genSpecialSelectorSend isMapped)	 " #* #/ #\\ #@ #bitShift: //"
		(1   94   94 genSpecialSelectorArithmetic isMapped AndRR)
		(1   95   95 genSpecialSelectorArithmetic isMapped OrRR)
		(1   96 101 genSpecialSelectorSend isMapped) "#at: #at:put: #size #next #nextPut: #atEnd"
		(1 102 102 genSpecialSelectorEqualsEquals needsFrameNever: notMapped -1) "not mapped because it is directly inlined (for now)"
		(1 103 103 genSpecialSelectorClass needsFrameIfStackGreaterThanOne: notMapped 0) "not mapped because it is directly inlined (for now)"
		(1 104 104 genSpecialSelectorNotEqualsEquals needsFrameNever: notMapped -1) "not mapped because it is directly inlined (for now)"
		(1 105 111 genSpecialSelectorSend isMapped) " #value #value: #do: #new #new: #x #y"

		(1 112 127 genSendLiteralSelector0ArgsBytecode isMapped)
		(1 128 143 genSendLiteralSelector1ArgBytecode isMapped)
		(1 144 159 genSendLiteralSelector2ArgsBytecode isMapped)
		(1 160 175	genSendAbsentImplicit0ArgsBytecode isMapped hasIRC)
			
		(1 176 183 genStoreAndPopReceiverVariableBytecode isInstVarRef is1ByteInstVarStore isMappedIfImmutability needsFrameIfImmutability: -1)
			
		(1 184 191 genStoreAndPopTemporaryVariableBytecode)

		(1 192 199 genShortUnconditionalJump	branch v3:ShortForward:Branch:Distance:)
		(1 200 207 genShortJumpIfTrue			branch isBranchTrue isMapped "because of mustBeBoolean"
													v3:ShortForward:Branch:Distance:)
		(1 208 215 genShortJumpIfFalse			branch isBranchFalse isMapped "because of mustBeBoolean"
													v3:ShortForward:Branch:Distance:)

		(1 216 216 genReturnReceiver				return needsFrameIfInBlock: isMappedInBlock 0)
		(1 217 217 genReturnTopFromMethod		return needsFrameIfInBlock: isMappedInBlock -1)
		(1 218 218 genExtReturnTopFromBlock	return needsFrameNever: -1)

		(1 219 219 duplicateTopBytecode			needsFrameNever: 1)
		(1 220 220 genPopStackBytecode			needsFrameNever: -1)
		(1 221 221 genExtNopBytecode			needsFrameNever: 0)
		(1 222 223	unknownBytecode)

		"2 byte bytecodes"
		(2 224 224 extABytecode extension					needsFrameNever: 0)
		(2 225 225 extBBytecode extension					needsFrameNever: 0)
		(2 226 226 genExtPushReceiverVariableBytecode isInstVarRef)
		(2 227 227 genExtPushLiteralVariableBytecode		needsFrameNever: 1)
		(2 228 228 genExtPushLiteralBytecode					needsFrameNever: 1)
		(2 229 229 genExtPushIntegerBytecode				needsFrameNever: 1)
		(2 230 230 genLongPushTemporaryVariableBytecode)
		(2 231 231 genPushNewArrayBytecode)
		(2 232 232 genExtStoreReceiverVariableBytecode isInstVarRef isMappedIfImmutability)
		(2 233 233 genExtStoreLiteralVariableBytecode isMappedIfImmutability)
		(2 234 234 genLongStoreTemporaryVariableBytecode)
		(2 235 235 genExtStoreAndPopReceiverVariableBytecode isInstVarRef isMappedIfImmutability)
		(2 236 236 genExtStoreAndPopLiteralVariableBytecode isMappedIfImmutability)
		(2 237 237 genLongStoreAndPopTemporaryVariableBytecode)

		(2 238 238 genExtSendBytecode isMapped)
		(2 239 239 genExtSendSuperBytecode isMapped)
		(2 240 240 genExtSendAbsentImplicitBytecode isMapped hasIRC)
		(2 241 241 genExtSendAbsentDynamicSuperBytecode isMapped hasIRC)

		(2 242 242 genExtUnconditionalJump	branch isMapped "because of interrupt check" v4:Long:Branch:Distance:)
		(2 243 243 genExtJumpIfTrue			branch isBranchTrue isMapped "because of mustBeBoolean" v4:Long:Branch:Distance:)
		(2 244 244 genExtJumpIfFalse			branch isBranchFalse isMapped "because of mustBeBoolean" v4:Long:Branch:Distance:)

		(2 245 245	genExtSendAbsentSelfBytecode isMapped hasIRC)

		(2 246 248	unknownBytecode)

		"3 byte bytecodes"
		(3 249 249 genCallPrimitiveBytecode)
		(3 250 250 genPushRemoteTempLongBytecode)
		(3 251 251 genStoreRemoteTempLongBytecode)
		(3 252 252 genStoreAndPopRemoteTempLongBytecode)
		(3 253 253 genExtPushClosureBytecode block v4:Block:Code:Size:)
		(3 254 254	genExtSendAbsentOuterBytecode isMapped hasIRC)

		(3 255 255	unknownBytecode))! !

!StackToRegisterMappingCogit class methodsFor: 'class initialization' stamp: 'eem 3/10/2018 12:39'!
initializeBytecodeTableForSistaV1
	"StackToRegisterMappingCogit initializeBytecodeTableForSistaV1"

	numPushNilsFunction := #sistaV1:Num:Push:Nils:.
	pushNilSizeFunction := #sistaV1PushNilSize:numInitialNils:.
	BytecodeSetHasDirectedSuperSend := true.
	FirstSpecialSelector := 96.
	NumSpecialSelectors := 32.
	self flag:
'Special selector send class must be inlined to agree with the interpreter, which
 inlines class.  If class is sent to e.g. a general instance of ProtoObject then unless
 class is inlined there will be an MNU.  It must be that the Cointerpreter and Cogit
 have identical semantics.  We get away with not hardwiring the other special
 selectors either because in the Cointerpreter they are not inlined or because they
 are inlined only to instances of classes for which there will always be a method.'.
	self generatorTableFrom: #(
		"1 byte bytecodes"
		"pushes"
		(1    0   15 genPushReceiverVariableBytecode isInstVarRef		needsFrameNever: 1)
		(1  16   31 genPushLiteralVariable16CasesBytecode			needsFrameNever: 1)
		(1  32   63 genPushLiteralConstantBytecode					needsFrameNever: 1)
		(1  64   75 genPushTemporaryVariableBytecode				needsFrameIfMod16GENumArgs: 1)
		(1  76   76 genPushReceiverBytecode							needsFrameNever: 1)
		(1  77   77 genPushConstantTrueBytecode						needsFrameNever: 1)
		(1  78   78 genPushConstantFalseBytecode					needsFrameNever: 1)
		(1  79   79 genPushConstantNilBytecode						needsFrameNever: 1)
		(1  80   80 genPushConstantZeroBytecode						needsFrameNever: 1)
		(1  81   81 genPushConstantOneBytecode						needsFrameNever: 1)
		(1  82   82 genExtPushPseudoVariable)
		(1  83   83 duplicateTopBytecode								needsFrameNever: 1)

		(1  84   87 unknownBytecode)

		"returns"
		(1  88   88 genReturnReceiver				return needsFrameIfInBlock: isMappedInBlock 0)
		(1  89   89 genReturnTrue					return needsFrameIfInBlock: isMappedInBlock 0)
		(1  90   90 genReturnFalse					return needsFrameIfInBlock: isMappedInBlock 0)
		(1  91   91 genReturnNil					return needsFrameIfInBlock: isMappedInBlock 0)
		(1  92   92 genReturnTopFromMethod		return needsFrameIfInBlock: isMappedInBlock -1)
		(1  93   93 genReturnNilFromBlock			return needsFrameNever: -1)
		(1  94   94 genReturnTopFromBlock		return needsFrameNever: -1)
		(1  95   95 genExtNopBytecode			needsFrameNever: 0)

		"sends"
		(1  96   96 genSpecialSelectorArithmetic isMapped AddRR)
		(1  97   97 genSpecialSelectorArithmetic isMapped SubRR)
		(1  98   98 genSpecialSelectorComparison isMapped JumpLess)
		(1  99   99 genSpecialSelectorComparison isMapped JumpGreater)
		(1 100 100 genSpecialSelectorComparison isMapped JumpLessOrEqual)
		(1 101 101 genSpecialSelectorComparison isMapped JumpGreaterOrEqual)
		(1 102 102 genSpecialSelectorComparison isMapped JumpZero)
		(1 103 103 genSpecialSelectorComparison isMapped JumpNonZero)
		(1 104 109 genSpecialSelectorSend isMapped)	 " #* #/ #\\ #@ #bitShift: //"
		(1 110 110 genSpecialSelectorArithmetic isMapped AndRR)
		(1 111 111 genSpecialSelectorArithmetic isMapped OrRR)
		(1 112 117 genSpecialSelectorSend isMapped) "#at: #at:put: #size #next #nextPut: #atEnd"
		(1 118 118 genSpecialSelectorEqualsEquals needsFrameNever: notMapped -1) "not mapped because it is directly inlined (for now)"
		(1 119 119 genSpecialSelectorClass needsFrameIfStackGreaterThanOne: notMapped 0) "not mapped because it is directly inlined (for now)"
		(1 120 120 genSpecialSelectorNotEqualsEquals needsFrameNever: notMapped -1) "not mapped because it is directly inlined (for now)"
		(1 121 127 genSpecialSelectorSend isMapped) "#value #value: #do: #new #new: #x #y"

		(1 128 143 genSendLiteralSelector0ArgsBytecode isMapped)
		(1 144 159 genSendLiteralSelector1ArgBytecode isMapped)
		(1 160 175 genSendLiteralSelector2ArgsBytecode isMapped)

		"jumps"
		(1 176 183 genShortUnconditionalJump	branch v3:ShortForward:Branch:Distance:)
		(1 184 191 genShortJumpIfTrue			branch isBranchTrue isMapped "because of mustBeBoolean"
													v3:ShortForward:Branch:Distance:)
		(1 192 199 genShortJumpIfFalse			branch isBranchFalse isMapped "because of mustBeBoolean"
													v3:ShortForward:Branch:Distance:)
		(1 200 207 genStoreAndPopReceiverVariableBytecode isInstVarRef is1ByteInstVarStore isMappedIfImmutability needsFrameIfImmutability: -1)
		
		(1 208 215 genStoreAndPopTemporaryVariableBytecode)

		(1 216 216 genPopStackBytecode needsFrameNever: -1)

		(1 217 217 genUnconditionalTrapBytecode isMapped)

		(1 218 223 unknownBytecode)

		"2 byte bytecodes"
		(2 224 224 extABytecode extension)
		(2 225 225 extBBytecode extension)

		"pushes"
		(2 226 226 genExtPushReceiverVariableBytecode isInstVarRef)		"Needs a frame for context inst var access"
		(2 227 227 genExtPushLiteralVariableBytecode			needsFrameNever: 1)
		(2 228 228 genExtPushLiteralBytecode					needsFrameNever: 1)
		(2 229 229 genLongPushTemporaryVariableBytecode)
		(2 230 230 unknownBytecode)
		(2 231 231 genPushNewArrayBytecode)
		(2 232 232 genExtPushIntegerBytecode				needsFrameNever: 1)
		(2 233 233 genExtPushCharacterBytecode				needsFrameNever: 1)

		"returns"
		"sends"
		(2 234 234 genExtSendBytecode isMapped)
		(2 235 235 genExtSendSuperBytecode isMapped)

		"sista bytecodes"
		(2 236 236 genCallMappedInlinedPrimitive isMapped hasUnsafeJump)

		"jumps"
		(2 237 237 genExtUnconditionalJump	branch isMapped "because of interrupt check" v4:Long:Branch:Distance:)
		(2 238 238 genExtJumpIfTrue			branch isBranchTrue isMapped "because of mustBeBoolean" v4:Long:Branch:Distance:)
		(2 239 239 genExtJumpIfFalse			branch isBranchFalse isMapped "because of mustBeBoolean" v4:Long:Branch:Distance:)

		"stores"
		(2 240 240 genExtStoreAndPopReceiverVariableBytecode isInstVarRef isMappedIfImmutability)
		(2 241 241 genExtStoreAndPopLiteralVariableBytecode isMappedIfImmutability)
		(2 242 242 genLongStoreAndPopTemporaryVariableBytecode)
		(2 243 243 genExtStoreReceiverVariableBytecode isInstVarRef isMappedIfImmutability)
		(2 244 244 genExtStoreLiteralVariableBytecode isMappedIfImmutability)
		(2 245 245 genLongStoreTemporaryVariableBytecode)

		(2 246 247	unknownBytecode)

		"3 byte bytecodes"
		(3 248 248 genCallPrimitiveBytecode hasUnsafeJump)
		(3 249 249 genExtPushFullClosureBytecode)
		(3 250 250 genExtPushClosureBytecode block v4:Block:Code:Size:)
		(3 251 251 genPushRemoteTempLongBytecode)
		(3 252 252 genStoreRemoteTempLongBytecode)
		(3 253 253 genStoreAndPopRemoteTempLongBytecode)
		
		(3 254 255	unknownBytecode))! !

!StackToRegisterMappingCogit class methodsFor: 'class initialization' stamp: 'eem 6/7/2018 15:23'!
initializeBytecodeTableForSqueakV3PlusClosures
	"StackToRegisterMappingCogit initializeBytecodeTableForSqueakV3PlusClosures"

	numPushNilsFunction := #v3:Num:Push:Nils:.
	pushNilSizeFunction := #v3PushNilSize:numInitialNils:.
	FirstSpecialSelector := 176.
	NumSpecialSelectors := 32.
	self flag:
'Special selector send class must be inlined to agree with the interpreter, which
 inlines class.  If class is sent to e.g. a general instance of ProtoObject then unless
 class is inlined there will be an MNU.  It must be that the Cointerpreter and Cogit
 have identical semantics.  We get away with not hardwiring the other special
 selectors either because in the Cointerpreter they are not inlined or because they
 are inlined only to instances of classes for which there will always be a method.'.
	self generatorTableFrom: #(
		(1    0   15 genPushReceiverVariableBytecode isInstVarRef needsFrameNever: 1)
		(1  16   31 genPushTemporaryVariableBytecode needsFrameIfMod16GENumArgs: 1)
		(1  32   63 genPushLiteralConstantBytecode needsFrameNever: 1)
		(1  64   95 genPushLiteralVariableBytecode needsFrameNever: 1)
		(1  96 103 genStoreAndPopReceiverVariableBytecode isInstVarRef is1ByteInstVarStore isMappedIfImmutability needsFrameIfImmutability: -1)
		(1 104 111 genStoreAndPopTemporaryVariableBytecode)
		(1 112 112 genPushReceiverBytecode needsFrameNever: 1)
		(1 113 113 genPushConstantTrueBytecode needsFrameNever: 1)
		(1 114 114 genPushConstantFalseBytecode needsFrameNever: 1)
		(1 115 115 genPushConstantNilBytecode needsFrameNever: 1)
		(1 116 119 genPushQuickIntegerConstantBytecode needsFrameNever: 1)
		"method returns in blocks need a frame because of nonlocalReturn:through:"
		(1 120 120 genReturnReceiver				return needsFrameIfInBlock: isMappedInBlock 0)
		(1 121 121 genReturnTrue					return needsFrameIfInBlock: isMappedInBlock 0)
		(1 122 122 genReturnFalse					return needsFrameIfInBlock: isMappedInBlock 0)
		(1 123 123 genReturnNil					return needsFrameIfInBlock: isMappedInBlock 0)
		(1 124 124 genReturnTopFromMethod		return needsFrameIfInBlock: isMappedInBlock -1)
		(1 125 125 genReturnTopFromBlock		return needsFrameNever: -1)

		(1 126 127 unknownBytecode)

		(2 128 128 extendedPushBytecode isInstVarRef) "well, maybe inst var ref"
		(2 129 129 extendedStoreBytecode isInstVarRef isMappedIfImmutability) "well, maybe inst var ref"
		(2 130 130 extendedStoreAndPopBytecode isInstVarRef isMappedIfImmutability) "well, maybe inst var ref"
		(2 131 131 genExtendedSendBytecode isMapped)
		(3 132 132 doubleExtendedDoAnythingBytecode isMapped) "well, maybe inst var ref"
		(2 133 133 genExtendedSuperBytecode isInstVarRef isMapped)
		(2 134 134 genSecondExtendedSendBytecode isMapped)
		(1 135 135 genPopStackBytecode needsFrameNever: -1)
		(1 136 136 duplicateTopBytecode needsFrameNever: 1)

		(1 137 137 genPushActiveContextBytecode)
		(2 138 138 genPushNewArrayBytecode)),

		((InitializationOptions at: #SpurObjectMemory ifAbsent: [false])
			ifTrue: [#((3 139 139 genCallPrimitiveBytecode))]
			ifFalse: [#((1 139 139 unknownBytecode))]),

	   #(
		(3 140 140 genPushRemoteTempLongBytecode)
		(3 141 141 genStoreRemoteTempLongBytecode)
		(3 142 142 genStoreAndPopRemoteTempLongBytecode)
		(4 143 143 genPushClosureCopyCopiedValuesBytecode block v3:Block:Code:Size:)

		(1 144 151 genShortUnconditionalJump			branch v3:ShortForward:Branch:Distance:)
		(1 152 159 genShortJumpIfFalse					branch isBranchFalse isMapped "because of mustBeBoolean"
															v3:ShortForward:Branch:Distance:)
		(2 160 163 genLongUnconditionalBackwardJump	branch isMapped "because of interrupt check"
															v3:Long:Branch:Distance:)
		(2 164 167 genLongUnconditionalForwardJump		branch v3:Long:Branch:Distance:)
		(2 168 171 genLongJumpIfTrue					branch isBranchTrue isMapped "because of mustBeBoolean"
															v3:LongForward:Branch:Distance:)
		(2 172 175 genLongJumpIfFalse					branch isBranchFalse isMapped "because of mustBeBoolean"
															v3:LongForward:Branch:Distance:)

		(1 176 176 genSpecialSelectorArithmetic isMapped AddRR)
		(1 177 177 genSpecialSelectorArithmetic isMapped SubRR)
		(1 178 178 genSpecialSelectorComparison isMapped JumpLess)
		(1 179 179 genSpecialSelectorComparison isMapped JumpGreater)
		(1 180 180 genSpecialSelectorComparison isMapped JumpLessOrEqual)
		(1 181 181 genSpecialSelectorComparison isMapped JumpGreaterOrEqual)
		(1 182 182 genSpecialSelectorComparison isMapped JumpZero)
		(1 183 183 genSpecialSelectorComparison isMapped JumpNonZero)
		(1 184 189 genSpecialSelectorSend isMapped)	 " #* #/ #\\ #@ #bitShift: //"
		(1 190 190 genSpecialSelectorArithmetic isMapped AndRR)
		(1 191 191 genSpecialSelectorArithmetic isMapped OrRR)
		(1 192 197 genSpecialSelectorSend isMapped) "#at: #at:put: #size #next #nextPut: #atEnd"
		(1 198 198 genSpecialSelectorEqualsEquals needsFrameNever: notMapped -1) "not mapped because it is directly inlined (for now)"
		(1 199 199 genSpecialSelectorClass needsFrameIfStackGreaterThanOne: notMapped 0) "not mapped because it is directly inlined (for now)"
		(1 200 200 genSpecialSelectorNotEqualsEquals needsFrameNever: notMapped -1) "not mapped because it is directly inlined (for now)"
		(1 201 207 genSpecialSelectorSend isMapped) " #value #value: #do: #new #new: #x #y"
		(1 208 223 genSendLiteralSelector0ArgsBytecode isMapped)
		(1 224 239 genSendLiteralSelector1ArgBytecode isMapped)
		(1 240 255 genSendLiteralSelector2ArgsBytecode isMapped))! !

!StackToRegisterMappingCogit class methodsFor: 'class initialization' stamp: 'eem 4/28/2015 14:50'!
initializeBytecodeTableForSqueakV3PlusClosuresNewspeakV4Hybrid
	"SimpleStackBasedCogit initializeBytecodeTableForSqueakV3PlusClosuresNewspeakV4Hybrid"
	"StackToRegisterMappingCogit initializeBytecodeTableForSqueakV3PlusClosuresNewspeakV4Hybrid"

	super initializeBytecodeTableForSqueakV3PlusClosuresNewspeakV4Hybrid.
	numPushNilsFunction := #v3or4:Num:Push:Nils:.
	pushNilSizeFunction := #v3or4PushNilSize:numInitialNils:! !

!StackToRegisterMappingCogit class methodsFor: 'class initialization' stamp: 'eem 4/2/2015 16:42'!
initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid
	"StackToRegisterMappingCogit initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid"

	super initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid.
	numPushNilsFunction := #squeakV3orSistaV1:Num:Push:Nils:.
	pushNilSizeFunction := #squeakV3orSistaV1PushNilSize:numInitialNils:! !

!StackToRegisterMappingCogit class methodsFor: 'class initialization' stamp: 'RonieSalgado 4/29/2016 17:22'!
initializeSimStackConstants
	"The simulation stack is used to delay code generation until operands are consumed by
	 some operation, thereby avoiding pushing operands to the real stack and enabling
	 mapping stack contents to registers, and cheaply apply various peephole optimizations.
	 The simulation stack is an array of CogSimStackEntry structs.  Each entry defines the
	 object on the virtual stack (Smalltalk context stack) as compilation proceeds.  See
	 stackToRegisterMapping in this class for documentation."

	SSIllegal := 0.
	SSBaseOffset := 1.
	SSConstant := 2.
	SSRegister := 3.
	SSSpill := 4.
	
	" Native simulation stack"
	SSNativeRegister := 5.
	SSRegisterPair := 6.
	SSRegisterDoubleFloat := 7.
	SSRegisterSingleFloat := 8.

	SSSpillNative := 9.
	SSSpillInt64 := 10.
	SSSpillFloat32 := 11.
	SSSpillFloat64 := 12.

	SSConstantInt32 := 13.
	SSConstantInt64 := 14.
	SSConstantFloat32 := 15.
	SSConstantFloat64 := 16.
	SSConstantNativePointer := 17.! !

!StackToRegisterMappingCogit class methodsFor: 'class initialization' stamp: 'eem 4/6/2012 14:25'!
initializeWithOptions: optionsDictionary

	super initializeWithOptions: optionsDictionary.
	self initializeSimStackConstants! !

!StackToRegisterMappingCogit class methodsFor: 'translation' stamp: 'eem 4/18/2016 09:09'!
isAcceptableAncilliaryClass: aClass
	^(aClass includesBehavior: CogBytecodeFixup)
		ifTrue: [aClass == self basicNew bytecodeFixupClass]
		ifFalse:
			[(aClass includesBehavior: CogSimStackEntry)
				ifTrue: [aClass == self basicNew simStackEntryClass]
				ifFalse: [true]]! !

!StackToRegisterMappingCogit class methodsFor: 'translation' stamp: 'eem 7/5/2014 18:41'!
mustBeGlobal: var
	"Answer if a variable must be global and exported.  Used for inst vars that are accessed from VM support code."

	^(super mustBeGlobal: var)
	   or: [#('ceCallCogCodePopReceiverArg0Regs' 'ceCallCogCodePopReceiverArg1Arg0Regs'
			'realCECallCogCodePopReceiverArg0Regs' 'realCECallCogCodePopReceiverArg1Arg0Regs'
			'ceCall0ArgsPIC' 'ceCall1ArgsPIC' 'ceCall2ArgsPIC') includes: var]! !

!StackToRegisterMappingCogit class methodsFor: 'accessing' stamp: 'eem 07/29/2014 06:48'!
numPushNilsFunction
	"Answer the value of numPushNilsFunction"

	^numPushNilsFunction! !

!StackToRegisterMappingCogit class methodsFor: 'trampoline support' stamp: 'eem 6/4/2015 12:22'!
numTrampolines
	^super numTrampolines + 12 "includes register args aborts"

	"Cogit withAllSubclasses collect: [:c| {c. (c instVarNames select: [:ea| ea beginsWith: 'ce']) size}]"
	"self instVarNames select: [:ea| ea beginsWith: 'ce']"! !

!StackToRegisterMappingCogit class methodsFor: 'accessing' stamp: 'eem 11/15/2012 16:28'!
pushNilSizeFunction
	"Answer the value of pushNilSizeFunction"

	^ pushNilSizeFunction! !

!StackToRegisterMappingCogit class methodsFor: 'translation' stamp: 'eem 07/29/2014 06:48'!
requiredMethodNames: options
	^(super requiredMethodNames: options)
		add: self numPushNilsFunction;
		add: self pushNilSizeFunction;
		yourself! !

!StackToRegisterMappingCogit class methodsFor: 'translation' stamp: 'eem 10/31/2013 00:05'!
shouldGenerateTypedefFor: aStructClass
	"Hack to work-around mutliple definitions.  Sometimes a type has been defined in an include."
	^aStructClass ~~ CogBytecodeFixup "overridden by CogSSBytecodeFixup"
	  and: [super shouldGenerateTypedefFor: aStructClass]! !

!StackToRegisterMappingCogit class methodsFor: 'translation' stamp: 'rsf 1/15/2017 21:50'!
simNativeStackSlots
	^ self basicNew simNativeStackSlots! !

!StackToRegisterMappingCogit class methodsFor: 'translation' stamp: 'cb 4/15/2016 16:15'!
simStackSlots
	^ self basicNew simStackSlots! !

!StackToRegisterMappingCogit class methodsFor: 'documentation' stamp: 'eem 09/28/2015 18:49'!
stackToRegisterMapping
	"Stack to register mapping is enabled via a simulation stack { simStack. simStackPtr, simSpillBase } of
	 operand descriptors (CogSimStackEntry) which serve
		- to avoid pushing operands to the actual stack by deferring operand manipulation until an
		  operand-consuming operation (send, store, run-time call)
		- to record operand type information for constants to avoid unnecessary type checks (e.g. tag checks)
		- as a simple register allocator since any live registers are recorded in descriptors on the stack.

	The operand types are
		SSBaseOffset - a value in memory at an offset relative to some register.  For method receiver args
						 and temps the base register is  FPReg (in a frameful method).  For indirect temps
						 the register could be any unassigned register.
		SSConstant - a method literal, hence a Smalltalk object
		SSRegister - the result of an expression assigned to a register
		SSSpill - a value spilled to the actual stack
	The special descriptor simSelf defines self in the current method, relative to FPReg in frameful
	 methods and  in a register in frameless methods.

	The register allocator aspect allocates registers by searching for SSBaseOffset and SSRegister
	 descriptors, computing the set of live registers, and then enumerating to find unused ones.
	 Simulation stack contents must be spilled to the actual stack
		- at a send (since at a suspension point the actual stack must be valid),
		- to make a register available if the code generator needs it
		- at a control flow join (since the two control flows could compute different stack contents and
		  we choose to avoid the complexity of saving stack contents to allow merging at join points).

	At a control-flow join we must discard type information for values pushed to the stack in either
	arm of the control-flow, but need /not/ for items pushed before the control flow diverged.  e.g. in
		self at: 1 put: (expr ifTrue: [v1] ifFalse: [v2]).
	the 1 is still valid after the control flow join for (expr ifTrue: [v1] ifFalse: [v2]).  So at a conditional
	branch we record simStackPtr in the target fixup and only void types between it and the
	simStackPtr at the join point.  This type voiding operation is called merge:.  For now we simply throw
	away all type info but would like to implement the baove scheme soon.

	 We can determine the stack depth at a conditional branch (if), but how do we determine the stack
	 depth following an unconditional jump (else)?  There are essentially three cases
		e ifTrue: [u] ifFalse: [v],
		e ifTrue: [^u] ifFalse: [v],
		e ifTrue: [u] ifFalse: [^v]

		1		expr
		2		jumpCond L1
		3		push
		4		jump L2
		5	L1:
		6		push
		7	L2:

		1		expr
		2		jumpCond L1
		3		ret
		4	L1:
		5		push

		1		expr
		2		jumpCond L1
		3		push
		4		jump L2
		5	L1:
		6		ret
		7	L2:

	In the first case we can know the merge base at L2 by propagating the merge base from 4 jump L2, which
	precedes the target of 2 jumpCond L1.  i.e. the merge base at 7 L2 is the stack pointer at 4 jump L2, which
	precedes the target of 2 jumpCond L1.  So at 2 jumpCond L1 we copy the stack pointer to the merge base
	at 5 L1, /and/ to the preceding 4 jump L2, and when we reach 4 jump L2, propagate the merge base to 7 L2.

	 Since we're conscious of JIT performance we restrict the live register search range by maintaining
	 simSpillBase, which is the index of the unspilled entry furthest from the end of simulation stack.
	 Only entries from simSpillBase to simStackPtr can contain unspilled, and hence live and volatile
	 registers (the FPReg is not volatile).

	 We further optimize by maintaining a simple optimization status for register contents.
	 We record whether ReceiverResultReg contains the receiver or an indirect temp vector
	 and merge this status at control-flow joins."! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/4/2015 09:53'!
addBlockStartAt: bytecodepc numArgs: numArgs numCopied: numCopied span: span
	"Add a blockStart for an embedded block.  For a binary tree walk block dispatch
	 blocks must be compiled in pc/depth-first order but are scanned in breadth-first
	 order, so do an insertion sort (which of course is really a bubble sort because we
	 have to move everything higher to make room)."
	<returnTypeC: #'BlockStart *'>
	| i blockStart |
	<var: #blockStart type: #'BlockStart *'>
	"Transcript ensureCr; nextPutAll: 'addBlockStartAt: '; print: bytecodepc; cr; flush."
	blockCount > 0
		ifTrue:
			[i := blockCount - 1.
			 [blockStart := self addressOf: (blockStarts at: i).
			  "check for repeat addition during recompilation due to initialNil miscount."
			  blockStart startpc = bytecodepc ifTrue:
				[^blockStart].
			  blockStart startpc > bytecodepc
			  and: [i > 0]] whileTrue:
				[i := i - 1].
			 blockCount to: i + 1 by: -1 do:
				[:j|
				blockStarts at: j put: (blockStarts at: j - 1)].
			blockStart := self cCode: [self addressOf: (blockStarts at: i + 1)]
								inSmalltalk: [blockStarts at: i + 1 put: CogBlockStart new]]
		ifFalse:
			[blockStart := self cCode: [self addressOf: (blockStarts at: blockCount)]
								inSmalltalk: [blockStarts at: blockCount put: CogBlockStart new]].
	
	blockCount := blockCount + 1.
	blockStart
		startpc: bytecodepc;
		numArgs: numArgs;
		numCopied: numCopied;
		numInitialNils: 0;
		stackCheckLabel: nil;
		hasInstVarRef: false;
		span: span.
	^blockStart! !

!StackToRegisterMappingCogit methodsFor: 'primitive generators' stamp: 'eem 2/23/2016 16:24'!
adjustArgumentsForPerform: numArgs
	"Generate code to adjust the possibly stacked arguments immediately
	 before jumping to a method looked up by a perform primitive."
	self assert: self numRegArgs <= 2.
	self assert: numArgs >= 1.
	numArgs <= self numRegArgs ifTrue:
		[numArgs = 2 ifTrue:
			[self MoveR: Arg1Reg R: Arg0Reg].
		 ^self].

	"If the arity is one more than the max numRegArgs, the receiver and all arguments have to be removed from the stack."
	self numRegArgs + 1 = numArgs ifTrue:
		[backEnd hasLinkRegister
			ifTrue:
				[self numRegArgs = 2
					ifTrue:
						[self MoveMw: 0 r: SPReg R: Arg1Reg.
						 self MoveMw: objectMemory wordSize r: SPReg R: Arg0Reg]
					ifFalse:
						[self MoveMw: 0 r: SPReg R: Arg0Reg].
				 self AddCq: numArgs + 1 * objectMemory wordSize R: SPReg]
			ifFalse:
				[self MoveMw: 0 r: SPReg R: TempReg. "save retpc"
				 self numRegArgs = 2
					ifTrue:
						[self MoveMw: objectMemory wordSize r: SPReg R: Arg1Reg.
						 self MoveMw: objectMemory wordSize * 2 r: SPReg R: Arg0Reg]
					ifFalse:
						[self MoveMw: objectMemory wordSize r: SPReg R: Arg0Reg].
				 self AddCq: numArgs + 1 * objectMemory wordSize R: SPReg.
				 self MoveR: TempReg Mw: 0 r: SPReg "Overwrite pushed receiver; ReceiverResultReg already contains receiver."].
		 ^self].

	"e.g.	Receiver				Receiver
			Selector/Arg0	=>		Arg1
			Arg1					Arg2
	 		Arg2			sp->	Arg3
	 sp->	Arg3"
	super adjustArgumentsForPerform: numArgs! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 12/10/2015 09:10'!
allocateEqualsEqualsRegistersArgNeedsReg: argNeedsReg rcvrNeedsReg: rcvrNeedsReg into: binaryBlock
	<inline: true>
	| argReg rcvrReg |
	self assert: (argNeedsReg or: [rcvrNeedsReg]).
	argReg := rcvrReg := NoReg.
	argNeedsReg
		ifTrue: 
			[rcvrNeedsReg
				ifTrue:
					[self allocateRegForStackTopTwoEntriesInto: [:rTop :rNext| argReg := rTop. rcvrReg := rNext].
					 self ssTop popToReg: argReg.
					 (self ssValue: 1) popToReg: rcvrReg]
				ifFalse:
					[argReg := self allocateRegForStackEntryAt: 0.
					 self ssTop popToReg: argReg.
					 "If the receiver is a spilled constant we need to pop it from the stack."
					 (self ssValue: 1) spilled ifTrue:
						[self AddCq: objectMemory wordSize R: SPReg]]]
		ifFalse:
			[self assert: rcvrNeedsReg.
			 self deny: self ssTop spilled.
			 rcvrReg := self allocateRegForStackEntryAt: 1.
			 (self ssValue: 1) popToReg: rcvrReg].
		
	self deny: (argNeedsReg and: [argReg = NoReg]).
	self deny: (rcvrNeedsReg and: [rcvrReg = NoReg]).

	binaryBlock value: rcvrReg value: argReg! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 3/6/2017 01:13'!
allocateFloatRegNotConflictingWith: regMask
	| reg |
	"if there's a free register, use it"
	reg := backEnd availableFloatRegisterOrNoneFor: (self liveFloatRegisters bitOr: regMask).
	reg = NoReg ifTrue: "No free register, choose one that does not conflict with regMask"
		[reg := self freeAnyFloatRegNotConflictingWith: regMask].
	^ reg! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'cb 4/26/2015 12:25'!
allocateRegForStackEntryAt: index
	"If the stack entry is already in a register, answers it,
	else allocate a new register for it"
	<inline: true>
	^ self allocateRegForStackEntryAt: index notConflictingWith: 0
	! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'cb 3/26/2017 19:30'!
allocateRegForStackEntryAt: index notConflictingWith: regMask
	"If the stack entry is already in a register not conflicting with regMask, answers it,
	else allocate a new register not conflicting with reg mask"
	<var: #stackEntry type: #'CogSimStackEntry *'>
	| stackEntry mask |
	stackEntry := self ssValue: index.
	mask := stackEntry registerMaskOrNone.
	(mask ~= 0 and: [mask noMask: regMask]) ifTrue: 
		[self flag: #TODO.
		"When one does pushDup on a SSRegister
		 followed by an operation reusing the register 
		 but mutating the value of the register, then the value that was 
		 dup is now refering to the register with a mutated value which
		 is incorrect. This problem is serious as it can happen also with 
		 multiple unsafe instructions (no dup involved)
		 This is not a problem if the operation reusing the register is not 
		 mutating the value of if all the dup values are used in the same 
		 operation.
		 => I think we should introduce allocateRegMutatingStackEntryAt:
		notConflictingWith:upThrough:, used by operations mutating the reg
		value and flushing partially the stack if somewhere between simSpill
		and the ssEntries used by the operation someone uses also the same
		register."
		 ^stackEntry registerOrNone].
	^self allocateRegNotConflictingWith: regMask! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 12/02/2016 16:00'!
allocateRegForStackTopThreeEntriesInto: trinaryBlock thirdIsReceiver: thirdIsReceiver
	"Answers registers for the 3 top values on stack. If the values are already in registers, answers
	these registers, else allocate registers not conflicting with each others.
	If thirdIsReceiver is true, allocate ReceiverResultReg for stackTop - 2 (for ceStoreCheck)."
	<inline: true>
	| topRegistersMask rTop rNext rThird |
	
	topRegistersMask := 0.
	rTop := rNext := rThird := NoReg.
	
	(self ssTop registerOrNone ~= NoReg and: [ thirdIsReceiver not or: [ self ssTop registerOrNone ~= ReceiverResultReg ] ]) ifTrue: 
		[ topRegistersMask := self registerMaskFor: (rTop := self ssTop registerOrNone)].
	((self ssValue: 1) registerOrNone ~= NoReg and: [ thirdIsReceiver not or: [ (self ssValue: 1) registerOrNone ~= ReceiverResultReg ] ]) ifTrue: 
		[ topRegistersMask := topRegistersMask bitOr: (self registerMaskFor: (rNext := (self ssValue: 1) registerOrNone))].
	((self ssValue: 2) registerOrNone ~= NoReg and: [thirdIsReceiver not or: [ (self ssValue: 2) registerOrNone = ReceiverResultReg ] ]) ifTrue: 
		[ topRegistersMask := topRegistersMask bitOr: (self registerMaskFor: (rThird := (self ssValue: 2) registerOrNone))].
	
	rThird = NoReg ifTrue:
		[ thirdIsReceiver 
			ifTrue:
				[ rThird := ReceiverResultReg.  "Free ReceiverResultReg if it was not free"
				self ssAllocateRequiredReg: ReceiverResultReg.
				self voidReceiverResultRegContainsSelf ]
			ifFalse: [ rThird := self allocateRegNotConflictingWith: topRegistersMask ].
		topRegistersMask := topRegistersMask bitOr: (self registerMaskFor: rThird) ].
	
	rTop = NoReg ifTrue:
		[ rTop := self allocateRegNotConflictingWith: topRegistersMask.
		  topRegistersMask := topRegistersMask bitOr: (self registerMaskFor: rTop) ].
	
	rNext = NoReg ifTrue:
		[ rNext := self allocateRegNotConflictingWith: topRegistersMask ].

	self deny: (rTop = NoReg or: [rNext = NoReg or: [rThird = NoReg]]).

	^ trinaryBlock value: rTop value: rNext value: rThird! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 11/28/2016 16:21'!
allocateRegForStackTopTwoEntriesInto: binaryBlock
	"Answers registers for the 2 top values on stack. If the values are already in registers, answers
	these registers, else allocate registers not conflicting with each others."
	<inline: true>
	| topRegistersMask rTop rNext |
	
	topRegistersMask := 0.
	rTop := rNext := NoReg.
	
	self ssTop  registerOrNone ~= NoReg ifTrue: 
		[ rTop := self ssTop registerOrNone].
	(self ssValue: 1)  registerOrNone ~= NoReg ifTrue: 
		[ topRegistersMask := self registerMaskFor: (rNext := (self ssValue: 1) registerOrNone)].
	
	rTop = NoReg ifTrue:
		[ rTop := self allocateRegNotConflictingWith: topRegistersMask ].
	
	rNext = NoReg ifTrue:
		[ rNext := self allocateRegNotConflictingWith: (self registerMaskFor: rTop) ].

	self deny: (rTop = NoReg or: [rNext = NoReg]).

	^ binaryBlock value: rTop value: rNext! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 12/02/2016 16:00'!
allocateRegNotConflictingWith: regMask
	| reg |
	"if there's a free register, use it"
	reg := backEnd availableRegisterOrNoneFor: (self liveRegisters bitOr: regMask).
	reg = NoReg ifTrue: "No free register, choose one that does not conflict with regMask"
		[reg := self freeAnyRegNotConflictingWith: regMask].
	reg = ReceiverResultReg ifTrue: "If we've allocated RcvrResultReg, it's not live anymore"
		[self voidReceiverResultRegContainsSelf].
	^ reg! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 02:17'!
allocateRegistersForLowcodeFloat2: aBlock
	<option: #LowcodeVM>
	<inline: true>

	| rNext rTop topRegistersMask |
	topRegistersMask := 0.
	rTop := rNext := NoReg.

	self ssNativeTop  nativeFloatRegisterOrNone ~= NoReg ifTrue: 
		[ rTop := self ssNativeTop nativeFloatRegisterOrNone].
	(self ssNativeValue: 1)  nativeFloatRegisterOrNone ~= NoReg ifTrue: 
		[ topRegistersMask := self registerMaskFor: (rNext := (self ssNativeValue: 1) nativeFloatRegisterOrNone)].
	
	rTop = NoReg ifTrue:
		[ rTop := self allocateFloatRegNotConflictingWith: topRegistersMask ].
	
	rNext = NoReg ifTrue:
		[ rNext := self allocateFloatRegNotConflictingWith: (self registerMaskFor: rTop) ].
	self deny: (rTop = NoReg or: [rNext = NoReg]).

	^ aBlock value: rTop value: rNext
! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/21/2017 18:10'!
allocateRegistersForLowcodeFloat2ResultInteger: aBlock
	<option: #LowcodeVM>
	<inline: true>

	| frNext frTop topRegistersMask rResult |
	topRegistersMask := 0.
	frTop := frNext := NoReg.
	rResult := NoReg.

	self ssNativeTop  nativeFloatRegisterOrNone ~= NoReg ifTrue: 
		[ frTop := self ssNativeTop nativeFloatRegisterOrNone].
	(self ssNativeValue: 1)  nativeFloatRegisterOrNone ~= NoReg ifTrue: 
		[ topRegistersMask := self registerMaskFor: (frNext := (self ssNativeValue: 1) nativeFloatRegisterOrNone)].
	
	frTop = NoReg ifTrue:
		[ frTop := self allocateFloatRegNotConflictingWith: topRegistersMask ].
	
	frNext = NoReg ifTrue:
		[ frNext := self allocateFloatRegNotConflictingWith: (self registerMaskFor: frTop) ].
	
	"Result"
	rResult := self allocateRegNotConflictingWith: 0.

	self deny: (frTop = NoReg or: [frNext = NoReg or: [rResult = NoReg]]).

	^ aBlock value: frTop value: frNext value: rResult
! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 02:19'!
allocateRegistersForLowcodeFloat: aBlock
	<option: #LowcodeVM>
	<inline: true>

	| frTop topRegistersMask |
	topRegistersMask := 0.
	frTop := NoReg.

	self ssNativeTop nativeFloatRegisterOrNone ~= NoReg ifTrue: 
		[ frTop := self ssNativeTop nativeFloatRegisterOrNone].
	
	frTop = NoReg ifTrue:
		[ frTop := self allocateFloatRegNotConflictingWith: topRegistersMask ].
	self deny: (frTop = NoReg).

	^ aBlock value: frTop
! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 05:24'!
allocateRegistersForLowcodeFloatInteger: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	"Used by storeFloat32ToMemory and storeFloat64ToMemory"
	| frTop rTop |
	frTop := rTop := NoReg.

	"Integer registers"
	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue: 
		[ rTop := self ssNativeTop nativeRegisterOrNone].

	rTop = NoReg ifTrue:
		[ rTop := self allocateRegNotConflictingWith: 0 ].

	"Float registers"
	(self ssNativeValue: 1) nativeFloatRegisterOrNone ~= NoReg ifTrue: 
		[ frTop := (self ssNativeValue: 1) nativeFloatRegisterOrNone].
	
	frTop = NoReg ifTrue:
		[ frTop := self allocateFloatRegNotConflictingWith: 0 ].
	
	self deny: (frTop = NoReg or: [rTop = NoReg]).

	^ aBlock value: frTop value: rTop! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 02:18'!
allocateRegistersForLowcodeFloatResultFloat: aBlock
	<option: #LowcodeVM>
	<inline: true>

	| frTop frResult |
	frTop := NoReg.
	frResult := NoReg.

	"Float argument"
	self ssNativeTop nativeFloatRegisterOrNone ~= NoReg ifTrue: 
		[ frTop := self ssNativeTop nativeFloatRegisterOrNone].
	
	frTop = NoReg ifTrue:
		[ frTop := self allocateFloatRegNotConflictingWith: self emptyRegisterMask ].
	
	"Float result".
	frResult := self allocateFloatRegNotConflictingWith: (self registerMaskFor: frTop).
	self deny: (frTop = NoReg or: [frResult = NoReg]).

	^ aBlock value: frTop value: frResult
! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 02:18'!
allocateRegistersForLowcodeFloatResultInteger2: aBlock
	<option: #LowcodeVM>
	<inline: true>

	| frTop rResult rResult2 |
	frTop := NoReg.
	rResult2 := rResult := NoReg.

	"Float argument"
	self ssNativeTop nativeFloatRegisterOrNone ~= NoReg ifTrue: 
		[ frTop := self ssNativeTop nativeFloatRegisterOrNone].
	
	frTop = NoReg ifTrue:
		[ frTop := self allocateFloatRegNotConflictingWith: self emptyRegisterMask ].
	
	"Integer result".
	rResult := self allocateRegNotConflictingWith: (self emptyRegisterMask).
	rResult2 := self allocateRegNotConflictingWith: (self registerMaskFor: rResult).
	
	self deny: (frTop = NoReg or: [rResult = NoReg or: [rResult2 = NoReg]]).

	^ aBlock value: frTop value: rResult value: rResult2

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 02:18'!
allocateRegistersForLowcodeFloatResultInteger: aBlock
	<option: #LowcodeVM>
	<inline: true>

	| frTop rResult |
	frTop := NoReg.
	rResult := NoReg.

	"Float argument"
	self ssNativeTop nativeFloatRegisterOrNone ~= NoReg ifTrue: 
		[ frTop := self ssNativeTop nativeFloatRegisterOrNone].
	
	frTop = NoReg ifTrue:
		[ frTop := self allocateFloatRegNotConflictingWith: self emptyRegisterMask ].
	
	"Integer result".
	rResult := self allocateRegNotConflictingWith: (self emptyRegisterMask).
	
	self deny: (frTop = NoReg or: [rResult = NoReg]).

	^ aBlock value: frTop value: rResult

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 02:18'!
allocateRegistersForLowcodeFloatResultOop: aBlock
	<option: #LowcodeVM>
	<inline: true>

	| frTop rResult |
	frTop := NoReg.
	rResult := NoReg.

	"Float argument"
	self ssNativeTop  nativeFloatRegisterOrNone ~= NoReg ifTrue: 
		[ frTop := self ssNativeTop nativeFloatRegisterOrNone].
	
	frTop = NoReg ifTrue:
		[ frTop := self allocateFloatRegNotConflictingWith: self emptyRegisterMask ].
	
	"Integer result".
	rResult := self allocateRegNotConflictingWith: (self emptyRegisterMask).
	
	self deny: (frTop = NoReg or: [rResult = NoReg]).

	^ aBlock value: frTop value: rResult

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 02:26'!
allocateRegistersForLowcodeInteger2: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rNext rTop topRegistersMask |
	topRegistersMask := 0.
	rTop := rNext := NoReg.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
		self ssNativeTop nativeRegisterSecondOrNone ~= NoReg ifTrue:
			[ rNext := self ssNativeTop nativeRegisterSecondOrNone]
	].

	rNext = NoReg ifTrue: [
		(self ssNativeValue: 1)  nativeRegisterOrNone ~= NoReg ifTrue: 
			[ topRegistersMask := self registerMaskFor: (rNext := (self ssNativeValue: 1) nativeRegisterOrNone)].
	].

	rTop = NoReg ifTrue:
		[ rTop := self allocateRegNotConflictingWith: topRegistersMask ].
	
	rNext = NoReg ifTrue:
		[ rNext := self allocateRegNotConflictingWith: (self registerMaskFor: rTop) ].

	self deny: (rTop = NoReg or: [rNext = NoReg]).
	^ aBlock value: rTop value: rNext
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/21/2017 16:56'!
allocateRegistersForLowcodeInteger2ResultFloat: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rNext rTop topRegistersMask frResult |
	topRegistersMask := 0.
	rTop := rNext := NoReg.
	frResult := NoReg.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
		self ssNativeTop nativeRegisterSecondOrNone ~= NoReg ifTrue:
			[ rNext := self ssNativeTop nativeRegisterSecondOrNone]
	].

	rNext = NoReg ifTrue: [
		(self ssNativeValue: 1)  nativeRegisterOrNone ~= NoReg ifTrue: 
			[ topRegistersMask := self registerMaskFor: (rNext := (self ssNativeValue: 1) nativeRegisterOrNone)].
	].

	rTop = NoReg ifTrue:
		[ rTop := self allocateRegNotConflictingWith: topRegistersMask ].
	
	rNext = NoReg ifTrue:
		[ rNext := self allocateRegNotConflictingWith: (self registerMaskFor: rTop) ].

	"Results"
	frResult := self allocateFloatRegNotConflictingWith: self emptyRegisterMask.
	
	self deny: (rTop = NoReg or: [rNext = NoReg or: [frResult = NoReg]]).
	
	^ aBlock value: rTop value: rNext value: frResult! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/21/2017 16:56'!
allocateRegistersForLowcodeInteger2ResultInteger2: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rNext rTop topRegistersMask rResult rResult2 |
	topRegistersMask := 0.
	rTop := rNext := NoReg.
	rResult := rResult2 := NoReg.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
		self ssNativeTop nativeRegisterSecondOrNone ~= NoReg ifTrue:
			[ rNext := self ssNativeTop nativeRegisterSecondOrNone]
	].

	rNext = NoReg ifTrue: [
		(self ssNativeValue: 1)  nativeRegisterOrNone ~= NoReg ifTrue: 
			[ topRegistersMask := self registerMaskFor: (rNext := (self ssNativeValue: 1) nativeRegisterOrNone)].
	].

	rTop = NoReg ifTrue:
		[ rTop := self allocateRegNotConflictingWith: topRegistersMask ].
	
	rNext = NoReg ifTrue:
		[ rNext := self allocateRegNotConflictingWith: (self registerMaskFor: rTop) ].

	self deny: (rTop = NoReg or: [rNext = NoReg]).

	"Results"
	rResult := self allocateFloatRegNotConflictingWith: (self registerMaskFor: rTop and: rNext).
	rResult2 := self allocateFloatRegNotConflictingWith: (self registerMaskFor: rTop and: rNext and: rResult).
	self deny: (rResult = NoReg or: [rResult2 = NoReg]).
	
	^ aBlock value: rTop value: rNext value: rResult value: rResult2! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/21/2017 16:56'!
allocateRegistersForLowcodeInteger2ResultInteger: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rNext rTop topRegistersMask rResult |
	topRegistersMask := 0.
	rTop := rNext := NoReg.
	rResult := NoReg.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
		self ssNativeTop nativeRegisterSecondOrNone ~= NoReg ifTrue:
			[ rNext := self ssNativeTop nativeRegisterSecondOrNone]
	].

	rNext = NoReg ifTrue: [
		(self ssNativeValue: 1)  nativeRegisterOrNone ~= NoReg ifTrue: 
			[ topRegistersMask := self registerMaskFor: (rNext := (self ssNativeValue: 1) nativeRegisterOrNone)].
	].

	rTop = NoReg ifTrue:
		[ rTop := self allocateRegNotConflictingWith: topRegistersMask ].
	
	rNext = NoReg ifTrue:
		[ rNext := self allocateRegNotConflictingWith: (self registerMaskFor: rTop) ].

	self deny: (rTop = NoReg or: [rNext = NoReg]).

	"Results"
	rResult := self allocateFloatRegNotConflictingWith: (self registerMaskFor: rTop and: rNext).
	self deny: (rResult = NoReg).
	
	^ aBlock value: rTop value: rNext value: rResult! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/21/2017 16:56'!
allocateRegistersForLowcodeInteger2ResultOop: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rNext rTop topRegistersMask rResult |
	topRegistersMask := 0.
	rTop := rNext := NoReg.
	rResult := NoReg.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
		self ssNativeTop nativeRegisterSecondOrNone ~= NoReg ifTrue:
			[ rNext := self ssNativeTop nativeRegisterSecondOrNone]
	].

	rNext = NoReg ifTrue: [
		(self ssNativeValue: 1)  nativeRegisterOrNone ~= NoReg ifTrue: 
			[ topRegistersMask := self registerMaskFor: (rNext := (self ssNativeValue: 1) nativeRegisterOrNone)].
	].

	rTop = NoReg ifTrue:
		[ rTop := self allocateRegNotConflictingWith: topRegistersMask ].
	
	rNext = NoReg ifTrue:
		[ rNext := self allocateRegNotConflictingWith: (self registerMaskFor: rTop) ].

	self deny: (rTop = NoReg or: [rNext = NoReg]).

	"Results"
	rResult := self allocateFloatRegNotConflictingWith: (self registerMaskFor: rTop and: rNext).
	self deny: (rResult = NoReg).
	
	^ aBlock value: rTop value: rNext value: rResult! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/21/2017 17:01'!
allocateRegistersForLowcodeInteger3: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rNext rNextNext nativeValueIndex rTop nextRegisterMask |
	rTop := rNext := rNextNext := NoReg.
	nativeValueIndex := 1.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
		self ssNativeTop nativeRegisterSecondOrNone ~= NoReg ifTrue:
			[ rNext := self ssNativeTop nativeRegisterSecondOrNone]
	].

	rNext = NoReg ifTrue: [
		(self ssNativeValue: nativeValueIndex)  nativeRegisterOrNone ~= NoReg ifTrue:  [
			rNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
			(self ssNativeValue: nativeValueIndex) nativeRegisterSecondOrNone ~= NoReg ifTrue: [
				rNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterSecondOrNone.
			].
			nativeValueIndex := nativeValueIndex + 1
		].
	].

	rNextNext = NoReg ifTrue: [
		(self ssNativeValue: nativeValueIndex)  nativeRegisterOrNone ~= NoReg ifTrue:  [
			rNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
		].
	].

	rTop = NoReg ifTrue: [
		nextRegisterMask := 0.
		rNext ~= NoReg ifTrue: [ nextRegisterMask := self registerMaskFor: rNext].
		rNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNext)].
		rTop := self allocateRegNotConflictingWith: nextRegisterMask
	].
	
	rNext = NoReg ifTrue: [
		nextRegisterMask := self registerMaskFor: rTop.
		rNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNext)].
		rNext := self allocateRegNotConflictingWith: nextRegisterMask
	].

	rNextNext = NoReg ifTrue: [
		nextRegisterMask := self registerMaskFor: rTop and: rNext.
		rNextNext := self allocateRegNotConflictingWith: nextRegisterMask
	].

	self deny: (rTop = NoReg or: [rNext = NoReg or: [rNextNext = NoReg]]).
	^ aBlock value: rTop value: rNext value: rNextNext
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/21/2017 17:00'!
allocateRegistersForLowcodeInteger3ResultInteger: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rNext rNextNext rResult nativeValueIndex rTop nextRegisterMask |
	rTop := rNext := rNextNext := NoReg.
	nativeValueIndex := 1.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
		self ssNativeTop nativeRegisterSecondOrNone ~= NoReg ifTrue:
			[ rNext := self ssNativeTop nativeRegisterSecondOrNone]
	].

	rNext = NoReg ifTrue: [
		(self ssNativeValue: nativeValueIndex)  nativeRegisterOrNone ~= NoReg ifTrue:  [
			rNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
			(self ssNativeValue: nativeValueIndex) nativeRegisterSecondOrNone ~= NoReg ifTrue: [
				rNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterSecondOrNone.
			].
			nativeValueIndex := nativeValueIndex + 1
		].
	].

	rNextNext = NoReg ifTrue: [
		(self ssNativeValue: nativeValueIndex)  nativeRegisterOrNone ~= NoReg ifTrue:  [
			rNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
		].
	].

	rTop = NoReg ifTrue: [
		nextRegisterMask := 0.
		rNext ~= NoReg ifTrue: [ nextRegisterMask := self registerMaskFor: rNext].
		rNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNext)].
		rTop := self allocateRegNotConflictingWith: nextRegisterMask
	].
	
	rNext = NoReg ifTrue: [
		nextRegisterMask := self registerMaskFor: rTop.
		rNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNext)].
		rNext := self allocateRegNotConflictingWith: nextRegisterMask
	].

	rNextNext = NoReg ifTrue: [
		nextRegisterMask := self registerMaskFor: rTop and: rNext.
		rNextNext := self allocateRegNotConflictingWith: nextRegisterMask
	].

	self deny: (rTop = NoReg or: [rNext = NoReg or: [rNextNext = NoReg]]).
	
	"Allocate the result"
	rResult := self allocateRegNotConflictingWith: (self registerMaskFor: rTop and: rNext and: rNextNext).
	self deny: (rResult = NoReg).
	
	^ aBlock value: rTop value: rNext value: rNextNext value: rResult
! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/21/2017 16:59'!
allocateRegistersForLowcodeInteger4: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rNext rNextNext rNextNextNext nativeValueIndex rTop nextRegisterMask |
	rTop := rNext := rNextNext := rNextNextNext := NoReg.
	nativeValueIndex := 1.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
		self ssNativeTop nativeRegisterSecondOrNone ~= NoReg ifTrue:
			[ rNext := self ssNativeTop nativeRegisterSecondOrNone]
	].

	rNext = NoReg ifTrue: [
		(self ssNativeValue: nativeValueIndex)  nativeRegisterOrNone ~= NoReg ifTrue:  [
			rNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
			(self ssNativeValue: nativeValueIndex) nativeRegisterSecondOrNone ~= NoReg ifTrue: [
				rNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
			].
			nativeValueIndex := nativeValueIndex + 1
		].
	].

	rNextNext = NoReg ifTrue: [
		(self ssNativeValue: nativeValueIndex)  nativeRegisterOrNone ~= NoReg ifTrue:  [
			rNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
			(self ssNativeValue: nativeValueIndex) nativeRegisterSecondOrNone ~= NoReg ifTrue: [
				rNextNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterSecondOrNone.
			].
			nativeValueIndex := nativeValueIndex + 1
		].
	].

	rNextNextNext = NoReg ifTrue: [
		(self ssNativeValue: nativeValueIndex)  nativeRegisterOrNone ~= NoReg ifTrue:  [
			rNextNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
			nativeValueIndex := nativeValueIndex + 1
		].
	].

	rTop = NoReg ifTrue: [
		nextRegisterMask := 0.
		rNext ~= NoReg ifTrue: [ nextRegisterMask := self registerMaskFor: rNext].
		rNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNext)].
		rNextNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNextNext)].
		rTop := self allocateRegNotConflictingWith: nextRegisterMask
	].
	
	rNext = NoReg ifTrue: [
		nextRegisterMask := self registerMaskFor: rTop.
		rNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNext)].
		rNextNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNextNext)].
		rNext := self allocateRegNotConflictingWith: nextRegisterMask
	].

	rNextNext = NoReg ifTrue: [
		nextRegisterMask := self registerMaskFor: rTop and: rNext.
		rNextNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNextNext)].
		rNextNext := self allocateRegNotConflictingWith: nextRegisterMask
	].

	rNextNextNext = NoReg ifTrue: [
		nextRegisterMask := self registerMaskFor: rTop and: rNext and: rNextNext.
		rNextNextNext := self allocateRegNotConflictingWith: nextRegisterMask
	].

	self deny: (rTop = NoReg or: [rNext = NoReg or: [rNextNext = NoReg or: [rNextNextNext = NoReg]]]).
	^ aBlock value: rTop value: rNext value: rNextNext value: rNextNextNext
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/21/2017 16:59'!
allocateRegistersForLowcodeInteger4ResultInteger2: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rNext rNextNext rNextNextNext nativeValueIndex rTop nextRegisterMask rResult rResult2 |
	rTop := rNext := rNextNext := rNextNextNext := NoReg.
	rResult := rResult2 := NoReg.
	nativeValueIndex := 1.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
		self ssNativeTop nativeRegisterSecondOrNone ~= NoReg ifTrue:
			[ rNext := self ssNativeTop nativeRegisterSecondOrNone]
	].

	rNext = NoReg ifTrue: [
		(self ssNativeValue: nativeValueIndex)  nativeRegisterOrNone ~= NoReg ifTrue:  [
			rNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
			(self ssNativeValue: nativeValueIndex) nativeRegisterSecondOrNone ~= NoReg ifTrue: [
				rNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
			].
			nativeValueIndex := nativeValueIndex + 1
		].
	].

	rNextNext = NoReg ifTrue: [
		(self ssNativeValue: nativeValueIndex)  nativeRegisterOrNone ~= NoReg ifTrue:  [
			rNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
			(self ssNativeValue: nativeValueIndex) nativeRegisterSecondOrNone ~= NoReg ifTrue: [
				rNextNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterSecondOrNone.
			].
			nativeValueIndex := nativeValueIndex + 1
		].
	].

	rNextNextNext = NoReg ifTrue: [
		(self ssNativeValue: nativeValueIndex)  nativeRegisterOrNone ~= NoReg ifTrue:  [
			rNextNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
			nativeValueIndex := nativeValueIndex + 1
		].
	].

	rTop = NoReg ifTrue: [
		nextRegisterMask := 0.
		rNext ~= NoReg ifTrue: [ nextRegisterMask := self registerMaskFor: rNext].
		rNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNext)].
		rNextNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNextNext)].
		rTop := self allocateRegNotConflictingWith: nextRegisterMask
	].
	
	rNext = NoReg ifTrue: [
		nextRegisterMask := self registerMaskFor: rTop.
		rNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNext)].
		rNextNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNextNext)].
		rNext := self allocateRegNotConflictingWith: nextRegisterMask
	].

	rNextNext = NoReg ifTrue: [
		nextRegisterMask := self registerMaskFor: rTop and: rNext.
		rNextNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNextNext)].
		rNextNext := self allocateRegNotConflictingWith: nextRegisterMask
	].

	rNextNextNext = NoReg ifTrue: [
		nextRegisterMask := self registerMaskFor: rTop and: rNext and: rNextNext.
		rNextNextNext := self allocateRegNotConflictingWith: nextRegisterMask
	].

	self deny: (rTop = NoReg or: [rNext = NoReg or: [rNextNext = NoReg or: [rNextNextNext = NoReg]]]).
	
	"Result registers"
	rResult := self allocateRegNotConflictingWith: (self registerMaskFor: rTop and: rNext and: rNextNext and: rNextNextNext).
	rResult2 := self allocateRegNotConflictingWith: (self registerMaskFor: rTop and: rNext and: rNextNext and: rNextNextNext and: rResult).
	
	^ aBlock value: rTop value: rNext value: rNextNext value: rNextNextNext value: rResult value: rResult2
! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/21/2017 17:00'!
allocateRegistersForLowcodeInteger4ResultInteger: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rNext rNextNext rNextNextNext nativeValueIndex rTop nextRegisterMask rResult |
	rTop := rNext := rNextNext := rNextNextNext := NoReg.
	rResult := NoReg.
	nativeValueIndex := 1.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
		self ssNativeTop nativeRegisterSecondOrNone ~= NoReg ifTrue:
			[ rNext := self ssNativeTop nativeRegisterSecondOrNone]
	].

	rNext = NoReg ifTrue: [
		(self ssNativeValue: nativeValueIndex)  nativeRegisterOrNone ~= NoReg ifTrue:  [
			rNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
			(self ssNativeValue: nativeValueIndex) nativeRegisterSecondOrNone ~= NoReg ifTrue: [
				rNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
			].
			nativeValueIndex := nativeValueIndex + 1
		].
	].

	rNextNext = NoReg ifTrue: [
		(self ssNativeValue: nativeValueIndex)  nativeRegisterOrNone ~= NoReg ifTrue:  [
			rNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
			(self ssNativeValue: nativeValueIndex) nativeRegisterSecondOrNone ~= NoReg ifTrue: [
				rNextNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterSecondOrNone.
			].
			nativeValueIndex := nativeValueIndex + 1
		].
	].

	rNextNextNext = NoReg ifTrue: [
		(self ssNativeValue: nativeValueIndex)  nativeRegisterOrNone ~= NoReg ifTrue:  [
			rNextNextNext := (self ssNativeValue: nativeValueIndex) nativeRegisterOrNone.
			nativeValueIndex := nativeValueIndex + 1
		].
	].

	rTop = NoReg ifTrue: [
		nextRegisterMask := 0.
		rNext ~= NoReg ifTrue: [ nextRegisterMask := self registerMaskFor: rNext].
		rNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNext)].
		rNextNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNextNext)].
		rTop := self allocateRegNotConflictingWith: nextRegisterMask
	].
	
	rNext = NoReg ifTrue: [
		nextRegisterMask := self registerMaskFor: rTop.
		rNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNext)].
		rNextNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNextNext)].
		rNext := self allocateRegNotConflictingWith: nextRegisterMask
	].

	rNextNext = NoReg ifTrue: [
		nextRegisterMask := self registerMaskFor: rTop and: rNext.
		rNextNextNext ~= NoReg ifTrue: [ nextRegisterMask := nextRegisterMask bitOr: (self registerMaskFor: rNextNextNext)].
		rNextNext := self allocateRegNotConflictingWith: nextRegisterMask
	].

	rNextNextNext = NoReg ifTrue: [
		nextRegisterMask := self registerMaskFor: rTop and: rNext and: rNextNext.
		rNextNextNext := self allocateRegNotConflictingWith: nextRegisterMask
	].

	self deny: (rTop = NoReg or: [rNext = NoReg or: [rNextNext = NoReg or: [rNextNextNext = NoReg]]]).
	
	"Result registers"
	rResult := self allocateRegNotConflictingWith: (self registerMaskFor: rTop and: rNext and: rNextNext and: rNextNextNext).
	
	^ aBlock value: rTop value: rNext value: rNextNext value: rNextNextNext value: rResult
! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 02:12'!
allocateRegistersForLowcodeInteger7ResultInteger: aBlock
	<option: #LowcodeVM>
	<inline: true>
	"Effective address 64. This instruction should not be generated on 32-bits mode"
	self abort
! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 04:13'!
allocateRegistersForLowcodeInteger: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rTop |
	rTop := NoReg.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
	].

	rTop = NoReg ifTrue:
		[ rTop := self allocateRegNotConflictingWith: self emptyRegisterMask].

	self deny: (rTop = NoReg).
	^ aBlock value: rTop
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 05:20'!
allocateRegistersForLowcodeIntegerOop2: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rTop rOopTop rOopNext topRegisterMask oopTopRegisterMask |
	rTop := rOopTop := rOopNext := NoReg.
	oopTopRegisterMask := topRegisterMask := 0.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
		oopTopRegisterMask := self registerMaskFor: rTop.
	].

	self ssTop registerOrNone ~= NoReg ifTrue:  [
		rOopTop := self ssTop registerOrNone.
		topRegisterMask := self registerMaskFor: rOopTop.
	].

	(self ssValue: 1) registerOrNone ~= NoReg ifTrue: [
		rOopNext := (self ssValue: 1) registerOrNone.
		topRegisterMask := topRegisterMask bitOr: (self registerMaskFor: rOopNext).
		oopTopRegisterMask := oopTopRegisterMask bitOr: (self registerMaskFor: rOopNext).
	].

	rTop = NoReg ifTrue: [
		rTop := self allocateRegNotConflictingWith: topRegisterMask
	].

	rOopTop = NoReg ifTrue: [
		rOopTop := self allocateRegNotConflictingWith: oopTopRegisterMask
	].

	rOopNext = NoReg ifTrue: [
		rOopNext := self allocateRegNotConflictingWith: (self registerMaskFor: rTop and: rOopTop)
	].

	self deny: (rTop = NoReg or: [rOopTop = NoReg or: [rOopNext = NoReg]]).
	^ aBlock value: rTop value: rOopTop value: rOopNext
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 05:20'!
allocateRegistersForLowcodeIntegerOop: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rTop rOopTop topRegisterMask |
	rTop := rOopTop := NoReg.
	topRegisterMask := 0.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
	].

	self ssTop registerOrNone ~= NoReg ifTrue:  [
		rOopTop := self ssTop registerOrNone.
		topRegisterMask := self registerMaskFor: rOopTop.
	].

	rTop = NoReg ifTrue: [
		rTop := self allocateRegNotConflictingWith: topRegisterMask
	].

	rOopTop = NoReg ifTrue: [
		rOopTop := self allocateRegNotConflictingWith: (self registerMaskFor: rTop)
	].

	self deny: (rTop = NoReg or: [rOopTop = NoReg]).
	^ aBlock value: rTop value: rOopTop
! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 05:20'!
allocateRegistersForLowcodeIntegerOopResultOop: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rTop rOopTop rOopResult topRegisterMask |
	rTop := rOopTop:= NoReg.
	rOopResult := NoReg.
	topRegisterMask := 0.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
	].

	self ssTop registerOrNone ~= NoReg ifTrue:  [
		rOopTop := self ssTop registerOrNone.
		topRegisterMask := self registerMaskFor: rOopTop.
	].

	rTop = NoReg ifTrue: [
		rTop := self allocateRegNotConflictingWith: topRegisterMask
	].

	rOopTop = NoReg ifTrue: [
		rOopTop := self allocateRegNotConflictingWith: (self registerMaskFor: rTop)
	].

	rOopResult := self allocateRegNotConflictingWith: (self registerMaskFor: rTop and: rOopTop).

	self deny: (rTop = NoReg or: [rOopTop = NoReg or: [rOopResult = NoReg]]).
	
	^ aBlock value: rTop value: rOopTop value: rOopResult
! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/21/2017 16:56'!
allocateRegistersForLowcodeIntegerResultFloat: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rTop frResult|
	rTop := NoReg.
	frResult := NoReg.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
	].

	rTop = NoReg ifTrue:
		[ rTop := self allocateRegNotConflictingWith: self emptyRegisterMask].
	
	frResult := self allocateFloatRegNotConflictingWith: self emptyRegisterMask.

	self deny: (rTop = NoReg or: [frResult = NoReg]).
	^ aBlock value: rTop value: frResult
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/21/2017 16:57'!
allocateRegistersForLowcodeIntegerResultInteger2: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rTop rResult rResult2 |
	rTop := NoReg.
	rResult := rResult := NoReg.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
	].

	rTop = NoReg ifTrue:
		[ rTop := self allocateRegNotConflictingWith: self emptyRegisterMask].
	
	rResult := self allocateRegNotConflictingWith: (self registerMaskFor: rTop).
	rResult2 := self allocateRegNotConflictingWith: (self registerMaskFor: rTop and: rResult).

	self deny: (rTop = NoReg or: [rResult = NoReg or: [rResult2 = NoReg]]).
	^ aBlock value: rTop value: rResult value: rResult2
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/21/2017 16:57'!
allocateRegistersForLowcodeIntegerResultInteger: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rTop rResult|
	rTop := NoReg.
	rResult := NoReg.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
	].

	rTop = NoReg ifTrue:
		[ rTop := self allocateRegNotConflictingWith: self emptyRegisterMask].
	
	rResult := self allocateRegNotConflictingWith: (self registerMaskFor: rTop).

	self deny: (rTop = NoReg or: [rResult = NoReg]).
	^ aBlock value: rTop value: rResult
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/21/2017 16:57'!
allocateRegistersForLowcodeIntegerResultOop: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rTop rResult|
	rTop := NoReg.
	rResult := NoReg.

	self ssNativeTop nativeRegisterOrNone ~= NoReg ifTrue:  [
		rTop := self ssNativeTop nativeRegisterOrNone.
	].

	rTop = NoReg ifTrue:
		[ rTop := self allocateRegNotConflictingWith: self emptyRegisterMask].
	
	rResult := self allocateRegNotConflictingWith: (self registerMaskFor: rTop).

	self deny: (rTop = NoReg or: [rResult = NoReg]).
	^ aBlock value: rTop value: rResult
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/26/2017 17:15'!
allocateRegistersForLowcodeOop2: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rOopTop rOopNext oopTopRegisterMask |
	rOopTop := rOopNext := NoReg.
	oopTopRegisterMask := 0.

	self ssTop registerOrNone ~= NoReg ifTrue:  [
		rOopTop := self ssTop registerOrNone.
		"Ensure we are not using a duplicated register."
		(self isOopRegister: rOopTop usedBefore: 0) ifTrue: [
			rOopTop := NoReg
		]
	].

	(self ssValue: 1) registerOrNone ~= NoReg ifTrue: [
		rOopNext := (self ssValue: 1) registerOrNone.
		"Ensure we are not using a duplicated register."
		(self isOopRegister: rOopNext usedBefore: 1) ifTrue: [
			rOopNext := NoReg
		].
		rOopNext ~= NoReg ifTrue: [ 
			oopTopRegisterMask := self registerMaskFor: rOopNext.
		].
	].

	rOopTop = NoReg ifTrue: [
		rOopTop := self allocateRegNotConflictingWith: oopTopRegisterMask
	].

	rOopNext = NoReg ifTrue: [
		rOopNext := self allocateRegNotConflictingWith: (self registerMaskFor: rOopTop)
	].

	self deny: (rOopTop = NoReg or: [rOopNext = NoReg ]).
	^ aBlock value: rOopTop value: rOopNext
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/26/2017 17:18'!
allocateRegistersForLowcodeOop2ResultInteger: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rOopTop rOopNext rResult oopTopRegisterMask |
	rOopTop := rOopNext := NoReg.
	rResult := NoReg.
	oopTopRegisterMask := 0.

	self ssTop registerOrNone ~= NoReg ifTrue:  [
		rOopTop := self ssTop registerOrNone.
		"Ensure we are not using a duplicated register."
		(self isOopRegister: rOopTop usedBefore: 0) ifTrue: [
			rOopTop := NoReg
		]
	].

	(self ssValue: 1) registerOrNone ~= NoReg ifTrue: [
		rOopNext := (self ssValue: 1) registerOrNone.
		"Ensure we are not using a duplicated register."
		(self isOopRegister: rOopNext usedBefore: 1) ifTrue: [
			rOopNext := NoReg
		].
		rOopNext ~= NoReg ifTrue: [ 
			oopTopRegisterMask := self registerMaskFor: rOopNext.
		].
	].

	rOopTop = NoReg ifTrue: [
		rOopTop := self allocateRegNotConflictingWith: oopTopRegisterMask
	].

	rOopNext = NoReg ifTrue: [
		rOopNext := self allocateRegNotConflictingWith: (self registerMaskFor: rOopTop).
	].

	rResult := self allocateRegNotConflictingWith: (self registerMaskFor: rOopTop and: rOopNext).
	
	self deny: (rOopTop = NoReg or: [rOopNext = NoReg or: [rResult = NoReg]]).
	^ aBlock value: rOopTop value: rOopNext value: rResult
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/26/2017 17:16'!
allocateRegistersForLowcodeOop: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rOopTop  |
	rOopTop := NoReg.

	self ssTop registerOrNone ~= NoReg ifTrue:  [
		rOopTop := self ssTop registerOrNone.
		"Ensure we are not using a duplicated register."
		(self isOopRegister: rOopTop usedBefore: 0) ifTrue: [
			rOopTop := NoReg
		]
	].

	rOopTop = NoReg ifTrue: [
		rOopTop := self allocateRegNotConflictingWith: (self emptyRegisterMask)
	].

	self deny: (rOopTop = NoReg).
	^ aBlock value: rOopTop
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/26/2017 17:16'!
allocateRegistersForLowcodeOopResultFloat: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rOopTop  frResult |
	rOopTop := NoReg.
	frResult := NoReg.

	self ssTop registerOrNone ~= NoReg ifTrue:  [
		rOopTop := self ssTop registerOrNone.
		"Ensure we are not using a duplicated register."
		(self isOopRegister: rOopTop usedBefore: 0) ifTrue: [
			rOopTop := NoReg
		]
	].

	rOopTop = NoReg ifTrue: [
		rOopTop := self allocateRegNotConflictingWith: (self emptyRegisterMask)
	].

	frResult := self allocateFloatRegNotConflictingWith: (self emptyRegisterMask).

	self deny: (rOopTop = NoReg or: [frResult = NoReg]).
	^ aBlock value: rOopTop value: frResult
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/26/2017 17:16'!
allocateRegistersForLowcodeOopResultInteger2: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rOopTop rResult rResult2 |
	rOopTop := NoReg.
	rResult := rResult2 := NoReg.

	self ssTop registerOrNone ~= NoReg ifTrue:  [
		rOopTop := self ssTop registerOrNone.
		"Ensure we are not using a duplicated register."
		(self isOopRegister: rOopTop usedBefore: 0) ifTrue: [
			rOopTop := NoReg
		]
	].

	rOopTop = NoReg ifTrue: [
		rOopTop := self allocateRegNotConflictingWith: (self emptyRegisterMask)
	].

	rResult := self allocateRegNotConflictingWith: (self registerMaskFor: rOopTop).
	rResult2 := self allocateRegNotConflictingWith: (self registerMaskFor: rOopTop and: rResult).

	self deny: (rOopTop = NoReg or: [rResult = NoReg]).
	^ aBlock value: rOopTop value: rResult value: rResult2
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/26/2017 17:16'!
allocateRegistersForLowcodeOopResultInteger: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rOopTop rResult |
	rOopTop := NoReg.
	rResult := NoReg.

	self ssTop registerOrNone ~= NoReg ifTrue:  [
		rOopTop := self ssTop registerOrNone.
		"Ensure we are not using a duplicated register."
		(self isOopRegister: rOopTop usedBefore: 0) ifTrue: [
			rOopTop := NoReg
		]
	].

	rOopTop = NoReg ifTrue: [
		rOopTop := self allocateRegNotConflictingWith: (self emptyRegisterMask)
	].

	rResult := self allocateRegNotConflictingWith: (self registerMaskFor: rOopTop).

	self deny: (rOopTop = NoReg or: [rResult = NoReg]).
	^ aBlock value: rOopTop value: rResult
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/26/2017 17:16'!
allocateRegistersForLowcodeOopResultOop: aBlock
	<option: #LowcodeVM>
	<inline: true>
	
	| rOopTop rResult |
	rOopTop := NoReg.
	rResult := NoReg.

	self ssTop registerOrNone ~= NoReg ifTrue:  [
		rOopTop := self ssTop registerOrNone.
		"Ensure we are not using a duplicated register."
		(self isOopRegister: rOopTop usedBefore: 0) ifTrue: [
			rOopTop := NoReg
		]
	].

	rOopTop = NoReg ifTrue: [
		rOopTop := self allocateRegNotConflictingWith: (self emptyRegisterMask)
	].

	rResult := self allocateRegNotConflictingWith: (self registerMaskFor: rOopTop).

	self deny: (rOopTop = NoReg or: [rResult = NoReg]).
	^ aBlock value: rOopTop value: rResult
	! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 02:07'!
allocateRegistersForLowcodeResultFloat: aBlock
	<option: #LowcodeVM>
	<inline: true>

	| frResult |
	frResult := NoReg.

	"Float result".
	frResult := self allocateFloatRegNotConflictingWith: (self emptyRegisterMask).
	self deny: (frResult = NoReg).

	^ aBlock value: frResult
! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 02:08'!
allocateRegistersForLowcodeResultInteger2: aBlock
	<option: #LowcodeVM>
	<inline: true>

	| rResult rResult2 |
	rResult := rResult2 := NoReg.

	"Float result".
	rResult := self allocateRegNotConflictingWith: (self emptyRegisterMask).
	rResult2 := self allocateRegNotConflictingWith: (self registerMaskFor: rResult).
	self deny: (rResult = NoReg or: [rResult2 = NoReg]).

	^ aBlock value: rResult value: rResult2
! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/6/2017 02:07'!
allocateRegistersForLowcodeResultInteger: aBlock
	<option: #LowcodeVM>
	<inline: true>

	| rResult |
	rResult := NoReg.

	"Float result".
	rResult := self allocateRegNotConflictingWith: (self emptyRegisterMask).
	self deny: (rResult = NoReg).

	^ aBlock value: rResult
! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 4/18/2016 18:59'!
annotateInstructionForBytecode
	"Annotate the current instruction as having a bytecode pc.  If the current instruction
	 is already annotated, insert a nop and annotate that.  We cannot have the same instruction
	 address carry multiple annotations."
	<inline: true>
	self annotateBytecode: (self prevInstIsPCAnnotated
								ifTrue: [self Nop]
								ifFalse: [self Label]).! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 12/5/2016 10:49'!
anyReferencesToRegister: reg inAllButTopNItems: n
	| regMask |
	regMask := self registerMaskFor: reg.
	simStackPtr - n to: 0 by: -1 do:
		[:i|
		((self simStackAt: i) registerMask anyMask: regMask) ifTrue:
			[^true]].
	^false! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/12/2013 10:48'!
anyReferencesToRegister: reg inTopNItems: n
	| regMask |
	regMask := self registerMaskFor: reg.
	simStackPtr to: simStackPtr - n + 1 by: -1 do:
		[:i|
		((self simStackAt: i) registerMask anyMask: regMask) ifTrue:
			[^true]].
	^false! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 2/6/2018 17:57'!
assertCorrectSimStackPtr
	<inline: true>
	"Would like to assert simply simSpillBase > methodOrBlockNumTemps but can't because
	 of the initialNils hack for nested blocks in SqueakV3PlusClosures"
	self assert: (simSpillBase >= methodOrBlockNumTemps
				or: [self maybeCompilingFirstPassOfBlockWithInitialPushNil and: [simSpillBase > methodOrBlockNumArgs]]).
	(needsFrame and: [simSpillBase > 0]) ifTrue:
		[self assert: (self simStackAt: simSpillBase - 1) spilled == true.
		 self assert: (simSpillBase > simStackPtr or: [(self simStackAt: simSpillBase) spilled == false])].
 	self cCode: '' inSmalltalk:
		[deadCode ifFalse:
			[self assert: simStackPtr + (needsFrame ifTrue: [0] ifFalse: [1])
						= (self debugStackPointerFor: bytecodePC)]].! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 12/10/2015 09:18'!
availableRegOrNoneNotConflictingWith: regMask
	<inline: true>
	"If there's a free register, answer it, otherwise answer NoReg."
	^backEnd availableRegisterOrNoneFor: (self liveRegisters bitOr: regMask)! !

!StackToRegisterMappingCogit methodsFor: 'inline ffi' stamp: 'eem 10/9/2019 14:33'!
beginHighLevelCall: alignment
	<option: #LowcodeVM>
	| mask actualAlignment |

	"Store the smalltalk pointers"
	self ssFlushAll.
	backEnd genSaveStackPointers.
	backEnd hasLinkRegister ifTrue:
		[self MoveAw: coInterpreter instructionPointerAddress R: LinkReg].
	"Load the C pointers"
	backEnd genLoadCStackPointer.

	"Load the native stack frame pointer"
	self MoveMw: self frameOffsetOfNativeFramePointer r: FPReg R: FPReg.
	self SubCq: 1 R: FPReg.

	actualAlignment := (alignment max: BytesPerWord).
	actualAlignment > BytesPerWord ifTrue:
		[mask := actualAlignment negated.
		self AndCq: mask R: SPReg].
	currentCallCleanUpSize := 0! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 11/22/2010 11:09'!
bytecodeFixupClass
	<doNotGenerate>
	^CogSSBytecodeFixup! !

!StackToRegisterMappingCogit methodsFor: 'trampolines' stamp: 'eem 12/4/2010 08:41'!
cPICMissTrampolineFor: numArgs
	^picMissTrampolines at: (numArgs min: self numRegArgs + 1)! !

!StackToRegisterMappingCogit methodsFor: 'debugging' stamp: 'eem 7/3/2014 17:05'!
callCogCodePopReceiverArg0Regs
	"This is a static version of ceCallCogCodePopReceiverArg0Regs
	 for break-pointing when debugging in C."
	<api>
	<inline: false>
	"This exists only for break-pointing."
	self cCode: [self realCECallCogCodePopReceiverArg0Regs]
		inSmalltalk: [self ceCallCogCodePopReceiverArg0Regs]! !

!StackToRegisterMappingCogit methodsFor: 'debugging' stamp: 'eem 7/3/2014 17:06'!
callCogCodePopReceiverArg1Arg0Regs
	"This is a static version of ceCallCogCodePopReceiverArg1Arg0Regs
	 for break-pointing when debugging in C."
	<api>
	<inline: false>
	"This exists only for break-pointing."
	self cCode: [self realCECallCogCodePopReceiverArg1Arg0Regs]
		inSmalltalk: [self ceCallCogCodePopReceiverArg1Arg0Regs]! !

!StackToRegisterMappingCogit methodsFor: 'inline ffi' stamp: 'rsf 8/10/2016 00:20'!
callSwitchToCStack
	<option: #LowcodeVM>
	self MoveAw: self cFramePointerAddress R: FPReg.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'inline ffi' stamp: 'eem 2/1/2020 18:51'!
callSwitchToSmalltalkStack
	<option: #LowcodeVM>
	"Restore the link register"
	backEnd hasVarBaseRegister ifTrue:
		[self MoveCq: self varBaseAddress R: VarBaseReg].
	backEnd hasLinkRegister ifTrue:
		[self MoveAw: coInterpreter instructionPointerAddress R: LinkReg].
	backEnd genLoadStackPointers! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 7/3/2014 17:00'!
ceCall0ArgsPIC
	<api: 'extern void (*ceCall0ArgsPIC)()'>
	<doNotGenerate>
	self simulateEnilopmart: ceCall0ArgsPIC numArgs: 1! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 7/3/2014 17:00'!
ceCall1ArgsPIC
	<api: 'extern void (*ceCall1ArgsPIC)()'>
	<doNotGenerate>
	self simulateEnilopmart: ceCall1ArgsPIC numArgs: 1! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 7/3/2014 17:01'!
ceCall2ArgsPIC
	<api: 'extern void (*ceCall2ArgsPIC)()'>
	<doNotGenerate>
	self simulateEnilopmart: ceCall2ArgsPIC numArgs: 1! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 7/3/2014 16:46'!
ceCallCogCodePopReceiverArg0Regs
	<api: 'extern void (*ceCallCogCodePopReceiverArg0Regs)()'>
	<doNotGenerate>
	self simulateEnilopmart: ceCallCogCodePopReceiverArg0Regs numArgs: 2! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 7/3/2014 16:46'!
ceCallCogCodePopReceiverArg1Arg0Regs
	<api: 'extern void (*ceCallCogCodePopReceiverArg1Arg0Regs)()'>
	<doNotGenerate>
	self simulateEnilopmart: ceCallCogCodePopReceiverArg1Arg0Regs numArgs: 3! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 1/25/2018 08:47'!
compileAbstractInstructionsFrom: start through: end
	"Loop over bytecodes, dispatching to the generator for each bytecode, handling fixups in due course."
	| nextOpcodeIndex descriptor nExts fixup result |
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<var: #fixup type: #'BytecodeFixup *'>
	self traceSimStack.
	bytecodePC := start.
	nExts := result := 0.
	descriptor := nil.
	deadCode := false.
	[self maybeHaltIfDebugPC.
	 self mergeWithFixupIfRequired: (fixup := self fixupAt: bytecodePC).
	 descriptor := self loadBytesAndGetDescriptor.
	 nextOpcodeIndex := opcodeIndex.
	 result := deadCode
				ifTrue: [self mapDeadDescriptorIfNeeded: descriptor]
				ifFalse: [self perform: descriptor generator].
	 result = 0 ifTrue: [self assertExtsAreConsumed: descriptor].
	 self traceDescriptor: descriptor; traceSimStack.
	 self patchFixupTargetIfNeeded: fixup nextOpcodeIndex: nextOpcodeIndex.
	 self maybeDumpLiterals: descriptor.
	 bytecodePC := self nextBytecodePCFor: descriptor exts: nExts.
	 result = 0 and: [bytecodePC <= end]] whileTrue:
		[nExts := descriptor isExtension ifTrue: [nExts + 1] ifFalse: [0]].
	self checkEnoughOpcodes.
	^result! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 03/04/2017 09:43'!
compileBlockBodies
	<inline: false>
	| result compiledBlocksCount blockStart savedNeedsFrame savedNumArgs savedNumTemps
	  initialStackPtr initialOpcodeIndex initialIndexOfIRC initialCounterIndex |
	<var: #blockStart type: #'BlockStart *'>
	self assert: blockCount > 0.
	"scanBlock: in compileBlockEntry: sets both of these appropriately for each block."
	savedNeedsFrame := needsFrame.
	savedNumArgs := methodOrBlockNumArgs.
	savedNumTemps := methodOrBlockNumTemps.
	inBlock := InVanillaBlock.
	compiledBlocksCount := 0.
	[compiledBlocksCount < blockCount] whileTrue:
		[compilationPass := 1.
		 blockStart := self blockStartAt: compiledBlocksCount.
		 (result := self scanBlock: blockStart) < 0 ifTrue: [^result].
		 initialOpcodeIndex := opcodeIndex.
		 initialCounterIndex := self maybeCounterIndex."for SistaCogit"
		 literalsManager saveForRecompile.
		 NewspeakVM ifTrue:
			[initialIndexOfIRC := indexOfIRC].
		 [self compileBlockEntry: blockStart.
		  initialStackPtr := simStackPtr.
		  (result := self compileAbstractInstructionsFrom: blockStart startpc + (self pushNilSize: methodObj numInitialNils: blockStart numInitialNils)
						through: blockStart startpc + blockStart span - 1) < 0 ifTrue:
			[^result].
		  "If the final simStackPtr is less than the initial simStackPtr then scanBlock: over-
		   estimated the number of initial nils (because it assumed one or more pushNils to
		   produce an operand were pushNils to initialize temps.  This is very rare, so
		   compensate by checking, adjusting numInitialNils and recompiling the block body.
		   N.B.  No need to reinitialize the literalsManager because it answers existing literals."
		  initialStackPtr = simStackPtr]
			whileFalse:
				[self assert: (initialStackPtr > simStackPtr or: [deadCode]).
				 compilationPass := compilationPass + 1. "for asserts"
				 blockStart numInitialNils: blockStart numInitialNils + simStackPtr - initialStackPtr.
				 blockStart fakeHeader dependent: nil.
				 self reinitializeFixupsFrom: blockStart startpc + blockStart numInitialNils
					through: blockStart startpc + blockStart span - 1.
				 self cCode: 'bzero(abstractOpcodes + initialOpcodeIndex,
									(opcodeIndex - initialOpcodeIndex) * sizeof(AbstractInstruction))'
					inSmalltalk: [initialOpcodeIndex to: opcodeIndex - 1 do:
									[:i| abstractOpcodes at: i put: (CogCompilerClass for: self)]].
				 opcodeIndex := initialOpcodeIndex.
				 self maybeSetCounterIndex: initialCounterIndex. "For SistaCogit"
				 literalsManager resetForRecompile.
				 NewspeakVM ifTrue:
					[indexOfIRC := initialIndexOfIRC]].
		compiledBlocksCount := compiledBlocksCount + 1].
	needsFrame := savedNeedsFrame.
	methodOrBlockNumArgs := savedNumArgs.
	methodOrBlockNumTemps := savedNumTemps.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/25/2016 18:58'!
compileBlockFrameBuild: blockStart
	"Build a frame for a block activation.  See CoInterpreter class>>initializeFrameIndices.
	 Override to push the register receiver and register arguments, if any, and to correctly
	 initialize the explicitly nilled/pushed temp entries (they are /not/ of type constant nil)."
	super compileBlockFrameBuild: blockStart.
	methodOrBlockNumTemps := blockStart numArgs + blockStart numCopied + blockStart numInitialNils.
	self initSimStackForFramefulMethod: blockStart startpc.
	blockStart numInitialNils > 0 ifTrue:
		[blockStart numInitialNils > 1
			ifTrue:
				[self genMoveNilR: TempReg.
				 1 to: blockStart numInitialNils do:
					[:ign| self PushR: TempReg]]
			ifFalse:
				[self genPushConstant: objectMemory nilObject]]! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/30/2011 17:26'!
compileBlockFramelessEntry: blockStart
	"Make sure ReceiverResultReg holds the receiver, loaded from
	 the closure, which is what is initially in ReceiverResultReg"
	<var: #blockStart type: #'BlockStart *'>
	methodOrBlockNumTemps := blockStart numArgs + blockStart numCopied + blockStart numInitialNils.
	self initSimStackForFramelessBlock: blockStart startpc.
	super compileBlockFramelessEntry: blockStart! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'cb 4/15/2016 11:09'!
compileCogFullBlockMethod: numCopied
	<option: #SistaV1BytecodeSet>
	methodOrBlockNumTemps := coInterpreter tempCountOf: methodObj.
	self cCode: '' inSmalltalk:
		[debugStackPointers := coInterpreter debugStackPointersFor: methodObj].
	^super compileCogFullBlockMethod: numCopied! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/30/2010 21:19'!
compileCogMethod: selector
	methodOrBlockNumTemps := coInterpreter tempCountOf: methodObj.
	self cCode: '' inSmalltalk:
		[debugStackPointers := coInterpreter debugStackPointersFor: methodObj].
	^super compileCogMethod: selector! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 1/18/2011 10:05'!
compileEntireMethod
	"Compile the abstract instructions for the entire method, including blocks."
	regArgsHaveBeenPushed := false.
	^super compileEntireMethod! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'cb 8/17/2016 15:59'!
compileFrameBuild
	"Build a frame for a CogMethod activation.  See CoInterpreter class>>initializeFrameIndices.
	 Override to push the register receiver and register arguments, if any."
	self cppIf: IMMUTABILITY ifTrue: 
		[useTwoPaths ifTrue: 
			[self compileTwoPathFrameBuild.
		 	^self]].
	needsFrame ifFalse:
		[useTwoPaths ifTrue: 
			[self compileTwoPathFramelessInit].
		 self initSimStackForFramelessMethod: initialPC.
		 ^self].
	self deny: useTwoPaths.
	self genPushRegisterArgs.
	super compileFrameBuild.
	self initSimStackForFramefulMethod: initialPC! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'cb 9/5/2016 15:41'!
compileFullBlockFramelessEntry: numCopied
	"Make sure ReceiverResultReg holds the receiver, loaded from
	 the closure, which is what is initially in ReceiverResultReg"
	self initSimStackForFramelessBlock: initialPC.
	super compileFullBlockFramelessEntry: numCopied! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 9/18/2019 15:01'!
compileFullBlockMethodFrameBuild: numCopied
	<option: #SistaV1BytecodeSet>
	useTwoPaths ifTrue: 
		[ "method with only inst var store, we compile only slow path for now" 
		 useTwoPaths := false.
		 self cppIf: IMMUTABILITY ifTrue: [ needsFrame := true ] ].
	needsFrame ifFalse:
		["it is OK for numCopied to be non-zero provided that the block does not actually use the copied values.
		 There are some blocks like this, e.g. that simply reference copied values to mark them as used for Slang.
		 See e.g. CroquetPlugin>>#primitiveGatherEntropy which contains the block [bufPtr. bufSize. false],
		 which the bytecode compiler optimizes to [false]."
		 false ifTrue: [self assert: numCopied = 0].
		 self compileFullBlockFramelessEntry: numCopied.
		 self initSimStackForFramelessBlock: initialPC.
		 ^self].
	super compileFullBlockMethodFrameBuild: numCopied.
	self initSimStackForFramefulMethod: initialPC! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'cb 7/25/2016 10:39'!
compileTwoPathFrameBuild
	"We are in a method where the frame is needed *only* for instance variable store, typically a setter method.
	This case has 20% overhead with Immutability compared to setter without immutability because of the stack
	frame creation. We compile two path, one where the object is immutable, one where it isn't. At the beginning 
	of the frame build, we take one path or the other depending on the receiver mutability.
	
	Note: this specific case happens only where there are only instance variabel stores. We could do something
	similar for literal variable stores, but we don't as it's too uncommon."
	<option: #IMMUTABILITY>
	| jumpImmutable jumpOld |
	self assert: useTwoPaths.
	self assert: blockCount = 0.
	jumpImmutable := objectRepresentation genJumpImmutable: ReceiverResultReg scratchReg: TempReg.
	jumpOld := objectRepresentation genJumpInOldSpace: ReceiverResultReg.
	"first path. The receiver is mutable"
	self assert: needsFrame not.
	self initSimStackForFramelessMethod: initialPC.
	self compileMethodBody.
	"second path. The receiver is mutable"
	useTwoPaths := false. "reset because it impacts inst var store compilation"
	needsFrame := true.
	jumpOld jmpTarget: (jumpImmutable jmpTarget: self Label).
	self genPushRegisterArgs.
	super compileFrameBuild.
	self initSimStackForFramefulMethod: initialPC! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/1/2016 16:25'!
compileTwoPathFramelessInit
	"We are in a frameless method with at least two inst var stores. We compile two paths,
	 one where the object is in new space, and one where it isn't. At the beginning 
	 of the method, we take one path or the other depending on the receiver being in newSpace."
	| jumpOld |
	self deny: IMMUTABILITY.
	self deny: needsFrame.
	self assert: useTwoPaths.
	jumpOld := objectRepresentation genJumpInOldSpace: ReceiverResultReg.
	"first path. The receiver is young"
	self initSimStackForFramelessMethod: initialPC.
	self compileMethodBody.
	"second path. The receiver is old"
	useTwoPaths := false. "reset because it impacts inst var store compilation"
	jumpOld jmpTarget: self Label! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 1/22/2018 18:02'!
debugStackPointerFor: bcpc
	<doNotGenerate>
	^(debugStackPointers at: bcpc) + (needsFrame ifTrue: [0] ifFalse: [1])! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 3/10/2018 10:25'!
doubleExtendedDoAnythingBytecode
	"Replaces the Blue Book double-extended send [132], in which the first byte was wasted on 8 bits of argument count. 
	Here we use 3 bits for the operation sub-type (opType),  and the remaining 5 bits for argument count where needed. 
	The last byte give access to 256 instVars or literals. 
	See also secondExtendedSendBytecode"
	| opType |
	opType := byte1 >> 5.
	opType = 0 ifTrue:
		[^self genSend: byte2 numArgs: (byte1 bitAnd: 31)].
	opType = 1 ifTrue:
		[^self genSendSuper: byte2 numArgs: (byte1 bitAnd: 31)].
	"We need a map entry for this bytecode for correct parsing.
	 The sends will get an IsSend entry anyway.  The other cases need a fake one."
	opType caseOf: {
			[2]	->	[(coInterpreter isReadMediatedContextInstVarIndex: byte2)
						ifTrue: [self genPushMaybeContextReceiverVariable: byte2]
						ifFalse: [self genPushReceiverVariable: byte2.
								self annotateInstructionForBytecode.
								^0]].
			[3]	->	[self genPushLiteralIndex: byte2.
					 self annotateInstructionForBytecode.
					 ^0].
			[4]	->	[self genPushLiteralVariable: byte2.].
			[7]	->	[self genStorePop: false LiteralVariable: byte2.
					self cppIf: IMMUTABILITY ifTrue: [ "genStorePop:LiteralVariable: annotates; don't annotate twice" ^0 ] ] }
		otherwise: "5 & 6"
			[(coInterpreter isWriteMediatedContextInstVarIndex: byte2)
				ifTrue: [self genStorePop: opType = 6 MaybeContextReceiverVariable: byte2]
				ifFalse: [self genStorePop: opType = 6 ReceiverVariable: byte2].
			self cppIf: IMMUTABILITY ifTrue: [ "genStorePop:...ReceiverVariable: annotate; don't annotate twice" ^0 ]].
	"We need a map entry for this bytecode for correct parsing (if the method builds a frame)."
	self assert: needsFrame.
	"genPushMaybeContextInstVar, pushLitVar, store & storePop all generate code"
	self assert: self prevInstIsPCAnnotated not.
	self annotateBytecode: self Label.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 4/18/2016 09:14'!
duplicateTopBytecode
	| desc |
	<var: #desc type: #SimStackEntry>
	desc := self ssTopDescriptor.
	^self ssPushDesc: desc! !

!StackToRegisterMappingCogit methodsFor: 'inline ffi' stamp: 'rsf 8/10/2016 00:20'!
endHighLevelCallWithCleanup
	<option: #LowcodeVM>
	currentCallCleanUpSize > 0 ifTrue: [
		self AddCq: currentCallCleanUpSize R: SPReg
	].
   self callSwitchToSmalltalkStack.
! !

!StackToRegisterMappingCogit methodsFor: 'inline ffi' stamp: 'rsf 8/10/2016 00:20'!
endHighLevelCallWithoutCleanup
	<option: #LowcodeVM>
	self callSwitchToSmalltalkStack.! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 1/21/2018 13:00'!
ensureFixupAt: targetPC
	"Make sure there's a flagged fixup at the target pc in fixups.
	 Initially a fixup's target is just a flag.  Later on it is replaced with a proper instruction."
	<returnTypeC: #'BytecodeFixup *'>
	| fixup |
	<var: #fixup type: #'BytecodeFixup *'>
	fixup := self fixupAt:  targetPC.
	self traceFixup: fixup merge: true.
	self cCode: '' inSmalltalk:
		[self assert: simStackPtr = (self debugStackPointerFor: targetPC).
		 (fixup isMergeFixupOrIsFixedUp
		  and: [fixup isBackwardBranchFixup not]) ifTrue: "ignore backward branch targets"
			[self assert: fixup simStackPtr = simStackPtr]].
	fixup isNonMergeFixupOrNotAFixup
		ifTrue: "convert a non-merge into a merge"
			[fixup becomeMergeFixup.
			 fixup simStackPtr: simStackPtr.
			 LowcodeVM ifTrue: [
				 fixup simNativeStackPtr: simNativeStackPtr.
				 fixup simNativeStackSize: simNativeStackSize]]
		ifFalse:
			[fixup isBackwardBranchFixup
				ifTrue: "this is the target of a backward branch and
						 so doesn't have a simStackPtr assigned yet."
					[fixup simStackPtr: simStackPtr.
		 			 LowcodeVM ifTrue:
			 			[fixup simNativeStackPtr: simNativeStackPtr.
			 			 fixup simNativeStackSize: simNativeStackSize]]
				ifFalse:
					[self assert: fixup simStackPtr = simStackPtr.
					 LowcodeVM ifTrue:
				 		[self assert: fixup simNativeStackPtr = simNativeStackPtr.
		 			 	 self assert: fixup simNativeStackSize = simNativeStackSize]]].
	fixup recordBcpc: bytecodePC.
	^fixup! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 2/12/2017 16:26'!
ensureNonMergeFixupAt: targetPC
	"Make sure there's a flagged fixup at the target pc in fixups.
	 Initially a fixup's target is just a flag.  Later on it is replaced with a proper instruction."
	<returnTypeC: #'BytecodeFixup *'>
	| fixup |
	<var: #fixup type: #'BytecodeFixup *'>
	fixup := self fixupAt:  targetPC.
	self traceFixup: fixup merge: true.
	fixup notAFixup ifTrue:
		[fixup becomeNonMergeFixup].
	self cCode: '' inSmalltalk:
		[fixup isMergeFixupOrIsFixedUp ifTrue:
			[self assert:
					(fixup isBackwardBranchFixup
					 or: [fixup simStackPtr = (self debugStackPointerFor: targetPC)])]].
	fixup recordBcpc: bytecodePC.
	^fixup! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/24/2018 15:13'!
ensureReceiverResultRegContainsSelf
	needsFrame
		ifTrue:
			[self receiverIsInReceiverResultReg ifFalse:
				[self ssAllocateRequiredReg: ReceiverResultReg.
				 self putSelfInReceiverResultReg.
				 self simSelf liveRegister: ReceiverResultReg]]
		ifFalse:
			[self assert: (self simSelf type = SSRegister
						  and: [self simSelf register = ReceiverResultReg
						  and: [self receiverIsInReceiverResultReg]])]! !

!StackToRegisterMappingCogit methodsFor: 'peephole optimizations' stamp: 'eem 6/1/2014 17:40'!
evaluate: descriptor at: pc
	<var: #descriptor type: #'BytecodeDescriptor *'>
	byte0 := objectMemory fetchByte: pc ofObject: methodObj.
	self assert: descriptor = (self generatorAt: bytecodeSetOffset + byte0).
	self loadSubsequentBytesForDescriptor: descriptor at: pc.
	self perform: descriptor generator! !

!StackToRegisterMappingCogit methodsFor: 'peephole optimizations' stamp: 'eem 1/21/2018 21:16'!
eventualTargetOf: targetBytecodePC
	"Attempt to follow a branch to a pc.  Handle branches to unconditional jumps and
	 branches to push: aBoolean; conditional branch pairs.  If the branch cannot be
	 followed answer targetBytecodePC. It is not possible to follow jumps to conditional
	 branches because the stack changes depth.  That following is left to the genJumpIf:to:
	 clients."

	| currentTarget nextPC nExts descriptor span cond |
	<var: #descriptor type: #'BytecodeDescriptor *'>
	nextPC := currentTarget := targetBytecodePC.
	[ nExts := 0.
	 [descriptor := self generatorForPC: nextPC.
	  descriptor isReturn ifTrue: [^currentTarget]. "avoid stepping off the end of methods"
	  descriptor isExtension]
		whileTrue:
			[nExts := nExts + 1.
			 nextPC := nextPC + descriptor numBytes].
	 descriptor isUnconditionalBranch
		ifTrue:
			[span := self spanFor: descriptor at: nextPC exts: nExts in: methodObj.
			 span < 0 ifTrue: "Do *not* follow backward branches; these are interrupt points and should not be elided."
				[^currentTarget].
			 nextPC := nextPC + descriptor numBytes + span]
		ifFalse:
			[descriptor generator == #genPushConstantTrueBytecode ifTrue: [ cond := true ]
			 ifFalse: [ descriptor generator == #genPushConstantFalseBytecode ifTrue: [ cond := false ]
			 ifFalse: [ ^currentTarget ] ].
			 "Don't step into loops across a pushTrue; jump:if: boundary, so as not to confuse stack depth fixup."
			 (self fixupAt: nextPC) isBackwardBranchFixup ifTrue:
				[^currentTarget].
			 nextPC := self eventualTargetOf: nextPC + descriptor numBytes.
			 nExts := 0.
			 [descriptor := self generatorForPC: nextPC.
			  descriptor isReturn ifTrue: [^currentTarget]. "avoid stepping off the end of methods"
			  descriptor isExtension]
				whileTrue:
					[nExts := nExts + 1.
					 nextPC := nextPC + descriptor numBytes].
			 descriptor isBranch ifFalse:
				[^currentTarget].
			 descriptor isUnconditionalBranch ifTrue:
				[^currentTarget].
			 nextPC := cond == descriptor isBranchTrue
									ifTrue: [nextPC
											+ descriptor numBytes
											+ (self spanFor: descriptor at: nextPC exts: nExts in: methodObj)]
									ifFalse: [nextPC + descriptor numBytes]].
	 currentTarget := nextPC]
		repeat! !

!StackToRegisterMappingCogit methodsFor: 'testing' stamp: 'cb 7/25/2016 13:08'!
extASpecifiesNoMustBeBoolean
	<inline: true>
	^ extA anyMask: 1! !

!StackToRegisterMappingCogit methodsFor: 'testing' stamp: 'cb 7/25/2016 13:16'!
extBSpecifiesImmCheck 
	"This is a negative check"
	<inline: true>
	^ extB noMask: 4! !

!StackToRegisterMappingCogit methodsFor: 'testing' stamp: 'cb 7/25/2016 13:08'!
extBSpecifiesInitializeInstance
	"This is a negative; do not intialize if 1 is present in extB.  This is also the no-store-check bit."
	<inline: true>
	^ extB noMask: 1! !

!StackToRegisterMappingCogit methodsFor: 'testing' stamp: 'cb 4/5/2016 15:54'!
extBSpecifiesMaybeContext
	<inline: true>
	^ extB anyMask: 2! !

!StackToRegisterMappingCogit methodsFor: 'testing' stamp: 'cb 8/17/2016 15:14'!
extBSpecifiesStoreCheck 
	"This is a negative check"
	<inline: true>
	^ extB noMask: 1! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/21/2018 16:15'!
extractMaybeBranchDescriptorInto: fourArgBlock
	"Looks one instruction ahead of the current bytecodePC and answers its bytecode descriptor and its pc.
	 If the instruction found is a branch, also answers the pc after the branch and the pc targeted by the branch.
	 For convenience, avoiding duplication in the senders, it follows those two pcs to their eventual targets."
	| primDescriptor nextPC nExts branchDescriptor targetBytecodePC postBranchPC |
	<inline: true>
	<var: #primDescriptor type: #'BytecodeDescriptor *'>
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>
	
	primDescriptor := self generatorAt: byte0.

	nextPC := bytecodePC + primDescriptor numBytes.
	nExts := 0.
	[[branchDescriptor := self generatorForPC: nextPC.
	  branchDescriptor isExtension] whileTrue:
		[nExts := nExts + 1.
		 nextPC := nextPC + branchDescriptor numBytes].
	 branchDescriptor isUnconditionalBranch]
		whileTrue:
			[nextPC := self eventualTargetOf: nextPC
											+ branchDescriptor numBytes
											+ (self spanFor: branchDescriptor at: nextPC exts: nExts in: methodObj)].

	targetBytecodePC := postBranchPC := 0.

	(branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse])
		ifTrue: 
			[targetBytecodePC := self eventualTargetOf: nextPC
														+ branchDescriptor numBytes
														+ (self spanFor: branchDescriptor at: nextPC exts: nExts in: methodObj).
			 postBranchPC := self eventualTargetOf: nextPC + branchDescriptor numBytes]
		ifFalse:
			[nextPC := bytecodePC + primDescriptor numBytes].

	fourArgBlock value: branchDescriptor value: nextPC value: postBranchPC value: targetBytecodePC! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 2/9/2017 09:58'!
fixupAtIndex: index
	"The fixups Array maps to bytecode pcs such that initialPC maps to index 0.
	 fixupAt: does the conversion.  Override to add breakpointing for fixups of bytecode PCs."
	<cmacro: '(index) (&fixups[index])'>
	<returnTypeC: #'BytecodeFixup *'>
	((debugFixupBreaks includes: index + initialPC)
	 and: [breakMethod isNil or: [methodObj = breakMethod]]) ifTrue:
		[self halt].
	^self addressOf: (fixups at: index)! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 3/6/2017 01:18'!
freeAnyFloatRegNotConflictingWith: regMask
	"Spill the closest register on stack not conflicting with regMask. 
	Assertion Failure if regMask has already all the registers"
	<var: #desc type: #'CogSimStackEntry *'>
	| reg index |
	self assert: needsFrame.
	reg := NoReg.
	index := simSpillBase max: 0.
	LowcodeVM ifTrue: [
		index := simNativeSpillBase max: 0.
		[reg = NoReg and: [index < simNativeStackPtr]] whileTrue: 
			[ | desc |
			 desc := self simNativeStackAt: index.
			(desc type = SSRegisterSingleFloat or: [desc type = SSRegisterDoubleFloat]) ifTrue:
				[(regMask anyMask: (self registerMaskFor: desc register)) ifFalse: 
					[reg := desc register]].
			 index := index + 1].
	].
	self deny: reg = NoReg.
	self ssAllocateRequiredFloatReg: reg.
	^reg! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 12/10/2015 09:19'!
freeAnyRegNotConflictingWith: regMask
	"Spill the closest register on stack not conflicting with regMask. 
	Assertion Failure if regMask has already all the registers"
	<var: #desc type: #'CogSimStackEntry *'>
	| reg index |
	self assert: needsFrame.
	reg := NoReg.
	index := simSpillBase max: 0.
	[reg = NoReg and: [index < simStackPtr]] whileTrue: 
		[ | desc |
		 desc := self simStackAt: index.
		 desc type = SSRegister ifTrue:
			[(regMask anyMask: (self registerMaskFor: desc register)) ifFalse: 
				[reg := desc register]].
		 index := index + 1].
	self deny: reg = NoReg.
	self ssAllocateRequiredReg: reg.
	^reg! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'cb 4/17/2016 12:14'!
genBlockReturn
	"Return from block, assuming result already loaded into ReceiverResultReg."
	super genBlockReturn.
	deadCode := true. "can't fall through"
	^0! !

!StackToRegisterMappingCogit methodsFor: 'initialization' stamp: 'eem 2/1/2020 18:52'!
genCallPICEnilopmartNumArgs: numArgs
	"Generate special versions of the ceCallCogCodePopReceiverAndClassRegs
	 enilopmart that also pop register args from the stack to undo the pushing of
	 register args in the abort/miss trampolines."
	<returnTypeC: 'void (*genCallPICEnilopmartNumArgs(sqInt numArgs))(void)'>
	| size endAddress enilopmart |
	self zeroOpcodeIndex.
	backEnd hasVarBaseRegister ifTrue:
		[self MoveCq: self varBaseAddress R: VarBaseReg]. "Must happen first; value may be used in genLoadStackPointers"
	backEnd genLoadStackPointers.
	self PopR: ClassReg. "cacheTag"
	self PopR: TempReg. "entry-point"
	self PopR: (backEnd hasLinkRegister ifTrue: [LinkReg] ifFalse: [SendNumArgsReg]). "retpc"
	numArgs > 0 ifTrue:
		[numArgs > 1 ifTrue:
			[self PopR: Arg1Reg.
			 self assert: self numRegArgs = 2].
		 self PopR: Arg0Reg].
	self PopR: ReceiverResultReg.
	backEnd hasLinkRegister ifFalse: [self PushR: SendNumArgsReg]. "retpc"
	self JumpR: TempReg.
	self computeMaximumSizes.
	size := self generateInstructionsAt: methodZoneBase.
	endAddress := self outputInstructionsAt: methodZoneBase.
	self assert: methodZoneBase + size = endAddress.
	enilopmart := methodZoneBase.
	methodZoneBase := self alignUptoRoutineBoundary: endAddress.
	backEnd stopsFrom: endAddress to: methodZoneBase - 1.
	self recordGeneratedRunTime: (self trampolineName: 'ceCallPIC' numRegArgs: numArgs) address: enilopmart.
	^self cCoerceSimple: enilopmart to: #'void (*)(void)'! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 5/14/2020 13:40'!
genCallPrimitiveBytecode
	"SistaV1:	**	248	(2)	11111000 	iiiiiiii		mssjjjjj		Call Primitive #iiiiiiii + (jjjjj * 256) 
								m=1 means inlined primitive, no hard return after execution. 
								ss defines the unsafe operation set used to encode the operations. 
								(ss = 0 means sista unsafe operations, ss = 01 means lowcode operations, other numbers are as yet used).
	 See SistaCogit genCallPrimitiveBytecode, EncoderForSistaV1's class comment and StackInterpreter>>#callPrimitiveBytecode for more information."
	| prim primSet |
	byte2 < 128 ifTrue:
		[^bytecodePC = initialPC
			ifTrue: [0]
			ifFalse: [EncounteredUnknownBytecode]].
	prim := byte2 - 128 << 8 + byte1.
	primSet := prim >> 13 bitAnd: 3.
	prim := prim bitAnd: 8191.
	LowcodeVM ifTrue:
		[primSet = 1 ifTrue:
			[prim < 1000 ifTrue:
				[^self genLowcodeNullaryInlinePrimitive: prim].

			prim < 2000 ifTrue:
				[^self genLowcodeUnaryInlinePrimitive: prim - 1000].
			
			prim < 3000 ifTrue:
				[^ self genLowcodeBinaryInlinePrimitive: prim - 2000].

			prim < 4000 ifTrue:
				[^self genLowcodeTrinaryInlinePrimitive: prim - 3000]]].

	^EncounteredUnknownBytecode! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'cb 11/24/2016 11:08'!
genCmpArgIsConstant: argIsConstant rcvrIsConstant: rcvrIsConstant argReg: argReg rcvrReg: rcvrReg
	"Generates the Cmp instruction for the top two ssStack values. 
	ssTop is called the argument and ssValue: 1 is called the receiver.
	0 or 1 value on ssStack is a constant (but not 2). If a value is a constant, then noReg is passed as argReg or rcvrReg.
	The instruction is different if one of the operands is a constant.
	In the case of the v3 memory manager, the constant could be annotable."
	<inline: true>
	self assert: (argReg ~= NoReg or: [rcvrReg ~= NoReg]).
	argIsConstant 
		ifTrue: [ self genCmpConstant: self ssTop constant R: rcvrReg ]
		ifFalse: [ rcvrIsConstant
			ifTrue: [ self genCmpConstant: (self ssValue: 1) constant R: argReg ]
			ifFalse: [ self CmpR: argReg R: rcvrReg ] ].! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:07'!
genExtPushClosureBytecode
	"Block compilation.  At this point in the method create the block.  Note its start
	 and defer generating code for it until after the method and any other preceding
	 blocks.  The block's actual code will be compiled later."
	"253		11111101 eei i i kkk	jjjjjjjj		Push Closure Num Copied iii (+ Ext A // 16 * 8) Num Args kkk (+ Ext A \\ 16 * 8) BlockSize jjjjjjjj (+ Ext B * 256). ee = num extensions"
	| startpc numArgs numCopied |
	self assert: needsFrame.
	startpc := bytecodePC + (self generatorAt: byte0) numBytes.
	self addBlockStartAt: startpc "0 relative"
		numArgs: (numArgs := (byte1 bitAnd: 16r7) + (extA \\ 16 * 8))
		numCopied: (numCopied := ((byte1 >> 3) bitAnd: 7) + (extA // 16 * 8))
		span: byte2 + (extB << 8).
	extA := numExtB := extB := 0.

	objectRepresentation createsClosuresInline
		ifTrue: [ self genInlineClosure: startpc numArgs: numArgs numCopied: numCopied ]
		ifFalse: [ self genOutlineClosure: startpc numArgs: numArgs numCopied: numCopied ].
		
	^ 0! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 1/22/2018 18:18'!
genExtPushFullClosureBytecode
	"Full Block creation compilation. The block's actual code will be compiled separatedly."
	"*	255		11111111	xxxxxxxx	siyyyyyy	push Closure Compiled block literal index xxxxxxxx (+ Extend A * 256) numCopied yyyyyy receiverOnStack: s = 1 ignoreOuterContext: i = 1"
	| numCopied ignoreContext receiverIsOnStack compiledBlock reg |
	self assert: needsFrame.
	compiledBlock := self getLiteral: byte1 + (extA << 8).
	extA := 0.
	numCopied := byte2 bitAnd: 1<< 6 - 1.
	receiverIsOnStack := byte2 anyMask: 1 << 7.
	ignoreContext := byte2 anyMask: 1 << 6.
	self voidReceiverResultRegContainsSelf.
	self ssAllocateCallReg: ReceiverResultReg
		and: SendNumArgsReg
		and: ClassReg.
	objectRepresentation
		genCreateFullClosure: compiledBlock
		numArgs: (coInterpreter argumentCountOf: compiledBlock)
		numCopied: numCopied
		ignoreContext: ignoreContext
		contextNumArgs: methodOrBlockNumArgs
		large: (coInterpreter methodNeedsLargeContext: methodObj)
		inBlock: inBlock.
	"Closure in ReceiverResultReg"
	1 to: numCopied do:
		[:i| 
		reg := self ssStorePop: true toPreferredReg: TempReg.
		 objectRepresentation
			genStoreSourceReg: reg
			slotIndex: FullClosureFirstCopiedValueIndex + numCopied - i
			intoNewObjectInDestReg: ReceiverResultReg].
	receiverIsOnStack
		ifTrue: [reg := self ssStorePop: true toPreferredReg: TempReg]
		ifFalse: [self simSelf storeToReg: (reg := TempReg)].
	objectRepresentation
			genStoreSourceReg: reg
			slotIndex: FullClosureReceiverIndex
			intoNewObjectInDestReg: ReceiverResultReg.
	self ssPushRegister: ReceiverResultReg.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:07'!
genExtPushRemoteTempOrInstVarLongBytecode
	| index maybeContext |
	^ (byte2 noMask: coInterpreter remoteIsInstVarAccess)
		ifTrue: [ self genPushRemoteTempLongBytecode ]
		ifFalse: 
			[ maybeContext := self extBSpecifiesMaybeContext.
			index := byte1 + (extA << 8).
			extA := 0.
			extB := 0.
			numExtB := 0.
			((coInterpreter isReadMediatedContextInstVarIndex: index) and: [ maybeContext ])
				ifTrue: [ self genPushMaybeContextRemoteInstVar: index inObjectAt: byte2 - coInterpreter remoteIsInstVarAccess ]
				ifFalse: [ self genPushRemoteInstVar: index inObjectAt: byte2 - coInterpreter remoteIsInstVarAccess ] ]! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 4/12/2017 11:26'!
genExtStorePopRemoteTempOrInstVarLongBytecodePopBoolean: boolean
	| index maybeContext needsStoreCheck needsImmCheck |
	needsStoreCheck := self sistaNeedsStoreCheck.
	maybeContext := self extBSpecifiesMaybeContext.
	needsImmCheck := self extBSpecifiesImmCheck.
	extB := 0.
	numExtB := 0.
	(byte2 noMask: coInterpreter remoteIsInstVarAccess)
		ifTrue: 
			[ self 
				genStorePop: boolean 
				RemoteTemp: byte1 
				At: byte2 
				needsStoreCheck: needsStoreCheck.
			self cppIf: IMMUTABILITY ifTrue: [ self annotateBytecode: self Label ] ]
		ifFalse: 
			[index := byte1 + (extA << 8).
			 extA := 0.
			 ((coInterpreter isWriteMediatedContextInstVarIndex: index) and: [ maybeContext ])
				ifTrue: [self 
						genStorePop: boolean 
						MaybeContextRemoteInstVar: index 
						ofObjectAt: byte2 - coInterpreter remoteIsInstVarAccess 
						needsStoreCheck: needsStoreCheck
						needsImmutabilityCheck: needsImmCheck ]
				ifFalse: [self 
						genStorePop: boolean 
						RemoteInstVar: index 
						ofObjectAt: byte2 - coInterpreter remoteIsInstVarAccess 
						needsStoreCheck: needsStoreCheck
						needsImmutabilityCheck: needsImmCheck ].
			self cppIf: IMMUTABILITY ifTrue: [ needsImmCheck ifFalse: [self annotateBytecode: self Label ] ] ].
	^ 0! !

!StackToRegisterMappingCogit methodsFor: 'trampoline support' stamp: 'eem 7/24/2021 18:58'!
genExternalizePointersForPrimitiveCall
	"Override to push the register receiver and register arguments, if any."
	self genPushRegisterArgs.
	^super genExternalizePointersForPrimitiveCall! !

!StackToRegisterMappingCogit methodsFor: 'trampoline support' stamp: 'eem 7/24/2021 18:58'!
genExternalizeStackPointerForFastPrimitiveCall
	"Override to push the register receiver and register arguments, if any."
	self genPushRegisterArgs.
	^super genExternalizeStackPointerForFastPrimitiveCall! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'cb 9/13/2019 11:56'!
genForwardersInlinedIdenticalOrNotIf: orNot
	| nextPC branchDescriptor unforwardRcvr argReg targetBytecodePC
	  unforwardArg  rcvrReg rcvrConstant argConstant postBranchPC label fixup |
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>
	<var: #label type: #'AbstractInstruction *'>
	
	unforwardRcvr := (self ssValue: 1) mayBeAForwarder.
	unforwardArg := self ssTop mayBeAForwarder.
	(unforwardRcvr not and: [unforwardArg not])
		ifTrue: [^self genVanillaInlinedIdenticalOrNotIf: orNot].
	self assert: (unforwardArg or: [unforwardRcvr]).
	"We use reg for non annotable constants to avoid duplicating objRef."
	rcvrConstant := objectRepresentation isUnannotatableConstant: (self ssValue: 1).
	argConstant := objectRepresentation isUnannotatableConstant: self ssTop.
		
	self extractMaybeBranchDescriptorInto: [ :descr :next :postBranch :target | 
		branchDescriptor := descr. nextPC := next. postBranchPC := postBranch. targetBytecodePC := target ].

	self 
		allocateEqualsEqualsRegistersArgNeedsReg: argConstant not 
		rcvrNeedsReg: rcvrConstant not 
		into: [ :rcvr :arg | rcvrReg:= rcvr. argReg := arg ].

	"If not followed by a branch, resolve to true or false."
	(branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse]) ifFalse:
		[^ self 
			genIdenticalNoBranchArgIsConstant: argConstant
			rcvrIsConstant: rcvrConstant
			argReg: argReg 
			rcvrReg: rcvrReg 
			orNotIf: orNot].
	
	"If branching the stack must be flushed for the merge"
	self ssFlushTo: simStackPtr - 2.
	
	label := self Label.
	self genCmpArgIsConstant: argConstant rcvrIsConstant: rcvrConstant argReg: argReg rcvrReg: rcvrReg.
	self ssPop: 2.

	"Since there is a following conditional jump bytecode (unless there is deadCode),
	 define non-merge fixups and leave the cond bytecode to set the mergeness."
	(self fixupAt: nextPC) notAFixup
		ifTrue: "The next instruction is dead.  we can skip it."
			[deadCode := true.
		 	 self ensureFixupAt: targetBytecodePC.
			 self ensureFixupAt: postBranchPC]
		ifFalse:
			[self deny: deadCode]. "push dummy value below"

	orNot == branchDescriptor isBranchTrue "orNot is true for ~~"
		ifFalse: "a == b ifTrue: ... or a ~~ b ifFalse: ... jump on equal to target pc"
			[fixup := self ensureNonMergeFixupAt: postBranchPC.
			 self JumpZero:  (self ensureNonMergeFixupAt: targetBytecodePC)]
		ifTrue: "a == b ifFalse: ... or a ~~ b ifTrue: ... jump on equal to post-branch pc"
			[fixup := self ensureNonMergeFixupAt: targetBytecodePC.
			 self JumpZero: (self ensureNonMergeFixupAt: postBranchPC)].
		
	"The forwarders checks need to jump back to the comparison (label) if a forwarder is found, else 
	 jump forward either to the next forwarder check or to the postBranch or branch target (fixup)."
	(unforwardArg and: [unforwardRcvr]) ifTrue:
		[objectRepresentation genEnsureOopInRegNotForwarded: argReg scratchReg: TempReg jumpBackTo: label].
	objectRepresentation 
		genEnsureOopInRegNotForwarded: (unforwardRcvr ifTrue: [rcvrReg] ifFalse: [argReg]) 
		scratchReg: TempReg 
		ifForwarder: label
		ifNotForwarder: fixup.
		
	"Not reached, execution flow has jumped to fixup"
	deadCode ifFalse:
		[self ssPushConstant: objectMemory trueObject]. "dummy value"
	^0! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator stores' stamp: 'cb 9/5/2016 16:48'!
genGenericStorePop: popBoolean MaybeContextSlotIndex: slotIndex needsStoreCheck: needsStoreCheck needsRestoreRcvr: needsRestoreReceiver needsImmutabilityCheck: needsImmCheck
	"Generates a store into an object that *may* be a context.
	Multiple settings:
	- needsStoreCheck (young into old object check)
	- needRestoreRcvr (ensures the recevier is live across the store)
	- needsImmCheck (do the call-back if the receiver is immutable)"
	<inline: true>
	<var: #mutableJump type: #'AbstractInstruction *'>
	<var: #immutabilityFailure type: #'AbstractInstruction *'>
	| immutabilityFailure mutableJump |
	"The reason we need a frame here is that assigning to an inst var of a context may
	 involve wholesale reorganization of stack pages, and the only way to preserve the
	 execution state of an activation in that case is if it has a frame."
	self assert: needsFrame.
	self 
		cppIf: IMMUTABILITY
		ifTrue:
			[needsImmCheck
				ifTrue: 
					[mutableJump := objectRepresentation genJumpMutable: ReceiverResultReg scratchReg: TempReg.
					 objectRepresentation genStoreTrampolineCall: slotIndex.
					 needsRestoreReceiver ifTrue: [ self putSelfInReceiverResultReg ].
					 immutabilityFailure := self Jump: 0.
					 mutableJump jmpTarget: self Label.]].
	self ssPop: 1.
	self ssAllocateCallReg: ClassReg and: SendNumArgsReg. "for ceStoreContextInstVarTrampoline"
	self ssPush: 1.
	objectRepresentation
		genLoadSlot: SenderIndex
		sourceReg: ReceiverResultReg
		destReg: TempReg.
	self ssStoreAndReplacePop: popBoolean toReg: ClassReg.
	self ssFlushTo: simStackPtr.
	self MoveCq: slotIndex R: SendNumArgsReg.
	self CallRT: ceStoreContextInstVarTrampoline.
	self 
		cppIf: IMMUTABILITY
		ifTrue:
			[needsImmCheck ifTrue:[immutabilityFailure jmpTarget: self Label]].
	^0! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator stores' stamp: 'eem 7/25/2018 11:51'!
genGenericStorePop: popBoolean slotIndex: slotIndex destReg: destReg needsStoreCheck: needsStoreCheck needsRestoreRcvr: needsRestoreReceiver needsImmutabilityCheck: needsImmCheck
	"Generates a store into an object that *cannot* be a context.
	 This code is common between multiple stores (litVar, instVar, remoteInstVar, RemoteTemp)
	 Multiple settings:
	- needsStoreCheck (young into old object check)
	- needRestoreRcvr (ensures the receiver is live across the store)
	- needsImmCheck (do the call-back if the receiver is immutable)"
	"We have two very different paths as only the immutability path requires a specific register 
	for the value on top of stack as well as the stack flush.
	N.B. If IMMUTABILITY then ReceiverResultReg/destReg will be smashed if needsImmCheck.
		If not IMMUTABILITY then ReceiverResultReg will be preserved by the ceStoreCheck trampoline."
	| topReg |
	<inline: true>
	self cppIf: IMMUTABILITY ifTrue:
		[needsImmCheck ifTrue: 
			[self ssAllocateRequiredReg: ClassReg upThrough: simStackPtr - 1. "If already classReg don't spill it"
			 "we replace the top value for the flush"
			 self ssStoreAndReplacePop: popBoolean toReg: ClassReg.
			 self ssFlushTo: simStackPtr.
			 ^objectRepresentation 
				genStoreWithImmutabilityCheckSourceReg: ClassReg 
				slotIndex: slotIndex 
				destReg: destReg 
				scratchReg: TempReg 
				needsStoreCheck: needsStoreCheck 
				needRestoreRcvr: needsRestoreReceiver]].
	topReg := self 
				allocateRegForStackEntryAt: 0 
				notConflictingWith: (self registerMaskFor: destReg). 
	self ssStorePop: popBoolean toReg: topReg.
	^objectRepresentation
		genStoreSourceReg: topReg
		slotIndex: slotIndex
		destReg: destReg
		scratchReg: TempReg
		inFrame: needsFrame
		needsStoreCheck: needsStoreCheck! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'cb 11/24/2016 11:09'!
genIdenticalNoBranchArgIsConstant: argIsConstant rcvrIsConstant: rcvrIsConstant argReg: argReg rcvrReg: rcvrRegOrNone orNotIf: orNot
	"Generates the machine code for #== in the case where the instruction is not followed by a branch"
	| label jumpEqual jumpNotEqual resultReg |
	<var: #label type: #'AbstractInstruction *'>
	<var: #jumpEqual type: #'AbstractInstruction *'>
	<var: #jumpNotEqual type: #'AbstractInstruction *'>
	label := self Label.
	self genCmpArgIsConstant: argIsConstant rcvrIsConstant: rcvrIsConstant argReg: argReg rcvrReg: rcvrRegOrNone.
	self ssPop: 2.
	resultReg := rcvrRegOrNone = NoReg ifTrue: [argReg] ifFalse: [rcvrRegOrNone].
	jumpEqual := self JumpZero: 0.
	 argIsConstant ifFalse:
		[objectRepresentation genEnsureOopInRegNotForwarded: argReg scratchReg: TempReg jumpBackTo: label].
	 rcvrIsConstant ifFalse:
		[objectRepresentation genEnsureOopInRegNotForwarded: rcvrRegOrNone scratchReg: TempReg jumpBackTo: label].
	 orNot ifFalse: [self genMoveFalseR: resultReg] ifTrue: [self genMoveTrueR: resultReg].
	 jumpNotEqual := self Jump: 0.
	 jumpEqual jmpTarget: (orNot ifFalse: [self genMoveTrueR: resultReg] ifTrue: [self genMoveFalseR: resultReg]).
	 jumpNotEqual jmpTarget: self Label.
	 self ssPushRegister: resultReg.
	 ^0! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 12/02/2016 16:00'!
genInlineClosure: startpc numArgs: numArgs numCopied: numCopied
	<inline: true>
	self assert: objectRepresentation getActiveContextAllocatesInMachineCode.
	 self voidReceiverResultRegContainsSelf.
	 self ssAllocateCallReg: ReceiverResultReg
		and: SendNumArgsReg
		and: ClassReg.
	 objectRepresentation
		genNoPopCreateClosureAt: startpc + 1 "1 relative"
		numArgs: numArgs
		numCopied: numCopied
		contextNumArgs: methodOrBlockNumArgs
		large: (coInterpreter methodNeedsLargeContext: methodObj)
		inBlock: inBlock.
	 1 to: numCopied do:
		[:i| | reg |
		 reg := self ssStorePop: true toPreferredReg: TempReg.
		 objectRepresentation
			genStoreSourceReg: reg
			slotIndex: ClosureFirstCopiedValueIndex + numCopied - i
			intoNewObjectInDestReg: ReceiverResultReg].
	 self ssPushRegister: ReceiverResultReg
! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'cb 9/13/2019 12:01'!
genInlinedIdenticalOrNotIf: orNot
	"Decompose code generation for #== into a common constant-folding version,
	 followed by a double dispatch through the objectRepresentation to a version
	 that doesn't deal with forwarders and a version that does."
	| primDescriptor result |
	<var: #primDescriptor type: #'BytecodeDescriptor *'>
	primDescriptor := self generatorAt: byte0.
	
	((objectRepresentation isUnannotatableConstant: self ssTop)
		and: [ objectRepresentation isUnannotatableConstant: (self ssValue: 1)]) ifTrue:
		[self assert: primDescriptor isMapped not.
		 result := (orNot
					ifFalse: [self ssTop constant = (self ssValue: 1) constant]
					ifTrue: [self ssTop constant ~= (self ssValue: 1) constant])
									ifTrue: [objectMemory trueObject]
									ifFalse: [objectMemory falseObject].
		 self ssPop: 2.
		 ^self ssPushConstant: result].

	^objectRepresentation genInlinedIdenticalOrNotIfGuts: orNot! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'cb 4/17/2016 13:33'!
genJumpBackTo: targetBytecodePC
	self ssFlushTo: simStackPtr.
	deadCode := true. "can't fall through"
	^super genJumpBackTo: targetBytecodePC! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/21/2018 15:43'!
genJumpIf: boolean to: targetBytecodePC
	<inline: false>
	| desc fixup ok eventualTarget |
	<var: #desc type: #'CogSimStackEntry *'>
	<var: #fixup type: #'BytecodeFixup *'>
	<var: #ok type: #'AbstractInstruction *'>
	eventualTarget := self eventualTargetOf: targetBytecodePC.
	self ssFlushTo: simStackPtr - 1.
	desc := self ssTop.
	self ssPop: 1.
	(self stackEntryIsBoolean: desc) ifTrue:
		["Must arrange there's a fixup at the target whether it is jumped to or
		  not so that the simStackPtr can be kept correct."
		 fixup := self ensureFixupAt: eventualTarget.
		 "Must annotate the bytecode for correct pc mapping."
		 self annotateBytecode: (desc constant = boolean
									ifTrue: [self Jump: fixup]
									ifFalse: [self prevInstIsPCAnnotated
												ifTrue: [self Nop]
												ifFalse: [self Label]]).
		 extA := 0.
		 ^0].
	desc popToReg: TempReg.
	"Cunning trick by LPD.  If true and false are contiguous subtract the smaller.
	 Correct result is either 0 or the distance between them.  If result is not 0 or
	 their distance send mustBeBoolean."
	self assert: (objectMemory objectAfter: objectMemory falseObject) = objectMemory trueObject.
	self genSubConstant: boolean R: TempReg.
	self JumpZero: (self ensureFixupAt: eventualTarget).
	
	self extASpecifiesNoMustBeBoolean ifTrue: 
		[ extA := 0. 
		self annotateBytecode: self lastOpcode.
		^ 0].
	extA := 0.
	
	self CmpCq: (boolean = objectMemory falseObject
					ifTrue: [objectMemory trueObject - objectMemory falseObject]
					ifFalse: [objectMemory falseObject - objectMemory trueObject])
		R: TempReg.
	ok := self JumpZero: 0.
	self genCallMustBeBooleanFor: boolean.
	ok jmpTarget: (self annotateBytecode: self Label).
	^0! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/21/2018 17:54'!
genJumpTo: targetBytecodePC
	| eventualTarget generator fixup |
	eventualTarget := self eventualTargetOf: targetBytecodePC.
	(eventualTarget > bytecodePC
	 and: [self stackTopIsBoolean
	 and: [(generator := self generatorForPC: eventualTarget) isConditionalBranch]])
		ifTrue:
			[eventualTarget := eventualTarget
							  + generator numBytes
							  + (generator isBranchTrue == (self ssTop constant = objectMemory trueObject)
									ifTrue: [self spanFor: generator at: eventualTarget exts: 0 in: methodObj]
									ifFalse: [0]).
			self ssPop: 1.
			self ssFlushTo: simStackPtr.
			fixup := self ensureFixupAt: eventualTarget.
			self ssPop: -1]
		ifFalse:
			[self ssFlushTo: simStackPtr.
			fixup := self ensureFixupAt: eventualTarget].
	deadCode := true. "can't fall through"
	self Jump: fixup.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 13:09'!
genLoadArgAtDepth: n into: reg
	"All machine code primitives apart from perform: have only
	 register arguments, hence no arg load code is necessary."
	<inline: true>
	self assert: n < objectRepresentation numRegArgs! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 9/9/2019 10:16'!
genLoadLiteralVariable: litVarIndex in: destReg
	<inline: true>
	| association |
	association := self getLiteral: litVarIndex.
	destReg = ReceiverResultReg ifTrue: [self voidReceiverResultRegContainsSelf].
	self ssAllocateRequiredReg: destReg.
	self genMoveConstant: association R: destReg.! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 12/02/2016 16:00'!
genLoadTemp: objectIndex in: destReg
	destReg = ReceiverResultReg ifTrue: [self voidReceiverResultRegContainsSelf].
	self ssAllocateRequiredReg: destReg. 
	self MoveMw: (self frameOffsetOfTemporary: objectIndex) r: FPReg R: destReg.! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeAdd32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self AddR: second R: first.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeAdd64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh first second firstLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger4: [:secondLowValue :secondHighValue :firstLowValue :firstHighValue |
			secondLow := secondLowValue.
			secondHigh := secondHighValue.
			firstLow := firstLowValue.
			firstHigh := firstHighValue.
		].

		self ssNativeTop nativePopToReg: secondLow secondReg: secondHigh.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: firstLow secondReg: firstHigh.
		self ssNativePop: 1.

		self AddR: secondLow R: firstLow.
		self AddcR: secondHigh R: firstHigh.
		self ssPushNativeRegister: firstLow secondRegister: firstHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
			second := secondValue.
			first := firstValue.
		].

		self ssNativeTop nativePopToReg: second.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: first.
		self ssNativePop: 1.

		self AddR: second R: first.
		self ssPushNativeRegister: first.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeAlloca32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| size |
	self allocateRegistersForLowcodeInteger: [:sizeValue |
		size := sizeValue.
	].

	self ssNativeTop nativePopToReg: size.
	self ssNativePop: 1.

	self MoveAw: coInterpreter nativeStackPointerAddress R: TempReg.
	self SubR: size R: TempReg.
	self AndCq: -16 R: TempReg.
	self MoveR: TempReg R: size.
	self MoveR: size Aw: coInterpreter nativeStackPointerAddress.
	self ssPushNativeRegister: size.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeAlloca64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| sizeHigh size sizeLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2: [:sizeLowValue :sizeHighValue |
			sizeLow := sizeLowValue.
			sizeHigh := sizeHighValue.
		].

		self ssNativeTop nativePopToReg: sizeLow secondReg: sizeHigh.
		self ssNativePop: 1.

		self SubR: sizeLow R: SPReg.
		self MoveR: SPReg R: sizeLow.
		self ssPushNativeRegister: sizeLow.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger: [:sizeValue |
			size := sizeValue.
		].

		self ssNativeTop nativePopToReg: size.
		self ssNativePop: 1.

		self SubR: size R: SPReg.
		self MoveR: SPReg R: size.
		self ssPushNativeRegister: size.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeAnd32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self AndR: second R: first.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeAnd64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh first second firstLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger4: [:secondLowValue :secondHighValue :firstLowValue :firstHighValue |
			secondLow := secondLowValue.
			secondHigh := secondHighValue.
			firstLow := firstLowValue.
			firstHigh := firstHighValue.
		].

		self ssNativeTop nativePopToReg: secondLow secondReg: secondHigh.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: firstLow secondReg: firstHigh.
		self ssNativePop: 1.

		self AndR: secondLow R: firstLow.
		self AndR: secondHigh R: firstHigh.
		self ssPushNativeRegister: firstLow secondRegister: firstHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
			second := secondValue.
			first := firstValue.
		].

		self ssNativeTop nativePopToReg: second.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: first.
		self ssNativePop: 1.

		self AndR: second R: first.
		self ssPushNativeRegister: first.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeArithmeticRightShift32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value shiftAmount |
	self allocateRegistersForLowcodeInteger2: [:shiftAmountValue :valueValue |
		shiftAmount := shiftAmountValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: shiftAmount.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self ArithmeticShiftRightR: shiftAmount R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeArithmeticRightShift64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value result shiftAmountHigh shiftAmount resultLow valueLow resultHigh shiftAmountLow valueHigh |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:shiftAmountValue :valueValue :resultValue |
		shiftAmount := shiftAmountValue.
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: shiftAmount.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeBeginCall
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| alignment |
	alignment := extA.

	self beginHighLevelCall: alignment.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators dispatch generated code' stamp: ''!
genLowcodeBinaryInlinePrimitive: prim
	<option: #LowcodeVM>	"Lowcode instruction generator dispatch"
	prim
		caseOf: {
			[0]	-> [ ^ self genLowcodeByteSizeOf ].
			[1]	-> [ ^ self genLowcodeFirstFieldPointer ].
			[2]	-> [ ^ self genLowcodeFirstIndexableFieldPointer ].
			[3]	-> [ ^ self genLowcodeIsBytes ].
			[4]	-> [ ^ self genLowcodeIsFloatObject ].
			[5]	-> [ ^ self genLowcodeIsIndexable ].
			[6]	-> [ ^ self genLowcodeIsIntegerObject ].
			[7]	-> [ ^ self genLowcodeIsPointers ].
			[8]	-> [ ^ self genLowcodeIsWords ].
			[9]	-> [ ^ self genLowcodeIsWordsOrBytes ].
			[10]	-> [ ^ self genLowcodeOopSmallIntegerToInt32 ].
			[11]	-> [ ^ self genLowcodeOopSmallIntegerToInt64 ].
			[12]	-> [ ^ self genLowcodeOopToBoolean32 ].
			[13]	-> [ ^ self genLowcodeOopToBoolean64 ].
			[14]	-> [ ^ self genLowcodeOopToFloat32 ].
			[15]	-> [ ^ self genLowcodeOopToFloat64 ].
			[16]	-> [ ^ self genLowcodeOopToInt32 ].
			[17]	-> [ ^ self genLowcodeOopToInt64 ].
			[18]	-> [ ^ self genLowcodeOopToPointer ].
			[19]	-> [ ^ self genLowcodeOopToPointerReinterpret ].
			[20]	-> [ ^ self genLowcodeOopToUInt32 ].
			[21]	-> [ ^ self genLowcodeOopToUInt64 ].
			[22]	-> [ ^ self genLowcodePin ].
			[23]	-> [ ^ self genLowcodeUnpin ].
		}
		otherwise: [ ^ EncounteredUnknownBytecode ].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeBoolean32ToOop
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value inst cont object trueJump |
	self allocateRegistersForLowcodeIntegerResultOop: [:valueValue :objectValue |
		value := valueValue.
		object := objectValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self CmpCq: 0 R: value.
	trueJump := self JumpNonZero: 0.
	"False"
	self annotate: (self MoveCw: objectMemory falseObject R: value) objRef: objectMemory falseObject.
	cont := self Jump: 0.
	"True"
	inst := self MoveCw: objectMemory trueObject R: value.
	trueJump jmpTarget: inst.
	self annotate: inst objRef: objectMemory trueObject.
	cont jmpTarget: self Label.
	self ssPushRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeBoolean64ToOop
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value inst trueJump2 cont valueLow object trueJump valueHigh |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2ResultOop: [:valueLowValue :valueHighValue :objectValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
			object := objectValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self OrR: valueLow R: valueHigh.
		trueJump := self JumpNonZero: 0.
		"False"
		self annotate: (self MoveCw: objectMemory falseObject R: valueLow) objRef: objectMemory falseObject.
		cont := self Jump: 0.
		"True"
		inst := self MoveCw: objectMemory trueObject R: valueLow.
		trueJump jmpTarget: inst.
		self annotate: inst objRef: objectMemory trueObject.
		cont jmpTarget: self Label.
		self ssPushRegister: valueLow.

	] ifFalse: [
		self allocateRegistersForLowcodeIntegerResultOop: [:valueValue :objectValue |
			value := valueValue.
			object := objectValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self CmpCq: 0 R: value.
		trueJump := self JumpNonZero: 0.
		"False"
		self annotate: (self MoveCw: objectMemory falseObject R: value) objRef: objectMemory falseObject.
		cont := self Jump: 0.
		"True"
		inst := self MoveCw: objectMemory trueObject R: value.
		trueJump jmpTarget: inst.
		self annotate: inst objRef: objectMemory trueObject.
		cont jmpTarget: self Label.
		self ssPushRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeByteSizeOf
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object value |
	self allocateRegistersForLowcodeOopResultInteger: [:objectValue :valueValue |
		object := objectValue.
		value := valueValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	self ssFlushAll.
	objectRepresentation genLcByteSizeOf: object to: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeCallArgumentFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssNativeTop nativeStackPopToReg: DPFPReg0.
	self ssNativePop: 1.
	self MoveRs: DPFPReg0 M32: BytesPerWord negated r: SPReg.
	self SubCq: BytesPerWord R: SPReg.
	currentCallCleanUpSize := currentCallCleanUpSize + BytesPerWord.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeCallArgumentFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssNativeTop nativeStackPopToReg: DPFPReg0.
	self ssNativePop: 1.
	self MoveRd: DPFPReg0 M64: -8 r: SPReg.
	self SubCq: 8 R: SPReg.
	currentCallCleanUpSize := currentCallCleanUpSize + 8.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeCallArgumentInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssNativeTop nativeStackPopToReg: TempReg.
	self ssNativePop: 1.
	self PushR: TempReg.
	currentCallCleanUpSize := currentCallCleanUpSize + BytesPerWord.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeCallArgumentInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	BytesPerWord = 4 ifTrue: [
	self ssNativeTop nativeStackPopToReg: TempReg secondReg: ReceiverResultReg.
	self ssNativePop: 1.
	self PushR: TempReg.
	self PushR: ReceiverResultReg.
	currentCallCleanUpSize := currentCallCleanUpSize + 8.
	] ifFalse: [
	self ssNativeTop nativeStackPopToReg: TempReg.
	self ssNativePop: 1.
	self PushR: TempReg.
	currentCallCleanUpSize := currentCallCleanUpSize + BytesPerWord.
	].

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeCallArgumentPointer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssNativeTop nativeStackPopToReg: TempReg.
	self ssNativePop: 1.
	self PushR: TempReg.
	currentCallCleanUpSize := currentCallCleanUpSize + BytesPerWord.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeCallArgumentSpace
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	"Allocate space"
	self SubCq: extA R: SPReg.
	currentCallCleanUpSize := currentCallCleanUpSize + extA.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeCallArgumentStructure
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	"Fetch the pointer"
	self ssNativeTop nativeStackPopToReg: TempReg.
	self ssNativePop: 1.
	"Allocate space"
	self SubCq: extA R: SPReg.
	currentCallCleanUpSize := currentCallCleanUpSize + extA.
	"Copy the structure"
	backEnd genMemCopy: TempReg to: SPReg constantSize: extA.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeCallInstruction
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| function |
	function := extA.

	self CallRT: function.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeCallPhysical
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| registerID |
	registerID := extA.

	self CallR: registerID.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeCheckSessionIdentifier
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| expectedSession |
	expectedSession := extA.

	self ssPushNativeConstantInt32: (expectedSession = coInterpreter getThisSessionID ifTrue: [1] ifFalse: [0]).

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeCompareAndSwap32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| newValue check oldValue value |
	self allocateRegistersForLowcodeInteger3ResultInteger: [:newValueValue :oldValueValue :checkValue :valueValue |
		newValue := newValueValue.
		oldValue := oldValueValue.
		check := checkValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: newValue.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: oldValue.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: check.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeDiv32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self DivR: second R: first Quo: first Rem: second.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeDiv64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh result resultLow resultHigh first second firstLow |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :resultValue |
		second := secondValue.
		first := firstValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeDuplicateFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value dup2 |
	self allocateRegistersForLowcodeFloatResultFloat: [:valueValue :dup2Value |
		value := valueValue.
		dup2 := dup2Value.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self MoveRs: value Rs: dup2.
	self ssPushNativeRegisterSingleFloat: value;
	ssPushNativeRegisterSingleFloat: dup2.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeDuplicateFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value dup2 |
	self allocateRegistersForLowcodeFloatResultFloat: [:valueValue :dup2Value |
		value := valueValue.
		dup2 := dup2Value.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self MoveRd: value Rd: dup2.
	self ssPushNativeRegisterDoubleFloat: value;
	ssPushNativeRegisterDoubleFloat: dup2.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeDuplicateInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value dup2 |
	self allocateRegistersForLowcodeIntegerResultInteger: [:valueValue :dup2Value |
		value := valueValue.
		dup2 := dup2Value.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self MoveR: value R: dup2.
	self ssPushNativeRegister: value;
	ssPushNativeRegister: dup2.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeDuplicateInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value dup2Low dup2 valueLow dup2High valueHigh |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2ResultInteger2: [:valueLowValue :valueHighValue :dup2LowValue :dup2HighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
			dup2Low := dup2LowValue.
			dup2High := dup2HighValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self MoveR: valueLow R: dup2Low.
		self MoveR: valueHigh R: dup2High.
		self ssPushNativeRegister: valueLow secondRegister: valueHigh.
		self ssPushNativeRegister: dup2Low secondRegister: dup2High.

	] ifFalse: [
		self allocateRegistersForLowcodeIntegerResultInteger: [:valueValue :dup2Value |
			value := valueValue.
			dup2 := dup2Value.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self MoveR: value R: dup2.
		self ssPushNativeRegister: value.
		self ssPushNativeRegister: dup2.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeDuplicatePointer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| dup2 pointerValue |
	self allocateRegistersForLowcodeIntegerResultInteger: [:pointerValueValue :dup2Value |
		pointerValue := pointerValueValue.
		dup2 := dup2Value.
	].

	self ssNativeTop nativePopToReg: pointerValue.
	self ssNativePop: 1.

	self MoveR: pointerValue R: dup2.
	self ssPushNativeRegister: pointerValue;
	ssPushNativeRegister: dup2.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeEffectiveAddress32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| index base offset scale |
	self allocateRegistersForLowcodeInteger4: [:offsetValue :scaleValue :indexValue :baseValue |
		offset := offsetValue.
		scale := scaleValue.
		index := indexValue.
		base := baseValue.
	].

	self ssNativeTop nativePopToReg: offset.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: scale.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: index.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: base.
	self ssNativePop: 1.

	self MulR: scale R: index.
	self AddR: index R: base.
	self AddR: offset R: base.
	self ssPushNativeRegister: base.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeEffectiveAddress64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| base index indexLow indexHigh scaleHigh offsetLow scaleLow offset offsetHigh result scale |
	self allocateRegistersForLowcodeInteger4ResultInteger: [:offsetValue :scaleValue :indexValue :baseValue :resultValue |
		offset := offsetValue.
		scale := scaleValue.
		index := indexValue.
		base := baseValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: offset.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: scale.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: index.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: base.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeEndCall
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self endHighLevelCallWithCleanup.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeEndCallNoCleanup
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self endHighLevelCallWithoutCleanup.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFirstFieldPointer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object |
	self allocateRegistersForLowcodeOop: [:objectValue |
		object := objectValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genLcFirstFieldPointer: object.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFirstIndexableFieldPointer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object |
	self allocateRegistersForLowcodeOop: [:objectValue |
		object := objectValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genLcFirstIndexableFieldPointer: object.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32Add
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeFloat2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self AddRs: second Rs: first.
	self ssPushNativeRegisterSingleFloat: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32Div
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeFloat2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self DivRs: second Rs: first.
	self ssPushNativeRegisterSingleFloat: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32Equal
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value contJump falseJump first second |
	self allocateRegistersForLowcodeFloat2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpRs: second Rs: first.
	falseJump := self JumpFPNotEqual: 0.
	"True result"
	self MoveCq: 1 R: value.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: value.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32Great
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value contJump falseJump first second |
	self allocateRegistersForLowcodeFloat2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpRs: second Rs: first.
	falseJump := self JumpFPLessOrEqual: 0.
	"True result"
	self MoveCq: 1 R: value.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: value.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32GreatEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value contJump falseJump first second |
	self allocateRegistersForLowcodeFloat2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpRs: second Rs: first.
	falseJump := self JumpFPLess: 0.
	"True result"
	self MoveCq: 1 R: value.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: value.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32Less
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value contJump falseJump first second |
	self allocateRegistersForLowcodeFloat2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpRs: second Rs: first.
	falseJump := self JumpFPGreaterOrEqual: 0.
	"True result"
	self MoveCq: 1 R: value.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: value.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32LessEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value contJump falseJump first second |
	self allocateRegistersForLowcodeFloat2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpRs: second Rs: first.
	falseJump := self JumpFPGreater: 0.
	"True result"
	self MoveCq: 1 R: value.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: value.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32Mul
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeFloat2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self MulRs: second Rs: first.
	self ssPushNativeRegisterSingleFloat: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32Neg
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value result |
	self allocateRegistersForLowcodeFloatResultFloat: [:valueValue :resultValue |
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self XorRs: result Rs: result.
	self SubRs: value Rs: result.
	self ssPushNativeRegisterSingleFloat: result.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32NotEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value contJump falseJump first second |
	self allocateRegistersForLowcodeFloat2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpRs: second Rs: first.
	falseJump := self JumpFPEqual: 0.
	"True result"
	self MoveCq: 1 R: value.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: value.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32Sqrt
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeFloat: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self SqrtRs: value.
	self ssPushNativeRegisterSingleFloat: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32Sub
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeFloat2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self SubRs: second Rs: first.
	self ssPushNativeRegisterSingleFloat: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32ToFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| singleFloatValue |
	self allocateRegistersForLowcodeFloat: [:singleFloatValueValue |
		singleFloatValue := singleFloatValueValue.
	].

	self ssNativeTop nativePopToReg: singleFloatValue.
	self ssNativePop: 1.

	self ConvertRs: singleFloatValue Rd: singleFloatValue.
	self ssPushNativeRegisterDoubleFloat: singleFloatValue.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32ToInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value result |
	self allocateRegistersForLowcodeFloatResultInteger: [:valueValue :resultValue |
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self ConvertRs: value R: result.
	self ssPushNativeRegister: result.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32ToInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| resultLow resultHigh value result |
	self allocateRegistersForLowcodeFloatResultInteger: [:valueValue :resultValue |
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32ToOop
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| singleFloatValue object |
	self allocateRegistersForLowcodeFloatResultOop: [:singleFloatValueValue :objectValue |
		singleFloatValue := singleFloatValueValue.
		object := objectValue.
	].

	self ssNativeTop nativePopToReg: singleFloatValue.
	self ssNativePop: 1.

	self ssFlushAll.
	objectRepresentation genLcFloat32: singleFloatValue toOop: object.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32ToUInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value result |
	self allocateRegistersForLowcodeFloatResultInteger: [:valueValue :resultValue |
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self ConvertRs: value R: result.
	self ssPushNativeRegister: result.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat32ToUInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| resultLow resultHigh value result |
	self allocateRegistersForLowcodeFloatResultInteger: [:valueValue :resultValue |
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64Add
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeFloat2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self AddRd: second Rd: first.
	self ssPushNativeRegisterDoubleFloat: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64Div
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeFloat2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self DivRd: second Rd: first.
	self ssPushNativeRegisterDoubleFloat: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64Equal
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value contJump falseJump first second |
	self allocateRegistersForLowcodeFloat2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpRd: second Rd: first.
	falseJump := self JumpFPNotEqual: 0.
	"True result"
	self MoveCq: 1 R: value.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: value.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64Great
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value contJump falseJump first second |
	self allocateRegistersForLowcodeFloat2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpRd: second Rd: first.
	falseJump := self JumpFPLessOrEqual: 0.
	"True result"
	self MoveCq: 1 R: value.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: value.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64GreatEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value contJump falseJump first second |
	self allocateRegistersForLowcodeFloat2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpRd: second Rd: first.
	falseJump := self JumpFPLess: 0.
	"True result"
	self MoveCq: 1 R: value.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: value.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64Less
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value contJump falseJump first second |
	self allocateRegistersForLowcodeFloat2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpRd: second Rd: first.
	falseJump := self JumpFPGreaterOrEqual: 0.
	"True result"
	self MoveCq: 1 R: value.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: value.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64LessEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value contJump falseJump first second |
	self allocateRegistersForLowcodeFloat2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpRd: second Rd: first.
	falseJump := self JumpFPGreater: 0.
	"True result"
	self MoveCq: 1 R: value.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: value.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64Mul
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeFloat2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self MulRd: second Rd: first.
	self ssPushNativeRegisterDoubleFloat: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64Neg
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value result |
	self allocateRegistersForLowcodeFloatResultFloat: [:valueValue :resultValue |
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self XorRd: result Rd: result.
	self SubRd: value Rd: result.
	self ssPushNativeRegisterDoubleFloat: result.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64NotEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value contJump falseJump first second |
	self allocateRegistersForLowcodeFloat2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpRd: second Rd: first.
	falseJump := self JumpFPEqual: 0.
	"True result"
	self MoveCq: 1 R: value.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: value.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64Sqrt
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeFloat: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self SqrtRd: value.
	self ssPushNativeRegisterDoubleFloat: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64Sub
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeFloat2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self SubRd: second Rd: first.
	self ssPushNativeRegisterDoubleFloat: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64ToFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| floatValue |
	self allocateRegistersForLowcodeFloat: [:floatValueValue |
		floatValue := floatValueValue.
	].

	self ssNativeTop nativePopToReg: floatValue.
	self ssNativePop: 1.

	self ConvertRd: floatValue Rs: floatValue.
	self ssPushNativeRegisterSingleFloat: floatValue.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64ToInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| floatValue int32Result |
	self allocateRegistersForLowcodeFloatResultInteger: [:floatValueValue :int32ResultValue |
		floatValue := floatValueValue.
		int32Result := int32ResultValue.
	].

	self ssNativeTop nativePopToReg: floatValue.
	self ssNativePop: 1.

	self ConvertRd: floatValue R: int32Result.
	self ssPushNativeRegister: int32Result.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64ToInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| floatValue int64ResultLow int64Result int64ResultHigh |
	self allocateRegistersForLowcodeFloatResultInteger: [:floatValueValue :int64ResultValue |
		floatValue := floatValueValue.
		int64Result := int64ResultValue.
	].

	self ssNativeTop nativePopToReg: floatValue.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64ToOop
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| floatValue object |
	self allocateRegistersForLowcodeFloatResultOop: [:floatValueValue :objectValue |
		floatValue := floatValueValue.
		object := objectValue.
	].

	self ssNativeTop nativePopToReg: floatValue.
	self ssNativePop: 1.

	self ssFlushAll.
	objectRepresentation genLcFloat64: floatValue toOop: object.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64ToUInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| floatValue int64Result |
	self allocateRegistersForLowcodeFloatResultInteger: [:floatValueValue :int64ResultValue |
		floatValue := floatValueValue.
		int64Result := int64ResultValue.
	].

	self ssNativeTop nativePopToReg: floatValue.
	self ssNativePop: 1.

	self ConvertRd: floatValue R: int64Result.
	self ssPushNativeRegister: int64Result.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFloat64ToUInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| floatValue int64ResultLow int64Result int64ResultHigh |
	self allocateRegistersForLowcodeFloatResultInteger: [:floatValueValue :int64ResultValue |
		floatValue := floatValueValue.
		int64Result := int64ResultValue.
	].

	self ssNativeTop nativePopToReg: floatValue.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeFree
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer |
	self allocateRegistersForLowcodeInteger: [:pointerValue |
		pointer := pointerValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.

	self ssFlushAll.
	pointer ~= ReceiverResultReg ifTrue: [self MoveR: pointer R: ReceiverResultReg ].
	self CallRT: ceFreeTrampoline.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInstantiateIndexable32Oop
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| indexableSize object classOop |
	self allocateRegistersForLowcodeIntegerOopResultOop: [:indexableSizeValue :classOopValue :objectValue |
		indexableSize := indexableSizeValue.
		classOop := classOopValue.
		object := objectValue.
	].

	self ssNativeTop nativePopToReg: indexableSize.
	self ssNativePop: 1.
	self ssTop popToReg: classOop.
	self ssPop: 1.

	self ssFlushAll.
	objectRepresentation genLcInstantiateOop: classOop indexableSize: indexableSize.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInstantiateIndexableOop
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| indexableSize object classOop |
	indexableSize := extA.
	self allocateRegistersForLowcodeOopResultOop: [:classOopValue :objectValue |
		classOop := classOopValue.
		object := objectValue.
	].

	self ssTop popToReg: classOop.
	self ssPop: 1.

	self ssFlushAll.
	objectRepresentation genLcInstantiateOop: classOop constantIndexableSize: indexableSize.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInstantiateOop
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| classOop |
	self allocateRegistersForLowcodeOop: [:classOopValue |
		classOop := classOopValue.
	].

	self ssTop popToReg: classOop.
	self ssPop: 1.

	self ssFlushAll.
	objectRepresentation genLcInstantiateOop: classOop.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt32Equal
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second falseJump contJump first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpR: second R: first.
	falseJump := self JumpNonZero: 0.
	"True result"
	self MoveCq: 1 R: first.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: first.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt32Great
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second falseJump contJump first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpR: second R: first.
	falseJump := self JumpLessOrEqual: 0.
	"True result"
	self MoveCq: 1 R: first.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: first.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt32GreatEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second falseJump contJump first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpR: second R: first.
	falseJump := self JumpLess: 0.
	"True result"
	self MoveCq: 1 R: first.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: first.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt32Less
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second falseJump contJump first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpR: second R: first.
	falseJump := self JumpGreaterOrEqual: 0.
	"True result"
	self MoveCq: 1 R: first.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: first.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt32LessEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second falseJump contJump first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpR: second R: first.
	falseJump := self JumpGreater: 0.
	"True result"
	self MoveCq: 1 R: first.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: first.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt32NotEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second falseJump contJump first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpR: second R: first.
	falseJump := self JumpZero: 0.
	"True result"
	self MoveCq: 1 R: first.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: first.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt32ToFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value result |
	self allocateRegistersForLowcodeIntegerResultFloat: [:valueValue :resultValue |
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self ConvertR: value Rs: result.
	self ssPushNativeRegisterSingleFloat: result.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt32ToFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value result |
	self allocateRegistersForLowcodeIntegerResultFloat: [:valueValue :resultValue |
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self ConvertR: value Rd: result.
	self ssPushNativeRegisterDoubleFloat: result.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt32ToOop
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self ssFlushAll.
	objectRepresentation genLcInt32ToOop: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt32ToPointer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	"TODO: Perform a NOP here"
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt64Equal
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| falseLabel first contJump falseJump secondLow secondHigh second falseJump2 firstHigh firstLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger4: [:secondLowValue :secondHighValue :firstLowValue :firstHighValue |
			secondLow := secondLowValue.
			secondHigh := secondHighValue.
			firstLow := firstLowValue.
			firstHigh := firstHighValue.
		].

		self ssNativeTop nativePopToReg: secondLow secondReg: secondHigh.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: firstLow secondReg: firstHigh.
		self ssNativePop: 1.

		self CmpR: secondHigh R: firstHigh.
		falseJump := self JumpNonZero: 0.
		self CmpR: secondLow R: firstLow.
		falseJump2 := self JumpNonZero: 0.
		"True result"
		self MoveCq: 1 R: firstLow.
		contJump := self Jump: 0.
		"False result"
		falseLabel := self MoveCq: 0 R: firstLow.
		falseJump jmpTarget: falseLabel.
		falseJump2 jmpTarget: falseLabel.
		contJump jmpTarget: self Label.
		self ssPushNativeRegister: firstLow.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
			second := secondValue.
			first := firstValue.
		].

		self ssNativeTop nativePopToReg: second.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: first.
		self ssNativePop: 1.

		self CmpR: second R: first.
		falseJump := self JumpNonZero: 0.
		"True result"
		self MoveCq: 1 R: first.
		contJump := self Jump: 0.
		"False result"
		falseJump jmpTarget: self Label.
		self MoveCq: 0 R: first.
		contJump jmpTarget: self Label.
		self ssPushNativeRegister: first.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt64Great
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh value first second firstLow |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt64GreatEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh value first second firstLow |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt64Less
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh value first second firstLow |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt64LessEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh value first second firstLow |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt64NotEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| falseLabel first contJump falseJump secondLow secondHigh second falseJump2 firstHigh firstLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger4: [:secondLowValue :secondHighValue :firstLowValue :firstHighValue |
			secondLow := secondLowValue.
			secondHigh := secondHighValue.
			firstLow := firstLowValue.
			firstHigh := firstHighValue.
		].

		self ssNativeTop nativePopToReg: secondLow secondReg: secondHigh.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: firstLow secondReg: firstHigh.
		self ssNativePop: 1.

		self CmpR: secondHigh R: firstHigh.
		falseJump := self JumpNonZero: 0.
		self CmpR: secondLow R: firstLow.
		falseJump2 := self JumpNonZero: 0.
		"False result"
		self MoveCq: 0 R: firstLow.
		contJump := self Jump: 0.
		"True result"
		falseLabel := self MoveCq: 1 R: firstLow.
		falseJump jmpTarget: falseLabel.
		falseJump2 jmpTarget: falseLabel.
		contJump jmpTarget: self Label.
		self ssPushNativeRegister: firstLow.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
			second := secondValue.
			first := firstValue.
		].

		self ssNativeTop nativePopToReg: second.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: first.
		self ssNativePop: 1.

		self CmpR: second R: first.
		falseJump := self JumpZero: 0.
		"True result"
		self MoveCq: 1 R: first.
		contJump := self Jump: 0.
		"False result"
		falseJump jmpTarget: self Label.
		self MoveCq: 0 R: first.
		contJump jmpTarget: self Label.
		self ssPushNativeRegister: first.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt64ToFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow result |
	self allocateRegistersForLowcodeIntegerResultFloat: [:valueValue :resultValue |
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt64ToFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow result |
	self allocateRegistersForLowcodeIntegerResultFloat: [:valueValue :resultValue |
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt64ToOop
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2: [:valueLowValue :valueHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self ssFlushAll.
		objectRepresentation genLcInt64ToOop: valueLow highPart: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger: [:valueValue |
			value := valueValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self ssFlushAll.
		objectRepresentation genLcInt64ToOop: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeInt64ToPointer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2: [:valueLowValue :valueHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self ssPushNativeRegister: valueLow.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger: [:valueValue |
			value := valueValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self ssPushNativeRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeIsBytes
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object value |
	self allocateRegistersForLowcodeOopResultInteger: [:objectValue :valueValue |
		object := objectValue.
		value := valueValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genLcIsBytes: object to: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeIsFloatObject
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object value |
	self allocateRegistersForLowcodeOopResultInteger: [:objectValue :valueValue |
		object := objectValue.
		value := valueValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genLcIsFloatObject: object to: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeIsIndexable
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object value |
	self allocateRegistersForLowcodeOopResultInteger: [:objectValue :valueValue |
		object := objectValue.
		value := valueValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genLcIsIndexable: object to: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeIsIntegerObject
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object value |
	self allocateRegistersForLowcodeOopResultInteger: [:objectValue :valueValue |
		object := objectValue.
		value := valueValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genLcIsIntegerObject: object to: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeIsPointers
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object value |
	self allocateRegistersForLowcodeOopResultInteger: [:objectValue :valueValue |
		object := objectValue.
		value := valueValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genLcIsPointers: object to: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeIsWords
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object value |
	self allocateRegistersForLowcodeOopResultInteger: [:objectValue :valueValue |
		object := objectValue.
		value := valueValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genLcIsWords: object to: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeIsWordsOrBytes
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object value |
	self allocateRegistersForLowcodeOopResultInteger: [:objectValue :valueValue |
		object := objectValue.
		value := valueValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genLcIsWordsOrBytes: object to: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLeftShift32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value shiftAmount |
	self allocateRegistersForLowcodeInteger2: [:shiftAmountValue :valueValue |
		shiftAmount := shiftAmountValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: shiftAmount.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self LogicalShiftLeftR: shiftAmount R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLeftShift64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value result shiftAmountHigh shiftAmount resultLow valueLow resultHigh shiftAmountLow valueHigh |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:shiftAmountValue :valueValue :resultValue |
		shiftAmount := shiftAmountValue.
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: shiftAmount.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadArgumentAddress
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:pointerValue |
		pointer := pointerValue.
	].


	self loadNativeArgumentAddress: baseOffset to: pointer.
	self ssPushNativeRegister: pointer.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadArgumentFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| floatValue baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultFloat: [:floatValueValue |
		floatValue := floatValueValue.
	].


	self loadNativeArgumentAddress: baseOffset to: TempReg.
	self MoveM32: 0 r: TempReg Rs: floatValue.
	self ssPushNativeRegisterSingleFloat: floatValue.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadArgumentFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| doubleValue baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultFloat: [:doubleValueValue |
		doubleValue := doubleValueValue.
	].


	self loadNativeArgumentAddress: baseOffset to: TempReg.
	self MoveM64: 0 r: TempReg Rd: doubleValue.
	self ssPushNativeRegisterDoubleFloat: doubleValue.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadArgumentInt16
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:valueValue |
		value := valueValue.
	].


	self loadNativeArgumentAddress: baseOffset to: TempReg.
	self MoveM16: 0 r: TempReg R: value.
	self SignExtend16R: value R: value.
	self ssPushNativeRegister: value.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadArgumentInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:valueValue |
		value := valueValue.
	].


	self loadNativeArgumentAddress: baseOffset to: TempReg.
	self MoveM32: 0 r: TempReg R: value.
	self ssPushNativeRegister: value.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadArgumentInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow baseOffset |
	baseOffset := extA.
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeResultInteger2: [:valueLowValue :valueHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].


		"TODO: Check the endianness"
		self loadNativeArgumentAddress: baseOffset to: TempReg.
		self MoveM32: 0 r: TempReg R: valueLow.
		self MoveM32: 4 r: TempReg R: valueHigh.
		self ssPushNativeRegister: valueLow secondRegister: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeResultInteger: [:valueValue |
			value := valueValue.
		].


		self loadNativeArgumentAddress: baseOffset to: TempReg.
		self MoveM64: 0 r: TempReg R: value.
		self ssPushNativeRegister: value.

	].
		extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadArgumentInt8
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:valueValue |
		value := valueValue.
	].


	self loadNativeArgumentAddress: baseOffset to: TempReg.
	self MoveM8: 0 r: TempReg R: value.
	self SignExtend8R: value R: value.
	self ssPushNativeRegister: value.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadArgumentPointer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointerResult baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:pointerResultValue |
		pointerResult := pointerResultValue.
	].


	self loadNativeArgumentAddress: baseOffset to: TempReg.
	self MoveMw: 0 r: TempReg R: pointerResult.
	self ssPushNativeRegister: pointerResult.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadArgumentUInt16
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:valueValue |
		value := valueValue.
	].


	self loadNativeArgumentAddress: baseOffset to: TempReg.
	self MoveM16: 0 r: TempReg R: value.
	self ssPushNativeRegister: value.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadArgumentUInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:valueValue |
		value := valueValue.
	].


	self loadNativeArgumentAddress: baseOffset to: TempReg.
	self MoveM32: 0 r: TempReg R: value.
	self ssPushNativeRegister: value.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadArgumentUInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow baseOffset |
	baseOffset := extA.
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeResultInteger2: [:valueLowValue :valueHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].


		"TODO: Check the endianness"
		self loadNativeArgumentAddress: baseOffset to: TempReg.
		self MoveM32: 0 r: TempReg R: valueLow.
		self MoveM32: 4 r: TempReg R: valueHigh.
		self ssPushNativeRegister: valueLow secondRegister: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeResultInteger: [:valueValue |
			value := valueValue.
		].


		self loadNativeArgumentAddress: baseOffset to: TempReg.
		self MoveM64: 0 r: TempReg R: value.
		self ssPushNativeRegister: value.

	].
		extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadArgumentUInt8
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:valueValue |
		value := valueValue.
	].


	self loadNativeArgumentAddress: baseOffset to: TempReg.
	self MoveM8: 0 r: TempReg R: value.
	self ssPushNativeRegister: value.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadFloat32FromMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer value |
	self allocateRegistersForLowcodeIntegerResultFloat: [:pointerValue :valueValue |
		pointer := pointerValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.

	self MoveM32: 0 r: pointer Rs: value.
	self ssPushNativeRegisterSingleFloat: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadFloat64FromMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer value |
	self allocateRegistersForLowcodeIntegerResultFloat: [:pointerValue :valueValue |
		pointer := pointerValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.

	self MoveM64: 0 r: pointer Rd: value.
	self ssPushNativeRegisterDoubleFloat: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadInt16FromMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer value |
	self allocateRegistersForLowcodeIntegerResultInteger: [:pointerValue :valueValue |
		pointer := pointerValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.

	self MoveM16: 0 r: pointer R: value.
	self SignExtend16R: value R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadInt32FromMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer value |
	self allocateRegistersForLowcodeIntegerResultInteger: [:pointerValue :valueValue |
		pointer := pointerValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.

	self MoveM32: 0 r: pointer R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadInt64FromMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh pointer value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeIntegerResultInteger2: [:pointerValue :valueLowValue :valueHighValue |
			pointer := pointerValue.
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssNativeTop nativePopToReg: pointer.
		self ssNativePop: 1.

		"TODO: Check the endianness"
		self MoveM32: 0 r: pointer R: valueLow.
		self MoveM32: 4 r: pointer R: valueHigh.
		self ssPushNativeRegister: valueLow secondRegister: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeIntegerResultInteger: [:pointerValue :valueValue |
			pointer := pointerValue.
			value := valueValue.
		].

		self ssNativeTop nativePopToReg: pointer.
		self ssNativePop: 1.

		self MoveM64: 0 r: pointer R: value.
		self ssPushNativeRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadInt8FromMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer value |
	self allocateRegistersForLowcodeIntegerResultInteger: [:pointerValue :valueValue |
		pointer := pointerValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.

	self MoveM8: 0 r: pointer R: value.
	self SignExtend8R: value R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadLocalAddress
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:pointerValue |
		pointer := pointerValue.
	].


	self loadNativeLocalAddress: baseOffset to: pointer.
	self ssPushNativeRegister: pointer.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadLocalFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| floatValue baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultFloat: [:floatValueValue |
		floatValue := floatValueValue.
	].


	self loadNativeLocalAddress: baseOffset to: TempReg.
	self MoveM32: 0 r: TempReg Rs: floatValue.
	self ssPushNativeRegisterSingleFloat: floatValue.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadLocalFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| doubleValue baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultFloat: [:doubleValueValue |
		doubleValue := doubleValueValue.
	].


	self loadNativeLocalAddress: baseOffset to: TempReg.
	self MoveM64: 0 r: TempReg Rd: doubleValue.
	self ssPushNativeRegisterDoubleFloat: doubleValue.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadLocalInt16
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:valueValue |
		value := valueValue.
	].


	self loadNativeLocalAddress: baseOffset to: TempReg.
	self MoveM16: 0 r: TempReg R: value.
	self SignExtend16R: value R: value.
	self ssPushNativeRegister: value.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadLocalInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:valueValue |
		value := valueValue.
	].


	self loadNativeLocalAddress: baseOffset to: TempReg.
	self MoveM32: 0 r: TempReg R: value.
	self ssPushNativeRegister: value.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadLocalInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow baseOffset |
	baseOffset := extA.
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeResultInteger2: [:valueLowValue :valueHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].


		"TODO: Check the endianness"
		self loadNativeLocalAddress: baseOffset to: TempReg.
		self MoveM32: 0 r: TempReg R: valueLow.
		self MoveM32: 4 r: TempReg R: valueHigh.
		self ssPushNativeRegister: valueLow secondRegister: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeResultInteger: [:valueValue |
			value := valueValue.
		].


		self loadNativeLocalAddress: baseOffset to: TempReg.
		self MoveM64: 0 r: TempReg R: value.
		self ssPushNativeRegister: value.

	].
		extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadLocalInt8
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:valueValue |
		value := valueValue.
	].


	self loadNativeLocalAddress: baseOffset to: TempReg.
	self MoveM8: 0 r: TempReg R: value.
	self SignExtend8R: value R: value.
	self ssPushNativeRegister: value.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadLocalPointer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointerResult baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:pointerResultValue |
		pointerResult := pointerResultValue.
	].


	self loadNativeLocalAddress: baseOffset to: TempReg.
	self MoveMw: 0 r: TempReg R: pointerResult.
	self ssPushNativeRegister: pointerResult.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadLocalUInt16
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:valueValue |
		value := valueValue.
	].


	self loadNativeLocalAddress: baseOffset to: TempReg.
	self MoveM16: 0 r: TempReg R: value.
	self ssPushNativeRegister: value.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadLocalUInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:valueValue |
		value := valueValue.
	].


	self loadNativeLocalAddress: baseOffset to: TempReg.
	self MoveM32: 0 r: TempReg R: value.
	self ssPushNativeRegister: value.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadLocalUInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow baseOffset |
	baseOffset := extA.
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeResultInteger2: [:valueLowValue :valueHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].


		"TODO: Check the endianness"
		self loadNativeLocalAddress: baseOffset to: TempReg.
		self MoveM32: 0 r: TempReg R: valueLow.
		self MoveM32: 4 r: TempReg R: valueHigh.
		self ssPushNativeRegister: valueLow secondRegister: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeResultInteger: [:valueValue |
			value := valueValue.
		].


		self loadNativeLocalAddress: baseOffset to: TempReg.
		self MoveM64: 0 r: TempReg R: value.
		self ssPushNativeRegister: value.

	].
		extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadLocalUInt8
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeResultInteger: [:valueValue |
		value := valueValue.
	].


	self loadNativeLocalAddress: baseOffset to: TempReg.
	self MoveM8: 0 r: TempReg R: value.
	self ssPushNativeRegister: value.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadObjectAt
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| fieldIndex object |
	self allocateRegistersForLowcodeIntegerOop: [:fieldIndexValue :objectValue |
		fieldIndex := fieldIndexValue.
		object := objectValue.
	].

	self ssNativeTop nativePopToReg: fieldIndex.
	self ssNativePop: 1.
	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genLcLoadObject: object at: fieldIndex.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadObjectField
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| fieldIndex object |
	fieldIndex := extA.
	self allocateRegistersForLowcodeOop: [:objectValue |
		object := objectValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genLcLoadObject: object field: fieldIndex.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadPointerFromMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointerResult pointer |
	self allocateRegistersForLowcodeIntegerResultInteger: [:pointerValue :pointerResultValue |
		pointer := pointerValue.
		pointerResult := pointerResultValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.

	self MoveMw: 0 r: pointer R: pointerResult.
	self ssPushNativeRegister: pointerResult.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadUInt16FromMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer value |
	self allocateRegistersForLowcodeIntegerResultInteger: [:pointerValue :valueValue |
		pointer := pointerValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.

	self MoveM16: 0 r: pointer R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadUInt32FromMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer value |
	self allocateRegistersForLowcodeIntegerResultInteger: [:pointerValue :valueValue |
		pointer := pointerValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.

	self MoveM32: 0 r: pointer R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadUInt64FromMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh pointer value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeIntegerResultInteger2: [:pointerValue :valueLowValue :valueHighValue |
			pointer := pointerValue.
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssNativeTop nativePopToReg: pointer.
		self ssNativePop: 1.

		"TODO: Check the endianness"
		self MoveM32: 0 r: pointer R: valueLow.
		self MoveM32: 4 r: pointer R: valueHigh.
		self ssPushNativeRegister: valueLow secondRegister: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeIntegerResultInteger: [:pointerValue :valueValue |
			pointer := pointerValue.
			value := valueValue.
		].

		self ssNativeTop nativePopToReg: pointer.
		self ssNativePop: 1.

		self MoveM64: 0 r: pointer R: value.
		self ssPushNativeRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLoadUInt8FromMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer value |
	self allocateRegistersForLowcodeIntegerResultInteger: [:pointerValue :valueValue |
		pointer := pointerValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.

	self MoveM8: 0 r: pointer R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLocalFrameSize
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| alignedSize size |
	size := extA.

	self assert: needsFrame.
	hasNativeFrame := true.
	"Align the size to 16 bytes."
	alignedSize := size + 15 bitAnd: -16.
	"Mark the stack frame"
	self annotate: (self MoveCw: (objectMemory splObj: LowcodeContextMark) R: TempReg) objRef: (objectMemory splObj: LowcodeContextMark).
	self MoveR: TempReg Mw: self frameOffsetOfNativeFrameMark r: FPReg.
	"Fetch the stack"
	self MoveAw: coInterpreter nativeStackPointerAddress R: TempReg.
	self AddCq: 1 R: TempReg.
	self MoveR: TempReg Mw: self frameOffsetOfPreviousNativeStackPointer r: FPReg.
	"Store the frame pointer"
	self SubCq: alignedSize R: TempReg.
	self MoveR: TempReg Mw: self frameOffsetOfNativeFramePointer r: FPReg.
	"Store the new stack pointer"
	self MoveR: TempReg Mw: self frameOffsetOfNativeStackPointer r: FPReg.
	"Allocate space for the locals"
	self SubCq: 1 + coInterpreter defaultNativeStackFrameSize R: TempReg.
	self MoveR: TempReg Aw: coInterpreter nativeStackPointerAddress.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLockRegisters
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssFlushAll.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeLockVM
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeMalloc32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer size |
	self allocateRegistersForLowcodeIntegerResultInteger: [:sizeValue :pointerValue |
		size := sizeValue.
		pointer := pointerValue.
	].

	self ssNativeTop nativePopToReg: size.
	self ssNativePop: 1.

	self ssFlushAll.
	size ~= ReceiverResultReg ifTrue: [self MoveR: size R: ReceiverResultReg ].
	self CallRT: ceMallocTrampoline.
	self MoveR: TempReg R: pointer.
	self ssPushNativeRegister: pointer.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeMalloc64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| sizeHigh size sizeLow pointer |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2ResultInteger: [:sizeLowValue :sizeHighValue :pointerValue |
			sizeLow := sizeLowValue.
			sizeHigh := sizeHighValue.
			pointer := pointerValue.
		].

		self ssNativeTop nativePopToReg: sizeLow secondReg: sizeHigh.
		self ssNativePop: 1.

		self ssFlushAll.
		sizeLow ~= ReceiverResultReg ifTrue: [self MoveR: sizeLow R: ReceiverResultReg ].
		self CallRT: ceMallocTrampoline.
		self MoveR: TempReg R: pointer.
		self ssPushNativeRegister: pointer.

	] ifFalse: [
		self allocateRegistersForLowcodeIntegerResultInteger: [:sizeValue :pointerValue |
			size := sizeValue.
			pointer := pointerValue.
		].

		self ssNativeTop nativePopToReg: size.
		self ssNativePop: 1.

		self ssFlushAll.
		size ~= ReceiverResultReg ifTrue: [self MoveR: size R: ReceiverResultReg ].
		self CallRT: ceMallocTrampoline.
		self MoveR: TempReg R: pointer.
		self ssPushNativeRegister: pointer.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeMemcpy32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| source size dest |
	self allocateRegistersForLowcodeInteger3: [:sizeValue :sourceValue :destValue |
		size := sizeValue.
		source := sourceValue.
		dest := destValue.
	].

	self ssNativeTop nativePopToReg: size.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: source.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: dest.
	self ssNativePop: 1.

	self ssFlushAll.
	backEnd genMemCopy: source to: dest size: size.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeMemcpy64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| source sizeLow sizeHigh dest size |
	self allocateRegistersForLowcodeInteger3: [:sizeValue :sourceValue :destValue |
		size := sizeValue.
		source := sourceValue.
		dest := destValue.
	].

	self ssNativeTop nativePopToReg: size.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: source.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: dest.
	self ssNativePop: 1.

	self ssFlushAll.
	backEnd genMemCopy: source to: dest size: sizeLow.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeMemcpyFixed
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| source size dest |
	size := extA.
	self allocateRegistersForLowcodeInteger2: [:sourceValue :destValue |
		source := sourceValue.
		dest := destValue.
	].

	self ssNativeTop nativePopToReg: source.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: dest.
	self ssNativePop: 1.

	size = BytesPerWord ifTrue: [
	self MoveMw: 0 r: source R: TempReg.
	self MoveR: TempReg Mw: 0 r: dest.
	] ifFalse: [
	self ssFlushAll.
	backEnd genMemCopy: source to: dest constantSize: size.
	].

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeMoveFloat32ToPhysical
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssNativeTop nativeStackPopToReg: extA.
	self ssNativePop: 1.
	currentCallCleanUpSize := currentCallCleanUpSize + BytesPerWord.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeMoveFloat64ToPhysical
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssNativeTop nativeStackPopToReg: extA.
	self ssNativePop: 1.
	currentCallCleanUpSize := currentCallCleanUpSize + 8.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeMoveInt32ToPhysical
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssNativeTop nativeStackPopToReg: extA.
	self ssNativePop: 1.
	currentCallCleanUpSize := currentCallCleanUpSize + BytesPerWord.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeMoveInt64ToPhysical
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssNativeTop nativeStackPopToReg: extA.
	self ssNativePop: 1.
	currentCallCleanUpSize := currentCallCleanUpSize + 8.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeMovePointerToPhysical
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssNativeTop nativeStackPopToReg: extA.
	self ssNativePop: 1.
	currentCallCleanUpSize := currentCallCleanUpSize + BytesPerWord.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeMul32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self MulR: second R: first.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeMul64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh result resultLow resultHigh first second firstLow |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :resultValue |
		second := secondValue.
		first := firstValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeNeg32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self NegateR: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeNeg64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2: [:valueLowValue :valueHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		"Two complement negation"
		self NotR: valueLow.
		self NotR: valueHigh.
		self AddCq: 1 R: valueLow.
		self AddcCq: 0 R: valueHigh.
		self ssPushNativeRegister: valueLow secondRegister: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger: [:valueValue |
			value := valueValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self NegateR: value.
		self ssPushNativeRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeNot32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self NotR: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeNot64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2: [:valueLowValue :valueHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self NotR: valueLow.
		self NotR: valueHigh.
		self ssPushNativeRegister: valueLow secondRegister: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger: [:valueValue |
			value := valueValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self NotR: value.
		self ssPushNativeRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators dispatch generated code' stamp: ''!
genLowcodeNullaryInlinePrimitive: prim
	<option: #LowcodeVM>	"Lowcode instruction generator dispatch"
	prim
		caseOf: {
			[0]	-> [ ^ self genLowcodeBoolean32ToOop ].
			[1]	-> [ ^ self genLowcodeBoolean64ToOop ].
			[2]	-> [ ^ self genLowcodeFloat32ToOop ].
			[3]	-> [ ^ self genLowcodeFloat64ToOop ].
			[4]	-> [ ^ self genLowcodeInt32ToOop ].
			[5]	-> [ ^ self genLowcodeInt64ToOop ].
			[6]	-> [ ^ self genLowcodePointerToOop ].
			[7]	-> [ ^ self genLowcodePointerToOopReinterprer ].
			[8]	-> [ ^ self genLowcodeSmallInt32ToOop ].
			[9]	-> [ ^ self genLowcodeUint32ToOop ].
			[10]	-> [ ^ self genLowcodeUint64ToOop ].
		}
		otherwise: [ ^ EncounteredUnknownBytecode ].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOopEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value contJump falseJump first second |
	self allocateRegistersForLowcodeOop2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssTop popToReg: second.
	self ssPop: 1.
	self ssTop popToReg: first.
	self ssPop: 1.

	self CmpR: second R: first.
	falseJump := self JumpNonZero: 0.
	"True result"
	self MoveCq: 1 R: first.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: first.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOopNotEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second falseJump contJump first |
	self allocateRegistersForLowcodeOop2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssTop popToReg: second.
	self ssPop: 1.
	self ssTop popToReg: first.
	self ssPop: 1.

	self CmpR: second R: first.
	falseJump := self JumpZero: 0.
	"True result"
	self MoveCq: 1 R: first.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: first.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOopSmallIntegerToInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object |
	self allocateRegistersForLowcodeOop: [:objectValue |
		object := objectValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genConvertSmallIntegerToIntegerInReg: object.
	self ssPushNativeRegister: object.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOopSmallIntegerToInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh object value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeOopResultInteger2: [:objectValue :valueLowValue :valueHighValue |
			object := objectValue.
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssTop popToReg: object.
		self ssPop: 1.

		objectRepresentation genConvertSmallIntegerToIntegerInReg: object.
		self MoveCq: 0 R: valueHigh.
		self ssPushNativeRegister: object secondRegister: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeOopResultInteger: [:objectValue :valueValue |
			object := objectValue.
			value := valueValue.
		].

		self ssTop popToReg: object.
		self ssPop: 1.

		objectRepresentation genConvertSmallIntegerToIntegerInReg: object.
		self ssPushNativeRegister: object.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOopToBoolean32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object |
	self allocateRegistersForLowcodeOop: [:objectValue |
		object := objectValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	self annotate: (self SubCw: objectMemory falseObject R: object) objRef: objectMemory falseObject.
	self ssPushNativeRegister: object.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOopToBoolean64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh object value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeOopResultInteger2: [:objectValue :valueLowValue :valueHighValue |
			object := objectValue.
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssTop popToReg: object.
		self ssPop: 1.

		self MoveCq: 0 R: valueHigh.
		self annotate: (self SubCw: objectMemory falseObject R: object) objRef: objectMemory falseObject.
		self ssPushNativeRegister: object secondRegister: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeOopResultInteger: [:objectValue :valueValue |
			object := objectValue.
			value := valueValue.
		].

		self ssTop popToReg: object.
		self ssPop: 1.

		self annotate: (self SubCw: objectMemory falseObject R: object) objRef: objectMemory falseObject.
		self ssPushNativeRegister: object.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOopToFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object value |
	self allocateRegistersForLowcodeOopResultFloat: [:objectValue :valueValue |
		object := objectValue.
		value := valueValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	self ssFlushAll.
	objectRepresentation genLcOop: object toFloat32: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOopToFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object value |
	self allocateRegistersForLowcodeOopResultFloat: [:objectValue :valueValue |
		object := objectValue.
		value := valueValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	self ssFlushAll.
	objectRepresentation genLcOop: object toFloat64: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOopToInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object |
	self allocateRegistersForLowcodeOop: [:objectValue |
		object := objectValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	self ssFlushAll.
	objectRepresentation genLcOopToInt32: object.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOopToInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh object value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeOopResultInteger2: [:objectValue :valueLowValue :valueHighValue |
			object := objectValue.
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssTop popToReg: object.
		self ssPop: 1.

		self ssFlushAll.
		objectRepresentation genLcOop: object toInt64: valueLow highPart: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeOopResultInteger: [:objectValue :valueValue |
			object := objectValue.
			value := valueValue.
		].

		self ssTop popToReg: object.
		self ssPop: 1.

		self ssFlushAll.
		objectRepresentation genLcOopToInt64: object.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOopToPointer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object |
	self allocateRegistersForLowcodeOop: [:objectValue |
		object := objectValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genLcOopToPointer: object.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOopToPointerReinterpret
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object pointer |
	self allocateRegistersForLowcodeOopResultInteger: [:objectValue :pointerValue |
		object := objectValue.
		pointer := pointerValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	"TODO: Generate a nop here"
	self ssPushNativeRegister: object.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOopToUInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object value |
	self allocateRegistersForLowcodeOopResultInteger: [:objectValue :valueValue |
		object := objectValue.
		value := valueValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	self ssFlushAll.
	objectRepresentation genLcOopToUInt32: object.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOopToUInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh object value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeOopResultInteger2: [:objectValue :valueLowValue :valueHighValue |
			object := objectValue.
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssTop popToReg: object.
		self ssPop: 1.

		self ssFlushAll.
		objectRepresentation genLcOop: object toUInt64: valueLow highPart: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeOopResultInteger: [:objectValue :valueValue |
			object := objectValue.
			value := valueValue.
		].

		self ssTop popToReg: object.
		self ssPop: 1.

		self ssFlushAll.
		objectRepresentation genLcOopToUInt64: object.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOr32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second result first |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :resultValue |
		second := secondValue.
		first := firstValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self OrR: second R: first.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeOr64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh first second firstLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger4: [:secondLowValue :secondHighValue :firstLowValue :firstHighValue |
			secondLow := secondLowValue.
			secondHigh := secondHighValue.
			firstLow := firstLowValue.
			firstHigh := firstHighValue.
		].

		self ssNativeTop nativePopToReg: secondLow secondReg: secondHigh.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: firstLow secondReg: firstHigh.
		self ssNativePop: 1.

		self OrR: secondLow R: firstLow.
		self OrR: secondHigh R: firstHigh.
		self ssPushNativeRegister: firstLow secondRegister: firstHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
			second := secondValue.
			first := firstValue.
		].

		self ssNativeTop nativePopToReg: second.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: first.
		self ssNativePop: 1.

		self OrR: second R: first.
		self ssPushNativeRegister: first.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePerformCallFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"

	self callSwitchToCStack.
	self MoveCw: extA R: TempReg.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: self Label.
	backEnd cFloatResultToRs: DPFPReg0.
	self ssPushNativeRegisterSingleFloat: DPFPReg0.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePerformCallFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"

	self callSwitchToCStack.
	self MoveCw: extA R: TempReg.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: self Label.
	backEnd cFloatResultToRd: DPFPReg0.
	self ssPushNativeRegisterDoubleFloat: DPFPReg0.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePerformCallIndirectFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"

	self ssNativeTop nativeStackPopToReg: TempReg.
	self ssNativePop: 1.
	self callSwitchToCStack.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: self Label.
	backEnd cFloatResultToRs: DPFPReg0.
	self ssPushNativeRegisterSingleFloat: DPFPReg0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePerformCallIndirectFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"

	self ssNativeTop nativeStackPopToReg: TempReg.
	self ssNativePop: 1.
	self callSwitchToCStack.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: self Label.
	backEnd cFloatResultToRd: DPFPReg0.
	self ssPushNativeRegisterDoubleFloat: DPFPReg0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePerformCallIndirectInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"

	self ssNativeTop nativeStackPopToReg: TempReg.
	self ssNativePop: 1.
	self callSwitchToCStack.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: self Label.
	self MoveR: ABIResultReg R: ReceiverResultReg.
	self ssPushNativeRegister: ReceiverResultReg.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: 'eem 2/4/2020 11:47'!
genLowcodePerformCallIndirectInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"

	self ssNativeTop nativeStackPopToReg: TempReg.
	self ssNativePop: 1.
	self callSwitchToCStack.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: (self MoveR: ABIResultReg R: ReceiverResultReg).
	BytesPerWord = 4
		ifTrue:
			[self MoveR: ABIResultRegHigh R: Arg0Reg.
			 self ssPushNativeRegister: ReceiverResultReg secondRegister: Arg0Reg]
		ifFalse:
			[self MoveR: ABIResultReg R: ReceiverResultReg.
			 self ssPushNativeRegister: ReceiverResultReg].
	^0! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePerformCallIndirectPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"

	self ssNativeTop nativeStackPopToReg: TempReg.
	self ssNativePop: 1.
	self callSwitchToCStack.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: self Label.
	self MoveR: ABIResultReg R: ReceiverResultReg.
	self ssPushNativeRegister: ReceiverResultReg.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePerformCallIndirectStructure
	<option: #LowcodeVM>	"Lowcode instruction generator"

	"Push the result space"
	self ssNativeTop nativeStackPopToReg: TempReg.
	self ssNativePop: 1.
	self PushR: TempReg.
	"Fetch the function pointer"
	self ssNativeTop nativeStackPopToReg: TempReg.
	self ssNativePop: 1.
	"Call the function"
	self callSwitchToCStack.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: self Label.
	"Fetch the result"
	self MoveR: ABIResultReg R: ReceiverResultReg.
	self ssPushNativeRegister: ReceiverResultReg.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePerformCallIndirectVoid
	<option: #LowcodeVM>	"Lowcode instruction generator"

	self ssNativeTop nativeStackPopToReg: TempReg.
	self ssNativePop: 1.
	self callSwitchToCStack.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: self Label.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePerformCallInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"

	self callSwitchToCStack.
	self MoveCw: extA R: TempReg.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: self Label.
	self MoveR: ABIResultReg R: ReceiverResultReg.
	self ssPushNativeRegister: ReceiverResultReg.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: 'eem 2/4/2020 11:48'!
genLowcodePerformCallInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"

	self callSwitchToCStack.
	self MoveCw: extA R: TempReg.
	extA := 0.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: (self MoveR: ABIResultReg R: ReceiverResultReg).
	BytesPerWord = 4
		ifTrue:
			[self MoveR: ABIResultRegHigh R: Arg0Reg.
			 self ssPushNativeRegister: ReceiverResultReg secondRegister: Arg0Reg]
		ifFalse:
			[self MoveR: ABIResultReg R: ReceiverResultReg.
			 self ssPushNativeRegister: ReceiverResultReg].
	^0! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePerformCallPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"

	self callSwitchToCStack.
	self MoveCw: extA R: TempReg.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: self Label.
	self MoveR: ABIResultReg R: ReceiverResultReg.
	self ssPushNativeRegister: ReceiverResultReg.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePerformCallStructure
	<option: #LowcodeVM>	"Lowcode instruction generator"

	"Push the result space"
	self ssNativeTop nativeStackPopToReg: TempReg.
	self ssNativePop: 1.
	self PushR: TempReg.
	"Call the function"
	self callSwitchToCStack.
	self MoveCw: extA R: TempReg.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: self Label.
	"Fetch the result"
	self MoveR: ABIResultReg R: ReceiverResultReg.
	self ssPushNativeRegister: ReceiverResultReg.
	extA := 0.
	extB := 0.
	numExtB := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePerformCallVoid
	<option: #LowcodeVM>	"Lowcode instruction generator"

	self callSwitchToCStack.
	self MoveCw: extA R: TempReg.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: self Label.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePerformCallout
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self callSwitchToCStack.
	self MoveCw: extA R: TempReg.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: self Label.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePerformCalloutIndirect
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssNativeTop nativeStackPopToReg: TempReg.
	self ssNativePop: 1.
	self callSwitchToCStack.
	self CallRT: ceFFICalloutTrampoline.
	self annotateBytecode: self Label.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePin
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object |
	self allocateRegistersForLowcodeOop: [:objectValue |
		object := objectValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePlaftormCode
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePointerAddConstantOffset
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| base offset |
	offset := extB.
	self allocateRegistersForLowcodeInteger: [:baseValue |
		base := baseValue.
	].

	self ssNativeTop nativePopToReg: base.
	self ssNativePop: 1.

	self AddCq: offset R: base.
	self ssPushNativeRegister: base.

	extB := 0.
	numExtB := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePointerAddOffset32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| base offset |
	self allocateRegistersForLowcodeInteger2: [:offsetValue :baseValue |
		offset := offsetValue.
		base := baseValue.
	].

	self ssNativeTop nativePopToReg: offset.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: base.
	self ssNativePop: 1.

	self AddR: offset R: base.
	self ssPushNativeRegister: base.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePointerAddOffset64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| offsetHigh base offset offsetLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger3: [:offsetLowValue :offsetHighValue :baseValue |
			offsetLow := offsetLowValue.
			offsetHigh := offsetHighValue.
			base := baseValue.
		].

		self ssNativeTop nativePopToReg: offsetLow secondReg: offsetHigh.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: base.
		self ssNativePop: 1.

		self AddR: offsetLow R: base.
		self ssPushNativeRegister: base.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger2: [:offsetValue :baseValue |
			offset := offsetValue.
			base := baseValue.
		].

		self ssNativeTop nativePopToReg: offset.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: base.
		self ssNativePop: 1.

		self AddR: offset R: base.
		self ssPushNativeRegister: base.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePointerEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second falseJump contJump first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpR: second R: first.
	falseJump := self JumpNonZero: 0.
	"True result"
	self MoveCq: 1 R: first.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: first.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePointerNotEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second falseJump contJump first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpR: second R: first.
	falseJump := self JumpZero: 0.
	"True result"
	self MoveCq: 1 R: first.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: first.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePointerToInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer |
	self allocateRegistersForLowcodeInteger: [:pointerValue |
		pointer := pointerValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.

	"TODO: Perform a NOP here"
	self ssPushNativeRegister: pointer.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePointerToInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| resultLow pointer result resultHigh |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeIntegerResultInteger2: [:pointerValue :resultLowValue :resultHighValue |
			pointer := pointerValue.
			resultLow := resultLowValue.
			resultHigh := resultHighValue.
		].

		self ssNativeTop nativePopToReg: pointer.
		self ssNativePop: 1.

		self MoveR: pointer R: resultLow.
		self MoveCq: 0 R: resultHigh.
		self ssPushNativeRegister: resultLow secondRegister: resultHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeIntegerResultInteger: [:pointerValue :resultValue |
			pointer := pointerValue.
			result := resultValue.
		].

		self ssNativeTop nativePopToReg: pointer.
		self ssNativePop: 1.

		self ssPushNativeRegister: pointer.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePointerToOop
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer pointerClassLiteral |
	pointerClassLiteral := self getLiteral: extA.
	self allocateRegistersForLowcodeInteger: [:pointerValue |
		pointer := pointerValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.

	self ssFlushAll.
	objectRepresentation genLcPointerToOop: pointer class: pointerClassLiteral.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePointerToOopReinterprer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer |
	self allocateRegistersForLowcodeInteger: [:pointerValue |
		pointer := pointerValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.

	"TODO: Generate a nop here"
	self ssPushRegister: pointer.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePopFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeFloat: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.


	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePopFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeFloat: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.


	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePopInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.


	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePopInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow |
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.


	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePopMultipleNative
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssPopNativeSize: extA.
	extA := 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePopPointer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointerValue |
	self allocateRegistersForLowcodeInteger: [:pointerValueValue |
		pointerValue := pointerValueValue.
	].

	self ssNativeTop nativePopToReg: pointerValue.
	self ssNativePop: 1.


	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushCalloutResultFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	backEnd cFloatResultToRs: DPFPReg0.
	self ssPushNativeRegisterSingleFloat: DPFPReg0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushCalloutResultFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	backEnd cFloatResultToRd: DPFPReg0.
	self ssPushNativeRegisterDoubleFloat: DPFPReg0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushCalloutResultInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self MoveR: ABIResultReg R: ReceiverResultReg.
	self ssPushNativeRegister: ReceiverResultReg.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: 'eem 2/4/2020 11:49'!
genLowcodePushCalloutResultInt64
	<option: #LowcodeVM>
	<inline: true>

	self MoveR: ABIResultReg R: ReceiverResultReg.
	BytesPerWord = 4
		ifTrue:
			[self MoveR: ABIResultRegHigh R: Arg0Reg.
			 self ssPushNativeRegister: ReceiverResultReg secondRegister: Arg0Reg]
		ifFalse:
			[self MoveR: ABIResultReg R: ReceiverResultReg.
			 self ssPushNativeRegister: ReceiverResultReg].
	^0! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushCalloutResultPointer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self MoveR: ABIResultReg R: ReceiverResultReg.
	self ssPushNativeRegister: ReceiverResultReg.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushConstantUInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| constant |
	constant := extA.

	self ssPushNativeConstantInt32: constant.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushConstantUInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| constant |
	constant := extA.

	self ssPushNativeConstantInt64: constant.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushNullPointer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssPushNativeConstantPointer: 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushOne32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssPushNativeConstantInt32: 1.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushOne64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssPushNativeConstantInt64: 1.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushOneFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssPushNativeConstantFloat32: 1.0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushOneFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssPushNativeConstantFloat64: 1.0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushPhysicalFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| registerID |
	registerID := extA.

	self abort.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushPhysicalFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| registerID |
	registerID := extA.

	self abort.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushPhysicalInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| registerID |
	registerID := extA.

	self abort.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushPhysicalInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| registerID |
	registerID := extA.

	self abort.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushPhysicalPointer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| registerID |
	registerID := extA.

	self abort.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushSessionIdentifier
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssPushNativeConstantInt32: coInterpreter getThisSessionID.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushZero32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssPushNativeConstantInt32: 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushZero64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssPushNativeConstantInt64: 0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushZeroFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssPushNativeConstantFloat32: 0.0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodePushZeroFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self ssPushNativeConstantFloat64: 0.0.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeRem32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self DivR: second R: first Quo: second Rem: first.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeRem64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh result resultLow resultHigh first second firstLow |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :resultValue |
		second := secondValue.
		first := firstValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeRightShift32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value shiftAmount |
	self allocateRegistersForLowcodeInteger2: [:shiftAmountValue :valueValue |
		shiftAmount := shiftAmountValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: shiftAmount.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self LogicalShiftRightR: shiftAmount R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeRightShift64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value result shiftAmountHigh shiftAmount resultLow valueLow resultHigh shiftAmountLow valueHigh |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:shiftAmountValue :valueValue :resultValue |
		shiftAmount := shiftAmountValue.
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: shiftAmount.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeSignExtend32From16
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self SignExtend16R: value R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeSignExtend32From8
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self SignExtend8R: value R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeSignExtend64From16
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value isNegative cont valueLow valueHigh |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2: [:valueLowValue :valueHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self SignExtend16R: valueLow R: valueLow.
		"Check the sign to set the high word"
		self CmpCq: 0 R: valueLow.
		"Positive"
		isNegative := self JumpLess: 0.
		self MoveCq: 0 R: valueHigh.
		cont := self Jump: 0.
		"Negative"
		isNegative jmpTarget: (self MoveCq: -1 R: valueHigh).
		cont jmpTarget: self Label.
		self ssPushNativeRegister: valueLow secondRegister: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger: [:valueValue |
			value := valueValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self SignExtend16R: value R: value.
		self ssPushNativeRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeSignExtend64From32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value result resultLow resultHigh isNegative cont |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeIntegerResultInteger2: [:valueValue :resultLowValue :resultHighValue |
			value := valueValue.
			resultLow := resultLowValue.
			resultHigh := resultHighValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self MoveR: value R: resultLow.
		"Check the sign to set the high word"
		self CmpCq: 0 R: value.
		"Positive"
		isNegative := self JumpLess: 0.
		self MoveCq: 0 R: resultHigh.
		cont := self Jump: 0.
		"Negative"
		isNegative jmpTarget: (self MoveCq: -1 R: resultHigh).
		cont jmpTarget: self Label.
		self ssPushNativeRegister: resultLow secondRegister: resultHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeIntegerResultInteger: [:valueValue :resultValue |
			value := valueValue.
			result := resultValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self SignExtend32R: value R: value.
		self ssPushNativeRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeSignExtend64From8
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value result resultLow resultHigh isNegative valueLow cont valueHigh |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2ResultInteger2: [:valueLowValue :valueHighValue :resultLowValue :resultHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
			resultLow := resultLowValue.
			resultHigh := resultHighValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self SignExtend8R: valueLow R: valueLow.
		"Check the sign to set the high word"
		self CmpCq: 0 R: valueLow.
		"Positive"
		isNegative := self JumpLess: 0.
		self MoveCq: 0 R: valueHigh.
		cont := self Jump: 0.
		"Negative"
		isNegative jmpTarget: (self MoveCq: -1 R: valueHigh).
		cont jmpTarget: self Label.
		self ssPushNativeRegister: valueLow secondRegister: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeIntegerResultInteger: [:valueValue :resultValue |
			value := valueValue.
			result := resultValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self ZeroExtend16R: value R: value.
		self ssPushNativeRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeSmallInt32ToOop
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	objectRepresentation genConvertIntegerToSmallIntegerInReg: value.
	self ssPushRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreFloat32ToMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| floatValue pointer |
	self allocateRegistersForLowcodeFloatInteger: [:floatValueValue :pointerValue |
		floatValue := floatValueValue.
		pointer := pointerValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: floatValue.
	self ssNativePop: 1.

	self MoveRs: floatValue M32: 0 r: pointer.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreFloat64ToMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| doubleValue pointer |
	self allocateRegistersForLowcodeFloatInteger: [:doubleValueValue :pointerValue |
		doubleValue := doubleValueValue.
		pointer := pointerValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: doubleValue.
	self ssNativePop: 1.

	self MoveRd: doubleValue M64: 0 r: pointer.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreInt16ToMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer value |
	self allocateRegistersForLowcodeInteger2: [:pointerValue :valueValue |
		pointer := pointerValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self MoveR: value R: TempReg.
	self MoveR: TempReg M16: 0 r: pointer.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreInt32ToMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer value |
	self allocateRegistersForLowcodeInteger2: [:pointerValue :valueValue |
		pointer := pointerValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self MoveR: value M32: 0 r: pointer.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreInt64ToMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh pointer value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger3: [:pointerValue :valueLowValue :valueHighValue |
			pointer := pointerValue.
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssNativeTop nativePopToReg: pointer.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self MoveR: valueLow M32: 0 r: pointer.
		self MoveR: valueHigh M32: 4 r: pointer.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger2: [:pointerValue :valueValue |
			pointer := pointerValue.
			value := valueValue.
		].

		self ssNativeTop nativePopToReg: pointer.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self MoveR: value M64: 0 r: pointer.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreInt8ToMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointer value |
	self allocateRegistersForLowcodeInteger2: [:pointerValue :valueValue |
		pointer := pointerValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: pointer.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self MoveR: value R: TempReg.
	self MoveR: TempReg M8: 0 r: pointer.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreLocalFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeFloat: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self loadNativeLocalAddress: baseOffset to: TempReg.
	self MoveRs: value M32: 0 r: TempReg.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreLocalFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeFloat: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self loadNativeLocalAddress: baseOffset to: TempReg.
	self MoveRd: value M64: 0 r: TempReg.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreLocalInt16
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self MoveR: value R: TempReg.
	self loadNativeLocalAddress: baseOffset to: value.
	self MoveR: TempReg M16: 0 r: value.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreLocalInt32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self loadNativeLocalAddress: baseOffset to: TempReg.
	self MoveR: value M32: 0 r: TempReg.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreLocalInt64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow baseOffset |
	baseOffset := extA.
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2: [:valueLowValue :valueHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self loadNativeLocalAddress: baseOffset to: TempReg.
		self MoveR: valueLow M32: 0 r: TempReg.
		self MoveR: valueHigh M32: 4 r: TempReg.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger: [:valueValue |
			value := valueValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self MoveR: value M64: 0 r: TempReg.

	].
		extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreLocalInt8
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self MoveR: value R: TempReg.
	self loadNativeLocalAddress: baseOffset to: value.
	self MoveR: TempReg M8: 0 r: value.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreLocalPointer
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| pointerValue baseOffset |
	baseOffset := extA.
	self allocateRegistersForLowcodeInteger: [:pointerValueValue |
		pointerValue := pointerValueValue.
	].

	self ssNativeTop nativePopToReg: pointerValue.
	self ssNativePop: 1.

	self loadNativeLocalAddress: baseOffset to: TempReg.
	self MoveR: pointerValue Mw: 0 r: TempReg.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreObjectField
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| fieldIndex object value |
	fieldIndex := extA.
	self allocateRegistersForLowcodeOop2: [:valueValue :objectValue |
		value := valueValue.
		object := objectValue.
	].

	self ssTop popToReg: value.
	self ssPop: 1.
	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genLcStore: value object: object field: fieldIndex.

	extA := 0.
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStoreObjectFieldAt
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| fieldIndex object value |
	self allocateRegistersForLowcodeIntegerOop2: [:fieldIndexValue :valueValue :objectValue |
		fieldIndex := fieldIndexValue.
		value := valueValue.
		object := objectValue.
	].

	self ssTop popToReg: value.
	self ssPop: 1.
	self ssNativeTop nativePopToReg: fieldIndex.
	self ssNativePop: 1.
	self ssTop popToReg: object.
	self ssPop: 1.

	objectRepresentation genLcStore: value object: object at: fieldIndex.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeStorePointerToMemory
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| memoryPointer pointerValue |
	self allocateRegistersForLowcodeInteger2: [:memoryPointerValue :pointerValueValue |
		memoryPointer := memoryPointerValue.
		pointerValue := pointerValueValue.
	].

	self ssNativeTop nativePopToReg: memoryPointer.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: pointerValue.
	self ssNativePop: 1.

	self MoveR: pointerValue Mw: 0 r: memoryPointer.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeSub32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self SubR: second R: first.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeSub64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh first second firstLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger4: [:secondLowValue :secondHighValue :firstLowValue :firstHighValue |
			secondLow := secondLowValue.
			secondHigh := secondHighValue.
			firstLow := firstLowValue.
			firstHigh := firstHighValue.
		].

		self ssNativeTop nativePopToReg: secondLow secondReg: secondHigh.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: firstLow secondReg: firstHigh.
		self ssNativePop: 1.

		self SubR: secondLow R: firstLow.
		self SubbR: secondHigh R: firstHigh.
		self ssPushNativeRegister: firstLow secondRegister: firstHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
			second := secondValue.
			first := firstValue.
		].

		self ssNativeTop nativePopToReg: second.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: first.
		self ssNativePop: 1.

		self SubR: second R: first.
		self ssPushNativeRegister: first.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators dispatch generated code' stamp: ''!
genLowcodeTrinaryInlinePrimitive: prim
	<option: #LowcodeVM>	"Lowcode instruction generator dispatch"
	prim
		caseOf: {
			[0]	-> [ ^ self genLowcodeOopEqual ].
			[1]	-> [ ^ self genLowcodeOopNotEqual ].
			[2]	-> [ ^ self genLowcodeStoreObjectField ].
			[3]	-> [ ^ self genLowcodeStoreObjectFieldAt ].
		}
		otherwise: [ ^ EncounteredUnknownBytecode ].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeTruncate32To16
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self AndCq: 16rFFFF R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeTruncate32To8
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self AndCq: 16rFF R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeTruncate64To16
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2: [:valueLowValue :valueHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self AndCq: 16rFFFF R: valueLow.
		self ssPushNativeRegister: valueLow.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger: [:valueValue |
			value := valueValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self AndCq: 16rFFFF R: value.
		self ssPushNativeRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeTruncate64To32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2: [:valueLowValue :valueHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self ssPushNativeRegister: valueLow.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger: [:valueValue |
			value := valueValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self AndCq: 16rFFFFFFFF R: value.
		self ssPushNativeRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeTruncate64To8
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow result |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2ResultInteger: [:valueLowValue :valueHighValue :resultValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
			result := resultValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self AndCq: 16rFF R: valueLow.
		self ssPushNativeRegister: valueLow.

	] ifFalse: [
		self allocateRegistersForLowcodeIntegerResultInteger: [:valueValue :resultValue |
			value := valueValue.
			result := resultValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self AndCq: 16rFF R: value.
		self ssPushNativeRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUdiv32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self DivR: second R: first Quo: first Rem: second.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUdiv64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh result resultLow resultHigh first second firstLow |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :resultValue |
		second := secondValue.
		first := firstValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUint32Great
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second falseJump contJump first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpR: second R: first.
	falseJump := self JumpBelowOrEqual: 0.
	"True result"
	self MoveCq: 1 R: first.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: first.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUint32GreatEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second falseJump contJump first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpR: second R: first.
	falseJump := self JumpBelow: 0.
	"True result"
	self MoveCq: 1 R: first.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: first.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUint32Less
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second falseJump contJump first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpR: second R: first.
	falseJump := self JumpAboveOrEqual: 0.
	"True result"
	self MoveCq: 1 R: first.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: first.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUint32LessEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second falseJump contJump first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self CmpR: second R: first.
	falseJump := self JumpAbove: 0.
	"True result"
	self MoveCq: 1 R: first.
	contJump := self Jump: 0.
	"False result"
	falseJump jmpTarget: self Label.
	self MoveCq: 0 R: first.
	contJump jmpTarget: self Label.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUint32ToFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value result |
	self allocateRegistersForLowcodeIntegerResultFloat: [:valueValue :resultValue |
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self ConvertR: value Rs: result.
	self ssPushNativeRegisterSingleFloat: result.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUint32ToFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value result |
	self allocateRegistersForLowcodeIntegerResultFloat: [:valueValue :resultValue |
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self ConvertR: value Rd: result.
	self ssPushNativeRegisterDoubleFloat: result.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUint32ToOop
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self ssFlushAll.
	objectRepresentation genLcUInt32ToOop: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUint64Great
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh value first second firstLow |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUint64GreatEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh value first second firstLow |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUint64Less
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh value first second firstLow |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUint64LessEqual
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh value first second firstLow |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :valueValue |
		second := secondValue.
		first := firstValue.
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUint64ToFloat32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow result |
	self allocateRegistersForLowcodeIntegerResultFloat: [:valueValue :resultValue |
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUint64ToFloat64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow result |
	self allocateRegistersForLowcodeIntegerResultFloat: [:valueValue :resultValue |
		value := valueValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUint64ToOop
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh object value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2ResultOop: [:valueLowValue :valueHighValue :objectValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
			object := objectValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self ssFlushAll.
		objectRepresentation genLcUInt64ToOop: valueLow highPart: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeIntegerResultOop: [:valueValue :objectValue |
			value := valueValue.
			object := objectValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self ssFlushAll.
		objectRepresentation genLcUInt64ToOop: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUmul32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self MulR: second R: first.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUmul64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh result resultLow resultHigh first second firstLow |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :resultValue |
		second := secondValue.
		first := firstValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators dispatch generated code' stamp: ''!
genLowcodeUnaryInlinePrimitive2: prim
	<option: #LowcodeVM>	"Lowcode instruction generator dispatch"
	prim
		caseOf: {
			[60]	-> [ ^ self genLowcodeFloat64ToFloat32 ].
			[61]	-> [ ^ self genLowcodeFloat64ToInt32 ].
			[62]	-> [ ^ self genLowcodeFloat64ToInt64 ].
			[63]	-> [ ^ self genLowcodeFloat64ToUInt32 ].
			[64]	-> [ ^ self genLowcodeFloat64ToUInt64 ].
			[65]	-> [ ^ self genLowcodeFree ].
			[66]	-> [ ^ self genLowcodeInstantiateIndexable32Oop ].
			[67]	-> [ ^ self genLowcodeInstantiateIndexableOop ].
			[68]	-> [ ^ self genLowcodeInstantiateOop ].
			[69]	-> [ ^ self genLowcodeInt32Equal ].
			[70]	-> [ ^ self genLowcodeInt32Great ].
			[71]	-> [ ^ self genLowcodeInt32GreatEqual ].
			[72]	-> [ ^ self genLowcodeInt32Less ].
			[73]	-> [ ^ self genLowcodeInt32LessEqual ].
			[74]	-> [ ^ self genLowcodeInt32NotEqual ].
			[75]	-> [ ^ self genLowcodeInt32ToFloat32 ].
			[76]	-> [ ^ self genLowcodeInt32ToFloat64 ].
			[77]	-> [ ^ self genLowcodeInt32ToPointer ].
			[78]	-> [ ^ self genLowcodeInt64Equal ].
			[79]	-> [ ^ self genLowcodeInt64Great ].
			[80]	-> [ ^ self genLowcodeInt64GreatEqual ].
			[81]	-> [ ^ self genLowcodeInt64Less ].
			[82]	-> [ ^ self genLowcodeInt64LessEqual ].
			[83]	-> [ ^ self genLowcodeInt64NotEqual ].
			[84]	-> [ ^ self genLowcodeInt64ToFloat32 ].
			[85]	-> [ ^ self genLowcodeInt64ToFloat64 ].
			[86]	-> [ ^ self genLowcodeInt64ToPointer ].
			[87]	-> [ ^ self genLowcodeLeftShift32 ].
			[88]	-> [ ^ self genLowcodeLeftShift64 ].
			[89]	-> [ ^ self genLowcodeLoadArgumentAddress ].
			[90]	-> [ ^ self genLowcodeLoadArgumentFloat32 ].
			[91]	-> [ ^ self genLowcodeLoadArgumentFloat64 ].
			[92]	-> [ ^ self genLowcodeLoadArgumentInt16 ].
			[93]	-> [ ^ self genLowcodeLoadArgumentInt32 ].
			[94]	-> [ ^ self genLowcodeLoadArgumentInt64 ].
			[95]	-> [ ^ self genLowcodeLoadArgumentInt8 ].
			[96]	-> [ ^ self genLowcodeLoadArgumentPointer ].
			[97]	-> [ ^ self genLowcodeLoadArgumentUInt16 ].
			[98]	-> [ ^ self genLowcodeLoadArgumentUInt32 ].
			[99]	-> [ ^ self genLowcodeLoadArgumentUInt64 ].
			[100]	-> [ ^ self genLowcodeLoadArgumentUInt8 ].
			[101]	-> [ ^ self genLowcodeLoadFloat32FromMemory ].
			[102]	-> [ ^ self genLowcodeLoadFloat64FromMemory ].
			[103]	-> [ ^ self genLowcodeLoadInt16FromMemory ].
			[104]	-> [ ^ self genLowcodeLoadInt32FromMemory ].
			[105]	-> [ ^ self genLowcodeLoadInt64FromMemory ].
			[106]	-> [ ^ self genLowcodeLoadInt8FromMemory ].
			[107]	-> [ ^ self genLowcodeLoadLocalAddress ].
			[108]	-> [ ^ self genLowcodeLoadLocalFloat32 ].
			[109]	-> [ ^ self genLowcodeLoadLocalFloat64 ].
			[110]	-> [ ^ self genLowcodeLoadLocalInt16 ].
			[111]	-> [ ^ self genLowcodeLoadLocalInt32 ].
			[112]	-> [ ^ self genLowcodeLoadLocalInt64 ].
			[113]	-> [ ^ self genLowcodeLoadLocalInt8 ].
			[114]	-> [ ^ self genLowcodeLoadLocalPointer ].
			[115]	-> [ ^ self genLowcodeLoadLocalUInt16 ].
			[116]	-> [ ^ self genLowcodeLoadLocalUInt32 ].
			[117]	-> [ ^ self genLowcodeLoadLocalUInt64 ].
			[118]	-> [ ^ self genLowcodeLoadLocalUInt8 ].
			[119]	-> [ ^ self genLowcodeLoadObjectAt ].
		}
		otherwise: [ ^ self genLowcodeUnaryInlinePrimitive3: prim ].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators dispatch generated code' stamp: ''!
genLowcodeUnaryInlinePrimitive3: prim
	<option: #LowcodeVM>	"Lowcode instruction generator dispatch"
	prim
		caseOf: {
			[120]	-> [ ^ self genLowcodeLoadObjectField ].
			[121]	-> [ ^ self genLowcodeLoadPointerFromMemory ].
			[122]	-> [ ^ self genLowcodeLoadUInt16FromMemory ].
			[123]	-> [ ^ self genLowcodeLoadUInt32FromMemory ].
			[124]	-> [ ^ self genLowcodeLoadUInt64FromMemory ].
			[125]	-> [ ^ self genLowcodeLoadUInt8FromMemory ].
			[126]	-> [ ^ self genLowcodeLocalFrameSize ].
			[127]	-> [ ^ self genLowcodeLockRegisters ].
			[128]	-> [ ^ self genLowcodeLockVM ].
			[129]	-> [ ^ self genLowcodeMalloc32 ].
			[130]	-> [ ^ self genLowcodeMalloc64 ].
			[131]	-> [ ^ self genLowcodeMemcpy32 ].
			[132]	-> [ ^ self genLowcodeMemcpy64 ].
			[133]	-> [ ^ self genLowcodeMemcpyFixed ].
			[134]	-> [ ^ self genLowcodeMoveFloat32ToPhysical ].
			[135]	-> [ ^ self genLowcodeMoveFloat64ToPhysical ].
			[136]	-> [ ^ self genLowcodeMoveInt32ToPhysical ].
			[137]	-> [ ^ self genLowcodeMoveInt64ToPhysical ].
			[138]	-> [ ^ self genLowcodeMovePointerToPhysical ].
			[139]	-> [ ^ self genLowcodeMul32 ].
			[140]	-> [ ^ self genLowcodeMul64 ].
			[141]	-> [ ^ self genLowcodeNeg32 ].
			[142]	-> [ ^ self genLowcodeNeg64 ].
			[143]	-> [ ^ self genLowcodeNot32 ].
			[144]	-> [ ^ self genLowcodeNot64 ].
			[145]	-> [ ^ self genLowcodeOr32 ].
			[146]	-> [ ^ self genLowcodeOr64 ].
			[147]	-> [ ^ self genLowcodePerformCallout ].
			[148]	-> [ ^ self genLowcodePerformCalloutIndirect ].
			[149]	-> [ ^ self genLowcodePushCalloutResultFloat32 ].
			[150]	-> [ ^ self genLowcodePushCalloutResultFloat64 ].
			[151]	-> [ ^ self genLowcodePushCalloutResultInt32 ].
			[152]	-> [ ^ self genLowcodePushCalloutResultInt64 ].
			[153]	-> [ ^ self genLowcodePushCalloutResultPointer ].
			[161]	-> [ ^ self genLowcodePlaftormCode ].
			[162]	-> [ ^ self genLowcodePointerAddConstantOffset ].
			[163]	-> [ ^ self genLowcodePointerAddOffset32 ].
			[164]	-> [ ^ self genLowcodePointerAddOffset64 ].
			[165]	-> [ ^ self genLowcodePointerEqual ].
			[166]	-> [ ^ self genLowcodePointerNotEqual ].
			[167]	-> [ ^ self genLowcodePointerToInt32 ].
			[168]	-> [ ^ self genLowcodePointerToInt64 ].
			[169]	-> [ ^ self genLowcodePopFloat32 ].
			[170]	-> [ ^ self genLowcodePopFloat64 ].
			[171]	-> [ ^ self genLowcodePopInt32 ].
			[172]	-> [ ^ self genLowcodePopInt64 ].
			[173]	-> [ ^ self genLowcodePopMultipleNative ].
			[174]	-> [ ^ self genLowcodePopPointer ].
			[175]	-> [ ^ self genLowcodePushConstantUInt32 ].
			[176]	-> [ ^ self genLowcodePushConstantUInt64 ].
			[177]	-> [ ^ self genLowcodePushNullPointer ].
			[178]	-> [ ^ self genLowcodePushOne32 ].
			[179]	-> [ ^ self genLowcodePushOne64 ].
			[180]	-> [ ^ self genLowcodePushOneFloat32 ].
			[181]	-> [ ^ self genLowcodePushOneFloat64 ].
			[182]	-> [ ^ self genLowcodePushPhysicalFloat32 ].
			[183]	-> [ ^ self genLowcodePushPhysicalFloat64 ].
			[184]	-> [ ^ self genLowcodePushPhysicalInt32 ].
			[185]	-> [ ^ self genLowcodePushPhysicalInt64 ].
			[186]	-> [ ^ self genLowcodePushPhysicalPointer ].
		}
		otherwise: [ ^ self genLowcodeUnaryInlinePrimitive4: prim ].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators dispatch generated code' stamp: ''!
genLowcodeUnaryInlinePrimitive4: prim
	<option: #LowcodeVM>	"Lowcode instruction generator dispatch"
	prim
		caseOf: {
			[187]	-> [ ^ self genLowcodePushSessionIdentifier ].
			[188]	-> [ ^ self genLowcodePushZero32 ].
			[189]	-> [ ^ self genLowcodePushZero64 ].
			[190]	-> [ ^ self genLowcodePushZeroFloat32 ].
			[191]	-> [ ^ self genLowcodePushZeroFloat64 ].
			[192]	-> [ ^ self genLowcodeRem32 ].
			[193]	-> [ ^ self genLowcodeRem64 ].
			[194]	-> [ ^ self genLowcodeRightShift32 ].
			[195]	-> [ ^ self genLowcodeRightShift64 ].
			[196]	-> [ ^ self genLowcodeSignExtend32From16 ].
			[197]	-> [ ^ self genLowcodeSignExtend32From8 ].
			[198]	-> [ ^ self genLowcodeSignExtend64From16 ].
			[199]	-> [ ^ self genLowcodeSignExtend64From32 ].
			[200]	-> [ ^ self genLowcodeSignExtend64From8 ].
			[201]	-> [ ^ self genLowcodeStoreFloat32ToMemory ].
			[202]	-> [ ^ self genLowcodeStoreFloat64ToMemory ].
			[203]	-> [ ^ self genLowcodeStoreInt16ToMemory ].
			[204]	-> [ ^ self genLowcodeStoreInt32ToMemory ].
			[205]	-> [ ^ self genLowcodeStoreInt64ToMemory ].
			[206]	-> [ ^ self genLowcodeStoreInt8ToMemory ].
			[207]	-> [ ^ self genLowcodeStoreLocalFloat32 ].
			[208]	-> [ ^ self genLowcodeStoreLocalFloat64 ].
			[209]	-> [ ^ self genLowcodeStoreLocalInt16 ].
			[210]	-> [ ^ self genLowcodeStoreLocalInt32 ].
			[211]	-> [ ^ self genLowcodeStoreLocalInt64 ].
			[212]	-> [ ^ self genLowcodeStoreLocalInt8 ].
			[213]	-> [ ^ self genLowcodeStoreLocalPointer ].
			[214]	-> [ ^ self genLowcodeStorePointerToMemory ].
			[215]	-> [ ^ self genLowcodeSub32 ].
			[216]	-> [ ^ self genLowcodeSub64 ].
			[217]	-> [ ^ self genLowcodeTruncate32To16 ].
			[218]	-> [ ^ self genLowcodeTruncate32To8 ].
			[219]	-> [ ^ self genLowcodeTruncate64To16 ].
			[220]	-> [ ^ self genLowcodeTruncate64To32 ].
			[221]	-> [ ^ self genLowcodeTruncate64To8 ].
			[222]	-> [ ^ self genLowcodeUdiv32 ].
			[223]	-> [ ^ self genLowcodeUdiv64 ].
			[224]	-> [ ^ self genLowcodeUint32Great ].
			[225]	-> [ ^ self genLowcodeUint32GreatEqual ].
			[226]	-> [ ^ self genLowcodeUint32Less ].
			[227]	-> [ ^ self genLowcodeUint32LessEqual ].
			[228]	-> [ ^ self genLowcodeUint32ToFloat32 ].
			[229]	-> [ ^ self genLowcodeUint32ToFloat64 ].
			[230]	-> [ ^ self genLowcodeUint64Great ].
			[231]	-> [ ^ self genLowcodeUint64GreatEqual ].
			[232]	-> [ ^ self genLowcodeUint64Less ].
			[233]	-> [ ^ self genLowcodeUint64LessEqual ].
			[234]	-> [ ^ self genLowcodeUint64ToFloat32 ].
			[235]	-> [ ^ self genLowcodeUint64ToFloat64 ].
			[236]	-> [ ^ self genLowcodeUmul32 ].
			[237]	-> [ ^ self genLowcodeUmul64 ].
			[238]	-> [ ^ self genLowcodeUnlockRegisters ].
			[239]	-> [ ^ self genLowcodeUnlockVM ].
			[240]	-> [ ^ self genLowcodeUrem32 ].
			[241]	-> [ ^ self genLowcodeUrem64 ].
			[242]	-> [ ^ self genLowcodeXor32 ].
			[243]	-> [ ^ self genLowcodeXor64 ].
			[244]	-> [ ^ self genLowcodeZeroExtend32From16 ].
			[245]	-> [ ^ self genLowcodeZeroExtend32From8 ].
			[246]	-> [ ^ self genLowcodeZeroExtend64From16 ].
		}
		otherwise: [ ^ self genLowcodeUnaryInlinePrimitive5: prim ].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators dispatch generated code' stamp: ''!
genLowcodeUnaryInlinePrimitive5: prim
	<option: #LowcodeVM>	"Lowcode instruction generator dispatch"
	prim
		caseOf: {
			[247]	-> [ ^ self genLowcodeZeroExtend64From32 ].
			[248]	-> [ ^ self genLowcodeZeroExtend64From8 ].
		}
		otherwise: [ ^ EncounteredUnknownBytecode ].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators dispatch generated code' stamp: ''!
genLowcodeUnaryInlinePrimitive: prim
	<option: #LowcodeVM>	"Lowcode instruction generator dispatch"
	prim
		caseOf: {
			[0]	-> [ ^ self genLowcodeAdd32 ].
			[1]	-> [ ^ self genLowcodeAdd64 ].
			[2]	-> [ ^ self genLowcodeAlloca32 ].
			[3]	-> [ ^ self genLowcodeAlloca64 ].
			[4]	-> [ ^ self genLowcodeAnd32 ].
			[5]	-> [ ^ self genLowcodeAnd64 ].
			[6]	-> [ ^ self genLowcodeArithmeticRightShift32 ].
			[7]	-> [ ^ self genLowcodeArithmeticRightShift64 ].
			[8]	-> [ ^ self genLowcodeBeginCall ].
			[9]	-> [ ^ self genLowcodeCallArgumentFloat32 ].
			[10]	-> [ ^ self genLowcodeCallArgumentFloat64 ].
			[11]	-> [ ^ self genLowcodeCallArgumentInt32 ].
			[12]	-> [ ^ self genLowcodeCallArgumentInt64 ].
			[13]	-> [ ^ self genLowcodeCallArgumentPointer ].
			[14]	-> [ ^ self genLowcodeCallArgumentSpace ].
			[15]	-> [ ^ self genLowcodeCallArgumentStructure ].
			[16]	-> [ ^ self genLowcodeCallInstruction ].
			[17]	-> [ ^ self genLowcodeCallPhysical ].
			[18]	-> [ ^ self genLowcodeCheckSessionIdentifier ].
			[19]	-> [ ^ self genLowcodeCompareAndSwap32 ].
			[20]	-> [ ^ self genLowcodeDiv32 ].
			[21]	-> [ ^ self genLowcodeDiv64 ].
			[22]	-> [ ^ self genLowcodeDuplicateFloat32 ].
			[23]	-> [ ^ self genLowcodeDuplicateFloat64 ].
			[24]	-> [ ^ self genLowcodeDuplicateInt32 ].
			[25]	-> [ ^ self genLowcodeDuplicateInt64 ].
			[26]	-> [ ^ self genLowcodeDuplicatePointer ].
			[27]	-> [ ^ self genLowcodeEffectiveAddress32 ].
			[28]	-> [ ^ self genLowcodeEffectiveAddress64 ].
			[29]	-> [ ^ self genLowcodeEndCall ].
			[30]	-> [ ^ self genLowcodeEndCallNoCleanup ].
			[31]	-> [ ^ self genLowcodeFloat32Add ].
			[32]	-> [ ^ self genLowcodeFloat32Div ].
			[33]	-> [ ^ self genLowcodeFloat32Equal ].
			[34]	-> [ ^ self genLowcodeFloat32Great ].
			[35]	-> [ ^ self genLowcodeFloat32GreatEqual ].
			[36]	-> [ ^ self genLowcodeFloat32Less ].
			[37]	-> [ ^ self genLowcodeFloat32LessEqual ].
			[38]	-> [ ^ self genLowcodeFloat32Mul ].
			[39]	-> [ ^ self genLowcodeFloat32Neg ].
			[40]	-> [ ^ self genLowcodeFloat32NotEqual ].
			[41]	-> [ ^ self genLowcodeFloat32Sqrt ].
			[42]	-> [ ^ self genLowcodeFloat32Sub ].
			[43]	-> [ ^ self genLowcodeFloat32ToFloat64 ].
			[44]	-> [ ^ self genLowcodeFloat32ToInt32 ].
			[45]	-> [ ^ self genLowcodeFloat32ToInt64 ].
			[46]	-> [ ^ self genLowcodeFloat32ToUInt32 ].
			[47]	-> [ ^ self genLowcodeFloat32ToUInt64 ].
			[48]	-> [ ^ self genLowcodeFloat64Add ].
			[49]	-> [ ^ self genLowcodeFloat64Div ].
			[50]	-> [ ^ self genLowcodeFloat64Equal ].
			[51]	-> [ ^ self genLowcodeFloat64Great ].
			[52]	-> [ ^ self genLowcodeFloat64GreatEqual ].
			[53]	-> [ ^ self genLowcodeFloat64Less ].
			[54]	-> [ ^ self genLowcodeFloat64LessEqual ].
			[55]	-> [ ^ self genLowcodeFloat64Mul ].
			[56]	-> [ ^ self genLowcodeFloat64Neg ].
			[57]	-> [ ^ self genLowcodeFloat64NotEqual ].
			[58]	-> [ ^ self genLowcodeFloat64Sqrt ].
			[59]	-> [ ^ self genLowcodeFloat64Sub ].
		}
		otherwise: [ ^ self genLowcodeUnaryInlinePrimitive2: prim ].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUnlockRegisters
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	"Do nothing for now"

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUnlockVM
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUnpin
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| object |
	self allocateRegistersForLowcodeOop: [:objectValue |
		object := objectValue.
	].

	self ssTop popToReg: object.
	self ssPop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUrem32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self DivR: second R: first Quo: second Rem: first.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeUrem64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh result resultLow resultHigh first second firstLow |
	self allocateRegistersForLowcodeInteger2ResultInteger: [:secondValue :firstValue :resultValue |
		second := secondValue.
		first := firstValue.
		result := resultValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self abort.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeXor32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| second first |
	self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
		second := secondValue.
		first := firstValue.
	].

	self ssNativeTop nativePopToReg: second.
	self ssNativePop: 1.
	self ssNativeTop nativePopToReg: first.
	self ssNativePop: 1.

	self XorR: second R: first.
	self ssPushNativeRegister: first.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeXor64
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| secondLow secondHigh firstHigh first second firstLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger4: [:secondLowValue :secondHighValue :firstLowValue :firstHighValue |
			secondLow := secondLowValue.
			secondHigh := secondHighValue.
			firstLow := firstLowValue.
			firstHigh := firstHighValue.
		].

		self ssNativeTop nativePopToReg: secondLow secondReg: secondHigh.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: firstLow secondReg: firstHigh.
		self ssNativePop: 1.

		self XorR: secondLow R: firstLow.
		self XorR: secondHigh R: firstHigh.
		self ssPushNativeRegister: firstLow secondRegister: firstHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger2: [:secondValue :firstValue |
			second := secondValue.
			first := firstValue.
		].

		self ssNativeTop nativePopToReg: second.
		self ssNativePop: 1.
		self ssNativeTop nativePopToReg: first.
		self ssNativePop: 1.

		self XorR: second R: first.
		self ssPushNativeRegister: first.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeZeroExtend32From16
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self ZeroExtend16R: value R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeZeroExtend32From8
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| value |
	self allocateRegistersForLowcodeInteger: [:valueValue |
		value := valueValue.
	].

	self ssNativeTop nativePopToReg: value.
	self ssNativePop: 1.

	self ZeroExtend8R: value R: value.
	self ssPushNativeRegister: value.

	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeZeroExtend64From16
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2: [:valueLowValue :valueHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self ZeroExtend16R: valueLow R: valueLow.
		self MoveCq: 0 R: valueHigh.
		self ssPushNativeRegister: valueLow secondRegister: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger: [:valueValue |
			value := valueValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self ZeroExtend16R: value R: value.
		self ssPushNativeRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeZeroExtend64From32
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| resultLow resultHigh value result |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeIntegerResultInteger2: [:valueValue :resultLowValue :resultHighValue |
			value := valueValue.
			resultLow := resultLowValue.
			resultHigh := resultHighValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self MoveR: value R: resultLow.
		self MoveCq: 0 R: resultHigh.
		self ssPushNativeRegister: resultLow secondRegister: resultHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeIntegerResultInteger: [:valueValue :resultValue |
			value := valueValue.
			result := resultValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self ZeroExtend32R: value R: value.
		self ssPushNativeRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive generators generated code' stamp: ''!
genLowcodeZeroExtend64From8
	<option: #LowcodeVM>
	<inline: true>
	"Generated by the Lowcode instruction generator."
	| valueHigh value valueLow |
	BytesPerWord = 4 ifTrue: [
		self allocateRegistersForLowcodeInteger2: [:valueLowValue :valueHighValue |
			valueLow := valueLowValue.
			valueHigh := valueHighValue.
		].

		self ssNativeTop nativePopToReg: valueLow secondReg: valueHigh.
		self ssNativePop: 1.

		self ZeroExtend8R: valueLow R: valueLow.
		self MoveCq: 0 R: valueHigh.
		self ssPushNativeRegister: valueLow secondRegister: valueHigh.

	] ifFalse: [
		self allocateRegistersForLowcodeInteger: [:valueValue |
			value := valueValue.
		].

		self ssNativeTop nativePopToReg: value.
		self ssNativePop: 1.

		self ZeroExtend8R: value R: value.
		self ssPushNativeRegister: value.

	].
	^ 0

! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/12/2018 10:32'!
genMarshalledSend: selectorIndex numArgs: numArgs sendTable: sendTable
	<inline: false>
	<var: #sendTable type: #'sqInt *'>
	| annotation |
	self assert: needsFrame.
	annotation := self annotationForSendTable: sendTable.
	"Deal with stale super sends; see SpurMemoryManager's class comment."
	(self annotationIsForUncheckedEntryPoint: annotation) ifTrue:
		[objectRepresentation genEnsureOopInRegNotForwarded: ReceiverResultReg scratchReg: TempReg].
	"0 through (NumSendTrampolines - 2) numArgs sends have the arg count implciti in the trampoline.
	 The last send trampoline (NumSendTrampolines - 1) passes numArgs in SendNumArgsReg."
	numArgs >= (NumSendTrampolines - 1) ifTrue:
		[self MoveCq: numArgs R: SendNumArgsReg].
	(BytecodeSetHasDirectedSuperSend
	 and: [annotation
			between: IsDirectedSuperSend
			and: IsDirectedSuperBindingSend]) ifTrue:
		[self genMoveConstant: tempOop R: TempReg].
	self genLoadInlineCacheWithSelector: selectorIndex.
	(self Call: (sendTable at: (numArgs min: NumSendTrampolines - 1))) annotation: annotation.
	self voidReceiverOptStatus.
	^self ssPushRegister: ReceiverResultReg! !

!StackToRegisterMappingCogit methodsFor: 'initialization' stamp: 'eem 03/03/2016 12:23'!
genMethodAbortTrampolineFor: numArgs
	"Generate the abort for a method.  This abort performs either a call of ceSICMiss:
	 to handle a single-in-line cache miss or a call of ceStackOverflow: to handle a
	 stack overflow.  It distinguishes the two by testing ResultReceiverReg.  If the
	 register is zero then this is a stack-overflow because a) the receiver has already
	 been pushed and so can be set to zero before calling the abort, and b) the
	 receiver must always contain an object (and hence be non-zero) on SIC miss."
	| jumpSICMiss |
	<var: #jumpSICMiss type: #'AbstractInstruction *'>
	self zeroOpcodeIndex.
	self CmpCq: 0 R: ReceiverResultReg.
	jumpSICMiss := self JumpNonZero: 0.

	"The abort sequence has pushed the LinkReg a second time - because a stack
	 overflow can only happen after building a frame, which pushes LinkReg anyway, and
	 we still need to push LinkReg in case we get to this routine from a sendMissAbort.
	 (On ARM there is a simpler way; use two separate abort calls since all instructions are 32-bits
	  but on x86 the zero receiver reg, call methodAbort sequence is smaller; we may fix this one day).
	 Overwrite that duplicate with the right one - the return address for the call to the abort trampoline.
	 The only reason it matters is an assert in ceStackOverflow: uses it"
	backEnd hasLinkRegister ifTrue:
		[self MoveR: LinkReg Mw: 0 r: SPReg].
	self compileTrampolineFor: #ceStackOverflow:
		numArgs: 1
		arg: SendNumArgsReg
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: false "The LinkReg has already been set above."
		resultReg: NoReg.
	jumpSICMiss jmpTarget: self Label.
	backEnd genPushRegisterArgsForAbortMissNumArgs: numArgs.
	^self genTrampolineFor: #ceSICMiss:
		called: (self trampolineName: 'ceMethodAbort' numRegArgs: numArgs)
		numArgs: 1
		arg: ReceiverResultReg
		arg: nil
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: false "The LinkReg will have been pushed in genPushRegisterArgsForAbortMissNumArgs: above."
		resultReg: NoReg
		appendOpcodes: true! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 9/9/2019 09:53'!
genNSSend: selectorIndex numArgs: numArgs depth: depth sendTable: sendTable
	<var: #sendTable type: #'sqInt *'>
	| selector nsSendCache |
	self assert: (selectorIndex between: 0 and: (objectMemory literalCountOf: methodObj) - 1).
	selector := self getLiteral: selectorIndex.
	self assert: (objectMemory addressCouldBeOop: selector).	
	(objectMemory isNonImmediate: selector) ifTrue:
		[self setHasMovableLiteral: true].
	(objectMemory isYoung: selector) ifTrue:
		[self setHasYoungReferent: true].

	nsSendCache := theIRCs + (NumOopsPerNSC * objectMemory bytesPerOop * indexOfIRC).
	indexOfIRC := indexOfIRC + 1.
	self assert: (objectMemory isInOldSpace: nsSendCache).
	self initializeNSSendCache: nsSendCache selector: selector numArgs: numArgs depth: depth.

	self ssAllocateCallReg: SendNumArgsReg.

	"This may leave the method receiver on the stack, which might not be the implicit receiver.
	 But the lookup trampoline will establish an on-stack receiver once it locates it."
	self marshallAbsentReceiverSendArguments: numArgs.

	"Load the cache last so it is a fixed distance from the call."
	self MoveUniqueCw: nsSendCache R: SendNumArgsReg.
	self CallNewspeakSend: (sendTable at: (numArgs min: NumSendTrampolines - 1)).

	self voidReceiverOptStatus.
	self ssPushRegister: ReceiverResultReg.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 12/02/2016 16:00'!
genOutlineClosure: startpc numArgs: numArgs numCopied: numCopied
	<inline: true>
	numCopied > 0 ifTrue:
		[self ssFlushTo: simStackPtr].
	self voidReceiverResultRegContainsSelf.
	objectRepresentation getActiveContextAllocatesInMachineCode
		ifTrue: [self ssAllocateCallReg: ReceiverResultReg
					and: SendNumArgsReg
					and: ClassReg]
		ifFalse: [self ssAllocateCallReg: SendNumArgsReg
					and: ReceiverResultReg].
	objectRepresentation
		genCreateClosureAt: startpc + 1 "1 relative"
		numArgs: numArgs
		numCopied: numCopied
		contextNumArgs: methodOrBlockNumArgs
		large: (coInterpreter methodNeedsLargeContext: methodObj)
		inBlock: inBlock.
	numCopied > 0 ifTrue:
		[self ssPop: numCopied].
	self ssPushRegister: ReceiverResultReg
! !

!StackToRegisterMappingCogit methodsFor: 'initialization' stamp: 'eem 11/13/2020 16:00'!
genPICAbortTrampolineFor: numArgs
	"Generate the abort for a PIC.  This abort performs either a call of
	 ceInterpretMethodFromPIC:receiver: to handle invoking an uncogged
	 target or a call of ceMNUFromPICMNUMethod:receiver: to handle an
	 MNU dispatch in a closed PIC.  It distinguishes the two by testing
	 ClassReg.  If the register is zero then this is an MNU."
	<inline: true>
	self zeroOpcodeIndex. 
	backEnd genPushRegisterArgsForAbortMissNumArgs: numArgs.
	^self genInnerPICAbortTrampoline: (self trampolineName: 'cePICAbort' numRegArgs: numArgs)! !

!StackToRegisterMappingCogit methodsFor: 'initialization' stamp: 'eem 03/03/2016 12:23'!
genPICMissTrampolineFor: numArgs
	<inline: false>
	| startAddress |
	startAddress := methodZoneBase.
	self zeroOpcodeIndex.
	"N.B. a closed PIC jumps to the miss routine, not calls it, so there is only one retpc on the stack."
	backEnd genPushRegisterArgsForNumArgs: numArgs scratchReg: SendNumArgsReg.
	self genTrampolineFor: #ceCPICMiss:receiver:
		called: (self trampolineName: 'cePICMiss' numRegArgs: numArgs)
		numArgs: 2
		arg: ClassReg
		arg: ReceiverResultReg
		arg: nil
		arg: nil
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: true.
	^startAddress! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'cb 4/18/2016 17:03'!
genPopStackBytecode
	self ssTop spilled ifTrue:
		[self AddCq: objectMemory wordSize R: SPReg].
	self ssPop: 1.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'primitive generators' stamp: 'cb 10/20/2017 16:39'!
genPrimReturn
	"Generate a return that cuts back the stack to remove the receiver
	 and arguments after an invocation of a primitive with nargs arguments.
	 Since all primitives that succeed in the normal way (i.e. don't execute a
	 method as do genPrimitiveClosureValue and genPrimitivePerform) take only
	 register arguments, there is nothing to do."
	<inline: true>
	^methodOrBlockNumArgs <= self numRegArgs
		ifTrue: [self RetN: 0]
		ifFalse: [super genPrimReturn]! !

!StackToRegisterMappingCogit methodsFor: 'primitive generators' stamp: 'eem 12/1/2010 10:12'!
genPrimitiveClosureValue
	"Check the argument count.  Fail if wrong.
	 Get the method from the outerContext and see if it is cogged.  If so, jump to the
	 block entry or the no-context-switch entry, as appropriate, and we're done.  If not,
	 invoke the interpreter primitive.
	 Override to push the register args first."
	self genPushRegisterArgs.
	^super genPrimitiveClosureValue! !

!StackToRegisterMappingCogit methodsFor: 'primitive generators' stamp: 'cb 4/11/2016 15:47'!
genPrimitiveFullClosureValue
	"Override to push the register args first."
	<option: #SistaV1BytecodeSet>
	self genPushRegisterArgs.
	^super genPrimitiveFullClosureValue! !

!StackToRegisterMappingCogit methodsFor: 'primitive generators' stamp: 'eem 1/7/2016 10:48'!
genPrimitivePerform
	"Generate an in-line perform primitive.  The lookup code requires the selector to be in Arg0Reg.
	 adjustArgumentsForPerform: adjusts the arguments once genLookupForPerformNumArgs:
	 has generated the code for the lookup."
	methodOrBlockNumArgs > self numRegArgs ifTrue:
		[self MoveMw: (backEnd hasLinkRegister
					ifTrue: [methodOrBlockNumArgs - 1]
					ifFalse: [methodOrBlockNumArgs]) * objectMemory wordSize
			r: SPReg
			R: Arg0Reg].
	^self genLookupForPerformNumArgs: methodOrBlockNumArgs! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 12/02/2016 16:00'!
genPushActiveContextBytecode
	self assert: needsFrame.
	self voidReceiverResultRegContainsSelf.
	objectRepresentation getActiveContextAllocatesInMachineCode
		ifTrue: [self ssAllocateCallReg: ReceiverResultReg
					and: SendNumArgsReg
					and: ClassReg]
		ifFalse: [self ssAllocateCallReg: ReceiverResultReg].
	objectRepresentation
		genGetActiveContextNumArgs: methodOrBlockNumArgs
		large: (coInterpreter methodNeedsLargeContext: methodObj)
		inBlock: inBlock.
	^self ssPushRegister: ReceiverResultReg! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 09/28/2015 18:48'!
genPushClosureCopyCopiedValuesBytecode
	"Block compilation.  At this point in the method create the block.  Note its start
	 and defer generating code for it until after the method and any other preceding
	 blocks.  The block's actual code will be compiled later."
	"143   10001111 llllkkkk jjjjjjjj iiiiiiii	Push Closure Num Copied llll Num Args kkkk BlockSize jjjjjjjjiiiiiiii"
	| startpc numArgs numCopied |
	self assert: needsFrame.
	startpc := bytecodePC + (self generatorAt: byte0) numBytes.
	self addBlockStartAt: startpc "0 relative"
		numArgs: (numArgs := byte1 bitAnd: 16rF)
		numCopied: (numCopied := byte1 >> 4)
		span: (byte2 << 8) + byte3.

	objectRepresentation createsClosuresInline 
		ifTrue: [ self genInlineClosure: startpc numArgs: numArgs numCopied: numCopied ]
		ifFalse: [ self genOutlineClosure: startpc numArgs: numArgs numCopied: numCopied ].
		
	^ 0

	! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 12/02/2016 16:00'!
genPushEnclosingObjectAt: level
	"Uncached push enclosing object"
	self voidReceiverResultRegContainsSelf.
	self ssAllocateCallReg: SendNumArgsReg and: ReceiverResultReg.
	self MoveCq: level R: SendNumArgsReg.
	self CallRT: ceEnclosingObjectTrampoline.
	^self ssPushRegister: ReceiverResultReg! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 11/21/2010 11:29'!
genPushLiteral: literal
	^self ssPushConstant: literal! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/12/2018 09:53'!
genPushLiteralIndex: literalIndex "<SmallInteger>"
	"Override to avoid the BytecodeSetHasDirectedSuperSend check, which is unnecessary
	 here given the simulation stack."
	<inline: false>
	| literal |
	literal := self getLiteral: literalIndex.
	^self genPushLiteral: literal! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 9/9/2019 10:16'!
genPushLiteralVariable: literalIndex
	<inline: false>
	| association freeReg |
	association := self getLiteral: literalIndex.
	"If followed by a directed super send bytecode, avoid generating any code yet.
	 The association will be passed to the directed send trampoline in a register
	 and fully dereferenced only when first linked.  It will be ignored in later sends."
	BytecodeSetHasDirectedSuperSend ifTrue:
		[self deny: directedSendUsesBinding.
		 self nextDescriptorExtensionsAndNextPCInto:
			[:descriptor :exta :extb :followingPC|
			(self isDirectedSuper: descriptor extA: exta extB: extb) ifTrue:
				[self ssPushConstant: association.
				 directedSendUsesBinding := true.
				 ^0]]].
	freeReg := self allocateRegNotConflictingWith: 0.
	"N.B. Do _not_ use ReceiverResultReg to avoid overwriting receiver in assignment in frameless methods."
	"So far descriptors are not rich enough to describe the entire dereference so generate the register
	 load but don't push the result.  There is an order-of-evaluation issue if we defer the dereference."
	self genMoveConstant: association R: TempReg.
	objectRepresentation
		genLoadSlot: ValueIndex
		sourceReg: TempReg
		destReg: freeReg.
	self ssPushRegister: freeReg.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'cb 4/5/2016 14:58'!
genPushMaybeContextReceiverVariable: slotIndex 
	<inline: false>
	self ssAllocateCallReg: ReceiverResultReg and: SendNumArgsReg.
	self ensureReceiverResultRegContainsSelf.
	^ self genPushMaybeContextSlotIndex: slotIndex! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'cb 6/30/2016 13:59'!
genPushMaybeContextRemoteInstVar: slotIndex inObjectAt: index
	<inline: false>
	self ssAllocateCallReg: ReceiverResultReg and: SendNumArgsReg.
	self genLoadTemp: index in: ReceiverResultReg.
	^ self genPushMaybeContextSlotIndex: slotIndex! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 12/02/2016 16:00'!
genPushMaybeContextSlotIndex: slotIndex
	<inline: true>
	"This method expects ReceiverResultReg to hold the object read"
	| jmpSingle jmpDone |
	<var: #jmpSingle type: #'AbstractInstruction *'>
	<var: #jmpDone type: #'AbstractInstruction *'>
	self assert: needsFrame.
	(self isCallerSavedReg: ReceiverResultReg) ifTrue:
		["We have no way of reloading ReceiverResultReg since we need the inst var value as the result."
		self voidReceiverResultRegContainsSelf].
	"See CoInterpreter>>contextInstructionPointer:frame: for an explanation
	 of the instruction pointer slot handling."
	slotIndex = InstructionPointerIndex ifTrue:
		[self MoveCq: slotIndex R: SendNumArgsReg.
		 self CallRT: ceFetchContextInstVarTrampoline.
		 ^self ssPushRegister: SendNumArgsReg].
	objectRepresentation
		genLoadSlot: SenderIndex
		sourceReg: ReceiverResultReg
		destReg: TempReg.
	jmpSingle := objectRepresentation genJumpNotSmallIntegerInScratchReg: TempReg.
	self MoveCq: slotIndex R: SendNumArgsReg.
	self CallRT: ceFetchContextInstVarTrampoline.
	jmpDone := self Jump: 0.
	jmpSingle jmpTarget: self Label.
	objectRepresentation
		genLoadSlot: slotIndex
		sourceReg: ReceiverResultReg
		destReg: SendNumArgsReg.
	jmpDone jmpTarget: self Label.
	^self ssPushRegister: SendNumArgsReg! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 12/02/2016 16:00'!
genPushNewArrayBytecode
	| size popValues |
	self assert: needsFrame.
	self voidReceiverResultRegContainsSelf.
	(popValues := byte1 > 127)
		ifTrue: [self ssFlushTo: simStackPtr]
		ifFalse: [self ssAllocateCallReg: SendNumArgsReg and: ReceiverResultReg].
	size := byte1 bitAnd: 127.
	popValues ifFalse:
		[(self tryCollapseTempVectorInitializationOfSize: size) ifTrue:
			[^0]].
	objectRepresentation genNewArrayOfSize: size initialized: popValues not.
	popValues ifTrue:
		[size - 1 to: 0 by: -1 do:
			[:i|
			self PopR: TempReg.
			objectRepresentation
				genStoreSourceReg: TempReg
				slotIndex: i
				intoNewObjectInDestReg: ReceiverResultReg].
		 self ssPop: size].
	^self ssPushRegister: ReceiverResultReg! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 3/6/2018 17:38'!
genPushReceiverBytecode
	self receiverIsInReceiverResultReg ifTrue:
		[^self ssPushRegister: ReceiverResultReg].
	^self ssPushDesc: self ssSelfDescriptor! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 10/1/2013 17:31'!
genPushReceiverVariable: index
	<inline: false>
	self ensureReceiverResultRegContainsSelf.
	^self ssPushBase: ReceiverResultReg
			offset: (objectRepresentation slotOffsetOfInstVarIndex: index)! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 7/24/2021 18:49'!
genPushRegisterArgs
	"Ensure that the register args are pushed before the retpc for methods with arity <= self numRegArgs."
	"This isn't as clumsy on a RISC.  But putting the receiver and
	 args above the return address means the CoInterpreter has a
	 single machine-code frame format which saves us a lot of work."
	(regArgsHaveBeenPushed
	 or: [methodOrBlockNumArgs > self numRegArgs]) ifFalse:
		[backEnd genPushRegisterArgsForNumArgs: methodOrBlockNumArgs scratchReg: SendNumArgsReg.
		regArgsHaveBeenPushed := true]! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'cb 6/30/2016 13:59'!
genPushRemoteInstVar: index inObjectAt: objectIndex
	<inline: false>
	| objectReg resultReg |
	self assert: needsFrame.
	objectReg := self allocateRegNotConflictingWith: 0.
	self genLoadTemp: objectIndex in: objectReg.
	resultReg := self availableRegOrNoneNotConflictingWith: (self registerMaskFor: objectReg). 
	resultReg = NoReg ifTrue: [resultReg := objectReg].
	objectRepresentation
		genLoadSlot: byte1
		sourceReg: objectReg
		destReg: resultReg.
	^self ssPushRegister: resultReg! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'cb 5/29/2018 08:43'!
genPushRemoteTempLongBytecode
	| tempVectReg remoteTempReg |
	tempVectReg := self allocateRegNotConflictingWith: 0.
	self MoveMw: (self frameOffsetOfTemporary: byte2) r: FPReg R: tempVectReg.
	remoteTempReg := self availableRegOrNoneNotConflictingWith: (self registerMaskFor: tempVectReg). 
	remoteTempReg = NoReg ifTrue: [remoteTempReg := tempVectReg].
	TempVectReadBarrier
		ifTrue: [objectRepresentation
				genEnsureObjInRegNotForwarded: tempVectReg
				scratchReg: TempReg].
	objectRepresentation
		genLoadSlot: byte1
		sourceReg: tempVectReg
		destReg: remoteTempReg.
	^self ssPushRegister: remoteTempReg! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/22/2018 18:15'!
genPushTemporaryVariable: index
	"If a frameless method (not a block), only argument temps can be accessed.
	 This is assured by the use of needsFrameIfMod16GENumArgs: in pushTemp."
	self assert: (inBlock > 0 or: [needsFrame or: [index < methodOrBlockNumArgs]]).
	^self ssPushDesc: (simStack at: index + 1)! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 4/21/2017 16:40'!
genReturnReceiver
	"In a frameless method ReceiverResultReg already contains self.
	 In a frameful method, ReceiverResultReg /may/ contain self."
	needsFrame ifTrue:
		[self receiverIsInReceiverResultReg ifFalse:
			[self putSelfInReceiverResultReg]].
	^self genUpArrowReturn! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 9/5/2016 14:13'!
genReturnTopFromBlock
	self assert: inBlock > 0.
	self ssTop popToReg: ReceiverResultReg.
	self ssPop: 1.
	^self genBlockReturn! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 12/1/2010 11:20'!
genReturnTopFromMethod
	self ssTop popToReg: ReceiverResultReg.
	self ssPop: 1.
	^self genUpArrowReturn! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 10/6/2015 14:53'!
genSend: selectorIndex numArgs: numArgs
	self marshallSendArguments: numArgs.
	^self genMarshalledSend: selectorIndex numArgs: numArgs sendTable: ordinarySendTrampolines! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'nice 11/8/2019 10:31'!
genSendDirectedSuper: selectorIndex numArgs: numArgs
	<option: #BytecodeSetHasDirectedSuperSend>
	| result |
	self assert: self ssTop type = SSConstant.
	tempOop := self ssTop constant.
	self ssPop: 1.
	self marshallSendArguments: numArgs.
	result := self
				genMarshalledSend: selectorIndex
				numArgs: numArgs
				sendTable: (directedSendUsesBinding
								ifTrue: [directedSuperBindingSendTrampolines]
								ifFalse: [directedSuperSendTrampolines]).
	directedSendUsesBinding := false.
	^result! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 10/6/2015 15:15'!
genSendDynamicSuper: selectorIndex numArgs: numArgs
	self marshallSendArguments: numArgs.
	^self genMarshalledSend: selectorIndex numArgs: numArgs sendTable: dynamicSuperSendTrampolines! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 10/6/2015 15:15'!
genSendSuper: selectorIndex numArgs: numArgs
	self marshallSendArguments: numArgs.
	^self genMarshalledSend: selectorIndex numArgs: numArgs sendTable: superSendTrampolines! !

!StackToRegisterMappingCogit methodsFor: 'initialization' stamp: 'eem 12/17/2019 12:12'!
genSendTrampolineFor: aRoutine numArgs: numArgs called: aString arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3
	"Generate a trampoline with four arguments.
	 Hack: a negative value indicates an abstract register, a non-negative value indicates a constant."
	<var: #aRoutine type: #'void *'>
	<var: #aString type: #'char *'>
	| startAddress |
	<inline: false>
	startAddress := methodZoneBase.
	self zeroOpcodeIndex.
	backEnd genPushRegisterArgsForNumArgs: numArgs scratchReg: SendNumArgsReg.
	"If we're on a 64-bit platform we use 32-bit inline caches.  These don't have room for a selector oop, only
	 a selector index.  If so, call the selectorIndexDereferenceRoutine to convert the index into a selector oop."
	objectRepresentation selectorIndexDereferenceRoutine ifNotNil:
		[:routine| "Explicitly save LinkReg via ExtraReg2; it's presumably faster than pushing/popping"
		 backEnd hasLinkRegister ifTrue: [self MoveR: LinkReg R: Extra2Reg].
		 self Call: routine.
		 backEnd hasLinkRegister ifTrue: [self MoveR: Extra2Reg R: LinkReg]].
	self genTrampolineFor: aRoutine
		called: aString
		numArgs: 4
		arg: regOrConst0
		arg: regOrConst1
		arg: regOrConst2
		arg: regOrConst3
		regsToSave: self emptyRegisterMask
		pushLinkReg: true
		resultReg: NoReg
		appendOpcodes: true.
	^startAddress! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 4/12/2017 11:29'!
genSistaExtStoreAndPopReceiverVariableBytecodePopBoolean: popBoolean
	<inline: true>
	| index needsStoreCheck needsImmCheck maybeContext result |
	needsStoreCheck := self sistaNeedsStoreCheck.
	needsImmCheck := self extBSpecifiesImmCheck.
	"Long form and short form exist for popInto. Only the long form exists for store.
	Store have an explicit flag to mark context accessing, while popInto context accessing are done through the long form,
	hence generate the context form if the flag is set or if this is a popInto."
	maybeContext := popBoolean or: [self extBSpecifiesMaybeContext].
	extB := 0.
	numExtB := 0.
	index := byte1 + (extA << 8).
	extA := 0.
	result := ((coInterpreter isWriteMediatedContextInstVarIndex: index) and: [maybeContext])
		ifTrue: [self 
				genStorePop: popBoolean 
				MaybeContextReceiverVariable: index 
				needsStoreCheck: needsStoreCheck 
				needsImmutabilityCheck: needsImmCheck]
		ifFalse: [self 
				 genStorePop: popBoolean 
				 ReceiverVariable: index 
				 needsStoreCheck: needsStoreCheck 
				 needsImmutabilityCheck: needsImmCheck].
	self cppIf: IMMUTABILITY ifTrue: [ needsImmCheck ifFalse: [self annotateBytecode: self Label ] ].
	^result! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'rsf 1/5/2017 16:07'!
genSistaExtStoreLiteralVariableBytecodePopBoolean: boolean
	<inline: true>
	| index needsStoreCheck needsImmCheck |
	needsStoreCheck := self sistaNeedsStoreCheck.
	needsImmCheck := self extBSpecifiesImmCheck.
	index := byte1 + (extA << 8).
	extA := numExtB := extB := 0.
	^self 
		genStorePop: boolean 
		LiteralVariable: index 
		needsStoreCheck: needsStoreCheck 
		needsImmutabilityCheck: needsImmCheck! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 1/22/2018 18:18'!
genSpecialSelectorArithmetic
	| primDescriptor rcvrIsConst argIsConst rcvrIsInt argIsInt rcvrInt argInt result
	 jumpNotSmallInts jumpContinue index |
	<var: #jumpContinue type: #'AbstractInstruction *'>
	<var: #primDescriptor type: #'BytecodeDescriptor *'>
	<var: #jumpNotSmallInts type: #'AbstractInstruction *'>
	primDescriptor := self generatorAt: byte0.
	argIsInt := (argIsConst := self ssTop type = SSConstant)
				 and: [objectMemory isIntegerObject: (argInt := self ssTop constant)].
	rcvrIsInt := ((rcvrIsConst := (self ssValue: 1) type = SSConstant)
				  and: [objectMemory isIntegerObject: (rcvrInt := (self ssValue: 1) constant)])
				or: [self mclassIsSmallInteger and: [(self ssValue: 1) isSameEntryAs: self simSelf]].

	(argIsInt and: [rcvrIsInt and: [rcvrIsConst]]) ifTrue:
		[rcvrInt := objectMemory integerValueOf: rcvrInt.
		 argInt := objectMemory integerValueOf: argInt.
		 primDescriptor opcode caseOf: {
			[AddRR]	-> [result := rcvrInt + argInt].
			[SubRR]	-> [result := rcvrInt - argInt].
			[AndRR]	-> [result := rcvrInt bitAnd: argInt].
			[OrRR]	-> [result := rcvrInt bitOr: argInt] }.
		(objectMemory isIntegerValue: result) ifTrue:
			["Must annotate the bytecode for correct pc mapping."
			^self ssPop: 2; ssPushAnnotatedConstant: (objectMemory integerObjectOf: result)].
		^self genSpecialSelectorSend].

	"If there's any constant involved other than a SmallInteger don't attempt to inline."
	((rcvrIsConst and: [rcvrIsInt not])
	 or: [argIsConst and: [argIsInt not]]) ifTrue:
		[^self genSpecialSelectorSend].

	"If we know nothing about the types then better not to inline as the inline cache and
	 primitive code is not terribly slow so wasting time on duplicating tag tests is pointless."
	(argIsInt or: [rcvrIsInt]) ifFalse:
		[^self genSpecialSelectorSend].

	argIsInt
		ifTrue:
			[self ssFlushTo: simStackPtr - 2.
			 (self ssValue: 1) popToReg: ReceiverResultReg.
			 self ssPop: 2]
		ifFalse:
			[self marshallSendArguments: 1].
	jumpNotSmallInts := (rcvrIsInt and: [argIsInt]) ifFalse:
							[argIsInt
								ifTrue: [objectRepresentation genJumpNotSmallInteger: ReceiverResultReg]
								ifFalse:
									[rcvrIsInt
										ifTrue: [objectRepresentation genJumpNotSmallInteger: Arg0Reg]
										ifFalse: [objectRepresentation genJumpNotSmallIntegersIn: ReceiverResultReg and: Arg0Reg scratch: TempReg]]].
	primDescriptor opcode caseOf: {
		[AddRR] -> [argIsInt
						ifTrue:
							[self AddCq: argInt - ConstZero R: ReceiverResultReg.
							 jumpContinue := self JumpNoOverflow: 0.
							 "overflow; must undo the damage before continuing"
							 self SubCq: argInt - ConstZero R: ReceiverResultReg]
						ifFalse:
							[objectRepresentation genRemoveSmallIntegerTagsInScratchReg: ReceiverResultReg.
							 self AddR: Arg0Reg R: ReceiverResultReg.
							jumpContinue := self JumpNoOverflow: 0.
							"overflow; must undo the damage before continuing"
							 (rcvrIsInt and: [rcvrIsConst])
								ifTrue: [self MoveCq: rcvrInt R: ReceiverResultReg]
								ifFalse:
									[self SubR: Arg0Reg R: ReceiverResultReg.
									 objectRepresentation genSetSmallIntegerTagsIn: ReceiverResultReg]]].
		[SubRR] -> [argIsInt
						ifTrue:
							[self SubCq: argInt - ConstZero R: ReceiverResultReg.
							 jumpContinue := self JumpNoOverflow: 0.
							 "overflow; must undo the damage before continuing"
							 self AddCq: argInt - ConstZero R: ReceiverResultReg]
						ifFalse:
							[objectRepresentation genRemoveSmallIntegerTagsInScratchReg: Arg0Reg.
							 self SubR: Arg0Reg R: ReceiverResultReg.
							 jumpContinue := self JumpNoOverflow: 0.
							 "overflow; must undo the damage before continuing"
							 self AddR: Arg0Reg R: ReceiverResultReg.
							 objectRepresentation genSetSmallIntegerTagsIn: Arg0Reg]].
		[AndRR] -> [argIsInt
						ifTrue: [self AndCq: argInt R: ReceiverResultReg]
						ifFalse: [self AndR: Arg0Reg R: ReceiverResultReg].
					jumpContinue := jumpNotSmallInts ifNotNil: [self Jump: 0]].
		[OrRR]	-> [argIsInt
						ifTrue: [self OrCq: argInt R: ReceiverResultReg]
						ifFalse: [self OrR: Arg0Reg R: ReceiverResultReg].
					jumpContinue := jumpNotSmallInts ifNotNil: [self Jump: 0]] }.
	jumpNotSmallInts
		ifNil: [jumpContinue ifNil: "overflow cannot happen"
				[self annotateInstructionForBytecode.
				 self ssPushRegister: ReceiverResultReg.
				 ^0]]
		ifNotNil:
			[jumpNotSmallInts jmpTarget: self Label].
	argIsInt ifTrue:
		[self MoveCq: argInt R: Arg0Reg].
	index := byte0 - self firstSpecialSelectorBytecodeOffset.
	self genMarshalledSend: index negated - 1 numArgs: 1 sendTable: ordinarySendTrampolines.
	jumpContinue jmpTarget: self Label.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'cb 9/11/2019 07:50'!
genSpecialSelectorClass
	| topReg |
	topReg := self ssTop registerOrNone.
	self ssPop: 1.
	(topReg = NoReg or: [topReg = ClassReg])
		ifTrue: [self ssAllocateRequiredReg: (topReg := SendNumArgsReg) and: ClassReg]
		ifFalse: [self ssAllocateRequiredReg: ClassReg].
	self ssPush: 1.
	self ssTop popToReg: topReg.
	objectRepresentation
		genGetClassObjectOf: topReg
		into: ClassReg
		scratchReg: TempReg
		mayBeAForwarder: self ssTop mayBeAForwarder.
	^self ssPop: 1; ssPushRegister: ClassReg! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 1/22/2018 18:19'!
genSpecialSelectorComparison
	| nextPC postBranchPC targetPC primDescriptor branchDescriptor
	  rcvrIsInt rcvrIsConst argIsIntConst argInt jumpNotSmallInts inlineCAB index |
	<var: #primDescriptor type: #'BytecodeDescriptor *'>
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>
	<var: #jumpNotSmallInts type: #'AbstractInstruction *'>
	self ssFlushTo: simStackPtr - 2.
	primDescriptor := self generatorAt: byte0.
	argIsIntConst := self ssTop type = SSConstant
				 and: [objectMemory isIntegerObject: (argInt := self ssTop constant)].
	rcvrIsInt := ((rcvrIsConst := (self ssValue: 1) type = SSConstant)
				  and: [objectMemory isIntegerObject: (self ssValue: 1) constant])
				or: [self mclassIsSmallInteger and: [(self ssValue: 1) isSameEntryAs: self simSelf]].

	(argIsIntConst and: [rcvrIsInt and: [rcvrIsConst]]) ifTrue:
		[^ self genStaticallyResolvedSpecialSelectorComparison].

	self extractMaybeBranchDescriptorInto: [ :descr :next :postBranch :target | 
		branchDescriptor := descr. nextPC := next. postBranchPC := postBranch. targetPC := target ].

	"Only interested in inlining if followed by a conditional branch."
	inlineCAB := branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse].
	"Further, only interested in inlining = and ~= if there's a SmallInteger constant involved.
	 The relational operators successfully statically predict SmallIntegers; the equality operators do not."
	(inlineCAB and: [primDescriptor opcode = JumpZero or: [primDescriptor opcode = JumpNonZero]]) ifTrue:
		[inlineCAB := argIsIntConst or: [rcvrIsInt]].
	inlineCAB ifFalse:
		[^self genSpecialSelectorSend].

	argIsIntConst
		ifTrue:
			[(self ssValue: 1) popToReg: ReceiverResultReg.
			 self ssPop: 2]
		ifFalse:
			[self marshallSendArguments: 1].
	jumpNotSmallInts := (rcvrIsInt and: [argIsIntConst]) ifFalse:
							[argIsIntConst
								ifTrue: [objectRepresentation genJumpNotSmallInteger: ReceiverResultReg]
								ifFalse:
									[rcvrIsInt
										ifTrue: [objectRepresentation genJumpNotSmallInteger: Arg0Reg]
										ifFalse: [objectRepresentation genJumpNotSmallIntegersIn: ReceiverResultReg and: Arg0Reg scratch: TempReg]]].
	argIsIntConst
		ifTrue: [self CmpCq: argInt R: ReceiverResultReg]
		ifFalse: [self CmpR: Arg0Reg R: ReceiverResultReg].
	"Cmp is weird/backwards so invert the comparison.  Further since there is a following conditional
	 jump bytecode define non-merge fixups and leave the cond bytecode to set the mergeness."
	self genConditionalBranch: (branchDescriptor isBranchTrue
				ifTrue: [primDescriptor opcode]
				ifFalse: [self inverseBranchFor: primDescriptor opcode])
		operand: (self ensureNonMergeFixupAt: targetPC) asUnsignedInteger.
	self Jump: (self ensureNonMergeFixupAt: postBranchPC).
	jumpNotSmallInts ifNil:
		[self annotateInstructionForBytecode.
		 self ensureFixupAt: postBranchPC.
		 self ensureFixupAt: targetPC.
		 deadCode := true.
		 ^0].
	jumpNotSmallInts jmpTarget: self Label.
	argIsIntConst ifTrue:
		[self MoveCq: argInt R: Arg0Reg].
	index := byte0 - self firstSpecialSelectorBytecodeOffset.
	^self genMarshalledSend: index negated - 1 numArgs: 1 sendTable: ordinarySendTrampolines! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'cb 4/18/2016 17:05'!
genStaticallyResolvedSpecialSelectorComparison
	"Assumes both operands are ints"
	<var: #primDescriptor type: #'BytecodeDescriptor *'>
	| rcvrInt argInt primDescriptor result |
	primDescriptor := self generatorAt: byte0.
	argInt := self ssTop constant.
	rcvrInt := (self ssValue: 1) constant.
	self cCode: '' inSmalltalk: "In Simulator ints are unsigned..."
		[rcvrInt := objectMemory integerValueOf: rcvrInt.
		argInt := objectMemory integerValueOf: argInt].
	 primDescriptor opcode caseOf: {
		[JumpLess]				-> [result := rcvrInt < argInt].
		[JumpLessOrEqual]		-> [result := rcvrInt <= argInt].
		[JumpGreater]			-> [result := rcvrInt > argInt].
		[JumpGreaterOrEqual]	-> [result := rcvrInt >= argInt].
		[JumpZero]				-> [result := rcvrInt = argInt].
		[JumpNonZero]			-> [result := rcvrInt ~= argInt] }.
	 "Must annotate the bytecode for correct pc mapping."
	 self ssPop: 2.
	 ^self ssPushAnnotatedConstant: (result
			ifTrue: [objectMemory trueObject]
			ifFalse: [objectMemory falseObject])! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator stores' stamp: 'cb 7/28/2016 11:05'!
genStorePop: popBoolean LiteralVariable: litVarIndex
	<inline: true>
	^self 
		genStorePop: popBoolean 
		LiteralVariable: litVarIndex 
		needsStoreCheck: self ssTopNeedsStoreCheck
		needsImmutabilityCheck: true "The generic store checks for IMMUTABILITY flag"
		! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator stores' stamp: 'cb 7/28/2016 11:06'!
genStorePop: popBoolean LiteralVariable: litVarIndex needsStoreCheck: needsStoreCheck needsImmutabilityCheck: needsImmCheck
	<inline: false>
	"We need a frame because the association has to be in ReceiverResultReg for the various trampolines
	and ReceiverResultReg holds only the receiver in frameless methods."
	self assert: needsFrame.
	self genLoadLiteralVariable: litVarIndex in: ReceiverResultReg.
	^self 
		genGenericStorePop: popBoolean 
		slotIndex: ValueIndex 
		destReg: ReceiverResultReg
		needsStoreCheck: needsStoreCheck
		needsRestoreRcvr: false
		needsImmutabilityCheck: needsImmCheck
		! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator stores' stamp: 'cb 7/28/2016 11:05'!
genStorePop: popBoolean MaybeContextReceiverVariable: slotIndex 
	<inline: true>
	^self 
		genStorePop: popBoolean 
		MaybeContextReceiverVariable: slotIndex 
		needsStoreCheck: self ssTopNeedsStoreCheck
		needsImmutabilityCheck: true "The generic store checks for IMMUTABILITY flag"! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator stores' stamp: 'cb 7/28/2016 11:06'!
genStorePop: popBoolean MaybeContextReceiverVariable: slotIndex needsStoreCheck: needsStoreCheck needsImmutabilityCheck: needsImmCheck
	<inline: false>
	"The reason we need a frame here is that assigning to an inst var of a context may
	 involve wholesale reorganization of stack pages, and the only way to preserve the
	 execution state of an activation in that case is if it has a frame."
	self assert: needsFrame.
	self ssFlushUpThroughReceiverVariable: slotIndex.
	self ensureReceiverResultRegContainsSelf.
	^self 
		genGenericStorePop: popBoolean 
		MaybeContextSlotIndex: slotIndex 
		needsStoreCheck: needsStoreCheck
		needsRestoreRcvr: true
		needsImmutabilityCheck: needsImmCheck! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator stores' stamp: 'cb 7/28/2016 11:06'!
genStorePop: popBoolean MaybeContextRemoteInstVar: slotIndex ofObjectAt: objectIndex needsStoreCheck: needsStoreCheck needsImmutabilityCheck: needsImmCheck
	<inline: false>
	"The reason we need a frame here is that assigning to an inst var of a context may
	 involve wholesale reorganization of stack pages, and the only way to preserve the
	 execution state of an activation in that case is if it has a frame."
	self assert: needsFrame.
	self genLoadTemp: objectIndex in: ReceiverResultReg.
	^self 
		genGenericStorePop: popBoolean 
		MaybeContextSlotIndex: slotIndex 
		needsStoreCheck: needsStoreCheck
		needsRestoreRcvr: false
		needsImmutabilityCheck: needsImmCheck! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator stores' stamp: 'cb 7/28/2016 11:05'!
genStorePop: popBoolean ReceiverVariable: slotIndex
	<inline: true>
	^self 
		genStorePop: popBoolean 
		ReceiverVariable: slotIndex 
		needsStoreCheck: self ssTopNeedsStoreCheck
		needsImmutabilityCheck: true "The generic store checks for IMMUTABILITY flag"
		! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator stores' stamp: 'cb 7/28/2016 11:06'!
genStorePop: popBoolean ReceiverVariable: slotIndex needsStoreCheck: needsStoreCheck needsImmutabilityCheck: needsImmCheck
	<inline: false>
	self ssFlushUpThroughReceiverVariable: slotIndex.
	self ensureReceiverResultRegContainsSelf.
	"In two path compilation the receiver is young AND mutable, hence no store check nor immutability check is needed"
	^self 
		genGenericStorePop: popBoolean 
		slotIndex: slotIndex 
		destReg: ReceiverResultReg 
		needsStoreCheck: (useTwoPaths not and: [needsStoreCheck])
		needsRestoreRcvr: true "ReceiverResultReg is kept live with the receiver across the operation"
		needsImmutabilityCheck: (needsImmCheck and: [useTwoPaths not])
		! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator stores' stamp: 'cb 7/28/2016 11:06'!
genStorePop: popBoolean RemoteInstVar: slotIndex ofObjectAt: objectIndex needsStoreCheck: needsStoreCheck needsImmutabilityCheck: needsImmCheck
	<inline: false>
	self assert: needsFrame.
	self genLoadTemp: objectIndex in: ReceiverResultReg.
	^self 
		genGenericStorePop: popBoolean 
		slotIndex: slotIndex 
		destReg: ReceiverResultReg 
		needsStoreCheck: needsStoreCheck
		needsRestoreRcvr: false
		needsImmutabilityCheck: needsImmCheck! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator stores' stamp: 'cb 7/28/2016 11:05'!
genStorePop: popBoolean RemoteTemp: slotIndex At: remoteTempIndex
	<inline: true>
	^ self genStorePop: popBoolean RemoteTemp: slotIndex At: remoteTempIndex needsStoreCheck: self ssTopNeedsStoreCheck! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator stores' stamp: 'cb 5/29/2018 08:43'!
genStorePop: popBoolean RemoteTemp: slotIndex At: remoteTempIndex needsStoreCheck: needsStoreCheck
	<inline: false>
	"The only reason we assert needsFrame here is that in a frameless method
	 ReceiverResultReg must and does contain only self, but the ceStoreCheck
	 trampoline expects the target of the store to be in ReceiverResultReg.  So
	 in a frameless method we would have a conflict between the receiver and
	 the temote temp store, unless we we smart enough to realise that
	 ReceiverResultReg was unused after the literal variable store, unlikely given
	 that methods return self by default."
	self assert: needsFrame.
	"N.B.  No need to check the stack for references because we generate code for
	 remote temp loads that stores the result in a register, deferring only the register push."
	self ssAllocateRequiredReg: ReceiverResultReg. 
	self voidReceiverResultRegContainsSelf.
	self MoveMw: (self frameOffsetOfTemporary: remoteTempIndex) r: FPReg R: ReceiverResultReg.
	TempVectReadBarrier
		ifTrue: [objectRepresentation
				genEnsureObjInRegNotForwarded: ReceiverResultReg
				scratchReg: TempReg].
	^self 
		genGenericStorePop: popBoolean 
		slotIndex: slotIndex 
		destReg: ReceiverResultReg 
		needsStoreCheck: needsStoreCheck
		needsRestoreRcvr: false "We don't keep ReceiverResultReg live with the receiver across this operation"
		needsImmutabilityCheck: false "never do immutability check on temp vectors"! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator stores' stamp: 'eem 1/22/2018 18:16'!
genStorePop: popBoolean TemporaryVariable: tempIndex
	<inline: false>
	| reg |
	self ssFlushUpThroughTemporaryVariable: tempIndex.
	reg := self ssStorePop: popBoolean toPreferredReg: TempReg.
	self MoveR: reg
		Mw: (self frameOffsetOfTemporary: tempIndex)
		r: FPReg.
	(self simStackAt: tempIndex + 1) bcptr: bytecodePC. "for debugging"
	^0! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator stores' stamp: 'cb 3/30/2016 14:39'!
genTraceStores
	<inline: true>
	traceStores > 0 ifTrue: 
		[ self MoveR: ClassReg R: TempReg.
		self CallRT: ceTraceStoreTrampoline ].! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 4/12/2017 12:02'!
genUpArrowReturn
	"Generate a method return from within a method or a block.
	 Frameless method activation looks like
	 CISCs (x86):
				receiver
				args
		sp->	ret pc.
	 RISCs (ARM):
				receiver
				args
				ret pc in LR.
	 A fully framed activation is described in CoInterpreter class>initializeFrameIndices.
	 Return pops receiver and arguments off the stack.  Callee pushes the result."
	deadCode := true. "can't fall through"
	inBlock > 0 ifTrue:
		[self assert: needsFrame.
		 self ssFlushTo: simStackPtr.
		 self CallRT: ceNonLocalReturnTrampoline.
		 self annotateBytecode: self Label.
		 ^0].
	(self cppIf: IMMUTABILITY ifTrue: [needsFrame and: [useTwoPaths not]] ifFalse: [needsFrame])
		ifTrue:
			[LowcodeVM ifTrue: [ hasNativeFrame ifTrue: [ self leaveNativeFrame ] ].
			 self MoveR: FPReg R: SPReg.
			 self PopR: FPReg.
			 backEnd hasLinkRegister ifTrue:
				[self PopR: LinkReg].
			 self RetN: methodOrBlockNumArgs + 1 * objectMemory wordSize]
		ifFalse:
			[self RetN: ((methodOrBlockNumArgs > self numRegArgs
						"A method with an interpreter prim will push its register args for the prim.  If the failure
						 body is frameless the args must still be popped, see e.g. Behavior>>nextInstance."
						or: [regArgsHaveBeenPushed])
							ifTrue: [methodOrBlockNumArgs + 1 * objectMemory wordSize]
							ifFalse: [0])].
	^0! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generators' stamp: 'eem 2/28/2017 09:44'!
genVanillaInlinedIdenticalOrNotIf: orNot
	| nextPC postBranchPC targetPC branchDescriptor
	  rcvrReg argReg argIsConstant rcvrIsConstant  |
	<var: #branchDescriptor type: #'BytecodeDescriptor *'>
	
	self extractMaybeBranchDescriptorInto: [ :descr :next :postBranch :target | 
		branchDescriptor := descr. nextPC := next. postBranchPC := postBranch. targetPC := target ].
	
	argIsConstant := self ssTop type = SSConstant.
	"They can't be both constants to use correct machine opcodes.
	 However annotable constants can't be resolved statically, hence we need to careful."
	rcvrIsConstant := argIsConstant not and: [(self ssValue: 1) type = SSConstant].
	
	self 
		allocateEqualsEqualsRegistersArgNeedsReg: argIsConstant not 
		rcvrNeedsReg: rcvrIsConstant not 
		into: [ :rcvr :arg | rcvrReg:= rcvr. argReg := arg ].
	
	"If not followed by a branch, resolve to true or false."
	(branchDescriptor isBranchTrue or: [branchDescriptor isBranchFalse]) ifFalse:
		[^ self 
			genIdenticalNoBranchArgIsConstant: argIsConstant 
			rcvrIsConstant: rcvrIsConstant 
			argReg: argReg 
			rcvrReg: rcvrReg 
			orNotIf: orNot].
	
	"If branching the stack must be flushed for the merge"
	self ssFlushTo: simStackPtr - 2.
	
	self genCmpArgIsConstant: argIsConstant rcvrIsConstant: rcvrIsConstant argReg: argReg rcvrReg: rcvrReg.
	self ssPop: 2.

	"Further since there is a following conditional jump bytecode, define
	 non-merge fixups and leave the cond bytecode to set the mergeness."
	(self fixupAt: nextPC) notAFixup
		ifTrue: "The next instruction is dead.  we can skip it."
			[deadCode := true.
		 	 self ensureFixupAt: targetPC.
			 self ensureFixupAt: postBranchPC]
		ifFalse:
			[self deny: deadCode]. "push dummy value below"
		

	orNot == branchDescriptor isBranchTrue "orNot is true for ~~"
		ifFalse: "a == b ifTrue: ... or a ~~ b ifFalse: ... jump on equal to target pc"
			[self JumpZero:  (self ensureNonMergeFixupAt: targetPC).
			 self Jump: (self ensureNonMergeFixupAt: postBranchPC)]
		ifTrue: "a == b ifFalse: ... or a ~~ b ifTrue: ... jump on equal to post-branch pc"
			[self ensureNonMergeFixupAt: targetPC.
			 self JumpZero: (self ensureNonMergeFixupAt: postBranchPC).
			 self Jump: (self ensureNonMergeFixupAt: targetPC)].

	"Not reached, execution flow has jumped to fixup"
	deadCode ifFalse:
		[self ssPushConstant: objectMemory trueObject]. "dummy value"
	^0! !

!StackToRegisterMappingCogit methodsFor: 'initialization' stamp: 'eem 3/18/2015 17:56'!
generateEnilopmarts
	"Enilopmarts transfer control from C into machine code (backwards trampolines).
	 Override to add version for generic and PIC-specific entry with reg args."
	super generateEnilopmarts.

	self cppIf: Debug
		ifTrue:
			[realCECallCogCodePopReceiverArg0Regs :=
				self genEnilopmartFor: ReceiverResultReg
					and: Arg0Reg
					forCall: true
					called: 'realCECallCogCodePopReceiverArg0Regs'.
			 ceCallCogCodePopReceiverArg0Regs := #callCogCodePopReceiverArg0Regs.
			 realCECallCogCodePopReceiverArg1Arg0Regs :=
				self genEnilopmartFor: ReceiverResultReg
					and: Arg0Reg
					and: Arg1Reg
					forCall: true
					called: 'realCECallCogCodePopReceiverArg1Arg0Regs'.
			 ceCallCogCodePopReceiverArg1Arg0Regs := #callCogCodePopReceiverArg1Arg0Regs]
		ifFalse:
			[ceCallCogCodePopReceiverArg0Regs :=
				self genEnilopmartFor: ReceiverResultReg
					and: Arg0Reg
					forCall: true
					called: 'ceCallCogCodePopReceiverArg0Regs'.
			 ceCallCogCodePopReceiverArg1Arg0Regs :=
				self genEnilopmartFor: ReceiverResultReg
					and: Arg0Reg
					and: Arg1Reg
					forCall: true
					called: 'ceCallCogCodePopReceiverArg1Arg0Regs'].

	"These are special versions of the ceCallCogCodePopReceiverAndClassRegs enilopmart that also
	 pop register args from the stack to undo the pushing of register args in the abort/miss trampolines."
	ceCall0ArgsPIC := self genCallPICEnilopmartNumArgs: 0.
	self numRegArgs >= 1 ifTrue:
		[ceCall1ArgsPIC := self genCallPICEnilopmartNumArgs: 1.
		 self numRegArgs >= 2 ifTrue:
			[ceCall2ArgsPIC := self genCallPICEnilopmartNumArgs: 2.
			 self assert: self numRegArgs = 2]]! !

!StackToRegisterMappingCogit methodsFor: 'generate machine code' stamp: 'eem 9/15/2020 13:46'!
generateInstructionsAt: eventualAbsoluteAddress
	"Size pc-dependent instructions and assign eventual addresses to all instructions.
	 Answer the size of the code.
	 Compute forward branches based on virtual address (abstract code starts at 0),
	 assuming that any branches branched over are long.
	 Compute backward branches based on actual address.
	 Reuse the fixups array to record the pc-dependent instructions that need to have
	 their code generation postponed until after the others.

	 Override to add handling for null branches (branches to the immediately following
	 instruction) occasioned by StackToRegisterMapping's following of jumps."
	| absoluteAddress pcDependentIndex abstractInstruction fixup |
	<var: #abstractInstruction type: #'AbstractInstruction *'>
	<var: #fixup type: #'BytecodeFixup *'>
	absoluteAddress := eventualAbsoluteAddress.
	pcDependentIndex := 0.
	0 to: opcodeIndex - 1 do:
		[:i|
		abstractInstruction := self abstractInstructionAt: i.
		"N.B. if you want to break in resizing, break here, note the instruction index, back up to the
		 sender, restart, and step into computeMaximumSizes, breaking at this instruction's index."
		self maybeBreakGeneratingFrom: absoluteAddress to: absoluteAddress + abstractInstruction maxSize.
		abstractInstruction isPCDependent
			ifTrue:
				[abstractInstruction sizePCDependentInstructionAt: absoluteAddress.
				 (abstractInstruction isJump
				  and: [(i + 1 < opcodeIndex
				  		and: [abstractInstruction getJmpTarget == (self abstractInstructionAt: i + 1)])
					or: [i + 2 < opcodeIndex
						and: [abstractInstruction getJmpTarget == (self abstractInstructionAt: i + 2)
						and: [(self abstractInstructionAt: i + 1) opcode = Nop]]]])
					ifTrue:
						[abstractInstruction
							opcode: Nop;
							concretizeAt: absoluteAddress]
					ifFalse:
						[fixup := self fixupAtIndex: pcDependentIndex.
						 pcDependentIndex := pcDependentIndex + 1.
						 fixup instructionIndex: i].
				 absoluteAddress := absoluteAddress + abstractInstruction machineCodeSize]
			ifFalse:
				[absoluteAddress := abstractInstruction concretizeAt: absoluteAddress.
				"N.B. if you want to break in resizing, break here, note the instruction index, back up to the
				 sender, restart, and step into computeMaximumSizes, breaking at this instruction's index."
				 self assert: abstractInstruction machineCodeSize = abstractInstruction maxSize]].
	0 to: pcDependentIndex - 1 do:
		[:j|
		fixup := self fixupAtIndex: j.
		abstractInstruction := self abstractInstructionAt: fixup instructionIndex.
		self maybeBreakGeneratingFrom: abstractInstruction address to: abstractInstruction address + abstractInstruction maxSize - 1.
		abstractInstruction concretizeAt: abstractInstruction address].
	^absoluteAddress - eventualAbsoluteAddress! !

!StackToRegisterMappingCogit methodsFor: 'initialization' stamp: 'eem 12/7/2017 17:01'!
generateMissAbortTrampolines
	"Generate the run-time entries for the various method and PIC entry misses and aborts.
	 Read the class-side method trampolines for documentation on the various trampolines"
	0 to: self numRegArgs + 1 do:
		[:numArgs|
		methodAbortTrampolines
			at: numArgs
			put: (self genMethodAbortTrampolineFor: numArgs)].
	0 to: self numRegArgs + 1 do:
		[:numArgs|
		picAbortTrampolines
			at: numArgs
			put: (self genPICAbortTrampolineFor: numArgs)].
	0 to: self numRegArgs + 1 do:
		[:numArgs|
		picMissTrampolines
			at: numArgs
			put: (self genPICMissTrampolineFor: numArgs)].
	ceReapAndResetErrorCodeTrampoline := self genTrampolineFor: #ceReapAndResetErrorCodeFor:
												called: 'ceReapAndResetErrorCodeTrampoline'
												arg: ClassReg! !

!StackToRegisterMappingCogit methodsFor: 'initialization' stamp: 'eem 1/11/2018 19:00'!
generateSendTrampolines
	"Override to generate code to push the register arg(s) for <= numRegArg arity sends."
	0 to: NumSendTrampolines - 1 do:
		[:numArgs|
		ordinarySendTrampolines
			at: numArgs
			put: (self genSendTrampolineFor: #ceSend:super:to:numArgs:
					  numArgs: numArgs
					  called: (self trampolineName: 'ceSend' numArgs: numArgs)
					  arg: ClassReg
					  arg: (self trampolineArgConstant: false)
					  arg: ReceiverResultReg
					  arg: (self numArgsOrSendNumArgsReg: numArgs))].

	"Generate these in the middle so they are within [firstSend, lastSend]."
	NewspeakVM ifTrue: [self generateNewspeakSendTrampolines].
	BytecodeSetHasDirectedSuperSend ifTrue:
		[0 to: NumSendTrampolines - 1 do:
			[:numArgs|
			directedSuperSendTrampolines
				at: numArgs
				put: (self genSendTrampolineFor: #ceSend:above:to:numArgs:
						  numArgs: numArgs
						  called: (self trampolineName: 'ceDirectedSuperSend' numArgs: numArgs)
						  arg: ClassReg
						  arg: TempReg
						  arg: ReceiverResultReg
						  arg: (self numArgsOrSendNumArgsReg: numArgs)).
			directedSuperBindingSendTrampolines
				at: numArgs
				put: (self genSendTrampolineFor: #ceSend:aboveClassBinding:to:numArgs:
						  numArgs: numArgs
						  called: (self trampolineName: 'ceDirectedSuperBindingSend' numArgs: numArgs)
						  arg: ClassReg
						  arg: TempReg
						  arg: ReceiverResultReg
						  arg: (self numArgsOrSendNumArgsReg: numArgs))]].

	0 to: NumSendTrampolines - 1 do:
		[:numArgs|
		superSendTrampolines
			at: numArgs
			put: (self genSendTrampolineFor: #ceSend:super:to:numArgs:
					  numArgs: numArgs
					  called: (self trampolineName: 'ceSuperSend' numArgs: numArgs)
					  arg: ClassReg
					  arg: (self trampolineArgConstant: true)
					  arg: ReceiverResultReg
					  arg: (self numArgsOrSendNumArgsReg: numArgs))].
	firstSend := ordinarySendTrampolines at: 0.
	lastSend := superSendTrampolines at: NumSendTrampolines - 1! !

!StackToRegisterMappingCogit methodsFor: 'initialization' stamp: 'RonieSalgado 4/29/2016 17:37'!
generateSistaRuntime
	"No sita vm"! !

!StackToRegisterMappingCogit methodsFor: 'initialization' stamp: 'eem 4/28/2020 13:36'!
generateTracingTrampolines
	"Generate trampolines for tracing.  In the simulator we can save a lot of time
	 and avoid noise instructions in the lastNInstructions log by short-cutting these
	 trampolines, but we need them in the real vm."
	ceTraceLinkedSendTrampoline :=
		self genTrampolineFor: #ceTraceLinkedSend:
			called: 'ceTraceLinkedSendTrampoline'
			arg: ReceiverResultReg
			regsToSave: CallerSavedRegisterMask.
	ceTraceBlockActivationTrampoline :=
		self genTrampolineFor: #ceTraceBlockActivation
			called: 'ceTraceBlockActivationTrampoline'
			regsToSave: CallerSavedRegisterMask.
	ceTraceStoreTrampoline :=
		self genTrampolineFor: #ceTraceStoreOf:into:
			called: 'ceTraceStoreTrampoline'
			arg: TempReg
			arg: ReceiverResultReg
			regsToSave: CallerSavedRegisterMask.
	self cCode: [] inSmalltalk:
		[ceTraceLinkedSendTrampoline := self simulatedTrampolineFor: #ceShortCutTraceLinkedSend:.
		 ceTraceBlockActivationTrampoline := self simulatedTrampolineFor: #ceShortCutTraceBlockActivation:.
		 ceTraceStoreTrampoline := self simulatedTrampolineFor: #ceShortCutTraceStore:]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/24/2018 10:57'!
initSimStackForFramefulMethod: startpc
	<var: #desc type: #'CogSimStackEntry *'>
	simStackPtr := methodOrBlockNumTemps. "N.B. Includes num args"
	simSpillBase := methodOrBlockNumTemps + 1.
	LowcodeVM ifTrue:
		[simNativeSpillBase := simNativeStackPtr := -1.
		 simNativeStackSize := 0].
	self simSelf
		type: SSBaseOffset;
		spilled: true;
		register: FPReg;
		offset: FoxMFReceiver;
		liveRegister: NoReg.
	"args"
	1 to: methodOrBlockNumArgs do:
		[:i| | desc |
		desc := self simStackAt: i.
		desc
			type: SSBaseOffset;
			spilled: true;
			register: FPReg;
			offset: FoxCallerSavedIP + ((methodOrBlockNumArgs - i + 1) * objectMemory wordSize);
			bcptr: startpc].
	"temps"
	methodOrBlockNumArgs + 1 to: simStackPtr do:
		[:i| | desc |
		desc := self simStackAt: i.
		desc
			type: SSBaseOffset;
			spilled: true;
			register: FPReg;
			offset: FoxMFReceiver - (i - methodOrBlockNumArgs * objectMemory wordSize);
			bcptr: startpc]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/24/2018 10:57'!
initSimStackForFramelessBlock: startpc
	"The register receiver (the closure itself) and args are pushed by the closure value primitive(s)
	 and hence a frameless block has all arguments and copied values pushed to the stack.  However,
	 the method receiver (self) is put in the ReceiverResultReg by the block entry."
	<var: #desc type: #'CogSimStackEntry *'>
	self simSelf
		type: SSRegister;
		spilled: false;
		register: ReceiverResultReg;
		liveRegister: ReceiverResultReg.
	self assert: methodOrBlockNumTemps >= methodOrBlockNumArgs.
	1 to: methodOrBlockNumTemps do:
		[:i| | desc |
		desc := self simStackAt: i.
		desc
			type: SSBaseOffset;
			spilled: true;
			register: SPReg;
			offset: ((backEnd hasLinkRegister
								ifTrue: [methodOrBlockNumArgs - i]
								ifFalse: [methodOrBlockNumArgs + 1 - i]) * objectMemory wordSize);
			bcptr: startpc].
	simStackPtr := methodOrBlockNumTemps. "N.B. Includes num args"
	simSpillBase := methodOrBlockNumTemps + 1.
	LowcodeVM ifTrue:
		[simNativeSpillBase := simNativeStackPtr := -1.
		 simNativeStackSize := 0]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/24/2018 10:51'!
initSimStackForFramelessMethod: startpc
	| desc |
	<var: #desc type: #'CogSimStackEntry *'>
	self simSelf
		type: SSRegister;
		spilled: false;
		register: ReceiverResultReg;
		liveRegister: ReceiverResultReg.
	self assert: methodOrBlockNumTemps = methodOrBlockNumArgs.
	self assert: self numRegArgs <= 2.
	(methodOrBlockNumArgs between: 1 and: self numRegArgs)
		ifTrue:
			[desc := self simStackAt: 1.
			 desc
				type: SSRegister;
				spilled: false;
				register: Arg0Reg;
				bcptr: startpc.
			 methodOrBlockNumArgs > 1 ifTrue:
				[desc := self simStackAt: 2.
				 desc
					type: SSRegister;
					spilled: false;
					register: Arg1Reg;
					bcptr: startpc]]
		ifFalse:
			[1 to: methodOrBlockNumArgs do:
				[:i|
				desc := self simStackAt: i.
				desc
					type: SSBaseOffset;
					register: SPReg;
					spilled: true;
					offset: ((backEnd hasLinkRegister
								ifTrue: [methodOrBlockNumArgs - i]
								ifFalse: [methodOrBlockNumArgs + 1 - i]) * objectMemory wordSize);
					bcptr: startpc]].
	simStackPtr := methodOrBlockNumArgs.
	simSpillBase := methodOrBlockNumArgs + 1.
	LowcodeVM ifTrue:
		[simNativeSpillBase := simNativeStackPtr := -1.
		 simNativeStackSize := 0]! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 3/4/2017 10:19'!
initializeFixup: fixup
	"Initialize a fixup.  These are the targets of backward branches.  A backward branch fixup's
	 simStackPtr needs to be set when generating the code for the bytecode at the targetPC.
	 Initially a fixup's target is just a flag.  Later on it is replaced with a proper instruction."
	<var: #fixup type: #'BytecodeFixup *'>
	<inline: true>
	fixup
		becomeMergeFixup;
		setIsBackwardBranchFixup! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 3/6/2018 17:50'!
initializeFixupAt: targetPC
	"Make sure there's a flagged fixup at the targetPC in fixups.
	 These are the targets of backward branches.  A backward branch fixup's simStackPtr
	 needs to be set when generating the code for the bytecode at the targetPC.
	 Initially a fixup's target is just a flag.  Later on it is replaced with a proper instruction."
	<inline: true>
	| fixup |
	fixup := self fixupAt: targetPC.
	self initializeFixup: fixup! !

!StackToRegisterMappingCogit methodsFor: 'testing' stamp: 'eem 9/18/2019 14:02'!
isNonForwarderReceiver: reg
	"Do not inline (inBlock access)"
	^self receiverIsInReceiverResultReg
		and: [inBlock = 0 "Method, rcvr may be forwarder in blocks without inst var access." 
		and: [reg = ReceiverResultReg]]! !

!StackToRegisterMappingCogit methodsFor: 'inline primitive register allocation' stamp: 'rsf 3/26/2017 17:16'!
isOopRegister: register usedBefore: stackIndex
	<option: #LowcodeVM>
	<inline: true>
	| index |
	index := simSpillBase max: 0.
	index to: simStackPtr - stackIndex do: [ :i |
		(self simStackAt: index) registerOrNone = register ifTrue: [ 
			^ true
		]
	].

	^ false! !

!StackToRegisterMappingCogit methodsFor: 'span functions' stamp: 'eem 07/29/2014 06:48'!
isPushNil: descriptor pc: pc nExts: nExts method: aMethodObj
	<inline: true>
	<var: #descriptor type: #'BytecodeDescriptor *'>
	^self perform: numPushNilsFunction
		with: descriptor
		with: pc
		with: nExts
		with: aMethodObj! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 8/10/2016 00:19'!
leaveNativeFrame
	<option: #LowcodeVM>
	self assert: needsFrame.
	self MoveMw: self frameOffsetOfPreviousNativeStackPointer r: FPReg R: TempReg.
	self SubCq: 1 R: TempReg.
	self MoveR: TempReg Aw: coInterpreter nativeStackPointerAddress! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 09/27/2016 11:20'!
liveFloatRegisters
	| regsSet |
	regsSet := 0.
	(simSpillBase max: 0) to: simStackPtr do:
		[:i|
		regsSet := regsSet bitOr: (self simStackAt: i) floatRegisterMask].
	LowcodeVM ifTrue: [ 
		(simNativeSpillBase max: 0) to: simNativeStackPtr do:
			[:i|
			regsSet := regsSet bitOr: (self simNativeStackAt: i) nativeFloatRegisterMask].
	].
	^regsSet! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 09/27/2016 11:20'!
liveRegisters
	| regsSet |
	needsFrame
		ifTrue: [regsSet := 0]
		ifFalse:
			[regsSet := self registerMaskFor: ReceiverResultReg.
			 (methodOrBlockNumArgs <= self numRegArgs
			  and: [methodOrBlockNumArgs > 0]) ifTrue:
				[regsSet := regsSet bitOr: (self registerMaskFor: Arg0Reg).
				 (self numRegArgs > 1 and: [methodOrBlockNumArgs > 1]) ifTrue:
					[regsSet := regsSet bitOr: (self registerMaskFor: Arg1Reg)]]].
	(simSpillBase max: 0) to: simStackPtr do:
		[:i|
		regsSet := regsSet bitOr: (self simStackAt: i) registerMask].
	LowcodeVM ifTrue: [ 
		(simNativeSpillBase max: 0) to: simNativeStackPtr do:
			[:i|
			regsSet := regsSet bitOr: (self simNativeStackAt: i) nativeRegisterMask].
	].
	^regsSet! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 9/5/2016 14:13'!
mapDeadDescriptorIfNeeded: descriptor 
	"insert nops for dead code that is mapped so that bc 
	 to mc mapping is not many to one"
	<var: #descriptor type: #'BytecodeDescriptor *'>
	self flag: #annotateInstruction .
	(descriptor isMapped
		or: [inBlock > 0 and: [descriptor isMappedInBlock]]) 
		ifTrue: [self annotateBytecode: self Nop].
	^ 0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/20/2016 14:23'!
marshallAbsentReceiverSendArguments: numArgs
	self assert: needsFrame.
	self ssAllocateCallReg: ReceiverResultReg.
	self putSelfInReceiverResultReg.

	"Spill everything on the simulated stack that needs spilling (that below arguments).
	 Marshall arguments to stack and/or registers depending on arg count.
	 If the args don't fit in registers push receiver and args (spill everything).  Assume
	 receiver already in ResultReceiverReg so shuffle args and push it if necessary."
	self ssFlushTo: simStackPtr - numArgs.
	numArgs > self numRegArgs
		ifTrue:
			["The arguments must be pushed to the stack, and hence the receiver
			   must be inserted beneath the args.  Reduce or eliminate the argument
			   shuffle by only moving already spilled items."
			| numSpilled |
			numSpilled := self numberOfSpillsInTopNItems: numArgs.
			numSpilled > 0
				ifTrue:
					[self MoveMw: 0 r: SPReg R: TempReg.
					 self PushR: TempReg.
					 2 to: numSpilled do:
						[:index|
						self MoveMw: index * objectMemory wordSize r: SPReg R: TempReg.
						self MoveR: TempReg Mw: index - 1 * objectMemory wordSize r: SPReg].
					 self MoveR: ReceiverResultReg Mw: numSpilled * objectMemory wordSize r: SPReg]
				ifFalse:
					[self PushR: ReceiverResultReg].
			self ssFlushTo: simStackPtr]
		"Move the args to the register arguments, being careful to do
		 so last to first so e.g. previous contents don't get overwritten.
		 Also check for any arg registers in use by other args."
		ifFalse:
			[numArgs > 0 ifTrue:
				[(self numRegArgs > 1 and: [numArgs > 1])
					ifTrue:
						[self ssAllocateRequiredReg: Arg0Reg upThrough: simStackPtr - 2.
						 self ssAllocateRequiredReg: Arg1Reg upThrough: simStackPtr - 1]
					ifFalse:
						[self ssAllocateRequiredReg: Arg0Reg upThrough: simStackPtr - 1]].
			 (self numRegArgs > 1 and: [numArgs > 1]) ifTrue:
				[(self simStackAt: simStackPtr) popToReg: Arg1Reg].
			 numArgs > 0 ifTrue:
				[(self simStackAt: simStackPtr - numArgs + 1)
					popToReg: Arg0Reg]].
	self ssPop: numArgs! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/12/2013 10:53'!
marshallSendArguments: numArgs 
	"Spill everything on the simulated stack that needs spilling (that below receiver and arguments).
	 Marshall receiver and arguments to stack and/or registers depending on arg count.
	 If the args don't fit in registers push receiver and args (spill everything), but still assign
	 the receiver to ReceiverResultReg."
	self ssFlushTo: simStackPtr - numArgs - 1.
	numArgs > self numRegArgs
		ifTrue:
			"If there are no spills and no references to ReceiverResultReg
			 the fetch of ReceiverResultReg from the stack can be avoided
			 by assigning directly to ReceiverResultReg and pushing it."
			[| numSpilled anyRefs |
			numSpilled := self numberOfSpillsInTopNItems: numArgs + 1.
			anyRefs := self anyReferencesToRegister: ReceiverResultReg inTopNItems: numArgs + 1.
			(numSpilled > 0 or: [anyRefs])
				ifTrue:
					[self ssFlushTo: simStackPtr.
					 (self simStackAt: simStackPtr - numArgs)
						storeToReg: ReceiverResultReg]
				ifFalse:
					[(self simStackAt: simStackPtr - numArgs)
						storeToReg: ReceiverResultReg;
					 	type: SSRegister;
						register: ReceiverResultReg.
					 self ssFlushTo: simStackPtr]]
		ifFalse:
			"Move the args to the register arguments, being careful to do
			 so last to first so e.g. previous contents don't get overwritten.
			 Also check for any arg registers in use by other args."
			[numArgs > 0 ifTrue:
				[(self numRegArgs > 1 and: [numArgs > 1])
					ifTrue:
						[self ssAllocateRequiredReg: Arg0Reg upThrough: simStackPtr - 2.
						 self ssAllocateRequiredReg: Arg1Reg upThrough: simStackPtr - 1]
					ifFalse:
						[self ssAllocateRequiredReg: Arg0Reg upThrough: simStackPtr - 1]].
			 (self numRegArgs > 1 and: [numArgs > 1]) ifTrue:
				[(self simStackAt: simStackPtr) popToReg: Arg1Reg].
			 numArgs > 0 ifTrue:
				[(self simStackAt: simStackPtr - numArgs + 1)
					popToReg: Arg0Reg].
			 (self simStackAt: simStackPtr - numArgs)
				popToReg: ReceiverResultReg].
	self ssPop: numArgs + 1! !

!StackToRegisterMappingCogit methodsFor: 'debugging' stamp: 'eem 3/3/2017 10:30'!
maybeCompilingFirstPassOfBlockWithInitialPushNil
	"For assert checking; or rather for avoiding assert fails when dealing with the hack for block temps in the SqueakV3PlusClosures bytecode set."
	^inBlock = InVanillaBlock and: [methodOrBlockNumTemps > methodOrBlockNumArgs and: [compilationPass = 1]]! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/1/2016 15:20'!
maybeCountCounter
	"This is a hook for SistaCogit"
	<inline: true>! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/28/2016 17:11'!
maybeCountFixup: descriptor
	"This is a hook for RegisterAllocatingCogit"
	<inline: true>! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/25/2016 17:57'!
maybeCounterIndex
	"Hook for SistaCogit to allow it to reuse compileBlockBodies"
	<inline: true>
	^0! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/1/2016 15:20'!
maybeInitNumCounters
	"This is a hook for SistaCogit"
	<inline: true>! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 6/1/2016 15:14'!
maybeInitNumFixups
	"This is a hook for RegisterAllocatingCogit"
	<inline: true>! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 11/25/2016 17:57'!
maybeSetCounterIndex: value
	"Hook for SistaCogit to allow it to reuse compileBlockBodies"
	<inline: true>! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 2/6/2018 17:57'!
mergeWithFixupIfRequired: fixup
	"If this bytecode has a fixup, some kind of merge needs to be done. There are 4 cases:
		1) the bytecode has no fixup (fixup isNotAFixup)
			do nothing
		2) the bytecode has a non merge fixup
			the fixup has needsNonMergeFixup.
			The code generating non merge fixup (currently only special selector code) is responsible
				for the merge so no need to do it.
			We set deadCode to false as the instruction can be reached from jumps.
		3) the bytecode has a merge fixup, but execution flow *cannot* fall through to the merge point.
			the fixup has needsMergeFixup and deadCode = true.
			ignores the current simStack as it does not mean anything 
			restores the simStack to the state the jumps to the merge point expects it to be.
		4) the bytecode has a merge fixup and execution flow *can* fall through to the merge point.
			the fixup has needsMergeFixup and deadCode = false.
			flushes the stack to the stack pointer so the fall through execution path simStack is 
				in the state the merge point expects it to be. 
			restores the simStack to the state the jumps to the merge point expects it to be.
			
	In addition, if this is a backjump merge point, we patch the fixup to hold the current simStackPtr 
	for later assertions."
	<var: #fixup type: #'BytecodeFixup *'>

	self assertCorrectSimStackPtr.

	"case 1"
	fixup notAFixup ifTrue:
		[^0].

	"case 2"
	fixup isNonMergeFixup ifTrue:
		[deadCode := false. ^0].

	"cases 3 and 4"
	self assert: fixup isMergeFixup.
	self traceMerge: fixup.
	deadCode
		ifTrue: "case 3"
			["Would like to assert fixup simStackPtr >= methodOrBlockNumTemps
			   but can't because of the initialNils hack."
			 self assert: (fixup simStackPtr >= methodOrBlockNumTemps
						or: [self maybeCompilingFirstPassOfBlockWithInitialPushNil]).
			 simStackPtr := fixup simStackPtr.
			 LowcodeVM ifTrue:
				[simNativeStackPtr := fixup simNativeStackPtr.
				simNativeStackSize := fixup simNativeStackSize]]
		ifFalse: "case 4"
			[self ssFlushTo: simStackPtr].

	"cases 3 and 4"
	deadCode := false.
	fixup isBackwardBranchFixup ifTrue:
		[fixup simStackPtr: simStackPtr.
		LowcodeVM ifTrue:
			[fixup simNativeStackPtr: simNativeStackPtr.
			 fixup simNativeStackSize: simNativeStackSize]].
	fixup targetInstruction: self Label.
	self assert: simStackPtr = fixup simStackPtr.
	LowcodeVM ifTrue:
		[self assert: simNativeStackPtr = fixup simNativeStackPtr.
		 self assert: simNativeStackSize = fixup simNativeStackSize].

	self cCode: '' inSmalltalk:
		[self assert: fixup simStackPtr = (self debugStackPointerFor: bytecodePC)].
	self restoreSimStackAtMergePoint: fixup.

	^0! !

!StackToRegisterMappingCogit methodsFor: 'trampolines' stamp: 'eem 12/3/2010 16:53'!
methodAbortTrampolineFor: numArgs
	^methodAbortTrampolines at: (numArgs min: self numRegArgs + 1)! !

!StackToRegisterMappingCogit methodsFor: 'testing' stamp: 'eem 1/25/2018 18:14'!
methodFoundInvalidPostScan
	"Frameless methods with local temporaries cause problems,
	 mostly in asserts, and yet they matter not at all for performance.
	 Shun them."
	needsFrame ifFalse:
		[^methodOrBlockNumTemps > methodOrBlockNumArgs].
	^super methodFoundInvalidPostScan! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 10/8/2014 11:31'!
needsFrameIfExtBGT2: stackDelta
	^extB < 0 or: [extB > 2]! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 8/17/2013 18:48'!
needsFrameIfFollowsSend: stackDelta
	"As of August 2013, the code generator can't deal with spills in frameless methods (the
	 issue is to do with the stack offset to get at an argument, which is changed when there's a spill).
	 The only context in a spill is needed in a frameless method that I can think of is sends
	 following sends as in e.g. TextColor>>#dominates: other ^other class == self class.
	 Only need to check for the frameless sends since all other sends will force a frame."

	self assert: (prevBCDescriptor notNil and: [prevBCDescriptor needsFrameFunction notNil]).
	^prevBCDescriptor generator == #genSpecialSelectorEqualsEquals
	  or: [prevBCDescriptor generator == #genSpecialSelectorClass]! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 8/17/2013 18:47'!
needsFrameIfMod16GENumArgs: stackDelta
	^byte0 \\ 16 >= methodOrBlockNumArgs! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 8/17/2013 18:50'!
needsFrameIfStackGreaterThanOne: stackDelta
	"As of August 2013, the code generator can't deal with spills in frameless methods (the
	 issue is to do with the stack offset to get at an argument, which is changed when there's a spill).
	 In e.g. TextColor>>#dominates: other ^other class == self class the second send of class
	 needs also rto allocate a register that the first one used, but the first one's register can't be
	 spilled.  So avoid this by only allowing class to be sent if the stack contains a single element."

	^stackDelta > 1! !

!StackToRegisterMappingCogit methodsFor: 'span functions' stamp: 'eem 07/29/2014 06:48'!
numPushNils: descriptor pc: pc nExts: nExts method: aMethodObj
	<inline: true>
	<var: #descriptor type: #'BytecodeDescriptor *'>
	^self perform: numPushNilsFunction
		with: descriptor
		with: pc
		with: nExts
		with: aMethodObj! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 5/1/2014 07:08'!
numRegArgs
	<doNotGenerate>
	^objectRepresentation numRegArgs! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/7/2013 11:53'!
numberOfSpillsInTopNItems: n
	simStackPtr to: simStackPtr - n + 1 by: -1 do:
		[:i| (self simStackAt: i) type = SSSpill ifTrue:
			[^n - (simStackPtr - i)]].
	^0! !

!StackToRegisterMappingCogit methodsFor: 'trampolines' stamp: 'eem 12/4/2010 08:42'!
picAbortTrampolineFor: numArgs
	^picAbortTrampolines at: (numArgs min: self numRegArgs + 1)! !

!StackToRegisterMappingCogit methodsFor: 'testing' stamp: 'eem 6/10/2015 17:29'!
prevInstIsPCAnnotated
	| prevIndex prevInst |
	<var: #prevInst type: #'AbstractInstruction *'>
	opcodeIndex > 0 ifFalse:
		[^false].
	prevIndex := opcodeIndex - 1.
	[prevIndex <= 0 ifTrue: [^false].
	 prevInst := self abstractInstructionAt: prevIndex.
	 (self isPCMappedAnnotation: (prevInst annotation ifNil: [0])) ifTrue:
		[^true].
	 prevInst opcode = Label]
		whileTrue:
			[prevIndex := prevIndex - 1].
	^false! !

!StackToRegisterMappingCogit methodsFor: 'debug printing' stamp: 'eem 2/1/2018 14:45'!
printRegisterMask: registerMask
	<doNotGenerate>
	^String streamContents: [:s| self printRegisterMask: registerMask on: s]! !

!StackToRegisterMappingCogit methodsFor: 'debug printing' stamp: 'eem 5/4/2017 12:50'!
printRegisterMask: registerMask on: aStream
	<doNotGenerate>
	| first |
	aStream nextPut: ${.
	registerMask = 0
		ifTrue:
			[aStream nextPutAll: 'NoReg']
		ifFalse:
			[first := true.
			 0 to: 31 do:
				[:reg|
				(registerMask anyMask: 1 << reg) ifTrue:
					[first ifFalse: [aStream space].
					 first := false.
					 aStream nextPutAll: (backEnd nameForRegister: reg)]]].
	aStream nextPut: $}; flush! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 11/25/2016 16:33'!
printSimStack
	<doNotGenerate>
	self printSimStack: simStack toDepth: simStackPtr spillBase: simSpillBase on: coInterpreter transcript! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 1/28/2018 19:35'!
printSimStack: aSimStackOrFixup
	<doNotGenerate>
	(aSimStackOrFixup isKindOf: CogRASSBytecodeFixup)
		ifTrue:
			[self printSimStack: aSimStackOrFixup mergeSimStack toDepth: aSimStackOrFixup simStackPtr spillBase: -1 on: coInterpreter transcript]
		ifFalse:
			[self printSimStack: aSimStackOrFixup toDepth: simStackPtr spillBase: simSpillBase on: coInterpreter transcript]! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 1/24/2018 16:52'!
printSimStack: aSimStack toDepth: limit spillBase: spillBase on: aStream
	<doNotGenerate>
	| tempNames width tabWidth |
	aStream ensureCr.
	limit < 0 ifTrue:
		[^aStream nextPutAll: 'simStackEmpty'; cr; flush].
	aSimStack ifNil:
		[^aStream nextPutAll: 'nil simStack'; cr; flush].
	(self class initializationOptions at: #tempNames ifAbsent: nil) ifNotNil:
		[:tempNamesDictOrNil |	 | tab longest |
		 tempNames := tempNamesDictOrNil at: (self class initializationOptions at: #startpc ifAbsent: [initialPC]) + 1.
		 longest := tempNames inject: '' into: [:m :t| m size >= t size ifTrue: [m] ifFalse: [t]].
		 tabWidth := self widthInDefaultFontOf: (tab := String with: Character tab).
		 width := self widthInDefaultFontOf: longest, tab.
		 width <= ((self widthInDefaultFontOf: longest, (String with: Character space)) + 4) ifTrue:
			[width := width + tabWidth]].
	0 to: limit do:
		[:i|
		width ifNotNil:
			[self put: (tempNames at: i ifAbsent: [i = 0 ifTrue: ['self'] ifFalse: ['']]) paddedTo: width tabWidth: tabWidth on: aStream].
		aStream print: i.
		i = simStackPtr ifTrue:
			[aStream nextPutAll: '<-' ].
		i = spillBase ifTrue:
			[aStream nextPutAll: '(sb)'].
		aStream tab: (i = spillBase ifTrue: [1] ifFalse: [2]).
		(aSimStack at: i) printStateOn: aStream.
		aStream cr; flush].
	simSpillBase > limit ifTrue:
		[aStream nextPutAll: '(sb: '; print: simSpillBase; nextPut: $); cr; flush]! !

!StackToRegisterMappingCogit methodsFor: 'span functions' stamp: 'eem 4/2/2015 16:44'!
pushNilSize: aMethodObj numInitialNils: numInitialNils
	<inline: true>
	^self perform: pushNilSizeFunction with: aMethodObj with: numInitialNils! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 2/28/2017 09:38'!
put: aString paddedTo: compositionWidth tabWidth: tabWidth on: aStream
	<doNotGenerate>
	| fittedString size width |
	fittedString := aString.
	size := fittedString size.
	[(width := self widthInDefaultFontOf: fittedString) > compositionWidth] whileTrue:
		[size := size - 2.
		 fittedString := aString contractTo: size].
	aStream
		nextPutAll: fittedString;
		tab: compositionWidth - width + (width \\ tabWidth) // tabWidth + 1! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/22/2018 18:07'!
putSelfInReceiverResultReg
	<inline: true>
	 self simSelf storeToReg: ReceiverResultReg! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/22/2018 18:06'!
receiverIsInReceiverResultReg
	"Used to mark ReceiverResultReg as dead or not containing simSelf.
	 Used when the simStack has already been flushed, e.g. for sends."
	<inline: true>
	^self simSelf liveRegister = ReceiverResultReg! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 02/09/2017 09:39'!
reinitializeFixupsFrom: start through: end
	"When a block must be recompiled due to overestimating the
	 numInitialNils fixups must be restored, which means rescannning
	 since backward branches need their targets initialized."
	| descriptor nExts pc distance targetPC |
	<var: #descriptor type: #'BytecodeDescriptor *'>
	pc := start.
	nExts := 0.
	[pc <= end] whileTrue:
		[(self fixupAt: pc) reinitialize.
		 byte0 := (objectMemory fetchByte: pc ofObject: methodObj) + bytecodeSetOffset.
		 descriptor := self generatorAt: byte0.
		 (descriptor isBranch
		  and: [self isBackwardBranch: descriptor at: pc exts: nExts in: methodObj]) ifTrue:
			[distance := self spanFor: descriptor at: pc exts: nExts in: methodObj.
			 targetPC := pc + descriptor numBytes + distance.
			 self initializeFixupAt: targetPC].
		 descriptor isBlockCreation
			ifTrue:
				[distance := self spanFor: descriptor at: pc exts: nExts in: methodObj.
				 pc := pc + descriptor numBytes + distance]
			ifFalse: [pc := pc + descriptor numBytes].
		 nExts := descriptor isExtension ifTrue: [nExts + 1] ifFalse: [0]]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/28/2018 09:19'!
restoreSimStackAtMergePoint: fixup
	<inline: true>
	"All the execution paths reaching a merge point expect everything
	 to be spilled on stack. Throw away all simStack optimization state."
	self voidReceiverOptStatus.
	methodOrBlockNumTemps + 1 to: simStackPtr do:
		[:i|
		 (self simStackAt: i)
			type: SSSpill;
			offset: FoxMFReceiver - (i - methodOrBlockNumArgs * objectMemory bytesPerOop);
			register: FPReg;
			spilled: true].
	simSpillBase := simStackPtr + 1.
	LowcodeVM ifTrue:
		[0 to: simNativeStackPtr do:
			[ :i |
			(self simNativeStackAt: i)
				ensureIsMarkedAsSpilled].
		simNativeSpillBase := simNativeStackPtr + 1].
	^ 0! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'eem 1/4/2019 16:46'!
scanBlock: blockStart
	"Scan the block to determine if the block needs a frame or not"
	| descriptor pc end framelessStackDelta nExts pushingNils numPushNils |
	<var: #blockStart type: #'BlockStart *'>
	<var: #descriptor type: #'BytecodeDescriptor *'>
	needsFrame := false.
	LowcodeVM ifTrue: [ hasNativeFrame := false ].
	prevBCDescriptor := nil.
	methodOrBlockNumArgs := blockStart numArgs.
	inBlock := InVanillaBlock.
	pc := blockStart startpc.
	end := blockStart startpc + blockStart span.
	framelessStackDelta := nExts := extA := numExtB := extB := 0.
	pushingNils := true.
	[pc < end] whileTrue:
		[byte0 := (objectMemory fetchByte: pc ofObject: methodObj) + bytecodeSetOffset.
		 descriptor := self generatorAt: byte0.
		 descriptor isExtension ifTrue:
			[self loadSubsequentBytesForDescriptor: descriptor at: pc.
			 self perform: descriptor generator].
		 needsFrame ifFalse:
			[(descriptor needsFrameFunction isNil
			  or: [self perform: descriptor needsFrameFunction with: framelessStackDelta])
				ifTrue: [needsFrame := true]
				ifFalse: [framelessStackDelta := framelessStackDelta + descriptor stackDelta]].
		 objectRepresentation maybeNoteDescriptor: descriptor blockStart: blockStart.
		 (pushingNils
		  and: [descriptor isExtension not]) ifTrue:
			["Count the initial number of pushed nils acting as temp initializers.  We can't tell
			  whether an initial pushNil is an operand reference or a temp initializer, except
			  when the pushNil is a jump target (has a fixup), which never happens:
					self systemNavigation browseAllSelect:
						[:m| | ebc |
						(ebc := m embeddedBlockClosures
									select: [:ea| ea decompile statements first isMessage]
									thenCollect: [:ea| ea decompile statements first selector]) notEmpty
						and: [(#(whileTrue whileFalse whileTrue: whileFalse:) intersection: ebc) notEmpty]]
			  or if the bytecode set has a push multiple nils bytecode.  We simply count initial nils.
			  Rarely we may end up over-estimating.  We will correct by checking the stack depth
			  at the end of the block in compileBlockBodies."
			 (numPushNils := self numPushNils: descriptor pc: pc nExts: nExts method: methodObj) > 0
				ifTrue:
					[self assert: descriptor numBytes = 1.
					 blockStart numInitialNils: blockStart numInitialNils + numPushNils]
				ifFalse:
					[pushingNils := false]].
		 pc := self nextBytecodePCFor: descriptor at: pc exts: nExts in: methodObj.
		 descriptor isExtension
			ifTrue: [nExts := nExts + 1]
			ifFalse: [nExts := extA := numExtB := extB := 0].
		 prevBCDescriptor := descriptor].
	"It would be nice of this wasn't necessary but alas we need to do the eager
	 scan for frameless methods so that we don't end up popping too much off
	 the simulated stack, e.g. for pushNil; returnTopFromBlock methods."
	needsFrame ifFalse:
		[self assert: (framelessStackDelta >= 0 and: [blockStart numInitialNils >= framelessStackDelta]).
		 blockStart numInitialNils: blockStart numInitialNils - framelessStackDelta].
	^0! !

!StackToRegisterMappingCogit methodsFor: 'compile abstract instructions' stamp: 'cb 1/3/2018 15:39'!
scanMethod
	"Scan the method (and all embedded blocks) to determine
		- what the last bytecode is; extra bytes at the end of a method are used to encode things like source pointers or temp names
		- if the method needs a frame or not
		- what are the targets of any backward branches.
		- how many blocks it creates
	 Answer the block count or on error a negative error code"
	| latestContinuation nExts descriptor pc numBlocks distance targetPC framelessStackDelta seenInstVarStore |
	<var: #descriptor type: #'BytecodeDescriptor *'>
	needsFrame := useTwoPaths := seenInstVarStore := false.
	LowcodeVM ifTrue: [ hasNativeFrame := false ].
	self maybeInitNumFixups.
	self maybeInitNumCounters.
	prevBCDescriptor := nil.
	NewspeakVM ifTrue:
		[numIRCs := 0].
	(primitiveIndex > 0
	 and: [coInterpreter isQuickPrimitiveIndex: primitiveIndex]) ifTrue:
		[^0].
	pc := latestContinuation := initialPC.
	numBlocks := framelessStackDelta := nExts := extA := numExtB := extB := 0.
	[pc <= endPC] whileTrue:
		[byte0 := (objectMemory fetchByte: pc ofObject: methodObj) + bytecodeSetOffset.
		 descriptor := self generatorAt: byte0.
		 descriptor isExtension ifTrue:
			[descriptor opcode = Nop ifTrue: "unknown bytecode tag; see Cogit class>>#generatorTableFrom:"
				[^EncounteredUnknownBytecode].
			 self loadSubsequentBytesForDescriptor: descriptor at: pc.
			 self perform: descriptor generator].
		 (descriptor isReturn
		  and: [pc >= latestContinuation]) ifTrue:
			[endPC := pc].
		
		  needsFrame ifFalse:
			[(descriptor needsFrameFunction isNil
			  or: [self perform: descriptor needsFrameFunction with: framelessStackDelta])
					ifTrue:
						["With immutability we win simply by avoiding a frame build if the receiver is young and not immutable."
						 self cppIf: IMMUTABILITY
							ifTrue: [descriptor is1ByteInstVarStore
									ifTrue: [useTwoPaths := true]
									ifFalse: [needsFrame := true. useTwoPaths := false]]
							ifFalse: [needsFrame := true. useTwoPaths := false]]
					ifFalse:
						[framelessStackDelta := framelessStackDelta + descriptor stackDelta.
						 "Without immutability we win if there are two or more stores and the receiver is new."
						 self cppIf: IMMUTABILITY
							ifTrue: []
							ifFalse:
								[descriptor is1ByteInstVarStore ifTrue:
									[seenInstVarStore
										ifTrue: [useTwoPaths := true]
										ifFalse: [seenInstVarStore := true]]]]].

		 descriptor isBranch ifTrue:
			[distance := self spanFor: descriptor at: pc exts: nExts in: methodObj.
			 targetPC := pc + descriptor numBytes + distance.
			 self maybeCountFixup: descriptor.
			 (self isBackwardBranch: descriptor at: pc exts: nExts in: methodObj)
				ifTrue: [self initializeFixupAt: targetPC]
				ifFalse:
					[latestContinuation := latestContinuation max: targetPC.
					 self maybeCountCounter]].
			
		 latestContinuation := self 
			maybeDealWithUnsafeJumpForDescriptor: descriptor 
			pc: pc 
			latestContinuation: latestContinuation.
		
		 descriptor isBlockCreation ifTrue:
			[numBlocks := numBlocks + 1.
			 distance := self spanFor: descriptor at: pc exts: nExts in: methodObj.
			 targetPC := pc + descriptor numBytes + distance.
			 latestContinuation := latestContinuation max: targetPC.
			 self maybeCountFixup: descriptor].

		 NewspeakVM ifTrue:
			[descriptor hasIRC ifTrue: [numIRCs := numIRCs + 1]].
		 pc := pc + descriptor numBytes.
		 nExts := descriptor isExtension ifTrue: [nExts + 1] ifFalse: [extA := numExtB := extB := 0].
		 prevBCDescriptor := descriptor].
	^numBlocks! !

!StackToRegisterMappingCogit methodsFor: 'initialization' stamp: 'eem 7/8/2021 14:43'!
setInterpreter: aCoInterpreter
	"Initialization of the code generator in the simulator.
	 These objects already exist in the generated C VM
	 or are used only in the simulation."
	<doNotGenerate>
	super setInterpreter: aCoInterpreter.

	methodAbortTrampolines := CArrayAccessor on: (Array new: self numRegArgs + 2).
	picAbortTrampolines := CArrayAccessor on: (Array new: self numRegArgs + 2).
	picMissTrampolines := CArrayAccessor on: (Array new: self numRegArgs + 2).

	simStack := CArrayAccessor on: ((1 to: self class simStackSlots) collect: [:i| self simStackEntryClass new cogit: self]).
	simNativeStack := CArrayAccessor on: ((1 to: self class simNativeStackSlots) collect: [:i| self simStackNativeEntryClass new cogit: self]).

	debugFixupBreaks := self class initializationOptions at: #debugFixupBreaks ifAbsent: [Set new].

	numPushNilsFunction := self class numPushNilsFunction.
	pushNilSizeFunction := self class pushNilSizeFunction.

	shortCutTrampolineBlocks
		at: #ceShortCutTraceStore:
			put: [coInterpreter
					ceTraceStoreOf: (processor registerAt: TempReg)
					into: (processor registerAt: ReceiverResultReg)]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'RonieSalgado 4/30/2016 17:57'!
simNativeStackAt: index
	<cmacro: '(index) (simNativeStack + (index))'>
	<returnTypeC: #'CogSimStackNativeEntry *'>
	^self addressOf: (simNativeStack at: index)! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 1/15/2017 21:52'!
simNativeStackSlots
	"Answer the number of slots toinclude in a simulated stack.
	 This needs to be big enough to include all slots in a context
	 plus some overflow for safety."
	<inline: true>
	^((LargeContextSlots - CtxtTempFrameStart max: 64) * 11 // 10)! !

!StackToRegisterMappingCogit methodsFor: 'accessing' stamp: 'eem 3/6/2018 16:35'!
simSelf
	<cmacro: '() simStack'>
	<returnTypeC: #'CogSimStackEntry *'>
	^self simStackAt: 0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 12/11/2010 18:01'!
simStackAt: index
	<cmacro: '(index) (simStack + (index))'>
	<returnTypeC: #'CogSimStackEntry *'>
	^self addressOf: (simStack at: index)! !

!StackToRegisterMappingCogit methodsFor: 'initialization' stamp: 'cb 4/16/2016 10:54'!
simStackEntryClass
	<doNotGenerate>
	^CogSimStackEntry! !

!StackToRegisterMappingCogit methodsFor: 'initialization' stamp: 'rsf 8/9/2016 23:44'!
simStackNativeEntryClass
	<doNotGenerate>
	^CogSimStackNativeEntry! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 1/24/2018 12:08'!
simStackPrintString
	<doNotGenerate>
	^String streamContents: [:s| self printSimStack: simStack toDepth: simStackPtr spillBase: simSpillBase on: s]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/22/2018 16:51'!
simStackSlots
	"Answer the number of slots to include in a simulated stack.
	 This needs to be big enough to include all stack slots in a context
	 plus a slot for the receiver and some overflow for safety."
	<inline: true>
	^((LargeContextSlots - CtxtTempFrameStart + 1 max: 64) * 11 // 10)! !

!StackToRegisterMappingCogit methodsFor: 'testing' stamp: 'cb 8/17/2016 15:14'!
sistaNeedsStoreCheck
	<inline: true>
	"The store check can be ignored if the value assigned doesn't need it (immediate, etc)
	In addition, the extB low bit is marked by the optimizer if the store check is not required"
	^ self ssTopNeedsStoreCheck and: [ self extBSpecifiesStoreCheck ]! !

!StackToRegisterMappingCogit methodsFor: 'span functions' stamp: 'cb 8/30/2016 13:57'!
sistaV1: descriptor Num: pc Push: nExts Nils: aMethodObj
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<inline: true>
	^descriptor generator == #genPushConstantNilBytecode
		ifTrue: [1]
		ifFalse: [0]! !

!StackToRegisterMappingCogit methodsFor: 'span functions' stamp: 'cb 8/30/2016 11:56'!
sistaV1PushNilSize: aMethodObj numInitialNils: numInitialNils
	"230		11100110	iiiiiiii		PushNClosureTemps iiiiiiii"
	<inline: true>
	^numInitialNils! !

!StackToRegisterMappingCogit methodsFor: 'span functions' stamp: 'eem 4/7/2015 15:24'!
squeakV3orSistaV1: descriptor Num: pc Push: nExts Nils: aMethodObj
	<var: #descriptor type: #'BytecodeDescriptor *'>
	^bytecodeSetOffset = 0
		ifTrue: [self v3: descriptor Num: pc Push: nExts Nils: aMethodObj]
		ifFalse: [self sistaV1: descriptor Num: pc Push: nExts Nils: aMethodObj]! !

!StackToRegisterMappingCogit methodsFor: 'span functions' stamp: 'eem 4/2/2015 16:46'!
squeakV3orSistaV1PushNilSize: aMethodObj numInitialNils: numInitialNils
	^(coInterpreter methodUsesAlternateBytecodeSet: aMethodObj)
		ifTrue: [self sistaV1PushNilSize: aMethodObj numInitialNils: numInitialNils]
		ifFalse: [self v3PushNilSize: aMethodObj numInitialNils: numInitialNils]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 7/24/2018 19:15'!
ssAllocateCallReg: requiredReg
	"Allocate a register needed in a run-time call (i.e. flush uses of the
	 register to the real stack).  Since the run-time can smash any and
	 all caller-saved registers also flush all caller-saved registers."
	<inline: true>
	self ssAllocateRequiredRegMask: (CallerSavedRegisterMask
										bitOr: (self registerMaskFor: requiredReg))
		upThrough: simStackPtr
		upThroughNative: simNativeStackPtr! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 7/24/2018 19:15'!
ssAllocateCallReg: requiredReg1 and: requiredReg2
	"Allocate registers needed in a run-time call (i.e. flush uses of the
	 registers to the real stack).  Since the run-time can smash any and
	 all caller-saved registers also flush all caller-saved registers."
	<inline: true>
	self ssAllocateRequiredRegMask: (CallerSavedRegisterMask
										bitOr: ((self registerMaskFor: requiredReg1)
										bitOr: (self registerMaskFor: requiredReg2)))
		upThrough: simStackPtr
		upThroughNative: simNativeStackPtr! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 7/24/2018 19:15'!
ssAllocateCallReg: requiredReg1 and: requiredReg2 and: requiredReg3
	"Allocate registers needed in a run-time call (i.e. flush uses of the
	 registers to the real stack).  Since the run-time can smash any and
	 all caller-saved registers also flush all caller-saved registers."
	<inline: true>
	self ssAllocateRequiredRegMask: (CallerSavedRegisterMask bitOr:
										(self registerMaskFor: requiredReg1 and: requiredReg2 and: requiredReg3))
		upThrough: simStackPtr
		upThroughNative: simNativeStackPtr! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 7/24/2018 19:15'!
ssAllocateCallReg: requiredReg1 and: requiredReg2 and: requiredReg3 and: requiredReg4
	"Allocate registers needed in a run-time call (i.e. flush uses of the
	 registers to the real stack).  Since the run-time can smash any and
	 all caller-saved registers also flush all caller-saved registers."
	<inline: true>
	self ssAllocateRequiredRegMask: (CallerSavedRegisterMask
										bitOr: ((self registerMaskFor: requiredReg1)
										bitOr: ((self registerMaskFor: requiredReg2)
										bitOr: ((self registerMaskFor: requiredReg3)
										bitOr: (self registerMaskFor: requiredReg4)))))
		upThrough: simStackPtr
		upThroughNative: simNativeStackPtr! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 8/10/2016 00:22'!
ssAllocateRequiredFloatReg: requiredReg
	self ssAllocateRequiredFloatRegMask: (self registerMaskFor: requiredReg)
		upThrough: simStackPtr upThroughNative: simNativeStackPtr! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/22/2018 17:57'!
ssAllocateRequiredFloatRegMask: requiredRegsMask upThrough: stackPtr upThroughNative: nativeStackPtr
	| lastRequired lastRequiredNative liveRegs |
	lastRequired := -1.
	lastRequiredNative := -1.
	"compute live regs while noting the last occurrence of required regs.
	 If these are not free we must spill from simSpillBase to last occurrence.
	 Note we are conservative here; we could allocate FPReg in frameless methods."
	liveRegs := NoReg.
	(simSpillBase max: 0) to: stackPtr do:
		[:i|
		liveRegs := liveRegs bitOr: (self simStackAt: i) registerMask.
		((self simStackAt: i) floatRegisterMask bitAnd: requiredRegsMask) ~= 0 ifTrue:
			[lastRequired := i]].
	self assert: lastRequiredNative = simNativeStackPtr.
	(simNativeSpillBase max: 0) to: nativeStackPtr do:
		[:i|
		liveRegs := liveRegs bitOr: (self simNativeStackAt: i) nativeRegisterMask.
		((self simNativeStackAt: i) floatRegisterMask bitAnd: requiredRegsMask) ~= 0 ifTrue:
			[lastRequiredNative := i]].

	"If any of requiredRegsMask are live we must spill."
	(liveRegs bitAnd: requiredRegsMask) = 0 ifFalse:
		["Some live, must spill"
		self ssFlushTo: lastRequired.
		self assert: (self liveFloatRegisters bitAnd: requiredRegsMask) = 0]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 7/24/2018 19:16'!
ssAllocateRequiredReg: requiredReg
	<inline: true>
	self ssAllocateRequiredRegMask: (self registerMaskFor: requiredReg)
		upThrough: simStackPtr
		upThroughNative: simNativeStackPtr! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 7/24/2018 19:15'!
ssAllocateRequiredReg: requiredReg1 and: requiredReg2
	<inline: true>
	self ssAllocateRequiredRegMask: ((self registerMaskFor: requiredReg1)
										bitOr: (self registerMaskFor: requiredReg2))
		upThrough: simStackPtr
		upThroughNative: simNativeStackPtr! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 7/24/2018 19:16'!
ssAllocateRequiredReg: requiredReg upThrough: stackPtr
	<inline: true>
	self ssAllocateRequiredRegMask: (self registerMaskFor: requiredReg)
		upThrough: stackPtr
		upThroughNative: simNativeStackPtr! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 2/6/2018 13:46'!
ssAllocateRequiredRegMask: requiredRegsMask upThrough: stackPtr upThroughNative: nativeStackPtr
	| lastRequired lastRequiredNative liveRegs |
	lastRequired := -1.
	lastRequiredNative := -1.
	"compute live regs while noting the last occurrence of required regs.
	 If these are not free we must spill from simSpillBase to last occurrence.
	 Note we are conservative here; we could allocate FPReg in frameless methods."
	liveRegs := self registerMaskFor: FPReg and: SPReg.
	(simSpillBase max: 0) to: stackPtr do:
		[:i|
		liveRegs := liveRegs bitOr: (self simStackAt: i) registerMask.
		((self simStackAt: i) registerMask anyMask: requiredRegsMask) ifTrue:
			[lastRequired := i]].
	LowcodeVM ifTrue:
		[self assert: nativeStackPtr = simNativeStackPtr.
		 (simNativeSpillBase max: 0) to: nativeStackPtr do:
			[:i|
			liveRegs := liveRegs bitOr: (self simNativeStackAt: i) nativeRegisterMask.
			((self simNativeStackAt: i) nativeRegisterMask anyMask: requiredRegsMask) ifTrue:
				[lastRequiredNative := i]]].
	"If any of requiredRegsMask are live we must spill."
	(liveRegs anyMask: requiredRegsMask) ifTrue:
		[self ssFlushTo: lastRequired.
		 self deny: (self liveRegisters anyMask: requiredRegsMask)]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/22/2018 17:50'!
ssFlushAll
	<inline: true>
	self ssFlushTo: simStackPtr.
	self voidReceiverResultRegContainsSelf.! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 3/9/2018 12:33'!
ssFlushTo: index
	<inline: true>
	self assert: self tempsValidAndVolatileEntriesSpilled.
	LowcodeVM ifTrue:
		[self ssNativeFlushTo: simNativeStackPtr].
	simSpillBase <= index ifTrue:
		[(((simSpillBase max: methodOrBlockNumTemps + 1) min: simStackPtr) min: index) to: index do:
			[:i|
			self assert: needsFrame.
			(self simStackAt: i)
				ensureSpilledAt: (self frameOffsetOfTemporary: i - 1) "frameOffsetOfTemporary: is 0-relative"
				from: FPReg].
		 simSpillBase := index + 1]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/25/2018 18:16'!
ssFlushUpThrough: unaryBlock
	"Any occurrences on the stack of the value being stored (which is the top of stack)
	 must be flushed, and hence any values colder than them stack."
	<inline: true>
	self assert: simSpillBase >= 0.
	simStackPtr - 1 to: simSpillBase by: -1 do:
		[ :index |
		(unaryBlock value: (self simStackAt: index)) ifTrue: [ ^ self ssFlushTo: index ] ]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 09/27/2016 11:20'!
ssFlushUpThroughReceiverVariable: slotIndex
	"Any occurrences on the stack of the value being stored (which is the top of stack)
	 must be flushed, and hence any values colder than them stack."
	<var: #desc type: #'CogSimStackEntry *'>
	LowcodeVM ifTrue: [ self ssNativeFlushTo: simNativeStackPtr. ].
	self ssFlushUpThrough: 
		[ :desc |  
			desc type = SSBaseOffset
			 and: [desc register = ReceiverResultReg
			 and: [desc offset = (objectRepresentation slotOffsetOfInstVarIndex: slotIndex) ] ] ]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 09/27/2016 11:20'!
ssFlushUpThroughRegister: reg
	"Any occurrences on the stack of the register must be
	 flushed, and hence any values colder than them stack."
	<var: #desc type: #'CogSimStackEntry *'>
	LowcodeVM ifTrue: [ self ssNativeFlushTo: simNativeStackPtr ].
	self ssFlushUpThrough: [ :desc | desc type = SSRegister and: [ desc register = reg ] ]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/26/2018 16:07'!
ssFlushUpThroughTemporaryVariable: tempIndex
	"Any occurrences on the stack of the value being stored (which is the top of stack)
	 must be flushed, and hence any values colder than them stack."
	<var: #desc type: #'CogSimStackEntry *'>
	| offset |
	LowcodeVM ifTrue: [ self ssNativeFlushTo: simNativeStackPtr ].
	offset := (self simStackAt: tempIndex + 1) offset.
	self assert: offset = (self frameOffsetOfTemporary: tempIndex).
	self ssFlushUpThrough: 
		[ :desc |
			desc type = SSBaseOffset
		 	and: [desc register = FPReg
		 	and: [desc offset = offset ] ] ]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 9/2/2016 04:22'!
ssNativeFlushTo: index
	<option: #LowcodeVM>
	| loadedPointer allocatedScratchRegister |
	simNativeSpillBase <= index ifTrue: [
		loadedPointer := false.
		allocatedScratchRegister := false.

		(simNativeSpillBase max: 0) to: index do:  [:i|
			loadedPointer ifFalse: [
				self MoveMw: self frameOffsetOfNativeFramePointer r: FPReg R: TempReg.
				loadedPointer := true.
			].

			((self simNativeStackAt: i) spillingNeedsScratchRegister and: [allocatedScratchRegister not ]) ifTrue: [
				self PushR: FPReg.
				allocatedScratchRegister := true.
			].

			(self simNativeStackAt: i)
				ensureSpilledSP: TempReg scratchRegister: FPReg
		].

		simNativeSpillBase := index + 1.

		allocatedScratchRegister ifTrue: [
			self PopR: FPReg.
		].

		loadedPointer ifTrue: [
			self SubCq: simNativeStackSize R: TempReg.
			self MoveR: TempReg Mw: self frameOffsetOfNativeStackPointer r: FPReg.
		]
	]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 8/10/2016 00:13'!
ssNativePop: n
	<option: #LowcodeVM>
	self assert: (simNativeStackPtr - n >= -1).
	simNativeStackPtr := simNativeStackPtr - n.
	simNativeStackPtr >= 0 ifTrue: [
		simNativeStackSize := self ssNativeTop offset
	] ifFalse: [
		simNativeStackSize := 0.
	]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 8/10/2016 00:13'!
ssNativePush: n 
	<option: #LowcodeVM>
	simNativeStackPtr := simNativeStackPtr + n! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 3/21/2017 16:54'!
ssNativeTop
	<option: #LowcodeVM>
	<inline: true>
	<returnTypeC: #'CogSimStackNativeEntry *'>
	self assert: simNativeStackPtr >= 0.
	^self simNativeStackAt: simNativeStackPtr! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 3/21/2017 21:34'!
ssNativeValue: n
	<returnTypeC: #'CogSimStackNativeEntry *'>
	<option: #LowcodeVM>
	^self simNativeStackAt: simNativeStackPtr - n! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/26/2018 08:59'!
ssPop: n
	self assert: (simStackPtr - n >= methodOrBlockNumTemps
				or: [(needsFrame not and: [simStackPtr - n >= 0])
				or: [self maybeCompilingFirstPassOfBlockWithInitialPushNil]]).
	simStackPtr := simStackPtr - n.
	self updateSimSpillBase! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 9/8/2016 16:58'!
ssPopNativeSize: popSize
	<option: #LowcodeVM>
	| poppingSize stackPosition popCount |
	poppingSize := 0.
	stackPosition := simNativeStackPtr.
	[ poppingSize < popSize and: [ stackPosition >= 0 ] ] whileTrue: [
		poppingSize := poppingSize + (self simNativeStackAt: stackPosition) stackSpillSize.
		stackPosition := stackPosition - 1
	].

	self assert: poppingSize = popSize.
	popCount := simNativeStackPtr - stackPosition.
	self ssNativePop: popCount.! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 11/20/2010 14:41'!
ssPush: n 
	simStackPtr := simStackPtr + n! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/26/2018 09:27'!
ssPushAnnotatedConstant: literal
	self ssPushConstant: literal.
	self annotateInstructionForBytecode.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/26/2018 09:28'!
ssPushBase: reg offset: offset
	self ssPush: 1.
	self ssTop
		type: SSBaseOffset;
		spilled: false;
		register: reg;
		offset: offset;
		bcptr: bytecodePC.
	self updateSimSpillBase.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/26/2018 09:27'!
ssPushConstant: literal
	self ssPush: 1.
	self ssTop
		type: SSConstant;
		spilled: false;
		constant: literal;
		bcptr: bytecodePC.
	self updateSimSpillBase.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'cb 4/18/2016 18:24'!
ssPushDesc: simStackEntry
	<var: #simStackEntry type: #SimStackEntry>
	self cCode:
			[simStackEntry type = SSSpill ifTrue:
				[simStackEntry type: SSBaseOffset].
			simStackEntry
				spilled: false;
				bcptr: bytecodePC.
			 simStack
				at: (simStackPtr := simStackPtr + 1)
				put: simStackEntry]
		inSmalltalk:
			[(simStack at: (simStackPtr := simStackPtr + 1))
				copyFrom: simStackEntry;
				type: (simStackEntry type = SSSpill
						ifTrue: [SSBaseOffset]
						ifFalse: [simStackEntry type]);
				spilled: false;
				bcptr: bytecodePC].
	self updateSimSpillBase.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 8/10/2016 00:14'!
ssPushNativeConstantFloat32: aFloat32
	<option: #LowcodeVM>
	<var: #aFloat32 type: #float>
	self ssNativePush: 1.
	simNativeSpillBase > simNativeStackPtr ifTrue:
		[simNativeSpillBase := simNativeStackPtr max: 0].
	simNativeStackSize := simNativeStackSize + BytesPerWord.
	self ssNativeTop
		type: SSConstantFloat32;
		spilled: false;
		offset: simNativeStackSize;
		constantFloat32: aFloat32;
		bcptr: bytecodePC.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 8/10/2016 00:14'!
ssPushNativeConstantFloat64: aFloat64
	<option: #LowcodeVM>
	<var: #aFloat64 type: #double>
	self ssNativePush: 1.
	simNativeSpillBase > simNativeStackPtr ifTrue:
		[simNativeSpillBase := simNativeStackPtr max: 0].
	simNativeStackSize := simNativeStackSize + 8.
	self ssNativeTop
		type: SSConstantFloat64;
		spilled: false;
		offset: simNativeStackSize;
		constantFloat64: aFloat64;
		bcptr: bytecodePC.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 8/10/2016 00:14'!
ssPushNativeConstantInt32: anInt32
	<option: #LowcodeVM>
	self ssNativePush: 1.
	simNativeSpillBase > simNativeStackPtr ifTrue:
		[simNativeSpillBase := simNativeStackPtr max: 0].
	simNativeStackSize := simNativeStackSize + BytesPerWord.
	self ssNativeTop
		type: SSConstantInt32;
		spilled: false;
		offset: simNativeStackSize;
		constantInt32: anInt32;
		bcptr: bytecodePC.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 8/10/2016 00:14'!
ssPushNativeConstantInt64: anInt64
	<option: #LowcodeVM>
	<var: #anInt64 type: #sqLong>
	self ssNativePush: 1.
	simNativeSpillBase > simNativeStackPtr ifTrue:
		[simNativeSpillBase := simNativeStackPtr max: 0].
	simNativeStackSize := simNativeStackSize + 8.
	self ssNativeTop
		type: SSConstantInt64;
		spilled: false;
		offset: simNativeStackSize;
		constantInt64: anInt64;
		bcptr: bytecodePC.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 8/10/2016 00:14'!
ssPushNativeConstantPointer: aNativePointer
	<option: #LowcodeVM>
	self ssNativePush: 1.
	simNativeSpillBase > simNativeStackPtr ifTrue:
		[simNativeSpillBase := simNativeStackPtr max: 0].
	simNativeStackSize := simNativeStackSize + BytesPerWord.
	self ssNativeTop
		type: SSConstantNativePointer;
		spilled: false;
		offset: simNativeStackSize;
		constantNativePointer: aNativePointer;
		bcptr: bytecodePC.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 8/10/2016 00:14'!
ssPushNativeRegister: reg
	<option: #LowcodeVM>
	self ssNativePush: 1.
	simNativeSpillBase > simNativeStackPtr ifTrue:
		[simNativeSpillBase := simNativeStackPtr max: 0].
	simNativeStackSize := simNativeStackSize + BytesPerWord.
	self ssNativeTop
		type: SSNativeRegister;
		spilled: false;
		offset: simNativeStackSize;
		register: reg;
		bcptr: bytecodePC.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 8/10/2016 00:14'!
ssPushNativeRegister: reg secondRegister: secondReg
	<option: #LowcodeVM>
	self ssNativePush: 1.
	simNativeSpillBase > simNativeStackPtr ifTrue:
		[simNativeSpillBase := simNativeStackPtr max: 0].
	simNativeStackSize := simNativeStackSize + 8.
	self ssNativeTop
		type: SSRegisterPair;
		spilled: false;
		offset: simNativeStackSize;
		register: reg;
		registerSecond: secondReg;
		bcptr: bytecodePC.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 8/10/2016 00:14'!
ssPushNativeRegisterDoubleFloat: reg
	<option: #LowcodeVM>
	self ssNativePush: 1.
	simNativeSpillBase > simNativeStackPtr ifTrue:
		[simNativeSpillBase := simNativeStackPtr max: 0].
	simNativeStackSize := simNativeStackSize + 8.
	self ssNativeTop
		type: SSRegisterDoubleFloat;
		spilled: false;
		offset: simNativeStackSize;
		register: reg;
		bcptr: bytecodePC.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'rsf 8/10/2016 00:14'!
ssPushNativeRegisterSingleFloat: reg
	<option: #LowcodeVM>
	self ssNativePush: 1.
	simNativeSpillBase > simNativeStackPtr ifTrue:
		[simNativeSpillBase := simNativeStackPtr max: 0].
	simNativeStackSize := simNativeStackSize + BytesPerWord.
	self ssNativeTop
		type: SSRegisterSingleFloat;
		spilled: false;
		offset: simNativeStackSize;
		register: reg;
		bcptr: bytecodePC.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 1/26/2018 09:26'!
ssPushRegister: reg
	self ssPush: 1.
	self ssTop
		type: SSRegister;
		spilled: false;
		register: reg;
		bcptr: bytecodePC.
	self updateSimSpillBase.
	^0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 3/6/2018 17:32'!
ssSelfDescriptor
	<returnTypeC: #SimStackEntry>
	<inline: true>
	^simStack at: 0! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'cb 1/19/2016 11:40'!
ssStoreAndReplacePop: popBoolean toReg: reg
	"In addition to ssStorePop:toReg:, if this is a store and not
	a popInto I change the simulated stack to use the register 
	for the top value"
	| topSpilled |
	topSpilled := self ssTop spilled.
	self ssStorePop: (popBoolean or: [topSpilled]) toReg: reg.
	popBoolean ifFalse: 
		[ topSpilled ifFalse: [self ssPop: 1].
		self ssPushRegister: reg ].! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'cb 4/18/2016 16:02'!
ssStorePop: popBoolean toPreferredReg: preferredReg
	"Store or pop the top simulated stack entry to a register.
	 Use preferredReg if the entry is not itself a register.
	 Answer the actual register the result ends up in."
	| actualReg |
	actualReg := preferredReg.
	self ssTop type = SSRegister ifTrue: 
		[self assert: self ssTop spilled not.
		 actualReg := self ssTop register].
	self ssStorePop: popBoolean toReg: actualReg. "generates nothing if ssTop is already in actualReg"
	^ actualReg! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'cb 4/23/2015 11:13'!
ssStorePop: popBoolean toReg: reg
	"Store or pop the top simulated stack entry to a register.
	N.B.: popToReg: and storeToReg: does not generate anything if 
	it moves a register to the same register."	
	popBoolean
		ifTrue: [self ssTop popToReg: reg.
				self ssPop: 1]
		ifFalse: [self ssTop storeToReg: reg].! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 12/11/2010 15:00'!
ssTop
	<returnTypeC: #'CogSimStackEntry *'>
	^self simStackAt: simStackPtr! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 3/6/2018 17:32'!
ssTopDescriptor
	<returnTypeC: #SimStackEntry>
	<inline: true>
	^simStack at: simStackPtr! !

!StackToRegisterMappingCogit methodsFor: 'testing' stamp: 'eem 7/12/2016 11:18'!
ssTopNeedsStoreCheck
	<inline: true>
	^self ssTop type ~= SSConstant
	  or: [(objectMemory isNonImmediate: self ssTop constant)
		and: [objectRepresentation shouldAnnotateObjectReference:  self ssTop constant]]! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 12/11/2010 15:00'!
ssValue: n
	<returnTypeC: #'CogSimStackEntry *'>
	^self simStackAt: simStackPtr - n! !

!StackToRegisterMappingCogit methodsFor: 'testing' stamp: 'eem 1/21/2018 15:42'!
stackEntryIsBoolean: simStackEntry
	<var: #simStackEntry type: #'CogSimStackEntry *'>
	<inline: true>
	^simStackEntry type = SSConstant
	  and: [simStackEntry constant = objectMemory trueObject
			or: [simStackEntry constant = objectMemory falseObject]]! !

!StackToRegisterMappingCogit methodsFor: 'testing' stamp: 'eem 1/24/2018 12:00'!
stackTopIsBoolean
	<inline: true>
	^simStackPtr >= methodOrBlockNumArgs and: [self stackEntryIsBoolean: self ssTop]! !

!StackToRegisterMappingCogit methodsFor: 'debugging' stamp: 'eem 1/24/2018 11:29'!
tempsValidAndVolatileEntriesSpilled
	"Answer if the stack is valid up to, but not including, simSpillBase."
	| culprit |
	1 to: methodOrBlockNumTemps do:
		[:i|
		 ((self simStackAt: i) type = SSBaseOffset
		   or: [self maybeCompilingFirstPassOfBlockWithInitialPushNil]) ifFalse:
			[culprit ifNil: [culprit := i].
			 ^false]].
	methodOrBlockNumTemps + 1 to: simSpillBase - 1 do:
		[:i|
		 (self simStackAt: i) spilled ifFalse:
			[culprit ifNil: [culprit := i].
			 ^false]].
	^true! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 1/18/2017 08:32'!
traceDescriptor: descriptor
	<cmacro: '(ign) 0'>
	(compilationTrace anyMask: 2) ifTrue:
		[coInterpreter transcript cr; print: bytecodePC; space; nextPutAll: descriptor generator.
		 deadCode ifTrue: [coInterpreter transcript nextPutAll: ' => deadCode'].
		 coInterpreter flush]! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 2/12/2017 16:25'!
traceFixup: fixup merge: isMerge
	<cmacro: '(igu,ana) 0'>
	| index |
	(compilationTrace anyMask: 32) ifTrue:
		[index := (fixups object identityIndexOf: fixup) - 1.
		 coInterpreter transcript
			ensureCr;
			print: bytecodePC; nextPutAll: ' -> '; print: index; nextPut: $/; print: index + initialPC;
			nextPut: $:; space.
		 isMerge
			ifTrue: [fixup printStateOn: coInterpreter transcript]
			ifFalse: [coInterpreter transcript nextPutAll: 'non-merge'].
		 coInterpreter transcript cr; flush]! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 2/2/2018 10:46'!
traceMerge: fixup
	<cmacro: '(ign) 0'>
	| index |
	(compilationTrace anyMask: 16) ifTrue:
		[index := (fixups object identityIndexOf: fixup) - 1.
		 coInterpreter transcript
			ensureCr;
			print: index; nextPut: $/; print: index + initialPC;
			nextPut: $:; space.
			fixup printStateOn: coInterpreter transcript.
			coInterpreter transcript cr; flush]! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 1/22/2018 18:04'!
traceSimStack
	<cmacro: '() 0'>
	(compilationTrace anyMask: 4) ifTrue:
		[self printSimStack]! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 11/21/2016 14:39'!
traceSpill: simStackEntry
	<cmacro: '(ign) 0'>
	(compilationTrace anyMask: 8) ifTrue:
		[coInterpreter transcript cr; print: bytecodePC; space; print: simStackEntry; flush]! !

!StackToRegisterMappingCogit methodsFor: 'peephole optimizations' stamp: 'eem 1/21/2018 16:14'!
tryCollapseTempVectorInitializationOfSize: slots
	"If the sequence of bytecodes is
		push: (Array new: 1)
		popIntoTemp: tempIndex
		pushConstant: const or pushTemp: n
		popIntoTemp: 0 inVectorAt: tempIndex
	 collapse this into
		tempAt: tempIndex put: {const or temp}
	 and answer true, otherwise answer false.
	 One might think that we should look for a sequence of more than
	 one pushes and pops but this is extremely rare.
	 Exclude pushRcvr: n to avoid potential complications with context inst vars."
	| pushArrayDesc storeArrayDesc pushValueDesc storeValueDesc tempIndex remoteTempIndex reg |
	<var: #pushArrayDesc type: #'BytecodeDescriptor *'>
	<var: #pushValueDesc type: #'BytecodeDescriptor *'>
	<var: #storeArrayDesc type: #'BytecodeDescriptor *'>
	<var: #storeValueDesc type: #'BytecodeDescriptor *'>
	slots ~= 1 ifTrue:
		[^false].
	pushArrayDesc := self generatorForPC: bytecodePC.
	self assert: pushArrayDesc generator == #genPushNewArrayBytecode.
	storeArrayDesc := self generatorForPC: bytecodePC + pushArrayDesc numBytes.
	storeArrayDesc generator == #genStoreAndPopTemporaryVariableBytecode
		ifTrue:
			[tempIndex := (objectMemory
								fetchByte: bytecodePC + pushArrayDesc numBytes
								ofObject: methodObj) bitAnd: 16r7]
		ifFalse:
			[storeArrayDesc generator == #genLongStoreAndPopTemporaryVariableBytecode ifFalse:
				[^false].
			 tempIndex := objectMemory
								fetchByte: bytecodePC + pushArrayDesc numBytes + 1
								ofObject: methodObj].
	pushValueDesc := self generatorForPC: bytecodePC
										+ pushArrayDesc numBytes
										+ storeArrayDesc numBytes.
	(pushValueDesc generator == #genPushLiteralConstantBytecode
	 or: [pushValueDesc generator == #genPushQuickIntegerConstantBytecode
	 or: [pushValueDesc generator == #genPushTemporaryVariableBytecode]]) ifFalse:
		[^false].
	storeValueDesc := self generatorForPC: bytecodePC
											+ pushArrayDesc numBytes
											+ storeArrayDesc numBytes
											+ pushValueDesc numBytes.
	remoteTempIndex := objectMemory
									fetchByte: bytecodePC
											+ pushArrayDesc numBytes
											+ storeArrayDesc numBytes
											+ pushValueDesc numBytes
											+ 2
									ofObject: methodObj.
	(storeValueDesc generator == #genStoreAndPopRemoteTempLongBytecode
	 and: [tempIndex = remoteTempIndex]) ifFalse:
		[^false].

	objectRepresentation genNewArrayOfSize: 1 initialized: false.
	self evaluate: pushValueDesc at: bytecodePC + pushArrayDesc numBytes + storeArrayDesc numBytes.
	reg := self ssStorePop: true toPreferredReg: TempReg.
	objectRepresentation
		genStoreSourceReg: reg
		slotIndex: 0
		intoNewObjectInDestReg: ReceiverResultReg.
	self ssPushRegister: ReceiverResultReg.
	self evaluate: storeArrayDesc at: bytecodePC + pushArrayDesc numBytes.
	bytecodePC := bytecodePC
					"+ pushArrayDesc numBytes this gets added by nextBytecodePCFor:at:exts:in:"
					+ storeArrayDesc numBytes
					+ pushValueDesc numBytes
					+ storeValueDesc numBytes.
	^true! !

!StackToRegisterMappingCogit methodsFor: 'simulation stack' stamp: 'eem 3/10/2018 13:08'!
updateSimSpillBase
	"Something volatile has been pushed on the stack; update simSpillBase accordingly."
	<inline: true>
	self assert: ((simSpillBase > methodOrBlockNumTemps
				and: [simStackPtr >= methodOrBlockNumTemps])
				or: [self maybeCompilingFirstPassOfBlockWithInitialPushNil]).
	simSpillBase > simStackPtr
		ifTrue:
			[simSpillBase := simStackPtr + 1.
			 [simSpillBase - 1 > methodOrBlockNumTemps
			   and: [(self simStackAt: simSpillBase - 1) spilled not]] whileTrue:
				[simSpillBase := simSpillBase - 1]]
		ifFalse:
			[[(self simStackAt: simSpillBase) spilled
			   and: [simSpillBase <= simStackPtr]] whileTrue:
				[simSpillBase := simSpillBase + 1]].
	methodOrBlockNumTemps + 1 to: (simSpillBase - 1 min: simStackPtr) do:
		[:i|
		self assert: (self simStackAt: i) spilled == true].
	self assert: (simSpillBase > simStackPtr or: [(self simStackAt: simSpillBase) spilled == false])! !

!StackToRegisterMappingCogit methodsFor: 'span functions' stamp: 'eem 7/29/2014 07:17'!
v3: descriptor Num: pc Push: nExts Nils: aMethodObj
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<inline: true>
	^descriptor generator == #genPushConstantNilBytecode
		ifTrue: [1]
		ifFalse: [0]! !

!StackToRegisterMappingCogit methodsFor: 'span functions' stamp: 'eem 4/2/2015 16:47'!
v3PushNilSize: aMethodObj numInitialNils: numInitialNils
	<inline: true>
	^numInitialNils! !

!StackToRegisterMappingCogit methodsFor: 'span functions' stamp: 'eem 4/7/2015 15:24'!
v3or4: descriptor Num: pc Push: nExts Nils: aMethodObj
	<var: #descriptor type: #'BytecodeDescriptor *'>
	^bytecodeSetOffset = 0
		ifTrue: [self v3: descriptor Num: pc Push: nExts Nils: aMethodObj]
		ifFalse: [self v4: descriptor Num: pc Push: nExts Nils: aMethodObj]! !

!StackToRegisterMappingCogit methodsFor: 'span functions' stamp: 'eem 4/2/2015 16:47'!
v3or4PushNilSize: aMethodObj numInitialNils: numInitialNils
	^(coInterpreter methodUsesAlternateBytecodeSet: aMethodObj)
		ifTrue: [self v4PushNilSize: aMethodObj numInitialNils: numInitialNils]
		ifFalse: [self v3PushNilSize: aMethodObj numInitialNils: numInitialNils]! !

!StackToRegisterMappingCogit methodsFor: 'span functions' stamp: 'eem 7/29/2014 07:24'!
v4: descriptor Num: pc Push: nExts Nils: aMethodObj
	"*	77			01001101		Push false [* 1:true, 2:nil, 3:thisContext, ..., -N: pushExplicitOuter: N, N = Extend B]"
	<var: #descriptor type: #'BytecodeDescriptor *'>
	<inline: true>
	^(descriptor generator == #genExtPushPseudoVariableOrOuterBytecode
	   and: [self assert: (objectMemory fetchByte: pc ofObject: aMethodObj) = 77.
			nExts = 1
	   and: [(objectMemory fetchByte: pc - 1 ofObject: aMethodObj) = 2]])
		ifTrue: [1]
		ifFalse: [0]! !

!StackToRegisterMappingCogit methodsFor: 'span functions' stamp: 'eem 4/2/2015 16:48'!
v4PushNilSize: aMethodObj numInitialNils: numInitialNils
	"77			01001101				Push false [* 1:true, 2:nil, 3:thisContext, ..., -N: pushExplicitOuter: N, N = Extend B]
	 225		11100001	sbbbbbbb	Extend B (Ext B = Ext B prev * 256 + Ext B)"
	<inline: true>
	^3 * numInitialNils! !

!StackToRegisterMappingCogit methodsFor: 'testing' stamp: 'eem 1/24/2018 13:35'!
violatesEnsureSpilledSpillAssert
	^true! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/22/2018 17:52'!
voidReceiverOptStatus
	"Used to mark ReceiverResultReg as dead or not containing simSelf.
	 Used when the simStack has already been flushed, e.g. for sends."
	<inline: true>
	self simSelf liveRegister: NoReg! !

!StackToRegisterMappingCogit methodsFor: 'bytecode generator support' stamp: 'eem 1/25/2018 09:21'!
voidReceiverResultRegContainsSelf
	"Used when ReceiverResultReg is allocated for other than simSelf, and
	 there may be references to ReceiverResultReg which need to be spilled."
	| spillIndex |
	self voidReceiverOptStatus.
	spillIndex := 0.
	(methodOrBlockNumTemps + 1 max: simSpillBase) to: simStackPtr do:
		[:i|
		(self simStackAt: i) registerOrNone = ReceiverResultReg ifTrue:
			[spillIndex := i]].
	spillIndex > 0 ifTrue:
		[self ssFlushTo: spillIndex]! !

!StackToRegisterMappingCogit methodsFor: 'simulation only' stamp: 'eem 12/6/2016 13:36'!
widthInDefaultFontOf: aString
	<doNotGenerate>
	^(NewParagraph new
		compose: aString asText
		style: TextStyle default
		from: 1
		in: Display boundingBox;
		adjustRightX)
		extent x! !

StackToRegisterMappingCogit subclass: #StackToRegisterMappingCogitChecker
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!StackToRegisterMappingCogitChecker methodsFor: 'compile abstract instructions' stamp: 'cb 8/30/2016 12:00'!
scanBlock: blockStart
	| result |
	 (result := super scanBlock: blockStart) < 0 ifTrue: [^result].
	needsFrame ifTrue:
		[| tempCount |
		 tempCount := coInterpreter tempCountForBlockStartingAt: blockStart startpc + 1 in: methodObj.
		 blockStart numInitialNils ~= tempCount ifTrue:
			[blockStart numInitialNils: tempCount]].
	^0! !

CogClass subclass: #CurrentImageCoInterpreterFacade
	instanceVariableNames: 'memory cogit coInterpreter objectMemory objectMap headerToMethodMap cachedObject cachedOop variables cFramePointer cStackPointer'
	classVariableNames: ''
	poolDictionaries: 'CogMethodConstants VMBasicConstants VMBytecodeConstants VMObjectIndices VMSqueakClassIndices'
	category: 'VMMaker-Support'!

!CurrentImageCoInterpreterFacade commentStamp: 'eem 8/6/2014 14:59' prior: 0!
A CurrentImageCoInterpreterFacade is a stand-in for an object memory (ObjectMemory, SpurMemoryManager, etc) that allows the Cogits to access image objects as if they were in the simulator VM's heap.  hence it allows the Cogits to generate code for methdos in the current image, for testing, etc.

Instance Variables
	cachedObject:			<Object>
	cachedOop:			<Integer>
	coInterpreter:			<CoInterpreter>
	cogit:					<Cogit>
	headerToMethodMap:	<Dictionary>
	memory:				<ByteArray>
	objectMap:				<IdentityDictionary>
	objectMemory:			<NewObjectMemory|SpurMemoryManager>
	variables:				<Dictionary>

cachedObject
	- the object matching cachedOop, to speed-up oop to obejct mapping

cachedOop
	- the last used oop

coInterpreter
	- the CoInterpreter simulator used by the cogit.

cogit
	- the code egnerator in use

headerToMethodMap
	- a map from header to CompiledMethod

memory
	- a rump memory for holding various interpreter variables (e.g. stackLimit) that are accessed as memory locations by generated code

objectMap
	- map from objects to their oops

objectMemory
	- the object memory used to encode various values, answer queries, etc

variables
	- a map from the names of variables to their addresses in memory
!

!CurrentImageCoInterpreterFacade class methodsFor: 'translation' stamp: 'eem 2/4/2020 09:30'!
clusteredVariableNames
	^CoInterpreter clusteredVariableNames! !

!CurrentImageCoInterpreterFacade class methodsFor: 'instance creation' stamp: 'eem 12/4/2014 16:17'!
forCogit: aCogit
	| class |
	class := self allSubclasses
				detect: [:subclass|
						aCogit class objectMemoryClass objectRepresentationClass
						== subclass objectRepresentationClass]
				ifNone: [self error: 'cannot find subclass for the Cogit''s objectRepresentation and/or objectMemory'].
	^class new
		cogit: aCogit;
		yourself! !

!CurrentImageCoInterpreterFacade class methodsFor: 'accessing class hierarchy' stamp: 'eem 6/1/2014 12:15'!
objectMemoryClass
	^self subclassResponsibility! !

!CurrentImageCoInterpreterFacade class methodsFor: 'accessing class hierarchy' stamp: 'eem 10/9/2013 17:52'!
objectRepresentationClass
	^self subclassResponsibility! !

!CurrentImageCoInterpreterFacade methodsFor: 'labels' stamp: 'eem 9/23/2015 08:52'!
addLabel: l
	(variables includesKey: l) ifFalse:
		[variables at: l put: variables size * objectMemory wordSize + 65536]! !

!CurrentImageCoInterpreterFacade methodsFor: 'debug support' stamp: 'eem 3/18/2013 11:50'!
addressCouldBeObj: address
	^(address bitAnd: 3) = 0
	  and: [self addressCouldBeOop: address]! !

!CurrentImageCoInterpreterFacade methodsFor: 'debug support' stamp: 'eem 11/15/2012 19:36'!
addressCouldBeOop: anOop 
	[self objectForOop: anOop]
		on: Error
		do: [:ex| ^false].
	^true! !

!CurrentImageCoInterpreterFacade methodsFor: 'labels' stamp: 'eem 2/4/2020 09:37'!
addressForLabel: l
	^variables at: l ifAbsentPut: [variables size + 1 * objectMemory wordSize + self variablesBase]! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/5/2015 14:25'!
alternateHeaderNumLiteralsMask
	^coInterpreter alternateHeaderNumLiteralsMask! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/8/2011 12:56'!
argumentCountAddress
	^self addressForLabel: #argumentCount! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/8/2011 14:43'!
argumentCountOf: anOop
	^(self objectForOop: anOop) numArgs! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 5/14/2015 18:08'!
argumentCountOfMethodHeader: headerIntegerOop
	^(headerToMethodMap at: headerIntegerOop) numArgs
	"a.k.a.
	^coInterpreter argumentCountOfMethodHeader: headerIntegerOop,
	 but only if the CompiledMethod header formats match"! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 4/2/2014 15:08'!
baseHeaderSize
	^self subclassResponsibility! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/25/2014 15:51'!
byteAt: index
	^objectMemory byteAt: index! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/25/2014 15:51'!
byteAt: index put: value
	^objectMemory byteAt: index put: value! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 7/28/2014 07:39'!
byteSizeOf: anOop
	| obj elementSize |
	obj := self objectForOop: anOop.
	([obj class isImmediateClass]
		on: MessageNotUnderstood
		do: [:ex| obj class == SmallInteger]) ifTrue:
		[^0].
	elementSize := 
		[obj class elementSize]
			on: MessageNotUnderstood
			do: [:ex| obj class isBytes ifTrue: [1] ifFalse: [Smalltalk wordSize]].
	^obj basicSize * elementSize! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/4/2014 16:22'!
bytesPerOop
	^objectMemory bytesPerOop! !

!CurrentImageCoInterpreterFacade methodsFor: 'debug support' stamp: 'cb 4/16/2016 19:20'!
cCoerceSimple: value to: cTypeString
	"Type coercion for translation and simulation.
	 For simulation answer a suitable surrogate for the struct types"
	^cTypeString
		caseOf:
		   {	[#'CogMethod *']		->	[value < 0 ifTrue: [value] ifFalse: [cogit cogMethodSurrogateAt: value asUnsignedInteger]].
			[#'CogBlockMethod *']	->	[cogit cogBlockMethodSurrogateAt: value asUnsignedInteger] }
		otherwise: [super cCoerceSimple: value to: cTypeString]! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/16/2019 23:40'!
cFramePointerAddress
	^self addressForLabel: #CFramePointer! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/30/2020 20:03'!
cReturnAddressAddress
	^self addressForLabel: #CReturnAddress! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/16/2019 23:40'!
cStackPointerAddress
	^self addressForLabel: #CStackPointer! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 1/8/2011 13:25'!
callForCogCompiledCodeCompaction
	cogit methodZone quickClearCogCompiledCode! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/26/2021 17:34'!
canContextSwitchIfActivating: method header: header
	"Would like to do
		^coInterpreter canContextSwitchIfActivating: method header: header
	 but the bytecode access to get at the primitive number defeats us :-(, so
	 the following may well get out-of-date..."
	^(self objectForOop: method) primitive ~= PrimNumberUnwindMarker
	  and: [(self objectForOop: method) primitive ~= PrimNumberNoContextSwitchMarker]! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 8/4/2021 18:47'!
ceTakeProfileSample: aCogMethod
	self placeholder! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/5/2016 15:13'!
characterObjectOf: anInteger
	^objectMemory characterObjectOf: anInteger! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 9/9/2013 11:53'!
characterTable
	^self subclassResponsibility! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 8/4/2021 18:48'!
checkForAndFollowForwardedPrimitiveState
	self placeholder! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 12/2/2019 16:48'!
checkIfCFramePointerInUse
	^true! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 2/23/2017 16:46'!
classCharacter
	^self oopForObject: Character! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 13:03'!
classFieldOffset
	^objectMemory classFieldOffset! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/8/2011 13:32'!
classFloat
	^self oopForObject: Float! !

!CurrentImageCoInterpreterFacade methodsFor: 'object map' stamp: 'eem 12/19/2010 14:59'!
classSmallInteger
	^self oopForObject: SmallInteger! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 1/8/2011 13:19'!
clearCogCompiledCodeCompactionCalledFor
	^self! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/9/2019 14:44'!
cogCodeSize
	^self startOfMemory! !

!CurrentImageCoInterpreterFacade methodsFor: 'initialize-release' stamp: 'eem 12/19/2010 15:07'!
cogit
	^cogit! !

!CurrentImageCoInterpreterFacade methodsFor: 'initialize-release' stamp: 'eem 7/8/2021 12:22'!
cogit: aCogit
	cogit := aCogit.
	cogit objectMemory ifNil:
		[cogit instVarNamed: 'objectMemory' put: objectMemory].
	coInterpreter cogit: aCogit.
	(objectMemory respondsTo: #cogit:) ifTrue:
		[objectMemory cogit: aCogit].
	(objectMemory respondsTo: #coInterpreter:) ifTrue:
		[objectMemory coInterpreter: coInterpreter].
	coInterpreter setUpForUseByFacade: self.
	objectMemory setUpForUseByFacade: self.
	self initializeObjectMap! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 13:02'!
compactClassFieldLSB
	^objectMemory compactClassFieldLSB! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 13:03'!
compactClassFieldWidth
	^objectMemory compactClassFieldWidth! !

!CurrentImageCoInterpreterFacade methodsFor: 'debug support' stamp: 'eem 3/20/2020 11:30'!
compilationBreakpoint: selectorOop classTag: classTag isMNUCase: isMNUCase
	^self! !

!CurrentImageCoInterpreterFacade methodsFor: 'debug support' stamp: 'eem 3/20/2020 11:26'!
compilationBreakpoint: aString isMNUCase: isMNUCase
	^self! !

!CurrentImageCoInterpreterFacade methodsFor: 'printing' stamp: 'eem 4/6/2015 14:54'!
cr
	coInterpreter transcript cr; flush! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/2/2014 09:04'!
debugStackPointersFor: anOop
	^CArrayAccessor on:
		((StackDepthFinder on: (objectMap keyAtValue: anOop))
			encoderClass: (coInterpreter encoderClassForHeader: (objectMap keyAtValue: anOop) header);
			stackPointers)! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 2/16/2011 11:40'!
eeInstantiateClass: classOop indexableSize: numSlots
	^self oopForObject: ((self objectForOop: classOop) new: numSlots)! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/5/2014 21:13'!
encoderClassForHeader: headerInteger
	^coInterpreter encoderClassForHeader: headerInteger! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/8/2011 11:28'!
endPCOf: methodOop
	^(objectMap keyAtValue: methodOop) endPC - 1! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 7/16/2014 11:21'!
ensureNoForwardedLiteralsIn: methodOop! !

!CurrentImageCoInterpreterFacade methodsFor: 'private' stamp: 'eem 8/1/2021 12:06'!
fakeLinkedExternalLiteralFor: methodOop
	| method |
	method := self objectForOop: methodOop.
	self assert: (method isCompiledMethod and: [method primitive = PrimNumberExternalCall]).
	self shouldBeImplemented! !

!CurrentImageCoInterpreterFacade methodsFor: 'object map' stamp: 'eem 12/19/2010 13:09'!
falseObject
	^objectMap at: false! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/8/2011 14:42'!
fetchByte: index ofObject: anOop
	^(self objectForOop: anOop)
		at: index + 1
		ifAbsent:
			[Transcript
				ensureCr;
				nextPutAll: 'warning, accessing past end of '; print: (objectMap keyAtValue: anOop);
				cr;
				flush.
			255]! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/11/2018 13:13'!
fetchPointer: index ofObject: anOop
	| obj |
	obj := (objectMap keyAtValue: anOop).
	^self oopForObject: (obj isCompiledCode
							ifTrue: [obj objectAt: index + 1]
							ifFalse: [obj instVarAt: index + 1])! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/15/2015 11:56'!
firstByteFormat
	^objectMemory firstByteFormat! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/15/2015 11:55'!
firstCompiledMethodFormat
	^objectMemory firstCompiledMethodFormat! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/15/2015 11:57'!
firstLongFormat
	^objectMemory firstLongFormat! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 5/9/2015 19:43'!
fixedFieldsFieldWidth
	^objectMemory fixedFieldsFieldWidth! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 7/12/2016 10:46'!
fixedFieldsOfClassFormat: classFormat
	^objectMemory fixedFieldsOfClassFormat: classFormat! !

!CurrentImageCoInterpreterFacade methodsFor: 'printing' stamp: 'eem 3/2/2016 17:01'!
flush
	coInterpreter transcript flush! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 9/9/2013 11:54'!
formatOfClass: classOop
	^self subclassResponsibility! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 14:25'!
framePointerAddress
	^self addressForLabel: #framePointer! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 7/11/2016 17:12'!
freeObject: anObj! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/8/2011 13:34'!
freeStartAddress
	^self addressForLabel: #freeStart! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/8/2011 12:52'!
functionPointerFor: primIndex inClass: lookupClass
	^primIndex = 0
		ifTrue: [#primitiveFail]
		ifFalse: [coInterpreter functionPointerFor: primIndex inClass: lookupClass]! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 7/24/2021 16:51'!
functionPointerForCompiledMethod: methodOop primitiveIndex: primIndex primitivePropertyFlagsInto: flagsPtr
	^([coInterpreter
		functionPointerForCompiledMethod: methodOop
		primitiveIndex: primIndex
		primitivePropertyFlagsInto: flagsPtr]
			on: Error
			do: [:ex|
				#someExternalPrimitive]) ifNotNil:
		[:symbol|
		self addressForLabel: symbol]! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 2/1/2020 15:59'!
getCFramePointer
	^coInterpreter getCFramePointer! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 3/31/2020 18:53'!
getCReturnAddress
	^coInterpreter getCReturnAddress! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 2/1/2020 15:59'!
getCStackPointer
	^coInterpreter getCStackPointer! !

!CurrentImageCoInterpreterFacade methodsFor: 'testing' stamp: 'eem 8/1/2021 11:50'!
hasSpurMemoryManagerAPI
	^objectMemory hasSpurMemoryManagerAPI! !

!CurrentImageCoInterpreterFacade methodsFor: 'testing' stamp: 'eem 11/1/2012 18:19'!
headerIndicatesAlternateBytecodeSet: methodHeader
	"A negative header selects the alternate bytecode set."
	^(self objectForOop: methodHeader) < 0! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 1/23/2020 15:45'!
highBit: anInteger
	^anInteger highBit! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 5/9/2015 19:43'!
identityHashHalfWordMask
	^objectMemory identityHashHalfWordMask! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/31/2021 13:05'!
ifTestProfilingAdvanceNextProfileTick
	^self! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 6/1/2014 12:11'!
indexablePointersFormat
	^objectMemory indexablePointersFormat! !

!CurrentImageCoInterpreterFacade methodsFor: 'initialize-release' stamp: 'eem 12/9/2019 14:42'!
initialize
	objectMemory := self class objectMemoryClass simulatorClass new.
	objectMemory allocateMemoryOfSize: 4*1024*1024.
	memory := objectMemory memory.
	objectMemory
		initializeFreeSpaceForFacadeFrom: self startOfMemory
		to: self variablesBase.
	coInterpreter := CoInterpreter new.
	coInterpreter
		instVarNamed: 'objectMemory'
			put: objectMemory;
		instVarNamed: 'primitiveTable'
			put: (CArrayAccessor on: CoInterpreter primitiveTable copy).
	variables := Dictionary new! !

!CurrentImageCoInterpreterFacade methodsFor: 'initialize-release' stamp: 'eem 8/16/2013 11:10'!
initializeObjectMap
	objectMap := IdentityDictionary new.
	headerToMethodMap := Dictionary new.
	{ nil. false. true. Smalltalk primitiveErrorTable. Float } do:
		[:o| self oopForObject: o]! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/22/2011 10:00'!
instFormatFieldLSB
	^objectMemory instFormatFieldLSB! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/22/2011 10:00'!
instFormatFieldWidth
	^objectMemory instFormatFieldWidth! !

!CurrentImageCoInterpreterFacade methodsFor: 'instantiation' stamp: 'eem 4/4/2014 11:59'!
instantiateClass: classPointer indexableSize: size
	(self objectForOop: classPointer) caseOf:
		{[Array]	->	[^self oopForObject: (Array new: size)]}! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 14:26'!
instructionPointerAddress
	^self addressForLabel: #instructionPointer! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 15:18'!
integerObjectOf: anInteger
	^objectMemory integerObjectOf: anInteger! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/22/2010 16:06'!
integerValueOf: anInteger
	^objectMemory integerValueOf: anInteger! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 11/11/2011 11:17'!
interpretAddress
	^self addressForLabel: #interpret! !

!CurrentImageCoInterpreterFacade methodsFor: 'testing' stamp: 'eem 8/2/2014 10:25'!
isArrayNonImm: anOop
	^(self objectForOop: anOop) class instSpec = Array instSpec! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 1/8/2011 13:18'!
isCogCompiledCodeCompactionCalledFor
	^true! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/22/2011 10:10'!
isCogMethodReference: methodHeader
	^coInterpreter isCogMethodReference: methodHeader! !

!CurrentImageCoInterpreterFacade methodsFor: 'testing' stamp: 'eem 11/2/2012 16:10'!
isCurrentImageFacade
	^true! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 4/18/2014 15:55'!
isImmediate: anOop
	^objectMemory isImmediate: anOop! !

!CurrentImageCoInterpreterFacade methodsFor: 'testing' stamp: 'eem 7/7/2015 12:35'!
isInOldSpace: address
	^objectMemory isInOldSpace: address! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 13:37'!
isIntegerObject: anOop
	^objectMemory isIntegerObject: anOop! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/22/2010 16:07'!
isIntegerValue: anInteger
	^objectMemory isIntegerValue: anInteger! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 4/18/2014 15:55'!
isNonImmediate: anOop
	^objectMemory isNonImmediate: anOop! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 13:50'!
isNonIntegerObject: anOop
	^objectMemory isNonIntegerObject: anOop! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/11/2018 13:13'!
isOopCompiledMethod: anOop
	^(objectMap keyAtValue: anOop) isCompiledCode! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 4/2/2015 16:51'!
isOptimizedMethod: methodObj
	^coInterpreter isOptimizedMethodHeader: (self methodHeaderOf: methodObj)! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 11/14/2012 15:04'!
isQuickPrimitiveIndex: primNum
	^coInterpreter isQuickPrimitiveIndex: primNum! !

!CurrentImageCoInterpreterFacade methodsFor: 'frame access' stamp: 'eem 8/16/2013 11:20'!
isReadMediatedContextInstVarIndex: index
	"Reading the sender, instructionPointer and stackPointer inst vars of a context must take
	 account of potentially married contexts and fetch the state from the frame. method,
	 closureOrNil and receiver can safely be fetched from the context without checking."
	^index <= StackPointerIndex! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 14:25'!
isThreadedVM
	^false! !

!CurrentImageCoInterpreterFacade methodsFor: 'frame access' stamp: 'eem 10/27/2014 09:59'!
isWriteMediatedContextInstVarIndex: index
	"Writing any inst vars of a context must take account of potentially married contexts
	 and set the state in the frame. Inst vars in subclasses don't need mediation; subclasses
	 can't marry."
	^index <= ReceiverIndex! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 15:18'!
isYoung: anOop
	^false! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 4/2/2014 11:12'!
isYoungObject: anOop
	^false! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/8/2011 14:43'!
lengthOf: anOop
	^(self objectForOop: anOop) basicSize! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 3/9/2018 14:21'!
literal: index ofMethod: anOop
	| lit |
	lit := (anOop isInteger ifTrue: [self objectForOop: anOop] ifFalse: [anOop]) literalAt: index + 1.
	^lit class == SmallInteger
		ifTrue: [objectMemory integerObjectOf: lit]
		ifFalse: [self oopForObject: lit]! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/15/2013 11:24'!
literalCountOf: anOop
	^(objectMap keyAtValue: anOop) numLiterals! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 5/14/2015 18:06'!
literalCountOfMethodHeader: headerIntegerOop
	^(headerToMethodMap at: headerIntegerOop) numLiterals
	"a.k.a.
	^coInterpreter literalCountOfMethodHeader: aSmallIntegerOop,
	 but only if the CompiledMethod header formats match"! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/25/2014 15:51'!
long32At: index
	^objectMemory long32At: index! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/25/2014 15:51'!
long32At: index put: value
	^objectMemory long32At: index put: value! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/25/2018 11:15'!
long64At: byteIndex
	^objectMemory long64At: byteIndex! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/25/2018 11:14'!
long64At: byteIndex put: aValue
	^objectMemory long64At: byteIndex put: aValue! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/25/2014 15:52'!
longAt: index
	^objectMemory longAt: index! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/25/2014 15:52'!
longAt: index put: value
	^objectMemory longAt: index put: value! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 5/14/2014 14:45'!
longStoreBytecodeForHeader: methodHeaderOop
	"Answer the relevant long store temp bytecode, which indicates it has a primitive error code."
	"234		11101010	i i i i i i i i	Store Temporary Variable #iiiiiiii"
	"129 		10000001 jjkkkkkk 	Store (Receiver Variable, Temporary Location, Illegal, Literal Variable) [jj] #kkkkkk"
	^coInterpreter longStoreBytecodeForHeader: methodHeaderOop
	"was: ^(headerToMethodMap at: methodHeaderOop) usesAlternateBytecodeSet
		ifTrue: [234]
		ifFalse: [129]"! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/28/2019 22:47'!
lookup32BitWordConstant: word
	^objectMemory lookup32BitWordConstant: word! !

!CurrentImageCoInterpreterFacade methodsFor: 'labels' stamp: 'eem 7/24/2018 17:24'!
lookupAddress: address
	| thing |
	thing := objectMap
				keyAtValue: address
				ifAbsent:
					[variables
						keyAtValue: address
						ifAbsent: [^nil]].
	^(((thing isLiteral
		ifTrue: [thing storeString]
		ifFalse: [thing asString]) contractTo: 64)
			copyReplaceAll: String cr with: '\r')
				copyReplaceAll: String lf with: '\n'! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 2/23/2017 16:55'!
lookupOrdinary: selectorOop receiver: receiverOop
	| rcvr selector |
	rcvr := self objectForOop: receiverOop.
	selector := self objectForOop: selectorOop.
	(rcvr class canUnderstand: selector) ifTrue:
		[^self oopForObject: ((rcvr class whichClassIncludesSelector: selector)
									compiledMethodAt: selector)].
	^SelectorDoesNotUnderstand! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 2/24/2017 13:05'!
lookupSelector: selectorOop inClass: classOop
	| class selector |
	class := self objectForOop: classOop.
	selector := self objectForOop: selectorOop.
	^(class canUnderstand: selector) ifTrue:
		[self oopForObject: ((class whichClassIncludesSelector: selector)
									compiledMethodAt: selector)]! !

!CurrentImageCoInterpreterFacade methodsFor: 'frame access' stamp: 'eem 8/23/2021 12:01'!
marryFrameCopiesTemps
	^coInterpreter marryFrameCopiesTemps! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 2/23/2017 16:56'!
maxLookupNoMNUErrorCode
	^coInterpreter maxLookupNoMNUErrorCode! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 4/12/2017 15:15'!
maybeSelectorOfMethod: methodOop
	^self oopForObject: (self objectForOop: methodOop) selector! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 14:52'!
memory
	^memory! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/9/2019 14:47'!
methodCacheAddress
	"Use the top half of memory for variables, method cache and rumpCStack,
	 and the bottom half for allocating code and objects:

		0Mb to 1Mb		code zone
		1Mb to 2Mb		object zone
		2Mb to 3Mb		variables
		3Mb to 4Mb - 1k	method cache
		4Mb - 1k to 4Mb	rump C stack"
	^memory byteSize * 3 / 4! !

!CurrentImageCoInterpreterFacade methodsFor: 'object map' stamp: 'eem 4/10/2015 12:06'!
methodClassOf: methodOop
	^self oopForObject: (self objectForOop: methodOop) methodClass! !

!CurrentImageCoInterpreterFacade methodsFor: 'testing' stamp: 'eem 12/19/2010 12:46'!
methodHasCogMethod: aCompiledMethod 
	^false! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 4/17/2015 15:13'!
methodHeaderOf: methodOop
	^self rawHeaderOf: methodOop! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 6/1/2014 12:28'!
methodNeedsLargeContext: aMethodOop 
	^(self objectForOop: aMethodOop) frameSize > CompiledMethod smallFrameSize! !

!CurrentImageCoInterpreterFacade methodsFor: 'testing' stamp: 'eem 4/6/2012 15:29'!
methodUsesAlternateBytecodeSet: methodObjOop
	"A negative header selects the alternate bytecode set."
	^(self objectForOop: methodObjOop) signFlag! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 11/21/2016 16:02'!
nameOfClass: objOop
	^(objectMap keyAtValue: objOop) name! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 2/4/2020 09:44'!
nativeSPAddress
	^self addressForLabel: #nativeSP! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'RonieSalgado 4/30/2016 20:34'!
nativeStackPointerAddress
	^self addressForLabel: #nativeStackPointer! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 14:27'!
newMethodAddress
	^self addressForLabel: #newMethod! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/5/2021 07:50'!
nextProfileTickAddress
	self addressForLabel: #nextProfileTickLow.
	self addressForLabel: #nextProfileTickHigh.
	^self addressForLabel: #nextProfileTickLow! !

!CurrentImageCoInterpreterFacade methodsFor: 'object map' stamp: 'eem 12/19/2010 13:09'!
nilObject
	^objectMap at: nil! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/16/2013 10:55'!
noAssertHeaderOf: aMethodOop
	^self rawHeaderOf: aMethodOop! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 5/9/2015 19:42'!
nonIndexablePointerFormat
	^objectMemory nonIndexablePointerFormat! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 15:08'!
nullHeaderForMachineCodeMethod
	^objectMemory nullHeaderForMachineCodeMethod! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 7/28/2014 08:06'!
numBytesOf: objOop 
	"Answer the number of indexable bytes in the given non-immediate object.
	 Does not adjust the size of contexts by stackPointer."
	| obj elementSize |
	obj := self objectForOop: objOop.
	self deny: ([obj class isImmediateClass]
				on: MessageNotUnderstood
				do: [:ex| obj class == SmallInteger]).
	elementSize := 
		[obj class elementSize]
			on: MessageNotUnderstood
			do: [:ex| obj class isBytes ifTrue: [1] ifFalse: [Smalltalk wordSize]].
	^obj basicSize * elementSize! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 5/9/2015 19:44'!
numSlotsHalfShift
	^objectMemory numSlotsHalfShift! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/25/2014 16:19'!
numSlotsOf: objOop
	"Answer the number of slots in the given non-immediate object.
	 Does not adjust the size of contexts by stackPointer."
	| obj elementSize wordSize |
	obj := self objectForOop: objOop.
	obj = objOop ifTrue:
		[^objectMemory numSlotsOf: objOop].
	self deny: ([obj class isImmediateClass]
				on: MessageNotUnderstood
				do: [:ex| obj class == SmallInteger]).
	wordSize := Smalltalk wordSize.
	elementSize := 
		[obj class elementSize]
			on: MessageNotUnderstood
			do: [:ex| obj class isBytes ifTrue: [1] ifFalse: [wordSize]].
	wordSize = 4 ifTrue:
		[^elementSize caseOf: {
			[1]	->	[obj basicSize + 3 // wordSize].
			[2]	->	[obj basicSize * 2 + 3 // wordSize].
			[4]	->	[obj basicSize + obj class instSize] }].
	^elementSize caseOf: {
		[1]	->	[obj basicSize + (wordSize - 1) // wordSize].
		[2]	->	[obj basicSize * 2 + (wordSize - 1) // wordSize].
		[4]	->	[obj basicSize * 2 + (wordSize - 1) // wordSize].
		[8]	->	[obj basicSize + obj class instSize] }! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 7/12/2016 10:34'!
objCouldBeClassObj: obj
	^(self objectForOop: obj) isBehavior! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 13:46'!
objectAfter: anOop
	^anOop + 8! !

!CurrentImageCoInterpreterFacade methodsFor: 'private-cacheing' stamp: 'eem 2/23/2017 16:52'!
objectForOop: anOop
	"This is a keyAtValue: search and so needs speeding up either by a reverse map or a simple cache."
	self subclassResponsibility! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 12:44'!
objectMemory
	^self! !

!CurrentImageCoInterpreterFacade methodsFor: 'object map' stamp: 'eem 3/22/2017 17:44'!
oopForObject: o
	self subclassResponsibility! !

!CurrentImageCoInterpreterFacade methodsFor: 'testing' stamp: 'eem 2/9/2011 15:09'!
picDataFor: cogMethod
	| cm nBytes nEntries oversizeData |
	cm := cogMethod methodObject.
	nBytes := (self byteSizeOf: cm) - (self startPCOfMethod: cm).
	oversizeData := Array new: nBytes.
	nEntries := cogit picDataFor: cogMethod into: (self oopForObject: oversizeData).
	^oversizeData copyFrom: 1 to: nEntries! !

!CurrentImageCoInterpreterFacade methodsFor: 'private' stamp: 'eem 8/4/2021 18:47'!
placeholder
	self error: 'should not be sent'! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/8/2011 13:01'!
primErrTable
	^objectMap at: Smalltalk primitiveErrorTable! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 14:26'!
primFailCodeAddress
	^self addressForLabel: #primFailCode! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 11/21/2016 14:31'!
primNumberExternalCall
	^coInterpreter primNumberExternalCall! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/22/2011 10:05'!
primTraceLogAddress
	^self addressForLabel: #primTraceLog! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/22/2011 10:01'!
primTraceLogIndexAddress
	^coInterpreter primTraceLogIndexAddress! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 9/5/2019 14:38'!
primitiveDoMixedArithmetic
	^coInterpreter primitiveDoMixedArithmetic ifNil: [true]! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 11/11/2011 11:17'!
primitiveFailAddress
	^self addressForLabel: #primitiveFail! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/8/2011 13:10'!
primitiveFunctionPointerAddress
	^self addressForLabel: #primitiveFunctionPointer! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/8/2011 14:43'!
primitiveIndexOf: anOop
	^(self objectForOop: anOop) primitive! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/2/2015 11:45'!
primitiveIndexOfMethod: anOop header: header
	| method |
	method := self objectForOop: anOop.
	self assert: (self objectForOop: header) = method header.
	^method primitive! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/8/2011 12:53'!
primitivePropertyFlags: primIndex
	^coInterpreter primitivePropertyFlags: primIndex! !

!CurrentImageCoInterpreterFacade methodsFor: 'printing' stamp: 'eem 4/6/2015 14:54'!
print: it
	it isString ifTrue: [coInterpreter transcript nextPutAll: it] ifFalse: [it printOn: coInterpreter transcript]! !

!CurrentImageCoInterpreterFacade methodsFor: 'printing' stamp: 'eem 3/2/2016 17:25'!
printChar: aCharacter
	coInterpreter transcript nextPut: aCharacter! !

!CurrentImageCoInterpreterFacade methodsFor: 'printing' stamp: 'eem 3/1/2016 13:33'!
printCogMethod: cogMethod
	| address primitive |
	address := cogMethod asInteger.
	self printHex: address;
		print: ' <-> ';
		printHex: address + cogMethod blockSize.
	cogMethod cmType = CMMethod ifTrue:
		[self print: ': method: ';
			printHex: cogMethod methodObject.
		 primitive := self primitiveIndexOfMethod: cogMethod methodObject
							header: cogMethod methodHeader.
		 primitive ~= 0 ifTrue:
			[self print: ' prim '; printNum: primitive]].
	cogMethod cmType = CMBlock ifTrue:
		[self print: ': block home: ';
			printHex: (self cCoerceSimple: cogMethod to: #'CogBlockMethod *') cmHomeMethod asUnsignedInteger].
	cogMethod cmType = CMClosedPIC ifTrue:
		[self print: ': Closed PIC N: ';
			printHex: cogMethod cPICNumCases].
	cogMethod cmType = CMOpenPIC ifTrue:
		[self print: ': Open PIC '].
	self print: ' selector: '; printHex: cogMethod selector.
	cogMethod selector = objectMemory nilObject
		ifTrue: [self print: ' (nil)']
		ifFalse: [self space; printStringOf: cogMethod selector].
	self cr! !

!CurrentImageCoInterpreterFacade methodsFor: 'printing' stamp: 'eem 3/2/2016 17:04'!
printHex: anInteger
	| it16 |
	it16 := anInteger radix: 16.
	coInterpreter transcript
		next: 8 - it16 size put: Character space;
		nextPutAll: (anInteger printStringBase: 16)! !

!CurrentImageCoInterpreterFacade methodsFor: 'printing' stamp: 'eem 3/2/2016 17:03'!
printHexnp: anInteger
	coInterpreter transcript nextPutAll: (anInteger printStringBase: 16)! !

!CurrentImageCoInterpreterFacade methodsFor: 'printing' stamp: 'eem 1/2/2017 17:39'!
printHexnpnp: anInteger
	coInterpreter transcript nextPutAll: (anInteger printStringBase: 16)! !

!CurrentImageCoInterpreterFacade methodsFor: 'printing' stamp: 'eem 3/2/2016 17:25'!
printNum: anInteger
	coInterpreter transcript print: anInteger! !

!CurrentImageCoInterpreterFacade methodsFor: 'printing' stamp: 'eem 3/1/2016 13:36'!
printStringOf: anOop
	Transcript nextPutAll: (self objectForOop: anOop)! !

!CurrentImageCoInterpreterFacade methodsFor: 'printing' stamp: 'eem 3/2/2016 17:01'!
putchar: aCharacter
	coInterpreter transcript nextPut: aCharacter! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/28/2011 17:56'!
quickPrimitiveConstantFor: aQuickPrimitiveIndex
	^self oopForObject: (coInterpreter quickPrimitiveConstantFor: aQuickPrimitiveIndex)! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/28/2011 17:50'!
quickPrimitiveGeneratorFor: aQuickPrimitiveIndex
	^coInterpreter quickPrimitiveGeneratorFor: aQuickPrimitiveIndex! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/28/2011 17:57'!
quickPrimitiveInstVarIndexFor: aQuickPrimitiveIndex
	^coInterpreter quickPrimitiveInstVarIndexFor: aQuickPrimitiveIndex! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/4/2021 22:44'!
rawHeaderOf: aMethodOop
	| method header correctedHeader headerOop |
	method := self objectForOop: aMethodOop.
	header := method header.

	"(index 0)		15 bits:	number of literals (#numLiterals)
	 (index 15)		  1 bit:	jit without counters - reserved for methods that have been optimized by Sista
	 (index 16)		  1 bit:	has primitive
	 (index 17)		  1 bit:	whether a large frame size is needed (#frameSize => either SmallFrame or LargeFrame)
	 (index 18)		  6 bits:	number of temporary variables (#numTemps)
	 (index 24)		  4 bits:	number of arguments to the method (#numArgs)
	 (index 28)		  2 bits:	reserved for an access modifier (00-unused, 01-private, 10-protected, 11-public), although accessors for bit 29 exist (see #flag).
	 sign bit:		  1 bit: selects the instruction set, >= 0 Primary, < 0 Secondary (#signFlag)"

	correctedHeader := self wordSize = Smalltalk wordSize
							ifTrue: [header]
							ifFalse:
								[(header bitAnd: (1 bitShift: 30) - 1)
								- (header < 0 ifTrue: [2 raisedTo: self wordSize * 8] ifFalse: [0])].
	headerOop := objectMemory integerObjectOf: correctedHeader.
	headerOop := headerOop bitAnd: (2 raisedTo: self wordSize * 8) - 1.
	self assert: (objectMemory literalCountOfMethodHeader: headerOop) = method numLiterals.
	self assert: (coInterpreter argumentCountOfMethodHeader: headerOop) = method numArgs.
	self assert: (coInterpreter temporaryCountOfMethodHeader: headerOop) = method numTemps.
	self assert: (objectMemory headerIndicatesAlternateBytecodeSet: headerOop) = method signFlag.
	headerToMethodMap
		at: headerOop
		ifPresent: [:existing| self assert: existing == method]
		ifAbsentPut: [method].
	^headerOop! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 15:20'!
rawHeaderOf: anOop put: aCogMethodSurrogate
	^self! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/4/2021 12:59'!
recordFastCCallPrimTraceForMethod: methodOop
	^cogit recordFastCCallPrimTrace! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/4/2021 13:01'!
recordPrimTraceForMethod: methodOop
	^cogit recordPrimTrace! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 7/12/2016 10:47'!
remoteIsInstVarAccess
	^coInterpreter remoteIsInstVarAccess! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/9/2019 14:48'!
rumpCStackAddress
	"Use the top half of memory for variables, method cache and rumpCStack,
	 and the bottom half for allocating code and objects:

		0Mb to 1Mb		code zone
		1Mb to 2Mb		object zone
		2Mb to 3Mb		variables
		3Mb to 4Mb - 1k	method cache
		4Mb - 1k to 4Mb	rump C stack"
	^memory byteSize - 1024! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/8/2011 13:35'!
scavengeThresholdAddress
	^self addressForLabel: #scavengeThreshold! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 2/1/2020 15:55'!
setCFramePointer: cfp
	^coInterpreter setCFramePointer: cfp! !

!CurrentImageCoInterpreterFacade methodsFor: 'cog jit support' stamp: 'eem 12/2/2019 16:46'!
setCFramePointer: fp setCStackPointer: sp
	cFramePointer := fp.
	cStackPointer := sp! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 2/1/2020 15:55'!
setCStackPointer: csp
	^coInterpreter setCStackPointer: csp! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'RonieSalgado 4/30/2016 20:35'!
shadowCallStackAddress
	^self addressForLabel: #shadowCallStackAddress! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 2/4/2020 09:44'!
shadowCallStackPointerAddress
	^self addressForLabel: #shadowCallStackPointer! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/2/2015 13:45'!
shiftForWord
	^objectMemory shiftForWord! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 11/21/2016 16:50'!
shortPrint: oop
	^(objectMemory isImmediate: oop)
		ifTrue: [coInterpreter shortPrint: oop]
		ifFalse: [(objectMap keyAtValue: oop) printString]! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/20/2017 14:37'!
sixtyFourBitIndexableFormat
	^objectMemory sixtyFourBitIndexableFormat! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 11/14/2012 15:01'!
sizeOfCallPrimitiveBytecode: methodHeaderOop
	^coInterpreter sizeOfCallPrimitiveBytecode: methodHeaderOop! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 11/14/2012 15:00'!
sizeOfLongStoreTempBytecode: methodHeaderOop 
	^coInterpreter sizeOfLongStoreTempBytecode: methodHeaderOop! !

!CurrentImageCoInterpreterFacade methodsFor: 'printing' stamp: 'eem 3/2/2016 17:00'!
space
	coInterpreter transcript space! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 3/11/2020 11:31'!
specialObjectsOopAddress
	self subclassResponsibility! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 3/21/2017 09:49'!
specialSelector: index 
	^self oopForObject: (self specialSelectors at: index * 2 + 1)! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 3/21/2017 09:48'!
specialSelectorNumArgs: index 
	^self specialSelectors at: index * 2 + 2! !

!CurrentImageCoInterpreterFacade methodsFor: 'private' stamp: 'eem 3/21/2017 09:49'!
specialSelectors
	^Smalltalk specialObjectsArray at: 24! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'RonieSalgado 5/31/2016 19:27'!
splObj: splObjIndex
	^splObjIndex caseOf: {
		[ClassArray]					-> [self oopForObject: Array].
		[ClassCharacter]				-> [self oopForObject: Character].
		[ClassLargeNegativeInteger]	-> [self oopForObject: LargeNegativeInteger].
		[ClassLargePositiveInteger]	-> [self oopForObject: LargePositiveInteger].
		[ClassSmallInteger]				-> [self oopForObject: SmallInteger].
		[CompactClasses]				-> [self oopForObject: Smalltalk compactClassesArray].
		[LowcodeContextMark]				-> [self oopForObject: (Smalltalk globals at: #LowcodeNativeContext ifPresent: [:cls | cls signalingObject ]) ]
		}! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 2/4/2020 09:35'!
stackLimitAddress
	^self addressForLabel: #stackLimit! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 14:26'!
stackPointerAddress
	^self addressForLabel: #stackPointer! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/9/2019 14:45'!
startOfMemory
	"Use the top half of memory for variables, method cache and rumpCStack,
	 and the bottom half for allocating code and objects:

		0Mb to 1Mb		code zone
		1Mb to 2Mb		object zone
		2Mb to 3Mb		variables
		3Mb to 4Mb - 1k	method cache
		4Mb - 1k to 4Mb	rump C stack"
	self assert: memory byteSize = (4 * 1024 * 1024).
	^memory byteSize / 4! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 15:15'!
startPCOfMethod: anOop
	^(objectMap keyAtValue: anOop) initialPC - 1! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 5/14/2015 18:08'!
startPCOfMethodHeader: headerIntegerOop
	^(headerToMethodMap at: headerIntegerOop) initialPC - 1
	"a.k.a.
	^coInterpreter startPCOfMethodHeader: headerIntegerOop,
	 but only if the CompiledMethod header formats match"! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/15/2013 11:30'!
startPCOrNilOfLiteral: literalOop in: methodOop
	^(literalOop isClosure
	   and: [literalOop method == (objectMap keyAtValue: methodOop)]) ifTrue:
		[literalOop startpc]! !

!CurrentImageCoInterpreterFacade methodsFor: 'testing' stamp: 'eem 2/9/2011 15:10'!
storePointer: index ofObject: objOop withValue: valueOop 
	(self objectForOop: objOop)
		at: index + 1
		put: (self objectForOop: valueOop)! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 2/16/2011 12:04'!
storePointerUnchecked: zeroRelativeIndex ofObject: targetOop withValue: valueOop
	(self objectForOop: targetOop) at: zeroRelativeIndex + 1 put: (self objectForOop: valueOop)! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 2/23/2017 10:22'!
stringOf: anOop
	| thing |
	thing := objectMap
				keyAtValue: anOop
				ifAbsent:
					[variables
						keyAtValue: anOop
						ifAbsent: [^nil]].
	^((thing isLiteral and: [thing isSymbol not])
		ifTrue: [thing storeString]
		ifFalse: [thing asString]) contractTo: 64! !

!CurrentImageCoInterpreterFacade methodsFor: 'testing' stamp: 'eem 11/6/2012 15:35'!
tempCountForBlockStartingAt: startpc in: methodOop
	^BlockLocalTempCounter
		tempCountForBlockStartingAt: startpc
		in: (self objectForOop: methodOop)! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/8/2011 14:43'!
tempCountOf: anOop
	^(self objectForOop: anOop) numTemps! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 5/14/2015 18:09'!
temporaryCountOfMethodHeader: headerIntegerOop
	^(headerToMethodMap at: headerIntegerOop) numTemps
	"a.k.a.
	^coInterpreter temporaryCountOfMethodHeader: headerIntegerOop,
	 but only if the CompiledMethod header formats match"! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 12:44'!
threadManager
	^nil! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/19/2010 14:56'!
transcript
	^Transcript! !

!CurrentImageCoInterpreterFacade methodsFor: 'object map' stamp: 'eem 12/19/2010 13:09'!
trueObject
	^objectMap at: true! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 9/6/2016 18:44'!
ultimateLiteralOf: methodObj
	| meth |
	meth := (objectMap keyAtValue: methodObj).
	^self oopForObject: (meth literalAt: meth numLiterals)! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/15/2015 12:03'!
unalignedLongAt: anInteger
	^objectMemory unalignedLongAt: anInteger! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 12/15/2015 12:04'!
unalignedLongAt: anInteger put: value
	^objectMemory unalignedLongAt: anInteger put: value! !

!CurrentImageCoInterpreterFacade methodsFor: 'private' stamp: 'eem 12/9/2019 14:48'!
variablesBase
	"Use the top half of memory for variables, method cache and rumpCStack,
	 and the bottom half for allocating code and objects:

		0Mb to 1Mb		code zone
		1Mb to 2Mb		object zone
		2Mb to 3Mb		variables
		3Mb to 4Mb - 1k	method cache
		4Mb - 1k to 4Mb	rump C stack"
	^memory byteSize / 2! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 1/15/2015 11:53'!
weakArrayFormat
	^objectMemory weakArrayFormat! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 8/25/2014 15:54'!
wordSize
	^objectMemory wordSize! !

!CurrentImageCoInterpreterFacade methodsFor: 'accessing' stamp: 'eem 9/9/2013 11:57'!
youngStartAddress
	^self subclassResponsibility! !

CurrentImageCoInterpreterFacade subclass: #CurrentImageCoInterpreterFacadeForSpurObjectRepresentation
	instanceVariableNames: 'hiddenRoots'
	classVariableNames: ''
	poolDictionaries: 'VMSpurObjectRepresentationConstants'
	category: 'VMMaker-Support'!

CurrentImageCoInterpreterFacadeForSpurObjectRepresentation subclass: #CurrentImageCoInterpreterFacadeFor64BitSpurObjectRepresentation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Support'!

!CurrentImageCoInterpreterFacadeFor64BitSpurObjectRepresentation class methodsFor: 'accessing class hierarchy' stamp: 'eem 9/22/2015 19:57'!
objectMemoryClass
	^Spur64BitCoMemoryManager! !

!CurrentImageCoInterpreterFacadeFor64BitSpurObjectRepresentation class methodsFor: 'accessing class hierarchy' stamp: 'eem 9/22/2015 19:58'!
objectRepresentationClass
	^CogObjectRepresentationFor64BitSpur! !

!CurrentImageCoInterpreterFacadeFor64BitSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 8/4/2021 12:56'!
functionPointerForCompiledMethod: methodOop primitiveIndex: primIndex primitivePropertyFlagsInto: flagsPtr
	| candidates literal method |
	primIndex = PrimNumberExternalCall ifTrue:
		[method := self objectForOop: methodOop.
		 self assert: method primitive = PrimNumberExternalCall.
		 literal := method literalAt: 1.
		 candidates := self sn
							allImplementorsOf: literal second
							localTo: (literal first basicSize = 0
										ifTrue: [InterpreterPrimitives]
										ifFalse: [InterpreterPlugin allSubclasses
													detect: [:pluginClass| pluginClass moduleName = literal first]]).
		 self assert: candidates size = 1.
		 flagsPtr
			at: 0
			put: (coInterpreter
					primitivePropertyFlagsFor: candidates first compiledMethod
					primitiveIndex: primIndex).
		 ^self oopForObject: literal second].
	^super functionPointerForCompiledMethod: methodOop primitiveIndex: primIndex primitivePropertyFlagsInto: flagsPtr! !

!CurrentImageCoInterpreterFacadeFor64BitSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 8/5/2021 07:51'!
nextProfileTickAddress
	^self addressForLabel: #nextProfileTick! !

!CurrentImageCoInterpreterFacadeFor64BitSpurObjectRepresentation methodsFor: 'private-cacheing' stamp: 'eem 11/30/2017 11:08'!
objectForOop: anOop
	"This is a keyAtValue: search and so needs speeding up either by a reverse map or a simple cache."
	^(anOop bitAnd: 7) caseOf: {
		[0] -> [anOop = cachedOop
				ifTrue: [cachedObject]
				ifFalse: [cachedObject := objectMap keyAtNonIdentityValue: anOop. "may raise Error"
						cachedOop := anOop. "Don't assign until accessed without error"
						cachedObject]].
		[1] -> [anOop signedIntFromLong64 >> 3].
		[2] -> [Character value: anOop >> 3].
		[4] -> [objectMemory smallFloatValueOf: anOop] }! !

!CurrentImageCoInterpreterFacadeFor64BitSpurObjectRepresentation methodsFor: 'object map' stamp: 'eem 11/30/2017 10:58'!
oopForObject: o
	o isNumber
		ifTrue:
			[o isInteger
				ifTrue:
					[(o between: objectMemory minSmallInteger and: objectMemory maxSmallInteger) ifTrue:
						[^objectMemory integerObjectOf: o asInteger]]
				ifFalse:
					[o isFloat ifTrue:
						[(objectMemory isSmallFloatValue: o) ifTrue:
							[^objectMemory smallFloatObjectOf: o]]]]
		ifFalse:
			[o isCharacter ifTrue:
				[^objectMemory characterObjectOf: o asInteger]].
	^objectMap at: o ifAbsentPut: [objectMap size * 8 + (1024 * 1024 * 1024 * 4)]! !

!CurrentImageCoInterpreterFacadeFor64BitSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 12/7/2015 22:36'!
smallFloatExponentOffset
	^objectMemory smallFloatExponentOffset! !

!CurrentImageCoInterpreterFacadeFor64BitSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 12/7/2015 22:36'!
smallFloatMantissaBits
	^objectMemory smallFloatMantissaBits! !

!CurrentImageCoInterpreterFacadeFor64BitSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 11/4/2015 09:18:58'!
smallFloatTag
	^objectMemory smallFloatTag! !

!CurrentImageCoInterpreterFacadeFor64BitSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 3/11/2020 11:31'!
specialObjectsOopAddress
	^self addressForLabel: #specialObjectsOop! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation class methodsFor: 'accessing class hierarchy' stamp: 'eem 6/1/2014 12:15'!
objectMemoryClass
	^Spur32BitCoMemoryManager! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation class methodsFor: 'accessing class hierarchy' stamp: 'eem 1/22/2014 15:37'!
objectRepresentationClass
	^CogObjectRepresentationFor32BitSpur! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'cog jit support' stamp: 'eem 8/4/2021 14:08'!
accessorDepthForExternalPrimitiveMethod: methodOop
	| metadata |
	metadata := StackInterpreter metadataFlagsForPrimitive: (self primitiveMethodForMethodContainingExternalPrimitive: methodOop).
	^metadata bitShift: SpurPrimitiveAccessorDepthShift negated! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 1/15/2015 12:02'!
accessorDepthForPrimitiveIndex: anInteger
	^StackInterpreterPrimitives primitiveAccessorDepthTable at: anInteger + 1! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'cog jit support' stamp: 'eem 8/25/2014 16:17'!
allocatePinnedSlots: nSlots
	^(objectMemory allocatePinnedSlots: nSlots)
		ifNil: [nil]
		ifNotNil:
			[:objOop |
			 objectMap at: objOop put: objOop]! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 6/1/2014 12:21'!
arrayFormat
	^objectMemory arrayFormat! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 4/2/2014 15:08'!
baseHeaderSize
	^SpurMemoryManager baseHeaderSize! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 6/3/2016 14:08'!
ceCheckAndMaybeRetryPrimitive: primIndex
	^coInterpreter ceCheckAndMaybeRetryPrimitive: primIndex! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 8/7/2014 05:04'!
characterTag
	^objectMemory characterTag! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 8/25/2014 15:54'!
classIndexMask
	^objectMemory classIndexMask! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 6/4/2014 14:48'!
classTableMajorIndexShift
	^objectMemory classTableMajorIndexShift! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 6/4/2014 14:49'!
classTableMinorIndexMask
	^objectMemory classTableMinorIndexMask! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 8/7/2014 05:05'!
classTablePageSize
	^objectMemory classTablePageSize! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 8/25/2014 15:54'!
classTableRootObj
	hiddenRoots ifNil:
		[hiddenRoots := Array new: objectMemory classTableRootSlots + objectMemory hiddenRootSlots].
	^self oopForObject: hiddenRoots! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 8/7/2014 05:01'!
classTagForClass: classOop
	^(self objectForOop: classOop) identityHash! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'RonieSalgado 4/30/2016 20:36'!
defaultNativeStackFrameSize
	^ coInterpreter defaultNativeStackFrameSize! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'cog jit support' stamp: 'eem 8/26/2014 09:14'!
eeInstantiateClassIndex: knownClassIndex format: objFormat numSlots: numSlots
	self assert: knownClassIndex = ClassArrayCompactIndex.
	^self oopForObject: (Array new: numSlots)! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 3/22/2017 17:23'!
fetchClassTagOf: anOop
	| class |
	self assert: SmalltalkImage current isRunningSpur.
	class := (self objectForOop: anOop) class.
	^SystemVersion current isPharo
		ifTrue: [class basicIdentityHash]
		ifFalse: [class identityHash]! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 1/15/2015 11:56'!
firstShortFormat
	^objectMemory firstShortFormat! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 1/15/2015 11:54'!
fixedFieldsOfClassFormatMask
	^objectMemory fixedFieldsOfClassFormatMask! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'forwarding' stamp: 'eem 7/10/2014 16:12'!
followForwardedObjectFields: objOop toDepth: depth 
	"This is a noop in the facade"! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 1/15/2015 11:51'!
formatMask
	^objectMemory formatMask! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 7/12/2016 10:45'!
formatOfClass: classOop 
	^(self objectForOop: classOop) format! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 1/15/2015 11:51'!
formatShift
	^objectMemory formatShift! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'cog jit support' stamp: 'eem 8/4/2021 14:30'!
functionPointerForCompiledMethod: methodOop primitiveIndex: primIndex primitivePropertyFlagsInto: flagsPtr
	primIndex = PrimNumberExternalCall ifTrue:
		[flagsPtr
			at: 0
			put: (coInterpreter
					primitivePropertyFlagsFor: (self primitiveMethodForMethodContainingExternalPrimitive: methodOop)
					primitiveIndex: primIndex).
		 ^self oopForObject: ((self objectForOop: methodOop) literalAt: 1) second].
	^super functionPointerForCompiledMethod: methodOop primitiveIndex: primIndex primitivePropertyFlagsInto: flagsPtr! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 6/29/2015 14:51'!
getScavengeThreshold
	^variables
		at: 'scavengeThreshold'
		ifAbsentPut: [objectMemory getScavengeThreshold ifNil: [16r24680]]! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 6/1/2014 12:12'!
headerForSlots: numSlots format: formatField classIndex: classIndex
	^objectMemory headerForSlots: numSlots format: formatField classIndex: classIndex! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 11/4/2015 09:18:12'!
identityHashFullWordShift
	^objectMemory identityHashFullWordShift! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'cb 11/26/2015 18:19'!
immutableBitMask
	^objectMemory immutableBitMask! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'cog jit support' stamp: 'eem 7/12/2016 10:42'!
instSpecOfClassFormat: classFormat
	^objectMemory instSpecOfClassFormat: classFormat! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'cog jit support' stamp: 'eem 7/12/2016 10:39'!
isFixedSizePointerFormat: formatOop
	^objectMemory isFixedSizePointerFormat: formatOop! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 6/2/2014 20:34'!
isForwardedObjectClassIndexPun
	^objectMemory isForwardedObjectClassIndexPun! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 8/7/2014 05:03'!
isImmediateClass: classOop
	"Can't rely on the host image; may be running on SqueakV3.  hence..."
	^#(Character SmallInteger SmallFloat) includes: (self objectForOop: classOop) name! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 1/25/2018 11:32'!
maybeMethodClassOf: methodObj seemsToBeInstantiating: format
	"Answers if the code is installed in a class instantiating objects with the format. Used in primitive 
	 generation to make a quick path based on where the method is installed. This method cannot
	 be used as a guarantee as there can be false positive, it's just a heuristic.
	 Tries to interpret the last literal of the method as a behavior (more than 3 fields, 3rd field a Smi).
	 If it can be interpreted as a behavior, answers if instSpec matches the format, else answers false."
	<api>
	| maybeClassObj|
	maybeClassObj := (self objectForOop: methodObj) methodClass.
	^maybeClassObj isBehavior
	 and: [maybeClassObj instSpec = format]! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 4/11/2017 17:11'!
needGCFlagAddress
	^self addressForLabel: #needGCFlag! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 12/2/2015 14:35'!
numSlotsFullShift
	^objectMemory numSlotsFullShift! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 6/1/2014 12:20'!
numSlotsMask
	^objectMemory numSlotsMask! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 11/4/2015 09:18:35'!
numTagBits
	^objectMemory numTagBits! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'private-cacheing' stamp: 'eem 3/22/2017 17:41'!
objectForOop: anOop
	"This is a keyAtValue: search and so needs speeding up either by a reverse map or a simple cache."
	^(anOop bitAnd: 3) caseOf: {
		[0] -> [anOop = cachedOop
				ifTrue: [cachedObject]
				ifFalse: [cachedObject := objectMap keyAtValue: anOop. "may raise Error"
						cachedOop := anOop. "Don't assign until accessed without error"
						cachedObject]].
		[1] -> [anOop signedIntFromLong >> 1].
		[2] -> [Character value: anOop >> 2].
		[3] -> [anOop signedIntFromLong >> 1] }! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'object map' stamp: 'eem 12/2/2015 14:54'!
oopForObject: o
	^(o isInteger and: [o between: objectMemory minSmallInteger and: objectMemory maxSmallInteger])
		ifTrue: [o << 1 + 1]
		ifFalse:
			[o isCharacter
				ifTrue: [o asInteger << 2 + 2]
				ifFalse: [objectMap at: o ifAbsentPut: [objectMap size * 8 + (1024 * 1024)]]]! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'cog jit support' stamp: 'eem 8/4/2021 14:03'!
primitiveMethodForMethodContainingExternalPrimitive: methodOop
	"Look up a primitive in a method and answer that method."
	| candidates literal method |
	method := self objectForOop: methodOop.
	self assert: method primitive = PrimNumberExternalCall.
	literal := method literalAt: 1.
	candidates := self sn
						allImplementorsOf: literal second
						localTo: (literal first basicSize = 0
									ifTrue: [InterpreterPrimitives]
									ifFalse: [InterpreterPlugin allSubclasses
												detect: [:pluginClass| pluginClass moduleName = literal first]]).
	self assert: candidates size = 1.
	^candidates first compiledMethod! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 7/12/2016 10:36'!
rawHashBitsOf: obj
	^(self objectForOop: obj) identityHash! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 8/31/2021 14:28'!
receiverTagBitsForMethod: methodOop
	| mclass |
	^(mclass := (self objectForOop: methodOop) methodClass)
		caseOf: {
			[SmallInteger]			-> [objectMemory smallIntegerTag].
			[Character]				-> [objectMemory characterTag].
			[SmallFloat64]			-> [objectMemory smallFloatTag].
			[Context]				-> [1 bitShift: objectMemory numTagBits] }
		otherwise:
			[(Context includesBehavior: mclass)
				ifTrue: [1 bitShift: objectMemory numTagBits]
				ifFalse: [0]]! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 6/1/2014 12:22'!
rememberedBitShift
	^objectMemory rememberedBitShift! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'cb 7/7/2016 17:01'!
remoteIsInstVarAccess
	^ 128! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 4/11/2017 17:11'!
scavengeThreshold
	^objectMemory scavengeThreshold! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 6/4/2014 14:48'!
shiftForWord
	^objectMemory shiftForWord! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 8/7/2014 05:03'!
smallIntegerTag
	^objectMemory smallIntegerTag! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 6/1/2014 12:18'!
smallObjectBytesForSlots: numSlots
	^objectMemory smallObjectBytesForSlots: numSlots! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'accessing' stamp: 'eem 6/1/2014 12:22'!
storeCheckBoundary
	^objectMemory storeCheckBoundary ifNil: [16r12345678]! !

!CurrentImageCoInterpreterFacadeForSpurObjectRepresentation methodsFor: 'testing' stamp: 'eem 8/25/2014 15:54'!
tagMask
	^objectMemory tagMask! !

CurrentImageCoInterpreterFacade subclass: #CurrentImageCoInterpreterFacadeForSqueakV3ObjectRepresentation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Support'!

!CurrentImageCoInterpreterFacadeForSqueakV3ObjectRepresentation class methodsFor: 'accessing class hierarchy' stamp: 'eem 6/2/2014 08:43'!
objectMemoryClass
	^NewCoObjectMemory! !

!CurrentImageCoInterpreterFacadeForSqueakV3ObjectRepresentation class methodsFor: 'accessing class hierarchy' stamp: 'eem 10/9/2013 17:52'!
objectRepresentationClass
	^CogObjectRepresentationForSqueakV3! !

!CurrentImageCoInterpreterFacadeForSqueakV3ObjectRepresentation methodsFor: 'accessing' stamp: 'eem 4/2/2014 15:09'!
baseHeaderSize
	^ObjectMemory baseHeaderSize! !

!CurrentImageCoInterpreterFacadeForSqueakV3ObjectRepresentation methodsFor: 'accessing' stamp: 'eem 9/9/2013 11:53'!
characterTable
	^self oopForObject: Character characterTable! !

!CurrentImageCoInterpreterFacadeForSqueakV3ObjectRepresentation methodsFor: 'accessing' stamp: 'eem 9/9/2013 11:54'!
formatOfClass: classOop 
	^(self objectForOop: classOop) format << 1! !

!CurrentImageCoInterpreterFacadeForSqueakV3ObjectRepresentation methodsFor: 'private-cacheing' stamp: 'eem 3/22/2017 17:41'!
objectForOop: anOop
	"This is a keyAtValue: search and so needs speeding up either by a reverse map or a simple cache."
	^(anOop bitAnd: 3) caseOf: {
		[0] -> [anOop = cachedOop
				ifTrue: [cachedObject]
				ifFalse: [cachedObject := objectMap keyAtValue: anOop. "may raise Error"
						cachedOop := anOop. "Don't assign until accessed without error"
						cachedObject]].
		[1] -> [anOop signedIntFromLong >> 1].
		[3] -> [anOop signedIntFromLong >> 1] }! !

!CurrentImageCoInterpreterFacadeForSqueakV3ObjectRepresentation methodsFor: 'object map' stamp: 'eem 3/22/2017 17:44'!
oopForObject: o
	^(o isInteger and: [o between: objectMemory minSmallInteger and: objectMemory maxSmallInteger])
		ifTrue: [o << 1 bitOr: 1]
		ifFalse: [objectMap at: o ifAbsentPut: [objectMap size * 8 + (1024 * 1024)]]! !

!CurrentImageCoInterpreterFacadeForSqueakV3ObjectRepresentation methodsFor: 'accessing' stamp: 'eem 8/31/2021 14:28'!
receiverTagBitsForMethod: methodOop
	^(self objectForOop: methodOop) methodClass = SmallInteger
		ifTrue: [1]
		ifFalse: [0]! !

!CurrentImageCoInterpreterFacadeForSqueakV3ObjectRepresentation methodsFor: 'accessing' stamp: 'eem 9/9/2013 11:57'!
youngStartAddress
	^16r4E5E400! !

CogClass subclass: #SpurCompactor
	instanceVariableNames: 'coInterpreter manager scavenger'
	classVariableNames: ''
	poolDictionaries: 'SpurMemoryManagementConstants VMBasicConstants VMSpurObjectRepresentationConstants'
	category: 'VMMaker-SpurMemoryManager'!

!SpurCompactor commentStamp: 'cb 4/27/2018 09:38' prior: 0!
Abstract common superclass of all compactors to define apis and simulation variables.

The full GC in Spur is split in two, the marking phase and the compactor phase. The subclasses of SpurCompactor are implementations of the second phase, so they are called once the marking phase is finished. SpurCompactor is reponsible for:
- freeing unmarked objects
- compacting the live old space objects (though each subclass define what it does, some spurCompactor may not compact)
- unmarking all objects remaining live
- updating oops directly referred by the VM when they are moved (remapObj:/shouldRemapObj: thingy)

The main apis are the following:
- biasForGC/biasForSnapshot: tells the compactor if the GC is performed for snapshots or not, in general we want to compact more aggressively for snapshots to avoid saving large files with many unused space.
- compact: main API, should free the unmarked object, unmark the objects remaining live and potentially compact the heap
- remapObj:/shouldRemapObj: => Not really sure what this does, it seems it has to do with updating oops directly referred by the VM when they are moved. 
- postSwizzleAction: if you want to do something at start-up after swizzle phase (typically useful if your compaction algo uses segInfos)

Instance Variables
	coInterpreter:				<StackInterpreter>
	scavenger:					<SpurGenerationScavenger>
	manager:					<SpurMemoryManager>!

!SpurCompactor class methodsFor: 'translation' stamp: 'eem 6/7/2018 15:40'!
classesForTranslation
	"Usually we want to add the SpurCompactor first, followed by the subclass.
	 SpurHybridCompactor is more complicated."
	^(self withAllSuperclasses copyUpThrough: SpurCompactor) reverse! !

!SpurCompactor class methodsFor: 'translation' stamp: 'eem 6/7/2018 15:59'!
declareCVarsIn: aCCodeGenerator
	"If subclasses are being used but are not the compactorClass itself
	 then statically resolve their api selectors."
	(self inheritsFrom: SpurCompactor) ifTrue: "leave SpurCompactor methods alone"
		[(SpurMemoryManager compactorClass includesBehavior: self) ifFalse: "only map methods not in the compactorClass"
			[#(compact biasForGC biasForSnapshot remapObj: shouldRemapObj:) do: "map the public api methods only"
				[:selectorToStaticallyResolve|
				 aCCodeGenerator
					staticallyResolveMethodNamed: selectorToStaticallyResolve
					forClass: self
					to: (self staticallyResolvePolymorphicSelector: selectorToStaticallyResolve)]]]! !

!SpurCompactor class methodsFor: 'translation' stamp: 'cb 10/1/2017 12:04'!
implicitReturnTypeFor: aSelector
	"Answer the return type for methods that don't have an explicit return."
	^#void! !

!SpurCompactor methodsFor: 'api' stamp: 'cb 10/2/2017 15:11'!
biasForGC
	self subclassResponsibility! !

!SpurCompactor methodsFor: 'api' stamp: 'cb 10/2/2017 15:11'!
biasForSnapshot
	self subclassResponsibility! !

!SpurCompactor methodsFor: 'initialization' stamp: 'cb 9/29/2017 16:38'!
coInterpreter: aVMSimulator
	<doNotGenerate>
	coInterpreter := aVMSimulator! !

!SpurCompactor methodsFor: 'api' stamp: 'cb 10/2/2017 15:11'!
compact
	self subclassResponsibility! !

!SpurCompactor methodsFor: 'initialization' stamp: 'cb 9/29/2017 16:38'!
manager: aSpurNBitMMXEndianSimulator
	<doNotGenerate>
	manager := aSpurNBitMMXEndianSimulator.
	aSpurNBitMMXEndianSimulator coInterpreter ifNotNil:
		[:coint| coInterpreter := coint].
	aSpurNBitMMXEndianSimulator scavenger ifNotNil:
		[:scav| scavenger := scav]! !

!SpurCompactor methodsFor: 'api' stamp: 'cb 4/26/2018 20:34'!
postSwizzleAction
	"do nothing"! !

!SpurCompactor methodsFor: 'debugging' stamp: 'cb 10/3/2017 11:00'!
printTheBogons: aBogon
	<inline: true>
	coInterpreter
		print: 'bogon '; printHexnp: aBogon; cr! !

!SpurCompactor methodsFor: 'api' stamp: 'cb 4/26/2018 20:44'!
remapObj: objOop
	<api>
	<inline: false>
	^manager vanillaRemapObj: objOop! !

!SpurCompactor methodsFor: 'initialization' stamp: 'eem 6/6/2018 18:40'!
scavenger: aSpurGenerationScavenger
	<doNotGenerate>
	scavenger := aSpurGenerationScavenger! !

!SpurCompactor methodsFor: 'api' stamp: 'cb 4/26/2018 20:44'!
shouldRemapObj: objOop
	<api>
	"Answer if the obj should be scavenged (or simply followed). The method is called
	 shouldRemapObj: for compatibility with ObjectMemory.  Defer to the compactor
	 to choose the actual test, there being a difference between the vanilla test and
	 that used with a sliding compactor where objects are not marked as forwarded."
	^manager vanillaShouldRemapObj: objOop! !

SpurCompactor subclass: #SpurHybridCompactor
	instanceVariableNames: 'planningCompactor selectiveCompactor planNotSelect'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManager'!

!SpurHybridCompactor commentStamp: 'eem 6/6/2018 19:25' prior: 0!
A SpurHybridCompactor is a compactor that uses SpurSelectiveCompactor for normal GC compactions, but uses SpurPlanningCompactor for snapshot.

Instance Variables
	planNotSelect:			<Boolean>
	planningCompactor:	<SpurPlanningCompactor>
	selectiveCompactor:	<SpurSelectiveCompactor>

planNotSelect
	- the boolean that selects between the two compactors; if true it chooses SpurPlanningCompactor

planningCompactor
	- the SpurPlanningCompactor

selectiveCompactor
	- the SpurSelectiveCompactor
!

!SpurHybridCompactor class methodsFor: 'translation' stamp: 'eem 6/7/2018 15:38'!
classesForTranslation
	^{	SpurPlanningCompactor.
		SpurSelectiveCompactor.
		SpurSweeper }, super classesForTranslation! !

!SpurHybridCompactor class methodsFor: 'translation' stamp: 'eem 6/6/2018 20:02'!
declareCVarsIn: aCCodeGenerator
	aCCodeGenerator
		staticallyResolvedPolymorphicReceiver: 'planningCompactor' to: SpurPlanningCompactor;
		staticallyResolvedPolymorphicReceiver: 'selectiveCompactor' to: SpurSweeper"SpurSelectiveCompactor" "The selector mangling is easier this way"! !

!SpurHybridCompactor methodsFor: 'api' stamp: 'eem 11/15/2019 09:03'!
biasForGC
	<inline: true>
	planNotSelect := false.
	selectiveCompactor biasForGC! !

!SpurHybridCompactor methodsFor: 'api' stamp: 'eem 6/6/2018 19:26'!
biasForSnapshot
	planNotSelect := true.
	planningCompactor biasForSnapshot! !

!SpurHybridCompactor methodsFor: 'initialization' stamp: 'eem 6/6/2018 18:45'!
coInterpreter: aVMSimulator
	<doNotGenerate>
	super coInterpreter: aVMSimulator.
	planningCompactor coInterpreter: aVMSimulator.
	selectiveCompactor coInterpreter: aVMSimulator! !

!SpurHybridCompactor methodsFor: 'api' stamp: 'eem 6/6/2018 18:49'!
compact
	planNotSelect
		ifTrue: [planningCompactor compact]
		ifFalse: [selectiveCompactor compact]! !

!SpurHybridCompactor methodsFor: 'initialization' stamp: 'eem 6/6/2018 19:04'!
manager: aSpurNBitMMXEndianSimulator
	<doNotGenerate>
	super manager: aSpurNBitMMXEndianSimulator.
	(planningCompactor ifNil: [planningCompactor := SpurPlanningCompactor simulatorClass new])
		manager: aSpurNBitMMXEndianSimulator.
	(selectiveCompactor ifNil: [selectiveCompactor := SpurSelectiveCompactor simulatorClass new])
		manager: aSpurNBitMMXEndianSimulator.

	planNotSelect := false! !

!SpurHybridCompactor methodsFor: 'api' stamp: 'eem 6/6/2018 20:10'!
remapObj: objOop
	<api>
	<inline: false>
	planNotSelect ifTrue:
		[^manager slidingCompactionRemapObj: objOop].
	^super remapObj: objOop! !

!SpurHybridCompactor methodsFor: 'initialization' stamp: 'eem 6/6/2018 18:45'!
scavenger: aSpurGenerationScavenger
	<doNotGenerate>
	super scavenger: aSpurGenerationScavenger. "I always thought the scavenger was super..."
	planningCompactor scavenger: aSpurGenerationScavenger.
	selectiveCompactor scavenger: aSpurGenerationScavenger! !

!SpurHybridCompactor methodsFor: 'api' stamp: 'eem 6/6/2018 20:09'!
shouldRemapObj: objOop
	<api>
	<inline: false>
	planNotSelect ifTrue:
		[^manager slidingCompactionShouldRemapObj: objOop].
	^super shouldRemapObj: objOop! !

SpurCompactor subclass: #SpurPlanningCompactor
	instanceVariableNames: 'anomaly biasForGC firstFieldOfRememberedSet firstFreeObject firstMobileObject lastMobileObject mobileStart objectAfterLastMobileObject savedFirstFieldsSpace savedFirstFieldsSpaceNotInOldSpace'
	classVariableNames: ''
	poolDictionaries: 'VMBytecodeConstants'
	category: 'VMMaker-SpurMemoryManager'!

!SpurPlanningCompactor commentStamp: 'eem 6/11/2018 12:43' prior: 0!
SpurPlanningCompactor implements the classic planning compaction algorithm for Spur.  It uses the fact that there is room for a forwarding pointer in all objects to store the eventual position of an object in the first field.  It therefore first locates a large free chunk, or eden or a memory segment, to use as the savedFirstFieldsSpace, which it uses to store the first fields of objects that will be compacted.  It then makes at least three passes through the heap.

The first pass plans where live movable objects will go, copying their first field to the next slot in savedFirstFieldsSpace, and setting their forwarding pointer to point to their eventual location (see planCompactSavingForwarders).  The second pass updates all pointers in live pointer objects to point to objects' final destinations, including the fields in savedFirstFieldsSpace (see updatePointers and updatePointersInMobileObjects).  The third pass moves objects to their final positions, unmarking objects, and restoring saved first fields as it does so (see copyAndUnmark: and copyAndUnmarkMobileObjects).  If the forwarding fields of live objects in the to-be-moved portion of the entire heap won't fit in savedFirstFieldsSpace, then additional passes can be made until the entire heap has been compacted.  When snapshotting multiple passes are made, but when doing a normal GC only one pass is made.

Each pass uses a three finger algorithm, a simple extension of the classic two finger algorithm with an extra finger used to identify the lowest pinned object between the to and from fingers.  Objects are moved down, starting at the first free object or chunk, provided that they fit below the lowest pinned object above the to finger.  When an object won't fit the to finger is moved above the pinned object and the third finger is reset to the next pinned object below the from finger, if any.

Instance Variables
	anomaly							<Oop>
	biasForGC							<Boolean>
	firstFieldOfRememberedSet			<Oop>
	firstFreeObject						<Oop>
	firstMobileObject					<Oop>
	lastMobileObject					<Oop>
	mobileStart							<Integer address>
	objectAfterLastMobileObject		<Oop|nil>
	savedFirstFieldsSpace				<SpurContiguousObjStack>
	savedFirstFieldsSpaceWasAllocated	<Boolean>

anomaly
	- if any bogus object is detected by asserts, etc, it is stored in anomaly

biasForGC
	- true if compacting for GC, in which case do only one pass, or false if compacting for snapshot, in which case do as many passes as necessary to compact the entire heap.

firstFieldOfRememberedSet
	- the saved first field of the rememberedSet.  The rememberedSet must be relocated specially because it is not a pointer object.  And hence the first field needs to be extracted for proper relocation.

firstFreeObject
	- the first free object in a compaction pass.

firstMobileObject
	- the first mobile object in a compaction.  Unpinned objects from the firstMobileObject through to the lastMobileObject are implicitly forwarded.

lastMobileObject
	- the last mobile object in a compaction.  Unpinned objects from the firstMobileObject through to the lastMobileObject are implicitly forwarded.

mobileStart
	- the address of the first byte in firstFreeObject

objectAfterLastMobileObject
	- the object following the last object that can be moved, used when more than one pass is needed.

savedFirstFieldsSpace
	- the space holding the saved first fields, each overwritten by a forwarding pointer, for the objects from firstMobileObject through to lastMobileObject.

savedFirstFieldsSpaceWasAllocated
	- if true, the memory for savedFirstFieldsSpace was obtained via a call of sqAllocateMemorySegmentOfSize:Above:AllocatedSizeInto:!

!SpurPlanningCompactor class methodsFor: 'translation' stamp: 'eem 6/6/2018 18:52'!
declareCVarsIn: aCCodeGenerator
	super declareCVarsIn: aCCodeGenerator.
	self declareCAsOop: (self instVarNames select: [:iv| iv endsWith: 'Object']) in: aCCodeGenerator.
	aCCodeGenerator
		var: 'savedFirstFieldsSpace' type: #SpurContiguousObjStack! !

!SpurPlanningCompactor class methodsFor: 'analysis' stamp: 'eem 3/3/2017 11:52'!
identify32BitSignedComparisons
	"self identify32BitSignedComparisons"
	self identifySignedComparisonsFor: #(ObjectMemory Spur32BitMemoryManager)
		noise: #('(manager bytesInObject: largestFreeChunk) >= spaceEstimate'
				'(self classIndexOf: o*) > self isForwardedObjectClassIndexPun'
				'GCModeFull > 0'
				'ReceiverIndex + (objectMemory integerValueOf: sp*) < (objectMemory lengthOf: o*)'
				'availableSpace > 0'
				'bytes + 2 * 8 > availableSpace'
				'fmt* < manager firstCompiledMethodFormat'
				'fmt* < self firstCompiledMethodFormat'
				'fmt* <= 5'
				'gcPhaseInProgress > 0'
				'i <= finishIndex'
				'numPointerSlots > 0'
				'scavenger rememberedSetSize > 0')! !

!SpurPlanningCompactor class methodsFor: 'analysis' stamp: 'eem 3/3/2017 11:49'!
identify64BitSignedComparisons
	"self identify64BitSignedComparisons"
	self identifySignedComparisonsFor: #(ObjectMemory Spur64BitMemoryManager)
		noise: #('(manager bytesInObject: largestFreeChunk) >= spaceEstimate'
				'(self classIndexOf: o*) > self isForwardedObjectClassIndexPun'
				'GCModeFull > 0'
				'ReceiverIndex + (objectMemory integerValueOf: sp*) < (objectMemory lengthOf: o*)'
				'availableSpace > 0'
				'bytes + 2 * 8 > availableSpace'
				'fmt* < manager firstCompiledMethodFormat'
				'fmt* < self firstCompiledMethodFormat'
				'fmt* <= 5'
				'gcPhaseInProgress > 0'
				'i <= finishIndex'
				'numPointerSlots > 0'
				'scavenger rememberedSetSize > 0')! !

!SpurPlanningCompactor class methodsFor: 'analysis' stamp: 'cb 5/28/2018 09:35'!
identifyMul: options noise: noise
	"Mul can overflow...
	SpurPlanningCompactor identifyMul: #(
	ObjectMemory Spur32BitMemoryManager
	compactorClass SpurSelectiveCompactor) noise: #()"
	| vmm cg |
	vmm := (VMMaker forPlatform: 'Cross')
				interpreterClass: StackInterpreter;
				options: options.
	cg := [vmm buildCodeGeneratorForInterpreter]
			on: Notification
			do: [:ex|
				ex tag == #getVMMaker
					ifTrue: [ex resume: vmm]
					ifFalse: [ex pass]].
	cg vmClass preGenerationHook: cg.
	cg inferTypesForImplicitlyTypedVariablesAndMethods.
	cg retainMethods: self selectors.
	cg prepareMethods.
	cg doInlining: true.
	self selectors sort do:
		[:sel|
		(cg methodNamed: sel) ifNotNil:
			[:m|
			m parseTree nodesDo:
				[:node|
				(node isSend
				 and: [#* = node selector]) ifTrue:
					[Transcript ensureCr; nextPutAll: sel; space; print: node; flush]]]]! !

!SpurPlanningCompactor class methodsFor: 'analysis' stamp: 'eem 3/3/2017 11:41'!
identifySignedComparisonsFor: options noise: noise
	"self identify32BitSignedComparisons"
	"self identify64BitSignedComparisons"
	| vmm cg |
	vmm := (VMMaker forPlatform: 'Cross')
				interpreterClass: StackInterpreter;
				options: options.
	cg := [vmm buildCodeGeneratorForInterpreter]
			on: Notification
			do: [:ex|
				ex tag == #getVMMaker
					ifTrue: [ex resume: vmm]
					ifFalse: [ex pass]].
	cg vmClass preGenerationHook: cg.
	cg inferTypesForImplicitlyTypedVariablesAndMethods.
	cg retainMethods: self selectors.
	cg prepareMethods.
	cg doInlining: true.
	self selectors sort do:
		[:sel|
		(cg methodNamed: sel) ifNotNil:
			[:m|
			m parseTree nodesDo:
				[:node|
				(node isSend
				 and: [(#(< > <= >=) includes: node selector)
				 and: [({node receiver. node args first } anySatisfy:
						[:o| (cg typeFor: o in: m)
								ifNil: [true]
								ifNotNil: [:t| (cg isIntegralCType: t) and: [t first ~= $u]]])
				 and: [noise noneSatisfy: [:n| n match: node printString]]]]) ifTrue:
					[Transcript ensureCr; nextPutAll: sel; space; print: node; flush]]]]! !

!SpurPlanningCompactor class methodsFor: 'simulation' stamp: 'eem 2/3/2017 09:39'!
simulatorClass
	^SpurPlanningCompactorSimulator! !

!SpurPlanningCompactor methodsFor: 'compaction - api' stamp: 'eem 11/15/2019 09:03'!
biasForGC
	<inline: true>
	biasForGC := true! !

!SpurPlanningCompactor methodsFor: 'compaction - api' stamp: 'eem 12/21/2016 11:06'!
biasForSnapshot
	biasForGC := false! !

!SpurPlanningCompactor methodsFor: 'private' stamp: 'eem 2/3/2017 09:41'!
check: obj
	"No-op in the real class."
	<inline: true>! !

!SpurPlanningCompactor methodsFor: 'private' stamp: 'eem 06/22/2021 15:31'!
coalesceFrom: maybeStartOfFree
	"manager printOopsFrom: maybeStartOfFree to: manager endOfMemory"
	<var: 'maybeStartOfFree' type: #usqInt>
	|obj next objBytes nextBytes|
	<var: 'obj' type: #usqInt>
	<var: 'next' type: #usqInt>
	maybeStartOfFree >= manager endOfMemory ifTrue:
		[^self].
	obj := manager objectStartingAt: maybeStartOfFree.
	[next := manager oldSpaceObjectAfter: obj.
	 next < manager endOfMemory] whileTrue:
		[((manager isFreeObject: obj) and: [manager isFreeObject: next])
			ifTrue:
				[objBytes := manager bytesInBody: obj.
				 nextBytes := manager bytesInBody: next.
				 manager unlinkFreeChunk: obj chunkBytes: objBytes.
				 manager unlinkFreeChunk: next chunkBytes: nextBytes.
				 obj := manager freeChunkWithBytes: objBytes + nextBytes at: (manager startOfObject: obj)]
			ifFalse:
				[obj := next]]! !

!SpurPlanningCompactor methodsFor: 'compaction - api' stamp: 'cb 11/28/2019 20:51'!
compact
	"Sweep all of old space, sliding unpinned marked objects down over free and unmarked objects.
	 Let the segmentManager mark which segments contain pinned objects via notePinned:."
	| finalPass firstPass |
	<inline: #never> "for profiling"
	self initializeScanCheckingForFullyCompactedHeap ifTrue:
		[^self unmarkObjectsInFullyCompactedHeap].
	self initializeCompaction.
	firstPass := true.
	[finalPass := self planCompactSavingForwarders.
	 self assert: (self validRelocationPlanInPass: finalPass) = 0.
	 self updatePointers.
	 self copyAndUnmark: firstPass.
	 "Would like to check here, but can't if multi-pass."
	 false ifTrue: [manager checkFreeSpace: GCModeFull].
	 "Currently we do only a single pass if a normal GC, assuming that a pass will
	  always compact plenty of space. But we should perhaps check this assumption
	  by looking at the large free tree and seeing that the ratio of the largest free
	  chunk to the total ammount of free space is high."
	 finalPass or: [biasForGC]] whileFalse:
		[firstPass := false.
		 self reinitializeScanFrom: firstFreeObject.
		 firstMobileObject > firstFreeObject ifFalse: 
			["We get here when firstFreeObject is the last object in memory. 
		   	 firstMobileObject may be random value from previous compaction
		   	 in this case, and heap is fully compacted. Abort compaction."
		  	 "We set firstFreeObject to firstMobileObject for unmarking."
			 firstFreeObject := firstMobileObject.
			 self unmarkObjectsFromFirstFreeObject.
		  	 manager checkFreeSpace: GCModeFull.
			 ^self endCompaction].
		 self updateSavedFirstFieldsSpaceIfNecessary].
	manager checkFreeSpace: GCModeFull.
	self endCompaction! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 2/17/2017 11:48'!
copyAndUnmark: firstPass
	"Sweep the heap, unmarking all objects and moving mobile objects to their correct positions,
	 restoring their savedFirstFields."
	<inline: #never>
	| finalPass |
	self logPhase: 'copying and unmarking...'.
	firstPass ifTrue:
		[self unmarkInitialImmobileObjects].
	finalPass := self copyAndUnmarkMobileObjects.
	(self thereAreObjectsToMove
	 and: [finalPass not
	 and: [biasForGC]]) ifTrue: "only ever one pass if biasForGC is true."
		[self unmarkObjectsFromFirstFreeObject]! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 06/22/2021 15:31'!
copyAndUnmarkMobileObjects
	"Sweep the mobile portion of the heap, moving objects to their eventual locations, and clearing their marked bits.
	 Remember to update the savedFirstFields of pointer objects, as these have been forwarded.
	 Answer if the end of the heap was reached (savedFirstFieldsSpace has not overflowed).

	 The enumerations in planCompactSavingForwarders, updatePointersInMobileObjects and copyAndUnmarkMobileObjects
	 match.  We could implement them as a single enumeration method taking several block arguments, but arguably that
	 would make understanding an already tricky algorithm more difficult.  Instead we tolerate the duplication and encourage
	 the reader to diff the three methods to see where they diverge (e.g. via Cmd-shift-C)."
	<inline: #never>
	| toFinger top previousPin startOfPreviousPin |
	<var: 'o' type: #usqInt>
	<var: 'top' type: #usqInt>
	<var: 'toFinger' type: #usqInt>
	<var: 'previousPin' type: #usqInt>
	<var: 'startOfPreviousPin' type: #usqInt>
	self deny: (manager isMarked: firstFreeObject).
	toFinger := manager startOfObject: firstFreeObject.
	top := savedFirstFieldsSpace start.
	startOfPreviousPin := 0.
	manager allOldSpaceEntitiesForCompactingFrom: firstFreeObject to: (lastMobileObject ifNil: manager nilObject) do:
		[:o :n|
		 self check: o.
		 self assert: (previousPin
						ifNil: [toFinger <= (manager startOfObject: o)]
						ifNotNil: [(manager isMarked: previousPin) and: [toFinger <= startOfPreviousPin]]).
		 self assert: (savedFirstFieldsSpaceNotInOldSpace or: [toFinger < top]).
		 (manager isMarked: o) ifTrue:
			[(manager isPinned: o)
				ifTrue:
					[previousPin ifNil:
						[previousPin := o. startOfPreviousPin := manager startOfObject: o]]
				ifFalse:
					[| availableSpace bytes |
					 bytes := manager bytesInBody: o.
					 [toFinger <= startOfPreviousPin
					  and: [bytes ~= (availableSpace := startOfPreviousPin - toFinger)
					  and: [bytes + (2 * manager allocationUnit) > availableSpace]]] whileTrue:
						["The object does not fit in the space between toFinger and previousPin.
						   Move toFinger up to point at the first unmarked or mobile object after
						   previousPin, or, if previousPin is contiguous with o, to the start of this
						   object.  Update previousPin to be the next pinned object above toFInger
						   and below this object, or nil if no such pinned object exists.
						   Any unfillable gaps between adjacent pinned objects will be freed."
						 availableSpace > 0 ifTrue:
							[manager addFreeChunkWithBytes: availableSpace at: toFinger].
					 	 [self assert: ((manager isMarked: previousPin) and: [manager isPinned: previousPin]).
						  self unmarkPinned: previousPin.
						  toFinger := manager addressAfter: previousPin.
						  previousPin := manager objectStartingAt: toFinger.
						  (manager isMarked: previousPin)
						   and: [(manager isPinned: previousPin)
						   and: [previousPin < o]]]
							whileTrue.
						 "Now previousPin is either equal to o or mobile.
						  Move it to the next pinned object below o"
						 [previousPin >= o
						  or: [(manager isMarked: previousPin)
						  and: [manager isPinned: previousPin]]] whileFalse:
							[previousPin := manager oldSpaceObjectAfter: previousPin].
						 previousPin >= o
							ifTrue: [previousPin := nil. startOfPreviousPin := 0]
							ifFalse: [startOfPreviousPin := manager startOfObject: previousPin]].
					 self copyAndUnmarkObject: o to: toFinger bytes: bytes firstField: (manager longAt: top).
					 toFinger := toFinger + bytes.
					 (top := top + manager bytesPerOop) >= savedFirstFieldsSpace limit ifTrue:
						[self assert: savedFirstFieldsSpace top = (top - manager bytesPerOop).
						 self assert: n = objectAfterLastMobileObject.
						 previousPin ifNil: [previousPin := n. startOfPreviousPin := manager startOfObject: n].
						 "Create a free object for firstFreeObject to be set to on the next pass, but
						   do not link it into the free tree as it will be written over in that next pass."
						 toFinger < startOfPreviousPin
							ifTrue:
								[firstFreeObject := manager initFreeChunkWithBytes: startOfPreviousPin - toFinger at: toFinger]
							ifFalse:
								[firstFreeObject := previousPin].
						^false]]]].
	self freeFrom: toFinger upTo: manager endOfMemory nextObject: (previousPin ifNil: [objectAfterLastMobileObject ifNil: [manager objectAfter: firstFreeObject]]).
	self coalesceFrom: toFinger.
	^true! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 9/19/2018 17:29'!
copyAndUnmarkObject: o to: toFinger bytes: bytes firstField: firstField
	"Copy the object to toFinger, clearing its mark bit both in the target and the corpse,
	 and restoring its firstField, which was overwritten with a forwarding pointer."
	<inline: true>
	| numSlots destObj start |
	manager setIsMarkedOf: o to: false.
	numSlots := manager rawNumSlotsOf: o.
	destObj := (manager objectWithRawSlotsHasOverflowHeader: numSlots)
					ifTrue: [toFinger + manager baseHeaderSize]
					ifFalse: [toFinger].
	start := manager startOfObject: o given: numSlots.
	"memmove must be used since the ranges may overlap."
	manager
		memmove: toFinger asVoidPointer _: start asVoidPointer _: bytes;
		storePointerUnchecked: 0 ofObject: destObj withValue: firstField! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 11/28/2019 14:49'!
endCompaction
	<inline: false>
	manager
		unmarkSurvivingObjectsForCompact;
		endSlidingCompaction.
	self repinRememberedSet.
	self releaseSavedFirstFieldsSpace! !

!SpurPlanningCompactor methodsFor: 'space management' stamp: 'eem 06/22/2021 15:31'!
findHighestSuitableFreeBlock: spaceEstimate
	"If a freeBlock of size at least spaceEstimate exists high enough in the heap, choose it.
	 Ignoring pinned objects for now, the total ammount of shrinkage is expected to be
	 at least totalFreeOldSpace (because of collected objects).  So any free chunk which is
	 at or above endOfMemory - totalFreeOldSpace should not be corrupted during compaction.
	 Let's play with this for a while and see how we get on."
	<inline: true>
	manager findLargestFreeChunk ifNotNil:
		[:largestFreeChunk|
		((manager bytesInBody: largestFreeChunk) >= spaceEstimate
		 and: [largestFreeChunk asUnsignedInteger > (manager endOfMemory - manager freeSize) asUnsignedInteger]) ifTrue:
			[^largestFreeChunk]].
	^nil! !

!SpurPlanningCompactor methodsFor: 'private' stamp: 'eem 2/3/2017 09:44'!
forwardMobileObject: o to: toFinger savedFirstFieldPtr: savedFirstFieldPtr
	"Forward a mobile object to some new location, saving its first field through savedFirstFieldPtr.
	 Don't use forward:to:; we dont want to alter the object in any way other than by setting the forwarding pointer."
	<inline: true>
	lastMobileObject := o.
	manager
		longAt: savedFirstFieldPtr
			put: (manager fetchPointer: 0 ofObject: o);
		storePointerUnchecked: 0
			ofObject: o
				withValue: ((manager hasOverflowHeader: o)
								ifTrue: [toFinger + manager baseHeaderSize]
								ifFalse: [toFinger]).
	self recordMovementOf: o to: toFinger savedFirstFieldPtr: savedFirstFieldPtr! !

!SpurPlanningCompactor methodsFor: 'private' stamp: 'eem 2/1/2017 13:40'!
freeFrom: initialToFinger upTo: limit nextObject: nextObject
	"Free from toFinger up to limit, dealing with possible intervening pinned objects."
	<inline: false>
	<var: 'limit' type: #usqInt>
	<var: 'initialToFinger' type: #usqInt>
	| toFinger obj objStart |
	<var: 'objStart' type: #usqInt>
	<var: 'toFinger' type: #usqInt>
	self cCode: [] inSmalltalk:
		[coInterpreter cr; cr; print: 'freeing at '; printHexnp: initialToFinger; print: ' up to '; printHexnp: limit; cr].
	toFinger := initialToFinger.
	objStart := manager startOfObject: nextObject.
	toFinger < objStart ifTrue:
		[manager addFreeChunkWithBytes: objStart - toFinger at: toFinger].
	toFinger := objStart.
	[objStart < limit] whileTrue:
		[obj := manager objectStartingAt: objStart.
		 ((manager isMarked: obj) and: [manager isPinned: obj])
			ifTrue:
				[self unmarkPinned: obj.
				 toFinger < objStart ifTrue:
					[manager addFreeChunkWithBytes: objStart - toFinger at: toFinger].
				 toFinger := objStart := manager addressAfter: obj]
			ifFalse:
				[objStart := manager addressAfter: obj]].
	limit > toFinger ifTrue:
		[manager addFreeChunkWithBytes: limit - toFinger at: toFinger]! !

!SpurPlanningCompactor methodsFor: 'instance initialization' stamp: 'eem 1/11/2017 12:27'!
initialize
	biasForGC := true.
	savedFirstFieldsSpace := SpurContiguousObjStack new.
	savedFirstFieldsSpaceNotInOldSpace := true! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 3/9/2020 19:56'!
initializeCompaction
	manager checkFreeSpace: GCModeFull.
	self selectSavedFirstFieldsSpace.
	self unpinRememberedSet.
	manager
		resetFreeListHeads;
		prepareObjStacksForPlanningCompactor;
		totalFreeOldSpace: 0;
		beginSlidingCompaction.
	savedFirstFieldsSpace top: savedFirstFieldsSpace start - manager bytesPerOop! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 2/3/2017 16:33'!
initializeScanCheckingForFullyCompactedHeap
	"Scan for firstFreeObject and firstMobileObject from the start of memory.
	 Answer if the heap is already fully compacted."
	firstMobileObject := lastMobileObject := objectAfterLastMobileObject := nil.
	self reinitializeScanFrom: manager hiddenRootsObject.
	firstFreeObject ifNil:
		[self error: 'uncompactable heap; no unmarked objects found'].
	^firstMobileObject >= manager endOfMemory! !

!SpurPlanningCompactor methodsFor: 'private' stamp: 'eem 12/30/2016 18:02'!
isMobile: obj
	<inline: true>
	^(self oop: obj isGreaterThanOrEqualTo: mobileStart andLessThanOrEqualTo: lastMobileObject)
	 and: [(manager isPinned: obj) not]! !

!SpurPlanningCompactor methodsFor: 'private' stamp: 'eem 12/30/2016 18:02'!
isPostMobile: obj
	"For asserts"
	^self oop: obj isGreaterThanOrEqualTo: mobileStart andLessThanOrEqualTo: lastMobileObject! !

!SpurPlanningCompactor methodsFor: 'private' stamp: 'eem 12/27/2016 17:55'!
logPhase: phaseName
	<inline: true>
	self cCode: '' inSmalltalk: [coInterpreter transcript nextPutAll: phaseName; flush].! !

!SpurPlanningCompactor methodsFor: 'private' stamp: 'eem 1/2/2017 19:53'!
numPointerSlotsWhileCompactingOf: obj withFormat: fmt savedFirstFieldPointer: firstFieldPtr
	"This is a version of SpurMemoryManager>>numPointerSlotsOf: that deals with the
	 possibility of obj being a CompiledMethod whose header is in savedFirstFieldsSpace.
	 Answer the number of pointer fields in the given object.
	 Works with CompiledMethods, as well as ordinary objects."
	<inline: true>
	| contextSize numLiterals header |
	self assert: (firstFieldPtr notNil and: [self isMobile: obj]).
	fmt <= manager lastPointerFormat ifTrue:
		[(fmt = manager indexablePointersFormat
		  and: [manager isContextNonImm: obj]) ifTrue:
			["contexts end at the stack pointer"
			contextSize := coInterpreter fetchStackPointerOf: obj.
			^CtxtTempFrameStart + contextSize].
		^manager numSlotsOf: obj  "all pointers"].
	self deny: fmt = manager forwardedFormat.
	fmt < manager firstCompiledMethodFormat ifTrue: [^0]. "no pointers"

	"CompiledMethod: contains both pointers and bytes"
	header := manager methodHeaderFromSavedFirstField: (manager longAt: firstFieldPtr).
	numLiterals := manager literalCountOfMethodHeader: header.
	^numLiterals + LiteralStart! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 06/22/2021 15:31'!
planCompactSavingForwarders
	"Sweep the heap from firstFreeObject forwarding marked objects to where they
	 can be moved to, saving their forwarding pointer in savedFirstFieldsSpace.
	 Continue until either the end of the heap is reached or savedFirstFieldsSpace is full.
	 Answer if the end of the heap was reached (savedFirstFieldsSpace has not overflowed).

	 The enumerations in planCompactSavingForwarders, updatePointersInMobileObjects and copyAndUnmarkMobileObjects
	 match.  We could implement them as a single enumeration method taking several block arguments, but arguably that
	 would make understanding an already tricky algorithm more difficult.  Instead we tolerate the duplication and encourage
	 the reader to diff the three methods to see where they diverge (e.g. via Cmd-shift-C)."
	<inline: #never>
	| toFinger top previousPin startOfPreviousPin |
	<var: 'o' type: #usqInt>
	<var: 'top' type: #usqInt>
	<var: 'toFinger' type: #usqInt>
	<var: 'previousPin' type: #usqInt>
	<var: 'startOfPreviousPin' type: #usqInt>
	savedFirstFieldsSpace top < savedFirstFieldsSpace start ifTrue:
		[self logPhase: 'planning...'].
	self deny: (manager isMarked: firstFreeObject).
	toFinger := manager startOfObject: firstFreeObject.
	top := savedFirstFieldsSpace start.
	startOfPreviousPin := 0.
	manager allOldSpaceEntitiesFrom: firstFreeObject do:
		[:o|
		 self check: o.
		 self assert: (previousPin
						ifNil: [toFinger <= (manager startOfObject: o)]
						ifNotNil: [(manager isMarked: previousPin) and: [toFinger <= startOfPreviousPin]]).
		 self assert: (savedFirstFieldsSpaceNotInOldSpace or: [toFinger < top]).
		 (manager isMarked: o) ifTrue:
			[(manager isPinned: o)
				ifTrue:
					[previousPin ifNil:
						[previousPin := o. startOfPreviousPin := manager startOfObject: o]]
				ifFalse:
					[| availableSpace bytes |
					 bytes := manager bytesInBody: o.
					 [toFinger <= startOfPreviousPin
					  and: [bytes ~= (availableSpace := startOfPreviousPin - toFinger)
					  and: [bytes + (2 * manager allocationUnit) > availableSpace]]] whileTrue:
						["The object does not fit in the space between toFinger and previousPin.
						   Move toFinger up to point at the first unmarked or mobile object after
						   previousPin, or, if previousPin is contiguous with o, to the start of this
						   object.  Update previousPin to be the next pinned object above toFInger
						   and below this object, or nil if no such pinned object exists.
						   Any unfillable gaps between adjacent pinned objects will be freed."
					 	 [toFinger := manager addressAfter: previousPin.
						  previousPin := manager objectStartingAt: toFinger.
						  (manager isMarked: previousPin)
						   and: [(manager isPinned: previousPin)
						   and: [previousPin < o]]]
							whileTrue.
						 "Now previousPin is either equal to o or mobile.
						  Move it to the next pinned object below o"
						 [previousPin >= o
						  or: [(manager isMarked: previousPin)
						  and: [manager isPinned: previousPin]]] whileFalse:
							[previousPin := manager oldSpaceObjectAfter: previousPin].
						 previousPin >= o
							ifTrue: [previousPin := nil. startOfPreviousPin := 0]
							ifFalse: [startOfPreviousPin := manager startOfObject: previousPin]].
					 self forwardMobileObject: o to: toFinger savedFirstFieldPtr: top.
					 toFinger := toFinger + bytes.
					 (top := top + manager bytesPerOop) >= savedFirstFieldsSpace limit ifTrue:
						[savedFirstFieldsSpace top: top - manager bytesPerOop.
						 objectAfterLastMobileObject := manager oldSpaceObjectAfter: lastMobileObject.
						 ^false]]]].
	"If the heap is already fully compacted there will be no lastMobileObject..."
	lastMobileObject ifNotNil:
		[savedFirstFieldsSpace top: top - manager bytesPerOop.
		 objectAfterLastMobileObject := manager oldSpaceObjectAfter: lastMobileObject].
	^true! !

!SpurPlanningCompactor methodsFor: 'debug support' stamp: 'eem 2/17/2017 10:50'!
printTheBogons: aBogon
	<inline: true>
	coInterpreter
		print: 'bogon '; printHexnp: aBogon; cr;
		print: 'anomaly '; printHexnp: anomaly; cr! !

!SpurPlanningCompactor methodsFor: 'private' stamp: 'eem 2/3/2017 09:44'!
recordMovementOf: o to: toFinger savedFirstFieldPtr: savedFirstFieldPtr
	"No-op in the real class."
	<inline: true>! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 11/28/2019 12:33'!
reinitializeScanFrom: initialObject
	"Search for firstFreeObject and firstMobileObject from initialObject, which is the
	 hiddenRootsObject on the first pass, and the objectAfterLastMobileObject on
	 subsequent passes)."
	firstFreeObject := self scanForFirstFreeAndFirstMobileObjectFrom: initialObject.
	firstFreeObject ifNotNil:
		[mobileStart := manager startOfObject: firstFreeObject].
	objectAfterLastMobileObject ifNotNil:
		[manager allOldSpaceEntitiesFrom: firstFreeObject to: objectAfterLastMobileObject do:
			[:o|
			 ((manager isPinned: o)
			  or: [(manager isMarked: o) not
			  or: [objectAfterLastMobileObject = o]]) ifFalse:
				[manager setIsMarkedOf: o to: false]].
		 firstMobileObject := objectAfterLastMobileObject]! !

!SpurPlanningCompactor methodsFor: 'space management' stamp: 'eem 1/11/2017 12:30'!
releaseSavedFirstFieldsSpace
	<inline: true>
	self savedFirstFieldsSpaceWasAllocated ifTrue:
		[manager
			sqDeallocateMemorySegmentAt: savedFirstFieldsSpace start asVoidPointer
			OfSize: savedFirstFieldsSpace limit - savedFirstFieldsSpace start]! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 12/24/2016 09:45'!
relocateObjectsInHeapEntity: heapEntity from: startIndex to: finishIndex
	"Sweep the fields in some non-pointer heap entity (objStackPage, rememberedSet),
	 updating all references to mobile objects to their eventual locations.  Answer the
	 heapEntity's eventual location."
	<inline: true>
	startIndex to: finishIndex do:
		[:i| | oop fwd |
		 oop := manager fetchPointer: i ofObject: heapEntity.
		 ((manager isNonImmediate: oop) and: [self isMobile: oop]) ifTrue:
			[self assert: (manager isMarked: oop).
			 fwd := manager fetchPointer: 0 ofObject: oop.
			 self assert: (self isPostMobile: fwd).
			 manager storePointerUnchecked: i ofObject: heapEntity withValue: fwd]].
	^(self isMobile: heapEntity)
		ifTrue: [manager fetchPointer: 0 ofObject: heapEntity]
		ifFalse: [heapEntity]! !

!SpurPlanningCompactor methodsFor: 'gc - scavenge/compact' stamp: 'eem 12/24/2016 14:35'!
remapObj: objOop
	"Scavenge or simply follow objOop.  Answer the new location of objOop.
	 The send should have been guarded by a send of shouldRemapOop:.
	 The method is called remapObj: for compatibility with ObjectMemory."
	<api>
	<inline: false>
	^manager slidingCompactionRemapObj: objOop! !

!SpurPlanningCompactor methodsFor: 'private' stamp: 'eem 12/23/2016 17:48'!
repinRememberedSet
	<inline: true>
	scavenger rememberedSetSize > 0 ifTrue:
		[manager storePointerUnchecked: 0 ofObject: manager rememberedSetObj withValue: firstFieldOfRememberedSet].
	manager setIsPinnedOf: manager rememberedSetObj to: true.
	scavenger relocateRememberedSet! !

!SpurPlanningCompactor methodsFor: 'space management' stamp: 'eem 1/25/2017 09:06'!
savedFirstFieldsSpaceInFreeChunk
	<inline: true>
	^savedFirstFieldsSpaceNotInOldSpace not
	 and: [self oop: savedFirstFieldsSpace start isGreaterThan: manager nilObject]! !

!SpurPlanningCompactor methodsFor: 'space management' stamp: 'eem 1/25/2017 09:29'!
savedFirstFieldsSpaceWasAllocated
	<inline: true>
	^savedFirstFieldsSpaceNotInOldSpace
	 and: [self oop: savedFirstFieldsSpace start isGreaterThan: manager nilObject]! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 11/28/2019 12:34'!
scanForFirstFreeAndFirstMobileObjectFrom: initialObject
	"Scan from initialObject, setting firstMobileObject to the first marked
	 object after the first free object found, or endOfMemory if none is found.
	 Answer the first free object found, or nil if none."
	<inline: false>
	| firstFree |
	firstMobileObject := manager endOfMemory.
	manager allOldSpaceEntitiesFrom: initialObject do:
		[:o|
		(manager isMarked: o)
			ifTrue:
				[firstFree ifNotNil:
					[firstMobileObject := o.
					 ^firstFree]]
			ifFalse:
				[firstFree ifNil:
					[firstFree := o]]].
	^firstFree! !

!SpurPlanningCompactor methodsFor: 'space management' stamp: 'eem 1/5/2017 14:32'!
selectSavedFirstFieldsSpace
	"To compact the heap the algorithm must save the first field (used for the forwarding pointer)
	 of all moved objects. This is done in savedFirstFieldsSpace, a contiguous block of memory borrowed
	 for the duration of compaction. In a 32-bit system a typical upper bound on the space needed
	 is about 1/40 of the heap size.  The default new space size of 4Mb provides an eden of about
	 3.6 Mb, which would serve the needs of a 144 Mb heap.  The default segment increment of
	 16 Mb would serve the needs of a 640 Mb heap. Make an estimate of the size needed, and
	 either use eden, a large free chunk, or a newly-allocated segment, falling back on eden if
	 the alternatives can't be had."
	<inline: true>
	| spaceEstimate sizeOfEden |
	spaceEstimate := manager endOfMemory - manager firstObject // 40.
	sizeOfEden := scavenger eden limit - scavenger eden start.
	spaceEstimate > sizeOfEden ifTrue:
		[(self findHighestSuitableFreeBlock: spaceEstimate) ifNotNil:
			[:highestSuitableFreeBlock|
			 (spaceEstimate > (manager sizeOfFree: highestSuitableFreeBlock)
			  and: [self useSegmentForSavedFirstFieldsSpace: spaceEstimate]) ifTrue:
				[^self].
			 (manager sizeOfFree: highestSuitableFreeBlock) > sizeOfEden ifTrue:
				[self useFreeChunkForSavedFirstFieldsSpace: highestSuitableFreeBlock.
				 ^self]].
		 (self useSegmentForSavedFirstFieldsSpace: spaceEstimate) ifTrue:
			[^self]].
	self useEdenForSavedFirstFieldsSpace! !

!SpurPlanningCompactor methodsFor: 'gc - scavenge/compact' stamp: 'eem 12/26/2016 10:34'!
shouldRemapObj: objOop
	"Answer if the obj should be scavenged, or simply followed. Sent via the compactor
	 from shouldRemapObj:.  We test for being already scavenged because mapStackPages
	 via mapInterpreterOops may be applied twice in the context of a global GC where a
	 scavenge, followed by a scan-mark-free, and final compaction passes may result in
	 scavenged fields being visited twice."
	<api>
	<inline: false>
	^manager slidingCompactionShouldRemapObj: objOop! !

!SpurPlanningCompactor methodsFor: 'private' stamp: 'eem 2/6/2017 10:17'!
thereAreObjectsToMove
	<inline: true>
	^lastMobileObject notNil! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 1/3/2017 09:12'!
unmarkInitialImmobileObjects
	"Sweep the initial immobile heap, unmarking all objects up to the first mobile object."
	manager allOldSpaceObjectsFrom: manager firstObject do:
		[:o|
		self check: o.
		(self oop: o isGreaterThanOrEqualTo: firstMobileObject) ifTrue:
			[^self].
		manager setIsMarkedOf: o to: false]! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 06/22/2021 15:31'!
unmarkObjectsFromFirstFreeObject
	"Sweep the final immobile heap, freeing and coalescing unmarked and free objects,
	 and unmarking all marked objects up to the end of memory."
	| startOfFree freeBytes |
	freeBytes := 0.
	manager allOldSpaceEntitiesFrom: firstFreeObject do:
		[:o|
		self check: o.
		(manager isMarked: o)
			ifFalse:
				[startOfFree ifNil: [startOfFree := manager startOfObject: o].
				 freeBytes := freeBytes + (manager bytesInBody: o)]
			ifTrue:
				[startOfFree ifNotNil:
					[manager addFreeChunkWithBytes: freeBytes at: startOfFree.
					 startOfFree := nil.
					 freeBytes := 0].
				 (manager isPinned: o)
					ifTrue: [self unmarkPinned: o]
					ifFalse: [manager setIsMarkedOf: o to: false]]].
	startOfFree ifNotNil:
		[manager addFreeChunkWithBytes: freeBytes at: startOfFree]! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 12/27/2016 12:57'!
unmarkObjectsInFullyCompactedHeap
	self unmarkInitialImmobileObjects.
	manager unmarkSurvivingObjectsForCompact! !

!SpurPlanningCompactor methodsFor: 'private' stamp: 'eem 1/12/2017 09:18'!
unmarkPinned: pinnedObj
	<inline: true>
	(manager isSegmentBridge: pinnedObj) ifFalse:
		[manager setIsMarkedOf: pinnedObj to: false.
		 manager segmentManager notePinned: pinnedObj]! !

!SpurPlanningCompactor methodsFor: 'private' stamp: 'eem 12/23/2016 17:45'!
unpinRememberedSet
	<inline: true>
	firstFieldOfRememberedSet := manager fetchPointer: 0 ofObject: manager rememberedSetObj.
	manager setIsPinnedOf: manager rememberedSetObj to: false! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 2/6/2017 10:19'!
updatePointers
	"Sweep the heap, updating all objects to their eventual locations.
	 Remember to update the savedFirstFields of pointer objects, as these have been forwarded."
	<inline: #never>
	| onePass |
	self logPhase: 'updating pointers...'.
	self thereAreObjectsToMove ifFalse:
		[^self].
	self assert: (manager startOfObject: firstFreeObject) = mobileStart.
	coInterpreter mapInterpreterOops.
	manager mapExtraRoots.
	self updatePointersInManagerHeapEntities.
	self updatePointersInSurvivingObjects.
	self updatePointersInInitialImmobileObjects.
	onePass := self updatePointersInMobileObjects.
	onePass ifFalse:
		[self updatePointersInObjectsOverflowingSavedFirstFieldsSpace]! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 12/27/2016 15:41'!
updatePointersIn: obj
	"Sweep the pointer fields in obj, updating all references to mobile objects to their eventual locations."
	<inline: true>
	| numPointerSlots |
	numPointerSlots := manager numPointerSlotsOf: obj.
	0 to: numPointerSlots - 1 do:
		[:i| | oop fwd |
		 oop := manager fetchPointer: i ofObject: obj.
		 ((manager isNonImmediate: oop) and: [self isMobile: oop]) ifTrue:
			[self assert: ((manager isMarked: oop) or: [obj = manager hiddenRootsObject]).
			 fwd := manager fetchPointer: 0 ofObject: oop.
			 self assert: (self isPostMobile: fwd).
			 manager storePointerUnchecked: i ofObject: obj withValue: fwd]]! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 1/2/2017 19:52'!
updatePointersIn: obj savedFirstFieldPointer: firstFieldPtr
	"Sweep the pointer fields in obj, updating all references to mobile objects to their eventual locations.
	 firstFieldPtr is supplied for mobile objects so that the saved first field can be updated, and so that
	 the first field of a compiled method (which is its header, or reference to a CogMethod holding its header)
	 can be retrieved."
	<inline: false>
	| fmt numPointerSlots oop fwd |
	fmt := manager formatOf: obj.
	numPointerSlots := self numPointerSlotsWhileCompactingOf: obj withFormat: fmt savedFirstFieldPointer: firstFieldPtr.
	(fmt <= manager lastPointerFormat "excludes CompiledMethod"
	 and: [numPointerSlots > 0]) ifTrue:
		["Relocate the saved first field; Note that CompiledMethods can be excluded since their
		  first field is either a SmallInteger or a reference to a CogMethod outside of oldSpace."
		 oop := manager longAt: firstFieldPtr.
		 ((manager isNonImmediate: oop) and: [self isMobile: oop]) ifTrue:
			[self assert: (manager isMarked: oop).
			 fwd := manager fetchPointer: 0 ofObject: oop.
			 self assert: (self isPostMobile: fwd).
			 manager longAt: firstFieldPtr put: fwd]].
	1 to: numPointerSlots - 1 do:
		[:i|
		 oop := manager fetchPointer: i ofObject: obj.
		 ((manager isNonImmediate: oop) and: [self isMobile: oop]) ifTrue:
			[self assert: ((manager isMarked: oop) or: [obj = manager hiddenRootsObject]).
			 fwd := manager fetchPointer: 0 ofObject: oop.
			 self assert: (self isPostMobile: fwd).
			 manager storePointerUnchecked: i ofObject: obj withValue: fwd]]! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 2/3/2017 10:31'!
updatePointersInInitialImmobileObjects
	"Sweep the initial immobile heap, updating all references to mobile objects to their eventual locations."
	manager allOldSpaceObjectsFrom: manager firstObject do:
		[:o|
		self check: o.
		(self oop: o isGreaterThanOrEqualTo: firstFreeObject) ifTrue:
			[^self].
		"would like to assert this, but it isn't true if more than one pass: self assert: (manager isMarked: o)."
		self updatePointersIn: o]! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 6/7/2017 18:42'!
updatePointersInManagerHeapEntities
	"The special non-pointer objects containing pointers, which are the objStacks and the rememberedSet,
	 must be updated manually since they will not be recognized as containing pointers in the normal sweep."
	manager relocateObjStacksForPlanningCompactor.
	(scavenger rememberedSetSize > 0
	 and: [self isMobile: firstFieldOfRememberedSet]) ifTrue:
		[firstFieldOfRememberedSet := manager fetchPointer: 0 ofObject: firstFieldOfRememberedSet].
	self relocateObjectsInHeapEntity: manager rememberedSetObj from: 1 to: scavenger rememberedSetSize - 1
	"Note that we /must not/ set the rememberedSetObj here since it is a slot in the hiddenRootsObj
	 and will be updated normally in updatePointersInInitialImmobileObjects.  So do not do
	(self isMobile: manager rememberedSetObj) ifTrue:
		[manager rememberedSetObj: (manager fetchPointer: 0 ofObject: manager rememberedSetObj)]"! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 06/22/2021 15:31'!
updatePointersInMobileObjects
	"Sweep the mobile portion of the heap, updating all references to objects to their eventual locations.
	 Remember to update the savedFirstFields of pointer objects, as these have been forwarded.
	 Answer if the end of the heap was reached (savedFirstFieldsSpace has not overflowed).

	 The enumerations in planCompactSavingForwarders, updatePointersInMobileObjects and copyAndUnmarkMobileObjects
	 match.  We could implement them as a single enumeration method taking several block arguments, but arguably that
	 would make understanding an already tricky algorithm more difficult.  Instead we tolerate the duplication and encourage
	 the reader to diff the three methods to see where they diverge (e.g. via Cmd-shift-C)."
	| toFinger top previousPin startOfPreviousPin |
	<var: 'o' type: #usqInt>
	<var: 'top' type: #usqInt>
	<var: 'toFinger' type: #usqInt>
	<var: 'previousPin' type: #usqInt>
	<var: 'startOfPreviousPin' type: #usqInt>
	self deny: (manager isMarked: firstFreeObject).
	toFinger := manager startOfObject: firstFreeObject.
	top := savedFirstFieldsSpace start.
	startOfPreviousPin := 0.
	manager allOldSpaceEntitiesFrom: firstFreeObject do:
		[:o|
		 self check: o.
		 self assert: (previousPin
						ifNil: [toFinger <= (manager startOfObject: o)]
						ifNotNil: [(manager isMarked: previousPin) and: [toFinger <= startOfPreviousPin]]).
		 (manager isMarked: o) ifTrue:
			[(manager isPinned: o)
				ifTrue:
					[previousPin ifNil:
						[previousPin := o. startOfPreviousPin := manager startOfObject: o].
					 self updatePointersIn: o]
				ifFalse:
					[| availableSpace bytes |
					 bytes := manager bytesInBody: o.
					 [toFinger <= startOfPreviousPin
					  and: [bytes ~= (availableSpace := startOfPreviousPin - toFinger)
					  and: [bytes + (2 * manager allocationUnit) > availableSpace]]] whileTrue:
						["The object does not fit in the space between toFinger and previousPin.
						   Move toFinger up to point at the first unmarked or mobile object after
						   previousPin, or, if previousPin is contiguous with o, to the start of this
						   object.  Update previousPin to be the next pinned object above toFInger
						   and below this object, or nil if no such pinned object exists.
						   Any unfillable gaps between adjacent pinned objects will be freed."
					 	 [toFinger := manager addressAfter: previousPin.
						  previousPin := manager objectStartingAt: toFinger.
						  (manager isMarked: previousPin)
						   and: [(manager isPinned: previousPin)
						   and: [previousPin < o]]]
							whileTrue.
						 "Now previousPin is either equal to o or mobile.
						  Move it to the next pinned object below o"
						 [previousPin >= o
						  or: [(manager isMarked: previousPin)
						  and: [manager isPinned: previousPin]]] whileFalse:
							[previousPin := manager oldSpaceObjectAfter: previousPin].
						 previousPin >= o
							ifTrue: [previousPin := nil. startOfPreviousPin := 0]
							ifFalse: [startOfPreviousPin := manager startOfObject: previousPin]].
					 self updatePointersIn: o savedFirstFieldPointer: top.
					 toFinger := toFinger + bytes.
					 (top := top + manager bytesPerOop) >= savedFirstFieldsSpace limit ifTrue:
						[self assert: savedFirstFieldsSpace top = (top - manager bytesPerOop).
						 ^false]]]].
	self assert: savedFirstFieldsSpace top = (top - manager bytesPerOop).
	^true! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 2/1/2017 11:12'!
updatePointersInObjectsOverflowingSavedFirstFieldsSpace
	"Sweep the final immobile heap, is any (those objects with no room in savedFirstFieldsSpace
	 in the current pass) updating all references to mobile objects to their eventual locations."
	manager allOldSpaceObjectsFrom: objectAfterLastMobileObject do:
		[:o|
		self check: o.
		(manager isMarked: o) ifTrue:
			[self updatePointersIn: o]]! !

!SpurPlanningCompactor methodsFor: 'compaction' stamp: 'eem 12/30/2016 20:13'!
updatePointersInSurvivingObjects
	"Sweep pastSpace, updating all references to mobile objects to their eventual locations."
	manager allPastSpaceObjectsDo:
		[:o|
		self check: o.
		self assert: (manager isMarked: o).
		self updatePointersIn: o]! !

!SpurPlanningCompactor methodsFor: 'space management' stamp: 'eem 3/27/2017 14:37'!
updateSavedFirstFieldsSpaceIfNecessary
	"If savedFirstFieldsSpace is a free chunk then it may need to be repositioned if there is more than one pass."
	<inline: true>
	(self savedFirstFieldsSpaceInFreeChunk
	 and: [savedFirstFieldsSpace start < lastMobileObject]) ifTrue:
		[(self findHighestSuitableFreeBlock: savedFirstFieldsSpace limit - savedFirstFieldsSpace start)
			ifNotNil:
				[:largestFreeChunk|
				 self useFreeChunkForSavedFirstFieldsSpace: largestFreeChunk]
			ifNil:
				[self useEdenForSavedFirstFieldsSpace]].

	savedFirstFieldsSpace top: savedFirstFieldsSpace start - manager bytesPerOop! !

!SpurPlanningCompactor methodsFor: 'space management' stamp: 'eem 1/11/2017 12:30'!
useEdenForSavedFirstFieldsSpace
	"Use teden to hold the savedFirstFieldsSpace."
	<inline: true>
	savedFirstFieldsSpace
		start: scavenger eden start;
		limit: scavenger eden limit.
	savedFirstFieldsSpaceNotInOldSpace := true.
	self deny: self savedFirstFieldsSpaceWasAllocated! !

!SpurPlanningCompactor methodsFor: 'space management' stamp: 'eem 3/27/2017 12:35'!
useFreeChunkForSavedFirstFieldsSpace: highestSuitableFreeBlock
	"Use the supplied free chunk to hold the savedFirstFieldsSpace. Invoked when
	 eden is found not to be big enough for the job. Avoid the first few fields so as
	 not to destroy the free chunk and there by confuse object enumeration."
	<inline: true>
	self assert: (manager validFreeTreeChunk: highestSuitableFreeBlock).
	savedFirstFieldsSpace
		start: highestSuitableFreeBlock + (manager freeChunkLargerIndex * manager bytesPerOop);
		limit: (manager addressAfter: highestSuitableFreeBlock).
	savedFirstFieldsSpaceNotInOldSpace := false.
	self deny: self savedFirstFieldsSpaceWasAllocated! !

!SpurPlanningCompactor methodsFor: 'space management' stamp: 'eem 7/7/2017 12:19'!
useSegmentForSavedFirstFieldsSpace: spaceEstimate
	"Attempt to allocate a memory segment large enough to hold the savedFirstFieldsSpace.
	 Invoked when neither eden nor a large free chunk are found to be big enough for the job."
	| roundedSize allocatedSize |
	<var: #segAddress type: #'void *'>
	roundedSize := spaceEstimate + 1023 // 1024 * 1024.
	(manager "sent to the manager so that the simulator can increase memory to simulate a new segment"
		sqAllocateMemorySegmentOfSize: roundedSize
		Above: (manager segmentManager firstGapOfSizeAtLeast: roundedSize)
		AllocatedSizeInto: (self cCode: [self addressOf: allocatedSize]
								inSmalltalk: [[:sz| allocatedSize := sz]])) ifNotNil:
		[:segAddress|
		 savedFirstFieldsSpace
			start: segAddress asUnsignedIntegerPtr;
			limit: segAddress asUnsignedIntegerPtr + allocatedSize.
		 savedFirstFieldsSpaceNotInOldSpace := true.
		 self assert: self savedFirstFieldsSpaceWasAllocated.
		 ^true].
	^false! !

!SpurPlanningCompactor methodsFor: 'private' stamp: 'eem 06/22/2021 15:31'!
validRelocationPlanInPass: onePass
	"Answer 0 if all the mobile objects from firstMobileObject to lastMobileObject
	 have sane forwarding addresses, and that savedFirstFieldsSpace is of
	 matching capacity.  Otherwise answer an error code identifying the anomaly."
	| nMobiles toFinger |
	<var: 'toFinger' type: #usqInt>
	<var: 'destination' type: #usqInt>
	nMobiles := 0.
	toFinger := mobileStart.
	anomaly := nil.
	manager allOldSpaceEntitiesFrom: firstMobileObject do:
		[:o| | destination |
		 self check: o.
		 (manager isMarked: o) ifTrue:
			[(manager isPinned: o) ifFalse:
				[nMobiles := nMobiles + 1.
				 destination := manager fetchPointer: 0 ofObject: o.
				 destination >= toFinger ifFalse:
					[anomaly := o. ^1].
				 toFinger := toFinger + (manager bytesInBody: o).
				 (self oop: o isGreaterThan: lastMobileObject) ifTrue:
					[anomaly := o. ^2].
				 o = lastMobileObject ifTrue:
					[^savedFirstFieldsSpace top + manager bytesPerOop - savedFirstFieldsSpace start / manager bytesPerOop
					   = nMobiles
						ifTrue: [0]
						ifFalse: [3]]]]].
	"N.B. written this way so that if there are no mobiles the expression evaluates to 0 in Smalltalk /and/ in C unsigned arithmetic."
	^savedFirstFieldsSpace top + manager bytesPerOop - savedFirstFieldsSpace start / manager bytesPerOop
	  = nMobiles
		ifTrue: [0]
		ifFalse: [4]! !

SpurPlanningCompactor subclass: #SpurPlanningCompactorSimulator
	instanceVariableNames: 'interestingObj relocationMap sffsMode'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManagerSimulation'!

!SpurPlanningCompactorSimulator methodsFor: 'private' stamp: 'eem 2/3/2017 14:16'!
check: obj
	<inline: true>
	sffsMode ifNotNil: [self deny: ((manager isGrey: obj) or: [(manager isImmutable: obj)  or: [manager isRemembered: obj]])].
	obj = interestingObj ifTrue: [self halt].
	"this debugged the misuse of the largest free chunk:"
	"(manager checkForLeaks bitAnd: GCModeFull+GCModeFreeSpace) = GCModeFull ifTrue:
		[self deny: ((manager isEnumerableObject: obj) and: [(manager heapMapAtWord: obj) = 0])]"! !

!SpurPlanningCompactorSimulator methodsFor: 'accessing' stamp: 'eem 2/3/2017 09:54'!
forceMultiPass
	sffsMode := #multiPass! !

!SpurPlanningCompactorSimulator methodsFor: 'accessing' stamp: 'eem 2/3/2017 09:41'!
interestingObj: obj
	interestingObj := obj! !

!SpurPlanningCompactorSimulator methodsFor: 'private' stamp: 'eem 2/3/2017 09:44'!
recordMovementOf: o to: toFinger savedFirstFieldPtr: savedFirstFieldPtr
	relocationMap ifNotNil: [:rm| rm at: o put: savedFirstFieldPtr]! !

!SpurPlanningCompactorSimulator methodsFor: 'accessing' stamp: 'eem 2/3/2017 09:42'!
recordMovements
	relocationMap := Dictionary new! !

!SpurPlanningCompactorSimulator methodsFor: 'space management' stamp: 'eem 2/3/2017 10:25'!
selectSavedFirstFieldsSpace
	"Override to make savedFirstFieldsSpace small enough for multi-pass compaction, if desired (for testing)."
	super selectSavedFirstFieldsSpace.
	"testRandomAssortment: et al create 1000 objects, aboput 5% pnned and about 33% reclaimable.  Sp 512 objects should force multi-pass."
	sffsMode == #multiPass ifTrue:
		[savedFirstFieldsSpace limit: savedFirstFieldsSpace start + (512 * manager bytesPerOop).
		 biasForGC := false]! !

SpurCompactor subclass: #SpurSweeper
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManager'!

!SpurSweeper commentStamp: 'cb 4/27/2018 09:43' prior: 0!
SpurSweeper is a sweep-only algorithm, setting the compactor to SpurSweeper effectively changes the fullGC to a mark-sweep non-moving algorithm. 

SpurSweeper is a reference implementation if one wants to evaluate GC performance and compare it to a Mark-Sweep. It's also the only non-moving GC available right now which can be convenient for some experiments. One of the main reason why it was implemented is because advanced compaction algorithm includes a sweep phase (See SelectiveCompactor for example) and SpurSweeper allows to debug the sweep phase separatedly.
!

SpurSweeper subclass: #SpurSelectiveCompactor
	instanceVariableNames: 'segmentToFill lastLilliputianChunk'
	classVariableNames: 'MaxOccupationForCompaction'
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManager'!

!SpurSelectiveCompactor commentStamp: 'cb 10/7/2018 19:54' prior: 0!
SpurSelectiveCompactor compacts memory by selecting the memory segments with the most free space and compacting only those, to limit fragmentation while being really quick to perform. The algorithm is fast mostly because it does not update pointers: they are updated lazily during the next marking phase, so there is no need to read the fields of objects in other memory segments that the one compacted.

The algorithm works as follow. First, a global sweep pass iterates over the memory linearly, changing unmarked objects to free chunks and concatenating free chunks. During the global sweep phase, the segments of the heap are analysed to determine the percentage of occupation. Second, the least occupied segments are compacted by copying the remaining live objects into an entirely free segment, called regionToFill (we detail later in the paragraph where regionToFill comes from), changing their values to forwarding objects and marking the free chunks as unavailable (removed from free list and marked as data objects). Third, the next marking phase removes all forwarders. Fourth, at the beginning of the next compaction phase the compacted segments from the previous GC can be entirely marked as free space (No need to check anything inside, there were only forwarders and trash data). One of the compacted segment is then selected as the segmentToFill, others are just marked as free chunks. 

The compaction is effectively partial, compacting only the most critical segments of the heap to limit fragmentation. Compaction time is crazy low, since a low number of objects are moved and pointer updated is lazily done during the next marking phase, while still preventing memory fragmentation.

Now this works well when biasForGC is true, but when performing a snapshot, the compactor is just total crap (we need to figure out a solution).

segmentToFill <SegInfo> the segment that will be filled through the copying algorithm
lastLilliputianChunk <Oop to FreeChunk> This is used as a performance trick for lilliputian free chunks. See below.

Segment abuse:
The swizzle field of segInfo is abused by using the low 8 bits for occupation and the 9th bit as isBeingCompacted bit.

Performance trick for lilliputian chunks:
Specific free chunks (called lilliputian, see isLilliputianSize:) are managed using a single linked list instead of a double linked list since there's not enough room in the free chunk for the back pointer. During the sweep phase this is not a problem since we're rebuilding the free chunk structure, but during selective compaction we're detaching free chunks from the free chunk structure and that can be horribly slow (10 seconds sometimes at 20Gb heap due to many iteration over the single linked list). To work around this problem, the sweep phase use lastLilliputianChunk variable to sort the lilliputian free chunk single linked list in ascending address order (See interceptAddFreeChunkWithBytes:at:). During the selective compation phase, the same variable is re-used to iterate at most once over the single linked list while detaching lilliputian chunks (See incrementalUnlinkSmallChunk:). In addition, each segment is annotated during the sweep phase with the last lilliputian chunk it holds. Hence, during the compaction phase, the linked list is iterated but the iteration can jump to the last chunk of the previous segment to compact.!

!SpurSelectiveCompactor class methodsFor: 'translation' stamp: 'eem 6/6/2018 20:14'!
declareCVarsIn: aCCodeGenerator
	super declareCVarsIn: aCCodeGenerator.
	aCCodeGenerator var: 'segmentToFill' type: #'SpurSegmentInfo *'! !

!SpurSelectiveCompactor class methodsFor: 'initialization' stamp: 'cb 5/28/2018 10:33'!
initialize
	super initialize.
	"If the segment is occupied by more than MaxOccupationForCompaction, 
	 it's not worth compacting it, whatever the rest of the system looks like.
	 MaxOccupationForCompaction is included in [0;16rFFFF]."
	MaxOccupationForCompaction := 16rD000. "81%"! !

!SpurSelectiveCompactor class methodsFor: 'simulation' stamp: 'cb 5/28/2018 15:44'!
simulatorClass
	^ SpurSelectiveCompactor"Simulator"! !

!SpurSelectiveCompactor methodsFor: 'compaction' stamp: 'cb 4/27/2018 10:55'!
assertNoSegmentBeingCompacted
	"Assertion only - no segment is being claimed at this point"
	| segInfo |
	<var: 'segInfo' type: #'SpurSegmentInfo *'>
	0 to: manager numSegments - 1 do:
		[:i|
		 segInfo := self addressOf: (manager segmentManager segments at: i).
		 self deny: (self isSegmentBeingCompacted: segInfo)].
	! !

!SpurSelectiveCompactor methodsFor: 'api' stamp: 'cb 10/5/2018 12:38'!
compact
	<inline: #never> "for profiling, though we profile selectiveCompaction and sweep separatly."
	self resetFreeLists.
	self freePastSegmentsAndSetSegmentToFill.
	self globalSweepAndSegmentOccupationAnalysis.
	self assert: self sortedLilliputianChunks.
	manager updateSweepEndUsecs.
	self selectiveCompaction.
	! !

!SpurSelectiveCompactor methodsFor: 'compaction' stamp: 'eem 06/22/2021 15:31'!
compactSegment: segInfo freeStart: initialFreeStart segIndex: segIndex
	<var: 'segInfo' type: #'SpurSegmentInfo *'>
	| currentEntity fillStart bytesToCopy bridge copy |
	fillStart := initialFreeStart.
	bridge := manager segmentManager bridgeFor: segInfo.
	currentEntity := manager objectStartingAt: segInfo segStart.
	self deny: segIndex = 0. "Cannot compact seg 0"
	lastLilliputianChunk := self lastLilliputianChunkAtIndex: segIndex - 1.
	[self oop: currentEntity isLessThan: bridge] whileTrue:
		[(manager isFreeObject: currentEntity)
			ifTrue: 
				["To avoid confusing too much Spur (especially the leak/free checks), we mark the free chunk as a word object."
				 (manager isLilliputianSize: (manager bytesInBody: currentEntity)) 
					ifTrue: [self incrementalUnlinkLilliputianChunk: currentEntity] "Performance hack for single linked list"
					ifFalse: [manager detachFreeObject: currentEntity].
				 manager set: currentEntity classIndexTo: manager wordSizeClassIndexPun formatTo: manager wordIndexableFormat]
			ifFalse: 
				["Copy the object in segmentToFill and replace it by a forwarder."
				 self assert: (manager isPinned: currentEntity) not. 
				 bytesToCopy := manager bytesInBody: currentEntity.
				 manager memcpy: fillStart asVoidPointer _: (manager startOfObject: currentEntity) asVoidPointer _: bytesToCopy.
				 copy := manager objectStartingAt: fillStart.
				 (manager isRemembered: copy) ifTrue: 
					["copy has the remembered bit set, but is not in the remembered table."
					 manager setIsRememberedOf: copy to: false.
					 scavenger remember: copy].
				 manager forward: currentEntity to: (manager objectStartingAt: fillStart).
				 fillStart := fillStart + bytesToCopy.
				 self assert: (self oop: fillStart isLessThan: (segmentToFill segLimit - manager bridgeSize))].
		 currentEntity := manager objectAfter: currentEntity limit: manager endOfMemory].
	self assert: currentEntity = bridge.
	^ fillStart! !

!SpurSelectiveCompactor methodsFor: 'compaction' stamp: 'cb 10/7/2018 19:51'!
compactSegmentsToCompact
	"Forwards all objects in segments to compact and removes their freechunks"
	| segInfo fillStart |
	<var: 'segInfo' type: #'SpurSegmentInfo *'>
	fillStart := segmentToFill segStart.
	
	 "Removes initial free chunk in segment to fill... (Segment is entirely free)"
	manager detachFreeObject: (manager objectStartingAt: fillStart).
	
	 "Compact each segment to compact..."
	0 to: manager numSegments - 1 do:
		[:i| 
		 segInfo := self addressOf: (manager segmentManager segments at: i).
		(self isSegmentBeingCompacted: segInfo)
			ifTrue: [fillStart := self compactSegment: segInfo freeStart: fillStart segIndex: i]].
		
	 "Final free chunk in segment to fill..."
	 manager 
		addFreeChunkWithBytes: segmentToFill segSize - manager bridgeSize + segmentToFill segStart - fillStart 
		at: fillStart.
	
	self postCompactionAction
	! !

!SpurSelectiveCompactor methodsFor: 'compaction' stamp: 'cb 11/27/2018 11:39'!
computeSegmentsToCompact
	"Compute segments to compact: least occupied.
	 Answers true if compaction should be done 
	 (at least 1 segment is being compacted and
	 there is a segment to compact into)."
	| canStillClaim aboutToClaim aboutToClaimSegment atLeastOneSegmentToCompact |
	<var: 'aboutToClaimSegment' type: #'SpurSegmentInfo *'>
	atLeastOneSegmentToCompact := false.
	aboutToClaimSegment := self findNextSegmentToCompact.
	"Segment to fill is one of the segment compacted last GC. 
	 If no segment were compacted last GC, and that there is 
	 at least one segment to compact, allocate a new one."
	aboutToClaimSegment ifNil: [^false].
	segmentToFill ifNil:
		[self findOrAllocateSegmentToFill.
		 segmentToFill ifNil: ["Abort compaction"^false]].
	canStillClaim := segmentToFill segSize - manager bridgeSize.
	[aboutToClaimSegment ifNil: [^atLeastOneSegmentToCompact].
	 aboutToClaim := self sizeClaimedIn: aboutToClaimSegment.
	 aboutToClaim < canStillClaim ] whileTrue: 
		[self markSegmentAsBeingCompacted: aboutToClaimSegment.
		 atLeastOneSegmentToCompact := true.
		 canStillClaim := canStillClaim - aboutToClaim.
		 aboutToClaimSegment := self findNextSegmentToCompact].
	^atLeastOneSegmentToCompact! !

!SpurSelectiveCompactor methodsFor: 'segment to fill' stamp: 'cb 10/8/2018 16:06'!
findAndSetSegmentToFill
	| segInfo firstEntity |
	<var: 'segInfo' type: #'SpurSegmentInfo *'>
	0 to: manager numSegments - 1 do:
		[:i| 
		 segInfo := self addressOf: (manager segmentManager segments at: i).
		 firstEntity := manager objectStartingAt: segInfo segStart.
		 ((manager isFreeObject: firstEntity) and: [(manager objectAfter: firstEntity limit: manager endOfMemory) = (manager segmentManager bridgeFor: segInfo)])
			ifTrue: [segmentToFill := segInfo. ^i]].
	^-1
	! !

!SpurSelectiveCompactor methodsFor: 'compaction' stamp: 'eem 6/6/2018 19:21'!
findNextSegmentToCompact
	"Answers the next segment to compact or nil if none.
	  The next segment to compact:
	 - cannot be segment 0 (Segment 0 has specific objects 
	  (nil, true, etc.) and special size computed at start-up 
	  that we don't want to deal with)
	 - cannot have a high occupation rate (> MaxOccupationForCompaction)"
	| leastOccupied leastOccupiedSegment tempOccupied segInfo |
	<var: 'segInfo' type: #'SpurSegmentInfo *'>
	leastOccupied := 16rFFFF.
	1 to: manager numSegments - 1 do:
		[:i|
		 segInfo := self addressOf: (manager segmentManager segments at: i).
		 ((self isSegmentBeingCompacted: segInfo) or: [segInfo containsPinned or: [manager segmentManager isEmptySegment: segInfo] ])
			ifFalse: 
				[(tempOccupied := self occupationOf: segInfo) <= leastOccupied
					ifTrue: [ leastOccupied := tempOccupied.
							 leastOccupiedSegment := segInfo ]]].
	leastOccupied > MaxOccupationForCompaction ifTrue:
		[^self cCoerceSimple: nil to: #'SpurSegmentInfo *'].
	^leastOccupiedSegment! !

!SpurSelectiveCompactor methodsFor: 'segment to fill' stamp: 'cb 11/27/2018 11:36'!
findOrAllocateSegmentToFill
	"There was no compacted segments from past GC that we can directly re-use.
	 We need either to find an empty segment or allocate a new one."
	| segIndex |
	self findAndSetSegmentToFill.
	segmentToFill ifNotNil: [^0].
	"No empty segment. We need to allocate a new one"
	(manager growOldSpaceByAtLeast: manager growHeadroom) ifNil: ["failed to allocate"^0].
	"We don't know which segment it is that we've just allocated... So we look for it... This is a bit dumb."
	segIndex := self findAndSetSegmentToFill.
	"Lilliputian performance hack management... Last lilliputian of new segment is same as prev because no lilliputian in new segment"
	self setLastLilliputianChunkAtindex: segIndex to: (self lastLilliputianChunkAtIndex: segIndex - 1).
	self assert: segmentToFill ~~ nil.
	
	! !

!SpurSelectiveCompactor methodsFor: 'segment access' stamp: 'cb 6/12/2018 07:37'!
freePastSegmentsAndSetSegmentToFill	
	"The first segment being claimed met becomes the segmentToFill. The others are just freed."
	| segInfo |
	<var: 'segInfo' type: #'SpurSegmentInfo *'>
	segmentToFill := nil.
	0 to: manager numSegments - 1 do:
		[:i|
		 segInfo := self addressOf: (manager segmentManager segments at: i).
		 (self isSegmentBeingCompacted: segInfo)
			ifTrue: 
				[manager 
					initFreeChunkWithBytes: segInfo segSize - manager bridgeSize 
					at: segInfo segStart.
				 segmentToFill ifNil: [segmentToFill := segInfo]]]! !

!SpurSelectiveCompactor methodsFor: 'sweep phase' stamp: 'eem 06/22/2021 15:31'!
globalSweepAndSegmentOccupationAnalysis
	<inline: #never> "profiling"
	"Iterate over old space, free unmarked objects, annotate each segment with each occupation"
	| currentEntity nextBridge segmentIndex currentUsed currentUnused |
	lastLilliputianChunk := 0. "performance hack for single linked list"
	currentEntity := manager firstObject.
	nextBridge := manager segmentManager bridgeAt: 0.
	segmentIndex := currentUnused := currentUsed := 0.
	self setLastLilliputianChunkAtindex: 0 to: 0.
	[self oop: currentEntity isLessThan: manager endOfMemory] whileTrue:
		[currentEntity = nextBridge "End of segment, set occupation"
			ifTrue: 
				[self 
					setOccupationAtIndex: segmentIndex
					used: currentUsed 
					unused: currentUnused.
				  self setLastLilliputianChunkAtindex: segmentIndex to: lastLilliputianChunk.
				  currentUnused := currentUsed := 0.
				  segmentIndex := segmentIndex + 1.
				  nextBridge := manager segmentManager bridgeAt: segmentIndex]
			ifFalse: 
				[(self canUseAsFreeSpace: currentEntity) "In-segment, sweep and compute occupation"
					ifTrue: 
						[currentEntity := self bulkFreeChunkFrom: currentEntity.
						 currentUnused := currentUnused + (manager bytesInBody: currentEntity)]
					ifFalse: 
						[self unmark: currentEntity.
						 currentUsed := currentUsed + (manager bytesInBody: currentEntity)]].
		 currentEntity := manager objectAfter: currentEntity limit: manager endOfMemory].
	"set last segment details"	
	self 
		setOccupationAtIndex: segmentIndex
		used: currentUsed 
		unused: currentUnused.
	self setLastLilliputianChunkAtindex: segmentIndex to: lastLilliputianChunk.
	"we set the nextFreeChunk of last chunk at the end of the loop to avoid to set it at each iteration"
	lastLilliputianChunk ~= 0 ifTrue:
	 	[manager setNextFreeChunkOf: lastLilliputianChunk withValue: 0 isLilliputianSize: true].
		
	manager checkFreeSpace: GCModeFull.
	manager unmarkSurvivingObjectsForCompact! !

!SpurSelectiveCompactor methodsFor: 'compaction' stamp: 'eem 06/22/2021 15:31'!
incrementalLilliputianSmallChunk: freeChunk
	"This is duplicate form #unlinkSmallChunk:index: for performance hack (single iteration of single linked list)"
	<inline: #never> "for profiling"
	| node next |
	self assert: (manager bytesInBody: freeChunk) = (manager baseHeaderSize + manager allocationUnit).
	self assert: manager lilliputianChunkIndex = ((manager bytesInBody: freeChunk) / manager allocationUnit).
	lastLilliputianChunk = 0 ifTrue: "first incremental unlink"
		[(freeChunk = manager firstLilliputianChunk) 
			ifTrue: [^manager unlinkFreeChunk: freeChunk atIndex: manager lilliputianChunkIndex isLilliputianSize: true]
			ifFalse: [lastLilliputianChunk := manager firstLilliputianChunk]].
	 node := manager fetchPointer: manager freeChunkNextIndex ofFreeChunk: lastLilliputianChunk.
	 [node ~= 0] whileTrue:
		[self assert: node = (manager startOfObject: node).
		 manager assertValidFreeObject: node.
		 next := manager fetchPointer: manager freeChunkNextIndex ofFreeChunk: node.
		 node = freeChunk ifTrue:
			[^manager setNextFreeChunkOf: lastLilliputianChunk withValue: next isLilliputianSize: true].
		 lastLilliputianChunk := node.
		 node := next].
	 self error: 'freeChunk not found in lilliputian chunk free list'

	
! !

!SpurSelectiveCompactor methodsFor: 'sweep phase' stamp: 'cb 10/5/2018 12:57'!
interceptAddFreeChunkWithBytes: bytes at: start
	<inline: true>
	| freeChunk |
	(manager isLilliputianSize: bytes) ifTrue: "build size 1 free chunk in ascending addresses order"
		[lastLilliputianChunk = 0
			ifTrue: [^lastLilliputianChunk := manager addFreeChunkWithBytes: bytes at: start]
			ifFalse: 
				[manager increaseFreeOldSpaceBy: bytes.
				 freeChunk := manager initFreeChunkWithBytes: bytes at: start.
				 manager setNextFreeChunkOf: lastLilliputianChunk withValue: freeChunk isLilliputianSize: true.
				 "we set the nextFreeChunk of last chunk at the end of the loop to avoid to set it at each iteration"
				 ^lastLilliputianChunk := freeChunk]].
	^manager addFreeChunkWithBytes: bytes at: start
! !

!SpurSelectiveCompactor methodsFor: 'segment access' stamp: 'cb 4/27/2018 10:51'!
isSegmentBeingCompacted: segInfo 
	<var: 'segInfo' type: #'SpurSegmentInfo *'>
	"Swizzle is abused bit 16 isBeingCompacted bits 0-15 occupation"
	^ segInfo swizzle anyMask: 1 << 16! !

!SpurSelectiveCompactor methodsFor: 'segment access' stamp: 'cb 10/7/2018 19:50'!
lastLilliputianChunkAtIndex: segIndex
	<inline: true>
	"Abuse lastFreeObject field, can be used during compaction only, used for different purpose during snapshot"
	^(self addressOf: (manager segmentManager segments at: 0)) lastFreeObject! !

!SpurSelectiveCompactor methodsFor: 'segment access' stamp: 'cb 4/27/2018 10:51'!
markSegmentAsBeingCompacted: segInfo 
	<var: 'segInfo' type: #'SpurSegmentInfo *'>
	"Swizzle is abused bit 16 isBeingCompacted bits 0-15 occupation"
	segInfo swizzle: (segInfo swizzle bitOr: 1 << 16)! !

!SpurSelectiveCompactor methodsFor: 'segment access' stamp: 'cb 4/27/2018 10:51'!
occupationOf: segInfo 
	<var: 'segInfo' type: #'SpurSegmentInfo *'>
	"Swizzle is abused bit 16 isBeingCompacted bits 0-15 occupation"
	^segInfo swizzle bitAnd: 16rFFFF! !

!SpurSelectiveCompactor methodsFor: 'compaction' stamp: 'cb 10/4/2018 09:33'!
postCompactionAction
	| allFlags |
	"For now we don't optimize and just follow everything everywhere on stack and in caches, let's see in the profiler if we need to optimize with those cases. My guess is that this is < 100 microSecond"
	manager followSpecialObjectsOop.
	allFlags := BecamePointerObjectFlag + BecameActiveClassFlag bitOr: BecameCompiledMethodFlag.
	"Note: there is not the OldBecameNewFlag"
	"gcMode flag is cleared after postBecomeAction, reset it."
	manager coInterpreter postBecomeAction: allFlags.
	manager coInterpreter setGCMode: GCModeFull.
	
	"Special to selective, crazy objects can be forwarded..."
	"manager postBecomeScanClassTable: allFlags. => Done in followClassTable"
	manager followClassTable.
	manager followProcessList.
	manager followForwardedObjStacks.
	
	"Not sure the following are needed...
	coInterpreter mapInterpreterOops.
	manager mapExtraRoots."
	self assert: manager validClassTableHashes.! !

!SpurSelectiveCompactor methodsFor: 'api' stamp: 'eem 6/6/2018 19:57'!
postSwizzleAction
	"Since the compact abuses the swizzle field of segment, it needs to be reset after start-up."
	| segInfo |
	<var: 'segInfo' type: #'SpurSegmentInfo *'>
	0 to: manager numSegments - 1 do:
		[:i|
		 segInfo := self addressOf: (manager segmentManager segments at: i).
		 segInfo swizzle: 0 ]! !

!SpurSelectiveCompactor methodsFor: 'compaction' stamp: 'cb 11/27/2018 11:38'!
selectiveCompaction
	"Figures out which segments to compact and compact them into segmentToFill"
	| shouldCompact |
	<inline: #never> "profiling"
	self assertNoSegmentBeingCompacted.
	"Should compact only if there is at least 1 segment to compact 
	 and there is a segment to compact into"
	shouldCompact := self computeSegmentsToCompact.
	"If no compaction we don't pay forwarding cost (stack scan, cache scan, etc.)
	 and we don't allocate segmentToFill if none available."
	shouldCompact ifTrue:
		[self assert: segmentToFill ~~ nil.
		 self compactSegmentsToCompact].
	manager checkFreeSpace: GCModeFull.! !

!SpurSelectiveCompactor methodsFor: 'segment access' stamp: 'cb 10/7/2018 19:48'!
setLastLilliputianChunkAtindex: segIndex to: chunk 
	<inline: true>
	"Abuse lastFreeObject field, can be used during compaction only, used for different purpose during snapshot"
	(self addressOf: (manager segmentManager segments at: 0)) lastFreeObject: chunk! !

!SpurSelectiveCompactor methodsFor: 'segment access' stamp: 'cb 5/28/2018 10:11'!
setOccupationAtIndex: segmentIndex used: used unused: unused
	"WARNING: Resets the isCompacted bit"
	"Swizzle is abused bit 16 isBeingCompacted bits 0-15 occupation
	 Setting occupation resets the claim bit"
	| occupation segInfo |
	<var: 'segInfo' type: #'SpurSegmentInfo *'>
	segInfo := self addressOf: (manager segmentManager segments at: segmentIndex).
	"careful with overflow here..."
	occupation := ((used asFloat / (used + unused)) * 16rFFFF) asInteger.
	self assert: (occupation between: 0 and: 16rFFFF).
	segInfo swizzle: occupation! !

!SpurSelectiveCompactor methodsFor: 'segment access' stamp: 'cb 5/28/2018 10:14'!
sizeClaimedIn: segment 
	<var: 'segment' type: #'SpurSegmentInfo *'>
	<var: 'ratio' type: #'double'>
	"careful with overflow here"
	"roundedup used ratio (+1 to round up)"
	| ratio |
	ratio := ((self occupationOf: segment) + 1) asFloat / 16rFFFF.
	^(ratio * (segment segSize - manager bridgeSize)) asInteger ! !

!SpurSelectiveCompactor methodsFor: 'sweep phase' stamp: 'cb 10/5/2018 12:42'!
sortedLilliputianChunks
	|current next|
	current := manager firstLilliputianChunk.
	current = 0 ifTrue: [^true]. "no node"
	[next := manager fetchPointer: manager freeChunkNextIndex ofFreeChunk: current.
	 next = 0] whileFalse:
		[(manager oop: current isLessThan: next) ifFalse: [^false].
		 current := next].
	^ true
! !

!SpurSelectiveCompactor methodsFor: 'segment access' stamp: 'cb 6/11/2018 10:01'!
unmarkSegmentAsBeingCompacted: segInfo 
	<inline: true> "So it is not generated if unused"
	<var: 'segInfo' type: #'SpurSegmentInfo *'>
	"Swizzle is abused bit 16 isBeingCompacted bits 0-15 occupation"
	segInfo swizzle: (segInfo swizzle bitAnd: 16rFFFF)! !

SpurSelectiveCompactor subclass: #SpurSelectiveCompactorSimulator
	instanceVariableNames: ''
	classVariableNames: 'Talking'
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManagerSimulation'!

!SpurSelectiveCompactorSimulator class methodsFor: 'initialization' stamp: 'cb 4/19/2018 08:53'!
initialize
	super initialize.
	Talking := true.! !

!SpurSelectiveCompactorSimulator methodsFor: 'debugging' stamp: 'cb 4/26/2018 08:53'!
checkOnlyForwardersOrWordNonImm: segInfo
	"Also check this is iterable until bridge 
	 prevEntity is here for debugging only"
	| currentEntity prevEntity bridge |
	self talk: 'Checking segment compacted has fwd/Word only ' , (manager segmentManager indexOfSegment: segInfo) printString.
	bridge := manager segmentManager bridgeFor: segInfo.
	currentEntity := manager objectStartingAt: segInfo segStart.
	[self oop: currentEntity isLessThan: bridge] whileTrue: 
		[self assert: ((manager isWordsNonImm: currentEntity) or: [manager isForwarded: currentEntity]).
		 prevEntity := currentEntity.
		 currentEntity := manager objectAfter: currentEntity limit: manager endOfMemory].
	self assert: currentEntity = bridge.
	self talkNoCr: ' OK'.! !

!SpurSelectiveCompactorSimulator methodsFor: 'debugging' stamp: 'cb 6/11/2018 07:33'!
checkSegmentToFillLooksAllRight
	"Check only 1 free object at the end or none.
	 Check segment is iterable until bridge"
	| currentEntity bridge |
	self talk: 'Checking segment to fill iterable/last is free ' , (manager segmentManager indexOfSegment: segmentToFill) printString.
	bridge := manager segmentManager bridgeFor: segmentToFill.
	currentEntity := manager objectStartingAt: segmentToFill segStart.
	[self oop: currentEntity isLessThan: bridge] whileTrue: 
		[(manager isFreeObject: currentEntity) 
			ifTrue: "should be last entity"
				[self assert: (manager objectAfter: currentEntity limit: manager endOfMemory) = bridge.
				 manager assertValidFreeObject: currentEntity].
		 currentEntity := manager objectAfter: currentEntity limit: manager endOfMemory].
	"End of iteration should end on bridge"
	self assert: currentEntity = bridge.
	self talkNoCr: ' OK'.! !

!SpurSelectiveCompactorSimulator methodsFor: 'debugging' stamp: 'cb 4/19/2018 14:39'!
checkSegmentsIterable
	"Check only 1 free object at the end or none.
	 Check segment is iterable until bridge"
	| currentEntity bridge |
	self talk: 'Checking segments iterable'.
	0 to: manager numSegments - 1 do:
		[:i| | segInfo |
		 segInfo := self addressOf: (manager segmentManager segments at: i).
		 bridge := manager segmentManager bridgeFor: segInfo.
		 currentEntity := manager objectStartingAt: segInfo segStart.
		 [self oop: currentEntity isLessThan: bridge] whileTrue: 
			[currentEntity := manager objectAfter: currentEntity limit: manager endOfMemory].
		 "End of iteration should end on bridge"
		 self assert: currentEntity = bridge.].
	self talkNoCr: ' OK'.! !

!SpurSelectiveCompactorSimulator methodsFor: 'api' stamp: 'eem 6/21/2021 12:04'!
compact
	self talk: '~~ Start selective compaction ~~'.
	self tellMeNumSegmentAndTotalFreeSpace.
	self tellMeWhichSegmentsAreBeingCompacted.
	super compact.
	self tellMeNumSegmentAndTotalFreeSpace.
	self tellMeWhichSegmentsAreBeingCompacted.
	manager runLeakCheckerForFreeSpace: GCCheckFreeSpace ignoring: nil.
	self talk: '~~ Finished selective compaction ~~'.
	! !

!SpurSelectiveCompactorSimulator methodsFor: 'compaction' stamp: 'cb 4/19/2018 16:55'!
compactSegmentsToCompact
	self talk: 'filling segment ' , (manager segmentManager indexOfSegment: segmentToFill).
	super compactSegmentsToCompact.
	"Now segments to compacts should have been compacted into segment to fill"
	self checkSegmentToFillLooksAllRight.
	! !

!SpurSelectiveCompactorSimulator methodsFor: 'compaction' stamp: 'cb 4/19/2018 14:37'!
computeSegmentsToCompact
	| res |
	res := super computeSegmentsToCompact.
	res	
		ifTrue: [self talk: 'There is something to compact!!'.
				self tellMeWhichSegmentsAreBeingCompacted]
		ifFalse: [self talk: 'Nothing to compact!!'].
	^ res! !

!SpurSelectiveCompactorSimulator methodsFor: 'freeing' stamp: 'cb 4/19/2018 11:23'!
freePastSegmentsAndSetSegmentToFill
	super freePastSegmentsAndSetSegmentToFill.
	segmentToFill
		ifNil: [self talk: 'No segment to fill so far - will allocate one later if anything to compact']
		ifNotNil: [self talk: 'filling segment ' , (manager segmentManager indexOfSegment: segmentToFill)].
	! !

!SpurSelectiveCompactorSimulator methodsFor: 'sweep phase' stamp: 'cb 5/28/2018 09:03'!
globalSweepAndSegmentOccupationAnalysis
	super globalSweepAndSegmentOccupationAnalysis.
	self talk: 'Occupation map: '.
	0 to: manager numSegments - 1 do:
		[:i| | segInfo |
		 segInfo := self addressOf: (manager segmentManager segments at: i).
		 self talkNoCr: i.
		 self talkNoCr: '->'.
		 self talkNoCr: (self occupationOf: segInfo) printString.
		 self talkNoCr: '('.
		 self talkNoCr: ((self occupationOf: segInfo) * 100 * 100 // 16rFFFF / 100) asFloat printString.
		 self talkNoCr: '%);'].
	self checkSegmentsIterable.
	! !

!SpurSelectiveCompactorSimulator methodsFor: 'compaction' stamp: 'cb 4/19/2018 13:31'!
markSegmentAsBeingCompacted: segInfo 
	super markSegmentAsBeingCompacted: segInfo.
	self talk: 'Marking as being compacted segment ' , (manager segmentManager indexOfSegment: segInfo) printString.! !

!SpurSelectiveCompactorSimulator methodsFor: 'compaction' stamp: 'cb 4/26/2018 11:02'!
selectiveCompaction
	super selectiveCompaction.
	manager allFreeObjectsDo: [:objOop | manager assertValidFreeObject: objOop]! !

!SpurSelectiveCompactorSimulator methodsFor: 'debugging' stamp: 'cb 4/19/2018 08:52'!
talk: string
	Talking ifTrue: [Transcript cr; show: string]! !

!SpurSelectiveCompactorSimulator methodsFor: 'debugging' stamp: 'cb 4/19/2018 09:18'!
talkNoCr: string
	Talking ifTrue: [Transcript show: string]! !

!SpurSelectiveCompactorSimulator methodsFor: 'debugging' stamp: 'cb 4/19/2018 11:33'!
tellMeNumSegmentAndTotalFreeSpace
	self talk: 'numSegments: ' , manager numSegments printString , ', free space: ' , manager totalFreeOldSpace printString.! !

!SpurSelectiveCompactorSimulator methodsFor: 'debugging' stamp: 'cb 4/19/2018 09:14'!
tellMeWhichSegmentsAreBeingCompacted
	<doNotGenerate> 
	| beingCompacted |
	beingCompacted := OrderedCollection new.
	0 to: manager numSegments - 1 do:
		[:i| | segInfo |
		 segInfo := self addressOf: (manager segmentManager segments at: i).
		(self isSegmentBeingCompacted: segInfo) ifTrue: [beingCompacted add: segInfo]].
	self 
		talk: 'Segment being compacted: ' , 
			(beingCompacted collect: [:seg | 
				manager segmentManager indexOfSegment: seg ]) asArray printString
	! !

!SpurSweeper methodsFor: 'api' stamp: 'eem 6/6/2018 20:23'!
biasForGC! !

!SpurSweeper methodsFor: 'api' stamp: 'eem 6/6/2018 20:23'!
biasForSnapshot! !

!SpurSweeper methodsFor: 'sweep phase' stamp: 'eem 06/22/2021 15:31'!
bulkFreeChunkFrom: objOop
	"The old space entity before objOop is necessarily a marked object. Attempts to free as many bytes 
	from objOop start as possible, looking ahead to free contiguous freechunks / unmarked objects"
	| bytes start next currentObj |
	self assert: ((manager isMarked: objOop) not or: [manager isFreeObject: objOop]).
	start := manager startOfObject: objOop.
	currentObj := objOop.
	bytes := 0.
	[bytes := bytes + (manager bytesInBody: currentObj).
	(manager isRemembered: currentObj)
		ifTrue: 
			[self assert: (manager isFreeObject: currentObj) not.
			 scavenger forgetObject: currentObj].
	next := manager objectStartingAt: start + bytes.
	self assert: ((manager oop: next isLessThan: manager endOfMemory)
		or: [next = manager endOfMemory and: [(self canUseAsFreeSpace: next) not]]).
	self canUseAsFreeSpace: next] 
		whileTrue: [currentObj := next].
	
	^self interceptAddFreeChunkWithBytes: bytes at: start! !

!SpurSweeper methodsFor: 'sweep phase' stamp: 'cb 10/1/2017 11:45'!
canUseAsFreeSpace: objOop
	<inline: true>
	^ (manager isFreeObject: objOop) or: [(manager isMarked: objOop) not]! !

!SpurSweeper methodsFor: 'api' stamp: 'cb 7/3/2018 09:18'!
compact
	<inline: #never> "for profiling"
	self resetFreeLists.
	self globalSweep.
	manager updateSweepEndUsecs! !

!SpurSweeper methodsFor: 'sweep phase' stamp: 'cb 6/12/2018 07:22'!
globalSweep
	"Iterate over all entities, in order, making large free chunks from free chunks and unmarked objects, 
	unmarking live objects and rebuilding the free lists."

	| currentEntity |
	currentEntity := manager firstObject.
	[self oop: currentEntity isLessThan: manager endOfMemory] whileTrue:
		[(self canUseAsFreeSpace: currentEntity) 
			ifTrue: [currentEntity := self bulkFreeChunkFrom: currentEntity]
			ifFalse: [self unmark: currentEntity].
		 currentEntity := manager objectAfter: currentEntity limit: manager endOfMemory].
			
	manager checkFreeSpace: GCModeFull.
	manager unmarkSurvivingObjectsForCompact.

	! !

!SpurSweeper methodsFor: 'sweep phase' stamp: 'cb 10/5/2018 12:42'!
interceptAddFreeChunkWithBytes: bytes at: start
	<inline: true>
	^manager addFreeChunkWithBytes: bytes at: start
! !

!SpurSweeper methodsFor: 'api' stamp: 'cb 6/12/2018 07:28'!
resetFreeLists
	manager resetFreeListHeads.
	manager totalFreeOldSpace: 0.! !

!SpurSweeper methodsFor: 'sweep phase' stamp: 'cb 10/3/2017 10:52'!
unmark: objOop
	self assert: ((manager isMarked: objOop) and: [(manager isFreeObject: objOop) not]).
	(manager isSegmentBridge: objOop) ifFalse: [manager setIsMarkedOf: objOop to: false].
	(manager isPinned: objOop) ifTrue: [manager segmentManager notePinned: objOop]! !

CogClass subclass: #SpurGenerationScavenger
	instanceVariableNames: 'coInterpreter manager eden futureSpace pastSpace futureSurvivorStart rememberedSet rememberedSetSize previousRememberedSetSize rememberedSetRedZone rememberedSetLimit refCountToShrinkRT weakList ephemeronList tenureCriterion tenureThreshold tenuringClassIndex tenuringProportion numRememberedEphemerons scavengeLog scavengeLogRecord statSurvivorCount statTenures'
	classVariableNames: ''
	poolDictionaries: 'SpurMemoryManagementConstants'
	category: 'VMMaker-SpurMemoryManager'!

!SpurGenerationScavenger commentStamp: 'eem 11/7/2017 17:26' prior: 0!
SpurGenerationScavenger is an implementation of David Ungar's Generation Scavenging garbage collection algorithm.  See
	Generation Scavenging, A Non-disruptive, High-Performance Storage Reclamation Algorithm
	David Ungar
	Proceeding
	SDE 1 Proceedings of the first ACM SIGSOFT/SIGPLAN software engineering symposium on Practical software development environments
	Pages 157 - 167 
	ACM New York, NY, USA ©1984 

Also relevant are
	An adaptive tenuring policy for generation scavengers
	David Ungar & Frank Jackson
	ACM Transactions on Programming Languages and Systems (TOPLAS) TOPLAS Homepage archive
	Volume 14 Issue 1, Jan. 1992 
	Pages 1 - 27 
	ACM New York, NY, USA ©1992
and
	Ephemerons: a new finalization mechanism
	Barry Hayes
	Proceedings of the 12th ACM SIGPLAN conference on Object-oriented programming, systems, languages, and applications
	Pages 176-183 
	ACM New York, NY, USA ©1997

See text below the variable definitions and explanation below for a full explanation of weak and ephemeron processing.

Instance Variables
	coInterpreter:					<StackInterpreterSimulator|CogVMSimulator>
	eden:							<SpurNewSpaceSpace>
	ephemeronList:					<Integer|nil>
	futureSpace:					<SpurNewSpaceSpace>
	futureSurvivorStart:				<Integer address>
	manager:						<SpurMemoryManager|Spur32BitMMLESimulator et al>
	numRememberedEphemerons:	<Integer>
	pastSpace:						<SpurNewSpaceSpace>
	previousRememberedSetSize:	<Integer>
	rememberedSet:				<CArrayAccessor on: Array>
	rememberedSetSize:			<Integer>
	tenuringProportion:				<Float>
	tenuringThreshold:				<Integer address>
	weakList:						<Integer|nil>

coInterpreter
	- the interpreter/vm, in this context, the mutator

manager
	- the Spur memory manager

eden
	- the space containing newly created objects

futureSpace
	- the space to which surviving objects are copied during a scavenge

futureSurvivorStart
	- the allocation pointer into futureSpace

pastSpace
	- the space surviving objects live in until the next scavenge

rememberedSet
	- the root old space objects that refer to objects in new space; a scavenge starts form these roots and the interpreter's stack

rememberedSetSize
	- the size of the remembered set, also the first unused index in the rememberedSet

previousRememberedSetSize:
	- the size of the remembered set before scavenging objects in future space.

numRememberedEphemerons
	- the number of unscavenged ephemerons at the front of the rememberedSet.

ephemeronList
	- the head of the list of corpses of unscavenged ephemerons reached in the current phase

weakList
	- the head of the list of corpses of weak arrays reached during the scavenge.

tenuringProportion
	- the amount of pastSpace below which the system will not tenure unless futureSpace fills up, and above which it will eagerly tenure

tenuringThreshold
	- the pointer into pastSpace below which objects will be tenured

Weakness and Ephemerality in the Scavenger.
Weak arrays should not hold onto their referents (except from their strong fileds, their named inst vars).  Ephemerons are objects that implement instance-based finalization; attaching an ephemeron to an object keeps that object alive and causes the ephemeron to "fire" when the object is only reachable from the ephemeron (or other ephemerons & weak arrays).  They are a special kind of Associations that detect when their keys are about to die, i.e. when an ephemeron's key is not reachable from the roots except from weak arrays and other ephemerons with about-to-die keys.  Note that if an ephemeron's key is not about to die then references from the rest of the ephemeron can indeed prevent ephemeron keys from dying.

The scavenger is concerned with collecting objects in new space, therefore it ony deals with weak arrays and ephemerons that are either in the remembered set or in new space.  By deferring scanning these objects until other reachable objects have been scavenged, the scavenger can detect dead or dying references.

Weak Array Processing
In the case of weak arrays this is simple.  The scavenger refuses to scavenge the referents of weak arrays in scavengeReferentsOf: until the entire scavenge is over.  It then scans the weak arrays in the remembered set and in future space and nils all fields in them that are referring to unforwarded objects in eden and past space, because these objects have not survived the scavenge.  The root weak arrays remaining to be scavenged are in the remembered table.  Surviving weak arrays in future space are collected on a list.  The list is threaded through the corpses of weak arrays in eden and/or past space.  weakList holds the slot offset of the first weak array found in eden and/or past space.  The next offset is stored in the weak array corpse's identityHash and format fields (22 bits & 5 bits of allocationUnits, for a max new space size of 2^28 bytes, 256Mb).  The list is threaded throguh corpses, but the surviving arrays are pointed to by the corpses' forwarding pointers.

Ephemeron Processing
The case of ephemerons is a little more complicated because an ephemeron's key should survive.  The scavenger is cyclical.  It scavenges the remembered set, which may copy and forward surviving objects in past and/or eden spaces to future space.  It then scavenges those promoted objects in future space until no more are promoted, which may in turn remember more objects.  The cycles continue until no more objects get promoted to future space and no more objects get remembered.  At this point all surviving objecta are in futureSpace.

So if the scavenger does not scan ephemerons in the remembered set or in future space until the scavenger finishes cycling, it can detect ephemerons whose keys are about to die because these will be unforwarded objects in eden and/or past space.  Ephemerons encountered in the remembered set are either processed like ordinary objects if their keys have been promoted to futureSpace, or are moved to the front of the rememberedSet (because, dear reader, it is a sequence) if their keys have not been promoted.  Ephemerons encountered in scavengeReferentsOf: are either scanned like normal objects if their keys have been promoted, or added to the ephemeronList, organized identically to the weakList, if their keys are yet to be promoted.  Since references from other ephemerons with surviving keys to ephemeron keys can and should prevent the ephemerons whose keys they are from firing the scavenger does not fire ephemerons unless all unscavenged ephemerons have unscavenged keys.  So the unscavenged ephemerons (they will be at the beginning of the remembered set and on the ephemeronList) are scanned and any that have promoted keys are scavenged.  But if no unscavenged ephemerons have surviving keys then all the unscavenged ephemerons are fired and then scavenged.  This in turn may remember more objects and promote more objects to future space, and encounter more unscavenged ephemerons.  So the scavenger continues until no more objects are remembered, no more objects are promoted to future space and no more unscavenged ephemerons exist.!

!SpurGenerationScavenger class methodsFor: 'translation' stamp: 'eem 11/13/2017 14:37'!
declareCVarsIn: aCCodeGenerator
	#(eden futureSpace pastSpace) do:
		[:var| aCCodeGenerator var: var type: #SpurNewSpaceSpace].
	aCCodeGenerator
		var: #rememberedSet type: #'sqInt *';
		var: #tenuringProportion type: #double;
		var: #scavengeLogRecord type: #SpurScavengeLogRecord;
		var: #scavengeLog type: #'FILE *'! !

!SpurGenerationScavenger class methodsFor: 'translation' stamp: 'eem 10/29/2013 23:43'!
implicitReturnTypeFor: aSelector
	"Answer the return type for methods that don't have an explicit return."
	^#void! !

!SpurGenerationScavenger class methodsFor: 'class initialization' stamp: 'eem 5/5/2015 17:37'!
initialize
	"SpurGenerationScavenger initialize"
	TenureByAge := 1.
	TenureByClass := 2.
	TenureToShrinkRT := 3.
	DontTenure := 4.
	MarkOnTenure := 5.
	MaxRTRefCount := 7 "The field comprised of {isGrey,isPinned,isRemembered}"! !

!SpurGenerationScavenger class methodsFor: 'translation' stamp: 'eem 10/13/2013 18:46'!
isNonArgumentImplicitReceiverVariableName: instVarName
	^#('self' 'coInterpreter' 'manager') includes: instVarName! !

!SpurGenerationScavenger class methodsFor: 'simulation' stamp: 'eem 2/3/2017 09:37'!
simulatorClass
	^SpurGenerationScavengerSimulator! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 3/25/2015 10:53'!
addToEphemeronList: ephemeronCorpse
	"ephemeronCorpse is the corpse of an ephemeron that was copied and forwarded.
	 Later on its surviving copy must be scanned to nil weak references.
	 Thread the corpse onto the weakList.  Later, the weakList can be followed, and
	 the forwarding pointer followed to locate the survivor."
	<inline: #never> "Should be too infrequent to lower icache density of copyAndForward:"
	| ephemeronListOffset |
	self assert: (self isScavengeSurvivor: (manager keyOfEphemeron: (manager followForwarded: ephemeronCorpse))) not.

	ephemeronListOffset := ephemeronList ifNil: 0.
	self setCorpseOffsetOf: ephemeronCorpse to: ephemeronListOffset.
	ephemeronList := self corpseOffsetOf: ephemeronCorpse.
	self assert: (self firstCorpse: ephemeronList) = ephemeronCorpse! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 3/25/2015 10:53'!
addToWeakList: weakCorpse
	"weakCorpse is the corpse of a weak array that was copied and forwarded.
	 Later on its surviving copy must be scanned to nil weak references.
	 Thread the corpse onto the weakList.  Later, the weakList can be followed, and
	 the forwarding pointer followed to locate the survivor."
	<inline: #never> "Should be too infrequent to lower icache density of copyAndForward:"
	| weakListOffset |

	weakListOffset := weakList ifNil: 0.
	self setCorpseOffsetOf: weakCorpse to: weakListOffset.
	weakList := self corpseOffsetOf: weakCorpse.
	self assert: (self firstCorpse: weakList) = weakCorpse! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 5/23/2014 16:15'!
allFutureSpaceEntitiesDo: aBlock
	"Enumerate all future space objects, including free objects."
	<inline: true>
	| prevObj prevPrevObj objOop limit |
	prevPrevObj := prevObj := nil.
	objOop := manager objectStartingAt: futureSpace start.
	limit := futureSurvivorStart.
	[self oop: objOop isLessThan: limit] whileTrue:
		[aBlock value: objOop.
		 prevPrevObj := prevObj.
		 prevObj := objOop.
		 objOop := manager objectAfter: objOop limit: limit]! !

!SpurGenerationScavenger methodsFor: 'remembered set' stamp: 'eem 5/5/2015 17:54'!
allNewSpaceObjectsHaveZeroRTRefCount
	manager allNewSpaceObjectsDo:
		[:obj|
		(manager rtRefCountOf: obj) > 0 ifTrue:
			[^false]].
	^true! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 5/23/2014 16:16'!
allWeakSurvivorsOnWeakList
	self allFutureSpaceEntitiesDo:
		[:survivor|
		(manager isWeakNonImm: survivor) ifTrue:
			[(self is: survivor onWeaklingList: weakList) ifFalse:
				[^false]]].
	^true! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 9/14/2013 11:19'!
coInterpreter: aCoInterpreter
	coInterpreter := aCoInterpreter! !

!SpurGenerationScavenger methodsFor: 'remembered set' stamp: 'eem 11/25/2019 16:17'!
computeRefCountToShrinkRT
	"Some time in every scavenger's life there may come a time when someone writes code that stresses
	 the remembered table.  One might conclude that if the remembered table is full, then the right thing
	 to do is simply to tenure everything, emptying the remembered table.  Bt in some circumstances this
	 can be counter-productive, and result in the same situation arising soon after tenuring everything.
	 Instead, we can try and selectively prune the remembered table, tenuring only those objects that
	 are referenced by many objects in the remembered table.  That's what this algorithm does.  It
	 reference counts young objects referenced from the remembered set, and then sets a threshold
	 used to tenure objects oft referenced from the remembered set, thereby allowing  the remembered
	 set to shrink, while not tenuring everything.

	 Once in a network monitoring application in a galaxy not dissimilar from the one this code inhabits,
	 a tree of nodes referring to large integers was in precisely this situation.  The nodes were old, and
	 the integers were in new space.  Some of the nodes referred to shared numbers, some their own
	 unique numbers.  The numbers were updated frequently. Were new space simply tenured when the
	 remembered table was full, the remembered table would soon fill up as new numbers were computed.
	 Only by selectively pruning the remembered table of nodes that shared data, was a balance achieved
	 whereby the remembered table population was kept small, and tenuring rates were low."
	<inline: #never>
	| population |
	<var: 'population' declareC: 'long population[MaxRTRefCount + 1]'>
	self cCode: [self memset: population _: 0 _: (self sizeof: #long) * (MaxRTRefCount + 1)]
		inSmalltalk: [population := CArrayAccessor on: (Array new: MaxRTRefCount + 1 withAll: 0)].
	self assert: self allNewSpaceObjectsHaveZeroRTRefCount.
	self referenceCountRememberedReferents: population.
	self setRefCountToShrinkRT: population

	"For debugging:
	(manager allNewSpaceObjectsDo: [:o| manager rtRefCountOf: o put: 0])"! !

!SpurGenerationScavenger methodsFor: 'scavenger' stamp: 'eem 10/12/2013 14:39'!
computeTenuringThreshold
	| fractionSurvived |
	<var: 'fractionSurvived' type: #float>
	fractionSurvived := futureSpace limit = futureSpace start
							ifTrue:
								[0.0]
							ifFalse:
								[(futureSurvivorStart - futureSpace start) asFloat
									/ (futureSpace limit - futureSpace start)].
	tenureThreshold := fractionSurvived > 0.9
							ifTrue: [((pastSpace limit - pastSpace start) * (1.0 - tenuringProportion)) rounded + pastSpace start]
							ifFalse: [0]! !

!SpurGenerationScavenger methodsFor: 'scavenger' stamp: 'eem 06/22/2021 15:31'!
copyAndForward: survivor
	"copyAndForward: survivor copies a survivor object either to
	 futureSurvivorSpace or, if it is to be promoted, to oldSpace.
	 It leaves a forwarding pointer behind.  If the object is weak
	 then corpse is threaded onto the weakList for later treatment."
	<inline: false>
	| bytesInObj format tenure newLocation |
	self assert: ((manager isInEden: survivor) "cog methods should be excluded."
				or: [manager isInPastSpace: survivor]).
	bytesInObj := manager bytesInBody: survivor.
	format := manager formatOf: survivor.
	tenure := self shouldBeTenured: survivor. "Allow Slang to inline."
	newLocation := (tenure or: [futureSurvivorStart + bytesInObj > futureSpace limit])
						ifTrue: [self copyToOldSpace: survivor bytes: bytesInObj format: format]
						ifFalse: [self copyToFutureSpace: survivor bytes: bytesInObj].
	manager forwardSurvivor: survivor to: newLocation.
	"if weak or ephemeron add to the relevant list for subsequent scanning."
	(manager isWeakFormat: format) ifTrue:
		[self addToWeakList: survivor].
	((manager isEphemeronFormat: format)
	 and: [(self isScavengeSurvivor: (manager keyOfEphemeron: newLocation)) not]) ifTrue:
		[self addToEphemeronList: survivor].
	^newLocation! !

!SpurGenerationScavenger methodsFor: 'scavenger' stamp: 'eem 06/22/2021 15:31'!
copyAndForwardMourner: mourner
	"A special version of copyAndForward: for objects in the mournQueue.  If we're
	 in the good times tenuring regime then copy to futureSpace, otherwise tenure.
	 Also, don't repeat any of the ephemeron processing."
	<inline: false>
	| bytesInObj format tenure newLocation |
	self assert: ((manager isInEden: mourner) "cog methods should be excluded."
				or: [manager isInPastSpace: mourner]).
	bytesInObj := manager bytesInBody: mourner.
	format := manager formatOf: mourner.
	tenure := self shouldMournerBeTenured: mourner. "Allow Slang to inline."
	newLocation := (tenure or: [futureSurvivorStart + bytesInObj > futureSpace limit])
						ifTrue: [self copyToOldSpace: mourner bytes: bytesInObj format: format]
						ifFalse: [self copyToFutureSpace: mourner bytes: bytesInObj].
	manager forwardSurvivor: mourner to: newLocation.
	"if weak or ephemeron add to the relevant list for subsequent scanning."
	(manager isWeakFormat: format) ifTrue:
		[self addToWeakList: mourner].
	^newLocation! !

!SpurGenerationScavenger methodsFor: 'scavenger' stamp: 'eem 9/19/2018 17:31'!
copyToFutureSpace: survivor bytes: bytesInObject
	"Copy survivor to futureSpace.  Assume it will fit (checked by sender).
	 Answer the new oop of the object (it may have an overflow size field)."
	<inline: true>
	| startOfSurvivor newStart |
	statSurvivorCount := statSurvivorCount + 1. "we hope writes are cheap..."
	self assert: futureSurvivorStart + bytesInObject <= futureSpace limit.
	startOfSurvivor := manager startOfObject: survivor.
	newStart := futureSurvivorStart.
	futureSurvivorStart := futureSurvivorStart + bytesInObject.
	manager memcpy: newStart asVoidPointer _: startOfSurvivor asVoidPointer _: bytesInObject.
	tenureCriterion = TenureToShrinkRT ifTrue:
		[manager rtRefCountOf: newStart + (survivor - startOfSurvivor) put: 0].
	^newStart + (survivor - startOfSurvivor)! !

!SpurGenerationScavenger methodsFor: 'scavenger' stamp: 'eem 9/19/2018 17:31'!
copyToOldSpace: survivor bytes: bytesInObject format: formatOfSurvivor
	"Copy survivor to oldSpace.  Answer the new oop of the object."
	<inline: #never> "Should be too infrequent to lower icache density of copyAndForward:"
	| nTenures startOfSurvivor newStart newOop |
	self assert: (formatOfSurvivor = (manager formatOf: survivor)
				and: [((manager isMarked: survivor) not or: [tenureCriterion = MarkOnTenure])
				and: [tenureCriterion = TenureToShrinkRT
					or: [(manager isPinned: survivor) not
						and: [(manager isRemembered: survivor) not]]]]).
	nTenures := statTenures.
	startOfSurvivor := manager startOfObject: survivor.
	newStart := manager allocateOldSpaceChunkOfBytes: bytesInObject.
	newStart ifNil:
		[manager growOldSpaceByAtLeast: 0. "grow by growHeadroom"
		 newStart := manager allocateOldSpaceChunkOfBytes: bytesInObject.
		 newStart ifNil:
			[self error: 'out of memory']].
	"manager checkFreeSpace."
	manager memcpy: newStart asVoidPointer _: startOfSurvivor asVoidPointer _: bytesInObject.
	newOop := newStart + (survivor - startOfSurvivor).
	tenureCriterion >= (TenureToShrinkRT min: MarkOnTenure) ifTrue:
		[tenureCriterion = TenureToShrinkRT ifTrue:
			[manager rtRefCountOf: newOop put: 0].
		 tenureCriterion = MarkOnTenure ifTrue:
			[manager setIsMarkedOf: newOop to: true]].
	statTenures := nTenures + 1.
	(manager isAnyPointerFormat: formatOfSurvivor) ifTrue:
		["A very quick and dirty scan to find young referents.  If we misidentify bytes
		  in a CompiledMethod as young we don't care; it's unlikely, and a subsequent
		  scan of the rt will filter the object out.  But it's good to filter here because
		  otherwise an attempt to shrink the RT may simply fill it up with new objects,
		  and here the data is likely in the cache."
		 manager baseHeaderSize to: bytesInObject - (survivor - startOfSurvivor) - manager wordSize by: manager wordSize do:
			[:p| | field |
			field := manager longAt: survivor + p.
			(manager isReallyYoung: field) ifTrue:
				[self remember: newOop.
				 ^newOop]]].
	^newOop! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 5/22/2014 13:29'!
corpseForCorpseOffset: corpseOffset
	"Use the identityHash and format fields to construct a 27 bit offset through
	 non-future newSpace and use this to implement lists for weak array and
	 ephemeron processing.  27 bits of 8 byte allocationUnits units is 2 ^ 30
	 bytes, or 1Gb, big enough for newSpace for a good few years yet."
	^corpseOffset - 1 << manager shiftForAllocationUnit + manager newSpaceStart! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 5/22/2014 13:30'!
corpseOffsetOf: corpse
	"Answer the offset of the corpse in newSpace as a multiple of allocationUnits.
	 Use the identityHash and format fields to construct a 27 bit offset through
	 non-future newSpace and use this to implement lists for weak array and
	 ephemeron processing.  27 bits of 8 byte allocationUnits units is 2 ^ 30
	 bytes or 1Gb, big enough for newSpace for a good few years yet.  Add
	 one to ensure that a corpse offset is always non-zero, even when it is
	 that of the first object in newSpace."
	^corpse - manager newSpaceStart >> manager shiftForAllocationUnit + 1! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 10/12/2013 15:49'!
eden
	<returnTypeC: #SpurNewSpaceSpace>
	<cmacro: '() GIV(eden)'>
	^eden! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 11/14/2019 14:25'!
edenBytes
	<inline: #always>
	^eden limit - eden start! !

!SpurGenerationScavenger methodsFor: 'scavenger' stamp: 'eem 10/14/2013 00:38'!
exchangeSurvivorSpaces
	| temp |
	<var: #temp type: #SpurNewSpaceSpace>
	temp := pastSpace.
	pastSpace := futureSpace.
	futureSpace := temp! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 5/25/2016 13:38'!
fireEphemeronsInRememberedSet
	"There are ephemerons to be fired in the remembered set.
	 Fire them and scavenge their keys.  Leave it to scavengeLoop
	 to remove any scavenged ephemerons that no longer have
	 new referents."
	| i |
	self assert: self noUnfiredEphemeronsAtEndOfRememberedSet.

	i := 0.
	[i < numRememberedEphemerons] whileTrue:
		[ | ephemeron key |
		 ephemeron := rememberedSet at: i.
		 self assert: (manager isEphemeron: ephemeron).
		 key := manager keyOfEphemeron: ephemeron.
		 (self isScavengeSurvivor: key) ifFalse:
			[coInterpreter fireEphemeron: ephemeron.
			 manager
				storePointerUnchecked: 0
				ofObject: ephemeron
				withValue: (self copyAndForward: key)].
		 "Fired ephemerons should have had their format changed."
		 self deny: ((self isScavengeSurvivor: key) and: [manager isEphemeron: ephemeron]).
		 (self scavengeReferentsOf: ephemeron)
			ifTrue: "keep in set"
				[i := i + 1]
			ifFalse:
				[manager setIsRememberedOf: ephemeron to: false.
				"remove from set by overwriting with next-to-be scanned"
				 numRememberedEphemerons := numRememberedEphemerons - 1.
				 previousRememberedSetSize := previousRememberedSetSize - 1.
				 rememberedSetSize := rememberedSetSize - 1.
				 "First overwrite with last firable ephemeron (could be a noop if this is the last one).
				  Then overwrite last firable entry with next unscanned rememberedSet entry (could also be a noop).
				  Then overwrite next unscanned entry with last unscanned rememberedSet entry (could also be a noop)."
				 rememberedSet
					at: i
						put: (rememberedSet at: numRememberedEphemerons);
					at: numRememberedEphemerons
						put: (rememberedSet at: previousRememberedSetSize);
					at: previousRememberedSetSize
						put: (rememberedSet at: rememberedSetSize)]].

	"no more firable ephemerons in this cycle.
	 scavengeRememberedSetStartingAt: may find new ones."
	numRememberedEphemerons := 0! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 5/25/2016 18:48'!
fireEphemeronsOnEphemeronList
	"There are ephemerons to be fired in the remembered set.
	 Fire them and scavenge their keys.  Be careful since copyAndForward:
	 can remember ephemerons (ephemerons pointing to ephemerons)."
	| ephemeron ephemeronCorpse key oldList oldCorpse | "old ones for debugging"
	ephemeronList ifNil:
		[^self].
	oldCorpse := nil.
	ephemeronCorpse := self firstCorpse: ephemeronList.
	"Reset the list head so that new ephemerons will get added
	 to a new list, not concatenated on the one we are scanning."
	oldList := ephemeronList.
	ephemeronList := nil.
	[ephemeronCorpse notNil] whileTrue:
		[self assert: ((manager isYoung: ephemeronCorpse) and: [manager isForwarded: ephemeronCorpse]).
		 ephemeron := manager followForwarded: ephemeronCorpse.
		 key := manager keyOfMaybeFiredEphemeron: ephemeron.
		 (self isScavengeSurvivor: key) ifFalse:
			[coInterpreter fireEphemeron: ephemeron.
			 manager
				storePointerUnchecked: 0
				ofObject: ephemeron
				withValue: (self copyAndForward: key)].
		 "Fired ephemerons should have had their format changed."
		 self deny: ((self isScavengeSurvivor: key) and: [manager isEphemeron: ephemeron]).
		 self cCoerceSimple: (self scavengeReferentsOf: ephemeron) to: #void.
		 oldCorpse := ephemeronCorpse.
		 ephemeronCorpse := self nextCorpseOrNil: ephemeronCorpse]! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 9/30/2013 11:45'!
firstCorpse: headOfCorpseList
	^self corpseForCorpseOffset: headOfCorpseList! !

!SpurGenerationScavenger methodsFor: 'gc - global' stamp: 'eem 10/29/2013 15:46'!
followRememberedForwardersAndForgetFreeObjects
	"Scan the remembered set. Follow any forwarded objects,
	 and remove free objects.  This is for global scan-mark GC."
	| index obj |
	index := 0.
	[index < rememberedSetSize] whileTrue:
		[obj := rememberedSet at: index.
		 (manager isFreeObject: obj) "free; remove by overwriting with last element"
			ifTrue:
				[rememberedSetSize := rememberedSetSize - 1.
				 rememberedSet at: index put: (rememberedSet at: rememberedSetSize)]
			ifFalse:
				[(manager isForwarded: obj) ifTrue:
					[obj := manager followForwarded: obj.
					 manager setIsRememberedOf: obj to: true.
					 rememberedSet at: index put: obj].
				 index := index + 1]]! !

!SpurGenerationScavenger methodsFor: 'gc - global' stamp: 'eem 4/30/2014 17:34'!
followRememberedForwardersAndForgetFreeObjectsForPigCompact
	"Scan the remembered set. Follow any forwarded objects,
	 and remove free objects.  This is for global scan-mark GC."
	| index obj |
	index := 0.
	[index < rememberedSetSize] whileTrue:
		[obj := rememberedSet at: index.
		 (manager isFreeObject: obj) "free; remove by overwriting with last element"
			ifTrue:
				[rememberedSetSize := rememberedSetSize - 1.
				 rememberedSet at: index put: (rememberedSet at: rememberedSetSize)]
			ifFalse:
				[(manager isForwarded: obj) ifTrue:
					[manager setIsRememberedOf: obj to: false.
					 obj := manager followForwarded: obj.
					 self assert: (manager isRemembered: obj).
					 rememberedSet at: index put: obj].
				 index := index + 1]]! !

!SpurGenerationScavenger methodsFor: 'gc - global' stamp: 'eem 7/11/2017 19:09'!
forgetObject: objOop
	"Forget the argument."
	self assert: rememberedSetSize > 0.
	self assert: (manager isRemembered: objOop).
	manager setIsRememberedOf: objOop to: false.
	objOop = (rememberedSet at: rememberedSetSize - 1) ifFalse:
		[| index |
		 index := 0.
		 [index < rememberedSetSize] whileTrue:
			[objOop = (rememberedSet at: index)
				ifTrue:
					[rememberedSet at: index put: (rememberedSet at: rememberedSetSize - 1).
					 index := rememberedSetSize]
				ifFalse: [index := index + 1]]].
	rememberedSetSize := rememberedSetSize - 1.
	self assert: rememberedSetSize >= 0! !

!SpurGenerationScavenger methodsFor: 'gc - global' stamp: 'eem 10/18/2013 16:28'!
forgetUnmarkedRememberedObjects
	"Remove all unmarked objects from the remembered set.
	 This is for global scan-mark GC."
	| index |
	index := 0.
	[index < rememberedSetSize] whileTrue:
		[| obj |
		 obj := rememberedSet at: index.
		 (manager isMarked: obj)
			ifTrue: [index := index + 1]
			ifFalse: "unmarked; remove by overwriting with last element."
				[manager setIsRememberedOf: obj to: false.
				 rememberedSetSize := rememberedSetSize - 1.
				 rememberedSet at: index put: (rememberedSet at: rememberedSetSize)]].
	self assert: rememberedSetSize >= 0! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 10/12/2013 15:49'!
futureSpace
	<returnTypeC: #SpurNewSpaceSpace>
	<cmacro: '() GIV(futureSpace)'>
	^futureSpace! !

!SpurGenerationScavenger methodsFor: 'debug support' stamp: 'eem 9/29/2013 10:21'!
futureSpaceObjectsDo: aBlock
	| obj |
	futureSurvivorStart > futureSpace start ifTrue:
		[obj := manager objectStartingAt: futureSpace start.
		 [obj < futureSurvivorStart] whileTrue:
			[aBlock value: obj.
			 obj := manager objectAfter: obj limit: futureSurvivorStart]]! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 5/7/2014 09:47'!
futureSurvivorStart
	<cmacro: '() GIV(futureSurvivorStart)'>
	^futureSurvivorStart! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 10/12/2013 14:39'!
getRawTenuringThreshold
	^tenureThreshold! !

!SpurGenerationScavenger methodsFor: 'remembered set' stamp: 'cb 10/23/2017 12:39'!
growRememberedSet
	| obj numSlots newObj base |
	<inline: false> "Don't ruin locality in remember:"
	<var: #base type: #'sqInt *'>
	obj := manager rememberedSetObj.
	numSlots := manager numSlotsOf: obj.
	self assert: numSlots >= 1024.
	newObj := manager allocatePinnedSlots: numSlots * 2.
	newObj ifNil:
		[newObj := manager allocatePinnedSlots: numSlots + 1024.
		 newObj ifNil:
			[(manager growOldSpaceByAtLeast: numSlots + 1024) ifNil: [self error: 'could not grow remembered set'].
			 newObj := manager allocatePinnedSlots: numSlots + 1024. "cannot fail"]].
	manager rememberedSetObj: newObj.
	base := manager firstIndexableField: newObj.
	0 to: rememberedSetSize - 1 do:
		[:i| base at: i put: (rememberedSet at: i)].
	"if growing in the middle of a GC, need to preserve marked status."
	(manager isMarked: obj) ifTrue:
		[manager
			setIsMarkedOf: newObj to: true;
			setIsMarkedOf: obj to: false].
	manager freeObject: obj.
	rememberedSet := base.
	rememberedSetLimit := manager numSlotsOf: newObj.
	self setRememberedSetRedZone! !

!SpurGenerationScavenger methodsFor: 'store check' stamp: 'eem 6/7/2017 18:35'!
indexInRememberedSet: objOop
	<doNotGenerate>
	0 to: rememberedSetSize - 1 do:
		[:i|
		(rememberedSet at: i) = objOop ifTrue:
			[^i]].
	^nil! !

!SpurGenerationScavenger methodsFor: 'initialization' stamp: 'eem 9/16/2013 14:29'!
initFutureSpaceStart
	| oldStart |
	oldStart := futureSurvivorStart.
	futureSurvivorStart := futureSpace start.
	^oldStart! !

!SpurGenerationScavenger methodsFor: 'initialization' stamp: 'eem 11/13/2017 12:00'!
initialize
	pastSpace := SpurNewSpaceSpace new.
	futureSpace := SpurNewSpaceSpace new.
	eden := SpurNewSpaceSpace new.
	rememberedSetSize := 0.
	tenureThreshold := 0.
	statSurvivorCount := statTenures := 0.
	scavengeLogRecord := SpurScavengeLogRecord new! !

!SpurGenerationScavenger methodsFor: 'initialization' stamp: 'eem 6/9/2017 14:46'!
initializeRememberedSet
	| obj |
	obj := manager rememberedSetObj.
	obj = manager nilObject
		ifTrue:
			[obj := manager allocatePinnedSlots: 1024.
			 manager rememberedSetObj: obj]
		ifFalse: "The Spur32to64BitBootstrap failed to set the type of rememberedSetObj to 64-bit indexability.
				 This is unimportant except for simulation; rememberedSet is declared as sqInt *, but in to have
				 firstIndexableField: below answer a suitable type the format must be wordIndexableFormat."
			[manager setFormatOf: obj to: manager wordIndexableFormat].
	self assert: (manager formatOf: obj) = manager wordIndexableFormat.
	self assert: (manager isPinned: obj).
	rememberedSet := manager firstIndexableField: obj.
	rememberedSetSize := 0.
	rememberedSetLimit := manager numSlotsOf: obj.
	self setRememberedSetRedZone! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 9/29/2013 11:47'!
is: oop onWeaklingList: listHead
	| corpse |
	corpse := self firstCorpse: listHead.
	[corpse notNil] whileTrue:
		[oop = (manager followForwarded: corpse) ifTrue:
			[^true].
		 corpse := self nextCorpseOrNil: corpse].
	^false! !

!SpurGenerationScavenger methodsFor: 'store check' stamp: 'eem 10/15/2013 00:24'!
isInRememberedSet: objOop
	0 to: rememberedSetSize - 1 do:
		[:i|
		(rememberedSet at: i) = objOop ifTrue:
			[^true]].
	^false! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 7/9/2014 06:49'!
isMaybeOldScavengeSurvivor: oop
	"Answer whether the oop has survived a scavenge.  This version is
	 for processing weak survivors and must cope with the scavenge in
	 freeUnmarkedObjectsAndSortAndCoalesceFreeSpaceForPigCompact."
	| target |
	(manager isImmediate: oop) ifTrue:
		[^true].
	(manager isForwarded: oop)
		ifTrue:
			[target := manager followForwarded: oop.
			 (manager isImmediate: oop) ifTrue:
				[^true]]
		ifFalse: [target := oop].
	^(manager isOldObject: target)
		ifTrue:
			[tenureCriterion ~= MarkOnTenure
			 or: [manager isMarked: target]]
		ifFalse:
			[manager isInFutureSpace: target]! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 11/29/2013 12:49'!
isScavengeSurvivor: oop
	"Answer whether the oop has survived a scavenge.  This is equivalent to
		| target |
		(manager isImmediate: oop) ifTrue:
			[^true].
		target := (manager isForwarded: oop)
					ifTrue: [manager followForwarded: oop]
					ifFalse: [oop].
	 	^((manager isInEden: target)
		  or: [(manager isInPastSpace: target)]) not"
	| target |
	(manager isImmediate: oop) ifTrue:
		[^true].
	(manager isForwarded: oop)
		ifTrue: [target := manager followForwarded: oop]
		ifFalse: [target := oop].
	^(manager isReallyYoung: target) not
	  or: [manager isInFutureSpace: target]! !

!SpurGenerationScavenger methodsFor: 'logging' stamp: 'eem 11/13/2017 14:21'!
logEndScavenge
	<inline: #always>
	scavengeLogRecord
		eSurvivorBytes: futureSurvivorStart - pastSpace start;
		eRememberedSetSize: rememberedSetSize;
		eStatTenures: statTenures! !

!SpurGenerationScavenger methodsFor: 'logging' stamp: 'eem 11/13/2017 12:01'!
logScavenge
	<inline: #always>
	scavengeLog ifNotNil:
		[self writeScavengeLog]! !

!SpurGenerationScavenger methodsFor: 'logging' stamp: 'eem 11/13/2017 16:09'!
logStartScavenge
	<inline: #always>
	scavengeLogRecord
		sEdenBytes: manager freeStart - eden start;
		sPastBytes: manager pastSpaceStart - pastSpace start;
		sRememberedSetSize: rememberedSetSize;
		sRememberedSetRedZone: rememberedSetRedZone;
		sRememberedSetLimit: rememberedSetLimit;
		sStatTenures: statTenures! !

!SpurGenerationScavenger methodsFor: 'logging' stamp: 'eem 11/7/2017 15:45'!
logStream
	<inline: #always>
	^scavengeLog! !

!SpurGenerationScavenger methodsFor: 'logging' stamp: 'eem 11/13/2017 14:21'!
logTenuringPolicy
	<inline: #always>
	scavengeLogRecord
		tTenureCriterion: tenureCriterion;
		tTenureThreshold: ((tenureCriterion = TenureByAge and: [tenureThreshold > pastSpace start])
								ifTrue: [tenureThreshold - pastSpace start]
								ifFalse: [0]);
		tRefCountToShrinkRT: refCountToShrinkRT! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 6/23/2014 18:02'!
newSpaceCapacity
	<inline: false>
	<returnTypeC: #usqInt>
	^eden limit - (futureSpace start min: pastSpace start)! !

!SpurGenerationScavenger methodsFor: 'initialization' stamp: 'eem 3/18/2020 12:59'!
newSpaceStart: startAddress newSpaceBytes: totalBytes survivorBytes: requestedSurvivorBytes
	| actualEdenBytes survivorBytes |
	survivorBytes := requestedSurvivorBytes truncateTo: manager allocationUnit.
	actualEdenBytes := totalBytes - survivorBytes - survivorBytes truncateTo: manager allocationUnit.
	self assert: totalBytes - actualEdenBytes - survivorBytes - survivorBytes < manager allocationUnit.

	"for tenuring we require older objects below younger objects.  since allocation
	 grows up this means that the survivor spaces must precede eden."

	pastSpace start: startAddress; limit: startAddress + survivorBytes.
	futureSpace start: pastSpace limit; limit: pastSpace limit + survivorBytes.
	eden start: futureSpace limit; limit: startAddress + totalBytes.

	self assert: self futureSpace limit <= (startAddress + totalBytes).
	self assert: self eden start \\ manager allocationUnit
				+ (self eden limit \\ manager allocationUnit) = 0.
	self assert: self pastSpace start \\ manager allocationUnit
				+ (self pastSpace limit \\ manager allocationUnit) = 0.
	self assert: self futureSpace start \\ manager allocationUnit
				+ (self futureSpace limit \\ manager allocationUnit) = 0.

	self initFutureSpaceStart.

	tenuringProportion := 0.9! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 11/15/2019 08:58'!
nextCorpseOffset: corpse
	"Answer the offset of the next corpse to corpse, which is zero if none.
	 Use the identityHash and format fields to construct a 27 bit offset through
	 non-future newSpace and use this to implement lists for weak array and
	 ephemeron processing.  27 bits of 8 byte allocationUnits units is 2 ^ 30 bytes
	 or 1Gb, big enough for newSpace for a good few years yet."
	<inline: true>
	^(manager rawHashBitsOf: corpse) << manager formatFieldWidthShift
	 + (manager formatOf: corpse)! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 9/30/2013 11:39'!
nextCorpseOrNil: corpse
	"corpse is the corpse of a weak array that has been added to the weakList.
	 Answer the next object on the list, or nil if none."
	| listOffset |
	self assert: (manager isYoung: corpse).
	listOffset := self nextCorpseOffset: corpse.
	^listOffset ~= 0 ifTrue:
		[self corpseForCorpseOffset: listOffset]! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 9/29/2013 11:42'!
noUnfiredEphemeronsAtEndOfRememberedSet
	"For assert checking only."
	numRememberedEphemerons to: rememberedSetSize - 1 do:
		[:i| | referrer |
		referrer := rememberedSet at: i.
		(manager isEphemeron: referrer) ifTrue:
			[(self isScavengeSurvivor: (manager keyOfEphemeron: referrer)) ifFalse:
				[^false]]].
	^true! !

!SpurGenerationScavenger methodsFor: 'logging' stamp: 'eem 11/9/2017 15:23'!
openScavengeLog
	<api>
	scavengeLog := self f: 'scavenge.log' open: 'a+'! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 10/12/2013 15:49'!
pastSpace
	<returnTypeC: #SpurNewSpaceSpace>
	<cmacro: '() GIV(pastSpace)'>
	^pastSpace! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 9/10/2013 17:52'!
pastSpaceBytes
	^pastSpace limit - pastSpace start! !

!SpurGenerationScavenger methodsFor: 'debug support' stamp: 'eem 5/19/2016 12:35'!
printRememberedSet
	"Print the objects in the remembered set."
	<api>
	0 to: rememberedSetSize - 1 do:
		[:i|
		coInterpreter printNum: i; space; shortPrintOop: (rememberedSet at: i)]! !

!SpurGenerationScavenger methodsFor: 'debug support' stamp: 'eem 6/21/2021 13:31'!
printSpaces
	coInterpreter print: 'eden start '; printHex: eden start; print: ' end/free '; printHex: manager freeStart; print: ' limit '; printHex: eden limit; cr.
	coInterpreter print: 'past start '; printHex: pastSpace start; print: ' end '; printHex: manager pastSpaceStart; print: ' limit '; printHex: pastSpace limit; cr.
	coInterpreter print: 'future start '; printHex: futureSpace start; print: ' survivorStart '; printHex: futureSurvivorStart; print: ' limit '; printHex: futureSpace limit; cr! !

!SpurGenerationScavenger methodsFor: 'debug support' stamp: 'eem 9/29/2013 10:16'!
printWeaklingList: listHead
	"Print the objects on either the weakList or the ephemeronList."
	| corpse |
	corpse := self firstCorpse: listHead.
	corpse ifNil:
		[coInterpreter print: 'empty'; cr.
		 ^self].
	[corpse notNil] whileTrue:
		[coInterpreter printHexnp: corpse; print: ' -> '; shortPrintOop: (manager followForwarded: corpse).
		 corpse := self nextCorpseOrNil: corpse]! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 10/25/2013 18:06'!
processEphemerons
	"There are ephemerons to be scavenged.  Scavenge them and fire any whose keys are
	 still in pastSpace and/or eden.  The unscavenged ephemerons in this cycle can only be
	 fired if all the unscavenged ephemerons in this cycle are firable, because references
	 to ephemeron keys from unfired ephemerons should prevent the ephemerons with
	 those keys from firing.  So scavenge ephemerons with surviving keys, and only if none
	 are found, fire ephemerons with unreferenced keys, and scavenge them.   Read the
	 class comment for a more in-depth description of the algorithm."
	<inline: false>
	| unfiredEphemeronsScavenged |
	unfiredEphemeronsScavenged := self scavengeUnfiredEphemeronsInRememberedSet.
	self scavengeUnfiredEphemeronsOnEphemeronList ifTrue:
		[unfiredEphemeronsScavenged := true].
	unfiredEphemeronsScavenged ifFalse:
		[self fireEphemeronsInRememberedSet.
		 self fireEphemeronsOnEphemeronList]! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 8/31/2015 11:11'!
processWeakSurvivor: weakObj
	"Process a weak survivor on the weakList.  Those of its fields
	 which have not survived the scavenge should be nilled, and if any
	 are, the coInterpreter should be informed via fireFinalization:.
	 Answer if the weakObj has any young referents."
	| weakObjShouldMourn hasYoungReferents numStrongSlots  |
	weakObjShouldMourn := hasYoungReferents := false.
	"N.B. generateToByDoLimitExpression:negative:on: guards against (unsigned)0 - 1 going +ve"
	numStrongSlots := manager numFixedSlotsOf: weakObj.
	0 to: numStrongSlots - 1 do:
		[:i| | referent |
		 referent := manager fetchPointer: i ofObject: weakObj.
		 ((manager isNonImmediate: referent)
		  and: [manager isYoungObject: referent]) ifTrue:
			[hasYoungReferents := true]].
	numStrongSlots
		to: (manager numSlotsOf: weakObj) - 1
		do: [:i| | referent |
			referent := manager fetchPointer: i ofObject: weakObj.
			"Referent could be forwarded due to scavenging or a become:, don't assume."
			(manager isNonImmediate: referent) ifTrue:
				[(manager isForwarded: referent) ifTrue:
					[referent := manager followForwarded: referent.
					 "weakObj is either young or already in remembered table; no need to check"
					 self assert: ((manager isReallyYoungObject: weakObj)
								or: [manager isRemembered: weakObj]).
					 manager storePointerUnchecked: i ofObject: weakObj withValue: referent].
				(self isMaybeOldScavengeSurvivor: referent)
					ifTrue:
						[(manager isYoungObject: referent) ifTrue:
							[hasYoungReferents := true]]
					ifFalse:
						[weakObjShouldMourn := true.
						 manager
							storePointerUnchecked: i
							ofObject: weakObj
							withValue: manager nilObject]]].
	weakObjShouldMourn ifTrue:
		[coInterpreter fireFinalization: weakObj].
	^hasYoungReferents! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 7/17/2014 20:56'!
processWeaklings
	"Go through the remembered set and the weak list, nilling references to
	 any objects that didn't survive the scavenge. Read the class comment
	 for a more in-depth description of the algorithm."
	<inline: false>
	| i rootObj weakCorpse weakObj |
	self assert: self allWeakSurvivorsOnWeakList.
	i := 0.
	[i < rememberedSetSize] whileTrue:
		[rootObj := rememberedSet at: i.
		(manager isWeakNonImm: rootObj)
			ifTrue:
				["If no more referents, remove by overwriting with the last element in the set."
				 (self processWeakSurvivor: rootObj)
					ifFalse:
						[manager setIsRememberedOf: rootObj to: false.
						 i + 1 < rememberedSetSize ifTrue:
							[rememberedSet at: i put: (rememberedSet at: rememberedSetSize - 1)].
						 rememberedSetSize := rememberedSetSize - 1]
					ifTrue: [i := i + 1]]
			ifFalse: [i := i + 1]].
	weakList ifNotNil:
		[weakCorpse := self firstCorpse: weakList.
		 [weakCorpse notNil] whileTrue:
			[self assert: (manager isForwarded: weakCorpse).
			 weakObj := manager followForwarded: weakCorpse.
			 "weakObj may have been tenured..."
			 ((self processWeakSurvivor: weakObj)
			  and: [(manager isOldObject: weakObj)
			  and: [(manager isRemembered: weakObj) not]]) ifTrue:
				[self remember: weakObj].
			 weakCorpse := self nextCorpseOrNil: weakCorpse].
		weakList := nil]! !

!SpurGenerationScavenger methodsFor: 'remembered set' stamp: 'eem 5/19/2016 12:41'!
referenceCountRememberedReferents: population
	"Both reference count young objects reachable from the RT,
	 and count the populations of each ref count, in a single pass."
	<var: 'population' declareC: 'long population[MaxRTRefCount + 1]'>
	<inline: true>
	0 to: rememberedSetSize - 1 do:
		[:i| | elephant |
		elephant := rememberedSet at: i.
		(manager isForwarded: elephant) ifTrue:
			[elephant := manager followForwarded: elephant.
			 (manager isImmediate: elephant) ifTrue: "take care if elephant forwarded to an immediate"
				[elephant := manager nilObject]].
		0 to: (manager numPointerSlotsOf: elephant) - 1 do:
			[:j| | referent refCount |
			referent := manager fetchPointer: j ofObject: elephant.
			(manager isReallyYoung: referent) ifTrue:
				[refCount := manager rtRefCountOf: referent.
				 refCount < MaxRTRefCount ifTrue:
					[refCount > 0 ifTrue:
						[population at: refCount put: (population at: refCount) - 1].
					 refCount := refCount + 1.
					 manager rtRefCountOf: referent put: refCount.
					 population at: refCount put: (population at: refCount) + 1]]]].! !

!SpurGenerationScavenger methodsFor: 'remembered set' stamp: 'eem 6/7/2017 19:14'!
relocateRememberedSet
	"For SpurPlanningCompactor"
	rememberedSet := manager firstIndexableField: manager rememberedSetObj! !

!SpurGenerationScavenger methodsFor: 'store check' stamp: 'eem 7/19/2014 11:33'!
remember: objOop
	"Add the argument to the remembered set and set its isRemembered bit to true.
	 Answer the argument for the benefit of the Cogit."
	<api>
	<inline: false>
	self assert: (manager isNonImmediate: objOop).
	self deny: (manager isYoungObject: objOop).
	self deny: (manager isRemembered: objOop).
	self deny: (self isInRememberedSet: objOop).
	manager setIsRememberedOf: objOop to: true.
	rememberedSetSize >= rememberedSetLimit ifTrue:
		[self growRememberedSet].
	rememberedSet at: rememberedSetSize put: objOop.
	(rememberedSetSize := rememberedSetSize + 1) >= rememberedSetRedZone ifTrue:
		[manager scheduleScavenge].
	^objOop! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 5/5/2015 16:14'!
rememberedSetLimit
	<cmacro: '() GIV(rememberedSetLimit)'>
	^rememberedSetLimit! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 10/12/2013 14:11'!
rememberedSetSize
	<cmacro: '() GIV(rememberedSetSize)'>
	^rememberedSetSize! !

!SpurGenerationScavenger methodsFor: 'debug support' stamp: 'eem 11/14/2019 13:40'!
rememberedSetWithIndexDo: aBlock
	<inline: #always>
	0 to: rememberedSetSize - 1 do:
		[:i| aBlock value: (rememberedSet at: i) value: i]! !

!SpurGenerationScavenger methodsFor: 'scavenger' stamp: 'eem 11/10/2017 10:59'!
scavenge: tenuringCriterion
	"The main routine, scavenge, scavenges young objects reachable from the roots (the stack zone
	 and the rememberedTable).  It first scavenges the new objects immediately reachable from old
	 ones (all in the remembered table), then the stack zone.  Then it scavenges those that are
	 transitively reachable.  If this results in a promotion, the promotee gets remembered, and it first
	 scavenges objects adjacent to the promotee, then scavenges the ones reachable from the
	 promoted.  This loop continues until no more reachable objects are left.  At that point,
	 pastSurvivorSpace is exchanged with futureSurvivorSpace.  Then any surviving weakArrays and
	 weakArrays in the remembered set can be processed and their dead elements nilled.

	 By default promotion (tenuring) is based on age and ammount of objects scavenged.  But
	 tenuring can be based on e.g. a particular class.  The argument selects the tenuring criterion.

	 Answer the limit of pastSpace, to allow the memory manager to bounds check survivors."
	statSurvivorCount := 0.
	tenureCriterion := tenuringCriterion.
	self logStartScavenge.
	self strategizeToLimitRememberedTable.
	self logTenuringPolicy.
	self scavengeLoop.
	self processWeaklings.
	self computeTenuringThreshold.
	self exchangeSurvivorSpaces.
	self logEndScavenge.
	^self initFutureSpaceStart! !

!SpurGenerationScavenger methodsFor: 'scavenger' stamp: 'eem 10/25/2013 18:04'!
scavengeFutureSurvivorSpaceStartingAt: initialAddress
	"scavengeFutureSurvivorSpaceStartingAt: does a depth-first traversal of the
	 new objects starting at the one at initialAddress in futureSurvivorSpace."
	<inline: false>
	| ptr obj |
	ptr := initialAddress.
	[ptr < futureSurvivorStart] whileTrue:
		[obj := manager objectStartingAt: ptr.
		 ptr := manager addressAfter: obj.
		 self cCoerceSimple: (self scavengeReferentsOf: obj) to: #void]! !

!SpurGenerationScavenger methodsFor: 'scavenger' stamp: 'eem 5/26/2016 08:47'!
scavengeLoop
	"This is the inner loop of the main routine, scavenge.  It first scavenges the new objects immediately
	 reachable from old ones. Then it scavenges those that are transitively reachable.  If this results in a
	 promotion, the promotee gets remembered, and it first scavenges objects adjacent to the promotee,
	 then scavenges the ones reachable from the promoted.  This loop continues until no more reachable
	 objects are left.  At that point, pastSurvivorSpace is exchanged with futureSurvivorSpace.

	 Notice that each pointer in a live object is inspected once and only once.  The previousRememberedSetSize
	 and previousFutureSurvivorSpaceLimit variables ensure that no object is scanned twice, as well as
	 detecting closure.  If this were not true, some pointers might get forwarded twice.

	 An extension of the algorithm presented in David's original paper is to handle weak arrays and ephemerons.
	 Weak arrays should not have their weak referents scavenged unless there are strong references to them.
	 Ephemerons should fire if their key is not reachable other than from ephemerons and weak arrays.
	 Handle this by maintaining a list for weak arrays and a list for ephemerons, which allow scavenging these
	 objects once all other objects in new space have been scavenged, hence allowing the scavenger to
	 detect which referents in new space of weak arrays are dead and of ephemeron keys are only live due to
	 ephemerons.  Read the class comment for a more in-depth description of the algorithm."
	<inline: false>
	| previousFutureSurvivorStart firstTime |
	self assert: futureSurvivorStart = futureSpace start. "future space should be empty at the start"

	weakList := ephemeronList := nil.
	numRememberedEphemerons := 0.
	firstTime := true.
	previousRememberedSetSize := 0.
	previousFutureSurvivorStart := futureSurvivorStart.

	coInterpreter initStackPageGC.

	[self scavengeRememberedSetStartingAt: previousRememberedSetSize.
	 previousRememberedSetSize := rememberedSetSize.
	 firstTime ifTrue:
		[coInterpreter mapInterpreterOops.
		 manager mapMournQueue.
		 manager mapExtraRoots.
		 firstTime := false].
	 "if nothing more copied and forwarded (or remembered by mapInterpreterOops)
	  to scavenge, and no ephemerons to process, scavenge is done."
	 (previousRememberedSetSize = rememberedSetSize
	  and: [previousFutureSurvivorStart = futureSurvivorStart
	  and: [numRememberedEphemerons = 0
	  and: [ephemeronList isNil]]]) ifTrue:
		[^self].

	 self scavengeFutureSurvivorSpaceStartingAt: previousFutureSurvivorStart.
	 previousFutureSurvivorStart := futureSurvivorStart.

	 "no more roots created to scavenge..."
	 previousRememberedSetSize = rememberedSetSize ifTrue:
		[(numRememberedEphemerons = 0
		  and: [ephemeronList isNil]) ifTrue:
			[^self]. "no ephemerons to process, scavenge is done."

		 "all reachable objects in this cycle have been promoted to futureSpace.
		  ephemerons can now be processed."
		 self processEphemerons]] repeat! !

!SpurGenerationScavenger methodsFor: 'scavenger' stamp: 'eem 6/16/2014 12:50'!
scavengeReferentsOf: referrer
	"scavengeReferentsOf: referrer inspects all the pointers in referrer.  If
	 any are new objects, it has them moved to FutureSurvivorSpace, and
	 answers truth. If there are no new referents, it answers falsity. To handle
	 weak arrays, if the referrer is weak only scavenge strong slots and answer
	 true so that it won't be removed from the remembered set until later."
	| fmt foundNewReferentOrIsWeakling numSlots |
	"forwarding objects should be followed by callers,
	 unless the forwarder is a root in the remembered table."
	self assert: ((manager isForwarded: referrer) not
				or: [manager isRemembered: referrer]).
	"unscanned ephemerons should be scanned later."
	self assert: ((manager isEphemeron: referrer) not
				or: [(self isScavengeSurvivor: (manager keyOfEphemeron: referrer))
				or: [self is: referrer onWeaklingList: ephemeronList]]).
	fmt := manager formatOf: referrer.
	foundNewReferentOrIsWeakling := manager isWeakFormat: fmt.
	numSlots := manager numStrongSlotsOf: referrer format: fmt ephemeronInactiveIf: #isScavengeSurvivor:.
	0 to: numSlots - 1 do:
		[:i| | referent newLocation |
		referent := manager fetchPointer: i ofMaybeForwardedObject: referrer.
		(manager isNonImmediate: referent) ifTrue:
			["a forwarding pointer could be because of become: or scavenging."
			 (manager isForwarded: referent) ifTrue:
				[referent := manager followForwarded: referent].
			 (manager isReallyYoungObject: referent)
				ifTrue: "if target is already in future space forwarding pointer was due to a become:."
					[(manager isInFutureSpace: referent)
						ifTrue: [newLocation := referent. foundNewReferentOrIsWeakling := true]
						ifFalse:
							[newLocation := self copyAndForward: referent.
							 (manager isYoung: newLocation) ifTrue:
								[foundNewReferentOrIsWeakling := true]].
					 manager storePointerUnchecked: i ofMaybeForwardedObject: referrer withValue: newLocation]
				ifFalse:
					[manager storePointerUnchecked: i ofMaybeForwardedObject: referrer withValue: referent]]].
	^foundNewReferentOrIsWeakling! !

!SpurGenerationScavenger methodsFor: 'scavenger' stamp: 'eem 10/25/2013 18:05'!
scavengeRememberedSetStartingAt: n
	"scavengeRememberedSetStartingAt: n traverses objects in the remembered
	 set starting at the nth one.  If the object does not refer to any new objects, it
	 is removed from the set.  Otherwise, its new referents are scavenged.  Defer
	 scavenging ephemerons until after a complete scavenge has been performed,
	 so that triggered ephemerons can be fired.  Move them to the front of the set
	 and count them in numRememberedEphemerons for later scanning."
	<inline: false>
	| destIndex sourceIndex referrer |
	sourceIndex := destIndex := n.
	[sourceIndex < rememberedSetSize] whileTrue:
		["*Don't* follow forwarding pointers here. oldSpace objects may refer
		  to these roots, and so they can't be removed in the scavenge."
		referrer := rememberedSet at: sourceIndex.
		"Any potential firing ephemerons should not be scanned yet.
		 Move any to the front of the set to save time in later scanning."
		((manager isEphemeron: referrer)
		 and: [(self isScavengeSurvivor: ((manager keyOfEphemeron: referrer))) not])
			ifTrue:
				[self assert: destIndex >= numRememberedEphemerons.
				 rememberedSet
					at: destIndex put: (rememberedSet at: numRememberedEphemerons);
					at: numRememberedEphemerons put: referrer.
				 numRememberedEphemerons := numRememberedEphemerons + 1.
				 destIndex := destIndex + 1]
			ifFalse:
				[(self scavengeReferentsOf: referrer)
					ifTrue:
						[rememberedSet at: destIndex put: referrer.
						 destIndex := destIndex + 1]
					ifFalse:
						[manager setIsRememberedOf: referrer to: false]].
		 sourceIndex := sourceIndex + 1].
	rememberedSetSize := destIndex.
	self assert: self noUnfiredEphemeronsAtEndOfRememberedSet! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 9/29/2013 11:43'!
scavengeUnfiredEphemeronsInRememberedSet
	"There are ephemerons to be scavenged in the remembered set.
	 Scavenge any with unfired (live) keys, removing them from the
	 unscavenged ephemerons, and from the set if they no longer refer
	 to new objects, and answer if any with unfired keys were found."
	| unfiredEphemeronsScavenged i referrer hasNewReferents |
	unfiredEphemeronsScavenged := false.
	i := 0.
	[i < numRememberedEphemerons] whileTrue:
		[referrer := rememberedSet at: i.
		 self assert: (manager isEphemeron: referrer).
		 (self isScavengeSurvivor: (manager keyOfEphemeron: referrer))
			ifTrue:
				[unfiredEphemeronsScavenged := true.
				 hasNewReferents := self scavengeReferentsOf: referrer.
				 "remove from unscanned ephemerons in set by swapping with last ephemeron"
				 numRememberedEphemerons := numRememberedEphemerons - 1.
				 rememberedSet
					at: i
					put: (rememberedSet at: numRememberedEphemerons).
				 hasNewReferents
					ifTrue: "keep in set"
						[rememberedSet
							at: numRememberedEphemerons
							put: referrer]
					ifFalse: "remove from set by overwriting with next-to-be scanned"
						[previousRememberedSetSize := previousRememberedSetSize - 1.
						 rememberedSetSize := rememberedSetSize - 1.
						 rememberedSet
							at: numRememberedEphemerons
								put: (rememberedSet at: previousRememberedSetSize);
							at: previousRememberedSetSize
								put: (rememberedSet at: rememberedSetSize).
						 manager setIsRememberedOf: referrer to: false]]
				ifFalse:
					[i := i + 1]].
	^unfiredEphemeronsScavenged! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 11/15/2019 08:59'!
scavengeUnfiredEphemeronsOnEphemeronList
	"There may be ephemerons to be scavenged on the ephemeronList.
	 Scavenge any with unfired (live) keys, removing them from the
	 list, and answer if any with unfired keys were found."
	| unfiredEphemeronsScavenged corpseOffset previousCorpse |
	ephemeronList ifNil:
		[^false].
	unfiredEphemeronsScavenged := false.
	corpseOffset := ephemeronList.
	[corpseOffset ~= 0] whileTrue:
		[| ephemeronCorpse ephemeron offsetToNextCorpse  |
		 ephemeronCorpse := self corpseForCorpseOffset: corpseOffset.
		 self assert: (manager isForwarded: ephemeronCorpse).
		 ephemeron := manager followForwarded: ephemeronCorpse.
		 offsetToNextCorpse := self nextCorpseOffset: ephemeronCorpse.
		 (self isScavengeSurvivor: (manager keyOfEphemeron: ephemeron))
			ifTrue:
				[corpseOffset = ephemeronList
					ifTrue: [ephemeronList := offsetToNextCorpse ~= 0 ifTrue: [offsetToNextCorpse]]
					ifFalse: [self setCorpseOffsetOf: previousCorpse to: offsetToNextCorpse].
				 unfiredEphemeronsScavenged := true.
				 self cCoerceSimple: (self scavengeReferentsOf: ephemeron) to: #void]
			ifFalse:
				[previousCorpse := ephemeronCorpse].
		 corpseOffset := offsetToNextCorpse].
	^unfiredEphemeronsScavenged! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 10/12/2013 14:39'!
scavengerTenuringThreshold "(Slang flattens so need unique selectors)"
	<returnTypeC: #float>
	^tenureThreshold >= pastSpace start
		ifTrue: [(tenureThreshold - pastSpace start) asFloat / (pastSpace limit - pastSpace start)]
		ifFalse: [0]! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'nice 7/20/2016 17:29'!
scavengerTenuringThreshold: aProportion "(Slang flattens so need unique selectors)"
	<var: 'aProportion' type: #double>
	tenuringProportion := aProportion.
	tenureThreshold := aProportion = 0.0
							ifTrue: [0]
							ifFalse: [((pastSpace limit - pastSpace start) * (1.0 - aProportion)) rounded + pastSpace start]! !

!SpurGenerationScavenger methodsFor: 'weakness and ephemerality' stamp: 'eem 5/27/2015 14:46'!
setCorpseOffsetOf: corpse to: offset
	"Set the offset of the corpse's next corpse to offset.  Use the identityHash
	 and format fields to construct a 27 bit offset through non-future newSpace
	 and use this to implement the list.  27 bits of 8 byte allocationUnits units is
	 2 ^ 30 bytes or 1Gb, big enough for newSpace for a good few years yet."
	<inline: true>
	self assert: (manager isYoung: corpse).
	self assert: (manager isForwarded: corpse).
	manager wordSize = 4
		ifTrue: "On 64-bit gcc 4.4.7-4 the following is miscompiled at -O2; the first assignment is lost"
			[manager
				setHashBitsOf: corpse
					to: offset >> manager formatFieldWidthShift;
				setFormatOf: corpse
					to: (offset bitAnd: manager formatMask)]
		ifFalse: "So we use the single assignment"
			[manager
				long64At: corpse
				put: (manager
						headerForSlots: (manager rawNumSlotsOf: corpse)
						hash: (offset >> manager formatFieldWidthShift)
						format: (offset bitAnd: manager formatMask)
						classIndex: manager isForwardedObjectClassIndexPun)]! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 10/12/2013 14:40'!
setRawTenuringThreshold: threshold
	tenureThreshold := threshold! !

!SpurGenerationScavenger methodsFor: 'remembered set' stamp: 'eem 5/6/2015 10:20'!
setRefCountToShrinkRT: population
	"Choose a refCount that should shrink the rt by at least  half.
	 i.e. find the maximum reference count that half the population have at least."
	<var: 'population' declareC: 'long population[MaxRTRefCount + 1]'>
	<inline: true>
	| entirePopulation i count |
	self assert: (population at: 0) = 0.
	entirePopulation := 0.
	1 to: MaxRTRefCount do:
		[:j| entirePopulation := entirePopulation + (population at: j)].
	count := 0.
	i := MaxRTRefCount + 1.
	[count < (entirePopulation // 2) and: [(i := i - 1) >= 0]] whileTrue:
		[count := count + (population at: i)].
	refCountToShrinkRT := i max: 0! !

!SpurGenerationScavenger methodsFor: 'remembered set' stamp: 'eem 5/5/2015 19:01'!
setRememberedSetRedZone
	| fudge |
	<inline: true>
	"fudge is a minimum below which we don't care about growing the RT.
	 It is chosen so that with a default 4Mb new space, the RT is not considered
	 full until it has from 512 to 1024 entries."
	fudge := eden limit - eden start / manager wordSize // 1024.
	rememberedSetRedZone := rememberedSetLimit * 3 // 4 max: fudge! !

!SpurGenerationScavenger methodsFor: 'scavenger' stamp: 'eem 5/5/2015 18:37'!
shouldBeTenured: survivor
	"Answer if an object should be tenured.  The default policy tenuring policy
	 is to use the tenuringThreshold to decide. If the survivors (measured in
	 bytes) are above some fraction of the survivor space then objects below
	 the threshold (older objects, since allocation grows upwards and hence
	 new objects are later than old) are scavenged.  Otherwise, the threshold
	 is set to 0 and no objects are tenured.  e.g. see
		An adaptive tenuring policy for generation scavengers,
		David Ungar & Frank Jackson.
		ACM TOPLAS, Volume 14 Issue 1, Jan. 1992, pp 1 - 27.

	 The other policies are for special purposes."
	^tenureCriterion
		caseOf: {
			[TenureByAge]	->
				[survivor < tenureThreshold]. 
			[TenureByClass] ->
				[(manager classIndexOf: survivor) = tenuringClassIndex].
			[TenureToShrinkRT]	->
				[(manager rtRefCountOf: survivor) >= refCountToShrinkRT]  }
		otherwise: [false]! !

!SpurGenerationScavenger methodsFor: 'scavenger' stamp: 'eem 5/26/2016 08:54'!
shouldMournerBeTenured: survivor
	"Answer if an object in the mourn queue should be tenured.  If we're
	 in the good times tenuring regime then copy to futureSpace, otherwise tenure."
	^(tenureCriterion = TenureByAge and: [tenureThreshold = 0]) not! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 1/25/2016 21:06'!
statSurvivorCount
	^statSurvivorCount! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 10/25/2013 16:39'!
statTenures
	^statTenures! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 6/15/2017 14:02'!
statTenures: aValue
	statTenures := aValue! !

!SpurGenerationScavenger methodsFor: 'remembered set' stamp: 'eem 5/5/2015 20:01'!
strategizeToLimitRememberedTable
	<inline: true>
	(tenureCriterion = TenureByAge
	 and: [rememberedSetSize >= rememberedSetRedZone]) ifTrue:
		[tenureCriterion := TenureToShrinkRT.
		 self computeRefCountToShrinkRT]! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 10/3/2013 21:01'!
tenuringClassIndex
	^tenuringClassIndex! !

!SpurGenerationScavenger methodsFor: 'accessing' stamp: 'eem 10/3/2013 21:01'!
tenuringClassIndex: anInteger
	tenuringClassIndex := anInteger! !

!SpurGenerationScavenger methodsFor: 'logging' stamp: 'nice 12/27/2018 14:57'!
writeScavengeLog
	"Output the entire record."
	<inline: #never>
	| policyNames |
	<var: 'policyNames' declareC: 'static char *policyNames[] = {"", "by age", "by class", "to shrink rt", "don''t tenure", "mark on tenure"}'>
	self cCode: []
		inSmalltalk: [policyNames := CLiteralArray on: #('' 'by age' 'by class' 'to shrink rt' 'don''t tenure' 'mark on tenure')].
	scavengeLog "log data collected by logStartScavenge"
		f: 'scavenge %ld eden bytes: 0x%lx/%ld past bytes: 0x%lx/%ld\n\trem set: %ld redzone: %ld size: %ld\n'
		printf:{	manager statScavenges.
				scavengeLogRecord sEdenBytes.
				scavengeLogRecord sEdenBytes.
				scavengeLogRecord sPastBytes.
				scavengeLogRecord sPastBytes.
				scavengeLogRecord sRememberedSetSize.
				scavengeLogRecord sRememberedSetRedZone.
				scavengeLogRecord sRememberedSetLimit }.
	scavengeLog "log data collected by logTenuringPolicy"
		f: (scavengeLogRecord tTenureCriterion = TenureToShrinkRT
				ifFalse: ['	tenure below 0x%lx/%ld %s\n']
				ifTrue: ['	tenure below 0x%lx/%ld %s refct %ld\n'])
		printf:{	scavengeLogRecord tTenureThreshold.
				scavengeLogRecord tTenureThreshold.
				policyNames at: scavengeLogRecord tTenureCriterion.
				scavengeLogRecord tRefCountToShrinkRT }.
	scavengeLog "log data collected by logEndScavenge"
		f: '	survivor bytes: 0x%lx/%ld rem set: %ld tenured: %ld usecs: %ld\n'
		printf:{	scavengeLogRecord eSurvivorBytes.
				scavengeLogRecord eSurvivorBytes.
				scavengeLogRecord eRememberedSetSize.
				scavengeLogRecord eStatTenures - scavengeLogRecord sStatTenures.
				manager statSGCDeltaUsecs}.
	scavengeLog fflush! !

SpurGenerationScavenger subclass: #SpurGenerationScavengerSimulator
	instanceVariableNames: 'cameFrom'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManagerSimulation'!

!SpurGenerationScavengerSimulator methodsFor: 'accessing' stamp: 'eem 8/13/2016 11:12'!
cameFrom

	^ cameFrom! !

!SpurGenerationScavengerSimulator methodsFor: 'accessing' stamp: 'eem 8/13/2016 11:12'!
cameFrom: anObject

	cameFrom := anObject! !

!SpurGenerationScavengerSimulator methodsFor: 'remembered set' stamp: 'eem 5/5/2015 19:18'!
computeRefCountToShrinkRT
	self halt.
	^super computeRefCountToShrinkRT! !

!SpurGenerationScavengerSimulator methodsFor: 'scavenger' stamp: 'eem 8/13/2016 11:11'!
copyAndForward: survivor
	| relocatedOop |
	relocatedOop := super copyAndForward: survivor.
	cameFrom ifNotNil:
		[cameFrom at: relocatedOop put: survivor].
	^relocatedOop! !

!SpurGenerationScavengerSimulator methodsFor: 'scavenger' stamp: 'eem 8/13/2016 11:11'!
copyAndForwardMourner: survivor
	| relocatedOop |
	relocatedOop := super copyAndForwardMourner: survivor.
	cameFrom ifNotNil:
		[cameFrom at: relocatedOop put: survivor].
	^relocatedOop! !

!SpurGenerationScavengerSimulator methodsFor: 'initialization' stamp: 'eem 10/9/2013 13:29'!
manager: aSpurNBitMMXEndianSimulator
	manager := aSpurNBitMMXEndianSimulator.
	aSpurNBitMMXEndianSimulator coInterpreter ifNotNil:
		[:coint| coInterpreter := coint]! !

!SpurGenerationScavengerSimulator methodsFor: 'scavenger' stamp: 'eem 8/13/2016 11:12'!
recordMovements
	cameFrom := Dictionary new! !

!SpurGenerationScavengerSimulator methodsFor: 'accessing' stamp: 'eem 6/8/2017 16:58'!
rememberedSet
	^rememberedSet! !

!SpurGenerationScavengerSimulator methodsFor: 'scavenger' stamp: 'eem 8/13/2016 11:13'!
scavenge: tenuringCriterion
	manager bootstrapping ifFalse:
		[coInterpreter transcript nextPutAll: 'scavenging('; print: manager statScavenges; nextPutAll: ')...'; flush.
		 cameFrom ifNotNil:
			[cameFrom := Dictionary new]].
	^super scavenge: tenuringCriterion! !

CogClass subclass: #SpurMemoryManager
	instanceVariableNames: 'coInterpreter scavenger segmentManager compactor memory freeStart freeOldSpaceStart scavengeThreshold newSpaceStart newSpaceLimit edenBytes oldSpaceStart nilObj falseObj trueObj specialObjectsOop hiddenRootsObj classTableFirstPage classTableIndex numClassTablePages endOfMemory mournQueue unscannedEphemerons objStackInvalidBecause invalidObjStackPage markStack weaklingStack freeLists freeListsMask lastHash signalLowSpace checkForLeaks needGCFlag heapMap becomeEffectsFlags pastSpaceStart heapGrowthToSizeGCRatio heapSizeAtPreviousGC oldSpaceUsePriorToScavenge growHeadroom shrinkThreshold marking remapBuffer remapBufferCount extraRootCount extraRoots lowSpaceThreshold totalFreeOldSpace maxOldSpaceSize gcPhaseInProgress gcStartUsecs gcMarkEndUsecs gcSweepEndUsecs compactionStartUsecs bogon statCoalesces statCompactPassCount statFGCDeltaUsecs statFullGCUsecs statCompactionUsecs statSweepUsecs statMarkUsecs statFullGCs statGCEndUsecs statGrowMemory statIGCDeltaUsecs statIncrGCUsecs statIncrGCs statMarkCount statRootTableCount statRootTableOverflows statSGCDeltaUsecs statScavengeGCUsecs statScavenges statShrinkMemory statAllocatedBytes statMaxAllocSegmentTime validatedIntegerClassFlags freeSpaceCheckOopToIgnore'
	classVariableNames: 'BitsPerByte ExtraRootsSize FirstValidClassIndex MarkObjectsForEnumerationPrimitives MarkStackRecord MarkStackRootIndex MournQueueRootIndex ObjStackFixedSlots ObjStackFreex ObjStackLimit ObjStackMyx ObjStackNextx ObjStackPageSlots ObjStackTopx RemapBufferSize RememberedSetRootIndex TopHashBit TopOopBit ValidatedClassDoubleByteArrayFlag ValidatedClassDoubleWordArrayFlag ValidatedClassFloat32ArrayFlag ValidatedClassFloat64ArrayFlag ValidatedClassWordArrayFlag WeaklingStackRootIndex'
	poolDictionaries: 'SpurMemoryManagementConstants VMBasicConstants VMBytecodeConstants VMObjectIndices VMSpurObjectRepresentationConstants VMSqueakClassIndices'
	category: 'VMMaker-SpurMemoryManager'!

!SpurMemoryManager commentStamp: 'eem 6/21/2021 08:54' prior: 0!
SpurMemoryManager is a new object representation and garbage collector for the Cog VM's.  Spur is dedicated to Andreas Raab, friend and colleague.  I miss you, Andreas.  The goal for Spur is an overall improvement in Cog of -50% (twice as fast) for memory-intensive benchmarks.

A detailed design sketch is included below after the instance variable descriptions.  The instance variable declarations can be checked for extra and missing declarations by making a literal array out of them and using
	(| ivs |
	ivs := (#(becomeEffectsFlags		<Integer> ...) piecesCutWhere: [:a :b| a = #>]) collect: [:ea| ea first].
	{ivs reject: [:iv| SpurMemoryManager instVarNames includes: iv].
	(SpurMemoryManager instVarNames reject: [:iv| ivs includes: iv]) sort})
The instance variable descriptions can be checked for extra and missing declarations by making a literal string out of them and using
	(| ivs |
	ivs := ('becomeEffectsFlags...' piecesCutWhere: [:a :b| a = Character cr]) select: [:s| s notEmpty and: [s first isLetter]] thenCollect: [:ea| ea withoutTrailingBlanks].
	{ivs reject: [:iv| SpurMemoryManager instVarNames includes: iv].
	(SpurMemoryManager instVarNames reject: [:iv| ivs includes: iv]) sort})

Instance Variables
	becomeEffectsFlags		<Integer>
	bogon						<Integer oop>
	checkForLeaks				<Boolean>
	classTableFirstPage		<Integer oop>
	classTableIndex			<Integer>
	coInterpreter				<StackInterpreter|CoInterpreter (StackInterpreterSimulator|CogVMSimulator)>
	compactionStartUsecs		<Integer>
	compactor					<subclass of SpurCompactor>
	edenBytes					<Integer>
	endOfMemory				<Integer address>
	mournQueue			<Integer objStack oop>
	extraRootCount			<Integer>
	extraRoots					<Array of: oop>
	falseObj					<Integer oop>
	freeLists					<CArray on: memory>
	freeListsMask				<Integer>
	freeOldSpaceStart			<Integer address>
	freeStart					<Integer address>
	gcStartUsecs				<Integer>
	growHeadroom				<Integer>
	heapMap					<CogCheck32BitHeapMap>
	hiddenRootsObj			<Integer oop>
	heapGrowthToSizeGCRatio	<Float>
	heapSizeAtPreviousGC		<Integer>
	invalidObjStackPage		<Integer oop|nil>
	lastHash					<Integer>
	lowSpaceThreshold		<Integer address>
	marking					<Boolean>
	markStack					<Integer objStack oop>
	maxOldSpaceSize			<Integer>
	memory					<Bitmap|LittleEndianBitmap>
	needGCFlag				<Boolean>
	newSpaceLimit				<Integer address>
	newSpaceStart				<Integer address>
	nilObj						<Integer oop>
	numClassTablePages		<Integer>
	objStackInvalidBecause	<String|nil>
	oldSpaceStart				<Integer address>
	oldSpaceUsePriorToScavenge	<Integer>
	pastSpaceStart				<Integer address>
	remapBuffer				<CArrayAccessor on: (Array new: RemapBufferSize)>
	remapBufferCount			<Integer>
	scavengeThreshold		<Integer address>
	scavenger					<SourGenerationScavenger>
	segmentManager			<SpurSegmentManager>
	signalLowSpace			<Boolean>
	shrinkThreshold			<Integer>
	specialObjectsOop			<Integer oop>
	statCoalesces				<Integer>
	statCompactPassCount		<Integer>
	statFGCDeltaUsecs			<Integer>
	statFullGCUsecs			<Integer>
	statFullGCs					<Integer>
	statGCEndUsecs			<Integer>
	statGrowMemory			<Integer>
	statIGCDeltaUsecs			<Integer>
	statIncrGCUsecs			<Integer>
	statIncrGCs					<Integer>
	statMarkCount				<Integer>
	statRootTableCount		<Integer>
	statRootTableOverflows	<Integer>
	statSGCDeltaUsecs		<Integer>
	statScavengeGCUsecs		<Integer>
	statScavenges				<Integer>
	statShrinkMemory			<Integer>
	totalFreeOldSpace			<Integer>
	trueObj						<Integer oop>
	unscannedEphemerons		<SpurContiguousObjStack>
	weaklingStack				<Integer objStack oop>

becomeEffectsFlags
	- a set of flags to limit the work done during become; one of BecameCompiledMethodFlag, BecamePointerObjectFlag

bogon
	- an oop that can be checked for.  Currently allObjectsUnmarked sets it to the first marked object found.

checkForLeaks
	- a set of flags determining when to run leak checks on the heap's contents

classTableFirstPage
	- the first page of the class table which contains all classes known to the VM at known indices

classTableIndex
	- the last known used index in the class table, used to limit the search for empty slots

coInterpreter
	- the VM interpreter using this heap

compactionStartUsecs
	- the microsecond clock value immediatel;y before a cmpaction p[hase starts.  For primitiveVMParameter reporting of time spent in the mark and compaction phases of full GC.

compactor
	- the compactor that compacts oldSpace

edenBytes
	- a variable holding the size of eden at start up only

endOfMemory
	- the address past the last oldSpace segment

mournQueue
	- the oopStack holding triggered ephemerons, and, depending on finalization scheme, bereaved weak arrays

extraRootCount
	- the count of extra roots

extraRoots
	- an array of pointers to variables holding oops that need updating.  used by plugins to update objects they hold on to.

falseObj
	- the oop of the false object; must be the second object in oldSpace

freeLists
	- an array of list heads of free chunks of each size (in allocationUnits).  the 0'th element
	  is the head of a tree of free chunks of size > 63 slots in 64-bits or 31 slots in 32-bits.
	  freeLists is the firstIndexableField of the fourth object in oldSpace.

freeListsMask
	- a bitmap of flags with a 1 set whereever the freeLists has a non-empty entry, used to limit the search for free chunks

freeOldSpaceStart
	- the last used address in oldSpace.  The space between freeOldSpaceStart and endOfMemory can be used for oldSpace allocations.
	  Given freeLists this probably isn't useful and shoud be discarded in favour of endOfMemory.

freeStart
	- the last used address in eden.  this is where new objects are allocated if there is room in eden.

gcPhaseInProgress
	- flag taking values 0, ScavengeInProgress, SlidingCompactionInProgress, for remapObj: to decide what to do.

gcStartUsecs
	- the time in microseconds of the start of the most recent gc, used to compute the duration of a gc.

growHeadroom
	- the minimum ammount of space to grow memory by when a grow is required.

heapMap
	- a bitmap used to check for leaks. a bit is set in the map at each object's header.  every pointer in the heap should point to the header of an object

heapGrowthToSizeGCRatio
	- determines how much heap growth since heapSizeAtPreviousGC at which a full GC will be performed after a scavenge.  This is a stop gap measure running the full GC occasionally until we have implemented an incremental scan-mark-compact GC [see Incremental Old Space Collection below].

heapSizeAtPreviousGC
	- the size of the heap at the most recent GC, used to decide if a full GC is needed after a scavenge.  See heapGrowthToSizeGCRatio.

hiddenRootsObj
	- the root page of the class table and some other roots (e.g. the mourn queue). its first
	  numClassTablePages elements are pages containing classes. Remaining slots hold other
	  roots the VM needs.  Must be the fifth object in oldSpace.

invalidObjStackPage
	- the offending objStack page, if isValidObjStack: failed, or nil. This is for debugging.

lastHash
	- the last object hash value.  the seed to the pseudo-random number generator for identity hashes.

lowSpaceThreshold
	- the ammount of free memory below which the low space condition should be signalled.

marking
	- a flag set and cleared in markAccessibleObjects to allow ensureRoomOnObjStackAtIndex: to know whether to mark a new obj stack page or not.

markStack
	- the objStack that houses the mark stack.

maxOldSpaceSize
	- the maximum allowed size of old space in bytes.  If zero, there is no limit.

memory
	- in the Simulator this is the single oldSpace segment that contains the codeZone, newSpace and oldSpace.  In C it is effectively the base address of used memory.

needGCFlag
	- a boolean flag set if a scavenge is needed

newSpaceStart
	- the first address in newSpace (the start address of the first of the two survivor spaces, past and future space)

newSpaceLimit
	- the last address in newSpace (the last address in eden)

nilObj
	- the oop of the nil object; must be the first object in oldSpace

numClassTablePages
	- the number of used pages in the classTable

objStackInvalidBecause
	- a string identifying the reason isValidObjStack: failed, or nil.  This is for debugging.

oldSpaceStart
	- the start of oldSpace, which happens to be nilObj.

oldSpaceUsePriorToScavenge
	- a count of oldSpace occupation prior to a scavenge, used to maintain an accurate byte allocation count from start-up.

pastSpaceStart
	- the address past the last object in pastSpace, used to enumerate objects in newSpace.

remapBuffer
	- an Array of objects used to implement pushRemappableOop: et al for compatibility with ObjectMemory.
	  Its functionality isnt needed because Spur will never GC during allocation.  But it muts be present and
	  correct, otherwise many primitives that use pushRemappableOop: et al would have to be rewritten.

remapBufferCount
	- the index of the last used entry in remampBuffer

scavengeThreshold
	- a tidemark in eden.  needGCFlag is set if a newSpace allocation pushes freeStart past scavengeThreshold

scavenger
	- the generation scavenger that collects objects in newSpace

segmentManager
	- the object that manages oldSpace segments.  Segments are largely invisible to the memory manager because
	  the gaps between segments are hidden by bridge objects.

shrinkThreshold
	- the amount of free oldSpace above which the manager should attempt to give memory back to the OS.

signalLowSpace
	- a boolean flag set if the lowSpaceSemaphore should be signalled

specialObjectsOop
	- the oop of the specialObjectsArray object

statCoalesces statCompactPassCount statFGCDeltaUsecs statFullGCUsecs statFullGCs statGCEndUsecs statGrowMemory statIGCDeltaUsecs statIncrGCUsecs statIncrGCs statMarkCount statRootTableCount statRootTableOverflows statSGCDeltaUsecs statScavengeGCUsecs statScavenges statShrinkMemory statSpecialMarkCount statSurvivorCount
	- various statistics

totalFreeOldSpace
	- the total free space on the free lists
	
trueObj
	- the oop of the false object; must be the third object in oldSpace

unscannedEphemerons
	- the set of ephemerons yet to be scanned.  used for identifying firable ephemerons during gc.

weaklingStack
	- the set of weak collections that must be scanned later on in gc to nil their collected referents.



Invariants (far from an exhaustive list; I'm adding to this as they arise; please prompt me to add undocumented ones):

On image load no object contains a forwarding pointer, and the image contains no forwarders. True because
a) the SpurBootstrap eliminates forwarders before saving its transformed image
b) snapshot does a fullGC which will follow forwarding pointers.
This is checked with an assert in swizzleFieldsOfObject:/lastPointerOfWhileSwizzling:.

classTableIndex ranges from classTablePageSize to classTableMask, since the first page is reserved for system classes, and the VM won't assign classes there-in.

Design
The design objectives for the Spur memory manager are

- efficient object representation a la Eliot Miranda's VisualWorks 64-bit object representation that uses a 64-bit header, eliminating direct class references so that all objects refer to their classes indirectly.  Instead the header contains a constant class index, in a field smaller than a full pointer, These class indices are used in inline and first-level method caches, hence they do not have to be updated on GC (although they do have to be traced to be able to GC classes).  Classes are held in a sparse strong table.  The class table needs only to be indexed by an instance's class index in class hierarchy search, in the class primitive, and in tracing live objects in the heap.  The additional header space is allocated to a much expanded identity hash field, reducing hash efficiency problems in identity collections due to the extremely small (11 bit) hash field in the old Squeak GC.  The identity hash field is also a key element of the class index scheme.  A class's identity hash is its index into the class table, so to create an instance of a class one merely copies its identity hash into the class index field of the new instance.  This implies that when classes gain their identity hash they are entered into the class table and their identity hash is that of a previously unused index in the table.  It also implies that there is a maximum number of classes in the table.  The classIndex field could be as narrow as 16 bits (for efficient access); at least for a few years 64k classes should be enough.  But currently we make it the same as the identityHash field, 22 bits, or 4M values.  A class is entered into the class table in the following operations:
	behaviorHash
	adoptInstance
	instantiate
	become  (i.e. if an old class becomes a new class)
		if target class field's = to original's id hash
		   and replacement's id hash is zero
			enter replacement in class table
behaviorHash is a special version of identityHash that must be implemented in the image by any object that can function as a class (i.e. Behavior).

The object representaion must also be simple enough to allow more operations (especially array at:put: and object instantiation to be implemented by the JIT in machine code.  The current Squea object representtion with three different header sizes and two different size formats has discouraged this.

- substantial commonality between 32-bit and 64-bit Implemetations
While the 32-bit and 64-bit implementations are different (32-bits per slot vs 64-bits per slot), the fact that the object header is 64-bit and common between them and that object lengths are rounded up to 64-bits means that most of the memory manager code is common.  As of 9/18/2014:

	SpurMemoryManager selectors size 765
	(SpurMemoryManager selectors reject: [:s| (SpurMemoryManager >> s) messages asArray = #(subclassResponsibility)]) size 721
	Spur32BitMemoryManager selectors size 46
	Spur64BitMemoryManager selectors size 44 (two selectors remaining to be implemented)

In fact the object header looks like

	 MSB:	| 8: numSlots		| (on a byte boundary)
			| 2 bits				|	(msb,lsb = {isMarked,?})
			| 22: identityHash	| (on a word boundary)
			| 3 bits				|	(msb <-> lsb = {isGrey,isPinned,isRemembered}
			| 5: format			| (on a byte boundary)
			| 2 bits				|	(msb,lsb = {isImmutable,?})
			| 22: classIndex		| (on a word boundary) : LSB
					
with the two unused bits being next to the 22-bit classIndex and identityHash fields, allowing expansion there-of.


- more immediate classes.  An immediate Character class would speed up String accessing, especially for WideString, since no instantiation needs to be done on at:put: and no dereference need be done on at:.  In a 32-bit system tag checking is complex since it is thought important to retain 31-bit SmallIntegers.  Hence, as in current Squeak, the least significant bit set implies a SmallInteger, but Characters would likely have a tag pattern of xxx10.  Hence masking with 11 results in two values for SmallInteger, xxx01 and xxx11 (for details see In-line cache probe for immediates below).  30-bit characters are more than adequate for Unicode.  In a 64-bit system we can use the full three bits and usefully implement an immediate Float.  As in VisualWorks a functional representation takes three bits away from the exponent.  Rotating to put the sign bit in the least significant non-tag bit makes expanding and contracting the 8-bit exponent to the 11-bit IEEE double exponent easy and makes comparing negative and positive zero easier (an immediate Float is zero if its unsigned 64-bits are < 16).  So the representation looks like
	| 8 bit exponent | 52 bit mantissa | sign bit | 3 tag bits |
For details see "60-bit immediate Floats" below.


- efficient scavenging.  The current Squeak GC uses a slow pointer-reversal collector that writes every field in live objects three times in each collection, twice in the pointer-reversing heap traversal to mark live objects and once to update the pointer to its new location.  A scavenger writes every field of live data twice in each collection, once as it does a block copy of the object when copying to to space, once as it traverses the live pointers in the to space objects.  Of course the block copy is a relatively cheap write.

- lazy become.  Squeak uses direct pointers and Spur retains them.  The JIT's use of inline cacheing provides a cheap way of avoiding scanning the heap as part of a become (which is the simple approach to implementing become in a system with direct pointers).  A becomeForward: on a (set of) non-zero-sized object(s) turns the object into a "forwarder" whose first (non-header) word/slot is replaced by a pointer to the target of the becomeForward:.  The forwarder's class index is set to one that identifies fowarders (in fact classIndex 8), and, because it is a special, hidden class index, will always fail an inline cache test.  The inline cache failure code is then responsible for following the forwarding pointer chain (these are Iliffe vectors :) ) and resolving to the actual target.  The system has two main faults, that for a forwarded selector and that for a forwarded receiver.  For the former we want to unforward the current method and for the latter we want to unforward the current frame  See become read barrier below on how we deal with becomes on objects with named inst vars.  We insist that objects are at least 16 bytes in size (see 8-byte alignment below) so that there will always be space for a forwarding pointer.  Since none of the immediate classes can have non-immediate instances and since we allocate the immediate class indices corresponding to their tag pattern (in 32-bits, SmallInteger = 1 & 3, Character = 2; in 64-bits, SmallInteger = , Character = 2, SmallFloat = 4) we can use all the class indices from 0 to 7 for special uses, 0 = free, and e.g. 1 = isForwarded.  In general what's going on here is the implemention of a partial read barrier. Certain operations require a read barrier to ensure access of the target of the forwarder, not the forwarder itself.  Read barriers stink (have poor performance), so we must restrict the read barrier to as few places as possible.  See become read barrier below.  See http://www.mirandabanda.org/cogblog/2013/09/13/lazy-become-and-a-partial-read-barrier/ & http://www.mirandabanda.org/cogblog/2014/02/08/primitives-and-the-partial-read-barrier/.  But we avoid the cost of a read barrier in almost all cases by folding that forwarding check into a dynamic type check that is performwed anyway (sends) or a validation check (primitives checking their arguments).  In the case of a send, or a primitive, the failure handling code checks for forwarders, follow any if found, and if any are found, retries either the send or the primitive.

- pinning.  To support a robust and easy-to-use FFI the memory manager must support temporary pinning where individual objects can be prevented from being moved by the GC for as long as required, either by being one of an in-progress FFI call's arguments, or by having pinning asserted by a primitive (allowing objects to be passed to external code that retains a reference to the object after returning).  Pinning probably implies a per-object "is-pinned" bit in the object header.  Objects are only pinnable in old space.  Pinning of new space objects is done via lazy become; i..e an object in new space is forwarded to a pinned copy of the object in old space.

- efficient old space collection.  An incremental collector (a la Dijkstra's three colour algorithm) collects old space, e.g. via an amount of tracing being hung off scavenges and/or old space allocations at an adaptive rate that keeps full garbage collections to a minimum.  It is also possible to provide cheap compaction by using lazy become: and best-fit (see free space/free list below). [see Incremental Old Space Collection below].

- 8-byte alignment.  It is advantageous for the FFI, for floating-point access, for object movement and for 32/64-bit compatibility to keep object sizes in units of 8 bytes.  For the FFI, 8-byte alignment means passing objects to code that expects that requirement (such as modern x86 numeric processing instructions).  This implies that
	- the starts of all spaces are aligned on 8-byte boundaries
	- object allocation rounds up the requested size to a multiple of 8 bytes
	- the overflow size field is also 8 bytes
But 8-byte alignment is not necessarily sufficient.  See extra alignment above 8-bytes below.

- heap segmentation. This allows incremental heap growth (the current Squeak VM requires a contiguous chunk of memory, meaning an unfortunate trade off between space for objects and space for external code), and returning memory to the OS after GC.

- no GC in any allocation.  In V3, the previous memory manager, any allocation other than the eeInstantiate... API I introduced for "execution engine" allocations, may cause a GC.  Consequently primitive smust use the clumsy and inefficient remapBuffer to store oops around allocations.  In Spur, no allocation can cause a GC, consequently primitives need not waste time or effort with the remapBuffer.  The primitive retry machinery is used to retry prmitives that fail with a PrimErrNoMemory error after a GC.  Currently we do this for plugin primitives, not core numbered primitives; the core numbered primities that do alloate are specifically designed to fail on allocation failure, allowing image-level code to control running the GC.
	

Object Size
We keep the minimum object size at 16 bytes so that there is always room for a forwarding pointer.  This implies we cannot support free chunks of 8 bytes in length and that the memory manager refuses to split objects when doing so would leave an 8-byte sliver.  We can do this using a special class index, e.g. 1, so that the method that answers the size of an object looks like

	rawNumSlotsOf: objOop
		^self byteAt: objOop + 7

	rawOverflowSlotsOf: objOop
		^self long32At: objOop - self baseHeaderSize

	numSlotsOf: objOop
		| numSlots |
		numSlots := self rawNumSlotsOf: objOop..
		^numSlots = self numSlotsMask	"overflow slots; (2^32)-1 slots are plenty"
			ifTrue: [self rawOverflowSlotsOf: objOop]
			ifFalse: [numSlots]

Note that the size field of an object (its slot size) reflects the logical size of the object e.g. 0-element array => 0 slot size, 1-element array => 1 slot size). The memory manager rounds up the slot size as appropriate, e.g. (self roundUp: (self slotSizeOf: obj) * 4 to: 8) min: 8 to obtain a byte size.

	bytesInObject: objOop
		| headerNumSlots numSlots |
		headerNumSlots := self rawNumSlotsOf: objOop.
		numSlots := headerNumSlots = self numSlotsMask
						ifTrue: [self rawOverflowSlotsOf: objOop]
						ifFalse: [headerNumSlots = 0 ifTrue: [1] ifFalse: [headerNumSlots]].
		^numSlots + (numSlots bitAnd: 1) << self shiftForWord
		+ (headerNumSlots = self numSlotsMask
			ifTrue: [self baseHeaderSize + self baseHeaderSize]
			ifFalse: [self baseHeaderSize])

Heap growth and shrinkage is handled by allocating and deallocating heap segments from/to the OS via e.g. memory-mapping.  This technique allows space to be released back to the OS by unmapping empty segments.  See "Segmented Old Space" below).

The basic approach is to use a fixed size new space and a segmented, growable old space.  The new space is a classic three-space nursery a la Ungar's Generation Scavenging, a large eden for new objects and two smaller survivor spaces that exchange roles on each collection, one being the to space to which surviving objects are copied, the other being the from space of the survivors of the previous collection, i.e. the previous to space.  (This basic algorithm is extended to handle weak arrays and ephemerons).

To provide apparent pinning in new space we rely on lazy become.  Since most pinned objects will be byte data and these do not require stack zone activation scanning, the overhead is simply an old space allocation and forwarding.

To provide pinning in old space, objects hav their "is pinned" bit set in the object header (which could be set automaticaly when allocating a large object).  Compaction moves objects from high memory down to low memory using forwarders, and stepping over pinned objects (see Compact below).

Free space in old space is organized by a number of free lists and a free tree.  There are 32 or 64 free lists, depending on word size, indices 1 through wordSize - 1 holding blocks of space of the index * allocationUnit, index 0 holding a semi-balanced ordered tree of free blocks, each node being the head of the list of free blocks of that size.  At the start of the mark phase the free list is thrown away and the sweep phase coalesces free space and reclaimed forwarders, stepping over pinned objects as it proceeds.  We can reuse the forwarding pointer compaction scheme used in the old collector for incremental collections.  Incremental collections merely move unmarked objects to the free lists (as well as nilling weak pointers in weak arrays and scheduling them for finalization).  The occupancy of the free lists is represented by a bitmap in an integer so that an allocation of size wordSize - 1 or less can know whether there exists a free chunk of that size, but more importantly can know whether a free chunk larger than it exists in the fixed size free lists without having to search all larger free list heads.

Incremental Old Space Collection
The (as yet unimplemented) incremental collector (a la Dijkstra's three colour algorithm) collects old space via an amount of tracing being hung off scavenges and/or old space allocations at an adaptive rate that keeps full garbage collections to a minimum.  [N.B. Not sure how to do this yet.  The incremental collector needs to complete a pass often enough to reclaim objects, but infrequent enough not to waste time.  So some form of feedback should work.  In VisualWorks tracing is broken into quanta or work where image-level code determines the size of a quantum based on how fast the machine is, and how big the heap is.  This code could easily live in the VM, controllable through vmParameterAt:put:.  An alternative would be to use the heartbeat to bound quanta by time.  But in any case some amount of incremental collection would be done on old space allocation and scavenging, the ammount being chosen to keep pause times acceptably short, and at a rate to reclaim old space before a full GC is required, i.e. at a rate proportional to the growth in old space]. The incremental collector compacts memory by exploiting lazy forwarding. [This has been implemented as a prototype and written up in Clément Béra, Eliot Miranda, and Elisa Gonzalez Boix. "Lazy Pointer Update for Low Heap Compaction Pause Times." In Proceedings of the 15th ACM SIGPLAN International Symposium on Dynamic Languages, 15-27. DLS '19. ACM, 2019. https://doi.org/10.1145/3359619.3359741, principally Clément Béra].

The incemental collector is a state machine, being either marking, nilling weak pointers, or freeing.  If nilling weak pointers is not done atomically then there must be a read barrier in weak array at: so that reading from an old space weak array that is holding stale un-nilled references to unmarked objects.  Tricks such as including the weak bit in bounds calculations can make this cheap for non-weak arrays.  Alternatively nilling weak pointers can be made an atomic part of incremental collection, which can be made cheaper by maintaining the set of weak arrays (e.g. on a list).  Note that the incremental collector also follows (and eliminates) forwarding pointers as it scans.

The incremental collector implies a more complex write barrier.  Objects are of three colours, black, having been scanned, grey, being scanned, and white, unreached.  A mark stack holds the grey objects.   If the incremental collector is marking and an unmarked white object is stored into a black object then the stored object must become grey, being added to the mark stack.  So the wrte barrier is essentially
	target isYoung ifFalse:
		[newValue isYoung
			ifTrue: [target isInRememberedSet ifFalse:
					[target addToRememberedSet]] "target now refers to a young object; it is a root for scavenges"
			ifFalse:
				[(target isBlack
				  and: [igc marking
				  and: [newValue isWhite]]) ifTrue:
					[newValue beGrey]]] "add newValue to IGC's markStack for subsequent scanning"

The incremental collector does not detect already marked objects all of whose references have been overwritten by other stores (e.g. in the above if newValue overwrites the sole remaining reference to a marked object).  So the incremental collector only guarantees to collect all garbage created in cycle N at the end of cycle N + 1.  The cost is hence slightly worse memory density but the benefit, provided the IGC works hard enough, is the elimination of long pauses due to full garbage collections, which become actions of last resort or programmer desire.

Compaction:
The main issue with compaction is handling pinned objects.  Our first two attempts, Incremental Best-Fit Compaction and Pig Compact, performed abysmally.  The current compaction algorithms are SpurPlanningCompactor, and SpurSelectiveCompactor.  SpurSelectiveCompactor is the prototype described in "Lazy Pointer Update for Low Heap Compaction Pause Times.".

Incremental Best-Fit Compaction
[A Failure]
The free list also looks like it allows efficient incremental compaction.  Currently in the 32-bit implementation, but easily affordable in the 64-bit implementation, objects have at least two fields, the first one being a forwarding pointer, the second one rounding up to 8-byte object alignment.  On the free list the first field is used to link the list of free chunks of a given size.  The second field could be used to link free chunks in memory order.  And of course the last free chunk is the chunk before the last run of non-free objects.  We compact by

a) keeping each free list in memory order (including the lists of free chunks off each node in the large free chunk tree)
b) sorting the free chunks in memory order by merge sorting the free lists
c) climbing the free list in memory order.  For each free chunk in the free list search memory from the last free chunk to the end (and from the previous chunk to the next chunk, and so on) looking for a best-fit live object.  That object is then copied into the free chunk, and its corpse is turned into a forwarding pointer.  This works because the compaction does not slide objects, and hence no forwarding blocks are necessary and the algorithm can be made incremental. Various optimizations are possible, e.g. using a bitmap to record the sizes of the first few free chunks on the list when looking for best fits.  The assumptions being
a) the number of objects on the free list is kept small because the IGC incrementally compacts, and so sorting and searching the list is not expensive
b) the incremental collector's following of forwarding pointers reclaims the corpses at the end of memory at a sufficient rate to keep the free list small
c) the rounding of objects to an 8-byte alignment improves the chances of finding a best fit.
Note that this incremental collection is easily amenable to leave pinned objects where they are; they are simply filtered out when looking for a best fit.

Pig Compact
[Another failure]
Pig Compact was an eager algorithm that moved as many objects for the top of memory to the bottom of memory as possible.  We called this "Pig" compact, and it was used as the compactor for global mark/sweep, used at least on snapshot.  Free memory was linked in memory ascending order through the first field, using the xor pointer trick to implement a doubly-linked list through a single field.  [Note this design was chosen for the benefit of the 64-bit image where there really is only one spare field.  In the 32-bit system there are at least two fields and the code could be simplified.].  With the doubly-linked list we can access objects at the top of memory by scanning up from the end of the list, and can move objects down into the free space at the front of the list, forwarding objects to their new locations.  Pinned objects are simply left where they are.  After a single pass there will be many forwarders at the top of memory and it will not have been compacted very much.  Memory is then swept to follow forwarders, and then again to free forwarders and coallesce free space.  The entire process is repeated (currently 3 times) to effectively compact memory.  Performance is dreadful.

Planning Compaction
This is an efficient algorithm for full GC, especially snapshot, where we want to fully compact the heap.  See SpurPlanningCompactor.  It uses the fact that there is room for a forwarding pointer in all objects to store the eventual position of an object in the first field.  It therefore first locates a large free chunk, or eden or a memory segment, and uses this to store the first fields of objects that will be compacted.  It then makes at least three passes through the heap.  The first pass plans where live movable objects will go, copying their first field to the next slot in savedFirstFieldsSpace, and setting their forwarding pointer to point to their eventual location.  The second pass updates all pointers in live pointer objects to point to objects' final destinations, including the fields in savedFirstFieldsSpace.  The third pass moves objects to their final positions, unmarking objects as it does so.  If the forwarding fields of live objects in the to-be-moved portion of the entire heap won't fit in savedFirstFieldsSpace, then additional passes are made until the entire heap has been compacted.  Each pass uses a three finger algorithm, a simple extension of the classic two finger algorithm with an extra finger used to identify the lowest pinned object between the to and from fingers.  Objects are moved down, starting at the first free object or chunk, provided that they fit below the lowest pinned object above the to finger.  When an object won't fit the to finger is moved above the pinned object and the third finger is reset to the next pinned object below the from finger, if any.

Incremental Compaction
An order of the segments is stored as a list, holding the segments in least occupied to most-occupied order.  An empty segment is obtained, or the most sparsely occupied segment is selected, and used as the target of incremental compaction. The lest occupied segments are scanned, and the objects in each sparsely occupied segment zare becomed into the target segment.  Compaction proceeds until some time quantum is reached or the target segment is fully occupied, or all sparsely occupied segments have had their objects forwarded to the target segment.  Given that scan-mark follows forwarders all the least occupied segments (except the target segment) will be fully vacated after the next cycle of incremental scan-mark, or a full GC.  See "Lazy Pointer Update for Low Heap Compaction Pause Times.", reference given above.

Segmented Old Space via Bridges
A segmented oldSpace is useful.  It allows growing oldSpace incrementally, adding a segment at a time, and freeing empty segments.  But such a scheme is likely to introduce complexity in object enumeration, and compaction (enumeration would apear to require visiting each segment, compaction must be within a segment, etc). Spur turns segmented old space into one that appears to be a single contiguous space by using fake pinned objects to bridge the gaps between segments.  We call these "bridges".  The last two words of each segment can be used to hold the header of a pinned object whose overflow size is the distance to the next segment.  The pinned object's classIndex is one of the puns so that it doesn't show up in allInstances or allObjects; this can perhaps also indicate to the incremental collector that it is not to reclaim the object, etc.  Bridge objects need a large enough overflow size field to stretch across large gaps in the address space.  This is no problem in the 32-bit system which has a 32-bit overflow slot coult, 4 times larger than necessary.  But the 64-bit sytsem only has a 56 bit overflow slot count, 5 bits short of spanning a full 64-bit address space.  We expect that this is adequate if we preferrentially allocate segments at one end of the address space or the other (in fact we try and allocate low).

This scheme has been expanded to enumerate objects in new space, using "slimbridges", shims between objects that can be one word long. [And the scheme could be extended to provide variable length slimbridges for spanning small gaps in object memory].  Depending on whether past and eden, or just eden objects need to be enumerated, and on whether past is full and the last object ends at the start of eden, and eden is full and the last object ends at the start of old space, a bridge, slim if needed, is inserted if required to bridge to the next object in the enumeration.  Now all objects can be enumerated in a single enumeration using the endOfMemory as the limit.


Lazy Become & the Partial Read Barrier
see http://www.mirandabanda.org/cogblog/2013/09/13/lazy-become-and-a-partial-read-barrier/ & http://www.mirandabanda.org/cogblog/2014/02/08/primitives-and-the-partial-read-barrier/.

As described earlier the basic idea behind lazy become is to use forwarding objects that are followed lazily during GC and inline cache miss (message send).  However, a lazy scheme would appear to require a read barrier to avoid accessing the forwarder and make sure the forwarding pointer is followed.  Without hardware support read barriers have poor performance (at least in code density), so we try and avoid the read barrier as much as possible.  The main goals are a) to avoid having to scan all of the heap to fix up pointers, as is done with ObjectMemory, and b) to avoid a read barrier on inst var read, class access, etc.  We're happy to do some scanning of a small subset of the heap, but become: cannot scale to large heaps if it must scan the entire heap.  Objects with named inst vars and CompiledMethods are accessed extensively by the interpreter and jitted code.  We must avoid as much checking of such accesses as possible; We judge an explicit read barrier on all accesses far too expensive.  The accesses the VM makes which notionally require a read barrier are:
- inst vars of thisContext, including stack contents (the receiver of a message and its arguments), which in Cog are the fields of the current stack frame, and the sender chain during (possibly non-local) return
- named inst vars of the receiver
- literals of the current method, in particular variable bindings (a.k.a. literal variables which are global associations), including the methodClass association.
- in primitives, the receiver and arguments, including possible sub-structure.
We have already discussed that we will catch an attempt to create a new activation on a forwarded object through method lookup failing for forwarded objects.  This would occur when e.g. some object referenced by the receiver via its inst vars is pushed on the stack as a message receiver, or e.g. answered as the result of some primitive which accesses object state such as primtive at:  So there is no need for a read barrier when accessing a new receiver or returning its state.  But there is a read barrier in primitives that inspect that sub-state.

We avoid a read barrier on access to receiver inst vars by scanning the stack zone (*) after a become and following pointers to the receiver. (* the Cog VM, following Peter Deutsch's PS Smalltalk VM, lazily maps stack frames to context objects, creating contexts as needed and using stack frames for all method activations.  The stack zone is small (320k default, of which at most half is used to execute Smalltalk, half being reserved for interrupt handlers running on the native stack) and organized as pages.  Context objects serve as proxy objects for stack frames in the stack zone, or as fully fledged activations when activations on a stack page are evicted to the heap as contexts to make room for new activations).

There is one wrinkle in this neat picture for sends that we call the "stale supersend" problem.  See the next section.

We can easily avoid read barriers in direct literal access, class hierarchy walking, and message dictionary searching in message lookup (but we do not do so, see next paragraph).  Whenever the become operation becomes one or more pointer objects (and e.g. it can know cheaply if it has becommed a CompiledMethod) both the class table and the stack zone could be scanned.  In the class table we can follow the forwarding pointers in all classes in the table, and we can follow their superclasses.  But we would like to avoid scanning classes many times.  Any superclass that has a non-zero hash must be in the table and will be encountered during the scan of the class table.  Any superclass with a zero hash can be entered into the table at a point subsequent to the current index and hence scanned.  The class scan can also scan method dictionary selectors and follow the method dictionary's array reference (so that dictionaries are valid for lookup) and scan the dictionary's method array iff a CompiledMethod was becommed. (Note that support for object-as-method implies an explicit read barrier in primitiveObjectAsMethod, which is a small overhead there-in).

In fact, scanning the class table is much more expensive than having a read barrier on message lookup, since
- becomes can be frequent (e.g. growing method dictionaries) and so the class table scanning becomes a significant cost in some circumstances
- method cacheing makes lookups rare and hence a small overhead
- the read barriers on fetching the superclass and method dictionary et al on lookup are very cheap because
	a) lookups are rare, in the interpreter at most 4% of message sends are looked up and in the JIT at most 1%
	b) the class indexes likely share cache lines with the state being fetched, and
	c) forwarding is rare so most of the time the VM accesses the class index, finds the object is not forwarded and then fetches the superclass or method dictionary from the same cache line as the class index.
- since lookups are rare (and since nil is never forwarded and explicitly tested for on method dictionary search) the read barrier on fetching selectors in method dictionaries has a small overhead.

Accessing possibly forwarded method literals is fine; these forwarding objects will be pushed on the stack and caught either by the message lookup trap or explicitly in primitives that access arguments.  However, push/store/pop literal variable cannot avoid an explicit check without requiring we scan all methods, which is far too expensive.  So accessing associations from methods (including the method class association for super sends) requires a read barrier.

And of course, all of the scavenger, the incremental scan-mark-compactor and the global garbage collector follow and eliminate forwarding pointers as part of their object graph traversals.

This means explicit read barriers in
- push/store/pop literal variable
- super send
- base frame return (accessing the sender context, its inst vars, and the method class association of its method)
- block activation (the JIT avoids a read barrier on block activation unless a block accesses inst vars)

This leaves primitives where explicit access to object state is done without message sending.  Since primitives check their arguments for validity, and fail without side effects if validation fails, they fail if encountering a forwarding object, and can be retried.  There were a few exceptions to this in the Semaphore installation primitives that assumed a non-Semaphore argument was nil, hence being fooled by forwarders to Semaphores.  These primitives were easily fixed.  So all primitives that examine state fail if they encounter a forwarder amongst that state.  Slang is used to analyse the parse trees of primitives and compute an "accessor depth", the depth to which a primitive traverses the object graph of its input arguments.  If a primitive fails then the primitive failure code retrieves the accessor depth for the primitive and, if >= 0, traverses the state to that depth, following (and fixing) any and all forwarders found.  If any are found the primitive is retried.

Stale Super Sends
Notionally we catch sends to forwarded objects by only checking for forwqarders when a send lookup cache fails, but Smalltalk supersends, once linked, are unchecked, and the lookup starts from a method's methodClass, which requires an explicit read barrier.  So there is no possibility of implicitly checking for forwarding of the receiver of a super send.  Since a super send looks like push: self, marshall arguments, super send, and receivers are never forwarded this may look like a non-problem, but any computation could happen as part of marshalling the supersend arguments, e.g. ^super doSomethingWith: (self become: self somethingElse).  Because self is pushed before argument marshalling, stale forwarded references to self can be left on the stack after a become and before the supersend.

There are three solutions for this problem we have considered.  One is to scan the entire stack after become (and the entire stack of a context faulted back into the stack zone).  This is expensive.  Another is to use checked entry-points for super.  This seems to add a lot of new code and complexity, and works only in jitted code.  The solution we have chosen is to explicitly test for a forwarded receiver.  In the interpreter this is done by sending ensureReceiverUnforwarded in the supersend lookup routines.  In jitted code it is done by adding code to the send site to unforward the receiver after arguments are marshalled and immediately before the send call.  Because super sends are extremely rare this adds little overhead.

61-bit Immediate Floats
Representation for immediate doubles, only used in the 64-bit implementation. Immediate doubles have the same 52 bit mantissa as IEEE double-precision  floating-point, but only have 8 bits of exponent.  So they occupy just less than the middle 1/8th of the double range.  They overlap the normal single-precision floats which also have 8 bit exponents, but exclude the single-precision denormals (exponent-127) and the single-precsion NaNs (exponent +127).  +/- zero is just a pair of values with both exponent and mantissa 0. 
So the non-zero immediate doubles range from 
        +/- 0x3800,0000,0000,0001 / 5.8774717541114d-39 
to      +/- 0x47ff,ffff,ffff,ffff / 6.8056473384188d+38 
The encoded tagged form has the sign bit moved to the least significant bit, which allows for faster encode/decode because offsetting the exponent can't overflow into the sign bit and because testing for +/- 0 is an unsigned compare for <= 0xf: 
    msb                                                                                        lsb 
    [8 exponent subset bits][52 mantissa bits ][1 sign bit][3 tag bits] 
So given the tag is 4, the tagged non-zero bit patterns are 
             0x0000,0000,0000,001[c(8+4)] 
to           0xffff,ffff,ffff,fff[c(8+4)] 
and +/- 0d is 0x0000,0000,0000,000[c(8+4)] 
Encode/decode of non-zero values in machine code looks like: 
						msb                                              lsb 
Decode:				[8expsubset][52mantissa][1s][3tags] 
shift away tags:			[ 000 ][8expsubset][52mantissa][1s] 
add exponent offset:	[     11 exponent     ][52mantissa][1s] 
rot sign:				[1s][     11 exponent     ][52mantissa]

Encode:					[1s][     11 exponent     ][52mantissa] 
rot sign:				[     11 exponent     ][52mantissa][1s] 
sub exponent offset:	[ 000 ][8expsubset][52 mantissa][1s] 
shift:					[8expsubset][52 mantissa][1s][ 000 ] 
or/add tags:			[8expsubset][52mantissa][1s][3tags] 
but is slower in C because 
a) there is no rotate, and 
b) raw conversion between double and quadword must (at least in the source) move bits through memory (quadword = *(q64 *)&doubleVariable). 


Heap Walking
In heap walking the memory manager needs to be able to detect the start of the next object.  This is complicated by the short and long header formats, short being for objects with 254 slots or less, long being for objects with 255 slots or more.  Since an object that has an overflow header must have 255 as its header slot count we can use this as the marker.  The overflow header word also has a numSlots field, set to 255.  The remainder of the overflow size field is used for the object's slot size, the least significant word in 32-bits (for 2^34 bytes, more than the address space), the remaining 56 bits in 64-bits (for 2^59 bytes, which we hope is big enough for bridge objects).  So if the word following an object contains 255 in its numSlots field, it must be the overflow size word of an object with a double header, and the word after that is the header, also with a saturated numSlots field.

We support extra alignment above 8 bytes, and bridging pastSpace and eden to oldSpace, by supporting "slim bridges", which occupy at least one world. Again for the FFI, on modern processors there exist media instructions that require greater than 8 byte alignment (for example the LD1/2/3/4R instructions in the ARMv8 ISA).  pastSpace is below eden and may be adjacent (because pastSpace can be completely full), or one word less than it.  Eden is unlikely to fill completely, but can be empty or full, and is adjacent to the first word of oldSpace.  So to bridge from the first used word in new space to old space may require a one-word slimbridge between an almost full pastSpace and a non-empty eden.  

This scheme is extended to provide greater than 8 byte alignment for objects.  An object that has greater than 8 byte alignment has the isExtraAligned but set and is (must be) preceded by a slimbridge.  The actual alignment the object has is held in the slimbridge.  Hence an object already aligned on an N word boundary, for N > 1, must be preceded by an N word slimbridge.  We support alignment slimbridges by noting that an overflow size word must have a slot count of at least 255.  Any value less than 128 indicates a slim bridge, which gives us 7 bits, organized as a 3-bit alignment field (in bits 4 to 6), and a 4-bit bridge size field (in bIts 0 to 3).  The bridge size field gives the size minus one in words of the slimbridge.  A size of zero specifies an 8-byte, one word slimbridge, a size of 15 implies an 128 byte, 16 word slimbridge. The two bit alignment field gives the power of two minus three in bytes.  An alignment field of 0 imples 8 byte alignment (and hence is only used with single word slimbridges). An alignment field of 1 imples 16 byte alignment (and hence may be used with one or two word slimbridges). An alignment field of 2 imples 32 byte alignment (and hence may be used with one to four word slimbridges). An alignment field of 3 imples 64 byte alignment (and hence may be used with one to eight word slimbridges).  An alignment field of 4 imples 128 byte alignment (and hence may be used with one to sixteen word slimbridges). Alignments of 256, 512, and 1024 bytes are representable, but require a larger size field than can fit in four bits, and so are not used.

So to align an object on a 16, 32, or 128 byte boundary, at allocation time a slimbridge of the size required to pad to the required alignment is written into memory, and immediatrly following it the object is allocated, with its isExtraAligned bit set. The required/actual alignment is written into the slimbridge's alignment field, and the entire field (saturated size count and slot count < 128) replicated in all words of the slimbridge.  When the object is moved, either during scavenging or compaction, when it is written into its new location, a new slimbridge is synthesized to align it in the new location on the same boundary.  But the size of slimbridge may be entirely different in the new location.  Only the alignment field and the object's actual alignment is guaranteed not to change.  Note also that objects with an overflow size field require one word less padding in the slimbridge, because it is the word following the normal header word that is aligned, not the header word and not the overflow header word.


Total Number of Classes and Instance-specific Behaviours
While the class index header field has advantages (saving significant header space, especially in 64-bits, providing a non-moving cache tag for inline caches, small constants for instantiating well-known classes instead of having to fetch them from a table such as the specialObjectsArray) it has the downside of limiting the number of classes.  For Smalltalk programs 2^20 to 2^24 classes is adequate for some time to come, but for prototype languages such as JavaScript this is clearly inadequate, and we woud like to support the ability to host prototype languages within Squeak. There is a solution in the form of "auto-instances", an idea of Claus Gittinger's.  The idea is to represent prototypes as behaviors that are instances of themselves.  In a classical Smalltalk system a Behavior is an object with the minimal amount of state to function as a class, and in Smalltalk-80 this state is the three instance variables of Behavior, superclass, methodDict and format, which are the only fields in a Behavior that are known to the virtual machine.  A prototype can therefore have its own behavior and inherit from other prototypes or classes, and have sub-prototypes derived from it if a) its first three instance variables are also superclass, methodDict, and format, and b) it is an instance of itself (one can create such objects in a normal Smalltalk system by creating an Array with the desired layout and using a change class primitive to change the class of the Array to itself).  The same effect can be achieved in a VM with class indexes by reserving one class index to indicate that the object is an instance of itself, hence not requiring the object be entered into the class table and in the code that derives the class of an object, requiring one simple test answering the object itself instead of indexing the class table.  There would probably need to be an auto-instantiation primitive that takes a behavior (or prototype) and an instance variable count and answers a new auto-instance with as many instance variables as the sum of the behavior (or prototype) and the instance variable count.  Using this scheme there can be as many auto-instances as available address space allows while retaining the benefits of class indices.

This scheme has obvious implications for the inline cache since all prototypes end up having the same inline cache tag.  Either the inline cache check checks for the auto-instance class tag and substitutes the receiver, or the cacheing machinery refuses to add the auto-instance class tag to any inline cache and failure path code checks for the special case.  Note that in V8 failing monomorphic sends are patched to open PICs (megamorphic sends); V8 does not use closed PICs due to the rationale that polymorphism is high in JavaScript.

Miscellanea:

In-line Cache Probe for Immediates
We would like to keep 31-bit SmallIntegers for the 32-bit system.  Lots of code could be impacted by a change to 30-bit SmallIntegers.  If so, then 
	isImmediate: oop	^(oop bitAnd: 3) ~= 0
	isSmallInteger: oop	^(oop bitAnd: 1) ~= 0
	isCharacter: oop	^(oop bitAnd: 2) = 2
If the in-line cache contains 0 for Characters then the in-line cache code (in x86 machine code) reads as follows, for a branch-free common-case:
	Limm: (aligned)
		andl $0x1, %eax
		j Lcmp
		nops...
	Lentry: (aligned)
		movl %edx, %eax
		andl $0x3, %eax
		jnz Limm
		movl 0(%edx), %eax
		andl $0x3fffff, %eax
	Lcmp:
		cmpl %ecx, %eax
		jnz Lmiss
	LuncheckedEntry:


64-Bit Sizes
We extend the original Squeak 4-bit per-object format field to 5 bits, providing space for 3 odd bits for byte objects (2 for short objects & 1 for 32-bit long objects).  Object sizes are slots, and byte (and short and 32-bit) lengths are computed by subtracting the odd bits from the shifted slot length.  Keeping the format field saves bits because it subsumes the isWeak,isEphemeron,isPointer bits that would be necessary otherwise.  The format field is organized as follows:
	 0 = 0 sized objects (UndefinedObject True False et al)
	 1 = non-indexable objects with inst vars (Point et al)
	 2 = indexable objects with no inst vars (Array et al)
	 3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
	 4 = weak indexable objects with inst vars (WeakArray et al)
	 5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
	 6 unused, reserved for exotic pointer objects?  e.g. contexts?
	 7 Forwarded Object, 1st field is pointer, rest of fields are ignored
	 8 unused, reserved for exotic non-pointer objects?
	 9 64-bit indexable
	10 - 11 32-bit indexable	(11 unused in 32 bits)
	12 - 15 16-bit indexable	(14 & 15 unused in 32-bits)
	16 - 23 byte indexable		(20-23 unused in 32-bits)
	24 - 31 compiled method	(28-21 unused in 32-bits)


One obvious optimization to images is to add image (de)compression to image loading and snapshot.  The image header remains unchanged but its contents could be compressed, compressing either on snapshot, if requested, or off-line via a special-purpose tool.

Simple Object Representation Improves Code Quality
Given that the Spur object representation is quite regular and simple it has been possible to implement many more performance-critcal operations in machine code.  The exsting VM allocated only Float results in the floating-point primitives, deferring to Interpreter primitives (in the Interpreter's C run-time, invoked by a system-call like stack-switching call) for object instantiation, and closure, indirection vector and context creation.  The Spur JIT now generates machine code for the object instantiation primitives, and closure, indirection vector and context creation.  It also implements the at:put: primitive in machine code and implements both at: and at:put: on WideString in machine code (the previous VM needs to instantiate characters with codes > 255 in at:).  As a result, Spur is substantially faster executing Smalltalk.  The scavenger also has a significant impact on GC-intensive benchmarks.


Issues:
Are class indices in inline caches strong references to classes or weak references?
If strong then they must be scanned during GC and the methodZone must be flushed on fullGC to reclaim all classes (this looks to be a bug in the V3 Cogit).
If weak then when the class table loses references, PICs containing freed classes must be freed and then sends to freed PICs or containing freed clases must be unlinked.
The second approach is faster; the common case is scanning the class table, the uncommon case is freeing classes.  The second approach is better for machine code; in-line caches do not prevent reclamation of classes.  However, the former is better for the scavenger which as a result doesn't have to scan the classes of objects.  The system scavenges much more frequently than it reclaims classes and the number of cog methods to be scanned during scavenge is small.  So I think a strong class table is better after all.  Indeed we use a strong class table with old pages.  So classes are only GCed by the incremental and global GCs, never by the scavenger.!

SpurMemoryManager subclass: #Spur32BitMemoryManager
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManager'!

Spur32BitMemoryManager subclass: #Spur32BitCoMemoryManager
	instanceVariableNames: 'cogit'
	classVariableNames: ''
	poolDictionaries: 'CogMethodConstants'
	category: 'VMMaker-SpurMemoryManager'!

!Spur32BitCoMemoryManager commentStamp: 'eem 11/25/2013 14:47' prior: 0!
Spur32BitCoMemoryManager is a refinement of Spur32BitMemoryManager that supports the CoInterpreter/Cogit just-in-time compiler.  The significant difference from Spur32BitMemoryManager is the memory layout.  Spur32BitCoMemoryManager adds the cogCodeZone beneath newSpace:

low address:
	cogCodeZone:
		generated run-time
		cog methods
		free space
		young referrers
	newSpace:
		past/future survivor space
		future/past survivor space
		eden
	first oldSpace segment
	...
	subsequent oldSpace segment
high address:

It would be convenient if the code zone were placed between newSpace and oldSpace; then Cog methods could be onsidered neither old nor young, filtering them out of copyAndForward: and the store check with single bounds checks.  But the CoInterpreter already assumes Cog methods are less than all objects (e.g. in its isMachineCodeFrame:).  If the dynamic frequency of isMachineCodeFrame: is higher (likely because this is used in e.g. scanning for unwind protects in non-local return) then it should keep the single bounds check.  So the coder zone remains beneath newSpace and Spur32BitCoMemoryManager ocerrides isReallyYoungObject: to filter-out Cog methods for copyAndForward:.

Instance Variables
	cogit:		<SimpleStackBasedCogit or subclass>

cogit
	- the just-in-time compiler!

!Spur32BitCoMemoryManager class methodsFor: 'simulation' stamp: 'eem 12/4/2015 22:20'!
defaultISA
	"Answer the default instruction set architecture to use for simulation."
	^#IA32! !

!Spur32BitCoMemoryManager class methodsFor: 'accessing class hierarchy' stamp: 'eem 10/9/2013 19:46'!
objectRepresentationClass
	^CogObjectRepresentationFor32BitSpur! !

!Spur32BitCoMemoryManager class methodsFor: 'simulation' stamp: 'eem 10/11/2013 16:22'!
simulatorClass
	^Spur32BitMMLECoSimulator! !

!Spur32BitCoMemoryManager methodsFor: 'trampolines' stamp: 'eem 3/12/2015 16:44'!
ceScheduleScavenge
	<api>
	self assert: freeStart >= scavengeThreshold.
	self scheduleScavenge! !

!Spur32BitCoMemoryManager methodsFor: 'debug support' stamp: 'eem 11/25/2013 14:34'!
checkMemoryMap
	"Override to check that Cog methods are considered neither young nor old.
	 Being young would cause them to be scavenged.
	 Being old would cause them to be remembered if stored into (but wait, they don't get stored into)."

	self assert: (self isYoungObject: newSpaceStart).
	self assert: (self isYoungObject: newSpaceLimit - self wordSize).
	self assert: (self isOldObject: newSpaceStart) not.
	self assert: (self isOldObject: newSpaceLimit - self wordSize) not.
	self assert: (self isYoungObject: newSpaceLimit) not.
	self assert: (self isYoungObject: oldSpaceStart) not.
	self assert: (self isYoungObject: endOfMemory) not.
	self assert: (self isOldObject: oldSpaceStart).
	self assert: (self isOldObject: endOfMemory).

	"we would like the following to be true, but we either choose one boundary check for
	 cogMethods vs objects (isMachineCodeFrame: et al) or one boundary check for
	 copyAndForward:.  We can't have both, and the former is likely the highest dynamic
	 frequency."
	false ifTrue:
		[self assert: (self isYoungObject: cogit minCogMethodAddress) not.
		 self assert: (self isYoungObject: cogit maxCogMethodAddress) not].
	self assert: (self isOldObject: cogit minCogMethodAddress) not.
	self assert: (self isOldObject: cogit maxCogMethodAddress) not! !

!Spur32BitCoMemoryManager methodsFor: 'debug support' stamp: 'eem 10/1/2013 16:58'!
clearLeakMapAndMapAccessibleObjects
	"Perform an integrity/leak check using the heapMap.  Set a bit at each object's header.
	 Override to set a bit at each Cog method"
	super clearLeakMapAndMapAccessibleObjects.
	cogit addCogMethodsToHeapMap! !

!Spur32BitCoMemoryManager methodsFor: 'cog jit support' stamp: 'eem 7/10/2014 17:09'!
ensureNoForwardedLiteralsIn: aMethodObj
	"Ensure there are no forwarded literals in the argument."
	<api>
	self followForwardedObjectFields: aMethodObj toDepth: 0! !

!Spur32BitCoMemoryManager methodsFor: 'forwarding' stamp: 'eem 6/24/2020 10:23'!
followForwardedObjectFields: objOop toDepth: depth
	"Follow pointers in the object to depth.
	 Answer if any forwarders were found.
	 How to avoid cyclic structures?? A temporary mark bit? eem 6/22/2020 no need since depth is always finite."
	<api>
	<inline: false>
	| found fmt numSlots |
	found := false.
	self assert: ((self isPointers: objOop) or: [self isOopCompiledMethod: objOop]).
	fmt := self formatOf: objOop.
	numSlots := self numPointerSlotsOf: objOop format: fmt.
	"It is essential to skip the first field of a method because it may be a
	 reference to a Cog method in the method zone, not a real object at all."
	((self isCompiledMethodFormat: fmt)
			ifTrue: [1]
			ifFalse: [0])
		to: numSlots - 1
		do: [:i| | oop |
			oop := self fetchPointer: i ofObject: objOop.
			(self isNonImmediate: oop) ifTrue:
				[(self isForwarded: oop) ifTrue:
					[found := true.
					 oop := self followForwarded: oop.
					 self storePointer: i ofObject: objOop withValue: oop].
				(depth > 0
				 and: [(self hasPointerFields: oop)
				 and: [self followForwardedObjectFields: oop toDepth: depth - 1]]) ifTrue:
					[found := true]]].
	^found! !

!Spur32BitCoMemoryManager methodsFor: 'cog jit support' stamp: 'eem 6/2/2014 18:47'!
freeStart: aValue
	self assert: (aValue >= scavenger eden start and: [aValue < (scavengeThreshold + 1024)]).
	self assert: (scavengeThreshold max: aValue) + coInterpreter interpreterAllocationReserveBytes <= scavenger eden limit.
	^freeStart := aValue! !

!Spur32BitCoMemoryManager methodsFor: 'trampoline support' stamp: 'eem 10/3/2013 10:53'!
freeStartAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: freeStart) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #freeStart in: self]! !

!Spur32BitCoMemoryManager methodsFor: 'cog jit support' stamp: 'eem 12/4/2013 12:55'!
getScavengeThreshold
	<api>
	<returnTypeC: #usqInt>
	^scavengeThreshold! !

!Spur32BitCoMemoryManager methodsFor: 'garbage collection' stamp: 'eem 10/1/2013 14:26'!
headerWhileForwardingOf: aCompiledMethodObjOop
	"Answer the header of the argument even though
	 it may have its header word in a forwarding block
	 (which shouldn't happen with Spur)."
	self assert: (self isForwarded: aCompiledMethodObjOop) not.
	^self baseHeader: aCompiledMethodObjOop! !

!Spur32BitCoMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 12/26/2014 17:01'!
ifAProxy: objOop updateCopy: copy
	"If the object being copied to the segment is weird and has exotic state,
	 i.e. a married context or a jitted method, update the copy with the vanilla state."

	super ifAProxy: objOop updateCopy: copy.
	(self isCompiledMethod: objOop) ifTrue:
		[| methodHeader |
		 methodHeader := coInterpreter rawHeaderOf: objOop.
		 (coInterpreter isCogMethodReference: methodHeader) ifTrue:
			[self storePointerUnchecked: HeaderIndex
				ofObject: copy
				withValue: (coInterpreter cCoerceSimple: methodHeader to: #'CogMethod *') methodHeader]]! !

!Spur32BitCoMemoryManager methodsFor: 'simulation only' stamp: 'eem 6/29/2015 14:42'!
initializeFreeSpaceForFacadeFrom: base to: limit
	"c.f. initializeFreeSpacePostLoad: freeListObj."
	| freeListObj freeBytes |
	newSpaceLimit := oldSpaceStart := freeStart := base.
	endOfMemory := limit.
	scavengeThreshold := limit * 3 // 4.
	segmentManager initSegmentForInImageCompilationFrom: base to: limit.
	freeListObj := self allocateSlots: self numFreeLists
						format: self wordIndexableFormat
						classIndex: self wordSizeClassIndexPun.
	freeLists := self firstIndexableField: freeListObj.
	freeListsMask := 0.
	0 to: self numFreeLists - 1 do:
		[:i|
		(freeLists at: i) ~= 0 ifTrue:
			[freeListsMask := freeListsMask bitOr: (1 << i).
			 freeLists at: i put: (segmentManager swizzleObj: (freeLists at: i))]].
	freeBytes := segmentManager lastSegment segLimit - self bridgeSize - freeStart.
	freeLists at: 0 put: (self initFreeChunkWithBytes: freeBytes at: freeStart).
	totalFreeOldSpace := freeBytes! !

!Spur32BitCoMemoryManager methodsFor: 'class table' stamp: 'eem 9/23/2014 18:56'!
isForwardedClassIndex: maybeClassIndex
	"A lenient tester of forwarded class indices for inline cache management in the Cogit."
	<api>
	| classTablePage entry |
	maybeClassIndex asUnsignedInteger >= self classTableRootSlots ifTrue:
		[^false].
	classTablePage := self fetchPointer: maybeClassIndex >> self classTableMajorIndexShift
							ofObject: hiddenRootsObj.
	classTablePage = nilObj ifTrue:
		[^false].
	entry := self
				fetchPointer: (maybeClassIndex bitAnd: self classTableMinorIndexMask)
				ofObject: classTablePage.
	^self isForwarded: entry! !

!Spur32BitCoMemoryManager methodsFor: 'cog jit support' stamp: 'eem 8/7/2014 04:34'!
isImmediateClass: classObj
	<api>
	^(self instSpecOfClass: classObj) = self instSpecForImmediateClasses! !

!Spur32BitCoMemoryManager methodsFor: 'debug support' stamp: 'eem 7/1/2021 11:25'!
isInHeapBounds: address 
	"Answer if the given address is within the entire range ST object memory.
	 For quick checking during leak checking only!!"
	^(self oop: address isGreaterThanOrEqualTo: cogit minCogMethodAddress)
	  and: [self oop: address isLessThan: endOfMemory]! !

!Spur32BitCoMemoryManager methodsFor: 'object testing' stamp: 'eem 11/25/2013 14:40'!
isReallyYoungObject: objOop
	<api>
	"Answer if obj is young. Require that obj is non-immediate. Override to filter-out Cog methods"
	self assert: (self isNonImmediate: objOop).
	^(self oop: objOop isLessThan: newSpaceLimit)
	  and: [self oop: objOop isGreaterThanOrEqualTo: newSpaceStart]! !

!Spur32BitCoMemoryManager methodsFor: 'compaction' stamp: 'eem 12/25/2016 10:40'!
methodHeaderFromSavedFirstField: field
	(self isIntegerObject: field) ifTrue:
		[^field].
	self assert: ((self isNonImmediate: field) and: [field < newSpaceStart]).
	self assert: (coInterpreter cCoerceSimple: field to: #'CogMethod *') objectHeader
				= self nullHeaderForMachineCodeMethod.
	^(coInterpreter cCoerceSimple: field to: #'CogMethod *') methodHeader! !

!Spur32BitCoMemoryManager methodsFor: 'growing/shrinking memory' stamp: 'eem 4/28/2015 14:47'!
methodHeaderOf: methodObj
	"Answer the method header of a CompiledMethod object.
	 If the method has been cogged then the header is a pointer to
	 the CogMethod and the real header will be stored in the CogMethod."
	<api>
	<inline: true>
	| header |
	self assert: (self isCompiledMethod: methodObj).
	header := self fetchPointer: HeaderIndex ofObject: methodObj.
	^(self isIntegerObject: header)
		ifTrue: [header]
		ifFalse:
			[self assert: header asUnsignedInteger < newSpaceStart.
			 self assert: (coInterpreter cCoerceSimple: header to: #'CogMethod *') objectHeader
						= self nullHeaderForMachineCodeMethod.
			(coInterpreter cCoerceSimple: header to: #'CogMethod *') methodHeader]! !

!Spur32BitCoMemoryManager methodsFor: 'trampoline support' stamp: 'eem 4/11/2017 16:12'!
needGCFlagAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: needGCFlag) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #needGCFlag in: self]! !

!Spur32BitCoMemoryManager methodsFor: 'garbage collection' stamp: 'eem 11/19/2013 12:31'!
nullHeaderForMachineCodeMethod
	<api>
	<returnTypeC: #sqLong>
	^(self firstLongFormat << self formatShift)
	+ (1 << self markedBitFullShift)
	+ ClassBitmapCompactIndex! !

!Spur32BitCoMemoryManager methodsFor: 'object enumeration' stamp: 'eem 5/31/2018 10:53'!
objectBytesForSlots: numSlots
	"Answer the total number of bytes in an object with the given
	 number of slots, including header and possible overflow size header."
	<api>
	<option: #SistaVM>
	^super objectBytesForSlots: numSlots! !

!Spur32BitCoMemoryManager methodsFor: 'cog jit support' stamp: 'eem 8/31/2021 14:23'!
receiverTagBitsForImmediateMethodClass: methodClassOrNil
	<inline: #always>
	^methodClassOrNil = (self fetchPointer: self smallIntegerTag ofObject: classTableFirstPage)
		ifTrue: [self smallIntegerTag]
		ifFalse: [self assert: methodClassOrNil = (self fetchPointer: self characterTag ofObject: classTableFirstPage).
				self characterTag]! !

!Spur32BitCoMemoryManager methodsFor: 'cog jit support' stamp: 'eem 7/29/2014 07:35'!
remember: objOop
	<doNotGenerate>
	^scavenger remember: objOop! !

!Spur32BitCoMemoryManager methodsFor: 'trampoline support' stamp: 'eem 5/31/2014 09:49'!
scavengeThresholdAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: scavengeThreshold) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #getScavengeThreshold in: self]! !

!Spur32BitCoMemoryManager methodsFor: 'debug support' stamp: 'eem 3/9/2016 17:48'!
setUpForUseByFacade: aCurrentImageCoInterpreterFacade
	"Make sure that eden etc are initialized, so that methods can be printed.
	 This is really to make addressCouldBeObj: et al work."
	<doNotGenerate>
	self edenBytes: 0.
	self setHeapBase: self freeStart
		memoryLimit: self endOfMemory
		endOfMemory: self endOfMemory! !

!Spur32BitCoMemoryManager methodsFor: 'accessing' stamp: 'eem 12/6/2013 18:50'!
startOfMemory
	"Return the start of object memory.  This is immediately after the native code zone.
	 N.B. the stack zone is alloca'ed. Use a macro so as not to punish the debug VM."
	<api>
	<cmacro: '() heapBase'>
	<returnTypeC: #usqInt>
	^coInterpreter heapBase! !

!Spur32BitCoMemoryManager methodsFor: 'simulation only' stamp: 'eem 5/14/2014 15:03'!
unalignedLongAt: byteAddress
	<doNotGenerate>
	| rem |
	rem := byteAddress \\ 4.
	^rem = 0
		ifTrue: [self longAt: byteAddress]
		ifFalse: [((self longAt: byteAddress - rem) + ((self longAt: byteAddress - rem + 4) bitShift: 32) bitShift: rem * -8) bitAnd: 16rFFFFFFFF]! !

!Spur32BitCoMemoryManager methodsFor: 'simulation only' stamp: 'eem 5/14/2014 15:04'!
unalignedLongAt: byteAddress put: aLong
	<doNotGenerate>
	| rem mask |
	rem := byteAddress \\ 4.
	^rem = 0
		ifTrue: [self longAt: byteAddress put: aLong]
		ifFalse:
			[mask := 16rFFFFFFFF bitAnd: (-1 bitShift: rem * 8).
			 self longAt: byteAddress - rem
				put: ((self longAt: byteAddress - rem) bitAnd: mask bitInvert)
					+ ((aLong bitShift: rem * 8) bitAnd: mask).
			 self longAt: byteAddress - rem + 4
				put: ((self longAt: byteAddress - rem + 4) bitAnd: mask)
					+ ((aLong bitShift: 4 - rem * -8) bitAnd: mask bitInvert).
			 aLong]! !

!Spur32BitCoMemoryManager methodsFor: 'cog jit support' stamp: 'eem 11/18/2013 18:57'!
withoutForwardingOn: obj1 and: obj2 with: aBool sendToCogit: selector
	"For the purposes of become: send selector to the cogit with obj1, obj2
	 and aBool and answer the result.  Undo forwarding for the selector."
	<api>
	<var: #selector declareC: 'sqInt (*selector)(sqInt,sqInt,sqInt)'>
	| targetA targetB |
	targetA := self followForwarded: obj1.
	targetB := self followForwarded: obj2.
	^cogit perform: selector with: targetA with: targetB with: aBool! !

Spur32BitCoMemoryManager subclass: #Spur32BitMMLECoSimulator
	instanceVariableNames: 'parent bootstrapping'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManagerSimulation'!

!Spur32BitMMLECoSimulator methodsFor: 'accessing' stamp: 'eem 11/14/2013 16:42'!
bootstrapping
	^bootstrapping! !

!Spur32BitMMLECoSimulator methodsFor: 'accessing' stamp: 'eem 11/15/2013 11:00'!
bootstrapping: aBoolean
	bootstrapping := aBoolean.
	segmentManager initForBootstrap! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:35'!
byteAt: byteAddress
	| lowBits long32 |
	lowBits := byteAddress bitAnd: 3.
	long32 := self long32At: byteAddress - lowBits.
	^(long32 bitShift: -8 * lowBits) bitAnd: 16rFF! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:35'!
byteAt: byteAddress put: byte
	| lowBits long32 longAddress mask value |
	lowBits := byteAddress bitAnd: 3.
	longAddress := byteAddress - lowBits.
	long32 := self long32At: longAddress.
	mask := (16rFF bitShift: 8 * lowBits) bitInvert.
	value := byte bitShift: 8 * lowBits.
	self long32At: longAddress put: ((long32 bitAnd: mask) bitOr: value).
	^byte! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 9/4/2013 16:23'!
byteAtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address."

	^self byteAt: pointer! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'nice 4/6/2016 00:14'!
byteSwapped32IfBigEndian: w
	^w bitAnd: 16rFFFFFFFF! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'nice 4/6/2016 00:15'!
byteSwapped64IfBigEndian: w
	^w bitAnd: 16rFFFFFFFFFFFFFFFF! !

!Spur32BitMMLECoSimulator methodsFor: 'debug support' stamp: 'eem 9/5/2021 18:54'!
checkHeapIntegrity: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid
	"Almost all of the time spent in SpurMemoryManager>>checkHeapIntegrity:classIndicesShouldBeValid: used to go into the asserts in fetchPointer:ofObject: in the simulator class overrides.
	 Since we know here the indices used are valid we temporarily remove them to claw back that performance."
	^self withSimulatorFetchPointerMovedAsideDo:
		[super checkHeapIntegrity: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid]! !

!Spur32BitMMLECoSimulator methodsFor: 'initialization' stamp: 'eem 12/16/2016 16:30'!
coInterpreter: aCoInterpreter cogit: aCogit
	coInterpreter := aCoInterpreter.
	cogit := aCogit.
	scavenger coInterpreter: aCoInterpreter.
	compactor coInterpreter: aCoInterpreter! !

!Spur32BitMMLECoSimulator methodsFor: 'simulation only' stamp: 'eem 11/25/2013 13:17'!
cogCodeBase
	^Cogit guardPageSize! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 8/30/2013 11:47'!
endianness
	^#little! !

!Spur32BitMMLECoSimulator methodsFor: 'float primitives' stamp: 'eem 9/10/2013 17:29'!
fetchFloatAt: floatBitsAddress into: aFloat
	aFloat at: 2 put: (self long32At: floatBitsAddress).
	aFloat at: 1 put: (self long32At: floatBitsAddress+4)! !

!Spur32BitMMLECoSimulator methodsFor: 'object access' stamp: 'cb 6/11/2018 09:55'!
fetchPointer: fieldIndex ofObject: objOop
	self assert: (self isForwarded: objOop) not.
	self assert: (fieldIndex >= 0 and: [fieldIndex < (self numSlotsOfAny: objOop)
				or: [fieldIndex = 0 "forwarders"]]).
	^super fetchPointer: fieldIndex ofObject: objOop! !

!Spur32BitMMLECoSimulator methodsFor: 'object format' stamp: 'eem 12/28/2016 21:08'!
firstIndexableField: objOop
	"NOTE: overridden from SpurMemoryManager to add coercion to CArray, so please duplicate any changes.
	 There are only two important cases, both for objects with named inst vars, i.e. formats 2,3 & 5.
	 The first indexable field for formats 2 & 5 is the slot count (by convention, even though that's off the end
	 of the object).  For 3 we must go to the class."
	| fmt classFormat |
	<returnTypeC: #'void *'>
	fmt := self formatOf: objOop.
	fmt <= self lastPointerFormat ifTrue: "pointer; may need to delve into the class format word"
		[(fmt between: self indexablePointersFormat and: self weakArrayFormat) ifTrue:
			[classFormat := self formatOfClass: (self fetchClassOfNonImm: objOop).
			 ^self cCoerce: (self pointerForOop: objOop
												+ self baseHeaderSize
												+ ((self fixedFieldsOfClassFormat: classFormat) << self shiftForWord))
					to: #'oop *'].
		^self cCoerce: (self pointerForOop: objOop + self baseHeaderSize)
				to: #'oop *'].
	"All bit objects, and indeed CompiledMethod, though this is a no-no, start at 0"
	self assert: (fmt >= self sixtyFourBitIndexableFormat and: [fmt < self firstCompiledMethodFormat]).
	^self
		cCoerce: (self pointerForOop: objOop + self baseHeaderSize)
		to: (fmt < self firstByteFormat
				ifTrue:
					[fmt = self sixtyFourBitIndexableFormat
						ifTrue: ["64 bit field objects" #'long long *']
						ifFalse:
							[fmt < self firstShortFormat
								ifTrue: ["32 bit field objects" #'int *']
								ifFalse: ["16-bit field objects" #'short *']]]
				ifFalse: ["byte objects (including CompiledMethod" #'char *'])! !

!Spur32BitMMLECoSimulator methodsFor: 'spur bootstrap' stamp: 'eem 10/8/2013 11:36'!
freeLists
	^freeLists! !

!Spur32BitMMLECoSimulator methodsFor: 'accessing' stamp: 'eem 6/1/2014 09:58'!
freeStart
	"freeStart = 16r1163E0 ifTrue: [self halt]."
	^super freeStart! !

!Spur32BitMMLECoSimulator methodsFor: 'gc - global' stamp: 'eem 5/30/2018 11:23'!
globalGarbageCollect
	"If we're /not/ a clone, clone the VM and push it over the cliff.
	 If it survives, destroy the clone and continue.  We should be OK until next time."
	parent ifNil:
		[coInterpreter cr; print: 'GC number '; print: statFullGCs; tab; flush.
		 CloneOnGC ifTrue:
			[coInterpreter cloneSimulation objectMemory globalGarbageCollect.
			 Smalltalk garbageCollect]].
	^super globalGarbageCollect! !

!Spur32BitMMLECoSimulator methodsFor: 'growing/shrinking memory' stamp: 'eem 11/14/2013 15:30'!
growOldSpaceByAtLeast: minAmmount
	"Attempt to grow memory by at least minAmmount.
	 Answer the size of the new segment, or nil if the attempt failed.
	 Override to not grow during the Spur image bootstrap."
	^bootstrapping ifFalse:
		[super growOldSpaceByAtLeast: minAmmount]! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 9/24/2013 13:50'!
halfWordHighInLong32: long32
	"Used by Balloon"

	^long32 bitAnd: 16rFFFF! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 9/24/2013 13:51'!
halfWordLowInLong32: long32
	"Used by Balloon"

	^long32 bitShift: -16! !

!Spur32BitMMLECoSimulator methodsFor: 'header format' stamp: 'eem 9/4/2013 11:08'!
headerForSlots: numSlots format: formatField classIndex: classIndex
	"The header format in LSB is
	 MSB:	| 2 bits				|
			| 22: identityHash	|
			| 8: slotSize			|
			| 3 bits				|
			| 5: format			|
			| 2 bits				|
			| 22: classIndex		| : LSB"
	self assert: (numSlots bitAnd: self numSlotsMask) = numSlots.
	self assert: (formatField bitAnd: self formatMask) = formatField.
	self assert: (classIndex bitAnd: self classIndexMask) = classIndex.
	^super headerForSlots: numSlots format: formatField classIndex: classIndex! !

!Spur32BitMMLECoSimulator methodsFor: 'debug support' stamp: 'eem 5/9/2014 20:28'!
heapMapAtWord: address
	^heapMap heapMapAtWord: address asInteger! !

!Spur32BitMMLECoSimulator methodsFor: 'initialization' stamp: 'eem 11/14/2013 15:28'!
initialize
	super initialize.
	bootstrapping := false! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'cb 6/30/2016 14:41'!
intAt: byteAddress 
	^self longAt: byteAddress! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 9/21/2013 07:43'!
intAt: byteAddress put: a32BitValue
	^self longAt: byteAddress put: (a32BitValue bitAnd: 16rFFFFFFFF)! !

!Spur32BitMMLECoSimulator methodsFor: 'image segment in/out' stamp: 'eem 10/22/2014 10:49'!
loadImageSegmentFrom: segmentWordArray outPointers: outPointerArray
	self leakCheckImageSegments ifTrue:
		[self halt].
	^super loadImageSegmentFrom: segmentWordArray outPointers: outPointerArray! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 9/10/2013 15:36'!
long32At: byteAddress
	"Answer the 32-bit word at byteAddress which must be 0 mod 4."

	^self longAt: byteAddress! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 8/31/2013 09:07'!
long32At: byteAddress put: a32BitValue
	"Store the 32-bit value at byteAddress which must be 0 mod 4."

	^self longAt: byteAddress put: a32BitValue! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 4/19/2014 13:40'!
long64At: byteAddress
	"memory is a Bitmap, a 32-bit indexable array of bits"
	| hiWord loWord |
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	loWord := memory at: byteAddress // 4 + 1.
	hiWord := memory at: byteAddress // 4 + 2.
	^hiWord = 0
		ifTrue: [loWord]
		ifFalse: [(hiWord bitShift: 32) + loWord]! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 4/19/2014 13:40'!
long64At: byteAddress put: a64BitValue
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	self
		longAt: byteAddress put: (a64BitValue bitAnd: 16rffffffff);
		longAt: byteAddress + 4 put: a64BitValue >> 32.
	^a64BitValue! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 8/30/2013 15:50'!
longAt: byteAddress
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	byteAddress \\ 4 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 4 + 1! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 4/21/2015 11:05'!
longAt: byteAddress put: a32BitValue
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	"(byteAddress = 16r101348 and: [a32BitValue = 16r53]) ifTrue:
		[self halt]."
	"((byteAddress between: 16rda8ac and: 16rda8c0)
	 or: [byteAddress between: 16r8eb98 and: 16r8ebb0]) ifTrue:
		[self halt]."
	byteAddress \\ 4 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 4 + 1 put: a32BitValue! !

!Spur32BitMMLECoSimulator methodsFor: 'image segment in/out' stamp: 'eem 5/10/2017 15:44'!
mapOopsFrom: segStart to: segAddr outPointers: outPointerArray outHashes: savedOutHashes
	self leakCheckImageSegments ifTrue:
		[self halt].
	^super mapOopsFrom: segStart to: segAddr outPointers: outPointerArray outHashes: savedOutHashes! !

!Spur32BitMMLECoSimulator methodsFor: 'gc - global' stamp: 'eem 11/6/2013 16:16'!
markAndTrace: objOop
	"objOop = 16rB26020 ifTrue: [self halt].
	objOop = 16rB25FD8 ifTrue: [self halt].
	objOop = 16rB26010 ifTrue: [self halt]."
	^super markAndTrace: objOop! !

!Spur32BitMMLECoSimulator methodsFor: 'snapshot' stamp: 'eem 11/15/2013 09:47'!
memoryBaseForImageRead
	"Answer the address to read the image into.  Override so that when bootstrapping,
	 the segmentManager's segments are undisturbed in adjustSegmentSwizzlesBy:"
	^bootstrapping
		ifTrue: [0] 
		ifFalse: [super memoryBaseForImageRead]! !

!Spur32BitMMLECoSimulator methodsFor: 'accessing' stamp: 'eem 4/12/2017 11:43'!
needGCFlag: anInteger
	needGCFlag := anInteger ~= 0! !

!Spur32BitMMLECoSimulator methodsFor: 'initialization' stamp: 'eem 12/6/2014 11:20'!
nextLongFrom: aStream
	"Read a 32- or 64-bit quantity from the given (binary) stream."

	^aStream nextLittleEndianNumber: self wordSize! !

!Spur32BitMMLECoSimulator methodsFor: 'initialization' stamp: 'eem 12/6/2014 11:18'!
nextWord32From: aStream
	"Read a 32-bit quantity from the given (binary) stream."

	^aStream nextLittleEndianNumber: 4! !

!Spur32BitMMLECoSimulator methodsFor: 'spur bootstrap' stamp: 'eem 10/16/2013 12:20'!
numClassTablePages
	^numClassTablePages! !

!Spur32BitMMLECoSimulator methodsFor: 'accessing' stamp: 'eem 5/24/2014 23:38'!
parent

	^ parent! !

!Spur32BitMMLECoSimulator methodsFor: 'accessing' stamp: 'eem 5/24/2014 23:38'!
parent: anObject

	parent := anObject! !

!Spur32BitMMLECoSimulator methodsFor: 'image segment in/out' stamp: 'eem 6/28/2017 08:43'!
return: errCode restoringObjectsIn: firstArray savedFirstFields: savedFirstFields and: secondArray savedHashes: savedHashes
	self leakCheckImageSegments ifTrue:
		[self halt: errCode printString].
	^super return: errCode restoringObjectsIn: firstArray savedFirstFields: savedFirstFields and: secondArray savedHashes: savedHashes! !

!Spur32BitMMLECoSimulator methodsFor: 'debug support' stamp: 'eem 12/25/2016 11:29'!
runLeakCheckerFor: gcModes excludeUnmarkedObjs: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid
	(coInterpreter displayView isNil
	 and: [gcModes anyMask: checkForLeaks]) ifTrue:
		[coInterpreter transcript nextPutAll: 'leak-checking...'; flush].
	^super
		runLeakCheckerFor: gcModes
		excludeUnmarkedObjs: excludeUnmarkedObjs
		classIndicesShouldBeValid: classIndicesShouldBeValid! !

!Spur32BitMMLECoSimulator methodsFor: 'gc - global' stamp: 'eem 5/30/2018 11:24'!
scavengingGCTenuringIf: tenuringCriterion
	"If we're /not/ a clone, clone the VM and push it over the cliff.
	 If it survives, destroy the clone and continue.  We should be OK until next time."
	(self leakCheckNewSpaceGC
	 and: [parent isNil]) ifTrue:
		[coInterpreter cr; print: 'scavenge '; print: statScavenges; tab; flush.
		 CloneOnScavenge ifTrue:
			[coInterpreter cloneSimulation objectMemory scavengingGCTenuringIf: tenuringCriterion.
			 Smalltalk garbageCollect]].
	^super scavengingGCTenuringIf: tenuringCriterion! !

!Spur32BitMMLECoSimulator methodsFor: 'free space' stamp: 'eem 12/2/2013 13:59'!
setFree: o
	"o = 16rB34D40 ifTrue: [self halt]."
	super setFree: o! !

!Spur32BitMMLECoSimulator methodsFor: 'header access' stamp: 'eem 11/6/2013 14:15'!
setIsMarkedOf: objOop to: aBoolean
	"objOop = 16rB26020 ifTrue: [self halt]."
	super setIsMarkedOf: objOop to: aBoolean.
	"(aBoolean
	 and: [(self isContextNonImm: objOop)
	 and: [(coInterpreter
			checkIsStillMarriedContext: objOop
			currentFP: coInterpreter framePointer)
	 and: [(coInterpreter stackPages stackPageFor: (coInterpreter frameOfMarriedContext: objOop)) trace = 0]]]) ifTrue:
		[self halt]"! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 9/20/2013 15:08'!
shortAt: byteAddress
    "Return the half-word at byteAddress which must be even."
	| lowBits long |
	lowBits := byteAddress bitAnd: 2.
	long := self longAt: byteAddress - lowBits.
	^ lowBits = 2
		ifTrue: [ long bitShift: -16 ]
		ifFalse: [ long bitAnd: 16rFFFF ]! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 12/4/2013 19:17'!
shortAt: byteAddress put: a16BitValue
    "Return the half-word at byteAddress which must be even."
	| lowBits long longAddress |
	lowBits := byteAddress bitAnd: 2.
	lowBits = 0
		ifTrue: "storing into LS word"
			[long := self longAt: byteAddress.
			 self longAt: byteAddress
				put: ((long bitAnd: 16rFFFF0000) bitOr: a16BitValue)]
		ifFalse: "storing into MS word"
			[longAddress := byteAddress - 2.
			long := self longAt: longAddress.
			self longAt: longAddress
				put: ((long bitAnd: 16rFFFF) bitOr: (a16BitValue bitShift: 16))].
	^a16BitValue! !

!Spur32BitMMLECoSimulator methodsFor: 'indexing primitive support' stamp: 'eem 6/21/2021 12:59'!
shorten: objOop toIndexableSize: indexableSize
	self leakCheckShorten ifTrue:
		[parent ifNil:
			[| result |
			 result := coInterpreter cloneSimulation objectMemory shorten: objOop toIndexableSize: indexableSize.
			 Smalltalk garbageCollect]].
	^super shorten: objOop toIndexableSize: indexableSize! !

!Spur32BitMMLECoSimulator methodsFor: 'interpreter access' stamp: 'eem 10/27/2020 21:47'!
sizeFieldOfAlien: alienObj
	"Answer the first field of alienObj which is assumed to be an Alien of at least 8 bytes"
	<inline: #always>
	^(super sizeFieldOfAlien: alienObj) signedIntFromLong! !

!Spur32BitMMLECoSimulator methodsFor: 'float primitives' stamp: 'eem 9/10/2013 17:30'!
storeFloatAt: floatBitsAddress from: aFloat
	self long32At: floatBitsAddress put: (aFloat at: 2).
	self long32At: floatBitsAddress+4 put: (aFloat at: 1)! !

!Spur32BitMMLECoSimulator methodsFor: 'image segment in/out' stamp: 'eem 7/5/2017 16:59'!
storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots
	self leakCheckImageSegments ifTrue:
		[parent ifNil:
			[| result |
			 self halt.
			 result := coInterpreter cloneSimulation objectMemory storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots.
			 Smalltalk garbageCollect]].
	^super storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots! !

!Spur32BitMMLECoSimulator methodsFor: 'ad-hoc tests' stamp: 'eem 1/9/2015 10:47'!
testObjStackDo
	| size them seqA seqB seqC rs |
	ExpensiveAsserts := true.
	self initializeWeaklingStack; emptyObjStack: weaklingStack.
	self assert: (self topOfObjStack: weaklingStack) isNil.
	self assert: (self capacityOfObjStack: weaklingStack) >= ObjStackLimit.
	seqA := (1 to: ObjStackLimit * 5 // 2) collect: [:i| self integerObjectOf: i].
	seqA do: [:it| self noCheckPush: it onObjStack: weaklingStack].
	them := Set new.
	size := self objStack: weaklingStack from: 0 do: [:it| them add: it].
	self assert: size = seqA size.
	self assert: (them asSortedCollection asArray = seqA).
	self assert: (self isValidObjStack: weaklingStack).
	seqB := (ObjStackLimit * 5 // 2 + 1 to: ObjStackLimit * 10 // 2) collect: [:i| self integerObjectOf: i].
	self assert: seqA size = seqB size.
	rs := seqB readStream.
	them := Set new.
	size := self objStack: weaklingStack from: 0 do:
				[:it|
				them add: it.
				self noCheckPush: rs next onObjStack: weaklingStack].
	self assert: size = seqA size.
	self assert: rs atEnd.
	self objStack: weaklingStack from: size do:
		[:it| them add: it].
	seqC := (seqA, seqB) sort.
	self assert: them asSortedCollection asArray = seqC! !

!Spur32BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 10/9/2013 14:37'!
vmEndianness
	"1 = big, 0 = little"
	^0! !

Spur32BitMemoryManager subclass: #Spur32BitMMLESimulator
	instanceVariableNames: 'parent bootstrapping'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManagerSimulation'!

!Spur32BitMMLESimulator methodsFor: 'accessing' stamp: 'eem 11/14/2013 16:42'!
bootstrapping
	^bootstrapping! !

!Spur32BitMMLESimulator methodsFor: 'accessing' stamp: 'eem 11/15/2013 11:00'!
bootstrapping: aBoolean
	bootstrapping := aBoolean.
	segmentManager initForBootstrap! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:35'!
byteAt: byteAddress
	| lowBits long32 |
	lowBits := byteAddress bitAnd: 3.
	long32 := self long32At: byteAddress - lowBits.
	^(long32 bitShift: -8 * lowBits) bitAnd: 16rFF! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:34'!
byteAt: byteAddress put: byte
	| lowBits long32 longAddress mask value |
	lowBits := byteAddress bitAnd: 3.
	longAddress := byteAddress - lowBits.
	long32 := self long32At: longAddress.
	mask := (16rFF bitShift: 8 * lowBits) bitInvert.
	value := byte bitShift: 8 * lowBits.
	self long32At: longAddress put: ((long32 bitAnd: mask) bitOr: value).
	^byte! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 9/4/2013 16:23'!
byteAtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address."

	^self byteAt: pointer! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'nice 4/6/2016 00:14'!
byteSwapped32IfBigEndian: w
	^w bitAnd: 16rFFFFFFFF! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'nice 4/6/2016 00:15'!
byteSwapped64IfBigEndian: w
	^w bitAnd: 16rFFFFFFFFFFFFFFFF! !

!Spur32BitMMLESimulator methodsFor: 'debug support' stamp: 'eem 9/5/2021 18:55'!
checkHeapIntegrity: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid
	"Almost all of the time spent in SpurMemoryManager>>checkHeapIntegrity:classIndicesShouldBeValid: used to go into the asserts in fetchPointer:ofObject: in the simulator class overrides.
	 Since we know here the indices used are valid we temporarily remove them to claw back that performance."
	^self withSimulatorFetchPointerMovedAsideDo:
		[super checkHeapIntegrity: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid]! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 8/30/2013 11:47'!
endianness
	^#little! !

!Spur32BitMMLESimulator methodsFor: 'float primitives' stamp: 'eem 9/10/2013 17:29'!
fetchFloatAt: floatBitsAddress into: aFloat
	aFloat at: 2 put: (self long32At: floatBitsAddress).
	aFloat at: 1 put: (self long32At: floatBitsAddress+4)! !

!Spur32BitMMLESimulator methodsFor: 'object access' stamp: 'cb 6/11/2018 09:55'!
fetchPointer: fieldIndex ofObject: objOop
	self assert: (self isForwarded: objOop) not.
	self assert: (fieldIndex >= 0 and: [fieldIndex < (self numSlotsOfAny: objOop)
				or: [fieldIndex = 0 "forwarders"]]).
	^super fetchPointer: fieldIndex ofObject: objOop! !

!Spur32BitMMLESimulator methodsFor: 'object format' stamp: 'eem 12/28/2016 21:08'!
firstIndexableField: objOop
	"NOTE: overridden from SpurMemoryManager to add coercion to CArray, so please duplicate any changes.
	 There are only two important cases, both for objects with named inst vars, i.e. formats 2,3 & 5.
	 The first indexable field for formats 2 & 5 is the slot count (by convention, even though that's off the end
	 of the object).  For 3 we must go to the class."
	| fmt classFormat |
	<returnTypeC: #'void *'>
	fmt := self formatOf: objOop.
	fmt <= self lastPointerFormat ifTrue: "pointer; may need to delve into the class format word"
		[(fmt between: self indexablePointersFormat and: self weakArrayFormat) ifTrue:
			[classFormat := self formatOfClass: (self fetchClassOfNonImm: objOop).
			 ^self cCoerce: (self pointerForOop: objOop
												+ self baseHeaderSize
												+ ((self fixedFieldsOfClassFormat: classFormat) << self shiftForWord))
					to: #'oop *'].
		^self cCoerce: (self pointerForOop: objOop + self baseHeaderSize)
				to: #'oop *'].
	"All bit objects, and indeed CompiledMethod, though this is a no-no, start at 0"
	self assert: (fmt >= self sixtyFourBitIndexableFormat and: [fmt < self firstCompiledMethodFormat]).
	^self
		cCoerce: (self pointerForOop: objOop + self baseHeaderSize)
		to: (fmt < self firstByteFormat
				ifTrue:
					[fmt = self sixtyFourBitIndexableFormat
						ifTrue: ["64 bit field objects" #'long long *']
						ifFalse:
							[fmt < self firstShortFormat
								ifTrue: ["32 bit field objects" #'int *']
								ifFalse: ["16-bit field objects" #'short *']]]
				ifFalse: ["byte objects (including CompiledMethod" #'char *'])! !

!Spur32BitMMLESimulator methodsFor: 'spur bootstrap' stamp: 'eem 10/8/2013 11:36'!
freeLists
	^freeLists! !

!Spur32BitMMLESimulator methodsFor: 'gc - global' stamp: 'eem 5/30/2018 11:23'!
globalGarbageCollect
	"If we're /not/ a clone, clone the VM and push it over the cliff.
	 If it survives, destroy the clone and continue.  We should be OK until next time."
	parent ifNil:
		[coInterpreter cr; print: 'GC number '; print: statFullGCs; tab; flush.
		 CloneOnGC ifTrue:
			[coInterpreter cloneSimulation objectMemory globalGarbageCollect.
			 Smalltalk garbageCollect]].
	^super globalGarbageCollect! !

!Spur32BitMMLESimulator methodsFor: 'growing/shrinking memory' stamp: 'eem 11/14/2013 15:30'!
growOldSpaceByAtLeast: minAmmount
	"Attempt to grow memory by at least minAmmount.
	 Answer the size of the new segment, or nil if the attempt failed.
	 Override to not grow during the Spur image bootstrap."
	^bootstrapping ifFalse:
		[super growOldSpaceByAtLeast: minAmmount]! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 9/24/2013 13:50'!
halfWordHighInLong32: long32
	"Used by Balloon"

	^long32 bitAnd: 16rFFFF! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 9/24/2013 13:51'!
halfWordLowInLong32: long32
	"Used by Balloon"

	^long32 bitShift: -16! !

!Spur32BitMMLESimulator methodsFor: 'header format' stamp: 'eem 9/4/2013 11:08'!
headerForSlots: numSlots format: formatField classIndex: classIndex
	"The header format in LSB is
	 MSB:	| 2 bits				|
			| 22: identityHash	|
			| 8: slotSize			|
			| 3 bits				|
			| 5: format			|
			| 2 bits				|
			| 22: classIndex		| : LSB"
	self assert: (numSlots bitAnd: self numSlotsMask) = numSlots.
	self assert: (formatField bitAnd: self formatMask) = formatField.
	self assert: (classIndex bitAnd: self classIndexMask) = classIndex.
	^super headerForSlots: numSlots format: formatField classIndex: classIndex! !

!Spur32BitMMLESimulator methodsFor: 'debug support' stamp: 'eem 5/9/2014 20:21'!
heapMapAtWord: address
	^heapMap heapMapAtWord: address! !

!Spur32BitMMLESimulator methodsFor: 'compaction' stamp: 'eem 5/9/2014 15:22'!
inSortedFreeListLink: freeChunk to: nextFree given: prevFree
	"thisContext sender selector = #sweepToCoallesceFreeSpaceForPigCompactFrom: ifTrue:
		[| pit |
			pit := [:label :thing|
					coInterpreter print: label; space; printHex: thing.
					(thing ~= 0 and: [self isFreeObject: thing]) ifTrue:
						[coInterpreter print: ' (free) ']].
			pit value: 'link ' value: freeChunk.
			pit value: ' to ' value: nextFree.
			pit value: ' from ' value: prevFree.
			coInterpreter cr]."
	"freeChunk = 16r10B0730 ifTrue:
		[self halt]."
	super inSortedFreeListLink: freeChunk to: nextFree given: prevFree! !

!Spur32BitMMLESimulator methodsFor: 'initialization' stamp: 'eem 11/14/2013 15:28'!
initialize
	super initialize.
	bootstrapping := false! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'cb 6/30/2016 16:16'!
intAt: byteAddress 
	^self longAt: byteAddress ! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 9/21/2013 07:43'!
intAt: byteAddress put: a32BitValue
	^self longAt: byteAddress put: (a32BitValue bitAnd: 16rFFFFFFFF)! !

!Spur32BitMMLESimulator methodsFor: 'image segment in/out' stamp: 'eem 7/11/2017 18:05'!
loadImageSegmentFrom: segmentWordArray outPointers: outPointerArray
	self leakCheckImageSegments ifTrue:
		[parent ifNil:
			[| result |
			 self halt.
			 result := coInterpreter cloneSimulation objectMemory loadImageSegmentFrom: segmentWordArray outPointers: outPointerArray.
			 Smalltalk garbageCollect]].
	^super loadImageSegmentFrom: segmentWordArray outPointers: outPointerArray! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 9/10/2013 15:36'!
long32At: byteAddress
	"Answer the 32-bit word at byteAddress which must be 0 mod 4."

	^self longAt: byteAddress! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 8/31/2013 09:07'!
long32At: byteAddress put: a32BitValue
	"Store the 32-bit value at byteAddress which must be 0 mod 4."

	^self longAt: byteAddress put: a32BitValue! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 4/19/2014 13:40'!
long64At: byteAddress
	"memory is a Bitmap, a 32-bit indexable array of bits"
	| hiWord loWord |
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	loWord := memory at: byteAddress // 4 + 1.
	hiWord := memory at: byteAddress // 4 + 2.
	^hiWord = 0
		ifTrue: [loWord]
		ifFalse: [(hiWord bitShift: 32) + loWord]! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 4/19/2014 13:41'!
long64At: byteAddress put: a64BitValue
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	self
		longAt: byteAddress put: (a64BitValue bitAnd: 16rffffffff);
		longAt: byteAddress + 4 put: a64BitValue >> 32.
	^a64BitValue! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 8/30/2013 15:50'!
longAt: byteAddress
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	byteAddress \\ 4 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 4 + 1! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 11/16/2013 08:25'!
longAt: byteAddress put: a32BitValue
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	"(byteAddress = 16r183FB00 and: [a32BitValue = 16r3FFFFC]) ifTrue:
		[self halt]."
	"(byteAddress between: 16r33FBB8 and: 16r33FBCF) ifTrue:
		[self halt]."
	byteAddress \\ 4 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 4 + 1 put: a32BitValue! !

!Spur32BitMMLESimulator methodsFor: 'image segment in/out' stamp: 'eem 5/10/2017 15:44'!
mapOopsFrom: segStart to: segAddr outPointers: outPointerArray outHashes: savedOutHashes
	self leakCheckImageSegments ifTrue:
		[self halt].
	^super mapOopsFrom: segStart to: segAddr outPointers: outPointerArray outHashes: savedOutHashes! !

!Spur32BitMMLESimulator methodsFor: 'gc - global' stamp: 'eem 11/6/2013 16:16'!
markAndTrace: objOop
	"objOop = 16rB26020 ifTrue: [self halt].
	objOop = 16rB25FD8 ifTrue: [self halt].
	objOop = 16rB26010 ifTrue: [self halt]."
	^super markAndTrace: objOop! !

!Spur32BitMMLESimulator methodsFor: 'snapshot' stamp: 'eem 11/15/2013 09:47'!
memoryBaseForImageRead
	"Answer the address to read the image into.  Override so that when bootstrapping,
	 the segmentManager's segments are undisturbed in adjustSegmentSwizzlesBy:"
	^bootstrapping
		ifTrue: [0] 
		ifFalse: [super memoryBaseForImageRead]! !

!Spur32BitMMLESimulator methodsFor: 'compaction' stamp: 'eem 5/28/2014 07:42'!
moveARunOfObjectsStartingAt: startAddress upTo: limit
	| result |.
	"self checkTraversableSortedFreeList."
	result := super moveARunOfObjectsStartingAt: startAddress upTo: limit.
	"self checkTraversableSortedFreeList."
	^result! !

!Spur32BitMMLESimulator methodsFor: 'initialization' stamp: 'eem 11/18/2014 13:49'!
nextLongFrom: aStream
	"Read a 32- or 64-bit quantity from the given (binary) stream."

	^aStream nextLittleEndianNumber: self wordSize! !

!Spur32BitMMLESimulator methodsFor: 'initialization' stamp: 'eem 11/18/2014 13:47'!
nextWord32From: aStream
	"Read a 32-bit quantity from the given (binary) stream."

	^aStream nextLittleEndianNumber: 4! !

!Spur32BitMMLESimulator methodsFor: 'spur bootstrap' stamp: 'eem 10/16/2013 12:20'!
numClassTablePages
	^numClassTablePages! !

!Spur32BitMMLESimulator methodsFor: 'accessing' stamp: 'eem 5/24/2014 23:38'!
parent

	^ parent! !

!Spur32BitMMLESimulator methodsFor: 'accessing' stamp: 'eem 5/24/2014 23:38'!
parent: anObject

	parent := anObject! !

!Spur32BitMMLESimulator methodsFor: 'image segment in/out' stamp: 'eem 6/28/2017 08:43'!
return: errCode restoringObjectsIn: firstArray savedFirstFields: savedFirstFields and: secondArray savedHashes: savedHashes
	self leakCheckImageSegments ifTrue:
		[self halt: errCode printString].
	^super return: errCode restoringObjectsIn: firstArray savedFirstFields: savedFirstFields and: secondArray savedHashes: savedHashes! !

!Spur32BitMMLESimulator methodsFor: 'debug support' stamp: 'eem 12/25/2016 11:29'!
runLeakCheckerFor: gcModes excludeUnmarkedObjs: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid
	(coInterpreter displayView isNil
	 and: [gcModes anyMask: checkForLeaks]) ifTrue:
		[coInterpreter transcript nextPutAll: 'leak-checking...'; flush].
	^super
		runLeakCheckerFor: gcModes
		excludeUnmarkedObjs: excludeUnmarkedObjs
		classIndicesShouldBeValid: classIndicesShouldBeValid! !

!Spur32BitMMLESimulator methodsFor: 'gc - global' stamp: 'eem 5/30/2018 11:24'!
scavengingGCTenuringIf: tenuringCriterion
	"If we're /not/ a clone, clone the VM and push it over the cliff.
	 If it survives, destroy the clone and continue.  We should be OK until next time."
	(self leakCheckNewSpaceGC
	 and: [parent isNil]) ifTrue:
		[coInterpreter cr; print: 'scavenge '; print: statScavenges; tab; flush.
		 CloneOnScavenge ifTrue:
			[coInterpreter cloneSimulation objectMemory scavengingGCTenuringIf: tenuringCriterion.
			 Smalltalk garbageCollect]].
	^super scavengingGCTenuringIf: tenuringCriterion! !

!Spur32BitMMLESimulator methodsFor: 'header access' stamp: 'eem 10/29/2014 16:32'!
setIsMarkedOf: objOop to: aBoolean
	"objOop = 16rB26020 ifTrue: [self halt]."
	"(#(16r1971D0 16r196EE0 16r197048 16r197148) includes: objOop) ifTrue:
		[self halt]."
	super setIsMarkedOf: objOop to: aBoolean.
	"(aBoolean
	 and: [(self isContextNonImm: objOop)
	 and: [(coInterpreter
			checkIsStillMarriedContext: objOop
			currentFP: coInterpreter framePointer)
	 and: [(coInterpreter stackPages stackPageFor: (coInterpreter frameOfMarriedContext: objOop)) trace = 0]]]) ifTrue:
		[self halt]"! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 9/20/2013 15:08'!
shortAt: byteAddress
    "Return the half-word at byteAddress which must be even."
	| lowBits long |
	lowBits := byteAddress bitAnd: 2.
	long := self longAt: byteAddress - lowBits.
	^ lowBits = 2
		ifTrue: [ long bitShift: -16 ]
		ifFalse: [ long bitAnd: 16rFFFF ]! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 12/4/2013 19:17'!
shortAt: byteAddress put: a16BitValue
    "Return the half-word at byteAddress which must be even."
	| lowBits long longAddress |
	lowBits := byteAddress bitAnd: 2.
	lowBits = 0
		ifTrue: "storing into LS word"
			[long := self longAt: byteAddress.
			 self longAt: byteAddress
				put: ((long bitAnd: 16rFFFF0000) bitOr: a16BitValue)]
		ifFalse: "storing into MS word"
			[longAddress := byteAddress - 2.
			long := self longAt: longAddress.
			self longAt: longAddress
				put: ((long bitAnd: 16rFFFF) bitOr: (a16BitValue bitShift: 16))].
	^a16BitValue! !

!Spur32BitMMLESimulator methodsFor: 'indexing primitive support' stamp: 'eem 6/21/2021 13:00'!
shorten: objOop toIndexableSize: indexableSize
	self leakCheckShorten ifTrue:
		[parent ifNil:
			[| result |
			 result := coInterpreter cloneSimulation objectMemory shorten: objOop toIndexableSize: indexableSize.
			 Smalltalk garbageCollect]].
	^super shorten: objOop toIndexableSize: indexableSize! !

!Spur32BitMMLESimulator methodsFor: 'interpreter access' stamp: 'eem 10/27/2020 21:46'!
sizeFieldOfAlien: alienObj
	"Answer the first field of alienObj which is assumed to be an Alien of at least 8 bytes"
	<inline: #always>
	^(super sizeFieldOfAlien: alienObj) signedIntFromLong! !

!Spur32BitMMLESimulator methodsFor: 'float primitives' stamp: 'eem 9/10/2013 17:30'!
storeFloatAt: floatBitsAddress from: aFloat
	self long32At: floatBitsAddress put: (aFloat at: 2).
	self long32At: floatBitsAddress+4 put: (aFloat at: 1)! !

!Spur32BitMMLESimulator methodsFor: 'image segment in/out' stamp: 'eem 7/5/2017 16:59'!
storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots
	self leakCheckImageSegments ifTrue:
		[parent ifNil:
			[| result |
			 self halt.
			 result := coInterpreter cloneSimulation objectMemory storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots.
			 Smalltalk garbageCollect]].
	^super storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots! !

!Spur32BitMMLESimulator methodsFor: 'ad-hoc tests' stamp: 'eem 1/9/2015 10:47'!
testObjStackDo
	| size them seqA seqB seqC rs |
	ExpensiveAsserts := true.
	self initializeWeaklingStack; emptyObjStack: weaklingStack.
	self assert: (self topOfObjStack: weaklingStack) isNil.
	self assert: (self capacityOfObjStack: weaklingStack) >= ObjStackLimit.
	seqA := (1 to: ObjStackLimit * 5 // 2) collect: [:i| self integerObjectOf: i].
	seqA do: [:it| self noCheckPush: it onObjStack: weaklingStack].
	them := Set new.
	size := self objStack: weaklingStack from: 0 do: [:it| them add: it].
	self assert: size = seqA size.
	self assert: (them asSortedCollection asArray = seqA).
	self assert: (self isValidObjStack: weaklingStack).
	seqB := (ObjStackLimit * 5 // 2 + 1 to: ObjStackLimit * 10 // 2) collect: [:i| self integerObjectOf: i].
	self assert: seqA size = seqB size.
	rs := seqB readStream.
	them := Set new.
	size := self objStack: weaklingStack from: 0 do:
				[:it|
				them add: it.
				self noCheckPush: rs next onObjStack: weaklingStack].
	self assert: size = seqA size.
	self assert: rs atEnd.
	self objStack: weaklingStack from: size do:
		[:it| them add: it].
	seqC := (seqA, seqB) sort.
	self assert: them asSortedCollection asArray = seqC! !

!Spur32BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 10/9/2013 14:37'!
vmEndianness
	"1 = big, 0 = little"
	^0! !

!Spur32BitMemoryManager class methodsFor: 'class initialization' stamp: 'eem 12/26/2015 03:04'!
initializeWithOptions: optionsDictionary
	"SpurMemoryManager initializeWithOptions: Dictionary new"

	optionsDictionary at: #Spur32BitMemoryManager put: true.
	super initializeWithOptions: optionsDictionary! !

!Spur32BitMemoryManager class methodsFor: 'simulation only' stamp: 'eem 10/30/2013 13:51'!
isImmediate: anObject
	^anObject class == SmallInteger
	  or: [anObject class == Character]! !

!Spur32BitMemoryManager class methodsFor: 'translation' stamp: 'eem 11/20/2014 13:01'!
maxSmallInteger
	^16r3FFFFFFF! !

!Spur32BitMemoryManager class methodsFor: 'translation' stamp: 'eem 11/20/2014 13:03'!
minSmallInteger
	^-16r40000000! !

!Spur32BitMemoryManager class methodsFor: 'simulation only' stamp: 'eem 11/18/2014 17:49'!
numSmallIntegerTagBits
	^1! !

!Spur32BitMemoryManager class methodsFor: 'simulation only' stamp: 'eem 9/30/2013 12:46'!
simulatorClass
	^Spur32BitMMLESimulator! !

!Spur32BitMemoryManager class methodsFor: 'word size' stamp: 'eem 11/13/2014 15:24'!
wordSize
	"Answer the manager's word size, which is the size of an oop, and which
	 is assumed to be equivalent to the underlying machine's word size."
	^4! !

!Spur32BitMemoryManager methodsFor: 'object enumeration' stamp: 'eem 1/15/2018 09:28'!
addressAfter: objOop
	"Answer the address immediately following an object."
	<returnTypeC: #usqInt>
	| numSlots |
	numSlots := self rawNumSlotsOf: objOop.
	numSlots = 0 ifTrue: [^objOop + self allocationUnit + self baseHeaderSize].
	numSlots = self numSlotsMask ifTrue:
		[numSlots := self rawOverflowSlotsOf: objOop].
	^objOop + self baseHeaderSize + (numSlots + (numSlots bitAnd: 1) bitShift: self shiftForWord)! !

!Spur32BitMemoryManager methodsFor: 'allocation' stamp: 'eem 09/28/2015 18:48'!
allocateSlots: numSlots format: formatField classIndex: classIndex
	"Allocate an object with numSlots space.  If there is room beneath scavengeThreshold
	 allocate in newSpace, otherwise alocate in oldSpace.  If there is not room in newSpace
	 and a scavenge is not already scheduled, schedule a scavenge."
	<inline: true>
	| numBytes newObj |
	"Object headers are 8 bytes in length if the slot size fits in the num slots field (max implies overflow),
	 16 bytes otherwise (num slots in preceding word).
	 Objects always have at least one slot, for the forwarding pointer,
	 and are multiples of 8 bytes in length."
	numSlots >= self numSlotsMask
		ifTrue:
			[newObj := freeStart + self baseHeaderSize.
			 numBytes := self largeObjectBytesForSlots: numSlots]
		ifFalse:
			[newObj := freeStart.
			 numBytes := self smallObjectBytesForSlots: numSlots].
	
	freeStart + numBytes > scavengeThreshold ifTrue:
		[needGCFlag ifFalse: [self scheduleScavenge].
		 ^self allocateSlotsInOldSpace: numSlots bytes: numBytes format: formatField classIndex: classIndex].
	numSlots >= self numSlotsMask
		ifTrue: "for header parsing we put a saturated slot count in the prepended overflow size word"
			[self flag: #endianness.
			 self longAt: freeStart put: numSlots.
			 self longAt: freeStart + 4 put: self numSlotsMask << self numSlotsHalfShift.
			 self long64At: newObj put: (self headerForSlots: self numSlotsMask format: formatField classIndex: classIndex)]
		ifFalse:
			[self long64At: newObj put: (self headerForSlots: numSlots format: formatField classIndex: classIndex)].
	self assert: numBytes \\ self allocationUnit = 0.
	self assert: newObj \\ self allocationUnit = 0.
	freeStart := freeStart + numBytes.
	^newObj! !

!Spur32BitMemoryManager methodsFor: 'allocation' stamp: 'eem 12/27/2016 12:06'!
allocateSlotsForPinningInOldSpace: numSlots bytes: totalBytes format: formatField classIndex: classIndex
	"Answer the oop of a chunk of space in oldSpace with numSlots slots.  Try and
	 allocate in a segment that already includes pinned objects.  The header of the
	 result will have been filled-in but not the contents."
	<var: #totalBytes type: #usqInt>
	<inline: false>
	| chunk |
	chunk := self allocateOldSpaceChunkOfBytes: totalBytes
				   suchThat: [:f| (segmentManager segmentContainingObj: f) containsPinned].
	chunk ifNil:
		[chunk := self allocateOldSpaceChunkOfBytes: totalBytes.
		 chunk ifNil:
			[^nil].
		 (segmentManager segmentContainingObj: chunk) containsPinned: true].
	numSlots >= self numSlotsMask ifTrue: "for header parsing we put a saturated slot count in the prepended overflow size word"
		[self flag: #endianness.
		 self longAt: chunk put: numSlots.
		 self longAt: chunk + 4 put: self numSlotsMask << self numSlotsHalfShift.
		 self long64At: chunk + self baseHeaderSize
			 put: ((self headerForSlots: self numSlotsMask format: formatField classIndex: classIndex)
					bitOr: 1 << self pinnedBitShift).
		 self checkFreeSpace: GCModeNewSpace.
		 ^chunk + self baseHeaderSize].
	self long64At: chunk
		put: ((self headerForSlots: numSlots format: formatField classIndex: classIndex)
					bitOr: 1 << self pinnedBitShift).
	self checkFreeSpace: GCModeNewSpace.
	^chunk! !

!Spur32BitMemoryManager methodsFor: 'allocation' stamp: 'eem 12/27/2016 12:00'!
allocateSlotsInOldSpace: numSlots bytes: totalBytes format: formatField classIndex: classIndex
	"Answer the oop of a chunk of space in oldSpace with numSlots slots.  The header
	 will have been filled-in but not the contents.  If no memory is available answer nil."
	<var: #totalBytes type: #usqInt>
	<inline: false>
	| chunk |
	chunk := self allocateOldSpaceChunkOfBytes: totalBytes.
	chunk ifNil:
		[^nil].
	numSlots >= self numSlotsMask ifTrue: "for header parsing we put a saturated slot count in the prepended overflow size word"
		[self flag: #endianness.
		 self longAt: chunk put: numSlots.
		 self longAt: chunk + 4 put: self numSlotsMask << self numSlotsHalfShift.
		 self long64At: chunk + self baseHeaderSize
			put: (self headerForSlots: self numSlotsMask format: formatField classIndex: classIndex).
		 self checkFreeSpace: GCModeNewSpace.
		 ^chunk + self baseHeaderSize].
	self long64At: chunk put: (self headerForSlots: numSlots format: formatField classIndex: classIndex).
	self checkFreeSpace: GCModeNewSpace.
	^chunk! !

!Spur32BitMemoryManager methodsFor: 'snapshot' stamp: 'nice 4/2/2016 22:49'!
byteSwapped: w
	"Answer the given integer with its bytes in the reverse order."
	<api>
	<returnTypeC: #sqInt>
	^w byteSwap32! !

!Spur32BitMemoryManager methodsFor: 'object enumeration' stamp: 'eem 06/22/2021 15:31'!
bytesInBody: objOop
	"Answer the total number of bytes in an object including header and possible overflow size header."
	<returnTypeC: #usqInt>
	| headerNumSlots numSlots |
	headerNumSlots := self rawNumSlotsOf: objOop.
	numSlots := headerNumSlots = self numSlotsMask
					ifTrue: [self rawOverflowSlotsOf: objOop]
					ifFalse: [headerNumSlots = 0 ifTrue: [1] ifFalse: [headerNumSlots]].
	^numSlots + (numSlots bitAnd: 1) << self shiftForWord
	+ (headerNumSlots = self numSlotsMask
		ifTrue: [self baseHeaderSize + self baseHeaderSize]
		ifFalse: [self baseHeaderSize])! !

!Spur32BitMemoryManager methodsFor: 'object enumeration' stamp: 'eem 12/24/2016 10:45'!
bytesInObject: objOop given: rawNumSotsOfObjOop
	"Answer the total number of bytes in an object including header and possible overflow size header, given the value of the object's numSlots field."
	<returnTypeC: #usqInt>
	<inline: true>
	| numSlots |
	numSlots := rawNumSotsOfObjOop = self numSlotsMask
					ifTrue: [self rawOverflowSlotsOf: objOop]
					ifFalse: [rawNumSotsOfObjOop = 0 ifTrue: [1] ifFalse: [rawNumSotsOfObjOop]].
	^numSlots + (numSlots bitAnd: 1) << self shiftForWord
	+ (rawNumSotsOfObjOop = self numSlotsMask
		ifTrue: [self baseHeaderSize + self baseHeaderSize]
		ifFalse: [self baseHeaderSize])! !

!Spur32BitMemoryManager methodsFor: 'accessing' stamp: 'eem 3/12/2014 10:26'!
bytesPerOop
	^4! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 6/21/2021 22:25'!
changeClassOf: rcvr to: argClass
	"Attempt to change the class of the receiver to the argument given that the
	 format of the receiver matches the format of the argument.  If successful,
	 answer 0, otherwise answer an error code indicating the reason for failure. 
	 Fail if the format of the receiver is incompatible with the format of the argument,
	 or if the argument is a fixed class and the receiver's size differs from the size
	 that an instance of the argument should have."
	<inline: false>
	| classFormat fixedFields instFormat newFormat classIndex instSlots instBytes |
	(self isObjImmutable: rcvr) ifTrue:
		[^PrimErrNoModification].
	classFormat := self formatOfClass: argClass.
	fixedFields := self fixedFieldsOfClassFormat: classFormat.
	classFormat := self instSpecOfClassFormat: classFormat.
	instFormat := self formatOf: rcvr.

	"Fail for inability to access classIndex before making contexts snapshot-safe."
	(classIndex := self ensureBehaviorHash: argClass) < 0 ifTrue:
		[^classIndex negated].

	"Now check the instance for compatibility and compute odd bits if necessary."
	classFormat <= self lastPointerFormat
		ifTrue:
			[instFormat > self lastPointerFormat ifTrue:
				[^PrimErrInappropriate].
			 ((instSlots := self numSlotsOf: rcvr) < fixedFields
			  or: [instSlots > fixedFields and: [self isFixedSizePointerFormat: classFormat]]) ifTrue:
				[^PrimErrBadReceiver].
			 (instFormat = self indexablePointersFormat
			  and: [self isContextNonImm: rcvr]) ifTrue:
				[coInterpreter makeContextSnapshotSafe: rcvr].
			 newFormat := classFormat]
		ifFalse:
			["Fail if the class's format is somehow invalid."
			 classFormat ~= (self classFormatFromInstFormat: classFormat) ifTrue:
				[^PrimErrBadArgument].

			 instBytes := self numBytesOf: rcvr.
			 classFormat
				caseOf: {
				[self sixtyFourBitIndexableFormat]
					-> [(instBytes anyMask: 7) ifTrue: [^PrimErrBadReceiver].
						newFormat := classFormat].
				[self firstLongFormat] 		
					-> [(instBytes anyMask: 3) ifTrue: [^PrimErrBadReceiver].
						newFormat := classFormat].
				[self firstShortFormat] 		
					-> [(instBytes anyMask: 1) ifTrue: [^PrimErrBadReceiver].
						newFormat := classFormat + (2 - (instBytes >> 1) bitAnd: 1)].
				[self firstByteFormat]
					-> [newFormat := classFormat + (4 - instBytes bitAnd: 3)].
				[self firstCompiledMethodFormat]
					-> [classFormat ~= self firstCompiledMethodFormat ifTrue:
							[^PrimErrInappropriate].
						newFormat := instFormat] }
				otherwise: "bits instances cannot be adopted by pointer-like classes..."
					[^PrimErrInappropriate]].

	self set: rcvr classIndexTo: classIndex formatTo: newFormat.

	self assert: (self numBytesOf: rcvr) = (classFormat <= self lastPointerFormat
											ifTrue: [instSlots * self bytesPerOop]
											ifFalse: [instBytes]).
	"ok"
	^0! !

!Spur32BitMemoryManager methodsFor: 'gc - global' stamp: 'eem 06/22/2021 15:31'!
coalesce: obj1 and: obj2
	| header1NumSlots header2NumSlots obj2slots newNumSlots |
	header1NumSlots := self rawNumSlotsOf: obj1.
	header2NumSlots := self rawNumSlotsOf: obj2.

	"compute total number of slots in obj2, including header"
	obj2slots := header2NumSlots = self numSlotsMask
					ifTrue: [(self rawOverflowSlotsOf: obj2) + (2 * self baseHeaderSize / self wordSize)]
					ifFalse: [(header2NumSlots = 0 ifTrue: [1] ifFalse: [header2NumSlots]) + (self baseHeaderSize / self wordSize)].
	obj2slots := obj2slots + (obj2slots bitAnd: 1).
	self assert: obj2slots * self wordSize = (self bytesInBody: obj2).

	"if obj1 already has a double header things are simple..."
	header1NumSlots = self numSlotsMask ifTrue:
		[self rawOverflowSlotsOf: obj1 put: obj2slots + (self rawOverflowSlotsOf: obj1).
		 ^obj1].

	"compute total number of slots in obj1, excluding header"
	header1NumSlots := header1NumSlots = 0
							ifTrue: [2]
							ifFalse: [header1NumSlots + (header1NumSlots bitAnd: 1)].
	self assert: header1NumSlots * self wordSize + self baseHeaderSize = (self bytesInBody: obj1).
	newNumSlots := obj2slots + header1NumSlots.

	"if obj1 still only requires a single header things are simple..."
	newNumSlots < self numSlotsMask ifTrue:
		[self rawNumSlotsOf: obj1 put: newNumSlots.
		 ^obj1].

	"convert from single to double header..."
	newNumSlots := newNumSlots - (self baseHeaderSize / self wordSize).
	self longAt: obj1 + self baseHeaderSize
			put: (self longAt: obj1);
		longAt: obj1 + 4 + self baseHeaderSize
			put: ((self longAt: obj1 + 4) bitOr: self numSlotsMask << self numSlotsHalfShift).
	self longAt: obj1
			put: newNumSlots.
	self longAt: obj1 + 4
			put: self numSlotsMask << self numSlotsHalfShift.
	^obj1 + self baseHeaderSize! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 12/1/2014 10:56'!
dbgFloatValueOf: oop
	"Answer the C double precision floating point value of the argument,
	 or if it is not, answer 0."

	| isFloat result |
	<returnTypeC: #double>
	<var: #result type: #double>
	isFloat := self isFloatInstance: oop.
	isFloat ifTrue:
		[self cCode: '' inSmalltalk: [result := Float new: 2].
		 self fetchFloatAt: oop + self baseHeaderSize into: result.
		 ^result].
	^0.0! !

!Spur32BitMemoryManager methodsFor: 'snapshot' stamp: 'eem 10/27/2016 12:14'!
defaultEdenBytes
	"Answer the default amount of memory to allocate for the eden space.
	 The actual value can be set via vmParameterAt: and/or a preference in the ini file.
	 The shootout tests seem to plateau at 5 or 6Mb.

	 Originally, both the 32-bit and 64-bit versions used the same 4Mb default.  Measuring
	 the simulator on image start-up, the 64-bit system's eden at the same point in start-up
	 (the first copyBits) is only 8% larger in bytes because it allocates 26% fewer objects.
	 Some 21% of the objects in the 32-bit version's eden are large integers and floats that
	 are representable as 64-bit immediates.

	 But when running benchmarks such as the computer language shootout's binary trees,
	 using the same amount of memory for the 64-bit system causes a significant slow-down
	 and a lot of compactions.  So we now use 4Mb for 32-bits and 7Mb for 64-bits."
	<inline: false>
	^4 * 1024 * 1024

	"Here are the expressions used to collect the above-mentioned figures:"
	"freeStart - self newSpaceStart"
	"| n | n := 0. self allNewSpaceObjectsDo: [:o| n := n + 1]. n"
	"| f | f := 0. self allNewSpaceObjectsDo: [:o| (self isFloatObject: o) ifTrue: [f := f + 1]]. f"
	"| l | l := 0. self allNewSpaceObjectsDo: [:o|
		(((self classIndexOf: o)
			between: ClassLargeNegativeIntegerCompactIndex
			and: ClassLargePositiveIntegerCompactIndex)
		 and: [(self numBytesOfBytes: o) <= 8
		 and: [self isIntegerValue: (coInterpreter signed64BitValueOf: o)]]) ifTrue: [l := l + 1]]. l"! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 2/24/2020 15:18'!
fetchClassTagOf: oop
	"Answer the tag used in lookup caches for a receiver.  This is the receiver's classIndex."
	<api>
	| tagBits |
	^(tagBits := oop bitAnd: self tagMask) ~= 0
		ifTrue: [(tagBits bitAnd: 1) ~= 0 ifTrue: [1] ifFalse: [tagBits]]
		ifFalse: [self classIndexOf: oop]! !

!Spur32BitMemoryManager methodsFor: 'object access' stamp: 'eem 12/17/2014 14:14'!
fetchLong32: fieldIndex ofFloatObject: oop
	"index by word size, and return a pointer as long as the word size"
	<inline: true>
	^self fetchLong32: fieldIndex ofObject: oop! !

!Spur32BitMemoryManager methodsFor: 'instantiation' stamp: 'eem 11/13/2014 12:15'!
fillObj: objOop numSlots: numSlots with: fillValue
	<inline: true>
	<var: #p type: #usqInt>
	self assert: (self oop: objOop + self baseHeaderSize + (numSlots * self bytesPerOop) - 1
					isLessThan: (self addressAfter: objOop)).
	(objOop + self baseHeaderSize) asUnsignedInteger
		to: (objOop + self baseHeaderSize + (numSlots * self bytesPerOop) - 1) asUnsignedInteger
		by: self allocationUnit
		do: [:p|
			self longAt: p put: fillValue;
				longAt: p + 4 put: fillValue]! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'RonieSalgado 5/1/2016 20:56'!
floatObjectOf: aFloat
	| newFloatObj |
	<inline: false>
	<var: #aFloat type: #double>
	<api>
	newFloatObj := self
						eeInstantiateSmallClassIndex: ClassFloatCompactIndex
						format: self firstLongFormat
						numSlots: (self sizeof: #double) / self bytesPerOop.
	self storeFloatAt: newFloatObj + self baseHeaderSize from: aFloat.
	^newFloatObj! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 1/18/2020 18:02'!
floatObjectOfBits: doubleFloatBits
	<var: 'doubleFloatBits' type: #sqLong>
	| newFloatObj |
	newFloatObj := self
						eeInstantiateSmallClassIndex: ClassFloatCompactIndex
						format: self firstLongFormat
						numSlots: (self sizeof: #double) / self bytesPerOop.
	self storeLong64: 0 ofObject: newFloatObj withValue: doubleFloatBits.
	^newFloatObj! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 1/18/2020 17:57'!
floatValueBitsOf: floatOop
	"Answer the 64-bit value of the argument."
	self assert: (self isFloatInstance: floatOop).
	^self fetchLong64: 0 ofObject: floatOop! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/14/2019 13:43'!
floatValueOf: oop
	"Answer the C double precision floating point value of the argument,
	 or fail if it is not a Float, and answer 0.
	 Note: May be called by translated primitive code."
	<api>
	<returnTypeC: #double>
	| isFloat result |
	<var: #result type: #double>
	isFloat := self isFloatInstance: oop.
	isFloat ifTrue:
		[self cCode: '' inSmalltalk: [result := Float new: 2].
		 self fetchFloatAt: oop + self baseHeaderSize into: result.
		 ^result].
	coInterpreter primitiveFail.
	^0.0! !

!Spur32BitMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 6/22/2021 14:04'!
hackSlimBridgeTo: objOop at: startAddress
	"This is a horrible hack for getting to the first object in eden if pastSpace is almost full.
	 If there is only one (64-bit) word at the end of pastSpace there is no room for a full
	 bridge, but there is room for this hack."
	self flag: #endianness.
	self assert: (self oop: startAddress isLessThan: objOop).
	self assert: (self oop: objOop isLessThanOrEqualTo: oldSpaceStart).
	self assert: (startAddress + self allocationUnit = objOop
				or: [startAddress + self allocationUnit + self allocationUnit = objOop]).
	self longAt: startAddress
		put: (startAddress + self allocationUnit = objOop ifTrue: [0] ifFalse: [1]).
	self longAt: startAddress + 4
		put: self numSlotsMask << self numSlotsHalfShift! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 2/12/2015 17:34'!
hasSixtyFourBitImmediates
	"The V3 64-bit memory manager has 64-bit oops, but its SmallIntegers only have
	 31 bits of precision.  The Spur 64-bit memory manager has 61-bit immediates."
	^false! !

!Spur32BitMemoryManager methodsFor: 'method access' stamp: 'eem 6/7/2016 14:53'!
headerIndicatesAlternateBytecodeSet: methodHeader
	"A negative header selects the alternate bytecode set."
	<api>
	<inline: true>
	^methodHeader signedIntFromLong < 0! !

!Spur32BitMemoryManager methodsFor: 'immediates' stamp: 'eem 3/10/2016 12:42'!
imageFormatVersion
	"Return a magic constant that changes when the image format changes.
	 Since the image reading code uses this to detect byte ordering, one
	 must avoid version numbers that are invariant under byte reversal.
	 N.B. Bit 4 (=16) is the isSpur bit"
	^6521! !

!Spur32BitMemoryManager methodsFor: 'free space' stamp: 'eem 11/13/2014 12:15'!
initFreeChunkWithBytes: numBytes at: address
	<var: #numBytes type: #usqLong>
	| numSlots |
	"must have room for a header (single or double) plus the next free pointer"
	self assert: (numBytes \\ self allocationUnit = 0
				 and: [numBytes >= (self baseHeaderSize + self bytesPerOop)]).
	self flag: #endianness.
	"double header"
	numBytes >= ((self numSlotsMask << self shiftForWord) + self baseHeaderSize) ifTrue:
		[numSlots := numBytes - self baseHeaderSize - self baseHeaderSize >> self shiftForWord.
		 self longAt: address put: numSlots;
			longAt: address + 4 put: self numSlotsMask << self numSlotsHalfShift;
			longAt: address + 8 put: 0; "0's classIndex; 0 = classIndex of free chunks"
			longAt: address + 12 put: self numSlotsMask << self numSlotsHalfShift.
		^address + 8].
	"single header"
	numSlots := numBytes - self baseHeaderSize >> self shiftForWord.
	self assert: numSlots < self numSlotsMask.
	self longAt: address put: 0; "0's classIndex; 0 = classIndex of free chunks"
		longAt: address + 4 put: numSlots << self numSlotsHalfShift.
	^address! !

!Spur32BitMemoryManager methodsFor: 'object enumeration' stamp: 'eem 9/24/2020 14:48'!
initSegmentBridgeWithBytes: numBytes at: address
	<var: #numBytes type: #usqLong>
	| numSlots |
	"Must have room for a double header or a short object with the forwarding slot (16 bytes either way)."
	self assert: (numBytes \\ self allocationUnit = 0
				 and: [numBytes >= (self baseHeaderSize + self baseHeaderSize)]).
	numSlots := numBytes - self baseHeaderSize - self baseHeaderSize >> self shiftForWord.
	self flag: #endianness.
	numSlots = 0
		ifTrue: "short bridge for adjacent segments"
			[self longAt: address put: (1 << self pinnedBitShift)
									+ (self wordIndexableFormat << self formatShift)
									+ self segmentBridgePun;
				longAt: address + 4 put: (1 << self markedBitHalfShift)]
		ifFalse: "long bridge"
			[self longAt: address put: numSlots;
				longAt: address + 4 put: self numSlotsMask << self numSlotsHalfShift;
				longAt: address + 8 put: (1 << self pinnedBitShift)
										+ (self wordIndexableFormat << self formatShift)
										+ self segmentBridgePun;
				longAt: address + 12 put: self numSlotsMask << self numSlotsHalfShift
										+ (1 << self markedBitHalfShift)]! !

!Spur32BitMemoryManager methodsFor: 'instantiation' stamp: 'eem 8/22/2017 14:46'!
instantiateClass: classObj indexableSize: nElements
	<api>
	<var: #nElements type: #usqInt>
	"Allocate an instance of a variable class, excepting CompiledMethod."
	| instSpec classFormat numSlots classIndex newObj fillValue |
	classFormat := self formatOfClass: classObj.
	instSpec := self instSpecOfClassFormat: classFormat.
	classIndex := self rawHashBitsOf: classObj.
	fillValue := 0.
	instSpec caseOf: {
		[self arrayFormat]	->
			[numSlots := nElements.
			 fillValue := nilObj].
		[self indexablePointersFormat]	->
			[numSlots := (self fixedFieldsOfClassFormat: classFormat) + nElements.
			 fillValue := nilObj].
		[self weakArrayFormat]	->
			[numSlots := (self fixedFieldsOfClassFormat: classFormat) + nElements.
			 fillValue := nilObj].
		[self sixtyFourBitIndexableFormat]	->
			[nElements > (self maxSlotsForAlloc / 2) ifTrue:
				[coInterpreter primitiveFailFor: PrimErrUnsupported.
				 ^nil].
			 numSlots := nElements * 2].
		[self firstLongFormat]	->
			[(classIndex = ClassFloatCompactIndex and: [nElements ~= 2]) ifTrue:
				[coInterpreter primitiveFailFor: PrimErrBadReceiver.
				 ^nil].
			 numSlots := nElements].
		[self firstShortFormat]	->
			[numSlots := nElements + 1 // 2.
			 instSpec := instSpec + (nElements bitAnd: 1)].
		[self firstByteFormat]	->
			[numSlots := nElements + 3 // 4.
			 instSpec := instSpec + (4 - nElements bitAnd: 3)] }
		otherwise: "non-indexable"
			["Some Squeak images include funky fixed subclasses of abstract variable
			  superclasses. e.g. DirectoryEntry as a subclass of ArrayedCollection.
			  The (Threaded)FFIPlugin expects to be able to instantiate ExternalData via
			  this method.
			  Hence allow fixed classes to be instantiated here iff nElements = 0."
			 (nElements ~= 0 or: [instSpec > self lastPointerFormat]) ifTrue:
				[^nil].
			 numSlots := self fixedFieldsOfClassFormat: classFormat.
			 fillValue := nilObj].
	classIndex = 0 ifTrue:
		[classIndex := self ensureBehaviorHash: classObj.
		 classIndex < 0 ifTrue:
			[coInterpreter primitiveFailFor: classIndex negated.
			 ^nil]].
	numSlots > self maxSlotsForNewSpaceAlloc
		ifTrue:
			[numSlots > self maxSlotsForAlloc ifTrue:
				[coInterpreter primitiveFailFor: PrimErrUnsupported.
				 ^nil].
			 newObj := self allocateSlotsInOldSpace: numSlots format: instSpec classIndex: classIndex]
		ifFalse:
			[newObj := self allocateSlots: numSlots format: instSpec classIndex: classIndex].
	newObj ifNotNil:
		[self fillObj: newObj numSlots: numSlots with: fillValue].
	^newObj! !

!Spur32BitMemoryManager methodsFor: 'instantiation' stamp: 'eem 8/12/2014 19:53'!
instantiateCompiledMethodClass: classObj indexableSize: nElements
	<var: #nElements type: #usqInt>
	"Allocate an instance of a CompiledMethod class."
	<inline: true>
	| instSpec classFormat numSlots classIndex newObj |
	classFormat := self formatOfClass: classObj.
	instSpec := self instSpecOfClassFormat: classFormat.
	instSpec ~= self firstCompiledMethodFormat ifTrue:
		[^nil].
	numSlots := nElements + 3 // 4.
	instSpec := instSpec + (4 - nElements bitAnd: 3).
	classIndex := self ensureBehaviorHash: classObj.
	classIndex < 0 ifTrue:
		[coInterpreter primitiveFailFor: classIndex negated.
		 ^nil].
	numSlots > self maxSlotsForNewSpaceAlloc
		ifTrue:
			[numSlots > self maxSlotsForAlloc ifTrue:
				[^nil].
			 newObj := self allocateSlotsInOldSpace: numSlots format: instSpec classIndex: classIndex]
		ifFalse:
			[newObj := self allocateSlots: numSlots format: instSpec classIndex: classIndex].
	newObj ifNotNil:
		[self fillObj: newObj numSlots: numSlots with: 0].
	^newObj! !

!Spur32BitMemoryManager methodsFor: 'immediates' stamp: 'nice 2/16/2020 19:43'!
integerObjectOf: value
	"Convert the integer value, assumed to be in SmallInteger range, into a tagged SmallInteger object.
	 In C, use a shift and an add to set the tag bit.
	 In Smalltalk we have to work harder because the simulator works with strictly positive bit patterns."
	<returnTypeC: #sqInt>
	^self
		cCode: [value asUnsignedInteger << 1 + 1]
		inSmalltalk: [value >= 0
						ifTrue: [value << 1 + 1]
						ifFalse: [16r80000000 + value << 1 + 1]]! !

!Spur32BitMemoryManager methodsFor: 'immediates' stamp: 'nice 7/8/2016 02:01'!
integerObjectOfCharacterObject: oop
	"Immediate characters are unsigned"
	^(self cCoerceSimple: oop to: #usqInt) >> 1! !

!Spur32BitMemoryManager methodsFor: 'immediates' stamp: 'eem 9/3/2013 17:39'!
integerValueOf: oop
	"Translator produces 'oop >> 1'"
	^(oop bitShift: -31) = 1 "tests top bit"
		ifTrue: "negative"
			[((oop bitShift: -1) bitAnd: 16r3FFFFFFF) - 16r3FFFFFFF - 1  "Faster than -16r40000000 (a LgInt)"]
		ifFalse: "positive"
			[oop bitShift: -1]! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/21/2014 17:02'!
isFloatInstance: oop
	<inline: true>
	^(self isNonImmediate: oop)
	  and: [(self classIndexOf: oop) = ClassFloatCompactIndex]! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 6/5/2017 10:31'!
isFloatOrInt: anOop
	"Answer if anOop is either a SmallInteger or a Float."

	<inline: true>
	^(self isImmediate: anOop)
		ifTrue: [self isIntegerObject: anOop]
		ifFalse: [(self classIndexOf: anOop) = ClassFloatCompactIndex]! !

!Spur32BitMemoryManager methodsFor: 'object testing' stamp: 'eem 3/9/2016 13:56'!
isImmediateCharacter: oop
	<inline: true>
	^(oop bitAnd: self tagMask) = self characterTag! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 12/1/2014 11:19'!
isImmediateFloat: oop
	^false! !

!Spur32BitMemoryManager methodsFor: 'object testing' stamp: 'eem 3/9/2016 13:58'!
isIntegerObject: oop
	^(oop bitAnd: self smallIntegerTag) ~= 0! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'nice 3/25/2015 00:34'!
isIntegerValue: intValue
	"Answer if the given value can be represented as a Smalltalk integer value.
	 In C, use a shift and XOR to set the sign bit if and only if the top two bits of the given
	 value are the same, then test the sign bit. Note that the top two bits are equal for
	 exactly those integers in the range that can be represented in 31-bits or 63-bits."
	<api>
	^self
		cCode: [(intValue asUnsignedInteger bitXor: (intValue asUnsignedInteger << 1)) asInteger >= 0]
		inSmalltalk: [intValue >= self minSmallInteger and: [intValue <= self maxSmallInteger]]! !

!Spur32BitMemoryManager methodsFor: 'free space' stamp: 'cb 10/5/2018 12:47'!
isLilliputianSize: chunkBytes
	"Allocation unit, the minimum size, is enough for 2 pointers"
	^ false! !

!Spur32BitMemoryManager methodsFor: 'header access' stamp: 'eem 11/18/2013 13:51'!
isMarked: objOop
	<api>
	self flag: #endianness.
	^((self longAt: objOop + 4) >> self markedBitHalfShift bitAnd: 1) ~= 0! !

!Spur32BitMemoryManager methodsFor: 'object testing' stamp: 'eem 10/11/2013 17:14'!
isNonIntegerObject: oop
	^(oop bitAnd: 1) = 0! !

!Spur32BitMemoryManager methodsFor: 'object testing' stamp: 'eem 9/10/2013 15:35'!
isWordsNonImm: objOop
	"Answer if the argument contains only indexable words (no oops). See comment in formatOf:"

	^(self formatOf: objOop) = self firstLongFormat! !

!Spur32BitMemoryManager methodsFor: 'allocation' stamp: 'eem 11/13/2014 12:15'!
largeObjectBytesForSlots: numSlots
	"Answer the total number of bytes in an object with an overflow header, including header bytes."
	<returnTypeC: #usqInt>
	^self baseHeaderSize + self baseHeaderSize "double header"
	+ (numSlots + (numSlots bitAnd: 1) * self bytesPerOop) "roundTo allocationUnit"! !

!Spur32BitMemoryManager methodsFor: 'object access' stamp: 'eem 2/6/2015 11:18'!
lengthOf: objOop format: fmt
	"Answer the number of indexable units in the given object.
	 For a CompiledMethod, the size of the method header (in bytes)
	 should be subtracted from the result of this method."
	| numSlots |
	<inline: true>
	numSlots := self numSlotsOfAny: objOop. "don't let forwarders freak us out..."
	fmt <= self ephemeronFormat ifTrue:
		[^numSlots].
	fmt >= self firstByteFormat ifTrue: "bytes, including CompiledMethod"
		[^numSlots << self shiftForWord - (fmt bitAnd: 7)].
	fmt >= self firstShortFormat ifTrue:
		[^numSlots << (self shiftForWord - 1) - (fmt bitAnd: 3)].
	fmt >= self firstLongFormat ifTrue:
		[^numSlots << (self shiftForWord - 2) - (fmt bitAnd: 1)].
	fmt = self sixtyFourBitIndexableFormat ifTrue:
		[^numSlots >> 1].
	"fmt = self forwardedFormat"
	^0! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 9/5/2019 09:27'!
loadFloatOrIntFrom: floatOrInt
	"If floatOrInt is an integer and we enable mixed arithmetic in primitives, then convert it to a C double float and return it.
	 If it is a Float, then load its value and return it.
	 Otherwise fail -- ie return with primErrorCode non-zero."

	<inline: true>
	<returnTypeC: #double>

	(coInterpreter primitiveDoMixedArithmetic and: [self isIntegerObject: floatOrInt]) ifTrue:
		[^(self integerValueOf: floatOrInt) asFloat].
	^self floatValueOf: floatOrInt! !

!Spur32BitMemoryManager methodsFor: 'header format' stamp: 'eem 11/14/2014 12:13'!
logBytesPerOop
	^2! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 1/8/2015 14:41'!
maxCInteger
	^16rFFFFFFFF! !

!Spur32BitMemoryManager methodsFor: 'instantiation' stamp: 'eem 6/28/2014 16:52'!
maxSlotsForAlloc
	"Answer the maximum number of slots we are willing to attempt to allocate in an object.
	 Must fit in 32-bits; c.f. bytesInObject:. Chosen so that maxSlotsForAlloc * self bytesPerWord is +ve."
	^512*1024*1024-1! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/19/2014 15:54'!
maxSmallInteger
	^16r3FFFFFFF! !

!Spur32BitMemoryManager methodsFor: 'simulation' stamp: 'eem 11/18/2016 16:19'!
memoryClass
	<doNotGenerate>
	^self endianness == #little
		ifTrue: [LittleEndianBitmap]
		ifFalse: [Bitmap]! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 6/5/2017 11:30'!
methodDictionaryHash: oop mask: mask
	<inline: true>
	^mask bitAnd: ((self isImmediate: oop)
						ifTrue: [(self isIntegerObject: oop)
									ifTrue: [self integerValueOf: oop]
									ifFalse: [self characterValueOf: oop]]
						ifFalse: [self rawHashBitsOf: oop]) "If no hash then it isn't in the dictionary..."! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/19/2014 15:12'!
minSmallInteger
	^-16r40000000! !

!Spur32BitMemoryManager methodsFor: 'simulation' stamp: 'eem 7/14/2015 12:27'!
newInputEventAccessorOfSize: numElements
	<doNotGenerate>
	^CArrayAccessor on: (IntegerArray new: 8)! !

!Spur32BitMemoryManager methodsFor: 'indexing primitive support' stamp: 'eem 6/22/2021 14:03'!
normalisedFormatFor: objOop indexableSize: indexableSize
	"Answer the format for an objOop clone with indexableSize.  Needs to work for longFormat and arrayFormat only."
	^self classFormatFromInstFormat: (self formatOf: objOop)! !

!Spur32BitMemoryManager methodsFor: 'free space' stamp: 'eem 11/18/2014 09:08'!
numFreeLists
	"Answer the number of free lists.  We use freeListsMask, a bitmap, to avoid
	 reading empty list heads.  This should fit in a machine word to end up in a
	 register during free chunk allocation."
	^32! !

!Spur32BitMemoryManager methodsFor: 'allocation' stamp: 'eem 6/21/2021 11:53'!
numSlotsForShortening: objOop toIndexableSize: indexableSize
	<inline: true>
	^(self formatOf: objOop)
		caseOf:
			{ [self arrayFormat]		-> [indexableSize].
			  [self firstLongFormat]		-> [self numSlotsForBytes: indexableSize * 4] }
		otherwise:
			[self assertf: 'invalid format for shortening'. 0]! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/19/2014 13:13'!
numSmallIntegerBits
	^31! !

!Spur32BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/18/2014 17:42'!
numSmallIntegerTagBits
	^1! !

!Spur32BitMemoryManager methodsFor: 'object access' stamp: 'eem 12/13/2015 10:16'!
numTagBits
	<api>
	<cmacro>
	^2! !

!Spur32BitMemoryManager methodsFor: 'object enumeration' stamp: 'eem 12/18/2016 16:58'!
objectAfter: objOop limit: limit
	"Object parsing.
	1. all objects have at least a word following the header, for a forwarding pointer.
	2. objects with an overflow size have a preceding word with a saturated numSlots.  If the word
	   following an object doesn't have a saturated numSlots field it must be a single-header object.
	   If the word following does have a saturated numSlots it must be the overflow size word."
	<inline: true>
	| followingWordAddress followingWord |
	followingWordAddress := self addressAfter: objOop.
	(self oop: followingWordAddress isGreaterThanOrEqualTo: limit) ifTrue:
		[^limit].
	self flag: #endianness.
	followingWord := self longAt: followingWordAddress + 4.
	^followingWord >> self numSlotsHalfShift = self numSlotsMask
		ifTrue: [followingWordAddress + self baseHeaderSize]
		ifFalse: [followingWordAddress]! !

!Spur32BitMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 9/6/2021 19:15'!
objectAfterMaybeSlimBridge: objOop limit: limit
	"Object parsing.
	1. all objects have at least a word following the header, for a forwarding pointer.
	2. objects with an overflow size have a preceding word with a saturated numSlots.  If the word
	   following an object doesn't have a saturated numSlots field it must be a single-header object.
	   If the word following does have a saturated numSlots it must be the overflow size word.
	This variation on objectAfter:limit: allows for a single (64-bit) word bridge which may be needed
	to bridge from an almost full pastSpace to eden.  It is only used in the flat enumerators that use
	startAddressForBridgedHeapEnumeration and enumerate over pastSpace, eden and oldSpace
	in that order.  Note that the order for allObjects, and allInstances enumerates over oldSpace first.

	This hack is cheap.  It increases the size of the objectAfter code, but saves two extra copies of
	the inner loop, since the inner loop now enumerates over all of pastSpace, eden and oldSpace.
	The test for a slim bridge is only performed if applied to an overflow header, and typically only
	1 in 400 objects have overflow headers in 32-bits, 1 in 500 in 64-bits.  The complication is that
	image segment loading evaporates the word array by setting the overflow slots to 1, and this
	is ambiguous with a slimbridge.  The resolution is that if the segmentArray has an overflow header,
	and is in new space, then it slot size can be zeroed and its overflow header changed to a slimbridge."
	<inline: true>
	| followingWordAddress followingWord |
	followingWordAddress := self addressAfter: objOop.
	(self oop: followingWordAddress isGreaterThanOrEqualTo: limit) ifTrue:
		[^limit].
	self flag: #endianness.
	followingWord := self longAt: followingWordAddress + 4.
	^followingWord >> self numSlotsHalfShift = self numSlotsMask
		ifTrue: [((self oop: objOop isLessThan: oldSpaceStart)
				 and: [1 = (self longAt: followingWordAddress)]) "i.e. the raw overflow slots in the overflow word"
					ifTrue: [followingWordAddress + self baseHeaderSize + self baseHeaderSize] "skip the one word slimbridge"
					ifFalse: [followingWordAddress + self baseHeaderSize]]
		ifFalse: [followingWordAddress]! !

!Spur32BitMemoryManager methodsFor: 'object enumeration' stamp: 'eem 5/10/2017 18:17'!
objectBytesForSlots: numSlots
	"Answer the total number of bytes in an object with the given
	 number of slots, including header and possible overflow size header."
	<returnTypeC: #usqInt>
	^numSlots = 0
		ifTrue: [self allocationUnit + self baseHeaderSize] "always at least one slot for forwarding pointer"
		ifFalse:
			[numSlots + (numSlots bitAnd: 1) << self shiftForWord
			 + (numSlots >= self numSlotsMask
					ifTrue: [self baseHeaderSize + self baseHeaderSize]
					ifFalse: [self baseHeaderSize])]! !

!Spur32BitMemoryManager methodsFor: 'object access' stamp: 'eem 4/5/2014 12:04'!
oldRawNumSlotsOf: objOop
	<returnTypeC: #usqInt>
	<inline: true>
	self flag: #endianness.
	^(self longAt: objOop + 4) asUnsignedInteger >> self numSlotsHalfShift! !

!Spur32BitMemoryManager methodsFor: 'object access' stamp: 'eem 10/10/2013 09:24'!
rawOverflowSlotsOf: objOop
	<returnTypeC: #usqInt>
	<inline: true>
	self flag: #endianness.
	^self longAt: objOop - self baseHeaderSize! !

!Spur32BitMemoryManager methodsFor: 'object access' stamp: 'eem 6/22/2021 14:05'!
rawOverflowSlotsOf: objOop put: numSlots
	<returnTypeC: #usqInt>
	<inline: true>
	self flag: #endianness.
	^self longAt: objOop - self baseHeaderSize put: numSlots! !

!Spur32BitMemoryManager methodsFor: 'header access' stamp: 'eem 5/5/2015 17:37'!
rtRefCountOf: obj put: refCount
	"Set the rt reference count of obj; this is the three bit field comprised
	 of isGrey,isPinned,isRemembered.  See computeRefCountToShrinkRT."
	| header |
	self assert: (refCount between: 0 and: MaxRTRefCount).
	header := self longAt: obj.
	header := header bitAnd: (7 << self rememberedBitShift) bitInvert32.
	header := header + (refCount << self rememberedBitShift).
	self longAt: obj put: header! !

!Spur32BitMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 5/9/2017 11:35'!
savedOutHashFillValue
	<inline: true>
	^self maxIdentityHash + 1! !

!Spur32BitMemoryManager methodsFor: 'header access' stamp: 'eem 1/10/2017 15:57'!
set: objOop classIndexTo: classIndex formatTo: format
	"0 = 0 sized objects (UndefinedObject True False et al)
	 1 = non-indexable objects with inst vars (Point et al)
	 2 = indexable objects with no inst vars (Array et al)
	 3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
	 4 = weak indexable objects with inst vars (WeakArray et al)
	 5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
	 6 unused, reserved for exotic pointer objects?
	 7 Forwarded Object, 1st field is pointer, rest of fields are ignored
	 8 unused, reserved for exotic non-pointer objects?
	 9 64-bit indexable
	 10 - 11 32-bit indexable
	 12 - 15 16-bit indexable
	 16 - 23 byte indexable
	 24 - 31 compiled method"
	<inline: true>
	self assert: (classIndex between: 0 and: self classIndexMask).
	self assert: (format between: 0 and: self formatMask).
	self flag: #endianness.
	self longAt: objOop
		put: ((self longAt: objOop) bitAnd: (self formatMask << self formatShift + self classIndexMask) bitInvert32)
			+ (classIndex
			+  (format << self formatShift))! !

!Spur32BitMemoryManager methodsFor: 'header access' stamp: 'eem 9/22/2013 15:47'!
setClassIndexOf: objOop to: classIndex
	self assert: (classIndex between: 0 and: self classIndexMask).
	self flag: #endianness.
	self longAt: objOop
		put: ((self longAt: objOop) bitAnd: self classIndexMask bitInvert32)
			+ classIndex! !

!Spur32BitMemoryManager methodsFor: 'header access' stamp: 'eem 1/10/2017 15:57'!
setFormatOf: objOop to: format
	"0 = 0 sized objects (UndefinedObject True False et al)
	 1 = non-indexable objects with inst vars (Point et al)
	 2 = indexable objects with no inst vars (Array et al)
	 3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
	 4 = weak indexable objects with inst vars (WeakArray et al)
	 5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
	 6 unused, reserved for exotic pointer objects?
	 7 Forwarded Object, 1st field is pointer, rest of fields are ignored
	 8 unused, reserved for exotic non-pointer objects?
	 9 64-bit indexable
	 10 - 11 32-bit indexable
	 12 - 15 16-bit indexable
	 16 - 23 byte indexable
	 24 - 31 compiled method"
	self assert: (format between: 0 and: self formatMask).
	self flag: #endianness.
	self longAt: objOop
		put: ((self longAt: objOop) bitAnd: (self formatMask << self formatShift) bitInvert32)
			+ (format << self formatShift)! !

!Spur32BitMemoryManager methodsFor: 'header access' stamp: 'eem 10/15/2013 07:37'!
setIsGreyOf: objOop to: aBoolean
	self flag: #endianness.
	self longAt: objOop
		put: (aBoolean
				ifTrue: [(self longAt: objOop) bitOr: 1 << self greyBitShift]
				ifFalse: [(self longAt: objOop) bitAnd: (1 << self greyBitShift) bitInvert32])! !

!Spur32BitMemoryManager methodsFor: 'header access' stamp: 'eem 10/15/2013 00:29'!
setIsImmutableOf: objOop to: aBoolean
	self flag: #endianness.
	self longAt: objOop
		put: (aBoolean
				ifTrue: [(self longAt: objOop) bitOr: 1 << self immutableBitShift]
				ifFalse: [(self longAt: objOop) bitAnd: (1 << self immutableBitShift) bitInvert32])! !

!Spur32BitMemoryManager methodsFor: 'header access' stamp: 'eem 10/21/2013 15:07'!
setIsMarkedOf: objOop to: aBoolean
	self assert: (self isFreeObject: objOop) not.
	self flag: #endianness.
	self longAt: objOop + 4
		put: (aBoolean
				ifTrue: [(self longAt: objOop + 4) bitOr: 1 << self markedBitHalfShift]
				ifFalse: [(self longAt: objOop + 4) bitAnd: (1 << self markedBitHalfShift) bitInvert32])! !

!Spur32BitMemoryManager methodsFor: 'header access' stamp: 'eem 10/7/2013 14:15'!
setIsPinnedOf: objOop to: aBoolean
	self flag: #endianness.
	self longAt: objOop
		put: (aBoolean
				ifTrue: [(self longAt: objOop) bitOr: 1 << self pinnedBitShift]
				ifFalse: [(self longAt: objOop) bitAnd: (1 << self pinnedBitShift) bitInvert32])! !

!Spur32BitMemoryManager methodsFor: 'header access' stamp: 'eem 9/22/2013 15:44'!
setIsRememberedOf: objOop to: aBoolean
	self flag: #endianness.
	self longAt: objOop
		put: (aBoolean
				ifTrue: [(self longAt: objOop) bitOr: 1 << self rememberedBitShift]
				ifFalse: [(self longAt: objOop) bitAnd: (1 << self rememberedBitShift) bitInvert32])! !

!Spur32BitMemoryManager methodsFor: 'word size' stamp: 'eem 12/13/2015 10:16'!
shiftForWord
	<api>
	<cmacro>
	^2! !

!Spur32BitMemoryManager methodsFor: 'allocation' stamp: 'eem 11/13/2014 12:15'!
smallObjectBytesForSlots: numSlots
	"Answer the total number of bytes in an object without an overflow header, including header bytes."
	<api>
	<returnTypeC: #usqInt>
	^self baseHeaderSize "single header"
	+ (numSlots <= 1
		ifTrue: [self allocationUnit] "at least one slot for the forwarding pointer"
		ifFalse: [numSlots + (numSlots bitAnd: 1) * self bytesPerOop]) "round up to allocationUnit"! !

!Spur32BitMemoryManager methodsFor: 'word size' stamp: 'eem 12/12/2015 14:12'!
tagMask
	<api>
	<cmacro>
	^3! !

!Spur32BitMemoryManager methodsFor: 'indexing primitive support' stamp: 'eem 6/21/2021 14:50'!
updateFormatOfShortenedObject: objOop to: indexableSize
	"This is currently a noop on 32-bits. It will need to be implemented if
	 shortening of 16-bit or 8-bit pure bits objects is ever supported."! !

!Spur32BitMemoryManager methodsFor: 'header formats' stamp: 'eem 10/8/2013 12:21'!
wordIndexableFormat
	^self firstLongFormat! !

!Spur32BitMemoryManager methodsFor: 'word size' stamp: 'eem 12/12/2015 14:12'!
wordSize
	"Answer the manager's word size, which is the size of an oop, and which
	 is assumed to be equivalent to the underlying machine's word size."
	<api>
	<cmacro>
	^4! !

!Spur32BitMemoryManager methodsFor: 'class table' stamp: 'eem 10/8/2013 12:17'!
wordSizeClassIndexPun
	^self thirtyTwoBitLongsClassIndexPun! !

SpurMemoryManager subclass: #Spur64BitMemoryManager
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManager'!

Spur64BitMemoryManager subclass: #Spur64BitCoMemoryManager
	instanceVariableNames: 'cogit'
	classVariableNames: ''
	poolDictionaries: 'CogMethodConstants'
	category: 'VMMaker-SpurMemoryManager'!

!Spur64BitCoMemoryManager commentStamp: 'eem 11/25/2013 14:47' prior: 0!
Spur64BitCoMemoryManager is a refinement of Spur64BitMemoryManager that supports the CoInterpreter/Cogit just-in-time compiler.  The significant difference from Spur64BitMemoryManager is the memory layout.  Spur64BitCoMemoryManager adds the cogCodeZone beneath newSpace:

low address:
	cogCodeZone:
		generated run-time
		cog methods
		free space
		young referrers
	newSpace:
		past/future survivor space
		future/past survivor space
		eden
	first oldSpace segment
	...
	subsequent oldSpace segment
high address:

It would be convenient if the code zone were placed between newSpace and oldSpace; then Cog methods could be onsidered neither old nor young, filtering them out of copyAndForward: and the store check with single bounds checks.  But the CoInterpreter already assumes Cog methods are less than all objects (e.g. in its isMachineCodeFrame:).  If the dynamic frequency of isMachineCodeFrame: is higher (likely because this is used in e.g. scanning for unwind protects in non-local return) then it should keep the single bounds check.  So the coder zone remains beneath newSpace and Spur64BitCoMemoryManager ocerrides isReallyYoungObject: to filter-out Cog methods for copyAndForward:.

Instance Variables
	cogit:		<SimpleStackBasedCogit or subclass>

cogit
	- the just-in-time compiler!

!Spur64BitCoMemoryManager class methodsFor: 'simulation' stamp: 'eem 12/4/2015 22:21'!
defaultISA
	"Answer the default instruction set architecture to use for simulation."
	^#X64! !

!Spur64BitCoMemoryManager class methodsFor: 'accessing class hierarchy' stamp: 'eem 10/9/2013 19:46'!
objectRepresentationClass
	^CogObjectRepresentationFor64BitSpur! !

!Spur64BitCoMemoryManager class methodsFor: 'simulation' stamp: 'eem 11/29/2017 17:47'!
simulatorClass
	^Spur64BitMMLECoSimulatorFor64Bits basicNew memoryClass
		ifNil: [Spur64BitMMLECoSimulator]
		ifNotNil: [Spur64BitMMLECoSimulatorFor64Bits]! !

!Spur64BitCoMemoryManager methodsFor: 'trampolines' stamp: 'eem 3/12/2015 16:44'!
ceScheduleScavenge
	<api>
	self assert: freeStart >= scavengeThreshold.
	self scheduleScavenge! !

!Spur64BitCoMemoryManager methodsFor: 'debug support' stamp: 'eem 11/25/2013 14:34'!
checkMemoryMap
	"Override to check that Cog methods are considered neither young nor old.
	 Being young would cause them to be scavenged.
	 Being old would cause them to be remembered if stored into (but wait, they don't get stored into)."

	self assert: (self isYoungObject: newSpaceStart).
	self assert: (self isYoungObject: newSpaceLimit - self wordSize).
	self assert: (self isOldObject: newSpaceStart) not.
	self assert: (self isOldObject: newSpaceLimit - self wordSize) not.
	self assert: (self isYoungObject: newSpaceLimit) not.
	self assert: (self isYoungObject: oldSpaceStart) not.
	self assert: (self isYoungObject: endOfMemory) not.
	self assert: (self isOldObject: oldSpaceStart).
	self assert: (self isOldObject: endOfMemory).

	"we would like the following to be true, but we either choose one boundary check for
	 cogMethods vs objects (isMachineCodeFrame: et al) or one boundary check for
	 copyAndForward:.  We can't have both, and the former is likely the highest dynamic
	 frequency."
	false ifTrue:
		[self assert: (self isYoungObject: cogit minCogMethodAddress) not.
		 self assert: (self isYoungObject: cogit maxCogMethodAddress) not].
	self assert: (self isOldObject: cogit minCogMethodAddress) not.
	self assert: (self isOldObject: cogit maxCogMethodAddress) not! !

!Spur64BitCoMemoryManager methodsFor: 'accessing' stamp: 'eem 2/23/2017 16:18'!
classSmallFloat
	<api>
	^self fetchPointer: self smallFloatTag ofObject: classTableFirstPage! !

!Spur64BitCoMemoryManager methodsFor: 'debug support' stamp: 'eem 10/1/2013 16:58'!
clearLeakMapAndMapAccessibleObjects
	"Perform an integrity/leak check using the heapMap.  Set a bit at each object's header.
	 Override to set a bit at each Cog method"
	super clearLeakMapAndMapAccessibleObjects.
	cogit addCogMethodsToHeapMap! !

!Spur64BitCoMemoryManager methodsFor: 'cog jit support' stamp: 'eem 7/10/2014 17:09'!
ensureNoForwardedLiteralsIn: aMethodObj
	"Ensure there are no forwarded literals in the argument."
	<api>
	self followForwardedObjectFields: aMethodObj toDepth: 0! !

!Spur64BitCoMemoryManager methodsFor: 'forwarding' stamp: 'eem 6/24/2020 10:23'!
followForwardedObjectFields: objOop toDepth: depth
	"Follow pointers in the object to depth.
	 Answer if any forwarders were found.
	 How to avoid cyclic structures?? A temporary mark bit? eem 6/22/2020 no need since depth is always finite."
	<api>
	<inline: false>
	| found fmt numSlots |
	found := false.
	self assert: ((self isPointers: objOop) or: [self isOopCompiledMethod: objOop]).
	fmt := self formatOf: objOop.
	numSlots := self numPointerSlotsOf: objOop format: fmt.
	"It is essential to skip the first field of a method because it may be a
	 reference to a Cog method in the method zone, not a real object at all."
	((self isCompiledMethodFormat: fmt)
			ifTrue: [1]
			ifFalse: [0])
		to: numSlots - 1
		do: [:i| | oop |
			oop := self fetchPointer: i ofObject: objOop.
			(self isNonImmediate: oop) ifTrue:
				[(self isForwarded: oop) ifTrue:
					[found := true.
					 oop := self followForwarded: oop.
					 self storePointer: i ofObject: objOop withValue: oop].
				(depth > 0
				 and: [(self hasPointerFields: oop)
				 and: [self followForwardedObjectFields: oop toDepth: depth - 1]]) ifTrue:
					[found := true]]].
	^found! !

!Spur64BitCoMemoryManager methodsFor: 'cog jit support' stamp: 'eem 6/2/2014 18:47'!
freeStart: aValue
	self assert: (aValue >= scavenger eden start and: [aValue < (scavengeThreshold + 1024)]).
	self assert: (scavengeThreshold max: aValue) + coInterpreter interpreterAllocationReserveBytes <= scavenger eden limit.
	^freeStart := aValue! !

!Spur64BitCoMemoryManager methodsFor: 'trampoline support' stamp: 'eem 10/3/2013 10:53'!
freeStartAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: freeStart) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #freeStart in: self]! !

!Spur64BitCoMemoryManager methodsFor: 'cog jit support' stamp: 'eem 12/4/2013 12:55'!
getScavengeThreshold
	<api>
	<returnTypeC: #usqInt>
	^scavengeThreshold! !

!Spur64BitCoMemoryManager methodsFor: 'garbage collection' stamp: 'eem 10/1/2013 14:26'!
headerWhileForwardingOf: aCompiledMethodObjOop
	"Answer the header of the argument even though
	 it may have its header word in a forwarding block
	 (which shouldn't happen with Spur)."
	self assert: (self isForwarded: aCompiledMethodObjOop) not.
	^self baseHeader: aCompiledMethodObjOop! !

!Spur64BitCoMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 12/26/2014 17:01'!
ifAProxy: objOop updateCopy: copy
	"If the object being copied to the segment is weird and has exotic state,
	 i.e. a married context or a jitted method, update the copy with the vanilla state."

	super ifAProxy: objOop updateCopy: copy.
	(self isCompiledMethod: objOop) ifTrue:
		[| methodHeader |
		 methodHeader := coInterpreter rawHeaderOf: objOop.
		 (coInterpreter isCogMethodReference: methodHeader) ifTrue:
			[self storePointerUnchecked: HeaderIndex
				ofObject: copy
				withValue: (coInterpreter cCoerceSimple: methodHeader to: #'CogMethod *') methodHeader]]! !

!Spur64BitCoMemoryManager methodsFor: 'simulation only' stamp: 'eem 6/29/2015 14:42'!
initializeFreeSpaceForFacadeFrom: base to: limit
	"c.f. initializeFreeSpacePostLoad: freeListObj."
	| freeListObj freeBytes |
	newSpaceLimit := oldSpaceStart := freeStart := base.
	endOfMemory := limit.
	scavengeThreshold := limit * 3 // 4.
	segmentManager initSegmentForInImageCompilationFrom: base to: limit.
	freeListObj := self allocateSlots: self numFreeLists
						format: self wordIndexableFormat
						classIndex: self wordSizeClassIndexPun.
	freeLists := self firstIndexableField: freeListObj.
	freeListsMask := 0.
	0 to: self numFreeLists - 1 do:
		[:i|
		(freeLists at: i) ~= 0 ifTrue:
			[freeListsMask := freeListsMask bitOr: (1 << i).
			 freeLists at: i put: (segmentManager swizzleObj: (freeLists at: i))]].
	freeBytes := segmentManager lastSegment segLimit - self bridgeSize - freeStart.
	freeLists at: 0 put: (self initFreeChunkWithBytes: freeBytes at: freeStart).
	totalFreeOldSpace := freeBytes! !

!Spur64BitCoMemoryManager methodsFor: 'class table' stamp: 'eem 9/23/2014 18:56'!
isForwardedClassIndex: maybeClassIndex
	"A lenient tester of forwarded class indices for inline cache management in the Cogit."
	<api>
	| classTablePage entry |
	maybeClassIndex asUnsignedInteger >= self classTableRootSlots ifTrue:
		[^false].
	classTablePage := self fetchPointer: maybeClassIndex >> self classTableMajorIndexShift
							ofObject: hiddenRootsObj.
	classTablePage = nilObj ifTrue:
		[^false].
	entry := self
				fetchPointer: (maybeClassIndex bitAnd: self classTableMinorIndexMask)
				ofObject: classTablePage.
	^self isForwarded: entry! !

!Spur64BitCoMemoryManager methodsFor: 'cog jit support' stamp: 'eem 8/7/2014 04:34'!
isImmediateClass: classObj
	<api>
	^(self instSpecOfClass: classObj) = self instSpecForImmediateClasses! !

!Spur64BitCoMemoryManager methodsFor: 'debug support' stamp: 'eem 7/1/2021 11:25'!
isInHeapBounds: address 
	"Answer if the given address is within the entire range ST object memory.
	 For quick checking during leak checking only!!"
	^(self oop: address isGreaterThanOrEqualTo: cogit minCogMethodAddress)
	  and: [self oop: address isLessThan: endOfMemory]! !

!Spur64BitCoMemoryManager methodsFor: 'object testing' stamp: 'eem 11/25/2013 14:40'!
isReallyYoungObject: objOop
	<api>
	"Answer if obj is young. Require that obj is non-immediate. Override to filter-out Cog methods"
	self assert: (self isNonImmediate: objOop).
	^(self oop: objOop isLessThan: newSpaceLimit)
	  and: [self oop: objOop isGreaterThanOrEqualTo: newSpaceStart]! !

!Spur64BitCoMemoryManager methodsFor: 'compaction' stamp: 'eem 12/25/2016 10:40'!
methodHeaderFromSavedFirstField: field
	(self isIntegerObject: field) ifTrue:
		[^field].
	self assert: ((self isNonImmediate: field) and: [field < newSpaceStart]).
	self assert: (coInterpreter cCoerceSimple: field to: #'CogMethod *') objectHeader
				= self nullHeaderForMachineCodeMethod.
	^(coInterpreter cCoerceSimple: field to: #'CogMethod *') methodHeader! !

!Spur64BitCoMemoryManager methodsFor: 'growing/shrinking memory' stamp: 'eem 4/28/2015 14:48'!
methodHeaderOf: methodObj
	"Answer the method header of a CompiledMethod object.
	 If the method has been cogged then the header is a pointer to
	 the CogMethod and the real header will be stored in the CogMethod."
	<api>
	<inline: true>
	| header |
	self assert: (self isCompiledMethod: methodObj).
	header := self fetchPointer: HeaderIndex ofObject: methodObj.
	^(self isIntegerObject: header)
		ifTrue: [header]
		ifFalse:
			[self assert: header asUnsignedInteger < newSpaceStart.
			 self assert: (coInterpreter cCoerceSimple: header to: #'CogMethod *') objectHeader
						= self nullHeaderForMachineCodeMethod.
			(coInterpreter cCoerceSimple: header to: #'CogMethod *') methodHeader]! !

!Spur64BitCoMemoryManager methodsFor: 'trampoline support' stamp: 'eem 4/11/2017 16:12'!
needGCFlagAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: needGCFlag) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #needGCFlag in: self]! !

!Spur64BitCoMemoryManager methodsFor: 'garbage collection' stamp: 'eem 11/19/2013 12:31'!
nullHeaderForMachineCodeMethod
	<api>
	<returnTypeC: #sqLong>
	^(self firstLongFormat << self formatShift)
	+ (1 << self markedBitFullShift)
	+ ClassBitmapCompactIndex! !

!Spur64BitCoMemoryManager methodsFor: 'object enumeration' stamp: 'eem 5/31/2018 10:53'!
objectBytesForSlots: numSlots
	"Answer the total number of bytes in an object with the given
	 number of slots, including header and possible overflow size header."
	<api>
	<option: #SistaVM>
	^super objectBytesForSlots: numSlots! !

!Spur64BitCoMemoryManager methodsFor: 'cog jit support' stamp: 'eem 8/31/2021 14:23'!
receiverTagBitsForImmediateMethodClass: methodClassOrNil
	<inline: #always>
	methodClassOrNil = (self fetchPointer: self smallIntegerTag ofObject: classTableFirstPage) ifTrue:
		[^self smallIntegerTag].
	methodClassOrNil = (self fetchPointer: self characterTag ofObject: classTableFirstPage) ifTrue:
		[^self characterTag].
	self assert: methodClassOrNil = (self fetchPointer: self smallFloatTag ofObject: classTableFirstPage).
	^self smallFloatTag! !

!Spur64BitCoMemoryManager methodsFor: 'cog jit support' stamp: 'eem 7/29/2014 07:35'!
remember: objOop
	<doNotGenerate>
	^scavenger remember: objOop! !

!Spur64BitCoMemoryManager methodsFor: 'trampoline support' stamp: 'eem 5/31/2014 09:49'!
scavengeThresholdAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: scavengeThreshold) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #getScavengeThreshold in: self]! !

!Spur64BitCoMemoryManager methodsFor: 'debug support' stamp: 'eem 3/9/2016 17:48'!
setUpForUseByFacade: aCurrentImageCoInterpreterFacade
	"Make sure that eden etc are initialized, so that methods can be printed.
	 This is really to make addressCouldBeObj: et al work."
	<doNotGenerate>
	self edenBytes: 0.
	self setHeapBase: self freeStart
		memoryLimit: self endOfMemory
		endOfMemory: self endOfMemory! !

!Spur64BitCoMemoryManager methodsFor: 'trampoline support' stamp: 'eem 3/11/2020 11:31'!
specialObjectsOopAddress
	<api>
	^self cCode: [(self addressOf: specialObjectsOop) asUnsignedInteger]
		inSmalltalk: [cogit simulatedVariableAddress: #specialObjectsOop in: self]! !

!Spur64BitCoMemoryManager methodsFor: 'accessing' stamp: 'eem 12/6/2013 18:50'!
startOfMemory
	"Return the start of object memory.  This is immediately after the native code zone.
	 N.B. the stack zone is alloca'ed. Use a macro so as not to punish the debug VM."
	<api>
	<cmacro: '() heapBase'>
	<returnTypeC: #usqInt>
	^coInterpreter heapBase! !

!Spur64BitCoMemoryManager methodsFor: 'simulation only' stamp: 'eem 12/12/2015 15:02'!
unalignedLongAt: byteAddress
	<doNotGenerate>
	| rem |
	rem := byteAddress \\ 8.
	^rem = 0
		ifTrue: [self longAt: byteAddress]
		ifFalse: [((self longAt: byteAddress - rem) + ((self longAt: byteAddress - rem + 8) bitShift: 64) bitShift: rem * -8) bitAnd: 16rFFFFFFFFFFFFFFFF]! !

!Spur64BitCoMemoryManager methodsFor: 'simulation only' stamp: 'eem 12/12/2015 15:01'!
unalignedLongAt: byteAddress put: aLong
	<doNotGenerate>
	| rem mask |
	rem := byteAddress \\ 8.
	^rem = 0
		ifTrue: [self longAt: byteAddress put: aLong]
		ifFalse:
			[mask := 16rFFFFFFFFFFFFFFFF bitAnd: (-1 bitShift: rem * 8).
			 self longAt: byteAddress - rem
				put: ((self longAt: byteAddress - rem) bitAnd: mask bitInvert)
					+ ((aLong bitShift: rem * 8) bitAnd: mask).
			 self longAt: byteAddress - rem + 8
				put: ((self longAt: byteAddress - rem + 8) bitAnd: mask)
					+ ((aLong bitShift: 8 - rem * -8) bitAnd: mask bitInvert).
			 aLong]! !

!Spur64BitCoMemoryManager methodsFor: 'cog jit support' stamp: 'eem 11/18/2013 18:57'!
withoutForwardingOn: obj1 and: obj2 with: aBool sendToCogit: selector
	"For the purposes of become: send selector to the cogit with obj1, obj2
	 and aBool and answer the result.  Undo forwarding for the selector."
	<api>
	<var: #selector declareC: 'sqInt (*selector)(sqInt,sqInt,sqInt)'>
	| targetA targetB |
	targetA := self followForwarded: obj1.
	targetB := self followForwarded: obj2.
	^cogit perform: selector with: targetA with: targetB with: aBool! !

Spur64BitCoMemoryManager subclass: #Spur64BitMMLECoSimulator
	instanceVariableNames: 'parent bootstrapping'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManagerSimulation'!

!Spur64BitMMLECoSimulator methodsFor: 'accessing' stamp: 'eem 11/14/2013 16:42'!
bootstrapping
	^bootstrapping! !

!Spur64BitMMLECoSimulator methodsFor: 'accessing' stamp: 'eem 11/15/2013 11:00'!
bootstrapping: aBoolean
	bootstrapping := aBoolean.
	segmentManager initForBootstrap! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:33'!
byteAt: byteAddress
	| lowBits long32 |
	lowBits := byteAddress bitAnd: 3.
	long32 := self long32At: byteAddress - lowBits.
	^(long32 bitShift: -8 * lowBits) bitAnd: 16rFF! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:34'!
byteAt: byteAddress put: byte
	| lowBits long32 longAddress mask value |
	lowBits := byteAddress bitAnd: 3.
	longAddress := byteAddress - lowBits.
	long32 := self long32At: longAddress.
	mask := (16rFF bitShift: 8 * lowBits) bitInvert.
	value := byte bitShift: 8 * lowBits.
	self long32At: longAddress put: ((long32 bitAnd: mask) bitOr: value).
	^byte! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 9/4/2013 16:23'!
byteAtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address."

	^self byteAt: pointer! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'nice 4/6/2016 00:14'!
byteSwapped32IfBigEndian: w
	^w bitAnd: 16rFFFFFFFF! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'nice 4/6/2016 00:15'!
byteSwapped64IfBigEndian: w
	^w bitAnd: 16rFFFFFFFFFFFFFFFF! !

!Spur64BitMMLECoSimulator methodsFor: 'debug support' stamp: 'eem 9/5/2021 18:54'!
checkHeapIntegrity: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid
	"Almost all of the time spent in SpurMemoryManager>>checkHeapIntegrity:classIndicesShouldBeValid: used to go into the asserts in fetchPointer:ofObject: in the simulator class overrides.
	 Since we know here the indices used are valid we temporarily remove them to claw back that performance."
	^self withSimulatorFetchPointerMovedAsideDo:
		[super checkHeapIntegrity: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid]! !

!Spur64BitMMLECoSimulator methodsFor: 'initialization' stamp: 'eem 12/16/2016 16:30'!
coInterpreter: aCoInterpreter cogit: aCogit
	coInterpreter := aCoInterpreter.
	cogit := aCogit.
	scavenger coInterpreter: aCoInterpreter.
	compactor coInterpreter: aCoInterpreter! !

!Spur64BitMMLECoSimulator methodsFor: 'simulation only' stamp: 'eem 11/25/2013 13:17'!
cogCodeBase
	^Cogit guardPageSize! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 8/30/2013 11:47'!
endianness
	^#little! !

!Spur64BitMMLECoSimulator methodsFor: 'float primitives' stamp: 'eem 9/10/2013 17:29'!
fetchFloatAt: floatBitsAddress into: aFloat
	aFloat at: 2 put: (self long32At: floatBitsAddress).
	aFloat at: 1 put: (self long32At: floatBitsAddress+4)! !

!Spur64BitMMLECoSimulator methodsFor: 'object access' stamp: 'cb 6/11/2018 09:55'!
fetchPointer: fieldIndex ofObject: objOop
	self assert: (self isForwarded: objOop) not.
	self assert: (fieldIndex >= 0 and: [fieldIndex < (self numSlotsOfAny: objOop)
				or: [fieldIndex = 0 "forwarders"]]).
	^super fetchPointer: fieldIndex ofObject: objOop! !

!Spur64BitMMLECoSimulator methodsFor: 'object format' stamp: 'eem 12/28/2016 21:08'!
firstIndexableField: objOop
	"NOTE: overridden from SpurMemoryManager to add coercion to CArray, so please duplicate any changes.
	 There are only two important cases, both for objects with named inst vars, i.e. formats 2,3 & 5.
	 The first indexable field for formats 2 & 5 is the slot count (by convention, even though that's off the end
	 of the object).  For 3 we must go to the class."
	| fmt classFormat |
	<returnTypeC: #'void *'>
	fmt := self formatOf: objOop.
	fmt <= self lastPointerFormat ifTrue: "pointer; may need to delve into the class format word"
		[(fmt between: self indexablePointersFormat and: self weakArrayFormat) ifTrue:
			[classFormat := self formatOfClass: (self fetchClassOfNonImm: objOop).
			 ^self cCoerce: (self pointerForOop: objOop
												+ self baseHeaderSize
												+ ((self fixedFieldsOfClassFormat: classFormat) << self shiftForWord))
					to: #'oop *'].
		^self cCoerce: (self pointerForOop: objOop + self baseHeaderSize)
				to: #'oop *'].
	"All bit objects, and indeed CompiledMethod, though this is a no-no, start at 0"
	self assert: (fmt >= self sixtyFourBitIndexableFormat and: [fmt < self firstCompiledMethodFormat]).
	^self
		cCoerce: (self pointerForOop: objOop + self baseHeaderSize)
		to: (fmt < self firstByteFormat
				ifTrue:
					[fmt = self sixtyFourBitIndexableFormat
						ifTrue: ["64 bit field objects" #'long long *']
						ifFalse:
							[fmt < self firstShortFormat
								ifTrue: ["32 bit field objects" #'int *']
								ifFalse: ["16-bit field objects" #'short *']]]
				ifFalse: ["byte objects (including CompiledMethod" #'char *'])! !

!Spur64BitMMLECoSimulator methodsFor: 'accessing' stamp: 'eem 1/16/2016 20:04'!
freeStart
	"freeStart = 16r328C68 ifTrue: [self halt]."
	^super freeStart! !

!Spur64BitMMLECoSimulator methodsFor: 'gc - global' stamp: 'eem 5/30/2018 11:23'!
globalGarbageCollect
	"If we're /not/ a clone, clone the VM and push it over the cliff.
	 If it survives, destroy the clone and continue.  We should be OK until next time."
	parent ifNil:
		[coInterpreter cr; print: 'GC number '; print: statFullGCs; tab; flush.
		 CloneOnGC ifTrue:
			[coInterpreter cloneSimulation objectMemory globalGarbageCollect.
			 Smalltalk garbageCollect]].
	^super globalGarbageCollect! !

!Spur64BitMMLECoSimulator methodsFor: 'growing/shrinking memory' stamp: 'eem 11/14/2013 15:30'!
growOldSpaceByAtLeast: minAmmount
	"Attempt to grow memory by at least minAmmount.
	 Answer the size of the new segment, or nil if the attempt failed.
	 Override to not grow during the Spur image bootstrap."
	^bootstrapping ifFalse:
		[super growOldSpaceByAtLeast: minAmmount]! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 9/24/2013 13:50'!
halfWordHighInLong32: long32
	"Used by Balloon"

	^long32 bitAnd: 16rFFFF! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 9/24/2013 13:51'!
halfWordLowInLong32: long32
	"Used by Balloon"

	^long32 bitShift: -16! !

!Spur64BitMMLECoSimulator methodsFor: 'header format' stamp: 'eem 9/4/2013 11:08'!
headerForSlots: numSlots format: formatField classIndex: classIndex
	"The header format in LSB is
	 MSB:	| 2 bits				|
			| 22: identityHash	|
			| 8: slotSize			|
			| 3 bits				|
			| 5: format			|
			| 2 bits				|
			| 22: classIndex		| : LSB"
	self assert: (numSlots bitAnd: self numSlotsMask) = numSlots.
	self assert: (formatField bitAnd: self formatMask) = formatField.
	self assert: (classIndex bitAnd: self classIndexMask) = classIndex.
	^super headerForSlots: numSlots format: formatField classIndex: classIndex! !

!Spur64BitMMLECoSimulator methodsFor: 'debug support' stamp: 'eem 5/9/2014 20:28'!
heapMapAtWord: address
	^heapMap heapMapAtWord: address asInteger! !

!Spur64BitMMLECoSimulator methodsFor: 'initialization' stamp: 'eem 11/14/2013 15:28'!
initialize
	super initialize.
	bootstrapping := false! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 7/1/2016 10:46'!
intAt: byteAddress 
	^self long32At: byteAddress! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 9/23/2015 08:39'!
intAt: byteAddress put: a64BitValue
	^self long32At: byteAddress put: (a64BitValue bitAnd: 16rFFFFFFFF)! !

!Spur64BitMMLECoSimulator methodsFor: 'image segment in/out' stamp: 'eem 10/22/2014 10:49'!
loadImageSegmentFrom: segmentWordArray outPointers: outPointerArray
	self leakCheckImageSegments ifTrue:
		[self halt].
	^super loadImageSegmentFrom: segmentWordArray outPointers: outPointerArray! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 11/17/2014 14:05'!
long32At: byteAddress
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	byteAddress \\ 4 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 4 + 1! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 11/17/2014 14:05'!
long32At: byteAddress put: a32BitValue
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	"(byteAddress = 16r183FB00 and: [a32BitValue = 16r3FFFFC]) ifTrue:
		[self halt]."
	"(byteAddress between: 16r33FBB8 and: 16r33FBCF) ifTrue:
		[self halt]."
	byteAddress \\ 4 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 4 + 1 put: a32BitValue! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 4/19/2014 13:40'!
long64At: byteAddress
	"memory is a Bitmap, a 32-bit indexable array of bits"
	| hiWord loWord |
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	loWord := memory at: byteAddress // 4 + 1.
	hiWord := memory at: byteAddress // 4 + 2.
	^hiWord = 0
		ifTrue: [loWord]
		ifFalse: [(hiWord bitShift: 32) + loWord]! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 11/17/2014 16:21'!
long64At: byteAddress put: a64BitValue
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	self
		long32At: byteAddress put: (a64BitValue bitAnd: 16rffffffff);
		long32At: byteAddress + 4 put: a64BitValue >> 32.
	^a64BitValue! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 11/17/2014 14:05'!
longAt: byteAddress
	"Answer the 64-bit word at byteAddress which must be 0 mod 4."

	^self long64At: byteAddress! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 3/9/2016 17:46'!
longAt: byteAddress put: a64BitValue
	"Store the 64-bit value at byteAddress which must be 0 mod 8."
	"byteAddress = 16r1F5AE8 ifTrue: [self halt]."
	^self long64At: byteAddress put: a64BitValue! !

!Spur64BitMMLECoSimulator methodsFor: 'image segment in/out' stamp: 'eem 5/10/2017 15:44'!
mapOopsFrom: segStart to: segAddr outPointers: outPointerArray outHashes: savedOutHashes
	self leakCheckImageSegments ifTrue:
		[self halt].
	^super mapOopsFrom: segStart to: segAddr outPointers: outPointerArray outHashes: savedOutHashes! !

!Spur64BitMMLECoSimulator methodsFor: 'gc - global' stamp: 'eem 11/6/2013 16:16'!
markAndTrace: objOop
	"objOop = 16rB26020 ifTrue: [self halt].
	objOop = 16rB25FD8 ifTrue: [self halt].
	objOop = 16rB26010 ifTrue: [self halt]."
	^super markAndTrace: objOop! !

!Spur64BitMMLECoSimulator methodsFor: 'snapshot' stamp: 'eem 11/15/2013 09:47'!
memoryBaseForImageRead
	"Answer the address to read the image into.  Override so that when bootstrapping,
	 the segmentManager's segments are undisturbed in adjustSegmentSwizzlesBy:"
	^bootstrapping
		ifTrue: [0] 
		ifFalse: [super memoryBaseForImageRead]! !

!Spur64BitMMLECoSimulator methodsFor: 'accessing' stamp: 'eem 4/12/2017 11:43'!
needGCFlag: anInteger
	needGCFlag := anInteger ~= 0! !

!Spur64BitMMLECoSimulator methodsFor: 'initialization' stamp: 'eem 12/6/2014 11:20'!
nextLongFrom: aStream
	"Read a 32- or 64-bit quantity from the given (binary) stream."

	^aStream nextLittleEndianNumber: self wordSize! !

!Spur64BitMMLECoSimulator methodsFor: 'initialization' stamp: 'eem 12/6/2014 11:18'!
nextWord32From: aStream
	"Read a 32-bit quantity from the given (binary) stream."

	^aStream nextLittleEndianNumber: 4! !

!Spur64BitMMLECoSimulator methodsFor: 'accessing' stamp: 'eem 5/24/2014 23:38'!
parent

	^ parent! !

!Spur64BitMMLECoSimulator methodsFor: 'accessing' stamp: 'eem 5/24/2014 23:38'!
parent: anObject

	parent := anObject! !

!Spur64BitMMLECoSimulator methodsFor: 'image segment in/out' stamp: 'eem 6/28/2017 08:43'!
return: errCode restoringObjectsIn: firstArray savedFirstFields: savedFirstFields and: secondArray savedHashes: savedHashes
	self leakCheckImageSegments ifTrue:
		[self halt: errCode printString].
	^super return: errCode restoringObjectsIn: firstArray savedFirstFields: savedFirstFields and: secondArray savedHashes: savedHashes! !

!Spur64BitMMLECoSimulator methodsFor: 'debug support' stamp: 'eem 12/25/2016 11:29'!
runLeakCheckerFor: gcModes excludeUnmarkedObjs: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid
	(coInterpreter displayView isNil
	 and: [gcModes anyMask: checkForLeaks]) ifTrue:
		[coInterpreter transcript nextPutAll: 'leak-checking...'; flush].
	^super
		runLeakCheckerFor: gcModes
		excludeUnmarkedObjs: excludeUnmarkedObjs
		classIndicesShouldBeValid: classIndicesShouldBeValid! !

!Spur64BitMMLECoSimulator methodsFor: 'gc - global' stamp: 'eem 5/30/2018 11:24'!
scavengingGCTenuringIf: tenuringCriterion
	"If we're /not/ a clone, clone the VM and push it over the cliff.
	 If it survives, destroy the clone and continue.  We should be OK until next time."
	(self leakCheckNewSpaceGC
	 and: [parent isNil]) ifTrue:
		[coInterpreter cr; print: 'scavenge '; print: statScavenges; tab; flush.
		 CloneOnScavenge ifTrue:
			[coInterpreter cloneSimulation objectMemory scavengingGCTenuringIf: tenuringCriterion.
			 Smalltalk garbageCollect]].
	^super scavengingGCTenuringIf: tenuringCriterion! !

!Spur64BitMMLECoSimulator methodsFor: 'free space' stamp: 'eem 12/2/2013 13:59'!
setFree: o
	"o = 16rB34D40 ifTrue: [self halt]."
	super setFree: o! !

!Spur64BitMMLECoSimulator methodsFor: 'header access' stamp: 'eem 11/6/2013 14:15'!
setIsMarkedOf: objOop to: aBoolean
	"objOop = 16rB26020 ifTrue: [self halt]."
	super setIsMarkedOf: objOop to: aBoolean.
	"(aBoolean
	 and: [(self isContextNonImm: objOop)
	 and: [(coInterpreter
			checkIsStillMarriedContext: objOop
			currentFP: coInterpreter framePointer)
	 and: [(coInterpreter stackPages stackPageFor: (coInterpreter frameOfMarriedContext: objOop)) trace = 0]]]) ifTrue:
		[self halt]"! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 11/17/2014 14:05'!
shortAt: byteAddress
    "Return the half-word at byteAddress which must be even."
	| lowBits long |
	lowBits := byteAddress bitAnd: 2.
	long := self long32At: byteAddress - lowBits.
	^ lowBits = 2
		ifTrue: [ long bitShift: -16 ]
		ifFalse: [ long bitAnd: 16rFFFF ]! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 2/12/2015 14:17'!
shortAt: byteAddress put: a16BitValue
    "Return the half-word at byteAddress which must be even."
	| lowBits long longAddress |
	lowBits := byteAddress bitAnd: 2.
	lowBits = 0
		ifTrue: "storing into LS word"
			[long := self long32At: byteAddress.
			 self long32At: byteAddress
				put: ((long bitAnd: 16rFFFF0000) bitOr: a16BitValue)]
		ifFalse: "storing into MS word"
			[longAddress := byteAddress - 2.
			long := self long32At: longAddress.
			self long32At: longAddress
				put: ((long bitAnd: 16rFFFF) bitOr: (a16BitValue bitShift: 16))].
	^a16BitValue! !

!Spur64BitMMLECoSimulator methodsFor: 'indexing primitive support' stamp: 'eem 6/21/2021 13:00'!
shorten: objOop toIndexableSize: indexableSize
	self leakCheckShorten ifTrue:
		[parent ifNil:
			[| result |
			 result := coInterpreter cloneSimulation objectMemory shorten: objOop toIndexableSize: indexableSize.
			 Smalltalk garbageCollect]].
	^super shorten: objOop toIndexableSize: indexableSize! !

!Spur64BitMMLECoSimulator methodsFor: 'interpreter access' stamp: 'eem 10/27/2020 21:47'!
sizeFieldOfAlien: alienObj
	"Answer the first field of alienObj which is assumed to be an Alien of at least 16 bytes"
	<inline: #always>
	^(super sizeFieldOfAlien: alienObj) signedIntFromLong64! !

!Spur64BitMMLECoSimulator methodsFor: 'float primitives' stamp: 'eem 9/10/2013 17:30'!
storeFloatAt: floatBitsAddress from: aFloat
	self long32At: floatBitsAddress put: (aFloat at: 2).
	self long32At: floatBitsAddress+4 put: (aFloat at: 1)! !

!Spur64BitMMLECoSimulator methodsFor: 'image segment in/out' stamp: 'eem 7/5/2017 16:59'!
storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots
	self leakCheckImageSegments ifTrue:
		[parent ifNil:
			[| result |
			 self halt.
			 result := coInterpreter cloneSimulation objectMemory storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots.
			 Smalltalk garbageCollect]].
	^super storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots! !

!Spur64BitMMLECoSimulator methodsFor: 'ad-hoc tests' stamp: 'eem 1/9/2015 10:47'!
testObjStackDo
	| size them seqA seqB seqC rs |
	ExpensiveAsserts := true.
	self initializeWeaklingStack; emptyObjStack: weaklingStack.
	self assert: (self topOfObjStack: weaklingStack) isNil.
	self assert: (self capacityOfObjStack: weaklingStack) >= ObjStackLimit.
	seqA := (1 to: ObjStackLimit * 5 // 2) collect: [:i| self integerObjectOf: i].
	seqA do: [:it| self noCheckPush: it onObjStack: weaklingStack].
	them := Set new.
	size := self objStack: weaklingStack from: 0 do: [:it| them add: it].
	self assert: size = seqA size.
	self assert: (them asSortedCollection asArray = seqA).
	self assert: (self isValidObjStack: weaklingStack).
	seqB := (ObjStackLimit * 5 // 2 + 1 to: ObjStackLimit * 10 // 2) collect: [:i| self integerObjectOf: i].
	self assert: seqA size = seqB size.
	rs := seqB readStream.
	them := Set new.
	size := self objStack: weaklingStack from: 0 do:
				[:it|
				them add: it.
				self noCheckPush: rs next onObjStack: weaklingStack].
	self assert: size = seqA size.
	self assert: rs atEnd.
	self objStack: weaklingStack from: size do:
		[:it| them add: it].
	seqC := (seqA, seqB) sort.
	self assert: them asSortedCollection asArray = seqC! !

!Spur64BitMMLECoSimulator methodsFor: 'memory access' stamp: 'eem 10/9/2013 14:37'!
vmEndianness
	"1 = big, 0 = little"
	^0! !

Spur64BitMMLECoSimulator subclass: #Spur64BitMMLECoSimulatorFor64Bits
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManagerSimulation'!

!Spur64BitMMLECoSimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:30'!
byteAt: byteAddress
	| lowBits long64 |
	lowBits := byteAddress bitAnd: 7.
	long64 := self long64At: byteAddress - lowBits.
	^(long64 bitShift: -8 * lowBits) bitAnd: 16rFF! !

!Spur64BitMMLECoSimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:27'!
byteAt: byteAddress put: byte
	| lowBits long64 longAddress mask value |
	lowBits := byteAddress bitAnd: 7.
	longAddress := byteAddress - lowBits.
	long64 := self long64At: longAddress.
	mask := (16rFF bitShift: 8 * lowBits) bitInvert.
	value := byte bitShift: 8 * lowBits.
	self long64At: longAddress put: ((long64 bitAnd: mask) bitOr: value).
	^byte! !

!Spur64BitMMLECoSimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 11/19/2016 08:52'!
long32At: byteAddress
	"Answer the 32-bit word at byteAddress which must be a multiple of four."
	| lowBits long |
	lowBits := byteAddress bitAnd: 4.
	long := self long64At: byteAddress - lowBits.
	^lowBits = 4
		ifTrue: [long bitShift: -32]
		ifFalse: [long bitAnd: 16rFFFFFFFF]! !

!Spur64BitMMLECoSimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 11/19/2016 08:54'!
long32At: byteAddress put: a32BitValue
 	"Store the 32-bit word at byteAddress which must be a multiple of four."
	| lowBits long longAddress |
	lowBits := byteAddress bitAnd: 4.
	lowBits = 0
		ifTrue: "storing into LS word"
			[long := self long64At: byteAddress.
			 self long64At: byteAddress
				put: ((long bitAnd: 16rFFFFFFFF00000000) bitOr: a32BitValue)]
		ifFalse: "storing into MS word"
			[longAddress := byteAddress - 4.
			long := self long64At: longAddress.
			self long64At: longAddress
				put: ((long bitAnd: 16rFFFFFFFF) bitOr: (a32BitValue bitShift: 32))].
	^a32BitValue! !

!Spur64BitMMLECoSimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 12/21/2017 17:18'!
long64At: byteAddress
	"memory is a DoubleWordArray, a 64-bit indexable array of bits"
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 8 + 1! !

!Spur64BitMMLECoSimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 12/21/2017 17:18'!
long64At: byteAddress put: a64BitValue
	"memory is a DoubleWordArray, a 64-bit indexable array of bits"
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 8 + 1 put: a64BitValue! !

!Spur64BitMMLECoSimulatorFor64Bits methodsFor: 'simulation' stamp: 'eem 11/18/2016 17:09'!
memoryClass
	"Answer the class to use for the memory inst var in simulation.
	 Answer nil if a suitable class isn't available.  This version uses a 64-bit element class if available."
	<doNotGenerate>
	^Smalltalk classNamed: #DoubleWordArray! !

!Spur64BitMMLECoSimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:31'!
shortAt: byteAddress
	| lowBits long64 |
	lowBits := byteAddress bitAnd: 6.
	long64 := self long64At: byteAddress - lowBits.
	^(long64 bitShift: -8 * lowBits) bitAnd: 16rFFFF! !

!Spur64BitMMLECoSimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:28'!
shortAt: byteAddress put: short
	| lowBits long64 longAddress mask value |
	lowBits := byteAddress bitAnd: 6.
	longAddress := byteAddress - lowBits.
	long64 := self long64At: longAddress.
	mask := (16rFFFF bitShift: 8 * lowBits) bitInvert.
	value := short bitShift: 8 * lowBits.
	self long64At: longAddress put: ((long64 bitAnd: mask) bitOr: value).
	^short! !

Spur64BitMemoryManager subclass: #Spur64BitMMLESimulator
	instanceVariableNames: 'parent bootstrapping'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManagerSimulation'!

!Spur64BitMMLESimulator methodsFor: 'accessing' stamp: 'eem 11/14/2013 16:42'!
bootstrapping
	^bootstrapping! !

!Spur64BitMMLESimulator methodsFor: 'accessing' stamp: 'eem 11/15/2013 11:00'!
bootstrapping: aBoolean
	bootstrapping := aBoolean.
	segmentManager initForBootstrap! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:33'!
byteAt: byteAddress
	| lowBits long32 |
	lowBits := byteAddress bitAnd: 3.
	long32 := self long32At: byteAddress - lowBits.
	^(long32 bitShift: -8 * lowBits) bitAnd: 16rFF! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:34'!
byteAt: byteAddress put: byte
	| lowBits long32 longAddress mask value |
	lowBits := byteAddress bitAnd: 3.
	longAddress := byteAddress - lowBits.
	long32 := self long32At: longAddress.
	mask := (16rFF bitShift: 8 * lowBits) bitInvert.
	value := byte bitShift: 8 * lowBits.
	self long32At: longAddress put: ((long32 bitAnd: mask) bitOr: value).
	^byte! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 9/4/2013 16:23'!
byteAtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address."

	^self byteAt: pointer! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'nice 4/6/2016 00:14'!
byteSwapped32IfBigEndian: w
	^w bitAnd: 16rFFFFFFFF! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'nice 4/6/2016 00:15'!
byteSwapped64IfBigEndian: w
	^w bitAnd: 16rFFFFFFFFFFFFFFFF! !

!Spur64BitMMLESimulator methodsFor: 'debug support' stamp: 'eem 9/5/2021 18:54'!
checkHeapIntegrity: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid
	"Almost all of the time spent in SpurMemoryManager>>checkHeapIntegrity:classIndicesShouldBeValid: used to go into the asserts in fetchPointer:ofObject: in the simulator class overrides.
	 Since we know here the indices used are valid we temporarily remove them to claw back that performance."
	^self withSimulatorFetchPointerMovedAsideDo:
		[super checkHeapIntegrity: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid]! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 8/30/2013 11:47'!
endianness
	^#little! !

!Spur64BitMMLESimulator methodsFor: 'float primitives' stamp: 'eem 11/19/2014 13:29'!
fetchFloatAt: floatBitsAddress into: aFloat
	aFloat at: 2 put: (self long32At: floatBitsAddress).
	aFloat at: 1 put: (self long32At: floatBitsAddress+4)! !

!Spur64BitMMLESimulator methodsFor: 'object access' stamp: 'cb 6/11/2018 09:55'!
fetchPointer: fieldIndex ofObject: objOop
	self assert: (self isForwarded: objOop) not.
	self assert: (fieldIndex >= 0 and: [fieldIndex < (self numSlotsOfAny: objOop)
				or: [fieldIndex = 0 "forwarders"]]).
	^super fetchPointer: fieldIndex ofObject: objOop! !

!Spur64BitMMLESimulator methodsFor: 'object format' stamp: 'eem 12/28/2016 21:08'!
firstIndexableField: objOop
	"NOTE: overridden from SpurMemoryManager to add coercion to CArray, so please duplicate any changes.
	 There are only two important cases, both for objects with named inst vars, i.e. formats 2,3 & 5.
	 The first indexable field for formats 2 & 5 is the slot count (by convention, even though that's off the end
	 of the object).  For 3 we must go to the class."
	| fmt classFormat |
	<returnTypeC: #'void *'>
	fmt := self formatOf: objOop.
	fmt <= self lastPointerFormat ifTrue: "pointer; may need to delve into the class format word"
		[(fmt between: self indexablePointersFormat and: self weakArrayFormat) ifTrue:
			[classFormat := self formatOfClass: (self fetchClassOfNonImm: objOop).
			 ^self cCoerce: (self pointerForOop: objOop
												+ self baseHeaderSize
												+ ((self fixedFieldsOfClassFormat: classFormat) << self shiftForWord))
					to: #'oop *'].
		^self cCoerce: (self pointerForOop: objOop + self baseHeaderSize)
				to: #'oop *'].
	"All bit objects, and indeed CompiledMethod, though this is a no-no, start at 0"
	self assert: (fmt >= self sixtyFourBitIndexableFormat and: [fmt < self firstCompiledMethodFormat]).
	^self
		cCoerce: (self pointerForOop: objOop + self baseHeaderSize)
		to: (fmt < self firstByteFormat
				ifTrue:
					[fmt = self sixtyFourBitIndexableFormat
						ifTrue: ["64 bit field objects" #'long long *']
						ifFalse:
							[fmt < self firstShortFormat
								ifTrue: ["32 bit field objects" #'int *']
								ifFalse: ["16-bit field objects" #'short *']]]
				ifFalse: ["byte objects (including CompiledMethod" #'char *'])! !

!Spur64BitMMLESimulator methodsFor: 'spur bootstrap' stamp: 'eem 10/8/2013 11:36'!
freeLists
	^freeLists! !

!Spur64BitMMLESimulator methodsFor: 'gc - global' stamp: 'eem 5/30/2018 11:23'!
globalGarbageCollect
	"If we're /not/ a clone, clone the VM and push it over the cliff.
	 If it survives, destroy the clone and continue.  We should be OK until next time."
	parent ifNil:
		[coInterpreter cr; print: 'GC number '; print: statFullGCs; tab; flush.
		 CloneOnGC ifTrue:
			[coInterpreter cloneSimulation objectMemory globalGarbageCollect.
			 Smalltalk garbageCollect]].
	^super globalGarbageCollect! !

!Spur64BitMMLESimulator methodsFor: 'growing/shrinking memory' stamp: 'eem 11/14/2013 15:30'!
growOldSpaceByAtLeast: minAmmount
	"Attempt to grow memory by at least minAmmount.
	 Answer the size of the new segment, or nil if the attempt failed.
	 Override to not grow during the Spur image bootstrap."
	^bootstrapping ifFalse:
		[super growOldSpaceByAtLeast: minAmmount]! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 9/24/2013 13:50'!
halfWordHighInLong32: long32
	"Used by Balloon"

	^long32 bitAnd: 16rFFFF! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 9/24/2013 13:51'!
halfWordLowInLong32: long32
	"Used by Balloon"

	^long32 bitShift: -16! !

!Spur64BitMMLESimulator methodsFor: 'header format' stamp: 'eem 9/4/2013 11:08'!
headerForSlots: numSlots format: formatField classIndex: classIndex
	"The header format in LSB is
	 MSB:	| 2 bits				|
			| 22: identityHash	|
			| 8: slotSize			|
			| 3 bits				|
			| 5: format			|
			| 2 bits				|
			| 22: classIndex		| : LSB"
	self assert: (numSlots bitAnd: self numSlotsMask) = numSlots.
	self assert: (formatField bitAnd: self formatMask) = formatField.
	self assert: (classIndex bitAnd: self classIndexMask) = classIndex.
	^super headerForSlots: numSlots format: formatField classIndex: classIndex! !

!Spur64BitMMLESimulator methodsFor: 'debug support' stamp: 'eem 5/9/2014 20:21'!
heapMapAtWord: address
	^heapMap heapMapAtWord: address! !

!Spur64BitMMLESimulator methodsFor: 'compaction' stamp: 'eem 5/9/2014 15:22'!
inSortedFreeListLink: freeChunk to: nextFree given: prevFree
	"thisContext sender selector = #sweepToCoallesceFreeSpaceForPigCompactFrom: ifTrue:
		[| pit |
			pit := [:label :thing|
					coInterpreter print: label; space; printHex: thing.
					(thing ~= 0 and: [self isFreeObject: thing]) ifTrue:
						[coInterpreter print: ' (free) ']].
			pit value: 'link ' value: freeChunk.
			pit value: ' to ' value: nextFree.
			pit value: ' from ' value: prevFree.
			coInterpreter cr]."
	"freeChunk = 16r10B0730 ifTrue:
		[self halt]."
	super inSortedFreeListLink: freeChunk to: nextFree given: prevFree! !

!Spur64BitMMLESimulator methodsFor: 'initialization' stamp: 'eem 11/14/2013 15:28'!
initialize
	super initialize.
	bootstrapping := false! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 7/1/2016 10:47'!
intAt: byteAddress 
	^self long32At: byteAddress! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 9/23/2015 08:39'!
intAt: byteAddress put: a64BitValue
	^self long32At: byteAddress put: (a64BitValue bitAnd: 16rFFFFFFFF)! !

!Spur64BitMMLESimulator methodsFor: 'image segment in/out' stamp: 'eem 10/22/2014 10:49'!
loadImageSegmentFrom: segmentWordArray outPointers: outPointerArray
	self leakCheckImageSegments ifTrue:
		[self halt].
	^super loadImageSegmentFrom: segmentWordArray outPointers: outPointerArray! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 11/17/2014 14:05'!
long32At: byteAddress
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	byteAddress \\ 4 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 4 + 1! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 11/17/2014 14:05'!
long32At: byteAddress put: a32BitValue
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	"(byteAddress = 16r183FB00 and: [a32BitValue = 16r3FFFFC]) ifTrue:
		[self halt]."
	"(byteAddress between: 16r33FBB8 and: 16r33FBCF) ifTrue:
		[self halt]."
	byteAddress \\ 4 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 4 + 1 put: a32BitValue! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 4/19/2014 13:40'!
long64At: byteAddress
	"memory is a Bitmap, a 32-bit indexable array of bits"
	| hiWord loWord |
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	loWord := memory at: byteAddress // 4 + 1.
	hiWord := memory at: byteAddress // 4 + 2.
	^hiWord = 0
		ifTrue: [loWord]
		ifFalse: [(hiWord bitShift: 32) + loWord]! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'nice 3/16/2016 01:33'!
long64At: byteAddress put: a64BitValue
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	self
		long32At: byteAddress put: (a64BitValue bitAnd: 16rffffffff);
		long32At: byteAddress + 4 put: (a64BitValue >> 32 bitAnd: 16rffffffff).
	^a64BitValue! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 11/17/2014 14:05'!
longAt: byteAddress
	"Answer the 64-bit word at byteAddress which must be 0 mod 4."

	^self long64At: byteAddress! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 3/9/2016 17:46'!
longAt: byteAddress put: a64BitValue
	"Store the 64-bit value at byteAddress which must be 0 mod 8."
	"byteAddress = 16r43C790 ifTrue: [self halt]."
	^self long64At: byteAddress put: a64BitValue! !

!Spur64BitMMLESimulator methodsFor: 'image segment in/out' stamp: 'eem 5/10/2017 15:44'!
mapOopsFrom: segStart to: segAddr outPointers: outPointerArray outHashes: savedOutHashes
	self leakCheckImageSegments ifTrue:
		[self halt].
	^super mapOopsFrom: segStart to: segAddr outPointers: outPointerArray outHashes: savedOutHashes! !

!Spur64BitMMLESimulator methodsFor: 'gc - global' stamp: 'eem 11/6/2013 16:16'!
markAndTrace: objOop
	"objOop = 16rB26020 ifTrue: [self halt].
	objOop = 16rB25FD8 ifTrue: [self halt].
	objOop = 16rB26010 ifTrue: [self halt]."
	^super markAndTrace: objOop! !

!Spur64BitMMLESimulator methodsFor: 'snapshot' stamp: 'eem 11/15/2013 09:47'!
memoryBaseForImageRead
	"Answer the address to read the image into.  Override so that when bootstrapping,
	 the segmentManager's segments are undisturbed in adjustSegmentSwizzlesBy:"
	^bootstrapping
		ifTrue: [0] 
		ifFalse: [super memoryBaseForImageRead]! !

!Spur64BitMMLESimulator methodsFor: 'compaction' stamp: 'eem 5/28/2014 07:42'!
moveARunOfObjectsStartingAt: startAddress upTo: limit
	| result |.
	"self checkTraversableSortedFreeList."
	result := super moveARunOfObjectsStartingAt: startAddress upTo: limit.
	"self checkTraversableSortedFreeList."
	^result! !

!Spur64BitMMLESimulator methodsFor: 'initialization' stamp: 'eem 11/18/2014 13:53'!
nextLongFrom: aStream
	"Read a 32- or 64-bit quantity from the given (binary) stream."

	^aStream nextLittleEndianNumber: self wordSize! !

!Spur64BitMMLESimulator methodsFor: 'initialization' stamp: 'eem 11/18/2014 13:52'!
nextWord32From: aStream
	"Read a 32-bit quantity from the given (binary) stream."

	^aStream nextLittleEndianNumber: 4! !

!Spur64BitMMLESimulator methodsFor: 'spur bootstrap' stamp: 'eem 10/16/2013 12:20'!
numClassTablePages
	^numClassTablePages! !

!Spur64BitMMLESimulator methodsFor: 'accessing' stamp: 'eem 5/24/2014 23:38'!
parent

	^ parent! !

!Spur64BitMMLESimulator methodsFor: 'accessing' stamp: 'eem 5/24/2014 23:38'!
parent: anObject

	parent := anObject! !

!Spur64BitMMLESimulator methodsFor: 'image segment in/out' stamp: 'eem 6/28/2017 08:43'!
return: errCode restoringObjectsIn: firstArray savedFirstFields: savedFirstFields and: secondArray savedHashes: savedHashes
	self leakCheckImageSegments ifTrue:
		[self halt: errCode printString].
	^super return: errCode restoringObjectsIn: firstArray savedFirstFields: savedFirstFields and: secondArray savedHashes: savedHashes! !

!Spur64BitMMLESimulator methodsFor: 'debug support' stamp: 'eem 12/25/2016 11:29'!
runLeakCheckerFor: gcModes excludeUnmarkedObjs: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid
	(coInterpreter displayView isNil
	 and: [gcModes anyMask: checkForLeaks]) ifTrue:
		[coInterpreter transcript nextPutAll: 'leak-checking...'; flush].
	^super
		runLeakCheckerFor: gcModes
		excludeUnmarkedObjs: excludeUnmarkedObjs
		classIndicesShouldBeValid: classIndicesShouldBeValid! !

!Spur64BitMMLESimulator methodsFor: 'gc - global' stamp: 'eem 5/30/2018 11:24'!
scavengingGCTenuringIf: tenuringCriterion
	"If we're /not/ a clone, clone the VM and push it over the cliff.
	 If it survives, destroy the clone and continue.  We should be OK until next time."
	(self leakCheckNewSpaceGC
	 and: [parent isNil]) ifTrue:
		[coInterpreter cr; print: 'scavenge '; print: statScavenges; tab; flush.
		 CloneOnScavenge ifTrue:
			[coInterpreter cloneSimulation objectMemory scavengingGCTenuringIf: tenuringCriterion.
			 Smalltalk garbageCollect]].
	^super scavengingGCTenuringIf: tenuringCriterion! !

!Spur64BitMMLESimulator methodsFor: 'header access' stamp: 'eem 10/29/2014 16:32'!
setIsMarkedOf: objOop to: aBoolean
	"objOop = 16rB26020 ifTrue: [self halt]."
	"(#(16r1971D0 16r196EE0 16r197048 16r197148) includes: objOop) ifTrue:
		[self halt]."
	super setIsMarkedOf: objOop to: aBoolean.
	"(aBoolean
	 and: [(self isContextNonImm: objOop)
	 and: [(coInterpreter
			checkIsStillMarriedContext: objOop
			currentFP: coInterpreter framePointer)
	 and: [(coInterpreter stackPages stackPageFor: (coInterpreter frameOfMarriedContext: objOop)) trace = 0]]]) ifTrue:
		[self halt]"! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 11/17/2014 14:05'!
shortAt: byteAddress
    "Return the half-word at byteAddress which must be even."
	| lowBits long |
	lowBits := byteAddress bitAnd: 2.
	long := self long32At: byteAddress - lowBits.
	^ lowBits = 2
		ifTrue: [ long bitShift: -16 ]
		ifFalse: [ long bitAnd: 16rFFFF ]! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 2/12/2015 14:17'!
shortAt: byteAddress put: a16BitValue
    "Return the half-word at byteAddress which must be even."
	| lowBits long longAddress |
	lowBits := byteAddress bitAnd: 2.
	lowBits = 0
		ifTrue: "storing into LS word"
			[long := self long32At: byteAddress.
			 self long32At: byteAddress
				put: ((long bitAnd: 16rFFFF0000) bitOr: a16BitValue)]
		ifFalse: "storing into MS word"
			[longAddress := byteAddress - 2.
			long := self long32At: longAddress.
			self long32At: longAddress
				put: ((long bitAnd: 16rFFFF) bitOr: (a16BitValue bitShift: 16))].
	^a16BitValue! !

!Spur64BitMMLESimulator methodsFor: 'indexing primitive support' stamp: 'eem 6/21/2021 13:00'!
shorten: objOop toIndexableSize: indexableSize
	self leakCheckShorten ifTrue:
		[parent ifNil:
			[| result |
			 result := coInterpreter cloneSimulation objectMemory shorten: objOop toIndexableSize: indexableSize.
			 Smalltalk garbageCollect]].
	^super shorten: objOop toIndexableSize: indexableSize! !

!Spur64BitMMLESimulator methodsFor: 'interpreter access' stamp: 'eem 10/27/2020 21:48'!
sizeFieldOfAlien: alienObj
	"Answer the first field of alienObj which is assumed to be an Alien of at least 16 bytes"
	<inline: #always>
	^(super sizeFieldOfAlien: alienObj) signedIntFromLong64! !

!Spur64BitMMLESimulator methodsFor: 'float primitives' stamp: 'eem 11/19/2014 22:03'!
storeFloatAt: floatBitsAddress from: aFloat
	self long32At: floatBitsAddress put: (aFloat at: 2).
	self long32At: floatBitsAddress+4 put: (aFloat at: 1)! !

!Spur64BitMMLESimulator methodsFor: 'image segment in/out' stamp: 'eem 7/5/2017 16:59'!
storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots
	self leakCheckImageSegments ifTrue:
		[parent ifNil:
			[| result |
			 self halt.
			 result := coInterpreter cloneSimulation objectMemory storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots.
			 Smalltalk garbageCollect]].
	^super storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots! !

!Spur64BitMMLESimulator methodsFor: 'memory access' stamp: 'eem 10/9/2013 14:37'!
vmEndianness
	"1 = big, 0 = little"
	^0! !

Spur64BitMMLESimulator subclass: #Spur64BitMMLESimulatorFor64Bits
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManagerSimulation'!

!Spur64BitMMLESimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:31'!
byteAt: byteAddress
	| lowBits long64 |
	lowBits := byteAddress bitAnd: 7.
	long64 := self long64At: byteAddress - lowBits.
	^(long64 bitShift: -8 * lowBits) bitAnd: 16rFF! !

!Spur64BitMMLESimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:28'!
byteAt: byteAddress put: byte
	| lowBits long64 longAddress mask value |
	lowBits := byteAddress bitAnd: 7.
	longAddress := byteAddress - lowBits.
	long64 := self long64At: longAddress.
	mask := (16rFF bitShift: 8 * lowBits) bitInvert.
	value := byte bitShift: 8 * lowBits.
	self long64At: longAddress put: ((long64 bitAnd: mask) bitOr: value).
	^byte! !

!Spur64BitMMLESimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 11/19/2016 08:52'!
long32At: byteAddress
	"Answer the 32-bit word at byteAddress which must be a multiple of four."
	| lowBits long |
	lowBits := byteAddress bitAnd: 4.
	long := self long64At: byteAddress - lowBits.
	^lowBits = 4
		ifTrue: [long bitShift: -32]
		ifFalse: [long bitAnd: 16rFFFFFFFF]! !

!Spur64BitMMLESimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 12/21/2017 15:21'!
long32At: byteAddress put: a32BitValue
 	"Store the 32-bit word at byteAddress which must be a multiple of four."
	| lowBits long longAddress |
	a32BitValue < 0 ifTrue:
		[self long32At: byteAddress put: (a32BitValue bitAnd: 16rFFFFFFFF).
		 ^a32BitValue].
	lowBits := byteAddress bitAnd: 4.
	lowBits = 0
		ifTrue: "storing into LS word"
			[long := self long64At: byteAddress.
			 self long64At: byteAddress
				put: ((long bitAnd: 16rFFFFFFFF00000000) bitOr: a32BitValue)]
		ifFalse: "storing into MS word"
			[longAddress := byteAddress - 4.
			long := self long64At: longAddress.
			self long64At: longAddress
				put: ((long bitAnd: 16rFFFFFFFF) bitOr: (a32BitValue bitShift: 32))].
	^a32BitValue! !

!Spur64BitMMLESimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 12/19/2017 16:59'!
long64At: byteAddress
	"memory is a DoubleWordArray, a 64-bit indexable array of bits"
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 8 + 1! !

!Spur64BitMMLESimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 12/21/2017 17:18'!
long64At: byteAddress put: a64BitValue
	"memory is a DoubleWordArray, a 64-bit indexable array of bits"
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 8 + 1 put: a64BitValue! !

!Spur64BitMMLESimulatorFor64Bits methodsFor: 'simulation' stamp: 'eem 11/18/2016 16:24'!
memoryClass
	"Answer the class to use for the memory inst var in simulation.
	 Answer nil if a suitable class isn't available.  This version uses a 64-bit element class if available."
	<doNotGenerate>
	^Smalltalk classNamed: #DoubleWordArray! !

!Spur64BitMMLESimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:30'!
shortAt: byteAddress
	| lowBits long64 |
	lowBits := byteAddress bitAnd: 6.
	long64 := self long64At: byteAddress - lowBits.
	^(long64 bitShift: -8 * lowBits) bitAnd: 16rFFFF! !

!Spur64BitMMLESimulatorFor64Bits methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:27'!
shortAt: byteAddress put: short
	| lowBits long64 longAddress mask value |
	lowBits := byteAddress bitAnd: 6.
	longAddress := byteAddress - lowBits.
	long64 := self long64At: longAddress.
	mask := (16rFFFF bitShift: 8 * lowBits) bitInvert.
	value := short bitShift: 8 * lowBits.
	self long64At: longAddress put: ((long64 bitAnd: mask) bitOr: value).
	^short! !

!Spur64BitMemoryManager class methodsFor: 'class initialization' stamp: 'eem 12/26/2015 03:05'!
initializeWithOptions: optionsDictionary
	"SpurMemoryManager initializeWithOptions: Dictionary new"

	optionsDictionary at: #Spur64BitMemoryManager put: true.
	super initializeWithOptions: optionsDictionary! !

!Spur64BitMemoryManager class methodsFor: 'simulation only' stamp: 'eem 11/20/2014 17:56'!
isImmediate: anObject
	self flag: 'The float range is close but probably wrong. Revisit when immediate floats are implemented'.
	^(anObject isInteger and: [anObject between: self minSmallInteger and: self maxSmallInteger])
	  or: [anObject isCharacter
	  or: [anObject isFloat and: [anObject exponent between: -126 and: 127]]]

	"1.0e-38 exponent -127
	 1.0e38 exponent 126

	 NumberParser new makeFloatFromMantissa: 1 exponent: 127 base: 2 1.7014118346046923e38
	 (NumberParser new makeFloatFromMantissa: 1 exponent: 127 base: 2) exponent 127
	 NumberParser new makeFloatFromMantissa: 1 exponent: -126 base: 2 1.1754943508222875e-38
	 (NumberParser new makeFloatFromMantissa: 1 exponent: -126 base: 2) exponent -126"! !

!Spur64BitMemoryManager class methodsFor: 'translation' stamp: 'eem 11/20/2014 13:01'!
maxSmallInteger
	^16rFFFFFFFFFFFFFFF! !

!Spur64BitMemoryManager class methodsFor: 'translation' stamp: 'eem 11/20/2014 13:03'!
minSmallInteger
	^-16r1000000000000000! !

!Spur64BitMemoryManager class methodsFor: 'simulation only' stamp: 'eem 11/18/2014 17:49'!
numSmallIntegerTagBits
	^3! !

!Spur64BitMemoryManager class methodsFor: 'simulation only' stamp: 'eem 11/18/2016 17:08'!
simulatorClass
	^Spur64BitMMLESimulatorFor64Bits basicNew memoryClass
		ifNil: [Spur64BitMMLESimulator]
		ifNotNil: [Spur64BitMMLESimulatorFor64Bits]! !

!Spur64BitMemoryManager class methodsFor: 'word size' stamp: 'eem 11/13/2014 15:24'!
wordSize
	"Answer the manager's word size, which is the size of an oop, and which
	 is assumed to be equivalent to the underlying machine's word size."
	^8! !

!Spur64BitMemoryManager methodsFor: 'object enumeration' stamp: 'eem 1/15/2018 09:28'!
addressAfter: objOop
	"Answer the address immediately following an object."
	<returnTypeC: #usqInt>
	| numSlots |
	numSlots := self rawNumSlotsOf: objOop.
	numSlots = 0 ifTrue: [^objOop + self allocationUnit + self baseHeaderSize].
	numSlots = self numSlotsMask ifTrue:
		[numSlots := self rawOverflowSlotsOf: objOop].
	^objOop + self baseHeaderSize + (numSlots bitShift: self shiftForWord)! !

!Spur64BitMemoryManager methodsFor: 'allocation' stamp: 'eem 09/28/2015 18:48'!
allocateSlots: numSlots format: formatField classIndex: classIndex
	"Allocate an object with numSlots space.  If there is room beneath scavengeThreshold
	 allocate in newSpace, otherwise alocate in oldSpace.  If there is not room in newSpace
	 and a scavenge is not already scheduled, schedule a scavenge."
	<inline: true>
	| numBytes newObj |
	"Object headers are 8 bytes in length if the slot size fits in the num slots field (max implies overflow),
	 16 bytes otherwise (num slots in preceding word).
	 Objects always have at least one slot, for the forwarding pointer,
	 and are multiples of 8 bytes in length."
	numSlots >= self numSlotsMask
		ifTrue:
			[numSlots >> 56 > 0 ifTrue:
				[^nil]. "overflow size must fit in 56-bits"
			 newObj := freeStart + self baseHeaderSize.
			 numBytes := self largeObjectBytesForSlots: numSlots]
		ifFalse:
			[newObj := freeStart.
			 numBytes := self smallObjectBytesForSlots: numSlots].
	
	freeStart + numBytes > scavengeThreshold ifTrue:
		[needGCFlag ifFalse: [self scheduleScavenge].
		 ^self allocateSlotsInOldSpace: numSlots bytes: numBytes format: formatField classIndex: classIndex].
	numSlots >= self numSlotsMask
		ifTrue: "for header parsing we put a saturated slot count in the prepended overflow size word"
			[self flag: #endianness.
			 self longAt: freeStart put: self numSlotsMask << self numSlotsFullShift + numSlots.
			 self longAt: newObj put: (self headerForSlots: self numSlotsMask format: formatField classIndex: classIndex)]
		ifFalse:
			[self longAt: newObj put: (self headerForSlots: numSlots format: formatField classIndex: classIndex)].
	self assert: numBytes \\ self allocationUnit = 0.
	self assert: newObj \\ self allocationUnit = 0.
	freeStart := freeStart + numBytes.
	^newObj! !

!Spur64BitMemoryManager methodsFor: 'allocation' stamp: 'eem 12/27/2016 12:06'!
allocateSlotsForPinningInOldSpace: numSlots bytes: totalBytes format: formatField classIndex: classIndex
	"Answer the oop of a chunk of space in oldSpace with numSlots slots.  Try and
	 allocate in a segment that already includes pinned objects.  The header of the
	 result will have been filled-in but not the contents."
	<var: #totalBytes type: #usqInt>
	<inline: false>
	| chunk |
	chunk := self allocateOldSpaceChunkOfBytes: totalBytes
				   suchThat: [:f| (segmentManager segmentContainingObj: f) containsPinned].
	chunk ifNil:
		[chunk := self allocateOldSpaceChunkOfBytes: totalBytes.
		 chunk ifNil:
			[^nil].
		 (segmentManager segmentContainingObj: chunk) containsPinned: true].
	numSlots >= self numSlotsMask ifTrue: "for header parsing we put a saturated slot count in the prepended overflow size word"
		[self longAt: chunk
			put: numSlots + (self numSlotsMask << self numSlotsFullShift).
		 self longAt: chunk + self baseHeaderSize
			put: ((self headerForSlots: self numSlotsMask format: formatField classIndex: classIndex)
					bitOr: 1 << self pinnedBitShift).
		 self checkFreeSpace: GCModeNewSpace.
		 ^chunk + self baseHeaderSize].
	self longAt: chunk
		put: ((self headerForSlots: numSlots format: formatField classIndex: classIndex)
				bitOr: 1 << self pinnedBitShift).
	self checkFreeSpace: GCModeNewSpace.
	^chunk! !

!Spur64BitMemoryManager methodsFor: 'allocation' stamp: 'eem 3/9/2020 16:13'!
allocateSlotsInOldSpace: numSlots bytes: totalBytes format: formatField classIndex: classIndex
	"Answer the oop of a chunk of space in oldSpace with numSlots slots.  The header
	 will have been filled-in but not the contents.  If no memory is available answer nil."
	<var: #totalBytes type: #usqInt>
	<inline: false>
	| chunk |
	chunk := self allocateOldSpaceChunkOfBytes: totalBytes.
	chunk ifNil:
		[^nil].
	numSlots >= self numSlotsMask ifTrue: "for header parsing we put a saturated slot count in the prepended overflow size word"
		[self longAt: chunk
			put: numSlots + (self numSlotsMask << self numSlotsFullShift).
		 self longAt: chunk + self baseHeaderSize
			put: (self headerForSlots: self numSlotsMask format: formatField classIndex: classIndex).
		 self checkFreeSpace: GCModeNewSpace ignoring: chunk + self baseHeaderSize.
		 ^chunk + self baseHeaderSize].
	self longAt: chunk
		put: (self headerForSlots: numSlots format: formatField classIndex: classIndex).
	self checkFreeSpace: GCModeNewSpace ignoring: chunk.
	^chunk! !

!Spur64BitMemoryManager methodsFor: 'snapshot' stamp: 'nice 4/2/2016 22:49'!
byteSwapped: w
	"Answer the given integer with its bytes in the reverse order."
	<api>
	<returnTypeC: #sqInt>
	^w byteSwap64! !

!Spur64BitMemoryManager methodsFor: 'object enumeration' stamp: 'eem 06/22/2021 15:31'!
bytesInBody: objOop
	"Answer the total number of bytes in an object including header and possible overflow size header."
	<returnTypeC: #usqInt>
	| header headerNumSlots numSlots |
	<var: 'header' type: #usqLong>
	self flag: #endianness.
	header := self longAt: objOop.
	headerNumSlots := header >> self numSlotsFullShift.
	numSlots := headerNumSlots = self numSlotsMask
					ifTrue: [self rawOverflowSlotsOf: objOop]
					ifFalse: [headerNumSlots = 0 ifTrue: [1] ifFalse: [headerNumSlots]].
	^numSlots << self shiftForWord
	+ (headerNumSlots = self numSlotsMask
		ifTrue: [self baseHeaderSize + self baseHeaderSize]
		ifFalse: [self baseHeaderSize])! !

!Spur64BitMemoryManager methodsFor: 'object enumeration' stamp: 'eem 12/24/2016 10:45'!
bytesInObject: objOop given: rawNumSotsOfObjOop
	"Answer the total number of bytes in an object including header and possible overflow size header, given the value of the object's numSlots field."
	<returnTypeC: #usqInt>
	<inline: true>
	| numSlots |
	numSlots := rawNumSotsOfObjOop = self numSlotsMask
					ifTrue: [self rawOverflowSlotsOf: objOop]
					ifFalse: [rawNumSotsOfObjOop = 0 ifTrue: [1] ifFalse: [rawNumSotsOfObjOop]].
	^numSlots << self shiftForWord
	+ (rawNumSotsOfObjOop = self numSlotsMask
		ifTrue: [self baseHeaderSize + self baseHeaderSize]
		ifFalse: [self baseHeaderSize])! !

!Spur64BitMemoryManager methodsFor: 'accessing' stamp: 'eem 3/12/2014 10:26'!
bytesPerOop
	^8! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 6/21/2021 22:25'!
changeClassOf: rcvr to: argClass
	"Attempt to change the class of the receiver to the argument given that the
	 format of the receiver matches the format of the argument.  If successful,
	 answer 0, otherwise answer an error code indicating the reason for failure. 
	 Fail if the format of the receiver is incompatible with the format of the argument,
	 or if the argument is a fixed class and the receiver's size differs from the size
	 that an instance of the argument should have."
	<inline: false>
	| classFormat fixedFields instFormat newFormat classIndex instSlots instBytes |
	(self isObjImmutable: rcvr) ifTrue:
		[^PrimErrNoModification].
	classFormat := self formatOfClass: argClass.
	fixedFields := self fixedFieldsOfClassFormat: classFormat.
	classFormat := self instSpecOfClassFormat: classFormat.
	instFormat := self formatOf: rcvr.

	"Fail for inability to access classIndex before making contexts snapshot-safe."
	(classIndex := self ensureBehaviorHash: argClass) < 0 ifTrue:
		[^classIndex negated].

	"Now check the instance for compatibility and compute odd bits if necessary."
	classFormat <= self lastPointerFormat
		ifTrue:
			[instFormat > self lastPointerFormat ifTrue:
				[^PrimErrInappropriate].
			 ((instSlots := self numSlotsOf: rcvr) < fixedFields
			  or: [instSlots > fixedFields and: [self isFixedSizePointerFormat: classFormat]]) ifTrue:
				[^PrimErrBadReceiver].
			 (instFormat = self indexablePointersFormat
			  and: [self isContextNonImm: rcvr]) ifTrue:
				[coInterpreter makeContextSnapshotSafe: rcvr].
			 newFormat := classFormat]
		ifFalse:
			["Fail if the class's format is somehow invalid."
			 classFormat ~= (self classFormatFromInstFormat: classFormat) ifTrue:
				[^PrimErrBadArgument].

			 instBytes := self numBytesOf: rcvr.
			 classFormat
				caseOf: {
				[self sixtyFourBitIndexableFormat]
					-> [(instBytes anyMask: 7) ifTrue: [^PrimErrBadReceiver].
						newFormat := classFormat].
				[self firstLongFormat] 		
					-> [(instBytes anyMask: 3) ifTrue: [^PrimErrBadReceiver].
						newFormat := classFormat + (2 - (instBytes >> 2) bitAnd: 1)].
				[self firstShortFormat] 		
					-> [(instBytes anyMask: 1) ifTrue: [^PrimErrBadReceiver].
						newFormat := classFormat + (4 - (instBytes >> 1) bitAnd: 3)].
				[self firstByteFormat]
					-> [newFormat := classFormat + (8 - instBytes bitAnd: 7)].
				[self firstCompiledMethodFormat]
					-> [classFormat ~= self firstCompiledMethodFormat ifTrue:
							[^PrimErrInappropriate].
						newFormat := instFormat] }
				otherwise: "bits instances cannot be adopted by pointer-like classes..."
					[^PrimErrInappropriate]].

	self set: rcvr classIndexTo: classIndex formatTo: newFormat.

	self assert: (self numBytesOf: rcvr) = (classFormat <= self lastPointerFormat
											ifTrue: [instSlots * self bytesPerOop]
											ifFalse: [instBytes]).
	"ok"
	^0! !

!Spur64BitMemoryManager methodsFor: 'gc - global' stamp: 'eem 06/22/2021 15:31'!
coalesce: obj1 and: obj2
	| header1NumSlots header2NumSlots obj2slots newNumSlots |
	header1NumSlots := self rawNumSlotsOf: obj1.
	header2NumSlots := self rawNumSlotsOf: obj2.

	"compute total number of slots in obj2, including header"
	obj2slots := header2NumSlots = self numSlotsMask
					ifTrue: [(self rawOverflowSlotsOf: obj2) + (2 * self baseHeaderSize / self wordSize)]
					ifFalse: [(header2NumSlots = 0 ifTrue: [1] ifFalse: [header2NumSlots]) + (self baseHeaderSize / self wordSize)].
	self assert: obj2slots * self wordSize = (self bytesInBody: obj2).

	"if obj1 already has a double header things are simple..."
	header1NumSlots = self numSlotsMask ifTrue:
		[self rawOverflowSlotsOf: obj1 put: obj2slots + (self rawOverflowSlotsOf: obj1).
		 ^obj1].

	"compute total number of slots in obj1, excluding header"
	header1NumSlots := header1NumSlots = 0 ifTrue: [1] ifFalse: [header1NumSlots].
	self assert: header1NumSlots * self wordSize + self baseHeaderSize = (self bytesInBody: obj1).
	newNumSlots := obj2slots + header1NumSlots.

	"if obj1 still only requires a single header things are simple..."
	newNumSlots < self numSlotsMask ifTrue:
		[self rawNumSlotsOf: obj1 put: newNumSlots.
		 ^obj1].

	"convert from single to double header..."
	newNumSlots := newNumSlots - (self baseHeaderSize / self wordSize).
	self
		rawNumSlotsOf: obj1 + self baseHeaderSize put: self numSlotsMask;
		rawOverflowSlotsOf: obj1 + self baseHeaderSize put: newNumSlots.
	^obj1 + self baseHeaderSize! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 12/1/2014 10:57'!
dbgFloatValueOf: oop
	"Answer the C double precision floating point value of the argument,
	 or if it is not, answer 0."

	| result tagBits |
	<returnTypeC: #double>
	<var: #result type: #double>
	(tagBits := oop bitAnd: self tagMask) ~= 0
		ifTrue:
			[tagBits = self smallFloatTag ifTrue:
				[^self smallFloatValueOf: oop]]
		ifFalse:
			[(self classIndexOf: oop) = ClassFloatCompactIndex ifTrue:
				[self cCode: '' inSmalltalk: [result := Float new: 2].
				 self fetchFloatAt: oop + self baseHeaderSize into: result.
				 ^result]].
	^0.0! !

!Spur64BitMemoryManager methodsFor: 'snapshot' stamp: 'eem 10/27/2016 12:15'!
defaultEdenBytes
	"Answer the default amount of memory to allocate for the eden space.
	 The actual value can be set via vmParameterAt: and/or a preference in the ini file.
	 The shootout tests seem to plateau at 5 or 6Mb.

	 Originally, both the 32-bit and 64-bit versions used the same 4Mb default.  Measuring
	 the simulator on image start-up, the 64-bit system's eden at the same point in start-up
	 (the first copyBits) is only 8% larger in bytes because it allocates 26% fewer objects.
	 Some 21% of the objects in the 32-bit version's eden are large integers and floats that
	 are representable as 64-bit immediates.

	 But when running benchmarks such as the computer language shootout's binary trees,
	 using the same amount of memory for the 64-bit system causes a significant slow-down
	 and a lot of compactions.  So we now use 4Mb for 32-bits and 7Mb for 64-bits."
	<inline: false>
	^7 * 1024 * 1024

	"Here are the expressions used to collect the above-mentioned figures:"
	"freeStart - self newSpaceStart"
	"| n | n := 0. self allNewSpaceObjectsDo: [:o| n := n + 1]. n"
	"| f | f := 0. self allNewSpaceObjectsDo: [:o| (self isFloatObject: o) ifTrue: [f := f + 1]]. f"
	"| l | l := 0. self allNewSpaceObjectsDo: [:o|
		(((self classIndexOf: o)
			between: ClassLargeNegativeIntegerCompactIndex
			and: ClassLargePositiveIntegerCompactIndex)
		 and: [(self numBytesOfBytes: o) <= 8
		 and: [self isIntegerValue: (coInterpreter signed64BitValueOf: o)]]) ifTrue: [l := l + 1]]. l"! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 12/18/2014 13:56'!
exponentOfSmallFloat: oop
	"Answer the exponent of the argument, a SmallFloat.
	 See section 61-bit Immediate Floats in the SpurMemoryManager class comment.
				msb                                              lsb 
				[8expsubset][52mantissa][1s][3tags]"
	| exp |
	self assert: (oop bitAnd: self tagMask) = self smallFloatTag.
	^oop asUnsignedInteger <= 15
		ifTrue: [0]
		ifFalse:
			[exp := oop asUnsignedInteger >> (self numTagBits + self smallFloatMantissaBits + 1).
			 exp + self smallFloatExponentOffset - 1022]! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 2/24/2020 15:18'!
fetchClassTagOf: oop
	"Answer the tag used in lookup caches for a receiver.  This is the receiver's classIndex."
	<api>
	| tagBits |
	^(tagBits := oop bitAnd: self tagMask) ~= 0
		ifTrue: [tagBits]
		ifFalse: [self classIndexOf: oop]! !

!Spur64BitMemoryManager methodsFor: 'object access' stamp: 'eem 8/20/2020 10:52'!
fetchLong32: fieldIndex ofFloatObject: oop
	"index by word size, and return a pointer as long as the word size"
	<returnTypeC: #'unsigned int'>
	| bits |
	(self isImmediateFloat: oop) ifFalse:
		[^self fetchLong32: fieldIndex ofObject: oop].
	bits := self smallFloatBitsOf: oop.
	^fieldIndex = 0
		ifTrue: [bits bitAnd: 16rFFFFFFFF]
		ifFalse: [bits >> 32]! !

!Spur64BitMemoryManager methodsFor: 'instantiation' stamp: 'eem 11/13/2014 12:15'!
fillObj: objOop numSlots: numSlots with: fillValue
	<inline: true>
	<var: #p type: #usqInt>
	self assert: (self oop: objOop + self baseHeaderSize + (numSlots * self bytesPerOop) - 1
					isLessThan: (self addressAfter: objOop)).
	(objOop + self baseHeaderSize) asUnsignedInteger
		to: (objOop + self baseHeaderSize + (numSlots * self bytesPerOop) - 1) asUnsignedInteger
		by: self allocationUnit
		do: [:p| self longAt: p put: fillValue]! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'RonieSalgado 5/1/2016 20:56'!
floatObjectOf: aFloat
	| newFloatObj |
	<inline: false>
	<var: #aFloat type: #double>
	<api>
	(self isSmallFloatValue: aFloat) ifTrue:
		[^self smallFloatObjectOf: aFloat].
	newFloatObj := self
						eeInstantiateSmallClassIndex: ClassFloatCompactIndex
						format: self firstLongFormat
						numSlots: (self sizeof: #double) / self bytesPerOop.
	self storeFloatAt: newFloatObj + self baseHeaderSize from: aFloat.
	^newFloatObj! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 1/18/2020 18:09'!
floatObjectOfBits: doubleFloatBits
	<var: 'doubleFloatBits' type: #sqLong>
	| newFloatObj |
	(self isSmallFloatValueBits: doubleFloatBits) ifTrue:
		[^self smallFloatObjectOfBits: doubleFloatBits].
	newFloatObj := self
						eeInstantiateSmallClassIndex: ClassFloatCompactIndex
						format: self firstLongFormat
						numSlots: (self sizeof: #double) / self bytesPerOop.
	self storeLong64: 0 ofObject: newFloatObj withValue: doubleFloatBits.
	^newFloatObj! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 1/18/2020 18:23'!
floatValueBitsOf: floatOop
	"Answer the 64-bit value of the argument as raw bits."
	<inline: false>
	self assert: (self isFloatInstance: floatOop).
	(floatOop bitAnd: self tagMask) ~= 0 ifTrue:
		[^self smallFloatBitsOf: floatOop].
	^self fetchLong64: 0 ofObject: floatOop! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/14/2019 13:43'!
floatValueOf: oop
	"Answer the C double precision floating point value of the argument,
	 or fail if it is not a Float, and answer 0.
	 Note: May be called by translated primitive code."
	<api>
	<returnTypeC: #double>
	| result tagBits |
	<var: #result type: #double>
	(tagBits := oop bitAnd: self tagMask) ~= 0
		ifTrue:
			[tagBits = self smallFloatTag ifTrue:
				[^self smallFloatValueOf: oop]]
		ifFalse:
			[(self classIndexOf: oop) = ClassFloatCompactIndex ifTrue:
				[self cCode: '' inSmalltalk: [result := Float new: 2].
				 self fetchFloatAt: oop + self baseHeaderSize into: result.
				 ^result]].
	coInterpreter primitiveFail.
	^0.0! !

!Spur64BitMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 6/22/2021 11:31'!
hackSlimBridgeTo: objOop at: startAddress
	"This is a horrible hack for getting to the first object in eden if pastSpace is almost full.
	 If there is only one (64-bit) word at the end of pastSpace there is no room for a full
	 bridge, but there is room for this hack."
	self assert: (self oop: startAddress isLessThan: objOop).
	self assert: (self oop: objOop isLessThanOrEqualTo: oldSpaceStart).
	self assert: (startAddress + self allocationUnit = objOop
				or: [startAddress + self allocationUnit + self allocationUnit = objOop]).
	self longAt: startAddress
		put: (startAddress + self allocationUnit = objOop
				ifTrue: [self numSlotsMask << self numSlotsFullShift]
				ifFalse: [self numSlotsMask << self numSlotsFullShift + 1])! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 2/12/2015 17:34'!
hasSixtyFourBitImmediates
	"The V3 64-bit memory manager has 64-bit oops, but its SmallIntegers only have
	 31 bits of precision.  The Spur 64-bit memory manager has 61-bit immediates."
	^true! !

!Spur64BitMemoryManager methodsFor: 'method access' stamp: 'eem 6/7/2016 14:52'!
headerIndicatesAlternateBytecodeSet: methodHeader
	"A negative header selects the alternate bytecode set."
	<api>
	<inline: true>
	^methodHeader signedIntFromLong64 < 0! !

!Spur64BitMemoryManager methodsFor: 'immediates' stamp: 'eem 3/10/2016 12:45'!
imageFormatVersion
	"Return a magic constant that changes when the image format changes.
	 Since the image reading code uses this to detect byte ordering, one
	 must avoid version numbers that are invariant under byte reversal.
	 N.B. Bit 4 (=16) is the isSpur bit"

	"^68019" "The first 64-bit Spur format, which unfortunately chose SmallFloat64's tag to be 3."
	^68021! !

!Spur64BitMemoryManager methodsFor: 'free space' stamp: 'eem 11/13/2014 12:15'!
initFreeChunkWithBytes: numBytes at: address
	<var: #numBytes type: #usqLong>
	| numSlots |
	"must have room for a header (single or double) plus the next free pointer"
	self assert: (numBytes \\ self allocationUnit = 0
				 and: [numBytes >= (self baseHeaderSize + self bytesPerOop)]).
	"double header"
	numBytes >= ((self numSlotsMask << self shiftForWord) + self baseHeaderSize) ifTrue:
		[numSlots := numBytes - self baseHeaderSize - self baseHeaderSize >> self shiftForWord.
		 self longAt: address put: self numSlotsMask << self numSlotsFullShift + numSlots;
			longAt: address + 8 put: self numSlotsMask << self numSlotsFullShift. "0's classIndex; 0 = classIndex of free chunks"
		^address + 8].
	"single header"
	numSlots := numBytes - self baseHeaderSize >> self shiftForWord.
	self assert: numSlots < self numSlotsMask.
	self longAt: address put: numSlots << self numSlotsFullShift. "0's classIndex; 0 = classIndex of free chunks"
	^address! !

!Spur64BitMemoryManager methodsFor: 'object enumeration' stamp: 'eem 9/24/2020 14:48'!
initSegmentBridgeWithBytes: numBytes at: address
	<var: #numBytes type: #usqLong>
	| numSlots |
	"Must have room for a double header or a short object with the forwarding slot (16 bytes either way)."
	self assert: (numBytes \\ self allocationUnit = 0
				 and: [numBytes >= (self baseHeaderSize + self baseHeaderSize)]).
	numSlots := numBytes - self baseHeaderSize - self baseHeaderSize >> self shiftForWord.
	numSlots = 0
		ifTrue: "short bridge for adjacent segments"
			[self longAt: address
					put:    (1 << self pinnedBitShift)
						+ (1 << self markedBitFullShift)
						+ (self wordIndexableFormat << self formatShift)
						+ self segmentBridgePun]
		ifFalse: "long bridge"
			[self longAt: address
					put: self numSlotsMask << self numSlotsFullShift + numSlots;
				longAt: address + self baseHeaderSize
					put: (self numSlotsMask << self numSlotsFullShift)
						+ (1 << self pinnedBitShift)
						+ (1 << self markedBitFullShift)
						+ (self wordIndexableFormat << self formatShift)
						+ self segmentBridgePun]! !

!Spur64BitMemoryManager methodsFor: 'instantiation' stamp: 'eem 8/22/2017 14:46'!
instantiateClass: classObj indexableSize: nElements
	<api>
	<var: #nElements type: #usqInt>
	"Allocate an instance of a variable class, excepting CompiledMethod."
	| instSpec classFormat numSlots classIndex newObj fillValue |
	classFormat := self formatOfClass: classObj.
	instSpec := self instSpecOfClassFormat: classFormat.
	classIndex := self rawHashBitsOf: classObj.
	fillValue := 0.
	instSpec caseOf: {
		[self arrayFormat]	->
			[numSlots := nElements.
			 fillValue := nilObj].
		[self indexablePointersFormat]	->
			[numSlots := (self fixedFieldsOfClassFormat: classFormat) + nElements.
			 fillValue := nilObj].
		[self weakArrayFormat]	->
			[numSlots := (self fixedFieldsOfClassFormat: classFormat) + nElements.
			 fillValue := nilObj].
		[self sixtyFourBitIndexableFormat]	->
			[numSlots := nElements].
		[self firstLongFormat]	->
			[(classIndex = ClassFloatCompactIndex and: [nElements ~= 2]) ifTrue:
				[coInterpreter primitiveFailFor: PrimErrBadReceiver.
				 ^nil].
			 numSlots := nElements + 1 // 2.
			 instSpec := instSpec + (nElements bitAnd: 1)].
		[self firstShortFormat]	->
			[numSlots := nElements + 3 // 4.
			 instSpec := instSpec + (4 - nElements bitAnd: 3)].
		[self firstByteFormat]	->
			[numSlots := nElements + 7 // 8.
			 instSpec := instSpec + (8 - nElements bitAnd: 7)] }
		otherwise: "non-indexable"
			["Some Squeak images include funky fixed subclasses of abstract variable
			  superclasses. e.g. DirectoryEntry as a subclass of ArrayedCollection.
			  The (Threaded)FFIPlugin expects to be able to instantiate ExternalData via
			  this method.
			  Hence allow fixed classes to be instantiated here iff nElements = 0."
			 (nElements ~= 0 or: [instSpec > self lastPointerFormat]) ifTrue:
				[^nil].
			 numSlots := self fixedFieldsOfClassFormat: classFormat.
			 fillValue := nilObj].
	classIndex = 0 ifTrue:
		[classIndex := self ensureBehaviorHash: classObj.
		 classIndex < 0 ifTrue:
			[coInterpreter primitiveFailFor: classIndex negated.
			 ^nil]].
	numSlots > self maxSlotsForNewSpaceAlloc
		ifTrue:
			[numSlots > self maxSlotsForAlloc ifTrue:
				[coInterpreter primitiveFailFor: PrimErrUnsupported.
				 ^nil].
			 newObj := self allocateSlotsInOldSpace: numSlots format: instSpec classIndex: classIndex]
		ifFalse:
			[newObj := self allocateSlots: numSlots format: instSpec classIndex: classIndex].
	newObj ifNotNil:
		[self fillObj: newObj numSlots: numSlots with: fillValue].
	^newObj! !

!Spur64BitMemoryManager methodsFor: 'instantiation' stamp: 'eem 8/12/2014 19:53'!
instantiateCompiledMethodClass: classObj indexableSize: nElements
	<var: #nElements type: #usqInt>
	"Allocate an instance of a CompiledMethod class."
	<inline: true>
	| instSpec classFormat numSlots classIndex newObj |
	classFormat := self formatOfClass: classObj.
	instSpec := self instSpecOfClassFormat: classFormat.
	instSpec ~= self firstCompiledMethodFormat ifTrue:
		[^nil].
	numSlots := nElements + 7 // 8.
	instSpec := instSpec + (8 - nElements bitAnd: 7).
	classIndex := self ensureBehaviorHash: classObj.
	classIndex < 0 ifTrue:
		[coInterpreter primitiveFailFor: classIndex negated.
		 ^nil].
	numSlots > self maxSlotsForNewSpaceAlloc
		ifTrue:
			[numSlots > self maxSlotsForAlloc ifTrue:
				[^nil].
			 newObj := self allocateSlotsInOldSpace: numSlots format: instSpec classIndex: classIndex]
		ifFalse:
			[newObj := self allocateSlots: numSlots format: instSpec classIndex: classIndex].
	newObj ifNotNil:
		[self fillObj: newObj numSlots: numSlots with: 0].
	^newObj! !

!Spur64BitMemoryManager methodsFor: 'immediates' stamp: 'nice 10/30/2015 01:27'!
integerObjectOf: value
	"Convert the integer value, assumed to be in SmallInteger range, into a tagged SmallInteger object.
	 In C, use a shift and an add to set the tag bit.
	 In Smalltalk we have to work harder because the simulator works with strictly positive bit patterns."
	<returnTypeC: #sqInt>
	^self
		cCode: [value asUnsignedInteger << self numTagBits + 1]
		inSmalltalk: [value << self numTagBits
					+ (value >= 0
						ifTrue: [1]
						ifFalse: [16r10000000000000001])]! !

!Spur64BitMemoryManager methodsFor: 'immediates' stamp: 'eem 11/19/2014 21:20'!
integerObjectOfCharacterObject: oop
	"Immediate characters are unsigned"
	^oop - (self characterTag - self smallIntegerTag)! !

!Spur64BitMemoryManager methodsFor: 'immediates' stamp: 'eem 11/18/2014 19:06'!
integerValueOf: oop
	"Translator produces 'oop >> 3'"
	^(oop bitShift: -63) = 1 "tests top bit"
		ifTrue: "negative"
			[((oop bitShift: self numTagBits negated) bitAnd: 16r1FFFFFFFFFFFFFFF) - 16r1FFFFFFFFFFFFFFF - 1  "Faster than -16r4000000000000000 (a LgInt)"]
		ifFalse: "positive"
			[oop bitShift: self numTagBits negated]! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/21/2014 17:05'!
isFloatInstance: oop
	<inline: true>
	| tagBits |
	^(tagBits := oop bitAnd: self tagMask) ~= 0
		ifTrue: [tagBits = self smallFloatTag]
		ifFalse: [(self classIndexOf: oop) = ClassFloatCompactIndex]! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 6/5/2017 10:30'!
isFloatOrInt: anOop
	"Answer if anOop is either a SmallInteger or a Float."

	<inline: true>
	^(self isImmediate: anOop)
		ifTrue: [(self isCharacterObject: anOop) not]
		ifFalse: [(self classIndexOf: anOop) = ClassFloatCompactIndex]! !

!Spur64BitMemoryManager methodsFor: 'object testing' stamp: 'eem 3/9/2016 13:56'!
isImmediateCharacter: oop
	<inline: true>
	^oop anyMask: self characterTag! !

!Spur64BitMemoryManager methodsFor: 'object testing' stamp: 'eem 3/9/2016 13:52'!
isImmediateFloat: oop
	<inline: true>
	^oop anyMask: self smallFloatTag! !

!Spur64BitMemoryManager methodsFor: 'object testing' stamp: 'eem 3/9/2016 13:53'!
isIntegerObject: oop
	^oop anyMask: self smallIntegerTag! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'nice 3/4/2016 23:40'!
isIntegerValue: intValue
	"Answer if the given value can be represented as a Smalltalk integer value.
	 In 64-bits we use a 3 bit tag which leaves 61 bits for 2's complement signed
	 integers. In C, use a shift add and mask to test if the top 4 bits are all the same.
	 Since 16rFFFFFFFFFFFFFFFF >> 60 = 16rF the computation intValue >> 60 + 1 bitAnd: 16rF
	 maps in-range -ve values to 0 and in-range +ve values to 1."
	<api>
	^self
		cCode: [(intValue >> 60 + 1 bitAnd: 16rF) <= 1] "N.B. (16rFFFFFFFFFFFFFFFF >> 60) + 1 = 16"
		inSmalltalk: [intValue >= self minSmallInteger and: [intValue <= self maxSmallInteger]]! !

!Spur64BitMemoryManager methodsFor: 'free space' stamp: 'cb 10/5/2018 12:48'!
isLilliputianSize: chunkBytes
	"To have a prev pointer, which follows the next pointer, we need at least two slots."
	self assert: chunkBytes >= (self baseHeaderSize + self allocationUnit).
	^chunkBytes = (self baseHeaderSize + self allocationUnit)! !

!Spur64BitMemoryManager methodsFor: 'header access' stamp: 'eem 11/18/2013 13:52'!
isMarked: objOop
	<api>
	^((self longAt: objOop) >> self markedBitFullShift bitAnd: 1) ~= 0! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 2/12/2015 09:37'!
isNegativeSmallFloat: aSmallFloat
	<inline: true>
	self assert: (self isImmediateFloat: aSmallFloat).
	^aSmallFloat anyMask: self smallFloatSignBit! !

!Spur64BitMemoryManager methodsFor: 'object testing' stamp: 'eem 3/9/2016 13:53'!
isNonIntegerObject: oop
	^oop noMask: self smallIntegerTag! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 9/19/2018 17:31'!
isSmallFloatValue: aFloat
	<inline: true>
	<var: #rawFloat type: #usqLong>
	<var: #aFloat type: #double>
	| exponent rawFloat |
	self
		cCode: [self memcpy: (self addressOf: rawFloat) _: (self addressOf: aFloat) _: (self sizeof: rawFloat)]
		inSmalltalk: [rawFloat := (aFloat at: 1) << 32 + (aFloat at: 2)].
	exponent := rawFloat >> self smallFloatMantissaBits bitAnd: 16r7FF.
	^exponent > self smallFloatExponentOffset
	 	ifTrue: [exponent <= (255 + self smallFloatExponentOffset)]
		ifFalse:
			[(rawFloat bitAnd: (1 << self smallFloatMantissaBits - 1)) = 0
				ifTrue: [exponent = 0]
				ifFalse: [exponent = self smallFloatExponentOffset]]! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 1/18/2020 18:20'!
isSmallFloatValueBits: rawFloatBits
	<inline: true>
	<var: #rawFloatBits type: #usqLong>
	| exponent |
	exponent := rawFloatBits >> self smallFloatMantissaBits bitAnd: 16r7FF.
	^exponent > self smallFloatExponentOffset
	 	ifTrue: [exponent <= (255 + self smallFloatExponentOffset)]
		ifFalse:
			[(rawFloatBits bitAnd: (1 << self smallFloatMantissaBits - 1)) = 0
				ifTrue: [exponent = 0]
				ifFalse: [exponent = self smallFloatExponentOffset]]! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 2/12/2015 10:37'!
isSmallFloatZero: aSmallFloat
	<inline: #always>
	self assert: (self isImmediateFloat: aSmallFloat).
	^aSmallFloat asUnsignedInteger <= self negativeSmallFloatZero! !

!Spur64BitMemoryManager methodsFor: 'object testing' stamp: 'eem 2/23/2016 21:36'!
isWordsNonImm: objOop
	"Answer if the argument contains only indexable words (no oops). See comment in formatOf:"

	^(self formatOf: objOop) between: self firstLongFormat and: self firstShortFormat - 1! !

!Spur64BitMemoryManager methodsFor: 'allocation' stamp: 'eem 11/13/2014 12:15'!
largeObjectBytesForSlots: numSlots
	"Answer the total number of bytes in an object with an overflow header, including header bytes."
	<returnTypeC: #usqInt>
	^self baseHeaderSize + self baseHeaderSize "double header"
	+ (numSlots * self bytesPerOop) "no rounding; bytesPerOop = allocationUnit"! !

!Spur64BitMemoryManager methodsFor: 'object access' stamp: 'eem 2/6/2015 11:17'!
lengthOf: objOop format: fmt
	"Answer the number of indexable units in the given object.
	 For a CompiledMethod, the size of the method header (in bytes)
	 should be subtracted from the result of this method."
	| numSlots |
	<inline: true>
	numSlots := self numSlotsOfAny: objOop. "don't let forwarders freak us out..."
	fmt <= self ephemeronFormat ifTrue:
		[^numSlots].
	fmt >= self firstByteFormat ifTrue: "bytes, including CompiledMethod"
		[^numSlots << self shiftForWord - (fmt bitAnd: 7)].
	fmt >= self firstShortFormat ifTrue:
		[^numSlots << (self shiftForWord - 1) - (fmt bitAnd: 3)].
	fmt >= self firstLongFormat ifTrue:
		[^numSlots << (self shiftForWord - 2) - (fmt bitAnd: 1)].
	fmt = self sixtyFourBitIndexableFormat ifTrue:
		[^numSlots].
	"fmt = self forwardedFormat"
	^0! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'nice 9/5/2019 21:32'!
loadFloatOrIntFrom: floatOrIntOop
	"If floatOrInt is an integer and we enable mixed arithmetic in primitives, then convert it to a C double float and return it.
	 If it is a Float, then load its value and return it.
	 Otherwise fail -- ie return with primErrorCode non-zero."

	<inline: true>
	<returnTypeC: #double>
	| result tagBits |
	<var: #result type: #double>

	(tagBits := floatOrIntOop bitAnd: self tagMask) ~= 0
		ifTrue:
			[tagBits = self smallFloatTag ifTrue:
				[^self smallFloatValueOf: floatOrIntOop].
			 (coInterpreter primitiveDoMixedArithmetic and: [tagBits = self smallIntegerTag]) ifTrue:
				[^(self integerValueOf: floatOrIntOop) asFloat]]
		ifFalse:
			[(self classIndexOf: floatOrIntOop) = ClassFloatCompactIndex ifTrue:
				[self cCode: '' inSmalltalk: [result := Float new: 2].
				 self fetchFloatAt: floatOrIntOop + self baseHeaderSize into: result.
				 ^result]].
	coInterpreter primitiveFail.
	^0.0! !

!Spur64BitMemoryManager methodsFor: 'accessing' stamp: 'eem 11/14/2014 12:13'!
logBytesPerOop
	^3! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 2/12/2015 09:42'!
mapSignedSmallFloatToSignedSmallFloatZero: aSmallFloat
	<inline: true>
	self assert: (self isImmediateFloat: aSmallFloat).
	^aSmallFloat bitAnd: self smallFloatSignBit + self tagMask! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 1/8/2015 14:41'!
maxCInteger
	^16rFFFFFFFFFFFFFFFF! !

!Spur64BitMemoryManager methodsFor: 'instantiation' stamp: 'eem 12/11/2014 11:27'!
maxSlotsForAlloc
	"Answer the maximum number of slots we are willing to attempt to allocate in an object.
	 Must fit in 56-bits; c.f. bytesInObject:"
	^1024*1024*1024*1024! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/19/2014 15:55'!
maxSmallInteger
	^16rFFFFFFFFFFFFFFF! !

!Spur64BitMemoryManager methodsFor: 'simulation' stamp: 'eem 11/18/2016 16:24'!
memoryClass
	"Answer the class to use for the memory inst var in simulation.
	 Answer nil if a suitable class isn't available.  This version emulates 64-bit access given a 32-bit element array."
	<doNotGenerate>
	^self endianness == #little
		ifTrue: [LittleEndianBitmap]
		ifFalse: [Bitmap]! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 6/5/2017 11:31'!
methodDictionaryHash: oop mask: mask
	<inline: true>
	^mask bitAnd: ((self isImmediate: oop)
						ifTrue: [self integerValueOf: oop] "this will fail for SmallFloat64 but we don't care"
						ifFalse: [self rawHashBitsOf: oop]) "If no hash then it isn't in the dictionary..."! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/19/2014 15:12'!
minSmallInteger
	^-16r1000000000000000! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 2/12/2015 09:36'!
negativeSmallFloatZero
	<inline: true>
	^self smallFloatSignBit + self smallFloatTag! !

!Spur64BitMemoryManager methodsFor: 'simulation' stamp: 'eem 10/23/2020 13:41'!
newInputEventAccessorOfSize: numElements
	<doNotGenerate>
	^CArrayAccessor on: (DoubleWordArray new: 8)! !

!Spur64BitMemoryManager methodsFor: 'indexing primitive support' stamp: 'eem 6/21/2021 22:33'!
normalisedFormatFor: objOop indexableSize: indexableSize
	"Answer the format for an objOop clone with indexableSize.  Needs to work for longFormat and arrayFormat only."
	| format |
	format := self classFormatFromInstFormat: (self formatOf: objOop).
	^(format = self firstLongFormat and: [indexableSize anyMask: 1])
		ifTrue: [format + 1]
		ifFalse: [format]! !

!Spur64BitMemoryManager methodsFor: 'free space' stamp: 'eem 11/18/2014 09:08'!
numFreeLists
	"Answer the number of free lists.  We use freeListsMask, a bitmap, to avoid
	 reading empty list heads.  This should fit in a machine word to end up in a
	 register during free chunk allocation."
	^64! !

!Spur64BitMemoryManager methodsFor: 'allocation' stamp: 'eem 6/21/2021 11:53'!
numSlotsForShortening: objOop toIndexableSize: indexableSize
	<inline: true>
	^(self formatOf: objOop) caseOf:
		{ [self arrayFormat]			-> [indexableSize].
		  [self firstLongFormat]		-> [self numSlotsForBytes: indexableSize * 4].
		  [self firstLongFormat + 1]	-> [self numSlotsForBytes: indexableSize * 4] }
		otherwise:
			[self assertf: 'invalid format for shortening'. 0]! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/19/2014 13:13'!
numSmallIntegerBits
	^61! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/18/2014 17:42'!
numSmallIntegerTagBits
	^3! !

!Spur64BitMemoryManager methodsFor: 'object access' stamp: 'eem 12/12/2015 14:12'!
numTagBits
	<api>
	<cmacro>
	"4th bit reserved for object alignment, which could imply e.g. what space the object is in."
	^3! !

!Spur64BitMemoryManager methodsFor: 'object enumeration' stamp: 'eem 12/18/2016 16:58'!
objectAfter: objOop limit: limit
	"Object parsing.
	1. all objects have at least a word following the header, for a forwarding pointer.
	2. objects with an overflow size have a preceding word with a saturated numSlots.  If the word
	   following an object doesn't have a saturated numSlots field it must be a single-header object.
	   If the word following does have a saturated numSlots it must be the overflow size word."
	<inline: true>
	| followingWordAddress followingWord |
	followingWordAddress := self addressAfter: objOop.
	(self oop: followingWordAddress isGreaterThanOrEqualTo: limit) ifTrue:
		[^limit].
	self flag: #endianness.
	followingWord := self longAt: followingWordAddress.
	^followingWord >> self numSlotsFullShift = self numSlotsMask
		ifTrue: [followingWordAddress + self baseHeaderSize]
		ifFalse: [followingWordAddress]! !

!Spur64BitMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 9/6/2021 19:15'!
objectAfterMaybeSlimBridge: objOop limit: limit
	"Object parsing.
	1. all objects have at least a word following the header, for a forwarding pointer.
	2. objects with an overflow size have a preceding word with a saturated numSlots.  If the word
	   following an object doesn't have a saturated numSlots field it must be a single-header object.
	   If the word following does have a saturated numSlots it must be the overflow size word.
	This variation on objectAfter:limit: allows for a single (64-bit) word bridge which may be needed
	to bridge from an almost full pastSpace to eden.  It is only used in the flat enumerators that use
	startAddressForBridgedHeapEnumeration and enumerate over pastSpace, eden and oldSpace
	in that order.  Note that the order for allObjects, and allInstances enumerates over oldSpace first.

	This hack is cheap.  It increases the size of the objectAfter code, but saves two extra copies of
	the inner loop, since the inner loop now enumerates over all of pastSpace, eden and oldSpace.
	The test for a slim bridge is only performed if applied to an overflow header, and typically only
	1 in 400 objects have overflow headers in 32-bits, 1 in 500 in 64-bits.  The complication is that
	image segment loading evaporates the word array by setting the overflow slots to 1, and this
	is ambiguous with a slimbridge.  The resolution is that if the segmentArray has an overflow header,
	and is in new space, then it slot size can be zeroed and its overflow header changed to a slimbridge."
	<inline: true>
	| followingWordAddress followingWord |
	followingWordAddress := self addressAfter: objOop.
	(self oop: followingWordAddress isGreaterThanOrEqualTo: limit) ifTrue:
		[^limit].
	self flag: #endianness.
	followingWord := self longAt: followingWordAddress.
	^followingWord >> self numSlotsFullShift = self numSlotsMask
		ifTrue:
			[((self oop: objOop isLessThan: oldSpaceStart)
			 and: [(followingWord bitAnd: 16rFFFFFFFFFFFFFF) = 1])
				ifTrue: [followingWordAddress + self baseHeaderSize + self baseHeaderSize] "skip the one word slimbridge"
				ifFalse: [followingWordAddress + self baseHeaderSize]]
		ifFalse: [followingWordAddress]! !

!Spur64BitMemoryManager methodsFor: 'object enumeration' stamp: 'eem 5/10/2017 18:18'!
objectBytesForSlots: numSlots
	"Answer the total number of bytes in an object with the given
	 number of slots, including header and possible overflow size header."
	<returnTypeC: #usqInt>
	^numSlots = 0
		ifTrue: [self allocationUnit + self baseHeaderSize] "always at least one slot for forwarding pointer"
		ifFalse:
			[numSlots << self shiftForWord
			 + (numSlots >= self numSlotsMask
					ifTrue: [self baseHeaderSize + self baseHeaderSize]
					ifFalse: [self baseHeaderSize])]! !

!Spur64BitMemoryManager methodsFor: 'object access' stamp: 'nice 7/8/2016 02:39'!
oldRawNumSlotsOf: objOop
	<returnTypeC: #usqInt>
	<inline: true>
	^(self longAt: objOop) asUnsignedInteger >> self numSlotsFullShift! !

!Spur64BitMemoryManager methodsFor: 'cog jit support' stamp: 'eem 3/9/2016 13:51'!
oldSmallFloatTag
	^3! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 2/12/2015 09:33'!
positiveSmallFloatZero
	<inline: true>
	^self smallFloatTag! !

!Spur64BitMemoryManager methodsFor: 'object access' stamp: 'nice 7/8/2016 02:39'!
rawOverflowSlotsOf: objOop
	<returnTypeC: #usqLong>
	<inline: true>
	self flag: #endianness.
	^self
		cCode: [((self longAt: objOop - self baseHeaderSize) << 8) asUnsignedInteger >> 8]
		inSmalltalk: [(self longAt: objOop - self baseHeaderSize) bitAnd: 16rFFFFFFFFFFFFFF]! !

!Spur64BitMemoryManager methodsFor: 'object access' stamp: 'eem 11/26/2014 16:05'!
rawOverflowSlotsOf: objOop put: numSlots
	<returnTypeC: #usqLong>
	<inline: true>
	self flag: #endianness.
	self longAt: objOop - self baseHeaderSize put: self numSlotsMask << 56 + numSlots.
	^numSlots! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/21/2014 16:56'!
rotateLeft: anInteger
	^(anInteger >> 63 bitAnd: 1) + ((self cCode: [anInteger] inSmalltalk: [anInteger bitAnd: 16r7FFFFFFFFFFFFFFF]) << 1)! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'nice 7/8/2016 02:40'!
rotateRight: anInteger
	^(self cCode: [anInteger] inSmalltalk: [anInteger bitAnd: 1]) << 63 + (anInteger asUnsignedInteger >> 1)! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'nice 10/24/2015 01:02'!
rotatedFloatBitsOf: oop
	"Answer the signed, but unadjusted value of a SmallFloat64, suitable for use as a hash.
	 Keeping the exponent unadjusted keeps the value in the SmallInteger range.
	 See section 61-bit Immediate Floats in the SpurMemoryManager class comment.
							msb                                             lsb 
	 Decode:				[8expsubset][52mantissa][1s][3tags] 
	 shift away tags & sign:	[   0000   ][8expsubset][52mantissa]
	 add sign:				[    ssss   ][8expsubset][52mantissa]"
	self assert: (self isImmediateFloat: oop).
	^oop asUnsignedInteger >> (self numTagBits + 1)
	 + ((oop anyMask: self smallFloatSignBit)
		ifTrue: [-1 asUnsignedInteger << (64 - self numTagBits - 1)]
		ifFalse: [0])! !

!Spur64BitMemoryManager methodsFor: 'header access' stamp: 'eem 5/5/2015 17:38'!
rtRefCountOf: obj put: refCount
	"Set the rt reference count of obj; this is the three bit field comprised
	 of isGrey,isPinned,isRemembered.  See computeRefCountToShrinkRT."
	| header |
	self assert: (refCount between: 0 and: MaxRTRefCount).
	header := self longAt: obj.
	header := header bitAnd: (7 << self rememberedBitShift) bitInvert64.
	header := header + (refCount << self rememberedBitShift).
	self longAt: obj put: header! !

!Spur64BitMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 5/9/2017 11:35'!
savedOutHashFillValue
	<inline: true>
	^self maxIdentityHash + 1 << 32 + (self maxIdentityHash + 1)! !

!Spur64BitMemoryManager methodsFor: 'header access' stamp: 'eem 1/10/2017 15:57'!
set: objOop classIndexTo: classIndex formatTo: format
	"0 = 0 sized objects (UndefinedObject True False et al)
	 1 = non-indexable objects with inst vars (Point et al)
	 2 = indexable objects with no inst vars (Array et al)
	 3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
	 4 = weak indexable objects with inst vars (WeakArray et al)
	 5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
	 6 unused, reserved for exotic pointer objects?
	 7 Forwarded Object, 1st field is pointer, rest of fields are ignored
	 8 unused, reserved for exotic non-pointer objects?
	 9 64-bit indexable
	 10 - 11 32-bit indexable
	 12 - 15 16-bit indexable
	 16 - 23 byte indexable
	 24 - 31 compiled method"
	<inline: true>
	self assert: (classIndex between: 0 and: self classIndexMask).
	self assert: (format between: 0 and: self formatMask).
	self longAt: objOop
		put: ((self longAt: objOop) bitAnd: (self formatMask << self formatShift + self classIndexMask) bitInvert64)
			+ (classIndex
			+  (format << self formatShift))! !

!Spur64BitMemoryManager methodsFor: 'header access' stamp: 'eem 9/22/2013 15:46'!
setClassIndexOf: objOop to: classIndex
	self assert: (classIndex between: 0 and: self classIndexMask).
	self longAt: objOop
		put: ((self longAt: objOop) bitAnd: self classIndexMask bitInvert64)
			+ classIndex! !

!Spur64BitMemoryManager methodsFor: 'header access' stamp: 'eem 1/10/2017 15:57'!
setFormatOf: objOop to: format
	"0 = 0 sized objects (UndefinedObject True False et al)
	 1 = non-indexable objects with inst vars (Point et al)
	 2 = indexable objects with no inst vars (Array et al)
	 3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
	 4 = weak indexable objects with inst vars (WeakArray et al)
	 5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
	 6 unused, reserved for exotic pointer objects?
	 7 Forwarded Object, 1st field is pointer, rest of fields are ignored
	 8 unused, reserved for exotic non-pointer objects?
	 9 64-bit indexable
	 10 - 11 32-bit indexable
	 12 - 15 16-bit indexable
	 16 - 23 byte indexable
	 24 - 31 compiled method"
	self assert: (format between: 0 and: self formatMask).
	self longAt: objOop
		put: ((self longAt: objOop) bitAnd: (self formatMask << self formatShift) bitInvert64)
			+ (format << self formatShift)! !

!Spur64BitMemoryManager methodsFor: 'header access' stamp: 'eem 10/15/2013 07:37'!
setIsGreyOf: objOop to: aBoolean
	self longAt: objOop
		put: (aBoolean
				ifTrue: [(self longAt: objOop) bitOr: 1 << self greyBitShift]
				ifFalse: [(self longAt: objOop) bitAnd: (1 << self greyBitShift) bitInvert64])! !

!Spur64BitMemoryManager methodsFor: 'header access' stamp: 'eem 10/15/2013 00:29'!
setIsImmutableOf: objOop to: aBoolean
	self longAt: objOop
		put: (aBoolean
				ifTrue: [(self longAt: objOop) bitOr: 1 << self immutableBitShift]
				ifFalse: [(self longAt: objOop) bitAnd: (1 << self immutableBitShift) bitInvert64])! !

!Spur64BitMemoryManager methodsFor: 'header access' stamp: 'eem 10/21/2013 15:07'!
setIsMarkedOf: objOop to: aBoolean
	self assert: (self isFreeObject: objOop) not.
	self longAt: objOop
		put: (aBoolean
				ifTrue: [(self longAt: objOop) bitOr: 1 << self markedBitFullShift]
				ifFalse: [(self longAt: objOop) bitAnd: (1 << self markedBitFullShift) bitInvert64])! !

!Spur64BitMemoryManager methodsFor: 'header access' stamp: 'eem 10/7/2013 14:15'!
setIsPinnedOf: objOop to: aBoolean
	self longAt: objOop
		put: (aBoolean
				ifTrue: [(self longAt: objOop) bitOr: 1 << self pinnedBitShift]
				ifFalse: [(self longAt: objOop) bitAnd: (1 << self pinnedBitShift) bitInvert64])! !

!Spur64BitMemoryManager methodsFor: 'header access' stamp: 'eem 9/22/2013 15:45'!
setIsRememberedOf: objOop to: aBoolean
	self longAt: objOop
		put: (aBoolean
				ifTrue: [(self longAt: objOop) bitOr: 1 << self rememberedBitShift]
				ifFalse: [(self longAt: objOop) bitAnd: (1 << self rememberedBitShift) bitInvert64])! !

!Spur64BitMemoryManager methodsFor: 'word size' stamp: 'eem 12/12/2015 14:12'!
shiftForWord
	<api>
	<cmacro>
	^3! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'nice 7/19/2016 09:43'!
smallFloatBitsOf: oop
	"Answer the ieee 754 double precision floating point bits of the argument, a SmallFloat.
	 See section 61-bit Immediate Floats in the SpurMemoryManager class comment.
							msb                                              lsb 
	 Decode:				[8expsubset][52mantissa][1s][3tags] 
	 shift away tags:		[ 000 ][8expsubset][52mantissa][1s] 
	 add exponent offset:	[     11 exponent     ][52mantissa][1s] 
	 rot sign:				[1s][     11 exponent     ][52mantissa]"
	| rot |
	<returnTypeC: #usqLong>
	<var: #rot type: #usqLong>
	self assert: (self isImmediateFloat: oop).
	rot := oop asUnsignedInteger >> self numTagBits.
	rot > 1 ifTrue: "a.k.a. ~= +/-0.0"
		[rot := rot + (self smallFloatExponentOffset << (self smallFloatMantissaBits + 1))].
	rot := self rotateRight: rot.
	^rot! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 12/24/2015 10:02'!
smallFloatExponentBits
	"SmallFloat64's have the same mantissa as IEEE single-precision floating point"
	<api>
	<cmacro>
	^8! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 12/12/2015 14:12'!
smallFloatExponentOffset
	"896 is 1023 - 127, where 1023 is the mid-point of the 11-bit double precision exponent range,
	 and 127 is the mid-point of the 8-bit SmallDouble exponent range."
	<api>
	<cmacro>
	^896! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 12/24/2015 10:01'!
smallFloatMantissaBits
	"SmallFloat64's have the same mantissa as IEEE double-precision floating point"
	<api>
	<cmacro>
	^52! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 1/18/2020 18:08'!
smallFloatObjectOf: aFloat
	"Encode the argument, aFloat in the SmallFloat range, as a tagged small float.
	 See section 61-bit Immediate Floats in the SpurMemoryManager class comment.

	 Encode:				[1s][     11 exponent     ][52mantissa] 
	 rot sign:				[     11 exponent     ][52mantissa][1s] 
	 sub exponent offset:	[ 000 ][8expsubset][52 mantissa][1s] 
	 shift:					[8expsubset][52 mantissa][1s][ 000 ] 
	 or/add tags:			[8expsubset][52mantissa][1s][3tags]"
	<inline: true>
	<returnTypeC: #sqInt>
	<var: #aFloat type: #double>
	| rawFloat |
	<var: #rawFloat type: #usqLong>
	self assert: (self isSmallFloatValue: aFloat).
	self
		cCode: [self memcpy: (self addressOf: rawFloat) _: (self addressOf: aFloat) _: (self sizeof: rawFloat)]
		inSmalltalk: [rawFloat := (aFloat at: 1) << 32 + (aFloat at: 2)].
	^self smallFloatObjectOfBits: rawFloat! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 1/18/2020 18:09'!
smallFloatObjectOfBits: rawFloatBits
	"Encode the argument, rawFloatBits in the SmallFloat range, as a tagged small float.
	 See section 61-bit Immediate Floats in the SpurMemoryManager class comment.

	 Encode:				[1s][     11 exponent     ][52mantissa] 
	 rot sign:				[     11 exponent     ][52mantissa][1s] 
	 sub exponent offset:	[ 000 ][8expsubset][52 mantissa][1s] 
	 shift:					[8expsubset][52 mantissa][1s][ 000 ] 
	 or/add tags:			[8expsubset][52mantissa][1s][3tags]"
	<inline: #always>
	<returnTypeC: #sqInt>
	| rot |
	<var: #rawFloatBits type: #usqLong>
	<var: #rot type: #usqLong>
	self assert: (self isSmallFloatValueBits: rawFloatBits).
	rot := self rotateLeft: rawFloatBits.
	rot > 1 ifTrue: "a.k.a. ~= +/-0.0"
		[rot := rot - (self smallFloatExponentOffset << (self smallFloatMantissaBits + 1)).
		 self assert: rot > 0].
	^self cCode: [rot << self numTagBits + self smallFloatTag]
		inSmalltalk: [((rot << self numTagBits) bitAnd: 16rFFFFFFFFFFFFFFFF) + self smallFloatTag]! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 2/12/2015 09:35'!
smallFloatSignBit
	<inline: true>
	^1 << self numTagBits! !

!Spur64BitMemoryManager methodsFor: 'cog jit support' stamp: 'eem 3/9/2016 13:51'!
smallFloatTag
	<api>
	<cmacro>
	^4! !

!Spur64BitMemoryManager methodsFor: 'interpreter access' stamp: 'eem 9/19/2018 17:31'!
smallFloatValueOf: oop
	"Answer the C double precision floating point value of the argument, a SmallFloat.
	 See section 61-bit Immediate Floats in the SpurMemoryManager class comment.
							msb                                              lsb 
	 Decode:				[8expsubset][52mantissa][1s][3tags] 
	 shift away tags:		[ 000 ][8expsubset][52mantissa][1s] 
	 add exponent offset:	[     11 exponent     ][52mantissa][1s] 
	 rot sign:				[1s][     11 exponent     ][52mantissa]"
	| bits value |
	<returnTypeC: #double>
	<var: #value type: #double>
	<var: #bits type: #usqLong>
	bits := self smallFloatBitsOf: oop.
	self cCode:
			[self memcpy: (self addressOf: value) _: (self addressOf: bits) _: (self sizeof: value).
			^value]
		inSmalltalk:
			[^(Float new: 2)
				at: 1 put: bits >> 32;
				at: 2 put: (bits bitAnd: 16rFFFFFFFF);
				* 1.0 "reduce to SmallFloat64 if possible"]! !

!Spur64BitMemoryManager methodsFor: 'allocation' stamp: 'eem 12/12/2015 13:42'!
smallObjectBytesForSlots: numSlots
	"Answer the total number of bytes in an object without an overflow header, including header bytes."
	<api>
	<returnTypeC: #usqInt>
	^self baseHeaderSize "single header"
	+ (numSlots < 1
		ifTrue: [self allocationUnit] "at least one slot for the forwarding pointer"
		ifFalse: [numSlots * self bytesPerOop])! !

!Spur64BitMemoryManager methodsFor: 'word size' stamp: 'eem 12/12/2015 14:13'!
tagMask
	<api>
	<cmacro>
	^7! !

!Spur64BitMemoryManager methodsFor: 'indexing primitive support' stamp: 'eem 6/21/2021 14:50'!
updateFormatOfShortenedObject: objOop to: indexableSize
	"Update the odd bits of a word object, if required. It will need to be extended if
	 shortening of 16-bit or 8-bit pure bits objects is ever supported."
	((self formatOf: objOop) between: self firstLongFormat and: self firstLongFormat + 1) ifTrue:
		[self setFormatOf: objOop to: self firstLongFormat + (indexableSize bitAnd: 1)]! !

!Spur64BitMemoryManager methodsFor: 'header formats' stamp: 'eem 10/8/2013 12:21'!
wordIndexableFormat
	^self sixtyFourBitIndexableFormat! !

!Spur64BitMemoryManager methodsFor: 'word size' stamp: 'eem 12/12/2015 14:13'!
wordSize
	"Answer the manager's word size, which is the size of an oop, and which
	 is assumed to be equivalent to the underlying machine's word size."
	<cmacro>
	^8! !

!Spur64BitMemoryManager methodsFor: 'class table' stamp: 'eem 10/8/2013 12:17'!
wordSizeClassIndexPun
	^self sixtyFourBitLongsClassIndexPun! !

!SpurMemoryManager class methodsFor: 'translation' stamp: 'eem 6/7/2018 15:35'!
ancilliaryClasses
	"Answer any extra classes to be included in the translation."
	^{	SpurGenerationScavenger. SpurScavengeLogRecord. SpurSegmentManager. SpurSegmentInfo }, 
		self compactorClass classesForTranslation,
		SpurNewSpaceSpace withAllSubclasses
		
	! !

!SpurMemoryManager class methodsFor: 'accessing' stamp: 'eem 10/1/2013 11:22'!
baseHeaderSize
	"For CogBlockMethod class>>instVarNamesAndTypesForTranslationDo:"
	^8! !

!SpurMemoryManager class methodsFor: 'word size' stamp: 'eem 4/2/2015 16:54'!
bytesPerOop
	"Answer the size of an oop for this manager, which is assumed
	 to be equivalent to the underlying machine's word size."
	^self wordSize! !

!SpurMemoryManager class methodsFor: 'simulation only' stamp: 'eem 10/30/2013 13:56'!
characterObjectOf: characterCode
	^characterCode asCharacter! !

!SpurMemoryManager class methodsFor: 'accessing' stamp: 'eem 11/6/2013 15:08'!
classIndicesMap
	"self classIndicesMap"
	| map |
	map := Dictionary new.
	self selectorsAndMethodsDo:
		[:s :m| | mn |
		(('*classindex*' match: s) or: ['*bridge*'match: s])
		and: [mn := m methodNode block.
			(mn statements size = 1
			 and: [mn statements first expr isConstantNumber]) ifTrue:
				[map at: mn statements first expr key put: s]]].
	^map keys sort collect:
		[:n|
		{ n. map at: n }]! !

!SpurMemoryManager class methodsFor: 'accessing' stamp: 'eem 10/15/2013 15:01'!
classTableBitmapBytes
	"Max size of the classTableBitmap.  A liottle too large to contemplate allocating statically."
	^1 << (self basicNew classIndexFieldWidth - (BitsPerByte log: 2) asInteger)! !

!SpurMemoryManager class methodsFor: 'accessing class hierarchy' stamp: 'eem 6/7/2018 15:20'!
compactorClass
	"Answer the compaction algorithm to use."
	^Smalltalk classNamed: (InitializationOptions at: #compactorClass ifAbsent: [#SpurPlanningCompactor])! !

!SpurMemoryManager class methodsFor: 'translation' stamp: 'eem 8/9/2021 12:43'!
declareCVarsIn: aCCodeGenerator
	self class == thisContext methodClass ifFalse: [^self]. "Don't duplicate decls in subclasses"
	aCCodeGenerator removeVariable: 'memory'. "memory is a simulation time thing only"
	self declareCAsOop: #(	freeStart scavengeThreshold newSpaceStart newSpaceLimit pastSpaceStart
							lowSpaceThreshold freeOldSpaceStart oldSpaceStart endOfMemory)
		in: aCCodeGenerator.
	self declareCAsUSqLong: (self allInstVarNames select: [:ivn| ivn endsWith: 'Usecs']), #(statAllocatedBytes)
		in: aCCodeGenerator.
	aCCodeGenerator
		var: #lastHash type: #usqInt;
		var: #freeListsMask type: #usqInt;
		var: #freeLists type: #'sqInt *';
		var: #objStackInvalidBecause type: #'char *';
		var: #unscannedEphemerons type: #SpurContiguousObjStack;
		var: #heapGrowthToSizeGCRatio type: #float;
		var: #heapSizeAtPreviousGC type: #usqInt;
		var: #totalFreeOldSpace type: #usqInt;
		var: #maxOldSpaceSize type: #usqInt.
	aCCodeGenerator
		var: #oldSpaceUsePriorToScavenge type: #sqLong.
	aCCodeGenerator
		var: #remapBuffer
		declareC: 'sqInt remapBuffer[RemapBufferSize + 1 /* ', (RemapBufferSize + 1) printString, ' */]'.
	aCCodeGenerator
		var: #extraRoots
		declareC: 'sqInt *extraRoots[ExtraRootsSize + 1 /* ', (ExtraRootsSize + 1) printString, ' */]'! !

!SpurMemoryManager class methodsFor: 'api characterization' stamp: 'eem 1/6/2014 13:33'!
hasSpurMemoryManagerAPI
	^true! !

!SpurMemoryManager class methodsFor: 'analysis' stamp: 'eem 7/7/2017 16:59'!
identify32BitSignedComparisonsInSegmentIO
	"self identify32BitSignedComparisonsInSegmentIO"
	self identifySignedComparisonsFor: #(ObjectMemory Spur32BitMemoryManager)
		in:	(self organization listAtCategoryNamed: #'image segment in/out')
		noise: #(	'classIndex* >= numOutPointers'
					'count > ptr - start / self bytesPerOop'
					'endSeg - segAddr < self baseHeaderSize + self bytesPerOop'
					'errorCode* > 0'
					'GCModeImageSegment > 0'
					'hash - TopHashBit <= outIndex'
					'limit - ptr <= 8'
					'num* >= self numSlotsMask'
					'num* <= 1'
					'outIndex >= (self numSlotsOf: outPointerArray)'
					'outIndex := self mapOopsFrom: * < 0'
					'segAddr - segStart / 8 + self lastClassIndexPun >= TopHashBit'
					'there > 0'
					'* > self identityHashHalfWordMask'
					'*segmentLimit >= self numSlotsMask*'
					'* > self isForwardedObjectClassIndexPun'
					'* > self lastClassIndexPun')! !

!SpurMemoryManager class methodsFor: 'analysis' stamp: 'eem 7/7/2017 17:02'!
identify64BitSignedComparisonsInSegmentIO
	"self identify64BitSignedComparisonsInSegmentIO"
	self identifySignedComparisonsFor: #(ObjectMemory Spur64BitMemoryManager)
		in:	(self organization listAtCategoryNamed: #'image segment in/out')
		noise: #(	'classIndex* >= numOutPointers'
					'count > ptr - start / self bytesPerOop'
					'endSeg - segAddr < self baseHeaderSize + self bytesPerOop'
					'errorCode* > 0'
					'GCModeImageSegment > 0'
					'hash - TopHashBit <= outIndex'
					'limit - ptr <= 8'
					'num* >= self numSlotsMask'
					'num* <= 1'
					'num* > 0'
					'num* < 1'
					'outIndex >= (self numSlotsOf: outPointerArray)'
					'outIndex := self mapOopsFrom: * < 0'
					'segAddr - segStart / 8 + self lastClassIndexPun >= TopHashBit'
					'there > 0'
					'* > self identityHashHalfWordMask'
					'*segmentLimit >= self numSlotsMask*'
					'* > self isForwardedObjectClassIndexPun'
					'* > self lastClassIndexPun')! !

!SpurMemoryManager class methodsFor: 'analysis' stamp: 'eem 7/7/2017 17:01'!
identifySignedComparisonsFor: options in: selectors noise: noise
	"self identify32BitSignedComparisonsInSegmentIO"
	"self identify64BitSignedComparisonsInSegmentIO"
	| vmm cg halt |
	halt := false.
	vmm := (VMMaker forPlatform: 'Cross')
				interpreterClass: StackInterpreter;
				options: options.
	cg := [vmm buildCodeGeneratorForInterpreter]
			on: Notification
			do: [:ex|
				ex tag == #getVMMaker
					ifTrue: [ex resume: vmm]
					ifFalse: [ex pass]].
	cg vmClass preGenerationHook: cg.
	cg inferTypesForImplicitlyTypedVariablesAndMethods.
	cg retainMethods: self selectors.
	cg prepareMethods.
	cg doInlining: true.
	selectors sort do:
		[:sel|
		(cg methodNamed: sel) ifNotNil:
			[:m|
			m parseTree nodesDo:
				[:node|
				(node isSend
				 and: [(#(< > <= >=) includes: node selector)
				 and: [({node receiver. node args first } anySatisfy:
						[:o| (cg typeFor: o in: m)
								ifNil: [true]
								ifNotNil: [:t| (cg isIntegralCType: t) and: [t first ~= $u]]])
				 and: [noise noneSatisfy: [:n| n match: node printString]]]]) ifTrue:
					[halt ifTrue: [self halt: node printString].
					 Transcript ensureCr; nextPutAll: sel; space; print: node; flush]]]]! !

!SpurMemoryManager class methodsFor: 'translation' stamp: 'eem 10/29/2013 23:43'!
implicitReturnTypeFor: aSelector
	"Answer the return type for methods that don't have an explicit return."
	^#void! !

!SpurMemoryManager class methodsFor: 'class initialization' stamp: 'eem 4/6/2016 17:28'!
initBytesPerWord: wordSize

	BytesPerWord := BytesPerOop := wordSize.
	"N.B.  This is *not* output when generating the interpreter file.
	 It is left to the various sqConfig.h files to define correctly."
	VMBIGENDIAN := Smalltalk endianness == #big.
	SPURVM := true! !

!SpurMemoryManager class methodsFor: 'class initialization' stamp: 'eem 3/9/2020 19:48'!
initialize
	"SpurMemoryManager initialize"
	BitsPerByte := 8.

	"Initialize at least the become constants for the Spur bootstrap where the
	 old ObjectMemory simulator is used before a Spur simulator is created.."
	self initializeSpurObjectRepresentationConstants.

	"An obj stack is a stack of objects stored in a hidden root slot, such as
	 the markStack or the ephemeronQueue.  It is a linked list of segments,
	 with the hot end at the head of the list.  It is a word object.  The stack
	 pointer is in ObjStackTopx and 0 means empty.  The list goes through
	 ObjStackNextx. We don't want to shrink objStacks, since they're used
	 in GC and it's good to keep their memory around.  So unused pages
	 created by popping emptied pages are kept on the ObjStackFreex list.
	 ObjStackNextx must be the last field for swizzleObjStackAt:."
	ObjStackPageSlots := 4092. "+ double header = 16k bytes per page in 32-bits, 32k bytes per page in 64 bits"
	ObjStackTopx := 0.
	ObjStackMyx := 1.
	ObjStackFreex := 2.
	ObjStackNextx := 3.
	ObjStackFixedSlots := 4.
	ObjStackLimit := ObjStackPageSlots - ObjStackFixedSlots.
	"The hiddenRootsObject contains the classTable pages and up to 8 additional objects.
	 Currently we use four; the three objStacks (the mark stack, the weaklings and the
	 mourn queue), and the rememberedSet."
	MarkStackRootIndex := self basicNew classTableRootSlots.
	WeaklingStackRootIndex := MarkStackRootIndex + 1.
	MournQueueRootIndex := MarkStackRootIndex + 2.
	RememberedSetRootIndex := MarkStackRootIndex + 3.

	MarkObjectsForEnumerationPrimitives := false.

	"The remap buffer support is for compatibility; Spur doesn't GC during allocation.
	 Eventually this should die."
	RemapBufferSize := 25.

	"Extra roots are for plugin support."
	ExtraRootsSize := 2048. "max. # of external roots"

	"gcPhaseInProgress takes these values to identify phases as required."
	ScavengeInProgress := 1.
	SlidingCompactionInProgress := 2! !

!SpurMemoryManager class methodsFor: 'class initialization' stamp: 'eem 1/22/2021 21:07'!
initializeCompactClassIndices
	"Reuse the compact class indices to name known classIndices.
	 This helps reduce the churn in the interpreters."
	"c.f. SpurBootstrap>>defineKnownClassIndices"
	FirstValidClassIndex :=
	ClassLargeNegativeIntegerCompactIndex := 32.
	ClassLargePositiveIntegerCompactIndex := 33.
	ClassFloatCompactIndex := 34.

	ClassMessageCompactIndex := 35.
	ClassMethodContextCompactIndex := 36.
	ClassBlockContextCompactIndex := 0.
	ClassBlockClosureCompactIndex := 37.
	ClassFullBlockClosureCompactIndex := 38.

	ClassByteArrayCompactIndex := 50.
	ClassArrayCompactIndex := 51.
	ClassByteStringCompactIndex := 52.
	ClassBitmapCompactIndex := 53.
	ClassPointCompactIndex := 54! !

!SpurMemoryManager class methodsFor: 'class initialization' stamp: 'eem 2/11/2015 11:08'!
initializeObjectHeaderConstants

	BytesPerWord ifNil: [BytesPerWord := 4].  "May get called on fileIn, so supply default"
	BaseHeaderSize := 8. "This is still needed for VM generation."

	"These are used in image segments"
	self ~~ SpurMemoryManager ifTrue:
		[TopHashBit := 1 << (self basicNew identityHashFieldWidth - 1).
		 TopOopBit := 1 << (self basicNew bytesPerOop * 8 - 1)]! !

!SpurMemoryManager class methodsFor: 'class initialization' stamp: 'eem 1/16/2020 19:54'!
initializeSpecialObjectIndices
	"Initialize indices into specialObjects array."

	NilObject := 0.
	FalseObject := 1.
	TrueObject := 2.
	SchedulerAssociation := 3.
	ClassBitmap := 4.
	ClassSmallInteger := 5.
	ClassByteString := ClassString := 6. "N.B.  Actually class ByteString"
	ClassArray := 7.
	"SmalltalkDictionary := 8."  "Do not delete!!"
	ClassFloat := 9.
	ClassMethodContext := 10.
	"ClassBlockContext := 11. unused by the VM"
	ClassPoint := 12.
	ClassLargePositiveInteger := 13.
	TheDisplay := 14.
	ClassMessage := 15.
	"ClassCompiledMethod := 16. unused by the VM"
	TheLowSpaceSemaphore := 17.
	ClassSemaphore := 18.
	ClassCharacter := 19.
	SelectorDoesNotUnderstand := 20.
	SelectorCannotReturn := 21.
	ProcessSignalingLowSpace := 22.	"was TheInputSemaphore"
	SpecialSelectors := 23.
	CharacterTable := nil.	"Must be unused by the VM"
	SelectorMustBeBoolean := 25.
	ClassByteArray := 26.
	"ClassProcess := 27. unused"
	CompactClasses := 28.
	TheTimerSemaphore := 29.
	TheInterruptSemaphore := 30.
	ClassDoubleByteArray := 31.	"was the prototype Float"
	ClassWordArray := 32.			"was the prototype 4-byte LargePositiveInteger"
	ClassDoubleWordArray := 33.	"was the prototype Point"
		ValidatedClassDoubleByteArrayFlag := 1.
		ValidatedClassWordArrayFlag := 2.
		ValidatedClassDoubleWordArrayFlag := 4.
		ValidatedClassFloat32ArrayFlag := 8.
		ValidatedClassFloat64ArrayFlag := 16.
	SelectorCannotInterpret := 34.
	"Was MethodContextProto := 35."
	ClassBlockClosure := 36.
	"Was BlockContextProto := 37."
	ExternalObjectsArray := 38.
	ClassMutex := 39.
	"Was: ClassTranslatedMethod := 40."
	ProcessInExternalCodeTag := 40.
	TheFinalizationSemaphore := 41.
	ClassLargeNegativeInteger := 42.

	ClassExternalAddress := 43.
	ClassExternalStructure := 44.
	ClassExternalData := 45.
	ClassExternalFunction := 46.
	ClassExternalLibrary := 47.

	SelectorAboutToReturn := 48.
	SelectorRunWithIn := 49.

	SelectorAttemptToAssign := 50.
	"PrimErrTableIndex := 51. in VMClass class>>initializePrimitiveErrorCodes"
	ClassAlien := 52.
	SelectorInvokeCallback := 53.
	ClassUnsafeAlien := 54.

	ClassWeakFinalizer := 55.

	ForeignCallbackProcess := 56.

	SelectorUnknownBytecode := 57.
	SelectorCounterTripped := 58.
	SelectorSistaTrap := 59.
	
	LowcodeContextMark := 60.
	LowcodeNativeContextClass := 61.! !

!SpurMemoryManager class methodsFor: 'class initialization' stamp: 'eem 9/23/2014 12:32'!
initializeSpurObjectRepresentationConstants
	"SpurMemoryManager initializeSpurObjectRepresentationConstants"
	BecamePointerObjectFlag := 1.
	BecameCompiledMethodFlag := 2.
	OldBecameNewFlag := 4.
	BecameActiveClassFlag := 8 "For flushing method caches"! !

!SpurMemoryManager class methodsFor: 'class initialization' stamp: 'eem 6/7/2018 15:06'!
initializeWithOptions: optionsDictionary
	"SpurMemoryManager initializeWithOptions: Dictionary new"

	super initializeWithOptions: optionsDictionary.
	InitializationOptions
		at: #Spur32BitMemoryManager ifAbsentPut: false;
		at: #Spur64BitMemoryManager ifAbsentPut: false.
	self initialize.
	self initBytesPerWord: (self == SpurMemoryManager
								ifTrue: [optionsDictionary at: #BytesPerWord ifAbsent: [4]]
								ifFalse: [self wordSize]).

	self initializeObjectHeaderConstants. "Initializes BaseHeaderSize so do early"
	self initializeSpurObjectRepresentationConstants.
	self initializeSpecialObjectIndices.
	self initializeCompactClassIndices.
	self initializePrimitiveErrorCodes.

	SpurGenerationScavenger initialize! !

!SpurMemoryManager class methodsFor: 'simulation only' stamp: 'eem 10/30/2013 13:50'!
isImmediate: anObject
	self subclassResponsibility! !

!SpurMemoryManager class methodsFor: 'translation' stamp: 'eem 6/6/2018 20:34'!
isNonArgumentImplicitReceiverVariableName: aString
	^#('self' 'coInterpreter' 'manager' 'scavenger' 'segmentManager' 'compactor' 'planningCompactor' 'selectiveCompactor' 'heapMap') includes: aString! !

!SpurMemoryManager class methodsFor: 'translation' stamp: 'eem 5/17/2015 16:40'!
isSameLevelObjectAccessor: selector
	"For accessor depth calculation, answer if selector doesn't traverse into an object, merely deriving a pointer from it."
	^#(arrayValueOf: firstFixedField: firstIndexableField:) includes: selector! !

!SpurMemoryManager class methodsFor: 'translation' stamp: 'eem 5/17/2015 17:07'!
isTerminalObjectAccessor: selector
	"For accessor depth calculation, answer if selector doesn't answer another object object; merely a value."
	^#(byteSizeOf: fetchFloat:ofObject: fetchInteger:ofObject: fetchLong32:ofObject: instanceSizeOf: slotSizeOf: stSizeOf:) includes: selector! !

!SpurMemoryManager class methodsFor: 'accessing' stamp: 'eem 11/15/2019 16:14'!
memoryManagerName
	^ 'Spur'! !

!SpurMemoryManager class methodsFor: 'translation' stamp: 'eem 6/29/2015 15:22'!
mustBeGlobal: var
	"Answer if a variable must be global and exported.  Used for inst vars that are accessed from VM support code."

	^#('checkForLeaks' 'maxOldSpaceSize') includes: var! !

!SpurMemoryManager class methodsFor: 'accessing' stamp: 'eem 11/18/2014 17:48'!
numSmallIntegerTagBits
	^self subclassResponsibility! !

!SpurMemoryManager class methodsFor: 'accessing' stamp: 'eem 11/6/2013 14:07'!
objectFormatsMap
	"self objectFormatsMap"
	| map |
	map := Dictionary new.
	self selectorsAndMethodsDo:
		[:s :m| | mn |
		('*format*' match: s)
		and: [mn := m methodNode block.
			(mn statements size = 1
			 and: [mn statements first expr isConstantNumber]) ifTrue:
				[map at: mn statements first expr key put: s]]].
	^map keys sort collect:
		[:n|
		{ n. map at: n }]! !

!SpurMemoryManager class methodsFor: 'accessing class hierarchy' stamp: 'eem 9/30/2013 14:12'!
objectRepresentationClass
	^self subclassResponsibility! !

!SpurMemoryManager class methodsFor: 'simulation only' stamp: 'eem 11/18/2016 17:07'!
simulatorClass
	^self subclassResponsibility! !

!SpurMemoryManager class methodsFor: 'api characterization' stamp: 'nice 5/1/2020 23:17'!
vmProxyMinorVersion
	"Define the  VM_PROXY_MINOR version for this VM as used to define
	 the api in platforms/Cross/vm/sqVirtualMachine.[ch] and in interp.h."
	^17 "isShorts isLong64s"! !

!SpurMemoryManager class methodsFor: 'word size' stamp: 'eem 11/13/2014 15:24'!
wordSize
	"Answer the manager's word size, which is the size of an oop, and which
	 is assumed to be equivalent to the underlying machine's word size."
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 12/21/2017 15:00'!
ISA
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter ISA! !

!SpurMemoryManager methodsFor: 'simulation tests support' stamp: 'eem 06/22/2021 15:31'!
abandonEmptySegmentForTests
	"Assume a freshly-loaded image. Eliminate the last segment."
	<doNotGenerate>
	| freeChunk emptySeg |
	freeChunk := self findLargestFreeChunk.
	self assert: totalFreeOldSpace = (self bytesInBody: freeChunk).
	self assert: endOfMemory = (self addressAfter: freeChunk).
	self unlinkSolitaryFreeTreeNode: freeChunk.
	segmentManager numSegments > 1
		ifTrue:
			[emptySeg := segmentManager findEmptySegNearestInSizeTo: (self bytesInBody: freeChunk).
			 segmentManager removeSegment: emptySeg]
		ifFalse:
			[(segmentManager segments at: 0)
				segSize: (segmentManager segments at: 0) segSize - (self bytesInBody: freeChunk).
			 self setLastSegment: (segmentManager segments at: 0);
			 	initSegmentBridgeWithBytes: self bridgeSize at: (self startOfObject: freeChunk)]! !

!SpurMemoryManager methodsFor: 'compaction - analysis' stamp: 'eem 06/22/2021 15:31'!
abstractCompaction
	"This method answers a rough estimate of compactibility."
	<doNotGenerate>
	| lowestFree freeChunks used movable |
	lowestFree := SmallInteger maxVal.
	freeChunks := Set new.
	used := Set new.
	movable := Set new.
	self allObjectsInFreeTreeDo:
		[:f|
		(self addressAfter: f) < endOfMemory ifTrue:
			[freeChunks add: f.
			 f < lowestFree ifTrue: [lowestFree := f]]].
	self allOldSpaceObjectsFrom: lowestFree do:
		[:o| | size delta best |
		size := self bytesInBody: o.
		delta := SmallInteger maxVal.
		freeChunks do: [:f| | fs |
			((fs := self bytesInBody: f) >= size) ifTrue:
				[delta > (fs - size) ifTrue:
					[delta := fs - size. best := f]]].
		 best ifNotNil:
			[movable add: o.
			 used add: (freeChunks remove: best)]].
	^{ totalFreeOldSpace. movable inject: 0 into: [:s :o| s + (self bytesInBody: o)]. used inject: 0 into: [:s :o| s + (self bytesInBody: o)] }! !

!SpurMemoryManager methodsFor: 'plugin support-private' stamp: 'eem 1/18/2020 11:51'!
accessIntegerClass: splArrayIdx withValidationFlag: validationFlag
	"Answer an integer array class from somewhere in the specialObjectsArray, if it appears
	 to be a class, otherwise nilObj.  Make delivering the answer fast since this is used for
	 allocations by using a flag to indicate that validation has already been done.  The
	 validation flags are cleared if the specialObjectsArray is becomed or scavenged."
	<inline: #always>
	| classOop hash |
	classOop := self splObj: splArrayIdx.
	(validatedIntegerClassFlags anyMask: validationFlag) ifTrue:
		[^classOop].
	((self isImmediate: classOop)
	or: [(hash := self rawHashBitsOf: classOop) = 0
	or: [(self classAtIndex: hash) ~= classOop]]) ifTrue:
		[^nilObj].
	validatedIntegerClassFlags := validatedIntegerClassFlags bitOr: validationFlag.
	^classOop! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 05/07/2020 16:13'!
accessibleObjectAfter: objOop
	"Answer the accessible object following the given object or 
	free chunk in the heap. Return nil when heap is exhausted.
	 This is for primitiveNextObject subsequent to primitiveSomeObject.
	 It also tries to handle more general use by ordering objects as
		eden
		past
		old
	 but this is tricky becaus ethe order in memory is
		past
		eden
		old"
	<inline: false>
	| objAfter |
	objAfter := objOop.
	(self oop: objAfter isLessThan: nilObj) ifTrue: "object in new space"
		[self assert: ((self isInEden: objOop) or: [self isInPastSpace: objOop]).
		 (self oop: objAfter isGreaterThan: pastSpaceStart) ifTrue:
			["Obj is in eden.  Answer next normal object in eden, if there is one."
			 [objAfter := self objectAfter: objAfter limit: freeStart.
			  self oop: objAfter isLessThan: freeStart] whileTrue:
				[(self isNormalObject: objAfter) ifTrue:
					[^objAfter]].
			 "There wasn't a next object in eden. If past space is empty answer nilObj."
			 pastSpaceStart <= scavenger pastSpace start ifTrue:
				[^nilObj].
			 "If the first object in pastSpace is OK, answer it, otherwise fall through to enumerate past space."
			 objAfter := self objectStartingAt: scavenger pastSpace start.
			 (self isNormalObject: objAfter) ifTrue:
				[^objAfter]].
		 "Either objOop was in pastSpace, or enumeration exhaused eden, so enumerate past space."
		 [objAfter := self objectAfter: objAfter limit: pastSpaceStart.
		  self oop: objAfter isLessThan: pastSpaceStart] whileTrue:
			[(self isNormalObject: objAfter) ifTrue:
				[^objAfter]].
		 ^nilObj].
	[objAfter := self objectAfter: objAfter limit: endOfMemory.
	 objAfter = endOfMemory ifTrue:
		[^nil].
	 (self isNormalObject: objAfter) ifTrue:
		[^objAfter]] repeat! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 7/20/2014 22:52'!
activeAndDeferredScan: anEphemeron
	"Answer whether an ephemeron is active (has an unmarked
	 key) and was pushed on the unscanned ephemerons stack."
	| key |
	<inline: #never>
	self assert: (self isEphemeron: anEphemeron).
	((self isImmediate: (key := self keyOfEphemeron: anEphemeron))
	 or: [self isMarked: key]) ifTrue:
		[^false].
	^self pushOnUnscannedEphemeronsStack: anEphemeron! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'cb 6/12/2018 07:20'!
addFreeChunkWithBytes: bytes at: address
	totalFreeOldSpace := totalFreeOldSpace + bytes.
	^self freeChunkWithBytes: bytes at: address! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
addFreeSubTree: freeTree
	"Add a freeChunk sub tree back into the large free chunk tree.
	 This is for allocateOldSpaceChunkOf[Exactly]Bytes:[suchThat:]."
	| bytesInArg treeNode bytesInNode subNode |
	"N.B. *can't* use numSlotsOfAny: because of rounding up of odd slots
	 and/or step in size at 1032 bytes in 32-bits or 2048 bytes in 64-bits."
	self assert: (self isFreeObject: freeTree).
	bytesInArg := self bytesInBody: freeTree.
	self assert: bytesInArg >= (self numFreeLists * self allocationUnit).
	treeNode := freeLists at: 0.
	self assert: treeNode ~= 0.
	[bytesInNode := self bytesInBody: treeNode.
	 "check for overlap; could write this as self oop: (self objectAfter: freeChunk) isLessThanOrEqualTo: child...
	  but that relies on headers being correct, etc.  So keep it clumsy..."
	 self assert: ((self oop: freeTree + bytesInArg - self baseHeaderSize isLessThanOrEqualTo: treeNode)
					or: [self oop: freeTree isGreaterThanOrEqualTo: treeNode + bytesInNode - self baseHeaderSize]).
	 self assert: bytesInNode >= (self numFreeLists * self allocationUnit).
	 self assert: bytesInArg ~= bytesInNode.
	 bytesInNode > bytesInArg
		ifTrue:
			[subNode := self fetchPointer: self freeChunkSmallerIndex ofFreeChunk: treeNode.
			 subNode = 0 ifTrue:
				[self storePointer: self freeChunkSmallerIndex ofFreeChunk: treeNode withValue: freeTree.
				 self storePointer: self freeChunkParentIndex ofFreeChunk: freeTree withValue: treeNode.
				 ^self]]
		ifFalse:
			[subNode := self fetchPointer: self freeChunkLargerIndex ofFreeChunk: treeNode.
			 subNode = 0 ifTrue:
				[self storePointer: self freeChunkLargerIndex ofFreeChunk: treeNode withValue: freeTree.
				 self storePointer: self freeChunkParentIndex ofFreeChunk: freeTree withValue: treeNode.
				 ^self]].
	 treeNode := subNode] repeat! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 10/15/2013 09:53'!
addGCRoot: varLoc
	"Add the given variable location to the extra roots table."
	<api>
	<var: #varLoc type: #'sqInt *'>
	extraRootCount >= ExtraRootsSize ifTrue: [^false]. "out of space"
	extraRoots at: (extraRootCount := extraRootCount + 1) put: varLoc.
	^true! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
addToFreeList: freeChunk bytes: chunkBytes
	"Add freeChunk to the relevant freeList.
	 For the benefit of sortedFreeObject:, if freeChunk is large, answer the treeNode it
	 is added to, if it is added to the next list of a freeTreeNode, otherwise answer 0."
	| index |
	"coInterpreter transcript ensureCr. coInterpreter print: 'freeing '. self printFreeChunk: freeChunk."
	self assert: (self isFreeObject: freeChunk).
	self assert: chunkBytes = (self bytesInBody: freeChunk).
	"Too slow to be enabled byt default but useful to debug Selective...
	 self deny: (compactor isSegmentBeingCompacted: (segmentManager segmentContainingObj: freeChunk))."
	index := chunkBytes / self allocationUnit.
	index < self numFreeLists ifTrue:
		[self setNextFreeChunkOf: freeChunk withValue: (freeLists at: index) chunkBytes: chunkBytes.
		(self isLilliputianSize: chunkBytes) ifFalse:
			[self storePointer: self freeChunkPrevIndex ofFreeChunk: freeChunk withValue: 0].
		 freeLists at: index put: freeChunk.
		 freeListsMask := freeListsMask bitOr: 1 << index.
		 ^0].

	^self addToFreeTree: freeChunk bytes: chunkBytes! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
addToFreeTree: freeChunk bytes: chunkBytes
	"Add freeChunk to the large free chunk tree.
	 For the benefit of sortedFreeObject:, answer the treeNode it is added
	 to, if it is added to the next list of a freeTreeNode, otherwise answer 0."
	| childBytes parent child |
	self assert: (self isFreeObject: freeChunk).
	self assert: chunkBytes = (self bytesInBody: freeChunk).
	self assert: chunkBytes >= (self numFreeLists * self allocationUnit).
	self
		storePointer: self freeChunkNextIndex ofFreeChunk: freeChunk withValue: 0;
		storePointer: self freeChunkPrevIndex ofFreeChunk: freeChunk withValue: 0;
		storePointer: self freeChunkParentIndex ofFreeChunk: freeChunk withValue: 0;
		storePointer: self freeChunkSmallerIndex ofFreeChunk: freeChunk withValue: 0;
		storePointer: self freeChunkLargerIndex ofFreeChunk: freeChunk withValue: 0.
	"Large chunk list organized as a tree, each node of which is a list of chunks of the same size.
	 Beneath the node are smaller and larger blocks."
	parent := 0.
	child := freeLists at: 0.
	[child ~= 0] whileTrue:
		[childBytes := self bytesInBody: child.
		 "check for overlap; could write this as self oop: (self objectAfter: freeChunk) isLessThanOrEqualTo: child...
		  but that relies on headers being correct, etc.  So keep it clumsy..."
		 self assert: ((self oop: freeChunk + chunkBytes - self baseHeaderSize isLessThanOrEqualTo: child)
						or: [self oop: freeChunk isGreaterThanOrEqualTo: child + childBytes - self baseHeaderSize]).
		 childBytes = chunkBytes ifTrue: "size match; add to list at node."
			[self setNextFreeChunkOf: freeChunk withValue: (self fetchPointer: self freeChunkNextIndex ofFreeChunk: child) isLilliputianSize: false. 
			 self setNextFreeChunkOf: child withValue: freeChunk isLilliputianSize: false.
			 ^child].
		 "walk down the tree"
		 parent := child.
		 child := self fetchPointer: (childBytes > chunkBytes
										ifTrue: [self freeChunkSmallerIndex]
										ifFalse: [self freeChunkLargerIndex])
					ofFreeChunk: child].
	parent = 0 ifTrue:
		[self assert: (freeLists at: 0) = 0.
		 freeLists at: 0 put: freeChunk.
		 freeListsMask := freeListsMask bitOr: 1.
		 ^0].
	self assert: (freeListsMask anyMask: 1).
	"insert in tree"
	self storePointer: self freeChunkParentIndex
			ofFreeChunk: freeChunk
				withValue: parent.
	self storePointer: (childBytes > chunkBytes
									ifTrue: [self freeChunkSmallerIndex]
									ifFalse: [self freeChunkLargerIndex])
			ofFreeChunk: parent
				withValue: freeChunk.
	^0! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 9/2/2013 13:07'!
addressAfter: objOop
	"Answer the address immediately following an object."
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 12/24/2016 13:55'!
addressCouldBeObj: address
	<api>
	<inline: false>
	^(address bitAnd: self baseHeaderSize - 1) = 0
	  and: [(self isInOldSpace: address)
		or: [(self isInEden: address)
		or: [(self isInPastSpace: address)
		or: [self scavengeInProgress and: [self isInFutureSpace: address]]]]]! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 6/12/2014 14:56'!
addressCouldBeOldObj: address
	^(address bitAnd: self baseHeaderSize - 1) = 0
	  and: [self isInOldSpace: address]! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 10/14/2014 12:23'!
addressCouldBeOop: address
	<api>
	"Answer if address appears to be that of either an immediate or an object.
	 For code disassembly and assertions."
	^(self isImmediate: address)
	  or: [self addressCouldBeObj: address]! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 1/26/2017 14:05'!
adjustAllOopsBy: bytesToShift
	"Adjust all oop references by the given number of bytes. This is
	 done just after reading in an image when the new base address
	 of the object heap is different from the base address in the image,
	 or when loading multiple segments that have been coalesced.  Also
	 set bits in the classTableBitmap corresponding to used classes."

	| obj classIndex |
	self assert: self newSpaceIsEmpty.
	self countNumClassPagesPreSwizzle: bytesToShift.
	(bytesToShift ~= 0
	 or: [segmentManager numSegments > 1]) ifTrue:
		[obj := self objectStartingAt: oldSpaceStart.
		 [self oop: obj isLessThan: freeOldSpaceStart] whileTrue:
			[classIndex := self classIndexOf: obj.
			 classIndex >= self isForwardedObjectClassIndexPun
				ifTrue:
					[self swizzleFieldsOfObject: obj]
				ifFalse:
					[classIndex = self isFreeObjectClassIndexPun ifTrue:
						[self swizzleFieldsOfFreeChunk: obj]].
			 obj := self objectAfter: obj limit: endOfMemory]]! !

!SpurMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 6/21/2021 12:45'!
allEntitiesFrom: initialObject do: aBlock
	<inline: true>
	| prevObj prevPrevObj obj |
	prevPrevObj := prevObj := nil.
	obj := initialObject.
	self enableObjectEnumerationFrom: initialObject.
	[self assert: obj \\ self allocationUnit = 0.
	 self oop: obj isLessThan: endOfMemory] whileTrue:
		[self assert: (self long64At: obj) ~= 0.
		 aBlock value: obj.
		 prevPrevObj := prevObj.
		 prevObj := obj.
		 obj := self objectAfterMaybeSlimBridge: obj limit: endOfMemory.
		 self assert: (self oop: obj isGreaterThan: prevObj)].
	self touch: prevPrevObj.
	self touch: prevObj! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 10/7/2013 09:00'!
allFreeObjects
	<doNotGenerate>
	| freeObjects |
	freeObjects := OrderedCollection new.
	self allFreeObjectsDo:
		[:f| freeObjects addLast: f].
	^freeObjects! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 10/18/2013 17:22'!
allFreeObjectsDo: aBlock
	| obj |
	1 to: self numFreeLists - 1 do:
		[:i|
		obj := freeLists at: i.
		[obj ~= 0] whileTrue:
			[aBlock value: obj.
			 obj := self fetchPointer: self freeChunkNextIndex ofFreeChunk: obj]].
	self allObjectsInFreeTreeDo: aBlock! !

!SpurMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 2/3/2014 16:35'!
allHeapEntitiesDo: aBlock
	"N.B. e.g. allObjects relies on the old/new order here."
	<inline: true>
	self allOldSpaceEntitiesDo: aBlock.
	self allNewSpaceEntitiesDo: aBlock! !

!SpurMemoryManager methodsFor: 'primitive support' stamp: 'eem 06/22/2021 15:31'!
allInstancesOf: aClass
	"Attempt to answer an array of all objects, excluding those that may
	 be garbage collected as a side effect of allocating the result array.
	 If no memory is available answer the number of instances as a SmallInteger.
	 Since objects are at least 16 bytes big, and the largest SmallInteger covers
	 1/4 of the address space, the count can never overflow."
	| classIndex freeChunk ptr start limit count bytes |
	classIndex := self rawHashBitsOf: aClass.
	classIndex = 0 ifTrue:
		[freeChunk := self allocateSlots: 0 format: self arrayFormat classIndex: ClassArrayCompactIndex.
		 ^freeChunk].
	MarkObjectsForEnumerationPrimitives ifTrue:
		[self markObjects: true]. "may not want to revive objects unnecessarily; but marking is sloooow."
	freeChunk := self allocateLargestFreeChunk. "N.B. Does /not/ update totalFreeOldSpace"
	start := freeChunk + self baseHeaderSize.
	limit := self addressAfter: freeChunk.
	(self isClassAtUniqueIndex: aClass)
		ifTrue:
			[self uniqueIndex: classIndex allInstancesInto: start limit: limit resultsInto: [:c :p| count := c. ptr := p]]
		ifFalse:
			[self ambiguousClass: aClass allInstancesInto: start limit: limit resultsInto: [:c :p| count := c. ptr := p]].
	self assert: (self isEmptyObjStack: markStack).
	MarkObjectsForEnumerationPrimitives
		ifTrue:
			[self assert: self allObjectsUnmarked.
			 self emptyObjStack: weaklingStack]
		ifFalse:
			[self assert: (self isEmptyObjStack: weaklingStack)].
	(count > (ptr - start / self bytesPerOop) "not enough room"
	 or: [limit ~= ptr and: [limit - ptr <= self allocationUnit]]) ifTrue: "can't split a single word"
		[self freeObject: freeChunk.
		 ^self integerObjectOf: count].
	count < self numSlotsMask ifTrue:
		[| smallObj |
		 smallObj := self allocateSlots: count format: self arrayFormat classIndex: ClassArrayCompactIndex.
		 0 to: count - 1 do:
			[:i|
			self storePointerUnchecked: i ofObject: smallObj withValue: (self fetchPointer: i ofFreeChunk: freeChunk)].
		 self freeChunkWithBytes: (self bytesInBody: freeChunk) at: (self startOfObject: freeChunk).
		 self beRootIfOld: smallObj.
		 self checkFreeSpace: GCModeFull.
		 ^smallObj].
	bytes := self largeObjectBytesForSlots: count.
	start := self startOfObject: freeChunk.
	self freeChunkWithBytes: limit - start - bytes at: start + bytes.
	totalFreeOldSpace := totalFreeOldSpace - bytes.
	self rawOverflowSlotsOf: freeChunk put: count.
	self set: freeChunk classIndexTo: ClassArrayCompactIndex formatTo: self arrayFormat.
	self possibleRootStoreInto: freeChunk.
	self checkFreeSpace: GCModeFull.
	self runLeakCheckerFor: GCModeFull.
	^freeChunk! !

!SpurMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 6/21/2021 10:33'!
allNewSpaceEntitiesDo: aBlock
	"Enumerate all new space objects, including free objects."
	<inline: true>
	| start prevObj prevPrevObj objOop |
	prevPrevObj := prevObj := nil.
	"After a scavenge eden is empty, futureSpace is empty, and all newSpace objects are
	  in pastSpace.  Objects are allocated in eden.  So enumerate only pastSpace and eden."
	self assert: (scavenger pastSpace start < scavenger eden start).
	start := self startAddressForBridgedHeapEnumeration.
	start > freeStart ifTrue: [^self].
	self bridgePastSpaceAndEden.
	objOop := self objectStartingAt: start.
	[self oop: objOop isLessThan: freeStart] whileTrue:
		[aBlock value: objOop.
		 prevPrevObj := prevObj.
		 prevObj := objOop.
		 objOop := self objectAfterMaybeSlimBridge: objOop limit: freeStart].
	self touch: prevPrevObj.
	self touch: prevObj! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 9/24/2020 12:50'!
allNewSpaceObjectsDo: aBlock
	"Enumerate all new space objects, excluding free objects."
	<inline: true>
	self allNewSpaceEntitiesDo:
		[:objOop|
		 self assert: (self isBridgeOrEnumerableObjectNoAssert: objOop).
		 aBlock value: objOop]! !

!SpurMemoryManager methodsFor: 'primitive support' stamp: 'eem 06/22/2021 15:31'!
allObjects
	"Attempt to answer an array of all objects, excluding those that may
	 be garbage collected as a side effect of allocating the result array.
	 If no memory is available answer the number of objects as a SmallInteger.
	 Since objects are at least 16 bytes big, and the largest SmallInteger covers
	 1/4 of the address space, the count can never overflow."
	| freeChunk ptr start limit count bytes |
	MarkObjectsForEnumerationPrimitives ifTrue:
		[self markObjects: true]. "may not want to revive objects unnecessarily; but marking is sloooow."
	freeChunk := self allocateLargestFreeChunk. "N.B. Does /not/ update totalFreeOldSpace"
	ptr := start := freeChunk + self baseHeaderSize.
	limit := self addressAfter: freeChunk.
	count := 0.
	self allHeapEntitiesDo:
		[:obj| "continue enumerating even if no room so as to unmark all objects."
		 (MarkObjectsForEnumerationPrimitives
				ifTrue: [self isMarked: obj]
				ifFalse: [true]) ifTrue:
			[(self isNormalObject: obj)
				ifTrue:
					[MarkObjectsForEnumerationPrimitives ifTrue:
						[self setIsMarkedOf: obj to: false].
					 count := count + 1.
					 ptr < limit ifTrue:
						[self longAt: ptr put: obj.
						 ptr := ptr + self bytesPerOop]]
				ifFalse:
					[MarkObjectsForEnumerationPrimitives ifTrue:
						[(self isSegmentBridge: obj) ifFalse:
							[self setIsMarkedOf: obj to: false]]]]].
	self assert: (self isEmptyObjStack: markStack).
	MarkObjectsForEnumerationPrimitives
		ifTrue:
			[self assert: self allObjectsUnmarked.
			 self emptyObjStack: weaklingStack]
		ifFalse:
			[self assert: (self isEmptyObjStack: weaklingStack)].
	self assert: count >= self numSlotsMask.
	(count > (ptr - start / self bytesPerOop) "not enough room"
	 or: [limit ~= ptr and: [limit - ptr <= self allocationUnit]]) ifTrue: "can't split a single word"
		[self freeChunkWithBytes: (self bytesInBody: freeChunk) at: (self startOfObject: freeChunk).
		 self checkFreeSpace: GCModeFull.
		 ^self integerObjectOf: count].
	bytes := self largeObjectBytesForSlots: count.
	start := self startOfObject: freeChunk.
	self freeChunkWithBytes: limit - start - bytes at: start + bytes.
	totalFreeOldSpace := totalFreeOldSpace - bytes.
	self rawOverflowSlotsOf: freeChunk put: count.
	self set: freeChunk classIndexTo: ClassArrayCompactIndex formatTo: self arrayFormat.
	self possibleRootStoreInto: freeChunk.
	self checkFreeSpace: GCModeFull.
	self runLeakCheckerFor: GCModeFull.
	^freeChunk! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 9/24/2020 17:03'!
allObjectsDo: aBlock
	<inline: true>
	| startObject |
	startObject := self objectStartingAt: self startAddressForBridgedHeapEnumeration.
	self enableObjectEnumerationFrom: startObject.
	self allEntitiesFrom: startObject
		do: [:objOop|
			 (self isEnumerableObject: objOop) ifTrue:
				[aBlock value: objOop]]! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 2/4/2014 11:14'!
allObjectsDoSafely: aBlock
	<inline: true>
	self allObjectsDo: aBlock! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 9/24/2020 12:14'!
allObjectsFrom: initialObject do: aBlock
	"Enumerate all objects (i.e. exclude bridges, forwarders and free chunks)
	 in oldSpace starting at initialObject."
	<inline: true>
	self allEntitiesFrom: initialObject
		do: [:objOop|
			 (self isEnumerableObject: objOop) ifTrue:
				[aBlock value: objOop]]! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 9/25/2013 16:38'!
allObjectsInFreeTree: freeNode do: aBlock
	| listNode |
	freeNode = 0 ifTrue: [^0].
	listNode := freeNode.
	[listNode ~= 0] whileTrue:
		[aBlock value: listNode.
		 listNode := self fetchPointer: self freeChunkNextIndex ofFreeChunk: listNode].
	self allObjectsInFreeTree: (self fetchPointer: self freeChunkSmallerIndex ofFreeChunk: freeNode)
		do: aBlock.
	self allObjectsInFreeTree: (self fetchPointer: self freeChunkLargerIndex ofFreeChunk: freeNode)
		do: aBlock! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 10/21/2013 10:30'!
allObjectsInFreeTreeDo: aBlock
	"Enumerate all objects in the free tree (in order, smaller to larger).
	 This is an iterative version so that the block argument can be
	 inlined by Slang. The trick to an iterative binary tree application is
	 to apply the function on the way back up when returning from a
	 particular direction, in this case up from the larger child."
	<inline: true>
	self freeTreeNodesDo:
		[:freeTreeNode| | next |
		 next := freeTreeNode.
		 [aBlock value: next.
		  next := self fetchPointer: self freeChunkNextIndex ofFreeChunk: next.
		  next ~= 0] whileTrue.
		 freeTreeNode]! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 2/17/2017 10:29'!
allObjectsUnmarked
	self allObjectsDo:
		[:o| (self isMarked: o) ifTrue: [bogon := o. ^false]].
	^true! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 5/23/2014 16:23'!
allOldMarkedWeakObjectsOnWeaklingStack
	self allOldSpaceEntitiesDo:
		[:o|
		((self isWeakNonImm: o)
		 and: [self isMarked: o]) ifTrue:
			[(self is: o onObjStack: weaklingStack) ifFalse:
				[^false]]].
	^true! !

!SpurMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 11/10/2013 14:59'!
allOldSpaceEntitiesDo: aBlock
	<inline: true>
	self allOldSpaceEntitiesFrom: self firstObject do: aBlock! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 5/9/2014 15:21'!
allOldSpaceEntitiesForCoalescingFrom: firstObj do: aBlock
	<inline: true>
	| prevObj prevPrevObj objOop rawNumSlots rawNumSlotsAfter |
	prevPrevObj := prevObj := nil.
	objOop := firstObj.
	[self assert: objOop \\ self allocationUnit = 0.
	 self oop: objOop isLessThan: endOfMemory] whileTrue:
		[self assert: (self long64At: objOop) ~= 0.
		 rawNumSlots := self rawNumSlotsOf: objOop.
		 aBlock value: objOop.
		 "If the number of slot changes coalescing changed an object from a single to a double header.
		  In future have the block return the vaue.  It should know when things change."
		 self flag: 'future work'.
		 rawNumSlotsAfter := self rawNumSlotsOf: objOop.
		 (rawNumSlotsAfter ~= rawNumSlots
		  and: [rawNumSlotsAfter = self numSlotsMask]) ifTrue:
			[objOop := objOop + self baseHeaderSize.
			 self assert: (self objectAfter: prevObj limit: endOfMemory) = objOop].
		 prevPrevObj := prevObj.
		 prevObj := objOop.
		 objOop := self objectAfter: objOop limit: endOfMemory].
	self touch: prevPrevObj.
	self touch: prevObj! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 2/2/2017 11:32'!
allOldSpaceEntitiesForCompactingFrom: initialObject to: finalObject do: aBlock
	<inline: true>
	| limit prevObj prevPrevObj objOop nextObj |
	self assert: (self isOldObject: initialObject).
	self assert: (self oop: finalObject isLessThanOrEqualTo: endOfMemory).
	prevPrevObj := prevObj := nil.
	objOop := initialObject.
	limit := (self oop: finalObject isLessThan: endOfMemory) ifTrue: [self addressAfter: finalObject] ifFalse: [endOfMemory].
	[self assert: objOop \\ self allocationUnit = 0.
	 self oop: objOop isLessThan: limit] whileTrue:
		[self assert: (self long64At: objOop) ~= 0.
		 nextObj := self objectAfter: objOop limit: endOfMemory.
		 aBlock value: objOop value: nextObj.
		 prevPrevObj := prevObj.
		 prevObj := objOop.
		 objOop := nextObj].
	self touch: prevPrevObj.
	self touch: prevObj! !

!SpurMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 7/16/2014 21:05'!
allOldSpaceEntitiesFrom: initialObject do: aBlock
	<inline: true>
	| prevObj prevPrevObj objOop |
	self assert: (self isOldObject: initialObject).
	prevPrevObj := prevObj := nil.
	objOop := initialObject.
	[self assert: objOop \\ self allocationUnit = 0.
	 self oop: objOop isLessThan: endOfMemory] whileTrue:
		[self assert: (self long64At: objOop) ~= 0.
		 aBlock value: objOop.
		 prevPrevObj := prevObj.
		 prevObj := objOop.
		 objOop := self objectAfter: objOop limit: endOfMemory].
	self touch: prevPrevObj.
	self touch: prevObj! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 12/31/2016 09:10'!
allOldSpaceEntitiesFrom: initialObject to: finalObject do: aBlock
	<inline: true>
	| prevObj prevPrevObj objOop |
	self assert: ((self isNonImmediate: initialObject) and: [segmentManager isInSegments: initialObject]).
	self assert: ((self isNonImmediate: finalObject) and: [segmentManager isInSegments: finalObject]).
	prevPrevObj := prevObj := nil.
	objOop := initialObject.
	[self assert: objOop \\ self allocationUnit = 0.
	 self oop: objOop isLessThanOrEqualTo: finalObject] whileTrue:
		[self assert: (self long64At: objOop) ~= 0.
		 aBlock value: objOop.
		 prevPrevObj := prevObj.
		 prevObj := objOop.
		 objOop := self objectAfter: objOop limit: endOfMemory].
	self touch: prevPrevObj.
	self touch: prevObj! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 10/7/2013 13:15'!
allOldSpaceObjectsDo: aBlock
	<inline: true>
	self allOldSpaceObjectsFrom: self firstObject do: aBlock! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 11/10/2013 15:06'!
allOldSpaceObjectsFrom: initialObject do: aBlock
	"Enumerate all objects (i.e. exclude bridges, forwarders and free chunks)
	 in oldSpace starting at initialObject."
	<inline: true>
	self allOldSpaceEntitiesFrom: initialObject
		do: [:objOop|
			 (self isEnumerableObject: objOop) ifTrue:
				[aBlock value: objOop]]! !

!SpurMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 6/21/2021 10:33'!
allPastSpaceEntitiesDo: aBlock
	"Enumerate all past space objects, including free objects."
	<inline: true>
	| prevObj prevPrevObj objOop |
	prevPrevObj := prevObj := nil.
	objOop := self objectStartingAt: scavenger pastSpace start.
	[self oop: objOop isLessThan: pastSpaceStart] whileTrue:
		[aBlock value: objOop.
		 prevPrevObj := prevObj.
		 prevObj := objOop.
		 objOop := self objectAfterMaybeSlimBridge: objOop limit: pastSpaceStart].
	self touch: prevPrevObj.
	self touch: prevObj! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 5/7/2014 09:50'!
allPastSpaceObjectsDo: aBlock
	"Enumerate all past space objects, excluding free objects."
	<inline: true>
	self allPastSpaceEntitiesDo:
		[:objOop|
		 self assert: (self isEnumerableObjectNoAssert: objOop).
		 aBlock value: objOop]! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 7/9/2014 10:34'!
allStrongSlotsOfWeaklingAreMarked: aWeakling
	"N.B. generateToByDoLimitExpression:negative:on: guards against (unsigned)0 - 1 going +ve"
	0 to: (self numStrongSlotsOfWeakling: aWeakling) - 1 do:
		[:i| | referent |
		referent := self fetchPointer: i ofObject: aWeakling.
		(self isNonImmediate: referent) ifTrue:
			[(self isMarked: referent) ifFalse:
				[^false]]].
	^true! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 5/25/2016 18:13'!
allUnscannedEphemeronsAreActive
	unscannedEphemerons start to: unscannedEphemerons top - self bytesPerOop by: self bytesPerOop do:
		[:p| | key |
		key := self keyOfMaybeFiredEphemeron: (self longAt: p).
		((self isImmediate: key) or: [self isMarked: key]) ifTrue:
			[^false]].
	^true! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 11/13/2014 12:15'!
allocateBytes: numBytes classIndex: classIndex
	"Allocate an object of numBytes.  Answer nil if no available memory.
	 classIndex must be that of a byte class (e.g. ByteString).
	 The object is *NOT FILLED*."
	<var: #numBytes type: #usqInt>
	self assert: (coInterpreter addressCouldBeClassObj: (self classAtIndex: classIndex)).
	self assert: (self instSpecOfClass: (self classAtIndex: classIndex)) = self firstByteFormat.
	^self
		allocateSlots: (numBytes + self bytesPerOop - 1 // self bytesPerOop)
		format: (self byteFormatForNumBytes: numBytes)
		classIndex: classIndex! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
allocateLargestFreeChunk
	"Answer the largest free chunk in the free lists."
	<inline: false>
	| freeChunk next |
	"would like to use ifNotNil: but the ^next inside the ^blah ifNotNil: confused Slang"
	freeChunk := self findLargestFreeChunk.
	freeChunk ifNil: [^nil].
	"This will be the node, not a list element.  Answer a list element in preference."
	next := self fetchPointer: self freeChunkNextIndex ofFreeChunk: freeChunk.
	next ~= 0 ifTrue:
		[self assert: (self bytesInBody: freeChunk) >= self numFreeLists. "findLargestFreeChunk searches only the tree"
		 self 
			setNextFreeChunkOf: freeChunk 
			withValue: (self fetchPointer: self freeChunkNextIndex ofFreeChunk: next) 
			isLilliputianSize: false.
		 ^next].
	self unlinkSolitaryFreeTreeNode: freeChunk.
	^freeChunk! !

!SpurMemoryManager methodsFor: 'testing' stamp: 'eem 11/18/2016 16:55'!
allocateMemoryOfSize: memoryBytes
	<doNotGenerate>
	| bytesPerElement |
	bytesPerElement := (self memoryClass basicNew: 0) bytesPerElement.
	memory := self memoryClass new: memoryBytes + bytesPerElement - 1 // bytesPerElement! !

!SpurMemoryManager methodsFor: 'spur bootstrap' stamp: 'eem 2/3/2017 09:38'!
allocateMemoryOfSize: memoryBytes newSpaceSize: newSpaceBytes stackSize: stackBytes codeSize: codeBytes
	"Intialize the receiver for bootsraping an image.
	 Set up a large oldSpace and an empty newSpace and set-up freeStart and scavengeThreshold
	 to allocate in oldSpace.  Later on (in initializePostBootstrap) freeStart and scavengeThreshold
	 will be set to sane values."
	<doNotGenerate>
	self assert: (memoryBytes \\ self allocationUnit = 0
				and: [newSpaceBytes \\ self allocationUnit = 0
				and: [codeBytes \\ self allocationUnit = 0]]).
	self allocateMemoryOfSize: memoryBytes + newSpaceBytes + codeBytes + stackBytes.
	newSpaceStart := codeBytes + stackBytes.
	endOfMemory := freeOldSpaceStart := memoryBytes + newSpaceBytes + codeBytes + stackBytes.
	"leave newSpace empty for the bootstrap"
	freeStart := newSpaceBytes + newSpaceStart.
	oldSpaceStart := newSpaceLimit := newSpaceBytes + newSpaceStart.
	scavengeThreshold := memory size * memory bytesPerElement. "i.e. /don't/ scavenge."
	scavenger := SpurGenerationScavenger simulatorClass new.
	scavenger manager: self.
	scavenger newSpaceStart: newSpaceStart
				newSpaceBytes: newSpaceBytes
				survivorBytes: newSpaceBytes // self scavengerDenominator.
	compactor := self class compactorClass simulatorClass new manager: self; yourself! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 09/28/2015 18:48'!
allocateNewSpaceSlots: numSlots format: formatField classIndex: classIndex
	"Allocate an object with numSlots in newSpace.  This is for the `ee' execution engine allocations,
	 and must be satisfied.  If no memory is available, abort.  If the allocation pushes freeStart past
	 scavengeThreshold and a scavenge is not already scheduled, schedule a scavenge."
	| numBytes newObj |
	"Object headers are 8 bytes in length if the slot size fits in the num slots field (max implies overflow),
	 16 bytes otherwise (num slots in preceding word).
	 Objects always have at least one slot, for the forwarding pointer,
	 and are multiples of 8 bytes in length."
	numSlots >= self numSlotsMask
		ifTrue:
			[(self wordSize >= 8 and: [numSlots > 16rffffffff]) ifTrue:
				[^nil]. "overflow size must fit in 32-bits"
			 newObj := freeStart + self baseHeaderSize.
			 numBytes := self largeObjectBytesForSlots: numSlots]
		ifFalse:
			[newObj := freeStart.
			 numBytes := self smallObjectBytesForSlots: numSlots].
	
	freeStart + numBytes > scavengeThreshold ifTrue:
		[needGCFlag ifFalse: [self scheduleScavenge].
		 freeStart + numBytes > scavenger eden limit ifTrue:
			[self error: 'no room in eden for allocateNewSpaceSlots:format:classIndex:'.
			 ^0]].
	numSlots >= self numSlotsMask
		ifTrue: "for header parsing we put a saturated slot count in the prepended overflow size word"
			[self flag: #endianness.
			 self longAt: freeStart put: numSlots.
			 self longAt: freeStart + 4 put: self numSlotsMask << self numSlotsHalfShift.
			 self long64At: newObj put: (self headerForSlots: self numSlotsMask format: formatField classIndex: classIndex)]
		ifFalse:
			[self long64At: newObj put: (self headerForSlots: numSlots format: formatField classIndex: classIndex)].
	self assert: numBytes \\ self allocationUnit = 0.
	self assert: newObj \\ self allocationUnit = 0.
	freeStart := freeStart + numBytes.
	^newObj! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
allocateOldSpaceChunkOfBytes: chunkBytes
	"Answer a chunk of oldSpace from the free lists, if available,
	 otherwise answer nil.  Break up a larger chunk if one of the
	 exact size does not exist.  N.B.  the chunk is simply a pointer, it
	 has no valid header.  The caller *must* fill in the header correctly."
	<var: #chunkBytes type: #usqInt>
	| initialIndex chunk index nodeBytes parent child |
	"for debugging:" "totalFreeOldSpace := self totalFreeListBytes"
	totalFreeOldSpace := totalFreeOldSpace - chunkBytes. "be optimistic (& don't wait for the write)"
	initialIndex := chunkBytes / self allocationUnit.
	(initialIndex < self numFreeLists and: [1 << initialIndex <= freeListsMask]) ifTrue:
		[(freeListsMask anyMask: 1 << initialIndex) ifTrue:
			[(chunk := freeLists at: initialIndex) ~= 0 ifTrue:
				[self assert: chunk = (self startOfObject: chunk).
				 self assertValidFreeObject: chunk.
				 self unlinkFreeChunk: chunk atIndex: initialIndex chunkBytes: chunkBytes.
				^ chunk].
			 freeListsMask := freeListsMask - (1 << initialIndex)].
		 "first search for free chunks of a multiple of chunkBytes in size"
		 index := initialIndex.
		 [(index := index + index) < self numFreeLists
		  and: [1 << index <= freeListsMask]] whileTrue:
			[(freeListsMask anyMask: 1 << index) ifTrue:
				[(chunk := freeLists at: index) ~= 0 ifTrue:
					[self assert: chunk = (self startOfObject: chunk).
					 self assertValidFreeObject: chunk.
					 self unlinkFreeChunk: chunk atIndex: index isLilliputianSize: false.
					 self assert: (self bytesInBody: chunk) = (index * self allocationUnit).
					 self freeChunkWithBytes: index * self allocationUnit - chunkBytes
						at: (self startOfObject: chunk) + chunkBytes.
					^chunk].
				 freeListsMask := freeListsMask - (1 << index)]].
		 "now get desperate and use the first that'll fit.
		  Note that because the minimum free size is 16 bytes (2 * allocationUnit), to
		  leave room for the forwarding pointer/next free link, we can only break chunks
		  that are at least 16 bytes larger, hence start at initialIndex + 2."
		 index := initialIndex + 1.
		 [(index := index + 1) < self numFreeLists
		  and: [1 << index <= freeListsMask]] whileTrue:
			[(freeListsMask anyMask: 1 << index) ifTrue:
				[(chunk := freeLists at: index) ~= 0 ifTrue:
					[self assert: chunk = (self startOfObject: chunk).
					 self assertValidFreeObject: chunk.
					 self unlinkFreeChunk: chunk atIndex: index isLilliputianSize: false.
					 self assert: (self bytesInBody: chunk) = (index * self allocationUnit).
					 self freeChunkWithBytes: index * self allocationUnit - chunkBytes
						at: (self startOfObject: chunk) + chunkBytes.
					^chunk].
				 freeListsMask := freeListsMask - (1 << index)]]].

	"Large chunk, or no space on small free lists.  Search the large chunk list.
	 Large chunk list organized as a tree, each node of which is a list of chunks
	 of the same size. Beneath the node are smaller and larger blocks.
	 When the search ends parent should hold the smallest chunk at least as
	 large as chunkBytes, or 0 if none."
	parent := 0.
	child := freeLists at: 0.
	[child ~= 0] whileTrue:
		[| childBytes |
		 self assertValidFreeObject: child.
		 childBytes := self bytesInBody: child.
		 childBytes = chunkBytes
			ifTrue: "size match; try to remove from list at node."
				[chunk := self fetchPointer: self freeChunkNextIndex
								ofFreeChunk: child.
				 chunk ~= 0 ifTrue:
					[self assertValidFreeObject: chunk.
					 self 
						setNextFreeChunkOf: child 
						withValue: (self fetchPointer: self freeChunkNextIndex ofFreeChunk: chunk) 
						isLilliputianSize: false.
					 ^self startOfObject: chunk].
				 nodeBytes := childBytes.
				 parent := child.
				 child := 0] "break out of loop to remove interior node"
			ifFalse:
				["Note that because the minimum free size is 16 bytes (2 * allocationUnit), to
				  leave room for the forwarding pointer/next free link, we can only break chunks
				  that are at least 16 bytes larger, hence reject chunks < 2 * allocationUnit larger."
				childBytes <= (chunkBytes + self allocationUnit)
					ifTrue: "node too small; walk down the larger size of the tree"
						[child := self fetchPointer: self freeChunkLargerIndex ofFreeChunk: child]
					ifFalse:
						[parent := child. "parent will be smallest node >= chunkBytes + allocationUnit"
						 nodeBytes := childBytes.
						 child := self fetchPointer: self freeChunkSmallerIndex ofFreeChunk: child]]].
	parent = 0 ifTrue:
		[totalFreeOldSpace := totalFreeOldSpace + chunkBytes. "optimism was unfounded"
		 ^nil].

	"self printFreeChunk: parent"
	self assert: (nodeBytes = chunkBytes or: [nodeBytes >= (chunkBytes + (2 * self allocationUnit))]).
	self assert: (self bytesInBody: parent) = nodeBytes.

	"attempt to remove from list"
	chunk := self fetchPointer: self freeChunkNextIndex ofFreeChunk: parent.
	chunk ~= 0 ifTrue:
		[self assert: (chunkBytes = nodeBytes or: [chunkBytes + self allocationUnit < nodeBytes]).
		self 
			setNextFreeChunkOf: parent 
			withValue: (self fetchPointer: self freeChunkNextIndex ofFreeChunk: chunk) 
			isLilliputianSize: false.
		 chunkBytes ~= nodeBytes ifTrue:
			[self freeChunkWithBytes: nodeBytes - chunkBytes
					at: (self startOfObject: chunk) + chunkBytes].
		 ^self startOfObject: chunk].

	"no list; remove the interior node"
	chunk := parent.
	self unlinkSolitaryFreeTreeNode: chunk.

	"if there's space left over, add the fragment back."
	chunkBytes ~= nodeBytes ifTrue:
		[self freeChunkWithBytes: nodeBytes - chunkBytes
				at: (self startOfObject: chunk) + chunkBytes].
	^self startOfObject: chunk! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
allocateOldSpaceChunkOfBytes: chunkBytes suchThat: acceptanceBlock
	"Answer a chunk of oldSpace from the free lists that satisfies acceptanceBlock,
	 if available, otherwise answer nil.  Break up a larger chunk if one of the exact
	 size cannot be found.  N.B.  the chunk is simply a pointer, it has no valid header.
	 The caller *must* fill in the header correctly."
	<var: #chunkBytes type: #usqInt>
	| initialIndex node next prev index child childBytes acceptedChunk acceptedNode |
	<inline: true> "must inline for acceptanceBlock"
	"for debugging:" "totalFreeOldSpace := self totalFreeListBytes"
	totalFreeOldSpace := totalFreeOldSpace - chunkBytes. "be optimistic (& don't wait for the write)"
	initialIndex := chunkBytes / self allocationUnit.
	(initialIndex < self numFreeLists and: [1 << initialIndex <= freeListsMask]) ifTrue:
		[(freeListsMask anyMask: 1 << initialIndex) ifTrue:
			[(node := freeLists at: initialIndex) = 0
				ifTrue: [freeListsMask := freeListsMask - (1 << initialIndex)]
				ifFalse:
					[prev := 0.
					 [node ~= 0] whileTrue:
						[self assert: node = (self startOfObject: node).
						 self assertValidFreeObject: node.
						 next := self fetchPointer: self freeChunkNextIndex ofFreeChunk: node.
						 (acceptanceBlock value: node) ifTrue:
							[prev = 0
								ifTrue: [self unlinkFreeChunk: node atIndex: initialIndex chunkBytes: chunkBytes]
								ifFalse: [self setNextFreeChunkOf: prev withValue: next chunkBytes: chunkBytes].
							 ^node].
						 prev := node.
						 node := next]]].
		 "first search for free chunks of a multiple of chunkBytes in size"
		 index := initialIndex.
		 [(index := index + initialIndex) < self numFreeLists
		  and: [1 << index <= freeListsMask]] whileTrue:
			[(freeListsMask anyMask: 1 << index) ifTrue:
				[(node := freeLists at: index) = 0
					ifTrue: [freeListsMask := freeListsMask - (1 << index)]
					ifFalse:
						[prev := 0.
						 [node ~= 0] whileTrue:
							[self assert: node = (self startOfObject: node).
							  self assertValidFreeObject: node.
							 next := self fetchPointer: self freeChunkNextIndex ofFreeChunk: node.
							 (acceptanceBlock value: node) ifTrue:
								[prev = 0
									ifTrue: [self unlinkFreeChunk: node atIndex: index isLilliputianSize: false.]
									ifFalse: [self setNextFreeChunkOf: prev withValue: next isLilliputianSize: false.]. 
								 self freeChunkWithBytes: index * self allocationUnit - chunkBytes
									at: (self startOfObject: node) + chunkBytes.
								 ^node].
							 prev := node.
							 node := next]]]].
		 "now get desperate and use the first that'll fit.
		  Note that because the minimum free size is 16 bytes (2 * allocationUnit), to
		  leave room for the forwarding pointer/next free link, we can only break chunks
		  that are at least 16 bytes larger, hence start at initialIndex + 2."
		 index := initialIndex + 1.
		 [(index := index + 1) < self numFreeLists
		  and: [1 << index <= freeListsMask]] whileTrue:
			[(freeListsMask anyMask: 1 << index) ifTrue:
				[(node := freeLists at: index) = 0
					ifTrue: [freeListsMask := freeListsMask - (1 << index)]
					ifFalse:
						[prev := 0.
						 [node ~= 0] whileTrue:
							[self assert: node = (self startOfObject: node).
							  self assertValidFreeObject: node.
							 next := self fetchPointer: self freeChunkNextIndex ofFreeChunk: node.
							 (acceptanceBlock value: node) ifTrue:
								[prev = 0
									ifTrue: [self unlinkFreeChunk: node atIndex: index isLilliputianSize: false.]
									ifFalse: [self setNextFreeChunkOf: prev withValue: next isLilliputianSize: false.]. 
								 self freeChunkWithBytes: index * self allocationUnit - chunkBytes
									at: (self startOfObject: node) + chunkBytes.
								 ^node].
							 prev := node.
							 node := next]]]]].

	"Large chunk, or no space on small free lists.  Search the large chunk list.
	 Large chunk list organized as a tree, each node of which is a list of chunks
	 of the same size. Beneath the node are smaller and larger blocks.
	 When the search ends parent should hold the smallest chunk at least as
	 large as chunkBytes, or 0 if none.  acceptedChunk and acceptedNode save
	 us from having to back-up when the acceptanceBlock filters-out all nodes
	 of the right size, but there are nodes of the wrong size it does accept."
	child := freeLists at: 0.
	node := acceptedChunk := acceptedNode := 0.
	[child ~= 0] whileTrue:
		[ self assertValidFreeObject: child.
		 childBytes := self bytesInBody: child.
		 childBytes = chunkBytes ifTrue: "size match; try to remove from list at node."
			[node := child.
			 [prev := node.
			  node := self fetchPointer: self freeChunkNextIndex ofFreeChunk: node.
			  node ~= 0] whileTrue:
				[(acceptanceBlock value: node) ifTrue:
					[self assertValidFreeObject: node.
					 self 
						setNextFreeChunkOf: prev 
						withValue: (self fetchPointer: self freeChunkNextIndex ofFreeChunk: node) 
						isLilliputianSize: false.
					 ^self startOfObject: node]].
			 (acceptanceBlock value: child) ifTrue:
				[next := self fetchPointer: self freeChunkNextIndex ofFreeChunk: child.
				 next = 0
					ifTrue: "no list; remove the interior node"
						[self unlinkSolitaryFreeTreeNode: child]
					ifFalse: "list; replace node with it"
						[self inFreeTreeReplace: child with: next].
				 ^self startOfObject: child]].
		 child ~= 0 ifTrue:
			["Note that because the minimum free size is 16 bytes (2 * allocationUnit), to
			  leave room for the forwarding pointer/next free link, we can only break chunks
			  that are at least 16 bytes larger, hence reject chunks < 2 * allocationUnit larger."
			childBytes <= (chunkBytes + self allocationUnit)
				ifTrue: "node too small; walk down the larger size of the tree"
					[child := self fetchPointer: self freeChunkLargerIndex ofFreeChunk: child]
				ifFalse:
					[self flag: 'we can do better here; preferentially choosing the lowest node. That would be a form of best-fit since we are trying to compact down'.
					 node := child.
					 child := self fetchPointer: self freeChunkSmallerIndex ofFreeChunk: node.
					 acceptedNode = 0 ifTrue:
						[acceptedChunk := node.
						 "first search the list."
						 [acceptedChunk := self fetchPointer: self freeChunkNextIndex
													ofFreeChunk: acceptedChunk.
						  (acceptedChunk ~= 0 and: [acceptanceBlock value: acceptedChunk]) ifTrue:
							[acceptedNode := node].
						  acceptedChunk ~= 0 and: [acceptedNode = 0]] whileTrue.
						 "nothing on the list; will the node do?  This prefers
						  acceptable nodes higher up the tree over acceptable
						  list elements further down, but we haven't got all day..."
						 (acceptedNode = 0
						  and: [acceptanceBlock value: node]) ifTrue:
							[acceptedNode := node.
							 child := 0 "break out of loop now we have an acceptedNode"]]]]].

	acceptedNode ~= 0 ifTrue:
		[acceptedChunk ~= 0 ifTrue:
			[self assert: (self bytesInBody: acceptedChunk) >= (chunkBytes + self allocationUnit).
			 [next := self fetchPointer: self freeChunkNextIndex ofFreeChunk: acceptedNode.
			  next ~= acceptedChunk] whileTrue:
				[acceptedNode := next].
			 self 
				setNextFreeChunkOf: acceptedNode 
				withValue: (self fetchPointer: self freeChunkNextIndex ofFreeChunk: acceptedChunk) 
				isLilliputianSize: false.
			self freeChunkWithBytes: (self bytesInBody: acceptedChunk) - chunkBytes
					at: (self startOfObject: acceptedChunk) + chunkBytes.
			^self startOfObject: acceptedChunk].
		next := self fetchPointer: self freeChunkNextIndex ofFreeChunk: acceptedNode.
		next = 0
			ifTrue: "no list; remove the interior node"
				[self unlinkSolitaryFreeTreeNode: acceptedNode]
			ifFalse: "list; replace node with it"
				[self inFreeTreeReplace: acceptedNode with: next].
		 self assert: (self bytesInBody: acceptedNode) >= (chunkBytes + self allocationUnit).
		 self freeChunkWithBytes: (self bytesInBody: acceptedNode) - chunkBytes
				at: (self startOfObject: acceptedNode) + chunkBytes.
		^self startOfObject: acceptedNode].

	totalFreeOldSpace := totalFreeOldSpace + chunkBytes. "optimism was unfounded"
	^nil! !

!SpurMemoryManager methodsFor: 'sista support' stamp: 'eem 7/19/2014 10:20'!
allocatePinnedSlots: nSlots
	<api>
	| obj |
	obj := self allocateSlotsForPinningInOldSpace: nSlots
				bytes: (self objectBytesForSlots: nSlots)
				format: self wordIndexableFormat
				classIndex: 	self wordSizeClassIndexPun.
	obj ifNotNil:
		[self fillObj: obj numSlots: nSlots with: 0].
	^obj! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 11/17/2014 17:43'!
allocateSlots: numSlots format: formatField classIndex: classIndex
	"Allocate an object with numSlots space.  If there is room beneath scavengeThreshold
	 allocate in newSpace, otherwise alocate in oldSpace.  If there is not room in newSpace
	 and a scavenge is not already scheduled, schedule a scavenge."
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 6/27/2014 15:50'!
allocateSlotsForPinningInOldSpace: numSlots bytes: totalBytes format: formatField classIndex: classIndex
	"Answer the oop of a chunk of space in oldSpace with numSlots slots.  Try and
	 allocate in a segment that already includes pinned objects.  The header of the
	 result will have been filled-in but not the contents."
	<var: #totalBytes type: #usqInt>
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 10/2/2014 10:53'!
allocateSlotsInOldSpace: numSlots bytes: totalBytes format: formatField classIndex: classIndex
	"Answer the oop of a chunk of space in oldSpace with numSlots slots.  The header
	 will have been filled-in but not the contents.  If no memory is available answer nil."
	<var: #totalBytes type: #usqInt>
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 9/24/2013 16:25'!
allocateSlotsInOldSpace: numSlots format: formatField classIndex: classIndex
	<inline: true>
	^self
		allocateSlotsInOldSpace: numSlots
		bytes: (self objectBytesForSlots: numSlots)
		format: formatField
		classIndex: classIndex! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 4/19/2014 13:33'!
allocateSmallNewSpaceSlots: numSlots format: formatField classIndex: classIndex
	"Allocate an object with numSlots in newSpace, where numSlots is known to be small.
	 This is for the `ee' execution engine allocations, and must be satisfied.  If no memory
	 is available, abort.  If the allocation pushes freeStart past scavengeThreshold and a
	 scavenge is not already scheduled, schedule a scavenge."
	<inline: true>
	| numBytes newObj |
	self assert: numSlots < self numSlotsMask.
	newObj := freeStart.
	numBytes := self smallObjectBytesForSlots: numSlots.
	self assert: numBytes \\ self allocationUnit = 0.
	self assert: newObj \\ self allocationUnit = 0.
	freeStart + numBytes > scavengeThreshold ifTrue:
		[needGCFlag ifFalse: [self scheduleScavenge].
		 freeStart + numBytes > scavenger eden limit ifTrue:
			[self error: 'no room in eden for allocateSmallNewSpaceSlots:format:classIndex:'.
			 ^0]].
	self long64At: newObj put: (self headerForSlots: numSlots format: formatField classIndex: classIndex).
	freeStart := freeStart + numBytes.
	^newObj! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 8/30/2013 11:54'!
allocationUnit
	"All objects are a multiple of 8 bytes in length"
	^8! !

!SpurMemoryManager methodsFor: 'primitive support' stamp: 'eem 11/13/2014 12:15'!
ambiguousClass: aClass allInstancesInto: start limit: limit resultsInto: binaryBlock
	"Dea with ambiguity and normalize indices."
	<inline: true>
	| expectedIndex count ptr |
	count := 0.
	ptr := start.
	expectedIndex := self rawHashBitsOf: aClass.
	self allHeapEntitiesDo:
		[:obj| | actualIndex | "continue enumerating even if no room so as to unmark all objects and/or normalize class indices."
		 (MarkObjectsForEnumerationPrimitives
				ifTrue: [self isMarked: obj]
				ifFalse: [true]) ifTrue:
			[(self isNormalObject: obj)
				ifTrue:
					[MarkObjectsForEnumerationPrimitives ifTrue:
						[self setIsMarkedOf: obj to: false].
					 actualIndex := self classIndexOf: obj.
					 (self classOrNilAtIndex: actualIndex) = aClass ifTrue:
					 	[actualIndex ~= expectedIndex ifTrue:
							[self setClassIndexOf: obj to: expectedIndex].
						 count := count + 1.
						 ptr < limit ifTrue:
							[self longAt: ptr put: obj.
							 ptr := ptr + self bytesPerOop]]]
				ifFalse:
					[MarkObjectsForEnumerationPrimitives ifTrue:
						[(self isSegmentBridge: obj) ifFalse:
							[self setIsMarkedOf: obj to: false]]]]].
	self purgeDuplicateClassTableEntriesFor: aClass.
	binaryBlock value: count value: ptr
! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 1/18/2019 16:20'!
areIntegers: oop1 and: oop2
	"Test oop1 and oop2 to make sure both are SmallIntegers.
	 In 32-bits the tags are 1 = SmallInteger & 3 = SmallInteger, and 2 = Character.
	 In 64-bits the  tags are 1 = SmallInteger, 2 = Character & 4 = SmallFloat64.
	 In both cases we can test for two SmallIntegers by anding them together and
	 if the result still has the least significant bit set then both are SmallIntegers."
	^(oop1 bitAnd: oop2) anyMask: self smallIntegerTag! !

!SpurMemoryManager methodsFor: 'class table puns' stamp: 'eem 3/9/2016 14:48'!
arrayClassIndexPun
	"Class puns are class indices not used by any class.  There is an entry
	 for the pun that refers to the notional class of objects with this class
	 index.  But because the index doesn't match the class it won't show up
	 in allInstances, hence hiding the object with a pun as its class index.
	 The puns occupy indices 16 through 31."
	<cmacro>
	^16! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 10/9/2013 19:36'!
arrayFormat
	<api>
	^2! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 12/12/2013 11:29'!
arrayValueOf: arrayOop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter arrayValueOf: arrayOop! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'cb 10/5/2018 12:49'!
assertFreeChunkPrevHeadZero
	|min|
	self wordSize = 8 ifTrue: [min := 3] ifFalse: [min := 2].
	min to: self numFreeLists - 1 do:
		[:i| 
		 	(freeLists at: i) ~= 0 ifTrue:
				[self deny: (self isLilliputianSize: (freeLists at: i)).
				 self assert: (self fetchPointer: self freeChunkPrevIndex ofFreeChunk: (freeLists at: i)) = 0]].
	"Large chunks"
	self freeTreeNodesDo: [:freeNode |
		self assert: (self fetchPointer: self freeChunkPrevIndex ofFreeChunk: freeNode) = 0.
		freeNode].
	^ true! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
assertInnerValidFreeObject: objOop
	<inline: #never> "we don't want to inline so we can nest that in an assertion with the return true so the production VM does not generate any code here, while in simulation, the code breaks on the assertion we want to."
	| chunk index |
	self assert: (self oop: (self addressAfter: objOop) isLessThanOrEqualTo: endOfMemory).
	chunk := self fetchPointer: self freeChunkNextIndex ofFreeChunk: objOop.
	self assert: (chunk = 0 or: [self isFreeOop: chunk]).
	(self isLilliputianSize: (self bytesInBody: objOop)) ifFalse:
		["double linkedlist assertions"
		 chunk := self fetchPointer: self freeChunkNextIndex ofFreeChunk: objOop.
		 chunk = 0 ifFalse: 
			[self assert: (self isFreeOop: chunk).
			 self assert: objOop = (self fetchPointer: self freeChunkPrevIndex ofFreeChunk: chunk)].
		chunk := self fetchPointer: self freeChunkPrevIndex ofFreeChunk: objOop.
		index := (self bytesInBody: objOop) / self allocationUnit.
		(index < self numFreeLists and: [1 << index <= freeListsMask]) 
			ifTrue: 
				[(freeLists at: index) = objOop ifTrue: [self assert: chunk = 0]]
			ifFalse: 
				[self freeTreeNodesDo: [:freeNode |
					freeNode = objOop ifTrue: [self assert: chunk = 0]. freeNode]].
		 chunk = 0 ifFalse: 
			[self assert: (self isFreeOop: chunk).
			 self assert: objOop = (self fetchPointer: self freeChunkNextIndex ofFreeChunk: chunk)]].
	(self isLargeFreeObject: objOop) ifTrue: 
		["Tree assertions"
		chunk := self fetchPointer: self freeChunkParentIndex ofFreeChunk: objOop.
		self assert: (chunk = 0 or: [(self isFreeOop: chunk) and: [self isLargeFreeObject: chunk]]).
		chunk := self fetchPointer: self freeChunkSmallerIndex ofFreeChunk: objOop.
		self assert: (chunk = 0 or: [(self isFreeOop: chunk) and: [self isLargeFreeObject: chunk]]).
		chunk := self fetchPointer: self freeChunkLargerIndex ofFreeChunk: objOop.
		self assert: (chunk = 0 or: [(self isFreeOop: chunk) and: [self isLargeFreeObject: chunk]])].
	^ true! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'cb 6/8/2018 09:10'!
assertValidFreeObject: objOop
	<inline: true>
	"assertInnerValidFreeObject: is never inlined and always returns true.
	 For the production VM, this is entirely removed.
	 For the other VMs and in simulation, the code breaks at the first warning/assertion failure"
	self assert: (self assertInnerValidFreeObject: objOop)! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 2/21/2019 09:29'!
assignClassIndicesAndPinFrom: segmentStart to: segmentLimit outPointers: outPointerArray filling: loadedObjectsArray
	"This is part of loadImageSegmentFrom:outPointers:.
	 Make a final pass, assigning the real class indices and/or pinning pinned objects."
	| fillIdx objOop |
	objOop := self objectStartingAt: segmentStart.
	fillIdx := 0.
	[self oop: objOop isLessThan: segmentLimit] whileTrue:
		[| classRef classOop classIndex |
		 self storePointerUnchecked: fillIdx ofObject: loadedObjectsArray withValue: objOop.
		 fillIdx := fillIdx + 1.
		 "In the segment, class indices are offset indexes into the segment data,
		  or into outPointers.  See mapOopsFrom:to:outPointers:outHashes: and
		  newOutPointer:at:in:hashes:."
		 classRef := self classIndexOf: objOop.
		 classOop := (classRef anyMask: TopHashBit)
						ifTrue: [self fetchPointer: classRef - TopHashBit ofObject: outPointerArray]
						ifFalse: [classRef - self firstClassIndexPun * self allocationUnit + segmentStart].
		 classIndex := self rawHashBitsOf: classOop.
		 classIndex = 0 ifTrue:
			[classIndex := self ensureBehaviorHash: classOop.
			 classIndex < 0 ifTrue: "Error code e.g. - PrimErrNoMemory"
				[^classIndex negated halt]].
		 self assert: (classIndex > self lastClassIndexPun
					  and: [(self classOrNilAtIndex: classIndex) = classOop]).
		 self setClassIndexOf: objOop to: classIndex.
		 ((self isInNewSpace: objOop)
		  and: [self isPinned: objOop]) ifTrue:
			[| oldClone |
			 oldClone := self cloneInOldSpace: objOop forPinning: true.
			 oldClone ~= 0 ifTrue:
				[self setIsPinnedOf: oldClone to: true.
				 self forward: objOop to: oldClone]].
		 objOop := self objectAfter: objOop limit: segmentLimit]! !

!SpurMemoryManager methodsFor: 'growing/shrinking memory' stamp: 'eem 11/26/2013 17:09'!
assimilateNewSegment: segInfo
	"Update after adding a segment.
	 Here we set freeOldSpaceStart & endOfMemory if required."
	<var: #segInfo type: #'SpurSegmentInfo *'>
	segInfo segLimit >= endOfMemory ifTrue:
		[freeOldSpaceStart :=
		 endOfMemory := segInfo segLimit - self bridgeSize]! !

!SpurMemoryManager methodsFor: 'compaction' stamp: 'eem 4/30/2014 09:11'!
atLeastClassIndexHalfHeader: obj
	"PRIVATE: For compaction, answer the bits contaning the
	 classIndex and isPinned bits in the most natural form."
	^self longAt: obj! !

!SpurMemoryManager methodsFor: 'growing/shrinking memory' stamp: 'eem 1/13/2017 10:56'!
attemptToShrink
	"Attempt to shrink memory after successfully reclaiming lots of memory.
	 If there's enough memory to shrink then be sure to attept to shrink by
	 at least growHeaqdroom because segments are typically of that size."
	(totalFreeOldSpace > shrinkThreshold
	 and: [totalFreeOldSpace > growHeadroom
	 and: [segmentManager shrinkObjectMemory: (totalFreeOldSpace - growHeadroom max: growHeadroom)]]) ifTrue:
		[statShrinkMemory := statShrinkMemory + 1]! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 11/13/2014 12:15'!
averageObjectSizeInBytes
	"Answer an approximation of the average object size.  This is a bit of an underestimate.
	 In the 32-bit system average object size is about 11 words per object, including header."
	^8 * self bytesPerOop! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 11/21/2013 11:32'!
baseAddressOfImage
	^oldSpaceStart! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 4/19/2014 13:33'!
baseHeader: obj
	^self long64At: obj! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 09/28/2015 18:48'!
baseHeaderSize
	"Object headers are 8 bytes in length if the slot size fits in the slot size field (max implies overflow),
	 16 bytes otherwise (slot size in preceding word)."
	^8! !

!SpurMemoryManager methodsFor: 'store check' stamp: 'eem 11/20/2013 13:55'!
beRootIfOld: oop 
	"If this object is old, mark it as a root (because a new object
	 may be stored into it)."
	<api>
	<inline: false>
	(self isOldObject: oop) ifTrue:"No, oop is an old object"
		[self possibleRootStoreInto: oop]! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 10/15/2013 09:21'!
become: array1 with: array2
	<api>
	^self become: array1 with: array2 twoWay: true copyHash: true! !

!SpurMemoryManager methodsFor: 'become api' stamp: 'eem 6/20/2016 12:58'!
become: array1 with: array2 twoWay: twoWayFlag copyHash: copyHashFlag
	"All references to each object in array1 are swapped with all references to the
	 corresponding object in array2. That is, all pointers to one object are replaced
	 with with pointers to the other. The arguments must be arrays of the same length. 
	 Answers PrimNoErr if the primitive succeeds, otherwise a relevant error code."
	"Implementation: Uses lazy forwarding to defer updating references until message send."
	<inline: false>
	| ec |
	self assert: becomeEffectsFlags = 0.
	self runLeakCheckerFor: GCModeBecome.
	(self isArray: array1) ifFalse:
		[^PrimErrBadReceiver].
	((self isArray: array2)
	 and: [(self numSlotsOf: array1) = (self numSlotsOf: array2)]) ifFalse:
		[^PrimErrBadArgument].
	ec := self containsOnlyValidBecomeObjects: array1 and: array2 twoWay: twoWayFlag copyHash: copyHashFlag.
	ec ~= 0 ifTrue:
		[becomeEffectsFlags := 0.
		 ^ec].

	coInterpreter preBecomeAction.
	twoWayFlag
		ifTrue:
			[self innerBecomeObjectsIn: array1 and: array2 copyHash: copyHashFlag]
		ifFalse:
			[self innerBecomeObjectsIn: array1 to: array2 copyHash: copyHashFlag].
	self followSpecialObjectsOop.
	"N.B. perform coInterpreter's postBecomeAction: *before* postBecomeScanClassTable:
	 to allow the coInterpreter to void method cache entries by spotting classIndices that
	 refer to forwarded objects. postBecomeScanClassTable: follows forwarders in the table."
	coInterpreter postBecomeAction: becomeEffectsFlags.
	self postBecomeScanClassTable: becomeEffectsFlags.
	becomeEffectsFlags := 0.

	self assert: self validClassTableHashes.
	self runLeakCheckerFor: GCModeBecome.

	^PrimNoErr "success"! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 9/23/2014 12:35'!
becomeEffectFlagsFor: objOop
	"Answer the appropriate become effect flags for objOop, or 0 if none.
	 The effect flags determine how much work is done after the become
	 in following forwarding pointers, voiding method caches, etc."
	<inline: false>
	^(self isPointersNonImm: objOop)
		ifTrue:
			[| hash |
			 ((hash := self rawHashBitsOf: objOop) ~= 0
			  and: [(self classAtIndex: hash) = objOop])
				ifTrue: [BecamePointerObjectFlag + BecameActiveClassFlag]
				ifFalse: [BecamePointerObjectFlag]]
		ifFalse:
			[(self isCompiledMethod: objOop)
				ifTrue: [BecameCompiledMethodFlag]
				ifFalse: [0]]! !

!SpurMemoryManager methodsFor: 'gc - scavenge/compact' stamp: 'eem 12/25/2016 09:11'!
beginSlidingCompaction
	gcPhaseInProgress := SlidingCompactionInProgress! !

!SpurMemoryManager methodsFor: 'compaction - analysis' stamp: 'eem 06/22/2021 15:31'!
biggies
	"This method answers a sorted collection of the objects >= 1,000,000 bytes long,
	 above the lowest large free chunk, sandwiched between nilObj and the end of memory."
	<doNotGenerate>
	| lowestFree biggies |
	lowestFree := SmallInteger maxVal.
	self allObjectsInFreeTreeDo:
		[:f| (self addressAfter: f) < endOfMemory ifTrue: [f < lowestFree ifTrue: [lowestFree := f]]].
	biggies := SortedCollection new.
	self allOldSpaceObjectsFrom: lowestFree do:
		[:f|
		(self bytesInBody: f) >= 1000000 ifTrue:
			[biggies add: f]].
	^{{nilObj hex. #nil}}, (biggies collect: [:f| {f hex. self bytesInBody: f}]), {{endOfMemory hex. #endOfMemory}}! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 10/8/2013 09:06'!
bitsSetInFreeSpaceMaskForAllFreeLists
	0 to: self numFreeLists - 1 do:
		[:i|
		((freeLists at: i) ~= 0
		 and: [1 << i noMask: freeListsMask]) ifTrue:
			[^false]].
	^true! !

!SpurMemoryManager methodsFor: 'primitive support' stamp: 'nice 11/8/2019 09:50'!
booleanObjectOf: aBool
	<inline: true>
	^aBool ifTrue: [trueObj] ifFalse: [falseObj]! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/19/2013 11:08'!
booleanValueOf: obj
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter booleanValueOf: obj! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 3/9/2016 16:55'!
bootstrapping
	<inline: true>
	^false! !

!SpurMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 9/24/2020 21:22'!
bridgeEdenAndOldSpace
	<inline: #always>
	self initSegmentBridgeWithBytes: oldSpaceStart - freeStart at: freeStart! !

!SpurMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 9/24/2020 22:05'!
bridgePastSpaceAndEden
	<inline: #always>
	pastSpaceStart < scavenger eden start ifTrue: "past space can be entirely full (!!!!)"
		[pastSpaceStart + self baseHeaderSize = scavenger eden start
			ifTrue: "No room for a full bridge (!!!!); use the slim bridge hack"
				[self hackSlimBridgeTo: (self objectStartingAt: scavenger eden start) at: pastSpaceStart.
				 "And carefully check the assumption"
				 self assert: (self objectAfterMaybeSlimBridge: (self objectInPastSpaceBefore: pastSpaceStart) limit: nilObj)
								= (self objectStartingAt: scavenger eden start)]
			ifFalse: "Room for a regular bridge; this is straight-forward"
				[self initSegmentBridgeWithBytes: scavenger eden start - pastSpaceStart at: pastSpaceStart]]! !

!SpurMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 9/24/2020 21:24'!
bridgePastSpaceAndOldSpace
	<inline: #always>
	self initSegmentBridgeWithBytes: oldSpaceStart - pastSpaceStart at: pastSpaceStart! !

!SpurMemoryManager methodsFor: 'segments' stamp: 'eem 10/10/2013 16:02'!
bridgeSize
	^2 * self baseHeaderSize! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 3/10/2014 12:17'!
byteFormatForNumBytes: numBytes
	^self firstByteFormat + (8 - numBytes bitAnd: self wordSize - 1)! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 9/10/2013 11:07'!
byteFormatMask
	^16r18! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 07/28/2014 08:07'!
byteSizeOf: oop
	<api>
	(self isImmediate: oop) ifTrue: [^0].
	^self numBytesOf: oop! !

!SpurMemoryManager methodsFor: 'indexing primitive support' stamp: 'eem 6/27/2014 16:09'!
byteSizeOfInstanceOf: classObj errInto: errBlock
	| instSpec classFormat numSlots |
	classFormat := self formatOfClass: classObj.
	instSpec := self instSpecOfClassFormat: classFormat.
	(self isFixedSizePointerFormat: instSpec) ifFalse:
		[^errBlock value: PrimErrBadReceiver]. "indexable"
	numSlots := self fixedFieldsOfClassFormat: classFormat.
	^self objectBytesForSlots: numSlots! !

!SpurMemoryManager methodsFor: 'indexing primitive support' stamp: 'eem 6/29/2017 18:05'!
byteSizeOfInstanceOf: classObj withIndexableSlots: nElements errInto: errorBlock
	| instSpec classFormat numSlots |
	<var: 'numSlots' type: #usqInt>
	classFormat := self formatOfClass: classObj.
	instSpec := self instSpecOfClassFormat: classFormat.
	instSpec caseOf: {
		[self arrayFormat]	->
			[numSlots := nElements].
		[self indexablePointersFormat]	->
			[numSlots := (self fixedFieldsOfClassFormat: classFormat) + nElements].
		[self weakArrayFormat]	->
			[numSlots := (self fixedFieldsOfClassFormat: classFormat) + nElements].
		[self sixtyFourBitIndexableFormat]	->
			[numSlots := self bytesPerOop = 4 ifTrue: [nElements * 2] ifFalse: [nElements]].
		[self firstLongFormat]	->
			[numSlots := self bytesPerOop = 4 ifTrue: [nElements] ifFalse: [nElements + 1 // 2]].
		[self firstShortFormat]	->
			[numSlots := self bytesPerOop = 4 ifTrue: [nElements + 1 // 2] ifFalse: [nElements + 3 // 4]].
		[self firstByteFormat]	->
			[numSlots := nElements + (self bytesPerOop - 1) // self bytesPerOop].
		[self firstCompiledMethodFormat]	-> "Assume nElements is derived from CompiledMethod>>basicSize."
			[numSlots := nElements + (self bytesPerOop - 1) // self bytesPerOop] }
		otherwise: [^errorBlock value: PrimErrBadReceiver]. "non-indexable"
	numSlots >= (1 asUnsignedInteger << (self bytesPerOop * 8 - self logBytesPerOop)) ifTrue:
		[^errorBlock value: (nElements < 0 ifTrue: [PrimErrBadArgument] ifFalse: [PrimErrLimitExceeded])].
	^self objectBytesForSlots: numSlots! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 10/14/2013 20:07'!
byteSwapped: w
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 06/22/2021 15:31'!
bytesInBody: objOop
	"Answer the total number of bytes in an object including header and possible overflow size header."
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 6/23/2017 12:52'!
bytesLeft: includeSwapSpace
	"Answer the amount of available free space. If includeSwapSpace is true, include
	 possibly available swap space. If includeSwapSpace is false, include possibly available
	 physical memory.  N.B. includeSwapSpace is ignored; answer total heap free space
	 minus the reserve available for flushing the tsack zone."
	^totalFreeOldSpace
	+ (scavenger eden limit - freeStart)
	+ (scavenger pastSpace limit - pastSpaceStart)
	+ (scavenger futureSpace limit - scavenger futureSpace limit)
	- coInterpreter interpreterAllocationReserveBytes! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 10/16/2013 13:17'!
bytesLeftInOldSpace
	"Answer the amount of available free old space.  Used by primitiveFullGC
	 to answer the current available memory."
	^totalFreeOldSpace! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 9/25/2020 11:08'!
bytesPerElement: oop
	"Answer the basic element size for the receiver.  Answer 0 for immediates and CompiledCode
	 (element size could be wordSize for literals or 1 for bytes, so its indeterminable).  Answer
	 wordSize for pointer objects.  Otherwise answer the actual element size of a bits container."
	<api>
	| fmt byteSizes |
	<var: #byteSizes declareC: 'unsigned char byteSizes[16] =
								   {	wordSize(),  wordSize(),  wordSize(),  wordSize(),
									wordSize(),  wordSize(),  wordSize(),  0, /* forwarder */ wordSize(),
									8,
									4, 4,
									2, 2, 2, 2 }'>
	(self isImmediate: oop) ifTrue:
		[^0].
	fmt := self formatOf: oop.
	fmt >= self firstByteFormat ifTrue:
		[fmt >= self firstCompiledMethodFormat ifTrue: [^0].
		 ^1].
	^(self cCode: [byteSizes] inSmalltalk:
			[CArrayAccessor
				on: {	self wordSize. self wordSize. self wordSize. self wordSize.
						self wordSize. self wordSize. self wordSize. 0. self wordSize.
						8.
						4. 4.
						2. 2. 2. 2 }])
		at: fmt! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 12/15/2014 14:12'!
bytesPerOop
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'indexing primitive support' stamp: 'eem 6/21/2021 11:23'!
canShorten: objOop
	^(self isNonImmediate: objOop)
	 and: [(self numSlotsOf: objOop) > 0
	 and: [(self isArrayNonImm: objOop) or: [self isWordsNonImm: objOop]]]! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 10/24/2013 11:24'!
capacityOfObjStack: objStack
	| total freeObj |
	objStack = nilObj ifTrue: [^0].
	total := ObjStackLimit negated.
	freeObj := objStack.
	[freeObj ~= 0] whileTrue:
		[total := total + ObjStackLimit.
		 freeObj := self fetchPointer: ObjStackFreex ofObject: freeObj].
	freeObj := objStack.
	[freeObj ~= 0] whileTrue:
		[total := total + ObjStackLimit.
		 freeObj := self fetchPointer: ObjStackNextx ofObject: freeObj].
	^total! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 10/14/2013 14:13'!
changeClassOf: rcvr to: argClass
	"Attempt to change the class of the receiver to the argument given that the
	 format of the receiver matches the format of the argument.  If successful,
	 answer 0, otherwise answer an error code indicating the reason for failure. 
	 Fail if the format of the receiver is incompatible with the format of the argument,
	 or if the argument is a fixed class and the receiver's size differs from the size
	 that an instance of the argument should have."
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 10/25/2013 15:01'!
characterObjectOf: characterCode
	<api>
	^characterCode << self numTagBits + self characterTag! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 11:48'!
characterTable
	self shouldNotImplement! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 12/13/2015 10:15'!
characterTag
	<api>
	<cmacro>
	^2! !

!SpurMemoryManager methodsFor: 'immediates' stamp: 'eem 10/14/2013 20:02'!
characterValueOf: oop
	"Immediate characters are unsigned"
	<api>
	^oop asUnsignedInteger >> self numTagBits! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 11/20/2013 13:29'!
cheapAddressCouldBeInHeap: address 
	^(address bitAnd: self wordSize - 1) = 0
	  and: [(self oop: address isGreaterThanOrEqualTo: newSpaceStart)
	  and: [self oop: address isLessThan: endOfMemory]]! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 11/25/2013 13:24'!
cheapIsInMemory: address
	"Answer if the given address is in ST object memory.  For simulation only."
	<doNotGenerate>
	^address >= newSpaceStart and: [address < endOfMemory]! !

!SpurMemoryManager methodsFor: 'initialization' stamp: 'eem 9/2/2013 17:52'!
checkCompactIndex: classIndex isClass: specialIndex named: name
	"Check that a class the VM assumes is compact has the right index."
	<inline: true> "macrofication of the name arg in invalidCompactClassError only works if this method is inlined so the name is a string literal not a parameter"
	(classIndex ~= 0
	 and: [(self splObj: specialIndex) ~= (self knownClassAtIndex: classIndex)]) ifTrue:
		[self invalidCompactClassError: name]! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 4/11/2017 15:50'!
checkForAvailableSlots: slots
	"Check for slots worth of free space being available.  Answer if that many slots are available.
	 If that many slots are not availabe, schedule a scavenge."
	<inline: true>
	freeStart + (self bytesPerOop * slots) <= scavengeThreshold ifTrue:
		[^true].
	needGCFlag := true.
	^false! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 06/22/2021 15:31'!
checkForCompactableObjects
	"self checkForCompactableObjects"
	<doNotGenerate>
	| firstFree them sizes |
	firstFree := 0.
	self allOldSpaceEntitiesDo: [:o| (firstFree = 0 and: [self isFreeObject: o]) ifTrue: [firstFree := o]].
	firstFree = 0 ifTrue: [^nil].
	sizes := Bag new.
	self allFreeObjectsDo:
		[:f| sizes add: (self bytesInBody: f)].
	them := OrderedCollection new.
	self allOldSpaceObjectsFrom: firstFree do:
		[:o| | b |
		b := self bytesInBody: o.
		(sizes includes: b) ifTrue:
			[them add: o.
			 sizes remove: b]].
	^them isEmpty ifFalse:
		[{them size. them first. them last}]! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 5/12/2016 11:29'!
checkForLeaks
	^checkForLeaks! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 1/24/2016 13:48'!
checkFreeSpace
	"Bootstrap only"
	<doNotGenerate>
	self checkFreeSpace: GCModeFull! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 6/21/2021 12:03'!
checkFreeSpace: gcModes
	<api>
	self assert: self bitsSetInFreeSpaceMaskForAllFreeLists.
	self assert: totalFreeOldSpace = self totalFreeListBytes.
	(gcModes > 0
	 and: [checkForLeaks allMask: (GCCheckFreeSpace bitOr: gcModes)]) ifTrue:
		[self runLeakCheckerForFreeSpace: GCCheckFreeSpace ignoring: nil]! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 6/21/2021 12:03'!
checkFreeSpace: gcModes ignoring: anOopToIgnore
	<api>
	self assert: self bitsSetInFreeSpaceMaskForAllFreeLists.
	self assert: totalFreeOldSpace = self totalFreeListBytes.
	(gcModes > 0
	 and: [checkForLeaks allMask: (GCCheckFreeSpace bitOr: gcModes)]) ifTrue:
		[self runLeakCheckerForFreeSpace: GCCheckFreeSpace ignoring: anOopToIgnore]! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 09/05/2021 19:12'!
checkHeapFreeSpaceIntegrity
	"Perform an integrity/leak check using the heapMap.  Assume clearLeakMapAndMapAccessibleFreeSpace
	 has set a bit at each free chunk's header.  Scan all objects in the heap checking that no pointer points
	 to a free chunk and that all free chunks that refer to others refer to marked chunks.  Answer if all checks pass."
	| ok total |
	<inline: false>
	<var: 'total' type: #usqInt>
	ok := true.
	total := 0.
	0 to: self numFreeLists - 1 do:
		[:i|
		(freeLists at: i) ~= 0 ifTrue:
			[(heapMap heapMapAtWord: (self pointerForOop: (freeLists at: i))) = 0 ifTrue:
				[coInterpreter print: 'leak in free list '; printNum: i; print: ' to non-free '; printHex: (freeLists at: i); eekcr.
				 ok := false]]].

	"Excuse the duplication but performance is at a premium and we avoid
	 some tests by splitting the newSpace and oldSpace enumerations."
	self allNewSpaceEntitiesDo:
		[:obj| | fieldOop |
		 (self isFreeObject: obj)
			ifTrue:
				[coInterpreter print: 'young object '; printHex: obj; print: ' is free'; eekcr.
				 ok := false]
			ifFalse:
				[obj ~= freeSpaceCheckOopToIgnore ifTrue:
					[0 to: (self numPointerSlotsOf: obj) - 1 do:
						[:fi|
						 fieldOop := self fetchPointer: fi ofObject: obj.
						 (self isNonImmediate: fieldOop) ifTrue:
							[(heapMap heapMapAtWord: (self pointerForOop: fieldOop)) ~= 0 ifTrue:
								[coInterpreter print: 'object leak in '; printHex: obj; print: ' @ '; printNum: fi; print: ' = '; printHex: fieldOop; print: ' is free'; eekcr.
								 ok := false]]]]]].
	self allOldSpaceEntitiesDo:
		[:obj| | fieldOop |
		(self isFreeObject: obj)
			ifTrue:
				[(heapMap heapMapAtWord: (self pointerForOop: obj)) = 0 ifTrue:
					[coInterpreter print: 'leak in free chunk '; printHex: obj; print: ' is unmapped?!! '; eekcr.
					 ok := false].
				 fieldOop := self fetchPointer: self freeChunkNextIndex ofFreeChunk: obj.
				 (fieldOop ~= 0
				 and: [(heapMap heapMapAtWord: (self pointerForOop: fieldOop)) = 0]) ifTrue:
					[coInterpreter print: 'leak in free chunk '; printHex: obj; print: ' @ 0 = '; printHex: fieldOop; print: ' is unmapped'; eekcr.
					 ok := false].
				(self isLilliputianSize: (self bytesInBody: obj)) ifFalse:
					[fieldOop := self fetchPointer: self freeChunkPrevIndex ofFreeChunk: obj.
					 (fieldOop ~= 0
					 and: [(heapMap heapMapAtWord: (self pointerForOop: fieldOop)) = 0]) ifTrue:
						[coInterpreter print: 'leak in free chunk '; printHex: obj; print: ' @ 0 = '; printHex: fieldOop; print: ' is unmapped'; eekcr.
						 ok := false]].
				(self isLargeFreeObject: obj) ifTrue:
					[self freeChunkParentIndex to: self freeChunkLargerIndex do:
						[:fi|
						 fieldOop := self fetchPointer: fi ofFreeChunk: obj.
						 (fieldOop ~= 0
						 and: [(heapMap heapMapAtWord: (self pointerForOop: fieldOop)) = 0]) ifTrue:
							[coInterpreter print: 'leak in free chunk '; printHex: obj; print: ' @ '; printNum: fi; print: ' = '; printHex: fieldOop; print: ' is unmapped'; eekcr.
							 ok := false]]].
				total := total + (self bytesInBody: obj)]
			ifFalse:
				[obj ~= freeSpaceCheckOopToIgnore ifTrue:
					[0 to: (self numPointerSlotsOf: obj) - 1 do:
						[:fi|
						 (self isForwarded: obj)
							ifTrue: 
								[self assert: fi = 0. "I'm now trying to use forwarders in GC algorithms..."
								 fieldOop := self fetchPointer: fi ofMaybeForwardedObject: obj] 
							ifFalse: "We keep #fetchPointer:ofObject: API here for assertions"
								[fieldOop := self fetchPointer: fi ofObject: obj].
						 (self isNonImmediate: fieldOop) ifTrue:
							[(heapMap heapMapAtWord: (self pointerForOop: fieldOop)) ~= 0 ifTrue:
								[coInterpreter print: 'object leak in '; printHex: obj; print: ' @ '; printNum: fi; print: ' = '; printHex: fieldOop; print: ' is free'; eekcr.
								 ok := false]]]]]].
	total ~= totalFreeOldSpace ifTrue:
		[coInterpreter print: 'incorrect totalFreeOldSpace; expected '; printNum: totalFreeOldSpace; print: ' found '; printNum: total; eekcr.
		 ok := false].
	^ok! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 09/05/2021 19:12'!
checkHeapIntegrity: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid
	"Perform an integrity/leak check using the heapMap.  Assume clearLeakMapAndMapAccessibleObjects
	 has set a bit at each (non-free) object's header.  Scan all objects in the heap checking that every
	 pointer points to a header.  Scan the rememberedSet, remapBuffer and extraRootTable checking
	 that every entry is a pointer to a header. Check that the number of roots is correct and that all
	 rememberedSet entries have their isRemembered: flag set.  Answer if all checks pass."
	| ok numRememberedObjectsInHeap |
	<inline: false>
	self cCode: []
		inSmalltalk:
			["Almost all of the time spent here used to go into the asserts in fetchPointer:ofObject: in the
			  simulator class overrides. Since we know here the indices used are valid we temporarily
			  remove them to claw back that performance."
			(self class whichClassIncludesSelector: #fetchPointer:ofObject:) ~= SpurMemoryManager ifTrue:
				[^self withSimulatorFetchPointerMovedAsideDo:
					[self checkHeapIntegrity: excludeUnmarkedObjs
						classIndicesShouldBeValid: classIndicesShouldBeValid]]].
	ok := true.
	numRememberedObjectsInHeap := 0.
	0 to: self numFreeLists - 1 do:
		[:i| | freeListOop |
		(freeListOop := freeLists at: i) ~= 0 ifTrue:
			[(self isInOldSpace: freeListOop)
				ifTrue:
					[(heapMap heapMapAtWord: (self pointerForOop: freeListOop)) ~= 0 ifTrue:
						[coInterpreter print: 'leak in free list '; printNum: i; print: ' to non-free '; printHex: freeListOop; eekcr.
						 ok := false]]
				ifFalse:
					[coInterpreter print: 'leak in free list '; printNum: i; print: ' to out-of-bounds '; printHex: freeListOop; eekcr.
					 ok := false]]].

	"Excuse the duplication but performance is at a premium and we avoid
	 some tests by splitting the newSpace and oldSpace enumerations."
	self allNewSpaceEntitiesDo:
		[:obj| | fieldOop classIndex classOop |
		(self isFreeObject: obj)
			ifTrue:
				[coInterpreter print: 'young object '; printHex: obj; print: ' is free'; eekcr.
				 ok := false]
			ifFalse:
				[((self isMarked: obj) not and: [excludeUnmarkedObjs]) ifFalse:
					[(self isRemembered: obj) ifTrue:
						[coInterpreter print: 'young object '; printHex: obj; print: ' is remembered'; eekcr.
						 ok := false]].
					 (self isForwarded: obj)
						ifTrue:
							[fieldOop := self fetchPointer: 0 ofMaybeForwardedObject: obj.
							(self isInHeapBounds: fieldOop)
								ifTrue:
									[(heapMap heapMapAtWord: (self pointerForOop: fieldOop)) = 0 ifTrue:
										[coInterpreter print: 'object leak in forwarder '; printHex: obj; print: ' to unmapped '; printHex: fieldOop; eekcr.
										 ok := false]]
								ifFalse:
									[coInterpreter print: 'object leak in forwarder '; printHex: obj; print: ' to out-of-bounds '; printHex: fieldOop; eekcr.
									 ok := false]]
						ifFalse:
							[classOop := self classOrNilAtIndex: (classIndex := self classIndexOf: obj).
							 (classIndicesShouldBeValid
							  and: [classOop = nilObj
							  and: [(self isHiddenObj: obj) not]]) ifTrue:
								[coInterpreter print: 'object leak in '; printHex: obj; print: ' invalid class index '; printHex: classIndex; print: ' -> '; print: (classOop ifNil: ['nil'] ifNotNil: ['nilObj']); eekcr.
								 ok := false].
							 0 to: (self numPointerSlotsOf: obj) - 1 do:
								[:fi|
								 fieldOop := self fetchPointer: fi ofObject: obj.
								 (self isNonImmediate: fieldOop) ifTrue:
									[(self isInHeapBounds: fieldOop)
										ifTrue:
											[(heapMap heapMapAtWord: (self pointerForOop: fieldOop)) = 0 ifTrue:
												[coInterpreter print: 'object leak in '; printHex: obj; print: ' @ '; printNum: fi; print: ' = '; printHex: fieldOop; eekcr.
												 ok := false]]
										ifFalse:
											[coInterpreter print: 'object leak in '; printHex: obj; print: ' to out-of-bounds '; printHex: fieldOop; eekcr.
											 ok := false]]]]]].
	self allOldSpaceEntitiesDo:
		[:obj| | containsYoung fieldOop classIndex classOop |
		(self isFreeObject: obj)
			ifTrue:
				[(heapMap heapMapAtWord: (self pointerForOop: obj)) ~= 0 ifTrue:
					[coInterpreter print: 'leak in free chunk '; printHex: obj; print: ' is mapped?!! '; eekcr.
					 ok := false].
				 fieldOop := self fetchPointer: self freeChunkNextIndex ofFreeChunk: obj.
				 (fieldOop ~= 0
				 and: [(heapMap heapMapAtWord: (self pointerForOop: fieldOop)) ~= 0]) ifTrue:
					[coInterpreter print: 'leak in free chunk '; printHex: obj; print: ' @ 0 = '; printHex: fieldOop; print: ' is mapped'; eekcr.
					 ok := false].
				(self isLilliputianSize: (self bytesInBody: obj)) ifFalse:
					[fieldOop := self fetchPointer: self freeChunkPrevIndex ofFreeChunk: obj.
					 (fieldOop ~= 0
					 and: [(heapMap heapMapAtWord: (self pointerForOop: fieldOop)) ~= 0]) ifTrue:
						[coInterpreter print: 'leak in free chunk '; printHex: obj; print: ' @ 1 = '; printHex: fieldOop; print: ' is mapped'; eekcr.
						 ok := false]].
				(self isLargeFreeObject: obj) ifTrue:
					[self freeChunkParentIndex to: self freeChunkLargerIndex do:
						[:fi|
						 fieldOop := self fetchPointer: fi ofFreeChunk: obj.
						 (fieldOop ~= 0
						 and: [(heapMap heapMapAtWord: (self pointerForOop: fieldOop)) ~= 0]) ifTrue:
							[coInterpreter print: 'leak in free chunk '; printHex: obj; print: ' @ '; printNum: fi; print: ' = '; printHex: fieldOop; print: ' is mapped'; eekcr.
							 ok := false].]]]
			ifFalse:
				[(excludeUnmarkedObjs and: [(self isMarked: obj) not]) ifFalse:
					[containsYoung := false.
					 (self isRemembered: obj) ifTrue:
						[numRememberedObjectsInHeap := numRememberedObjectsInHeap + 1.
						 (scavenger isInRememberedSet: obj) ifFalse:
							[coInterpreter print: 'remembered object '; printHex: obj; print: ' is not in remembered table'; eekcr.
							 ok := false]].
					 (self isForwarded: obj)
						ifTrue:
							[fieldOop := self fetchPointer: 0 ofMaybeForwardedObject: obj.
							(self isInHeapBounds: fieldOop)
								ifTrue:
									[(heapMap heapMapAtWord: (self pointerForOop: fieldOop)) = 0 ifTrue:
										[coInterpreter print: 'object leak in forwarder '; printHex: obj; print: ' to unmapped '; printHex: fieldOop; eekcr.
										 ok := false]]
								ifFalse:
									[coInterpreter print: 'object leak in forwarder '; printHex: obj; print: ' to out-of-bounds '; printHex: fieldOop; eekcr.
									 ok := false].
							 (self isReallyYoung: fieldOop) ifTrue:
								[containsYoung := true]]
						ifFalse:
							[classOop := self classOrNilAtIndex: (classIndex := self classIndexOf: obj).
							 (classIndicesShouldBeValid
							  and: [classOop = nilObj
							  and: [classIndex > self lastClassIndexPun]]) ifTrue:
								[coInterpreter print: 'object leak in '; printHex: obj; print: ' invalid class index '; printHex: classIndex; print: ' -> '; print: (classOop ifNil: ['nil'] ifNotNil: ['nilObj']); eekcr.
								 ok := false].
							 0 to: (self numPointerSlotsOf: obj) - 1 do:
								[:fi|
								 fieldOop := self fetchPointer: fi ofObject: obj.
								 (self isNonImmediate: fieldOop) ifTrue:
									[(self isInHeapBounds: fieldOop)
										ifTrue:
											[(heapMap heapMapAtWord: (self pointerForOop: fieldOop)) = 0 ifTrue:
												[coInterpreter print: 'object leak in '; printHex: obj; print: ' @ '; printNum: fi; print: ' = '; printHex: fieldOop; eekcr.
												 ok := false]]
										ifFalse:
											[coInterpreter print: 'object leak in '; printHex: obj; print: ' to out-of-bounds '; printHex: fieldOop; eekcr.
											 ok := false].
									 "don't be misled by CogMethods; they appear to be young, but they're not"
									 (self isReallyYoung: fieldOop) ifTrue:
										[containsYoung := true]]]].
					 containsYoung ifTrue:
						[(self isRemembered: obj) ifFalse:
							[coInterpreter print: 'unremembered object '; printHex: obj; print: ' contains young oop(s)'; eekcr.
							 ok := false]]]]].
	numRememberedObjectsInHeap ~= scavenger rememberedSetSize ifTrue:
		[coInterpreter
			print: 'root count mismatch. #heap roots ';
			printNum: numRememberedObjectsInHeap;
			print: '; #roots ';
			printNum: scavenger rememberedSetSize;
			eekcr.
		"But the system copes with overflow..."
		self flag: 'no support for remembered set overflow yet'.
		"ok := rootTableOverflowed and: [needGCFlag]"].
	scavenger rememberedSetWithIndexDo:
		[:obj :i|
		(obj bitAnd: self wordSize - 1) ~= 0
			ifTrue:
				[coInterpreter print: 'misaligned oop in remembered set @ '; printNum: i; print: ' = '; printHex: obj; eekcr.
				 ok := false]
			ifFalse:
				[(heapMap heapMapAtWord: (self pointerForOop: obj)) = 0
					ifTrue:
						[coInterpreter print: 'object leak in remembered set @ '; printNum: i; print: ' = '; printHex: obj; eekcr.
						 ok := false]
					ifFalse:
						[(self isYoung: obj) ifTrue:
							[coInterpreter print: 'non-root in remembered set @ '; printNum: i; print: ' = '; printHex: obj; eekcr.
							 ok := false]]]].
	self objStack: mournQueue do:
		[:i :page| | obj |
		obj := self fetchPointer: i ofObject: page.
		(obj bitAnd: self wordSize - 1) ~= 0
			ifTrue:
				[coInterpreter print: 'misaligned oop in mournQueue @ '; printNum: i; print: ' in '; printHex: page; print: ' = '; printHex: obj; eekcr.
				 ok := false]
			ifFalse:
				[(excludeUnmarkedObjs and: [(self isMarked: obj) not]) ifFalse:
					[(heapMap heapMapAtWord: (self pointerForOop: obj)) = 0 ifTrue:
						[coInterpreter print: 'object leak in mournQueue @ '; printNum: i; print: ' in '; printHex: page; print: ' = '; printHex: obj; eekcr.
						 ok := false]]]].
	1 to: remapBufferCount do:
		[:ri| | obj |
		obj := remapBuffer at: ri.
		(obj bitAnd: self wordSize - 1) ~= 0
			ifTrue:
				[coInterpreter print: 'misaligned remapRoot @ '; printNum: ri; print: ' = '; printHex: obj; eekcr.
				 ok := false]
			ifFalse:
				[(heapMap heapMapAtWord: (self pointerForOop: obj)) = 0 ifTrue:
					[coInterpreter print: 'object leak in remapRoots @ '; printNum: ri; print: ' = '; printHex: obj; eekcr.
					 ok := false]]].
	1 to: extraRootCount do:
		[:ri| | obj |
		obj := (extraRoots at: ri) at: 0.
		(obj bitAnd: self wordSize - 1) ~= 0
			ifTrue:
				[coInterpreter print: 'misaligned extraRoot @ '; printNum: ri; print: ' => '; printHex: obj; eekcr.
				 ok := false]
			ifFalse:
				[(heapMap heapMapAtWord: (self pointerForOop: obj)) = 0 ifTrue:
					[coInterpreter print: 'object leak in extraRoots @ '; printNum: ri; print: ' => '; printHex: obj; eekcr.
					 ok := false]]].
	^ok! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 11/20/2013 13:59'!
checkMemoryMap
	self assert: (self isYoungObject: newSpaceStart).
	self assert: (self isYoungObject: newSpaceLimit - self wordSize).
	self assert: (self isOldObject: newSpaceStart) not.
	self assert: (self isOldObject: newSpaceLimit - self wordSize) not.
	self assert: (self isYoungObject: newSpaceLimit) not.
	self assert: (self isYoungObject: oldSpaceStart) not.
	self assert: (self isYoungObject: endOfMemory) not.
	self assert: (self isOldObject: newSpaceLimit).
	self assert: (self isOldObject: oldSpaceStart).
	self assert: (self isOldObject: endOfMemory)! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 3/21/2015 10:45'!
checkOkayOop: oop
	"Verify that the given oop is legitimate. Check address, header, and size but not class.
	 Answer true if OK.  Otherwise print reason and answer false."
	<api>
	<var: #oop type: #usqInt>
	| classIndex fmt unusedBits unusedBitsInYoungObjects |
	<var: #unusedBits type: #usqLong>

	"address and size checks"
	(self isImmediate: oop) ifTrue: [^true].
	(self addressCouldBeObj: oop) ifFalse:
		[self print: 'oop '; printHex: oop; print: ' is not a valid address'. ^false].

	(self oop: (self addressAfter: oop) isLessThanOrEqualTo: endOfMemory) ifFalse:
		[self print: 'oop '; printHex: oop; print: ' size would make it extend beyond the end of memory'. ^false].

	"header type checks"
	(classIndex := self classIndexOf: oop) >= self firstClassIndexPun ifFalse:
		[self print: 'oop '; printHex: oop; print: ' is a free chunk, or bridge, not an object'. ^false].
	((self rawNumSlotsOf: oop) = self numSlotsMask
	 and: [(self rawNumSlotsOf: oop - self baseHeaderSize) ~= self numSlotsMask]) ifTrue:
		[self print: 'oop '; printHex: oop; print: ' header has overflow header word, but overflow word does not have a saturated numSlots field'. ^false].

	"format check"
	fmt := self formatOf: oop.
	(fmt = 6) | (fmt = 8) ifTrue:
		[self print: 'oop '; printHex: oop; print: ' has an unknown format type'. ^false].
	(fmt = self forwardedFormat) ~= (classIndex = self isForwardedObjectClassIndexPun) ifTrue:
		[self print: 'oop '; printHex: oop; print: ' has mis-matched format/classIndex fields; only one of them is the isForwarded value'. ^false].

	"specific header bit checks"
	unusedBits := (1 << self classIndexFieldWidth)
				   bitOr: (1 << (self identityHashFieldWidth + 32)).
	((self long64At: oop) bitAnd: unusedBits) ~= 0 ifTrue:
		[self print: 'oop '; printHex: oop; print: ' has some unused header bits set; should be zero'. ^false].

	unusedBitsInYoungObjects := self newSpaceRefCountMask.
	((self longAt: oop) bitAnd: unusedBitsInYoungObjects) ~= 0 ifTrue:
		[self print: 'oop '; printHex: oop; print: ' has some header bits unused in young objects set; should be zero'. ^false].
	^true! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 10/15/2013 00:25'!
checkOkayYoungReferrer: obj
	"Verify that the given obj is a valid youngReferrer. Check remembered is set and
	 is in remembered set.  Answer true if OK.  Otherwise print reason and answer false.
	 Assumes the object contains young references."

	(self oop: obj isLessThan: newSpaceLimit) ifTrue:
		[^true].

	(self isRemembered: obj) ifFalse:
		[ self print: 'remembered bit is not set in '; printHex: obj; cr. ^false ].

	(scavenger isInRememberedSet: obj) ifTrue: [^true].

	self printHex: obj; print: ' has remembered bit set but is not in remembered set'; cr.

	^false
! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 10/31/2013 08:10'!
checkOopHasOkayClass: obj
	"Attempt to verify that the given obj has a reasonable behavior. The class must be a
	 valid, non-integer oop and must not be nilObj. It must be a pointers object with three
	 or more fields. Finally, the instance specification field of the behavior must match that
	 of the instance. If OK answer true.  If  not, print reason and answer false."

	<api>
	<var: #obj type: #usqInt>
	| objClass objFormat |
	<var: #objClass type: #usqInt>

	(self checkOkayOop: obj) ifFalse:
		[^false].
	objClass := self cCoerce: (self fetchClassOfNonImm: obj) to: #usqInt.

	(self isImmediate: objClass) ifTrue:
		[self print: 'obj '; printHex: obj; print: ' an immediate is not a valid class or behavior'; cr. ^false].
	(self okayOop: objClass) ifFalse:
		[self print: 'obj '; printHex: obj; print: ' class obj is not ok'; cr. ^false].
	((self isPointersNonImm: objClass) and: [(self numSlotsOf: objClass) >= 3]) ifFalse:
		[self print: 'obj '; printHex: obj; print: ' a class (behavior) must be a pointers object of size >= 3'; cr. ^false].
	objFormat := (self isBytes: obj)
						ifTrue: [(self formatOf: obj) bitClear: 7]  "ignore extra bytes size bits"
						ifFalse: [self formatOf: obj].

	(self instSpecOfClass: objClass) ~= objFormat ifTrue:
		[self print: 'obj '; printHex: obj; print: ' and its class (behavior) formats differ'; cr. ^false].
	^true! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 09/05/2021 19:12'!
checkOopIntegrity: obj named: name
	<inline: false>
	<var: #name type: #'char *'>
	((self oop: obj isLessThan: endOfMemory)
	 and: [(heapMap heapMapAtWord: (self pointerForOop: obj)) ~= 0]) ifTrue:
		[^true].
	coInterpreter print: name; print: ' leak '; printHex: obj; eekcr.
	^false! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 09/05/2021 19:12'!
checkOopIntegrity: obj named: name index: i
	<inline: false>
	<var: #name type: #'char *'>
	((self oop: obj isLessThan: endOfMemory)
	 and: [(heapMap heapMapAtWord: (self pointerForOop: obj)) ~= 0]) ifTrue:
		[^true].
	coInterpreter print: name; print: ' leak @ '; printNum: i; print: ' = '; printHex: obj; eekcr.
	^false! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 10/11/2013 16:02'!
checkedIntegerValueOf: intOop
	<doNotGenerate>
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	^coInterpreter checkedIntegerValueOf: intOop! !

!SpurMemoryManager methodsFor: 'memory access' stamp: 'eem 11/19/2013 17:57'!
checkedLongAt: byteAddress
	"Assumes zero-based array indexing."
	<api>
	(self isInMemory: byteAddress) ifFalse:
		[self warning: 'checkedLongAt bad address'.
		 coInterpreter primitiveFail].
	^self longAt: byteAddress! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 11:24'!
classAlien
	^self splObj: ClassAlien! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 12/16/2015 15:04'!
classArray
	<api>
	^self splObj: ClassArray! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 1/31/2017 11:03'!
classAtIndex: classIndex
	<api>
	<inline: true>
	| classTablePage |
	self assert: (classIndex >= 0 and: [classIndex <= self tagMask or: [classIndex >= self arrayClassIndexPun and: [classIndex <= self classIndexMask]]]).
	classTablePage := self fetchPointer: classIndex >> self classTableMajorIndexShift
							ofObject: hiddenRootsObj.
	classTablePage = nilObj ifTrue:
		[^nil].
	^self
		fetchPointer: (classIndex bitAnd: self classTableMinorIndexMask)
		ofObject: classTablePage! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 7/23/2014 11:32'!
classAtIndex: classIndex put: objOop
	"for become & GC of classes"
	| classTablePage |
	self assert: (classIndex <= self tagMask or: [classIndex >= self arrayClassIndexPun]).
	self assert: (objOop = nilObj
				 or: [((self rawHashBitsOf: objOop) = classIndex						"normal entry"
					  or: [(self classAtIndex: (self rawHashBitsOf: objOop)) = objOop])	"become forces duplicate entry"
					and: [coInterpreter objCouldBeClassObj: objOop]]).
	classTablePage := self fetchPointer: classIndex >> self classTableMajorIndexShift
							ofObject: hiddenRootsObj.
	classTablePage = nilObj ifTrue:
		[self error: 'attempt to add class to empty page'].
	^self
		storePointer: (classIndex bitAnd: self classTableMinorIndexMask)
		ofObject: classTablePage
		withValue: objOop! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 9/20/2013 18:14'!
classBitmap
	^self splObj: ClassBitmap! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 9/20/2013 15:36'!
classByteArray
	"a.k.a. self fetchPointer: ClassByteArrayCompactIndex ofObject: classTableFirstPage"
	^self splObj: ClassByteArray! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 10/14/2013 19:24'!
classCharacter
	^self splObj: ClassCharacter! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 1/16/2020 20:04'!
classDoubleByteArray
	^self accessIntegerClass: ClassDoubleByteArray withValidationFlag: ValidatedClassDoubleByteArrayFlag! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 1/16/2020 20:04'!
classDoubleWordArray
	^self accessIntegerClass: ClassDoubleWordArray withValidationFlag: ValidatedClassDoubleWordArrayFlag! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 11:25'!
classExternalAddress
	^self splObj: ClassExternalAddress! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 11:25'!
classExternalData
	^self splObj: ClassExternalData! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 11:25'!
classExternalFunction
	^self splObj: ClassExternalFunction! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 12:50'!
classExternalLibrary
	^self splObj: ClassExternalLibrary! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 19:30'!
classExternalStructure
	^self splObj: ClassExternalStructure! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 10/14/2013 19:24'!
classFloat
	^self splObj: ClassFloat! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 1/16/2020 20:17'!
classFloat32Array
	^nilObj
	"should be this but we haven't assigned unused specialObjectArray indices yet
		^self accessIntegerClass: ClassFloat32Array withValidationFlag: ValidatedClassFloat32ArrayFlag"! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 1/16/2020 20:17'!
classFloat64Array
	^nilObj
	"should be this but we haven't assigned unused specialObjectArray indices yet
		^self accessIntegerClass: ClassFloat64Array withValidationFlag: ValidatedClassFloat64ArrayFlag"! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 6/21/2015 13:49'!
classForClassTag: classIndex
	<inline: true>
	"No need to check this; classAtIndex: has a stricter assert:
		self assert: classIndex ~= self isForwardedObjectClassIndexPun."
	^self classAtIndex: classIndex! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 1/10/2017 15:59'!
classFormatFromInstFormat: instFormat
	"0 = 0 sized objects (UndefinedObject True False et al)
	 1 = non-indexable objects with inst vars (Point et al)
	 2 = indexable objects with no inst vars (Array et al)
	 3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
	 4 = weak indexable objects with inst vars (WeakArray et al)
	 5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
	 6 unused, reserved for exotic pointer objects?
	 7 Forwarded Object, 1st field is pointer, rest of fields are ignored
	 8 unused, reserved for exotic non-pointer objects?
	 9 64-bit indexable
	 10 - 11 32-bit indexable
	 12 - 15 16-bit indexable
	 16 - 23 byte indexable
	 24 - 31 compiled method"
	<inline: true>
	instFormat >= self firstByteFormat ifTrue: "this is likely the common case"
		[^instFormat bitAnd: -8].
	instFormat <= self sixtyFourBitIndexableFormat ifTrue:
		[^instFormat].
	instFormat < self firstShortFormat ifTrue:
		[^instFormat bitAnd: -2].
	^instFormat bitAnd: -4! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 10/8/2013 09:37'!
classIndexFieldWidth
	"22-bit class mask => ~ 4M classes"
	^22! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 12/12/2015 14:18'!
classIndexMask
	<api>
	<cmacro>
	"22-bit class mask => ~ 4M classes"
	^16r3fffff! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 10/26/2016 17:16'!
classIndexOf: objOop
	"Answer objOop's classIndex from the object header.
	 A note on performance:
		This routine uses longAt:, which does a 32-bit load on the 32-bit system, and a 64-bit load
		on the 64-bit system.  Since the only 64-bit implementation(s) is/are little-endian, and
		since all of isGrey,isPinned,isRemembered,format,isUmmutable and classIndex fit within
		the least significant 32-bits we could use long32At: to access these, in the hope that the
		32-bit access will be quicker on 64-bits by virtue of fetching half the bits.  But experiments
		show that doing this does not produce any increase; at least any signal is lost in the noise."
	<api>
	^(self longAt: objOop) bitAnd: self classIndexMask! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 9/11/2013 07:32'!
classIndexOfHeader: aHeader
	<inline: true>
	^aHeader bitAnd: self classIndexMask! !

!SpurMemoryManager methodsFor: 'class table puns' stamp: 'eem 3/9/2016 14:48'!
classIsItselfClassIndexPun
	"Class puns are class indices not used by any class.  There is an entry
	 for the pun that refers to the notional class of objects with this class
	 index.  But because the index doesn't match the class it won't show up
	 in allInstances, hence hiding the object with a pun as its class index.
	 The puns occupy indices 16 through 31."
	<cmacro>
	^31! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 9/19/2013 11:16'!
classLargeNegativeInteger
	^self knownClassAtIndex: ClassLargeNegativeIntegerCompactIndex! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 9/19/2013 11:16'!
classLargePositiveInteger
	^self knownClassAtIndex: ClassLargePositiveIntegerCompactIndex! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 9/22/2013 09:06'!
classMutex
	^self splObj: ClassMutex! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 2/26/2014 15:21'!
classOrNilAtIndex: classIndex
	<api>
	| classTablePage |
	self assert: (classIndex <= self tagMask or: [classIndex >= self arrayClassIndexPun]).
	classTablePage := self fetchPointer: classIndex >> self classTableMajorIndexShift
							ofObject: hiddenRootsObj.
	classTablePage = nilObj ifTrue:
		[^nilObj].
	^self
		fetchPointer: (classIndex bitAnd: self classTableMinorIndexMask)
		ofObject: classTablePage! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 9/20/2013 18:08'!
classPoint
	^self splObj: ClassPoint! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 9/22/2013 09:06'!
classSemaphore
	^self splObj: ClassSemaphore! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 9/13/2014 15:09'!
classSmallInteger
	<api>
	^self splObj: ClassSmallInteger! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 12:05'!
classString
	^self splObj: ClassByteString! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 9/24/2014 11:41'!
classTableEntriesDo: binaryBlock
	"Evaluate binaryBlock with all non-nil entries in the classTable and their index."
	<inline: true>
	0 to: numClassTablePages - 1 do:
		[:i| | page |
		 page := self fetchPointer: i ofObject: hiddenRootsObj.
		 0 to: self classTablePageSize - 1 do:
			[:j| | classOrNil |
			classOrNil := self fetchPointer: j ofObject: page.
			classOrNil ~= nilObj ifTrue:
				[binaryBlock
					value: classOrNil
					value: i << self classTableMajorIndexShift + j]]]! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 12/16/2016 16:55'!
classTableFirstPage
	<cmacro>
	^classTableFirstPage! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 9/11/2013 06:01'!
classTableIndex
	^classTableIndex! !

!SpurMemoryManager methodsFor: 'spur bootstrap' stamp: 'eem 9/11/2013 06:22'!
classTableIndex: n
	classTableIndex := n! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 12/12/2015 14:18'!
classTableMajorIndexShift
	<api>
	<cmacro>
	"1024 entries per page (2^10); 22 bit classIndex implies 2^12 pages"
	^10! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 11/18/2013 13:43'!
classTableMinorIndexMask
	<api>
	"1024 entries per page (2^10); 22 bit classIndex implies 2^12 pages"
	"self basicNew classTableMinorIndexMask"
	^1 << self classTableMajorIndexShift - 1! !

!SpurMemoryManager methodsFor: 'spur bootstrap' stamp: 'eem 10/17/2013 05:49'!
classTableObjectsDo: aBlock
	"for the bootstrap..."
	<doNotGenerate>
	0 to: self classTableRootSlots - 1 do:
		[:i| | page |
		page := self fetchPointer: i ofObject: hiddenRootsObj.
		0 to: (self numSlotsOf: page) - 1 do:
			[:j| | classOrNil |
			classOrNil := self fetchPointer: j ofObject: page.
			classOrNil ~= nilObj ifTrue:
				[aBlock value: classOrNil]]]! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 8/7/2014 04:41'!
classTablePageSize
	"1024 entries per page (2^10); 22 bit classIndex implies 2^12 pages"
	"self basicNew classTablePageSize"
	<api>
	^1 << self classTableMajorIndexShift! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 11/18/2013 13:43'!
classTableRootObj
	<api>
	"For Cogit & bootstrap"
	^hiddenRootsObj! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 10/17/2013 05:46'!
classTableRootSlots
	"Answer the number of slots for class table pages in the hidden root object."
	^1 << (self classIndexFieldWidth - self classTableMajorIndexShift)! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'nice 4/27/2016 20:22'!
classTagForClass: classObj
	"Answer the classObj's identityHash to use as a tag in the first-level method lookup cache."
	<api>
	^self ensureBehaviorHash: classObj! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 9/11/2013 20:42'!
classTagForSpecialObjectsIndex: splObjIndex compactClassIndex: compactClassIndex
	"Answer the compactClassIndex to use as a tag in the first-level method lookup cache."
	^compactClassIndex! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 11:25'!
classUnsafeAlien
	^self splObj: ClassUnsafeAlien! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 1/16/2020 20:05'!
classWordArray
	^self accessIntegerClass: ClassWordArray withValidationFlag: ValidatedClassWordArrayFlag! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 1/24/2016 13:17'!
clearLeakMapAndMapAccessibleFreeSpace
	"Perform an integrity/leak check using the heapMap.  Set a bit at each free chunk's header."
	<inline: false>
	heapMap clearHeapMap.
	self allOldSpaceEntitiesFrom: self firstObject
		do: [:objOop|
			(self isFreeObject: objOop) ifTrue:
				[heapMap heapMapAtWord: (self pointerForOop: objOop) Put: 1]]! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 9/16/2013 10:43'!
clearLeakMapAndMapAccessibleObjects
	"Perform an integrity/leak check using the heapMap.  Set a bit at each object's header."
	<inline: false>
	heapMap clearHeapMap.
	self allObjectsDo:
		[:oop| heapMap heapMapAtWord: (self pointerForOop: oop) Put: 1]! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 8/31/2017 10:11'!
cleverSwapHeaders: obj1 and: obj2 copyHashFlag: copyHashFlag
	"swap headers, but swapping headers swaps remembered bits and hashes;
	 remembered bits must be unswapped and hashes may be unswapped if
	 copyHash is false."
	"This variant doesn't tickle a compiler bug in gcc and clang.  See naiveSwapHeaders:and:copyHashFlag:"
	<inline: true>
	| header1 header2 remembered |
	header1 := self long64At: obj1.
	header2 := self long64At: obj2.
	remembered := (header1 bitXor: header2) bitAnd: 1 << self rememberedBitShift.
	remembered ~= 0 ifTrue:
		[header1 := header1 bitXor: remembered.
		 header2 := header2 bitXor: remembered].
	"swapping headers swaps hash; if not copyHashFlag then unswap hash"
	copyHashFlag ifFalse:
		[| hashBits |
		 hashBits := (header1 bitXor: header2) bitAnd: self identityHashFullWordMask.
		 hashBits ~= 0 ifTrue:
			[header1 := header1 bitXor: hashBits.
			 header2 := header2 bitXor: hashBits]].
	self long64At: obj1 put: header2.
	self long64At: obj2 put: header1! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 1/10/2018 23:54'!
cloneInOldSpace: objOop forPinning: forPinning
	<inline: false>
	| numSlots fmt newObj hash |
	numSlots := self numSlotsOf: objOop.
	fmt := self formatOf: objOop.
	
	forPinning
		ifTrue:
			[newObj := self allocateSlotsForPinningInOldSpace: numSlots
							bytes: (self objectBytesForSlots: numSlots)
							format: fmt
							classIndex: (self classIndexOf: objOop)]
		ifFalse:
			[newObj := self allocateSlotsInOldSpace: numSlots
							bytes: (self objectBytesForSlots: numSlots)
							format: fmt
							classIndex: (self classIndexOf: objOop)].
	newObj ifNil:
		[^0].
	(self isPointersFormat: fmt)
		ifTrue:
			[| hasYoung |
			 hasYoung := false.
			 0 to: numSlots - 1 do:
				[:i| | oop |
				oop := self fetchPointer: i ofObject: objOop.
				((self isNonImmediate: oop)
				 and: [self isForwarded: oop]) ifTrue:
					[oop := self followForwarded: oop].
				((self isNonImmediate: oop)
				 and: [self isYoungObject: oop]) ifTrue:
					[hasYoung := true].
				self storePointerUnchecked: i
					ofObject: newObj
					withValue: oop].
			hasYoung ifTrue:
				[scavenger remember: newObj]]
		ifFalse:
			[0 to: numSlots - 1 do:
				[:i|
				self storePointerUnchecked: i
					ofObject: newObj
					withValue: (self fetchPointer: i ofObject: objOop)].
			 fmt >= self firstCompiledMethodFormat ifTrue:
				[coInterpreter maybeFixClonedCompiledMethod: newObj.
				 ((self isYoungObject: objOop) or: [self isRemembered: objOop]) ifTrue:
					[scavenger remember: newObj]]].
	(hash := self rawHashBitsOf: objOop) ~= 0 ifTrue:
		[self setHashBitsOf: newObj to: hash].
	(self isObjImmutable: objOop) ifTrue:
		[self setIsImmutableOf: newObj to: true].
	^newObj! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 01/07/2021 10:40'!
cloneObject: objOop
	| numSlots fmt newObj |
	numSlots := self numSlotsOf: objOop.
	fmt := self formatOf: objOop.
	numSlots > self maxSlotsForNewSpaceAlloc
		ifTrue:
			[newObj := self allocateSlotsInOldSpace: numSlots
							format: fmt
							classIndex: (self classIndexOf: objOop)]
		ifFalse:
			[newObj := self allocateSlots: numSlots
							format: fmt
							classIndex: (self classIndexOf: objOop)].
	newObj ifNil:
		[^0].
	(self isPointersFormat: fmt)
		ifTrue:
			[| hasYoung |
			 hasYoung := false.
			 0 to: numSlots - 1 do:
				[:i| | oop |
				oop := self fetchPointer: i ofObject: objOop.
				(self isNonImmediate: oop) ifTrue:
					[(self isForwarded: oop) ifTrue:
						[oop := self followForwarded: oop].
					((self isNonImmediate: oop)
					 and: [self isYoungObject: oop]) ifTrue:
						[hasYoung := true]].
				self storePointerUnchecked: i
					ofObject: newObj
					withValue: oop].
			(hasYoung
			 and: [(self isYoungObject: newObj) not]) ifTrue:
				[scavenger remember: newObj]]
		ifFalse:
			[0 to: numSlots - 1 do:
				[:i|
				self storePointerUnchecked: i
					ofObject: newObj
					withValue: (self fetchPointer: i ofObject: objOop)].
			 fmt >= self firstCompiledMethodFormat ifTrue:
				[coInterpreter maybeFixClonedCompiledMethod: newObj.
				 ((self isOldObject: newObj)
				  and: [(self isYoungObject: objOop) or: [self isRemembered: objOop]]) ifTrue:
					[scavenger remember: newObj]]].
	^newObj! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 9/18/2013 16:47'!
coInterpreter
	<doNotGenerate>
	^coInterpreter! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 12/16/2016 16:29'!
coInterpreter: aCoInterpreter
	<doNotGenerate>
	coInterpreter := aCoInterpreter.
	scavenger ifNotNil:
		[scavenger coInterpreter: aCoInterpreter].
	compactor ifNotNil:
		[compactor coInterpreter: aCoInterpreter]! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 10/21/2013 13:05'!
coalesce: obj1 and: obj2
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 4/24/2014 12:11'!
coallesceFreeChunk: objOop
	"Attempt to coallesce objOop with the following objects in memory.
	 Answer the possibly changed start of objOop after coallescing."
	| here next |
	here := objOop.
	self assert: (self isRemembered: here) not.
	"Because lastBridge is marked loop below will terminate on reaching lastBridge."
	self assert: (self isMarked: segmentManager lastBridge).
	next := self objectAfter: here limit: endOfMemory.
	[self isMarked: next] whileFalse: "coalescing; rare case"
		[self assert: (self isRemembered: next) not.
		 statCoalesces := statCoalesces + 1.
		 here := self coalesce: here and: next.
		 next := self objectAfter: here limit: endOfMemory].
	^here! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 11/18/2013 13:43'!
compactClassIndexOf: objOop
	<api>
	<inline: true>
	^self classIndexOf: objOop! !

!SpurMemoryManager methodsFor: 'class membership' stamp: 'eem 9/11/2013 05:07'!
compactIndexOfClass: objOop
	self assert: (self rawHashBitsOf: objOop) ~= 0.
	^self rawHashBitsOf: objOop! !

!SpurMemoryManager methodsFor: 'compaction - analysis' stamp: 'eem 2/13/2014 11:50'!
compactionIssues
	<doNotGenerate>
	"Compaction isn't working well.  It rarely moves more than a few tens of kilobytes.  Why?
	 Load an image and before you run it, or just before a GC, run these anaylsis routines.
	 e.g.
		self abstractCompaction #(63230272 75456 63210648)
	 shows we can move 75456 bytes of objects, but that would use 63210648 of free space.
	 i.e. there are lots of big free chunks in play, not many small ones that fit the bill.

		self largeFreeChunkDistribution
			#(	#('16r31C788' #nil)
				#('16r1423AC0' 2061864)
				#('16r1B705E8' 1515200)
				#('16r1D31D20' 2011152)
				#('16r1F37818' 1491480)
				#('16r2225968' 1450512)
				#('16r24C92C8' 48575672)  (16r24C92C8 + 48575672) hex '16r531C780' a free chunk
				#('16r531C788' #endOfMemory))
	 shows there's plenty of large free chunks.  And the trailing 16-byte free chunk shows coallescing is not working properly.

		self biggies #(#('16r31C788' #nil) #('16r531C788' #endOfMemory))
	 shows there are no large objects to be moved.

	 So... looks like compaction should hold onto the lowest large chunk and preferentially move objects into that.
	 Let's call it a pig.  Compaction needs to whittle away at the pig.

	 e.g.
		self abstractPigCompaction #(2061864 0 2061864 18759 2018224 34757)
	 shows we can move 18759 objects that will occupy 2018224 bytes into that
	 low pig of 2061864 bytes."! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 1/11/2017 09:42'!
compactor
	"This is really only for tests..."
	^compactor! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 8/9/2014 12:25'!
compiledMethodFormatForNumBytes: numBytes
	^self firstCompiledMethodFormat + (8 - numBytes bitAnd: self wordSize - 1)! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 1/24/2016 15:54'!
computeFreeSpacePostSwizzle
	totalFreeOldSpace := self totalFreeListBytes.
	self checkFreeSpace: 0! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 06/22/2021 15:31'!
containsOnlyValidBecomeObjects: array1 and: array2 twoWay: isTwoWay copyHash: copyHash
	"Answer 0 if neither array contains an object inappropriate for the become operation.
	 Otherwise answer an informative error code for the first offending object found:
		Can't become: immediates => PrimErrInappropriate
		Shouldn't become pinned objects => PrimErrObjectIsPinned.
		Shouldn't become immutable objects => PrimErrNoModification.
		Can't copy hash into immediates => PrimErrInappropriate.
		Two-way become may require memory to create copies => PrimErrNoMemory.
	 As a side-effect unforward any forwarders in the two arrays if answering 0."
	<inline: true>
	| fieldOffset effectsFlags oop1 oop2 size |
	fieldOffset := self lastPointerOf: array1.
	effectsFlags := size := 0.
	"array1 is known to be the same size as array2"
	[fieldOffset >= self baseHeaderSize] whileTrue:
		[oop1 := self longAt: array1 + fieldOffset.
		 (self isOopForwarded: oop1) ifTrue:
			[oop1 := self followForwarded: oop1.
			 self longAt: array1 + fieldOffset put: oop1].
		 self ifOopInvalidForBecome: oop1 errorCodeInto: [:errCode| ^errCode].
		 oop2 := self longAt: array2 + fieldOffset.
		 (self isOopForwarded: oop2) ifTrue:
			[oop2 := self followForwarded: oop2.
			 self longAt: array2 + fieldOffset put: oop2].
		 oop1 ~= oop2 ifTrue:
			[isTwoWay
				ifTrue:
					[self ifOopInvalidForBecome: oop2 errorCodeInto: [:errCode| ^errCode].
					 size := size + (self bytesInBody: oop1) + (self bytesInBody: oop2).
					 effectsFlags := (effectsFlags
										bitOr: (self becomeEffectFlagsFor: oop1))
										bitOr: (self becomeEffectFlagsFor: oop2)]
				ifFalse:
					[copyHash ifTrue:
						[(self isImmediate: oop2) ifTrue:
							[^PrimErrInappropriate].
						 (self isObjImmutable: oop2) ifTrue:
							[^PrimErrNoModification]].
					 effectsFlags := effectsFlags bitOr: (self becomeEffectFlagsFor: oop1)]].
		 fieldOffset := fieldOffset - self bytesPerOop].
	size >= (totalFreeOldSpace + (scavengeThreshold - freeStart)) ifTrue:
		[^PrimErrNoMemory].
	"only set flags after checking all args."
	becomeEffectsFlags := effectsFlags.
	^0! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 11/18/2016 17:01'!
convertToArray
	<doNotGenerate>
	"I dont believe it -- this *just works*"

	memory:= memory as: Array! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 06/22/2021 15:31'!
copyObj: objOop toAddr: segAddr stopAt: endSeg savedFirstFields: savedFirstFields index: i
	"This is part of storeImageSegmentInto:outPointers:roots:.
	 Copy objOop into the segment beginning at segAddr, and forward it to the copy,
	 saving its first field in savedFirstField, and setting its marked bit to indicate it has
	 been copied.  If it is a class in the class table, set the copy's hash to 0 for reassignment
	 on load, and mark it as a class by setting its isRemembered bit.
	 Answer the next segmentAddr if successful.  Answer an appropriate error code if not"

	"Copy the object..."
	| bodySize copy hash |
	<inline: false>
	self deny: (self isCopiedIntoSegment: objOop).
	bodySize := self bytesInBody: objOop.
	(self oop: segAddr + bodySize isGreaterThanOrEqualTo: endSeg) ifTrue:
		[^PrimErrWritePastObject halt].
	self memcpy: segAddr asVoidPointer _: (self startOfObject: objOop) asVoidPointer _: bodySize.
	copy := self objectStartingAt: segAddr.

	"Clear remembered, mark bits of all headers copied into the segment (except classes)"
	self
		setIsRememberedOf: copy to: false;
		setIsMarkedOf: copy to: false.

	"Make any objects with hidden dynamic state (contexts, methods) look like normal objects."
	self ifAProxy: objOop updateCopy: copy.

	"If the object is a class, zero its identityHash (which is its classIndex) and set its
	 isRemembered bit.  It will be assigned a new hash and entered into the table on load."
	hash := self rawHashBitsOf: objOop.
	(hash > self lastClassIndexPun and: [(self classOrNilAtIndex: hash) = objOop]) ifTrue:
		[self setHashBitsOf: copy to: 0.
		 self setIsRememberedOf: copy to: true].

	"Now forward the object to its copy in the segment."
	self storePointerUnchecked: i ofObject: savedFirstFields withValue: (self fetchPointer: 0 ofObject: objOop);
		storePointerUnchecked: 0 ofObject: objOop withValue: copy;
		markAsCopiedIntoSegment: objOop.

	"Answer the new end of segment"
	^segAddr + bodySize! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 11/13/2013 12:55'!
countMarkedAndUnmarkdObjects: printFlags
	"print the count of marked and unmarked objects.
	 In addition if 1 is set in printFlags, short-print marked objects,
	 and/or if 2 is set, short-print unmarked obejcts."
	<api>
	| nm nu |
	nm := nu := 0.
	self allObjectsDo:
		[:o|
		(self isMarked: o)
			ifTrue:
				[nm := nm + 1.
				 (printFlags anyMask: 1) ifTrue:
					[coInterpreter shortPrintOop: o]]
			ifFalse:
				[nu := nu + 1.
				 (printFlags anyMask: 2) ifTrue:
					[coInterpreter shortPrintOop: o]]].
	self print: 'n marked: '; printNum: nm; cr.
	self print: 'n unmarked: '; printNum: nu; cr! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 3/10/2016 13:43'!
countNumClassPagesPreSwizzle: bytesToShift
	"Compute the used size of the class table before swizzling.  Needed to
	 initialize the classTableBitmap which is populated during adjustAllOopsBy:"
	| firstObj classTableRoot nilObjPreSwizzle |
	firstObj := self objectStartingAt: oldSpaceStart. "a.k.a. nilObj"
	"first five objects are nilObj, falseObj, trueObj, freeListsObj, classTableRootObj"
	classTableRoot := self noInlineObjectAfter:
							(self noInlineObjectAfter:
									(self noInlineObjectAfter:
											(self noInlineObjectAfter: firstObj
												limit: endOfMemory)
										limit: endOfMemory)
								limit: endOfMemory)
							limit: endOfMemory.
	nilObjPreSwizzle := oldSpaceStart - bytesToShift.
	numClassTablePages := self numSlotsOf: classTableRoot.
	self assert: numClassTablePages = (self classTableRootSlots + self hiddenRootSlots).
	2 to: numClassTablePages - 1 do:
		[:i|
		(self fetchPointer: i ofObject: classTableRoot) = nilObjPreSwizzle ifTrue:
			[numClassTablePages := i.
			 ^self]]
	! !

!SpurMemoryManager methodsFor: 'allocation accounting' stamp: 'eem 8/31/2017 10:36'!
currentAllocatedBytes
	"Compute the current allocated bytes since last set.
	 This is the cumulative total in statAllocatedBytes plus the allocation since the last scavenge."
	| use |
	"Slang infers the type of the difference between two unsigned variables as signed.
	 In this case we want it to be unsigned."
	<var: 'use' type: #usqInt>
	use := segmentManager totalOldSpaceCapacity - totalFreeOldSpace.
	^statAllocatedBytes
	 + (freeStart - scavenger eden start)
	 + (use - oldSpaceUsePriorToScavenge)! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 12/1/2014 10:56'!
dbgFloatValueOf: oop
	"Answer the C double precision floating point value of the argument,
	 or if it is not, answer 0."

	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 10/27/2016 16:54'!
defaultEdenBytes
	"Answer the default amount of memory to allocate for the eden space.
	 The actual value can be set via vmParameterAt: and/or a preference in the ini file.
	 The shootout tests seem to plateau at 5 or 6Mb.

	 Originally, both the 32-bit and 64-bit versions used the same 4Mb default.  Measuring
	 the simulator on image start-up, the 64-bit system's eden at the same point in start-up
	 (the first copyBits) is only 8% larger in bytes because it allocates 26% fewer objects.
	 Some 21% of the objects in the 32-bit version's eden are large integers and floats that
	 are representable as 64-bit immediates.

	 But when running benchmarks such as the computer language shootout's binary trees,
	 using the same amount of memory for the 64-bit system causes a significant slow-down
	 and a lot of compactions.  So we now use 4Mb for 32-bits and 7Mb for 64-bits."
	<inline: false>
	^self subclassResponsibility

	"self systemNavigation browseAllCallsOn: #start: and: #limit:"

	"Here are the expressions used to collect the above-mentioned figures:"
	"freeStart - self newSpaceStart"
	"| n | n := 0. self allNewSpaceObjectsDo: [:o| n := n + 1]. n"
	"| f | f := 0. self allNewSpaceObjectsDo: [:o| (self isFloatObject: o) ifTrue: [f := f + 1]]. f"
	"| l | l := 0. self allNewSpaceObjectsDo: [:o|
		(((self classIndexOf: o)
			between: ClassLargeNegativeIntegerCompactIndex
			and: ClassLargePositiveIntegerCompactIndex)
		 and: [(self numBytesOfBytes: o) <= 8
		 and: [self isIntegerValue: (coInterpreter signed64BitValueOf: o)]]) ifTrue: [l := l + 1]]. l"! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 4/13/2016 21:45'!
dequeueMourner
	"Answer the top mourner (ephemeron or weak array) from the queue or
	 nil if the queue is empty. We don't care about order; ephemerons are
	 fired in an arbitrary order based on where they are in the heap."
	^mournQueue ~= nilObj ifTrue:
		[self popObjStack: mournQueue]! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
detachFreeObject: freeChunk
	<inline: true>
	| chunkBytes |
	chunkBytes := self bytesInBody: freeChunk.
	totalFreeOldSpace := totalFreeOldSpace - chunkBytes.
	self unlinkFreeChunk: freeChunk chunkBytes: chunkBytes! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 8/24/2018 16:36'!
disownVM: flags
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter disownVM: flags! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 19:07'!
displayObject
	^self splObj: TheDisplay! !

!SpurMemoryManager methodsFor: 'allocation accounting' stamp: 'eem 6/7/2017 13:09'!
doAllocationAccountingForScavenge
	<inline: true>
	statAllocatedBytes := self currentAllocatedBytes! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 9/23/2014 13:06'!
doBecome: obj1 and: obj2 copyHash: copyHashFlag
	"Inner dispatch for two-way become.
	 N.B. At least in current two-way become use copyHashFlag is false."
	| o1ClassIndex o2ClassIndex |
	"in-lined
			classIndex := (self isInClassTable: obj) ifTrue: [self rawHashBitsOf: obj] ifFalse: [0]
	 for speed."
	o1ClassIndex := self rawHashBitsOf: obj1.
	(o1ClassIndex ~= 0 and: [(self classAtIndex: o1ClassIndex) ~= obj1]) ifTrue:
		[o1ClassIndex := 0].
	o2ClassIndex := self rawHashBitsOf: obj2.
	(o2ClassIndex ~= 0 and: [(self classAtIndex: o2ClassIndex) ~= obj2]) ifTrue:
		[o2ClassIndex := 0].

	"Refuse to do an in-place become on classes since their being
	 forwarded is used in the flush method cache implementations."
	((self numSlotsOf: obj1) = (self numSlotsOf: obj2)
	 and: [o1ClassIndex = 0
	 and: [o2ClassIndex = 0]]) ifTrue:
		[self inPlaceBecome: obj1 and: obj2 copyHashFlag: copyHashFlag.
		 ^self].
	self outOfPlaceBecome: obj1 and: obj2 copyHashFlag: copyHashFlag.
	"if copyHashFlag then nothing changes, since hashes were also swapped."
	copyHashFlag ifTrue:
		[^self].
	"if copyHash is false then the classTable entries must be updated.
	 We leave the following until postBecomeScanClassTable:, but must
	 swap the forwarders if two active classes have been becommed,
	 and assign hashes if not."
	o1ClassIndex ~= 0
		ifTrue:
			[o2ClassIndex ~= 0
				ifTrue:
					[self classAtIndex: o1ClassIndex put: obj2.
					 self classAtIndex: o2ClassIndex put: obj1]
				ifFalse: "o2 wasn't in the table; set its hash"
					[| newObj2 |
					 newObj2 := self followForwarded: obj1.
					 self assert: (self rawHashBitsOf: newObj2) = 0.
					 self setHashBitsOf: newObj2 to: o1ClassIndex]]
		ifFalse:
			[o2ClassIndex ~= 0 ifTrue: "o1 wasn't in the table; set its hash"
				[| newObj1 |
				 newObj1 := self followForwarded: obj2.
				 self assert: (self rawHashBitsOf: newObj1) = 0.
				 self setHashBitsOf: newObj1 to: o2ClassIndex]]! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 5/19/2016 12:29'!
doBecome: obj1 to: obj2 copyHash: copyHashFlag
	"one-way become with or without copying obj1's hash into obj2.
	 Straight-forward, even for classes.  With classes we end up with two entries
	 for obj2.  Which is current depends on copyHashFlag.  If copyHashFlag is true
	 then the entry at obj1's hash is valid, otherwise the the existing one at obj2's
	 hash.  When all the instances with the old hash have been collected, the GC will
	 discover this and expunge obj2 at the unused index (see markAndTraceClassOf:)."
	self forward: obj1 to: obj2.
	copyHashFlag ifTrue: [self setHashBitsOf: obj2 to: (self rawHashBitsOf: obj1)].
	((self isOldObject: obj1)
	 and: [self isYoung: obj2]) ifTrue:
		[becomeEffectsFlags := becomeEffectsFlags bitOr: OldBecameNewFlag].
	self deny: (self isOopForwarded: obj2)! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 3/18/2020 12:59'!
doScavenge: tenuringCriterion
	"The inner shell for scavenge, abstrascted out so globalGarbageCollect can use it."
	<inline: false>
	self doAllocationAccountingForScavenge.
	gcPhaseInProgress := ScavengeInProgress.
	pastSpaceStart := scavenger scavenge: tenuringCriterion.
	self assert: (self
					oop: pastSpaceStart
					isGreaterThanOrEqualTo: scavenger pastSpace start
					andLessThanOrEqualTo: scavenger pastSpace limit).
	freeStart := scavenger eden start.
	gcPhaseInProgress := 0.
	self resetAllocationAccountingAfterGC! !

!SpurMemoryManager methodsFor: 'indexing primitive support' stamp: 'eem 06/22/2021 15:31'!
doShorten: objOop toIndexableSize: indexableSize
	"Reduce the number of indexable fields in objOop, a pointer object, to nSlots. Convert the
	 unused residual to a free chunk. Without changes to numSlotsForShortening:toIndexableSize:
	 this only works for arrayFormat and longFormat objects.
	 Answer the number of bytes returned to free memory, which may be zero if no change
	 was possible."
	<inline: false>
	<api>
	| followingAddress numSlots bytesBefore delta copy freeChunk |
	followingAddress := self addressAfter: objOop. "for assert checking"
	self assert: (self oop: followingAddress isLessThanOrEqualTo: endOfMemory).
	numSlots := self numSlotsForShortening: objOop toIndexableSize: indexableSize.
	bytesBefore := self bytesInBody: objOop.
	delta := bytesBefore - (self objectBytesForSlots: numSlots).

	"Since the system rounds objects up to 64-bits, losing a 32-bit
	 slot may not actually change the bytes occupied by the object."
	delta = 0 ifTrue:
		[(self hasOverflowHeader: objOop)
			ifTrue: [self rawOverflowSlotsOf: objOop put: numSlots]
			ifFalse: [self rawNumSlotsOf: objOop put: numSlots].
		 self updateFormatOfShortenedObject: objOop to: indexableSize.
		 ^0].

	"Currently we can't have one word gaps in oldSpace but we can in newSpace.  So only create
	 a copy and forward if objOop is in old space. If in newSpace we'll create a slimbridge below."
	(delta <= self allocationUnit
	 and: [self isOldObject: objOop]) ifTrue:
		[| format |
		 format := self normalisedFormatFor: objOop indexableSize: indexableSize.
		 copy := self allocateSlots: numSlots
					format: format
					classIndex: (self classIndexOf: objOop).
		 copy ifNil: [self error: 'shorten:toIndexableSize: attempted to shorten by allocationUnit and failed to allocate space for copy!!'].
		 (self isPureBitsFormat: format)
			ifTrue:
				[self memcpy: (self firstIndexableField: copy)
							_: (self firstIndexableField: objOop)
							_: (self numBytesOfBits: copy format: format)]
			ifFalse:
				[0 to: numSlots - 1 do:
					[:i|
					self storePointerUnchecked: i
						ofObject: copy
						withValue: (self fetchPointer: i ofObject: objOop)]].
		 (self isRemembered: objOop) ifTrue:
			[scavenger remember: copy].
		 self forward: objOop to: copy.
		 ^0].

	(self hasOverflowHeader: objOop)
		ifTrue:
			[self rawOverflowSlotsOf: objOop put: numSlots.
			 "Setting an overflow slot count to 0 or 1 in newSpace creates a slimbridge.
			  So we must also change the normal slot count so that the object no
			  longer has the overflow header word, which has become the slimbridge."
			 numSlots <= 1 ifTrue:
			 	[(self oop: objOop isLessThan: oldSpaceStart) ifTrue:
					[self rawNumSlotsOf: objOop put: numSlots.
					 self hackSlimBridgeTo: objOop at: objOop - self allocationUnit]]]
		ifFalse:
			[self assert: numSlots < self numSlotsMask.
			 self rawNumSlotsOf: objOop put: numSlots].

	"{self addressAfter: objOop. delta. (self addressAfter: objOop) + delta. followingAddress} collect: #hex"
	delta := followingAddress - (self addressAfter: objOop).
	self assert: (delta >= self allocationUnit and: [delta \\ self allocationUnit = 0]).
	delta = self allocationUnit
		ifTrue:
			[self deny: (self isOldObject: objOop).
			 delta := 0.
			 followingAddress = freeStart
				ifTrue: [freeStart := self addressAfter: objOop]
				ifFalse: [self hackSlimBridgeTo: followingAddress at: (self addressAfter: objOop)]]
		ifFalse:
			[freeChunk := self initFreeChunkWithBytes: delta at: (self addressAfter: objOop).
			self assert: (self objectAfter: objOop) = freeChunk.
			self assert: (self addressAfter: freeChunk) = followingAddress.
			(self isInOldSpace: objOop)
				ifTrue:
					[totalFreeOldSpace := totalFreeOldSpace + delta.
					 self addToFreeList: freeChunk bytes: delta]
				ifFalse:
					[delta := 0.
					 self
						setClassIndexOf: freeChunk to: self wordSizeClassIndexPun;
						setFormatOf: freeChunk to: self firstLongFormat]].
	self wordSize = 8 ifTrue:
		[self setFormatOf: objOop to: (self normalisedFormatFor: objOop indexableSize: indexableSize)].
	^delta! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/9/2013 20:23'!
edenBytes
	<doNotGenerate>
	^scavenger edenBytes! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 11/21/2013 08:59'!
edenBytes: bytes
	edenBytes := bytes! !

!SpurMemoryManager methodsFor: 'instantiation' stamp: 'eem 11/24/2019 19:24'!
eeInstantiateAnySmallClassIndex: classIndex format: objFormat numSlots: numSlots
	"Instantiate a small instance of a class.  ee stands for execution engine and
	 implies that this allocation will *NOT* cause a GC.  small implies the object will have
	 less than 255 slots. N.B. the instantiated object IS NOT FILLED and must be completed
	 before returning it to Smalltalk. Since this call is used in routines that do just that we
	 are safe.  Break this rule and die in GC.  Result is guaranteed to be young."
	<inline: true>
	self assert: (numSlots >= 0 and: [classIndex ~= 0 and: [(self classAtIndex: classIndex) ~= nilObj]]).
	self assert: (objFormat < self firstByteFormat
					ifTrue: [objFormat]
					ifFalse: [objFormat bitAnd: self byteFormatMask])
				= (self instSpecOfClass: (self classAtIndex: classIndex)).
	^self allocateSmallNewSpaceSlots: numSlots format: objFormat classIndex: classIndex! !

!SpurMemoryManager methodsFor: 'instantiation' stamp: 'eem 1/1/2018 18:18'!
eeInstantiateClassIndex: knownClassIndex format: objFormat numSlots: numSlots
	"Instantiate an instance of a compact class.  ee stands for execution engine and
	 implies that this allocation will *NOT* cause a GC.  N.B. the instantiated object
	 IS NOT FILLED and must be completed before returning it to Smalltalk. Since this
	 call is used in routines that do just that we are safe.  Break this rule and die in GC.
	 Result is guaranteed to be young."
	<api>
	<inline: true>
	self assert: (numSlots >= 0 and: [knownClassIndex ~= 0 and: [(self knownClassAtIndex: knownClassIndex) ~= nilObj]]).
	self assert: (objFormat < self firstByteFormat
					ifTrue: [objFormat]
					ifFalse: [objFormat bitAnd: self byteFormatMask])
				= (self instSpecOfClass: (self knownClassAtIndex: knownClassIndex)).
	^self allocateNewSpaceSlots: numSlots format: objFormat classIndex: knownClassIndex! !

!SpurMemoryManager methodsFor: 'instantiation' stamp: 'eem 9/24/2013 23:16'!
eeInstantiateMethodContextSlots: numSlots
	"Allocate a new MethodContext.  ee stands for execution engine and
	 implies that this allocation will *NOT* cause a GC.  N.B. the instantiated object
	 IS NOT FILLED and must be completed before returning it to Smalltalk. Since this
	 call is used in routines that do just that we are safe.  Break this rule and die in GC.
	 Result is guaranteed to be young."
	<inline: true>
	<inline: true>
	^self
		allocateNewSpaceSlots: numSlots
		format: self indexablePointersFormat
		classIndex: ClassMethodContextCompactIndex! !

!SpurMemoryManager methodsFor: 'instantiation' stamp: 'eem 11/13/2014 10:20'!
eeInstantiateSmallClass: classObj numSlots: numSlots
	"Instantiate an instance of a class, with only a few slots.  ee stands for execution
	 engine and implies that this allocation will *NOT* cause a GC.  N.B. the instantiated
	 object IS NOT FILLED and must be completed before returning it to Smalltalk. Since
	 this call is used in routines that do just that we are safe.  Break this rule and die in GC.
	 Result is guaranteed to be young.  We assume this is only used on classes that are
	 already in the class table."
	<inline: true>
	self assert: (self rawHashBitsOf: classObj) ~= 0.
	^self
		eeInstantiateSmallClassIndex: (self rawHashBitsOf: classObj)
		format: (self instSpecOfClass: classObj)
		numSlots: numSlots! !

!SpurMemoryManager methodsFor: 'instantiation' stamp: 'eem 1/1/2018 18:17'!
eeInstantiateSmallClassIndex: knownClassIndex format: objFormat numSlots: numSlots
	"Instantiate a small instance of a compact class.  ee stands for execution engine and
	 implies that this allocation will *NOT* cause a GC.  small implies the object will have
	 less than 255 slots. N.B. the instantiated object IS NOT FILLED and must be completed
	 before returning it to Smalltalk. Since this call is used in routines that do just that we
	 are safe.  Break this rule and die in GC.  Result is guaranteed to be young."
	<inline: true>
	self assert: (numSlots >= 0 and: [knownClassIndex ~= 0 and: [(self knownClassAtIndex: knownClassIndex) ~= nilObj]]).
	self assert: (objFormat < self firstByteFormat
					ifTrue: [objFormat]
					ifFalse: [objFormat bitAnd: self byteFormatMask])
				= (self instSpecOfClass: (self knownClassAtIndex: knownClassIndex)).
	^self allocateSmallNewSpaceSlots: numSlots format: objFormat classIndex: knownClassIndex! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 5/15/2015 10:25'!
emptyObjStack: objStack
	"Remove all the entries on the stack.  Do so by setting Topx to 0
	 on the first page, and adding all subsequent pages to the free list."
	| nextPage nextNextPage |
	objStack = nilObj ifTrue:
		[^self].
	self eassert: [self isValidObjStack: objStack].
	self storePointer: ObjStackTopx ofObjStack: objStack withValue: 0.
	nextPage := self fetchPointer: ObjStackNextx ofObject: objStack.
	[nextPage ~= 0] whileTrue:
		[nextNextPage := self fetchPointer: ObjStackNextx ofObject: nextPage.
		 self storePointer: ObjStackFreex
			ofObjStack: nextPage
			withValue: (self fetchPointer: ObjStackFreex ofObject: objStack).
		 self storePointer: ObjStackNextx ofObjStack: nextPage withValue: 0.
		 self storePointer: ObjStackFreex ofObjStack: objStack withValue: nextPage.
		 nextPage := nextNextPage].
	self storePointer: ObjStackNextx ofObjStack: objStack withValue: 0.
	self eassert: [self isValidObjStack: objStack]! !

!SpurMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 9/24/2020 22:06'!
enableObjectEnumerationFrom: initialObject
	"We use bridges to stitch segments together to make it appear that the heap is one contiguous space.
	 Bridges at the end of oldSpace segments are maintained.  Bridges at the end of pastSpace and eden
	 are temporary, and are established here, depending on the current sizes of pastSpace end eden.

	 N.B. this introduces complications.  Either or both pastSpace and eden may be empty, so the 
	 bridge from pastSpace may skip eden.  pastSpace may be full, so there may be no bridge at
	 the end of pastSpace.  Most difficult, pastSpace could be one 64-bit word short of full, but normal
	 bridges are two word objects.  To make this work we introduce a hack, objectAfterMaybeSlimBridge:limit:,
	 which uses a fake overflow slot count to get to the start of the next object, which is either one or two
	 words away, depending on whether the first object in eden has a normal or an overflow header."
	<inline: false>
	(self oop: initialObject isLessThan: oldSpaceStart) ifTrue:
		[freeStart > scavenger eden start
			ifTrue:
				[self bridgeEdenAndOldSpace.
				 self bridgePastSpaceAndEden]
			ifFalse: "If eden is empty (e.g. at snapshot time), skip it entirely"
				[self bridgePastSpaceAndOldSpace]]! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 11/16/2013 08:37'!
endOfMemory
	<cmacro: '() GIV(endOfMemory)'>
	^endOfMemory! !

!SpurMemoryManager methodsFor: 'gc - scavenge/compact' stamp: 'eem 12/25/2016 09:12'!
endSlidingCompaction
	gcPhaseInProgress := 0! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 2/4/2015 16:44'!
ensureBehaviorHash: aBehavior
	| hash err |
	<inline: true>
	self assert: (coInterpreter addressCouldBeClassObj: aBehavior).
	^(hash := self rawHashBitsOf: aBehavior) ~= 0
		ifTrue:
			[hash]
		ifFalse:
			[(coInterpreter objCouldBeClassObj: aBehavior)
				ifFalse: [PrimErrBadReceiver negated]
				ifTrue:
					[(err := self enterIntoClassTable: aBehavior) ~= 0
						ifTrue: [err negated]
						ifFalse: [self rawHashBitsOf: aBehavior]]]! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 12/14/2017 12:17'!
ensureNoNewObjectsIn: outPointerArray
	"This is part of loadImageSegmentFrom:outPointers:.
	 Since the remembered bit is currently used to identify classes in the segment, setting remembered bits
	 in new objects in the segment is difficult.  Instead simply arrange that there are no new objects in
	 outPointerArray, obviating the need to remember any of the loaded objects in the segment."
	| scanClassTable |
	scanClassTable := false.
	0 to: (self numSlotsOf: outPointerArray) - 1 do:
		[:i| | oop clone hash |
		 oop := self fetchPointer: i ofObject: outPointerArray.
		 (self isYoung: oop) ifTrue:
			[clone := self cloneInOldSpace: oop forPinning: false.
			 clone = 0 ifTrue:
				[^PrimErrNoMemory halt].
			 ((hash := self rawHashBitsOf: oop) ~= 0
			  and: [(self classOrNilAtIndex: hash) = oop]) ifTrue:
				[scanClassTable := true].
			 self forward: oop to: clone.
			 self storePointerUnchecked: i ofObject: outPointerArray withValue: clone]].
	scanClassTable ifTrue:
		[self postBecomeScanClassTable: BecamePointerObjectFlag].
	^0! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'cb 10/2/2018 14:56'!
ensureRoomOnObjStackAt: objStackRootIndex
	"An obj stack is a stack of objects stored in a hidden root slot, such as
	 the markStack or the ephemeronQueue.  It is a linked list of segments,
	 with the hot end at the head of the list.  It is a word object.  The stack
	 pointer is in ObjStackTopx and 0 means empty.  The list goes through
	 ObjStackNextx. We don't want to shrink objStacks, since they're used
	 in GC and its good to keep their memory around.  So unused pages
	 created by popping emptying pages are kept on the ObjStackFreex list."
	| stackOrNil freeOrNewPage |
	stackOrNil := self fetchPointer: objStackRootIndex ofObject: hiddenRootsObj.
	(stackOrNil = nilObj
	 or: [(self fetchPointer: ObjStackTopx ofObject: stackOrNil) >= ObjStackLimit]) ifTrue:
		[freeOrNewPage := stackOrNil = nilObj
								ifTrue: [0]
								ifFalse: [self fetchPointer: ObjStackFreex ofObject: stackOrNil].
		 freeOrNewPage ~= 0
			ifTrue: "the free page list is always on the new page."
				[self storePointer: ObjStackFreex ofObjStack: stackOrNil withValue: 0.
				 self assert: (marking not or: [self isMarked: freeOrNewPage])]
			ifFalse:
				[freeOrNewPage := self allocateSlotsInOldSpace: ObjStackPageSlots
										format: self wordIndexableFormat
										classIndex: self wordSizeClassIndexPun.
				 freeOrNewPage ifNil: 
					["Allocate a new segment an retry. This is very uncommon. But it happened to me (Clement)."
					 self growOldSpaceByAtLeast: ObjStackPageSlots.
					 freeOrNewPage := self allocateSlotsInOldSpace: ObjStackPageSlots
										format: self wordIndexableFormat
										classIndex: self wordSizeClassIndexPun.
					freeOrNewPage ifNil: [self error: 'no memory to allocate or extend obj stack']].
				 self storePointer: ObjStackFreex ofObjStack: freeOrNewPage withValue: 0.
				 marking ifTrue: [self setIsMarkedOf: freeOrNewPage to: true]].
		 self storePointer: ObjStackMyx ofObjStack: freeOrNewPage withValue: objStackRootIndex;
			  storePointer: ObjStackNextx ofObjStack: freeOrNewPage withValue: (stackOrNil = nilObj ifTrue: [0] ifFalse: [stackOrNil]);
			  storePointer: ObjStackTopx ofObjStack: freeOrNewPage withValue: 0;
			  storePointer: objStackRootIndex ofObject: hiddenRootsObj withValue: freeOrNewPage.
		 self assert: (self isValidObjStackAt: objStackRootIndex).
		 "Added a new page; now update and answer the relevant cached first page."
		 stackOrNil := self updateRootOfObjStackAt: objStackRootIndex with: freeOrNewPage].
	self assert: (self isValidObjStackAt: objStackRootIndex).
	^stackOrNil! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 1/17/2015 13:54'!
ensureSemaphoreUnforwardedThroughContext: aSemaphore
	"Make sure that the aSemaphore is followed through to the suspendedContext of the first link."
	<inline: true>
	| proc ctxt |
	proc := self fetchPointer: FirstLinkIndex ofObject: aSemaphore.
	(self isForwarded: proc) ifTrue:
		[self followForwardedObjectFields: aSemaphore toDepth: 1.
		 proc := self fetchPointer: FirstLinkIndex ofObject: aSemaphore].
	self deny: (self isForwarded: proc).
	ctxt := self fetchPointer: SuspendedContextIndex ofObject: proc.
	(self isForwarded: ctxt) ifTrue:
		[ctxt := self followForwarded: ctxt.
		 self storePointer: SuspendedContextIndex ofObject: proc withValue: ctxt]! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 7/11/2017 19:37'!
enterClassesIntoClassTableFrom: segmentStart to: segmentLimit
 	"This is part of loadImageSegmentFrom:outPointers:.
	 Scan for classes contained in the segment, entering them into the class table,
	 and clearing their isRemembered: bit. Classes are at the front, after the root
	 array and have the remembered bit set. If the attempt succeeds, answer 0,
	 otherwise remove all entered entries and answer an error code."
	| objOop errorCode|
	objOop := self objectAfter: (self objectStartingAt: segmentStart).
	[(self oop: objOop isLessThan: segmentLimit)
	 and: [self isRemembered: objOop]] whileTrue:
		[self setIsRememberedOf: objOop to: false.
		 (errorCode := self enterIntoClassTable: objOop) ~= 0 ifTrue:
			[| oop |
			 oop := objOop.
			 objOop := self objectAfter: (self objectStartingAt: segmentStart).
			 [self oop: objOop isLessThan: oop] whileTrue:
				[self expungeFromClassTable: objOop.
				 objOop := self objectAfter: objOop limit: segmentLimit].
			 ^errorCode halt].
		 objOop := self objectAfter: objOop limit: segmentLimit].
	^0! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 4/28/2015 14:49'!
enterIntoClassTable: aBehavior
	"Enter aBehavior into the class table and answer 0.  Otherwise answer a primitive failure code."
	<inline: false>
	| initialMajorIndex majorIndex minorIndex page |
	majorIndex := classTableIndex >> self classTableMajorIndexShift.
	initialMajorIndex := majorIndex.
	"classTableIndex should never index the first page; it's reserved for known classes"
	self assert: initialMajorIndex > 0.
	minorIndex := classTableIndex bitAnd: self classTableMinorIndexMask.

	[page := self fetchPointer: majorIndex ofObject: hiddenRootsObj.
	 page = nilObj ifTrue:
		[page := self allocateSlotsInOldSpace: self classTablePageSize
					format: self arrayFormat
					classIndex: self arrayClassIndexPun.
		 page ifNil:
			[^PrimErrNoMemory].
		 self fillObj: page numSlots: self classTablePageSize with: nilObj.
		 self storePointer: majorIndex
			ofObject: hiddenRootsObj
			withValue: page.
		 numClassTablePages := numClassTablePages + 1.
		 minorIndex := 0].
	 minorIndex to: self classTablePageSize - 1 do:
		[:i|
		(self fetchPointer: i ofObject: page) = nilObj ifTrue:
			[classTableIndex := majorIndex << self classTableMajorIndexShift + i.
			 "classTableIndex must never index the first page, which is reserved for classes known to the VM."
			 self assert: classTableIndex >= (1 << self classTableMajorIndexShift).
			 self storePointer: i
				ofObject: page
				withValue: aBehavior.
			 self setHashBitsOf: aBehavior to: classTableIndex.
			 self assert: (self classAtIndex: (self rawHashBitsOf: aBehavior)) = aBehavior.
			 ^0]].
	 majorIndex := (majorIndex + 1 bitAnd: self classIndexMask) max: 1.
	 majorIndex = initialMajorIndex ifTrue: "wrapped; table full"
		[^PrimErrLimitExceeded]] repeat! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 9/9/2013 11:15'!
ephemeronFormat
	^5! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 10/3/2013 21:17'!
existInstancesInNewSpaceOf: classObj
	| classIndex |
	classIndex := self rawHashBitsOf: classObj.
	self allNewSpaceObjectsDo:
		[:obj|
		(self classIndexOf: obj) = classIndex ifTrue:
			[^true]].
	^false! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 1/8/2015 12:31'!
expungeDuplicateAndUnmarkedClasses: expungeUnmarked
	"Bits have been set in the classTableBitmap corresponding to
	 used classes.  Any class in the class table that does not have a
	 bit set has no instances with that class index.  However, becomeForward:
	 can create duplicate entries, and these duplicate entries wont match their
	 identityHash. So expunge duplicates by eliminating unmarked entries that
	 don't occur at their identityHash."
	1 to: numClassTablePages - 1 do: "Avoid expunging the puns by not scanning the 0th page."
		[:i| | classTablePage |
		classTablePage := self fetchPointer: i ofObject: hiddenRootsObj.
		 0 to: self classTablePageSize - 1 do:
			[:j| | classOrNil classIndex |
			 classOrNil := self fetchPointer: j ofObject: classTablePage.
			 classIndex := i << self classTableMajorIndexShift + j.
			 self assert: (classOrNil = nilObj or: [coInterpreter addressCouldBeClassObj: classOrNil]).
			 "only remove a class if it is at a duplicate entry or it is unmarked and we're expunging unmarked classes."
			 classOrNil = nilObj
				ifTrue:
					[classIndex < classTableIndex ifTrue:
						[classTableIndex := classIndex]]
				ifFalse:
					[((expungeUnmarked and: [(self isMarked: classOrNil) not])
					   or: [(self rawHashBitsOf: classOrNil) ~= classIndex]) ifTrue:
						[self storePointerUnchecked: j
							ofObject: classTablePage
							withValue: nilObj.
						 "but if it is marked, it should still be in the table at its correct index."
						 self assert: ((expungeUnmarked and: [(self isMarked: classOrNil) not])
									or: [(self classAtIndex: (self rawHashBitsOf: classOrNil)) = classOrNil]).
						 "If the removed class is before the classTableIndex, set the
						  classTableIndex to point to the empty slot so as to reuse it asap."
						 classIndex < classTableIndex ifTrue:
							[classTableIndex := classIndex]]]]].
	"classTableIndex must never index the first page, which is reserved for classes known to the VM."
	self assert: classTableIndex >= (1 << self classTableMajorIndexShift)! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 1/8/2015 12:34'!
expungeFromClassTable: aBehavior
	"Remove aBehavior from the class table."
	<inline: false>
	| classIndex majorIndex minorIndex classTablePage |
	self assert: (self isInClassTable: aBehavior).
	classIndex := self rawHashBitsOf: aBehavior.
	majorIndex := classIndex >> self classTableMajorIndexShift.
	minorIndex := classIndex bitAnd: self classTableMinorIndexMask.
	classTablePage := self fetchPointer: majorIndex ofObject: hiddenRootsObj.
	self assert: classTablePage ~= classTableFirstPage.
	self assert: (self numSlotsOf: classTablePage) = self classTablePageSize.
	self assert: (self fetchPointer: minorIndex ofObject: classTablePage) = aBehavior.
	self storePointerUnchecked: minorIndex ofObject: classTablePage withValue: nilObj.
	"If the removed class is before the classTableIndex, set the
	 classTableIndex to point to the empty slot so as to reuse it asap."
	classIndex < classTableIndex ifTrue:
		[classTableIndex := classIndex].
	"classTableIndex must never index the first page, which is reserved for classes known to the VM."
	self assert: classTableIndex >= (1 << self classTableMajorIndexShift)! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 10/4/2020 12:34'!
extraAlignedBitShift
	<cmacro>
	"bit 2 of 2-bit field above classIndex (little endian)"
	^24! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/19/2013 11:08'!
failed
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter failed! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 12/16/2015 09:21'!
falseObject
	<api>
	^falseObj! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 8/31/2013 12:55'!
falseObject: anOop
	"For mapInterpreterOops"
	falseObj := anOop! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 9/5/2013 13:34'!
fetchByte: byteIndex ofObject: objOop
	<api>
	^self byteAt: objOop + self baseHeaderSize + byteIndex! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 2/4/2015 15:38'!
fetchClassOf: oop
	| tagBits |
	^(tagBits := oop bitAnd: self tagMask) ~= 0
		ifTrue: [self fetchPointer: tagBits ofObject: classTableFirstPage]
		ifFalse: [self fetchClassOfNonImm: oop]! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 2/4/2015 13:49'!
fetchClassOfNonImm: objOop
	<inline: #never>
	| classIndex |
	classIndex := self classIndexOf: objOop.
	classIndex <= self classIsItselfClassIndexPun ifTrue:
		[classIndex = self classIsItselfClassIndexPun ifTrue:
			[^objOop].
		 "Answer nil to avoid the assert failure in classOrNilAtIndex:"
		 classIndex = self isForwardedObjectClassIndexPun ifTrue:
			[^nilObj]].
	self assert: classIndex >= self arrayClassIndexPun.
	^self classOrNilAtIndex: classIndex! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 2/24/2020 15:18'!
fetchClassTagOf: oop
	"Answer the tag used in lookup caches for a receiver.  This is the receiver's classIndex."
	<api>
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 2/25/2020 18:18'!
fetchClassTagOfClass: classObj
	<inline: #always>
	^self rawHashBitsOf: classObj! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 9/11/2013 18:45'!
fetchClassTagOfNonImm: obj
	"In Spur an object's classIndex is the tag in all method caches."
	^self classIndexOf: obj! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 1/18/2020 14:18'!
fetchFloat32: wordIndex ofObject: objOop
	<returnTypeC: #float>
	^self singleFloatAtPointer: objOop + self baseHeaderSize + (wordIndex << 2)! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 1/18/2020 13:52'!
fetchFloat64: longIndex ofObject: objOop
	<returnTypeC: #double>
	^self floatAtPointer: objOop + self baseHeaderSize + (longIndex << 3)! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 15:37'!
fetchInteger: fieldIndex ofObject: objectPointer
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter fetchInteger: fieldIndex ofObject: objectPointer! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 12/15/2014 13:23'!
fetchLong32: fieldIndex ofFloatObject: oop
	"index by word size, and return a pointer as long as the word size"
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 8/31/2013 09:00'!
fetchLong32: fieldIndex ofObject: oop
	"index by 32-bit units, and return a 32-bit value. Intended to replace fetchWord:ofObject:"

	^self long32At: oop + self baseHeaderSize + (fieldIndex << 2)! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 7/29/2014 08:11'!
fetchLong64: longIndex ofObject: objOop
	<returnTypeC: #sqLong>
	^self long64At: objOop + self baseHeaderSize + (longIndex << 3)! !

!SpurMemoryManager methodsFor: 'heap management' stamp: 'cb 6/8/2018 11:27'!
fetchPointer: fieldIndex ofFreeChunk: objOop
	^self longAt: objOop + self baseHeaderSize + (fieldIndex << self shiftForWord)! !

!SpurMemoryManager methodsFor: 'heap management' stamp: 'eem 9/22/2013 15:12'!
fetchPointer: fieldIndex ofMaybeForwardedObject: objOop
	^self longAt: objOop + self baseHeaderSize + (fieldIndex << self shiftForWord)! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 12/15/2015 22:02'!
fetchPointer: fieldIndex ofObject: objOop
	<api>
	^self longAt: objOop + self baseHeaderSize + (fieldIndex << self shiftForWord)! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 7/29/2014 08:11'!
fetchShort16: shortIndex ofObject: objOop
	^self shortAt: objOop + self baseHeaderSize + (shortIndex << 1)! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'nice 11/14/2016 22:32'!
fetchUnsignedShort16: shortIndex ofObject: objOop
	^self cCoerceSimple: (self shortAt: objOop + self baseHeaderSize + (shortIndex << 1)) to: #'unsigned short'! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 11/8/2020 21:11'!
fileTimesInUTC
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter fileTimesInUTC! !

!SpurMemoryManager methodsFor: 'instantiation' stamp: 'eem 9/26/2013 10:39'!
fillObj: objOop numSlots: numSlots with: fillValue
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
findLargestFreeChunk
	"Answer, but do not remove, the largest free chunk in the free lists."
	| treeNode childNode |
	treeNode := freeLists at: 0.
	treeNode = 0 ifTrue:
		[^nil].
	[self assertValidFreeObject: treeNode.
	 self assert: (self bytesInBody: treeNode) >= (self numFreeLists * self allocationUnit).
	 childNode := self fetchPointer: self freeChunkLargerIndex ofFreeChunk: treeNode.
	 childNode ~= 0] whileTrue:
		[treeNode := childNode].
	^treeNode! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 8/8/2021 07:51'!
findString: aCString
	"Print the oops of all string-like things that have the same characters as aCString"
	<api>
	<var: #aCString type: #'char *'>
	| aCStringStrlen |
	aCStringStrlen := self strlen: aCString.
	self allObjectsDo:
		[:obj|
		 (self object: obj equalsString: aCString ofSize: aCStringStrlen) ifTrue:
			[coInterpreter printHex: obj; space; printOopShort: obj; cr]]! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 8/9/2021 13:35'!
findStringBeginningWith: aCString
	"Print the oops of all string-like things that start with the same characters as aCString"
	<api>
	<var: #aCString type: #'char *'>
	| aCStringStrlen |
	aCStringStrlen := self strlen: aCString.
	self allObjectsDo:
		[:obj|
		 (self object: obj equalsString: aCString ofSize: aCStringStrlen) ifTrue:
			[coInterpreter printHex: obj; space; printNum: (self lengthOf: obj); space; printOopShort: obj; cr]]! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 5/23/2016 15:40'!
fireAllUnscannedEphemerons
	self assert: (self noUnscannedEphemerons) not.
	self assert: self allUnscannedEphemeronsAreActive.
	unscannedEphemerons start to: unscannedEphemerons top - self bytesPerOop by: self bytesPerOop do:
		[:p|
		coInterpreter fireEphemeron: (self longAt: p)]! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 11/21/2013 11:32'!
firstAccessibleObject
	<inline: false>
	self assert: nilObj = oldSpaceStart.
	"flush newSpace to settle the enumeration."
	self flushNewSpace.
	^nilObj! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 12/12/2015 14:18'!
firstByteFormat
	<api>
	<cmacro>
	^16! !

!SpurMemoryManager methodsFor: 'class table puns' stamp: 'eem 3/9/2016 14:48'!
firstClassIndexPun
	"Class puns are class indices not used by any class.  There is an entry
	 for the pun that refers to the notional class of objects with this class
	 index.  But because the index doesn't match the class it won't show up
	 in allInstances, hence hiding the object with a pun as its class index.
	 The puns occupy indices 16 through 31."
	<cmacro>
	^16! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 12/12/2015 14:18'!
firstCompiledMethodFormat
	<api>
	<cmacro>
	^24! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 9/3/2013 14:53'!
firstFixedField: objOop
	<returnTypeC: #'void *'>
	^ self pointerForOop: objOop + self baseHeaderSize! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 10/31/2013 09:38'!
firstFixedFieldOfMaybeImmediate: oop
	"for the message send breakpoint; selectors can be immediates."
	<inline: false>
	^(self isImmediate: oop)
		ifTrue: [oop asVoidPointer]
		ifFalse: [self firstFixedField: oop]! !

!SpurMemoryManager methodsFor: 'object format' stamp: 'eem 6/28/2016 13:53'!
firstIndexableField: objOop
	"NOTE: overridden in various simulator subclasses to add coercion to CArray, so please duplicate any changes.
	 There are only two important cases, both for objects with named inst vars, i.e. formats 2,3 & 5.
	 The first indexable field for formats 2 & 5 is the slot count (by convention, even though that's off the end
	 of the object).  For 3 we must go to the class."
	| fmt classFormat |
	<returnTypeC: #'void *'>
	fmt := self formatOf: objOop.
	fmt <= self weakArrayFormat ifTrue:
		[fmt = self arrayFormat ifTrue: "array starts at 0."
			[^self pointerForOop: objOop + self baseHeaderSize].
		 fmt >= self indexablePointersFormat ifTrue: "indexable with inst vars; need to delve into the class format word"
			[classFormat := self formatOfClass: (self fetchClassOfNonImm: objOop).
			 ^self pointerForOop: objOop
								+ self baseHeaderSize
								+ ((self fixedFieldsOfClassFormat: classFormat) << self shiftForWord)].
		 "otherwise not indexable"
		 ^0].
	"All bit objects, and indeed CompiledMethod, though this is a no-no, start at 0"
	(fmt >= self sixtyFourBitIndexableFormat
	 and: [fmt < self firstCompiledMethodFormat]) ifTrue:
		[^self pointerForOop: objOop + self baseHeaderSize].
	"otherwise not indexable"
	^0! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'cb 10/5/2018 12:34'!
firstLilliputianChunk
	^freeLists at: self lilliputianChunkIndex! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 12/12/2015 13:56'!
firstLongFormat
	<api>
	<cmacro>
	^10! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 9/1/2013 14:32'!
firstObject
	"Return the first object or free chunk in the heap."

	^nilObj! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 10/9/2013 10:55'!
firstSegmentBytes
	<doNotGenerate>
	^segmentManager firstSegmentBytes! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 10/10/2013 06:09'!
firstSegmentSize: firstSegmentSize
	<doNotGenerate>
	"even an empty segment needs a bridge ;-)"
	self assert: firstSegmentSize >= (2 * self baseHeaderSize).
	segmentManager firstSegmentSize: firstSegmentSize! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 12/12/2015 14:18'!
firstShortFormat
	<api>
	<cmacro>
	^12! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 9/9/2013 15:07'!
firstStringyFakeFormat
	"A fake format for the interpreter used to mark indexable strings in
	 the interpreter's at cache.  This is larger than any format."
	^32! !

!SpurMemoryManager methodsFor: 'indexing primitive support' stamp: 'eem 9/11/2013 07:35'!
firstValidIndexOfIndexableObject: obj withFormat: fmt
	"Answer the one-relative index of the first valid index in an indexbale object
	 with the given format.  This is 1 for all objects except compiled methods
	 where the first index is beyond the last literal.
	 Used for safer bounds-checking on methods."
	^fmt >= self firstCompiledMethodFormat
		ifTrue: [coInterpreter firstByteIndexOfMethod: obj]
		ifFalse: [1]! !

!SpurMemoryManager methodsFor: 'forwarding' stamp: 'eem 2/15/2016 15:55'!
fixFollowedField: fieldIndex ofObject: anObject withInitialValue: initialValue
	"Private helper for followField:ofObject: to avoid code duplication for rare case."
	<inline: #never>
	| objOop |
	self assert: (self isOopForwarded: initialValue).
	"inlined followForwarded: for speed (one less test)"
	objOop := initialValue.
	[objOop := self fetchPointer: 0 ofMaybeForwardedObject: objOop.
	 self isOopForwarded: objOop] whileTrue.
	self storePointer: fieldIndex ofObject: anObject withValue: objOop.
	^objOop! !

!SpurMemoryManager methodsFor: 'object format' stamp: 'eem 12/12/2015 14:18'!
fixedFieldsFieldWidth
	<api>
	<cmacro>
	^16! !

!SpurMemoryManager methodsFor: 'object format' stamp: 'eem 11/10/2014 12:49'!
fixedFieldsOf: objOop format: fmt length: wordLength
	| class |
	<inline: true>
	"N.B. written to fall through to fetchClassOfNonImm: et al for forwarders
	 so as to trigger an assert fail."
	(fmt >= self sixtyFourBitIndexableFormat or: [fmt = self arrayFormat]) ifTrue:
		[^0].  "indexable fields only"
	fmt < self arrayFormat ifTrue:
		[^wordLength].  "fixed fields only (zero or more)"
	class := self fetchClassOfNonImm: objOop.
	^self fixedFieldsOfClassFormat: (self formatOfClass: class)! !

!SpurMemoryManager methodsFor: 'object format' stamp: 'eem 9/26/2013 14:34'!
fixedFieldsOfClass: objOop
	^self fixedFieldsOfClassFormat: (self formatOfClass: objOop)! !

!SpurMemoryManager methodsFor: 'object format' stamp: 'eem 7/12/2016 13:55'!
fixedFieldsOfClassFormat: classFormat
	<api>
	^classFormat bitAnd: self fixedFieldsOfClassFormatMask! !

!SpurMemoryManager methodsFor: 'object format' stamp: 'eem 11/18/2013 13:44'!
fixedFieldsOfClassFormatMask
	<api>
	^1 << self fixedFieldsFieldWidth - 1! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 1/18/2020 14:26'!
floatAtPointer: ptr
	"See platforms/Cross/vm/sqMemoryAccess.h for the production implementation."
	<doNotGenerate>
	^Float fromIEEE64BitWord: (self long64At: ptr)! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 1/18/2020 14:25'!
floatAtPointer: ptr put: aDouble
	"See platforms/Cross/vm/sqMemoryAccess.h for the production implementation."
	<doNotGenerate>
	self long64At: ptr put: aDouble asIEEE64BitWord.
	^aDouble! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/21/2014 16:22'!
floatObjectOf: aFloat
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/21/2014 17:18'!
floatValueOf: oop
	"Answer the C double precision floating point value of the argument,
	 or fail if it is not a Float, and answer 0.
	 Note: May be called by translated primitive code."

	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 5/6/2014 19:10'!
flushEden
	"Fush everything in eden.  Do so by doing a non-tenuring scavenge."
	self scavengingGCTenuringIf: DontTenure.
	self assert: pastSpaceStart = scavenger pastSpace start.
	self assert: freeStart = scavenger eden start! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 5/6/2014 19:10'!
flushNewSpace
	"Fush everything in new space.  Do so by setting the tenure
	 threshold above everything in newSpace, i.e. newSpaceLimit."
	| savedTenuringThreshold |
	savedTenuringThreshold := scavenger getRawTenuringThreshold.
	scavenger setRawTenuringThreshold: newSpaceLimit.
	self scavengingGCTenuringIf: TenureByAge.
	scavenger setRawTenuringThreshold: savedTenuringThreshold.
	self assert: scavenger rememberedSetSize = 0.
	self assert: pastSpaceStart = scavenger pastSpace start.
	self assert: freeStart = scavenger eden start! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 10/3/2013 21:18'!
flushNewSpaceInstancesOf: aClass
	| classIndex |
	classIndex := self rawHashBitsOf: aClass.
	classIndex = 0 ifTrue: "no instances; nothing to do"
		[^self].
	scavenger tenuringClassIndex: classIndex.
	self scavengingGCTenuringIf: TenureByClass.
	self assert: (self existInstancesInNewSpaceOf: aClass) not! !

!SpurMemoryManager methodsFor: 'selective compaction' stamp: 'eem 10/9/2018 15:51'!
followClassTable
	"In addition to postBecomeScanClassTable:, I follow the pages in the class table.
	 Because hiddenRootsObj follows nil, false, true and the freeLists, it can never be forwarded."
	self deny: (self isForwarded: hiddenRootsObj).
	0 to: numClassTablePages - 1 do:
		[:i| | page |
		page := self followField: i ofObject: hiddenRootsObj.
		0 to: (self numSlotsOf: page) - 1 do:
			[:j| | classOrNil |
			classOrNil := self fetchPointer: j ofObject: page.
			classOrNil ~= nilObj ifTrue:
				[(self isForwarded: classOrNil) ifTrue:
					[classOrNil := self followForwarded: classOrNil.
					 self storePointer: j ofObject: page withValue: classOrNil].
				 (self rawHashBitsOf: classOrNil) = 0 ifTrue:
					[self storePointerUnchecked: j ofObject: page withValue: nilObj.
					 "If the removed class is before the classTableIndex, set the
					  classTableIndex to point to the empty slot so as to reuse it asap."
					 (i << self classTableMajorIndexShift + j) < classTableIndex ifTrue:
						[classTableIndex := i << self classTableMajorIndexShift + j]]]]].
	"classTableIndex must never index the first page, which is reserved for classes known to the VM."
	self assert: classTableIndex >= (1 << self classTableMajorIndexShift).
	self assert: self validClassTableRootPages! !

!SpurMemoryManager methodsFor: 'forwarding' stamp: 'eem 12/4/2013 15:53'!
followField: fieldIndex ofObject: anObject
	"Make sure the oop at fieldIndex in anObject is not forwarded (follow the
	 forwarder there-in if so).  Answer the (possibly followed) oop at fieldIndex."
	| objOop |
	objOop := self fetchPointer: fieldIndex ofObject: anObject.
	(self isOopForwarded: objOop) ifTrue:
		[objOop := self fixFollowedField: fieldIndex ofObject: anObject withInitialValue: objOop].
	^objOop! !

!SpurMemoryManager methodsFor: 'forwarding' stamp: 'eem 7/9/2014 11:32'!
followForwarded: objOop
	"Follow a forwarding pointer.  THis must be a loop because we cannot prevent forwarders to
	 forwarders being created by lazy become.  Consider the following example by Igor Stasenko:
		array := { a. b. c }.
		- array at: 1 points to &a. array at: 2 points to &b. array at: 3 points to &c
		a becomeForward: b
		- array at: 1 still points to &a. array at: 2 still points to &b. array at: 3 still points to &c
		b becomeForward: c.
		- array at: 1 still points to &a. array at: 2 still points to &b. array at: 3 still points to &c
		- when accessing array first one has to follow a forwarding chain:
		&a -> &b -> c"
	<api>
	| referent |
	self assert: (self isUnambiguouslyForwarder: objOop).
	referent := self fetchPointer: 0 ofMaybeForwardedObject: objOop.
	[(self isOopForwarded: referent)] whileTrue:
		[referent := self fetchPointer: 0 ofMaybeForwardedObject: referent].
	^referent! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 10/31/2013 21:25'!
followForwardedInObjStack: objStack atIndex: objStackRootIndex
	"Post-compact, follow forwarders in an obj stack."
	| firstPage stackOrNil index field |
	objStack = nilObj ifTrue:
		[^objStack].
	stackOrNil := objStack.
	(self isForwarded: stackOrNil) ifTrue:
		[stackOrNil := self followForwarded: stackOrNil.
		 self updateRootOfObjStackAt: objStackRootIndex with: stackOrNil].
	firstPage := stackOrNil.
	[self assert: (self numSlotsOfAny: stackOrNil) = ObjStackPageSlots.
	 self assert: (self fetchPointer: ObjStackMyx ofObject: stackOrNil) = objStackRootIndex.
	 "There are four fixed slots in an obj stack, and a Topx of 0 indicates empty, so
	   if there were 5 slots in an oop stack, full would be 2, and the last 0-rel index is 4.
	   Hence the last index is topx + fixed slots - 1, or topx + ObjStackNextx"
	 index := (self fetchPointer: ObjStackTopx ofObject: stackOrNil) + ObjStackNextx.
	 "follow forwarded fields including ObjStackNextx and leave field containing the next link."
	 [field := self fetchPointer: index ofObject: stackOrNil.
	  (field ~= 0 and: [(self isNonImmediate: field) and: [self isForwarded: field]]) ifTrue:
		[field := self followForwarded: field.
		 self storePointer: index ofObjStack: stackOrNil withValue: field].
	  (index := index - 1) > ObjStackFreex] whileTrue.
	  self assert: field = (self fetchPointer: ObjStackNextx ofObject: stackOrNil).
	 (stackOrNil := field) ~= 0] whileTrue.
	[stackOrNil := self fetchPointer: ObjStackFreex ofObject: firstPage.
	 stackOrNil ~= 0] whileTrue:
		[(self isForwarded: stackOrNil) ifTrue:
			[stackOrNil := self followForwarded: stackOrNil.
			 self storePointer: ObjStackFreex ofObjStack: firstPage withValue: stackOrNil].
		 firstPage := stackOrNil].
	self assert: (self isValidObjStackAt: objStackRootIndex)! !

!SpurMemoryManager methodsFor: 'compaction' stamp: 'eem 8/31/2015 10:58'!
followForwardedObjStacks
	"Compaction will move objStack pages as well as ordinary objects.
	 So they need their slots followed."
	self followForwardedInObjStack: markStack atIndex: MarkStackRootIndex.
	self followForwardedInObjStack: weaklingStack atIndex: WeaklingStackRootIndex.
	self followForwardedInObjStack: mournQueue atIndex: MournQueueRootIndex! !

!SpurMemoryManager methodsFor: 'forwarding' stamp: 'eem 6/24/2020 10:22'!
followForwardedObjectFields: objOop toDepth: depth
	"Follow pointers in the object to depth.
	 Answer if any forwarders were found.
	 How to avoid cyclic structures?? A temporary mark bit? eem 6/22/2020 no need since depth is always finite."
	<api>
	<inline: false>
	| found numSlots |
	found := false.
	self assert: ((self isPointers: objOop) or: [self isOopCompiledMethod: objOop]).
	numSlots  := self numPointerSlotsOf: objOop.
	0 to: numSlots - 1 do:
		[:i| | oop |
		 oop := self fetchPointer: i ofObject: objOop.
		 (self isNonImmediate: oop) ifTrue:
			[(self isForwarded: oop) ifTrue:
				[found := true.
				 oop := self followForwarded: oop.
				 self storePointer: i ofObject: objOop withValue: oop].
			(depth > 0
			 and: [(self hasPointerFields: oop)
			 and: [self followForwardedObjectFields: oop toDepth: depth - 1]]) ifTrue:
				[found := true]]].
	^found! !

!SpurMemoryManager methodsFor: 'forwarding' stamp: 'eem 10/14/2014 12:09'!
followMaybeForwarded: objOop
	<inline: true>
	^(self isOopForwarded: objOop)
		ifTrue: [self noInlineFollowForwarded: objOop]
		ifFalse: [objOop]! !

!SpurMemoryManager methodsFor: 'forwarding' stamp: 'eem 1/17/2014 10:59'!
followObjField: fieldIndex ofObject: anObject
	"Make sure the obj at fieldIndex in anObject is not forwarded (follow the
	 forwarder there-in if so).  Answer the (possibly followed) obj at fieldIndex."
	| objOop |
	objOop := self fetchPointer: fieldIndex ofObject: anObject.
	self assert: (self isNonImmediate: objOop).
	(self isForwarded: objOop) ifTrue:
		[objOop := self fixFollowedField: fieldIndex ofObject: anObject withInitialValue: objOop].
	^objOop! !

!SpurMemoryManager methodsFor: 'forwarding' stamp: 'eem 2/16/2016 16:02'!
followOopField: fieldIndex ofObject: anObject
	"Make sure the oop at fieldIndex in anObject is not forwarded (follow the
	 forwarder there-in if so).  Answer the (possibly followed) oop at fieldIndex."
	| oop |
	oop := self fetchPointer: fieldIndex ofObject: anObject.
	(self isOopForwarded: oop) ifTrue:
		[oop := self fixFollowedField: fieldIndex ofObject: anObject withInitialValue: oop].
	^oop! !

!SpurMemoryManager methodsFor: 'selective compaction' stamp: 'cb 10/4/2018 09:33'!
followProcessList
	"Eagerly patch all process related forwarders, except suspended contexts which are lazily patched in wakeHighestPriority"
	| scheduler processLists processList proc |
	scheduler := self followField: ValueIndex ofObject: (self splObj: SchedulerAssociation).
	self followField: ActiveProcessIndex ofObject: scheduler.
	processLists := self followField: ProcessListsIndex ofObject: scheduler.
	0 to: (self numSlotsOf: processLists) - 1 do: [:i |
		processList := self followField: i ofObject: processLists.
		self followField: LastLinkIndex ofObject: processList.
		proc := self followField: FirstLinkIndex ofObject: processList.
		[proc = nilObj] whileFalse: [proc := self followField: NextLinkIndex ofObject: proc]].! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 1/16/2020 19:50'!
followSpecialObjectsOop
	(self isForwarded: specialObjectsOop) ifTrue:
		[validatedIntegerClassFlags := 0.
		 specialObjectsOop := self followForwarded: specialObjectsOop].
	self followForwardedObjectFields: specialObjectsOop toDepth: 0.! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 2/16/2016 16:02'!
followedKeyOfEphemeron: objOop
	"Answer the object the ephemeron guards.  This is its first element."
	self assert: ((self isNonImmediate: objOop) and: [self isEphemeron: objOop]).
	^self followOopField: 0 ofObject: objOop! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 5/25/2016 18:35'!
followedKeyOfMaybeFiredEphemeron: objOop
	"Answer the object the ephemeron guards.  This is its first element."
	self assert: ((self isNonImmediate: objOop) and: [self isMaybeFiredEphemeron: objOop]).
	^self followOopField: 0 ofObject: objOop! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 12/28/2016 13:06'!
formatFieldWidthShift
	<cmacro>
	"The format field contains 5 bits."
	^5! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 1/10/2017 15:56'!
formatMask
	<api>
	<cmacro>
	"0 = 0 sized objects (UndefinedObject True False et al)
	 1 = non-indexable objects with inst vars (Point et al)
	 2 = indexable objects with no inst vars (Array et al)
	 3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
	 4 = weak indexable objects with inst vars (WeakArray et al)
	 5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
	 6,7,8 unused
	 9 64-bit indexable
	 10 - 11 32-bit indexable
	 12 - 15 16-bit indexable
	 16 - 23 byte indexable
	 24 - 31 compiled method"
	^16r1f! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 1/10/2017 15:56'!
formatOf: objOop
	"0 = 0 sized objects (UndefinedObject True False et al)
	 1 = non-indexable objects with inst vars (Point et al)
	 2 = indexable objects with no inst vars (Array et al)
	 3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
	 4 = weak indexable objects with inst vars (WeakArray et al)
	 5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
	 6 unused, reserved for exotic pointer objects?
	 7 Forwarded Object, 1st field is pointer, rest of fields are ignored
	 8 unused, reserved for exotic non-pointer objects?
	 9 64-bit indexable
	 10 - 11 32-bit indexable	(11 unused in 32 bits)
	 12 - 15 16-bit indexable	(14 & 15 unused in 32-bits)
	 16 - 23 byte indexable		(20-23 unused in 32-bits)
	 24 - 31 compiled method	(28-31 unused in 32-bits)"
	"A note on performance.  Since the format field is, by design, aligned on a byte boundary
	 in the fourth byte of the header we could access it via
		^(self byteAt: objOop + 3) bitAnd: self formatMask
	 but al least on e.g. Core i7 x86-64 using the clang 6 compiler, this makes no difference,
	 or at least any change is in the noise."
	^(self longAt: objOop) >> self formatShift bitAnd: self formatMask! !

!SpurMemoryManager methodsFor: 'object format' stamp: 'eem 9/9/2013 11:58'!
formatOfClass: classPointer
	<api>
	<inline: true>
	^self integerValueOf: (self fetchPointer: InstanceSpecificationIndex ofObject: classPointer)! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 1/10/2017 15:56'!
formatOfHeader: header
	"0 = 0 sized objects (UndefinedObject True False et al)
	 1 = non-indexable objects with inst vars (Point et al)
	 2 = indexable objects with no inst vars (Array et al)
	 3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
	 4 = weak indexable objects with inst vars (WeakArray et al)
	 5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
	 6 unused, reserved for exotic pointer objects?
	 7 Forwarded Object, 1st field is pointer, rest of fields are ignored
	 8 unused, reserved for exotic non-pointer objects?
	 9 64-bit indexable
	 10 - 11 32-bit indexable
	 12 - 15 16-bit indexable
	 16 - 23 byte indexable
	 24 - 31 compiled method"
	<var: 'header' type: #usqLong>
	^header >> self formatShift bitAnd: self formatMask! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 12/12/2015 14:18'!
formatShift
	<api>
	<cmacro>
	^24! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'cb 12/1/2015 14:54'!
forward: obj1 to: obj2
	self set: obj1 classIndexTo: self isForwardedObjectClassIndexPun formatTo: self forwardedFormat.
	self cppIf: IMMUTABILITY ifTrue: [ self setIsImmutableOf: obj1 to: false ].
	self storePointer: 0 ofForwarder: obj1 withValue: obj2.
	"For safety make sure the forwarder has a slot count that includes its contents."
	(self rawNumSlotsOf: obj1) = 0 ifTrue:
		[self rawNumSlotsOf: obj1 put: 1]! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 5/3/2014 22:25'!
forwardSurvivor: obj1 to: obj2
	"This version of forward:to: can use an unchecked store because it is known that obj1 is young."
	self assert: (self isInNewSpace: obj1).
	self assert: ((self isInFutureSpace: obj2) or: [self isInOldSpace: obj2]).
	self storePointerUnchecked: 0 ofObject: obj1 withValue: obj2.
	self set: obj1 classIndexTo: self isForwardedObjectClassIndexPun formatTo: self forwardedFormat! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 5/3/2014 22:25'!
forwardUnchecked: obj1 to: obj2
	"This version of forward:to: can use an unchecked store because it is known that both obj1 and obj2 are old."
	self assert: ((self isInOldSpace: obj1) and: [self isInOldSpace: obj2]).
	self storePointerUnchecked: 0 ofObject: obj1 withValue: obj2.
	self set: obj1 classIndexTo: self isForwardedObjectClassIndexPun formatTo: self forwardedFormat! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 9/11/2013 12:03'!
forwardedFormat
	"A special format used by the GC to follow only the first pointer."
	^7! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 7/5/2017 16:40'!
forwardersIn: anObject
	"Answer if anObject is itself forwarded, or is a pointer object containing any references to forwarded objects."
	(self isForwarded: anObject) ifTrue:
		[^true].
	0 to: (self numPointerSlotsOf: anObject) - 1 do:
		[:i| | oop |
		 oop := self fetchPointer: i ofMaybeForwardedObject: anObject.
		 ((self isNonImmediate: oop)
		  and: [self isForwarded: oop]) ifTrue:
			[^true]].
	^false! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 9/11/2013 13:51'!
freeChunkLargerIndex
	"for organizing the tree of large free chunks."
	^4! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 4/23/2014 16:43'!
freeChunkNextIndex
	"for linking objecs on each free list, or, during pigCompact, doubly-
	 linking the free objects in address order using the xor link hack."
	^0! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 9/11/2013 13:51'!
freeChunkParentIndex
	"for organizing the tree of large free chunks."
	^2! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'cb 6/7/2018 10:02'!
freeChunkPrevIndex
	"For linking objecs on each free list, doubly-linking the free objects.
	 Free chunks of size 1 do not have a prev index."
	^1! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 9/11/2013 13:51'!
freeChunkSmallerIndex
	"for organizing the tree of large free chunks."
	^3! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 1/13/2017 12:17'!
freeChunkWithBytes: bytes at: address
	<inline: false>
	| freeChunk |
	self assert: (self isInOldSpace: address).
	self assert: (segmentManager segmentContainingObj: address) = (segmentManager segmentContainingObj: address + bytes).
	freeChunk := self initFreeChunkWithBytes: bytes at: address.
	self addToFreeList: freeChunk bytes: bytes.
	self assert: freeChunk = (self objectStartingAt: address).
	^freeChunk! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 5/28/2014 14:25'!
freeListHeadsEmpty
	0 to: self numFreeLists - 1 do:
		[:i| (freeLists at: i) ~= 0 ifTrue: [^false]].
	^true! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 1/26/2017 15:03'!
freeListsObj
	self assert: (self firstIndexableField: (self oldSpaceObjectAfter: trueObj)) = freeLists.
	^self oldSpaceObjectAfter: trueObj! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 10/8/2013 09:43'!
freeListsObject
	^self objectAfter: trueObj! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
freeObject: objOop
	"Free an object in oldSpace.  Coalesce if possible to reduce fragmentation."
	<api>
	<inline: false>
	| bytes start next |
	self assert: (self isInOldSpace: objOop).
	(self isRemembered: objOop) ifTrue:
		[scavenger forgetObject: objOop].
	bytes := self bytesInBody: objOop.
	start := self startOfObject: objOop.
	next := self objectStartingAt: start + bytes.
	(self isFreeObject: next) ifTrue:
		[self detachFreeObject: next.
		 bytes := bytes + (self bytesInBody: next)].
	totalFreeOldSpace := totalFreeOldSpace + bytes.
	^self freeChunkWithBytes: bytes at: start! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 10/10/2013 11:09'!
freeOldSpaceStart
	^freeOldSpaceStart! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 9/22/2013 08:31'!
freeSize
	^totalFreeOldSpace! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 06/22/2021 15:31'!
freeSpaceCharacterisation
	<doNotGenerate>
	| n s |
	n := 0.
	s := Bag new.
	self allFreeObjectsDo:
		[:f| n := n + 1. s add: (self bytesInBody: f)].
	^{ n. s sortedCounts. s sortedElements }! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/12/2013 09:29'!
freeStart
	"This is a horrible hack and only works because C macros are generated after Interpreter variables."
	<cmacro: '() GIV(freeStart)'>
	^freeStart! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
freeTreeNodesDo: aBlock
	"Enumerate all nodes in the free tree (in order, smaller to larger),
	 but *not* including the next nodes of the same size off each tree node.
	 This is an iterative version so that the block argument can be
	 inlined by Slang. The trick to an iterative binary tree application is
	 to apply the function on the way back up when returning from a
	 particular direction, in this case up from the larger child.

	 N.B For the convenience of rebuildFreeTreeFromSortedFreeChunks
	 aBlock *MUST* answer the freeTreeNode it was invoked with, or
	 its replacement if it was replaced by aBlock."
	<inline: true>
	| treeNode cameFrom |
	treeNode := freeLists at: 0.
	treeNode = 0 ifTrue:
		[^self].
	cameFrom := -1.
	[| smallChild largeChild |
	 self assert: (self bytesInBody: treeNode) >= (self numFreeLists * self allocationUnit).
	 smallChild := self fetchPointer: self freeChunkSmallerIndex ofFreeChunk: treeNode.
	 largeChild := self fetchPointer: self freeChunkLargerIndex ofFreeChunk: treeNode.
	 self assert: (smallChild = 0 or: [treeNode = (self fetchPointer: self freeChunkParentIndex ofFreeChunk: smallChild)]).
	 self assert: (largeChild = 0 or: [treeNode = (self fetchPointer: self freeChunkParentIndex ofFreeChunk: largeChild)]).
	 "apply if the node has no children, or it has no large children and we're
	  returning from the small child, or we're returning from the large child."
	 ((smallChild = 0 and: [largeChild = 0])
	  or: [largeChild = 0
			ifTrue: [cameFrom = smallChild]
			ifFalse: [cameFrom = largeChild]])
		ifTrue:
			[treeNode := aBlock value: treeNode.
			 "and since we've applied we must move on up"
			 cameFrom := treeNode.
			 treeNode := self fetchPointer: self freeChunkParentIndex ofFreeChunk: treeNode]
		ifFalse:
			[(smallChild ~= 0 and: [cameFrom ~= smallChild])
				ifTrue:
					[treeNode := smallChild]
				ifFalse:
					[self assert: largeChild ~= 0.
					 treeNode := largeChild].
			 cameFrom := -1].
	 treeNode ~= 0] whileTrue! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
freeTreeOverlapCheck
	<doNotGenerate>
	"Assumes no 2 consecutive free chunks"
	self allObjectsInFreeTreeDo: [:freeNode1|
		self allObjectsInFreeTreeDo: [:freeNode2|
			freeNode1 == freeNode2
				ifFalse: 
					[|start1 start2 end1 end2|
					start1 := self startOfObject: freeNode1.
					start2 := self startOfObject: freeNode2.
					end1 := start1 + (self bytesInBody: freeNode1).
					end2 := start2 + (self bytesInBody: freeNode2).
					"
					Transcript 
						show: '['; show: start1; show: ';'; 
						show: end1; show: '];'; cr; show: '['; show: start2;
						show: ';'; show: end2; show: ']'; cr.
					"
					self assert: (start2 > end1 or: [end2 < start1]).
					self assert: (start1 > end2 or: [start1 < start2])]]]! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 3/14/2014 16:48'!
fullDisplayUpdate
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter fullDisplayUpdate! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 06/22/2021 15:31'!
fullGC
	"Perform a full eager compacting GC.  Answer the size of the largest free chunk."
	<returnTypeC: #usqLong>
	<inline: #never> "for profiling"
	needGCFlag := false.
	gcStartUsecs := coInterpreter ioUTCMicrosecondsNow.
	statMarkCount := 0.
	coInterpreter preGCAction: GCModeFull.
	self globalGarbageCollect.
	coInterpreter postGCAction: GCModeFull.
	statGCEndUsecs := coInterpreter ioUTCMicrosecondsNow.
	self updateFullGCStats.
	^(freeLists at: 0) ~= 0
		ifTrue: [self bytesInBody: self findLargestFreeChunk]
		ifFalse: [0]! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 12/16/2016 15:48'!
garbageCollectForSnapshot
	self flushNewSpace. "There is no place to put newSpace in the snapshot file."
	self flag: 'If we wanted to shrink the rememberedSet prior to snapshot this is the place to do it.'.
	compactor biasForSnapshot.
	self fullGC.
	compactor biasForGC.
	segmentManager prepareForSnapshot.
	self checkFreeSpace: GCModeFull! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/15/2013 09:23'!
gcStartUsecs
	^gcStartUsecs! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 5/22/2014 10:34'!
getHeapGrowthToSizeGCRatio
	<returnTypeC: #float>
	^heapGrowthToSizeGCRatio! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 10/26/2018 20:41'!
getStackPointer
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter getStackPointer! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 12/12/2013 11:30'!
getThisSessionID
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter getThisSessionID! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'cb 7/3/2018 09:15'!
globalGarbageCollect
	<inline: true> "inline into fullGC"
	self assert: self validObjStacks.
	self assert: (self isEmptyObjStack: markStack).
	self assert: (self isEmptyObjStack: weaklingStack).

	"Mark objects /before/ scavenging, to empty the rememberedTable of unmarked roots."
	self markObjects: true.
	gcMarkEndUsecs := coInterpreter ioUTCMicrosecondsNow.
	
	scavenger forgetUnmarkedRememberedObjects.
	self doScavenge: MarkOnTenure.

	"Mid-way the leak check must be more lenient.  Unmarked classes will have been
	 expunged from the table, but unmarked instances will not yet have been reclaimed."
	self runLeakCheckerFor: GCModeFull
		excludeUnmarkedObjs: true
		classIndicesShouldBeValid: true.

	compactionStartUsecs := coInterpreter ioUTCMicrosecondsNow.
	segmentManager prepareForGlobalSweep. "for notePinned:"
	compactor compact.
	self attemptToShrink.
	self setHeapSizeAtPreviousGC.

	self assert: self validObjStacks.
	self assert: (self isEmptyObjStack: markStack).
	self assert: (self isEmptyObjStack: weaklingStack).
	self assert: self allObjectsUnmarked.
	self runLeakCheckerFor: GCModeFull! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/3/2013 15:36'!
goodContextSize: oop
	| numSlots |
	numSlots := self numSlotsOf: oop.
	^numSlots = SmallContextSlots or: [numSlots = LargeContextSlots]! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 12/28/2016 13:06'!
greyBitShift
	<cmacro>
	"bit 2 of 3-bit field above format (little endian)"
	^31! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/16/2013 09:59'!
growHeadroom
	^growHeadroom! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/16/2013 09:58'!
growHeadroom: aValue
	^growHeadroom := aValue! !

!SpurMemoryManager methodsFor: 'growing/shrinking memory' stamp: 'eem 6/21/2021 12:03'!
growOldSpaceByAtLeast: minAmmount
	"Attempt to grow memory by at least minAmmount.
	 Answer the size of the new segment, or nil if the attempt failed."
	| ammount headroom total start interval |
	<var: #segInfo type: #'SpurSegmentInfo *'>
	"statGrowMemory counts attempts, not successes."
	statGrowMemory := statGrowMemory + 1."we need to include overhead for a new object header plus the segment bridge."
	ammount := minAmmount + (self baseHeaderSize * 2 + self bridgeSize).
	"round up to the nearest power of two."
	ammount := 1 << (ammount - 1) highBit.
	"and grow by at least growHeadroom."
	ammount := ammount max: growHeadroom.

	"Now apply the maxOldSpaceSize limit, if one is in effect."
	maxOldSpaceSize > 0 ifTrue:
		[total := segmentManager totalBytesInSegments.
		 total >= maxOldSpaceSize ifTrue:
			[needGCFlag := true.
			 ^nil].
		 headroom := maxOldSpaceSize - total.
		 headroom < ammount ifTrue:
			[headroom < (minAmmount + (self baseHeaderSize * 2 + self bridgeSize)) ifTrue:
				[needGCFlag := true.
				 ^nil].
			 ammount := headroom]].
		 
	start := coInterpreter ioUTCMicrosecondsNow.
	^(segmentManager addSegmentOfSize: ammount)
		ifNil: [needGCFlag := true. nil]
		ifNotNil:
			[:segInfo|
			 self assimilateNewSegment: segInfo.
			 "and add the new free chunk to the free list; done here
			  instead of in assimilateNewSegment: for the assert"
			 self addFreeChunkWithBytes: segInfo segSize - self bridgeSize at: segInfo segStart.
			 self assert: (self addressAfter: (self objectStartingAt: segInfo segStart))
						= (segInfo segLimit - self bridgeSize).
			 self checkFreeSpace: GCCheckFreeSpace.
			 segmentManager checkSegments.
			 interval := coInterpreter ioUTCMicrosecondsNow - start.
			 interval > statMaxAllocSegmentTime ifTrue: [statMaxAllocSegmentTime := interval].
			 segInfo segSize]! !

!SpurMemoryManager methodsFor: 'growing/shrinking memory' stamp: 'eem 3/11/2014 13:04'!
growToAccomodateContainerWithNumSlots: numSlots
	"Grow memory to accomodate a container (an Array) with numSlots.
	 Grow by at least the growHeadroom.  Supports allInstancesOf: and allObjects."
	| delta |
	delta := self baseHeaderSize * 2 + (numSlots * self bytesPerOop).
	self growOldSpaceByAtLeast: (growHeadroom max: delta)! !

!SpurMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 9/24/2020 20:01'!
hackSlimBridgeTo: firstEdenObject at: lastWordInPastSpace
	"This is a horrible hack for getting to the first object in eden if pastSpace is almost full.
	 If there is only one (64-bit) word at the end of pastSpace there is no room for a full
	 bridge, but there is room for this hack."
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 9/14/2015 10:35'!
hasIdentityHash: objOop
	<inline: true>
	^self cCode: [self rawHashBitsOf: objOop]
		inSmalltalk: [(self rawHashBitsOf: objOop) ~= 0]! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 10/7/2013 13:46'!
hasOverflowHeader: objOop
	^(self rawNumSlotsOf: objOop) = self numSlotsMask! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 3/21/2014 04:40'!
hasPointerFields: oop
	<inline: true>
	^(self isNonImmediate: oop)
	  and: [self hasPointerFieldsNonImm: oop]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 3/21/2014 04:07'!
hasPointerFieldsNonImm: oop
	<inline: true>
	^self isAnyPointerFormat: (self formatOf: oop)! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 2/12/2015 17:34'!
hasSixtyFourBitImmediates
	"The V3 64-bit memory manager has 64-bit oops, but its SmallIntegers only have
	 31 bits of precision.  The Spur 64-bit memory manager has 61-bit immediates."
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'api characterization' stamp: 'eem 9/9/2013 11:00'!
hasSpurMemoryManagerAPI
	^true! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 11/18/2013 15:31'!
hasYoungReferents: objOop
	0 to: (self numPointerSlotsOf: objOop) - 1 do:
		[:i| | oop |
		oop := self fetchPointer: i ofObject: objOop.
		(self isYoung: oop) ifTrue:
			[^true]].
	^false! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'cb 1/19/2017 18:34'!
hashBitsOf: objOop
	| hash |
	hash := self rawHashBitsOf: objOop.
	hash = 0 ifTrue:
		["would like to assert
			self assert: (coInterpreter addressCouldBeClassObj: objOop) not
		  but instance-specific behaviors that are instances of themselves may
		  fail this test."
		 hash := self newHashBitsOf: objOop].
	^hash! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 10/4/2020 12:29'!
headerForSlots: numSlots format: formatField classIndex: classIndex
	<api>
	"The header format in LSB is
	 MSB:	| 8: numSlots		| (on a byte boundary)
			| 2 bits				|	(msb,lsb = {isMarked,?})
			| 22: identityHash	| (on a word boundary)
			| 3 bits				|	(msb <-> lsb = {isGrey,isPinned,isRemembered}
			| 5: format			| (on a byte boundary)
			| 2 bits				|	(msb,lsb = {isImmutable,isExtraAligned})
			| 22: classIndex		| (on a word boundary) : LSB
	 The remaining bits (7) are used for
		isImmutable	(bit 23)
		isRemembered	(bit 29)
		isPinned		(bit 30)
		isGrey			(bit 31)
		isMarked		(bit 55)
	 leaving 2 unused bits, each next to a 22-bit field, allowing those fields to be
	 expanded to 23 bits..  The three bit field { isGrey, isPinned, isRemembered }
	 is for bits that are never set in young objects.  This allows the remembered
	 table to be pruned when full by using these bits as a reference count of
	 newSpace objects from the remembered table. Objects with a high count
	 should be tenured to prune the remembered table."
	<returnTypeC: #usqLong>
	<inline: true>
	^ ((self cCoerceSimple: numSlots to: #usqLong) << self numSlotsFullShift)
	+ (formatField << self formatShift)
	+ classIndex! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 3/25/2015 13:13'!
headerForSlots: numSlots hash: hash format: formatField classIndex: classIndex
	"The header format in LSB is
	 MSB:	| 8: numSlots		| (on a byte boundary)
			| 2 bits				|	(msb,lsb = {isMarked,?})
			| 22: identityHash	| (on a word boundary)
			| 3 bits				|	(msb <-> lsb = {isGrey,isPinned,isRemembered}
			| 5: format			| (on a byte boundary)
			| 2 bits				|	(msb,lsb = {isImmutable,?})
			| 22: classIndex		| (on a word boundary) : LSB
	 The remaining bits (7) are used for
		isImmutable	(bit 23)
		isRemembered	(bit 29)
		isPinned		(bit 30)
		isGrey			(bit 31)
		isMarked		(bit 55)
	 leaving 2 unused bits, each next to a 22-bit field, allowing those fields to be
	 expanded to 23 bits..  The three bit field { isGrey, isPinned, isRemembered }
	 is for bits that are never set in young objects.  This allows the remembered
	 table to be pruned when full by using these bits as a reference count of
	 newSpace objects from the remembered table. Objects with a high count
	 should be tenured to prune the remembered table."
	<returnTypeC: #usqLong>
	<inline: true>
	^ ((self cCoerceSimple: numSlots to: #usqLong) << self numSlotsFullShift)
	+ ((self cCoerceSimple: hash to: #usqLong) << self identityHashFullWordShift)
	+ (formatField << self formatShift)
	+ classIndex! !

!SpurMemoryManager methodsFor: 'method access' stamp: 'eem 6/7/2016 14:52'!
headerIndicatesAlternateBytecodeSet: methodHeader
	"A negative header selects the alternate bytecode set."
	<api>
	<inline: true>
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 9/16/2013 11:58'!
heapMap
	^heapMap! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 10/17/2013 05:47'!
hiddenRootSlots
	"Answer the number of extra root slots in the root of the hidden root object."
	^8! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 12/16/2016 16:51'!
hiddenRootsObject
	^hiddenRootsObj! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 10/15/2013 08:54'!
identityHashFieldWidth
	^22! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 11/18/2014 16:54'!
identityHashFullWordMask
	^16r3fffff00000000! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 12/12/2015 14:20'!
identityHashFullWordShift
	<api>
	<cmacro>
	^32! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 12/12/2015 14:18'!
identityHashHalfWordMask
	<api>
	<cmacro>
	^16r3fffff! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 12/11/2017 16:09'!
ifAProxy: objOop updateCopy: copy
	"This is part of storeImageSegmentInto:outPointers:roots:.
	 If the object being copied to the segment is weird and has exotic state,
	 i.e. a married context or a jitted method, update the copy with the vanilla state."

	(self isContext: objOop) ifTrue:
		[| numMediatedSlots |
		 (coInterpreter isMarriedOrWidowedContext: objOop)
			ifTrue:
				["Since the context is here via objectsReachableFromRoots: we know it cannot be divorced.
				  I'd like to assert coInterpreter checkIsStillMarriedContext: objOop currentFP: framePointer,
				  here but that requires access to framePointer."
				 numMediatedSlots := coInterpreter numSlotsOfMarriedContext: objOop.
				 0 to: numMediatedSlots - 1 do:
					[:i| | oop |
					 oop := coInterpreter fetchPointer: i ofMarriedContext: objOop.
					 self storePointerUnchecked: i ofObject: copy withValue: oop]]
			ifFalse:
				[numMediatedSlots := self numPointerSlotsOf: objOop].
		 "And make sure to nil the slots beyond the top of stack..."
		 numMediatedSlots to: (self numSlotsOf: objOop) - 1 do:
			[:i| self storePointerUnchecked: i ofObject: copy withValue: nilObj]]! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 6/20/2016 12:17'!
ifOopInvalidForBecome: oop errorCodeInto: aBlock
	"Evaluates aBlock with an appropriate error if oop is invalid for become."
	<inline: true>
	(self isImmediate: oop) ifTrue:
		[^aBlock value: PrimErrInappropriate].
	(self isPinned: oop) ifTrue:
		[^aBlock value: PrimErrObjectIsPinned].
	(self isObjImmutable: oop) ifTrue:
		[^aBlock value: PrimErrNoModification]! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 3/10/2016 12:41'!
imageFormatVersion
	"Return a magic constant that changes when the image format changes.
	 Since the image reading code uses this to detect byte ordering, one
	 must avoid version numbers that are invariant under byte reversal."
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 9/12/2014 16:21'!
imageSegmentVersion
	| wholeWord |
	"a more complex version that tells both the word reversal and the endianness of the machine
	 it came from.  Low half of word is e.g. 6521.  Top byte is top byte of #doesNotUnderstand: on
	 this machine. ($d on the Mac or $s on the PC)"

	wholeWord := self long32At: (self splObj: SelectorDoesNotUnderstand) + self baseHeaderSize. "first data word, 'does'"
	^coInterpreter imageFormatVersion bitOr: (wholeWord bitAnd: 16rFF000000)! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 3/7/2014 11:53'!
imageSizeToWrite
	"when asked, newSpace should be empty."
	self assert: self newSpaceIsEmpty.
	^segmentManager totalBytesInNonEmptySegments! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'cb 1/20/2016 10:49'!
immutableBitMask
	"mask the immutable bit in the base header word"
	<api>
	<option: #IMMUTABILITY>
	^ 1 << self immutableBitShift! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 12/28/2016 13:06'!
immutableBitShift
	<cmacro>
	"bit 1 of 2-bit field above classIndex (little endian)"
	^23! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'cb 6/7/2018 11:51'!
inFreeTreeReplace: treeNode with: newNode
	"Part of reorderReversedTreeList:.  Switch treeNode with newNode in
	 the tree, but do nothing to the list linked through freeChunkNextIndex."
	| relative |
	self storePointer: self freeChunkPrevIndex ofFreeChunk: newNode withValue: 0.
	"copy parent, smaller, larger"
	self freeChunkParentIndex to: self freeChunkLargerIndex do:
		[:i|
		relative := self fetchPointer: i ofFreeChunk: treeNode.
		i = self freeChunkParentIndex
			ifTrue:
				[relative = 0
					ifTrue: "update root to point to newNode"
						[self assert: (freeLists at: 0) = treeNode.
						 freeLists at: 0 put: newNode]
					ifFalse: "replace link from parent to treeNode with link to newNode."
						[self storePointer: (treeNode = (self fetchPointer: self freeChunkSmallerIndex ofFreeChunk: relative)
												ifTrue: [self freeChunkSmallerIndex]
												ifFalse: [self freeChunkLargerIndex])
							ofFreeChunk: relative
							withValue: newNode]]
			ifFalse:
				[relative ~= 0 ifTrue:
					[self assert: (self fetchPointer: self freeChunkParentIndex ofFreeChunk: relative) = treeNode.
					 self storePointer: self freeChunkParentIndex ofFreeChunk: relative withValue: newNode]].
		self storePointer: i ofFreeChunk: newNode withValue: relative.
		self storePointer: i ofFreeChunk: treeNode withValue: 0]! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 6/21/2021 12:04'!
inLineRunLeakCheckerFor: gcModes excludeUnmarkedObjs: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid
	<inline: true>
	(gcModes anyMask: checkForLeaks) ifTrue:
		[(gcModes anyMask: GCModeFull)
			ifTrue: [coInterpreter reverseDisplayFrom: 0 to: 7]
			ifFalse: [coInterpreter reverseDisplayFrom: 8 to: 15].
		 self clearLeakMapAndMapAccessibleObjects.
		 self asserta: (self checkHeapIntegrity: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid).
		 self asserta: coInterpreter checkInterpreterIntegrity = 0.
		 self asserta: coInterpreter checkStackIntegrity.
		 self asserta: (coInterpreter checkCodeIntegrity: gcModes).
		 (gcModes anyMask: GCCheckFreeSpace) ifTrue:
			[self clearLeakMapAndMapAccessibleFreeSpace.
			 self asserta: self checkHeapFreeSpaceIntegrity]]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
inOrderPrintFreeTree: freeChunk printList: printNextList
	"print free chunks in freeTree in order."
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| next |
	(next := self fetchPointer: self freeChunkSmallerIndex ofFreeChunk: freeChunk) ~= 0 ifTrue:
		[self inOrderPrintFreeTree: next printList: printNextList].
	self printFreeChunk: freeChunk printAsTreeNode: true.
	printNextList ifTrue:
		[next := freeChunk.
		 [(next := self fetchPointer: self freeChunkNextIndex ofFreeChunk: next) ~= 0] whileTrue:
			[coInterpreter tab.
			 self printFreeChunk: next printAsTreeNode: false]].
	(next := self fetchPointer: self freeChunkLargerIndex ofFreeChunk: freeChunk) ~= 0 ifTrue:
		[self inOrderPrintFreeTree: next printList: printNextList]! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 1/8/2016 22:33'!
inPlaceBecome: obj1 and: obj2 copyHashFlag: copyHashFlag
	<inline: #never> "in an effort to fix a compiler bug with two-way become post r3427"
	"Do become in place by swapping object contents."
	| o1HasYoung o2HasYoung fmt |
	self assert: (self numSlotsOf: obj1) = (self numSlotsOf: obj2).
	self assert: ((self rawHashBitsOf: obj1) = 0
				 or: [(self classOrNilAtIndex: (self rawHashBitsOf: obj1)) ~= obj1]).
	self assert: ((self rawHashBitsOf: obj2) = 0
				 or: [(self classOrNilAtIndex: (self rawHashBitsOf: obj2)) ~= obj2]).
	"swap headers, but swapping headers swaps remembered bits and hashes;
	 remembered bits must be unswapped and hashes may be unswapped if
	 copyHash is false."
	false
		ifTrue: [self naiveSwapHeaders: obj1 and: obj2 copyHashFlag: copyHashFlag]
		ifFalse: [self cleverSwapHeaders: obj1 and: obj2 copyHashFlag: copyHashFlag].
	o1HasYoung := o2HasYoung := false.
	0 to: (self numSlotsOf: obj1) - 1 do:
		[:i| | temp1 temp2 |
		temp1 := self fetchPointer: i ofObject: obj1.
		temp2 := self fetchPointer: i ofObject: obj2.
		self storePointerUnchecked: i ofObject: obj1 withValue: temp2.
		self storePointerUnchecked: i ofObject: obj2 withValue: temp1.
		(self isYoung: temp2) ifTrue:
			[o1HasYoung := true].
		(self isYoung: temp1) ifTrue:
			[o2HasYoung := true]].
	(o1HasYoung and: [self isOldObject: obj1]) ifTrue:
		[fmt := self formatOf: obj1.
		 (self isPureBitsFormat: fmt) ifFalse:
			[self possibleRootStoreInto: obj1]].
	(o2HasYoung and: [self isOldObject: obj2]) ifTrue:
		[fmt := self formatOf: obj2.
		 (self isPureBitsFormat: fmt) ifFalse:
			[self possibleRootStoreInto: obj2]]! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 10/18/2013 11:29'!
inactiveOrFailedToDeferScan: anEphemeron
	"Answer whether an ephemeron is inactive (has a marked key) or,
	 if active, failed to fit on the unscanned ephemerons stack."
	| key |
	self assert: (self isEphemeron: anEphemeron).
	((self isImmediate: (key := self keyOfEphemeron: anEphemeron))
	 or: [self isMarked: key]) ifTrue:
		[^true].
	^(self pushOnUnscannedEphemeronsStack: anEphemeron) not! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'cb 10/5/2018 12:34'!
increaseFreeOldSpaceBy: bytes 
	totalFreeOldSpace := totalFreeOldSpace + bytes
! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 4/15/2021 11:51'!
incrementalGC
	"The selector incrementalGC refers to the V3 incremental collector, which is the collector for young objects, equivalent in function to the scavenger.
	 The Spur incremental collector is called the global incremental garbage collector or globalIGC for short."
	self shouldNotImplement! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 7/7/2017 11:26'!
indexOf: anElement in: anObject
	<api>
	| fmt numSlots |
	fmt := self formatOf: anObject.

	fmt <= self lastPointerFormat ifTrue:
		[numSlots := self numSlotsOf: anObject.
		 0 to: numSlots do:
			[:i| anElement = (self fetchPointer: i ofMaybeForwardedObject: anObject) ifTrue: [^i]].
		-1].

	fmt >= self firstByteFormat ifTrue:
		[fmt >= self firstCompiledMethodFormat ifTrue:
			[^self primitiveFailFor: PrimErrUnsupported].
		 numSlots := self numBytesOfBytes: anObject.
		 0 to: numSlots do:
			[:i| (self fetchByte: i ofObject: anObject) ifTrue: [^i]].
		-1].

	fmt >= self firstShortFormat ifTrue:
		[numSlots := self num16BitUnitsOf: anObject.
		 0 to: numSlots do:
			[:i| anElement = (self fetchUnsignedShort16: i ofObject: anObject) ifTrue: [^i]].
		-1].

	fmt = self sixtyFourBitIndexableFormat ifTrue:
		[numSlots := self num64BitUnitsOf: anObject.
		 0 to: numSlots do:
			[:i| anElement = (self fetchLong64: i ofObject: anObject) ifTrue: [^i]].
		-1].

	fmt >= self firstLongFormat ifTrue:
		[numSlots := self num32BitUnitsOf: anObject.
		 0 to: numSlots do:
			[:i| anElement = (self fetchLong32: i ofObject: anObject) ifTrue: [^i]].
		-1].

	^-1! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 12/12/2015 14:18'!
indexablePointersFormat
	<api>
	<cmacro>
	^3! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 9/19/2013 09:15'!
initFreeChunkWithBytes: numBytes at: address
	<var: #numBytes type: #usqLong>
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'segments' stamp: 'eem 9/24/2020 14:49'!
initSegmentBridgeWithBytes: numBytes at: address
	<var: #numBytes type: #usqLong>
	"Must have room for a double header or a short object with the forwarding slot (16 bytes either way)."
	self assert: (numBytes \\ self allocationUnit = 0
				 and: [numBytes >= (self baseHeaderSize + self baseHeaderSize)]).

	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 3/23/2015 12:17'!
initialHeadroom: extraVmMemory givenFreeOldSpaceInImage: freeOldSpaceInImage
	"Answer how much headroom to allocate, if any, on loading the image.
	 If the image already conatins lots of free space, we should not allocate lots more."
	<inline: true>
	| headroom |
	headroom := extraVmMemory = 0
					ifTrue: [growHeadroom ifNil: [16*1024*1024]]
					ifFalse: [extraVmMemory].
	freeOldSpaceInImage >= headroom ifTrue:
		[^0].
	freeOldSpaceInImage >= (headroom * 7 // 8) ifTrue:
		[^headroom // 8].
	freeOldSpaceInImage >= (headroom * 3 // 4) ifTrue:
		[^headroom // 4].
	freeOldSpaceInImage >= (headroom * 5 // 8) ifTrue:
		[^headroom * 3 // 8].
	freeOldSpaceInImage >= (headroom // 2) ifTrue:
		[^headroom // 2].
	^headroom! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 10/4/2013 14:21'!
initialInstanceOf: classObj
	<inline: false>
	| classIndex |
	classIndex := self rawHashBitsOf: classObj.
	classIndex = 0 ifTrue:
		[^nil].
	"flush instances in newSpace to settle the enumeration."
	self flushNewSpaceInstancesOf: classObj.
	self allObjectsDo:
		[:objOop|
		classIndex = (self classIndexOf: objOop) ifTrue:
			[^objOop]].
	^nil! !

!SpurMemoryManager methodsFor: 'initialization' stamp: 'eem 9/25/2020 07:46'!
initialize
	"We can put all initializations that set something to 0 or to false here.
	 In C all global variables are initialized to 0, and 0 is false."
	| moreThanEnough |
	remapBuffer := Array new: RemapBufferSize.
	remapBufferCount := extraRootCount := 0. "see below"
	freeListsMask := totalFreeOldSpace := lowSpaceThreshold := 0.
	checkForLeaks := 0.
	needGCFlag := signalLowSpace := marking := false.
	becomeEffectsFlags := gcPhaseInProgress := validatedIntegerClassFlags := 0.
	statScavenges := statIncrGCs := statFullGCs := 0.
	statMaxAllocSegmentTime := 0.
	statMarkUsecs := statSweepUsecs := statScavengeGCUsecs := statIncrGCUsecs := statFullGCUsecs := statCompactionUsecs := statGCEndUsecs := gcSweepEndUsecs := 0.
	statSGCDeltaUsecs := statIGCDeltaUsecs := statFGCDeltaUsecs := 0.
	statGrowMemory := statShrinkMemory := statRootTableCount := statAllocatedBytes := 0.
	statRootTableOverflows := statMarkCount := statCompactPassCount := statCoalesces := 0.

	"We can initialize things that are allocated but are lazily initialized."
	unscannedEphemerons := SpurContiguousObjStack new.

	"we can initialize things that are virtual in C."
	scavenger := SpurGenerationScavenger simulatorClass new manager: self; yourself.
	segmentManager := SpurSegmentManager simulatorClass new manager: self; yourself.
	compactor := self class compactorClass simulatorClass new manager: self; yourself.

	"We can also initialize here anything that is only for simulation."
	heapMap := CogCheck32BitHeapMap new.

	"N.B. We *don't* initialize extraRoots because we don't simulate it."

	"This is needed on 64-bits. We don't want a simulation creating a huge heap by default.
	 By default use 512Mb on 64-bits, 256Mb on 32-bits."
	moreThanEnough := 1024 * 1024 * 1024 / (16 / self wordSize). "One million dollars, ha ha ha ha ha,... ha, ha ha ha ha, ..."
	maxOldSpaceSize := self class initializationOptions
							ifNotNil: [:initOpts| initOpts at: #maxOldSpaceSize ifAbsent: [moreThanEnough]]
							ifNil: [moreThanEnough]! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 11/28/2019 12:39'!
initializeFreeSpacePostLoad: freeListObj
	"Reinitialize the free list info.  The freeLists object needs to be swizzled
	 because its neither a free, nor a pointer object.  Free objects have already
	 been swizzled in adjustAllOopsBy:"
	
	self assert: (self numSlotsOf: freeListObj) = self numFreeLists.
	self assert: (self formatOf: freeListObj) = self wordIndexableFormat.
	freeLists := self firstIndexableField: freeListObj.
	freeListsMask := 0.
	0 to: self numFreeLists - 1 do:
		[:i|
		(freeLists at: i) ~= 0 ifTrue:
			[freeListsMask := freeListsMask bitOr: (1 << i).
			 freeLists at: i put: (segmentManager swizzleObj: (freeLists at: i) in: freeListObj)]]! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 10/18/2013 10:07'!
initializeMarkStack
	self ensureRoomOnObjStackAt: MarkStackRootIndex! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 3/18/2020 12:59'!
initializeNewSpaceVariables
	<inline: #never>
	freeStart := scavenger eden start.
	pastSpaceStart := scavenger pastSpace start.
	scavengeThreshold := scavenger eden limit
							- (scavenger edenBytes // 64)
							- coInterpreter interpreterAllocationReserveBytes.
	newSpaceStart := scavenger pastSpace start min: scavenger futureSpace start.
	self assert: newSpaceStart < scavenger eden start! !

!SpurMemoryManager methodsFor: 'initialization' stamp: 'cb 6/7/2018 15:16'!
initializeObjectMemory: bytesToShift
	"Initialize object memory variables at startup time. Assume endOfMemory at al are
	 initialised by the image-reading code via setHeapBase:memoryLimit:endOfMemory:.
	 endOfMemory is assumed to point to the end of the last object in the image.
	 Assume: image reader also initializes the following variables:
		specialObjectsOop
		lastHash"
	<inline: false>
	| freeListObj |
	"Catch mis-initializations leading to bad translations to C"
	self assert: self baseHeaderSize = self baseHeaderSize.
	self assert: (self maxSlotsForAlloc * self wordSize) asInteger > 0.
	self bootstrapping ifFalse:
		[self
			initSegmentBridgeWithBytes: self bridgeSize
			at: endOfMemory - self bridgeSize].
	segmentManager adjustSegmentSwizzlesBy: bytesToShift.
	"image may be at a different address; adjust oops for new location"
	self adjustAllOopsBy: bytesToShift.
	specialObjectsOop := segmentManager swizzleObj: specialObjectsOop.

	"heavily used special objects"
	nilObj		:= self splObj: NilObject.
	falseObj	:= self splObj: FalseObject.
	trueObj		:= self splObj: TrueObject.

	"In Cog we insist that nil, true & false are next to each other (Cogit generates tighter
	 conditional branch code as a result).  In addition, Spur places the free lists and
	 class table root page immediately following them."
	self assert: nilObj = oldSpaceStart.
	self assert: falseObj = (self oldSpaceObjectAfter: nilObj).
	self assert: trueObj = (self oldSpaceObjectAfter: falseObj).
	freeListObj := self oldSpaceObjectAfter: trueObj.
	self setHiddenRootsObj: (self oldSpaceObjectAfter: freeListObj).
	markStack := self swizzleObjStackAt: MarkStackRootIndex.
	weaklingStack := self swizzleObjStackAt: WeaklingStackRootIndex.
	mournQueue := self swizzleObjStackAt: MournQueueRootIndex.
	self assert: self validObjStacks.
	self assert: (self isEmptyObjStack: markStack).
	self assert: (self isEmptyObjStack: weaklingStack).

	self initializeFreeSpacePostLoad: freeListObj.
	segmentManager collapseSegmentsPostSwizzle.
	self updateFreeLists.
	self computeFreeSpacePostSwizzle.
	compactor postSwizzleAction.
	self initializeOldSpaceFirstFree: freeOldSpaceStart. "initializes endOfMemory, freeStart, free space"
	self initializeNewSpaceVariables.
	scavenger initializeRememberedSet.
	segmentManager checkSegments.
	compactor biasForGC.

	"These defaults should depend on machine size; e.g. too small on a powerful laptop, too big on a Pi."
	growHeadroom := 16*1024*1024.		"headroom when growing"
	shrinkThreshold := 32*1024*1024.		"free space before shrinking"
	self setHeapSizeAtPreviousGC.
	heapGrowthToSizeGCRatio := 0.333333. "By default GC after scavenge if heap has grown by a third since the last GC"! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 6/21/2021 12:04'!
initializeOldSpaceFirstFree: startOfFreeOldSpace
	<var: 'startOfFreeOldSpace' type: #usqInt>
	| limit freeOldStart freeChunk |
	<var: 'limit' type: #usqInt>
	<var: 'freeOldStart' type: #usqInt>
	limit := endOfMemory - self bridgeSize.
	limit > startOfFreeOldSpace ifTrue:
		[totalFreeOldSpace := totalFreeOldSpace + (limit - startOfFreeOldSpace).
		 freeOldStart := startOfFreeOldSpace.
		 self wordSize > 4 ifTrue:
			[[limit - freeOldStart >= (1 << 32)] whileTrue:
				[freeChunk := self freeChunkWithBytes: (1 << 32) at: freeOldStart.
				 freeOldStart := freeOldStart + (1 << 32).
				 self assert: freeOldStart = (self addressAfter: freeChunk)]].
		freeOldStart < limit ifTrue:
			[freeChunk := self freeChunkWithBytes: limit - freeOldStart at: freeOldStart.
			 self assert: (self addressAfter: freeChunk) = limit]].
	endOfMemory := endOfMemory - self bridgeSize.
	freeOldSpaceStart := endOfMemory.
	self checkFreeSpace: GCCheckFreeSpace! !

!SpurMemoryManager methodsFor: 'spur bootstrap' stamp: 'eem 12/11/2013 14:52'!
initializePostBootstrap
	"The heap has just been bootstrapped into a modified newSpace occupying all of memory
	 above newSpace (and the codeZone). Put things back to some kind of normalcy."
	freeOldSpaceStart := freeStart.
	freeStart := scavenger eden start.
	pastSpaceStart := scavenger pastSpace start.
	scavengeThreshold := scavenger eden limit - (scavenger edenBytes // 64)! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 2/3/2014 16:28'!
initializeUnscannedEphemerons
	"Initialize unscannedEphemerons to use the largest free chunk
	 or unused eden space, which ever is the larger."
	
	| largestFree sizeOfUnusedEden |
	largestFree := self findLargestFreeChunk.
	sizeOfUnusedEden := scavenger eden limit - freeStart.
	(largestFree notNil
	 and: [(self numSlotsOfAny: largestFree) > (sizeOfUnusedEden / self wordSize)])
		ifTrue:
			[unscannedEphemerons
				start: largestFree
					+ self baseHeaderSize
					+ (self freeChunkLargerIndex + 1 * self wordSize);
				limit: (self addressAfter: largestFree)]
		ifFalse:
			[unscannedEphemerons
				start: freeStart;
				limit: scavenger eden limit].
	unscannedEphemerons top: unscannedEphemerons start! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 10/31/2013 11:57'!
initializeWeaklingStack
	self ensureRoomOnObjStackAt: WeaklingStackRootIndex! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 7/23/2014 09:08'!
innerBecomeObjectsIn: array1 and: array2 copyHash: copyHashFlag
	"Inner loop of two-way become."
	0 to: (self numSlotsOf: array1) - 1 do:
		[:i| | obj1 obj2 |
		"At first blush it would appear unnecessary to use followField: here since
		 the validation in become:with:twoWay:copyHash: follows forwarders.  But
		 there's nothing to ensure all elements of each array are unique and don't
		 appear in the other array.  So the enumeration could encounter an object
		 already becommed earlier in the same enumeration."
		obj1 := self followField: i ofObject: array1.
		obj2 := self followField: i ofObject: array2.
		obj1 ~= obj2 ifTrue:
			[self doBecome: obj1 and: obj2 copyHash: copyHashFlag.
			 self followField: i ofObject: array1.
			 self followField: i ofObject: array2]]! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 5/19/2016 12:30'!
innerBecomeObjectsIn: array1 to: array2 copyHash: copyHashFlag
	"Inner loop of one-way become."
	0 to: (self numSlotsOf: array1) - 1 do:
		[:i| | obj1 obj2 |
		"At first blush it would appear unnecessary to use followField: here since
		 the validation in become:with:twoWay:copyHash: follows forwarders.  But
		 there's nothing to ensure all elements of each array is unique and doesn't
		 appear in the other array.  So the enumeration could encounter an object
		 already becommed earlier in the same enumeration."
		obj1 := self followField: i ofObject: array1.
		obj2 := self followField: i ofObject: array2.
		obj1 ~= obj2 ifTrue:
			[self doBecome: obj1 to: obj2 copyHash: copyHashFlag.
			 self followField: i ofObject: array1.
			 self assert: (self isOopForwarded: obj2) not]]! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 12/17/2013 16:46'!
instSpecForImmediateClasses
	"Use the format for forwarded objects for immediate classes.  Immediate classes (SmallInteger,
	 Character, etc) can't be instantiated, so the inst spec should be an invalid one to cause the
	 instantiation primitives to fail.  The forwardedFormat is internal to the SpurMemoryManager,
	 used to mark forwarded objects only, and so is suitable."
	^self forwardedFormat! !

!SpurMemoryManager methodsFor: 'object format' stamp: 'eem 8/7/2014 04:20'!
instSpecOfClass: classPointer
	"This field in a class's format inst var corresponds to the 5-bit format field stored in every object header"

	^self instSpecOfClassFormat: (self formatOfClass: classPointer)! !

!SpurMemoryManager methodsFor: 'object format' stamp: 'eem 7/12/2016 13:55'!
instSpecOfClassFormat: classFormat
	<api>
	^classFormat >> self fixedFieldsFieldWidth bitAnd: self formatMask! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 10/29/2014 10:45'!
instanceAfter: objOop
	| actualObj classIndex |
	actualObj := objOop.
	classIndex := self classIndexOf: objOop.

	(self isInEden: objOop) ifTrue:
		[[actualObj := self objectAfter: actualObj limit: freeStart.
		  self oop: actualObj isLessThan: freeStart] whileTrue:
			[classIndex = (self classIndexOf: actualObj) ifTrue:
				[^actualObj]].
		 actualObj := (self oop: pastSpaceStart isGreaterThan: scavenger pastSpace start)
						ifTrue: [self objectStartingAt: scavenger pastSpace start]
						ifFalse: [nilObj]].

	(self isInPastSpace: actualObj) ifTrue:
		[[actualObj := self objectAfter: actualObj limit: pastSpaceStart.
		  self oop: actualObj isLessThan: pastSpaceStart] whileTrue:
			[classIndex = (self classIndexOf: actualObj) ifTrue:
				[^actualObj]].
		 actualObj := nilObj].

	[actualObj := self objectAfter: actualObj limit: endOfMemory.
	 self oop:actualObj isLessThan: endOfMemory] whileTrue:
		[classIndex = (self classIndexOf: actualObj) ifTrue:
			[^actualObj]].
	^nil! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 9/10/2013 15:52'!
instanceSizeOf: classObj
	<api>
	"Answer the number of slots in a class.  For example the instanceSizeOf: 
	 ClassPoint is 2, for the x & y slots. The instance size of non-pointer classes is 0."
	self assert: (coInterpreter addressCouldBeClassObj: classObj).

	^(self formatOfClass: classObj) bitAnd: self fixedFieldsOfClassFormatMask! !

!SpurMemoryManager methodsFor: 'instantiation' stamp: 'eem 9/11/2013 05:14'!
instantiateClass: classObj
	| instSpec classFormat numSlots classIndex newObj |
	classFormat := self formatOfClass: classObj.
	instSpec := self instSpecOfClassFormat: classFormat.
	(self isFixedSizePointerFormat: instSpec) ifFalse:
		[^nil].
	classIndex := self ensureBehaviorHash: classObj.
	classIndex < 0 ifTrue:
		[coInterpreter primitiveFailFor: classIndex negated.
		 ^nil].
	numSlots := self fixedFieldsOfClassFormat: classFormat.
	newObj := self allocateSlots: numSlots format: instSpec classIndex: classIndex.
	newObj ifNotNil:
		[self fillObj: newObj numSlots: numSlots with: nilObj].
	^newObj! !

!SpurMemoryManager methodsFor: 'instantiation' stamp: 'eem 8/12/2014 19:52'!
instantiateClass: classObj indexableSize: nElements
	<var: #nElements type: #usqInt>
	"Allocate an instance of a variable class, excepting CompiledMethod."
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'instantiation' stamp: 'eem 8/12/2014 19:53'!
instantiateCompiledMethodClass: classObj indexableSize: nElements
	<var: #nElements type: #usqInt>
	"Allocate an instance of a CompiledMethod class."
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'immediates' stamp: 'eem 9/3/2013 17:41'!
integerObjectOf: value
	"Convert the integer value, assumed to be in SmallInteger range, into a tagged SmallInteger object.
	 In C, use a shift and an add to set the tag bit.
	 In Smalltalk we have to work harder because the simulator works with strictly positive bit patterns."
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'immediates' stamp: 'eem 11/19/2014 21:16'!
integerObjectOfCharacterObject: oop
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'immediates' stamp: 'eem 9/3/2013 17:29'!
integerValueOf: oop
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 9/19/2013 16:01'!
interpreter
	<doNotGenerate>
	^coInterpreter! !

!SpurMemoryManager methodsFor: 'initialization' stamp: 'eem 10/18/2020 20:57'!
invalidCompactClassError: className
	<var: 'className' type: #'const char *'>
	<inline: false>
	self cCode:
			['\nClass %s does not have the required class index\n' printf: className.
			 self exit: -1]
		inSmalltalk:
			[self error: 'Class ', className, ' does not have the required class index']! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 1/9/2014 10:37'!
ioLoadFunction: functionName From: moduleName
	"hack around the CoInterpreter/ObjectMemory split refactoring.
	 provide accurate types for the VMPluginCodeGenerator."
	<doNotGenerate>
	<returnTypeC: #'void *'>
	<var: #functionName type: #'char *'>
	<var: #moduleName type: #'char *'>
	^coInterpreter ioLoadFunction: functionName From: moduleName! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 2/18/2014 15:07'!
ioLoadModule: moduleNameIndex OfLength: moduleLength
	<doNotGenerate>
	<returnTypeC: #'void *'>
	"Dummy - provided by support code"
	^0! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 10/25/2013 20:33'!
ioUTCMicrosecondsNow
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter ioUTCMicrosecondsNow! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/19/2013 11:07'!
is: oop KindOf: classNameString
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter is: oop KindOf: classNameString! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 8/22/2018 12:06'!
is: oop KindOfClass: aClass
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter is: oop KindOfClass: aClass! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 7/25/2014 21:42'!
is: oop instanceOf: classOop
	"Answer if oop is an instance of the given class."

	<inline: true>
	| tag |
	tag := self fetchClassTagOf: oop.
	^tag = (self rawHashBitsOf: classOop)! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 9/5/2013 13:29'!
is: oop instanceOf: classOop compactClassIndex: compactClassIndex
	"Answer if oop is an instance of the given class. If the class has a (non-zero)
	 compactClassIndex use that to speed up the check.  N.B. Inlining should
	 result in classOop not being accessed if oop's compact class index and
	 compactClassIndex are non-zero."

	<inline: true>
	(self isImmediate: oop) ifTrue:
		[^false].

	^self isClassOfNonImm: oop equalTo: classOop compactClassIndex: compactClassIndex! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 5/25/2016 21:30'!
is: oop onObjStack: objStack
	<inline: false>
	| index nextPage |
	objStack = nilObj ifTrue:
		[^false].
	self assert: (self numSlotsOfAny: objStack) = ObjStackPageSlots.
	"There are four fixed slots in an obj stack, and a Topx of 0 indicates empty, so
	  if there were 6 slots in an oop stack, full would be 2, and the last 0-rel index is 5."
	index := (self fetchPointer: ObjStackTopx ofObject: objStack) + ObjStackNextx.
	[index >= ObjStackFixedSlots] whileTrue:
		[oop = (self fetchPointer: index ofObject: objStack) ifTrue:
			[^true].
		 index := index - 1].
	nextPage := self fetchPointer: ObjStackNextx ofObject: objStack.
	nextPage ~= 0 ifTrue:
		[(self is: oop onObjStack: nextPage) ifTrue:
			[^true]].
	^false! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 7/11/2017 19:39'!
is: hash outPointerClassHashFor: oop in: outPointerArray limit: outIndex
	"This is part of storeImageSegmentInto:outPointers:roots:.
	 suspect; what about false positives?"
	^(hash anyMask: TopHashBit)
	  and: [hash - TopHashBit <= outIndex
	  and: [oop = (self fetchPointer: hash - TopHashBit ofObject: outPointerArray)]]! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 3/21/2014 04:07'!
isAnyPointerFormat: format
	"the inverse of isPureBitsFormat:"
	<inline: true>
	^format <= self lastPointerFormat or: [format >= self firstCompiledMethodFormat]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/9/2013 16:42'!
isArray: oop
	"Answer true if this is an indexable object with pointer elements, e.g., an array"
	^(self isNonImmediate: oop) and: [self isArrayNonImm: oop]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 7/29/2014 20:14'!
isArrayNonImm: oop
	<api>
	"Answer if this is an indexable object with pointer elements, e.g., an array"
	^ (self formatOf: oop) = self arrayFormat! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 12/18/2018 18:08'!
isBooleanObject: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter isBooleanObject: oop! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 9/24/2020 12:49'!
isBridgeOrEnumerableObjectNoAssert: objOop
	"Answer if objOop should be included in an allObjects...Do: enumeration.
	 This is for assert-checking only."
	| classIndex |
	classIndex := self classIndexOf: objOop.
	^classIndex >= self isForwardedObjectClassIndexPun
	  	ifTrue: [classIndex < (numClassTablePages * self classTablePageSize)]
		ifFalse: [classIndex = self segmentBridgePun]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/3/2013 14:51'!
isBytes: oop
	"Answer true if the argument contains indexable bytes. See comment in formatOf:"
	"Note: Includes CompiledMethods."
	^(self isNonImmediate: oop) and: [self isBytesNonImm: oop]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/3/2013 14:51'!
isBytesNonImm: objOop
	"Answer true if the argument contains indexable bytes. See comment in formatOf:"
	^(self formatOf: objOop) >= self firstByteFormat! !

!SpurMemoryManager methodsFor: 'immediates' stamp: 'eem 10/14/2013 20:05'!
isCharacterObject: oop
	<api>
	^self isImmediateCharacter: oop! !

!SpurMemoryManager methodsFor: 'immediates' stamp: 'eem 10/14/2013 20:04'!
isCharacterValue: anInteger
	<api>
	^self isInRangeCharacterCode: anInteger! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 1/8/2015 12:43'!
isClassAtUniqueIndex: aClass
	"Answer if aClass exists at only one index in the class table.  Be careful not to
	 be misled by classes that have puns, such as Array."
	| expectedIndex |
	expectedIndex := self rawHashBitsOf: aClass.
	self classTableEntriesDo:
		[:entry :index|
		 (entry = aClass
		  and: [index ~= expectedIndex
		  and: [index > self lastClassIndexPun]]) ifTrue:
			[^false]].
	^true! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 3/17/2015 13:24'!
isClassOfNonImm: objOop equalTo: classOop
	<inline: true>
	self assert: (self isNonImmediate: objOop).
	^(self classIndexOf: objOop) = (self rawHashBitsOf: classOop)! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 3/6/2015 10:19'!
isClassOfNonImm: oop equalTo: classOop compactClassIndex: knownClassIndex
	"Answer if the given (non-immediate) object is an instance of the given class
	 that may have a knownClassIndex (if knownClassIndex is non-zero).  This method
	 is misnamed given SPur's architecture (where all objects have ``compact'' class indices)
	 but is so-named for compatibility with ObjectMemory.
	 N.B. Inlining and/or compiler optimization should result in classOop not being
	 accessed if knownClassIndex is non-zero."

	| ccIndex |
	<inline: true>
	self assert: (self isImmediate: oop) not.

	ccIndex := self classIndexOf: oop.
	knownClassIndex ~= 0
		ifTrue:
			[^knownClassIndex = ccIndex]
		ifFalse:
			[^classOop = (self classAtIndex: ccIndex)]! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 2/28/2017 10:44'!
isCompactInstance: oop
	"For assert checking"
	^false! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 12/1/2014 11:31'!
isCompiledMethod: objOop
    "Answer whether the argument object is of compiled method format"
	<api>
    ^(self formatOf: objOop) >= self firstCompiledMethodFormat! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 8/9/2014 10:09'!
isCompiledMethodFormat: format
	^format >= self firstCompiledMethodFormat! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 10/1/2013 14:27'!
isCompiledMethodHeader: objHeader
    "Answer whether the argument header has compiled method format"
    ^(self formatOfHeader: objHeader) >= self firstCompiledMethodFormat! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/3/2013 22:03'!
isContext: oop
	<inline: true>
	^(self isNonImmediate: oop)
	   and: [(self classIndexOf: oop) = ClassMethodContextCompactIndex]! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 9/11/2013 07:31'!
isContextHeader: aHeader
	<inline: true>
	^(self classIndexOfHeader: aHeader) = ClassMethodContextCompactIndex! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/4/2013 11:36'!
isContextNonImm: oop
	<inline: true>
	^(self classIndexOf: oop) = ClassMethodContextCompactIndex! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 7/11/2017 19:39'!
isCopiedIntoSegment: anObjectInTheHeap
	"This is part of storeImageSegmentInto:outPointers:roots:."
	<inline: true>
	^self isMarked: anObjectInTheHeap! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 5/15/2015 10:25'!
isEmptyObjStack: objStack
	objStack = nilObj ifTrue:
		[^true].
	self eassert: [self isValidObjStack: objStack].
	^0 = (self fetchPointer: ObjStackTopx ofObject: objStack)
	  and: [0 = (self fetchPointer: ObjStackNextx ofObject: objStack)]! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 9/24/2020 14:35'!
isEnumerableObject: objOop
	"Answer if objOop should be included in an allObjects...Do: enumeration.
	 Non-objects should be excluded; these are bridges and free chunks."
	| classIndex |
	<inline: true>
	classIndex := self classIndexOf: objOop.
	self assert: (classIndex = self segmentBridgePun
				or: [classIndex = self isForwardedObjectClassIndexPun
				or: [(self long64At: objOop) ~= 0
					 and: [classIndex < (numClassTablePages * self classTablePageSize)]]]).
	^classIndex >= self isForwardedObjectClassIndexPun! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 3/11/2014 16:10'!
isEnumerableObjectNoAssert: objOop
	"Answer if objOop should be included in an allObjects...Do: enumeration.
	 This is for assert-checking only."
	| classIndex |
	classIndex := self classIndexOf: objOop.
	^classIndex >= self isForwardedObjectClassIndexPun
	  and: [classIndex < (numClassTablePages * self classTablePageSize)]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 3/21/2014 04:04'!
isEphemeron: objOop
	self assert: (self isNonImmediate: objOop).
	^self isEphemeronFormat: (self formatOf: objOop)! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 3/21/2014 03:57'!
isEphemeronFormat: format
	<inline: true>
	^format = self ephemeronFormat! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 10/4/2020 15:04'!
isExtraAligned: objOop
	"An object that has greater than 8 byte alignment has the isExtraAligned but set and is (must be) preceded
	 by a slimbridge.  The actual alignment the object has is held in the slimbridge.  Hence an object already
	 aligned on an N word boundary, for N > 1, must be preceded by an N word slimbridge.  We support
	 alignment slimbridges by noting that an overflow size word must have a slot count of at least 255. Any value
	 less than 128 indicates a slim bridge, which gives us 7 bits, organized as a 3-bit alignment field (in bits 4 to 6),
	 and a 4-bit bridge size field (in bIts 0 to 3).  The bridge size field gives the size minus one in words of the
	 slimbridge.  A size of zero specifies an 8-byte, one word slimbridge, a size of 15 implies an 128 byte, 16 word
	 slimbridge. The two bit alignment field gives the power of two minus three in bytes.  An alignment field of 0
	 imples 8 byte alignment (and hence is only used with single word slimbridges). An alignment field of 1 imples
	 16 byte alignment (and hence may be used with one or two word slimbridges). An alignment field of 2 imples
	 32 byte alignment (and hence may be used with one to four word slimbridges). An alignment field of 3 imples
	 64 byte alignment (and hence may be used with one to eight word slimbridges). An alignment field of 4 imples
	 128 byte alignment (and hence may be used with one to sixteed word slimbridges). Alignments of 256, 512,
	 and 1024 bytes are representable, but require a larger size field than can fit in four bits, and so are not used.

	 So to align an object on a 16, 32, or 128 byte boundary, at allocation time a slimbridge of the size required
	 to pad to the required alignment is written into memory, and immediatrly following it the object is allocated,
	 with its isExtraAligned bit set. The required/actual alignment is written into the slimbridge's alignment field,
	 and the entire field (saturated size count and slot count < 128) replicated in all words of the slimbridge.
	 When the object is moved, either during scavenging or compaction, when it is written into its new location,
	 a new slimbridge is synthesized to align it in the new location on the same boundary.  But the size of slimbridge
	 may be entirely different in the new location.  Only the alignment field and the object's actual alignment is
	 guaranteed not to change.  Note also that objects with an overflow size field require one word less padding in
	 the slimbridge, because it is the word following the normal header word that is aligned, not the header word
	 and not the overflow header word."
	^((self longAt: objOop) >> self extraAlignedBitShift bitAnd: 1) ~= 0! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 7/12/2016 09:45'!
isFixedSizePointerFormat: format
	<api>
	^format <= self nonIndexablePointerFormat
	  or: [format = self ephemeronFormat]! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/21/2014 17:01'!
isFloatInstance: oop
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/24/2013 14:58'!
isFloatObject: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter isFloatObject: oop! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 6/5/2017 10:28'!
isFloatOrInt: anOop
	"Answer if anOop is either a SmallInteger or a Float."

	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 7/10/2014 14:59'!
isForwarded: objOop
	"Answer if objOop is that if a forwarder.  Take advantage of isForwardedObjectClassIndexPun
	 being a power of two to generate a more efficient test than the straight-forward
		(self classIndexOf: objOop) = self isForwardedObjectClassIndexPun
	 at the cost of this being ambiguous with free chunks.  So either never apply this to free chunks
	 or guard with (self isFreeObject: foo) not.  So far the idiom has been to guard with isFreeObject:"
	<api>
	<inline: true>
	"self assert: (self isFreeObject: objOop) not."
	^(self longAt: objOop) noMask: self classIndexMask - self isForwardedObjectClassIndexPun! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 9/13/2013 15:11'!
isForwardedClassTag: classIndex
	^classIndex = self isForwardedObjectClassIndexPun! !

!SpurMemoryManager methodsFor: 'class table puns' stamp: 'eem 12/12/2015 14:19'!
isForwardedObjectClassIndexPun
	"Answer the class index of a forwarder.  We choose 8 so as not to
	 be confused with any immediate class (whose classIndex matches
	 its instances tag pattern), and because it is a power of two, which
	 allows us to generate a slightly slimmer test for isForwarded:."
	<api>
	<cmacro>
	^8! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 5/4/2021 14:32'!
isForwarderFormat: format
	^format = self forwardedFormat! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/8/2013 13:48'!
isFreeObject: objOop
	^(self classIndexOf: objOop) = self isFreeObjectClassIndexPun! !

!SpurMemoryManager methodsFor: 'class table puns' stamp: 'eem 3/9/2016 14:48'!
isFreeObjectClassIndexPun
	<cmacro>
	^0! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 1/16/2015 11:31'!
isFreeOop: oop
	^(self isNonImmediate: oop) and: [self isFreeObject: oop]! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 10/15/2013 07:55'!
isGrey: objOop
	^((self longAt: objOop) >> self greyBitShift bitAnd: 1) ~= 0! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 10/19/2014 01:45'!
isHiddenObj: objOop
	^(self classIndexOf: objOop) <= self lastClassIndexPun! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 3/9/2016 14:13'!
isImmediate: oop
	<api>
	^oop anyMask: self tagMask! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 3/9/2016 13:55'!
isImmediateCharacter: oop
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 12/1/2014 11:19'!
isImmediateFloat: oop
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 10/15/2013 00:29'!
isImmutable: objOop
	^((self longAt: objOop) >> self immutableBitShift bitAnd: 1) ~= 0! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 7/23/2014 09:06'!
isInClassTable: objOop
	| hash |
	hash := self rawHashBitsOf: objOop.
	^hash ~= 0
	 and: [(self classAtIndex: hash) = objOop]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 10/25/2013 15:53'!
isInEden: objOop
	^self
		oop: objOop
		isGreaterThanOrEqualTo: scavenger eden start
		andLessThan: freeStart! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 10/25/2013 15:56'!
isInFutureSpace: address
	^self
		oop: address
		isGreaterThanOrEqualTo: scavenger futureSpace start
		andLessThan: scavenger futureSurvivorStart! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 11/28/2019 13:40'!
isInHeapBounds: address 
	"Answer if the given address is within the entire range ST object memory.
	 For quick checking during leak checking only!!"
	^(self oop: address isGreaterThanOrEqualTo: newSpaceStart)
	  and: [self oop: address isLessThan: endOfMemory]! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 12/24/2016 13:55'!
isInMemory: address 
	"Answer if the given address is in ST object memory."
	(self isInNewSpace: address) ifTrue:
		[^(self isInEden: address)
			or: [(self isInPastSpace: address)
			or: [self scavengeInProgress and: [self isInFutureSpace: address]]]].
	^segmentManager isInSegments: address! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 11/20/2013 13:31'!
isInNewSpace: objOop
	^(self oop: objOop isLessThan: newSpaceLimit)
	  and: [self oop: objOop isGreaterThanOrEqualTo: newSpaceStart]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'rmacnak 3/26/2015 22:04'!
isInOldSpace: address
	<api>
	^self
		oop: address
		isGreaterThanOrEqualTo: oldSpaceStart
		andLessThan: endOfMemory! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 11/22/2013 14:21'!
isInPastSpace: address
	^self
		oop: address
		isGreaterThanOrEqualTo: scavenger pastSpace start
		andLessThan: pastSpaceStart! !

!SpurMemoryManager methodsFor: 'immediates' stamp: 'eem 12/24/2013 16:59'!
isInRangeCharacterCode: characterCode
	^characterCode between: 0 and: (1 << 30) - 1! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/10/2013 20:46'!
isIndexable: objOop
	| fmt |
	fmt := self formatOf: objOop.
	^self isIndexableFormat: fmt! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 9/20/2013 14:52'!
isIndexableFormat: format
	^format >= self arrayFormat
	  and: [format <= self weakArrayFormat
			or: [format >= self sixtyFourBitIndexableFormat]]! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'nice 4/19/2016 23:48'!
isInstanceOfClassLargeNegativeInteger: oop
	"Answer if the oop is a large negative integer instance."
	^(self isNonImmediate: oop) and: [(self classIndexOf: oop) = ClassLargeNegativeIntegerCompactIndex]! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'nice 4/19/2016 23:49'!
isInstanceOfClassLargePositiveInteger: oop
	"Answer if the oop is a large positive integer instance."
	^(self isNonImmediate: oop) and: [(self classIndexOf: oop) = ClassLargePositiveIntegerCompactIndex]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 11/17/2014 14:13'!
isIntegerObject: oop
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 9/9/2013 13:59'!
isIntegerValue: intValue
	"Answer if the given value can be represented as a Smalltalk integer value."
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'nice 4/20/2016 01:58'!
isKindOfInteger: oop
	<doNotGenerate>
	^coInterpreter isKindOfInteger: oop! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
isLargeFreeObject: objOop
	^(self bytesInBody: objOop) >= (self numFreeLists * self allocationUnit)! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'nice 4/20/2016 00:36'!
isLargeIntegerInstance: oop
	"Answer if the oop is a large positive or negative integer instance."
	^(self isNonImmediate: oop) and: [((self classIndexOf: oop) - ClassLargeNegativeIntegerCompactIndex) asUnsignedInteger <= 1]! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'nice 4/20/2016 01:59'!
isLargeNegativeIntegerObject: oop
	<doNotGenerate>
	^coInterpreter isLargeNegativeIntegerObject: oop! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'nice 4/20/2016 01:59'!
isLargePositiveIntegerObject: oop
	<doNotGenerate>
	^coInterpreter isLargePositiveIntegerObject: oop! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'cb 10/5/2018 12:57'!
isLilliputianSize: chunkBytes
	"Answer if chunkBytes (which includes an object header) is too small to hold both
	 a next free chunk and a previous free chunk pointer. This is always false in 32-bits, 
	but in 64 bits small chunk of size 2 are lilliputian."
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'nice 5/1/2020 21:49'!
isLong64s: oop
	"Answer if the argument contains only indexable 64-bit double words (no oops). See comment in formatOf:"

	<api>
	^(self isNonImmediate: oop)
	  and: [self isLong64sNonImm: oop]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 2/23/2016 21:39'!
isLong64sNonImm: objOop
	"Answer if the argument contains only indexable 64-bit double words (no oops). See comment in formatOf:"

	^(self formatOf: objOop) = self sixtyFourBitIndexableFormat! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 11/18/2013 13:51'!
isMarked: objOop
	<api>
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 5/25/2016 11:52'!
isMaybeFiredEphemeron: objOop
	^self isMaybeFiredEphemeronFormat: (self formatOf: objOop)! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 5/25/2016 11:52'!
isMaybeFiredEphemeronFormat: format
	"Answer if an object's format could be that of an ephemeron.  When ephemerons are born
	 their format is ephemeronFormat, but when they fire their format is changed to 3 (inst vars
	 plus indexable fields) or 1 (objects with inst vars), we haven't decided which yet."
	^format <= self lastPointerFormat and: [format odd]! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 4/29/2014 13:49'!
isMobileObject: objOop
	"Answer if objOop should be moved during compaction.  Non-objects
	 (free chunks & bridges), forwarders and pinned objects are excluded."
	| classIndex |
	<inline: true>
	classIndex := self classIndexOf: objOop.
	self assert: ((self long64At: objOop) ~= 0
				  and: [classIndex < (numClassTablePages * self classTablePageSize)]).
	^classIndex > self isForwardedObjectClassIndexPun
	  and: [(self isPinned: objOop) not]! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 5/1/2014 16:07'!
isMobileObjectHeader: objHeader
	"Answer if an object with header objHeader should be moved during compaction.
	 Non-objects (free chunks & bridges), forwarders and pinned objects are excluded."
	<inline: true>
	^(objHeader >> self pinnedBitShift bitAnd: 1) = 0
	  and: [(self classIndexOfHeader: objHeader) > self isForwardedObjectClassIndexPun]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 3/9/2016 14:13'!
isNonImmediate: oop
	<api>
	^oop noMask: self tagMask! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 10/31/2014 11:37'!
isNonIntegerImmediate: oop
	<inline: true>
	^self isImmediate: oop! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 10/31/2014 11:39'!
isNonIntegerNonImmediate: oop
	<inline: true>
	^self isNonImmediate: oop! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 11/17/2014 14:13'!
isNonIntegerObject: oop
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 10/14/2013 19:46'!
isNormalObject: objOop
	^(self classIndexOf: objOop) > self lastClassIndexPun! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 5/25/2016 11:37'!
isObjEphemeron: objOop
	^self isEphemeronFormat: (self formatOf: objOop)! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 1/20/2016 14:25'!
isObjImmutable: anOop
	<inline: true>
	^self cppIf: IMMUTABILITY
		ifTrue: [self isImmutable: anOop]
		ifFalse: [false]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 11/20/2013 13:57'!
isOldObject: objOop
	<api>
	"Answer if obj is old. Require that obj is non-immediate."
	self assert: (self isNonImmediate: objOop).
	^self oop: objOop isGreaterThanOrEqualTo: oldSpaceStart! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/4/2013 16:22'!
isOopCompiledMethod: oop 
    "Answer whether the oop is an object of compiled method format"
	<api>
    ^(self isNonImmediate: oop)
	 and: [(self formatOf: oop) >= self firstCompiledMethodFormat]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'cb 5/28/2018 18:34'!
isOopForwarded: oop
	<api>
	^(self isNonImmediate: oop) and: [self isForwarded: oop]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 10/15/2013 00:31'!
isOopImmutable: oop
	<api>
	^(self isImmediate: oop)
	  or: [self isImmutable: oop]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 10/15/2013 00:31'!
isOopMutable: oop
	<api>
	^(self isNonImmediate: oop)
	  and: [(self isImmutable: oop) not]! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 10/25/2013 15:17'!
isPinned: objOop
	<api>
	^((self longAt: objOop) >> self pinnedBitShift bitAnd: 1) ~= 0! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/9/2013 14:32'!
isPointers: oop
	"Answer if the argument has only fields that can hold oops. See comment in formatOf:"

	^(self isNonImmediate: oop) and: [self isPointersNonImm: oop]! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 9/9/2013 16:05'!
isPointersFormat: format
	^format <= self lastPointerFormat! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/9/2013 16:05'!
isPointersNonImm: objOop
	"Answer if the argument has only fields that can hold oops. See comment in formatOf:"
	^(self formatOf: objOop) <= self lastPointerFormat! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 12/18/2018 17:44'!
isPositiveMachineIntegerObject: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter isPositiveMachineIntegerObject: oop! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 3/21/2014 04:07'!
isPureBitsFormat: format
	"the inverse of isAnyPointerFormat:"
	^format >= self sixtyFourBitIndexableFormat
	  and: [format < self firstCompiledMethodFormat]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 10/25/2016 13:25'!
isPureBitsNonImm: objOop
	"Answer if the argument contains only indexable words (no oops). See comment in formatOf:"

	^self isPureBitsFormat: (self formatOf: objOop)! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 11/29/2013 12:38'!
isReallyYoung: oop
	<api>
	"Answer if oop is young."
	^(self isNonImmediate: oop)
	 and: [self isReallyYoungObject: oop]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 11/18/2013 15:43'!
isReallyYoungObject: obj
	<api>
	"Answer if obj is young. This for compatibility with SqueakV3 where
	 the GC makes all objects young during full GC.  Spur doesn't do so."
	^self isYoungObject: obj! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 9/22/2013 15:37'!
isRemembered: objOop
	^((self longAt: objOop) >> self rememberedBitShift bitAnd: 1) ~= 0! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 4/30/2014 15:53'!
isRememberedObjectHeader: objHeader
	"Answer if an object with header objHeader is remembered."
	<inline: true>
	^(objHeader >> self rememberedBitShift bitAnd: 1) ~= 0! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 9/29/2013 11:23'!
isScavengeSurvivor: oop
	<doNotGenerate>
	^scavenger isScavengeSurvivor: oop! !

!SpurMemoryManager methodsFor: 'segments' stamp: 'eem 10/8/2013 12:41'!
isSegmentBridge: objOop
	"Maybe this should be in SpurSegmentManager only"
	^(self classIndexOf: objOop) = self segmentBridgePun! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 6/29/2014 10:22'!
isSemaphoreObj: anObj
	^(self classIndexOf: anObj) = (self rawHashBitsOf: (self splObj: ClassSemaphore))! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 6/29/2014 10:21'!
isSemaphoreOop: anOop
	^(self isNonImmediate: anOop)
	 and: [self isSemaphoreObj: anOop]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'nice 5/1/2020 21:49'!
isShorts: oop
	"Answer if the argument contains only indexable 16-bit half words (no oops). See comment in formatOf:"

	<api>
	^(self isNonImmediate: oop)
	  and: [self isShortsNonImm: oop]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 2/23/2016 21:35'!
isShortsNonImm: objOop
	"Answer if the argument contains only indexable 16-bit half words (no oops). See comment in formatOf:"

	^(self formatOf: objOop) between: self firstShortFormat and: self firstByteFormat - 1! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 7/9/2014 11:34'!
isUnambiguouslyForwarder: objOop
	"This version is private to SpurMemoryManager (for asserts, etc).  It does not
	 take advantage of the power-of0two optimization in isForwarded:."
	<api>
	^(self classIndexOf: objOop) = self isForwardedObjectClassIndexPun! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 2/17/2017 10:31'!
isUnmarked: objOop
	"For debugging using printOopsSuchThat:"
	<api>
	^(self isMarked: objOop) not! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 1/4/2019 17:19'!
isValidClassIndex: classIndex
	| classTablePage |
	classIndex <= 0 ifTrue:
		[^false].
	(classIndex <= self tagMask or: [classIndex >= self arrayClassIndexPun]) ifFalse:
		[^false].
	classIndex >= (1 << self classIndexFieldWidth) ifTrue:
		[^false].
	classTablePage := self fetchPointer: classIndex >> self classTableMajorIndexShift
							ofObject: hiddenRootsObj.
	classTablePage = nilObj ifTrue:
		[^false].
	(self addressCouldBeObj: classTablePage) ifFalse:
		[^false].
	^coInterpreter addressCouldBeClassObj:
		(self
			fetchPointer: (classIndex bitAnd: self classTableMinorIndexMask)
			ofObject: classTablePage)! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 4/9/2015 10:40'!
isValidClassTag: classIndex
	<api>
	| classOrNil |
	self assert: (classIndex between: 0 and: 1 << self classIndexFieldWidth - 1).
	classOrNil := self classOrNilAtIndex: classIndex.
	^classOrNil ~= nilObj
	 and: [(self rawHashBitsOf: classOrNil) = classIndex]! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 10/31/2013 20:41'!
isValidObjStack: objStack
	"Answer if the obj stack at objStackRootIndex is valid."
	((self addressCouldBeObj: objStack)
	 and: [(self numSlotsOfAny: objStack) = ObjStackPageSlots]) ifFalse:
		[objStackInvalidBecause := 'first page not obj or wrong size'.
		 invalidObjStackPage := objStack.
		 ^false].
	^self isValidObjStackPage: objStack
		myIndex: (self fetchPointer: ObjStackMyx ofObject: objStack)
		firstPage: true! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 10/18/2013 10:16'!
isValidObjStackAt: objStackRootIndex
	"Answer if the obj stack at objStackRootIndex is valid."
	| stackOrNil |
	stackOrNil := self fetchPointer: objStackRootIndex ofObject: hiddenRootsObj.
	^stackOrNil = nilObj
	  or: [self isValidObjStackPage: stackOrNil myIndex: objStackRootIndex firstPage: true]! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 11/18/2014 14:44'!
isValidObjStackPage: objStackPage
	"for the Spur 32- to 64-bit bootstrap"
	<doNotGenerate>
	^self
		isValidObjStackPage: objStackPage
		myIndex: ((self numSlotsOfAny: objStackPage) = ObjStackPageSlots ifTrue:
					[self fetchPointer: ObjStackMyx ofObject: objStackPage])! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 11/13/2013 13:09'!
isValidObjStackPage: objStackPage myIndex: myx
	"Just check the page itself."
	<inline: false>
	(self classIndexOf: objStackPage) = self wordSizeClassIndexPun ifFalse:
		[objStackInvalidBecause := 'wrong class index'.
		 invalidObjStackPage := objStackPage.
		 ^false].
	(self formatOf: objStackPage) = self wordIndexableFormat ifFalse:
		[objStackInvalidBecause := 'wrong format'.
		 invalidObjStackPage := objStackPage.
		 ^false].
	(self numSlotsOfAny: objStackPage) = ObjStackPageSlots ifFalse:
		[objStackInvalidBecause := 'wrong num slots'.
		 invalidObjStackPage := objStackPage.
		 ^false].
	myx = (self fetchPointer: ObjStackMyx ofObject: objStackPage) ifFalse:
		[objStackInvalidBecause := 'wrong myx'.
		 invalidObjStackPage := objStackPage.
		 ^false].
	(marking and: [(self isMarked: objStackPage) not]) ifTrue:
		[objStackInvalidBecause := 'marking but page is unmarked'.
		 invalidObjStackPage := objStackPage.
		 ^false].
	^true! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 12/12/2013 13:07'!
isValidObjStackPage: objStackPage myIndex: myx firstPage: isFirstPage
	"Answer if the obj stack at stackRootIndex is valid."
	| page freeOrNextPage index |
	<inline: false>
	(self isValidObjStackPage: objStackPage myIndex: myx) ifFalse:
		[^false].
	freeOrNextPage := self fetchPointer: ObjStackFreex ofObject: objStackPage.
	[freeOrNextPage ~= 0] whileTrue:
		[isFirstPage ifFalse:
			[objStackInvalidBecause := 'free page on other than first page'.
			 invalidObjStackPage := objStackPage.
			 ^false].
		 freeOrNextPage = (self fetchPointer: ObjStackNextx ofObject: objStackPage) ifTrue:
			[objStackInvalidBecause := 'free page = next page'.
			 invalidObjStackPage := freeOrNextPage.
			^false].
		 (self isValidObjStackPage: freeOrNextPage myIndex: myx) ifFalse:
			[objStackInvalidBecause := self stretch: objStackInvalidBecause cat: ', on next page'.
			^false].
		 page := self fetchPointer: ObjStackFreex ofObject: freeOrNextPage.
		 (page = freeOrNextPage
		  or: [page = objStackPage]) ifTrue:
			[objStackInvalidBecause := 'circularity in free page list'.
			 invalidObjStackPage := page.
			^false].
		 freeOrNextPage := page].
	isFirstPage ifTrue:
		[(myx between: self classTableRootSlots and: self classTableRootSlots + self hiddenRootSlots - 1) ifFalse:
			[objStackInvalidBecause := 'myx out of range'.
			 invalidObjStackPage := objStackPage.
			 ^false].
		 (self fetchPointer: myx ofObject: hiddenRootsObj) = objStackPage ifFalse:
			[objStackInvalidBecause := 'firstPage is not root'.
			 invalidObjStackPage := objStackPage.
			 ^false]].
	index := self fetchPointer: ObjStackTopx ofObject: objStackPage.
	(index between: 0 and: ObjStackLimit) ifFalse:
		[objStackInvalidBecause := 'bad topx'.
		 invalidObjStackPage := objStackPage.
		 ^false].
	freeOrNextPage := self fetchPointer: ObjStackNextx ofObject: objStackPage.
	freeOrNextPage = 0 ifTrue:
		[^true].
	freeOrNextPage = objStackPage ifTrue:
		[objStackInvalidBecause := 'circularity in objStack page list'.
		 invalidObjStackPage := objStackPage.
		 ^false].
	^self isValidObjStackPage: freeOrNextPage myIndex: myx firstPage: false! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 10/14/2013 19:30'!
isWeak: oop
	"Answer if the argument has only weak fields that can hold oops. See comment in formatOf:"
	^(self isNonImmediate: oop) and: [self isWeakNonImm: oop]! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 3/21/2014 04:00'!
isWeakFormat: format
	<inline: true>
	^format = self weakArrayFormat! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 3/21/2014 04:02'!
isWeakNonImm: objOop
	<inline: true>
	^self isWeakFormat: (self formatOf: objOop)! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/10/2013 15:33'!
isWords: oop
	"Answer if the argument contains only indexable words (no oops). See comment in formatOf:"

	^(self isNonImmediate: oop)
	  and: [self isWordsNonImm: oop]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/10/2013 15:35'!
isWordsNonImm: objOop
	"Answer if the argument contains only indexable words (no oops). See comment in formatOf:"

	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 12/1/2014 13:18'!
isWordsOrBytes: oop
	"Answer if the contains only indexable words or bytes (no oops). See comment in formatOf:"
	"Note: Excludes CompiledMethods."
	^(self isNonImmediate: oop)
	  and: [self isWordsOrBytesNonImm: oop]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 12/1/2014 13:19'!
isWordsOrBytesNonImm: objOop
	"Answer if the contains only indexable words or bytes (no oops). See comment in formatOf:"
	"Note: Excludes CompiledMethods."
	^self isPureBitsFormat: (self formatOf: objOop)! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/10/2020 13:00'!
isWordsOrShorts: oop
	"Answer if the argument contains only indexable 16-bit half words or 32-bit indexable words (no oops).
	 See comment in formatOf:"

	<api>
	^(self isNonImmediate: oop)
	  and: [self isWordsOrShortsNonImm: oop]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 9/10/2020 13:01'!
isWordsOrShortsNonImm: objOop
	"Answer if the argument contains only indexable 16-bit half words or 32-bit words (no oops).
	 See comment in formatOf:"

	^(self formatOf: objOop) between: self firstLongFormat and: self firstByteFormat - 1! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 11/18/2013 15:19'!
isYoung: oop
	<api>
	"Answer if oop is young."
	^(self isNonImmediate: oop)
	 and: [self oop: oop isLessThan: newSpaceLimit]! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'eem 11/20/2013 13:57'!
isYoungObject: objOop
	<api>
	"Answer if obj is young. Require that obj is non-immediate."
	self assert: (self isNonImmediate: objOop).
	^self oop: objOop isLessThan: newSpaceLimit! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 5/25/2016 11:39'!
keyOfEphemeron: objOop
	"Answer the object the ephemeron guards.  This is its first element."
	self assert: ((self isNonImmediate: objOop) and: [self isObjEphemeron: objOop]).
	^self fetchPointer: 0 ofObject: objOop! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 5/25/2016 11:52'!
keyOfMaybeFiredEphemeron: objOop
	"Answer the object the ephemeron guards.  This is its first element."
	self assert: ((self isNonImmediate: objOop) and: [self isMaybeFiredEphemeron: objOop]).
	^self fetchPointer: 0 ofObject: objOop! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 9/2/2013 17:53'!
knownClassAtIndex: classIndex
	self assert: (classIndex between: 1 and: self classTablePageSize).
	^self fetchPointer: classIndex ofObject: classTableFirstPage! !

!SpurMemoryManager methodsFor: 'compaction - analysis' stamp: 'eem 06/22/2021 15:31'!
largeFreeChunkDistribution
	"This method answers a sorted collection of the free chunks >= 1,000,000 bytes long,
	 sandwiched between nilObj and the end of memory (ignoring the large chunk often found at the end of the heap)."
	<doNotGenerate>
	| freeChunks |
	freeChunks := SortedCollection new.
	self allObjectsInFreeTreeDo:
		[:f|
		((self addressAfter: f) < endOfMemory
		 and: [(self bytesInBody: f) >= 1000000]) ifTrue:
			[freeChunks add: f]].
	^{{nilObj hex. #nil}}, (freeChunks collect: [:f| {f hex. self bytesInBody: f}]), {{endOfMemory hex. #endOfMemory}}! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 6/27/2014 15:29'!
largeObjectBytesForSlots: numSlots
	"Answer the total number of bytes in an object with an overflow header, including header bytes."
	<returnTypeC: #usqInt>
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'class table puns' stamp: 'eem 3/9/2016 14:49'!
lastClassIndexPun
	"Class puns are class indices not used by any class.  There is an entry
	 for the pun that refers to the notional class of objects with this class
	 index.  But because the index doesn't match the class it won't show up
	 in allInstances, hence hiding the object with a pun as its class index.
	 The puns occupy indices 16 through 31."
	<cmacro>
	^31! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 9/11/2013 05:18'!
lastHash
	^lastHash! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 9/11/2013 05:17'!
lastHash: seed
	lastHash := seed! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 9/2/2013 11:12'!
lastPointerFormat
	^5! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 11/10/2014 12:49'!
lastPointerOf: objOop 
	"Answer the byte offset of the last pointer field of the given object.
	 Works with CompiledMethods, as well as ordinary objects."
	<api>
	<inline: true>
	| fmt contextSize header |
	fmt := self formatOf: objOop.
	self assert: fmt ~= self forwardedFormat.
	fmt <= self lastPointerFormat ifTrue:
		[(fmt = self indexablePointersFormat
		  and: [self isContextNonImm: objOop]) ifTrue:
			["contexts end at the stack pointer"
			contextSize := coInterpreter fetchStackPointerOf: objOop.
			^CtxtTempFrameStart - 1 + contextSize * self bytesPerOop + self baseHeaderSize].
		^(self numSlotsOf: objOop) - 1 * self bytesPerOop + self baseHeaderSize  "all pointers"].
	fmt < self firstCompiledMethodFormat ifTrue: [^0]. "no pointers"

	"CompiledMethod: contains both pointers and bytes"
	header := self methodHeaderOf: objOop.
	^self lastPointerOfMethodHeader: header! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 2/10/2015 17:19'!
lastPointerOfArray: objOop 
	"Answer the byte offset of the last pointer field of the given array."
	self assert: (self isArray: objOop).
	^(self numSlotsOf: objOop) - 1 * self bytesPerOop + self baseHeaderSize! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 11/10/2014 12:49'!
lastPointerOfMethodHeader: methodHeader 
	"Answer the byte offset of the last pointer field of a
	 CompiledMethod with the given header."
	<inline: true>
	^(self literalCountOfMethodHeader: methodHeader)
	  + LiteralStart - 1 * self bytesPerOop + self baseHeaderSize! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 11/10/2014 12:49'!
lastPointerOfWhileSwizzling: objOop 
	"Answer the byte offset of the last pointer field of the given object.
	 Works with CompiledMethods, as well as ordinary objects.
	 Does not examine the stack pointer of contexts to be sure to swizzle
	 the nils that fill contexts on snapshot.
	 It is invariant that on image load no object contains a forwarding pointer,
	 and the image contains no forwarders (see class comment)."
	<api>
	<inline: true>
	| fmt header |
	fmt := self formatOf: objOop.
	self assert: fmt ~= self forwardedFormat.
	fmt <= self lastPointerFormat ifTrue:
		[^(self numSlotsOf: objOop) - 1 * self bytesPerOop + self baseHeaderSize  "all pointers"].
	fmt < self firstCompiledMethodFormat ifTrue: [^0]. "no pointers"

	"CompiledMethod: contains both pointers and bytes"
	header := self methodHeaderOf: objOop.
	^self lastPointerOfMethodHeader: header! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 6/21/2021 12:53'!
leakCheckBecome
	<api>
	^checkForLeaks anyMask: GCModeBecome! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 6/21/2021 12:53'!
leakCheckFullGC
	<api>
	^checkForLeaks anyMask: GCModeFull! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 6/21/2021 12:53'!
leakCheckImageSegments
	^checkForLeaks anyMask: GCCheckImageSegment! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 6/21/2021 12:54'!
leakCheckIncremental
	<api>
	^checkForLeaks anyMask: GCModeIncremental! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 6/21/2021 12:54'!
leakCheckNewSpaceGC
	<api>
	^checkForLeaks anyMask: GCModeNewSpace! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 6/21/2021 12:54'!
leakCheckRuntimeCalls
	<doNotGenerate>
	^checkForLeaks anyMask: GCCheckPrimCall! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 6/21/2021 12:52'!
leakCheckShorten
	^checkForLeaks anyMask: GCCheckShorten! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 3/20/2020 11:33'!
lengthOf: objOop
	"Answer the number of indexable units in the given object.
	 For a CompiledMethod, the size of the method header (in bytes) should
	 be subtracted from the result."

	<inline: true>
	^self lengthOf: objOop format: (self formatOf: objOop)! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 9/19/2013 11:58'!
lengthOf: objOop baseHeader: header format: fmt 
	<var: #header type: #usqLong>
	"Compatibility; does not really suit the Spur format.
	 Answer the number of indexable bytes or words in the given object.
	 For a CompiledMethod, the size of the method header (in bytes) should
	 be subtracted from the result of this method."
	^self lengthOf: objOop format: fmt! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 2/6/2015 11:18'!
lengthOf: objOop format: fmt
	"Answer the number of indexable units in the given object.
	 For a CompiledMethod, the size of the method header (in bytes)
	 should be subtracted from the result of this method."
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 9/17/2013 13:57'!
lengthOfMaybeImmediate: oop
	"for the message send breakpoint; selectors can be immediates."
	<inline: false>
	(self isImmediate: oop) ifTrue: [^0].
	^self lengthOf: oop! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'cb 10/5/2018 12:59'!
lilliputianChunkIndex
	"See isLilliputianSize:"
	^(self baseHeaderSize + self allocationUnit) // self allocationUnit
! !

!SpurMemoryManager methodsFor: 'method access' stamp: 'eem 8/8/2014 15:08'!
literalCountOf: methodPointer
	<api>
	^self literalCountOfMethodHeader: (self methodHeaderOf: methodPointer)! !

!SpurMemoryManager methodsFor: 'method access' stamp: 'eem 1/22/2015 15:36'!
literalCountOfMethodHeader: header
	<api>
	<inline: true>
	self assert: (self isIntegerObject: header).
	^coInterpreter literalCountOfAlternateHeader: header! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 12/18/2014 13:26'!
loadFloatOrIntFrom: floatOrInt
	"If floatOrInt is an integer, then convert it to a C double float and return it.
	 If it is a Float, then load its value and return it.
	 Otherwise fail -- ie return with primErrorCode non-zero."

	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 9/6/2021 19:29'!
loadImageSegmentFrom: segmentWordArray outPointers: outPointerArray
	"This primitive is called from Smalltalk as...
		<imageSegment> loadSegmentFrom: aWordArray outPointers: anArray."

	"This primitive will load a binary image segment created by primitiveStoreImageSegment.
	 It expects the outPointer array to be of the proper size, and the wordArray to be well formed.
	 It will return as its value the original array of roots, and the segmentWordArray will become an
	 array of the loaded objects.  If this primitive should fail, the segmentWordArray will, sadly, have
	 been reduced to an unrecognizable and unusable jumble.  But what more could you have done
	 with it anyway?

	 The primitive, if it succeeds, also becomes the segmentWordArray into the array of loaded objects.
	 This allows fixing up of loaded objects directly, without nextObject, which Spur doesn't support."

	<inline: #never>
	| segmentLimit segmentStart segVersion errorCode numLoadedObjects loadedObjectsArray |

	segmentLimit := self numSlotsOf: segmentWordArray.
	(self objectBytesForSlots: segmentLimit) < (self allocationUnit "version info" + self baseHeaderSize "one object header") ifTrue:
		[^PrimErrBadArgument halt].

	"Verify format.  If the format is wrong, word-swap (since ImageSegment data are 32-bit longs).
	 If it is still wrong, undo the damage and fail."
	segVersion := self longAt: segmentWordArray + self baseHeaderSize.
	(coInterpreter readableFormat: (segVersion bitAnd: 16rFFFFFF "low 3 bytes")) ifFalse:
		[self reverseBytesIn32BitWordsIn: segmentWordArray.
		 segVersion := self longAt: segmentWordArray + self baseHeaderSize.
		 (coInterpreter readableFormat: (segVersion bitAnd: 16rFFFFFF "low 3 bytes")) ifFalse:
			[self reverseBytesIn32BitWordsIn: segmentWordArray.
			 ^PrimErrBadArgument halt]].

	segmentStart := segmentWordArray + self baseHeaderSize + self allocationUnit.
	segmentLimit := segmentLimit * self bytesPerOop + segmentWordArray + self baseHeaderSize.

	"Notionally reverse the Byte type objects if the data is from opposite endian machine.
	 Test top byte.  $d on the Mac or $s on the PC.  Rest of word is equal.  If Spur is ever
	 ported to big-endian machines then the segment may have to be byte/word swapped,
	 but so far it only runs on little-endian machines, so for now just fail if endianness is wrong."
	self flag: #endianness.
	(segVersion >> 24 bitAnd: 16rFF) ~= (self imageSegmentVersion >> 24 bitAnd: 16rFF) ifTrue:
		"Reverse the byte-type objects once"
		[true
			ifTrue: [^PrimErrBadArgument halt]
			ifFalse:
				[self byteSwapByteObjectsFrom: (self objectStartingAt: segmentStart)
					to: segmentLimit
					flipFloatsIf: false]].

	"Avoid having to remember by arranging that there are no young outPointers if segment is in old space."
	(self isOldObject: segmentWordArray) ifTrue:
		[errorCode := self ensureNoNewObjectsIn: outPointerArray.
		 errorCode ~= 0 ifTrue:
			[^errorCode]].

	"scan through mapping oops and validating class references. Defer entering any
	 class objects into the class table and/or pinning objects until a second pass."
	errorCode := self mapOopsAndValidateClassRefsFrom: segmentStart to: segmentLimit outPointers: outPointerArray.
	errorCode > 0 ifTrue:
		[^errorCode].
	numLoadedObjects := errorCode negated.
	loadedObjectsArray := self allocateSlots: numLoadedObjects format: self arrayFormat classIndex: ClassArrayCompactIndex.
	loadedObjectsArray ifNil:
		[^PrimErrNoMemory halt].

	"Scan for classes contained in the segment, entering them into the class table.
	 Classes are at the front, after the root array and have the remembered bit set."
	errorCode := self enterClassesIntoClassTableFrom: segmentStart to: segmentLimit.
	errorCode ~= 0 ifTrue:
		[^errorCode].

	"Make a final pass, assigning class indices and/or pinning pinned objects and collecting the loaded objects in loadedObjectsArray"
	self assignClassIndicesAndPinFrom: segmentStart to: segmentLimit outPointers: outPointerArray filling: loadedObjectsArray.

	"Evaporate the container, leaving the newly loaded objects in place."
	(self hasOverflowHeader: segmentWordArray)
		ifTrue: "N.B. setting the overflow slots to 1 creates a slimbridge in eden, so we also need to delete the overflow slot count in the segment itself"
			[(self oop: segmentWordArray isLessThan: oldSpaceStart)
				ifTrue:
					[self rawOverflowSlotsOf: segmentWordArray put: 0.
					 self rawNumSlotsOf: segmentWordArray put: 0]
				ifFalse:
					[self rawOverflowSlotsOf: segmentWordArray put: self allocationUnit / self bytesPerOop]]
		ifFalse: [self rawNumSlotsOf: segmentWordArray put: self allocationUnit / self bytesPerOop].

	"Finally forward the segmentWordArray to the loadedObjectsArray"
	self forward: segmentWordArray to: loadedObjectsArray.
	
	self runLeakCheckerFor: GCCheckImageSegment.

	^self objectStartingAt: segmentStart! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 12/15/2014 14:12'!
logBytesPerOop
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
longPrintInstancesOf: aClassOop
	"Scan the heap printing the oops of any and all objects that are instances of aClassOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| classIndex |
	classIndex := self rawHashBitsOf: aClassOop.
	classIndex ~= self isFreeObjectClassIndexPun ifTrue:
		[self longPrintInstancesWithClassIndex: classIndex]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
longPrintInstancesWithClassIndex: classIndex
	"Scan the heap printing any and all objects whose classIndex equals the argument."
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	<inline: false>
	self allHeapEntitiesDo:
		[:obj|
		 (self classIndexOf: obj) = classIndex ifTrue:
			[coInterpreter longPrintOop: obj; cr]]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:04'!
longPrintReferencesTo: anOop
	"Scan the heap long printing the oops of any and all objects that refer to anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| prntObj |
	prntObj := false.
	self allObjectsDo:
		[:obj| | i |
		((self isPointersNonImm: obj) or: [self isCompiledMethod: obj]) ifTrue:
			[(self isCompiledMethod: obj)
				ifTrue:
					[i := (self literalCountOf: obj) + LiteralStart]
				ifFalse:
					[(self isContextNonImm: obj)
						ifTrue: [i := CtxtTempFrameStart + (coInterpreter fetchStackPointerOf: obj)]
						ifFalse: [i := self numSlotsOf: obj]].
			[(i := i - 1) >= 0] whileTrue:
				[anOop = (self fetchPointer: i ofObject: obj) ifTrue:
					[coInterpreter printHex: obj; print: ' @ '; printNum: i; cr.
					 prntObj := true.
					 i := 0]].
			prntObj ifTrue:
				[prntObj := false.
				 coInterpreter longPrintOop: obj]]]! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 1/6/2020 11:37'!
lookup32BitWordConstant: word
	"If word appears to be a class index answer the class name, otherwise answer nil.
	 For code disassembly"
	<doNotGenerate>
	| class classSize classNameIndex thisClassIndex |
	(hiddenRootsObj notNil
	 and: [word between: 1 and: self classIndexMask]) ifFalse:
		[^nil].
	class := self noCheckClassAtIndex: word.
	(class isNil or: [class = nilObj]) ifTrue:
		[^nil].	"address is either a class or a metaclass, or an instance of a class or invalid.  determine which."
	classNameIndex := coInterpreter classNameIndex.
	thisClassIndex := coInterpreter thisClassIndex.
	((classSize := self numSlotsOf: class) <= (classNameIndex max: thisClassIndex)
	 or: [classSize > 255]) ifTrue:
		[^nil].
	"Address could be a class or a metaclass"
	(self lookupAddress: (self fetchPointer: classNameIndex ofObject: class)) ifNotNil:
		[:maybeClassName| ^maybeClassName allButFirst].
	(self lookupAddress: (self fetchPointer: thisClassIndex ofObject: class)) ifNotNil:
		[:maybeClassName| ^maybeClassName allButFirst, ' class'].! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 5/9/2016 14:43'!
lookupAddress: address
	"If address appears to be that of a Symbol or a few well-known objects (such as classes) answer it, otherwise answer nil.
	 For code disassembly"
	<doNotGenerate>
	| fmt size string class classSize maybeThisClass classNameIndex thisClassIndex |
	((self addressCouldBeObj: address)
	 and: [(thisClassIndex := self classIndexOf: address) > 0]) ifFalse:
		[^address = scavengeThreshold ifTrue:
			['scavengeThreshold']].
	address - self baseHeaderSize = hiddenRootsObj ifTrue:
		[^'(hiddenRootsObj+baseHeaderSize)'].
	fmt := self formatOf: address.
	size := self lengthOf: address baseHeader: (self baseHeader: address) format: fmt.
	size = 0 ifTrue:
		[^address caseOf: { [nilObj] -> ['nil']. [trueObj] -> ['true']. [falseObj] -> ['false'] } otherwise: []].
	((fmt between: self firstByteFormat and: self firstCompiledMethodFormat - 1) "indexable byte fields"
	and: [(size between: 1 and: 64)
	and: [(Scanner isLiteralSymbol: (string := (0 to: size - 1) collect: [:i| Character value: (self fetchByte: i ofObject: address)]))
		or: [NewspeakVM and: [string noneSatisfy: [:c| c isSeparator or: [c asInteger > 126]]]]]]) ifTrue:
		[^'#', (ByteString withAll: string)].
	class := self noCheckClassAtIndex: thisClassIndex.
	(class isNil or: [class = nilObj]) ifTrue:
		[^nil].
	"address is either a class or a metaclass, or an instance of a class or invalid.  determine which."
	classNameIndex := coInterpreter classNameIndex.
	thisClassIndex := coInterpreter thisClassIndex.
	((classSize := self numSlotsOf: class) <= (classNameIndex max: thisClassIndex)
	 or: [classSize > 255]) ifTrue:
		[^nil].
	"Address could be a class or a metaclass"
	(fmt = 1 and: [size > classNameIndex]) ifTrue:
		["Is address a class? If so class's thisClass is address."
		 (self lookupAddress: (self fetchPointer: classNameIndex ofObject: address)) ifNotNil:
			[:maybeClassName|
			(self fetchPointer: thisClassIndex ofObject: class) = address ifTrue:
				[^maybeClassName allButFirst]].
		"Is address a Metaclass?  If so class's name is Metaclass and address's thisClass holds the class name"
		((self isBytes: (self fetchPointer: classNameIndex ofObject: class))
		 and: [(self lookupAddress: (self fetchPointer: classNameIndex ofObject: class)) = '#Metaclass'
		 and: [size >= thisClassIndex]]) ifTrue:
			[maybeThisClass := self fetchPointer: thisClassIndex ofObject: address.
			(self lookupAddress: (self fetchPointer: classNameIndex ofObject: maybeThisClass)) ifNotNil:
				[:maybeThisClassName| ^maybeThisClassName allButFirst, ' class']]].
	^(self lookupAddress: (self fetchPointer: classNameIndex ofObject: class)) ifNotNil:
		[:maybeClassName|
		(size = 2
		 and: [fmt = self nonIndexablePointerFormat
		 and: ['#ClassBinding' = maybeClassName]]) ifTrue:
			[^'a ClassBinding '
				, ((self lookupAddress: (self fetchPointer: KeyIndex ofObject: address)) ifNil: [''])
				, ((self lookupAddress: (self fetchPointer: ValueIndex ofObject: address)) ifNil: [''] ifNotNil: [:val| ' -> ', val])].
		'a(n) ', maybeClassName allButFirst]! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 4/26/2016 16:15'!
lowSpaceThreshold: threshold
	lowSpaceThreshold := threshold.
	"N.B. The threshold > 0 guard eliminates a warning when
		self lowSpaceThreshold: 0
	 is inlined into setSignalLowSpaceFlagAndSaveProcess"
	(threshold > 0
	 and: [totalFreeOldSpace < threshold]) ifTrue:
		[self growOldSpaceByAtLeast: threshold - totalFreeOldSpace].
	self assert: totalFreeOldSpace >= lowSpaceThreshold! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 1/16/2020 19:51'!
mapExtraRoots
	(self shouldRemapObj: specialObjectsOop) ifTrue:
		[validatedIntegerClassFlags := 0.
		 specialObjectsOop := self remapObj: specialObjectsOop].
	self assert: remapBufferCount = 0.
	"1 to: remapBufferCount do:
		[:i | | oop |
		oop := remapBufferCount at: i.
		((self isImmediate: oop) or: [self isFreeObject: oop]) ifFalse:
			[(self shouldRemapObj: oop) ifTrue:
				[remapBuffer at: i put: (self remapObj: oop)]]]."
	1 to: extraRootCount do:
		[:i | | oop |
		oop := (extraRoots at: i) at: 0.
		((self isImmediate: oop) or: [self isFreeObject: oop]) ifFalse:
			[(self shouldRemapObj: oop) ifTrue:
				[(extraRoots at: i) at: 0 put: (self remapObj: oop)]]]! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 5/26/2016 11:43'!
mapMournQueue
	<inline: #never>
	self objStack: mournQueue do:
		[:i :page| | mourner |
		mourner := self fetchPointer: i ofObject: page.
		(self isNonImmediate: mourner) ifTrue: "someone could try and become weaklings into immediates..."
			[(self isForwarded: mourner) ifTrue:
				[mourner := self followForwarded: mourner].
			 (scavenger isScavengeSurvivor: mourner) ifFalse:
				[mourner := scavenger copyAndForwardMourner: mourner].
			 "we could check for change but writes are cheasp with write buffers..."
			 self storePointerUnchecked: i ofObject: page withValue: mourner]]! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 2/21/2019 09:23'!
mapOopsAndValidateClassRefsFrom: segmentStart to: segmentLimit outPointers: outPointerArray
	"This is part of loadImageSegmentFrom:outPointers:.
	 Scan through mapping oops and validating class references.  Defer
	 entering any class objects into the class table and/or pinning objects
	 until the second pass in assignClassIndicesAndPinFrom:to:outPointers:."
	<var: 'segmentLimit' type: #usqInt>
	| numOutPointers numSegObjs objOop |
	<var: #oop type: #usqInt>
	numOutPointers := self numSlotsOf: outPointerArray.
	numSegObjs := 0.
	objOop := self objectStartingAt: segmentStart.
	[self oop: objOop isLessThan: segmentLimit] whileTrue:
		[| classIndex hash oop mappedOop |
		 numSegObjs := numSegObjs + 1.
		 "No object in the segment should be marked.  If is is something is wrong."
		 (self isMarked: objOop) ifTrue:
			[^PrimErrInappropriate halt].
		 classIndex := self classIndexOf: objOop.
		 "validate the class ref, but don't update it until any internal classes have been added to the class table."
		 (classIndex anyMask: TopHashBit)
			ifTrue:
				[classIndex := classIndex - TopHashBit.
				 classIndex >= numOutPointers ifTrue:
					[^PrimErrBadIndex halt].
				 mappedOop := self fetchPointer: classIndex ofObject: outPointerArray.
				 hash := self rawHashBitsOf: mappedOop.
				 (hash = 0 "class has yet to be instantiated"
				  or: [hash > self lastClassIndexPun and: [(self classOrNilAtIndex: hash) = mappedOop]]) ifFalse:
					[^PrimErrInappropriate halt]]
			ifFalse: "The class is contained within the segment."
				[(oop := classIndex - self firstClassIndexPun * self allocationUnit + segmentStart) >= segmentLimit ifTrue:
					[^PrimErrBadIndex halt].
				 (self rawHashBitsOf: oop) ~= 0 ifTrue:
					[^PrimErrInappropriate halt]].
		 0 to: (self numPointerSlotsOf: objOop) - 1 do:
			[:i|
			 oop := self fetchPointer: i ofObject: objOop.
			 (self isNonImmediate: oop) ifTrue:
				[(oop anyMask: TopOopBit)
					ifTrue:
						[(oop := oop - TopOopBit / self bytesPerOop) >= numOutPointers ifTrue:
							[^PrimErrBadIndex halt].
						 mappedOop := self fetchPointer: oop ofObject: outPointerArray]
					ifFalse:
						[(oop bitAnd: self allocationUnit - 1) ~= 0 ifTrue:
							[^PrimErrInappropriate halt].
						 (mappedOop := oop + segmentStart) >= segmentLimit ifTrue:
							[^PrimErrBadIndex halt]].
				 self storePointerUnchecked: i ofObject: objOop withValue: mappedOop]].
		 objOop := self objectAfter: objOop limit: segmentLimit].
	^numSegObjs negated! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 12/11/2017 15:39'!
mapOopsFrom: segStart to: segAddr outPointers: outPointerArray outHashes: savedOutHashes
	"This is part of storeImageSegmentInto:outPointers:roots:.
	 Now scan, adding out pointers to the outPointersArray; all objects in arrayOfObjects
	 have had their first fields set to point to their copies in segmentWordArray.  Answer
	 the outIndex if the scan succeded.  Fail if outPointers is too small and answer -1.

	 As established by copyObj:toAddr:startAt:stopAt:savedFirstFields:index:,
	 the marked bit is set for all objects in the segment
	 the remembered bit is set for all classes in the segment.

	 Class indices should be set as follows (see assignClassIndicesAndPinFrom:to:outPointers:filling:)
	 - class indices for classes in the segment "
	| objOop objIndex outIndex |
	outIndex := objIndex := 0. "objIndex is for debugging; it mirrors indices in the sender's arrayOfObjects."
	self fillObj: outPointerArray numSlots: (self numSlotsOf: outPointerArray) with: nilObj.
	objOop := self objectStartingAt: segStart.
	[self oop: objOop isLessThan: segAddr] whileTrue:
		[| heapOop oop hash segIndex |
		 heapOop := self fetchClassOfNonImm: objOop.
		 "Set the classIndex of the instance.  This is a segment offset (segAddr - segStart / allocationUnit) for instances of
		  classes within the segment, and an outPointer index (index in outPointers + TopHashBit) for classes outside the segment."
		 (self isCopiedIntoSegment: heapOop)
			ifTrue: "oop is a class in the segment; storeImageSegmentInto:outPointers:roots: established offset is within range."
				[oop := self fetchPointer: 0 ofObject: heapOop.
				 self assert: (self oop: oop isGreaterThanOrEqualTo: segStart andLessThan: segAddr).
				 segIndex := oop - segStart / self allocationUnit + self firstClassIndexPun.
				 (segIndex anyMask: TopHashBit) ifTrue: "Too many classes in the segment"
					[^-1 halt]]
			ifFalse: "oop is an outPointer; locate or allocate its oop"
				[hash := self rawHashBitsOf: heapOop.
				 (self is: hash outPointerClassHashFor: heapOop in: outPointerArray limit: outIndex)
					ifTrue: [segIndex := hash]
					ifFalse: "oop is a new outPointer; allocate its oop"
						[outIndex := self newOutPointer: heapOop at: outIndex in: outPointerArray hashes: savedOutHashes.
						 outIndex = 0 ifTrue: "no room in outPointers; fail"
							[^-1 halt].
						 segIndex := self rawHashBitsOf: heapOop].
				 self assert: (segIndex anyMask: TopHashBit)].
		 self setClassIndexOf: objOop to: segIndex.
		 0 to: (self numPointerSlotsOf: objOop) - 1 do:
			[:i|
			 heapOop := self fetchPointer: i ofObject: objOop.
			 (self isNonImmediate: heapOop) ifTrue:
				[(self isCopiedIntoSegment: heapOop)
					ifTrue: "oop is an object in the segment."
						[oop := self fetchPointer: 0 ofObject: heapOop.
						 self assert: (self oop: oop isGreaterThanOrEqualTo: segStart andLessThan: segAddr).
						 oop := oop - segStart]
					ifFalse: "oop is an outPointer; locate or allocate its oop"
						[hash := self rawHashBitsOf: heapOop.
						(self is: hash outPointerClassHashFor: heapOop in: outPointerArray limit: outIndex)
							ifTrue: [oop := hash - TopHashBit * self bytesPerOop + TopOopBit]
							ifFalse: "oop is a new outPointer; allocate its oop"
								[outIndex := self newOutPointer: heapOop at: outIndex in: outPointerArray hashes: savedOutHashes.
								 outIndex = 0 ifTrue: "no room in outPointers; fail"
									[^-1 halt].
								 self assert: ((self rawHashBitsOf: heapOop) anyMask: TopHashBit).
								 oop := (self rawHashBitsOf: heapOop) - TopHashBit * self bytesPerOop + TopOopBit]].
				 self storePointerUnchecked: i ofObject: objOop withValue: oop]].
		 objOop := self objectAfter: objOop limit: segAddr.
		 objIndex := objIndex + 1].
	^outIndex! !

!SpurMemoryManager methodsFor: 'primitive support' stamp: 'eem 3/27/2014 15:29'!
mark: objOop
	<inline: true>
	self setIsMarkedOf: objOop to: true! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 5/23/2016 15:35'!
markAccessibleObjectsAndFireEphemerons
	self assert: marking.
	self assert: self validClassTableRootPages.
	self assert: segmentManager allBridgesMarked.
	self cCode: [] "for debugging markAndTrace: set (MarkStackRecord := OrderedCollection new)"
		inSmalltalk: [MarkStackRecord ifNotNil: [MarkStackRecord resetTo: 1]].

	"This must come first to enable stack page reclamation.  It clears
	  the trace flags on stack pages and so must precede any marking.
	  Otherwise it will clear the trace flags of reached pages."
	coInterpreter initStackPageGC.
	self markAndTraceHiddenRoots.
	self markAndTraceExtraRoots.
	self assert: self validClassTableRootPages.
	coInterpreter markAndTraceInterpreterOops: true.
	self assert: self validObjStacks.
	self markWeaklingsAndMarkAndFireEphemerons.
	self assert: self validObjStacks! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 5/25/2016 18:35'!
markAllUnscannedEphemerons
	"After firing the unscanned ephemerons we must scan-mark them.
	 The wrinkle is that doing so may add more ephemerons to the set.
	 So we remove the first element, by overwriting it with the last element,
	 and decrementing the top, and then markAndTrace its contents."
	self assert: (self noUnscannedEphemerons) not.
	self assert: self allUnscannedEphemeronsAreActive.
	[unscannedEphemerons top > unscannedEphemerons start] whileTrue:
		[| ephemeron key lastptr |
		 ephemeron := self longAt: unscannedEphemerons start.
		 lastptr := unscannedEphemerons top - self bytesPerOop.
		 lastptr > unscannedEphemerons start ifTrue:
			[self longAt: unscannedEphemerons start put: (self longAt: lastptr)].
		 unscannedEphemerons top: lastptr.
		 key := self followedKeyOfMaybeFiredEphemeron: ephemeron.
		 self setIsMarkedOf: ephemeron to: false. "to get it to be fully scanned in markAndTrace:"
		 self
			markAndTrace: key;
			markAndTrace: ephemeron]! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 10/2/2014 15:26'!
markAndShouldScan: objOop
	"Helper for markAndTrace:.
	 Mark the argument, and answer if its fields should be scanned now.
	 Immediate objects don't need to be marked.
	 Already marked objects have already been processed.
	 Pure bits objects don't need scanning, although their class does.
	 Weak objects should be pushed on the weakling stack.
	 Anything else need scanning."
	| format |
	<inline: true>
	(self isImmediate: objOop) ifTrue:
		[^false].
	"if markAndTrace: is to follow and eliminate forwarding pointers
	 in its scan it cannot be handed an r-value which is forwarded."
	self assert: (self isForwarded: objOop) not.
	(self isMarked: objOop) ifTrue:
		[^false].
	self setIsMarkedOf: objOop to: true.
	format := self formatOf: objOop.
	(self isPureBitsFormat: format) ifTrue: "avoid pushing non-pointer objects on the markStack."
		["Avoid tracing classes of non-objects on the heap, e.g. IRC caches, Sista counters."
		 (self classIndexOf: objOop) > self lastClassIndexPun ifTrue:
			[self markAndTraceClassOf: objOop].
		 ^false].
	format = self weakArrayFormat ifTrue: "push weaklings on the weakling stack to scan later"
		[self push: objOop onObjStack: weaklingStack.
		 ^false].
	(format = self ephemeronFormat
	 and: [self activeAndDeferredScan: objOop]) ifTrue:
		[^false].
	^true! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 8/11/2016 09:49'!
markAndTrace: objOop
	"Mark the argument, and all objects reachable from it, and any remaining objects
	 on the mark stack. Follow forwarding pointers in the scan."
	<api>
	<inline: #never>
	"if markAndTrace: is to follow and eliminate forwarding pointers
	 in its scan it cannot be handed an r-value which is forwarded.
	 The assert for this is in markAndShouldScan:"
	(self markAndShouldScan: objOop) ifFalse:
		[^self].

	"Now scan the object, and any remaining objects on the mark stack."
	self markLoopFrom: objOop! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 3/9/2016 17:43'!
markAndTraceClassOf: objOop
	"Ensure the class of the argument is marked, pushing it on the markStack if not already marked.
	 And for one-way become, which can create duplicate entries in the class table, make sure
	 objOop's classIndex refers to the classObj's actual classIndex.
	 Note that this is recursive, but the metaclass chain should terminate quickly."
	<inline: false>
	| classIndex classObj realClassIndex |
	classIndex := self classIndexOf: objOop.
	classObj := self classOrNilAtIndex: classIndex.
	self assert: (coInterpreter objCouldBeClassObj: classObj).
	realClassIndex := self rawHashBitsOf: classObj.
	(classIndex ~= realClassIndex
	 and: [classIndex > self lastClassIndexPun]) ifTrue:
		[self setClassIndexOf: objOop to: realClassIndex].
	(self isMarked: classObj) ifFalse:
		[self setIsMarkedOf: classObj to: true.
		 self markAndTraceClassOf: classObj.
		 self push: classObj onObjStack: markStack]! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 12/11/2013 09:36'!
markAndTraceExtraRoots
	| oop |
	self assert: remapBufferCount = 0.
	"1 to: remapBufferCount do:
		[:i|
		 oop := remapBuffer at: i.
		 ((self isImmediate: oop) or: [self isFreeObject: oop]) ifFalse:
			[self markAndTrace: oop]]."
	1 to: extraRootCount do:
		[:i|
		oop := (extraRoots at: i) at: 0.
		((self isImmediate: oop) or: [self isFreeObject: oop]) ifFalse:
			[self markAndTrace: oop]]! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 8/31/2015 11:00'!
markAndTraceHiddenRoots
	"The hidden roots hold both the class table pages and the obj stacks,
	 and hence need special treatment.  The obj stacks must be marked
	 specially; their pages must be marked, but only the contents of the
	 mournQueue should be marked.

	 If a class table page is weak we can mark and trace the hiddenRoots,
	 which will not trace through class table pages because they are weak.
	 But if class table pages are strong, we must mark the pages and *not*
	 trace them so that only classes reachable from the true roots will be
	 marked, and unreachable classes will be left unmarked."

	self markAndTraceObjStack: markStack andContents: false.
	self markAndTraceObjStack: weaklingStack andContents: false.
	self markAndTraceObjStack: mournQueue andContents: true.

	self setIsMarkedOf: self rememberedSetObj to: true.
	self setIsMarkedOf: self freeListsObj to: true.

	(self isWeakNonImm: classTableFirstPage) ifTrue:
		[^self markAndTrace: hiddenRootsObj].

	self setIsMarkedOf: hiddenRootsObj to: true.
	self markAndTrace: classTableFirstPage.
	1 to: numClassTablePages - 1 do:
		[:i| self setIsMarkedOf: (self fetchPointer: i ofObject: hiddenRootsObj)
				to: true]! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 5/25/2016 14:34'!
markAndTraceObjStack: stackOrNil andContents: markAndTraceContents
	"An obj stack is a stack of objects stored in a hidden root slot, such
	 as the markStack or the ephemeronQueue.  It is a linked list of
	 segments, with the hot end at the head of the list.  It is a word object.
	 The stack pointer is in ObjStackTopx and 0 means empty."
	<inline: false>
	| index field |
	stackOrNil = nilObj ifTrue:
		[^self].
	self setIsMarkedOf: stackOrNil to: true.
	self assert: (self numSlotsOfAny: stackOrNil) = ObjStackPageSlots.
	field := self fetchPointer: ObjStackNextx ofObject: stackOrNil.
	field ~= 0 ifTrue:
		[self markAndTraceObjStack: field andContents: markAndTraceContents].
	field := stackOrNil.
	[field := self fetchPointer: ObjStackFreex ofObject: field.
	 field ~= 0] whileTrue:
		[self setIsMarkedOf: field to: true].
	markAndTraceContents ifFalse:
		[^self].
	"There are four fixed slots in an obj stack, and a Topx of 0 indicates empty, so
	  if there were 6 slots in an oop stack, full would be 2, and the last 0-rel index is 5."
	index := (self fetchPointer: ObjStackTopx ofObject: stackOrNil) + ObjStackNextx.
	[index >= ObjStackFixedSlots] whileTrue:
		[field := self followObjField: index ofObject: stackOrNil.
		 (self isImmediate: field) ifFalse:
			[self markAndTrace: field].
		 index := index - 1]! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 2/18/2016 13:00'!
markAndTraceWeaklingsFrom: startIndex
	"Mark weaklings on the weaklingStack, ignoring startIndex
	 number of elements on the bottom of the stack.  Answer
	 the size of the stack *before* the enumeration began."
	^self objStack: weaklingStack from: startIndex do:
		[:weakling|
		 self deny: (self isForwarded: weakling).
		 self markAndTraceClassOf: weakling.
		"N.B. generateToByDoLimitExpression:negative:on: guards against (unsigned)0 - 1 going +ve"
		 0 to: (self numStrongSlotsOfWeakling: weakling) - 1 do:
			[:i| | field |
			field := self followOopField: i ofObject: weakling.
			((self isImmediate: field) or: [self isMarked: field]) ifFalse:
				[self markAndTrace: field]]]! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 7/11/2017 19:40'!
markAsCopiedIntoSegment: anObjectInTheHeap
	"This is part of storeImageSegmentInto:outPointers:roots:."
	<inline: true>
	self setIsMarkedOf: anObjectInTheHeap to: true! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 5/23/2016 15:40'!
markInactiveEphemerons
	"Go through the unscanned ephemerons, marking the inactive ones, and
	 removing them from the unscanned ephemerons. Answer if any inactive
	 ones were found. We cannot fire the ephemerons until all are found to
	 be active since scan-marking an inactive ephemeron later in the set may
	 render a previously-observed active ephemeron as inactive."
	| foundInactive ptr |
	foundInactive := false.
	ptr := unscannedEphemerons start.
	[ptr < unscannedEphemerons top] whileTrue:
		[| ephemeron key |
		 key := self followedKeyOfEphemeron: (ephemeron := self longAt: ptr).
		 ((self isImmediate: key) or: [self isMarked: key])
			ifTrue:
				[foundInactive := true.
				 "Now remove the inactive ephemeron from the set, and scan-mark it.
				  Scan-marking it may add more ephemerons to the set."
				 unscannedEphemerons top: unscannedEphemerons top - self bytesPerOop.
				 unscannedEphemerons top > ptr ifTrue:
					[self longAt: ptr put: (self longAt: unscannedEphemerons top)].
				 self markAndTrace: ephemeron]
			ifFalse:
				[ptr := ptr + self bytesPerOop]].
	^foundInactive! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 10/20/2014 10:16'!
markLoopFrom: objOop
	"Scan objOop and all objects on the mark stack, until the mark stack is empty.
	 N.B. When the incremental GC is written this will probably be refactored as
	 markLoopFrom: objOop while: aBlock"
	<inline: true>
	| objToScan field index numStrongSlots scanLargeObject |

	"Now scan the object, and any remaining objects on the mark stack."
	objToScan := objOop.
	"To avoid overflowing the mark stack when we encounter large objects, we
	 push the obj, then its numStrongSlots, and then index the object from the stack."
	[(self isImmediate: objToScan)
		ifTrue: [scanLargeObject := true]
		ifFalse:
			[numStrongSlots := self numStrongSlotsOfInephemeral: objToScan.
			 scanLargeObject := numStrongSlots > self traceImmediatelySlotLimit].
	 scanLargeObject
		ifTrue: "scanning a large object. scan until hitting an unmarked object, then switch to it, if any."
			[(self isImmediate: objToScan)
				ifTrue:
					[index := self integerValueOf: objToScan.
					 objToScan := self topOfObjStack: markStack]
				ifFalse:
					[index := numStrongSlots.
					 self markAndTraceClassOf: objToScan].
			 [index > 0] whileTrue:
				[index := index - 1.
				 field := self fetchPointer: index ofObject: objToScan.
				 (self isNonImmediate: field) ifTrue:
					[(self isForwarded: field) ifTrue: "fixFollowedField: is /not/ inlined"
						[field := self fixFollowedField: index ofObject: objToScan withInitialValue: field].
					 (self markAndShouldScan: field) ifTrue:
						[index > 0 ifTrue:
							[(self topOfObjStack: markStack) ~= objToScan ifTrue: 
								[self push: objToScan onObjStack: markStack].
							 self push: (self integerObjectOf: index) onObjStack: markStack].
						 objToScan := field.
						 index := -1]]].
			 index >= 0 ifTrue: "if loop terminated without finding an unmarked referent, switch to top of stack."
				[objToScan := self popObjStack: markStack.
				 objToScan = objOop ifTrue:
					[objToScan := self popObjStack: markStack]]]
		ifFalse: "scanning a small object. scan, marking, pushing unmarked referents, then switch to the top of the stack."
			[index := numStrongSlots.
			 self markAndTraceClassOf: objToScan.
			 [index > 0] whileTrue:
				[index := index - 1.
				 field := self fetchPointer: index ofObject: objToScan.
				 (self isNonImmediate: field) ifTrue:
					[(self isForwarded: field) ifTrue: "fixFollowedField: is /not/ inlined"
						[field := self fixFollowedField: index ofObject: objToScan withInitialValue: field].
					 (self markAndShouldScan: field) ifTrue:
						[self push: field onObjStack: markStack.
						 ((self rawNumSlotsOf: field) > self traceImmediatelySlotLimit
						  and: [(numStrongSlots := self numStrongSlotsOfInephemeral: field) > self traceImmediatelySlotLimit]) ifTrue:
							[self push: (self integerObjectOf: numStrongSlots) onObjStack: markStack]]]].
			 objToScan := self popObjStack: markStack].
	 objToScan notNil] whileTrue! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 4/15/2021 11:46'!
markObjects: objectsShouldBeUnmarkedAndUnmarkedClassesShouldBeExpunged
	<inline: #never> "for profiling"
	"Mark all accessible objects.  objectsShouldBeUnmarkedAndUnmarkedClassesShouldBeExpunged
	 is true if all objects are unmarked and/or if unmarked classes shoud be removed from the class table."
	"If the incremental collector is running mark bits may be set; stop it and clear them if necessary."
	self cCode: '' inSmalltalk: [coInterpreter transcript nextPutAll: 'marking...'; flush].
	self runLeakCheckerFor: GCModeFull.

	self shutDownGlobalIncrementalGC: objectsShouldBeUnmarkedAndUnmarkedClassesShouldBeExpunged.
	self initializeUnscannedEphemerons.
	self initializeMarkStack.
	self initializeWeaklingStack.
	marking := true.
	self markAccessibleObjectsAndFireEphemerons.
	self expungeDuplicateAndUnmarkedClasses: objectsShouldBeUnmarkedAndUnmarkedClassesShouldBeExpunged.
	self nilUnmarkedWeaklingSlots.
	marking := false! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 2/11/2015 10:18'!
markObjectsIn: arrayOfRoots
	"This is part of storeImageSegmentInto:outPointers:roots:."
	self setIsMarkedOf: arrayOfRoots to: true.
	0 to: (self numSlotsOf: arrayOfRoots) - 1 do:
		[:i| | oop |
		oop := self followField: i ofObject: arrayOfRoots.
		(self isNonImmediate: oop) ifTrue:
			[self setIsMarkedOf: oop to: true]]! !

!SpurMemoryManager methodsFor: 'spur bootstrap' stamp: 'eem 11/18/2014 15:11'!
markStack
	^markStack! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 11/27/2018 15:35'!
markWeaklingsAndMarkAndFireEphemerons
	"After the initial scan-mark is complete ephemerons can be processed.
	 Weaklings have accumulated on the weaklingStack, but more may be
	 uncovered during ephemeron processing.  So trace the strong slots
	 of the weaklings, and as ephemerons are processed ensure any newly
	 reached weaklings are also traced."
	| numTracedWeaklings |
	<inline: false>
	numTracedWeaklings := 0.
	[coInterpreter markAndTraceUntracedReachableStackPages.
	 coInterpreter markAndTraceMachineCodeOfMarkedMethods.
	 "Make sure all reached weaklings have their strong slots traced before firing ephemerons..."
	 [numTracedWeaklings := self markAndTraceWeaklingsFrom: numTracedWeaklings.
	  (self sizeOfObjStack: weaklingStack) > numTracedWeaklings] whileTrue.
	 self noUnscannedEphemerons ifTrue:
		[coInterpreter
			markAndTraceUntracedReachableStackPages;
	 		markAndTraceMachineCodeOfMarkedMethods;
			freeUntracedStackPages;
			freeUnmarkedMachineCode.
		 ^self].
	 self markInactiveEphemerons ifFalse:
		[self fireAllUnscannedEphemerons].
	 self markAllUnscannedEphemerons]
		repeat! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 12/28/2016 13:06'!
markedBitFullShift
	<cmacro>
	"bit 1 of 2-bit field above identityHash (little endian)"
	^55! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 12/28/2016 13:04'!
markedBitHalfShift
	<cmacro>
	"bit 1 of 2-bit field above identityHash (little endian)"
	^23! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 1/8/2015 14:40'!
maxCInteger
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 11:45'!
maxIdentityHash
	^self identityHashHalfWordMask! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 6/29/2015 15:37'!
maxOldSpaceSize
	<cmacro: '() maxOldSpaceSize'>
	^maxOldSpaceSize! !

!SpurMemoryManager methodsFor: 'instantiation' stamp: 'eem 6/28/2014 16:35'!
maxSlotsForAlloc
	"Answer the maximum number of slots we are willing to attempt to allocate in an object.
	 Must fit in 32-bits; c.f. bytesInObject:"
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'instantiation' stamp: 'eem 10/3/2013 11:22'!
maxSlotsForNewSpaceAlloc
	"Almost entirely arbitrary, but we dont want 1Mb bitmaps allocated in eden.
	 But this choice means no check for numSlots > maxSlotsForNewSpaceAlloc
	 for non-variable allocations."
	^self fixedFieldsOfClassFormatMask! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/20/2014 13:00'!
maxSmallInteger
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'cog jit support' stamp: 'eem 8/31/2021 14:16'!
maybeContextMClassTagBits
	<inline: #always>
	^1 bitShift: self numTagBits! !

!SpurMemoryManager methodsFor: 'object testing' stamp: 'cb 10/23/2017 11:16'!
maybeMethodClassOf: methodObj seemsToBeInstantiating: format
	"Answers if the code is installed in a class instantiating objects with the format. Used in primitive 
	 generation to make a quick path based on where the method is installed. This method cannot
	 be used as a guarantee as there can be false positive, it's just a heuristic.
	 Tries to interpret the last literal of the method as a behavior (more than 3 fields, 3rd field a Smi).
	 If it can be interpreted as a behavior, answers if instSpec matches the format, else answers false."
	<api>
	| maybeClassObj maybeFormat instSpec|
	maybeClassObj := coInterpreter methodClassOf: methodObj.
	(self isPointersNonImm: maybeClassObj) ifFalse: [^false].
	(self numSlotsOfAny: maybeClassObj) > InstanceSpecificationIndex ifFalse: [^false].
	maybeFormat := self fetchPointer: InstanceSpecificationIndex ofObject: maybeClassObj.
	(self isIntegerObject: maybeFormat) ifFalse: [^false].
	instSpec := self instSpecOfClassFormat: (self integerValueOf: maybeFormat).
	^ instSpec = format! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 10/14/2013 12:50'!
maybeSplObj: index
	<api>
	"Answer one of the objects in the SpecialObjectsArray, if in range, otherwise answer nil."
	^index < (self numSlotsOf: specialObjectsOop) ifTrue:
		[self fetchPointer: index ofObject: specialObjectsOop]! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 8/1/2021 12:16'!
memcpy: destAddress _: sourceAddress _: bytes
	"For SpurGenerationScavenger>>copyToFutureSpace:bytes:. N.B. If ranges overlap, must use memmove."
	<doNotGenerate>
	| nToCopy offset |
	(destAddress isInteger and: [sourceAddress isInteger]) ifFalse: "CogMethodProxies..."
		[(sourceAddress isCollection "String, etc..."
		 or: [sourceAddress isCArray
		 or: [destAddress isCArray]]) ifTrue: 
			[^super memcpy: destAddress _: sourceAddress _: bytes].
		 ^self memcpy: destAddress asInteger _: sourceAddress asInteger _: bytes].
	self deny: ((destAddress <= sourceAddress and: [destAddress + bytes > sourceAddress])
				or: [sourceAddress <= destAddress and: [sourceAddress + bytes > destAddress]]).
	self assert: (destAddress \\ 8) + (sourceAddress \\ 8) = 0. "for now..."
	nToCopy := bytes.
	offset := 0.
	memory bytesPerElement = 8 ifTrue:
		[0 to: nToCopy - 8 by: 8 do:
			[:i| self long64At: destAddress + i put: (self long64At: sourceAddress + i)].
		 offset := nToCopy - (nToCopy \\ 8).
		 nToCopy := nToCopy \\ 8].
	nToCopy >= 4 ifTrue:
		[0 to: nToCopy - 4 by: 4 do:
			[:i| self long32At: destAddress + i + offset put: (self long32At: sourceAddress + i + offset)].
		 offset := offset + nToCopy - (nToCopy \\ 4).
		 nToCopy := nToCopy \\ 4].
	0 to: nToCopy - 1 do:
		[:i| self byteAt: destAddress + i + offset put: (self byteAt: sourceAddress + i + offset)].
	^destAddress! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 10/25/2018 11:34'!
memmove: destAddress _: sourceAddress _: bytes
	"Emulate the c library memmove function"
	<doNotGenerate>
	| dst src  |
	dst := destAddress asInteger.
	src := sourceAddress asInteger.
	self assert: bytes \\ 8 + (dst \\ 8) + (src \\ 8) = 0.
	memory bytesPerElement = 8
		ifTrue:
			[destAddress > sourceAddress
				ifTrue:
					[bytes - 8 to: 0 by: -8 do:
						[:i| self long64At: dst + i put: (self long64At: src + i)]]
				ifFalse:
					[0 to: bytes - 8 by: 8 do:
						[:i| self long64At: dst + i put: (self long64At: src + i)]]]
		ifFalse:
			[destAddress > sourceAddress
				ifTrue:
					[bytes - 4 to: 0 by: -4 do:
						[:i| self long32At: dst + i put: (self long32At: src + i)]]
				ifFalse:
					[0 to: bytes - 4 by: 4 do:
						[:i| self long32At: dst + i put: (self long32At: src + i)]]]! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 6/30/2021 18:48'!
memory
	"memory is a simulation thing, but much more convenient to declare here than in every simulator subclass."
	^memory! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 9/30/2013 20:05'!
memory: aValue
	^memory := aValue! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 11/21/2013 11:30'!
memoryBaseForImageRead
	"Answer the address to read the image into."
	^oldSpaceStart! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 11/18/2016 16:23'!
memoryClass
	"Answer the class to use for the memory inst var in simulation.
	 Answer nil if a suitable class isn't available."
	<doNotGenerate>
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 10/9/2013 14:36'!
memoryLimit
	^endOfMemory! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/19/2013 11:09'!
methodArgumentCount
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter methodArgumentCount! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 8/4/2016 11:35'!
methodCacheHashOf: selector with: classTag
	"Sicne class tags are indices or immediate tags, it is necessary to shift the class
	 tag w.r.t. the selector to include the full class index/immediate tags in the hash."
	<inline: true>
	^selector bitXor: classTag << 2 "num cache probes - 1"! !

!SpurMemoryManager methodsFor: 'compaction' stamp: 'eem 12/22/2016 14:48'!
methodHeaderFromSavedFirstField: oop
	self assert: (self isIntegerObject: oop).
	^oop! !

!SpurMemoryManager methodsFor: 'memory access' stamp: 'eem 8/8/2014 09:44'!
methodHeaderOf: methodObj
	"Answer the method header of a CompiledMethod object."
	self assert: (self isCompiledMethod: methodObj).
	^self fetchPointer: HeaderIndex ofObject: methodObj! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 03/16/2018 14:10'!
methodReturnBool: boolean
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter methodReturnBool: boolean! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 3/16/2018 15:22'!
methodReturnFloat: aFloat
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter methodReturnFloat: aFloat! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 03/16/2018 14:10'!
methodReturnInteger: integer
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter methodReturnInteger: integer! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'ul 4/17/2018 16:02'!
methodReturnReceiver
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter methodReturnReceiver! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 03/16/2018 14:10'!
methodReturnString: aCString
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter methodReturnString: aCString! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 1/24/2014 12:42'!
methodReturnValue: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter methodReturnValue: oop! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'cb 5/22/2017 16:52'!
minSlotsForShortening
	"Answer the minimum number of additional slots to allocate in an object to always be able to shorten it.
	 This is enough slots to allocate a minimum-sized object."
	<api>
	^self allocationUnit * 2 / self bytesPerOop! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/19/2014 15:10'!
minSmallInteger
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'spur bootstrap' stamp: 'eem 8/31/2015 10:56'!
mournQueue
	^mournQueue! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 5/10/2017 09:22'!
moveClassesForwardsIn: arrayOfObjects
	"This is part of storeImageSegmentInto:outPointers:roots:.
	 Both to expand the max size of segment and to reduce the length of the
	 load-time pass that adds classes to the class table, move classes to the
	 front of arrayOfObjects, leaving the root array as the first element.
	 Answer the number of classes in the segment."
	| nClasses there |
	nClasses := there := 0. "if > 0, this is the index of the first non-class past the first element."
	1 to: (self numSlotsOf: arrayOfObjects) - 1 do:
		[:here| | objOop hash tempObjOop |
		 objOop := self fetchPointer: here ofObject: arrayOfObjects.
		 hash := self rawHashBitsOf: objOop.
		 (hash > self lastClassIndexPun and: [(self classOrNilAtIndex: hash) = objOop])
			ifTrue:
				[nClasses := nClasses + 1.
				 there > 0 ifTrue: "if there is zero we're in a run of classes at the start so don't move"
					[tempObjOop := self fetchPointer: there ofObject: arrayOfObjects.
					 self storePointerUnchecked: there ofObject: arrayOfObjects withValue: objOop.
					 self storePointerUnchecked: here ofObject: arrayOfObjects withValue: tempObjOop.
					 there := there + 1]]
			ifFalse:
				[there = 0 ifTrue:
					[there := here]]].
	^nClasses! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 1/8/2016 22:37'!
naiveSwapHeaders: obj1 and: obj2 copyHashFlag: copyHashFlag
	"swap headers, but swapping headers swaps remembered bits and hashes;
	 remembered bits must be unswapped and hashes may be unswapped if
	 copyHash is false."
	"This variant tickles a compiler bug in gcc and clang. See cleverSwapHeaders:and:copyHashFlag:"
	<inline: true>
	| headerTemp remembered1 remembered2 hashTemp |
	remembered1 := self isRemembered: obj1.
	remembered2 := self isRemembered: obj2.
	headerTemp := self long64At: obj1.
	self long64At: obj1 put: (self long64At: obj2).
	self long64At: obj2 put: headerTemp.
	self setIsRememberedOf: obj1 to: remembered1.
	self setIsRememberedOf: obj2 to: remembered2.
	"swapping headers swaps hash; if !!copyHashFlag undo hash copy"
	copyHashFlag ifFalse:
		[hashTemp := self rawHashBitsOf: obj1.
		 self setHashBitsOf: obj1 to: (self rawHashBitsOf: obj2).
		 self setHashBitsOf: obj2 to: hashTemp]! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 9/10/2013 15:06'!
needGCFlag
	^needGCFlag! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'cb 1/19/2017 18:34'!
newHashBitsOf: objOop
	| hash |
	hash := self newObjectHash bitAnd: self identityHashHalfWordMask.
	self setHashBitsOf: objOop to: hash.
	^hash! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'nice 1/13/2020 20:57'!
newObjectHash
	"Use a slight variation on D.H. Lehmer's linear congruential generator from 1951.
	 See e.g. http://en.wikipedia.org/wiki/Linear_congruential_generator."
	| hash |
	<var: #hash type: #usqInt>
	[lastHash := self cCode: [lastHash * 16807] "7 raisedTo: 5"
					inSmalltalk: [lastHash := lastHash * 16807 bitAnd: 16rFFFFFFFF].
	 hash := lastHash + (lastHash >> 4). "adding the top bits gives much better spread.  See below:"
	 (hash bitAnd: self identityHashHalfWordMask) = 0] whileTrue.
	^hash

	"the standard algorithm doesn't vary well in the least significant bits:"
	"| r s n |
	r := 1.
	n := 256 * 256 * 256.
	s := Set new: n * 2.
	n timesRepeat:
		[s add: (r bitAnd: n - 1).
		r := r * 16807 bitAnd: 16rFFFFFFFF].
	{ s size. s size / n asFloat. s includes: 0. r hex }
=>	 #(2097152 0.125 false '16r38000001')"

	"this can be improved by adding in shifted upper bits"
	"| r s n |
	r := 1.
	n := 256 * 256 * 256.
	s := Set new: n * 2.
	n timesRepeat:
		[s add: (r + (r bitShift: -8) bitAnd: n - 1).
		r := r * 16807 bitAnd: 16rFFFFFFFF].
	{ s size. s size / n asFloat. s includes: 0. r hex }
=>	#(10702109 0.637895405292511 false '16r38000001')"


"but which shift is best?  Looks like -4:"
"(-2 to: -15 by: -1) do:
	[:shift|
	| r s n |
	r := 1.
	n := 256 * 256 * 256.
	s := Set new: n * 2.
	n timesRepeat:
		[s add: (r + (r bitShift: shift) bitAnd: n - 1).
		r := r * 16807 bitAnd: 16rFFFFFFFF].
	Transcript cr; print: { shift. s size. s size / n asFloat. s includes: 0. r hex }; flush]

#(-2 8388608 0.5 true '16r38000001')
#(-3 8388608 0.5 false '16r38000001')
#(-4 12582503 0.749975621700287 true '16r38000001')
#(-5 11468379 0.6835686564445495 false '16r38000001')
#(-6 11013442 0.656452298164368 true '16r38000001')
#(-7 10804094 0.643974184989929 true '16r38000001')
#(-8 10702109 0.637895405292511 false '16r38000001')
#(-9 10703730 0.637992024421692 false '16r38000001')
#(-10 7865201 0.468802511692047 false '16r38000001')
#(-11 8444092 0.503307104110718 false '16r38000001')
#(-12 10703317 0.6379674077034 true '16r38000001')
#(-13 10701116 0.637836217880249 true '16r38000001')
#(-14 10689443 0.637140452861786 true '16r38000001')
#(-15 7853923 0.4681302905082702 true '16r38000001')"! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 5/10/2017 19:10'!
newOutPointer: oop at: outIndex in: outPointerArray hashes: savedOutHashes
	"This is part of storeImageSegmentInto:outPointers:roots:.
	 oop is a new outPointer; allocate its oop, and answer the new outIndex.
	 If outPointerArray is full, answer 0."
	<inline: true>
	outIndex >= (self numSlotsOf: outPointerArray) ifTrue:
					["no room in outPointers; fail"
					 ^0].
	self storePointer: outIndex ofObject: outPointerArray withValue: oop.
	self storeLong32: outIndex ofObject: savedOutHashes withValue: (self rawHashBitsOf: oop).
	self setHashBitsOf: oop to: outIndex + TopHashBit.
	^outIndex + 1! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 11/21/2013 08:54'!
newSpaceBytes
	^edenBytes! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 9/30/2013 20:07'!
newSpaceIsEmpty
	^freeStart = scavenger eden start
	  and: [pastSpaceStart = scavenger pastSpace start]! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 11/22/2013 18:07'!
newSpaceLimit
	<cmacro: '() GIV(newSpaceLimit)'>
	^newSpaceLimit! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 3/21/2015 10:46'!
newSpaceRefCountMask
	"The three bit field { isGrey, isPinned, isRemembered } is for bits
	 that are never set in young objects.  This allows the remembered
	 table to be pruned when full by using these bits as a reference
	 count of newSpace objects from the remembered table. Objects
	 with a high count should be tenured to prune the remembered table."
	^1 << self greyBitShift
	  bitOr: (1 << self pinnedBitShift
	  bitOr: 1 << self rememberedBitShift)! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/12/2013 09:28'!
newSpaceSize
	^(freeStart - scavenger eden start)
	 + (pastSpaceStart - scavenger pastSpace start)! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 11/22/2013 15:16'!
newSpaceStart
	^newSpaceStart! !

!SpurMemoryManager methodsFor: 'primitive support' stamp: 'eem 9/2/2013 15:21'!
nilFieldsOf: obj 
	0 to: (self numSlotsOf: obj) - 1 do:
		[:i|
		self storePointerUnchecked: i ofObject: obj withValue: nilObj]! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 12/16/2015 09:21'!
nilObject
	<api>
	^nilObj! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 8/30/2013 15:41'!
nilObject: anOop
	"For mapInterpreterOops"
	nilObj := anOop! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 5/23/2016 15:37'!
nilUnmarkedWeaklingSlots
	"Nil the unmarked slots in the weaklings on the
	 weakling stack, finalizing those that lost references.
	 Finally, empty the weaklingStack."
	<inline: #never> "for profiling"
	self cCode: '' inSmalltalk: [coInterpreter transcript nextPutAll: 'nilling...'; flush].
	self eassert: [self allOldMarkedWeakObjectsOnWeaklingStack].
	weaklingStack = nilObj ifTrue:
		[^self].
	self objStack: weaklingStack from: 0 do:
		[:weakling| | anyUnmarked |
		anyUnmarked := self nilUnmarkedWeaklingSlotsIn: weakling.
		anyUnmarked ifTrue:
			["fireFinalization: could grow the mournQueue and if so,
			  additional pages must be marked to avoid being GC'ed."
			 self assert: marking.
			 coInterpreter fireFinalization: weakling]].
	self emptyObjStack: weaklingStack! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 7/9/2014 11:33'!
nilUnmarkedWeaklingSlotsIn: aWeakling
	"Nil the unmarked slots in aWeakling and
	 answer if any unmarked slots were found."
	<inline: true>
	| anyUnmarked |
	anyUnmarked := false.
	self assert: (self allStrongSlotsOfWeaklingAreMarked: aWeakling).
	"N.B. generateToByDoLimitExpression:negative:on: guards against (unsigned)0 - 1 going +ve"
	(self numStrongSlotsOfWeakling: aWeakling) to: (self numSlotsOf: aWeakling) - 1 do:
		[:i| | referent |
		referent := self fetchPointer: i ofObject: aWeakling.
		(self isNonImmediate: referent) ifTrue:
			[(self isUnambiguouslyForwarder: referent) ifTrue:
				[referent := self fixFollowedField: i ofObject: aWeakling withInitialValue: referent].
			 ((self isImmediate: referent) or: [self isMarked: referent]) ifFalse:
				[self storePointerUnchecked: i ofObject: aWeakling withValue: nilObj.
				 anyUnmarked := true]]].
	^anyUnmarked! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 7/23/2014 07:40'!
noCheckClassAtIndex: classIndex
	| classTablePage |
	classTablePage := self fetchPointer: classIndex >> self classTableMajorIndexShift
							ofObject: hiddenRootsObj.
	classTablePage = nilObj ifTrue:
		[^nil].
	^self
		fetchPointer: (classIndex bitAnd: self classTableMinorIndexMask)
		ofObject: classTablePage! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 5/11/2017 14:42'!
noCheckPush: objOop onObjStack: objStack
	<inline: false>
	"Push an element on an objStack.  Split from push:onObjStack: for testing."
	| topx |
	self eassert: [self isValidObjStack: objStack].
	self cCode: '' "for debugging markAndTrace: set (MarkStackRecord := OrderedCollection new)"
		inSmalltalk:
			[MarkStackRecord ifNotNil:
				[(self fetchPointer: ObjStackMyx ofObject: objStack) = MarkStackRootIndex ifTrue:
					[MarkStackRecord addLast: {#push. objOop}]]].
	topx := self fetchPointer: ObjStackTopx ofObject: objStack.
	topx >= ObjStackLimit
		ifTrue:
			[self noCheckPush: objOop
				onObjStack: (self ensureRoomOnObjStackAt: (self fetchPointer: ObjStackMyx ofObject: objStack))]
		ifFalse:
			[self storePointer: ObjStackFixedSlots + topx ofObjStack: objStack withValue: objOop.
			 self storePointer: ObjStackTopx ofObjStack: objStack withValue: topx + 1].
	^objOop! !

!SpurMemoryManager methodsFor: 'forwarding' stamp: 'eem 8/23/2014 11:37'!
noFixupFollowField: fieldIndex ofObject: anObject
	"Make sure the oop at fieldIndex in anObject is not forwarded (follow the
	 forwarder there-in if so).  Answer the (possibly followed) oop at fieldIndex."
	| objOop |
	objOop := self fetchPointer: fieldIndex ofObject: anObject.
	(self isOopForwarded: objOop) ifTrue:
		[objOop := self followForwarded: objOop].
	^objOop! !

!SpurMemoryManager methodsFor: 'forwarding' stamp: 'eem 10/14/2014 12:08'!
noInlineFollowForwarded: objOop
	<inline: false>
	^self followForwarded: objOop! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 3/10/2016 13:14'!
noInlineObjectAfter: objOop limit: limit
	"Object parsing.
	1. all objects have at least a word following the header, for a forwarding pointer.
	2. objects with an overflow size have a preceeing word with a saturated numSlots.  If the word
	   following an object doesn't have a saturated numSlots field it must be a single-header object.
	   If the word following does have a saturated numSlots it must be the overflow size word."
	<inline: false>
	^self objectAfter: objOop limit: limit! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 10/18/2013 14:52'!
noUnscannedEphemerons
	^unscannedEphemerons top = unscannedEphemerons start! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 11/18/2013 15:55'!
nonIndexablePointerFormat
	<api>
	^1! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 7/28/2014 08:11'!
num16BitUnitsOf: objOop 
	"Answer the number of 16-bit units in the given non-immediate object.
	 N..B. Rounds down 8-bit units, so a 5 byte object has 2 16-bit units.
	 Does not adjust the size of contexts by stackPointer."
	^(self numBytesOf: objOop) >> 1! !

!SpurMemoryManager methodsFor: 'object access' stamp: ' nice 4/12/2018 10:38'!
num32BitUnitsOf: objOop 
	"Answer the number of 32-bit units in the given non-immediate object.
	 N..B. Rounds down 8-bit units, so a 7 byte object has 1 32-bit unit.
	 Does not adjust the size of contexts by stackPointer."
	^(self numBytesOf: objOop) >> 2! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 7/28/2014 08:21'!
num64BitUnitsOf: objOop 
	"Answer the number of 16-bit units in the given non-immediate object.
	 N..B. Rounds down 8-bit units, so a 15 byte object has 1 64-bit unit.
	 Does not adjust the size of contexts by stackPointer."
	^(self numBytesOf: objOop) >> 3! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 11/19/2014 08:45'!
numBytesOf: objOop 
	"Answer the number of indexable bytes in the given non-immediate object.
	 Does not adjust the size of contexts by stackPointer."
	<api>
	| fmt numBytes |
	<inline: true>
	fmt := self formatOf: objOop.
	numBytes := self numSlotsOf: objOop.
	numBytes := numBytes << self shiftForWord.
	fmt >= self firstByteFormat ifTrue: "bytes (the common case), including CompiledMethod"
		[^numBytes - (fmt bitAnd: 7)].
	fmt <= self sixtyFourBitIndexableFormat ifTrue:
		[^numBytes].
	fmt >= self firstShortFormat ifTrue:
		[^numBytes - ((fmt bitAnd: 3) << 1)].
	"fmt >= self firstLongFormat"
	^numBytes - ((fmt bitAnd: 1) << 2)! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 6/22/2021 09:20'!
numBytesOfBits: objOop format: format
	"Answer the number of indexable bytes in the given non-immediate pure bits object with format."
	| numBytes |
	<inline: true>
	self assert: ((self isPureBitsFormat: format) and: [(self formatOf: objOop) = format]).
	numBytes := self numSlotsOf: objOop.
	numBytes := numBytes << self shiftForWord.
	format < self firstShortFormat ifTrue:
		[^format <= self sixtyFourBitIndexableFormat
			ifTrue: [numBytes]
			ifFalse: [numBytes - ((format bitAnd: 1) << 2)]].
	^format < self firstByteFormat
		ifTrue: [numBytes - ((format bitAnd: 3) << 1)]
		ifFalse: [numBytes - (format bitAnd: 7)]! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 6/20/2021 16:48'!
numBytesOfBytes: objOop
	"Answer the number of indexable bytes in the given non-immediate byte-indexable object."
	| fmt |
	<inline: true>
	fmt := self formatOf: objOop.
	self assert: fmt >= self firstByteFormat.
	^(self numSlotsOf: objOop) << self shiftForWord - (fmt bitAnd: 7)! !

!SpurMemoryManager methodsFor: 'object format' stamp: 'eem 11/10/2014 12:49'!
numFixedSlotsOf: objOop
	<inline: true>
	^self fixedFieldsOfClassFormat: (self formatOfClass: (self fetchClassOfNonImm: objOop))! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 11/18/2014 09:08'!
numFreeLists
	"Answer the number of free lists.  We use freeListsMask, a bitmap, to avoid
	 reading empty list heads.  This should fit in a machine word to end up in a
	 register during free chunk allocation."
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 6/22/2020 20:37'!
numPointerSlotsOf: objOop
	"Answer the number of pointer fields in the given object.
	 Works with CompiledMethods, as well as ordinary objects."
	<api>
	<inline: true>
	| fmt |
	fmt := self formatOf: objOop.
	^self numPointerSlotsOf: objOop format: fmt! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 6/22/2020 20:36'!
numPointerSlotsOf: objOop format: fmt
	"Answer the number of pointer fields in the given object.
	 Works with CompiledMethods, as well as ordinary objects."
	<inline: #always>
	| contextSize numLiterals header |
	fmt <= self lastPointerFormat ifTrue:
		[(fmt = self indexablePointersFormat
		  and: [self isContextNonImm: objOop]) ifTrue:
			["contexts end at the stack pointer"
			contextSize := coInterpreter fetchStackPointerOf: objOop.
			^CtxtTempFrameStart + contextSize].
		^self numSlotsOf: objOop  "all pointers"].
	fmt = self forwardedFormat ifTrue: [^1].
	fmt < self firstCompiledMethodFormat ifTrue: [^0]. "no pointers"

	"CompiledMethod: contains both pointers and bytes"
	header := self methodHeaderOf: objOop.
	numLiterals := self literalCountOfMethodHeader: header.
	^numLiterals + LiteralStart! !

!SpurMemoryManager methodsFor: 'segments' stamp: 'eem 7/19/2014 11:56'!
numSegments
	<doNotGenerate>
	^segmentManager numSegments! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 8/9/2014 12:38'!
numSlotsForBytes: numBytes
	^numBytes + (self wordSize - 1) // self wordSize! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 10/18/2014 18:40'!
numSlotsForShortening: objOop toIndexableSize: indexableSize
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 12/12/2015 14:19'!
numSlotsFullShift
	<api>
	<cmacro>
	^56! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 12/12/2015 14:19'!
numSlotsHalfShift
	<api>
	<cmacro>
	^24! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 12/12/2015 14:19'!
numSlotsMask
	<api>
	<cmacro>
	"8-bit slot count
		max 64-bit small obj size 254 * 8 =  2032 bytes
		max 32-bit small obj size 254 * 4 =   1016 bytes"
	^255! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 4/28/2015 14:49'!
numSlotsOf: objOop
	<returnTypeC: #usqInt>
	<api>
	| numSlots |
	self flag: #endianness.
	"numSlotsOf: should not be applied to free or forwarded objects."
	self assert: (self classIndexOf: objOop) > self isForwardedObjectClassIndexPun.
	numSlots := self rawNumSlotsOf: objOop.
	^numSlots = self numSlotsMask	"overflow slots; (2^32)-1 slots are plenty"
		ifTrue: [self rawOverflowSlotsOf: objOop]
		ifFalse: [numSlots]! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 4/28/2015 14:49'!
numSlotsOfAny: objOop
	"A private internal version of numSlotsOf: that is happy to be applied to free or forwarded objects."
	<returnTypeC: #usqInt>
	| numSlots |
	numSlots := self rawNumSlotsOf: objOop.
	^numSlots = self numSlotsMask
		ifTrue: [self rawOverflowSlotsOf: objOop] "overflow slots; (2^32)-1 slots are plenty"
		ifFalse: [numSlots]! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/19/2014 13:13'!
numSmallIntegerBits
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 11/18/2014 17:42'!
numSmallIntegerTagBits
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 11/10/2014 12:49'!
numStrongSlotsOf: objOop format: fmt ephemeronInactiveIf: criterion
	"Answer the number of strong pointer fields in the given object.
	 Works with CompiledMethods, as well as ordinary objects."
	<var: 'criterion' declareC: 'int (*criterion)(sqInt key)'>
	<inline: true>
	| numSlots  contextSize numLiterals header |
	fmt <= self lastPointerFormat ifTrue:
		[numSlots := self numSlotsOf: objOop.
		 fmt <= self arrayFormat ifTrue:
			[^numSlots].
		 fmt = self indexablePointersFormat ifTrue:
			[(self isContextNonImm: objOop) ifTrue:
				[coInterpreter setTraceFlagOnContextsFramesPageIfNeeded: objOop.
				 "contexts end at the stack pointer"
				 contextSize := coInterpreter fetchStackPointerOf: objOop.
				 ^CtxtTempFrameStart + contextSize].
			 ^numSlots].
		 fmt = self weakArrayFormat ifTrue:
			[^self fixedFieldsOfClass: (self fetchClassOfNonImm: objOop)].
		 self assert: fmt = self ephemeronFormat.
		 ^(self perform: criterion with: (self keyOfEphemeron: objOop))
			ifTrue: [numSlots]
			ifFalse: [0]].
	fmt = self forwardedFormat ifTrue: [^1].
	fmt < self firstCompiledMethodFormat ifTrue: [^0]. "no pointers"

	"CompiledMethod: contains both pointers and bytes"
	header := self methodHeaderOf: objOop.
	numLiterals := self literalCountOfMethodHeader: header.
	^numLiterals + LiteralStart! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 11/10/2014 12:49'!
numStrongSlotsOfInephemeral: objOop
	"Answer the number of strong pointer fields in the given object,
	 which is .expected not to be an active ephemeron.
	 Works with CompiledMethods, as well as ordinary objects."
	<inline: true>
	| fmt numSlots  contextSize numLiterals header |
	fmt := self formatOf: objOop.
	self assert: (fmt ~= self ephemeronFormat or: [self isMarked: (self keyOfEphemeron: objOop)]).
	fmt <= self lastPointerFormat ifTrue:
		[numSlots := self numSlotsOf: objOop.
		 fmt <= self arrayFormat ifTrue:
			[^numSlots].
		 fmt = self indexablePointersFormat ifTrue:
			[(self isContextNonImm: objOop) ifTrue:
				[coInterpreter setTraceFlagOnContextsFramesPageIfNeeded: objOop.
				 "contexts end at the stack pointer"
				 contextSize := coInterpreter fetchStackPointerOf: objOop.
				 ^CtxtTempFrameStart + contextSize].
			 ^numSlots].
		 fmt = self weakArrayFormat ifTrue:
			[^self fixedFieldsOfClass: (self fetchClassOfNonImm: objOop)]].
	fmt = self forwardedFormat ifTrue: [^1].
	fmt < self firstCompiledMethodFormat ifTrue: [^0]. "no pointers"

	"CompiledMethod: contains both pointers and bytes"
	header := self methodHeaderOf: objOop.
	numLiterals := self literalCountOfMethodHeader: header.
	^numLiterals + LiteralStart! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 11/10/2014 12:49'!
numStrongSlotsOfWeakling: objOop
	"Answer the number of strong pointer fields in the given weakling."
	<api>
	<inline: true>
	self assert: (self formatOf: objOop) = self weakArrayFormat.
	^self fixedFieldsOfClass: (self fetchClassOfNonImm: objOop)! !

!SpurMemoryManager methodsFor: 'scavenger' stamp: 'eem 10/9/2013 13:37'!
numSurvivorSpaces
	"there are two survivor spaces, futureSPace & pastSpace."
	^2! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 12/12/2015 11:33'!
numTagBits
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 7/9/2014 11:33'!
numberOfForwarders
	| n |
	n := 0.
	self allHeapEntitiesDo:
		[:o|
		(self isUnambiguouslyForwarder: o) ifTrue:
			[n := n + 1]].
	^n! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 5/26/2016 10:44'!
objStack: objStack do: aBlock
	"Evaluate aBinaryBlock with all indices and pages of elements in objStack"
	<inline: true>
	| objStackPage |
	objStack = nilObj ifTrue:
		[^self].
	self eassert: [self isValidObjStack: objStack].
	objStackPage := objStack.
	[objStackPage ~= 0] whileTrue:
		[| numOnThisPage |
		 numOnThisPage := self fetchPointer: ObjStackTopx ofObject: objStackPage.
		 numOnThisPage + ObjStackFixedSlots - 1 to: ObjStackFixedSlots by: -1 do:
			[:i| aBlock value: i value: objStackPage].
		 objStackPage := self fetchPointer: ObjStackNextx ofObject: objStackPage]! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 5/15/2015 10:26'!
objStack: objStack from: start do: aBlock
	"Evaluate aBlock with all elements from start (0-relative) in objStack.
	 Answer the size of the stack *before* the enumeration commences.
	 This evaluates in top-of-stack-to-bottom order.  N.B. this is also stable
	 if aBlock causes new elements to be added to the objStack, but
	 unstable if aBlock causes elements to be removed."
	<inline: true>
	| size objStackPage numToEnumerate |
	self eassert: [self isValidObjStack: weaklingStack].
	size := self fetchPointer: ObjStackTopx ofObject: objStack.
	objStackPage := self fetchPointer: ObjStackNextx ofObject: objStack.
	[objStackPage ~= 0] whileTrue:
		[size := size + ObjStackLimit.
		 self assert: (self fetchPointer: ObjStackTopx ofObject: objStackPage) = ObjStackLimit.
		 objStackPage := self fetchPointer: ObjStackNextx ofObject: objStackPage].
	numToEnumerate := size - start.
	objStackPage := objStack.
	[numToEnumerate > 0] whileTrue:
		[| numOnThisPage numToEnumerateOnThisPage topIndex |
		 numOnThisPage := self fetchPointer: ObjStackTopx ofObject: objStackPage.
		 numToEnumerateOnThisPage := numToEnumerate min: numOnThisPage.
		 topIndex := numOnThisPage + ObjStackFixedSlots - 1.
		 topIndex
			to: topIndex - numToEnumerateOnThisPage + 1
			by: -1
			do:	[:i|
				self assert: (self isWeak: (self fetchPointer: i ofObject: objStackPage)).
				aBlock value: (self fetchPointer: i ofObject: objStackPage)].
		 numToEnumerate := numToEnumerate - numToEnumerateOnThisPage.
		 objStackPage := self fetchPointer: ObjStackNextx ofObject: objStackPage].
	^size! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 10/29/2014 10:45'!
objectAfter: objOop
	<api>
	"Object parsing.
	1. all objects have at least a word following the header, for a forwarding pointer.
	2. objects with an overflow size have a preceeing word with a saturated slotSize.  If the word following
	    an object doesn't have a saturated size field it must be a single-header object.  If the word following
	   does have a saturated slotSize it must be the overflow size word."
	<inline: false>
	(self oop: objOop isLessThan: newSpaceLimit) ifTrue:
		[(self isInEden: objOop) ifTrue:
			[^self objectAfter: objOop limit: freeStart].
		 (self isInPastSpace: objOop) ifTrue:
			[^self objectAfter: objOop limit: pastSpaceStart].
		 ^self objectAfter: objOop limit: scavenger futureSurvivorStart].
	^self objectAfter: objOop limit: endOfMemory! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 12/18/2016 16:58'!
objectAfter: objOop limit: limit
	"Object parsing.
	1. all objects have at least a word following the header, for a forwarding pointer.
	2. objects with an overflow size have a preceding word with a saturated numSlots.  If the word
	   following an object doesn't have a saturated numSlots field it must be a single-header object.
	   If the word following does have a saturated numSlots it must be the overflow size word."
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 5/10/2021 19:15'!
objectAfterMaybeSlimBridge: objOop limit: limit
	"Object parsing.
	1. all objects have at least a word following the header, for a forwarding pointer.
	2. objects with an overflow size have a preceding word with a saturated numSlots.  If the word
	   following an object doesn't have a saturated numSlots field it must be a single-header object.
	   If the word following does have a saturated numSlots it must be the overflow size word.
	This variation on objectAfter:limit: allows for a single (64-bit) word bridge which may be needed
	to bridge from an almost full pastSpace to eden.  It is only used in the flat enumerators that use
	startAddressForBridgedHeapEnumeration and enumerate over pastSpace, eden and oldSpace
	in that order.  Note that the order for allObjects, and allInstances enumerates over oldSpace first.

	This hack is cheap.  It increases the size of the objectAfter code, but saves two extra copies of
	the inner loop, since the inner loop now enumerates over all of pastSpace, eden and oldSpace.
	The test for a slim bridge is only performed if applied to an overflow header, and typically only
	 1 in 400 objects have overflow headers in 32-bits, 1 in 500 in 64-bits."
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 4/25/2014 12:33'!
objectBefore: objOop
	<api>
	| prev |
	prev := nil.
	(self oop: objOop isLessThan: newSpaceLimit) ifTrue:
		[self allNewSpaceEntitiesDo:
			[:o|
			 (self oop: o isGreaterThanOrEqualTo: objOop) ifTrue:
				[^prev].
			 prev := o].
		 ^prev].
	self allOldSpaceEntitiesDo:
		[:o|
		 (self oop: o isGreaterThanOrEqualTo: objOop) ifTrue:
			[^prev].
		 prev := o].
	^prev! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 6/27/2014 15:33'!
objectBytesForSlots: numSlots
	"Answer the total number of bytes in an object with the given
	 number of slots, including header and possible overflow size header."
	<returnTypeC: #usqInt>
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 9/24/2020 21:56'!
objectInPastSpaceBefore: objOop
	"For assertions only... This ends the recursion in setting up the bridged
	 enumerations of new space that need objectBefore: for assertion checking."
	| prev obj |
	pastSpaceStart <= scavenger pastSpace start ifTrue:
		[^nil].
	prev := nil.
	obj := self objectStartingAt: scavenger pastSpace start.
	[self oop: obj isLessThan: pastSpaceStart] whileTrue:
		[(self oop: obj isGreaterThanOrEqualTo: objOop) ifTrue:
			[^prev].
		prev := obj.
		obj := self objectAfter: obj limit: pastSpaceStart].
	^prev! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 10/26/2018 13:10'!
objectMemory
	<doNotGenerate>
	^self! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 10/7/2013 13:33'!
objectStartingAt: address
	"For enumerating objects find the header of the first object in a space.
	 If the object starts with an overflow size field it will start at the next allocationUnit.
	 c.f. numSlotsOf:"
	| numSlots |
	numSlots := self rawNumSlotsOf: address.
	^numSlots = self numSlotsMask
		ifTrue: [address + self baseHeaderSize]
		ifFalse: [address]! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 12/23/2016 12:24'!
objectWithRawSlotsHasOverflowHeader: rawNumSlots
	^rawNumSlots = self numSlotsMask! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 06/22/2021 15:31'!
objectsReachableFromRoots: arrayOfRoots
	"This is part of storeImageSegmentInto:outPointers:roots:.
	 Answer an Array of all the objects only reachable from the argument, an Array of root objects,
	 starting with arrayOfRoots.  If there is no space, answer a SmallInteger whose value is the
	 number of slots required.  This is used to collect the objects to include in an image segment
	 on Spur, separate from creating the segment, hence simplifying the implementation.
	 Thanks to Igor Stasenko for this idea."

	| freeChunk ptr start limit count oop objOop |
	<var: #freeChunk type: #usqInt> "& hence start & ptr are too; limit is also because of addressAfter:"
	<inline: #never>
	self assert: (self isArray: arrayOfRoots).
	"Mark all objects except those only reachable from the arrayOfRoots by marking
	 each object in arrayOfRoots and then marking all reachable objects (from the
	 system roots).  This leaves unmarked only objects reachable from the arrayOfRoots.
	 N.B. A side-effect of the marking is that all forwarders in arrayOfRoots will be followed."
 	self assert: self allObjectsUnmarked.
	self markObjectsIn: arrayOfRoots.
	self markObjects: false.

	"After the mark phase all unreachable weak slots will have been nilled
	 and all active ephemerons fired."
	self assert: (self isEmptyObjStack: markStack).
	self assert: (self isEmptyObjStack: weaklingStack).
	self assert: self noUnscannedEphemerons.

	"Now unmark the roots before collecting the transitive closure of unmarked objects accessible from the roots."
	self unmarkObjectsIn: arrayOfRoots.

	"Use the largest free chunk to answer the result."
	freeChunk := self allocateLargestFreeChunk. "N.B. Does /not/ update totalFreeOldSpace"
	totalFreeOldSpace := totalFreeOldSpace - (self bytesInBody: freeChunk). "but must update so that growth in the markStack does not cause assert fails."
	ptr := start := freeChunk + self baseHeaderSize.
	limit := self addressAfter: freeChunk.
	count := 0.

	"First put the arrayOfRoots; order is important."
	self noCheckPush: arrayOfRoots onObjStack: markStack.

	"Now collect the roots and the transitive closure of unmarked objects from them."
	[self isEmptyObjStack: markStack] whileFalse:
		[objOop := self popObjStack: markStack.
		 self assert: (self isMarked: objOop).
		 count := count + 1.
		 ptr < limit ifTrue:
			[self longAt: ptr put: objOop.
			 ptr := ptr + self bytesPerOop].
		 oop := self fetchClassOfNonImm: objOop.
		 (self isMarked: oop) ifFalse:
			[self setIsMarkedOf: oop to: true.
			 self noCheckPush: oop onObjStack: markStack].
		 ((self isContextNonImm: objOop)
		  and: [coInterpreter isStillMarriedContext: objOop]) "widow now, before the copy loop"
			ifTrue:
				[0 to: (coInterpreter numSlotsOfMarriedContext: objOop) - 1 do:
					[:i|
					 oop := coInterpreter fetchPointer: i ofMarriedContext: objOop.
					 ((self isImmediate: oop)
					  or: [self isMarked: oop]) ifFalse:
						[self setIsMarkedOf: oop to: true.
						 self noCheckPush: oop onObjStack: markStack]]]
			ifFalse:
				[0 to: (self numPointerSlotsOf: objOop) - 1 do:
					[:i|
					 oop := self fetchPointer: i ofObject: objOop.
					 ((self isImmediate: oop)
					  or: [self isMarked: oop]) ifFalse:
						[self setIsMarkedOf: oop to: true.
						 self noCheckPush: oop onObjStack: markStack]]]].

	self unmarkAllObjects.

	"Now try and allocate the result"
	(count > (ptr - start / self bytesPerOop) "not enough room"
	 or: [limit ~= ptr and: [limit - ptr <= self allocationUnit]]) ifTrue: "can't split a single word"
		[self freeObject: freeChunk.
		 self checkFreeSpace: GCCheckImageSegment.
		 ^self integerObjectOf: count].
	"There's room; set the format, & classIndex and shorten."
	self setFormatOf: freeChunk to: self arrayFormat.
	self setClassIndexOf: freeChunk to: ClassArrayCompactIndex.
	self shorten: freeChunk toIndexableSize: count.
	(self isForwarded: freeChunk) ifTrue:
		[freeChunk := self followForwarded: freeChunk].
	self possibleRootStoreInto: freeChunk.
	self checkFreeSpace: GCCheckImageSegment.
	self runLeakCheckerFor: GCCheckImageSegment.
	^freeChunk! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'nice 3/18/2015 23:35'!
okayOop: signedOop
	"Verify that the given oop is legitimate. Check address, header, and size but not class."

	| oop classIndex fmt unusedBits unusedBitsInYoungObjects |
	<var: #oop type: #usqInt>
	<var: #unusedBits type: #usqLong>
	oop := self cCoerce: signedOop to: #usqInt.

	"address and size checks"
	(self isImmediate: oop) ifTrue: [^true].
	(self addressCouldBeObj: oop) ifFalse:
		[self error: 'oop is not a valid address'. ^false].

	(self oop: (self addressAfter: oop) isLessThanOrEqualTo: endOfMemory) ifFalse:
		[self error: 'oop size would make it extend beyond the end of memory'. ^false].

	"header type checks"
	(classIndex := self classIndexOf: oop) >= self firstClassIndexPun ifFalse:
		[self error: 'oop is a free chunk, or bridge, not an object'. ^false].
	((self rawNumSlotsOf: oop) = self numSlotsMask
	 and: [(self rawNumSlotsOf: oop - self baseHeaderSize) ~= self numSlotsMask]) ifTrue:
		[self error: 'oop header has overflow header word, but overflow word does not have a saturated numSlots field'. ^false].

	"format check"
	fmt := self formatOf: oop.
	(fmt = 6) | (fmt = 8) ifTrue:
		[self error: 'oop has an unknown format type'. ^false].
	(fmt = self forwardedFormat) ~= (classIndex = self isForwardedObjectClassIndexPun) ifTrue:
		[self error: 'oop has mis-matched format/classIndex fields; only one of them is the isForwarded value'. ^false].

	"specific header bit checks"
	unusedBits := (1 << self classIndexFieldWidth)
				   bitOr: (1 << (self identityHashFieldWidth + 32)).
	((self long64At: oop) bitAnd: unusedBits) ~= 0 ifTrue:
		[self error: 'some unused header bits are set; should be zero'. ^false].

	unusedBitsInYoungObjects := ((1 << self greyBitShift)
								   bitOr: (1 << self pinnedBitShift))
								   bitOr: (1 << self rememberedBitShift).
	((self longAt: oop) bitAnd: unusedBitsInYoungObjects) ~= 0 ifTrue:
		[self error: 'some header bits unused in young objects are set; should be zero'. ^false].
	^true
! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 4/5/2014 12:04'!
oldRawNumSlotsOf: objOop
	<returnTypeC: #usqInt>
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 1/26/2017 15:03'!
oldSpaceObjectAfter: objOop
	<api>
	"Object parsing.
	1. all objects have at least a word following the header, for a forwarding pointer.
	2. objects with an overflow size have a preceeing word with a saturated slotSize.  If the word following
	    an object doesn't have a saturated size field it must be a single-header object.  If the word following
	   does have a saturated slotSize it must be the overflow size word."
	<inline: false>
	^self objectAfter: objOop limit: endOfMemory! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 10:59'!
oldSpaceSize
	^segmentManager totalBytesInSegments! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 11/22/2013 18:09'!
oldSpaceStart
	<cmacro: '() GIV(oldSpaceStart)'>
	^oldSpaceStart! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 01/07/2021 10:40'!
outOfPlaceBecome: obj1 and: obj2 copyHashFlag: copyHashFlag
	<inline: #never> "in an effort to fix a compiler bug with two-way become post r3427"
	"Allocate two new objects, n1 & n2.  Copy the contents appropriately. Convert
	 obj1 and obj2 into forwarding objects pointing to n2 and n1 respectively"
	| clone1 clone2 |
	clone1 := (self isContextNonImm: obj1)
				ifTrue: [coInterpreter cloneContext: obj1]
				ifFalse: [self cloneObject: obj1].
	clone2 := (self isContextNonImm: obj2)
				ifTrue: [coInterpreter cloneContext: obj2]
				ifFalse: [self cloneObject: obj2].
	(self isObjImmutable: obj1) ifTrue:
		[self setIsImmutableOf: clone1 to: true].
	(self isObjImmutable: obj2) ifTrue:
		[self setIsImmutableOf: clone2 to: true].
	copyHashFlag
		ifTrue:
			[self setHashBitsOf: clone1 to: (self rawHashBitsOf: obj1).
			 self setHashBitsOf: clone2 to: (self rawHashBitsOf: obj2)]
		ifFalse:
			[self setHashBitsOf: clone1 to: (self rawHashBitsOf: obj2).
			 self setHashBitsOf: clone2 to: (self rawHashBitsOf: obj1)].
	self
		forward: obj1 to: clone2;
		forward: obj2 to: clone1.
	((self isYoungObject: obj1) ~= (self isYoungObject: clone2)
	 or: [(self isYoungObject: obj2) ~= (self isYoungObject: clone1)]) ifTrue:
		[becomeEffectsFlags := becomeEffectsFlags bitOr: OldBecameNewFlag]! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 8/24/2018 16:37'!
ownVM: flags
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter ownVM: flags! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 11/9/2017 17:03'!
pastSpaceStart
	<cmacro: '() GIV(pastSpaceStart)'>
	^pastSpaceStart! !

!SpurMemoryManager methodsFor: 'primitive support' stamp: 'eem 7/11/2017 18:57'!
pinObject: objOop
	"Attempt to pin objOop, which must not be immediate.
	 If the attempt succeeds answer objOop's (possibly moved) oop.
	 If the attempt fails, which can only occur if there is no memory, answer 0."
	<inline: false>
	| oldClone seg |
	<var: #seg type: #'SpurSegmentInfo *'>
	self assert: (self isNonImmediate: objOop).
	self flag: 'policy decision here. if already old, do we clone in a segment containing pinned objects or merely pin?'.
	"We choose to clone to keep pinned objects together to reduce fragmentation,
	 if the object is not too large, assuming that pinning is rare and that fragmentation is a bad thing.
	 Too large is defined as over 1mb.  The size of a 640x480x4 bitmap is 1228800."
	(self isOldObject: objOop) ifTrue:
		[(self numBytesOf: objOop) > (1024 * 1024) ifTrue:
			[self setIsPinnedOf: objOop to: true.
			 ^objOop].
		 seg := segmentManager segmentContainingObj: objOop.
		 seg containsPinned ifTrue:
			[self setIsPinnedOf: objOop to: true.
			 ^objOop].
		 segmentManager someSegmentContainsPinned ifFalse:
			[self setIsPinnedOf: objOop to: true.
			 seg containsPinned: true.
			 ^objOop]].
	oldClone := self cloneInOldSpace: objOop forPinning: true.
	oldClone ~= 0 ifTrue:
		[becomeEffectsFlags := self becomeEffectFlagsFor: objOop.
		 self setIsPinnedOf: oldClone to: true.
		 self forward: objOop to: oldClone.
		 self followSpecialObjectsOop.
		 coInterpreter postBecomeAction: becomeEffectsFlags.
		 self postBecomeScanClassTable: becomeEffectsFlags.
		 becomeEffectsFlags := 0].
	^oldClone! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 12/28/2016 13:06'!
pinnedBitShift
	<cmacro>
	"bit 1 of 3-bit field above format (little endian)"
	^30! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/19/2013 15:58'!
pop: nItems
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter pop: nItems! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/19/2013 12:00'!
pop: nItems thenPush: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter pop: nItems thenPush: oop! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 1/27/2020 09:06'!
popObjStack: objStack
	| topx top nextPage myx |
	self eassert: [self isValidObjStack: objStack].
	topx := self fetchPointer: ObjStackTopx ofObject: objStack.
	topx = 0 ifTrue:
		[self assert: (self fetchPointer: ObjStackNextx ofObject: objStack) = 0.
		 self cCode: '' "for debugging markAndTrace: set (MarkStackRecord := OrderedCollection new)"
			inSmalltalk:
				[MarkStackRecord ifNotNil:
					[(self fetchPointer: ObjStackMyx ofObject: objStack) = MarkStackRootIndex ifTrue:
						[MarkStackRecord addLast: {#EMPTY. nil}]]].
		^nil].
	topx := topx - 1.
	top := self fetchPointer: topx + ObjStackFixedSlots ofObject: objStack.
	self cCode: '' "for debugging markAndTrace: set (MarkStackRecord := OrderedCollection new)"
		inSmalltalk:
			[MarkStackRecord ifNotNil:
				[(self fetchPointer: ObjStackMyx ofObject: objStack) = MarkStackRootIndex ifTrue:
					[(MarkStackRecord last first = #push and: [MarkStackRecord last last = top])
						ifTrue: [MarkStackRecord removeLast]
						ifFalse: [MarkStackRecord addLast: {#pop. top}]]]].
	self storePointer: ObjStackTopx ofObjStack: objStack withValue: topx.
	(topx = 0
	 and: [(nextPage := self fetchPointer: ObjStackNextx ofObject: objStack) ~= 0])
		ifTrue:
			[self storePointer: ObjStackFreex ofObjStack: nextPage withValue: objStack.
			 self storePointer: ObjStackNextx ofObjStack: objStack withValue: 0.
			 myx := self fetchPointer: ObjStackMyx ofObject: objStack.
			 self updateRootOfObjStackAt: myx with: nextPage.
			 self eassert: [self isValidObjStack: nextPage]]
		ifFalse:
			[self eassert: [self isValidObjStack: objStack]].
	^top! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 9/20/2013 15:19'!
popRemappableOop
	"Pop and return the possibly remapped object from the remap buffer.
	 We support this excessence for compatibility with ObjectMemory.
	 Spur doesn't GC during allocation."
	<api>
	| oop |
	oop := remapBuffer at: remapBufferCount.
	remapBufferCount := remapBufferCount - 1.
	^oop! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:15'!
positive32BitIntegerFor: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter positive32BitIntegerFor: integerValue! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:16'!
positive32BitValueOf: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter positive32BitValueOf: oop! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:20'!
positive64BitIntegerFor: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter positive64BitIntegerFor: integerValue! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:20'!
positive64BitValueOf: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter positive64BitValueOf: oop! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 11/19/2016 09:38'!
positiveMachineIntegerFor: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter positiveMachineIntegerFor: integerValue! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 11/19/2016 09:38'!
positiveMachineIntegerValueOf: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter positiveMachineIntegerValueOf: integerValue! !

!SpurMemoryManager methodsFor: 'store check' stamp: 'eem 7/14/2014 16:47'!
possibleRootStoreInto: destObj
	<inline: true>
	(self isRemembered: destObj) ifFalse:
		[scavenger remember: destObj]! !

!SpurMemoryManager methodsFor: 'become implementation' stamp: 'eem 4/14/2020 14:01'!
postBecomeScanClassTable: effectsFlags
	"Scan the class table post-become (iff an active class object was becommed) to ensure no
	 forwarding pointers, and no unhashed classes exist in the class table.

	 Note that one-way become can cause duplications in the class table.
	 So if pointer objects have been becommed, scan all pages in the class table
	 and eliminate ay entries that have a zero hash. doBecome:and:copyHash: has
	 already ensured that any becomed class has been stored at the right index."
	self assert: self validClassTableRootPages.

	(effectsFlags anyMask: BecamePointerObjectFlag) ifFalse: [^self].

	0 to: numClassTablePages - 1 do:
		[:i| | page |
		page := self fetchPointer: i ofObject: hiddenRootsObj.
		self assert: (self isForwarded: page) not.
		0 to: (self numSlotsOf: page) - 1 do:
			[:j| | classOrNil |
			classOrNil := self fetchPointer: j ofObject: page.
			classOrNil ~= nilObj ifTrue:
				[(self isForwarded: classOrNil) ifTrue:
					[classOrNil := self followForwarded: classOrNil.
					 self storePointer: j ofObject: page withValue: classOrNil].
				 (self rawHashBitsOf: classOrNil) = 0 ifTrue:
					[self storePointerUnchecked: j ofObject: page withValue: nilObj.
					 "If the removed class is before the classTableIndex, set the
					  classTableIndex to point to the empty slot so as to reuse it asap."
					 (i << self classTableMajorIndexShift + j) < classTableIndex ifTrue:
						[classTableIndex := i << self classTableMajorIndexShift + j]]]]].
	"classTableIndex must never index the first page, which is reserved for classes known to the VM."
	self assert: classTableIndex >= (1 << self classTableMajorIndexShift)! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 2/12/2014 17:09'!
postSnapshot
	<doNotGenerate>
	segmentManager postSnapshot! !

!SpurMemoryManager methodsFor: 'compaction' stamp: 'eem 12/23/2016 17:22'!
prepareObjStackForPlanningCompactor: objStack
	"SpurPlanningCompactor overwrites the first fields of all moved objects, and saves these
	 fields in a data structure from which they can only be retrieved while scanning the heap.
	 The first field of an objStack page is its stack index, and so to know how many fields in an
	 objStack page to update it is necessary to save the ObjStackTopx field somewhere temporarily.
	 We use the hash field."

	| stackOrNil |
	objStack = nilObj ifTrue:
		[^self].
	stackOrNil := objStack.
	[self assert: (self numSlotsOfAny: stackOrNil) = ObjStackPageSlots.
	 self setHashBitsOf: stackOrNil to: (self fetchPointer: ObjStackTopx ofObject: stackOrNil).
	 (stackOrNil := self fetchPointer: ObjStackNextx ofObject: stackOrNil) ~= 0] whileTrue! !

!SpurMemoryManager methodsFor: 'compaction' stamp: 'eem 12/23/2016 17:19'!
prepareObjStacksForPlanningCompactor
	"SpurPlanningCompactor overwrites the first fields of all moved objects, and saves these
	 fields in a data structure from which they can only be retrieved while scanning the heap.
	 The first field of an objStack page is its stack index, and so to know how many fields in an
	 objStack page to update it is necessary to save the ObjStackTopx field somewhere temporarily.
	 We use the hash field."

	self
		prepareObjStackForPlanningCompactor: markStack;
		prepareObjStackForPlanningCompactor: weaklingStack;
		prepareObjStackForPlanningCompactor: mournQueue! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 11:03'!
primitiveErrorTable
	<api>
	^self splObj: PrimErrTableIndex! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 15:12'!
primitiveFail
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter primitiveFail! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 2/7/2014 09:59'!
primitiveFailFor: reasonCode
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter primitiveFailFor: reasonCode! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 8/22/2018 12:34'!
primitiveFailForFFIException: errorCode at: pc
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter primitiveFailForFFIException: errorCode at: pc! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'AlistairGrant 10/18/2018 22:34'!
primitiveFailForOSError: reasonCode
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter primitiveFailForOSError: reasonCode! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:23'!
primitiveFailureCode
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter primitiveFailureCode! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 11/15/2014 11:11'!
print64BitMemoryFrom: start to: end
	<doNotGenerate>
	| wordSize64 address |
	wordSize64 := 8.
	address := start bitAnd: (wordSize64 - 1) bitInvert.
	[address < end] whileTrue:
		[coInterpreter printHex: address; printChar: $:; space; printHex: (self long64At: address); cr.
		 address := address + wordSize64]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printActivationsOf: aMethodObj
	"Scan the heap printing the oops of any and all contexts that refer to anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	self allObjectsDo:
		[:obj| 
		 ((self isContextNonImm: obj)
		  and: [aMethodObj = (self fetchPointer: MethodIndex ofObject: obj)]) ifTrue:
			[coInterpreter
				printHex: obj; space; printOopShort: obj; print: ' pc ';
				printHex: (self fetchPointer: InstructionPointerIndex ofObject: obj); cr]]! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 11/26/2013 12:34'!
printAdjacentFreeChunks
	"self printAdjacentFreeChunks"
	<doNotGenerate>
	| uncoalesced |
	uncoalesced := OrderedCollection new.
	self allOldSpaceEntitiesDo:
		[:e| | s |
		((self isFreeObject: e)
		 and: [(s := self objectAfter: e limit: endOfMemory) < endOfMemory
		 and: [self isFreeObject: s]]) ifTrue:
			[uncoalesced addLast: e]].
	uncoalesced do:
		[:f|
		self printFreeChunk: f. coInterpreter printHexnp: (self objectAfter: f limit: endOfMemory); cr] 
! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printBogons
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	compactor printTheBogons: bogon! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printContextReferencesTo: anOop
	"Scan the heap printing the oops of any and all contexts that refer to anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	self allObjectsDo:
		[:obj| | i |
		 (self isContextNonImm: obj) ifTrue:
			[i := CtxtTempFrameStart + (coInterpreter fetchStackPointerOf: obj).
			 [(i := i - 1) >= 0] whileTrue:
				[anOop = (self fetchPointer: i ofObject: obj) ifTrue:
					[coInterpreter
						printHex: obj; print: ' @ '; printNum: i; space; printOopShort: obj;
						print: ' pc '; printHex: (self fetchPointer: InstructionPointerIndex ofObject: obj); cr.
					 i := 0]]]]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 06/22/2021 15:31'!
printEntity: oop
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| printFlags |
	printFlags := false.
	coInterpreter printHex: oop; space.
	(self addressCouldBeObj: oop) ifFalse:
		[^coInterpreter print: ((self isImmediate: oop) ifTrue: ['immediate'] ifFalse: ['unknown'])].
	coInterpreter
		print: ((self isFreeObject: oop) ifTrue: ['free'] ifFalse:
				[(self isSegmentBridge: oop) ifTrue: ['bridge'] ifFalse:
				[(self isForwarded: oop) ifTrue: ['forwarder'] ifFalse:
				[(self classIndexOf: oop) <= self lastClassIndexPun ifTrue: [printFlags := true. 'pun/obj stack'] ifFalse:
				[printFlags := true. 'object']]]]);
		space; printHexnpnp: (self rawNumSlotsOf: oop); print: '/'; printHexnpnp: (self bytesInBody: oop); print: '/'; printNum: (self bytesInBody: oop).
	printFlags ifTrue:
		[coInterpreter
			space;
			print: ((self formatOf: oop) <= 16rF ifTrue: ['f:0'] ifFalse: ['f:']);
			printHexnpnp: (self formatOf: oop);
			print: ((self isGrey: oop) ifTrue: [' g'] ifFalse: [' .']);
			print: ((self isImmutable: oop) ifTrue: ['i'] ifFalse: ['.']);
			print: ((self isMarked: oop) ifTrue: ['m'] ifFalse: ['.']);
			print: ((self isPinned: oop) ifTrue: ['p'] ifFalse: ['.']);
			print: ((self isRemembered: oop) ifTrue: ['r'] ifFalse: ['.'])].
	coInterpreter cr! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printForwarders
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	self allHeapEntitiesDo:
		[:objOop|
		 (self isUnambiguouslyForwarder: objOop) ifTrue:
			[coInterpreter printHex: objOop; cr]]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printFreeChunk: freeChunk
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	self printFreeChunk: freeChunk printAsTreeNode: true! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 06/22/2021 15:31'!
printFreeChunk: freeChunk printAsTreeNode: printAsTreeNode
	| numBytes |
	numBytes := self bytesInBody: freeChunk.
	coInterpreter
		print: 'freeChunk '; printHexPtrnp: freeChunk.
	printAsTreeNode ifTrue:
		[coInterpreter
			print: ((freeChunk = (freeLists at: 0)) ifTrue: [' + '] ifFalse: [' - ']);
			printHexPtrnp:(self addressAfter: freeChunk)].
	coInterpreter
		print: ' bytes '; printNum: numBytes;
		print: ' next '; printHexPtrnp: (self fetchPointer: self freeChunkNextIndex
											ofFreeChunk: freeChunk).
	(self isLilliputianSize: numBytes) ifFalse: 
		[coInterpreter
			print: ' prev '; printHexPtrnp: (self fetchPointer: self freeChunkPrevIndex
											ofFreeChunk: freeChunk).].
	(numBytes >= (self numFreeLists * self allocationUnit)
	 and: [printAsTreeNode]) ifTrue:
		[coInterpreter
			print: ' ^ '; printHexPtrnp: (self fetchPointer: self freeChunkParentIndex
											ofFreeChunk: freeChunk);
			print: ' < '; printHexPtrnp: (self fetchPointer: self freeChunkSmallerIndex
											ofFreeChunk: freeChunk);
			print: ' > '; printHexPtrnp: (self fetchPointer: self freeChunkLargerIndex
											ofFreeChunk: freeChunk)].
	coInterpreter cr! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 5/22/2014 11:37'!
printFreeChunks
	"This version goes through memory, printing all free chunks.
	 Other versions go through the free lists.  This one will show
	 all free chunks even if the free lists are corrupt."
	<api>
	| seenNewFreeChunk |
	seenNewFreeChunk := false.
	self allNewSpaceEntitiesDo:
		[:o|
		(self isFreeObject: o) ifTrue:
			[seenNewFreeChunk ifFalse:
				[coInterpreter print: 'NewSpace CONTAINS FREE OBJECT(S)!!!!'; cr.
				 seenNewFreeChunk := true].
			 self printFreeChunk: o]].
	self allOldSpaceEntitiesDo:
		[:o|
		(self isFreeObject: o) ifTrue:
			[self printFreeChunk: o]]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printFreeList: chunkOrIndex
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| freeChunk |
	(chunkOrIndex >= 0 and: [chunkOrIndex < self numFreeLists]) ifTrue:
		[^self printFreeList: (freeLists at: chunkOrIndex)].
	freeChunk := chunkOrIndex.
	[freeChunk ~= 0] whileTrue:
		[self printFreeChunk: freeChunk.
		 freeChunk := self fetchPointer: self freeChunkNextIndex ofFreeChunk: freeChunk]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printFreeListHeads
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| expectedMask |
	expectedMask := 0.
	0 to: self numFreeLists - 1 do:
		[:i|
		coInterpreter printHex: (freeLists at: i).
		(freeLists at: i) ~= 0 ifTrue:
			[expectedMask := expectedMask + (1 << i)].
		i + 1 \\ (32 >> self logBytesPerOop) = 0
			ifTrue: [coInterpreter cr]
			ifFalse: [coInterpreter print: '  ']].
	coInterpreter
		cr;
		print: 'mask: '; printHexnp: freeListsMask;
		print: ' expected: '; printHexnp: expectedMask;
		cr! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printFreeTree
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	self printFreeTreeChunk: (freeLists at: 0)! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 6/12/2014 14:52'!
printFreeTreeChunk: chunkOrZero
	<inline: false> "Slang is blind-sided by the inlining of printFreeTreeChunk: into printFreeTree."
	chunkOrZero > 0 ifTrue:
		[self printFreeTreeChunk: (self fetchPointer: self freeChunkSmallerIndex ofFreeChunk: chunkOrZero).
		 self printFreeChunk: chunkOrZero.
		 self printFreeTreeChunk: (self fetchPointer: self freeChunkLargerIndex ofFreeChunk: chunkOrZero)]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printHeaderOf: objOop
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	"N.B. No safety bounds checks!!!!  We need to look e.g. at corpses."
	coInterpreter printHexnp: objOop.
	(self numSlotsOfAny: objOop) >= self numSlotsMask
		ifTrue: [coInterpreter
					print: ' hdr16 slotf '; printHexnp: (self numSlotsOfAny: objOop - self allocationUnit);
					print: ' slotc '; printHexnp: (self rawOverflowSlotsOf: objOop); space]
		ifFalse: [coInterpreter print: ' hdr8 slots '; printHexnp: (self numSlotsOfAny: objOop)].
	coInterpreter
		space;
		printChar: ((self isMarked: objOop) ifTrue: [$M] ifFalse: [$m]);
		printChar: ((self isGrey: objOop) ifTrue: [$G] ifFalse: [$g]);
		printChar: ((self isPinned: objOop) ifTrue: [$P] ifFalse: [$p]);
		printChar: ((self isRemembered: objOop) ifTrue: [$R] ifFalse: [$r]);
		printChar: ((self isImmutable: objOop) ifTrue: [$I] ifFalse: [$i]);
		print: ' hash '; printHexnp: (self rawHashBitsOf: objOop);
		print: ' fmt '; printHexnp: (self formatOf: objOop);
		print: ' cidx '; printHexnp: (self classIndexOf: objOop);
		cr! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 2/6/2015 15:28'!
printHeaderTypeOf: objOop
	coInterpreter
		print: ((self numSlotsOfAny: objOop) >= self numSlotsMask
					ifTrue: [' hdr16 ']
					ifFalse: [' hdr8 ']);
		printChar: ((self isImmutable: objOop) ifTrue: [$i] ifFalse: [$.]);
		printChar: ((self isRemembered: objOop) ifTrue: [$r] ifFalse: [$.]);
		printChar: ((self isPinned: objOop) ifTrue: [$p] ifFalse: [$.]);
		printChar: ((self isMarked: objOop) ifTrue: [$m] ifFalse: [$.]);
		printChar: ((self isGrey: objOop) ifTrue: [$g] ifFalse: [$.])! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printInstancesOf: aClassOop
	"Scan the heap printing the oops of any and all objects that are instances of aClassOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| classIndex |
	classIndex := self rawHashBitsOf: aClassOop.
	classIndex ~= self isFreeObjectClassIndexPun ifTrue:
		[self printInstancesWithClassIndex: classIndex]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printInstancesWithClassIndex: classIndex
	"Scan the heap printing the oops of any and all objects whose classIndex equals the argument."
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	<inline: false>
	self allHeapEntitiesDo:
		[:obj|
		 (self classIndexOf: obj) = classIndex ifTrue:
			[coInterpreter printHex: obj; cr]]! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 9/25/2014 09:28'!
printInvalidClassTableEntries
	"Print the objects in the classTable that have bad hashes."
	<api>
	self validClassTableRootPages ifFalse:
		[coInterpreter print: 'class table invalid; cannot print'; cr.
		 ^self].

	self classTableEntriesDo:
		[:classOrNil :index| | hash |
		 ((self isForwarded: classOrNil)
		  or: [(hash := self rawHashBitsOf: classOrNil) = 0
		  or: [(self noCheckClassAtIndex: hash) ~= classOrNil]]) ifTrue:
			[coInterpreter
				print: 'entry '; printHex: index;
				print: ' oop '; printHex: classOrNil;
				print: ' hash '; printHex: hash; print: ' => '; printHex: (self classAtIndex: hash);
				cr]]! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 9/24/2014 11:40'!
printInvalidClassTableHashes
	"Print the entries in the classTable that have invalid hashes."

	self validClassTableRootPages ifFalse:
		[^false].

	self classTableEntriesDo:
		[:entry :index| | hash |
		 (self isForwarded: entry)
			ifTrue:
				[coInterpreter printHex: entry; print: ' @ '; printHex: index; print: ' forwarder']
			ifFalse:
				[hash := self rawHashBitsOf: entry.
				 hash = 0
					ifTrue:
						[coInterpreter printHex: entry; print: ' @ '; printHex: index; print: ' no hash']
					ifFalse:
						[(self noCheckClassAtIndex: hash) ~= entry ifTrue:
							[coInterpreter printHex: entry; print: ' @ '; printHex: index; print: ' bad hash: '; printHex: hash]]]]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printMarkedOops
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	<option: #LLDB>
	self printOopsSuchThat: #isMarked! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem (const to msg rewrite) 11/5/2014 18:28'!
printMemoryFrom: start to: end
	<doNotGenerate>
	| address |
	address := start bitAnd: (self wordSize - 1) bitInvert.
	[address < end] whileTrue:
		[coInterpreter printHex: address; printChar: $:; space; printHex: (self longAt: address); cr.
		 address := address + self wordSize]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printMethodImplementorsOf: anOop
	"Scan the heap printing the oops of any and all methods that implement anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	self allObjectsDo:
		[:obj|
		 ((self isCompiledMethod: obj)
		  and: [(coInterpreter maybeSelectorOfMethod: obj) = anOop]) ifTrue:
			[coInterpreter printHex: obj; space; printOopShort: obj; cr]]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printMethodReferencesTo: anOop
	"Scan the heap printing the oops of any and all methods that refer to anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	self allObjectsDo:
		[:obj| | i |
		 (self isCompiledMethod: obj) ifTrue:
			[i := (self literalCountOf: obj) + LiteralStart - 1.
			[(i := i - 1) >= 0] whileTrue:
				[anOop = (self fetchPointer: i ofObject: obj) ifTrue:
					[coInterpreter printHex: obj; print: ' @ '; printNum: i; space; printOopShort: obj; cr.
					 i := 0]]]]! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 3/9/2020 17:51'!
printObjStack: objStack
	<api>
	self printObjStack: objStack printContents: false! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 3/9/2020 17:51'!
printObjStack: objStack printContents: printContents
	<api>
	objStack = nilObj
		ifTrue:
			[coInterpreter print: 'nil'; cr]
		ifFalse:
			[self printObjStackPage: objStack
				myIndex: (self fetchPointer: ObjStackMyx ofObject: objStack)
				pageType: ObjStackMyx
				printContents: printContents]! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 12/12/2013 11:53'!
printObjStackPage: objStackPage myIndex: myx pageType: pageType
	| freeOrNextPage page isFirstPage isNextPage isFreePage |
	<inline: false>
	isFirstPage := pageType = ObjStackMyx.
	isNextPage := pageType = ObjStackNextx.
	isFreePage := pageType = ObjStackFreex.
	self printObjStackPage: objStackPage
		myIndex: myx
		tag: (isFirstPage ifTrue: ['head'] ifFalse: [isFreePage ifTrue: ['free'] ifFalse: ['next']]).
	(isFirstPage or: [isNextPage]) ifTrue:
		[coInterpreter tab; print: 'topx: '; printNum: (self fetchPointer: ObjStackTopx ofObject: objStackPage); print: ' next: '; printHex: (self fetchPointer: ObjStackNextx ofObject: objStackPage).
		 isFirstPage ifTrue:
			[coInterpreter print: ' free: '; printHex: (self fetchPointer: ObjStackFreex ofObject: objStackPage)].
		 coInterpreter cr].
	isFirstPage ifTrue:
		[freeOrNextPage := self fetchPointer: ObjStackFreex ofObject: objStackPage.
		 [freeOrNextPage ~= 0] whileTrue:
			[self printObjStackPage: freeOrNextPage myIndex: myx pageType: ObjStackFreex.
			 page := self fetchPointer: ObjStackFreex ofObject: freeOrNextPage.
			 (page = freeOrNextPage
			  or: [page = objStackPage]) ifTrue:
				[coInterpreter print: 'circularity in free page list!!!!'; cr.
				 page := 0].
			 freeOrNextPage := page]].
	freeOrNextPage := self fetchPointer: ObjStackNextx ofObject: objStackPage.
	freeOrNextPage ~= 0 ifTrue:
		[self printObjStackPage: freeOrNextPage myIndex: myx pageType: ObjStackNextx]! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 3/9/2020 17:50'!
printObjStackPage: objStackPage myIndex: myx pageType: pageType printContents: printContents
	| freeOrNextPage page isFirstPage isNextPage isFreePage |
	<inline: false>
	isFirstPage := pageType = ObjStackMyx.
	isNextPage := pageType = ObjStackNextx.
	isFreePage := pageType = ObjStackFreex.
	self printObjStackPage: objStackPage
		myIndex: myx
		tag: (isFirstPage ifTrue: ['head'] ifFalse: [isFreePage ifTrue: ['free'] ifFalse: ['next']]).
	(isFirstPage or: [isNextPage]) ifTrue:
		[coInterpreter tab; print: 'topx: '; printNum: (self fetchPointer: ObjStackTopx ofObject: objStackPage); print: ' next: '; printHex: (self fetchPointer: ObjStackNextx ofObject: objStackPage).
		 isFirstPage ifTrue:
			[coInterpreter print: ' free: '; printHex: (self fetchPointer: ObjStackFreex ofObject: objStackPage)].
		 coInterpreter cr].
	isFirstPage ifTrue:
		[freeOrNextPage := self fetchPointer: ObjStackFreex ofObject: objStackPage.
		 [freeOrNextPage ~= 0] whileTrue:
			[self printObjStackPage: freeOrNextPage myIndex: myx pageType: ObjStackFreex printContents: false.
			 page := self fetchPointer: ObjStackFreex ofObject: freeOrNextPage.
			 (page = freeOrNextPage
			  or: [page = objStackPage]) ifTrue:
				[coInterpreter print: 'circularity in free page list!!!!'; cr.
				 page := 0].
			 freeOrNextPage := page]].
	freeOrNextPage := self fetchPointer: ObjStackNextx ofObject: objStackPage.
	freeOrNextPage ~= 0 ifTrue:
		[self printObjStackPage: freeOrNextPage myIndex: myx pageType: ObjStackNextx printContents: printContents].
	printContents ifTrue:
		[| index |
		index := (self fetchPointer: ObjStackTopx ofObject: objStackPage) + ObjStackNextx.
		[index >= ObjStackFixedSlots] whileTrue:
			[coInterpreter space; printHex: (self fetchPointer: index ofObject: objStackPage).
			 index := index - 1].
		(self fetchPointer: ObjStackTopx ofObject: objStackPage) + ObjStackNextx >= ObjStackFixedSlots ifTrue:
			[coInterpreter cr]]! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 12/6/2013 10:56'!
printObjStackPage: objStackPage myIndex: myx tag: pageType
	<var: 'pageType' type: #'char *'>
	<inline: false>
	coInterpreter
		print: pageType; space; printHex: objStackPage; space;
		print: 'cx '; printNum: (self classIndexOf: objStackPage);
		print: ' ('; printNum: self wordSizeClassIndexPun;
		print: ') fmt '; printNum: (self formatOf: objStackPage);
		print: ' ('; printNum: self wordIndexableFormat;
		print: ') sz '; printNum: (self numSlotsOfAny: objStackPage);
		print: ' ('; printNum: ObjStackPageSlots;
		print: ') myx: '; printNum: (self fetchPointer: ObjStackMyx ofObject: objStackPage);
		print: ' ('; printNum: myx;
		print: ((self isMarked: objStackPage) ifTrue: [') mkd'] ifFalse: [') unmkd']);
		cr! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printObjectsFrom: startAddress to: endAddress
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| oop |
	oop := self objectBefore: startAddress.
	oop := oop
				ifNil: [startAddress]
				ifNotNil: [(self objectAfter: oop) = startAddress
							ifTrue: [startAddress]
							ifFalse: [oop]].
	[self oop: oop isLessThan: endAddress] whileTrue:
		[((self isFreeObject: oop)
		 or: [self isSegmentBridge: oop]) ifFalse:
			[coInterpreter printOop: oop].
		oop := self objectAfter: oop]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printObjectsWithHash: hash
	"Scan the heap printing the oops of any and all objects whose hash equals the argument."
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	self allHeapEntitiesDo:
		[:obj|
		 (self rawHashBitsOf: obj) = hash ifTrue:
			[coInterpreter shortPrintOop: obj; cr]]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printOopsExcept: function
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	<var: #function declareC: 'sqInt (*function)(sqInt)'>
	<inline: #never>
	| n |
	n := 0.
	self allHeapEntitiesDo:
		[:o|
		(self perform: function with: o) ifFalse:
			[n := n + 1.
			 self printEntity: o]].
	n > 4 ifTrue: "rabbits"
		[self printNum: n; print: ' objects'; cr]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printOopsFrom: startAddress to: endAddress
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| oop limit firstNonEntity inEmptySpace lastNonEntity |
	oop := self objectBefore: startAddress.
	limit := endAddress asUnsignedIntegerPtr min: endOfMemory.
	oop := oop
				ifNil: [startAddress]
				ifNotNil: [(self objectAfter: oop) = startAddress
							ifTrue: [startAddress]
							ifFalse: [oop]].
	inEmptySpace := false.
	[self oop: oop isLessThan: limit] whileTrue:
		[self printEntity: oop.
		 [oop := self objectAfter: oop.
		  (self long64At: oop) = 0] whileTrue:
			[inEmptySpace ifFalse:
				[inEmptySpace := true.
				 firstNonEntity := oop].
			 lastNonEntity := oop].
		 inEmptySpace ifTrue:
			[inEmptySpace := false.
			 coInterpreter
				print: 'skipped empty space from '; printHexPtrnp: firstNonEntity;
				print:' to '; printHexPtrnp: lastNonEntity; cr.
			 oop := self objectStartingAt: oop]]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printOopsSuchThat: function
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	<var: #function declareC: 'sqInt (*function)(sqInt)'>
	<inline: #never>
	| n |
	n := 0.
	self allHeapEntitiesDo:
		[:o|
		(self perform: function with: o) ifTrue:
			[n := n + 1.
			 self printEntity: o]].
	n > 4 ifTrue: "rabbits"
		[self printNum: n; print: ' objects'; cr]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printReferencesTo: anOop
	"Scan the heap printing the oops of any and all objects that refer to anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	self allObjectsDo:
		[:obj| | i |
		 i := self numPointerSlotsOf: obj.
		 [(i := i - 1) >= 0] whileTrue:
			[anOop = (self fetchPointer: i ofMaybeForwardedObject: obj) ifTrue:
				[coInterpreter printHex: obj; print: ' @ '; printNum: i; space; printOopShort: obj; cr.
				 i := 0]]]! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printUnmarkedOops
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	<option: #LLDB>
	self printOopsExcept: #isMarked! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 3/8/2016 15:05'!
purgeDuplicateClassTableEntriesFor: aClass
	"Given that either marking or allInstances has ensured that
	 all instances of aClass have the class's hash as their class
	 index, ensure aClass is in the table only at its hash (or a pun)."
	| expectedIndex |
	expectedIndex := self rawHashBitsOf: aClass.
	self classTableEntriesDo:
		[:entry :index|
		 (entry = aClass
		  and: [index ~= expectedIndex
		  and: [index > self lastClassIndexPun]]) ifTrue:
			[self classAtIndex: index put: nilObj.
			 index < classTableIndex ifTrue:
				[classTableIndex := index]]].
	"classTableIndex must never index the first page, which is reserved for classes known to the VM."
	self assert: classTableIndex >= (1 << self classTableMajorIndexShift)! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:05'!
push: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter push: oop! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 5/11/2017 14:41'!
push: objOop onObjStack: objStack
	<inline: true>
	self assert: (self addressCouldBeOop: objOop).
	(self isImmediate: objOop)
		ifTrue:
			[self assert: objStack = markStack.
			 self assert: (self addressCouldBeObj: (self topOfObjStack:
							(0 = (self fetchPointer: ObjStackTopx ofObject: objStack)
								ifTrue: [self fetchPointer: ObjStackNextx ofObject: objStack]
								ifFalse: [objStack])))]
		ifFalse: "There should be no weaklings on the mark stack."
			[self assert: (objStack = markStack and: [self isWeakNonImm: objOop]) not.
			"There should only be weaklings on the weaklingStack"
			 self assert: (objStack ~= weaklingStack or: [self isWeakNonImm: objOop])].
	^self noCheckPush: objOop onObjStack: objStack! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 14:59'!
pushBool: trueOrFalse
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter pushBool: trueOrFalse! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/24/2013 13:52'!
pushFloat: f
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter pushFloat: f! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:10'!
pushInteger: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter pushInteger: integerValue! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 5/23/2016 15:41'!
pushOnUnscannedEphemeronsStack: anEphemeron
	"Attempt to push anEphemeron on the unscanned ephemerons stack
	 and answer if the attempt succeeded.  Note that the ephemeron
	 stack overflowing isn't a disaster; it simply means treating the
	 ephemeron as strong in this GC cycle."
	<inline: false>
	self assert: (self isEphemeron: anEphemeron).
	unscannedEphemerons top >= unscannedEphemerons limit ifTrue:
		[^false].
	self longAt: unscannedEphemerons top put: anEphemeron.
	unscannedEphemerons top: unscannedEphemerons top + self bytesPerOop.
	^true! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 9/20/2013 15:18'!
pushRemappableOop: oop
	"Record the given object in a the remap buffer. Objects in this buffer are remapped
	 when a compaction occurs. This facility is used by the interpreter to ensure that
	 objects in temporary variables are properly remapped.
	 We support this excessence for compatibility with ObjectMemory.
	 Spur doesn't GC during allocation."
	<api>
	self assert: (self addressCouldBeOop: oop).
	remapBuffer at: (remapBufferCount := remapBufferCount + 1) put: oop.
	remapBufferCount <= RemapBufferSize ifFalse:
		[self error: 'remapBuffer overflow']! !

!SpurMemoryManager methodsFor: 'weakness and ephemerality' stamp: 'eem 3/17/2020 19:20'!
queueMourner: anEphemeronOrWeakArray
	"Add the ephemeron or weak array to the queue."
	self assert: ((self isNonImmediate: anEphemeronOrWeakArray)
				and: [(self formatOf: anEphemeronOrWeakArray) = self ephemeronFormat
				   or: [(self formatOf: anEphemeronOrWeakArray) = self weakArrayFormat]]).
	self deny: ((self formatOf: anEphemeronOrWeakArray) = self ephemeronFormat
				and: [self is: anEphemeronOrWeakArray onObjStack: mournQueue]).
	self ensureRoomOnObjStackAt: MournQueueRootIndex.
	"There is no point queueing weak arrays more than once.  Note that it should be impossible
	 for ephemerons to be enqueued more than once since they are turned into non-ephemerons
	 in the relevant sender.  Alas this means that other ephemerons on the same object not
	 identified in this scavenge or GC will not fire until later.  But that's life."
	(self isEphemeron: anEphemeronOrWeakArray) ifFalse:
		[(self is: anEphemeronOrWeakArray onObjStack: mournQueue) ifTrue:
			[^self]].
	self push: anEphemeronOrWeakArray onObjStack: mournQueue! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 7/25/2014 21:48'!
rawClassTagForClass: classObj
	"Answer the classObj's identityHash to use as a tag in a class comparison."
	^self rawHashBitsOf: classObj! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 7/12/2016 09:47'!
rawHashBitsOf: objOop
	<api>
	self flag: #endianness.
	^(self long32At: objOop + 4) bitAnd: self identityHashHalfWordMask! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 4/5/2014 12:05'!
rawNumSlotsOf: objOop
	<returnTypeC: #usqInt>
	<inline: true>
	self flag: #endian.
	^self byteAt: objOop + 7! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 4/5/2014 12:07'!
rawNumSlotsOf: objOop put: aByte
	<returnTypeC: #usqInt>
	<inline: true>
	self flag: #endian.
	^self byteAt: objOop + 7 put: aByte! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 10/10/2013 09:21'!
rawOverflowSlotsOf: objOop
	<returnTypeC: #usqInt>
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 4/5/2014 12:07'!
rawOverflowSlotsOf: objOop put: numSlots
	<returnTypeC: #usqInt>
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 10/10/2013 15:24'!
readHeapFromImageFile: f dataBytes: numBytes
	"Read numBytes of image data from f into memory at memoryBaseForImageRead.
	 Answer the number of bytes written."
	<doNotGenerate>
	^segmentManager readHeapFromImageFile: f dataBytes: numBytes! !

!SpurMemoryManager methodsFor: 'cog jit support' stamp: 'eem 9/5/2021 06:55'!
receiverTagBitsForMethod: aMethodObj
	"Answer the tag bits for the receiver based on the method's methodClass, if any.
	 These bits are extended with a bit that says that the method may have a Context receiver, i.e.
	 mclass is Context or its superclasses.  The absence of this bit is used to avoid expensive store checks, etc."
	<api>
	<option: #Cogit>
	| methodClassOrNil classContextOrSuperclass |
	methodClassOrNil := coInterpreter methodClassOf: aMethodObj.
	methodClassOrNil = nilObj ifTrue: "If we don't know the methodClass be pessimal"
		[^self maybeContextMClassTagBits].
	NewspeakVM "Mixins don't necessarily have a format inst var; filter out non-integer format."
		ifTrue:
			[| instSpec |
			((self isIntegerObject: (instSpec := self fetchPointer: InstanceSpecificationIndex ofObject: methodClassOrNil))
			 and: [(self instSpecOfClassFormat: (self integerValueOf: instSpec)) = self forwardedFormat]) ifTrue:
				[^self receiverTagBitsForImmediateMethodClass: methodClassOrNil]]
		ifFalse:
			[(self instSpecOfClass: methodClassOrNil) = self forwardedFormat ifTrue:
				[^self receiverTagBitsForImmediateMethodClass: methodClassOrNil]].
	"Now check if mclass is one of Context's superclasses..."
	classContextOrSuperclass := self knownClassAtIndex: ClassMethodContextCompactIndex.
	[classContextOrSuperclass = methodClassOrNil ifTrue:
		[^self maybeContextMClassTagBits].
	 classContextOrSuperclass := self fetchPointer: SuperclassIndex ofObject: classContextOrSuperclass.
	 classContextOrSuperclass ~= nilObj]
		whileTrue.
	^0! !

!SpurMemoryManager methodsFor: 'compaction' stamp: 'eem 3/10/2020 19:28'!
relocateObjStackForPlanningCompactor: objStack andContents: relocateContents
	"Relocate all objStack pages that comprise objStack, including contents if requested."
	| stackOrNil freeList next relocated result |
	objStack = nilObj ifTrue:
		[^objStack].
	self assert: (relocateContents ifTrue: [(self rawHashBitsOf: objStack) < ObjStackLimit] ifFalse: [(self rawHashBitsOf: objStack) = 0]).
	stackOrNil := objStack.
	freeList := self fetchPointer: ObjStackFreex ofObject: objStack.
	[self assert: (self numSlotsOfAny: stackOrNil) = ObjStackPageSlots.
	 "There are four fixed slots in an obj stack, and a Topx of 0 indicates empty, so
	   if there were 5 slots in an oop stack, full would be 2, and the last 0-rel index is 4.
	   Hence the last index is topx + fixed slots - 1, or topx + ObjStackNextx. The first
	   two slots, Topx and Myx are not object references."
	 next := self fetchPointer: ObjStackNextx ofObject: stackOrNil.
	 relocated := compactor
					relocateObjectsInHeapEntity: stackOrNil
					from: ObjStackFreex
					to: ObjStackNextx + (relocateContents ifTrue: [self rawHashBitsOf: stackOrNil] ifFalse: [0]).
	 stackOrNil = objStack ifTrue:
		[result := relocated].
	 next ~= 0]
		whileTrue:
			[stackOrNil := next].
	[freeList ~= 0] whileTrue:
		[self assert: (self numSlotsOfAny: freeList) = ObjStackPageSlots.
		 next := self fetchPointer: ObjStackFreex ofObject: freeList.
		 compactor
			relocateObjectsInHeapEntity: freeList
			from: ObjStackFreex
			to: ObjStackFreex.
		 freeList := next].
	^relocated! !

!SpurMemoryManager methodsFor: 'compaction' stamp: 'eem 3/10/2020 12:22'!
relocateObjStacksForPlanningCompactor
	"Relocate all non-empty objStack pages, following the objStacks from the roots."

	markStack := self relocateObjStackForPlanningCompactor: markStack andContents: false.
	weaklingStack := self relocateObjStackForPlanningCompactor: weaklingStack andContents: false.
	mournQueue := self relocateObjStackForPlanningCompactor: mournQueue andContents: true! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 9/15/2013 22:56'!
remap: oop
	self shouldNotImplement! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 9/20/2013 15:22'!
remapBuffer
	"We support this excessence for compatibility with ObjectMemory.
	 Spur doesn't GC during allocation."
	^remapBuffer! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/13/2014 14:30'!
remapBufferCount
	<cmacro: '() GIV(remapBufferCount)'>
	^remapBufferCount! !

!SpurMemoryManager methodsFor: 'gc - scavenge/compact' stamp: 'eem 12/24/2016 14:32'!
remapObj: objOop
	"Scavenge or simply follow objOop.  Answer the new location of objOop.
	 The send should have been guarded by a send of shouldRemapOop:.
	 The method is called remapObj: for compatibility with ObjectMemory.
	 Defer to the compactor to choose the actual method, there being a
	 difference between the vanilla method and that used with a sliding
	 compactor where objects are not marked as forwarded."
	<doNotGenerate>
	^compactor remapObj: objOop! !

!SpurMemoryManager methodsFor: 'header format' stamp: 'eem 12/12/2015 14:19'!
rememberedBitShift
	<api>
	<cmacro>
	"bit 0 of 3-bit field above format (little endian)"
	^29! !

!SpurMemoryManager methodsFor: 'scavenger' stamp: 'eem 11/14/2019 14:18'!
rememberedSetObj
	<inline: #always>
	^self fetchPointer: RememberedSetRootIndex ofObject: hiddenRootsObj! !

!SpurMemoryManager methodsFor: 'scavenger' stamp: 'eem 7/19/2014 10:14'!
rememberedSetObj: anObj
	self assert: (self isOldObject: anObj).
	self storePointerUnchecked: RememberedSetRootIndex ofObject: hiddenRootsObj withValue: anObj! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 10/15/2013 09:53'!
removeGCRoot: varLoc
	"Remove the given variable location to the extra roots table."
	<api>
	<var: #varLoc type: #'sqInt *'>
	1 to: extraRootCount do:
		[:i|
		varLoc = (extraRoots at: i) ifTrue: "swap varLoc with last entry"
			[extraRoots at: i put: (extraRoots at: extraRootCount).
			 extraRootCount := extraRootCount - 1.
			 ^true]].
	^false "not found"! !

!SpurMemoryManager methodsFor: 'allocation accounting' stamp: 'eem 6/7/2017 14:09'!
resetAllocationAccountingAfterGC
	"oldSpaceUsePriorToScavenge is used to maintain an accurate allocation count.
	 Since scavenging may tenure objects and tenuring does not count as allocation (that
	 would count twice) we must compute heapSizeAtPreviousGC after any tenuring.
	 fullGC reclaims space which does not count as deallocation (that would not count
	 allocations at all), so we must reset heapSizeAtPreviousGC after GC also."
	<inline: true>
	oldSpaceUsePriorToScavenge := segmentManager totalOldSpaceCapacity - totalFreeOldSpace! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 1/30/2017 11:11'!
resetFreeListHeads
	freeListsMask := 0.
	0 to: self numFreeLists - 1 do:
		[:i| freeLists at: i put: 0]! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 5/9/2017 12:10'!
restoreObjectsIn: objArray savedHashes: savedHashes
	"This is part of storeImageSegmentInto:outPointers:roots:.
	 Enumerate the objects in objArray, unmarking them and restoring their hashes
	 from the corresponding 32-bit slots in savedHashes.  The first unused entry in
	 objArray will have a non-hash value entry in savedHashes.  Free savedHashes."
	<inline: true>
	0 to: (self numSlotsOf: objArray) - 1 do:
		[:i| | hash oop |
		(hash := self fetchLong32: i ofObject: savedHashes) > self maxIdentityHash ifTrue:
			[(self isInOldSpace: savedHashes) ifTrue:
				[self freeObject: savedHashes].
			 ^self].
		oop := self fetchPointer: i ofObject: objArray.
		self setHashBitsOf: oop to: hash.
		self setIsMarkedOf: oop to: false].
	(self isInOldSpace: savedHashes) ifTrue:
		[self freeObject: savedHashes]! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 5/9/2017 13:45'!
restoreObjectsIn: objArray upTo: limitOrTag savedFirstFields: savedFirstFields
	"This is part of storeImageSegmentInto:outPointers:roots:.
	 Enumerate the objects in objArray, unmarking them and restoring their saved first fields
	 from the corresponding slot in savedFirstFields.  The first unused entry in
	 objArray will have a non-hash value entry in savedHashes.  Free savedFirstFields."
	<inline: true>
	| numSlots |
	numSlots := limitOrTag = -1 ifTrue: [self numSlotsOf: objArray] ifFalse: [limitOrTag].
	0 to: numSlots - 1 do:
		[:i| | oop |
		oop := self fetchPointer: i ofObject: objArray.
		self storePointerUnchecked: 0 ofObject: oop withValue: (self fetchPointer: i ofObject: savedFirstFields).
		self setIsMarkedOf: oop to: false].
	(self isInOldSpace: savedFirstFields) ifTrue:
		[self freeObject: savedFirstFields]! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 6/21/2021 12:01'!
return: errCode restoringObjectsIn: firstArray savedFirstFields: savedFirstFields and: secondArray savedHashes: savedHashes
	<inline: false>
	"This is part of storeImageSegmentInto:outPointers:roots:."
	self cCode: [] inSmalltalk: [errCode ~= 0 ifTrue: [self halt]].
	self restoreObjectsIn: firstArray upTo: -1 savedFirstFields: savedFirstFields.
	self restoreObjectsIn: secondArray savedHashes: savedHashes.
	self runLeakCheckerFor: GCCheckImageSegment.
	self assert: self allObjectsUnmarked.
	^errCode! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 6/21/2021 12:01'!
return: errCode restoringObjectsIn: firstArray upTo: limitOrTag savedFirstFields: savedFirstFields
	<inline: false>
	"This is part of storeImageSegmentInto:outPointers:roots:."
	self cCode: [] inSmalltalk: [errCode ~= 0 ifTrue: [self halt]].
	self restoreObjectsIn: firstArray upTo: limitOrTag savedFirstFields: savedFirstFields.
	self runLeakCheckerFor: GCCheckImageSegment.
	self assert: self allObjectsUnmarked.
	^errCode! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem (const to msg rewrite) 11/5/2014 18:28'!
reverseBytesFrom: startAddr to: stopAddr
	"Byte-swap the given range of memory (not inclusive of stopAddr!!)."
	| addr |
	addr := startAddr.
	[self oop: addr isLessThan: stopAddr] whileTrue:
		[self longAt: addr put: (self byteSwapped: (self longAt: addr)).
		addr := addr + self wordSize]! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'nice 4/2/2016 22:45'!
reverseBytesIn32BitWordsFrom: startAddr to: stopAddr
	"Byte-swap the given range of memory (not inclusive of stopAddr!!)."
	| addr |
	addr := startAddr.
	[self oop: addr isLessThan: stopAddr] whileTrue:
		[self long32At: addr put: ((self long32At: addr) byteSwap32).
		 addr := addr + 4]! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 5/2/2018 17:14'!
reverseBytesIn32BitWordsIn: segmentWordArray
	"This exists to get around a compiler bug in Apple LLVM version 7.0.0 (clang-700.1.76)
	 that was avoiding the second comparison of segVersion after the first byte swap."
	<inline: #never>
	self reverseBytesIn32BitWordsFrom: segmentWordArray + self baseHeaderSize
		to: (self addressAfter: segmentWordArray)! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 11/21/2013 11:30'!
reverseBytesInMemory
	self reverseBytesFrom: oldSpaceStart to: endOfMemory! !

!SpurMemoryManager methodsFor: 'scavenger' stamp: 'eem 5/5/2015 16:14'!
rootTableCapacity
	<cmacro: '() GIV(rememberedSetLimit)'>
	^scavenger rememberedSetLimit! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 5/5/2015 16:14'!
rootTableCount
	<cmacro: '() GIV(rememberedSetSize)'>
	^scavenger rememberedSetSize! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 2/3/2014 17:40'!
roundBytesToAllocationUnit: bytes
	^bytes + (self allocationUnit - 1) bitClear: self allocationUnit - 1! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 3/23/2015 12:07'!
roundUpHeapSize: heapSize
	<var: 'heapSize' type: #usqInt>
	| bit |
	bit := heapSize highBit - 1 * 3 // 4.
	^(heapSize anyMask: (1 << bit - 1))
		ifTrue: [(heapSize bitClear: (1 << bit - 1)) + (1 << bit)]
		ifFalse: [heapSize]! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 5/5/2015 17:37'!
rtRefCountOf: obj
	"Answer the rt reference count of obj; this is the three bit field comprised
	 of isGrey,isPinned,isRemembered.  See computeRefCountToShrinkRT."
	^(self longAt: obj) >> self rememberedBitShift bitAnd: MaxRTRefCount! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 5/5/2015 17:08'!
rtRefCountOf: obj put: refCount
	"Set the rt reference count of obj; this is the three bit field comprised
	 of isGrey,isPinned,isRemembered.  See computeRefCountToShrinkRT."
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 12/25/2016 11:29'!
runLeakCheckerFor: gcModes
	<inline: false>
	^self
		inLineRunLeakCheckerFor: gcModes
		excludeUnmarkedObjs: false
		classIndicesShouldBeValid: true! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 12/25/2016 11:29'!
runLeakCheckerFor: gcModes excludeUnmarkedObjs: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid
	<inline: false>
	self inLineRunLeakCheckerFor: gcModes
		excludeUnmarkedObjs: excludeUnmarkedObjs
		classIndicesShouldBeValid: classIndicesShouldBeValid! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 6/21/2021 12:04'!
runLeakCheckerForFreeSpace: gcModes ignoring: anOopOrNil
	"Check free space integrity by setting bits in the map corresponding to all free space objects
	 and checking tat no pointer field refers to a free object.  anOopOrNil is provided to filter-out
	 the as-yet-to-be initialized object in primitiveShallowCopy/primitiveClone."
	<inline: false>
	(gcModes anyMask: GCCheckFreeSpace) ifTrue:
		[coInterpreter reverseDisplayFrom: 16 to: 19.
		 self clearLeakMapAndMapAccessibleFreeSpace.
		 freeSpaceCheckOopToIgnore := anOopOrNil.
		 self asserta: self checkHeapFreeSpaceIntegrity.
		 freeSpaceCheckOopToIgnore := nil]! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 2/7/2015 10:26'!
runLeakCheckerForFullGC
	<doNotGenerate>
	"Support for the Spur bootstrap"
	self runLeakCheckerFor: GCModeFull! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 12/28/2019 13:32'!
safeLastPointerOf: objOop 
	"Answer the byte offset of the last pointer field of the given object.
	 Works with CompiledMethods, as well as ordinary objects."
	<inline: true>
	| fmt contextSize header byteSize |
	byteSize := (self numSlotsOf: objOop) - 1 * self bytesPerOop + self baseHeaderSize.
	fmt := self formatOf: objOop.
	self assert: fmt ~= self forwardedFormat.
	fmt <= self lastPointerFormat ifTrue:
		[(fmt = self indexablePointersFormat
		  and: [self isContextNonImm: objOop]) ifTrue:
			["contexts end at the stack pointer"
			contextSize := coInterpreter fetchStackPointerOf: objOop.
			^CtxtTempFrameStart - 1 + contextSize * self bytesPerOop + self baseHeaderSize min: byteSize].
		^byteSize  "all pointers"].
	fmt < self firstCompiledMethodFormat ifTrue: [^0]. "no pointers"

	"CompiledMethod: contains both pointers and bytes"
	header := self methodHeaderOf: objOop.
	^(self lastPointerOfMethodHeader: header) min: byteSize! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 11/17/2013 15:51'!
safePrintStringOf: oop
	| target |
	(self isOopForwarded: oop)
		ifTrue: [target := self followForwarded: oop]
		ifFalse: [target := oop].
	^coInterpreter printStringOf: target! !

!SpurMemoryManager methodsFor: 'testing' stamp: 'eem 12/24/2016 13:59'!
scavengeInProgress
	^gcPhaseInProgress == ScavengeInProgress! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 6/16/2014 13:16'!
scavengeThreshold
	^scavengeThreshold! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/27/2016 10:51'!
scavengeThresholdAsExtent
	^scavengeThreshold - scavenger eden start! !

!SpurMemoryManager methodsFor: 'spur bootstrap' stamp: 'eem 9/10/2013 17:09'!
scavenger
	<doNotGenerate>
	^scavenger! !

!SpurMemoryManager methodsFor: 'scavenger' stamp: 'eem 10/9/2013 13:14'!
scavengerDenominator
	"David's paper uses 140Kb eden + 2 x 28kb survivor spaces,
	 which is 5 7ths for eden and 1 7th each for the survivor spaces.
	 So express scavenger sizes in 7ths"
	^7! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 10/3/2013 21:05'!
scavengingGC
	"Run the scavenger."

	self scavengingGCTenuringIf: TenureByAge! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 11/13/2017 11:50'!
scavengingGCTenuringIf: tenuringCriterion
	"Run the scavenger."
	<inline: false>
	self assert: remapBufferCount = 0.
	(self asserta: scavenger eden limit - freeStart > coInterpreter interpreterAllocationReserveBytes) ifFalse:
		[coInterpreter tab;
			printNum: scavenger eden limit - freeStart; space;
			printNum: coInterpreter interpreterAllocationReserveBytes; space;
			printNum: coInterpreter interpreterAllocationReserveBytes - (scavenger eden limit - freeStart); cr].
	self checkMemoryMap.
	self checkFreeSpace: GCModeNewSpace.
	self runLeakCheckerFor: GCModeNewSpace.

	coInterpreter
		preGCAction: GCModeNewSpace;
		"would prefer this to be in mapInterpreterOops, but
		 compatibility with ObjectMemory dictates it goes here."
		flushMethodCacheFrom: newSpaceStart to: newSpaceLimit.
	needGCFlag := false.

	gcStartUsecs := coInterpreter ioUTCMicrosecondsNow.

	self doScavenge: tenuringCriterion.

	statScavenges := statScavenges + 1.
	statGCEndUsecs := coInterpreter ioUTCMicrosecondsNow.
	statSGCDeltaUsecs := statGCEndUsecs - gcStartUsecs.
	statScavengeGCUsecs := statScavengeGCUsecs + statSGCDeltaUsecs.
	statRootTableCount := scavenger rememberedSetSize.

	scavenger logScavenge.

	coInterpreter postGCAction: GCModeNewSpace.

	self runLeakCheckerFor: GCModeNewSpace.
	self checkFreeSpace: GCModeNewSpace! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 9/14/2013 08:50'!
scheduleScavenge
	needGCFlag := true.
	coInterpreter forceInterruptCheck! !

!SpurMemoryManager methodsFor: 'class table puns' stamp: 'eem 3/9/2016 14:49'!
segmentBridgePun
	<cmacro>
	^3! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/8/2013 11:37'!
segmentManager
	^segmentManager! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 1/10/2017 15:56'!
set: objOop classIndexTo: classIndex formatTo: format
	"0 = 0 sized objects (UndefinedObject True False et al)
	 1 = non-indexable objects with inst vars (Point et al)
	 2 = indexable objects with no inst vars (Array et al)
	 3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
	 4 = weak indexable objects with inst vars (WeakArray et al)
	 5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
	 6 unused, reserved for exotic pointer objects?
	 7 Forwarded Object, 1st field is pointer, rest of fields are ignored
	 8 unused, reserved for exotic non-pointer objects?
	 9 64-bit indexable
	 10 - 11 32-bit indexable
	 12 - 15 16-bit indexable
	 16 - 23 byte indexable
	 24 - 31 compiled method"
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'spur bootstrap' stamp: 'eem 6/21/2021 12:00'!
setCheckForLeaks: integerFlags
	" 0 = do nothing.
	  1 = check for leaks on fullGC (GCModeFull).
	  2 = check for leaks on scavenger (GCModeNewSpace).
	  4 = check for leaks on incremental (GCModeIncremental)
	  8 = check for leaks on become (GCModeBecome)
	 16 = check for leaks on image segments (GCCheckImageSegment)
	 32 = check for leaks in free space (GCCheckFreeSpace)
	 64 = check for leaks on object shortening (GCCheckShorten)
	 128 = check for leaks on simulated run-time call (GCCheckPrimCall)"
	checkForLeaks := integerFlags! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 9/22/2013 15:47'!
setClassIndexOf: objOop to: classIndex
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'allocation accounting' stamp: 'eem 6/7/2017 14:03'!
setCurrentAllocatedBytesTo: n
	"Reset the allocation count to n (n will typically be zero).  Since we wish to
	 discount the current use we must set statAllocatedBytes to n and update
	 oldSpaceUsePriorToScavenge to discount the current allocated bytes."
	| delta |
	delta := self currentAllocatedBytes - statAllocatedBytes.
	statAllocatedBytes := n.
	oldSpaceUsePriorToScavenge := oldSpaceUsePriorToScavenge + delta.
	self assert: self currentAllocatedBytes = n! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 10/10/2013 17:32'!
setEndOfMemory: newEndOfMemory
	"Set by the segment manager after swizzling the image,
	 and by the SpurBootstrap on writing out the transformed image."
	endOfMemory := newEndOfMemory.
	freeOldSpaceStart > newEndOfMemory ifTrue:
		[freeOldSpaceStart := newEndOfMemory]! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 1/10/2017 15:56'!
setFormatOf: objOop to: format
	"0 = 0 sized objects (UndefinedObject True False et al)
	 1 = non-indexable objects with inst vars (Point et al)
	 2 = indexable objects with no inst vars (Array et al)
	 3 = indexable objects with inst vars (MethodContext AdditionalMethodState et al)
	 4 = weak indexable objects with inst vars (WeakArray et al)
	 5 = weak non-indexable objects with inst vars (ephemerons) (Ephemeron)
	 6 unused, reserved for exotic pointer objects?
	 7 Forwarded Object, 1st field is pointer, rest of fields are ignored
	 8 unused, reserved for exotic non-pointer objects?
	 9 64-bit indexable
	 10 - 11 32-bit indexable
	 12 - 15 16-bit indexable
	 16 - 23 byte indexable
	 24 - 31 compiled method"
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 11/26/2013 14:15'!
setFree: objOop
	<inline: true>
	"turn the object into a free chunk, zeroing classIndex, format, isGrey,isPinned,isRemembered,isImmutable & ?."
	self long32At: objOop put: 0! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 10/11/2013 16:07'!
setFreeOldSpaceStart: newFreeOldSpaceStart
	"Set by the segment manager on parsing the image."
	freeOldSpaceStart := newFreeOldSpaceStart! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 11/18/2014 16:56'!
setHashBitsOf: objOop to: hash
	self flag: #endianness.
	self assert: (hash between: 0 and: self identityHashHalfWordMask).
	self long32At: objOop + 4
		put: ((self long32At: objOop + 4) bitClear: self identityHashHalfWordMask) + hash! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 10/28/2016 11:51'!
setHeapBase: baseOfHeap memoryLimit: memLimit endOfMemory: memEnd
	"Set the dimensions of the heap, answering the start of oldSpace. edenBytes holds the desired ``size of eden''
	 which is actually the total size of new space minus the reserve.  edenBytes is then divided up between eden
	 and the two survivor spaces, where each survivor space is a scavengerDenominator (one seventh) of the total."
	"Transcript
		cr; nextPutAll: 'heapBase: '; print: baseOfHeap; nextPut: $/; nextPutAll: baseOfHeap hex;
		nextPutAll: ' memLimit '; print: memLimit; nextPut: $/; nextPutAll: memLimit hex;
		nextPutAll: ' memEnd '; print: memEnd; nextPut: $/; nextPutAll: memEnd hex; cr; flush."
	"This is more than a little counter-intuitive.  Eden must include interpreterAllocationReserveBytes."
	<inline: #never>
	| reserve |
	reserve := coInterpreter interpreterAllocationReserveBytes.
	newSpaceStart := baseOfHeap.
	newSpaceLimit := baseOfHeap + edenBytes + reserve.
	scavenger newSpaceStart: newSpaceStart
				newSpaceBytes: newSpaceLimit - newSpaceStart
				survivorBytes: newSpaceLimit - newSpaceStart - reserve // self scavengerDenominator.

	freeStart := scavenger eden start.
	pastSpaceStart := scavenger pastSpace start.

	oldSpaceStart := newSpaceLimit.
	freeOldSpaceStart := memEnd.
	endOfMemory := memLimit.

	^baseOfHeap! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 7/23/2014 10:16'!
setHeapGrowthToSizeGCRatio: aDouble
	<var: #aDouble type: #double>
	heapGrowthToSizeGCRatio := aDouble.
	^0! !

!SpurMemoryManager methodsFor: 'allocation accounting' stamp: 'eem 6/7/2017 14:10'!
setHeapSizeAtPreviousGC
	"heapSizeAtPreviousGC is used to invoke full GC when lots of oldSpace objects are created.
	 Also reset oldSpaceUsePriorToScavenge."
	<inline: true>
	heapSizeAtPreviousGC := segmentManager totalOldSpaceCapacity - totalFreeOldSpace.
	self resetAllocationAccountingAfterGC! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 8/19/2020 18:13'!
setHiddenRootsObj: anOop
	hiddenRootsObj := anOop.
	self cCode: [self assert: self validClassTableRootPages]
		inSmalltalk: [numClassTablePages ifNotNil:
						[self assert: self validClassTableRootPages]].
	classTableFirstPage := self fetchPointer: 0 ofObject: hiddenRootsObj.
	self assert: (self numSlotsOf: classTableFirstPage) - 1 = self classTableMinorIndexMask.
	"Hack fix.  A bug in markAndTraceClassOf: caused the class of the first class table page
	 to be changed from its pun.  This can be restored manually, but we do it here too."
	self flag: 'remove at some stage'.
	(self classIndexOf: classTableFirstPage) ~= self arrayClassIndexPun ifTrue:
		[self setClassIndexOf: classTableFirstPage to: self arrayClassIndexPun].
	numClassTablePages := self classTableRootSlots.
	self bootstrapping ifTrue:
		["Set classTableIndex to the start of the last used page (excepting first page).
		  Set numClassTablePages to the number of used pages."
		2 to: numClassTablePages - 1 do:
			[:i|
			(self fetchPointer: i ofObject: hiddenRootsObj) = nilObj ifTrue:
				[numClassTablePages := i.
				 classTableIndex := (i - 1 max: 1) << self classTableMajorIndexShift.
				 ^self]].
		"no unused pages; set it to the start of the second page."
		classTableIndex := 1 << self classTableMajorIndexShift.
		^self].
	"If loading an image, set the classTableIndex to the first unused slot in the class table after the first page.
	 Set numClassTablePages to the number of used pages.
	 Set classTableIndex to point at the first unused entry. First set it to the max as a sentinel."
	classTableIndex := numClassTablePages << self classTableMajorIndexShift.
	1 to: numClassTablePages - 1 do:
		[:i| | page j |
		(page := self fetchPointer: i ofObject: hiddenRootsObj) = nilObj
			ifFalse:
				[classTableIndex >> self classTableMajorIndexShift > i ifTrue:
					[j := 0.
					 [j < self classTablePageSize] whileTrue:
						[(self fetchPointer: j ofObject: page) = nilObj ifTrue:
							[classTableIndex := i << self classTableMajorIndexShift + j.
							 j := self classTablePageSize].
						 j := j + 1]]]
			ifTrue:
				[classTableIndex >> self classTableMajorIndexShift > i ifTrue:
					[classTableIndex := (i - 1 max: 1) << self classTableMajorIndexShift].
				 numClassTablePages := i.
				 self assert: (self classOrNilAtIndex: classTableIndex) = nilObj.
				 ^self]].
	"no unused slots; set it to the start of the second page."
	classTableIndex >> self classTableMajorIndexShift >= numClassTablePages ifTrue:
		[classTableIndex := 1 << self classTableMajorIndexShift].
	 self assert: (self classOrNilAtIndex: classTableIndex) = nilObj! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 10/28/2020 01:31'!
setInterruptCheckChain: ignored
	<doNotGenerate>
	^0! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 10/15/2013 07:37'!
setIsGreyOf: objOop to: aBoolean
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 10/15/2013 00:28'!
setIsImmutableOf: objOop to: aBoolean
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 10/15/2013 07:48'!
setIsMarkedOf: objOop to: aBoolean
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 10/7/2013 14:15'!
setIsPinnedOf: objOop to: aBoolean
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'header access' stamp: 'eem 9/22/2013 15:45'!
setIsRememberedOf: objOop to: aBoolean
	self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'growing/shrinking memory' stamp: 'eem 5/21/2014 10:17'!
setLastSegment: segInfo
	"Update after removing a segment.
	 Here we cut back endOfMemory if required."
	<var: #segInfo type: #'SpurSegmentInfo *'>
	| currentEnd |
	<var: #currentEnd type: #usqInt>
	currentEnd := segInfo segLimit - self bridgeSize.
	currentEnd <= endOfMemory ifTrue:
		[endOfMemory := currentEnd.
		 freeOldSpaceStart > currentEnd ifTrue:
			[freeOldSpaceStart :=currentEnd]]! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'nice 7/15/2016 23:40'!
setMaxOldSpaceSize: limit
	<var: #limit type: #usqInt>
	maxOldSpaceSize := limit.
	^0! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'cb 10/5/2018 12:54'!
setNextFreeChunkOf: freeChunk withValue: nextFreeChunk chunkBytes: chunkBytes
	self 
		setNextFreeChunkOf: freeChunk 
		withValue: nextFreeChunk 
		isLilliputianSize: (self isLilliputianSize: chunkBytes) 
	
	! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'cb 10/5/2018 12:59'!
setNextFreeChunkOf: freeChunk withValue: nextFreeChunk isLilliputianSize: lilliputian 
	<inline: true> "Inlining is quite important since isLilliputianSize: is often true/false"
	self 
		storePointer: self freeChunkNextIndex 
		ofFreeChunk: freeChunk 
		withValue: nextFreeChunk.
	(nextFreeChunk ~= 0 and: [lilliputian not]) ifTrue:
		[self 
			storePointer: self freeChunkPrevIndex 
			ofFreeChunk: nextFreeChunk 
			withValue: freeChunk]
	
	! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 12/15/2014 12:15'!
setObjectFree: objOop
	"Mark an object free, but do not add it to the free lists.  The wrinkle here
	 is that we don't tolerate a zero-slot count in a free object so that the
	 (self long64At: objOop) ~= 0 assert in isEnumerableObject: isn't triggered."
		 
	(self rawNumSlotsOf: objOop) = 0 ifTrue:
		[self rawNumSlotsOf: objOop put: 1].
	self setFree: objOop! !

!SpurMemoryManager methodsFor: 'word size' stamp: 'eem 9/30/2013 10:30'!
shiftForAllocationUnit
	^3! !

!SpurMemoryManager methodsFor: 'word size' stamp: 'eem 11/18/2013 15:50'!
shiftForWord
	<api>
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
shortPrintObjectsFrom: startAddress to: endAddress
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| oop |
	oop := self objectBefore: startAddress.
	oop := oop
				ifNil: [startAddress]
				ifNotNil: [(self objectAfter: oop) = startAddress
							ifTrue: [startAddress]
							ifFalse: [oop]].
	[self oop: oop isLessThan: endAddress] whileTrue:
		[(self isFreeObject: oop) ifFalse:
			[coInterpreter shortPrintOop: oop].
		oop := self objectAfter: oop]! !

!SpurMemoryManager methodsFor: 'indexing primitive support' stamp: 'eem 6/22/2021 15:23'!
shorten: objOop toIndexableSize: indexableSize
	"Reduce the number of indexable fields in objOop, an arrayFormat or longFormat object, to nSlots.
	 Convert the unused residual to a free chunk (if in oldSpace).
	 Answer the number of bytes returned to free memory, which may be zero."
	<api>
	| delta |
	self assert: (indexableSize >= 0 and: [indexableSize < (self lengthOf: objOop)]).
	false ifTrue: [self runLeakCheckerFor: GCCheckShorten]. "assume no leaks before hand..."
	delta := self doShorten: objOop toIndexableSize: indexableSize.
	self assert: (self lengthOf: (self followMaybeForwarded: objOop)) = indexableSize.
	self cCode: [] inSmalltalk:
		[(checkForLeaks anyMask: GCCheckShorten) ifTrue:
			[coInterpreter cr; print: 'leak checking shorten...'; flush]].
	self runLeakCheckerFor: GCCheckShorten.
	^delta

	"coInterpreter printOop: objOop"
	"{ objOop. self objectAfter: objOop } collect: [:ea| ea hex]"
	"coInterpreter printOop: (self objectAfter: objOop)"! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 6/5/2017 11:22'!
shouldBreakForLookupIn: lookupClass given: breakClassTag
	<inline: true>
	^breakClassTag notNil
	  and: [lookupClass == breakClassTag
			or: [(self rawHashBitsOf: lookupClass) == breakClassTag]]! !

!SpurMemoryManager methodsFor: 'gc - scavenge/compact' stamp: 'eem 12/24/2016 14:32'!
shouldRemapObj: objOop
	"Answer if the obj should be scavenged (or simply followed). The method is called
	 shouldRemapObj: for compatibility with ObjectMemory.  Defer to the compactor
	 to choose the actual test, there being a difference between the vanilla test and
	 that used with a sliding compactor where objects are not marked as forwarded."
	<doNotGenerate>
	^compactor shouldRemapObj: objOop! !

!SpurMemoryManager methodsFor: 'gc - scavenge/compact' stamp: 'eem 9/15/2013 22:46'!
shouldRemapOop: oop
	<api>
	"Answer if the oop should be scavenged.. The method is called
	 shouldRemapOop: for compatibility with ObjectMemory."
	<inline: true>
	^(self isNonImmediate: oop)
	   and: [self shouldRemapObj: oop]! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 15:40'!
showDisplayBits: aForm Left: l Top: t Right: r Bottom: b
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter showDisplayBits: aForm Left: l Top: t Right: r Bottom: b! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 10/13/2013 12:53'!
shrinkThreshold
	^shrinkThreshold! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 10/14/2013 00:57'!
shrinkThreshold: aValue
	shrinkThreshold := aValue! !

!SpurMemoryManager methodsFor: 'gc - incremental' stamp: 'eem 4/15/2021 11:46'!
shutDownGlobalIncrementalGC: objectsShouldBeUnmarked
	"If the incremental collector is running mark bits may be set; stop it and clear them if necessary."
	self flag: 'need to implement the global inc GC first...'.
	objectsShouldBeUnmarked ifTrue:
		[self assert: self allObjectsUnmarked]! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 9/10/2013 15:06'!
signalLowSpace
	^signalLowSpace! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/12/2013 10:23'!
signalLowSpace: aValue
	^signalLowSpace := aValue! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 12/30/2016 14:59'!
signalSemaphoreWithIndex: index
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter signalSemaphoreWithIndex: index! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 1/24/2014 12:41'!
signed32BitIntegerFor: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter signed32BitIntegerFor: integerValue! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 2/18/2014 15:05'!
signed32BitValueOf: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter signed32BitValueOf: oop! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 12/21/2017 20:39'!
signed64BitIntegerFor: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter signed64BitIntegerFor: integerValue! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 12/21/2017 15:17'!
signed64BitValueOf: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter signed64BitValueOf: oop! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 1/24/2015 16:41'!
signedMachineIntegerValueOf: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter signedMachineIntegerValueOf: oop! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 1/18/2020 14:26'!
singleFloatAtPointer: ptr
	"See platforms/Cross/vm/sqMemoryAccess.h for the production implementation."
	<doNotGenerate>
	^Float fromIEEE32Bit: (self long32At: ptr)! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 1/18/2020 14:26'!
singleFloatAtPointer: ptr put: aFloat
	"See platforms/Cross/vm/sqMemoryAccess.h for the production implementation."
	<doNotGenerate>
	self long32At: ptr put: aFloat asIEEE32BitWord.
	^aFloat! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 12/13/2015 10:21'!
sixtyFourBitIndexableFormat
	<api>
	<cmacro>
	^9! !

!SpurMemoryManager methodsFor: 'class table puns' stamp: 'eem 3/9/2016 14:49'!
sixtyFourBitLongsClassIndexPun
	"Class puns are class indices not used by any class.  There may be
	 an entry for the pun that refers to the notional class of objects with
	 this class index.  But because the index doesn't match the class it
	 won't show up in allInstances, hence hiding the object with a pun as
	 its class index. The puns occupy indices 16 through 31."
	<cmacro>
	^19! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 9/3/2013 14:35'!
sizeBitsOf: objOop
	"Answer the number of bytes in the given object, including its base header, rounded up to an integral number of words.
	 Hence, were it not for the fact that zero-sized objects have at least room for a forwarding pointer,
	 objOop + (self sizeBitsOf: objOop) is the address immediately following objOop."
	"Note: byte indexable objects need to have low bits subtracted from this size to find the address beyond the last byte."
	^(self numSlotsOf: objOop) << self shiftForWord + self baseHeaderSize! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 9/9/2013 14:34'!
sizeBitsOfSafe: objOop
	^self sizeBitsOf: objOop! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 10/27/2020 21:45'!
sizeFieldOfAlien: alienObj
	"Answer the first field of alienObj which is assumed to be an Alien of at least 8/16 bytes, depending on word size"
	<inline: #always>
	^self longAt: alienObj + self baseHeaderSize! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
sizeOfFree: objOop
	"For compatibility with ObjectMemory, answer the size of a free chunk in bytes.
	 Do *not* use internally."
	self assert: (self isFreeObject: objOop).
	^self bytesInBody: objOop! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
sizeOfLargestFreeChunk
	"Answer the size of largest free chunk in oldSpace."
	| freeChunk |
	freeChunk := self findLargestFreeChunk.
	freeChunk ifNil:
		[63 to: 1 by: -1 do:
			[:i|
			(freeLists at: i) ifNotNil:
				[:chunk| ^self bytesInBody: chunk]].
		 ^0].
	^self bytesInBody: freeChunk! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 7/9/2014 11:26'!
sizeOfObjStack: objStack
	| total objStackPage |
	objStack = nilObj ifTrue: [^0].
	total := self fetchPointer: ObjStackTopx ofObject: objStack.
	objStackPage := objStack.
	[objStackPage := self fetchPointer: ObjStackNextx ofObject: objStackPage.
	 objStackPage ~= 0] whileTrue:
		[total := total + ObjStackLimit.
		 self assert: (self fetchPointer: ObjStackTopx ofObject: objStackPage) = ObjStackLimit].
	^total! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 2/21/2018 15:19'!
sizeOfSTArrayFromCPrimitive: cPtr
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter sizeOfSTArrayFromCPrimitive: cPtr! !

!SpurMemoryManager methodsFor: 'testing' stamp: 'eem 12/24/2016 13:59'!
slidingCompactionInProgress
	^gcPhaseInProgress == SlidingCompactionInProgress! !

!SpurMemoryManager methodsFor: 'gc - scavenge/compact' stamp: 'eem 12/24/2016 14:32'!
slidingCompactionRemapObj: objOop
	"Scavenge or simply follow objOop.  Answer the new location of objOop.
	 The send should have been guarded by a send of shouldRemapOop:.
	 The method is called remapObj: for compatibility with ObjectMemory."
	<inline: true>
	| resolvedObj |
	self assert: (self shouldRemapOop: objOop).
	(self isForwarded: objOop)
		ifTrue:
			[resolvedObj := self followForwarded: objOop]
		ifFalse:
			[self deny: (self isInFutureSpace: objOop).
			 resolvedObj := objOop].
	gcPhaseInProgress > 0 ifTrue:
		[self scavengeInProgress
			ifTrue:
				[((self isReallyYoung: resolvedObj) "don't scavenge immediate, old, or CogMethod objects."
				  and: [(self isInFutureSpace: resolvedObj) not]) ifTrue: 
					[^scavenger copyAndForward: resolvedObj]]
			ifFalse:
				[self assert: self slidingCompactionInProgress.
				 (compactor isMobile: objOop) ifTrue:
					[^self fetchPointer: 0 ofObject: objOop]]].
	^resolvedObj! !

!SpurMemoryManager methodsFor: 'gc - scavenge/compact' stamp: 'eem 12/26/2016 17:54'!
slidingCompactionShouldRemapObj: objOop
	<inline: true>
	"Answer if the obj should be scavenged, or simply followed. Sent via the compactor
	 from shouldRemapObj:.  We test for being already scavenged because mapStackPages
	 via mapInterpreterOops may be applied twice in the context of a global GC where a
	 scavenge, followed by a scan-mark-free, and final compaction passes may result in
	 scavenged fields being visited twice."
	^(self isForwarded: objOop)
	   or: [gcPhaseInProgress > 0 "Hence either scavengeInProgress or slidingCompactionInProgress"
		   and: [self scavengeInProgress
					ifTrue: [(self isReallyYoungObject: objOop)
							and: [(self isInFutureSpace: objOop) not]]
					ifFalse: [compactor isMobile: objOop]]]! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 5/27/2021 20:16'!
slotSizeOf: oop
	"*DO NOT CONFUSE THIS WITH numSlotsOf:.
	 This is an ObjectMemory compatibility method with questionable semantics.
	 Answers the number of slots in the receiver.
	 If the receiver is a byte object, return the number of bytes.
	 If the receiver is a short object, return the number of shorts.
	 If the receiver is a word object, return the number of words.
	 If the receiver is a double word object, return the number of double words.
	 Otherwise return the number of pointers."
	(self isImmediate: oop) ifTrue: [^0].
	^self lengthOf: oop! !

!SpurMemoryManager methodsFor: 'cog jit support' stamp: 'eem 2/23/2017 16:14'!
smallIntegerTag
	<api>
	<cmacro>
	^1! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 6/27/2014 15:28'!
smallObjectBytesForSlots: numSlots
	"Answer the total number of bytes in an object without an overflow header, including header bytes."
	<returnTypeC: #usqInt>
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 8/31/2013 12:55'!
specialObjectsOop
	^specialObjectsOop! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 8/31/2013 12:56'!
specialObjectsOop: anObject
	"For mapInterpreterOops"
	specialObjectsOop := anObject! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 9/2/2013 12:49'!
splObj: index
	<api>
	<inline: true>
	"Return one of the objects in the specialObjectsArray"
	^self fetchPointer: index ofObject: specialObjectsOop! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 9/19/2013 14:36'!
splObj: index put: anObject
	"Set one of the objects in the SpecialObjectsArray"
	self storePointer: index ofObject: specialObjectsOop withValue: anObject! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 06/22/2021 15:31'!
sqAllocateMemorySegmentOfSize: segmentSize Above: minAddress AllocatedSizeInto: allocSizePtrOrBlock
	<doNotGenerate>
	"Simulate heap growth by growing memory by segmentSize + a delta.
	 To test bridges alternate the delta between 0 bytes and 1M bytes
	 depending on the number of segments.
	 The delta will be the distance between segments to be bridged."
	| delta newMemory start |
	self assert: segmentSize \\ memory bytesPerElement = 0.
	delta := segmentManager numSegments odd ifTrue: [1024 * 1024] ifFalse: [0].
	"A previous shrink may have freed up memory.  Don't bother to grow if there's already room.
	 At minAddress there is a hole of segmentSize or it is the segLimit of the last segment.
	 However there is no hole of segmentSize + delta guaranteed..."
	0 to: segmentManager numSegments - 1 do:
			[:i| | segment bridge |
			segment := segmentManager segments at: i.
			bridge := segmentManager bridgeAt: i.
			(segment segLimit >= minAddress 
				and: [(self bytesInBody: bridge) - self bridgeSize >= (segmentSize + delta)]) ifTrue: [ 
					allocSizePtrOrBlock value: segmentSize.
					^ segment segLimit + delta] ].
	start := memory size * memory bytesPerElement + delta.
	newMemory := memory class new: memory size + (segmentSize + delta / memory bytesPerElement).
	newMemory replaceFrom: 1 to: memory size with: memory startingAt: 1.
	memory := newMemory.
	allocSizePtrOrBlock value: segmentSize.
	^start! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 12/20/2016 16:06'!
sqDeallocateMemorySegmentAt: startAddress OfSize: ammount
	"This is a nop in the simulator, except for SpurPlanningCompactor which may
	 release at the end of memory, allowing the simulator to shrink memory."
	<doNotGenerate>
	startAddress >= endOfMemory ifTrue:
		[self halt]! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/23/2013 14:24'!
stObject: objOop at: indexOop put: valueOop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stObject: objOop at: indexOop put: valueOop! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 3/23/2016 11:08'!
stSizeOf: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stSizeOf: oop! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/24/2013 13:51'!
stackFloatValue: offset
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stackFloatValue: offset! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:07'!
stackIntegerValue: offset
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stackIntegerValue: offset! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:05'!
stackObjectValue: offset
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stackObjectValue: offset! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 10/27/2020 20:51'!
stackPositiveMachineIntegerValue: offset
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stackPositiveMachineIntegerValue: offset! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/19/2013 11:06'!
stackValue: offset
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stackValue: offset! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 10/18/2013 11:13'!
stackValue: offset ofObjStack: objStackPage
	| topx nextPage |
	self assert: offset >= 0.
	topx := self fetchPointer: ObjStackTopx ofObject: objStackPage.
	offset < topx ifTrue:
		[^self fetchPointer: ObjStackTopx + offset ofObject: objStackPage].
	nextPage := self fetchPointer: ObjStackNextx ofObject: objStackPage.
	nextPage = 0 ifTrue:
		[^nil].
	^self stackValue: offset - topx ofObjStack: nextPage! !

!SpurMemoryManager methodsFor: 'object enumeration-private' stamp: 'eem 9/24/2020 17:04'!
startAddressForBridgedHeapEnumeration
	"Answer the address to start enumeration at if using the bridged enumerator.
	 This is not suitable for the enumerators which enumerate old space before new space.
	 The complication here is that either or both of pastSpace and eden may be empty."
	<inline: #always>
	^pastSpaceStart > scavenger pastSpace start
		ifTrue: [scavenger pastSpace start]
		ifFalse:
			[freeStart > scavenger eden start
				ifTrue: [scavenger eden start]
				ifFalse: [oldSpaceStart]]! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 12/24/2013 20:51'!
startOfMemory
	"Return the start of object memory. Use a macro so as not to punish the debug VM."
	<cmacro: '() GIV(memory)'>
	<returnTypeC: #usqInt>
	^0! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 1/25/2017 09:31'!
startOfObject: objOop
	"Answer the start of objOop, which is either the address of the overflow
	 size word, or objOop itself, depending on the size of the object.  This may
	 be applied to any kind of object, normal, forwarders or free chunks."
	<returnTypeC: #usqInt>
	^(self hasOverflowHeader: objOop)
		ifTrue: [objOop - self baseHeaderSize]
		ifFalse: [objOop]! !

!SpurMemoryManager methodsFor: 'object enumeration' stamp: 'eem 1/25/2017 09:32'!
startOfObject: objOop given: rawNumSlots
	"Answer the start of objOop, which is either the address of the overflow
	 size word, or objOop itself, depending on the size of the object.  This may
	 be applied to any kind of object, normal, forwarders or free chunks."
	<returnTypeC: #usqInt>
	^(self objectWithRawSlotsHasOverflowHeader: rawNumSlots)
		ifTrue: [objOop - self baseHeaderSize]
		ifFalse: [objOop]! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/25/2013 17:05'!
statCompMoveCount
	"Spur never compacts by moving; but it does make compaction passes."
	^statCompactPassCount! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 12/16/2016 16:28'!
statCompactPassCount
	^statCompactPassCount! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 12/16/2016 16:28'!
statCompactPassCount: anInteger
	^statCompactPassCount := anInteger! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 2/6/2017 09:12'!
statCompactionUsecs
	^statCompactionUsecs! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 19:05'!
statFGCDeltaUsecs
	^statFGCDeltaUsecs! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/12/2013 16:17'!
statFullGCUsecs
	^statFullGCUsecs! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/12/2013 16:16'!
statFullGCs
	^statFullGCs! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 19:05'!
statGCEndUsecs
	^statGCEndUsecs! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 19:05'!
statGrowMemory
	^statGrowMemory! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/12/2013 16:18'!
statIGCDeltaUsecs
	^statIGCDeltaUsecs! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/12/2013 16:17'!
statIncrGCUsecs
	^statIncrGCUsecs! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/12/2013 16:18'!
statIncrGCs
	^statIncrGCs! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 19:05'!
statMarkCount
	^statMarkCount! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'cb 7/3/2018 10:17'!
statMarkUsecs
	^statMarkUsecs! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'cb 10/4/2018 08:42'!
statMaxAllocSegmentTime
	^statMaxAllocSegmentTime! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'cb 10/4/2018 08:42'!
statMaxAllocSegmentTime: aValue
	statMaxAllocSegmentTime := aValue! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/25/2013 17:06'!
statMkFwdCount
	^0! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 3/14/2017 13:58'!
statNumGCs
	"Part of InterpreterProxy's 1.14 API"
	<export: true>
	^statScavenges + statIncrGCs + statFullGCs! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 18:51'!
statRootTableCount
	^statRootTableCount! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 19:05'!
statRootTableOverflows
	^statRootTableOverflows! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 19:05'!
statSGCDeltaUsecs
	^statSGCDeltaUsecs! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 19:05'!
statScavengeGCUsecs
	^statScavengeGCUsecs! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 9/24/2013 16:53'!
statScavenges
	^statScavenges! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 19:05'!
statShrinkMemory
	^statShrinkMemory! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 1/25/2016 21:23'!
statSpecialMarkCount
	^0! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 1/25/2016 20:57'!
statSurvivorCount
	<doNotGenerate>
	^scavenger statSurvivorCount! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/25/2013 16:58'!
statSweepCount
	^0! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'cb 7/3/2018 10:17'!
statSweepUsecs
	^statSweepUsecs! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/25/2013 16:39'!
statTenures
	<doNotGenerate>
	^scavenger statTenures! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 6/15/2017 14:02'!
statTenures: aValue
	<doNotGenerate>
	scavenger statTenures: aValue! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 9/22/2013 15:50'!
storeByte: byteIndex ofObject: oop withValue: valueByte
	^self byteAt: oop + self baseHeaderSize + byteIndex put: valueByte! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 11/22/2013 18:08'!
storeCheckBoundary
	"A renaming for the Cogit, which couldn't make sense of GIV(newSpaceLimit)"
	<api>
	^newSpaceLimit! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 1/18/2020 14:19'!
storeFloat32: wordIndex ofObject: objOop withValue: aFloat
	<returnTypeC: #float>
	<var: 'aFloat' type: #float>
	^self singleFloatAtPointer: objOop + self baseHeaderSize + (wordIndex << 2) put: aFloat! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 1/18/2020 14:15'!
storeFloat64: longIndex ofObject: objOop withValue: aDouble
	<returnTypeC: #double>
	<var: 'aDouble' type: #double>
	^self floatAtPointer: objOop + self baseHeaderSize + (longIndex << 3) put: aDouble! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 6/21/2021 12:01'!
storeImageSegmentInto: segmentWordArrayArg outPointers: outPointerArrayArg roots: arrayOfRootsArg
	"This primitive is called from Squeak as...
		<imageSegment> storeSegmentFor: arrayOfRoots into: aWordArray outPointers: anArray.

	 This primitive will store a binary image segment (in the same format as objects in the heap) of the
	 set of objects in arrayOfObjects.  All pointers from within the set to objects outside the set will be
	 copied into the array of outPointers.  In their place in the image segment will be an oop equal to the
	 offset in the outPointer array (the first would be 8), but with the high bit set.

	 Since Spur has a class table the load primitive must insert classes that have instances into the
	 class table.  This primitive marks such classes using the isRemembered bit, which isn't meaningful
	 as a remembered bit in the segment.

	 The primitive expects the segmentWordArray and outPointerArray to be more than adequately long.
	 In this case it returns normally, and truncates the two arrays to exactly the right size.

	 The primitive can fail for the following reasons with the specified failure codes:
		PrimErrGenericError:		the segmentWordArray is too small for the version stamp
		PrimErrWritePastObject:	the segmentWordArray is too small to contain the reachable objects
		PrimErrBadIndex:			the outPointerArray is too small
		PrimErrNoMemory:			additional allocations failed
		PrimErrLimitExceeded:		there is no room in the hash field to store out pointer indices or class references."
	<inline: false>
	| segmentWordArray outPointerArray arrayOfRoots
	  arrayOfObjects savedFirstFields savedOutHashes segStart segAddr endSeg outIndex numClassesInSegment |
	<var: 'segAddr' type: #usqInt>
	((self isObjImmutable: segmentWordArrayArg)
	 or: [self isObjImmutable: outPointerArrayArg]) ifTrue:
		[^PrimErrNoModification].
	"Since segmentWordArrayArg & outPointerArrayArg may get shortened, they can't be pinned."
	((self isPinned: segmentWordArrayArg)
	 or: [self isPinned: outPointerArrayArg]) ifTrue:
		[^PrimErrObjectIsPinned].
	(self numSlotsOf: outPointerArrayArg) > self maxIdentityHash ifTrue:
		[^PrimErrLimitExceeded].

	self runLeakCheckerFor: GCCheckImageSegment.

	"First scavenge to collect any new space garbage that refers to the graph."
	self scavengingGC.
	segmentWordArray := self updatePostScavenge: segmentWordArrayArg.
	outPointerArray := self updatePostScavenge: outPointerArrayArg.
	arrayOfRoots := self updatePostScavenge: arrayOfRootsArg.
	self deny: (self forwardersIn: outPointerArray).
	self deny: (self forwardersIn: arrayOfRoots).
	
	"Now compute the transitive closure, collecting the sequence of objects to be stored in the arrayOfObjects array.
	 Included in arrayOfObjects are the arrayOfRoots and all its contents.  All objects have been unmarked."
	arrayOfObjects := self objectsReachableFromRoots: arrayOfRoots.
	arrayOfObjects ifNil:
		[^PrimErrNoMemory].
	"If objectsReachableFromRoots: answers an integer there is not enough continuous free space in which to allocate the
	 reachable objects.  If there is sufficient free space then answer an error code to prompt a compacting GC and a retry."
	(self isIntegerObject: arrayOfObjects) ifTrue:
		[totalFreeOldSpace - self allocationUnit >= (self integerValueOf: arrayOfObjects) ifTrue:
			[^PrimErrNeedCompaction].
		 ^PrimErrNoMemory].

	self assert: self allObjectsUnmarked. "work to be done when the incremental GC is written"
	self deny: (self forwardersIn: arrayOfObjects).

	"Both to expand the max size of segment and to reduce the length of the
	 load-time pass that adds classes to the class table, move classes to the
	 front of arrayOfObjects, leaving the root array as the first element."
	numClassesInSegment := self moveClassesForwardsIn: arrayOfObjects.

	"The scheme is to copy the objects into segmentWordArray, and then map the oops in segmentWordArray.
	 Therefore the primitive needs to both map efficiently originals to copies in segmentWordArray and
	 be able to undo any side-effects if the primitive fails because either segmentWordArray or outPointerArray
	 is too small.  The mapping is done by having the objects to be stored in arrayOfObjects refer to their mapped
	 locations through their first field, just like a forwarding pointer, but without becoming a forwarder, saving their
	 first field in savedFirstFields, and the objects in outPointerArray pointing to their locations in the outPointerArray
	 through their identityHashes, saved in savedOutHashes.
	 Since arrayOfObjects and its savedFirstFields, and outPointerArray and its saved hashes, can be enumerated
	 side-by-side, the hashes can be restored to the originals.  So the first field of the heap object corresponding to
	 an object in arrayOfObjects is set to its location in segmentWordArray, and the hash of an object in outPointerArray
	 is set to its index in outPointerArray plus the top hash bit.  Classes in arrayOfObjects have their marked bit set.
	 Oops in objects in segmentWordArray are therefore mapped by accessing the original oop, and following its first
	 field. Class indices in segmentWordArray are mapped by fetching the original class, and testing its marked bit.
	 If marked, the first field is followed to access the class copy in the segment.  Out pointers (objects and classes,
	 which are unmarked), the object's identityHash is set (eek!!!!) to its index in the outPointerArray. So savedOutHashes
	 parallels the outPointerArray. The saved hash array is initialized with an out-of-range hash value so that the first
	 unused entry can be identified."

	savedFirstFields := self allocateSlots: (self numSlotsOf: arrayOfObjects)
							format: self wordIndexableFormat
							classIndex: self wordSizeClassIndexPun.
	savedOutHashes := self allocateSlots: (self numSlotsForBytes: (self numSlotsOf: outPointerArray) * 4)
							format: self firstLongFormat
							classIndex: self thirtyTwoBitLongsClassIndexPun.
	(savedFirstFields isNil or: [savedOutHashes isNil]) ifTrue:
		[self freeObject: arrayOfObjects.
		 (savedFirstFields notNil and: [self isInOldSpace: savedFirstFields]) ifTrue:
			[self freeObject: savedFirstFields].
		 (savedOutHashes notNil and: [self isInOldSpace: savedOutHashes]) ifTrue:
			[self freeObject: savedOutHashes].
		 ^PrimErrNoMemory].

	self fillObj: savedFirstFields numSlots: (self numSlotsOf: savedFirstFields) with: 0.
	self fillObj: savedOutHashes numSlots: (self numSlotsOf: savedOutHashes) with: self savedOutHashFillValue.

	segAddr := segmentWordArray + self baseHeaderSize.
	endSeg := self addressAfter: segmentWordArray.

	"Write a version number for byte order and version check."
	segAddr >= endSeg ifTrue: [^PrimErrGenericFailure].
	self long32At: segAddr put: self imageSegmentVersion.
	self long32At: segAddr + 4 put: self imageSegmentVersion.
	segStart := segAddr := segAddr + self allocationUnit.

	self assert: arrayOfRoots = (self fetchPointer: 0 ofObject: arrayOfObjects).

	"Copy all reachable objects to the segment, setting the marked bit for all objects (clones) in the segment,
	 and the remembered bit for all classes (clones) in the segment."
	0 to: (self numSlotsOf: arrayOfObjects) - 1 do:
		[:i| | newSegAddrOrError objOop |
		"Check that classes in the segment are addressable.  Since the top bit of the hash field is used to tag
		 classes external to the segment, the segment offset must not inadvertently set this bit.  This limit still
		 allows for a million or more classes."
		(i = numClassesInSegment
		 and: [segAddr - segStart / self allocationUnit + self lastClassIndexPun >= TopHashBit]) ifTrue:
			[^self return: PrimErrLimitExceeded
					restoringObjectsIn: arrayOfObjects upTo: i savedFirstFields: savedFirstFields].
		objOop := self fetchPointer: i ofObject: arrayOfObjects.
		self deny: ((self isImmediate: objOop) or: [self isForwarded: objOop]).
		newSegAddrOrError := self copyObj: objOop
									toAddr: segAddr
									stopAt: endSeg
									savedFirstFields: savedFirstFields
									index: i.
		(self oop: newSegAddrOrError isLessThan: segStart) ifTrue:
			[^self return: newSegAddrOrError
					restoringObjectsIn: arrayOfObjects upTo: i savedFirstFields: savedFirstFields].
		 segAddr := newSegAddrOrError].

	"Check that it can be safely shortened."
	(endSeg ~= segAddr
	 and: [endSeg - segAddr < (self baseHeaderSize + self bytesPerOop)]) ifTrue:
		[^self return: PrimErrWritePastObject
				restoringObjectsIn: arrayOfObjects upTo: -1 savedFirstFields: savedFirstFields].

	"Now scan, adding out pointers to the outPointersArray; all objects in arrayOfObjects
	 have their first field pointing to the corresponding copy in segmentWordArray."
	(outIndex := self mapOopsFrom: segStart
					to: segAddr
					outPointers: outPointerArray
					outHashes: savedOutHashes) < 0 ifTrue: "no room in outPointers; fail"
		[^self return: PrimErrBadIndex
				restoringObjectsIn: arrayOfObjects savedFirstFields: savedFirstFields
				and: outPointerArray savedHashes: savedOutHashes].

	"We're done.  Shorten the results, restore hashes and return."
	self shorten: segmentWordArray toIndexableSize: segAddr - (segmentWordArray + self baseHeaderSize) / 4.
	self shorten: outPointerArray toIndexableSize: outIndex.
	^self return: PrimNoErr
		restoringObjectsIn: arrayOfObjects savedFirstFields: savedFirstFields
		and: outPointerArray savedHashes: savedOutHashes! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/21/2013 07:15'!
storeInteger: fieldIndex ofObject: objectPointer withValue: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter storeInteger: fieldIndex ofObject: objectPointer withValue: integerValue! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 8/31/2013 08:59'!
storeLong32: fieldIndex ofObject: obj withValue: valueWord
	^self long32At: obj + self baseHeaderSize + (fieldIndex << 2) put: valueWord! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 7/29/2014 08:11'!
storeLong64: longIndex ofObject: objOop withValue: value
	<var: #value type: #sqLong>
	^self long64At: objOop + self baseHeaderSize + (longIndex << 3) put: value! !

!SpurMemoryManager methodsFor: 'heap management' stamp: 'eem 11/20/2013 13:56'!
storePointer: fieldIndex ofForwarder: objOop withValue: valuePointer

	self assert: (self isForwarded: objOop).
	self assert: (self isOopForwarded: valuePointer) not.

	(self isOldObject: objOop) ifTrue: "most stores into young objects"
		[(self isYoung: valuePointer) ifTrue:
			[self possibleRootStoreInto: objOop]].

	^self
		longAt: objOop + self baseHeaderSize + (fieldIndex << self shiftForWord)
		put: valuePointer! !

!SpurMemoryManager methodsFor: 'heap management' stamp: 'eem 4/22/2014 18:37'!
storePointer: fieldIndex ofFreeChunk: objOop withUncheckedValue: valuePointer

	self assert: (self isFreeObject: objOop).
	^self
		longAt: objOop + self baseHeaderSize + (fieldIndex << self shiftForWord)
		put: valuePointer! !

!SpurMemoryManager methodsFor: 'heap management' stamp: 'cb 6/8/2018 11:27'!
storePointer: fieldIndex ofFreeChunk: objOop withValue: valuePointer

	self assert: (self isFreeObject: objOop).
	self assert: (valuePointer = 0 or: [self isFreeObject: valuePointer]).		
	^self
		longAt: objOop + self baseHeaderSize + (fieldIndex << self shiftForWord)
		put: valuePointer! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 1/27/2020 09:07'!
storePointer: fieldIndex ofObjStack: objStackPage withValue: thang
	self assert: (self formatOf: objStackPage) = self wordIndexableFormat.
	self cCode: ''
		inSmalltalk:
			[fieldIndex caseOf: {
				[ObjStackTopx]		->	[self assert: (thang between: 0 and: ObjStackLimit)].
				[ObjStackMyx]		->	[self assert: (thang between: MarkStackRootIndex and: MournQueueRootIndex)].
				[ObjStackFreex]	->	[self assert: (thang = 0
														or: [(self addressCouldBeObj: thang)
															and: [(self numSlotsOfAny: thang) = ObjStackPageSlots
															and: [(self formatOf: thang) = self wordIndexableFormat]]])].
				[ObjStackNextx]	->	[self assert: (thang = 0
														or: [(self addressCouldBeObj: thang)
															and: [(self numSlotsOfAny: thang) = ObjStackPageSlots
															and: [(self formatOf: thang) = self wordIndexableFormat]]])]. }
				otherwise: []].
	^self
		longAt: objStackPage + self baseHeaderSize + (fieldIndex << self shiftForWord)
		put: thang! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 11/20/2013 13:56'!
storePointer: fieldIndex ofObject: objOop withValue: valuePointer
	"Note must check here for stores of young objects into old ones."
	self assert: (self isForwarded: objOop) not.

	(self isOldObject: objOop) ifTrue: "most stores into young objects"
		[(self isYoung: valuePointer) ifTrue:
			[self possibleRootStoreInto: objOop]].

	^self
		longAt: objOop + self baseHeaderSize + (fieldIndex << self shiftForWord)
		put: valuePointer! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 5/11/2017 10:18'!
storePointerImmutabilityCheck: fieldIndex ofObject: objOop withValue: valuePointer
	"Note must check here for stores of young objects into old ones."
	<inline: true> "Must be inlined for the normal send in cannotAssign:to:withIndex:"

	self cppIf: IMMUTABILITY ifTrue: 
		[self deny: (self isImmediate: objOop).
		 (self isImmutable: objOop) ifTrue: 
			[^coInterpreter cannotAssign: valuePointer to: objOop withIndex: fieldIndex]].

	self storePointer: fieldIndex ofObject: objOop withValue: valuePointer! !

!SpurMemoryManager methodsFor: 'heap management' stamp: 'eem 11/9/2013 10:37'!
storePointerNoAssert: fieldIndex ofFreeChunk: objOop withValue: valuePointer

	^self
		longAt: objOop + self baseHeaderSize + (fieldIndex << self shiftForWord)
		put: valuePointer! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 9/22/2013 15:23'!
storePointerUnchecked: fieldIndex ofMaybeForwardedObject: objOop withValue: valuePointer
	^self
		longAt: objOop + self baseHeaderSize + (fieldIndex << self shiftForWord)
		put: valuePointer! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 8/15/2016 17:11'!
storePointerUnchecked: fieldIndex ofObject: objOop withValue: valuePointer
	<api>
	self assert: (self isOopForwarded: objOop) not.
	^self
		longAt: objOop + self baseHeaderSize + (fieldIndex << self shiftForWord)
		put: valuePointer! !

!SpurMemoryManager methodsFor: 'object access' stamp: 'eem 10/29/2014 11:20'!
storeShort16: shortIndex ofObject: objOop withValue: value
	^self
		shortAt: objOop + self baseHeaderSize + (shortIndex << 1)
		put: (self cCode: [value] inSmalltalk: [value bitAnd: 16rFFFF])! !

!SpurMemoryManager methodsFor: 'primitive support' stamp: 'eem 9/27/2020 14:13'!
stringForCString: aCString
	"Answer a new String copied from a null-terminated C string,
	 or nil if out of memory."
	<api>
	<var: 'aCString' type: #'const char *'>
	<inline: false>
	| len newString |
	len := self strlen: aCString.
	newString := self
					allocateSlots: (self numSlotsForBytes: len)
					format: (self byteFormatForNumBytes: len)
					classIndex: ClassByteStringCompactIndex.
	newString ifNotNil:
		[self strncpy: (self cCoerceSimple: newString + self baseHeaderSize to: #'char *')
			_: aCString
			_: len]. "(char *)strncpy()"
	^newString! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 15:57'!
stringOf: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stringOf: oop! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/19/2013 11:07'!
success: boolean
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter success: boolean! !

!SpurMemoryManager methodsFor: 'gc - scavenging' stamp: 'eem 9/25/2020 09:53'!
sufficientSpaceAfterGC: numBytes
	"This is ObjectMemory's funky entry-point into its incremental GC,
	 which is a stop-the-world a young generation reclaimer.  In Spur
	 we run the scavenger.  Answer if space is not low."

	| heapSizePostGC |
	self assert: numBytes = 0.
	self scavengingGCTenuringIf: TenureByAge.
	heapSizePostGC := segmentManager totalOldSpaceCapacity - totalFreeOldSpace.
	(heapSizePostGC - heapSizeAtPreviousGC) asFloat / heapSizeAtPreviousGC >= heapGrowthToSizeGCRatio
		ifTrue: [self fullGC] "fullGC will attempt to shrink"
		ifFalse: "Also attempt to shrink if there is plenty of free space and no need to GC"
			[totalFreeOldSpace > (shrinkThreshold * 2) ifTrue:
				[self attemptToShrink.
				 ^true]].
	[totalFreeOldSpace < growHeadroom
	 and: [(self growOldSpaceByAtLeast: 0) notNil]] whileTrue:
		[totalFreeOldSpace >= growHeadroom ifTrue:
			[^true]].
	lowSpaceThreshold > totalFreeOldSpace ifTrue: "space is low"
		[lowSpaceThreshold := 0. "avoid signalling low space twice"
		 ^false].
	^true! !

!SpurMemoryManager methodsFor: 'allocation' stamp: 'eem 9/9/2013 10:59'!
sufficientSpaceToInstantiate: classObj indexableSize: indexableFields
	self shouldNotImplement! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 2/3/2015 18:36'!
superclassOf: classPointer
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter superclassOf: classPointer! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 06/22/2021 15:31'!
swizzleFieldsOfFreeChunk: chunk
	<inline: true>
	| field chunkBytes |
	field := self fetchPointer: self freeChunkNextIndex ofFreeChunk: chunk.
	field ~= 0 ifTrue:
		[self storePointerNoAssert: self freeChunkNextIndex
			ofFreeChunk: chunk
			withValue: (segmentManager swizzleObj: field in: chunk)].
	chunkBytes := self bytesInBody: chunk.
	false ifTrue: "The prevPointer is not guaranteed to be valid in older images.
				 updateListStartingAt: via updateFreeLists does restore the prev pointer
				 in all small free lists, so simply avoid swizzling it now."
		[(self isLilliputianSize: chunkBytes) ifFalse:
			[field := self fetchPointer: self freeChunkPrevIndex ofFreeChunk: chunk.
			 field ~= 0 ifTrue:
				[self storePointerNoAssert: self freeChunkPrevIndex
					ofFreeChunk: chunk
					withValue: (segmentManager swizzleObj: field in: chunk)]]].
	chunkBytes >= (self numFreeLists * self allocationUnit) ifTrue:
		[self freeChunkParentIndex to: self freeChunkLargerIndex do:
			[:index|
			 field := self fetchPointer: index ofFreeChunk: chunk.
			 field ~= 0 ifTrue:
				[self storePointerNoAssert: index
					ofFreeChunk: chunk
					withValue: (segmentManager swizzleObj: field in: chunk)]]]! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 11/28/2019 12:37'!
swizzleFieldsOfObject: oop
	| fieldAddr fieldOop |
	<inline: true>
	fieldAddr := oop + (self lastPointerOfWhileSwizzling: oop).
	[self oop: fieldAddr isGreaterThanOrEqualTo: oop + self baseHeaderSize] whileTrue:
		[fieldOop := self longAt: fieldAddr.
		 (self isNonImmediate: fieldOop) ifTrue:
			[self longAt: fieldAddr put: (segmentManager swizzleObj: fieldOop in: oop)].
		 fieldAddr := fieldAddr - self bytesPerOop]! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 11/28/2019 12:38'!
swizzleObjStackAt: objStackRootIndex
	"On load, swizzle the pointers in an obj stack. Answer the obj stack's oop."
	| firstPage page stackOrNil index field |
	firstPage := stackOrNil := self fetchPointer: objStackRootIndex ofObject: hiddenRootsObj.
	stackOrNil = nilObj ifTrue:
		[^stackOrNil].
	[self assert: (self numSlotsOfAny: stackOrNil) = ObjStackPageSlots.
	 self assert: (self fetchPointer: ObjStackMyx ofObject: stackOrNil) = objStackRootIndex.
	 "There are four fixed slots in an obj stack, and a Topx of 0 indicates empty, so
	   if there were 5 slots in an oop stack, full would be 2, and the last 0-rel index is 4.
	   Hence the last index is topx + fixed slots - 1, or topx + ObjStackNextx"
	 index := (self fetchPointer: ObjStackTopx ofObject: stackOrNil) + ObjStackNextx.
	 "swizzle fields including ObjStackNextx, excluding ObjStackFreex and leave field containing the next link."
	 [field := self fetchPointer: index ofObject: stackOrNil.
	  (field = 0 or: [self isImmediate: field]) ifFalse:
		[field := segmentManager swizzleObj: field in: stackOrNil.
		 self storePointer: index ofObjStack: stackOrNil withValue: field].
	  (index := index - 1) >= ObjStackNextx] whileTrue.
	 (stackOrNil := field) ~= 0] whileTrue.
	(stackOrNil := self fetchPointer: ObjStackFreex ofObject: firstPage) ~=  0 ifTrue:
		[page := firstPage.
		 [stackOrNil := segmentManager swizzleObj: stackOrNil in: firstPage.
		  self storePointer: ObjStackFreex ofObjStack: page withValue: stackOrNil.
		  page := stackOrNil.
		  (stackOrNil := self fetchPointer: ObjStackFreex ofObject: page) ~=  0] whileTrue].
	self assert: (self isValidObjStackAt: objStackRootIndex).
	^self fetchPointer: objStackRootIndex ofObject: hiddenRootsObj! !

!SpurMemoryManager methodsFor: 'word size' stamp: 'eem 12/12/2015 11:34'!
tagMask
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'plugin support' stamp: 'eem 4/15/2021 11:51'!
tenuringIncrementalGC
	"Do an incremental GC that tenures all surviving young objects to old space.
	 The selector tenuringIncrementalGC refers to the V3 incremental collector, which is the collector for young objects, equivalent in function to the scavenger.
	 The Spur incremental collector is called the global incremental garbage collector or globalIGC for short."
	<api>
	self flushNewSpace! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 4/29/2014 14:51'!
tenuringThreshold
	"In the scavenger the tenuring threshold is effectively a number of bytes of objects,
	 accessed as a proportion of pastSpace from 0 to 1.   In the Squeak image the tenuring
	 threshold is an object count. Marry the two notions by multiplying the proportion by
	 the size of pastSpace and dividing by the average object size, as derived from observation."
	^(scavenger scavengerTenuringThreshold * scavenger pastSpaceBytes // self averageObjectSizeInBytes) asInteger! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 3/16/2017 13:59'!
tenuringThreshold: threshold
	"c.f. tenuringThreshold"
	threshold < 0 ifTrue:
		[^PrimErrBadArgument].
	scavenger scavengerTenuringThreshold:
		(threshold * self averageObjectSizeInBytes) asFloat
		/ scavenger pastSpaceBytes asFloat.
	^0! !

!SpurMemoryManager methodsFor: 'class table puns' stamp: 'eem 3/9/2016 14:49'!
thirtyTwoBitLongsClassIndexPun
	"Class puns are class indices not used by any class.  There may be
	 an entry for the pun that refers to the notional class of objects with
	 this class index.  But because the index doesn't match the class it
	 won't show up in allInstances, hence hiding the object with a pun as
	 its class index. The puns occupy indices 16 through 31."
	<cmacro>
	^18! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 6/24/2015 11:38'!
topOfObjStack: objStack
	| topx |
	"This assert is tricky.  push:onObjStack: may call topOfObjStack: just after pushing an
	 empty page on the stack, and will ask if the second page is valid."
	self eassert: [self isValidObjStackPage: objStack
					myIndex: (self fetchPointer: ObjStackMyx ofObject: objStack)
					firstPage: (objStack = (self fetchPointer: (self fetchPointer: ObjStackMyx ofObject: objStack) ofObject: hiddenRootsObj))].
	topx := self fetchPointer: ObjStackTopx ofObject: objStack.
	topx = 0 ifTrue:
		[self assert: (self fetchPointer: ObjStackNextx ofObject: objStack) = 0.
		^nil].
	^self fetchPointer: topx + ObjStackFixedSlots - 1 ofObject: objStack! !

!SpurMemoryManager methodsFor: 'interpreter access' stamp: 'eem 9/20/2013 15:17'!
topRemappableOop
	<api>
	"Answers the top of the remappable oop stack. Useful when writing loops.
	 We support this excessence for compatibility with ObjectMemory.
	 Spur doesn't GC during allocation."
	^remapBuffer at: remapBufferCount! !

!SpurMemoryManager methodsFor: 'indexing primitive support' stamp: 'eem 06/22/2021 15:31'!
totalByteSizeOf: oop
	<returnTypeC: #usqLong>
	^(self isImmediate: oop)
		ifTrue: [0]
		ifFalse: [self bytesInBody: oop]! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
totalFreeListBytes
	"This method both computes the actual number of free bytes by traversing all free objects
	 on the free lists/tree, and checks that the tree is valid.  It is used mainly by checkFreeSpace."
	| totalFreeBytes bytesInChunk listNode nextNode |
	totalFreeBytes := 0.
	1 to: self numFreeLists - 1 do:
		[:i| 
		bytesInChunk := i * self allocationUnit.
		listNode := freeLists at: i.
		[listNode ~= 0] whileTrue:
			[totalFreeBytes := totalFreeBytes + bytesInChunk.
			 self assertValidFreeObject: listNode.
			 self assert: bytesInChunk = (self bytesInBody: listNode).
			 nextNode := self fetchPointer: self freeChunkNextIndex ofFreeChunk: listNode.
			 self assert: nextNode ~= listNode.
			 listNode := nextNode]].

	self freeTreeNodesDo:
		[:treeNode|
		 bytesInChunk := self bytesInBody: treeNode.
		 self assert: bytesInChunk / self allocationUnit >= self numFreeLists.
		 listNode := treeNode.
		 [listNode ~= 0] whileTrue:
			["self printFreeChunk: listNode"
			 self assertValidFreeObject: listNode.
			 self assert: (listNode = treeNode
						  or: [(self fetchPointer: self freeChunkParentIndex ofFreeChunk: listNode) = 0]).
			 totalFreeBytes := totalFreeBytes + bytesInChunk.
			 self assert: bytesInChunk = (self bytesInBody: listNode).
			 nextNode := self fetchPointer: self freeChunkNextIndex ofFreeChunk: listNode.
			 self assert: nextNode ~= listNode.
			 listNode := nextNode].
		 treeNode].
	^totalFreeBytes! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'cb 4/19/2018 09:48'!
totalFreeOldSpace
	<doNotGenerate>
	^ totalFreeOldSpace! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 12/16/2016 16:36'!
totalFreeOldSpace: anInteger
	totalFreeOldSpace := anInteger! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 10/14/2013 20:15'!
totalMemorySize
	^scavenger newSpaceCapacity + segmentManager totalBytesInSegments! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'eem 7/20/2014 23:02'!
traceImmediatelySlotLimit
	"Arbitrary level at which to defer tracing large objects until later.
	 The average slot size of Smalltalk objects is typically near 8.
	 We do require traceImmediatelySlotLimit to be < numSlotsMask."
	^64! !

!SpurMemoryManager methodsFor: 'simulation only' stamp: 'eem 9/20/2013 15:13'!
transcript
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter transcript! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 12/16/2015 09:21'!
trueObject
	<api>
	^trueObj! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'eem 8/31/2013 12:54'!
trueObject: anOop
	"For mapInterpreterOops"
	trueObj := anOop! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 12/21/2017 17:41'!
unalignedLong32At: index
	"Support for primitiveFFIIntegerAt[Put]"
	<doNotGenerate>
	| odd hi lo |
	(odd := index bitAnd: 3) = 0 ifTrue:
		[^self long32At: index].
	lo := self long32At: index - odd.
	hi := self long32At: index + 4 - odd.
	^lo >> (odd * 8) + ((hi bitAnd: 1 << (odd * 8) - 1) << (4 - odd * 8))! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 12/21/2017 20:38'!
unalignedLong32At: index put: aValue
	"Support for primitiveFFIIntegerAt[Put]"
	<doNotGenerate>
	| odd hi lo mask |
	aValue < 0 ifTrue:
		[self unalignedLong64At: index put: (aValue bitAnd: 16rFFFFFFFF).
		 ^aValue].
	(odd := index bitAnd: 3) = 0 ifTrue:
		[^self long32At: index put: aValue].
	mask := 1 << (odd * 8) - 1.
	lo := self long32At: index - odd.
	self long32At: index - odd
		put: (lo bitAnd: mask)
			+ ((aValue bitAnd: 1 << (4 - odd * 8) - 1) << (odd * 8)).
	hi := self long32At: index + 4 - odd.
	self long32At: index + 4 - odd
		put: (hi bitClear: mask) + (aValue >> (4 - odd * 8) bitAnd: mask).
	^aValue! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 12/21/2017 17:41'!
unalignedLong64At: index
	"Support for primitiveFFIIntegerAt[Put]"
	<doNotGenerate>
	| odd hi lo |
	(odd := index bitAnd: 7) = 0 ifTrue:
		[^self long64At: index].
	lo := self long64At: index - odd.
	hi := self long64At: index + 8 - odd.
	^lo >> (odd * 8) + ((hi bitAnd: 1 << (odd * 8) - 1) << (8 - odd * 8))! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 12/21/2017 20:37'!
unalignedLong64At: index put: aValue
	"Support for primitiveFFIIntegerAt[Put]"
	<doNotGenerate>
	| odd hi lo mask |
	aValue < 0 ifTrue:
		[self unalignedLong64At: index put: (aValue bitAnd: 16rFFFFFFFFFFFFFFFF).
		 ^aValue].
	(odd := index bitAnd: 7) = 0 ifTrue:
		[^self long64At: index put: aValue].
	mask := 1 << (odd * 8) - 1.
	lo := self long64At: index - odd.
	self long64At: index - odd
		put: (lo bitAnd: mask)
			+ ((aValue bitAnd: 1 << (8 - odd * 8) - 1) << (odd * 8)).
	hi := self long64At: index + 8 - odd.
	self long64At: index + 8 - odd
		put: (hi bitClear: mask) + (aValue >> (8 - odd * 8) bitAnd: mask).
	^aValue! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 12/21/2017 17:32'!
unalignedShortAt: index
	"Support for primitiveFFIIntegerAt[Put]"
	<doNotGenerate>
	| hi lo |
	(index bitAnd: 1) = 0 ifTrue:
		[^self shortAt: index].
	lo := self shortAt: index - 1.
	hi := self shortAt: index + 1.
	^lo >> 8 + ((hi bitAnd: 16rFF) << 8)! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 12/21/2017 17:45'!
unalignedShortAt: index put: aValue
	"Support for primitiveFFIIntegerAt[Put]"
	<doNotGenerate>
	(index bitAnd: 1) = 0 ifTrue:
		[^self shortAt: index put: aValue].
	self shouldBeImplemented.
	^aValue! !

!SpurMemoryManager methodsFor: 'primitive support' stamp: 'eem 11/13/2014 12:15'!
uniqueIndex: classIndex allInstancesInto: start limit: limit resultsInto: binaryBlock
	<inline: true>
	| count ptr |
	count := 0.
	ptr := start.
	self allHeapEntitiesDo:
		[:obj| "continue enumerating even if no room so as to unmark all objects."
		 (MarkObjectsForEnumerationPrimitives
				ifTrue: [self isMarked: obj]
				ifFalse: [true]) ifTrue:
			[(self isNormalObject: obj)
				ifTrue:
					[MarkObjectsForEnumerationPrimitives ifTrue:
						[self setIsMarkedOf: obj to: false].
					 (self classIndexOf: obj) = classIndex ifTrue:
					 	[count := count + 1.
						 ptr < limit ifTrue:
							[self longAt: ptr put: obj.
							 ptr := ptr + self bytesPerOop]]]
				ifFalse:
					[MarkObjectsForEnumerationPrimitives ifTrue:
						[(self isSegmentBridge: obj) ifFalse:
							[self setIsMarkedOf: obj to: false]]]]].
	binaryBlock value: count value: ptr
! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'cb 10/5/2018 12:53'!
unlinkFreeChunk: chunk atIndex: index chunkBytes: chunkBytes
	^self 
		unlinkFreeChunk: chunk 
		atIndex: index 
		isLilliputianSize: (self isLilliputianSize: chunkBytes) ! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
unlinkFreeChunk: chunk atIndex: index isLilliputianSize: lilliputian 
	"Unlink and answer a small chunk from one of the fixed size freeLists"
	<inline: true> "inlining is important because isLilliputianSize: is often true"
	|next|
	self assert: ((self bytesInBody: chunk) = (index * self allocationUnit)
				and: [index > 1 "a.k.a. (self bytesInBody: chunk) > self allocationUnit"
				and: [(self startOfObject: chunk) = chunk]]).
			
	"For some reason the assertion is not compiled correctly"
	self cCode: '' inSmalltalk: [self assert: (self isLilliputianSize: (self bytesInBody: chunk)) = lilliputian].
	
	freeLists
		at: index 
		put: (next := self
				fetchPointer: self freeChunkNextIndex
				ofFreeChunk: chunk).
	(lilliputian not and: [next ~= 0]) ifTrue:
		[self storePointer: self freeChunkPrevIndex ofFreeChunk: next withValue: 0].
	^chunk! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'cb 10/5/2018 12:56'!
unlinkFreeChunk: freeChunk chunkBytes: chunkBytes
	"Unlink a free object from the free lists. Do not alter totalFreeOldSpace. Used for coalescing."
	| index next prev |
	index := chunkBytes / self allocationUnit.
	
	"Pathological 64 bits case - size 1 - single linked list"
	(self isLilliputianSize: chunkBytes) ifTrue:
		[^self unlinkLilliputianChunk: freeChunk index: index].
	
	prev := self fetchPointer: self freeChunkPrevIndex ofFreeChunk: freeChunk.
	"Has prev element: update double linked list"
	prev ~= 0 ifTrue:
		[self 
			setNextFreeChunkOf: prev 
			withValue: (self fetchPointer: self freeChunkNextIndex ofFreeChunk: freeChunk) 
			chunkBytes: chunkBytes.
		 ^freeChunk].
	
	"Is the beginning of a list"
	"Small chunk"
	(index < self numFreeLists and: [1 << index <= freeListsMask]) ifTrue: 
		[self unlinkFreeChunk: freeChunk atIndex: index isLilliputianSize: false.
		 ^freeChunk].
	"Large chunk"
	 next := self fetchPointer: self freeChunkNextIndex ofFreeChunk: freeChunk.
	 next = 0
		ifTrue: "no list; remove the interior node"
			[self unlinkSolitaryFreeTreeNode: freeChunk]
		ifFalse: "list; replace node with it"
			[self inFreeTreeReplace: freeChunk with: next].
	^freeChunk
	
	

	! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'cb 10/5/2018 12:58'!
unlinkLilliputianChunk: freeChunk index: index
	| node prev next |
	<inline: #never> "for profiling"
	 node := freeLists at: index.
	 prev := 0.
	 [node ~= 0] whileTrue:
		[self assert: node = (self startOfObject: node).
		 self assertValidFreeObject: node.
		 next := self fetchPointer: self freeChunkNextIndex ofFreeChunk: node.
		 node = freeChunk ifTrue:
			[prev = 0
				ifTrue: [self unlinkFreeChunk: freeChunk atIndex: index isLilliputianSize: true]
				ifFalse: [self setNextFreeChunkOf: prev withValue: next isLilliputianSize: true].
			 ^freeChunk].
		 prev := node.
		 node := next].
	 self error: 'freeChunk not found in lilliputian chunk free list'
	
	

	! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 6/4/2014 12:28'!
unlinkSolitaryFreeTreeNode: freeTreeNode
	"Unlink a freeTreeNode.  Assumes the node has no list (null next link)."
	| parent smaller larger |
	self assert: (self fetchPointer: self freeChunkNextIndex ofFreeChunk: freeTreeNode) = 0.

	"case 1. interior node has one child, P = parent, N = node, S = subtree (mirrored for large vs small)
			___				  ___
			| P |				  | P |
		    _/_				_/_
		    | N |		=>		| S |
		 _/_
		 | S |

	 case 2: interior node has two children, , P = parent, N = node, L = smaller, left subtree, R = larger, right subtree.
	 add the left subtree to the bottom left of the right subtree (mirrored for large vs small) 
			___				  ___
			| P |				  | P |
		    _/_				_/_
		    | N |		=>		| R |
		 _/_  _\_		    _/_
		 | L | | R |		    | L |"

	smaller := self fetchPointer: self freeChunkSmallerIndex ofFreeChunk: freeTreeNode.
	larger := self fetchPointer: self freeChunkLargerIndex ofFreeChunk: freeTreeNode.
	parent := self fetchPointer: self freeChunkParentIndex ofFreeChunk: freeTreeNode.
	parent = 0
		ifTrue: "no parent; stitch the subnodes back into the root"
			[smaller = 0
				ifTrue:
					[larger ~= 0 ifTrue:
						[self storePointer: self freeChunkParentIndex ofFreeChunk: larger withValue: 0].
					 freeLists at: 0 put: larger]
				ifFalse:
					[self storePointer: self freeChunkParentIndex ofFreeChunk: smaller withValue: 0.
					 freeLists at: 0 put: smaller.
					 larger ~= 0 ifTrue:
						[self addFreeSubTree: larger]]]
		ifFalse: "parent; stitch back into appropriate side of parent."
			[smaller = 0
				ifTrue: [self storePointer: (freeTreeNode = (self fetchPointer: self freeChunkSmallerIndex ofFreeChunk: parent)
											ifTrue: [self freeChunkSmallerIndex]
											ifFalse: [self freeChunkLargerIndex])
							ofFreeChunk: parent
							withValue: larger.
						larger ~= 0 ifTrue:
							[self storePointer: self freeChunkParentIndex
								ofFreeChunk: larger
								withValue: parent]]
				ifFalse:
					[self storePointer: (freeTreeNode = (self fetchPointer: self freeChunkSmallerIndex ofFreeChunk: parent)
											ifTrue: [self freeChunkSmallerIndex]
											ifFalse: [self freeChunkLargerIndex])
						ofFreeChunk: parent
						withValue: smaller.
					 self storePointer: self freeChunkParentIndex
						ofFreeChunk: smaller
						withValue: parent.
					 larger ~= 0 ifTrue:
						[self addFreeSubTree: larger]]]! !

!SpurMemoryManager methodsFor: 'primitive support' stamp: 'eem 3/27/2014 15:34'!
unmarkAllObjects
	self allHeapEntitiesDo:
		[:obj|
		 (self isMarked: obj) ifTrue:
			[(self isNormalObject: obj)
				ifTrue:
					[self setIsMarkedOf: obj to: false]
				ifFalse:
					[(self isSegmentBridge: obj) ifFalse:
						[self setIsMarkedOf: obj to: false]]]].
! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 12/11/2017 14:23'!
unmarkObjectsIn: arrayOfRoots
	"This is part of storeImageSegmentInto:outPointers:roots:."
	0 to: (self numSlotsOf: arrayOfRoots) - 1 do:
		[:i| | oop |
		oop := self followField: i ofObject: arrayOfRoots.
		(self isNonImmediate: oop) ifTrue:
			[self setIsMarkedOf: oop to: false]]! !

!SpurMemoryManager methodsFor: 'compaction' stamp: 'eem 12/24/2016 11:01'!
unmarkSurvivingObjectsForCompact
	self allPastSpaceObjectsDo:
		[:objOop|
		(self isMarked: objOop) ifTrue:
			[self setIsMarkedOf: objOop to: false]]! !

!SpurMemoryManager methodsFor: 'primitive support' stamp: 'eem 7/28/2014 07:14'!
unpinObject: objOop
	self assert: (self isNonImmediate: objOop).
	self setIsPinnedOf: objOop to: false.
	^0! !

!SpurMemoryManager methodsFor: 'initialization' stamp: 'cb 6/8/2018 14:03'!
updateFreeLists
	"Snapshot did not guarantee the state of the freelist prevLink, so we need to update it.
	 Effectively transforms the freechunk single linked list in double linked list."
	|min|
	"Small chunks"
	"Skip in 64 bits size 1 which is single linked list - pathological case"
	self wordSize = 8 ifTrue: [min := 3] ifFalse: [min := 2].
	min to: self numFreeLists - 1 do:
		[:i| self updateListStartingAt: (freeLists at: i)].
	"Large chunks"
	self freeTreeNodesDo: [:freeNode |
		self updateListStartingAt: freeNode.
		freeNode]! !

!SpurMemoryManager methodsFor: 'gc - global' stamp: 'cb 7/3/2018 09:25'!
updateFullGCStats	
	"Stats updated (since VM start-up): 
	1) number of full GCs,
	2) time spent in full GC,
	3) time spent in compaction (included in 2)
	4) time spent in sweep phase (included in 2 & 3, 0 if no sweep phase)
	5) time spent in mark phase (included in 2)"
	statFullGCs := statFullGCs + 1.
	statFullGCUsecs := statFullGCUsecs + (statGCEndUsecs - gcStartUsecs).
	statCompactionUsecs := statCompactionUsecs + (statGCEndUsecs - compactionStartUsecs).
	gcSweepEndUsecs = 0 ifFalse: [statSweepUsecs := statSweepUsecs + (gcSweepEndUsecs - compactionStartUsecs)].
	statMarkUsecs := statMarkUsecs + (gcMarkEndUsecs - gcStartUsecs).! !

!SpurMemoryManager methodsFor: 'initialization' stamp: 'eem 06/22/2021 15:31'!
updateListStartingAt: freeNode 
	|prev obj|
	freeNode = 0 ifTrue: [^self].
	self deny: (self isLilliputianSize: (self bytesInBody: freeNode)).
	prev := freeNode.
	self storePointer: self freeChunkPrevIndex ofFreeChunk: prev withValue: 0.
	[obj := self fetchPointer: self freeChunkNextIndex ofFreeChunk: prev.
	 obj ~= 0] whileTrue:
		[self storePointer: self freeChunkPrevIndex ofFreeChunk: obj withValue: prev.
		 prev := obj]! !

!SpurMemoryManager methodsFor: 'image segment in/out' stamp: 'eem 5/10/2017 14:53'!
updatePostScavenge: anObj
	<inline: true>
	^(self isForwarded: anObj) ifTrue: [self followForwarded: anObj] ifFalse: [anObj]! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 8/31/2015 11:03'!
updateRootOfObjStackAt: objStackRootIndex with: newRootPage
	self storePointer: objStackRootIndex
		ofObject: hiddenRootsObj
		withValue: newRootPage.
	objStackRootIndex caseOf: {
		[MarkStackRootIndex]		->	[markStack := newRootPage].
		[WeaklingStackRootIndex]	->	[weaklingStack := newRootPage].
		[MournQueueRootIndex]	->	[mournQueue := newRootPage] }.
	^newRootPage! !

!SpurMemoryManager methodsFor: 'accessing' stamp: 'cb 7/3/2018 09:17'!
updateSweepEndUsecs
	gcSweepEndUsecs := coInterpreter ioUTCMicrosecondsNow.! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 9/24/2014 11:43'!
validClassTableHashes
	"Check the hashes of classes in the table.  The tricky thing here is that classes may be duplicated
	 in the table.  So each entry must be in the table at its hash, even if it is elsewhere in the table."

	self validClassTableRootPages ifFalse:
		[^false].

	self classTableEntriesDo:
		[:classOrNil :ignored| | hash |
		 (self isForwarded: classOrNil) ifTrue:
			[^0].
		  hash := self rawHashBitsOf: classOrNil.
		  hash = 0 ifTrue:
			[^false].
		  (self noCheckClassAtIndex: hash) ~= classOrNil ifTrue:
			[^false]].

	^true! !

!SpurMemoryManager methodsFor: 'class table' stamp: 'eem 10/17/2013 13:48'!
validClassTableRootPages
	"Answer if hiddenRootsObj is of the right size with the
	 expected contents, and if numClassTablePages is correct."

	(self numSlotsOf: hiddenRootsObj) = (self classTableRootSlots + self hiddenRootSlots) ifFalse:
		[^false].

	"is it in range?"
	(numClassTablePages > 1 and: [numClassTablePages <= self classTableRootSlots]) ifFalse:
		[^false].
	"are all pages the right size?"
	0 to: numClassTablePages - 1 do:
		[:i| | obj |
		 obj := self fetchPointer: i ofObject: hiddenRootsObj.
		 ((self addressCouldBeObj: obj)
		  and: [(self numSlotsOf: obj) = self classTablePageSize]) ifFalse:
			[^false]].
	"are all entries beyond numClassTablePages nil?"
	numClassTablePages to: self classTableRootSlots - 1 do:
		[:i|
		(self fetchPointer: i ofObject: hiddenRootsObj) ~= nilObj ifTrue:
			[^false]].
	^true! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 1/18/2015 16:15'!
validFreeTree
	<api>
	^(self validFreeTreeChunk: (freeLists at: 0) parent: 0) isNil! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 3/27/2017 12:35'!
validFreeTreeChunk: chunk
	<inline: false>
	(segmentManager segmentContainingObj: chunk) ifNil:
		[^false].
	^(self validFreeTreeChunk: chunk parent: (self fetchPointer: self freeChunkParentIndex ofFreeChunk: chunk)) isNil! !

!SpurMemoryManager methodsFor: 'free space' stamp: 'eem 06/22/2021 15:31'!
validFreeTreeChunk: chunk parent: parent
	<var: 'reason' type: #'const char *'>
	<returnTypeC: #'const char *'>
	chunk = 0 ifTrue:
		[^nil].
	(self addressCouldBeOldObj: chunk) ifFalse:
		[^'not in old space'].
	(self bytesInBody: chunk) / self allocationUnit < self numFreeLists ifTrue:
		[^'too small'].
	parent ~= (self fetchPointer: self freeChunkParentIndex ofFreeChunk: chunk) ifTrue:
		[^'bad parent'].

	(segmentManager segmentContainingObj: chunk) ~~ (segmentManager segmentContainingObj: (self addressAfter: chunk)) ifTrue:
		[^'not in one segment'].
	(self validFreeTreeChunk: (self fetchPointer: self freeChunkSmallerIndex ofFreeChunk: chunk) parent: chunk) ifNotNil:
		[:reason| ^reason].
	(self validFreeTreeChunk: (self fetchPointer: self freeChunkLargerIndex ofFreeChunk: chunk) parent: chunk) ifNotNil:
		[:reason| ^reason].
	^nil! !

!SpurMemoryManager methodsFor: 'obj stacks' stamp: 'eem 8/31/2015 11:03'!
validObjStacks
	^(markStack = nilObj or: [self isValidObjStack: markStack])
	  and: [(weaklingStack = nilObj or: [self isValidObjStack: weaklingStack])
	  and: [mournQueue = nilObj or: [self isValidObjStack: mournQueue]]]! !

!SpurMemoryManager methodsFor: 'gc - scavenge/compact' stamp: 'eem 12/24/2016 14:32'!
vanillaRemapObj: objOop
	"Scavenge or simply follow objOop.  Answer the new location of objOop.
	 The send should have been guarded by a send of shouldRemapOop:.
	 The method is called remapObj: for compatibility with ObjectMemory."
	<inline: true>
	| resolvedObj |
	self assert: (self shouldRemapOop: objOop).
	(self isForwarded: objOop)
		ifTrue:
			[resolvedObj := self followForwarded: objOop]
		ifFalse:
			[self deny: (self isInFutureSpace: objOop).
			 resolvedObj := objOop].
	(self scavengeInProgress
	 and: [(self isReallyYoung: resolvedObj) "don't scavenge immediate, old, or CogMethod objects."
	 and: [(self isInFutureSpace: resolvedObj) not]]) ifTrue: 
		[^scavenger copyAndForward: resolvedObj].
	^resolvedObj! !

!SpurMemoryManager methodsFor: 'gc - scavenge/compact' stamp: 'eem 12/26/2016 17:54'!
vanillaShouldRemapObj: objOop
	<inline: true>
	"Answer if the obj should be scavenged, or simply followed. Sent via the compactor
	 from shouldRemapObj:.  We test for being already scavenged because mapStackPages
	 via mapInterpreterOops may be applied twice in the context of a global GC where a
	 scavenge, followed by a scan-mark-free, and final compaction passes may result in
	 scavenged fields being visited twice."
	^(self isForwarded: objOop)
	  or: [(self isReallyYoungObject: objOop)
		 and: [(self isInFutureSpace: objOop) not]]! !

!SpurMemoryManager methodsFor: 'memory access' stamp: 'eem 10/15/2013 10:20'!
vmEndianness
	<api>
	"1 = big, 0 = little"
	^self cCode: [VMBIGENDIAN] inSmalltalk: [self subclassResponsibility]! !

!SpurMemoryManager methodsFor: 'class table puns' stamp: 'eem 3/9/2016 14:49'!
weakArrayClassIndexPun
	"Class puns are class indices not used by any class.  There is an entry
	 for the pun that refers to the notional class of objects with this class
	 index.  But because the index doesn't match the class it won't show up
	 in allInstances, hence hiding the object with a pun as its class index.
	 The puns occupy indices 16 through 31."
	<cmacro>
	^17! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 12/12/2015 14:19'!
weakArrayFormat
	<api>
	<cmacro>
	^4! !

!SpurMemoryManager methodsFor: 'spur bootstrap' stamp: 'eem 11/18/2014 15:11'!
weaklingStack
	^weaklingStack! !

!SpurMemoryManager methodsFor: 'debug printing' stamp: 'eem 3/19/2014 17:21'!
whereIsMaybeHeapThing: anOop
	<returnTypeC: 'char *'>
	(self isInNewSpace: anOop) ifTrue:
		[(self isInEden: anOop) ifTrue: [^' is in eden'].
		 (self isInFutureSpace: anOop) ifTrue: [^' is in future space'].
		 (self isInPastSpace: anOop) ifTrue: [^' is in past space'].
		 ^' is in new space'].
	(self isInOldSpace: anOop) ifTrue:
		[(segmentManager segmentContainingObj: anOop) ifNotNil:
			[^' is in old space'].
		 ^' is between old space segments'].
	^nil! !

!SpurMemoryManager methodsFor: 'debug support' stamp: 'eem 3/9/2020 17:32'!
withSimulatorFetchPointerMovedAsideDo: aBlock
	"For performance, remove the simulator implementation of fetchPointer:ofObject:
	 while aBlock is running and answer the block's result."
	| theMethod |
	theMethod := self class lookupSelector: #fetchPointer:ofObject:.
	self deny: theMethod isNil.
	theMethod methodClass == SpurMemoryManager
		ifTrue: [theMethod := nil]
		ifFalse: [theMethod methodClass basicRemoveSelector: #fetchPointer:ofObject:].
	^aBlock ensure:
		[theMethod ifNotNil:
			[theMethod methodClass
				basicAddSelector: #fetchPointer:ofObject:
				withMethod: theMethod]]! !

!SpurMemoryManager methodsFor: 'header formats' stamp: 'eem 10/8/2013 12:21'!
wordIndexableFormat
	"Either firstLongFormat or sixtyFourBitIndexableFormat"
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'word size' stamp: 'eem 11/13/2014 15:23'!
wordSize
	"Answer the manager's word size, which is the size of an oop, and which
	 is assumed to be equivalent to the underlying machine's word size."
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'class table puns' stamp: 'eem 10/8/2013 12:17'!
wordSizeClassIndexPun
	^self subclassResponsibility! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 11/18/2016 16:59'!
writeEnableMemory
	<doNotGenerate>
	memory class == ReadOnlyArrayWrapper ifTrue:
		[memory := memory array]! !

!SpurMemoryManager methodsFor: 'snapshot' stamp: 'eem 3/9/2014 08:17'!
writeImageSegmentsToFile: aBinaryStream
	<doNotGenerate>
	^segmentManager writeImageSegmentsToFile: aBinaryStream! !

!SpurMemoryManager methodsFor: 'simulation' stamp: 'eem 11/18/2016 16:59'!
writeProtectMemory
	<doNotGenerate>
	memory class ~~ ReadOnlyArrayWrapper ifTrue:
		[memory := ReadOnlyArrayWrapper around: memory]! !

CogClass subclass: #SpurSegmentManager
	instanceVariableNames: 'manager numSegments numSegInfos segments firstSegmentSize canSwizzle sweepIndex preferredPinningSegment totalHeapSizeIncludingBridges'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManager'!

!SpurSegmentManager commentStamp: 'eem 6/7/2017 14:00' prior: 0!
Instances of SpurSegmentManager manage oldSpace, which is organized as a sequence of segments.  Segments can be obtained from the operating system and returned to the operating system when empty and shrinkage is required.  Segments are kept invisible from the SpurMemoryManager by using "bridge" objects, "fake" pinned objects to bridge the gaps between segments.  A pinned object header occupies the last 16 bytes of each segment, and the pinned object's size is the distance to the start of the next segment.  So when the memory manager enumerates objects it skips over these bridges and memory appears linear.  The constraint is that segments obtained from the operating system must be at a higher address than the first segment.  The maximum size of large objects, being an overflow slot size, should be big enough to bridge the gaps, because in 32-bits the maximum size is 2^32 slots.  In 64-bits the maximum size of large objects is 2^56 slots, or 2^59 bits, which we hope will suffice.

When an image is written to a snapshot file the second word of the header of the bridge at the end of each segment is replaced by the size of the following segment, the segments are written to the file, and the second word of each bridge is restored.  Hence the length of each segment is derived from the bridge at the end of the preceeding segment.  The length of the first segment is stored in the image header as firstSegmentBytes.  The start of each segment is also derived from the bridge as a delta from the start of the previous segment.  The start of The first segment is stored in the image header as startOfMemory.

On load all segments are read into one single segment, eliminating the bridge objects, and computing the swizzle distance for each segment, based on where the segments were in memory when the image file was written, and where the coalesced segment ends up on load.  Then the segment is traversed, swizzling pointers by selecting the relevant swizzle for each oop's segment.

Instance Variables
	manager						<SpurMemoryManager>
	numSegments					<Integer>
	numSegInfos					<Integer>
	segments						<Array of SpurSegmentInfo>
	firstSegmentSize				<Integer>
	canSwizzle						<Boolean>
	sweepIndex					<Integer>
	preferredPinningSegment		<SpurSegmentInfo>
	totalHeapSizeIncludingBridges	<integer>

canSwizzle
	- a flag set and cleared during initialization to validate that swizzling is only performed at the right time

firstSegmentSize
	- the size of the first segment when loading an image

manager
	- the memory manager the receiver manages segments for (simulation only)
	
numSegInfos
	- the size of the segments array in units of SpurSegmentInfo size
	
numSegments
	- the number of segments (the number of used entries in segments, <= numSegInfos)

preferredPinningSegment
	- the segment in which objects should be copied when pinned, so as to cluster pinned objects in as few segments as possible.  As yet unimplemented.

segments
	- the start addresses, lengths and offsets to adjust oops on image load, for each segment

sweepIndex
	- a segment index used to optimize setting the containsPinned flag on segments during freeUnmarkedObjectsAndSortAndCoalesceFreeSpace

totalHeapSizeIncludingBridges
	- the total size of all segments, used to compute heap usage!

SpurSegmentManager subclass: #SpurLeakCheckingSegmentManager
	instanceVariableNames: 'oldHeapBase oldHeapLimit detectedInvalidOop subsequentReport'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Utilities'!

!SpurLeakCheckingSegmentManager methodsFor: 'accessing' stamp: 'eem 11/28/2019 12:58'!
detectedInvalidOop
	^detectedInvalidOop! !

!SpurLeakCheckingSegmentManager methodsFor: 'initialization' stamp: 'eem 11/28/2019 13:15'!
initialize
	super initialize.
	detectedInvalidOop := false.
	subsequentReport := false! !

!SpurLeakCheckingSegmentManager methodsFor: 'initialization' stamp: 'eem 11/28/2019 13:12'!
noteValidHeapExtent: bytesToShift
	oldHeapBase :=  manager memoryBaseForImageRead - bytesToShift.
	oldHeapLimit := self lastSegment segStart + self lastSegment segSize - self lastSegment swizzle! !

!SpurLeakCheckingSegmentManager methodsFor: 'snapshot' stamp: 'eem 11/28/2019 13:17'!
swizzleObj: objOop in: containerOopOrNil
	<inline: false>
	"Override to check that objOop is in bounds."
	(self oop: objOop isGreaterThanOrEqualTo: oldHeapBase andLessThan: oldHeapLimit) ifFalse:
		[manager coInterpreter verbose ifTrue:
			[subsequentReport ifFalse:
				['Swizzling:\n' printf.
				 subsequentReport := true].
			 containerOopOrNil
				ifNil: ['Oop %p is out of bounds\n' printf: { objOop }]
				ifNotNil: ['Oop %p is out of bounds at/in %p\n' printf: { objOop. containerOopOrNil - manager nilObject }]].
		 detectedInvalidOop := true].
	"And now swizzle inline"
	numSegments - 1 to: 1 by: -1 do:
		[:i|
		objOop >= (segments at: i) segStart ifTrue:
			[^objOop + (segments at: i) swizzle]].
	^objOop + (segments at: 0) swizzle! !

!SpurSegmentManager class methodsFor: 'translation' stamp: 'eem 6/7/2017 12:20'!
declareCVarsIn: aCCodeGenerator
	aCCodeGenerator
		var: #segments type: #'SpurSegmentInfo *';
		var: #totalHeapSizeIncludingBridges type: #usqInt! !

!SpurSegmentManager class methodsFor: 'translation' stamp: 'eem 10/29/2013 23:43'!
implicitReturnTypeFor: aSelector
	"Answer the return type for methods that don't have an explicit return."
	^#void! !

!SpurSegmentManager class methodsFor: 'translation' stamp: 'eem 10/13/2013 18:45'!
isNonArgumentImplicitReceiverVariableName: instVarName
	^instVarName = 'manager'! !

!SpurSegmentManager methodsFor: 'growing/shrinking memory' stamp: 'cb 6/11/2018 09:54'!
addSegmentOfSize: ammount
	<returnTypeC: #'SpurSegmentInfo *'>
	<inline: false>
	| allocatedSize |
	<var: #newSeg type: #'SpurSegmentInfo *'>
	<var: #segAddress type: #'void *'>
	<var: #allocatedSize type: #'usqInt'>
	self cCode: [] inSmalltalk: [segments ifNil: [^nil]]. "bootstrap"
	(manager "sent to the manager so that the simulator can increase memory to simulate a new segment"
			sqAllocateMemorySegmentOfSize: ammount
			Above: (self firstGapOfSizeAtLeast: ammount)
			AllocatedSizeInto: (self cCode: [self addressOf: allocatedSize]
									inSmalltalk: [[:sz| allocatedSize := sz]])) ifNotNil:
		[:segAddress| | newSegIndex newSeg |
		 newSegIndex := self insertSegmentFor: segAddress asUnsignedIntegerPtr.
		 "Simulation insertion code duplicates entries if newSegIndex ~= numSegments - 1"
		 self cCode: '' inSmalltalk: [segments at: newSegIndex put: SpurSegmentInfo new].
		 newSeg := self addressOf: (segments at: newSegIndex).
		 newSeg
			segStart: segAddress asUnsignedIntegerPtr;
			segSize: allocatedSize;
			swizzle: 0. "Required in the C version only"
		 self assert: self segmentOverlap not. "self printSegmentAddresses."
		 self bridgeFrom: (self addressOf: (segments at: newSegIndex - 1)) to: newSeg.
		 self bridgeFrom: newSeg to: (newSegIndex = (numSegments - 1) ifFalse:
										[self addressOf: (segments at: newSegIndex + 1)]).
		 totalHeapSizeIncludingBridges := totalHeapSizeIncludingBridges + allocatedSize.
		 "test isInMemory:"
		 0 to: numSegments - 1 do:
			[:i|
			self assert: (self isInSegments: (segments at: i) segStart).
			self assert: (self isInSegments: (segments at: i) segLimit - manager wordSize).
			self assert: ((self isInSegments: (segments at: i) segLimit) not
						or: [i < (numSegments - 1)
							and: [(segments at: i) segLimit = (segments at: i + 1) segStart]]).
			self assert: ((self isInSegments: (segments at: i) segStart - manager wordSize) not
							or: [i > 0
								and: [(segments at: i - 1) segLimit = (segments at: i) segStart]])].
		 ^newSeg].
	^nil! !

!SpurSegmentManager methodsFor: 'snapshot' stamp: 'eem 3/7/2014 16:53'!
adjustSegmentSwizzlesBy: firstSegmentShift
	"Adjust swizzles by firstSegmentShift.  Also computes segStarts as
	 they were in the image when it was written, so that oops' segments
	 can be determined and hence oops correctly swizzled."
	<inline: false>
	<var: 'segInfo' type: #'SpurSegmentInfo *'>
	| oldBaseAddr |
	oldBaseAddr := manager memoryBaseForImageRead - firstSegmentShift.
	0 to: numSegments - 1 do:
		[:i| | segInfo |
		 segInfo := self addressOf: (segments at: i).
		 segInfo
			segStart: segInfo segStart + oldBaseAddr;
			swizzle: segInfo swizzle - oldBaseAddr].
	canSwizzle := true! !

!SpurSegmentManager methodsFor: 'debug support' stamp: 'eem 11/29/2013 11:37'!
allBridgesMarked
	0 to: numSegments - 1 do:
		[:i| | bridgeObj |
		 bridgeObj := self bridgeAt: i.
		 self assert: (self isValidSegmentBridge: bridgeObj).
		 (manager isMarked: bridgeObj) ifFalse:
			[^false]].
	^true

	"for debugging:"
	"(0 to: numSegments - 1) select:
		[:i| | bridgeObj |
		 bridgeObj := self bridgeAt: i.
		 self assert: (self isValidSegmentBridge: bridgeObj).
		 manager isMarked: bridgeObj]"! !

!SpurSegmentManager methodsFor: 'private' stamp: 'eem 11/25/2019 16:19'!
allocateOrExtendSegmentInfos
	"Increase the number of allocated segInfos by 16."
	| newNumSegs |
	numSegInfos = 0 ifTrue:
		[numSegInfos := 16.
		 segments := self
						cCode: [self calloc: numSegInfos _: (self sizeof: SpurSegmentInfo)]
						inSmalltalk: [CArrayAccessor on: ((1 to: numSegInfos) collect: [:i| SpurSegmentInfo new])].
		 ^self].
	newNumSegs := numSegInfos + 16.
	segments := self
						cCode: [self realloc: segments _: newNumSegs * (self sizeof: SpurSegmentInfo)]
						inSmalltalk: [CArrayAccessor on: segments object,
									((numSegInfos to: newNumSegs) collect: [:i| SpurSegmentInfo new])].
	self cCode:
		[segments = 0 ifTrue:
			[self error: 'out of memory; cannot allocate more segments'].
		 self
			memset: segments + numSegInfos
			_: 0
			_: newNumSegs - numSegInfos * (self sizeof: SpurSegmentInfo)].
	numSegInfos := newNumSegs! !

!SpurSegmentManager methodsFor: 'bridges' stamp: 'eem 11/29/2013 11:18'!
bridgeAt: segIndex
	^self bridgeFor: (self addressOf: (segments at: segIndex))! !

!SpurSegmentManager methodsFor: 'bridges' stamp: 'eem 11/29/2013 11:17'!
bridgeFor: aSegment
	<var: 'aSegment' type: #'SpurSegmentInfo *'>
	^manager objectStartingAt: aSegment segLimit - manager bridgeSize! !

!SpurSegmentManager methodsFor: 'growing/shrinking memory' stamp: 'cb 5/29/2018 14:44'!
bridgeFrom: aSegment to: nextSegmentOrNil
	"Create a bridge from aSegment to the next segment,
	 or create a terminating bridge if there is no next segment."
	<var: #aSegment type: #'SpurSegmentInfo *'>
	<var: #nextSegmentOrNil type: #'SpurSegmentInfo *'>
	| segEnd clifton bridgeSpan |
	segEnd := aSegment segLimit.
	clifton := segEnd - manager bridgeSize. "clifton is where the Avon bridge begins..."
	bridgeSpan := nextSegmentOrNil
					ifNil: [manager bridgeSize]
					ifNotNil: [nextSegmentOrNil segStart - segEnd + manager bridgeSize].
	self assert: bridgeSpan >= 0.
	manager initSegmentBridgeWithBytes: bridgeSpan at: clifton.
	"the revised bridge should get us to the new segment"
	self assert: (manager addressAfter: (manager objectStartingAt: clifton))
				= (nextSegmentOrNil
						ifNil: [aSegment segLimit]
						ifNotNil: [nextSegmentOrNil segStart])
! !

!SpurSegmentManager methodsFor: 'debug support' stamp: 'eem 5/28/2014 10:11'!
checkSegments
	self assert: numSegments >= 1.
	0 to: numSegments - 1 do:
		[:i|
		self assert: (manager addressCouldBeObj: (segments at: i) segStart).
		self assert: (self isValidSegmentBridge: (self bridgeAt: i))].
	self assert: (segments at: numSegments - 1) segLimit - manager bridgeSize = manager endOfMemory! !

!SpurSegmentManager methodsFor: 'snapshot' stamp: 'eem 6/7/2017 12:21'!
collapseSegmentsPostSwizzle
	"The image has been loaded, old segments reconstructed, and the heap
	 swizzled into a single contiguous segment.  Collapse the segments into one."
	<inline: false>
	canSwizzle := false.
	self cCode: []
		inSmalltalk:
			[segments ifNil:
				[self allocateOrExtendSegmentInfos]].
	numSegments := 1.
	(segments at: 0)
		segStart: manager oldSpaceStart;
		segSize: (totalHeapSizeIncludingBridges := manager endOfMemory - manager oldSpaceStart).
	manager bootstrapping ifTrue:
		["finally plant a bridge at the end of the coalesced segment and cut back the
		  manager's notion of the end of memory to immediately before the bridge."
		 self assert: manager endOfMemory = (segments at: 0) segLimit.
		 manager
			initSegmentBridgeWithBytes: manager bridgeSize
			at: manager endOfMemory - manager bridgeSize].
	self assert: (manager isSegmentBridge: (self bridgeAt: 0)).
	self assert: (manager numSlotsOfAny: (self bridgeAt: 0)) = 0! !

!SpurSegmentManager methodsFor: 'simulation only' stamp: 'eem 6/29/2017 15:59'!
computeTotalHeapSizeIncludingBridges
	totalHeapSizeIncludingBridges := manager endOfMemory - manager oldSpaceStart! !

!SpurSegmentManager methodsFor: 'growing/shrinking memory' stamp: 'nice 11/9/2016 21:49'!
findEmptySegNearestInSizeTo: size
	| seg best delta |
	<var: #seg type: #'SpurSegmentInfo *'>
	<var: #best type: #'SpurSegmentInfo *'>
	best := nil.
	delta := size.
	0 to: numSegments - 1 do:
		[:i|
		seg := self addressOf: (segments at: i).
		(self isEmptySegment: seg) ifTrue:
			[best
				ifNil: [best := seg]
				ifNotNil:
					[(size >= (seg segSize * 0.75)
					 and: [(self cCoerce: (seg segSize - size) to: #sqInt ) abs < delta]) ifTrue:
						[best := seg. delta := (self cCoerce: (seg segSize - size) to: #sqInt ) abs]]]].
	^best! !

!SpurSegmentManager methodsFor: 'growing/shrinking memory' stamp: 'eem 06/22/2021 15:31'!
firstGapOfSizeAtLeast: size
	"Answer the segment limit of the first segment followed by a gap of at least size bytes."
	0 to: numSegments - 2 do:
		[:i| | bridge |
		bridge := self bridgeAt: i.
		(manager bytesInBody: bridge) - manager bridgeSize >= size ifTrue:
			[^(segments at: i) segLimit asVoidPointer]].
	^(segments at: numSegments - 1) segLimit asVoidPointer! !

!SpurSegmentManager methodsFor: 'snapshot' stamp: 'eem 10/9/2013 10:56'!
firstSegmentBytes
	^(segments at: 0) segSize! !

!SpurSegmentManager methodsFor: 'initialization' stamp: 'eem 10/9/2013 21:19'!
firstSegmentSize: firstSegSize
	"Remember firstSegSize for subsequent image segment parsing."
	firstSegmentSize := firstSegSize! !

!SpurSegmentManager methodsFor: 'growing/shrinking memory' stamp: 'eem 5/21/2014 11:40'!
indexOfSegment: seg
	<var: #seg type: #'SpurSegmentInfo *'>
	0 to: numSegments - 1 do:
		[:i|
		seg segStart = (segments at: i) segStart ifTrue:
			[^i]].
	self error: 'segment not found'! !

!SpurSegmentManager methodsFor: 'spur bootstrap' stamp: 'eem 12/17/2013 17:01'!
initForBootstrap
	numSegInfos = 0 ifTrue:
		[self allocateOrExtendSegmentInfos]! !

!SpurSegmentManager methodsFor: 'simulation only' stamp: 'eem 8/26/2014 09:32'!
initSegmentForInImageCompilationFrom: base to: limit
	<doNotGenerate>
	| bridge |
	self allocateOrExtendSegmentInfos.
	numSegments := 1.
	bridge := manager initSegmentBridgeWithBytes: manager memory byteSize - limit at: limit - manager bridgeSize.
	segments := {SpurSegmentInfo new
						segStart: base;
						segSize: limit - base;
						yourself}.
	segments := CArrayAccessor on: segments! !

!SpurSegmentManager methodsFor: 'initialization' stamp: 'eem 6/7/2017 12:20'!
initialize
	numSegments := numSegInfos := sweepIndex := totalHeapSizeIncludingBridges := 0.
	canSwizzle := false! !

!SpurSegmentManager methodsFor: 'simulation only' stamp: 'eem 06/22/2021 15:31'!
initializeFromFreeChunks: freeChunks
	<doNotGenerate>
	"For testing, create a set of segments using the freeChunks as bridges."
	self assert: (freeChunks allSatisfy: [:f| manager hasOverflowHeader: f]).
	numSegments := freeChunks size.
	freeChunks do:
		[:f|
		manager initSegmentBridgeWithBytes: (manager bytesInBody: f) at: (manager startOfObject: f).
		self assert: (manager isSegmentBridge: f)].
	segments := (1 to: numSegments) collect:
					[:i| | bridge start size |
					bridge := freeChunks at: i.
					start := i = 1
								ifTrue: [manager newSpaceLimit]
								ifFalse: [manager addressAfter: (freeChunks at: i - 1)].
					size := bridge + manager baseHeaderSize - start.
					SpurSegmentInfo new
						segStart: start;
						segSize: size;
						yourself].
	manager setEndOfMemory: segments last segLimit.
	segments := CArrayAccessor on: segments.
	freeChunks with: segments object do:
		[:bridge :segment|
		self assert: (self isValidSegmentBridge: bridge).
		self assert: bridge = (self bridgeFor: segment)]! !

!SpurSegmentManager methodsFor: 'growing/shrinking memory' stamp: 'eem 3/24/2015 13:08'!
insertSegmentFor: segAddress
	"Reserve a new segInfo for segAddress.  If segAddress
	 is in the middle of the existing segments, shuffle them
	 up to make room. Answer the new segment's index."
	| segIndex lastSegIndex |
	self assert: segAddress >= (segments at: 0) segLimit.
	numSegments = numSegInfos ifTrue:
		[self allocateOrExtendSegmentInfos].
	self assert: numSegments < numSegInfos.
	segIndex := lastSegIndex := numSegments - 1.
	numSegments := numSegments + 1.
	[segAddress >= (segments at: segIndex) segLimit ifTrue:
		[segIndex := segIndex + 1.
		 lastSegIndex to: segIndex by: -1 do:
			[:idx| segments at: idx + 1 put: (segments at: idx)].
		 ^segIndex].
	 segIndex := segIndex - 1]
		repeat! !

!SpurSegmentManager methodsFor: 'growing/shrinking memory' stamp: 'eem 5/21/2014 11:40'!
isEmptySegment: seg
	<var: #seg type: #'SpurSegmentInfo *'>
	| firstObj |
	firstObj := manager objectStartingAt: seg segStart.
	^(manager isFreeObject: firstObj)
	  and: [(manager addressAfter: firstObj) = (seg segLimit - manager bridgeSize)]! !

!SpurSegmentManager methodsFor: 'testing' stamp: 'eem 11/11/2013 16:48'!
isInSegments: address
	<var: #address type: #usqInt>
	0 to: numSegments - 1 do:
		[:i|
		address < (segments at: i) segStart ifTrue:
			[^false].
		address < (segments at: i) segLimit ifTrue:
			[^true]].
	^false! !

!SpurSegmentManager methodsFor: 'testing' stamp: 'eem 1/29/2014 16:34'!
isValidSegmentBridge: objOop
	"bridges bridge the gaps between segments. They are the last object in each segment."
	^((manager addressCouldBeObj: objOop) or: [objOop = manager endOfMemory])
	  and: [(manager isSegmentBridge: objOop)
	  and: [(manager hasOverflowHeader: objOop)
		or: [(manager numSlotsOfAny: objOop) = 0]]]! !

!SpurSegmentManager methodsFor: 'accessing' stamp: 'eem 4/24/2014 12:07'!
lastBridge
	^self bridgeAt: numSegments - 1! !

!SpurSegmentManager methodsFor: 'accessing' stamp: 'eem 5/21/2014 10:54'!
lastSegment
	^segments at: numSegments - 1! !

!SpurSegmentManager methodsFor: 'initialization' stamp: 'eem 10/10/2013 15:39'!
manager: aSpurMemoryManager
	manager := aSpurMemoryManager! !

!SpurSegmentManager methodsFor: 'snapshot' stamp: 'eem 3/9/2014 09:16'!
nextNonEmptySegmentAfter: i
	"Answer the the next non-empty segment or nil. The size of a segment includes
	 that of its bridge. A segment containing just a free object and a bridge will still
	 have a size of manager bridgeSize after shortening it in prepareForSnapshot."
	<returnTypeC: #'SpurSegmentInfo *'>
	| nextx |
	nextx := i.
	[(nextx := nextx + 1) >= numSegments ifTrue:
		[^nil].
	 (segments at: nextx) segSize > manager bridgeSize ifTrue:
		[^self addressOf: (segments at: nextx)]] repeat! !

!SpurSegmentManager methodsFor: 'pinning' stamp: 'eem 1/25/2017 10:05'!
notePinned: objOop
	"Let the segmentManager mark which segments contain pinned objects"
	self assert: (manager isPinned: objOop).
	(manager isSegmentBridge: objOop)
		ifTrue:
			[manager setIsMarkedOf: objOop to: true]
		ifFalse:
			[[self oop: (segments at: sweepIndex) segLimit isLessThan: objOop] whileTrue:
				[sweepIndex := sweepIndex + 1].
			 (segments at: sweepIndex) containsPinned: true]! !

!SpurSegmentManager methodsFor: 'accessing' stamp: 'eem 10/12/2013 14:37'!
numSegments
	<cmacro: '() GIV(numSegments)'>
	^numSegments! !

!SpurSegmentManager methodsFor: 'snapshot' stamp: 'eem 1/24/2016 13:42'!
postSnapshot
	"Restore all shortened segments to their proper size, re-freeing the trailing space."
	<inline: false>
	| seg |
	<var: #seg type: #'SpurSegmentInfo *'>
	"Set endOfMemory first, to avoid assert fails in freeChunkWithBytes:at:."
	seg := self addressOf: (segments at: numSegments - 1).
	seg lastFreeObject
		ifNil: [self assert: manager endOfMemory = (seg segLimit - manager bridgeSize)]
		ifNotNil: [manager setEndOfMemory: seg savedSegSize + seg segStart - manager bridgeSize].

	numSegments - 1 to: 0 by: -1 do:
		[:i|
		 seg := self addressOf: (segments at: i).
		 seg lastFreeObject ifNotNil:
			[:freeChunk| | address |
			address := seg segLimit - manager bridgeSize.
			seg segSize: seg savedSegSize.
			self bridgeFrom: seg
				to: (i < (numSegments - 1) ifTrue: [self addressOf: (segments at: i + 1)]).
			manager
				addFreeChunkWithBytes: seg segLimit - address - manager bridgeSize
				at: address]].
	self checkSegments.
	manager checkFreeSpace: GCModeFull! !

!SpurSegmentManager methodsFor: 'pinning' stamp: 'eem 4/24/2014 11:08'!
prepareForGlobalSweep
	"Let the segmentManager mark which segments contain pinned objects via notePinned:.
	 For coallesceFreeChunk:, ensure that the last bridge is marked."
	sweepIndex := 0.
	0 to: numSegments - 1 do:
		[:i| (segments at: i) containsPinned: false].
	manager
		setIsMarkedOf: (self bridgeAt: numSegments - 1)
		to: true! !

!SpurSegmentManager methodsFor: 'snapshot' stamp: 'eem 1/24/2016 14:31'!
prepareForSnapshot
	"shorten all segments by any trailing free space."
	<inline: #never>
	<var: #seg type: #'SpurSegmentInfo *'>
	self checkSegments.
	0 to: numSegments - 1 do:
		[:i|
		 (segments at: i)
			savedSegSize: (segments at: i) segSize;
			lastFreeObject: nil].

	"Ideally finding the lastFreeObject of each segment would be
	 done in some linear pass through the heap.  But for now KISS."
	manager freeTreeNodesDo:
		[:freeChunk| | node next seg |
		 node := freeChunk.
		 [node ~= 0] whileTrue:
			[next := manager objectAfter: node limit: manager endOfMemory.
			 (manager isSegmentBridge: next)
				ifTrue:
					[seg := self segmentContainingObj: node.
					 seg lastFreeObject: node.
					 node := 0]
				ifFalse:
					[node := manager
								fetchPointer: manager freeChunkNextIndex
								ofFreeChunk: node]].
		 freeChunk].

	0 to: numSegments - 1 do:
		[:i|
		 (segments at: i) lastFreeObject ifNotNil:
			[:freeChunk|
			manager detachFreeObject: freeChunk.
			(segments at: i)
				segSize: (manager startOfObject: freeChunk)
						+ manager bridgeSize
						- (segments at: i) segStart.
			self bridgeFrom: (self addressOf: (segments at: i))
				to: (i < (numSegments - 1) ifTrue: [self addressOf: (segments at: i + 1)])]].

	"perhaps this should read
		manager setEndOfMemory: 0; assimilateNewSegment: (segments at: numSegments - 1)"
	manager setEndOfMemory: (segments at: numSegments - 1) segLimit - manager bridgeSize! !

!SpurSegmentManager methodsFor: 'simulation only' stamp: 'cb 5/29/2018 15:57'!
printSegmentAddresses
	<doNotGenerate>
	"Convenient when debugging segment overlap"
	Transcript show: 'Seg addresses: '; cr.
	0 to: numSegments - 1 do:
		[:i|
			Transcript
				show: (segments at: i) segStart;
				show: ' -> ';
				show: (segments at: i) segLimit;
				show: ' ; '].
			Transcript cr.! !

!SpurSegmentManager methodsFor: 'private' stamp: 'eem 11/15/2019 10:52'!
readHeapFrom: f at: location dataBytes: numBytes
	"Read numBytes from f into mmory at location.  Answer the number of bytes read."
	<inline: true>
	^self cCode:
			[self
				sq: (self pointerForOop: location)
				Image: (self sizeof: #char)
				File: numBytes
				Read: f]
		inSmalltalk:
			[| bytesPerElement |
			 bytesPerElement := manager memory bytesPerElement.
			 (f	readInto: manager memory
				startingAt: location // bytesPerElement + 1
				count: numBytes // bytesPerElement)
			  * bytesPerElement]! !

!SpurSegmentManager methodsFor: 'snapshot' stamp: 'eem 1/25/2016 16:50'!
readHeapFromImageFile: f dataBytes: numBytes
	"Read numBytes of image data from f into memory at memoryBaseForImageRead.
	 Answer the number of bytes written.  In addition, read each segment, build up the
	 segment info for swizzling, while eliminating the bridge objects at the end of each
	 segment that specify the distance to and the size of the subsequent segment."
	<var: #f type: #sqImageFile>
	<inline: false>
	| bytesRead totalBytesRead bridgehead bridge nextSegmentSize oldBase newBase segInfo bridgeSpan |
	<var: 'segInfo' type: #'SpurSegmentInfo *'>
	self allocateOrExtendSegmentInfos.

	"segment sizes include the two-header-word bridge at the end of each segment."
	numSegments := totalBytesRead := 0.
	oldBase := 0. "N.B. still must be adjusted by oldBaseAddr."
	newBase := manager oldSpaceStart.
	nextSegmentSize := firstSegmentSize.
	bridgehead := firstSegmentSize + manager oldSpaceStart - manager bridgeSize.
	[segInfo := self addressOf: (segments at: numSegments).
	 segInfo
		segStart: oldBase;					"N.B. still must be adjusted by oldBaseAddr."
		segSize: nextSegmentSize;
		swizzle: newBase - oldBase.	"N.B. still must be adjusted by oldBaseAddr."
	 bytesRead := self readHeapFrom: f at: newBase dataBytes: nextSegmentSize.
	 bytesRead > 0 ifTrue:
			[totalBytesRead := totalBytesRead + bytesRead].
	 bytesRead ~= nextSegmentSize ifTrue:
		[^totalBytesRead].
	 (numSegments := numSegments + 1) >= numSegInfos ifTrue:
		[self allocateOrExtendSegmentInfos].
	 bridge := bridgehead + manager baseHeaderSize.
	 bridgeSpan := (manager rawNumSlotsOf: bridgehead) = 0
						ifTrue: [0]
						ifFalse: [manager bytesPerOop * (manager rawOverflowSlotsOf: bridge)].
	 oldBase := oldBase + nextSegmentSize + bridgeSpan.
	 newBase := newBase + nextSegmentSize - manager bridgeSize.
	 nextSegmentSize := (manager long64At: bridge) asUnsignedInteger.
	 nextSegmentSize ~= 0] whileTrue:
		[bridgehead := bridgehead - manager bridgeSize + nextSegmentSize].
	"newBase should point just past the last bridge. all others should have been eliminated."
	self assert: newBase - manager oldSpaceStart
				= (totalBytesRead - (numSegments * manager bridgeSize)).
	"set freeOldSpaceStart now for adjustAllOopsBy:"
	manager setFreeOldSpaceStart: newBase.
	"we're done. nil firstSegmentSize for a subsequent snapshot."
	firstSegmentSize := nil.
	^totalBytesRead! !

!SpurSegmentManager methodsFor: 'growing/shrinking memory' stamp: 'cb 10/4/2018 09:37'!
removeSegment: emptySeg
	<var: #emptySeg type: #'SpurSegmentInfo *'>
	| i |
	i := self indexOfSegment: emptySeg.
	self assert: i > 0.

	totalHeapSizeIncludingBridges := totalHeapSizeIncludingBridges - emptySeg segSize.
	manager sqDeallocateMemorySegmentAt: emptySeg segStart asVoidPointer OfSize: emptySeg segSize.

	i to: numSegments - 2 do:
		[:j| segments at: j put: (segments at: j + 1)].
	self cCode: [] inSmalltalk: [segments at: numSegments - 1 put: SpurSegmentInfo new].
	numSegments := numSegments - 1.

	self bridgeFrom: (self addressOf: (segments at: i - 1))
		to: (i <= (numSegments - 1) ifTrue: [self addressOf: (segments at: i)]).

	manager setLastSegment: (self addressOf: (segments at: numSegments - 1))! !

!SpurSegmentManager methodsFor: 'accessing' stamp: 'eem 3/7/2014 17:12'!
segmentContainingObj: objOop
	<export: true>
	<returnTypeC: #'SpurSegmentInfo *'>
	numSegments - 1 to: 0 by: -1 do:
		[:i|
		objOop >= (segments at: i) segStart ifTrue:
			[^self addressOf: (segments at: i)]].
	^nil! !

!SpurSegmentManager methodsFor: 'growing/shrinking memory' stamp: 'cb 5/29/2018 15:28'!
segmentOverlap
	"Answers true if a segment overlaps with another one."
	0 to: numSegments - 1 do:
		[:i| | starti endi |
			starti := (segments at: i) segStart.
			endi := (segments at: i) segLimit.
			0 to: numSegments - 1 do:
				[:j| | startj endj |
					startj := (segments at: j) segStart.
					endj := (segments at: j) segLimit.
					i = j ifFalse: 
						[(starti < startj or: [starti >= endj])
							ifFalse: 
								[self cCode: '' inSmalltalk: [self error: 'segment overlap ' , i printString , ' and ' , j printString].
								 ^true].
						 (endi <= startj or: [endi >= endj])
							ifFalse: [self cCode: '' inSmalltalk: [self error: 'segment overlap ' , i printString , ' and ' , j printString].
								 ^true].
						"self printSegmentAddresses"
						]]].
	^false! !

!SpurSegmentManager methodsFor: 'accessing' stamp: 'cb 4/26/2018 13:16'!
segments
	<cmacro: '() GIV(segments)'>
	^segments! !

!SpurSegmentManager methodsFor: 'growing/shrinking memory' stamp: 'eem 6/21/2021 12:04'!
shrinkObjectMemory: delta
	"Answer if any shrinkage was achieved."
	<inline: false>
	<var: #delta type: #usqInt>
	| shrinkage emptySeg |
	<var: #shrinkage type: #usqInt>
	<var: #emptySeg type: #'SpurSegmentInfo *'>
	manager checkFreeSpace: GCCheckFreeSpace.
	shrinkage := delta.
	[emptySeg := self findEmptySegNearestInSizeTo: shrinkage.
	 (emptySeg isNil
	  or: [emptySeg segSize > shrinkage]) ifTrue:
		[manager checkFreeSpace: GCCheckFreeSpace.
		 ^shrinkage < delta].
	 shrinkage := shrinkage - emptySeg segSize.
	 manager detachFreeObject: (manager objectStartingAt: emptySeg segStart).
	 self removeSegment: emptySeg] repeat! !

!SpurSegmentManager methodsFor: 'pinning' stamp: 'eem 5/21/2014 15:36'!
someSegmentContainsPinned
	0 to: numSegments - 1 do:
		[:i| (segments at: i) containsPinned ifTrue: [^true]].
	^false! !

!SpurSegmentManager methodsFor: 'snapshot' stamp: 'eem 10/12/2013 11:51'!
swizzleObj: objOop
	<inline: false>
	self assert: canSwizzle.
	numSegments - 1 to: 1 by: -1 do:
		[:i|
		objOop >= (segments at: i) segStart ifTrue:
			[^objOop + (segments at: i) swizzle]].
	^objOop + (segments at: 0) swizzle! !

!SpurSegmentManager methodsFor: 'snapshot' stamp: 'eem 11/28/2019 12:36'!
swizzleObj: objOop in: containerOrNil
	"Hook for error reporting in SpurLeakCheckingSegmentManager"
	<inline: #always>
	^self swizzleObj: objOop! !

!SpurSegmentManager methodsFor: 'snapshot' stamp: 'eem 3/7/2014 11:53'!
totalBytesInNonEmptySegments
	| total |
	total := 0.
	0 to: numSegments - 1 do:
		[:i|
		(segments at: i) segSize > manager bridgeSize ifTrue:
			[total := total + (segments at: i) segSize]].
	^total! !

!SpurSegmentManager methodsFor: 'accessing' stamp: 'eem 6/7/2017 12:22'!
totalBytesInSegments
	| total |
	<var: #total type: #usqInt>
	total := 0.
	0 to: numSegments - 1 do:
		[:i|
		total := total + (segments at: i) segSize].
	self assert: totalHeapSizeIncludingBridges = total.
	^total! !

!SpurSegmentManager methodsFor: 'accessing' stamp: 'eem 6/7/2017 13:28'!
totalOldSpaceCapacity
	^totalHeapSizeIncludingBridges - (numSegments * manager bridgeSize)! !

!SpurSegmentManager methodsFor: 'accessing' stamp: 'eem 9/25/2020 09:43'!
totalOldSpaceSize
	^totalHeapSizeIncludingBridges! !

!SpurSegmentManager methodsFor: 'snapshot' stamp: 'eem 3/9/2014 09:14'!
writeImageSegmentsToFile: aBinaryStream
	<var: 'aBinaryStream' type: #sqImageFile>
	<inline: false>
	| total |
	self assert: (manager endOfMemory = (segments at: numSegments - 1) segLimit
				 or: [manager endOfMemory + manager bridgeSize = (segments at: numSegments - 1) segLimit]).
	firstSegmentSize ifNotNil:
		[self assert: firstSegmentSize = (segments at: 0) segSize].
	self assert: (segments at: 0) segSize > 0.
	total := 0.
	0 to: numSegments - 1 do:
		[:i|
		(segments at: i) segSize > manager bridgeSize ifTrue:
			[total := total + (self writeSegment: (self addressOf: (segments at: i))
								nextSegment: (self nextNonEmptySegmentAfter: i)
								toFile: aBinaryStream)]].
	^total! !

!SpurSegmentManager methodsFor: 'snapshot' stamp: 'eem 1/14/2018 11:26'!
writeSegment: segment nextSegment: nextSegment toFile: aBinaryStream
	"Write the segment contents, the size of and the distance to the next segment to aBinaryStream."
	<var: 'segment' type: #'SpurSegmentInfo *'>
	<var: 'nextSegment' type: #'SpurSegmentInfo *'>
	<var: 'aBinaryStream' type: #sqImageFile>
	| pier1 pier2 firstSavedBridgeWord secondSavedBridgeWord nWritten |
	<var: 'firstSavedBridgeWord' type: #usqLong>
	<var: 'secondSavedBridgeWord' type: #usqLong>
	pier1 := segment segLimit - manager bridgeSize.
	pier2 := pier1 + manager baseHeaderSize.
	self assert: (self isValidSegmentBridge: (self bridgeFor: segment)).
	self assert: (manager startOfObject: (self bridgeFor: segment)) = pier1.
	"Temporarily change the bridge to bridge to the next non-empty segment.
	 The first double word of the bridge includes the bridge size in slots, and
	 hence specifies the distance to the next segment. The following double
	 word is replaced by the size of the next segment, or 0 if there isn't one."
	firstSavedBridgeWord := manager long64At: pier1.
	secondSavedBridgeWord := manager long64At: pier2.
	self bridgeFrom: segment to: nextSegment.
	manager
		long64At: pier2
		put: (nextSegment ifNil: [0] ifNotNil: [nextSegment segSize]).
	nWritten := self cCode:
						[self
							sq: segment segStart asVoidPointer
							Image: 1
							File: segment segSize
							Write: aBinaryStream]
					inSmalltalk:
						[| bytesPerElement |
						 bytesPerElement := manager memory bytesPerElement.
						 aBinaryStream
							next: segment segSize / bytesPerElement
							putAll: manager memory
							startingAt: segment segStart / bytesPerElement + 1.
						 segment segSize].
	manager
		long64At: pier1 put: firstSavedBridgeWord;
		long64At: pier2 put: secondSavedBridgeWord.
	^nWritten! !

VMClass subclass: #CogMethodSurrogate
	instanceVariableNames: 'address memory baseHeaderSize cogit'
	classVariableNames: ''
	poolDictionaries: 'CogMethodConstants VMBytecodeConstants'
	category: 'VMMaker-JITSimulation'!

CogMethodSurrogate subclass: #CogBlockMethodSurrogate32
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CogBlockMethodSurrogate32 class methodsFor: 'accessing' stamp: 'eem 10/1/2013 14:06'!
alignedByteSize
	^4 + self baseHeaderSize! !

!CogBlockMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 2/26/2009 16:48'!
cmIsUnlinked
	^(((memory unsignedByteAt: address + 6) bitShift: -4) bitAnd: 16r1) ~= 0! !

!CogBlockMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 2/26/2009 16:48'!
cmIsUnlinked: aValue
	memory
		unsignedByteAt: address + 6
		put: (((memory unsignedByteAt: address + 6) bitAnd: 16rEF) + ((aValue ifTrue: [1] ifFalse: [0]) bitShift: 4)).
	^aValue! !

!CogBlockMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 10/1/2013 14:05'!
padToWord
	^memory unsignedLongAt: address + 5! !

!CogBlockMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
padToWord: aValue
	
	| index delta |
	index := address + 5.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedLongAt: index - delta
			put: aValue].
	^memory
		unsignedLongAt: index
		put: aValue! !

CogBlockMethodSurrogate32 subclass: #CogMethodSurrogate32
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CogMethodSurrogate32 commentStamp: '<historical>' prior: 0!
Here's a doit to generate my code:
	| offset crtab |
	offset := 1.
	crtab := String with: Character cr with: Character tab.
	CogMethod instVarNamesAndTypesForTranslationDo:
		[:name :type| | isByte class |
		(isByte := type = 'unsigned char') ifFalse:
			[offset := (offset - 1 + 3 bitAnd: -4) + 1].
		class := (CogBlockMethod instVarNames includes: name)
					ifTrue: [CogBlockMethodSurrogate32]
					ifFalse: [CogMethodSurrogate32].
		class
			compile: name, crtab, '^memory unsigned', (isByte ifTrue: ['ByteAt:'] ifFalse: ['LongAt:']), ' address + ', offset printString
			classified: #accessing.
		class
			compile: name, ': aValue', crtab, '^memory unsigned', (isByte ifTrue: ['ByteAt:'] ifFalse: ['LongAt:']), ' address + ', offset printString,
					' put: aValue'
			classified: #accessing.
		offset := offset + (isByte ifTrue: [1] ifFalse: [4])].
	CogMethodSurrogate32 class compile: 'alignedByteSize', crtab, '^', (offset - 1 + 3 bitAnd: -4) printString classified: #'instance creation'
!

!CogMethodSurrogate32 class methodsFor: 'accessing' stamp: 'eem 10/1/2013 13:58'!
alignedByteSize
	^20 + self baseHeaderSize! !

!CogMethodSurrogate32 class methodsFor: 'accessing' stamp: 'eem 12/5/2015 14:24'!
offsetOf: aByteSymbol
	"These should be generated!!!!"
	| baseHeaderSize |
	baseHeaderSize := self objectMemoryClass baseHeaderSize.
	^aByteSymbol caseOf:
		{	[#methodObject]		-> [8 + baseHeaderSize].
			[#methodHeader]		-> [12 + baseHeaderSize].
			[#selector]				-> [16 + baseHeaderSize].
			[#blockEntryOffset]	-> [6 + baseHeaderSize].
		}! !

!CogMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 1/28/2020 16:46'!
blockEntryOffset
	^memory unsignedShortAt: address + 7 + baseHeaderSize! !

!CogMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:03'!
blockEntryOffset: aValue
	
	| index delta |
	index := address + baseHeaderSize + 7.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedShortAt: index - delta
			put: aValue].
	^memory
		unsignedShortAt: index
		put: aValue! !

!CogMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 1/28/2020 16:46'!
blockSize
	^memory unsignedShortAt: address + 5 + baseHeaderSize! !

!CogMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:03'!
blockSize: aValue
	
	| index delta |
	index := address + baseHeaderSize + 5.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedShortAt: index - delta
			put: aValue].
	^memory
		unsignedShortAt: index
		put: aValue! !

!CogMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 10/1/2013 13:58'!
methodHeader
	^memory unsignedLongAt: address + 13 + baseHeaderSize! !

!CogMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:03'!
methodHeader: aValue
	
	| index delta |
	index := address + baseHeaderSize + 13.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedLongAt: index - delta
			put: aValue].
	^memory
		unsignedLongAt: index
		put: aValue! !

!CogMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 10/1/2013 13:58'!
methodObject
	^memory unsignedLongAt: address + 9 + baseHeaderSize! !

!CogMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:03'!
methodObject: aValue
	
	| index delta |
	index := address + baseHeaderSize + 9.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedLongAt: index - delta
			put: aValue].
	^memory
		unsignedLongAt: index
		put: aValue! !

!CogMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 10/1/2013 13:58'!
selector
	^memory unsignedLongAt: address + 17 + baseHeaderSize! !

!CogMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:03'!
selector: aValue
	
	| index delta |
	index := address + baseHeaderSize + 17.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedLongAt: index - delta
			put: aValue].
	^memory
		unsignedLongAt: index
		put: aValue! !

CogMethodSurrogate32 subclass: #CogNewspeakMethodSurrogate32
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CogNewspeakMethodSurrogate32 class methodsFor: 'accessing' stamp: 'eem 10/30/2013 16:44'!
alignedByteSize
	^24 + self baseHeaderSize! !

!CogNewspeakMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 7/9/2014 19:37'!
nextMethodOrIRCs
	^memory unsignedLongAt: address + 21 + baseHeaderSize! !

!CogNewspeakMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:06'!
nextMethodOrIRCs: aValue
	
	| index delta |
	index := address + baseHeaderSize + 21.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedLongAt: index - delta
			put: aValue].
	^memory
		unsignedLongAt: index
		put: aValue! !

CogMethodSurrogate32 subclass: #CogSistaMethodSurrogate32
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CogSistaMethodSurrogate32 class methodsFor: 'accessing' stamp: 'eem 7/5/2014 17:45'!
alignedByteSize
	^24 + self baseHeaderSize! !

!CogSistaMethodSurrogate32 class methodsFor: 'accessing' stamp: 'eem 11/6/2014 11:17'!
offsetOf: aByteSymbol
	"These should be generated!!!!"
	^aByteSymbol == #counters
		ifTrue: [20 + self objectMemoryClass baseHeaderSize]
		ifFalse: [super offsetOf: aByteSymbol]! !

!CogSistaMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 7/4/2014 12:02'!
counters
	^memory unsignedLongAt: address + 21 + baseHeaderSize! !

!CogSistaMethodSurrogate32 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
counters: aValue
	
	| index delta |
	index := address + baseHeaderSize + 21.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedLongAt: index - delta
			put: aValue].
	^memory
		unsignedLongAt: index
		put: aValue! !

CogMethodSurrogate subclass: #CogBlockMethodSurrogate64
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CogBlockMethodSurrogate64 class methodsFor: 'accessing' stamp: 'eem 11/3/2015 17:41:12'!
alignedByteSize
	^8 + self baseHeaderSize! !

!CogBlockMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 2/26/2009 16:48'!
cmIsUnlinked
	^(((memory unsignedByteAt: address + 10) bitShift: -4) bitAnd: 16r1) ~= 0! !

!CogBlockMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 2/26/2009 16:48'!
cmIsUnlinked: aValue
	memory
		unsignedByteAt: address + 10
		put: (((memory unsignedByteAt: address + 10) bitAnd: 16rEF) + ((aValue ifTrue: [1] ifFalse: [0]) bitShift: 4)).
	^aValue! !

!CogBlockMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
padToWord
	^memory unsignedLong64At: address + 5! !

!CogBlockMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
padToWord: aValue
	
	| index delta |
	index := address + 5.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedLong64At: index - delta
			put: aValue].
	^memory
		unsignedLong64At: index
		put: aValue! !

CogBlockMethodSurrogate64 subclass: #CogMethodSurrogate64
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CogMethodSurrogate64 commentStamp: '<historical>' prior: 0!
Here's a doit to generate my code:
	| offset crtab |
	offset := 1.
	crtab := String with: Character cr with: Character tab.
	CogMethod instVarNamesAndTypesForTranslationDo:
		[:name :type| | isByte class |
		(isByte := type = 'unsigned char') ifFalse:
			[offset := (offset - 1 + 7 bitAnd: -8) + 1].
		class := (CogBlockMethod instVarNames includes: name)
					ifTrue: [CogBlockMethodSurrogate64]
					ifFalse: [CogMethodSurrogate64].
		class
			compile: name, crtab, '^memory unsigned', (isByte ifTrue: ['ByteAt:'] ifFalse: ['LongLongAt:']), ' address + ', offset printString
			classified: #accessing.
		class
			compile: name, ': aValue', crtab, '^memory unsigned', (isByte ifTrue: ['ByteAt:'] ifFalse: ['LongLongAt:']), ' address + ', offset printString,
					' put: aValue'
			classified: #accessing.
		offset := offset + (isByte ifTrue: [1] ifFalse: [8])].
	CogMethodSurrogate64 class compile: 'alignedByteSize', crtab, '^', (offset + 7 bitAnd: -8) printString classified: #'instance creation'
!

!CogMethodSurrogate64 class methodsFor: 'accessing' stamp: 'eem 10/9/2013 16:00'!
alignedByteSize
	^32 + self baseHeaderSize! !

!CogMethodSurrogate64 class methodsFor: 'accessing' stamp: 'eem 12/5/2015 14:24'!
offsetOf: aByteSymbol
	"These should be generated!!!!"
	| baseHeaderSize |
	baseHeaderSize := self objectMemoryClass baseHeaderSize.
	^aByteSymbol caseOf:
		{	[#methodObject]		-> [8 + baseHeaderSize].
			[#methodHeader]		-> [16 + baseHeaderSize].
			[#selector]				-> [24 + baseHeaderSize].
			[#blockEntryOffset]	-> [6 + baseHeaderSize].
		}! !

!CogMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2013 16:00'!
blockEntryOffset
	^memory unsignedShortAt: address + 7 + baseHeaderSize! !

!CogMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:03'!
blockEntryOffset: aValue
	
	| index delta |
	index := address + baseHeaderSize + 7.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedShortAt: index - delta
			put: aValue].
	^memory
		unsignedShortAt: index
		put: aValue! !

!CogMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 10/9/2013 16:00'!
blockSize
	^memory unsignedShortAt: address + 5 + baseHeaderSize! !

!CogMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:03'!
blockSize: aValue
	
	| index delta |
	index := address + baseHeaderSize + 5.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedShortAt: index - delta
			put: aValue].
	^memory
		unsignedShortAt: index
		put: aValue! !

!CogMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
methodHeader
	^memory unsignedLong64At: address + 17 + baseHeaderSize! !

!CogMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:03'!
methodHeader: aValue
	
	| index delta |
	index := address + baseHeaderSize + 17.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedLong64At: index - delta
			put: aValue].
	^memory
		unsignedLong64At: index
		put: aValue! !

!CogMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
methodObject
	^memory unsignedLong64At: address + 9 + baseHeaderSize! !

!CogMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:03'!
methodObject: aValue
	
	| index delta |
	index := address + baseHeaderSize + 9.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedLong64At: index - delta
			put: aValue].
	^memory
		unsignedLong64At: index
		put: aValue! !

!CogMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
selector
	^memory unsignedLong64At: address + 25 + baseHeaderSize! !

!CogMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:03'!
selector: aValue
	
	| index delta |
	index := address + baseHeaderSize + 25.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedLong64At: index - delta
			put: aValue].
	^memory
		unsignedLong64At: index
		put: aValue! !

CogMethodSurrogate64 subclass: #CogNewspeakMethodSurrogate64
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CogNewspeakMethodSurrogate64 class methodsFor: 'accessing' stamp: 'eem 10/30/2013 16:44'!
alignedByteSize
	^40 + self baseHeaderSize! !

!CogNewspeakMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
nextMethodOrIRCs
	^memory unsignedLong64At: address + 33 + baseHeaderSize! !

!CogNewspeakMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:06'!
nextMethodOrIRCs: aValue
	
	| index delta |
	index := address + baseHeaderSize + 33.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedLong64At: index - delta
			put: aValue].
	^memory
		unsignedLong64At: index
		put: aValue! !

CogMethodSurrogate64 subclass: #CogSistaMethodSurrogate64
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CogSistaMethodSurrogate64 class methodsFor: 'accessing' stamp: 'eem 7/5/2014 17:46'!
alignedByteSize
	^40 + self baseHeaderSize! !

!CogSistaMethodSurrogate64 class methodsFor: 'accessing' stamp: 'eem 11/6/2014 11:17'!
offsetOf: aByteSymbol
	"These should be generated!!!!"
	^aByteSymbol == #counters
		ifTrue: [32 + self objectMemoryClass baseHeaderSize]
		ifFalse: [super offsetOf: aByteSymbol]! !

!CogSistaMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 1/23/2017 13:18'!
counters
	^memory unsignedLong64At: address + 33 + baseHeaderSize! !

!CogSistaMethodSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
counters: aValue
	
	| index delta |
	index := address + baseHeaderSize + 33.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedLong64At: index - delta
			put: aValue].
	^memory
		unsignedLong64At: index
		put: aValue! !

!CogMethodSurrogate class methodsFor: 'accessing' stamp: 'eem 9/2/2011 14:03'!
alignedByteSizeOf: anObject forClient: aVMClass
	^self alignedByteSize! !

!CogMethodSurrogate class methodsFor: 'accessing' stamp: 'eem 10/1/2013 14:09'!
baseHeaderSize
	^self objectMemoryClass baseHeaderSize! !

!CogMethodSurrogate class methodsFor: 'accessing class hierarchy' stamp: 'eem 10/1/2013 14:10'!
defaultObjectMemoryClass
	^NewCoObjectMemory! !

!CogMethodSurrogate methodsFor: 'arithmetic' stamp: 'eem 9/4/2018 17:03'!
+ anInteger
	"N.B. Mimic C semantics so that aCogMethod + 1 is a pointer to the byte past the CogMethod header."
	^address + (anInteger * self class alignedByteSize)! !

!CogMethodSurrogate methodsFor: 'comparing' stamp: 'eem 2/15/2009 23:17'!
< aCogMethodOrAddress
	^address < aCogMethodOrAddress asInteger! !

!CogMethodSurrogate methodsFor: 'comparing' stamp: 'eem 2/15/2009 23:18'!
<= aCogMethodOrAddress
	^address <= aCogMethodOrAddress asInteger! !

!CogMethodSurrogate methodsFor: 'comparing' stamp: 'eem 6/26/2010 09:19'!
= aCogMethodOrAddressOrNil
	^address = (aCogMethodOrAddressOrNil
					ifNil: [0]
					ifNotNil: [aCogMethodOrAddressOrNil asInteger])! !

!CogMethodSurrogate methodsFor: 'comparing' stamp: 'eem 6/15/2009 18:12'!
> aCogMethodOrAddress
	^address > aCogMethodOrAddress asInteger! !

!CogMethodSurrogate methodsFor: 'comparing' stamp: 'eem 3/9/2009 10:22'!
>= aCogMethodOrAddress
	^address >= aCogMethodOrAddress asInteger! !

!CogMethodSurrogate methodsFor: 'arithmetic' stamp: 'eem 2/15/2009 23:20'!
adaptToInteger: rcvr andSend: selector
	"If no method has been provided for adapting an object to a Integer,
	then it may be adequate to simply adapt it to a number."
	^rcvr perform: selector with: address! !

!CogMethodSurrogate methodsFor: 'breakpoints' stamp: 'eem 2/25/2020 11:06'!
addBreakpoint: bkpt
	^{self. bkpt}! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 12/26/2008 11:02'!
address
	^address! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 1/28/2020 11:01'!
asInteger
	"Answer the surrogate's address. This is equivalent to a C cast to sqInt,
	 which is precisely what Slang generates for asInteger"
	^address! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 1/28/2020 11:01'!
asIntegerPtr
	"Answer the surrogate's address. This is equivalent to a C cast to usqIntptr_t,
	 which is precisely what Slang generates for asIntegerPtr"
	^address! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 1/28/2020 11:01'!
asUnsignedInteger
	"Answer the surrogate's address. This is equivalent to a C cast to usqInt,
	 which is precisely what Slang generates for asUnsignedInteger"
	^address! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 1/28/2020 11:01'!
asUnsignedIntegerPtr
	"Answer the surrogate's address. This is equivalent to a C cast to usqIntptr_t,
	 which is precisely what Slang generates for asUnsignedIntegerPtr"
	^address! !

!CogMethodSurrogate methodsFor: 'coercing' stamp: 'eem 4/4/2014 09:53'!
asVoidPointer
	^self! !

!CogMethodSurrogate methodsFor: 'instance initialization' stamp: 'eem 4/15/2016 11:57'!
at: anAddress memory: memoryArray headerSize: headerSize cogit: aCogit
	address := anAddress.
	memory := memoryArray.
	baseHeaderSize := headerSize.
	cogit := aCogit! !

!CogMethodSurrogate methodsFor: 'instance initialization' stamp: 'eem 10/1/2013 10:53'!
at: anAddress objectMemory: objectMemory cogit: aCogit
	address := anAddress.
	memory := objectMemory memory.
	baseHeaderSize := objectMemory baseHeaderSize.
	cogit := aCogit! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 2/26/2009 09:37'!
cPICNumCases
	^self stackCheckOffset! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 2/26/2009 09:38'!
cPICNumCases: n
	^self stackCheckOffset: n! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 1/28/2020 11:00'!
cbUsesInstVars
	^(((memory unsignedByteAt: address + 3 + baseHeaderSize) bitShift: -1) bitAnd: 16r1) ~= 0! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
cbUsesInstVars: aValue
	
	| index delta |
	index := address + baseHeaderSize + 3.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedByteAt: index - delta
			put: (((memory unsignedByteAt: index - delta) bitAnd: 16rFD) + (aValue ifTrue: [2] ifFalse: [0]))].
	memory
		unsignedByteAt: index
		put: (((memory unsignedByteAt: index) bitAnd: 16rFD) + (aValue ifTrue: [2] ifFalse: [0])).
	^aValue! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 1/28/2020 11:00'!
cmHasMovableLiteral
	^(((memory unsignedByteAt: address + 3 + baseHeaderSize) bitShift: -2) bitAnd: 16r1) ~= 0! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
cmHasMovableLiteral: aValue
	
	| index delta |
	index := address + baseHeaderSize + 3.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedByteAt: index - delta
			put: (((memory unsignedByteAt: index - delta) bitAnd: 16rFB) + (aValue ifTrue: [4] ifFalse: [0]))].
	memory
		unsignedByteAt: index
		put: (((memory unsignedByteAt: index) bitAnd: 16rFB) + (aValue ifTrue: [4] ifFalse: [0])).
	^aValue! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 4/14/2016 09:21'!
cmHomeMethod
	^cogit cogMethodSurrogateAt: ((SistaV1BytecodeSet and: [self cmIsFullBlock])
										ifTrue: [address]
										ifFalse: [address - self homeOffset])! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 4/14/2016 10:30'!
cmIsFullBlock
	"Answer the value of cpicHasMNUCaseOrCMIsFullBlock"

	^SistaV1BytecodeSet
		ifTrue: [self cpicHasMNUCaseOrCMIsFullBlock]
		ifFalse: [false]! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 1/28/2020 11:00'!
cmNumArgs
	^memory unsignedByteAt: address + 1 + baseHeaderSize! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
cmNumArgs: aValue
	
	| index delta |
	index := address + baseHeaderSize + 1.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedByteAt: index - delta
			put: aValue].
	^memory
		unsignedByteAt: index
		put: aValue! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 1/28/2020 11:00'!
cmRefersToYoung
	^(((memory unsignedByteAt: address + 2 + baseHeaderSize) bitShift: -3) bitAnd: 16r1) ~= 0! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
cmRefersToYoung: aValue
	
	| index delta |
	index := address + baseHeaderSize + 2.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedByteAt: index - delta
			put: (((memory unsignedByteAt: index - delta) bitAnd: 16rF7) + (aValue ifTrue: [8] ifFalse: [0]))].
	memory
		unsignedByteAt: index
		put: (((memory unsignedByteAt: index) bitAnd: 16rF7) + (aValue ifTrue: [8] ifFalse: [0])).
	^aValue! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 1/28/2020 11:00'!
cmType
	^(memory unsignedByteAt: address + 2 + baseHeaderSize) bitAnd: 16r7! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
cmType: aValue
	
	| index delta |
	index := address + baseHeaderSize + 2.
	self assert: (aValue between: 0 and: 16r7).
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedByteAt: index - delta
			put: ((memory unsignedByteAt: index - delta) bitAnd: 16rF8) + aValue].
	memory
		unsignedByteAt: index
		put: ((memory unsignedByteAt: index) bitAnd: 16rF8) + aValue.
	^aValue! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 1/28/2020 11:00'!
cmUsageCount
	^((memory unsignedByteAt: address + 2 + baseHeaderSize) bitShift: -5) bitAnd: 16r7! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
cmUsageCount: aValue
	
	| index delta |
	index := address + baseHeaderSize + 2.
	self assert: (aValue between: 0 and: 16r7).
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedByteAt: index - delta
			put: ((memory unsignedByteAt: index - delta) bitAnd: 16r1F) + (aValue bitShift: 5)].
	memory
		unsignedByteAt: index
		put: ((memory unsignedByteAt: index) bitAnd: 16r1F) + (aValue bitShift: 5).
	^aValue! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 1/28/2020 11:00'!
cmUsesPenultimateLit
	^((memory unsignedByteAt: address + 3 + baseHeaderSize) bitAnd: 16r1) ~= 0! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
cmUsesPenultimateLit: aValue
	
	| index delta |
	index := address + baseHeaderSize + 3.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedByteAt: index - delta
			put: (((memory unsignedByteAt: index - delta) bitAnd: 16rFE) + (aValue ifTrue: [1] ifFalse: [0]))].
	memory
		unsignedByteAt: index
		put: (((memory unsignedByteAt: index) bitAnd: 16rFE) + (aValue ifTrue: [1] ifFalse: [0])).
	^aValue! !

!CogMethodSurrogate methodsFor: 'testing' stamp: 'eem 9/30/2015 13:27'!
containsAddress: anAddress
	^address <= anAddress asUnsignedInteger
	  and: [address + self blockSize >= anAddress asUnsignedInteger]! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 4/14/2016 10:30'!
cpicHasMNUCase
	"Answer if the receiver has an MNU case."
	<inline: true>

	^SistaV1BytecodeSet
		ifTrue: [self cpicHasMNUCaseOrCMIsFullBlock and: [self cmType = CMClosedPIC]]
		ifFalse: [self cpicHasMNUCaseOrCMIsFullBlock]! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 4/11/2016 17:51'!
cpicHasMNUCase: anObject
	"Set if the receiver has an MNU case."

	^self cpicHasMNUCaseOrCMIsFullBlock: anObject! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 1/28/2020 11:00'!
cpicHasMNUCaseOrCMIsFullBlock
	^(((memory unsignedByteAt: address + 2 + baseHeaderSize) bitShift: -4) bitAnd: 16r1) ~= 0! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
cpicHasMNUCaseOrCMIsFullBlock: aValue
	
	| index delta |
	index := address + baseHeaderSize + 2.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedByteAt: index - delta
			put: (((memory unsignedByteAt: index - delta) bitAnd: 16rEF) + (aValue ifTrue: [16] ifFalse: [0]))].
	memory
		unsignedByteAt: index
		put: (((memory unsignedByteAt: index) bitAnd: 16rEF) + (aValue ifTrue: [16] ifFalse: [0])).
	^aValue! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 1/28/2020 11:00'!
homeOffset
	^memory unsignedShortAt: address + 1! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
homeOffset: aValue
	
	| index delta |
	index := address + 1.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedShortAt: index - delta
			put: aValue].
	^memory
		unsignedShortAt: index
		put: aValue! !

!CogMethodSurrogate methodsFor: 'breakpoints' stamp: 'eem 2/24/2020 21:45'!
isActiveBreakpoint
	^address < cogit methodZone freeStart! !

!CogMethodSurrogate methodsFor: 'breakpoints' stamp: 'eem 2/24/2020 22:01'!
isBreakpointFor: anAddress
	^anAddress >= (address + self class baseHeaderSize)
	  and: [anAddress < (address + self blockSize)]! !

!CogMethodSurrogate methodsFor: 'testing' stamp: 'akg 10/24/2018 15:55'!
isVMSimulationAddress
	"Answer a boolean indicating whether the receiver is a kind of address in the VM simulation"

	^true! !

!CogMethodSurrogate methodsFor: 'comparing' stamp: 'eem 7/1/2010 12:53'!
isZero
	^address = 0! !

!CogMethodSurrogate methodsFor: 'breakpoints' stamp: 'eem 2/25/2020 11:22'!
menuPrompt
	^String streamContents:
		[:s|
		 s
			nextPut: $(;
			nextPutAll: (#('CMFree ' 'CMMethod' 'CMBlock' 'CMClosedPIC' 'CMOpenPIC') at: self cmType);
			space;
			nextPutAll: address hex;
			space.
		 (self cmType = CMMethod or: [self cmType = CMClosedPIC or: [self cmType = CMOpenPIC]]) ifTrue:
			[s nextPutAll: ((cogit objectMemory isBytes: self selector)
							ifTrue: [cogit coInterpreter stringOf: self selector]
							ifFalse: [self selector = cogit objectMemory nilObject
										ifTrue: ['(nil)']
										ifFalse: [self selector hex]])].
		 s nextPut: $)]! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 3/21/2013 15:14'!
nextOpenPIC
	| moField |
	moField := self methodObject.
	^moField ~= 0 ifTrue:
		[cogit cogMethodSurrogateAt: moField]! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 6/22/2009 14:08'!
nextOpenPIC: cogMethodSurrogateOrNil
	^self methodObject: (cogMethodSurrogateOrNil ifNil: [0] ifNotNil: [cogMethodSurrogateOrNil asUnsignedInteger])! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 1/28/2020 16:54'!
objectHeader
	"N.B. This *must* be signed.  We use it for relocation when compacting the code zone."
	^baseHeaderSize = 8
		ifTrue: [memory long64At: address + 1]
		ifFalse: [memory longAt: address + 1]! !

!CogMethodSurrogate methodsFor: 'accessing' stamp: 'eem 2/25/2020 19:05'!
objectHeader: aValue
	| delta |
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 baseHeaderSize = 8
			ifTrue: [memory long64At: address - delta + 1 put: aValue]
			ifFalse: [memory longAt: address - delta + 1 put: aValue]].
	^baseHeaderSize = 8
		ifTrue: [memory long64At: address + 1 put: aValue]
		ifFalse: [memory longAt: address + 1 put: aValue]! !

!CogMethodSurrogate methodsFor: 'printing' stamp: 'eem 1/23/2009 13:09'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: ' @'; print: address; nextPut: $/; nextPutAll: address hex! !

!CogMethodSurrogate methodsFor: 'breakpoints' stamp: 'eem 2/24/2020 21:56'!
removeBreakpoint: address
	^nil! !

!CogMethodSurrogate methodsFor: 'breakpoints' stamp: 'eem 2/24/2020 21:36'!
singleStepRequiredToTriggerIn: aCogit
	^true! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 1/28/2020 11:00'!
stackCheckOffset
	^((memory unsignedShortAt: address + 3 + baseHeaderSize) bitShift: -4) bitAnd: 16rFFF! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
stackCheckOffset: aValue
	
	| index delta |
	index := address + baseHeaderSize + 3.
	self assert: (aValue between: 0 and: 16rFFF).
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedShortAt: index - delta
			put: ((memory unsignedShortAt: index - delta) bitAnd: 16rF) + (aValue bitShift: 4)].
	memory
		unsignedShortAt: index
		put: ((memory unsignedShortAt: index) bitAnd: 16rF) + (aValue bitShift: 4).
	^aValue! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 1/28/2020 11:00'!
startpc
	^memory unsignedShortAt: address + 3! !

!CogMethodSurrogate methodsFor: 'accessing generated' stamp: 'eem 2/25/2020 19:04'!
startpc: aValue
	
	| index delta |
	index := address + 3.
	(delta := cogit getCodeToDataDelta) > 0 ifTrue:
		[self assert: (cogit addressIsInCodeZone: address - delta).
		 memory
			unsignedShortAt: index - delta
			put: aValue].
	^memory
		unsignedShortAt: index
		put: aValue! !

VMClass subclass: #InLineLiteralsManager
	instanceVariableNames: 'cogit objectMemory objectRepresentation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!InLineLiteralsManager commentStamp: 'eem 6/7/2015 12:07' prior: 0!
An InLineLiteralsManager is a dummy class that understands the OutOfLineLiteralsManager API but does nothing.  It is used to allow the Cogits to work with back-ends that generate either in-line or out-of-line literals.!

!InLineLiteralsManager class methodsFor: 'translation' stamp: 'eem 6/16/2015 10:25'!
isNonArgumentImplicitReceiverVariableName: aString
	^Cogit isNonArgumentImplicitReceiverVariableName: aString! !

!InLineLiteralsManager class methodsFor: 'translation' stamp: 'eem 1/23/2020 14:46'!
translationClasses
	^{self}! !

!InLineLiteralsManager methodsFor: 'initialization' stamp: 'eem 6/8/2015 17:06'!
allocateLiterals: initialNumLiterals
	<inline: true>! !

!InLineLiteralsManager methodsFor: 'generate machine code' stamp: 'eem 6/11/2015 16:05'!
assertValidAnnotation: annotation for: instruction
	<var: #instruction type: #'AbstractInstruction *'>
	<inline: true>! !

!InLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 12/17/2019 14:12'!
checkLiteral32: literal forInstruction: anInstruction
	<var: #anInstruction type: #'AbstractInstruction *'>
	<inline: true>
	^anInstruction! !

!InLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 6/8/2015 11:20'!
checkLiteral: literal forInstruction: anInstruction
	<var: #anInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	^anInstruction! !

!InLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 6/8/2015 11:20'!
checkQuickConstant: literal forInstruction: anInstruction
	<var: #anInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	^anInstruction! !

!InLineLiteralsManager methodsFor: 'initialization' stamp: 'eem 11/18/2015 17:49'!
cogit: aCogit
	<doNotGenerate>
	cogit := aCogit.
	objectMemory := aCogit objectMemory.
	objectRepresentation := aCogit objectRepresentation.! !

!InLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 6/11/2015 14:12'!
dumpLiterals: generateBranchAround
	<inline: true>! !

!InLineLiteralsManager methodsFor: 'closed PIC parsing' stamp: 'tpr 10/15/2015 17:01'!
endSizeOffset
	"return the offset need from the cPICEndSize in order to point to just after the last instruction - here that means 0"
	^0! !

!InLineLiteralsManager methodsFor: 'garbage collection' stamp: 'eem 6/11/2015 16:35'!
fetchLiteralAtAnnotatedAddress: address using: instruction
	"Normally literals are embedded in instructions and the annotation is at the start of
	 the following instruction, to cope with literals embedded in variable-length instructions,
	 since, e.g. on x86, the literal typically comes at the end of the instruction."
	<var: 'instruction' type: #'AbstractInstruction *'>
	<inline: true>
	^instruction literalBeforeFollowingAddress: address! !

!InLineLiteralsManager methodsFor: 'garbage collection' stamp: 'eem 01/10/2016 10:01'!
literalBytesFollowingBranchInClosedPIC
	^0! !

!InLineLiteralsManager methodsFor: 'garbage collection' stamp: 'eem 01/10/2016 10:01'!
literalBytesFollowingJumpInClosedPIC
	^0! !

!InLineLiteralsManager methodsFor: 'testing' stamp: 'eem 6/11/2015 13:15'!
mustDumpLiterals: currentOpcodeIndex
	<inline: true>
	^false! !

!InLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 03/04/2017 09:43'!
resetForRecompile! !

!InLineLiteralsManager methodsFor: 'initialization' stamp: 'eem 01/10/2016 10:01'!
resetLiterals! !

!InLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 03/04/2017 09:43'!
saveForRecompile! !

!InLineLiteralsManager methodsFor: 'garbage collection' stamp: 'eem 2/24/2020 09:16'!
storeLiteral: literal atAnnotatedAddress: address using: instruction
	"Normally literals are embedded in instructions and the annotation is at the start of
	 the following instruction, to cope with literals embedded in variable-length instructions,
	 since, e.g. on x86, the literal typically comes at the end of the instruction."
	<var: 'address' type: #usqInt>
	<var: 'instruction' type: #'AbstractInstruction *'>
	<inline: #always>
	^instruction storeLiteral: literal beforeFollowingAddress: address! !

!InLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 12/17/2019 13:46'!
uniqueLiteral32: literal forInstruction: anInstruction
	<var: #anInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	^anInstruction! !

!InLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 6/10/2015 19:38'!
uniqueLiteral: literal forInstruction: anInstruction
	<var: #anInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	^anInstruction! !

VMClass subclass: #InterpreterPlugin
	instanceVariableNames: 'interpreterProxy moduleName translatedMethodCache'
	classVariableNames: ''
	poolDictionaries: 'VMBasicConstants'
	category: 'VMMaker-Plugins'!

!InterpreterPlugin commentStamp: 'tpr 5/5/2003 11:43' prior: 0!
This class provides the basic framework for creating VM plugins. Most of the useful methods are on the class side; particularly take note of the messages like #shouldBeTranslated and #requiresPlatformFiles.!

InterpreterPlugin subclass: #ADPCMCodecPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!ADPCMCodecPlugin commentStamp: 'tpr 3/24/2004 14:48' prior: 0!
This is a kludgy interface to the translated primitive code in ADPCMCodec. The translate.... method gathers the code and writes it to a file. No methods in this class actually implement the codec.!

!ADPCMCodecPlugin class methodsFor: 'translation' stamp: 'eem 2/16/2015 18:51'!
methodsToBePruned
	"Since sharing between plugins means that normal pruning
	 can't be done, allow plugins that want to prune specific methods."
	^#(indexForDeltaFrom:to: nextBits: nextBits:put:)! !

!ADPCMCodecPlugin class methodsFor: 'translation' stamp: 'eem 1/3/2013 11:42'!
monticelloDescription
	"Override to include the ADPCMCodec class."
	"self monticelloDescription"
	^super monticelloDescription, '\' withCRs, (CCodeGenerator monticelloDescriptionFor: ADPCMCodec)! !

!ADPCMCodecPlugin class methodsFor: 'translation' stamp: 'eem 2/16/2015 14:19'!
translatedPrimitives
	"Answer an Array of Class, selector pair Arrays for any primitives to
	 be translated from Smalltalk methods elsewhere in the system."
	^ADPCMCodec translatedPrimitives! !

InterpreterPlugin subclass: #B3DAcceleratorPlugin
	instanceVariableNames: 'doRangeChecks'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!B3DAcceleratorPlugin commentStamp: '<historical>' prior: 0!
B3DAcceleratorPlugin translate!

!B3DAcceleratorPlugin class methodsFor: 'translation' stamp: 'tpr 5/23/2001 17:08'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!B3DAcceleratorPlugin class methodsFor: 'translation' stamp: 'ar 4/19/2001 02:21'!
moduleName
	^'B3DAcceleratorPlugin'! !

!B3DAcceleratorPlugin class methodsFor: 'translation' stamp: 'tpr 7/4/2001 15:12'!
requiresCrossPlatformFiles
	"default is ok for most, any plugin needing platform specific files must say so"
	^true! !

!B3DAcceleratorPlugin class methodsFor: 'translation' stamp: 'tpr 3/20/2001 12:16'!
requiresPlatformFiles
	"default is ok for most, any plugin needing platform specific files must say so"
	^true! !

!B3DAcceleratorPlugin class methodsFor: 'testing' stamp: 'eem 5/1/2017 10:22'!
testLoadPlugin
	"B3DAcceleratorPlugin testLoadPlugin"
	<primitive: 'primitiveRendererVersion' module: 'B3DAcceleratorPlugin' error: ec>
	self primitiveFailed! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dDrawArrays: handle _: mode _: minIdx _: maxIdx
	"int b3dDrawArrays(int handle, int mode, int minIdx, int maxIdx)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dDrawElements: handle _: mode _: faceSize _: facePtr
	"int b3dDrawElements(int handle, int mode, int nFaces, unsigned int *facePtr)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dDrawRangeElements: handle _: mode _: minIdx _: maxIdx _: faceSize _: facePtr
	"int b3dDrawRangeElements(int handle, int mode, int minIdx, int maxIdx, int nFaces, unsigned int *facePtr)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dLoadClientState: handle _: vtxData _: vtxSize _: colorData _: colorSize _: normalData _: normalSize _: txData _: txSize
	"int b3dLoadClientState(int handle, float *vtxData, int vtxSize, float *colorData, int colorSize, float *normalData, int normalSize, float *txData, int txSize)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxActualTextureDepth: renderer _: handle
	"int b3dxActualTextureDepth(int renderer, int handle)"
	<doNotGenerate>
	^-1! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxAllocateTexture: renderer _: w _: h _: d
	"int b3dxAllocateTexture(int renderer, int w, int h, int d)"
	<doNotGenerate>
	^-1! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxCompositeTexture: rendererHandle _: texHandle _: x _: y _: w _: h _: translucent
	"int b3dxCompositeTexture(int renderer, int handle, int x, int y, int w, int h, int translucent)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxDestroyRenderer: handle
	"int b3dxDestroyRenderer(int handle)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxDestroyTexture: renderer _: handle
	"int b3dxDestroyTexture(int renderer, int handle)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxDisableLights: handle
	"int b3dxDisableLights(int handle)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxFinishRenderer: handle
	"int b3dxFinishRenderer(int handle)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxFlushRenderer: handle
	"int b3dxFlushRenderer(int handle)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxGetRendererColorMasks: handle _: masks
	"int b3dxGetRendererColorMasks(int handle, int *masks)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxGetRendererSurfaceDepth: handle
	"int b3dxGetRendererSurfaceDepth(int handle)"
	<doNotGenerate>
	^-1! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxGetRendererSurfaceHandle: handle
	"int b3dxGetRendererSurfaceHandle(int handle)"
	<doNotGenerate>
	^-1! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxGetRendererSurfaceHeight: handle
	"int b3dxGetRendererSurfaceHeight(int handle)"
	<doNotGenerate>
	^-1! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxGetRendererSurfaceWidth: handle
	"int b3dxGetRendererSurfaceWidth(int handle)"
	<doNotGenerate>
	^-1! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxIsOverlayRenderer: handle
	"int b3dxIsOverlayRenderer(int handle)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxLoadLight: handle _: i _: light
	"int b3dxLoadLight(int handle, int index, B3DPrimitiveLight *light)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxLoadMaterial: handle _: material
	"int b3dxLoadMaterial(int handle, B3DPrimitiveMaterial *material)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxRenderVertexBuffer: handle _: primType _: flags _: texHandle _: vtxArray _: vtxCount _: idxArray _: idxCount
	"int b3dxRenderVertexBuffer(int handle, int primType, int flags, int texHandle, float *vtxArray, int vtxSize, int *idxArray, int idxSize)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxSwapRendererBuffers: handle
	"int b3dxSwapRendererBuffers(int handle)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxTextureByteSex: renderer _: handle
	"int b3dxTextureByteSex(int renderer, int handle)"
	<doNotGenerate>
	^-1! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxTextureColorMasks: renderer _: handle _: masks
	"int b3dxTextureColorMasks(int renderer, int handle, int masks[4])"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxTextureSurfaceHandle: renderer _: handle
	"int b3dxTextureSurfaceHandle(int renderer, int handle)"
	<doNotGenerate>
	^-1! !

!B3DAcceleratorPlugin methodsFor: 'simulation' stamp: 'eem 10/30/2018 07:48'!
b3dxUploadTexture: renderer _: handle _: w _: h _: d _: bitsPtr
	"int b3dxUploadTexture(int renderer, int handle, int w, int h, int d, void* bits)"
	<doNotGenerate>
	^false! !

!B3DAcceleratorPlugin methodsFor: 'primitives-qwaq' stamp: 'ar (auto pragmas 12/08) 12/5/2008 14:34'!
checkBoundsRange: vertices faces: facePtr count: faceCount
	"Verify the bounds condition on the entire faces array"
	| vtxSize |
	<var: #facePtr type: 'unsigned int *'>
	vtxSize := (interpreterProxy slotSizeOf: vertices) / 3.
	0 to: faceCount-1 do:[:i|
		(facePtr at: i) > vtxSize
			ifTrue:[^interpreterProxy primitiveFail]].
! !

!B3DAcceleratorPlugin methodsFor: 'primitives-qwaq' stamp: 'ar (auto pragmas 12/08) 12/5/2008 12:33'!
checkVertexData: vertices
	"Check the entire vertex data to ensure no nan/inf values"
	| vtxSize vtxPtr f |
	<var: #vtxPtr type: 'float *'>
	<var: #f type: 'float'>
	vtxSize := interpreterProxy slotSizeOf: vertices.
	vtxPtr := interpreterProxy firstIndexableField: vertices.
	0 to: vtxSize-1 do:[:i|
		f := vtxPtr at: i.
		(self isnan: (f-f)) ifTrue:[^interpreterProxy primitiveFail].
	].
! !

!B3DAcceleratorPlugin methodsFor: 'primitive support' stamp: 'ar (auto pragmas 12/08) 9/6/2000 22:14'!
fetchLightSource: index ofObject: anArray
	"Fetch the primitive light source from the given array.
	Note: No checks are done within here - that happened in stackLightArrayValue:"
	| lightOop |
	<inline: true>
	<returnTypeC:'void*'>
	lightOop := interpreterProxy fetchPointer: index ofObject: anArray.
	^interpreterProxy firstIndexableField: lightOop! !

!B3DAcceleratorPlugin methodsFor: 'initialize-release' stamp: 'ar (auto pragmas 12/08) 12/5/2008 12:35'!
initialiseModule
	<export: true>
	doRangeChecks := true.
	^self b3dxInitialize! !

!B3DAcceleratorPlugin methodsFor: 'primitives-qwaq' stamp: 'eem 10/29/2018 17:56'!
loadClientState: handle vertices: vertices colors: colors normals: normals texCoords: texCoords
	"Common method to set up client state for some render ops"
	| nilOop vtxSize sz colorPtr normalPtr txPtr vertexPtr ok |
	<var: #colorPtr type: 'void *'>
	<var: #normalPtr type: 'void *'>
	<var: #txPtr type: 'void *'>
	<var: #vertexPtr type: 'void *'>

	colorPtr := normalPtr := txPtr := vertexPtr := nil.
	sz := 0.

	"Verify vertex data"
	(interpreterProxy isWords: vertices) ifFalse:
		[^interpreterProxy primitiveFail].
	vtxSize := (interpreterProxy slotSizeOf: vertices) / 3.

	"Verify assumptions of color, normal, texCoords data"
	nilOop := interpreterProxy nilObject.
	(colors = nilOop
	 or: [(interpreterProxy isWords: colors)
		and: [(interpreterProxy slotSizeOf: colors) = (vtxSize * 4)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	(normals = nilOop
	 or: [(interpreterProxy isWords: normals)
		and: [(interpreterProxy slotSizeOf: normals) = (vtxSize * 3)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	"Don't check size for texCoords since they can be 2,3,4 elements"
	(texCoords = nilOop
	 or: [(interpreterProxy isWords: texCoords)]) ifFalse:
		[^interpreterProxy primitiveFail].

	"Finally submit the data to OpenGL"
	colors = nilOop ifFalse:
		[colorPtr := interpreterProxy firstIndexableField: colors].
	normals = nilOop ifFalse:
		[normalPtr := interpreterProxy firstIndexableField: normals].
	texCoords = nilOop ifFalse:
		[sz := (interpreterProxy slotSizeOf: texCoords) / vtxSize.
		txPtr := interpreterProxy firstIndexableField: texCoords].
	vertexPtr := interpreterProxy firstIndexableField: vertices.
	interpreterProxy failed ifFalse:
		[ok := self b3dLoadClientState: handle _: vertexPtr _: 3 _: colorPtr _: 4 _: normalPtr _: 3 _: txPtr _: sz.
		 ok ifFalse: [interpreterProxy primitiveFail]].
	^nil "keep compiler quiet"
! !

!B3DAcceleratorPlugin methodsFor: 'primitives-textures' stamp: 'eem 10/29/2018 17:42'!
primitiveAllocateTexture
	| h w d result renderer |
	<export: true>
	interpreterProxy methodArgumentCount = 4
		ifFalse:[^interpreterProxy primitiveFail].
	h := interpreterProxy stackIntegerValue: 0.
	w := interpreterProxy stackIntegerValue: 1.
	d := interpreterProxy stackIntegerValue: 2.
	renderer := interpreterProxy stackIntegerValue: 3.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxAllocateTexture: renderer _: w _: h _: d.
	result = -1 ifTrue:[^interpreterProxy primitiveFail].
	interpreterProxy pop: 5. "args+rcvr"
	^interpreterProxy pushInteger: result.! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 9/19/2018 14:23'!
primitiveClearDepthBuffer
	| handle |
	<export: true>
	interpreterProxy methodArgumentCount = 1 ifFalse:
		[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifFalse:
		[(self b3dxClearDepthBuffer: handle)
			ifTrue: [interpreterProxy pop: 1] "pop args; return rcvr"
			ifFalse:[interpreterProxy primitiveFail]]! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 9/19/2018 14:23'!
primitiveClearViewport
	| handle pv rgba |
	<export: true>
	interpreterProxy methodArgumentCount = 3 ifFalse:
		[^interpreterProxy primitiveFail].
	pv := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	rgba := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 1).
	handle := interpreterProxy stackIntegerValue: 2.
	interpreterProxy failed ifTrue:[^nil].
	(self b3dxClearViewport: handle _: rgba _: pv)
		ifTrue: [interpreterProxy pop: 3] "pop args; return rcvr"
		ifFalse:[interpreterProxy primitiveFail]! !

!B3DAcceleratorPlugin methodsFor: 'primitives-textures' stamp: 'eem 10/29/2018 17:42'!
primitiveCompositeTexture
	| result translucent y x w h texHandle rendererHandle |
	<export: true>
	interpreterProxy methodArgumentCount = 7
		ifFalse:[^interpreterProxy primitiveFail].
	translucent := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
	h := interpreterProxy stackIntegerValue: 1.
	w := interpreterProxy stackIntegerValue: 2.
	y := interpreterProxy stackIntegerValue: 3.
	x := interpreterProxy stackIntegerValue: 4.
	texHandle := interpreterProxy stackIntegerValue: 5.
	rendererHandle := interpreterProxy stackIntegerValue: 6.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxCompositeTexture: rendererHandle _: texHandle _: x _: y _: w _: h _: translucent.
	result ifFalse:[^interpreterProxy primitiveFail].
	^interpreterProxy pop: 7. "args"
! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 9/19/2018 14:25'!
primitiveCreateRenderer
	"NOTE: This primitive is obsolete but should be supported for older images"
	| h w y x result allowHardware allowSoftware |
	<export: true>
	interpreterProxy methodArgumentCount = 6 ifFalse:
		[^interpreterProxy primitiveFail].
	h := interpreterProxy stackIntegerValue: 0.
	w := interpreterProxy stackIntegerValue: 1.
	y := interpreterProxy stackIntegerValue: 2.
	x := interpreterProxy stackIntegerValue: 3.
	allowHardware := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 4).
	allowSoftware := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 5).
	interpreterProxy failed ifFalse:
		[(self b3dxCreateRenderer: allowSoftware _: allowHardware _: x _: y _: w _: h) >= 0
			ifTrue: [interpreterProxy methodReturnInteger: result]
			ifFalse: [interpreterProxy primitiveFail]]! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 9/19/2018 14:27'!
primitiveCreateRendererFlags
	| flags h w y x result  |
	<export: true>
	interpreterProxy methodArgumentCount = 5 ifFalse:
		[^interpreterProxy primitiveFail].
	h := interpreterProxy stackIntegerValue: 0.
	w := interpreterProxy stackIntegerValue: 1.
	y := interpreterProxy stackIntegerValue: 2.
	x := interpreterProxy stackIntegerValue: 3.
	flags := interpreterProxy stackIntegerValue: 4.
	interpreterProxy failed ifFalse:
		[(self b3dxCreateRendererFlags: x _: y _: w _: h _: flags) >= 0
			ifTrue: [interpreterProxy methodReturnInteger: result]
			ifFalse: [interpreterProxy primitiveFail]]! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 10/29/2018 17:42'!
primitiveDestroyRenderer
	| handle result |
	<export: true>
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxDestroyRenderer: handle.
	result ifFalse:[^interpreterProxy primitiveFail].
	^interpreterProxy pop: 1. "pop arg; return rcvr"! !

!B3DAcceleratorPlugin methodsFor: 'primitives-textures' stamp: 'eem 10/29/2018 17:42'!
primitiveDestroyTexture
	| handle result renderer |
	<export: true>
	interpreterProxy methodArgumentCount = 2
		ifFalse:[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	renderer := interpreterProxy stackIntegerValue: 1.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxDestroyTexture: renderer _: handle.
	result ifFalse:[^interpreterProxy primitiveFail].
	^interpreterProxy pop: 2. "pop arg; return rcvr"! !

!B3DAcceleratorPlugin methodsFor: 'primitives-qwaq' stamp: 'eem 10/29/2018 17:57'!
primitiveDrawArrays
	"Primitive. Setup non-VBO client state and call drawRangeElements in one go to avoid
	garbage collection to move the buffers underneith."
	| maxIdx minIdx mode texCoords normals colors vertices handle vtxSize ok |
	<export: true>

	interpreterProxy methodArgumentCount = 8 ifFalse:
		[^interpreterProxy primitiveFail].

	maxIdx := interpreterProxy stackIntegerValue: 0.
	minIdx := interpreterProxy stackIntegerValue: 1.
	mode := interpreterProxy stackIntegerValue: 2.
	texCoords := interpreterProxy stackValue: 3.
	normals := interpreterProxy stackValue: 4.
	colors := interpreterProxy stackValue: 5.
	vertices := interpreterProxy stackValue: 6.
	handle := interpreterProxy stackIntegerValue: 7.

	self loadClientState: handle vertices: vertices colors: colors normals: normals texCoords: texCoords.
	interpreterProxy failed ifTrue:[^nil].
	doRangeChecks ifTrue:
		["Verify the vertex data itself"
		self checkVertexData: vertices.

		"Verify min-max range in bounds for given vertex array"
		vtxSize := (interpreterProxy slotSizeOf: vertices) / 3.
		(minIdx < 0 or:[minIdx > maxIdx or:[maxIdx > vtxSize]]) ifTrue:
			[interpreterProxy primitiveFail]].
	interpreterProxy failed ifFalse:
		[ok := self b3dDrawArrays: handle _: mode _: minIdx _: maxIdx.
		 ok ifTrue:
			[interpreterProxy pop: interpreterProxy methodArgumentCount]].
	^nil "keep compiler quiet"
! !

!B3DAcceleratorPlugin methodsFor: 'primitives-qwaq' stamp: 'eem 10/29/2018 17:57'!
primitiveDrawElements
	"Primitive. Setup non-VBO client state and call drawElements in one go to avoid
	garbage collection to move the buffers underneith."
	| faces mode texCoords normals colors vertices handle ok facePtr faceSize |
	<export: true>
	<var: #facePtr type: 'unsigned int *'>

	interpreterProxy methodArgumentCount = 7 ifFalse:
		[^interpreterProxy primitiveFail].

	faces := interpreterProxy stackValue: 0.
	(interpreterProxy isWords: faces) ifFalse:
		[^interpreterProxy primitiveFail].
	faceSize := interpreterProxy slotSizeOf: faces.
	facePtr := interpreterProxy firstIndexableField: faces.

	mode := interpreterProxy stackIntegerValue: 1.
	texCoords := interpreterProxy stackValue: 2.
	normals := interpreterProxy stackValue: 3.
	colors := interpreterProxy stackValue: 4.
	vertices := interpreterProxy stackValue: 5.
	handle := interpreterProxy stackIntegerValue: 6.

	self loadClientState: handle vertices: vertices colors: colors normals: normals texCoords: texCoords.
	interpreterProxy failed ifTrue:[^nil].

	doRangeChecks ifTrue:
		["Verify the vertex data itself"
		self checkVertexData: vertices.
		"Change bounds range to make sure the data is valid"
		self checkBoundsRange: vertices faces: facePtr count: faceSize].

	interpreterProxy failed ifFalse:
		[ok := self b3dDrawElements: handle _: mode _: faceSize _: facePtr.
		 ok ifTrue:
			[interpreterProxy pop: interpreterProxy methodArgumentCount]].
	^nil "keep compiler quiet"
! !

!B3DAcceleratorPlugin methodsFor: 'primitives-qwaq' stamp: 'eem 10/29/2018 17:59'!
primitiveDrawRangeElements
	"Primitive. Setup non-VBO client state and call drawRangeElements in one go to avoid
	garbage collection to move the buffers underneith."
	| faces maxIdx minIdx mode texCoords normals colors vertices handle vtxSize ok facePtr faceSize |
	<export: true>
	<var: #facePtr type: 'unsigned int *'>

	interpreterProxy methodArgumentCount = 9 ifFalse:
		[^interpreterProxy primitiveFail].

	faces := interpreterProxy stackValue: 0.
	(interpreterProxy isWords: faces) ifFalse:
		[^interpreterProxy primitiveFail].
	faceSize := interpreterProxy slotSizeOf: faces.
	facePtr := interpreterProxy firstIndexableField: faces.

	maxIdx := interpreterProxy stackIntegerValue: 1.
	minIdx := interpreterProxy stackIntegerValue: 2.
	mode := interpreterProxy stackIntegerValue: 3.
	texCoords := interpreterProxy stackValue: 4.
	normals := interpreterProxy stackValue: 5.
	colors := interpreterProxy stackValue: 6.
	vertices := interpreterProxy stackValue: 7.
	handle := interpreterProxy stackIntegerValue: 8.

	self loadClientState: handle vertices: vertices colors: colors normals: normals texCoords: texCoords.
	interpreterProxy failed ifTrue:[^nil].

	doRangeChecks ifTrue:
		["Verify the vertex data itself"
		self checkVertexData: vertices.
		"Change bounds range to make sure the data is valid"
		self checkBoundsRange: vertices faces: facePtr count: faceSize.
		"Verify min-max range in bounds for given vertex array"
		vtxSize := (interpreterProxy slotSizeOf: vertices) / 3.
		(minIdx < 0 or:[minIdx > maxIdx or:[maxIdx > vtxSize]]) ifTrue:
			[interpreterProxy primitiveFail]].

	interpreterProxy failed ifFalse:
		[ok := self b3dDrawRangeElements: handle _: mode _: minIdx _: maxIdx _: faceSize _: facePtr.
		ok ifTrue:
			[interpreterProxy pop: interpreterProxy methodArgumentCount]].
	^nil "keep compiler quiet"
! !

!B3DAcceleratorPlugin methodsFor: 'primitives-qwaq' stamp: 'eem 1/30/2015 17:10'!
primitiveEnableDrawRangeChecks
	"Primitive. Enable/disable draw (range) checks"
	| enabled |
	<export: true>
	interpreterProxy methodArgumentCount = 0 ifTrue:
		[interpreterProxy pop: 1.
		^interpreterProxy pushBool: doRangeChecks].
	interpreterProxy methodArgumentCount = 1 ifTrue:
		[enabled := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
		 interpreterProxy failed ifFalse:
			[doRangeChecks := enabled.
			^interpreterProxy pop: 1]]. "pop arg; return rcvr"
	^nil "keep compiler quiet"! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 10/29/2018 17:42'!
primitiveFinishRenderer
	| handle result |
	<export: true>
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxFinishRenderer: handle.
	result ifFalse:[^interpreterProxy primitiveFail].
	^interpreterProxy pop: 1. "pop arg; return rcvr"! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 10/29/2018 17:42'!
primitiveFlushRenderer
	| handle result |
	<export: true>
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxFlushRenderer: handle.
	result ifFalse:[^interpreterProxy primitiveFail].
	^interpreterProxy pop: 1. "pop arg; return rcvr"! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 9/19/2018 14:32'!
primitiveGetIntProperty
	| handle prop result |
	<export: true>
	interpreterProxy methodArgumentCount = 2 ifFalse:
		[^interpreterProxy primitiveFail].
	prop := interpreterProxy stackIntegerValue: 0.
	handle := interpreterProxy stackIntegerValue: 1.
	result := self b3dxGetIntProperty: handle _: prop.
	interpreterProxy methodReturnInteger: result! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 10/29/2018 17:42'!
primitiveGetRendererColorMasks
	| handle masks array |
	<export: true>
	<var: #masks declareC:'unsigned int masks[4]'>
	interpreterProxy methodArgumentCount = 2
		ifFalse:[^interpreterProxy primitiveFail].
	array := interpreterProxy stackObjectValue: 0.
	handle := interpreterProxy stackIntegerValue: 1.
	interpreterProxy failed ifTrue:[^nil].
	((interpreterProxy isArray: array)
	and: [(interpreterProxy slotSizeOf: array) = 4
	and: [self b3dxGetRendererColorMasks: handle _: masks]])
		ifFalse:[^interpreterProxy primitiveFail].
	interpreterProxy pushRemappableOop: array.
	0 to: 3 do: [:i|
		interpreterProxy
			storePointer: i
			ofObject: interpreterProxy topRemappableOop
			withValue: (interpreterProxy positive32BitIntegerFor: (masks at: i))].
	interpreterProxy popRemappableOop.
	^interpreterProxy pop: 2 "pop args return receiver"! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 10/29/2018 17:42'!
primitiveGetRendererSurfaceDepth
	| handle result |
	<export: true>
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxGetRendererSurfaceDepth: handle.
	result < 0 ifTrue:[^interpreterProxy primitiveFail].
	interpreterProxy pop: 2. "args+rcvr"
	^interpreterProxy pushInteger: result! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 10/29/2018 17:42'!
primitiveGetRendererSurfaceHandle
	| handle result |
	<export: true>
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxGetRendererSurfaceHandle: handle.
	result < 0 ifTrue:[^interpreterProxy primitiveFail].
	interpreterProxy pop: 2. "args+rcvr"
	^interpreterProxy pushInteger: result! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 10/29/2018 17:42'!
primitiveGetRendererSurfaceHeight
	| handle result |
	<export: true>
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxGetRendererSurfaceHeight: handle.
	result < 0 ifTrue:[^interpreterProxy primitiveFail].
	interpreterProxy pop: 2. "args+rcvr"
	^interpreterProxy pushInteger: result! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 10/29/2018 17:42'!
primitiveGetRendererSurfaceWidth
	| handle result |
	<export: true>
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxGetRendererSurfaceWidth: handle.
	result < 0 ifTrue:[^interpreterProxy primitiveFail].
	interpreterProxy pop: 2. "args+rcvr"
	^interpreterProxy pushInteger: result! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 10/29/2018 17:42'!
primitiveIsOverlayRenderer
	| handle result |
	<export: true>
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxIsOverlayRenderer: handle.
	interpreterProxy pop: 2. "args+rcvr"
	^interpreterProxy pushBool: result.! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 10/29/2018 17:42'!
primitiveRenderVertexBuffer
	| idxCount vtxCount vtxArray idxArray texHandle primType result flags handle |
	<export: true>
	<var: #idxArray type: 'int *'>
	<var: #vtxArray type: 'float *'>

	interpreterProxy methodArgumentCount = 8
		ifFalse:[^interpreterProxy primitiveFail].
	idxCount := interpreterProxy stackIntegerValue: 0.
	vtxCount := interpreterProxy stackIntegerValue: 2.
	texHandle := interpreterProxy stackIntegerValue: 4.
	flags := interpreterProxy stackIntegerValue: 5.
	primType := interpreterProxy stackIntegerValue: 6.
	handle := interpreterProxy stackIntegerValue: 7.

	interpreterProxy failed ifTrue:[^nil].
	vtxArray := self stackPrimitiveVertexArray: 3 ofSize: vtxCount.
	idxArray := self stackPrimitiveIndexArray: 1 ofSize: idxCount validate: true forVertexSize: vtxCount.

	(vtxArray == nil or:[idxArray == nil 
		or:[primType < 1 or:[primType > 6 
			or:[interpreterProxy failed]]]])
				ifTrue:[^interpreterProxy primitiveFail].

	result := self b3dxRenderVertexBuffer: handle _: primType _: flags _: texHandle _: vtxArray _: vtxCount _: idxArray _: idxCount.
	result ifFalse:[^interpreterProxy primitiveFail].
	^interpreterProxy pop: 8. "pop args; return rcvr"! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'ar (auto pragmas 12/08) 9/9/2000 15:52'!
primitiveRendererVersion
	<export: true>
	interpreterProxy methodArgumentCount = 0
		ifFalse:[^interpreterProxy primitiveFail].
	interpreterProxy pop: 1.
	^interpreterProxy pushInteger: 1.! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 9/19/2018 14:30'!
primitiveSetBufferRect
	"Primitive. Set the buffer rectangle (e.g., the pixel area on screen) to use for this renderer.
	The viewport is positioned within the buffer rectangle."
	| h w y x handle |
	<export: true>
	interpreterProxy methodArgumentCount = 5 ifFalse:
		[^interpreterProxy primitiveFail].
	h := interpreterProxy stackIntegerValue: 0.
	w := interpreterProxy stackIntegerValue: 1.
	y := interpreterProxy stackIntegerValue: 2.
	x := interpreterProxy stackIntegerValue: 3.
	handle := interpreterProxy stackIntegerValue: 4.
	interpreterProxy failed ifFalse:
		[(self b3dxSetBufferRect: handle _: x _: y _: w _: h)
			ifTrue: [interpreterProxy pop: 5] "pop args; return rcvr"
			ifFalse:[interpreterProxy primitiveFail]]! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 9/19/2018 14:32'!
primitiveSetFog
	| handle rgba density fogType stop start |
	<export: true>
	<var: #density type: #double>
	<var: #start type: #double>
	<var: #stop type: #double>
	interpreterProxy methodArgumentCount = 6 ifFalse:
		[^interpreterProxy primitiveFail].
	rgba := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	stop := interpreterProxy floatValueOf: (interpreterProxy stackValue: 1).
	start := interpreterProxy floatValueOf: (interpreterProxy stackValue: 2).
	density := interpreterProxy floatValueOf: (interpreterProxy stackValue: 3).
	fogType := interpreterProxy stackIntegerValue: 4.
	handle := interpreterProxy stackIntegerValue: 5.
	interpreterProxy failed ifFalse:
		[(self b3dxSetFog: handle _: fogType _: density _: start _: stop _: rgba)
			ifTrue: [interpreterProxy pop: 6] "pop args; return rcvr"
			ifFalse:[interpreterProxy primitiveFail]]! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 9/19/2018 14:33'!
primitiveSetIntProperty
	| handle prop value |
	<export: true>
	interpreterProxy methodArgumentCount = 3 ifFalse:
		[^interpreterProxy primitiveFail].
	value := interpreterProxy stackIntegerValue: 0.
	prop := interpreterProxy stackIntegerValue: 1.
	handle := interpreterProxy stackIntegerValue: 2.
	(self b3dxSetIntProperty: handle _: prop _: value)
		ifTrue: [interpreterProxy pop: 3] "args; return rcvr"
		ifFalse: [interpreterProxy primitiveFail]
! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 10/29/2018 17:31'!
primitiveSetLights
	| lightArray lightCount light handle |
	<export: true>
	<var: #light type: #'void *'>

	interpreterProxy methodArgumentCount = 2
		ifFalse:[^interpreterProxy primitiveFail].

	lightArray := self stackLightArrayValue: 0.
	handle := interpreterProxy stackIntegerValue: 1.
	interpreterProxy failed ifTrue:[^nil].
	(self b3dxDisableLights: handle)
		ifFalse:[^interpreterProxy primitiveFail].
	lightArray ifNil: [^nil].
	lightCount := interpreterProxy slotSizeOf: lightArray.
	"For each enabled light source"
	0 to: lightCount-1 do:[:i|
		light := self fetchLightSource: i ofObject: lightArray.
		(self b3dxLoadLight: handle _: i _: light) ifFalse:
			[^interpreterProxy primitiveFail].
	].
	^interpreterProxy pop: 2. "args; return rcvr"! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 10/29/2018 17:42'!
primitiveSetMaterial
	| material handle |
	<export: true>
	<inline: false>
	<var: #material type: 'void*'>

	interpreterProxy methodArgumentCount = 2
		ifFalse:[^interpreterProxy primitiveFail].
	material := self stackMaterialValue: 0.
	handle := interpreterProxy stackIntegerValue: 1.
	(self b3dxLoadMaterial: handle _: material)
		ifFalse:[^interpreterProxy primitiveFail].
	^interpreterProxy pop: 2. "args; return rcvr"! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 9/19/2018 14:35'!
primitiveSetTransform
	"Transform an entire vertex buffer using the supplied modelview and projection matrix."
	| projectionMatrix modelViewMatrix handle |
	<export: true>
	<var: #projectionMatrix type: #'float *'>
	<var: #modelViewMatrix type: #'float *'>

	interpreterProxy methodArgumentCount = 3 ifFalse:
		[^interpreterProxy primitiveFail].
	projectionMatrix := self stackMatrix: 0.
	modelViewMatrix := self stackMatrix: 1.
	handle := interpreterProxy stackIntegerValue: 2.
	interpreterProxy failed ifFalse:
		[self b3dxSetTransform: handle _: modelViewMatrix _: projectionMatrix.
		 interpreterProxy pop: 3] "Leave rcvr on stack"! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 9/19/2018 14:36'!
primitiveSetVerboseLevel
	| result level |
	<export: true>
	interpreterProxy methodArgumentCount = 1 ifFalse:
		[^interpreterProxy primitiveFail].
	level := interpreterProxy stackIntegerValue: 0.
	result := self b3dxSetVerboseLevel: level.
	interpreterProxy methodReturnInteger: result! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 9/19/2018 14:38'!
primitiveSetViewport
	| h w y x handle |
	<export: true>
	interpreterProxy methodArgumentCount = 5 ifFalse:
		[^interpreterProxy primitiveFail].
	h := interpreterProxy stackIntegerValue: 0.
	w := interpreterProxy stackIntegerValue: 1.
	y := interpreterProxy stackIntegerValue: 2.
	x := interpreterProxy stackIntegerValue: 3.
	handle := interpreterProxy stackIntegerValue: 4.
	interpreterProxy failed ifFalse:
		[(self b3dxSetViewport: handle _: x _: y _: w _: h)
			ifTrue: [interpreterProxy pop: 5] "pop args; return rcvr"
			ifFalse:[interpreterProxy primitiveFail]]! !

!B3DAcceleratorPlugin methodsFor: 'primitives-renderer' stamp: 'eem 10/29/2018 17:42'!
primitiveSwapRendererBuffers
	| handle result |
	<export: true>
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxSwapRendererBuffers: handle.
	result ifFalse:[^interpreterProxy primitiveFail].
	^interpreterProxy pop: 1. "pop arg; return rcvr"! !

!B3DAcceleratorPlugin methodsFor: 'primitives-textures' stamp: 'eem 10/29/2018 17:42'!
primitiveTextureByteSex
	| handle result renderer |
	<export: true>
	interpreterProxy methodArgumentCount = 2
		ifFalse:[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	renderer := interpreterProxy stackIntegerValue: 1.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxTextureByteSex: renderer _: handle.
	result < 0 ifTrue:[^interpreterProxy primitiveFail].
	interpreterProxy pop: 3.
	^interpreterProxy pushBool: result.! !

!B3DAcceleratorPlugin methodsFor: 'primitives-textures' stamp: 'eem 10/29/2018 17:42'!
primitiveTextureDepth
	| handle result renderer |
	<export: true>
	interpreterProxy methodArgumentCount = 2
		ifFalse:[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	renderer := interpreterProxy stackIntegerValue: 1.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxActualTextureDepth: renderer _: handle.
	result < 0 ifTrue:[^interpreterProxy primitiveFail].
	interpreterProxy pop: 3.
	^interpreterProxy pushInteger: result.! !

!B3DAcceleratorPlugin methodsFor: 'primitives-textures' stamp: 'eem 10/29/2018 17:42'!
primitiveTextureGetColorMasks
	| handle masks array renderer |
	<export: true>
	<var: #masks declareC:'unsigned int masks[4]'>
	interpreterProxy methodArgumentCount = 3
		ifFalse:[^interpreterProxy primitiveFail].
	array := interpreterProxy stackObjectValue: 0.
	handle := interpreterProxy stackIntegerValue: 1.
	renderer := interpreterProxy stackIntegerValue: 2.
	interpreterProxy failed ifTrue:[^nil].
	((interpreterProxy isArray: array)
	 and: [(interpreterProxy slotSizeOf: array) = 4
	 and: [self b3dxTextureColorMasks: renderer _: handle _: masks]])
		ifFalse:[^interpreterProxy primitiveFail].
	interpreterProxy pushRemappableOop: array.
	0 to: 3 do: [:i|
		interpreterProxy
			storePointer: i
			ofObject: interpreterProxy topRemappableOop
			withValue: (interpreterProxy positive32BitIntegerFor: (masks at: i))].
	interpreterProxy popRemappableOop.
	^interpreterProxy pop: 3 "pop args return receiver"! !

!B3DAcceleratorPlugin methodsFor: 'primitives-textures' stamp: 'eem 10/29/2018 17:42'!
primitiveTextureSurfaceHandle
	| handle result renderer |
	<export: true>
	interpreterProxy methodArgumentCount = 2
		ifFalse:[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	renderer := interpreterProxy stackIntegerValue: 1.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxTextureSurfaceHandle: renderer _: handle.
	result < 0 ifTrue:[^interpreterProxy primitiveFail].
	interpreterProxy pop: 3.
	^interpreterProxy pushInteger: result! !

!B3DAcceleratorPlugin methodsFor: 'primitives-textures' stamp: 'eem 10/29/2018 17:42'!
primitiveTextureUpload
	| h w d result form bits ppw bitsPtr handle renderer |
	<export: true>
	<var: #bitsPtr type: 'void*'>
	interpreterProxy methodArgumentCount = 3
		ifFalse:[^interpreterProxy primitiveFail].
	form := interpreterProxy stackValue: 0.
	((interpreterProxy isPointers: form) and:[(interpreterProxy slotSizeOf: form) >= 4])
		ifFalse:[^interpreterProxy primitiveFail].
	bits := interpreterProxy fetchPointer: 0 ofObject: form.
	w := interpreterProxy fetchInteger: 1 ofObject: form.
	h := interpreterProxy fetchInteger: 2 ofObject: form.
	d := interpreterProxy fetchInteger: 3 ofObject: form.
	ppw := 32 // d.
	(interpreterProxy isWords: bits)
		ifFalse:[^interpreterProxy primitiveFail].
	(interpreterProxy slotSizeOf: bits) = (w + ppw - 1 // ppw * h)
		ifFalse:[^interpreterProxy primitiveFail].
	bitsPtr := interpreterProxy firstIndexableField: bits.
	handle := interpreterProxy stackIntegerValue: 1.
	renderer := interpreterProxy stackIntegerValue: 2.
	interpreterProxy failed ifTrue:[^nil].
	result := self b3dxUploadTexture: renderer _: handle _: w _: h _: d _: bitsPtr.
	result ifFalse:[^interpreterProxy primitiveFail].
	^interpreterProxy pop: 3. "args; return rcvr"! !

!B3DAcceleratorPlugin methodsFor: 'initialize-release' stamp: 'ar (auto pragmas 12/08) 5/26/2000 17:23'!
shutdownModule
	<export: true>
	^self b3dxShutdown! !

!B3DAcceleratorPlugin methodsFor: 'primitive support' stamp: 'eem 10/29/2018 17:19'!
stackLightArrayValue: stackIndex
	"Load an Array of B3DPrimitiveLights from the given stack index"
	| oop array arraySize |
	<inline: false>
	array := interpreterProxy stackValue: stackIndex.
	array = interpreterProxy nilObject ifTrue:[^nil].
	(interpreterProxy isArray: array) ifFalse:
		[^interpreterProxy primitiveFail].
	arraySize := interpreterProxy slotSizeOf: array.
	0 to: arraySize-1 do:[:i|
		oop := interpreterProxy fetchPointer: i ofObject: array.
		((interpreterProxy isWords: oop) and:[(interpreterProxy slotSizeOf: oop) = 32])
			ifFalse:[^interpreterProxy primitiveFail].
	].
	^array! !

!B3DAcceleratorPlugin methodsFor: 'primitive support' stamp: 'ar (auto pragmas 12/08) 3/26/2006 22:38'!
stackMaterialValue: stackIndex
	"Load a B3DMaterial from the given stack index"
	| oop |
	<inline: false>
	<returnTypeC:'void *'>
	oop := interpreterProxy stackObjectValue: stackIndex.
	oop = nil ifTrue:[^nil].
	oop = interpreterProxy nilObject ifTrue:[^nil].
	((interpreterProxy isWords: oop) and:[(interpreterProxy slotSizeOf: oop) =  17])
		ifTrue:[^interpreterProxy firstIndexableField: oop].
	^nil! !

!B3DAcceleratorPlugin methodsFor: 'primitive support' stamp: 'eem 8/2/2021 16:25'!
stackMatrix: index
	"Load a 4x4 transformation matrix from the interpreter stack.
	Return a pointer to the matrix data if successful, nil otherwise."
	| oop |
	<inline: #always>
	oop := interpreterProxy stackValue: index.
	((interpreterProxy isWords: oop) and:[(interpreterProxy slotSizeOf: oop) = 16]) ifTrue:
		[^interpreterProxy firstIndexableField: oop].
	interpreterProxy primitiveFail.
	^nil! !

!B3DAcceleratorPlugin methodsFor: 'primitive support' stamp: 'ar (auto pragmas 12/08) 5/26/2000 12:37'!
stackPrimitiveIndexArray: stackIndex ofSize: nItems validate: aBool forVertexSize: maxIndex
	"Load a primitive index array from the interpreter stack.
	If aBool is true then check that all the indexes are in the range (1,maxIndex).
	Return a pointer to the index data if successful, nil otherwise."
	| oop oopSize idxPtr index |
	<inline: false>
	<returnTypeC:'void*'>
	<var: #idxPtr declareC:'int *idxPtr'>

	oop := interpreterProxy stackObjectValue: stackIndex.
	oop = nil ifTrue:[^nil].
	(interpreterProxy isWords: oop) ifFalse:[^nil].
 	oopSize := interpreterProxy slotSizeOf: oop.
	oopSize < nItems ifTrue:[^nil].
	idxPtr := self cCoerce: (interpreterProxy firstIndexableField: oop) to:'int *'.
	aBool ifTrue:[
		0 to: nItems-1 do:[:i|
			index := idxPtr at: i.
			(index < 0 or:[index > maxIndex]) ifTrue:[^nil]]].
	^idxPtr! !

!B3DAcceleratorPlugin methodsFor: 'primitive support' stamp: 'ar (auto pragmas 12/08) 3/26/2006 22:40'!
stackPrimitiveVertexArray: index ofSize: nItems
	"Load a primitive vertex array from the interpreter stack.
	Return a pointer to the vertex data if successful, nil otherwise."
	| oop oopSize |
	<inline: false>
	<returnTypeC:'void*'>
	oop := interpreterProxy stackObjectValue: index.
	oop = nil ifTrue:[^nil].
	(interpreterProxy isWords: oop) ifTrue:[
 		oopSize := interpreterProxy slotSizeOf: oop.
		(oopSize >= nItems * 16 and:[oopSize \\ 16 = 0])
			ifTrue:[^interpreterProxy firstIndexableField: oop]].
	^nil! !

InterpreterPlugin subclass: #BMPReadWriterPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!BMPReadWriterPlugin commentStamp: 'tpr 5/5/2003 16:45' prior: 0!
A plugin to provide fast read and write of .bmp files!

!BMPReadWriterPlugin methodsFor: 'primitives' stamp: 'ar (auto pragmas dtl 2010-09-27) 8/1/2006 09:54'!
primitiveRead24BmpLine
	| width formBitsIndex formBitsOop pixelLineOop formBitsSize formBits pixelLineSize pixelLine |
	<export: true>
	<inline: false>
	<var: #formBits type: 'unsigned int *'>
	<var: #pixelLine type: 'unsigned char *'>
	interpreterProxy methodArgumentCount = 4 
		ifFalse:[^interpreterProxy primitiveFail].
	width := interpreterProxy stackIntegerValue: 0.
	width <= 0 ifTrue:[^interpreterProxy primitiveFail].
	formBitsIndex := interpreterProxy stackIntegerValue: 1.
	formBitsOop := interpreterProxy stackObjectValue: 2.
	pixelLineOop := interpreterProxy stackObjectValue: 3.
	interpreterProxy failed ifTrue:[^nil].
	(interpreterProxy isWords: formBitsOop) 
		ifFalse:[^interpreterProxy primitiveFail].
	(interpreterProxy isBytes: pixelLineOop)
		ifFalse:[^interpreterProxy primitiveFail].
	formBitsSize := interpreterProxy slotSizeOf: formBitsOop.
	formBits := interpreterProxy firstIndexableField: formBitsOop.
	pixelLineSize := interpreterProxy slotSizeOf: pixelLineOop.
	pixelLine := interpreterProxy firstIndexableField: pixelLineOop.
	(formBitsIndex + width <= formBitsSize and:[width*3 <= pixelLineSize])
		ifFalse:[^interpreterProxy primitiveFail].

	"do the actual work"
	self cCode:'
	formBits += formBitsIndex-1;
	while(width--) {
		unsigned int rgb;
		rgb = (*pixelLine++);
		rgb += (*pixelLine++) << 8;
		rgb += (*pixelLine++) << 16;
		if(rgb) rgb |= 0xFF000000; else rgb |= 0xFF000001;
		*formBits++ = rgb;
	}
	' inSmalltalk:[formBits. pixelLine. ^interpreterProxy primitiveFail].
	interpreterProxy pop: 4. "args"
! !

!BMPReadWriterPlugin methodsFor: 'primitives' stamp: 'ar (auto pragmas dtl 2010-09-27) 8/1/2006 09:53'!
primitiveWrite24BmpLine

	| width formBitsIndex formBitsOop pixelLineOop formBitsSize formBits pixelLineSize pixelLine |
	<export: true>
	<inline: false>
	<var: #formBits type: 'unsigned int *'>
	<var: #pixelLine type: 'unsigned char *'>
	interpreterProxy methodArgumentCount = 4 
		ifFalse:[^interpreterProxy primitiveFail].
	width := interpreterProxy stackIntegerValue: 0.
	width <= 0 ifTrue:[^interpreterProxy primitiveFail].
	formBitsIndex := interpreterProxy stackIntegerValue: 1.
	formBitsOop := interpreterProxy stackObjectValue: 2.
	pixelLineOop := interpreterProxy stackObjectValue: 3.
	interpreterProxy failed ifTrue:[^nil].
	(interpreterProxy isWords: formBitsOop) 
		ifFalse:[^interpreterProxy primitiveFail].
	(interpreterProxy isBytes: pixelLineOop)
		ifFalse:[^interpreterProxy primitiveFail].
	formBitsSize := interpreterProxy slotSizeOf: formBitsOop.
	formBits := interpreterProxy firstIndexableField: formBitsOop.
	pixelLineSize := interpreterProxy slotSizeOf: pixelLineOop.
	pixelLine := interpreterProxy firstIndexableField: pixelLineOop.

	(formBitsIndex + width <= formBitsSize and:[width*3 <= pixelLineSize])
		ifFalse:[^interpreterProxy primitiveFail].

	"do the actual work. Read 32 bit at a time from formBits, and store the low order 24 bits 
	or each word into pixelLine in little endian order."

	self cCode:'
	formBits += formBitsIndex-1;

	while(width--) {
		unsigned int rgb;
		rgb = *formBits++;
		(*pixelLine++) = (rgb      ) & 0xFF;
		(*pixelLine++) = (rgb >> 8 ) & 0xFF;
		(*pixelLine++) = (rgb >> 16) & 0xFF;
	}

	' inSmalltalk:[formBits. pixelLine. ^interpreterProxy primitiveFail].
	interpreterProxy pop: 4. "args"
! !

InterpreterPlugin subclass: #BalloonEngineBase
	instanceVariableNames: 'workBuffer objBuffer getBuffer aetBuffer spanBuffer engine formArray engineStopped geProfileTime dispatchedValue dispatchReturnValue objUsed doProfileStats copyBitsFn loadBBFn bbPluginName'
	classVariableNames: 'EdgeInitTable EdgeStepTable FillTable WideLineFillTable WideLineWidthTable'
	poolDictionaries: 'BalloonEngineConstants'
	category: 'VMMaker-Plugins'!

!BalloonEngineBase commentStamp: 'tpr 5/5/2003 11:45' prior: 0!
This is the main class for the Balloon graphics Engine.

BalloonEnginePlugin should be translated but its superclass should not since it is incorporated within that class's translation process. Nor should the simulation subclass be translated!

!BalloonEngineBase class methodsFor: 'documentation' stamp: 'ar 11/7/1998 03:33'!
a1EngineOutline
	"The following is a brief outline on how the engine works.

	In general, we're using a pretty straight-forward active edge approach, e.g., 
	we classify all edges into three different states:
		a) Waiting for processing
		b) Active (e.g., being processed)
		c) Finished
	Before the engine starts all edges are sorted by their y-value in a so-called
	'global edge table' (furthermore referred to as GET) and processed in top 
	to bottom order (the edges are also sorted by x-value but this is only for 
	simplifying the insertion when adding edges).

	Then, we start at the first visible scan line and execute the following steps:

	1) Move all edges starting at the current scan line from state a) to state b)

	This step requires the GET to be sorted so that we only need to check
	the first edges of the GET. After the initial state of the edge (e.g., it's current
	pixel value and data required for incremental updates) the edges are then 
	inserted in the 'active edge table' (called AET). The sort order in the AET is 
	defined by the pixel position of each edge at the current scan line and thus 
	edges are kept in increasing x-order.

	This step does occur for every edge only once and is therefore not the most
	time-critical part of the approach.

	2) Draw the current scan line

	This step includes two sub-parts. In the first part, the scan line is assembled.
	This involves walking through the AET and drawing the pixels between
	each two neighbour edges. Since each edge can have two associated fills
	(a 'left' and a 'right' fill) we need to make sure that edges falling on the
	same pixel position do not affect the painted image. This issue is discussed
	in the aetScanningProblems documentation.

	Wide edges (e.g., edges having an associated width) are also handled during
	this step. Wide edges are always preferred over interior fills - this ensures
	that the outline of an object cannot be overdrawn by any interior fill of
	a shape that ends very close to the edge (for more information see wideEdges 
	documentation).

	After the scan is assembled it is blitted to the screen. This only happens all
	'aaLevel' scan lines (for further information see the antiAliasing documentation).

	This second step is done at each scan line in the image, and is usually the most
	time-critical part.

	3) Update all currently active edges

	Updating the active edges basically means either to remove the edge from the AET
	(if it is at the end y value) or incrementally computing the pixel value for the
	next scan line. Based on the information gathered in the first step, this part
	should be executed as fast as possible - it happens for each edge in the AET
	at each scan line and may be the bottleneck if many edges are involved in
	the drawing operations (see the TODO list; part of it probably deals with the
	issue).

"
	^self error:'Comment only'! !

!BalloonEngineBase class methodsFor: 'documentation' stamp: 'ar 11/7/1998 03:55'!
a2AntiAliasing
	"The engine currently used a very simple, but efficient anti-aliasing scheme. It is based on a square unweighted filter of size 1, 2, or 4 resulting in three levels of anti-aliasing:

	* No anti-aliasing (filter size 1)
	This simply draws each pixel 'as is' on the screen

	* Slight anti-aliasing (filter size 2)
	Doubles the rasterization size in each direction and assembles the pixel value as the medium of the four sub-pixels falling into the full pixel

	* Full anti-aliasing (filter size 4)
	Quadruples the rasterization in each direction and assembles the pixel value as the medium of the sixteen sub-pixels falling into the full pixel


The reason for using these three AA levels is simply efficiency of computing. Since the above filters (1x1, 2x2, 4x4) have all power of two elements (1, 4, and 16) we can compute the weighted sum of the final pixel by computing

	destColor := destColor + (srcColor // subPixels)

And, since we're only working on 32bit destination buffer we do not need to compute the components of each color separately but can neatly put the entire color into a single formula:

	destPixel32 := destPixel32 + ((srcPixel32 bitAnd: aaMask) >> aaShift).

with aaMask = 16rFFFFFFFF for aaLevel = 1, aaMask = 16rFCFCFCFC for aaLevel = 2, aaMask = 16rF0F0F0F0 for aaLevel = 4 and aaShift = 0, 2, or 4 for the different levels. However, while the above is efficient to compute, it also drops accuracy. So, for the 4x4 anti-aliasing we're effectively only using the high 4 bits of each color component. While is generally not a problem (we add 16 sub-pixels into this value) there is a simple arithmetic difficulty because the above cannot fill the entire range of values, e.g.,

	16 * (255 // 16) = 16 * 15 = 240

and not 255 as expected. We solve this problem by replicating the top n (n=0, 2, 4) bits of each component as the low bits in an adjustment step before blitting to scan line to the screen. This has the nice effect that a zero pixel value (e.g., transparent) will remain zero, a white pixel (as computed above) will result in a value of 255 for each component (defining opaque white) and each color inbetween linearly mapped between 0 and 255. 

"
	^self error:'Comment only'! !

!BalloonEngineBase class methodsFor: 'documentation' stamp: 'ar 11/7/1998 03:35'!
a3RasterizationRules


	^self error:'Comment only'! !

!BalloonEngineBase class methodsFor: 'documentation' stamp: 'ar 11/7/1998 03:35'!
a4WideEdges! !

!BalloonEngineBase class methodsFor: 'documentation' stamp: 'ar 11/7/1998 03:36'!
a5AETScanningProblems
	"Due to having two fill entries (one left and one right) there can be problems while scanning the active edge table. In general, the AET should look like the following (ri - regions, ei - edges, fi - fills):

			|				\				|
	r1		|		r2		 \		r3		|		r4
			|				  \				|
			e1				 e2				e3		

	with:
		f(r1) = fLeft(e1) = 0				(empty fill, denoted -)
		f(r2) = fRight(e1) = fLeft(e2)		(denoted x)
		f(r3) = fRight(e2) = fLeft(e3)	(denoted o)
		f(r4) = fRight(e3) = 0

	However, due to integer arithmetic used during computations the AET may look like the following:
			X
			\|						|
			 | \						|
			 |   \					|
	r1		 | r2 \			r3		|		r4
			 |	   \					|
			e1		e2				e3		

	In this case, the starting point of e1 and e2 have the same x value at the first scan line but e2 has been sorted before e1 (Note: This can happen in *many* cases - the above is just a very simple example). Given the above outlined fill relations we have a problem. So, for instance, using the left/right fills as defined by the edges would lead to the effect that in the first scan line region r3 is actually filled with the right fill of e1 while it should actually be filled with the right fill of e2. This leads to noticable artifacts in the image and increasing resolution does not help.

	What we do here is defining an arbitrary sort order between fills (you can think of it as a depth value but the only thing that matters is that you can order the fills by this number and that the empty fill is always sorted at the end), and toggle the fills between an 'active' and an 'inactive' state at each edge. This is done as follows:
	For each edge ei in the AET do:
		* if fLeft(ei) isActive then removeActive(fLeft(ei)) else addActive(fLeft(ei))
		* if fRight(ei) isActive then removeActive(fRight(ei)) else addActive(fRight(ei))
		* draw the span from ei to ei+1 with currentActive
	where addActive adds the fill to the list of currently active fills, removeActive() removes the fill from the active list and currentActive returns the fill AS DEFINED BY THE SORT ORDER from the list of active fills. Note that this does not change anything in the first example above because the list will only contain one entry (besides the empty fill). In the second case however, it will lead to the following sequence:

	* toggle fLeft(e2) = f(r2) = 'x'
		- makes fLeft(e2) active
		- activeList = 'x'
	* toggle fRight(e2) = f(r3) = 'o'
		- makes fRight(e2) active
		- activeList = 'xo'
	* draw span from e2 to e1
		Depending on the sort order between 'x' and 'o' the region will be drawn with either one of the fills. It is significant to note here that the occurence of such a problem is generally only *very* few pixels large (in the above example zero pixels) and will therefore not be visually noticable. In any case, there is a unique decision for the fill to use here and that is what we need if the problem did not happen accidentally (e.g., someone has manually changed one fill of an edge but not the fill of the opposite edge).

	* toggle fLeft(e1) = f(r1) = '-'
		- makes fLeft(r1) visible
		- activeList = 'xo-'
		[Note: empty fills are a special case. 
		They can be ignored since they sort last
		and the activeList can return the empty
		fill if it is itself empty].
	* toggle fRight(e1) = f(r2) = 'x'
		- makes fRight(e1) invisible
		- activeList = 'o-'
	* draw span from e2 to e3
		Since the active list contains (besides the empty fill) only one fill value this will be used. Fortunately, this is the correct fill because it is the fill we had initially defined for the region r2.

An interesting side effect of the above is that there is no such notion as a 'left' or 'right' fill anymore. Another (not-so-nice) side effect is that the entire AET has to be scanned from the beginning even if only the last few edges actually affect the visible region.

PS. I need to find a way of clipping the edges for this. More on it later...
"
	^self error:'Comment only'! !

!BalloonEngineBase class methodsFor: 'documentation' stamp: 'ar 11/8/1998 00:06'!
a6StuffTODO
	"This is an unordered list of things to do:

BalloonEnginePlugin>>stepToFirstBezierIn:at:
	1)	Check if reducing maxSteps from 2*deltaY to deltaY 
		brings a *significant* performance improvement.
		In theory this should make for double step performance
		but will cost in quality. Might be that the AA stuff will
		compensate for this - but I'm not really sure.

BalloonEngineBase>>dispatchOn:in:
	1)	Check what dispatches cost most and must be inlined
		by an #inlinedDispatchOn:in: Probably this will be
		stepping and eventually wide line stuff but we'll see.

BalloonEngineBase
	1)	Check which variables should become inst vars, if any.
		This will remove an indirection during memory access
		and might allow a couple of optimizations by the C compiler.

Anti-Aliasing:
	1)	Check if we can use a weighted 3x3 filter function of the form
				1	2	1
				2	4	2
				1	2	1
		Which should be *extremely* nice for fonts (it's sharpening
		edges). The good thing about the above is that it sums up to
		16 (as in the 4x4 case) but I don't know how to keep a history
		without needing two extra scan lines.

	2)	Check if we can - somehow - integrate more general filters.

	3) Unroll the loops during AA so we can copy and mask aaLevel pixels
	   in each step between start and end. This should speed up filling
	   by a factor of 2-4 (in particular for difficult stuff like radial gradients).

Clipping
	1)	Find a way of clipping edges left of the clip rectangle
		or at least ignoring most of them after the first scan line.
		The AET scanning problems discuss the issue but it should be
		possible to keep the color list between spans (if not empty)
		and speed up drawing at the very right (such as in the
		Winnie Pooh example where a lot of stuff is between the
		left border and the clipping rect.

	2)	Check if we can determine empty states of the color list and
		an edge that is longer than anything left of it. This should
		work in theory but might be relatively expensive to compute.

"
	^self error:'Comment only'! !

!BalloonEngineBase class methodsFor: 'translation' stamp: 'ikp 6/14/2004 15:19'!
declareCVarsIn: cg

	"Buffers"
	cg var: #workBuffer type: #'int*'.
	cg var: #objBuffer type: #'int*'.
	cg var: #getBuffer type: #'int*'.
	cg var: #aetBuffer type: #'int*'.
	cg var: #spanBuffer type: #'unsigned int*'.
	cg var: #edgeTransform declareC: 'float edgeTransform[6]'.
	cg var: #doProfileStats declareC: 'int doProfileStats = 0'.
	cg var: 'bbPluginName' declareC:'char bbPluginName[256] = "BitBltPlugin"'.
	"Functions"
	cg var: 'copyBitsFn'	type: 'void *'.
	cg var: 'loadBBFn'	type: 'void *'.! !

!BalloonEngineBase class methodsFor: 'class initialization' stamp: 'eem 8/27/2009 15:37'!
initialize
	"BalloonEngineBase initialize"
	"BalloonEnginePlugin translateDoInlining: true."
	EdgeInitTable := self initializeEdgeInitTable.
	EdgeStepTable := self initializeEdgeStepTable.
	WideLineWidthTable := self initializeWideLineWidthTable.
	WideLineFillTable := self initializeWideLineFillTable.
	FillTable := self initializeFillTable.

	(Smalltalk classNamed: #BalloonEngineConstants) ifNotNil:
		[:balloonEngineConstants|
		(balloonEngineConstants classPool anySatisfy: [:classVarValue| classVarValue isNil]) ifTrue:
			[balloonEngineConstants initialize]]! !

!BalloonEngineBase class methodsFor: 'class initialization' stamp: 'ar 11/4/1998 21:52'!
initializeEdgeInitTable
	"BalloonEngineBase initialize"
	^#(
		errorWrongIndex
		errorWrongIndex
		errorWrongIndex
		errorWrongIndex

		stepToFirstLine
		stepToFirstWideLine
		stepToFirstBezier
		stepToFirstWideBezier
	)! !

!BalloonEngineBase class methodsFor: 'class initialization' stamp: 'ar 11/4/1998 21:52'!
initializeEdgeStepTable
	"BalloonEngineBase initialize"
	^#(
		errorWrongIndex
		errorWrongIndex
		errorWrongIndex
		errorWrongIndex

		stepToNextLine
		stepToNextWideLine
		stepToNextBezier
		stepToNextWideBezier
	)! !

!BalloonEngineBase class methodsFor: 'class initialization' stamp: 'ar 11/25/1998 19:46'!
initializeFillTable
	"BalloonEngineBase initialize"
	^#(
		errorWrongIndex "Type zero - undefined"
		errorWrongIndex "Type one - external fill"

		fillLinearGradient "Linear gradient fill"
		fillRadialGradient "Radial gradient fill"

		fillBitmapSpan	"Clipped bitmap fill"
		fillBitmapSpan	"Repeated bitmap fill"
	)! !

!BalloonEngineBase class methodsFor: 'class initialization' stamp: 'ar 11/4/1998 23:03'!
initializeWideLineFillTable
	"BalloonEngineBase initialize"
	^#(
		errorWrongIndex
		errorWrongIndex
		returnWideLineFill
		returnWideBezierFill
	)! !

!BalloonEngineBase class methodsFor: 'class initialization' stamp: 'ar 11/4/1998 23:03'!
initializeWideLineWidthTable
	"BalloonEngineBase initialize"
	^#(
		errorWrongIndex
		errorWrongIndex
		returnWideLineWidth
		returnWideBezierWidth
	)! !

!BalloonEngineBase class methodsFor: 'translation' stamp: 'ar 5/11/2000 23:48'!
moduleName
	^'B2DPlugin'! !

!BalloonEngineBase class methodsFor: 'translation' stamp: 'tpr 5/14/2001 12:14'!
shouldBeTranslated
"BalloonEnginePlugin should be translated but its superclasse should not since it is incorporated within this class's translation process. Nor should the simulation subclass be translated"
	^self == BalloonEnginePlugin! !

!BalloonEngineBase class methodsFor: 'simulation' stamp: 'ar 11/11/1998 21:56'!
simulatorClass
	^BalloonEngineSimulation! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 00:34'!
aaColorMaskGet
	^workBuffer at: GWAAColorMask! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 00:34'!
aaColorMaskPut: value
	^workBuffer at: GWAAColorMask put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 00:34'!
aaColorShiftGet
	^workBuffer at: GWAAColorShift! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 00:34'!
aaColorShiftPut: value
	^workBuffer at: GWAAColorShift put: value! !

!BalloonEngineBase methodsFor: 'displaying' stamp: 'ar (auto pragmas 12/08) 11/9/1998 00:53'!
aaFirstPixelFrom: leftX to: rightX
	"Common function to compute the first full pixel for AA drawing"
	| firstPixel |
	<inline: true>
	firstPixel := (leftX + self aaLevelGet - 1) bitAnd: (self aaLevelGet - 1) bitInvert32.
	firstPixel > rightX 
		ifTrue:[^rightX]
		ifFalse:[^firstPixel]! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 00:34'!
aaHalfPixelGet
	^workBuffer at: GWAAHalfPixel! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 00:35'!
aaHalfPixelPut: value
	^workBuffer at: GWAAHalfPixel put: value! !

!BalloonEngineBase methodsFor: 'displaying' stamp: 'ar (auto pragmas 12/08) 11/9/1998 00:53'!
aaLastPixelFrom: leftX to: rightX
	"Common function to compute the last full pixel for AA drawing"
	<inline: true>
	^(rightX - 1) bitAnd: (self aaLevelGet - 1) bitInvert32.! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 00:35'!
aaLevelGet
	^workBuffer at: GWAALevel! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 00:35'!
aaLevelPut: value
	^workBuffer at: GWAALevel put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 00:35'!
aaScanMaskGet
	^workBuffer at: GWAAScanMask! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 00:35'!
aaScanMaskPut: value
	^workBuffer at: GWAAScanMask put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 00:35'!
aaShiftGet
	^workBuffer at: GWAAShift! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 00:36'!
aaShiftPut: value
	^workBuffer at: GWAAShift put: value! !

!BalloonEngineBase methodsFor: 'other' stamp: 'ar 11/9/1998 02:06'!
accurateLengthOf: deltaX with: deltaY
	"Return the accurate length of the vector described by deltaX and deltaY"
	| length2 |
	deltaX = 0 ifTrue:[deltaY < 0 ifTrue:[^0-deltaY] ifFalse:[^deltaY]].
	deltaY = 0 ifTrue:[deltaX < 0 ifTrue:[^0-deltaX] ifFalse:[^deltaX]].
	length2 := (deltaX * deltaX) + (deltaY * deltaY).
	^self computeSqrt: length2! !

!BalloonEngineBase methodsFor: 'GET processing' stamp: 'ar (auto pragmas 12/08) 11/1/1998 01:07'!
addEdgeToGET: edge
	<inline: false>
	(self allocateGETEntry: 1) ifFalse:[^0].
	"Install edge in the GET"
	getBuffer at: self getUsedGet put: edge.
	self getUsedPut: self getUsedGet + 1.! !

!BalloonEngineBase methodsFor: 'displaying' stamp: 'ar (auto pragmas 12/08) 11/9/1998 00:50'!
adjustAALevel

		"NOTE: 	This method is (hopefully) obsolete due to unrolling 
				the fill loops to deal with full pixels."

	"Adjust the span buffers values by the appropriate color offset for anti-aliasing.
	We do this by replicating the top bits of each color in the lower bits. The idea is that we can scale each color value uniquely from 0 to 255 and thus fill the entire range of colors."
	| adjustShift adjustMask x0 x1 pixelValue |
	<inline: false>
	adjustShift := 8 - self aaColorShiftGet.
	adjustMask := self aaColorMaskGet bitInvert32.
	x0 := self spanStartGet >> self aaShiftGet.
	x1 := self spanEndGet >> self aaShiftGet.
	[x0 < x1] whileTrue:[
		pixelValue := spanBuffer at: x0.
		spanBuffer at: x0 put: (pixelValue bitOr: (pixelValue >> adjustShift bitAnd: adjustMask)).
		x0 := x0 + 1].! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:27'!
aetStartGet
	^workBuffer at: GWAETStart! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:29'!
aetStartPut: value
	^workBuffer at: GWAETStart put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:27'!
aetUsedGet
	^workBuffer at: GWAETUsed! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:29'!
aetUsedPut: value
	^workBuffer at: GWAETUsed put: value! !

!BalloonEngineBase methodsFor: 'allocating' stamp: 'ar 10/29/1998 18:37'!
allocateAETEntry: nSlots
	"Allocate n slots in the active edge table"
	^self needAvailableSpace: nSlots! !

!BalloonEngineBase methodsFor: 'allocating' stamp: 'ar (auto pragmas 12/08) 10/28/1998 21:06'!
allocateGETEntry: nSlots
	"Allocate n slots in the global edge table"
	| srcIndex dstIndex |
	<inline: false>
	"First allocate nSlots in the AET"
	(self allocateAETEntry: nSlots) ifFalse:[^false].
	self aetUsedGet = 0 ifFalse:["Then move the AET upwards"
		srcIndex := self aetUsedGet.
		dstIndex := self aetUsedGet + nSlots.
		1 to: self aetUsedGet do:[:i|
			aetBuffer at: (dstIndex := dstIndex - 1) put: (aetBuffer at: (srcIndex := srcIndex - 1))].
	].
	aetBuffer := aetBuffer + nSlots.
	^true! !

!BalloonEngineBase methodsFor: 'allocating' stamp: 'ar (auto pragmas 12/08) 10/28/1998 21:16'!
allocateObjEntry: nSlots
	"Allocate n slots in the object buffer"
	| srcIndex dstIndex |
	<inline: false>
	"First allocate nSlots in the GET"
	(self allocateGETEntry: nSlots) ifFalse:[^false].
	self getUsedGet = 0 ifFalse:["Then move the GET upwards"
		srcIndex := self getUsedGet.
		dstIndex := self getUsedGet + nSlots.
		1 to: self getUsedGet do:[:i|
			getBuffer at: (dstIndex := dstIndex - 1) put: (getBuffer at: (srcIndex := srcIndex - 1))].
	].
	getBuffer := getBuffer + nSlots.
	^true! !

!BalloonEngineBase methodsFor: 'allocating' stamp: 'ar 10/29/1998 18:37'!
allocateStackEntry: nSlots
	"AET and Stack allocation are symmetric"
	^self needAvailableSpace: nSlots! !

!BalloonEngineBase methodsFor: 'allocating' stamp: 'ar 10/30/1998 19:24'!
allocateStackFillEntry
	^self wbStackPush: self stackFillEntryLength! !

!BalloonEngineBase methodsFor: 'testing' stamp: 'ar 11/9/1998 15:34'!
areEdgeFillsValid: edge

	^((self objectHeaderOf: edge) bitAnd: GEEdgeFillsInvalid) = 0! !

!BalloonEngineBase methodsFor: 'displaying' stamp: 'ar (auto pragmas 12/08) 11/14/1998 19:31'!
clearSpanBuffer
	"Clear the current span buffer.
	The span buffer is only cleared in the area that has been used by the previous scan line."
	| x0 x1 |
	<inline: false>
	x0 := self spanStartGet >> self aaShiftGet.
	x1 := self spanEndGet >> self aaShiftGet + 1.
	x0 < 0 ifTrue:[x0 := 0].
	x1 > self spanSizeGet ifTrue:[x1 := self spanSizeGet].
	[x0 < x1] whileTrue:[
		spanBuffer at: x0 put: 0.
		x0 := x0 + 1].
	self spanStartPut: self spanSizeGet.
	self spanEndPut: 0.! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/29/1998 00:43'!
clearSpanBufferGet
	^workBuffer at: GWClearSpanBuffer! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/29/1998 00:44'!
clearSpanBufferPut: value
	^workBuffer at: GWClearSpanBuffer put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 18:46'!
clipMaxXGet
	^workBuffer at: GWClipMaxX! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 18:46'!
clipMaxXPut: value
	^workBuffer at: GWClipMaxX put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 18:46'!
clipMaxYGet
	^workBuffer at: GWClipMaxY! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 18:46'!
clipMaxYPut: value
	^workBuffer at: GWClipMaxY put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 18:46'!
clipMinXGet
	^workBuffer at: GWClipMinX! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 18:47'!
clipMinXPut: value
	^workBuffer at: GWClipMinX put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 18:47'!
clipMinYGet
	^workBuffer at: GWClipMinY! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 18:47'!
clipMinYPut: value
	^workBuffer at: GWClipMinY put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar (auto pragmas 12/08) 11/24/1998 21:36'!
colorTransform
	<returnTypeC:'float *'>
	^self cCoerce: workBuffer + GWColorTransform to:'float *'! !

!BalloonEngineBase methodsFor: 'other' stamp: 'ar 11/7/1998 14:26'!
computeSqrt: length2
	length2 < 32 
		ifTrue:[^self smallSqrtTable at: length2]
		ifFalse:[^(length2 asFloat sqrt + 0.5) asInteger]! !

!BalloonEngineBase methodsFor: 'private' stamp: 'ikp 6/14/2004 15:14'!
copyBitsFrom: x0 to: x1 at: yValue

	copyBitsFn = 0 ifTrue: [
		"We need copyBits here so try to load it implicitly"
		self initialiseModule ifFalse: [^false].
	].
	^self cCode: '((sqInt (*)(sqInt, sqInt, sqInt))copyBitsFn)(x0, x1, yValue)'! !

!BalloonEngineBase methodsFor: 'GET processing' stamp: 'ar (auto pragmas 12/08) 11/25/1998 00:41'!
createGlobalEdgeTable
	"Create the global edge table"
	| object end |
	<inline: false>
	object := 0.
	end := objUsed.
	[object < end] whileTrue:[
		"Note: addEdgeToGET: may fail on insufficient space but that's not a problem here"
		(self isEdge: object) ifTrue:[
			"Check if the edge starts below fillMaxY."
			(self edgeYValueOf: object) >= self fillMaxYGet ifFalse:[
				self checkedAddEdgeToGET: object.
			].
		].
		object := object + (self objectLengthOf: object).
	].! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:27'!
currentYGet
	^workBuffer at: GWCurrentY! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 21:27'!
currentYPut: value
	^workBuffer at: GWCurrentY put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 11/24/1998 20:28'!
currentZGet
	^workBuffer at: GWCurrentZ! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 11/24/1998 20:29'!
currentZPut: value
	^workBuffer at: GWCurrentZ put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/29/1998 16:23'!
destOffsetXGet
	^workBuffer at: GWDestOffsetX! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/29/1998 16:24'!
destOffsetXPut: value
	^workBuffer at: GWDestOffsetX put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/29/1998 16:23'!
destOffsetYGet
	^workBuffer at: GWDestOffsetY! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/29/1998 16:24'!
destOffsetYPut: value
	^workBuffer at: GWDestOffsetY put: value! !

!BalloonEngineBase methodsFor: 'displaying' stamp: 'ar (auto pragmas 12/08) 5/12/2000 16:42'!
displaySpanBufferAt: y
	"Display the span buffer at the current scan line."
	| targetX0 targetX1 targetY |
	<inline: false>
	"self aaLevelGet > 1 ifTrue:[self adjustAALevel]."
	targetX0 := self spanStartGet >> self aaShiftGet.
	targetX0 < self clipMinXGet ifTrue:[targetX0 := self clipMinXGet].
	targetX1 := (self spanEndGet + self aaLevelGet - 1) >> self aaShiftGet.
	targetX1 > self clipMaxXGet ifTrue:[targetX1 := self clipMaxXGet].
	targetY := y >> self aaShiftGet.
	(targetY < self clipMinYGet or:[targetY >= self clipMaxYGet or:[
		targetX1 < self clipMinXGet or:[targetX0 >= self clipMaxXGet]]]) ifTrue:[^0].
	self copyBitsFrom: targetX0 to: targetX1 at: targetY.! !

!BalloonEngineBase methodsFor: 'displaying' stamp: 'ar (auto pragmas 12/08) 11/25/1998 02:34'!
drawWideEdge: edge from: leftX
	"Draw the given edge starting from leftX with the edge's fill.
	Return the end value of the drawing operation."
	| rightX fill type lineWidth |
	<inline: false> "Not for the moment"
	type := self edgeTypeOf: edge.
	dispatchedValue := edge.
	self dispatchOn: type in: WideLineWidthTable.
	lineWidth := dispatchReturnValue.
	self dispatchOn: type in: WideLineFillTable.
	fill := self makeUnsignedFrom: dispatchReturnValue.
	fill = 0 ifTrue:[^leftX].
	"Check if this line is only partially visible"
	"self assert:(self isFillColor: fill)."
	rightX := leftX + lineWidth.
	self fillSpan: fill from: leftX to: rightX.
	^rightX! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/9/1998 15:35'!
edgeFillsInvalidate: edge

	^self objectTypeOf: edge put: 
		((self objectTypeOf: edge) bitOr: GEEdgeFillsInvalid)! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/9/1998 15:35'!
edgeFillsValidate: edge

	^self objectTypeOf: edge put: 
		((self objectTypeOf: edge) bitAnd: GEEdgeFillsInvalid bitInvert32)! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/24/1998 22:04'!
edgeLeftFillOf: edge

	^self obj: edge at: GEFillIndexLeft! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/24/1998 22:04'!
edgeLeftFillOf: edge put: value

	^self obj: edge at: GEFillIndexLeft put: value! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/24/1998 22:04'!
edgeNumLinesOf: edge

	^self obj: edge at: GENumLines! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/24/1998 22:04'!
edgeNumLinesOf: edge put: value

	^self obj: edge at: GENumLines put: value! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/24/1998 22:05'!
edgeRightFillOf: edge

	^self obj: edge at: GEFillIndexRight! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/24/1998 22:05'!
edgeRightFillOf: edge put: value

	^self obj: edge at: GEFillIndexRight put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar (auto pragmas 12/08) 11/24/1998 21:36'!
edgeTransform
	<returnTypeC:'float *'>
	^self cCoerce: workBuffer + GWEdgeTransform to:'float *'! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/9/1998 15:35'!
edgeTypeOf: edge
	"Return the edge type (e.g., witout the wide edge flag)"

	^(self objectTypeOf: edge) >> 1! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/24/1998 22:05'!
edgeXValueOf: edge

	^self obj: edge at: GEXValue! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/24/1998 22:05'!
edgeXValueOf: edge put: value

	^self obj: edge at: GEXValue put: value! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/24/1998 22:05'!
edgeYValueOf: edge

	^self obj: edge at: GEYValue! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/24/1998 22:06'!
edgeYValueOf: edge put: value

	^self obj: edge at: GEYValue put: value! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/24/1998 22:06'!
edgeZValueOf: edge

	^self obj: edge at: GEZValue! !

!BalloonEngineBase methodsFor: 'accessing edges' stamp: 'ar 11/24/1998 22:06'!
edgeZValueOf: edge put: value

	^self obj: edge at: GEZValue put: value! !

!BalloonEngineBase methodsFor: 'private' stamp: 'eem 4/27/2020 13:39'!
errorWrongIndex
	interpreterProxy error:'BalloonEngine: Fatal dispatch error'! !

!BalloonEngineBase methodsFor: 'other' stamp: 'ar 11/8/1998 14:33'!
estimatedLengthOf: deltaX with: deltaY
	"Estimate the length of the vector described by deltaX and deltaY.
	This method may be extremely inaccurate - use it only
	if you know exactly that this doesn't matter. Otherwise
	use #accurateLengthOf:width:"
	| absDx absDy |
	deltaX >= 0 ifTrue:[absDx := deltaX] ifFalse:[absDx := 0 - deltaX].
	deltaY >= 0 ifTrue:[absDy := deltaY] ifFalse:[absDy := 0 - deltaY].
	absDx > absDy 
		ifTrue:[^absDx + (absDy // 2)]
		ifFalse:[^absDy + (absDx // 2)]

! !

!BalloonEngineBase methodsFor: 'displaying' stamp: 'ar (auto pragmas 12/08) 11/25/1998 15:12'!
fillAllFrom: leftX to: rightX
	"Fill the span buffer from leftX to rightX with the given fill."
	| fill startX stopX |
	<inline: true>
	fill := self topFill.
	startX := leftX.
	stopX := self topRightX.
	[stopX < rightX] whileTrue:[
		fill := self makeUnsignedFrom: self topFill.
		fill = 0 ifFalse:[
			(self fillSpan: fill from: startX to: stopX) ifTrue:[^true]].
		self quickRemoveInvalidFillsAt: stopX.
		startX := stopX.
		stopX := self topRightX].
	fill := self makeUnsignedFrom: self topFill.
	fill = 0 ifFalse:[^self fillSpan: fill from: startX to: rightX].
	^false! !

!BalloonEngineBase methodsFor: 'displaying' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 14:20'!
fillBitmapSpan: bits from: leftX to: rightX
	"Fill the span buffer between leftEdge and rightEdge using the given bits.
	Note: We always start from zero - this avoids using huge bitmap buffers if the bitmap is to be displayed at the very far right hand side and also gives us a chance of using certain bitmaps (e.g., those with depth 32) directly."
	| x0 x1 x bitX colorMask colorShift baseShift fillValue |
	<inline: false>
	<var: #bits type:'int *'>

	x0 := leftX.
	x1 := rightX.
	bitX := -1. "Hack for pre-increment"
	self aaLevelGet = 1 ifTrue:["Speedy version for no anti-aliasing"
		[x0 < x1] whileTrue:[
			fillValue := (self cCoerce: bits to: 'int *') at: (bitX := bitX + 1).
			spanBuffer at: x0 put: fillValue.
			x0 := x0 + 1.
		].
	] ifFalse:["Generic version with anti-aliasing"
		colorMask := self aaColorMaskGet.
		colorShift := self aaColorShiftGet.
		baseShift := self aaShiftGet.
		[x0 < x1] whileTrue:[
			x := x0 >> baseShift.
			fillValue := (self cCoerce: bits to: 'int *') at: (bitX := bitX + 1).
			fillValue := (fillValue bitAnd: colorMask) >> colorShift.
			spanBuffer at: x put: (spanBuffer at: x) + fillValue.
			x0 := x0 + 1.
		].
	].
	x1 > self spanEndGet ifTrue:[self spanEndPut: x1].
	x1 > self spanEndAAGet ifTrue:[self spanEndAAPut: x1].! !

!BalloonEngineBase methodsFor: 'displaying' stamp: 'ar (auto pragmas 12/08) 11/8/1998 03:30'!
fillColorSpan: pixelValue32 from: leftX to: rightX
	"Fill the span buffer between leftEdge and rightEdge with the given pixel value."
	| x0 x1 |
	<inline: true>
	"Use a unrolled version for anti-aliased fills..."
	self aaLevelGet = 1
		ifFalse:[^self fillColorSpanAA: pixelValue32 x0: leftX x1: rightX].
	x0 := leftX.
	x1 := rightX.
	"Unroll the inner loop four times, since we're only storing data."
	[x0 + 4 < x1] whileTrue:[
		spanBuffer at: x0 put: pixelValue32.
		spanBuffer at: x0+1 put: pixelValue32.
		spanBuffer at: x0+2 put: pixelValue32.
		spanBuffer at: x0+3 put: pixelValue32.
		x0 := x0+4.
	].
	[x0 < x1] whileTrue:[
		spanBuffer at: x0 put: pixelValue32.
		x0 := x0 + 1.
	].! !

!BalloonEngineBase methodsFor: 'displaying' stamp: 'ar (auto pragmas 12/08) 11/9/1998 00:52'!
fillColorSpanAA: pixelValue32 x0: leftX x1: rightX
	"This is the inner loop for solid color fills with anti-aliasing.
	This loop has been unrolled for speed and quality into three parts:
		a) copy all pixels that fall into the first full pixel.
		b) copy aaLevel pixels between the first and the last full pixel
		c) copy all pixels that fall in the last full pixel"
	| colorMask baseShift x idx firstPixel lastPixel aaLevel pv32 |
	<inline: false> "Not now -- maybe later"
	"Compute the pixel boundaries."
	firstPixel := self aaFirstPixelFrom: leftX to: rightX.
	lastPixel := self aaLastPixelFrom: leftX to: rightX.
	aaLevel := self aaLevelGet.
	baseShift := self aaShiftGet.
	x := leftX.

	"Part a: Deal with the first n sub-pixels"
	x < firstPixel ifTrue:[
		pv32 := (pixelValue32 bitAnd: self aaColorMaskGet) >> self aaColorShiftGet.
		[x < firstPixel] whileTrue:[
			idx := x >> baseShift.
			spanBuffer at: idx put: (spanBuffer at: idx) + pv32.
			x := x + 1.
		].
	].

	"Part b: Deal with the full pixels"
	x < lastPixel ifTrue:[
		colorMask := (self aaColorMaskGet >> self aaShiftGet) bitOr: 16rF0F0F0F0.
		pv32 := (pixelValue32 bitAnd: colorMask) >> self aaShiftGet.
		[x < lastPixel] whileTrue:[
			idx := x >> baseShift.
			spanBuffer at: idx put: (spanBuffer at: idx) + pv32.
			x := x + aaLevel.
		].
	].

	"Part c: Deal with the last n sub-pixels"
	x < rightX ifTrue:[
		pv32 := (pixelValue32 bitAnd: self aaColorMaskGet) >> self aaColorShiftGet.
		[x < rightX] whileTrue:[
			idx := x >> baseShift.
			spanBuffer at: idx put: (spanBuffer at: idx) + pv32.
			x := x + 1.
		].
	].! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:27'!
fillMaxXGet
	^workBuffer at: GWFillMaxX! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:29'!
fillMaxXPut: value
	^workBuffer at: GWFillMaxX put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:27'!
fillMaxYGet
	^workBuffer at: GWFillMaxY! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:30'!
fillMaxYPut: value
	^workBuffer at: GWFillMaxY put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:27'!
fillMinXGet
	^workBuffer at: GWFillMinX! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:30'!
fillMinXPut: value
	^workBuffer at: GWFillMinX put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:27'!
fillMinYGet
	^workBuffer at: GWFillMinY! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:30'!
fillMinYPut: value
	^workBuffer at: GWFillMinY put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:27'!
fillOffsetXGet
	^workBuffer at: GWFillOffsetX! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:30'!
fillOffsetXPut: value
	^workBuffer at: GWFillOffsetX put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:27'!
fillOffsetYGet
	^workBuffer at: GWFillOffsetY! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:30'!
fillOffsetYPut: value
	^workBuffer at: GWFillOffsetY put: value! !

!BalloonEngineBase methodsFor: 'FILL processing' stamp: 'ar (auto pragmas 12/08) 11/24/1998 22:42'!
fillSorts: fillEntry1 before: fillEntry2
	"Return true if fillEntry1 should be drawn before fillEntry2"
	| diff |
	<inline: false>
	"First check the depth value"
	diff := (self stackFillDepth: fillEntry1) - (self stackFillDepth: fillEntry2).
	diff = 0 ifFalse:[^diff > 0].
	"See the class comment for aetScanningProblems"
	^(self cCoerce: (self makeUnsignedFrom: (self stackFillValue: fillEntry1)) to:'unsigned') <
		(self cCoerce: (self makeUnsignedFrom: (self stackFillValue: fillEntry2)) to: 'unsigned')! !

!BalloonEngineBase methodsFor: 'displaying' stamp: 'ikp (auto pragmas 12/08) 8/9/2004 18:22'!
fillSpan: fill from: leftX to: rightX
	"Fill the span buffer from leftX to rightX with the given fill.
	Clip before performing any operations. Return true if the fill must
	be handled by some Smalltalk code."
	| x0 x1 type |
	<var: #fill type: 'unsigned int'>
	<inline: false>
	fill = 0 ifTrue:[^false]. "Nothing to do"
	"Start from spEnd - we must not paint pixels twice at a scan line"
	leftX < self spanEndAAGet 
		ifTrue:[x0 := self spanEndAAGet]
		ifFalse:[x0 := leftX].
	rightX > (self spanSizeGet << self aaShiftGet) 
		ifTrue:[x1 := (self spanSizeGet << self aaShiftGet)]
		ifFalse:[x1 := rightX].

	"Clip left and right values"
	x0 < self fillMinXGet ifTrue:[x0 := self fillMinXGet].
	x1 > self fillMaxXGet ifTrue:[x1 := self fillMaxXGet].

	"Adjust start and end values of span"
	x0 < self spanStartGet ifTrue:[self spanStartPut: x0].
	x1 > self spanEndGet ifTrue:[self spanEndPut: x1].
	x1 > self spanEndAAGet ifTrue:[self spanEndAAPut: x1].

	x0 >= x1 ifTrue:[^false]. "Nothing to do"

	(self isFillColor: fill) ifTrue:[
		self fillColorSpan: fill from: x0 to: x1.
	] ifFalse:[
		"Store the values for the dispatch"
		self lastExportedFillPut: fill.
		self lastExportedLeftXPut: x0.
		self lastExportedRightXPut: x1.
		type := self fillTypeOf: fill.
		type <= 1 ifTrue:[^true].
		self dispatchOn: type in: FillTable.
	].
	^false! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 11/7/1998 22:25'!
fillTypeOf: fill
	^((self objectTypeOf: fill) bitAnd: GEPrimitiveFillMask) >> 8! !

!BalloonEngineBase methodsFor: 'AET processing' stamp: 'ar 11/24/1998 22:47'!
findNextAETEdgeFrom: leftEdge
	| depth rightEdge |
	depth := self edgeZValueOf: leftEdge.
	[self aetStartGet < self aetUsedGet] whileTrue:[
		rightEdge := aetBuffer at: self aetStartGet.
		(self edgeZValueOf: rightEdge) >= depth ifTrue:[^rightEdge].
		self aetStartPut: self aetStartGet + 1.
	].
	^nil! !

!BalloonEngineBase methodsFor: 'GET processing' stamp: 'ar 11/9/1998 15:36'!
findNextExternalEntryFromGET
	"Check the global edge table for any entries that cannot be handled by the engine itself.
	If there are any, return true. Otherwise, initialize the the edge and add it to the AET"
	| yValue edge type |
	yValue := self currentYGet.
	"As long as we have entries in the GET"
	[self getStartGet < self getUsedGet] whileTrue:[
		edge := getBuffer at: self getStartGet.
		(self edgeYValueOf: edge) > yValue ifTrue:[^false]. "No more edges to add"
		type := self objectTypeOf: edge.
		(type bitAnd: GEPrimitiveWideMask) = GEPrimitiveEdge 
			ifTrue:[^true]. "This is an external edge"
		"Note: We must make sure not to do anything with the edge if there is not
		enough room in the AET"
		(self needAvailableSpace: 1) ifFalse:[^false]. "No more room"
		"Process the edge in the engine itself"
		self dispatchOn: type in: EdgeInitTable.
		"Insert the edge into the AET"
		self insertEdgeIntoAET: edge.
		self getStartPut: self getStartGet + 1.
	].
	"No entries in GET"
	^false! !

!BalloonEngineBase methodsFor: 'AET processing' stamp: 'ar (auto pragmas 12/08) 11/25/1998 23:21'!
findNextExternalFillFromAET
	"Scan the active edge table. If there is any fill that cannot be handled by the engine itself,  return true. Otherwise handle the fills and return false."
	| leftEdge rightEdge leftX rightX |
"self currentYGet >= 680 ifTrue:[
self printAET.
self halt.
]."

	<inline: false>
	leftX := rightX := self fillMaxXGet.
	[self aetStartGet < self aetUsedGet] whileTrue:[
		leftEdge := rightEdge := aetBuffer at: self aetStartGet.
		"TODO: We should check if leftX from last operation 
			is  greater than leftX from next edge.
			Currently, we rely here on spanEndAA
			from the span buffer fill."
		leftX := rightX := self edgeXValueOf: leftEdge.
		leftX >= self fillMaxXGet ifTrue:[^false]. "Nothing more visible"
		self quickRemoveInvalidFillsAt: leftX.
		"Check if we need to draw the edge"
		(self isWide: leftEdge) ifTrue:[
			self toggleWideFillOf: leftEdge.
			"leftX := rightX := self drawWideEdge: leftEdge from: leftX."
		].
		(self areEdgeFillsValid: leftEdge) ifTrue:[
			self toggleFillsOf: leftEdge. "Adjust the fills"
			engineStopped ifTrue:[^false].
		].
		self aetStartPut: self aetStartGet + 1.
		self aetStartGet < self aetUsedGet ifTrue:[
			rightEdge := aetBuffer at: self aetStartGet.
			rightX := self edgeXValueOf: rightEdge.
			rightX >= self fillMinXGet ifTrue:["This is the visible portion"
				self fillAllFrom: leftX to: rightX.
				"Fetch the currently active fill"
				"fill := self makeUnsignedFrom: self topFill.
				fill = 0 ifFalse:[self fillSpan: fill from: leftX to: rightX max: self topRightX]"
			].
		].
	].
	"Note: Due to pre-clipping we may have to draw remaining stuff with the last fill"
	rightX < self fillMaxXGet ifTrue:[
		self fillAllFrom: rightX to: self fillMaxXGet.
		"fill := self makeUnsignedFrom: self topFill.
		fill = 0 ifFalse:[self fillSpan: fill from: rightX to: self fillMaxXGet max: self topRightX]."
	].
	^false! !

!BalloonEngineBase methodsFor: 'AET processing' stamp: 'ar (auto pragmas 12/08) 11/9/1998 15:36'!
findNextExternalUpdateFromAET
	"Check the active edge table for any entries that cannot be handled by the engine itself.
	If there are any, return true. Otherwise, step the the edge to the next y value."
	| edge count type |
	<inline: false>
	[self aetStartGet < self aetUsedGet] whileTrue:[
		edge := aetBuffer at: self aetStartGet.
		count := (self edgeNumLinesOf: edge) - 1.
		count = 0 ifTrue:[
			"Edge at end -- remove it"
			self removeFirstAETEntry
		] ifFalse:[
			"Store remaining lines back"
			self edgeNumLinesOf: edge put: count.
			type := self objectTypeOf: edge.
			(type bitAnd: GEPrimitiveWideMask) = GEPrimitiveEdge 
				ifTrue:[^true]. "This is an external edge"
			self dispatchOn: type in: EdgeStepTable.
			self resortFirstAETEntry.
			self aetStartPut: self aetStartGet+1.
		].
	].
	^false! !

!BalloonEngineBase methodsFor: 'FILL processing' stamp: 'ar 11/25/1998 15:47'!
findStackFill: fillIndex depth: depth
	| index |
	index := 0.
	[index < self stackFillSize and:[
		(self stackFillValue: index) ~= fillIndex or:[
			(self stackFillDepth: index) ~= depth]]]
				whileTrue:[index := index + self stackFillEntryLength].
	index >= self stackFillSize 
		ifTrue:[^-1]
		ifFalse:[^index].
! !

!BalloonEngineBase methodsFor: 'testing' stamp: 'ar 10/31/1998 17:06'!
finishedProcessing
	"Return true if processing is finished"
	^self stateGet = GEStateCompleted! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/29/1998 17:08'!
firstPointListGet
	^workBuffer at: GWPointListFirst! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/29/1998 17:08'!
firstPointListPut: value
	^workBuffer at: GWPointListFirst put: value! !

!BalloonEngineBase methodsFor: 'allocating' stamp: 'ar 10/30/1998 19:24'!
freeStackFillEntry
	self wbStackPop: self stackFillEntryLength.! !

!BalloonEngineBase methodsFor: 'GET processing' stamp: 'ar (auto pragmas 12/08) 10/28/1998 21:07'!
getSorts: edge1 before: edge2
	"Return true if the edge at index i should sort before the edge at index j."
	| diff |
	<inline: false>
	edge1 = edge2 ifTrue:[^true].
	"First, sort by Y"
	diff := (self edgeYValueOf: edge1) - (self edgeYValueOf: edge2).
	diff = 0 ifFalse:[^diff < 0].
	"Then, by X"
	diff := (self edgeXValueOf: edge1) - (self edgeXValueOf: edge2).
	^diff < 0! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:27'!
getStartGet
	^workBuffer at: GWGETStart! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:30'!
getStartPut: value
	^workBuffer at: GWGETStart put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:27'!
getUsedGet
	^workBuffer at: GWGETUsed! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:31'!
getUsedPut: value
	^workBuffer at: GWGETUsed put: value! !

!BalloonEngineBase methodsFor: 'testing' stamp: 'ar 11/24/1998 19:39'!
hasColorTransform
	^self hasColorTransformGet ~= 0! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 11/24/1998 21:36'!
hasColorTransformGet
	^workBuffer at: GWHasColorTransform! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 11/24/1998 21:36'!
hasColorTransformPut: value
	^workBuffer at: GWHasColorTransform put: value! !

!BalloonEngineBase methodsFor: 'testing' stamp: 'ar 11/24/1998 19:38'!
hasEdgeTransform
	^self hasEdgeTransformGet ~= 0! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 11/24/1998 21:36'!
hasEdgeTransformGet
	^workBuffer at: GWHasEdgeTransform! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 11/24/1998 21:35'!
hasEdgeTransformPut: value
	^workBuffer at: GWHasEdgeTransform put: value! !

!BalloonEngineBase methodsFor: 'FILL processing' stamp: 'ar (auto pragmas 12/08) 11/25/1998 15:48'!
hideFill: fillIndex depth: depth
	"Make the fill style with the given index invisible"
	| index newTopIndex newTop newDepth newRightX |
	<inline: false>
	index := self findStackFill: fillIndex depth: depth.
	index = -1 ifTrue:[^false].

	index = 0 ifTrue:[
		self freeStackFillEntry.
		^true].

	"Fill is visible - replace it with the last entry on the stack"
	self stackFillValue: index put: (self stackFillValue: 0).
	self stackFillDepth: index put: (self stackFillDepth: 0).
	self stackFillRightX: index put: (self stackFillRightX: 0).
	self freeStackFillEntry.
	(self stackFillSize <= self stackFillEntryLength) ifTrue:[^true]. "Done"

	"Find the new top fill"
	newTopIndex := 0.
	index := self stackFillEntryLength.
	[index < self stackFillSize] whileTrue:[
		(self fillSorts: index before: newTopIndex)
			ifTrue:[newTopIndex := index].
		index := index + self stackFillEntryLength.
	].
	(newTopIndex + self stackFillEntryLength = self stackFillSize) 
		ifTrue:[^true]. "Top fill not changed"
	newTop := self stackFillValue: newTopIndex.
	self stackFillValue: newTopIndex put: self topFillValue.
	self topFillValuePut: newTop.
	newDepth := self stackFillDepth: newTopIndex.
	self stackFillDepth: newTopIndex put: self topFillDepth.
	self topFillDepthPut: newDepth.
	newRightX := self stackFillRightX: newTopIndex.
	self stackFillRightX: newTopIndex put: self topFillRightX.
	self topFillRightXPut: newRightX.
	^true! !

!BalloonEngineBase methodsFor: 'transforming' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 14:20'!
incrementPoint: point by: delta
	<var: #point type:'int *'>
	point at: 0 put: (point at: 0) + delta.
	point at: 1 put: (point at: 1) + delta.! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 11/9/1998 15:36'!
incrementStat: statIndex by: value

	^workBuffer at: statIndex put: (workBuffer at: statIndex) + value! !

!BalloonEngineBase methodsFor: 'AET processing' stamp: 'ar (auto pragmas 12/08) 10/28/1998 21:07'!
indexForInsertingIntoAET: edge
	"Find insertion point for the given edge in the AET"
	| initialX index |
	<inline: false>
	initialX := self edgeXValueOf: edge.
	index := 0.
	[index < self aetUsedGet and:[
		(self edgeXValueOf: (aetBuffer at: index)) < initialX]]
			whileTrue:[index := index + 1].
	[index < self aetUsedGet and:[
		(self edgeXValueOf: (aetBuffer at: index)) = initialX and:[
			(self getSorts: (aetBuffer at: index) before: edge)]]]
				whileTrue:[index := index + 1].
	^index! !

!BalloonEngineBase methodsFor: 'other' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 14:20'!
initColorTransform
	| transform |
	<inline: false>
	<var: #transform type:'float *'>
	transform := self colorTransform.
	transform at: 0 put: (self cCoerce: 1.0 to: 'float').
	transform at: 1 put: (self cCoerce: 0.0 to: 'float').
	transform at: 2 put: (self cCoerce: 1.0 to: 'float').
	transform at: 3 put: (self cCoerce: 0.0 to: 'float').
	transform at: 4 put: (self cCoerce: 1.0 to: 'float').
	transform at: 5 put: (self cCoerce: 0.0 to: 'float').
	transform at: 6 put: (self cCoerce: 1.0 to: 'float').
	transform at: 7 put: (self cCoerce: 0.0 to: 'float').
	self hasColorTransformPut: 0.! !

!BalloonEngineBase methodsFor: 'other' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 14:20'!
initEdgeTransform
	| transform |
	<inline: false>
	<var: #transform type:'float *'>
	transform := self edgeTransform.
	transform at: 0 put: (self cCoerce: 1.0 to: 'float').
	transform at: 1 put: (self cCoerce: 0.0 to: 'float').
	transform at: 2 put: (self cCoerce: 0.0 to: 'float').
	transform at: 3 put: (self cCoerce: 0.0 to: 'float').
	transform at: 4 put: (self cCoerce: 1.0 to: 'float').
	transform at: 5 put: (self cCoerce: 0.0 to: 'float').
	self hasEdgeTransformPut: 0.! !

!BalloonEngineBase methodsFor: 'initialize-release' stamp: 'tpr (auto pragmas 12/08) 4/7/2004 21:10'!
initialiseModule
	<export: true>

	loadBBFn := interpreterProxy ioLoadFunction: 'loadBitBltFrom' From: bbPluginName.
	copyBitsFn := interpreterProxy ioLoadFunction: 'copyBitsFromtoat' From: bbPluginName.
	^(loadBBFn ~= 0 and:[copyBitsFn ~= 0])! !

!BalloonEngineBase methodsFor: 'GET processing' stamp: 'di (auto pragmas 12/08) 7/14/2004 13:09'!
initializeGETProcessing
	"Initialization stuff that needs to be done before any processing can take place."
	<inline: false>

	"Make sure aaLevel is initialized"
	self setAALevel: self aaLevelGet.

	self clipMinXGet < 0 ifTrue:[self clipMinXPut: 0].
	self clipMaxXGet > self spanSizeGet ifTrue:[self clipMaxXPut: self spanSizeGet].
	"Convert clipRect to aaLevel"
	self fillMinXPut: self clipMinXGet << self aaShiftGet.
	self fillMinYPut: self clipMinYGet << self aaShiftGet.
	self fillMaxXPut: self clipMaxXGet << self aaShiftGet.
	self fillMaxYPut: self clipMaxYGet << self aaShiftGet.

	"Reset GET and AET"
	self getUsedPut: 0.
	self aetUsedPut: 0.
	getBuffer := objBuffer + objUsed.
	aetBuffer := objBuffer + objUsed.

	"Create the global edge table"
	self createGlobalEdgeTable.
	engineStopped ifTrue:[^nil].

	self getUsedGet = 0 ifTrue:[
		"Nothing to do"
		self currentYPut: self fillMaxYGet.
		^0].

	"Sort entries in the GET"
	self sortGlobalEdgeTable.

	"Find the first y value to be processed"
	self currentYPut: (self edgeYValueOf: (getBuffer at: 0)).
	self currentYGet < self fillMinYGet ifTrue:[self currentYPut: self fillMinYGet].

	"Load and clear the span buffer"
	self spanStartPut: 0.
	self spanEndPut: (self spanSizeGet << self aaShiftGet) - 1.
	self clearSpanBuffer. "@@: Is this really necessary?!!"! !

!BalloonEngineBase methodsFor: 'AET processing' stamp: 'ar (auto pragmas 12/08) 10/28/1998 19:52'!
insertEdgeIntoAET: edge
	"Insert the edge with the given index from the global edge table into the active edge table.
	The edge has already been stepped to the initial yValue -- thus remainingLines and rasterX
	are both set."
	| index |
	<inline: false>

	"Check for the number of lines remaining"
	(self edgeNumLinesOf: edge) <= 0 ifTrue:[^nil]. "Nothing to do"

	"Find insertion point"
	index := self indexForInsertingIntoAET: edge.

	"And insert edge"
	self insertToAET: edge beforeIndex: index.! !

!BalloonEngineBase methodsFor: 'AET processing' stamp: 'ar (auto pragmas 12/08) 10/28/1998 21:07'!
insertToAET: edge beforeIndex: index
	"Insert the given edge into the AET."
	| i |
	<inline: false>
	"Make sure we have space in the AET"
	(self allocateAETEntry: 1) ifFalse:[^nil]. "Insufficient space in AET"

	i := self aetUsedGet-1.
	[i < index] whileFalse:[
		aetBuffer at: i+1 put: (aetBuffer at: i).
		i := i - 1.
	].
	aetBuffer at: index put: edge.
	self aetUsedPut: self aetUsedGet + 1.! !

!BalloonEngineBase methodsFor: 'testing' stamp: 'ar 10/29/1998 19:36'!
isEdge: edge
	| type |
	type := self objectTypeOf: edge.
	type > GEPrimitiveEdgeMask ifTrue:[^false].
	^((self objectTypeOf: edge) bitAnd: GEPrimitiveEdgeMask) ~= 0! !

!BalloonEngineBase methodsFor: 'testing' stamp: 'ar 11/7/1998 21:28'!
isFill: fill
	^(self isFillColor: fill) or:[self isRealFill: fill]! !

!BalloonEngineBase methodsFor: 'testing' stamp: 'ar 10/29/1998 19:31'!
isFillColor: fill
	^((self makeUnsignedFrom: fill) bitAnd: 16rFF000000) ~= 0! !

!BalloonEngineBase methodsFor: 'testing' stamp: 'ar 11/25/1998 00:43'!
isObject: obj
	^obj >= 0 and:[obj < objUsed]! !

!BalloonEngineBase methodsFor: 'testing' stamp: 'ar 11/7/1998 21:28'!
isRealFill: fill
	^((self objectTypeOf: fill) bitAnd: GEPrimitiveFillMask) ~= 0! !

!BalloonEngineBase methodsFor: 'testing' stamp: 'ar 10/31/1998 23:12'!
isStackEntry: entry
	^entry >= self wbTopGet and:[entry < self wbSizeGet]! !

!BalloonEngineBase methodsFor: 'testing' stamp: 'ar 10/30/1998 17:38'!
isStackIndex: index
	^index >= 0 and:[index < self wbStackSize]! !

!BalloonEngineBase methodsFor: 'testing' stamp: 'ar 11/9/1998 15:36'!
isWide: object

	^((self objectTypeOf: object) bitAnd: GEPrimitiveWide) ~= 0! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:28'!
lastExportedEdgeGet
	^workBuffer at: GWLastExportedEdge! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 20:11'!
lastExportedEdgePut: value
	^workBuffer at: GWLastExportedEdge put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 14:24'!
lastExportedFillGet
	^workBuffer at: GWLastExportedFill! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 14:24'!
lastExportedFillPut: value
	^workBuffer at: GWLastExportedFill put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 16:51'!
lastExportedLeftXGet
	^workBuffer at: GWLastExportedLeftX! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 16:51'!
lastExportedLeftXPut: value
	^workBuffer at: GWLastExportedLeftX put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 16:51'!
lastExportedRightXGet
	^workBuffer at: GWLastExportedRightX! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/28/1998 16:51'!
lastExportedRightXPut: value
	^workBuffer at: GWLastExportedRightX put: value! !

!BalloonEngineBase methodsFor: 'loading state' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 14:20'!
loadArrayTransformFrom: transformOop into: destPtr length: n
	"Load a transformation from the given array."
	| value |
	<inline: false>
	<var: #destPtr type:'float *'>
	0 to: n-1 do:[:i|
		value := interpreterProxy fetchPointer: i ofObject: transformOop.
		((interpreterProxy isIntegerObject: value) or:[interpreterProxy isFloatObject: value])
			ifFalse:[^interpreterProxy primitiveFail].
		(interpreterProxy isIntegerObject: value)
			ifTrue:[destPtr at: i put: 
				(self cCoerce: (interpreterProxy integerValueOf: value) asFloat to:'float')]
			ifFalse:[destPtr at: i put: 
				(self cCoerce: (interpreterProxy floatValueOf: value) to: 'float')].
	].! !

!BalloonEngineBase methodsFor: 'private' stamp: 'ikp 6/14/2004 15:14'!
loadBitBltFrom: bbObj

	loadBBFn = 0 ifTrue: [
		"We need copyBits here so try to load it implicitly"
		self initialiseModule ifFalse:[^false].
	].
	^self cCode: '((sqInt (*)(sqInt))loadBBFn)(bbObj)'! !

!BalloonEngineBase methodsFor: 'loading state' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 14:20'!
loadColorTransformFrom: transformOop
	"Load a 2x3 transformation matrix from the given oop.
	Return true if the matrix is not nil, false otherwise"
	| okay transform |
	<var: #transform type:'float *'>
	transform := self colorTransform.
	self hasColorTransformPut: 0.
	okay := self loadTransformFrom: transformOop into: transform length: 8.
	okay ifFalse:[^false].
	self hasColorTransformPut: 1.
	"Scale transform to be in 0-256 range"
	transform at: 1 put: (transform at: 1) * (self cCoerce: 256.0 to:'float').
	transform at: 3 put: (transform at: 3) * (self cCoerce: 256.0 to:'float').
	transform at: 5 put: (transform at: 5) * (self cCoerce: 256.0 to:'float').
	transform at: 7 put: (transform at: 7) * (self cCoerce: 256.0 to:'float').
	^okay! !

!BalloonEngineBase methodsFor: 'loading state' stamp: 'eem 6/3/2009 14:00'!
loadEdgeStateFrom: edgeOop
	| edge |
	<inline: false>
	edge := self lastExportedEdgeGet.
	(interpreterProxy slotSizeOf: edgeOop) < ETBalloonEdgeDataSize 
		ifTrue:[^nil].
	self edgeXValueOf: edge 
		put: (interpreterProxy fetchInteger: ETXValueIndex ofObject: edgeOop).
	self edgeYValueOf: edge 
		put: (interpreterProxy fetchInteger: ETYValueIndex ofObject: edgeOop).
	self edgeZValueOf: edge 
		put: (interpreterProxy fetchInteger: ETZValueIndex ofObject: edgeOop).
	self edgeNumLinesOf: edge 
		put: (interpreterProxy fetchInteger: ETLinesIndex ofObject: edgeOop).
	^edge! !

!BalloonEngineBase methodsFor: 'loading state' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 14:21'!
loadEdgeTransformFrom: transformOop
	"Load a 2x3 transformation matrix from the given oop.
	Return true if the matrix is not nil, false otherwise"
	| transform okay |
	<inline: false>
	<var: #transform type:'float *'>
	self hasEdgeTransformPut: 0.
	transform := self edgeTransform.
	okay := self loadTransformFrom: transformOop into: transform length: 6.
	interpreterProxy failed ifTrue:[^nil].
	okay ifFalse:[^false].
	self hasEdgeTransformPut: 1.
	"Add the fill offset to the matrix"
	transform at: 2 put: 
		(self cCoerce: (transform at: 2) + self destOffsetXGet asFloat to:'float').
	transform at: 5 put: 
		(self cCoerce: (transform at: 5) + self destOffsetYGet asFloat to:'float').
	^true! !

!BalloonEngineBase methodsFor: 'loading state' stamp: 'eem 1/30/2014 16:03'!
loadFormsFrom: arrayOop
	"Check all the forms from arrayOop."
	| formOop bmBits bmBitsSize bmWidth bmHeight bmDepth ppw bmRaster |
	(interpreterProxy isArray: arrayOop) ifFalse:[^false].
	formArray := arrayOop.
	0 to: (interpreterProxy slotSizeOf: formArray) - 1 do:[:i|
		formOop := interpreterProxy fetchPointer: i ofObject: formArray.
		(interpreterProxy isPointers: formOop) ifFalse:[^false].
		(interpreterProxy slotSizeOf: formOop) < 5 ifTrue:[^false].
		bmBits := interpreterProxy fetchPointer: 0 ofObject: formOop.
		(interpreterProxy fetchClassOf: bmBits) = interpreterProxy classBitmap
			ifFalse:[^false].
		bmBitsSize := interpreterProxy slotSizeOf: bmBits.
		bmWidth := interpreterProxy fetchInteger: 1 ofObject: formOop.
		bmHeight := interpreterProxy fetchInteger: 2 ofObject: formOop.
		bmDepth := interpreterProxy fetchInteger: 3 ofObject: formOop.
		interpreterProxy failed ifTrue:[^false].
		(bmWidth >= 0 and:[bmHeight >= 0]) ifFalse:[^false].
		ppw := 32 // bmDepth.
		bmRaster := bmWidth + (ppw-1) // ppw.
		bmBitsSize = (bmRaster * bmHeight)
			ifFalse:[^false].
	].
	^true! !

!BalloonEngineBase methodsFor: 'loading state' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 14:21'!
loadPoint: pointArray from: pointOop
	"Load the contents of pointOop into pointArray"
	| value |
	<inline: false>
	<var: #pointArray type:'int *'>
	(interpreterProxy fetchClassOf: pointOop) = interpreterProxy classPoint 
		ifFalse:[^interpreterProxy primitiveFail].
	value := interpreterProxy fetchPointer: 0 ofObject: pointOop.
	((interpreterProxy isIntegerObject: value) or:[interpreterProxy isFloatObject: value])
		ifFalse:[^interpreterProxy primitiveFail].
	(interpreterProxy isIntegerObject: value)
		ifTrue:[pointArray at: 0 put: (interpreterProxy integerValueOf: value)]
		ifFalse:[pointArray at: 0 put: (interpreterProxy floatValueOf: value) asInteger].
	value := interpreterProxy fetchPointer: 1 ofObject: pointOop.
	((interpreterProxy isIntegerObject: value) or:[interpreterProxy isFloatObject: value])
		ifFalse:[^interpreterProxy primitiveFail].
	(interpreterProxy isIntegerObject: value)
		ifTrue:[pointArray at: 1 put: (interpreterProxy integerValueOf: value)]
		ifFalse:[pointArray at: 1 put: (interpreterProxy floatValueOf: value) asInteger].
! !

!BalloonEngineBase methodsFor: 'primitives-rendering' stamp: 'eem 6/8/2009 10:33'!
loadRenderingState
	"Load the entire state from the interpreter for the rendering primitives.
	 Answer 0 on success or a non-zero failure code on failure."
	| failCode edgeOop fillOop state |
	<inline: false>
	interpreterProxy methodArgumentCount = 2
		ifFalse:[^PrimErrBadNumArgs].

	(failCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 2)) ~= 0 ifTrue:
		[^failCode].

	fillOop := interpreterProxy stackObjectValue: 0.
	edgeOop := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue: [^PrimErrBadArgument].

	"Load span buffer and bitBlt"
	(failCode := self loadSpanBufferFrom:
		(interpreterProxy fetchPointer: BESpanIndex ofObject: engine)) = 0 ifFalse:
		[^failCode].
	(self loadBitBltFrom: 
		(interpreterProxy fetchPointer: BEBitBltIndex ofObject: engine)) ifFalse:
		[^GEFBitBltLoadFailed].
	(self loadFormsFrom:
		(interpreterProxy fetchPointer: BEFormsIndex ofObject: engine)) ifFalse:
		[^GEFFormLoadFailed].
	"Check edgeOop and fillOop"
	(interpreterProxy slotSizeOf: edgeOop) < ETBalloonEdgeDataSize 
		ifTrue:[^GEFEdgeDataTooSmall].
	(interpreterProxy slotSizeOf: fillOop) < FTBalloonFillDataSize 
		ifTrue:[^GEFFillDataTooSmall].

	"Note: Rendering can only take place if we're not in one of the intermediate
	(e.g., external) states."
	state := self stateGet.
	(state = GEStateWaitingForEdge or:[
		state = GEStateWaitingForFill or:[
			state = GEStateWaitingChange]]) ifTrue:[^GEFWrongState].

	^0! !

!BalloonEngineBase methodsFor: 'loading state' stamp: 'eem 6/3/2009 14:15'!
loadSpanBufferFrom: spanOop
	"Load the span buffer from the given oop.
	 Answer 0 on success or a non-zero failure code on failure."
	<inline: false>
	(interpreterProxy fetchClassOf: spanOop) = (interpreterProxy classBitmap) ifFalse:[^GEFClassMismatch].
	spanBuffer := interpreterProxy firstIndexableField: spanOop.
	"Leave last entry unused to avoid complications"
	self spanSizePut: (interpreterProxy slotSizeOf: spanOop) - 1.
	^0! !

!BalloonEngineBase methodsFor: 'loading state' stamp: 'eem 9/24/2013 13:49'!
loadTransformFrom: transformOop into: destPtr length: n
	"Load a transformation from transformOop into the float array
	defined by destPtr. The transformation is assumed to be either
	an array or a FloatArray of length n."
	<inline: false>
	<var: #destPtr type:'float *'>
	transformOop = interpreterProxy nilObject ifTrue:[^false].
	(interpreterProxy isImmediate: transformOop)
		ifTrue:[^interpreterProxy primitiveFail].
	(interpreterProxy slotSizeOf: transformOop) = n 
		ifFalse:[^interpreterProxy primitiveFail].
	(interpreterProxy isWords: transformOop) 
		ifTrue:[self loadWordTransformFrom: transformOop into: destPtr length: n]
		ifFalse:[self loadArrayTransformFrom: transformOop into: destPtr length: n].
	^true! !

!BalloonEngineBase methodsFor: 'loading state' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 14:21'!
loadWordTransformFrom: transformOop into: destPtr length: n
	"Load a float array transformation from the given oop"
	| srcPtr |
	<inline: false>
	<var: #srcPtr type:'float *'>
	<var: #destPtr type:'float *'>
	srcPtr := self cCoerce: (interpreterProxy firstIndexableField: transformOop) to: 'float *'.
	0 to: n-1 do:[:i| destPtr at: i put: (srcPtr at: i)].! !

!BalloonEngineBase methodsFor: 'loading state' stamp: 'eem 9/20/2013 18:06'!
loadWorkBufferFrom: wbOop
	"Load the working buffer from the given oop"
	<inline: false>
	(interpreterProxy isImmediate: wbOop) ifTrue:[^GEFWorkBufferIsInteger].
	(interpreterProxy isWords: wbOop) ifFalse:[^GEFWorkBufferIsPointers].
	(interpreterProxy slotSizeOf: wbOop) < GWMinimalSize ifTrue:[^GEFWorkBufferTooSmall].
	self workBufferPut: wbOop.
	self magicNumberGet = GWMagicNumber ifFalse:[^GEFWorkBufferBadMagic].
	"Sanity checks"
	(self wbSizeGet = (interpreterProxy slotSizeOf: wbOop)) ifFalse:[^GEFWorkBufferWrongSize].
	self objStartGet = GWHeaderSize ifFalse:[^GEFWorkBufferStartWrong].

	"Load buffers"
	objBuffer := workBuffer + self objStartGet.
	getBuffer := objBuffer + self objUsedGet.
	aetBuffer := getBuffer + self getUsedGet.

	"Make sure we don't exceed the work buffer"
	GWHeaderSize + self objUsedGet + self getUsedGet + self aetUsedGet
	> self wbSizeGet ifTrue:[^GEFWorkTooBig].

	^0! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:28'!
magicNumberGet
	^workBuffer at: GWMagicIndex! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:39'!
magicNumberPut: value
	^workBuffer at: GWMagicIndex put: value! !

!BalloonEngineBase methodsFor: 'private' stamp: 'ar 10/28/1998 20:58'!
makeUnsignedFrom: someIntegerValue
	^someIntegerValue! !

!BalloonEngineBase methodsFor: 'initialize-release' stamp: 'ar (auto pragmas 12/08) 5/16/2000 19:57'!
moduleUnloaded: aModuleName
	"The module with the given name was just unloaded.
	Make sure we have no dangling references."
	<export: true>
	<var: #aModuleName type: 'char *'>
	(aModuleName strcmp: bbPluginName) = 0 ifTrue:[
		"BitBlt just shut down. How nasty."
		loadBBFn := 0.
		copyBitsFn := 0.
	].! !

!BalloonEngineBase methodsFor: 'AET processing' stamp: 'ar (auto pragmas 12/08) 10/28/1998 01:39'!
moveAETEntryFrom: index edge: edge x: xValue
	"The entry at index is not in the right position of the AET. 
	Move it to the left until the position is okay."
	| newIndex |
	<inline: false>
	newIndex := index.
	[newIndex > 0 and:[(self edgeXValueOf: (aetBuffer at: newIndex-1)) > xValue]]
		whileTrue:[	aetBuffer at: newIndex put: (aetBuffer at: newIndex-1).
					newIndex := newIndex - 1].
	aetBuffer at: newIndex put: edge.! !

!BalloonEngineBase methodsFor: 'allocating' stamp: 'ar 11/25/1998 02:19'!
needAvailableSpace: nSlots
	"Check if we have n slots available"
	GWHeaderSize + objUsed + self getUsedGet + self aetUsedGet + nSlots > self wbTopGet ifTrue:[
		self stopBecauseOf: GErrorNoMoreSpace.
		^false
	].
	^true! !

!BalloonEngineBase methodsFor: 'testing' stamp: 'ar 11/25/1998 00:21'!
needsFlush
	^self needsFlushGet ~= 0! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 11/25/1998 00:20'!
needsFlushGet
	^workBuffer at: GWNeedsFlush! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 11/25/1998 00:20'!
needsFlushPut: value
	^workBuffer at: GWNeedsFlush put: value! !

!BalloonEngineBase methodsFor: 'accessing objects' stamp: 'ar 11/24/1998 22:22'!
obj: object at: index
	^objBuffer at: object + index! !

!BalloonEngineBase methodsFor: 'accessing objects' stamp: 'ar 11/24/1998 22:22'!
obj: object at: index put: value
	^objBuffer at: object + index put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:28'!
objStartGet
	^workBuffer at: GWObjStart! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:32'!
objStartPut: value
	^workBuffer at: GWObjStart put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:28'!
objUsedGet
	^workBuffer at: GWObjUsed! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:32'!
objUsedPut: value
	^workBuffer at: GWObjUsed put: value! !

!BalloonEngineBase methodsFor: 'accessing objects' stamp: 'ar 11/24/1998 22:02'!
objectHeaderOf: obj

	^self makeUnsignedFrom:(self obj: obj at: GEObjectType)! !

!BalloonEngineBase methodsFor: 'accessing objects' stamp: 'ar 11/24/1998 22:03'!
objectIndexOf: obj

	^self obj: obj at: GEObjectIndex! !

!BalloonEngineBase methodsFor: 'accessing objects' stamp: 'ar 11/24/1998 22:03'!
objectIndexOf: obj put: value

	^self obj: obj at: GEObjectIndex put: value! !

!BalloonEngineBase methodsFor: 'accessing objects' stamp: 'ar 11/24/1998 22:03'!
objectLengthOf: obj

	^self obj: obj at: GEObjectLength! !

!BalloonEngineBase methodsFor: 'accessing objects' stamp: 'ar 11/24/1998 22:03'!
objectLengthOf: obj put: value

	^self obj: obj at: GEObjectLength put: value! !

!BalloonEngineBase methodsFor: 'accessing objects' stamp: 'ar 11/24/1998 22:03'!
objectTypeOf: obj

	^(self makeUnsignedFrom:(self obj: obj at: GEObjectType)) bitAnd: GEPrimitiveTypeMask! !

!BalloonEngineBase methodsFor: 'accessing objects' stamp: 'ar 11/24/1998 22:03'!
objectTypeOf: obj put: value

	^self obj: obj at: GEObjectType put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar (auto pragmas 12/08) 10/28/1998 16:33'!
point1Get
	<returnTypeC:'int *'>
	^self cCoerce: workBuffer + GWPoint1 to:'int *'! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar (auto pragmas 12/08) 10/28/1998 16:34'!
point2Get
	<returnTypeC:'int *'>
	^self cCoerce: workBuffer + GWPoint2 to:'int *'! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar (auto pragmas 12/08) 10/28/1998 16:34'!
point3Get
	<returnTypeC:'int *'>
	^self cCoerce: workBuffer + GWPoint3 to:'int *'! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar (auto pragmas 12/08) 11/1/1998 03:13'!
point4Get
	<returnTypeC:'int *'>
	^self cCoerce: workBuffer + GWPoint4 to:'int *'! !

!BalloonEngineBase methodsFor: 'displaying' stamp: 'ar (auto pragmas 12/08) 11/8/1998 15:13'!
postDisplayAction
	"We have just blitted a scan line to the screen.
	Do whatever seems to be a good idea here."
	"Note: In the future we may check the time needed for this scan line and interrupt processing to give the Smalltalk code a chance to run at a certain time."

	<inline: false>

	"Check if there is any more work to do."
	(self getStartGet >= self getUsedGet and:[self aetUsedGet = 0]) ifTrue:[
		"No more entries to process"
		self statePut: GEStateCompleted.
	].
	(self currentYGet >= self fillMaxYGet) ifTrue:[
		"Out of clipping range"
		self statePut: GEStateCompleted.
	].! !

!BalloonEngineBase methodsFor: 'primitives-other' stamp: 'eem 6/3/2009 16:43'!
primitiveAbortProcessing
	<export: true>
	<inline: false>
	| failureCode |
	interpreterProxy methodArgumentCount = 0
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 0)) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].
	self statePut: GEStateCompleted.
	self storeEngineStateInto: engine.! !

!BalloonEngineBase methodsFor: 'primitives-incremental' stamp: 'eem 6/3/2009 16:42'!
primitiveAddActiveEdgeEntry
	"Note: No need to load either bitBlt or spanBuffer"
	| failureCode edgeOop edge |
	<export: true>
	<inline: false>
	doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)
						requiredState: GEStateWaitingForEdge) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	edgeOop := interpreterProxy stackObjectValue: 0.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	edge := self loadEdgeStateFrom: edgeOop.
	edge = nil ifTrue:[^interpreterProxy primitiveFailFor: GEFEdgeDataTooSmall].

	(self needAvailableSpace: 1) 
		ifFalse:[^interpreterProxy primitiveFailFor: GEFWorkTooBig].

	(self edgeNumLinesOf: edge) > 0 ifTrue:[
		self insertEdgeIntoAET: edge.
	].

	engineStopped ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineStopped].

	self statePut: GEStateAddingFromGET. "Back to adding edges from GET"
	self storeEngineStateInto: engine.
	interpreterProxy pop: 1. "Leave rcvr on stack"
	doProfileStats ifTrue:[
		self incrementStat: GWCountAddAETEntry by: 1.
		self incrementStat: GWTimeAddAETEntry by: (interpreterProxy ioMicroMSecs - geProfileTime)].
! !

!BalloonEngineBase methodsFor: 'primitives-incremental' stamp: 'eem 6/3/2009 16:42'!
primitiveChangedActiveEdgeEntry
	"Note: No need to load either bitBlt or spanBuffer"
	| failureCode edgeOop edge |
	<export: true>
	<inline: false>
	doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)
						requiredState: GEStateWaitingChange) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	edgeOop := interpreterProxy stackObjectValue: 0.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	edge := self loadEdgeStateFrom: edgeOop.
	edge = nil ifTrue:[^interpreterProxy primitiveFailFor: GEFEdgeDataTooSmall].

	(self edgeNumLinesOf: edge) = 0 
		ifTrue:[	self removeFirstAETEntry]
		ifFalse:[	self resortFirstAETEntry.
				self aetStartPut: self aetStartGet + 1].

	self statePut: GEStateUpdateEdges. "Back to updating edges"
	self storeEngineStateInto: engine.
	interpreterProxy pop: 1. "Leave rcvr on stack"
	doProfileStats ifTrue:[
		self incrementStat: GWCountChangeAETEntry by: 1.
		self incrementStat: GWTimeChangeAETEntry by: (interpreterProxy ioMicroMSecs - geProfileTime)].
! !

!BalloonEngineBase methodsFor: 'primitives-other' stamp: 'eem 6/4/2009 17:02'!
primitiveCopyBuffer
	| failCode buf1 buf2 diff src dst |
	<export: true>
	<inline: false>
	<var: #src type:'int * '>
	<var: #dst type:'int * '>

	interpreterProxy methodArgumentCount = 2
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	buf2 := interpreterProxy stackValue: 0.
	buf1 := interpreterProxy stackValue: 1.
	"Make sure the old buffer is properly initialized"
	(failCode := self loadWorkBufferFrom: buf1) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failCode].
	"Make sure the buffers are of the same type"
	(interpreterProxy fetchClassOf: buf1) = (interpreterProxy fetchClassOf: buf2)
		ifFalse:[^interpreterProxy primitiveFailFor: GEFClassMismatch].
	"Make sure buf2 is at least of the size of buf1"
	diff := (interpreterProxy slotSizeOf: buf2) - (interpreterProxy slotSizeOf: buf1).
	diff < 0 ifTrue:[^interpreterProxy primitiveFailFor: GEFSizeMismatch].

	"Okay - ready for copying. First of all just copy the contents up to wbTop"
	src := workBuffer.
	dst := interpreterProxy firstIndexableField: buf2.
	0 to: self wbTopGet-1 do:[:i|
		dst at: i put: (src at: i).
	].
	"Adjust wbSize and wbTop in the new buffer"
	dst at: GWBufferTop put: self wbTopGet + diff.
	dst at: GWSize put: self wbSizeGet + diff.
	"Now copy the entries from wbTop to wbSize"
	src := src + self wbTopGet.
	dst := dst + self wbTopGet + diff.
	0 to: (self wbSizeGet - self wbTopGet - 1) do:[:i|
		dst at: i put: (src at: i).
	].
	"Okay, done. Check the new buffer by loading the state from it"
	(failCode := self loadWorkBufferFrom: buf2) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failCode].
	interpreterProxy pop: 2. "Leave rcvr on stack"
! !

!BalloonEngineBase methodsFor: 'primitives-incremental' stamp: 'eem 6/4/2009 17:56'!
primitiveDisplaySpanBuffer
	"Note: Must load bitBlt and spanBuffer"
	<export: true>
	<inline: false>
	| failureCode |
	doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
	interpreterProxy methodArgumentCount = 0
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 0)
						requiredState: GEStateBlitBuffer) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].
	"Load span buffer and bitBlt"
	(failureCode := self loadSpanBufferFrom:
		(interpreterProxy fetchPointer: BESpanIndex ofObject: engine)) = 0
			ifFalse:[^interpreterProxy primitiveFailFor: failureCode].
	(self loadBitBltFrom: 
		(interpreterProxy fetchPointer: BEBitBltIndex ofObject: engine))
			ifFalse:[^interpreterProxy primitiveFailFor: GEFBitBltLoadFailed].
	(self currentYGet bitAnd: self aaScanMaskGet) = self aaScanMaskGet ifTrue:[
		self displaySpanBufferAt: self currentYGet.
		self postDisplayAction.
	].
	self finishedProcessing ifFalse:[
		self aetStartPut: 0.
		self currentYPut: self currentYGet + 1.
		self statePut: GEStateUpdateEdges].
	self storeEngineStateInto: engine.
	doProfileStats ifTrue:[
		self incrementStat: GWCountDisplaySpan by: 1.
		self incrementStat: GWTimeDisplaySpan by: (interpreterProxy ioMicroMSecs - geProfileTime)].
! !

!BalloonEngineBase methodsFor: 'primitives-other' stamp: 'ar (auto pragmas 12/08) 5/11/2000 23:05'!
primitiveDoProfileStats
	"Turn on/off profiling. Return the old value of the flag."
	| oldValue newValue |
	<inline: false>
	<export: true>
	oldValue := doProfileStats.
	newValue := interpreterProxy stackObjectValue: 0.
	newValue := interpreterProxy booleanValueOf: newValue.
	interpreterProxy failed ifFalse:[
		doProfileStats := newValue.
		interpreterProxy pop: 2. "Pop rcvr, arg"
		interpreterProxy pushBool: oldValue.
	].! !

!BalloonEngineBase methodsFor: 'primitives-other' stamp: 'eem 6/3/2009 16:43'!
primitiveFinishedProcessing
	| finished failureCode |
	<export: true>
	<inline: false>
	doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
	interpreterProxy methodArgumentCount = 0
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 0)) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].
	finished := self finishedProcessing.
	self storeEngineStateInto: engine.
	interpreterProxy pop: 1.
	interpreterProxy pushBool: finished.
	doProfileStats ifTrue:[
		self incrementStat: GWCountFinishTest by: 1.
		self incrementStat: GWTimeFinishTest by: (interpreterProxy ioMicroMSecs - geProfileTime)].
! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 6/3/2009 16:43'!
primitiveGetAALevel
	<export: true>
	<inline: false>
	| failureCode |
	interpreterProxy methodArgumentCount = 0
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 0)) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self aaLevelGet.! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 6/3/2009 16:44'!
primitiveGetClipRect
	| failureCode rectOop pointOop |
	<export: true>
	<inline: false>

	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	rectOop := interpreterProxy stackObjectValue: 0.
	(interpreterProxy failed not
	and: [(interpreterProxy isPointers: rectOop)
	and: [(interpreterProxy slotSizeOf: rectOop) >= 2]])
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	interpreterProxy pushRemappableOop: rectOop.
	pointOop := interpreterProxy makePointwithxValue: self clipMinXGet yValue: self clipMinYGet.
	interpreterProxy storePointer: 0 ofObject: interpreterProxy topRemappableOop withValue: pointOop.
	pointOop := interpreterProxy makePointwithxValue: self clipMaxXGet yValue: self clipMaxYGet.
	rectOop := interpreterProxy popRemappableOop.
	interpreterProxy storePointer: 1 ofObject: rectOop withValue: pointOop.

	interpreterProxy pop: 2 thenPush: rectOop.! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 6/3/2009 16:44'!
primitiveGetCounts
	| failureCode statOop stats |
	<export: true>
	<inline: false>
	<var: #stats type:'int *'>

	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	statOop := interpreterProxy stackObjectValue: 0.
	(interpreterProxy failed not
	and: [(interpreterProxy isWords: statOop)
	and: [(interpreterProxy slotSizeOf: statOop) >= 9]])
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	stats := interpreterProxy firstIndexableField: statOop.
	stats at: 0 put: (stats at: 0) + (workBuffer at: GWCountInitializing).
	stats at: 1 put: (stats at: 1) + (workBuffer at: GWCountFinishTest).
	stats at: 2 put: (stats at: 2) + (workBuffer at: GWCountNextGETEntry).
	stats at: 3 put: (stats at: 3) + (workBuffer at: GWCountAddAETEntry).
	stats at: 4 put: (stats at: 4) + (workBuffer at: GWCountNextFillEntry).
	stats at: 5 put: (stats at: 5) + (workBuffer at: GWCountMergeFill).
	stats at: 6 put: (stats at: 6) + (workBuffer at: GWCountDisplaySpan).
	stats at: 7 put: (stats at: 7) + (workBuffer at: GWCountNextAETEntry).
	stats at: 8 put: (stats at: 8) + (workBuffer at: GWCountChangeAETEntry).

	interpreterProxy pop: 1. "Leave rcvr on stack"! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 6/3/2009 16:44'!
primitiveGetDepth
	<export: true>
	<inline: false>
	| failureCode |
	interpreterProxy methodArgumentCount = 0
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 0)) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self currentZGet.! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 1/30/2014 16:02'!
primitiveGetFailureReason
	"Return the reason why the last operation failed."
	<export: true>
	<inline: false>
	| failCode |
	interpreterProxy methodArgumentCount = 0
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	engine := interpreterProxy stackValue: 0.
	"Note -- don't call loadEngineFrom here because this will override the stopReason with Zero"
	(interpreterProxy isImmediate: engine) ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineIsInteger].
	(interpreterProxy isPointers: engine) ifFalse:[^interpreterProxy primitiveFailFor: GEFEngineIsWords].
	(interpreterProxy slotSizeOf: engine) < BEBalloonEngineSize ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineTooSmall].
	(failCode := self loadWorkBufferFrom: 
		(interpreterProxy fetchPointer: BEWorkBufferIndex ofObject: engine)) = 0
			ifFalse:[^interpreterProxy primitiveFailFor: failCode].
	interpreterProxy pop: 1.
	interpreterProxy pushInteger: self stopReasonGet.! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 6/3/2009 16:46'!
primitiveGetOffset
	| failureCode pointOop |
	<export: true>
	<inline: false>

	interpreterProxy methodArgumentCount = 0
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 0)) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].
	pointOop := interpreterProxy makePointwithxValue: self destOffsetXGet yValue: self destOffsetYGet.
	interpreterProxy pop: 1 thenPush: pointOop.! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 6/3/2009 16:46'!
primitiveGetTimes
	| failureCode statOop stats |
	<export: true>
	<inline: false>
	<var: #stats type:'int *'>

	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	statOop := interpreterProxy stackObjectValue: 0.
	(interpreterProxy failed not
	and: [(interpreterProxy isWords: statOop)
	and: [(interpreterProxy slotSizeOf: statOop) >= 9]])
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	stats := interpreterProxy firstIndexableField: statOop.
	stats at: 0 put: (stats at: 0) + (workBuffer at: GWTimeInitializing).
	stats at: 1 put: (stats at: 1) + (workBuffer at: GWTimeFinishTest).
	stats at: 2 put: (stats at: 2) + (workBuffer at: GWTimeNextGETEntry).
	stats at: 3 put: (stats at: 3) + (workBuffer at: GWTimeAddAETEntry).
	stats at: 4 put: (stats at: 4) + (workBuffer at: GWTimeNextFillEntry).
	stats at: 5 put: (stats at: 5) + (workBuffer at: GWTimeMergeFill).
	stats at: 6 put: (stats at: 6) + (workBuffer at: GWTimeDisplaySpan).
	stats at: 7 put: (stats at: 7) + (workBuffer at: GWTimeNextAETEntry).
	stats at: 8 put: (stats at: 8) + (workBuffer at: GWTimeChangeAETEntry).

	interpreterProxy pop: 1. "Leave rcvr on stack"! !

!BalloonEngineBase methodsFor: 'primitives-other' stamp: 'eem 10/22/2018 13:45'!
primitiveInitializeBuffer
	<export: true>
	| wbOop size |
	(interpreterProxy methodArgumentCount = 1
	 and: [(interpreterProxy isWords: (wbOop := interpreterProxy stackValue: 0)) 
	 and: [(size := interpreterProxy slotSizeOf: wbOop) >= GWMinimalSize]]) ifFalse:
		[^interpreterProxy primitiveFail].
	self workBufferPut: wbOop.
	objBuffer := workBuffer + GWHeaderSize.
	self magicNumberPut: GWMagicNumber.
	self wbSizePut: size.
	self wbTopPut: size.
	self statePut: GEStateUnlocked.
	self objStartPut: GWHeaderSize.
	self objUsedPut: 4.	"Dummy fill object"
	self objectTypeOf: 0 put: GEPrimitiveFill.
	self objectLengthOf: 0 put: 4.
	self objectIndexOf: 0 put: 0.
	self getStartPut: 0.
	self getUsedPut: 0.
	self aetStartPut: 0.
	self aetUsedPut: 0.
	self stopReasonPut: 0.
	self needsFlushPut: 0.
	self clipMinXPut: 0.
	self clipMaxXPut: 0.
	self clipMinYPut: 0.
	self clipMaxYPut: 0.
	self currentZPut: 0.
	self resetGraphicsEngineStats.
	self initEdgeTransform.
	self initColorTransform.
	interpreterProxy pop: 2 thenPush: wbOop! !

!BalloonEngineBase methodsFor: 'primitives-incremental' stamp: 'eem 6/3/2009 16:41'!
primitiveInitializeProcessing
	"Note: No need to load bitBlt but must load spanBuffer"
	<export: true>
	<inline: false>
	| failureCode |
	doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
	interpreterProxy methodArgumentCount = 0
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 0)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].
	"Load span buffer for clear operation"
	(failureCode := self loadSpanBufferFrom:
		(interpreterProxy fetchPointer: BESpanIndex ofObject: engine)) = 0
			ifFalse:[^interpreterProxy primitiveFailFor: failureCode].
	self initializeGETProcessing.
	engineStopped ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineStopped].
	self statePut: GEStateAddingFromGET. "Initialized"
	interpreterProxy failed ifFalse:[self storeEngineStateInto: engine].
	doProfileStats ifTrue:[
		self incrementStat: GWCountInitializing by: 1.
		self incrementStat: GWTimeInitializing by: (interpreterProxy ioMicroMSecs - geProfileTime)].
! !

!BalloonEngineBase methodsFor: 'primitives-incremental' stamp: 'eem 6/3/2009 16:41'!
primitiveMergeFillFrom
	"Note: No need to load bitBlt but must load spanBuffer"
	| failureCode fillOop bitsOop value |
	<export: true>
	<inline: false>
	doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
	interpreterProxy methodArgumentCount = 2
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 2)
						requiredState: GEStateWaitingForFill) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].
	"Load span buffer for merging the fill"
	(failureCode := self loadSpanBufferFrom:
		(interpreterProxy fetchPointer: BESpanIndex ofObject: engine)) = 0
			ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	fillOop := interpreterProxy stackObjectValue: 0.
	bitsOop := interpreterProxy stackObjectValue: 1.
	"Check bitmap"
	(interpreterProxy failed not
	and: [(interpreterProxy fetchClassOf: bitsOop) = interpreterProxy classBitmap])
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	"Check fillOop"
	(interpreterProxy slotSizeOf: fillOop) < FTBalloonFillDataSize
		ifTrue:[^interpreterProxy primitiveFailFor: GEFFillDataTooSmall].
	"Check if this was the fill we have exported"
	value := interpreterProxy fetchInteger: FTIndexIndex ofObject: fillOop.
	(self objectIndexOf: self lastExportedFillGet) = value
		ifFalse:[^interpreterProxy primitiveFailFor: GEFWrongFill].
	value := interpreterProxy fetchInteger: FTMinXIndex ofObject: fillOop.
	self lastExportedLeftXGet = value
		ifFalse:[^interpreterProxy primitiveFailFor: GEFWrongFill].
	value := interpreterProxy fetchInteger: FTMaxXIndex ofObject: fillOop.
	self lastExportedRightXGet = value
		ifFalse:[^interpreterProxy primitiveFailFor: GEFWrongFill].

	(interpreterProxy slotSizeOf: bitsOop) < (self lastExportedRightXGet - self lastExportedLeftXGet)
		ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	interpreterProxy failed ifTrue:[^nil].

	self fillBitmapSpan: (interpreterProxy firstIndexableField: bitsOop)
		from: self lastExportedLeftXGet
		to: self lastExportedRightXGet.

	self statePut: GEStateScanningAET. "Back to scanning AET"
	self storeEngineStateInto: engine.
	interpreterProxy pop: 2. "Leave rcvr on stack"
	doProfileStats ifTrue:[
		self incrementStat: GWCountMergeFill by: 1.
		self incrementStat: GWTimeMergeFill by: (interpreterProxy ioMicroMSecs - geProfileTime)].
! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 6/3/2009 16:46'!
primitiveNeedsFlush
	| failureCode needFlush |
	<export: true>
	<inline: false>
	interpreterProxy methodArgumentCount = 0
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 0)) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].
	needFlush := self needsFlush.
	self storeEngineStateInto: engine.
	interpreterProxy pop: 1.
	interpreterProxy pushBool: needFlush.! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 6/3/2009 16:47'!
primitiveNeedsFlushPut
	| failureCode needFlush |
	<export: true>
	<inline: false>
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	needFlush := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	needFlush == true 
		ifTrue:[self needsFlushPut: 1]
		ifFalse:[self needsFlushPut: 0].
	self storeEngineStateInto: engine.
	interpreterProxy pop: 1. "Leave rcvr on stack"
! !

!BalloonEngineBase methodsFor: 'primitives-incremental' stamp: 'eem 6/3/2009 16:49'!
primitiveNextActiveEdgeEntry
	"Note: No need to load either bitBlt or spanBuffer"
	| failureCode edgeOop hasEdge edge |
	<export: true>
	<inline: false>
	doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)
						requiredState: GEStateUpdateEdges or: GEStateCompleted) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	edgeOop := interpreterProxy stackObjectValue: 0.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	hasEdge := false.
	self stateGet = GEStateCompleted ifFalse:[
		hasEdge := self findNextExternalUpdateFromAET.
		hasEdge ifTrue:[
			edge := aetBuffer at: self aetStartGet.
			self storeEdgeStateFrom: edge into: edgeOop.
			"Do not advance to the next aet entry yet"
			"self aetStartPut: self aetStartGet + 1."
			self statePut: GEStateWaitingChange. "Wait for changed edge"
		] ifFalse:[self statePut: GEStateAddingFromGET]. "Start over"
	].
	interpreterProxy failed ifTrue:[^nil].

	self storeEngineStateInto: engine.

	interpreterProxy pop: 2.
	interpreterProxy pushBool: hasEdge not.
	doProfileStats ifTrue:[
		self incrementStat: GWCountNextAETEntry by: 1.
		self incrementStat: GWTimeNextAETEntry by: (interpreterProxy ioMicroMSecs - geProfileTime)].
! !

!BalloonEngineBase methodsFor: 'primitives-incremental' stamp: 'eem 6/4/2009 18:08'!
primitiveNextFillEntry
	"Note: No need to load bitBlt but must load spanBuffer"
	| failureCode fillOop hasFill |
	<export: true>
	<inline: false>
	doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)
						requiredState: GEStateScanningAET) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].
	"Load span buffer for internal handling of fills"
	(failureCode := self loadSpanBufferFrom:
		(interpreterProxy fetchPointer: BESpanIndex ofObject: engine)) = 0
			ifFalse:[^interpreterProxy primitiveFailFor: failureCode].
	(self loadFormsFrom:
		(interpreterProxy fetchPointer: BEFormsIndex ofObject: engine))
			ifFalse:[^interpreterProxy primitiveFailFor: GEFFormLoadFailed].

	"Check if we have to clear the span buffer before proceeding"
	(self clearSpanBufferGet = 0) ifFalse:[
		(self currentYGet bitAnd: self aaScanMaskGet) = 0
			ifTrue:[self clearSpanBuffer].
		self clearSpanBufferPut: 0].

	fillOop := interpreterProxy stackObjectValue: 0.
	hasFill := self findNextExternalFillFromAET.
	engineStopped ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineStopped].
	hasFill ifTrue:[self storeFillStateInto: fillOop].
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: GEFWrongFill].
	hasFill
		ifTrue:[	self statePut: GEStateWaitingForFill]
		ifFalse:[	self wbStackClear.
				self spanEndAAPut: 0.
				self statePut: GEStateBlitBuffer].
	self storeEngineStateInto: engine.
	interpreterProxy pop: 2.
	interpreterProxy pushBool: hasFill not.
	doProfileStats ifTrue:[
		self incrementStat: GWCountNextFillEntry by: 1.
		self incrementStat: GWTimeNextFillEntry by: (interpreterProxy ioMicroMSecs - geProfileTime)]! !

!BalloonEngineBase methodsFor: 'primitives-incremental' stamp: 'eem 6/3/2009 16:40'!
primitiveNextGlobalEdgeEntry
	"Note: No need to load either bitBlt or spanBuffer"
	| failureCode edgeOop hasEdge edge |
	<export: true>
	<inline: false>
	doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)
						requiredState: GEStateAddingFromGET) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	edgeOop := interpreterProxy stackObjectValue: 0.
	hasEdge := self findNextExternalEntryFromGET.
	hasEdge ifTrue:[
		edge := getBuffer at: self getStartGet.
		self storeEdgeStateFrom: edge into: edgeOop.
		self getStartPut: self getStartGet + 1].

	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: GEFWrongEdge].

	hasEdge
		ifTrue:[	self statePut: GEStateWaitingForEdge] "Wait for adding edges"
		ifFalse:[ "Start scanning the AET"
				self statePut: GEStateScanningAET.
				self clearSpanBufferPut: 1. "Clear span buffer at next entry"
				self aetStartPut: 0.
				self wbStackClear].
	self storeEngineStateInto: engine.

	interpreterProxy pop: 2.
	interpreterProxy pushBool: hasEdge not.
	doProfileStats ifTrue:[
		self incrementStat: GWCountNextGETEntry by: 1.
		self incrementStat: GWTimeNextGETEntry by: (interpreterProxy ioMicroMSecs - geProfileTime)].
! !

!BalloonEngineBase methodsFor: 'primitives-other' stamp: 'eem 6/4/2009 18:16'!
primitiveRegisterExternalEdge
	| failureCode rightFillIndex leftFillIndex initialZ initialY initialX index  edge |
	<export: true>
	<inline: false>
	interpreterProxy methodArgumentCount = 6 
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 6)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	rightFillIndex := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	leftFillIndex := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 1).
	initialZ := interpreterProxy stackIntegerValue: 2.
	initialY := interpreterProxy stackIntegerValue: 3.
	initialX := interpreterProxy stackIntegerValue: 4.
	index := interpreterProxy stackIntegerValue: 5.
	interpreterProxy failed
		ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	(self allocateObjEntry: GEBaseEdgeSize) 
		ifFalse:[^interpreterProxy primitiveFailFor: GEFWorkTooBig].

	"Make sure the fills are okay"
	((self isFillOkay: leftFillIndex)
	and: [self isFillOkay: rightFillIndex])
		ifFalse:[^interpreterProxy primitiveFailFor: GEFWrongFill].

	edge := objUsed.
	objUsed := edge + GEBaseEdgeSize.
	"Install type and length"
	self objectTypeOf: edge put: GEPrimitiveEdge.
	self objectLengthOf: edge put: GEBaseEdgeSize.
	self objectIndexOf: edge put: index.
	"Install remaining stuff"
	self edgeXValueOf: edge put: initialX.
	self edgeYValueOf: edge put: initialY.
	self edgeZValueOf: edge put: initialZ.
	self edgeLeftFillOf: edge put: (self transformColor: leftFillIndex).
	self edgeRightFillOf: edge put: (self transformColor: rightFillIndex).
	engineStopped ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineStopped].

	interpreterProxy failed ifFalse:[
		self storeEngineStateInto: engine.
		interpreterProxy pop: 6. "Leave rcvr on stack"
	].! !

!BalloonEngineBase methodsFor: 'primitives-other' stamp: 'eem 6/3/2009 16:39'!
primitiveRegisterExternalFill
	| failureCode index  fill |
	<export: true>
	<inline: false>
	interpreterProxy methodArgumentCount = 1 
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	index := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	"Note: We *must* not allocate any fill with index 0"
	fill := 0.
	[fill = 0] whileTrue:[
		(self allocateObjEntry: GEBaseEdgeSize) 
			ifFalse:[^interpreterProxy primitiveFailFor: GEFWorkTooBig].
		fill := objUsed.
		objUsed := fill + GEBaseFillSize.
		"Install type and length"
		self objectTypeOf: fill put: GEPrimitiveFill.
		self objectLengthOf: fill put: GEBaseFillSize.
		self objectIndexOf: fill put: index.
	].

	interpreterProxy failed ifFalse:[
		self storeEngineStateInto: engine.
		interpreterProxy pop: 2.
		interpreterProxy pushInteger: fill.
	].! !

!BalloonEngineBase methodsFor: 'primitives-rendering' stamp: 'eem 6/8/2009 10:46'!
primitiveRenderImage
	"Start/Proceed rendering the entire image"
	<export: true>
	<inline: false>
	| failCode |
	(failCode := self loadRenderingState) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failCode].

	self proceedRenderingScanline. "Finish this scan line"
	engineStopped ifTrue:[^self storeRenderingState].
	self proceedRenderingImage. "And go on as usual"

	self storeRenderingState.! !

!BalloonEngineBase methodsFor: 'primitives-rendering' stamp: 'eem 6/8/2009 10:47'!
primitiveRenderScanline
	"Start rendering the entire image"
	<export: true>
	<inline: false>
	| failCode |
	(failCode := self loadRenderingState) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failCode].

	self proceedRenderingScanline. "Finish the current scan line"

	self storeRenderingState.! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 6/4/2009 17:09'!
primitiveSetAALevel
	| failureCode level |
	<export: true>
	<inline: false>
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].
	level := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed
		ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	self setAALevel: level.
	self storeEngineStateInto: engine.
	interpreterProxy pop: 1. "Leave rcvr on stack"! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:38'!
primitiveSetBitBltPlugin
	"Primitive. Set the BitBlt plugin to use."
	| pluginName length ptr needReload |
	<export: true>
	<var: #ptr type:'char *'>
	pluginName := interpreterProxy stackValue: 0.
	"Must be string to work"
	(interpreterProxy isBytes: pluginName) 
		ifFalse:[^interpreterProxy primitiveFail].
	length := interpreterProxy byteSizeOf: pluginName.
	length >= 256 
		ifTrue:[^interpreterProxy primitiveFail].
	ptr := interpreterProxy firstIndexableField: pluginName.
	needReload := false.
	0 to: length-1 do:[:i|
		"Compare and store the plugin to be used"
		(bbPluginName at: i) = (ptr at: i) ifFalse:[
			bbPluginName at: i put: (ptr at: i).
			needReload := true]].
	(bbPluginName at: length) = 0 ifFalse:[
		bbPluginName at: length put: 0.
		needReload := true].
	needReload ifTrue:[
		self initialiseModule 
			ifFalse:[^interpreterProxy primitiveFail]].
	interpreterProxy pop: 1. "Return receiver"! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 6/3/2009 16:39'!
primitiveSetClipRect
	| failureCode rectOop |
	<export: true>
	<inline: false>

	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	rectOop := interpreterProxy stackObjectValue: 0.
	(interpreterProxy failed not
	and: [(interpreterProxy isPointers: rectOop)
	and: [(interpreterProxy slotSizeOf: rectOop) >= 2]])
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	self loadPoint: self point1Get from: (interpreterProxy fetchPointer: 0 ofObject: rectOop).
	self loadPoint: self point2Get from: (interpreterProxy fetchPointer: 1 ofObject: rectOop).
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	self clipMinXPut: (self point1Get at: 0).
	self clipMinYPut: (self point1Get at: 1).
	self clipMaxXPut: (self point2Get at: 0).
	self clipMaxYPut: (self point2Get at: 1).
	self storeEngineStateInto: engine.
	interpreterProxy pop: 1. "Leave rcvr on stack"! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 6/3/2009 17:25'!
primitiveSetColorTransform
	| failureCode transformOop |
	<export: true>
	<inline: false>

	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	transformOop := interpreterProxy stackObjectValue: 0.
	interpreterProxy failed
		ifTrue: [^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	self loadColorTransformFrom: transformOop.
	interpreterProxy failed
		ifTrue: [^interpreterProxy primitiveFailFor: GEFEntityLoadFailed].
	self storeEngineStateInto: engine.
	interpreterProxy pop: 1. "Leave rcvr on stack"! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 6/3/2009 16:38'!
primitiveSetDepth
	| failureCode depth |
	<export: true>
	<inline: false>

	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	depth := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed
		ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	self currentZPut: depth.

	self storeEngineStateInto: engine.
	interpreterProxy pop: 1. "Leave rcvr on stack"! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 6/4/2009 18:16'!
primitiveSetEdgeTransform
	| failureCode transformOop |
	<export: true>
	<inline: false>

	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	transformOop := interpreterProxy stackObjectValue: 0.
	interpreterProxy failed
		ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	self loadEdgeTransformFrom: transformOop.
	interpreterProxy failed
		ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	self storeEngineStateInto: engine.
	interpreterProxy pop: 1. "Leave rcvr on stack"! !

!BalloonEngineBase methodsFor: 'primitives-access' stamp: 'eem 6/3/2009 16:50'!
primitiveSetOffset
	| failureCode pointOop |
	<export: true>
	<inline: false>

	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	pointOop := interpreterProxy stackValue: 0.
	(interpreterProxy fetchClassOf: pointOop) = interpreterProxy classPoint
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	self loadPoint: self point1Get from: pointOop.
	interpreterProxy failed
		ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	self destOffsetXPut: (self point1Get at: 0).
	self destOffsetYPut: (self point1Get at: 1).
	self storeEngineStateInto: engine.
	interpreterProxy pop: 1. "Leave rcvr on stack"! !

!BalloonEngineBase methodsFor: 'primitives-rendering' stamp: 'ar (auto pragmas 12/08) 5/13/2000 15:00'!
proceedRenderingImage
	"This is the main rendering entry"
	| external |
	<inline: false>
	[self finishedProcessing] whileFalse:[
		doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
		external := self findNextExternalEntryFromGET.
		doProfileStats ifTrue:[
			self incrementStat: GWCountNextGETEntry by: 1.
			self incrementStat: GWTimeNextGETEntry by: (interpreterProxy ioMicroMSecs - geProfileTime)].
		engineStopped ifTrue:[^self statePut: GEStateAddingFromGET].
		external ifTrue:[
			self statePut: GEStateWaitingForEdge.
			^self stopBecauseOf: GErrorGETEntry.
		]. 
		self aetStartPut: 0.
		self wbStackClear.
		self clearSpanBufferPut: 1.

		doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
		(self clearSpanBufferGet ~= 0 and:[(self currentYGet bitAnd: self aaScanMaskGet) = 0])
			ifTrue:[self clearSpanBuffer].
		self clearSpanBufferPut: 0.
		external := self findNextExternalFillFromAET.
		doProfileStats ifTrue:[
			self incrementStat: GWCountNextFillEntry by: 1.
			self incrementStat: GWTimeNextFillEntry by: (interpreterProxy ioMicroMSecs - geProfileTime)].
		engineStopped ifTrue:[^self statePut: GEStateScanningAET].
		external ifTrue:[
			self statePut: GEStateWaitingForFill.
			^self stopBecauseOf: GErrorFillEntry.
		].
		self wbStackClear.
		self spanEndAAPut: 0.

		doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
		(self currentYGet bitAnd: self aaScanMaskGet) = self aaScanMaskGet ifTrue:[
			self displaySpanBufferAt: self currentYGet.
			self postDisplayAction.
		].
		doProfileStats ifTrue:[
			self incrementStat: GWCountDisplaySpan by: 1.
			self incrementStat: GWTimeDisplaySpan by: (interpreterProxy ioMicroMSecs - geProfileTime)].
		engineStopped ifTrue:[^self statePut: GEStateBlitBuffer].
		self finishedProcessing ifTrue:[^0].
		self aetStartPut: 0.
		self currentYPut: self currentYGet + 1.

		doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
		external := self findNextExternalUpdateFromAET.
		doProfileStats ifTrue:[
			self incrementStat: GWCountNextAETEntry by: 1.
			self incrementStat: GWTimeNextAETEntry by: (interpreterProxy ioMicroMSecs - geProfileTime)].
		engineStopped ifTrue:[^self statePut: GEStateUpdateEdges].
		external ifTrue:[
			self statePut: GEStateWaitingChange.
			^self stopBecauseOf: GErrorAETEntry.
		].
	].! !

!BalloonEngineBase methodsFor: 'primitives-rendering' stamp: 'ar (auto pragmas 12/08) 5/13/2000 15:00'!
proceedRenderingScanline
	"Proceed rendering the current scan line.
	This method may be called after some Smalltalk code has been executed inbetween."
	"This is the main rendering entry"
	| external state |
	<inline: false>
	state := self stateGet.

	state = GEStateUnlocked ifTrue:[
		self initializeGETProcessing.
		engineStopped ifTrue:[^0].
		state := GEStateAddingFromGET.
	]. 

	state = GEStateAddingFromGET ifTrue:[
		doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
		external := self findNextExternalEntryFromGET.
		doProfileStats ifTrue:[
			self incrementStat: GWCountNextGETEntry by: 1.
			self incrementStat: GWTimeNextGETEntry by: (interpreterProxy ioMicroMSecs - geProfileTime)].
		engineStopped ifTrue:[^self statePut: GEStateAddingFromGET].
		external ifTrue:[
			self statePut: GEStateWaitingForEdge.
			^self stopBecauseOf: GErrorGETEntry.
		]. 
		self aetStartPut: 0.
		self wbStackClear.
		self clearSpanBufferPut: 1.
		state := GEStateScanningAET.
	].

	state = GEStateScanningAET ifTrue:[
		doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
		(self clearSpanBufferGet ~= 0 and:[(self currentYGet bitAnd: self aaScanMaskGet) = 0])
			ifTrue:[self clearSpanBuffer].
		self clearSpanBufferPut: 0.
		external := self findNextExternalFillFromAET.
		doProfileStats ifTrue:[
			self incrementStat: GWCountNextFillEntry by: 1.
			self incrementStat: GWTimeNextFillEntry by: (interpreterProxy ioMicroMSecs - geProfileTime)].
		engineStopped ifTrue:[^self statePut: GEStateScanningAET].
		external ifTrue:[
			self statePut: GEStateWaitingForFill.
			^self stopBecauseOf: GErrorFillEntry.
		].
		state := GEStateBlitBuffer.
		self wbStackClear.
		self spanEndAAPut: 0.
	].

	state = GEStateBlitBuffer ifTrue:[
		doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
		(self currentYGet bitAnd: self aaScanMaskGet) = self aaScanMaskGet ifTrue:[
			self displaySpanBufferAt: self currentYGet.
			self postDisplayAction.
		].
		doProfileStats ifTrue:[
			self incrementStat: GWCountDisplaySpan by: 1.
			self incrementStat: GWTimeDisplaySpan by: (interpreterProxy ioMicroMSecs - geProfileTime)].
		engineStopped ifTrue:[^self statePut: GEStateBlitBuffer].
		self finishedProcessing ifTrue:[^0].
		state := GEStateUpdateEdges.
		self aetStartPut: 0.
		self currentYPut: self currentYGet + 1.
	].

	state = GEStateUpdateEdges ifTrue:[
		doProfileStats ifTrue:[geProfileTime := interpreterProxy ioMicroMSecs].
		external := self findNextExternalUpdateFromAET.
		doProfileStats ifTrue:[
			self incrementStat: GWCountNextAETEntry by: 1.
			self incrementStat: GWTimeNextAETEntry by: (interpreterProxy ioMicroMSecs - geProfileTime)].
		engineStopped ifTrue:[^self statePut: GEStateUpdateEdges].
		external ifTrue:[
			self statePut: GEStateWaitingChange.
			^self stopBecauseOf: GErrorAETEntry.
		].
		self statePut: GEStateAddingFromGET.
	].! !

!BalloonEngineBase methodsFor: 'loading state' stamp: 'eem 9/20/2013 18:06'!
quickLoadEngineFrom: engineOop
	"Load the minimal required state from the engineOop, e.g., just the work buffer.
	 Answer 0 on success or non-zero a failure code on failure"
	<inline: false>
	| failCode |
	interpreterProxy failed ifTrue:[^GEFAlreadyFailed].
	(interpreterProxy isImmediate: engineOop) ifTrue:[^GEFEngineIsInteger].
	(interpreterProxy isPointers: engineOop) ifFalse:[^GEFEngineIsWords].
	(interpreterProxy slotSizeOf: engineOop) < BEBalloonEngineSize ifTrue:[^GEFEngineTooSmall].
	engine := engineOop.
	(failCode := self loadWorkBufferFrom: (interpreterProxy
											fetchPointer: BEWorkBufferIndex
											ofObject: engineOop)) = 0
		ifFalse: [^failCode].
	self stopReasonPut: 0.
	objUsed := self objUsedGet.
	engineStopped := false.
	^0! !

!BalloonEngineBase methodsFor: 'loading state' stamp: 'eem 6/3/2009 13:56'!
quickLoadEngineFrom: oop requiredState: requiredState
	<inline: false>
	| failureCode |
	(failureCode := self quickLoadEngineFrom: oop) = 0 ifFalse:[^failureCode].
	self stateGet = requiredState ifTrue:[^0].
	self stopReasonPut: GErrorBadState.
	^GEFWrongState! !

!BalloonEngineBase methodsFor: 'loading state' stamp: 'eem 6/3/2009 13:56'!
quickLoadEngineFrom: oop requiredState: requiredState or: alternativeState
	<inline: false>
	| failureCode |
	(failureCode := self quickLoadEngineFrom: oop) = 0 ifFalse:[^failureCode].
	self stateGet = requiredState ifTrue:[^0].
	self stateGet = alternativeState ifTrue:[^0].
	self stopReasonPut: GErrorBadState.
	^GEFWrongState! !

!BalloonEngineBase methodsFor: 'FILL processing' stamp: 'ar 11/25/1998 15:16'!
quickRemoveInvalidFillsAt: leftX
	"Remove any top fills if they have become invalid."
	self stackFillSize = 0 ifTrue:[^nil].
	[self topRightX <= leftX] whileTrue:[
		self hideFill: self topFill depth: self topDepth.
		self stackFillSize = 0 ifTrue:[^nil].
	].! !

!BalloonEngineBase methodsFor: 'GET processing' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:38'!
quickSortGlobalEdgeTable: array from: i to: j 
	"Sort elements i through j of self to be nondescending according to
	sortBlock."
	"Note: The original loop has been heavily re-written for C translation"
	| di dij dj tt ij k l n tmp again before |
	<var: #array type:'int *'>
	<inline: false>
	"The prefix d means the data at that index."
	(n := j + 1  - i) <= 1 ifTrue: [^0].	"Nothing to sort." 
	 "Sort di,dj."
	di := array at: i.
	dj := array at: j.
	before := self getSorts: di before: dj. "i.e., should di precede dj?"
	before ifFalse:[
		tmp := array at: i.
		array at: i put: (array at: j).
		array at: j put: tmp.
		tt := di.	di := dj.	dj := tt].
	n <= 2 ifTrue:[^0].

	"More than two elements."
	ij := (i + j) // 2.  "ij is the midpoint of i and j."
	dij := array at: ij.  "Sort di,dij,dj.  Make dij be their median."
	before := (self getSorts: di before: dij). "i.e. should di precede dij?"
	before ifTrue:[
		before := (self getSorts: dij before: dj). "i.e., should dij precede dj?"
		before ifFalse:["i.e., should dij precede dj?"
			tmp := array at: j.
			array at: j put: (array at: ij).
			array at: ij put: tmp.
			dij := dj]
	] ifFalse:[  "i.e. di should come after dij"
		tmp := array at: i.
		array at: i put: (array at: ij).
		array at: ij put: tmp.
		 dij := di].
	n <= 3 ifTrue:[^0].

	 "More than three elements."
	"Find k>i and l<j such that dk,dij,dl are in reverse order.
	Swap k and l.  Repeat this procedure until k and l pass each other."
	k := i.
	l := j.

	again := true.
	[again] whileTrue:[
		before := true.
		[before] whileTrue:[
			k <= (l := l - 1)
				ifTrue:[	tmp := array at: l.
						before := self getSorts: dij before: tmp]
				ifFalse:[before := false].
		].
		before := true.
		[before] whileTrue:[
			(k := k + 1) <= l
				ifTrue:[	tmp := array at: k.
						before := self getSorts: tmp before: dij]
				ifFalse:[before := false]].

		again := k <= l.
		again ifTrue:[
			tmp := array at: k.
			array at: k put: (array at: l).
			array at: l put: tmp]].

	"Now l<k (either 1 or 2 less), and di through dl are all less than or equal to dk
	through dj.  Sort those two segments."
	self quickSortGlobalEdgeTable: array from: i to: l.
	self quickSortGlobalEdgeTable: array from: k to: j.! !

!BalloonEngineBase methodsFor: 'AET processing' stamp: 'ar (auto pragmas 12/08) 10/28/1998 21:07'!
removeFirstAETEntry
	| index |
	<inline: false>
	index := self aetStartGet.
	self aetUsedPut: self aetUsedGet - 1.
	[index < self aetUsedGet] whileTrue:[
		aetBuffer at: index put: (aetBuffer at: index + 1).
		index := index + 1.
	].! !

!BalloonEngineBase methodsFor: 'other' stamp: 'ar (auto pragmas 12/08) 11/7/1998 14:26'!
resetGraphicsEngineStats
	<inline: false>
	workBuffer at: GWTimeInitializing put: 0.
	workBuffer at: GWTimeFinishTest put: 0.
	workBuffer at: GWTimeNextGETEntry put: 0.
	workBuffer at: GWTimeAddAETEntry put: 0.
	workBuffer at: GWTimeNextFillEntry put: 0.
	workBuffer at: GWTimeMergeFill put: 0.
	workBuffer at: GWTimeDisplaySpan put: 0.
	workBuffer at: GWTimeNextAETEntry put: 0.
	workBuffer at: GWTimeChangeAETEntry put: 0.

	workBuffer at: GWCountInitializing put: 0.
	workBuffer at: GWCountFinishTest put: 0.
	workBuffer at: GWCountNextGETEntry put: 0.
	workBuffer at: GWCountAddAETEntry put: 0.
	workBuffer at: GWCountNextFillEntry put: 0.
	workBuffer at: GWCountMergeFill put: 0.
	workBuffer at: GWCountDisplaySpan put: 0.
	workBuffer at: GWCountNextAETEntry put: 0.
	workBuffer at: GWCountChangeAETEntry put: 0.

	workBuffer at: GWBezierMonotonSubdivisions put: 0.
	workBuffer at: GWBezierHeightSubdivisions put: 0.
	workBuffer at: GWBezierOverflowSubdivisions put: 0.
	workBuffer at: GWBezierLineConversions put: 0.
! !

!BalloonEngineBase methodsFor: 'AET processing' stamp: 'ar (auto pragmas 12/08) 10/28/1998 21:07'!
resortFirstAETEntry
	| edge xValue leftEdge |
	<inline: false>
	self aetStartGet = 0 ifTrue:[^nil]. "Nothing to resort"
	edge := aetBuffer at: self aetStartGet.
	xValue := self edgeXValueOf: edge.
	leftEdge := aetBuffer at: (self aetStartGet - 1).
	(self edgeXValueOf: leftEdge) <= xValue ifTrue:[^nil]. "Okay"
	self moveAETEntryFrom: self aetStartGet edge: edge x: xValue.! !

!BalloonEngineBase methodsFor: 'other' stamp: 'ar (auto pragmas 12/08) 11/7/1998 14:26'!
setAALevel: level
	"Set the anti-aliasing level. Three levels are supported:
		1 - No antialiasing
		2 - 2x2 unweighted anti-aliasing
		4 - 4x4 unweighted anti-aliasing.
	"
	| aaLevel |
	<inline: false>
	level >= 4 ifTrue:[aaLevel := 4].
	(level >= 2) & (level < 4) ifTrue:[aaLevel := 2].
	level < 2 ifTrue:[aaLevel := 1].
	self aaLevelPut: aaLevel.
	aaLevel = 1 ifTrue:[
		self aaShiftPut: 0.
		self aaColorMaskPut: 16rFFFFFFFF.
		self aaScanMaskPut: 0.
	].
	aaLevel = 2 ifTrue:[
		self aaShiftPut: 1.
		self aaColorMaskPut: 16rFCFCFCFC.
		self aaScanMaskPut: 1.
	].
	aaLevel = 4 ifTrue:[
		self aaShiftPut: 2.
		self aaColorMaskPut: 16rF0F0F0F0.
		self aaScanMaskPut: 3.
	].
	self aaColorShiftPut: self aaShiftGet * 2.
	self aaHalfPixelPut: self aaShiftGet.
! !

!BalloonEngineBase methodsFor: 'FILL processing' stamp: 'ar (auto pragmas 12/08) 11/25/1998 14:38'!
showFill: fillIndex depth: depth rightX: rightX
	<inline: false>
	(self allocateStackFillEntry) ifFalse:[^nil]. "Insufficient space"
	self stackFillValue: 0 put: fillIndex.
	self stackFillDepth: 0 put: depth.
	self stackFillRightX: 0 put: rightX.
	self stackFillSize = self stackFillEntryLength ifTrue:[^nil]. "No need to update"

	(self fillSorts: 0 before: self stackFillSize - self stackFillEntryLength) ifTrue:[
		"New top fill"
		self stackFillValue: 0 put: self topFillValue.
		self stackFillDepth: 0 put: self topFillDepth.
		self stackFillRightX: 0 put: self topFillRightX.
		self topFillValuePut: fillIndex.
		self topFillDepthPut: depth.
		self topFillRightXPut: rightX.
	].! !

!BalloonEngineBase methodsFor: 'other' stamp: 'ar (auto pragmas 12/08) 11/8/1998 15:25'!
smallSqrtTable
	| theTable |
	<inline: false>
	<returnTypeC:'int *'>
	<var: #theTable declareC:'static int theTable[32] = 
	{0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6}'>
	^theTable! !

!BalloonEngineBase methodsFor: 'GET processing' stamp: 'ar 10/27/1998 23:34'!
sortGlobalEdgeTable
	"Sort the entire global edge table"
	self quickSortGlobalEdgeTable: getBuffer from: 0 to: self getUsedGet-1.! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:28'!
spanEndAAGet
	^workBuffer at: GWSpanEndAA! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:31'!
spanEndAAPut: value
	^workBuffer at: GWSpanEndAA put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:28'!
spanEndGet
	^workBuffer at: GWSpanEnd! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:31'!
spanEndPut: value
	^workBuffer at: GWSpanEnd put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:28'!
spanSizeGet
	^workBuffer at: GWSpanSize! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:31'!
spanSizePut: value
	^workBuffer at: GWSpanSize put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:28'!
spanStartGet
	^workBuffer at: GWSpanStart! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:31'!
spanStartPut: value
	^workBuffer at: GWSpanStart put: value! !

!BalloonEngineBase methodsFor: 'other' stamp: 'ar 11/8/1998 20:57'!
squaredLengthOf: deltaX with: deltaY
	^(deltaX * deltaX) + (deltaY * deltaY)! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 10/30/1998 19:08'!
stackFillDepth: index
	^self wbStackValue: index+1! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 10/30/1998 19:08'!
stackFillDepth: index put: value
	^self wbStackValue: index+1 put: value! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 11/25/1998 14:31'!
stackFillEntryLength
	^3! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 11/25/1998 14:35'!
stackFillRightX: index
	^self wbStackValue: index+2! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 11/25/1998 14:35'!
stackFillRightX: index put: value
	^self wbStackValue: index+2 put: value! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 10/30/1998 19:10'!
stackFillSize
	^self wbStackSize! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 10/30/1998 19:09'!
stackFillValue: index
	^self wbStackValue: index! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 10/30/1998 19:09'!
stackFillValue: index put: value
	^self wbStackValue: index put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:28'!
stateGet
	^workBuffer at: GWState! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:31'!
statePut: value
	^workBuffer at: GWState put: value! !

!BalloonEngineBase methodsFor: 'other' stamp: 'ar 11/25/1998 02:22'!
stopBecauseOf: stopReason
	self stopReasonPut: stopReason.
	engineStopped := true.! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:28'!
stopReasonGet
	^workBuffer at: GWStopReason! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:31'!
stopReasonPut: value
	^workBuffer at: GWStopReason put: value! !

!BalloonEngineBase methodsFor: 'storing state' stamp: 'ar (auto pragmas 12/08) 11/11/1998 22:21'!
storeEdgeStateFrom: edge into: edgeOop

	<inline: false>
	(interpreterProxy slotSizeOf: edgeOop) < ETBalloonEdgeDataSize 
		ifTrue:[^interpreterProxy primitiveFail].
	interpreterProxy storeInteger: ETIndexIndex ofObject: edgeOop withValue: 
		(self objectIndexOf: edge).
	interpreterProxy storeInteger: ETXValueIndex ofObject: edgeOop withValue: 
		(self edgeXValueOf: edge).
	interpreterProxy storeInteger: ETYValueIndex ofObject: edgeOop withValue: 
		(self currentYGet).
	interpreterProxy storeInteger: ETZValueIndex ofObject: edgeOop withValue: 
		(self edgeZValueOf: edge).
	interpreterProxy storeInteger: ETLinesIndex ofObject: edgeOop withValue: 
		(self edgeNumLinesOf: edge).
	self lastExportedEdgePut: edge.! !

!BalloonEngineBase methodsFor: 'storing state' stamp: 'ar 11/25/1998 00:36'!
storeEngineStateInto: oop
	self objUsedPut: objUsed.! !

!BalloonEngineBase methodsFor: 'storing state' stamp: 'ar (auto pragmas 12/08) 11/11/1998 22:24'!
storeFillStateInto: fillOop
	| fillIndex leftX rightX |
	<inline: false>
	fillIndex := self lastExportedFillGet.
	leftX := self lastExportedLeftXGet.
	rightX := self lastExportedRightXGet.

	(interpreterProxy slotSizeOf: fillOop) < FTBalloonFillDataSize 
		ifTrue:[^interpreterProxy primitiveFail].
	interpreterProxy storeInteger: FTIndexIndex ofObject: fillOop withValue: 
		(self objectIndexOf: fillIndex).
	interpreterProxy storeInteger: FTMinXIndex ofObject: fillOop withValue: leftX.
	interpreterProxy storeInteger: FTMaxXIndex ofObject: fillOop withValue: rightX.
	interpreterProxy storeInteger: FTYValueIndex ofObject: fillOop withValue: self currentYGet.! !

!BalloonEngineBase methodsFor: 'primitives-rendering' stamp: 'ar (auto pragmas 12/08) 10/31/1998 23:54'!
storeRenderingState
	<inline: false>
	interpreterProxy failed ifTrue:[^nil].
	engineStopped ifTrue:[
		"Check the stop reason and store the required information"
		self storeStopStateIntoEdge: (interpreterProxy stackObjectValue: 1) 
			fill: (interpreterProxy stackObjectValue: 0).
	].
	self storeEngineStateInto: engine.
	interpreterProxy pop: 3.
	interpreterProxy pushInteger: self stopReasonGet.! !

!BalloonEngineBase methodsFor: 'storing state' stamp: 'ar 11/9/1998 15:34'!
storeStopStateIntoEdge: edgeOop fill: fillOop
	| reason edge |
	reason := self stopReasonGet.

	reason = GErrorGETEntry ifTrue:[
		edge := getBuffer at: self getStartGet.
		self storeEdgeStateFrom: edge into: edgeOop.
		self getStartPut: self getStartGet + 1.
	].

	reason = GErrorFillEntry ifTrue:[
		self storeFillStateInto: fillOop.
	].

	reason = GErrorAETEntry ifTrue:[
		edge := aetBuffer at: self aetStartGet.
		self storeEdgeStateFrom: edge into: edgeOop.
		"Do not advance to the next aet entry yet"
		"self aetStartPut: self aetStartGet + 1."
	].! !

!BalloonEngineBase methodsFor: 'FILL processing' stamp: 'ar (auto pragmas 12/08) 11/25/1998 14:38'!
toggleFill: fillIndex depth: depth rightX: rightX
	"Make the fill style with the given index either visible or invisible"
	| hidden |
	<inline: false>

	self stackFillSize = 0 ifTrue:[
		(self allocateStackFillEntry) ifTrue:[
			self topFillValuePut: fillIndex.
			self topFillDepthPut: depth.
			self topFillRightXPut: rightX.
		].
	] ifFalse:[
		hidden := self hideFill: fillIndex depth: depth.
		hidden ifFalse:[self showFill: fillIndex depth: depth rightX: rightX].
	].! !

!BalloonEngineBase methodsFor: 'FILL processing' stamp: 'ar (auto pragmas 12/08) 11/25/1998 15:19'!
toggleFillsOf: edge
	| depth fillIndex |
	<inline: false>

	(self needAvailableSpace: self stackFillEntryLength * 2) 
		ifFalse:[^nil]. "Make sure we have enough space left"
	depth := (self edgeZValueOf: edge) << 1.
	fillIndex := self edgeLeftFillOf: edge.
	fillIndex = 0 ifFalse:[self toggleFill: fillIndex depth: depth rightX: 999999999].
	fillIndex := self edgeRightFillOf: edge.
	fillIndex = 0 ifFalse:[self toggleFill: fillIndex depth: depth rightX: 999999999].
	self quickRemoveInvalidFillsAt: (self edgeXValueOf: edge).! !

!BalloonEngineBase methodsFor: 'FILL processing' stamp: 'ar (auto pragmas 12/08) 11/25/1998 15:50'!
toggleWideFillOf: edge
	| fill type lineWidth depth rightX index |
	<inline: false>
	type := self edgeTypeOf: edge.
	dispatchedValue := edge.
	self dispatchOn: type in: WideLineWidthTable.
	lineWidth := dispatchReturnValue.
	self dispatchOn: type in: WideLineFillTable.
	fill := dispatchReturnValue.
	fill = 0 ifTrue:[^nil].
	(self needAvailableSpace: self stackFillEntryLength) 
		ifFalse:[^nil]. "Make sure we have enough space left"
	depth := (self edgeZValueOf: edge) << 1 + 1. "So lines sort before interior fills"
	rightX := (self edgeXValueOf: edge) + lineWidth.
	index := self findStackFill: fill depth: depth.
	index = -1 ifTrue:[
		self showFill: fill 
			depth: depth
			rightX: rightX.
	] ifFalse:[
		(self stackFillRightX: index) < rightX
			ifTrue:[self stackFillRightX: index put: rightX].
	].
	self quickRemoveInvalidFillsAt: (self edgeXValueOf: edge).! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 11/24/1998 22:49'!
topDepth
	self stackFillSize = 0
		ifTrue:[^-1]
		ifFalse:[^self topFillDepth].! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 10/30/1998 19:28'!
topFill
	self stackFillSize = 0
		ifTrue:[^0]
		ifFalse:[^self topFillValue].! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 10/30/1998 19:27'!
topFillDepth
	^self stackFillDepth: self stackFillSize - self stackFillEntryLength! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 10/30/1998 19:27'!
topFillDepthPut: value
	^self stackFillDepth: self stackFillSize - self stackFillEntryLength put: value! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 11/25/1998 14:36'!
topFillRightX
	^self stackFillRightX: self stackFillSize - self stackFillEntryLength! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 11/25/1998 14:36'!
topFillRightXPut: value
	^self stackFillRightX: self stackFillSize - self stackFillEntryLength put: value! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 10/30/1998 19:27'!
topFillValue
	^self stackFillValue: self stackFillSize - self stackFillEntryLength! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 10/30/1998 19:27'!
topFillValuePut: value
	^self stackFillValue: self stackFillSize - self stackFillEntryLength put: value! !

!BalloonEngineBase methodsFor: 'accessing fills' stamp: 'ar 11/25/1998 15:19'!
topRightX
	self stackFillSize = 0
		ifTrue:[^999999999]
		ifFalse:[^self topFillRightX].! !

!BalloonEngineBase methodsFor: 'transforming' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:42'!
transformColor: fillIndex
	| r g b a transform alphaScale |
	<var: #transform type:'float *'>
	<var: #alphaScale type:'double '>
	(fillIndex = 0 or:[self isFillColor: fillIndex]) ifFalse:[^fillIndex].
	b := fillIndex bitAnd: 255.
	g := (fillIndex >> 8) bitAnd: 255.
	r := (fillIndex >> 16) bitAnd: 255.
	a := (fillIndex >> 24) bitAnd: 255.
	(self hasColorTransform) ifTrue:[
		transform := self colorTransform.
		alphaScale := (a * (transform at: 6) + (transform at: 7)) / a.
		r := (r * (transform at: 0) + (transform at: 1) * alphaScale) asInteger.
		g := (g * (transform at: 2) + (transform at: 3) * alphaScale) asInteger.
		b := (b * (transform at: 4) + (transform at: 5) * alphaScale) asInteger.
		a := a * alphaScale.
		r := r max: 0. r := r min: 255.
		g := g max: 0. g := g min: 255.
		b := b max: 0. b := b min: 255.
		a := a max: 0. a := a min: 255.
	].
	a < 1 ifTrue:[^0]."ALWAYS return zero for transparent fills"
	"If alpha is not 255 (or close thereto) then we need to flush the engine before proceeding"
	(a < 255 and:[self needsFlush]) 
		ifTrue:[self stopBecauseOf: GErrorNeedFlush].
	^b + (g << 8) + (r << 16) + (a << 24)! !

!BalloonEngineBase methodsFor: 'transforming' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:43'!
transformPoint: point
	"Transform the given point. If haveMatrix is true then use the current transformation."
	<var:#point type:'int *'>
	self hasEdgeTransform ifFalse:[
		"Multiply each component by aaLevel and add a half pixel"
		point at: 0 put: (point at: 0) + self destOffsetXGet * self aaLevelGet.
		point at: 1 put: (point at: 1) + self destOffsetYGet * self aaLevelGet.
	] ifTrue:[
		"Note: AA adjustment is done in #transformPoint: for higher accuracy"
		self transformPoint: point into: point.
	].! !

!BalloonEngineBase methodsFor: 'transforming' stamp: 'ar (auto pragmas 12/08) 11/1/1998 16:59'!
transformPoint: srcPoint into: dstPoint
	"Transform srcPoint into dstPoint by using the currently loaded matrix"
	"Note: This method has been rewritten so that inlining works (e.g., removing
	the declarations and adding argument coercions at the appropriate points)"
	<inline: true>
	self transformPointX: ((self cCoerce: srcPoint to: 'int *') at: 0) asFloat 
		y: ((self cCoerce: srcPoint to:'int *') at: 1) asFloat
		into: (self cCoerce: dstPoint to: 'int *')! !

!BalloonEngineBase methodsFor: 'transforming' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:23'!
transformPointX: xValue y: yValue into: dstPoint
	"Transform srcPoint into dstPoint by using the currently loaded matrix"
	"Note: This should be rewritten so that inlining works (e.g., removing
	the declarations and adding argument coercions at the appropriate points)"
	| x y transform |
	<inline: true>
	<var: #dstPoint type:'int *'>
	<var: #xValue type: 'double '>
	<var: #yValue type: 'double '>
	<var: #transform type:'float *'>
	transform := self edgeTransform.
	x := ((((transform at: 0) * xValue) +
		((transform at: 1) * yValue) +
		(transform at: 2)) * self aaLevelGet asFloat) asInteger.
	y := ((((transform at: 3) * xValue) +
		((transform at: 4) * yValue) +
		(transform at: 5)) * self aaLevelGet asFloat) asInteger.
	dstPoint at: 0 put: x.
	dstPoint at: 1 put: y.! !

!BalloonEngineBase methodsFor: 'transforming' stamp: 'ar (auto pragmas 12/08) 11/24/1998 19:48'!
transformPoints: n
	"Transform n (n=1,2,3) points.
	If haveMatrix is true then the matrix contains the actual transformation."
	<inline: true>
	n > 0 ifTrue:[self transformPoint: self point1Get].
	n > 1 ifTrue:[self transformPoint: self point2Get].
	n > 2 ifTrue:[self transformPoint: self point3Get].
	n > 3 ifTrue:[self transformPoint: self point4Get].! !

!BalloonEngineBase methodsFor: 'transforming' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:43'!
transformWidth: w
	"Transform the given width"
	| deltaX deltaY dstWidth dstWidth2 |
	<inline: false>
	<var: #deltaX type:'double '>
	<var: #deltaY type:'double '>
	w = 0 ifTrue:[^0].
	self point1Get at: 0 put: 0.
	self point1Get at: 1 put: 0.
	self point2Get at: 0 put: w * 256.
	self point2Get at: 1 put: 0.
	self point3Get at: 0 put: 0.
	self point3Get at: 1 put: w * 256.
	self transformPoints: 3.
	deltaX := ((self point2Get at: 0) - (self point1Get at: 0)) asFloat.
	deltaY := ((self point2Get at: 1) - (self point1Get at: 1)) asFloat.
	dstWidth := (((deltaX * deltaX) + (deltaY * deltaY)) sqrt asInteger + 128) // 256.
	deltaX := ((self point3Get at: 0) - (self point1Get at: 0)) asFloat.
	deltaY := ((self point3Get at: 1) - (self point1Get at: 1)) asFloat.
	dstWidth2 := (((deltaX * deltaX) + (deltaY * deltaY)) sqrt asInteger + 128) // 256.
	dstWidth2 < dstWidth ifTrue:[dstWidth := dstWidth2].
	dstWidth = 0
		ifTrue:[^1]
		ifFalse:[^dstWidth]! !

!BalloonEngineBase methodsFor: 'transforming' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:43'!
uncheckedTransformColor: fillIndex
	| r g b a transform |
	<var: #transform type:'float *'>
	(self hasColorTransform) ifFalse:[^fillIndex].
	b := fillIndex bitAnd: 255.
	g := (fillIndex >> 8) bitAnd: 255.
	r := (fillIndex >> 16) bitAnd: 255.
	a := (fillIndex >> 24) bitAnd: 255.
	transform := self colorTransform.
	r := (r * (transform at: 0) + (transform at: 1)) asInteger.
	g := (g * (transform at: 2) + (transform at: 3)) asInteger.
	b := (b * (transform at: 4) + (transform at: 5)) asInteger.
	a := (a * (transform at: 6) + (transform at: 7)) asInteger.
	r := r max: 0. r := r min: 255.
	g := g max: 0. g := g min: 255.
	b := b max: 0. b := b min: 255.
	a := a max: 0. a := a min: 255.
	a < 16 ifTrue:[^0]."ALWAYS return zero for transparent fills"
	^b + (g << 8) + (r << 16) + (a << 24)! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:28'!
wbSizeGet
	^workBuffer at: GWSize! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:37'!
wbSizePut: value
	^workBuffer at: GWSize put: value! !

!BalloonEngineBase methodsFor: 'accessing stack' stamp: 'ar 10/31/1998 00:43'!
wbStackClear
	self wbTopPut: self wbSizeGet.! !

!BalloonEngineBase methodsFor: 'accessing stack' stamp: 'ar 11/9/1998 15:34'!
wbStackPop: nItems

	self wbTopPut: self wbTopGet + nItems.! !

!BalloonEngineBase methodsFor: 'accessing stack' stamp: 'ar 10/30/1998 17:16'!
wbStackPush: nItems
	(self allocateStackEntry: nItems) ifFalse:[^false].
	self wbTopPut: self wbTopGet - nItems.
	^true! !

!BalloonEngineBase methodsFor: 'accessing stack' stamp: 'ar 10/30/1998 17:17'!
wbStackSize
	^self wbSizeGet - self wbTopGet! !

!BalloonEngineBase methodsFor: 'accessing stack' stamp: 'ar 11/9/1998 15:34'!
wbStackValue: index

	^workBuffer at: self wbTopGet + index! !

!BalloonEngineBase methodsFor: 'accessing stack' stamp: 'ar 11/9/1998 15:34'!
wbStackValue: index put: value

	^workBuffer at: self wbTopGet + index put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 23:29'!
wbTopGet
	^workBuffer at: GWBufferTop! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 10/27/1998 18:32'!
wbTopPut: value
	^workBuffer at: GWBufferTop put: value! !

!BalloonEngineBase methodsFor: 'accessing state' stamp: 'ar 7/11/2004 13:43'!
workBufferPut: wbOop
	workBuffer := interpreterProxy firstIndexableField: wbOop.! !

BalloonEngineBase subclass: #BalloonEnginePlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!BalloonEnginePlugin commentStamp: 'tpr 5/5/2003 11:46' prior: 0!
This class adds the plugin functionality for the Balloon graphics engine.
BalloonEnginePlugin should be translated but its superclass should not since it is incorporated within this class's translation process. Nor should the simulation subclass be translated!

!BalloonEnginePlugin class methodsFor: 'class initialization' stamp: 'ar 11/11/1998 22:01'!
declareCVarsIn: cg
	"Nothing to declare"! !

!BalloonEnginePlugin methodsFor: 'private' stamp: 'nice 11/18/2016 21:57'!
absoluteSquared8Dot24: value
	"Compute the squared value of a 8.24 number with 0.0 <= value < 1.0,
	e.g., compute (value * value) bitShift: -24"
	| word1 word2 |
	<inline: true>
	<var: 'word1' type: #'unsigned int'>
	<var: 'word2' type: #'unsigned int'>
	word1 := value bitAnd: 16rFFFF.
	word2 := (value bitShift: -16) bitAnd: 255.
	^(( (word1 * word1) bitShift: -16) +
		((word1 * word2) * 2) +
			((word2 * word2) bitShift: 16)) bitShift: -8! !

!BalloonEnginePlugin methodsFor: 'beziers-wide' stamp: 'ar (auto pragmas 12/08) 11/8/1998 15:18'!
adjustWideBezierLeft: bezier width: lineWidth offset: lineOffset endX: endX
	"Adjust the wide bezier curve (dx < 0) to start/end at the right point"
	| lastX lastY |
	<inline: false>
	(self bezierUpdateDataOf: bezier) at: GBUpdateX put: 
		(((self bezierUpdateDataOf: bezier) at: GBUpdateX) - (lineOffset * 256)).
	"Set the lastX/Y value of the second curve lineWidth pixels right/down"
	lastX := (self wideBezierUpdateDataOf: bezier) at: GBUpdateX.
	(self wideBezierUpdateDataOf: bezier) at: GBUpdateX put: lastX + (lineWidth - lineOffset * 256).
	"Set lineWidth pixels down"
	lastY := (self wideBezierUpdateDataOf: bezier) at: GBUpdateY.
	(self wideBezierUpdateDataOf: bezier) at: GBUpdateY put: lastY + (lineWidth * 256).
	"Record the last X value"
	self bezierFinalXOf: bezier put: endX - lineOffset.
! !

!BalloonEnginePlugin methodsFor: 'beziers-wide' stamp: 'ar (auto pragmas 12/08) 11/8/1998 15:18'!
adjustWideBezierRight: bezier width: lineWidth offset: lineOffset endX: endX
	"Adjust the wide bezier curve (dx >= 0) to start/end at the right point"
	| lastX lastY |
	<inline: false>
	(self bezierUpdateDataOf: bezier) at: GBUpdateX put: 
		(((self bezierUpdateDataOf: bezier) at: GBUpdateX) + (lineOffset * 256)).
	"Set the lastX/Y value of the second curve lineWidth pixels right/down"
	"Set lineWidth-lineOffset pixels left"
	lastX := (self wideBezierUpdateDataOf: bezier) at: GBUpdateX.
	(self wideBezierUpdateDataOf: bezier) at: GBUpdateX put: lastX - (lineWidth - lineOffset * 256).
	lastY := (self wideBezierUpdateDataOf: bezier) at: GBUpdateY.
	"Set lineWidth pixels down"
	(self wideBezierUpdateDataOf: bezier) at: GBUpdateY put: lastY + (lineWidth * 256).
	"Record the last X value"
	self bezierFinalXOf: bezier put: endX - lineOffset + lineWidth.! !

!BalloonEnginePlugin methodsFor: 'lines-wide' stamp: 'ar (auto pragmas 12/08) 11/9/1998 15:34'!
adjustWideLine: line afterSteppingFrom: lastX to: nextX
	"Adjust the wide line after it has been stepped from lastX to nextX.
	Special adjustments of line width and start position are made here
	to simulate a rectangular brush"
	|  yEntry yExit lineWidth lineOffset deltaX xDir baseWidth |
	<inline: false>	"Don't inline this"


	"Fetch the values the adjustment decisions are based on"
	yEntry := (self wideLineEntryOf: line).
	yExit := (self wideLineExitOf: line).
	baseWidth := self wideLineExtentOf: line.
	lineOffset := self offsetFromWidth: baseWidth.
	lineWidth := self wideLineWidthOf: line.
	xDir := self lineXDirectionOf: line.
	deltaX := nextX - lastX.

	"Adjust the start of the line to fill an entire rectangle"
	yEntry < baseWidth ifTrue:[
		xDir < 0
			ifTrue:[	lineWidth := lineWidth - deltaX] "effectively adding"
			ifFalse:[	lineWidth := lineWidth + deltaX.
					self edgeXValueOf: line put: lastX].
	].

	"Adjust the end of x-major lines"
	((yExit + lineOffset) = 0) ifTrue:[
		xDir > 0
			ifTrue:[lineWidth := lineWidth - (self lineXIncrementOf: line)]
			ifFalse:[lineWidth := lineWidth + (self lineXIncrementOf: line).	"effectively subtracting"
					self edgeXValueOf: line put: lastX].
	].

	"Adjust the end of the line to fill an entire rectangle"
	(yExit + lineOffset) > 0 ifTrue:[
		xDir < 0
			ifTrue:[	lineWidth := lineWidth + deltaX. "effectively subtracting"
					self edgeXValueOf: line put: lastX]
			ifFalse:[	lineWidth := lineWidth - deltaX]
	].

	"Store the manipulated line width back"
	self wideLineWidthOf: line put: lineWidth.! !

!BalloonEnginePlugin methodsFor: 'allocation' stamp: 'ar 11/25/1998 00:38'!
allocateBezier
	| bezier |
	(self allocateObjEntry: GBBaseSize) ifFalse:[^0].
	bezier := objUsed.
	objUsed := bezier + GBBaseSize.
	self objectTypeOf: bezier put: GEPrimitiveBezier.
	self objectIndexOf: bezier put: 0.
	self objectLengthOf: bezier put: GBBaseSize.
	^bezier! !

!BalloonEnginePlugin methodsFor: 'allocation' stamp: 'ar 10/30/1998 20:52'!
allocateBezierStackEntry
	self wbStackPush: 6.
	^self wbStackSize! !

!BalloonEnginePlugin methodsFor: 'allocation' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:44'!
allocateBitmapFill: cmSize colormap: cmBits
	| fill fillSize cm |
	<var:#cm type:'int *'>
	<var:#cmBits type:'int *'>
	fillSize := GBMBaseSize + cmSize.
	(self allocateObjEntry: fillSize) ifFalse:[^0].
	fill := objUsed.
	objUsed := fill + fillSize.
	self objectTypeOf: fill put: GEPrimitiveClippedBitmapFill.
	self objectIndexOf: fill put: 0.
	self objectLengthOf: fill put: fillSize.
	cm := self colormapOf: fill.
	self hasColorTransform ifTrue:[
		0 to: cmSize-1 do:[:i| cm at: i put: (self transformColor: (cmBits at: i))].
	] ifFalse:[
		0 to: cmSize-1 do:[:i| cm at: i put: (cmBits at: i)].
	].
	self bitmapCmSizeOf: fill put: cmSize.
	^fill! !

!BalloonEnginePlugin methodsFor: 'allocation' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:44'!
allocateGradientFill: ramp rampWidth: rampWidth isRadial: isRadial
	| fill fillSize rampPtr |
	<var:#ramp type:'int *'>
	<var:#rampPtr type:'int *'>
	fillSize := GGBaseSize + rampWidth.
	(self allocateObjEntry: fillSize) ifFalse:[^0].
	fill := objUsed.
	objUsed := fill + fillSize.
	isRadial
		ifTrue:[self objectTypeOf: fill put: GEPrimitiveRadialGradientFill]
		ifFalse:[self objectTypeOf: fill put: GEPrimitiveLinearGradientFill].
	self objectIndexOf: fill put: 0.
	self objectLengthOf: fill put: fillSize.
	rampPtr := self gradientRampOf: fill.
	self hasColorTransform ifTrue:[
		0 to: rampWidth-1 do:[:i| rampPtr at: i put: (self transformColor: (ramp at: i))].
	] ifFalse:[
		0 to: rampWidth-1 do:[:i| rampPtr at: i put: (ramp at: i)].
	].
	self gradientRampLengthOf: fill put: rampWidth.
	^fill! !

!BalloonEnginePlugin methodsFor: 'allocation' stamp: 'ar 11/25/1998 00:39'!
allocateLine
	| line |
	(self allocateObjEntry: GLBaseSize) ifFalse:[^0].
	line := objUsed.
	objUsed := line + GLBaseSize.
	self objectTypeOf: line put: GEPrimitiveLine.
	self objectIndexOf: line put: 0.
	self objectLengthOf: line put: GLBaseSize.
	^line! !

!BalloonEnginePlugin methodsFor: 'allocation' stamp: 'ar 11/25/1998 00:39'!
allocateWideBezier
	| bezier |
	(self allocateObjEntry: GBWideSize) ifFalse:[^0].
	bezier := objUsed.
	objUsed := bezier + GBWideSize.
	self objectTypeOf: bezier put: GEPrimitiveWideBezier.
	self objectIndexOf: bezier put: 0.
	self objectLengthOf: bezier put: GBWideSize.
	^bezier! !

!BalloonEnginePlugin methodsFor: 'allocation' stamp: 'ar 11/25/1998 00:39'!
allocateWideLine
	| line |
	(self allocateObjEntry: GLWideSize) ifFalse:[^0].
	line := objUsed.
	objUsed := line + GLWideSize.
	self objectTypeOf: line put: GEPrimitiveWideLine.
	self objectIndexOf: line put: 0.
	self objectLengthOf: line put: GLWideSize.
	^line! !

!BalloonEnginePlugin methodsFor: 'bezier-loading' stamp: 'ar (auto pragmas 12/08) 11/9/1998 01:56'!
assureValue: val1 between: val2 and: val3
	"Make sure that val1 is between val2 and val3."
	<inline: true>
	val2 > val3 ifTrue:[
		val1 > val2 ifTrue:[^val2].
		val1 < val3 ifTrue:[^val3].
	] ifFalse:[
		val1 < val2 ifTrue:[^val2].
		val1 > val3 ifTrue:[^val3].
	].
	^val1	! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:20'!
bezierEndXOf: bezier

	^self obj: bezier at: GBEndX! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:17'!
bezierEndXOf: bezier put: value

	^self obj: bezier at: GBEndX put: value! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:17'!
bezierEndYOf: bezier

	^self obj: bezier at: GBEndY! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:18'!
bezierEndYOf: bezier put: value

	^self obj: bezier at: GBEndY put: value! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:19'!
bezierFinalXOf: bezier

	^self obj: bezier at: GBFinalX! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:18'!
bezierFinalXOf: bezier put: value

	^self obj: bezier at: GBFinalX put: value! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar (auto pragmas 12/08) 11/24/1998 22:24'!
bezierUpdateDataOf: bezier
	<returnTypeC: 'int *'>

	^objBuffer + bezier + GBUpdateData! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:20'!
bezierViaXOf: bezier

	^self obj: bezier at: GBViaX! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:17'!
bezierViaXOf: bezier put: value

	^self obj: bezier at: GBViaX put: value! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:17'!
bezierViaYOf: bezier

	^self obj: bezier at: GBViaY! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:18'!
bezierViaYOf: bezier put: value

	^self obj: bezier at: GBViaY put: value! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar 11/24/1998 22:20'!
bitmapCmSizeOf: bmFill

	^self obj: bmFill at: GBColormapSize! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar 11/24/1998 22:19'!
bitmapCmSizeOf: bmFill put: value

	^self obj: bmFill at: GBColormapSize put: value! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar 11/24/1998 22:17'!
bitmapDepthOf: bmFill

	^self obj: bmFill at: GBBitmapDepth! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar 11/24/1998 22:20'!
bitmapDepthOf: bmFill put: value

	^self obj: bmFill at: GBBitmapDepth put: value! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar 11/24/1998 22:17'!
bitmapHeightOf: bmFill

	^self obj: bmFill at: GBBitmapHeight! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar 11/24/1998 22:17'!
bitmapHeightOf: bmFill put: value

	^self obj: bmFill at: GBBitmapHeight put: value! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar 11/24/1998 22:18'!
bitmapRasterOf: bmFill

	^self obj: bmFill at: GBBitmapRaster! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar 11/24/1998 22:18'!
bitmapRasterOf: bmFill put: value

	^self obj: bmFill at: GBBitmapRaster put: value! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar 11/24/1998 22:19'!
bitmapSizeOf: bmFill

	^self obj: bmFill at: GBBitmapSize! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar 11/24/1998 22:18'!
bitmapSizeOf: bmFill put: value

	^self obj: bmFill at: GBBitmapSize put: value! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar 11/27/1998 14:20'!
bitmapTileFlagOf: bmFill

	^self obj: bmFill at: GBTileFlag! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar 11/27/1998 14:20'!
bitmapTileFlagOf: bmFill put: value

	^self obj: bmFill at: GBTileFlag put: value! !

!BalloonEnginePlugin methodsFor: 'fills-bitmaps' stamp: 'eem 6/21/2021 19:13'!
bitmapValue: bmFill bits: bits atX: xp y: yp

	| bmDepth bmRaster value rShift cMask r g b a |
	<inline: true>

	bmDepth := self bitmapDepthOf: bmFill.
	bmRaster := self bitmapRasterOf: bmFill.
	bmDepth = 32 ifTrue: [
		value := (self cCoerce: bits to: 'int *') at: (bmRaster * yp) + xp.
		(value ~= 0 and: [(value bitAnd: 16rFF000000) = 0])
				ifTrue: [value := value bitOr: 16rFF000000].
		^self uncheckedTransformColor: value].
	"rShift - shift value to convert from pixel to word index"
	rShift := self rShiftTable at: bmDepth.
	value := self makeUnsignedFrom: 
		((self cCoerce: bits to: 'int *') at: (bmRaster * yp) + (xp >> rShift)).
	"cMask - mask out the pixel from the word"
	cMask := (1 << bmDepth) - 1.
	"rShift - shift value to move the pixel in the word to the lowest bit position"
	rShift := 32 - bmDepth - ((xp bitAnd: (1 << rShift - 1)) * bmDepth).
	value := (value >> rShift) bitAnd: cMask.
	bmDepth = 16 ifTrue: [
		"Must convert by expanding bits"
		value = 0 ifFalse: [
			b := (value bitAnd: 31) << 3.		b := b + (b >> 5).
			g := (value >> 5 bitAnd: 31) << 3.	g := g + (g >> 5).
			r := (value >> 10 bitAnd: 31) << 3.	r := r + (r >> 5).
			a := 255.
			value := b + (g << 8) + (r << 16) + (a << 24)].
	] ifFalse: [
		"Must convert by using color map"
		(self bitmapCmSizeOf: bmFill) = 0
			ifTrue: [value := 0]
			ifFalse: [value := self makeUnsignedFrom: ((self colormapOf: bmFill) at: value)].
	].
	^self uncheckedTransformColor: value! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar 11/24/1998 22:18'!
bitmapWidthOf: bmFill

	^self obj: bmFill at: GBBitmapWidth! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar 11/24/1998 22:17'!
bitmapWidthOf: bmFill put: value

	^self obj: bmFill at: GBBitmapWidth put: value! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 10/30/1998 20:53'!
bzEndX: index
	^self wbStackValue: self wbStackSize - index + 4! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 10/30/1998 20:53'!
bzEndX: index put: value
	^self wbStackValue: self wbStackSize - index + 4 put: value! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 10/30/1998 20:53'!
bzEndY: index
	^self wbStackValue: self wbStackSize - index + 5! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 10/30/1998 20:53'!
bzEndY: index put: value
	^self wbStackValue: self wbStackSize - index + 5 put: value! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 10/30/1998 20:54'!
bzStartX: index
	^self wbStackValue: self wbStackSize - index + 0! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 10/30/1998 20:54'!
bzStartX: index put: value
	^self wbStackValue: self wbStackSize - index + 0 put: value! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 10/30/1998 20:53'!
bzStartY: index
	^self wbStackValue: self wbStackSize - index + 1! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 10/30/1998 20:54'!
bzStartY: index put: value
	^self wbStackValue: self wbStackSize - index + 1 put: value! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 10/30/1998 20:53'!
bzViaX: index
	^self wbStackValue: self wbStackSize - index + 2! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 10/30/1998 20:54'!
bzViaX: index put: value
	^self wbStackValue: self wbStackSize - index + 2 put: value! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 10/30/1998 20:53'!
bzViaY: index
	^self wbStackValue: self wbStackSize - index + 3! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 10/30/1998 20:54'!
bzViaY: index put: value
	^self wbStackValue: self wbStackSize - index + 3 put: value! !

!BalloonEnginePlugin methodsFor: 'shapes-compressed' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:45'!
checkCompressedFillIndexList: fillList max: maxIndex segments: nSegs
	"Check the fill indexes in the run-length encoded fillList"
	| length runLength runValue nFills fillPtr |
	<inline: false>
	<var: #fillPtr type:'int *'>
	length := interpreterProxy slotSizeOf: fillList.
	fillPtr := interpreterProxy firstIndexableField: fillList.
	nFills := 0.
	0 to: length-1 do:[:i |
		runLength := self shortRunLengthAt: i from: fillPtr.
		runValue := self shortRunValueAt: i from: fillPtr.
		(runValue >= 0 and:[runValue <= maxIndex]) ifFalse:[^false].
		nFills := nFills + runLength.
	].
	^nFills = nSegs! !

!BalloonEnginePlugin methodsFor: 'shapes-compressed' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:45'!
checkCompressedFills: indexList
	"Check if the indexList (containing fill handles) is okay."
	| fillPtr length fillIndex |
	<inline: false>
	<var: #fillPtr type:'int *'>
	"First check if the oops have the right format"
	(interpreterProxy isWords: indexList) ifFalse:[^false].

	"Then check the fill entries"
	length := interpreterProxy slotSizeOf: indexList.
	fillPtr := interpreterProxy firstIndexableField: indexList.
	0 to: length-1 do:[:i |
		fillIndex := fillPtr at: i.
		"Make sure the fill is okay"
		(self isFillOkay: fillIndex) ifFalse:[^false]].

	^ true! !

!BalloonEnginePlugin methodsFor: 'shapes-compressed' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:45'!
checkCompressedLineWidths: lineWidthList segments: nSegments
	"Check the run-length encoded lineWidthList matches nSegments"
	| length runLength nItems ptr |
	<inline: false>
	<var: #ptr type:'int *'>
	length := interpreterProxy slotSizeOf: lineWidthList.
	ptr := interpreterProxy firstIndexableField: lineWidthList.
	nItems := 0.
	0 to: length-1 do:[:i|
		runLength := self shortRunLengthAt: i from: ptr.
		nItems := nItems + runLength.
	].
	^nItems = nSegments! !

!BalloonEnginePlugin methodsFor: 'shapes-compressed' stamp: 'ar (auto pragmas 12/08) 11/8/1998 15:19'!
checkCompressedPoints: points segments: nSegments
	"Check if the given point array can be handled by the engine."
	| pSize |
	<inline: false>
	(interpreterProxy isWords: points) ifFalse:[^false].
	pSize := interpreterProxy slotSizeOf: points.
	"The points must be either in PointArray format or ShortPointArray format.
	Also, we currently handle only quadratic segments (e.g., 3 points each) and thus either
		pSize = nSegments * 3,		for ShortPointArrays or,
		pSize = nSegments * 6,		for PointArrays"
	(pSize = (nSegments * 3) or:[pSize = (nSegments * 6)]) 
		ifFalse:[^false]. "Can't handle this"
	^true! !

!BalloonEnginePlugin methodsFor: 'shapes-compressed' stamp: 'ar (auto pragmas 12/08) 11/12/1998 21:22'!
checkCompressedShape: points segments: nSegments leftFills: leftFills rightFills: rightFills lineWidths: lineWidths lineFills: lineFills fillIndexList: fillIndexList
	"Check if the given shape can be handled by the engine. 
	Since there are a number of requirements this is an extra method."
	| maxFillIndex |
	<inline: false>
	(self checkCompressedPoints: points segments: nSegments) 
		ifFalse:[^false].
	(self checkCompressedFills: fillIndexList)
		ifFalse:[^false].
	maxFillIndex := interpreterProxy slotSizeOf: fillIndexList.
	(self checkCompressedFillIndexList: leftFills max: maxFillIndex segments: nSegments)
		ifFalse:[^false].
	(self checkCompressedFillIndexList: rightFills max: maxFillIndex segments: nSegments)
		ifFalse:[^false].
	(self checkCompressedFillIndexList: lineFills max: maxFillIndex segments: nSegments)
		ifFalse:[^false].
	(self checkCompressedLineWidths: lineWidths segments: nSegments)
		ifFalse:[^false].
	^true! !

!BalloonEnginePlugin methodsFor: 'GET processing' stamp: 'ar (auto pragmas 12/08) 11/9/1998 15:37'!
checkedAddBezierToGET: bezier
	"Add the bezier to the global edge table if it intersects the clipping region"
	| lineWidth |
	<inline: true>

	(self isWide: bezier) 
		ifTrue:[lineWidth := (self wideBezierExtentOf: bezier)]
		ifFalse:[lineWidth := 0].
	(self bezierEndYOf: bezier) + lineWidth < (self fillMinYGet) ifTrue:[^0].
	"Overlaps in Y but may still be entirely right of clip region"
	((self edgeXValueOf: bezier) - lineWidth >= self fillMaxXGet and:[
		(self bezierEndXOf: bezier) - lineWidth >= self fillMaxXGet]) ifTrue:[^0].
	self addEdgeToGET: bezier.
! !

!BalloonEnginePlugin methodsFor: 'GET processing' stamp: 'ar (auto pragmas 12/08) 11/9/1998 15:37'!
checkedAddEdgeToGET: edge
	"Add the edge to the global edge table.
	For known edge types, check if the edge intersects the visible region"
	<inline: true>

	(self isLine: edge) ifTrue:[^self checkedAddLineToGET: edge].
	(self isBezier: edge) ifTrue:[^self checkedAddBezierToGET: edge].
	self addEdgeToGET: edge.
! !

!BalloonEnginePlugin methodsFor: 'GET processing' stamp: 'ar (auto pragmas 12/08) 11/9/1998 15:37'!
checkedAddLineToGET: line
	"Add the line to the global edge table if it intersects the clipping region"
	| lineWidth |
	<inline: true>

	(self isWide: line) 
		ifTrue:[lineWidth := (self wideLineExtentOf: line)]
		ifFalse:[lineWidth := 0].
	(self lineEndYOf: line) + lineWidth < (self fillMinYGet) ifTrue:[^0].
	"Overlaps in Y but may still be entirely right of clip region"
	((self edgeXValueOf: line) - lineWidth >= self fillMaxXGet and:[
		(self lineEndXOf: line) - lineWidth >= self fillMaxXGet]) ifTrue:[^0].
	self addEdgeToGET: line.
! !

!BalloonEnginePlugin methodsFor: 'private' stamp: 'ar (auto pragmas 12/08) 11/1/1998 17:06'!
circleCosTable
	| theTable |
	<returnTypeC:'double *'>
	<inline: false>
	<var:#theTable declareC:'static double theTable[33] =
		{1.0, 0.98078528040323, 0.923879532511287, 0.831469612302545,
		0.7071067811865475, 0.555570233019602, 0.38268343236509, 0.1950903220161286,
		0.0, -0.1950903220161283, -0.3826834323650896, -0.555570233019602,
		-0.707106781186547, -0.831469612302545, -0.9238795325112865, -0.98078528040323,
		-1.0, -0.98078528040323, -0.923879532511287, -0.831469612302545,
		-0.707106781186548, -0.555570233019602, -0.3826834323650903, -0.1950903220161287,
		0.0, 0.1950903220161282, 0.38268343236509, 0.555570233019602,
		0.707106781186547, 0.831469612302545, 0.9238795325112865, 0.98078528040323,
		1.0 }'>
	^theTable! !

!BalloonEnginePlugin methodsFor: 'private' stamp: 'ar (auto pragmas 12/08) 11/1/1998 17:06'!
circleSinTable
	| theTable |
	<returnTypeC:'double *'>
	<inline: false>
	<var:#theTable declareC:'static double theTable[33] =
		{0.0, 0.1950903220161282, 0.3826834323650897, 0.555570233019602,
		0.707106781186547, 0.831469612302545, 0.923879532511287, 0.98078528040323,
		1.0, 0.98078528040323, 0.923879532511287, 0.831469612302545,
		0.7071067811865475, 0.555570233019602, 0.38268343236509, 0.1950903220161286,
		0.0, -0.1950903220161283, -0.3826834323650896, -0.555570233019602,
		-0.707106781186547, -0.831469612302545, -0.9238795325112865, -0.98078528040323,
		-1.0, -0.98078528040323, -0.923879532511287, -0.831469612302545,
		-0.707106781186548, -0.555570233019602, -0.3826834323650903, -0.1950903220161287,
		 0.0 }'>
	^theTable! !

!BalloonEnginePlugin methodsFor: 'fills-bitmaps' stamp: 'ar (auto pragmas 12/08) 11/27/1998 14:19'!
clampValue: value max: maxValue

	<inline: true>
	value < 0 
		ifTrue:[^0]
		ifFalse:[value >= maxValue 
					ifTrue:[^maxValue-1]
					ifFalse:[^value]]! !

!BalloonEnginePlugin methodsFor: 'accessing bitmaps' stamp: 'ar (auto pragmas 12/08) 11/25/1998 16:39'!
colormapOf: bmFill
	<returnTypeC:'int *'>
	^objBuffer + bmFill + GBColormapOffset! !

!BalloonEnginePlugin methodsFor: 'bezier-loading' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:46'!
computeBezier: index splitAt: param
	"Split the bezier curve at the given parametric value.
	Note: Since this method is only invoked to make non-monoton
		beziers monoton we must check for the resulting y values
		to be *really* between the start and end value."
	| startX startY viaX viaY endX endY newIndex 
	leftViaX  leftViaY rightViaX rightViaY sharedX sharedY |
	<inline: false>
	<var: #param type:'double '>
	leftViaX := startX := self bzStartX: index.
	leftViaY := startY := self bzStartY: index.
	rightViaX := viaX := self bzViaX: index.
	rightViaY := viaY := self bzViaY: index.
	endX := self bzEndX: index.
	endY := self bzEndY: index.

	"Compute intermediate points"
	sharedX := leftViaX := leftViaX + ((viaX - startX) asFloat * param) asInteger.
	sharedY := leftViaY := leftViaY + ((viaY - startY) asFloat * param) asInteger.
	rightViaX := rightViaX + ((endX - viaX) asFloat * param) asInteger.
	rightViaY := rightViaY + ((endY - viaY) asFloat * param) asInteger.
	"Compute new shared point"
	sharedX := sharedX + ((rightViaX - leftViaX) asFloat * param) asInteger.
	sharedY := sharedY + ((rightViaY - leftViaY) asFloat * param) asInteger.

	"Check the new via points"
	leftViaY := self assureValue: leftViaY between: startY and: sharedY.
	rightViaY := self assureValue: rightViaY between: sharedY and: endY.

	newIndex := self allocateBezierStackEntry.
	engineStopped ifTrue:[^0]. "Something went wrong"

	"Store the first part back"
	self bzViaX: index put: leftViaX.
	self bzViaY: index put: leftViaY.
	self bzEndX: index put: sharedX.
	self bzEndY: index put: sharedY.
	"Store the second point back"
	self bzStartX: newIndex put: sharedX.
	self bzStartY: newIndex put: sharedY.
	self bzViaX: newIndex put: rightViaX.
	self bzViaY: newIndex put: rightViaY.
	self bzEndX: newIndex put: endX.
	self bzEndY: newIndex put: endY.

	^newIndex! !

!BalloonEnginePlugin methodsFor: 'bezier-loading' stamp: 'ar (auto pragmas 12/08) 11/6/1998 01:26'!
computeBezierSplitAtHalf: index
	"Split the bezier curve at 0.5."
	| startX startY viaX viaY endX endY newIndex 
	leftViaX  leftViaY rightViaX rightViaY sharedX sharedY |
	<inline: false>

	newIndex := self allocateBezierStackEntry.
	engineStopped ifTrue:[^0]. "Something went wrong"

	leftViaX := startX := self bzStartX: index.
	leftViaY := startY := self bzStartY: index.
	rightViaX := viaX := self bzViaX: index.
	rightViaY := viaY := self bzViaY: index.
	endX := self bzEndX: index.
	endY := self bzEndY: index.
	"Compute intermediate points"
	leftViaX := leftViaX + ((viaX - startX) // 2).
	leftViaY := leftViaY + ((viaY - startY) // 2).
	sharedX := rightViaX := rightViaX + ((endX - viaX) // 2).
	sharedY := rightViaY := rightViaY + ((endY - viaY) // 2).
	"Compute new shared point"
	sharedX := sharedX + ((leftViaX - rightViaX) // 2).
	sharedY := sharedY + ((leftViaY - rightViaY) // 2).
	"Store the first part back"
	self bzViaX: index put: leftViaX.
	self bzViaY: index put: leftViaY.
	self bzEndX: index put: sharedX.
	self bzEndY: index put: sharedY.
	"Store the second point back"
	self bzStartX: newIndex put: sharedX.
	self bzStartY: newIndex put: sharedY.
	self bzViaX: newIndex put: rightViaX.
	self bzViaY: newIndex put: rightViaY.
	self bzEndX: newIndex put: endX.
	self bzEndY: newIndex put: endY.

	^newIndex! !

!BalloonEnginePlugin methodsFor: 'beziers-wide' stamp: 'ar 11/8/1998 03:44'!
computeFinalWideBezierValues: bezier width: lineWidth
	"Get both values from the two boundaries of the given bezier 
	and compute the actual position/width of the line"
	| leftX rightX temp |
	leftX := ((self bezierUpdateDataOf: bezier) at: GBUpdateX) // 256.
	rightX := ((self wideBezierUpdateDataOf: bezier) at: GBUpdateX) // 256.
	leftX > rightX ifTrue:[temp := leftX. leftX := rightX. rightX := temp].
	self edgeXValueOf: bezier put: leftX.
	(rightX - leftX) > lineWidth ifTrue:[
		self wideBezierWidthOf: bezier put: (rightX - leftX).
	] ifFalse:[
		self wideBezierWidthOf: bezier put: lineWidth.
	].! !

!BalloonEnginePlugin methodsFor: 'fills-bitmaps' stamp: 'ar (auto pragmas 12/08) 11/25/1998 19:46'!
fillBitmapSpan
	<inline: true>
	^self fillBitmapSpan: self lastExportedFillGet from: self lastExportedLeftXGet to: self lastExportedRightXGet at: self currentYGet! !

!BalloonEnginePlugin methodsFor: 'fills-bitmaps' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:46'!
fillBitmapSpan: bmFill from: leftX to: rightX at: yValue
	| x x1 dsX ds dtX dt deltaX deltaY bits xp yp bmWidth bmHeight fillValue tileFlag |
	<inline: false>
	<var: #bits type:'int *'>
	self aaLevelGet = 1
		ifFalse:[^self fillBitmapSpanAA: bmFill from: leftX to: rightX at: yValue].

	bits := self loadBitsFrom: bmFill.
	bits == nil ifTrue:[^nil].
	bmWidth := self bitmapWidthOf: bmFill.
	bmHeight := self bitmapHeightOf: bmFill.
	tileFlag := (self bitmapTileFlagOf: bmFill) = 1.
	deltaX := leftX - (self fillOriginXOf: bmFill).
	deltaY := yValue - (self fillOriginYOf: bmFill).
	dsX := self fillDirectionXOf: bmFill.
	dtX := self fillNormalXOf: bmFill.

	ds := (deltaX * dsX) + (deltaY * (self fillDirectionYOf: bmFill)).
	dt := (deltaX * dtX) + (deltaY * (self fillNormalYOf: bmFill)).

	x := leftX.
	x1 := rightX.
	[x < x1] whileTrue:[
		tileFlag ifTrue:[
			ds := self repeatValue: ds max: bmWidth << 16.
			dt := self repeatValue: dt max: bmHeight << 16].
		xp := ds // 16r10000.
		yp := dt // 16r10000.
		tileFlag ifFalse:[
			xp := self clampValue: xp max: bmWidth.
			yp := self clampValue: yp max: bmHeight].
		(xp >= 0 and:[yp >= 0 and:[xp < bmWidth and:[yp < bmHeight]]]) ifTrue:[
			fillValue := self bitmapValue: bmFill bits: bits atX: xp y: yp.
			spanBuffer at: x put: fillValue.
		].
		ds := ds + dsX.
		dt := dt + dtX.
		x := x + 1.
	].! !

!BalloonEnginePlugin methodsFor: 'fills-bitmaps' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:46'!
fillBitmapSpanAA: bmFill from: leftX to: rightX at: yValue
	| x dsX ds dtX dt deltaX deltaY bits xp yp bmWidth bmHeight fillValue baseShift cMask cShift idx aaLevel firstPixel lastPixel tileFlag |
	<inline: false>
	<var: #bits type:'int *'>
	bits := self loadBitsFrom: bmFill.
	bits == nil ifTrue:[^nil].
	bmWidth := self bitmapWidthOf: bmFill.
	bmHeight := self bitmapHeightOf: bmFill.
	tileFlag := (self bitmapTileFlagOf: bmFill) = 1.
	deltaX := leftX - (self fillOriginXOf: bmFill).
	deltaY := yValue - (self fillOriginYOf: bmFill).
	dsX := self fillDirectionXOf: bmFill.
	dtX := self fillNormalXOf: bmFill.

	ds := (deltaX * dsX) + (deltaY * (self fillDirectionYOf: bmFill)).
	dt := (deltaX * dtX) + (deltaY * (self fillNormalYOf: bmFill)).

	aaLevel := self aaLevelGet.
	firstPixel := self aaFirstPixelFrom: leftX to: rightX.
	lastPixel := self aaLastPixelFrom: leftX to: rightX.
	baseShift := self aaShiftGet.
	cMask := self aaColorMaskGet.
	cShift := self aaColorShiftGet.
	x := leftX.
	[x < firstPixel] whileTrue:[
		tileFlag ifTrue:[
			ds := self repeatValue: ds max: bmWidth << 16.
			dt := self repeatValue: dt max: bmHeight << 16].
		xp := ds // 16r10000.
		yp := dt // 16r10000.
		tileFlag ifFalse:[
			xp := self clampValue: xp max: bmWidth.
			yp := self clampValue: yp max: bmHeight].
		(xp >= 0 and:[yp >= 0 and:[xp < bmWidth and:[yp < bmHeight]]]) ifTrue:[
			fillValue := self bitmapValue: bmFill bits: bits atX: xp y: yp.
			fillValue := (fillValue bitAnd: cMask) >> cShift.
			idx := x >> baseShift.
			spanBuffer at: idx put: (spanBuffer at: idx) + fillValue.
		].
		ds := ds + dsX.
		dt := dt + dtX.
		x := x + 1.
	].

	cMask := (self aaColorMaskGet >> self aaShiftGet) bitOr: 16rF0F0F0F0.
	cShift := self aaShiftGet.
	[x < lastPixel] whileTrue:[
		tileFlag ifTrue:[
			ds := self repeatValue: ds max: bmWidth << 16.
			dt := self repeatValue: dt max: bmHeight << 16].
		xp := ds // 16r10000.
		yp := dt // 16r10000.
		tileFlag ifFalse:[
			xp := self clampValue: xp max: bmWidth.
			yp := self clampValue: yp max: bmHeight].
		(xp >= 0 and:[yp >= 0 and:[xp < bmWidth and:[yp < bmHeight]]]) ifTrue:[
			fillValue := self bitmapValue: bmFill bits: bits atX: xp y: yp.
			fillValue := (fillValue bitAnd: cMask) >> cShift.
			idx := x >> baseShift.
			spanBuffer at: idx put: (spanBuffer at: idx) + fillValue.
		].
		ds := ds + (dsX << cShift).
		dt := dt + (dtX << cShift).
		x := x + aaLevel.
	].

	cMask := self aaColorMaskGet.
	cShift := self aaColorShiftGet.
	[x < rightX] whileTrue:[
		tileFlag ifTrue:[
			ds := self repeatValue: ds max: bmWidth << 16.
			dt := self repeatValue: dt max: bmHeight << 16].
		xp := ds // 16r10000.
		yp := dt // 16r10000.
		tileFlag ifFalse:[
			xp := self clampValue: xp max: bmWidth.
			yp := self clampValue: yp max: bmHeight].
		(xp >= 0 and:[yp >= 0 and:[xp < bmWidth and:[yp < bmHeight]]]) ifTrue:[
			fillValue := self bitmapValue: bmFill bits: bits atX: xp y: yp.
			fillValue := (fillValue bitAnd: cMask) >> cShift.
			idx := x >> baseShift.
			spanBuffer at: idx put: (spanBuffer at: idx) + fillValue.
		].
		ds := ds + dsX.
		dt := dt + dtX.
		x := x + 1.
	].
! !

!BalloonEnginePlugin methodsFor: 'accessing fills' stamp: 'ar 11/24/1998 22:18'!
fillDirectionXOf: fill

	^self obj: fill at: GFDirectionX! !

!BalloonEnginePlugin methodsFor: 'accessing fills' stamp: 'ar 11/24/1998 22:18'!
fillDirectionXOf: fill put: value

	^self obj: fill at: GFDirectionX put: value! !

!BalloonEnginePlugin methodsFor: 'accessing fills' stamp: 'ar 11/24/1998 22:19'!
fillDirectionYOf: fill

	^self obj: fill at: GFDirectionY! !

!BalloonEnginePlugin methodsFor: 'accessing fills' stamp: 'ar 11/24/1998 22:18'!
fillDirectionYOf: fill put: value

	^self obj: fill at: GFDirectionY put: value! !

!BalloonEnginePlugin methodsFor: 'fills-gradient' stamp: 'ar (auto pragmas 12/08) 11/8/1998 15:20'!
fillLinearGradient
	<inline: true>
	^self fillLinearGradient: self lastExportedFillGet from: self lastExportedLeftXGet to: self lastExportedRightXGet at: self currentYGet! !

!BalloonEnginePlugin methodsFor: 'fills-gradient' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:46'!
fillLinearGradient: fill from: leftX to: rightX at: yValue
	"Draw a linear gradient fill."
	| x0 x1 ramp rampSize dsX ds x rampIndex |
	<inline: false>
	<var: #ramp type:'int *'>
	ramp := self gradientRampOf: fill.
	rampSize := self gradientRampLengthOf: fill.

	dsX := self fillDirectionXOf: fill.
	ds := ((leftX - (self fillOriginXOf: fill)) * dsX) + 
			((yValue - (self fillOriginYOf: fill)) * (self fillDirectionYOf: fill)).

	x := x0 := leftX.
	x1 := rightX.

	"Note: The inner loop has been divided into three parts for speed"
	"Part one: Fill everything outside the left boundary"
	[((rampIndex := ds // 16r10000) < 0 or:[rampIndex >= rampSize]) and:[x < x1]] 
		whileTrue:[	x := x + 1.
					ds := ds + dsX].
	x > x0 ifTrue:[
		rampIndex < 0 ifTrue:[rampIndex := 0].
		rampIndex >= rampSize ifTrue:[rampIndex := rampSize - 1].
		self fillColorSpan: (self makeUnsignedFrom: (ramp at: rampIndex)) from: x0 to: x].

	"Part two: Fill everything inside the boundaries"
	self aaLevelGet = 1 ifTrue:[
		"Fast version w/o anti-aliasing"
		[((rampIndex := ds // 16r10000) < rampSize and:[rampIndex >= 0]) and:[x < x1]] whileTrue:[
			spanBuffer at: x put: (self makeUnsignedFrom: (ramp at: rampIndex)).
			x := x + 1.
			ds := ds + dsX.
		].
	] ifFalse:[x := self fillLinearGradientAA: fill ramp: ramp ds: ds dsX: dsX from: x to: rightX].
	"Part three fill everything outside right boundary"
	x < x1 ifTrue:[
		rampIndex < 0 ifTrue:[rampIndex := 0].
		rampIndex >= rampSize ifTrue:[rampIndex := rampSize-1].
		self fillColorSpan: (self makeUnsignedFrom: (ramp at: rampIndex)) from: x to: x1].
! !

!BalloonEnginePlugin methodsFor: 'fills-gradient' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:46'!
fillLinearGradientAA: fill ramp: ramp ds: deltaS dsX: dsX from: leftX to: rightX
	"This is the AA version of linear gradient filling."
	| colorMask colorShift baseShift rampIndex ds rampSize x idx rampValue 
	 aaLevel firstPixel lastPixel |
	<inline: false>
	<var: #ramp type:'int *'>

	aaLevel := self aaLevelGet.
	baseShift := self aaShiftGet.
	rampSize := self gradientRampLengthOf: fill.
	ds := deltaS.
	x := leftX.
	rampIndex := ds // 16r10000.

	firstPixel := self aaFirstPixelFrom: leftX to: rightX.
	lastPixel := self aaLastPixelFrom: leftX to: rightX.

	"Deal with the first n sub-pixels"
	colorMask := self aaColorMaskGet.
	colorShift := self aaColorShiftGet.
	[x < firstPixel and:[rampIndex < rampSize and:[rampIndex >= 0]]] whileTrue:[
		rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
		rampValue := (rampValue bitAnd: colorMask) >> colorShift.
		"Copy as many pixels as possible"
		[x < firstPixel and:[(ds//16r10000) = rampIndex]] whileTrue:[
			idx := x >> baseShift.
			spanBuffer at: idx put: (spanBuffer at: idx) + rampValue.
			x := x + 1.
			ds := ds + dsX].
		rampIndex := ds // 16r10000.
	].

	"Deal with the full pixels"
	colorMask := (self aaColorMaskGet >> self aaShiftGet) bitOr: 16rF0F0F0F0.
	colorShift := self aaShiftGet.
	[x < lastPixel and:[rampIndex < rampSize and:[rampIndex >= 0]]] whileTrue:[
		rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
		rampValue := (rampValue bitAnd: colorMask) >> colorShift.
		"Copy as many pixels as possible"
		[x < lastPixel and:[(ds//16r10000) = rampIndex]] whileTrue:[
			idx := x >> baseShift.
			spanBuffer at: idx put: (spanBuffer at: idx) + rampValue.
			x := x + aaLevel.
			ds := ds + (dsX << colorShift)].
		rampIndex := ds // 16r10000.
	].

	"Deal with the last n sub-pixels"
	colorMask := self aaColorMaskGet.
	colorShift := self aaColorShiftGet.
	[x < rightX and:[rampIndex < rampSize and:[rampIndex>=0]]] whileTrue:[
		rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
		rampValue := (rampValue bitAnd: colorMask) >> colorShift.
		"Copy as many pixels as possible"
		[x < rightX and:[(ds//16r10000) = rampIndex]] whileTrue:[
			idx := x >> baseShift.
			spanBuffer at: idx put: (spanBuffer at: idx) + rampValue.
			x := x + 1.
			ds := ds + dsX].
		rampIndex := ds // 16r10000.
	].
	^x! !

!BalloonEnginePlugin methodsFor: 'accessing fills' stamp: 'ar 11/24/1998 22:18'!
fillNormalXOf: fill

	^self obj: fill at: GFNormalX! !

!BalloonEnginePlugin methodsFor: 'accessing fills' stamp: 'ar 11/24/1998 22:18'!
fillNormalXOf: fill put: value

	^self obj: fill at: GFNormalX put: value! !

!BalloonEnginePlugin methodsFor: 'accessing fills' stamp: 'ar 11/24/1998 22:18'!
fillNormalYOf: fill

	^self obj: fill at: GFNormalY! !

!BalloonEnginePlugin methodsFor: 'accessing fills' stamp: 'ar 11/24/1998 22:16'!
fillNormalYOf: fill put: value

	^self obj: fill at: GFNormalY put: value! !

!BalloonEnginePlugin methodsFor: 'accessing fills' stamp: 'ar 11/24/1998 22:17'!
fillOriginXOf: fill

	^self obj: fill at: GFOriginX! !

!BalloonEnginePlugin methodsFor: 'accessing fills' stamp: 'ar 11/24/1998 22:17'!
fillOriginXOf: fill put: value

	^self obj: fill at: GFOriginX put: value! !

!BalloonEnginePlugin methodsFor: 'accessing fills' stamp: 'ar 11/24/1998 22:17'!
fillOriginYOf: fill

	^self obj: fill at: GFOriginY! !

!BalloonEnginePlugin methodsFor: 'accessing fills' stamp: 'ar 11/24/1998 22:18'!
fillOriginYOf: fill put: value

	^self obj: fill at: GFOriginY put: value! !

!BalloonEnginePlugin methodsFor: 'fills-gradient' stamp: 'ar (auto pragmas 12/08) 11/24/1998 19:02'!
fillRadialDecreasing: fill ramp: ramp deltaST: deltaST dsX: dsX dtX: dtX from: leftX to: rightX
	"Part 2a) Compute the decreasing part of the ramp"
	| ds dt rampIndex rampValue length2 x x1 nextLength |
	<inline: true>
	ds := (self cCoerce: deltaST to:'int*') at: 0.
	dt := (self cCoerce: deltaST to:'int*') at: 1.
	rampIndex := self accurateLengthOf: ds // 16r10000 with: dt // 16r10000.
	rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
	length2 := (rampIndex-1) * (rampIndex-1).

	x := leftX.
	x1 := rightX.
	x1 > (self fillOriginXOf: fill) ifTrue:[x1 := self fillOriginXOf: fill].
	[x < x1] whileTrue:[
		"Try to copy the current value more than just once"
		[x < x1 and:[(self squaredLengthOf: ds //  16r10000 with: dt // 16r10000) >= length2]]
			whileTrue:[	spanBuffer at: x put: rampValue.
						x := x + 1.
						ds := ds + dsX.
						dt := dt + dtX].
		"Step to next ramp value"
		nextLength := self squaredLengthOf: ds //  16r10000 with: dt // 16r10000.
		[nextLength < length2] whileTrue:[
			rampIndex := rampIndex - 1.
			rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
			length2 := (rampIndex-1) * (rampIndex-1).
		].
	].

	(self cCoerce: deltaST to: 'int *') at: 0 put: ds.
	(self cCoerce: deltaST to: 'int *') at: 1 put: dt.
	^x! !

!BalloonEnginePlugin methodsFor: 'fills-gradient' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:47'!
fillRadialDecreasingAA: fill ramp: ramp deltaST: deltaST dsX: dsX dtX: dtX from: leftX to: rightX
	"Part 2a) Compute the decreasing part of the ramp"
	| ds dt rampIndex rampValue length2 x nextLength x1
	aaLevel colorMask colorShift baseShift index firstPixel lastPixel |
	<inline: false>
	<var: #ramp type:'int *'>
	<var: #deltaST type:' int *'>

	ds := (self cCoerce: deltaST to:'int*') at: 0.
	dt := (self cCoerce: deltaST to:'int*') at: 1.
	aaLevel := self aaLevelGet.
	baseShift := self aaShiftGet.
	rampIndex := self accurateLengthOf: ds // 16r10000 with: dt // 16r10000.
	length2 := (rampIndex-1) * (rampIndex-1).

	x := leftX.
	x1 := self fillOriginXOf: fill.
	x1 > rightX ifTrue:[x1 := rightX].
	firstPixel := self aaFirstPixelFrom: leftX to: x1.
	lastPixel := self aaLastPixelFrom: leftX to: x1.

	"Deal with the first n sub-pixels"
	(x < firstPixel) ifTrue:[
		colorMask := self aaColorMaskGet.
		colorShift := self aaColorShiftGet.
		rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
		rampValue := (rampValue bitAnd: colorMask) >> colorShift.
		[x < firstPixel] whileTrue:[
			"Try to copy the current value more than just once"
			[x < firstPixel and:[(self squaredLengthOf: ds //  16r10000 with: dt // 16r10000) >= length2]]
				whileTrue:[	index := x >> baseShift.
							spanBuffer at: index put: (spanBuffer at: index) + rampValue.
							x := x + 1.
							ds := ds + dsX.
							dt := dt + dtX].
			"Step to next ramp value"
			nextLength := self squaredLengthOf: ds //  16r10000 with: dt // 16r10000.
			[nextLength < length2] whileTrue:[
				rampIndex := rampIndex - 1.
				rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
				rampValue := (rampValue bitAnd: colorMask) >> colorShift.
				length2 := (rampIndex-1) * (rampIndex-1).
			].
		].
	].

	"Deal with the full pixels"
	(x < lastPixel) ifTrue:[
		colorMask := (self aaColorMaskGet >> self aaShiftGet) bitOr: 16rF0F0F0F0.
		colorShift := self aaShiftGet.
		rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
		rampValue := (rampValue bitAnd: colorMask) >> colorShift.
		[x < lastPixel] whileTrue:[
			"Try to copy the current value more than just once"
			[x < lastPixel and:[(self squaredLengthOf: ds //  16r10000 with: dt // 16r10000) >= length2]]
				whileTrue:[	index := x >> baseShift.
							spanBuffer at: index put: (spanBuffer at: index) + rampValue.
							x := x + aaLevel.
							ds := ds + (dsX << colorShift).
							dt := dt + (dtX << colorShift)].
			"Step to next ramp value"
			nextLength := self squaredLengthOf: ds //  16r10000 with: dt // 16r10000.
			[nextLength < length2] whileTrue:[
				rampIndex := rampIndex - 1.
				rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
				rampValue := (rampValue bitAnd: colorMask) >> colorShift.
				length2 := (rampIndex-1) * (rampIndex-1).
			].
		].
	].

	"Deal with the last n sub-pixels"
	(x < x1) ifTrue:[
		colorMask := self aaColorMaskGet.
		colorShift := self aaColorShiftGet.
		rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
		rampValue := (rampValue bitAnd: colorMask) >> colorShift.
		[x < x1] whileTrue:[
			"Try to copy the current value more than just once"
			[x < x1 and:[(self squaredLengthOf: ds //  16r10000 with: dt // 16r10000) >= length2]]
				whileTrue:[	index := x >> baseShift.
							spanBuffer at: index put: (spanBuffer at: index) + rampValue.
							x := x + 1.
							ds := ds + dsX.
							dt := dt + dtX].
			"Step to next ramp value"
			nextLength := self squaredLengthOf: ds //  16r10000 with: dt // 16r10000.
			[nextLength < length2] whileTrue:[
				rampIndex := rampIndex - 1.
				rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
				rampValue := (rampValue bitAnd: colorMask) >> colorShift.
				length2 := (rampIndex-1) * (rampIndex-1).
			].
		].
	].
	"Done -- store stuff back"
	(self cCoerce: deltaST to: 'int *') at: 0 put: ds.
	(self cCoerce: deltaST to: 'int *') at: 1 put: dt.
	^x! !

!BalloonEnginePlugin methodsFor: 'fills-gradient' stamp: 'ar (auto pragmas 12/08) 11/8/1998 15:20'!
fillRadialGradient
	<inline: true>
	^self fillRadialGradient: self lastExportedFillGet from: self lastExportedLeftXGet to: self lastExportedRightXGet at: self currentYGet! !

!BalloonEnginePlugin methodsFor: 'fills-gradient' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:47'!
fillRadialGradient: fill from: leftX to: rightX at: yValue
	"Draw a radial gradient fill."
	| x x1 ramp rampSize dsX ds dtX dt length2 deltaX deltaY deltaST |
	<inline: false>
	<var: #ramp type:'int *'>
	<var: #deltaST type:'int *'>

	ramp := self gradientRampOf: fill.
	rampSize := self gradientRampLengthOf: fill.

	deltaX := leftX - (self fillOriginXOf: fill).
	deltaY := yValue - (self fillOriginYOf: fill).

	dsX := self fillDirectionXOf: fill.
	dtX := self fillNormalXOf: fill.

	ds := (deltaX * dsX) + (deltaY * (self fillDirectionYOf: fill)).
	dt := (deltaX * dtX) + (deltaY * (self fillNormalYOf: fill)).

	x := leftX.
	x1 := rightX.

	"Note: The inner loop has been divided into three parts for speed"
	"Part one: Fill everything outside the left boundary"
	length2 := (rampSize-1) * (rampSize-1). "This is the upper bound"
	[(self squaredLengthOf: ds // 16r10000 with: dt // 16r10000) >= length2 and:[x < x1]]
		whileTrue:[	x := x + 1.	ds := ds + dsX.	dt := dt + dtX].
	x > leftX ifTrue:[self fillColorSpan: (self makeUnsignedFrom: (ramp at: rampSize-1)) from: leftX to: x].

	"Part two: Fill everything inside the boundaries"
	deltaST := self point1Get.
	deltaST at: 0 put: ds.
	deltaST at: 1 put: dt.
	(x < (self fillOriginXOf: fill)) ifTrue:[
		"Draw the decreasing part"
		self aaLevelGet = 1 
			ifTrue:[x := self fillRadialDecreasing: fill ramp: ramp deltaST: deltaST 
							dsX: dsX dtX: dtX from: x to: x1]
			ifFalse:[x := self fillRadialDecreasingAA: fill ramp: ramp deltaST: deltaST 
							dsX: dsX dtX: dtX from: x to: x1].
	].
	x < x1 ifTrue:[
		"Draw the increasing part"
		self aaLevelGet = 1
			ifTrue:[x := self fillRadialIncreasing: fill ramp: ramp deltaST: deltaST
							dsX: dsX dtX: dtX from: x to: x1]
			ifFalse:[x := self fillRadialIncreasingAA: fill ramp: ramp deltaST: deltaST
							dsX: dsX dtX: dtX from: x to: x1].
	].

	"Part three fill everything outside right boundary"
	x < rightX ifTrue:[self fillColorSpan: (self makeUnsignedFrom: (ramp at: rampSize-1)) from: x to: rightX].
! !

!BalloonEnginePlugin methodsFor: 'fills-gradient' stamp: 'ar (auto pragmas 12/08) 11/9/1998 01:21'!
fillRadialIncreasing: fill ramp: ramp deltaST: deltaST dsX: dsX dtX: dtX from: leftX to: rightX
	"Part 2b) Compute the increasing part of the ramp"
	| ds dt rampIndex rampValue length2 x x1 nextLength rampSize lastLength |
	<inline: true>
	ds := (self cCoerce: deltaST to:'int*') at: 0.
	dt := (self cCoerce: deltaST to:'int*') at: 1.
	rampIndex := self accurateLengthOf: ds // 16r10000 with: dt // 16r10000.
	rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
	rampSize := self gradientRampLengthOf: fill.
	length2 := (rampSize-1) * (rampSize-1). "This is the upper bound"
	nextLength := (rampIndex+1) * (rampIndex+1).
	lastLength := self squaredLengthOf: ds //  16r10000 with: dt // 16r10000.

	x := leftX.
	x1 := rightX.

	[x < x1 and:[lastLength < length2]] whileTrue:[
		"Try to copy the current value more than once"
		[x < x1 and:[(self squaredLengthOf: ds //  16r10000 with: dt // 16r10000) <= nextLength]]
			whileTrue:[	spanBuffer at: x put: rampValue.
						x := x + 1.
						ds := ds + dsX.
						dt := dt + dtX].
		lastLength := self squaredLengthOf: ds //  16r10000 with: dt // 16r10000.
		[lastLength > nextLength] whileTrue:[
			rampIndex := rampIndex + 1.
			rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
			nextLength := (rampIndex+1) * (rampIndex+1).
		].
	].

	(self cCoerce: deltaST to: 'int *') at: 0 put: ds.
	(self cCoerce: deltaST to: 'int *') at: 1 put: dt.
	^x! !

!BalloonEnginePlugin methodsFor: 'fills-gradient' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:47'!
fillRadialIncreasingAA: fill ramp: ramp deltaST: deltaST dsX: dsX dtX: dtX from: leftX to: rightX
	"Part 2b) Compute the increasing part of the ramp"
	| ds dt rampIndex rampValue length2 x nextLength rampSize lastLength 
	aaLevel colorMask colorShift baseShift index firstPixel lastPixel |
	<inline: false>
	<var: #ramp type:'int *'>
	<var: #deltaST type:' int *'>

	ds := (self cCoerce: deltaST to:'int*') at: 0.
	dt := (self cCoerce: deltaST to:'int*') at: 1.
	aaLevel := self aaLevelGet.
	baseShift := self aaShiftGet.
	rampIndex := self accurateLengthOf: ds // 16r10000 with: dt // 16r10000.
	rampSize := self gradientRampLengthOf: fill.
	length2 := (rampSize-1) * (rampSize-1). "This is the upper bound"
	nextLength := (rampIndex+1) * (rampIndex+1).
	lastLength := self squaredLengthOf: ds //  16r10000 with: dt // 16r10000.

	x := leftX.

	firstPixel := self aaFirstPixelFrom: leftX to: rightX.
	lastPixel := self aaLastPixelFrom: leftX to: rightX.

	"Deal with the first n subPixels"
	(x < firstPixel and:[lastLength < length2]) ifTrue:[
		colorMask := self aaColorMaskGet.
		colorShift := self aaColorShiftGet.
		rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
		rampValue := (rampValue bitAnd: colorMask) >> colorShift.
		[x < firstPixel and:[lastLength < length2]] whileTrue:[
			"Try to copy the current value more than once"
			[x < firstPixel and:[(self squaredLengthOf: ds //  16r10000 with: dt // 16r10000) <= nextLength]]
				whileTrue:[	index := x >> baseShift.
							spanBuffer at: index put: (spanBuffer at: index) + rampValue.
							x := x + 1.
							ds := ds + dsX.
							dt := dt + dtX].
			lastLength := self squaredLengthOf: ds //  16r10000 with: dt // 16r10000.
			[lastLength > nextLength] whileTrue:[
				rampIndex := rampIndex + 1.
				rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
				rampValue := (rampValue bitAnd: colorMask) >> colorShift.
				nextLength := (rampIndex+1) * (rampIndex+1).
			].
		].
	].

	"Deal with the full pixels"
	(x < lastPixel and:[lastLength < length2]) ifTrue:[
		colorMask := (self aaColorMaskGet >> self aaShiftGet) bitOr: 16rF0F0F0F0.
		colorShift := self aaShiftGet.
		rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
		rampValue := (rampValue bitAnd: colorMask) >> colorShift.
		[x < lastPixel and:[lastLength < length2]] whileTrue:[
			"Try to copy the current value more than once"
			[x < lastPixel and:[(self squaredLengthOf: ds //  16r10000 with: dt // 16r10000) <= nextLength]]
				whileTrue:[	index := x >> baseShift.
							spanBuffer at: index put: (spanBuffer at: index) + rampValue.
							x := x + aaLevel.
							ds := ds + (dsX << colorShift).
							dt := dt + (dtX << colorShift)].
			lastLength := self squaredLengthOf: ds //  16r10000 with: dt // 16r10000.
			[lastLength > nextLength] whileTrue:[
				rampIndex := rampIndex + 1.
				rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
				rampValue := (rampValue bitAnd: colorMask) >> colorShift.
				nextLength := (rampIndex+1) * (rampIndex+1).
			].
		].
	].

	"Deal with last n sub-pixels"
	(x < rightX and:[lastLength < length2]) ifTrue:[
		colorMask := self aaColorMaskGet.
		colorShift := self aaColorShiftGet.
		rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
		rampValue := (rampValue bitAnd: colorMask) >> colorShift.
		[x < rightX and:[lastLength < length2]] whileTrue:[
			"Try to copy the current value more than once"
			[x < rightX and:[(self squaredLengthOf: ds //  16r10000 with: dt // 16r10000) <= nextLength]]
				whileTrue:[	index := x >> baseShift.
							spanBuffer at: index put: (spanBuffer at: index) + rampValue.
							x := x + 1.
							ds := ds + dsX.
							dt := dt + dtX].
			lastLength := self squaredLengthOf: ds //  16r10000 with: dt // 16r10000.
			[lastLength > nextLength] whileTrue:[
				rampIndex := rampIndex + 1.
				rampValue := self makeUnsignedFrom: ((self cCoerce: ramp to:'int *') at: rampIndex).
				rampValue := (rampValue bitAnd: colorMask) >> colorShift.
				nextLength := (rampIndex+1) * (rampIndex+1).
			].
		].
	].
	"Done -- store stuff back"
	(self cCoerce: deltaST to: 'int *') at: 0 put: ds.
	(self cCoerce: deltaST to: 'int *') at: 1 put: dt.
	^x! !

!BalloonEnginePlugin methodsFor: 'accessing gradients' stamp: 'ar 11/24/1998 22:18'!
gradientRampLengthOf: fill

	^self obj: fill at: GFRampLength! !

!BalloonEnginePlugin methodsFor: 'accessing gradients' stamp: 'ar 11/24/1998 22:17'!
gradientRampLengthOf: fill put: value

	^self obj: fill at: GFRampLength put: value! !

!BalloonEnginePlugin methodsFor: 'accessing gradients' stamp: 'ar (auto pragmas 12/08) 11/24/1998 22:25'!
gradientRampOf: fill
	<returnTypeC:'int *'>

	^objBuffer + fill +  GFRampOffset! !

!BalloonEnginePlugin methodsFor: 'testing' stamp: 'ar 11/4/1998 21:46'!
isBezier: bezier
	^((self objectTypeOf: bezier) bitAnd: GEPrimitiveWideMask) = GEPrimitiveBezier! !

!BalloonEnginePlugin methodsFor: 'testing' stamp: 'ar (auto pragmas 12/08) 11/8/1998 15:14'!
isFillOkay: fill
	<inline: false>
	^(fill = 0 or:[(self isFillColor: fill) or:[((self isObject: fill) and:[self isFill: fill])]]) 
! !

!BalloonEnginePlugin methodsFor: 'testing' stamp: 'ar 11/4/1998 21:46'!
isLine: line
	^((self objectTypeOf: line) bitAnd: GEPrimitiveWideMask) = GEPrimitiveLine! !

!BalloonEnginePlugin methodsFor: 'testing' stamp: 'ar 11/6/1998 01:53'!
isWideBezier: bezier
	^(self isBezier: bezier) and:[self isWide: bezier]! !

!BalloonEnginePlugin methodsFor: 'testing' stamp: 'ar 11/4/1998 22:08'!
isWideLine: line
	^(self isLine: line) and:[self isWide: line]! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:07'!
lineEndXOf: line

	^self obj: line at: GLEndX! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:07'!
lineEndXOf: line put: value

	^self obj: line at: GLEndX put: value! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:07'!
lineEndYOf: line

	^self obj: line at: GLEndY! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:07'!
lineEndYOf: line put: value

	^self obj: line at: GLEndY put: value! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:07'!
lineErrorAdjDownOf: line

	^self obj: line at: GLErrorAdjDown! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:07'!
lineErrorAdjDownOf: line put: value

	^self obj: line at: GLErrorAdjDown put: value! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:08'!
lineErrorAdjUpOf: line

	^self obj: line at: GLErrorAdjUp! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:08'!
lineErrorAdjUpOf: line put: value

	^self obj: line at: GLErrorAdjUp put: value! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:08'!
lineErrorOf: line

	^self obj: line at: GLError! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:08'!
lineErrorOf: line put: value

	^self obj: line at: GLError put: value! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:08'!
lineXDirectionOf: line

	^self obj: line at: GLXDirection! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:09'!
lineXDirectionOf: line put: value

	^self obj: line at: GLXDirection put: value! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:09'!
lineXIncrementOf: line

	^self obj: line at: GLXIncrement! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:09'!
lineXIncrementOf: line put: value

	^self obj: line at: GLXIncrement put: value! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:09'!
lineYDirectionOf: line

	^self obj: line at: GLYDirection! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:09'!
lineYDirectionOf: line put: value

	^self obj: line at: GLYDirection put: value! !

!BalloonEnginePlugin methodsFor: 'bezier-loading' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:47'!
loadAndSubdivideBezierFrom: point1 via: point2 to: point3 isWide: wideFlag
	"Load and subdivide the bezier curve from point1/point2/point3.
	If wideFlag is set then make sure the curve is monoton in X."
	| bz1 bz2 index2 index1 |
	<inline: false>
	<var: #point1 type:'int *'>
	<var: #point2 type:'int *'>
	<var: #point3 type:'int *'>
	bz1 := self allocateBezierStackEntry.	
	engineStopped ifTrue:[^0].
	"Load point1/point2/point3 on the top of the stack"
	self bzStartX: bz1 put: (point1 at: 0).
	self bzStartY: bz1 put: (point1 at: 1).
	self bzViaX: bz1 put: (point2 at: 0).
	self bzViaY: bz1 put: (point2 at: 1).
	self bzEndX: bz1 put: (point3 at: 0).
	self bzEndY: bz1 put: (point3 at: 1).

	"Now check if the bezier curve is monoton. If not, subdivide it."
	index2 := bz2 := self subdivideToBeMonoton: bz1 inX: wideFlag.
	bz1 to: bz2 by: 6 do:[:index|
		index1 := self subdivideBezierFrom: index.
		index1 > index2 ifTrue:[index2 := index1].
		engineStopped ifTrue:[^0]. "Something went wrong"
	].
	"Return the number of segments"
	^index2 // 6! !

!BalloonEnginePlugin methodsFor: 'shapes-polygons' stamp: 'ar 11/24/1998 23:09'!
loadArrayPolygon: points nPoints: nPoints fill: fillIndex lineWidth: lineWidth lineFill: lineFill
	| x0 y0 x1 y1 |
	self loadPoint: self point1Get from: (interpreterProxy fetchPointer: 0 ofObject: points).
	interpreterProxy failed ifTrue:[^nil].
	x0 := self point1Get at: 0.
	y0 := self point1Get at: 1.
	1 to: nPoints-1 do:[:i|
		self loadPoint: self point1Get from: (interpreterProxy fetchPointer: i ofObject: points).
		interpreterProxy failed ifTrue:[^nil].
		x1 := self point1Get at: 0.
		y1 := self point1Get at: 1.
		self point1Get at: 0 put: x0.
		self point1Get at: 1 put: y0.
		self point2Get at: 0 put: x1.
		self point2Get at: 1 put: y1.
		self transformPoints: 2.
		self loadWideLine: lineWidth 
			from: self point1Get
			to: self point2Get
			lineFill: lineFill 
			leftFill: fillIndex
			rightFill: 0.
		engineStopped ifTrue:[^nil].
		x0 := x1.
		y0 := y1.
	].! !

!BalloonEnginePlugin methodsFor: 'shapes-polygons' stamp: 'ar (auto pragmas 12/08) 11/24/1998 23:14'!
loadArrayShape: points nSegments: nSegments fill: fillIndex lineWidth: lineWidth lineFill: lineFill 
	| pointOop x0 y0 x1 y1 x2 y2 segs |
	<inline: false>
	0 to: nSegments-1 do:[:i|
		pointOop := interpreterProxy fetchPointer: (i * 3) ofObject: points.
		self loadPoint: self point1Get from: pointOop.
		pointOop := interpreterProxy fetchPointer: (i * 3 + 1) ofObject: points.
		self loadPoint: self point2Get from: pointOop.
		pointOop := interpreterProxy fetchPointer: (i * 3 + 2) ofObject: points.
		self loadPoint: self point3Get from: pointOop.
		interpreterProxy failed ifTrue:[^nil].
		self transformPoints: 3.
		x0 := self point1Get at: 0.
		y0 := self point1Get at: 1.
		x1 := self point2Get at: 0.
		y1 := self point2Get at: 1.
		x2 := self point3Get at: 0.
		y2 := self point3Get at: 1.
		"Check if we can use a line"
		((x0 = y0 and:[x1 = y1]) or:[x1 = x2 and:[y1 = y2]]) ifTrue:[
			self loadWideLine: lineWidth
				from: self point1Get
				to: self point3Get
				lineFill: lineFill
				leftFill: fillIndex
				rightFill: 0.
		] ifFalse:["Need bezier"
			segs := self loadAndSubdivideBezierFrom: self point1Get
					via: self point2Get
					to: self point3Get
					isWide: (lineWidth ~= 0 and:[lineFill ~= 0]).
			engineStopped ifTrue:[^nil].
			self loadWideBezier: lineWidth
				lineFill: lineFill
				leftFill: fillIndex
				rightFill: 0
				n: segs.
		].
		engineStopped ifTrue:[^nil].
	].! !

!BalloonEnginePlugin methodsFor: 'bezier-loading' stamp: 'ar (auto pragmas 12/08) 11/24/1998 23:15'!
loadBezier: bezier segment: index leftFill: leftFillIndex rightFill: rightFillIndex offset: yOffset
	"Initialize the bezier segment stored on the stack"
	<inline: false>
	(self bzEndY: index) >= (self bzStartY: index) ifTrue:[
		"Top to bottom"
		self edgeXValueOf: bezier put: (self bzStartX: index).
		self edgeYValueOf: bezier put: (self bzStartY: index) - yOffset.
		self bezierViaXOf: bezier put: (self bzViaX: index).
		self bezierViaYOf: bezier put: (self bzViaY: index) - yOffset.
		self bezierEndXOf: bezier put: (self bzEndX: index).
		self bezierEndYOf: bezier put: (self bzEndY: index) - yOffset.
	] ifFalse:[
		self edgeXValueOf: bezier put: (self bzEndX: index).
		self edgeYValueOf: bezier put: (self bzEndY: index) - yOffset.
		self bezierViaXOf: bezier put: (self bzViaX: index).
		self bezierViaYOf: bezier put: (self bzViaY: index) - yOffset.
		self bezierEndXOf: bezier put: (self bzStartX: index).
		self bezierEndYOf: bezier put: (self bzStartY: index) - yOffset.
	].
	self edgeZValueOf: bezier put: self currentZGet.
	self edgeLeftFillOf: bezier put: leftFillIndex.
	self edgeRightFillOf: bezier put: rightFillIndex.
	"self debugDrawBezier: bezier."! !

!BalloonEnginePlugin methodsFor: 'fills-bitmaps' stamp: 'eem 1/30/2014 16:02'!
loadBitmapFill: formOop colormap: cmOop tile: tileFlag from: point1 along: point2 normal: point3 xIndex: xIndex
	"Load the bitmap fill."
	| bmFill cmSize cmBits bmBits bmBitsSize bmWidth bmHeight bmDepth ppw bmRaster |
	<var: #cmBits type: #'int *'>
	<var: #point1 type: #'int *'>
	<var: #point2 type: #'int *'>
	<var: #point3 type: #'int *'>

	cmOop == interpreterProxy nilObject ifTrue:[
		cmSize := 0.
		cmBits := nil.
	] ifFalse:[
		(interpreterProxy fetchClassOf: cmOop) = interpreterProxy classBitmap
			ifFalse:[^interpreterProxy primitiveFail].
		cmSize := interpreterProxy slotSizeOf: cmOop.
		cmBits := interpreterProxy firstIndexableField: cmOop.
	].
	(interpreterProxy isPointers: formOop) 
		ifFalse:[^interpreterProxy primitiveFail].
	(interpreterProxy slotSizeOf: formOop) < 5 
		ifTrue:[^interpreterProxy primitiveFail].
	bmBits := interpreterProxy fetchPointer: 0 ofObject: formOop.
	(interpreterProxy fetchClassOf: bmBits) = interpreterProxy classBitmap
		ifFalse:[^interpreterProxy primitiveFail].
	bmBitsSize := interpreterProxy slotSizeOf: bmBits.
	bmWidth := interpreterProxy fetchInteger: 1 ofObject: formOop.
	bmHeight := interpreterProxy fetchInteger: 2 ofObject: formOop.
	bmDepth := interpreterProxy fetchInteger: 3 ofObject: formOop.
	interpreterProxy failed ifTrue:[^nil].
	(bmWidth >= 0 and:[bmHeight >= 0]) ifFalse:[^interpreterProxy primitiveFail].
	(bmDepth = 32) | (bmDepth = 8) | (bmDepth = 16) | 
		(bmDepth = 1) | (bmDepth = 2) | (bmDepth = 4)
			ifFalse:[^interpreterProxy primitiveFail].
	(cmSize = 0 or:[cmSize = (1 << bmDepth)])
		ifFalse:[^interpreterProxy primitiveFail].
	ppw := 32 // bmDepth.
	bmRaster := bmWidth + (ppw-1) // ppw.
	bmBitsSize = (bmRaster * bmHeight)
		ifFalse:[^interpreterProxy primitiveFail].
	bmFill := self allocateBitmapFill: cmSize colormap: cmBits.
	engineStopped ifTrue:[^nil].
	self bitmapWidthOf: bmFill put: bmWidth.
	self bitmapHeightOf: bmFill put: bmHeight.
	self bitmapDepthOf: bmFill put: bmDepth.
	self bitmapRasterOf: bmFill put: bmRaster.
	self bitmapSizeOf: bmFill put: bmBitsSize.
	self bitmapTileFlagOf: bmFill put: tileFlag.
	self objectIndexOf: bmFill put: xIndex.
	self loadFillOrientation: bmFill
		from: point1 along: point2 normal: point3
		width: bmWidth height: bmHeight.
	^bmFill! !

!BalloonEnginePlugin methodsFor: 'fills-bitmaps' stamp: 'ar (auto pragmas 12/08) 11/25/1998 17:25'!
loadBitsFrom: bmFill
	"Note: Assumes that the contents of formArray has been checked before"
	| xIndex formOop bitsOop bitsLen |
	<returnTypeC:'int *'>
	xIndex := self objectIndexOf: bmFill.
	xIndex > (interpreterProxy slotSizeOf: formArray) ifTrue:[^nil].
	formOop := interpreterProxy fetchPointer: xIndex ofObject: formArray.
	bitsOop := interpreterProxy fetchPointer: 0 ofObject: formOop.
	bitsLen := interpreterProxy slotSizeOf: bitsOop.
	bitsLen = (self bitmapSizeOf: bmFill) ifFalse:[^nil].
	^interpreterProxy firstIndexableField: bitsOop! !

!BalloonEnginePlugin methodsFor: 'shapes-compressed' stamp: 'ar (auto pragmas 12/08) 11/24/1998 21:13'!
loadCompressedSegment: segmentIndex from: points short: pointsShort leftFill: leftFill rightFill: rightFill lineWidth: lineWidth lineColor: lineFill 
	"Load the compressed segment identified by segment index"
	| x0 y0 x1 y1 x2 y2 index segs |
	<inline: true>

	"Check if have anything to do at all"
	(leftFill = rightFill and:[lineWidth = 0 or:[lineFill = 0]]) 
		ifTrue:[^nil]. "Nothing to do"

	index := segmentIndex * 6. "3 points with x/y each"
	pointsShort ifTrue:["Load short points"
		x0 := self loadPointShortAt: (index+0) from: points.
		y0 := self loadPointShortAt: (index+1) from: points.
		x1 := self loadPointShortAt: (index+2) from: points.
		y1 := self loadPointShortAt: (index+3) from: points.
		x2 := self loadPointShortAt: (index+4) from: points.
		y2 := self loadPointShortAt: (index+5) from: points.
	] ifFalse:[
		x0 := self loadPointIntAt: (index+0) from: points.
		y0 := self loadPointIntAt: (index+1) from: points.
		x1 := self loadPointIntAt: (index+2) from: points.
		y1 := self loadPointIntAt: (index+3) from: points.
		x2 := self loadPointIntAt: (index+4) from: points.
		y2 := self loadPointIntAt: (index+5) from: points.
	].
	"Briefly check if can represent the bezier as a line"
	((x0 = x1 and:[y0 = y1]) or:[x1 = x2 and:[y1 = y2]]) ifTrue:[
		"We can use a line from x0/y0 to x2/y2"
		(x0 = x2 and:[y0 = y2]) ifTrue:[^nil]. "Nothing to do"
		"Load and transform points"
		self point1Get at: 0 put: x0.
		self point1Get at: 1 put: y0.
		self point2Get at: 0 put: x2.
		self point2Get at: 1 put: y2.
		self transformPoints: 2.
		^self loadWideLine: lineWidth 
			from: self point1Get
			to: self point2Get
			lineFill: lineFill 
			leftFill: leftFill 
			rightFill: rightFill.
	].
	"Need bezier curve"
	"Load and transform points"
	self point1Get at: 0 put: x0.
	self point1Get at: 1 put: y0.
	self point2Get at: 0 put: x1.
	self point2Get at: 1 put: y1.
	self point3Get at: 0 put: x2.
	self point3Get at: 1 put: y2.
	self transformPoints: 3.
	segs := self loadAndSubdivideBezierFrom: self point1Get 
				via: self point2Get 
				to: self point3Get 
				isWide: (lineWidth ~= 0 and:[lineFill ~= 0]).
	engineStopped ifTrue:[^nil].
	self loadWideBezier: lineWidth lineFill: lineFill leftFill: leftFill rightFill: rightFill n: segs.
! !

!BalloonEnginePlugin methodsFor: 'shapes-compressed' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:48'!
loadCompressedShape: points segments: nSegments leftFills: leftFills rightFills: rightFills lineWidths: lineWidths lineFills: lineFills fillIndexList: fillIndexList pointShort: pointsShort
	"Load a compressed shape into the engine.
		WARNING: THIS METHOD NEEDS THE FULL FRAME SIZE!!!!!!!!
	"
	| leftRun rightRun widthRun lineFillRun
	leftLength rightLength widthLength lineFillLength
	leftValue rightValue widthValue lineFillValue |

	<inline: false>
	<var: #points type:'int *'>
	<var: #leftFills type:'int *'>
	<var: #rightFills type:'int *'>
	<var: #lineWidths type:'int *'>
	<var: #lineFills type:'int *'>
	<var: #fillIndexList type:'int *'>

	nSegments = 0 ifTrue:[^0].

	"Initialize run length encodings"
	leftRun :=  rightRun := widthRun := lineFillRun := -1.
	leftLength := rightLength := widthLength := lineFillLength := 1.
	leftValue := rightValue := widthValue := lineFillValue := 0.

	1 to: nSegments do:[:i|
		"Decrement current run length and load new stuff"
		(leftLength := leftLength - 1) <= 0 ifTrue:[
			leftRun := leftRun + 1.
			leftLength := self shortRunLengthAt: leftRun from: leftFills.
			leftValue := self shortRunValueAt: leftRun from: leftFills.
			leftValue = 0 ifFalse:[
				leftValue := fillIndexList at: leftValue-1.
				leftValue := self transformColor: leftValue.
				engineStopped ifTrue:[^nil]]].
		(rightLength := rightLength - 1) <= 0 ifTrue:[
			rightRun := rightRun + 1.
			rightLength := self shortRunLengthAt: rightRun from: rightFills.
			rightValue := self shortRunValueAt: rightRun from: rightFills.
			rightValue = 0 ifFalse:[
				rightValue := fillIndexList at: rightValue-1.
				rightValue := self transformColor: rightValue]].
		(widthLength := widthLength - 1) <= 0 ifTrue:[
			widthRun := widthRun + 1.
			widthLength := self shortRunLengthAt: widthRun from: lineWidths.
			widthValue := self shortRunValueAt: widthRun from: lineWidths.
			widthValue = 0 ifFalse:[widthValue := self transformWidth: widthValue]].
		(lineFillLength := lineFillLength - 1) <= 0 ifTrue:[
			lineFillRun := lineFillRun + 1.
			lineFillLength := self shortRunLengthAt: lineFillRun from: lineFills.
			lineFillValue := self shortRunValueAt: lineFillRun from: lineFills.
			lineFillValue = 0 ifFalse:[lineFillValue := fillIndexList at: lineFillValue-1]].
		self loadCompressedSegment: i - 1
			from: points 
			short: pointsShort 
			leftFill: leftValue 
			rightFill: rightValue 
			lineWidth: widthValue 
			lineColor: lineFillValue.
		engineStopped ifTrue:[^nil].
	].! !

!BalloonEnginePlugin methodsFor: 'fills-gradient' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:48'!
loadFillOrientation: fill from: point1 along: point2 normal: point3 width: fillWidth height: fillHeight
	"Transform the points"
	| dirX dirY nrmX nrmY dsLength2 dsX dsY dtLength2 dtX dtY |
	<var: #point1 type:'int *'>
	<var: #point2 type:'int *'>
	<var: #point3 type:'int *'>

	point2 at: 0 put: (point2 at: 0) + (point1 at: 0).
	point2 at: 1 put: (point2 at: 1) + (point1 at: 1).
	point3 at: 0 put: (point3 at: 0) + (point1 at: 0).
	point3 at: 1 put: (point3 at: 1) + (point1 at: 1).
	self transformPoint: point1.
	self transformPoint: point2.
	self transformPoint: point3.
	dirX := (point2 at: 0) - (point1 at: 0).
	dirY := (point2 at: 1) - (point1 at: 1).
	nrmX := (point3 at: 0) - (point1 at: 0).
	nrmY := (point3 at: 1) - (point1 at: 1).

	"Compute the scale from direction/normal into ramp size"
	dsLength2 := (dirX * dirX) + (dirY * dirY).
	dsLength2 > 0 ifTrue:[
		dsX := (dirX asFloat * fillWidth asFloat * 65536.0 / dsLength2 asFloat) asInteger.
		dsY := (dirY asFloat * fillWidth asFloat * 65536.0 / dsLength2 asFloat) asInteger.
	] ifFalse:[ dsX := 0. dsY := 0].
	dtLength2 := (nrmX * nrmX) + (nrmY * nrmY).
	dtLength2 > 0 ifTrue:[
		dtX := (nrmX asFloat * fillHeight asFloat * 65536.0 / dtLength2 asFloat) asInteger.
		dtY := (nrmY asFloat * fillHeight asFloat * 65536.0 / dtLength2 asFloat) asInteger.
	] ifFalse:[dtX := 0. dtY := 0].
	self fillOriginXOf: fill put: (point1 at: 0).
	self fillOriginYOf: fill put: (point1 at: 1).
	self fillDirectionXOf: fill put: dsX.
	self fillDirectionYOf: fill put: dsY.
	self fillNormalXOf: fill put: dtX.
	self fillNormalYOf: fill put: dtY.
! !

!BalloonEnginePlugin methodsFor: 'fills-gradient' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:50'!
loadGradientFill: rampOop from: point1 along: point2 normal: point3 isRadial: isRadial
	"Load the gradient fill as defined by the color ramp."
	| rampWidth fill |
	<inline: false>
	<var: #point1 type:'int *'>
	<var: #point2 type:'int *'>
	<var: #point3 type:'int *'>
	(interpreterProxy fetchClassOf: rampOop) = interpreterProxy classBitmap
		ifFalse:[^interpreterProxy primitiveFail].
	rampWidth := interpreterProxy slotSizeOf: rampOop.
	fill := self allocateGradientFill: (interpreterProxy firstIndexableField: rampOop)
				rampWidth: rampWidth isRadial: isRadial.
	engineStopped ifTrue:[^nil].
	self loadFillOrientation: fill 
		from: point1 along: point2 normal: point3 
		width: rampWidth height: rampWidth.
	^fill! !

!BalloonEnginePlugin methodsFor: 'lines-loading' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:50'!
loadLine: line from: point1 to: point2 offset: yOffset leftFill: leftFill rightFill: rightFill
	"Load the line defined by point1 and point2."
	| p1 p2 yDir |
	<var: #point1 type:'int *'>
	<var: #point2 type:'int *'>
	<var: #p1 type:'int *'>
	<var: #p2 type:'int *'>

	(point1 at: 1) <= (point2 at: 1) 
		ifTrue:[	p1 := point1.
				p2 := point2.
				yDir := 1]
		ifFalse:[	p1 := point2.
				p2 := point1.
				yDir := -1].
	self edgeXValueOf: line put: (p1 at: 0).
	self edgeYValueOf: line put: (p1 at: 1) - yOffset.
	self edgeZValueOf: line put: self currentZGet.
	self edgeLeftFillOf: line put: leftFill.
	self edgeRightFillOf: line put: rightFill.
	self lineEndXOf: line put: (p2 at: 0).
	self lineEndYOf: line put: (p2 at: 1) - yOffset.
	self lineYDirectionOf: line put: yDir.! !

!BalloonEnginePlugin methodsFor: 'bezier-loading' stamp: 'ar (auto pragmas 12/08) 11/25/1998 23:21'!
loadOval: lineWidth lineFill: lineFill leftFill: leftFill rightFill: rightFill
	"Load a rectangular oval currently defined by point1/point2"
	| w h cx cy nSegments |
	<inline: false>
	w := ((self point2Get at: 0) - (self point1Get at: 0)) // 2.
	h := ((self point2Get at: 1) - (self point1Get at: 1)) // 2.
	cx := ((self point2Get at: 0) + (self point1Get at: 0)) // 2.
	cy := ((self point2Get at: 1) + (self point1Get at: 1)) // 2.
	0 to: 15 do:[:i|
		self loadOvalSegment: i w: w h: h cx: cx cy: cy.
		self transformPoints: 3.
		nSegments := self loadAndSubdivideBezierFrom: self point1Get 
							via: self point2Get to: self point3Get
							isWide: (lineWidth ~= 0 and:[lineFill ~= 0]).
		engineStopped ifTrue:[^nil].
		self loadWideBezier: lineWidth lineFill: lineFill 
			leftFill: leftFill rightFill: rightFill n: nSegments.
		engineStopped ifTrue:[^nil].
	].! !

!BalloonEnginePlugin methodsFor: 'bezier-loading' stamp: 'ar (auto pragmas 12/08) 11/8/1998 15:17'!
loadOvalSegment: seg w: w h: h cx: cx cy: cy
	| x0 y0 x2 y2 x1 y1 |
	<inline: false>
	"Load start point of segment"
	x0 := ((self circleCosTable at: seg * 2 + 0) * w asFloat + cx) asInteger.
	y0 := ((self circleSinTable at: seg * 2 + 0) * h asFloat + cy) asInteger.
	self point1Get at: 0 put: x0.
	self point1Get at: 1 put: y0.
	"Load end point of segment"
	x2 := ((self circleCosTable at: seg * 2 + 2) * w asFloat + cx) asInteger.
	y2 := ((self circleSinTable at: seg * 2 + 2) * h asFloat + cy) asInteger.
	self point3Get at: 0 put: x2.
	self point3Get at: 1 put: y2.
	"Load intermediate point of segment"
	x1 := ((self circleCosTable at: seg * 2 + 1) * w asFloat + cx) asInteger.
	y1 := ((self circleSinTable at: seg * 2 + 1) * h asFloat + cy) asInteger.
	"NOTE: The intermediate point is the point ON the curve
	and not yet the control point (which is OFF the curve)"
	x1 := (x1 * 2) - (x0 + x2 // 2).
	y1 := (y1 * 2) - (y0 + y2 // 2).
	self point2Get at: 0 put: x1.
	self point2Get at: 1 put: y1.! !

!BalloonEnginePlugin methodsFor: 'private' stamp: 'ar 11/3/1998 23:24'!
loadPointIntAt: index from: intArray
	"Load the int value from the given index in intArray"
	^(self cCoerce: intArray to: 'int *') at: index! !

!BalloonEnginePlugin methodsFor: 'private' stamp: 'ar (auto pragmas 12/08) 11/3/1998 23:23'!
loadPointShortAt: index from: shortArray
	"Load the short value from the given index in shortArray"
	<returnTypeC:'short'>
	^(self cCoerce: shortArray to: 'short *') at: index! !

!BalloonEnginePlugin methodsFor: 'shapes-polygons' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:51'!
loadPolygon: points nPoints: nPoints fill: fillIndex lineWidth: lineWidth lineFill: lineFill pointsShort: isShort
	| x0 y0 x1 y1 |
	<var:#points type:'int *'>
	isShort ifTrue:[
		x0 := self loadPointShortAt: 0 from: points.
		y0 := self loadPointShortAt: 1 from: points.
	] ifFalse:[
		x0 := self loadPointIntAt: 0 from: points.
		y0 := self loadPointIntAt: 1 from: points.
	].
	1 to: nPoints-1 do:[:i|
		isShort ifTrue:[
			x1 := self loadPointShortAt: i*2 from: points.
			y1 := self loadPointShortAt: i*2+1 from: points.
		] ifFalse:[
			x1 := self loadPointIntAt: i*2 from: points.
			y1 := self loadPointIntAt: i*2+1 from: points.
		].
		self point1Get at: 0 put: x0.
		self point1Get at: 1 put: y0.
		self point2Get at: 0 put: x1.
		self point2Get at: 1 put: y1.
		self transformPoints: 2.
		self loadWideLine: lineWidth 
			from: self point1Get
			to: self point2Get
			lineFill: lineFill 
			leftFill: fillIndex
			rightFill: 0.
		engineStopped ifTrue:[^nil].
		x0 := x1.
		y0 := y1.
	].! !

!BalloonEnginePlugin methodsFor: 'lines-loading' stamp: 'ar 11/6/1998 17:07'!
loadRectangle: lineWidth lineFill: lineFill leftFill: leftFill rightFill: rightFill
	"Load a rectangle currently defined by point1-point4"

	self loadWideLine: lineWidth from: self point1Get to: self point2Get
		lineFill: lineFill leftFill: leftFill rightFill: rightFill.
	self loadWideLine: lineWidth from: self point2Get to: self point3Get
		lineFill: lineFill leftFill: leftFill rightFill: rightFill.
	self loadWideLine: lineWidth from: self point3Get to: self point4Get
		lineFill: lineFill leftFill: leftFill rightFill: rightFill.
	self loadWideLine: lineWidth from: self point4Get to: self point1Get
		lineFill: lineFill leftFill: leftFill rightFill: rightFill.
! !

!BalloonEnginePlugin methodsFor: 'shapes-polygons' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:51'!
loadShape: points nSegments: nSegments fill: fillIndex lineWidth: lineWidth lineFill: lineFill  pointsShort: pointsShort
	<inline: false>
	<var:#points type:'int *'>
	1 to: nSegments do:[:i|
		self loadCompressedSegment: i-1
			from: points
			short: pointsShort
			leftFill: fillIndex
			rightFill: 0
			lineWidth: lineWidth
			lineColor: lineFill.
		engineStopped ifTrue:[^nil].
	].! !

!BalloonEnginePlugin methodsFor: 'bezier-loading' stamp: 'ar (auto pragmas 12/08) 11/8/1998 03:41'!
loadWideBezier: lineWidth lineFill: lineFill leftFill: leftFill rightFill: rightFill n: nSegments
	"Load the (possibly wide) bezier from the segments currently on the bezier stack."
	| index bezier wide offset |
	<inline: false>
	(lineWidth = 0 or:[lineFill = 0])
		ifTrue:[wide := false.
				offset := 0]
		ifFalse:[wide := true.
				offset := self offsetFromWidth: lineWidth].
	index := nSegments * 6.
	[index > 0] whileTrue:[
		wide 
			ifTrue:[bezier := self allocateWideBezier]
			ifFalse:[bezier := self allocateBezier].
		engineStopped ifTrue:[^0].
		self loadBezier: bezier 
			segment: index 
			leftFill: leftFill 
			rightFill: rightFill 
			offset: offset.
		wide ifTrue:[
			self wideBezierFillOf: bezier put: lineFill.
			self wideBezierWidthOf: bezier put: lineWidth.
			self wideBezierExtentOf: bezier put: lineWidth.
		].
		index := index - 6.
	].
	self wbStackClear.! !

!BalloonEnginePlugin methodsFor: 'lines-loading' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:51'!
loadWideLine: lineWidth from: p1 to: p2 lineFill: lineFill leftFill: leftFill rightFill: rightFill
	"Load a (possibly wide) line defined by the points p1 and p2"
	| line offset |
	<var: #p1 type:'int *'>
	<var: #p2 type:'int *'>
	(lineWidth = 0 or:[lineFill = 0])
		ifTrue:[	line := self allocateLine.
				offset := 0]
		ifFalse:[	line := self allocateWideLine.
				offset := self offsetFromWidth: lineWidth].
	engineStopped ifTrue:[^0].
	self loadLine: line 
		from: p1
		to: p2
		offset: offset 
		leftFill: leftFill
		rightFill: rightFill.
	(self isWide: line) ifTrue:[
		self wideLineFillOf: line put: lineFill.
		self wideLineWidthOf: line put: lineWidth.
		self wideLineExtentOf: line put: lineWidth].! !

!BalloonEnginePlugin methodsFor: 'private' stamp: 'ar 11/1/1998 03:16'!
makeRectFromPoints
	self point2Get at: 0 put: (self point3Get at: 0).
	self point2Get at: 1 put: (self point1Get at: 1).
	self point4Get at: 0 put: (self point1Get at: 0).
	self point4Get at: 1 put: (self point3Get at: 1).! !

!BalloonEnginePlugin methodsFor: 'private' stamp: 'ar (auto pragmas 12/08) 11/6/1998 17:55'!
offsetFromWidth: lineWidth
	"Common function so that we don't compute that wrong in any place
	and can easily find all the places where we deal with one-pixel offsets."
	<inline: true>
	^lineWidth // 2! !

!BalloonEnginePlugin methodsFor: 'primitives' stamp: 'eem 6/3/2009 17:14'!
primitiveAddBezier
	| failureCode leftFill rightFill viaOop endOop startOop nSegments |
	<export: true>
	<inline: false>

	"Fail if we have the wrong number of arguments"
	interpreterProxy methodArgumentCount = 5 
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	rightFill := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	leftFill := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 1).
	viaOop := interpreterProxy stackObjectValue: 2.
	endOop := interpreterProxy stackObjectValue: 3.
	startOop := interpreterProxy stackObjectValue: 4.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 5)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	"Make sure the fills are okay"
	((self isFillOkay: leftFill) and:[self isFillOkay: rightFill])
		ifFalse:[^interpreterProxy primitiveFailFor: GEFWrongFill].

	"Do a quick check if the fillIndices are equal - if so, just ignore it"
	leftFill = rightFill & false ifTrue:[
		^interpreterProxy pop: 6. "Leave rcvr on stack"
	].


	self loadPoint: self point1Get from: startOop.
	self loadPoint: self point2Get from: viaOop.
	self loadPoint: self point3Get from: endOop.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	self transformPoints: 3.

	nSegments := self loadAndSubdivideBezierFrom: self point1Get 
						via: self point2Get 
						to: self point3Get 
						isWide: false.
	self needAvailableSpace: nSegments * GBBaseSize.
	engineStopped ifFalse:[
		leftFill := self transformColor: leftFill.
		rightFill := self transformColor: rightFill].
	engineStopped ifFalse:[
		self loadWideBezier: 0 lineFill: 0 leftFill: leftFill rightFill: rightFill n: nSegments.
	].
	engineStopped ifTrue:[
		"Make sure the stack is okay"
		self wbStackClear.
		^interpreterProxy primitiveFailFor: GEFEngineStopped].

	interpreterProxy failed
		ifTrue:[^interpreterProxy primitiveFailFor: GEFEntityLoadFailed].
	self storeEngineStateInto: engine.
	interpreterProxy pop: 5. "Leave rcvr on stack"! !

!BalloonEnginePlugin methodsFor: 'primitives' stamp: 'eem 12/6/2012 12:37'!
primitiveAddBezierShape
	| failureCode points lineFill lineWidth fillIndex length pointsIsArray segSize nSegments |
	<export: true>
	<inline: false>

	"Fail if we have the wrong number of arguments"
	interpreterProxy methodArgumentCount = 5 
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	lineFill := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	lineWidth := interpreterProxy stackIntegerValue: 1.
	fillIndex := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 2).
	nSegments := interpreterProxy stackIntegerValue: 3.
	points := interpreterProxy stackObjectValue: 4.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 5)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	"First, do a check if the points look okay"
	length := interpreterProxy slotSizeOf: points.
	(interpreterProxy isWords: points) ifTrue:[
		pointsIsArray := false.
		"Either PointArray or ShortPointArray"
		(length = (nSegments * 3) or:[length = (nSegments * 6)])
			ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	] ifFalse:["Must be Array of points"
		(interpreterProxy isArray: points)
			ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
		length = (nSegments * 3)
			ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
		pointsIsArray := true.
	].

	"Now check that we have some hope to have enough free space.
	Do this by assuming nPoints boundaries of maximum size,
	hoping that most of the fills will be colors and many boundaries
	will be line segments"

	(lineWidth = 0 or:[lineFill = 0])
		ifTrue:[segSize := GLBaseSize]
		ifFalse:[segSize := GLWideSize].
	(self needAvailableSpace: segSize * nSegments)
		ifFalse:[^interpreterProxy primitiveFailFor: GEFWorkTooBig].

	"Check the fills"
	((self isFillOkay: lineFill) and:[self isFillOkay: fillIndex])
		ifFalse:[^interpreterProxy primitiveFailFor: GEFWrongFill]. 

	"Transform colors"
	lineFill := self transformColor: lineFill.
	fillIndex := self transformColor: fillIndex.
	engineStopped ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineStopped].

	"Check if have anything at all to do"
	((lineFill = 0 or:[lineWidth = 0]) and:[fillIndex = 0])
		ifTrue:[^interpreterProxy pop: 5].

	"Transform the lineWidth"
	lineWidth = 0 ifFalse:[
		lineWidth := self transformWidth: lineWidth.
		lineWidth < 1 ifTrue:[lineWidth := 1]].

	"And load the actual shape"
	pointsIsArray ifTrue:[
		self loadArrayShape: points nSegments: nSegments
			fill: fillIndex lineWidth: lineWidth lineFill: lineFill.
	] ifFalse:[
		self loadShape: (interpreterProxy firstIndexableField: points) nSegments: nSegments
			fill: fillIndex lineWidth: lineWidth lineFill: lineFill 
			pointsShort: (nSegments * 3 = length)].

	engineStopped ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineStopped].

	interpreterProxy failed
		ifTrue:[^interpreterProxy primitiveFailFor: GEFEntityLoadFailed].
	self needsFlushPut: 1.
	self storeEngineStateInto: engine.
	interpreterProxy pop: 5. "Leave rcvr on stack"! !

!BalloonEnginePlugin methodsFor: 'primitives' stamp: 'eem 6/3/2009 17:11'!
primitiveAddBitmapFill

	| failureCode nrmOop dirOop originOop tileFlag fill xIndex cmOop formOop |
	<export: true>
	<inline: false>

	"Fail if we have the wrong number of arguments"
	interpreterProxy methodArgumentCount = 7 
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	xIndex := interpreterProxy stackIntegerValue: 0.
	xIndex <= 0 ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	nrmOop := interpreterProxy stackObjectValue: 1.
	dirOop := interpreterProxy stackObjectValue: 2.
	originOop := interpreterProxy stackObjectValue: 3.
	tileFlag := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 4).
	cmOop := interpreterProxy stackObjectValue: 5.
	formOop := interpreterProxy stackObjectValue: 6.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 7)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	self loadPoint: self point1Get from: originOop.
	self loadPoint: self point2Get from: dirOop.
	self loadPoint: self point3Get from: nrmOop.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: GEFBadPoint].

	fill := self loadBitmapFill: formOop 
				colormap: cmOop
				tile: (tileFlag ifTrue:[1] ifFalse:[0])
				from: self point1Get 
				along: self point2Get 
				normal: self point3Get 
				xIndex: xIndex-1.
	engineStopped ifTrue:[
		"Make sure the stack is okay"
		^interpreterProxy primitiveFailFor: GEFEngineStopped].

	interpreterProxy failed
		ifTrue:[^interpreterProxy primitiveFailFor: GEFEntityLoadFailed].
	self storeEngineStateInto: engine.
	interpreterProxy pop: 8 thenPush: (interpreterProxy positive32BitIntegerFor: fill)! !

!BalloonEnginePlugin methodsFor: 'primitives' stamp: 'eem 6/3/2009 17:11'!
primitiveAddCompressedShape
	| failureCode fillIndexList lineFills lineWidths rightFills leftFills nSegments points pointsShort |
	<export: true>
	<inline: false>

	"Fail if we have the wrong number of arguments"
	interpreterProxy methodArgumentCount = 7 
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	fillIndexList := interpreterProxy stackObjectValue: 0.
	lineFills := interpreterProxy stackObjectValue: 1.
	lineWidths := interpreterProxy stackObjectValue: 2.
	rightFills := interpreterProxy stackObjectValue: 3.
	leftFills := interpreterProxy stackObjectValue: 4.
	nSegments := interpreterProxy stackIntegerValue: 5.
	points := interpreterProxy stackObjectValue: 6.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 7)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	"First, do a check if the compressed shape is okay"
	(self checkCompressedShape: points 
			segments: nSegments 
			leftFills: leftFills 
			rightFills: rightFills 
			lineWidths: lineWidths 
			lineFills: lineFills 
			fillIndexList: fillIndexList) ifFalse:[^interpreterProxy primitiveFailFor: GEFEntityCheckFailed].

	"Now check that we have some hope to have enough free space.
	Do this by assuming nSegments boundaries of maximum size,
	hoping that most of the fills will be colors and many boundaries
	will be line segments"

	(self needAvailableSpace: (GBBaseSize max: GLBaseSize) * nSegments)
		ifFalse:[^interpreterProxy primitiveFailFor: GEFWorkTooBig].

	"Check if the points are short"
	pointsShort := (interpreterProxy slotSizeOf: points) = (nSegments * 3).

	"Then actually load the compressed shape"
	self loadCompressedShape: (interpreterProxy firstIndexableField: points)
			segments: nSegments 
			leftFills: (interpreterProxy firstIndexableField: leftFills)
			rightFills: (interpreterProxy firstIndexableField: rightFills)
			lineWidths: (interpreterProxy firstIndexableField: lineWidths)
			lineFills: (interpreterProxy firstIndexableField: lineFills)
			fillIndexList: (interpreterProxy firstIndexableField: fillIndexList)
			pointShort: pointsShort.

	engineStopped ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineStopped].

	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: GEFEntityLoadFailed].
	self needsFlushPut: 1.
	self storeEngineStateInto: engine.
	interpreterProxy pop: 7. "Leave rcvr on stack"! !

!BalloonEnginePlugin methodsFor: 'primitives' stamp: 'eem 6/3/2009 17:09'!
primitiveAddGradientFill

	| failureCode isRadial nrmOop dirOop originOop rampOop fill |
	<export: true>
	<inline: false>

	"Fail if we have the wrong number of arguments"
	interpreterProxy methodArgumentCount = 5 
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	isRadial := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
	nrmOop := interpreterProxy stackValue: 1.
	dirOop := interpreterProxy stackValue: 2.
	originOop := interpreterProxy stackValue: 3.
	rampOop := interpreterProxy stackValue: 4.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 5)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	self loadPoint: self point1Get from: originOop.
	self loadPoint: self point2Get from: dirOop.
	self loadPoint: self point3Get from: nrmOop.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: GEFBadPoint].

	fill := self loadGradientFill: rampOop 
				from: self point1Get 
				along: self point2Get 
				normal: self point3Get 
				isRadial: isRadial.
	engineStopped ifTrue:[
		"Make sure the stack is okay"
		^interpreterProxy primitiveFailFor: GEFEngineStopped].

	interpreterProxy failed
		ifTrue: [^interpreterProxy primitiveFailFor: GEFEntityLoadFailed].
	self storeEngineStateInto: engine.
	interpreterProxy pop: 6 thenPush: (interpreterProxy positive32BitIntegerFor: fill)! !

!BalloonEnginePlugin methodsFor: 'primitives' stamp: 'eem 6/3/2009 17:09'!
primitiveAddLine
	| failureCode leftFill rightFill endOop startOop |
	<export: true>
	<inline: false>

	"Fail if we have the wrong number of arguments"
	interpreterProxy methodArgumentCount = 4 
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	rightFill := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	leftFill := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 1).
	endOop := interpreterProxy stackObjectValue: 2.
	startOop := interpreterProxy stackObjectValue: 3.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 4)
			requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	"Make sure the fills are okay"
	((self isFillOkay: leftFill) and:[self isFillOkay: rightFill])
			ifFalse:[^interpreterProxy primitiveFailFor: GEFWrongFill].

	"Load the points"
	self loadPoint: self point1Get from: startOop.
	self loadPoint: self point2Get from: endOop.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: GEFBadPoint].

	"Transform points"
	self transformPoints: 2.

	"Transform colors"
	leftFill := self transformColor: leftFill.
	rightFill := self transformColor: rightFill.
	engineStopped ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineStopped].

	"Load line"
	self loadWideLine: 0 from: self point1Get to: self point2Get 
		lineFill: 0 leftFill: leftFill rightFill: rightFill.
	engineStopped ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineStopped].

	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: GEFEntityLoadFailed].
	self storeEngineStateInto: engine.
	interpreterProxy pop: 4. "Leave rcvr on stack"! !

!BalloonEnginePlugin methodsFor: 'primitives' stamp: 'eem 6/3/2009 17:09'!
primitiveAddOval
	| failureCode fillIndex borderWidth borderIndex endOop startOop |
	<export: true>
	<inline: false>

	"Fail if we have the wrong number of arguments"
	interpreterProxy methodArgumentCount = 5 
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	borderIndex := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	borderWidth := interpreterProxy stackIntegerValue: 1.
	fillIndex := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 2).
	endOop := interpreterProxy stackObjectValue: 3.
	startOop := interpreterProxy stackObjectValue: 4.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 5)
					requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	"Make sure the fills are okay"
	((self isFillOkay: borderIndex) and:[self isFillOkay: fillIndex])
			ifFalse:[^interpreterProxy primitiveFailFor: GEFWrongFill].

	"Transform colors"
	fillIndex := self transformColor: fillIndex.
	borderIndex := self transformColor: borderIndex.
	engineStopped ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineStopped].

	"Check if we have anything at all to do"
	(fillIndex = 0 and:[borderIndex = 0 or:[borderWidth <= 0]]) ifTrue:[
		^interpreterProxy pop: 5. "Leave rcvr on stack"
	].

	"Make sure we have some space"
	(self needAvailableSpace: (16 * GBBaseSize)) 
		ifFalse:[^interpreterProxy primitiveFailFor: GEFWorkTooBig].

	"Check if we need a border"
	(borderWidth > 0 and:[borderIndex ~= 0]) 
		ifTrue:[borderWidth := self transformWidth: borderWidth]
		ifFalse:[borderWidth := 0].


	"Load the rectangle points"
	self loadPoint: self point1Get from: startOop.
	self loadPoint: self point2Get from: endOop.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: GEFBadPoint].

	self loadOval: borderWidth lineFill: borderIndex 
		leftFill: 0 rightFill: fillIndex.

	engineStopped ifTrue:[
		self wbStackClear.
		^interpreterProxy primitiveFailFor: GEFEngineStopped.
	].
	interpreterProxy failed
		ifTrue:[^interpreterProxy primitiveFailFor: GEFEntityLoadFailed].
	self needsFlushPut: 1.
	self storeEngineStateInto: engine.
	interpreterProxy pop: 5. "Leave rcvr on stack"! !

!BalloonEnginePlugin methodsFor: 'primitives' stamp: 'eem 12/6/2012 12:37'!
primitiveAddPolygon
	| failureCode points lineFill lineWidth fillIndex nPoints length pointsIsArray segSize |
	<export: true>
	<inline: false>

	"Fail if we have the wrong number of arguments"
	interpreterProxy methodArgumentCount = 5 
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	lineFill := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	lineWidth := interpreterProxy stackIntegerValue: 1.
	fillIndex := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 2).
	nPoints := interpreterProxy stackIntegerValue: 3.
	points := interpreterProxy stackObjectValue: 4.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 5)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	"First, do a check if the points look okay"
	length := interpreterProxy slotSizeOf: points.
	(interpreterProxy isWords: points) ifTrue:[
		pointsIsArray := false.
		"Either PointArray or ShortPointArray"
		(length = nPoints or:[nPoints * 2 = length])
			ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	] ifFalse:["Must be Array of points"
		(interpreterProxy isArray: points)
			ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
		length = nPoints
			ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
		pointsIsArray := true.
	].

	"Now check that we have some hope to have enough free space.
	Do this by assuming nPoints boundaries of maximum size,
	hoping that most of the fills will be colors and many boundaries
	will be line segments"

	(lineWidth = 0 or:[lineFill = 0])
		ifTrue:[segSize := GLBaseSize]
		ifFalse:[segSize := GLWideSize].
	(self needAvailableSpace: segSize * nPoints)
		ifFalse:[^interpreterProxy primitiveFail].

	"Check the fills"
	((self isFillOkay: lineFill) and:[self isFillOkay: fillIndex])
		ifFalse:[^interpreterProxy primitiveFailFor: GEFWrongFill]. 

	"Transform colors"
	lineFill := self transformColor: lineFill.
	fillIndex := self transformColor: fillIndex.
	engineStopped ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineStopped].

	"Check if have anything at all to do"
	((lineFill = 0 or:[lineWidth = 0]) and:[fillIndex = 0])
		ifTrue:[^interpreterProxy pop: 5].

	"Transform the lineWidth"
	lineWidth = 0 ifFalse:[lineWidth := self transformWidth: lineWidth].

	"And load the actual polygon"
	pointsIsArray ifTrue:[
		self loadArrayPolygon: points nPoints: nPoints
			fill: fillIndex lineWidth: lineWidth lineFill: lineFill
	] ifFalse:[
		self loadPolygon: (interpreterProxy firstIndexableField: points) nPoints: nPoints 
			fill: fillIndex lineWidth: lineWidth lineFill: lineFill 
			pointsShort: (nPoints = length)].

	engineStopped ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineStopped].

	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: GEFEntityLoadFailed].
	self needsFlushPut: 1.
	self storeEngineStateInto: engine.
	interpreterProxy pop: 5. "Leave rcvr on stack"! !

!BalloonEnginePlugin methodsFor: 'primitives' stamp: 'eem 6/4/2009 18:15'!
primitiveAddRect
	| failureCode fillIndex borderWidth borderIndex endOop startOop |
	<export: true>
	<inline: false>

	"Fail if we have the wrong number of arguments"
	interpreterProxy methodArgumentCount = 5 
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	borderIndex := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	borderWidth := interpreterProxy stackIntegerValue: 1.
	fillIndex := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 2).
	endOop := interpreterProxy stackObjectValue: 3.
	startOop := interpreterProxy stackObjectValue: 4.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 5)
						requiredState: GEStateUnlocked) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	"Make sure the fills are okay"
	((self isFillOkay: borderIndex) and:[self isFillOkay: fillIndex])
		ifFalse:[^interpreterProxy primitiveFailFor: GEFWrongFill].

	"Transform colors"
	borderIndex := self transformColor: borderIndex.
	fillIndex := self transformColor: fillIndex.
	engineStopped ifTrue:[^interpreterProxy primitiveFailFor: GEFEngineStopped].

	"Check if we have anything at all to do"
	(fillIndex = 0 and:[borderIndex = 0 or:[borderWidth = 0]]) ifTrue:[
		^interpreterProxy pop: 5. "Leave rcvr on stack"
	].

	"Make sure we have some space"
	(self needAvailableSpace: (4 * GLBaseSize)) 
		ifFalse:[^interpreterProxy primitiveFailFor: GEFWorkTooBig].

	"Check if we need a border"
	(borderWidth > 0 and:[borderIndex ~= 0]) 
		ifTrue:[borderWidth := self transformWidth: borderWidth]
		ifFalse:[borderWidth := 0].

	"Load the rectangle"
	self loadPoint: self point1Get from: startOop.
	self loadPoint: self point3Get from: endOop.
	interpreterProxy failed ifTrue:[^interpreterProxy primitiveFailFor: GEFBadPoint].
	self point2Get at: 0 put: (self point3Get at: 0).
	self point2Get at: 1 put: (self point1Get at: 1).
	self point4Get at: 0 put: (self point1Get at: 0).
	self point4Get at: 1 put: (self point3Get at: 1).
	"Transform the points"
	self transformPoints: 4.

	self loadRectangle: borderWidth lineFill: borderIndex leftFill: 0 rightFill: fillIndex.

	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: GEFEntityLoadFailed].
	self needsFlushPut: 1.
	self storeEngineStateInto: engine.
	interpreterProxy pop: 5. "Leave rcvr on stack"! !

!BalloonEnginePlugin methodsFor: 'primitives' stamp: 'eem 6/3/2009 16:50'!
primitiveGetBezierStats
	| failureCode statOop stats |
	<export: true>
	<inline: false>
	<var: #stats type:'int *'>

	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	(failureCode := self quickLoadEngineFrom: (interpreterProxy stackValue: 1)) = 0
		ifFalse:[^interpreterProxy primitiveFailFor: failureCode].

	statOop := interpreterProxy stackObjectValue: 0.
	(interpreterProxy failed not
	and: [(interpreterProxy isWords: statOop)
	and: [(interpreterProxy slotSizeOf: statOop) >= 4]])
		ifFalse:[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	stats := interpreterProxy firstIndexableField: statOop.
	stats at: 0 put: (stats at: 0) + (workBuffer at: GWBezierMonotonSubdivisions).
	stats at: 1 put: (stats at: 1) + (workBuffer at: GWBezierHeightSubdivisions).
	stats at: 2 put: (stats at: 2) + (workBuffer at: GWBezierOverflowSubdivisions).
	stats at: 3 put: (stats at: 3) + (workBuffer at: GWBezierLineConversions).

	interpreterProxy pop: 1. "Leave rcvr on stack"! !

!BalloonEnginePlugin methodsFor: 'private' stamp: 'ar (auto pragmas 12/08) 11/25/1998 19:27'!
rShiftTable
	| theTable |
	<returnTypeC:'int *'>
	<inline: false>
	<var:#theTable declareC:'static int theTable[17] =
		{0, 5, 4, 0, 3, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 1}'>
	^theTable! !

!BalloonEnginePlugin methodsFor: 'fills-bitmaps' stamp: 'ar (auto pragmas 12/08) 11/27/1998 14:14'!
repeatValue: delta max: maxValue
	| newDelta |
	<inline: true>
	newDelta := delta.
	[newDelta < 0] whileTrue:[newDelta := newDelta + maxValue].
	[newDelta >= maxValue] whileTrue:[newDelta := newDelta - maxValue].
	^newDelta! !

!BalloonEnginePlugin methodsFor: 'beziers-wide' stamp: 'ar 11/6/1998 01:54'!
returnWideBezierFill
	^(dispatchReturnValue := self wideBezierFillOf: dispatchedValue).! !

!BalloonEnginePlugin methodsFor: 'beziers-wide' stamp: 'ar 11/6/1998 01:54'!
returnWideBezierWidth
	^(dispatchReturnValue := self wideBezierWidthOf: dispatchedValue).! !

!BalloonEnginePlugin methodsFor: 'lines-wide' stamp: 'ar 11/6/1998 17:08'!
returnWideLineFill
	"Return the fill of the (wide) line - this method is called from a case."
	^(dispatchReturnValue := self wideLineFillOf: dispatchedValue).! !

!BalloonEnginePlugin methodsFor: 'lines-wide' stamp: 'ar 11/6/1998 17:08'!
returnWideLineWidth
	"Return the width of the (wide) line - this method is called from a case."
	^(dispatchReturnValue := self wideLineWidthOf: dispatchedValue).! !

!BalloonEnginePlugin methodsFor: 'private' stamp: 'eem 5/20/2010 15:05'!
shortRunLengthAt: i from: runArray
	"Return the run-length value from the given ShortRunArray."
	^((self cCoerce: runArray to:'int *') at: i) bitShift: -16! !

!BalloonEnginePlugin methodsFor: 'private' stamp: 'ar 11/3/1998 22:54'!
shortRunValueAt: i from: runArray
	"Return the run-length value from the given ShortRunArray.
	Note: We don't need any coercion to short/int here, since
	we deal basically only with unsigned values."
	^(((self cCoerce: runArray to:'int *') at: i) bitAnd: 16rFFFF)! !

!BalloonEnginePlugin methodsFor: 'beziers-simple' stamp: 'ar (auto pragmas 12/08) 11/6/1998 00:07'!
stepToFirstBezier
	"Initialize the current entry in the GET by stepping to the current scan line"
	<inline: true>
	^self stepToFirstBezierIn: (getBuffer at: self getStartGet) at: self currentYGet! !

!BalloonEnginePlugin methodsFor: 'beziers-simple' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:52'!
stepToFirstBezierIn: bezier at: yValue
	"Initialize the bezier at yValue.
	TODO: Check if reducing maxSteps from 2*deltaY to deltaY 
		brings a *significant* performance improvement.
		In theory this should make for double step performance
		but will cost in quality. Might be that the AA stuff will
		compensate for this - but I'm not really sure."
	| updateData deltaY maxSteps scaledStepSize squaredStepSize 
	startX startY viaX viaY endX endY 
	fwX1 fwX2 fwY1 fwY2 
	fwDx fwDDx fwDy fwDDy |
	<inline: false>
	<var: #updateData type:'int *'>


	"Do a quick check if there is anything at all to do"
	((self isWide: bezier) not and:[yValue >= (self bezierEndYOf: bezier)])
		ifTrue:[^self edgeNumLinesOf: bezier put: 0].

	"Now really initialize bezier"
	startX := self edgeXValueOf: bezier.
	startY := self edgeYValueOf: bezier.
	viaX := self bezierViaXOf: bezier.
	viaY := self bezierViaYOf: bezier.
	endX := self bezierEndXOf: bezier.
	endY := self bezierEndYOf: bezier.
	deltaY := endY - startY.

	"Initialize integer forward differencing"
	fwX1 := (viaX - startX) * 2.
	fwX2 := startX + endX - (viaX * 2).
	fwY1 := (viaY - startY) * 2.
	fwY2 := startY + endY - (viaY * 2).
	maxSteps := deltaY * 2.
	maxSteps < 2 ifTrue:[maxSteps := 2].
	scaledStepSize := 16r1000000 // maxSteps.
	squaredStepSize := self absoluteSquared8Dot24: scaledStepSize.
	fwDx := fwX1 * scaledStepSize.
	fwDDx := fwX2 * squaredStepSize * 2.
	fwDx := fwDx + (fwDDx // 2).
	fwDy := fwY1 * scaledStepSize.
	fwDDy := fwY2 * squaredStepSize * 2.
	fwDy := fwDy + (fwDDy // 2).

	"Store the values"
	self edgeNumLinesOf: bezier put: deltaY.

	updateData := self bezierUpdateDataOf: bezier.
	updateData at: GBUpdateX put: (startX * 256).
	updateData at: GBUpdateY put: (startY * 256).
	updateData at: GBUpdateDX put: fwDx.
	updateData at: GBUpdateDY put: fwDy.
	updateData at: GBUpdateDDX put: fwDDx.
	updateData at: GBUpdateDDY put: fwDDy.

	"And step to the first scan line"
	(startY := self edgeYValueOf: bezier) = yValue ifFalse:[
		self stepToNextBezierIn: bezier at: yValue.
		"Adjust number of lines remaining"
		self edgeNumLinesOf: bezier put: deltaY - (yValue - startY).
	].! !

!BalloonEnginePlugin methodsFor: 'lines-simple' stamp: 'ar (auto pragmas 12/08) 11/4/1998 21:52'!
stepToFirstLine
	"Initialize the current entry in the GET by stepping to the current scan line"
	<inline: true>
	^self stepToFirstLineIn: (getBuffer at: self getStartGet) at: self currentYGet! !

!BalloonEnginePlugin methodsFor: 'lines-simple' stamp: 'ar (auto pragmas 12/08) 11/9/1998 15:38'!
stepToFirstLineIn: line at: yValue
	"Initialize the line at yValue"
	| deltaX deltaY xDir widthX error xInc errorAdjUp startY |
	<inline: false>

	"Do a quick check if there is anything at all to do"
	((self isWide: line) not and:[yValue >= (self lineEndYOf: line)])
		ifTrue:[^self edgeNumLinesOf: line put: 0].

	deltaX := (self lineEndXOf: line) - (self edgeXValueOf: line).
	deltaY := (self lineEndYOf: line) - (self edgeYValueOf: line).

	"Check if edge goes left to right"
	deltaX >= 0 
		ifTrue:[	xDir := 1.
				widthX := deltaX.
				error := 0]
		ifFalse:[	xDir := -1.
				widthX := 0 - deltaX.
				error := 1 - deltaY].

	"Check if deltaY is zero.
	Note: We could actually get out here immediately 
	but wide lines rely on an accurate setup in this case"
	deltaY = 0
		ifTrue:[	error := 0.			"No error for horizontal edges"
				xInc := deltaX.		"Encodes width and direction"
				errorAdjUp := 0]
		ifFalse:["Check if edge is y-major"
				deltaY > widthX "Note: The '>' instead of '>=' could be important here..."
					ifTrue:[	xInc := 0.
							errorAdjUp := widthX]
					ifFalse:[	xInc := (widthX // deltaY) * xDir.
							errorAdjUp := widthX \\ deltaY]].

	"Store the values"
	self edgeNumLinesOf: line put: deltaY.
	self lineXDirectionOf: line put: xDir.
	"self lineYDirectionOf: line put: yDir." "<-- Already set"
	self lineXIncrementOf: line put: xInc.
	self lineErrorOf: line put: error.
	self lineErrorAdjUpOf: line put: errorAdjUp.
	self lineErrorAdjDownOf: line put: deltaY.

	"And step to the first scan line"
	(startY := self edgeYValueOf: line) = yValue ifFalse:[
		startY to: yValue-1 do:[:i| self stepToNextLineIn: line at: i].
		"Adjust number of lines remaining"
		self edgeNumLinesOf: line put: deltaY - (yValue - startY).
	].! !

!BalloonEnginePlugin methodsFor: 'beziers-wide' stamp: 'ar (auto pragmas 12/08) 11/6/1998 02:00'!
stepToFirstWideBezier
	"Initialize the current entry in the GET by stepping to the current scan line"
	<inline: true>
	^self stepToFirstWideBezierIn: (getBuffer at: self getStartGet) at: self currentYGet! !

!BalloonEnginePlugin methodsFor: 'beziers-wide' stamp: 'nice 11/18/2016 03:45'!
stepToFirstWideBezierIn: bezier at: yValue
	"Initialize the bezier at yValue"	
	| lineWidth startY nLines yEntry yExit lineOffset endX xDir |
	<inline: false>

	"Get some values"
	lineWidth := self wideBezierExtentOf: bezier.
	lineOffset := self offsetFromWidth: lineWidth.

	"Compute the incremental values of the bezier"
	endX := self bezierEndXOf: bezier.
	startY := self edgeYValueOf: bezier.
	self stepToFirstBezierIn: bezier at: startY.
	nLines := (self edgeNumLinesOf: bezier).

	"Copy the incremental update data"
	0 to: 5 do:[:i|
		(self wideBezierUpdateDataOf: bezier) at: i put:
			((self bezierUpdateDataOf: bezier) at: i).
	].

	"Compute primary x direction of curve (e.g., 1: left to right; -1: right to left)."
	xDir := ((self bezierUpdateDataOf: bezier) at: GBUpdateDX).
	xDir = 0 ifTrue:[xDir := ((self bezierUpdateDataOf: bezier) at: GBUpdateDDX)].
	xDir >= 0 ifTrue:[xDir := 1] ifFalse:[xDir := -1].

	"Adjust the curve to start/end at the right position"
	xDir < 0
		ifTrue:[self adjustWideBezierLeft: bezier width: lineWidth offset: lineOffset endX: endX]
		ifFalse:[self adjustWideBezierRight: bezier width: lineWidth offset: lineOffset endX: endX].

	"Adjust the last value for horizontal lines"
	nLines = 0 ifTrue:[(self bezierUpdateDataOf: bezier) at: GBUpdateX put: 
						(self bezierFinalXOf: bezier) * 256].
	"Adjust the number of lines to include the lineWidth"
	self edgeNumLinesOf: bezier put: nLines + lineWidth.

	"Compute the points where we have to turn on/off the fills"
	yEntry := 0.						"turned on at lineOffset"
	yExit := 0 - nLines - lineOffset.	"turned off at zero"
	self wideBezierEntryOf: bezier put: yEntry.
	self wideBezierExitOf: bezier put: yExit.

	"Turn the fills on/off as necessary"
	(yEntry >= lineOffset and:[yExit < 0])
		ifTrue:[self edgeFillsValidate: bezier]
		ifFalse:[self edgeFillsInvalidate: bezier].

	self computeFinalWideBezierValues: bezier width: lineWidth.

	"And step to the first scan line"
	startY = yValue ifFalse:[
		"Note: Must single step here so that entry/exit works"
		startY to: yValue-1 do:[:i| self stepToNextWideBezierIn: bezier at: i].
		"Adjust number of lines remaining"
		self edgeNumLinesOf: bezier put: (self edgeNumLinesOf: bezier) - (yValue - startY).
	].! !

!BalloonEnginePlugin methodsFor: 'lines-wide' stamp: 'ar (auto pragmas 12/08) 11/4/1998 21:54'!
stepToFirstWideLine
	"Initialize the current entry in the GET by stepping to the current scan line"
	<inline: true>
	^self stepToFirstWideLineIn: (getBuffer at: self getStartGet) at: self currentYGet! !

!BalloonEnginePlugin methodsFor: 'lines-wide' stamp: 'ar (auto pragmas 12/08) 11/9/1998 15:38'!
stepToFirstWideLineIn: line at: yValue
	"Initialize the wide line at yValue."
	| startY yEntry yExit lineWidth nLines lineOffset startX xDir |
	<inline: false>

	"Get some values"
	lineWidth := self wideLineExtentOf: line.
	lineOffset := self offsetFromWidth: lineWidth.

	"Compute the incremental values of the line"
	startX := self edgeXValueOf: line.
	startY := self edgeYValueOf: line.
	self stepToFirstLineIn: line at: startY.
	nLines := (self edgeNumLinesOf: line).
	xDir := self lineXDirectionOf: line.

	"Adjust the line to start at the correct X position"
	self edgeXValueOf: line put: startX - lineOffset.

	"Adjust the number of lines to include the lineWidth"
	self edgeNumLinesOf: line put: nLines + lineWidth.

	"Adjust the values for x-major lines"
	xDir > 0 ifTrue:[
		self wideLineWidthOf: line put: (self lineXIncrementOf: line) + lineWidth.
	] ifFalse:[
		self wideLineWidthOf: line put: lineWidth - (self lineXIncrementOf: line). "adding"
		self edgeXValueOf: line put: (self edgeXValueOf: line) + (self lineXIncrementOf: line).
	].

	"Compute the points where we have to turn on/off the fills"
	yEntry := 0.						"turned on at lineOffset"
	yExit := 0 - nLines - lineOffset.	"turned off at zero"
	self wideLineEntryOf: line put: yEntry.
	self wideLineExitOf: line put: yExit.

	"Turn the fills on/off as necessary"
	(yEntry >= lineOffset and:[yExit < 0])
		ifTrue:[self edgeFillsValidate: line]
		ifFalse:[self edgeFillsInvalidate: line].

	"And step to the first scan line"
	startY = yValue ifFalse:[
		startY to: yValue-1 do:[:i| self stepToNextWideLineIn: line at: i].
		"Adjust number of lines remaining"
		self edgeNumLinesOf: line put: (self edgeNumLinesOf: line) - (yValue - startY).
	].
! !

!BalloonEnginePlugin methodsFor: 'beziers-simple' stamp: 'ar (auto pragmas 12/08) 11/6/1998 00:08'!
stepToNextBezier
	"Process the current entry in the AET by stepping to the next scan line"
	<inline: true>
	^self stepToNextBezierIn: (aetBuffer at: self aetStartGet) at: self currentYGet! !

!BalloonEnginePlugin methodsFor: 'beziers-simple' stamp: 'eem 12/12/2014 10:16'!
stepToNextBezierForward: updateData at: yValue
	"Incrementally step to the next scan line in the given bezier update data.
	Note: This method has been written so that inlining works, e.g.,
		not declaring updateData as 'int*' but casting it on every use."
	<var: #updateData type:'int *'>
	| minY lastX lastY fwDx fwDy |
	<inline: true>
	lastX := updateData at: GBUpdateX.
	lastY := updateData at: GBUpdateY.
	fwDx := updateData at: GBUpdateDX.
	fwDy := updateData at: GBUpdateDY.
	minY := yValue * 256.
	"Step as long as we haven't yet reached minY and also
	as long as fwDy is greater than zero thus stepping down.
	Note: The test for fwDy should not be necessary in theory
		but is a good insurance in practice."
	[minY > lastY and:[fwDy >= 0]] whileTrue:[
		lastX := lastX + ((fwDx + 16r8000) signedBitShift: -16).
		lastY := lastY + ((fwDy + 16r8000) signedBitShift: -16).
		fwDx := fwDx + (updateData at: GBUpdateDDX).
		fwDy := fwDy + (updateData at: GBUpdateDDY).
	].
	updateData at: GBUpdateX put: lastX.
	updateData at: GBUpdateY put: lastY.
	updateData at: GBUpdateDX put: fwDx.
	updateData at: GBUpdateDY put: fwDy.
	^lastX signedBitShift: -8
! !

!BalloonEnginePlugin methodsFor: 'beziers-simple' stamp: 'ar (auto pragmas 12/08) 11/9/1998 15:39'!
stepToNextBezierIn: bezier at: yValue
	"Incrementally step to the next scan line in the given bezier"
	|  xValue |
	<inline: true>
	xValue := self stepToNextBezierForward: (self bezierUpdateDataOf: bezier) at: yValue.
	self edgeXValueOf: bezier put: xValue.! !

!BalloonEnginePlugin methodsFor: 'lines-simple' stamp: 'ar (auto pragmas 12/08) 11/4/1998 21:53'!
stepToNextLine
	"Process the current entry in the AET by stepping to the next scan line"
	<inline: true>
	^self stepToNextLineIn: (aetBuffer at: self aetStartGet) at: self currentYGet! !

!BalloonEnginePlugin methodsFor: 'lines-simple' stamp: 'ar (auto pragmas 12/08) 11/9/1998 15:39'!
stepToNextLineIn: line at: yValue
	"Incrementally step to the next scan line in the given line"
	| x  err |
	<inline: true>
	x := (self edgeXValueOf: line) + (self lineXIncrementOf: line).
	err := (self lineErrorOf: line) + (self lineErrorAdjUpOf: line).
	err > 0 ifTrue:[
		x := x + (self lineXDirectionOf: line).
		err := err - (self lineErrorAdjDownOf: line).
	].
	self lineErrorOf: line put: err.
	self edgeXValueOf: line put: x.! !

!BalloonEnginePlugin methodsFor: 'beziers-wide' stamp: 'ar (auto pragmas 12/08) 11/6/1998 02:34'!
stepToNextWideBezier
	"Initialize the current entry in the GET by stepping to the current scan line"
	<inline: true>
	self stepToNextWideBezierIn: (aetBuffer at: self aetStartGet) at: self currentYGet.! !

!BalloonEnginePlugin methodsFor: 'beziers-wide' stamp: 'ar (auto pragmas 12/08) 11/9/1998 15:39'!
stepToNextWideBezierIn: bezier at: yValue
	"Incrementally step to the next scan line in the given wide bezier"
	|  yEntry yExit lineWidth lineOffset |
	<inline: false>	"Don't inline this"

	lineWidth := self wideBezierExtentOf: bezier.
	lineOffset := self offsetFromWidth: lineWidth.

	yEntry := (self wideBezierEntryOf: bezier) + 1.
	yExit := (self wideBezierExitOf: bezier) + 1.
	self wideBezierEntryOf: bezier put: yEntry.
	self wideBezierExitOf: bezier put: yExit.
	yEntry >= lineOffset ifTrue:[self edgeFillsValidate: bezier].
	yExit >= 0 ifTrue:[self edgeFillsInvalidate: bezier].

	"Check if we have to step the upper curve"
	(yExit + lineOffset < 0) ifTrue:[
		self stepToNextBezierForward: (self bezierUpdateDataOf: bezier) at: yValue.
	] ifFalse:[
		"Adjust the last x value to the final x recorded previously"
		(self bezierUpdateDataOf: bezier) at: GBUpdateX put: (self bezierFinalXOf: bezier) * 256.
	].
	"Step the lower curve"
	self stepToNextBezierForward: (self wideBezierUpdateDataOf: bezier) at: yValue.

	self computeFinalWideBezierValues: bezier width: lineWidth.! !

!BalloonEnginePlugin methodsFor: 'lines-wide' stamp: 'ar (auto pragmas 12/08) 11/4/1998 21:55'!
stepToNextWideLine
	"Process the current entry in the AET by stepping to the next scan line"
	<inline: true>
	^self stepToNextWideLineIn: (aetBuffer at: self aetStartGet) at: self currentYGet! !

!BalloonEnginePlugin methodsFor: 'lines-wide' stamp: 'ar (auto pragmas 12/08) 11/9/1998 15:39'!
stepToNextWideLineIn: line at: yValue
	"Incrementally step to the next scan line in the given wide line"
	|  yEntry yExit lineWidth lineOffset lastX nextX |
	<inline: true>

	"Adjust entry/exit values"
	yEntry := (self wideLineEntryOf: line) + 1.
	yExit := (self wideLineExitOf: line) + 1.
	self wideLineEntryOf: line put: yEntry.
	self wideLineExitOf: line put: yExit.

	"Turn fills on/off"
	lineWidth := self wideLineExtentOf: line.
	lineOffset := self offsetFromWidth: lineWidth.
	yEntry >= lineOffset ifTrue:[self edgeFillsValidate: line].
	yExit >= 0 ifTrue:[self edgeFillsInvalidate: line].

	"Step to the next scan line"
	lastX := self edgeXValueOf: line.
	self stepToNextLineIn: line at: yValue.
	nextX := self edgeXValueOf: line.

	"Check for special start/end adjustments"
	(yEntry <= lineWidth or:[yExit+lineOffset >= 0]) ifTrue:[
		"Yes, need an update"
		self adjustWideLine: line afterSteppingFrom: lastX to: nextX.
	].! !

!BalloonEnginePlugin methodsFor: 'bezier-loading' stamp: 'ar (auto pragmas 12/08) 11/8/1998 14:36'!
subdivideBezier: index
	"Subdivide the given bezier curve if necessary"
	| startX startY endX endY deltaX deltaY |
	<inline: false>
	startY := self bzStartY: index.
	endY := self bzEndY: index.

	"If the receiver is horizontal, don't do anything"
	(endY = startY) ifTrue:[^index].

	"TODO: If the curve can be represented as a line, then do so"

	"If the height of the curve exceeds 256 pixels, subdivide 
	(forward differencing is numerically not very stable)"
	deltaY := endY - startY.
	deltaY < 0 ifTrue:[deltaY := 0 - deltaY].
	(deltaY > 255) ifTrue:[
		self incrementStat: GWBezierHeightSubdivisions by: 1.
		^self computeBezierSplitAtHalf: index].

	"Check if the incremental values could possibly overflow the scaled integer range"
	startX := self bzStartX: index.
	endX := self bzEndX: index.
	deltaX := endX - startX.
	deltaX < 0 ifTrue:[deltaX := 0 - deltaX].
	deltaY * 32 < deltaX ifTrue:[
		self incrementStat: GWBezierOverflowSubdivisions by: 1.
		^self computeBezierSplitAtHalf: index].
	^index
! !

!BalloonEnginePlugin methodsFor: 'bezier-loading' stamp: 'ar (auto pragmas 12/08) 11/8/1998 03:43'!
subdivideBezierFrom: index
	"Recursively subdivide the curve on the bezier stack."
	| otherIndex index1 index2 |
	<inline: false>
	otherIndex := self subdivideBezier: index.
	otherIndex = index ifFalse:[
		index1 := self subdivideBezierFrom: index.
		engineStopped ifTrue:[^0].
		index2 := self subdivideBezierFrom: otherIndex.
		engineStopped ifTrue:[^0].
		index1 >= index2
			ifTrue:[^index1]
			ifFalse:[^index2]
	].
	^index! !

!BalloonEnginePlugin methodsFor: 'bezier-loading' stamp: 'ar (auto pragmas 12/08) 11/8/1998 15:17'!
subdivideToBeMonoton: base inX: doTestX
	"Check if the given bezier curve is monoton in Y, and, if desired in X. 
	If not, subdivide it"
	| index1 index2 base2 |
	<inline: false>
	base2 := index1 := index2 := self subdivideToBeMonotonInY: base.
	doTestX ifTrue:[index1 := self subdivideToBeMonotonInX: base].
	index1 > index2 ifTrue:[index2 := index1].
	(base ~= base2 and:[doTestX]) ifTrue:[index1 := self subdivideToBeMonotonInX: base2].
	index1 > index2 ifTrue:[index2 := index1].
	^index2! !

!BalloonEnginePlugin methodsFor: 'bezier-loading' stamp: 'ar (auto pragmas 12/08) 11/7/1998 19:42'!
subdivideToBeMonotonInX: index
	"Check if the given bezier curve is monoton in X. If not, subdivide it"
	| denom num startX viaX endX dx1 dx2 |
	<inline: false>
	startX := self bzStartX: index.
	viaX := self bzViaX: index.
	endX := self bzEndX: index.

	dx1 := viaX - startX.
	dx2 := endX - viaX.
	(dx1 * dx2) >= 0 ifTrue:[^index]. "Bezier is monoton"

	self incrementStat: GWBezierMonotonSubdivisions by: 1.
	"Compute split value"
	denom := dx2 - dx1.
	num := dx1.
	num < 0 ifTrue:[num := 0 - num].
	denom < 0 ifTrue:[denom := 0 - denom].
	^self computeBezier: index splitAt: (num asFloat / denom asFloat).! !

!BalloonEnginePlugin methodsFor: 'bezier-loading' stamp: 'ar (auto pragmas 12/08) 11/7/1998 19:42'!
subdivideToBeMonotonInY: index
	"Check if the given bezier curve is monoton in Y. If not, subdivide it"
	| startY viaY endY dy1 dy2 denom num |
	<inline: false>
	startY := self bzStartY: index.
	viaY := self bzViaY: index.
	endY := self bzEndY: index.

	dy1 := viaY - startY.
	dy2 := endY - viaY.
	(dy1 * dy2) >= 0 ifTrue:[^index]. "Bezier is monoton"

	self incrementStat: GWBezierMonotonSubdivisions by: 1.
	"Compute split value"
	denom := dy2 - dy1.
	num := dy1.
	num < 0 ifTrue:[num := 0 - num].
	denom < 0 ifTrue:[denom := 0 - denom].
	^self computeBezier: index splitAt: (num asFloat / denom asFloat).! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:18'!
wideBezierEntryOf: line

	^self obj: line at: GBWideEntry! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:18'!
wideBezierEntryOf: line put: value

	^self obj: line at: GBWideEntry put: value! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:18'!
wideBezierExitOf: line

	^self obj: line at: GBWideExit! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:18'!
wideBezierExitOf: line put: value

	^self obj: line at: GBWideExit put: value! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:20'!
wideBezierExtentOf: bezier

	^self obj: bezier at: GBWideExtent! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:20'!
wideBezierExtentOf: bezier put: value

	^self obj: bezier at: GBWideExtent put: value! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:19'!
wideBezierFillOf: bezier

	^self obj: bezier at: GBWideFill! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:17'!
wideBezierFillOf: bezier put: value

	^self obj: bezier at: GBWideFill put: value! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar (auto pragmas 12/08) 11/24/1998 22:25'!
wideBezierUpdateDataOf: bezier
	<returnTypeC: 'int *'>

	^objBuffer + bezier +  GBWideUpdateData! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:20'!
wideBezierWidthOf: line

	^self obj: line at: GBWideWidth! !

!BalloonEnginePlugin methodsFor: 'accessing beziers' stamp: 'ar 11/24/1998 22:20'!
wideBezierWidthOf: line put: value

	^self obj: line at: GBWideWidth put: value! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:09'!
wideLineEntryOf: line

	^self obj: line at: GLWideEntry! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:10'!
wideLineEntryOf: line put: value

	^self obj: line at: GLWideEntry put: value! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:10'!
wideLineExitOf: line

	^self obj: line at: GLWideExit! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:10'!
wideLineExitOf: line put: value

	^self obj: line at: GLWideExit put: value! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:10'!
wideLineExtentOf: line

	^self obj: line at: GLWideExtent! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:10'!
wideLineExtentOf: line put: value

	^self obj: line at: GLWideExtent put: value! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:10'!
wideLineFillOf: line

	^self obj: line at: GLWideFill! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:11'!
wideLineFillOf: line put: value

	^self obj: line at: GLWideFill put: value! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:11'!
wideLineWidthOf: line

	^self obj: line at: GLWideWidth! !

!BalloonEnginePlugin methodsFor: 'accessing lines' stamp: 'ar 11/24/1998 22:11'!
wideLineWidthOf: line put: value

	^self obj: line at: GLWideWidth put: value! !

BalloonEnginePlugin subclass: #BalloonEngineSimulation
	instanceVariableNames: 'bbObj savedBBObj workBufferArray'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!BalloonEngineSimulation commentStamp: 'tpr 5/5/2003 11:48' prior: 0!
Support for the VM simulator Balloon graphics calls!

!BalloonEngineSimulation class methodsFor: 'simulation' stamp: 'tfel 3/30/2015 16:10'!
simulatePrimitive: aString receiver: rcvr args: args

	| proxy bb |
	proxy := InterpreterProxy new.
	proxy synthesizeStackFor: rcvr with: args.
	bb := self simulatorClass new.
	bb setInterpreter: proxy.
	bb initialiseModule.
	"rendering state is loaded in the primitive implementations"
	[bb perform: aString asSymbol] on: Exception do: [:ex |
		proxy success: false].
	^ proxy stackValue: 0
! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'ar 11/9/1998 01:23'!
circleCosTable
	^CArrayAccessor on:
#(1.0 0.98078528040323 0.923879532511287 0.831469612302545 0.7071067811865475 0.555570233019602 0.38268343236509 0.1950903220161286 0.0 -0.1950903220161283 -0.3826834323650896 -0.555570233019602 -0.707106781186547 -0.831469612302545 -0.9238795325112865 -0.98078528040323 -1.0 -0.98078528040323 -0.923879532511287 -0.831469612302545 -0.707106781186548 -0.555570233019602 -0.3826834323650903 -0.1950903220161287 0.0 0.1950903220161282 0.38268343236509 0.555570233019602 0.707106781186547 0.831469612302545 0.9238795325112865 0.98078528040323 1.0 )! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'ar 11/9/1998 01:23'!
circleSinTable
	^CArrayAccessor on:
#(0.0 0.1950903220161282 0.3826834323650897 0.555570233019602 0.707106781186547 0.831469612302545 0.923879532511287 0.98078528040323 1.0 0.98078528040323 0.923879532511287 0.831469612302545 0.7071067811865475 0.555570233019602 0.38268343236509 0.1950903220161286 0.0 -0.1950903220161283 -0.3826834323650896 -0.555570233019602 -0.707106781186547 -0.831469612302545 -0.9238795325112865 -0.98078528040323 -1.0 -0.98078528040323 -0.923879532511287 -0.831469612302545 -0.707106781186548 -0.555570233019602 -0.3826834323650903 -0.1950903220161287 0.0 )! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'ar 11/24/1998 20:50'!
colorTransform
	^super colorTransform asPluggableAccessor:
		(Array 
			with:[:obj :index| obj floatAt: index]
			with:[:obj :index :value| obj floatAt: index put: value])! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'eem 9/17/2009 11:40'!
copyBitsFrom: x0 to: x1 at: y
	"Simulate the copyBits primitive"

	| bb |
	bbObj isInteger ifTrue:
		["Create a proxy object to handle BitBlt calls"
		bb := savedBBObj
				ifNil: [BitBltSimulator new
						initialiseModule;
						setInterpreter: interpreterProxy;
						yourself]
				ifNotNil: [savedBBObj].
		(bb loadBitBltFrom: bbObj)
			ifTrue: [bbObj := bb]
			ifFalse: [^ self]].
	bbObj copyBitsFrom: x0 to: x1 at: y.
"
	interpreterProxy showDisplayBits: bbObj destForm
		Left: bb affectedLeft Top: bb affectedTop
		Right: bb affectedRight Bottom: bb affectedBottom.
"! !

!BalloonEngineSimulation methodsFor: 'debug support' stamp: 'ar 5/25/2000 17:58'!
debugDrawBezier: line
	| canvas p1 p2 p3 |
	self assert:(self isBezier: line).
	p1 := (self edgeXValueOf: line) @ (self edgeYValueOf: line) // self aaLevelGet.
	p2 := (self bezierViaXOf: line) @ (self bezierViaYOf: line) // self aaLevelGet.
	p3 := (self bezierEndXOf: line) @ (self bezierEndYOf: line) // self aaLevelGet.
	canvas := Display getCanvas.
	canvas
		line: p1 to: p2 width: 2 color: Color blue;
		line: p2 to: p3 width: 2 color: Color blue.! !

!BalloonEngineSimulation methodsFor: 'debug support' stamp: 'ar 11/1/1998 01:16'!
debugDrawEdge: edge
	self assert: (self isEdge: edge).
	(self isLine: edge) ifTrue:[^self debugDrawLine: edge].
	(self isBezier: edge) ifTrue:[^self debugDrawBezier: edge].
	self halt.! !

!BalloonEngineSimulation methodsFor: 'debug support' stamp: 'ar 5/25/2000 17:58'!
debugDrawHLine: yValue
	| canvas |
	canvas := Display getCanvas.
	canvas
		line: 0 @ (yValue // self aaLevelGet)
		to: Display extent x @ (yValue // self aaLevelGet)
		width: 2
		color: Color green.! !

!BalloonEngineSimulation methodsFor: 'debug support' stamp: 'ar 5/25/2000 17:58'!
debugDrawLine: line
	| canvas |
	self assert: (self isLine: line).
	canvas := Display getCanvas.
	canvas
		line: (self edgeXValueOf: line) @ (self edgeYValueOf: line) // self aaLevelGet
		to: (self lineEndXOf: line) @ (self lineEndYOf: line) // self aaLevelGet
		width: 2
		color: Color red.! !

!BalloonEngineSimulation methodsFor: 'debug support' stamp: 'ar 5/25/2000 17:58'!
debugDrawLineFrom: pt1 to: pt2
	| canvas |
	canvas := Display getCanvas.
	canvas
		line: (pt1 at: 0) @ (pt1 at: 1) // self aaLevelGet
		to: (pt2 at: 0) @ (pt2 at: 1) // self aaLevelGet
		width: 1
		color: Color red.! !

!BalloonEngineSimulation methodsFor: 'debug support' stamp: 'ar 5/25/2000 17:58'!
debugDrawPt: pt
	| canvas |
	canvas := Display getCanvas.
	canvas
		fillRectangle:((pt-2) corner: pt+2) color: Color red! !

!BalloonEngineSimulation methodsFor: 'debug support' stamp: 'ar 5/25/2000 17:58'!
debugDrawPtLineFrom: pt1 to: pt2
	| canvas |
	canvas := Display getCanvas.
	canvas
		line: pt1
		to: pt2
		width: 1
		color: Color red.! !

!BalloonEngineSimulation methodsFor: 'debug support' stamp: 'ar (auto pragmas 12/08) 11/25/1998 00:43'!
debugPrintObjects
	| object end |
	<inline: false>
	object := 0.
	end := objUsed.
	[object < end] whileTrue:[
		Transcript cr; 
			nextPut:$#; print: object; space;
			print: (self objectHeaderOf: object); space.
		(self isEdge: object) 
			ifTrue:[Transcript nextPutAll:'(edge) '].
		(self isFill:object)
			ifTrue:[Transcript nextPutAll:'(fill) '].
		Transcript print: (self objectLengthOf: object); space.
		Transcript endEntry.
		object := object + (self objectLengthOf: object).
	].! !

!BalloonEngineSimulation methodsFor: 'debug support' stamp: 'ar 11/1/1998 17:21'!
debugPrintPoints: n
	Transcript cr.
	n > 0 ifTrue:[
		Transcript print: (self point1Get at: 0) @ (self point1Get at: 1); space.
	].
	n > 1 ifTrue:[
		Transcript print: (self point2Get at: 0) @ (self point2Get at: 1); space.
	].
	n > 2 ifTrue:[
		Transcript print: (self point3Get at: 0) @ (self point3Get at: 1); space.
	].
	n > 3 ifTrue:[
		Transcript print: (self point4Get at: 0) @ (self point4Get at: 1); space.
	].
	Transcript endEntry.! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'ar 10/29/1998 18:44'!
dispatchOn: anInteger in: selectorArray
	"Simulate a case statement via selector table lookup.
	The given integer must be between 0 and selectorArray size-1, inclusive.
	For speed, no range test is done, since it is done by the at: operation."
	self perform: (selectorArray at: (anInteger + 1)).! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'ar 11/24/1998 20:50'!
edgeTransform
	^super edgeTransform asPluggableAccessor:
		(Array 
			with:[:obj :index| obj floatAt: index]
			with:[:obj :index :value| obj floatAt: index put: value])! !

!BalloonEngineSimulation methodsFor: 'simulation support' stamp: 'eem 10/25/2018 17:29'!
ensureUpToDateOverrideFor: aSelector
	"Ensure that an override that uses 32-bit integer arithmetic exists for aSelector"
	| source |
	source := self transformedSourceFor: self class superclass >> aSelector.
	(self class sourceCodeAt: aSelector ifAbsent: ['']) ~= source ifTrue:
		[self class compile: source classified: 'generated overrides']! !

!BalloonEngineSimulation methodsFor: 'simulation support' stamp: 'eem 10/26/2018 19:20'!
ensureUpToDateOverrides
	"Ensure that all methods that require int and unsigned int arithmetic are overridden."
	self selectorsForIntMethods do:
		[:selector| self ensureUpToDateOverrideFor: selector]! !

!BalloonEngineSimulation methodsFor: 'initialize' stamp: 'di 7/12/2004 15:54'!
initialiseModule
	super initialiseModule.
	^ true! !

!BalloonEngineSimulation methodsFor: 'initialize' stamp: 'eem 10/26/2018 15:14'!
initialize
	doProfileStats := false.
	bbPluginName := 'BitBltPlugin'.
	self ensureUpToDateOverrides! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'ar 10/29/1998 19:19'!
ioMicroMSecs
	^Time millisecondClockValue! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'eem 9/17/2009 11:35'!
loadBitBltFrom: oop
	(bbObj isInteger or: [bbObj isNil]) ifFalse:
		[savedBBObj := bbObj].
	bbObj := oop.
	^true! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'ar 11/4/1998 14:05'!
loadPointIntAt: index from: intArray
	"Load the int value from the given index in intArray"
	^(index bitAnd: 1) = 0
		ifTrue:[(intArray getObject at: (index // 2) + 1) x]
		ifFalse:[(intArray getObject at: (index // 2) + 1) y]! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'bak 6/24/2015 16:33'!
loadPointShortAt: index from: intArray 
	"Load the short value from the given index in intArray"
	| long |
	long := intArray at: index // 2.
	long isPoint
		ifTrue: [^ (index bitAnd: 1)
					= 0
				ifTrue: [long x]
				ifFalse: [long y]].
	^ ((index bitAnd: 1)
			= 0
		ifTrue: [interpreterProxy halfWordHighInLong32: long]
		ifFalse: [interpreterProxy halfWordLowInLong32: long]) signedIntFromShort
! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'eem 9/30/2015 11:56'!
loadWordTransformFrom: transformOop into: destPtr length: n 
	"Load a float array transformation from the given oop"
	| srcPtr |
	srcPtr := interpreterProxy firstIndexableField: transformOop.
	"Use BalloonArray's float conversion shell.  It stores both the float bits into
	 the heap object, and the float object into the simArray.  This is needed for
	 correct simulation, otherwise there will be errors in BalloonArray>>floatAt:"
	0 to: n - 1 do: [:i| destPtr at: i put: (srcPtr floatAt: i)]! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'ar 10/28/1998 01:05'!
makeUnsignedFrom: integer
	integer < 0 
		ifTrue:[^(0 - integer - 1) bitInvert32]
		ifFalse:[^integer]! !

!BalloonEngineSimulation methodsFor: 'accessing objects' stamp: 'eem 10/26/2018 20:51'!
obj: object at: index
	"Override to correctly map objBuffer access, which has type int *"
	^(objBuffer at: object + index) signedIntFromLong! !

!BalloonEngineSimulation methodsFor: 'accessing objects' stamp: 'eem 10/26/2018 20:52'!
obj: object at: index put: value
	"Override to correctly map objBuffer access, which has type int *"
	objBuffer at: object + index put: value signedIntToLong! !

!BalloonEngineSimulation methodsFor: 'initialize' stamp: 'di 7/16/2004 12:06'!
primitiveInitializeBuffer

	"Fix an uninitialized variable (should probably go into the real engine too)"
	super primitiveInitializeBuffer.
	self spanEndAAPut: 0.! !

!BalloonEngineSimulation methodsFor: 'initialize' stamp: 'di 7/12/2004 16:15'!
primitiveSetBitBltPlugin
	"Primitive. Set the BitBlt plugin to use."
	| pluginName |
	pluginName := interpreterProxy stackValue: 0.
	"Must be string to work"
	(interpreterProxy isBytes: pluginName) 
		ifFalse:[^interpreterProxy primitiveFail].
	(interpreterProxy stringOf: pluginName) = bbPluginName
		ifTrue: [interpreterProxy pop: 1. "Return receiver"]
		ifFalse: [^interpreterProxy primitiveFail]! !

!BalloonEngineSimulation methodsFor: 'debug support' stamp: 'ar 11/5/1998 21:15'!
printAET

	| edge |
	Transcript cr; show:'************* ActiveEdgeTable **************'.
	0 to: self aetUsedGet - 1 do:[:i|
		edge := aetBuffer at: i.
		Transcript cr;
			print: i; space;
			nextPutAll:'edge #';print: edge; space;
			nextPutAll:'x: '; print: (self edgeXValueOf: edge); space;
			nextPutAll:'y: '; print: (self edgeYValueOf: edge); space;
			nextPutAll:'z: '; print: (self edgeZValueOf: edge); space;
			nextPutAll:'fill0: '; print: (self edgeLeftFillOf: edge); space;
			nextPutAll:'fill1: '; print: (self edgeRightFillOf: edge); space;
			nextPutAll:'lines: '; print: (self edgeNumLinesOf: edge); space.
		(self areEdgeFillsValid: edge) ifFalse:[Transcript nextPutAll:' disabled'].
		Transcript endEntry.
	].! !

!BalloonEngineSimulation methodsFor: 'debug support' stamp: 'ar 11/5/1998 21:14'!
printGET

	| edge |
	Transcript cr; show:'************* GlobalEdgeTable **************'.
	0 to: self getUsedGet - 1 do:[:i|
		edge := getBuffer at: i.
		Transcript cr;
			print: i; space;
			nextPutAll:'edge #';print: edge; space;
			nextPutAll:'x: '; print: (self edgeXValueOf: edge); space;
			nextPutAll:'y: '; print: (self edgeYValueOf: edge); space;
			nextPutAll:'z: '; print: (self edgeZValueOf: edge); space;
			nextPutAll:'fill0: '; print: (self edgeLeftFillOf: edge); space;
			nextPutAll:'fill1: '; print: (self edgeRightFillOf: edge); space;
			nextPutAll:'lines: '; print: (self edgeNumLinesOf: edge); space.
		(self areEdgeFillsValid: edge) ifFalse:[Transcript nextPutAll:' disabled'].
		Transcript endEntry.
	].! !

!BalloonEngineSimulation methodsFor: 'debug support' stamp: 'ar 10/30/1998 21:57'!
quickPrint: curve
	Transcript nextPut:$(;
		print: curve start;
		space;
		print: curve via;
		space;
		print: curve end;
		nextPut:$).! !

!BalloonEngineSimulation methodsFor: 'debug support' stamp: 'ar 10/30/1998 22:18'!
quickPrintBezier: bezier
	Transcript cr.
	Transcript nextPut:$(;
		print: (self edgeXValueOf: bezier)@(self edgeYValueOf: bezier);
		space;
		print: (self bezierViaXOf: bezier)@(self bezierViaYOf: bezier);
		space;
		print: (self bezierEndXOf: bezier)@(self bezierEndYOf: bezier);
		nextPut:$).
	Transcript endEntry.! !

!BalloonEngineSimulation methodsFor: 'debug support' stamp: 'ar 10/30/1998 22:00'!
quickPrintBezier: index first: aBool
	aBool ifTrue:[Transcript cr].
	Transcript nextPut:$(;
		print: (self bzStartX: index)@(self bzStartY: index);
		space;
		print: (self bzViaX: index)@(self bzViaY: index);
		space;
		print: (self bzEndX: index)@(self bzEndY: index);
		nextPut:$).
	Transcript endEntry.! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'ar 11/25/1998 19:24'!
rShiftTable
	^CArrayAccessor on: #(0 5 4 0 3 0 0 0 2 0 0 0 0 0 0 0 1).! !

!BalloonEngineSimulation methodsFor: 'simulation support' stamp: 'eem 10/26/2018 20:31'!
selectorsForIntMethods
	"Ensure that all methods that require int and unsigned int arithmetic are overridden."
	| selectors cg |
	selectors := OrderedCollection new.
	cg := CCodeGenerator new.
	^(BalloonEnginePlugin organization listAtCategoryNamed: 'beziers-simple') select:
		[:selector|
		(BalloonEnginePlugin >> selector) pragmas anySatisfy:
			[:p| | type |
			 p keyword == #var:type: ifTrue:
				[type := p arguments last].
			 p keyword == #var:declareC: ifTrue:
				[type := cg
							extractTypeFor: p arguments first
							fromDeclaration: p arguments last].
			 type notNil
			 and: [type last = $* ifTrue:
					[type := cg baseTypeForPointerType: type].
				#(int #'unsigned int') includes: (cg baseTypeForType: type)]]]! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'di 7/16/2004 15:11'!
shortRunLengthAt: index from: runArray
	"Load the short value from the given index in intArray"

	^ interpreterProxy halfWordHighInLong32: (runArray at: index)! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'di 7/16/2004 15:10'!
shortRunValueAt: index from: runArray
	"Load the short value from the given index in intArray"

	^ (interpreterProxy halfWordLowInLong32: (runArray at: index)) signedIntFromShort ! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'ar 10/31/1998 23:07'!
showDisplayBits
	"Do nothing."! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'ar 11/4/1998 19:51'!
smallSqrtTable
	"Return a lookup table for rounded integer square root values from 0 to 31"
	^CArrayAccessor on:#(0 1 1 2 2 2 2 3 3 3 3 3 3 4 4 4 4 4 4 4 4 5 5 5 5 5 5 5 5 5 5 6 )! !

!BalloonEngineSimulation methodsFor: 'generated overrides' stamp: 'eem 10/26/2018 20:38'!
stepToFirstBezierIn: bezier at: yValue
	"Initialize the bezier at yValue.
	TODO: Check if reducing maxSteps from 2*deltaY to deltaY 
		brings a *significant* performance improvement.
		In theory this should make for double step performance
		but will cost in quality. Might be that the AA stuff will
		compensate for this - but I'm not really sure."

	<inline: false>
	<var: #updateData type: 'int *'>
	| updateData deltaY maxSteps scaledStepSize squaredStepSize startX startY viaX viaY endX endY fwX1 fwX2 fwY1 fwY2 fwDx fwDDx fwDy fwDDy |
	((self isWide: bezier) not and: [ yValue >= (self bezierEndYOf: bezier) ])
		ifTrue: [ ^ self edgeNumLinesOf: bezier put: 0 ].	"Do a quick check if there is anything at all to do"	"Now really initialize bezier"
	startX := self edgeXValueOf: bezier.
	startY := self edgeYValueOf: bezier.
	viaX := self bezierViaXOf: bezier.
	viaY := self bezierViaYOf: bezier.
	endX := self bezierEndXOf: bezier.
	endY := self bezierEndYOf: bezier.
	deltaY := endY - startY.	"Initialize integer forward differencing"
	fwX1 := (viaX - startX) * 2.
	fwX2 := startX + endX - (viaX * 2).
	fwY1 := (viaY - startY) * 2.
	fwY2 := startY + endY - (viaY * 2).
	maxSteps := deltaY * 2.
	maxSteps < 2
		ifTrue: [ maxSteps := 2 ].
	scaledStepSize := 16r1000000 // maxSteps.
	squaredStepSize := self absoluteSquared8Dot24: scaledStepSize.
	fwDx := fwX1 * scaledStepSize.
	fwDDx := fwX2 * squaredStepSize * 2.
	fwDx := fwDx + (fwDDx // 2).
	fwDy := fwY1 * scaledStepSize.
	fwDDy := fwY2 * squaredStepSize * 2.
	fwDy := fwDy + (fwDDy // 2).	"Store the values"
	self edgeNumLinesOf: bezier put: deltaY.
	updateData := self bezierUpdateDataOf: bezier.
	updateData at: GBUpdateX put: (startX * 256) asC_int.
	updateData at: GBUpdateY put: (startY * 256) asC_int.
	updateData at: GBUpdateDX put: fwDx asC_int.
	updateData at: GBUpdateDY put: fwDy asC_int.
	updateData at: GBUpdateDDX put: fwDDx asC_int.
	updateData at: GBUpdateDDY put: fwDDy asC_int.	"And step to the first scan line"
	(startY := self edgeYValueOf: bezier) = yValue
		ifFalse: [ 
			self stepToNextBezierIn: bezier at: yValue.	"Adjust number of lines remaining"
			self edgeNumLinesOf: bezier put: deltaY - (yValue - startY) ]! !

!BalloonEngineSimulation methodsFor: 'generated overrides' stamp: 'eem 10/26/2018 20:38'!
stepToNextBezierForward: updateData at: yValue
	"Incrementally step to the next scan line in the given bezier update data.
	Note: This method has been written so that inlining works, e.g.,
		not declaring updateData as 'int*' but casting it on every use."

	<var: #updateData type: 'int *'>
	<inline: true>
	| minY lastX lastY fwDx fwDy |
	lastX := updateData at: GBUpdateX.
	lastY := updateData at: GBUpdateY.
	fwDx := updateData at: GBUpdateDX.
	fwDy := updateData at: GBUpdateDY.
	minY := yValue * 256.	"Step as long as we haven't yet reached minY and also
	as long as fwDy is greater than zero thus stepping down.
	Note: The test for fwDy should not be necessary in theory
		but is a good insurance in practice."
	[ minY > lastY and: [ fwDy >= 0 ] ]
		whileTrue: [ 
			lastX := lastX + (fwDx + 16r8000 signedBitShift: -16).
			lastY := lastY + (fwDy + 16r8000 signedBitShift: -16).
			fwDx := fwDx + (updateData at: GBUpdateDDX).
			fwDy := fwDy + (updateData at: GBUpdateDDY) ].
	updateData at: GBUpdateX put: lastX asC_int.
	updateData at: GBUpdateY put: lastY asC_int.
	updateData at: GBUpdateDX put: fwDx asC_int.
	updateData at: GBUpdateDY put: fwDy asC_int.
	^ lastX signedBitShift: -8! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'ar 11/25/1998 02:23'!
stopBecauseOf: stopReason
	"Don't stop because of need to flush."
	stopReason = GErrorNeedFlush ifFalse:[
		^super stopBecauseOf: stopReason.
	].! !

!BalloonEngineSimulation methodsFor: 'simulation support' stamp: 'eem 10/26/2018 20:22'!
transformedSourceFor: m
	"Transform all references to #int variables in the source code of m into var asC_int
	 and all references to #'unsigned int' variables into var asC_unsigned_int"
	| tm rules cgen |
	tm := m asTranslationMethodOfClass: TMethod.
	tm recordDeclarationsIn: (cgen := CCodeGenerator new).
	rules := RBParseTreeRewriter new.
	#(#'int *' #'unsigned int *')
		with: #(asC_int asC_unsigned_int)
		do: [:type :coercionMessage|
			(tm declarations keys select: [:v| (tm typeFor: v in: cgen) = type]) do:
				[:var| "make this use ``@expr and editSource breaks"
				rules replace: var, ' at: `@index put: `@expr' with: var, ' at: `@index put: `@expr ', coercionMessage]].
	#(#'int' #'unsigned int')
		with: #(asC_int asC_unsigned_int)
		do: [:type :coercionMessage|
			(tm declarations keys select: [:v| (tm typeFor: v in: cgen) = type]) do:
				[:var|
				rules
					replace: var, ' := `@args' with: var, ' := `@args ', coercionMessage;
					replace: var with: var, ' ', coercionMessage]].
	(rules executeTree: (m methodClass parseTreeFor: m selector)) ifTrue:
		[^rules tree newSource].
	self error: 'parse tree rewrite failed'! !

!BalloonEngineSimulation methodsFor: 'simulation' stamp: 'eem 11/5/2013 15:15'!
workBufferPut: wbOop
	interpreterProxy isInterpreterProxy 
		ifTrue:[^super workBufferPut: wbOop].
	workBuffer := ((interpreterProxy firstIndexableField: wbOop) as: BalloonArray) asCArrayAccessor.
	workBufferArray ifNil:
		[workBufferArray := Array new: (interpreterProxy slotSizeOf: wbOop) withAll: 0].
	workBuffer getObject setSimArray: workBufferArray! !

InterpreterPlugin subclass: #BrokenPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!BrokenPlugin commentStamp: '<historical>' prior: 0!
A test plugin to exercise imbalanced stack access.!

!BrokenPlugin class methodsFor: 'tests' stamp: 'ar 3/5/2009 11:42'!
testNoPopButPush
	"BrokenPlugin testNoPopButPush"
	<primitive: 'primitiveNoPopButPush' module: 'BrokenPlugin'>
! !

!BrokenPlugin class methodsFor: 'tests' stamp: 'ar 3/5/2009 11:42'!
testPopAndFail
	"BrokenPlugin testPopAndFail"
	<primitive: 'primitivePopAndFail' module: 'BrokenPlugin'>
! !

!BrokenPlugin class methodsFor: 'tests' stamp: 'ar 3/5/2009 12:01'!
testPopLessThanExpected: arg1 with: arg2
	"BrokenPlugin testPopLessThanExpected: 3 with: 4"
	<primitive: 'primitivePopLessThanExpected' module: 'BrokenPlugin'>
! !

!BrokenPlugin class methodsFor: 'tests' stamp: 'ar 3/5/2009 11:41'!
testPopMoreThanExpected
	"BrokenPlugin testPopMoreThanExpected"
	<primitive: 'primitivePopMoreThanExpected' module: 'BrokenPlugin'>
! !

!BrokenPlugin methodsFor: 'primitives' stamp: 'ar 3/5/2009 11:35'!
primitiveNoPopButPush
	"Doesn't pop anything but pushes return value"
	<export: true>
	interpreterProxy pushBool: true.! !

!BrokenPlugin methodsFor: 'primitives' stamp: 'ar 3/5/2009 11:39'!
primitivePopAndFail
	"Pops in a failing primitive"
	<export: true>
	interpreterProxy pop: interpreterProxy methodArgumentCount + 1.
	interpreterProxy primitiveFail.! !

!BrokenPlugin methodsFor: 'primitives' stamp: 'ar 3/5/2009 11:39'!
primitivePopLessThanExpected
	"Pops less than expected; call this with two arguments."
	<export: true>
	interpreterProxy pop: interpreterProxy methodArgumentCount - 1.
! !

!BrokenPlugin methodsFor: 'primitives' stamp: 'ar 3/5/2009 11:39'!
primitivePopMoreThanExpected
	"Pops more than expected"
	<export: true>
	interpreterProxy pop: interpreterProxy methodArgumentCount + 4.
! !

InterpreterPlugin subclass: #CameraPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!CameraPlugin class methodsFor: 'translation' stamp: 'ikp 7/30/2012 14:39'!
hasHeaderFile

	^true! !

!CameraPlugin methodsFor: 'initialize-release' stamp: 'eem 10/14/2020 07:52'!
initialiseModule
	<export: true>
	^self cameraInit! !

!CameraPlugin methodsFor: 'primitives' stamp: 'eem 4/18/2021 16:10'!
primAcknowledgeFrame
	"Acknowledge the receipt of one or more buffered frames, installed via primSetCameraBuffers.
	 Answer the number of times a buffer was filled since the last invocation.  This is a version of
	 primGetFrame that does not need to supply a buffer to receiver the result."
	<export: true>
	| cameraNum result |

	cameraNum := interpreterProxy stackValue: 0.
	(interpreterProxy isIntegerObject: cameraNum) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	result := self Camera: cameraNum Get: 0 Frame: 0.
	result < 0 ifTrue:
		[^interpreterProxy primitiveFailFor: result negated].

	interpreterProxy methodReturnInteger: result! !

!CameraPlugin methodsFor: 'primitives' stamp: 'eem 10/13/2020 18:16'!
primCameraName
	"Answer the name for the camera with the given number.
	 Fail if the camera number is less than one or greater than the number of available cameras."

	<export: true>
	| cameraNum |
	cameraNum := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnString: (self CameraName: cameraNum)] "fails for a NULL string..."! !

!CameraPlugin methodsFor: 'primitives' stamp: 'eem 10/13/2020 18:16'!
primCameraUID
	"Answer a UID for the camera with the given number.
	 Fail if the camera number is less than one or greater than the number of available cameras."

	<export: true>
	| cameraNum |
	cameraNum := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnString: (self CameraUID: cameraNum)] "fails for a NULL string..."! !

!CameraPlugin methodsFor: 'primitives' stamp: 'eem 10/7/2020 14:36'!
primCloseCamera
	"Close the camera. Do nothing if it was not open."

	<export: true>
	| cameraNum |
	cameraNum := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue: [^ 0].

	self CameraClose: cameraNum.

	interpreterProxy methodReturnReceiver! !

!CameraPlugin methodsFor: 'primitives' stamp: 'eem 3/16/2018 14:25'!
primFrameExtent
	"Answer the frame extent of the given camera, or zero if the camera is not open. The extent is 16 bits of width and height packed into a single integer."

	| cameraNum |
	<export: true>
	cameraNum := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnInteger: (self CameraExtent: cameraNum)]! !

!CameraPlugin methodsFor: 'primitives' stamp: 'eem 10/16/2020 09:13'!
primGetCameraSemaphore
	"Answers the semaphoreIndex set through primSetCameraSemaphore.
	 Fails if no semaphore has been set."

	<export: true>
	| cameraNum |
	cameraNum := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifFalse:
		[(self CameraGetSemaphore: cameraNum)
			ifNil: [interpreterProxy primitiveFail]
			ifNotNil: [:index| interpreterProxy methodReturnInteger: index]]! !

!CameraPlugin methodsFor: 'primitives' stamp: 'eem 4/12/2021 13:00'!
primGetFrame
	"Copy a camera frame into the given Bitmap. The Bitmap should be for a Form of depth 32 that is the same width and height as the current camera frame. Fail if the camera is not open or if the bitmap is not the right size. If successful, answer the number of frames received from the camera since the last call. If this is zero, then there has been no change."

	| cameraNum bitmapOop bitmap pixCount result |
	<export: true>
	<var: 'bitmap' type: #'unsigned char *'>

	cameraNum := interpreterProxy stackIntegerValue: 1.
	bitmapOop := interpreterProxy stackValue: 0.
	interpreterProxy success: (interpreterProxy isWords: bitmapOop).
	interpreterProxy failed ifTrue:
		[^ 0].
	bitmap := self cCoerce: (interpreterProxy firstIndexableField: bitmapOop) to: #'unsigned char *'.
	pixCount := interpreterProxy stSizeOf: bitmapOop.

	result := self Camera: cameraNum Get: bitmap Frame: pixCount.
	result < 0 ifTrue:
		[^interpreterProxy primitiveFailFor: result negated].

	interpreterProxy methodReturnInteger: result! !

!CameraPlugin methodsFor: 'primitives' stamp: 'eem 3/16/2018 14:28'!
primGetParam
	"Answer the given integer parameter of the given camera."

	| cameraNum paramNum |
	<export: true>
	cameraNum := interpreterProxy stackIntegerValue: 1.
	paramNum := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnInteger: (self Camera: cameraNum GetParam: paramNum)]! !

!CameraPlugin methodsFor: 'primitives' stamp: 'eem 4/18/2021 16:29'!
primOpenCamera
	"Open a camera. Takes one argument, the index of the device to open.
	 CameraOpen should choose the default size for the device if both desiredFrameWidth
	 and desiredFrameHeight are zero, and the closest to the non-zero value(s) if either or both
	 of desiredFrameWidth and desiredFrameHeight are non-zero."

	<export: true>
	| cameraNum desiredFrameWidth desiredFrameHeight ok |

	cameraNum := interpreterProxy stackIntegerValue: 2.
	desiredFrameWidth := interpreterProxy stackIntegerValue: 1.
	desiredFrameHeight := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue: [^0].

	ok := self CameraOpen: cameraNum _: desiredFrameWidth _: desiredFrameHeight.
	ok <= 0 ifTrue:
		[interpreterProxy success: false.
		^0].

	interpreterProxy methodReturnReceiver! !

!CameraPlugin methodsFor: 'primitives' stamp: 'eem 4/12/2021 12:10'!
primSetCameraBuffers
	"Sets one or two frame buffers to receive camera frames. The buffer(s) must be big enough and pinned.
	Provides these two interfaces:
	camera: cameraNum setFrameBuffer: pinnedNonPointers
	camera: cameraNum setFrameBufferA: pinnedNonPointers B: pinnedNonPointersOrNil"

	<export: true>
	| cameraNum failCode bufferOne bufferTwo |
	interpreterProxy methodArgumentCount
		caseOf: {
			[3] ->	[cameraNum := interpreterProxy stackValue: 2.
					 bufferOne := interpreterProxy stackValue: 1.
					 bufferTwo := interpreterProxy stackValue: 0].
			[2] ->	[cameraNum := interpreterProxy stackValue: 1.
					 bufferOne := interpreterProxy stackValue: 0.
					 bufferTwo := nil] }
		otherwise:
			[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	((interpreterProxy isIntegerObject: cameraNum)
	and: [(interpreterProxy isWordsOrBytes: bufferOne)
	and: [(interpreterProxy isPinned: bufferOne)
	and: [bufferTwo isNil
		or: [bufferTwo = interpreterProxy nilObject
		or: [(interpreterProxy isWordsOrBytes: bufferTwo)
			and: [interpreterProxy isPinned: bufferTwo]]]]]]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	failCode := self Camera: (interpreterProxy integerValueOf: cameraNum)
					SetFrame: bufferOne
					Buffers: (bufferTwo = interpreterProxy nilObject ifFalse: [bufferTwo]). "coerce nilObject to nil"
	failCode ~= 0 ifTrue: [^interpreterProxy primitiveFailFor: failCode].

	interpreterProxy methodReturnReceiver! !

!CameraPlugin methodsFor: 'primitives' stamp: 'eem 10/7/2020 14:34'!
primSetCameraSemaphore
	"Supplies a semaphoreIndex through which to signal a semaphore when a frame is available from a camera.
	 Fails if the platform does not provide an interrupt-driven camera interface."

	<export: true>
	| cameraNum failCode semaphoreIndex |

	cameraNum := interpreterProxy stackIntegerValue: 1.
	semaphoreIndex := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue: [^0].

	failCode := self Camera: cameraNum SetSemaphore: semaphoreIndex.
	failCode ~= 0 ifTrue: [^interpreterProxy primitiveFailFor: failCode].

	interpreterProxy methodReturnReceiver! !

!CameraPlugin methodsFor: 'initialize-release' stamp: 'eem 10/14/2020 07:53'!
shutdownModule
	<export: true>
	^self cameraShutdown! !

InterpreterPlugin subclass: #CroquetPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!CroquetPlugin commentStamp: '<historical>' prior: 0!
An assorted list of useful primitives for Croquet.!

!CroquetPlugin class methodsFor: 'translation' stamp: 'ar 3/26/2006 19:37'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!CroquetPlugin class methodsFor: 'translation' stamp: 'ar 3/26/2006 19:37'!
requiresCrossPlatformFiles
	"default is ok for most, any plugin needing platform specific files must say so"
	^true! !

!CroquetPlugin methodsFor: 'simulation support' stamp: 'eem 8/3/2021 00:53'!
ioGatherEntropy: bytes _: length
	<doNotGenerate>
	| entropy |
	(InitializationOptions at: #entropy ifAbsent: nil)
		ifNotNil: [:repeatableEntropy|
			self assert: repeatableEntropy size = length.
			entropy := repeatableEntropy]
		ifNil:
			[entropy := ByteArray new: length.
			(Random gatherEntropyInto: entropy) ifFalse:
				[^false]].
	self memcpy: bytes _: entropy _: length.
	^true! !

!CroquetPlugin methodsFor: 'cryptography' stamp: 'eem 8/2/2021 17:05'!
primitiveARC4Transform
	"Perform an ARC4 transform of input.
	Arguments:
		buffer		<ByteArray> transformed data
		startIndex 	<Integer>	start of transform
		stopIndex	<Integer>	end of transform
		m			<ByteArray>	key stream data
		x			<Integer>	key state value
		y			<Integer>	key state value
	Return value:
		x@y - updated key state value
	"

	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)> "Align in case the compiler identifies and generates vector instructions"
	| y x mOop stopIndex startIndex bufOop bufSize buffer a m b mask ptOop xOop yOop |
	<var: 'buffer' type: #'unsigned char *'>
	<var: 'm' type: #'unsigned char *'>

	interpreterProxy methodArgumentCount = 6
		ifFalse:[^interpreterProxy primitiveFail].
	"pick up arguments"
	y := interpreterProxy stackIntegerValue: 0.
	x := interpreterProxy stackIntegerValue: 1.
	mOop := interpreterProxy stackValue: 2.
	stopIndex := interpreterProxy stackIntegerValue: 3.
	startIndex := interpreterProxy stackIntegerValue: 4.
	bufOop := interpreterProxy stackValue: 5.
	interpreterProxy failed ifTrue:[^nil].
	((interpreterProxy isBytes: mOop)
	and: [(interpreterProxy isBytes: bufOop)
	and: [(interpreterProxy byteSizeOf: mOop) = 256
	and: [startIndex > 0 and: [startIndex <= (bufSize := interpreterProxy byteSizeOf: bufOop)
	and: [stopIndex >= startIndex and: [stopIndex <= bufSize]]]]]]) ifFalse:
		[^interpreterProxy primitiveFail].
	m := interpreterProxy firstIndexableField: mOop.
	buffer := interpreterProxy firstIndexableField: bufOop.
	startIndex-1 to: stopIndex-1 do:[:i|
		x := (x + 1) bitAnd: 255.
		a := m at: x.
		y := (y + a) bitAnd: 255.
		b := m at: y.
		m at: x put: b.
		m at: y put: a.
		mask := m at: ((a + b) bitAnd: 255).
		buffer at: i put: ((buffer at: i) bitXor: mask).
	].
	ptOop := interpreterProxy instantiateClass: interpreterProxy classPoint indexableSize: 0.
	self cppIf: #SPURVM
		ifTrue:
			[xOop := interpreterProxy positive32BitIntegerFor: x.
			yOop := interpreterProxy positive32BitIntegerFor: y.
			(ptOop isNil or: [xOop isNil or: [yOop isNil]]) ifTrue:
				[^interpreterProxy primitiveFailFor: PrimErrNoMemory].
			interpreterProxy storePointer: 0 ofObject: ptOop withValue: xOop.
			interpreterProxy storePointer: 1 ofObject: ptOop withValue: yOop]
		ifFalse:
			[interpreterProxy pushRemappableOop: ptOop.
			xOop := interpreterProxy positive32BitIntegerFor: x.
			interpreterProxy pushRemappableOop: xOop.
			yOop := interpreterProxy positive32BitIntegerFor: y.
			xOop := interpreterProxy popRemappableOop.
			ptOop := interpreterProxy popRemappableOop.
			interpreterProxy storePointer: 0 ofObject: ptOop withValue: xOop.
			interpreterProxy storePointer: 1 ofObject: ptOop withValue: yOop].
	^interpreterProxy methodReturnValue: ptOop! !

!CroquetPlugin methodsFor: 'transforms' stamp: 'eem 8/2/2021 17:21'!
primitiveAdj3
	"Computes the adjoint of the Matrix4x4 receiver,
	placing the results the the Matrix4x4 argument,
	"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	|
		srcOop src
		dstOop dst
		m11 m12 m13 m21 m22 m23 m31 m32 m33
		c11 c12 c13 c21 c22 c23 c31 c32 c33  
	|
	<var: #c11 declareC: 'const int c11 = 0'>
	<var: #c12 declareC: 'const int c12 = 1'>
	<var: #c13 declareC: 'const int c13 = 2'>
	"<var: #c14 declareC: 'const int c14 = 3'>"
	<var: #c21 declareC: 'const int c21 = 4'>
	<var: #c22 declareC: 'const int c22 = 5'>
	<var: #c23 declareC: 'const int c23 = 6'>
	"<var: #c24 declareC: 'const int c24 = 7'>"
	<var: #c31 declareC: 'const int c31 = 8'>
	<var: #c32 declareC: 'const int c32 = 9'>
	<var: #c33 declareC: 'const int c33 = 10'>
	"<var: #c34 declareC: 'const int c34 = 11'>"
	<var: #m11 type: #double>
	<var: #m12 type: #double>
	<var: #m13 type: #double>
	<var: #m21 type: #double>
	<var: #m22 type: #double>
	<var: #m23 type: #double>
	<var: #m31 type: #double>
	<var: #m32 type: #double>
	<var: #m33 type: #double>

	"then we need the following no-op to make Smalltalk shut up about vars not being initted."
	self cCode: '' inSmalltalk: [ 
		c11 := 0. 
		c12 := 1.
		c13 := 2.
		"c14 := 3."
		c21 := 4.
		c22 := 5.
		c23 := 6.
		"c24 := 7."
		c31 := 8.
		c32 := 9.
		c33 := 10.
		"c34 := 11."
	].

	"NOTE: the bottom row of a OpenGL-ordered matrix is always 0 0 0 1, 
	so we don't need consts here for those elements."

	srcOop := interpreterProxy stackObjectValue: 1.	
	dstOop := interpreterProxy stackObjectValue: 0.
	interpreterProxy failed ifTrue:
		[^nil].
	src := self cCoerce: (interpreterProxy firstIndexableField: srcOop) to: #'float *'.
	dst := self cCoerce: (interpreterProxy firstIndexableField: dstOop) to: #'float *'.

	"read in the source matrix 3x3, which contains the encoded rotation and scale factors"
	m11 := src at: c11.
	m12 := src at: c12.
	m13 := src at: c13.
	m21 := src at: c21.
	m22 := src at: c22.
	m23 := src at: c23.
	m31 := src at: c31.
	m32 := src at: c32.
	m33 := src at: c33.

	"do the actual work"

	"compute our cofactors and transpose.  adj = transpose of cofactors"
	dst at: c11 put:  ((m22 * m33) - (m23 *  m32)) .
	dst at: c21 put: (0.0 - ((m21 * m33) - (m23 * m31))).
	dst at: c31 put: ((m21 * m32) - (m22 * m31)).

	dst at: c12 put: (0.0 - ((m12 * m33) - (m13 * m32))).
	dst at: c22 put: ((m11 * m33) - (m13 * m31)).
	dst at: c32 put: (0.0 - ((m11 * m32) - (m12 * m31))).

	dst at: c13 put: ((m12 * m23) - (m13 * m22)).
	dst at: c23 put: (0.0 - ((m11 * m23) - (m13 * m21))).
	dst at: c33 put: ((m11 * m22) - (m12 * m21)).
	
	^interpreterProxy methodReturnValue: dstOop! !

!CroquetPlugin methodsFor: 'transforms' stamp: 'eem 8/2/2021 17:21'!
primitiveDet3
	"Computes the determinant of the upper 3x3 of a Matrix4x4"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| srcOop src m11 m12 m13 m21 m22 m23 m31 m32 m33 |
	<var: #m11 type: #double>
	<var: #m12 type: #double>
	<var: #m13 type: #double>
	<var: #m21 type: #double>
	<var: #m22 type: #double>
	<var: #m23 type: #double>
	<var: #m31 type: #double>
	<var: #m32 type: #double>
	<var: #m33 type: #double>

	"
		Load element vars using C version of Matrix4x4 storage, as 0-based, 1-dimensional array:
			0 1 2 3
			4 5 6 7
			8 9 10 11
	"
	srcOop := interpreterProxy stackObjectValue: 0.	
	interpreterProxy failed ifTrue:
		[^nil].
	src := self cCoerce: (interpreterProxy firstIndexableField: srcOop) to: #'float *'.

	m11 := src at: 0.
	m12 := src at: 1.
	m13 := src at: 2.
	m21 := src at: 4.
	m22 := src at: 5.
	m23 := src at: 6.
	m31 := src at: 8.
	m32 := src at: 9.
	m33 := src at: 10.

	^interpreterProxy methodReturnFloat: (m11 * ((m22 * m33) - (m23 * m32)))
										+ (m12 * ((m23 * m31) - (m21 * m33)))
										+ (m13 * ((m21 * m32) - (m22 * m31)))
! !

!CroquetPlugin methodsFor: 'cryptography' stamp: 'eem 8/2/2021 17:34'!
primitiveGatherEntropy
	"Primitive. Gather good random entropy from a system source."
	<export: true>
	| bufOop |
	bufOop := interpreterProxy stackValue: 0.
	((interpreterProxy isBytes: bufOop)
	 and: [self ioGatherEntropy: (interpreterProxy firstIndexableField: bufOop) _: (interpreterProxy byteSizeOf: bufOop)]) ifFalse:
		[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnBool: true! !

!CroquetPlugin methodsFor: 'transforms' stamp: 'eem 8/2/2021 17:11'!
primitiveInplaceHouseHolderInvert
	"Primitive. Perform an inplace house holder matrix inversion"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr d x sigma beta sum s m |
	<var: #rcvr type: #'float *'>
	<var: #m declareC:'double m[4][4]'>
	<var: #x declareC:'double x[4][4] = { {1, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 1, 0}, {0, 0, 0, 1} }'>
	<var: #d declareC:'double d[4][4]'>
	<var: #sigma type: #double>
	<var: #beta type: #double>
	<var: #sum type: #double>
	<var: #s type: #double>

	self cCode:'' inSmalltalk:[
		m := CArrayAccessor on: 
				((1 to: 4) collect:[:i| CArrayAccessor on: (Array new: 4)]).
		x := CArrayAccessor on: (Array
				with: (CArrayAccessor on: #(1.0 0.0 0.0 0.0) copy)
				with: (CArrayAccessor on: #(0.0 1.0 0.0 0.0) copy)
				with: (CArrayAccessor on: #(0.0 0.0 1.0 0.0) copy)
				with: (CArrayAccessor on: #(0.0 0.0 0.0 1.0) copy)).
		d := CArrayAccessor on: 
				((1 to: 4) collect:[:i| CArrayAccessor on: (Array new: 4)]).
	].
	rcvr := self stackMatrix: 0.
	rcvr ifNil:
		[^interpreterProxy primitiveFail].
	0 to: 3 do:[:i| 0 to: 3 do:[:j|
		(m at: i) at: j put: (rcvr at: i*4+j)]].
	0 to: 3 do:[:j|
		sigma := 0.0.
		j to: 3 do:[:i| sigma := sigma + (((m at: i) at: j)  * ((m at: i) at: j))].
		sigma < 1.0e-10 ifTrue:[^interpreterProxy primitiveFail]. "matrix is singular"
		(((m at: j) at: j) < 0.0) 
			ifTrue:[ s:= sigma sqrt]
			ifFalse:[ s:= 0.0 - sigma sqrt].
		0 to: 3 do:[:r| (d at: j) at: r put: s].
		beta := 1.0 / ( s * ((m at: j) at: j) - sigma).
		(m at: j) at: j put: (((m at: j) at: j) - s).
		"update remaining columns"
		j+1 to: 3 do:[:k|
			sum := 0.0.
			j to: 3 do:[:i| sum := sum + (((m at: i) at: j) * ((m at: i) at: k))].
			sum := sum * beta.
			j to: 3 do:[:i| 
				(m at: i) at: k put: (((m at: i) at: k) + (((m at: i) at: j) * sum))]].
		"update vector"
		0 to: 3 do:[:r|
			sum := 0.0.
			j to: 3 do:[:i| 
				sum := sum + (((x at: i) at: r) * ((m at: i) at: j))].
			sum := sum * beta.
			j to: 3 do:[:i| 
				(x at: i) at: r put:(((x at: i) at: r) + (sum * ((m at: i) at: j)))].
		].
	].
	"Now calculate result"
	0 to: 3 do:[:r|
		3 to: 0 by: -1 do:[:i|
			i+1 to: 3 do:[:j|
				(x at: i) at: r put: (((x at: i) at: r) - (((x at: j) at: r) * ((m at: i) at: j))) ].
			(x at: i) at: r put: (((x at: i) at: r) / ((d at: i) at: r))].
	].
	0 to: 3 do:[:i| 0 to: 3 do:[:j|
		rcvr at: i*4+j put: (self cCoerce: ((x at: i) at: j) to:'float')]].
	"Return receiver"
	^nil! !

!CroquetPlugin methodsFor: 'transforms' stamp: 'eem 8/2/2021 17:18'!
primitiveInverseByAdjoint
	"Computes the inverse of the Matrix4x4 receiver, using the 'classical adjoint' method,
	placing the results the the Matrix4x4 argument,
	"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	|
		srcOop src
		dstOop dst
		det
		m11 m12 m13 m21 m22 m23 m31 m32 m33
		c11 c12 c13 c14 c21 c22 c23 c24 c31 c32 c33 c34
		x y z
	 |
	<var: #c11 declareC: 'const int c11 = 0'>
	<var: #c12 declareC: 'const int c12 = 1'>
	<var: #c13 declareC: 'const int c13 = 2'>
	<var: #c14 declareC: 'const int c14 = 3'>
	<var: #c21 declareC: 'const int c21 = 4'>
	<var: #c22 declareC: 'const int c22 = 5'>
	<var: #c23 declareC: 'const int c23 = 6'>
	<var: #c24 declareC: 'const int c24 = 7'>
	<var: #c31 declareC: 'const int c31 = 8'>
	<var: #c32 declareC: 'const int c32 = 9'>
	<var: #c33 declareC: 'const int c33 = 10'>
	<var: #c34 declareC: 'const int c34 = 11'>
	<var: #m11 type: #double>
	<var: #m12 type: #double>
	<var: #m13 type: #double>
	<var: #m21 type: #double>
	<var: #m22 type: #double>
	<var: #m23 type: #double>
	<var: #m31 type: #double>
	<var: #m32 type: #double>
	<var: #m33 type: #double>
	<var: #x type:#double>
	<var: #y type:#double>
	<var: #z type:#double>
	<var: #det type:#double>

	"then we need the following no-op to make Smalltalk shut up about vars not being initted."
	self cCode: '' inSmalltalk: [ 
		c11 := 0. 
		c12 := 1.
		c13 := 2.
		c14 := 3.
		c21 := 4.
		c22 := 5.
		c23 := 6.
		c24 := 7.
		c31 := 8.
		c32 := 9.
		c33 := 10.
		c34 := 11.
	].

	"NOTE: the bottom row of a OpenGL-ordered matrix is always 0 0 0 1, 
	so we don't need consts here for those elements."

	"do the dance to get our receiver and argument"
	srcOop := interpreterProxy stackValue: 1.
	dstOop := interpreterProxy stackValue: 0.
	interpreterProxy failed ifTrue:
		[^nil].
	src := self cCoerce: (interpreterProxy firstIndexableField: srcOop) to: #'float *'.
	dst := self cCoerce: (interpreterProxy firstIndexableField: dstOop) to: #'float *'.


	"read in the source matrix 3x3, which contains the encoded rotation and scale factors"
	m11 := src at: c11.
	m12 := src at: c12.
	m13 := src at: c13.
	m21 := src at: c21.
	m22 := src at: c22.
	m23 := src at: c23.
	m31 := src at: c31.
	m32 := src at: c32.
	m33 := src at: c33.

	"read in the source translation vector"
	x := src at: c14.
	y := src at: c24.
	z := src at: c34.

	"do the actual work"

	"first, compute the determinant of the upper 3x3 of the source"
	det := 
		( m11 * ((m22 * m33) - (m23 * m32))) +
			(m12 * ((m23 * m31) - (m21 * m33))) +
				(m13 * ((m21 * m32) - (m22 * m31))).

	"Compute the classical adjunct of the source, and divide by the source determinant
	storing in the destination.  adjoint = transpose of cofactors, so we'll transpose as we store."

	det := 1 / det.		"let's make div by det a multiply"
	dst at: c11 put:  ((m22 * m33) - (m23 *  m32)) * det .
	dst at: c21 put: (0.0 - ((m21 * m33) - (m23 * m31))) * det.
	dst at: c31 put: ((m21 * m32) - (m22 * m31)) * det.

	dst at: c12 put: (0.0 - ((m12 * m33) - (m13 * m32))) * det.
	dst at: c22 put: ((m11 * m33) - (m13 * m31)) * det.
	dst at: c32 put: (0.0 - ((m11 * m32) - (m12 * m31))) * det.

	dst at: c13 put: ((m12 * m23) - (m13 * m22)) * det.
	dst at: c23 put: (0.0 - ((m11 * m23) - (m13 * m21))) * det.
	dst at: c33 put: ((m11 * m22) - (m12 * m21)) * det.
	
	"finally, apply the inversed rotation transform to our translation"
	
	"read in the source matrix 3x3"
	m11 := dst at: c11.
	m12 := dst at: c12.
	m13 := dst at: c13.
	m21 := dst at: c21.
	m22 := dst at: c22.
	m23 := dst at: c23.
	m31 := dst at: c31.
	m32 := dst at: c32.
	m33 := dst at: c33.

	dst at: c14 put: 0.0 - ((x * m11) + (y * m12) + (z * m13)).
	dst at: c24 put: 0.0 - ((x * m21) + (y * m22) + (z * m23)).
	dst at: c34 put: 0.0 - ((x * m31) + (y * m32) + (z * m33)).

	^interpreterProxy methodReturnValue: dstOop! !

!CroquetPlugin methodsFor: 'cryptography' stamp: 'eem 8/2/2021 17:20'!
primitiveMD5Transform
	"Perform an MD5 transform of input"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)> "Align in case the compiler identifies and generates vector instructions"
	| bufOop hashOop |
	hashOop := interpreterProxy stackValue: 0.
	bufOop := interpreterProxy stackValue: 1.
	((interpreterProxy isWords: hashOop) and: [(interpreterProxy slotSizeOf: hashOop) = 4
	and: [(interpreterProxy isWords: bufOop) and: [(interpreterProxy slotSizeOf: bufOop) = 16]]])
		ifFalse:
			[^interpreterProxy primitiveFail].

	self MD5Transform: (interpreterProxy firstIndexableField: hashOop) _: (interpreterProxy firstIndexableField: bufOop).
	^interpreterProxy methodReturnValue: bufOop! !

!CroquetPlugin methodsFor: 'mesh processing' stamp: 'eem 8/2/2021 17:06'!
primitiveOptimizeVertexIndicesForCacheLocality
	"Given a list of integer indices for rendering a triangle-mesh in indexed-triangles mode, reorganize the indices in-place to provide better vertex cache locality.
	We use Tom Forsyth's algorithm:
		http://home.comcast.net/~tom_forsyth/papers/fast_vert_cache_opt.html
	... and the MIT-licensed implementation by Michael Georgoulpoulos at:
		http://code.google.com/p/vcacne/"
	<export: true>
	| indicesOop byteSize triCount |

	indicesOop := interpreterProxy stackValue: 0.
	(interpreterProxy isWords: indicesOop) ifFalse: [^interpreterProxy primitiveFail].
	
	"Ensure that the number of indices is a multiple of 3."
	byteSize := interpreterProxy byteSizeOf: indicesOop.
	triCount := byteSize // 12.
	(triCount * 12) = byteSize ifFalse: [^interpreterProxy primitiveFail].

	"N.B. as of 8/2021 optimizeVertexIndices is simply a stub."
	(self optimizeVertexIndices: (interpreterProxy firstIndexableField: indicesOop) _: triCount) ~= 0 ifTrue:
		[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnReceiver! !

!CroquetPlugin methodsFor: 'transforms' stamp: 'eem 8/2/2021 17:23'!
primitiveOrthoNormInverseMatrix
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| srcOop dstOop src dst x y z rx ry rz |
	<var: #x type: #double>
	<var: #y type: #double>
	<var: #z type: #double>
	<var: #rx type: #double>
	<var: #ry type: #double>
	<var: #rz type: #double>

	srcOop := interpreterProxy stackValue: 0.
	((interpreterProxy isWords: srcOop) and:[(interpreterProxy slotSizeOf: srcOop) = 16]) ifFalse:
		[^interpreterProxy primitiveFail].
	dstOop := interpreterProxy cloneObject: srcOop.
	dstOop = 0 ifTrue:
		[^interpreterProxy primitiveFail].
	"reload srcOop in case of GC"
	self cppIf: #SPURVM ifFalse: [srcOop := interpreterProxy stackValue: 0].
	src := self cCoerce: (interpreterProxy firstIndexableField: srcOop) to: #'float *'.
	dst := self cCoerce: (interpreterProxy firstIndexableField: dstOop) to: #'float *'.

	"Transpose upper 3x3 matrix"
	"dst at: 0 put: (src at: 0)."	dst at: 1 put: (src at: 4). 	dst at: 2 put: (src at: 8). 
	dst at: 4 put: (src at: 1). 	"dst at: 5 put: (src at: 5)."	dst at: 6 put: (src at: 9). 
	dst at: 8 put: (src at: 2). 	dst at: 9 put: (src at: 6). 	"dst at: 10 put: (src at: 10)."

	"Compute inverse translation vector"
	x := src at: 3.
	y := src at: 7.
	z := src at: 11.
	rx := (x * (dst at: 0)) + (y * (dst at: 1)) + (z * (dst at: 2)).
	ry := (x * (dst at: 4)) + (y * (dst at: 5)) + (z * (dst at: 6)).
	rz := (x * (dst at: 8)) + (y * (dst at: 9)) + (z * (dst at: 10)).

	dst at: 3 put: 0.0 - rx.
	dst at: 7 put: 0.0 - ry.
	dst at: 11 put: 0.0 - rz.

	^interpreterProxy methodReturnValue: dstOop! !

!CroquetPlugin methodsFor: 'transforms' stamp: 'eem 8/2/2021 17:20'!
primitiveTransformDirection
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| x y z rx ry rz matrix vertex v3Oop |
	<var: #x type: #double>
	<var: #y type: #double>
	<var: #z type: #double>
	<var: #rx type: #double>
	<var: #ry type: #double>
	<var: #rz type: #double>

	matrix := self stackMatrix: 1.
	v3Oop := interpreterProxy stackValue: 0.
	(matrix notNil and: [(interpreterProxy isWords: v3Oop) and:[(interpreterProxy slotSizeOf: v3Oop) = 3]]) ifFalse:
		[^interpreterProxy primitiveFail].
	vertex := self cCoerce: (interpreterProxy firstIndexableField: v3Oop) to: #'float *'.

	x := vertex at: 0.
	y := vertex at: 1.
	z := vertex at: 2.

	rx := (x * (matrix at: 0)) + (y * (matrix at: 1)) + (z * (matrix at: 2)).
	ry := (x * (matrix at: 4)) + (y * (matrix at: 5)) + (z * (matrix at: 6)).
	rz := (x * (matrix at: 8)) + (y * (matrix at: 9)) + (z * (matrix at: 10)).

	v3Oop := interpreterProxy cloneObject: v3Oop.
	v3Oop = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoMemory].
	vertex := self cCoerce: (interpreterProxy firstIndexableField: v3Oop) to: #'float *'.

	vertex at: 0 put: rx.
	vertex at: 1 put: ry.
	vertex at: 2 put: rz.

	^interpreterProxy methodReturnValue: v3Oop! !

!CroquetPlugin methodsFor: 'transforms' stamp: 'eem 8/2/2021 17:16'!
primitiveTransformMatrixWithInto
	"Transform two matrices into the third"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| m1 m2 m3 |

	m3 := self stackMatrix: 0.
	m2 := self stackMatrix: 1.
	m1 := self stackMatrix: 2.
	(m1 isNil or: [m2 isNil or: [m3 isNil
	 or: [m2 = m3]]]) ifTrue:
		[^interpreterProxy primitiveFail].
	self transformMatrix: m1 with: m2 into: m3.
	^interpreterProxy methodReturnReceiver! !

!CroquetPlugin methodsFor: 'transforms' stamp: 'eem 8/2/2021 17:20'!
primitiveTransformVector3
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| x y z rx ry rz rw matrix vertex v3Oop |
	<var: #x type: #double>
	<var: #y type: #double>
	<var: #z type: #double>
	<var: #rx type: #double>
	<var: #ry type: #double>
	<var: #rz type: #double>
	<var: #rw type: #double>

	matrix := self stackMatrix: 1.
	v3Oop := interpreterProxy stackValue: 0.
	(matrix notNil and: [(interpreterProxy isWords: v3Oop) and:[(interpreterProxy slotSizeOf: v3Oop) = 3]]) ifFalse:
		[^interpreterProxy primitiveFail].
	vertex := self cCoerce: (interpreterProxy firstIndexableField: v3Oop) to: #'float *'.

	x := vertex at: 0.
	y := vertex at: 1.
	z := vertex at: 2.

	rx := (x * (matrix at: 0)) + (y * (matrix at: 1)) + (z * (matrix at: 2)) + (matrix at: 3).
	ry := (x * (matrix at: 4)) + (y * (matrix at: 5)) + (z * (matrix at: 6)) + (matrix at: 7).
	rz := (x * (matrix at: 8)) + (y * (matrix at: 9)) + (z * (matrix at: 10)) + (matrix at: 11).
	rw := (x * (matrix at: 12)) + (y * (matrix at: 13)) + (z * (matrix at: 14)) + (matrix at: 15).

	v3Oop := interpreterProxy cloneObject: v3Oop.
	v3Oop = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoMemory].
	vertex := self cCoerce: (interpreterProxy firstIndexableField: v3Oop) to: #'float *'.

	rw = 1.0 ifTrue:[
		vertex at: 0 put: rx.
		vertex at: 1 put: ry.
		vertex at: 2 put: rz.
	] ifFalse:[
		rw = 0.0 
			ifTrue:[rw := 0.0]
			ifFalse:[rw := 1.0 / rw].
		vertex at: 0 put: rx * rw.
		vertex at: 1 put: ry * rw.
		vertex at: 2 put: rz * rw.
	].
	^interpreterProxy methodReturnValue: v3Oop! !

!CroquetPlugin methodsFor: 'transforms' stamp: 'eem 01/07/2021 10:40'!
primitiveTransposeMatrix
	| srcOop dstOop src dst |
	<export: true>
	<var: #src declareC:'float *src'>
	<var: #dst declareC:'float *dst'>

	interpreterProxy methodArgumentCount = 0
		ifFalse:[^interpreterProxy primitiveFail].
	srcOop := interpreterProxy stackObjectValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	((interpreterProxy isWords: srcOop) and:[(interpreterProxy slotSizeOf: srcOop) = 16])
		ifFalse:[^interpreterProxy primitiveFail].
	dstOop := interpreterProxy cloneObject: srcOop.
	"reload srcOop in case of GC"
	srcOop := interpreterProxy stackObjectValue: 0.
	src := interpreterProxy firstIndexableField: srcOop.
	dst := interpreterProxy firstIndexableField: dstOop.

	"dst at: 0 put: (src at: 0)."
	dst at: 1 put: (src at: 4). 
	dst at: 2 put: (src at: 8). 
	dst at: 3 put: (src at: 12).

	dst at: 4 put: (src at: 1). 
	"dst at: 5 put: (src at: 5)."
	dst at: 6 put: (src at: 9). 
	dst at: 7 put: (src at: 13).

	dst at: 8 put: (src at: 2). 
	dst at: 9 put: (src at: 6). 
	"dst at: 10 put: (src at: 10)."
	dst at: 11 put: (src at: 14).

	dst at: 12 put: (src at: 3). 
	dst at: 13 put: (src at: 7). 
	dst at: 14 put: (src at: 11). 
	"dst at: 15 put: (src at: 15)."

	interpreterProxy pop: 1.
	^interpreterProxy push: dstOop.
! !

!CroquetPlugin methodsFor: 'transforms' stamp: 'eem 8/2/2021 18:06'!
primitiveTriBoxIntersects
	"Primitive. Answer whether an AABB intersects with a given triangle"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| minCorner maxCorner v0 v1 v2 result |

	v2 := self stackVector3: 0.
	v1 := self stackVector3: 1.
	v0 := self stackVector3: 2.
	maxCorner := self stackVector3: 3.
	minCorner := self stackVector3: 4.

	(v0 isNil or: [v1 isNil or: [v2 isNil
	 or: [maxCorner isNil or: [minCorner isNil]]]]) ifTrue:
		[^interpreterProxy primitiveFail].
	"N.B. as of 8/2021 triBoxOverlap is simply a stub."
	result := self triBoxOverlap: minCorner _: maxCorner _: v0 _: v1 _: v2.
	result < 0 ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnBool: result > 0! !

!CroquetPlugin methodsFor: 'transforms' stamp: 'eem 8/2/2021 17:16'!
stackMatrix: index
	"Load a 4x4 transformation matrix from the interpreter stack.
	Return a pointer to the matrix data if successful, nil otherwise."
	<returnTypeC: #'float *'>
	| oop |
	<inline: #always>
	oop := interpreterProxy stackValue: index.
	^((interpreterProxy isWords: oop) and:[(interpreterProxy slotSizeOf: oop) = 16]) ifTrue:
		[self cCoerce: (interpreterProxy firstIndexableField: oop) to: #'float *']! !

!CroquetPlugin methodsFor: 'transforms' stamp: 'eem 8/2/2021 17:19'!
stackVector3: index
	"Load a Vector3 from the interpreter stack.
	Return a pointer to the float data if successful, nil otherwise."
	<returnTypeC: #'float *'>
	| oop |
	<inline: #always>
	oop := interpreterProxy stackValue: index.
	^((interpreterProxy isWords: oop) and:[(interpreterProxy slotSizeOf: oop) = 3]) ifTrue:
		[self cCoerce: (interpreterProxy firstIndexableField: oop) to: #'float *']! !

!CroquetPlugin methodsFor: 'transforms' stamp: 'eem 8/2/2021 17:28'!
transformMatrix: src with: arg into: dst
	"Transform src with arg into dst.
	It is allowed that src == dst but not arg == dst"
	<var: #src type: #'float *'>
	<var: #arg type: #'float *'>
	<var: #dst type: #'float *'>
	| m1 m2 m3 c1 c2 c3 c4 |
	<var: #m1 type: #'float *'>
	<var: #m2 type: #'float *'>
	<var: #m3 type: #'float *'>

	<var: #c1 type: #double>
	<var: #c2 type: #double>
	<var: #c3 type: #double>
	<var: #c4 type: #double>

	m1 := self cCoerce: src to: #'float *'.
	m2 := self cCoerce: arg to: #'float *'.
	m3 := self cCoerce: dst to: #'float *'.

	0 to: 3 do:[:i|

		"Compute next row"
		c1 :=	((m1 at: 0) * (m2 at: 0)) + ((m1 at: 1) * (m2 at: 4)) +
				((m1 at: 2) * (m2 at: 8)) + ((m1 at: 3) * (m2 at: 12)).

		c2 :=	((m1 at: 0) * (m2 at: 1)) + ((m1 at: 1) * (m2 at: 5)) +
				((m1 at: 2) * (m2 at: 9)) + ((m1 at: 3) * (m2 at: 13)).

		c3 :=	((m1 at: 0) * (m2 at: 2)) + ((m1 at: 1) * (m2 at: 6)) +
				((m1 at: 2) * (m2 at: 10)) + ((m1 at: 3) * (m2 at: 14)).

		c4 :=	((m1 at: 0) * (m2 at: 3)) + ((m1 at: 1) * (m2 at: 7)) +
				((m1 at: 2) * (m2 at: 11)) + ((m1 at: 3) * (m2 at: 15)).

		"Store result"
		m3 at: 0 put: c1.
		m3 at: 1 put: c2.
		m3 at: 2 put: c3.
		m3 at: 3 put: c4.

		"Skip src and dst to next row"
		m1 := m1 + 4.
		m3 := m3 + 4.
	].

	^nil
! !

InterpreterPlugin subclass: #DropPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!DropPlugin commentStamp: '<historical>' prior: 0!
This class defines the necessary primitives for dropping files from the OS onto Squeak.

Implementation notes:
The drop support is really a two phase process. The first thing the OS code needs to do is to signal an event of type EventTypeDragDropFiles to Squeak. This event needs to include the following information (see sq.h for the definition of sqDragDropFilesEvent):
* dragType:
		DragEnter - dragging mouse entered Squeak window
		DragMove - dragging mouse moved within Squeak window
		DragLeave - dragging mouse left Squeak window
		DragDrop - dropped files onto Squeak window
* numFiles:
		The number of files in the drop operation.
* x, y, modifiers:
		Associated mouse state.

When these events are received, the primitives implemented by this plugin come into play. The two primitives can be used to either receive a list of file names or to receive a list of (read-only) file handles. Because drag and drop operations are intended to work in a restricted (plugin) environment, certain security precautions need to be taken:
* Access to the contents of the files (e.g., the file streams) must only be granted after a drop occured. Simply dragging the file over the Squeak window is not enough to grant access.
* Access to the contents of the files after a drop is allowed to bypass the file sandbox and create a read-only file stream directly.
* Access to the names of files can be granted even if the files are only dragged over Squeak (but not dropped). This is so that appropriate user feedback can be given.

If somehow possible, the support code should track the location of the drag-and-drop operation and generate appropriate DragMove type events. While not important right now, it will allow us to integrate OS DnD operations with Morphic DnD operation in a seemless manner.
!

!DropPlugin class methodsFor: 'translation' stamp: 'tpr 5/23/2001 17:09'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!DropPlugin class methodsFor: 'translation' stamp: 'tpr 3/20/2001 12:15'!
requiresPlatformFiles
	"default is ok for most, any plugin needing platform specific files must say so"
	^true! !

!DropPlugin methodsFor: 'initialize' stamp: 'ar (auto pragmas 12/08) 1/10/2001 19:57'!
initialiseModule
	<export: true>
	^self cCode: 'dropInit()' inSmalltalk:[true]! !

!DropPlugin methodsFor: 'primitives' stamp: 'eem 1/4/2021 19:50'!
primitiveDropRequestFileHandle
	"Note: File handle creation needs to be handled by specific support code explicitly bypassing the plugin file sand box."
	<export: true>
	| dropIndex handleOop |
	interpreterProxy methodArgumentCount = 1 ifFalse:
		[^interpreterProxy primitiveFail].
	dropIndex := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifFalse:
		[handleOop := self dropRequestFileHandle: dropIndex.
		 interpreterProxy failed ifFalse:
			[interpreterProxy methodReturnValue: handleOop]]! !

!DropPlugin methodsFor: 'primitives' stamp: 'eem 1/4/2021 19:53'!
primitiveDropRequestFileName
	"Note: File handle creation needs to be handled by specific support code explicitly bypassing the plugin file sand box."
	<export: true>
	| dropIndex |
	interpreterProxy methodArgumentCount = 1 ifFalse:
		[^interpreterProxy primitiveFail].
	dropIndex := interpreterProxy stackIntegerValue: 0.
	"dropRequestFileName returns name or NULL on error; methodReturnString: fails for nil"
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnString: (self dropRequestFileName: dropIndex)]! !

!DropPlugin methodsFor: 'primitives' stamp: 'eem 1/5/2021 12:13'!
primitiveDropRequestURI
	"On platforms that don't support URIs (currently Windows) the idea is to use primitiveDropRequestURI,
	 which will fail, and then fall back on primitiveDropRequestFileName."
	<export: true>
	| dropIndex |
	interpreterProxy methodArgumentCount = 1 ifFalse:
		[^interpreterProxy primitiveFail].
	dropIndex := interpreterProxy stackIntegerValue: 0.
	"dropRequestURI returns name or NULL on error; methodReturnString: fails for nil"
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnString: (self dropRequestURI: dropIndex)]! !

!DropPlugin methodsFor: 'initialize' stamp: 'ar (auto pragmas 12/08) 1/10/2001 19:57'!
shutdownModule
	<export: true>
	^self cCode: 'dropShutdown()' inSmalltalk:[true]! !

InterpreterPlugin subclass: #FFIPlugin
	instanceVariableNames: 'ffiLastError ffiArgSpec ffiArgSpecSize ffiArgHeader ffiRetOop ffiRetClass ffiRetSpec ffiRetSpecSize ffiRetHeader ffiLogEnabled externalFunctionInstSize'
	classVariableNames: ''
	poolDictionaries: 'FFIConstants'
	category: 'VMMaker-Plugins-FFI'!

!FFIPlugin commentStamp: 'eem 11/24/2014 11:00' prior: 0!
This plugin provides access to foreign function interfaces on those platforms that provide such. For example Windows DLLs and unix .so's.  It is obsolete, having been superceded by the ThreadedFFIPlugin.!

!FFIPlugin class methodsFor: 'accessing' stamp: 'ar 11/28/1999 18:55'!
declareCVarsIn: aCCodeGen
	aCCodeGen addHeaderFile: '"sqFFI.h"'! !

!FFIPlugin class methodsFor: 'C support code' stamp: 'tpr 5/23/2001 17:09'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!FFIPlugin class methodsFor: 'translation' stamp: 'eem 12/21/2017 14:27'!
moduleName "FFIPlugin translate"
	"IMPORTANT: IF YOU CHANGE THE NAME OF THIS PLUGIN YOU MUST CHANGE
		Interpreter>>primitiveCalloutToFFI
	TO REFLECT THE CHANGE."
	^'SqueakFFIPrims (Obsolete)'! !

!FFIPlugin class methodsFor: 'translation' stamp: 'tpr 11/29/2000 22:40'!
requiresPlatformFiles
	"this plugin requires platform specific files in order to work"
	^true! !

!FFIPlugin methodsFor: 'primitive support' stamp: 'ar 12/2/1999 21:02'!
atomicTypeOf: value
	^(value bitAnd: FFIAtomicTypeMask) >> FFIAtomicTypeShift! !

!FFIPlugin methodsFor: 'primitive support' stamp: 'tpr 12/21/2005 16:58'!
ffiAddressOf: rcvr startingAt: byteOffset size: byteSize
"return an int of the address of the byteSize slot (byte, short, int, whatever) at byteOffset in rcvr. Nominally intended for use with ExternalAddress objects, this code will work (for obscure historical reasons) with plain Byte or Word Arrays as well. "
	| rcvrClass rcvrSize addr |
	(interpreterProxy isBytes: rcvr) ifFalse:[^interpreterProxy primitiveFail].
	(byteOffset > 0) ifFalse:[^interpreterProxy primitiveFail].
	rcvrClass := interpreterProxy fetchClassOf: rcvr.
	rcvrSize := interpreterProxy byteSizeOf: rcvr.
	rcvrClass == interpreterProxy classExternalAddress ifTrue:[
		(rcvrSize = 4) ifFalse:[^interpreterProxy primitiveFail].
		addr := interpreterProxy fetchPointer: 0 ofObject: rcvr.
		"don't you dare to read from object memory!!"
		(addr == 0 or:[interpreterProxy isInMemory: addr])
			ifTrue:[^interpreterProxy primitiveFail].
	] ifFalse:[
		(byteOffset+byteSize-1 <= rcvrSize)
			ifFalse:[^interpreterProxy primitiveFail].
		addr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: 'int'.
	].
	addr := addr + byteOffset - 1.
	^addr! !

!FFIPlugin methodsFor: 'callout support' stamp: 'mt 8/13/2021 15:30'!
ffiArgByValue: oop
	"Support for generic callout. Prepare an argument by value for a callout."
	| atomicType intValue floatValue |
	<inline: true>
	<var: #floatValue type: 'double'>
	atomicType := self atomicTypeOf: ffiArgHeader.
	"check if the range is valid"
	(atomicType < 0 or:[atomicType > FFITypeDoubleFloat])
		ifTrue:[^self ffiFail: FFIErrorBadAtomicType].
	atomicType < FFITypeSingleFloat ifTrue:["integer types"
		(atomicType >> 1) = (FFITypeSignedInt64 >> 1)
			ifTrue:[intValue := oop] "ffi support code must coerce longlong"
			ifFalse:[intValue := self ffiIntegerValueOf: oop]. "does all the coercions"
		interpreterProxy failed ifTrue:[^self ffiFail: FFIErrorCoercionFailed].
		self dispatchOn: atomicType
			in: #(
				ffiPushVoid:
				ffiPushUnsignedInt:
				ffiPushUnsignedByte:
				ffiPushSignedByte:
				ffiPushUnsignedShort:
				ffiPushSignedShort:
				ffiPushUnsignedInt:
				ffiPushSignedInt:
				ffiPushUnsignedLongLongOop:
				ffiPushSignedLongLongOop:
				ffiPushUnsignedChar:
				ffiPushSignedChar:)
		with: intValue.
	] ifFalse:[
		"either float or double"
		floatValue := self ffiFloatValueOf: oop.
		interpreterProxy failed ifTrue:[^self ffiFail: FFIErrorCoercionFailed].
		atomicType = FFITypeSingleFloat
			ifTrue:[self ffiPushSingleFloat: floatValue]
			ifFalse:[self ffiPushDoubleFloat: floatValue].
	].
	^0! !

!FFIPlugin methodsFor: 'callout support' stamp: 'eem 1/30/2014 15:40'!
ffiArgument: oop Spec: argSpec Class: argClass
	"Callout support. Prepare the given oop as argument.
	argSpec defines the compiled spec for the argument.
	argClass (if non-nil) defines the required (super)class for the argument."
	| valueOop oopClass isStruct nilOop |
	<inline: false>
	oopClass := interpreterProxy fetchClassOf: oop. "Prefetch class (we'll need it)"
	nilOop :=  interpreterProxy nilObject.
	"Do the necessary type checks"
	argClass == nilOop ifFalse:[
		"Type check 1: 
		Is the required class of the argument a subclass of ExternalStructure?"
		(interpreterProxy includesBehavior: argClass 
						ThatOf: interpreterProxy classExternalStructure)
			ifFalse:[^self ffiFail: FFIErrorWrongType]. "Nope. Fail."
		"Type check 2:
		Is the class of the argument a subclass of required class?"
		((nilOop == oop) or:[interpreterProxy includesBehavior: oopClass ThatOf: argClass])
				ifFalse:[^self ffiFail: FFIErrorCoercionFailed]. "Nope. Fail."
		"Okay, we've passed the type check (so far)"
	].

	"Check if oopClass is a subclass of ExternalStructure.
	If this is the case we'll work on it's handle and not the actual oop."
	isStruct := false.
	((interpreterProxy isImmediate: oop) or:[oop == nilOop]) ifFalse:[
		"#isPointers: will fail if oop is immediate so don't even attempt to use it"
		(interpreterProxy isPointers: oop) 
			ifTrue:[isStruct := interpreterProxy includesBehavior: oopClass 
								ThatOf: interpreterProxy classExternalStructure.
					(argClass == nilOop or:[isStruct]) 
						ifFalse:[^self ffiFail: FFIErrorCoercionFailed]].
		"note: the test for #isPointers: above should speed up execution since no pointer type ST objects are allowed in external calls and thus if #isPointers: is true then the arg must be ExternalStructure to work. If it isn't then the code fails anyways so speed isn't an issue"
	].

	"Determine valueOop (e.g., the actual oop to pass as argument)"
	isStruct
		ifTrue:[valueOop := interpreterProxy fetchPointer: 0 ofObject: oop]
		ifFalse:[valueOop := oop].

	"Fetch and check the contents of the compiled spec"
	(interpreterProxy isWords: argSpec)
		ifFalse:[self ffiFail: FFIErrorWrongType. ^nil].
	ffiArgSpecSize := interpreterProxy slotSizeOf: argSpec.
	ffiArgSpecSize = 0 ifTrue:[self ffiFail: FFIErrorWrongType. ^nil].
	ffiArgSpec := self cCoerce: (interpreterProxy firstIndexableField: argSpec) to: 'int'.
	ffiArgHeader := interpreterProxy longAt: ffiArgSpec.

	"Do the actual preparation of the argument"
	"Note: Order is important since FFIFlagStructure + FFIFlagPointer is used to represent 'typedef void* VoidPointer' and VoidPointer really is *struct* not pointer."

	(ffiArgHeader anyMask: FFIFlagStructure) ifTrue:[
		"argument must be ExternalStructure"
		isStruct ifFalse:[^self ffiFail: FFIErrorCoercionFailed].
		(ffiArgHeader anyMask: FFIFlagAtomic) 
			ifTrue:[^self ffiFail: FFIErrorWrongType]. "bad combination"
		^self ffiPushStructureContentsOf: valueOop].

	(ffiArgHeader anyMask: FFIFlagPointer) ifTrue:[
		"no integers (or characters) for pointers please"
		(interpreterProxy isImmediate: oop) 
			ifTrue:[^self ffiFail: FFIErrorIntAsPointer].

		"but allow passing nil pointer for any pointer type"
		oop == interpreterProxy nilObject ifTrue:[^self ffiPushPointer: nil].

		"argument is reference to either atomic or structure type"
		(ffiArgHeader anyMask: FFIFlagAtomic) ifTrue:[
			isStruct "e.g., ExternalData"
				ifTrue:[^self ffiAtomicStructByReference: oop Class: oopClass]
				ifFalse:[^self ffiAtomicArgByReference: oop Class: oopClass].
			"********* NOTE: The above uses 'oop' not 'valueOop' (for ExternalData) ******"
		].

		"Needs to be external structure here"
		isStruct ifFalse:[^self ffiFail: FFIErrorCoercionFailed].
		^self ffiPushPointerContentsOf: valueOop].

	(ffiArgHeader anyMask: FFIFlagAtomic) ifTrue:[
		"argument is atomic value"
		self ffiArgByValue: valueOop.
		^0].
	"None of the above - bad spec"
	^self ffiFail: FFIErrorWrongType! !

!FFIPlugin methodsFor: 'callout support' stamp: 'mt 8/16/2021 09:32'!
ffiAtomicArgByReference: oop Class: oopClass
	"Support for generic callout. Prepare a pointer reference to an atomic type for callout. Note: for type 'void*' we allow either one of ByteArray/String/Symbol or wordVariableSubclass."
	| atomicType isString |
	<inline: true>
	atomicType := self atomicTypeOf: ffiArgHeader.
	(atomicType = FFITypeBool) "No bools on input"
		ifTrue:[^self ffiFail: FFIErrorCoercionFailed].
	((atomicType >> 1) = (FFITypeSignedChar8 >> 1)) ifTrue:["string value (char*)"
		"note: the only types allowed for passing into char* types are
		ByteArray, String, Symbol and *no* other byte indexed objects
		(e.g., CompiledMethod, LargeInteger). We only check for strings
		here and fall through to the byte* check otherwise."
		isString := interpreterProxy 
					includesBehavior: oopClass 
					ThatOf: interpreterProxy classString.
		isString ifTrue:["String/Symbol"
			"Strings must be allocated by the ffi support code"
			^self ffiPushString: (self cCoerce: (interpreterProxy firstIndexableField: oop) to: 'int') OfLength: (interpreterProxy byteSizeOf: oop)].
		"Fall through to byte* test"
		atomicType := FFITypeUnsignedInt8].

	(atomicType = FFITypeVoid or:[(atomicType >> 1) = (FFITypeSignedInt8 >> 1)]) ifTrue:[
		"byte* -- see comment on string above"
		oopClass = interpreterProxy classByteArray ifTrue:["ByteArray"
			^self ffiPushPointer: (self cCoerce: (interpreterProxy firstIndexableField: oop) to:'int')].
		isString := interpreterProxy includesBehavior: oopClass 
					ThatOf: interpreterProxy classString.
		isString ifTrue:["String/Symbol"
			^self ffiPushPointer: (self cCoerce: (interpreterProxy firstIndexableField: oop) to:'int')].
		atomicType = FFITypeVoid ifFalse:[^self ffiFail: FFIErrorCoercionFailed].
		"note: type void falls through"
	].

	(atomicType <= FFITypeSignedInt32 "void/short/int"
		or:[atomicType = FFITypeSingleFloat]) ifTrue:[
			"require a word subclass to work"
			(interpreterProxy isWords: oop) ifTrue:[
				^self ffiPushPointer: (self cCoerce: (interpreterProxy firstIndexableField: oop) to:'int')]].

	^self ffiFail: FFIErrorCoercionFailed.! !

!FFIPlugin methodsFor: 'callout support' stamp: 'ar (auto pragmas 12/08) 1/26/2000 23:43'!
ffiAtomicStructByReference: oop Class: oopClass
	"Support for generic callout. Prepare an external pointer reference to an atomic type for callout."
	| atomicType valueOop |
	<inline: true>
	"must be external data to pass pointers to atomic type"
	oopClass == interpreterProxy classExternalData 
		ifFalse:[^self ffiFail: FFIErrorCoercionFailed].
	atomicType := self atomicTypeOf: ffiArgHeader.
	"no type checks for void pointers"
	atomicType ~= FFITypeVoid ifTrue:[
		self ffiValidateExternalData: oop AtomicType: atomicType.
		interpreterProxy failed ifTrue:[^nil].
	].
	"and push pointer contents"
	valueOop := interpreterProxy fetchPointer: 0 ofObject: oop.
	^self ffiPushPointerContentsOf: valueOop! !

!FFIPlugin methodsFor: 'callout support' stamp: 'ar (auto pragmas 12/08) 12/2/1999 22:19'!
ffiCall: address WithFlags: callType AndTypes: argTypeArray
	"Generic callout. Does the actual work."
	| stackIndex argType argTypes oop nArgs argClass argSpec |
	<inline: true>
	"check if the calling convention is supported"
	(self ffiSupportsCallingConvention: callType)
		ifFalse:[^self ffiFail: FFIErrorCallType].
	argTypes := argTypeArray.
	"Fetch return type and args"
	argType := interpreterProxy fetchPointer: 0 ofObject: argTypes.
	argSpec := interpreterProxy fetchPointer: 0 ofObject: argType.
	argClass := interpreterProxy fetchPointer: 1 ofObject: argType.
	self ffiCheckReturn: argSpec With: argClass.
	interpreterProxy failed ifTrue:[^0]. "cannot return"
	ffiRetOop := argType.
	nArgs := interpreterProxy methodArgumentCount.
	stackIndex := nArgs - 1. "stack index goes downwards"
	1 to: nArgs do:[:i|
		argType := interpreterProxy fetchPointer: i ofObject: argTypes.
		argSpec := interpreterProxy fetchPointer: 0 ofObject: argType.
		argClass := interpreterProxy fetchPointer: 1 ofObject: argType.
		oop := interpreterProxy stackValue: stackIndex.
		self ffiArgument: oop Spec: argSpec Class: argClass.
		interpreterProxy failed ifTrue:[^0]. "coercion failed"
		stackIndex := stackIndex - 1.
	].
	"Go out and call this guy"
	^self ffiCalloutTo: address WithFlags: callType! !

!FFIPlugin methodsFor: 'callout support' stamp: 'ar (auto pragmas 12/08) 12/2/1999 22:19'!
ffiCall: address WithFlags: callType Args: argArray AndTypes: argTypeArray OfSize: nArgs
	"Generic callout. Does the actual work."
	| argType argTypes oop argSpec argClass |
	<inline: true>
	"check if the calling convention is supported"
	(self ffiSupportsCallingConvention: callType)
		ifFalse:[^self ffiFail: FFIErrorCallType].
	argTypes := argTypeArray.
	"Fetch return type and args"
	argType := interpreterProxy fetchPointer: 0 ofObject: argTypes.
	argSpec := interpreterProxy fetchPointer: 0 ofObject: argType.
	argClass := interpreterProxy fetchPointer: 1 ofObject: argType.
	self ffiCheckReturn: argSpec With: argClass.
	interpreterProxy failed ifTrue:[^0]. "cannot return"
	ffiRetOop := argType.
	1 to: nArgs do:[:i|
		argType := interpreterProxy fetchPointer: i ofObject: argTypes.
		argSpec := interpreterProxy fetchPointer: 0 ofObject: argType.
		argClass := interpreterProxy fetchPointer: 1 ofObject: argType.
		oop := interpreterProxy fetchPointer: i-1 ofObject: argArray.
		self ffiArgument: oop Spec: argSpec Class: argClass.
		interpreterProxy failed ifTrue:[^0]. "coercion failed"
	].
	"Go out and call this guy"
	^self ffiCalloutTo: address WithFlags: callType! !

!FFIPlugin methodsFor: 'callout support' stamp: 'ar (auto pragmas 12/08) 1/26/2000 23:11'!
ffiCalloutTo: address WithFlags: callType
	"Go out, call this guy and create the return value"
	| retVal |
	<inline: false>
	"Note: Order is important here since FFIFlagPointer + FFIFlagStructure is used to represent 'typedef void* VoidPointer' and VoidPointer must be returned as pointer *not* as struct"
	(ffiRetHeader anyMask: FFIFlagPointer) ifTrue:[
		retVal := self ffiCallAddressOf: address WithPointerReturn: callType.
		^self ffiCreateReturnPointer: retVal.
	].
	(ffiRetHeader anyMask: FFIFlagStructure) ifTrue:[
		self 
			ffiCallAddressOf: address 
			With: callType 
			Struct: (self cCoerce: ffiRetSpec to:'int*')
			Return: ffiRetSpecSize.
		^self ffiCreateReturnStruct.
	].
	retVal := self ffiCallAddressOf: address With: callType ReturnType: ffiRetHeader.
	^self ffiCreateReturn: retVal.! !

!FFIPlugin methodsFor: 'callout support' stamp: 'eem 1/30/2014 15:38'!
ffiCheckReturn: retSpec With: retClass
	"Make sure we can return an object of the given type"
	<inline: true>
	retClass == interpreterProxy nilObject ifFalse:[
		(interpreterProxy includesBehavior: retClass 
						ThatOf: interpreterProxy classExternalStructure)
			ifFalse:[^self ffiFail: FFIErrorBadReturn]].
	ffiRetClass := retClass.

	(interpreterProxy isWords: retSpec)
		ifFalse:[self ffiFail: FFIErrorWrongType. ^nil].
	ffiRetSpecSize := interpreterProxy slotSizeOf: retSpec.
	ffiRetSpecSize = 0 ifTrue:[self ffiFail: FFIErrorWrongType. ^nil].
	ffiRetSpec := self cCoerce: (interpreterProxy firstIndexableField: retSpec) to: 'int'.
	ffiRetHeader := interpreterProxy longAt: ffiRetSpec.
	(self isAtomicType: ffiRetHeader) ifFalse:[
		(ffiRetClass == interpreterProxy nilObject)
			ifTrue:[^self ffiFail: FFIErrorBadReturn]].
	(self ffiCan: (self cCoerce: ffiRetSpec to:'int*') Return: ffiRetSpecSize)
		ifFalse:[self ffiFail: FFIErrorBadReturn]. "cannot return this type"
	^0! !

!FFIPlugin methodsFor: 'callout support' stamp: 'eem 1/30/2014 15:37'!
ffiContentsOfHandle: oop errCode: errCode
	"Make sure that the given oop is a valid external handle"
	<inline: true>
	(interpreterProxy isBytes: oop)
		ifFalse:[^self ffiFail: errCode].
	((interpreterProxy byteSizeOf: oop) == 4)
		ifFalse:[^self ffiFail: errCode].
	^interpreterProxy fetchPointer: 0 ofObject: oop! !

!FFIPlugin methodsFor: 'callout support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:02'!
ffiCreateLongLongReturn: isSigned
	"Create a longlong return value from a previous call out"
	| lowWord highWord largeClass nBytes largeInt ptr |
	<var: #ptr type:'unsigned char *'>
	lowWord := self ffiLongLongResultLow.
	highWord := self ffiLongLongResultHigh.
	isSigned ifTrue:["check for 32 bit signed"
		(highWord = 0 and:[lowWord >= 0])
			ifTrue:[^interpreterProxy signed32BitIntegerFor: lowWord].
		(highWord = -1 and:[lowWord < 0])
			ifTrue:[^interpreterProxy signed32BitIntegerFor: lowWord].
		"negate value for negative longlong"
		highWord < 0 
			ifTrue:[	largeClass := interpreterProxy classLargeNegativeInteger.
					lowWord := lowWord bitInvert32.
					highWord := highWord bitInvert32.
					lowWord = -1 "e.g., overflow when adding one"
						ifTrue:[highWord := highWord + 1].
					lowWord := lowWord + 1]
			ifFalse:[largeClass := interpreterProxy classLargePositiveInteger].
			"fall through"
	] ifFalse:["check for 32 bit unsigned"
		highWord = 0 ifTrue:[
			^interpreterProxy positive32BitIntegerFor: lowWord].
		largeClass := interpreterProxy classLargePositiveInteger.
		"fall through"
	].
	"Create LargeInteger result"
	nBytes := 8.
	(highWord anyMask: 255 << 24) ifFalse:[
		nBytes := 7.
		highWord < (1 << 16) ifTrue:[nBytes := 6].
		highWord < (1 << 8) ifTrue:[nBytes := 5].
		highWord = 0 ifTrue:[nBytes := 4]].
	"now we know how many bytes to create"
	largeInt := interpreterProxy instantiateClass: largeClass indexableSize: nBytes.
	(interpreterProxy isBytes: largeInt) 
		ifFalse:[^self ffiFail: FFIErrorBadReturn]. "Hossa!!"
	ptr := interpreterProxy firstIndexableField: largeInt.
	4 to: nBytes-1 do:[:i|
		ptr at: i put: (highWord >> (i-4*8) bitAnd: 255)].
	ptr at: 3 put: (lowWord >> 24 bitAnd: 255).
	ptr at: 2 put: (lowWord >> 16 bitAnd: 255).
	ptr at: 1 put: (lowWord >> 8 bitAnd: 255).
	ptr at: 0 put: (lowWord bitAnd: 255).
	^largeInt! !

!FFIPlugin methodsFor: 'callout support' stamp: 'ar (auto pragmas 12/08) 1/26/2000 19:47'!
ffiCreateReturn: retVal
	"Generic callout support. Create an atomic return value from an external function call"
	| atomicType retOop oop |
	<inline: true>
	interpreterProxy failed ifTrue:[^nil].
	atomicType := self atomicTypeOf: ffiRetHeader.
	"void returns self"
	atomicType <= FFITypeVoid ifTrue:[
		^interpreterProxy pop: interpreterProxy methodArgumentCount].
	"everything else returns value"
	interpreterProxy pop: 
		interpreterProxy methodArgumentCount+1.
	interpreterProxy pushRemappableOop: ffiRetClass.
	retOop := self ffiCreateReturnOop: retVal.
	ffiRetClass := interpreterProxy popRemappableOop.
	ffiRetClass == interpreterProxy nilObject ifTrue:[
		"Just return oop"
		^interpreterProxy push: retOop].
	"Otherwise create an instance of external structure and store the return oop"
	interpreterProxy pushRemappableOop: retOop.
	retOop := interpreterProxy instantiateClass: ffiRetClass indexableSize: 0.
	oop := interpreterProxy popRemappableOop.
	interpreterProxy storePointer: 0 ofObject: retOop withValue: oop.
	^interpreterProxy push: retOop.! !

!FFIPlugin methodsFor: 'callout support' stamp: 'mt 8/13/2021 15:27'!
ffiCreateReturnOop: retVal
	"Callout support. Return the appropriate oop for the given atomic value"
	| atomicType shift value mask byteSize |
	atomicType := self atomicTypeOf: ffiRetHeader.
	atomicType = FFITypeBool ifTrue:[
			"Make sure bool honors the byte size requested"
			byteSize := ffiRetHeader bitAnd: FFIStructSizeMask.
			byteSize = 4
				ifTrue:[value := retVal]
				ifFalse:[value := retVal bitAnd: 1 << (byteSize * 8) - 1].
			value = 0
				ifTrue:[^interpreterProxy falseObject]
				ifFalse:[^interpreterProxy trueObject]].
	atomicType <= FFITypeSignedInt32 ifTrue:[
		"these are all generall integer returns"
		atomicType <= FFITypeSignedInt16 ifTrue:[
			"byte/short. first extract partial word, then sign extend"
			shift := (atomicType >> 1) * 8. "# of significant bits"
			value := retVal bitAnd: (1 << shift - 1). 
			(atomicType anyMask: 1) ifTrue:[
				"make the guy signed"
				mask := 1 << (shift-1).
				value := (value bitAnd: mask-1) - (value bitAnd: mask)].
			^interpreterProxy integerObjectOf: value].
		"32bit integer return"
		(atomicType anyMask: 1)
			ifTrue:[^(interpreterProxy signed32BitIntegerFor: retVal)] "signed return"
			ifFalse:[^(interpreterProxy positive32BitIntegerFor: retVal)]]. "unsigned return"

	atomicType < FFITypeSingleFloat ifTrue:[
		"longlong, char"
		(atomicType >> 1) = (FFITypeSignedInt64 >> 1) 
			ifTrue:[^self ffiCreateLongLongReturn: (atomicType anyMask: 1)]
			ifFalse:[^(interpreterProxy 
						fetchPointer: (retVal bitAnd: 255)
						ofObject: interpreterProxy characterTable)]].
	"float return"
	^interpreterProxy floatObjectOf: (self ffiReturnFloatValue).! !

!FFIPlugin methodsFor: 'callout support' stamp: 'mt 8/16/2021 09:32'!
ffiCreateReturnPointer: retVal
	"Generic callout support. Create a pointer return value from an external function call"
	| atomicType retOop oop ptr classOop |
	<var: #ptr type:'int *'>
	interpreterProxy failed ifTrue:[^nil].
	interpreterProxy pop: interpreterProxy methodArgumentCount+1.
	(ffiRetClass == interpreterProxy nilObject) ifTrue:[
		"Create ExternalData upon return"
		atomicType := self atomicTypeOf: ffiRetHeader.
		(atomicType >> 1) = (FFITypeSignedChar8 >> 1) ifTrue:["String return"
			^self ffiReturnCStringFrom: retVal].
		"generate external data"
		interpreterProxy pushRemappableOop: ffiRetOop.
		oop := interpreterProxy 
				instantiateClass: interpreterProxy classExternalAddress 
				indexableSize: 4.
		ptr := interpreterProxy firstIndexableField: oop.
		ptr at: 0 put: retVal.
		interpreterProxy pushRemappableOop: oop. "preserve for gc"
		retOop := interpreterProxy 
				instantiateClass: interpreterProxy classExternalData 
				indexableSize: 0.
		oop := interpreterProxy popRemappableOop. "external address"
		interpreterProxy storePointer: 0 ofObject: retOop withValue: oop.
		oop := interpreterProxy popRemappableOop. "return type"
		interpreterProxy storePointer: 1 ofObject: retOop withValue: oop.
		^interpreterProxy push: retOop.
	].
	"non-atomic pointer return"
	interpreterProxy pushRemappableOop: ffiRetClass. "preserve for gc"
	(ffiRetHeader anyMask: FFIFlagStructure)
		ifTrue:[classOop := interpreterProxy classByteArray]
		ifFalse:[classOop := interpreterProxy classExternalAddress].
	oop := interpreterProxy 
			instantiateClass: classOop
			indexableSize: 4.
	ptr := interpreterProxy firstIndexableField: oop.
	ptr at: 0 put: retVal.
	ffiRetClass := interpreterProxy popRemappableOop. "return class"
	interpreterProxy pushRemappableOop: oop. "preserve for gc"
	retOop := interpreterProxy instantiateClass: ffiRetClass indexableSize: 0.
	oop := interpreterProxy popRemappableOop. "external address"
	interpreterProxy storePointer: 0 ofObject: retOop withValue: oop.
	^interpreterProxy push: retOop.! !

!FFIPlugin methodsFor: 'callout support' stamp: 'ar (auto pragmas 12/08) 12/2/1999 21:33'!
ffiCreateReturnStruct
	"Generic callout support. Create a structure return value from an external function call"
	| retOop structSize oop |
	<inline: true>
	interpreterProxy failed ifTrue:[^nil].
	interpreterProxy pop: interpreterProxy methodArgumentCount+1.
	structSize := ffiRetHeader bitAnd: FFIStructSizeMask.
	interpreterProxy pushRemappableOop: ffiRetClass.
	oop := interpreterProxy 
			instantiateClass: interpreterProxy classByteArray 
			indexableSize: structSize.
	self ffiStore: (self cCoerce: (interpreterProxy firstIndexableField: oop) to:'int') 
		Structure: structSize.
	ffiRetClass := interpreterProxy popRemappableOop.
	interpreterProxy pushRemappableOop: oop. "secure byte array"
	retOop := interpreterProxy instantiateClass: ffiRetClass indexableSize: 0.
	oop := interpreterProxy popRemappableOop.
	interpreterProxy storePointer: 0 ofObject: retOop withValue: oop.
	^interpreterProxy push: retOop.! !

!FFIPlugin methodsFor: 'callout support' stamp: 'eem 11/19/2009 11:00'!
ffiFail: reason
	<inline: false>
	"Map the FFI error code into a primitive error code.  If reason is negative it encodes one of the
	 standard PrimErr... codes, negated to distinguish it from the FFIError codes.  If it is an FFIError...
	 code then add the size of the primitive error table + 2 to disambiguate it from the PrimErr... codes.
	 For historic reasons the FFIError codes range from -1 on up hence adding size + 2 maps them to
	 size of table + 1 on up.  This OFFSET IS undone by ExternalFunction class>>externalCallFailedWith:.
	 Thus we can communicate back both PrimErr.. and FFIError codes.  Complex but necessary in the ThreadedFFIPlugin."
	self ffiSetLastError: reason.
	^interpreterProxy primitiveFailFor:
		(reason >= FFINoCalloutAvailable
			ifTrue: [reason + 2 + (interpreterProxy slotSizeOf: interpreterProxy primitiveErrorTable)]
			ifFalse: [reason negated])! !

!FFIPlugin methodsFor: 'callout support' stamp: 'ar (auto pragmas 12/08) 12/2/1999 21:36'!
ffiFloatValueOf: oop
	"Support for generic callout. Return a float value that is coerced as C would do."
	| oopClass |
	<returnTypeC:'double'>
	oopClass := interpreterProxy fetchClassOf: oop.
	oopClass == interpreterProxy classFloat
		ifTrue:[^interpreterProxy floatValueOf: oop].
	"otherwise try the integer coercions and return its float value"
	^(self ffiIntegerValueOf: oop) asFloat! !

!FFIPlugin methodsFor: 'callout support' stamp: 'ar 11/28/1999 18:06'!
ffiGetLastError
	^ffiLastError! !

!FFIPlugin methodsFor: 'callout support' stamp: 'eem 1/30/2014 15:39'!
ffiIntegerValueOf: oop
	"Support for generic callout. Return an integer value that is coerced as C would do."
	| oopClass |
	<inline: true>
	(interpreterProxy isIntegerObject: oop) ifTrue:[^interpreterProxy integerValueOf: oop].
	oop == interpreterProxy nilObject ifTrue:[^0]. "@@: should we really allow this????"
	oop == interpreterProxy falseObject ifTrue:[^0].
	oop == interpreterProxy trueObject ifTrue:[^1].
	oopClass := interpreterProxy fetchClassOf: oop.
	oopClass == interpreterProxy classFloat
		ifTrue:[^(interpreterProxy floatValueOf: oop) asInteger].
	oopClass == interpreterProxy classCharacter
		ifTrue:[^interpreterProxy characterValueOf: oop].
	oopClass == interpreterProxy classLargePositiveInteger
		ifTrue:[^interpreterProxy positive32BitValueOf: oop].
	^interpreterProxy signed32BitValueOf: oop "<- will fail if not integer"! !

!FFIPlugin methodsFor: 'symbol loading' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:03'!
ffiLoadCalloutAddress: lit
	"Load the address of the foreign function from the given object"
	| addressPtr address ptr |
	<var: #ptr type:'int *'>
	"Lookup the address"
	addressPtr := interpreterProxy fetchPointer: 0 ofObject: lit.
	"Make sure it's an external handle"
	address := self ffiContentsOfHandle: addressPtr errCode: FFIErrorBadAddress.
	interpreterProxy failed ifTrue:[^0].
	address = 0 ifTrue:["Go look it up in the module"
		(interpreterProxy slotSizeOf: lit) < 5 ifTrue:[^self ffiFail: FFIErrorNoModule].
		address := self ffiLoadCalloutAddressFrom: lit.
		interpreterProxy failed ifTrue:[^0].
		"Store back the address"
		ptr := interpreterProxy firstIndexableField: addressPtr.
		ptr at: 0 put: address].
	^address! !

!FFIPlugin methodsFor: 'symbol loading' stamp: 'eem 11/18/2009 16:09'!
ffiLoadCalloutAddressFrom: oop
	"Load the function address for a call out to an external function"
	| module moduleHandle functionName functionLength address |
	<inline: false>
	"First find and load the module"
	module := interpreterProxy fetchPointer: externalFunctionInstSize + 1 ofObject: oop.
	moduleHandle := self ffiLoadCalloutModule: module.
	interpreterProxy failed ifTrue:
		[^0]. "failed"
	"fetch the function name"
	functionName := interpreterProxy fetchPointer: externalFunctionInstSize ofObject: oop.
	(interpreterProxy isBytes: functionName) ifFalse:
		[^self ffiFail: FFIErrorBadExternalFunction].
	functionLength := interpreterProxy byteSizeOf: functionName.
	address := interpreterProxy
					ioLoadSymbol: (self cCoerce: (interpreterProxy firstIndexableField: functionName) to: #int)
					OfLength: functionLength 
					FromModule: moduleHandle.
	(interpreterProxy failed or: [address = 0]) ifTrue:
		[^self ffiFail: FFIErrorAddressNotFound].
	^address! !

!FFIPlugin methodsFor: 'symbol loading' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:03'!
ffiLoadCalloutModule: module
	"Load the given module and return its handle"
	| moduleHandlePtr moduleHandle ffiModuleName moduleLength rcvr theClass ptr |
	<var: #ptr type:'int *'>
	(interpreterProxy isBytes: module) ifTrue:[
		"plain module name"
		ffiModuleName := module.
		moduleLength := interpreterProxy byteSizeOf: ffiModuleName.
		moduleHandle := interpreterProxy ioLoadModule: (self cCoerce: (interpreterProxy firstIndexableField: ffiModuleName) to:'int') OfLength: moduleLength.
		interpreterProxy failed ifTrue:[^self ffiFail: FFIErrorModuleNotFound]. "failed"
		^moduleHandle].
	"Check if the external method is defined in an external library"
	rcvr := interpreterProxy stackValue: interpreterProxy methodArgumentCount.
	theClass := interpreterProxy fetchClassOf: rcvr.
	(interpreterProxy includesBehavior: theClass 
			ThatOf: interpreterProxy classExternalLibrary) ifFalse:[^0].
	"external library"
	moduleHandlePtr := interpreterProxy fetchPointer: 0 ofObject: rcvr.
	moduleHandle := self ffiContentsOfHandle: moduleHandlePtr errCode: FFIErrorBadExternalLibrary.
	interpreterProxy failed ifTrue:[^0].
	moduleHandle = 0 ifTrue:["need to reload module"
		ffiModuleName := interpreterProxy fetchPointer: 1 ofObject: rcvr.
		(interpreterProxy isBytes: ffiModuleName) ifFalse:[^self ffiFail: FFIErrorBadExternalLibrary].
		moduleLength := interpreterProxy byteSizeOf: ffiModuleName.
		moduleHandle := interpreterProxy ioLoadModule: (self cCoerce: (interpreterProxy firstIndexableField: ffiModuleName) to:'int') OfLength: moduleLength.
		interpreterProxy failed ifTrue:[^self ffiFail: FFIErrorModuleNotFound]. "failed"
		"and store back"
		ptr := interpreterProxy firstIndexableField: moduleHandlePtr.
		ptr at: 0 put: moduleHandle].
	^moduleHandle! !

!FFIPlugin methodsFor: 'symbol loading' stamp: 'eem 11/18/2009 16:13'!
ffiLogCallout: lit
	"fetch the function name"
	| functionName |
	ffiLogEnabled ifTrue:[
		functionName := interpreterProxy fetchPointer: externalFunctionInstSize ofObject: lit.
		(interpreterProxy isBytes: functionName) ifFalse:[^nil].
		self ffiLogCall: (interpreterProxy firstIndexableField: functionName)
			OfLength: (interpreterProxy byteSizeOf: functionName).
	].! !

!FFIPlugin methodsFor: 'initialize' stamp: 'ar (auto pragmas 12/08) 7/27/2007 11:18'!
ffiLogCallsTo: fileName
	"This is a special entry point exposed such that client code can 
	enable and disable logging of FFI calls."
	| ok |
	<export: true>
	<var: #fileName type: 'char*'>
	fileName == nil ifTrue:[ "disable logging"
		ok := self ffiLogFileName: nil OfLength: 0.
		ok ifFalse:[^false].
		ffiLogEnabled := false.
	] ifFalse:[ "enable logging"
		ok := self ffiLogFileName: fileName OfLength: (self strlen: fileName).
		ok ifFalse:[^false].
		ffiLogEnabled := true.
	].
	^true! !

!FFIPlugin methodsFor: 'callout support' stamp: 'di (auto pragmas 12/08) 7/4/2004 08:40'!
ffiPushPointerContentsOf: oop
	"Push the contents of the given external structure"
	| ptrValue ptrClass ptrAddress |
	<inline: false>
	ptrValue := oop.
	ptrClass := interpreterProxy fetchClassOf: ptrValue.
	ptrClass == interpreterProxy classExternalAddress ifTrue:[
		ptrAddress := interpreterProxy fetchPointer: 0 ofObject: ptrValue.
		"Don't you dare to pass pointers into object memory"
		(interpreterProxy isInMemory: ptrAddress)
			ifTrue:[^self ffiFail: FFIErrorInvalidPointer].
		^self ffiPushPointer: ptrAddress].
	ptrClass == interpreterProxy classByteArray ifTrue:[
		ptrAddress := self cCoerce: (interpreterProxy firstIndexableField: ptrValue) to: 'int'.
		^self ffiPushPointer: ptrAddress].
	^self ffiFail: FFIErrorBadArg! !

!FFIPlugin methodsFor: 'callout support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:03'!
ffiPushSignedLongLongOop: oop
	"Push a longlong type (e.g., a 64bit integer).
	Note: Coercions from float are *not* supported."
	| lowWord highWord length oopClass negative ptr |
	<var: #ptr type:'unsigned char *'>
	oop == interpreterProxy nilObject 
		ifTrue:[^self ffiPushSignedLong: 0 Long: 0.]. "@@: check this"
	oop == interpreterProxy falseObject
		ifTrue:[^self ffiPushSignedLong: 0 Long: 0].
	oop == interpreterProxy trueObject
		ifTrue:[^self ffiPushSignedLong: 0 Long: 1].
	(interpreterProxy isIntegerObject: oop) ifTrue:[
		lowWord := interpreterProxy integerValueOf: oop.
		lowWord < 0 
			ifTrue:[highWord := -1]
			ifFalse:[highWord := 0].
	] ifFalse:[
		oopClass := interpreterProxy fetchClassOf: oop.
		oopClass == interpreterProxy classLargePositiveInteger 
			ifTrue:[negative := false]
			ifFalse:[oopClass == interpreterProxy classLargeNegativeInteger 
				ifTrue:[negative := true]
				ifFalse:[^self ffiFail: FFIErrorCoercionFailed]].
		(interpreterProxy isBytes: oop) ifFalse:[^self ffiFail: FFIErrorCoercionFailed].
		length := interpreterProxy byteSizeOf: oop.
		length > 8 ifTrue:[^self ffiFail: FFIErrorCoercionFailed].
		lowWord := highWord := 0.
		ptr := interpreterProxy firstIndexableField: oop.
		0 to: (length min: 4)-1 do:[:i|
			lowWord := lowWord + ((ptr at: i) << (i*8))].
		0 to: (length-5) do:[:i|
			highWord := highWord + ((ptr at: i+4) << (i*8))].
		negative ifTrue:[
			lowWord := lowWord bitInvert32.
			highWord := highWord bitInvert32.
			lowWord = -1 "e.g., will overflow when adding one"
				ifTrue:[highWord := highWord + 1].
			lowWord := lowWord + 1].
	].
	^self ffiPushSignedLong: lowWord Long: highWord.! !

!FFIPlugin methodsFor: 'callout support' stamp: 'di (auto pragmas 12/08) 7/4/2004 08:43'!
ffiPushStructureContentsOf: oop
	"Push the contents of the given external structure"
	| ptrValue ptrClass ptrAddress |
	<inline: true>
	ptrValue := oop.
	ptrClass := interpreterProxy fetchClassOf: ptrValue.
	ptrClass == interpreterProxy classExternalAddress ifTrue:[
		ptrAddress := interpreterProxy fetchPointer: 0 ofObject: ptrValue.
		"There is no way we can make sure the structure is valid.
		But we can at least check for attempts to pass pointers to ST memory."
		(interpreterProxy isInMemory: ptrAddress)
			ifTrue:[^self ffiFail: FFIErrorInvalidPointer].
		^self ffiPush: ptrAddress 
				Structure: (self cCoerce: ffiArgSpec to:'int*')
				OfLength: ffiArgSpecSize].
	ptrClass == interpreterProxy classByteArray ifTrue:[
		"The following is a somewhat pessimistic test but I like being sure..."
		(interpreterProxy byteSizeOf: ptrValue) = (ffiArgHeader bitAnd: FFIStructSizeMask)
			ifFalse:[^self ffiFail: FFIErrorStructSize].
		ptrAddress := self cCoerce: (interpreterProxy firstIndexableField: ptrValue) to: 'int'.
		(ffiArgHeader anyMask: FFIFlagPointer) ifFalse:[
			^self ffiPush: ptrAddress 
					Structure: (self cCoerce: ffiArgSpec to: 'int*')
					OfLength: ffiArgSpecSize].
		"If FFIFlagPointer + FFIFlagStructure is set use ffiPushPointer on the contents"
		(ffiArgHeader bitAnd: FFIStructSizeMask) = 4
			ifFalse:[^self ffiFail: FFIErrorStructSize].
		ptrAddress := interpreterProxy fetchPointer: 0 ofObject: ptrValue.
		(interpreterProxy isInMemory: ptrAddress)
			ifTrue:[^self ffiFail: FFIErrorInvalidPointer].
		^self ffiPushPointer: ptrAddress].
	^self ffiFail: FFIErrorBadArg! !

!FFIPlugin methodsFor: 'callout support' stamp: 'nice 4/19/2016 22:08'!
ffiPushUnsignedLongLongOop: oop
	"Push a longlong type (e.g., a 64bit integer).
	Note: Coercions from float are *not* supported."
	| lowWord highWord length ptr |
	<var: #ptr type:'unsigned char *'>
	oop == interpreterProxy nilObject 
		ifTrue:[^self ffiPushUnsignedLong: 0 Long: 0.]. "@@: check this"
	oop == interpreterProxy falseObject 
		ifTrue:[^self ffiPushUnsignedLong: 0 Long: 0].
	oop == interpreterProxy trueObject 
		ifTrue:[^self ffiPushUnsignedLong: 0 Long: 1].
	(interpreterProxy isIntegerObject: oop) ifTrue:[
		lowWord := interpreterProxy integerValueOf: oop.
		lowWord < 0 ifTrue:[^self ffiFail: FFIErrorCoercionFailed].
		highWord := 0.
	] ifFalse:[
		(interpreterProxy isLargePositiveIntegerObject: oop)
			ifFalse:[^interpreterProxy primitiveFail].
		(interpreterProxy isBytes: oop) ifFalse:[^self ffiFail: FFIErrorCoercionFailed].
		length := interpreterProxy byteSizeOf: oop.
		length > 8 ifTrue:[^self ffiFail: FFIErrorCoercionFailed].
		lowWord := highWord := 0.
		ptr := interpreterProxy firstIndexableField: oop.
		0 to: (length min: 4)-1 do:[:i|
			lowWord := lowWord + ((ptr at: i) << (i*8))].
		0 to: (length-5) do:[:i|
			highWord := highWord + ((ptr at: i+4) << (i*8))].
	].
	^self ffiPushUnsignedLong: lowWord Long: highWord.! !

!FFIPlugin methodsFor: 'callout support' stamp: 'ar 11/28/1999 19:25'!
ffiPushVoid: ignored
	"This is a fallback in case somebody tries to pass a 'void' value.
	We could simply ignore the argument but I think it's better to let
	the caller know what he did"
	^self ffiFail: FFIErrorAttemptToPassVoid.! !

!FFIPlugin methodsFor: 'callout support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:04'!
ffiReturnCStringFrom: cPointer
	"Create a Smalltalk string from a zero terminated C string"
	| strLen strOop cString strPtr |
	<var: #cString type:'char *'>
	<var: #strPtr type:'char *'>
	cPointer = nil ifTrue:[
		^interpreterProxy push: interpreterProxy nilObject]. "nil always returs as nil"
	cString := self cCoerce: cPointer to:'char *'.
	strLen := 0.
	[(cString at: strLen) = 0] whileFalse:[strLen := strLen+1].
	strOop := interpreterProxy 
				instantiateClass: interpreterProxy classString 
				indexableSize: strLen.
	strPtr := interpreterProxy firstIndexableField: strOop.
	0 to: strLen-1 do:[:i| strPtr at: i put: (cString at: i)].
	^interpreterProxy push: strOop! !

!FFIPlugin methodsFor: 'callout support' stamp: 'ar 11/28/1999 18:25'!
ffiSetLastError: errCode
	^ffiLastError := errCode! !

!FFIPlugin methodsFor: 'callout support' stamp: 'eem 1/30/2014 15:37'!
ffiValidateExternalData: oop AtomicType: atomicType
	"Validate if the given oop (an instance of ExternalData) can be passed as a pointer to the given atomic type."
	| ptrType specOop spec specType |
	<inline: true>
	ptrType := interpreterProxy fetchPointer: 1 ofObject: oop.
	(interpreterProxy isPointers: ptrType)
		ifFalse:[^self ffiFail: FFIErrorWrongType].
	(interpreterProxy slotSizeOf: ptrType) < 2
		ifTrue:[^self ffiFail: FFIErrorWrongType].
	specOop := interpreterProxy fetchPointer: 0 ofObject: ptrType.
	(interpreterProxy isWords: specOop)
		ifFalse:[^self ffiFail: FFIErrorWrongType].
	(interpreterProxy slotSizeOf: specOop) = 0
		ifTrue:[^self ffiFail: FFIErrorWrongType].
	spec := interpreterProxy fetchPointer: 0 ofObject: specOop.
	(self isAtomicType: spec)
		ifFalse:[^self ffiFail: FFIErrorWrongType].
	specType := self atomicTypeOf: spec.
	specType ~= atomicType ifTrue:[
		"allow for signed/unsigned conversion but nothing else"
		(atomicType > FFITypeBool and:[atomicType < FFITypeSingleFloat])
			ifFalse:[^self ffiFail: FFIErrorCoercionFailed].
		((atomicType >> 1) = (specType >> 1))
			ifFalse:[^self ffiFail: FFIErrorCoercionFailed]].
	^0! !

!FFIPlugin methodsFor: 'initialize' stamp: 'eem 9/30/2010 15:28'!
initialiseModule
	<export: true>
	self initSurfacePluginFunctionPointers.
	"By default, disable logging"
	ffiLogEnabled := false.
	"Get the instSize of ExternalFunction to know whether it contains a cache of the stackSize,
	 and what the offset of ExternalLibraryFunction's functionName and moduleName slots are."
	externalFunctionInstSize := interpreterProxy instanceSizeOf: interpreterProxy classExternalFunction.
	^1! !

!FFIPlugin methodsFor: 'primitive support' stamp: 'ar 12/2/1999 21:04'!
isAtomicType: typeSpec
	^typeSpec anyMask: FFIFlagAtomic! !

!FFIPlugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 7/25/2007 10:41'!
primitiveCallout

	"IMPORTANT: IF YOU CHANGE THE NAME OF THIS METHOD YOU MUST CHANGE
		Interpreter>>primitiveCalloutToFFI
	TO REFLECT THE CHANGE."

	"Perform a function call to a foreign function.
	Only invoked from method containing explicit external call spec."
	| lit address flags argTypes litClass nArgs meth |
	<export: true>
	<inline: false>
	self ffiSetLastError: FFIErrorGenericError. "educated guess if we fail silently"
	lit := nil.
	"Look if the method is itself a callout function"
	meth := interpreterProxy primitiveMethod.
	(interpreterProxy literalCountOf: meth) > 0 ifFalse:[^interpreterProxy primitiveFail].
	lit := interpreterProxy literal: 0 ofMethod: meth.
	litClass := interpreterProxy fetchClassOf: lit.
	(interpreterProxy includesBehavior: litClass 
						ThatOf: interpreterProxy classExternalFunction) 
		ifFalse:[^self ffiFail: FFIErrorNotFunction].
	address := self ffiLoadCalloutAddress: lit.
	interpreterProxy failed ifTrue:[^0].
	"Load and check the other values before we call out"
	flags := interpreterProxy fetchInteger: 1 ofObject: lit.
	interpreterProxy failed ifTrue:[^self ffiFail: FFIErrorBadArgs].
	argTypes := interpreterProxy fetchPointer: 2 ofObject: lit.
	"must be array of arg types"
	(interpreterProxy isArray: argTypes)
		ifFalse:[^self ffiFail: FFIErrorBadArgs].
	nArgs := interpreterProxy slotSizeOf: argTypes.
	"must be argumentCount+1 arg types"
	nArgs = (interpreterProxy methodArgumentCount+1) 
		ifFalse:[^self ffiFail: FFIErrorBadArgs].
	self ffiLogCallout: lit.
	self ffiInitialize. "announce the execution of an external call"
	self ffiCall: address 
		WithFlags: flags 
		AndTypes: argTypes.
	self ffiCleanup. "cleanup temp allocations"
	^0! !

!FFIPlugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 12/5/2003 20:09'!
primitiveCalloutWithArgs
	"Perform a function call to a foreign function.
	Only invoked from ExternalFunction>>invokeWithArguments:"
	| lit address flags argTypes litClass nArgs argArray |
	<export: true>
	<inline: false>
	self ffiSetLastError: FFIErrorGenericError. "educated guess if we fail silently"
	lit := nil.
	"Look if the method is itself a callout function"
	lit := interpreterProxy stackValue: interpreterProxy methodArgumentCount.
	litClass := interpreterProxy fetchClassOf: lit.
	(interpreterProxy includesBehavior: litClass 
						ThatOf: interpreterProxy classExternalFunction) 
		ifFalse:[^self ffiFail: FFIErrorNotFunction].
	address := self ffiLoadCalloutAddress: lit.
	interpreterProxy failed ifTrue:[^nil].
	"Load and check the other values before we call out"
	flags := interpreterProxy fetchInteger: 1 ofObject: lit.
	interpreterProxy failed ifTrue:[^self ffiFail: FFIErrorBadArgs].
	argTypes := interpreterProxy fetchPointer: 2 ofObject: lit.
	"must be array of arg types"
	(interpreterProxy isArray: argTypes) 
		ifFalse:[^self ffiFail: FFIErrorBadArgs].
	nArgs := interpreterProxy slotSizeOf: argTypes.
	(interpreterProxy methodArgumentCount = 1) 
		ifFalse:[^self ffiFail: FFIErrorBadArgs].
	argArray := interpreterProxy stackValue: 0.
	(interpreterProxy isArray: argArray)
		ifFalse:[^self ffiFail: FFIErrorBadArgs].
	nArgs = ((interpreterProxy slotSizeOf: argArray) + 1)
		ifFalse:[^self ffiFail: FFIErrorBadArgs].
	self ffiInitialize. "announce the execution of an external call"
	self ffiCall: address 
		WithFlags: flags 
		Args: argArray
		AndTypes: argTypes
		OfSize: nArgs-1.
	self ffiCleanup. "cleanup temp allocations"
	^0! !

!FFIPlugin methodsFor: 'primitives - surfaces' stamp: 'jcg 6/14/2010 14:08'!
primitiveCreateManualSurface
	"arguments: name(type, stack offset)
		width(Integer, 4)
		height(Integer, 3)
		rowPitch(Integer, 2)
		depth(Integer, 1)
		isMSB(Boolean, 0)"
	| width height rowPitch depth isMSB result |
	<export: true>
	
	interpreterProxy methodArgumentCount == 5 ifFalse: [^interpreterProxy primitiveFail].
	width := interpreterProxy stackIntegerValue: 4.
	height := interpreterProxy stackIntegerValue: 3.
	rowPitch := interpreterProxy stackIntegerValue: 2.
	depth := interpreterProxy stackIntegerValue: 1.
	isMSB := interpreterProxy stackObjectValue: 0.
	isMSB := interpreterProxy booleanValueOf: isMSB. 
	interpreterProxy failed ifTrue: [^nil].
	
	self touch: width; touch: height; touch: rowPitch; touch: depth; touch: isMSB.
	
	result := self cCode: 'createManualSurface(width, height, rowPitch, depth, isMSB)'.
	result < 0 ifTrue: [^interpreterProxy primitiveFail].
	result := interpreterProxy signed32BitIntegerFor: result.
	^interpreterProxy pop: 6 thenPush: result
	! !

!FFIPlugin methodsFor: 'primitives - surfaces' stamp: 'jcg 6/14/2010 14:54'!
primitiveDestroyManualSurface
	"arguments: name(type, stack offset)
		surfaceID(Integer, 0)"
	| surfaceID result |
	<export: true>
	
	interpreterProxy methodArgumentCount == 1 ifFalse: [^interpreterProxy primitiveFail].
	surfaceID := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue: [^nil].
	result := self destroyManualSurface: surfaceID.
	result = 0 ifTrue: [^interpreterProxy primitiveFail].
	^interpreterProxy pop: 1
	! !

!FFIPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:04'!
primitiveFFIAllocate
	"Primitive. Allocate an object on the external heap."
	| byteSize addr oop ptr |
	<export: true>
	<inline: false>
	<var: #ptr type:'int *'>
	byteSize := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	addr := self ffiAlloc: byteSize.
	addr = 0 ifTrue:[^interpreterProxy primitiveFail].
	oop := interpreterProxy 
			instantiateClass: interpreterProxy classExternalAddress 
			indexableSize: 4.
	ptr := interpreterProxy firstIndexableField: oop.
	ptr at: 0 put: addr.
	interpreterProxy pop: 2.
	^interpreterProxy push: oop.
! !

!FFIPlugin methodsFor: 'primitives' stamp: 'eem 9/19/2018 17:31'!
primitiveFFIDoubleAt
	"Return a (signed or unsigned) n byte integer from the given byte offset."
	| byteOffset rcvr addr floatValue |
	<export: true>
	<inline: false>
	<var: #floatValue type: #double>
	byteOffset := interpreterProxy stackIntegerValue: 0.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:[^0].
	addr := self ffiAddressOf: rcvr startingAt: byteOffset size: 8.
	interpreterProxy failed ifTrue:[^0].
	self memcpy: (self addressOf: floatValue) _: addr _: (self sizeof: floatValue).
	interpreterProxy pop: 2.
	^interpreterProxy pushFloat: floatValue
! !

!FFIPlugin methodsFor: 'primitives' stamp: 'eem 9/19/2018 17:31'!
primitiveFFIDoubleAtPut
	"Return a (signed or unsigned) n byte integer from the given byte offset."
	| byteOffset rcvr addr floatValue floatOop |
	<export: true>
	<inline: false>
	<var: #floatValue type: #double>
	floatOop := interpreterProxy stackValue: 0.
	(interpreterProxy isIntegerObject: floatOop)
		ifTrue:[floatValue := self cCoerce: (interpreterProxy integerValueOf: floatOop) to:'double']
		ifFalse:[floatValue := self cCoerce: (interpreterProxy floatValueOf: floatOop) to:'double'].
	byteOffset := interpreterProxy stackIntegerValue: 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	interpreterProxy failed ifTrue:[^0].
	addr := self ffiAddressOf: rcvr startingAt: byteOffset size: 8.
	interpreterProxy failed ifTrue:[^0].
	self memcpy: addr _: (self addressOf: floatValue) _: (self sizeof: floatValue).
	interpreterProxy pop: 3.
	^interpreterProxy push: floatOop! !

!FFIPlugin methodsFor: 'primitives' stamp: 'eem 9/19/2018 17:31'!
primitiveFFIFloatAt
	"Return a (signed or unsigned) n byte integer from the given byte offset."
	| byteOffset rcvr addr floatValue |
	<export: true>
	<inline: false>
	<var: #floatValue type: #float>
	byteOffset := interpreterProxy stackIntegerValue: 0.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:[^0].
	addr := self ffiAddressOf: rcvr startingAt: byteOffset size: 4.
	interpreterProxy failed ifTrue:[^0].
	self memcpy: (self addressOf: floatValue) _: addr _: (self sizeof: floatValue).
	interpreterProxy pop: 2.
	^interpreterProxy pushFloat: floatValue! !

!FFIPlugin methodsFor: 'primitives' stamp: 'eem 9/19/2018 17:31'!
primitiveFFIFloatAtPut
	"Return a (signed or unsigned) n byte integer from the given byte offset."
	| byteOffset rcvr addr floatValue floatOop |
	<export: true>
	<inline: false>
	<var: #floatValue type: #float>
	floatOop := interpreterProxy stackValue: 0.
	(interpreterProxy isIntegerObject: floatOop)
		ifTrue:[floatValue := self cCoerce: (interpreterProxy integerValueOf: floatOop) to:'float']
		ifFalse:[floatValue := self cCoerce: (interpreterProxy floatValueOf: floatOop) to:'float'].
	byteOffset := interpreterProxy stackIntegerValue: 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	interpreterProxy failed ifTrue:[^0].
	addr := self ffiAddressOf: rcvr startingAt: byteOffset size: 4.
	interpreterProxy failed ifTrue:[^0].
	self memcpy: addr _: (self addressOf: floatValue) _: (self sizeof: floatValue).
	interpreterProxy pop: 3.
	^interpreterProxy push: floatOop! !

!FFIPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:05'!
primitiveFFIFree
	"Primitive. Free the object pointed to on the external heap."
	| addr oop ptr |
	<export: true>
	<inline: false>
	<var: #ptr type:'int *'>
	oop := interpreterProxy stackObjectValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	(interpreterProxy fetchClassOf: oop) = (interpreterProxy classExternalAddress)
		ifFalse:[^interpreterProxy primitiveFail].
	(interpreterProxy byteSizeOf: oop) = 4
		ifFalse:[^interpreterProxy primitiveFail].
	ptr := interpreterProxy firstIndexableField: oop.
	addr := ptr at: 0.
	"Don't you dare to free Squeak's memory!!"
	(addr = 0 or:[interpreterProxy isInMemory: addr])
		ifTrue:[^interpreterProxy primitiveFail].
	self ffiFree: addr.
	^ptr at: 0 put: 0. "cleanup"
! !

!FFIPlugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 11/28/1999 19:26'!
primitiveFFIGetLastError
	"Primitive. Return the error code from a failed call to the foreign function interface."
	<export: true>
	<inline: false>
	interpreterProxy pop: 1.
	^interpreterProxy pushInteger: self ffiGetLastError.! !

!FFIPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas dtl 2010-09-27) 4/6/2006 11:33'!
primitiveFFIIntegerAt
	"Return a (signed or unsigned) n byte integer from the given byte offset."
	| isSigned byteSize byteOffset rcvr addr value mask |
	<export: true>
	<inline: false>
	isSigned := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
	byteSize := interpreterProxy stackIntegerValue: 1.
	byteOffset := interpreterProxy stackIntegerValue: 2.
	rcvr := interpreterProxy stackObjectValue: 3.
	interpreterProxy failed ifTrue:[^0].
	(byteOffset > 0 and:[byteSize = 1 or:[byteSize = 2 or:[byteSize = 4]]])
		ifFalse:[^interpreterProxy primitiveFail].
	addr := self ffiAddressOf: rcvr startingAt: byteOffset size: byteSize.
	interpreterProxy failed ifTrue:[^0].
	byteSize < 4 ifTrue:[
		"short/byte"
		byteSize = 1 
			ifTrue:[value := interpreterProxy byteAt: addr]
			ifFalse:[	value := self cCode: '*((unsigned short int *) addr)' 
								inSmalltalk: [interpreterProxy shortAt: addr]].
		isSigned ifTrue:["sign extend value"
			mask := 1 << (byteSize * 8 - 1).
			value := (value bitAnd: mask-1) - (value bitAnd: mask)].
		"note: byte/short never exceed SmallInteger range"
		value := interpreterProxy integerObjectOf: value.
	] ifFalse:[
		"general 32 bit integer"
		value := interpreterProxy longAt: addr.
		isSigned
			ifTrue:[value := interpreterProxy signed32BitIntegerFor: value]
			ifFalse:[value := interpreterProxy positive32BitIntegerFor: value].
	].
	interpreterProxy pop: 4.
	^interpreterProxy push: value
! !

!FFIPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/21/2005 17:03'!
primitiveFFIIntegerAtPut
	"Store a (signed or unsigned) n byte integer at the given byte offset."
	| isSigned byteSize byteOffset rcvr addr value max valueOop |
	<export: true>
	<inline: false>
	isSigned := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
	byteSize := interpreterProxy stackIntegerValue: 1.
	valueOop := interpreterProxy stackValue: 2.
	byteOffset := interpreterProxy stackIntegerValue: 3.
	rcvr := interpreterProxy stackObjectValue: 4.
	interpreterProxy failed ifTrue:[^0].
	(byteOffset > 0 and:[byteSize = 1 or:[byteSize = 2 or:[byteSize = 4]]])
		ifFalse:[^interpreterProxy primitiveFail].
	addr := self ffiAddressOf: rcvr startingAt: byteOffset size: byteSize.
	interpreterProxy failed ifTrue:[^0].
	isSigned 
		ifTrue:[value := interpreterProxy signed32BitValueOf: valueOop]
		ifFalse:[value := interpreterProxy positive32BitValueOf: valueOop].
	interpreterProxy failed ifTrue:[^0].
	byteSize < 4 ifTrue:[
		isSigned ifTrue:[
			max := 1 << (8 * byteSize - 1).
			value >= max ifTrue:[^interpreterProxy primitiveFail].
			value < (0 - max) ifTrue:[^interpreterProxy primitiveFail].
		] ifFalse:[
			value >= (1 << (8*byteSize)) ifTrue:[^interpreterProxy primitiveFail].
		].
		"short/byte"
		byteSize = 1 
			ifTrue:[interpreterProxy byteAt: addr put: value]
			ifFalse:[	self cCode: '*((short int *) addr) = value' 
						inSmalltalk: [interpreterProxy shortAt: addr put: value]].
	] ifFalse:[interpreterProxy longAt: addr put: value].
	interpreterProxy pop: 5.
	^interpreterProxy push: valueOop.! !

!FFIPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:05'!
primitiveForceLoad
	"Primitive. Force loading the receiver (an instance of ExternalLibrary)."
	| rcvr theClass moduleHandlePtr moduleHandle ffiModuleName moduleLength ptr |
	<export: true>
	<inline: false>
	<var: #ptr type:'int *'>
	self ffiSetLastError: FFIErrorGenericError. "educated guess if we fail silently"
	interpreterProxy methodArgumentCount = 0
		ifFalse:[^interpreterProxy primitiveFail].
	rcvr := interpreterProxy stackValue: 0.
	theClass := interpreterProxy fetchClassOf: rcvr.
	(interpreterProxy includesBehavior: theClass 
			ThatOf: interpreterProxy classExternalLibrary) 
				ifFalse:[^self ffiFail: FFIErrorBadExternalLibrary].
	moduleHandlePtr := interpreterProxy fetchPointer: 0 ofObject: rcvr.
	moduleHandle := self ffiContentsOfHandle: moduleHandlePtr errCode: FFIErrorBadExternalLibrary.
	interpreterProxy failed ifTrue:[^0].
	ffiModuleName := interpreterProxy fetchPointer: 1 ofObject: rcvr.
	(interpreterProxy isBytes: ffiModuleName) 
		ifFalse:[^self ffiFail: FFIErrorBadExternalLibrary].
	moduleLength := interpreterProxy byteSizeOf: ffiModuleName.
	moduleHandle := interpreterProxy ioLoadModule: (self cCoerce: (interpreterProxy firstIndexableField: ffiModuleName) to:'int') OfLength: moduleLength.
	interpreterProxy failed ifTrue:[^self ffiFail: FFIErrorModuleNotFound]. "failed"
	"and store back"
	ptr := interpreterProxy firstIndexableField: moduleHandlePtr.
	ptr at: 0 put: moduleHandle.
	^0 "done"! !

!FFIPlugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 7/27/2007 11:17'!
primitiveLogCallsTo
	"Enable logging of FFI calls by providing it with a log file name."
	| logFile ok |
	<export: true>
	interpreterProxy methodArgumentCount = 1 
		ifFalse:[^interpreterProxy primitiveFail].
	logFile := interpreterProxy stackObjectValue: 0.
	logFile == interpreterProxy nilObject ifTrue:[ "disable logging"
		ok := self ffiLogFileName: nil OfLength: 0.
		ok ifFalse:[^interpreterProxy primitiveFail].
		ffiLogEnabled := false.
	] ifFalse:[ "enable logging"
		(interpreterProxy isBytes: logFile) ifFalse:[^interpreterProxy primitiveFail].
		ok := self ffiLogFileName: (interpreterProxy firstIndexableField: logFile)
					OfLength: (interpreterProxy byteSizeOf: logFile).
		ok ifFalse:[^interpreterProxy primitiveFail].
		ffiLogEnabled := true.
	].
	^interpreterProxy pop: 1. "pop arg; return rcvr"
! !

!FFIPlugin methodsFor: 'primitives - surfaces' stamp: 'jcg 6/14/2010 14:14'!
primitiveSetManualSurfacePointer
	"Create a 'manual surface' data-structure.  See the ExternalForm class in the FFI package for example usage."
	"arguments: name(type, stack offset)
		surfaceID(Integer, 1)
		ptr(uint32, 0)"
	| surfaceID ptr result |
	<export: true>
	<var: #ptr type: #'unsigned int'>
	
	interpreterProxy methodArgumentCount == 2 ifFalse: [^interpreterProxy primitiveFail].
	surfaceID := interpreterProxy stackIntegerValue: 1.
	ptr := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifTrue: [^nil].

	self touch: surfaceID; touch: ptr.
	
	result := self cCode: 'setManualSurfacePointer(surfaceID, (void*)ptr)'.
	result = 0 ifTrue: [^interpreterProxy primitiveFail].
	^interpreterProxy pop: 2
	! !

InterpreterPlugin subclass: #FFTPlugin
	instanceVariableNames: 'nu fftSize sinTable sinTableSize permTable permTableSize realData realDataSize imagData imagDataSize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!FFTPlugin commentStamp: '<historical>' prior: 0!
FFTPlugin is an example  of how plugins are written. It shows the use of FloatArray for heavy numerical stuff as well as the simulation of plugins from Squeak.

See also:
		FFT pluginTransformData:
!

!FFTPlugin class methodsFor: 'translation to C' stamp: 'eem 3/5/2009 13:18'!
declareCVarsIn: cg
	cg var: #sinTable type: #'float *'.
	cg var: #realData type:  #'float *'.
	cg var: #imagData type: #'float *'.
	cg var: #permTable type: #'unsigned int *'! !

!FFTPlugin methodsFor: 'private' stamp: 'ar (auto pragmas 12/08) 9/16/1998 21:40'!
checkedFloatPtrOf: oop
	"Return the first indexable word of oop which is assumed to be variableWordSubclass"
	<returnTypeC:'float *'>
	interpreterProxy success: (interpreterProxy isWords: oop).
	interpreterProxy failed ifTrue:[^0].
	^self cCoerce: (interpreterProxy firstIndexableField: oop) to:'float *'! !

!FFTPlugin methodsFor: 'private' stamp: 'ar (auto pragmas 12/08) 9/16/1998 21:40'!
checkedWordPtrOf: oop
	"Return the first indexable word of oop which is assumed to be variableWordSubclass"
	<returnTypeC:'unsigned int *'>
	interpreterProxy success: (interpreterProxy isWords: oop).
	^self cCoerce: (interpreterProxy firstIndexableField: oop) to: 'unsigned int *'! !

!FFTPlugin methodsFor: 'private' stamp: 'eem 9/5/2021 10:51'!
loadFFTFrom: fftOop
	| oop |
	"N.B. no need to check for an object here because there is a guard in
	 slotSizeOf: and slotSizeOf: anImmediate = 0"
	interpreterProxy success: (interpreterProxy slotSizeOf: fftOop) >= 6.
	interpreterProxy failed ifTrue:[^false].
	nu := interpreterProxy fetchInteger: 0 ofObject: fftOop.
	fftSize := interpreterProxy fetchInteger: 1 ofObject: fftOop.

	oop := interpreterProxy fetchPointer: 2 ofObject: fftOop.
	sinTableSize := interpreterProxy stSizeOf: oop.
	sinTable := self checkedFloatPtrOf: oop.

	oop := interpreterProxy fetchPointer: 3 ofObject: fftOop.
	permTableSize := interpreterProxy stSizeOf: oop.
	permTable := self checkedWordPtrOf: oop.

	oop := interpreterProxy fetchPointer: 4 ofObject: fftOop.
	realDataSize := interpreterProxy stSizeOf: oop.
	realData := self checkedFloatPtrOf: oop.

	oop := interpreterProxy fetchPointer: 5 ofObject: fftOop.
	imagDataSize := interpreterProxy stSizeOf: oop.
	imagData := self checkedFloatPtrOf: oop.

	"Check assumptions about sizes"
	interpreterProxy success:
		(1 << nu = fftSize) & 
		(fftSize // 4 + 1 = sinTableSize) & 
		(fftSize = realDataSize) & 
		(fftSize = imagDataSize) &
		(realDataSize = imagDataSize).

	^interpreterProxy failed == false! !

!FFTPlugin methodsFor: 'transforming' stamp: 'eem 9/5/2021 10:46'!
permuteData
	| i end a b tmp |
	<var: 'tmp' type:  #float>
	i := 0.
	end := permTableSize.
	[i < end] whileTrue:
		[a := (permTable at: i) - 1.
		b := (permTable at: i+1) - 1.

		(a < realDataSize and:[b < realDataSize]) ifFalse:[^interpreterProxy success: false].

		tmp := realData at: a.
		realData at: a put: (realData at: b).
		realData at: b put: tmp.

		tmp := imagData at: a.
		imagData at: a put: (imagData at: b).
		imagData at: b put: tmp.

		i := i + 2]! !

!FFTPlugin methodsFor: 'primitives' stamp: 'eem 9/5/2021 10:50'!
primitiveFFTPermuteData
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	(self loadFFTFrom: (interpreterProxy stackValue: 0)) ifFalse:[^nil].
	self permuteData.
	interpreterProxy failed ifTrue:
		"permuteData went wrong. Do the permutation again -- this will restore the original order"
		[self permuteData].! !

!FFTPlugin methodsFor: 'primitives' stamp: 'eem 9/5/2021 10:50'!
primitiveFFTScaleData
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	(self loadFFTFrom: (interpreterProxy stackValue: 0)) ifFalse:[^nil].
	self scaleData! !

!FFTPlugin methodsFor: 'primitives' stamp: 'eem 9/5/2021 10:50'!
primitiveFFTTransformData
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| forward |
	forward := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
	(self loadFFTFrom: (interpreterProxy stackValue: 1)) ifFalse:[^nil].
	self transformData: forward.
	interpreterProxy failed ifFalse:
		[interpreterProxy pop: 1] "Leave rcvr on stack"! !

!FFTPlugin methodsFor: 'transforming' stamp: 'eem 9/5/2021 10:46'!
scaleData
	"Scale all elements by 1/n when doing inverse"
	| realN |
	<var: 'realN' type:  #float>
	fftSize <= 1 ifTrue:[^nil].
	realN := self cCoerce: (1.0 / (self cCoerce: fftSize to: #double)) to: #float.
	0 to: fftSize-1 do:
		[:i |
		realData at: i put: (realData at: i) * realN.
		imagData at: i put: (imagData at: i) * realN]! !

!FFTPlugin methodsFor: 'transforming' stamp: 'ar 9/16/1998 20:21'!
transformData: forward
	self permuteData.
	interpreterProxy failed ifTrue:[
		"permuteData went wrong. Do the permutation again -- this will restore the original order"
		self permuteData.
		^nil].
	self transformForward: forward.
	forward ifFalse: [self scaleData]  "Reverse transform must scale to be an inverse"! !

!FFTPlugin methodsFor: 'transforming' stamp: 'eem 9/5/2021 10:46'!
transformForward: forward
	| lev lev1 ip theta realU imagU realT imagT i fftSize2 fftSize4 fftScale ii |
	<var: 'realU' type: #float>
	<var: 'realT' type: #float>
	<var: 'imagU' type: #float>
	<var: 'imagT' type: #float>
	fftSize2 := fftSize // 2.
	fftSize4 := fftSize // 4.
	1 to: nu do:
		[:level |
		lev := 1 << level.
		lev1 := lev // 2.
		fftScale := fftSize // lev.
		1 to: lev1 do:
			[:j |
			theta := j-1 * fftScale.   "pi * (j-1) / lev1 mapped onto 0..n/2"
			theta < fftSize4  "Compute U, the complex multiplier for each level"
				ifTrue:
					[realU := sinTable at: sinTableSize - theta - 1.
					imagU := sinTable at: theta]
				ifFalse:
					[realU := 0.0 - (sinTable at: theta - fftSize4).
					imagU := sinTable at: fftSize2 - theta].
			forward ifFalse: [imagU := 0.0 - imagU].
"
			Here is the inner loop...
			j to: n by: lev do:
				[:i |   hand-transformed to whileTrue...
"
			i := j.
			[i <= fftSize] whileTrue:
				[ip := i + lev1 - 1.
				ii := i-1.
				realT := ((realData at: ip) * realU) - ((imagData at: ip) * imagU).
				imagT := ((realData at: ip) * imagU) + ((imagData at: ip) * realU).
				realData at: ip put: (realData at: ii) - realT.
				imagData at: ip put: (imagData at: ii) - imagT.
				realData at: ii put: (realData at: ii) + realT.
				imagData at: ii put: (imagData at: ii) + imagT.
				i := i + lev]]].! !

InterpreterPlugin subclass: #FilePlugin
	instanceVariableNames: 'sCCPfn sCDPfn sCGFTfn sCLPfn sCSFTfn sDFAfn sCDFfn sCOFfn sCRFfn sHFAfn'
	classVariableNames: 'DirBadPath DirEntryFound DirNoMoreEntries'
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!FilePlugin commentStamp: 'tpr 5/5/2003 12:01' prior: 0!
Provide access to the host machine file system. Requires both the Cross platform support files from platforms - Cross - plugins - FilePlugin (or some suitable replacement) and the platform specific fils from platforms - {your platform} - plugins - FilePlugin.!

!FilePlugin class methodsFor: 'translation' stamp: 'eem 1/7/2018 10:15'!
declareCVarsIn: aCCodeGenerator

	"Hmmm, we could use (self instVarNames select: [:ivn| ivn first = $s and: [ivn second isUppercase]])"
	self declareC:  #('sCCPfn' 'sCDFfn' 'sCDPfn' 'sCGFTfn' 'sCLPfn' 'sCOFfn' 'sCRFfn' 'sCSFTfn' 'sDFAfn' 'sHFAfn')
		as: #'void *'
		in: aCCodeGenerator.
	aCCodeGenerator addHeaderFile: '"FilePlugin.h"'! !

!FilePlugin class methodsFor: 'translation' stamp: 'tpr 5/23/2001 17:09'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!FilePlugin class methodsFor: 'class initialization' stamp: 'ar 5/12/2000 16:04'!
initialize
	"FilePlugin initialize"
	DirEntryFound := 0.
	DirNoMoreEntries := 1.
	DirBadPath := 2.! !

!FilePlugin class methodsFor: 'translation' stamp: 'tpr 7/4/2001 15:12'!
requiresCrossPlatformFiles
	"this plugin requires cross platform files in order to work"
	^true! !

!FilePlugin class methodsFor: 'translation' stamp: 'tpr 11/29/2000 22:37'!
requiresPlatformFiles
	"this plugin requires platform specific files in order to work"
	^true! !

!FilePlugin class methodsFor: 'simulation' stamp: 'ar 5/11/2000 22:13'!
simulatorClass
	^FilePluginSimulator! !

!FilePlugin methodsFor: 'private' stamp: 'eem 1/15/2021 18:41'!
connectToFd: fd write: writeFlag
	"Connect to the supplied file descriptor. Answer the file oop.
	On POSIX platforms this translates to fdopen().
	writeFlag must be compatible with the existing file access."
	| file fileOop |
	<var: 'fd' type: #int>
	fileOop := interpreterProxy instantiateClass: interpreterProxy classByteArray indexableSize: self fileRecordSize.
	file := self fileValueOf: fileOop.
	interpreterProxy failed ifFalse:
		[self sqConnect: file ToFile: fd Descriptor: writeFlag].
	^fileOop! !

!FilePlugin methodsFor: 'private' stamp: 'eem 1/15/2021 18:42'!
connectToFile: cfile write: writeFlag
	"Open the FILE* as file. Answer the file oop.
	writeFlag must be compatible with the existing file access."
	| file fileOop |
	<var: 'cfile' type: #'void *'>
	fileOop := interpreterProxy instantiateClass: interpreterProxy classByteArray indexableSize: self fileRecordSize.
	file := self fileValueOf: fileOop.
	interpreterProxy failed ifFalse:
		[self sqConnect: file To: cfile File: writeFlag].
	^fileOop! !

!FilePlugin methodsFor: 'directory primitives' stamp: 'eem 10/29/2018 12:35'!
dir_Delimitor
	<doNotGenerate>
	^(Smalltalk classNamed: #FileSystem)
		ifNotNil: [:fileSystem| fileSystem disk delimiter asInteger]
		ifNil: [FileDirectory pathNameDelimiter asInteger]! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:50'!
fileOpenName: nameIndex size: nameSize write: writeFlag secure: secureFlag
	"Open the named file, possibly checking security. Answer the file oop."
	| file fileOop okToOpen |
	<var: 'nameIndex' type: #'char *'>
	<export: true>
	fileOop := interpreterProxy instantiateClass: interpreterProxy classByteArray indexableSize: self fileRecordSize.
	file := self fileValueOf: fileOop.
	interpreterProxy failed ifFalse:
		[secureFlag ifTrue:
			"If the security plugin can be loaded, use it to check for permission.
			 If not, assume it's ok"
			[sCOFfn ~= 0  ifTrue:
				[okToOpen := self cCode: '((sqInt (*) (char *, sqInt, sqInt)) sCOFfn)(nameIndex, nameSize, writeFlag)'.
				 okToOpen ifFalse:
					[^interpreterProxy primitiveFailFor: PrimErrOperationFailed]]].
		 self sqFileOpen: file _: nameIndex _: nameSize _: writeFlag].
	^fileOop! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:39'!
fileOpenNewName: nameIndex size: nameSize secure: secureFlag 
	"Open the new named file, possibly checking security. Answer the file oop."
	| file fileOop okToOpen exists |
	<var: 'nameIndex' type: #'char *'>
	<export: true>
	fileOop := interpreterProxy instantiateClass: interpreterProxy classByteArray indexableSize: self fileRecordSize.
	file := self fileValueOf: fileOop.
	interpreterProxy failed ifFalse:
		[secureFlag ifTrue:
			"If the security plugin can be loaded, use it to check for permission.
			 If not, assume it's ok"
			[sCOFfn ~= 0  ifTrue:
				[okToOpen := self cCode: '((sqInt (*) (char *, sqInt, sqInt)) sCOFfn)(nameIndex, nameSize, true)'.
				 okToOpen ifFalse:
					[^interpreterProxy primitiveFailFor: PrimErrOperationFailed]]].
		 exists := false.
		 self sqFileOpenNew: file _: nameIndex _: nameSize _: (self addressOf: exists).
		 (interpreterProxy failed or: [exists]) ifTrue: "worry not; it is sqFileOpenNew's responsibility to close the file if it exists"
			[interpreterProxy primitiveFailFor: PrimErrInappropriate]].
	^fileOop! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 12/12/2012 11:43'!
fileRecordSize
	"Return the size of a Smalltalk file record in bytes."
	<static: false>
	^self sizeof: #SQFile! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:21'!
fileValueOf: objectPointer
	"Return a pointer to the first byte of of the file record within the given Smalltalk object, or nil if objectPointer is not a file record."
	<returnTypeC: #'SQFile *'>
	<static: false>
	(((interpreterProxy isBytes: objectPointer)
	 and: [(interpreterProxy byteSizeOf: objectPointer) = self fileRecordSize])) ifFalse:
		[interpreterProxy primitiveFailFor: PrimErrBadArgument. ^nil].
	^interpreterProxy firstIndexableField: objectPointer! !

!FilePlugin methodsFor: 'initialize-release' stamp: 'eem 10/29/2018 15:07'!
initialiseModule
	<export: true>
	sCCPfn := interpreterProxy ioLoadFunction: 'secCanCreatePathOfSize' From: 'SecurityPlugin'.
	sCDPfn := interpreterProxy ioLoadFunction: 'secCanDeletePathOfSize' From: 'SecurityPlugin'.
	sCGFTfn := interpreterProxy ioLoadFunction: 'secCanGetFileTypeOfSize' From: 'SecurityPlugin'.
	sCLPfn := interpreterProxy ioLoadFunction: 'secCanListPathOfSize' From: 'SecurityPlugin'.
	sCSFTfn := interpreterProxy ioLoadFunction: 'secCanSetFileTypeOfSize' From: 'SecurityPlugin'.
	sDFAfn := interpreterProxy ioLoadFunction: 'secDisableFileAccess' From: 'SecurityPlugin'.
	sCDFfn := interpreterProxy ioLoadFunction: 'secCanDeleteFileOfSize' From: 'SecurityPlugin'.
	sCOFfn := interpreterProxy ioLoadFunction: 'secCanOpenFileOfSizeWritable' From: 'SecurityPlugin'.
	sCRFfn := interpreterProxy ioLoadFunction: 'secCanRenameFileOfSize' From: 'SecurityPlugin'.
	sHFAfn := interpreterProxy ioLoadFunction: 'secHasFileAccess' From: 'SecurityPlugin'.
	^self sqFileInit! !

!FilePlugin methodsFor: 'directory primitives' stamp: 'eem 5/16/2017 18:45'!
makeDirEntryName: entryName size: entryNameSize
	createDate: createDate modDate: modifiedDate
		isDir: dirFlag fileSize: fileSize
	<var: 'entryName' type: #'char *'>
	<var: 'fileSize' type: #squeakFileOffsetType>
	| modDateOop createDateOop nameString results stringPtr fileSizeOop |
	<var: 'stringPtr' type: #'char *'>

	"allocate storage for results, remapping newly allocated
	 oops if required in case GC happens during allocation"
	results := interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 5.
	self remapOop: results in:
		[nameString := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: entryNameSize.
		 self remapOop: nameString in:
			[createDateOop := interpreterProxy positive32BitIntegerFor: createDate.
			 self remapOop: createDateOop in:
				[modDateOop := interpreterProxy positive32BitIntegerFor: modifiedDate.
				 self remapOop: modDateOop in:
					[fileSizeOop := interpreterProxy positive64BitIntegerFor: fileSize]]]].

	"copy name into Smalltalk string"
	stringPtr := interpreterProxy firstIndexableField: nameString.
	0 to: entryNameSize - 1 do:
		[ :i |
		self cCode: [stringPtr at: i put: (entryName at: i)]
			inSmalltalk: [interpreterProxy storeByte: i ofObject: nameString withValue: (entryName at: i+1) asciiValue]].

	interpreterProxy
		storePointer: 0 ofObject: results withValue: nameString;
		storePointer: 1 ofObject: results withValue: createDateOop;
		storePointer: 2 ofObject: results withValue: modDateOop;
		storePointer: 3 ofObject: results withValue: (dirFlag
														ifTrue: [interpreterProxy trueObject]
														ifFalse: [interpreterProxy falseObject]);
		storePointer: 4 ofObject: results withValue: fileSizeOop.
	^results! !

!FilePlugin methodsFor: 'directory primitives' stamp: 'eem 5/16/2017 18:46'!
makeDirEntryName: entryName size: entryNameSize createDate: createDate modDate: modifiedDate isDir: dirFlag fileSize: fileSize posixPermissions: posixPermissions isSymlink: symlinkFlag
	<var: 'entryName' type: #'char *'>
	<var: 'fileSize' type: #squeakFileOffsetType>
	<option: #PharoVM>
	| modDateOop createDateOop nameString results stringPtr posixPermissionsOop fileSizeOop |
	<var: 'stringPtr' type: #'char *'>

	"allocate storage for results, remapping newly allocated
	 oops if required in case GC happens during allocation"
	results := interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 7.
	self remapOop: results in:
		[nameString := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: entryNameSize.
		 self remapOop: nameString in:
			[createDateOop := interpreterProxy positive32BitIntegerFor: createDate.
			 self remapOop: createDateOop in:
				[modDateOop := interpreterProxy positive32BitIntegerFor: modifiedDate.
				 self remapOop: modDateOop in:
					[fileSizeOop := interpreterProxy positive64BitIntegerFor: fileSize.
					 self remapOop: fileSizeOop in:
						[posixPermissionsOop := interpreterProxy positive32BitIntegerFor: posixPermissions]]]]].

	"copy name into Smalltalk string"
	stringPtr := interpreterProxy firstIndexableField: nameString.
	0 to: entryNameSize - 1 do:
		[ :i |
		self cCode: [stringPtr at: i put: (entryName at: i)]
			inSmalltalk: [interpreterProxy storeByte: i ofObject: nameString withValue: (entryName at: i+1) asciiValue]].

	interpreterProxy
		storePointer: 0 ofObject: results withValue: nameString;
		storePointer: 1 ofObject: results withValue: createDateOop;
		storePointer: 2 ofObject: results withValue: modDateOop;
		storePointer: 3 ofObject: results withValue: (dirFlag
														ifTrue: [interpreterProxy trueObject]
														ifFalse: [interpreterProxy falseObject]);
		storePointer: 4 ofObject: results withValue: fileSizeOop;
		storePointer: 5 ofObject: results withValue: posixPermissionsOop;
		storePointer: 6 ofObject: results withValue: (symlinkFlag
														ifTrue: [interpreterProxy trueObject]
														ifFalse: [interpreterProxy falseObject]).
	^results! !

!FilePlugin methodsFor: 'initialize-release' stamp: 'JMM (auto pragmas 12/08) 1/21/2002 11:03'!
moduleUnloaded: aModuleName
	"The module with the given name was just unloaded.
	Make sure we have no dangling references."
	<export: true>
	<var: #aModuleName type: 'char *'>
	(aModuleName strcmp: 'SecurityPlugin') = 0 ifTrue:[
		"The security plugin just shut down. How odd."
		sCCPfn := sCDPfn := sCGFTfn := sCLPfn := sCSFTfn := sDFAfn := sCDFfn := sCOFfn := sCRFfn := sHFAfn := 0.
	].! !

!FilePlugin methodsFor: 'private' stamp: 'eem 1/15/2021 18:52'!
pointerFrom: pointerByteArray
	"Answer the machine address contained in anExternalAddressOop."
	<inline: #always>
	| ptr |
	<var: 'ptr' type: #'void **'>
	((interpreterProxy isBytes: pointerByteArray)
	 and: [(interpreterProxy stSizeOf: pointerByteArray) = (self sizeof: #'void *')]) ifFalse:
		[interpreterProxy primitiveFailFor: PrimErrBadArgument.
		 ^nil].
	ptr := interpreterProxy firstIndexableField: pointerByteArray.
	^ptr at: 0! !

!FilePlugin methodsFor: 'file primitives' stamp: 'VB 5/12/2018 10:16'!
primitiveConnectToFile
	"Connect to the file with the supplied FILE* and writeFlag.
	FILE* must be supplied in a byte object (ByteArray) with the platform address size.
	writeFlag must be a boolean and compatible with the existing file access."
	| writeFlag cfileOop cfile filePointer |
	<var: 'cfile' type: #'void *'>
	<export: true>
	writeFlag := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
	cfileOop := interpreterProxy stackValue: 1.
	cfile := self pointerFrom: cfileOop.
	interpreterProxy failed ifTrue: [
		"Ensure that the appropriate failure code has been set"
		^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	filePointer := self connectToFile: cfile write: writeFlag.
	interpreterProxy failed ifFalse: 
		[interpreterProxy methodReturnValue: filePointer]! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:24'!
primitiveConnectToFileDescriptor
	"Connect to the existing file identified by fileDescriptor.
	fileDescriptor must be an integer.
	writeFlag is aboolean indicating whether to open in read or write mode and must be compatible with the existing file access."
	| writeFlag fdPointer filePointer |
	<export: true>
	writeFlag := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
	fdPointer := interpreterProxy stackValue: 1.
	(interpreterProxy failed not
	 and: [interpreterProxy isIntegerObject: fdPointer]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	filePointer := self connectToFd: (interpreterProxy integerValueOf: fdPointer) write: writeFlag.
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnValue: filePointer]! !

!FilePlugin methodsFor: 'directory primitives' stamp: 'eem 10/9/2018 16:02'!
primitiveDirectoryCreate
	| dirName dirNameIndex dirNameSize okToCreate |
	<var: #dirNameIndex type: #'char *'>
	<export: true>

	dirName := interpreterProxy stackValue: 0.
	(interpreterProxy isBytes: dirName) ifFalse:
		[^interpreterProxy primitiveFail].
	dirNameIndex := interpreterProxy firstIndexableField: dirName.
	dirNameSize := interpreterProxy byteSizeOf: dirName.
	"If the security plugin can be loaded, use it to check for permission.
	If not, assume it's ok"
	sCCPfn ~= 0 ifTrue:
		[okToCreate := self cCode: ' ((sqInt (*)(char *, sqInt))sCCPfn)(dirNameIndex, dirNameSize)'
							inSmalltalk: [true].
		 okToCreate ifFalse:
			[^interpreterProxy primitiveFail]].
	(self dir_Create: dirNameIndex _: dirNameSize) ifFalse:
		[^interpreterProxy primitiveFail].
	interpreterProxy pop: 1! !

!FilePlugin methodsFor: 'directory primitives' stamp: 'eem 1/15/2021 18:40'!
primitiveDirectoryDelete

	| dirName dirNameIndex dirNameSize okToDelete |
	<var: #dirNameIndex type: #'char *'>
	<export: true>

	dirName := interpreterProxy stackValue: 0.
	(interpreterProxy isBytes: dirName)
		ifFalse: [^interpreterProxy primitiveFail].
	dirNameIndex := interpreterProxy firstIndexableField: dirName.
	dirNameSize := interpreterProxy byteSizeOf: dirName.
	"If the security plugin can be loaded, use it to check for permission.
	If not, assume it's ok"
	sCDPfn ~= 0
		ifTrue: [okToDelete := self cCode: ' ((sqInt (*)(char *, sqInt))sCDPfn)(dirNameIndex, dirNameSize)' inSmalltalk: [false].
			okToDelete
				ifFalse: [^interpreterProxy primitiveFail]].
	(self
			cCode: 'dir_Delete(dirNameIndex, dirNameSize)'
			inSmalltalk: [false])
		ifFalse: [^interpreterProxy primitiveFail].
	interpreterProxy pop: 1! !

!FilePlugin methodsFor: 'directory primitives' stamp: 'eem 10/29/2018 17:28'!
primitiveDirectoryDelimitor
	<export: true>
	interpreterProxy minorVersion >= 13
		ifTrue:
			[interpreterProxy
				pop: 1
				thenPush: (interpreterProxy characterObjectOf: self dir_Delimitor)]
		ifFalse:
			[| ascii |
			 ascii := self dir_Delimitor.
			 (ascii >= 0 and: [ascii <= 255])
				ifTrue:
					[interpreterProxy
						pop: 1
						thenPush: (interpreterProxy
										fetchPointer: ascii
										ofObject: interpreterProxy characterTable)]
				ifFalse:
					[interpreterProxy primitiveFail]]! !

!FilePlugin methodsFor: 'directory primitives' stamp: 'eem 1/15/2021 18:40'!
primitiveDirectoryEntry

	"Two arguments - directory path, and simple file name;
 	 returns an array (see primitiveDirectoryLookup) describing the file or directory,
 	 or nil if it does not exist.  
	 Primitive fails if the outer path does not identify a readable directory.
	 (This is a lookup-by-name variant of primitiveDirectoryLookup.)"

	| requestedName pathName pathNameIndex pathNameSize status entryName entryNameSize createDate modifiedDate dirFlag posixPermissions symlinkFlag fileSize okToList reqNameIndex reqNameSize |
	
	<var: 'entryName' declareC: 'char entryName[256]'>
	<var: 'pathNameIndex' type: #'char *'>
	<var: 'reqNameIndex' type: #'char *'>
	<var: 'fileSize' type: #squeakFileOffsetType>
	<export: true>

	requestedName := interpreterProxy stackValue: 0.
	pathName := interpreterProxy stackValue: 1.
	(interpreterProxy isBytes: pathName) ifFalse:
		[^interpreterProxy primitiveFail].

	"Outbound string parameters"
	pathNameIndex := interpreterProxy firstIndexableField: pathName.
	pathNameSize := interpreterProxy byteSizeOf: pathName.

	reqNameIndex := interpreterProxy firstIndexableField: requestedName.
	reqNameSize := interpreterProxy byteSizeOf: requestedName.
	self cCode: '' inSmalltalk:
		[entryName := ByteString new: 256.
		 entryNameSize := createDate := modifiedDate := dirFlag := fileSize := posixPermissions := symlinkFlag := nil].
	"If the security plugin can be loaded, use it to check for permission. 
	 If not, assume it's ok"
	okToList := sCLPfn ~= 0
					ifTrue: [self cCode: '((sqInt (*)(char *, sqInt))sCLPfn)(pathNameIndex, pathNameSize)' inSmalltalk: [true]]
					ifFalse: [true].
	status := okToList
		ifTrue:
			[self dir_EntryLookup: pathNameIndex _: pathNameSize
					_: reqNameIndex _: reqNameSize
					_: entryName _: (self addressOf: entryNameSize put: [:v| entryNameSize := v])
					_: (self addressOf: createDate put: [:v| createDate := v])
					_: (self addressOf: modifiedDate put: [:v| modifiedDate := v])
					_: (self addressOf: dirFlag put: [:v| dirFlag := v])
					_: (self addressOf: fileSize put: [:v| fileSize := v])
					_: (self addressOf: posixPermissions put: [:v| posixPermissions := v])
					_: (self addressOf: symlinkFlag put: [:v| symlinkFlag := v])]
		ifFalse:
			[DirNoMoreEntries].

	interpreterProxy failed ifTrue:
		[^nil].
	status = DirNoMoreEntries ifTrue: "no entry; return nil"
		[interpreterProxy "pop pathName, index, rcvr"
			pop: 3 thenPush: interpreterProxy nilObject.
			^nil].
	status = DirBadPath ifTrue:
		[^interpreterProxy primitiveFail]."bad path"

	interpreterProxy 
		pop: 3	"pop pathName, index, rcvr" 
		thenPush:
			(self 
				cppIf: PharoVM 
				ifTrue:
					[self
						makeDirEntryName: entryName
						size: entryNameSize
						createDate: createDate
						modDate: modifiedDate
						isDir: dirFlag
						fileSize: fileSize
						posixPermissions: posixPermissions
						isSymlink: symlinkFlag]
				ifFalse:
					[self
						makeDirEntryName: entryName
						size: entryNameSize
						createDate: createDate
						modDate: modifiedDate
						isDir: dirFlag
						fileSize: fileSize])! !

!FilePlugin methodsFor: 'directory primitives' stamp: 'eem 1/15/2021 18:41'!
primitiveDirectoryGetMacTypeAndCreator

	| creatorString typeString fileName creatorStringIndex typeStringIndex fileNameIndex fileNameSize okToGet |
	<var: 'creatorStringIndex' type: #'char *'>
	<var: 'typeStringIndex' type: #'char *'>
	<var: 'fileNameIndex' type: #'char *'>
	<export: true>

	creatorString := interpreterProxy stackValue: 0.
	typeString := interpreterProxy stackValue: 1.
	fileName := interpreterProxy stackValue: 2.
	((interpreterProxy isBytes: creatorString)
			and: [(interpreterProxy byteSizeOf: creatorString) = 4])
		ifFalse: [^interpreterProxy primitiveFail].
	((interpreterProxy isBytes: typeString)
			and: [(interpreterProxy byteSizeOf: typeString) = 4])
		ifFalse: [^interpreterProxy primitiveFail].
	(interpreterProxy isBytes: fileName)
		ifFalse: [^interpreterProxy primitiveFail].
	creatorStringIndex := interpreterProxy firstIndexableField: creatorString.
	typeStringIndex := interpreterProxy firstIndexableField: typeString.
	fileNameIndex := interpreterProxy firstIndexableField: fileName.
	fileNameSize := interpreterProxy byteSizeOf: fileName.
	"If the security plugin can be loaded, use it to check for permission.
	If not, assume it's ok"
	sCGFTfn ~= 0
		ifTrue: [okToGet := self cCode: ' ((sqInt (*)(char *, sqInt))sCGFTfn)(fileNameIndex, fileNameSize)'.
			okToGet
				ifFalse: [^interpreterProxy primitiveFail]].
	(self
			cCode: 'dir_GetMacFileTypeAndCreator(fileNameIndex, fileNameSize, typeStringIndex, creatorStringIndex)'
			inSmalltalk: [true])
		ifFalse: [^interpreterProxy primitiveFail].
	interpreterProxy pop: 3! !

!FilePlugin methodsFor: 'directory primitives' stamp: 'eem 1/15/2021 18:41'!
primitiveDirectoryLookup

	| index pathName pathNameIndex pathNameSize status entryName entryNameSize createDate modifiedDate dirFlag symlinkFlag posixPermissions fileSize okToList |
	
	<var: 'entryName' declareC: 'char entryName[256]'>
	<var: 'pathNameIndex' type: #'char *'>
	<var: 'fileSize' type: #squeakFileOffsetType>
	<export: true>

	index := interpreterProxy stackIntegerValue: 0.
	pathName := interpreterProxy stackValue: 1.
	(interpreterProxy isBytes: pathName)
		ifFalse: [^interpreterProxy primitiveFail].
	pathNameIndex := interpreterProxy firstIndexableField: pathName.
	pathNameSize := interpreterProxy byteSizeOf: pathName.
	self cCode: '' inSmalltalk:
		[entryName := ByteString new: 256.
		 entryNameSize := createDate := modifiedDate := dirFlag := fileSize := posixPermissions := symlinkFlag := nil].
	"If the security plugin can be loaded, use it to check for permission. 
	If not, assume it's ok"
	okToList := sCLPfn ~= 0
					ifTrue: [self cCode: '((sqInt (*)(char *, sqInt))sCLPfn)(pathNameIndex, pathNameSize)' inSmalltalk: [true]]
					ifFalse: [true].
	status := okToList
		ifTrue:
			[self dir_Lookup: pathNameIndex _: pathNameSize
					_: index
					_: entryName _: (self addressOf: entryNameSize put: [:v| entryNameSize := v])
					_: (self addressOf: createDate put: [:v| createDate := v])
					_: (self addressOf: modifiedDate put: [:v| modifiedDate := v])
					_: (self addressOf: dirFlag put: [:v| dirFlag := v])
					_: (self addressOf: fileSize put: [:v| fileSize := v])
					_: (self addressOf: posixPermissions put: [:v| posixPermissions := v])
					_: (self addressOf: symlinkFlag put: [:v| symlinkFlag := v])]
		ifFalse: [DirNoMoreEntries].
	interpreterProxy failed ifTrue:
		[^nil].
	status = DirNoMoreEntries ifTrue: "no more entries; return nil"
		[interpreterProxy "pop pathName, index, rcvr"
			pop: 3 thenPush: interpreterProxy nilObject.
		^nil].
	status = DirBadPath ifTrue:
		[^interpreterProxy primitiveFail]."bad path"

	interpreterProxy 
		pop: 3	"pop pathName, index, rcvr" 
		thenPush:
			(self 
				cppIf: PharoVM 
				ifTrue:
					[self
						makeDirEntryName: entryName
						size: entryNameSize
						createDate: createDate
						modDate: modifiedDate
						isDir: dirFlag
						fileSize: fileSize
						posixPermissions: posixPermissions
						isSymlink: symlinkFlag]
				ifFalse:
					[self
						makeDirEntryName: entryName
						size: entryNameSize
						createDate: createDate
						modDate: modifiedDate
						isDir: dirFlag
						fileSize: fileSize])! !

!FilePlugin methodsFor: 'directory primitives' stamp: 'eem 1/15/2021 18:41'!
primitiveDirectorySetMacTypeAndCreator

	| creatorString typeString fileName creatorStringIndex typeStringIndex fileNameIndex fileNameSize okToSet |
	<var: 'creatorStringIndex' type: #'char *'>
	<var: 'typeStringIndex' type: #'char *'>
	<var: 'fileNameIndex' type: #'char *'>
	<export: true>

	creatorString := interpreterProxy stackValue: 0.
	typeString := interpreterProxy stackValue: 1.
	fileName := interpreterProxy stackValue: 2.
	((interpreterProxy isBytes: creatorString)
	 and: [(interpreterProxy isBytes: typeString)
	 and: [(interpreterProxy isBytes: fileName)
	 and: [(interpreterProxy byteSizeOf: creatorString) = 4
	 and: [(interpreterProxy byteSizeOf: typeString) = 4]]]]) ifFalse:
		[^interpreterProxy primitiveFail].
	creatorStringIndex := interpreterProxy firstIndexableField: creatorString.
	typeStringIndex := interpreterProxy firstIndexableField: typeString.
	fileNameIndex := interpreterProxy firstIndexableField: fileName.
	fileNameSize := interpreterProxy byteSizeOf: fileName.
	"If the security plugin can be loaded, use it to check for permission.
	If not, assume it's ok"
	sCSFTfn ~= 0 ifTrue:
		[okToSet := self
						cCode: '((sqInt (*)(char *, sqInt))sCSFTfn)(fileNameIndex, fileNameSize)'
						inSmalltalk: [true].
		 okToSet ifFalse:
			[^interpreterProxy primitiveFail]].
	(self
		cCode: 'dir_SetMacFileTypeAndCreator(fileNameIndex, fileNameSize, typeStringIndex, creatorStringIndex)'
		inSmalltalk: [true]) ifFalse:
			[^interpreterProxy primitiveFail].
	interpreterProxy pop: 3! !

!FilePlugin methodsFor: 'security primitives' stamp: 'eem 7/10/2009 12:11'!
primitiveDisableFileAccess
	<export: true>
	"If the security plugin can be loaded, use it to turn off file access
	 If not, assume it's ok"
	sDFAfn ~= 0 ifTrue:
		[self cCode: '((sqInt (*)(void))sDFAfn)()']! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:22'!
primitiveFileAtEnd
	| file atEnd |
	<export: true>
	file := self fileValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifFalse:
		[atEnd := self sqFileAtEnd: file.
		interpreterProxy failed ifFalse:
			[interpreterProxy methodReturnBool: atEnd]]! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:19'!
primitiveFileClose
	| file |
	<export: true>
	file := self fileValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifFalse:
		[self sqFileClose: file.
		 interpreterProxy failed ifFalse:
			[interpreterProxy pop: 1]]  "pop file; leave rcvr on stack"! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:19'!
primitiveFileDelete

	| namePointer nameIndex nameSize  okToDelete |
	<var: 'nameIndex' type: #'char *'>
	<export: true>

	namePointer := interpreterProxy stackValue: 0.
	(interpreterProxy isBytes: namePointer) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	nameIndex := interpreterProxy firstIndexableField: namePointer.
	nameSize := interpreterProxy byteSizeOf: namePointer.
	"If the security plugin can be loaded, use it to check for permission.
	If not, assume it's ok"
	sCDFfn ~= 0
		ifTrue: [okToDelete := self cCode: ' ((sqInt (*)(char *, sqInt))sCDFfn)(nameIndex, nameSize)' inSmalltalk: [true].
			okToDelete
				ifFalse: [^ interpreterProxy primitiveFail]].
	self
		sqFileDeleteName: nameIndex
		Size: nameSize.
	interpreterProxy failed
		ifFalse: [interpreterProxy pop: 1]! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:14'!
primitiveFileDescriptorType
	| fdPointer |
	"Allow to test if the standard input/output files are from a console or not
	Return values:
	* -1 - Error
	* 0 - no console (windows only)
	* 1 - normal terminal (unix terminal / windows console)
	* 2 - pipe
	* 3 - file
	* 4 - cygwin terminal (windows only)"
	<var: 'fd' type: #int>
	<export: true>
	fdPointer := interpreterProxy stackValue: 0.
	(interpreterProxy isIntegerObject: fdPointer) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	interpreterProxy methodReturnInteger: (self sqFileDescriptorType: (interpreterProxy integerValueOf: fdPointer))! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:11'!
primitiveFileFlush
	| file |
	<export: true>
	file := self fileValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifFalse:
		[self sqFileFlush: file.
		 interpreterProxy failed ifFalse:
			[interpreterProxy pop: 1]]! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:11'!
primitiveFileGetPosition
	| file position |
	<var: 'position' type: #squeakFileOffsetType>
	<export: true>
	file := self fileValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifFalse:
		[position := self sqFileGetPosition: file.
		 interpreterProxy failed ifFalse:
			[interpreterProxy methodReturnValue: (interpreterProxy positive64BitIntegerFor: position)]]! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:05'!
primitiveFileOpen
	| writeFlag namePointer filePointer nameIndex nameSize |
	<var: 'nameIndex' type: #'char *'>
	<export: true>
	writeFlag := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
	namePointer := interpreterProxy stackValue: 1.
	(interpreterProxy isBytes: namePointer) ifFalse:
		[^interpreterProxy primitiveFail].
	nameIndex := interpreterProxy firstIndexableField: namePointer.
	nameSize := interpreterProxy byteSizeOf: namePointer.
	filePointer := self fileOpenName: nameIndex size: nameSize write: writeFlag secure: true.
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnValue: filePointer]! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:05'!
primitiveFileOpenNew
	| namePointer filePointer nameIndex nameSize |
	<var: 'nameIndex' type: #'char *'>
	<export: true>
	namePointer := interpreterProxy stackValue: 0.
	(interpreterProxy isBytes: namePointer) ifFalse:
		[^interpreterProxy primitiveFail].
	nameIndex := interpreterProxy firstIndexableField: namePointer.
	nameSize := interpreterProxy byteSizeOf: namePointer.
	filePointer := self fileOpenNewName: nameIndex size: nameSize secure: true.
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnValue: filePointer]! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 10/28/2020 19:47'!
primitiveFileRead
	<export: true>
	self cppIf: SPURVM
		ifTrue: [self cppIf: COGMTVM
					ifTrue: [self primitiveFileReadPinningAndDisowning]
					ifFalse: [self primitiveFileReadWithPinning]]
		ifFalse: [self primitiveFileReadWithoutPinning]! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 11/9/2020 19:13'!
primitiveFileReadPinningAndDisowning
	"This version of primitiveFileRead is for garbage collectors that support pinning
	 and the multi-threaded VM.  It actually does the own/disown dance."
	| count startIndex array file slotSize elementSize bytesRead threadIndexAndFlags wasPinned |
	<inline: true>
	<var: 'file' type: #'SQFile *'>
	<var: 'count' type: #'size_t'>
	<var: 'startIndex' type: #'size_t'>
	<var: 'slotSize' type: #'size_t'>
	<var: 'elementSize' type: #'size_t'>
	count		:= interpreterProxy positiveMachineIntegerValueOf: (interpreterProxy stackValue: 0).
	startIndex	:= interpreterProxy positiveMachineIntegerValueOf: (interpreterProxy stackValue: 1).
 	array		:= interpreterProxy stackValue: 2.
	file			:= self fileValueOf: (interpreterProxy stackValue: 3).

	(interpreterProxy failed
	"buffer can be any indexable words or bytes object except CompiledMethod"
	 or: [(interpreterProxy isWordsOrBytes: array) not]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	slotSize := interpreterProxy slotSizeOf: array.
	(startIndex >= 1 and: [startIndex + count - 1 <= slotSize]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(wasPinned := interpreterProxy isPinned: array) ifFalse:
		[array := interpreterProxy pinObject: array].
	threadIndexAndFlags := interpreterProxy disownVM: DisownVMForThreading.
	"Note: adjust startIndex for zero-origin byte indexing"
	elementSize := slotSize = 0
						ifTrue: [1]
						ifFalse: [(interpreterProxy byteSizeOf: array) // slotSize].
	bytesRead := self
					sqFile: file
					Read: count * elementSize
					Into: (interpreterProxy cCoerce: (interpreterProxy firstIndexableField: array) to: #'char *')
					At: startIndex - 1 * elementSize.
	interpreterProxy ownVM: threadIndexAndFlags.
	wasPinned ifFalse:
		[interpreterProxy unpinObject: array].
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnInteger: bytesRead // elementSize] "answer # of elements read"! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:10'!
primitiveFileReadWithPinning
	"This version of primitiveFileRead is for garbage collectors that support pinning."
	| count startIndex array file slotSize elementSize bytesRead |
	<inline: true>
	<var: 'count' type: #'size_t'>
	<var: 'startIndex' type: #'size_t'>
	<var: 'slotSize' type: #'size_t'>
	<var: 'elementSize' type: #'size_t'>
	count		:= interpreterProxy positiveMachineIntegerValueOf: (interpreterProxy stackValue: 0).
	startIndex	:= interpreterProxy positiveMachineIntegerValueOf: (interpreterProxy stackValue: 1).
 	array		:= interpreterProxy stackValue: 2.
	file			:= self fileValueOf: (interpreterProxy stackValue: 3).

	(interpreterProxy failed
	"buffer can be any indexable words or bytes object except CompiledMethod"
	 or: [(interpreterProxy isWordsOrBytes: array) not]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	slotSize := interpreterProxy slotSizeOf: array.
	(startIndex >= 1 and: [startIndex + count - 1 <= slotSize]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].

	"Note: adjust startIndex for zero-origin byte indexing"
	elementSize := slotSize = 0
						ifTrue: [1]
						ifFalse: [(interpreterProxy byteSizeOf: array) // slotSize].
	bytesRead := self
					sqFile: file
					Read: count * elementSize
					Into: (interpreterProxy cCoerce: (interpreterProxy firstIndexableField: array) to: #'char *')
					At: startIndex - 1 * elementSize.
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnInteger: bytesRead // elementSize] "answer # of elements read"! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:10'!
primitiveFileReadWithoutPinning
	"This version of primitiveFileRead is for garbage collectors without support for pinning."
	| retryCount count startIndex array file elementSize bytesRead |
	<inline: true>
	<var: 'count' type: #'size_t'>
	<var: 'startIndex' type: #'size_t'>
	<var: 'elementSize' type: #'size_t'>
	retryCount	:= 0.
	count		:= interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	startIndex	:= interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 1).
 
	[array		:= interpreterProxy stackValue: 2.
	 file			:= self fileValueOf: (interpreterProxy stackValue: 3).

	 (interpreterProxy failed
	 "buffer can be any indexable words or bytes object except CompiledMethod"
	 or: [(interpreterProxy isWordsOrBytes: array) not]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	 elementSize := (interpreterProxy isWords: array) ifTrue: [4] ifFalse: [1].
	 (startIndex >= 1
	  and: [(startIndex + count - 1) <= (interpreterProxy slotSizeOf: array)]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].

	 "Note: adjust startIndex for zero-origin indexing"
	 bytesRead := self
					sqFile: file
					Read: count * elementSize
					Into: (interpreterProxy cCoerce: (interpreterProxy firstIndexableField: array) to: #'char *')
					At: (startIndex - 1) * elementSize.
	 interpreterProxy primitiveFailureCode = PrimErrObjectMayMove
	 and: [(retryCount := retryCount + 1) <= 2] "Two objects, the file and the array can move"] whileTrue:
		[interpreterProxy
			tenuringIncrementalGC;
			primitiveFailFor: PrimNoErr].
	interpreterProxy failed ifFalse:
		[interpreterProxy
				methodReturnInteger: bytesRead // elementSize "push # of elements read"]! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:04'!
primitiveFileRename

	| oldNamePointer newNamePointer oldNameIndex oldNameSize newNameIndex newNameSize  okToRename |
	<var: 'oldNameIndex' type: #'char *'>
	<var: 'newNameIndex' type: #'char *'>
	<export: true>

	newNamePointer := interpreterProxy stackValue: 0.
	oldNamePointer := interpreterProxy stackValue: 1.
	((interpreterProxy isBytes: newNamePointer)
	 and: [interpreterProxy isBytes: oldNamePointer]) ifFalse:
		[^interpreterProxy primitiveFail].
	newNameIndex := interpreterProxy firstIndexableField: newNamePointer.
	newNameSize := interpreterProxy byteSizeOf: newNamePointer.
	oldNameIndex := interpreterProxy firstIndexableField: oldNamePointer.
	oldNameSize := interpreterProxy byteSizeOf: oldNamePointer.
	"If the security plugin can be loaded, use it to check for rename permission.
	If not, assume it's ok"
	sCRFfn ~= 0 ifTrue: 
		[okToRename := self cCode: ' ((sqInt (*)(char *, sqInt))sCRFfn)(oldNameIndex, oldNameSize)'
			inSmalltalk: [true].
		okToRename ifFalse: [^interpreterProxy primitiveFail]].
	self
		sqFileRenameOld: oldNameIndex Size: oldNameSize
		New: newNameIndex Size: newNameSize.
	interpreterProxy failed
		ifFalse: [interpreterProxy pop: 2]! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:09'!
primitiveFileSetPosition
	| newPosition file |
	<var: 'newPosition' type: #squeakFileOffsetType>
	<export: true>
	(interpreterProxy byteSizeOf: (interpreterProxy stackValue: 0)) > (self sizeof: #squeakFileOffsetType) ifTrue:
		[^interpreterProxy primitiveFail].
	newPosition := interpreterProxy positive64BitValueOf: (interpreterProxy stackValue: 0).
	file := self fileValueOf: (interpreterProxy stackValue: 1).
	interpreterProxy failed ifFalse:
		[self sqFile: file SetPosition: newPosition.
		 interpreterProxy failed ifFalse:
			[interpreterProxy pop: 2]] "pop position, file; leave rcvr on stack"! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:09'!
primitiveFileSize
	| file size |
	<var: 'size' type: #squeakFileOffsetType>
	<export: true>
	file := self fileValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifFalse:[size := self sqFileSize: file].
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnValue: (interpreterProxy positive64BitIntegerFor: size)].! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 9/19/2018 17:31'!
primitiveFileStdioHandles
	"Answer an Array of file handles for standard in, standard out and standard error,
	 with nil in entries that are unvailable, e.g. because the platform does not provide
	 standard error, etc.  Fail if an error occurs determining the stdio handles,
	 if the security plugin denies access or if memory runs out."
	| fileRecords result validMask |
	<export: true>
	<var: 'fileRecords' declareC: 'SQFile fileRecords[3]'>
	sHFAfn ~= 0 ifTrue:
		[(self cCode: ' ((sqInt (*)(void))sHFAfn)()' inSmalltalk: [true]) ifFalse:
			[^interpreterProxy primitiveFailFor: PrimErrUnsupported]].
	self cCode: '' inSmalltalk: [fileRecords := Array new: 3].
	validMask := self sqFileStdioHandlesInto: fileRecords.
	validMask < 0 ifTrue:
		[^interpreterProxy primitiveFailForOSError: validMask].
	result := interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 3.
	result = nil ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoMemory].
	interpreterProxy pushRemappableOop: result.
	0 to: 2 do:
		[:index|
		(validMask bitAnd: (1 << index)) ~= 0 ifTrue:
			[result := interpreterProxy instantiateClass: interpreterProxy classByteArray indexableSize: self fileRecordSize.
			 result = nil ifTrue:
				[interpreterProxy popRemappableOop.
				^interpreterProxy primitiveFailFor: PrimErrNoMemory].
			 interpreterProxy storePointer: index ofObject: interpreterProxy topRemappableOop withValue: result.
			 self
				cCode:
					[self memcpy: (interpreterProxy firstIndexableField: result)
						_: (self addressOf: (fileRecords at: index))
						_: self fileRecordSize]
				inSmalltalk:
					[(interpreterProxy firstIndexableField: result)
						unitSize: interpreterProxy wordSize;
						at: 0 put: (fileRecords at: index + 1)]]].
	 "In the non-Spur threaded VM ensure the handles are old, so that sqFileReadIntoAt is unaffected
	  by incremental GCs.  See platforms/Cross/plugins/FilePlugin/sqFilePluginBasicPrims.c.  The Spur
	  VM uses pinning, so it doesn't need the GC."
	self cppIf: COGMTVM
		ifTrue: [self cppIf: SPURVM
					ifTrue: []
					ifFalse: [interpreterProxy fullGC]].
	result := interpreterProxy popRemappableOop.
	interpreterProxy methodReturnValue: result! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:08'!
primitiveFileSync
	| file |
	<export: true>
	file := self fileValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifFalse:[self sqFileSync: file].
	interpreterProxy failed ifFalse: [interpreterProxy pop: 1].! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:08'!
primitiveFileTruncate
	"ftruncate is not an ansi function so we have a macro to point to a suitable platform implementation" 
	| truncatePosition file |
	<var: 'truncatePosition' type: #squeakFileOffsetType>
	<export: true>
	(interpreterProxy isIntegerObject: (interpreterProxy stackValue: 0)) ifFalse:
		[(interpreterProxy byteSizeOf: (interpreterProxy stackValue: 0)) > (self sizeof: #squeakFileOffsetType) ifTrue:
			[^interpreterProxy primitiveFail]].
	truncatePosition := interpreterProxy positive64BitValueOf: (interpreterProxy stackValue: 0).
	file := self fileValueOf: (interpreterProxy stackValue: 1).
	interpreterProxy failed ifFalse:
		[self sqFile: file Truncate: truncatePosition].
		 interpreterProxy failed ifFalse:
			[interpreterProxy pop: 2 "pop position, file; leave rcvr on stack"]! !

!FilePlugin methodsFor: 'file primitives' stamp: 'eem 1/15/2021 18:07'!
primitiveFileWrite
	| count startIndex array file slotSize elementSize bytesWritten |
	<var: 'count' type: #'size_t'>
	<var: 'startIndex' type: #'size_t'>
	<var: 'slotSize' type: #'size_t'>
	<var: 'elementSize' type: #'size_t'>
	<export: true>
	count		:= interpreterProxy positiveMachineIntegerValueOf: (interpreterProxy stackValue: 0).
	startIndex	:= interpreterProxy positiveMachineIntegerValueOf: (interpreterProxy stackValue: 1).
	array		:= interpreterProxy stackValue: 2.
	file			:= self fileValueOf: (interpreterProxy stackValue: 3).

	 (interpreterProxy failed
	 "buffer can be any indexable words or bytes object except CompiledMethod"
	 or: [(interpreterProxy isWordsOrBytes: array) not]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	slotSize := interpreterProxy slotSizeOf: array.
	(startIndex >= 1 and: [startIndex + count - 1 <= slotSize]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].

	"Note: adjust startIndex for zero-origin byte indexing"
	elementSize := slotSize = 0
						ifTrue: [1]
						ifFalse: [(interpreterProxy byteSizeOf: array) // slotSize].
	bytesWritten := self
						sqFile: file
						Write: count * elementSize
						From: (interpreterProxy cCoerce: (interpreterProxy firstIndexableField: array) to: #'char *')
						At: startIndex - 1 * elementSize.
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnInteger: bytesWritten // elementSize] "answer # of elements written"! !

!FilePlugin methodsFor: 'security primitives' stamp: 'VB 5/12/2018 10:19'!
primitiveHasFileAccess
	|  hasAccess |
	<export: true>
	"If the security plugin can be loaded, use it to check . 
	If not, assume it's ok"
	sHFAfn ~= 0
		ifTrue: [hasAccess := self cCode: ' ((sqInt (*)(void))sHFAfn)()' inSmalltalk: [true]]
		ifFalse: [hasAccess := true].
	interpreterProxy methodReturnBool: hasAccess! !

!FilePlugin methodsFor: 'file primitives' stamp: 'ar (auto pragmas 12/08) 5/13/2000 14:51'!
setMacFile: fileName Type: typeString AndCreator: creatorString
	"Exported entry point for the VM. Needed for image saving only and no-op on anything but Macs."
	<export: true>
	<var: #fileName type: 'char *'>
	<var: #typeString type: 'char *'>
	<var: #creatorString type: 'char *'>
	^self cCode: 'dir_SetMacFileTypeAndCreator(fileName, strlen(fileName), typeString, creatorString)'.! !

!FilePlugin methodsFor: 'initialize-release' stamp: 'eem 11/10/2020 11:53'!
shutdownModule
	<export: true>
	^self sqFileShutdown! !

!FilePlugin methodsFor: 'private' stamp: 'akg 3/6/2018 22:12'!
sizeOfPointer
	"Return the size of a pointer (address size)"
	<inline: #always>
	^self sizeof: #'void *'! !

FilePlugin subclass: #FilePluginSimulator
	instanceVariableNames: 'openFiles states maxOpenFiles'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!FilePluginSimulator commentStamp: 'tpr 5/5/2003 12:02' prior: 0!
File plugin simulation for the VM simulator!

!FilePluginSimulator class methodsFor: 'translation' stamp: 'tpr 5/14/2001 12:34'!
shouldBeTranslated
"This class should not be translated"
	^false! !

!FilePluginSimulator methodsFor: 'initialize-release' stamp: 'eem 1/6/2020 11:35'!
close
	"close any files that ST may have opened"
	openFiles do:
		[:f |
		 [f close]
			on: Error
			do: [:ex|
				ex messageText ~= 'primClose: failed' ifTrue:
					[ex pass]]]! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 10/9/2018 16:03'!
dir_Create: dirNameIndex _: dirNameSize
	^[FileDirectory default
		primCreateDirectory: (interpreterProxy interpreter
								asString: dirNameIndex
								size: dirNameSize).
	   true]
		on: Error
		do: [:ex| false]! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 9/18/2018 15:02'!
dir_EntryLookup: pathString _: pathStringLength _: entryNameString _: entryNameStringLength _: name _: nameLength _: creationDate _: modificationDate _: isDirectory _: sizeIfFile _: posixPermissions _: isSymlink
	"sqInt dir_EntryLookup(char *pathString, sqInt pathStringLength, char *nameString, sqInt nameStringLength,
		/* outputs: */		char *name, sqInt *nameLength, sqInt *creationDate, sqInt *modificationDate,
  						      sqInt *isDirectory, squeakFileOffsetType *sizeIfFile, sqInt *posixPermissions, sqInt *isSymlink)"
	| result pathName entryName |
	pathName := ((0 to: pathStringLength - 1) collect: [:i| (pathString at: i) asCharacter]) as: ByteString.
	entryName := ((0 to: entryNameStringLength - 1) collect: [:i| (entryNameString at: i) asCharacter]) as: ByteString.
	result := self primLookupEntryIn: pathName name: entryName.
	result ifNil: [^DirNoMoreEntries].
	result isInteger ifTrue:
		[result > 1 ifTrue:
			[interpreterProxy primitiveFailFor: result].
		 ^DirBadPath].
	name replaceFrom: 1 to: result first size with: result first startingAt: 1.
	nameLength at: 0 put: result first size.
	creationDate at: 0 put: (result at: 2).
	modificationDate at: 0 put: (result at: 3).
	isDirectory at: 0 put: (result at: 4).
	sizeIfFile at: 0 put: (result at: 5).
	posixPermissions at: 0 put: (result at: 6 ifAbsent: [(result at: 4) ifTrue: [8r755] ifFalse: [8r644]]).
	isSymlink at: 0 put: (result at: 7 ifAbsent: [false]).
	^DirEntryFound! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 9/18/2018 14:59'!
dir_Lookup: pathString _: pathStringLength _: index _: name _: nameLength _: creationDate _: modificationDate _: isDirectory _: sizeIfFile _: posixPermissions _: isSymlink
	"sqInt dir_Lookup(	char *pathString, sqInt pathStringLength, sqInt index,
		/* outputs: */	char *name, sqInt *nameLength, sqInt *creationDate, sqInt *modificationDate,
		   				sqInt *isDirectory, squeakFileOffsetType *sizeIfFile, sqInt * posixPermissions, sqInt *isSymlink)"
	| result pathName |
	pathName := ((0 to: pathStringLength - 1) collect: [:i| (pathString at: i) asCharacter]) as: ByteString.
	result := self primLookupEntryIn: pathName index: index.
	result ifNil: [^DirNoMoreEntries].
	result isInteger ifTrue:
		[result > 1 ifTrue:
			[interpreterProxy primitiveFailFor: result].
		 ^DirBadPath].
	name replaceFrom: 1 to: result first size with: result first startingAt: 1.
	nameLength at: 0 put: result first size.
	creationDate at: 0 put: (result at: 2).
	modificationDate at: 0 put: (result at: 3).
	isDirectory at: 0 put: (result at: 4).
	sizeIfFile at: 0 put: (result at: 5).
	posixPermissions at: 0 put: (result at: 6 ifAbsent: [(result at: 4) ifTrue: [8r755] ifFalse: [8r644]]).
	isSymlink at: 0 put: (result at: 7 ifAbsent: [false]).
	^DirEntryFound! !

!FilePluginSimulator methodsFor: 'file primitives' stamp: 'eem 3/5/2019 06:08'!
fileOpenName: nameIndex size: nameSize write: writeFlag secure: secureFlag
	"Open the named file, possibly checking security. Answer the file oop."
	| path f index |
	openFiles size >= maxOpenFiles ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrLimitExceeded].
	path := (interpreterProxy asByteArray: nameIndex size: nameSize) utf8Decoded.
	(writeFlag
	 or: [StandardFileStream isAFileNamed: path]) ifFalse:
		[^interpreterProxy primitiveFail].
	f := StandardFileStream new open: path forWrite: writeFlag.
	f ifNil: [^interpreterProxy primitiveFail].
	f binary.
	self assert: f isReadOnly = writeFlag not.
	index := (3 to: openFiles size + 1) detect: [:n| (openFiles includesKey: n) not].
	openFiles at: index put: f.
	^interpreterProxy integerObjectOf: index! !

!FilePluginSimulator methodsFor: 'file primitives' stamp: 'eem 3/3/2019 16:07'!
fileOpenNewName: nameIndex size: nameSize secure: secureFlag
	"Open the new named file, possibly checking security. Answer the file oop."
	| path f index |
	openFiles size >= maxOpenFiles ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrLimitExceeded].
	path := (interpreterProxy interpreter asByteArray: nameIndex size: nameSize) utf8Decoded.
	"the #defaultAction for FileExistsException creates a dialog,
	so it is caught and resignaled as a generic Error"
	[f := StandardFileStream newFileNamed: path]
		on: FileExistsException
		do: [:error | ^ interpreterProxy primitiveFailFor: PrimErrInappropriate].
	f ifNil: [^interpreterProxy primitiveFail].
	self deny: f isReadOnly.
	f binary.
	index := (3 to: openFiles size + 1) detect: [:n| (openFiles includesKey: n) not].
	openFiles at: index put: f.
	^interpreterProxy integerObjectOf: index! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'cb 4/17/2015 13:39'!
fileValueOf: objectPointer
	| index file |
	index := (interpreterProxy isIntegerObject: objectPointer)
				ifTrue: [interpreterProxy integerValueOf: objectPointer]
				ifFalse:
					[((interpreterProxy isBytes: objectPointer)
					  and: [(interpreterProxy byteSizeOf: objectPointer) = (self sizeof: #SQFile)]) ifFalse:
						[interpreterProxy primitiveFail.
						 ^nil].
					interpreterProxy longAt: objectPointer + interpreterProxy baseHeaderSize].
	file := openFiles at: index ifAbsent: [ interpreterProxy primitiveFail. ^ nil ].
	"this attempts to preserve file positions across snapshots when debugging the VM
	 requires saving an image in full flight and pushing it over the cliff time after time..."
	(file closed and: [states includesKey: file]) ifTrue:
		[[:pos :isBinary|
		  file reopen; position: pos.
		  isBinary ifTrue:
			[file binary]] valueWithArguments: (states at: file)].
	^file! !

!FilePluginSimulator methodsFor: 'file security' stamp: 'ar 2/5/2001 19:23'!
ioCanCreatePath: dirNameIndex OfSize: dirNameSize
	"Return true if we're allowed to create a directory with the given name"
	^true! !

!FilePluginSimulator methodsFor: 'file security' stamp: 'ar 2/5/2001 18:16'!
ioCanDeleteFile: nameIndex OfSize: nameSize
	"Return true if we're allowed to delete the file with the given name"
	^true! !

!FilePluginSimulator methodsFor: 'file security' stamp: 'ar 2/5/2001 18:16'!
ioCanDeletePath: dirNameIndex OfSize: dirNameSize
	"Return true if we're allowed to delete the directory with the given name"
	^true! !

!FilePluginSimulator methodsFor: 'file security' stamp: 'ar 2/5/2001 18:16'!
ioCanGetFileType: fileNameIndex OfSize: fileNameSize
	"Return true if we're allowed to retrieve the (mac) file type of the given file."
	^true! !

!FilePluginSimulator methodsFor: 'file security' stamp: 'ar 2/5/2001 18:17'!
ioCanListPath: pathNameIndex OfSize: pathNameSize
	"Return true if we're allowed to list the contents of the given directory"
	^true! !

!FilePluginSimulator methodsFor: 'file security' stamp: 'ar 2/5/2001 18:17'!
ioCanOpenFile: nameIndex OfSize: nameSize Writable: writeFlag
	"Return true if we're allowed to open the given file (possibly in write mode)"
	^true! !

!FilePluginSimulator methodsFor: 'file security' stamp: 'ar 2/5/2001 18:17'!
ioCanRenameFile: oldNameIndex OfSize: oldNameSize
	"Return true if we're allowed to rename the given file"
	^true! !

!FilePluginSimulator methodsFor: 'file security' stamp: 'ar 2/5/2001 18:18'!
ioCanSetFileType: fileNameIndex OfSize: fileNameSize
	"Return true if we're allowed to set the (mac) file type and creator on the given file"
	^true! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 9/18/2018 14:38'!
primLookupEntryIn: fullPath index: index
	"Look up the index-th entry of the directory with the given fully-qualified path (i.e., starting from the root of the file hierarchy) and return an array containing:

	<name> <creationTime> <modificationTime> <dirFlag> <fileSize>

	The empty string enumerates the top-level files or drives. (For example, on Unix, the empty path enumerates the contents of '/'. On Macs and PCs, it enumerates the mounted volumes/drives.)

	The creation and modification times are in seconds since the start of the Smalltalk time epoch. DirFlag is true if the entry is a directory. FileSize the file size in bytes or zero for directories. The primitive returns nil when index is past the end of the directory. It fails if the given path is bad."

 	<primitive: 'primitiveDirectoryLookup' module: 'FilePlugin' error: ec>
	^ec isInteger
		ifTrue: [ec]
		ifFalse:
			[Smalltalk primitiveErrorTable
				indexOf: ec
				ifAbsent: [Smalltalk primitiveErrorTable size + 1]]! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 9/18/2018 14:39'!
primLookupEntryIn: fullPath name: fName
	"Look up <fName> (a simple file name) in the directory identified by <fullPath>
 	 and return an array containing:

	<fName> <creationTime> <modificationTime> <dirFlag> <fileSize>

	On Unix, the empty path denotes '/'. 
      On Macs and PCs, it is the container of the system volumes.)

	The creation and modification times are in seconds since the start of the Smalltalk time epoch. DirFlag is true if the entry is a directory. FileSize the file size in bytes or zero for directories. The primitive returns nil when index is past the end of the directory. It fails if the given path is bad."

 	<primitive: 'primitiveDirectoryEntry' module: 'FilePlugin' error: ec>
	^ec isInteger
		ifTrue: [ec]
		ifFalse:
			[Smalltalk primitiveErrorTable
				indexOf: ec
				ifAbsent: [Smalltalk primitiveErrorTable size + 1]]! !

!FilePluginSimulator methodsFor: 'file primitives' stamp: 'eem 3/3/2019 14:14'!
primitiveFileGetPosition
	super primitiveFileGetPosition.
	interpreterProxy failed ifTrue:
		[self halt]! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 3/18/2014 14:09'!
recordStateOf: file
	([file position]
			on: Error
			do: [:ex| nil]) ifNotNil:
		[:position|
		states at: file put: {position. file isBinary}]! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 10/29/2020 17:47'!
sqFile: file Read: count Into: byteArrayIndexArg At: startIndex
	| byteArrayIndex |
	file isFakeStdinStream ifTrue:
		[^file sqFile: file Read: count Into: byteArrayIndexArg At: startIndex].
	byteArrayIndex := byteArrayIndexArg asInteger. "Coerces CArray et al correctly"
	[[startIndex to: startIndex + count - 1 do:
		[ :i |
		interpreterProxy byteAt: byteArrayIndex + i put: (file next ifNil: [^i - startIndex] ifNotNil: [:charOrByte| charOrByte asInteger])]]
			on: Error
			do: [:ex|
				(file isStream and: [file isTranscriptStream]) ifFalse:
					[ex pass].
				^0]]
		ensure: [self recordStateOf: file].
	^count! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 3/18/2014 14:11'!
sqFile: file SetPosition: newPosition
	file position: newPosition.
	self recordStateOf: file! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 3/18/2014 14:11'!
sqFile: file Truncate: truncatePosition
	file truncate: truncatePosition.
	self recordStateOf: file! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 6/7/2017 14:47'!
sqFile: file Write: count From: byteArrayIndexArg At: startIndex
	| byteArrayIndex |
	byteArrayIndex := byteArrayIndexArg asInteger. "Coerces CArray et al correctly"
	file isBinary
		ifTrue:
			[startIndex to: startIndex + count - 1 do:
				[ :i | file nextPut: (interpreterProxy byteAt: byteArrayIndex + i)]]
		ifFalse:
			[startIndex to: startIndex + count - 1 do:
				[ :i | | byte |
				byte := interpreterProxy byteAt: byteArrayIndex + i.
				file nextPut: (Character value: (byte == 12 "lf" ifTrue: [15"cr"] ifFalse: [byte]))]].
	self recordStateOf: file.
	^count! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 6/22/2011 12:14'!
sqFileAtEnd: file
	^file atEnd! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 1/17/2018 01:09'!
sqFileClose: file
	| index |
	index := openFiles keyAtValue: file.
	openFiles removeKey: index.
	states removeKey: file ifAbsent: [].
	file close! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 9/25/2020 06:55'!
sqFileDeleteName: nameIndex Size: nameSize
	| path |
	path := (interpreterProxy asString: nameIndex size: nameSize) asByteArray utf8Decoded.
	(StandardFileStream isAFileNamed: path) ifFalse:
		[^interpreterProxy primitiveFail].
	(InitializationOptions at: #haltOnFileDelete ifAbsent: [true])
		ifTrue: [self halt: 'Deleting ', (path contractTo: 64)]
		ifFalse: [interpreterProxy transcript cr; show: 'Deleting ', (path contractTo: 64); cr].
	[FileDirectory deleteFilePath: path]
		on: Error
		do: [:ex| interpreterProxy primitiveFail]! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 1/15/2021 18:18'!
sqFileDescriptorType: fd
	"Allow to test if the standard input/output files are from a console or not
		Return values:
			-1 - Error
			0 - no console (windows only)
			1 - normal terminal (unix terminal / windows console)
			2 - pipe
			3 - file
			4 - cygwin terminal (windows only)"
	^fd < 0
		ifTrue: [-1]
		ifFalse:
			[fd <= 2 "i.e. stdio,stdout,stderr"
				ifTrue: [1]
				ifFalse: [3]]! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 6/22/2011 12:48'!
sqFileFlush: file
	^file flush! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 6/22/2011 12:49'!
sqFileGetPosition: file
	^file position! !

!FilePluginSimulator methodsFor: 'initialize-release' stamp: 'eem 10/29/2018 15:08'!
sqFileInit
	"See FilePluginSimulator>>sqFileStdioHandlesInto:"
	(openFiles := Dictionary new)
		at: 0 put: (FakeStdinStream for: interpreterProxy interpreter); "stdin"
		at: 1 put: interpreterProxy interpreter transcript; "stdout"
		at: 2 put: interpreterProxy interpreter transcript. "stderr"
	states := IdentityDictionary new.
	maxOpenFiles := VMClass initializationOptions at: #MaxFileDescriptors ifAbsent: [1024].
	^true! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 6/22/2011 12:56'!
sqFileRenameOld: oldNameIndex Size: oldNameSize New: newNameIndex Size: newNameSize
	| oldPath newPath |
	oldPath := FileDirectory default fullNameFor: (interpreterProxy interpreter asString: oldNameIndex size: oldNameSize).
	newPath := FileDirectory default fullNameFor: (interpreterProxy interpreter asString: newNameIndex size: newNameSize).
	((StandardFileStream isAFileNamed: oldPath)
	 and: [(StandardFileStream isAFileNamed: newPath) not]) ifFalse:
		[^interpreterProxy primitiveFail].
	[FileDirectory default primRename: oldPath to: newPath]
		on: Error
		do: [:ex| interpreterProxy primitiveFail]! !

!FilePluginSimulator methodsFor: 'initialize-release' stamp: 'eem 11/10/2020 11:54'!
sqFileShutdown
	self close.
	^true! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 6/22/2011 12:56'!
sqFileSize: file
	^file size! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'eem 6/21/2012 12:58'!
sqFileStdioHandlesInto: anArray
	(interpreterProxy transcript ~~ Transcript
	 or: [UIManager default confirm: 'clear transcript?']) ifTrue:
		[interpreterProxy transcript clear].
	"See FilePluginSimulator>>initialiseModule"
	anArray
		at: 1 put: 0;
		at: 2 put: 1;
		at: 3 put: 2.
	^7! !

!FilePluginSimulator methodsFor: 'simulation' stamp: 'monty 9/26/2015 22:26'!
sqFileSync: file
	^file sync! !

InterpreterPlugin subclass: #Float64ArrayPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!Float64ArrayPlugin commentStamp: 'tpr 5/2/2003 15:42' prior: 0!
Float64ArrayPlugin provides fast access to Float64Arrays for batch processing of double precision Float numbers!

!Float64ArrayPlugin class methodsFor: 'translation to C' stamp: 'ar 9/15/1998 00:30'!
declareCVarsIn: cg
	"Nothing to declare..."! !

!Float64ArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveAddFloat64Array
	"Primitive. Add the receiver and the argument, both Float64Arrays and store the result into the receiver."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg rcvrPtr argPtr length |
	<var: #rcvrPtr type: #'double *'>
	<var: #argPtr type: #'double *'>
	arg := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	((interpreterProxy isLong64s: arg)
	 and: [(interpreterProxy isLong64s: rcvr)
	 and: [(length := interpreterProxy stSizeOf: arg) = (interpreterProxy stSizeOf: rcvr)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	argPtr := self cCoerce: (interpreterProxy firstIndexableField: arg) to: #'double *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (rcvrPtr at: i) + (argPtr at: i)].
	^interpreterProxy methodReturnReceiver! !

!Float64ArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveAddScalar
	"Primitive. Add the argument, a scalar value to the receiver, a Float64Array"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr rcvrPtr value length |
	<var: #value type: #double>
	<var: #rcvrPtr type:#'double *'>
	value := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isLong64s: rcvr) ifFalse:
		[^interpreterProxy primitiveFail].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (rcvrPtr at: i) + value].
	^interpreterProxy methodReturnReceiver! !

!Float64ArrayPlugin methodsFor: 'access primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveAt
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| index rcvr doublePtr |
	<var: #doublePtr type: #'double *'>
	index := interpreterProxy stackIntegerValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy failed not
	 and: [(interpreterProxy isLong64s: rcvr)
	 and: [index > 0 and: [index <= (interpreterProxy slotSizeOf: rcvr)]]]) ifFalse:
		[^interpreterProxy primitiveFail].
	doublePtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	^interpreterProxy methodReturnFloat: (doublePtr at: index - 1)! !

!Float64ArrayPlugin methodsFor: 'access primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveAtPut
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| value floatValue index rcvr doublePtr |
	<var: #floatValue type: #double>
	<var: #doublePtr type: #'double *'>
	value := interpreterProxy stackValue: 0.
	floatValue := (interpreterProxy isIntegerObject: value)
					ifTrue: [(interpreterProxy integerValueOf: value) asFloat]
					ifFalse: [interpreterProxy floatValueOf: value].
	index := interpreterProxy stackIntegerValue: 1.
	rcvr := interpreterProxy stackValue: 2.
	(interpreterProxy failed not
	 and: [(interpreterProxy isLong64s: rcvr)
	 and: [index > 0 and: [index <= (interpreterProxy slotSizeOf: rcvr)]]]) ifFalse:
		[^interpreterProxy primitiveFail].
	doublePtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	doublePtr at: index-1 put: floatValue.
	^interpreterProxy methodReturnValue: value! !

!Float64ArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveDivFloat64Array
	"Primitive. Divide each element in the receiver by the corresponding element in the argument, both Float64Arrays, and store the result into the receiver."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg rcvrPtr argPtr length |
	<var: #rcvrPtr type: #'double *'>
	<var: #argPtr type: #'double *'>
	arg := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	((interpreterProxy isLong64s: arg)
	 and: [(interpreterProxy isLong64s: rcvr)
	 and: [(length := interpreterProxy stSizeOf: arg) = (interpreterProxy stSizeOf: rcvr)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	argPtr := self cCoerce: (interpreterProxy firstIndexableField: arg) to: #'double *'.
	"Check if any of the argument's values is zero"
	0 to: length - 1 do:
		[:i| (argPtr at: i) = 0.0 ifTrue: "i.e. check for both 0.0 and -0.0"
			[^interpreterProxy primitiveFail]].
	0 to: length - 1 do:
		[:i| rcvrPtr at: i put: (rcvrPtr at: i) / (argPtr at: i)].
	^interpreterProxy methodReturnReceiver! !

!Float64ArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveDivScalar
	"Primitive. Divide each element in the receiver by the argument, a scalar, and store the result into the receiver, a Float64Array"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr rcvrPtr value length |
	<var: #value type: #double>
	<var: #rcvrPtr type:#'double *'>
	value := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isLong64s: rcvr) ifFalse:
		[^interpreterProxy primitiveFail].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (rcvrPtr at: i) / value.
	].
	^interpreterProxy methodReturnReceiver! !

!Float64ArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveDotProduct
	"Primitive. Compute the dot product of the receiver and the argument, both Float64Arrays.
	The dot product is defined as the sum of the products of the individual elements."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg rcvrPtr argPtr length result |
	<var: #result type: #double>
	<var: #rcvrPtr type: #'double *'>
	<var: #argPtr type: #'double *'>
	arg := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	((interpreterProxy isLong64s: arg)
	 and: [(interpreterProxy isLong64s: rcvr)
	 and: [(length := interpreterProxy stSizeOf: arg) = (interpreterProxy stSizeOf: rcvr)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	argPtr := self cCoerce: (interpreterProxy firstIndexableField: arg) to: #'double *'.
	result := 0.0.
	0 to: length-1 do:[:i|
		result := result + ((rcvrPtr at: i) * (argPtr at: i)).
	].
	^interpreterProxy methodReturnFloat: result! !

!Float64ArrayPlugin methodsFor: 'access primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveEqual
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg rcvrPtr argPtr length |
	<var: #rcvrPtr type: #'double *'>
	<var: #argPtr type: #'double *'>
	arg := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	((interpreterProxy isLong64s: arg)
	 and: [(interpreterProxy isLong64s: rcvr)]) ifFalse:
		[^interpreterProxy primitiveFail].
	(length := interpreterProxy stSizeOf: arg) = (interpreterProxy stSizeOf: rcvr) ifFalse:
		[^interpreterProxy methodReturnBool: false].
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	argPtr := self cCoerce: (interpreterProxy firstIndexableField: arg) to: #'double *'.
	0 to: length-1 do:[:i|
		(rcvrPtr at: i) = (argPtr at: i) ifFalse:[^interpreterProxy methodReturnBool: false].
	].
	^interpreterProxy methodReturnBool: true! !

!Float64ArrayPlugin methodsFor: 'access primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveFromFloatArray
	"Primitive. Set each element of the receiver, a Float64Array with that of the argument, a FloatArray and return the receiver.
	Fail if both have different size"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg rcvrPtr argPtr length |
	<var: #rcvrPtr type: #'double *'>
	<var: #argPtr type: #'float *'>
	arg := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	((interpreterProxy isWords: arg)
	 and: [(interpreterProxy isLong64s: rcvr)
	 and: [(length := interpreterProxy stSizeOf: arg) = (interpreterProxy stSizeOf: rcvr)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	argPtr := self cCoerce: (interpreterProxy firstIndexableField: arg) to: #'float *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (self cCoerce: (argPtr at: i) to: #double)].
	^interpreterProxy methodReturnReceiver! !

!Float64ArrayPlugin methodsFor: 'access primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveHashArray
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr rcvrPtr length result |
	<var: #rcvrPtr type: #'unsigned int *'>
	<var: #result type: #'unsigned int'>
	rcvr := interpreterProxy stackValue: 0.
	(interpreterProxy isLong64s: rcvr) ifFalse:
		[^interpreterProxy primitiveFail].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'unsigned int *'.
	result := 0.
	0 to: length*2-1 do:[:i|
		result := result + (rcvrPtr at: i).
	].
	^interpreterProxy methodReturnInteger: (result bitAnd: 16r1FFFFFFF)! !

!Float64ArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveLength
	"Primitive. Compute the length of the argument (sqrt of sum of component squares)."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr rcvrPtr length len |
	<var: #rcvrPtr type: #'double *'>
	<var: #len type: #double>
	rcvr := interpreterProxy stackValue: 0.
	(interpreterProxy isLong64s: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	len := 0.0.
	0 to: length-1 do:
		[:i| len := len + ((rcvrPtr at: i) * (rcvrPtr at: i)) ].
	len > 0.0 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	^interpreterProxy methodReturnFloat: (self sqrt: len)! !

!Float64ArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveMulFloat64Array
	"Primitive. Multiply the receiver and the argument, both Float64Arrays and store the result into the receiver."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg rcvrPtr argPtr length |
	<var: #rcvrPtr type: #'double *'>
	<var: #argPtr type: #'double *'>
	arg := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	((interpreterProxy isLong64s: arg)
	 and: [(interpreterProxy isLong64s: rcvr)
	 and: [(length := interpreterProxy stSizeOf: arg) = (interpreterProxy stSizeOf: rcvr)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	argPtr := self cCoerce: (interpreterProxy firstIndexableField: arg) to: #'double *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (rcvrPtr at: i) * (argPtr at: i).
	].
	^interpreterProxy methodReturnReceiver! !

!Float64ArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveMulScalar
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	"Primitive. Multiply elements in the receiver, a Float64Array, by argument, a scalar value, and store the result into the receiver."
	| rcvr rcvrPtr value length |
	<var: #value type: #double>
	<var: #rcvrPtr type:#'double *'>
	value := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isLong64s: rcvr) ifFalse:
		[^interpreterProxy primitiveFail].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (rcvrPtr at: i) * value.
	].
	^interpreterProxy methodReturnReceiver! !

!Float64ArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveNormalize
	"Primitive. Normalize the argument (A Float64Array) in place."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr rcvrPtr length len |
	<var: #rcvrPtr type: #'double *'>
	<var: #len type: #double>
	rcvr := interpreterProxy stackValue: 0.
	(interpreterProxy isLong64s: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	len := 0.0.
	0 to: length - 1 do:
		[:i| len := len + ((rcvrPtr at: i) * (rcvrPtr at: i)) ].
	len > 0.0 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].

	len := self sqrt: len.
	0 to: length - 1 do:
		[:i| rcvrPtr at: i put: (rcvrPtr at: i) / len ].

	"Leave receiver on the stack."! !

!Float64ArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveSubFloat64Array
	"Primitive. Subtract each element in the argument from each element in the receiver, both Float64Arrays and store the result into the receiver."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg rcvrPtr argPtr length |
	<var: #rcvrPtr type: #'double *'>
	<var: #argPtr type: #'double *'>
	arg := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	((interpreterProxy isLong64s: arg)
	 and: [(interpreterProxy isLong64s: rcvr)
	 and: [(length := interpreterProxy stSizeOf: arg) = (interpreterProxy stSizeOf: rcvr)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	argPtr := self cCoerce: (interpreterProxy firstIndexableField: arg) to: #'double *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (rcvrPtr at: i) - (argPtr at: i).
	].
	^interpreterProxy methodReturnReceiver! !

!Float64ArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveSubScalar
	"Primitive. Subtract the argument, a scalar value from  each element in the receiver, a Float64Array"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr rcvrPtr value length |
	<var: #value type: #double>
	<var: #rcvrPtr type:#'double *'>
	value := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isLong64s: rcvr) ifFalse:
		[^interpreterProxy primitiveFail].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (rcvrPtr at: i) - value.
	].
	^interpreterProxy methodReturnReceiver! !

!Float64ArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveSum
	"Primitive. Answer the sum of each float in the receiver, a Float64Array."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr rcvrPtr length sum |
	<var: #sum type: #double>
	<var: #rcvrPtr type: #'double *'>
	rcvr := interpreterProxy stackValue: 0.
	(interpreterProxy isLong64s: rcvr) ifFalse:
		[^interpreterProxy primitiveFail].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'double *'.
	sum := 0.0.
	0 to: length-1 do:[:i|
		sum := sum + (rcvrPtr at: i).
	].
	^interpreterProxy methodReturnFloat: sum! !

!Float64ArrayPlugin methodsFor: 'simulation' stamp: 'eem 5/26/2016 12:38'!
sqrt: aFloat
	<doNotGenerate>
	^aFloat sqrt! !

InterpreterPlugin subclass: #FloatArrayPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!FloatArrayPlugin commentStamp: 'tpr 5/2/2003 15:42' prior: 0!
FloatArrayPlugin provides fast access to FloatArrays for batch processing of float numbers!

!FloatArrayPlugin class methodsFor: 'translation to C' stamp: 'ar 9/15/1998 00:30'!
declareCVarsIn: cg
	"Nothing to declare..."! !

!FloatArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveAddFloatArray
	"Primitive. Add the receiver and the argument, both FloatArrays and store the result into the receiver."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg rcvrPtr argPtr length |
	<var: #rcvrPtr type: #'float *'>
	<var: #argPtr type: #'float *'>
	arg := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	((interpreterProxy isWords: arg)
	 and: [(interpreterProxy isWords: rcvr)
	 and: [(length := interpreterProxy stSizeOf: arg) = (interpreterProxy stSizeOf: rcvr)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	argPtr := self cCoerce: (interpreterProxy firstIndexableField: arg) to: #'float *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (self cCoerce: (rcvrPtr at: i) to: #double) + (self cCoerce: (argPtr at: i) to: #double)].
	^interpreterProxy methodReturnReceiver! !

!FloatArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveAddScalar
	"Primitive. Add the argument, a scalar value to the receiver, a FloatArray"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr rcvrPtr value length |
	<var: #value type: #double>
	<var: #rcvrPtr type:#'float *'>
	value := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isWords: rcvr) ifFalse:
		[^interpreterProxy primitiveFail].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (self cCoerce: (rcvrPtr at: i) to: #double) + value].
	^interpreterProxy methodReturnReceiver! !

!FloatArrayPlugin methodsFor: 'access primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveAt
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| index rcvr floatPtr |
	<var: #floatPtr type: #'float *'>
	index := interpreterProxy stackIntegerValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy failed not
	 and: [(interpreterProxy isWords: rcvr)
	 and: [index > 0 and: [index <= (interpreterProxy slotSizeOf: rcvr)]]]) ifFalse:
		[^interpreterProxy primitiveFail].
	floatPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	^interpreterProxy methodReturnFloat: (floatPtr at: index - 1)! !

!FloatArrayPlugin methodsFor: 'access primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveAtPut
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| value floatValue index rcvr floatPtr |
	<var: #floatValue type: #double>
	<var: #floatPtr type: #'float *'>
	value := interpreterProxy stackValue: 0.
	floatValue := (interpreterProxy isIntegerObject: value)
					ifTrue: [(interpreterProxy integerValueOf: value) asFloat]
					ifFalse: [interpreterProxy floatValueOf: value].
	index := interpreterProxy stackIntegerValue: 1.
	rcvr := interpreterProxy stackValue: 2.
	(interpreterProxy failed not
	 and: [(interpreterProxy isWords: rcvr)
	 and: [index > 0 and: [index <= (interpreterProxy slotSizeOf: rcvr)]]]) ifFalse:
		[^interpreterProxy primitiveFail].
	floatPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	floatPtr at: index-1 put: (self cCoerce: floatValue to: #float).
	^interpreterProxy methodReturnValue: value! !

!FloatArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveDivFloatArray
	"Primitive. Divide each element in the receiver by the corresponding element in the argument, both FloatArrays, and store the result into the receiver."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg rcvrPtr argPtr length |
	<var: #rcvrPtr type: #'float *'>
	<var: #argPtr type: #'float *'>
	arg := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	((interpreterProxy isWords: arg)
	 and: [(interpreterProxy isWords: rcvr)
	 and: [(length := interpreterProxy stSizeOf: arg) = (interpreterProxy stSizeOf: rcvr)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	argPtr := self cCoerce: (interpreterProxy firstIndexableField: arg) to: #'float *'.
	"Check if any of the argument's values is zero"
	0 to: length - 1 do:
		[:i| (argPtr at: i) = (self cCoerce: 0.0 to: #float) ifTrue: "i.e. check for both 0.0 and -0.0"
			[^interpreterProxy primitiveFail]].
	0 to: length - 1 do:
		[:i| rcvrPtr at: i put: (self cCoerce: (rcvrPtr at: i) to: #double) / (self cCoerce: (argPtr at: i) to: #double)].
	^interpreterProxy methodReturnReceiver! !

!FloatArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveDivScalar
	"Primitive. Divide each element in the receiver by the argument, a scalar, and store the result into the receiver, a FloatArray"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr rcvrPtr value inverse length |
	<var: #inverse type:'double '>
	<var: #value type: #double>
	<var: #rcvrPtr type:#'float *'>
	value := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isWords: rcvr) ifFalse:
		[^interpreterProxy primitiveFail].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	inverse := 1.0 / value.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (self cCoerce: (rcvrPtr at: i) to: #double) * inverse.
	].
	^interpreterProxy methodReturnReceiver! !

!FloatArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveDotProduct
	"Primitive. Compute the dot product of the receiver and the argument.
	The dot product is defined as the sum of the products of the individual elements."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg rcvrPtr argPtr length result |
	<var: #result type: #double>
	<var: #rcvrPtr type: #'float *'>
	<var: #argPtr type: #'float *'>
	arg := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	((interpreterProxy isWords: arg)
	 and: [(interpreterProxy isWords: rcvr)
	 and: [(length := interpreterProxy stSizeOf: arg) = (interpreterProxy stSizeOf: rcvr)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	argPtr := self cCoerce: (interpreterProxy firstIndexableField: arg) to: #'float *'.
	result := 0.0.
	0 to: length-1 do:[:i|
		result := result + ((self cCoerce: (rcvrPtr at: i) to: #double) * (self cCoerce: (argPtr at: i) to: #double)).
	].
	^interpreterProxy methodReturnFloat: result! !

!FloatArrayPlugin methodsFor: 'access primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveEqual
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg rcvrPtr argPtr length |
	<var: #rcvrPtr type: #'float *'>
	<var: #argPtr type: #'float *'>
	arg := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	((interpreterProxy isWords: arg)
	 and: [(interpreterProxy isWords: rcvr)]) ifFalse:
		[^interpreterProxy primitiveFail].
	(length := interpreterProxy stSizeOf: arg) = (interpreterProxy stSizeOf: rcvr) ifFalse:
		[^interpreterProxy methodReturnBool: false].
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	argPtr := self cCoerce: (interpreterProxy firstIndexableField: arg) to: #'float *'.
	0 to: length-1 do:[:i|
		(rcvrPtr at: i) = (argPtr at: i) ifFalse:[^interpreterProxy methodReturnBool: false].
	].
	^interpreterProxy methodReturnBool: true! !

!FloatArrayPlugin methodsFor: 'access primitives' stamp: 'eem 9/5/2021 06:54'!
primitiveFromFloat64Array
	"Primitive. Set each element of the receiver, a FloatArray with that of the argument, a Float64Array and return the receiver.
	Note that this conversion might loose bits, or generate overflow.
	Fail if both have different size"
	<option: #SpurObjectMemory>
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg rcvrPtr argPtr length |
	<var: #rcvrPtr type: #'float *'>
	<var: #argPtr type: #'double *'>
	arg := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	((interpreterProxy isLong64s: arg)
	 and: [(interpreterProxy isWords: rcvr)
	 and: [(length := interpreterProxy stSizeOf: arg) = (interpreterProxy stSizeOf: rcvr)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	argPtr := self cCoerce: (interpreterProxy firstIndexableField: arg) to: #'double *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (self cCoerce: (argPtr at: i) to: #float)].
	^interpreterProxy methodReturnReceiver! !

!FloatArrayPlugin methodsFor: 'access primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveHashArray
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr rcvrPtr length result |
	<var: #rcvrPtr type: #'unsigned int *'>
	<var: #result type: #'unsigned int'>
	rcvr := interpreterProxy stackValue: 0.
	(interpreterProxy isWords: rcvr) ifFalse:
		[^interpreterProxy primitiveFail].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'unsigned int *'.
	result := 0.
	0 to: length-1 do:[:i|
		result := result + (rcvrPtr at: i).
	].
	^interpreterProxy methodReturnInteger: (result bitAnd: 16r1FFFFFFF)! !

!FloatArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveLength
	"Primitive. Compute the length of the argument (sqrt of sum of component squares)."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr rcvrPtr length len |
	<var: #rcvrPtr type: #'float *'>
	<var: #len type: #double>
	rcvr := interpreterProxy stackValue: 0.
	(interpreterProxy isWords: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	len := 0.0.
	0 to: length-1 do:
		[:i| len := len + ((self cCoerce: (rcvrPtr at: i) to: #double) * (self cCoerce: (rcvrPtr at: i) to: #double)) ].
	len > 0.0 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	^interpreterProxy methodReturnFloat: (self sqrt: len)! !

!FloatArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveMulFloatArray
	"Primitive. Multiply the receiver and the argument, both FloatArrays and store the result into the receiver."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg rcvrPtr argPtr length |
	<var: #rcvrPtr type: #'float *'>
	<var: #argPtr type: #'float *'>
	arg := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	((interpreterProxy isWords: arg)
	 and: [(interpreterProxy isWords: rcvr)
	 and: [(length := interpreterProxy stSizeOf: arg) = (interpreterProxy stSizeOf: rcvr)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	argPtr := self cCoerce: (interpreterProxy firstIndexableField: arg) to: #'float *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (self cCoerce: (rcvrPtr at: i) to: #double) * (self cCoerce: (argPtr at: i) to: #double).
	].
	^interpreterProxy methodReturnReceiver! !

!FloatArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveMulScalar
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	"Primitive. Multiply elements in the receiver, a FloatArray, by argument, a scalar value, and store the result into the receiver."
	| rcvr rcvrPtr value length |
	<var: #value type: #double>
	<var: #rcvrPtr type:#'float *'>
	value := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isWords: rcvr) ifFalse:
		[^interpreterProxy primitiveFail].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (self cCoerce: (rcvrPtr at: i) to: #double) * value.
	].
	^interpreterProxy methodReturnReceiver! !

!FloatArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveNormalize
	"Primitive. Normalize the argument (A FloatArray) in place."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr rcvrPtr length len |
	<var: #rcvrPtr type: #'float *'>
	<var: #len type: #double>
	rcvr := interpreterProxy stackValue: 0.
	(interpreterProxy isWords: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	len := 0.0.
	0 to: length - 1 do:
		[:i| len := len + ((self cCoerce: (rcvrPtr at: i) to: #double) * (self cCoerce: (rcvrPtr at: i) to: #double)) ].
	len > 0.0 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].

	len := self sqrt: len.
	0 to: length - 1 do:
		[:i| rcvrPtr at: i put: ((self cCoerce: (rcvrPtr at: i) to: #double) / len) ].

	"Leave receiver on the stack."! !

!FloatArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveSubFloatArray
	"Primitive. Subtract each element in the argument from each element in the receiver, both FloatArrays and store the result into the receiver."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg rcvrPtr argPtr length |
	<var: #rcvrPtr type: #'float *'>
	<var: #argPtr type: #'float *'>
	arg := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	((interpreterProxy isWords: arg)
	 and: [(interpreterProxy isWords: rcvr)
	 and: [(length := interpreterProxy stSizeOf: arg) = (interpreterProxy stSizeOf: rcvr)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	argPtr := self cCoerce: (interpreterProxy firstIndexableField: arg) to: #'float *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (self cCoerce: (rcvrPtr at: i) to: #double) - (self cCoerce: (argPtr at: i) to: #double).
	].
	^interpreterProxy methodReturnReceiver! !

!FloatArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveSubScalar
	"Primitive. Subtract the argument, a scalar value from  each element in the receiver, a FloatArray"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr rcvrPtr value length |
	<var: #value type: #double>
	<var: #rcvrPtr type:#'float *'>
	value := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isWords: rcvr) ifFalse:
		[^interpreterProxy primitiveFail].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	0 to: length-1 do:[:i|
		rcvrPtr at: i put: (self cCoerce: (rcvrPtr at: i) to: #double) - value.
	].
	^interpreterProxy methodReturnReceiver! !

!FloatArrayPlugin methodsFor: 'arithmetic primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveSum
	"Primitive. Answer the sum of each float in the receiver, a FloatArray."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr rcvrPtr length sum |
	<var: #sum type: #double>
	<var: #rcvrPtr type: #'float *'>
	rcvr := interpreterProxy stackValue: 0.
	(interpreterProxy isWords: rcvr) ifFalse:
		[^interpreterProxy primitiveFail].
	length := interpreterProxy stSizeOf: rcvr.
	rcvrPtr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'float *'.
	sum := 0.0.
	0 to: length-1 do:[:i|
		sum := sum + (self cCoerce: (rcvrPtr at: i) to: #double).
	].
	^interpreterProxy methodReturnFloat: sum! !

!FloatArrayPlugin methodsFor: 'simulation' stamp: 'eem 5/26/2016 12:38'!
sqrt: aFloat
	<doNotGenerate>
	^aFloat sqrt! !

InterpreterPlugin subclass: #FloatMathPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!FloatMathPlugin class methodsFor: 'translation' stamp: 'ar 11/6/2005 02:48'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!FloatMathPlugin class methodsFor: 'translation' stamp: 'ar 11/6/2005 16:20'!
requiresCrossPlatformFiles
	"default is ok for most, any plugin needing platform specific files must say so"
	^true! !

!FloatMathPlugin class methodsFor: 'simulation' stamp: 'eem 10/21/2008 10:32'!
simulatorClass
	^FloatMathPluginSimulator! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 2/12/2015 11:15'!
floatExponentBits
	"Answer the number of bits in the double-precision exponent.  This is an 11-bit field."
	<inline: true>
	^11! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 2/11/2015 17:49'!
isFinite: aDouble
	<var: #aDouble type: #double>
	<inline: true>
	^aDouble - aDouble = 0.0! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveArcCos
	"Computes acos(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_acos(rcvr)' inSmalltalk: [rcvr arcCos].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveArcCosH
	"Computes acosh(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_acosh(rcvr)' inSmalltalk: [rcvr arcCosH].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveArcSin
	"Computes asin(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_asin(rcvr)' inSmalltalk: [rcvr arcSin].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveArcSinH
	"Computes asinh(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_asinh(rcvr)' inSmalltalk: [rcvr arcSinH].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveArcTan
	"Computes atan(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_atan(rcvr)' inSmalltalk: [rcvr arcTan].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveArcTan2
	"Computes atan2(receiver, arg)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg result |
	<var: #rcvr type: #double>
	<var: #arg type: #double>
	<var: #result type: #double>
	arg := interpreterProxy stackFloatValue: 0.
	rcvr := interpreterProxy stackFloatValue: 1.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_atan2(rcvr, arg)' inSmalltalk: [rcvr arcTan: arg].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveArcTanH
	"Computes atanh(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_atanh(rcvr)' inSmalltalk: [rcvr arcTanH].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveCos
	"Computes cos(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_cos(rcvr)' inSmalltalk: [rcvr cos].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveCosH
	"Computes cosh(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_cosh(rcvr)' inSmalltalk: [rcvr cosH].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveExp
	"Computes E raised to the receiver power."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := (self cCode: '__ieee754_exp(rcvr)' inSmalltalk: [rcvr exp]).
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveFMod
	"Computes receiver \\ arg"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg result |
	<var: #rcvr type: #double>
	<var: #arg type: #double>
	<var: #result type: #double>
	arg := interpreterProxy stackFloatValue: 0.
	rcvr := interpreterProxy stackFloatValue: 1.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_fmod(rcvr, arg)' inSmalltalk: [rcvr \\ arg].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveFractionalPart
	"Computes receiver \\ 1.0"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result trunc |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	<var: #trunc type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_modf(rcvr, &trunc)' inSmalltalk: [rcvr fractionPart].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveHypot
	"hypot(x,y) returns sqrt(x^2+y^2) with error less  than 1 ulps"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg result |
	<var: #rcvr type: #double>
	<var: #arg type: #double>
	<var: #result type: #double>
	arg := interpreterProxy stackFloatValue: 0.
	rcvr := interpreterProxy stackFloatValue: 1.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_hypot(rcvr, arg)' inSmalltalk: [rcvr hypot: arg].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveLog10
	"Computes log10(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	rcvr < 0.0 ifTrue:[^interpreterProxy primitiveFail].
	result := self cCode: '__ieee754_log10(rcvr)' inSmalltalk: [rcvr log: 10].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveLogN
	"Computes log(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	rcvr < 0.0 ifTrue:[^interpreterProxy primitiveFail].
	result := self cCode: '__ieee754_log(rcvr)' inSmalltalk: [rcvr ln].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveRaisedToPower
	"Computes receiver**arg"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg result |
	<var: #rcvr type: #double>
	<var: #arg type: #double>
	<var: #result type: #double>
	arg := interpreterProxy stackFloatValue: 0.
	rcvr := interpreterProxy stackFloatValue: 1.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_pow(rcvr, arg)' inSmalltalk: [rcvr raisedTo: arg].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveSin
	"Computes sin(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_sin(rcvr)' inSmalltalk: [rcvr sin].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveSinH
	"Computes sinh(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_sinh(rcvr)' inSmalltalk: [rcvr sinH].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveSqrt
	"Computes sqrt(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	rcvr < 0.0 ifTrue:[^interpreterProxy primitiveFail].
	result := self cCode: '__ieee754_sqrt(rcvr)' inSmalltalk: [rcvr sqrt].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveTan
	"Computes tan(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_tan(rcvr)' inSmalltalk: [rcvr tan].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveTanH
	"Computes tanh(receiver)"
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	rcvr := interpreterProxy stackFloatValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: '__ieee754_tanh(rcvr)' inSmalltalk: [rcvr tanH].
	(self isnan: result) ifTrue:[^interpreterProxy primitiveFail].
	^interpreterProxy methodReturnFloat: result! !

!FloatMathPlugin methodsFor: 'float primitives' stamp: 'eem 08/02/2021 14:36'!
primitiveTimesTwoPower
	"Multiply the receiver by the power of the argument."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| rcvr arg result |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	arg := interpreterProxy stackIntegerValue: 0.
	rcvr := interpreterProxy stackFloatValue: 1.
	interpreterProxy failed ifTrue:
		[^nil].
	interpreterProxy bytesPerOop > 4 ifTrue:
		[| twiceMaxExponent | "clip arg to at most int range; ldexp's last arg is of type int"
		 twiceMaxExponent := 2 * (1 << self floatExponentBits).
	 	 arg < twiceMaxExponent negated
			ifTrue: [arg := twiceMaxExponent negated]
			ifFalse: [arg > twiceMaxExponent ifTrue:
						[arg := twiceMaxExponent]]].
	result := self cCode: '__ieee754_ldexp(rcvr, (int)arg)'
					inSmalltalk: [rcvr timesTwoPower: arg].
	(self isnan: result) ifFalse:
		[interpreterProxy methodReturnFloat: result].
	^nil! !

FloatMathPlugin subclass: #FloatMathPluginSimulator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!FloatMathPluginSimulator commentStamp: '<historical>' prior: 0!
Float math plugin simulation for the VM simulator!

!FloatMathPluginSimulator class methodsFor: 'translation' stamp: 'eem 10/21/2008 12:30'!
shouldBeTranslated
"This class should not be translated"
	^false! !

!FloatMathPluginSimulator methodsFor: 'float primitives' stamp: 'eem 10/21/2008 10:34'!
isnan: result
	^result isNaN! !

InterpreterPlugin subclass: #IA32ABIPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: 'VMBasicConstants'
	category: 'VMMaker-Plugins-FFI'!

!IA32ABIPlugin commentStamp: 'eem 1/25/2016 11:59' prior: 0!
This plugin implements the Alien foreign-function interface, a small elaboration on the Strongtalk FFI.

Call-outs are performed by a small number of primitives, one each for the four different kinds of return linkage on x86.  The primitives are var-args.  Each primitive has a signature something like:
primFFICall: functionAddress <Alien> result: result <Alien> with: firstArg <Alien> ... with: lastArg <Alien>
	<primitive: 'primCallOutIntegralReturn' module: 'IA32ABI'>
which arranges to call-out supplying the arguments to the function pointed to by functionAddress, copying its return value into result.  The call-out primitives are as follows:

primCallOutIntegralReturn call a function which returns up to 8 bytes in %eax & %edx, taking up to the first 4 bytes from %eax.  i.e. if the sizeof(result) is 4 or less only bytes from %eax will be returned, but if more then the first 4 bytes of result will be assigned with %eax and subsequent bytes with %edx, up to a total of 8 bytes.

primCallOutPointerReturn call a function which returns a pointer in %eax.  Assign sizeof(result) bytes from this pointer into the result.

primCallOutFloatReturn call a function which returns a 4 byte single-precision float in %f0, assigning the 4 bytes of %f0 into result.

primCallOutDoubleReturn call a function which returns an 8 byte double-precision float in %f0, assigning the 8 bytes of %f0 into result.

!

!IA32ABIPlugin class methodsFor: 'translation' stamp: 'eem 6/7/2018 15:34'!
ancilliaryClasses
	"Answer any extra classes to be included in the translation."
	^{ VMCallbackContext. VMCallbackReturnValue }! !

!IA32ABIPlugin class methodsFor: 'translation' stamp: 'eem 3/11/2017 14:27'!
declareCVarsIn: aCCodeGen
	aCCodeGen
		addHeaderFile: '<setjmp.h>';
		addHeaderFile: '"vmCallback.h"';
		addHeaderFile: '"ia32abi.h"';
		addHeaderFile: '"sqAssert.h"'! !

!IA32ABIPlugin class methodsFor: 'accessing' stamp: 'John M McIntosh 11/21/2008 13:14'!
hasHeaderFile
	^true! !

!IA32ABIPlugin class methodsFor: 'translation' stamp: 'eem 10/31/2007 18:05'!
moduleName
	^'IA32ABI'! !

!IA32ABIPlugin class methodsFor: 'translation' stamp: 'eem 10/14/2010 09:45'!
shouldGenerateTypedefFor: aStructClass
	"Hack to work-around mutliple definitions.  Sometimes a type has been defined in an include."
	^aStructClass ~~ VMCallbackContext! !

!IA32ABIPlugin class methodsFor: 'simulation only' stamp: 'eem 11/15/2011 13:08'!
simulatorClass
	^NewspeakVM ifFalse: [IA32ABIPluginSimulator]! !

!IA32ABIPlugin methodsFor: 'private-support' stamp: 'eem 10/20/2015 16:00:58'!
index: byteIndex length: length inRange: rcvr
	"Answer if the indices byteIndex to byteIndex + length - 1 are valid zero-relative indices into the rcvr.
	 Beware!!  There be dragons here.  The form below (byteIndex <= (dataSize abs - length)) is used
	 because byteIndex + length could overflow, whereas (dataSize abs - length) can't.  We *don't* use the
	 obvious optimization
		^dataSize = 0 or: [byteIndex asUnsignedInteger <= (dataSize abs - length)]
	 because with C's Usual Arithmetic Conversions
		5. Otherwise, both operands are converted to the unsigned integer type corresponding to the type of the operand with signed integer type.
	 means that the comparison will be unsigned, and if length > dataSize abs then dataSize abs - length is large and positive."
	| dataSize |
	<inline: true>
	dataSize := self sizeField: rcvr.
	^dataSize = 0 or: [byteIndex >= 0 and: [byteIndex <= (dataSize abs - length)]]! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 9/27/2020 14:02'!
primAddressField
	"Answer the unsigned 32-bit (or 64-bit) integer comprising the address field (the second 32-bit or 64-bit field)."
	"<Alien> primAddressField ^<Integer>
		<primitive: 'primAddressField' error: errorCode module: 'IA32ABI'>"
	| rcvr value |
	<export: true>
	rcvr := interpreterProxy stackValue: 0.
	value := self longAt: rcvr + interpreterProxy baseHeaderSize + interpreterProxy bytesPerOop.
	^interpreterProxy methodReturnValue: (self positiveMachineIntegerFor: value)! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 12/12/2014 15:46'!
primAddressFieldPut
	"Store an unsigned integer into the size field (the second 32/64 bit field; little endian)."
	"<Alien> addressFieldPut: value <Integer> ^<Integer>
		<primitive: 'primAddressFieldPut' error: errorCode module: 'IA32ABI'>"
	| rcvr value valueOop |
	<export: true>
	valueOop := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	value := interpreterProxy positiveMachineIntegerValueOf: valueOop.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	self longAt: rcvr + interpreterProxy baseHeaderSize + interpreterProxy bytesPerOop put: value.
	^interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 7/19/2021 15:02'!
primAlienCopyInto
	"Copy some number of bytes from the receiver starting at the first index into some destination
	 object starting at the second index.  The  destination may be an Aliens or a bit-indexable object.
	 The primitive will have the following signature:
	<Alien>
		primCopyFrom: start <Integer>
		to: stop <Integer>
		into: destination <Alien | indexableByteSubclass et al>
		startingAt: destStart <Integer> ^<self>
		<primitive: 'primitiveAlienReplace' error: errorCode module: 'IA32ABI'>
	"
	<export: true>
	| alien start stop dest destStart src totalLength destAddr myLength |
	alien		:= interpreterProxy stackValue: 4.  "Unchecked!!"
	start		:= interpreterProxy stackValue: 3.
	stop		:= interpreterProxy stackValue: 2.
	dest		:= interpreterProxy stackValue: 1.
	destStart	:= interpreterProxy stackValue: 0.

	((interpreterProxy isIntegerObject: start)
	 and: [(interpreterProxy isIntegerObject: stop)
	 and: [(interpreterProxy isIntegerObject: destStart)
	 and: [interpreterProxy isWordsOrBytes: dest]]]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	start		:= interpreterProxy integerValueOf: start.
	stop		:= interpreterProxy integerValueOf: stop.
	destStart	:= interpreterProxy integerValueOf: destStart.

	myLength := self sizeField: alien.
	src := (self startOfData: alien withSize: myLength) + start - 1.

	(self isAlien: dest)
		ifTrue:
			[totalLength := self sizeField: dest.
			 destAddr := (self startOfData: dest withSize: totalLength) + destStart - 1.
			 totalLength = 0 "no bounds checks for zero-sized (pointer) Aliens"
				ifTrue: [totalLength := stop]
				ifFalse: [totalLength := totalLength abs]]
		ifFalse:
			[totalLength := interpreterProxy byteSizeOf: dest.
			 destAddr := (self startOfByteData: dest) + destStart - 1].

	((start >= 1 and: [start - 1 <= stop and: [stop <= myLength abs]])
	 and: [stop - start + 1 <= totalLength]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].

	(interpreterProxy isOopImmutable: dest) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].

	"Use memmove to allow source and destination to overlap"
	self memmove: destAddr asVoidPointer _: src asVoidPointer _: stop - start + 1.

	interpreterProxy methodReturnReceiver! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 7/19/2021 15:02'!
primAlienReplace
	"Copy some number of bytes from some source object starting at the index
	 into the receiver destination object from startIndex to stopIndex.  The  source
	 and destination may be Aliens or byte-indexable objects.  The primitive will have
	 either of the following signatures:
	<Alien | indexableByteSubclass | indexableWordSubclass>
		primReplaceFrom: start <Integer>
		to: stop <Integer>
		with: replacement <Alien | indexableByteSubclass | indexableWordSubclass | Integer>
		startingAt: repStart <Integer> ^<self>
		<primitive: 'primitiveAlienReplace' error: errorCode module: 'IA32ABI'>
	<Anywhere>
		primReplaceIn: dest <Alien | indexableByteSubclass | indexableWordSubclass>
		from: start <Integer>
		to: stop <Integer>
		with: replacement <Alien | indexableByteSubclass | indexableWordSubclass | Integer>
		startingAt: repStart <Integer> ^<self>
		<primitive: 'primitiveAlienReplace' error: errorCode module: 'IA32ABI'>
	"
	| array start stop repl replStart dest src totalLength isOnExecutablePage |
	<export: true>
	array		:= interpreterProxy stackValue: 4.  "Unchecked!!"
	start		:= interpreterProxy stackValue: 3.
	stop		:= interpreterProxy stackValue: 2.
	repl			:= interpreterProxy stackValue: 1.
	replStart	:= interpreterProxy stackValue: 0.

	((interpreterProxy isIntegerObject: start)
	 and: [(interpreterProxy isIntegerObject: stop)
	 and: [(interpreterProxy isIntegerObject: replStart)
	 and: [interpreterProxy isWordsOrBytes: repl]]]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	start		:= interpreterProxy integerValueOf: start.
	stop		:= interpreterProxy integerValueOf: stop.
	replStart	:= interpreterProxy integerValueOf: replStart.

	(self isAlien: array)
		ifTrue:
			[totalLength := self sizeField: array.
			 dest := (self startOfData: array withSize: totalLength) + start - 1.
			 totalLength = 0 "no bounds checks for zero-sized (pointer) Aliens"
				ifTrue: [totalLength := stop]
				ifFalse: [totalLength := totalLength abs]]
		ifFalse:
			[totalLength := interpreterProxy byteSizeOf: array.
			 dest := (self startOfByteData: array) + start - 1].
	(start >= 1 and: [start - 1 <= stop and: [stop <= totalLength]]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].

	(interpreterProxy isKindOfInteger: repl)
		ifTrue:
			[src := (interpreterProxy positiveMachineIntegerValueOf: repl) + replStart - 1.
			 interpreterProxy failed ifTrue:
				[^interpreterProxy primitiveFailFor: PrimErrBadArgument]]
		ifFalse:
			[(self isAlien: repl)
				ifTrue:
					[totalLength := self sizeField: repl.
					 src := (self startOfData: repl withSize: totalLength) + replStart - 1.
					 totalLength = 0 "no bounds checks for zero-sized (pointer) Aliens"
						ifTrue: [totalLength := stop - start + replStart]
						ifFalse: [totalLength := totalLength abs]]
				ifFalse:
					[(interpreterProxy isWordsOrBytes: repl) ifFalse:
						[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
					 totalLength := interpreterProxy byteSizeOf: repl.
					 src := (self startOfByteData: repl) + replStart - 1].
			(replStart >= 1 and: [stop - start + replStart <= totalLength]) ifFalse:
				[^interpreterProxy primitiveFailFor: PrimErrBadIndex]].

	(interpreterProxy isOopImmutable: array) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].

	isOnExecutablePage := self ifIsWithinExecutablePageMakePageWritable: dest asVoidPointer.
	"Use memmove to allow source and destination to overlap"
	self memmove: dest asVoidPointer _: src asVoidPointer _: stop - start + 1.
	isOnExecutablePage ifTrue: [self makePageExecutableAgain: dest asVoidPointer].

	interpreterProxy methodReturnReceiver! !

!IA32ABIPlugin methodsFor: 'primitives-memory management' stamp: 'eem 4/27/2020 13:16'!
primAllocateExecutablePage
	"Answer an Alien for an executable page; for thunks"
	"primAllocateExecutablePage ^<Alien>
		<primitive: 'primAllocateExecutablePage' error: errorCode module: 'IA32ABI'>"
	| byteSize ptr mem alien |
	<export: true>
	<var: #byteSize type: #'sqIntptr_t'>
	<var: #ptr type: #'sqIntptr_t *'>
	<var: #mem type: #'void *'>

	mem := self allocateExecutablePage: (self addressOf: byteSize put: [:v| byteSize := v]).
	mem = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoCMemory].
	alien := interpreterProxy
				instantiateClass: interpreterProxy classAlien
				indexableSize: 2 * interpreterProxy bytesPerOop.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoMemory].
	ptr := interpreterProxy firstIndexableField: alien.
	ptr at: 0 put: 0 - byteSize. "indirect mem indicated by negative size. Slang doesn't grok negated"
	ptr at: 1 put: (self cCoerce: mem to: #'sqIntptr_t').
	interpreterProxy methodReturnValue: alien! !

!IA32ABIPlugin methodsFor: 'primitives-memory management' stamp: 'eem 9/27/2020 14:03'!
primBoxedFree
	"Free the memory referenced by the receiver, an Alien."
	"proxy <Alien> primFree ^<Alien>
		<primitive: 'primBoxedFree' error: errorCode module: 'IA32ABI'>"
	| addr rcvr ptr sizeField |
	<export: true>
	<var: #ptr type: #'sqIntptr_t *'>
	<var: #addr type: #'sqIntptr_t'>
	<var: #sizeField type: #'sqIntptr_t'>

	rcvr := interpreterProxy stackValue: 0.
	(interpreterProxy byteSizeOf: rcvr) >= (2 * interpreterProxy bytesPerOop) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	ptr := interpreterProxy firstIndexableField: rcvr.
	sizeField := ptr at: 0.
	addr := ptr at: 1.
	"Don't you dare to free Squeak's memory!!"
	(sizeField >= 0 or: [addr = 0 or: [interpreterProxy isInMemory: addr]]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrInappropriate].
	self free: addr asVoidPointer.
	ptr
		at: 0 put: 0;
		at: 1 put: 0 "cleanup"! !

!IA32ABIPlugin methodsFor: 'primitives-callouts' stamp: 'eem 12/10/2012 14:39'!
primCallOutDoubleReturn
	"Call a foreign function that answers a double-precision floating-point result in %f0
	 according to IA32-ish ABI rules. The primitive will have a signature of the form
	functionAddress <Alien> primFFICallResult: result <Alien|Object> with: firstArg <Alien | Integer> ... with: lastArg <Alien | Integer> ^<Alien>
		<primitive: 'primCallOutDoubleReturn' error: errorCode module: 'IA32ABI'>.
	Answer result. If result is an Alien the value answered by the call will be assigned to result."
	| errCode mac result |
	<export: true>
	mac := interpreterProxy methodArgumentCount.
	self cppIf: STACKVM
		ifTrue: "In the STACKVM stacks grow down"
			[errCode := self call: interpreterProxy getStackPointer + mac - 2 "ptr to 0th arg"
							IA32: 1 - mac	"nargs negated to imply stack direction"
							Double: mac	"funcOffset"
							Return: mac - 1	"resultOffset"]
		ifFalse:
			[errCode := self call: interpreterProxy getStackPointer - mac + 2 "ptr to 0th arg"
							IA32: mac - 1	"nargs"
							Double: mac	"funcOffset"
							Return: mac - 1	"resultOffset"].
	errCode ~= 0 ifTrue:
		[^interpreterProxy primitiveFailFor: errCode].
	result := interpreterProxy stackValue: mac - 1.
	interpreterProxy pop: mac + 1 thenPush: result! !

!IA32ABIPlugin methodsFor: 'primitives-callouts' stamp: 'eem 12/10/2012 14:39'!
primCallOutFloatReturn
	"Call a foreign function that answers a single-precision floating-point result in %f0
	 according to IA32-ish ABI rules. The primitive will have a signature of the form
	functionAddress <Alien> primFFICallResult: result <Alien|Object> with: firstArg <Alien | Integer> ... with: lastArg <Alien | Integer> ^<Alien>
		<primitive: 'primCallOutFloatReturn' error: errorCode module: 'IA32ABI'>.
	Answer result. If result is an Alien the value answered by the call will be assigned to result."
	| errCode mac result |
	<export: true>
	mac := interpreterProxy methodArgumentCount.
	self cppIf: STACKVM
		ifTrue: "In the STACKVM stacks grow down"
			[errCode := self call: interpreterProxy getStackPointer + mac - 2 "ptr to 0th arg"
							IA32: 1 - mac	"nargs negated to imply stack direction"
							Float: mac		"funcOffset"
							Return: mac - 1	"resultOffset"]
		ifFalse:
			[errCode := self call: interpreterProxy getStackPointer - mac + 2 "ptr to 0th arg"
							IA32: mac - 1	"nargs"
							Float: mac		"funcOffset"
							Return: mac - 1	"resultOffset"].
	errCode ~= 0 ifTrue:
		[^interpreterProxy primitiveFailFor: errCode].
	result := interpreterProxy stackValue: mac - 1.
	interpreterProxy pop: mac + 1 thenPush: result! !

!IA32ABIPlugin methodsFor: 'primitives-callouts' stamp: 'eem 12/10/2012 14:40'!
primCallOutIntegralReturn
	"Call a foreign function that answers an integral result in %eax (and possibly %edx)
	 according to IA32-ish ABI rules. The primitive will have a signature of the form
	functionAddress <Alien> primFFICallResult: result <Alien|Object> with: firstArg <Alien | Integer> ... with: lastArg <Alien | Integer> ^<Alien>
		<primitive: 'primCallOutIntegralReturn' error: errorCode module: 'IA32ABI'>.
	Answer result. If result is an Alien the value answered by the call will be assigned to result."
	| errCode mac result |
	<export: true>
	mac := interpreterProxy methodArgumentCount.
	self cppIf: STACKVM
		ifTrue: "In the STACKVM stacks grow down"
			[errCode := self call: interpreterProxy getStackPointer + mac - 2 "ptr to 0th arg"
							IA32: 1 - mac	"nargs negated to imply stack direction"
							Integral: mac	"funcOffset"
							Return: mac - 1	"resultOffset"]
		ifFalse:
			[errCode := self call: interpreterProxy getStackPointer - mac + 2 "ptr to 0th arg"
							IA32: mac - 1	"nargs"
							Integral: mac	"funcOffset"
							Return: mac - 1	"resultOffset"].
	errCode ~= 0 ifTrue:
		[^interpreterProxy primitiveFailFor: errCode].
	result := interpreterProxy stackValue: mac - 1.
	interpreterProxy pop: mac + 1 thenPush: result! !

!IA32ABIPlugin methodsFor: 'primitives-memory management' stamp: 'eem 9/27/2020 14:04'!
primCalloc
	"calloc (malloc + zero-fill) arg bytes."
	"primCalloc: byteSize <Integer> ^<Integer>
		<primitive: 'primCalloc' error: errorCode module: 'IA32ABI'>"
	| byteSize addr |
	<export: true>

	byteSize := interpreterProxy stackIntegerValue: 0.
	(interpreterProxy failed
	 or: [byteSize <= 0 "some mallocs can't deal with malloc(0) bytes"]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	addr := (self calloc: 1 _: byteSize) asUnsignedInteger.
	addr = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoCMemory].
	interpreterProxy methodReturnValue: (self positiveMachineIntegerFor: addr)! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 9/27/2020 14:05'!
primDoubleAt
	"Answer the 64-bit double starting at the given byte offset (little endian)."
	"<Alien> doubleAt: index <Integer> ^<Float>
		<primitive: 'primDoubleAt' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr addr floatValue |
	<export: true>
	<var: #floatValue type: #double>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 8 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	addr := startAddr + byteOffset.
	self memcpy: (self addressOf: floatValue put: [:v| floatValue := v]) _: addr _: (self sizeof: floatValue).
	interpreterProxy methodReturnFloat: floatValue! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 9/19/2018 17:31'!
primDoubleAtPut
	"Store a double into 64 bits starting at the given byte offset (little endian)."
	"<Alien> doubleAt: index <Integer> put: value <Float | Integer> ^<Float | Integer>
		<primitive: 'primDoubleAtPut' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr addr valueOop floatValue |
	<export: true>
	<var: #floatValue type: #double>

	valueOop := interpreterProxy stackValue: 0.
	(interpreterProxy isIntegerObject: valueOop)
		ifTrue:[floatValue := self cCoerce: (interpreterProxy integerValueOf: valueOop) to: #double]
		ifFalse:[floatValue := self cCoerce: (interpreterProxy floatValueOf: valueOop) to: #double].
	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 1) - 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 8 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	addr := startAddr + byteOffset.
	self memcpy: addr _: (self addressOf: floatValue) _: (self sizeof: floatValue).
	interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-Windows-VM-specific' stamp: 'eem 1/29/2016 14:56'!
primDrainOSEventQueue
	<export: true>
	self cppIf: #NewspeakVM
		ifTrue: [self ioDrainEventQueue]! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 9/27/2020 14:06'!
primFloatAt
	"Answer the 32-bit float starting at the given byte offset (little endian)."
	"<Alien> floatAt: index <Integer> ^<Float>
		<primitive: 'primFloatAt' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr addr floatValue |
	<export: true>
	<var: #floatValue type: #float>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 4 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	addr := startAddr + byteOffset.
	self memcpy: (self addressOf: floatValue put: [:v| floatValue := v]) _: addr _: (self sizeof: floatValue).
	interpreterProxy methodReturnFloat: floatValue! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 9/19/2018 17:31'!
primFloatAtPut
	"Store a float into 32 bits starting at the given byte offset (little endian)."
	"<Alien> floatAt: index <Integer> put: value <Float | Integer> ^<Float | Integer>
		<primitive: 'primFloatAtPut' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr addr valueOop floatValue |
	<export: true>
	<var: #floatValue type: #float>

	valueOop := interpreterProxy stackValue: 0.
	(interpreterProxy isIntegerObject: valueOop)
		ifTrue:[floatValue := self cCoerce: (interpreterProxy integerValueOf: valueOop) to: #double]
		ifFalse:[floatValue := self cCoerce: (interpreterProxy floatValueOf: valueOop) to: #double].
	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 1) - 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 4 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	addr := startAddr + byteOffset.
	self memcpy: addr _: (self addressOf: floatValue) _: (self sizeof: floatValue).
	interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-memory management' stamp: 'eem 9/27/2020 14:06'!
primFree
	"Free the memory referenced by the argument, an integer."
	"<Anywhere> primFree: address <Integer>
		<primitive: 'primFree' error: errorCode module: 'IA32ABI'>"
	| addr |
	<export: true>
	addr := interpreterProxy stackPositiveMachineIntegerValue: 0.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	"Don't you dare to free Squeak's memory!!"
	(addr = 0 or: [interpreterProxy isInMemory: addr]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrInappropriate].
	self free: addr asVoidPointer.
	interpreterProxy pop: 1! !

!IA32ABIPlugin methodsFor: 'primitives-Windows-VM-specific' stamp: 'eem 1/30/2015 11:38'!
primInIOProcessEventsFlagAddress
	"Answer the address of the int inIOProcessEvents flag.  This can be used to
	 disable invocation of ioProcessEvents and is for backward-compatibility.
	 Please use the core VM primitiveEventProcessingControl in new code."
	| inIOProcessEvents |
	<export: true>
	<var: 'inIOProcessEvents' declareC: 'extern int inIOProcessEvents'>
	self cCode: '' inSmalltalk: [inIOProcessEvents = 0].
	interpreterProxy methodReturnValue: (self positiveMachineIntegerFor: (self addressOf: inIOProcessEvents) asUnsignedInteger)! !

!IA32ABIPlugin methodsFor: 'primitives-library loading' stamp: 'eem 12/12/2014 15:38'!
primInLibraryFindSymbol
	"Attempt to find the address of a symbol in a loaded library.
	 The primitive can have a signature  either of the form:
		<Anywhere> primInLibrary: libraryHandle <Alien> findSymbol: symbolName <String> ^<Integer>
			<primitive: 'primInLibraryFindSymbol' error: errorCode module: 'IA32ABI'>
	 or:
		libraryHandle <Alien>  primFindSymbol: symbolName <String> ^<Integer>
			<primitive: 'primInLibraryFindSymbol' error: errorCode module: 'IA32ABI'>"
	<export: true>
	| functionName libraryProxy address |
	<var: #address type: #'void *'>
	functionName := interpreterProxy stackValue: 0.
	libraryProxy := interpreterProxy stackValue: 1.
	((self isAlien: libraryProxy)
	 and: [(interpreterProxy byteSizeOf: libraryProxy) >= (2 * interpreterProxy bytesPerOop)
	 and: [interpreterProxy isBytes: functionName]]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	address := interpreterProxy
					ioLoadSymbol: (self cCoerce: (interpreterProxy firstIndexableField: functionName)
										to: #sqInt)
					OfLength: (interpreterProxy byteSizeOf: functionName)
					FromModule: (self longAt: libraryProxy + interpreterProxy baseHeaderSize + interpreterProxy bytesPerOop).
	(interpreterProxy failed
	 or: [address = 0]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNotFound].
	interpreterProxy methodReturnValue: (self positiveMachineIntegerFor: address asUnsignedInteger)! !

!IA32ABIPlugin methodsFor: 'primitives-library loading' stamp: 'eem 12/12/2014 15:39'!
primLoadLibrary
	"Attempt to load a library of the given name.  The primitive will have a signature
	 of the form:
		<Anywhere>  primLoadLibrary: libraryName <String> ^<Integer>
			<primitive: 'primLoadLibrary' error: errorCode module: 'IA32ABI'>"
	<export: true>
	| libraryName libraryHandle |
	<var: #libraryHandle type: #'void *'>
	libraryName := interpreterProxy stackValue: 0.
	(interpreterProxy isBytes: libraryName)
		ifFalse: [^ interpreterProxy primitiveFailFor: PrimErrBadArgument].
	libraryHandle := interpreterProxy
					ioLoadModule: (self cCoerce: (interpreterProxy firstIndexableField: libraryName) to: 'sqInt')
					OfLength: (interpreterProxy byteSizeOf: libraryName).
	libraryHandle = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNotFound].
	interpreterProxy methodReturnValue: (self positiveMachineIntegerFor: libraryHandle asUnsignedInteger)! !

!IA32ABIPlugin methodsFor: 'primitives-memory management' stamp: 'eem 9/27/2020 14:06'!
primMalloc
	"Malloc arg bytes."
	"primMalloc: byteSize <Integer> <^Integer>
		<primitive: 'primMalloc' error: errorCode module: 'IA32ABI'>"
	| byteSize addr |
	<export: true>

	byteSize := interpreterProxy stackIntegerValue: 0.
	(interpreterProxy failed
	 or: [byteSize <= 0 "some mallocs can't deal with malloc(0) bytes"]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	addr := (self malloc: byteSize) asUnsignedInteger.
	addr = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoCMemory].
	interpreterProxy methodReturnValue: (self positiveMachineIntegerFor: addr)! !

!IA32ABIPlugin methodsFor: 'primitives-callbacks' stamp: 'eem 10/17/2020 15:24'!
primMostRecentCallbackContext
	"This is here only for debugging; it is really useful in predicting the behaviour of primReturnAsFromContextThrough."
	<export: true>
	interpreterProxy methodReturnValue: (self positiveMachineIntegerFor: self getMostRecentCallbackContext)! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/26/2020 20:16'!
primOopAt
	"Fetch an oop from 32 or 64 bits starting at the given byte offset (little endian)."
	"THIS IS A HUGE SECURITY HOLE, BUT FULL CALLBACKS DEMAND IT"
	"<Alien> oopAt: index <Integer> ^<Object>
		<primitive: 'oopAtPut' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr addr |
	<export: true>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: BytesPerOop inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	addr := self cCoerce: startAddr + byteOffset to: #'sqInt *'.
	"WE SHOULD DO VALIDATION HERE!!  FOR NOW WE PUNT, BUT THIS REALLY DOES REQUIRE A GOOD ANSWER!!!!"
	^interpreterProxy methodReturnValue: (addr at: 0)! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/26/2020 20:09'!
primOopAtPut
	"Store an oop into 32 or 64 bits starting at the given byte offset (little endian)."
	"THIS IS A HUGE SECURITY HOLE, BUT FULL CALLBACKS DEMAND IT"
	"<Alien> oopAt: index <Integer> put: value <Object> ^<Object>
		<primitive: 'oopAtPut' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr addr oop |
	<export: true>

	oop := interpreterProxy stackValue: 0.
	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 1) - 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: BytesPerOop inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	addr := self cCoerce: startAddr + byteOffset to: #'sqInt *'.
	addr at: 0 put: oop.
	^interpreterProxy methodReturnValue: oop! !

!IA32ABIPlugin methodsFor: 'primitives-callbacks' stamp: 'eem 10/17/2020 17:37'!
primReturnAsFromContextThrough
	"Return a result from a callback to the callback's callee.  The primitive
	 has a signature of either of the forms:
		result <VMCallbackContext32/64>
				primReturnAs: returnTypeCode <Integer>
				FromContext: callbackContext <Context>
		result <VMCallbackContext32/64>
				primSignal: aSemaphore <Semaphore>
				andReturnAs: returnTypeCode <Integer>
				FromContext: callbackContext <Context>
			<primitive: 'primReturnAsFromContextThrough' error: errorCode module: 'IA32ABI'>.
	 If of the second form answer false if this is not the most recent callback, and in any case
	 signal aSemaphore (Alien's lifoCallbackSemaphore), releasing all processes waiting on it,
	 so as to implement LIFO ordering of callbacks."
	<export: true>
	| argCount vmCallbackContext isMostRecentCallback lifoCallbackSemaphore |
	argCount := interpreterProxy methodArgumentCount.
	self assert: (interpreterProxy isWordsOrBytes: (interpreterProxy stackValue: argCount)). "Receiver, a VMCallbackContext32/64 Alien ..."
	vmCallbackContext := self cCoerceSimple: (self startOfData: (interpreterProxy stackValue: argCount))
								to: #'VMCallbackContext *'.
	isMostRecentCallback := vmCallbackContext = self getMostRecentCallbackContext.
	argCount = 3
		ifTrue: "More error checking is done in StackInterpreter>>#returnAs:ThroughCallback:Context: below..."
			[lifoCallbackSemaphore := interpreterProxy stackValue: 2.
			(interpreterProxy fetchClassOf: lifoCallbackSemaphore) = interpreterProxy classSemaphore ifFalse:
				[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
			 "Allow the next waiting callback to attempt to return..."
			 interpreterProxy signalNoResume: lifoCallbackSemaphore.
			 "If this is not the most recent, return false, and proceed to wait..."
			 isMostRecentCallback ifFalse:
				[^interpreterProxy methodReturnValue: interpreterProxy falseObject]]
		ifFalse:
			[self assert: interpreterProxy methodArgumentCount = 2.
			isMostRecentCallback ifFalse:
				[^interpreterProxy primitiveFailFor: PrimErrInappropriate]].
	(interpreterProxy
		returnAs: (interpreterProxy stackValue: 1)
		ThroughCallback: vmCallbackContext
		Context: (interpreterProxy stackValue: 0)) ifFalse:
			[^interpreterProxy primitiveFailFor: PrimErrBadArgument]
	"NOTREACHED"! !

!IA32ABIPlugin methodsFor: 'primitives-callbacks' stamp: 'eem 10/17/2020 16:56'!
primReturnFromContextThrough
	"Return a result from a callback to the callback's callee.  The primitive
	 has a signature of either of the forms:
		result <FFICallbackResult> primReturnFromContext: callbackContext <MethodContext> through: jmpBuf <Integer>
		result <FFICallbackResult> primSignal: aSemaphore <Semaphore> andReturnFromContext: callbackContext <MethodContext> through: jmpBuf <Integer>
			<primitive: 'primReturnFromContextThrough' error: errorCode module: 'IA32ABI'>.
	 If of the second form answer true if this is not the most recent callback, and signal aSemaphore
	 if it is, so as to implement LIFO ordering of callbacks."
	<export: true>
	<option: #'OBSOLETE_ALIEN_PRIMITIVES'>
	<legacy>
	| mac vmCallbackContext vmCallbackReturnValue isMostRecent |
	<var: #vmCallbackContext type: #'VMCallbackContext *'>
	<var: #vmCallbackReturnValue type: #'VMCallbackReturnValue *'>
	vmCallbackContext := self cCoerceSimple: (interpreterProxy positiveMachineIntegerValueOf: (interpreterProxy stackValue: 0))
								to: #'VMCallbackContext *'.
	(interpreterProxy failed or: [vmCallbackContext = 0]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	(mac := interpreterProxy methodArgumentCount) = 3 ifTrue:
		[isMostRecent := vmCallbackContext = self getMostRecentCallbackContext.
		isMostRecent ifFalse:
			[interpreterProxy methodReturnValue: interpreterProxy trueObject.
			^nil].
		(interpreterProxy fetchClassOf: (interpreterProxy stackValue: 2)) = interpreterProxy classSemaphore
			ifFalse: [^interpreterProxy primitiveFailFor: PrimErrBadArgument].
		[interpreterProxy signalNoResume: (interpreterProxy stackValue: 2)] whileFalse].
	vmCallbackReturnValue := self cCoerceSimple: (self startOfData: (interpreterProxy stackValue: mac))
									to: #'VMCallbackReturnValue *'.
	self cCode: "C needs a typedef for structs to be assigned, but that implies a struct class for just one assignment."
		[self memcpy: (self addressOf: vmCallbackContext rvs)
			_: (self addressOf: vmCallbackReturnValue crvrvs)
			_: (self sizeof: vmCallbackContext rvs)]
		inSmalltalk: [vmCallbackContext rvs: vmCallbackReturnValue crvrvs].
	(interpreterProxy
		returnAs: (interpreterProxy integerObjectOf: vmCallbackReturnValue type + 1)
		ThroughCallback: vmCallbackContext
		Context: (interpreterProxy stackValue: 1)) ifFalse:
			[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	"NOTREACHED"! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:35'!
primSignedByteAt
	"Answer the signed 8-bit integer starting at the given byte offset (little endian)."
	"<Alien> unsignedByteAt: index <Integer> ^<Integer>
		<primitive: 'primSignedByteAt' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value |
	<export: true>
	<var: #value type: #'signed char'>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 1 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	value := self cCode: [self byteAt: startAddr + byteOffset]
				inSmalltalk: [(self alienFor: rcvr at: startAddr) signedByteAt: byteOffset + 1].
	^interpreterProxy methodReturnInteger: value! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/28/2020 00:52'!
primSignedByteAtPut
	"Store a signed integer into 8 bits starting at the given byte offset (little endian)."
	"<Alien> signedByteAt: index <Integer> put: value <Integer> ^<Integer>
		<primitive: 'primSignedByteAtPut' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value valueOop |
	<export: true>

	valueOop := interpreterProxy stackValue: 0.
	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 1) - 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	value := interpreterProxy signed32BitValueOf: valueOop.
	(interpreterProxy failed
	or: [value < -128
	or: [value > 127]]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 1 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	self cCode: [self byteAt: startAddr + byteOffset put: value]
		inSmalltalk: [(self alienFor: rcvr at: startAddr) signedByteAt: byteOffset + 1 put: value].
	^interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:30'!
primSignedLongAt
	"Answer the signed 32-bit integer starting at the given byte offset (little endian)."
	"<Alien> signedLongAt: index <Integer> ^<Integer>
		<primitive: 'primSignedLongAt' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value |
	<export: true>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 4 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	value := self cCode: [self long32At: startAddr + byteOffset]
				inSmalltalk: [(self alienFor: rcvr at: startAddr) signedLongAt: byteOffset + 1].
	^interpreterProxy methodReturnValue: (interpreterProxy signed32BitIntegerFor: value)! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:32'!
primSignedLongAtPut
	"Store a signed integer into 32 bits starting at the given byte offset (little endian)."
	"<Alien> signedLongAt: index <Integer> put: value <Integer> ^<Integer>
		<primitive: 'primSignedLongAtPut' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value valueOop |
	<export: true>

	valueOop := interpreterProxy stackValue: 0.
	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 1) - 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	value := interpreterProxy signed32BitValueOf: valueOop.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 4 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	self cCode: [self long32At: startAddr + byteOffset put: value]
		inSmalltalk: [(self alienFor: rcvr at: startAddr) signedLongAt: byteOffset + 1 put: value].
	^interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:28'!
primSignedLongLongAt
	"Answer the signed 64-bit integer starting at the given byte offset (little endian)."
	"<Alien> signedLongLongAt: index <Integer> ^<Integer>
		<primitive: 'primSignedLongLongAt' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value |
	<export: true>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 8 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	value := self cCode: [(self cCoerce: startAddr + byteOffset to: #'long long *') at: 0]
				inSmalltalk: [(self alienFor: rcvr at: startAddr) signedLongLongAt: byteOffset + 1].
	^interpreterProxy methodReturnValue: (interpreterProxy signed64BitIntegerFor: value)! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:29'!
primSignedLongLongAtPut
	"Store a signed integer into 64 bits starting at the given byte offset (little endian)."
	"<Alien> signedLongLongAt: index <Integer> put: value <Integer> ^<Integer>
		<primitive: 'primSignedLongLongAtPut' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr valueOop value |
	<export: true>

	valueOop := interpreterProxy stackValue: 0.
	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 1) - 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	value := interpreterProxy signed64BitValueOf: valueOop.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 8 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	self cCode: [(self cCoerce: startAddr + byteOffset to: #'long long *') at: 0 put: value]
		inSmalltalk: [(self alienFor: rcvr at: startAddr) signedLongLongAt: byteOffset + 1 put: value].
	^interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:33'!
primSignedShortAt
	"Answer the signed 32-bit integer starting at the given byte offset (little endian)."
	"<Alien> signedShortAt: index <Integer> ^<Integer>
		<primitive: 'primSignedShortAt' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value |
	<export: true>
	<var: #value type: #short>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 2 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	value := self cCode: [self shortAt: startAddr + byteOffset]
				inSmalltalk: [(self alienFor: rcvr at: startAddr) signedShortAt: byteOffset + 1].
	^interpreterProxy methodReturnInteger: value! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:34'!
primSignedShortAtPut
	"Store a signed integer into 16 bits starting at the given byte offset (little endian)."
	"<Alien> signedShortAt: index <Integer> put: value <Integer> ^<Integer>
		<primitive: 'primSignedShortAtPut' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value valueOop |
	<export: true>

	valueOop := interpreterProxy stackValue: 0.
	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 1) - 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	value := interpreterProxy signed32BitValueOf: valueOop.
	(interpreterProxy failed
	or: [value < -32768
	or: [value > 32767]]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 2 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	self cCode: [self shortAt: startAddr + byteOffset put: value]
		inSmalltalk: [(self alienFor: rcvr at: startAddr) signedShortAt: byteOffset + 1 put: value].
	^interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'rmacnak 9/28/2015 21:58'!
primSignedWordAt
	"Answer the signed word starting at the given byte offset (little endian)."
	"<Alien> signedWordAt: index <Integer> ^<Integer>
		<primitive: 'primSignedWordAt' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr addr value valueOop |
	<export: true>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: interpreterProxy bytesPerOop inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	addr := startAddr + byteOffset.
	value := self longAt: addr.
	valueOop := interpreterProxy signedMachineIntegerValueOf: value.
	^interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'rmacnak 10/19/2015 20:59:01'!
primSignedWordAtPut
	"Store a signed integer into the word starting at the given byte offset (little endian)."
	"<Alien> signedWordAt: index <Integer> put: value <Integer> ^<Integer>
		<primitive: 'primSignedWordAtPut' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr addr value valueOop |
	<export: true>

	valueOop := interpreterProxy stackValue: 0.
	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 1) - 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	value := interpreterProxy signedMachineIntegerValueOf: valueOop.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: interpreterProxy bytesPerOop inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	addr := startAddr + byteOffset.
	self longAt: addr put: value.
	^interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 9/26/2020 18:13'!
primSizeField
	"Answer the signed 32- or 64-bit integer comprising the size field (the first 32- or 64-bit field)."
	"<Alien> primSizeField ^<Integer>
		<primitive: 'primSizeField' error: errorCode module: 'IA32ABI'>"
	| rcvr valueOop |
	<export: true>

	rcvr := interpreterProxy stackValue: 0.
	valueOop := interpreterProxy wordSize = 8
					ifTrue: [interpreterProxy signed64BitIntegerFor: (self longAt: rcvr + interpreterProxy baseHeaderSize) signedIntFromLong64]
					ifFalse: [interpreterProxy signed32BitIntegerFor:(self longAt: rcvr + interpreterProxy baseHeaderSize) signedIntFromLong].
	^interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:41'!
primSizeFieldPut
	"Store a signed integer into the size field (the first 32 bit field; little endian)."
	"<Alien> sizeFieldPut: value <Integer> ^<Integer>
		<primitive: 'primSizeFieldPut' error: errorCode module: 'IA32ABI'>"
	| rcvr value valueOop |
	<export: true>

	valueOop := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	value := interpreterProxy signedMachineIntegerValueOf: valueOop.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	interpreterProxy wordSize = 8
		ifTrue: [self longAt: rcvr + interpreterProxy baseHeaderSize put: value signedIntToLong64]
		ifFalse: [self longAt: rcvr + interpreterProxy baseHeaderSize put: value signedIntToLong].
	^interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 9/27/2020 14:09'!
primStrlenFromStartIndex
	"Answer the number of non-null bytes starting at index.  If
	 there isn't a null byte before the end of the object then the
	 result will be the number of bytes from index to the end of
	 the object, i.e. the result will be within the bounds of the object."
	"<Alien> primStrlenFrom: index <Integer> ^<Integer>
		<primitive: 'primStrlenFromStartIndex' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr index limit ptr |
	<export: true>
	<var: #ptr type: #'char *'>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 1 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	limit := self sizeField: rcvr.
	ptr := self cCoerce: ((self startOfData: rcvr withSize: limit) + byteOffset) to: #'char *'.
	limit = 0
		ifTrue: [index := self strlen: ptr]
		ifFalse:
			[limit := limit abs.
			 index := 0.
			 [index < limit and: [(ptr at: index) ~= 0]] whileTrue:
				[index := index + 1]].
	interpreterProxy methodReturnInteger: index! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/17/2020 15:39'!
primStrlenThroughPointerAtIndex
	"Answer the number of non-null bytes starting at the byte addressed by
	 the 4-byte pointer at index."
	"<Alien> strlenThroughPointerAt: index <Integer> ^<Integer>
		<primitive: 'primStrlenThroughPointerAtIndex' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr addr |
	<export: true>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: interpreterProxy bytesPerOop inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	addr := (self startOfData: rcvr) + byteOffset.
	^interpreterProxy methodReturnInteger: (self strlen: (self cCoerce: (self longAt: addr) to: #'char *'))! !

!IA32ABIPlugin methodsFor: 'primitives-callbacks' stamp: 'nice 8/11/2016 01:05'!
primThunkEntryAddress
	"Answer the address of the entry-point for thunk callbacks:
		long thunkEntry(void *thunkp, long *stackp);
	 This could be derived via loadModule: findSymbol: etc but that would
	preclude making the plugin internal."
	| address |
	<export: true>
	address := self cCode: [#thunkEntry asIntegerPtr] inSmalltalk: [0].
	interpreterProxy methodReturnValue: (self positiveMachineIntegerFor: address)! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:18'!
primUnsignedByteAt
	"Answer the unsigned 8-bit integer starting at the given byte offset (little endian)."
	"<Alien> unsignedByteAt: index <Integer> ^<Integer>
		<primitive: 'primUnsignedByteAt' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value |
	<export: true>
	<var: #value type: 'unsigned char '>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 1 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	value := self cCode: [self byteAt: startAddr + byteOffset]
				inSmalltalk: [(self alienFor: rcvr at: startAddr) unsignedByteAt: byteOffset + 1].
	^interpreterProxy methodReturnInteger: value! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:25'!
primUnsignedByteAtPut
	"Store an unsigned integer into 8 bits starting at the given byte offset (little endian)."
	"<Alien> unsignedByteAt: index <Integer> put: value <Integer> ^<Integer>
		<primitive: 'primUnsignedByteAtPut' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value valueOop |
	<export: true>

	valueOop := interpreterProxy stackValue: 0.
	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 1) - 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	value := interpreterProxy positive32BitValueOf: valueOop.
	(interpreterProxy failed
	or: [value > 255]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 1 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	self cCode: [self byteAt: startAddr + byteOffset put: value]
		inSmalltalk: [(self alienFor: rcvr at: startAddr) unsignedByteAt: byteOffset + 1 put: value].
	^interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:16'!
primUnsignedLongAt
	"Answer the unsigned 32-bit integer starting at the given byte offset (little endian)."
	"<Alien> unsignedLongAt: index <Integer> ^<Integer>
		<primitive: 'primUnsignedLongAt' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value |
	<export: true>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 4 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	value := self cCode: [self long32At: startAddr + byteOffset]
				inSmalltalk: [(self alienFor: rcvr at: startAddr) unsignedLongAt: byteOffset + 1].
	^interpreterProxy methodReturnValue: (interpreterProxy positive32BitIntegerFor: value)! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:13'!
primUnsignedLongAtPut
	"Store an unsigned integer into 32 bits starting at the given byte offset (little endian)."
	"<Alien> unsignedLongAt: index <Integer> put: value <Integer> ^<Integer>
		<primitive: 'primUnsignedLongAtPut' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value valueOop |
	<export: true>

	valueOop := interpreterProxy stackValue: 0.
	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 1) - 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	value := interpreterProxy positive32BitValueOf: valueOop.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 4 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	self cCode: [self long32At: startAddr + byteOffset put: value]
		inSmalltalk: [(self alienFor: rcvr at: startAddr) unsignedLongAt: byteOffset + 1 put: value].
	^interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:24'!
primUnsignedLongLongAt
	"Answer the unsigned 64-bit integer starting at the given byte offset (little endian)."
	"<Alien>unsignedLongLongAt: index <Integer> ^<Integer>
		<primitive: 'primUnsignedLongLongAt' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value |
	<export: true>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 8 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	value := self cCode: [(self cCoerce: startAddr + byteOffset to: #'unsigned long long *') at: 0]
				inSmalltalk: [(self alienFor: rcvr at: startAddr) unsignedLongLongAt: byteOffset + 1].
	^interpreterProxy methodReturnValue: (interpreterProxy positive64BitIntegerFor: value)! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:24'!
primUnsignedLongLongAtPut
	"Store a signed integer into 64 bits starting at the given byte offset (little endian)."
	"<Alien> unsignedLongLongAt: index <Integer> put: value <Integer> ^<Integer>
		<primitive: 'primUnSignedLongLongAtPut' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr valueOop value |
	<export: true>

	valueOop := interpreterProxy stackValue: 0.
	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 1) - 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	value := interpreterProxy positive64BitValueOf: valueOop.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 8 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	self cCode: [(self cCoerce: startAddr + byteOffset to: #'unsigned long long *') at: 0 put: value]
		inSmalltalk: [(self alienFor: rcvr at: startAddr) unsignedLongLongAt: byteOffset + 1 put: value].
	^interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:17'!
primUnsignedShortAt
	"Answer the unsigned 16-bit integer starting at the given byte offset (little endian)."
	"<Alien> unsignedShortAt: index <Integer> ^<Integer>
		<primitive: 'primUnsignedShortAt' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value |
	<export: true>
	<var: #value type: 'unsigned short'>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 2 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	value := self cCode: [self shortAt: startAddr + byteOffset]
				inSmalltalk: [(self alienFor: rcvr at: startAddr) unsignedShortAt: byteOffset + 1].
	^interpreterProxy methodReturnInteger: value! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:20'!
primUnsignedShortAtPut
	"Store an unsigned integer into 16 bits starting at the given byte offset (little endian)."
	"<Alien> unsignedShortAt: index <Integer> put: value <Integer> ^<Integer>
		<primitive: 'primUnsignedShortAtPut' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value valueOop |
	<export: true>

	valueOop := interpreterProxy stackValue: 0.
	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 1) - 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	value := interpreterProxy positive32BitValueOf: valueOop.
	(interpreterProxy failed
	or: [value > 65535]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: 2 inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	self cCode: [self shortAt: startAddr + byteOffset put: value]
		inSmalltalk: [(self alienFor: rcvr at: startAddr) unsignedShortAt: byteOffset + 1 put: value].
	^interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:22'!
primUnsignedWordAt
	"Answer the unsigned word starting at the given byte offset (little endian)."
	"<Alien> unsignedWordAt: index <Integer> ^<Integer>
		<primitive: 'primUnsignedWordAt' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value |
	<export: true>

	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 0) - 1.
	rcvr := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: interpreterProxy bytesPerOop inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	value := self cCode: [self longAt: startAddr + byteOffset]
				inSmalltalk: [(self alienFor: rcvr at: startAddr) unsignedWordAt: byteOffset + 1].
	^interpreterProxy methodReturnValue: (interpreterProxy positiveMachineIntegerFor: value)! !

!IA32ABIPlugin methodsFor: 'primitives-accessing' stamp: 'eem 10/27/2020 23:23'!
primUnsignedWordAtPut
	"Store an unsigned integer into 32 bits starting at the given byte offset (little endian)."
	"<Alien> unsignedWordAt: index <Integer> put: value <Integer> ^<Integer>
		<primitive: 'primUnsignedWordAtPut' error: errorCode module: 'IA32ABI'>"
	| byteOffset rcvr startAddr value valueOop |
	<export: true>

	valueOop := interpreterProxy stackValue: 0.
	byteOffset := (interpreterProxy stackPositiveMachineIntegerValue: 1) - 1.
	rcvr := interpreterProxy stackObjectValue: 2.
	value := interpreterProxy positiveMachineIntegerValueOf: valueOop.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self index: byteOffset length: interpreterProxy bytesPerOop inRange: rcvr) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	(interpreterProxy isOopImmutable: rcvr) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(startAddr := self startOfData: rcvr) = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadReceiver].
	self cCode: [self longAt: startAddr + byteOffset put: value]
		inSmalltalk: [(self alienFor: rcvr at: startAddr) unsignedWordAt: byteOffset + 1 put: value].
	^interpreterProxy methodReturnValue: valueOop! !

!IA32ABIPlugin methodsFor: 'primitives-callouts' stamp: 'eem 12/10/2012 14:42'!
primVarArgsCallOutDoubleReturn
	"Call a foreign function that answers a double-precision floating-point result in %f0
	 according to IA32-ish ABI rules. The primitive will have a signature of the form
	functionAddress <Alien> primFFICallResult: result <Alien|Object> withArguments: args <Array of: Alien | Integer> ^<Alien>
		<primitive: 'primVarArgsCallOutDoubleReturn' error: errorCode module: 'IA32ABI'>.
	Answer result. If result is an Alien the value answered by the call will be assigned to result."
	| array mac errCode result |
	<export: true>
	array := interpreterProxy stackValue: 0.
	(interpreterProxy isArray: array) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	mac := interpreterProxy methodArgumentCount.
	errCode := self call: (interpreterProxy firstIndexableField: array)	"ptr to 0th arg"
					IA32: (interpreterProxy stSizeOf: array)				"nargs"
					Double: 2											"func stackValue"
					Return: 1.											"result stackValue"
	errCode ~= 0 ifTrue:
		[^interpreterProxy primitiveFailFor: errCode].
	result := interpreterProxy stackValue: 1.
	interpreterProxy pop: mac + 1 thenPush: result! !

!IA32ABIPlugin methodsFor: 'primitives-callouts' stamp: 'eem 12/10/2012 14:42'!
primVarArgsCallOutFloatReturn
	"Call a foreign function that answers a single-precision floating-point result in %f0
	 according to IA32-ish ABI rules. The primitive will have a signature of the form
	functionAddress <Alien> primFFICallResult: result <Alien|Object> withArguments: args <Array of: Alien | Integer> ^<Alien>
		<primitive: 'primVarArgsCallOutFloatReturn' error: errorCode module: 'IA32ABI'>.
	Answer result. If result is an Alien the value answered by the call will be assigned to result."
	| array mac errCode result |
	<export: true>
	array := interpreterProxy stackValue: 0.
	(interpreterProxy isArray: array) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	mac := interpreterProxy methodArgumentCount.
	errCode := self call: (interpreterProxy firstIndexableField: array)	"ptr to 0th arg"
					IA32: (interpreterProxy stSizeOf: array)				"nargs"
					Float: 2												"func stackValue"
					Return: 1.											"result stackValue"
	errCode ~= 0 ifTrue:
		[^interpreterProxy primitiveFailFor: errCode].
	result := interpreterProxy stackValue: 1.
	interpreterProxy pop: mac + 1 thenPush: result! !

!IA32ABIPlugin methodsFor: 'primitives-callouts' stamp: 'eem 12/10/2012 14:43'!
primVarArgsCallOutIntegralReturn
	"Call a foreign function that answers an integral result in %eax (and possibly %edx)
	 according to IA32-ish ABI rules. The primitive will have a signature of the form
	functionAddress <Alien> primFFICallResult: result <Alien|Object> withArguments: args <Array of: Alien | Integer> ^<Alien>
		<primitive: 'primVarArgsCallOutIntegralReturn' error: errorCode module: 'IA32ABI'>.
	Answer result. If result is an Alien the value answered by the call will be assigned to result."
	| array mac errCode result |
	<export: true>
	array := interpreterProxy stackValue: 0.
	(interpreterProxy isArray: array) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	mac := interpreterProxy methodArgumentCount.
	errCode := self call: (interpreterProxy firstIndexableField: array)	"ptr to 0th arg"
					IA32: (interpreterProxy stSizeOf: array)				"nargs"
					Integral: 2											"func stackValue"
					Return: 1.											"result stackValue"
	errCode ~= 0 ifTrue:
		[^interpreterProxy primitiveFailFor: errCode].
	result := interpreterProxy stackValue: 1.
	interpreterProxy pop: mac + 1 thenPush: result! !

!IA32ABIPlugin methodsFor: 'initialize' stamp: 'eem 3/14/2017 13:56'!
setInterpreter: anInterpreter 
	"Note: This is coded so that is can be run from Squeak."
	| ok |
	<export: true>
	<var: #anInterpreter type: #'struct VirtualMachine*'>
	interpreterProxy := anInterpreter.
	ok := interpreterProxy majorVersion > 1
			or: [interpreterProxy minorVersion >= 12]. "There was no version 0"
	ok ifTrue:
		[self expandDereferenceInterpreterProxyFunctionTable].
	^ok! !

!IA32ABIPlugin methodsFor: 'private-support' stamp: 'eem (const to msg rewrite) 11/5/2014 18:17'!
startOfByteData: rcvr "<byte indexable oop> ^<Integer>"
	"Answer the start of rcvr's data, given that it is not an alien."
	<inline: true>
	^rcvr + interpreterProxy baseHeaderSize! !

!IA32ABIPlugin methodsFor: 'private-support' stamp: 'eem (const to msg rewrite) 11/5/2014 18:17'!
startOfData: rcvr "<Alien oop>" withSize: sizeField "<Integer> ^<Integer>"
	"Answer the start of rcvr's data.  For direct aliens this is the address of
	 the second field.  For indirect and pointer aliens it is what the second field points to."
	<inline: true>
	^sizeField > 0
	 	ifTrue: [rcvr + interpreterProxy baseHeaderSize + interpreterProxy bytesPerOop]
		ifFalse: [self longAt: rcvr + interpreterProxy baseHeaderSize + interpreterProxy bytesPerOop]! !

IA32ABIPlugin subclass: #IA32ABIPluginSimulator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!IA32ABIPluginSimulator class methodsFor: 'translation' stamp: 'eem 10/3/2010 08:43'!
shouldBeTranslated
	^false! !

!IA32ABIPluginSimulator methodsFor: 'simulation support' stamp: 'eem 10/27/2020 22:01'!
alienFor: alienObj at: addr
	| size |
	size := interpreterProxy sizeFieldOfAlien: alienObj.
	size > 0 ifTrue:
		[self error: 'cannot easily simulate internal aliens'].
	^Alien forPointer: addr! !

!IA32ABIPluginSimulator methodsFor: 'platform support' stamp: 'eem 4/27/2020 13:15'!
allocateExecutablePage: byteSizePtrBlock
	"void *allocateExecutablePage(sqIntptr_t *byteSizePtrBlock)"
	self error: 'not yet implemented'.
	byteSizePtrBlock value: 0.
	^0! !

!IA32ABIPluginSimulator methodsFor: 'simulation support' stamp: 'eem 10/27/2020 21:25'!
cCode: thing
	"These we want to rewrite into functional simulation code..."
	self halt! !

!IA32ABIPluginSimulator methodsFor: 'simulation support' stamp: 'eem 7/19/2021 14:02'!
ifIsWithinExecutablePageMakePageWritable: anAddress
	^false! !

!IA32ABIPluginSimulator methodsFor: 'memory access' stamp: 'eem 11/15/2011 13:15'!
longAt: byteAddress
	^interpreterProxy longAt: byteAddress! !

!IA32ABIPluginSimulator methodsFor: 'memory access' stamp: 'eem 11/15/2011 13:26'!
longAt: byteAddress put: a32BitValue
	^interpreterProxy longAt: byteAddress put: a32BitValue! !

!IA32ABIPluginSimulator methodsFor: 'simulation support' stamp: 'eem 10/27/2020 21:55'!
sizeField: alienOop
	"This way correctly coerces to a signed integer..."
	^interpreterProxy sizeFieldOfAlien: alienOop! !

InterpreterPlugin subclass: #InflatePlugin
	instanceVariableNames: 'zipCollection zipReadLimit zipPosition zipState zipBitBuf zipBitPos zipSource zipSourcePos zipSourceLimit zipLitTable zipDistTable zipCollectionSize zipLitTableSize zipDistTableSize readStreamInstSize'
	classVariableNames: 'MaxBits StateNoMoreData'
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!InflatePlugin commentStamp: '<historical>' prior: 0!
This plugin implements the one crucial function for efficiently decompressing streams.!

InflatePlugin subclass: #DeflatePlugin
	instanceVariableNames: 'zipHashHead zipHashTail zipHashValue zipBlockPos zipBlockStart zipLiterals zipDistances zipLiteralFreq zipDistanceFreq zipLiteralCount zipLiteralSize zipMatchCount zipMatchLengthCodes zipDistanceCodes zipCrcTable zipExtraLengthBits zipExtraDistanceBits zipBaseLength zipBaseDistance writeStreamInstSize'
	classVariableNames: 'DeflateHashBits DeflateHashMask DeflateHashShift DeflateHashTableSize DeflateMaxDistance DeflateMaxDistanceCodes DeflateMaxLiteralCodes DeflateMaxMatch DeflateMinMatch DeflateWindowMask DeflateWindowSize'
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!DeflatePlugin commentStamp: 'tpr 5/5/2003 11:52' prior: 0!
This adds Zip deflating support.
InflatePlugin should not be translated but this subclass should since it is incorporated within that class's translation process!

!DeflatePlugin class methodsFor: 'translation' stamp: 'sma 3/3/2000 12:33'!
declareCVarsIn: cg
	super declareCVarsIn: cg. "Required since we share some instVars"
	cg var: #zipHashHead type: #'unsigned int*'.
	cg var: #zipHashTail type: #'unsigned int*'.
	cg var: #zipLiterals type: #'unsigned char*'.
	cg var: #zipDistances type: #'unsigned int*'.
	cg var: #zipLiteralFreq type: #'unsigned int*'.
	cg var: #zipDistanceFreq type: #'unsigned int*'.
	cg var: #zipMatchLengthCodes type: #'unsigned int' array: ZipWriteStream matchLengthCodes.
	cg var: #zipDistanceCodes type: #'unsigned int' array: ZipWriteStream distanceCodes.
	cg var: #zipCrcTable type: #'unsigned int' array: GZipWriteStream crcTable.
	cg var: #zipExtraLengthBits type: #'unsigned int' array: ZipWriteStream extraLengthBits.
	cg var: #zipExtraDistanceBits type: #'unsigned int' array: ZipWriteStream extraDistanceBits.
	cg var: #zipBaseLength type: #'unsigned int' array: ZipWriteStream baseLength.
	cg var: #zipBaseDistance type: #'unsigned int' array: ZipWriteStream baseDistance! !

!DeflatePlugin class methodsFor: 'class initialization' stamp: 'ar 12/29/1999 20:54'!
initialize
	"DeflatePlugin initialize"
	DeflateWindowSize := 16r8000.
	DeflateWindowMask := DeflateWindowSize - 1.
	DeflateMinMatch := 3.
	DeflateMaxMatch := 258.
	DeflateMaxDistance := DeflateWindowSize.
	DeflateHashBits := 15.
	DeflateHashTableSize := 1 << DeflateHashBits.
	DeflateHashMask := DeflateHashTableSize - 1.
	DeflateHashShift := (DeflateHashBits + DeflateMinMatch - 1) // DeflateMinMatch.
	DeflateMaxLiteralCodes := ZipWriteStream maxLiteralCodes.
	DeflateMaxDistanceCodes := ZipWriteStream maxDistanceCodes.! !

!DeflatePlugin methodsFor: 'deflating' stamp: 'ar (auto pragmas 12/08) 12/29/1999 21:59'!
compare: here with: matchPos min: minLength
	"Compare the two strings and return the length of matching characters.
	minLength is a lower bound for match lengths that will be accepted.
	Note: here and matchPos are zero based."
	| length |
	<inline: true>
	"First test if we can actually get longer than minLength"
	(zipCollection at: here+minLength) = (zipCollection at: matchPos+minLength)
		ifFalse:[^0].
	(zipCollection at: here+minLength-1) = (zipCollection at: matchPos+minLength-1)
		ifFalse:[^0].
	"Then test if we have an initial match at all"
	(zipCollection at: here) = (zipCollection at: matchPos)
		ifFalse:[^0].
	(zipCollection at: here+1) = (zipCollection at: matchPos+1)
		ifFalse:[^1].
	"Finally do the real comparison"
	length := 2.
	[length < DeflateMaxMatch and:[
		(zipCollection at: here+length) = (zipCollection at: matchPos+length)]]
			whileTrue:[length := length + 1].
	^length! !

!DeflatePlugin methodsFor: 'deflating' stamp: 'ar (auto pragmas 12/08) 12/29/1999 22:00'!
deflateBlock: lastIndex chainLength: chainLength goodMatch: goodMatch
	"Continue deflating the receiver's collection from blockPosition to lastIndex.
	Note that lastIndex must be at least MaxMatch away from the end of collection"
	| here matchResult flushNeeded hereMatch hereLength newMatch newLength hasMatch |
	<inline: false>
	zipBlockPos > lastIndex ifTrue:[^false]. "Nothing to deflate"
	zipLiteralCount >= zipLiteralSize ifTrue:[^true].
	hasMatch := false.
	here := zipBlockPos.
	[here <= lastIndex] whileTrue:[
		hasMatch ifFalse:[
			"Find the first match"
			matchResult := self findMatch: here
								lastLength: DeflateMinMatch-1
								lastMatch: here
								chainLength: chainLength
								goodMatch: goodMatch.
			self insertStringAt: here. "update hash table"
			hereMatch := matchResult bitAnd: 16rFFFF.
			hereLength := matchResult bitShift: -16].

		"Look ahead if there is a better match at the next position"
		matchResult := self findMatch: here+1
							lastLength: hereLength
							lastMatch: hereMatch
							chainLength: chainLength
							goodMatch: goodMatch.
		newMatch := matchResult bitAnd: 16rFFFF.
		newLength := matchResult bitShift: -16.

		"Now check if the next match is better than the current one.
		If not, output the current match (provided that the current match
		is at least MinMatch long)"
		(hereLength >= newLength and:[hereLength >= DeflateMinMatch]) ifTrue:[
			"Encode the current match"
			flushNeeded := self
				encodeMatch: hereLength
				distance: here - hereMatch.
			"Insert all strings up to the end of the current match.
			Note: The first string has already been inserted."
			1 to: hereLength-1 do:[:i| self insertStringAt: (here := here + 1)].
			hasMatch := false.
			here := here + 1.
		] ifFalse:[
			"Either the next match is better than the current one or we didn't
			have a good match after all (e.g., current match length < MinMatch).
			Output a single literal."
			flushNeeded := self encodeLiteral: (zipCollection at: here).
			here := here + 1.
			(here <= lastIndex and:[flushNeeded not]) ifTrue:[
				"Cache the results for the next round"
				self insertStringAt: here.
				hasMatch := true.
				hereMatch := newMatch.
				hereLength := newLength].
		].
		flushNeeded ifTrue:[zipBlockPos := here. ^true].
	].
	zipBlockPos := here.
	^false! !

!DeflatePlugin methodsFor: 'primitive support' stamp: 'eem 7/18/2014 13:56'!
determineSizeOfWriteStream: rcvr
	"Determine the inst size of the class above DeflateStream or
	 ZipEncoder by looking for the first class whose inst size is less than 7."
	| class |
	class := interpreterProxy fetchClassOf: rcvr.
	[class ~= interpreterProxy nilObject
	 and: [(interpreterProxy instanceSizeOf: class) >= 7]] whileTrue:
		[class := interpreterProxy superclassOf: class].
	class = interpreterProxy nilObject ifTrue:
		[^false].
	writeStreamInstSize := interpreterProxy instanceSizeOf: class.
	^true
! !

!DeflatePlugin methodsFor: 'encoding' stamp: 'ar (auto pragmas 12/08) 12/29/1999 20:37'!
encodeLiteral: lit
	"Encode the given literal"
	<inline: true>
	zipLiterals at: zipLiteralCount put: lit.
	zipDistances at: zipLiteralCount put: 0.
	zipLiteralFreq at: lit put: (zipLiteralFreq at: lit) + 1.
	zipLiteralCount := zipLiteralCount + 1.
	^zipLiteralCount = zipLiteralSize "We *must* flush"
		or:[(zipLiteralCount bitAnd: 16rFFF) = 0 "Only check every N kbytes"
			and:[self shouldFlush]]! !

!DeflatePlugin methodsFor: 'encoding' stamp: 'ar (auto pragmas 12/08) 12/29/1999 20:37'!
encodeMatch: length distance: dist
	"Encode the given match of length length starting at dist bytes ahead"
	| literal distance |
	<inline: true>
	zipLiterals at: zipLiteralCount put: length - DeflateMinMatch.
	zipDistances at: zipLiteralCount put: dist.
	literal := (zipMatchLengthCodes at: length - DeflateMinMatch).
	zipLiteralFreq at: literal put: (zipLiteralFreq at: literal) + 1.
	dist < 257
		ifTrue:[distance := zipDistanceCodes at: dist - 1]
		ifFalse:[distance := zipDistanceCodes at: 256 + (dist - 1 bitShift: -7)].
	zipDistanceFreq at: distance put: (zipDistanceFreq at: distance) + 1.
	zipLiteralCount := zipLiteralCount + 1.
	zipMatchCount := zipMatchCount + 1.
	^zipLiteralCount = zipLiteralSize "We *must* flush"
		or:[(zipLiteralCount bitAnd: 16rFFF) = 0 "Only check every N kbytes"
			and:[self shouldFlush]]! !

!DeflatePlugin methodsFor: 'deflating' stamp: 'ar (auto pragmas 12/08) 12/29/1999 22:00'!
findMatch: here lastLength: lastLength lastMatch: lastMatch chainLength: maxChainLength goodMatch: goodMatch
	"Find the longest match for the string starting at here.
	If there is no match longer than lastLength return lastMatch/lastLength.
	Traverse at most maxChainLength entries in the hash table.
	Stop if a match of at least goodMatch size has been found."
	| matchResult matchPos distance chainLength limit bestLength length |
	<inline: false>
	"Compute the default match result"
	matchResult := (lastLength bitShift: 16) bitOr: lastMatch.

	"There is no way to find a better match than MaxMatch"
	lastLength >= DeflateMaxMatch ifTrue:[^matchResult].

	"Start position for searches"
	matchPos := zipHashHead at: (self updateHashAt: here + DeflateMinMatch - 1).

	"Compute the distance to the (possible) match"
	distance := here - matchPos.

	"Note: It is required that 0 < distance < MaxDistance"
	(distance > 0 and:[distance < DeflateMaxDistance]) ifFalse:[^matchResult].

	chainLength := maxChainLength.	"Max. nr of match chain to search"
	here > DeflateMaxDistance	"Limit for matches that are too old"
		ifTrue:[limit := here - DeflateMaxDistance]
		ifFalse:[limit := 0].

	"Best match length so far (current match must be larger to take effect)"
	bestLength := lastLength.

	[true] whileTrue:[
		"Compare the current string with the string at match position"
		length := self compare: here with: matchPos min: bestLength.
		"Truncate accidental matches beyound stream position"
		(here + length > zipPosition) ifTrue:[length := zipPosition - here].
		"Ignore very small matches if they are too far away"
		(length = DeflateMinMatch and:[(here - matchPos) > (DeflateMaxDistance // 4)])
			ifTrue:[length := DeflateMinMatch - 1].
		length > bestLength ifTrue:["We have a new (better) match than before"
			"Compute the new match result"
			matchResult := (length bitShift: 16) bitOr: matchPos.
			bestLength := length.
			"There is no way to find a better match than MaxMatch"
			bestLength >= DeflateMaxMatch ifTrue:[^matchResult].
			"But we may have a good, fast match"
			bestLength > goodMatch ifTrue:[^matchResult].
		].
		(chainLength := chainLength - 1) > 0 ifFalse:[^matchResult].
		"Compare with previous entry in hash chain"
		matchPos := zipHashTail at: (matchPos bitAnd: DeflateWindowMask).
		matchPos <= limit ifTrue:[^matchResult]. "Match position is too old"
	].! !

!DeflatePlugin methodsFor: 'initialize-release' stamp: 'eem 2/3/2015 18:34'!
initialize
	writeStreamInstSize := 0.
	super initialize! !

!DeflatePlugin methodsFor: 'deflating' stamp: 'ar (auto pragmas 12/08) 12/29/1999 21:59'!
insertStringAt: here
	"Insert the string at the given start position into the hash table.
	Note: The hash value is updated starting at MinMatch-1 since
	all strings before have already been inserted into the hash table
	(and the hash value is updated as well)."
	| prevEntry |
	<inline: true>
	zipHashValue := self updateHashAt: (here + DeflateMinMatch - 1).
	prevEntry := zipHashHead at: zipHashValue.
	zipHashHead at: zipHashValue put: here.
	zipHashTail at: (here bitAnd: DeflateWindowMask) put: prevEntry.! !

!DeflatePlugin methodsFor: 'primitive support' stamp: 'eem 7/18/2014 14:15'!
loadDeflateStreamFrom: rcvr
	| oop |
	<inline: false>
	((interpreterProxy isPointers: rcvr)
	 and: [(interpreterProxy slotSizeOf: rcvr) >= 15]) ifFalse:
		[^false].
	oop := interpreterProxy fetchPointer: 0 ofObject: rcvr.
	(interpreterProxy isBytes: oop) ifFalse:
		[^false].
	writeStreamInstSize = 0 ifTrue:
		[(self determineSizeOfWriteStream: rcvr) ifFalse:
			[^false].
		 "If the receiver wasn't valid then we derived writeStreamInstSize from an invalid source.  discard it."
		 (interpreterProxy slotSizeOf: rcvr) < (writeStreamInstSize + 5) ifTrue:
			[writeStreamInstSize := 0.
			 ^false]].
	zipCollection := interpreterProxy firstIndexableField: oop.
	zipCollectionSize := interpreterProxy byteSizeOf: oop.

	zipPosition := interpreterProxy fetchInteger: 1 ofObject: rcvr.
	zipReadLimit := interpreterProxy fetchInteger: 2 ofObject: rcvr.
	"zipWriteLimit := interpreterProxy fetchInteger: 3 ofObject: rcvr."

	oop := interpreterProxy fetchPointer: writeStreamInstSize + 0 ofObject: rcvr.
	((interpreterProxy isWords: oop)
	 and: [(interpreterProxy slotSizeOf: oop) = DeflateHashTableSize]) ifFalse:
		[^false].
	zipHashHead := interpreterProxy firstIndexableField: oop.
	oop := interpreterProxy fetchPointer: writeStreamInstSize + 1 ofObject: rcvr.
	((interpreterProxy isWords: oop)
	 and: [(interpreterProxy slotSizeOf: oop) = DeflateWindowSize]) ifFalse:
		[^false].
	zipHashTail := interpreterProxy firstIndexableField: oop.
	zipHashValue := interpreterProxy fetchInteger: writeStreamInstSize + 2 ofObject: rcvr.
	zipBlockPos := interpreterProxy fetchInteger: writeStreamInstSize + 3 ofObject: rcvr.
	"zipBlockStart := interpreterProxy fetchInteger: writeStreamInstSize + 4 ofObject: rcvr."
	oop := interpreterProxy fetchPointer: writeStreamInstSize + 5 ofObject: rcvr.
	(interpreterProxy isBytes: oop) ifFalse:
		[^false].
	zipLiteralSize := interpreterProxy slotSizeOf: oop.
	zipLiterals := interpreterProxy firstIndexableField: oop.

	oop := interpreterProxy fetchPointer: writeStreamInstSize + 6 ofObject: rcvr.
	((interpreterProxy isWords: oop)
	 and: [(interpreterProxy slotSizeOf: oop) >= zipLiteralSize]) ifFalse:
		[^false].
	zipDistances := interpreterProxy firstIndexableField: oop.

	oop := interpreterProxy fetchPointer: writeStreamInstSize + 7 ofObject: rcvr.
	((interpreterProxy isWords: oop)
	 and: [(interpreterProxy slotSizeOf: oop) = DeflateMaxLiteralCodes]) ifFalse:
		[^false].
	zipLiteralFreq := interpreterProxy firstIndexableField: oop.

	oop := interpreterProxy fetchPointer: writeStreamInstSize + 8 ofObject: rcvr.
	((interpreterProxy isWords: oop)
	 and: [(interpreterProxy slotSizeOf: oop) = DeflateMaxDistanceCodes]) ifFalse:
		[^false].
	zipDistanceFreq := interpreterProxy firstIndexableField: oop.

	zipLiteralCount := interpreterProxy fetchInteger: writeStreamInstSize + 9 ofObject: rcvr.
	zipMatchCount := interpreterProxy fetchInteger: writeStreamInstSize + 10 ofObject: rcvr.

	^interpreterProxy failed not! !

!DeflatePlugin methodsFor: 'primitive support' stamp: 'eem 7/18/2014 17:04'!
loadZipEncoderFrom: rcvr
	| oop |
	<inline: false>
	writeStreamInstSize = 0 ifTrue:
		[(self determineSizeOfWriteStream: rcvr) ifFalse:
			[^false].
		 "If the receiver wasn't valid then we derived writeStreamInstSize from an invalid source.  discard it."
		 (interpreterProxy slotSizeOf: rcvr) < (writeStreamInstSize + 3) ifTrue:
			[writeStreamInstSize := 0.
			 ^false]].
	((interpreterProxy isPointers: rcvr)
	 and: [(interpreterProxy slotSizeOf: rcvr) >= (writeStreamInstSize + 3)]) ifFalse:
		[^false].
	oop := interpreterProxy fetchPointer: 0 ofObject: rcvr.
	(interpreterProxy isBytes: oop) ifFalse:
		[^interpreterProxy primitiveFail].
	zipCollection := interpreterProxy firstIndexableField: oop.
	zipCollectionSize := interpreterProxy byteSizeOf: oop.

	zipPosition := interpreterProxy fetchInteger: 1 ofObject: rcvr.
	zipReadLimit := interpreterProxy fetchInteger: 2 ofObject: rcvr.
	"zipWriteLimit := interpreterProxy fetchInteger: 3 ofObject: rcvr."
	zipBitBuf := interpreterProxy fetchInteger: writeStreamInstSize + 1 ofObject: rcvr.
	zipBitPos := interpreterProxy fetchInteger: writeStreamInstSize + 2 ofObject: rcvr.

	^interpreterProxy failed not! !

!DeflatePlugin methodsFor: 'encoding' stamp: 'nice 12/27/2012 21:03'!
nextZipBits: nBits put: value
	"Require:
		zipCollection, zipCollectionSize, zipPosition,
		zipBitBuf, zipBitPos.
	"
	<inline: true>
	(value >= 0 and:[(1 << nBits) > value])
		ifFalse:[^interpreterProxy primitiveFail].
	zipBitBuf := zipBitBuf bitOr: value << zipBitPos.
	zipBitPos := zipBitPos + nBits.
	[zipBitPos >= 8 and:[zipPosition < zipCollectionSize]] whileTrue:[
		zipCollection at: zipPosition put: (zipBitBuf bitAnd: 255).
		zipPosition := zipPosition + 1.
		zipBitBuf := zipBitBuf >> 8.
		zipBitPos := zipBitPos - 8].
! !

!DeflatePlugin methodsFor: 'primitives' stamp: 'eem 7/18/2014 14:28'!
primitiveDeflateBlock
	"Primitive. Deflate the current contents of the receiver."
	| goodMatch chainLength lastIndex rcvr result |
	<export: true>
	<inline: false>
	interpreterProxy methodArgumentCount = 3
		ifFalse:[^interpreterProxy primitiveFail].
	goodMatch := interpreterProxy stackIntegerValue: 0.
	chainLength := interpreterProxy stackIntegerValue: 1.
	lastIndex := interpreterProxy stackIntegerValue: 2.
	rcvr := interpreterProxy stackObjectValue: 3.
	interpreterProxy failed ifTrue:[^nil].
	self cCode:'' inSmalltalk:[
		zipMatchLengthCodes := CArrayAccessor on: ZipWriteStream matchLengthCodes.
		zipDistanceCodes := CArrayAccessor on: ZipWriteStream distanceCodes].
	(self loadDeflateStreamFrom: rcvr)
		ifFalse:[^interpreterProxy primitiveFail].
	result := self deflateBlock: lastIndex chainLength: chainLength goodMatch: goodMatch.
	interpreterProxy failed ifFalse:[
		"Store back modified values"
		interpreterProxy storeInteger: writeStreamInstSize + 2 ofObject: rcvr withValue: zipHashValue.
		interpreterProxy storeInteger: writeStreamInstSize + 3 ofObject: rcvr withValue: zipBlockPos.
		interpreterProxy storeInteger: writeStreamInstSize + 9 ofObject: rcvr withValue: zipLiteralCount.
		interpreterProxy storeInteger: writeStreamInstSize + 10 ofObject: rcvr withValue: zipMatchCount].
	interpreterProxy failed ifFalse:[
		interpreterProxy pop: 4.
		interpreterProxy pushBool: result.
	].! !

!DeflatePlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:00'!
primitiveDeflateUpdateHashTable
	"Primitive. Update the hash tables after data has been moved by delta."
	| delta table tableSize tablePtr entry |
	<export: true>
	<var: #tablePtr type:'int *'>
	interpreterProxy methodArgumentCount = 2
		ifFalse:[^interpreterProxy primitiveFail].
	delta := interpreterProxy stackIntegerValue: 0.
	table := interpreterProxy stackObjectValue: 1.
	interpreterProxy failed ifTrue:[^nil].
	(interpreterProxy isWords: table)
		ifFalse:[^interpreterProxy primitiveFail].
	tableSize := interpreterProxy slotSizeOf: table.
	tablePtr := interpreterProxy firstIndexableField: table.
	0 to: tableSize-1 do:[:i|
		entry := tablePtr at: i.
		entry >= delta
			ifTrue:[tablePtr at: i put: entry - delta]
			ifFalse:[tablePtr at: i put: 0]].
	interpreterProxy pop: 2. "Leave rcvr on stack"! !

!DeflatePlugin methodsFor: 'primitives' stamp: 'eem 7/18/2014 17:10'!
primitiveUpdateAdler32
	"Primitive. Update a 32bit CRC value."
	| collection stopIndex startIndex length bytePtr s1 adler32 s2 b |
	<export: true>
	<var: #adler32 type:'unsigned int '>
	<var: #bytePtr type:'unsigned char *'>
	interpreterProxy methodArgumentCount = 4
		ifFalse:[^interpreterProxy primitiveFail].
	collection := interpreterProxy stackObjectValue: 0.
	stopIndex := interpreterProxy stackIntegerValue: 1.
	startIndex := interpreterProxy stackIntegerValue: 2.
	adler32 := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 3).
	interpreterProxy failed ifTrue:[^0].
	((interpreterProxy isBytes: collection) and:[stopIndex >= startIndex and:[startIndex > 0]])
		ifFalse:[^interpreterProxy primitiveFail].
	length := interpreterProxy byteSizeOf: collection.
	(stopIndex <= length) ifFalse:[^interpreterProxy primitiveFail].
	bytePtr := interpreterProxy firstIndexableField: collection.
	startIndex := startIndex - 1.
	stopIndex := stopIndex - 1.
	s1 := adler32 bitAnd: 16rFFFF.
	s2 := (adler32 >> 16) bitAnd: 16rFFFF.
	startIndex to: stopIndex do:[:i|
		b := bytePtr at: i.
		s1 := (s1 + b) \\ 65521.
		s2 := (s2 + s1) \\ 65521.
	].
	adler32 := (s2 bitShift: 16) + s1.
	interpreterProxy
		pop: 5 "args + rcvr"
		thenPush: (interpreterProxy positive32BitIntegerFor: adler32)! !

!DeflatePlugin methodsFor: 'primitives' stamp: 'eem 11/2/2013 00:34'!
primitiveUpdateGZipCrc32
	"Primitive. Update a 32bit CRC value."
	| collection stopIndex startIndex crc length bytePtr |
	<export: true>
	<var: #bytePtr type: #'unsigned char *'>
	interpreterProxy methodArgumentCount = 4
		ifFalse:[^interpreterProxy primitiveFail].
	collection := interpreterProxy stackObjectValue: 0.
	stopIndex := interpreterProxy stackIntegerValue: 1.
	startIndex := interpreterProxy stackIntegerValue: 2.
	crc := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 3).
	interpreterProxy failed ifTrue: [^self].
	((interpreterProxy isBytes: collection) and:[stopIndex >= startIndex and:[startIndex > 0]])
		ifFalse:[^interpreterProxy primitiveFail].
	length := interpreterProxy byteSizeOf: collection.
	(stopIndex <= length) ifFalse:[^interpreterProxy primitiveFail].
	bytePtr := interpreterProxy firstIndexableField: collection.
	self cCode:'' inSmalltalk:[zipCrcTable := CArrayAccessor on: GZipWriteStream crcTable].
	startIndex := startIndex - 1.
	stopIndex := stopIndex - 1.
	startIndex to: stopIndex do:
		[:i|
		crc := (zipCrcTable at: ((crc bitXor: (bytePtr at: i)) bitAnd: 255)) bitXor: (crc >> 8)].
	interpreterProxy
		pop: 5 "args + rcvr"
		thenPush: (interpreterProxy positive32BitIntegerFor: crc)! !

!DeflatePlugin methodsFor: 'primitives' stamp: 'eem 7/18/2014 14:24'!
primitiveZipSendBlock
	| distTree litTree distStream litStream rcvr result |
	<export: true>
	interpreterProxy methodArgumentCount = 4 
		ifFalse:[^interpreterProxy primitiveFail].
	distTree := interpreterProxy stackObjectValue: 0.
	litTree := interpreterProxy stackObjectValue: 1.
	distStream := interpreterProxy stackObjectValue: 2.
	litStream := interpreterProxy stackObjectValue: 3.
	rcvr := interpreterProxy stackObjectValue: 4.
	interpreterProxy failed ifTrue:[^nil].
	(self loadZipEncoderFrom: rcvr)
		ifFalse:[^interpreterProxy primitiveFail].
	((interpreterProxy isPointers: distTree) and:[
		(interpreterProxy slotSizeOf: distTree) >= 2])
			ifFalse:[^interpreterProxy primitiveFail].
	((interpreterProxy isPointers: litTree) and:[
		(interpreterProxy slotSizeOf: litTree) >= 2])
			ifFalse:[^interpreterProxy primitiveFail].
	((interpreterProxy isPointers: litStream) and:[
		(interpreterProxy slotSizeOf: litStream) >= 3])
			ifFalse:[^interpreterProxy primitiveFail].
	((interpreterProxy isPointers: distStream) and:[
		(interpreterProxy slotSizeOf: distStream) >= 3])
			ifFalse:[^interpreterProxy primitiveFail].
	self cCode:'' inSmalltalk:[
		zipMatchLengthCodes := CArrayAccessor on: ZipWriteStream matchLengthCodes.
		zipDistanceCodes := CArrayAccessor on: ZipWriteStream distanceCodes.
		zipExtraLengthBits := CArrayAccessor on: ZipWriteStream extraLengthBits.
		zipExtraDistanceBits := CArrayAccessor on: ZipWriteStream extraDistanceBits.
		zipBaseLength := CArrayAccessor on: ZipWriteStream baseLength.
		zipBaseDistance := CArrayAccessor on: ZipWriteStream baseDistance].
	result := self sendBlock: litStream with: distStream with: litTree with: distTree.
	interpreterProxy failed ifFalse:[
		interpreterProxy storeInteger: 1 ofObject: rcvr withValue: zipPosition.
		interpreterProxy storeInteger: readStreamInstSize + 1 ofObject: rcvr withValue: zipBitBuf.
		interpreterProxy storeInteger: readStreamInstSize + 2 ofObject: rcvr withValue: zipBitPos.
	].
	interpreterProxy failed ifFalse:[
		interpreterProxy pop: 5. "rcvr + args"
		interpreterProxy pushInteger: result.
	].! !

!DeflatePlugin methodsFor: 'encoding' stamp: 'eem 8/30/2020 18:08'!
sendBlock: literalStream with: distanceStream with: litTree with: distTree
	"Require: 
		zipCollection, zipCollectionSize, zipPosition,
		zipBitBuf, zipBitPos.
	"
	| oop litPos litLimit litArray distArray lit dist sum llBitLengths llCodes distBitLengths distCodes code extra litBlCount distBlCount |
	<var: #dist type: #sqInt> "must be signed"
	<var: #litArray type: #'unsigned char *'>
	<var: #distArray type: #'unsigned int *'>
	<var: #llBitLengths type: #'unsigned int *'>
	<var: #llCodes type: #'unsigned int *'>
	<var: #distBitLengths type: #'unsigned int *'>
	<var: #distCodes type: #'unsigned int *'>
	oop := interpreterProxy fetchPointer: 0 ofObject: literalStream.
	litPos := interpreterProxy fetchInteger: 1 ofObject: literalStream.
	litLimit := interpreterProxy fetchInteger: 2 ofObject: literalStream.
	(litPos <= litLimit
	 and: [(interpreterProxy isBytes: oop)
	 and: [litLimit <= (interpreterProxy byteSizeOf: oop)]]) ifFalse:
		[^interpreterProxy primitiveFail].
	litArray := interpreterProxy firstIndexableField: oop.

	oop := interpreterProxy fetchPointer: 0 ofObject: distanceStream.
	((interpreterProxy isWords: oop)
	 and: [litLimit <= (interpreterProxy slotSizeOf: oop)
	 and: [(interpreterProxy fetchInteger: 1 ofObject: distanceStream) = litPos
	 and: [(interpreterProxy fetchInteger: 2 ofObject: distanceStream) = litLimit]]]) ifFalse:
		[^interpreterProxy primitiveFail].
	distArray := interpreterProxy firstIndexableField: oop.

	oop := interpreterProxy fetchPointer: 0 ofObject: litTree.
	(interpreterProxy isWords: oop) ifFalse:
		[^interpreterProxy primitiveFail].
	litBlCount := interpreterProxy slotSizeOf: oop.
	llBitLengths := interpreterProxy firstIndexableField: oop.

	oop := interpreterProxy fetchPointer: 1 ofObject: litTree.
	((interpreterProxy isWords: oop)
	 and: [litBlCount = (interpreterProxy slotSizeOf: oop)]) ifFalse:
		[^interpreterProxy primitiveFail].
	llCodes := interpreterProxy firstIndexableField: oop.

	oop := interpreterProxy fetchPointer: 0 ofObject: distTree.
	(interpreterProxy isWords: oop) ifFalse:
		[^interpreterProxy primitiveFail].
	distBlCount := interpreterProxy slotSizeOf: oop.
	distBitLengths := interpreterProxy firstIndexableField: oop.

	oop := interpreterProxy fetchPointer: 1 ofObject: distTree.
	((interpreterProxy isWords: oop)
	 and: [distBlCount = (interpreterProxy slotSizeOf: oop)]) ifFalse:
		[^interpreterProxy primitiveFail].
	distCodes := interpreterProxy firstIndexableField: oop.

	self nextZipBits: 0 put: 0. "Flush pending bits if necessary"
	sum := 0.
	[litPos < litLimit and:[zipPosition + 4 < zipCollectionSize]] whileTrue:[
		lit := litArray at: litPos.
		dist := distArray at: litPos.
		litPos := litPos + 1.
		dist = 0 ifTrue:["literal"
			sum := sum + 1.
			lit < litBlCount ifFalse:[^interpreterProxy primitiveFail].
			self nextZipBits: (llBitLengths at: lit) put: (llCodes at: lit).
		] ifFalse:["match"
			sum := sum + lit + DeflateMinMatch.
			"eem 8/30/2020 Can't happen; litArray has type unsigned char *, lit unsigned char. Leaving this in causes a C compiler warning."
			false ifTrue: [lit < 256 ifFalse:[^interpreterProxy primitiveFail]].
			code := zipMatchLengthCodes at: lit.
			code < litBlCount ifFalse:[^interpreterProxy primitiveFail].
			self nextZipBits: (llBitLengths at: code) put: (llCodes at: code).
			extra := zipExtraLengthBits at: code - 257.
			extra = 0 ifFalse:[
				lit := lit - (zipBaseLength at: code - 257).
				self nextZipBits: extra put: lit].
			dist := dist - 1.
			dist < 16r8000 ifFalse:[^interpreterProxy primitiveFail].
			dist < 256
				ifTrue:[code := zipDistanceCodes at: dist]
				ifFalse:[code := zipDistanceCodes at: 256 + (dist >> 7)].
			code < distBlCount ifFalse:[^interpreterProxy primitiveFail].
			self nextZipBits: (distBitLengths at: code) put: (distCodes at: code).
			extra := zipExtraDistanceBits at: code.
			extra = 0 ifFalse:[
				dist := dist - (zipBaseDistance at: code).
				self nextZipBits: extra put: dist].
		].
	].
	interpreterProxy failed ifTrue:[^nil].
	interpreterProxy storeInteger: 1 ofObject: literalStream withValue: litPos.
	interpreterProxy storeInteger: 1 ofObject: distanceStream withValue: litPos.
	^sum! !

!DeflatePlugin methodsFor: 'encoding' stamp: 'ar (auto pragmas 12/08) 12/29/1999 22:00'!
shouldFlush
	"Check if we should flush the current block.
	Flushing can be useful if the input characteristics change."
	| nLits |
	<inline: false>
	zipLiteralCount = zipLiteralSize ifTrue:[^true]. "We *must* flush"
	(zipLiteralCount bitAnd: 16rFFF) = 0 ifFalse:[^false]. "Only check every N kbytes"
	zipMatchCount * 10 <= zipLiteralCount ifTrue:[
		"This is basically random data. 
		There is no need to flush early since the overhead
		for encoding the trees will add to the overall size"
		^false].
	"Try to adapt to the input data.
	We flush if the ratio between matches and literals
	changes beyound a certain threshold"
	nLits := zipLiteralCount - zipMatchCount.
	nLits <= zipMatchCount ifTrue:[^false]. "whow!! so many matches"
	^nLits * 4 <= zipMatchCount! !

!DeflatePlugin methodsFor: 'deflating' stamp: 'nice 12/27/2012 02:58'!
updateHash: nextValue
	"Update the running hash value based on the next input byte.
	Return the new updated hash value."
	^((zipHashValue << DeflateHashShift) bitXor: nextValue) bitAnd: DeflateHashMask.! !

!DeflatePlugin methodsFor: 'deflating' stamp: 'ar 12/29/1999 20:29'!
updateHashAt: here
	"Update the hash value at position here (one based)"
	^self updateHash: (zipCollection at: here)! !

!InflatePlugin class methodsFor: 'translation' stamp: 'sma 3/3/2000 12:31'!
declareCVarsIn: cg
	cg var: #zipCollection type: #'unsigned char*'.
	cg var: #zipSource type: #'unsigned char*'.
	cg var: #zipLitTable type: #'unsigned int*'.
	cg var: #zipDistTable type: #'unsigned int*'! !

!InflatePlugin class methodsFor: 'class initialization' stamp: 'ar 12/21/1999 23:02'!
initialize
	"InflatePlugin initialize"
	MaxBits := 16.
	StateNoMoreData := 1.! !

!InflatePlugin class methodsFor: 'translation' stamp: 'ar 5/11/2000 23:58'!
moduleName
	^'ZipPlugin'! !

!InflatePlugin class methodsFor: 'translation' stamp: 'tpr 5/14/2001 12:27'!
shouldBeTranslated
"InflatePlugin should not be translated but its subclass should since it is incorporated within that class's translation process"
	^self ~= InflatePlugin! !

!InflatePlugin class methodsFor: 'simulation' stamp: 'eem 9/21/2013 07:10'!
simulatorClass
	"For running from Smalltalk - answer a class that can be used to simulate the receiver,
	 or nil if you want the primitives in this module to always fail, causing simulation to fall
	 through to the Smalltalk code.  By default every non-TestInterpreterPlugin can simulate itself."

	^DeflatePlugin! !

!InflatePlugin methodsFor: 'primitive support' stamp: 'eem 7/18/2014 17:05'!
determineSizeOfReadStream: rcvr
	"Determine the inst size of the class above DeflateStream by
	 looking for the first class whose inst size is less than 13."
	| class |
	class := interpreterProxy fetchClassOf: rcvr.
	[class ~= interpreterProxy nilObject
	 and: [(interpreterProxy instanceSizeOf: class) >= 13]] whileTrue:
		[class := interpreterProxy superclassOf: class].
	class = interpreterProxy nilObject ifTrue:
		[^false].
	readStreamInstSize := interpreterProxy instanceSizeOf: class.
	^true! !

!InflatePlugin methodsFor: 'initialize-release' stamp: 'eem 7/18/2014 16:52'!
initialize
	readStreamInstSize := 0! !

!InflatePlugin methodsFor: 'primitives' stamp: 'eem 7/18/2014 17:06'!
primitiveInflateDecompressBlock
	"Primitive. Inflate a single block."
	| oop rcvr |
	<export: true>
	interpreterProxy methodArgumentCount = 2 ifFalse:
		[^interpreterProxy primitiveFail].
	"distance table"
	oop := interpreterProxy stackValue: 0.
	(interpreterProxy isWords: oop) ifFalse:
		[^interpreterProxy primitiveFail].
	zipDistTable := interpreterProxy firstIndexableField: oop.
	zipDistTableSize := interpreterProxy slotSizeOf: oop.

	"literal table"
	oop := interpreterProxy stackValue: 1.
	(interpreterProxy isWords: oop) ifFalse:
		[^interpreterProxy primitiveFail].
	zipLitTable := interpreterProxy firstIndexableField: oop.
	zipLitTableSize := interpreterProxy slotSizeOf: oop.


	"Receiver (InflateStream)"
	rcvr := interpreterProxy stackValue: 2.
	(interpreterProxy isPointers: rcvr) ifFalse:
		[^interpreterProxy primitiveFail].
	"All the integer instvars"
	readStreamInstSize = 0 ifTrue:
		[(self determineSizeOfReadStream: rcvr) ifFalse:
			[^interpreterProxy primitiveFail].
		 "If the receiver wasn't valid then we derived readStreamInstSize from an invalid source.  discard it."
		 (interpreterProxy slotSizeOf: rcvr) < (readStreamInstSize + 8) ifTrue:
			[readStreamInstSize := 0.
			 ^interpreterProxy primitiveFail]].
	(interpreterProxy slotSizeOf: rcvr) < (readStreamInstSize + 8) ifTrue:
		[^interpreterProxy primitiveFail].

	zipReadLimit := interpreterProxy fetchInteger: 2 ofObject: rcvr.
	zipState := interpreterProxy fetchInteger: readStreamInstSize + 0 ofObject: rcvr.
	zipBitBuf := interpreterProxy fetchInteger: readStreamInstSize + 1 ofObject: rcvr.
	zipBitPos := interpreterProxy fetchInteger: readStreamInstSize + 2 ofObject: rcvr.
	zipSourcePos := interpreterProxy fetchInteger: readStreamInstSize + 4 ofObject: rcvr.
	zipSourceLimit := interpreterProxy fetchInteger: readStreamInstSize + 5 ofObject: rcvr.
	interpreterProxy failed ifTrue:[^nil].
	zipReadLimit := zipReadLimit - 1.
	zipSourcePos := zipSourcePos - 1.
	zipSourceLimit := zipSourceLimit - 1.

	"collection"
	oop := interpreterProxy fetchPointer: 0 ofObject: rcvr.
	(interpreterProxy isBytes: oop) ifFalse:
		[^interpreterProxy primitiveFail].
	zipCollection := interpreterProxy firstIndexableField: oop.
	zipCollectionSize := interpreterProxy byteSizeOf: oop.

	"source"
	oop := interpreterProxy fetchPointer: readStreamInstSize + 3 ofObject: rcvr.
	(interpreterProxy isBytes: oop) ifFalse:
		[^interpreterProxy primitiveFail].
	zipSource := interpreterProxy firstIndexableField: oop.

	"do the primitive"
	self zipDecompressBlock.
	interpreterProxy failed ifFalse: "store modified values back"
		[interpreterProxy storeInteger: 2 ofObject: rcvr withValue: zipReadLimit + 1.
		interpreterProxy storeInteger: readStreamInstSize + 0 ofObject: rcvr withValue: zipState.
		interpreterProxy storeInteger: readStreamInstSize + 1 ofObject: rcvr withValue: zipBitBuf.
		interpreterProxy storeInteger: readStreamInstSize + 2 ofObject: rcvr withValue: zipBitPos.
		interpreterProxy storeInteger: readStreamInstSize + 4 ofObject: rcvr withValue: zipSourcePos + 1.
		interpreterProxy pop: 2]! !

!InflatePlugin methodsFor: 'inflating' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:22'!
zipDecodeValueFrom: table size: tableSize
	"Decode the next value in the receiver using the given huffman table."
	| bits bitsNeeded tableIndex value index |
	<var: #table type:'unsigned int *'>
	bitsNeeded := (table at: 0) bitShift: -24.	"Initial bits needed"
	bitsNeeded > MaxBits ifTrue:[interpreterProxy primitiveFail. ^0].
	tableIndex := 2.							"First real table"
	[true] whileTrue:[
		bits := self zipNextBits: bitsNeeded.		"Get bits"
		index := tableIndex + bits - 1.
		index >= tableSize ifTrue:[interpreterProxy primitiveFail. ^0].
		value := table at: index.					"Lookup entry in table"
		(value bitAnd: 16r3F000000) = 0 ifTrue:[^value]. "Check if it is a leaf node"
		"Fetch sub table"
		tableIndex := value bitAnd: 16rFFFF.	"Table offset in low 16 bit"
		bitsNeeded := (value bitShift: -24) bitAnd: 255. "Additional bits in high 8 bit"
		bitsNeeded > MaxBits ifTrue:[interpreterProxy primitiveFail. ^0]].
	^0! !

!InflatePlugin methodsFor: 'inflating' stamp: 'ar (auto pragmas 12/08) 12/22/1999 00:04'!
zipDecompressBlock
	| value extra length distance oldPos oldBits oldBitPos dstPos srcPos max |
	<inline: false>
	max := zipCollectionSize - 1.
	[zipReadLimit < max and:[zipSourcePos <= zipSourceLimit]] whileTrue:[
		"Back up stuff if we're running out of space"
		oldBits := zipBitBuf.
		oldBitPos := zipBitPos.
		oldPos := zipSourcePos.
		value := self zipDecodeValueFrom: zipLitTable size: zipLitTableSize.
		value < 256 ifTrue:[ "A literal"
			zipCollection at: (zipReadLimit := zipReadLimit + 1) put: value.
		] ifFalse:["length/distance or end of block"
			value = 256 ifTrue:["End of block"
				zipState := zipState bitAnd: StateNoMoreData.
				^0].
			"Compute the actual length value (including possible extra bits)"
			extra := (value bitShift: -16) - 1.
			length := value bitAnd: 16rFFFF.
			extra > 0 ifTrue:[length := length + (self zipNextBits: extra)].
			"Compute the distance value"
			value := self zipDecodeValueFrom: zipDistTable size: zipDistTableSize.
			extra := (value bitShift: -16).
			distance := value bitAnd: 16rFFFF.
			extra > 0 ifTrue:[distance := distance + (self zipNextBits: extra)].
			(zipReadLimit + length >= max) ifTrue:[
				zipBitBuf := oldBits.
				zipBitPos := oldBitPos.
				zipSourcePos := oldPos.
				^0].
			dstPos := zipReadLimit.
			srcPos := zipReadLimit - distance.
			1 to: length do:[:i|
				zipCollection at: dstPos+i put: (zipCollection at: srcPos+i)].
			zipReadLimit := zipReadLimit + length.
		].
	].! !

!InflatePlugin methodsFor: 'inflating' stamp: 'ar (auto pragmas 12/08) 12/21/1999 23:06'!
zipNextBits: n
	| bits byte |
	<inline: true>
	[zipBitPos < n] whileTrue:[
		byte := zipSource at: (zipSourcePos := zipSourcePos + 1).
		zipBitBuf := zipBitBuf + (byte << zipBitPos).
		zipBitPos := zipBitPos + 8].
	bits := zipBitBuf bitAnd: (1 << n)-1.
	zipBitBuf := zipBitBuf >> n.
	zipBitPos := zipBitPos - n.
	^bits! !

!InterpreterPlugin class methodsFor: 'accessing' stamp: 'tpr 12/17/2003 16:52'!
allCallsOn
	"Answer a SortedCollection of all the methods that refer to me. Most classes simply defer to SystemDictionary>allCallsOn: but some have special requirements - plugins may have a module name that does not match the class name"

	self theNonMetaClass name ~= self moduleName asSymbol
		ifTrue:[^super allCallsOn, (self systemNavigation allCallsOn: self moduleName asSymbol)]
		ifFalse:[^super allCallsOn]! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'tpr 12/3/2015 10:26'!
allCodeOlderThan: modificationTime
	^((self pluginClassesUpToRoot) allSatisfy:
			[:aPluginClass| aPluginClass timeStamp < modificationTime])
	  and: [self translatedPrimitives allSatisfy:
			[:pair| | c m stamp |
			c := Smalltalk classNamed: pair first.
			m := c compiledMethodAt: pair last ifAbsent: [c class >> pair last].
			stamp := (m timeStamp subStrings: {Character space}) last: 2.
			stamp := TimeStamp date: (Date fromString: stamp first) time: (Time fromString: stamp last).
			stamp asSeconds < modificationTime]]! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'ar 5/12/2000 00:24'!
baseDirectoryName
	"Return the directory into which plugins should be generated by default."
	^FileDirectory default pathName! !

!InterpreterPlugin class methodsFor: 'translated primitives' stamp: 'eem 3/6/2015 14:47'!
browseTranslatedPrimitives
	"InterpreterPlugin browseTranslatedPrimitives"
	| methodRefs |
	methodRefs := OrderedCollection new.
	self withAllSubclasses do:
		[:class|
		methodRefs addAll:
			(class translatedPrimitives collect:
				[:tuple|
				(self methodForTranslatedPrimitiveTuple: tuple) methodReference])].
	self systemNavigation browseMessageList: methodRefs
		name: 'Translated primitives'! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'eem 7/27/2021 08:33'!
buildCodeGenerator
	"Build a CCodeGenerator for the plugin"
	^self buildCodeGeneratorInto: nil! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'eem 7/27/2021 08:33'!
buildCodeGeneratorInto: aBlockOrNil
	"Build a CCodeGenerator for the plugin"
	| cg pluginClasses |
	cg := self codeGeneratorClass new initialize.
	cg pluginClass: self.
	aBlockOrNil ifNotNil: [aBlockOrNil value: cg].
	(pluginClasses := self pluginClassesUpToRoot) do:
		[:aClass| cg addClass: aClass].
	(cg structClassesForTranslationClasses: pluginClasses) do:
		[:structClasss| cg addStructClass: structClasss].
	cg addMethodsForTranslatedPrimitives: self translatedPrimitives.
	^cg! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'tpr 12/3/2015 10:30'!
buildCodeGeneratorUpTo: aPluginClass
	"Build a CCodeGenerator for the plugin - Deprecated and here only in case old plugin code tries to use it"
	self deprecated.
	^self buildCodeGenerator! !

!InterpreterPlugin class methodsFor: 'private' stamp: 'tpr 6/9/2003 16:41'!
codeGeneratorClass
	"return the appropriate class of code generator for this kind ofplugin"

	^VMPluginCodeGenerator! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'eem 6/7/2016 20:27'!
declareCVarsIn: aCCodeGenerator
	"Note: This method must be implemented by all subclasses to declare variables."

	aCCodeGenerator 
		var: #interpreterProxy type: #'struct VirtualMachine*';
		removeVariable: 'translatedMethodCache' ifAbsent: nil.
	self declareHeaderFilesIn: aCCodeGenerator! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'tpr 5/23/2001 17:14'!
declareHeaderFilesIn: aCCodeGenerator
	self hasHeaderFile ifTrue:[
		aCCodeGenerator addHeaderFile: '"', self moduleName,'.h"'].! !

!InterpreterPlugin class methodsFor: 'instance creation' stamp: 'ar 12/31/2001 01:36'!
doPrimitive: primitiveName 
	| proxy plugin |
	proxy := InterpreterProxy new.
	proxy loadStackFrom: thisContext sender.
	plugin := self simulatorClass new.
	plugin setInterpreter: proxy.
	(plugin respondsTo: #initialiseModule) ifTrue:[plugin initialiseModule].
	plugin perform: primitiveName asSymbol.
	^ proxy stackValue: 0! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'eem 12/14/2012 11:32'!
exportBuildInfoOrNil
	"A hook for classes to emit some descriptive build string.
	 See CCodeGenerator>>fileHeaderVersionStampForSourceClass:"
	^nil! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'tpr 5/23/2001 17:03'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^false! !

!InterpreterPlugin class methodsFor: 'initialization' stamp: 'eem 1/28/2019 15:01'!
initialize
	DisownVMForFFICall ifNil:
		[StackInterpreter initializeMiscConstants]! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'eem 9/18/2020 11:01'!
internalModuleName
	"Answer the name to include in receiver's internal plugin exports.
	 This is the value of the module: argument in named primitives.
	 By default answer the moduleName."

	^self moduleName! !

!InterpreterPlugin class methodsFor: 'accessing' stamp: 'RMF 3/27/2000 09:39'!
isCPP
	^ false! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'eem 12/14/2012 11:29'!
isPluginClass
	^true! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'eem 9/1/2017 14:58'!
isStructType: typeName "<String>"
	"Subclasses should override to answer trye for struct types they use."
	^false! !

!InterpreterPlugin class methodsFor: 'translated primitives' stamp: 'eem 1/20/2017 17:41'!
methodForTranslatedPrimitiveTuple: tuple
	| class |
	class := Smalltalk classNamed: tuple first.
	^class
		compiledMethodAt: tuple last
		ifAbsent:
			[class class
				compiledMethodAt: tuple last
				ifAbsent: [tuple = #(String findSubstringViaPrimitive:in:startingAt:matchTable:) ifTrue:
							[ByteString compiledMethodAt: #findSubstring:in:startingAt:matchTable:]]]! !

!InterpreterPlugin class methodsFor: 'translated primitives' stamp: 'eem 3/6/2015 14:48'!
methodOrNilForTranslatedPrimitiveSelector: selector
	self translatedPrimitives do:
		[:tuple| | method |
		method := self methodForTranslatedPrimitiveTuple: tuple.
		method pragmas do:
			[:p|
			((p keyword beginsWith: 'primitive:') and: [p arguments first = selector]) ifTrue:
				[^method]]].
	^nil! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'eem 2/16/2015 18:50'!
methodsToBePruned
	"Since sharing between plugins means that normal pruning
	 can't be done, allow plugins that want to prune specific methods."
	^#()! !

!InterpreterPlugin class methodsFor: 'accessing' stamp: 'sma 4/22/2000 12:32'!
moduleExtension
	^ self isCPP ifTrue: ['.cpp'] ifFalse: ['.c']! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'tpr 12/3/2015 10:35'!
moduleFileName
	"Answer the receiver's module name that is used for the plugin's C code."

	^ self moduleName, self moduleExtension! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'sma 3/3/2000 12:24'!
moduleName
	"Answer the receiver's module name that is used for the plugin's C code."

	^ self name asString! !

!InterpreterPlugin class methodsFor: 'accessing' stamp: 'eem 2/10/2009 14:48'!
moduleNameAndVersion
	"Answer the receiver's module name and version info that is used for the plugin's C code. The default is to append the code generation date, but any useful text is ok (keep it short)"

	^ self moduleName! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'tpr 12/3/2015 10:26'!
pluginClassesUpToRoot
	"Answer the classes to include for translation of aPluginClass, superclasses first, and the root (VMClass in general, possibly Object)  last."
	| theClass classes |

	classes := OrderedCollection new.
	theClass := self.
	[theClass == Object
	 or: [theClass == VMClass]] whileFalse:
		[classes addLast: theClass.
		theClass := theClass superclass].
	^classes reverse! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'eem 11/4/2009 15:30'!
preambleCCode
	"Defining this method to answer a string dumps that string in VMPluginCodeGenerator>>emitCHeaderOn:"
	^nil! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'eem 12/12/2014 16:46'!
pruneUnusedInterpreterPluginMethodsIn: aCodeGen
	aCodeGen unreachableMethods do:
		[:m|
		 m definingClass = InterpreterPlugin ifTrue:
			[aCodeGen removeMethodForSelector: m selector]]! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'tpr 7/2/2001 16:33'!
requiresCrossPlatformFiles
	"default is ok for most, any plugin needing cross platform files aside from a normal header must say so. See SoundCodecPlugin for example"
	^self hasHeaderFile! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'tpr 11/21/2000 11:53'!
requiresPlatformFiles
	"default is ok for most, any plugin needing platform specific files must say so"
	^false! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'tpr 5/14/2001 12:05'!
shouldBeTranslated
"is this class intended to be translated as a plugin? Most subclasses should answer true, but some such as:-
	TestInterpreterPlugin
	FlippArrayPlugin2
	InflatePlugin
	should answer false for various reasons."
	^true! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'eem 6/3/2011 09:09'!
shouldBeTranslatedFor: platformName
	"Is this class intended to be translated as a plugin, perhaps specific to a platform?
	 Most subclasses should answer true, but some such as simulation-only versions
	 should answer false for various reasons."
	^self shouldBeTranslated! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'eem 11/22/2013 15:59'!
shouldGenerateDeadCode
	"Answer if the code generator should generate dead code, e.g. in false ifTrue: [dead] ifFalse: [live].
	 Since plugin source is shared between different VM builds it is unsafe to assume any code is dead."

	^true! !

!InterpreterPlugin class methodsFor: 'simulation' stamp: 'ajh 8/21/2002 21:43'!
simulatorClass
	"For running from Smalltalk - answer a class that can be used to simulate the receiver, or nil if you want the primitives in this module to always fail, causing simulation to fall through to the Smalltalk code.  By default every non-TestInterpreterPlugin can simulate itself."

	^ self! !

!InterpreterPlugin class methodsFor: 'simulation' stamp: 'eem 10/26/2018 09:13'!
simulatorForInterpreterInterface: objectMemoryOrInterpreterProxy
	"Answer an uninitialized (in the sense that the plugin has been sent initialize, but not initialiseModule)
	 simulator for the receiver, or nil, if the receiver answers nil to simulatorClass.  Overridden by
	 SmartSyntaxPluginSimulator to answer an instance of SmartSyntaxPluginSimulator wrapped around
	 an uniniialized simulator for the receiver."

	^self simulatorClass ifNotNil:
		[:simClass|
		 simClass new
			setInterpreter: objectMemoryOrInterpreterProxy;
			yourself]! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'eem 5/17/2011 12:20'!
storeString: s onFileNamed: fileName
	"Store the given string in a file of the given name."

	| f |
	f := VMMaker forceNewFileNamed: fileName.
	f nextPutAll: s.
	f close.! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'tpr 12/3/2015 10:27'!
translateInDirectory: directory doInlining: inlineFlag
"This is the default method for writing out sources for a plugin. Several classes need special handling, so look at all implementors of this message"
	| cg fname |
	 fname := self moduleName, '.c'.

	"don't translate if the file is newer than my timeStamp"
	(directory entryAt: fname ifAbsent: nil) ifNotNil:
		[:fstat| | mTime |
		mTime := fstat modificationTime.
		mTime isInteger ifFalse: [mTime := mTime asSeconds].
		 (self allCodeOlderThan: mTime) ifTrue:
			[^nil]].

	self initialize.
	cg := self buildCodeGenerator.
	cg inferTypesForImplicitlyTypedVariablesAndMethods.
	self pruneUnusedInterpreterPluginMethodsIn: cg.
	cg storeCodeOnFile:  (directory fullNameFor: fname) doInlining: inlineFlag.
	^cg exportedPrimitiveNames asArray! !

!InterpreterPlugin class methodsFor: 'translation' stamp: 'eem 2/16/2015 14:19'!
translatedPrimitives
	"Answer an Array of Class, selector pair Arrays for any primitives to
	 be translated from Smalltalk methods elsewhere in the system.
	 By default answer none.  Some subclasses redefine as required."
	^#()! !

!InterpreterPlugin methodsFor: 'simulation' stamp: 'eem 10/13/2020 09:08'!
cCoerce: value to: cTypeString
	<doNotGenerate>
	^value coerceTo: cTypeString sim: interpreterProxy! !

!InterpreterPlugin methodsFor: 'simulation' stamp: 'eem 12/22/2017 13:56'!
cCoerceSimple: value to: cType
	<doNotGenerate>
	"Coercion without type mapping.  Don't even bother to check for valid types..."
	^value! !

!InterpreterPlugin methodsFor: 'simulation support' stamp: 'eem 12/29/2016 13:01'!
close
	"Simulation subclasses needing some specific close action override as required."
	<doNotGenerate>! !

!InterpreterPlugin methodsFor: 'simulation support' stamp: 'eem 8/3/2021 18:42'!
doesNotUnderstand: aMessage
	<doNotGenerate>
	"Override doesNotUnderstand: to intercept sends of translated primitive selectors.
	 The translated primitives are primitives derived from the primitive methods themselves
	 translating their failure code/method body into Slang code.  There used to be several
	 translated primitives in the MiscPrimitivePlugin, but these have been replaced.  As of
	 mid 2021 only ADPCMCodec[Plugin] class and AbstractSound/SoundGenerationPlugin class
	 define translatedPrimitives."
	(self methodAndTypesOrNilForTranslatedPrimitiveSelector: aMessage selector)
		ifNil: [^super doesNotUnderstand: aMessage]
		ifNotNil:
			[:tuple| | method |
			 "First check the cache for validity; if the last element of the tuple is the actual method
			 then the cache is up-to-date.  if it is not, the method has changed and should be regenerated."
			 method := tuple last.
			 method == (method methodClass >> method selector) ifFalse:
				[translatedMethodCache removeKey: aMessage selector.
				 ^self doesNotUnderstand: aMessage].
			 method := tuple first.
			 tuple second
				ifNil: [interpreterProxy primitiveFail]
				ifNotNil:
					[:types|
					 self tryToRunTranslatedPrimitive: method types: types subsidiaries: tuple third].
			 interpreterProxy transcript print: method; cr.
			 interpreterProxy coInterpreter printExternalHeadFrame.
			 interpreterProxy failed ifTrue:
				[interpreterProxy transcript
					nextPutAll: 'WARNING!! Failing translated primitive ';
					nextPutAll: aMessage selector;
					nextPutAll: ' implemented by ';
					nextPutAll: method methodClass name;
					nextPutAll: '>>';
					nextPutAll: method selector;
					cr;
					flush]]! !

!InterpreterPlugin methodsFor: 'simulation' stamp: 'eem 05/07/2020 16:13'!
evaluateIfFailed: aBlock
	"Evaluate aBlock, catching primitive failure, and failing if so.
	 Answer if evaluating aBlock caused primitive failure."
	<doNotGenerate>
	aBlock
		on: Error
		do: [:ex|
			((ex signalerContext selector beginsWith: #primitiveFailed) "e.g. could be error: sent from primitiveFailed:"
			 or: [ex signalerContext sender selector beginsWith: #primitiveFailed]) ifFalse:
				[ex pass].
			interpreterProxy primitiveFail.
			^true].
	^false! !

!InterpreterPlugin methodsFor: 'simulation support' stamp: 'eem 3/10/2015 17:41'!
executeTranslatedPrimitiveMethod: method arguments: args subsidiaries: subsidiaryMethods
	<doNotGenerate>
	"Execute the method within a doesNotUnderstand: handler that will
	 catch sends of asciiValue and evaluate any subsidiary methods.."
	^[interpreterProxy withArgs: args executeMethod: method]
		on: MessageNotUnderstood
		do: [:ex|
			ex receiver == interpreterProxy ifTrue:
				[ex resume:
					(self executeTranslatedPrimitiveMethod:
							(subsidiaryMethods
								at: ex message selector
								ifAbsent: [ex pass])
						arguments: ex message arguments
						subsidiaries: subsidiaryMethods)].
			(ex receiver isInteger
			and: [ex message selector == #asciiValue]) ifTrue:
				[ex resume: ex receiver].
			ex pass]! !

!InterpreterPlugin methodsFor: 'initialize' stamp: 'eem 11/13/2014 16:42'!
expandDereferenceInterpreterProxyFunctionTable
	"This is a dummy function that the VMPluginCodeGenerator expands into a
	 sequence of assignments from interpreterProxy funcitons to local function pointers."
	<doNotGenerate>! !

!InterpreterPlugin methodsFor: 'initialize' stamp: 'ikp (auto pragmas 12/08) 8/3/2004 19:18'!
getInterpreter
	"Note: This is coded so that plugins can be run from Squeak."

	<returnTypeC: 'VirtualMachine *'>
	^interpreterProxy! !

!InterpreterPlugin methodsFor: 'initialize' stamp: 'eem 9/27/2020 15:00'!
getModuleName
	"Note: This is hardcoded so it can be run from Squeak.
	The module name is used for validating a module *after*
	it is loaded to check if it does really contain the module
	we're thinking it contains. This is important!!"
	<returnTypeC: #'const char*'>
	<export: true>
	^self cCode: [moduleName]
		inSmalltalk:
			[| string index |
			 string := ((self class codeGeneratorClass new pluginClass: self class) variableDeclarationStringsForVariable: 'moduleName') first.
			 index := (string indexOfSubCollection: 'moduleName = "') + 14.
			 (string copyFrom: index to: (string indexOf: $" startingAt: index + 1) - 1), '(i)']! !

!InterpreterPlugin methodsFor: 'debugging' stamp: 'eem 10/18/2018 11:49'!
halt
	self cCode: '' inSmalltalk: [^super halt "avoid the ^0 below"].
	^0! !

!InterpreterPlugin methodsFor: 'alien support' stamp: 'eem 9/27/2020 14:46'!
isAlien: oop
	"Answer if oop is an Alien.  We could ask if isWordsOrBytes: first, but that doesn't help.  We still have to do the is:KindOf: walk.
	 We're not interested in fast falsehood, but as fast as possible truth, and with the current API this is it."
	<inline: true>
	^interpreterProxy is: oop KindOfClass: interpreterProxy classAlien! !

!InterpreterPlugin methodsFor: 'testing' stamp: 'eem 3/17/2014 12:26'!
isInterpreterPlugin
	<doNotGenerate>
	"Simulation only"
	^true! !

!InterpreterPlugin methodsFor: 'accessing' stamp: 'eem 7/2/2014 19:06'!
isSmartSyntaxPluginSimulator
	<doNotGenerate>
	^false! !

!InterpreterPlugin methodsFor: 'simulation support' stamp: 'eem 12/7/2012 14:39'!
majorVersion
	"This is implemented in sqVirtualMachine.c, so this form is for simulation only."
	<doNotGenerate>
	^1! !

!InterpreterPlugin methodsFor: 'simulation support' stamp: 'akg 10/22/2018 18:37'!
memcpy: dest _: src _: bytes
	<doNotGenerate>
	"implementation of memcpy(3). N.B. If ranges overlap, must use memmove."

	^interpreterProxy memcpy: dest _: src _: bytes
! !

!InterpreterPlugin methodsFor: 'simulation support' stamp: 'akg 10/22/2018 18:38'!
memmove: destAddress _: sourceAddress _: bytes
	<doNotGenerate>
	"implementation of memcpy(3). N.B. If ranges overlap, must use memmove."

	^interpreterProxy memmove: destAddress _: sourceAddress _: bytes! !

!InterpreterPlugin methodsFor: 'simulation' stamp: 'eem 3/7/2015 18:41'!
methodAndTypesOrNilForTranslatedPrimitiveSelector: selector
	<doNotGenerate>
	"If selector maps to a simulateable translated primitive method, then
	 answer the method and its types for selector, the selector of a translated primitive.
	 Otherwise answer nil.  This caches the results of analysis in translatedMethodCache."
	translatedMethodCache ifNil:
		[translatedMethodCache := IdentityDictionary new.
		 translatedMethodCache at: #CCodeGenerator put: CCodeGenerator new].
	^translatedMethodCache
		at: selector
		ifAbsentPut: [self tupleOrNilForTranslatedPrimitiveSelector: selector]! !

!InterpreterPlugin methodsFor: 'API access' stamp: 'eem 8/14/2020 13:53'!
methodReturnStringOrNil: cString
	"methodReturnString: fails for a nil value; this convenience answers nil for a nil value."
	<var: 'cString' type: #'char *'>
	<inline: #always>
	cString
		ifNil: [interpreterProxy methodReturnValue: interpreterProxy nilObject]
		ifNotNil: [interpreterProxy methodReturnString: cString]! !

!InterpreterPlugin methodsFor: 'simulation' stamp: 'eem 3/10/2015 17:42'!
methodWithoutPrimitive: method
	"Answer either aMethod or a copy of aMehtod, such that the result does /not/ have a primitive."
	<doNotGenerate>
	| mn properties newProperties |
	method primitive = 0 ifTrue:
		[^method].
	mn := method methodNode.
	properties := mn properties.
	newProperties := properties copy.
	properties pragmas do:
		[:pragma|
		(pragma keyword beginsWith: #primitive:) ifTrue:
			[newProperties := newProperties copyWithout: pragma]].
	^mn
		instVarNamed: 'primitive' put: 0;
		instVarNamed: 'properties' put: newProperties;
		generate: method trailer using: method class! !

!InterpreterPlugin methodsFor: 'simulation support' stamp: 'eem 12/7/2012 14:40'!
minorVersion
	"This is implemented in sqVirtualMachine.c, so this form is for simulation only."
	<doNotGenerate>
	^8! !

!InterpreterPlugin methodsFor: 'debugging' stamp: 'jcg (auto pragmas 12/08) 6/30/2008 17:36'!
msg: s 
	<var: #s type: 'char *'>
	self cCode: 'fprintf(stderr, "\n%s: %s", moduleName, s)' inSmalltalk: [Transcript cr; show: self class moduleName , ': ' , s; endEntry].
	^0! !

!InterpreterPlugin methodsFor: 'simulation support' stamp: 'eem 8/1/2021 12:36'!
objectMemory
	^interpreterProxy objectMemory! !

!InterpreterPlugin methodsFor: 'API access' stamp: 'eem 7/21/2017 11:26'!
positiveMachineIntegerFor: value
	<var: #value type: #'usqIntptr_t'>
	<inline: #always>
	^interpreterProxy wordSize = 8
		ifTrue: [interpreterProxy positive64BitIntegerFor: value]
		ifFalse: [interpreterProxy positive32BitIntegerFor: value]! !

!InterpreterPlugin methodsFor: 'simulation' stamp: 'eem 7/2/2014 15:29'!
remapOop: oopOrList in: aBlock
	"Call remapOop: for the variable oopOrList (or all of the variables in oopOrList) before evaluating
	 aBlock, and restore them after.  If this is Spur, do nothing, since Spur does not GC on allocation
	 and the  SmartSyntaxPluginCodeGenerator generates null code for this op in Spur."
	<doNotGenerate>
	| ctxt tempNames tempIndices |
	interpreterProxy hasSpurMemoryManagerAPI ifTrue:
		[^aBlock value].
	ctxt := thisContext sender.
	tempNames := ctxt tempNames.
	oopOrList isArray
		ifTrue:
			[tempIndices := oopOrList collect: [:tempName| tempNames indexOf: tempName].
			 tempIndices do:
				[:index| interpreterProxy pushRemappableOop: (ctxt namedTempAt: index)]]
		ifFalse: [interpreterProxy pushRemappableOop: oopOrList].
	^aBlock ensure:
		[oopOrList isArray
			ifTrue:
				[tempIndices reverseDo:
					[:index| ctxt namedTempAt: index put: interpreterProxy popRemappableOop]]
			ifFalse:
				[1 to: ctxt numTemps do:
					[:index|
					(ctxt tempAt: index) = oopOrList ifTrue:
						[ctxt tempAt: index put: interpreterProxy topRemappableOop]].
				 interpreterProxy popRemappableOop]]! !

!InterpreterPlugin methodsFor: 'initialize' stamp: 'eem 1/18/2020 11:24'!
setInterpreter: anInterpreter 
	"Note: This is coded so that it can be run in Squeak."

	| ok |
	<export: true>
	<var: #anInterpreter type: #'struct VirtualMachine*'>
	interpreterProxy := anInterpreter.
	"This may seem tautological, but in a real plugin it checks that the VM provides
	 the version the plugin was compiled against which is the version the plugin expects."
	ok := interpreterProxy majorVersion = (self cCode: 'VM_PROXY_MAJOR' inSmalltalk: [self majorVersion])
		and: [interpreterProxy minorVersion >= (self cCode: 'VM_PROXY_MINOR' inSmalltalk: [self minorVersion])].
	ok ifTrue:
		[self expandDereferenceInterpreterProxyFunctionTable].
	^ok! !

!InterpreterPlugin methodsFor: 'API access' stamp: 'eem 7/21/2017 11:29'!
signedMachineIntegerFor: value
	<var: #value type: #'sqIntptr_t'>
	<inline: #always>
	^interpreterProxy wordSize = 8
		ifTrue: [interpreterProxy signed64BitIntegerFor: value]
		ifFalse: [interpreterProxy signed32BitIntegerFor: value]! !

!InterpreterPlugin methodsFor: 'alien support' stamp: 'eem 11/23/2020 18:14'!
sizeField: alienOop
	"Answer the size field of an alienOop which is assumed to be an Alien of at least 8 bytes (32-bits) or 16 bytes (64 bits)"
	<inline: true>
	^self cCode:
			[self longAt: alienOop + interpreterProxy baseHeaderSize]
		inSmalltalk:
			[| field |
			 field := interpreterProxy longAt: alienOop + interpreterProxy baseHeaderSize.
			 interpreterProxy wordSize = 8
				ifTrue: [field signedIntFromLong64]
				ifFalse: [field signedIntFromLong]]! !

!InterpreterPlugin methodsFor: 'simulation support' stamp: 'nice 11/4/2016 21:33'!
sizeof: objectSymbolOrClass
	<doNotGenerate>
	objectSymbolOrClass isInteger ifTrue:
		[^interpreterProxy wordSize].
	objectSymbolOrClass isSymbol ifTrue:
		["In the simulator file handles are just integer indices into openFiles and so need
		 only be BytesPerWord big. But in the actual VM they are at least 5 words long."
		objectSymbolOrClass == #SQFile ifTrue:
			[^interpreterProxy wordSize * 5].
		"SQSocket is typedef struct { int sessionID; int socketType; void *privateSocketPtr; } SQSocket"
		objectSymbolOrClass == #SQSocket ifTrue:
			[^8 + interpreterProxy wordSize].
		"We assume the file offset type is always 64-bits."
		objectSymbolOrClass == #squeakFileOffsetType ifTrue:
			[^8].
		(objectSymbolOrClass last == $*
		 or: [#long == objectSymbolOrClass
		 or: [#'unsigned long' == objectSymbolOrClass
		 or: [#'sqIntptr_t' == objectSymbolOrClass
		 or: [#'usqIntptr_t' == objectSymbolOrClass
		 or: [#'size_t' == objectSymbolOrClass]]]]]) ifTrue:
			[^interpreterProxy wordSize].
		(#(usqInt sqInt) includes: objectSymbolOrClass) ifTrue:
			[^interpreterProxy bytesPerOop]].
	^super sizeof: objectSymbolOrClass! !

!InterpreterPlugin methodsFor: 'API access' stamp: 'eem 11/10/2020 11:31'!
stackBooleanValue: index
	<inline: #always>
	^interpreterProxy booleanValueOf: (interpreterProxy stackValue: index)! !

!InterpreterPlugin methodsFor: 'API access' stamp: 'eem 11/24/2020 19:07'!
stackEphemeralStringValue: index
	"Convenience to answer a given argument as a C string using stack allocation.
	 Fails if the argument is not a string or there is no memory."
	<returnTypeC: #'char *'>
	<inline: #always>
	| obj sz dstPtr |
	<var: 'dstPtr' type: #'char *'>
	obj := interpreterProxy stackValue: index.
	(interpreterProxy isBytes: obj) ifFalse:
		[interpreterProxy primitiveFailFor: PrimErrBadArgument.
		^nil].
	sz := interpreterProxy byteSizeOf: obj.
	dstPtr := self alloca: sz+1.
	dstPtr ifNil:
		[interpreterProxy primitiveFailFor: PrimErrNoCMemory.
		^nil].
	self memcpy: dstPtr _: (interpreterProxy firstIndexableField: obj) _: sz.
	dstPtr at: sz put: 0.
	^dstPtr! !

!InterpreterPlugin methodsFor: 'API access' stamp: 'eem 8/14/2020 14:13'!
stackStringValue: index
	"Convenience to answer a given argument as a C string.
	 Fails if the argument is not a string or there is no memory."
	<returnTypeC: #'char *'>
	| obj sz dstPtr |
	<var: 'dstPtr' type: #'char *'>
	obj := interpreterProxy stackValue: index.
	(interpreterProxy isBytes: obj) ifFalse:
		[interpreterProxy primitiveFailFor: PrimErrBadArgument.
		^nil].
	sz := interpreterProxy byteSizeOf: obj.
	dstPtr := self malloc: sz+1.
	dstPtr ifNil:
		[interpreterProxy primitiveFailFor: PrimErrNoCMemory.
		^nil].
	self memcpy: dstPtr _: (interpreterProxy firstIndexableField: obj) _: sz.
	dstPtr at: sz put: 0.
	^dstPtr! !

!InterpreterPlugin methodsFor: 'alien support' stamp: 'eem 10/28/2020 01:04'!
startOfData: alienOop
	"Answer the start of an Alien's data.  For direct aliens this is the address of the second field.
	 For indirect and pointer aliens it is what the second field points to."
	^(self sizeField: alienOop) > 0
	 	ifTrue: [alienOop + interpreterProxy baseHeaderSize + interpreterProxy bytesPerOop]
		ifFalse: [interpreterProxy longAt: alienOop + interpreterProxy baseHeaderSize + interpreterProxy bytesPerOop]! !

!InterpreterPlugin methodsFor: 'simulation support' stamp: 'eem 2/5/2015 14:42'!
strlen: aCString
	<doNotGenerate>
	^interpreterProxy strlen: aCString! !

!InterpreterPlugin methodsFor: 'simulation support' stamp: 'eem 9/19/2018 18:02'!
strncpy: aString _: bString _: n
	<doNotGenerate>
	^interpreterProxy strncpy: aString _: bString _: n! !

!InterpreterPlugin methodsFor: 'simulation' stamp: 'eem 12/22/2017 13:55'!
translatedPrimitiveArgument: index ofType: cTypeString using: aCCodeGenerator
	<doNotGenerate>
	| oop unitSize |
	oop := interpreterProxy stackValue: interpreterProxy methodArgumentCount - index.
	(interpreterProxy isOopForwarded: oop) ifTrue: [^nil]. 
	cTypeString last == $* ifTrue:
		[unitSize := self sizeof: (aCCodeGenerator baseTypeForPointerType: cTypeString) asSymbol.
		 unitSize caseOf: {
			[1]	->	[(interpreterProxy isBytes: oop) ifFalse: [^nil]].
			[2]	->	[(interpreterProxy isShorts: oop) ifFalse: [^nil]].
			[4]	->	[(interpreterProxy isWords: oop) ifFalse: [^nil]].
			[8]	->	[(interpreterProxy isLong64s: oop) ifFalse: [^nil]] }
			otherwise: [^nil].
		^ObjectProxyForTranslatedPrimitiveSimulation new
			interpreter: interpreterProxy
			oop: oop
			unitSize: unitSize].
	((interpreterProxy isIntegerObject: oop)
	 and: [aCCodeGenerator isIntegralCType: cTypeString]) ifTrue:
		[^interpreterProxy integerValueOf: oop].
	self halt! !

!InterpreterPlugin methodsFor: 'simulation' stamp: 'eem 3/9/2015 10:34'!
tryToRunTranslatedPrimitive: method types: types subsidiaries: subsidiaryMethods
	"Evaluate a translated primitive method using the receiver as its receiver.
	 Supply integers or proxies for the arguments as necessary."
	<doNotGenerate>
	| cg args result |
	interpreterProxy methodArgumentCount ~= method numArgs ifTrue:
		[^interpreterProxy primitiveFail].
	cg := translatedMethodCache at: #CCodeGenerator.
	args := (1 to: method numArgs)
				with: types
				collect:
					[:i :type |
					 (self translatedPrimitiveArgument: i ofType: type using: cg) ifNil:
						[^interpreterProxy primitiveFail]].
	result := self executeTranslatedPrimitiveMethod: method arguments: args subsidiaries: subsidiaryMethods.
	interpreterProxy failed ifFalse:
		[result == interpreterProxy
			ifTrue: [interpreterProxy pop: interpreterProxy methodArgumentCount]
			ifFalse:
				[result isInteger
					ifTrue: [interpreterProxy methodReturnValue: (interpreterProxy integerObjectOf: result)]
					ifFalse: [self halt]]]! !

!InterpreterPlugin methodsFor: 'simulation' stamp: 'eem 3/13/2015 18:18'!
tupleOrNilForTranslatedPrimitiveSelector: selector
	"Answer a tuple of {simulation method, types, subsidiary methods, original method}
	 for the selector of a translated primitive.  If the method cannot be simulated, for
	 example if it accesses instance variables, answer a tuple whose types element is nil."
	<doNotGenerate>
	^(self class methodOrNilForTranslatedPrimitiveSelector: selector) ifNotNil:
		[:method| | argNames argPragmas cg types subsidiaryMethods |
		"Since the plugin itself runs the method, and the method is on some
		 distant class, if the method accesses inst vars, the mechanism can't work."
		method hasInstVarRef ifTrue: [^{method. nil. nil. method}].
		argNames := [method methodClass newParser parseParameterNames: method getSource]
						on: Error
						do: [:ex|
							^{method. nil. nil. method}].
		argPragmas := method pragmas select:
							[:p|
							(p keyword beginsWith: 'var:')
							and: [argNames includes: p arguments first]].
		cg := translatedMethodCache at: #CCodeGenerator.
		types := (1 to: method numArgs) collect:
					[:i|
					(argPragmas detect: [:p| p arguments first = (argNames at: i)] ifNone: [])
						ifNil: [#sqInt]
						ifNotNil:
							[:pragma|
							cg extractTypeFor: (argNames at: i) fromDeclaration: pragma arguments second]].
		 "Subsidiary methods are typically implemented for the primitive method only."
		 subsidiaryMethods :=
			method messages
				select:
					[:subsidiary|
					 (method methodClass includesSelector: subsidiary)
					 and: [(Object includesSelector: subsidiary) not]]
				thenCollect:
					[:subsidiary| | subsidiaryMethod |
					subsidiaryMethod := method methodClass >> subsidiary.
					subsidiaryMethod hasInstVarRef ifTrue:
						[^{method. nil. nil. method}].
					subsidiaryMethod].
		 { self methodWithoutPrimitive: method.
		   types.
		   Dictionary withAll:
			(subsidiaryMethods collect:
				[:m| m selector -> (self methodWithoutPrimitive: m)]).
		   method}]! !

InterpreterPlugin subclass: #JPEGReaderPlugin
	instanceVariableNames: 'yComponent crComponent cbComponent ySampleStream crSampleStream cbSampleStream yBlocks crBlocks cbBlocks residuals ditherMask jpegBits jpegBitsSize jpegNaturalOrder jsCollection jsPosition jsReadLimit jsBitBuffer jsBitCount acTable dcTable acTableSize dcTableSize'
	classVariableNames: 'BlockWidthIndex BlueIndex ConstBits CurrentXIndex CurrentYIndex DCTSize DCTSize2 FIXn0n298631336 FIXn0n34414 FIXn0n390180644 FIXn0n541196100 FIXn0n71414 FIXn0n765366865 FIXn0n899976223 FIXn1n175875602 FIXn1n40200 FIXn1n501321110 FIXn1n77200 FIXn1n847759065 FIXn1n961570560 FIXn2n053119869 FIXn2n562915447 FIXn3n072711026 GreenIndex HScaleIndex LookaheadBitsIndex LookaheadSymbolIndex MCUBlockIndex MCUWidthIndex MaxBits MaxMCUBlocks MaxSample MaxcodeIndex MinComponentSize Pass1Bits Pass1Div Pass2Div PriorDCValueIndex RedIndex SampleOffset VScaleIndex'
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!JPEGReaderPlugin commentStamp: 'tpr 5/5/2003 12:10' prior: 0!
This is another JPEG reader plugin, this time not requiring jpeglib support. !

!JPEGReaderPlugin class methodsFor: 'translation' stamp: 'ar 3/4/2001 21:13'!
declareCVarsIn: cg
	cg var: 'yComponent'
		declareC: 'int yComponent[' , MinComponentSize printString , ']'.
	cg var: 'crComponent'
		declareC: 'int crComponent[' , MinComponentSize printString , ']'.
	cg var: 'cbComponent'
		declareC: 'int cbComponent[' , MinComponentSize printString , ']'.
	cg var: 'yBlocks'
		declareC: 'int *yBlocks[' , MaxMCUBlocks printString , ']'.
	cg var: 'crBlocks'
		declareC: 'int *crBlocks[' , MaxMCUBlocks printString  , ']'.
	cg var: 'cbBlocks'
		declareC: 'int *cbBlocks[' , MaxMCUBlocks printString  , ']'.
	cg var: 'residuals'
		declareC: 'int *residuals'.
	cg var: 'jpegBits'
		declareC: 'int *jpegBits'.

	cg var: 'jpegNaturalOrder'
		declareC: 'int jpegNaturalOrder[64] = {
	0, 1, 8, 16, 9, 2, 3, 10, 
	17, 24, 32, 25, 18, 11, 4, 5, 
	12, 19, 26, 33, 40, 48, 41, 34, 
	27, 20, 13, 6, 7, 14, 21, 28, 
	35, 42, 49, 56, 57, 50, 43, 36, 
	29, 22, 15, 23, 30, 37, 44, 51, 
	58, 59, 52, 45, 38, 31, 39, 46, 
	53, 60, 61, 54, 47, 55, 62, 63
}'.

	cg var: 'jsCollection' 
		declareC:'unsigned char *jsCollection'.
	cg var: 'acTable' 
		declareC:'int *acTable'.
	cg var: 'dcTable' 
		declareC:'int *dcTable'.
! !

!JPEGReaderPlugin class methodsFor: 'class initialization' stamp: 'ar 3/4/2001 19:13'!
initialize
	"JPEGReaderPlugin initialize"
	DCTSize := 8.
	MaxSample := (2 raisedToInteger: DCTSize) - 1.
	SampleOffset := MaxSample // 2.
	DCTSize2 := DCTSize squared.
	ConstBits := 13.
	Pass1Bits := 2.
	Pass1Div := 1 bitShift: ConstBits - Pass1Bits.
	Pass2Div := 1 bitShift: ConstBits + Pass1Bits + 3.

	"fixed-point Inverse Discrete Cosine Transform (IDCT) constants"
	FIXn0n298631336 := 2446.
	FIXn0n390180644 := 3196.
	FIXn0n541196100 := 4433.
	FIXn0n765366865 := 6270.
	FIXn0n899976223 := 7373.
	FIXn1n175875602 := 9633.
	FIXn1n501321110 := 12299.
	FIXn1n847759065 := 15137.
	FIXn1n961570560 := 16069.
	FIXn2n053119869 := 16819.
	FIXn2n562915447 := 20995.
	FIXn3n072711026 := 25172.

	"fixed-point color conversion constants"
	FIXn0n34414 := 22554.
	FIXn0n71414 := 46802.
	FIXn1n40200 := 91881.
	FIXn1n77200 :=  116130.

	CurrentXIndex := 0.
	CurrentYIndex := 1.
	HScaleIndex := 2.
	VScaleIndex := 3.
	MCUBlockIndex := 4.
	BlockWidthIndex := 5.
	MCUWidthIndex := 8.
	PriorDCValueIndex := 10.
	MinComponentSize := 11.

	RedIndex := 0.
	GreenIndex := 1.
	BlueIndex := 2.

	MaxMCUBlocks := 128.
	MaxBits := 16.! !

!JPEGReaderPlugin methodsFor: 'decoding' stamp: 'ar 3/4/2001 21:11'!
cbColorComponentFrom: oop
	^(self colorComponent: cbComponent from: oop)
		and:[self colorComponentBlocks: cbBlocks from: oop]! !

!JPEGReaderPlugin methodsFor: 'decoding' stamp: 'eem 1/30/2014 15:34'!
colorComponent: aColorComponent from: oop
	<var: #aColorComponent type: 'int *'>
	(interpreterProxy isPointers: oop) ifFalse:[^false].
	(interpreterProxy slotSizeOf: oop) < MinComponentSize ifTrue:[^false].
	aColorComponent at: CurrentXIndex put: 
		(interpreterProxy fetchInteger: CurrentXIndex ofObject: oop).
	aColorComponent at: CurrentYIndex put: 
		(interpreterProxy fetchInteger: CurrentYIndex ofObject: oop).
	aColorComponent at: HScaleIndex put: 
		(interpreterProxy fetchInteger: HScaleIndex ofObject: oop).
	aColorComponent at: VScaleIndex put: 
		(interpreterProxy fetchInteger: VScaleIndex ofObject: oop).
	aColorComponent at: BlockWidthIndex put: 
		(interpreterProxy fetchInteger: BlockWidthIndex ofObject: oop).
	aColorComponent at: MCUWidthIndex put: 
		(interpreterProxy fetchInteger: MCUWidthIndex ofObject: oop).
	aColorComponent at: PriorDCValueIndex put: 
		(interpreterProxy fetchInteger: PriorDCValueIndex ofObject: oop).
	^interpreterProxy failed not! !

!JPEGReaderPlugin methodsFor: 'decoding' stamp: 'eem 1/30/2014 15:33'!
colorComponentBlocks: blocks from: oop
	| arrayOop max blockOop |
	<var: #blocks type: 'int **'>
	(interpreterProxy isPointers: oop) ifFalse:[^false].
	(interpreterProxy slotSizeOf: oop) < MinComponentSize ifTrue:[^false].

	arrayOop := interpreterProxy fetchPointer: MCUBlockIndex ofObject: oop.
	(interpreterProxy isPointers: arrayOop) ifFalse:[^false].
	max := interpreterProxy slotSizeOf: arrayOop.
	max > MaxMCUBlocks ifTrue:[^false].
	0 to: max-1 do:[:i|
		blockOop := interpreterProxy fetchPointer: i ofObject: arrayOop.
		(interpreterProxy isWords: blockOop) ifFalse:[^false].
		(interpreterProxy slotSizeOf: blockOop) = DCTSize2 ifFalse:[^false].
		blocks at: i put: (interpreterProxy firstIndexableField: blockOop).
	].
	^interpreterProxy failed not! !

!JPEGReaderPlugin methodsFor: 'decoding' stamp: 'ar 3/4/2001 22:24'!
colorConvertGrayscaleMCU
	| y |
	yComponent at: CurrentXIndex put: 0.
	yComponent at: CurrentYIndex put: 0.
	0 to: jpegBitsSize-1 do:[:i|
		y := self nextSampleY.
		y := y + (residuals at: GreenIndex).
		y := y min: MaxSample.
		residuals at: GreenIndex put: (y bitAnd: ditherMask).
		y := y bitAnd: MaxSample - ditherMask.
		y := y max: 1.
		jpegBits at: i put: 16rFF000000 + (y<<16) + (y<<8) + y.
	].! !

!JPEGReaderPlugin methodsFor: 'decoding' stamp: 'ar 3/4/2001 21:11'!
colorConvertMCU
	| y cb cr red green blue |
	yComponent at: CurrentXIndex put: 0.
	yComponent at: CurrentYIndex put: 0.
	cbComponent at: CurrentXIndex put: 0.
	cbComponent at: CurrentYIndex put: 0.
	crComponent at: CurrentXIndex put: 0.
	crComponent at: CurrentYIndex put: 0.
	0 to: jpegBitsSize-1 do:[:i|
		y := self nextSampleY.
		cb := self nextSampleCb.
		cb := cb - SampleOffset.
		cr := self nextSampleCr.
		cr := cr - SampleOffset.
		red := y + ((FIXn1n40200 * cr) // 65536) + (residuals at: RedIndex).
		red := red min: MaxSample. red := red max: 0.
		residuals at: RedIndex put: (red bitAnd: ditherMask).
		red := red bitAnd: MaxSample - ditherMask.
		red := red max: 1.
		green := y - ((FIXn0n34414 * cb) // 65536) -
			((FIXn0n71414 * cr) // 65536) + (residuals at: GreenIndex).
		green := green min: MaxSample. green := green max: 0.
		residuals at: GreenIndex put: (green bitAnd: ditherMask).
		green := green bitAnd: MaxSample - ditherMask.
		green := green max: 1.
		blue := y + ((FIXn1n77200 * cb) // 65536) + (residuals at: BlueIndex).
		blue := blue min: MaxSample. blue := blue max: 0.
		residuals at: BlueIndex put: (blue bitAnd: ditherMask).
		blue := blue bitAnd: MaxSample - ditherMask.
		blue := blue max: 1.
		jpegBits at: i put: 16rFF000000 + (red bitShift: 16) + (green bitShift: 8) + blue.
	].! !

!JPEGReaderPlugin methodsFor: 'decoding' stamp: 'ar 3/4/2001 21:11'!
crColorComponentFrom: oop
	^(self colorComponent: crComponent from: oop)
		and:[self colorComponentBlocks: crBlocks from: oop]! !

!JPEGReaderPlugin methodsFor: 'decoding' stamp: 'ar (auto pragmas 12/08) 3/4/2001 21:17'!
decodeBlockInto: anArray component: aColorComponent
	| byte zeroCount bits index |
	<var: #anArray type: 'int *'>
	<var: #aColorComponent type: 'int *'>
	byte := self jpegDecodeValueFrom: dcTable size: dcTableSize.
	byte < 0 ifTrue:[^interpreterProxy primitiveFail].
	byte ~= 0 ifTrue: [
		bits := self getBits: byte.
		byte := self scaleAndSignExtend: bits inFieldWidth: byte].
	byte := aColorComponent 
				at: PriorDCValueIndex 
				put: (aColorComponent at: PriorDCValueIndex) + byte.
	anArray at: 0 put: byte.
	1 to: DCTSize2 - 1 do:[:i| anArray at: i put: 0].
	index := 1.
	[index < DCTSize2] whileTrue:[
		byte := self jpegDecodeValueFrom: acTable size: acTableSize.
		byte < 0 ifTrue:[^interpreterProxy primitiveFail].
		zeroCount := byte >> 4.
		byte := byte bitAnd: 16r0F.
		byte ~= 0 ifTrue:[
			index := index + zeroCount.
			bits :=  self getBits: byte.
			byte := self scaleAndSignExtend: bits inFieldWidth: byte.
			(index < 0 or:[index >= DCTSize2]) ifTrue:[^interpreterProxy primitiveFail].
			anArray at:	 (jpegNaturalOrder at: index) put: byte.
		] ifFalse:[
			zeroCount = 15 ifTrue: [index := index + zeroCount] ifFalse: [^ nil].
		].
		index := index + 1
	].! !

!JPEGReaderPlugin methodsFor: 'stream support' stamp: 'ar 3/4/2001 20:44'!
fillBuffer
	| byte |
	[jsBitCount <= 16] whileTrue:[
		jsPosition < jsReadLimit ifFalse:[^jsBitCount].
		byte := jsCollection at: jsPosition.
		jsPosition := jsPosition + 1.
		byte = 16rFF ifTrue:["peek for 00"
			((jsPosition < jsReadLimit) and:[(jsCollection at: jsPosition) = 16r00]) ifFalse:[
				jsPosition := jsPosition - 1.
				^jsBitCount].
			jsPosition := jsPosition + 1].
		jsBitBuffer := (jsBitBuffer bitShift: 8) bitOr: byte.
		jsBitCount := jsBitCount + 8].
	^jsBitCount! !

!JPEGReaderPlugin methodsFor: 'stream support' stamp: 'nice 12/23/2012 14:00'!
getBits: requestedBits
	| value |
	requestedBits > jsBitCount ifTrue:[
		self fillBuffer.
		requestedBits > jsBitCount ifTrue:[^-1]].
	jsBitCount := jsBitCount - requestedBits.
	value := jsBitBuffer >> jsBitCount.
	jsBitBuffer := jsBitBuffer bitAnd: (1 << jsBitCount) -1.
	^ value! !

!JPEGReaderPlugin methodsFor: 'decoding' stamp: 'nice 12/27/2012 03:03'!
idctBlockInt: anArray qt: qt
	| ws anACTerm dcval z2 z3 z1 t2 t3 t0 t1 t10 t13 t11 t12 z4 z5 v |
	<var: #anArray type: 'int *'>
	<var: #qt type: 'int *'>
	<var: #ws declareC: 'int ws[64]'>
	self cCode:'' inSmalltalk:[ws := CArrayAccessor on: (IntegerArray new: 64)].
	"Pass 1: process columns from anArray, store into work array"
	0 to: DCTSize-1 do:[:i |
		anACTerm := -1.
		1 to: DCTSize-1 do:[:row|
			anACTerm = -1 ifTrue:[
				(anArray at: row * DCTSize + i) = 0 ifFalse:[anACTerm := row]]].
		anACTerm = -1 ifTrue:[
			dcval := (anArray at: i) * (qt at: 0) << Pass1Bits.
			0 to: DCTSize-1 do: [:j | ws at: (j * DCTSize + i) put: dcval]
		] ifFalse:[
			z2 := (anArray at: (DCTSize * 2 + i)) * (qt at: (DCTSize * 2 + i)).
			z3 := (anArray at: (DCTSize * 6 + i)) * (qt at: (DCTSize * 6 + i)).
			z1 := (z2 + z3) * FIXn0n541196100.
			t2 := z1 + (z3 * (0 - FIXn1n847759065)).
			t3 := z1 + (z2 * FIXn0n765366865).
			z2 := (anArray at: i) * (qt at: i).
			z3 := (anArray at: (DCTSize * 4 + i)) * (qt at: (DCTSize * 4 + i)).
			t0 := (z2 + z3) << ConstBits.
			t1 := (z2 - z3) << ConstBits.
			t10 := t0 + t3.
			t13 := t0 - t3.
			t11 := t1 + t2.
			t12 := t1 - t2.
			t0 := (anArray at: (DCTSize * 7 + i)) * (qt at: (DCTSize * 7 + i)).
			t1 := (anArray at: (DCTSize * 5 + i)) * (qt at: (DCTSize * 5 + i)).
			t2 := (anArray at: (DCTSize * 3 + i)) * (qt at: (DCTSize * 3 + i)).
			t3 := (anArray at: (DCTSize + i)) * (qt at: (DCTSize + i)).
			z1 := t0 + t3.
			z2 := t1 + t2.
			z3 := t0 + t2.
			z4 := t1 + t3.
			z5 := (z3 + z4) * FIXn1n175875602.
			t0 := t0 * FIXn0n298631336.
			t1 := t1 * FIXn2n053119869.
			t2 := t2 * FIXn3n072711026.
			t3 := t3 * FIXn1n501321110.
			z1 := z1 * (0 - FIXn0n899976223).
			z2 := z2 * (0 - FIXn2n562915447).
			z3 := z3 * (0 - FIXn1n961570560).
			z4 := z4 * (0 - FIXn0n390180644).
			z3 := z3 + z5.
			z4 := z4 + z5.
			t0 := t0 + z1 + z3.
			t1 := t1 +z2 +z4.
			t2 := t2 + z2 + z3.
			t3 := t3 + z1 + z4.
			ws at: i put: (t10 + t3) // Pass1Div.
			ws at: (DCTSize * 7 + i) put: (t10 - t3) // Pass1Div.
			ws at: (DCTSize * 1 + i) put: (t11 + t2) // Pass1Div.
			ws at: (DCTSize * 6 + i) put: (t11 - t2) // Pass1Div.
			ws at: (DCTSize * 2 + i) put: (t12 + t1) // Pass1Div.
			ws at: (DCTSize * 5 + i) put: (t12 - t1) // Pass1Div.
			ws at: (DCTSize * 3 + i) put: (t13 + t0) // Pass1Div.
			ws at: (DCTSize * 4 + i) put: (t13 - t0) // Pass1Div]].

	"Pass 2: process rows from work array, store back into anArray"
	0 to: DCTSize2-DCTSize by: DCTSize do:[:i |
		z2 := ws at: i + 2.
		z3 := ws at: i + 6.
		z1 := (z2 + z3) * FIXn0n541196100.
		t2 := z1 + (z3 * (0-FIXn1n847759065)).
		t3 := z1 + (z2 * FIXn0n765366865).
		t0 := (ws at: i) + (ws at: (i + 4)) << ConstBits.
		t1 := (ws at: i) - (ws at: (i + 4)) << ConstBits.
		t10 := t0 + t3.
		t13 := t0 - t3.
		t11 := t1 + t2.
		t12 := t1 -t2.
		t0 := ws at: (i + 7).
		t1 := ws at: (i + 5).
		t2 := ws at: (i + 3).
		t3 := ws at: (i + 1).
		z1 := t0 + t3.
		z2 := t1 + t2.
		z3 := t0 + t2.
		z4 := t1 + t3.
		z5 := (z3 + z4) * FIXn1n175875602.
		t0 := t0 * FIXn0n298631336.
		t1 := t1 * FIXn2n053119869.
		t2 := t2 * FIXn3n072711026.
		t3 := t3 * FIXn1n501321110.
		z1 := z1 * (0-FIXn0n899976223).
		z2 := z2 * (0-FIXn2n562915447).
		z3 := z3 * (0-FIXn1n961570560).
		z4 := z4 * (0-FIXn0n390180644).
		z3 := z3 + z5.
		z4 := z4 + z5.
		t0 := t0 + z1 + z3.
		t1 := t1 + z2 + z4.
		t2 := t2 + z2 + z3.
		t3 := t3 + z1 + z4.
		v := (t10 + t3) // Pass2Div + SampleOffset.
		v := v min: MaxSample. v := v max: 0.
		anArray at: i put: v.
		v := (t10 - t3) // Pass2Div + SampleOffset.
		v := v min: MaxSample. v := v max: 0.
		anArray at: (i + 7) put: v.
		v := (t11 + t2) // Pass2Div + SampleOffset. 
		v := v min: MaxSample. v := v max: 0.
		anArray at: (i + 1) put: v.
		v := (t11 - t2) // Pass2Div + SampleOffset.
		v := v min: MaxSample. v := v max: 0.
		anArray at: (i + 6) put: v.
		v :=  (t12 + t1) // Pass2Div + SampleOffset.
		v := v min: MaxSample. v := v max: 0.
		anArray at: (i + 2) put: v.
		v :=  (t12 - t1) // Pass2Div + SampleOffset.
		v := v min: MaxSample. v := v max: 0.
		anArray at: (i + 5) put: v.
		v := (t13 + t0) // Pass2Div + SampleOffset.
		v := v min: MaxSample. v := v max: 0.
		anArray at: (i + 3) put: v.
		v := (t13 - t0) // Pass2Div + SampleOffset.
		v := v min: MaxSample. v := v max: 0.
		anArray at: (i + 4) put: v].! !

!JPEGReaderPlugin methodsFor: 'stream support' stamp: 'tpr (auto pragmas dtl 2010-09-28) 12/29/2005 16:36'!
jpegDecodeValueFrom: table size: tableSize
	"Decode the next value in the receiver using the given huffman table."
	| bits bitsNeeded tableIndex value index |
	<var: #table type: 'int *'>
	bitsNeeded := (table at: 0) >> 24.	"Initial bits needed"
	bitsNeeded > MaxBits ifTrue:[^-1].
	tableIndex := 2.							"First real table"
	[true] whileTrue:[
		bits := self getBits: bitsNeeded.		"Get bits"
		bits < 0 ifTrue:[^-1].
		index := tableIndex + bits - 1.
		index >= tableSize ifTrue:[^-1].
		value := table at: index.					"Lookup entry in table"
		(value bitAnd: 16r3F000000) = 0 ifTrue:[^value]. "Check if it is a leaf node"
		"Fetch sub table"
		tableIndex := value bitAnd: 16rFFFF.	"Table offset in low 16 bit"
		bitsNeeded := (value >> 24) bitAnd: 255. "Additional bits in high 8 bit"
		bitsNeeded > MaxBits ifTrue:[^-1]].
	^-1! !

!JPEGReaderPlugin methodsFor: 'stream support' stamp: 'eem 1/30/2014 15:33'!
loadJPEGStreamFrom: streamOop
	| oop sz |
	(interpreterProxy isPointers: streamOop) ifFalse:[^false].
	(interpreterProxy slotSizeOf: streamOop) < 5 ifTrue:[^false].
	oop := interpreterProxy fetchPointer: 0 ofObject: streamOop.
	(interpreterProxy isBytes: oop) ifFalse:[^false].
	jsCollection := interpreterProxy firstIndexableField: oop.
	sz := interpreterProxy byteSizeOf: oop.
	jsPosition := interpreterProxy fetchInteger: 1 ofObject: streamOop.
	jsReadLimit := interpreterProxy fetchInteger: 2 ofObject: streamOop.
	jsBitBuffer := interpreterProxy fetchInteger: 3 ofObject: streamOop.
	jsBitCount := interpreterProxy fetchInteger: 4 ofObject: streamOop.
	interpreterProxy failed ifTrue:[^false].
	sz < jsReadLimit ifTrue:[^false].
	(jsPosition < 0 or:[jsPosition >= jsReadLimit]) ifTrue:[^false].
	^true! !

!JPEGReaderPlugin methodsFor: 'decoding' stamp: 'ar (auto pragmas 12/08) 3/4/2001 21:11'!
nextSampleCb
	| dx dy blockIndex sampleIndex sample curX sx sy |
	<inline: true>
	dx := curX := cbComponent at: CurrentXIndex.
	dy := cbComponent at: CurrentYIndex.
	sx := cbComponent at: HScaleIndex.
	sy := cbComponent at: VScaleIndex.
	(sx = 0 and:[sy = 0]) ifFalse:[
		dx := dx // sx.
		dy := dy // sy.
	].
	blockIndex := (dy bitShift: -3) * (cbComponent at: BlockWidthIndex) + (dx bitShift: -3).
	sampleIndex := ((dy bitAnd: 7) bitShift: 3) + (dx bitAnd: 7).
	sample := (cbBlocks at: blockIndex) at: sampleIndex.
	curX := curX + 1.
	curX < ((cbComponent at: MCUWidthIndex) * 8) ifTrue:[
		cbComponent at: CurrentXIndex put: curX.
	] ifFalse:[
		cbComponent at: CurrentXIndex put: 0.
		cbComponent at: CurrentYIndex put: (cbComponent at: CurrentYIndex) + 1.
	].
	^ sample! !

!JPEGReaderPlugin methodsFor: 'decoding' stamp: 'ar (auto pragmas 12/08) 3/4/2001 21:11'!
nextSampleCr
	| dx dy blockIndex sampleIndex sample curX sx sy |
	<inline: true>
	dx := curX := crComponent at: CurrentXIndex.
	dy := crComponent at: CurrentYIndex.
	sx := crComponent at: HScaleIndex.
	sy := crComponent at: VScaleIndex.
	(sx = 0 and:[sy = 0]) ifFalse:[
		dx := dx // sx.
		dy := dy // sy.
	].
	blockIndex := (dy bitShift: -3) * (crComponent at: BlockWidthIndex) + (dx bitShift: -3).
	sampleIndex := ((dy bitAnd: 7) bitShift: 3) + (dx bitAnd: 7).
	sample := (crBlocks at: blockIndex) at: sampleIndex.
	curX := curX + 1.
	curX < ((crComponent at: MCUWidthIndex) * 8) ifTrue:[
		crComponent at: CurrentXIndex put: curX.
	] ifFalse:[
		crComponent at: CurrentXIndex put: 0.
		crComponent at: CurrentYIndex put: (crComponent at: CurrentYIndex) + 1.
	].
	^ sample! !

!JPEGReaderPlugin methodsFor: 'decoding' stamp: 'ar (auto pragmas 12/08) 3/4/2001 21:16'!
nextSampleFrom: aComponent blocks: aBlockArray
	| dx dy blockIndex sampleIndex sample curX sx sy |
	<var: #aComponent type: 'int *'>
	<var: #aBlockArray type: 'int **'>
	<inline: true>
	dx := curX := aComponent at: CurrentXIndex.
	dy := aComponent at: CurrentYIndex.
	sx := aComponent at: HScaleIndex.
	sy := aComponent at: VScaleIndex.
	(sx = 0 and:[sy = 0]) ifFalse:[
		dx := dx // sx.
		dy := dy // sy.
	].
	blockIndex := (dy bitShift: -3) * (aComponent at: BlockWidthIndex) + (dx bitShift: -3).
	sampleIndex := ((dy bitAnd: 7) bitShift: 3) + (dx bitAnd: 7).
	sample := (aBlockArray at: blockIndex) at: sampleIndex.
	curX := curX + 1.
	curX < ((aComponent at: MCUWidthIndex) * 8) ifTrue:[
		aComponent at: CurrentXIndex put: curX.
	] ifFalse:[
		aComponent at: CurrentXIndex put: 0.
		aComponent at: CurrentYIndex put: (aComponent at: CurrentYIndex) + 1.
	].
	^ sample! !

!JPEGReaderPlugin methodsFor: 'decoding' stamp: 'ar (auto pragmas 12/08) 3/4/2001 21:10'!
nextSampleY
	| dx dy blockIndex sampleIndex sample curX sx sy |
	<inline: true>
	dx := curX := yComponent at: CurrentXIndex.
	dy := yComponent at: CurrentYIndex.
	sx := yComponent at: HScaleIndex.
	sy := yComponent at: VScaleIndex.
	(sx = 0 and:[sy = 0]) ifFalse:[
		dx := dx // sx.
		dy := dy // sy.
	].
	blockIndex := (dy bitShift: -3) * (yComponent at: BlockWidthIndex) + (dx bitShift: -3).
	sampleIndex := ((dy bitAnd: 7) bitShift: 3) + (dx bitAnd: 7).
	sample := (yBlocks at: blockIndex) at: sampleIndex.
	curX := curX + 1.
	curX < ((yComponent at: MCUWidthIndex) * 8) ifTrue:[
		yComponent at: CurrentXIndex put: curX.
	] ifFalse:[
		yComponent at: CurrentXIndex put: 0.
		yComponent at: CurrentYIndex put: (yComponent at: CurrentYIndex) + 1.
	].
	^ sample! !

!JPEGReaderPlugin methodsFor: 'primitives' stamp: 'eem 6/28/2016 13:19'!
primitiveColorConvertGrayscaleMCU
	"Requires:
		JPEGColorComponent
		bits
		WordArray with: 3*Integer (residuals)
		ditherMask
	"
	| arrayOop |
	<export: true>
	self stInit.
	interpreterProxy methodArgumentCount = 4
		ifFalse:[^interpreterProxy primitiveFail].
	ditherMask := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	arrayOop := interpreterProxy stackValue: 1.
	((interpreterProxy isWords: arrayOop) and:[(interpreterProxy slotSizeOf: arrayOop) = 3])
		ifFalse:[^interpreterProxy primitiveFail].
	residuals := interpreterProxy firstIndexableField: arrayOop.
	arrayOop := interpreterProxy stackValue: 2.
	(interpreterProxy isWords: arrayOop)
		ifFalse:[^interpreterProxy primitiveFail].
	jpegBitsSize := interpreterProxy slotSizeOf: arrayOop.
	jpegBits := interpreterProxy firstIndexableField: arrayOop.
	arrayOop := interpreterProxy stackValue: 3.
	(self yColorComponentFrom: arrayOop)
		ifFalse:[^interpreterProxy primitiveFail].
	self colorConvertGrayscaleMCU.
	interpreterProxy pop: 4! !

!JPEGReaderPlugin methodsFor: 'primitives' stamp: 'eem 6/28/2016 13:18'!
primitiveColorConvertMCU
	"Requires:
		Array with: 3*JPEGColorComponent
		bits
		WordArray with: 3*Integer (residuals)
		ditherMask
	"
	| arrayOop |
	<export: true>
	self stInit.
	interpreterProxy methodArgumentCount = 4
		ifFalse:[^interpreterProxy primitiveFail].
	ditherMask := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	arrayOop := interpreterProxy stackValue: 1.
	((interpreterProxy isWords: arrayOop) and:[(interpreterProxy slotSizeOf: arrayOop) = 3])
		ifFalse:[^interpreterProxy primitiveFail].
	residuals := interpreterProxy firstIndexableField: arrayOop.
	arrayOop := interpreterProxy stackValue: 2.
	(interpreterProxy isWords: arrayOop)
		ifFalse:[^interpreterProxy primitiveFail].
	jpegBitsSize := interpreterProxy slotSizeOf: arrayOop.
	jpegBits := interpreterProxy firstIndexableField: arrayOop.
	arrayOop := interpreterProxy stackValue: 3.
	((interpreterProxy isPointers: arrayOop) and:[(interpreterProxy slotSizeOf: arrayOop) = 3])
		ifFalse:[^interpreterProxy primitiveFail].
	(self yColorComponentFrom: (interpreterProxy fetchPointer: 0 ofObject: arrayOop))
		ifFalse:[^interpreterProxy primitiveFail].
	(self cbColorComponentFrom: (interpreterProxy fetchPointer: 1 ofObject: arrayOop))
		ifFalse:[^interpreterProxy primitiveFail].
	(self crColorComponentFrom: (interpreterProxy fetchPointer: 2 ofObject: arrayOop))
		ifFalse:[^interpreterProxy primitiveFail].
	self colorConvertMCU.
	interpreterProxy pop: 4! !

!JPEGReaderPlugin methodsFor: 'primitives' stamp: 'eem 6/28/2016 13:16'!
primitiveDecodeMCU
	"In:
		anArray 		WordArray of: DCTSize2
		aColorComponent JPEGColorComponent
		dcTable			WordArray
		acTable			WordArray
		stream			JPEGStream
	"
	| arrayOop oop anArray |
	<export: true>
	<var: #anArray type: #'int *'>
	self cCode:'' inSmalltalk:[self stInit].

	interpreterProxy methodArgumentCount = 5 
		ifFalse:[^interpreterProxy primitiveFail].

	oop := interpreterProxy stackValue: 0.
	(self loadJPEGStreamFrom: oop)
		ifFalse:[^interpreterProxy primitiveFail].

	arrayOop := interpreterProxy stackValue: 1.
	(interpreterProxy isWords: arrayOop)
		ifFalse:[^interpreterProxy primitiveFail].
	acTableSize := interpreterProxy slotSizeOf: arrayOop.
	acTable := interpreterProxy firstIndexableField: arrayOop.

	arrayOop := interpreterProxy stackValue: 2.
	(interpreterProxy isWords: arrayOop)
		ifFalse:[^interpreterProxy primitiveFail].
	dcTableSize := interpreterProxy slotSizeOf: arrayOop.
	dcTable := interpreterProxy firstIndexableField: arrayOop.

	oop := interpreterProxy stackValue: 3.
	(self colorComponent: yComponent from: oop)
		ifFalse:[^interpreterProxy primitiveFail].

	arrayOop := interpreterProxy stackValue: 4.
	((interpreterProxy isWords: arrayOop)
	 and: [(interpreterProxy slotSizeOf: arrayOop) = DCTSize2])
		ifFalse:[^interpreterProxy primitiveFail].
	anArray := interpreterProxy firstIndexableField: arrayOop.

	interpreterProxy failed ifTrue:[^nil].

	self decodeBlockInto: anArray component: yComponent.

	interpreterProxy failed ifTrue:[^nil].
	self storeJPEGStreamOn: (interpreterProxy stackValue: 0).
	interpreterProxy 
		storeInteger: PriorDCValueIndex 
		ofObject: (interpreterProxy stackValue: 3) 
		withValue: (yComponent at: PriorDCValueIndex).

	interpreterProxy pop: 5! !

!JPEGReaderPlugin methodsFor: 'primitives' stamp: 'eem 6/28/2016 13:20'!
primitiveIdctInt
	"In:
		anArray: IntegerArray new: DCTSize2
		qt: IntegerArray new: DCTSize2.
	"
	| arrayOop anArray qt |
	<export: true>
	<var: #anArray type: #'int *'>
	<var: #qt type: #'int *'>
	self cCode:'' inSmalltalk:[self stInit].
	interpreterProxy methodArgumentCount = 2
		ifFalse:[^interpreterProxy primitiveFail].
	arrayOop := interpreterProxy stackValue: 0.
	((interpreterProxy isWords: arrayOop) and:[(interpreterProxy slotSizeOf: arrayOop) = DCTSize2])
		ifFalse:[^interpreterProxy primitiveFail].
	qt := interpreterProxy firstIndexableField: arrayOop.
	arrayOop := interpreterProxy stackValue: 1.
	((interpreterProxy isWords: arrayOop) and:[(interpreterProxy slotSizeOf: arrayOop) = DCTSize2])
		ifFalse:[^interpreterProxy primitiveFail].
	anArray := interpreterProxy firstIndexableField: arrayOop.
	self idctBlockInt: anArray qt: qt.
	interpreterProxy pop: 2! !

!JPEGReaderPlugin methodsFor: 'decoding' stamp: 'nice 12/22/2012 23:32'!
scaleAndSignExtend: aNumber inFieldWidth: w
	<inline: true>
	aNumber < (1 << (w - 1))
		ifTrue: [^aNumber - (1 << w) + 1]
		ifFalse: [^aNumber]! !

!JPEGReaderPlugin methodsFor: 'initialize' stamp: 'ar 3/4/2001 21:10'!
stInit
	self cCode:'' inSmalltalk:[
		yComponent := CArrayAccessor on: (IntegerArray new: MinComponentSize).
		cbComponent := CArrayAccessor on: (IntegerArray new: MinComponentSize).
		crComponent := CArrayAccessor on: (IntegerArray new: MinComponentSize).
		yBlocks := CArrayAccessor on: (Array new: MaxMCUBlocks).
		crBlocks := CArrayAccessor on: (Array new: MaxMCUBlocks).
		cbBlocks := CArrayAccessor on: (Array new: MaxMCUBlocks).
		jpegNaturalOrder := CArrayAccessor on: #(
			0 1 8 16 9 2 3 10 
			17 24 32 25 18 11 4 5 
			12 19 26 33 40 48 41 34 
			27 20 13 6 7 14 21 28 
			35 42 49 56 57 50 43 36 
			29 22 15 23 30 37 44 51 
			58 59 52 45 38 31 39 46 
			53 60 61 54 47 55 62 63).
	].! !

!JPEGReaderPlugin methodsFor: 'stream support' stamp: 'ar 3/4/2001 19:04'!
storeJPEGStreamOn: streamOop
	interpreterProxy storeInteger: 1 ofObject: streamOop withValue: jsPosition.
	interpreterProxy storeInteger: 3 ofObject: streamOop withValue: jsBitBuffer.
	interpreterProxy storeInteger: 4 ofObject: streamOop withValue: jsBitCount.! !

!JPEGReaderPlugin methodsFor: 'decoding' stamp: 'ar 3/4/2001 21:10'!
yColorComponentFrom: oop
	^(self colorComponent: yComponent from: oop)
		and:[self colorComponentBlocks: yBlocks from: oop]! !

InterpreterPlugin subclass: #KlattSynthesizerPlugin
	instanceVariableNames: 'resonators frame pitch t0 nper nopen nmod a1 a2 x1 x2 b1 c1 glast vlast nlast periodCount samplesCount seed cascade samplesPerFrame samplingRate'
	classVariableNames: 'A1v A2f A2v A3f A3v A4f A4v A5f A6f Anv Aspiration Atv B1 B2 B2f B3 B3f B4 B4f B5 B5f B6 B6f Bnp Bnz Btp Btz Bypass Db1 Df1 Diplophonia Epsilon F0 F1 F2 F3 F4 F5 F6 Flutter Fnp Fnz Friction Ftp Ftz Gain Jitter PI Ra Rk Ro Shimmer Turbulence Voicing'
	poolDictionaries: 'KlattResonatorIndices'
	category: 'VMMaker-Plugins'!

!KlattSynthesizerPlugin commentStamp: '<historical>' prior: 0!
This is a pluggable primitive implementation of the KlattSynthesizer.!

!KlattSynthesizerPlugin class methodsFor: 'translation' stamp: 'eem 3/5/2009 13:18'!
declareCVarsIn: cg 
	cg var: #resonators type: #'float *'.
	cg var: #frame type: #'float *'.
	cg var: #pitch type: #float.
	cg var: #a1 type: #float.
	cg var: #a2 type: #float.
	cg var: #x1 type: #float.
	cg var: #x2 type: #float.
	cg var: #b1 type: #float.
	cg var: #c1 type: #float.
	cg var: #glast type: #float.
	cg var: #vlast type: #float.
	cg var: #nlast type: #float! !

!KlattSynthesizerPlugin class methodsFor: 'class initialization' stamp: 'eem 1/19/2021 20:44'!
initialize
	"
	KlattSynthesizerPlugin initialize
	"
	| parameterClass |
	parameterClass := (Smalltalk hasClassNamed: #KlattFrame)
		ifTrue: [Smalltalk at: #KlattFrame "original package structure"]
		ifFalse: [Smalltalk at: #KlattResonatorIndices "repackaged to separate plugin from Speech dependency"].
	
	parameterClass parameterNames withIndexDo:
		[ :each :i | self classPool at: each capitalized asSymbol put: i - 1].
	PI := Float pi.
	Epsilon := 1.0e-04! !

!KlattSynthesizerPlugin class methodsFor: 'translation' stamp: 'len 9/25/1999 23:14'!
moduleName
	"
	KlattSynthesizerPlugin translateDoInlining: true
	"
	^ 'Klatt'! !

!KlattSynthesizerPlugin methodsFor: 'processing-voice' stamp: 'len (auto pragmas 12/08) 12/17/1999 03:16'!
addAmplitudeDiplophonia
	"Add diplophonia (bicyclic voice). Change voicing amplitude."
	<returnTypeC: 'void'>
	periodCount \\ 2 = 0
		ifFalse: [x1 := x1 * (1.0 - (frame at: Diplophonia)).
				"x1 must be <= 0"
				x1 > 0 ifTrue: [x1 := 0]]! !

!KlattSynthesizerPlugin methodsFor: 'processing-voice' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:49'!
addFlutter
	"Add F0 flutter, as specified in:
		'Analysis, synthesis and perception of voice quality variations among
		female and male talkers' D.H. Klatt and L.C. Klatt JASA 87(2) February 1990.
	Flutter is added by applying a quasi-random element constructed from three
	slowly varying sine waves."
	| timeCount asin bsin csin deltaF0 |
	<returnTypeC: 'void'>
	<var: 'timeCount' type: 'float '>
	<var: 'asin' type: 'float '>
	<var: 'bsin' type: 'float '>
	<var: 'csin' type: 'float '>
	<var: 'deltaF0' type: 'double '>
	timeCount := (self cCoerce: samplesCount to: 'float') / (self cCoerce: samplingRate to: 'float').
	asin := (2.0 * PI * 12.7 * timeCount) sin.
	bsin := (2.0 * PI * 7.1 * timeCount) sin.
	csin := (2.0 * PI * 4.7 * timeCount) sin.
	deltaF0 := (frame at: Flutter) * 2.0 * (frame at: F0) / 100.0 * (asin + bsin + csin).
	pitch := pitch + deltaF0! !

!KlattSynthesizerPlugin methodsFor: 'processing-voice' stamp: 'len (auto pragmas 12/08) 12/17/1999 03:11'!
addFrequencyDiplophonia
	"Add diplophonia (bicyclic voice). Change F0."
	<returnTypeC: 'void'>
	periodCount \\ 2 = 0
		ifTrue: [pitch := pitch + ((frame at: Diplophonia) * (frame at: F0) * (1.0 - (frame at: Ro)))]
		ifFalse: [pitch := pitch - ((frame at: Diplophonia) * (frame at: F0) * (1.0 - (frame at: Ro)))]! !

!KlattSynthesizerPlugin methodsFor: 'processing-voice' stamp: 'len (auto pragmas 12/08) 11/22/1999 02:44'!
addJitter
	"Add jitter (random F0 perturbation)."
	<returnTypeC: 'void'>
	pitch := pitch + (self nextRandom - 32767 * (frame at: Jitter) / 32768.0 * (frame at: F0))! !

!KlattSynthesizerPlugin methodsFor: 'processing-voice' stamp: 'len (auto pragmas 12/08) 11/22/1999 02:45'!
addShimmer
	"Add shimmer (random voicing amplitude perturbation)."
	<returnTypeC: 'void'>
	x1 := x1 + (self nextRandom - 32767 * (frame at: Shimmer) / 32768.0 * x1).
	"x1 must be <= 0"
	x1 > 0 ifTrue: [x1 := 0]! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:50'!
antiResonator: index frequency: freq bandwidth: bw
	"Set up an anti-resonator"
	| arg r a b c |
	<returnTypeC: 'void'>
	<var: 'freq' type: 'float '>
	<var: 'bw' type: 'float '>
	<var: 'arg' type: 'double '>
	<var: 'a' type: 'float '>
	<var: 'b' type: 'float '>
	<var: 'c' type: 'float '>
	<var: 'r' type: 'float '>
	arg := 0.0 - PI / samplingRate * bw.
	r := arg exp.
	c := 0.0 - (r * r).
	arg := PI * 2.0 / samplingRate * freq.
	b := r * arg cos * 2.0.
	a := 1.0 - b - c.
	a := 1.0 / a.
	b := 0.0 - b * a.
	c := 0.0 - c * a.
	self resonatorA: index put: a.
	self resonatorB: index put: b.
	self resonatorC: index put: c! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:50'!
antiResonator: index value: aFloat
	| answer p1 |
	<inline: true>
	<returnTypeC: 'float'>
	<var: 'aFloat' type: 'double '>
	<var: 'answer' type: 'double '>
	<var: 'p1' type: 'double '>
	answer := (self resonatorA: index) * aFloat
			+ ((self resonatorB: index) * (p1 := self resonatorP1: index))
			+ ((self resonatorC: index) * (self resonatorP2: index)).
	self resonatorP2: index put: p1.
	self resonatorP1: index put: aFloat.
	^ answer! !

!KlattSynthesizerPlugin methodsFor: 'processing' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:50'!
cascadeBranch: source
	"Cascade vocal tract, excited by laryngeal sources.
	Nasal antiresonator, nasal resonator, tracheal antirresonator,
	tracheal resonator, then formants F8, F7, F6, F5, F4, F3, F2, F1."
	| out |
	<inline: true>
	<returnTypeC: 'float'>
	<var: 'source' type: 'float '>
	<var: 'out' type: 'float '>
	cascade > 0 ifFalse: [^ 0.0].
	out := self antiResonator: Rnz value: source.
	out := self resonator: Rnpc value: out.
	out := self antiResonator: Rtz value: out.
	out := self resonator: Rtpc value: out.
	"Do not use unless sample rate >= 16000"
	cascade >= 8 ifTrue: [out := self resonator: R8c value: out].
	"Do not use unless sample rate >= 16000"
	cascade >= 7 ifTrue: [out := self resonator: R7c value: out].
	"Do not use unless long vocal tract or sample rate increased"
	cascade >= 6 ifTrue: [out := self resonator: R6c value: out].
	cascade >= 5 ifTrue: [out := self resonator: R5c value: out].
	cascade >= 4 ifTrue: [out := self resonator: R4c value: out].
	cascade >= 3 ifTrue: [out := self resonator: R3c value: out].
	cascade >= 2 ifTrue: [out := self resonator: R2c value: out].
	cascade >= 1 ifTrue: [out := self resonator: R1c value: out].
	^ out! !

!KlattSynthesizerPlugin methodsFor: 'private' stamp: 'hmm (auto pragmas dtl 2010-09-28) 5/10/1999 06:49'!
checkedFloatPtrOf: oop
	"Return the first indexable word of oop which is assumed to be variableWordSubclass"
	<returnTypeC: 'float *'>
	interpreterProxy success: (interpreterProxy isWords: oop).
	interpreterProxy failed ifTrue:[^0].
	^self cCoerce: (interpreterProxy firstIndexableField: oop) to:'float *'! !

!KlattSynthesizerPlugin methodsFor: 'private' stamp: 'hmm (auto pragmas dtl 2010-09-28) 5/10/1999 06:50'!
checkedShortPtrOf: oop
	"Return the first indexable word of oop which is assumed to be variableWordSubclass"
	<returnTypeC: 'short *'>
	interpreterProxy success: (interpreterProxy isWords: oop).
	interpreterProxy failed ifTrue:[^0].
	^self cCoerce: (interpreterProxy firstIndexableField: oop) to:'short *'! !

!KlattSynthesizerPlugin methodsFor: 'processing-LF' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:50'!
glottalSource
	| x0 |
	<inline: true>
	<returnTypeC: 'float'>
	<var: 'x0' type: 'float '>
	t0 = 0 ifTrue: [^ 0].
	nper < nopen
		ifTrue: [x0 := a1 * x1 + (a2 * x2).
				x2 := x1.
				x1 := x0]
		ifFalse: [x0 := b1 * x1 - c1.
				x1 := x0].
	"Reset period when 'nper' reaches t0."
	nper >= t0 ifTrue: [nper := 0. self pitchSynchronousReset].
	nper := nper + 1.
	^ x0! !

!KlattSynthesizerPlugin methodsFor: 'processing' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:50'!
linearFromdB: aNumber
	<inline: true>
	<returnTypeC: 'float'>
	<var: 'aNumber' type: 'double '>
	^ (2.0 raisedTo: (aNumber-87.0/6.0)) * 32.767! !

!KlattSynthesizerPlugin methodsFor: 'private' stamp: 'len 12/17/1999 03:27'!
loadFrom: klattOop
	| oop |
	interpreterProxy success: (interpreterProxy slotSizeOf: klattOop) = 22.
	interpreterProxy failed ifTrue:[^ false].

	oop := interpreterProxy fetchPointer: 0 ofObject: klattOop.
	resonators := self checkedFloatPtrOf: oop.

	pitch := interpreterProxy fetchFloat: 2 ofObject: klattOop.
	t0 := interpreterProxy fetchInteger: 3 ofObject: klattOop.
	nper := interpreterProxy fetchInteger: 4 ofObject: klattOop.
	nopen := interpreterProxy fetchInteger: 5 ofObject: klattOop.
	nmod := interpreterProxy fetchInteger: 6 ofObject: klattOop.
	a1 := interpreterProxy fetchFloat: 7 ofObject: klattOop.
	a2 := interpreterProxy fetchFloat: 8 ofObject: klattOop.
	x1 := interpreterProxy fetchFloat: 9 ofObject: klattOop.
	x2 := interpreterProxy fetchFloat: 10 ofObject: klattOop.
	b1 := interpreterProxy fetchFloat: 11 ofObject: klattOop.
	c1 := interpreterProxy fetchFloat: 12 ofObject: klattOop.
	glast := interpreterProxy fetchFloat: 13 ofObject: klattOop.
	vlast := interpreterProxy fetchFloat: 14 ofObject: klattOop.
	nlast := interpreterProxy fetchFloat: 15 ofObject: klattOop.
	periodCount := interpreterProxy fetchInteger: 16 ofObject: klattOop.
	samplesCount := interpreterProxy fetchInteger: 17 ofObject: klattOop.
	seed := interpreterProxy fetchInteger: 18 ofObject: klattOop.
	cascade := interpreterProxy fetchInteger: 19 ofObject: klattOop.
	samplesPerFrame := interpreterProxy fetchInteger: 20 ofObject: klattOop.
	samplingRate := interpreterProxy fetchInteger: 21 ofObject: klattOop.

	^ interpreterProxy failed == false! !

!KlattSynthesizerPlugin methodsFor: 'processing' stamp: 'len (auto pragmas 12/08) 11/9/1999 22:40'!
nextRandom
	"Answer a random number between 0 and 65535."
	<inline: true>
	seed := (seed * 1309) + 13849 bitAnd: 65535.
	^ seed! !

!KlattSynthesizerPlugin methodsFor: 'processing-LF' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:51'!
normalizeGlottalPulse
	| s1 s2 s0 |
	<inline: true>
	<returnTypeC: 'void'>
	<var: 's0' type: 'float '>
	<var: 's1' type: 'float '>
	<var: 's2' type: 'float '>
	s0 := 0.0.
	s1 := x1.
	s2 := x2.
	1 to: nopen do: [ :ingore |
		s0 := a1 * s1 + (a2 * s2).
		s2 := s1.
		s1 := s0].
	s0 = 0.0 ifFalse: [x1 := x1 / s0 * 10000.0]! !

!KlattSynthesizerPlugin methodsFor: 'processing' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:51'!
parallelFrictionBranch: source
	"Friction-excited parallel vocal tract formants F6, F5, F4, F3, F2,
	outputs added with alternating sign. Sound source for other
	parallel resonators is friction plus first difference of
	voicing waveform."
	<inline: true>
	<returnTypeC: 'float'>
	<var: 'source' type: 'float '>
	^ (self resonator: R2fp value: source) - (self resonator: R3fp value: source) + (self resonator: R4fp value: source) - (self resonator: R5fp value: source) + (self resonator: R6fp value: source)! !

!KlattSynthesizerPlugin methodsFor: 'processing' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:51'!
parallelVoicedBranch: source
	"Voice-excited parallel vocal tract F1, F2, F3, F4, FNP and FTP."
	<inline: true>
	<returnTypeC: 'float'>
	<var: 'source' type: 'float '>
	^ (self resonator: R1vp value: source) + (self resonator: R2vp value: source) + (self resonator: R3vp value: source) + (self resonator: R4vp value: source) + (self resonator: Rnpp value: source) + (self resonator: Rtpp value: source)! !

!KlattSynthesizerPlugin methodsFor: 'processing-voice' stamp: 'len (auto pragmas 12/08) 12/21/1999 02:24'!
pitchSynchronousReset
	<returnTypeC: 'void'>
	(frame at: F0) > 0
		ifTrue: [self voicedPitchSynchronousReset.
				periodCount := periodCount + 1 \\ 65535]
		ifFalse: [t0 := 1.
				nmod := t0]! !

!KlattSynthesizerPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:51'!
primitiveSynthesizeFrameIntoStartingAt
	| aKlattFrame buffer startIndex rcvr bufferOop |
	<export: true>
	<var: 'aKlattFrame' type: 'float *'>
	<var: 'buffer' type: 'short *'>
	aKlattFrame := self checkedFloatPtrOf: (interpreterProxy stackValue: 2).
	buffer := self checkedShortPtrOf: (bufferOop := interpreterProxy stackValue: 1).
	startIndex := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue: [^nil].
	rcvr := interpreterProxy stackObjectValue: 3.
	(self loadFrom: rcvr) ifFalse:[^nil].
	interpreterProxy success: (interpreterProxy stSizeOf: bufferOop) * 2 >= samplesPerFrame.
	interpreterProxy failed ifTrue: [^nil].
	self synthesizeFrame: aKlattFrame into: buffer startingAt: startIndex.
	(self saveTo: rcvr) ifFalse: [^nil].
	interpreterProxy pop: 3! !

!KlattSynthesizerPlugin methodsFor: 'processing-LF' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:51'!
qu: u phi: phi cosphi: cosphi sinphi: sinphi rphid: rphid
	| expuphi |
	<returnTypeC: 'float'>
	<var: 'u' type: 'float '>
	<var: 'phi' type: 'float '>
	<var: 'cosphi' type: 'float '>
	<var: 'sinphi' type: 'float '>
	<var: 'rphid' type: 'float '>
	<var: 'expuphi' type: 'float '>
	expuphi := (u * phi) exp.
	^ expuphi * ((rphid * (u*u + 1.0) + u) * sinphi - cosphi) + 1.0! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:52'!
resonator: index frequency: freq bandwidth: bw
	"Convert formant frequencies and bandwidth into
	resonator difference equation coefficients."
	| arg r a b c |
	<returnTypeC: 'void'>
	<var: 'freq' type: 'float '>
	<var: 'bw' type: 'float '>
	<var: 'arg' type: 'double '>
	<var: 'a' type: 'float '>
	<var: 'b' type: 'float '>
	<var: 'c' type: 'float '>
	<var: 'r' type: 'float '>
	arg := 0.0 - PI / samplingRate * bw.
	r := arg exp.
	c := 0.0 - (r * r).
	arg := PI * 2.0 / samplingRate * freq.
	b := r * arg cos * 2.0.
	a := 1.0 - b - c.
	self resonatorA: index put: a.
	self resonatorB: index put: b.
	self resonatorC: index put: c! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:52'!
resonator: index frequency: freq bandwidth: bw gain: gain
	"Convert formant frequencies and bandwidth into
	resonator difference equation coefficients."
	<returnTypeC: 'void'>
	<var: 'freq' type: 'float '>
	<var: 'bw' type: 'float '>
	<var: 'gain' type: 'float '>
	self resonator: index frequency: freq bandwidth: bw.
	self resonatorA: index put: (self resonatorA: index) * gain! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:53'!
resonator: index value: aFloat
	| answer p1 |
	<inline: true>
	<returnTypeC: 'float'>
	<var: 'aFloat' type: 'float '>
	<var: 'answer' type: 'float '>
	<var: 'p1' type: 'float '>
	answer := (self resonatorA: index) * aFloat
			+ ((self resonatorB: index) * (p1 := self resonatorP1: index))
			+ ((self resonatorC: index) * (self resonatorP2: index)).
	"(p1 between: -100000 and: 100000) ifFalse: [self halt].
	(answer between: -100000 and: 100000) ifFalse: [self halt]."
	self resonatorP2: index put: p1.
	self resonatorP1: index put: answer.
	^ answer! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'len (auto pragmas 12/08) 9/6/1999 00:16'!
resonatorA: index
	<inline: true>
	<returnTypeC: 'float'>
	^resonators at: index*5-5! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:51'!
resonatorA: index put: aFloat
	<inline: true>
	<returnTypeC: 'void'>
	<var: 'aFloat' type: 'float '>
	resonators at: index*5-5 put: aFloat! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'len (auto pragmas 12/08) 9/6/1999 00:16'!
resonatorB: index
	<inline: true>
	<returnTypeC: 'float'>
	^resonators at: index*5-4! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:52'!
resonatorB: index put: aFloat
	<inline: true>
	<returnTypeC: 'void'>
	<var: 'aFloat' type: 'float '>
	resonators at: index*5-4 put: aFloat! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'len (auto pragmas 12/08) 9/6/1999 00:16'!
resonatorC: index
	<inline: true>
	<returnTypeC: 'float'>
	^resonators at: index*5-3! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:52'!
resonatorC: index put: aFloat
	<inline: true>
	<returnTypeC: 'void'>
	<var: 'aFloat' type: 'float '>
	resonators at: index*5-3 put: aFloat! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'len (auto pragmas 12/08) 9/6/1999 00:16'!
resonatorP1: index
	<inline: true>
	<returnTypeC: 'float'>
	^resonators at: index*5-2! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:52'!
resonatorP1: index put: aFloat
	<inline: true>
	<returnTypeC: 'void'>
	<var: 'aFloat' type: 'float '>
	resonators at: index*5-2 put: aFloat! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'len (auto pragmas 12/08) 9/6/1999 00:16'!
resonatorP2: index
	<inline: true>
	<returnTypeC: 'float'>
	^resonators at: index*5-1! !

!KlattSynthesizerPlugin methodsFor: 'resonators' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:52'!
resonatorP2: index put: aFloat
	<inline: true>
	<returnTypeC: 'void'>
	<var: 'aFloat' type: 'float '>
	resonators at: index*5-1 put: aFloat! !

!KlattSynthesizerPlugin methodsFor: 'processing-LF' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:53'!
ro: roNumber ra: raNumber rk: rkNumber
	| r d phi cosphi sinphi rphid u theta rho gamma gammapwr te ro ra rk |

	<returnTypeC: 'void'>
	<var: 'roNumber' type: 'float '>
	<var: 'raNumber' type: 'float '>
	<var: 'rkNumber' type: 'float '>
	<var: 'r' type: 'float '>
	<var: 'd' type: 'float '>
	<var: 'phi' type: 'float '>
	<var: 'cosphi' type: 'float '>
	<var: 'sinphi' type: 'float '>
	<var: 'rphid' type: 'float '>
	<var: 'u' type: 'float '>
	<var: 'theta' type: 'float '>
	<var: 'rho' type: 'float '>
	<var: 'gamma' type: 'float '>
	<var: 'gammapwr' type: 'float '>
	<var: 'ro' type: 'float '>
	<var: 'ra' type: 'float '>
	<var: 'rk' type: 'float '>

	te := (t0 * roNumber) asInteger.
	ro := te asFloat / t0 asFloat.
	rk := rkNumber.
	ra := raNumber.

	ra <= 0.0
		ifTrue: [d := 1.0]
		ifFalse: [r := 1.0 - ro / ra.
				d := 1.0 - (r / (r exp - 1.0))].

	phi := PI * (rk + 1.0).
	cosphi := phi cos.
	sinphi := phi sin.
	rphid := ra / ro * phi * d.

	u := self zeroQphi: phi cosphi: cosphi sinphi: sinphi rphid: rphid.
	theta := phi / te.
	rho := (u * theta) exp.
	a1 := 2.0 * theta cos * rho.
	a2 := 0.0 - (rho * rho).
	x2 := 0.0.
	x1 := rho * theta sin.

	gamma := (-1.0 / (ra * t0)) exp.
	gammapwr := gamma raisedTo: t0 - te.

	b1 := gamma.
	c1 := (1.0 - gamma) * gammapwr / (1.0 - gammapwr).

	self normalizeGlottalPulse! !

!KlattSynthesizerPlugin methodsFor: 'private' stamp: 'len 12/19/1999 05:25'!
saveTo: origKlattOop
	| pitchOop a1Oop a2Oop x1Oop x2Oop b1Oop c1Oop glastOop vlastOop nlastOop klattOop |
	interpreterProxy pushRemappableOop: origKlattOop.
	interpreterProxy pushRemappableOop: (interpreterProxy floatObjectOf: pitch).
	interpreterProxy pushRemappableOop: (interpreterProxy floatObjectOf: a1).
	interpreterProxy pushRemappableOop: (interpreterProxy floatObjectOf: a2).
	interpreterProxy pushRemappableOop: (interpreterProxy floatObjectOf: x1).
	interpreterProxy pushRemappableOop: (interpreterProxy floatObjectOf: x2).
	interpreterProxy pushRemappableOop: (interpreterProxy floatObjectOf: b1).
	interpreterProxy pushRemappableOop: (interpreterProxy floatObjectOf: c1).
	interpreterProxy pushRemappableOop: (interpreterProxy floatObjectOf: glast).
	interpreterProxy pushRemappableOop: (interpreterProxy floatObjectOf: vlast).
	nlastOop := interpreterProxy floatObjectOf: nlast.
	vlastOop := interpreterProxy popRemappableOop.
	glastOop := interpreterProxy popRemappableOop.
	c1Oop := interpreterProxy popRemappableOop.
	b1Oop := interpreterProxy popRemappableOop.
	x2Oop := interpreterProxy popRemappableOop.
	x1Oop := interpreterProxy popRemappableOop.
	a2Oop := interpreterProxy popRemappableOop.
	a1Oop := interpreterProxy popRemappableOop.
	pitchOop := interpreterProxy popRemappableOop.
	klattOop := interpreterProxy popRemappableOop.
	interpreterProxy failed ifTrue:[^ false].

	interpreterProxy storePointer: 2 ofObject: klattOop withValue: pitchOop.
	interpreterProxy storeInteger: 3 ofObject: klattOop withValue: t0.
	interpreterProxy storeInteger: 4 ofObject: klattOop withValue: nper.
	interpreterProxy storeInteger: 5 ofObject: klattOop withValue: nopen.
	interpreterProxy storeInteger: 6 ofObject: klattOop withValue: nmod.
	interpreterProxy storePointer: 7 ofObject: klattOop withValue: a1Oop.
	interpreterProxy storePointer: 8 ofObject: klattOop withValue: a2Oop.
	interpreterProxy storePointer: 9 ofObject: klattOop withValue: x1Oop.
	interpreterProxy storePointer: 10 ofObject: klattOop withValue: x2Oop.
	interpreterProxy storePointer: 11 ofObject: klattOop withValue: b1Oop.
	interpreterProxy storePointer: 12 ofObject: klattOop withValue: c1Oop.
	interpreterProxy storePointer: 13 ofObject: klattOop withValue: glastOop.
	interpreterProxy storePointer: 14 ofObject: klattOop withValue: vlastOop.
	interpreterProxy storePointer: 15 ofObject: klattOop withValue: nlastOop.
	interpreterProxy storeInteger: 16 ofObject: klattOop withValue: periodCount.
	interpreterProxy storeInteger: 17 ofObject: klattOop withValue: samplesCount.
	interpreterProxy storeInteger: 18 ofObject: klattOop withValue: seed.

	^ interpreterProxy failed == false! !

!KlattSynthesizerPlugin methodsFor: 'processing' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:54'!
setCurrentFrame: aKlattFrame
	| ampFNV ampFTV ampF1V ampF2V ampF3V ampF4V ampF2F ampF3F ampF4F ampF5F ampF6F |
	<returnTypeC: 'void'>
	<var: 'aKlattFrame' type: 'float *'>
	<var: 'ampFNV' type: 'float '>
	<var: 'ampFTV' type: 'float '>
	<var: 'ampF1V' type: 'float '>
	<var: 'ampF2V' type: 'float '>
	<var: 'ampF3V' type: 'float '>
	<var: 'ampF4V' type: 'float '>
	<var: 'ampF2F' type: 'float '>
	<var: 'ampF3F' type: 'float '>
	<var: 'ampF4F' type: 'float '>
	<var: 'ampF5F' type: 'float '>
	<var: 'ampF6F' type: 'float '>

	frame := aKlattFrame.

	"Fudge factors..."
	ampFNV := (self linearFromdB: (frame at: Anv)) * 0.6.	"-4.44 dB"
	ampFTV := (self linearFromdB: (frame at: Atv)) * 0.6.		"-4.44 dB"
	ampF1V := (self linearFromdB: (frame at: A1v)) * 0.4.		"-7.96 dB"
	ampF2V := (self linearFromdB: (frame at: A2v)) * 0.15.	"-16.5 dB"
	ampF3V := (self linearFromdB: (frame at: A3v)) * 0.06.	"-24.4 dB"
	ampF4V := (self linearFromdB: (frame at: A4v)) * 0.04.	"-28.0 dB"
	ampF2F := (self linearFromdB: (frame at: A2f)) * 0.15.		"-16.5 dB"
	ampF3F := (self linearFromdB: (frame at: A3f)) * 0.06.	"-24.4 dB"
	ampF4F := (self linearFromdB: (frame at: A4f)) * 0.04.	"-28.0 dB"
	ampF5F := (self linearFromdB: (frame at: A5f)) * 0.022.	"-33.2 dB"
	ampF6F := (self linearFromdB: (frame at: A6f)) * 0.03.	"-30.5 dB"

	"Set coefficients of variable cascade resonators"
	cascade >= 8
		ifTrue: [samplingRate >= 16000	"Inside Nyquist rate?"
					ifTrue: [self resonator: R8c frequency: 7500 bandwidth: 600]
					ifFalse: [cascade := 6]].
	cascade >= 7
		ifTrue: [samplingRate >= 16000	"Inside Nyquist rate?"
					ifTrue: [self resonator: R7c frequency: 6500 bandwidth: 500]
					ifFalse: [cascade := 6]].
	cascade >= 6 ifTrue: [self resonator: R6c frequency: (frame at: F6) bandwidth: (frame at: B6)].
	cascade >= 5 ifTrue: [self resonator: R5c frequency: (frame at: F5) bandwidth: (frame at: B5)].
	self resonator: R4c frequency: (frame at: F4) bandwidth: (frame at: B4).
	self resonator: R3c frequency: (frame at: F3) bandwidth: (frame at: B3).
	self resonator: R2c frequency: (frame at: F2) bandwidth: (frame at: B2).
	self resonator: R1c frequency: (frame at: F1) bandwidth: (frame at: B1).

	"Set coefficients of nasal and tracheal resonators and antiresonators"
	self resonator: Rnpc frequency: (frame at: Fnp) bandwidth: (frame at: Bnp).
	self resonator: Rtpc frequency: (frame at: Ftp) bandwidth: (frame at: Btp).
	self antiResonator: Rnz frequency: (frame at: Fnz) bandwidth: (frame at: Bnz).
	self antiResonator: Rtz frequency: (frame at: Ftz) bandwidth: (frame at: Btz).

	"Set coefficients of parallel resonators, and amplitude of outputs"
	self resonator: Rnpp frequency: (frame at: Fnp) bandwidth: (frame at: Bnp) gain: ampFNV.
	self resonator: Rtpp frequency: (frame at: Ftp) bandwidth: (frame at: Btp) gain: ampFTV.
	self resonator: R1vp frequency: (frame at: F1) bandwidth: (frame at: B1) gain: ampF1V.
	self resonator: R2vp frequency: (frame at: F2) bandwidth: (frame at: B2) gain: ampF2V.
	self resonator: R3vp frequency: (frame at: F3) bandwidth: (frame at: B3) gain: ampF3V.
	self resonator: R4vp frequency: (frame at: F4) bandwidth: (frame at: B4) gain: ampF4V.
	self resonator: R2fp frequency: (frame at: F2) bandwidth: (frame at: B2f) gain: ampF2F.
	self resonator: R3fp frequency: (frame at: F3) bandwidth: (frame at: B3f) gain: ampF3F.
	self resonator: R4fp frequency: (frame at: F4) bandwidth: (frame at: B4f) gain: ampF4F.
	self resonator: R5fp frequency: (frame at: F5) bandwidth: (frame at: B5f) gain: ampF5F.
	self resonator: R6fp frequency: (frame at: F6) bandwidth: (frame at: B6f) gain: ampF6F! !

!KlattSynthesizerPlugin methodsFor: 'processing' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:54'!
synthesizeFrame: aKlattFrame into: buffer startingAt: startIndex
	| noise voice frictionNoise aspirationNoise glotout
	parGlotout source temp out
	index top
	voicing parVoicing turbulence friction aspiration bypass
	gain ampGain |

	<returnTypeC: 'void'>
	<var: 'aKlattFrame' type: 'float *'>
	<var: 'buffer' type: 'short *'>
	<var: 'noise' type: 'float '>
	<var: 'voice' type: 'float '>
	<var: 'frictionNoise' type: 'float '>
	<var: 'aspirationNoise' type: 'float '>
	<var: 'voicing' type: 'float '>
	<var: 'parVoicing' type: 'float '>
	<var: 'turbulence' type: 'float '>
	<var: 'friction' type: 'float '>
	<var: 'aspiration' type: 'float '>
	<var: 'bypass' type: 'float '>
	<var: 'glotout' type: 'float '>
	<var: 'parGlotout' type: 'float '>
	<var: 'source' type: 'float '>
	<var: 'gain' type: 'float '>
	<var: 'ampGain' type: 'float '>
	<var: 'out' type: 'float '>

	self setCurrentFrame: aKlattFrame.

	pitch > 0
		ifTrue: [voicing := self linearFromdB: (frame at: Voicing) - 7.
				parVoicing := self linearFromdB: (frame at: Voicing).
				turbulence := (self linearFromdB: (frame at: Turbulence)) * 0.1]
		ifFalse: [voicing := parVoicing := turbulence := 0.0].

	friction := (self linearFromdB: (frame at: Friction)) * 0.25.
	aspiration := (self linearFromdB: (frame at: Aspiration)) * 0.05.
	bypass := (self linearFromdB: (frame at: Bypass)) * 0.05.		"-26.0 dB"

	"Flod overall gain into output resonator (low-pass filter)"
	gain := (frame at: Gain) - 3.
	gain <= 0 ifTrue: [gain := 57].
	ampGain := self linearFromdB: gain.
	self resonator: Rout frequency: 0 bandwidth: samplingRate gain: ampGain.

	noise := nlast.
	index := startIndex.
	top := samplesPerFrame + startIndex - 1.
	[index <= top] whileTrue: [
		"Get low-passed random number for aspiration and friction noise"
		noise := (self nextRandom - 32768) asFloat / 4.0. "radom number between -8196.0 and 8196.0"

		"Tilt down noise spectrum by soft low-pass filter having
		 a pole near the origin in the z-plane."
		noise := noise + (0.75 * nlast).
		nlast := noise.

		"Amplitude modulate noise (reduce noise amplitude during second
		 half of glottal period) if voicing  simultaneously present."
		nper > nmod ifTrue: [noise := noise * 0.5].

		"Compute frictation noise"
		frictionNoise := friction * noise.

		"Compute voicing waveform."
		voice := self glottalSource.
		vlast := voice.

		"Add turbulence during glottal open phase.
		 Use random rather than noise because noise is low-passed."
		nper < nopen ifTrue: [voice := voice + (turbulence * (self nextRandom - 32768) asFloat / 4.0)].

		"Set amplitude of voicing."
		glotout := voicing * voice.
		parGlotout := parVoicing * voice.

		"Compute aspiration amplitude and add to voicing source."
		aspirationNoise := aspiration * noise.
		glotout := glotout + aspirationNoise.
		parGlotout := parGlotout + aspirationNoise.

		"Cascade vocal tract, excited by laryngeal sources.
		 Nasal antiresonator, nasal resonator, trachearl antirresonator,
		 tracheal resonator, then formants F8, F7, F6, F5, F4, F3, F2, F1."
		out := self cascadeBranch: glotout.

		"Voice-excited parallel vocal tract F1, F2, F3, F4, FNP and FTP."
		source := parGlotout.	"Source is voicing plus aspiration."
		out := out + (self parallelVoicedBranch: source).

		"Friction-excited parallel vocal tract formants F6, F5, F4, F3, F2,
		 outputs added with alternating sign. Sound source for other
		 parallel resonators is friction plus first difference of
		 voicing waveform."
		source := frictionNoise + parGlotout - glast.
		glast := parGlotout.
		out := (self parallelFrictionBranch: source) - out.

		"Apply bypas and output low-pass filter"
		out := bypass * source - out.
		out := self resonator: Rout value: out.
		temp := (out * ampGain) asInteger.
		temp < -32768 ifTrue: [temp := -32768].
		temp > 32767 ifTrue: [temp := 32767].
		buffer at: index - 1 put: temp.
		index := index + 1.
		samplesCount := samplesCount + 1]! !

!KlattSynthesizerPlugin methodsFor: 'processing-voice' stamp: 'len (auto pragmas 12/08) 12/19/1999 04:52'!
voicedPitchSynchronousReset
	<returnTypeC: 'void'>

	"Set the pitch."
	pitch := frame at: F0.

	"Add flutter and jitter (F0 perturbations)."
	self addFlutter.
	self addJitter.
	self addFrequencyDiplophonia.
	pitch < 0 ifTrue: [pitch := 0].

	"Recompute t0 (it is the number of samples in one pitch period)."
	t0 := (samplingRate / pitch) asInteger.

	"Duration of period before amplitude modulation."
	nmod := t0.
	(frame at: Voicing) > 0 ifTrue: [nmod := nmod // 2].

	"Set open phase of glottal period."
	nopen := (t0 * (frame at: Ro)) asInteger.

	"Set the LF glottal pulse model parameters."
	self ro: (frame at: Ro) ra: (frame at: Ra) rk: (frame at: Rk).

	"Add shimmer and diplophonia amplitude pertirbations.
	(This must be done AFTER the actual computation of the LF parameters.)"
	self addShimmer.
	self addAmplitudeDiplophonia! !

!KlattSynthesizerPlugin methodsFor: 'processing-LF' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:55'!
zeroQphi: phi cosphi: cosphi sinphi: sinphi rphid: rphid
	| qzero ua ub qa qb uc qc |
	<returnTypeC: 'float'>
	<var: 'qzero' type: 'float '>
	<var: 'ua' type: 'float '>
	<var: 'ub' type: 'float '>
	<var: 'qa' type: 'float '>
	<var: 'qb' type: 'float '>
	<var: 'uc' type: 'float '>
	<var: 'qc' type: 'float '>
	<var: 'phi' type: 'float '>
	<var: 'cosphi' type: 'float '>
	<var: 'sinphi' type: 'float '>
	<var: 'rphid' type: 'float '>

	qzero := self qu: 0 phi: phi cosphi: cosphi sinphi: sinphi rphid: rphid.

	qzero > 0
		ifTrue: [ua := 0. ub := 1.
				qa := qzero. qb := self qu: ub phi: phi cosphi: cosphi sinphi: sinphi rphid: rphid.
				[qb > 0]
					whileTrue: [ua := ub. qa := qb.
								ub := ub * 2. qb := self qu: ub phi: phi cosphi: cosphi sinphi: sinphi rphid: rphid]]
		ifFalse: [ua := -1. ub := 0.
				qa := self qu: ua phi: phi cosphi: cosphi sinphi: sinphi rphid: rphid. qb := qzero.
				[qa < 0]
					whileTrue: [ub := ua. qb := qa.
								ua := ua * 2. qa := self qu: ua phi: phi cosphi: cosphi sinphi: sinphi rphid: rphid]].
	[ub - ua > Epsilon]
		whileTrue: [uc := ub + ua / 2.0. qc := self qu: uc phi: phi cosphi: cosphi sinphi: sinphi rphid: rphid.
					qc > 0 ifTrue: [ua := uc. qa := qc] ifFalse: [ub := uc. qb := qc]].
	^ ub + ua / 2.0! !

InterpreterPlugin subclass: #Matrix2x3Plugin
	instanceVariableNames: 'm23ResultX m23ResultY m23ArgX m23ArgY'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!Matrix2x3Plugin class methodsFor: 'class initialization' stamp: 'sma 3/3/2000 12:39'!
declareCVarsIn: cg 
	cg var: #m23ResultX type: #double.
	cg var: #m23ResultY type: #double.
	cg var: #m23ArgX type: #double.
	cg var: #m23ArgY type: #double! !

!Matrix2x3Plugin class methodsFor: 'simulation' stamp: 'eem 10/21/2018 12:32'!
simulatorClass
	^Matrix2x3PluginSimulator! !

!Matrix2x3Plugin methodsFor: 'private' stamp: 'eem 10/21/2018 11:10'!
loadArgumentMatrix: matrix
	"Load the argument matrix"
	<returnTypeC: #'float *'>
	<inline: #always>
	((interpreterProxy isWords: matrix) and: [(interpreterProxy slotSizeOf: matrix) = 6]) ifFalse:
		[interpreterProxy primitiveFail.
		 ^nil].
	^self cCoerce: (interpreterProxy firstIndexableField: matrix) to: #'float *'! !

!Matrix2x3Plugin methodsFor: 'private' stamp: 'eem 10/21/2018 12:12'!
loadArgumentPoint: point
	"Load the argument point into m23ArgX and m23ArgY"
	| oop |
	"Check class of point"
	(interpreterProxy fetchClassOf: point) = interpreterProxy classPoint ifFalse:
		[^interpreterProxy primitiveFail].
	"Load X value"
	oop := interpreterProxy fetchPointer: 0 ofObject: point.
	(interpreterProxy isIntegerObject: oop)
		ifTrue: [m23ArgX := interpreterProxy integerValueOf: oop]
		ifFalse:
			[(interpreterProxy isFloatObject: oop)
				ifTrue: [m23ArgX := interpreterProxy floatValueOf: oop]
				ifFalse: [interpreterProxy primitiveFail]].
	"Load Y value"
	oop := interpreterProxy fetchPointer: 1 ofObject: point.
	(interpreterProxy isIntegerObject: oop)
		ifTrue: [m23ArgY := interpreterProxy integerValueOf: oop]
		ifFalse:
			[(interpreterProxy isFloatObject: oop)
				ifTrue: [m23ArgY := interpreterProxy floatValueOf: oop]
				ifFalse: [interpreterProxy primitiveFail]]! !

!Matrix2x3Plugin methodsFor: 'transforming' stamp: 'eem 10/21/2018 12:49'!
matrix2x3ComposeMatrix: m1 with: m2 into: m3
	"Multiply matrix m1 with m2 and store the result into m3."
	| a11 a12 a13 a21 a22 a23 |
	<var: #m1 type: #'const float *'>
	<var: #m2 type: #'const float *'>
	<var: #m3 type: #'float *'>
	<var: #a11 type: #double>
	<var: #a12 type: #double>
	<var: #a13 type: #double>
	<var: #a21 type: #double>
	<var: #a22 type: #double>
	<var: #a23 type: #double>

	a11 := ((m1 at: 0) * (m2 at: 0)) + ((m1 at: 1) * (m2 at: 3)).
	a12 := ((m1 at: 0) * (m2 at: 1)) + ((m1 at: 1) * (m2 at: 4)).
	a13 := ((m1 at: 0) * (m2 at: 2)) + ((m1 at: 1) * (m2 at: 5)) + (m1 at: 2).
	a21 := ((m1 at: 3) * (m2 at: 0)) + ((m1 at: 4) * (m2 at: 3)).
	a22 := ((m1 at: 3) * (m2 at: 1)) + ((m1 at: 4) * (m2 at: 4)).
	a23 := ((m1 at: 3) * (m2 at: 2)) + ((m1 at: 4) * (m2 at: 5)) + (m1 at: 5).

	m3 at: 0 put: (self cCoerce: a11 to: #float).
	m3 at: 1 put: (self cCoerce: a12 to: #float).
	m3 at: 2 put: (self cCoerce: a13 to: #float).
	m3 at: 3 put: (self cCoerce: a21 to: #float).
	m3 at: 4 put: (self cCoerce: a22 to: #float).
	m3 at: 5 put: (self cCoerce: a23 to: #float)! !

!Matrix2x3Plugin methodsFor: 'transforming' stamp: 'tpr (auto pragmas dtl 2010-09-28) 12/29/2005 17:02'!
matrix2x3InvertPoint: m
	"Invert the pre-loaded argument point by the given matrix"
	| x y det detX detY |
	<var: #m type: 'float *'>
	<var: #x type: 'double '>
	<var: #y type: 'double '>
	<var: #det type: 'double '>
	<var: #detX type: 'double '>
	<var: #detY type: 'double '>

	x := m23ArgX - (m at: 2).
	y := m23ArgY - (m at: 5).
	det := ((m at: 0) * (m at: 4)) - ((m at: 1) * (m at: 3)).
	det = 0.0 ifTrue:[^interpreterProxy primitiveFail]."Matrix is singular."
	det := 1.0 / det.
	detX := (x * (m at: 4)) - ((m at: 1) * y).
	detY := ((m at: 0) * y) - (x * (m at: 3)).
	m23ResultX := detX * det.
	m23ResultY := detY * det.! !

!Matrix2x3Plugin methodsFor: 'transforming' stamp: 'tpr (auto pragmas dtl 2010-09-28) 12/29/2005 17:03'!
matrix2x3TransformPoint: m
	"Transform the pre-loaded argument point by the given matrix"
	<var: #m type: 'float *'>
	m23ResultX := (m23ArgX * (m at: 0)) + (m23ArgY * (m at: 1)) + (m at: 2).
	m23ResultY := (m23ArgX * (m at: 3)) + (m23ArgY * (m at: 4)) + (m at: 5).! !

!Matrix2x3Plugin methodsFor: 'private' stamp: 'eem 9/14/2015 14:15'!
okayIntValue: value
	^(value >= interpreterProxy minSmallInteger asFloat
	  and: [value <= interpreterProxy maxSmallInteger asFloat])! !

!Matrix2x3Plugin methodsFor: 'primitives' stamp: 'eem 10/21/2018 12:15'!
primitiveComposeMatrix
	<export: true>
	| m1 m2 m3 result |
	<var: #m1 type: #'float *'>
	<var: #m2 type: #'float *'>
	<var: #m3 type: #'float *'>
	m3 := self loadArgumentMatrix: (result := interpreterProxy stackValue: 0).
	m2 := self loadArgumentMatrix: (interpreterProxy stackValue: 1).
	m1 := self loadArgumentMatrix: (interpreterProxy stackValue: 2).
	interpreterProxy failed ifFalse:
		[self matrix2x3ComposeMatrix: m1 with: m2 into: m3.
		 interpreterProxy pop: 3 thenPush: result]! !

!Matrix2x3Plugin methodsFor: 'primitives' stamp: 'eem 10/21/2018 12:14'!
primitiveInvertPoint
	<export: true>
	| matrix |
	<var: #matrix type: #'float *'>
	self loadArgumentPoint: (interpreterProxy stackValue: 0).
	matrix := self loadArgumentMatrix: (interpreterProxy stackValue: 1).
	interpreterProxy failed ifTrue:[^nil].
	self matrix2x3InvertPoint: matrix.
	interpreterProxy failed ifFalse:[self roundAndStoreResultPoint: 2]! !

!Matrix2x3Plugin methodsFor: 'primitives' stamp: 'eem 10/21/2018 12:12'!
primitiveInvertRectInto
	<export: true>
	| matrix srcOop dstOop originX originY cornerX cornerY minX maxX minY maxY |
	<var: #matrix type: #'float *'>
	<var: #originX type: #double>
	<var: #originY type: #double>
	<var: #cornerX type: #double>
	<var: #cornerY type: #double>
	<var: #minX type: #double>
	<var: #maxX type: #double>
	<var: #minY type: #double>
	<var: #maxY type: #double>

	dstOop := interpreterProxy stackValue: 0.
	srcOop := interpreterProxy stackValue: 1.
	matrix := self loadArgumentMatrix: (interpreterProxy stackValue: 2).

	(matrix notNil
	and: [(interpreterProxy fetchClassOf: srcOop) = (interpreterProxy fetchClassOf: dstOop)
	and: [(interpreterProxy isPointers: srcOop)
	and: [(interpreterProxy slotSizeOf: srcOop) = 2]]]) ifFalse:
		[^interpreterProxy primitiveFail].

	"Load top-left point"
	self loadArgumentPoint: (interpreterProxy fetchPointer: 0 ofObject: srcOop).
	interpreterProxy failed ifTrue: [^nil].
	originX := m23ArgX.
	originY := m23ArgY.
	self matrix2x3InvertPoint: matrix.
	minX := maxX := m23ResultX.
	minY := maxY := m23ResultY.

	"Load bottom-right point"
	self loadArgumentPoint: (interpreterProxy fetchPointer: 1 ofObject: srcOop).
	interpreterProxy failed ifTrue: [^nil].
	cornerX := m23ArgX.
	cornerY := m23ArgY.
	self matrix2x3InvertPoint: matrix.
	minX := minX min: m23ResultX.
	maxX := maxX max: m23ResultX.
	minY := minY min: m23ResultY.
	maxY := maxY max: m23ResultY.

	"Load top-right point"
	m23ArgX := cornerX.
	m23ArgY := originY.
	self matrix2x3InvertPoint: matrix.
	minX := minX min: m23ResultX.
	maxX := maxX max: m23ResultX.
	minY := minY min: m23ResultY.
	maxY := maxY max: m23ResultY.

	"Load bottom-left point"
	m23ArgX := originX.
	m23ArgY := cornerY.
	self matrix2x3InvertPoint: matrix.
	minX := minX min: m23ResultX.
	maxX := maxX max: m23ResultX.
	minY := minY min: m23ResultY.
	maxY := maxY max: m23ResultY.

	dstOop := self roundAndStoreResultRect: dstOop x0: minX y0: minY x1: maxX y1: maxY.
	interpreterProxy failed ifFalse:
		[interpreterProxy pop: 3 thenPush: dstOop]! !

!Matrix2x3Plugin methodsFor: 'primitives' stamp: 'eem 10/21/2018 12:14'!
primitiveIsIdentity
	<export: true>
	| matrix |
	<var: #matrix type: #'float *'>
	matrix := self loadArgumentMatrix: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnBool:(
			((matrix at: 0) = (self cCoerce: 1.0 to: #float)) &
			((matrix at: 1) = (self cCoerce: 0.0 to: #float)) &
			((matrix at: 2) = (self cCoerce: 0.0 to: #float)) &
			((matrix at: 3) = (self cCoerce: 0.0 to: #float)) &
			((matrix at: 4) = (self cCoerce: 1.0 to: #float)) &
			((matrix at: 5) = (self cCoerce: 0.0 to: #float)))]! !

!Matrix2x3Plugin methodsFor: 'primitives' stamp: 'eem 10/21/2018 12:06'!
primitiveIsPureTranslation
	<export: true>
	| matrix |
	<var: #matrix type: #'float *'>
	matrix := self loadArgumentMatrix: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnBool:(
			((matrix at: 0) = (self cCoerce: 1.0 to: #float)) &
			((matrix at: 1) = (self cCoerce: 0.0 to: #float)) &
			((matrix at: 3) = (self cCoerce: 0.0 to: #float)) &
			((matrix at: 4) = (self cCoerce: 1.0 to: #float)))]! !

!Matrix2x3Plugin methodsFor: 'primitives' stamp: 'eem 10/21/2018 11:20'!
primitiveTransformPoint
	<export: true>
	| matrix |
	<var: #matrix type: #'float *'>
	self loadArgumentPoint: (interpreterProxy stackValue: 0).
	matrix := self loadArgumentMatrix: (interpreterProxy stackValue: 1).
	interpreterProxy failed ifFalse:
		[self matrix2x3TransformPoint: matrix.
		 self roundAndStoreResultPoint: 2]! !

!Matrix2x3Plugin methodsFor: 'primitives' stamp: 'eem 10/21/2018 11:15'!
primitiveTransformRectInto
	<export: true>
	| matrix srcOop dstOop originX originY cornerX cornerY minX maxX minY maxY |
	<var: #matrix type: #'float *'>
	<var: #originX type: #double>
	<var: #originY type: #double>
	<var: #cornerX type: #double>
	<var: #cornerY type: #double>
	<var: #minX type: #double>
	<var: #maxX type: #double>
	<var: #minY type: #double>
	<var: #maxY type: #double>

	dstOop := interpreterProxy stackValue: 0.
	srcOop := interpreterProxy stackValue: 1.
	matrix := self loadArgumentMatrix: (interpreterProxy stackValue: 2).

	(matrix notNil
	 and: [(interpreterProxy fetchClassOf: srcOop) = (interpreterProxy fetchClassOf: dstOop)
	 and: [(interpreterProxy isPointers: srcOop)
	 and: [(interpreterProxy slotSizeOf: srcOop) = 2]]]) ifFalse:
		[^interpreterProxy primitiveFail].

	"Load top-left point"
	self loadArgumentPoint: (interpreterProxy fetchPointer: 0 ofObject: srcOop).
	interpreterProxy failed ifTrue:[^nil].
	originX := m23ArgX.
	originY := m23ArgY.
	self matrix2x3TransformPoint: matrix.
	minX := maxX := m23ResultX.
	minY := maxY := m23ResultY.

	"Load bottom-right point"
	self loadArgumentPoint:(interpreterProxy fetchPointer: 1 ofObject: srcOop).
	interpreterProxy failed ifTrue:[^nil].
	cornerX := m23ArgX.
	cornerY := m23ArgY.
	self matrix2x3TransformPoint: matrix.
	minX := minX min: m23ResultX.
	maxX := maxX max: m23ResultX.
	minY := minY min: m23ResultY.
	maxY := maxY max: m23ResultY.

	"Load top-right point"
	m23ArgX := cornerX.
	m23ArgY := originY.
	self matrix2x3TransformPoint: matrix.
	minX := minX min: m23ResultX.
	maxX := maxX max: m23ResultX.
	minY := minY min: m23ResultY.
	maxY := maxY max: m23ResultY.

	"Load bottom-left point"
	m23ArgX := originX.
	m23ArgY := cornerY.
	self matrix2x3TransformPoint: matrix.
	minX := minX min: m23ResultX.
	maxX := maxX max: m23ResultX.
	minY := minY min: m23ResultY.
	maxY := maxY max: m23ResultY.

	dstOop := self roundAndStoreResultRect: dstOop x0: minX y0: minY x1: maxX y1: maxY.
	interpreterProxy failed ifFalse:
		[interpreterProxy pop: 3 thenPush: dstOop]
! !

!Matrix2x3Plugin methodsFor: 'private' stamp: 'eem 10/22/2018 14:00'!
roundAndStoreResultPoint: nItemsToPop
	"Store the result of a previous operation.
	Fail if we cannot represent the result as SmallInteger"
	m23ResultX := m23ResultX + 0.5.
	m23ResultY := m23ResultY + 0.5.
	((self okayIntValue: m23ResultX)
	 and: [self okayIntValue: m23ResultY]) ifFalse:
		[^interpreterProxy primitiveFail].
	interpreterProxy
		pop: nItemsToPop
		thenPush: (interpreterProxy
					makePointwithxValue: m23ResultX asInteger
					yValue: m23ResultY asInteger)! !

!Matrix2x3Plugin methodsFor: 'private' stamp: 'eem 10/21/2018 11:31'!
roundAndStoreResultRect: dstOop x0: x0 y0: y0 x1: x1 y1: y1
	"Check, round and store the result of a rectangle operation"
	<var: #x0 type: #double>
	<var: #y0 type: #double>
	<var: #x1 type: #double>
	<var: #y1 type: #double>
	| minX maxX minY maxY originOop cornerOop |
	<var: #minX type: #double>
	<var: #maxX type: #double>
	<var: #minY type: #double>
	<var: #maxY type: #double>

	((self okayIntValue: (minX := x0 + 0.5))
	and: [(self okayIntValue: (maxX := x1 + 0.5))
	and: [(self okayIntValue: (minY := y0 + 0.5))
	and: [self okayIntValue: (maxY := y1 + 0.5)]]]) ifFalse:
		[^interpreterProxy primitiveFail].

	self remapOop: dstOop in:
		[originOop := interpreterProxy makePointwithxValue: minX asInteger yValue: minY asInteger.
		 self remapOop: originOop in:
			[cornerOop := interpreterProxy makePointwithxValue: maxX asInteger yValue: maxY asInteger]].
	interpreterProxy
		storePointer: 0 ofObject: dstOop withValue: originOop;
		storePointer: 1 ofObject: dstOop withValue: cornerOop.
	^dstOop! !

Matrix2x3Plugin subclass: #Matrix2x3PluginSimulator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!Matrix2x3PluginSimulator methodsFor: 'primitives' stamp: 'eem 10/21/2018 12:25'!
primitiveComposeMatrix
	| m1 m2 m3 |
	m2 := self stackMatrixValue: 1.
	m1 := self stackMatrixValue: 2.
	super primitiveComposeMatrix.
	interpreterProxy failed ifFalse:
		[m3 := self stackMatrixValue: 0.
		 self assert: m3 = (m1 composedWithLocal: m2)]! !

!Matrix2x3PluginSimulator methodsFor: 'simulation support' stamp: 'eem 10/21/2018 18:34'!
stackMatrixValue: n
	| maybeMatrixOop |
	maybeMatrixOop := interpreterProxy stackValue: n.
	^((interpreterProxy isWords: maybeMatrixOop)
	  and: [(interpreterProxy slotSizeOf: maybeMatrixOop) = 6]) ifTrue:
		[MatrixTransform2x3 new
			basicAt: 1 put: (interpreterProxy fetchLong32: 0 ofObject: maybeMatrixOop);
			basicAt: 2 put: (interpreterProxy fetchLong32: 1 ofObject: maybeMatrixOop);
			basicAt: 3 put: (interpreterProxy fetchLong32: 2 ofObject: maybeMatrixOop);
			basicAt: 4 put: (interpreterProxy fetchLong32: 3 ofObject: maybeMatrixOop);
			basicAt: 5 put: (interpreterProxy fetchLong32: 4 ofObject: maybeMatrixOop);
			basicAt: 6 put: (interpreterProxy fetchLong32: 5 ofObject: maybeMatrixOop);
			yourself]! !

InterpreterPlugin subclass: #MiscPrimitivePlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!MiscPrimitivePlugin commentStamp: 'eem 2/21/2018 15:10' prior: 0!
This plugin pulls together a number of string and array related primitives.  In the olden days these primitives were generated as "translated primitives", which generated the primitive versions form methods in the image.  In 2018 they were rewritten to use conventional Slang.!

!MiscPrimitivePlugin methodsFor: 'helper functions' stamp: 'eem 2/21/2018 14:51'!
encodeBytesOf: anInt in: ba at: i
	"Copy the integer anInt into byteArray ba at index i, and answer the next index"
	<inline: #always>
	<var: #ba type: #'unsigned char *'>
	0 to: 3 do:
		[:j | ba at: i + j put: (anInt >> (3 - j * 8) bitAnd: 16rFF)].
	^i + 4! !

!MiscPrimitivePlugin methodsFor: 'helper functions' stamp: 'eem 2/21/2018 14:51'!
encodeInt: anInt in: ba at: i
	"Encode the integer anInt in byteArray ba at index i, and answer the next index.
	 The encoding is as follows...
		0-223		0-223
		224-254	(0-30) * 256 + next byte (0-7935)
		255		next 4 bytes"		

	<inline: #always>
	<var: #ba declareC: 'unsigned char *ba'>
	anInt <= 223 ifTrue: [ba at: i put: anInt. ^i + 1].
	anInt <= 7935 ifTrue: [ba at: i put: anInt // 256 + 224. ba at: i + 1 put: anInt \\ 256. ^i + 2].
	ba at: i put: 255.
	^self encodeBytesOf: anInt in: ba at: i + 1! !

!MiscPrimitivePlugin methodsFor: 'primitives' stamp: 'eem 7/22/2021 18:47'!
primitiveCompareString
	"ByteString (class) compare: string1 with: string2 collated: order"
	<export: true flags: #FastCPrimitiveFlag>
	| len1 len2 order string1 string2 orderOop string1Oop string2Oop |

	<var: 'order' type: #'unsigned char *'>
	<var: 'string1' type: #'unsigned char *'>
	<var: 'string2' type: #'unsigned char *'>
	orderOop := interpreterProxy stackValue: 0.
	string2Oop := interpreterProxy stackValue: 1.
	string1Oop := interpreterProxy stackValue: 2.
	((interpreterProxy isBytes: orderOop)
	and: [(interpreterProxy isBytes: string2Oop)
	and: [interpreterProxy isBytes: string1Oop]]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	order := interpreterProxy firstIndexableField: orderOop.
	(interpreterProxy sizeOfSTArrayFromCPrimitive: order) < 256 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	string1 := interpreterProxy firstIndexableField: string1Oop.
	string2 := interpreterProxy firstIndexableField: string2Oop.
	len1 := interpreterProxy sizeOfSTArrayFromCPrimitive: string1.
	len2 := interpreterProxy sizeOfSTArrayFromCPrimitive: string2.
	interpreterProxy failed ifTrue: "the sizeOfSTArrayFromCPrimitive:'s fail for e.g. CompiledMethod"
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	0 to: (len1 min: len2) - 1 do: 
		[ :i | | c1 c2 |
		c1 := order at: (string1 at: i).
		c2 := order at: (string2 at: i).
		c1 = c2 ifFalse:
			[^interpreterProxy methodReturnInteger: (c1 < c2 ifTrue: [1] ifFalse: [3])]].
	interpreterProxy methodReturnInteger:
		(len1 = len2 ifTrue: [2] ifFalse: [len1 < len2 ifTrue: [1] ifFalse: [3]])! !

!MiscPrimitivePlugin methodsFor: 'primitives' stamp: 'eem 7/22/2021 19:23'!
primitiveCompressToByteArray
	"Bitmap compress: bm toByteArray: ba"
	<export: true flags: #FastCPrimitiveFlag>
	| bm ba eqBytes i j k lowByte size destSize word |
	<var: 'ba' type: #'unsigned char *'>
	<var: 'bm' type: #'int *'>
	bm := self cCode: [interpreterProxy arrayValueOf: (interpreterProxy stackValue: 1)]
				inSmalltalk: [interpreterProxy
								cCoerce: (interpreterProxy arrayValueOf: (interpreterProxy stackValue: 1))
								to: #'int *'].
	interpreterProxy failed ifTrue: [^nil].
	(interpreterProxy isBytes: (interpreterProxy stackValue: 0)) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(interpreterProxy isOopImmutable: (interpreterProxy stackValue: 0)) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	ba := interpreterProxy firstIndexableField: (interpreterProxy stackValue: 0).
	size := interpreterProxy sizeOfSTArrayFromCPrimitive: bm.
	destSize := interpreterProxy sizeOfSTArrayFromCPrimitive: ba.
	interpreterProxy failed ifTrue: "the sizeOfSTArrayFromCPrimitive:'s fail for e.g. CompiledMethod"
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	destSize < ((size * 4) + 7 + (size // 1984 * 3)) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrUnsupported]. "Size may be OK but we don't know, hence fail with unsupported"
	i := self encodeInt: size in: ba at: 0.
	k := 0.
	[k < size] whileTrue: 
		[word := bm at: k.
		lowByte := word bitAnd: 255.
		eqBytes := (word >> 8 bitAnd: 255) = lowByte and: [(word >> 16 bitAnd: 255) = lowByte and: [(word >> 24 bitAnd: 255) = lowByte]].
		j := k.
		[j + 1 < size and: [word = (bm at: j + 1)]] whileTrue: [j := j + 1].
		j > k
			ifTrue: 
				[eqBytes
					ifTrue: 
						[i := self encodeInt: j - k + 1 * 4 + 1 in: ba at: i.
						ba at: i put: lowByte.
						i := i + 1]
					ifFalse: 
						[i := self encodeInt: j - k + 1 * 4 + 2 in: ba at: i.
						i := self encodeBytesOf: word in: ba at: i].
				k := j + 1]
			ifFalse:
				[eqBytes
					ifTrue: 
						[i := self encodeInt: 1 * 4 + 1 in: ba at: i.
						ba at: i put: lowByte.
						i := i + 1.
						k := k + 1]
					ifFalse: 
						[[j + 1 < size and: [(bm at: j) ~= (bm at: j + 1)]] whileTrue: [j := j + 1].
						j + 1 = size ifTrue: [j := j + 1].
						i := self encodeInt: j - k * 4 + 3 in: ba at: i.
						k to: j - 1 by: 1 do: [ :m | i := self encodeBytesOf: (bm at: m) in: ba at: i].
						k := j]]].
	interpreterProxy methodReturnInteger: i! !

!MiscPrimitivePlugin methodsFor: 'primitives' stamp: 'eem 7/22/2021 19:23'!
primitiveConvert8BitSigned
	"SampledSound (class) convert8bitSignedFrom: aByteArray to16Bit: aSoundBuffer"
	<export: true flags: #FastCPrimitiveFlag>
	| aByteArray aSoundBuffer arraySize byteArrayOop soundBufferOop |

	<var: 'aByteArray' type: #'unsigned char *'>
	<var: 'aSoundBuffer' type: #'unsigned short *'>
	byteArrayOop := interpreterProxy stackValue: 1.
	(interpreterProxy isBytes: byteArrayOop) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	aByteArray := interpreterProxy firstIndexableField: byteArrayOop.
	soundBufferOop := interpreterProxy stackValue: 0.
	(interpreterProxy isOopImmutable: soundBufferOop) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	aSoundBuffer := self
						cCode: [interpreterProxy arrayValueOf: soundBufferOop]
						inSmalltalk: [interpreterProxy
										cCoerce: (interpreterProxy arrayValueOf: soundBufferOop)
										to: #'unsigned short *'].
	arraySize := interpreterProxy sizeOfSTArrayFromCPrimitive: aByteArray.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(interpreterProxy byteSizeOf: soundBufferOop) < (2 * arraySize) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	0 to: arraySize - 1 do: 
		[ :i | | s |
		s := aByteArray at: i.
		aSoundBuffer
			at: i
			put: (s > 127
					ifTrue: [s - 256 bitShift: 8]
					ifFalse: [s bitShift: 8])].
	interpreterProxy methodReturnReceiver! !

!MiscPrimitivePlugin methodsFor: 'primitives' stamp: 'eem 7/22/2021 19:24'!
primitiveDecompressFromByteArray
	"Bitmap decompress: bm fromByteArray: ba at: index"
	<export: true flags: #FastCPrimitiveFlag>
	| bm ba index i anInt code data end k n pastEnd |
	<var: 'ba' type: #'unsigned char *'>
	<var: 'bm' type: #'int *'>
	<var: 'anInt' type: #'unsigned int'>
	<var: 'code' type: #'unsigned int'>
	<var: 'data' type: #'unsigned int'>
	<var: 'n' type: #'unsigned int'>
	bm := self cCode: [interpreterProxy arrayValueOf: (interpreterProxy stackValue: 2)]
				inSmalltalk: [interpreterProxy
								cCoerce: (interpreterProxy arrayValueOf: (interpreterProxy stackValue: 2))
								to: #'int *'].
	(interpreterProxy isOopImmutable: (interpreterProxy stackValue: 2)) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	(interpreterProxy isBytes: (interpreterProxy stackValue: 1)) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	ba := interpreterProxy firstIndexableField: (interpreterProxy stackValue: 1).
	index := interpreterProxy stackIntegerValue: 0.
	end := interpreterProxy sizeOfSTArrayFromCPrimitive: ba.
	pastEnd := interpreterProxy sizeOfSTArrayFromCPrimitive: bm.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	i := index - 1.
	k := 0.
	[i < end] whileTrue: 
		[anInt := ba at: i.
		i := i + 1.
		anInt <= 223 ifFalse:
			[anInt <= 254
				ifTrue: 
					[anInt := anInt - 224 * 256 + (ba at: i).
					i := i + 1]
				ifFalse: 
					[anInt := 0.
					1 to: 4 by: 1 do: 
						[ :j | anInt := (anInt bitShift: 8) + (ba at: i).
						i := i + 1]]].
		n := anInt >> 2.
		k + n > pastEnd ifTrue:
			[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
		code := anInt bitAnd: 3.
		"code = 0 ifTrue: [nil]."
		code = 1 ifTrue: 
			[data := ba at: i.
			i := i + 1.
			data := data bitOr: (data bitShift: 8).
			data := data bitOr: (data bitShift: 16).
			1 to: n do: 
				[ :j |
				bm at: k put: data.
				k := k + 1]].
		code = 2 ifTrue: 
			[data := 0.
			1 to: 4 do: 
				[ :j |
				data := (data bitShift: 8) bitOr: (ba at: i).
				i := i + 1].
			1 to: n do: 
				[ :j |
				bm at: k put: data.
				k := k + 1]].
		code = 3 ifTrue:
			[1 to: n do: 
				[ :m |
				data := 0.
				1 to: 4 do: 
					[ :j |
					data := (data bitShift: 8) bitOr: (ba at: i).
					i := i + 1].
				bm at: k put: data.
				k := k + 1]]].
	interpreterProxy pop: interpreterProxy methodArgumentCount! !

!MiscPrimitivePlugin methodsFor: 'primitives' stamp: 'eem 7/22/2021 19:24'!
primitiveFindFirstInString
	"ByteString (class) findFirstInString: aString inSet: inclusionMap  startingAt: start"
	<export: true flags: #FastCPrimitiveFlag>
	
	|  aString i inclusionMap stringSize aStringOop inclusionMapOop |
	<var: 'aString' type: #'unsigned char *'>
	<var: 'inclusionMap' type: #'unsigned char *'>
	aStringOop := interpreterProxy stackValue: 2.
	(interpreterProxy isBytes: aStringOop) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	inclusionMapOop :=  interpreterProxy stackValue: 1.
	(interpreterProxy isBytes: inclusionMapOop) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	i := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	i := i - 1. "Convert to 0-based index."
	i < 0 ifTrue: [^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	inclusionMap := interpreterProxy firstIndexableField: inclusionMapOop.
	(interpreterProxy sizeOfSTArrayFromCPrimitive: inclusionMap) ~= 256 ifTrue:
		[^interpreterProxy methodReturnInteger: 0].
	aString := interpreterProxy firstIndexableField: aStringOop.
	stringSize := interpreterProxy sizeOfSTArrayFromCPrimitive: aString.
	interpreterProxy failed ifTrue: "the sizeOfSTArrayFromCPrimitive:'s fail for e.g. CompiledMethod"
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	[i < stringSize and: [(inclusionMap at: (aString at: i)) = 0]] whileTrue:
		[i := i + 1].
	interpreterProxy methodReturnInteger: (i >= stringSize ifTrue: [0] ifFalse: [i + 1])! !

!MiscPrimitivePlugin methodsFor: 'primitives' stamp: 'eem 7/22/2021 19:23'!
primitiveFindSubstring
	"ByteString findSubstring: key in: body startingAt: start matchTable: matchTable"
	<export: true flags: #FastCPrimitiveFlag>

	| body key keySize bodySize matchTable start bodyOop keyOop matchTableOop |
	<var: #key type: #'unsigned char *'>
	<var: #body type: #'unsigned char *'>
	<var: #matchTable type: #'unsigned char *'>
	keyOop := interpreterProxy stackValue: 3.
	(interpreterProxy isBytes: keyOop) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	bodyOop := interpreterProxy stackValue: 2.
	(interpreterProxy isBytes: bodyOop) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	start := interpreterProxy stackIntegerValue: 1.
	interpreterProxy failed ifTrue: 
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	matchTableOop := interpreterProxy stackValue: 0.
	(interpreterProxy isBytes: matchTableOop) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	matchTable := interpreterProxy firstIndexableField: matchTableOop.
	(interpreterProxy sizeOfSTArrayFromCPrimitive: matchTable) < 256 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].		
	key := interpreterProxy firstIndexableField: keyOop.
	(keySize := interpreterProxy sizeOfSTArrayFromCPrimitive: key) > 0 ifTrue:
		[keySize := keySize - 1. "adjust for zero relative indexes"
		start := start - 1 max: 0. "adjust for zero relative indexes"
		body := interpreterProxy firstIndexableField: bodyOop.
		bodySize := interpreterProxy sizeOfSTArrayFromCPrimitive: body.
		interpreterProxy failed ifTrue: "the sizeOfSTArrayFromCPrimitive:'s fail for e.g. CompiledMethod"
			[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
		start to: bodySize - 1 - keySize do: 
			[ :startIndex | | index |
			index := 0.
			[(matchTable at: (body at: startIndex + index)) = (matchTable at: (key at: index))] whileTrue: 
				[index = keySize ifTrue:
					[^interpreterProxy methodReturnInteger: startIndex + 1].
				index := index + 1]]].
	^interpreterProxy methodReturnInteger: 0! !

!MiscPrimitivePlugin methodsFor: 'primitives' stamp: 'eem 7/22/2021 19:24'!
primitiveIndexOfAsciiInString
	"ByteString indexOfAscii: anInteger inString: aString startingAt: start"
	<export: true flags: #FastCPrimitiveFlag>

	| integerOop startOop anInteger aString start stringSize stringOop |
	<var: #aString type: #'unsigned char *'>
	integerOop := interpreterProxy stackValue: 2.
	stringOop := interpreterProxy stackValue: 1.
	startOop := interpreterProxy stackValue: 0.
	((interpreterProxy isIntegerObject: integerOop)
	 and: [(interpreterProxy isIntegerObject: startOop)
	 and: [(interpreterProxy isBytes: stringOop)
	 and: [interpreterProxy isWordsOrBytes: stringOop]]]) ifFalse: "sizeOfSTArrayFromCPrimitive: is defined only for words or bytes"
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(start := interpreterProxy integerValueOf: startOop) >= 1 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	anInteger := interpreterProxy integerValueOf: integerOop.
	aString := interpreterProxy firstIndexableField: stringOop.
	stringSize := interpreterProxy sizeOfSTArrayFromCPrimitive: aString.
	start - 1 to: stringSize - 1 do:
		[:pos |
		(aString at: pos) = anInteger ifTrue:
			[^interpreterProxy methodReturnInteger: pos + 1]].
	^interpreterProxy methodReturnInteger: 0! !

!MiscPrimitivePlugin methodsFor: 'primitives' stamp: 'eem 7/25/2021 20:01'!
primitiveStringHash
	"ByteArray (class) hashBytes: aByteArray startingWith: speciesHash"
	<export: true flags: #FastCPrimitiveFlag>

	| aByteArray hash byteArrayOop |
	<var: 'aByteArray' type: #'unsigned char *'>
	<var: 'hash' type: #'unsigned int'>
	hash := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	byteArrayOop := interpreterProxy stackValue: 1.
	((interpreterProxy isBytes: byteArrayOop)
	and: [interpreterProxy isWordsOrBytes: byteArrayOop]) ifFalse: "filters out CompiledMethods"
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	aByteArray := interpreterProxy firstIndexableField: byteArrayOop.
	0 to: (interpreterProxy sizeOfSTArrayFromCPrimitive: aByteArray) - 1 do: 
		[ :pos |
		hash := hash + (aByteArray at: pos) * 16r19660D ].
	interpreterProxy methodReturnInteger: (hash bitAnd: 16r0FFFFFFF)! !

!MiscPrimitivePlugin methodsFor: 'primitives' stamp: 'eem 7/22/2021 19:24'!
primitiveTranslateStringWithTable
	"ByteString (class) translate: aString from: start to: stop table: table"
	<export: true flags: #FastCPrimitiveFlag>

	| aString start stop table aStringOop tableOop |
	<var: #table type: #'unsigned char *'>
	<var: #aString type: #'unsigned char *'>
	aStringOop := interpreterProxy stackValue: 3.
	(interpreterProxy isBytes: aStringOop) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(interpreterProxy isOopImmutable: aStringOop) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoModification].
	start := interpreterProxy stackIntegerValue: 2.
	stop := interpreterProxy stackIntegerValue: 1.
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	tableOop := interpreterProxy stackValue: 0.
	(interpreterProxy isBytes: tableOop) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	aString := interpreterProxy firstIndexableField: aStringOop.
	(start >= 1 and: [stop <= (interpreterProxy sizeOfSTArrayFromCPrimitive: aString)]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].
	table := interpreterProxy firstIndexableField: tableOop.
	(interpreterProxy sizeOfSTArrayFromCPrimitive: table) < 256 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	start - 1 to: stop - 1 do: [ :i | aString at: i put: (table at: (aString at: i))].
	interpreterProxy methodReturnReceiver! !

InterpreterPlugin subclass: #ScratchPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!ScratchPlugin commentStamp: 'jm 11/8/2006 18:33' prior: 0!
This plugin combines a number of primitives needed by Scratch including:

  a. primitives that manipulate 24-bit color images (i.e. 32-bit deep Forms but alpha is ignored)
  b. primitive to open browsers, find the user's documents folder, set the window title and other host OS functions

This plugin includes new serial port primitives, including support for named serial ports. The underlying plugin code can support up to 32 simultaenously open ports.

Port options for Set/GetOption primitives:
  1. baud rate
  2. data bits
  3. stop bits
  4. parity type
  5. input flow control type
  6. output flow control type

Handshake lines (options 20-25 for Set/GetOption primitives):
  20. DTR	(output line)
  21. RTS	(output line)
  22. CTS	(input line)
  23. DSR	(input line)
  24. CD		(input line)
  25. RI		(input line)

!

!ScratchPlugin class methodsFor: 'translation' stamp: 'ikp 7/30/2012 14:38'!
hasHeaderFile

	^true! !

!ScratchPlugin methodsFor: 'private' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/8/2006 05:30'!
bitmap: bitmap at: i putH: hue s: saturation v: brightness

	| hI hF p q t v outPix |
	<inline: true>
	<var: 'bitmap' declareC: 'unsigned int *bitmap'>

	hI := hue // 60.  "integer part of hue (0..5)"
	hF := hue \\ 60.  "fractional part ofhue"
	p := (1000 - saturation) * brightness.
	q := (1000 - ((saturation * hF) // 60)) * brightness.
	t := (1000 - ((saturation * (60 - hF)) // 60)) * brightness.

	v := (brightness * 1000) // 3922.
	p := p // 3922.
	q := q // 3922.
	t := t // 3922.

	0 = hI ifTrue: [outPix := ((v bitShift: 16) + (t bitShift: 8) + p)].
	1 = hI ifTrue: [outPix := ((q bitShift: 16) + (v bitShift: 8) + p)].
	2 = hI ifTrue: [outPix := ((p bitShift: 16) + (v bitShift: 8) + t)].
	3 = hI ifTrue: [outPix := ((p bitShift: 16) + (q bitShift: 8) + v)].
	4 = hI ifTrue: [outPix := ((t bitShift: 16) + (p bitShift: 8) + v)].
	5 = hI ifTrue: [outPix := ((v bitShift: 16) + (p bitShift: 8) + q)].

	outPix = 0 ifTrue: [outPix := 1].  "convert transparent to 1"
	bitmap at: i put: outPix.
	^ 0
! !

!ScratchPlugin methodsFor: 'private' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/8/2006 05:30'!
checkedFloatPtrOf: oop
	"Return an unsigned int pointer to the first indexable word of oop, which must be a words object."

	<inline: true>
	<returnTypeC: 'double *'>

	interpreterProxy success: (interpreterProxy isWordsOrBytes: oop).
	interpreterProxy failed ifTrue: [^ 0].
	^ self cCoerce: (interpreterProxy firstIndexableField: oop) to: 'double *'
! !

!ScratchPlugin methodsFor: 'private' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/8/2006 05:30'!
checkedUnsignedIntPtrOf: oop
	"Return an unsigned int pointer to the first indexable word of oop, which must be a words object."

	<inline: true>
	<returnTypeC: 'unsigned int *'>

	interpreterProxy success: (interpreterProxy isWords: oop).
	interpreterProxy failed ifTrue: [^ 0].
	^ self cCoerce: (interpreterProxy firstIndexableField: oop) to: 'unsigned int *'
! !

!ScratchPlugin methodsFor: 'private' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/8/2006 05:30'!
hueFromR: r G: g B: b min: min max: max
	"Answer the hue, an angle between 0 and 360."

	| span result |
	<inline: true>

	span := max - min.
	span = 0 ifTrue: [^ 0].

	r = max
		ifTrue: [result := ((60 * (g - b)) // span)]
		ifFalse: [
			g = max
				ifTrue: [result := 120 + ((60 * (b - r)) // span)]
				ifFalse: [result := 240 + ((60 * (r - g)) // span)]].

	result < 0 ifTrue: [^ result + 360].
	^ result
! !

!ScratchPlugin methodsFor: 'private' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/8/2006 05:30'!
interpolate: pix1 and: pix2 frac: frac2
	"Answer the interpolated pixel value between the given two pixel values. If either pixel is zero (transparent) answer the other pixel. If both pixels are  transparent, answer transparent. The fraction is between 0 and 1023, out of a total range of 1024."

	| frac1 r g b result |
	<inline: true>

	pix1 = 0 ifTrue: [^ pix2].  "pix1 is transparent"
	pix2 = 0 ifTrue: [^ pix1].  "pix2 is transparent"

	frac1 := 1024 - frac2.
	r := ((frac1 * ((pix1 bitShift: -16) bitAnd: 16rFF)) + (frac2 * ((pix2 bitShift: -16) bitAnd: 16rFF))) // 1024.
	g := ((frac1 * ((pix1 bitShift: -8) bitAnd: 16rFF)) + (frac2 * ((pix2 bitShift: -8) bitAnd: 16rFF))) // 1024.
	b := ((frac1 * (pix1 bitAnd: 16rFF)) + (frac2 * (pix2 bitAnd: 16rFF))) // 1024.
	result := (r bitShift: 16) + (g bitShift: 8) + b.
	result = 0 ifTrue: [result := 1].
	^ result
! !

!ScratchPlugin methodsFor: 'private' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/8/2006 05:30'!
interpolatedFrom: bitmap x: xFixed y: yFixed width: w height: h
	"Answer the interpolated pixel value from the given bitmap at the given point. The x and y coordinates are fixed-point integers with 10 bits of fraction (i.e. they were multiplied by 1024, then truncated). If the given point is right on an edge, answer the nearest edge pixel value. If it is entirely outside of the image, answer 0 (transparent)."

	| x y xFrac yFrac index topPix bottomPix |
	<inline: true>
	<var: 'bitmap' declareC: 'unsigned int *bitmap'>

	x := xFixed bitShift: -10.
	(x < -1 or: [x >= w]) ifTrue: [^ 0].
	y := yFixed bitShift: -10.
	(y < -1 or: [y >= h]) ifTrue: [^ 0].

	xFrac := xFixed bitAnd: 1023.
	x = -1 ifTrue: [x := 0. xFrac := 0].  "left edge"
	x = (w - 1) ifTrue: [xFrac := 0].  "right edge"

	yFrac := yFixed bitAnd: 1023.
	y = -1 ifTrue: [y := 0. yFrac := 0].  "top edge"
	y = (h - 1) ifTrue: [yFrac := 0].  "bottom edge"

	index := (y * w) + x "for squeak: + 1".
	topPix := (bitmap at: index) bitAnd: 16rFFFFFF.
	xFrac > 0 ifTrue: [
		topPix := self interpolate: topPix and: ((bitmap at: index + 1) bitAnd: 16rFFFFFF) frac: xFrac].

	yFrac = 0 ifTrue: [^ topPix].  "no y fraction, so just use value from top row"

	index := ((y + 1) * w) + x "for squeak: + 1".
	bottomPix := (bitmap at: index) bitAnd: 16rFFFFFF.
	xFrac > 0 ifTrue: [
		bottomPix := self interpolate: bottomPix and: ((bitmap at: index + 1) bitAnd: 16rFFFFFF) frac: xFrac].

	^ self interpolate: topPix and: bottomPix frac: yFrac
! !

!ScratchPlugin methodsFor: 'serial port' stamp: 'jm (auto pragmas dtl 2014-01-04) 8/28/2005 10:57'!
primClose
	"Close the given serial port."

	| portNum |
	<export: true>
	portNum := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue: [^ nil].

	self cCode: 'SerialPortClose(portNum)'.

	interpreterProxy pop: 1.  "pop arg, leave rcvr on stack"
	^ 0
! !

!ScratchPlugin methodsFor: 'serial port' stamp: 'jm (auto pragmas dtl 2014-01-04) 4/8/2007 14:50'!
primGetOption
	"Return the given option value for the given serial port."

	| portNum attrNum result |
	<export: true>
	portNum := interpreterProxy stackIntegerValue: 1.
	attrNum := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue: [^ nil].

	result := self cCode: 'SerialPortGetOption(portNum, attrNum)'.
	result = -1 ifTrue: [interpreterProxy success: false. ^ 0].

	interpreterProxy pop: 3.  "pop args and rcvr, push result"
	interpreterProxy pushInteger: result.
	^ 0
! !

!ScratchPlugin methodsFor: 'serial port' stamp: 'jm (auto pragmas dtl 2014-01-04) 8/28/2005 10:57'!
primIsPortOpen
	"Answer the true if the given port is open."

	| portNum result |
	<export: true>
	portNum := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue: [^ 0].

	result := self cCode: 'SerialPortIsOpen(portNum)'.

	interpreterProxy pop: 2.  "pop arg and rcvr"
	interpreterProxy pushBool: result ~= 0.  "push result"
	^ 0
! !

!ScratchPlugin methodsFor: 'serial port' stamp: 'eem 3/16/2018 14:39'!
primOpenPortNamed
	"Open the port with the given name and baud rate."

	| nameStr src nameOop baudRate count portNum |
	<export: true>
	<var: 'nameStr' declareC: 'char nameStr[1000]'>
	<var: 'src' type: #'char *'>

	nameOop := interpreterProxy stackValue: 1.
	baudRate := interpreterProxy stackIntegerValue: 0.

	interpreterProxy success: (interpreterProxy isBytes: nameOop).

	interpreterProxy failed ifTrue: [^ 0].

	src := self cCoerce: (interpreterProxy firstIndexableField: nameOop) to: #'char *'.
	count := interpreterProxy stSizeOf: nameOop.
	0 to: count - 1 do: [:i | nameStr at: i put: (src at: i)].
	nameStr at: count put: 0.

	portNum := self SerialPortOpen: nameStr PortNamed: baudRate.
	portNum = -1 ifTrue: [interpreterProxy success: false. ^ 0].

	^interpreterProxy methodReturnInteger: portNum! !

!ScratchPlugin methodsFor: 'serial port' stamp: 'eem 3/16/2018 14:40'!
primPortCount
	"Answer the number of serial ports."

	| result |
	<export: true>

	result := self SerialPortCount.
	result = -1 ifTrue: [interpreterProxy success: false. ^ 0].

	^interpreterProxy methodReturnInteger: result
! !

!ScratchPlugin methodsFor: 'serial port' stamp: 'jm (auto pragmas dtl 2014-01-04) 8/28/2005 11:02'!
primPortName
	"Get the name for the port with the given number. Fail if the port number is greater than the number of available ports. Port numbering starts with 1."

	| portIndex nameStr count resultOop dst |
	<export: true>
	<var: 'nameStr' declareC: 'char nameStr[1000]'>
	<var: 'dst' declareC: 'char* dst'>

	portIndex := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue: [^ 0].

	self cCode: 'SerialPortName(portIndex, nameStr, 1000)'.

	count := self cCode: 'strlen(nameStr)'.
	count = 0 ifTrue: [
		interpreterProxy success: false.
		^ 0].

	resultOop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: count.
	dst := self cCoerce: (interpreterProxy firstIndexableField: resultOop) to: 'char *'.
	0 to: count - 1 do: [:i | dst at: i put: (nameStr at: i)].

	interpreterProxy pop: 2 thenPush: resultOop.  "pop arg and rcvr, push result"
	^ 0
! !

!ScratchPlugin methodsFor: 'serial port' stamp: 'eem 1/30/2014 15:24'!
primRead
	"Read data from the given serial port into the given buffer (a ByteArray or String). Answer the number of bytes read."

	| portNum bufOop bytesRead |
	<export: true>

	portNum := interpreterProxy stackIntegerValue: 1.
	bufOop := interpreterProxy stackValue: 0.

	interpreterProxy success: (interpreterProxy isBytes: bufOop).
	interpreterProxy failed ifTrue: [^ 0].

	bytesRead := self Serial: portNum
						Port: (self cCoerce: (interpreterProxy firstIndexableField: bufOop) to: 'char *')
						Read: (interpreterProxy stSizeOf: bufOop).

	interpreterProxy pop: 3.  					"pop args and rcvr"
	interpreterProxy pushInteger: bytesRead.	"push result"
	^ 0! !

!ScratchPlugin methodsFor: 'serial port' stamp: 'jm (auto pragmas dtl 2014-01-04) 4/8/2007 14:55'!
primSetOption
	"Return the given option value for the given serial port."

	| portNum attrNum attrValue result |
	<export: true>
	portNum := interpreterProxy stackIntegerValue: 2.
	attrNum := interpreterProxy stackIntegerValue: 1.
	attrValue := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue: [^ nil].

	result := self cCode: 'SerialPortSetOption(portNum, attrNum, attrValue)'.
	result = -1 ifTrue: [interpreterProxy success: false. ^ 0].

	interpreterProxy pop: 3.  "pop args; leave rcvr on stack"
	^ 0
! !

!ScratchPlugin methodsFor: 'serial port' stamp: 'eem 1/30/2014 15:25'!
primWrite
	"Write data to the given serial port from the given buffer (a ByteArray or String). Answer the number of bytes written."

	| portNum bufOop bytesWritten |
	<export: true>

	portNum := interpreterProxy stackIntegerValue: 1.
	bufOop := interpreterProxy stackValue: 0.

	interpreterProxy success: (interpreterProxy isBytes: bufOop).
	interpreterProxy failed ifTrue: [^ 0].

	bytesWritten := self Serial: portNum
						Port: (self cCoerce: (interpreterProxy firstIndexableField: bufOop) to: 'char *')
						Write: (interpreterProxy stSizeOf: bufOop).

	interpreterProxy pop: 3.  						"pop args and rcvr"
	interpreterProxy pushInteger: bytesWritten.	"push result"
	^ 0! !

!ScratchPlugin methodsFor: 'other filters' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/8/2006 05:30'!
primitiveBlur

	| inOop outOop width in out sz height n rTotal gTotal bTotal pix outPix |
	<export: true>
	<var: 'in' declareC: 'unsigned int *in'>
	<var: 'out' declareC: 'unsigned int *out'>

	inOop := interpreterProxy stackValue: 2.
	outOop := interpreterProxy stackValue: 1.
	width := interpreterProxy stackIntegerValue: 0.
	in := self checkedUnsignedIntPtrOf: inOop.
	out := self checkedUnsignedIntPtrOf: outOop.
	sz := interpreterProxy stSizeOf: inOop.
	interpreterProxy success: ((interpreterProxy stSizeOf: outOop) = sz).
	interpreterProxy failed ifTrue: [^ nil].

	height := sz // width.
	1 to: height - 2 do: [:y |
		1 to: width - 2 do: [:x |
			n := rTotal := gTotal := bTotal := 0.
			-1 to: 1 do: [:dY |
				-1 to: 1 do: [:dX |
					pix := (in at: ((y + dY) * width) + (x + dX) "add 1 when testing in Squeak") bitAnd: 16rFFFFFF.
					pix = 0 ifFalse: [  "skip transparent pixels"
						rTotal := rTotal + ((pix bitShift: -16) bitAnd: 16rFF).
						gTotal := gTotal + ((pix bitShift: -8) bitAnd: 16rFF).
						bTotal := bTotal + (pix bitAnd: 16rFF).
						n := n + 1]]].
			n = 0
				ifTrue: [outPix :=  0]
				ifFalse: [outPix := ((rTotal // n) bitShift: 16) + ((gTotal // n) bitShift: 8) +  (bTotal // n)].
			out at: ((y * width) + x "add 1 when testing in Squeak") put: outPix]].

	interpreterProxy pop: 3.  "pop args, leave rcvr on stack"
	^ 0
! !

!ScratchPlugin methodsFor: 'hsv filters' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/8/2006 05:30'!
primitiveBrightnessShift

	| inOop outOop shift in sz out pix r g b max min hue saturation brightness |
	<export: true>
	<var: 'in' declareC: 'unsigned int *in'>
	<var: 'out' declareC: 'unsigned int *out'>

	inOop := interpreterProxy stackValue: 2.
	outOop := interpreterProxy stackValue: 1.
	shift := interpreterProxy stackIntegerValue: 0.
	in := self checkedUnsignedIntPtrOf: inOop.
	sz := interpreterProxy stSizeOf: inOop.
	out := self checkedUnsignedIntPtrOf: outOop.
	interpreterProxy success: ((interpreterProxy stSizeOf: outOop) = sz).
	interpreterProxy failed ifTrue: [^ nil].

	0 to: sz - 1 do: [:i |
		pix := (in at: i) bitAnd: 16rFFFFFF.
		pix = 0 ifFalse: [  "skip pixel values of 0 (transparent)"
			r := (pix bitShift: -16) bitAnd: 16rFF.
			g := (pix bitShift: -8) bitAnd: 16rFF.
			b := pix bitAnd: 16rFF.

			"find min and max color components"
			max := min := r.
			g > max ifTrue: [max := g].
			b > max ifTrue: [max := b].
			g < min ifTrue: [min := g].
			b < min ifTrue: [min := b].

			"find current hue with range 0 to 360"
			hue := self hueFromR: r G: g B: b min: min max: max.

			"find current saturation and brightness with range 0 to 1000"
			max = 0 ifTrue: [saturation := 0] ifFalse: [saturation := ((max - min) * 1000) // max].
			brightness := (max * 1000) // 255.

			"compute new brigthness"
			brightness := brightness + (shift * 10).
			brightness > 1000 ifTrue: [brightness := 1000].
			brightness < 0 ifTrue: [brightness := 0].

			self bitmap: out at: i putH: hue s: saturation v: brightness]].

	interpreterProxy pop: 3.  "pop args, leave rcvr on stack"
	^ 0
! !

!ScratchPlugin methodsFor: 'sound' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/21/2006 21:21'!
primitiveCondenseSound

	| srcOop dstOop factor sz src dst count max v |
	<export: true>
	<var: 'src' declareC: 'short *src'>
	<var: 'dst' declareC: 'short *dst'>

	srcOop := interpreterProxy stackValue: 2.
	dstOop := interpreterProxy stackValue: 1.
	factor := interpreterProxy stackIntegerValue: 0.
	interpreterProxy success: (interpreterProxy isWords: srcOop).
	interpreterProxy success: (interpreterProxy isWords: dstOop).

	count := (2 * (interpreterProxy stSizeOf: srcOop)) // factor.
	sz := 2 * (interpreterProxy stSizeOf: dstOop).
	interpreterProxy success: (sz >= count).
	interpreterProxy failed ifTrue: [^ nil].

	src := self cCoerce: (interpreterProxy firstIndexableField: srcOop) to: 'short *'.
	dst := self cCoerce: (interpreterProxy firstIndexableField: dstOop) to: 'short *'.

	1 to: count do: [:i |
		max := 0.
		1 to: factor do: [:j |
			v := self cCode: '*src++'.
			v < 0 ifTrue: [v := 0 - v].
			v > max ifTrue: [max := v]].
		self cCode: '*dst++ = max'].

	interpreterProxy pop: 3.  "pop args, leave rcvr on stack"
	^ 0
! !

!ScratchPlugin methodsFor: 'scaling' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/8/2006 05:30'!
primitiveDoubleSize

	| in out inOop outOop inW inH outW outH dstX dstY baseIndex pix i |
	<export: true>
	<var: 'in' declareC: 'int *in'>
	<var: 'out' declareC: 'int *out'>

	inOop := interpreterProxy stackValue: 7.
	inW := interpreterProxy stackIntegerValue: 6.
	inH := interpreterProxy stackIntegerValue: 5.
	outOop := interpreterProxy stackValue: 4.
	outW := interpreterProxy stackIntegerValue: 3.
	outH := interpreterProxy stackIntegerValue: 2.
	dstX := interpreterProxy stackIntegerValue: 1.
	dstY := interpreterProxy stackIntegerValue: 0.

	in := self checkedUnsignedIntPtrOf: inOop.
	out := self checkedUnsignedIntPtrOf: outOop.
	interpreterProxy success: (dstX + (2 * inW)) < outW.
	interpreterProxy success: (dstY + (2 * inH)) < outH.
	interpreterProxy failed ifTrue: [^ nil].

	0 to: inH - 1 do: [:y |
		baseIndex := ((dstY + (2 * y)) * outW) + dstX.
		0 to: inW - 1 do: [:x |
			pix := in at: x + (y * inW).
			i := baseIndex + (2 * x).
			out at: i put: pix.
			out at: i + 1 put: pix.
			out at: i + outW put: pix.
			out at: i + outW + 1 put: pix]].

	interpreterProxy pop: 8.  "pop args, leave rcvr on stack"
	^ 0
! !

!ScratchPlugin methodsFor: 'sound' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/21/2006 20:53'!
primitiveExtractChannel

	| srcOop dstOop rightFlag sz src dst |
	<export: true>
	<var: 'src' declareC: 'short *src'>
	<var: 'dst' declareC: 'short *dst'>

	srcOop := interpreterProxy stackValue: 2.
	dstOop := interpreterProxy stackValue: 1.
	rightFlag := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy success: (interpreterProxy isWords: srcOop).
	interpreterProxy success: (interpreterProxy isWords: dstOop).

	sz := interpreterProxy stSizeOf: srcOop.
	interpreterProxy success: ((interpreterProxy stSizeOf: dstOop) >= (sz // 2)).
	interpreterProxy failed ifTrue: [^ nil].

	src := self cCoerce: (interpreterProxy firstIndexableField: srcOop) to: 'short *'.
	dst := self cCoerce: (interpreterProxy firstIndexableField: dstOop) to: 'short *'.

	rightFlag ifTrue: [self cCode: 'src++']. 
	1 to: sz do: [:i | self cCode: '*dst++ = *src; src += 2'].

	interpreterProxy pop: 3.  "pop args, leave rcvr on stack"
	^ 0
! !

!ScratchPlugin methodsFor: 'other filters' stamp: 'eem 1/30/2014 17:05'!
primitiveFisheye

	| inOop outOop width in out sz height centerX centerY dx dy ang pix power r srcX srcY scaledPower |
	<export: true>
	<var: 'in' type: #'unsigned int *'>
	<var: 'out' type: #'unsigned int *'>
	<var: 'dx' type: #double>
	<var: 'dy' type: #double>
	<var: 'ang' type: #double>
	<var: 'r' type: #double>
	<var: 'scaledPower' type: #double>

	inOop := interpreterProxy stackValue: 3.
	outOop := interpreterProxy stackValue: 2.
	width := interpreterProxy stackIntegerValue: 1.
	power := interpreterProxy stackIntegerValue: 0.
	in := self checkedUnsignedIntPtrOf: inOop.
	out := self checkedUnsignedIntPtrOf: outOop.
	sz := interpreterProxy stSizeOf: inOop.
	interpreterProxy success: ((interpreterProxy stSizeOf: outOop) = sz).
	interpreterProxy failed ifTrue: [^ nil].

	"calculate height, center, scales, radius, whirlRadians, and radiusSquared"
	height := sz // width.
	centerX := width // 2.
	centerY := height // 2.

	height := sz // width.
	centerX := width // 2.
	centerY := height // 2.
	scaledPower := power / 100.0.

	0 to: width - 1 do: [:x |
		0 to: height - 1 do: [:y |
			dx := (x - centerX) / centerX asFloat.
			dy := (y - centerY) / centerY asFloat.
			r := ((dx * dx) + (dy * dy)) sqrt raisedTo: scaledPower.
			r <= 1.0
				ifTrue: [
					ang := self cCode: 'atan2(dy,dx)'.
					srcX := (1024 * (centerX + ((r * ang cos) * centerX))) asInteger.
					srcY := (1024 * (centerY + ((r * ang sin) * centerY))) asInteger]
				ifFalse: [
					srcX := 1024 * x.
					srcY := 1024 * y].
			pix := self interpolatedFrom: in
					x: srcX
					y: srcY
					width: width
					height: height.
			out at: ((y * width) + x "+ 1 for Squeak") put: pix]].

	interpreterProxy pop: 4.  "pop args, leave rcvr on stack"
	^ 0
! !

!ScratchPlugin methodsFor: 'os functions' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/8/2006 17:47'!
primitiveGetFolderPath
	"Get the path for the special folder with given ID. Fail if the folder ID is out of range."

	| nameStr dst folderID count resultOop |
	<export: true>
	<var: 'nameStr' declareC: 'char nameStr[2000]'>
	<var: 'dst' declareC: 'char* dst'>

	folderID := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue: [^ 0].

	self cCode: 'GetFolderPathForID(folderID, nameStr, 2000)'.

	count := self cCode: 'strlen(nameStr)'.
	resultOop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: count.
	dst := self cCoerce: (interpreterProxy firstIndexableField: resultOop) to: 'char *'.
	0 to: count - 1 do: [:i | dst at: i put: (nameStr at: i)].

	interpreterProxy pop: 2 thenPush: resultOop.  "pop arg and rcvr, push result"
	^ 0
! !

!ScratchPlugin methodsFor: 'scaling' stamp: 'jm (auto pragmas dtl 2014-01-04) 1/1/2009 17:08'!
primitiveHalfSizeAverage

	| in inW inH out outW outH srcX srcY dstX dstY dstW dstH srcIndex dstIndex pixel r g b |
	<export: true>
	<var: 'in' declareC: 'int *in'>
	<var: 'out' declareC: 'int *out'>

	in := self checkedUnsignedIntPtrOf: (interpreterProxy stackValue: 11).
	inW := interpreterProxy stackIntegerValue: 10.
	inH := interpreterProxy stackIntegerValue: 9.
	out := self checkedUnsignedIntPtrOf: (interpreterProxy stackValue: 8).
	outW := interpreterProxy stackIntegerValue: 7.
	outH := interpreterProxy stackIntegerValue: 6.
	srcX := interpreterProxy stackIntegerValue: 5.
	srcY := interpreterProxy stackIntegerValue: 4.
	dstX := interpreterProxy stackIntegerValue: 3.
	dstY := interpreterProxy stackIntegerValue: 2.
	dstW := interpreterProxy stackIntegerValue: 1.
	dstH := interpreterProxy stackIntegerValue: 0.

	interpreterProxy success: (srcX >= 0) & (srcY >= 0).
	interpreterProxy success: (srcX + (2 * dstW)) <= inW.
	interpreterProxy success: (srcY + (2 * dstH)) <= inH.
	interpreterProxy success: (dstX >= 0) & (dstY >= 0).
	interpreterProxy success: (dstX + dstW) <= outW.
	interpreterProxy success: (dstY + dstH) <= outH.
	interpreterProxy failed ifTrue: [^ nil].

	0 to: dstH - 1 do: [:y |
		srcIndex := (inW * (srcY + (2 * y))) + srcX.
		dstIndex := (outW * (dstY + y)) + dstX.
		0 to: dstW - 1 do: [:x |
			pixel := in at: srcIndex.
			r := pixel bitAnd: 16rFF0000.
			g := pixel bitAnd: 16rFF00.
			b := pixel bitAnd: 16rFF.

			pixel := in at: srcIndex + 1.
			r := r + (pixel bitAnd: 16rFF0000).
			g := g + (pixel bitAnd: 16rFF00).
			b := b + (pixel bitAnd: 16rFF).

			pixel := in at: srcIndex + inW.
			r := r + (pixel bitAnd: 16rFF0000).
			g := g + (pixel bitAnd: 16rFF00).
			b := b + (pixel bitAnd: 16rFF).

			pixel := in at: srcIndex + inW + 1.
			r := r + (pixel bitAnd: 16rFF0000).
			g := g + (pixel bitAnd: 16rFF00).
			b := b + (pixel bitAnd: 16rFF).

			"store combined RGB into target bitmap"
			out at: dstIndex put:
				(((r bitShift: -2) bitAnd: 16rFF0000) bitOr:
				(((g bitShift: -2) bitAnd: 16rFF00) bitOr: (b bitShift: -2))).

			srcIndex := srcIndex + 2.
			dstIndex := dstIndex + 1]].

	interpreterProxy pop: 12.  "pop args, leave rcvr on stack"
	^ 0
! !

!ScratchPlugin methodsFor: 'scaling' stamp: 'jm (auto pragmas dtl 2014-01-04) 1/1/2009 17:08'!
primitiveHalfSizeDiagonal

	| in inW inH out outW outH srcX srcY dstX dstY dstW dstH srcIndex dstIndex p1 p2 r g b |
	<export: true>
	<var: 'in' declareC: 'int *in'>
	<var: 'out' declareC: 'int *out'>

	in := self checkedUnsignedIntPtrOf: (interpreterProxy stackValue: 11).
	inW := interpreterProxy stackIntegerValue: 10.
	inH := interpreterProxy stackIntegerValue: 9.
	out := self checkedUnsignedIntPtrOf: (interpreterProxy stackValue: 8).
	outW := interpreterProxy stackIntegerValue: 7.
	outH := interpreterProxy stackIntegerValue: 6.
	srcX := interpreterProxy stackIntegerValue: 5.
	srcY := interpreterProxy stackIntegerValue: 4.
	dstX := interpreterProxy stackIntegerValue: 3.
	dstY := interpreterProxy stackIntegerValue: 2.
	dstW := interpreterProxy stackIntegerValue: 1.
	dstH := interpreterProxy stackIntegerValue: 0.

	interpreterProxy success: (srcX >= 0) & (srcY >= 0).
	interpreterProxy success: (srcX + (2 * dstW)) <= inW.
	interpreterProxy success: (srcY + (2 * dstH)) <= inH.
	interpreterProxy success: (dstX >= 0) & (dstY >= 0).
	interpreterProxy success: (dstX + dstW) <= outW.
	interpreterProxy success: (dstY + dstH) <= outH.
	interpreterProxy failed ifTrue: [^ nil].

	0 to: dstH - 1 do: [:y |
		srcIndex := (inW * (srcY + (2 * y))) + srcX.
		dstIndex := (outW * (dstY + y)) + dstX.
		0 to: dstW - 1 do: [:x |
			p1 := in at: srcIndex.
			p2 := in at: srcIndex + inW + 1.

			r := (((p1 bitAnd: 16rFF0000) + (p2 bitAnd: 16rFF0000)) bitShift: -1) bitAnd: 16rFF0000.
			g := (((p1 bitAnd: 16rFF00) + (p2 bitAnd: 16rFF00)) bitShift: -1) bitAnd: 16rFF00.
			b := ((p1 bitAnd: 16rFF) + (p2 bitAnd: 16rFF)) bitShift: -1.

			"store combined RGB into target bitmap"
			out at: dstIndex put: (r bitOr: (g bitOr: b)).

			srcIndex := srcIndex + 2.
			dstIndex := dstIndex + 1]].

	interpreterProxy pop: 12.  "pop args, leave rcvr on stack"
	^ 0
! !

!ScratchPlugin methodsFor: 'hsv filters' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/8/2006 05:30'!
primitiveHueShift

	| inOop outOop shift in sz out pix r g b max min brightness saturation hue |
	<export: true>
	<var: 'in' declareC: 'unsigned int *in'>
	<var: 'out' declareC: 'unsigned int *out'>

	inOop := interpreterProxy stackValue: 2.
	outOop := interpreterProxy stackValue: 1.
	shift := interpreterProxy stackIntegerValue: 0.
	in := self checkedUnsignedIntPtrOf: inOop.
	sz := interpreterProxy stSizeOf: inOop.
	out := self checkedUnsignedIntPtrOf: outOop.
	interpreterProxy success: ((interpreterProxy stSizeOf: outOop) = sz).
	interpreterProxy failed ifTrue: [^ nil].

	0 to: sz - 1 do: [:i |
		pix := (in at: i) bitAnd: 16rFFFFFF.
		pix = 0 ifFalse: [  "skip pixel values of 0 (transparent)"
			r := (pix bitShift: -16) bitAnd: 16rFF.
			g := (pix bitShift: -8) bitAnd: 16rFF.
			b := pix bitAnd: 16rFF.

			"find min and max color components"
			max := min := r.
			g > max ifTrue: [max := g].
			b > max ifTrue: [max := b].
			g < min ifTrue: [min := g].
			b < min ifTrue: [min := b].

			"find current brightness (v) and  saturation with range 0 to 1000"
			brightness := (max * 1000) // 255.
			max = 0 ifTrue: [saturation := 0] ifFalse: [saturation := ((max - min) * 1000) // max].

			brightness < 110 ifTrue: [					"force black to a very dark, saturated gray"
				brightness := 110. saturation := 1000].	
			saturation < 90 ifTrue: [saturation := 90].		"force a small color change on grays"
			((brightness = 110) | (saturation = 90))		"tint all blacks and grays the same"
				ifTrue: [hue := 0]
				ifFalse: [hue := self hueFromR: r G: g B: b min: min max: max].

			hue := (hue + shift + 360000000) \\ 360.  "compute new hue"
			self bitmap: out at: i putH: hue s: saturation v: brightness]].

	interpreterProxy pop: 3.  "pop args, leave rcvr on stack"
	^ 0
! !

!ScratchPlugin methodsFor: 'bilinear interpolation' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/8/2006 05:30'!
primitiveInterpolate

	| inOop xFixed yFixed width in sz result |
	<export: true>
	<var: 'in' declareC: 'unsigned int *in'>

	inOop := interpreterProxy stackValue: 3.
	width := interpreterProxy stackIntegerValue: 2.
	xFixed := interpreterProxy stackIntegerValue: 1.
	yFixed := interpreterProxy stackIntegerValue: 0.
	in := self checkedUnsignedIntPtrOf: inOop.
	sz := interpreterProxy stSizeOf: inOop.
	interpreterProxy failed ifTrue: [^ nil].

	result := self interpolatedFrom: in x: xFixed y: yFixed width: width height: sz // width.

	interpreterProxy pop: 5.  "pop args and rcvr"
	interpreterProxy pushInteger: result.
	^ 0
! !

!ScratchPlugin methodsFor: 'os functions' stamp: 'eem 1/30/2014 15:19'!
primitiveIsHidden
	"Answer true if the file or folder with the given path should be hidden from the user. On Windows, this is the value of the 'hidden' file property."

	| pathOop src count fullPath result |
	<export: true>
	<var: 'fullPath' declareC: 'char fullPath[1000]'>
	<var: 'src' type: #'char *'>

	pathOop := interpreterProxy stackValue: 0.

	interpreterProxy success: (interpreterProxy isBytes: pathOop).

	interpreterProxy failed ifTrue: [^ 0].

	src := self cCoerce: (interpreterProxy firstIndexableField: pathOop) to: #'char *'.
	count := interpreterProxy stSizeOf: pathOop.
	count >= 1000 ifTrue: [interpreterProxy success: false. ^ 0].
	0 to: count - 1 do: [:i | fullPath at: i put: (src at: i)].
	fullPath at: count put: 0.

	result := self IsFileOrFolderHidden: fullPath.

	interpreterProxy pop: 2.  "pop arg and rcvr"
	interpreterProxy pushBool: result ~= 0.  "push result"
	^ 0! !

!ScratchPlugin methodsFor: 'os functions' stamp: 'eem 1/30/2014 15:19'!
primitiveOpenURL
	"Open a web browser on the given URL."

	| urlStr src urlOop count |
	<export: true>
	<var: 'urlStr' declareC: 'char urlStr[2000]'>
	<var: 'src' type: #'char *'>

	urlOop := interpreterProxy stackValue: 0.

	interpreterProxy success: (interpreterProxy isBytes: urlOop).

	interpreterProxy failed ifTrue: [^ 0].

	src := self cCoerce: (interpreterProxy firstIndexableField: urlOop) to: #'char *'.
	count := interpreterProxy stSizeOf: urlOop.
	count >= 2000 ifTrue: [interpreterProxy success: false. ^ 0].
	0 to: count - 1 do: [:i | urlStr at: i put: (src at: i)].
	urlStr at: count put: 0.

	self OpenURL: urlStr.

	interpreterProxy pop: 1.  "pop arg, leave rcvr on stack"
	^ 0! !

!ScratchPlugin methodsFor: 'hsv filters' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/8/2006 05:30'!
primitiveSaturationShift

	| inOop outOop shift in sz out pix r g b max min brightness saturation hue |
	<export: true>
	<var: 'in' declareC: 'unsigned int *in'>
	<var: 'out' declareC: 'unsigned int *out'>

	inOop := interpreterProxy stackValue: 2.
	outOop := interpreterProxy stackValue: 1.
	shift := interpreterProxy stackIntegerValue: 0.
	in := self checkedUnsignedIntPtrOf: inOop.
	sz := interpreterProxy stSizeOf: inOop.
	out := self checkedUnsignedIntPtrOf: outOop.
	interpreterProxy success: ((interpreterProxy stSizeOf: outOop) = sz).
	interpreterProxy failed ifTrue: [^ nil].

	0 to: sz - 1 do: [:i |
		pix := (in at: i) bitAnd: 16rFFFFFF.
		pix < 2 ifFalse: [  "skip pixel values of 0 (transparent) and 1 (black)"
			r := (pix bitShift: -16) bitAnd: 16rFF.
			g := (pix bitShift: -8) bitAnd: 16rFF.
			b := pix bitAnd: 16rFF.

			"find min and max color components"
			max := min := r.
			g > max ifTrue: [max := g].
			b > max ifTrue: [max := b].
			g < min ifTrue: [min := g].
			b < min ifTrue: [min := b].

			"find current brightness (v) and  saturation with range 0 to 1000"
			brightness := (max * 1000) // 255.
			max = 0 ifTrue: [saturation := 0] ifFalse: [saturation := ((max - min) * 1000) // max].

			saturation > 0 ifTrue: [  "do nothing if pixel is unsaturated (gray)"
				hue := self hueFromR: r G: g B: b min: min max: max.

				"compute new saturation"
				saturation := saturation + (shift * 10).
				saturation > 1000 ifTrue: [saturation := 1000].
				saturation < 0 ifTrue: [saturation := 0].
				self bitmap: out at: i putH: hue s: saturation v: brightness]]].

	interpreterProxy pop: 3.  "pop args, leave rcvr on stack"
	^ 0
! !

!ScratchPlugin methodsFor: 'scaling' stamp: 'tpr 12/10/2014 13:09'!
primitiveScale
	"Scale using bilinear interpolation."
	"This version isn't really able to do much with full ARGB based images; the A channel will be ignored and only fully transparent pixels will be treated as transparent. The output pixel will be either fully transparent or fully opaque."

	| inOop inW inH outOop outW outH in out inX inY xIncr yIncr outPix w1 w2 w3 w4 t p1 p2 p3 p4 tWeight |
	<export: true>
	<var: 'in' declareC: 'int *in'>
	<var: 'out' declareC: 'int *out'>

	inOop := interpreterProxy stackValue: 5.
	inW := interpreterProxy stackIntegerValue: 4.
	inH := interpreterProxy stackIntegerValue: 3.
	outOop := interpreterProxy stackValue: 2.
	outW := interpreterProxy stackIntegerValue: 1.
	outH := interpreterProxy stackIntegerValue: 0.

	interpreterProxy success: (interpreterProxy stSizeOf: inOop) = (inW * inH).
	interpreterProxy success: (interpreterProxy stSizeOf: outOop) = (outW * outH).
	in := self checkedUnsignedIntPtrOf: inOop.
	out := self checkedUnsignedIntPtrOf: outOop.
	interpreterProxy failed ifTrue: [^ nil].

	inX := inY := 0.					"source x and y, scaled by 1024"
	xIncr := (inW * 1024) // outW.	"source x increment, scaled by 1024"
	yIncr := (inH * 1024) // outH.		"source y increment, scaled by 1024"

	0 to: (outH - 1) do: [:outY |
		inX := 0.
		0 to: (outW - 1) do: [:outX |
			"compute weights, scaled by 2^20"
			w1 := (1024 - (inX bitAnd: 1023))	* (1024 - (inY bitAnd: 1023)).
			w2 := (inX bitAnd: 1023)			* (1024 - (inY bitAnd: 1023)).
			w3 := (1024 - (inX bitAnd: 1023))	* (inY bitAnd: 1023).
			w4 := (inX bitAnd: 1023)			* (inY bitAnd: 1023).

			"get source pixels"
			t := ((inY >> 10) * inW) + (inX >> 10).
			p1 := in at: t.
			((inX >> 10) < (inW - 1)) ifTrue: [p2 := in at: t + 1] ifFalse: [p2 := p1].
			(inY >> 10) < (inH - 1) ifTrue: [t := t + inW].  "next row"
			p3 := in at: t.
			((inX >> 10) < (inW - 1)) ifTrue: [p4 := in at: t + 1] ifFalse: [p4 := p3].

			"deal with transparent pixels"
			"Note as above - only transparent not translucent"
			tWeight := 0.
			p1 = 0 ifTrue: [p1 := p2. tWeight := tWeight + w1].
			p2 = 0 ifTrue: [p2 := p1. tWeight := tWeight + w2].
			p3 = 0 ifTrue: [p3 := p4. tWeight := tWeight + w3].
			p4 = 0 ifTrue: [p4 := p3. tWeight := tWeight + w4].
			p1 = 0 ifTrue: [p1 := p3. p2 := p4].  "both top pixels were transparent; use bottom row"
			p3 = 0 ifTrue: [p3 := p1. p4 := p2].  "both bottom pixels were transparent; use top row"

			outPix := 0.
			tWeight < 500000 ifTrue: [  "compute an (opaque) output pixel if less than 50% transparent"
				t := (w1 * ((p1 >> 16) bitAnd: 255)) + (w2 * ((p2 >> 16) bitAnd: 255)) + (w3 * ((p3 >> 16) bitAnd: 255)) + (w4 * ((p4 >> 16) bitAnd: 255)).
				outPix := ((t >> 20) bitAnd: 255) << 16.
				t := (w1 * ((p1 >> 8) bitAnd: 255)) + (w2 * ((p2 >> 8) bitAnd: 255)) + (w3 * ((p3 >> 8) bitAnd: 255)) + (w4 * ((p4 >> 8) bitAnd: 255)).
				outPix := outPix bitOr: (((t >> 20) bitAnd: 255) << 8).
				t := (w1 * (p1 bitAnd: 255)) + (w2 * (p2 bitAnd: 255)) + (w3 * (p3 bitAnd: 255)) + (w4 * (p4 bitAnd: 255)).
				outPix := outPix bitOr: ((t >> 20) bitAnd: 255).
				"If the result is black, remember to make it Squeak-standard-fake-black"
				outPix = 0 ifTrue: [outPix := 1].
				"add the A channel to make it really opaque"
				outPix := outPix bitOr: 16rFF000000].

			out at: (outY * outW) + outX put: outPix.
			inX := inX + xIncr].
		inY := inY + yIncr].

	interpreterProxy pop: 6.  "pop args, leave rcvr on stack"
	^ 0
! !

!ScratchPlugin methodsFor: 'os functions' stamp: 'eem 1/30/2014 15:19'!
primitiveSetUnicodePasteBuffer
	"Set the Mac OS X Unicode paste buffer."

	| utf16 strOop count |
	<export: true>
	<var: 'utf16' declareC: 'short *utf16'>

	strOop := interpreterProxy stackValue: 0.

	interpreterProxy success: (interpreterProxy isBytes: strOop).

	interpreterProxy failed ifTrue: [^ 0].

	utf16 := self cCoerce: (interpreterProxy firstIndexableField: strOop) to: #'short *'.
	count := interpreterProxy stSizeOf: strOop.

	self SetUnicodePaste: utf16 Buffer: count.

	interpreterProxy pop: 1.  "pop arg, leave rcvr on stack"
	^ 0! !

!ScratchPlugin methodsFor: 'os functions' stamp: 'eem 1/30/2014 15:19'!
primitiveSetWindowTitle
	"Set the title of the Scratch window."

	| titleStr src titleOop count |
	<export: true>
	<var: 'titleStr' declareC: 'char titleStr[1000]'>
	<var: 'src' type: #'char *'>

	titleOop := interpreterProxy stackValue: 0.

	interpreterProxy success: (interpreterProxy isBytes: titleOop).

	interpreterProxy failed ifTrue: [^ 0].

	src := self cCoerce: (interpreterProxy firstIndexableField: titleOop) to: #'char *'.
	count := interpreterProxy stSizeOf: titleOop.
	count >= 1000 ifTrue: [interpreterProxy success: false. ^ 0].
	0 to: count - 1 do: [:i | titleStr at: i put: (src at: i)].
	titleStr at: count put: 0.

	self SetScratchWindowTitle: titleStr.

	interpreterProxy pop: 1.  "pop arg, leave rcvr on stack"
	^ 0! !

!ScratchPlugin methodsFor: 'os functions' stamp: 'eem 1/30/2014 15:19'!
primitiveShortToLongPath
	"On Windows, convert a short file/path name into a long one. Fail on other platforms."

	| shortPath longPath ptr shortPathOop result count resultOop |
	<export: true>
	<var: 'shortPath' declareC: 'char shortPath[1000]'>
	<var: 'longPath' declareC: 'char longPath[1000]'>
	<var: 'ptr' type: #'char *'>

	shortPathOop := interpreterProxy stackValue: 0.

	(interpreterProxy isBytes: shortPathOop) ifFalse:
		[interpreterProxy success: false. ^ 0].

	ptr := self cCoerce: (interpreterProxy firstIndexableField: shortPathOop) to: #'char *'.
	count := interpreterProxy stSizeOf: shortPathOop.
	count >= 1000 ifTrue: [interpreterProxy success: false. ^ 0].
	0 to: count - 1 do: [:i | shortPath at: i put: (ptr at: i)].
	shortPath at: count put: 0.

	result := self cCode: 'WinShortToLongPath(shortPath, longPath, 1000)'.
	result = -1 ifTrue: [interpreterProxy success: false. ^ 0].

	count := self strlen: longPath.
	resultOop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: count.
	ptr := self cCoerce: (interpreterProxy firstIndexableField: resultOop) to: #'char *'.
	0 to: count - 1 do: [:i | ptr at: i put: (longPath at: i)].

	interpreterProxy pop: 2 thenPush: resultOop.  "pop arg and rcvr, push result"
	^ 0! !

!ScratchPlugin methodsFor: 'other filters' stamp: 'eem 1/6/2014 18:39'!
primitiveWaterRipples1
 
	| in out aArray bArray ripply temp pix dx dy dist inOop outOop width allPix aArOop bArOop height t1 blops x y power val val2 dx2 dy2 newLoc |
	<export: true>
	<var: 'in' declareC: 'unsigned int *in'>
	<var: 'out' declareC: 'unsigned int *out'>
	<var: 'aArray' declareC: 'double *aArray'>
	<var: 'bArray' declareC: 'double *bArray'>
	<var: 'ripply' declareC: 'int ripply'>
	<var: 'temp' declareC: 'double temp'>
	<var: 'pix' declareC: 'unsigned int pix'>
	<var: 'dist' declareC: 'double dist'>
	<var: 'dx2' declareC: 'double dx2'>
	<var: 'dy2' declareC: 'double dy2'>

	inOop := interpreterProxy stackValue: 5.
	outOop := interpreterProxy stackValue: 4.
	width := interpreterProxy stackIntegerValue: 3.
	in := self checkedUnsignedIntPtrOf: inOop.
	out := self checkedUnsignedIntPtrOf: outOop.
	allPix := interpreterProxy stSizeOf: inOop.
	ripply := interpreterProxy stackIntegerValue: 2.
	aArOop := interpreterProxy stackValue: 1.
	bArOop := interpreterProxy stackValue: 0.
	aArray := self checkedFloatPtrOf: aArOop.
	bArray := self checkedFloatPtrOf: bArOop.
	interpreterProxy success: ((interpreterProxy stSizeOf: outOop) = allPix).
	interpreterProxy failed ifTrue: [^ nil].

	height := allPix // width.

	t1 := self cCode: 'rand()'.
	blops := t1 \\ ripply -1.
	0 to: blops /2-1 do: [:t |
		t1 := self cCode: 'rand()'.
		x := t1 \\ width.
		t1 := self cCode: 'rand()'.
		y := t1 \\ height.
		t1 := self cCode: 'rand()'.
		power := t1 \\ 8.
		-4 to: 4 do: [:g |
			-4 to: 4 do: [:h |
				dist := ((g*g) + (h*h)) asFloat.
				((dist < 25) and: [dist > 0]) ifTrue: [
						dx := (x + g) asInteger.
						dy := (y + h) asInteger.
						((dx >0) and: [(dy>0) and: [(dy < height) and: [dx < width]]]) ifTrue: [
							aArray at: ((dy)*width + dx) put: (power *(1.0 asFloat -(dist/(25.0 asFloat))) asFloat).
						].
					].
				].
			].
		].
	
		1 to: width -2 do: [:f |
			1 to: height -2 do: [:d |
			val := (d)*width + f.
			aArray at: val put: (((
				(bArray at: (val+1)) + (bArray at: (val-1)) + (bArray at: (val + width)) + (bArray at: (val - width)) +
				((bArray at: (val -1 -width))/2) + ((bArray at: (val-1+width))/2) + ((bArray at: (val+1-width))/2) + ((bArray at: (val+1+width))/2)) /4) - (aArray at: (val))).
			aArray at: (val) put: ((aArray at: (val))*(0.9 asFloat)).
			].
		].
	
		"temp := bArray.
		bArray := aArray.
		aArray := temp."
		0 to: width*height do: [:q |
			temp := bArray at: q.
			bArray at: q put: (aArray at: q).
			aArray at: q put: temp.
		].

		0 to: height-1 do: [:j |
			0 to: width-1 do: [:i |
				((i > 1) and: [(i<(width-1)) and: [(j>1) and: [(j<(height-1))]]]) ifTrue: [
					val2 := (j)*width + i.
					dx2 := ((((aArray at: (val2)) - (aArray at: (val2-1))) + ((aArray at: (val2+1)) - (aArray at: (val2)))) *64) asFloat.
					dy2 := ((((aArray at: (val2)) - (aArray at: (val2-width))) + ((aArray at: (val2+width)) - (aArray at: (val2)))) /64) asFloat.
					(dx2 < -2) ifTrue: [dx2 := -2].
					(dx2 >  2) ifTrue: [dx2 :=  2].
					(dy2 < -2) ifTrue: [dy2 := -2].
					(dy2 >  2) ifTrue: [dy2 :=  2].
					newLoc := ((j+dy2)*width + (i+dx2)) asInteger.
					((newLoc < (width*height)) and: [newLoc >=0]) ifTrue: [
						pix := in at: newLoc]
					ifFalse: [
						pix := in at: (i +(j*width)) ].
				]
				ifFalse: [
					pix := in at: (i +(j*width)) ].
			out at: (i + (j*width)) put: pix.
		]].

	interpreterProxy pop: 6.  "pop args, leave rcvr on stack"
	^ 0
! !

!ScratchPlugin methodsFor: 'other filters' stamp: 'jm (auto pragmas dtl 2014-01-04) 11/8/2006 05:30'!
primitiveWhirl

	| inOop outOop width degrees in out sz height centerX centerY radius scaleX scaleY whirlRadians radiusSquared dx dy d factor ang sina cosa pix |
	<export: true>
	<var: 'in' declareC: 'unsigned int *in'>
	<var: 'out' declareC: 'unsigned int *out'>
	<var: 'scaleX' declareC: 'double scaleX'>
	<var: 'scaleY' declareC: 'double scaleY'>
	<var: 'whirlRadians' declareC: 'double whirlRadians'>
	<var: 'radiusSquared' declareC: 'double radiusSquared'>
	<var: 'dx' declareC: 'double dx'>
	<var: 'dy' declareC: 'double dy'>
	<var: 'd' declareC: 'double d'>
	<var: 'factor' declareC: 'double factor'>
	<var: 'ang' declareC: 'double ang'>
	<var: 'sina' declareC: 'double sina'>
	<var: 'cosa' declareC: 'double cosa'>

	inOop := interpreterProxy stackValue: 3.
	outOop := interpreterProxy stackValue: 2.
	width := interpreterProxy stackIntegerValue: 1.
	degrees := interpreterProxy stackIntegerValue: 0.
	in := self checkedUnsignedIntPtrOf: inOop.
	out := self checkedUnsignedIntPtrOf: outOop.
	sz := interpreterProxy stSizeOf: inOop.
	interpreterProxy success: ((interpreterProxy stSizeOf: outOop) = sz).
	interpreterProxy failed ifTrue: [^ nil].

	"calculate height, center, scales, radius, whirlRadians, and radiusSquared"
	height := sz // width.
	centerX := width // 2.
	centerY := height // 2.
	centerX < centerY
		ifTrue: [
			radius := centerX.
			scaleX := centerY asFloat / centerX. 
			scaleY := 1.0]
		ifFalse: [
			radius := centerY.
			scaleX := 1.0.
			centerY < centerX
				ifTrue: [scaleY := centerX asFloat / centerY]
				ifFalse: [scaleY := 1.0]].
	whirlRadians := (-3.141592653589793 * degrees) / 180.0.
	radiusSquared := (radius * radius) asFloat.

	0 to: width - 1 do: [:x |
		0 to: height - 1 do: [:y |
			dx := scaleX * (x - centerX) asFloat.
			dy := scaleY * (y - centerY) asFloat.
			d := (dx * dx) + (dy * dy).
			d < radiusSquared ifTrue: [  "inside the whirl circle"
				factor := 1.0 - (d sqrt / radius).
				ang := whirlRadians * (factor * factor).
				sina := ang sin.
				cosa := ang cos.
				pix := self interpolatedFrom: in
					x: (1024.0 * ((((cosa * dx) - (sina * dy)) / scaleX) + centerX)) asInteger
					y: (1024.0 * ((((sina * dx) + (cosa * dy)) / scaleY) + centerY)) asInteger
					width: width
					height: height.
				out at: ((width * y) + x "for Squeak: + 1") put: pix]]].

	interpreterProxy pop: 4.  "pop args, leave rcvr on stack"
	^ 0
! !

InterpreterPlugin subclass: #SecurityPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!SecurityPlugin commentStamp: 'tpr 5/5/2003 12:19' prior: 0!
IMPLEMENTATION NOTES:
The support should assume a trusted directory based on which access to files is granted when running in restricted mode. If necessary, links need to be resolved before granting access (currently, this applies only to platforms on which links can be created by simply creating the right kind of file).

The untrusted user directory returned MUST be different from the image and VM location. Otherwise a Badlet could attempt to overwrite these by using simple file primitives. The secure directory location returned by the primitive is a place to store per-user security information. Again, this place needs to be outside the untrusted space. Examples:

[Windows]
	* VM+image location: "C:\Program Files\Squeak\"
	* secure directory location: "C:\Program Files\Squeak\username\"
	* untrusted user directory: "C:\My Squeak\username\"
[Unix]
	* VM+image location: "/user/local/squeak"
	* secure directory location: "~username/.squeak/
	* untrusted user directory: "~username/squeak/"
[Mac]
	* plugin VM location: "MacHD:Netscape:Plugins:"
	* standalone VM and image location: "MacHD:Squeak:"
	* secure directory location: "MacHD:Squeak:username:"
	* untrusted user directory: "MacHD:My Squeak:username:"

Restoring the rights revoked by an image might be possible by some interaction with the VM directly. Any such action should be preceeded by a BIG FAT WARNING - the average user will never need that ability (if she does, we did something wrong) so this is a last resort in cases where something fundamtally tricky happened.
!

!SecurityPlugin class methodsFor: 'translation' stamp: 'tpr 5/23/2001 17:10'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!SecurityPlugin class methodsFor: 'translation' stamp: 'tpr 3/20/2001 12:33'!
requiresPlatformFiles
	"default is ok for most, any plugin needing platform specific files must say so"
	^true! !

!SecurityPlugin methodsFor: 'initialize' stamp: 'JMM (auto pragmas 12/08) 8/15/2001 11:59'!
initialiseModule
	<export: true>
	^self cCode: 'ioInitSecurity()' inSmalltalk:[true]! !

!SecurityPlugin methodsFor: 'simulation' stamp: 'eem 1/14/2018 11:11'!
ioCanRenameImage
	<doNotGenerate>
	^true! !

!SecurityPlugin methodsFor: 'simulation' stamp: 'eem 8/14/2020 14:28'!
ioCanWriteImage
	<doNotGenerate>
	^true! !

!SecurityPlugin methodsFor: 'simulation' stamp: 'eem 8/14/2020 14:29'!
ioDisableImageWrite
	<doNotGenerate>! !

!SecurityPlugin methodsFor: 'simulation' stamp: 'eem 8/14/2020 14:33'!
ioGetSecureUserDirectory
	<doNotGenerate>
	^SecurityManager default primSecureUserDirectory! !

!SecurityPlugin methodsFor: 'simulation' stamp: 'eem 8/14/2020 14:35'!
ioGetUntrustedUserDirectory
	<doNotGenerate>
	^SecurityManager default primUntrustedUserDirectory! !

!SecurityPlugin methodsFor: 'primitives' stamp: 'eem 8/14/2020 14:28'!
primitiveCanWriteImage
	<export: true>
	interpreterProxy methodReturnBool: self ioCanWriteImage! !

!SecurityPlugin methodsFor: 'primitives' stamp: 'eem 8/14/2020 14:29'!
primitiveDisableImageWrite
	<export: true>
	self ioDisableImageWrite! !

!SecurityPlugin methodsFor: 'primitives' stamp: 'eem 8/14/2020 14:30'!
primitiveGetSecureUserDirectory
	"Primitive. Return the secure directory for the current user."
	| dirName |
	<export: true>
	<var: #dirName type: #'char *'>
	dirName := self ioGetSecureUserDirectory.
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnString: dirName]! !

!SecurityPlugin methodsFor: 'primitives' stamp: 'eem 8/14/2020 14:34'!
primitiveGetUntrustedUserDirectory
	"Primitive. Return the untrusted user directory name."
	| dirName |
	<export: true>
	<var: #dirName type: #'char *'>
	dirName := self ioGetUntrustedUserDirectory.
	interpreterProxy failed ifFalse:
		[interpreterProxy methodReturnString: dirName]! !

!SecurityPlugin methodsFor: 'exported functions' stamp: '(auto pragmas 12/08) '!
secCan: socket ListenOnPort: port
	<export: true>
	^self cCode: 'ioCanListenOnPort(socket, port)'! !

!SecurityPlugin methodsFor: 'exported functions' stamp: '(auto pragmas 12/08) '!
secCanConnect: addr ToPort: port
	<export: true>
	^self cCode: 'ioCanConnectToPort(addr, port)'! !

!SecurityPlugin methodsFor: 'exported functions' stamp: '(auto pragmas 12/08) '!
secCanCreate: netType SocketOfType: socketType
	<export: true>
	^self cCode: 'ioCanCreateSocketOfType(netType, socketType)'! !

!SecurityPlugin methodsFor: 'exported functions' stamp: '(auto pragmas 12/08) '!
secCanCreatePath: dirName OfSize: dirNameSize
	<export: true>
	<var: #dirName type: 'char *'>
	^self cCode: 'ioCanCreatePathOfSize(dirName, dirNameSize)'! !

!SecurityPlugin methodsFor: 'exported functions' stamp: '(auto pragmas 12/08) '!
secCanDeleteFile: fileName OfSize: fileNameSize
	<export: true>
	<var: #fileName type: 'char *'>
	^self cCode: 'ioCanDeleteFileOfSize(fileName, fileNameSize)'! !

!SecurityPlugin methodsFor: 'exported functions' stamp: '(auto pragmas 12/08) '!
secCanDeletePath: dirName OfSize: dirNameSize
	<export: true>
	<var: #dirName type: 'char *'>
	^self cCode: 'ioCanDeletePathOfSize(dirName, dirNameSize)'! !

!SecurityPlugin methodsFor: 'exported functions' stamp: '(auto pragmas 12/08) '!
secCanGetFileType: fileName OfSize: fileNameSize
	<export: true>
	<var: #fileName type: 'char *'>
	^self cCode: 'ioCanGetFileTypeOfSize(fileName, fileNameSize)'! !

!SecurityPlugin methodsFor: 'exported functions' stamp: '(auto pragmas 12/08) '!
secCanListPath: pathName OfSize: pathNameSize
	<export: true>
	<var: #pathName type: 'char *'>
	^self cCode: 'ioCanListPathOfSize(pathName, pathNameSize)'! !

!SecurityPlugin methodsFor: 'exported functions' stamp: '(auto pragmas 12/08) '!
secCanOpenAsyncFile: fileName OfSize: fileNameSize Writable: writeFlag
	<export: true>
	<var: #fileName type: 'char *'>
	^self cCode: 'ioCanOpenAsyncFileOfSizeWritable(fileName, fileNameSize, writeFlag)'! !

!SecurityPlugin methodsFor: 'exported functions' stamp: '(auto pragmas 12/08) '!
secCanOpenFile: fileName OfSize: fileNameSize Writable: writeFlag
	<export: true>
	<var: #fileName type: 'char *'>
	^self cCode: 'ioCanOpenFileOfSizeWritable(fileName, fileNameSize, writeFlag)'! !

!SecurityPlugin methodsFor: 'exported functions' stamp: '(auto pragmas 12/08) '!
secCanRenameFile: fileName OfSize: fileNameSize
	<export: true>
	<var: #fileName type: 'char *'>
	^self cCode: 'ioCanRenameFileOfSize(fileName, fileNameSize)'! !

!SecurityPlugin methodsFor: 'exported functions' stamp: 'eem 3/17/2017 10:44'!
secCanRenameImage
	<export: true>
	^self ioCanRenameImage! !

!SecurityPlugin methodsFor: 'exported functions' stamp: '(auto pragmas 12/08) '!
secCanSetFileType: fileName OfSize: fileNameSize
	<export: true>
	<var: #fileName type: 'char *'>
	^self cCode: 'ioCanSetFileTypeOfSize(fileName, fileNameSize)'! !

!SecurityPlugin methodsFor: 'exported functions' stamp: 'eem 3/17/2017 10:44'!
secCanWriteImage
	<export: true>
	^self ioCanWriteImage! !

!SecurityPlugin methodsFor: 'exported functions' stamp: 'eem 3/17/2017 10:44'!
secDisableFileAccess
	<export: true>
	^self ioDisableFileAccess! !

!SecurityPlugin methodsFor: 'exported functions' stamp: 'eem 3/17/2017 10:44'!
secDisableSocketAccess
	<export: true>
	^self ioDisableSocketAccess! !

!SecurityPlugin methodsFor: 'exported functions' stamp: 'eem 3/17/2017 10:43'!
secHasEnvironmentAccess
	<export: true>
	^self ioHasEnvironmentAccess! !

!SecurityPlugin methodsFor: 'exported functions' stamp: 'eem 3/17/2017 10:45'!
secHasFileAccess
	<export: true>
	^self ioHasFileAccess! !

!SecurityPlugin methodsFor: 'exported functions' stamp: 'eem 3/17/2017 10:45'!
secHasSocketAccess
	<export: true>
	^self ioHasSocketAccess! !

InterpreterPlugin subclass: #SlangBasicTypeInferenceTestClass
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!SlangBasicTypeInferenceTestClass methodsFor: 'constant' stamp: ''!
aBigNegativeIntegerConstantNode
	"SmallInteger maxVal on a 64bit system"
	-1152921504606846975! !

!SlangBasicTypeInferenceTestClass methodsFor: 'constant' stamp: ''!
aFalseConstantNode
	false! !

!SlangBasicTypeInferenceTestClass methodsFor: 'constant' stamp: ''!
aFloatConstantNode
	1.0! !

!SlangBasicTypeInferenceTestClass methodsFor: 'constant' stamp: ''!
aNilConstantNode
	nil! !

!SlangBasicTypeInferenceTestClass methodsFor: 'constant' stamp: ''!
aSmallNegativeIntegerConstantNode
	-30! !

!SlangBasicTypeInferenceTestClass methodsFor: 'constant' stamp: ''!
aStringConstantNode
	'aString'! !

!SlangBasicTypeInferenceTestClass methodsFor: 'constant' stamp: ''!
aTrueConstantNode
	true! !

!SlangBasicTypeInferenceTestClass methodsFor: 'constant' stamp: ''!
anIntEqual32ConstantNode
	" 2 ^ 32 - 1"
	4294967295! !

!SlangBasicTypeInferenceTestClass methodsFor: 'constant' stamp: ''!
anIntEqual64ConstantNode
	" 2 raisedTo: 64 "
	18446744073709551615! !

!SlangBasicTypeInferenceTestClass methodsFor: 'constant' stamp: ''!
anIntGreater64ConstantNode
	"SmallInteger maxVal on a 64bit system"
	1152921504606846975! !

!SlangBasicTypeInferenceTestClass methodsFor: 'constant' stamp: ''!
anIntLesser32ConstantNode
	30! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-constant' stamp: ''!
returnABigNegativeIntegerConstantNode
"SmallInteger maxVal on a 64bit system"
	^ -1152921504606846975
	! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-message-send' stamp: ''!
returnABigNegativeIntegerMessageSend
	^ self returnABigNegativeIntegerConstantNode! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-constant' stamp: ''!
returnAFalseConstantNode
	^ false! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-message-send' stamp: ''!
returnAFalseMessageSend
	^ self returnAFalseConstantNode! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-constant' stamp: ''!
returnAFloatConstantNode
	^ 1.0! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-message-send' stamp: ''!
returnAFloatMessageSend
	^ self returnAFloatConstantNode! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-constant' stamp: ''!
returnANilConstantNode
	^ nil! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-message-send' stamp: ''!
returnANilMessageSend
	^ self returnANilConstantNode! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-constant' stamp: ''!
returnASmallNegativeIntegerConstantNode
	^ -30! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-message-send' stamp: ''!
returnASmallNegativeIntegerMessageSend
	^ self returnASmallNegativeIntegerConstantNode! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-constant' stamp: ''!
returnAStringConstantNode
	^ 'aString'! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-message-send' stamp: ''!
returnAStringMessageSend
	^ self returnAStringConstantNode! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-constant' stamp: ''!
returnATrueConstantNode
	^ true! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-message-send' stamp: ''!
returnATrueMessageSend
	^ self returnATrueConstantNode! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-constant' stamp: ''!
returnAnIntEqual32ConstantNode
	" 2 ^ 32 - 1"
	^ 4294967295! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-message-send' stamp: ''!
returnAnIntEqual32MessageSend
	^ self returnAnIntEqual32ConstantNode! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-constant' stamp: ''!
returnAnIntEqual64ConstantNode
	" 2 raisedTo: 64 "
	^ 18446744073709551615! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-message-send' stamp: ''!
returnAnIntEqual64MessageSend
	^ self returnAnIntEqual64ConstantNode! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-constant' stamp: ''!
returnAnIntGreater64ConstantNode
	"SmallInteger maxVal on a 64bit system"
	^ 1152921504606846975! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-message-send' stamp: ''!
returnAnIntGreater64MessageSend
	^ self returnAnIntGreater64ConstantNode! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-constant' stamp: ''!
returnAnIntLesser32ConstantNode
	^ 30! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-message-send' stamp: ''!
returnAnIntLesser32MessageSend
	^ self returnAnIntLesser32ConstantNode! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-arg' stamp: ''!
returnExplicitArgDouble: t
	<var:#t type:#double>
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-arg' stamp: ''!
returnExplicitArgFloat: t
	<var:#t type:#float>
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-arg' stamp: ''!
returnExplicitArgInt: t
	<var:#t type:#int>
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-arg' stamp: ''!
returnExplicitArgLongLong: t
	<var:#t type:#'long long'>
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-arg' stamp: ''!
returnExplicitArgRandom: t
	<var:#t type:#completelyRandom>
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-arg' stamp: ''!
returnExplicitArgString: t
	<var:#t type:#'char *'>
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-arg' stamp: ''!
returnExplicitArgUnsignedInt: t
	<var:#t type:#'unsigned int'>
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-arg' stamp: ''!
returnExplicitArgUnsignedLongLong: t
	<var:#t type:#'unsigned long long'>
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-temp' stamp: ''!
returnExplicitTempDouble
	<var:#t type:#double>
	| t |
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-temp' stamp: ''!
returnExplicitTempFloat
	<var:#t type:#float>
	| t |
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-temp' stamp: ''!
returnExplicitTempInt
	<var:#t type:#int>
	| t |
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-temp' stamp: ''!
returnExplicitTempLongLong
	<var:#t type:#'long long'>
	| t | 
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-temp' stamp: ''!
returnExplicitTempRandom
	<var:#t type:#completelyRandom>
	| t |
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-temp' stamp: ''!
returnExplicitTempString
	<var:#t type:#'char *'>
	| t |
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-temp' stamp: ''!
returnExplicitTempUnsignedInt
	<var:#t type:#'unsigned int'>
	| t |
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-explicit-temp' stamp: ''!
returnExplicitTempUnsignedLongLong
	<var:#t type:#'unsigned long long'>
	| t |
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-const' stamp: ''!
returnTempBigNegativeIntegerConstantNode
	| t |
	"SmallInteger maxVal on a 64bit system"
	t := -1152921504606846975.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-message' stamp: ''!
returnTempBigNegativeIntegerMessageNode
	| t |
	t := self returnABigNegativeIntegerConstantNode.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-const' stamp: ''!
returnTempFalseConstantNode
	| t |
	t := false.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-message' stamp: ''!
returnTempFalseMessageNode
	| t |
	t := self returnAFalseConstantNode.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-const' stamp: ''!
returnTempFloatConstantNode
	| t |
	t := 1.0.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-message' stamp: ''!
returnTempFloatMessageNode
	| t |
	t := self returnAFloatMessageSend.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-const' stamp: ''!
returnTempIntEqual32ConstantNode
	" 2 ^ 32 - 1"
	| t |
	t := 4294967295.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-message' stamp: ''!
returnTempIntEqual32MessageNode
	| t |
	t := self returnAnIntEqual32ConstantNode.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-const' stamp: ''!
returnTempIntEqual64ConstantNode
	" 2 raisedTo: 64 "
	| t | 
	t := 18446744073709551615.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-message' stamp: ''!
returnTempIntEqual64MessageNode
	| t | 
	t := self returnAnIntEqual64ConstantNode.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-const' stamp: ''!
returnTempIntGreater64ConstantNode
	| t |
	"SmallInteger maxVal on a 64bit system"
	t := 1152921504606846975.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-message' stamp: ''!
returnTempIntGreater64MessageNode
	| t |
	t := self returnAnIntGreater64ConstantNode.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-const' stamp: ''!
returnTempIntLesser32ConstantNode
	| t |
	t := 30.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-message' stamp: ''!
returnTempIntLesser32MessageNode
	| t |
	t := self returnAnIntLesser32ConstantNode.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-const' stamp: ''!
returnTempNilConstantNode
	| t |
	t := nil.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-message' stamp: ''!
returnTempNilMessageNode
	| t |
	t := self returnANilConstantNode.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-const' stamp: ''!
returnTempSmallNegativeIntegerConstantNode
	| t |
	t := -30.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-message' stamp: ''!
returnTempSmallNegativeIntegerMessageNode
	| t |
	t := self returnASmallNegativeIntegerConstantNode.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-const' stamp: ''!
returnTempStringConstantNode
	| t |
	t := 'aString'.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-message' stamp: ''!
returnTempStringMessageNode
	| t |
	t := self returnAStringConstantNode.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-const' stamp: ''!
returnTempTrueConstantNode
	| t |
	t := true.
	^ t! !

!SlangBasicTypeInferenceTestClass methodsFor: 'return-temp-assigned-message' stamp: ''!
returnTempTrueMessageNode
	| t |
	t := self returnATrueConstantNode.
	^ t! !

InterpreterPlugin subclass: #SlangTestPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!SlangTestPlugin methodsFor: 'primitives' stamp: 'ar 3/11/2009 00:34'!
getByteFunctional
	"Return a byte from a (semi-)functional method"
	<returnTypeC: 'unsigned char'>
	^12345! !

!SlangTestPlugin methodsFor: 'primitives' stamp: 'ar 3/11/2009 00:42'!
getByteFunctionalAt: addr
	"Return a byte from a (semi-)functional method"
	<returnTypeC: 'unsigned char'>
	^interpreterProxy longAt: addr! !

!SlangTestPlugin methodsFor: 'primitives' stamp: 'ar 3/11/2009 00:39'!
getByteNonFunctional
	"Return a byte from a non-functional method"
	<returnTypeC: 'unsigned char'>
	self msg: 'yo'.
	^12345! !

!SlangTestPlugin methodsFor: 'primitives' stamp: 'ar 3/11/2009 00:42'!
getByteNonFunctionalAt: addr
	"Return a byte from a non-functional method"
	<returnTypeC: 'unsigned char'>
	self msg: 'yo'.
	^interpreterProxy longAt: addr! !

!SlangTestPlugin methodsFor: 'primitives' stamp: 'ar 3/11/2009 00:43'!
testTypePropagationBasic
	"Ensure that types are propagated correctly"
	| shouldBeInt |
	<export: true>
	"the resulting byte should be truncated by the functions and expanded back to sqInt"
	shouldBeInt := self getByteNonFunctional.
	interpreterProxy pushBool: shouldBeInt < 256.
! !

!SlangTestPlugin methodsFor: 'primitives' stamp: 'ar 3/11/2009 00:43'!
testTypePropagationBasicAt
	"Ensure that types are propagated correctly"
	| buffer shouldBeInt |
	<export: true>
	<var: #buffer declareC: 'sqInt buffer[1] = { 12345 }'>
	"the resulting byte should be truncated by the functions and expanded back to sqInt"
	shouldBeInt := self getByteNonFunctionalAt: buffer.
	interpreterProxy pushBool: shouldBeInt < 256.
! !

!SlangTestPlugin methodsFor: 'primitives' stamp: 'ar 3/11/2009 00:43'!
testTypePropagationBasicFunctional
	"Ensure that types are propagated correctly"
	| shouldBeInt |
	<export: true>
	"the resulting byte should be truncated by the functions and expanded back to sqInt"
	shouldBeInt := self getByteFunctional.
	interpreterProxy pushBool: shouldBeInt < 256.
! !

!SlangTestPlugin methodsFor: 'primitives' stamp: 'ar 3/11/2009 00:43'!
testTypePropagationBasicFunctionalAt
	"Ensure that types are propagated correctly"
	| buffer shouldBeInt |
	<export: true>
	<var: #buffer declareC: 'sqInt buffer[1] = { 12345 }'>
	"the resulting byte should be truncated by the functions and expanded back to sqInt"
	shouldBeInt := self getByteFunctionalAt: buffer.
	interpreterProxy pushBool: shouldBeInt < 256.
! !

!SlangTestPlugin methodsFor: 'primitives' stamp: 'ar 3/11/2009 00:43'!
testTypePropagationBasicNoPropagate
	"Ensure that types are propagated correctly"
	| shouldBeInt |
	<export: true>
	shouldBeInt := self getByteNonFunctional.
	"shoudBeInt should be int, not unsigned char"
	shouldBeInt := 12345.
	interpreterProxy pushBool: shouldBeInt > 256.! !

!SlangTestPlugin methodsFor: 'primitives' stamp: 'ar 3/11/2009 00:43'!
testTypePropagationBasicNoPropagateAt
	"Ensure that types are propagated correctly"
	| buffer shouldBeInt |
	<export: true>
	<var: #buffer declareC: 'sqInt buffer[1] = { 12345 }'>
	shouldBeInt := self getByteNonFunctionalAt: buffer.
	"shoudBeInt should be int, not unsigned char"
	shouldBeInt := 12345.
	interpreterProxy pushBool: shouldBeInt > 256.! !

!SlangTestPlugin methodsFor: 'primitives' stamp: 'ar 3/11/2009 00:43'!
testTypePropagationBasicNoPropagateFunctional
	"Ensure that types are propagated correctly"
	| shouldBeInt |
	<export: true>
	shouldBeInt := self getByteFunctional.
	"shoudBeInt should be int, not unsigned char"
	shouldBeInt := 12345.
	interpreterProxy pushBool: shouldBeInt > 256.! !

!SlangTestPlugin methodsFor: 'primitives' stamp: 'ar 3/11/2009 00:43'!
testTypePropagationBasicNoPropagateFunctionalAt
	"Ensure that types are propagated correctly"
	| buffer shouldBeInt |
	<export: true>
	<var: #buffer declareC: 'sqInt buffer[1] = { 12345 }'>
	shouldBeInt := self getByteFunctionalAt: buffer.
	"shoudBeInt should be int, not unsigned char"
	shouldBeInt := 12345.
	interpreterProxy pushBool: shouldBeInt > 256.! !

InterpreterPlugin subclass: #SlangTypeForDereferenceTestClass
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!SlangTypeForDereferenceTestClass methodsFor: 'as yet unclassified' stamp: ''!
aMethodReturningAMessageSendToSelf
	^ self aMethodReturningAPointerType! !

!SlangTypeForDereferenceTestClass methodsFor: 'as yet unclassified' stamp: ''!
aMethodReturningAPointerType
	<var:#a type:#'int*'>
	| a |
	^ a! !

!SlangTypeForDereferenceTestClass methodsFor: 'as yet unclassified' stamp: ''!
aMethodReturningAnAtOnAMatrix
	<var:#a type:#'int**'>
	| a |
	^ a at: 2! !

!SlangTypeForDereferenceTestClass methodsFor: 'as yet unclassified' stamp: ''!
aMethodReturningAnAtOnAnArray
	<var:#a type:#'int*'>
	| a |
	^ a at: 2! !

!SlangTypeForDereferenceTestClass methodsFor: 'as yet unclassified' stamp: ''!
aMethodReturningTwoAtOnAMatrix
	<var:#a type:#'int**'>
	| a |
	^ (a at: 2) at: 2! !

InterpreterPlugin subclass: #SmartSyntaxInterpreterPlugin
	instanceVariableNames: 'simulator'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SmartSyntaxPlugins'!

!SmartSyntaxInterpreterPlugin commentStamp: '<historical>' prior: 0!
Subclass of InterpreterPlugin, used in connection with TestCodeGenerator for named primitives with type coercion specifications!

SmartSyntaxInterpreterPlugin subclass: #AsynchFilePlugin
	instanceVariableNames: 'sCOAFfn'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!AsynchFilePlugin commentStamp: 'tpr 5/2/2003 15:22' prior: 0!
Implements the asynchronous file primitives available on a few platforms. See the platform specific files in platforms- {your platform} - plugins - Asynchplugin!

!AsynchFilePlugin class methodsFor: 'translation' stamp: 'JMM 5/23/2005 18:59'!
declareCVarsIn: cg
	super declareCVarsIn: cg.
	cg var: #sCOAFfn type: #'void *'.
! !

!AsynchFilePlugin class methodsFor: 'translation' stamp: 'tpr 5/23/2001 17:04'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!AsynchFilePlugin class methodsFor: 'translation' stamp: 'tpr 11/29/2000 22:37'!
requiresPlatformFiles
	"this plugin requires platform specific files in order to work"
	^true! !

!AsynchFilePlugin methodsFor: 'primitives' stamp: 'eem 12/10/2014 14:30'!
asyncFileValueOf: oop
	"Answer a pointer to the first byte of the async file record within the given
	 Smalltalk bytes object, or fail and answer nil if oop is not an async file record."

	<returnTypeC: 'AsyncFile *'>
	((interpreterProxy isBytes: oop)
	 and: [(interpreterProxy byteSizeOf: oop) = (self sizeof: #AsyncFile)]) ifTrue:
		[^self cCoerceSimple: (interpreterProxy firstIndexableField: oop) to: 'AsyncFile *'].
	interpreterProxy primitiveFailFor: PrimErrBadArgument.
	^nil! !

!AsynchFilePlugin methodsFor: 'initialize-release' stamp: 'JMM (auto pragmas 12/08) 1/20/2002 22:00'!
initialiseModule
	"Initialise the module"
	<export: true>
	sCOAFfn := interpreterProxy ioLoadFunction: 'secCanOpenAsyncFileOfSizeWritable' From: 'SecurityPlugin'.
	^self cCode: 'asyncFileInit()' inSmalltalk:[true]! !

!AsynchFilePlugin methodsFor: 'initialize-release' stamp: 'tpr (auto pragmas 12/08) 3/24/2004 14:51'!
moduleUnloaded: aModuleName 
	"The module with the given name was just unloaded. 
	Make sure we have no dangling references."
	<export: true>
	<var: #aModuleName type: 'char *'>
	(aModuleName strcmp: 'SecurityPlugin') = 0
		ifTrue: ["The security plugin just shut down. How odd. Zero the function pointer we have into it"
			sCOAFfn := 0]! !

!AsynchFilePlugin methodsFor: 'primitives' stamp: 'eem 1/30/2015 12:43'!
primitiveAsyncFileClose: fh 
	| f |
	<var: #f type: 'AsyncFile *'>
	self primitive: 'primitiveAsyncFileClose' parameters: #(Oop ).
	f := self asyncFileValueOf: fh.
	interpreterProxy failed ifTrue: [^nil].
	self asyncFileClose: f! !

!AsynchFilePlugin methodsFor: 'primitives' stamp: 'eem 12/10/2014 14:40'!
primitiveAsyncFileOpen: fileName forWrite: writeFlag semaIndex: semaIndex 
	| fileNameSize fOop f okToOpen |
	<var: #f type: 'AsyncFile *'>
	self primitive: 'primitiveAsyncFileOpen' parameters: #(#String #Boolean #SmallInteger ).
	fileNameSize := interpreterProxy byteSizeOf: (fileName asOop: String).
	"If the security plugin can be loaded, use it to check for permission.
	If not, assume it's ok"
	sCOAFfn ~= 0 ifTrue:
		[okToOpen := self cCode: '((sqInt (*) (char *, sqInt, sqInt)) sCOAFfn)(fileName, fileNameSize, writeFlag)'
							inSmalltalk: [true].
		 okToOpen ifFalse:
			[^interpreterProxy primitiveFail]].
	fOop := interpreterProxy
				instantiateClass: interpreterProxy classByteArray
				indexableSize: (self sizeof: #AsyncFile).
	f := self asyncFileValueOf: fOop.
	interpreterProxy failed ifFalse:
		[self cCode: 'asyncFileOpen(f, fileName, fileNameSize, writeFlag, semaIndex)'].
	^fOop! !

!AsynchFilePlugin methodsFor: 'primitives' stamp: 'eem 1/30/2015 12:44'!
primitiveAsyncFileReadResult: fhandle intoBuffer: buffer at: start count: num 
	| bufferSize bufferPtr r f count startIndex |
	<var: #f type: 'AsyncFile *'>
	self primitive: 'primitiveAsyncFileReadResult' parameters: #(Oop Oop SmallInteger SmallInteger ).

	f := self asyncFileValueOf: fhandle.
	count := num.
	startIndex := start.
	bufferSize := interpreterProxy slotSizeOf: buffer. "in bytes or words"
	(interpreterProxy isWords: buffer) ifTrue: "covert word counts to byte counts"
		[count := count * 4.
		 startIndex := startIndex - 1 * 4 + 1.
		 bufferSize := bufferSize * 4].
	interpreterProxy success: (startIndex >= 1 and: [startIndex + count - 1 <= bufferSize]).

	interpreterProxy failed ifTrue: [^nil].
	"adjust for zero-origin indexing"
	bufferPtr := (self cCoerce: (interpreterProxy firstIndexableField: buffer) to:#sqInt) + startIndex - 1.
	r := self asyncFile: f Read: bufferPtr asVoidPointer Result: count.
	^r asOop: SmallInteger! !

!AsynchFilePlugin methodsFor: 'primitives' stamp: 'eem 1/30/2015 12:44'!
primitiveAsyncFileReadStart: fHandle fPosition: fPosition count: count
	| f |
	<var: #f type: 'AsyncFile *'>
	self primitive: 'primitiveAsyncFileReadStart' parameters: #(Oop SmallInteger SmallInteger).
	f := self asyncFileValueOf: fHandle.
	interpreterProxy failed ifTrue: [^nil].
	self asyncFile: f Read: fPosition Start: count! !

!AsynchFilePlugin methodsFor: 'primitives' stamp: 'eem 1/30/2015 12:44'!
primitiveAsyncFileWriteResult: fHandle
	| f r |
	<var: #f type: 'AsyncFile *'>
	self primitive: 'primitiveAsyncFileWriteResult' parameters:#(Oop).

	f := self asyncFileValueOf: fHandle.
	interpreterProxy failed ifTrue: [^nil].
	r := self asyncFileWriteResult: f.
	^r asOop: SmallInteger! !

!AsynchFilePlugin methodsFor: 'primitives' stamp: 'eem 1/30/2015 12:43'!
primitiveAsyncFileWriteStart: fHandle fPosition: fPosition fromBuffer: buffer at: start count: num 
	| f bufferSize bufferPtr count startIndex |
	<var: #f type: 'AsyncFile *'>
	self primitive: 'primitiveAsyncFileWriteStart' parameters: #(Oop SmallInteger Oop SmallInteger SmallInteger ).
	f := self asyncFileValueOf: fHandle.
	count := num.
	startIndex := start.
	bufferSize := interpreterProxy slotSizeOf: buffer.	"in bytes or words"
	(interpreterProxy isWords: buffer) ifTrue: "covert word counts to byte counts"
		[count := count * 4.
		 startIndex := startIndex - 1 * 4 + 1.
		 bufferSize := bufferSize * 4].
	interpreterProxy success: (startIndex >= 1 and: [startIndex + count - 1 <= bufferSize]).

	interpreterProxy failed ifTrue: [^nil].
	"adjust for zero-origin indexing"
	bufferPtr := (self cCoerce: (interpreterProxy firstIndexableField: buffer) to: #sqInt) + startIndex - 1.
	self async: f File: fPosition Write: bufferPtr asVoidPointer Start: count! !

!AsynchFilePlugin methodsFor: 'initialize-release' stamp: 'eem 11/10/2020 11:52'!
shutdownModule
	"Initialise the module"
	<export: true>
	^self asyncFileShutdown! !

!AsynchFilePlugin methodsFor: 'simulation' stamp: 'eem 12/10/2014 14:26'!
sizeof: objectSymbolOrClass
	<doNotGenerate>
	objectSymbolOrClass isInteger ifTrue:
		[^interpreterProxy wordSize].
	objectSymbolOrClass == #AsyncFile ifTrue:
		[^interpreterProxy wordSize * 2].
	^super sizeof: objectSymbolOrClass! !

SmartSyntaxInterpreterPlugin subclass: #BitBltSimulation
	instanceVariableNames: 'destForm sourceForm halftoneForm combinationRule destX destY width height sourceX sourceY clipX clipY clipWidth clipHeight sourceWidth sourceHeight sourceDepth sourcePitch sourceBits sourcePPW sourceMSB destWidth destHeight destDepth destPitch destBits destPPW destMSB bitCount skew mask1 mask2 preload nWords destMask hDir vDir sourceIndex sourceDelta destIndex destDelta sx sy dx dy bbW bbH halftoneHeight noSource noHalftone halftoneBase sourceAlpha srcBitShift dstBitShift bitBltOop affectedL affectedR affectedT affectedB opTable maskTable ditherMatrix4x4 ditherThresholds16 ditherValues16 hasSurfaceLock warpSrcShift warpSrcMask warpAlignShift warpAlignMask warpBitShiftTable querySurfaceFn lockSurfaceFn unlockSurfaceFn isWarping cmFlags cmMask cmShiftTable cmMaskTable cmLookupTable cmBitsPerColor dither8Lookup componentAlphaModeColor componentAlphaModeAlpha ungammaLookupTable gammaLookupTable numGCsOnInvocation bitBltIsReceiver endOfDestination endOfSource'
	classVariableNames: 'AllOnes AlphaIndex BBClipHeightIndex BBClipWidthIndex BBClipXIndex BBClipYIndex BBColorMapIndex BBDestFormIndex BBDestXIndex BBDestYIndex BBHalftoneFormIndex BBHeightIndex BBLastIndex BBRuleIndex BBSourceFormIndex BBSourceXIndex BBSourceYIndex BBWarpBase BBWidthIndex BBXTableIndex BEBitBltIndex BinaryPoint BlueIndex ColorMapFixedPart ColorMapIndexedPart ColorMapNewStyle ColorMapPresent CrossedX EndOfRun FixedPt1 FormBitsIndex FormDepthIndex FormHeightIndex FormWidthIndex GreenIndex OpTable OpTableSize RedIndex'
	poolDictionaries: ''
	category: 'VMMaker-Interpreter'!

!BitBltSimulation commentStamp: 'nice 10/31/2020 23:39' prior: 0!
This class implements BitBlt, much as specified in the Blue Book spec.

Performance has been enhanced through the use of pointer variables such as sourceIndex and destIndex, and by separating several special cases of the inner loop.

Operation has been extended to color, with support for 1, 2, 4, 8, 16, and 32-bit pixel sizes.  Conversion between different pixel sizes is facilitated by accepting an optional color map.

In addition to the original 16 combination rules, this BitBlt supports
	16	fail (for old paint mode)
	17	fail (for old mask mode)
	18	sourceWord + destinationWord
	19	sourceWord - destinationWord
	20	rgbAdd: sourceWord with: destinationWord
	21	rgbSub: sourceWord with: destinationWord
	22	OLDrgbDiff: sourceWord with: destinationWord
	23	OLDtallyIntoMap: destinationWord -- old vers doesn't clip to bit boundary
	24	alphaBlend: sourceWord with: destinationWord
	25	pixPaint: sourceWord with: destinationWord
	26	pixMask: sourceWord with: destinationWord
	27	rgbMax: sourceWord with: destinationWord
	28	rgbMin: sourceWord with: destinationWord
	29	rgbMin: sourceWord bitInvert32 with: destinationWord
	30	alphaBlendConst: sourceWord with: destinationWord -- alpha passed as an arg
	31	alphaPaintConst: sourceWord with: destinationWord -- alpha passed as an arg
	32	rgbDiff: sourceWord with: destinationWord
	33	tallyIntoMap: destinationWord
	34	alphaBlendScaled: sourceWord with: destinationWord
	35 alphaBlendScaled: sourceWord with:	"unused here - only used by FXBlt"
	36 alphaBlendScaled: sourceWord with:	"unused here - only used by FXBlt"
	37 rgbMul: sourceWord with: destinationWord
	38 pixSwap: sourceWord with: destinationWord
	39 pixClear: sourceWord with: destinationWord
	40 fixAlpha: sourceWord with: destinationWord
	41 rgbComponentAlpha: sourceWord with: destinationWord
	42 alphaScale: ignoredSourceWord with: destinationWord
	43 alphaUnscale: ignoredSourceWord with: destinationWord
	44	alphaBlendUnscaled: sourceWord with: destinationWord

This implementation has also been fitted with an experimental "warp drive" that allows abritrary scaling and rotation (and even limited affine deformations) with all BitBlt storage modes supported.

To add a new rule to BitBlt...
	1.  add the new rule method or methods in the category 'combination rules' of BBSim
	2.  describe it in the class comment  of BBSim and in the class comment for BitBlt
	3.  add refs to initializeRuleTable in proper positions
	4.  add refs to initBBOpTable, following the pattern
!

!BitBltSimulation class methodsFor: 'system simulation' stamp: 'tfel 2/17/2016 13:57'!
copyBitsFrom: aBitBlt
	"Simulate the copyBits primitive"
	| proxy bb |
	proxy := InterpreterProxy new.
	proxy loadStackFrom: thisContext sender home.
	bb := self simulatorClass new.
	bb initialiseModule.
	bb setInterpreter: proxy.
	proxy success: (bb loadBitBltFrom: aBitBlt).
	bb primitiveCopyBits.
	^ proxy stackValue: 0! !

!BitBltSimulation class methodsFor: 'translation' stamp: 'nice 11/21/2019 23:41'!
declareCVarsIn: aCCodeGenerator

	"add option of  fast path BitBLT code header"
	aCCodeGenerator
		addHeaderFile:'"sqAssert.h"';
		addHeaderFile:'#ifdef ENABLE_FAST_BLT
#include "BitBltDispatch.h"
#else
// to handle the unavoidable decl in the spec of copyBitsFallback();
#define operation_t void
#endif'.
		
	aCCodeGenerator var: 'opTable'
		declareC: 'void *opTable[' , OpTableSize printString , ']'.
	aCCodeGenerator var: 'maskTable'
		declareC:'int maskTable[33] = {
0, 1, 3, 0, 15, 31, 0, 0, 255, 0, 0, 0, 0, 0, 0, 0, 65535,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1
}'.
	aCCodeGenerator var: 'ditherMatrix4x4'
		declareC:'const int ditherMatrix4x4[16] = {
0,	8,	2,	10,
12,	4,	14,	6,
3,	11,	1,	9,
15,	7,	13,	5
}'.
	aCCodeGenerator var: 'ditherThresholds16'
		declareC:'const int ditherThresholds16[8] = { 0, 2, 4, 6, 8, 12, 14, 16 }'.
	aCCodeGenerator var: 'ditherValues16'
		declareC:'const int ditherValues16[32] = {
0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14,
15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30
}'.

	aCCodeGenerator var: 'warpBitShiftTable'
		declareC:'int warpBitShiftTable[32]'.

	aCCodeGenerator var:'cmShiftTable' 
		type:'int *'.
	aCCodeGenerator var:'cmMaskTable' 
		type:'unsigned int *'.
	aCCodeGenerator var:'cmLookupTable' 
		type:'unsigned int *'.

	aCCodeGenerator var: 'dither8Lookup'
		declareC:' unsigned char dither8Lookup[4096]'.

	aCCodeGenerator var:'ungammaLookupTable' 
		type: 'unsigned char *'.
	aCCodeGenerator var:'gammaLookupTable' 
		type: 'unsigned char *'.

	aCCodeGenerator
		var: 'querySurfaceFn' declareC: 'int (*querySurfaceFn)(sqIntptr_t, int*, int*, int*, int*)';
		var: 'lockSurfaceFn' declareC: 'sqIntptr_t (*lockSurfaceFn)(sqIntptr_t, int*, int, int, int, int)';
		var: 'unlockSurfaceFn' declareC: 'int (*unlockSurfaceFn)(sqIntptr_t, int, int, int, int)'.
	
	#(sourcePitch sourceWidth sourceHeight sourceDepth sourceMSB
		destPitch destWidth destHeight destDepth destMSB bbW bbH)
		do: [:ivar | aCCodeGenerator var: ivar type: #int].

	#(dx dy sx sy) do:
		[:ivar| aCCodeGenerator var: ivar type: #sqInt].
	
	#('sourceIndex' 'destIndex' 'endOfDestination' 'endOfSource') do:
		[:ivar| aCCodeGenerator var: ivar type: #usqInt]! !

!BitBltSimulation class methodsFor: 'initialization' stamp: 'eem 3/14/2017 15:45'!
initialize
	"BitBltSimulation initialize"

	self initializeRuleTable.

	"Mask constants"
	AllOnes := 16rFFFFFFFF.
	BinaryPoint := 14.
	FixedPt1 := 1 << BinaryPoint.  "Value of 1.0 in Warp's fixed-point representation"
 
	"Indices into stopConditions for scanning"
	EndOfRun := 257.
	CrossedX := 258.
 
	"Form fields"
	FormBitsIndex := 0.
	FormWidthIndex := 1.
	FormHeightIndex := 2.
	FormDepthIndex := 3.
 
	"BitBlt fields"
	BBDestFormIndex := 0.
	BBSourceFormIndex := 1.
	BBHalftoneFormIndex := 2.
	BBRuleIndex := 3.
	BBDestXIndex := 4.
	BBDestYIndex := 5.
	BBWidthIndex := 6.
	BBHeightIndex := 7.
	BBSourceXIndex := 8.
	BBSourceYIndex := 9.
	BBClipXIndex := 10.
	BBClipYIndex := 11.
	BBClipWidthIndex := 12.
	BBClipHeightIndex := 13.
	BBColorMapIndex := 14.
	BBWarpBase := 15.
	BBLastIndex := 15.
	BBXTableIndex := 16.

	"BalloonEngineFields"
	BEBitBltIndex := (Smalltalk classNamed: #BalloonEngine)
						ifNil: [2]
						ifNotNil: [:be| (be allInstVarNames indexOf: 'bitBlt') - 1].

	"RGBA indexes"
	RedIndex := 0.
	GreenIndex := 1.
	BlueIndex := 2.
	AlphaIndex := 3.

	"Color map flags"
	ColorMapPresent := 1.		"do we have one?"
	ColorMapFixedPart := 2.		"does it have a fixed part?"
	ColorMapIndexedPart := 4.	"does it have an indexed part?"
	ColorMapNewStyle := 8.		"new style color map"! !

!BitBltSimulation class methodsFor: 'initialization' stamp: 'nice 10/31/2020 18:55'!
initializeRuleTable
	"BitBltSimulation initializeRuleTable"
	"**WARNING** You MUST change initBBOpTable if you change this"
	OpTable := #(
		"0" clearWord:with:
		"1" bitAnd:with:
		"2" bitAndInvert:with:
		"3" sourceWord:with:
		"4" bitInvertAnd:with:
		"5" destinationWord:with:
		"6" bitXor:with:
		"7" bitOr:with:
		"8" bitInvertAndInvert:with:
		"9" bitInvertXor:with:
		"10" bitInvertDestination:with:
		"11" bitOrInvert:with:
		"12" bitInvertSource:with:
		"13" bitInvertOr:with:
		"14" bitInvertOrInvert:with:
		"15" destinationWord:with:
		"16" destinationWord:with: "unused - was old paint"
		"17" destinationWord:with: "unused - was old mask"
		"18" addWord:with:
		"19" subWord:with:
		"20" rgbAdd:with:
		"21" rgbSub:with:
		"22" OLDrgbDiff:with:
		"23" OLDtallyIntoMap:with:
		"24" alphaBlend:with:
		"25" pixPaint:with:
		"26" pixMask:with:
		"27" rgbMax:with:
		"28" rgbMin:with:
		"29" rgbMinInvert:with:
		"30" alphaBlendConst:with:
		"31" alphaPaintConst:with:
		"32" rgbDiff:with:
		"33" tallyIntoMap:with:
		"34" alphaBlendScaled:with:

		"35" alphaBlendScaled:with:	"unused here - only used by FXBlt"
		"36" alphaBlendScaled:with:	"unused here - only used by FXBlt"
		"37" rgbMul:with:
		"38" pixSwap:with:
		"39" pixClear:with:
		"40" fixAlpha:with:
		"41" rgbComponentAlpha:with:
		"42" alphaScale:with:
		"43" alphaUnscale:with:
		"44" alphaBlendUnscaled:with:
	).
	OpTableSize := OpTable size + 1.  "0-origin indexing"
! !

!BitBltSimulation class methodsFor: 'translation' stamp: 'ar 2/19/2000 20:55'!
moduleName
	^'BitBltPlugin'! !

!BitBltSimulation class methodsFor: 'translation' stamp: 'jm 5/12/1999 12:02'!
opTable

	^ OpTable
! !

!BitBltSimulation class methodsFor: 'translation' stamp: 'eem 8/21/2013 13:18'!
requiredMethodNames: options
	^self opTable asSet! !

!BitBltSimulation class methodsFor: 'system simulation' stamp: 'ar 10/27/1999 14:06'!
simulatorClass
	^BitBltSimulator! !

!BitBltSimulation class methodsFor: 'testing' stamp: 'tpr 3/24/2004 13:03'!
test2
	"BitBltSimulation test2"
	| f |
	Display fillWhite: (0 @ 0 extent: 300 @ 140).
	1 to: 12 do: [:i | 
			f := (Form extent: i @ 5) fillBlack.
			0 to: 20 do: [:x | f displayOn: Display at: x * 13 @ (i * 10)]]! !

!BitBltSimulation class methodsFor: 'testing' stamp: 'tpr 3/24/2004 13:03'!
timingTest: extent 
	"BitBltSimulation timingTest: 640@480"
	| f f2 map |
	f := Form extent: extent depth: 8.
	f2 := Form extent: extent depth: 8.
	map := Bitmap new: 1 << f2 depth.
	^ Array
		with: (Time millisecondsToRun: [100 timesRepeat: [f fillWithColor: Color white]])
		with: (Time millisecondsToRun: [100 timesRepeat: [f copy: f boundingBox from: 0 @ 0 in: f2 rule: Form over]])
		with: (Time millisecondsToRun: [100 timesRepeat: [f copyBits: f boundingBox from: f2 at: 0 @ 0 colorMap: map]])! !

!BitBltSimulation class methodsFor: 'system simulation' stamp: 'tpr 3/24/2004 13:02'!
warpBitsFrom: aBitBlt
	"Simulate the warpBits primitive"
	| proxy bb |
	proxy := InterpreterProxy new.
	proxy loadStackFrom: thisContext sender home.
	bb := self simulatorClass new.
	bb initialiseModule.
	bb setInterpreter: proxy.
	proxy success: (bb loadWarpBltFrom: aBitBlt).
	bb warpBits.
	proxy failed ifFalse:[
		proxy showDisplayBits: aBitBlt destForm Left: bb affectedLeft Top: bb affectedTop Right: bb affectedRight Bottom: bb affectedBottom].
	^proxy stackValue: 0! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'ar (auto pragmas 12/08) 4/18/2001 20:36'!
OLDrgbDiff: sourceWord with: destinationWord
	"Subract the pixels in the source and destination, color by color,
	and return the sum of the absolute value of all the differences.
	For non-rgb, XOR the two and return the number of differing pixels.
	Note that the region is not clipped to bit boundaries, but only to the
	nearest (enclosing) word.  This is because copyLoop does not do
	pre-merge masking.  For accurate results, you must subtract the
	values obtained from the left and right fringes."
	| diff pixMask |
	<inline: false>
	destDepth < 16 ifTrue:
		["Just xor and count differing bits if not RGB"
		diff := sourceWord bitXor: destinationWord.
		pixMask := maskTable at: destDepth.
		[diff = 0] whileFalse:
			[(diff bitAnd: pixMask) ~= 0 ifTrue: [bitCount := bitCount + 1].
			diff := diff >> destDepth].
		^ destinationWord "for no effect"].
 	destDepth = 16
		ifTrue:
		[diff := (self partitionedSub: sourceWord from: destinationWord
						nBits: 5 nPartitions: 3).
		bitCount := bitCount + (diff bitAnd: 16r1F)
							+ (diff>>5 bitAnd: 16r1F)
							+ (diff>>10 bitAnd: 16r1F).
		diff := (self partitionedSub: sourceWord>>16 from: destinationWord>>16
						nBits: 5 nPartitions: 3).
		bitCount := bitCount + (diff bitAnd: 16r1F)
							+ (diff>>5 bitAnd: 16r1F)
							+ (diff>>10 bitAnd: 16r1F)]
		ifFalse:
		[diff := (self partitionedSub: sourceWord from: destinationWord
						nBits: 8 nPartitions: 3).
		bitCount := bitCount + (diff bitAnd: 16rFF)
							+ (diff>>8 bitAnd: 16rFF)
							+ (diff>>16 bitAnd: 16rFF)].
	^ destinationWord  "For no effect on dest"! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'ar 5/17/2001 15:15'!
OLDtallyIntoMap: sourceWord with: destinationWord
	"Tally pixels into the color map.  Note that the source should be 
	specified = destination, in order for the proper color map checks 
	to be performed at setup.
	Note that the region is not clipped to bit boundaries, but only to the
	nearest (enclosing) word.  This is because copyLoop does not do
	pre-merge masking.  For accurate results, you must subtract the
	values obtained from the left and right fringes."
	| mapIndex pixMask shiftWord |
	(cmFlags bitAnd: (ColorMapPresent bitOr: ColorMapIndexedPart)) =
		(ColorMapPresent bitOr: ColorMapIndexedPart)
			ifFalse: [^ destinationWord "no op"].
	destDepth < 16 ifTrue:
		["loop through all packed pixels."
		pixMask := (maskTable at: destDepth) bitAnd: cmMask.
		shiftWord := destinationWord.
		1 to: destPPW do:
			[:i |
			mapIndex := shiftWord bitAnd: pixMask.
			self tallyMapAt: mapIndex put: (self tallyMapAt: mapIndex) + 1.
			shiftWord := shiftWord >> destDepth].
		^ destinationWord].
	destDepth = 16 ifTrue:
		["Two pixels  Tally the right half..."
		mapIndex := self rgbMap: (destinationWord bitAnd: 16rFFFF) from: 5 to: cmBitsPerColor.
		self tallyMapAt: mapIndex put: (self tallyMapAt: mapIndex) + 1.
		"... and then left half"
		mapIndex := self rgbMap: destinationWord>>16 from: 5 to: cmBitsPerColor.
		self tallyMapAt: mapIndex put: (self tallyMapAt: mapIndex) + 1]
	ifFalse:
		["Just one pixel."
		mapIndex := self rgbMap: destinationWord from: 8 to: cmBitsPerColor.
		self tallyMapAt: mapIndex put: (self tallyMapAt: mapIndex) + 1].
	^ destinationWord  "For no effect on dest"! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
addWord: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^sourceWord + destinationWord! !

!BitBltSimulation methodsFor: 'accessing' stamp: ''!
affectedBottom

	^affectedB! !

!BitBltSimulation methodsFor: 'accessing' stamp: ''!
affectedLeft

	^affectedL! !

!BitBltSimulation methodsFor: 'accessing' stamp: ''!
affectedRight

	^affectedR! !

!BitBltSimulation methodsFor: 'accessing' stamp: ''!
affectedTop

	^affectedT! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
alphaBlend: sourceWord with: destinationWord
	"Blend sourceWord with destinationWord, assuming both are 32-bit pixels.
	The source is assumed to have 255*alpha in the high 8 bits of each pixel,
	while the high 8 bits of the destinationWord will be ignored.
	The blend produced is alpha*source + (1-alpha)*dest, with
	the computation being performed independently on each color
	component.  The high byte of the result will be 0."
	| alpha unAlpha result blendRB blendAG |
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	<var: #blendRB type: 'unsigned int'>
	<var: #blendAG type: 'unsigned int'>
	<var: #result type: 'unsigned int'>
	<var: #alpha type: 'unsigned int'>
	<var: #unAlpha type: 'unsigned int'>
	alpha := sourceWord >> 24.  "High 8 bits of source pixel"
	alpha = 0 ifTrue: [ ^ destinationWord ].
	alpha = 255 ifTrue: [ ^ sourceWord ].
	unAlpha := 255 - alpha.

	blendRB := ((sourceWord bitAnd: 16rFF00FF) * alpha) +
				((destinationWord bitAnd: 16rFF00FF) * unAlpha)
				+ 16rFF00FF.	"blend red and blue"

	blendAG := (((sourceWord>> 8 bitOr: 16rFF0000) bitAnd: 16rFF00FF) * alpha) +
				((destinationWord>>8 bitAnd: 16rFF00FF) * unAlpha)
				+ 16rFF00FF.	"blend alpha and green"

	blendRB := blendRB + (blendRB - 16r10001 >> 8 bitAnd: 16rFF00FF) >> 8 bitAnd: 16rFF00FF.	"divide by 255"
	blendAG := blendAG + (blendAG - 16r10001 >> 8 bitAnd: 16rFF00FF) >> 8 bitAnd: 16rFF00FF.
	result := blendRB bitOr: blendAG<<8.
	^ result
! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
alphaBlendConst: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>

	^ self alphaBlendConst: sourceWord with: destinationWord paintMode: false! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
alphaBlendConst: sourceWord with: destinationWord paintMode: paintMode
	"Blend sourceWord with destinationWord using a constant alpha.
	Alpha is encoded as 0 meaning 0.0, and 255 meaning 1.0.
	The blend produced is alpha*source + (1.0-alpha)*dest, with the
	computation being performed independently on each color component.
	This function could eventually blend into any depth destination,
	using the same color averaging and mapping as warpBlt.
	paintMode = true means do nothing if the source pixel value is zero."

	"This first implementation works with dest depths of 16 and 32 bits only.
	Normal color mapping will allow sources of lower depths in this case,
	and results can be mapped directly by truncation, so no extra color maps are needed.
	To allow storing into any depth will require subsequent addition of two other
	colormaps, as is the case with WarpBlt."

	| pixMask destShifted sourceShifted destPixVal rgbMask sourcePixVal unAlpha result pixBlend shift blend maskShifted bitsPerColor blendAG blendRB |
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	<var: #blendRB type: 'unsigned int'>
	<var: #blendAG type: 'unsigned int'>
	<var: #result type: 'unsigned int'>
	<var: #sourceAlpha type: 'unsigned int'>
	<var: #unAlpha type: 'unsigned int'>
	<var: #sourceShifted type: 'unsigned int'>
	<var: #destShifted type: 'unsigned int'>
	<var: #maskShifted type: 'unsigned int'>
	<var: #pixMask type: 'unsigned int'>
	<var: #rgbMask type: 'unsigned int'>
	<var: #pixBlend type: 'unsigned int'>
	<var: #blend type: 'unsigned int'>
	destDepth < 16 ifTrue: [^ destinationWord "no-op"].
	unAlpha := 255 - sourceAlpha.
	result := destinationWord.
	destPPW = 1 ifTrue:["32bpp blends include alpha"
		paintMode & (sourceWord = 0)  "painting a transparent pixel" ifFalse:[

				blendRB := ((sourceWord bitAnd: 16rFF00FF) * sourceAlpha) +
						((destinationWord bitAnd: 16rFF00FF) * unAlpha) + 16rFF00FF.	"blendRB red and blue"

				blendAG := ((sourceWord>> 8 bitAnd: 16rFF00FF) * sourceAlpha) +
						((destinationWord>>8 bitAnd: 16rFF00FF) * unAlpha) + 16rFF00FF.	"blendRB alpha and green"

				blendRB := blendRB + (blendRB - 16r10001 >> 8 bitAnd: 16rFF00FF) >> 8 bitAnd: 16rFF00FF.	"divide by 255"
				blendAG := blendAG + (blendAG - 16r10001 >> 8 bitAnd: 16rFF00FF) >> 8 bitAnd: 16rFF00FF.
				result := blendRB bitOr: blendAG<<8.
		].
	] ifFalse:[
		pixMask := maskTable at: destDepth.
		bitsPerColor := 5.
		rgbMask := 16r1F.
		maskShifted := destMask.
		destShifted := destinationWord.
		sourceShifted := sourceWord.
		1 to: destPPW do:[:j |
			sourcePixVal := sourceShifted bitAnd: pixMask.
			((maskShifted bitAnd: pixMask) = 0  "no effect if outside of dest rectangle"
				or: [paintMode & (sourcePixVal = 0)  "or painting a transparent pixel"])
			ifFalse:
				[destPixVal := destShifted bitAnd: pixMask.
				pixBlend := 0.
				1 to: 3 do:
					[:i | shift := (i-1)*bitsPerColor.
					blend := (((sourcePixVal>>shift bitAnd: rgbMask) * sourceAlpha)
								+ ((destPixVal>>shift bitAnd: rgbMask) * unAlpha))
						 	+ 254 // 255 bitAnd: rgbMask.
					pixBlend := pixBlend bitOr: blend<<shift].
				result := (result bitAnd: (pixMask << (j-1*16)) bitInvert32)
								bitOr: pixBlend << (j-1*16)].
			maskShifted := maskShifted >> destDepth.
			sourceShifted := sourceShifted >> destDepth.
			destShifted := destShifted >> destDepth].
	].
	^ result
! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
alphaBlendScaled: sourceWord with: destinationWord
	"Blend sourceWord with destinationWord using the alpha value from sourceWord.
	Alpha is encoded as 0 meaning 0.0, and 255 meaning 1.0.
	In contrast to alphaBlend:with: the color produced is

		srcColor + (1-srcAlpha) * dstColor

	e.g., it is assumed that the source color is already scaled."
	| unAlpha rb ag |
	<inline: false>	"Do NOT inline this into optimized loops"
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	<var: #rb type: 'unsigned int'>
	<var: #ag type: 'unsigned int'>
	<var: #unAlpha type: 'unsigned int'>
	unAlpha := 255 - (sourceWord >> 24).  "High 8 bits of source pixel is source opacity (ARGB format)"
	rb := ((destinationWord bitAnd: 16rFF00FF) * unAlpha >> 8 bitAnd: 16rFF00FF) + (sourceWord bitAnd: 16rFF00FF). "blend red and blue components"
	ag := ((destinationWord>>8 bitAnd: 16rFF00FF) * unAlpha >> 8 bitAnd: 16rFF00FF) + (sourceWord>>8 bitAnd: 16rFF00FF). "blend alpha and green components"
	rb := (rb bitAnd: 16rFF00FF) bitOr: (rb bitAnd: 16r1000100) * 16rFF >> 8. "saturate red and blue components if there is a carry"
	ag := (ag bitAnd: 16rFF00FF) << 8 bitOr: (ag bitAnd: 16r1000100) * 16rFF. "saturate alpha and green components if there is a carry"
	^ag bitOr: rb "recompose"! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/31/2020 23:35'!
alphaBlendUnscaled: sourceWord with: destinationWord
	"Blend sourceWord with destinationWord using the alpha value from both sourceWord and destinationWord.
	Alpha is encoded as 0 meaning 0.0, and 255 meaning 1.0.
	The alpha channel and color produced are

		srcAlpha + (destAlpha*(1-srcAlpha))
		(srcAlpha*srcColor + (destAlpha*(1-srcAlpha)*dstColor)) / (srcAlpha + (destAlpha*(1-srcAlpha)))

	In contrast to alphaBlend:with: the method does not assume that destination form is opaque.
	In contrast to alphaBlendScaled:with: the method does not assume that colors have been pre-scaled (muliplied) by alpha channel."
	| alpha blendA result blendRB blendG |
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	<var: #blendRB type: 'unsigned int'>
	<var: #blendG type: 'unsigned int'>
	<var: #result type: 'unsigned int'>
	<var: #alpha type: 'unsigned int'>
	<var: #blendA type: 'unsigned int'>
	alpha := sourceWord >> 24.  "High 8 bits of source pixel, assuming ARGB encoding"
	alpha = 0 ifTrue: [ ^ destinationWord ].
	alpha = 255 ifTrue: [ ^ sourceWord ].
	
	blendA := 16rFF * alpha + (16rFF - alpha * (destinationWord >> 24)) + 16rFF. "blend alpha channels"
	blendA := blendA + (blendA - 1 >> 8 bitAnd: 16rFF) >> 8 bitAnd: 16rFF. "divide by 255"

	blendRB := ((sourceWord bitAnd: 16rFF00FF) * alpha) +
				((destinationWord bitAnd: 16rFF00FF) * (blendA-alpha))
				/ blendA.	"blend red and blue"

	blendG := ((sourceWord bitAnd: 16r00FF00) * alpha) +
				((destinationWord bitAnd: 16r00FF00) * (blendA-alpha))
				/ blendA.	"blend green"
	result := (blendRB bitOr: blendG) bitOr: blendA << 24.
	^ result
! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
alphaPaintConst: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>

	sourceWord = 0 ifTrue: [^ destinationWord  "opt for all-transparent source"].
	^ self alphaBlendConst: sourceWord with: destinationWord paintMode: true! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/31/2020 19:47'!
alphaScale: sourceWord with: destinationWord
	"Scale (premultiply) the destination with its alpha channel.
	Note that sourceWord is ignored."
	| alpha rb g |
	<inline: false>	"Do NOT inline this into optimized loops"
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	<var: #rb type: 'unsigned int'>
	<var: #g type: 'unsigned int'>
	<var: #alpha type: 'unsigned int'>
	alpha := destinationWord >> 24.  "High 8 bits is opacity (ARGB format)"
	rb := ((destinationWord bitAnd: 16rFF00FF) * alpha >> 8 bitAnd: 16rFF00FF). "scale red and blue components"
	g := ((destinationWord bitAnd: 16r00FF00) * alpha >> 8 bitAnd: 16r00FF00). "scale green component"
	^(g bitOr: rb) bitOr: (destinationWord bitAnd: 16rFF000000) "recompose"! !

!BitBltSimulation methodsFor: 'inner loop' stamp: 'nice 10/29/2015 12:22'!
alphaSourceBlendBits16
	"This version assumes 
		combinationRule = 34
		sourcePixSize = 32
		destPixSize = 16
		sourceForm ~= destForm.
	"
	| srcIndex dstIndex sourceWord srcAlpha destWord deltaX deltaY 
	srcY dstY dstMask srcShift ditherBase ditherIndex ditherThreshold |
	<inline: false> "This particular method should be optimized in itself"
	<var: #sourceWord type: #'unsigned int'>
	<var: #destWord type: #'unsigned int'>
	<var: #dstMask type: #'unsigned int'>
	deltaY := bbH + 1. "So we can pre-decrement"
	srcY := sy.
	dstY := dy.
	srcShift := (dx bitAnd: 1) * 16.
	destMSB ifTrue:[srcShift := 16 - srcShift].
	mask1 := 16rFFFF << (16 - srcShift).
	"This is the outer loop"
	[(deltaY := deltaY - 1) ~= 0] whileTrue:[
		srcIndex := sourceBits + (srcY * sourcePitch) + (sx * 4).
		dstIndex := destBits + (dstY * destPitch) + (dx // 2 * 4).
		ditherBase := (dstY bitAnd: 3) * 4.
		ditherIndex := (sx bitAnd: 3) - 1. "For pre-increment"
		deltaX := bbW + 1. "So we can pre-decrement"
		dstMask := mask1.
		dstMask = 16rFFFF ifTrue:[srcShift := 16] ifFalse:[srcShift := 0].

		"This is the inner loop"
		[(deltaX := deltaX - 1) ~= 0] whileTrue:[
			ditherThreshold := ditherMatrix4x4 at: ditherBase + (ditherIndex := ditherIndex + 1 bitAnd: 3).
			sourceWord := self srcLongAt: srcIndex.
			srcAlpha := sourceWord >> 24.
			srcAlpha = 255 ifTrue:[
				"Dither from 32 to 16 bit"
				sourceWord := self dither32To16: sourceWord threshold: ditherThreshold.
				sourceWord = 0 
					ifTrue:[sourceWord := 1 << srcShift]
					ifFalse: [sourceWord := sourceWord << srcShift].
				"Store masked value"
				self dstLongAt: dstIndex put: sourceWord mask: dstMask.
			] ifFalse:[ "srcAlpha ~= 255"
				srcAlpha = 0 ifFalse:[ "0 < srcAlpha < 255"
					"If we have to mix colors then just copy a single word"
					destWord := self dstLongAt: dstIndex.
					destWord := destWord bitAnd: dstMask bitInvert32.
					destWord := destWord >> srcShift.
					"Expand from 16 to 32 bit by adding zero bits"
					destWord := (((destWord bitAnd: 16r7C00) bitShift: 9) bitOr:
									((destWord bitAnd: 16r3E0) bitShift: 6)) bitOr:
								(((destWord bitAnd: 16r1F) bitShift: 3) bitOr:
									16rFF000000).
					"Mix colors"
					sourceWord := self alphaBlendScaled: sourceWord with: destWord.
					"And dither"
					sourceWord := self dither32To16: sourceWord threshold: ditherThreshold.
					sourceWord = 0 
						ifTrue:[sourceWord := 1 << srcShift]
						ifFalse:[sourceWord := sourceWord << srcShift].
					"Store back"
					self dstLongAt: dstIndex put: sourceWord mask: dstMask.
				].
			].
			srcIndex := srcIndex + 4.
			destMSB
				ifTrue:[srcShift = 0 ifTrue:[dstIndex := dstIndex + 4]]
				ifFalse:[srcShift = 0 ifFalse:[dstIndex := dstIndex + 4]].
			srcShift := srcShift bitXor: 16. "Toggle between 0 and 16"
			dstMask := dstMask bitInvert32. "Mask other half word"
		].
		srcY := srcY + 1.
		dstY := dstY + 1.
	].! !

!BitBltSimulation methodsFor: 'inner loop' stamp: 'nice 10/29/2015 12:21'!
alphaSourceBlendBits32
	"This version assumes 
		combinationRule = 34
		sourcePixSize = destPixSize = 32
		sourceForm ~= destForm.
	Note: The inner loop has been optimized for dealing
		with the special cases of srcAlpha = 0.0 and srcAlpha = 1.0 
	"
	| srcIndex dstIndex sourceWord srcAlpha destWord deltaX deltaY srcY dstY |
	<inline: false> "This particular method should be optimized in itself"
	<var: #sourceWord type: #'unsigned int'>
	<var: #destWord type: #'unsigned int'>

	"Give the compile a couple of hints"

	"The following should be declared as pointers so the compiler will
	notice that they're used for accessing memory locations 
	(good to know on an Intel architecture) but then the increments
	would be different between ST code and C code so must hope the
	compiler notices what happens (MS Visual C does)"

	deltaY := bbH + 1. "So we can pre-decrement"
	srcY := sy.
	dstY := dy.

	"This is the outer loop"
	[(deltaY := deltaY - 1) ~= 0] whileTrue:[
		srcIndex := sourceBits + (srcY * sourcePitch) + (sx * 4).
		dstIndex := destBits + (dstY * destPitch) + (dx * 4).
		deltaX := bbW + 1. "So we can pre-decrement"

		"This is the inner loop"
		[(deltaX := deltaX - 1) ~= 0] whileTrue:[
			sourceWord := self srcLongAt: srcIndex.
			srcAlpha := sourceWord >> 24.
			srcAlpha = 255 ifTrue:[
				self dstLongAt: dstIndex put: sourceWord.
				srcIndex := srcIndex + 4.
				dstIndex := dstIndex + 4.
				"Now copy as many words as possible with alpha = 255"
				[(deltaX := deltaX - 1) ~= 0 and:[
					(sourceWord := self srcLongAt: srcIndex) >> 24 = 255]]
						whileTrue:[
							self dstLongAt: dstIndex put: sourceWord.
							srcIndex := srcIndex + 4.
							dstIndex := dstIndex + 4.
						].
				"Adjust deltaX"
				deltaX := deltaX + 1.
			] ifFalse:[ "srcAlpha ~= 255"
				srcAlpha = 0 ifTrue:[
					srcIndex := srcIndex + 4.
					dstIndex := dstIndex + 4.
					"Now skip as many words as possible,"
					[(deltaX := deltaX - 1) ~= 0 and:[
						(sourceWord := self srcLongAt: srcIndex) >> 24 = 0]]
						whileTrue:[
							srcIndex := srcIndex + 4.
							dstIndex := dstIndex + 4.
						].
					"Adjust deltaX"
					deltaX := deltaX + 1.
				] ifFalse:[ "0 < srcAlpha < 255"
					"If we have to mix colors then just copy a single word"
					destWord := self dstLongAt: dstIndex.
					destWord := self alphaBlendScaled: sourceWord with: destWord.
					self dstLongAt: dstIndex put: destWord.
					srcIndex := srcIndex + 4.
					dstIndex := dstIndex + 4.
				].
			].
		].
		srcY := srcY + 1.
		dstY := dstY + 1.
	].! !

!BitBltSimulation methodsFor: 'inner loop' stamp: 'nice 10/29/2015 12:21'!
alphaSourceBlendBits8
	"This version assumes 
		combinationRule = 34
		sourcePixSize = 32
		destPixSize = 8
		sourceForm ~= destForm.
	Note: This is not real blending since we don't have the source colors available.
	"
	| srcIndex dstIndex sourceWord srcAlpha destWord deltaX deltaY 
	srcY dstY dstMask srcShift adjust mappingTable mapperFlags |
	<inline: false>
	<var: #mappingTable type:'unsigned int *'>
	<var: #sourceWord type: #'unsigned int'>
	<var: #destWord type: #'unsigned int'>
	<var: #dstMask type: #'unsigned int'>
	mappingTable := self default8To32Table.
	mapperFlags := cmFlags bitAnd: ColorMapNewStyle bitInvert32.
	deltaY := bbH + 1. "So we can pre-decrement"
	srcY := sy.
	dstY := dy.
	mask1 := ((dx bitAnd: 3) * 8).
	destMSB ifTrue:[mask1 := 24 - mask1].
	mask2 := AllOnes bitXor:(16rFF << mask1).
	(dx bitAnd: 1) = 0 
		ifTrue:[adjust := 0]
		ifFalse:[adjust := 16r1F1F1F1F].
	(dy bitAnd: 1) = 0
		ifTrue:[adjust := adjust bitXor: 16r1F1F1F1F].
	"This is the outer loop"
	[(deltaY := deltaY - 1) ~= 0] whileTrue:[
		adjust := adjust bitXor: 16r1F1F1F1F.
		srcIndex := sourceBits + (srcY * sourcePitch) + (sx * 4).
		dstIndex := destBits + (dstY * destPitch) + (dx // 4 * 4).
		deltaX := bbW + 1. "So we can pre-decrement"
		srcShift := mask1.
		dstMask := mask2.

		"This is the inner loop"
		[(deltaX := deltaX - 1) ~= 0] whileTrue:[
			sourceWord := ((self srcLongAt: srcIndex) bitAnd: (adjust bitInvert32)) + adjust.
			srcAlpha := sourceWord >> 24.
			srcAlpha > 31 ifTrue:["Everything below 31 is transparent"
				srcAlpha < 224 ifTrue:["Everything above 224 is opaque"
					destWord := self dstLongAt: dstIndex.
					destWord := destWord bitAnd: dstMask bitInvert32.
					destWord := destWord >> srcShift.
					destWord := mappingTable at: destWord.
					sourceWord := self alphaBlendScaled: sourceWord with: destWord.
				].
				sourceWord := self mapPixel: sourceWord flags: mapperFlags.
				sourceWord := sourceWord << srcShift.
				"Store back"
				self dstLongAt: dstIndex put: sourceWord mask: dstMask.
			].
			srcIndex := srcIndex + 4.
			destMSB ifTrue:[
				srcShift = 0 
					ifTrue:[dstIndex := dstIndex + 4.
							srcShift := 24.
							dstMask := 16r00FFFFFF]
					ifFalse:[srcShift := srcShift - 8.
							dstMask := (dstMask >> 8) bitOr: 16rFF000000].
			] ifFalse:[
				srcShift = 24
					ifTrue:[dstIndex := dstIndex + 4.
							srcShift := 0.
							dstMask := 16rFFFFFF00]
					ifFalse:[srcShift := srcShift + 8.
							dstMask := dstMask << 8 bitOr: 255].
			].
			adjust := adjust bitXor: 16r1F1F1F1F.
		].
		srcY := srcY + 1.
		dstY := dstY + 1.
	].! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 11/1/2020 00:05'!
alphaUnscale: sourceWord with: destinationWord
	"Unscale (divide) the destination with its alpha channel.
	Note that sourceWord is ignored."
	| alpha rb g rgb carry |
	<inline: false>	"Do NOT inline this into optimized loops"
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	<var: #rb type: 'unsigned int'>
	<var: #g type: 'unsigned int'>
	<var: #rgb type: 'unsigned int'>
	<var: #alpha type: 'unsigned int'>
	<var: #carry type: 'unsigned int'>
	alpha := destinationWord >> 24.  "High 8 bits is opacity (ARGB format)"
	alpha = 0 ifTrue: [^0].
	rb := (destinationWord bitAnd: 16rFF00FF) << 8 / alpha. "unscale red and blue components"
	g := (destinationWord bitAnd: 16r00FF00) / alpha. "unscale green component"
	carry := ((rb >> 8 bitAnd: 16rAA00AA) >> 1 bitOr: (rb >> 8 bitAnd: 16r550055) << 1)
		bitOr: ((g bitAnd: 16r00AA00) >> 1 bitOr: (g bitAnd: 16r005500) << 1).
	carry := (carry bitAnd: 16rCCCCCC) >> 2 bitOr: (carry bitAnd: 16r333333) << 2.
	carry := (carry bitAnd: 16rF0F0F0) >> 4 bitOr: (carry bitAnd: 16r0F0F0F) << 4.
	rgb := (rb bitAnd: 16rFF00FF) bitOr: (g << 8 bitAnd: 16r00FF00).
	rgb := rgb bitOr: carry.  "saturate RGB components if division overflows"
	^rgb bitOr: (destinationWord bitAnd: 16rFF000000) "restore alpha"! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
bitAnd: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^sourceWord bitAnd: destinationWord! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
bitAndInvert: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^sourceWord bitAnd: destinationWord bitInvert32! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
bitInvertAnd: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^sourceWord bitInvert32 bitAnd: destinationWord! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
bitInvertAndInvert: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^sourceWord bitInvert32 bitAnd: destinationWord bitInvert32! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
bitInvertDestination: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^destinationWord bitInvert32! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
bitInvertOr: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^sourceWord bitInvert32 bitOr: destinationWord! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
bitInvertOrInvert: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^sourceWord bitInvert32 bitOr: destinationWord bitInvert32! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
bitInvertSource: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^sourceWord bitInvert32! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
bitInvertXor: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^sourceWord bitInvert32 bitXor: destinationWord! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
bitOr: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^sourceWord bitOr: destinationWord! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
bitOrInvert: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^sourceWord bitOr: destinationWord bitInvert32! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
bitXor: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^sourceWord bitXor: destinationWord! !

!BitBltSimulation methodsFor: 'setup' stamp: 'ar (auto pragmas 12/08) 4/18/2001 20:39'!
checkSourceOverlap
	"check for possible overlap of source and destination"
	"ar 10/19/1999: This method requires surfaces to be locked."
	| t |
	<inline: true>
	(sourceForm = destForm and: [dy >= sy]) ifTrue:
		[dy > sy ifTrue:
			["have to start at bottom"
			vDir := -1.
			sy := sy + bbH - 1.
			dy := dy + bbH - 1]
		ifFalse:
			[(dy = sy) & (dx > sx) ifTrue:
				["y's are equal, but x's are backward"
				hDir := -1.
				sx := sx + bbW - 1.
				"start at right"
				dx := dx + bbW - 1.
				"and fix up masks"
				nWords > 1 ifTrue: 
					[t := mask1.
					mask1 := mask2.
					mask2 := t]]].
		"Dest inits may be affected by this change"
		destIndex := destBits + (dy * destPitch) + ((dx // destPPW) *4).
		destDelta := (destPitch * vDir) - (4 * (nWords * hDir))]! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
clearWord: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^ 0! !

!BitBltSimulation methodsFor: 'setup' stamp: 'ar 4/18/2001 21:15'!
clipRange
	"clip and adjust source origin and extent appropriately"
	"first in x"
	destX >= clipX
		ifTrue: [sx := sourceX.
				dx := destX.
				bbW := width]
		ifFalse: [sx := sourceX + (clipX - destX).
				bbW := width - (clipX - destX).
				dx := clipX].
	(dx + bbW) > (clipX + clipWidth)
		ifTrue: [bbW := bbW - ((dx + bbW) - (clipX + clipWidth))].
	"then in y"
	destY >= clipY
		ifTrue: [sy := sourceY.
				dy := destY.
				bbH := height]
		ifFalse: [sy := sourceY + clipY - destY.
				bbH := height - (clipY - destY).
				dy := clipY].
	(dy + bbH) > (clipY + clipHeight)
		ifTrue: [bbH := bbH - ((dy + bbH) - (clipY + clipHeight))].
	noSource ifTrue: [^ nil].
	sx < 0
		ifTrue: [dx := dx - sx.
				bbW := bbW + sx.
				sx := 0].
	sx + bbW > sourceWidth
		ifTrue: [bbW := bbW - (sx + bbW - sourceWidth)].
	sy < 0
		ifTrue: [dy := dy - sy.
				bbH := bbH + sy.
				sy := 0].
	sy + bbH > sourceHeight
		ifTrue: [bbH := bbH - (sy + bbH - sourceHeight)]! !

!BitBltSimulation methodsFor: 'setup' stamp: 'tpr 6/19/2013 14:34'!
copyBits
	"This function is exported for the Balloon engine"
	<export: true>
	<inline: false>
	self clipRange.
	(bbW <= 0 or: [bbH <= 0]) ifTrue:
		["zero width or height; noop"
		affectedL := affectedR := affectedT := affectedB := 0.
		^ nil].
	"Lock the surfaces"
	self lockSurfaces ifFalse:[^interpreterProxy primitiveFail].

	self
		cppIf: #'ENABLE_FAST_BLT'
		ifTrue:["you really, really mustn't call this unless you have the rest of the code to link to"
			self copyBitsFastPathSpecialised]
		ifFalse:[self copyBitsLockedAndClipped].

	self unlockSurfaces.! !

!BitBltSimulation methodsFor: 'setup' stamp: 'tpr 5/15/2021 14:23'!
copyBits: op Fallback: flags
	"Recover from the fast path specialised code saying Help-I-cant-cope"
	|done |
	<static: false>
	<returnTypeC: 'void'>
	<inline: false>
	<var: #op type: 'operation_t *'>
	<var: #flags type:'unsigned int'>
	self cppIf: #'ENABLE_FAST_BLT'
		ifTrue:[
			"recover values from the operation struct used by the fast ARM code"
			self cCode:'
	combinationRule = op->combinationRule;
	noSource = op->noSource;
	sourceBits = (sqInt) op->src.bits;
	sourcePitch = op->src.pitch;
	sourceDepth = op->src.depth;
	sourceMSB = op->src.msb;
	sx = op->src.x;
	sy = op->src.y;
	destBits = (sqInt) op->dest.bits;
	destPitch = op->dest.pitch;
	destDepth = op->dest.depth;
	destMSB = op->dest.msb;
	dx = op->dest.x;
	dy = op->dest.y;
	bbW = op->width;
	bbH = op->height;
	cmFlags = op->cmFlags;
	cmShiftTable = (void *) op->cmShiftTable;
	cmMaskTable = (void *) op->cmMaskTable;
	cmMask = op->cmMask;
	cmLookupTable = (void *) op->cmLookupTable;
	noHalftone = op->noHalftone;
	halftoneHeight = op->halftoneHeight;
	halftoneBase = (sqInt) op->halftoneBase;
	if (combinationRule == 30 || combinationRule == 31) {
		sourceAlpha = op->opt.sourceAlpha;
	}
	if (combinationRule == 41) {
		componentAlphaModeColor = op->opt.componentAlpha.componentAlphaModeColor;
		componentAlphaModeAlpha = op->opt.componentAlpha.componentAlphaModeAlpha;
		gammaLookupTable = (void *) op->opt.componentAlpha.gammaLookupTable;
		ungammaLookupTable = (void *) op->opt.componentAlpha.ungammaLookupTable;
	}'.
	
				destPPW := 32 / destDepth.
				sourcePPW := 32 / sourceDepth.
				cmBitsPerColor := 0.
				cmMask = 16r1FF ifTrue: [cmBitsPerColor := 3].
				cmMask = 16rFFF ifTrue: [cmBitsPerColor := 4].
				cmMask = 16r7FFF ifTrue: [cmBitsPerColor := 5].
				" In some places, sourceForm and destForm are compared in order to detect
				 whether we're reading and writing the same image. However, these have
				 not always been initialised by the time we get here, so substitute
				 sourceBits and destBits if so. "
				(sourceForm == 0 and: [destForm == 0])
					ifTrue:
						[sourceForm := sourceBits.
						destForm := destBits].
	
				"Try a shortcut for stuff that should be run as quickly as possible"
				done := self tryCopyingBitsQuickly.
				done ifTrue:[^nil].

				bitCount := 0.
				"Choose and perform the actual copy loop."
				self performCopyLoop]

	


! !

!BitBltSimulation methodsFor: 'setup' stamp: 'tpr 6/19/2013 14:39'!
copyBitsFastPathSpecialised
	"Perform the actual copyBits operation using the fast path specialised code; fail some cases by falling back to normal code.
	Assume: Surfaces have been locked and clipping was performed."
	<inline: false>

	self
		cppIf: #'ENABLE_FAST_BLT'
		ifTrue:[
	"set the affected area to 0 first"
	affectedL := affectedR := affectedT := affectedB := 0.
	
	self copyBitsRule41Test.	
	(interpreterProxy failed not)
		ifFalse: [^ interpreterProxy primitiveFail].

 	"we skip the tryCopyingBitsQuickly and leave that to falback code"
	 
	(combinationRule = 30) | (combinationRule = 31) ifTrue:
		["Check and fetch source alpha parameter for alpha blend"
		interpreterProxy methodArgumentCount = 1
			ifTrue: [sourceAlpha := interpreterProxy stackIntegerValue: 0.
					(interpreterProxy failed not and: [(sourceAlpha >= 0) & (sourceAlpha <= 255)])
						ifFalse: [^ interpreterProxy primitiveFail]]
			ifFalse: [^ interpreterProxy primitiveFail]].

	"we don't worry about bitCount"
	"bitCount := 0."

	"We don't  do - Choose and perform the actual copy loop."
	"self performCopyLoop."

	"this is done inversely to plain copyBitsLockedAndClipped"
	(combinationRule ~= 22) & (combinationRule ~= 32) ifTrue:
		["zero width and height; return the count"
		affectedL := dx.
		affectedR := dx + bbW.
		affectedT := dy.
		affectedB := dy + bbH].
	
	"Now we fill the 'operation' structure and pass it to the sneaky ARM code"
	self cCode:'
	// fill the operation structure
	operation_t op;
	op.combinationRule = combinationRule;
	op.noSource = noSource;
	op.src.bits = (void *) sourceBits;
	op.src.pitch = sourcePitch;
	op.src.depth = sourceDepth;
	op.src.msb = sourceMSB;
	op.src.x = sx;
	op.src.y = sy;
	op.dest.bits = (void *) destBits;
	op.dest.pitch = destPitch;
	op.dest.depth = destDepth;
	op.dest.msb = destMSB;
	op.dest.x = dx;
	op.dest.y = dy;
	op.width = bbW;
	op.height = bbH;
	op.cmFlags = cmFlags;
	op.cmShiftTable = (void *) cmShiftTable;
	op.cmMaskTable = (void *) cmMaskTable;
	op.cmMask = cmMask;
	op.cmLookupTable = (void *) cmLookupTable;
	op.noHalftone = noHalftone;
	op.halftoneHeight = halftoneHeight;
	op.halftoneBase = (void *) halftoneBase;
	if (combinationRule == 30 || combinationRule == 31) {
		op.opt.sourceAlpha = sourceAlpha;
	}
	if (combinationRule == 41) {
		op.opt.componentAlpha.componentAlphaModeColor = componentAlphaModeColor;
		op.opt.componentAlpha.componentAlphaModeAlpha = componentAlphaModeAlpha;
		op.opt.componentAlpha.gammaLookupTable = (void *) gammaLookupTable;
		op.opt.componentAlpha.ungammaLookupTable = (void *) ungammaLookupTable;
	}
	// call the sneaky code
	copyBitsDispatch(&op)'
	]! !

!BitBltSimulation methodsFor: 'setup' stamp: 'ar (auto pragmas 12/08) 2/20/2000 19:42'!
copyBitsFrom: startX to: stopX at: yValue
	"Support for the balloon engine."
	<export: true>
	destX := startX.
	destY := yValue.
	sourceX := startX.
	width := (stopX - startX).
	self copyBits.
	self showDisplayBits.! !

!BitBltSimulation methodsFor: 'setup' stamp: 'eem 10/16/2018 10:13'!
copyBitsLockedAndClipped
	"Perform the actual copyBits operation.
	Assume: Surfaces have been locked and clipping was performed."
	<inline: false>
	
	self copyBitsRule41Test.	
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFail].

 	"Try a shortcut for stuff that should be run as quickly as possible"
	self tryCopyingBitsQuickly ifTrue:
		[^nil].

	(combinationRule between: 30 and: 31) ifTrue:
		["Check and fetch source alpha parameter for alpha blend"
		 interpreterProxy methodArgumentCount = 1 ifFalse:
			[^interpreterProxy primitiveFail].
		 sourceAlpha := interpreterProxy stackIntegerValue: 0.
		 (interpreterProxy failed
		  or: [sourceAlpha < 0
		  or: [sourceAlpha > 255]]) ifTrue:
			[^interpreterProxy primitiveFail]].

	bitCount := 0.
	"Choose and perform the actual copy loop."
	self performCopyLoop.

	(combinationRule between: 30 and: 31)
		ifTrue:"zero width and height; just return the count"
			[affectedL := affectedR := affectedT := affectedB := 0]
		ifFalse:
			[hDir > 0
				ifTrue: [affectedL := dx.
						affectedR := dx + bbW]
				ifFalse: [affectedL := dx - bbW + 1.
						affectedR := dx + 1].
			 vDir > 0
				ifTrue: [affectedT := dy.
						affectedB := dy + bbH]
				ifFalse: [affectedT := dy - bbH + 1.
						affectedB := dy + 1]]! !

!BitBltSimulation methodsFor: 'setup' stamp: 'tpr 5/27/2013 18:12'!
copyBitsRule41Test
	"Test possible use of rule 41, rgbComponentAlpha:with: Nothing to return, just set up some variables"
	| gammaLookupTableOop ungammaLookupTableOop |
	<inline: false>
	
	combinationRule = 41
		ifTrue:["fetch the forecolor into componentAlphaModeColor."
			componentAlphaModeAlpha := 255.
			componentAlphaModeColor := 16777215.
			gammaLookupTable := nil.
			ungammaLookupTable := nil.
			interpreterProxy methodArgumentCount >= 2
				ifTrue:[
					componentAlphaModeAlpha := interpreterProxy stackIntegerValue: (interpreterProxy methodArgumentCount - 2).
					(interpreterProxy failed not)
						ifFalse: [^ interpreterProxy primitiveFail].
					componentAlphaModeColor := interpreterProxy stackIntegerValue: (interpreterProxy methodArgumentCount - 1).
					(interpreterProxy failed not)
						ifFalse: [^ interpreterProxy primitiveFail].
					interpreterProxy methodArgumentCount = 4
						ifTrue:[
							gammaLookupTableOop := interpreterProxy stackObjectValue: 1.
							(interpreterProxy isBytes: gammaLookupTableOop) 
								ifTrue:[gammaLookupTable := interpreterProxy firstIndexableField: gammaLookupTableOop.].
							ungammaLookupTableOop := interpreterProxy stackObjectValue: 0.
							(interpreterProxy isBytes: ungammaLookupTableOop) 
								ifTrue:[ungammaLookupTable := interpreterProxy firstIndexableField: ungammaLookupTableOop]]]
				ifFalse:[
					interpreterProxy methodArgumentCount = 1
						ifTrue: [
							componentAlphaModeColor := interpreterProxy stackIntegerValue: 0.
							(interpreterProxy failed not)
								ifFalse: [^ interpreterProxy primitiveFail]]
						ifFalse:[^ interpreterProxy primitiveFail]]].	


! !

!BitBltSimulation methodsFor: 'inner loop' stamp: 'nice 10/30/2019 23:04'!
copyLoop
	| prevWord thisWord skewWord halftoneWord mergeWord hInc y unskew skewMask notSkewMask mergeFnwith destWord |
	"This version of the inner loop assumes noSource = false."
	<inline: false>
	<var: #prevWord type: #'unsigned int'>
	<var: #thisWord type: #'unsigned int'>
	<var: #skewWord type: #'unsigned int'>
	<var: #halftoneWord type: #'unsigned int'>
	<var: #mergeWord type: #'unsigned int'>
	<var: #destWord type: #'unsigned int'>
	<var: #skewMask type: #'unsigned int'>
	<var: #notSkewMask type: #'unsigned int'>
	<var: #unskew type: #int> "unskew is a bitShift and MUST remain signed, while skewMask is unsigned."
	<var: #mergeFnwith declareC: 'unsigned int (*mergeFnwith)(unsigned int, unsigned int)'>
	mergeFnwith := self cCoerce: (opTable at: combinationRule+1) to: 'unsigned int (*)(unsigned int, unsigned int)'.
	mergeFnwith.  "null ref for compiler"

	self deny: (preload and: [skew = 0]).
	self assert: (skew between: -31 and: 31).

	hInc := hDir * 4.  "Byte delta"
	skew < 0
		ifTrue: [unskew := skew + 32. skewMask := AllOnes << (0 - skew).
				self cCode: [] inSmalltalk: [skewMask := skewMask bitAnd: 16rFFFFFFFF]]
		ifFalse:
			[skew = 0
				ifTrue: [unskew := 0. skewMask := AllOnes]
				ifFalse: [unskew := skew - 32. skewMask := AllOnes >> skew]].

	notSkewMask := skewMask bitInvert32.
	noHalftone
		ifTrue: [halftoneWord := AllOnes.  halftoneHeight := 0]
		ifFalse: [halftoneWord := self halftoneAt: 0].

	y := dy.
	"Here is the vertical loop, in two versions, one for the combinationRule = 3 copy mode, one for the general case."
	combinationRule = 3
		ifTrue:
			[1 to: bbH do: "here is the vertical loop for combinationRule = 3 copy mode; no need to call merge"
				[ :i |
				halftoneHeight > 1 ifTrue:  "Otherwise, its always the same"
					[halftoneWord := self halftoneAt: y.
					y := y + vDir].
				preload
					ifTrue: "load the 64-bit shifter"
						[prevWord := self srcLongAt: sourceIndex.
						self incSrcIndex: hInc]
					ifFalse:
						[prevWord := 0].

				"Note: the horizontal loop has been expanded into three parts for speed:"

				"This first section requires masking of the destination store..."
				destMask := mask1.
				thisWord := self srcLongAt: sourceIndex.  "pick up next word"
				self incSrcIndex: hInc.
				skewWord := ((prevWord bitAnd: notSkewMask) bitShift: unskew)
								bitOr:  "32-bit rotate"
									((thisWord bitAnd: skewMask) bitShift: skew).
				prevWord := thisWord.
				destWord := self dstLongAt: destIndex.
				destWord := (destMask bitAnd: (skewWord bitAnd: halftoneWord))
								bitOr: (destWord bitAnd: destMask bitInvert32).
				self dstLongAt: destIndex put: destWord.
				self incDestIndex: hInc.

				"This central horizontal loop requires no store masking"
				destMask := AllOnes.
				(skew = 0 and: [halftoneWord = AllOnes])
					ifTrue: "Very special inner loop for STORE mode with no skew -- just move words"
						[(preload and: [hDir = 1])
							ifTrue:
								[2 to: nWords-1 do: 
									[ :word |  "Note loop starts with prevWord loaded (due to preload)"
									self dstLongAt: destIndex put: prevWord.
									self incDestIndex: hInc.
									prevWord := self srcLongAt: sourceIndex.
									self incSrcIndex: hInc]]
							ifFalse:
								[2 to: nWords-1 do: 
									[ :word |
									thisWord := self srcLongAt: sourceIndex.
									self incSrcIndex: hInc.
									self dstLongAt: destIndex put: thisWord.
									self incDestIndex: hInc].
								 prevWord := thisWord]]
						ifFalse:
							[2 to: nWords-1 do: 
								[ :word |
								thisWord := self srcLongAt: sourceIndex.
								self incSrcIndex: hInc.
								skewWord := ((prevWord bitAnd: notSkewMask) bitShift: unskew)
												bitOr:  "32-bit rotate"
											((thisWord bitAnd: skewMask) bitShift: skew).
								prevWord := thisWord.
								self dstLongAt: destIndex put: (skewWord bitAnd: halftoneWord).
								self incDestIndex: hInc]].

				"This last section, if used, requires masking of the destination store..."
				nWords > 1 ifTrue:
					[destMask := mask2.
					thisWord :=((skewMask bitShift: skew) bitAnd: mask2) = 0
						ifTrue: [0 "we don't need more bits, they will all come from prevWord"]
						ifFalse: [self srcLongAt: sourceIndex.  "pick up last bits from next word".].
					self incSrcIndex: hInc. "Note: this will be undone by inncSrcIndex: sourceDelta below if undue"
					skewWord := ((prevWord bitAnd: notSkewMask) bitShift: unskew)
									bitOr:  "32-bit rotate"
										((thisWord bitAnd: skewMask) bitShift: skew).
					destWord := self dstLongAt: destIndex.
					destWord := (destMask bitAnd: (skewWord bitAnd: halftoneWord))
									bitOr: (destWord bitAnd: destMask bitInvert32).
					self dstLongAt: destIndex put: destWord.
					self incDestIndex: hInc].

				self incSrcIndex: sourceDelta.
				self incDestIndex: destDelta]]
		ifFalse:
			[1 to: bbH do: "here is the vertical loop for the general case (combinationRule ~= 3)"
				[ :i |
				halftoneHeight > 1 ifTrue:  "Otherwise, its always the same"
					[halftoneWord := self halftoneAt: y.
					y := y + vDir].
				preload
					ifTrue: "load the 64-bit shifter"
						[prevWord := self srcLongAt: sourceIndex.
						self incSrcIndex: hInc]
					ifFalse:
						[prevWord := 0].

				"Note: the horizontal loop has been expanded into three parts for speed:"

				"This first section requires masking of the destination store..."
				destMask := mask1.
				thisWord := self srcLongAt: sourceIndex.  "pick up next word"
				self incSrcIndex: hInc.
				skewWord := ((prevWord bitAnd: notSkewMask) bitShift: unskew)
								bitOr:  "32-bit rotate"
							((thisWord bitAnd: skewMask) bitShift: skew).
				prevWord := thisWord.
				destWord := self dstLongAt: destIndex.
				mergeWord := self mergeFn: (skewWord bitAnd: halftoneWord) with: destWord.
				destWord := (destMask bitAnd: mergeWord)
								bitOr: (destWord bitAnd: destMask bitInvert32).
				self dstLongAt: destIndex put: destWord.
				self incDestIndex: hInc.

				"This central horizontal loop requires no store masking"
				destMask := AllOnes.
				2 to: nWords-1 do: "Normal inner loop does merge:"
					[ :word |
					thisWord := self srcLongAt: sourceIndex.  "pick up next word"
					self incSrcIndex: hInc.
					skewWord := ((prevWord bitAnd: notSkewMask) bitShift: unskew)
									bitOr:  "32-bit rotate"
								((thisWord bitAnd: skewMask) bitShift: skew).
					prevWord := thisWord.
					mergeWord := self mergeFn: (skewWord bitAnd: halftoneWord)
									with: (self dstLongAt: destIndex).
					self dstLongAt: destIndex put: mergeWord.
					self incDestIndex: hInc].

				"This last section, if used, requires masking of the destination store..."
				nWords > 1 ifTrue:
					[destMask := mask2.
					thisWord :=((skewMask bitShift: skew) bitAnd: mask2) = 0
						ifTrue: [0 "we don't need more bits, they will all come from prevWord"]
						ifFalse: [self srcLongAt: sourceIndex.  "pick up last bits from next word".].
					self incSrcIndex: hInc. "Note: this will be undone by incSrcIndex: sourceDelta below if undue"
					skewWord := ((prevWord bitAnd: notSkewMask) bitShift: unskew)
									bitOr:  "32-bit rotate"
								((thisWord bitAnd: skewMask) bitShift: skew).
					destWord := self dstLongAt: destIndex.
					mergeWord := self mergeFn: (skewWord bitAnd: halftoneWord) with: destWord.
					destWord := (destMask bitAnd: mergeWord)
									bitOr: (destWord bitAnd: destMask bitInvert32).
					self dstLongAt: destIndex put: destWord.
					self incDestIndex: hInc].

				self incSrcIndex: sourceDelta.
				self incDestIndex: destDelta]]! !

!BitBltSimulation methodsFor: 'inner loop' stamp: 'eem 10/16/2018 16:27'!
copyLoopNoSource
	"Faster copyLoop when source not used.  hDir and vDir are both
	positive, and perload and skew are unused"
	| halftoneWord mergeWord mergeFnwith destWord |
	<inline: false>
	<var: #mergeFnwith declareC: 'unsigned int (*mergeFnwith)(unsigned int, unsigned int)'>
	<var: #halftoneWord type: #'unsigned int'>
	<var: #mergeWord type: #'unsigned int'>
	<var: #destWord type: #'unsigned int'>
	mergeFnwith := self cCoerce: (opTable at: combinationRule+1) to: 'unsigned int (*)(unsigned int, unsigned int)'.
	mergeFnwith.  "null ref for compiler"

	noHalftone ifTrue:
		[halftoneWord := AllOnes].
	1 to: bbH do: "here is the vertical loop"
		[ :i |
		noHalftone ifFalse:
			[halftoneWord := self halftoneAt: dy+i-1].

	"Note: the horizontal loop has been expanded into three parts for speed:"

		"This first section requires masking of the destination store..."
		destMask := mask1.
		destWord := self dstLongAt: destIndex.
		mergeWord := self mergeFn: halftoneWord
						with: destWord.
		destWord := (destMask bitAnd: mergeWord) bitOr: 
						(destWord bitAnd: destMask bitInvert32).
		self dstLongAt: destIndex put: destWord.
		self incDestIndex: 4.

	"This central horizontal loop requires no store masking"
		destMask := AllOnes.
		combinationRule = 3
			ifTrue: "Special inner loop for STORE"
				[destWord := halftoneWord.
				2 to: nWords-1 do:[ :word |
					self dstLongAt: destIndex put: destWord.
					self incDestIndex: 4]]
			ifFalse: "Normal inner loop does merge"
				[2 to: nWords-1 do:[ :word | "Normal inner loop does merge"
					destWord := self dstLongAt: destIndex.
					mergeWord := self mergeFn: halftoneWord with: destWord.
					self dstLongAt: destIndex put: mergeWord.
					self incDestIndex: 4]].

		"This last section, if used, requires masking of the destination store..."
		nWords > 1 ifTrue:
			[destMask := mask2.
			destWord := self dstLongAt: destIndex.
			mergeWord := self mergeFn: halftoneWord with: destWord.
			destWord := (destMask bitAnd: mergeWord) bitOr:
							(destWord bitAnd: destMask bitInvert32).
			self dstLongAt: destIndex put: destWord.
			self incDestIndex: 4].

		self incDestIndex: destDelta]! !

!BitBltSimulation methodsFor: 'inner loop' stamp: 'eem 10/16/2018 14:56'!
copyLoopPixMap
	"This version of the inner loop maps source pixels
	to a destination form with different depth.  Because it is already
	unweildy, the loop is not unrolled as in the other versions.
	Preload, skew and skewMask are all overlooked, since pickSourcePixels
	delivers its destination word already properly aligned.
	Note that pickSourcePixels could be copied in-line at the top of
	the horizontal loop, and some of its inits moved out of the loop."
	"ar 12/7/1999:
	The loop has been rewritten to use only one pickSourcePixels call.
	The idea is that the call itself could be inlined. If we decide not
	to inline pickSourcePixels we could optimize the loop instead."
	| skewWord halftoneWord mergeWord scrStartBits nSourceIncs startBits endBits sourcePixMask destPixMask mergeFnwith nPix srcShift dstShift destWord words srcShiftInc dstShiftInc dstShiftLeft mapperFlags |
	<inline: false>
	<var: #mergeFnwith declareC: 'unsigned int (*mergeFnwith)(unsigned int, unsigned int)'>
	<var: #skewWord type: #'unsigned int'>
	<var: #halftoneWord type: #'unsigned int'>
	<var: #mergeWord type: #'unsigned int'>
	<var: #destWord type: #'unsigned int'>
	<var: #sourcePixMask type: #'unsigned int'>
	<var: #destPixMask type: #'unsigned int'>
	mergeFnwith := self cCoerce: (opTable at: combinationRule+1) to: 'unsigned int (*)(unsigned int, unsigned int)'.
	mergeFnwith.  "null ref for compiler"

	"Additional inits peculiar to unequal source and dest pix size..."
	sourcePPW := 32//sourceDepth.
	sourcePixMask := maskTable at: sourceDepth.
	destPixMask := maskTable at: destDepth.
	mapperFlags := cmFlags bitAnd: ColorMapNewStyle bitInvert32.
	sourceIndex := sourceBits +
					(sy * sourcePitch) + ((sx // sourcePPW) *4).
	scrStartBits := sourcePPW - (sx bitAnd: sourcePPW-1).
	bbW < scrStartBits
		ifTrue: [nSourceIncs := 0]
		ifFalse: [nSourceIncs := (bbW - scrStartBits)//sourcePPW + 1].
	sourceDelta := sourcePitch - (nSourceIncs * 4).

	"Note following two items were already calculated in destmask setup!!"
	startBits := destPPW - (dx bitAnd: destPPW-1).
	endBits := ((dx + bbW - 1) bitAnd: destPPW-1) + 1.

	bbW < startBits ifTrue:[startBits := bbW].

	"Precomputed shifts for pickSourcePixels"
	srcShift := ((sx bitAnd: sourcePPW - 1) * sourceDepth).
	dstShift := ((dx bitAnd: destPPW - 1) * destDepth).
	srcShiftInc := sourceDepth.
	dstShiftInc := destDepth.
	dstShiftLeft := 0.
	sourceMSB ifTrue:[
		srcShift := 32 - sourceDepth - srcShift.
		srcShiftInc := 0 - srcShiftInc].
	destMSB ifTrue:[
		dstShift := 32 - destDepth - dstShift.
		dstShiftInc := 0 - dstShiftInc.
		dstShiftLeft := 32 - destDepth].
	noHalftone ifTrue:
		[halftoneWord := AllOnes].
	1 to: bbH do: "here is the vertical loop"
		[ :i |
		noHalftone ifFalse:
			[halftoneWord := self halftoneAt: dy+i-1].
		"setup first load"
		srcBitShift := srcShift.
		dstBitShift := dstShift.
		destMask := mask1.
		nPix := startBits.
		"Here is the horizontal loop..."
		words := nWords.
			["pick up the word"
			skewWord := self pickSourcePixels: nPix flags: mapperFlags 
								srcMask: sourcePixMask destMask: destPixMask
								srcShiftInc: srcShiftInc dstShiftInc: dstShiftInc.
			"align next word to leftmost pixel"
			dstBitShift := dstShiftLeft.

			destMask = AllOnes ifTrue:["avoid read-modify-write"
				mergeWord := self mergeFn: (skewWord bitAnd: halftoneWord)
								with: (self dstLongAt: destIndex).
				self dstLongAt: destIndex put: (destMask bitAnd: mergeWord).
			] ifFalse:[ "General version using dest masking"
				destWord := self dstLongAt: destIndex.
				mergeWord := self mergeFn: (skewWord bitAnd: halftoneWord)
								with: (destWord bitAnd: destMask).
				destWord := (destMask bitAnd: mergeWord) bitOr:
								(destWord bitAnd: destMask bitInvert32).
				self dstLongAt: destIndex put: destWord.
			].
			self incDestIndex: 4.
			words = 2 "e.g., is the next word the last word?"
				ifTrue:["set mask for last word in this row"
						destMask := mask2.
						nPix := endBits]
				ifFalse:["use fullword mask for inner loop"
						destMask := AllOnes.
						nPix := destPPW].
			(words := words - 1) = 0] whileFalse.
		"--- end of inner loop ---"
		self incSrcIndex: sourceDelta.
		self incDestIndex: destDelta]
! !

!BitBltSimulation methodsFor: 'pixel mapping' stamp: 'ar (auto pragmas 12/08) 11/16/1998 00:23'!
default8To32Table
	"Return the default translation table from 1..8 bit indexed colors to 32bit"
	"The table has been generated by the following statements"
	"| pvs hex |
	String streamContents:[:s|
		s nextPutAll:'static unsigned int theTable[256] = { '.
		pvs := (Color colorMapIfNeededFrom: 8 to: 32) asArray.
		1 to: pvs size do:[:i|
			i > 1 ifTrue:[s nextPutAll:', '].
			(i-1 \\ 8) = 0 ifTrue:[s cr].
			s nextPutAll:'0x'.
			hex := (pvs at: i) printStringBase: 16.
			s nextPutAll: (hex copyFrom: 4 to: hex size).
		].
		s nextPutAll:'};'.
	]."
	| theTable |
	<returnTypeC:'unsigned int *'>
	<var: #theTable declareC:'static unsigned int theTable[256] = { 
0x0, 0xFF000001, 0xFFFFFFFF, 0xFF808080, 0xFFFF0000, 0xFF00FF00, 0xFF0000FF, 0xFF00FFFF, 
0xFFFFFF00, 0xFFFF00FF, 0xFF202020, 0xFF404040, 0xFF606060, 0xFF9F9F9F, 0xFFBFBFBF, 0xFFDFDFDF, 
0xFF080808, 0xFF101010, 0xFF181818, 0xFF282828, 0xFF303030, 0xFF383838, 0xFF484848, 0xFF505050, 
0xFF585858, 0xFF686868, 0xFF707070, 0xFF787878, 0xFF878787, 0xFF8F8F8F, 0xFF979797, 0xFFA7A7A7, 
0xFFAFAFAF, 0xFFB7B7B7, 0xFFC7C7C7, 0xFFCFCFCF, 0xFFD7D7D7, 0xFFE7E7E7, 0xFFEFEFEF, 0xFFF7F7F7, 
0xFF000001, 0xFF003300, 0xFF006600, 0xFF009900, 0xFF00CC00, 0xFF00FF00, 0xFF000033, 0xFF003333, 
0xFF006633, 0xFF009933, 0xFF00CC33, 0xFF00FF33, 0xFF000066, 0xFF003366, 0xFF006666, 0xFF009966, 
0xFF00CC66, 0xFF00FF66, 0xFF000099, 0xFF003399, 0xFF006699, 0xFF009999, 0xFF00CC99, 0xFF00FF99, 
0xFF0000CC, 0xFF0033CC, 0xFF0066CC, 0xFF0099CC, 0xFF00CCCC, 0xFF00FFCC, 0xFF0000FF, 0xFF0033FF, 
0xFF0066FF, 0xFF0099FF, 0xFF00CCFF, 0xFF00FFFF, 0xFF330000, 0xFF333300, 0xFF336600, 0xFF339900, 
0xFF33CC00, 0xFF33FF00, 0xFF330033, 0xFF333333, 0xFF336633, 0xFF339933, 0xFF33CC33, 0xFF33FF33, 
0xFF330066, 0xFF333366, 0xFF336666, 0xFF339966, 0xFF33CC66, 0xFF33FF66, 0xFF330099, 0xFF333399, 
0xFF336699, 0xFF339999, 0xFF33CC99, 0xFF33FF99, 0xFF3300CC, 0xFF3333CC, 0xFF3366CC, 0xFF3399CC, 
0xFF33CCCC, 0xFF33FFCC, 0xFF3300FF, 0xFF3333FF, 0xFF3366FF, 0xFF3399FF, 0xFF33CCFF, 0xFF33FFFF, 
0xFF660000, 0xFF663300, 0xFF666600, 0xFF669900, 0xFF66CC00, 0xFF66FF00, 0xFF660033, 0xFF663333, 
0xFF666633, 0xFF669933, 0xFF66CC33, 0xFF66FF33, 0xFF660066, 0xFF663366, 0xFF666666, 0xFF669966, 
0xFF66CC66, 0xFF66FF66, 0xFF660099, 0xFF663399, 0xFF666699, 0xFF669999, 0xFF66CC99, 0xFF66FF99, 
0xFF6600CC, 0xFF6633CC, 0xFF6666CC, 0xFF6699CC, 0xFF66CCCC, 0xFF66FFCC, 0xFF6600FF, 0xFF6633FF, 
0xFF6666FF, 0xFF6699FF, 0xFF66CCFF, 0xFF66FFFF, 0xFF990000, 0xFF993300, 0xFF996600, 0xFF999900, 
0xFF99CC00, 0xFF99FF00, 0xFF990033, 0xFF993333, 0xFF996633, 0xFF999933, 0xFF99CC33, 0xFF99FF33, 
0xFF990066, 0xFF993366, 0xFF996666, 0xFF999966, 0xFF99CC66, 0xFF99FF66, 0xFF990099, 0xFF993399, 
0xFF996699, 0xFF999999, 0xFF99CC99, 0xFF99FF99, 0xFF9900CC, 0xFF9933CC, 0xFF9966CC, 0xFF9999CC, 
0xFF99CCCC, 0xFF99FFCC, 0xFF9900FF, 0xFF9933FF, 0xFF9966FF, 0xFF9999FF, 0xFF99CCFF, 0xFF99FFFF, 
0xFFCC0000, 0xFFCC3300, 0xFFCC6600, 0xFFCC9900, 0xFFCCCC00, 0xFFCCFF00, 0xFFCC0033, 0xFFCC3333, 
0xFFCC6633, 0xFFCC9933, 0xFFCCCC33, 0xFFCCFF33, 0xFFCC0066, 0xFFCC3366, 0xFFCC6666, 0xFFCC9966, 
0xFFCCCC66, 0xFFCCFF66, 0xFFCC0099, 0xFFCC3399, 0xFFCC6699, 0xFFCC9999, 0xFFCCCC99, 0xFFCCFF99, 
0xFFCC00CC, 0xFFCC33CC, 0xFFCC66CC, 0xFFCC99CC, 0xFFCCCCCC, 0xFFCCFFCC, 0xFFCC00FF, 0xFFCC33FF, 
0xFFCC66FF, 0xFFCC99FF, 0xFFCCCCFF, 0xFFCCFFFF, 0xFFFF0000, 0xFFFF3300, 0xFFFF6600, 0xFFFF9900, 
0xFFFFCC00, 0xFFFFFF00, 0xFFFF0033, 0xFFFF3333, 0xFFFF6633, 0xFFFF9933, 0xFFFFCC33, 0xFFFFFF33, 
0xFFFF0066, 0xFFFF3366, 0xFFFF6666, 0xFFFF9966, 0xFFFFCC66, 0xFFFFFF66, 0xFFFF0099, 0xFFFF3399, 
0xFFFF6699, 0xFFFF9999, 0xFFFFCC99, 0xFFFFFF99, 0xFFFF00CC, 0xFFFF33CC, 0xFFFF66CC, 0xFFFF99CC, 
0xFFFFCCCC, 0xFFFFFFCC, 0xFFFF00FF, 0xFFFF33FF, 0xFFFF66FF, 0xFFFF99FF, 0xFFFFCCFF, 0xFFFFFFFF};'>
	^theTable! !

!BitBltSimulation methodsFor: 'pixel mapping' stamp: 'ar (auto pragmas 12/08) 10/27/1999 17:54'!
deltaFrom: x1 to: x2 nSteps: n
	"Utility routine for computing Warp increments."
	<inline: true>
	x2 > x1
		ifTrue: [^ x2 - x1 + FixedPt1 // (n+1) + 1]
		ifFalse: [x2 = x1 ifTrue: [^ 0].
				^ 0 - (x1 - x2 + FixedPt1 // (n+1) + 1)]! !

!BitBltSimulation methodsFor: 'setup' stamp: 'nice 2/22/2020 00:45'!
destMaskAndPointerInit
	"Compute masks for left and right destination words"
	| startBits pixPerM1 endBits |
	<inline: true>
	pixPerM1 := destPPW - 1.  "A mask, assuming power of two"
	"how many pixels in first word"
	startBits := destPPW - (dx bitAnd: pixPerM1).
	"how many pixels in last word"
	endBits := (dx + bbW - 1 bitAnd: pixPerM1) + 1.
	destMSB
		ifTrue:
			[mask1 := AllOnes >> (32 - (startBits * destDepth)).
			 mask2 := AllOnes << (32 - (endBits * destDepth)) bitAnd: AllOnes] 
		ifFalse:
			[mask1 := AllOnes << (32 - (startBits * destDepth)) bitAnd: AllOnes.
			 mask2 := AllOnes >> (32 - (endBits * destDepth))].

	"determine number of words stored per line; merge masks if only 1"
	bbW <= startBits
		ifTrue: [mask1 := mask1 bitAnd: mask2.
				mask2 := 0.
				nWords := 1]
		ifFalse: [nWords := bbW - startBits + pixPerM1 // destPPW + 1].
	hDir := vDir := 1. "defaults for no overlap with source"

	"calculate byte addr and delta, based on first word of data"
	"Note pitch is bytes and nWords is longs, not bytes"
	destIndex := destBits + (dy * destPitch) + ((dx // destPPW) * 4).
	destDelta := destPitch * vDir - (4 * (nWords * hDir)) "byte addr delta"! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:23'!
destinationWord: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^destinationWord! !

!BitBltSimulation methodsFor: 'pixel mapping' stamp: 'nice 10/29/2015 22:24'!
dither32To16: srcWord threshold: ditherValue
	"Dither the given 32bit word to 16 bit. Ignore alpha."
	| addThreshold  |
	<inline: true> "You bet"
	<returnTypeC: 'unsigned int'>
	<var: #srcWord type: 'unsigned int'>
	addThreshold := ditherValue bitShift: 8.
	^((dither8Lookup at: (addThreshold+((srcWord bitShift: -16) bitAnd: 255))) bitShift: 10) + 
		((dither8Lookup at: (addThreshold+((srcWord bitShift: -8) bitAnd: 255))) bitShift: 5) + 
		(dither8Lookup at: (addThreshold+(srcWord bitAnd: 255))).
! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'ar 2/19/2000 21:27'!
drawLoopX: xDelta Y: yDelta 
	"This is the primitive implementation of the line-drawing loop.
	See the comments in BitBlt>>drawLoopX:Y:"
	| dx1 dy1 px py P affL affR affT affB |
	xDelta > 0
		ifTrue: [dx1 := 1]
		ifFalse: [xDelta = 0
				ifTrue: [dx1 := 0]
				ifFalse: [dx1 := -1]].
	yDelta > 0
		ifTrue: [dy1 := 1]
		ifFalse: [yDelta = 0
				ifTrue: [dy1 := 0]
				ifFalse: [dy1 := -1]].
	px := yDelta abs.
	py := xDelta abs.
	affL := affT := 9999.  "init null rectangle"
	affR := affB := -9999.
	py > px
		ifTrue: 
			["more horizontal"
			P := py // 2.
			1 to: py do: 
				[:i |
				destX := destX + dx1.
				(P := P - px) < 0 ifTrue: 
					[destY := destY + dy1.
					P := P + py].
				i < py ifTrue:
					[self copyBits.
					interpreterProxy failed ifTrue:
						[^ nil "bail out now on failure -- avoid storing x,y"].
					(affectedL < affectedR and: [affectedT < affectedB]) ifTrue:
						["Affected rectangle grows along the line"
						affL := affL min: affectedL.
						affR := affR max: affectedR.
						affT := affT min: affectedT.
						affB := affB max: affectedB.
						(affR - affL) * (affB - affT) > 4000 ifTrue:
							["If affected rectangle gets large, update it in chunks"
							affectedL := affL.  affectedR := affR.
							affectedT := affT.  affectedB := affB.
							self showDisplayBits.
							affL := affT := 9999.  "init null rectangle"
							affR := affB := -9999]].
					]]]
		ifFalse: 
			["more vertical"
			P := px // 2.
			1 to: px do:
				[:i |
				destY := destY + dy1.
				(P := P - py) < 0 ifTrue: 
					[destX := destX + dx1.
					P := P + px].
				i < px ifTrue:
					[self copyBits.
					interpreterProxy failed ifTrue:
						[^ nil "bail out now on failure -- avoid storing x,y"].
					(affectedL < affectedR and: [affectedT < affectedB]) ifTrue:
						["Affected rectangle grows along the line"
						affL := affL min: affectedL.
						affR := affR max: affectedR.
						affT := affT min: affectedT.
						affB := affB max: affectedB.
						(affR - affL) * (affB - affT) > 4000 ifTrue:
							["If affected rectangle gets large, update it in chunks"
							affectedL := affL.  affectedR := affR.
							affectedT := affT.  affectedB := affB.
							self showDisplayBits.
							affL := affT := 9999.  "init null rectangle"
							affR := affB := -9999]].
					]]].

	"Remaining affected rect"
	affectedL := affL.  affectedR := affR.
	affectedT := affT.  affectedB := affB.

	"store destX, Y back"	
	interpreterProxy storeInteger: BBDestXIndex ofObject: bitBltOop withValue: destX.
	interpreterProxy storeInteger: BBDestYIndex ofObject: bitBltOop withValue: destY.! !

!BitBltSimulation methodsFor: 'memory access' stamp: 'eem 10/15/2018 16:41'!
dstLongAt: idx
	<inline: #always>
	self assert: idx asUnsignedInteger < endOfDestination.
	^self long32At: idx! !

!BitBltSimulation methodsFor: 'memory access' stamp: 'eem 10/15/2018 16:43'!
dstLongAt: idx put: value
	<inline: #always>
	"We omit the assert here since dstLongAt:put: is always preceded by a dstLongAt: which does have an assert.
	self assert: idx asUnsignedInteger < endOfDestination."
	^self long32At: idx put: value! !

!BitBltSimulation methodsFor: 'memory access' stamp: 'eem 10/15/2018 16:29'!
dstLongAt: idx put: srcValue mask: dstMask
	"Store the given value back into destination form, using dstMask
	 to mask out the bits to be modified. This is an essential
	 read-modify-write operation on the destination form."
	<inline: #always>
	| dstValue |
	dstValue := self dstLongAt: idx.
	dstValue := dstValue bitAnd: dstMask.
	dstValue := dstValue bitOr: srcValue.
	self dstLongAt: idx put: dstValue.! !

!BitBltSimulation methodsFor: 'setup' stamp: 'eem 3/14/2017 16:38'!
ensureDestAndSourceFormsAreValid
	"If a GC has occurred, update destForm and sourceForm from the stack."
	<inline: #always>
	numGCsOnInvocation ~= interpreterProxy statNumGCs ifTrue:
		[self reloadDestAndSourceForms]! !

!BitBltSimulation methodsFor: 'pixel mapping' stamp: 'nice 10/29/2015 22:24'!
expensiveDither32To16: srcWord threshold: ditherValue
	"Dither the given 32bit word to 16 bit. Ignore alpha."
	| pv threshold value out |
	<inline: true> "You bet"
	<returnTypeC: 'unsigned int'>
	<var: #srcWord type: 'unsigned int'>
	pv := srcWord bitAnd: 255.
	threshold := ditherThresholds16 at: (pv bitAnd: 7).
	value := ditherValues16 at: (pv bitShift: -3).
	ditherValue < threshold
		ifTrue:[out := value + 1]
		ifFalse:[out := value].
	pv := (srcWord bitShift: -8) bitAnd: 255.
	threshold := ditherThresholds16 at: (pv bitAnd: 7).
	value := ditherValues16 at: (pv bitShift: -3).
	ditherValue < threshold
		ifTrue:[out := out bitOr: (value+1 bitShift:5)]
		ifFalse:[out := out bitOr: (value bitShift: 5)].
	pv := (srcWord bitShift: -16) bitAnd: 255.
	threshold := ditherThresholds16 at: (pv bitAnd: 7).
	value := ditherValues16 at: (pv bitShift: -3).
	ditherValue < threshold
		ifTrue:[out := out bitOr: (value+1 bitShift:10)]
		ifFalse:[out := out bitOr: (value bitShift: 10)].
	^out! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:53'!
fetchIntOrFloat: fieldIndex ofObject: objectPointer
	"Return the integer value of the given field of the given object. If the field contains a Float, truncate it and return its integral part. Fail if the given field does not contain a small integer or Float, or if the truncated Float is out of the range of small integers."
	| fieldOop floatValue |
	<var: #floatValue type:'double '>
	fieldOop := interpreterProxy fetchPointer: fieldIndex ofObject: objectPointer.
	(interpreterProxy isIntegerObject: fieldOop)
		ifTrue:[^interpreterProxy integerValueOf: fieldOop].
	floatValue := interpreterProxy floatValueOf: fieldOop.
	(-2147483648.0 <= floatValue and:[floatValue <= 2147483647.0])
		ifFalse:[interpreterProxy primitiveFail. ^0].
	^floatValue asInteger! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:53'!
fetchIntOrFloat: fieldIndex ofObject: objectPointer ifNil: defaultValue
	"Return the integer value of the given field of the given object. If the field contains a Float, truncate it and return its integral part. Fail if the given field does not contain a small integer or Float, or if the truncated Float is out of the range of small integers."
	| fieldOop floatValue |
	<var: #floatValue type:'double '>
	fieldOop := interpreterProxy fetchPointer: fieldIndex ofObject: objectPointer.
	(interpreterProxy isIntegerObject: fieldOop)
		ifTrue:[^interpreterProxy integerValueOf: fieldOop].
	(fieldOop = interpreterProxy nilObject) ifTrue:[^defaultValue].
	floatValue := interpreterProxy floatValueOf: fieldOop.
	(-2147483648.0 <= floatValue and:[floatValue <= 2147483647.0])
		ifFalse:[interpreterProxy primitiveFail. ^0].
	^floatValue asInteger! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:24'!
fixAlpha: sourceWord with: destinationWord
	"For any non-zero pixel value in destinationWord with zero alpha channel take the alpha from sourceWord and fill it in. Intended for fixing alpha channels left at zero during 16->32 bpp conversions."
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	destDepth = 32 ifFalse:[^destinationWord]. "no-op for non 32bpp"
	destinationWord = 0 ifTrue:[^0].
	(destinationWord bitAnd: 16rFF000000) = 0 ifFalse:[^destinationWord].
	^destinationWord bitOr: (sourceWord bitAnd: 16rFF000000)
! !

!BitBltSimulation methodsFor: 'memory access' stamp: 'eem 10/15/2018 16:30'!
halftoneAt: idx
	"Return a value from the halftone pattern."
	<inline: #always>
	^self long32At: halftoneBase + (idx \\ halftoneHeight * 4)! !

!BitBltSimulation methodsFor: 'setup' stamp: ''!
ignoreSourceOrHalftone: formPointer

	formPointer = interpreterProxy nilObject ifTrue: [ ^true ].
	combinationRule = 0 ifTrue: [ ^true ].
	combinationRule = 5 ifTrue: [ ^true ].
	combinationRule = 10 ifTrue: [ ^true ].
	combinationRule = 15 ifTrue: [ ^true ].
	^false! !

!BitBltSimulation methodsFor: 'memory access' stamp: 'eem 10/15/2018 17:00'!
incDestIndex: offset
	<inline: #always>
	^destIndex := destIndex + offset
! !

!BitBltSimulation methodsFor: 'memory access' stamp: 'eem 10/15/2018 17:00'!
incSrcIndex: offset
	<inline: #always>
	^sourceIndex := sourceIndex + offset
! !

!BitBltSimulation methodsFor: 'initialize-release' stamp: 'nice 10/31/2020 18:57'!
initBBOpTable
	self cCode: 'opTable[0+1] = (void *)clearWordwith'.
	self cCode: 'opTable[1+1] = (void *)bitAndwith'.
	self cCode: 'opTable[2+1] = (void *)bitAndInvertwith'.
	self cCode: 'opTable[3+1] = (void *)sourceWordwith'.
	self cCode: 'opTable[4+1] = (void *)bitInvertAndwith'.
	self cCode: 'opTable[5+1] = (void *)destinationWordwith'.
	self cCode: 'opTable[6+1] = (void *)bitXorwith'.
	self cCode: 'opTable[7+1] = (void *)bitOrwith'.
	self cCode: 'opTable[8+1] = (void *)bitInvertAndInvertwith'.
	self cCode: 'opTable[9+1] = (void *)bitInvertXorwith'.
	self cCode: 'opTable[10+1] = (void *)bitInvertDestinationwith'.
	self cCode: 'opTable[11+1] = (void *)bitOrInvertwith'.
	self cCode: 'opTable[12+1] = (void *)bitInvertSourcewith'.
	self cCode: 'opTable[13+1] = (void *)bitInvertOrwith'.
	self cCode: 'opTable[14+1] = (void *)bitInvertOrInvertwith'.
	self cCode: 'opTable[15+1] = (void *)destinationWordwith'.
	self cCode: 'opTable[16+1] = (void *)destinationWordwith'.
	self cCode: 'opTable[17+1] = (void *)destinationWordwith'.
	self cCode: 'opTable[18+1] = (void *)addWordwith'.
	self cCode: 'opTable[19+1] = (void *)subWordwith'.
	self cCode: 'opTable[20+1] = (void *)rgbAddwith'.
	self cCode: 'opTable[21+1] = (void *)rgbSubwith'.
	self cCode: 'opTable[22+1] = (void *)OLDrgbDiffwith'.
	self cCode: 'opTable[23+1] = (void *)OLDtallyIntoMapwith'.
	self cCode: 'opTable[24+1] = (void *)alphaBlendwith'.
	self cCode: 'opTable[25+1] = (void *)pixPaintwith'.
	self cCode: 'opTable[26+1] = (void *)pixMaskwith'.
	self cCode: 'opTable[27+1] = (void *)rgbMaxwith'.
	self cCode: 'opTable[28+1] = (void *)rgbMinwith'.
	self cCode: 'opTable[29+1] = (void *)rgbMinInvertwith'.
	self cCode: 'opTable[30+1] = (void *)alphaBlendConstwith'.
	self cCode: 'opTable[31+1] = (void *)alphaPaintConstwith'.
	self cCode: 'opTable[32+1] = (void *)rgbDiffwith'.
	self cCode: 'opTable[33+1] = (void *)tallyIntoMapwith'.
	self cCode: 'opTable[34+1] = (void *)alphaBlendScaledwith'.
	self cCode: 'opTable[35+1] = (void *)alphaBlendScaledwith'.
	self cCode: 'opTable[36+1] = (void *)alphaBlendScaledwith'.	
	self cCode: 'opTable[37+1] = (void *)rgbMulwith'.
	self cCode: 'opTable[38+1] = (void *)pixSwapwith'.
	self cCode: 'opTable[39+1] = (void *)pixClearwith'.
	self cCode: 'opTable[40+1] = (void *)fixAlphawith'.
	self cCode: 'opTable[41+1] = (void *)rgbComponentAlphawith'.
	self cCode: 'opTable[42+1] = (void *)alphaScalewith'.
	self cCode: 'opTable[43+1] = (void *)alphaUnscalewith'.
	self cCode: 'opTable[44+1] = (void *)alphaBlendUnscaledwith'.! !

!BitBltSimulation methodsFor: 'initialize-release' stamp: 'JMM (auto pragmas 12/08) 7/4/2003 11:14'!
initDither8Lookup	
	<inline: false> 
	0 to: 255 do: [:b | 
		0 to: 15 do: [:t | | value |
			value := self expensiveDither32To16: b threshold: t.
			dither8Lookup at: ((t << 8)+b)put: value]].
	! !

!BitBltSimulation methodsFor: 'initialize-release' stamp: 'tpr 11/20/2014 15:36'!
initialiseModule
	<export: true>
	self initBBOpTable.
	self initDither8Lookup.
	self cppIf: #'ENABLE_FAST_BLT'
		ifTrue:[self initialiseCopyBits].
	^true! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 15:53'!
isIdentityMap: shifts with: masks
	"Return true if shiftTable/maskTable define an identity mapping."
	<var: #shifts type:'int *'>
	<var: #masks type:'unsigned int *'>
	(shifts == nil or:[masks == nil]) ifTrue:[^true].
	((shifts at: RedIndex) = 0 
		and:[(shifts at: GreenIndex) = 0
		and:[(shifts at: BlueIndex) = 0 
		and:[(shifts at: AlphaIndex) = 0
			and:[((masks at: RedIndex) = 16rFF0000)
			and:[((masks at: GreenIndex) = 16r00FF00)
			and:[((masks at: BlueIndex) = 16r0000FF)
			and:[((masks at: AlphaIndex) = 16rFF000000)]]]]]]])
		ifTrue:[^true].
	^false! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'eem 10/1/2018 14:27'!
loadBitBltDestForm
	"Load the dest form for BitBlt. Answer false if anything is wrong, true otherwise."

	| destBitsSize |
	<inline: true>
	((interpreterProxy isPointers: destForm)
	 and: [(interpreterProxy slotSizeOf: destForm) >= 4]) ifFalse:
		[^false].
	destBits := interpreterProxy fetchPointer: FormBitsIndex ofObject: destForm.
	destWidth := interpreterProxy fetchInteger: FormWidthIndex ofObject: destForm.
	destHeight := interpreterProxy fetchInteger: FormHeightIndex ofObject: destForm.
	(destWidth >= 0 and: [destHeight >= 0]) ifFalse:
		[^false].
	destDepth := interpreterProxy fetchInteger: FormDepthIndex ofObject: destForm.
	(destMSB := destDepth > 0) ifFalse:
		[destDepth := 0 - destDepth].
	"Ignore an integer bits handle for Display in which case 
	the appropriate values will be obtained by calling ioLockSurfaceBits()."
	(interpreterProxy isIntegerObject: destBits)
		ifTrue: "Query for actual surface dimensions"
			[querySurfaceFn = 0 ifTrue: [self loadSurfacePlugin ifFalse: [^false]].
			 (self querySurfaceFn: (interpreterProxy integerValueOf: destBits)
								_: (self addressOf: destWidth put: [:v| destWidth := v])
								_: (self addressOf: destHeight put: [:v| destHeight := v])
								_: (self addressOf: destDepth put: [:v| destDepth := v])
								_: (self addressOf: destMSB put: [:v| destMSB := v])) ifFalse:
					[interpreterProxy primitiveFailFor: PrimErrCallbackError.
					 ^false].
			destPPW := 32 // destDepth.
			destBits := destPitch := 0]
		ifFalse:
			[(interpreterProxy isWordsOrBytes: destBits) ifFalse:
				[^false].
			destPPW := 32 // destDepth.
			destPitch := destWidth + (destPPW-1) // destPPW * 4.
			destBitsSize := interpreterProxy byteSizeOf: destBits.
			destBitsSize >= (destPitch * destHeight) ifFalse:
				[^false].
			"Skip header since external bits don't have one"
			destBits := self oopForPointer: (interpreterProxy firstIndexableField: destBits)].
	^true! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'ar (auto pragmas 12/08) 2/20/2000 19:42'!
loadBitBltFrom: bbObj
	"Load BitBlt from the oop.
	This function is exported for the Balloon engine."
	<export: true>
	^self loadBitBltFrom: bbObj warping: false.! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'eem 6/6/2018 09:30'!
loadBitBltFrom: bbObj warping: aBool
	"Load context from BitBlt instance.  Return false if anything is amiss"
	"NOTE this should all be changed to minX/maxX coordinates for simpler clipping
		-- once it works!!"
	| ok |
	<inline: false>
	bitBltOop := bbObj.
	isWarping := aBool.
	bitBltIsReceiver := bbObj = (interpreterProxy stackValue: interpreterProxy methodArgumentCount).
	numGCsOnInvocation := interpreterProxy statNumGCs.
	combinationRule := interpreterProxy fetchInteger: BBRuleIndex ofObject: bitBltOop.
	(interpreterProxy failed
		or: [combinationRule < 0 or: [combinationRule > (OpTableSize - 2)]])
		 ifTrue: [^false  "operation out of range"].
	(combinationRule >= 16 and: [combinationRule <= 17])
		 ifTrue: [^false  "fail for old simulated paint, erase modes"].
	sourceForm := interpreterProxy fetchPointer: BBSourceFormIndex ofObject: bitBltOop.
	noSource := self ignoreSourceOrHalftone: sourceForm.
	halftoneForm := interpreterProxy fetchPointer: BBHalftoneFormIndex ofObject: bitBltOop.
	noHalftone := self ignoreSourceOrHalftone: halftoneForm.

	destForm := interpreterProxy fetchPointer: BBDestFormIndex ofObject: bbObj.
	ok := self loadBitBltDestForm.
	ok ifFalse:[^false].

	destX := self fetchIntOrFloat: BBDestXIndex ofObject: bitBltOop ifNil: 0.
	destY := self fetchIntOrFloat: BBDestYIndex ofObject: bitBltOop ifNil: 0.
	width := self fetchIntOrFloat: BBWidthIndex ofObject: bitBltOop ifNil: destWidth.
	height := self fetchIntOrFloat: BBHeightIndex ofObject: bitBltOop ifNil: destHeight.
	interpreterProxy failed ifTrue: [^false].

	noSource
		ifTrue:
			[sourceX := sourceY := 0]
		ifFalse: 
			[ok := self loadBitBltSourceForm.
			ok ifFalse:[^false].
			ok := self loadColorMap.
			ok ifFalse:[^false].
			"Need the implicit setup here in case of 16<->32 bit conversions"
			(cmFlags bitAnd: ColorMapNewStyle) = 0 ifTrue:[self setupColorMasks].
			sourceX := self fetchIntOrFloat: BBSourceXIndex ofObject: bitBltOop ifNil: 0.
			sourceY := self fetchIntOrFloat: BBSourceYIndex ofObject: bitBltOop ifNil: 0].

	ok := self loadHalftoneForm.
	ok ifFalse:[^false].
	clipX := self fetchIntOrFloat: BBClipXIndex ofObject: bitBltOop ifNil: 0.
	clipY := self fetchIntOrFloat: BBClipYIndex ofObject: bitBltOop ifNil: 0.
	clipWidth := self fetchIntOrFloat: BBClipWidthIndex ofObject: bitBltOop ifNil: destWidth.
	clipHeight := self fetchIntOrFloat: BBClipHeightIndex ofObject: bitBltOop ifNil: destHeight.
	interpreterProxy failed ifTrue: [^ false  "non-integer value"].
	clipX < 0 ifTrue: [clipWidth := clipWidth + clipX.  clipX := 0].
	clipY < 0 ifTrue: [clipHeight := clipHeight + clipY.  clipY := 0].
	clipX+clipWidth > destWidth ifTrue: [clipWidth := destWidth - clipX].
	clipY+clipHeight > destHeight ifTrue: [clipHeight := destHeight - clipY].
	numGCsOnInvocation ~= interpreterProxy statNumGCs ifTrue: "querySurface could be a callback in loadSourceFor: and loadDestForm:"
		[interpreterProxy primitiveFailFor: PrimErrObjectMoved.
		 ^false].
	^true! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'eem 10/1/2018 14:25'!
loadBitBltSourceForm
	"Load the source form for BitBlt. Return false if anything is wrong, true otherwise."
	| sourceBitsSize |
	<inline: true>
	((interpreterProxy isPointers: sourceForm)
	 and: [(interpreterProxy slotSizeOf: sourceForm) >= 4]) ifFalse:
		[^false].
	sourceBits := interpreterProxy fetchPointer: FormBitsIndex ofObject: sourceForm.
	sourceWidth := self fetchIntOrFloat: FormWidthIndex ofObject: sourceForm.
	sourceHeight := self fetchIntOrFloat: FormHeightIndex ofObject: sourceForm.
	(sourceWidth >= 0 and: [sourceHeight >= 0]) ifFalse:
		[^false].
	sourceDepth := interpreterProxy fetchInteger: FormDepthIndex ofObject: sourceForm.
	(sourceMSB := sourceDepth > 0) ifFalse:
		[sourceDepth := 0 - sourceDepth].
	"Ignore an integer bits handle for Display in which case
	 the appropriate values will be obtained by calling ioLockSurfaceBits()."
	(interpreterProxy isIntegerObject: sourceBits)
		ifTrue: "Query for actual surface dimensions"
			[querySurfaceFn = 0 ifTrue: [self loadSurfacePlugin ifFalse:[^false]].
			 (self querySurfaceFn: (interpreterProxy integerValueOf: sourceBits)
								_: (self addressOf: sourceWidth put: [:v| sourceWidth := v])
								_: (self addressOf: sourceHeight put: [:v| sourceHeight := v])
								_: (self addressOf: sourceDepth put: [:v| sourceDepth := v])
								_: (self addressOf: sourceMSB put: [:v| sourceMSB := v])) ifFalse:
					[interpreterProxy primitiveFailFor: PrimErrCallbackError.
					 ^false].
			sourcePPW := 32 // sourceDepth.
			sourceBits := sourcePitch := 0]
		ifFalse:
			[(interpreterProxy isWordsOrBytes: sourceBits) ifFalse:
				[^false].
			sourcePPW := 32 // sourceDepth.
			sourcePitch := sourceWidth + (sourcePPW-1) // sourcePPW * 4.
			sourceBitsSize := interpreterProxy byteSizeOf: sourceBits.
			sourceBitsSize >= (sourcePitch * sourceHeight) ifFalse:
				[^false].
			"Skip header since external bits don't have one"
			sourceBits := self oopForPointer: (interpreterProxy firstIndexableField: sourceBits)].
	^true! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'eem 4/27/2020 11:03'!
loadColorMap
	"ColorMap, if not nil, must be longWords, and 
	2^N long, where N = sourceDepth for 1, 2, 4, 8 bits, 
	or N = 9, 12, or 15 (3, 4, 5 bits per color) for 16 or 32 bits."
	| cmSize oldStyle oop cmOop |
	<inline: true>
	cmFlags := cmMask := cmBitsPerColor := 0.
	cmShiftTable := nil.
	cmMaskTable := nil.
	cmLookupTable := nil.
	cmOop := interpreterProxy fetchPointer: BBColorMapIndex ofObject: bitBltOop.
	cmOop = interpreterProxy nilObject ifTrue:[^true].
	cmFlags := ColorMapPresent. "even if identity or somesuch - may be cleared later"
	(interpreterProxy isWords: cmOop)
		ifTrue: "This is an old-style color map (indexed only, with implicit RGBA conversion)"
			[oldStyle := true.
			cmSize := interpreterProxy slotSizeOf: cmOop.
			cmLookupTable := interpreterProxy firstIndexableField: cmOop]
		ifFalse: "A new-style color map (fully qualified)"
			[oldStyle := false.
			((interpreterProxy isPointers: cmOop) 
			 and:[(interpreterProxy slotSizeOf: cmOop) >= 3]) ifFalse:[^false].
			cmShiftTable := self loadColorMapShiftOrMaskFrom:
				(interpreterProxy fetchPointer: 0 ofObject: cmOop).
			cmMaskTable := self loadColorMapShiftOrMaskFrom:
				(interpreterProxy fetchPointer: 1 ofObject: cmOop).
			oop := interpreterProxy fetchPointer: 2 ofObject: cmOop.
			oop = interpreterProxy nilObject 
				ifTrue:[cmSize := 0]
				ifFalse:[(interpreterProxy isWords: oop) ifFalse:[^false].
						cmSize := (interpreterProxy slotSizeOf: oop).
						cmLookupTable := interpreterProxy firstIndexableField: oop].
			cmFlags := cmFlags bitOr: ColorMapNewStyle].
	(cmSize bitAnd: cmSize - 1) = 0 ifFalse:[^false].
	cmMask := cmSize - 1.
	cmBitsPerColor := 0.
	cmSize = 0
		ifTrue:[cmLookupTable := nil. cmMask := 0]
		ifFalse:
			[cmFlags := cmFlags bitOr: ColorMapIndexedPart.
			 cmSize = 512 ifTrue: [cmBitsPerColor := 3].
			 cmSize = 4096 ifTrue: [cmBitsPerColor := 4].
			 cmSize = 32768 ifTrue: [cmBitsPerColor := 5]].
	oldStyle ifTrue: "needs implicit conversion"
		[self setupColorMasks].
	"Check if colorMap is just identity mapping for RGBA parts"
	(self isIdentityMap: cmShiftTable with: cmMaskTable)
		ifTrue: [cmMaskTable := nil. cmShiftTable := nil]
		ifFalse: [cmFlags := cmFlags bitOr: ColorMapFixedPart].
	^true! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'eem 9/21/2013 07:24'!
loadColorMapShiftOrMaskFrom: mapOop
	<returnTypeC:'void *'>
	mapOop = interpreterProxy nilObject ifTrue:[^nil].
	((interpreterProxy isWords: mapOop) 
		and:[(interpreterProxy slotSizeOf: mapOop) = 4])
			ifFalse:[interpreterProxy primitiveFail. ^nil].
	^interpreterProxy firstIndexableField: mapOop! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'eem 6/6/2018 09:20'!
loadHalftoneForm
	"Load the halftone form"
	| halftoneBits |
	<inline: true>
	noHalftone ifTrue:
		[halftoneBase := nil.
		^true].
	((interpreterProxy isPointers: halftoneForm)
	 and: [(interpreterProxy slotSizeOf: halftoneForm) >= 4])
		ifTrue:
			["Old-style 32xN monochrome halftone Forms"
			halftoneBits := interpreterProxy fetchPointer: FormBitsIndex ofObject: halftoneForm.
			halftoneHeight := interpreterProxy fetchInteger: FormHeightIndex ofObject: halftoneForm.
			(interpreterProxy isWords: halftoneBits) ifFalse:
				[noHalftone := true]]
		ifFalse:
			["New spec accepts, basically, a word array"
			(interpreterProxy isWords: halftoneForm) ifFalse:
				[^false].
			halftoneBits := halftoneForm.
			halftoneHeight := interpreterProxy slotSizeOf: halftoneBits].
	halftoneBase := self oopForPointer: (interpreterProxy firstIndexableField: halftoneBits).
	^true! !

!BitBltSimulation methodsFor: 'surface support' stamp: 'ar 4/18/2001 21:23'!
loadSurfacePlugin
	"Load the surface support plugin"
	querySurfaceFn := interpreterProxy ioLoadFunction:'ioGetSurfaceFormat' From:'SurfacePlugin'.
	lockSurfaceFn := interpreterProxy ioLoadFunction:'ioLockSurface' From:'SurfacePlugin'.
	unlockSurfaceFn := interpreterProxy ioLoadFunction:'ioUnlockSurface' From:'SurfacePlugin'.
	^querySurfaceFn ~= 0 and:[lockSurfaceFn ~= 0 and:[unlockSurfaceFn ~= 0]]! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'ar 10/27/1999 16:03'!
loadWarpBltFrom: bbObj
	^self loadBitBltFrom: bbObj warping: true! !

!BitBltSimulation methodsFor: 'surface support' stamp: 'eem 10/15/2018 15:54'!
lockSurfaceFn: sourceHandle _: pitchPtr _: x _: y _: w _: h
	"Simulate the lockSurfaceFn function call as a failure to load the surface."
	<doNotGenerate>
	^0! !

!BitBltSimulation methodsFor: 'surface support' stamp: 'eem 10/16/2018 10:04'!
lockSurfaces
	"Get a pointer to the bits of any OS surfaces."
	"Notes: 
	* For equal source/dest handles only one locking operation is performed.
	This is to prevent locking of overlapping areas which does not work with
	certain APIs (as an example, DirectDraw prevents locking of overlapping areas). 
	A special case for non-overlapping but equal source/dest handle would 
	be possible but we would have to transfer this information over to 
	unlockSurfaces somehow (currently, only one unlock operation is 
	performed for equal source and dest handles). Also, this would require
	a change in the notion of ioLockSurface() which is right now interpreted
	as a hint and not as a requirement to lock only the specific portion of
	the surface.

	* The arguments in ioLockSurface() provide the implementation with
	an explicit hint what area is affected. It can be very useful to
	know the max. affected area beforehand if getting the bits requires expensive
	copy operations (e.g., like a roundtrip to the X server or a glReadPixel op).
	However, the returned pointer *MUST* point to the virtual origin of the surface
	and not to the beginning of the rectangle. The promise made by BitBlt
	is to never access data outside the given rectangle (aligned to 4byte boundaries!!)
	so it is okay to return a pointer to the virtual origin that is actually outside
	the valid memory area.

	* The area provided in ioLockSurface() is already clipped (e.g., it will always
	be inside the source and dest boundingBox) but it is not aligned to word boundaries
	yet. It is up to the support code to compute accurate alignment if necessary.

	* Warping always requires the entire source surface to be locked because
	there is no beforehand knowledge about what area will actually be traversed.

	* Fail if a GC has occurred since the primitive started (presumably in the lockSurface
	   function), because one or more of the primitives' parameters may have been moved.
	"
	| sourceHandle destHandle l r t b |
	<inline: true>
	self assert: numGCsOnInvocation = interpreterProxy statNumGCs.
	hasSurfaceLock := false.
	destBits = 0 ifTrue: "Blitting *to* OS surface"
		[lockSurfaceFn = 0 ifTrue: [self loadSurfacePlugin ifFalse: [^false]].
		 destHandle := interpreterProxy fetchInteger: FormBitsIndex ofObject: destForm.
		 (sourceBits ~= 0 or: [noSource]) ifFalse:
			[sourceHandle := interpreterProxy fetchInteger: FormBitsIndex ofObject: sourceForm.
			 "Handle the special case of equal source and dest handles"
			 sourceHandle = destHandle ifTrue:
				"If we have overlapping source/dest we lock the entire area
				so that there is only one area transmitted"
				[isWarping
					ifFalse: "When warping we always need the entire surface for the source"
						[sourceBits := self lockSurfaceFn: sourceHandle
														_: (self addressOf: sourcePitch put: [:v| sourcePitch := v])
														_: 0
														_: 0
														_: sourceWidth
														_: sourceHeight]
					ifTrue: "Otherwise use overlapping area"
						[l := sx min: dx. r := (sx max: dx) + bbW.
						 t := sy min: dy. b := (sy max: dy) + bbH.
						 sourceBits := self lockSurfaceFn: sourceHandle
														_: (self addressOf: sourcePitch put: [:v| sourcePitch := v])
														_: l
														_: t
														_: r - l
														_: b - t].
				destBits := sourceBits.
				destPitch := sourcePitch.
				hasSurfaceLock := true.
				numGCsOnInvocation ~= interpreterProxy statNumGCs ifTrue:
					[self unlockSurfaces.
					 interpreterProxy primitiveFailFor: PrimErrObjectMoved.
					 ^false].
				destBits = 0 ifTrue:
					[self unlockSurfaces.
					 interpreterProxy primitiveFailFor: PrimErrCallbackError.
					 ^false].
				endOfDestination := endOfSource := sourceBits + (sourcePitch * sourceHeight).
				^true]].
		"Fall through - if not equal it'll be handled below"
		destBits := self lockSurfaceFn: destHandle
									_: (self addressOf: destPitch put: [:v| destPitch := v])
									_: dx
									_: dy
									_: bbW
									_: bbH.
		hasSurfaceLock := true.
		numGCsOnInvocation ~= interpreterProxy statNumGCs ifTrue:
			[self unlockSurfaces.
			 interpreterProxy primitiveFailFor: PrimErrObjectMoved.
			 ^false].
		destBits = 0 ifTrue:
			[interpreterProxy primitiveFailFor: PrimErrCallbackError]].

	(sourceBits ~= 0 or: [noSource]) ifFalse: "Blitting *from* OS surface"
		[sourceHandle := interpreterProxy fetchInteger: FormBitsIndex ofObject: sourceForm.
		 interpreterProxy failed ifTrue: [^false]. "fetch sourceHandle could fail"
		lockSurfaceFn = 0 ifTrue:[self loadSurfacePlugin ifFalse:[^false]].
		"Warping requiring the entire surface"
		isWarping
			ifTrue:
				[sourceBits := self lockSurfaceFn: sourceHandle
												_: (self addressOf: sourcePitch put: [:v| sourcePitch := v])
												_: 0
												_: 0
												_: sourceWidth
												_: sourceHeight]
			ifFalse:
				[sourceBits := self lockSurfaceFn: sourceHandle
												_: (self addressOf: sourcePitch put: [:v| sourcePitch := v])
												_: sx
												_: sy
												_: bbW
												_: bbH].
		hasSurfaceLock := true.
		numGCsOnInvocation ~= interpreterProxy statNumGCs ifTrue:
			[self unlockSurfaces.
			 interpreterProxy primitiveFailFor: PrimErrObjectMoved.
			 ^false].
		sourceBits = 0 ifTrue:
			[interpreterProxy primitiveFailFor: PrimErrCallbackError]].
	endOfSource := (noSource or: [sourceBits = 0]) ifTrue: [0] ifFalse: [sourceBits + (sourcePitch * sourceHeight)].
	endOfDestination := destBits + (destPitch * destHeight).
	^destBits ~= 0 and: [sourceBits ~= 0 or: [noSource]]! !

!BitBltSimulation methodsFor: 'color mapping' stamp: 'ar (auto pragmas 12/08) 4/26/2001 21:57'!
mapPixel: sourcePixel flags: mapperFlags
	"Color map the given source pixel."
	| pv |
	<inline: true>
	pv := sourcePixel.
	(mapperFlags bitAnd: ColorMapPresent) ~= 0 ifTrue:[
		(mapperFlags bitAnd: ColorMapFixedPart) ~= 0 ifTrue:[
			pv := self rgbMapPixel: sourcePixel flags: mapperFlags.
			"avoid introducing transparency by color reduction"
			(pv = 0 and:[sourcePixel ~= 0]) ifTrue:[pv := 1]].
		(mapperFlags bitAnd: ColorMapIndexedPart) ~= 0
			ifTrue:[pv := cmLookupTable at: (pv bitAnd: cmMask)].
	].
	^pv! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:24'!
merge: sourceWord with: destinationWord
	| mergeFnwith |
	"Sender warpLoop is too big to include this in-line"
	<var: #mergeFnwith declareC: 'unsigned int (*mergeFnwith)(unsigned int, unsigned int)'>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	mergeFnwith := self cCoerce: (opTable at: combinationRule+1) to: 'unsigned int (*)(unsigned int, unsigned int)'.
	mergeFnwith.  "null ref for compiler"

	^ self mergeFn: sourceWord with: destinationWord! !

!BitBltSimulation methodsFor: 'initialize-release' stamp: 'eem 3/15/2017 14:50'!
moduleUnloaded: aModuleName
	"The module with the given name was just unloaded.
	 Make sure we have no dangling references."
	<export: true>
	<var: #aModuleName type: 'char *'>
	(aModuleName strcmp: 'SurfacePlugin') = 0 ifTrue: "The surface plugin just shut down. How nasty."
		[querySurfaceFn := 0.
		 lockSurfaceFn := 0.
		 unlockSurfaceFn := 0]! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 1/13/2020 22:43'!
partitionedAND: word1 to: word2 nBits: nBits nPartitions: nParts
	"AND word1 to word2 as nParts partitions of nBits each.
	Any field of word1 not all-ones is treated as all-zeroes.
	Used for erasing, eg, brush shapes prior to ORing in a color"
	| mask result |
	<returnTypeC: 'unsigned int'>
	<var: #word1 type: 'unsigned int'>
	<var: #word2 type: 'unsigned int'>
	<var: #result type: 'unsigned int'>
	<var: #mask type: 'unsigned int'>
	mask := maskTable at: nBits.  "partition mask starts at the right"
	result := 0.
	nBits = 32
		ifTrue:
			[word1 = mask ifTrue: [result := result bitOr: word2]]
		ifFalse:
			[1 to: nParts do:
				[:i |
				(word1 bitAnd: mask) = mask
					ifTrue: [result := result bitOr: (word2 bitAnd: mask)].
				mask := mask << nBits  "slide left to next partition"]].
	^ result! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:24'!
partitionedAdd: word1 to: word2 nBits: nBits componentMask: componentMask carryOverflowMask: carryOverflowMask
	"Add word1 to word2 as nParts partitions of nBits each.
	This is useful for packed pixels, or packed colors"
	| carryOverflow sum w1 w2 |
	"Use unsigned int everywhere because it has a well known arithmetic model without undefined behavior w.r.t. overflow and shifts"
	<returnTypeC: 'unsigned int'>
	 <var: #word1 type: 'unsigned int'>
	<var: #word2 type: 'unsigned int'>
	 <var: #w1 type: 'unsigned int'>
	<var: #w2 type: 'unsigned int'>
	<var: #componentMask type: 'unsigned int'>
	<var: #carryOverflowMask type: 'unsigned int'>
	<var: #carryOverflow type: 'unsigned int'>
	<var: #sum type: 'unsigned int'>
	w1 := word1 bitAnd: carryOverflowMask. "mask to remove high bit of each component"
	w2 := word2 bitAnd: carryOverflowMask.
	sum := (word1 bitXor: w1)+(word2 bitXor: w2). "sum without high bit to avoid overflowing over next component"
	carryOverflow := (w1 bitAnd: w2) bitOr: ((w1 bitOr: w2) bitAnd: sum). "detect overflow condition for saturating"
	^((sum bitXor: w1)bitXor:w2) "sum high bit without overflow"
		bitOr: carryOverflow>>(nBits-1) * componentMask "saturate in case of overflow"! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 1/13/2020 22:02'!
partitionedMax: word1 with: word2 nBits: nBits nPartitions: nParts
	"Max word1 to word2 as nParts partitions of nBits each"
	| mask result |
	"In C, most arithmetic operations answer the same bit pattern regardless of the operands being signed or unsigned ints
	(this is due to the way 2's complement numbers work). However, comparisions might fail. Add the proper declaration of
	words as unsigned int in those cases where comparisions are done (jmv)"
	<returnTypeC: 'unsigned int'>
	<var: #word1 type: 'unsigned int'>
	<var: #word2 type: 'unsigned int'>
	<var: #mask type: 'unsigned int'>
	<var: #result type: 'unsigned int'>
	nBits = 32
		ifTrue:
			[result := word2 max: word1]
		ifFalse:
			[result := 0.
			mask := maskTable at: nBits.  "partition mask starts at the right"
			1 to: nParts do:
				[:i |
				result := result bitOr: ((word2 bitAnd: mask) max: (word1 bitAnd: mask)).
				mask := mask << nBits  "slide left to next partition"]].
	^ result
! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 1/13/2020 22:03'!
partitionedMin: word1 with: word2 nBits: nBits nPartitions: nParts
	"Min word1 to word2 as nParts partitions of nBits each"
	| mask result |
	"In C, most arithmetic operations answer the same bit pattern regardless of the operands being signed or unsigned ints
	(this is due to the way 2's complement numbers work). However, comparisions might fail. Add the proper declaration of
	words as unsigned int in those cases where comparisions are done (jmv)"
	<returnTypeC: 'unsigned int'>
	<var: #word1 type: 'unsigned int'>
	<var: #word2 type: 'unsigned int'>
	<var: #mask type: 'unsigned int'>
	<var: #result type: 'unsigned int'>
	nBits = 32
		ifTrue:
			[result := word2 min: word1]
		ifFalse:
			[result := 0.
			mask := maskTable at: nBits.  "partition mask starts at the right"
			1 to: nParts do:
				[:i |
				result := result bitOr: ((word2 bitAnd: mask) min: (word1 bitAnd: mask)).
				mask := mask << nBits  "slide left to next partition"]].
	^ result
! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:24'!
partitionedMul: word1 with: word2 nBits: nBits nPartitions: nParts
	"Multiply word1 with word2 as nParts partitions of nBits each.
	This is useful for packed pixels, or packed colors.
	Bug in loop version when non-white background"

	| sMask product result dMask |
	"In C, integer multiplication might answer a wrong value if the unsigned values are declared as signed.
	This problem does not affect this method, because the most significant bit (i.e. the sign bit) will
	always be zero (jmv)"
	<returnTypeC: 'unsigned int'>
	<var: #word1 type: 'unsigned int'>
	<var: #word2 type: 'unsigned int'>
	<var: #sMask type: 'unsigned int'>
	<var: #dMask type: 'unsigned int'>
	<var: #result type: 'unsigned int'>
	<var: #product type: 'unsigned int'>
	sMask := maskTable at: nBits.  "partition mask starts at the right"
	dMask :=  sMask << nBits.
	result := (((word1 bitAnd: sMask)+1) * ((word2 bitAnd: sMask)+1) - 1 
				bitAnd: dMask) >> nBits.	"optimized first step"
	nParts = 1
		ifTrue: [ ^result ].
	product := (((word1>>nBits bitAnd: sMask)+1) * ((word2>>nBits bitAnd: sMask)+1) - 1 bitAnd: dMask).
	result := result bitOr: product.
	nParts = 2
		ifTrue: [ ^result ].
	product := (((word1>>(2*nBits) bitAnd: sMask)+1) * ((word2>>(2*nBits) bitAnd: sMask)+1) - 1 bitAnd: dMask).
	result := result bitOr: product << nBits.
	nParts = 3
		ifTrue: [ ^result ].
	product := (((word1>>(3*nBits) bitAnd: sMask)+1) * ((word2>>(3*nBits) bitAnd: sMask)+1) - 1 bitAnd: dMask).
	result := result bitOr: product << (2*nBits).
	^ result

"	| sMask product result dMask |
	sMask := maskTable at: nBits.  'partition mask starts at the right'
	dMask :=  sMask << nBits.
	result := (((word1 bitAnd: sMask)+1) * ((word2 bitAnd: sMask)+1) - 1 
				bitAnd: dMask) >> nBits.	'optimized first step'
	nBits to: nBits * (nParts-1) by: nBits do: [:ofs |
		product := (((word1>>ofs bitAnd: sMask)+1) * ((word2>>ofs bitAnd: sMask)+1) - 1 bitAnd: dMask).
		result := result bitOr: (product bitAnd: dMask) << (ofs-nBits)].
	^ result"! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:24'!
partitionedRgbComponentAlpha: sourceWord dest: destWord nBits: nBits nPartitions: nParts
	| mask result p1 p2 v |
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destWord type: 'unsigned int'>
	<var: #p1 type: 'unsigned int'>
	<var: #p2 type: 'unsigned int'>
	<var: #mask type: 'unsigned int'>
	<var: #result type: 'unsigned int'>
	mask := maskTable at: nBits.  "partition mask starts at the right"
	result := 0.
	1 to: nParts do:
		[:i |
		p1 := (sourceWord bitAnd: mask) >> ((i - 1)*nBits).
		p2 := (destWord bitAnd: mask) >> ((i - 1)*nBits).
		nBits = 32
			ifFalse:[
				nBits = 16
					ifTrue:[
						p1 := (self rgbMap16To32: p1) bitOr: 16rFF000000.
						p2 := (self rgbMap16To32: p2) bitOr: 16rFF000000]
					ifFalse:[
						p1 := (self rgbMap: p1 from: nBits to: 32) bitOr: 16rFF000000.
						p2 := (self rgbMap: p2 from: nBits to: 32) bitOr: 16rFF000000.]].
		v := self rgbComponentAlpha32: p1 with: p2.
		nBits = 32
			ifFalse:[
				v := self rgbMap: v from: 32 to: nBits].
		result := result bitOr: (v <<  ((i - 1)*nBits)). 
		mask := mask << nBits  "slide left to next partition"].
	^ result
! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 1/13/2020 22:02'!
partitionedSub: word1 from: word2 nBits: nBits nPartitions: nParts
	"Subtract word1 from word2 as nParts partitions of nBits each.
	This is useful for packed pixels, or packed colors"
	| mask result p1 p2 |
	"In C, most arithmetic operations answer the same bit pattern regardless of the operands being signed or unsigned ints
	(this is due to the way 2's complement numbers work). However, comparisions might fail. Add the proper declaration of
	words as unsigned int in those cases where comparisions are done (jmv)"
	<returnTypeC: 'unsigned int'>
	<var: #word1 type: 'unsigned int'>
	<var: #word2 type: 'unsigned int'>
	<var: #p1 type: 'unsigned int'>
	<var: #p2 type: 'unsigned int'>
	<var: #mask type: 'unsigned int'>
	<var: #result type: 'unsigned int'>
	nBits = 32
		ifTrue:
			[result := word1 < word2
				ifTrue: [word2 - word1]
				ifFalse: [word1 - word2]]
		ifFalse:
			[result := 0.
			mask := maskTable at: nBits.  "partition mask starts at the right"
			1 to: nParts do:
				[:i |
				p1 := word1 bitAnd: mask.
				p2 := word2 bitAnd: mask.
				p1 < p2  "result is really abs value of thedifference"
					ifTrue: [result := result bitOr: p2 - p1]
					ifFalse: [result := result bitOr: p1 - p2].
				mask := mask << nBits  "slide left to next partition"]].
	^ result
! !

!BitBltSimulation methodsFor: 'setup' stamp: 'eem 10/16/2018 10:16'!
performCopyLoop
	"Based on the values provided during setup choose and
	perform the appropriate inner loop function."
	<inline: true> "Should be inlined into caller for speed"
	self destMaskAndPointerInit.
	noSource
		ifTrue: "Simple fill loop"
			[self copyLoopNoSource]
		ifFalse: "Loop using source and dest"
			[self checkSourceOverlap.
			 (sourceDepth ~= destDepth or: [cmFlags ~= 0 or: [sourceMSB ~= destMSB]])
				ifTrue: "If we must convert between pixel depths or use
						color lookups or swap pixels use the general version"
					[self copyLoopPixMap]
				ifFalse: "Otherwise we simply copy pixels and can use a faster version"
					[self sourceSkewAndPointerInit.
					 self copyLoop]]! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'eem 10/16/2018 15:35'!
pickSourcePixels: nPixels flags: mapperFlags srcMask: srcMask destMask: dstMask srcShiftInc: srcShiftInc dstShiftInc: dstShiftInc
	"Pick nPix pixels starting at srcBitIndex from the source, map by the
	color map, and justify them according to dstBitIndex in the resulting destWord."
	| sourceWord destWord sourcePix destPix srcShift dstShift nPix |
	<inline: true> "oh please"
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: #'unsigned int'>
	<var: #destWord type: #'unsigned int'>
	destWord := 0.
	srcShift := srcBitShift. "Hint: Keep in register"
	dstShift := dstBitShift. "Hint: Keep in register"
	nPix := nPixels. "always > 0 so we can use do { } while(--nPix);"
	(mapperFlags = (ColorMapPresent bitOr: ColorMapIndexedPart)) ifTrue:[
		"a little optimization for (pretty crucial) blits using indexed lookups only"
		[	"grab, colormap and mix in pixel"
			sourceWord := self srcLongAt: sourceIndex.
			sourcePix := sourceWord >> srcShift bitAnd: srcMask.
			destPix := cmLookupTable at: (sourcePix bitAnd: cmMask).
			destWord := destWord bitOr: (destPix bitAnd: dstMask) << dstShift.
			"adjust dest pix index"
			dstShift := dstShift + dstShiftInc.
			"adjust source pix index"
			((srcShift := srcShift + srcShiftInc) bitAnd: 16rFFFFFFE0) = 0 ifFalse:
				[srcShift := sourceMSB ifTrue: [srcShift + 32] ifFalse: [srcShift - 32].
				 self incSrcIndex: 4].
		(nPix := nPix - 1) = 0] whileFalse.
	] ifFalse:[
		[	"grab, colormap and mix in pixel"
			sourceWord := self srcLongAt: sourceIndex.
			sourcePix := sourceWord >> srcShift bitAnd: srcMask.
			destPix := self mapPixel: sourcePix flags: mapperFlags.
			destWord := destWord bitOr: (destPix bitAnd: dstMask) << dstShift.
			"adjust dest pix index"
			dstShift := dstShift + dstShiftInc.
			"adjust source pix index"
			((srcShift := srcShift + srcShiftInc) bitAnd: 16rFFFFFFE0) = 0 ifFalse:
				[srcShift := sourceMSB ifTrue: [srcShift + 32] ifFalse: [srcShift - 32].
				 self incSrcIndex: 4].
		(nPix := nPix - 1) = 0] whileFalse.
	].
	srcBitShift := srcShift. "Store back"
	^destWord
! !

!BitBltSimulation methodsFor: 'pixel mapping' stamp: 'nice 10/29/2015 22:24'!
pickWarpPixelAtX: xx y: yy
	"Pick a single pixel from the source for WarpBlt.
	Note: This method is crucial for WarpBlt speed w/o smoothing
	and still relatively important when smoothing is used."
	| x y srcIndex sourceWord sourcePix |
	<inline: true> "*please*"
	<returnTypeC: #'unsigned int'>
	<var: #sourceWord type: #'unsigned int'>

	"note: it would be much faster if we could just
	avoid these stupid tests for being inside sourceForm."
	(xx < 0 or:[yy < 0 or:[
		(x := xx >> BinaryPoint) >= sourceWidth or:[
			(y := yy >> BinaryPoint) >= sourceHeight]]]) ifTrue:[^0]. "out of bounds"

	"Fetch source word.
	Note: We should really update srcIndex with sx and sy so that
	we don't have to do the computation below. We might even be
	able to simplify the out of bounds test from above."
	srcIndex := sourceBits + (y * sourcePitch) + (x >> warpAlignShift * 4).
	sourceWord := self srcLongAt: srcIndex.

	"Extract pixel from word"
	srcBitShift := warpBitShiftTable at: (x bitAnd: warpAlignMask).
	sourcePix := sourceWord >> srcBitShift bitAnd: warpSrcMask.
	^sourcePix! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:24'!
pixClear: sourceWord with: destinationWord
	"Clear all pixels in destinationWord for which the pixels of sourceWord have the same values. Used to clear areas of some constant color to zero."
	| mask result nBits pv |
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	<var: #mask type: 'unsigned int'>
	<var: #pv type: 'unsigned int'>
	<var: #result type: 'unsigned int'>
	destDepth = 32 ifTrue:[
		sourceWord = destinationWord ifTrue:[^0] ifFalse:[^destinationWord].
	].
	nBits := destDepth.
	mask := maskTable at: nBits.  "partition mask starts at the right"
	result := 0.
	1 to: destPPW do:[:i |
		pv := destinationWord bitAnd: mask.
		(sourceWord bitAnd: mask) = pv ifTrue:[pv := 0].
		result := result bitOr: pv.
		mask := mask << nBits "slide left to next partition"].
	^ result! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:24'!
pixMask: sourceWord with: destinationWord
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^ self partitionedAND: sourceWord bitInvert32 to: destinationWord
					nBits: destDepth nPartitions: destPPW! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:24'!
pixPaint: sourceWord with: destinationWord
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	sourceWord = 0 ifTrue: [^ destinationWord].
	^ sourceWord bitOr:
		(self partitionedAND: sourceWord bitInvert32 to: destinationWord
						nBits: destDepth nPartitions: destPPW)! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:24'!
pixSwap: sourceWord with: destWord
	"Swap the pixels in destWord"
	| result shift lowMask highMask |
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destWord type: 'unsigned int'>
	<var: #lowMask type: 'unsigned int'>
	<var: #highMask type: 'unsigned int'>
	<var: #result type: 'unsigned int'>
	destPPW = 1 ifTrue:[^destWord]. "a single pixel per word"
	result := 0.
	lowMask := (1 << destDepth) - 1. "mask low pixel"
	highMask := lowMask << (destPPW-1 * destDepth). "mask high pixel"
	shift := 32 - destDepth.
	result := result bitOr: (
				(destWord bitAnd: lowMask) << shift bitOr:
					(destWord bitAnd: highMask) >> shift).
	destPPW <= 2 ifTrue:[^result].
	2 to: destPPW // 2 do:[:i|
		lowMask := lowMask << destDepth.
		highMask := highMask >> destDepth.
		shift := shift - (destDepth * 2).
		result := result bitOr: (
					(destWord bitAnd: lowMask) << shift bitOr:
						(destWord bitAnd: highMask) >> shift)].
	^result! !

!BitBltSimulation methodsFor: 'primitives' stamp: 'tpr 12/19/2014 11:01'!
primitiveCompareColorA: colorA to: colorB test: testID 
	"Invoke the pixel color comparing primitive.Only applicable if compiling
	with ENABLE_FAST_BLT"
	| rcvr val |
	<export: true>
	rcvr := self
				primitive: 'primitiveCompareColors'
				parameters: #(#Unsigned #Unsigned #SmallInteger )
				receiver: #Oop.
	self
		cppIf: #'ENABLE_FAST_BLT'
		ifTrue: [(self loadBitBltFrom: rcvr)
					ifFalse: [^ interpreterProxy primitiveFail].
			self clipRange.
			(bbW <= 0 or: [bbH <= 0])
				ifTrue: ["zero width or height; noop"
					^ interpreterProxy primitiveFail].
			self cCode: '
	compare_operation_t op;
	op.matchRule = testID & 3;
	op.tally = testID & (1u<<3);
	op.srcA.bits = (void *) sourceBits;
	op.srcA.pitch = sourcePitch;
	op.srcA.depth = sourceDepth;
	op.srcA.msb = sourceMSB;
	op.srcA.x = sx;
	op.srcA.y = sy;
	op.srcB.bits = (void *) destBits;
	op.srcB.pitch = destPitch;
	op.srcB.depth = destDepth;
	op.srcB.msb = destMSB;
	op.srcB.x = dx;
	op.srcB.y = dy;
	op.width = bbW;
	op.height = bbH;
	op.colorA = colorA;
	op.colorB = colorB;

	val = compareColorsDispatch(&op);'.
	^val asPositiveIntegerObj]
		ifFalse: [interpreterProxy primitiveFail]! !

!BitBltSimulation methodsFor: 'primitives' stamp: 'eem 9/28/2018 14:49'!
primitiveCopyBits
	"Invoke the copyBits primitive. If the destination is the display, then copy it to the screen."
	| rcvr |
	<export: true>
	rcvr := interpreterProxy stackValue: interpreterProxy methodArgumentCount.
	(self loadBitBltFrom: rcvr) ifFalse:
		[^interpreterProxy primitiveFail].
	self copyBits.
	interpreterProxy failed ifTrue: [^nil].
	self showDisplayBits.
	interpreterProxy failed ifTrue: [^nil].
	(combinationRule = 22 or: [combinationRule = 32])
		ifTrue: [interpreterProxy methodReturnInteger: bitCount]
		ifFalse: [interpreterProxy methodReturnReceiver]! !

!BitBltSimulation methodsFor: 'primitives' stamp: 'eem 10/18/2018 14:07'!
primitiveDisplayString
	<export: true>
	| kernDelta xTable glyphMap stopIndex startIndex sourceString bbObj maxGlyph ascii glyphIndex sourcePtr left quickBlt |
	<var: #sourcePtr type: #'char *'>
	interpreterProxy methodArgumentCount = 6 ifFalse:
		[^interpreterProxy primitiveFail].
	kernDelta := interpreterProxy stackIntegerValue: 0.
	xTable := interpreterProxy stackValue: 1.
	glyphMap := interpreterProxy stackValue: 2.
	stopIndex := interpreterProxy stackIntegerValue: 3.
	startIndex := interpreterProxy stackIntegerValue: 4.
	sourceString := interpreterProxy stackValue: 5.
	bbObj := interpreterProxy stackObjectValue: 6.
	interpreterProxy failed ifTrue:
		[^nil].

	((interpreterProxy isArray: xTable)
	 and: [(interpreterProxy isArray: glyphMap)
	 and: [(interpreterProxy slotSizeOf: glyphMap) = 256
	 and: [(interpreterProxy isBytes: sourceString)
	 and: [startIndex > 0
	 and: [stopIndex >= 0 "to avoid failing for empty strings..."
	 and: [stopIndex <= (interpreterProxy byteSizeOf: sourceString)
	 and: [(self loadBitBltFrom: bbObj)
	 and: [combinationRule ~= 30 "these two need extra source alpha"
	 and: [combinationRule ~= 31]]]]]]]]]) ifFalse:
		[^interpreterProxy primitiveFail].
	stopIndex = 0 ifTrue:
		[^interpreterProxy pop: 6 "the string is empty; pop args, return rcvr"].
	maxGlyph := (interpreterProxy slotSizeOf: xTable) - 2.
	"See if we can go directly into copyLoopPixMap (usually we can)"
	quickBlt := destBits ~= 0 "no OS surfaces please"
				and:[sourceBits ~= 0 "and again"
				and:[noSource = false "needs a source"
				and:[sourceForm ~= destForm "no blits onto self"
				and:[cmFlags ~= 0 
					 or:[sourceMSB ~= destMSB 
					 or:[sourceDepth ~= destDepth]]]]]]. "no point using slower version"
	quickBlt
		ifTrue:
			[endOfSource := sourceBits + (sourcePitch * sourceHeight).
			 endOfDestination := destBits + (destPitch * destHeight)]
		ifFalse:
			[self lockSurfaces ifFalse:
				[^interpreterProxy primitiveFail]].
	left := destX.
	sourcePtr := interpreterProxy firstIndexableField: sourceString.
	startIndex to: stopIndex do:
		[:charIndex|
		ascii := interpreterProxy byteAtPointer: sourcePtr + charIndex - 1.
		glyphIndex := interpreterProxy fetchInteger: ascii ofObject: glyphMap.
		(glyphIndex < 0 or: [glyphIndex > maxGlyph])  ifTrue:
			[^interpreterProxy primitiveFail].
		sourceX := interpreterProxy fetchInteger: glyphIndex ofObject: xTable.
		width := (interpreterProxy fetchInteger: glyphIndex + 1 ofObject: xTable) - sourceX.
		interpreterProxy failed ifTrue:
			[^nil].
		self clipRange.	"Must clip here"
		(bbW > 0 and: [bbH > 0]) ifTrue:
			[quickBlt
				ifTrue:
					[self destMaskAndPointerInit.
					 self copyLoopPixMap.
					 "both, hDir and vDir are known to be > 0"
					 affectedL := dx.
					 affectedR := dx + bbW.
					 affectedT := dy.
					 affectedB := dy + bbH]
				ifFalse:
					[self copyBitsLockedAndClipped]].
		interpreterProxy failed ifTrue:
			[^nil].
		destX := destX + width + kernDelta].
	affectedL := left.
	quickBlt ifFalse:
		[self unlockSurfaces].
	self showDisplayBits.
	"store destX back"	
	interpreterProxy storeInteger: BBDestXIndex ofObject: bbObj withValue: destX.
	interpreterProxy pop: 6 "pop args, return rcvr"! !

!BitBltSimulation methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 3/24/2004 13:56'!
primitiveDrawLoop
	"Invoke the line drawing primitive."
	| rcvr xDelta yDelta |
	<export: true>
	rcvr := interpreterProxy stackValue: 2.
	xDelta := interpreterProxy stackIntegerValue: 1.
	yDelta := interpreterProxy stackIntegerValue: 0.
	(self loadBitBltFrom: rcvr) ifFalse:[^interpreterProxy primitiveFail].
	interpreterProxy failed ifFalse:[
		self drawLoopX: xDelta Y: yDelta.
		self showDisplayBits].
	interpreterProxy failed ifFalse:[interpreterProxy pop: 2].! !

!BitBltSimulation methodsFor: 'primitives' stamp: 'lpc 7/15/2016 11:19'!
primitivePixelValueAtX: xVal y: yVal
	"returns the single pixel at x@y.
	It does not handle LSB bitmaps right now.
	If x or y are < 0, return 0 to indicate transparent (cf BitBlt>bitPeekerFromForm: usage).
	Likewise if x>width or y>depth.
	Fail if the rcvr doesn't seem to be a Form, or x|y seem wrong"
	| rcvr bitmap depth ppW stride bitsSize word mask shift pixel |
	rcvr := self primitive: 'primitivePixelValueAt' parameters: #(SmallInteger SmallInteger) receiver: #Oop.
	
	"possible quick exit if x or y is -ve"
	(xVal < 0 or: [ yVal < 0 ] ) ifTrue:[^interpreterProxy integerObjectOf: 0].
	"check that rcvr is plausibly a Form or subclass"	
	rcvr := interpreterProxy stackValue: interpreterProxy methodArgumentCount.
	((interpreterProxy isPointers: rcvr) and: [(interpreterProxy slotSizeOf: rcvr) >= 4])
		ifFalse: [^interpreterProxy primitiveFail].

	"get the bits oop and width/height/depth"
	bitmap := interpreterProxy fetchPointer: FormBitsIndex ofObject: rcvr.
	(interpreterProxy isWordsOrBytes: bitmap) ifFalse: [^interpreterProxy primitiveFail].
	width := interpreterProxy fetchInteger: FormWidthIndex ofObject: rcvr.
	height := interpreterProxy fetchInteger: FormHeightIndex ofObject: rcvr.
	depth := interpreterProxy fetchInteger: FormDepthIndex ofObject: rcvr.
	"if width/height/depth are not integer, fail"
	interpreterProxy failed ifTrue:[^nil].

	"possible quick exit if x or y is >= extent of form. This also catches cases where the width/height are < 0"
	(xVal >= width or: [ yVal >= height ] ) ifTrue:[^interpreterProxy integerObjectOf: 0].

	"we don't handle LSB Forms yet"
	depth < 0 ifTrue:[^interpreterProxy primitiveFail].
	
	"OK so now we know we have a plausible Form, the width/height/depth/x/y are all reasonable and it's time to plunder the bitmap"
	ppW := 32//depth. "pixels in each word"
	stride := (width + (ppW  -1)) // ppW. "how many words per row of pixels"
	bitsSize := interpreterProxy byteSizeOf: bitmap.
	bitsSize >= (stride * height * 4 "bytes per word")
		ifFalse: [^interpreterProxy primitiveFail].
	word := interpreterProxy fetchLong32:(yVal * stride) + (xVal//ppW) ofObject: bitmap. "load the word that contains our target"
	mask := 16rFFFFFFFF >> (32 - depth). "make a mask to isolate the pixel within that word"
	shift := 32 - (((xVal bitAnd: ppW-1) + 1) * depth). "this is the tricky MSB part - we mask the xVal to find how far into the word we need, then add 1 for the pixel we're looking for, then * depth to get the bit shift"
	pixel := (word >> shift) bitAnd: mask. "shift, mask and dim the lights"
	^ pixel asPositiveIntegerObj "pop the incoming and push our answer"
! !

!BitBltSimulation methodsFor: 'primitives' stamp: 'eem 9/28/2018 14:50'!
primitiveWarpBits
	"Invoke the warpBits primitive. If the destination is the display, then copy it to the screen."
	| rcvr |
	<export: true>
	rcvr := interpreterProxy stackValue: interpreterProxy methodArgumentCount.
	(self loadWarpBltFrom: rcvr) 
		ifFalse:[^interpreterProxy primitiveFail].
	self warpBits.
	interpreterProxy failed ifTrue:[^nil].
	self showDisplayBits.
	interpreterProxy failed ifTrue:[^nil].
	interpreterProxy methodReturnReceiver! !

!BitBltSimulation methodsFor: 'surface support' stamp: 'eem 10/1/2018 14:23'!
querySurfaceFn: handle _: widthPtr _: heightPtr _: depthPtr _: endianPtr
	"Query the dimension of an OS surface.
	This method is provided so that in case the inst vars of the
	source form are broken, *actual* values of the OS surface
	can be obtained. This might, for instance, happen if the user
	resizes the main window.
	This is a simulation of the querySurfaceFn function call; simulate as a failure."
	<doNotGenerate>
	^false! !

!BitBltSimulation methodsFor: 'setup' stamp: 'eem 3/14/2017 16:00'!
reloadDestAndSourceForms
	"A GC has occurred.  The destForm must be updated.  But where to derive it from?
	  For copyBits and warpBits it is derived from the receiver.  But for a BalloonEnginePlugin
	  it should be obtained from (interpreterProxy fetchPointer: BEBitBltIndex ofObject: engine).
	  For the moment implement something that works for these two cases."
	<inline: false>
	| receiver |
	receiver := interpreterProxy stackValue: interpreterProxy methodArgumentCount.
	bitBltIsReceiver ifFalse: [receiver := interpreterProxy fetchPointer: BEBitBltIndex ofObject: receiver].
	destForm := interpreterProxy fetchPointer: BBDestFormIndex ofObject: receiver.
	sourceForm := interpreterProxy fetchPointer: BBSourceFormIndex ofObject: receiver! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:24'!
rgbAdd: sourceWord with: destinationWord
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	<var: #carryOverflowMask type: 'unsigned int'>
	<var: #componentMask type: 'unsigned int'>
	| componentMask carryOverflowMask |
	destDepth < 16 ifTrue:
		["Add each pixel separately"
		componentMask := 1<<destDepth-1.
		carryOverflowMask := 16rFFFFFFFF//componentMask<<(destDepth-1).
		^ self partitionedAdd: sourceWord to: destinationWord
						nBits: destDepth componentMask: componentMask carryOverflowMask: carryOverflowMask].
	destDepth = 16 ifTrue:
		["Add RGB components of each pixel separately"
		componentMask := 16r1F.
		carryOverflowMask := 16r42104210.
		^ (self partitionedAdd: (sourceWord bitAnd: 16r7FFF7FFF) to: (destinationWord bitAnd: 16r7FFF7FFF) "make sure that the unused bit is at 0"
						nBits: 5 componentMask: componentMask carryOverflowMask: carryOverflowMask)]
	ifFalse:
		["Add RGBA components of the pixel separately"
		componentMask := 16rFF.
		carryOverflowMask := 16r80808080.
		^ self partitionedAdd: sourceWord to: destinationWord
						nBits: 8 componentMask: componentMask carryOverflowMask: carryOverflowMask]! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 12:21'!
rgbComponentAlpha16
	"This version assumes 
		combinationRule = 41
		sourcePixSize = 32
		destPixSize = 16
		sourceForm ~= destForm.
	"
	<inline: false>  "This particular method should be optimized in itself"
	<var: #sourceWord type: #'unsigned int'>
	<var: #destWord type: #'unsigned int'>
	<var: #dstMask type: #'unsigned int'>
	
	| srcIndex dstIndex sourceWord srcAlpha destWord deltaX deltaY 
	srcY dstY dstMask srcShift ditherBase ditherIndex ditherThreshold |

	deltaY := bbH + 1. "So we can pre-decrement"
	srcY := sy.
	dstY := dy.
	srcShift := (dx bitAnd: 1) * 16.
	destMSB ifTrue:[srcShift := 16 - srcShift].
	mask1 := 16rFFFF << (16 - srcShift).
	"This is the outer loop"
	[(deltaY := deltaY - 1) ~= 0] whileTrue:[
		srcIndex := sourceBits + (srcY * sourcePitch) + (sx * 4).
		dstIndex := destBits + (dstY * destPitch) + (dx // 2 * 4).
		ditherBase := (dstY bitAnd: 3) * 4.
		ditherIndex := (sx bitAnd: 3) - 1. "For pre-increment"
		deltaX := bbW + 1. "So we can pre-decrement"
		dstMask := mask1.
		dstMask = 16rFFFF ifTrue:[srcShift := 16] ifFalse:[srcShift := 0].

		"This is the inner loop"
		[(deltaX := deltaX - 1) ~= 0] whileTrue:[
			ditherThreshold := ditherMatrix4x4 at: ditherBase + (ditherIndex := ditherIndex + 1 bitAnd: 3).
			sourceWord := self srcLongAt: srcIndex.
			srcAlpha := sourceWord bitAnd: 16rFFFFFF.
				srcAlpha = 0 ifFalse:[ "0 < srcAlpha"
					"If we have to mix colors then just copy a single word"
					destWord := self dstLongAt: dstIndex.
					destWord := destWord bitAnd: dstMask bitInvert32.
					destWord := destWord >> srcShift.
					"Expand from 16 to 32 bit by adding zero bits"
					destWord := (((destWord bitAnd: 16r7C00) bitShift: 9) bitOr:
									((destWord bitAnd: 16r3E0) bitShift: 6)) bitOr:
								(((destWord bitAnd: 16r1F) bitShift: 3) bitOr:
									16rFF000000).
					"Mix colors"
					sourceWord := self rgbComponentAlpha32: sourceWord with: destWord.
					"And dither"
					sourceWord := self dither32To16: sourceWord threshold: ditherThreshold.
					sourceWord = 0 
						ifTrue:[sourceWord := 1 << srcShift]
						ifFalse:[sourceWord := sourceWord << srcShift].
					"Store back"
					self dstLongAt: dstIndex put: sourceWord mask: dstMask.
				].
			srcIndex := srcIndex + 4.
			destMSB
				ifTrue:[srcShift = 0 ifTrue:[dstIndex := dstIndex + 4]]
				ifFalse:[srcShift = 0 ifFalse:[dstIndex := dstIndex + 4]].
			srcShift := srcShift bitXor: 16. "Toggle between 0 and 16"
			dstMask := dstMask bitInvert32. "Mask other half word"
		].
		srcY := srcY + 1.
		dstY := dstY + 1.
	].
! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 7/13/2016 17:04'!
rgbComponentAlpha32
	"This version assumes 
		combinationRule = 41
		sourcePixSize = destPixSize = 32
		sourceForm ~= destForm.
	Note: The inner loop has been optimized for dealing
		with the special case of aR = aG = aB = 0 
	"
	| srcIndex dstIndex sourceWord srcAlpha destWord deltaX deltaY srcY dstY |

	<inline: false> "This particular method should be optimized in itself"

	"Give the compile a couple of hints"
	<var: #deltaX type: 'register sqInt'>
	<var: #sourceWord type: #'unsigned int'>
	<var: #destWord type: #'unsigned int'>

	"The following should be declared as pointers so the compiler will
	notice that they're used for accessing memory locations 
	(good to know on an Intel architecture) but then the increments
	would be different between ST code and C code so must hope the
	compiler notices what happens (MS Visual C does)"
	<var: #srcIndex type: 'register sqIntptr_t'>
	<var: #dstIndex type: 'register sqIntptr_t'>
	
	deltaY := bbH + 1. "So we can pre-decrement"
	srcY := sy.
	dstY := dy.

	"This is the outer loop"
	[(deltaY := deltaY - 1) ~= 0] whileTrue:[
		srcIndex := sourceBits + (srcY * sourcePitch) + (sx * 4).
		dstIndex := destBits + (dstY * destPitch) + (dx * 4).
		deltaX := bbW + 1. "So we can pre-decrement"

		"This is the inner loop"
		[(deltaX := deltaX - 1) ~= 0] whileTrue:[
			sourceWord := self srcLongAt: srcIndex.
			srcAlpha := sourceWord bitAnd:16rFFFFFF.
				srcAlpha = 0 ifTrue:[
					srcIndex := srcIndex + 4.
					dstIndex := dstIndex + 4.
					"Now skip as many words as possible,"
					[(deltaX := deltaX - 1) ~= 0 and:[
						((sourceWord := self srcLongAt: srcIndex) bitAnd:16rFFFFFF) = 0]]
						whileTrue:[
							srcIndex := srcIndex + 4.
							dstIndex := dstIndex + 4.
						].
					"Adjust deltaX"
					deltaX := deltaX + 1.
				] ifFalse:[ "0 < srcAlpha"
					"If we have to mix colors then just copy a single word"
					destWord := self dstLongAt: dstIndex.
					destWord := self rgbComponentAlpha32: sourceWord with: destWord.
					self dstLongAt: dstIndex put: destWord.
					srcIndex := srcIndex + 4.
					dstIndex := dstIndex + 4.
				].
		].
		srcY := srcY + 1.
		dstY := dstY + 1.
	].! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'IgorStasenko 7/12/2011 08:32'!
rgbComponentAlpha32: sourceWord with: destinationWord
	"
	componentAlphaModeColor is the color,
	sourceWord contains an alpha value for each component of RGB
	each of which is encoded as0 meaning 0.0 and 255 meaning 1.0 .
	the rule is...
	
	color = componentAlphaModeColor.
	colorAlpha = componentAlphaModeAlpha.
	mask = sourceWord.
	dst.A =  colorAlpha + (1 - colorAlpha) * dst.A
      dst.R = color.R * mask.R * colorAlpha + (1 - (mask.R * colorAlpha)) * dst.R
      dst.G = color.G * mask.G * colorAlpha + (1 - (mask.G* colorAlpha)) * dst.G
      dst.B = color.B * mask.B * colorAlpha + (1 - (mask.B* colorAlpha)) * dst.B
	"
	<inline: false> "Do NOT inline this into optimized loops"
	| alpha dstMask srcColor srcAlpha b g r a aB aG aR aA answer s d |
		
	alpha := sourceWord.
	alpha = 0 ifTrue:[^destinationWord].
	srcColor := componentAlphaModeColor.
	srcAlpha := componentAlphaModeAlpha bitAnd: 255.
	
	aB := alpha bitAnd: 255.
	alpha := alpha >> 8.
	aG := alpha bitAnd: 255.
	alpha := alpha >> 8.
	aR := alpha bitAnd: 255.
	alpha := alpha >> 8.
	aA := alpha bitAnd: 255.	

	srcAlpha = 255 
		ifFalse:[
			aA := aA * srcAlpha >> 8.
			aR := aR * srcAlpha >> 8.
			aG := aG * srcAlpha >> 8.
			aB := aB * srcAlpha >> 8].
			
	dstMask := destinationWord.
	d := dstMask bitAnd: 255.
	s := srcColor bitAnd: 255.
	ungammaLookupTable == nil
		ifFalse:[
			d := ungammaLookupTable at: d.
			s := ungammaLookupTable at: s.].
	b := (d * (255 - aB) >> 8) + (s * aB >> 8).
	b > 255 ifTrue:[b := 255].
	gammaLookupTable == nil
		ifFalse:[	
			b := gammaLookupTable at: b].
	dstMask := dstMask >> 8.
	srcColor := srcColor >> 8.
	d := dstMask bitAnd: 255.
	s := srcColor bitAnd: 255.
	ungammaLookupTable == nil
		ifFalse:[
			d := ungammaLookupTable at: d.
			s := ungammaLookupTable at: s.].
	g := (d * (255 - aG) >> 8) + (s * aG >> 8).
	g > 255 ifTrue:[g := 255].
	gammaLookupTable == nil
		ifFalse:[	
			g := gammaLookupTable at: g].
	dstMask := dstMask >> 8.
	srcColor := srcColor >> 8.
	d := dstMask bitAnd: 255.
	s := srcColor bitAnd: 255.
	ungammaLookupTable == nil
		ifFalse:[
			d := ungammaLookupTable at: d.
			s := ungammaLookupTable at: s.].
	r := (d * (255 - aR) >> 8) + (s * aR >> 8).
	r > 255 ifTrue:[r := 255].
	gammaLookupTable == nil
		ifFalse:[	
			r := gammaLookupTable at: r].
	dstMask := dstMask >> 8.
	srcColor := srcColor >> 8.
	a := ((dstMask bitAnd: 255) * (255 - aA) >> 8) + aA. "no need to gamma correct alpha value ?"
	a > 255 ifTrue:[a := 255].
	answer := (((((a << 8) + r) << 8) + g) << 8) + b.
	^answer	! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 12:20'!
rgbComponentAlpha8
	"This version assumes 
		combinationRule = 41
		sourcePixSize = 32
		destPixSize = 8
		sourceForm ~= destForm.
	Note: This is not real blending since we don't have the source colors available.
	"
	
	| srcIndex dstIndex sourceWord srcAlpha destWord deltaX deltaY 
	srcY dstY dstMask srcShift adjust mappingTable mapperFlags |
	
	<inline: false>  "This particular method should be optimized in itself"
	<var: #mappingTable declareC:'unsigned int *mappingTable'>
	<var: #sourceWord type: #'unsigned int'>
	<var: #destWord type: #'unsigned int'>
	<var: #dstMask type: #'unsigned int'>
	
	mappingTable := self default8To32Table.
	mapperFlags := cmFlags bitAnd: ColorMapNewStyle bitInvert32.
	deltaY := bbH + 1. "So we can pre-decrement"
	srcY := sy.
	dstY := dy.
	mask1 := ((dx bitAnd: 3) * 8).
	destMSB ifTrue:[mask1 := 24 - mask1].
	mask2 := AllOnes bitXor:(16rFF << mask1).
	(dx bitAnd: 1) = 0 
		ifTrue:[adjust := 0]
		ifFalse:[adjust := 16r1F1F1F1F].
	(dy bitAnd: 1) = 0
		ifTrue:[adjust := adjust bitXor: 16r1F1F1F1F].
	"This is the outer loop"
	[(deltaY := deltaY - 1) ~= 0] whileTrue:[
		adjust := adjust bitXor: 16r1F1F1F1F.
		srcIndex := sourceBits + (srcY * sourcePitch) + (sx * 4).
		dstIndex := destBits + (dstY * destPitch) + (dx // 4 * 4).
		deltaX := bbW + 1. "So we can pre-decrement"
		srcShift := mask1.
		dstMask := mask2.

		"This is the inner loop"
		[(deltaX := deltaX - 1) ~= 0] whileTrue:[
			sourceWord := ((self srcLongAt: srcIndex) bitAnd: (adjust bitInvert32)) + adjust.
			srcAlpha := sourceWord bitAnd: 16rFFFFFF.
			"set srcAlpha to the average of the 3 separate aR,Ag,AB values"
			srcAlpha := ((srcAlpha >> 16) + (srcAlpha >> 8 bitAnd: 16rFF) + (srcAlpha bitAnd: 16rFF)) // 3.
			srcAlpha > 31 ifTrue:["Everything below 31 is transparent"
				srcAlpha > 224 
					ifTrue: ["treat everything above 224 as opaque"
						sourceWord := 16rFFFFFFFF].
				destWord := self dstLongAt: dstIndex.
				destWord := destWord bitAnd: dstMask bitInvert32.
				destWord := destWord >> srcShift.
				destWord := mappingTable at: destWord.
				sourceWord := self rgbComponentAlpha32: sourceWord with: destWord.
				sourceWord := self mapPixel: sourceWord flags: mapperFlags.
				sourceWord := sourceWord << srcShift.
				"Store back"
				self dstLongAt: dstIndex put: sourceWord mask: dstMask.
			].
			srcIndex := srcIndex + 4.
			destMSB ifTrue:[
				srcShift = 0 
					ifTrue:[dstIndex := dstIndex + 4.
							srcShift := 24.
							dstMask := 16r00FFFFFF]
					ifFalse:[srcShift := srcShift - 8.
							dstMask := (dstMask >> 8) bitOr: 16rFF000000].
			] ifFalse:[
				srcShift = 32
					ifTrue:[dstIndex := dstIndex + 4.
							srcShift := 0.
							dstMask := 16rFFFFFF00]
					ifFalse:[srcShift := srcShift + 8.
							dstMask := dstMask << 8 bitOr: 255].
			].
			adjust := adjust bitXor: 16r1F1F1F1F.
		].
		srcY := srcY + 1.
		dstY := dstY + 1.
	].
! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'tpr 5/15/2021 14:27'!
rgbComponentAlpha: sourceWord with: destinationWord
	"
	componentAlphaModeColor is the color,
	sourceWord contains an alpha value for each component of RGB
	each of which is encoded as0 meaning 0.0 and 255 meaning 1.0 .
	the rule is...
	
	color = componentAlphaModeColor.
	colorAlpha = componentAlphaModeAlpha.
	mask = sourceWord.
	dst.A =  colorAlpha + (1 - colorAlpha) * dst.A
      dst.R = color.R * mask.R * colorAlpha + (1 - (mask.R * colorAlpha)) * dst.R
      dst.G = color.G * mask.G * colorAlpha + (1 - (mask.G* colorAlpha)) * dst.G
      dst.B = color.B * mask.B * colorAlpha + (1 - (mask.B* colorAlpha)) * dst.B
	"
	<inline: false> "Do NOT inline this into optimized loops"
	| alpha |
		
	alpha := sourceWord.
		"This is not a valid optimisation because alpha == 0 can change the destination
		 due to rounding errors in blending and/or in gamma lookup round-trip
		alpha = 0 ifTrue:[^destinationWord]."
	^self partitionedRgbComponentAlpha: sourceWord dest: destinationWord nBits: destDepth nPartitions: destPPW.! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:24'!
rgbDiff: sourceWord with: destinationWord
	"Subract the pixels in the source and destination, color by color,
	and return the sum of the absolute value of all the differences.
	For non-rgb, return the number of differing pixels."
	| pixMask destShifted sourceShifted destPixVal bitsPerColor rgbMask sourcePixVal diff maskShifted |
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	<var: #sourceShifted type: 'unsigned int'>
	<var: #destShifted type: 'unsigned int'>
	<var: #maskShifted type: 'unsigned int'>
	<var: #pixMask type: 'unsigned int'>
	<var: #rgbMask type: 'unsigned int'>
	pixMask := maskTable at: destDepth.
	destDepth = 16
		ifTrue: [bitsPerColor := 5.  rgbMask := 16r1F]
		ifFalse: [bitsPerColor := 8.  rgbMask := 16rFF].
	maskShifted := destMask.
	destShifted := destinationWord.
	sourceShifted := sourceWord.
	1 to: destPPW do:
		[:i |
		(maskShifted bitAnd: pixMask) > 0 ifTrue:
			["Only tally pixels within the destination rectangle"
			destPixVal := destShifted bitAnd: pixMask.
			sourcePixVal := sourceShifted bitAnd: pixMask.
			destDepth < 16
				ifTrue: [sourcePixVal = destPixVal
							ifTrue: [diff := 0]
							ifFalse: [diff := 1]]
				ifFalse: [diff := (self partitionedSub: sourcePixVal from: destPixVal
								nBits: bitsPerColor nPartitions: 3).
						diff := (diff bitAnd: rgbMask)
							+ (diff>>bitsPerColor bitAnd: rgbMask)
							+ ((diff>>bitsPerColor)>>bitsPerColor bitAnd: rgbMask)].
			bitCount := bitCount + diff].
		maskShifted := maskShifted >> destDepth.
		sourceShifted := sourceShifted >> destDepth.
		destShifted := destShifted >> destDepth].
	^ destinationWord  "For no effect on dest"
! !

!BitBltSimulation methodsFor: 'color mapping' stamp: 'ar 10/28/1999 16:02'!
rgbMap16To32: sourcePixel
	"Convert the given 16bit pixel value to a 32bit RGBA value.
 	Note: This method is intended to deal with different source formats."
	^(((sourcePixel bitAnd: 31) << 3) bitOr:
		((sourcePixel bitAnd: 16r3E0) << 6)) bitOr:
			((sourcePixel bitAnd: 16r7C00) << 9)! !

!BitBltSimulation methodsFor: 'color mapping' stamp: 'ar 10/27/1999 14:28'!
rgbMap32To32: sourcePixel
	"Convert the given 32bit pixel value to a 32bit RGBA value.
 	Note: This method is intended to deal with different source formats."
	^sourcePixel "For now do it simple"! !

!BitBltSimulation methodsFor: 'pixel mapping' stamp: 'di (auto pragmas 12/08) 4/10/1999 17:27'!
rgbMap: sourcePixel from: nBitsIn to: nBitsOut
	"Convert the given pixel value with nBitsIn bits for each color component to a pixel value with nBitsOut bits for each color component. Typical values for nBitsIn/nBitsOut are 3, 5, or 8."
	| mask d srcPix destPix |
	<inline: true>
	(d := nBitsOut - nBitsIn) > 0
		ifTrue:
			["Expand to more bits by zero-fill"
			mask := (1 << nBitsIn) - 1.  "Transfer mask"
			srcPix := sourcePixel << d.
			mask := mask << d.
			destPix := srcPix bitAnd: mask.
			mask := mask << nBitsOut.
			srcPix := srcPix << d.
			^ destPix + (srcPix bitAnd: mask)
				 	+ (srcPix << d bitAnd: mask << nBitsOut)]
		ifFalse:
			["Compress to fewer bits by truncation"
			d = 0 ifTrue:
				[nBitsIn = 5 ifTrue:
					["Sometimes called with 16 bits, though pixel is 15,
					but we must never return more than 15."
					^ sourcePixel bitAnd: 16r7FFF].
				nBitsIn = 8 ifTrue:
					["Sometimes called with 32 bits, though pixel is 24,
					but we must never return more than 24."
					^ sourcePixel bitAnd: 16rFFFFFF].
				^ sourcePixel].  "no compression"
			sourcePixel = 0 ifTrue: [^ sourcePixel].  "always map 0 (transparent) to 0"
			d := nBitsIn - nBitsOut.
			mask := (1 << nBitsOut) - 1.  "Transfer mask"
			srcPix := sourcePixel >> d.
			destPix := srcPix bitAnd: mask.
			mask := mask << nBitsOut.
			srcPix := srcPix >> d.
			destPix := destPix + (srcPix bitAnd: mask)
					+ (srcPix >> d bitAnd: mask << nBitsOut).
			destPix = 0 ifTrue: [^ 1].  "Dont fall into transparent by truncation"
			^ destPix]! !

!BitBltSimulation methodsFor: 'color mapping' stamp: 'nice 3/16/2016 13:21'!
rgbMapPixel: sourcePixel flags: mapperFlags
	"Perform the RGBA conversion for the given source pixel"
	| val |
	<inline: true>
	val := 			((sourcePixel bitAnd: (cmMaskTable at: 0)) bitShift: (cmShiftTable at: 0) signedIntFromLong).
	val := val bitOr: ((sourcePixel bitAnd: (cmMaskTable at: 1)) bitShift: (cmShiftTable at: 1) signedIntFromLong).
	val := val bitOr: ((sourcePixel bitAnd: (cmMaskTable at: 2)) bitShift: (cmShiftTable at: 2) signedIntFromLong).
		  ^val bitOr: ((sourcePixel bitAnd: (cmMaskTable at: 3)) bitShift: (cmShiftTable at: 3) signedIntFromLong).
! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:24'!
rgbMax: sourceWord with: destinationWord
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	destDepth < 16 ifTrue:
		["Max each pixel separately"
		^ self partitionedMax: sourceWord with: destinationWord
						nBits: destDepth nPartitions: destPPW].
	destDepth = 16 ifTrue:
		["Max RGB components of each pixel separately"
		^ (self partitionedMax: sourceWord with: destinationWord
						nBits: 5 nPartitions: 3)
		+ ((self partitionedMax: sourceWord>>16 with: destinationWord>>16
						nBits: 5 nPartitions: 3) << 16)]
	ifFalse:
		["Max RGBA components of the pixel separately"
		^ self partitionedMax: sourceWord with: destinationWord
						nBits: 8 nPartitions: 4]! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:24'!
rgbMin: sourceWord with: destinationWord
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	destDepth < 16 ifTrue:
		["Min each pixel separately"
		^ self partitionedMin: sourceWord with: destinationWord
						nBits: destDepth nPartitions: destPPW].
	destDepth = 16 ifTrue:
		["Min RGB components of each pixel separately"
		^ (self partitionedMin: sourceWord with: destinationWord
						nBits: 5 nPartitions: 3)
		+ ((self partitionedMin: sourceWord>>16 with: destinationWord>>16
						nBits: 5 nPartitions: 3) << 16)]
	ifFalse:
		["Min RGBA components of the pixel separately"
		^ self partitionedMin: sourceWord with: destinationWord
						nBits: 8 nPartitions: 4]! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:25'!
rgbMinInvert: wordToInvert with: destinationWord
	| sourceWord |
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #wordToInvert type: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	sourceWord := wordToInvert bitInvert32.
	destDepth < 16 ifTrue:
		["Min each pixel separately"
		^ self partitionedMin: sourceWord with: destinationWord
						nBits: destDepth nPartitions: destPPW].
	destDepth = 16 ifTrue:
		["Min RGB components of each pixel separately"
		^ (self partitionedMin: sourceWord with: destinationWord
						nBits: 5 nPartitions: 3)
		+ ((self partitionedMin: sourceWord>>16 with: destinationWord>>16
						nBits: 5 nPartitions: 3) << 16)]
	ifFalse:
		["Min RGBA components of the pixel separately"
		^ self partitionedMin: sourceWord with: destinationWord
						nBits: 8 nPartitions: 4]! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:25'!
rgbMul: sourceWord with: destinationWord
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	destDepth < 16 ifTrue:
		["Mul each pixel separately"
		^ self partitionedMul: sourceWord with: destinationWord
						nBits: destDepth nPartitions: destPPW].
	destDepth = 16 ifTrue:
		["Mul RGB components of each pixel separately"
		^ (self partitionedMul: sourceWord with: destinationWord
						nBits: 5 nPartitions: 3)
		+ ((self partitionedMul: sourceWord>>16 with: destinationWord>>16
						nBits: 5 nPartitions: 3) << 16)]
	ifFalse:
		["Mul RGBA components of the pixel separately"
		^ self partitionedMul: sourceWord with: destinationWord
						nBits: 8 nPartitions: 4]

"	| scanner |
	Display repaintMorphicDisplay.
	scanner := DisplayScanner quickPrintOn: Display.
	MessageTally time: [0 to: 760 by: 4 do:  [:y |scanner drawString: 'qwrepoiuasfd=)(/&()=#!!lkjzxv.,mn124+09857907QROIYTOAFDJZXNBNB,M-.,Mqwrepoiuasfd=)(/&()=#!!lkjzxv.,mn124+09857907QROIYTOAFDJZXNBNB,M-.,M1234124356785678' at: 0@y]]. "! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:25'!
rgbSub: sourceWord with: destinationWord
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	destDepth < 16 ifTrue:
		["Sub each pixel separately"
		^ self partitionedSub: sourceWord from: destinationWord
						nBits: destDepth nPartitions: destPPW].
	destDepth = 16 ifTrue:
		["Sub RGB components of each pixel separately"
		^ (self partitionedSub: sourceWord from: destinationWord
						nBits: 5 nPartitions: 3)
		+ ((self partitionedSub: sourceWord>>16 from: destinationWord>>16
						nBits: 5 nPartitions: 3) << 16)]
	ifFalse:
		["Sub RGBA components of the pixel separately"
		^ self partitionedSub: sourceWord from: destinationWord
						nBits: 8 nPartitions: 4]! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'ar 4/26/2001 19:43'!
setupColorMasks
	"WARNING: For WarpBlt w/ smoothing the source depth is wrong here!!"
	| bits targetBits |
	bits := targetBits := 0.
	sourceDepth <= 8 ifTrue:[^nil].
	sourceDepth = 16 ifTrue:[bits := 5].
	sourceDepth = 32 ifTrue:[bits := 8].

	cmBitsPerColor = 0
		ifTrue:["Convert to destDepth"
				destDepth <= 8 ifTrue:[^nil].
				destDepth = 16 ifTrue:[targetBits := 5].
				destDepth = 32 ifTrue:[targetBits := 8]]
		ifFalse:[targetBits := cmBitsPerColor].

	self setupColorMasksFrom: bits to: targetBits! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'ar (auto pragmas 12/08) 5/4/2001 14:53'!
setupColorMasksFrom: srcBits to: targetBits
	"Setup color masks for converting an incoming RGB pixel value from srcBits to targetBits."
	| mask shifts masks deltaBits |
	<var: #shifts declareC:'static int shifts[4] = {0, 0, 0, 0}'>
	<var: #masks declareC:'static unsigned int masks[4] = {0, 0, 0, 0}'>
	self cCode:'' inSmalltalk:[
		shifts := CArrayAccessor on: (IntegerArray new: 4).
		masks := CArrayAccessor on: (WordArray new: 4).
	].
	deltaBits := targetBits - srcBits.
	deltaBits = 0 ifTrue:[^0].
	deltaBits <= 0
		ifTrue:[	mask := 1 << targetBits - 1.
				"Mask for extracting a color part of the source"
				masks at: RedIndex put: mask << (srcBits*2 - deltaBits).
				masks at: GreenIndex put: mask << (srcBits - deltaBits).
				masks at: BlueIndex put: mask << (0 - deltaBits).
				masks at: AlphaIndex put: 0]
		ifFalse:[	mask := 1 << srcBits - 1.
				"Mask for extracting a color part of the source"
				masks at: RedIndex put: mask << (srcBits*2).
				masks at: GreenIndex put: mask << srcBits.
				masks at: BlueIndex put: mask].

	"Shifts for adjusting each value in a cm RGB value"
	shifts at: RedIndex put: deltaBits * 3.
	shifts at: GreenIndex put: deltaBits * 2.
	shifts at: BlueIndex put: deltaBits.
	shifts at: AlphaIndex put: 0.

	cmShiftTable := shifts.
	cmMaskTable := masks.
	cmFlags := cmFlags bitOr: (ColorMapPresent bitOr: ColorMapFixedPart).
! !

!BitBltSimulation methodsFor: 'interpreter interface' stamp: 'eem 3/14/2017 16:01'!
showDisplayBits
	self ensureDestAndSourceFormsAreValid.
	interpreterProxy 
		showDisplayBits: destForm
		Left: affectedL
		Top: affectedT
		Right: affectedR
		Bottom: affectedB! !

!BitBltSimulation methodsFor: 'setup' stamp: 'nice 2/22/2020 09:06'!
sourceSkewAndPointerInit
	"This is only used when source and dest are same depth,
	ie, when the barrel-shift copy loop is used."
	| sxLowBits dxLowBits pixPerM1 startBits m1 |
	<inline: true>
	<var: 'm1' type: #'unsigned int'>
	self assert: (destPPW = sourcePPW and: [destMSB = sourceMSB and: [destDepth = sourceDepth]]).
	pixPerM1 := destPPW - 1.  "A mask, assuming power of two"
	sxLowBits := sx bitAnd: pixPerM1.
	dxLowBits := dx bitAnd: pixPerM1.
	"how many pixels in first word"
	startBits := hDir > 0
					ifTrue: [sourcePPW - (sx bitAnd: pixPerM1)]
					ifFalse: [(sx bitAnd: pixPerM1) + 1].	
	m1 := destMSB = (hDir > 0)
			ifTrue: [AllOnes >> (32 - (startBits * destDepth))] 
			ifFalse: [AllOnes << (32 - (startBits * destDepth)) bitAnd: AllOnes].
	preload := bbW > startBits and: [(m1 bitAnd: mask1) ~= mask1]. "i.e. there are some missing bits"

	"calculate right-shift skew from source to dest"
	skew := destDepth * (sourceMSB ifTrue: [sxLowBits - dxLowBits] ifFalse: [dxLowBits - sxLowBits]).  " -32..32 "
	preload ifTrue: 
		[skew := skew < 0 ifTrue: [skew + 32] ifFalse: [skew - 32]].

	"Calc byte addr and delta from longWord info"
	sourceIndex := sourceBits + (sy * sourcePitch) + ((sx // sourcePPW) * 4).
	"calculate increments from end of 1 line to start of next"
	sourceDelta := (sourcePitch * vDir) - (4 * (nWords * hDir)).

	preload ifTrue: "Compensate for extra source word fetched"
		[sourceDelta := sourceDelta - (4 * hDir)].

	self deny: (preload and: [skew = 0]).
	self assert: (skew between: -31 and: 31)! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:25'!
sourceWord: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^sourceWord! !

!BitBltSimulation methodsFor: 'memory access' stamp: 'eem 10/15/2018 16:13'!
srcLongAt: idx
	<inline: #always>
	self assert: idx asUnsignedInteger < endOfSource.
	^self long32At: idx! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:25'!
subWord: sourceWord with: destinationWord
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	^sourceWord - destinationWord! !

!BitBltSimulation methodsFor: 'combination rules' stamp: 'nice 10/29/2015 22:25'!
tallyIntoMap: sourceWord with: destinationWord
	"Tally pixels into the color map.  Those tallied are exactly those
	in the destination rectangle.  Note that the source should be 
	specified == destination, in order for the proper color map checks 
	to be performed at setup."
	| mapIndex pixMask destShifted maskShifted pixVal |
	<inline: false>
	<returnTypeC: 'unsigned int'>
	<var: #sourceWord type: 'unsigned int'>
	<var: #destinationWord type: 'unsigned int'>
	<var: #pixMask type: 'unsigned int'>
	<var: #destShifted type: 'unsigned int'>
	<var: #maskShifted type: 'unsigned int'>
	(cmFlags bitAnd: (ColorMapPresent bitOr: ColorMapIndexedPart)) = 
		(ColorMapPresent bitOr: ColorMapIndexedPart)
			ifFalse: [^ destinationWord "no op"].
	pixMask := maskTable at: destDepth.
	destShifted := destinationWord.
	maskShifted := destMask.
	1 to: destPPW do:
		[:i |
		(maskShifted bitAnd: pixMask) = 0 ifFalse:
			["Only tally pixels within the destination rectangle"
			pixVal := destShifted bitAnd: pixMask.
			destDepth < 16
				ifTrue: [mapIndex := pixVal]
				ifFalse: [destDepth = 16
					ifTrue: [mapIndex := self rgbMap: pixVal from: 5 to: cmBitsPerColor]
					ifFalse: [mapIndex := self rgbMap: pixVal from: 8 to: cmBitsPerColor]].
			self tallyMapAt: mapIndex put: (self tallyMapAt: mapIndex) + 1].
		maskShifted := maskShifted >> destDepth.
		destShifted := destShifted >> destDepth].
	^ destinationWord  "For no effect on dest"! !

!BitBltSimulation methodsFor: 'memory access' stamp: 'eem 10/15/2018 16:30'!
tallyMapAt: idx
	"Return the word at position idx from the colorMap"
	<inline: #always>
	^cmLookupTable at: (idx bitAnd: cmMask)! !

!BitBltSimulation methodsFor: 'memory access' stamp: 'eem 10/15/2018 16:30'!
tallyMapAt: idx put: value
	"Store the word at position idx in the colorMap"
	<inline: #always>
	^cmLookupTable at: (idx bitAnd: cmMask) put: value! !

!BitBltSimulation methodsFor: 'setup' stamp: 'IgorStasenko 7/12/2011 08:16'!
tryCopyingBitsQuickly
	"Shortcut for stuff that's being run from the balloon engine.
	Since we do this at each scan line we should avoid the expensive 
	setup for source and destination."
	<inline: true>
	"We need a source."
	noSource ifTrue:[^false].
	"We handle only combinationRule 34 and combinationRule 41"	
	(combinationRule = 34 or:[combinationRule = 41]) ifFalse:[^false].
	"We handle only sourceDepth 32"
	(sourceDepth = 32) ifFalse:[^false].
	"We don't handle overlaps"
	(sourceForm = destForm) ifTrue:[^false].
	
	(combinationRule = 41)
		ifTrue:[
			destDepth = 32 
				ifTrue:[
					self rgbComponentAlpha32.
					affectedL := dx.
					affectedR := dx + bbW.
					affectedT := dy.
					affectedB := dy + bbH.
					^true].
			destDepth = 16 
				ifTrue:[
					self rgbComponentAlpha16.
					affectedL := dx.
					affectedR := dx + bbW.
					affectedT := dy.
					affectedB := dy + bbH.
					^true].
			destDepth = 8 
				ifTrue:[
					self rgbComponentAlpha8.
					affectedL := dx.
					affectedR := dx + bbW.
					affectedT := dy.
					affectedB := dy + bbH.
					^true].
			^false].
	
	"We need at least 8bit deep dest forms"
	(destDepth < 8) ifTrue:[^false].
	"If 8bit, then we want a color map"
	(destDepth = 8 and:[(cmFlags bitAnd: ColorMapPresent) = 0]) ifTrue:[^false].
	destDepth = 32 
		ifTrue:[self alphaSourceBlendBits32].
	destDepth = 16
		ifTrue:[self alphaSourceBlendBits16].
	destDepth = 8
		ifTrue:[self alphaSourceBlendBits8].
	affectedL := dx.
	affectedR := dx + bbW.
	affectedT := dy.
	affectedB := dy + bbH.
	^true! !

!BitBltSimulation methodsFor: 'surface support' stamp: 'eem 10/18/2018 12:00'!
unlockSurfaces
	"Unlock the bits of any OS surfaces."
	"See the comment in lockSurfaces. Similar rules apply. That is, the area provided in ioUnlockSurface
	 can be used to determine the dirty region after drawing. If a source is unlocked, then the area will
	 be (0,0,0,0) to indicate that no portion is dirty.  Note that if a GC happens during unlockSourceFn
	 (if it is effectively a callback) no matter.  No bits are touched after unlock."
	| sourceHandle destHandle destLocked |
	hasSurfaceLock ifFalse: [^self].
	unlockSurfaceFn = 0 ifTrue:
		[self loadSurfacePlugin ifFalse:
			[^self]].
	self ensureDestAndSourceFormsAreValid.
	destLocked := false.
	destHandle := interpreterProxy fetchPointer: FormBitsIndex ofObject: destForm.
	(interpreterProxy isIntegerObject: destHandle) ifTrue: "The destBits are always assumed to be dirty"
		[self unlockSurfaceFn: (interpreterProxy integerValueOf: destHandle)
							_: affectedL
							_: affectedT
							_: affectedR - affectedL
							_: affectedB - affectedT.
		 destBits := destPitch := 0.
		 destLocked := true].
	noSource ifFalse:
		[self ensureDestAndSourceFormsAreValid.
		 sourceHandle := interpreterProxy fetchPointer: FormBitsIndex ofObject: sourceForm.
		 (interpreterProxy isIntegerObject: sourceHandle) ifTrue:
			["Only unlock sourceHandle if different from destHandle"
			 (destLocked and: [sourceHandle = destHandle]) ifFalse:
				[self unlockSurfaceFn: (interpreterProxy integerValueOf: sourceHandle) _: 0 _: 0 _: 0 _: 0].
			sourceBits := sourcePitch := 0]].
	hasSurfaceLock := false! !

!BitBltSimulation methodsFor: 'setup' stamp: 'eem 1/28/2010 14:46'!
warpBits
	| ns |
	<inline: true>
	ns := noSource.  noSource := true.
		self clipRange.  "noSource suppresses sourceRect clipping"
		noSource := ns.
	(noSource or: [bbW <= 0 or: [bbH <= 0]]) ifTrue:
		["zero width or height; noop"
		affectedL := affectedR := affectedT := affectedB := 0.
		^ nil].

	self lockSurfaces ifFalse:[^interpreterProxy primitiveFail].
	self destMaskAndPointerInit.
	self warpLoop.
 
	hDir > 0
		ifTrue: [affectedL := dx.
				affectedR := dx + bbW]
		ifFalse: [affectedL := dx - bbW + 1.
				affectedR := dx + 1].
	vDir > 0
		ifTrue: [affectedT := dy.
				affectedB := dy + bbH]
		ifFalse: [affectedT := dy - bbH + 1.
				affectedB := dy + 1].
	self unlockSurfaces.! !

!BitBltSimulation methodsFor: 'inner loop' stamp: 'eem 10/16/2018 16:33'!
warpLoop
	"This version of the inner loop traverses an arbirary quadrilateral
	source, thus producing a general affine transformation."
	| skewWord halftoneWord mergeWord startBits
	  deltaP12x deltaP12y deltaP43x deltaP43y pAx pAy pBx pBy
	  xDelta yDelta smoothingCount sourceMapOop
	  nSteps nPix words destWord endBits mergeFnwith dstShiftInc dstShiftLeft mapperFlags |
	<inline: false>	
	<var: #mergeFnwith declareC: 'unsigned int (*mergeFnwith)(unsigned int, unsigned int)'>
	<var: #skewWord type: #'unsigned int'>
	<var: #halftoneWord type: #'unsigned int'>
	<var: #mergeWord type: #'unsigned int'>
	<var: #destWord type: #'unsigned int'>
	mergeFnwith := self cCoerce: (opTable at: combinationRule+1) to: 'unsigned int (*)(unsigned int, unsigned int)'.
	mergeFnwith.  "null ref for compiler"

	(interpreterProxy slotSizeOf: bitBltOop) >= (BBWarpBase+12)
		ifFalse: [^ interpreterProxy primitiveFail].
	nSteps := height-1.  nSteps <= 0 ifTrue: [nSteps := 1].

	pAx := self fetchIntOrFloat: BBWarpBase ofObject: bitBltOop.
	words := self fetchIntOrFloat: BBWarpBase+3 ofObject: bitBltOop.
	deltaP12x := self deltaFrom: pAx to: words nSteps: nSteps.
	deltaP12x < 0 ifTrue: [pAx := words - (nSteps*deltaP12x)].

	pAy := self fetchIntOrFloat: BBWarpBase+1 ofObject: bitBltOop.
	words := self fetchIntOrFloat: BBWarpBase+4 ofObject: bitBltOop.
	deltaP12y := self deltaFrom: pAy to: words nSteps: nSteps.
	deltaP12y < 0 ifTrue: [pAy := words - (nSteps*deltaP12y)].

	pBx := self fetchIntOrFloat: BBWarpBase+9 ofObject: bitBltOop.
	words := self fetchIntOrFloat: BBWarpBase+6 ofObject: bitBltOop.
	deltaP43x := self deltaFrom: pBx to: words nSteps: nSteps.
	deltaP43x < 0 ifTrue: [pBx := words - (nSteps*deltaP43x)].

	pBy := self fetchIntOrFloat: BBWarpBase+10 ofObject: bitBltOop.
	words := self fetchIntOrFloat: BBWarpBase+7 ofObject: bitBltOop.
	deltaP43y := self deltaFrom: pBy to: words nSteps: nSteps.
	deltaP43y < 0 ifTrue: [pBy := words - (nSteps*deltaP43y)].

	interpreterProxy failed ifTrue: [^ false].  "ie if non-integers above"
	interpreterProxy methodArgumentCount = 2
		ifTrue:
			[smoothingCount := interpreterProxy stackIntegerValue: 1.
			 sourceMapOop := interpreterProxy stackValue: 0.
			 sourceMapOop = interpreterProxy nilObject
				ifTrue:
					[sourceDepth < 16 ifTrue: "color map is required to smooth non-RGB dest"
						[^interpreterProxy primitiveFail]]
				ifFalse:
					[(interpreterProxy slotSizeOf: sourceMapOop) < (1 << sourceDepth) ifTrue: "sourceMap must be long enough for sourceDepth"
						[^interpreterProxy primitiveFail].
					sourceMapOop := self oopForPointer: (interpreterProxy firstIndexableField: sourceMapOop)]]
		ifFalse:
			[smoothingCount := 1.
			 sourceMapOop := interpreterProxy nilObject].
	nSteps := width-1.  nSteps <= 0 ifTrue: [nSteps := 1].
	startBits := destPPW - (dx bitAnd: destPPW-1).
	endBits := ((dx + bbW - 1) bitAnd: destPPW-1) + 1.
 	bbW < startBits ifTrue:[startBits := bbW].

	destY < clipY ifTrue:[
		"Advance increments if there was clipping in y"
		pAx := pAx + (clipY - destY * deltaP12x).
		pAy := pAy + (clipY - destY * deltaP12y).
		pBx := pBx + (clipY - destY * deltaP43x).
		pBy := pBy + (clipY - destY * deltaP43y)].

	"Setup values for faster pixel fetching."
	self warpLoopSetup.
	"Setup color mapping if not provided"
	(smoothingCount > 1 and:[(cmFlags bitAnd: ColorMapNewStyle) = 0]) ifTrue:
		[cmLookupTable
			ifNil: [destDepth = 16 ifTrue: [self setupColorMasksFrom: 8 to: 5]]
			ifNotNil: [self setupColorMasksFrom: 8 to: cmBitsPerColor]].
	mapperFlags := cmFlags bitAnd: ColorMapNewStyle bitInvert32.

	destMSB
		ifTrue:[	dstShiftInc := 0 - destDepth.
				dstShiftLeft := 32 - destDepth]
		ifFalse:[	dstShiftInc := destDepth.
				dstShiftLeft := 0].
	noHalftone ifTrue:
		[halftoneWord := AllOnes].
	1 to: bbH do:
		[ :i | "here is the vertical loop..."
		xDelta := self deltaFrom: pAx to: pBx nSteps: nSteps.
 		xDelta >= 0 ifTrue: [sx := pAx] ifFalse: [sx := pBx - (nSteps*xDelta)].
		yDelta := self deltaFrom: pAy to: pBy nSteps: nSteps.
 		yDelta >= 0 ifTrue: [sy := pAy] ifFalse: [sy := pBy - (nSteps*yDelta)].

		destMSB
			ifTrue:[dstBitShift := 32 - ((dx bitAnd: destPPW - 1) + 1 * destDepth)]
			ifFalse:[dstBitShift := (dx bitAnd: destPPW - 1) * destDepth].

		(destX < clipX) ifTrue:[
			"Advance increments if there was clipping in x"
			sx := sx + (clipX - destX * xDelta).
			sy := sy + (clipX - destX * yDelta).
		].

		noHalftone ifFalse:
			[halftoneWord := self halftoneAt: dy + i - 1].
		destMask := mask1.
		nPix := startBits.
		"Here is the inner loop..."
		words := nWords.
			["pick up word"
			smoothingCount = 1 ifTrue:["Faster if not smoothing"
				skewWord := self warpPickSourcePixels: nPix
								xDeltah: xDelta yDeltah: yDelta
								xDeltav: deltaP12x yDeltav: deltaP12y
								dstShiftInc: dstShiftInc flags: mapperFlags.
			] ifFalse:["more difficult with smoothing"
				skewWord := self warpPickSmoothPixels: nPix
						xDeltah: xDelta yDeltah: yDelta
						xDeltav: deltaP12x yDeltav: deltaP12y
						sourceMap: sourceMapOop
						smoothing: smoothingCount
						dstShiftInc: dstShiftInc.
			].
			"align next word access to left most pixel"
			dstBitShift := dstShiftLeft.
			destMask = AllOnes ifTrue:["avoid read-modify-write"
				mergeWord := self mergeFn: (skewWord bitAnd: halftoneWord)
								with: (self dstLongAt: destIndex).
				self dstLongAt: destIndex put: (destMask bitAnd: mergeWord).
			] ifFalse:[ "General version using dest masking"
				destWord := self dstLongAt: destIndex.
				mergeWord := self mergeFn: (skewWord bitAnd: halftoneWord)
								with: (destWord bitAnd: destMask).
				destWord := (destMask bitAnd: mergeWord) bitOr:
								(destWord bitAnd: destMask bitInvert32).
				self dstLongAt: destIndex put: destWord.
			].
			self incDestIndex: 4.
			words = 2 "e.g., is the next word the last word?"
				ifTrue:["set mask for last word in this row"
						destMask := mask2.
						nPix := endBits]
				ifFalse:["use fullword mask for inner loop"
						destMask := AllOnes.
						nPix := destPPW].
			(words := words - 1) = 0] whileFalse.
		"--- end of inner loop ---"
		pAx := pAx + deltaP12x.
		pAy := pAy + deltaP12y.
		pBx := pBx + deltaP43x.
		pBy := pBy + deltaP43y.
		self incDestIndex: destDelta]! !

!BitBltSimulation methodsFor: 'pixel mapping' stamp: 'ar (auto pragmas 12/08) 4/19/2001 12:42'!
warpLoopSetup
	"Setup values for faster pixel fetching."
	| words |
	<inline: true>

	"warpSrcShift = log2(sourceDepth)"
	warpSrcShift := 0.
	words := sourceDepth. "recycle temp"
	[words = 1] whileFalse:[
		warpSrcShift := warpSrcShift + 1.
		words := words >> 1].

	"warpSrcMask = mask for extracting one pixel from source word"
	warpSrcMask := maskTable at: sourceDepth.

	"warpAlignShift: Shift for aligning x position to word boundary"
	warpAlignShift := 5 - warpSrcShift.

	"warpAlignMask: Mask for extracting the pixel position from an x position"
	warpAlignMask := 1 << warpAlignShift - 1.

	"Setup the lookup table for source bit shifts"
	"warpBitShiftTable: given an sub-word x value what's the bit shift?"
	0 to: warpAlignMask do:[:i|
		sourceMSB
			ifTrue:[warpBitShiftTable at: i put: 32 - ( i + 1 << warpSrcShift )]
			ifFalse:[warpBitShiftTable at: i put: (i << warpSrcShift)]].
! !

!BitBltSimulation methodsFor: 'pixel mapping' stamp: 'nice 10/29/2015 12:10'!
warpPickSmoothPixels: nPixels
	xDeltah: xDeltah yDeltah: yDeltah
	xDeltav: xDeltav yDeltav: yDeltav
	sourceMap: sourceMap
	smoothing: n
	dstShiftInc: dstShiftInc
	"Pick n (sub-) pixels from the source form, mapped by sourceMap,
	average the RGB values, map by colorMap and return the new word.
	This version is only called from WarpBlt with smoothingCount > 1"
	| rgb x y a r g b xx yy xdh ydh xdv ydv dstMask destWord i j k nPix |
	<inline: false> "nope - too much stuff in here"
	<var: #rgb type: #'unsigned int'>
	dstMask := maskTable at: destDepth.
	destWord := 0.
	n = 2 "Try avoiding divides for most common n (divide by 2 is generated as shift)"
		ifTrue:[xdh := xDeltah // 2. ydh := yDeltah // 2. 
				xdv := xDeltav // 2. ydv := yDeltav // 2]
		ifFalse:[xdh := xDeltah // n. ydh := yDeltah // n. 
				xdv := xDeltav // n. ydv := yDeltav // n].
	i := nPixels.
	[
		x := sx. y := sy.
		a := r := g := b := 0.
		"Pick and average n*n subpixels"
		nPix := 0.  "actual number of pixels (not clipped and not transparent)"
		j := n.
		[
			xx := x. yy := y.
			k := n.
			[
				"get a single subpixel"
				rgb := self pickWarpPixelAtX: xx y: yy.
				(combinationRule=25 "PAINT" and: [rgb = 0]) ifFalse:[
					"If not clipped and not transparent, then tally rgb values"
					nPix := nPix + 1.
					sourceDepth < 16 ifTrue:[
						"Get RGBA values from sourcemap table"
						rgb := self long32At: sourceMap + (rgb << 2).
					] ifFalse:["Already in RGB format"
						sourceDepth = 16 
								ifTrue:[rgb := self rgbMap16To32: rgb]
								ifFalse:[rgb := self rgbMap32To32: rgb]].
					b := b + (rgb bitAnd: 255).
					g := g + (rgb >> 8 bitAnd: 255).
					r := r + (rgb >> 16 bitAnd: 255).
					a := a + (rgb >> 24)].
				xx := xx + xdh.
				yy := yy + ydh.
			(k := k - 1) = 0] whileFalse.
			x := x + xdv.
			y := y + ydv.
		(j := j - 1) = 0] whileFalse.

		(nPix = 0 or: [combinationRule=25 "PAINT" and: [nPix < (n * n // 2)]]) ifTrue:[
			rgb := 0  "All pixels were 0, or most were transparent"
		] ifFalse:[
			"normalize rgba sums"
			nPix = 4 "Try to avoid divides for most common n"
				ifTrue:[r := r >> 2.	g := g >> 2.	b := b >> 2.	a := a >> 2]
				ifFalse:[	r := r // nPix.	g := g // nPix.	b := b // nPix.	a := a // nPix].
			rgb := (a << 24) + (r << 16) + (g << 8) + b.

			"map the pixel"
			rgb = 0 ifTrue: [
				"only generate zero if pixel is really transparent"
				(r + g + b + a) > 0 ifTrue: [rgb := 1]].
			rgb := self mapPixel: rgb flags: cmFlags.
		].
		"Mix it in"
		destWord := destWord bitOr: (rgb bitAnd: dstMask) << dstBitShift.
		dstBitShift := dstBitShift + dstShiftInc.
		sx := sx + xDeltah.
		sy := sy + yDeltah.
	(i := i - 1) = 0] whileFalse.

	^destWord
! !

!BitBltSimulation methodsFor: 'pixel mapping' stamp: 'ar (auto pragmas 12/08) 4/26/2001 00:58'!
warpPickSourcePixels: nPixels
	xDeltah: xDeltah yDeltah: yDeltah
	xDeltav: xDeltav yDeltav: yDeltav
	dstShiftInc: dstShiftInc
	flags: mapperFlags
	"Pick n pixels from the source form,
	map by colorMap and return aligned by dstBitShift.
	This version is only called from WarpBlt with smoothingCount = 1"
	| dstMask destWord nPix sourcePix destPix |
	<inline: true> "Yepp - this should go into warpLoop"
	dstMask := maskTable at: destDepth.
	destWord := 0.
	nPix := nPixels.
	(mapperFlags = (ColorMapPresent bitOr: ColorMapIndexedPart)) ifTrue:[
		"a little optimization for (pretty crucial) blits using indexed lookups only"
		[	"grab, colormap and mix in pixel"
			sourcePix := self pickWarpPixelAtX: sx y: sy.
			destPix := cmLookupTable at: (sourcePix bitAnd: cmMask).
			destWord := destWord bitOr: (destPix bitAnd: dstMask) << dstBitShift.
			dstBitShift := dstBitShift + dstShiftInc.
			sx := sx + xDeltah.
			sy := sy + yDeltah.
		(nPix := nPix - 1) = 0] whileFalse.
	] ifFalse:[
		[	"grab, colormap and mix in pixel"
			sourcePix := self pickWarpPixelAtX: sx y: sy.
			destPix := self mapPixel: sourcePix flags: mapperFlags.
			destWord := destWord bitOr: (destPix bitAnd: dstMask) << dstBitShift.
			dstBitShift := dstBitShift + dstShiftInc.
			sx := sx + xDeltah.
			sy := sy + yDeltah.
		(nPix := nPix - 1) = 0] whileFalse.
	].
	^destWord
! !

BitBltSimulation subclass: #BitBltSimulator
	instanceVariableNames: 'destinationHashes copyBitsCount copyBitsBreakCount'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

BitBltSimulator class
	instanceVariableNames: 'opTable maskTable warpBitShiftTable ditherMatrix4x4 ditherThresholds16 ditherValues16 dither8Lookup isInitialised'!

!BitBltSimulator commentStamp: 'eem 10/18/2018 17:17' prior: 0!
Provide bitblt support for the vm simulator.

Instance Variables
	copyBitsBreakCount:	<Integer|nil>
	copyBitsCount:			<Integer|nil>
	destinationHashes:		<OrderedCollection|nil>

copyBitsBreakCount
	- if set, and destinationHashes is non-nil, copyBits will halt at this count

copyBitsCount
	- if destinationHashes is non-nil this counts copyBits invocations

destinationHashes
	- if an empty OrderedCollection then this will collect the hash of the destination bitmap after each copyBits operation.
	  if a non-empty OrderedCollection, then the hash of the destination bitmap will be compared to the corresponding hash
	  in destinationHashes after each copyBits operation, halting if there is a mismatch.
	  if nil, then nothing is done.  In this way one can track down regressions by collecting hashes in the unregressed version
	  and comparing against results in the regressed version.!

BitBltSimulator class
	instanceVariableNames: 'opTable maskTable warpBitShiftTable ditherMatrix4x4 ditherThresholds16 ditherValues16 dither8Lookup isInitialised'!

!BitBltSimulator class methodsFor: 'accessing' stamp: 'tfel 3/11/2015 18:26'!
dither8Lookup

	^ dither8Lookup! !

!BitBltSimulator class methodsFor: 'accessing' stamp: 'tfel 3/11/2015 18:26'!
ditherMatrix4x4

	^ ditherMatrix4x4! !

!BitBltSimulator class methodsFor: 'accessing' stamp: 'tfel 3/11/2015 18:26'!
ditherThresholds16

	^ ditherThresholds16! !

!BitBltSimulator class methodsFor: 'accessing' stamp: 'tfel 3/11/2015 18:26'!
ditherValues16

	^ ditherValues16! !

!BitBltSimulator class methodsFor: 'class initialization' stamp: 'tfel 3/11/2015 18:31'!
initialize
	"self initialize"
	super initialize.
	isInitialised := false.
! !

!BitBltSimulator class methodsFor: 'accessing' stamp: 'tfel 3/11/2015 18:31'!
isInitialised

	^ isInitialised! !

!BitBltSimulator class methodsFor: 'accessing' stamp: 'tfel 3/11/2015 18:26'!
maskTable

	^ maskTable! !

!BitBltSimulator class methodsFor: 'instance creation' stamp: 'ar 5/11/2000 22:06'!
new
	^super new! !

!BitBltSimulator class methodsFor: 'accessing' stamp: 'tfel 3/11/2015 18:32'!
setInitialised

	isInitialised := true.! !

!BitBltSimulator class methodsFor: 'translation' stamp: 'tpr 5/14/2001 12:19'!
shouldBeTranslated
"This class should not be translated "
	^false! !

!BitBltSimulator class methodsFor: 'accessing' stamp: 'tfel 3/11/2015 18:26'!
warpBitShiftTable

	^ warpBitShiftTable! !

!BitBltSimulator methodsFor: 'debug support' stamp: 'eem 10/18/2018 10:16'!
copyBits
	(destinationHashes notNil and: [copyBitsCount + 1 = copyBitsBreakCount]) ifTrue:
		[self halt: 'reached copyBitsBreakCount ', copyBitsBreakCount printString].
	super copyBits.
	(interpreterProxy failed
	 or: [destinationHashes isNil]) ifFalse:
		[copyBitsCount ifNil: [copyBitsCount := 0].
		 (copyBitsCount := copyBitsCount + 1) <= destinationHashes size
			ifTrue:
				[(destinationHashes at: copyBitsCount) ~= self destinationHash ifTrue:
					[self halt: 'destination different']]
			ifFalse:
				[destinationHashes addLast: self destinationHash]]! !

!BitBltSimulator methodsFor: 'debug support' stamp: 'eem 10/18/2018 10:09'!
copyBitsBreakCount

	^ copyBitsBreakCount! !

!BitBltSimulator methodsFor: 'debug support' stamp: 'eem 10/18/2018 10:09'!
copyBitsBreakCount: anObject

	copyBitsBreakCount := anObject! !

!BitBltSimulator methodsFor: 'debug support' stamp: 'eem 10/18/2018 09:03'!
copyBitsCount

	^ copyBitsCount! !

!BitBltSimulator methodsFor: 'debug support' stamp: 'eem 10/18/2018 10:09'!
copyBitsCount: anObject

	copyBitsCount := anObject! !

!BitBltSimulator methodsFor: 'translation support' stamp: 'eem 10/16/2018 17:20'!
cppIf: conditionBlockOrSymbolValue ifTrue: trueExpressionOrBlock ifFalse: falseExpressionOrBlockOrNil
	"The simulator does not have fast blt defines"
	^ conditionBlockOrSymbolValue == #'ENABLE_FAST_BLT'
		ifTrue: falseExpressionOrBlockOrNil
		ifFalse: [super
					cppIf: conditionBlockOrSymbolValue
					ifTrue: trueExpressionOrBlock
					ifFalse: falseExpressionOrBlockOrNil]! !

!BitBltSimulator methodsFor: 'debug support' stamp: 'eem 10/18/2018 09:21'!
destinationHash
	| mem bits |
	mem := interpreterProxy memory.
	bits := mem copyFrom: destBits // mem bytesPerElement + 1 to: endOfDestination // mem bytesPerElement.
	ByteArray adoptInstance: bits.
	"This must be (and is) true; but the copyFrom:to: short cut is 100x faster than the byteAt: version"
	"self assert: (((destBits to: endOfDestination - 1) collect: [:i| interpreterProxy byteAt: i]) as: ByteArray) = bits"
	^bits hash! !

!BitBltSimulator methodsFor: 'debug support' stamp: 'eem 10/18/2018 09:03'!
destinationHashes

	^ destinationHashes! !

!BitBltSimulator methodsFor: 'debug support' stamp: 'eem 10/18/2018 10:11'!
destinationHashes: anObject

	destinationHashes := anObject.
	copyBitsCount ifNil: [copyBitsCount := 0].
	copyBitsBreakCount ifNil: [copyBitsBreakCount := -1]! !

!BitBltSimulator methodsFor: 'debug support' stamp: 'eem 10/15/2018 17:00'!
dstLongAt: dstIndex
	interpreterProxy isInterpreterProxy ifTrue:
		[^dstIndex long32At: 0].
	^super dstLongAt: dstIndex! !

!BitBltSimulator methodsFor: 'debug support' stamp: 'eem 10/15/2018 16:44'!
dstLongAt: dstIndex put: value
	interpreterProxy isInterpreterProxy ifTrue:
		[^dstIndex long32At: 0 put: value].
	^super dstLongAt: dstIndex put: value! !

!BitBltSimulator methodsFor: 'memory access' stamp: 'eem 10/15/2018 17:01'!
halftoneAt: idx
	interpreterProxy isInterpreterProxy ifTrue:
		[^(halftoneBase + (idx \\ halftoneHeight * 4)) long32At: 0].
	^super halftoneAt: idx! !

!BitBltSimulator methodsFor: 'memory access' stamp: 'tfrel 3/9/2016 23:01'!
incDestIndex: offset
	"if offset is a CObject, this avoids a copy, making the simulation massively faster"
	^ destIndex isCObjectAccessor
		ifTrue: [destIndex += offset]
		ifFalse: [super incDestIndex: offset]
! !

!BitBltSimulator methodsFor: 'memory access' stamp: 'tfrel 3/9/2016 23:01'!
incSrcIndex: offset
	"if offset is a CObject, this avoids a copy, making the simulation massively faster"
	^ sourceIndex isCObjectAccessor
		ifTrue: [sourceIndex += offset]
		ifFalse: [super incSrcIndex: offset]
! !

!BitBltSimulator methodsFor: 'simulation' stamp: 'tpr 4/3/2004 23:16'!
initBBOpTable
	opTable := OpTable.
	maskTable := Array new: 32.
	#(1 2 4 5 8 16 32) do:[:i| maskTable at: i put: (1 << i)-1].
	self initializeDitherTables.
	warpBitShiftTable := CArrayAccessor on: (Array new: 32).! !

!BitBltSimulator methodsFor: 'initialize-release' stamp: 'eem 3/31/2015 18:07'!
initialiseModule

	self class isInitialised ifFalse: [| ivars |
		ivars := #(opTable maskTable warpBitShiftTable ditherMatrix4x4 ditherThresholds16 ditherValues16 dither8Lookup).
		super initialiseModule.
		ivars do: [:symbol | self class instVarNamed: symbol put: (self instVarNamed: symbol)].
		self class setInitialised].
	opTable := self class opTable.
	maskTable := self class maskTable.
	warpBitShiftTable := self class warpBitShiftTable.
	ditherMatrix4x4 := self class ditherMatrix4x4.
	ditherThresholds16 := self class ditherThresholds16.
	ditherValues16 := self class ditherValues16.
	dither8Lookup := self class dither8Lookup.
	^true! !

!BitBltSimulator methodsFor: 'simulation' stamp: 'tfel 3/10/2015 23:14'!
initializeDitherTables
	ditherMatrix4x4 := CArrayAccessor on:
		#(	0	8	2	10
			12	4	14	6
			3	11	1	9
			15	7	13	5).
	ditherThresholds16 := CArrayAccessor on:#(0 2 4 6 8 10 12 14 16).
	ditherValues16 := CArrayAccessor on: 
		#(0 0 1 2 3 4 5 6 7 8 9 10 11 12 13 14
		15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30).
	dither8Lookup := CArrayAccessor on: (Array new: 4096).! !

!BitBltSimulator methodsFor: 'memory access' stamp: 'eem 7/1/2008 17:29'!
long32At: byteAddress
	^interpreterProxy long32At: byteAddress! !

!BitBltSimulator methodsFor: 'memory access' stamp: 'eem 7/1/2008 17:28'!
long32At: byteAddress put: a32BitValue
	^interpreterProxy long32At: byteAddress put: a32BitValue! !

!BitBltSimulator methodsFor: 'simulation' stamp: 'di 12/30/97 11:07'!
mergeFn: arg1 with: arg2
	^ self perform: (opTable at: combinationRule+1) with: arg1 with: arg2! !

!BitBltSimulator methodsFor: 'simulation' stamp: 'eem 10/26/2018 10:02'!
primitive: primitiveName parameters: parameterTypesArray receiver: rcvrType
	"This exists just to check the set of methods that use smart syntax for which
	 marshallers have been written.  If a case cannot be found you'll have to write
	 a marshaller for the case in question that follows the approach taken in the
	 marshallers below."
	simulator forMap ifTrue:
		[Notification new
			tag: {#forMap. primitiveName. parameterTypesArray. rcvrType};
			signal].
	primitiveName caseOf: {
		['primitivePixelValueAt']		-> [^self].
		['primitiveCompareColors']	-> [^self] }! !

!BitBltSimulator methodsFor: 'primitives' stamp: 'eem 10/26/2018 09:57'!
primitiveCompareColors
	"This is a hack to mimic the SmartSyntaxPlugin support for primitive:parameters:receiver:
	 which is hacked here at BitBltSimulator>>primitive:parameters:receiver:"
	"rcvr := self
				primitive: 'primitiveCompareColors'
				parameters: #(#Unsigned #Unsigned #SmallInteger )
				receiver: #Oop."
	| unsignedBlock colorA colorB testID |
	unsignedBlock := simulator ccgLoad: nil expr: nil asUnsignedValueFrom: nil.
	colorA := unsignedBlock value: (interpreterProxy stackValue: 2).
	colorB := unsignedBlock value: (interpreterProxy stackValue: 1).
	testID := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifFalse:
		[self primitiveCompareColorA: colorA to: colorB test: testID]! !

!BitBltSimulator methodsFor: 'primitives' stamp: 'eem 12/10/2015 11:20'!
primitiveCopyBits
	"Override to provide a convenience to investigate primitive failures.
	 Comment out the first statement to be able to loop, examining e.g. why a BitBlt fails to load."
	true ifTrue:
		[^super primitiveCopyBits].
	[super primitiveCopyBits.
	 interpreterProxy failed] whileTrue:
		[self halt. interpreterProxy coInterpreter initPrimCall]! !

!BitBltSimulator methodsFor: 'primitives' stamp: 'eem 10/18/2018 13:14'!
primitivePixelValueAt
	"This is a hack to mimic the SmartSyntaxPlugin support for primitive:parameters:receiver:
	 which is hacked here at BitBltSimulator>>primitive:parameters:receiver:"
	| xVal yVal |
	xVal := interpreterProxy stackIntegerValue: 1.
	yVal := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifFalse:
		[self primitivePixelValueAtX: xVal y: yVal]! !

!BitBltSimulator methodsFor: 'debug support' stamp: 'eem 10/15/2018 16:08'!
srcLongAt: srcIndex
	interpreterProxy isInterpreterProxy ifTrue:
		[^srcIndex long32At: 0].
	^super srcLongAt: srcIndex! !

SmartSyntaxInterpreterPlugin subclass: #ClipboardExtendedPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!ClipboardExtendedPlugin class methodsFor: 'simulation' stamp: 'eem 10/29/2018 15:17'!
simulatorClass
	"This should be easy to simulate, but I'm short of time.  Feel free to have a go, following e.g. the pattern in JPEGReadWriter2Plugin."
	^nil! !

!ClipboardExtendedPlugin methodsFor: 'io' stamp: 'nice 7/8/2016 02:12'!
ioAddClipboardData: clipboard data: data dataFormat: aFormat
	| clipboardAddress formatLength dataLength |
	<var: #clipboardAddress type: #'usqIntptr_t'>
	self primitive: 'ioAddClipboardData' parameters: #(Oop ByteArray String).

	clipboardAddress := interpreterProxy positiveMachineIntegerValueOf: clipboard.
	dataLength := interpreterProxy slotSizeOf: data cPtrAsOop.
	formatLength := interpreterProxy slotSizeOf: aFormat cPtrAsOop.

	self sqPasteboardPutItemFlavor: clipboardAddress data: data length: dataLength formatType: aFormat formatLength: formatLength.
! !

!ClipboardExtendedPlugin methodsFor: 'io' stamp: 'nice 7/8/2016 02:12'!
ioClearClipboard: clipboard
	| clipboardAddress |
	<var: #clipboardAddress type: #'usqIntptr_t'>
	self primitive: 'ioClearClipboard' parameters: #(Oop).
	clipboardAddress :=  interpreterProxy positiveMachineIntegerValueOf: clipboard.
	self sqPasteboardClear: clipboardAddress.! !

!ClipboardExtendedPlugin methodsFor: 'io' stamp: 'nice 4/2/2015 22:19'!
ioCreateClipboard
	| clipboardAddress |
	self primitive: 'ioCreateClipboard' parameters: #().

	clipboardAddress := interpreterProxy positiveMachineIntegerFor: self sqCreateClipboard.

	^ clipboardAddress.! !

!ClipboardExtendedPlugin methodsFor: 'io' stamp: 'nice 7/8/2016 02:12'!
ioGetClipboardFormat: clipboard formatNumber: formatNumber 
	| clipboardAddress itemCount |
	<var: #clipboardAddress type: #'usqIntptr_t'>
	self primitive: 'ioGetClipboardFormat' parameters: #(#Oop #SmallInteger ).
	clipboardAddress := interpreterProxy positiveMachineIntegerValueOf: clipboard.
	itemCount := self sqPasteboardGetItemCount: clipboardAddress.
	itemCount > 0
		ifTrue: [^ self sqPasteboardCopyItemFlavors: clipboardAddress itemNumber: formatNumber].
	^ interpreterProxy nilObject! !

!ClipboardExtendedPlugin methodsFor: 'io' stamp: 'nice 7/8/2016 02:12'!
ioReadClipboardData: clipboard format: format
	| clipboardAddress formatLength |
	<var: #clipboardAddress type: #'usqIntptr_t'>
	self primitive: 'ioReadClipboardData' parameters: #(Oop String).
	clipboardAddress := interpreterProxy positiveMachineIntegerValueOf: clipboard.
	formatLength := interpreterProxy slotSizeOf: format cPtrAsOop.
	^ self sqPasteboardCopyItemFlavorData: clipboardAddress format: format formatLength: formatLength.
! !

SmartSyntaxInterpreterPlugin subclass: #FileCopyPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!FileCopyPlugin commentStamp: 'tpr 5/2/2003 15:48' prior: 0!
This plugin is a simple workaround for the lamentable state of the Squeak file handling system; it provides a primitive to copy a file or tree via the OS facility and thus preserve all the OS attributes in the most trivial possible manner. Not intended for a long life and should be replaced by better code as soon as possible. The key benefit it offers is maintenance of any OS flags, tags, bits and bobs belonging to the file.  Since it requires platform support it will only be built when supported on your platform!

!FileCopyPlugin class methodsFor: 'translation' stamp: 'mt 7/27/2021 12:51'!
hasHeaderFile

	^ true! !

!FileCopyPlugin class methodsFor: 'translation' stamp: 'tpr 4/27/2001 11:02'!
requiresPlatformFiles
	"this plugin requires platform specific files in order to work"
	^true! !

!FileCopyPlugin methodsFor: 'system primitives' stamp: 'tpr 4/10/2002 19:34'!
primitiveFile: srcName copyTo: dstName

	|srcSz dstSz ok |
	self primitive: 'primitiveFileCopyNamedTo'
		parameters: #(String String).

	srcSz := interpreterProxy slotSizeOf: srcName cPtrAsOop.
	dstSz := interpreterProxy slotSizeOf: dstName cPtrAsOop.
	ok := self sqCopyFile: srcName size: srcSz to: dstName size: dstSz.
	ok ifFalse:[interpreterProxy primitiveFail].
! !

SmartSyntaxInterpreterPlugin subclass: #GeniePlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!GeniePlugin commentStamp: '<historical>' prior: 0!
This plugin implements the functionality of
	CRStrokeFeature>>sameClassAbsoluteStrokeDistance: aCRFeature forReference: aBoolean
. This means that changes there should be mirrored here!!

GeniePlugin>>majorNO should be in sync with version number of Genie.
!

!GeniePlugin class methodsFor: 'check installed plugin' stamp: 'sr 6/25/2001 20:49'!
majorNO
	| no |
	^ (no := self versionNO) ifNotNil: [no // 1000]
! !

!GeniePlugin class methodsFor: 'check installed plugin' stamp: 'sr 6/25/2001 20:50'!
minorNO
	| no |
	^ (no := self versionNO) ifNotNil: [no \\ 1000]
! !

!GeniePlugin class methodsFor: 'translation' stamp: 'sr 4/15/2001 17:44'!
moduleNameAndVersion
	"Answer the receiver's module name and version info that is used for the plugin's C code. The default is to append the code generation date, but any useful text is ok (keep it short)"

	^ self moduleName, Character space asString, self version, Character space asString, Date today asString! !

!GeniePlugin class methodsFor: 'translation' stamp: 'sr 6/25/2001 20:42'!
version
	"Answer the receiver's version info as String."
	"Somewhat a hack, but calling class methods from inst methods doesn't result in usable C-code..."

	| inst |
	inst := self new.
	^ 'v', inst majorNO asString, '.', inst minorNO asString! !

!GeniePlugin class methodsFor: 'check installed plugin' stamp: 'sr 6/25/2001 20:46'!
versionNO
	<primitive: 'primVersionNO' module: 'GeniePlugin'>
	^ nil
! !

!GeniePlugin class methodsFor: 'check installed plugin' stamp: 'NS 8/8/2001 15:31'!
versionString
	^ 'v', (self versionNO / 1000 asFloat) asString! !

!GeniePlugin methodsFor: 'computation' stamp: 'eem 9/9/2011 16:04'!
cSquaredDistanceFrom: aPoint to: bPoint
	"arguments are pointer to ints paired as x,y coordinates of points"
	| aPointX aPointY bPointX bPointY xDiff yDiff |
	<var: #aPoint type: #'int *'>
	<var: #bPoint type: #'int *'>
	aPointX := aPoint at: 0.
	aPointY := aPoint at: 1.
	bPointX := bPoint at: 0.
	bPointY := bPoint at: 1.

	xDiff := bPointX - aPointX.
	yDiff := bPointY - aPointY.
	^ xDiff * xDiff + (yDiff * yDiff)! !

!GeniePlugin methodsFor: 'computation' stamp: 'sr 6/5/2001 06:53'!
cSubstAngleFactorFrom: startDegreeNumber to: endDegreeNumber 
	| absDiff |
	absDiff := (endDegreeNumber - startDegreeNumber) abs.
	absDiff > 180 ifTrue: [absDiff := 360 - absDiff].
	^ absDiff * absDiff bitShift: -6! !

!GeniePlugin methodsFor: 'version' stamp: 'NS 7/12/2001 11:54'!
majorNO
	^ 2! !

!GeniePlugin methodsFor: 'version' stamp: 'NS 7/12/2001 11:54'!
minorNO
	^ 0! !

!GeniePlugin methodsFor: 'computation' stamp: 'eem 9/9/2011 16:05'!
primSameClassAbsoluteStrokeDistanceMyPoints: myPointsOop otherPoints: otherPointsOop myVectors: myVectorsOop otherVectors: otherVectorsOop mySquaredLengths: mySquaredLengthsOop otherSquaredLengths: otherSquaredLengthsOop myAngles: myAnglesOop otherAngles: otherAnglesOop maxSizeAndReferenceFlag: maxSizeAndRefFlag rowBase: rowBaseOop rowInsertRemove: rowInsertRemoveOop rowInsertRemoveCount: rowInsertRemoveCountOop
	| base insertRemove jLimiT substBase insert remove subst removeBase insertBase insertRemoveCount additionalMultiInsertRemoveCost myPoints otherPoints myVectors otherVectors rowInsertRemoveCount mySquaredLengths otherSquaredLengths myAngles otherAngles rowBase rowInsertRemove otherPointsSize myVectorsSize otherVectorsSize otherSquaredLengthsSize rowBaseSize maxDist maxSize forReference jM1 iM1 iM1T2 jM1T2 |
	<var: #myPoints type: #'int *'>
	<var: #otherPoints type: #'int *'>
	<var: #myVectors type: #'int *'>
	<var: #otherVectors type: #'int *'>
	<var: #mySquaredLengths type: #'int *'>
	<var: #otherSquaredLengths type: #'int *'>
	<var: #myAngles type: #'int *'>
	<var: #otherAngles type: #'int *'>
	<var: #rowBase type: #'int *'>
	<var: #rowInsertRemove type: #'int *'>
	<var: #rowInsertRemoveCount type: #'int *'>
	self
		primitive: 'primSameClassAbsoluteStrokeDistanceMyPoints_otherPoints_myVectors_otherVectors_mySquaredLengths_otherSquaredLengths_myAngles_otherAngles_maxSizeAndReferenceFlag_rowBase_rowInsertRemove_rowInsertRemoveCount'
		parameters: #(#Oop #Oop #Oop #Oop #Oop #Oop #Oop #Oop #SmallInteger #Oop #Oop #Oop)
		receiver: #Oop.
	interpreterProxy failed
		ifTrue: [self msg: 'failed 1'.
			^ nil].

	interpreterProxy success: (interpreterProxy isWords: myPointsOop)
			& (interpreterProxy isWords: otherPointsOop)
			& (interpreterProxy isWords: myVectorsOop)
			& (interpreterProxy isWords: otherVectorsOop)
			& (interpreterProxy isWords: mySquaredLengthsOop)
			& (interpreterProxy isWords: otherSquaredLengthsOop)
			& (interpreterProxy isWords: myAnglesOop)
			& (interpreterProxy isWords: otherAnglesOop)
			& (interpreterProxy isWords: rowBaseOop)
			& (interpreterProxy isWords: rowInsertRemoveOop)
			& (interpreterProxy isWords: rowInsertRemoveCountOop).
	interpreterProxy failed
		ifTrue: [self msg: 'failed 2'.
			^ nil].
	interpreterProxy success: (interpreterProxy is: myPointsOop MemberOf: 'PointArray')
			& (interpreterProxy is: otherPointsOop MemberOf: 'PointArray').
	interpreterProxy failed
		ifTrue: [self msg: 'failed 3'.
			^ nil].
	myPoints := interpreterProxy firstIndexableField: myPointsOop.
	otherPoints := interpreterProxy firstIndexableField: otherPointsOop.
	myVectors := interpreterProxy firstIndexableField: myVectorsOop.
	otherVectors := interpreterProxy firstIndexableField: otherVectorsOop.
	mySquaredLengths := interpreterProxy firstIndexableField: mySquaredLengthsOop.
	otherSquaredLengths := interpreterProxy firstIndexableField: otherSquaredLengthsOop.
	myAngles := interpreterProxy firstIndexableField: myAnglesOop.
	otherAngles := interpreterProxy firstIndexableField: otherAnglesOop.
	rowBase := interpreterProxy firstIndexableField: rowBaseOop.
	rowInsertRemove := interpreterProxy firstIndexableField: rowInsertRemoveOop.
	rowInsertRemoveCount := interpreterProxy firstIndexableField: rowInsertRemoveCountOop.
	"Note: myPointsSize and mySquaredLengthsSize variables eliminated to reduce
	method temporary variable count for closure-enabled images"
	"PointArrays"
	"myPointsSize := (interpreterProxy stSizeOf: myPointsOop) bitShift: -1."
	otherPointsSize := (interpreterProxy stSizeOf: otherPointsOop) bitShift: -1.
	myVectorsSize := (interpreterProxy stSizeOf: myVectorsOop) bitShift: -1.
	otherVectorsSize := (interpreterProxy stSizeOf: otherVectorsOop) bitShift: -1.
	"IntegerArrays"
	"mySquaredLengthsSize := interpreterProxy stSizeOf: mySquaredLengthsOop."
	otherSquaredLengthsSize := interpreterProxy stSizeOf: otherSquaredLengthsOop.
	rowBaseSize := interpreterProxy stSizeOf: rowBaseOop.

	interpreterProxy success: rowBaseSize
			= (interpreterProxy stSizeOf: rowInsertRemoveOop) & (rowBaseSize
				= (interpreterProxy stSizeOf: rowInsertRemoveCountOop)) & (rowBaseSize > otherVectorsSize).
	interpreterProxy failed
		ifTrue: [self msg: 'failed 4'.
			^ nil].
	interpreterProxy success: (interpreterProxy stSizeOf: mySquaredLengthsOop) >= (myVectorsSize - 1)
				& (((interpreterProxy stSizeOf: myPointsOop) bitShift: -1) >= myVectorsSize)
				& (otherSquaredLengthsSize >= (otherVectorsSize - 1))
				& (otherPointsSize >= otherVectorsSize) & ((interpreterProxy stSizeOf: myAnglesOop)
				>= (myVectorsSize - 1)) & ((interpreterProxy stSizeOf: otherAnglesOop)
				>= (otherVectorsSize - 1)).
	interpreterProxy failed
		ifTrue: [self msg: 'failed 5'.
			^ nil].

	"maxSizeAndRefFlag contains the maxium feature size (pixel) and also indicates whether
	the reference flag (boolean) is set. Therefore the maximum size is moved to the left 
	and the reference flag is stored in the LSB.
	Note: This is necessary to avoid more than 12 primitive parameters"
	forReference := maxSizeAndRefFlag bitAnd: 1.
	maxSize := maxSizeAndRefFlag bitShift: -1.
	maxDist := 1 bitShift: 29.
	forReference
		ifTrue: [additionalMultiInsertRemoveCost := 0]
		ifFalse: [additionalMultiInsertRemoveCost := maxSize * maxSize bitShift: -10].
	"C indices!!!!"
	rowBase
		at: 0
		put: 0.
	rowInsertRemove
		at: 0
		put: 0.
	rowInsertRemoveCount
		at: 0
		put: 2.
	insertRemove := 0 - additionalMultiInsertRemoveCost.
	jLimiT := otherVectorsSize.
	otherPointsSize >= (jLimiT - 1) & (otherSquaredLengthsSize >= (jLimiT - 1))
		ifFalse: [^ interpreterProxy primitiveFail].
	1
		to: jLimiT
		do: [:j |
			jM1 := j - 1.
			insertRemove := insertRemove + ((otherSquaredLengths at: jM1)
							+ (self
									cSquaredDistanceFrom: (otherPoints + (jM1 bitShift: 1))
									to: myPoints) bitShift: -7) + additionalMultiInsertRemoveCost.
			rowInsertRemove
				at: j
				put: insertRemove.
			rowBase
				at: j
				put: insertRemove * j.
			rowInsertRemoveCount
				at: j
				put: j + 1].
	insertRemove := (rowInsertRemove at: 0)
				- additionalMultiInsertRemoveCost.
	1
		to: myVectorsSize
		do: [:i |
			iM1 := i - 1.
			iM1T2 := iM1 bitShift: 1.
			substBase := rowBase at: 0.
			insertRemove := insertRemove + ((mySquaredLengths at: iM1)
							+ (self
									cSquaredDistanceFrom: (myPoints + iM1T2)
									to: otherPoints) bitShift: -7) + additionalMultiInsertRemoveCost.
			rowInsertRemove
				at: 0
				put: insertRemove.
			rowBase
				at: 0
				put: insertRemove * i.
			rowInsertRemoveCount
				at: 0
				put: i + 1.
			jLimiT := otherVectorsSize.
			1
				to: jLimiT
				do: [:j |
					jM1 := j - 1.
					jM1T2 := jM1 bitShift: 1.
					removeBase := rowBase at: j.
					insertBase := rowBase at: jM1.
					remove := (mySquaredLengths at: iM1)
								+ (self
										cSquaredDistanceFrom: (myPoints + iM1T2)
										to: (otherPoints + (j bitShift: 1))) bitShift: -7.
					(insertRemove := rowInsertRemove at: j) = 0
						ifTrue: [removeBase := removeBase + remove]
						ifFalse: [removeBase := removeBase + insertRemove + (remove
											* (rowInsertRemoveCount at: j)).
							remove := remove + insertRemove].
					insert := (otherSquaredLengths at: jM1)
								+ (self
										cSquaredDistanceFrom: (otherPoints + jM1T2)
										to: (myPoints + (i bitShift: 1))) bitShift: -7.
					(insertRemove := rowInsertRemove at: jM1) = 0
						ifTrue: [insertBase := insertBase + insert]
						ifFalse: [insertBase := insertBase + insertRemove + (insert
											* (rowInsertRemoveCount at: jM1)).
							insert := insert + insertRemove].
					forReference
						ifTrue: [substBase := maxDist]
						ifFalse: [subst := (self
										cSquaredDistanceFrom: (otherVectors + jM1T2)
										to: (myVectors + iM1T2))
										+ (self
												cSquaredDistanceFrom: (otherPoints + jM1T2)
												to: (myPoints + iM1T2)) * (16
											+ (self
													cSubstAngleFactorFrom: (otherAngles at: jM1)
													to: (myAngles at: iM1))) bitShift: -11.
							substBase := substBase + subst].
					(substBase <= removeBase
							and: [substBase <= insertBase])
						ifTrue: [base := substBase.
							insertRemove := 0.
							insertRemoveCount := 1]
						ifFalse: [removeBase <= insertBase
								ifTrue: [base := removeBase.
									insertRemove := remove + additionalMultiInsertRemoveCost.
									insertRemoveCount := (rowInsertRemoveCount at: j)
												+ 1]
								ifFalse: [base := insertBase.
									insertRemove := insert + additionalMultiInsertRemoveCost.
									insertRemoveCount := (rowInsertRemoveCount at: jM1)
												+ 1]].
					substBase := rowBase at: j.
					rowBase
						at: j
						put: (base min: maxDist).
					rowInsertRemove
						at: j
						put: (insertRemove min: maxDist).
					rowInsertRemoveCount
						at: j
						put: insertRemoveCount].
			insertRemove := rowInsertRemove at: 0].
	^ base asOop: SmallInteger
! !

!GeniePlugin methodsFor: 'version' stamp: 'sr 6/25/2001 20:39'!
primVersionNO
	"majorNO * 1000 + minorNO"
	self
		primitive: 'primVersionNO'
		parameters: #()
		receiver: #Oop.
	^ (self majorNO * 1000 + self minorNO) asOop: SmallInteger! !

SmartSyntaxInterpreterPlugin subclass: #HostWindowPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!HostWindowPlugin commentStamp: 'tpr 10/14/2004 15:57' prior: 0!
This plugin provides access to creating, destroying and manipulating host OS windows. See the Cross/plugins/HostWindowPlugin/HostWindowPlugin.h file for documented declarations for the C functions you need to provide.!

!HostWindowPlugin class methodsFor: 'translation' stamp: 'JMM 8/17/2004 20:24'!
hasHeaderFile
	^true! !

!HostWindowPlugin class methodsFor: 'translation' stamp: 'tpr 7/20/2004 13:09'!
requiresPlatformFiles
	^true! !

!HostWindowPlugin methodsFor: 'support' stamp: 'eem 10/13/2020 08:45'!
pointFromCompactPointEncoding: encodedPoint
	"Answer a point from one of the funky encoded x,y position/size values the VM uses.
	 The issue here is that the values can be negative, each value being in the range -32768 to 32767"
	<inline: #always>
	^interpreterProxy
		makePointwithxValue: (self cCoerceSimple: encodedPoint >> 16 to: #short)
		yValue: (self cCoerce: (encodedPoint bitAnd: 16rFFFF) to: #short)! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'tpr 9/20/2004 12:32'!
primitiveCloseHostWindow: windowIndex
"Close a host window. windowIndex is the SmallInt handle returned previously by primitiveCreateHostWindow. Fail if the index is invalid or the platform code fails"
	| ok |
	self primitive: 'primitiveCloseHostWindow'
		parameters: #(SmallInteger).
	
	ok := self closeWindow: windowIndex.
	ok ifFalse:[interpreterProxy primitiveFail].
! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'tpr 10/3/2004 10:34'!
primitiveCreateHostWindowWidth: w height: h originX: x y: y attributes: list
"Create a host window of width 'w' pixels, height 'h' with the origin of the
user area at 'x@y' from the topleft corner of the screen.
Return the SmallInt value of the internal index to the window description block
- which is whatever the host platform code needs it to be."
	| windowIndex listLength |
	self primitive: 'primitiveCreateHostWindow'
		parameters: #(SmallInteger SmallInteger SmallInteger SmallInteger ByteArray).

	"createWindowWidthheightoriginXyattr(int w, int h, int x, int y, int*
attributeList) must create a hostwindow and return an integer index. Return 0 if
failed"
	listLength := interpreterProxy slotSizeOf: list cPtrAsOop.
	windowIndex := self createWindowWidth: w height: h originX: x y: y attr: list
length: listLength.
	windowIndex > 0 ifTrue:[^windowIndex asSmallIntegerObj]
		ifFalse:[^interpreterProxy primitiveFail].
! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 9/19/2018 14:42'!
primitiveHostWindowIcon: id path: pathString
	"Set the icon of the window by fetching the logo in given path. Fail if the windowIndex is invalid or the platform routine returns -1 to indicate failure"
	| res pathLength |
	self primitive: 'primitiveHostWindowIcon'
		parameters: #(SmallInteger ByteArray).
	pathLength := interpreterProxy slotSizeOf: pathString cPtrAsOop.
	res := self ioSetIconOfWindow: id _: pathString _: pathLength.
	res = 0 ifTrue: [^ self ].
	res = -1 ifTrue: [^ interpreterProxy primitiveFailFor: PrimErrBadArgument ].
	"As the GetLastError function can return 0, but it is still an error, 1 is added to all the GetLastError error codes."
	^ interpreterProxy primitiveFail! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 11/19/2008 17:29'!
primitiveHostWindowPosition: windowIndex 
	"Return the origin position of the user area of the window in pixels from the topleft corner of the screen. Fail if the windowIndex is invalid or the platform routine returns -1 to indicate failure"
	| pos |
	self primitive: 'primitiveHostWindowPosition'
		parameters: #(SmallInteger ).
	pos := self ioPositionOfWindow: windowIndex.
	pos = -1 ifTrue:
		[^interpreterProxy primitiveFail].
	^self pointFromCompactPointEncoding: pos! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 11/19/2008 17:29'!
primitiveHostWindowPositionSet: windowIndex x: x y: y
	"Set the origin position of the user area of the window in pixels from the topleft corner of the screen- return the position actually set by the OS/GUI/window manager. Fail if the windowIndex is invalid or the platform routine returns -1 to indicate failure"
	| pos |
	self primitive: 'primitiveHostWindowPositionSet'
		parameters: #(SmallInteger SmallInteger SmallInteger).
	pos := self ioPositionOfWindowSet: windowIndex x: x y: y.
	pos = -1 ifTrue:
		[^interpreterProxy primitiveFail].
	^self pointFromCompactPointEncoding: pos! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 07/21/2017 11:16'!
primitiveHostWindowScreenWorkAreaPosition: windex
	<option: #TerfVM>
	"Answer the platform's notion of the available work area on the screen containing the given window."
	| size |
	self primitive: 'primitiveHostWindowScreenWorkAreaPosition'
		parameters: #(SmallInteger).
	size := self ioPositionOfScreenWorkArea: windex.
	size = -1 ifTrue:
		[^interpreterProxy primitiveFail].
	^self pointFromCompactPointEncoding: size! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 07/21/2017 11:16'!
primitiveHostWindowScreenWorkAreaSize: windex
	<option: #TerfVM>
	"Answer the platform's notion of the available work area on the screen containing the given window."
	| size |
	self primitive: 'primitiveHostWindowScreenWorkAreaSize'
		parameters: #(SmallInteger).
	size := self ioSizeOfScreenWorkArea: windex.
	size = -1 ifTrue:
		[^interpreterProxy primitiveFail].
	^self pointFromCompactPointEncoding: size! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 11/19/2008 17:30'!
primitiveHostWindowSize: windowIndex 
	"Return the size of the user area of the window in pixels. Fail if the windowIndex is invalid or the platform routine returns -1 to indicate failure"
	| size |
	self primitive: 'primitiveHostWindowSize'
		parameters: #(SmallInteger ).
	size := self ioSizeOfWindow: windowIndex.
	size = -1 ifTrue:
		[^interpreterProxy primitiveFail].
	^self pointFromCompactPointEncoding: size! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 11/19/2008 17:29'!
primitiveHostWindowSizeSet: windowIndex x: x y: y
	"Set the size of the user area of the window in pixels - return what is actually set by the OS/GUI/window manager. Fail if the windowIndex is invalid or the platform routine returns -1 to indicate failure"
	| size |
	self primitive: 'primitiveHostWindowSizeSet'
		parameters: #(SmallInteger SmallInteger SmallInteger).
	size := self ioSizeOfWindowSet: windowIndex x: x y: y.
	size = -1 ifTrue:
		[^interpreterProxy primitiveFail].
	^self pointFromCompactPointEncoding: size! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 9/19/2018 14:43'!
primitiveHostWindowTitle: id string: titleString
	"Set the title bar label of the window. Fail if the windowIndex is invalid or the platform routine returns -1 to indicate failure"
	| titleLength |
	self primitive: 'primitiveHostWindowTitle'
		parameters: #(SmallInteger String).
	titleLength := interpreterProxy slotSizeOf: titleString cPtrAsOop.
	(self ioSetTitleOfWindow: id _: titleString _: titleLength) = -1 ifTrue:
		[interpreterProxy primitiveFail]! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 07/21/2017 11:16'!
primitiveNativeDisplayPosition: windowHandle
	<option: #TerfVM>
	"Answer the origin position of the user area of the given native window in pixels from the topleft corner of the screen.
	 Fail if the windowHandle is invalid or the platform routine returns -1 to indicate failure"
	| pos |
	self primitive: 'primitiveNativeDisplayPosition'
		parameters: #(Unsigned).
	pos := self ioPositionOfNativeDisplay: windowHandle.
	pos = -1 ifTrue:
		[^interpreterProxy primitiveFail].
	^self pointFromCompactPointEncoding: pos! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 07/21/2017 11:16'!
primitiveNativeDisplaySize: windowHandle
	<option: #TerfVM>
	"Answer the size of the user area of the given native window in pixels.
	 Fail if the windowHandle is invalid or the platform routine returns -1 to indicate failure"
	| size |
	self primitive: 'primitiveNativeDisplaySize'
		parameters: #(Unsigned).
	size := self ioSizeOfNativeDisplay: windowHandle.
	size = -1 ifTrue:
		[^interpreterProxy primitiveFail].
	^self pointFromCompactPointEncoding: size! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 07/21/2017 11:16'!
primitiveNativeWindowPosition: windowHandle
	<option: #TerfVM>
	"Answer the origin position of the given native window in pixels from the topleft corner of the screen.
	 Fail if the windowHandle is invalid or the platform routine returns -1 to indicate failure"
	| pos |
	self primitive: 'primitiveNativeWindowPosition'
		parameters: #(Unsigned).
	pos := self ioPositionOfNativeWindow: windowHandle.
	pos = -1 ifTrue:
		[^interpreterProxy primitiveFail].
	^self pointFromCompactPointEncoding: pos! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 07/21/2017 11:16'!
primitiveNativeWindowSize: windowHandle
	<option: #TerfVM>
	"Answer the size of the given native window in pixels.
	 Fail if the windowHandle is invalid or the platform routine returns -1 to indicate failure"
	| size |
	self primitive: 'primitiveNativeWindowSize'
		parameters: #(Unsigned).
	size := self ioSizeOfNativeWindow: windowHandle.
	size = -1 ifTrue:
		[^interpreterProxy primitiveFail].
	^self pointFromCompactPointEncoding: size! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 1/22/2021 11:15'!
primitiveScreenRectangles
	<option: #TerfVM>
	"Answer an Array of pairs of packed x,y repreeentation origins and extents for all the available monitors."
	self primitive: 'primitiveScreenRectangles'.
	self ioScreenRectangles
		ifNil: [interpreterProxy primitiveFailFor: PrimErrNoMemory]
		ifNotNil: [:screenRectangles| interpreterProxy methodReturnValue: screenRectangles]! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 10/4/2020 18:39'!
primitiveSetCursorPositionX: x Y: y
	"Set the position of the cursor to the specified position on the desktop (*not* the display).
	 Fail if the platform routine returns -1 to indicate failure."
	| result |
	self primitive: 'primitiveSetCursorPosition'
		parameters: #(SmallInteger SmallInteger).
	result := self ioSetCursorPositionX: x Y: y.
	result = -1 ifTrue:
		[^interpreterProxy primitiveFail]! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 9/19/2018 14:43'!
primitiveShowHostWindow: windowIndex bits: dispBits width: w height: h depth: d
left: left right: right top: top bottom: bottom
"Host window analogue of DisplayScreen> primShowRectLeft:right:top:bottom:
(Interpreter>primitiveShowDisplayRect) which takes the window index, bitmap
details and the rectangle bounds. Fail if the windowIndex is invalid or the
platform routine returns false to indicate failure"
	self primitive: 'primitiveShowHostWindowRect'
		parameters: #(SmallInteger WordArray SmallInteger SmallInteger SmallInteger
SmallInteger SmallInteger SmallInteger SmallInteger).

	"Tell the vm to copy pixel's from dispBits to the screen - this is just
ioShowDisplay with the extra parameter of the windowIndex integer"
	(self ioShowDisplayOnWindow: dispBits _: w _: h _: d _: left _: right _: top _: bottom _: windowIndex) ifFalse:
		[interpreterProxy primitiveFail]! !

!HostWindowPlugin methodsFor: 'system primitives' stamp: 'eem 07/21/2017 11:16'!
primitiveSqueakWindowHandle
	<option: #TerfVM>
	"Answer the native Squeak window handle, presumably a pointer, as an unsigned integer.
	 This value should be usable in native calls to the window manager as well as
	 the argument to primitiveNativeWindowPosition:"
	| handle |
	<var: #handle type: 'void *'>
	self primitive: 'primitiveSqueakWindowHandle'.
	handle := self ioGetWindowHandle.
	handle = 0 ifTrue:
		[^interpreterProxy primitiveFail].
	^interpreterProxy positiveMachineIntegerFor: (self cCoerceSimple: handle to: #sqInt)! !

!HostWindowPlugin methodsFor: 'initialize-release' stamp: 'eem 11/10/2020 12:02'!
shutdownModule
"do any window related VM closing down work your platform requires."
	<export: true>
	^self ioCloseAllWindows! !

HostWindowPlugin subclass: #HostWindowPluginSimulator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!HostWindowPluginSimulator methodsFor: 'primitive simulation' stamp: 'eem 10/16/2020 22:43'!
ioGetWindowHandle
	<primitive: 'primitiveSqueakWindowHandle' module: 'HostWindowPlugin' error: ec>
	self primitiveFailed

	"self basicNew ioGetWindowHandle hex"! !

!HostWindowPluginSimulator methodsFor: 'primitive simulation' stamp: 'eem 10/16/2020 23:26'!
ioPositionOfNativeDisplay: windowHandle
	<primitive: 'primitiveNativeDisplayPosition' module: 'HostWindowPlugin' error: ec>
	self primitiveFailed

	"self basicNew ioPositionOfNativeDisplay: self basicNew ioGetWindowHandle"
	"self basicNew ioPositionOfNativeWindow: self basicNew ioGetWindowHandle"! !

!HostWindowPluginSimulator methodsFor: 'primitive simulation' stamp: 'eem 10/16/2020 23:27'!
ioPositionOfNativeWindow: windowHandle
	<primitive: 'primitiveNativeWindowPosition' module: 'HostWindowPlugin' error: ec>
	self primitiveFailed

	"self basicNew ioPositionOfNativeWindow: self basicNew ioGetWindowHandle"
	"self basicNew ioPositionOfNativeDisplay: self basicNew ioGetWindowHandle"! !

!HostWindowPluginSimulator methodsFor: 'primitive simulation' stamp: 'eem 10/4/2020 18:43'!
ioPositionOfWindow: windowIndex
	<primitive: 'primitiveHostWindowPosition' module: 'HostWindowPlugin' error: ec>
	^-1

	"self basicNew ioPositionOfWindow: 1"! !

!HostWindowPluginSimulator methodsFor: 'primitive simulation' stamp: 'eem 10/4/2020 19:15'!
ioPositionOfWindowSet: windowIndex x: x y: y
	<primitive: 'primitiveHostWindowPositionSet' module: 'HostWindowPlugin' error: ec>
	^-1

	"self basicNew ioPositionOfWindowSet: 1 x: 500 y: 500"
	"self basicNew ioPositionOfWindowSet: 1 x: 64 y: 64"! !

!HostWindowPluginSimulator methodsFor: 'primitive simulation' stamp: 'eem 1/22/2021 11:34'!
ioScreenRectangles
	self ioScreenRectanglesPrimitive
		ifNil: [interpreterProxy primitiveFailFor: PrimErrNoMemory]
		ifNotNil:
			[:a|
			(interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: a size)
				ifNil: [interpreterProxy primitiveFailFor: PrimErrNoMemory]
				ifNotNil:
					[:result|
					1 to: a size do:
						[:i| interpreterProxy storeInteger: i - 1 ofObject: result withValue: (a at: i)].
					^result]].
	^nil! !

!HostWindowPluginSimulator methodsFor: 'primitive simulation' stamp: 'eem 1/22/2021 11:46'!
ioScreenRectanglesPrimitive
	<primitive: 'primitiveScreenRectangles' module: 'HostWindowPlugin' error: ec>
	^nil

	"self basicNew ioScreenRectanglesPrimitive"
	"self basicNew ioScreenRectanglesPrimitive collect:
		[:pp| | x y |
		x := pp bitShift: -16.
		y := pp bitAnd: 16rFFFF.
		y > 32767 ifTrue: [y := y - 65536].
		x@y]"! !

!HostWindowPluginSimulator methodsFor: 'primitive simulation' stamp: 'eem 10/4/2020 18:19'!
ioSetCursorPositionX: x Y: y
	<primitive: 'primitiveSetCursorPosition' module: 'HostWindowPlugin' error: ec>
	^-1

	"self basicNew ioSetCursorPositionX: Display width // 2 Y: Display height // 2"! !

!HostWindowPluginSimulator methodsFor: 'primitive simulation' stamp: 'eem 10/16/2020 23:25'!
ioSizeOfNativeDisplay: windowHandle
	<primitive: 'primitiveNativeDisplaySize' module: 'HostWindowPlugin' error: ec>
	self primitiveFailed

	"self basicNew ioSizeOfNativeDisplay: self basicNew ioGetWindowHandle"
	"self basicNew ioSizeOfNativeWindow: self basicNew ioGetWindowHandle"! !

!HostWindowPluginSimulator methodsFor: 'primitive simulation' stamp: 'eem 10/16/2020 23:25'!
ioSizeOfNativeWindow: windowHandle
	<primitive: 'primitiveNativeWindowSize' module: 'HostWindowPlugin' error: ec>
	self primitiveFailed

	"self basicNew ioSizeOfNativeWindow: self basicNew ioGetWindowHandle"
	"self basicNew ioSizeOfNativeDisplay: self basicNew ioGetWindowHandle"! !

!HostWindowPluginSimulator methodsFor: 'primitive simulation' stamp: 'eem 10/4/2020 18:44'!
ioSizeOfWindow: windowIndex
	<primitive: 'primitiveHostWindowSize' module: 'HostWindowPlugin' error: ec>
	^-1

	"self basicNew ioSizeOfWindow: 1"! !

!HostWindowPluginSimulator methodsFor: 'primitive simulation' stamp: 'eem 10/4/2020 19:53'!
ioSizeOfWindowSet: windowIndex x: x y: y
	<primitive: 'primitiveHostWindowSizeSet' module: 'HostWindowPlugin' error: ec>
	^-1

	"self basicNew ioSizeOfWindowSet: 1 x: Display width - 32 y: Display height - 32"
	"self basicNew ioSizeOfWindowSet: 1 x: Display width + 32 y: Display height + 32"! !

!HostWindowPluginSimulator methodsFor: 'support' stamp: 'eem 10/4/2020 19:37'!
pointFromCompactPointEncoding: encodedPoint
	"For simplicity in the simulation invoking primitives that may or may not answer a point..."
	<inline: #always>
	encodedPoint isPoint ifTrue:
		[^interpreterProxy
			makePointwithxValue: encodedPoint x
			yValue: encodedPoint y].
	^super pointFromCompactPointEncoding: encodedPoint! !

SmartSyntaxInterpreterPlugin subclass: #IOSPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: 'VMBasicConstants'
	category: 'VMMaker-Plugins-IOS'!

!IOSPlugin methodsFor: 'primitives' stamp: 'EstebanLorenzano 3/7/2012 11:08'!
primitiveShowKeyboard: showBoolean
	<export: true>
	self primitive: #primitiveShowKeyboard parameters: #(Boolean).
	self sqShowKeyboard: showBoolean.! !

SmartSyntaxInterpreterPlugin subclass: #ImmX11Plugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!ImmX11Plugin class methodsFor: 'as yet unclassified' stamp: 'dtl 1/7/2018 12:05'!
declareHeaderFilesIn: cg

	cg addHeaderFile: '<locale.h>'.
	cg addHeaderFile: '"sqUnixMain.h"'.
	cg addHeaderFile: '"sqUnixCharConv.h"', Character cr asString,
		'extern char *setLocale(char *, size_t);', Character cr asString,
		'extern int setCompositionWindowPosition(int, int);', Character cr asString,
		'extern int setCompositionFocus(int);'.! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'Tsutomu (auto pragmas dtl 2011-02-01) Hiroshima 11/30/2003 04:49'!
primGetEncoding

	| ret array len |

	<export: true>
	<var: #len type: 'int'>
	<var: #array type: 'char *'>
	len := self cCode: 'strlen(sqTextEncoding)'.

	ret := interpreterProxy
			instantiateClass: interpreterProxy classString
			indexableSize: len.
	array := self cCoerce: (interpreterProxy firstIndexableField: ret) to: 'char *'.	
	self cCode: 'strncpy(array, (char *)sqTextEncoding, len)'.
	
	^ ret.! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'eem 12/11/2012 13:14'!
primGetLocale

	| ret array len locale |

	<export: true>
	<var: #len type: 'int'>
	<var: #array type: 'char *'>
	<var: #locale type: 'char *'>
	
	locale := self cCode: 'setlocale(LC_CTYPE, "")'.
	(self cCode: 'locale')
		ifFalse: [ ret := interpreterProxy nilObject ]
		ifTrue: [
			len := self cCode: 'strlen(locale)'.
			ret := interpreterProxy
				instantiateClass: interpreterProxy classString
				indexableSize: len.
			array := self cCoerce: (interpreterProxy firstIndexableField: ret) to: 'char *'.
			self cCode: 'strncpy(array, (char *)locale, len)'.
		].	
	^ ret.! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'Tsutomu (auto pragmas dtl 2011-02-01) Hiroshima 11/30/2003 04:51'!
primGetLocaleEncoding

	| ret array len |

	<export: true>
	<var: #len type: 'int'>
	<var: #array type: 'char *'>

	(self cCode: '(int) localeEncoding')
		ifFalse: [ ret := interpreterProxy nilObject ]
		ifTrue: [
			len := self cCode: 'strlen(localeEncoding)'.
			ret := interpreterProxy
				instantiateClass: interpreterProxy classString
				indexableSize: len.
			array := self cCoerce: (interpreterProxy firstIndexableField: ret) to: 'char *'.
			self cCode: 'strncpy(array, (char *)localeEncoding, len)'.
		].	
	^ ret.! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'Tsutomu (auto pragmas dtl 2011-02-01) Hiroshima 12/5/2003 14:43'!
primGetPathEnc

	| ret array len |

	<export: true>
	<var: #len type: 'int'>
	<var: #array type: 'char *'>
	len := self cCode: 'strlen(uxPathEncoding)'.

	ret := interpreterProxy
			instantiateClass: interpreterProxy classString
			indexableSize: len.
	array := self cCoerce: (interpreterProxy firstIndexableField: ret) to: 'char *'.	
	self cCode: 'strncpy(array, (char *)uxPathEncoding, len)'.
	
	^ ret.! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'Tsutomu (auto pragmas dtl 2011-02-01) Hiroshima 11/30/2003 04:51'!
primGetTextEnc

	| ret array len |

	<export: true>
	<var: #len type: 'int'>
	<var: #array type: 'char *'>
	len := self cCode: 'strlen(uxTextEncoding)'.

	ret := interpreterProxy
			instantiateClass: interpreterProxy classString
			indexableSize: len.
	array := self cCoerce: (interpreterProxy firstIndexableField: ret) to: 'char *'.	
	self cCode: 'strncpy(array, (char *)uxTextEncoding, len)'.
	
	^ ret.! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'Tsutomu (auto pragmas dtl 2011-02-01) Hiroshima 12/1/2003 18:34'!
primGetXWinEnc

	| ret array len |

	<export: true>
	<var: #len type: 'int'>
	<var: #array type: 'char *'>
	len := self cCode: 'strlen(uxXWinEncoding)'.

	ret := interpreterProxy
			instantiateClass: interpreterProxy classString
			indexableSize: len.
	array := self cCoerce: (interpreterProxy firstIndexableField: ret) to: 'char *'.	
	self cCode: 'strncpy(array, (char *)uxXWinEncoding, len)'.
	
	^ ret.! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'Tsutomu (auto pragmas dtl 2011-02-01) Hiroshima 11/28/2003 16:24'!
primIsTextEncUTF8

	<export: true>
	self primitive: 'primIsTextEncUTF8'.
	^ (self cCode: 'textEncodingUTF8') asOop: Boolean.! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'nice 11/8/2019 09:24'!
primSetCompositionFocus: aBoolean

	| ret |
	<export: true>
	self primitive: 'primSetCompositionFocus'
		parameters: #(Boolean).
	
	ret := self cCode: 'setCompositionFocus(aBoolean)'.

	ret = false ifTrue: [
		interpreterProxy primitiveFail.
		^ nil
	].

	^ ret asOop: Boolean
! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'Tsutomu (auto pragmas dtl 2011-02-01) Hiroshima 1/17/2004 08:14'!
primSetCompositionWindowPositionX: x y: y

	| ret |

	<export: true>
	self primitive: 'primSetCompositionWindowPosition'
		parameters: #(SmallInteger SmallInteger).
	

	ret := self cCode: 'setCompositionWindowPosition(x, y)'.

	ret = false ifTrue: [
		interpreterProxy primitiveFail.
		^ nil
	].

	^ ret asOop: Boolean

! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'eem 12/11/2012 13:15'!
primSetEncoding: encoding

	| name len |
	<export: true>
	<var: #name type: 'char *'>
	<var: #len type: 'size_t'>
	len := interpreterProxy stSizeOf: encoding.
	name := self cCoerce: (interpreterProxy firstIndexableField: encoding) to: 'char *'.
	self cCode: 'setNEncoding(&sqTextEncoding, name, len)'.
	^ encoding.
! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'Tsutomu (auto pragmas dtl 2011-02-01) Hiroshima 11/30/2003 05:07'!
primSetEncodingToLocale

	| ret |
	<export: true>
	self primitive: 'primSetEncodingToLocale'.
	
	(self cCode: '(int) localeEncoding')
		ifFalse: [ ret := interpreterProxy falseObject ] 
		ifTrue: [
			self cCode: 'sqTextEncoding = (void *)localeEncoding'.
			ret := interpreterProxy trueObject
		].
	^ ret! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'eem 12/11/2012 13:15'!
primSetLocale: locale

	| name len ret array |
	<export: true>
	<var: #name type: 'char *'>
	<var: #array type: 'char *'>
	<var: #len type: 'size_t'>
	len :=  interpreterProxy stSizeOf: locale.
	name := self cCoerce: (interpreterProxy arrayValueOf: locale) to: 'char *'.
	name := self cCode: 'setLocale(name, len)'.
	(self cCode: '(int) name')
		ifFalse: [^ interpreterProxy nilObject ]
		ifTrue: [
			len := self cCode: 'strlen(name)'.
			ret := interpreterProxy
				instantiateClass: interpreterProxy classString
				indexableSize: len.
			array := self cCoerce: (interpreterProxy firstIndexableField: ret) to: 'char *'.
			self cCode: 'strncpy(array, (char *)name, len)'.
			^ ret ].
! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'eem 12/11/2012 13:15'!
primSetLocaleEncoding: encoding

	| name len |
	<export: true>
	<var: #name type: 'char *'>
	<var: #len type: 'size_t'>
	len := interpreterProxy stSizeOf: encoding.
	name := self cCoerce: (interpreterProxy firstIndexableField: encoding) to: 'char *'.
	self cCode: 'setNEncoding(&localeEncoding, name, len)'.
	self cCode: 'sqTextEncoding= uxTextEncoding= uxPathEncoding= uxXWinEncoding= localeEncoding'.
	^ encoding.
! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'eem 12/11/2012 13:15'!
primSetPathEnc: encoding

	| name len |
	<export: true>
	<var: #name type: 'char *'>
	<var: #len type: 'size_t'>
	len := interpreterProxy stSizeOf: encoding.
	name := self cCoerce: (interpreterProxy firstIndexableField: encoding) to: 'char *'.
	self cCode: 'setNEncoding(&uxPathEncoding, name, len)'.
	^ encoding.! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'Tsutomu (auto pragmas dtl 2011-02-01) Hiroshima 12/5/2003 14:40'!
primSetPathEncToLocale

	| ret |
	<export: true>
	self primitive: 'primSetPathEncToLocale'.
	
	(self cCode: '(int) localeEncoding')
		ifFalse: [ ret := interpreterProxy falseObject ] 
		ifTrue: [
			self cCode: 'uxPathEncoding = (void *)localeEncoding'.
			ret := interpreterProxy trueObject
		].
	^ ret! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'eem 12/11/2012 13:16'!
primSetTextEnc: encoding

	| name len |
	<export: true>
	<var: #name type: 'char *'>
	<var: #len type: 'size_t'>
	len := interpreterProxy stSizeOf: encoding.
	name := self cCoerce: (interpreterProxy firstIndexableField: encoding) to: 'char *'.
	self cCode: 'setNEncoding(&uxTextEncoding, name, len)'.
	^ encoding.! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'Tsutomu (auto pragmas dtl 2011-02-01) Hiroshima 11/30/2003 05:08'!
primSetTextEncToLocale

	| ret |
	<export: true>
	self primitive: 'primSetTextEncToLocale'.
	
	(self cCode: '(int) localeEncoding')
		ifFalse: [ ret := interpreterProxy falseObject ] 
		ifTrue: [
			self cCode: 'uxTextEncoding = (void *)localeEncoding'.
			ret := interpreterProxy trueObject
		].
	^ ret! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'nice 11/8/2019 09:23'!
primSetTextEncUTF8: aBoolean

	<export: true>
	self primitive: 'primSetTextEncUTF8'
		parameters: #(Boolean).
	self cCode: 'textEncodingUTF8 = aBoolean'.
	^ aBoolean asOop: Boolean.! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'eem 12/11/2012 13:16'!
primSetXWinEnc: encoding

	| name len |
	<export: true>
	<var: #name type: 'char *'>
	<var: #len type: 'size_t'>
	len := interpreterProxy stSizeOf: encoding.
	name := self cCoerce: (interpreterProxy firstIndexableField: encoding) to: 'char *'.
	self cCode: 'setNEncoding(&uxXWinEncoding, name, len)'.
	^ encoding.! !

!ImmX11Plugin methodsFor: 'primitives' stamp: 'Tsutomu (auto pragmas dtl 2011-02-01) Hiroshima 12/1/2003 18:35'!
primSetXWinEncToLocale

	| ret |
	<export: true>
	self primitive: 'primSetXWinEncToLocale'.
	
	(self cCode: '(int) localeEncoding')
		ifFalse: [ ret := interpreterProxy falseObject ] 
		ifTrue: [
			self cCode: 'uxXWinEncoding = (void *)localeEncoding'.
			ret := interpreterProxy trueObject
		].
	^ ret! !

SmartSyntaxInterpreterPlugin subclass: #InternetConfigPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!InternetConfigPlugin commentStamp: 'tpr 5/5/2003 12:05' prior: 0!
This plugin provides access to the Mac's internet configuration toolkit - so long as you are running on a Mac.!

!InternetConfigPlugin class methodsFor: 'translation' stamp: 'JMM 9/26/2001 12:21'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!InternetConfigPlugin class methodsFor: 'translation' stamp: 'JMM 9/26/2001 12:21'!
requiresPlatformFiles
	"this plugin requires platform specific files in order to work"
	^true! !

!InternetConfigPlugin methodsFor: 'initialize' stamp: 'JMM (auto pragmas 12/08) 9/26/2001 12:22'!
initialiseModule
	<export: true>
	^self cCode: 'sqInternetConfigurationInit()' inSmalltalk:[true]! !

!InternetConfigPlugin methodsFor: 'system primitives' stamp: 'eem 11/1/2013 23:48'!
primitiveGetMacintoshFileTypeAndCreatorFrom: aFileName
	| oop ptr keyLength creator |

	<var: #creator declareC: 'char creator[8]'>
	<var: #ptr type: 'char *'>
	self primitive: 'primitiveGetMacintoshFileTypeAndCreatorFrom'
		parameters: #(String).

	keyLength := interpreterProxy byteSizeOf: aFileName cPtrAsOop.
	self sqInternetGetMacintoshFileTypeAndCreatorFrom: aFileName keySize: keyLength into: creator.
	oop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: 8.
	ptr := interpreterProxy firstIndexableField: oop.
	0 to: 7 do:[:i|
		ptr at: i put: (creator at: i)].
	^oop.
! !

!InternetConfigPlugin methodsFor: 'system primitives' stamp: 'eem 10/25/2018 11:54'!
primitiveGetStringKeyedBy: aKey
	| oop ptr size aString keyLength |

	<var: #aString declareC: 'char aString[1025]'>
	<var: #ptr type: 'char *'>
	self primitive: 'primitiveGetStringKeyedBy'
		parameters: #(String).
	self cCode: '' inSmalltalk: [aString := ByteString new: 1025].
	keyLength := interpreterProxy byteSizeOf: aKey cPtrAsOop.
	size := self sqInternetConfigurationGetStringKeyedBy: aKey keySize: keyLength into: aString.
	oop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: size.
	ptr := interpreterProxy firstIndexableField: oop.
	0 to: size-1 do:[:i|
		ptr at: i put: (aString at: i)].
	^oop.
! !

!InternetConfigPlugin methodsFor: 'initialize' stamp: 'JMM (auto pragmas 12/08) 9/26/2001 12:22'!
shutdownModule
	<export: true>
	^self cCode: 'sqInternetConfigurationShutdown()' inSmalltalk:[true]! !

!InternetConfigPlugin methodsFor: 'simulation' stamp: 'eem 10/25/2018 11:56'!
sqInternetConfigurationGetStringKeyedBy: keyCArray keySize: keySize into: aString
	<doNotGenerate>
	| result |
	result := InternetConfiguration primitiveGetStringKeyedBy: (self asString: keyCArray size: keySize).
	self strncpy: aString _: result _: (aString size min: result size).
	^result size! !

SmartSyntaxInterpreterPlugin subclass: #JPEGReadWriter2Plugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!JPEGReadWriter2Plugin commentStamp: '<historical>' prior: 0!
This work is a Squeak VM Plugin version of LibJPEG.

The following sentence applies to this class:
  "This software is based in part on the work of the Independent JPEG Group".

You can read more about it at www.ijg.org

In addition to the code generated from this class, the plugin uses the following files (from LibJPEG ver. 6b):

jerror.c
jcmarker.c
jdmarker.c
jctrans.c
jcparam.c
jdapimin.c
jcapimin.c
jidctred.c
jidctflt.c
jidctfst.c
jidctint.c
jccoefct.c
jdinput.c
jdmaster.c
jdcoefct.c
jdhuff.c
jdphuff.c
jcphuff.c
jchuff.c
jcomapi.c
jcinit.c
jcmaster.c
jdcolor.c
jdtrans.c
jmemmgr.c
jutils.c
jddctmgr.c
jcdctmgr.c
jquant2.c
jquant1.c
jmemnobs.c
jfdctint.c
jfdctfst.c
jfdctflt.c
jdsample.c
jdpostct.c
jdmerge.c
jdmainct.c
jdapistd.c
jcsample.c
jcprepct.c
jcmainct.c
jccolor.c
jcapistd.c
jversion.h
jpeglib.h
jdhuff.h
jchuff.h
jerror.h
jmorecfg.h
jmemsys.h
jpegint.h
jdct.h
jinclude.h

!

!JPEGReadWriter2Plugin class methodsFor: 'translation' stamp: 'JMM 10/3/2001 11:44'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!JPEGReadWriter2Plugin class methodsFor: 'translation' stamp: 'tpr 3/1/2002 17:03'!
requiresCrossPlatformFiles
	"default is ok for most, any plugin needing cross platform files must say so"
	^true! !

!JPEGReadWriter2Plugin class methodsFor: 'translation' stamp: 'JMM 10/3/2001 11:48'!
requiresPlatformFiles
	"default is ok for most, any plugin needing platform specific files must say so"
	^true! !

!JPEGReadWriter2Plugin methodsFor: 'plugin support' stamp: 'eem 10/18/2018 14:49'!
compressStructSize
	<inline: #always>
	^self
		cCode: [self sizeof: #'struct jpeg_compress_struct']
		inSmalltalk: [JPEGReadWriter2 new primJPEGCompressStructSize]! !

!JPEGReadWriter2Plugin methodsFor: 'plugin support' stamp: 'eem 10/18/2018 14:53'!
decompressStructSize
	<inline: #always>
	^self
		cCode: [self sizeof: #'struct jpeg_decompress_struct']
		inSmalltalk: [JPEGReadWriter2 new primJPEGDecompressStructSize]! !

!JPEGReadWriter2Plugin methodsFor: 'plugin support' stamp: 'eem 10/18/2018 14:49'!
errorMgr2StructSize
	<inline: #always>
	^self
		cCode: [self sizeof: #'struct error_mgr2']
		inSmalltalk: [JPEGReadWriter2 new primJPEGErrorMgr2StructSize]! !

!JPEGReadWriter2Plugin methodsFor: 'plugin support' stamp: 'eem 10/18/2018 19:57'!
isValidCompressionStruct: aByteArray
	<inline: #always>
	^(interpreterProxy byteSizeOf: aByteArray cPtrAsOop) >= self compressStructSize! !

!JPEGReadWriter2Plugin methodsFor: 'plugin support' stamp: 'eem 10/18/2018 19:57'!
isValidDecompressionStruct: aByteArray
	<inline: #always>
	^(interpreterProxy byteSizeOf: aByteArray cPtrAsOop) >= self decompressStructSize! !

!JPEGReadWriter2Plugin methodsFor: 'plugin support' stamp: 'eem 10/18/2018 19:57'!
isValidErrorMessageStruct: aByteArray
	<inline: #always>
	^(interpreterProxy byteSizeOf: aByteArray cPtrAsOop) >= self errorMgr2StructSize! !

!JPEGReadWriter2Plugin methodsFor: 'primitives' stamp: 'eem 10/18/2018 15:57'!
primImageHeight: aJPEGDecompressStruct
	<export: true>
	self primitive: 'primImageHeight' parameters: #(ByteArray).

	"Various parameter checks"
	(self isValidDecompressionStruct: aJPEGDecompressStruct) ifFalse:
		[^interpreterProxy primitiveFail].

	^(self
		cCode: '((j_decompress_ptr)aJPEGDecompressStruct)->image_height'
		inSmalltalk: [JPEGReadWriter2 new primImageHeight: aJPEGDecompressStruct asByteArray])
			asOop: SmallInteger! !

!JPEGReadWriter2Plugin methodsFor: 'primitives' stamp: 'eem 10/18/2018 15:56'!
primImageNumComponents: aJPEGDecompressStruct
	<export: true>
	self primitive: 'primImageNumComponents' parameters: #(ByteArray).

	"Various parameter checks"
	(self isValidDecompressionStruct: aJPEGDecompressStruct) ifFalse:
		[^interpreterProxy primitiveFail].

	^(self
		cCode: '((j_decompress_ptr)aJPEGDecompressStruct)->num_components'
		inSmalltalk: [JPEGReadWriter2 new primImageNumComponents: aJPEGDecompressStruct asByteArray])
			asOop: SmallInteger! !

!JPEGReadWriter2Plugin methodsFor: 'primitives' stamp: 'eem 10/18/2018 15:55'!
primImageWidth: aJPEGDecompressStruct
	<export: true>
	self primitive: 'primImageWidth' parameters: #(ByteArray).

	"Various parameter checks"
	(self isValidDecompressionStruct: aJPEGDecompressStruct) ifFalse:
		[^interpreterProxy primitiveFail].

	^(self
		cCode: '((j_decompress_ptr)aJPEGDecompressStruct)->image_width'
		inSmalltalk: [JPEGReadWriter2 new primImageWidth: aJPEGDecompressStruct asByteArray])
			asOop: SmallInteger! !

!JPEGReadWriter2Plugin methodsFor: 'primitives' stamp: 'eem 10/18/2018 13:43'!
primJPEGCompressStructSize
	<export: true>
	self primitive: #primJPEGCompressStructSize parameters: #().

	^self compressStructSize asOop: SmallInteger! !

!JPEGReadWriter2Plugin methodsFor: 'primitives' stamp: 'eem 10/18/2018 13:43'!
primJPEGDecompressStructSize
	<export: true>
	self primitive: #primJPEGDecompressStructSize parameters: #().

	^self decompressStructSize asOop: SmallInteger! !

!JPEGReadWriter2Plugin methodsFor: 'primitives' stamp: 'eem 10/18/2018 13:43'!
primJPEGErrorMgr2StructSize
	<export: true>
	self primitive: #primJPEGErrorMgr2StructSize parameters: #().

	^self errorMgr2StructSize asOop: SmallInteger! !

!JPEGReadWriter2Plugin methodsFor: 'primitives' stamp: 'eem 10/18/2018 13:41'!
primJPEGPluginIsPresent
	<export: true>
	self primitive: #primJPEGPluginIsPresent parameters: #().

	^(self cCode: [true] inSmalltalk: [JPEGReadWriter2 primJPEGPluginIsPresent]) asOop: Boolean! !

!JPEGReadWriter2Plugin methodsFor: 'primitives' stamp: 'eem 10/18/2018 17:38'!
primJPEGReadHeader: aJPEGDecompressStruct fromByteArray: source errorMgr: aJPEGErrorMgr2Struct
	<export: true>
	
	| sourceSize |
	self
		primitive: 'primJPEGReadHeaderfromByteArrayerrorMgr'
		parameters: #(ByteArray ByteArray ByteArray).

	"Various parameter checks"
	((self isValidDecompressionStruct: aJPEGDecompressStruct)
	 and: [self isValidErrorMessageStruct: aJPEGErrorMgr2Struct]) ifFalse:
		[^interpreterProxy primitiveFail].
	
	sourceSize := interpreterProxy byteSizeOf: source cPtrAsOop.
	sourceSize > 0 ifTrue:
		[self primJPEGReadHeader: aJPEGDecompressStruct
			fromByteArray: source
			size: sourceSize
			errorMgrReadHeader: aJPEGErrorMgr2Struct]! !

!JPEGReadWriter2Plugin methodsFor: 'simulation' stamp: 'eem 10/18/2018 17:00'!
primJPEGReadHeader: jpegDecompressStruct fromByteArray: source size: sourceSize errorMgrReadHeader: jpegErrorMgr2Struct
	"void primJPEGReadHeaderfromByteArraysizeerrorMgrReadHeader(
			char* jpegDecompressStruct,
			char* source,
			unsigned int sourceSize,
			char* jpegErrorMgr2Struct)"
	<doNotGenerate>
	| decompressStruct errorStruct sourceBytes |
	sourceBytes := source asByteArray.
	self assert: sourceBytes size >= sourceSize.
	decompressStruct := jpegDecompressStruct asByteArray.
	errorStruct := jpegErrorMgr2Struct asByteArray.
	(self evaluateIfFailed:
		[JPEGReadWriter2 new
			primJPEGReadHeader: decompressStruct
			fromByteArray: sourceBytes
			errorMgr: errorStruct]) ifTrue: [^nil].
	jpegDecompressStruct overwriteContentsWith: decompressStruct.
	jpegErrorMgr2Struct overwriteContentsWith: errorStruct.
	source overwriteContentsWith: sourceBytes.
	^nil! !

!JPEGReadWriter2Plugin methodsFor: 'primitives' stamp: 'eem 10/18/2018 13:47'!
primJPEGReadImage: aJPEGDecompressStruct fromByteArray: source onForm: form doDithering: ditherFlag errorMgr: aJPEGErrorMgr2Struct
	<export: true>
	| formBitmap formNativeDepth formDepth formWidth formHeight pixelsPerWord formPitch formBitmapSizeInBytes sourceSize formBitmapOOP formComponentBitSize formComponents wordsPerRow |
	<var: #formBitmap type: #'unsigned int *'>

	self
		primitive: 'primJPEGReadImagefromByteArrayonFormdoDitheringerrorMgr'
		parameters: #(ByteArray ByteArray Form Boolean ByteArray).

	formBitmapOOP := interpreterProxy fetchPointer: 0 ofObject: form. 
	formNativeDepth := interpreterProxy fetchInteger: 3 ofObject: form.
	formWidth := interpreterProxy fetchInteger: 1 ofObject: form.
	formHeight := interpreterProxy fetchInteger: 2 ofObject: form.
	formDepth := formNativeDepth abs.
	
	"Various parameter checks"
	((self isValidDecompressionStruct: aJPEGDecompressStruct)
	 and: [self isValidErrorMessageStruct: aJPEGErrorMgr2Struct]) ifFalse:
		[^interpreterProxy primitiveFail].
	
	formComponents := formDepth ~= 8 ifTrue: [4] ifFalse: [1].
	formComponentBitSize := formDepth ~= 16 ifTrue: [8] ifFalse: [4].
	pixelsPerWord := 32 // (formComponents * formComponentBitSize).
	wordsPerRow := (formWidth + pixelsPerWord - 1) // pixelsPerWord.
	formPitch := formWidth + (pixelsPerWord-1) // pixelsPerWord * 4.
	formBitmapSizeInBytes := interpreterProxy byteSizeOf: formBitmapOOP.
	
	interpreterProxy success: 
		((interpreterProxy isWordsOrBytes: formBitmapOOP)
		 and: [formBitmapSizeInBytes >= (formPitch * formHeight)]).
	interpreterProxy failed ifTrue: [^nil].
	
	sourceSize := interpreterProxy stSizeOf: (interpreterProxy stackValue: 3).
	
	interpreterProxy success: (sourceSize ~= 0).
	interpreterProxy failed ifTrue: [^nil].
	
	formBitmap := interpreterProxy firstIndexableField: formBitmapOOP.
	
	self primJPEGReadImagefromByteArrayonFormdoDitheringerrorMgrReadScanlines: aJPEGDecompressStruct
																			    		_: aJPEGErrorMgr2Struct
																					_: source
																			    		_: sourceSize
																			    		_: ditherFlag
																			    		_: formBitmap
																			   		_: pixelsPerWord
																			   		_: wordsPerRow
																			    		_: formNativeDepth! !

!JPEGReadWriter2Plugin methodsFor: 'simulation' stamp: 'eem 10/18/2018 17:00'!
primJPEGReadImagefromByteArrayonFormdoDitheringerrorMgrReadScanlines: jpegDecompressStruct _: jpegErrorMgr2Struct _: source _: sourceSize _: ditherFlag _: formBitmap _: pixelsPerWord _: wordsPerRow _: formNativeDepth
	"void primJPEGReadImagefromByteArrayonFormdoDitheringerrorMgrReadScanlines(
		    char* jpegDecompressStruct,
		    char* jpegErrorMgr2Struct,
		    char* source,
		    unsigned int sourceSize,
		    int ditherFlag,
		    unsigned int* bitmap,
		    unsigned int pixelsPerWord,
		    unsigned int wordsPerRow,
		    int nativeDepth)"
	<doNotGenerate>
	| sourceBytes decompressStruct errorStruct form |
	sourceBytes := source asByteArray.
	decompressStruct := jpegDecompressStruct asByteArray.
	errorStruct := jpegErrorMgr2Struct asByteArray.
	(self evaluateIfFailed:
		[| slave height |
		 slave := JPEGReadWriter2 new.
		 height := slave primImageWidth: decompressStruct. "cheating ;-)"
		 form := Form extent: wordsPerRow * pixelsPerWord @ height depth: formNativeDepth.
		 slave
			primJPEGReadImage: decompressStruct
			fromByteArray: sourceBytes
			onForm: form
			doDithering: true
			errorMgr: errorStruct]) ifTrue: [^nil].
	jpegDecompressStruct overwriteContentsWith: decompressStruct.
	jpegErrorMgr2Struct overwriteContentsWith: errorStruct.
	source overwriteContentsWith: sourceBytes. "could happen..."
	ByteArray adoptInstance: form bits.
	formBitmap overwriteContentsWith: form bits.
	^nil! !

!JPEGReadWriter2Plugin methodsFor: 'primitives' stamp: 'eem 10/18/2018 13:47'!
primJPEGWriteImage: aJPEGCompressStruct onByteArray: destination form: form quality: quality progressiveJPEG: progressiveFlag errorMgr: aJPEGErrorMgr2Struct

	| formBitmap formWidth formHeight formNativeDepth formDepth destinationSize pixelsPerWord wordsPerRow formPitch formBitmapSizeInBytes formBitmapOOP formComponentBitSize formComponents |
	<export: true>
	<var: #formBitmap type: 'unsigned int *'> 
	<var: #destinationSize type: 'unsigned int'>

	self
		primitive: 'primJPEGWriteImageonByteArrayformqualityprogressiveJPEGerrorMgr'
		parameters: #(ByteArray ByteArray Form SmallInteger Boolean ByteArray).

	formBitmapOOP := interpreterProxy fetchPointer: 0 ofObject: form.
	formWidth := interpreterProxy fetchInteger: 1 ofObject: form.
	formHeight := interpreterProxy fetchInteger: 2 ofObject: form.
	formNativeDepth := interpreterProxy fetchInteger: 3 ofObject: form.
	formDepth := formNativeDepth abs.

	"Various parameter checks"
	((self isValidCompressionStruct: aJPEGCompressStruct)
	 and: [self isValidErrorMessageStruct: aJPEGErrorMgr2Struct]) ifFalse:
		[^interpreterProxy primitiveFail].

	
	formComponents := formDepth ~= 8 ifTrue: [4] ifFalse: [1].
	formComponentBitSize := formDepth ~= 16 ifTrue: [8] ifFalse: [4].
	pixelsPerWord := 32 // (formComponents * formComponentBitSize).
	wordsPerRow := (formWidth + pixelsPerWord - 1) // pixelsPerWord.
	formPitch := wordsPerRow * 4.
	formBitmapSizeInBytes := interpreterProxy byteSizeOf: formBitmapOOP.
	interpreterProxy success: 
		((interpreterProxy isWordsOrBytes: formBitmapOOP) and: 
		[formBitmapSizeInBytes >= (formPitch * formHeight)]).
	interpreterProxy failed ifTrue: [ ^ nil ].
	
	formBitmap := interpreterProxy firstIndexableField: formBitmapOOP.
	destinationSize := interpreterProxy stSizeOf: (interpreterProxy stackValue: 4).
	destinationSize > 0 ifTrue:
		[self primJPEGWriteImageonByteArrayformqualityprogressiveJPEGerrorMgrWriteScanlines: formWidth
																								_: formHeight
																								_: formNativeDepth
																								_: formBitmap
																								_: aJPEGCompressStruct
																								_: aJPEGErrorMgr2Struct
																								_: quality
																								_: progressiveFlag
																								_: pixelsPerWord
																								_: wordsPerRow
																								_: destination
																								_: (self addressOf: destinationSize put: [:v| destinationSize := v])].
	
	^destinationSize asOop: SmallInteger! !

!JPEGReadWriter2Plugin methodsFor: 'simulation' stamp: 'eem 10/18/2018 17:00'!
primJPEGWriteImageonByteArrayformqualityprogressiveJPEGerrorMgrWriteScanlines: width _: height _: nativeDepth _: bitmap _: jpegCompressStruct _: jpegErrorMgr2Struct _: quality _: progressiveFlag _: pixelsPerWord _: wordsPerRow _: destination _: destinationSizePtr
	"void primJPEGWriteImageonByteArrayformqualityprogressiveJPEGerrorMgrWriteScanlines(
			unsigned int width,
			unsigned int height,
			int nativeDepth,
			unsigned int* bitmap,
			char* jpegCompressStruct,
			char* jpegErrorMgr2Struct,
			int quality,
			int progressiveFlag,
			unsigned int pixelsPerWord,
			unsigned int wordsPerRow,
			char* destination,
			unsigned int* destinationSizePtr)"
	<doNotGenerate>
	| bits form compressStruct errorStruct destinationBytes destinationSize |
	bits := bitmap asByteArray.
	Bitmap adoptInstance: bits.
	form := Form extent: width @ height depth: nativeDepth bits: bits.
	compressStruct := jpegCompressStruct asByteArray.
	errorStruct := jpegErrorMgr2Struct asByteArray.
	destinationBytes := destination asByteArray.
	(self evaluateIfFailed:
		[destinationSize :=JPEGReadWriter2 new
			primJPEGWriteImage: compressStruct
			onByteArray: destinationBytes
			form: form
			quality: quality
			progressiveJPEG: progressiveFlag
			errorMgr: errorStruct]) ifTrue: [^nil].
	jpegCompressStruct overwriteContentsWith: compressStruct.
	jpegErrorMgr2Struct overwriteContentsWith: errorStruct.
	destination overwriteContentsWith: destinationBytes.
	destinationSizePtr at: 0 put: destinationSize.
	^nil! !

!JPEGReadWriter2Plugin methodsFor: 'primitives' stamp: 'lpc 6/3/2016 14:13'!
primSupports8BitGrayscaleJPEGs
	<export: true>
	self
		primitive: 'primSupports8BitGrayscaleJPEGs'
		parameters: #().
	^ true asOop: Boolean! !

SmartSyntaxInterpreterPlugin subclass: #JoystickTabletPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!JoystickTabletPlugin commentStamp: 'tpr 5/2/2003 15:48' prior: 0!
This plugin implements the interface to the joystick and tablet input devices, if you have one.  Since it requires platform support it will only be built when supported on your platform!

!JoystickTabletPlugin class methodsFor: 'translation' stamp: 'tpr 5/23/2001 17:08'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!JoystickTabletPlugin class methodsFor: 'translation' stamp: 'tpr 3/26/2002 15:22'!
requiresPlatformFiles
	" this plugin requires platform specific files in order to work"
	^true! !

!JoystickTabletPlugin methodsFor: 'initialize-release' stamp: 'ar (auto pragmas 12/08) 5/12/2000 16:53'!
initialiseModule
	<export: true>
	^self cCode: 'joystickInit()' inSmalltalk:[true]! !

!JoystickTabletPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:35'!
primitiveGetTabletParameters: cursorIndex
	"Get information on the pen tablet attached to this machine. Fail if there is no tablet. If successful, the result is an array of integers; see the Smalltalk call on this primitive for its interpretation."

	| resultSize result resultPtr |
	<var: #resultPtr type: 'int * '>
	self primitive: 'primitiveGetTabletParameters'
		parameters: #(SmallInteger).
	resultSize := self tabletResultSize.
	result := interpreterProxy instantiateClass: interpreterProxy classBitmap indexableSize: resultSize.
	resultPtr := result asIntPtr.
	interpreterProxy success: (self cCode: 'tabletGetParameters(cursorIndex, resultPtr)').
	^result! !

!JoystickTabletPlugin methodsFor: 'primitives' stamp: 'TPR 3/24/2000 18:33'!
primitiveReadJoystick: index
	"Read an input word from the joystick with the given index."

	self primitive: 'primitiveReadJoystick'
		parameters: #(SmallInteger).
	^(self joystickRead: index) asPositiveIntegerObj! !

!JoystickTabletPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:36'!
primitiveReadTablet: cursorIndex
	"Get the current state of the cursor of the pen tablet specified by my argument. Fail if there is no tablet. If successful, the result is an array of integers; see the Smalltalk call on this primitive for its interpretation."

	| resultSize result resultPtr|
	<var: #resultPtr type: 'int * '>
	self primitive: 'primitiveReadTablet'
		parameters: #(SmallInteger).
	resultSize := self tabletResultSize.
	result := interpreterProxy instantiateClass: interpreterProxy classBitmap indexableSize: resultSize.
	resultPtr := result asIntPtr.
	interpreterProxy success: (self cCode: 'tabletRead(cursorIndex, resultPtr)').
	^result! !

!JoystickTabletPlugin methodsFor: 'initialize-release' stamp: 'tpr (auto pragmas 12/08) 3/26/2002 15:22'!
shutdownModule
	<export: true>
	^self cCode: 'joystickShutdown()' inSmalltalk:[true]! !

SmartSyntaxInterpreterPlugin subclass: #LargeIntegersPlugin
	instanceVariableNames: 'andOpIndex orOpIndex xorOpIndex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!LargeIntegersPlugin commentStamp: 'nice 4/5/2016 22:27' prior: 0!
LargeIntegersPlugin provides functions for speeding up LargeInteger arithmetics. Usually it is part of your installation as 'LargeIntegers' plugin (a C-compiled binary).


Correctly installed?
----------------------
Probably you are just working with it.

To be really sure try
	100 factorial. "to force plugin loading"
	SmalltalkImage current listLoadedModules.
Then you should see 'LargeIntegers' somewhere in the output strings. If this should not be the case, you probably have a problem.


Variables
-----------

Inst vars:
	andOpIndex			C constant
	orOpIndex			C constant
	xorOpIndex 			C constant
Used like an enum, in ST one would use symbols instead.

Class vars:
	none


History
--------

v2.0

Rewrite primitives to use 32-bits digits instead of 8-bits digits. This can fasten many operation up to 2x and multiplication up to 7x.
Since Large Integers are still variable byte object, care is taken to allocate just enough bytes to contain integer value (not necessarily a multiple of 4 bytes).
Primitives will use the extra bytes beyond necessary bytes up to next multiple of 4, so it's mandatory that those extra bytes be correctly set to zero
At image side, digits are still 8-bits so that fallback code involve only SmallInteger.

Primitives assume that byte ordering (endianness) is allways little endian, because LargeIntegers remain regular byte Objects.
On (hypothetic) big endian machines, conversion is performed at fetch/store time thru usage of dedicated macro.

v1.5

- no code change at all compared to v1.4
- made to outsource testing code (LargeIntegersPluginTest) introduced in earlier versions
- updated class comment: reference to LargeIntegersPluginTest removed

v1.4

- no semantic change compared to v1.3
- >>cHighBit: improved (could be faster now)
- fix: class comment
- improved class comment
- >>flag: introduced to allow #flag: messages (does nothing)
- new: class>>buildCodeGeneratorUpTo: as hook for switching debugMode (default is not to change anything)
- removed: class>>new (obsolete)
- minor cleanup of source code layout

v1.3

- fix: >>primDigitDiv:negative: now checks if its Integer args are normalized; without this change the plugin crashes, if a division by zero through a non normalized - filled with zero bytes - arg occurs. This can happen through printing by the inspector windows after changing the bytes of a LargeInteger manually.

v1.2

- fix: >>anyBitOfBytes: aBytesOop from: start to: stopArg

v1.1

- >>primGetModuleName for checking the version of the plugin;

- >>primDigitBitShiftMagnitude and >>primAnyBitFrom:to: for supporting - not installing!! - unification of shift semantics of negative Integers;

v1.0

- speeds up digitDiv:neg: at about 20%.
	In >>cCoreDigitDivDiv:len:rem:len:quo:len: the 'nibble' arithmetic is removed.
!

!LargeIntegersPlugin class methodsFor: 'translation' stamp: 'sr 4/8/2000 05:40'!
declareCVarsIn: cg 
	cg var: 'andOpIndex' declareC: 'const int  andOpIndex = 0'.
	cg var: 'orOpIndex' declareC: 'const int  orOpIndex = 1'.
	cg var: 'xorOpIndex' declareC: 'const int  xorOpIndex = 2'! !

!LargeIntegersPlugin class methodsFor: 'translation' stamp: 'ar 5/17/2000 16:08'!
moduleName
	^'LargeIntegers'! !

!LargeIntegersPlugin class methodsFor: 'translation' stamp: 'eem 2/10/2009 14:48'!
moduleNameAndVersion
	"Answer the receiver's module name and version info that is used for the plugin's C code. The default is to append the code generation date, but any useful text is ok (keep it short)"

	^ self moduleName, ' ', self version! !

!LargeIntegersPlugin class methodsFor: 'translation' stamp: 'nice 1/15/2013 00:40'!
version
	"Answer the receiver's version info as String."

	^ 'v2.0'! !

!LargeIntegersPlugin methodsFor: 'util' stamp: 'nice 5/2/2020 12:02'!
anyBitOfLargeInt: anOop from: start to: stopArg 
	"Argument has to be a Large Integer!!"
	"Tests for any magnitude bits in the interval from start to stopArg."
	| magnitude stop firstDigitIx lastDigitIx firstMask lastMask |
	<var: #digit type: #'unsigned int'>
	<var: #firstMask type: #'unsigned int'>
	<var: #lastMask type: #'unsigned int'>
	<var: #firstDigitIx type: #usqInt>
	<var: #lastDigitIx type: #usqInt>
	<var: #ix type: #usqInt>
	magnitude := anOop.
	stop := stopArg min: (self highBitOfLargeInt: magnitude).
	start > stop
		ifTrue: [^ false].
	firstDigitIx := start - 1 // 32 + 1.
	lastDigitIx := stop - 1 // 32 + 1.
	firstMask := 16rFFFFFFFF << (start - 1 bitAnd: 31).
	lastMask := 16rFFFFFFFF >> (31 - (stop - 1 bitAnd: 31)).
	firstDigitIx = lastDigitIx
		ifTrue: [| digit | 
			digit := self unsafeDigitOfLargeInt: magnitude at: firstDigitIx.
			^ (digit bitAnd: (firstMask bitAnd: lastMask))
				~= 0].
	((self unsafeDigitOfLargeInt: magnitude at: firstDigitIx) bitAnd: firstMask)
			~= 0
		ifTrue: [^ true].
	firstDigitIx + 1
		to: lastDigitIx - 1
		do: [:ix | (self unsafeDigitOfLargeInt: magnitude at: ix)
					~= 0
				ifTrue: [^ true]].
	((self unsafeDigitOfLargeInt: magnitude at: lastDigitIx)  bitAnd: lastMask)
			~= 0
		ifTrue: [^ true].
	^ false! !

!LargeIntegersPlugin methodsFor: 'util' stamp: 'nice 4/5/2016 22:57'!
byteSizeOfCSI: csi 
	"Answer the number of bytes required to represent the value of a C-SmallInteger."
	csi >= 0 ifTrue:
		[csi < 256 ifTrue:
			[^1].
		 csi < 65536 ifTrue:
			[^2].
		 csi < 16777216 ifTrue:
			[^3].
		 self cppIf: interpreterProxy bytesPerOop = 4
			ifTrue:
				[^4]
			ifFalse:
				[csi < 4294967296 ifTrue:
					[^4].
				 csi < 1099511627776 ifTrue:
					[^5].
				 csi < 281474976710656 ifTrue:
					[^6].
				 csi < 72057594037927936 ifTrue:
					[^7].
				 ^8]].
	csi > -256 ifTrue:
		[^1].
	csi > -65536 ifTrue:
		[^2].
	csi > -16777216 ifTrue:
		[^3].
	self cppIf: interpreterProxy bytesPerOop = 4
		ifTrue:
			[^4]
		ifFalse:
			[csi > -4294967296 ifTrue:
				[^4].
			 csi > -1099511627776 ifTrue:
				[^5].
			 csi > -281474976710656 ifTrue:
				[^6].
			 csi > -72057594037927936 ifTrue:
				[^7].
			^8]! !

!LargeIntegersPlugin methodsFor: 'C core util' stamp: 'nice 4/6/2016 01:26'!
byteSwapped32IfBigEndian: anInteger
	"This is only for simulation purposes"
	<doNotGenerate>
	^interpreterProxy byteSwapped32IfBigEndian: anInteger! !

!LargeIntegersPlugin methodsFor: 'C core util' stamp: 'nice 4/6/2016 04:19'!
byteSwapped64IfBigEndian: anInteger
	"This is only for simulation purposes"
	<doNotGenerate>
	^interpreterProxy byteSwapped64IfBigEndian: anInteger! !

!LargeIntegersPlugin methodsFor: 'C core' stamp: 'nice 4/6/2016 03:25'!
cDigitAdd: pWordShort len: shortLen with: pWordLong len: longLen into: pWordRes 
	"pWordRes len = longLen; returns over.."
	| accum |
	<returnTypeC: #'unsigned int'>
	<var: #pWordShort type: #'unsigned int *'>
	<var: #pWordLong type: #'unsigned int *'>
	<var: #pWordRes type: #'unsigned int *'>
	<var: #accum type: #'unsigned long long'>
	accum := 0.
	0 to: shortLen - 1 do: 
		[:i | 
		accum := (accum >> 32)
					+ (self cDigitOf: pWordShort at: i)
					+ (self cDigitOf: pWordLong at: i).
		self cDigitOf: pWordRes at: i put: (accum bitAnd: 16rFFFFFFFF)].
	shortLen to: longLen - 1 do: 
		[:i | 
		accum := (accum >> 32)
					+ (self cDigitOf: pWordLong at: i).
		self cDigitOf: pWordRes at: i put: (accum bitAnd: 16rFFFFFFFF)].
	^ accum >> 32! !

!LargeIntegersPlugin methodsFor: 'C core' stamp: 'nice 4/6/2016 03:25'!
cDigitCompare: pFirst with: pSecond len: len 
	"Precondition: pFirst len = pSecond len."
	| secondDigit ix firstDigit |
	<var: #pFirst type: #'unsigned int *'>
	<var: #pSecond type: #'unsigned int *'>
	<var: #firstDigit type: #'unsigned int'>
	<var: #secondDigit type: #'unsigned int'>
	ix := len - 1.
	[ix >= 0]
		whileTrue: 
			[(secondDigit := pSecond at: ix) ~= (firstDigit := pFirst at: ix)
				ifTrue: [(self byteSwapped32IfBigEndian: secondDigit) < (self byteSwapped32IfBigEndian: firstDigit)
						ifTrue: [^ 1]
						ifFalse: [^ -1]].
			ix := ix - 1].
	^ 0! !

!LargeIntegersPlugin methodsFor: 'C core util' stamp: 'nice 4/6/2016 03:31'!
cDigitCopyFrom: pFrom to: pTo len: len 
	<var: #pFrom type: #'unsigned int *'>
	<var: #pTo type: #'unsigned int *'>

	self cCode: '' inSmalltalk: [
		(interpreterProxy isKindOf: InterpreterSimulator) ifTrue: [
			"called from InterpreterSimulator"
				0 to: (len - 1) * 4 do: [:i |
					interpreterProxy byteAt: pTo + i
						put: (interpreterProxy byteAt: pFrom + i)
				].
			^ 0
		].
	].	
	"Note: don't care about endianness here, copy operation is endian neutral"
	0 to: len - 1 do: [:i | pTo at: i put: (pFrom at: i)].
	^ 0! !

!LargeIntegersPlugin methodsFor: 'C core' stamp: 'nice 11/11/2016 17:21'!
cDigitDiv: pDiv len: divLen rem: pRem len: remLen quo: pQuo len: quoLen 
	| dl ql dh dnh j t hi lo r3 l a b cond q r1r2 mul |
	<var: #pDiv type: #'unsigned int *'>
	<var: #pRem type: #'unsigned int *'>
	<var: #pQuo type: #'unsigned int *'>
	<var: #dh type: #'unsigned int'>
	<var: #dnh type: #'unsigned int'>
	<var: #r3 type: #'unsigned int'>
	<var: #q type: #'unsigned long long'>
	<var: #a type: #'unsigned long long'>
	<var: #b type: #'unsigned long long'>
	<var: #t type: #'unsigned long long'>
	<var: #mul type: #'unsigned long long'>
	<var: #hi type: #'unsigned long long'>
	<var: #lo type: #'unsigned long long'>
	<var: #r1r2 type: #'unsigned long long'>
	dl := divLen - 1.
	"Last actual byte of data (ST ix)"
	ql := quoLen.
	dh := self cDigitOf: pDiv at: dl - 1.
	dl = 1
		ifTrue: [dnh := 0]
		ifFalse: [dnh := self cDigitOf: pDiv at: dl - 2].
	1 to: ql do: 
		[:k | 
		"maintain quo*arg+rem=self"
		"Estimate rem/div by dividing the leading two unint32 of rem by dh."
		"The estimate is q = qhi*16r100000000+qlo, where qhi and qlo are uint32."
		j := remLen + 1 - k.
		"r1 := rem digitAt: j."
		(self cDigitOf: pRem at: j - 1)
			= dh
			ifTrue: [q := 16rFFFFFFFF]
			ifFalse: 
				["Compute q = (r1,r2)//dh, t = (r1,r2)\\dh.
				Note that r1,r2 are uint64, not uint32."
				"r2 := (rem digitAt: j - 2)."
				r1r2 := self cDigitOf: pRem at: j - 1.
				r1r2 := (r1r2 << 32) + (self cDigitOf: pRem at: j - 2).
				t := r1r2 \\ dh.
				q := r1r2 // dh.
				"Next compute (hi,lo) := q*dnh"
				mul := q * dnh.
				hi := mul >> 32.
				lo := mul bitAnd: 16rFFFFFFFF.
				"Correct overestimate of q.                
				Max of 2 iterations through loop -- see Knuth vol. 2"
				j < 3
					ifTrue: [r3 := 0]
					ifFalse: [r3 := self cDigitOf: pRem at: j - 3].
				
				[(t < hi
					or: [t = hi and: [r3 < lo]])
					ifTrue: 
						["i.e. (t,r3) < (hi,lo)"
						q := q - 1.
						hi = 0 "since hi is unsigned we must have this guard"
							ifTrue: [cond := false]
							ifFalse:
								[lo < dnh
									ifTrue: 
										[hi := hi - 1.
										lo := lo + 16r100000000 - dnh]
									ifFalse:
										[lo := lo - dnh].
								cond := hi >= dh]]
					ifFalse: [cond := false].
				cond]
					whileTrue: [hi := hi - dh]].
		"Subtract q*div from rem"
		l := j - dl.
		a := 0.
		1 to: divLen do: 
			[:i | 
			hi := (self cDigitOf: pDiv at: i - 1) * (q >> 32).
			lo := (self cDigitOf: pDiv at: i - 1) * (q bitAnd: 16rFFFFFFFF).
			b := (self cDigitOf: pRem at: l - 1) - a - (lo bitAnd: 16rFFFFFFFF).
			self cDigitOf: pRem at: l - 1 put: (b bitAnd: 16rFFFFFFFF).
			"simulate arithmetic shift (preserving sign of b)"
			b := b >> 32 bitOr: (0 - (b >> 63) bitAnd: 16rFFFFFFFF00000000).
			a := hi + (lo >> 32) - b.
			l := l + 1].
		a > 0
			ifTrue: 
				["Add div back into rem, decrease q by 1"
				q := q - 1.
				l := j - dl.
				a := 0.
				1 to: divLen do: 
					[:i | 
					a := (a >> 32)
								+ (self cDigitOf: pRem at: l - 1) + (self cDigitOf: pDiv at: i - 1).
					self cDigitOf: pRem at: l - 1 put: (a bitAnd: 16rFFFFFFFF).
					l := l + 1]].
		self cDigitOf: pQuo at: quoLen - k put: (self cCoerceSimple: q to: #'unsigned int')].
	^0! !

!LargeIntegersPlugin methodsFor: 'C core util' stamp: 'nice 4/6/2016 03:23'!
cDigitHighBit: pUint32 len: len 
	"Answer the index (in bits) of the high order bit of the receiver, or zero if the    
	 receiver is zero. This method is allowed (and needed) for     
	LargeNegativeIntegers as well, since Squeak's LargeIntegers are sign/magnitude.
	Work with 32 bits digits."
	| realLength lastDigit |
	<var: #pUint32 type: #'unsigned int *'>
	<var: #lastDigit type: #'unsigned int'>
	realLength := len.
	[realLength = 0 ifTrue: [^0].
	(lastDigit := self cDigitOf: pUint32 at: (realLength := realLength - 1)) = 0]
		whileTrue.
	^  (self cHighBit32: lastDigit) + (32 * realLength)! !

!LargeIntegersPlugin methodsFor: 'C core' stamp: 'nice 4/6/2016 03:26'!
cDigitLshift: shiftCount from: pFrom len: lenFrom to: pTo len: lenTo 
	"C indexed!!"
	| digitShift bitShift carry limit digit rshift |
	<var: #pTo type: #'unsigned int *'>
	<var: #pFrom type: #'unsigned int *'>
	<var: #carry type: #'unsigned int'>
	<var: #digit type: #'unsigned int'>
	digitShift := shiftCount // 32.
	bitShift := shiftCount \\ 32.
	
	limit := digitShift - 1.
	"Note: 0 is endian neutral, use direct access"
	0 to: limit do: [:i | pTo at: i put: 0].
	
	bitShift = 0 ifTrue: ["Fast version for digit-aligned shifts"
		"C indexed!!"
		^ self
			cDigitReplace: pTo
			from: digitShift
			to: lenTo - 1
			with: pFrom
			startingAt: 0].
		
	"This implementation use at most 31 bits of carry.
	bitAnd: 16rFFFFFFFF is only for simulator, useless in C"
	rshift := 32 - bitShift.
	carry := 0.
	limit := lenFrom - 1.
	0 to: limit do: 
		[:i | 
		digit := self cDigitOf: pFrom at: i.
		self cDigitOf: pTo at: i + digitShift put: ((carry bitOr: digit << bitShift) bitAnd: 16rFFFFFFFF).
		carry := digit >> rshift].
	carry = 0 ifFalse: [self cDigitOf: pTo at: lenTo - 1 put: carry].
	^0! !

!LargeIntegersPlugin methodsFor: 'C core' stamp: 'nice 4/6/2016 03:26'!
cDigitMontgomery: pFirst
				len: firstLen
				times: pSecond
				len: secondLen
				modulo: pThird
				len: thirdLen
				mInvModB: mInv
				into: pRes
				
	| u limit1 limit2 limit3 accum accum2 accum3 lastDigit |
	<var: #pFirst type: #'unsigned int *'>
	<var: #pSecond type: #'unsigned int *'>
	<var: #pThird type: #'unsigned int *'>
	<var: #mInv type: #'unsigned int'>
	<var: #pRes type: #'unsigned int *'>
	<var: #accum type: #'unsigned long long'>
	<var: #accum2 type: #'unsigned long long'>
	<var: #accum3 type: #'unsigned long long'>
	<var: #u type: #'unsigned long long '>
	<var: #lastDigit type: #'unsigned int '>
	limit1 := firstLen - 1.
	limit2 := secondLen - 1.
	limit3 := thirdLen - 1.
	lastDigit := 0.
	0 to: limit1 do: 
		[:i | 
		accum3 := self cDigitOf: pFirst at: i.
		accum3 := accum3*(self cDigitOf: pSecond at: 0) + (self cDigitOf: pRes at: 0).
		u := accum3 * mInv bitAnd: 16rFFFFFFFF.
		accum2 :=  u * (self cDigitOf: pThird at: 0).
		accum := (accum2 bitAnd: 16rFFFFFFFF) + (accum3 bitAnd: 16rFFFFFFFF).
		accum := (accum >> 32) + (accum2 >> 32) + (accum3 >> 32).
		1 to: limit2 do: [:k |
			accum3 := self cDigitOf: pFirst at: i.
			accum3 := accum3*(self cDigitOf: pSecond at: k) + (self cDigitOf: pRes at: k).
			accum2 :=  u * (self cDigitOf: pThird at: k).
			accum := accum + (accum2 bitAnd: 16rFFFFFFFF) + (accum3 bitAnd: 16rFFFFFFFF).
			self cDigitOf: pRes at: k-1 put: (accum bitAnd: 16rFFFFFFFF).
			accum := (accum >> 32) + (accum2 >> 32) + (accum3 >> 32)].
		secondLen to: limit3 do: [:k |
			accum2 :=  u * (self cDigitOf: pThird at: k).
			accum := accum + (self cDigitOf: pRes at: k) + (accum2 bitAnd: 16rFFFFFFFF).
			self cDigitOf: pRes at: k-1 put: (accum bitAnd: 16rFFFFFFFF).
			accum := (accum >> 32) + (accum2 >> 32)].
		accum := accum + lastDigit.
		self cDigitOf: pRes at: limit3 put: (accum bitAnd: 16rFFFFFFFF).
		lastDigit := accum >> 32].
	firstLen to: limit3 do: 
		[:i | 
		accum := (self cDigitOf: pRes at: 0).
		u := accum * mInv bitAnd: 16rFFFFFFFF.
		accum := accum + (u * (self cDigitOf: pThird at: 0)).
		accum := accum >> 32.
		1 to: limit3 do: [:k |
			accum2 :=  u * (self cDigitOf: pThird at: k).
			accum := accum + (self cDigitOf: pRes at: k) + (accum2 bitAnd: 16rFFFFFFFF).
			self cDigitOf: pRes at: k-1 put: (accum bitAnd: 16rFFFFFFFF).
			accum := (accum >> 32) + (accum2 >> 32)].
		accum := accum + lastDigit.
		self cDigitOf: pRes at: limit3 put: (accum bitAnd: 16rFFFFFFFF).
		lastDigit := accum >> 32].
	(lastDigit = 0 and: [(self cDigitCompare: pThird with: pRes len: thirdLen) = 1]) ifFalse: [
		"self cDigitSub: pThird len: thirdLen with: pRes len: thirdLen into: pRes"
		accum := 0.
		0 to: limit3 do: 
			[:i | 
			accum := accum + (self cDigitOf: pRes at: i) - (self cDigitOf: pThird at: i).
			self cDigitOf: pRes at: i put: (accum bitAnd: 16rFFFFFFFF).
			accum := 0 - (accum >> 63)]].
	^0! !

!LargeIntegersPlugin methodsFor: 'C core' stamp: 'nice 4/6/2016 03:27'!
cDigitMultiply: pWordShort len: shortLen with: pWordLong len: longLen into: pWordRes len: resLen
	| limitLong digit k carry limitShort ab |
	<returnTypeC: #'unsigned int'>
	<var: #pWordShort type: #'unsigned int *'>
	<var: #pWordLong type: #'unsigned int *'>
	<var: #pWordRes type: #'unsigned int *'>
	<var: #digit type: #'unsigned int'>
	<var: #carry type: #'unsigned int'>
	<var: #ab type: #'unsigned long long'>
	(shortLen = 1 and: [(pWordShort at: 0) = 0])
		ifTrue: [^ 0].
	(longLen = 1 and: [(pWordLong at: 0) = 0])
		ifTrue: [^ 0].
	"prod starts out all zero"
	limitShort := shortLen - 1.
	limitLong := longLen - 1.
	0 to: limitShort do: [:i | (digit := self cDigitOf: pWordShort at: i) ~= 0
			ifTrue: 
				[k := i.
				carry := 0.
				"Loop invariant: 0<=carry<=16rFFFFFFFF, k=i+j-1 (ST)"
				"-> Loop invariant: 0<=carry<=16rFFFFFFFF, k=i+j (C) (?)"
				0 to: limitLong do: 
					[:j | 
					ab := (self cDigitOf: pWordLong at: j).
					ab := ab * digit + carry + (self cDigitOf: pWordRes at: k).
					carry := ab >> 32.
					self cDigitOf: pWordRes at: k put: (ab bitAnd: 16rFFFFFFFF).
					k := k + 1].
				k < resLen ifTrue: [self cDigitOf: pWordRes at: k put: carry]]].
	^ 0! !

!LargeIntegersPlugin methodsFor: 'C core util' stamp: 'eem 4/6/2016 16:56'!
cDigitOf: cPointer at: zeroBasedDigitIndex
	<inline: true>
	<returnTypeC: #'unsigned int'>
	<var: 'cPointer' type: #'unsigned int *'>
	^self byteSwapped32IfBigEndian: ((self cCode: [cPointer] inSmalltalk: [interpreterProxy cCoerce: cPointer to: #'unsigned int *']) at: zeroBasedDigitIndex)! !

!LargeIntegersPlugin methodsFor: 'C core util' stamp: 'eem 4/6/2016 16:58'!
cDigitOf: cPointer at: zeroBasedDigitIndex put: aValue
	<inline: true>
	<returnTypeC: #'unsigned int'>
	<var: 'cPointer' type: #'unsigned int *'>
	<var: 'aValue' type: #'unsigned int'>
	^(self cCode: [cPointer] inSmalltalk: [interpreterProxy cCoerce: cPointer to: #'unsigned int *'])
		at: zeroBasedDigitIndex
		put: (self byteSwapped32IfBigEndian: aValue)! !

!LargeIntegersPlugin methodsFor: 'C core' stamp: 'nice 4/6/2016 03:27'!
cDigitOp: opIndex short: pWordShort len: shortLen long: pWordLong len: longLen into: pWordRes 
	"pWordRes len = longLen.
	NOTE: we don't bother with endianness here, those bit opes are endian-neutral"
	| limit |
	<var: #pWordShort type: #'unsigned int *'>
	<var: #pWordLong type: #'unsigned int *'>
	<var: #pWordRes type: #'unsigned int *'>
	limit := shortLen - 1.
	opIndex = andOpIndex
		ifTrue: 
			[0 to: limit do: [:i | pWordRes at: i put: ((pWordShort at: i)
						bitAnd: (pWordLong at: i))].
			limit := longLen - 1.
			shortLen to: limit do: [:i | pWordRes at: i put: 0].
			^ 0].
	opIndex = orOpIndex
		ifTrue: 
			[0 to: limit do: [:i | pWordRes at: i put: ((pWordShort at: i)
						bitOr: (pWordLong at: i))].
			limit := longLen - 1.
			shortLen to: limit do: [:i | pWordRes at: i put: (pWordLong at: i)].
			^ 0].
	opIndex = xorOpIndex
		ifTrue: 
			[0 to: limit do: [:i | pWordRes at: i put: ((pWordShort at: i)
						bitXor: (pWordLong at: i))].
			limit := longLen - 1.
			shortLen to: limit do: [:i | pWordRes at: i put: (pWordLong at: i)].
			^ 0].
	^ interpreterProxy primitiveFail! !

!LargeIntegersPlugin methodsFor: 'C core util' stamp: 'nice 4/6/2016 03:30'!
cDigitReplace: pTo from: start to: stop with: pFrom startingAt: repStart 
	"C indexed!!"
	<var: #pTo type: #'unsigned int *'>
	<var: #pFrom type: #'unsigned int *'>
	^ self
		cDigitCopyFrom: pFrom + repStart
		to: pTo + start
		len: stop - start + 1! !

!LargeIntegersPlugin methodsFor: 'C core' stamp: 'nice 4/6/2016 03:27'!
cDigitRshift: shiftCount from: pFrom len: lenFrom to: pTo len: lenTo 
	| carry digit digitShift bitShift leftShift limit start |
	<var: #pTo type: #'unsigned int *'>
	<var: #pFrom type: #'unsigned int *'>
	<var: #carry type: #'unsigned int'>
	<var: #digit type: #'unsigned int'>
	digitShift := shiftCount // 32.
	bitShift := shiftCount \\ 32.
	bitShift = 0 ifTrue: ["Fast version for digit-aligned shifts"
		"C indexed!!"
		^self
			cDigitReplace: pTo
			from: 0
			to: lenTo - 1
			with: pFrom
			startingAt: digitShift].
		
	"This implementation use at most 31 bits of carry.
	bitAnd: 16rFFFFFFFF is only for simulator, useless in C"
	leftShift := 32 - bitShift.
	carry := (self cDigitOf: pFrom at: digitShift) >> bitShift.
	start := digitShift + 1.
	limit := lenFrom - 1.
	start to: limit do: 
		[:j | 
		digit := self cDigitOf: pFrom at: j.
		self cDigitOf: pTo at: j - start put: ((carry bitOr: digit << leftShift) bitAnd: 16rFFFFFFFF).
		carry := digit >> bitShift].
	carry = 0 ifFalse: [self cDigitOf: pTo at: lenTo - 1 put: carry].
	^0! !

!LargeIntegersPlugin methodsFor: 'C core' stamp: 'eem 12/21/2017 14:00'!
cDigitSub: pWordSmall
		len: smallLen
		with: pWordLarge
		len: largeLen
		into: pWordRes
	| z |
	<var: #pWordSmall type: #'unsigned int *'>
	<var: #pWordLarge type: #'unsigned int *'>
	<var: #pWordRes type: #'unsigned int *'>
	<var: #z type: #'unsigned long long'>

	z := 0.
	0 to: smallLen - 1 do: 
		[:i | 
		z := z + (self cDigitOf: pWordLarge at: i) - (self cDigitOf: pWordSmall at: i).
		self cDigitOf: pWordRes at: i put: (z bitAnd: 16rFFFFFFFF).
		z := 0 - (self cCode: [z >> 63] inSmalltalk: [z >> 63 bitAnd: 1])].
	smallLen to: largeLen - 1 do: 
		[:i | 
		z := z + (self cDigitOf: pWordLarge at: i) .
		self cDigitOf: pWordRes at: i put: (z bitAnd: 16rFFFFFFFF).
		z := 0 - (self cCode: [z >> 63] inSmalltalk: [z >> 63 bitAnd: 1])].
	^0! !

!LargeIntegersPlugin methodsFor: 'C core util' stamp: 'nice 4/6/2016 03:31'!
cHighBit32: anUnsignedInt32 
	"Answer the index of the high order bit of the argument, or zero if the  
	argument is zero."
	| shifted bitNo |

	<var: #anUnsignedInt32 type: #'unsigned int'>
	<var: #shifted type: #'unsigned int'>
	shifted := anUnsignedInt32.
	bitNo := 0.
	shifted < (1 << 16)
		ifFalse: [shifted := shifted >> 16.
			bitNo := bitNo + 16].
	shifted < (1 << 8)
		ifFalse: [shifted := shifted >> 8.
			bitNo := bitNo + 8].
	shifted < (1 << 4)
		ifFalse: [shifted := shifted >> 4.
			bitNo := bitNo + 4].
	shifted < (1 << 2)
		ifFalse: [shifted := shifted >> 2.
			bitNo := bitNo + 2].
	shifted < (1 << 1)
		ifFalse: [shifted := shifted >> 1.
			bitNo := bitNo + 1].
	"shifted 0 or 1 now"
	^ bitNo + shifted! !

!LargeIntegersPlugin methodsFor: 'oop util' stamp: 'nice 4/9/2016 03:54'!
createLargeFromSmallInteger: anOop 
	"anOop has to be a SmallInteger!!"
	| val res pDigit byteSize digitSize |
	<var: #pDigit type: #'unsigned int *'>
	val := interpreterProxy integerValueOf: anOop.
	byteSize := self byteSizeOfCSI: val.
	res := self createLargeIntegerNeg: val < 0 byteLength: byteSize.
	pDigit := self pointerToFirstDigitOfLargeInt: res.
	digitSize := byteSize + 3 // 4.
	1 to: digitSize do: [:ix | self cDigitOf: pDigit at: ix - 1 put: (self digitOfCSI: val at: ix)].
	^ res! !

!LargeIntegersPlugin methodsFor: 'util' stamp: 'eem 8/10/2017 15:30'!
createLargeIntegerNeg: neg byteLength: byteLength
	<inline: #always>
	^interpreterProxy
		instantiateClass: (neg
				ifTrue: [interpreterProxy classLargeNegativeInteger]
				ifFalse: [interpreterProxy classLargePositiveInteger])
		indexableSize: byteLength! !

!LargeIntegersPlugin methodsFor: 'util' stamp: 'eem 8/10/2017 15:30'!
createLargeIntegerNeg: neg digitLength: digitLength
	<inline: #always>
	^self createLargeIntegerNeg: neg byteLength: digitLength * 4! !

!LargeIntegersPlugin methodsFor: 'oop functions' stamp: 'eem 8/10/2017 12:51'!
digit: anOop Lshift: shiftCount 
	"Attention: this method invalidates all oop's!! Only newOop is valid at return."
	"Does not normalize."
	| newOop highBit newDigitLen newByteLen oldDigitLen |
	oldDigitLen := self digitSizeOfLargeInt: anOop.
	(highBit := self cDigitHighBit: (self pointerToFirstDigitOfLargeInt: anOop)
				len: oldDigitLen) = 0 ifTrue: [^  interpreterProxy instantiateClass: (interpreterProxy fetchClassOf: anOop) indexableSize: 1].
	newByteLen := highBit + shiftCount + 7 // 8.
	self remapOop: anOop in:
		[newOop := interpreterProxy
						instantiateClass: (interpreterProxy fetchClassOf: anOop)
						indexableSize: newByteLen].
	newOop ifNil:
		[interpreterProxy primitiveFailFor: PrimErrNoMemory.
		 ^nil].
	newDigitLen := newByteLen + 3 // 4.
	self
		cDigitLshift: shiftCount
		from: (self pointerToFirstDigitOfLargeInt: anOop)
		len: oldDigitLen
		to: (self pointerToFirstDigitOfLargeInt: newOop)
		len: newDigitLen.
	^ newOop! !

!LargeIntegersPlugin methodsFor: 'oop functions' stamp: 'eem 6/1/2016 17:44'!
digit: anOop Rshift: shiftCount lookfirst: a 
	"Attention: this method invalidates all oop's!! Only newBytes is valid at return."
	"Shift right 32*digitShift+bitShift bits, 0<=bitShift<32.         
	Discard all digits beyond a, and all zeroes at or below a."
	"Does not normalize."
	| newByteLen newOop oldBitLen newBitLen oldDigitLen newDigitLen |
	oldBitLen := self cDigitHighBit: (self pointerToFirstDigitOfLargeInt: anOop) len: a.
	oldDigitLen := oldBitLen + 31 // 32.
	newBitLen := oldBitLen - shiftCount.
	newBitLen <= 0 ifTrue: ["All bits lost"
		^ interpreterProxy
			instantiateClass: (interpreterProxy fetchClassOf: anOop)
			indexableSize: 0].
	newByteLen := newBitLen + 7 // 8.
	newDigitLen := newByteLen + 3 // 4.
	self remapOop: anOop in:
		[newOop := interpreterProxy
						instantiateClass: (interpreterProxy fetchClassOf: anOop)
						indexableSize: newByteLen].
	newOop ifNil: [^interpreterProxy primitiveFailFor: PrimErrNoMemory].
	self
		cDigitRshift: shiftCount
		from: (self pointerToFirstDigitOfLargeInt: anOop)
		len: oldDigitLen
		to: (self pointerToFirstDigitOfLargeInt: newOop)
		len: newDigitLen.
	^ newOop! !

!LargeIntegersPlugin methodsFor: 'oop functions' stamp: 'eem 6/1/2016 17:40'!
digitAddLarge: firstInteger with: secondInteger 
	"Does not need to normalize!!"
	| over firstDigitLen secondDigitLen shortInt shortDigitLen longInt longDigitLen sum newSum neg |
	<var: #over type: #'unsigned int'>
	firstDigitLen := self digitSizeOfLargeInt: firstInteger.
	secondDigitLen := self digitSizeOfLargeInt: secondInteger.
	neg := interpreterProxy isLargeNegativeIntegerObject: firstInteger.
	firstDigitLen <= secondDigitLen
		ifTrue: 
			[shortInt := firstInteger.
			shortDigitLen := firstDigitLen.
			longInt := secondInteger.
			longDigitLen := secondDigitLen]
		ifFalse: 
			[shortInt := secondInteger.
			shortDigitLen := secondDigitLen.
			longInt := firstInteger.
			longDigitLen := firstDigitLen].
	"	sum := Integer new: len neg: firstInteger negative."
	self remapOop: #(shortInt longInt ) in: [sum := self createLargeIntegerNeg: neg digitLength: longDigitLen].
	sum ifNil: [^interpreterProxy primitiveFailFor: PrimErrNoMemory].
	over := self
				cDigitAdd: (self pointerToFirstDigitOfLargeInt: shortInt)
				len: shortDigitLen
				with: (self pointerToFirstDigitOfLargeInt: longInt)
				len: longDigitLen
				into: (self pointerToFirstDigitOfLargeInt: sum).
	over > 0
		ifTrue: 
			["sum := sum growby: 1."
			self remapOop: sum in: [newSum := self createLargeIntegerNeg: neg byteLength: longDigitLen * 4 + 1].
			newSum ifNil: [^interpreterProxy primitiveFailFor: PrimErrNoMemory].
			self
				cDigitCopyFrom: (self pointerToFirstDigitOfLargeInt: sum)
				to: (self pointerToFirstDigitOfLargeInt: newSum)
				len: longDigitLen.
			sum := newSum.
			"C index!!"
			self cDigitOf: (self pointerToFirstDigitOfLargeInt: sum)
				at: longDigitLen put: over]
		ifFalse:
			[sum := neg 
				ifTrue: [self normalizeNegative: sum]
				ifFalse: [self normalizePositive: sum]].
	^ sum! !

!LargeIntegersPlugin methodsFor: 'oop functions' stamp: 'eem 7/27/2021 08:10'!
digitBitLogic: firstInteger with: secondInteger opIndex: opIx 
	"Bit logic here is only implemented for positive integers or Zero;
	if rec or arg is negative, it fails."
	| firstLarge secondLarge firstLen secondLen shortLen shortLarge longLen longLarge result |
	(interpreterProxy isIntegerObject: firstInteger)
		ifTrue: 
			[(interpreterProxy integerValueOf: firstInteger) < 0 ifTrue:
				[^interpreterProxy primitiveFail].
			"convert it to a not normalized LargeInteger"
			self remapOop: secondInteger in:
				[firstLarge := self createLargeFromSmallInteger: firstInteger]]
		ifFalse: 
			[(interpreterProxy isLargePositiveIntegerObject: firstInteger) ifFalse:
				[^interpreterProxy primitiveFail].
			firstLarge := firstInteger].
	(interpreterProxy isIntegerObject: secondInteger)
		ifTrue: 
			[(interpreterProxy integerValueOf: secondInteger) < 0 ifTrue:
				[^interpreterProxy primitiveFail].
			"convert it to a not normalized LargeInteger"
			self remapOop: firstLarge in:
				[secondLarge := self createLargeFromSmallInteger: secondInteger]]
		ifFalse: 
			[(interpreterProxy isLargePositiveIntegerObject: secondInteger) ifFalse:
				[^interpreterProxy primitiveFail].
			secondLarge := secondInteger].
	firstLen := interpreterProxy byteSizeOf: firstLarge.
	secondLen := interpreterProxy byteSizeOf: secondLarge.
	firstLen < secondLen
		ifTrue: 
			[shortLen := firstLen.
			shortLarge := firstLarge.
			longLen := secondLen.
			longLarge := secondLarge]
		ifFalse: 
			[shortLen := secondLen.
			shortLarge := secondLarge.
			longLen := firstLen.
			longLarge := firstLarge].
	self remapOop: #(shortLarge longLarge) in:
		[result := interpreterProxy instantiateClass: interpreterProxy classLargePositiveInteger indexableSize: longLen].
	result ifNil: [^interpreterProxy primitiveFailFor: PrimErrNoMemory].
	self
		cDigitOp: opIx
		short: (self pointerToFirstDigitOfLargeInt: shortLarge)
		len: shortLen + 3 // 4
		long: (self pointerToFirstDigitOfLargeInt: longLarge)
		len: longLen + 3 // 4
		into: (self pointerToFirstDigitOfLargeInt: result).
	interpreterProxy failed ifTrue: [^ 0].
	^self normalizePositive: result! !

!LargeIntegersPlugin methodsFor: 'oop functions' stamp: 'nice 4/9/2016 03:57'!
digitCompareLarge: firstInteger with: secondInteger 
	"Compare the magnitude of firstInteger with that of secondInteger.      
	Return a code of 1, 0, -1 for firstInteger >, = , < secondInteger"
	| firstDigitLen secondDigitLen |
	firstDigitLen := self digitSizeOfLargeInt: firstInteger.
	secondDigitLen := self digitSizeOfLargeInt: secondInteger.
	secondDigitLen ~= firstDigitLen
		ifTrue: [secondDigitLen > firstDigitLen
				ifTrue: [^ -1 asOop: SmallInteger]
				ifFalse: [^ 1 asOop: SmallInteger]].
	^ (self
		cDigitCompare: (self pointerToFirstDigitOfLargeInt: firstInteger)
		with: (self pointerToFirstDigitOfLargeInt: secondInteger)
		len: firstDigitLen)
		asOop: SmallInteger! !

!LargeIntegersPlugin methodsFor: 'oop functions' stamp: 'nice 2/18/2020 01:19'!
digitDivLarge: firstInteger with: secondInteger negative: neg 
	"Does not normalize."
	"Division by zero has to be checked in caller."
	| firstDigitLen secondDigitLen quoDigitLen d div rem quo result |
	firstDigitLen := self digitSizeOfLargeInt: firstInteger.
	secondDigitLen := self digitSizeOfLargeInt: secondInteger.
	quoDigitLen := firstDigitLen - secondDigitLen + 1.
	quoDigitLen <= 0
		ifTrue: 
			[self remapOop: firstInteger in: [result := interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 2].
			 result ifNotNil:
				[interpreterProxy stObject: result at: 1 put: (0 asOop: SmallInteger).
				 interpreterProxy stObject: result at: 2 put: firstInteger].
			^ result].
	"set rem and div to copies of firstInteger and secondInteger, respectively. 
	 However,  to facilitate use of Knuth's algorithm, multiply rem and div by 2
	 (that is, shift) until the high word of div is >=16r80000000"
	d := 32 - (self cHighBit32: (self unsafeDigitOfLargeInt: secondInteger at: secondDigitLen)).
	self remapOop: firstInteger
		in:	[div := self digit: secondInteger Lshift: d.
			 div ifNotNil:
				[div := self largeInt: div growTo: (self digitSizeOfLargeInt: div) + 1 * 4].
			 div ifNil: [^div]].
	self remapOop: div
		in:	[rem := self digit: firstInteger Lshift: d.
			 rem ifNotNil:
				[(self digitSizeOfLargeInt: rem) = firstDigitLen ifTrue:
					[rem := self largeInt: rem growTo: firstDigitLen + 1 * 4]].
			 rem ifNil: [^rem]].
	self remapOop: #(div rem)
		in: [quo := self createLargeIntegerNeg: neg digitLength: quoDigitLen.
			quo ifNil:
				[interpreterProxy primitiveFailFor: PrimErrNoMemory. ^quo]].
	self
		cDigitDiv: (self pointerToFirstDigitOfLargeInt: div)
		len: (self digitSizeOfLargeInt: div)
		rem: (self pointerToFirstDigitOfLargeInt: rem)
		len: (self digitSizeOfLargeInt: rem)
		quo: (self pointerToFirstDigitOfLargeInt: quo)
		len: (self digitSizeOfLargeInt: quo).
	self remapOop: quo
		in: [rem := self digit: rem Rshift: d lookfirst: (self digitSizeOfLargeInt: div) - 1].
	"^ Array with: quo with: rem"
	self remapOop: #(quo rem)
		in: [result := interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 2].
	result ifNotNil:
		[interpreterProxy stObject: result at: 1 put: quo.
		 interpreterProxy stObject: result at: 2 put: rem].
	^result! !

!LargeIntegersPlugin methodsFor: 'oop functions' stamp: 'eem 6/1/2016 17:46'!
digitMontgomery: firstLarge times: secondLarge modulo: thirdLarge mInvModB: mInv
	<var: #mInv type: #'unsigned int'>
	| firstLen secondLen thirdLen prod |
	firstLen := self digitSizeOfLargeInt: firstLarge.
	secondLen := self digitSizeOfLargeInt: secondLarge.
	thirdLen := self digitSizeOfLargeInt: thirdLarge.
	(firstLen <= thirdLen and: [secondLen <= thirdLen]) ifFalse: [^interpreterProxy primitiveFail].

	self remapOop: #(firstLarge secondLarge thirdLarge)
		in: [prod := interpreterProxy instantiateClass: interpreterProxy classLargePositiveInteger indexableSize: thirdLen * 4].
	prod ifNil: [^interpreterProxy primitiveFailFor: PrimErrNoMemory].
	self
		cDigitMontgomery: (self pointerToFirstDigitOfLargeInt: firstLarge)
		len: firstLen
		times: (self pointerToFirstDigitOfLargeInt: secondLarge)
		len: secondLen
		modulo: (self pointerToFirstDigitOfLargeInt: thirdLarge)
		len: thirdLen
		mInvModB: mInv
		into: (self pointerToFirstDigitOfLargeInt: prod).
	^self normalizePositive: prod! !

!LargeIntegersPlugin methodsFor: 'oop functions' stamp: 'eem 7/27/2021 08:09'!
digitMultiplyLarge: firstInteger with: secondInteger negative: neg 
	"Normalizes."
	| firstLen secondLen shortInt shortLen longInt longLen prod |
	firstLen := interpreterProxy byteSizeOf: firstInteger.
	secondLen := interpreterProxy byteSizeOf: secondInteger.
	firstLen <= secondLen
		ifTrue: 
			[shortInt := firstInteger.
			shortLen := firstLen.
			longInt := secondInteger.
			longLen := secondLen]
		ifFalse: 
			[shortInt := secondInteger.
			shortLen := secondLen.
			longInt := firstInteger.
			longLen := firstLen].
	self remapOop: #(shortInt longInt) in:
		[prod := self createLargeIntegerNeg: neg byteLength: longLen + shortLen].
	prod ifNil: [^interpreterProxy primitiveFailFor: PrimErrNoMemory].
	self
		cDigitMultiply: (self pointerToFirstDigitOfLargeInt: shortInt)
		len: shortLen + 3 // 4
		with: (self pointerToFirstDigitOfLargeInt: longInt)
		len: longLen + 3 // 4
		into: (self pointerToFirstDigitOfLargeInt: prod)
		len: longLen + shortLen + 3 // 4.
	^neg 
		ifTrue: [self normalizeNegative: prod]
		ifFalse: [self normalizePositive: prod]! !

!LargeIntegersPlugin methodsFor: 'util' stamp: 'nice 4/5/2016 22:55'!
digitOfCSI: csi at: ix 
	"Answer the value of a 32 bits digit in a C-SmallInteger."
	"ST indexed!!"
	^(csi < 0
		ifTrue:  [0 - csi]
		ifFalse: [csi]) >> (ix - 1 * 32)! !

!LargeIntegersPlugin methodsFor: 'util' stamp: 'nice 4/5/2016 22:41'!
digitOfLargeInt: aBytesOop at: ix 
	"Argument has to be aLargeInteger!!"
	ix > (self digitSizeOfLargeInt: aBytesOop)
		ifTrue: [^ 0]
		ifFalse: [^ self unsafeDigitOfLargeInt: aBytesOop at: ix]! !

!LargeIntegersPlugin methodsFor: 'util' stamp: 'nice 7/8/2016 02:26'!
digitSizeOfCSI: csi 
	"Answer the number of 32-bits fields of a C-SmallInteger. This value is 
	   the same as the largest legal subscript."
	^(interpreterProxy maxSmallInteger <= 16r3FFFFFFF)
		ifTrue: [1]
		ifFalse: [csi > 16rFFFFFFFF asIntegerPtr "conversion is not really needed here, but avoid generating a warning in 32bits, and harmless in 64bits"
			ifTrue: [2]
			ifFalse: [csi < -16rFFFFFFFF
				ifTrue: [2]
				ifFalse: [1]]]! !

!LargeIntegersPlugin methodsFor: 'util' stamp: 'eem 7/27/2021 08:10'!
digitSizeOfLargeInt: anOop
	"answer number of 32 bits digits of a Large Integer"
	<inline: true>
	^(interpreterProxy byteSizeOf: anOop) + 3 // 4! !

!LargeIntegersPlugin methodsFor: 'oop functions' stamp: 'eem 6/1/2016 17:48'!
digitSubLarge: firstInteger with: secondInteger 
	"Normalizes."
	| firstDigitLen secondDigitLen larger largeDigitLen smaller smallerDigitLen neg resDigitLen res firstNeg |
	firstNeg := interpreterProxy isLargeNegativeIntegerObject: firstInteger.
	firstDigitLen := self digitSizeOfLargeInt: firstInteger.
	secondDigitLen := self digitSizeOfLargeInt: secondInteger.
	firstDigitLen = secondDigitLen ifTrue: 
		[[firstDigitLen > 1
		  and: [(self unsafeDigitOfLargeInt: firstInteger at: firstDigitLen) = (self unsafeDigitOfLargeInt: secondInteger at: firstDigitLen)]]
			whileTrue: [firstDigitLen := firstDigitLen - 1].
		secondDigitLen := firstDigitLen].
	(firstDigitLen < secondDigitLen
	 or: [firstDigitLen = secondDigitLen
		 and: [(self unsafeDigitOfLargeInt: firstInteger at: firstDigitLen) < (self unsafeDigitOfLargeInt: secondInteger at: firstDigitLen)]])
		ifTrue: 
			[larger := secondInteger.
			largeDigitLen := secondDigitLen.
			smaller := firstInteger.
			smallerDigitLen := firstDigitLen.
			neg := firstNeg == false]
		ifFalse: 
			[larger := firstInteger.
			largeDigitLen := firstDigitLen.
			smaller := secondInteger.
			smallerDigitLen := secondDigitLen.
			neg := firstNeg].
	resDigitLen := largeDigitLen.
	self remapOop: #(smaller larger)
		in: [res := self createLargeIntegerNeg: neg digitLength: resDigitLen].
	res ifNil: [^interpreterProxy primitiveFailFor: PrimErrNoMemory].
	self
		cDigitSub: (self pointerToFirstDigitOfLargeInt: smaller)
		len: smallerDigitLen
		with: (self pointerToFirstDigitOfLargeInt: larger)
		len: largeDigitLen
		into: (self pointerToFirstDigitOfLargeInt: res).
	^neg 
		ifTrue: [self normalizeNegative: res]
		ifFalse: [self normalizePositive: res]! !

!LargeIntegersPlugin methodsFor: 'util' stamp: 'nice 4/9/2016 03:59'!
highBitOfLargeInt: anOop 
	<inline: true>
	^ self
		cDigitHighBit: (self pointerToFirstDigitOfLargeInt: anOop)
		len: (self digitSizeOfLargeInt: anOop)! !

!LargeIntegersPlugin methodsFor: 'ST initialize' stamp: 'sr (auto pragmas 12/08) 3/15/2000 00:57'!
initialize
	"Initializes ST constants; C's are set by class>>declareCVarsIn:."
	<returnTypeC: 'void'>
	self cCode: '"nothing to do here"'
		inSmalltalk: 
			[andOpIndex := 0.
			orOpIndex := 1.
			xorOpIndex := 2]! !

!LargeIntegersPlugin methodsFor: 'oop functions' stamp: 'eem 7/27/2021 08:10'!
isNormalized: aLargeInteger 
	| len |
	"Check for leading zero of LargeInteger"
	len := interpreterProxy byteSizeOf: aLargeInteger.
	len = 0 ifTrue:
		[^ false].
	(self unsafeByteOfLargeInt: aLargeInteger at: len) = 0 ifTrue:
		[^ false].
	^true! !

!LargeIntegersPlugin methodsFor: 'oop util' stamp: 'eem 8/10/2017 11:58'!
largeInt: aBytesObject growTo: newByteLen 
	"Attention: this method invalidates all oop's!! Only newBytes is valid at return."
	"Does not normalize."
	| newBytes oldDigitLen newDigitLen copyLen |
	self remapOop: aBytesObject in:
		[newBytes := interpreterProxy
						instantiateClass: (interpreterProxy fetchClassOf: aBytesObject)
						indexableSize: newByteLen.
		 newBytes ifNil:
			[interpreterProxy primitiveFailFor: PrimErrNoMemory. ^nil]].
	newDigitLen := newByteLen + 3 // 4.
	oldDigitLen := self digitSizeOfLargeInt: aBytesObject.
	copyLen := oldDigitLen < newDigitLen
					ifTrue: [oldDigitLen]
					ifFalse: [newDigitLen].
	self
		cDigitCopyFrom: (self pointerToFirstDigitOfLargeInt: aBytesObject)
		to: (self pointerToFirstDigitOfLargeInt: newBytes)
		len: copyLen.
	^newBytes! !

!LargeIntegersPlugin methodsFor: 'oop functions' stamp: 'nice 5/2/2020 12:02'!
normalize: aLargeInteger 
	"Check for leading zeroes and return shortened copy if so."
	(interpreterProxy isLargePositiveIntegerObject: aLargeInteger)
		ifTrue: [^ self normalizePositive: aLargeInteger]
		ifFalse: [^ self normalizeNegative: aLargeInteger]! !

!LargeIntegersPlugin methodsFor: 'oop functions' stamp: 'eem 7/27/2021 08:10'!
normalizeNegative: aLargeNegativeInteger 
	"Check for leading zeroes and return shortened copy if so."
	"First establish len = significant length."
	| val val2 sLen digitLen byteLen oldByteLen minVal |
	<var: #val type: #usqInt>
	<var: #val2 type: #usqInt>
	<var: #minVal type: #usqInt>
	digitLen := self digitSizeOfLargeInt: aLargeNegativeInteger.
	[digitLen ~= 0 and: [(self unsafeDigitOfLargeInt: aLargeNegativeInteger at: digitLen) = 0]]
		whileTrue: [digitLen := digitLen - 1].
	digitLen = 0 ifTrue: [^ 0 asOop: SmallInteger].
	"Now check if in SmallInteger range"
	val := self unsafeDigitOfLargeInt: aLargeNegativeInteger at: digitLen.
	sLen := interpreterProxy minSmallInteger < -16r40000000
				ifTrue: [2]
				ifFalse: [1]. "SmallInteger minVal digitLength"
	digitLen <= sLen
		ifTrue: 
			[minVal := 0 - interpreterProxy minSmallInteger.
			val2 := val.
			digitLen > 1 ifTrue: [val2 := val2 << 32 + (self unsafeDigitOfLargeInt: aLargeNegativeInteger at: 1)].
			val2 <= minVal
				ifTrue: [^0 -  val2 asOop: SmallInteger]].
	"Return self, or a shortened copy"
	byteLen := digitLen * 4.
	val <= 16rFFFF
		ifTrue: [byteLen := byteLen - 2]
		ifFalse: [val := val >> 16].
	val <= 16rFF
		ifTrue: [byteLen := byteLen - 1].
	oldByteLen := interpreterProxy byteSizeOf: aLargeNegativeInteger.
	byteLen < oldByteLen
		ifTrue: [^ self largeInt: aLargeNegativeInteger growTo: byteLen]
		ifFalse: [^ aLargeNegativeInteger]! !

!LargeIntegersPlugin methodsFor: 'oop functions' stamp: 'eem 7/27/2021 08:11'!
normalizePositive: aLargePositiveInteger 
	"Check for leading zeroes and return shortened copy if so."
	"First establish len = significant length."
	| val val2 sLen digitLen byteLen oldByteLen maxVal |
	<var: #val type: #usqInt>
	<var: #val2 type: #usqInt>
	<var: #maxVal type: #usqInt>
	digitLen := self digitSizeOfLargeInt: aLargePositiveInteger.
	[digitLen ~= 0 and: [(self unsafeDigitOfLargeInt: aLargePositiveInteger at: digitLen) = 0]]
		whileTrue: [digitLen := digitLen - 1].
	digitLen = 0 ifTrue: [^ 0 asOop: SmallInteger].
	"Now check if in SmallInteger range"
	val := self unsafeDigitOfLargeInt: aLargePositiveInteger at: digitLen.
	sLen := interpreterProxy maxSmallInteger > 16r3FFFFFFF
				ifTrue: [2]
				ifFalse: [1]. "SmallInteger maxVal digitLength"
	digitLen <= sLen
		ifTrue: 
			[maxVal := interpreterProxy maxSmallInteger.
			val2 := val.
			digitLen > 1 ifTrue:
				["Note: asUnsignedLongLong is not necessary because this branch is for 64 bits only.
				but we want to avoid a C Compiler warning on 32 bits"
				val2 := val2 asUnsignedLongLong << 32 + (self unsafeDigitOfLargeInt: aLargePositiveInteger at: 1)].
			val2 <= maxVal
				ifTrue: [^val2 asOop: SmallInteger]].
	"Return self, or a shortened copy"
	byteLen := digitLen * 4.
	val <= 16rFFFF
		ifTrue: [byteLen := byteLen - 2]
		ifFalse: [val := val >> 16].
	val <= 16rFF
		ifTrue: [byteLen := byteLen - 1].
	oldByteLen := interpreterProxy byteSizeOf: aLargePositiveInteger.
	byteLen < oldByteLen
		ifTrue: [^ self largeInt: aLargePositiveInteger growTo: byteLen]
		ifFalse: [^ aLargePositiveInteger]! !

!LargeIntegersPlugin methodsFor: 'util' stamp: 'eem 8/10/2017 15:25'!
pointerToFirstDigitOfLargeInt: bytesObject
	<inline: #always>
	<returnTypeC: #'unsigned int *'>
	^interpreterProxy cCoerce: (interpreterProxy firstIndexableField: bytesObject) to: #'unsigned int *'! !

!LargeIntegersPlugin methodsFor: 'Integer primitives' stamp: 'eem 7/22/2021 21:12'!
primAnyBitFrom: from to: to
	<export: true flags: #FastCPrimitiveFlag>
	| integer someBitIsSet val mask |
	integer := self primitive: 'primAnyBitFromTo' parameters: #(SmallInteger SmallInteger) receiver: #Integer.
	from < 1 | (to < 1) ifTrue:
		[^ interpreterProxy primitiveFail].
	(interpreterProxy isIntegerObject: integer)
		ifTrue: "For small integers, use a single bit mask operation"
			[from <= to
				ifTrue: 
					[val := interpreterProxy integerValueOf: integer.
					val < 0 ifTrue: ["Get the bits of magnitude" val := 0 - val].
					mask := (1 asUnsignedInteger << (to min: (self sizeof: #usqInt) * 8 - 1))
						- (1 asUnsignedInteger << (from - 1 min: (self sizeof: #usqInt) * 8 - 1)).
					someBitIsSet := val anyMask: mask]
				ifFalse: [someBitIsSet := 0]]
		ifFalse:
			[someBitIsSet := self anyBitOfLargeInt: integer from: from to: to].
	^someBitIsSet asOop: Boolean! !

!LargeIntegersPlugin methodsFor: 'Integer primitives' stamp: 'eem 7/22/2021 21:12'!
primDigitAdd: secondInteger
	<export: true flags: #FastCPrimitiveFlag>
	| firstLarge secondLarge firstInteger |
	firstInteger := self primitive: 'primDigitAdd' parameters: #(Integer) receiver: #Integer.
	(interpreterProxy isIntegerObject: firstInteger)
		ifTrue: ["convert it to a not normalized LargeInteger"
			self remapOop: secondInteger in: [firstLarge := self createLargeFromSmallInteger: firstInteger]]
		ifFalse: [firstLarge := firstInteger].
	(interpreterProxy isIntegerObject: secondInteger)
		ifTrue: ["convert it to a not normalized LargeInteger"
			self remapOop: firstLarge in: [secondLarge := self createLargeFromSmallInteger: secondInteger]]
		ifFalse: [secondLarge := secondInteger].
	^ self digitAddLarge: firstLarge with: secondLarge! !

!LargeIntegersPlugin methodsFor: 'Integer primitives' stamp: 'eem 7/22/2021 21:13'!
primDigitBitAnd: secondInteger 
	"Bit logic here is only implemented for positive integers or Zero; if rec 
	or arg is negative, it fails."
	<export: true flags: #FastCPrimitiveFlag>
	| firstInteger |
	firstInteger := self primitive: 'primDigitBitAnd' parameters: #(Integer) receiver: #Integer.
	^self
		digitBitLogic: firstInteger
		with: secondInteger
		opIndex: andOpIndex! !

!LargeIntegersPlugin methodsFor: 'Integer primitives' stamp: 'eem 7/22/2021 21:13'!
primDigitBitOr: secondInteger 
	"Bit logic here is only implemented for positive integers or Zero; if rec 
	or arg is negative, it fails."
	<export: true flags: #FastCPrimitiveFlag>
	| firstInteger |
	firstInteger := self primitive: 'primDigitBitOr' parameters: #(Integer) receiver: #Integer.
	^self
		digitBitLogic: firstInteger
		with: secondInteger
		opIndex: orOpIndex! !

!LargeIntegersPlugin methodsFor: 'Integer primitives' stamp: 'eem 7/22/2021 21:14'!
primDigitBitShiftMagnitude: shiftCount
	<export: true flags: #FastCPrimitiveFlag>
	| rShift aLarge anInteger |
	anInteger := self primitive: 'primDigitBitShiftMagnitude' parameters: #(SmallInteger) receiver: #Integer.
	(interpreterProxy isIntegerObject: anInteger)
		ifTrue: "convert it to a not normalized LargeInteger"
			[aLarge := self createLargeFromSmallInteger: anInteger]
		ifFalse: [aLarge := anInteger].
	shiftCount >= 0 ifTrue:
		[^ self digit: aLarge Lshift: shiftCount].
	rShift := 0 - shiftCount.
	^self normalize: (self
						digit: aLarge
						Rshift: rShift
						lookfirst: (self digitSizeOfLargeInt: aLarge))! !

!LargeIntegersPlugin methodsFor: 'Integer primitives' stamp: 'eem 7/22/2021 21:15'!
primDigitBitXor: secondInteger
	"Bit logic here is only implemented for positive integers or Zero; if rec 
	or arg is negative, it fails."
	<export: true flags: #FastCPrimitiveFlag>
	| firstInteger |
	firstInteger := self primitive: 'primDigitBitXor' parameters: #(Integer) receiver: #Integer.
	^self
		digitBitLogic: firstInteger
		with: secondInteger
		opIndex: xorOpIndex! !

!LargeIntegersPlugin methodsFor: 'Integer primitives' stamp: 'eem 7/22/2021 21:15'!
primDigitCompare: secondInteger 
	"Compare the magnitude of self with that of arg.   
	 Answer a code of 1, 0, -1 for self >, = , < arg"
	<export: true flags: #FastCPrimitiveFlag>
	| firstVal secondVal firstInteger |
	firstInteger := self primitive: 'primDigitCompare' parameters: #(Integer) receiver: #Integer.
	"shortcut: aSmallInteger has to be smaller in Magnitude than aLargeInteger"
	(interpreterProxy isIntegerObject: firstInteger) ifTrue:
		[(interpreterProxy isIntegerObject: secondInteger) ifTrue:
			[firstVal := interpreterProxy integerValueOf: firstInteger.
			 secondVal := interpreterProxy integerValueOf: secondInteger.
			 "Compute their magnitudes.  Since SmallIntegers are tagged they have
			  fewer bits than an integer on the platform; therefore in computing their
			  magnitude they cannot overflow."
			 firstVal < 0 ifTrue: [firstVal := 0 - firstVal].
			 secondVal < 0 ifTrue: [secondVal := 0 - secondVal].
			 ^firstVal = secondVal
				ifTrue: [0 asOop: SmallInteger]
				ifFalse:
					[firstVal < secondVal
						ifTrue: [-1 asOop: SmallInteger]
						ifFalse: [1 asOop: SmallInteger]]].
			^-1 asOop: SmallInteger]. "first < second"
	(interpreterProxy isIntegerObject: secondInteger) ifTrue:
		[^1 asOop: SmallInteger]. "first > second"
	^ self digitCompareLarge: firstInteger with: secondInteger! !

!LargeIntegersPlugin methodsFor: 'Integer primitives' stamp: 'eem 7/22/2021 21:17'!
primDigitDiv: secondInteger negative: neg 
	"Answer the result of dividing firstInteger by secondInteger. 
	Fail if parameters are not integers, not normalized or secondInteger is zero. "
	<export: true flags: #FastCPrimitiveFlag>
	| firstAsLargeInteger secondAsLargeInteger firstInteger |
	firstInteger := self primitive: 'primDigitDivNegative' parameters: #(Integer Boolean) receiver: #Integer.
	"Coerce SmallIntegers to corresponding (not normalized) large integers  
	and check for zerodivide."
	(interpreterProxy isIntegerObject: firstInteger)
		ifTrue: "convert to LargeInteger"
			[self
				remapOop: secondInteger
				in: [firstAsLargeInteger := self createLargeFromSmallInteger: firstInteger]]
		ifFalse:
			["Avoid crashes in case of getting unnormalized args."
			(self isNormalized: firstInteger)
				ifFalse:
					[self debugCode:
						[self msg: 'ERROR in primDigitDiv: secondInteger negative: neg'.
						self msg: '------> receiver *not* normalized!!'].
					^ interpreterProxy primitiveFail].
			firstAsLargeInteger := firstInteger].
	(interpreterProxy isIntegerObject: secondInteger)
		ifTrue: "check for zerodivide and convert to LargeInteger"
			[(interpreterProxy integerValueOf: secondInteger)
					= 0
				ifTrue: [^ interpreterProxy primitiveFail].
			self
				remapOop: firstAsLargeInteger
				in: [secondAsLargeInteger := self createLargeFromSmallInteger: secondInteger]]
		ifFalse:
			["Avoid crashes in case of getting unnormalized args."
			(self isNormalized: secondInteger) ifFalse:
				[self debugCode:
					[self msg: 'ERROR in primDigitDiv: secondInteger negative: neg'.
					self msg: '------> argument *not* normalized!!'].
				^ interpreterProxy primitiveFail].
			secondAsLargeInteger := secondInteger].
	^ self
		digitDivLarge: firstAsLargeInteger
		with: secondAsLargeInteger
		negative: neg! !

!LargeIntegersPlugin methodsFor: 'Integer primitives' stamp: 'eem 7/22/2021 21:17'!
primDigitMultiply: secondInteger negative: neg
	<export: true flags: #FastCPrimitiveFlag>
	| firstLarge secondLarge firstInteger |
	firstInteger := self primitive: 'primDigitMultiplyNegative' parameters: #(Integer Boolean) receiver: #Integer.
	(interpreterProxy isIntegerObject: firstInteger)
		ifTrue: "convert it to a not normalized LargeInteger"
			[self
				remapOop: secondInteger
				in: [firstLarge := self createLargeFromSmallInteger: firstInteger]]
		ifFalse: [firstLarge := firstInteger].
	(interpreterProxy isIntegerObject: secondInteger)
		ifTrue: "convert it to a not normalized LargeInteger"
			[self
				remapOop: firstLarge
				in: [secondLarge := self createLargeFromSmallInteger: secondInteger]]
		ifFalse: [secondLarge := secondInteger].
	^ self
		digitMultiplyLarge: firstLarge
		with: secondLarge
		negative: neg! !

!LargeIntegersPlugin methodsFor: 'Integer primitives' stamp: 'eem 7/22/2021 21:18'!
primDigitSubtract: secondInteger
	<export: true flags: #FastCPrimitiveFlag>
	| firstLarge secondLarge firstInteger |
	firstInteger := self primitive: 'primDigitSubtract' parameters: #(Integer) receiver: #Integer.
	(interpreterProxy isIntegerObject: firstInteger)
		ifTrue: "convert it to a not normalized LargeInteger"
			[self
				remapOop: secondInteger
				in: [firstLarge := self createLargeFromSmallInteger: firstInteger]]
		ifFalse: [firstLarge := firstInteger].
	(interpreterProxy isIntegerObject: secondInteger)
		ifTrue: "convert it to a not normalized LargeInteger"
			[self
				remapOop: firstLarge
				in: [secondLarge := self createLargeFromSmallInteger: secondInteger]]
		ifFalse: [secondLarge := secondInteger].
	^ self digitSubLarge: firstLarge with: secondLarge! !

!LargeIntegersPlugin methodsFor: 'control & support primitives' stamp: 'nice 5/2/2020 12:03'!
primGetModuleName
	"If calling this primitive fails, then C module does not exist."
	| strLen strOop |
	self
		primitive: 'primGetModuleName'
		parameters: #()
		receiver: #Oop.
	strLen := self strlen: self getModuleName.
	strOop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: strLen.
	self strncpy: (interpreterProxy firstIndexableField: strOop)
		_: self getModuleName
		_: strLen.
	^strOop! !

!LargeIntegersPlugin methodsFor: 'Integer primitives' stamp: 'eem 7/22/2021 21:21'!
primMontgomeryDigitLength
	<export: true flags: #FastCPrimitiveFlag>
	self primitive: #primMontgomeryDigitLength parameters: #() receiver: #Integer.
	^interpreterProxy integerObjectOf: 32! !

!LargeIntegersPlugin methodsFor: 'Integer primitives' stamp: 'eem 7/22/2021 21:19'!
primMontgomeryTimes: secondOperandInteger modulo: thirdModuloInteger mInvModB: mInverseInteger
	<export: true flags: #FastCPrimitiveFlag>
	| firstLarge secondLarge firstInteger thirdLarge mInv |
	<var: #mInv type: #'unsigned int'>
	firstInteger := self primitive: 'primMontgomeryTimesModulo' parameters: #(Integer Integer Integer) receiver: #Integer.
	 mInv := interpreterProxy positive32BitValueOf: mInverseInteger.
	(interpreterProxy isIntegerObject: firstInteger)
		ifTrue: "convert it to a not normalized LargeInteger"
			[self remapOop: #(secondOperandInteger thirdModuloInteger) in: [firstLarge := self createLargeFromSmallInteger: firstInteger]]
		ifFalse: [firstLarge := firstInteger].
	(interpreterProxy isIntegerObject: secondOperandInteger)
		ifTrue: "convert it to a not normalized LargeInteger"
			[self remapOop: #(firstLarge thirdModuloInteger) in: [secondLarge := self createLargeFromSmallInteger: secondOperandInteger]]
		ifFalse: [secondLarge := secondOperandInteger].
	(interpreterProxy isIntegerObject: thirdModuloInteger)
		ifTrue: "convert it to a not normalized LargeInteger"
			[self remapOop: #(firstLarge secondLarge) in: [thirdLarge := self createLargeFromSmallInteger: thirdModuloInteger]]
		ifFalse: [thirdLarge := thirdModuloInteger].
	^ self digitMontgomery: firstLarge times: secondLarge modulo: thirdLarge mInvModB: mInv! !

!LargeIntegersPlugin methodsFor: 'Integer primitives' stamp: 'eem 7/22/2021 21:20'!
primNormalizeNegative
	<export: true flags: #FastCPrimitiveFlag>
	| rcvr |
	rcvr := self primitive: #primNormalizeNegative parameters: #() receiver: #LargeNegativeInteger.
	^ self normalizeNegative: rcvr! !

!LargeIntegersPlugin methodsFor: 'Integer primitives' stamp: 'eem 7/22/2021 21:20'!
primNormalizePositive
	<export: true flags: #FastCPrimitiveFlag>
	| rcvr |
	rcvr := self primitive: #primNormalizePositive parameters: #() receiver: #LargePositiveInteger.
	^ self normalizePositive: rcvr! !

!LargeIntegersPlugin methodsFor: 'util' stamp: 'nice 4/9/2016 04:01'!
unsafeByteOfLargeInt: bytesObj at: ix
	"Argument bytesObj must not be aSmallInteger!!"
	<inline: true>
	<returnTypeC: #'unsigned char'>
	^(interpreterProxy cCoerce: (interpreterProxy firstIndexableField: bytesObj) to: #'unsigned char *') at: ix - 1! !

!LargeIntegersPlugin methodsFor: 'util' stamp: 'eem 8/10/2017 15:29'!
unsafeDigitOfLargeInt: anOop at: ix
	"Argument must not be aSmallInteger!!"
	<inline: #always>
	<returnTypeC: #'unsigned int'>
	^self cDigitOf: (self pointerToFirstDigitOfLargeInt: anOop) at: ix - 1! !

SmartSyntaxInterpreterPlugin subclass: #LocalePlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!LocalePlugin commentStamp: '<historical>' prior: 0!
LocalePlugin provides access to some localization info.
primLanguage - returns a string describing the language in use as per ISO 639
primCountry - returns a string with country tag as per ISO 639
primVMOffsetToUTC - returns offset from UTC to time as provided by the VM. integer of minutes to allow for those odd places with halkf-hour offeset.
primTimeZone - returns UTC offset (? why two?)
primDST - returns boolean to indicate DST in use
primDecimalSymbol - return string with '.' or ',' etc
primDigitGrouping - return string with ',' or '.' etc for thousands type separation
primTimeFormat - return string with time dispaly format string - eg 'hh:mm:ss' etc
primLongDateFOrmat - return string with long date formatting - eg 'dd/mm/yyyy'
primShortDateFOrmat - similar but shortform
primCurrencySymbol - return string of currency name
primCurrencyNotation - return boolean for pre or postfix currency symbol
primMeasurement - return boolean for imperial or metric

!

!LocalePlugin class methodsFor: 'translation' stamp: 'tpr 5/31/2005 18:41'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!LocalePlugin class methodsFor: 'translation' stamp: 'tpr 5/31/2005 17:00'!
requiresPlatformFiles
	"this plugin requires platform specific files in order to work"
	^true! !

!LocalePlugin methodsFor: 'initialize' stamp: 'tpr (auto pragmas 12/08) 6/1/2005 18:20'!
initialiseModule
	<export: true>
	^self sqLocInitialize! !

!LocalePlugin methodsFor: 'system primitives' stamp: 'tpr 5/31/2005 18:59'!
primitiveCountry
	"return a 3 char string describing the country in use. ISO 3166 is the relevant source here; see http://www.unicode.org/onlinedat/countries.html for details. Using the 3 character Alpha-3 codes"
	| oop |
	self primitive:'primitiveCountry'.
	oop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: 3.
	self sqLocGetCountryInto: (interpreterProxy firstIndexableField: oop).
	^oop
! !

!LocalePlugin methodsFor: 'system primitives' stamp: 'tpr 5/31/2005 18:25'!
primitiveCurrencyNotation
	"return a boolean specifying whether the currency symbol is pre or post fix. true -> pre"
	self primitive:'primitiveCurrencyNotation'.
	^self sqLocCurrencyNotation asOop: Boolean! !

!LocalePlugin methodsFor: 'system primitives' stamp: 'tpr 6/1/2005 13:22'!
primitiveCurrencySymbol
	"return a string describing the currency symbol used 
Still need to find details on standard symbols - ISO 4217 is supposed to be it but cannot find on web"
	| oop length |
	self primitive:'primitiveCurrencySymbol'.
	length := self sqLocCurrencySymbolSize.
	oop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: length.
	self sqLocGetCurrencySymbolInto: (interpreterProxy firstIndexableField: oop).
	^oop
! !

!LocalePlugin methodsFor: 'system primitives' stamp: 'tpr 5/31/2005 18:24'!
primitiveDaylightSavings
	"return a boolean specifying the DST setting. true -> active"
	self primitive:'primitiveDaylightSavings'.
	^self sqLocDaylightSavings asOop: Boolean! !

!LocalePlugin methodsFor: 'system primitives' stamp: 'tpr 6/1/2005 12:23'!
primitiveDecimalSymbol
	"return a 1 char string describing the decimal symbol used - usually a . or a ,"
	| oop |
	self primitive:'primitiveDecimalSymbol'.
	oop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: 1.
	self sqLocGetDecimalSymbolInto: (interpreterProxy firstIndexableField: oop).
	^oop
! !

!LocalePlugin methodsFor: 'system primitives' stamp: 'tpr 6/1/2005 12:24'!
primitiveDigitGroupingSymbol
	"return a 1 char string describing the digitGrouping symbol used - usually a . or a , between triples of digits"
	| oop |
	self primitive:'primitiveDigitGroupingSymbol'.
	oop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: 1.
	self sqLocGetDigitGroupingSymbolInto: (interpreterProxy firstIndexableField: oop).
	^oop
! !

!LocalePlugin methodsFor: 'system primitives' stamp: 'tpr 6/1/2005 12:24'!
primitiveLanguage
	"return a 3 char string describing the language in use. ISO 639 is the relevant source here; see http://www.w3.org/WAI/ER/IG/ert/iso639.html for details"
	| oop |
	self primitive:'primitiveLanguage'.
	oop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: 3.
	self sqLocGetLanguageInto: (interpreterProxy firstIndexableField: oop).
	^oop
! !

!LocalePlugin methodsFor: 'system primitives' stamp: 'tpr 6/1/2005 12:24'!
primitiveLongDateFormat
	"return a string describing the long date formatting.
Format is made up of
d day, m month, y year,
double symbol is null padded, single not padded (m=6, mm=06)
dddd weekday
mmmm month name
"
	| oop length |
	self primitive:'primitiveLongDateFormat'.
	length := self sqLocLongDateFormatSize.
	oop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: length.
	self sqLocGetLongDateFormatInto: (interpreterProxy firstIndexableField: oop).
	^oop
! !

!LocalePlugin methodsFor: 'system primitives' stamp: 'tpr 5/31/2005 18:28'!
primitiveMeasurementMetric
	"return a boolean specifying whether the currency symbol is pre or post fix. true -> pre"
	self primitive:'primitiveMeasurementMetric'.
	^self sqLocMeasurementMetric asOop: Boolean! !

!LocalePlugin methodsFor: 'system primitives' stamp: 'tpr 6/1/2005 12:25'!
primitiveShortDateFormat
	"return a string describing the long date formatting.
Format is made up of
d day, m month, y year,
double symbol is null padded, single not padded (m=6, mm=06)
dddd weekday
mmmm month name
"
	| oop length |
	self primitive:'primitiveShortDateFormat'.
	length := self sqLocShortDateFormatSize.
	oop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: length.
	self sqLocGetShortDateFormatInto: (interpreterProxy firstIndexableField: oop).
	^oop
! !

!LocalePlugin methodsFor: 'system primitives' stamp: 'tpr 6/1/2005 12:25'!
primitiveTimeFormat
	"return a string describing the time formatting.
Format is made up of
h hour (h 12, H 24), m minute, s seconds, x (am/pm String)
double symbol is null padded, single not padded (h=6, hh=06)"
	| oop length |
	self primitive:'primitiveTimeFormat'.
	length := self sqLocTimeFormatSize.
	oop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: length.
	self sqLocGetTimeFormatInto: (interpreterProxy firstIndexableField: oop).
	^oop
! !

!LocalePlugin methodsFor: 'system primitives' stamp: 'tpr 6/1/2005 19:43'!
primitiveTimezoneOffset
	"return the number of minutes this VM's time value is offset from UTC"

	self primitive:'primitiveTimezoneOffset'.
	^self sqLocGetTimezoneOffset asSmallIntegerObj
! !

!LocalePlugin methodsFor: 'system primitives' stamp: 'tpr 6/1/2005 19:04'!
primitiveVMOffsetToUTC
	"return the number of minutes this VM's time value is offset from UTC"

	self primitive:'primitiveVMOffsetToUTC'.
	^self sqLocGetVMOffsetToUTC asSmallIntegerObj
! !

!LocalePlugin methodsFor: 'simulation support' stamp: 'eem 10/25/2018 11:11'!
sqLocGetTimezoneOffset
	<doNotGenerate>
	^(DateAndTime localTimeZone offset asSeconds / 60) rounded! !

!LocalePlugin methodsFor: 'simulation support' stamp: 'eem 10/25/2018 11:11'!
sqLocInitialize
	<doNotGenerate>
	^true! !

SmartSyntaxInterpreterPlugin subclass: #MIDIPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!MIDIPlugin commentStamp: 'tpr 5/5/2003 12:15' prior: 0!
Provide MIDI support, if your platform provides it. !

!MIDIPlugin class methodsFor: 'translation' stamp: 'tpr 5/23/2001 17:10'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!MIDIPlugin class methodsFor: 'translation' stamp: 'tpr 11/29/2000 22:38'!
requiresPlatformFiles
	"this plugin requires platform specific files in order to work"
	^true! !

!MIDIPlugin methodsFor: 'initialize-release' stamp: 'ar (auto pragmas 12/08) 5/12/2000 16:53'!
initialiseModule
	<export: true>
	^self cCode: 'midiInit()' inSmalltalk:[true]! !

!MIDIPlugin methodsFor: 'primitives' stamp: 'ar 5/12/2000 17:14'!
primitiveMIDIClosePort: portNum

	self primitive: 'primitiveMIDIClosePort'
		parameters: #(SmallInteger).
	self sqMIDIClosePort: portNum! !

!MIDIPlugin methodsFor: 'primitives' stamp: 'ar 5/12/2000 17:14'!
primitiveMIDIGetClock
	"Return the value of the MIDI clock as a SmallInteger. The range is limited to SmallInteger maxVal / 2 to allow scheduling MIDI events into the future without overflowing a SmallInteger. The sqMIDIGetClock function is assumed to wrap at or before 16r20000000."

	| clockValue |
	self primitive: 'primitiveMIDIGetClock'.
	clockValue := self sqMIDIGetClock bitAnd: 16r1FFFFFFF.
	^clockValue asSmallIntegerObj! !

!MIDIPlugin methodsFor: 'primitives' stamp: 'ar 5/12/2000 17:14'!
primitiveMIDIGetPortCount

	| n |
	self primitive: 'primitiveMIDIGetPortCount'.
	n := self sqMIDIGetPortCount.
	^n asSmallIntegerObj
! !

!MIDIPlugin methodsFor: 'primitives' stamp: 'ar 5/12/2000 17:14'!
primitiveMIDIGetPortDirectionality: portNum

	|  dir |
	self primitive: 'primitiveMIDIGetPortDirectionality'
		parameters: #(SmallInteger).
	dir := self sqMIDIGetPortDirectionality: portNum.
	^dir asSmallIntegerObj! !

!MIDIPlugin methodsFor: 'primitives' stamp: 'eem 9/19/2018 17:31'!
primitiveMIDIGetPortName: portNum
	| portName sz nameObj |
	<var: #portName declareC: 'char portName[256]'>
	self primitive: 'primitiveMIDIGetPortName' parameters: #(SmallInteger).

	sz := self sqMIDIGet: portNum Port: portName Name: 255.
	nameObj := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: sz.
	interpreterProxy failed ifTrue: [^nil].
	self memcpy: (nameObj asValue: String) _: portName _: sz.
	^nameObj! !

!MIDIPlugin methodsFor: 'primitives' stamp: 'ar 5/12/2000 17:15'!
primitiveMIDIOpenPort: portNum sema: semaIndex speed: clockRate

	self primitive: 'primitiveMIDIOpenPort'
		parameters: #(SmallInteger SmallInteger SmallInteger).
	self cCode: 'sqMIDIOpenPort(portNum, semaIndex, clockRate)'! !

!MIDIPlugin methodsFor: 'primitives' stamp: 'ar 5/12/2000 17:15'!
primitiveMIDIParameterGet: whichParameter

	|  currentValue |
	"read parameter"
	self primitive: 'primitiveMIDIParameterGet'
		parameters: #(SmallInteger).
	currentValue := self cCode: 'sqMIDIParameterGet(whichParameter)'.
	^currentValue asSmallIntegerObj! !

!MIDIPlugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 5/12/2000 17:15'!
primitiveMIDIParameterGetOrSet
	"Backward compatibility"
	<export: true>
	interpreterProxy methodArgumentCount = 1
		ifTrue:[^self primitiveMIDIParameterGet]
		ifFalse:[^self primitiveMIDIParameterSet]! !

!MIDIPlugin methodsFor: 'primitives' stamp: 'ar 5/12/2000 17:15'!
primitiveMIDIParameterSet: whichParameter value: newValue

	"write parameter"
	self primitive:'primitiveMIDIParameterSet'
		parameters:#(SmallInteger SmallInteger).
	self cCode: 'sqMIDIParameterSet(whichParameter, newValue)'! !

!MIDIPlugin methodsFor: 'primitives' stamp: 'nice 11/11/2016 20:42'!
primitiveMIDIRead: portNum into: array

	| arrayLength bytesRead |
	self primitive: 'primitiveMIDIRead'
		parameters: #(SmallInteger ByteArray).
	arrayLength := interpreterProxy byteSizeOf: array cPtrAsOop.
	bytesRead := self sqMIDIPort: portNum
			Read: arrayLength
			Into: array.
	^bytesRead asSmallIntegerObj! !

!MIDIPlugin methodsFor: 'primitives' stamp: 'nice 11/11/2016 20:43'!
primitiveMIDIWrite: portNum from: array at: time

	| arrayLength bytesWritten |
	self primitive: 'primitiveMIDIWrite'
		parameters: #(SmallInteger ByteArray SmallInteger).
	arrayLength := interpreterProxy byteSizeOf: array cPtrAsOop.
	bytesWritten := self sqMIDIPort: portNum
			Write: arrayLength
			From: array
			At: time.
	^bytesWritten asSmallIntegerObj! !

!MIDIPlugin methodsFor: 'initialize-release' stamp: 'ar (auto pragmas 12/08) 5/12/2000 16:55'!
shutdownModule
	<export: true>
	^self cCode: 'midiShutdown()' inSmalltalk:[true]! !

SmartSyntaxInterpreterPlugin subclass: #MacMenubarPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!MacMenubarPlugin class methodsFor: 'translation' stamp: 'JMM 8/27/2004 12:50'!
hasHeaderFile
	^true! !

!MacMenubarPlugin class methodsFor: 'translation' stamp: 'JMM 8/27/2004 12:50'!
requiresCrossPlatformFiles
	^false! !

!MacMenubarPlugin class methodsFor: 'translation' stamp: 'JMM 8/16/2004 13:17'!
requiresPlatformFiles
	^true! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:32'!
primitiveAppendMenu: menuHandleOop data: str255
	<var: #menuHandle type: #MenuHandle>
	<var: #constStr255 type: #ConstStr255Param>
	| menuHandle constStr255 |
	self primitive: 'primitiveAppendMenu'
		parameters: #(Oop ByteArray).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	constStr255 := self cCoerce: str255 to: #ConstStr255Param.	
	self cCode: 'AppendMenu(menuHandle,constStr255)' inSmalltalk:[menuHandle].
	^nil! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:33'!
primitiveAppendMenuItemText: menuHandleOop data: str255
	<var: #menuHandle type: #MenuHandle>
	<var: #constStr255 type: #ConstStr255Param>
	| menuHandle constStr255 |
	self primitive: 'primitiveAppendMenuItemText'
		parameters: #(Oop ByteArray).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	constStr255 := self cCoerce: str255 to: #ConstStr255Param.	
	self cCode: 'AppendMenuItemText(menuHandle,constStr255)' inSmalltalk:[menuHandle].
	^nil! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:34'!
primitiveCheckMenuItem: menuHandleOop item: anInteger checked: aBoolean
	<var: #menuHandle type: #MenuHandle>
	| menuHandle |
	self primitive: 'primitiveCheckMenuItem'
		parameters: #(Oop SmallInteger Boolean).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'CheckMenuItem(menuHandle,anInteger,aBoolean)' inSmalltalk:[menuHandle].
	^nil

! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'JMM 8/16/2004 13:51'!
primitiveClearMenuBar 
	self primitive: 'primitiveClearMenuBar'
		parameters: #().
	
	self cCode: 'ClearMenuBar()' inSmalltalk:[].
	^nil! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:34'!
primitiveCountMenuItems: menuHandleOop 
	<var: #menuHandle type: #MenuHandle>
	| menuHandle returnValue |
	self primitive: 'primitiveCountMenuItems'
		parameters: #(Oop).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	returnValue := self cCode: 'CountMenuItems(menuHandle)' inSmalltalk:[0].
	^returnValue asSmallIntegerObj
! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 7/22/2016 00:28'!
primitiveCreateStandardWindowMenu: inOptions 
	<var: #menuHandle type: #MenuHandle>
	| menuHandle result |
	self primitive: 'primitiveCreateStandardWindowMenu'
		parameters: #(SmallInteger).
	self cppIf: #'TARGET_API_MAC_CARBON'
		ifTrue: [result := self cCode: 'CreateStandardWindowMenu(inOptions,&menuHandle);' inSmalltalk:[0]].
	^interpreterProxy positiveMachineIntegerFor: (self cCoerce: menuHandle to: #'usqIntptr_t')! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:36'!
primitiveDeleteMenu: menuID 
	<var: #menuID type: #MenuID>
	self primitive: 'primitiveDeleteMenu'
		parameters: #(SmallInteger).
	self cCode: 'DeleteMenu(menuID)' inSmalltalk:[].
	^nil! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:36'!
primitiveDeleteMenuItem: menuHandleOop item: anInteger
	<var: #menuHandle type: #MenuHandle>
	| menuHandle |
	self primitive: 'primitiveDeleteMenuItem'
		parameters: #(Oop SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'DeleteMenuItem(menuHandle,anInteger)' inSmalltalk:[menuHandle].
	^nil
! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 4/20/2016 11:31'!
primitiveDisableMenuCommand: menuHandleOop item: anInteger
	<var: #menuHandle type: #MenuHandle>
	<var: #commandID type: #MenuCommand>
	| menuHandle commandID |
	self primitive: 'primitiveDisableMenuCommand'
		parameters: #(Oop Oop).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	commandID := self cCoerce: (interpreterProxy positive32BitValueOf: anInteger) to: #MenuCommand.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cppIf: #'TARGET_API_MAC_CARBON'
		ifTrue: [self cCode: 'DisableMenuCommand(menuHandle,commandID);' inSmalltalk: [menuHandle]].
	^nil
! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:37'!
primitiveDisableMenuItem: menuHandleOop item: anInteger
	<var: #menuHandle type: #MenuHandle>
	| menuHandle |
	self primitive: 'primitiveDisableMenuItem'
		parameters: #(Oop SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'DisableMenuItem(menuHandle,anInteger)' inSmalltalk:[menuHandle].
	^nil
! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:37'!
primitiveDisableMenuItemIcon: menuHandleOop item: anInteger
	<var: #menuHandle type: #MenuHandle>
	| menuHandle |
	self primitive: 'primitiveDisableMenuItemIcon'
		parameters: #(Oop SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'DisableMenuItemIcon(menuHandle,anInteger)' inSmalltalk:[menuHandle].
	^nil
! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:38'!
primitiveDisposeMenu: menuHandleOop 
	<var: #menuHandle type: #MenuHandle>
	| menuHandle |
	self primitive: 'primitiveDisposeMenu'
		parameters: #(Oop).	
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'DisposeMenu(menuHandle)' inSmalltalk:[menuHandle].
	^nil
! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 8/8/2016 12:13'!
primitiveDisposeMenuBar: menuHandleOop 
	<var: #menuBarHandle type: #Handle>
	| menuBarHandle |
	self primitive: 'primitiveDisposeMenuBar'
		parameters: #(Oop).
	menuBarHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #Handle.
	self cppIf: #'TARGET_API_MAC_CARBON'
		ifTrue: [self cCode: 'DisposeMenuBar(menuBarHandle);' inSmalltalk: [menuBarHandle]]
		ifFalse: [self cCode: 'DisposeHandle(menuBarHandle);' inSmalltalk: [menuBarHandle]].
	^nil
! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'JMM 8/16/2004 13:53'!
primitiveDrawMenuBar
	self primitive: 'primitiveDrawMenuBar'
		parameters: #().
	
	self cCode: 'DrawMenuBar()' inSmalltalk:[].
	^nil! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 4/20/2016 11:32'!
primitiveEnableMenuCommand: menuHandleOop item: anInteger
	<var: #menuHandle type: #MenuHandle>
	<var: #commandID type: #MenuCommand>
	| menuHandle commandID |
	self primitive: 'primitiveEnableMenuCommand'
		parameters: #(Oop Oop).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	commandID := self cCoerce: (interpreterProxy positive32BitValueOf: anInteger) to: #MenuCommand.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cppIf: #'TARGET_API_MAC_CARBON'
		ifTrue: [self cCode: 'EnableMenuCommand(menuHandle,commandID);' inSmalltalk: [menuHandle]].
	^nil! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:38'!
primitiveEnableMenuItem: menuHandleOop item: anInteger
	<var: #menuHandle type: #MenuHandle>
	| menuHandle |
	self primitive: 'primitiveEnableMenuItem'
		parameters: #(Oop SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'EnableMenuItem(menuHandle,anInteger)' inSmalltalk:[menuHandle].
	^nil
! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:39'!
primitiveEnableMenuItemIcon: menuHandleOop item: anInteger
	<var: #menuHandle type: #MenuHandle>
	| menuHandle |
	self primitive: 'primitiveEnableMenuItemIcon'
		parameters: #(Oop SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'EnableMenuItemIcon(menuHandle,anInteger)' inSmalltalk:[menuHandle].
	^nil
! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 4/20/2016 11:33'!
primitiveGetIndMenuItemWithCommandID: menuHandleOop commandID: aCommandID
	<var: #menuHandle type: #MenuHandle>
	<var: #commandID type: #MenuCommand>
	<var: #applicationMenu type: #MenuHandle>
	<var: #outIndex type: #MenuItemIndex>
	| menuHandle commandID applicationMenu outIndex |
	self primitive: 'primitiveGetIndMenuItemWithCommandID'
		parameters: #(Oop Oop).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	commandID := self cCoerce: (interpreterProxy positive32BitValueOf: aCommandID) to: #MenuCommand.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cppIf: #'TARGET_API_MAC_CARBON'
		ifTrue: [self cCode: 'GetIndMenuItemWithCommandID(menuHandle, kHICommandHide, 1,
                   &applicationMenu, &outIndex);' inSmalltalk: [menuHandle]].
	^outIndex asSmallIntegerObj
! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 7/22/2016 00:28'!
primitiveGetIndMenuWithCommandID: menuHandleOop commandID: aCommandID
	<var: #menuHandle type: #MenuHandle>
	<var: #commandID type: #MenuCommand>
	<var: #applicationMenu type: #MenuHandle>
	<var: #outIndex type: #MenuItemIndex>
	| menuHandle commandID applicationMenu outIndex |
	self primitive: 'primitiveGetIndMenuWithCommandID'
		parameters: #(Oop Oop).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: 'MenuHandle'.
	commandID := self cCoerce: (interpreterProxy positive32BitValueOf: aCommandID) to: 'MenuCommand'.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cppIf: #'TARGET_API_MAC_CARBON'
		ifTrue: [self cCode: 'GetIndMenuItemWithCommandID(menuHandle, kHICommandHide, 1,
                   &applicationMenu, &outIndex);' inSmalltalk: [menuHandle]].
	outIndex asSmallIntegerObj. "to avoid elimination of the variable..."
	^interpreterProxy positiveMachineIntegerFor: (self cCoerce: applicationMenu to: #'usqIntptr_t')

! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 22:02'!
primitiveGetItemCmd: menuHandleOop item: anInteger
	| menuHandle aCharacter |
	<var: #menuHandle type: #MenuHandle>
	<var: #aCharacter type: #CharParameter>
	self primitive: 'primitiveGetItemCmd'
		parameters: #(Oop SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	aCharacter := 0.
	self cCode: 'GetItemCmd(menuHandle,anInteger,&aCharacter)' inSmalltalk:[menuHandle].
	^aCharacter asSmallIntegerObj

! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:40'!
primitiveGetItemIcon: menuHandleOop item: anInteger
	<var: #menuHandle type: #MenuHandle>
	<var: #iconIndex type: #short>
	| menuHandle iconIndex |
	self primitive: 'primitiveGetItemIcon'
		parameters: #(Oop SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	iconIndex := 0.
	self cCode: 'GetItemIcon(menuHandle,anInteger,&iconIndex)' inSmalltalk:[menuHandle].
	^iconIndex asSmallIntegerObj
! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 22:02'!
primitiveGetItemMark: menuHandleOop item: anInteger
	| menuHandle aCharacter |
	<var: #menuHandle type: #MenuHandle>
	<var: #aCharacter type: #CharParameter>
	self primitive: 'primitiveGetItemMark'
		parameters: #(Oop SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	aCharacter := 0.
	self cCode: 'GetItemMark(menuHandle,anInteger,&aCharacter)' inSmalltalk:[menuHandle].
	^aCharacter asSmallIntegerObj

! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:41'!
primitiveGetItemStyle: menuHandleOop item: anInteger 
	<var: #menuHandle type: #MenuHandle>
	<var: #chStyle type: #Style>
	| menuHandle chStyle |
	self primitive: 'primitiveGetItemStyle'
		parameters: #(Oop SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	chStyle := 0.
	self cCode: 'GetItemStyle(menuHandle,anInteger,&chStyle)' inSmalltalk:[menuHandle].
	^chStyle asSmallIntegerObj

! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 7/22/2016 00:28'!
primitiveGetMenuBar 
	<var: #menuHandle type: #Handle>
	| menuHandle |
	self primitive: 'primitiveGetMenuBar'
		parameters: #().
	menuHandle := self cCode: 'GetMenuBar()' inSmalltalk:[0].
	^interpreterProxy positiveMachineIntegerFor: (self cCoerce: menuHandle to: #'usqIntptr_t')! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 7/22/2016 00:29'!
primitiveGetMenuHandle: menuID 
	<var: #menuHandle type: #MenuHandle>
	<var: #menuID type: #MenuID>
	| menuHandle |
	self primitive: 'primitiveGetMenuHandle'
		parameters: #(SmallInteger).
	menuHandle := self cCode: 'GetMenuHandle(menuID)' inSmalltalk:[0].
	^interpreterProxy positiveMachineIntegerFor: (self cCoerce: menuHandle to: #'usqIntptr_t')! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:43'!
primitiveGetMenuID: menuHandleOop 
	<var: #menuHandle type: #MenuHandle>
	<var: #menuID type: #MenuID>
	| menuHandle menuID |
	self primitive: 'primitiveGetMenuID'
		parameters: #(Oop ).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'menuID = GetMenuID(menuHandle)' inSmalltalk:[menuHandle].
	^menuID asSmallIntegerObj
	! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:44'!
primitiveGetMenuItemCommandID: menuHandleOop item: anInteger 
	<var: #menuHandle type: #MenuHandle>
	<var: #outCommandID type: #MenuCommand>
	| menuHandle outCommandID |
	self primitive: 'primitiveGetMenuItemCommandID'
		parameters: #(Oop SmallInteger).
	outCommandID := 0.
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'GetMenuItemCommandID(menuHandle,anInteger,&outCommandID)' inSmalltalk:[menuHandle].
	^interpreterProxy positive32BitIntegerFor: outCommandID
	! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:45'!
primitiveGetMenuItemFontID: menuHandleOop item: anInteger 
	<var: #menuHandle type: #MenuHandle>
	<var: #outFontID type: #SInt16>
	| menuHandle outFontID |
	self primitive: 'primitiveGetMenuItemFontID'
		parameters: #(Oop SmallInteger).
	outFontID := 0.
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'GetMenuItemFontID(menuHandle,anInteger,&outFontID)' inSmalltalk:[menuHandle].
	^outFontID asSmallIntegerObj
	! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:46'!
primitiveGetMenuItemHierarchicalID: menuHandleOop item: anInteger 
	<var: #menuHandle type: #MenuHandle>
	<var: #outHierID type: #MenuID>
	| menuHandle outHierID |
	self primitive: 'primitiveGetMenuItemHierarchicalID'
		parameters: #(Oop SmallInteger ).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	outHierID := 0.
	self cCode: 'GetMenuItemHierarchicalID(menuHandle,anInteger,&outHierID)' inSmalltalk:[menuHandle].
	^outHierID asSmallIntegerObj

! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:46'!
primitiveGetMenuItemKeyGlyph: menuHandleOop item: anInteger 
	<var: #menuHandle type: #MenuHandle>
	<var: #outGlyph type: #SInt16>
	| menuHandle outGlyph |
	self primitive: 'primitiveGetMenuItemKeyGlyph'
		parameters: #(Oop SmallInteger).
	outGlyph := 0.
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'GetMenuItemKeyGlyph(menuHandle,anInteger,&outGlyph)' inSmalltalk:[menuHandle].
	^interpreterProxy positive32BitIntegerFor: outGlyph
	! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:47'!
primitiveGetMenuItemModifiers: menuHandleOop item: anInteger 
	<var: #menuHandle type: #MenuHandle>
	<var: #outModifers type: #Style>
	| menuHandle outModifers |
	self primitive: 'primitiveGetMenuItemModifiers'
		parameters: #(Oop SmallInteger ).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	outModifers := 0.
	self cCode: 'GetMenuItemModifiers(menuHandle,anInteger,&outModifers)' inSmalltalk:[menuHandle].
	^outModifers asSmallIntegerObj
	

! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 22:02'!
primitiveGetMenuItemText: menuHandleOop item: anInteger
	| menuHandle size oop ptr aString |
	<var: #menuHandle type: #MenuHandle>
	<var: #aString type: #Str255>
	<var: #ptr type: 'char *'>
	self primitive: 'primitiveGetMenuItemText'
		parameters: #(Oop SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	aString at: 0 put: 0.
	self cCode: 'GetMenuItemText(menuHandle,anInteger,aString)' inSmalltalk:[menuHandle].
	size := self cCode: 'aString[0]' inSmalltalk: [0].
	oop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize:  size.
	ptr := interpreterProxy firstIndexableField: oop.
	0 to: size-1 do:[:i|
		ptr at: i put: (aString at: (i+1))].
	^oop

! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:48'!
primitiveGetMenuItemTextEncoding: menuHandleOop item: anInteger
	<var: #menuHandle type: #MenuHandle>
	<var: #outScriptID type: #TextEncoding>
	| menuHandle outScriptID |
	self primitive: 'primitiveGetMenuItemTextEncoding'
		parameters: #(Oop SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'GetMenuItemTextEncoding(menuHandle,anInteger,&outScriptID)' inSmalltalk:[menuHandle].
	^interpreterProxy positive32BitIntegerFor: outScriptID! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 22:02'!
primitiveGetMenuTitle: menuHandleOop
	| menuHandle size oop ptr aString |
	<var: #menuHandle type: #MenuHandle>
	<var: #aString type: #Str255>
	<var: #ptr type: 'char *'>
	self primitive: 'primitiveGetMenuTitle'
		parameters: #(Oop).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	aString at: 0 put: 0.
	self cCode: 'GetMenuTitle(menuHandle,aString)' inSmalltalk:[menuHandle].
	size := self cCode: 'aString[0]' inSmalltalk: [0].
	oop := interpreterProxy instantiateClass: interpreterProxy classString indexableSize:  size.
	ptr := interpreterProxy firstIndexableField: oop.
	0 to: size-1 do:[:i|
		ptr at: i put: (aString at: (i+1))].
	^oop

! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'JMM 8/17/2004 16:29'!
primitiveHideMenuBar 
	self primitive: 'primitiveHideMenuBar'
		parameters: #().
	
	self cCode: 'HideMenuBar()' inSmalltalk:[].
	^nil! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:50'!
primitiveHiliteMenu: menuID 
	<var: #menuID type: #MenuID>
	self primitive: 'primitiveHiliteMenu'
		parameters: #(SmallInteger).
	self cCode: 'HiliteMenu(menuID)' inSmalltalk:[].
	^nil! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:50'!
primitiveInsertFontResMenu: menuHandleOop afterItem: afterItemInteger scriptFilter:  scriptFilterInteger
	<var: #menuHandle type: #MenuHandle>
	| menuHandle |
	self primitive: 'primitiveInsertFontResMenu'
		parameters: #(Oop SmallInteger SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'InsertFontResMenu(menuHandle,afterItemInteger,scriptFilterInteger)' inSmalltalk:[menuHandle].
	^nil
	! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:51'!
primitiveInsertIntlResMenu: menuHandleOop theType: aResType afterItem: afterItemInteger scriptFilter:  scriptFilterInteger
	<var: #menuHandle type: #MenuHandle>
	<var: #resType type: #ResType>
	| menuHandle resType |
	self primitive: 'primitiveInsertIntlResMenu'
		parameters: #(Oop SmallInteger SmallInteger SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	resType := self cCoerce: (interpreterProxy positive32BitValueOf: aResType) to: #ResType.
	self cCode: 'InsertIntlResMenu(menuHandle,resType,afterItemInteger,scriptFilterInteger)' inSmalltalk:[menuHandle].
	^nil
	! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:51'!
primitiveInsertMenu: menuHandleOop beforeID: anInteger
	<var: #menuHandle type: #MenuHandle>
	<var: #anInteger type: #MenuID>
	| menuHandle |
	self primitive: 'primitiveInsertMenu'
		parameters: #(Oop SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'InsertMenu(menuHandle,anInteger)' inSmalltalk:[menuHandle].
	^nil! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:52'!
primitiveInsertMenuItem: menuHandleOop itemString: str255 afterItem: anInteger
	<var: #menuHandle type: #MenuHandle>
	<var: #constStr255 type: #ConstStr255Param>
	| menuHandle constStr255 |
	self primitive: 'primitiveInsertMenuItem'
		parameters: #(Oop ByteArray SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	constStr255 := self cCoerce: str255 to: #ConstStr255Param.
	self cCode: 'InsertMenuItem(menuHandle,constStr255,anInteger)' inSmalltalk:[menuHandle].
	^nil
	! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'JMM 8/16/2004 13:54'!
primitiveInvalMenuBar
	self primitive: 'primitiveInvalMenuBar'
		parameters: #().
	
	self cCode: 'InvalMenuBar()' inSmalltalk:[].
	^nil! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'JMM 8/17/2004 16:58'!
primitiveIsMenuBarVisible
	| result |
	self primitive: 'primitiveIsMenuBarVisible'
		parameters: #().
	result := self cCode: 'IsMenuBarVisible()' inSmalltalk:[true].
	^result asOop: Boolean! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:53'!
primitiveIsMenuItemEnabled: menuHandleOop item: anInteger
	<var: #menuHandle type: #MenuHandle>
	| menuHandle result |
	self primitive: 'primitiveIsMenuItemEnabled'
		parameters: #(Oop SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	result := self cCode: 'IsMenuItemEnabled(menuHandle,anInteger)' inSmalltalk:[0].
	^result asOop: Boolean! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:53'!
primitiveIsMenuItemIconEnabled: menuHandleOop item: anInteger
	<var: #menuHandle type: #MenuHandle>
	| menuHandle result |
	self primitive: 'primitiveIsMenuItemIconEnabled'
		parameters: #(Oop SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	result := self cCode: 'IsMenuItemIconEnabled(menuHandle,anInteger)' inSmalltalk:[0].
	^result asOop: Boolean! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 7/22/2016 00:29'!
primitiveNewMenu: menuID menuTitle: menuTitle
	<var: #menuHandle type: #MenuHandle>
	<var: #constStr255 type: #ConstStr255Param>
	<var: #menuID type: #MenuID>
	| menuHandle constStr255 |
	self primitive: 'primitiveNewMenu'
		parameters: #(SmallInteger ByteArray).
	constStr255 := self cCoerce: menuTitle to: #ConstStr255Param.	
	menuHandle := self cCode: 'NewMenu(menuID,constStr255)' inSmalltalk:[0].
	^interpreterProxy positiveMachineIntegerFor: (self cCoerce: menuHandle to: #'usqIntptr_t')! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 22:03'!
primitiveSetItemCmd: menuHandleOop item: anInteger cmdChar: anIntegerCmdChar
	<var: #menuHandle type: #MenuHandle>
	<var: #aCharacter type: #CharParameter>
	| menuHandle aCharacter |
	self primitive: 'primitiveSetItemCmd'
		parameters: #(Oop SmallInteger SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	aCharacter := anIntegerCmdChar.
	self cCode: 'SetItemCmd(menuHandle,anInteger,aCharacter)' inSmalltalk:[menuHandle].
	^nil

! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:55'!
primitiveSetItemIcon: menuHandleOop item: anInteger iconIndex: aIconIndexInteger
	<var: #menuHandle type: #MenuHandle>
	| menuHandle |
	self primitive: 'primitiveSetItemIcon'
		parameters: #(Oop SmallInteger SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'SetItemIcon(menuHandle,anInteger,aIconIndexInteger)' inSmalltalk:[menuHandle].
	^nil

! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:56'!
primitiveSetItemMark: menuHandleOop item: anInteger markChar: aMarkChar
	<var: #menuHandle type: #MenuHandle>
	<var: #aCharacter type: #CharParameter>
	| menuHandle aCharacter |
	self primitive: 'primitiveSetItemMark'
		parameters: #(Oop SmallInteger SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	aCharacter := aMarkChar.
	self cCode: 'SetItemMark(menuHandle,anInteger,aCharacter)' inSmalltalk:[menuHandle].
	^nil

! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:56'!
primitiveSetItemStyle: menuHandleOop item: anInteger styleParameter: chStyleInteger
	<var: #menuHandle type: #MenuHandle>
	| menuHandle |
	self primitive: 'primitiveSetItemStyle'
		parameters: #(Oop SmallInteger SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'SetItemStyle(menuHandle,anInteger,chStyleInteger)' inSmalltalk:[menuHandle].
	^nil

! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 22:24'!
primitiveSetMenuBar: menuHandleOop
	<var: #menuBarHandle type: #MenuBarHandle>
	| menuBarHandle |
	self primitive: 'primitiveSetMenuBar'
		parameters: #(Oop).
	menuBarHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuBarHandle.
	self cCode: 'SetMenuBar(menuBarHandle)' inSmalltalk:[menuBarHandle].
	^nil! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:57'!
primitiveSetMenuItemCommandID: menuHandleOop item: anInteger menuCommand:  inCommandID
	<var: #menuHandle type: #MenuHandle>
	<var: #commandID type: #MenuCommand>
	| menuHandle commandID |
	self primitive: 'primitiveSetMenuItemCommandID'
		parameters: #(Oop SmallInteger Oop).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	commandID := self cCoerce: (interpreterProxy positive32BitValueOf: inCommandID) to: #MenuCommand.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'SetMenuItemCommandID(menuHandle,anInteger,commandID)' inSmalltalk:[menuHandle].
	^nil
	! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:58'!
primitiveSetMenuItemFontID: menuHandleOop item: anInteger fontID: aFontIDInteger  
	<var: #menuHandle type: #MenuHandle>
	| menuHandle |
	self primitive: 'primitiveSetMenuItemFontID'
		parameters: #(Oop SmallInteger SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'SetMenuItemFontID(menuHandle,anInteger,aFontIDInteger)' inSmalltalk:[menuHandle].
	^nil
	! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 22:11'!
primitiveSetMenuItemHierarchicalID: menuHandleOop item: anInteger hierID: aMenuID
	<var: #menuHandle type: #MenuHandle>
	<var: #aMenuID type: #MenuID>
	| menuHandle |
	self primitive: 'primitiveSetMenuItemHierarchicalID'
		parameters: #(Oop SmallInteger SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'SetMenuItemHierarchicalID(menuHandle,anInteger,aMenuID)' inSmalltalk:[menuHandle].
	^nil! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:59'!
primitiveSetMenuItemKeyGlyph: menuHandleOop item: anInteger glyph:  inGlyphInteger
	<var: #menuHandle type: #MenuHandle>
	| menuHandle |
	self primitive: 'primitiveSetMenuItemKeyGlyph'
		parameters: #(Oop SmallInteger SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'SetMenuItemKeyGlyph(menuHandle,anInteger,inGlyphInteger)' inSmalltalk:[menuHandle].
	^nil
	! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:59'!
primitiveSetMenuItemModifiers: menuHandleOop item: anInteger inModifiers: aUInt8
	<var: #menuHandle type: #MenuHandle>
	| menuHandle |
	self primitive: 'primitiveSetMenuItemModifiers'
		parameters: #(Oop SmallInteger SmallInteger).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	self cCode: 'SetMenuItemModifiers(menuHandle,anInteger,aUInt8)' inSmalltalk:[menuHandle].
	^nil! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 21:59'!
primitiveSetMenuItemText: menuHandleOop item: anInteger itemString: str255
	<var: #menuHandle type: #MenuHandle>
	<var: #constStr255 type: #ConstStr255Param>
	| menuHandle constStr255 |
	self primitive: 'primitiveSetMenuItemText'
		parameters: #(Oop SmallInteger ByteArray).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	constStr255 := self cCoerce: str255 to: #ConstStr255Param.
	self cCode: 'SetMenuItemText(menuHandle,anInteger,constStr255)' inSmalltalk:[menuHandle].
	^nil
	! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 22:00'!
primitiveSetMenuItemTextEncoding: menuHandleOop item: anInteger inScriptID: aTextEncodingOop
	<var: #menuHandle type: #MenuHandle>
	<var: #inScriptID type: #TextEncoding>
	| menuHandle inScriptID |
	self primitive: 'primitiveSetMenuItemTextEncoding'
		parameters: #(Oop SmallInteger Oop).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	inScriptID := self cCoerce: (interpreterProxy positive32BitValueOf: aTextEncodingOop) to: #TextEncoding.
	self cCode: 'SetMenuItemTextEncoding(menuHandle,anInteger,inScriptID)' inSmalltalk:[menuHandle].
	^nil! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'nice 3/12/2016 22:01'!
primitiveSetMenuTitle: menuHandleOop  title: str255
	<var: #menuHandle type: #MenuHandle>
	<var: #constStr255 type: #ConstStr255Param>
	| menuHandle constStr255 |
	self primitive: 'primitiveSetMenuTitle'
		parameters: #(Oop ByteArray).
	menuHandle := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: menuHandleOop) to: #MenuHandle.
	(self ioCheckMenuHandle: menuHandle) ifFalse: [^interpreterProxy success: false].
	constStr255 := self cCoerce: str255 to: #ConstStr255Param.
	self cCode: 'SetMenuTitle(menuHandle,constStr255)' inSmalltalk:[menuHandle].
	^nil
	! !

!MacMenubarPlugin methodsFor: 'system primitives' stamp: 'JMM 8/17/2004 16:29'!
primitiveShowMenuBar 
	self primitive: 'primitiveShowMenuBar'
		parameters: #().
	
	self cCode: 'ShowMenuBar()' inSmalltalk:[].
	^nil! !

SmartSyntaxInterpreterPlugin subclass: #Mpeg3Plugin
	instanceVariableNames: 'maximumNumberOfFilesToWatch mpegFiles'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!Mpeg3Plugin commentStamp: '<historical>' prior: 0!
/********************************************************
 * An interface to LibMPEG3
 * Author: Adam Williams <broadcast@earthling.net>
 * Page: heroine.linuxbox.com
 *
 * Changed for Squeak to work with Squeak and to work on the Macintosh
 * Sept 2000, by John M McIntosh johnmci@smalltalkconsulting.com
 * The smalltalk code and the C code it produces is released under the 
 * Squeak licence. The libmpeg3 C code is co-licenced under either the Squeak licence or
 * the GNU LGPL!

!Mpeg3Plugin class methodsFor: 'initialize-release' stamp: 'JMM 10/2/2000 12:56'!
declareCVarsIn: cg 
	super declareCVarsIn: cg.
	cg var: 'mpegFiles' declareC: 'mpeg3_t *mpegFiles[1024+1]'.
! !

!Mpeg3Plugin class methodsFor: 'initialize-release' stamp: 'tpr 5/23/2001 17:10'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!Mpeg3Plugin class methodsFor: 'initialize-release' stamp: 'tpr 7/4/2001 15:13'!
requiresCrossPlatformFiles
	"If there cross platform files to be associated with the plugin, here is where you want to flag"
	^true! !

!Mpeg3Plugin class methodsFor: 'initialize-release' stamp: 'tpr 3/13/2002 18:05'!
requiresPlatformFiles
	"If there platform files to be associated with the plugin, here is where you want to flag"
	^true! !

!Mpeg3Plugin methodsFor: 'support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:05'!
checkFileEntry: aMpegFile
	<var: #aMpegFile type: 'mpeg3_t * '>
	1 to: maximumNumberOfFilesToWatch do: 
		[:i | ((mpegFiles at: i) = aMpegFile) ifTrue: 
				[^true]].
	^false.
		
	! !

!Mpeg3Plugin methodsFor: 'support' stamp: 'eem 10/18/2018 17:09'!
initialiseModule
	<export: true>
	maximumNumberOfFilesToWatch := 1024.
	1 to: maximumNumberOfFilesToWatch do: [:i | mpegFiles at: i put: 0].
	^true! !

!Mpeg3Plugin methodsFor: 'support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:05'!
makeFileEntry: aMpegFile
	<var: #aMpegFile type: 'mpeg3_t * '>
	1 to: maximumNumberOfFilesToWatch do: 
		[:i | ((mpegFiles at: i) = 0) ifTrue: 
				[mpegFiles at: i put: aMpegFile.
				^true]].
	^false
	"Ok no room just ignore, we'll get a primitive failure later"
		
	! !

!Mpeg3Plugin methodsFor: 'support' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:02'!
mpeg3tValueOf: mpeg3tHandle 
	"Return a pointer to the first byte of of the mpeg3_t record within the  
	given Smalltalk object, or nil if socketOop is not a mpeg3_t record."
	| index check |

	<returnTypeC: 'mpeg3_t *'>
	<var: #index type: 'mpeg3_t ** '>
	interpreterProxy success: ((interpreterProxy isBytes: mpeg3tHandle)
			and: [(interpreterProxy byteSizeOf: mpeg3tHandle)
					= 4]).
	interpreterProxy failed
		ifTrue: [^ nil]
		ifFalse: 
			[index := self cCoerce: (interpreterProxy firstIndexableField: mpeg3tHandle)
						to: 'mpeg3_t **'.
			self cCode: 'check = checkFileEntry(*index)'.
			check = 0 ifTrue: [^nil]. 
			^ self cCode: '*index']! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:03'!
primitiveMPEG3AudioChannels: fileHandle stream: aNumber
	| file result |

	"int mpeg3_audio_channels(mpeg3_t *file,int stream)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3AudioChannels'
		parameters: #(Oop SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^0].
	aNumber >= (self cCode: 'mpeg3_total_astreams(file)') ifTrue: [
		interpreterProxy success: false. ^0.
	].

	result := self cCode: 'mpeg3_audio_channels(file,aNumber)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:00'!
primitiveMPEG3AudioSamples: fileHandle stream: aNumber
	| file result |

	"long mpeg3_audio_samples(mpeg3_t *file, int stream)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3AudioSamples'
		parameters: #(Oop SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'mpeg3_total_astreams(file)') ifTrue: [
		interpreterProxy success: false. ^0.
	].

	self cCode: 'result = mpeg3_audio_samples(file,aNumber)'.
	^result asOop: Float
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas dtl 2010-09-28) 4/12/2006 12:08'!
primitiveMPEG3CheckSig: path
	| result sz storage |

	"int mpeg3_check_sig(char *path)"
	<var: #storage declareC: 'char storage[1024] '>
	self primitive: 'primitiveMPEG3CheckSig'
		parameters: #(String).
	sz := interpreterProxy byteSizeOf: path cPtrAsOop.
	interpreterProxy ioFilename: storage fromString: path ofLength: sz resolveAliases: true.
	self cCode: 'result = mpeg3_check_sig(storage)'.
	^result asOop: Boolean
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:00'!
primitiveMPEG3Close: fileHandle
	| file index |

	"int mpeg3_close(mpeg3_t *file)"
	<var: #file type: 'mpeg3_t * '>
	<var: #index type: 'mpeg3_t ** '>
	self primitive: 'primitiveMPEG3Close'
		parameters: #(Oop).
	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^nil].
	self cCode: 'removeFileEntry(file); mpeg3_close(file)'.
	index := self cCoerce: (interpreterProxy firstIndexableField: fileHandle)
						to: 'mpeg3_t **'.
	self cCode: '*index = 0'.
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:01'!
primitiveMPEG3DropFrames: fileHandle frames: aFrameNumber stream: aNumber
	| file result |

	"int mpeg3_drop_frames(mpeg3_t *file, long frames, int stream)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3DropFrames'
		parameters: #(Oop SmallInteger SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_vstreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].


	self cCode: 'result = mpeg3_drop_frames(file,aFrameNumber,aNumber)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:03'!
primitiveMPEG3EndOfAudio: fileHandle stream: aNumber
	| file result |

	"int mpeg3_end_of_audio(mpeg3_t *file, int stream)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3EndOfAudio'
		parameters: #(Oop SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_astreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].

	self cCode: 'result = mpeg3_end_of_audio(file,aNumber)'.
	^result asOop: Boolean
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas dtl 2010-09-28) 4/12/2006 12:13'!
primitiveMPEG3EndOfVideo: fileHandle stream: aNumber
	| file result |

	"int mpeg3_end_of_video(mpeg3_t *file, int stream)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3EndOfVideo'
		parameters: #(Oop SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_vstreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].


	self cCode: 'result = mpeg3_end_of_video(file,aNumber)'.
	^result asOop: Boolean
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:03'!
primitiveMPEG3FrameRate: fileHandle stream: aNumber
	| file result |

	"float mpeg3_frame_rate(mpeg3_t *file, int stream)"
	<var: #result type: 'double '>
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3FrameRate'
		parameters: #(Oop SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_vstreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].


	self cCode: 'result =  mpeg3_frame_rate(file,aNumber)'.
	^result asOop: Float
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 20:58'!
primitiveMPEG3GenerateToc: fileHandle useSearch: timecode doStreams: streams buffer: aString
	| file bufferSize |

	"int mpeg3_generate_toc_for_Squeak(FILE *output, char *path, int timecode_search, int print_streams, char *buffer)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3GenerateToc'
		parameters: #(Oop SmallInteger Boolean  String).
	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^nil].
	bufferSize := interpreterProxy slotSizeOf: (interpreterProxy stackValue: 0).
	self cCode: 'mpeg3_generate_toc_for_Squeak(file,timecode,streams,aString,bufferSize)'.
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 20:59'!
primitiveMPEG3GetFrame: fileHandle stream: aNumber
	| file result |

	"long mpeg3_get_frame(mpeg3_t *file,int stream)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3GetFrame'
		parameters: #(Oop SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_vstreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].

	self cCode: 'result = mpeg3_get_frame(file,aNumber)'.
	^result asOop: Float.
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 20:59'!
primitiveMPEG3GetSample: fileHandle stream: aNumber
	| file result |

	"int mpeg3_video_width(mpeg3_t *file, int stream)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3GetSample'
		parameters: #(Oop SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_astreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].


	self cCode: 'result = mpeg3_get_sample(file,aNumber)'.
	^result asOop: Float
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas dtl 2010-09-28) 4/12/2006 12:07'!
primitiveMPEG3GetTime: fileHandle 
	| file result |

	"double mpeg3_get_time(mpeg3_t *file)"
	<var: #result type: 'double '>
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3GetTime'
		parameters: #(Oop).
	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^nil].
	self cCode: 'result = mpeg3_get_time(file)'.
	^result asOop: Float.
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:01'!
primitiveMPEG3HasAudio: fileHandle
	| file result |

	"int mpeg3_has_audio(mpeg3_t *file)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3HasAudio'
		parameters: #(Oop).
	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^nil].
	self cCode: 'result = mpeg3_has_audio(file)'.
	^result asOop: Boolean
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 20:59'!
primitiveMPEG3HasVideo: fileHandle
	| file result |

	"int mpeg3_has_video(mpeg3_t *file)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3HasVideo'
		parameters: #(Oop).
	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^nil].
	self cCode: 'result = mpeg3_has_video(file)'.
	^result asOop: Boolean
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'JMM (auto pragmas dtl 2010-09-28) 3/29/2006 16:50'!
primitiveMPEG3Open: path
	| mpeg3Oop index sz storage |

	"mpeg3_t* mpeg3_open(char *path)"
	<var: #index declareC: 'mpeg3_t ** index'>
	<var: #storage declareC: 'char storage[1024]'>
	self primitive: 'primitiveMPEG3Open'
		parameters: #(String).
	sz := interpreterProxy byteSizeOf: path cPtrAsOop.
	interpreterProxy ioFilename: storage fromString: path ofLength: sz resolveAliases: true.
	mpeg3Oop := interpreterProxy instantiateClass: interpreterProxy classByteArray
					indexableSize: 4.
	index := self cCoerce: (interpreterProxy firstIndexableField: mpeg3Oop)
						to: 'mpeg3_t **'.
	self cCode: '*index = mpeg3_open(storage,0); makeFileEntry(*index)'.
	^mpeg3Oop.
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'JMM (auto pragmas 12/08) 1/20/2006 18:38'!
primitiveMPEG3OpenABuffer: path size: size
	| mpeg3Oop index |

	<var: #index declareC: 'mpeg3_t ** index'>
	self primitive: 'primitiveMPEG3OpenABuffer'
		parameters: #(String SmallInteger).
	mpeg3Oop := interpreterProxy instantiateClass: interpreterProxy classByteArray
					indexableSize: 4.	
	index := self cCoerce: (interpreterProxy firstIndexableField: mpeg3Oop)
						to: 'mpeg3_t **'.
	self cCode: '*index = mpeg3_open(path,size); makeFileEntry(*index)'.
	^mpeg3Oop.
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:01'!
primitiveMPEG3PreviousFrame: fileHandle stream: aNumber
	| file result |

	"int mpeg3_previous_frame(mpeg3_t *file, int stream)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3PreviousFrame'
		parameters: #(Oop SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	file = nil ifTrue: [^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_vstreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].


	self cCode: 'result = mpeg3_previous_frame(file,aNumber)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'tpr 12/13/2020 15:35'!
primitiveMPEG3ReReadAudio: fileHandle shortArray: anArray channel: aChannelNumber samples: aSampleNumber stream: aNumber
	| file result arrayBase |

	"int mpeg3_reread_audio(mpeg3_t *file, 
		float *output_f, 
		short *output_i, 
		int channel, 
		long samples,
		int stream)"
	<var: #file type: 'mpeg3_t * '>
	<var: #arrayBase type: 'short * '>
	self primitive: 'primitiveMPEG3ReReadAudio'
		parameters: #(Oop WordArray SmallInteger SmallInteger SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_astreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].


	arrayBase := self cCoerce: anArray to: 'short *'.
	interpreterProxy failed ifTrue: [^nil].
	self cCode: 'result = mpeg3_reread_audio(file,(float *) NULL,arrayBase,aChannelNumber,aSampleNumber,aNumber)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'tpr 12/13/2020 15:34'!
primitiveMPEG3ReadAudio: fileHandle shortArray: anArray channel: aChannelNumber samples: aSampleNumber stream: aNumber
	| file result arrayBase |

	"int mpeg3_read_audio(mpeg3_t *file, 
		float *output_f, 
		short *output_i, 
		int channel, 
		long samples,
		int stream)"
	<var: #file type: 'mpeg3_t * '>
	<var: #arrayBase type: 'short * '>
	self primitive: 'primitiveMPEG3ReadAudio'
		parameters: #(Oop WordArray SmallInteger SmallInteger SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_astreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].
	arrayBase := self cCoerce: anArray to: 'short *'.
	interpreterProxy failed ifTrue: [^nil].

	self cCode: 'result = mpeg3_read_audio(file,(float *) NULL,arrayBase,aChannelNumber,aSampleNumber,aNumber)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'dtl 9/28/2010 19:17'!
primitiveMPEG3ReadFrame: fileHandle buffer: aBuffer bufferOffset: aBufferOffset x: xNumber y: yNumber w: width h: height ow: outWidth oh: outHeight colorModel: model stream: aNumber bytesPerRow: aByteNumber 
	| file result outputRowsPtr bufferBaseAddr |
	<var: #file declareC: 'mpeg3_t * file'>
	<var: #bufferBaseAddr declareC: 'unsigned char *bufferBaseAddr'>
	<var: #outputRowsPtr declareC: 'unsigned char  ** outputRowsPtr'>

	"int mpeg3_read_frame(mpeg3_t *file, 
		unsigned char **output_rows, 
		int in_x, 
		int in_y, 
		int in_w, 
		int in_h, 
		int out_w, 
		int out_h, 
		int color_model,
		int stream)"

	self primitive: 'primitiveMPEG3ReadFrameBufferOffset'
		parameters: #(Oop WordArray  SmallInteger SmallInteger  SmallInteger  SmallInteger  SmallInteger  SmallInteger  SmallInteger  SmallInteger  SmallInteger SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [ interpreterProxy success: false.  ^nil ].
	aNumber >= (self cCode: 'result = mpeg3_total_vstreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].

	bufferBaseAddr := self cCoerce: aBuffer to: 'unsigned char *'.
	self cCode: 'outputRowsPtr = (unsigned char **) memoryAllocate(1,sizeof(unsigned char*) * outHeight)'.

	0 to: outHeight-1 do: [:i | outputRowsPtr at: i put: (bufferBaseAddr + aBufferOffset + (aByteNumber*i))].
		
	self cCode: 'result = mpeg3_read_frame(file,outputRowsPtr,xNumber,yNumber,width,height,outWidth,outHeight,model,aNumber)'.
	self cCode: 'memoryFree(outputRowsPtr)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'dtl 9/28/2010 19:17'!
primitiveMPEG3ReadFrame: fileHandle buffer: aBuffer x: xNumber y: yNumber w: width h: height ow: outWidth oh: outHeight colorModel: model stream: aNumber bytesPerRow: aByteNumber 
	| file result outputRowsPtr bufferBaseAddr |
	<var: #file type: 'mpeg3_t * '>
	<var: #bufferBaseAddr type: 'unsigned char *'>
	<var: #outputRowsPtr type: 'unsigned char  ** '>

	"int mpeg3_read_frame(mpeg3_t *file, 
		unsigned char **output_rows, 
		int in_x, 
		int in_y, 
		int in_w, 
		int in_h, 
		int out_w, 
		int out_h, 
		int color_model,
		int stream)"

	self primitive: 'primitiveMPEG3ReadFrame'
		parameters: #(Oop WordArray  SmallInteger  SmallInteger  SmallInteger  SmallInteger  SmallInteger  SmallInteger  SmallInteger  SmallInteger SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [ interpreterProxy success: false.  ^nil ].
	aNumber >= (self cCode: 'result = mpeg3_total_vstreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].

	bufferBaseAddr := self cCoerce: aBuffer to: 'unsigned char *'.
	self cCode: 'outputRowsPtr = (unsigned char **) memoryAllocate(1,sizeof(unsigned char*) * outHeight)'.

	0 to: outHeight-1 do: [:i | outputRowsPtr at: i put: (bufferBaseAddr + (aByteNumber*i))].
		
	self cCode: 'result = mpeg3_read_frame(file,outputRowsPtr,xNumber,yNumber,width,height,outWidth,outHeight,model,aNumber)'.
	self cCode: 'memoryFree(outputRowsPtr)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:02'!
primitiveMPEG3SampleRate: fileHandle stream: aNumber
	| file result |

	"int mpeg3_sample_rate(mpeg3_t *file,int stream)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3SampleRate'
		parameters: #(Oop SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_astreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].


	self cCode: 'result = mpeg3_sample_rate(file,aNumber)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:03'!
primitiveMPEG3SeekPercentage: fileHandle percentage: aNumber
	| file result |

	"int mpeg3_seek_percentage(mpeg3_t *file, double percentage)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3SeekPercentage'
		parameters: #(Oop Float).
	file := self mpeg3tValueOf: fileHandle.
	aNumber < 0.0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber > 1.0 ifTrue: [interpreterProxy success: false. ^nil].
	file = nil ifTrue: [^nil].
	self cCode: 'result = mpeg3_seek_percentage(file,aNumber)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:00'!
primitiveMPEG3SetCpus: fileHandle number: cpus
	| file |

	"int mpeg3_set_cpus(mpeg3_t *file, int cpus)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3SetCpus'
		parameters: #(Oop SmallInteger).
	file := self mpeg3tValueOf: fileHandle.
	cpus < 0 ifTrue: [interpreterProxy success: false. ^nil].
	file = nil ifTrue: [^nil].
	self cCode: 'mpeg3_set_cpus(file,cpus)'.
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 20:58'!
primitiveMPEG3SetFrame: fileHandle frame: aFrameNumber stream: aNumber
	| file result |

	"int mpeg3_set_frame(mpeg3_t *file, long frame, int stream)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3SetFrame'
		parameters: #(Oop Float SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_vstreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].


	self cCode: 'result = mpeg3_set_frame(file,(long) aFrameNumber,aNumber)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 20:59'!
primitiveMPEG3SetMmx: fileHandle useMmx: mmx
	| file |

	"int mpeg3_set_mmx(mpeg3_t *file, int use_mmx)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3SetMmx'
		parameters: #(Oop Boolean).
	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^nil].
	self cCode: 'mpeg3_set_mmx(file,mmx)'.
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:02'!
primitiveMPEG3SetSample: fileHandle sample: aSampleNumber stream: aNumber
	| file result |

	"int mpeg3_set_sample(mpeg3_t *file, long sample, int stream)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3SetSample'
		parameters: #(Oop Float SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_astreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].


	aSampleNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	self cCode: 'result = mpeg3_set_sample(file,aSampleNumber,aNumber)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas dtl 2010-09-28) 4/12/2006 12:11'!
primitiveMPEG3TellPercentage: fileHandle
	| file result |

	"double mpeg3_tell_percentage(mpeg3_t *file)"
	<var: #result type: 'double '>
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3TellPercentage'
		parameters: #(Oop).
	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^nil].
	self cCode: 'result = mpeg3_tell_percentage(file)'.
	^result asOop: Float.
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:00'!
primitiveMPEG3TotalAStreams: fileHandle
	| file result |

	"int mpeg3_total_astreams(mpeg3_t *file)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3TotalAStreams'
		parameters: #(Oop).
	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	self cCode: 'result = mpeg3_total_astreams(file)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:03'!
primitiveMPEG3TotalVStreams: fileHandle
	| file result |

	"int mpeg3_total_vstreams(mpeg3_t *file)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3TotalVStreams'
		parameters: #(Oop).
	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^nil].
	self cCode: 'result = mpeg3_total_vstreams(file)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 20:58'!
primitiveMPEG3VideoFrames: fileHandle stream: aNumber
	| file result |

	"long mpeg3_video_frames(mpeg3_t *file, int stream)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3VideoFrames'
		parameters: #(Oop SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_vstreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].


	self cCode: 'result = mpeg3_video_frames(file,aNumber)'.
	^result asOop: Float.
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:01'!
primitiveMPEG3VideoHeight: fileHandle stream: aNumber
	| file result |

	"int mpeg3_video_height(mpeg3_t *file,int stream)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3VideoHeight'
		parameters: #(Oop SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_vstreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].


	self cCode: 'result = mpeg3_video_height(file,aNumber)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:01'!
primitiveMPEG3VideoWidth: fileHandle stream: aNumber
	| file result |

	"int mpeg3_video_width(mpeg3_t *file, int stream)"
	<var: #file type: 'mpeg3_t * '>
	self primitive: 'primitiveMPEG3VideoWidth'
		parameters: #(Oop SmallInteger).

	file := self mpeg3tValueOf: fileHandle.
	file = nil ifTrue: [^0].
	aNumber < 0 ifTrue: [interpreterProxy success: false. ^nil].
	aNumber >= (self cCode: 'result = mpeg3_total_vstreams(file)') ifTrue: [
		interpreterProxy success: false.  ^0 ].


	self cCode: 'result = mpeg3_video_width(file,aNumber)'.
	^result asSmallIntegerObj
! !

!Mpeg3Plugin methodsFor: 'support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:12'!
removeFileEntry: aMpegFile
	<var: #aMpegFile type: 'mpeg3_t * '>
	1 to: maximumNumberOfFilesToWatch do: 
		[:i | ((mpegFiles at: i) = aMpegFile) ifTrue: 
				[mpegFiles at: i put: 0.
				^true]].
	"Just ignore"
	^false
		
	! !

!Mpeg3Plugin methodsFor: 'support' stamp: 'eem 11/10/2020 11:20'!
shutdownModule
	<export: true>
	1 to: maximumNumberOfFilesToWatch do: 
		[:i |
		((mpegFiles at: i) ~= 0) ifTrue:
			[self mpeg3_close: (mpegFiles at: i).
			mpegFiles at: i put: 0]].
	^true! !

SmartSyntaxInterpreterPlugin subclass: #QuicktimePlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!QuicktimePlugin class methodsFor: 'translation' stamp: 'JMM 1/17/2006 20:24'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!QuicktimePlugin class methodsFor: 'translation' stamp: 'JMM 1/17/2006 20:24'!
requiresCrossPlatformFiles
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!QuicktimePlugin methodsFor: 'system primitives' stamp: 'JMM (auto pragmas 12/08) 1/17/2006 15:53'!
initialiseModule
	<export: true>
	^self sqQuicktimeInitialize ! !

!QuicktimePlugin methodsFor: 'system primitives' stamp: 'JMM (auto pragmas 12/08) 1/17/2006 15:53'!
moduleUnloaded: aModuleName 
	"The module with the given name was just unloaded. 
	Make sure we have no dangling references."
	<export: true>
	<var: #aModuleName type: 'char *'>
	(aModuleName strcmp: 'QuicktimePlugin') = 0
		ifTrue: [self sqQuicktimeShutdown]! !

!QuicktimePlugin methodsFor: 'system primitives' stamp: 'JMM 1/17/2006 23:05'!
primitiveClearFrameCompletedSemaphore: data

	self primitive: 'primitiveClearFrameCompletedSemaphore' parameters: #(SmallInteger).
	self stQuicktimeClearSemaphore: data.
	^nil! !

!QuicktimePlugin methodsFor: 'system primitives' stamp: 'JMM 1/17/2006 23:05'!
primitiveDestroyHandle: data 

	self primitive: 'primitiveDestroyHandle' parameters: #(SmallInteger).
	self stQuicktimeDestroy: data.
	^nil! !

!QuicktimePlugin methodsFor: 'system primitives' stamp: 'JMM 3/16/2006 09:32'!
primitiveDestroySurface: data 

	self primitive: 'primitiveDestroySurface' parameters: #(SmallInteger).
	self stQuicktimeDestroySurface: data.
	^nil! !

!QuicktimePlugin methodsFor: 'system primitives' stamp: 'JMM 1/17/2006 23:06'!
primitiveSetFrameCompletedSemaphore: semaIndex for: data

	self primitive: 'primitiveSetFrameCompletedSemaphore'  parameters:#(SmallInteger SmallInteger).
	self stQuicktimeSetSemaphore: semaIndex for: data.
	^nil! !

!QuicktimePlugin methodsFor: 'system primitives' stamp: 'nice 7/8/2016 03:14'!
primitiveSetGWorldPtrOntoExistingSurface: surfaceID gWorld: bitMapPtr width: width height: height rowBytes: rowBytes depth: depth movie: moviePtr
	| buffer movie |

	<var: #buffer type: #'char *'>
	<var: #movie type: #'sqIntptr_t'>
	self primitive: 'primitiveSetGWorldPtrOntoExistingSurface'  parameters:#(SmallInteger Oop SmallInteger SmallInteger SmallInteger SmallInteger Oop).
	buffer := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: bitMapPtr) to: 'char *'.
	movie := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: moviePtr) to: #'sqIntptr_t'.
	self stQuicktimeSetToExistingSurface: surfaceID gworld: buffer width: width height: height rowBytes: rowBytes depth: depth movie: movie.
	! !

!QuicktimePlugin methodsFor: 'system primitives' stamp: 'nice 7/8/2016 03:14'!
primitiveSetGWorldPtrOntoSurface: bitMapPtr width: width height: height rowBytes: rowBytes depth: depth movie: moviePtr
	| buffer movie results |

	<var: #buffer type: #'char *'>
	<var: #movie type: #'sqIntptr_t'>
	self primitive: 'primitiveSetGWorldPtrOntoSurface'  parameters:#(Oop SmallInteger SmallInteger SmallInteger SmallInteger Oop).
	buffer := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: bitMapPtr) to: 'char *'.
	movie := self cCoerce: (interpreterProxy positiveMachineIntegerValueOf: moviePtr) to: #'sqIntptr_t'.
	results := self stQuicktimeSetSurface: buffer width: width height: height rowBytes: rowBytes depth: depth movie: movie.
	^results asOop: SmallInteger ! !

SmartSyntaxInterpreterPlugin subclass: #RePlugin
	instanceVariableNames: 'netMemory numAllocs numFrees lastAlloc patternStr rcvr compileFlags pcrePtr extraPtr errorStr errorOffset matchFlags patternStrPtr errorStrBuffer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!RePlugin commentStamp: '<historical>' prior: 0!
/*	Regular Expression Plugin (This class comment becomes part of rePlugin.c)

	RePlugin translate: 'RePlugin.c' doInlining: true.

See documentation and source code for the PCRE C Library Code.  This plugin is designed to serve an object such as RePattern:

	patternStr		A 0-terminated string comprising the pattern to be compiled.
	compileFlags	An Integer representing re compiler options
	PCREBuffer		A ByteArray of regular expression bytecodes
	extraPtr			A ByteArray of match optimization data (or nil)
	errorString		A String Object For Holding an Error Message (when compile failed)
	errorOffset		The index in patternStr (0-based) where the error ocurred (when compile failed)
	matchFlags		An Integer representing re matcher options
	matchSpaceObj	An Integer array for match results and workspace during matching.

The instance variables must appear in the preceding order.  MatchSpaceObj must be allocated by the calling routine and contain at least 6*(numGroups+1) bytes.
*/
#include "pcre.h"
#include "internal.h"

/* Slight machine-specific hack for MacOS Memory Management */
#ifdef TARGET_OS_MAC
#define	malloc(ptr) NewPtr(ptr)
#define free(ptr) DisposePtr(aPointer)
#endif

/* Adjust malloc and free routines as used by PCRE */
void rePluginFree(void * aPointer);
void * rePluginMalloc(size_t anInteger);
void *(*pcre_malloc)(size_t) = rePluginMalloc;
void  (*pcre_free)(void *) = rePluginFree;
!

!RePlugin class methodsFor: 'plugin code generation' stamp: 'eem 7/15/2021 13:51'!
declareCVarsIn: cg

	cg addHeaderFile:'"rePlugin.h"'.

	"Memory Management Error Checking"
	cg var: 'netMemory' 	declareC: 'int netMemory = 0'.
	cg var: 'numAllocs' 		declareC: 'int numAllocs = 0'.
	cg var: 'numFrees' 		declareC: 'int numFrees = 0'.
	cg var: 'lastAlloc'		declareC: 'int lastAlloc = 0'.

	"Support Variables for Access to Receiver Instance Variables"
	cg var: 'patternStrPtr' type: #'const char *'.
	cg var: 'errorStrBuffer'	type: #'const char *'.
	cg var: 'errorOffset'	type: #int! !

!RePlugin class methodsFor: 'plugin code generation' stamp: 'acg 8/16/2002 22:51'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!RePlugin class methodsFor: 'translation' stamp: 'nk 11/21/2002 15:54'!
moduleName

	^'RePlugin'! !

!RePlugin class methodsFor: 'plugin code generation' stamp: 'acg 7/27/2002 20:09'!
requiresCrossPlatformFiles
	"default is ok for most, any plugin needing cross platform files must say so"
	^true! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:07'!
allocateByteArrayAndSetRcvrExtraPtrFrom: anExtraPtr

	| extraObject extraByteArrayPtr |
	<var: #extraByteArrayPtr type: 'void *'>

	anExtraPtr
		ifFalse: [extraObject := interpreterProxy nilObject]
		ifTrue: [
			"Allocate a Smalltalk ByteArray -- lastAlloc contains the length"
			extraObject := interpreterProxy
						instantiateClass: (interpreterProxy classByteArray) 
						indexableSize: (self cCode: 'sizeof(real_pcre_extra)').
			self loadRcvrFromStackAt: 0. "Assume garbage collection after instantiation"

			"Copy from the C bytecode buffer to the Smalltalk ByteArray"
			extraByteArrayPtr := interpreterProxy arrayValueOf: extraObject.	
			self cCode:'memcpy(extraByteArrayPtr, (void *) anExtraPtr, sizeof(real_pcre_extra))'].
 
	"Set rcvrErrorStr from errorStr and Return"
	self rcvrExtraPtrFrom: extraObject.
	self touch: extraByteArrayPtr.	
	^extraObject.
! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:13'!
allocateByteArrayAndSetRcvrPCREPtrFromPCRE: aPCREPtr

	| patObject patByteArrayPtr |
	<var: #patByteArrayPtr type: 'void *'>

	"Allocate a Smalltalk ByteArray -- lastAlloc contains the length"
	patObject := interpreterProxy
				instantiateClass: (interpreterProxy classByteArray) 
				indexableSize: lastAlloc.
	self loadRcvrFromStackAt: 0. "Assume garbage collection after instantiation"

	"Copy from the C bytecode buffer to the Smalltalk ByteArray"
	patByteArrayPtr := interpreterProxy arrayValueOf: patObject.	
	self cCode:'memcpy(patByteArrayPtr, (void *) aPCREPtr, lastAlloc)'.
 
	"Set rcvrErrorStr from errorStr and Return"
	self rcvrPCREBufferFrom: patObject.
	self touch: patByteArrayPtr.	
	^patObject.
! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:13'!
allocateStringAndSetRcvrErrorStrFromCStr: aCStrBuffer

	|length errorStrObj errorStrObjPtr |
	<var: #aCStrBuffer type: 'const char *'>
	<var: #errorStrObjPtr type: 'void *'>
	"Allocate errorStrObj"
	length := self cCode: 'strlen(aCStrBuffer)'.
	errorStrObj := interpreterProxy
				instantiateClass: (interpreterProxy classString) 
				indexableSize: length.
	self loadRcvrFromStackAt: 0. "Assume garbage collection after instantiation"

	"Copy aCStrBuffer to errorStrObj's buffer"
	errorStrObjPtr := interpreterProxy arrayValueOf: errorStrObj.	
	self cCode:'memcpy(errorStrObjPtr,aCStrBuffer,length)'.
	self touch: errorStrObjPtr; touch: errorStrObj.
	"Set rcvrErrorStr from errorStrObj and Return"
	self rcvrErrorStrFrom: errorStrObj.
	^errorStrObj.! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'acg (auto pragmas 12/08) 2/21/1999 22:58'!
loadRcvrFromStackAt: stackInteger

	<inline:true>
	rcvr := interpreterProxy stackObjectValue: stackInteger.
! !

!RePlugin methodsFor: 'memory management' stamp: 'acg (auto pragmas 12/08) 2/25/1999 08:36'!
primLastAlloc
	
	<export: true>
	interpreterProxy pop:1; pushInteger: lastAlloc
! !

!RePlugin methodsFor: 'memory management' stamp: 'acg (auto pragmas 12/08) 2/21/1999 23:20'!
primNetMemory 
	
	<export: true>
	interpreterProxy pop:1; pushInteger: netMemory
! !

!RePlugin methodsFor: 'memory management' stamp: 'acg (auto pragmas 12/08) 2/21/1999 23:20'!
primNumAllocs

	<export: true>
	interpreterProxy pop:1; pushInteger: numAllocs
! !

!RePlugin methodsFor: 'memory management' stamp: 'acg (auto pragmas 12/08) 2/21/1999 23:20'!
primNumFrees 
	
	<export: true>
	interpreterProxy pop:1; pushInteger: numFrees
! !

!RePlugin methodsFor: 're primitives' stamp: 'eem 12/10/2014 16:04'!
primPCRECompile

"<rcvr primPCRECompile>, where rcvr is an object with instance variables:

	'patternStr compileFlags pcrePtr extraPtr errorStr errorOffset matchFlags'	

Compile the regular expression in patternStr, and if the compilation is successful, attempt to optimize the compiled expression.  Store the results in <pcrePtr> and <extratr>, or fill errorStr with a meaningful errorString and errorOffset with an indicator where the error was found, applying compileFlags throughout.  Answer nil with a clean compile (regardless of whether an optimization is possible, and answer with the string otherwise."


	<export: true>
	self loadRcvrFromStackAt: 0.
	patternStrPtr := self rcvrPatternStrPtr.
	compileFlags := self rcvrCompileFlags.
	interpreterProxy failed ifTrue:[^ nil].

	pcrePtr := self cCode: '(sqInt) pcre_compile(patternStrPtr, compileFlags, &errorStrBuffer, &errorOffset, NULL)'.
	pcrePtr
		ifTrue: [
			self allocateByteArrayAndSetRcvrPCREPtrFromPCRE: pcrePtr.
			extraPtr := self cCode: '(sqInt) pcre_study((pcre *)pcrePtr, compileFlags, &errorStrBuffer)'.
			self allocateByteArrayAndSetRcvrExtraPtrFrom: extraPtr.
			self rePluginFree: pcrePtr asVoidPointer.
			extraPtr ifTrue: [self rePluginFree: extraPtr asVoidPointer].
			interpreterProxy failed ifTrue:[^ nil].
			interpreterProxy pop: 1 thenPush: interpreterProxy nilObject]
		ifFalse: [
			errorStr := self allocateStringAndSetRcvrErrorStrFromCStr: errorStrBuffer.
			self rcvrErrorOffsetFrom: errorOffset.
			interpreterProxy failed ifTrue:[^ nil].
			interpreterProxy pop: 1 thenPush: errorStr].! !

!RePlugin methodsFor: 're primitives' stamp: 'ar (auto pragmas 12/08) 4/4/2006 21:08'!
primPCREExec

"<rcvr primPCREExec: searchObject>, where rcvr is an object with instance variables:

	'patternStr compileFlags pcrePtr extraPtr errorStr errorOffset matchFlags'	

Apply the regular expression (stored in <pcrePtr> and <extratr>, generated from calls to primPCRECompile), to smalltalk String searchObject using <matchOptions>.  If there is no match, answer nil.  Otherwise answer a ByteArray of offsets representing the results of the match."

	| searchObject searchBuffer length  result matchSpacePtr matchSpaceSize |
	<export: true>
	<var:#searchBuffer	type: 'char *'>
	<var:#matchSpacePtr	type: 'int *'>
	
	"Load Parameters"
	searchObject := interpreterProxy stackObjectValue: 0.	
	searchBuffer := interpreterProxy arrayValueOf: searchObject.
	length := interpreterProxy byteSizeOf: searchObject.
	self loadRcvrFromStackAt: 1.
	"Load Instance Variables"
	pcrePtr := self rcvrPCREBufferPtr.
	extraPtr := self rcvrExtraPtr.
	matchFlags := self rcvrMatchFlags.
	matchSpacePtr := self rcvrMatchSpacePtr.
	matchSpaceSize := self rcvrMatchSpaceSize.

	interpreterProxy failed ifTrue:[^ nil].
	
	result := self 
		cCode: 'pcre_exec((pcre *)pcrePtr, (pcre_extra *)extraPtr, 
				searchBuffer, length, 0, matchFlags, matchSpacePtr, matchSpaceSize)'.

	interpreterProxy pop: 2; pushInteger: result.

	"empty call so compiler doesn't bug me about variables not used"
	self touch: searchBuffer; touch: matchSpacePtr; touch: matchSpaceSize; touch: length
! !

!RePlugin methodsFor: 're primitives' stamp: 'eem 8/26/2013 11:46'!
primPCREExecfromto

"<rcvr primPCREExec: searchObject from: fromInteger to: toInteger>, where rcvr is an object with instance variables:

	'patternStr compileFlags pcrePtr extraPtr errorStr errorOffset matchFlags'	

Apply the regular expression (stored in <pcrePtr> and <extratr>, generated from calls to primPCRECompile), to smalltalk String searchObject using <matchOptions>, beginning at offset <fromInteger> and continuing until offset <toInteger>.  If there is no match, answer nil.  Otherwise answer a ByteArray of offsets representing the results of the match."

	| searchObject searchBuffer length  result matchSpacePtr matchSpaceSize fromInteger toInteger |
	<export: true>
	<var:#searchBuffer	type: 'char *'>
	<var:#matchSpacePtr	type: 'int *'>
	
	"Load Parameters"
	toInteger := interpreterProxy stackIntegerValue: 0.
	fromInteger := interpreterProxy stackIntegerValue: 1.
	searchObject := interpreterProxy stackObjectValue: 2.	
	searchBuffer := interpreterProxy arrayValueOf: searchObject.
	length := interpreterProxy byteSizeOf: searchObject.
	self loadRcvrFromStackAt: 3.

	"Validate parameters"
	interpreterProxy success: (1 <= fromInteger).
	interpreterProxy success: (toInteger<=length).
	fromInteger := fromInteger - 1. "Smalltalk offsets are 1-based"
	interpreterProxy success: (fromInteger<=toInteger).

	"adjust length, searchBuffer"
	length := toInteger - fromInteger.
	searchBuffer := searchBuffer + fromInteger.

	"Load Instance Variables"
	pcrePtr := self rcvrPCREBufferPtr.
	extraPtr := self rcvrExtraPtr.
	matchFlags := self rcvrMatchFlags.
	matchSpacePtr := self rcvrMatchSpacePtr.
	matchSpaceSize := self rcvrMatchSpaceSize.
	interpreterProxy failed ifTrue:[^ nil].
	
	result := self 
		cCode: 'pcre_exec((pcre *)pcrePtr, (pcre_extra *)extraPtr, 
				searchBuffer, length, 0, matchFlags, matchSpacePtr, matchSpaceSize)'.
	interpreterProxy pop: 4; pushInteger: result.

	"empty call so compiler doesn't bug me about variables not used"
	self touch: searchBuffer; touch: matchSpacePtr; touch: matchSpaceSize; touch: length
! !

!RePlugin methodsFor: 're primitives' stamp: 'eem 8/28/2013 10:47'!
primPCRENumSubPatterns

"<rcvr primPCRENumSubPatterns>, where rcvr is an object with instance variables:

	'patternStr compileFlags pcrePtr extraPtr errorStr errorOffset matchFlags'	

Return the number of subpatterns captured by the compiled pattern."

	<export: true>
	| ncap |
	"Load Parameters"
	self loadRcvrFromStackAt: 0.
	"Load Instance Variables"
	pcrePtr := self rcvrPCREBufferPtr.
	self cCode: 'pcre_fullinfo((const pcre *)pcrePtr, NULL, PCRE_INFO_CAPTURECOUNT, &ncap)'
		inSmalltalk: [ncap := -1].
	interpreterProxy pop: 1; pushInteger: ncap.
! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'acg (auto pragmas 12/08) 2/21/1999 21:20'!
rcvrCompileFlags

	<inline:true>
	^interpreterProxy fetchInteger: 1 ofObject: rcvr.
! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'acg (auto pragmas 12/08) 2/21/1999 22:46'!
rcvrErrorOffsetFrom: anInteger

	<inline: true>
	interpreterProxy storeInteger: 5 ofObject: rcvr withValue: anInteger.
! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'acg (auto pragmas 12/08) 2/24/1999 20:53'!
rcvrErrorStrFrom: aString

	<inline: true>
	interpreterProxy 
		storePointer: 4
		ofObject: rcvr 
		withValue: aString.
! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'eem 12/10/2014 16:00'!
rcvrExtraPtr

	|extraObj|
	<inline: true>
	extraObj := interpreterProxy fetchPointer: 3 ofObject: rcvr.
	^self 
		cCoerce: (extraObj = interpreterProxy nilObject
					ifTrue: [nil]
					ifFalse: [interpreterProxy arrayValueOf: extraObj])
		to: #sqInt! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'acg (auto pragmas 12/08) 2/27/1999 23:42'!
rcvrExtraPtrFrom: aByteArrayOrNilObject

	<inline: true>
	interpreterProxy 
		storePointer: 3 
		ofObject: rcvr 
		withValue: aByteArrayOrNilObject! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'acg (auto pragmas 12/08) 2/21/1999 21:19'!
rcvrMatchFlags

	<inline: true>
	^interpreterProxy fetchInteger: 6 ofObject: rcvr.
! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'eem 7/31/2020 11:59'!
rcvrMatchSpacePtr

	<inline: true>
	<returnTypeC: 'int *'>
	^self
		cCoerce: (interpreterProxy fetchArray: 7 ofObject: rcvr)
		to: #'int *'! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'acg (auto pragmas 12/08) 2/25/1999 00:52'!
rcvrMatchSpaceSize

	<inline: true>
	^(interpreterProxy byteSizeOf: (interpreterProxy fetchPointer: 7 ofObject: rcvr))//4.! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'acg (auto pragmas 12/08) 2/24/1999 21:33'!
rcvrPCREBufferFrom: aByteArray

	<inline: true>
	interpreterProxy 
		storePointer: 2 
		ofObject: rcvr 
		withValue: aByteArray! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'eem 12/10/2014 16:00'!
rcvrPCREBufferPtr
	<inline: true>
	^self
		cCoerce: (interpreterProxy fetchArray: 2 ofObject: rcvr)
		to: #sqInt! !

!RePlugin methodsFor: 'rcvr linkage' stamp: 'eem 7/31/2020 11:58'!
rcvrPatternStrPtr

	<inline: true>
	<returnTypeC: 'char *'>
	^self 
		cCoerce: (interpreterProxy fetchArray: 0 ofObject: rcvr) 
		to: #'char *'.! !

!RePlugin methodsFor: 'memory management' stamp: 'eem 11/7/2012 12:43'!
rePluginFree: aPointer
	"Free a block of fixed memory allocated with rePluginMalloc.  Instrumented version of C free() to facilitate leak analysis from Smalltalk.   OS-specific variations on malloc/free, such as with MacOS, are handled by adding a C macro to the header file redefining malloc/free -- see the class comment"

	<inline: true>
	<var: #aPointer type: 'void * '>
	<returnTypeC: 'void'>

	numFrees := numFrees + 1.
	aPointer notNil ifTrue: [self free: aPointer]	! !

!RePlugin methodsFor: 'memory management' stamp: 'eem 11/7/2012 12:43'!
rePluginMalloc: anInteger
	"Allocate a block of fixed memory using C calls to malloc().  Instrumented to facilitate leak analysis from Smalltalk.  Set global lastAlloc to anInteger.  OS-specific variations on malloc/free, such as with MacOS, are handled by adding a C macro to the header file redefining malloc/free -- see the class comment"

	| aPointer |
	<inline: true>
	<var: #anInteger type: 'size_t '>
	<var: #aPointer type: 'void *'>
	<returnTypeC: 'void *'>
	numAllocs := numAllocs + 1.
	(aPointer := self malloc: anInteger) notNil ifTrue:
		[lastAlloc := anInteger].
	^aPointer
! !

SmartSyntaxInterpreterPlugin subclass: #SerialPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!SerialPlugin commentStamp: '<historical>' prior: 0!
Implement the serial port primitives.  Since it requires platform support it will only be built when supported on your platform.

IMPORTANT: So far, we are converting everytime a string into a char* and then we look for it in the ports array. That can be optimized a lot by just answering the external handler (the position in the array perhaps) after open and using it instead the name.
Also, using open by id functions doesn't help because internally they are also converted into a char* (using sprintf).

If needed, that can be optimized then. !

!SerialPlugin class methodsFor: 'translation' stamp: 'tpr 5/23/2001 17:11'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!SerialPlugin class methodsFor: 'translation' stamp: 'tpr 11/29/2000 22:38'!
requiresPlatformFiles
	"this plugin requires platform specific files in order to work"
	^true! !

!SerialPlugin methodsFor: 'private-support' stamp: 'eem 12/6/2018 15:09'!
copyPortNameToCString: portName
	<returnTypeC: #'char *'>
	| port portNameSize |
	<inline: #always>
	<var: 'port' type: #'char *'>
	portNameSize := interpreterProxy slotSizeOf: (portName asOop: String).
	port := self alloca: portNameSize + 1.
	self memcpy: port _: portName _: portNameSize.
	port at: portNameSize put: 0.
	^port! !

!SerialPlugin methodsFor: 'initialize-release' stamp: 'ar (auto pragmas 12/08) 5/12/2000 16:53'!
initialiseModule
	<export: true>
	^self cCode: 'serialPortInit()' inSmalltalk:[true]! !

!SerialPlugin methodsFor: 'primitives' stamp: 'TPR 2/17/2000 18:16'!
primitiveSerialPortClose: portNum

	self primitive: 'primitiveSerialPortClose'
		parameters: #(SmallInteger).
	self serialPortClose: portNum! !

!SerialPlugin methodsFor: 'primitives' stamp: 'eem 12/6/2018 14:46'!
primitiveSerialPortCloseByName: portName
	| port |
	self primitive: 'primitiveSerialPortCloseByName'
		parameters: #(String).
	port := self copyPortNameToCString: portName.
	self serialPortCloseByName: port! !

!SerialPlugin methodsFor: 'primitives' stamp: 'eem 12/6/2018 14:42'!
primitiveSerialPortOpen: portNum baudRate: baudRate stopBitsType: stopBitsType parityType: parityType dataBits: dataBits inFlowControlType: inFlowControl outFlowControlType: outFlowControl xOnByte: xOnChar xOffByte: xOffChar

	self primitive: 'primitiveSerialPortOpen'
		parameters: #(SmallInteger SmallInteger SmallInteger SmallInteger SmallInteger SmallInteger SmallInteger SmallInteger SmallInteger).

	self serialPortOpen: portNum _: baudRate _: stopBitsType _: parityType _: dataBits
					_: inFlowControl _: outFlowControl _: xOnChar _: xOffChar! !

!SerialPlugin methodsFor: 'primitives' stamp: 'eem 12/6/2018 14:41'!
primitiveSerialPortOpenByName: portName baudRate: baudRate stopBitsType: stopBitsType parityType: parityType dataBits: dataBits inFlowControlType: inFlowControl outFlowControlType: outFlowControl xOnByte: xOnChar xOffByte: xOffChar

	| port |
	self primitive: 'primitiveSerialPortOpenByName'
		parameters: #(String SmallInteger SmallInteger SmallInteger SmallInteger SmallInteger SmallInteger SmallInteger SmallInteger).

	port := self copyPortNameToCString: portName.
	self serialPortOpenByName: port _: baudRate _: stopBitsType _: parityType _: dataBits
							_: inFlowControl _: outFlowControl _: xOnChar _: xOffChar! !

!SerialPlugin methodsFor: 'primitives' stamp: 'eem 12/6/2018 14:43'!
primitiveSerialPortRead: portNum into: array startingAt: startIndex count: count 
	| bytesRead |
	self primitive: 'primitiveSerialPortRead'
		parameters: #(SmallInteger ByteArray SmallInteger SmallInteger).

	(startIndex >= 1 and: [startIndex + count - 1 <= (interpreterProxy byteSizeOf: array cPtrAsOop)]) ifTrue:
		[bytesRead := self serialPortReadInto: portNum _: count _: array + startIndex - 1.
		 ^bytesRead asSmallIntegerObj].
	interpreterProxy success: false! !

!SerialPlugin methodsFor: 'primitives' stamp: 'eem 12/6/2018 14:38'!
primitiveSerialPortReadByName: portName into: array startingAt: startIndex count: count
	| port bytesRead |
	<var: #port type: #'char *'>
	self primitive: 'primitiveSerialPortReadByName'
		parameters: #(String  ByteArray SmallInteger SmallInteger).

	(startIndex >= 1 and: [startIndex + count - 1 <= (interpreterProxy byteSizeOf: array cPtrAsOop)]) ifTrue:
		[port := self copyPortNameToCString: portName.
		 bytesRead := self serialPortReadIntoByName: port _: count _: array + startIndex - 1.
		 ^bytesRead asSmallIntegerObj].
	interpreterProxy success: false! !

!SerialPlugin methodsFor: 'primitives' stamp: 'eem 12/6/2018 14:38'!
primitiveSerialPortWrite: portNum from: array startingAt: startIndex count: count 
	| bytesWritten |
	self primitive: 'primitiveSerialPortWrite'
		parameters: #(SmallInteger ByteArray SmallInteger SmallInteger).

	(startIndex >= 1 and: [startIndex + count - 1 <= (interpreterProxy byteSizeOf: array cPtrAsOop)]) ifTrue:
		[bytesWritten := self serialPort: portNum Write: count From: array + startIndex - 1.
		 ^bytesWritten asSmallIntegerObj].
	interpreterProxy success: false! !

!SerialPlugin methodsFor: 'primitives' stamp: 'eem 12/6/2018 14:38'!
primitiveSerialPortWriteByName: portName from: array startingAt: startIndex count: count 
	| bytesWritten port |
	self primitive: 'primitiveSerialPortWriteByName'
		parameters: #(String ByteArray SmallInteger SmallInteger).

	(startIndex >= 1 and: [startIndex + count - 1 <= (interpreterProxy byteSizeOf: array cPtrAsOop)]) ifTrue:
		[port := self copyPortNameToCString: portName.
		 bytesWritten := self serialPortWriteFromByName: port _: count _: array + startIndex - 1.
		 ^bytesWritten asSmallIntegerObj].
	interpreterProxy success: false! !

!SerialPlugin methodsFor: 'initialize-release' stamp: 'ar (auto pragmas 12/08) 5/12/2000 16:55'!
shutdownModule
	<export: true>
	^self cCode: 'serialPortShutdown()' inSmalltalk:[true]! !

!SmartSyntaxInterpreterPlugin class methodsFor: 'private' stamp: 'tpr 6/9/2003 16:36'!
codeGeneratorClass
	"return the appropriate class of code generator for this kind ofplugin"

	^SmartSyntaxPluginCodeGenerator! !

!SmartSyntaxInterpreterPlugin class methodsFor: 'instance creation' stamp: 'tpr 6/28/2003 17:28'!
doPrimitive: primitiveName withArguments: argArray
	| proxy plugin |
	proxy := InterpreterProxy new.
	proxy loadStackFrom: thisContext sender.
	plugin := (self simulatorClass ifNil: [self]) new.
	plugin setInterpreter: proxy.
	^plugin perform: primitiveName asSymbol withArguments: argArray! !

!SmartSyntaxInterpreterPlugin class methodsFor: 'translation' stamp: 'eem 3/4/2009 16:44'!
implicitReturnTypeFor: aSelector
	"Answer the return type for methods that don't have an explicit return."
	^#void! !

!SmartSyntaxInterpreterPlugin class methodsFor: 'translation' stamp: 'eem 10/2/2010 10:01'!
prepareToBeAddedToCodeGenerator: aCodeGen
	aCodeGen removeVariable: 'simulator'! !

!SmartSyntaxInterpreterPlugin class methodsFor: 'translation' stamp: 'tpr 6/9/2003 16:44'!
shouldBeTranslated
"SmartSyntaxInterpreterPlugin should not be translated but its subclasses should"
	^self ~= SmartSyntaxInterpreterPlugin! !

!SmartSyntaxInterpreterPlugin class methodsFor: 'simulation' stamp: 'eem 10/25/2018 11:13'!
simulatorClass
	"For running from Smalltalk - answer a class that can be used to simulate the receiver, or nil if you want the primitives in this module to always fail, causing simulation to fall through to the Smalltalk code."

	^SmartSyntaxPluginSimulator! !

!SmartSyntaxInterpreterPlugin class methodsFor: 'simulation' stamp: 'eem 10/26/2018 09:21'!
simulatorForInterpreterInterface: objectMemoryOrInterpreterProxy
	"Answer an uninitialized (in the sense that the plugin has been sent initialize, but not initialiseModule)
	 simulator for the receiver, or nil, if the receiver answers nil to simulatorClass.  In this case, answer an
	 instance of SmartSyntaxPluginSimulator wrapped around an uniniialized simulator for the receiver."

	^self simulatorClass ifNotNil: "Maybe specific, or the general SmartSyntaxPluginSimulator"
		[:simClass| | actualPluginSimClass |
		 actualPluginSimClass := simClass == SmartSyntaxPluginSimulator
									ifTrue: [self subclasses 
											detect: [:sc| sc name endsWith: 'Simulator']
											ifNone: [self]]
									ifFalse: [simClass].
		 SmartSyntaxPluginSimulator new
			actualPlugin: actualPluginSimClass new signatureClass: self;
			setInterpreter: objectMemoryOrInterpreterProxy;
			yourself]! !

!SmartSyntaxInterpreterPlugin class methodsFor: 'translation' stamp: 'sr 12/23/2001 22:24'!
translateDoInlining: inlineFlag locally: localFlag debug: debugFlag 
	^ self
		translate: self moduleName , '.c'
		doInlining: inlineFlag
		locally: localFlag
		debug: debugFlag! !

!SmartSyntaxInterpreterPlugin methodsFor: 'simulation' stamp: 'eem 12/28/2016 22:06'!
primitive: primName
	<doNotGenerate>
	^simulator primitive: primName! !

!SmartSyntaxInterpreterPlugin methodsFor: 'simulation' stamp: 'eem 10/2/2010 10:05'!
primitive: primName parameters: parms
	<doNotGenerate>
	^simulator primitive: primName parameters: parms! !

!SmartSyntaxInterpreterPlugin methodsFor: 'simulation' stamp: 'eem 10/2/2010 10:13'!
primitive: primName parameters: anArray receiver: aClassSymbol
	<doNotGenerate>
	^simulator primitive: primName parameters: anArray receiver: aClassSymbol! !

!SmartSyntaxInterpreterPlugin methodsFor: 'accessing' stamp: 'eem 10/3/2010 08:31'!
simulator
	<doNotGenerate>
	^simulator! !

!SmartSyntaxInterpreterPlugin methodsFor: 'accessing' stamp: 'eem 10/2/2010 10:02'!
simulator: aSmartSyntaxPluginSimulator
	<doNotGenerate>
	simulator := aSmartSyntaxPluginSimulator! !

SmartSyntaxInterpreterPlugin subclass: #SocketPlugin
	instanceVariableNames: 'sDSAfn sHSAfn sCCTPfn sCCLOPfn sCCSOTfn'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!SocketPlugin commentStamp: 'tpr 5/2/2003 15:49' prior: 0!
Implement the socket and resolver primitives.  Since it requires platform support it will only be built when supported on your platform!

!SocketPlugin class methodsFor: 'translation' stamp: 'eem 1/7/2018 10:16'!
declareCVarsIn: aCCodeGenerator

	"Hmmm, we could use 
		self declareC:  (self instVarNames select: [:ivn| ivn first = $s and: [ivn second isUppercase]])
			as: #'void *'
			in: aCCodeGenerator."
	aCCodeGenerator var: 'sDSAfn'	type: 'void *'.
	aCCodeGenerator var: 'sHSAfn'	type: 'void *'.
	aCCodeGenerator var: 'sCCTPfn'	type: 'void *'.
	aCCodeGenerator var: 'sCCLOPfn'	type: 'void *'.
	aCCodeGenerator var: 'sCCSOTfn'	type: 'void *'.
	aCCodeGenerator addHeaderFile: '"SocketPlugin.h"'! !

!SocketPlugin class methodsFor: 'translation' stamp: 'tpr 5/23/2001 17:11'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!SocketPlugin class methodsFor: 'translation' stamp: 'tpr 11/29/2000 22:38'!
requiresPlatformFiles
	"this plugin requires platform specific files in order to work"
	^true! !

!SocketPlugin methodsFor: 'initialize-release' stamp: 'eem 12/28/2016 22:55'!
initialiseModule
	<export: true>
	sDSAfn := interpreterProxy ioLoadFunction: 'secDisableSocketAccess' From: 'SecurityPlugin'.
	sHSAfn := interpreterProxy ioLoadFunction: 'secHasSocketAccess' From: 'SecurityPlugin'.
	sCCTPfn := interpreterProxy ioLoadFunction: 'secCanConnectToPort' From: 'SecurityPlugin'.
	sCCLOPfn := interpreterProxy ioLoadFunction: 'secCanListenOnPort' From: 'SecurityPlugin'.
	sCCSOTfn := interpreterProxy ioLoadFunction: 'secCanCreateSocketOfType' From: 'SecurityPlugin'.
	^self socketInit! !

!SocketPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:16'!
intToNetAddress: addr
	"Convert the given 32-bit integer into an internet network address represented as a four-byte ByteArray."

	| netAddressOop naPtr |
	<var: #naPtr type: 'char * '>

	netAddressOop :=
		interpreterProxy instantiateClass: interpreterProxy classByteArray
			indexableSize: 4.
	naPtr := netAddressOop asCharPtr.
	naPtr at: 0 put: (self cCoerce: ((addr >> 24) bitAnd: 16rFF) to: 'char').
	naPtr at: 1 put: (self cCoerce: ((addr >> 16) bitAnd: 16rFF) to: 'char').
	naPtr at: 2 put: (self cCoerce: ((addr >> 8) bitAnd: 16rFF) to: 'char').
	naPtr at: 3 put: (self cCoerce: (addr bitAnd: 16rFF) to: 'char').
	^ netAddressOop! !

!SocketPlugin methodsFor: 'initialize-release' stamp: 'JMM (auto pragmas 12/08) 1/21/2002 11:10'!
moduleUnloaded: aModuleName
	"The module with the given name was just unloaded.
	Make sure we have no dangling references."
	<export: true>
	<var: #aModuleName type: 'char *'>
	(aModuleName strcmp: 'SecurityPlugin') = 0 ifTrue:[
		"The security plugin just shut down. How odd."
		sDSAfn := sHSAfn := sCCTPfn := sCCLOPfn := sCCSOTfn := 0.
	].! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 8/30/2020 17:41'!
netAddressToInt: ptrToByteArray
	"Convert the given internet network address (represented as a four-byte ByteArray) into a 32-bit integer. Fail if the given ptrToByteArray does not appear to point to a four-byte ByteArray."

	<var: #ptrToByteArray type: #'unsigned char *'>
	| sz |
	<inline: #always>
	sz := interpreterProxy byteSizeOf: ptrToByteArray cPtrAsOop.
	sz = 4 ifFalse: [^interpreterProxy primitiveFail].
	^ (ptrToByteArray at: 3	) +
		((ptrToByteArray at: 2) <<8) +
		((ptrToByteArray at: 1) <<16) +
		((ptrToByteArray at: 0) <<24)! !

!SocketPlugin methodsFor: 'security primitives' stamp: 'eem 2/25/2016 13:26'!
primitiveDisableSocketAccess
	<export: true>
	"If the security plugin can be loaded, use it to turn off socket access
	 If not, assume it's ok"
	sDSAfn ~= 0 ifTrue:
		[self cCode: '((sqInt (*) (void)) sDSAfn)()']! !

!SocketPlugin methodsFor: 'security primitives' stamp: 'eem 12/30/2016 14:17'!
primitiveHasSocketAccess
	self primitive: 'primitiveHasSocketAccess'.
	"If the security plugin can be loaded, use it to check . 
	 If not, assume it's ok"
	^(sHSAfn = 0
	  or: [self cCode: ' ((sqInt (*) (void)) sHSAfn)()' inSmalltalk: [true]]) asBooleanObj! !

!SocketPlugin methodsFor: 'primitives' stamp: 'TPR 2/18/2000 10:39'!
primitiveInitializeNetwork: resolverSemaIndex

	| err |
	self primitive: 'primitiveInitializeNetwork'
		parameters: #(SmallInteger).
	err := self sqNetworkInit: resolverSemaIndex.
	interpreterProxy success: err = 0! !

!SocketPlugin methodsFor: 'primitives' stamp: 'TPR 2/18/2000 11:47'!
primitiveResolverAbortLookup

	self primitive: 'primitiveResolverAbortLookup'.

	self sqResolverAbort! !

!SocketPlugin methodsFor: 'primitives' stamp: 'TPR 3/21/2000 16:47'!
primitiveResolverAddressLookupResult
	| sz s |
	self primitive: 'primitiveResolverAddressLookupResult'.
	sz := self sqResolverAddrLookupResultSize.

	interpreterProxy failed
		ifFalse: [s := interpreterProxy instantiateClass: interpreterProxy classString indexableSize: sz.
			self sqResolverAddrLookup: s asCharPtr Result: sz].
	^ s! !

!SocketPlugin methodsFor: 'primitives' stamp: 'TPR 2/18/2000 10:10'!
primitiveResolverError

	self primitive: 'primitiveResolverError'.
	^ self sqResolverError asSmallIntegerObj! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp 6/8/2007 18:01'!
primitiveResolverGetAddressInfoFamily

	| family |
	self primitive: 'primitiveResolverGetAddressInfoFamily'
		parameters: #().
	interpreterProxy failed
		ifFalse:
			[family := self sqResolverGetAddressInfoFamily.
			 ^family asSmallIntegerObj]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp 6/8/2007 17:23'!
primitiveResolverGetAddressInfoHost: hostName service: servName flags: flags family: family type: type protocol: protocol

	| hostSize servSize |
	self primitive: 'primitiveResolverGetAddressInfo'
		parameters: #(String String SmallInteger SmallInteger SmallInteger SmallInteger).
	interpreterProxy failed
		ifFalse:
			[hostSize := interpreterProxy byteSizeOf: hostName cPtrAsOop.
			 servSize := interpreterProxy byteSizeOf: servName cPtrAsOop.
			 self
			 	sqResolverGetAddressInfoHost: hostName Size: hostSize Service: servName Size: servSize
					Flags: flags Family: family Type: type Protocol: protocol]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp 6/8/2007 18:17'!
primitiveResolverGetAddressInfoNext

	| more |
	self primitive: 'primitiveResolverGetAddressInfoNext'
		parameters: #().
	more := self sqResolverGetAddressInfoNext.
	interpreterProxy failed ifTrue: [^nil].
	^more asBooleanObj! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp 6/8/2007 18:01'!
primitiveResolverGetAddressInfoProtocol

	| protocol |
	self primitive: 'primitiveResolverGetAddressInfoProtocol'
		parameters: #().
	interpreterProxy failed
		ifFalse:
			[protocol := self sqResolverGetAddressInfoProtocol.
			 ^protocol asSmallIntegerObj]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp 6/8/2007 17:33'!
primitiveResolverGetAddressInfoResult: socketAddress

	| addrSize |
	self primitive: 'primitiveResolverGetAddressInfoResult'
		parameters: #(ByteArray).
	interpreterProxy failed
		ifFalse:
			[addrSize := interpreterProxy byteSizeOf: socketAddress cPtrAsOop.
			 self sqResolverGetAddressInfoResult: socketAddress Size: addrSize]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp 6/8/2007 17:40'!
primitiveResolverGetAddressInfoSize

	| size |
	self primitive: 'primitiveResolverGetAddressInfoSize'
		parameters: #().
	interpreterProxy failed
		ifFalse:
			[size := self sqResolverGetAddressInfoSize.
			 ^size asSmallIntegerObj]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp 6/8/2007 18:01'!
primitiveResolverGetAddressInfoType

	| type |
	self primitive: 'primitiveResolverGetAddressInfoType'
		parameters: #().
	interpreterProxy failed
		ifFalse:
			[type := self sqResolverGetAddressInfoType.
			 ^type asSmallIntegerObj]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp (auto pragmas dtl 2010-09-28) 6/8/2007 18:59'!
primitiveResolverGetNameInfo: socketAddress flags: flags

	| addrSize addrBase |
	<var: #addrBase type: 'char *'>
	self primitive: 'primitiveResolverGetNameInfo'
		parameters: #(Oop SmallInteger).
	interpreterProxy failed
		ifFalse:
			[addrSize := interpreterProxy byteSizeOf: socketAddress.
			 addrBase := self cCoerce: (interpreterProxy firstIndexableField: socketAddress) to: 'char *'.
			 self sqResolverGetNameInfo: addrBase Size: addrSize Flags: flags]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp 6/8/2007 18:49'!
primitiveResolverGetNameInfoHostResult: socketName

	| addrSize |
	self primitive: 'primitiveResolverGetNameInfoHostResult'
		parameters: #(String).
	interpreterProxy failed
		ifFalse:
			[addrSize := interpreterProxy byteSizeOf: socketName cPtrAsOop.
			 self sqResolverGetNameInfoHostResult: socketName Size: addrSize]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp 6/8/2007 18:38'!
primitiveResolverGetNameInfoHostSize

	| size |
	self primitive: 'primitiveResolverGetNameInfoHostSize'
		parameters: #().
	interpreterProxy failed
		ifFalse:
			[size := self sqResolverGetNameInfoHostSize.
			 ^size asSmallIntegerObj]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp 6/8/2007 18:50'!
primitiveResolverGetNameInfoServiceResult: socketName

	| addrSize |
	self primitive: 'primitiveResolverGetNameInfoServiceResult'
		parameters: #(String).
	interpreterProxy failed
		ifFalse:
			[addrSize := interpreterProxy byteSizeOf: socketName cPtrAsOop.
			 self sqResolverGetNameInfoServiceResult: socketName Size: addrSize]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp 6/8/2007 18:38'!
primitiveResolverGetNameInfoServiceSize

	| size |
	self primitive: 'primitiveResolverGetNameInfoServiceSize'
		parameters: #().
	interpreterProxy failed
		ifFalse:
			[size := self sqResolverGetNameInfoServiceSize.
			 ^size asSmallIntegerObj]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp 6/9/2007 09:03'!
primitiveResolverHostNameResult: nameString

	| nameSize |
	self primitive: 'primitiveResolverHostNameResult'
		parameters: #(String).
	interpreterProxy failed
		ifFalse:
			[nameSize := interpreterProxy byteSizeOf: nameString cPtrAsOop.
			 self sqResolverHostNameResult: nameString Size: nameSize]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp 6/9/2007 09:00'!
primitiveResolverHostNameSize

	| size |
	self primitive: 'primitiveResolverHostNameSize'
		parameters: #().
	interpreterProxy failed
		ifFalse:
			[size := self sqResolverHostNameSize.
			 interpreterProxy failed ifFalse: [^size asSmallIntegerObj]]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'TPR 2/18/2000 10:10'!
primitiveResolverLocalAddress

	| addr |
	self primitive: 'primitiveResolverLocalAddress'.
	addr := self sqResolverLocalAddress.
	^self intToNetAddress: addr! !

!SocketPlugin methodsFor: 'primitives' stamp: 'TPR 2/18/2000 10:11'!
primitiveResolverNameLookupResult

	| addr |
	self primitive: 'primitiveResolverNameLookupResult'.
	addr := self sqResolverNameLookupResult.
	^self intToNetAddress: addr! !

!SocketPlugin methodsFor: 'primitives' stamp: 'TPR 4/25/2000 14:42'!
primitiveResolverStartAddressLookup: address

	| addr |
	self primitive: 'primitiveResolverStartAddressLookup'
		parameters: #(ByteArray).
	addr := self netAddressToInt: (self cCoerce: address to: 'unsigned char *').
	interpreterProxy failed ifFalse: [
		self sqResolverStartAddrLookup: addr]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'TPR 2/18/2000 10:19'!
primitiveResolverStartNameLookup: name

	| sz |
	self primitive: 'primitiveResolverStartNameLookup'
		parameters: #(String).
	interpreterProxy failed ifFalse:  [
		sz := interpreterProxy byteSizeOf: name cPtrAsOop.
		self sqResolverStartName: name Lookup: sz]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'TPR 2/18/2000 10:20'!
primitiveResolverStatus

	| status |
	self primitive: 'primitiveResolverStatus'.
	status := self sqResolverStatus.
	^status asSmallIntegerObj! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp (auto pragmas dtl 2010-09-28) 6/9/2007 09:36'!
primitiveSocket: socket bindTo: socketAddress

	| addrSize addrBase s  |
	<var: #addrBase type: 'char *'>
	<var: #s type: 'SocketPtr'>
	self primitive: 'primitiveSocketBindTo' parameters: #(#Oop #Oop).
	s := self socketValueOf: socket.
	addrSize := interpreterProxy byteSizeOf: socketAddress.
	addrBase := self cCoerce: (interpreterProxy firstIndexableField: socketAddress) to: 'char *'.
	interpreterProxy failed ifFalse: [self sqSocket: s BindToAddress: addrBase Size: addrSize]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 3/26/2006 21:32'!
primitiveSocket: socket bindTo: address port: port 
	| addr s  |
	<var: #s declareC: 'SocketPtr s'>
	self primitive: 'primitiveSocketBindToPort' parameters: #(#Oop #ByteArray #SmallInteger ).
	addr := self
				netAddressToInt: (self cCoerce: address to: 'unsigned char *').
	s := self socketValueOf: socket.
	interpreterProxy failed 
		ifFalse:[self sqSocket: s BindTo: addr Port: port]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp (auto pragmas dtl 2010-09-28) 6/8/2007 22:21'!
primitiveSocket: socket connectTo: socketAddress

	| addrSize addrBase s  |
	<var: #addrBase type: 'char *'>
	<var: #s type: 'SocketPtr'>
	self primitive: 'primitiveSocketConnectTo' parameters: #(#Oop #Oop).
	s := self socketValueOf: socket.
	addrSize := interpreterProxy byteSizeOf: socketAddress.
	addrBase := self cCoerce: (interpreterProxy firstIndexableField: socketAddress) to: 'char *'.
	interpreterProxy failed ifFalse: [self sqSocket: s ConnectToAddress: addrBase Size: addrSize]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 2/25/2016 13:28'!
primitiveSocket: socket connectTo: address port: port 
	| addr s okToConnect  |
	<var: #s type: 'SocketPtr'>
	self primitive: 'primitiveSocketConnectToPort' parameters: #(#Oop #ByteArray #SmallInteger ).
	addr := self netAddressToInt: (self cCoerce: address to: 'unsigned char *').
	"If the security plugin can be loaded, use it to check for permission.
	If not, assume it's ok"
	sCCTPfn ~= 0 ifTrue:
		[okToConnect := self cCode: ' ((sqInt (*) (sqInt, sqInt)) sCCTPfn)(addr, port)'.
		 okToConnect ifFalse:
			[^ interpreterProxy primitiveFail]].
	s := self socketValueOf: socket.
	interpreterProxy failed ifFalse:
		[self sqSocket: s ConnectTo: addr Port: port]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 12:11'!
primitiveSocket: socket getOptions: optionName
	| s returnedValue errorCode results |
	<var: #s type: #SocketPtr>
	self primitive: 'primitiveSocketGetOptions' parameters: #(Oop Oop).

	s := self socketValueOf: socket.
	interpreterProxy success: (interpreterProxy isBytes: optionName).
	interpreterProxy failed ifTrue: [^nil].
	returnedValue := 0.
	errorCode := self sqSocketGetOptions: s 
					optionNameStart: (self cCoerce: (interpreterProxy firstIndexableField: optionName) to: #'char *') 
					optionNameSize: (interpreterProxy byteSizeOf: optionName)
					returnedValue: (self addressOf: returnedValue put: [:val| returnedValue := val]).

	results := interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 2.
	interpreterProxy
		storeInteger: 0 ofObject: results withValue: errorCode;
		storeInteger: 1 ofObject: results withValue: returnedValue.
	^results! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 2/25/2016 13:32'!
primitiveSocket: socket listenOnPort: port 
	"one part of the wierdass dual prim primitiveSocketListenOnPort which 
	was warped by some demented evil person determined to twist the very 
	nature of reality"
	| s  okToListen |
	<var: #s type: 'SocketPtr '>
	self primitive: 'primitiveSocketListenOnPort' parameters: #(#Oop #SmallInteger ).
	s := self socketValueOf: socket.
	"If the security plugin can be loaded, use it to check for permission.
	If  not, assume it's ok"
	sCCLOPfn ~= 0 ifTrue:
		[okToListen := self cCode: ' ((sqInt (*) (sqInt, sqInt)) sCCLOPfn)((sqInt)s, port)'.
		 okToListen ifFalse:
			[^ interpreterProxy primitiveFail]].
	interpreterProxy failed ifFalse:
		[self sqSocket: s ListenOnPort: port]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 12:09'!
primitiveSocket: socket listenOnPort: port backlogSize: backlog 
	"second part of the wierdass dual prim primitiveSocketListenOnPort 
	which was warped by some demented evil person determined to twist the 
	very nature of reality"
	| s okToListen |
	<var: #s type: 'SocketPtr'>
	self primitive: 'primitiveSocketListenOnPortBacklog' parameters: #(#Oop #SmallInteger #SmallInteger ).
	"If the security plugin can be loaded, use it to check for permission.
	If not, assume it's ok"
	sCCLOPfn ~= 0 ifTrue:
		[okToListen := self cCode: ' ((sqInt (*) (sqInt, sqInt)) sCCLOPfn)((sqInt)s, port)'.
		 okToListen ifFalse:
			[^interpreterProxy primitiveFail]].
	s := self socketValueOf: socket.
	interpreterProxy failed ifFalse:
		[self sqSocket: s ListenOnPort: port BacklogSize: backlog]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 12:09'!
primitiveSocket: socket listenOnPort: port backlogSize: backlog interface: ifAddr
	"Bind a socket to the given port and interface address with no more than backlog pending connections.  The socket can be UDP, in which case the backlog should be specified as zero."

	| s okToListen addr |
	<var: #s type: #SocketPtr>
	self primitive: 'primitiveSocketListenOnPortBacklogInterface' parameters: #(#Oop #SmallInteger #SmallInteger #ByteArray).
	"If the security plugin can be loaded, use it to check for permission.
	If  not, assume it's ok"
	sCCLOPfn ~= 0 ifTrue:
		[okToListen := self cCode: ' ((sqInt (*) (sqInt, sqInt)) sCCLOPfn)((sqInt)s, port)'.
		 okToListen ifFalse:
			[^ interpreterProxy primitiveFail]].
	s := self socketValueOf: socket.
	addr := self netAddressToInt: (self cCoerce: ifAddr to: #'unsigned char *').
	interpreterProxy failed ifFalse:
		[self sqSocket: s ListenOnPort: port BacklogSize: backlog Interface: addr]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp (auto pragmas dtl 2010-09-28) 6/9/2007 09:46'!
primitiveSocket: socket listenWithBacklog: backlogSize

	| s  |
	<var: #s type: 'SocketPtr'>
	self primitive: 'primitiveSocketListenWithBacklog' parameters: #(#Oop #SmallInteger).
	s := self socketValueOf: socket.
	interpreterProxy failed ifFalse: [self sqSocket: s ListenBacklog: backlogSize]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp (auto pragmas dtl 2010-09-28) 6/9/2007 10:26'!
primitiveSocket: socket localAddressResult: socketAddress

	| addrSize addrBase s  |
	<var: #addrBase type: 'char *'>
	<var: #s type: 'SocketPtr'>
	self primitive: 'primitiveSocketLocalAddressResult' parameters: #(#Oop #Oop).
	s := self socketValueOf: socket.
	addrSize := interpreterProxy byteSizeOf: socketAddress.
	addrBase := self cCoerce: (interpreterProxy firstIndexableField: socketAddress) to: 'char *'.
	interpreterProxy failed ifFalse: [self sqSocket: s LocalAddressResult: addrBase Size: addrSize]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 12:08'!
primitiveSocket: socket receiveDataBuf: array start: startIndex count: count 
	| s elementSize arrayBase bytesReceived |
	<var: #s type: #SocketPtr>
	self primitive: 'primitiveSocketReceiveDataBufCount'
		parameters: #(Oop Oop SmallInteger SmallInteger).

	"buffer can be any indexable bits object"
	elementSize := self cppIf: SPURVM
						ifTrue: [interpreterProxy bytesPerElement: array]
						ifFalse: [(interpreterProxy isWords: array) ifTrue: [4] ifFalse: [1]].
	((interpreterProxy isWordsOrBytes: array)
	 and: [elementSize > 0
	 and: [startIndex >= 1
	 and: [count >= 0
	 and: [startIndex + count - 1 <= (interpreterProxy slotSizeOf: array)]]]]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	s := self socketValueOf: socket.
	interpreterProxy failed ifFalse: "Note: adjust bufStart for zero-origin indexing"
		[arrayBase := self cCoerce: (interpreterProxy firstIndexableField: array) to: #'char *'.
		bytesReceived := self
					sqSocket: s
					ReceiveDataBuf: arrayBase + (startIndex - 1 * elementSize)
					Count: count * elementSize.
		^(bytesReceived // elementSize) asSmallIntegerObj]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 12:01'!
primitiveSocket: socket receiveUDPDataBuf: array start: startIndex count: count 
	| s elementSize arrayBase bytesReceived results address port moreFlag addressOop |
	<var: #s type: #SocketPtr>
	<var: #arrayBase type: #'char *'>
	self primitive: 'primitiveSocketReceiveUDPDataBufCount'
		parameters: #(Oop Oop SmallInteger SmallInteger).
	s := self socketValueOf: socket.

	"buffer can be any indexable bits object"
	elementSize := self cppIf: SPURVM
						ifTrue: [interpreterProxy bytesPerElement: array]
						ifFalse: [(interpreterProxy isWords: array) ifTrue: [4] ifFalse: [1]].
	((interpreterProxy isWordsOrBytes: array)
	 and: [elementSize > 0
	 and: [startIndex >= 1
	 and: [count >= 0
	 and: [startIndex + count - 1 <= (interpreterProxy slotSizeOf: array)]]]]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	"Note: adjust for zero-origin indexing"
	arrayBase := self cCoerce: (interpreterProxy firstIndexableField: array) to: #'char *'.
	address	 := port := moreFlag := 0.
	bytesReceived := self sqSocket: s
							ReceiveUDPDataBuf: arrayBase + (startIndex - 1 * elementSize)
							Count: count * elementSize
							address: (self addressOf: address put: [:v| address := v])
							port: (self addressOf: port put: [:v| port := v])
							moreFlag: (self addressOf: moreFlag put: [:v| moreFlag := v]).

	"allocate storage for results, remapping newly allocated
	 oops in case GC happens during allocation"
	addressOop := self intToNetAddress: address.
	self remapOop: addressOop
		in: [results := interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 4].
	interpreterProxy
		storeInteger: 0 ofObject: results withValue: bytesReceived // elementSize;
		storePointer: 1 ofObject: results withValue: addressOop;
		storeInteger: 2 ofObject: results withValue: port;
		storePointer: 3 ofObject: results withValue: (moreFlag
														ifTrue: [interpreterProxy trueObject]
														ifFalse: [interpreterProxy falseObject]).
	^results! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'ikp (auto pragmas dtl 2010-09-28) 6/9/2007 10:26'!
primitiveSocket: socket remoteAddressResult: socketAddress

	| addrSize addrBase s  |
	<var: #addrBase type: 'char *'>
	<var: #s type: 'SocketPtr'>
	self primitive: 'primitiveSocketRemoteAddressResult' parameters: #(#Oop #Oop).
	s := self socketValueOf: socket.
	addrSize := interpreterProxy byteSizeOf: socketAddress.
	addrBase := self cCoerce: (interpreterProxy firstIndexableField: socketAddress) to: 'char *'.
	interpreterProxy failed ifFalse: [self sqSocket: s RemoteAddressResult: addrBase Size: addrSize]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 13:46'!
primitiveSocket: socket sendData: array start: startIndex count: count 
	| s arrayBase bytesSent elementSize |
	<var: #s type: #SocketPtr>
	<var: #arrayBase type: #'char *'>
	self primitive: 'primitiveSocketSendDataBufCount'
		parameters: #(Oop Oop SmallInteger SmallInteger ).
	s := self socketValueOf: socket.

	"buffer can be any indexable bits object"
	elementSize := self cppIf: SPURVM
						ifTrue: [interpreterProxy bytesPerElement: array]
						ifFalse: [(interpreterProxy isWords: array) ifTrue: [4] ifFalse: [1]].
	((interpreterProxy isWordsOrBytes: array)
	 and: [elementSize > 0
	 and: [startIndex >= 1
	 and: [count >= 0
	 and: [startIndex + count - 1 <= (interpreterProxy slotSizeOf: array)]]]]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	interpreterProxy failed
		ifFalse: ["Note: adjust bufStart for zero-origin indexing"
			arrayBase := self cCoerce: (interpreterProxy firstIndexableField: array) to: #'char *'.
			bytesSent := self
						sqSocket: s
						SendDataBuf: arrayBase + (startIndex - 1 * elementSize)
						Count: count * elementSize].
	^(bytesSent // elementSize) asSmallIntegerObj! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 14:25'!
primitiveSocket: socket sendUDPData: array toHost: hostAddress  port: portNumber start: startIndex count: count 
	| s elementSize arrayBase bytesSent hostAddressInteger |
	<var: #s type: #SocketPtr>
	self primitive: 'primitiveSocketSendUDPDataBufCount'
		parameters: #(Oop Oop ByteArray SmallInteger SmallInteger SmallInteger).

	"buffer can be any indexable bits object"
	elementSize := self cppIf: SPURVM
						ifTrue: [interpreterProxy bytesPerElement: array]
						ifFalse: [(interpreterProxy isWords: array) ifTrue: [4] ifFalse: [1]].
	((interpreterProxy isWordsOrBytes: array)
	 and: [elementSize > 0
	 and: [startIndex >= 1
	 and: [count >= 0
	 and: [startIndex + count - 1 <= (interpreterProxy slotSizeOf: array)]]]]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	s := self socketValueOf: socket.
	hostAddressInteger := self netAddressToInt: (self cCoerce: hostAddress to: #'unsigned char *').
	interpreterProxy failed ifFalse:
		["Note: adjust bufStart for zero-origin indexing"
		arrayBase := self cCoerce: (interpreterProxy firstIndexableField: array) to: #'char *'.
		bytesSent := self
						sqSocket: s
						toHost: hostAddressInteger
						port: portNumber
						SendDataBuf: arrayBase + (startIndex - 1 * elementSize)
						Count: count * elementSize].
	^(bytesSent // elementSize) asSmallIntegerObj! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 12:14'!
primitiveSocket: socket setOptions: optionName value: optionValue
	"THIS BADLY NEEDS TO BE REWRITTEN TO TAKE Booleans AND Integers AS WELL AS (OR INSTEAD OF) Strings.
	 It is only used with booleans and integers and parsing these back out of strings in
	 sqSocketSetOptions:optionNameStart:optionNameSize:optionValueStart:optionValueSize:returnedValue:
	 is STUPID."
	| s returnedValue errorCode results |
	<var: #s type: #SocketPtr>
	self primitive: 'primitiveSocketSetOptions' parameters: #(Oop Oop Oop).

	s := self socketValueOf: socket.
	interpreterProxy success: ((interpreterProxy isBytes: optionName) and: [interpreterProxy isBytes: optionValue]).
	interpreterProxy failed ifTrue: [^nil].
	returnedValue := 0.

	errorCode := self sqSocketSetOptions: s 
					optionNameStart: (self cCoerce: (interpreterProxy firstIndexableField: optionName) to: #'char *') 
					optionNameSize: (interpreterProxy byteSizeOf: optionName)
					optionValueStart: (self cCoerce: (interpreterProxy firstIndexableField: optionValue) to: #'char *')
					optionValueSize: (interpreterProxy byteSizeOf: optionValue)
					returnedValue: (self addressOf: returnedValue put: [:val| returnedValue := val]).

	results := interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 2.
	interpreterProxy
		storeInteger: 0 ofObject: results withValue: errorCode;
		storeInteger: 1 ofObject: results withValue: returnedValue.
	^results! !

!SocketPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:17'!
primitiveSocketAbortConnection: socket

	| s |
	<var: #s type: 'SocketPtr '>
	self primitive: 'primitiveSocketAbortConnection'
		parameters: #(Oop).
	s := self socketValueOf: socket.
	interpreterProxy failed ifFalse: [
		self sqSocketAbortConnection: s]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 12:25'!
primitiveSocketAcceptFrom: sockHandle rcvBufferSize: recvBufSize sndBufSize: sendBufSize semaIndex: semaIndex 
	| socketOop s serverSocket |
	<var: #s type: 'SocketPtr '>
	<var: #serverSocket type: 'SocketPtr '>
	self primitive: 'primitiveSocketAccept'
		parameters: #(Oop SmallInteger SmallInteger SmallInteger ).
	serverSocket := self socketValueOf: sockHandle.

	interpreterProxy failed ifFalse:
		[socketOop := interpreterProxy
						instantiateClass: interpreterProxy classByteArray
						indexableSize: self socketRecordSize.
		s := self socketValueOf: socketOop.
		interpreterProxy failed ifFalse:
			[self
				sqSocket: s
				AcceptFrom: serverSocket
				RecvBytes: recvBufSize
				SendBytes: sendBufSize
				SemaID: semaIndex.
			^ socketOop]]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 12:25'!
primitiveSocketAcceptFrom: sockHandle rcvBufferSize: recvBufSize sndBufSize: sendBufSize semaIndex: semaIndex readSemaIndex: aReadSema writeSemaIndex: aWriteSema
	| socketOop s serverSocket |
	<var: #s type: 'SocketPtr '>
	<var: #serverSocket type: 'SocketPtr '>
	self primitive: 'primitiveSocketAccept3Semaphores'
		parameters: #(Oop SmallInteger SmallInteger SmallInteger SmallInteger SmallInteger).
	serverSocket := self socketValueOf: sockHandle.

	interpreterProxy failed ifFalse:
		[socketOop := interpreterProxy instantiateClass: interpreterProxy classByteArray indexableSize: self socketRecordSize.
		s := self socketValueOf: socketOop.
		interpreterProxy failed ifFalse:
			[self
				sqSocket: s
				AcceptFrom: serverSocket
				RecvBytes: recvBufSize
				SendBytes: sendBufSize
				SemaID: semaIndex
				ReadSemaID: aReadSema
				WriteSemaID: aWriteSema.
			^ socketOop]]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'dtl (auto pragmas dtl 2010-09-28) 9/28/2010 20:15'!
primitiveSocketAddressGetPort

	| addr addrSize addrBase port |
	<var: #addrBase type: 'char *'>
	addr := self primitive: 'primitiveSocketAddressGetPort' parameters: #() receiver: #Oop.
	addrSize := interpreterProxy byteSizeOf: addr.
	addrBase := self cCoerce: (interpreterProxy firstIndexableField: addr) to: 'char *'.
	interpreterProxy failed
		ifFalse:
			[port := self sqSocketAddress: addrBase SizeGetPort: addrSize.
			 interpreterProxy failed ifFalse: [^port asSmallIntegerObj]]! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'dtl (auto pragmas dtl 2010-09-28) 9/28/2010 20:16'!
primitiveSocketAddressSetPort: portNumber

	| addr addrSize addrBase |
	<var: #addrBase type: 'char *'>
	addr := self primitive: 'primitiveSocketAddressSetPort' parameters: #(SmallInteger) receiver: #Oop.
	addrSize := interpreterProxy byteSizeOf: addr.
	addrBase := self cCoerce: (interpreterProxy firstIndexableField: addr) to: 'char *'.
	interpreterProxy failed ifFalse: [self sqSocketAddress: addrBase Size: addrSize SetPort: portNumber]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:17'!
primitiveSocketCloseConnection: socket

	| s |
	<var: #s type: 'SocketPtr '>
	self primitive: 'primitiveSocketCloseConnection'
		parameters: #(Oop).
	s := self socketValueOf: socket.
	interpreterProxy failed ifFalse: [
		self sqSocketCloseConnection: s]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:17'!
primitiveSocketConnectionStatus: socket

	| s status |
	<var: #s type: 'SocketPtr '>
	self primitive: 'primitiveSocketConnectionStatus'
		parameters: #(Oop).
	s := self socketValueOf: socket.
	interpreterProxy failed ifFalse: [
		status := self sqSocketConnectionStatus: s].
	^ status asSmallIntegerObj! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 12:17'!
primitiveSocketCreateNetwork: netType type: socketType receiveBufferSize: recvBufSize sendBufSize: sendBufSize semaIndex: semaIndex 
	| socketOop s okToCreate |
	<var: #s type: 'SocketPtr '>
	self primitive: 'primitiveSocketCreate' parameters: #(#SmallInteger #SmallInteger #SmallInteger #SmallInteger #SmallInteger ).
	"If the security plugin can be loaded, use it to check for permission.
	If  not, assume it's ok"
	sCCSOTfn ~= 0 ifTrue:
		[okToCreate := self cCode: ' ((sqInt (*) (sqInt, sqInt)) sCCSOTfn)(netType, socketType)'.
		 okToCreate ifFalse:
			[^ interpreterProxy primitiveFail]].
	socketOop := interpreterProxy instantiateClass: interpreterProxy classByteArray indexableSize: self socketRecordSize.
	s := self socketValueOf: socketOop.
	interpreterProxy failed ifFalse:
		[self
			sqSocket: s
			CreateNetType: netType
			SocketType: socketType
			RecvBytes: recvBufSize
			SendBytes: sendBufSize
			SemaID: semaIndex.
		^socketOop]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 12:17'!
primitiveSocketCreateNetwork: netType type: socketType receiveBufferSize: recvBufSize sendBufSize: sendBufSize semaIndex: semaIndex readSemaIndex: aReadSema writeSemaIndex: aWriteSema 
	| socketOop s okToCreate |
	<var: #s type: 'SocketPtr '>
	self primitive: 'primitiveSocketCreate3Semaphores' parameters: #(#SmallInteger #SmallInteger #SmallInteger #SmallInteger #SmallInteger #SmallInteger #SmallInteger ).
	"If the security plugin can be loaded, use it to check for permission.
	If not, assume it's ok"
	sCCSOTfn ~= 0 ifTrue:
		[okToCreate := self cCode: ' ((sqInt (*) (sqInt, sqInt)) sCCSOTfn)(netType, socketType)'.
		 okToCreate ifFalse:
			[^ interpreterProxy primitiveFail]].
	socketOop := interpreterProxy instantiateClass: interpreterProxy classByteArray indexableSize: self socketRecordSize.
	s := self socketValueOf: socketOop.
	interpreterProxy failed ifFalse:
		[self
			sqSocket: s
			CreateNetType: netType
			SocketType: socketType
			RecvBytes: recvBufSize
			SendBytes: sendBufSize
			SemaID: semaIndex
			ReadSemaID: aReadSema
			WriteSemaID: aWriteSema.
		^socketOop]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 12:18'!
primitiveSocketCreateRaw: netType type: protoType receiveBufferSize: recvBufSize sendBufSize: sendBufSize semaIndex: semaIndex readSemaIndex: aReadSema writeSemaIndex: aWriteSema 
	| socketOop s okToCreate |
	<var: #s type: 'SocketPtr '>
	self primitive: 'primitiveSocketCreateRAW' parameters: #(#SmallInteger #SmallInteger #SmallInteger #SmallInteger #SmallInteger #SmallInteger #SmallInteger ).
	"If the security plugin can be loaded, use it to check for permission.
	If not, assume it's ok"
	sCCSOTfn ~= 0 ifTrue:
		[okToCreate := self cCode: ' ((sqInt (*) (sqInt, sqInt)) sCCSOTfn)(netType, protoType)'.
		 okToCreate ifFalse:
			[^ interpreterProxy primitiveFail]].
	socketOop := interpreterProxy instantiateClass: interpreterProxy classByteArray indexableSize: self socketRecordSize.
	s := self socketValueOf: socketOop.
	interpreterProxy failed ifFalse:
		[self
			sqSocket: s
			CreateRaw: netType
			ProtoType: protoType
			RecvBytes: recvBufSize
			SendBytes: sendBufSize
			SemaID: semaIndex
			ReadSemaID: aReadSema
			WriteSemaID: aWriteSema.
		^ socketOop]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:18'!
primitiveSocketDestroy: socket

	| s |
	<var: #s type: 'SocketPtr'>
	self primitive: 'primitiveSocketDestroy'
		parameters: #(Oop).
	s := self socketValueOf: socket.
	interpreterProxy failed ifFalse: [
		self sqSocketDestroy: s]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 12:18'!
primitiveSocketError: socket

	| s |
	<var: #s type: #SocketPtr>
	self primitive: 'primitiveSocketError'
		parameters: #(Oop).
	s := self socketValueOf: socket.
	interpreterProxy failed ifFalse:
		[^(self sqSocketError: s) asSmallIntegerObj]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'ar (auto pragmas 12/08) 5/24/2000 13:38'!
primitiveSocketListenWithOrWithoutBacklog
	"Backward compatibility"
	<export: true>
	interpreterProxy methodArgumentCount = 2
		ifTrue:[^self primitiveSocketListenOnPort]
		ifFalse:[^self primitiveSocketListenOnPortBacklog]
! !

!SocketPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:18'!
primitiveSocketLocalAddress: socket

	| s addr |
	<var: #s type: 'SocketPtr'>
	self primitive: 'primitiveSocketLocalAddress'
		parameters: #(Oop).
	s := self socketValueOf: socket.
	addr := self sqSocketLocalAddress: s.
	^self intToNetAddress: addr! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'eem 9/25/2020 12:19'!
primitiveSocketLocalAddressSize: socket

	| s size |
	<var: #s type: #SocketPtr>
	self primitive: 'primitiveSocketLocalAddressSize' parameters: #(#Oop).
	s := self socketValueOf: socket.
	interpreterProxy failed ifFalse:
		[size := self sqSocketLocalAddressSize: s.
		 interpreterProxy failed ifFalse:
			[^size asSmallIntegerObj]]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:18'!
primitiveSocketLocalPort: socket

	| s port |
	<var: #s type: 'SocketPtr '>
	self primitive: 'primitiveSocketLocalPort'
		parameters: #(Oop).
	s := self socketValueOf: socket.
	port := self sqSocketLocalPort: s.
	^port asSmallIntegerObj! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 13:39'!
primitiveSocketReceiveDataAvailable: socket

	| s |
	<var: #s type: #SocketPtr>
	self primitive: 'primitiveSocketReceiveDataAvailable'
		parameters: #(Oop).
	s := self socketValueOf: socket.
	interpreterProxy failed ifFalse:
		[^(self sqSocketReceiveDataAvailable: s) asBooleanObj]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:18'!
primitiveSocketRemoteAddress: socket

	| s addr |
	<var: #s type: 'SocketPtr'>
	self primitive: 'primitiveSocketRemoteAddress'
		parameters: #(Oop).
	s := self socketValueOf: socket.
	addr := self sqSocketRemoteAddress: s.
	^self intToNetAddress: addr! !

!SocketPlugin methodsFor: 'ipv6 primitives' stamp: 'eem 9/25/2020 12:21'!
primitiveSocketRemoteAddressSize: socket

	| s size |
	<var: #s type: #SocketPtr>
	self primitive: 'primitiveSocketRemoteAddressSize' parameters: #(#Oop).
	s := self socketValueOf: socket.
	interpreterProxy failed ifFalse:
		[size := self sqSocketRemoteAddressSize: s.
		interpreterProxy failed ifFalse:
			[^size asSmallIntegerObj]]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 13:38'!
primitiveSocketRemotePort: socket

	| s |
	<var: #s type: #SocketPtr>
	self primitive: 'primitiveSocketRemotePort'
		parameters: #(Oop).
	s := self socketValueOf: socket.
	interpreterProxy failed ifFalse:
		[^(self sqSocketRemotePort: s) asSmallIntegerObj]! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 9/25/2020 13:38'!
primitiveSocketSendDone: socket

	| s |
	<var: #s type: #SocketPtr>
	self primitive: 'primitiveSocketSendDone'
		parameters: #(Oop).
	s := self socketValueOf: socket.
	interpreterProxy failed ifFalse:
		[^(self sqSocketSendDone: s) asBooleanObj]! !

!SocketPlugin methodsFor: 'initialize-release' stamp: 'eem 11/10/2020 11:18'!
shutdownModule
	<export: true>
	^self socketShutdown! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 8/30/2020 17:38'!
socketRecordSize
	"Answer the size of a Smalltalk socket record in bytes."
	<inline: #always>
	^self sizeof: #SQSocket! !

!SocketPlugin methodsFor: 'primitives' stamp: 'eem 8/30/2020 17:37'!
socketValueOf: socketOop 
	"Answer a pointer to the first byte of of the socket record within the  
	 given Smalltalk object, or nil if socketOop is not a socket record."
	<returnTypeC: #SocketPtr>
	<inline: #always>
	^((interpreterProxy isBytes: socketOop)
	   and: [(interpreterProxy byteSizeOf: socketOop) = self socketRecordSize])
		ifTrue: [self cCoerce: (interpreterProxy firstIndexableField: socketOop) to: #SocketPtr]
		ifFalse: [interpreterProxy primitiveFailFor: PrimErrBadArgument. nil]! !

SocketPlugin subclass: #SocketPluginSimulator
	instanceVariableNames: 'openSocketHandles externalSemaphores hostSocketToSimSocketMap simSocketToHostSocketMap fakeAddressCounter resolverSemaphoreIndex ipv6support resolverStatus addressForName'
	classVariableNames: 'ResolverBusy ResolverError ResolverReady ResolverUninitialized'
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!SocketPluginSimulator class methodsFor: 'class initialization' stamp: 'eem 1/15/2018 13:30'!
initialize
	ResolverUninitialized := 0.	"network is not initialized"
	ResolverReady := 1.		"resolver idle, last request succeeded"
	ResolverBusy := 2.			"lookup in progress"
	ResolverError := 3.			"resolver idle, last request failed"! !

!SocketPluginSimulator class methodsFor: 'translation' stamp: 'eem 12/18/2018 21:01'!
shouldBeTranslated
	^false! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 10/13/2020 09:09'!
cCoerce: value to: cTypeString
	"Type coercion for translation only; just return the value when running in Smalltalk.
	 This overrides the generic coercion method in VMClass.  For some reason we are the exception.
	 If we want that style of coercion we can send cCoerce:to: to interpreterProxy, not self."

	^value
		coerceTo: (cTypeString = #SocketPtr ifTrue: [#'char *'] ifFalse: [cTypeString])
		sim: interpreterProxy! !

!SocketPluginSimulator methodsFor: 'initialize-release' stamp: 'eem 12/29/2016 15:54'!
close  "close any sockets that ST may have opened"
	openSocketHandles do: [:h | self closeAndDestroy: h].
	Smalltalk unregisterExternalObjects: externalSemaphores! !

!SocketPluginSimulator methodsFor: 'initialize-release' stamp: 'eem 9/18/2019 21:07'!
closeAndDestroy: socketHandle
	"c.f. Socket closeAndDestroy: timeoutSeconds"
	| fakeSocket |
	fakeSocket := Socket basicNew.
	[(fakeSocket primSocketConnectionStatus: socketHandle) = (Socket classPool at: #Connected) ifTrue:
		[fakeSocket primSocketCloseConnection: socketHandle].
	 fakeSocket
		primSocketAbortConnection: socketHandle;
		primSocketDestroy: socketHandle]
		on: SocketPrimitiveFailed
		do: [:ex| Transcript cr; show: ex messageText]! !

!SocketPluginSimulator methodsFor: 'simulation support' stamp: 'eem 12/30/2016 15:05'!
hostSocketHandleFromSimSocketHandle: socketHandleCArray
	"Answer the corresponding host socketHandle for the simulation socketHandle, or nil if none, failing the primitive."
	^simSocketToHostSocketMap
		at: (self simSocketHandleFrom: socketHandleCArray)
		ifAbsent: [interpreterProxy primitiveFail. nil]! !

!SocketPluginSimulator methodsFor: 'accessing' stamp: 'eem 12/30/2016 14:29'!
ipv6support
	
	^ ipv6support
! !

!SocketPluginSimulator methodsFor: 'accessing' stamp: 'eem 12/30/2016 14:29'!
ipv6support: anObject
	
	ipv6support := anObject.
! !

!SocketPluginSimulator methodsFor: 'simulation support' stamp: 'eem 10/1/2019 19:53'!
map: hostSocketHandle to: simSockPtr type: socketType register: semaphores spawning: blockOne and: blockTwo and: blockThree
	| simSocket ourPriority |
	"SQSocket is typedef struct { int sessionID; int socketType; void *privateSocketPtr; } SQSocket"
	simSocket := ByteArray new: (self sizeof: #SQSocket).
	simSocket
		unsignedLongAt: 1 put: interpreterProxy getThisSessionID;
		unsignedLongAt: 5 put: socketType.
	simSocket size = 12
		ifTrue: [simSocket unsignedLongAt: 9 put: (fakeAddressCounter := fakeAddressCounter + 64)]
		ifFalse: [simSocket unsignedLongLongAt: 9 put: (fakeAddressCounter := fakeAddressCounter + 80)].
	self assert: ((interpreterProxy isBytes: simSockPtr cPtrAsOop)
				and: [(interpreterProxy numBytesOf: simSockPtr cPtrAsOop) = simSocket size]).
	1 to: simSocket size do:
		[:i| simSockPtr at: i - 1 put: (simSocket at: i)].
	self assert: (self simSocketHandleFrom: simSockPtr) = simSocket.
	openSocketHandles add: hostSocketHandle.
	hostSocketToSimSocketMap at: hostSocketHandle put: simSocket.
	simSocketToHostSocketMap at: simSocket put: hostSocketHandle.
	externalSemaphores addAll: semaphores.
	"N.B. These don't need registering.  Eventually they will end up
	 waiting on semaphores that have been unregistered, and hence
	 will get garbage collected, along with these processes. But a) run them at one higher than the current priority, and
	 b) ensure that they all effectively resume at the same time; i.e. ensure this code doesn't get preempted by the first
	 process to start running."
	ourPriority := Processor activePriority.
	[({blockOne. blockTwo. blockThree}
				collect: [:b| b newProcess priority: ourPriority + 1; yourself])
			do: [:each| each resume]]
		valueAt: ourPriority + 2! !

!SocketPluginSimulator methodsFor: 'simulation support' stamp: 'eem 12/30/2016 14:54'!
netAddressAsByteArrayFromInt: netAddress
	^ByteArray
		with: ((netAddress bitShift: -24) bitAnd: 16rFF)
		with: ((netAddress bitShift: -16) bitAnd: 16rFF)
		with: ((netAddress bitShift: -8) bitAnd: 16rFF)
		with:  (netAddress bitAnd: 16rFF)! !

!SocketPluginSimulator methodsFor: 'primitives' stamp: 'eem 9/21/2019 09:47'!
primitiveSocketListenWithOrWithoutBacklog
	"This is a hack to get around limitations in the simulated dispatch of smart syntax primitives."
	interpreterProxy methodArgumentCount = 2 ifTrue:
		["self primitive: 'primitiveSocketListenOnPort' parameters: #(#Oop #SmallInteger)"
		 ^self primitiveSocket: (interpreterProxy stackValue: 1)
				listenOnPort: (interpreterProxy stackIntegerValue: 0)].
	"self primitive: 'primitiveSocketListenOnPortBacklog' parameters: #(#Oop #SmallInteger #SmallInteger)"
	^self primitiveSocket: (interpreterProxy stackValue: 2)
		listenOnPort: (interpreterProxy stackIntegerValue: 1)
		backlogSize: (interpreterProxy stackIntegerValue: 0)! !

!SocketPluginSimulator methodsFor: 'simulation support' stamp: 'eem 12/30/2016 14:48'!
simSocketHandleFrom: socketHandleCArray
	| simSocket |
	"SQSocket is typedef struct { int sessionID; int socketType; void *privateSocketPtr; } SQSocket"
	simSocket := ByteArray new: (self sizeof: #SQSocket).
	1 to: simSocket size do:
		[:i|
		simSocket at: i put: (socketHandleCArray at: i - 1)].
	^simSocket! !

!SocketPluginSimulator methodsFor: 'accessing' stamp: 'eem 12/30/2016 18:13'!
simulator: aSmartSyntaxPluginSimulator
	super simulator: aSmartSyntaxPluginSimulator.
	"aSmartSyntaxPluginSimulator logging: true"! !

!SocketPluginSimulator methodsFor: 'initialize-release' stamp: 'eem 1/15/2018 13:31'!
socketInit
	openSocketHandles := Set new.
	externalSemaphores := Set new.
	hostSocketToSimSocketMap := Dictionary new.
	simSocketToHostSocketMap := Dictionary new.
	fakeAddressCounter := 16r50C4E70. "Socket, if you squint at it right..."
	"Set all the security functions to zero so simulation does't need to work fully."
	sDSAfn := sHSAfn := sCCTPfn := sCCLOPfn := sCCSOTfn := 0.
	"for now..."
	ipv6support := false.
	resolverStatus := ResolverUninitialized.
	^true! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 11/10/2020 11:18'!
socketShutdown
	self close.
	^true! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 1/15/2018 13:56'!
sqNetworkInit: resolverSemaIndex
	"Simply assume the network is initialized."
	(NetNameResolver classPool at: #HaveNetwork) ifFalse:
		[NetNameResolver initializeNetwork].
	resolverSemaphoreIndex
		ifNil: [resolverSemaphoreIndex := resolverSemaIndex]
		ifNotNil: [self assert: resolverSemaphoreIndex = resolverSemaIndex].
	resolverStatus := ResolverReady.
	^0! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/30/2016 14:28'!
sqResolverHostNameSize
	ipv6support ifTrue: [^NetNameResolver primHostNameSize].
	interpreterProxy primitiveFail! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'cb 3/13/2017 16:18'!
sqResolverLocalAddress
	"For now don't simulate the implicit semaphore."
	| bytes |
	bytes := NetNameResolver primLocalAddress.
	self assert: bytes size = 4.
	"Effectively netAddressToInt: bytes"
	^	((bytes at: 4)) +
		((bytes at: 3) <<8) +
		((bytes at: 2) <<16) +
		((bytes at: 1) <<24)! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 10/30/2020 09:38'!
sqResolverNameLookupResult
	"For now don't simulate the implicit semaphore."
	addressForName ifNil: [^interpreterProxy primitiveFail].
	self assert: (addressForName size >= 4 and: [addressForName size \\ 4 = 0]).
	"Effectively netAddressToInt: bytes"
	^	((addressForName at: 4)) +
		((addressForName at: 3) <<8) +
		((addressForName at: 2) <<16) +
		((addressForName at: 1) <<24)! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 9/19/2018 18:02'!
sqResolverStartName: aCArray Lookup: size
	"For now don't simulate the implicit semaphore."
	| hostName |
	hostName := self strncpy: (String new: size) _: aCArray _: size.
	addressForName := NetNameResolver addressForName: hostName timeout: 30.
	resolverSemaphoreIndex ifNotNil:
		[resolverStatus := ResolverReady.
		 interpreterProxy signalSemaphoreWithIndex: resolverSemaphoreIndex]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 1/16/2018 11:13'!
sqResolverStatus
	^resolverStatus ! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 9/18/2019 15:40'!
sqSocket: sockPtr AcceptFrom: serverSocket RecvBytes: recvBufSize SendBytes: sendBufSize SemaID: semaIndex ReadSemaID: readSemaIndex WriteSemaID: writeSemaIndex
	"Simulate the sqSocketAcceptFromRecvBytesSendBytesSemaIDReadSemaIDWriteSemaID function.
	 We descend beneath the Socket abstraction to simulate as accurately as possible."
	| semaphoresAndIndexes semaphores indexes socketHandle |
	semaphoresAndIndexes := Smalltalk newExternalSemaphores: 3.
	semaphores := semaphoresAndIndexes first.
	indexes := semaphoresAndIndexes second.
	socketHandle := [Socket basicNew
						primAcceptFrom: serverSocket
						receiveBufferSize: recvBufSize
						sendBufSize: sendBufSize
						semaIndex: indexes first
						readSemaIndex: indexes second
						writeSemaIndex: indexes third]
			on: SocketPrimitiveFailed
			do: [:ex|
				#failed].
	socketHandle == #failed ifTrue:
		[interpreterProxy primitiveFail.
		 Smalltalk unregisterExternalObjects: semaphores.
		 ^self].
	"N.B. There is now a Processor yield in doSignalExternalSemaphores: every 100 virtual microseconds.
	 This allows these to make progress.  Their job is to map a host signal into a signal of the relevant index."
	self map:		socketHandle
		to:			sockPtr
		type:		nil
		register:	semaphores
		spawning:	[[semaphores first wait. interpreterProxy  signalSemaphoreWithIndex: semaIndex] repeat]
		and:		[[semaphores second wait. interpreterProxy  signalSemaphoreWithIndex: readSemaIndex] repeat]
		and:		[[semaphores third wait. interpreterProxy  signalSemaphoreWithIndex: writeSemaIndex] repeat]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/30/2016 17:27'!
sqSocket: socketHandle ConnectTo: addr Port: port
	^[Socket basicNew
			primSocket: ((self hostSocketHandleFromSimSocketHandle: socketHandle) ifNil: [^self])
			connectTo: (self netAddressAsByteArrayFromInt: addr)
			port: port]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			0]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/29/2016 16:08'!
sqSocket: sockPtr CreateNetType: netType SocketType: socketType RecvBytes: recvBufSize SendBytes: sendBufSize SemaID: semaIndex ReadSemaID: readSemaIndex WriteSemaID: writeSemaIndex
	"Simulate the sqSocketCreateNetTypeSocketTypeRecvBytesSendBytesSemaIDReadSemaIDWriteSemaID function.
	 We descend beneath the Socket abstraftion to simulate as accurately as possible."
	| semaphoresAndIndexes semaphores indexes socketHandle |
	semaphoresAndIndexes := Smalltalk newExternalSemaphores: 3.
	semaphores := semaphoresAndIndexes first.
	indexes := semaphoresAndIndexes second.
	socketHandle := [Socket basicNew
						primSocketCreateNetwork: netType
						type: socketType
						receiveBufferSize: recvBufSize
						sendBufSize: sendBufSize
						semaIndex: indexes first
						readSemaIndex: indexes second
						writeSemaIndex: indexes third]
			on: SocketPrimitiveFailed
			do: [:ex|
				#failed].
	socketHandle == #failed ifTrue:
		[interpreterProxy primitiveFail.
		 Smalltalk unregisterExternalObjects: semaphores.
		 ^self].
	"N.B. There is now a Processor yield in doSignalExternalSemaphores: every 100 virtual microseconds.
	 This allows these to make progress.  Their job is to map a host signal into a signal of the relevant index."
	self map:		socketHandle
		to:			sockPtr
		type:		socketType
		register:	semaphores
		spawning:	[[semaphores first wait. interpreterProxy  signalSemaphoreWithIndex: semaIndex] repeat]
		and:		[[semaphores second wait. interpreterProxy  signalSemaphoreWithIndex: readSemaIndex] repeat]
		and:		[[semaphores third wait. interpreterProxy  signalSemaphoreWithIndex: writeSemaIndex] repeat]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 9/18/2019 20:44'!
sqSocket: socketHandle ListenOnPort: portNumber
	^[Socket basicNew
			primSocket: ((self hostSocketHandleFromSimSocketHandle: socketHandle) ifNil: [^self])
			listenOn: portNumber]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			0]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/22/2020 16:25'!
sqSocket: socketHandle ListenOnPort: portNumber BacklogSize: backlog
	^[Socket basicNew
			primSocket: ((self hostSocketHandleFromSimSocketHandle: socketHandle) ifNil: [^self])
			listenOn: portNumber
			backlogSize: backlog]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			0]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 9/18/2019 15:31'!
sqSocket: socketHandle ListenOnPort: portNumber BacklogSize: backlog Interface: addr
	^[Socket basicNew
			primSocket: ((self hostSocketHandleFromSimSocketHandle: socketHandle) ifNil: [^self])
			listenOn: portNumber
			backlogSize: backlog
			interface: (self netAddressAsByteArrayFromInt: addr)]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			0]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/30/2016 17:31'!
sqSocket: socketHandleCArray ReceiveDataBuf: bufferStartCArray Count: numBytes
	^[| buffer n |
	  buffer := ByteArray new: numBytes.
	  n := Socket basicNew
			primSocket: ((self hostSocketHandleFromSimSocketHandle: socketHandleCArray) ifNil: [^0])
			receiveDataInto: buffer
			startingAt: 1
			count: numBytes.
	  1 to: n do:
		[:i|
		bufferStartCArray at: i - 1 put: (buffer at: i)].
	  n]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			0]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/30/2016 17:31'!
sqSocket: socketHandleCArray SendDataBuf: bufferStartCArray Count: numBytes
	| data |
	data := ByteArray new: numBytes.
	1 to: numBytes do:
		[:i| data at: i put: (bufferStartCArray at: i - 1)].
	^[Socket basicNew
			primSocket: ((self hostSocketHandleFromSimSocketHandle: socketHandleCArray) ifNil: [^0])
			sendData: data
			startIndex: 1
			count: numBytes]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			0]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 1/15/2018 14:33'!
sqSocketAbortConnection: socketHandleCArray
	^[Socket basicNew
			primSocketAbortConnection: ((self hostSocketHandleFromSimSocketHandle: socketHandleCArray) ifNil: [^false])]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			false]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 1/15/2018 14:31'!
sqSocketCloseConnection: socketHandleCArray
	^[Socket basicNew
			primSocketCloseConnection: ((self hostSocketHandleFromSimSocketHandle: socketHandleCArray) ifNil: [^false])]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			false]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/30/2016 17:34'!
sqSocketConnectionStatus: socketHandleCArray
	^[Socket basicNew
			primSocketConnectionStatus: ((self hostSocketHandleFromSimSocketHandle: socketHandleCArray) ifNil: [^-1])]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			-1]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/30/2016 16:24'!
sqSocketDestroy: socketHandleCArray 
	| simHandle hostHandle |
	simHandle := self simSocketHandleFrom: socketHandleCArray.
	hostHandle := simSocketToHostSocketMap removeKey: simHandle ifAbsent: [].
	hostHandle ifNil:
		[interpreterProxy primitiveFail.
		 ^self].
	hostSocketToSimSocketMap removeKey: hostHandle ifAbsent: [].
	[Socket basicNew primSocketDestroy: hostHandle]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 9/18/2019 17:55'!
sqSocketError: socketHandleCArray 
	^[Socket basicNew
			primSocketError: ((self hostSocketHandleFromSimSocketHandle: socketHandleCArray) ifNil: [^false])]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			false]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 9/18/2019 17:58'!
sqSocketGetOptions: socketHandle
	optionNameStart: optionNameStart optionNameSize: optionNameSize
		returnedValue: valuePtrBlock
	"Simulate sqSocketGetOptionsoptionNameStartoptionNameSizereturnedValue.
	 Answer the error code and assign the value through valuePtrBlock."
	| optionName |
	optionName := self asString: optionNameStart size: optionNameSize.
	^[| errAndValue |
		errAndValue := Socket basicNew
						primSocket: ((self hostSocketHandleFromSimSocketHandle: socketHandle) ifNil: [^self])
						getOption: optionName.
		valuePtrBlock at: 0 put: errAndValue second.
		errAndValue first]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			0]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 9/18/2019 20:46'!
sqSocketLocalPort: socketHandleCArray 
	^[Socket basicNew
			primSocketLocalPort: ((self hostSocketHandleFromSimSocketHandle: socketHandleCArray) ifNil: [^false])]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			false]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/30/2016 17:33'!
sqSocketReceiveDataAvailable: socketHandleCArray 
	^[Socket basicNew
			primSocketReceiveDataAvailable: ((self hostSocketHandleFromSimSocketHandle: socketHandleCArray) ifNil: [^false])]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			false]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 9/18/2019 15:58'!
sqSocketRemoteAddress: socketHandleCArray 
	^[| remoteAddress |
	   remoteAddress := Socket basicNew
							primSocketRemoteAddress: ((self hostSocketHandleFromSimSocketHandle: socketHandleCArray) ifNil: [^false]).
		"Effectively netAddressToInt: remoteAddress"
		((remoteAddress at: 4)) +
		((remoteAddress at: 3) <<8) +
		((remoteAddress at: 2) <<16) +
		((remoteAddress at: 1) <<24)]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			false]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/30/2016 17:34'!
sqSocketSendDone: socketHandleCArray
	^[Socket basicNew
			primSocketSendDone: ((self hostSocketHandleFromSimSocketHandle: socketHandleCArray) ifNil: [^false])]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			false]! !

!SocketPluginSimulator methodsFor: 'simulation' stamp: 'eem 9/18/2019 17:53'!
sqSocketSetOptions: socketHandle
	optionNameStart: optionNameStart optionNameSize: optionNameSize optionValueStart: optionValueStart optionValueSize: optionValueSize
		returnedValue: valuePtrBlock
	"Simulate sqSocketSetOptionsoptionNameStartoptionNameSizeoptionValueStartoptionValueSizereturnedValue.
	 Answer the error code and assign the value through valuePtrBlock."
	| optionName optionValue |
	optionName := self asString: optionNameStart size: optionNameSize.
	optionValue := self asString: optionValueStart size: optionValueSize.
	^[| errAndValue |
		errAndValue := Socket basicNew
						primSocket: ((self hostSocketHandleFromSimSocketHandle: socketHandle) ifNil: [^self])
						setOption: optionName
						value: optionValue.
		valuePtrBlock at: 0 put: errAndValue second.
		errAndValue first]
		on: SocketPrimitiveFailed
		do: [:ex|
			interpreterProxy primitiveFail.
			0]! !

SmartSyntaxInterpreterPlugin subclass: #SoundPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!SoundPlugin commentStamp: 'tpr 5/2/2003 15:50' prior: 0!
This plugin implements the main sound related primiives.  Since it requires platform support it will only be built when supported on your platform


FORMAT OF SOUND DATA

Squeak uses 16-bit signed samples encoded in the host's endian order.  A sound buffer is a sequence of "frames", or "slices", where each frame usually includes one sample per channel.  The exception is that for playback, each frame always includes 2 samples; for monaural playback, every other sample is ignored.
!

!SoundPlugin class methodsFor: 'translation' stamp: 'tpr 5/23/2001 17:12'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!SoundPlugin class methodsFor: 'translation' stamp: 'tpr 11/29/2000 22:38'!
requiresPlatformFiles
	"this plugin requires platform specific files in order to work"
	^true! !

!SoundPlugin methodsFor: 'support' stamp: 'eem 9/18/2020 18:15'!
frameCountOf: buf
	"Answer the number of 16-bit sound sample pairs in buf, a pointer to the first indexable field of a sound buffer."
	<inline: #always>
	^(interpreterProxy byteSizeOf: buf cPtrAsOop) / ((self sizeof: #short) * 2)! !

!SoundPlugin methodsFor: 'initialize-release' stamp: 'eem 9/1/2020 14:52'!
initialiseModule
	<export: true>
	^self soundInit! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 8/14/2020 13:58'!
primitiveGetDefaultSoundPlayer
	"Answer a String with the operating system name of the default output device, or nil"
	"no arguments"
	<export: true>
	self methodReturnStringOrNil: self getDefaultSoundPlayer! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 8/14/2020 13:59'!
primitiveGetDefaultSoundRecorder
	"Answer a String with the operating system name of the default input device, or nil"
	"no arguments"
	<export: true>
	self methodReturnStringOrNil: self getDefaultSoundRecorder! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 7/30/2020 19:10'!
primitiveGetNumberOfSoundPlayerDevices
	<export: true>

	^interpreterProxy methodReturnInteger: self getNumberOfSoundPlayerDevices! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 7/30/2020 19:11'!
primitiveGetNumberOfSoundRecorderDevices
	<export: true>

	^interpreterProxy methodReturnInteger: self getNumberOfSoundRecorderDevices! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 8/14/2020 13:59'!
primitiveGetSoundPlayerDeviceName
	"arguments: name(type, stack offset)
		deviceNumber(Integer, 0)"
	"answers a string or nil"
	<export: true>
	| deviceNumber |

	"Parse arguments" 
	interpreterProxy methodArgumentCount = 1 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	deviceNumber := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	
	self methodReturnStringOrNil: (self getSoundPlayerDeviceName: deviceNumber - 1)! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 2/3/2021 19:54'!
primitiveGetSoundPlayerDeviceUID
	"arguments: name(type, stack offset)
		deviceNumber(Integer, 0)"
	"answers a string or nil"
	<export: true>
	<option: #TerfVM>
	| deviceNumber |

	"Parse arguments" 
	interpreterProxy methodArgumentCount = 1 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	deviceNumber := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	
	self methodReturnStringOrNil: (self getSoundPlayerDeviceUID: deviceNumber - 1)! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 8/14/2020 14:00'!
primitiveGetSoundRecorderDeviceName
	"arguments: name(type, stack offset)
		deviceNumber(Integer, 0)"
	"answers a string or nil"
	<export: true>
	| deviceNumber |

	"Parse arguments" 
	interpreterProxy methodArgumentCount = 1 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	deviceNumber := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	
	self methodReturnStringOrNil: (self getSoundRecorderDeviceName: deviceNumber - 1)! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 2/3/2021 19:54'!
primitiveGetSoundRecorderDeviceUID
	"arguments: name(type, stack offset)
		deviceNumber(Integer, 0)"
	"answers a string or nil"
	<export: true>
	<option: #TerfVM>
	| deviceNumber |

	"Parse arguments" 
	interpreterProxy methodArgumentCount = 1 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	deviceNumber := interpreterProxy positive32BitValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	
	self methodReturnStringOrNil: (self getSoundRecorderDeviceUID: deviceNumber - 1)! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 2/3/2021 19:00'!
primitiveSetDefaultSoundPlayer
	"Tell the operating system to use the specified device name as the output device for sound."
	"arg at top of stack is the String"
	<export: true>
	| deviceName |
	"Parse arguments"
	interpreterProxy methodArgumentCount = 1 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	deviceName := interpreterProxy stackEphemeralStringValue: 0.
	interpreterProxy failed ifFalse:
		[self setDefaultSoundPlayer: deviceName.
		 interpreterProxy failed ifFalse:
			[interpreterProxy methodReturnReceiver]]! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 2/3/2021 18:59'!
primitiveSetDefaultSoundRecorder
	"Tell the operating system to use the specified device name as the input device for sound."
	"arg at top of stack is the String"
	<export: true>
	| deviceName |
	"Parse arguments"
	interpreterProxy methodArgumentCount = 1 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	deviceName := interpreterProxy stackEphemeralStringValue: 0.
	interpreterProxy failed ifFalse:
		[self setDefaultSoundRecorder: deviceName.
		 interpreterProxy failed ifFalse:
			[interpreterProxy methodReturnReceiver]]! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 7/31/2020 19:44'!
primitiveSoundAvailableSpace
	"Returns the number of bytes of available sound output buffer space.
	 This should be (frames*4) if the device is in stereo mode, or (frames*2) otherwise"

	<export: true>
	| frames |
	frames := self snd_AvailableSpace.  "-1 if sound output not started"
	frames >= 0
		ifTrue: [interpreterProxy methodReturnInteger: frames]
		ifFalse: [interpreterProxy primitiveFail]! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 9/24/2020 13:27'!
primitiveSoundEnableAEC
	"Enable or disable acoustic echo-cancellation (AEC).
	 Arg is a boolean or 1 for true and 0 for false."
	<export: true>
	| arg trueOrFalse errorCode |
	interpreterProxy methodArgumentCount = 1 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	"Parse arguments"
	(interpreterProxy isIntegerObject: (arg := interpreterProxy stackValue: 0))
		ifTrue:
			[arg := interpreterProxy integerValueOf: arg.
			 (interpreterProxy cCoerce: arg to: #unsigned) > 1 ifTrue:
				[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
			 trueOrFalse := arg = 1]
		ifFalse:
			[(interpreterProxy isBooleanObject: arg) ifFalse:
				[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
			trueOrFalse := interpreterProxy booleanValueOf: arg].
	"Set AEC"
	(errorCode := self snd_EnableAEC: trueOrFalse) ~= 0 ifTrue:
		[interpreterProxy primitiveFailFor: (errorCode < 0 ifTrue: [PrimErrGenericFailure] ifFalse: [errorCode])].
	interpreterProxy methodReturnReceiver! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 7/31/2020 19:44'!
primitiveSoundGetRecordLevel
	"Get the default input device's volume level in the range 0-1000."
	<export: true>
	| level |
	<var: 'level' type: #int>
	level := self snd_GetRecordLevel.
	^level >= 0
		ifTrue: [interpreterProxy methodReturnInteger: level]
		ifFalse: [interpreterProxy primitiveFail]! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 7/31/2020 12:15'!
primitiveSoundGetRecordingSampleRate
	"Return a float representing the actual sampling rate during recording. Fail if not currently recording."

	<export: true>
	| rate |
	<var: #rate type: #double>
	rate := self snd_GetRecordingSampleRate.  "fails if not recording"
	interpreterProxy failed ifFalse:
		[^interpreterProxy methodReturnFloat: rate]! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 7/31/2020 18:18'!
primitiveSoundGetVolume
	"Get the default output device's volume level as a left/right pair of floats in the range 0-1."
	<export: true>
	| left right leftOop rightOop results |
	<var: #left type: #double>
	<var: #right type: #double>
	left := 0.0.
	right := 0.0.
	self snd_Volume: (self addressOf: left put: [:v| left := v]) _: (self addressOf: right put: [:v| right := v]).
	interpreterProxy failed ifTrue:
		[^self].
	results := interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: 2.
	results ifNil:
		[^interpreterProxy primitiveFailFor: PrimErrNoMemory].
	self remapOop: results in:
		[leftOop := interpreterProxy floatObjectOf: left.
		 self remapOop: leftOop in:
			[rightOop := interpreterProxy floatObjectOf: right]].
	interpreterProxy
		storePointer: 0 ofObject: results withValue: leftOop;
		storePointer: 1 ofObject: results withValue: rightOop;
		methodReturnValue: results! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 9/18/2020 18:16'!
primitiveSoundInsertSamples: frameCount from: buf leadTime: leadTime 
	"Insert a buffer's worth of sound samples into the currently playing  
	buffer. Used to make a sound start playing as quickly as possible. The  
	new sound is mixed with the previously buffered sampled."
	"Details: Unlike primitiveSoundPlaySamples, this primitive always starts  
	with the first sample the given sample buffer. Its third argument  
	specifies the number of samples past the estimated sound output buffer  
	position the inserted sound should start. If successful, it returns the  
	number of samples inserted."
	
	"N.B. At least on Windows and MacOS Cocoa circa late 2020 this is unsupported.
	 snd_InsertSamplesFromLeadTime merely answers zero."
	| framesPlayed |
	self primitive: 'primitiveSoundInsertSamples'
		parameters: #(SmallInteger WordsOrShorts SmallInteger).
	(self cCoerce: frameCount to: #usqInt) > (self frameCountOf: buf) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	framesPlayed := self snd_InsertSamplesFromLeadTime: frameCount _: buf _: leadTime.
	framesPlayed >= 0
		ifTrue: [interpreterProxy methodReturnInteger: framesPlayed]
		ifFalse: [interpreterProxy primitiveFail]! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 9/18/2020 18:16'!
primitiveSoundPlaySamples: frameCount from: buf startingAt: startIndex 
	"Output a buffer's worth of stereo sound samples.  The frameCount is the number of 16-bit sample pairs.
	 See e.g. platforms/iOS/plugins/SoundPlugin/sqSqueakSoundCoreAudio.m:
		#define SqueakFrameSize	4	// guaranteed (see class SoundPlayer)
	 Of course there is no FrameSize in Squeak 5.x/6.x; sigh..."
	| framesPlayed |
	self primitive: 'primitiveSoundPlaySamples'
		parameters: #(SmallInteger WordsOrShorts SmallInteger).
	(startIndex >= 1
	 and: [startIndex + frameCount - 1 <= (self frameCountOf: buf)]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].

	framesPlayed := self snd_PlaySamplesFromAtLength: frameCount _: buf _: startIndex - 1.
	framesPlayed >= 0
		ifTrue: [interpreterProxy methodReturnInteger: framesPlayed]
		ifFalse: [interpreterProxy primitiveFail]! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 7/31/2020 19:45'!
primitiveSoundPlaySilence
	"Output a buffer's worth of silence. Returns the number of sample frames played."

	<export: true>
	| framesPlayed |
	framesPlayed := self snd_PlaySilence.  "-1 if sound output not started"
	framesPlayed >= 0
		ifTrue: [interpreterProxy methodReturnInteger: framesPlayed]
		ifFalse: [interpreterProxy primitiveFail]! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 9/18/2020 18:25'!
primitiveSoundRecordSamplesInto: buf startingAt: startWordIndex 
	"Record a buffer's worth of (mono?) 16-bit sound samples."
	| bufSizeInBytes samplesRecorded byteOffset |
	self primitive: 'primitiveSoundRecordSamples'
		parameters: #(WordsOrShorts SmallInteger).

	bufSizeInBytes := interpreterProxy byteSizeOf: buf cPtrAsOop.
	byteOffset := (startWordIndex - 1) * 2.

	(startWordIndex >= 1 and: [byteOffset < bufSizeInBytes]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadIndex].

	samplesRecorded := self snd_RecordSamplesIntoAtLength: (self cCoerce: buf to: #'char *') + byteOffset
							_: 0
							_: bufSizeInBytes - byteOffset.
	interpreterProxy failed ifFalse:
		[^samplesRecorded asPositiveIntegerObj]! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 7/31/2020 12:59'!
primitiveSoundSetLeftVolume: aLeftVolume rightVolume: aRightVolume
	"Set the sound input recording level."

	self primitive: 'primitiveSoundSetLeftVolume'
		parameters: #(Float Float).
	self snd_SetVolume: aLeftVolume _: aRightVolume! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 7/31/2020 13:16'!
primitiveSoundSetRecordLevel: level 
	"Set the sound input recording level."
	self primitive: 'primitiveSoundSetRecordLevel'
		parameters: #(SmallInteger).
	self snd_SetRecordLevel: level! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 7/31/2020 12:56'!
primitiveSoundStartBufferSize: bufFrames rate: samplesPerSec stereo: stereoFlag
	"Start the double-buffered sound output with the given buffer size, sample rate, and stereo flag."

	self primitive: 'primitiveSoundStart'
		parameters: #(SmallInteger SmallInteger Boolean).
	interpreterProxy success: (self snd_Start: bufFrames _: samplesPerSec _: stereoFlag _: 0)! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 7/31/2020 12:56'!
primitiveSoundStartBufferSize: bufFrames rate: samplesPerSec stereo: stereoFlag semaIndex: semaIndex
	"Start the double-buffered sound output with the given buffer size, sample rate, stereo flag, and semaphore index."

	self primitive: 'primitiveSoundStartWithSemaphore'
		parameters: #(SmallInteger SmallInteger Boolean SmallInteger).
	interpreterProxy success: (self snd_Start: bufFrames _: samplesPerSec _: stereoFlag _: semaIndex)! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 7/31/2020 12:54'!
primitiveSoundStartRecordingDesiredSampleRate: desiredSamplesPerSec stereo: stereoFlag semaIndex: semaIndex
	"Start recording sound with the given parameters."

	self primitive: 'primitiveSoundStartRecording'
		parameters: #(SmallInteger Boolean SmallInteger).
	self snd_StartRecording: desiredSamplesPerSec _: stereoFlag _: semaIndex! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 7/31/2020 13:16'!
primitiveSoundStop
	"Stop double-buffered sound output."
	<export: true>
	self snd_Stop! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 7/31/2020 13:16'!
primitiveSoundStopRecording
	"Stop recording sound."
	<export: true>
	self snd_StopRecording! !

!SoundPlugin methodsFor: 'primitives' stamp: 'eem 9/24/2020 13:31'!
primitiveSoundSupportsAEC
	"Answer if the OS/hardware supports echo-cancellation."
	<export: true>
	interpreterProxy methodReturnBool: self snd_SupportsAEC ~= 0! !

!SoundPlugin methodsFor: 'initialize-release' stamp: 'eem 9/1/2020 14:51'!
shutdownModule
	<export: true>
	^self soundShutdown! !

!SoundPlugin methodsFor: 'simulation' stamp: 'eem 9/1/2020 14:51'!
snd_Stop
	<doNotGenerate>! !

!SoundPlugin methodsFor: 'simulation' stamp: 'eem 9/1/2020 14:52'!
soundInit
	<doNotGenerate>
	^true! !

!SoundPlugin methodsFor: 'simulation' stamp: 'eem 9/1/2020 14:52'!
soundShutdown
	<doNotGenerate>
	^true! !

SmartSyntaxInterpreterPlugin subclass: #TestOSAPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!TestOSAPlugin commentStamp: 'tpr 5/2/2003 15:51' prior: 0!
I am the Plugin for the Squeak/Applescript Interface. Since it requires platform support it will only be built when supported on your platform!

!TestOSAPlugin class methodsFor: 'translation' stamp: 'EstebanLorenzano 6/3/2014 10:37'!
declareCVarsIn: cg
	cg addHeaderFile: '<Carbon/Carbon.h>'.
! !

!TestOSAPlugin class methodsFor: 'translation' stamp: 'JMM 5/30/2001 19:43'!
requiresPlatformFiles
	"this plugin requires platform specific files in order to work"
	^true! !

!TestOSAPlugin class methodsFor: 'translation' stamp: 'eem 12/18/2018 11:40'!
shouldBeTranslated
	"As yet this can't be translated because a DescType parameter can't be handled."
	^false! !

!TestOSAPlugin class methodsFor: 'simulation' stamp: 'eem 10/26/2018 09:34'!
simulatorClass
	"As yet this can't be simulated because a DescType parameter can't be handled."
	^nil! !

!TestOSAPlugin methodsFor: 'AppleEvents prims' stamp: 'acg 9/23/1999 22:49'!
primAECoerceDesc: typeCode to: result

	|rcvr |
	rcvr := self 	primitive: 	'primAECoerceDesc'
				parameters:	#(DescType AEDesc)
				receiver:	#AEDesc.
	^(self 
		cCode: 'AECoerceDesc(rcvr,*typeCode,result)'
		inSmalltalk: [[rcvr]. -1]) asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'AppleEvents prims' stamp: 'acg 9/20/1999 12:37'!
primAECreateDesc: typeCode from: aString

	|rcvr size |
	rcvr := self 	primitive: 	'primAECreateDesc'
				parameters:	#(DescType String)
				receiver:	#AEDesc.
	size := aString size.
	^(self 
		cCode: 'AECreateDesc(*typeCode, aString, size, rcvr)'
		inSmalltalk: [[rcvr. size]. -1]) asOop: Unsigned
! !

!TestOSAPlugin methodsFor: 'AppleEvents prims' stamp: 'acg 9/20/1999 14:16'!
primAEDescToString: aString

	| rcvr size |
	rcvr := self	primitive: 	'primAEDescToString'
				parameters: #(String)
				receiver: #AEDesc.
	size := aString size.
	self cCode: 'BlockMove(*(rcvr->dataHandle), aString, size)'
		 inSmalltalk: [rcvr. size].
	^aString asOop: String

! !

!TestOSAPlugin methodsFor: 'AppleEvents prims' stamp: 'acg 9/20/1999 12:38'!
primAEDisposeDesc
	
	|rcvr|
	rcvr :=	self	primitive: 	'primAEDisposeDesc'
				parameters:	#()
				receiver: 	#AEDesc.
	^(self 
		cCode: 'AEDisposeDesc(rcvr)'
		inSmalltalk: [[rcvr]. -1]) asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'AppleEvents prims' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:22'!
primAEGetKeyPtr: key type: type actual: ignoreDesc to: bytes

	| rcvr size ignoreSize |
	<var: #ignoreSize type: 'Size '>
	rcvr := self	primitive: 	'primAEGetKeyPtr'
				parameters: #(DescType DescType DescType ByteArray)
				receiver: #AEDesc.
	size := ignoreSize := bytes size.
	^(self cCode: 'AEGetKeyPtr(rcvr, *key, *type, ignoreDesc, bytes, size, &ignoreSize)'
		 inSmalltalk: [[rcvr. size. ignoreSize]. -1]) asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'Gen''l Mac OS prims' stamp: 'acg 9/20/1999 12:56'!
primGetHandleSize: anIndex

	|rcvr|
	rcvr := self	primitive: 'primGetHandleSize'
				parameters: #(SmallInteger)
				receiver:	#WordArray.
	^(self
		cCode: 'GetHandleSize((Handle) *(rcvr+anIndex))'
		inSmalltalk: [[rcvr]. -1]) asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'OSA prims' stamp: 'acg 9/20/1999 16:08'!
primOSACompile: source mode: mode to: object

	|component|
	component := self primitive: 	'primOSACompile'
					parameters: #(AEDesc SmallInteger OSAID)
					receiver:	#ComponentInstance.

	^(self cCode: 'OSACompile(*component,source,mode,object)'
			inSmalltalk: [[component]. -1]) asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'OSA prims' stamp: 'acg 9/20/1999 16:08'!
primOSADisplay: source as: type mode: mode to: result

	|component|
	component := self primitive: 	'primOSADisplay'
					parameters: #(OSAID DescType SmallInteger AEDesc)
					receiver:	#ComponentInstance.

	^(self cCode: 'OSADisplay(*component,*source,*type,mode,result)'
			inSmalltalk: [[component]. -1]) asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'OSA prims' stamp: 'acg 9/20/1999 15:39'!
primOSADispose: anOSAID

	|component|
	component := self primitive: 	'primOSADispose'
					parameters: #(OSAID)
					receiver:	#ComponentInstance.

	^(self cCode: 'OSADispose(*component,*anOSAID)'
			inSmalltalk: [[component]. -1]) asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'OSA prims' stamp: 'JMM 10/31/2005 11:59'!
primOSADoScript: source in: context mode: mode resultType: type to: result

	|component resultsOfCall giLocker |
	component := self primitive: 	'primOSADoScript'
					parameters: #(AEDesc OSAID SmallInteger DescType AEDesc)
					receiver:	#ComponentInstance.

	self cCode: '
        giLocker = interpreterProxy->ioLoadFunctionFrom("getUIToLock", "");
        if (giLocker !!= 0) {
            long *foo;
            foo = malloc(sizeof(long)*9);
            foo[0] = 6;
            foo[1] = OSADoScript;
            foo[2] = *component;
            foo[3] = source;
            foo[4] = *context;
            foo[5] = *type;
            foo[6] = mode;
            foo[7] = result;
            foo[8] = 0;
            ((int (*) (void *)) giLocker)(foo);
            resultsOfCall = interpreterProxy->positive32BitIntegerFor(foo[8]);
            free(foo);}'
			inSmalltalk: [[component. giLocker].  resultsOfCall := -1].
	^resultsOfCall asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'OSA prims' stamp: 'JMM 10/31/2005 11:59'!
primOSAExecute: script in: context mode: mode to: result

	|component giLocker resultsOfCall |
	component := self primitive: 	'primOSAExecute'
					parameters: #(OSAID OSAID SmallInteger OSAID)
					receiver:	#ComponentInstance.

	self cCode: '
         giLocker = interpreterProxy->ioLoadFunctionFrom("getUIToLock", "");
         if (giLocker !!= 0) {
            long *foo;
            foo = malloc(sizeof(long)*8);
            foo[0] = 5;
            foo[1] = OSAExecute;
            foo[2] = *component;
            foo[3] = *script;
            foo[4] = *context;
            foo[5] = mode;
            foo[6] = result;
            foo[7] = 0;
            ((int (*) (void *)) giLocker)(foo);
            resultsOfCall = interpreterProxy->positive32BitIntegerFor(foo[7]);
            free(foo); }
'
			inSmalltalk: [[component. giLocker]. resultsOfCall := -1].
	^resultsOfCall asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'OSA prims' stamp: 'acg 9/25/1999 15:01'!
primOSAGetScriptInfo: aScriptID type: aDescType to: resultData

	|component|
	component := self	primitive: 	'primOSAGetScriptInfo'
						parameters: #(OSAID DescType IntegerArray)
						receiver:	#ComponentInstance.
	
	^(self cCode: 'OSAGetScriptInfo(*component,*aScriptID,*aDescType, (long *)resultData)'
			inSmalltalk: [[component]. -1]) asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'OSA prims' stamp: 'acg 9/25/1999 17:25'!
primOSAGetSource: aScriptID type: aDescType to: resultData

	|component|
	component := self	primitive: 	'primOSAGetSource'
						parameters: #(OSAID DescType AEDesc)
						receiver:	#ComponentInstance.
	
	^(self cCode: 'OSAGetSource(*component,*aScriptID,*aDescType, resultData)'
			inSmalltalk: [[component]. -1]) asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'OSA prims' stamp: 'acg 9/22/1999 03:08'!
primOSALoad: source mode: mode to: result

	|component|
	component := self primitive: 	'primOSALoad'
					parameters: #(AEDesc SmallInteger OSAID)
					receiver:	#ComponentInstance.

	^(self cCode: 'OSALoad(*component,source,mode,result)'
			inSmalltalk: [[component]. -1]) asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'OSA prims' stamp: 'acg 9/25/1999 22:55'!
primOSAMakeContext: name parent: parent to: result

	|component|
	component := self primitive: 	#primOSAMakeContext
					parameters: #(AEDesc OSAID OSAID)
					receiver:	#ComponentInstance.

	^(self cCode: 'OSAMakeContext(*component,name,*parent,result)'
			inSmalltalk: [[component]. -1]) asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'OSA prims' stamp: 'acg 9/23/1999 20:39'!
primOSAScriptError: selector type: type to: result

	|component|
	component := self primitive: 	'primOSAScriptError'
					parameters: #(DescType DescType AEDesc)
					receiver:	#ComponentInstance.

	^(self cCode: 'OSAScriptError(*component,*selector,*type,result)'
			inSmalltalk: [[component]. -1]) asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'OSA prims' stamp: 'acg 9/20/1999 21:53'!
primOSAScriptingComponentNameTo: anAEDesc

	|component|
	component := self	primitive: 	'primOSAScriptingComponentName'
						parameters: #(AEDesc)
						receiver:	#ComponentInstance.
	
	^(self cCode: 'OSAScriptingComponentName(*component,anAEDesc)'
			inSmalltalk: [[component]. -1]) asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'OSA prims' stamp: 'acg 9/22/1999 07:51'!
primOSAStore: source resultType: type mode: mode to: result

	|component|
	component := self primitive: 	#primOSAStore
					parameters: #(OSAID DescType SmallInteger AEDesc)
					receiver:	#ComponentInstance.

	^(self cCode: 'OSAStore(*component,*source,*type,mode,result)'
			inSmalltalk: [[component]. -1]) asOop: Unsigned! !

!TestOSAPlugin methodsFor: 'Component Mgr prims' stamp: 'acg 9/20/1999 23:42'!
primOpenDefaultConfiguration: type subtype: subtype

	| component |
	component := self	primitive: 	'primOpenDefaultConfiguration'
						parameters: #(DescType DescType)
						receiver:	#ComponentInstance.
	self	cCode: '*component = OpenDefaultComponent(*type,*subtype)'
		inSmalltalk: [component at: 0 put: 0].
	^component asOop: ComponentInstance! !

SmartSyntaxInterpreterPlugin subclass: #UUIDPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!UUIDPlugin commentStamp: 'tpr 5/2/2003 15:52' prior: 0!
A class to generate UUID
by John M McIntosh johnmci@smalltalkconsulting.com, since it requires platform support it will only be built when supported on your platform.
See http://www.webdav.org/specs/draft-leach-uuids-guids-01.txt

If a plugin does not exist then we generate a UUID version 4 type GUUID in Smalltalk!

!UUIDPlugin class methodsFor: 'translation' stamp: 'JMM 10/9/2001 12:46'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!UUIDPlugin class methodsFor: 'translation' stamp: 'JMM 10/9/2001 13:37'!
requiresPlatformFiles
	"this plugin requires platform specific files in order to work"
	^true! !

!UUIDPlugin methodsFor: 'initialize' stamp: 'JMM (auto pragmas 12/08) 10/9/2001 12:47'!
initialiseModule
	<export: true>
	^self cCode: 'sqUUIDInit()' inSmalltalk:[true]! !

!UUIDPlugin methodsFor: 'system primitives' stamp: 'eem 6/28/2016 12:59'!
primitiveMakeUUID
	| oop location |
	<export: true>
	<var: #location type: #'char*'>
	oop := interpreterProxy stackValue: 0.
	(interpreterProxy methodArgumentCount = 0
	 and: [(interpreterProxy isBytes: oop)
	 and: [(interpreterProxy byteSizeOf: oop) = 16]]) ifFalse:
		[^interpreterProxy primitiveFail].
	location := interpreterProxy firstIndexableField: oop.
	self cCode: [self MakeUUID: location]
		inSmalltalk:
			[| uuid |
			uuid := UUID new.
			1 to: 16 do:
				[:i| location at: i - 1 put: (uuid at: i)]].
	^oop! !

!UUIDPlugin methodsFor: 'initialize' stamp: 'ar (auto pragmas 12/08) 2/3/2002 20:03'!
shutdownModule
	<export: true>
	^self cCode: 'sqUUIDShutdown()' inSmalltalk:[true]! !

InterpreterPlugin subclass: #SmartSyntaxPluginSimulator
	instanceVariableNames: 'actualPlugin signatureMap forMap pluginClass logging'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SmartSyntaxPlugins'!

!SmartSyntaxPluginSimulator class methodsFor: 'simulation' stamp: 'eem 10/3/2010 08:28'!
shouldBeTranslated
	^false! !

!SmartSyntaxPluginSimulator methodsFor: 'accessing' stamp: 'eem 10/1/2010 09:22'!
actualPlugin
	"Answer the value of actualPlugin"

	^ actualPlugin! !

!SmartSyntaxPluginSimulator methodsFor: 'accessing' stamp: 'eem 12/28/2016 21:47'!
actualPlugin: aSmartSyntaxInterpreterPlugin signatureClass: signatureClass
	pluginClass := signatureClass.
	actualPlugin := aSmartSyntaxInterpreterPlugin.
	actualPlugin simulator: self! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 10/2/2010 22:00'!
ccgLoad: forProlog expr: failBlock asBooleanValueFrom: anInteger 
	^[:oop| interpreterProxy booleanValueOf: oop]! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/28/2016 21:16'!
ccgLoad: codeGen expr: exprBlock asCharPtrFrom: stackIndex andThen: validateBlock
	^[:oop|
	   validateBlock value: oop.
	   interpreterProxy cCoerce: (interpreterProxy firstIndexableField: oop) asInteger to: #'char *']! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 10/26/2018 09:31'!
ccgLoad: codeGen expr: exprBlock asFloatValueFrom: stackIndex
	^[:oop| interpreterProxy floatValueOf: oop]! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/28/2016 21:17'!
ccgLoad: codeGen expr: exprBlock asIntPtrFrom: stackIndex andThen: validateBlock
	^[:oop|
	   validateBlock value: oop.
	   interpreterProxy cCoerce: (interpreterProxy firstIndexableField: oop) asInteger to: #'int *']! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 10/2/2010 21:48'!
ccgLoad: forProlog expr: failBlock asIntegerValueFrom: anInteger 
	^[:oop| interpreterProxy checkedIntegerValueOf: oop]! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 10/2/2010 13:29'!
ccgLoad: forProlog expr: failBlock asKindOf: aClass from: argIndexOrNil 
	^[:oop|
	   interpreterProxy success: (interpreterProxy is: oop KindOf: aClass name asString).
	   oop]! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'nice 4/18/2016 23:43'!
ccgLoad: forProlog expr: failBlock asKindOfIntegerFrom: argIndexOrNil 
	^[:oop|
	   interpreterProxy success: (interpreterProxy isKindOfInteger: oop).
	   oop]! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'nice 4/19/2016 21:51'!
ccgLoad: forProlog expr: failBlock asMemberOfLargeNegativeIntegerFrom: argIndexOrNil 
	^[:oop|
	   interpreterProxy success: (interpreterProxy isLargeNegativeIntegerObject: oop).
	   oop]! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'nice 4/19/2016 21:51'!
ccgLoad: forProlog expr: failBlock asMemberOfLargePositiveIntegerFrom: argIndexOrNil 
	^[:oop|
	   interpreterProxy success: (interpreterProxy isLargePositiveIntegerObject: oop).
	   oop]! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/28/2016 21:14'!
ccgLoad: codeGen expr: exprBlock asOopPtrFrom: stackIndex andThen: validateBlock
	^[:oop|
	   validateBlock value: oop.
	   interpreterProxy cCoerce: (interpreterProxy firstIndexableField: oop) asInteger to: #'sqInt *']! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 10/2/2010 21:46'!
ccgLoad: forProlog expr: failBlock asRawOopFrom: anUndefinedObject 
	^[:oop| oop]! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/1/2020 18:08'!
ccgLoad: codeGen expr: exprBlock asShortPtrFrom: stackIndex andThen: validateBlock
	^[:oop|
	   validateBlock value: oop.
	   interpreterProxy cCoerce: (interpreterProxy firstIndexableField: oop) asInteger to: #'short *']! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'nice 5/2/2020 11:11'!
ccgLoad: codeGen expr: exprBlock asUnsignedLong64PtrFrom: stackIndex andThen: validateBlock
	^[:oop|
	   validateBlock value: oop.
	   interpreterProxy cCoerce: (interpreterProxy firstIndexableField: oop) asInteger to: #'unsigned long long *']! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/28/2016 21:19'!
ccgLoad: codeGen expr: exprBlock asUnsignedPtrFrom: stackIndex andThen: validateBlock
	^[:oop|
	   validateBlock value: oop.
	   interpreterProxy cCoerce: (interpreterProxy firstIndexableField: oop) asInteger to: #'unsigned *']! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'nice 5/1/2020 21:04'!
ccgLoad: codeGen expr: exprBlock asUnsignedShortPtrFrom: stackIndex andThen: validateBlock
	^[:oop|
	   validateBlock value: oop.
	   interpreterProxy cCoerce: (interpreterProxy firstIndexableField: oop) asInteger to: #'unsigned short *']! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 10/26/2018 09:30'!
ccgLoad: codeGen expr: exprBlock asUnsignedValueFrom: stackIndex
	^[:oop|
	  interpreterProxy bytesPerOop = 4
		ifTrue: [interpreterProxy positive32BitValueOf: oop]
		ifFalse: [interpreterProxy positive64BitValueOf: oop]]! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/28/2016 21:16'!
ccgLoad: codeGen expr: exprBlock asWBFloatPtrFrom: stackIndex
	self shouldBeImplemented! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 10/27/2020 16:24'!
ccgValBlock: aString 
	^aString caseOf: {
		['isBytes']			-> [	[:oop|
								 interpreterProxy success: (interpreterProxy isBytes: oop).
								 oop]].
		['isShorts']			-> [	[:oop|
								 interpreterProxy success: (interpreterProxy isShorts: oop).
								 oop]].
		['isWords']			-> [	[:oop|
								 interpreterProxy success: (interpreterProxy isWords: oop).
								 oop]].
		['isLong64s']		-> [	[:oop|
								 interpreterProxy success: (interpreterProxy isLong64s: oop).
								 oop]].
		['isWordsOrBytes']	-> [	[:oop|
								 interpreterProxy success: (interpreterProxy isWordsOrBytes: oop).
								 oop]].
		['isIndexable']		-> [	[:oop|
								 interpreterProxy success: (interpreterProxy isIndexable: oop).
								 oop]].
		['isWordsOrShorts']	-> [	[:oop|
								 interpreterProxy success: (interpreterProxy isWordsOrShorts: oop).
								 oop]] }! !

!SmartSyntaxPluginSimulator methodsFor: 'initialize' stamp: 'eem 10/2/2010 22:31'!
computeSignatureFor: selector from: tuple
	| signature |
	self assert: tuple first == #forMap.
	signature := tuple third collect:
					[:className|
					(Smalltalk classNamed: className)
						ifNil: [self error: 'Argument class' , className, ' does not exist']
						ifNotNil:
							[:argClass|
							argClass
								ccg: self
								prolog: true
								expr: [interpreterProxy primitiveFail]
								index: nil]].
	^signatureMap
		at: tuple second asSymbol
		put: {	selector.
				signature.
				tuple fourth
					ifNil: [[:oop| oop]]
					ifNotNil:
						[:rcvrClassSymbol|
						(Smalltalk classNamed: rcvrClassSymbol)
							ifNil: [self error: 'Receiver class' , rcvrClassSymbol, ' does not exist']
							ifNotNil:
								[:rcvrClass|
								rcvrClass
									ccg: self
									prolog: false
									expr: [interpreterProxy primitiveFail]
									index: nil]] }! !

!SmartSyntaxPluginSimulator methodsFor: 'initialize' stamp: 'eem 10/27/2020 22:28'!
computeSignatureMap
	forMap := true. "true only while we compute the signatureMap"
	signatureMap := Dictionary new.
	(pluginClass withAllSuperclasses copyUpTo: SmartSyntaxInterpreterPlugin) do:
		[:theClass|
		 theClass selectorsAndMethodsDo:
			[:s :m|
			(m messages includesAnyOf: #(primitive: primitive:parameters: primitive:parameters:receiver:))
				ifTrue: [self getPrimitiveSignatureFor: s]
				ifFalse:
					[(m pragmaAt: #export:) ifNotNil:
						[:exportPragma|
						(exportPragma argumentAt: 1) ifTrue:
							[self computeSignatureFor: s from: { #forMap. s. #(). nil }]]]]].
	forMap := false! !

!SmartSyntaxPluginSimulator methodsFor: 'message forwarding' stamp: 'eem 7/27/2021 11:36'!
doesNotUnderstand: aMessage
	| signature selector parameters result sp |
	signature := signatureMap
					at: aMessage selector
					ifAbsent: [^super doesNotUnderstand: aMessage].
	self log: [interpreterProxy coInterpreter printExternalHeadFrame; print: aMessage selector; cr].
	"record the stack pointer to avoid cutting back the stack twice in plugins that mix smart syntax and traditional style."
	sp := interpreterProxy getStackPointer.
	selector := signature first.
	parameters := signature second.
	signature third "receiver block" value: (interpreterProxy stackValue: parameters size).
	interpreterProxy failed ifTrue:
		[self log: 'failed in receiver marshalling'.
		 ^nil].
	result := [| pluginArgs |
			   pluginArgs := parameters withIndexCollect:
										[:block :index|
										block value: (interpreterProxy stackValue: parameters size - index)].
			   interpreterProxy failed ifTrue:
				[self log: 'failed in argument marshalling'.
				 ^nil].
			   actualPlugin perform: selector withArguments: pluginArgs]
					on: Notification
					do: [:ex|
						ex tag == #getInterpreter ifTrue: [ex resume: interpreterProxy] ifFalse:
						[ex tag == #getSimulator ifTrue: [ex resume: self]
							ifFalse: [ex pass]]].
	interpreterProxy failed ifTrue:
		[self log: 'failed in execution'.
		 ^nil].
	result == actualPlugin ifTrue:
		[self log: '^self'.
		 "For methods in BitBltSimulator that do their own marshalling, don't double pop"
		 sp = interpreterProxy getStackPointer ifTrue:
			[interpreterProxy pop: interpreterProxy methodArgumentCount].
		 ^nil].
	self log: [interpreterProxy coInterpreter print: '^'; shortPrintOop: result; flush].
	"For methods in BitBltSimulator that do their own marshalling, don't double pop"
	sp = interpreterProxy getStackPointer ifTrue:
		[interpreterProxy
			pop: interpreterProxy methodArgumentCount + 1
			thenPush: result].
	^nil "SmartSyntaxPluginPrimitives return null"! !

!SmartSyntaxPluginSimulator methodsFor: 'accessing' stamp: 'eem 10/26/2018 09:59'!
forMap
	^forMap! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 10/11/2014 12:42'!
generateCoerceToBooleanObjectFrom: aBoolean on: ignored
	^interpreterProxy booleanObjectOf: aBoolean! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 10/11/2014 12:42'!
generateCoerceToFloatObjectFrom: aNumber on: ignored
	^interpreterProxy floatObjectOf: aNumber! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 10/2/2010 14:03'!
generateCoerceToSmallIntegerObjectFrom: aSmallInteger on: ignored 
	^interpreterProxy integerObjectOf: aSmallInteger! !

!SmartSyntaxPluginSimulator methodsFor: 'initialize' stamp: 'eem 10/2/2010 22:21'!
getPrimitiveSignatureFor: s
	"Execute the primitive until the send of #primitive:parameters: or primitive:parameters:receiver:,
	collect the processed signature and store it in the map"
	[actualPlugin perform: s withArguments: (1 to: s numArgs) asArray]
		on: Notification
		do: [:ex|
			(ex tag isArray
			 and: [ex tag first == #forMap]) ifTrue:
				[^self computeSignatureFor: s from: ex tag]].
	self error: 'can''t find primitive name in ', s! !

!SmartSyntaxPluginSimulator methodsFor: 'initialize' stamp: 'eem 10/2/2010 22:44'!
initialiseModule
	^(actualPlugin respondsTo: #initialiseModule) not
	  or: [actualPlugin initialiseModule]! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 7/2/2014 17:56'!
isPluginClass
	"This stands in for SmartSyntaxInterpreterPlugin classes during simulation."
	^true! !

!SmartSyntaxPluginSimulator methodsFor: 'accessing' stamp: 'eem 7/2/2014 19:05'!
isSmartSyntaxPluginSimulator
	<doNotGenerate>
	^true! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation support' stamp: 'eem 12/30/2016 13:17'!
log: aBlockOrString
	logging ifTrue:
		[aBlockOrString isBlock
			ifTrue: [aBlockOrString value]
			ifFalse: [interpreterProxy coInterpreter print: aBlockOrString; cr; flush]]! !

!SmartSyntaxPluginSimulator methodsFor: 'accessing' stamp: 'eem 12/30/2016 13:05'!
logging
	^logging! !

!SmartSyntaxPluginSimulator methodsFor: 'accessing' stamp: 'eem 12/30/2016 13:05'!
logging: anObject
	logging := anObject! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 12/28/2016 21:49'!
primitive: primNameString
	"If initializing, pass back the type signature.  If executing, answer nil."
	^self primitive: primNameString parameters: #() receiver: nil! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 10/2/2010 21:36'!
primitive: primNameString parameters: argumentClassNames
	"If initializing, pass back the type signature.  If executing, answer nil."
	^self primitive: primNameString parameters: argumentClassNames receiver: nil! !

!SmartSyntaxPluginSimulator methodsFor: 'simulation' stamp: 'eem 10/2/2010 21:35'!
primitive: primNameString parameters: argumentClassNames receiver: rcvrClassSymbolOrNil
	"If initializing, pass back the type signature.  If executing, answer the receiver or nil."
	forMap
		ifTrue:
			[Notification new
				tag: {#forMap. primNameString. argumentClassNames. rcvrClassSymbolOrNil};
				signal]
		ifFalse:
			[^rcvrClassSymbolOrNil ifNotNil:
				[interpreterProxy stackValue: interpreterProxy methodArgumentCount]]! !

!SmartSyntaxPluginSimulator methodsFor: 'message forwarding' stamp: 'eem 10/2/2010 22:31'!
respondsTo: aSelector
	^(signatureMap notNil and: [signatureMap includesKey: aSelector])
	  or: [super respondsTo: aSelector]! !

!SmartSyntaxPluginSimulator methodsFor: 'initialize' stamp: 'eem 12/30/2016 13:15'!
setInterpreter: anInterpreterProxy
	interpreterProxy := anInterpreterProxy.
	actualPlugin setInterpreter: anInterpreterProxy.
	logging ifNil: [logging := false].
	self computeSignatureMap! !

!SmartSyntaxPluginSimulator methodsFor: 'initialize' stamp: 'eem 11/10/2020 12:13'!
shutdownModule
	^(actualPlugin respondsTo: #shutdownModule)
		ifTrue: [actualPlugin shutdownModule]
		ifFalse: [true]! !

InterpreterPlugin subclass: #SoundCodecPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!SoundCodecPlugin commentStamp: 'tpr 5/5/2003 12:20' prior: 0!
This plugin provide GSM typecodec capabilities.!

!SoundCodecPlugin class methodsFor: 'accessing' stamp: 'tpr 5/23/2001 17:11'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!SoundCodecPlugin class methodsFor: 'translation' stamp: 'tpr 4/18/2002 15:56'!
moduleName

	^ 'SoundCodecPrims' "Needs to be the name used for module specification..."
! !

!SoundCodecPlugin class methodsFor: 'accessing' stamp: 'tpr 7/2/2001 16:34'!
requiresCrossPlatformFiles
	"If there cross platform files to be associated with the plugin, here is where you want to flag"
	^true! !

!SoundCodecPlugin methodsFor: 'gsm 6.10 codec' stamp: 'eem 9/12/2020 13:32'!
primitiveGSMDecode

	| dstIndex dst srcIndex src frameCount state srcSize dstSize result srcDelta dstDelta |
	<export: true>
	dstIndex := interpreterProxy stackIntegerValue: 0.
	dst := interpreterProxy stackValue: 1.
	srcIndex := interpreterProxy stackIntegerValue: 2.
	src := interpreterProxy stackValue: 3.
	frameCount := interpreterProxy stackIntegerValue: 4.
	state := interpreterProxy stackValue: 5.
	interpreterProxy success: ((interpreterProxy isWordsOrShorts: dst)
								and: [(interpreterProxy isBytes: src)
								and: [interpreterProxy isBytes: state]]).
	interpreterProxy failed ifTrue:[^ nil].
	srcSize := interpreterProxy byteSizeOf: src.
	dstSize := (interpreterProxy byteSizeOf: dst) / 2.
	self gsmDecode: state + BaseHeaderSize _: frameCount _: src _: srcIndex _: srcSize _: dst _: dstIndex _: dstSize _: (self addressOf: srcDelta) _: (self addressOf: dstDelta).
	interpreterProxy failed ifTrue:[^ nil].
	result := interpreterProxy makePointwithxValue: srcDelta yValue: dstDelta.
	interpreterProxy failed ifTrue:[^ nil].
	interpreterProxy methodReturnValue: result! !

!SoundCodecPlugin methodsFor: 'gsm 6.10 codec' stamp: 'eem 9/12/2020 13:37'!
primitiveGSMEncode

	| dstIndex dst srcIndex src frameCount state srcSize dstSize result srcDelta dstDelta |
	<export: true>
	dstIndex := interpreterProxy stackIntegerValue: 0.
	dst := interpreterProxy stackValue: 1.
	srcIndex := interpreterProxy stackIntegerValue: 2.
	src := interpreterProxy stackValue: 3.
	frameCount := interpreterProxy stackIntegerValue: 4.
	state := interpreterProxy stackValue: 5.
	interpreterProxy success: ((interpreterProxy isBytes: dst)
								and: [(interpreterProxy isWordsOrShorts: src)
								and: [interpreterProxy isBytes: state]]).
	interpreterProxy failed ifTrue:[^ nil].
	srcSize := (interpreterProxy byteSizeOf: src) / 2.
	dstSize := interpreterProxy byteSizeOf: dst.
	self gsmEncode: state + BaseHeaderSize _: frameCount _: src _: srcIndex _: srcSize _: dst _: dstIndex _: dstSize _: (self addressOf: srcDelta) _: (self addressOf: dstDelta).
	interpreterProxy failed ifTrue:[^ nil].
	result := interpreterProxy makePointwithxValue: srcDelta yValue: dstDelta.
	interpreterProxy failed ifTrue:[^ nil].
	interpreterProxy methodReturnValue: result! !

!SoundCodecPlugin methodsFor: 'gsm 6.10 codec' stamp: 'eem 9/13/2020 15:37'!
primitiveGSMNewState

	<export: true>
	(interpreterProxy
				instantiateClass: interpreterProxy classByteArray
				indexableSize: self gsmStateBytes)
		ifNil: [interpreterProxy primitiveFailFor: PrimErrNoMemory]
		ifNotNil:
			[:state|
			 self gsmInitState: state + BaseHeaderSize.
			 interpreterProxy methodReturnValue: state]! !

InterpreterPlugin subclass: #SoundGenerationPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!SoundGenerationPlugin commentStamp: '<historical>' prior: 0!
This class is a stub for the directly generated primitives in AbstractSound and subclasses.!

!SoundGenerationPlugin class methodsFor: 'translation' stamp: 'ar 2/3/2001 18:16'!
declareCVarsIn: cg
	cg addHeaderFile: '"SoundGenerationPlugin.h"'.! !

!SoundGenerationPlugin class methodsFor: 'accessing' stamp: 'tpr 5/23/2001 17:11'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!SoundGenerationPlugin class methodsFor: 'translation' stamp: 'eem 1/3/2013 11:51'!
monticelloDescription
	"Override to include the AbstractSound class."
	"self monticelloDescription"
	^super monticelloDescription, '\' withCRs, (CCodeGenerator monticelloDescriptionFor: AbstractSound)! !

!SoundGenerationPlugin class methodsFor: 'translation' stamp: 'eem 2/16/2015 14:19'!
translatedPrimitives
	"Answer an Array of Class, selector pair Arrays for any primitives to
	 be translated from Smalltalk methods elsewhere in the system."
	^AbstractSound translatedPrimitives! !

InterpreterPlugin subclass: #SqueakSSLPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!SqueakSSLPlugin class methodsFor: 'translation' stamp: 'ar 6/23/2010 21:27'!
hasHeaderFile
	^true! !

!SqueakSSLPlugin class methodsFor: 'translation' stamp: 'ar 7/16/2010 00:38'!
moduleName
	^'SqueakSSL'! !

!SqueakSSLPlugin methodsFor: 'primitives' stamp: 'eem 6/28/2016 12:55'!
primitiveAccept
	"Primitive. Starts or continues a server handshake using the current session.
	Will eventually produce output to be sent to the client. Requires the host
	and cert name to be set for the session. Returns a code indicating the sate
	of the connection:
		> 0	 - Number of bytes to be sent to the client.
		0	 - Success. The connection is established.
		-1 	 - More input is required.
		< -1 - Other errors.
	"
	| start srcLen dstLen srcOop dstOop handle srcPtr dstPtr result |
	<var: #srcPtr type: #'char *'>
	<var: #dstPtr type: #'char *'>
	<export: true>
	interpreterProxy methodArgumentCount = 5
		ifFalse:[^interpreterProxy primitiveFail].
	dstOop := interpreterProxy stackValue: 0.
	srcLen := interpreterProxy stackIntegerValue: 1.
	start := interpreterProxy stackIntegerValue: 2.
	srcOop := interpreterProxy stackValue: 3.
	handle := interpreterProxy stackIntegerValue: 4.
	interpreterProxy failed ifTrue:[^nil].
	((start > 0 and:[srcLen >= 0])
		and:[(interpreterProxy isBytes: srcOop) 
		and:[(interpreterProxy isBytes: dstOop) 
		and:[(interpreterProxy byteSizeOf: srcOop) >= (start + srcLen - 1)]]])
			ifFalse:[^interpreterProxy primitiveFail].
	srcPtr := interpreterProxy firstIndexableField: srcOop.
	dstPtr := interpreterProxy firstIndexableField: dstOop.
	srcPtr := srcPtr + start - 1.
	dstLen := interpreterProxy byteSizeOf: dstOop.
	result := self cCode: 'sqAcceptSSL(handle, srcPtr, srcLen, dstPtr, dstLen)' 
					inSmalltalk:[handle. srcPtr. srcLen. dstPtr. dstLen. -2].
	interpreterProxy failed ifTrue:[^nil].
	interpreterProxy pop: interpreterProxy methodArgumentCount+1.
	interpreterProxy pushInteger: result.! !

!SqueakSSLPlugin methodsFor: 'primitives' stamp: 'eem 6/28/2016 12:55'!
primitiveConnect
	"Primitive. Starts or continues a client handshake using the provided data.
	Will eventually produce output to be sent to the server. Requires the host
	name to be set for the session. 
	Returns:
		> 0	 - Number of bytes to be sent to the server
		0	 - Success. The connection is established.
		-1 	 - More input is required.
		< -1 - Other errors.
	"
	| start srcLen dstLen srcOop dstOop handle srcPtr dstPtr result |
	<var: #srcPtr type: #'char *'>
	<var: #dstPtr type: #'char *'>
	<export: true>
	interpreterProxy methodArgumentCount = 5
		ifFalse:[^interpreterProxy primitiveFail].
	dstOop := interpreterProxy stackValue: 0.
	srcLen := interpreterProxy stackIntegerValue: 1.
	start := interpreterProxy stackIntegerValue: 2.
	srcOop := interpreterProxy stackValue: 3.
	handle := interpreterProxy stackIntegerValue: 4.
	interpreterProxy failed ifTrue:[^nil].
	((start > 0 and:[srcLen >= 0])
		and:[(interpreterProxy isBytes: srcOop) 
		and:[(interpreterProxy isBytes: dstOop) 
		and:[(interpreterProxy byteSizeOf: srcOop) >= (start + srcLen - 1)]]])
			ifFalse:[^interpreterProxy primitiveFail].
	srcPtr := interpreterProxy firstIndexableField: srcOop.
	dstPtr := interpreterProxy firstIndexableField: dstOop.
	srcPtr := srcPtr + start - 1.
	dstLen := interpreterProxy byteSizeOf: dstOop.
	result := self cCode: 'sqConnectSSL(handle, srcPtr, srcLen, dstPtr, dstLen)' 
					inSmalltalk:[handle. srcPtr. srcLen. dstPtr. dstLen. -2].
	interpreterProxy failed ifTrue:[^nil].
	interpreterProxy pop: interpreterProxy methodArgumentCount+1.
	interpreterProxy pushInteger: result.! !

!SqueakSSLPlugin methodsFor: 'primitives' stamp: 'ar (auto pragmas eem 2014-01-09) 6/23/2010 20:56'!
primitiveCreate
	"Primitive. Creates a new SSL session and returns its handle."
	| handle |
	<export: true>
	interpreterProxy methodArgumentCount = 0 
		ifFalse:[^interpreterProxy primitiveFail].
	handle := self cCode: 'sqCreateSSL()' inSmalltalk:[0].
	handle = 0 ifTrue:[^interpreterProxy primitiveFail].
	interpreterProxy pop: interpreterProxy methodArgumentCount+1.
	interpreterProxy pushInteger: handle.
! !

!SqueakSSLPlugin methodsFor: 'primitives' stamp: 'eem 6/28/2016 12:56'!
primitiveDecrypt
	"Primitive. Decrypts a buffer sent via the connection.
	Requires the session to be established.
	Returns:
		>=0 - Number of bytes decrypted in the result buffer
		< -1 - Other errors.
	"
	| start srcLen dstLen srcOop dstOop handle srcPtr dstPtr result |
	<var: #srcPtr type: 'char *'>
	<var: #dstPtr type: 'char *'>
	<export: true>
	interpreterProxy methodArgumentCount = 5
		ifFalse:[^interpreterProxy primitiveFail].
	dstOop := interpreterProxy stackValue: 0.
	srcLen := interpreterProxy stackIntegerValue: 1.
	start := interpreterProxy stackIntegerValue: 2.
	srcOop := interpreterProxy stackValue: 3.
	handle := interpreterProxy stackIntegerValue: 4.
	interpreterProxy failed ifTrue:[^nil].
	((start > 0 and:[srcLen >= 0])
		and:[(interpreterProxy isBytes: srcOop) 
		and:[(interpreterProxy isBytes: dstOop) 
		and:[(interpreterProxy byteSizeOf: srcOop) >= (start + srcLen - 1)]]])
			ifFalse:[^interpreterProxy primitiveFail].
	srcPtr := interpreterProxy firstIndexableField: srcOop.
	dstPtr := interpreterProxy firstIndexableField: dstOop.
	srcPtr := srcPtr + start - 1.
	dstLen := interpreterProxy byteSizeOf: dstOop.
	result := self cCode: 'sqDecryptSSL(handle, srcPtr, srcLen, dstPtr, dstLen)' 
					inSmalltalk:[handle. srcPtr. srcLen. dstPtr. dstLen. -2].
	interpreterProxy failed ifTrue:[^nil].
	interpreterProxy pop: interpreterProxy methodArgumentCount+1.
	interpreterProxy pushInteger: result.! !

!SqueakSSLPlugin methodsFor: 'primitives' stamp: 'ar (auto pragmas eem 2014-01-09) 7/15/2010 20:55'!
primitiveDestroy
	"Primitive. Destroys an SSL session."

	| handle result |
	<export: true>
	interpreterProxy methodArgumentCount = 1
		ifFalse:[^interpreterProxy primitiveFail].
	handle := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: 'sqDestroySSL(handle)' inSmalltalk:[handle. 0].
	result = 0 ifTrue:[^interpreterProxy primitiveFail].
	interpreterProxy pop: interpreterProxy methodArgumentCount.

! !

!SqueakSSLPlugin methodsFor: 'primitives' stamp: 'eem 6/28/2016 12:56'!
primitiveEncrypt
	"Primitive. Encrypts a buffer to be sent to the via the connection.
	Requires the session to be established.
	Returns:
		>=0 - Number of bytes encrypted in the result buffer
		< -1 - Other errors.
	"
	| start srcLen dstLen srcOop dstOop handle srcPtr dstPtr result |
	<var: #srcPtr type: 'char *'>
	<var: #dstPtr type: 'char *'>
	<export: true>
	interpreterProxy methodArgumentCount = 5
		ifFalse:[^interpreterProxy primitiveFail].
	dstOop := interpreterProxy stackValue: 0.
	srcLen := interpreterProxy stackIntegerValue: 1.
	start := interpreterProxy stackIntegerValue: 2.
	srcOop := interpreterProxy stackValue: 3.
	handle := interpreterProxy stackIntegerValue: 4.
	interpreterProxy failed ifTrue:[^nil].
	((start > 0 and:[srcLen >= 0])
		and:[(interpreterProxy isBytes: srcOop) 
		and:[(interpreterProxy isBytes: dstOop) 
		and:[(interpreterProxy byteSizeOf: srcOop) >= (start + srcLen - 1)]]])
			ifFalse:[^interpreterProxy primitiveFail].
	srcPtr := interpreterProxy firstIndexableField: srcOop.
	dstPtr := interpreterProxy firstIndexableField: dstOop.
	srcPtr := srcPtr + start - 1.
	dstLen := interpreterProxy byteSizeOf: dstOop.
	result := self cCode: 'sqEncryptSSL(handle, srcPtr, srcLen, dstPtr, dstLen)' 
					inSmalltalk:[handle. srcPtr. srcLen. dstPtr. dstLen. -2].
	interpreterProxy failed ifTrue:[^nil].
	interpreterProxy pop: interpreterProxy methodArgumentCount+1.
	interpreterProxy pushInteger: result.! !

!SqueakSSLPlugin methodsFor: 'primitives' stamp: 'ar (auto pragmas eem 2014-01-09) 7/16/2010 22:14'!
primitiveGetIntProperty
	"Primitive. Returns an integer property for the session"

	| propID handle value |
	<export: true>
	interpreterProxy methodArgumentCount = 2
		ifFalse:[^interpreterProxy primitiveFail].
	propID := interpreterProxy stackIntegerValue: 0.
	handle := interpreterProxy stackIntegerValue: 1.
	interpreterProxy failed ifTrue:[^nil].
	value := self cCode: 'sqGetIntPropertySSL(handle, propID)' 
					inSmalltalk:[handle. propID. 0].
	interpreterProxy failed ifTrue:[^nil].
	interpreterProxy 
		pop: interpreterProxy methodArgumentCount + 1
		thenPush: (interpreterProxy signed32BitIntegerFor: value)
! !

!SqueakSSLPlugin methodsFor: 'primitives' stamp: 'ar (auto pragmas eem 2014-01-09) 7/16/2010 22:16'!
primitiveGetStringProperty
	"Primitive. Returns a string property for the session"

	| stringLen stringOop propID handle stringPtr oopPtr |
	<var: #stringPtr type: 'char *'>
	<var: #oopPtr type: 'char *'>
	<export: true>
	interpreterProxy methodArgumentCount = 2
		ifFalse:[^interpreterProxy primitiveFail].
	propID := interpreterProxy stackIntegerValue: 0.
	handle := interpreterProxy stackIntegerValue: 1.
	interpreterProxy failed ifTrue:[^nil].
	stringPtr := self cCode: 'sqGetStringPropertySSL(handle, propID)' 
					inSmalltalk:[handle. propID. nil].
	interpreterProxy failed ifTrue:[^nil].
	stringPtr == nil ifTrue:[
		stringOop := interpreterProxy nilObject.
	] ifFalse:[
		stringLen := self strlen: stringPtr.
		stringOop := interpreterProxy 
			instantiateClass: (interpreterProxy classString)
			indexableSize: stringLen.
		oopPtr := interpreterProxy firstIndexableField: stringOop.
		0 to: stringLen-1 do:[:i| oopPtr at: i put: (stringPtr at: i)].
	].
	interpreterProxy 
		pop: interpreterProxy methodArgumentCount + 1
		thenPush: stringOop.
! !

!SqueakSSLPlugin methodsFor: 'primitives' stamp: 'ar (auto pragmas eem 2014-01-09) 7/16/2010 22:16'!
primitiveSetIntProperty
	"Primitive. Sets a integer property for the session"

	| propID handle result value |
	<export: true>
	interpreterProxy methodArgumentCount = 3
		ifFalse:[^interpreterProxy primitiveFail].
	value := interpreterProxy signed32BitValueOf: (interpreterProxy stackValue: 0).
	propID := interpreterProxy stackIntegerValue: 1.
	handle := interpreterProxy stackIntegerValue: 2.
	interpreterProxy failed ifTrue:[^nil].
	result := self cCode: 'sqSetIntPropertySSL(handle, propID, value)' 
					inSmalltalk:[handle. propID. value. false].
	result ifFalse:[^interpreterProxy primitiveFail].
	interpreterProxy failed ifTrue:[^nil].
	interpreterProxy pop: interpreterProxy methodArgumentCount.
! !

!SqueakSSLPlugin methodsFor: 'primitives' stamp: 'eem 6/28/2016 12:57'!
primitiveSetStringProperty
	"Primitive. Sets a string property for the session"

	| srcLen srcOop propID handle srcPtr result |
	<var: #srcPtr type: 'char *'>
	<export: true>
	interpreterProxy methodArgumentCount = 3
		ifFalse:[^interpreterProxy primitiveFail].
	srcOop := interpreterProxy stackValue: 0.
	propID := interpreterProxy stackIntegerValue: 1.
	handle := interpreterProxy stackIntegerValue: 2.
	interpreterProxy failed ifTrue:[^nil].
	(interpreterProxy isBytes: srcOop) 
		ifFalse:[^interpreterProxy primitiveFail].
	srcPtr := interpreterProxy firstIndexableField: srcOop.
	srcLen := interpreterProxy byteSizeOf: srcOop.
	result := self cCode: 'sqSetStringPropertySSL(handle, propID, srcPtr, srcLen)' 
					inSmalltalk:[handle. srcPtr. propID. srcLen. false].
	result ifFalse:[^interpreterProxy primitiveFail].
	interpreterProxy failed ifTrue:[^nil].
	interpreterProxy pop: interpreterProxy methodArgumentCount.
! !

InterpreterPlugin subclass: #StarSqueakPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!StarSqueakPlugin commentStamp: '<historical>' prior: 0!
This plugin defines primitives accelerators to support StarSqueak. The plugin library, usually named "StarSqueakPlugin", should be put in the same folder as the Squeak interpreter. If this plugin is not available the primitives will still work, but they will be run much more slowly, since they will be running as Squeak code.
!

!StarSqueakPlugin methodsFor: 'all' stamp: 'jm (auto pragmas 12/08) 1/20/2001 11:01'!
checkedUnsignedIntPtrOf: oop
	"Return an unsigned int pointer to the first indexable word of oop, which must be a words object."

	<returnTypeC: 'unsigned int *'>
	interpreterProxy success: (interpreterProxy isWords: oop).
	interpreterProxy failed ifTrue: [^ 0].
	^ self cCoerce: (interpreterProxy firstIndexableField: oop) to: 'unsigned int *'
! !

!StarSqueakPlugin methodsFor: 'all' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:22'!
primitiveDiffuseFromToWidthHeightDelta
	"Diffuse the integer values of the source patch variable Bitmap into the output Bitmap. Each cell of the output is the average of the NxN area around it in the source, where N = (2 * delta) + 1."

	| srcOop dstOop height width delta src dst area startY endY startX endX sum rowStart |
	<export: true>
	<var: 'src' type: 'unsigned int *'>
	<var: 'dst' type: 'unsigned int *'>

	srcOop := interpreterProxy stackValue: 4.
	dstOop := interpreterProxy stackValue: 3.
	width := interpreterProxy stackIntegerValue: 2.
	height := interpreterProxy stackIntegerValue: 1.
	delta := interpreterProxy stackIntegerValue: 0.
	src := self checkedUnsignedIntPtrOf: srcOop.
	dst := self checkedUnsignedIntPtrOf: dstOop.
	interpreterProxy success:
		(interpreterProxy stSizeOf: srcOop) = (interpreterProxy stSizeOf: dstOop).
	interpreterProxy success:
		(interpreterProxy stSizeOf: srcOop) = (width * height).
	interpreterProxy failed ifTrue: [^ nil].

	area := ((2 * delta) + 1) * ((2 * delta) + 1).
	0 to: height - 1 do: [:y |
		startY := y - delta.
		startY < 0 ifTrue: [startY := 0].
		endY := y + delta.
		endY >= height ifTrue: [endY := height - 1].
		0 to: width - 1 do: [:x |
			startX := x - delta.
			startX < 0 ifTrue: [startX := 0].
			endX := x + delta.
			endX >= width ifTrue: [endX := width - 1].

			sum := 0.
			startY to: endY do: [:y2 |
				rowStart := y2 * width.
				startX to: endX do: [:x2 |
					sum := sum + (src at: rowStart + x2)]].

			dst at: ((y * width) + x) put: (sum // area)]].

	interpreterProxy pop: 5.  "pop args, leave rcvr on stack"
! !

!StarSqueakPlugin methodsFor: 'all' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:22'!
primitiveEvaporateRate
	"Evaporate the integer values of the source Bitmap at the given rate. The rate is an integer between 0 and 1024, where 1024 is a scale factor of 1.0 (i.e., no evaporation)."

	| patchVarOop rate patchVar sz |
	<export: true>
	<var: 'patchVar' type: 'unsigned int *'>

	patchVarOop := interpreterProxy stackValue: 1.
	rate := interpreterProxy stackIntegerValue: 0.
	patchVar := self checkedUnsignedIntPtrOf: patchVarOop.
	sz := interpreterProxy stSizeOf: patchVarOop.
	interpreterProxy failed ifTrue: [^ nil].

	0 to: sz - 1 do: [:i |
		patchVar at: i put: (((patchVar at: i) * rate) >> 10)].

	interpreterProxy pop: 2.  "pop args, leave rcvr on stack"
! !

!StarSqueakPlugin methodsFor: 'all' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:22'!
primitiveMapFromToWidthHeightPatchSizeRgbFlagsShift

	| srcOop dstOop w h patchSize rgbFlags shiftAmount src dst rgbMult srcIndex level pixel offset |
	<export: true>
	<var: 'src' type: 'unsigned int *'>
	<var: 'dst' type: 'unsigned int *'>

	srcOop := interpreterProxy stackValue: 6.
	dstOop := interpreterProxy stackValue: 5.
	w := interpreterProxy stackIntegerValue: 4.
	h := interpreterProxy stackIntegerValue: 3.
	patchSize := interpreterProxy stackIntegerValue: 2.
	rgbFlags := interpreterProxy stackIntegerValue: 1.
	shiftAmount := interpreterProxy stackIntegerValue: 0.

	src := self checkedUnsignedIntPtrOf: srcOop.
	dst := self checkedUnsignedIntPtrOf: dstOop.
	interpreterProxy success:
		(interpreterProxy stSizeOf: dstOop) = (w * h).
	interpreterProxy success:
		(interpreterProxy stSizeOf: dstOop) = ((interpreterProxy stSizeOf: srcOop) * patchSize * patchSize).
	interpreterProxy failed ifTrue: [^ nil].

	rgbMult := 0.
	(rgbFlags bitAnd: 2r100) > 0 ifTrue: [rgbMult := rgbMult + 16r10000].
	(rgbFlags bitAnd: 2r10) > 0 ifTrue: [rgbMult := rgbMult + 16r100].
	(rgbFlags bitAnd: 2r1) > 0 ifTrue: [rgbMult := rgbMult + 16r1].
	srcIndex := -1.
	0 to: (h // patchSize) - 1 do: [:y |
		0 to: (w // patchSize) - 1 do: [:x |
			level := (src at: (srcIndex := srcIndex + 1)) bitShift: shiftAmount.
			level > 255 ifTrue: [level := 255].
			level <= 0
				ifTrue: [pixel := 1]  "non-transparent black"
				ifFalse: [pixel := level * rgbMult].

			"fill a patchSize x patchSize square with the pixel value"
			offset := ((y * w) + x) * patchSize.
			offset to: offset + ((patchSize - 1) * w) by: w do: [:rowStart |
				rowStart to: rowStart + patchSize - 1 do: [:dstIndex |
					dst at: dstIndex put: pixel]] ]].

	interpreterProxy pop: 7.  "pop args, leave rcvr on stack"
! !

InterpreterPlugin subclass: #SurfacePlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!SurfacePlugin commentStamp: '<historical>' prior: 0!
This plugin is a fake. It doesn't do anything useful. It's sole purpose is to wrap the C code that's associated with the SurfacePlugin into the main VM generation process. Since the C code isn't easily generated from ST code this is (unfortunately) necessary. But look on the bright side - you don't have to define any weird stuff for the C compiler. Isn't that great?!! (just kidding...) !

!SurfacePlugin class methodsFor: 'translation' stamp: 'tpr 5/23/2001 17:12'!
hasHeaderFile
	"If there is a single intrinsic header file to be associated with the plugin, here is where you want to flag"
	^true! !

!SurfacePlugin class methodsFor: 'translation' stamp: 'tpr 7/4/2001 15:15'!
requiresCrossPlatformFiles
	"If there cross platform files to be associated with the plugin, here is where you want to flag"
	^true! !

!SurfacePlugin class methodsFor: 'translation' stamp: 'tpr 12/3/2015 10:28'!
translateInDirectory: directory doInlining: inlineFlag
"handle a special case external file rather than normal generated code."
	| cg |
	self initialize.

	cg := self buildCodeGenerator.

	"We rely on the fake entry points implemented on the instance side to allow the export list to be accurate. Please update them if you change the code"
	^cg exportedPrimitiveNames asArray! !

!SurfacePlugin methodsFor: 'fake entry points' stamp: 'ar (auto pragmas 12/08) 5/26/2000 22:33'!
initialiseModule
	"Fake entry point"
	<export: true>! !

!SurfacePlugin methodsFor: 'fake entry points' stamp: 'ar (auto pragmas 12/08) 5/26/2000 22:34'!
ioFindSurface
	"Fake entry point"
	<export: true>! !

!SurfacePlugin methodsFor: 'fake entry points' stamp: 'ar (auto pragmas 12/08) 5/26/2000 22:34'!
ioGetSurfaceFormat
	"Fake entry point"
	<export: true>! !

!SurfacePlugin methodsFor: 'fake entry points' stamp: 'ar (auto pragmas 12/08) 5/26/2000 22:34'!
ioLockSurface
	"Fake entry point"
	<export: true>! !

!SurfacePlugin methodsFor: 'fake entry points' stamp: 'ar (auto pragmas 12/08) 5/26/2000 22:34'!
ioRegisterSurface
	"Fake entry point"
	<export: true>! !

!SurfacePlugin methodsFor: 'fake entry points' stamp: 'ar (auto pragmas 12/08) 5/26/2000 22:34'!
ioShowSurface
	"Fake entry point"
	<export: true>! !

!SurfacePlugin methodsFor: 'fake entry points' stamp: 'ar (auto pragmas 12/08) 5/26/2000 22:34'!
ioUnlockSurface
	"Fake entry point"
	<export: true>! !

!SurfacePlugin methodsFor: 'fake entry points' stamp: 'ar (auto pragmas 12/08) 5/26/2000 22:34'!
ioUnregisterSurface
	"Fake entry point"
	<export: true>! !

!SurfacePlugin methodsFor: 'fake entry points' stamp: 'ar (auto pragmas 12/08) 5/26/2000 22:33'!
shutdownModule
	"Fake entry point"
	<export: true>! !

InterpreterPlugin subclass: #ThreadedFFIPlugin
	instanceVariableNames: 'ffiLogEnabled externalFunctionInstSize ffiLastError allocationMap'
	classVariableNames: 'DefaultMaxStackSize ExternalFunctionAddressIndex ExternalFunctionArgTypesIndex ExternalFunctionFlagsIndex ExternalFunctionStackSizeIndex MaxNumArgs PluginVersionInfo'
	poolDictionaries: 'FFIConstants'
	category: 'VMMaker-Plugins-FFI'!

!ThreadedFFIPlugin commentStamp: 'eem 2/6/2019 16:40' prior: 0!
This plugin provides access to foreign function interfaces on those platforms that provide such. For example Windows DLLs and unix .so's.  This version is designed to support reentrancy and threading, and so uses alloca to stack allocate all memory needed for a given callout.  Specific platforms are implemented by concrete subclasses.  Threaded calls can only be provided within the context of the threaded VM; othewise calls must be blocking.  So code specific to threading is guarded with a
	self cppIf: COGMTVM
		ifTrue: [...]
form to arrange that it is only compiled in the threaded VM context.

The callout primitives consume a type spec that defines the signature of the function to be called and a vector of arguments.  The type spec may be extracted from the method containing an FFI pragma or as an explicit parameter.  The arguments to be passed may either be arguments to the the method containing an FFI pragma or as an explicit Array of parameters.

Space is allocated to house the marshalled parameters and the type spec and arguments are then parsed to marshall the actual parameters into that space.  The space is some combination of alloca'ed (stack allocated) memory for parameters passed on the stack and a "callout state" struct (an instance of ThreadedFFICalloutState) to hold any parameters to be passed in registers.

By using C's facilities appropriately we can arrange that the C compiler generates code for passing all parameters, avoiding having to descend to the assembler or machine-code level (*). The basic scheme is to use alloca to stack allocate space for passing stacked parameters, since the memory allocated by alloca is at top-of-stack and in exactly the right place for parameter passing (*), and to invoke the function to be called with as many arguments as there are integer register parameters in the calling convention.  For example, on x86/IA32 there are no register parameters and all arguments are passed on the stack, while on ARMv4/5/6 there are four integer register parameters.  Since float results are typically answered through a floating-point register and integer/pointer results answered through one (or, on 32-bits for 64-bit results, two) register(s).  So on x86 the function to be called is invoked using

	floatRet := self dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'double (*)()')
or
	intRet := self dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'usqLong (*)()')

On ARMv4/5/6 it is invoked with

	floatRet := self 
			dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'float (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
			with: (calloutState integerRegisters at: 0)
			with: (calloutState integerRegisters at: 1)
			with: (calloutState integerRegisters at: 2)
			with: (calloutState integerRegisters at: 3)
or
	floatRet := self 
				dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'double (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
				with: (calloutState integerRegisters at: 0)
				with: (calloutState integerRegisters at: 1)
				with: (calloutState integerRegisters at: 2)
				with: (calloutState integerRegisters at: 3)
or
	intRet := self 
				dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'usqIntptr_t (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
				with: (calloutState integerRegisters at: 0)
				with: (calloutState integerRegisters at: 1)
				with: (calloutState integerRegisters at: 2)
				with: (calloutState integerRegisters at: 3)

Hence the C compiler generates the code to pass register parameters appropriately.  All we have to do is provide sufficient register parameters and give the function (pointer) to be called a suitable type.

Likewise we also persuade the C compiler to generate code to load any floating-point register arguments by preceding any call of a function that takes floating-point arguments passed in registers with a call to loadFloatRegs with as many floating-point parameters as there are floating-point parameter registers.  loadFloatRegs is implemented in platforms//Cross/plugins/SqueakFFIPrims/sqFFIPlugin.c as an empty function.  So floating-point registers are passed via calls such as:

	calloutState floatRegisterIndex > 0 ifTrue:
		[self loadFloatRegs:
			   ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 0)) to: #'double *') at: 0)
			_: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 2)) to: #'double *') at: 0)
			_: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 4)) to: #'double *') at: 0)
			_: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 6)) to: #'double *') at: 0)
			_: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 8)) to: #'double *') at: 0)
			_: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 10)) to: #'double *') at: 0)
			_: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 12)) to: #'double *') at: 0)
			_: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 14)) to: #'double *') at: 0)

(*) some implementations of alloca do not answer the actual top-of-stack, but may answer an address a word or two away.  On these implementations we merely derive the top-of-stack instead of using the result of alloca, or perhaps set the stack pointer to the result of alloca, as befits the platform.

Callbacks are not handled by this plugin.  Instead they are handled by image-level code to create Alien "thunks", small sequences of machine code, that invoke a function thunkEntry, implemented as required by each platform in one of the files in platforms/Cross/plugins/IA32ABI (a regrettable name; it should be something like platforms/Cross/plugins/AlienCallbacks).  Each platform's thunkEntry has a signature that takes all integer register parameters (if any), all floating-point register parameters (if any), a pointer to the thunk, and the stack pointer at the point the thunk was invoiked.  To pass a callback to foreign code, the marshaller in this plugin passes the address of the thunk.  When external code calls that address the think is invoked, and it invokes thunkEntry as required.  thunkEntry then saves its parameters locally in an instance of VMCallbackContext, which includes a jmpbuf.  thunkEntry then does a setjmp and enters the VM via sendInvokeCallbackContext[:], which creates an activation of invokeCallbackContext: to execute the callback.  This method then extracts the address of the thunk from the VMCallbackContext and invokes machinery in Callback to match the address with a Smalltalk block.  The block is then evaluated with the parameters extracted from the VMCallbackContext, marshalled by methods in Callback subclasses signatures protocol, which know whether a parameter would be passed in a register or on the stack.  The result of the block is then passed back to thunkEntry by storing it in a field in the VMCallbackContext and invoking primSignal:andReturnAs:fromContext:, which uses longjmp to jump back to thunkEntry which then extracts the result from its VMCallbackContext and returns.

For  example, the signature of thunkEntry on x86/IA32 (platforms/Cross/plugins/IA32ABI/ia32abicc.c) is
	long
	thunkEntry(void *thunkp, sqIntptr_t *stackp)
whereas on ARMv4/5/6 (platforms/Cross/plugins/IA32ABI/arm32abicc.c) it is
	long long
	thunkEntry(long r0, long r1, long r2, long r3,
		            double d0, double d1, double d2, double d3,
		            double d4, double d5, double d6, double d7,
		            void *thunkpPlus16, sqIntptr_t *stackp)!

ThreadedFFIPlugin subclass: #ThreadedARM32FFIPlugin
	instanceVariableNames: ''
	classVariableNames: 'NumFloatRegArgs NumIntRegArgs'
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedARM32FFIPlugin commentStamp: '<historical>' prior: 0!
This subclass is for the 32-bit ARM ABI.  It typically has 4 integer registers.!

!ThreadedARM32FFIPlugin class methodsFor: 'translation' stamp: 'eem 1/20/2019 19:02'!
calloutStateClass
	^ThreadedFFICalloutStateForARM32! !

!ThreadedARM32FFIPlugin class methodsFor: 'translation' stamp: 'eem 6/12/2019 18:42'!
identifyingPredefinedMacros
	^#('__ARM_ARCH_5__' '__ARM_ARCH_6__' '__ARM_ARCH_7__' '__arm32__' 'ARM32' '_M_ARM')! !

!ThreadedARM32FFIPlugin class methodsFor: 'class initialization' stamp: 'KenD 12/23/2018 01:41'!
initialize
	super initialize.
	NumIntRegArgs := 4.
	NumFloatRegArgs := 16! !

!ThreadedARM32FFIPlugin class methodsFor: 'translation' stamp: 'eem 4/21/2014 11:04'!
moduleName
	^'ARM32FFIPlugin'! !

!ThreadedARM32FFIPlugin class methodsFor: 'accessing' stamp: 'djm 4/12/2014 19:13'!
numFloatRegArgs
	^NumFloatRegArgs! !

!ThreadedARM32FFIPlugin class methodsFor: 'accessing' stamp: 'djm 4/12/2014 19:13'!
numIntRegArgs
	^NumIntRegArgs! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
canReturnInRegistersStructOfSize: returnStructSize
	"Answer if a struct result of a given size is returned in registers or not."
	<inline: #always>
	^returnStructSize <= self wordSize! !

!ThreadedARM32FFIPlugin methodsFor: 'callout support' stamp: 'eem 7/27/2021 16:46'!
encodeStructReturnTypeIn: calloutState
	"Set the return type to true if returning the struct via register"
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>

	calloutState structReturnType: (self canReturnInRegistersStructOfSize: calloutState structReturnSize)! !

!ThreadedARM32FFIPlugin methodsFor: 'callout support' stamp: 'eem 8/29/2020 13:20'!
ffiCalloutTo: procAddr SpecOnStack: specOnStack in: calloutState
	<var: #procAddr type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	<var: #loadFloatRegs declareC: 'extern void loadFloatRegs(double, double, double, double, double, double, double, double)'>
	"Go out, call this guy and create the return value.  This *must* be inlined because of
	 the alloca of the outgoing stack frame in ffiCall:WithFlags:NumArgs:Args:AndTypes:"
	| myThreadIndex atomicType floatRet intRet |
	<var: #floatRet type: #double>
	<var: #intRet type: #usqLong>
	<inline: #always>
	myThreadIndex := interpreterProxy disownVM: (self disownFlagsFor: calloutState).

	calloutState floatRegisterIndex > 0 ifTrue:
		[self 
			load: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 0)) to: 'double *') at: 0)
			Flo: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 2)) to: 'double *') at: 0)
			a: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 4)) to: 'double *') at: 0)
			t: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 6)) to: 'double *') at: 0)
			R: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 8)) to: 'double *') at: 0)
			e: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 10)) to: 'double *') at: 0)
			g: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 12)) to: 'double *') at: 0)
			s: ((self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: 14)) to: 'double *') at: 0)].

	(self allocaLiesSoSetSpBeforeCall or: [self mustAlignStack]) ifTrue:
		[self setsp: calloutState argVector].

	atomicType := self atomicTypeOf: calloutState ffiRetHeader.
	(atomicType >> 1) = (FFITypeSingleFloat >> 1) ifTrue:
		[atomicType = FFITypeSingleFloat
			ifTrue:
				[floatRet := self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'float (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)]
			ifFalse: "atomicType = FFITypeDoubleFloat"
				[floatRet := self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'double (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)].

		 "undo any callee argument pops because it may confuse stack management with the alloca."
		 (self isCalleePopsConvention: calloutState callFlags) ifTrue:
			[self setsp: calloutState argVector].
		 interpreterProxy ownVM: myThreadIndex.

		 ^interpreterProxy floatObjectOf: floatRet].

	intRet := self 
				dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'usqIntptr_t (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
				with: (calloutState integerRegisters at: 0)
				with: (calloutState integerRegisters at: 1)
				with: (calloutState integerRegisters at: 2)
				with: (calloutState integerRegisters at: 3).

	"undo any callee argument pops because it may confuse stack management with the alloca."
	(self isCalleePopsConvention: calloutState callFlags) ifTrue:
		[self setsp: calloutState argVector].
	interpreterProxy ownVM: myThreadIndex.

	(calloutState ffiRetHeader anyMask: FFIFlagPointer+FFIFlagStructure) ifTrue:
		["Note: Order is important here since FFIFlagPointer + FFIFlagStructure is used to represent
		 'typedef void* VoidPointer' and VoidPointer must be returned as pointer *not* as struct."
		 (calloutState ffiRetHeader anyMask: FFIFlagPointer) ifTrue:
			[^self ffiReturnPointer: intRet ofType: (self ffiReturnType: specOnStack) in: calloutState].
		 ^self ffiReturnStruct: (self addressOf: intRet) ofType: (self ffiReturnType: specOnStack) in: calloutState].
	
	^self ffiCreateIntegralResultOop: intRet ofAtomicType: atomicType in: calloutState! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 11/3/2018 13:21'!
ffiPushDoubleFloat: value in: calloutState
	<var: #value type: #double>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>

	calloutState floatRegisterIndex < (NumFloatRegArgs - 1)
		ifTrue:
			[(calloutState floatRegisterIndex bitAnd: 1) = 1
				ifTrue: 
					[calloutState backfillFloatRegisterIndex: calloutState floatRegisterIndex.
					 calloutState floatRegisterIndex: (calloutState floatRegisterIndex + 1)].
		  	 (self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: calloutState floatRegisterIndex)) 
				 	to: 'double*')
						at: 0
						put: value.
			 calloutState floatRegisterIndex: calloutState floatRegisterIndex + 2]
		ifFalse:
			[calloutState currentArg + 8 > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 calloutState floatRegisterIndex: NumFloatRegArgs.
			 interpreterProxy storeFloatAtPointer: calloutState currentArg from: value.
			 calloutState currentArg: calloutState currentArg + 8].
	^0! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushPointer: pointer in: calloutState
	<var: #pointer type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: pointer asInteger.
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: pointer.
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 7/27/2021 16:54'!
ffiPushSignedByte: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #'signed char').
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'signed char').
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedChar: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #'signed char').
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'signed char').
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedInt: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: value.
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: value.
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedLongLong: value in: calloutState
	<var: #value type: #sqLong>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < (NumIntRegArgs - 1)
		ifTrue:
			[calloutState integerRegisterIndex: (calloutState integerRegisterIndex + 1 bitClear: 1).
			 calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #usqInt).
			 calloutState integerRegisters at: calloutState integerRegisterIndex + 1 put: (self cCoerceSimple: value >> 32 to: #usqInt).
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 2]
		ifFalse:
			[calloutState currentArg + 8 > calloutState limit ifTrue: [^FFIErrorCallFrameTooBig].
			 calloutState integerRegisterIndex: NumIntRegArgs.
			 interpreterProxy
				longAt: calloutState currentArg put: (self cCoerceSimple: value to: #usqInt);
				longAt: calloutState currentArg + self wordSize put: (self cCoerceSimple: value >> 32 to: #usqInt).
			 calloutState currentArg: calloutState currentArg + 8].
	^0! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedShort: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #'signed short').
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'signed short').
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'KenD 12/23/2018 01:49'!
ffiPushSingleFloat: value in: calloutState
	<var: #value type: #float>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState floatRegisterIndex < NumFloatRegArgs
		ifTrue: 
			[calloutState backfillFloatRegisterIndex > 0
				ifTrue: 
					[calloutState floatRegisters at: calloutState backfillFloatRegisterIndex  put: value.
					 calloutState backfillFloatRegisterIndex: 0]
				ifFalse: 
					[calloutState floatRegisters at: calloutState floatRegisterIndex  put: value.
					 calloutState floatRegisterIndex: calloutState floatRegisterIndex + 1]]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy storeSingleFloatAtPointer: calloutState currentArg from: value.
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushStructure: pointer ofSize: structSize typeSpec: argSpec ofLength: argSpecSize in: calloutState
	<var: #pointer type: #'void *'>
	<var: #argSpec type: #'sqInt *'>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	| availableRegisterSpace stackPartSize roundedSize |

	availableRegisterSpace := (NumIntRegArgs - calloutState integerRegisterIndex) * 4.
	stackPartSize := structSize.
	availableRegisterSpace > 0
		ifTrue: 
			[structSize <= availableRegisterSpace
				ifTrue:
					["all in registers"
					 stackPartSize := 0.
					 self 
						memcpy: (self cCoerceSimple: (self addressOf: (calloutState integerRegisters at: calloutState integerRegisterIndex)) to: 'void *') 
						_: pointer 
						_: structSize.
					 calloutState integerRegisterIndex: calloutState integerRegisterIndex + (structSize + 3 bitShift: -2) ]
				ifFalse:
					["If no previous co-processor candidate arg has already been pushed on the stack, then split the struct between registers and stack.
					  Otherwise push entire struct on stack."
					 calloutState currentArg = calloutState argVector
						ifTrue: 
					 		[stackPartSize := structSize - availableRegisterSpace.
					 		self 
								memcpy: (self cCoerceSimple: (self addressOf: (calloutState integerRegisters at: calloutState integerRegisterIndex)) to: 'void *') 
								_: pointer 
								_: availableRegisterSpace]
						ifFalse:
							[availableRegisterSpace := 0].
					calloutState integerRegisterIndex: NumIntRegArgs]].

	stackPartSize > 0
		ifTrue: 
			[roundedSize := stackPartSize + 3 bitClear: 3.
			 calloutState currentArg + roundedSize > calloutState limit ifTrue:
				 [^FFIErrorCallFrameTooBig].
			 self memcpy: calloutState currentArg _: (self addressOf: ((self cCoerceSimple: pointer to: 'char *') at: availableRegisterSpace)) _: stackPartSize.
			 calloutState currentArg: calloutState currentArg + roundedSize].
	^0! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedByte: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #'unsigned char').
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'unsigned char').
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedChar: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #'unsigned char').
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'unsigned char').
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedInt: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: value.
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: value.
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedLongLong: value in: calloutState
	<var: #value type: #usqLong>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < (NumIntRegArgs - 1)
		ifTrue:
			[calloutState integerRegisterIndex: (calloutState integerRegisterIndex + 1 bitClear: 1).
			 calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #usqInt).
			 calloutState integerRegisters at: calloutState integerRegisterIndex + 1 put: (self cCoerceSimple: value >> 32 to: #usqInt).
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 2]
		ifFalse:
			[calloutState currentArg + 8 > calloutState limit ifTrue: [^FFIErrorCallFrameTooBig].
			 calloutState integerRegisterIndex: NumIntRegArgs.
			 interpreterProxy
				longAt: calloutState currentArg put: (self cCoerceSimple: value to: #usqInt);
				longAt: calloutState currentArg + self wordSize put: (self cCoerceSimple: value >> 32 to: #usqInt).
			 calloutState currentArg: calloutState currentArg + 8].
	^0! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedShort: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #'unsigned short').
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'unsigned short').
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM32FFIPlugin methodsFor: 'callout support' stamp: 'nice 1/30/2020 21:50'!
ffiReturnStruct: longLongRetPtr ofType: ffiRetType in: calloutState
	<var: #longLongRetPtr type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	"Create a structure return value from an external function call.  The value has been stored in
	 alloca'ed space pointed to by the calloutState or in the return value passed by pointer."
	| retOop retClass oop |
	<inline: true>
	retClass := interpreterProxy fetchPointer: 1 ofObject: ffiRetType.
	retOop := interpreterProxy instantiateClass: retClass indexableSize: 0.
	self remapOop: retOop
		in: [oop := interpreterProxy 
					instantiateClass: interpreterProxy classByteArray 
					indexableSize: calloutState structReturnSize].
	self memcpy: (interpreterProxy firstIndexableField: oop)
		_: ((self returnStructInRegisters: calloutState)
				ifTrue: [longLongRetPtr]
				ifFalse: [calloutState limit])
		 _: calloutState structReturnSize.
	interpreterProxy storePointer: 0 ofObject: retOop withValue: oop.
	^retOop! !

!ThreadedARM32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
returnStructInRegisters: calloutState
	"Return through register if structReturnType is true"
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	^calloutState structReturnType! !

!ThreadedARM32FFIPlugin methodsFor: 'simulation support' stamp: 'KenD 12/23/2018 01:42'!
wordSize

	^ 4  "arm32"! !

ThreadedARM32FFIPlugin subclass: #ThreadedARM64FFIPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedARM64FFIPlugin commentStamp: 'KenD 12/17/2018 14:38' prior: 0!
A ThreadedARM64FFIPlugin is for the 64-bit ARM ABI.  It typically has 8 integer registers

Instance Variables
!

!ThreadedARM64FFIPlugin class methodsFor: 'translation' stamp: 'KenD 1/8/2019 17:00'!
calloutStateClass
	^ThreadedFFICalloutStateForARM64! !

!ThreadedARM64FFIPlugin class methodsFor: 'translation' stamp: 'eem 6/12/2019 18:41'!
identifyingPredefinedMacros
	^#('__ARM_ARCH_ISA_A64' '__aarch64__' '__arm64__' 'ARM64' '_M_ARM64')! !

!ThreadedARM64FFIPlugin class methodsFor: 'class initialization' stamp: 'KenD 1/23/2019 08:14'!
initialize
	super initialize.
	NumIntRegArgs    := 8.
	NumFloatRegArgs := 8! !

!ThreadedARM64FFIPlugin class methodsFor: 'translation' stamp: 'eem 8/28/2020 19:49'!
isStructType: typeName
	| space |
	^(space := typeName indexOf: Character space) > 0
	 and: [#(union struct) includes: (typeName copyFrom: 1 to: space - 1)]! !

!ThreadedARM64FFIPlugin class methodsFor: 'translation' stamp: 'KenD 12/17/2018 14:41'!
moduleName
	^'ARM64FFIPlugin'! !

!ThreadedARM64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
canReturnInRegistersStructOfSize: returnStructSize
	"Answer if a struct result of a given size is able to be returned in registers.
	NB: this is a predicate!! #canReturnInRegistersStructOfSize: does NOT return a struct in anything!!"
	<inline: #always>
	^returnStructSize <= (2 * self wordSize)! !

!ThreadedARM64FFIPlugin methodsFor: 'translation support' stamp: 'eem 8/29/2020 18:01'!
d
	"Hack for floatRet in ffiCalloutTo:SpecOnStack:in:"
	<returnTypeC: 'struct { double doubles[4]; }'>
	<inline: #always> "to make it disappear..."
	^0! !

!ThreadedARM64FFIPlugin methodsFor: 'translation support' stamp: 'eem 8/29/2020 18:01'!
doubles
	"Hack for floatRet in ffiCalloutTo:SpecOnStack:in:"
	<returnTypeC: #'double *'>
	<inline: #always> "to make it disappear..."
	^0! !

!ThreadedARM64FFIPlugin methodsFor: 'callout support' stamp: 'eem 7/27/2021 16:46'!
encodeStructReturnTypeIn: calloutState
	"Set the return type to true if returning the struct via register"
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>

	calloutState structReturnType: (self canReturnInRegistersStructOfSize: calloutState structReturnSize)! !

!ThreadedARM64FFIPlugin methodsFor: 'translation support' stamp: 'eem 8/29/2020 18:01'!
f
	"Hack for floatRet in ffiCalloutTo:SpecOnStack:in:"
	<returnTypeC: 'struct { float floats[8]; }'>
	<inline: #always> "to make it disappear..."
	^0! !

!ThreadedARM64FFIPlugin methodsFor: 'callout support' stamp: 'eem 8/29/2020 13:28'!
ffiCall: externalFunction ArgArrayOrNil: argArrayOrNil NumArgs: nArgs
	"Generic callout. Does the actual work.  If argArrayOrNil is nil it takes args from the stack
	 and the spec from the method.  If argArrayOrNil is not nil takes args from argArrayOrNil
	 and the spec from the receiver."
	| flags argTypeArray address argType oop argSpec argClass err theCalloutState calloutState requiredStackSize stackSize allocation result primNumArgs |
	<inline: false>
	<var: #theCalloutState type: #'CalloutState'>
	<var: #calloutState type: #'CalloutState *'>
	<var: #allocation type: #'char *'>

	primNumArgs := interpreterProxy methodArgumentCount.
	(interpreterProxy is: externalFunction KindOfClass: interpreterProxy classExternalFunction) ifFalse:
		[^self ffiFail: FFIErrorNotFunction].
	"Load and check the values in the externalFunction before we call out"
	flags := interpreterProxy fetchInteger: ExternalFunctionFlagsIndex ofObject: externalFunction.
	interpreterProxy failed ifTrue:
		[^self ffiFail: FFIErrorBadArgs].

	"This must come early for compatibility with the old FFIPlugin.  Image-level code
	 may assume the function pointer is loaded eagerly.  Thanks to Nicolas Cellier."
	address := self ffiLoadCalloutAddress: externalFunction.
	interpreterProxy failed ifTrue:
		[^0 "error code already set by ffiLoadCalloutAddress:"].
	
	argTypeArray := interpreterProxy fetchPointer: ExternalFunctionArgTypesIndex ofObject: externalFunction.
	"must be array of arg types"
	((interpreterProxy isArray: argTypeArray)
	and: [(interpreterProxy slotSizeOf: argTypeArray) = (nArgs + 1)]) ifFalse:
		[^self ffiFail: FFIErrorBadArgs].
	"check if the calling convention is supported"
	self cppIf: COGMTVM
		ifTrue:
			[(self ffiSupportsCallingConvention: (flags bitAnd: FFICallTypesMask)) ifFalse:
				[^self ffiFail: FFIErrorCallType]]
		ifFalse: "not masking causes threaded calls to fail, which is as they should if the plugin is not threaded."
			[(self ffiSupportsCallingConvention: flags) ifFalse:
				[^self ffiFail: FFIErrorCallType]].
		
	requiredStackSize := self externalFunctionHasStackSizeSlot
							ifTrue: [interpreterProxy
										fetchInteger: ExternalFunctionStackSizeIndex
										ofObject: externalFunction]
							ifFalse: [-1].
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: (argArrayOrNil isNil
												ifTrue: [PrimErrBadMethod]
												ifFalse: [PrimErrBadReceiver])].
	stackSize := requiredStackSize < 0 ifTrue: [DefaultMaxStackSize] ifFalse: [requiredStackSize].
	self cCode: [] inSmalltalk: [theCalloutState := self class calloutStateClass new].
	calloutState := self addressOf: theCalloutState.
	self cCode: [self memset: calloutState _: 0 _: (self sizeof: #CalloutState)].
	calloutState callFlags: flags.
	"Fetch return type and args"
	argType := interpreterProxy fetchPointer: 0 ofObject: argTypeArray.
	argSpec := interpreterProxy fetchPointer: 0 ofObject: argType.
	argClass := interpreterProxy fetchPointer: 1 ofObject: argType.
	(err := self ffiCheckReturn: argSpec With: argClass in: calloutState) ~= 0 ifTrue:
		[^self ffiFail: err]. "cannot return"
	"alloca the outgoing stack frame, leaving room for marshalling args, and including space for the return struct, if any.
	Additional space reserved for saving register args like mandated by Win64 X64 or PPC ABI, will be managed by the call itself"
	allocation := self alloca: stackSize + calloutState structReturnSize + self cStackAlignment.
	self mustAlignStack ifTrue:
		[allocation := self cCoerce: (allocation asUnsignedIntegerPtr bitClear: self cStackAlignment - 1) to: #'char *'].
	calloutState
		argVector: allocation;
		currentArg: allocation;
		limit: allocation + stackSize.
	1 to: nArgs do:
		[:i|
		argType := interpreterProxy fetchPointer: i ofObject: argTypeArray.
		argSpec := interpreterProxy fetchPointer: 0 ofObject: argType.
		argClass := interpreterProxy fetchPointer: 1 ofObject: argType.
		oop := argArrayOrNil isNil
				ifTrue: [interpreterProxy stackValue: nArgs - i]
				ifFalse: [interpreterProxy fetchPointer: i - 1 ofObject: argArrayOrNil].
		err := self ffiArgument: oop Spec: argSpec Class: argClass in: calloutState.
		err ~= 0 ifTrue:
			[self cleanupCalloutState: calloutState.
			 self cppIf: COGMTVM ifTrue:
			 [err = PrimErrObjectMayMove negated ifTrue:
				[^PrimErrObjectMayMove]]. "N.B. Do not fail if object may move because caller will GC and retry."
			 ^self ffiFail: err]]. "coercion failed or out of stack space"
	"Failures must be reported back from ffiArgument:Spec:Class:in:.
	 Should not fail from here on in."
	self assert: interpreterProxy failed not.
	self ffiLogCallout: externalFunction.
	(requiredStackSize < 0
	 and: [self externalFunctionHasStackSizeSlot]) ifTrue:
		[stackSize := calloutState currentArg - calloutState argVector.
		 interpreterProxy storeInteger: ExternalFunctionStackSizeIndex ofObject: externalFunction withValue: stackSize].
	"Go out and call this guy"
	result := self ffiCalloutTo: address SpecOnStack: argArrayOrNil notNil in: calloutState.
	self cleanupCalloutState: calloutState.
	"Can not safely use argumentCount (via e.g. methodReturnValue:) since it may have been changed by a callback."
	interpreterProxy pop: primNumArgs + 1 thenPush: result. 
	^result! !

!ThreadedARM64FFIPlugin methodsFor: 'callout support' stamp: 'eem 8/29/2020 15:29'!
ffiCalloutTo: procAddr SpecOnStack: specOnStack in: calloutState
	<var: #procAddr type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	<var: #loadFloatRegs declareC: 'extern void loadFloatRegs(double, double, double, double, double, double, double, double)'>
	"Go out, call this guy and create the return value.  This *must* be inlined because of
	 the alloca of the outgoing stack frame in ffiCall:WithFlags:NumArgs:Args:AndTypes:"
	| myThreadIndex atomicType floatRet intRet x1Ret specSize |
	<var: #floatRet type: 'union { struct { float floats[8]; } f; struct dprr { double doubles[4]; } d; }'>
	<var: #intRet type: #usqLong>
	<var: #x1Ret type: #usqLong>
	<inline: #always>
	self cCode: [] inSmalltalk: [floatRet := ByteArray new: 32]. "Just a hack to placate the Smalltalk compiler; this shoudl be a proper struct..."
	myThreadIndex := interpreterProxy disownVM: (self disownFlagsFor: calloutState).
	
	calloutState floatRegisterIndex > 0 ifTrue:
		[self loadFloatRegs:
			   (calloutState floatRegisters at: 0)
			_: (calloutState floatRegisters at: 1)
			_: (calloutState floatRegisters at: 2)
			_: (calloutState floatRegisters at: 3)
			_: (calloutState floatRegisters at: 4)
			_: (calloutState floatRegisters at: 5)
			_: (calloutState floatRegisters at: 6)
			_: (calloutState floatRegisters at: 7)].

	(self allocaLiesSoSetSpBeforeCall or: [self mustAlignStack]) ifTrue:
		[self setsp: calloutState argVector].

	atomicType := self atomicTypeOf: calloutState ffiRetHeader.
	((atomicType >> 1) = (FFITypeSingleFloat >> 1)
	or: [(calloutState ffiRetHeader bitAnd: FFIFlagPointer+FFIFlagStructure) = FFIFlagStructure
		and: [self structIsHomogenousFloatArrayOfSize: (calloutState ffiRetHeader bitAnd: FFIStructSizeMask)
				typeSpec: (self cCoerce: (interpreterProxy firstIndexableField: calloutState ffiRetSpec) to: #'unsigned int *')
				ofLength: (specSize := interpreterProxy byteSizeOf: calloutState ffiRetSpec) / (self sizeof: #'unsigned int')]]) ifTrue:
		[floatRet d: (self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'struct dprr (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)')
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)
					with: (calloutState integerRegisters at: 4)
					with: (calloutState integerRegisters at: 5)
					with: (calloutState integerRegisters at: 6)
					with: (calloutState integerRegisters at: 7)).

		 "undo any callee argument pops because it may confuse stack management with the alloca."
		 (self isCalleePopsConvention: calloutState callFlags) ifTrue:
			[self setsp: calloutState argVector].
		 interpreterProxy ownVM: myThreadIndex.

		 atomicType = FFITypeDoubleFloat ifTrue:
			[^interpreterProxy floatObjectOf: (floatRet d doubles at: 0)].
		 atomicType = FFITypeSingleFloat ifTrue:
			[^interpreterProxy floatObjectOf: (floatRet f floats at: 0)].
		"If the struct is a vector of floats then move float[2] to float[1], float[4] to float[2] and float[6] to float[3],
		 to pack the float data in the double fields.  We can tell if the struct is composed of floats if its size is less
		 than the spec size, since the spec size is (1 + n fields) * 4 bytes, and the struct size is n fields * 4 bytes
		 for floats and n fields * 8 bytes for doubles.  We can't access the spec post call because it may have moved."
		specSize > calloutState structReturnSize ifTrue:
			[floatRet f floats at: 1 put: (floatRet f floats at: 2).
			 floatRet f floats at: 2 put: (floatRet f floats at: 4).
			 floatRet f floats at: 3 put: (floatRet f floats at: 6)].
		^self ffiReturnStruct: (self addressOf: floatRet) ofType: (self ffiReturnType: specOnStack) in: calloutState].

	"If struct address used for return value, call is special"
	(self mustReturnStructOnStack: calloutState structReturnSize) 
	ifTrue: [
		intRet := 0.
		self setReturnRegister: (self cCoerceSimple: calloutState limit to: #sqLong) "stack alloca'd struct"
			 andCall: (self cCoerceSimple: procAddr to: #sqLong)
			 withArgsArray: (self cCoerceSimple: (self addressOf: calloutState integerRegisters) to: #sqLong).
	] ifFalse: [
		intRet := self 
				dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'usqIntptr_t (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)')
				with: (calloutState integerRegisters at: 0)
				with: (calloutState integerRegisters at: 1)
				with: (calloutState integerRegisters at: 2)
				with: (calloutState integerRegisters at: 3)
				with: (calloutState integerRegisters at: 4)
				with: (calloutState integerRegisters at: 5)
				with: (calloutState integerRegisters at: 6)
				with: (calloutState integerRegisters at: 7).
	
	 x1Ret := self getX1register. "Capture x1 immediately. No problem if unused"
	].
	"If struct returned in registers, 
	 place register values into calloutState integerRegisters"
	(calloutState structReturnSize > 0
	 and: [self returnStructInRegisters: calloutState]) ifTrue: 
		["Only 2 regs used in ARMv8/Aarch64 current"
		 calloutState integerRegisters at: 0 put: intRet. "X0"
		 calloutState integerRegisters at: 1 put: x1Ret]. "X1"

	"undo any callee argument pops because it may confuse stack management with the alloca."
	(self isCalleePopsConvention: calloutState callFlags) ifTrue:
		[self setsp: calloutState argVector].
	interpreterProxy ownVM: myThreadIndex.

	(calloutState ffiRetHeader anyMask: FFIFlagPointer+FFIFlagStructure) ifTrue:
		["Note: Order is important here since FFIFlagPointer + FFIFlagStructure is used to represent
		 'typedef void* VoidPointer' and VoidPointer must be returned as pointer *not* as struct."
		 (calloutState ffiRetHeader anyMask: FFIFlagPointer) ifTrue:
			[^self ffiReturnPointer: intRet ofType: (self ffiReturnType: specOnStack) in: calloutState].
		 ^self ffiReturnStruct: (((self returnStructInRegisters: calloutState)
								ifTrue: [self addressOf: calloutState integerRegisters]
								ifFalse: [calloutState limit]))
				ofType: (self ffiReturnType: specOnStack)
				in: calloutState].
	
	^self ffiCreateIntegralResultOop: intRet ofAtomicType: atomicType in: calloutState! !

!ThreadedARM64FFIPlugin methodsFor: 'marshalling' stamp: 'KenD 1/23/2019 08:17'!
ffiPushDoubleFloat: value in: calloutState
	<var: #value type: #double>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>

	calloutState floatRegisterIndex < NumFloatRegArgs
		ifTrue:
			[calloutState floatRegisters
				at: calloutState floatRegisterIndex
				put: value.
			 calloutState floatRegisterIndex: calloutState floatRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 calloutState floatRegisterIndex: NumFloatRegArgs.
			 interpreterProxy storeFloatAtPointer: calloutState currentArg from: value.
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedInt: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters 
				at: calloutState integerRegisterIndex 
				put: (self cCoerceSimple: value to: #sqLong).
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: value.
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedLongLong: value in: calloutState
	<var: #value type: #sqLong>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters 
				at: calloutState integerRegisterIndex 
				put: (self cCoerceSimple: value to: #sqLong).
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue: [^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: value.
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM64FFIPlugin methodsFor: 'marshalling' stamp: 'KenD 1/23/2019 08:19'!
ffiPushSingleFloat: value in: calloutState
	<var: #value type: #float>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState floatRegisterIndex < NumFloatRegArgs
		ifTrue: "Note: this is a 'memcopy', so size is preserved. Casting to #double changes the size"
			[(self cCoerceSimple: 
				(self addressOf: (calloutState floatRegisters at: calloutState floatRegisterIndex)) 
					 to: 'float*')
				at: 0
				put: value.
			 calloutState floatRegisterIndex: calloutState floatRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy storeSingleFloatAtPointer: calloutState currentArg from: value.
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushStructure: pointer ofSize: structSize typeSpec: argSpec ofLength: argSpecSize in: calloutState
	<var: #pointer type: #'void *'>
	<var: #argSpec type: #'unsigned int *'>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	| availableRegisterSpace stackPartSize roundedSize |
	"See IHI0055B_aapcs64.pdf sections 4.3.5 & 5.4.2 Stage C; we don't yet support HVA's"
	(self structIsHomogenousFloatArrayOfSize: structSize typeSpec: argSpec ofLength: argSpecSize)
		ifTrue:
			[availableRegisterSpace := (NumFloatRegArgs - calloutState floatRegisterIndex) * self wordSize.
			 structSize <= availableRegisterSpace ifTrue: "Stage C, step C.2, all in floating-point registers (!!!!)"
				[self 
					memcpy: (self cCoerceSimple: (self addressOf: (calloutState floatRegisters at: calloutState floatRegisterIndex)) to: #'void *') 
					_: pointer 
					_: structSize.
					"Round structSize up and divide by 8 ( NB: _not_ 4 !!)"
				 calloutState floatRegisterIndex: calloutState floatRegisterIndex + (structSize + 7 bitShift: -3).
				 ^0].
			 "Stage C, step C.3"
			 availableRegisterSpace := 0.
			 calloutState floatRegisterIndex: 8]

		ifFalse:
			[availableRegisterSpace := (NumIntRegArgs - calloutState integerRegisterIndex) * self wordSize].
	stackPartSize := structSize.
	availableRegisterSpace > 0 ifTrue: 
		[structSize <= availableRegisterSpace ifTrue:"all in integer registers"
			[self 
				memcpy: (self cCoerceSimple: (self addressOf: (calloutState integerRegisters at: calloutState integerRegisterIndex)) to: #'void *') 
				_: pointer 
				_: structSize.
				"Round structSize up and divide by 8 ( NB: _not_ 4 !!)"
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + (structSize + 7 bitShift: -3).
			 ^0].
		 "If no previous co-processor candidate arg has already been pushed on the stack, then split the struct between registers and stack.
		  Otherwise push entire struct on stack."
		 calloutState currentArg = calloutState argVector
			ifTrue: 
		 		[stackPartSize := structSize - availableRegisterSpace.
		 		self 
					memcpy: (self cCoerceSimple: (self addressOf: (calloutState integerRegisters at: calloutState integerRegisterIndex)) to: 'void *') 
					_: pointer 
					_: availableRegisterSpace]
			ifFalse:
				[availableRegisterSpace := 0].
		"Stage C, step C.11"
		calloutState integerRegisterIndex: NumIntRegArgs].

	stackPartSize > 0 ifTrue: 
		[roundedSize := stackPartSize + 3 bitClear: 3.
		 calloutState currentArg + roundedSize > calloutState limit ifTrue:
			 [^FFIErrorCallFrameTooBig].
		 self memcpy: calloutState currentArg _: (self addressOf: ((self cCoerceSimple: pointer to: #'char *') at: availableRegisterSpace)) _: stackPartSize.
		 calloutState currentArg: calloutState currentArg + roundedSize].
	^0! !

!ThreadedARM64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedInt: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters 
				at: calloutState integerRegisterIndex 
				put:  (self cCoerceSimple: value to: #usqLong).
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: value.
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedLongLong: value in: calloutState
	<var: #value type: #usqLong>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters 
				at: calloutState integerRegisterIndex 
				put:(self cCoerceSimple: value to: #usqLong).
			 calloutState integerRegisterIndex: calloutState integerRegisterIndex + 1]
		ifFalse:
			[calloutState currentArg + self wordSize > calloutState limit ifTrue: [^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: value.
			 calloutState currentArg: calloutState currentArg + self wordSize].
	^0! !

!ThreadedARM64FFIPlugin methodsFor: 'callout support' stamp: 'eem 8/29/2020 14:30'!
ffiReturnStruct: longLongRetPtr ofType: ffiRetType in: calloutState
	<var: #longLongRetPtr type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	"Create a structure return value from an external function call.  The value has been stored in
	 alloca'ed space pointed to by the calloutState or in the integer registers."
	| retOop retClass oop |
	<inline: true>
	retClass := interpreterProxy fetchPointer: 1 ofObject: ffiRetType.
	retOop := interpreterProxy instantiateClass: retClass indexableSize: 0.
	self remapOop: retOop
		in: [oop := interpreterProxy 
					instantiateClass: interpreterProxy classByteArray 
					indexableSize: calloutState structReturnSize].
	self memcpy: (interpreterProxy firstIndexableField: oop)
		_: longLongRetPtr
		 _: calloutState structReturnSize.
	interpreterProxy storePointer: 0 ofObject: retOop withValue: oop.
	^retOop! !

!ThreadedARM64FFIPlugin methodsFor: 'translation support' stamp: 'eem 8/29/2020 18:01'!
floats
	"Hack for floatRet in ffiCalloutTo:SpecOnStack:in:"
	<returnTypeC: #'float *'>
	<inline: #always> "to make it disappear..."
	^0! !

!ThreadedARM64FFIPlugin methodsFor: 'callout support' stamp: 'KenD 1/30/2019 16:16'!
getX1register

	<inline: true>
	<var: #returnX1value declareC: 'extern sqLong returnX1value()'>

	^self returnX1value! !

!ThreadedARM64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
mustReturnStructOnStack: returnStructSize
	"Answer if a struct result of a given size is unable to be returned in registers."
	<inline: #always>
	^returnStructSize > (2 * self wordSize)! !

!ThreadedARM64FFIPlugin methodsFor: 'marshalling' stamp: 'KenD 1/25/2019 20:31'!
nonRegisterStructReturnIsViaImplicitFirstArgument
	"Answer if a struct returned in memory is returned to the
	 referent of a pointer passed as an implciit first argument.
	 It almost always is.  Subclasses can override if not."
	^false! !

!ThreadedARM64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
returnStructInRegisters: calloutState
	"Return through register if structReturnType is true"
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	^calloutState structReturnType! !

!ThreadedARM64FFIPlugin methodsFor: 'callout support' stamp: 'KenD 2/6/2019 08:37'!
setReturnRegister: structAddr andCall: procAddr withArgsArray: arrayAddr

	<inline: true>
	<var: #structAddr type: #'sqLong'>
	<var: #procAddr  type: #'sqLong'>
	<var: #arrayAddr type: #'sqLong'>
	<var: #callAndReturnWithStructAddr declareC: 'extern void callAndReturnWithStructAddr(sqLong structAddr,sqLong procAddr,sqLong arrayAddr)'>

	self callAndReturnWithStructAddr: structAddr _: procAddr _: arrayAddr! !

!ThreadedARM64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 8/29/2020 15:00'!
structIsHomogenousFloatArrayOfSize: structSize typeSpec: argSpec ofLength: argSpecSize
	<var: #argSpec type: #'unsigned int *'>
	| firstField typeOfFirstField |
	(structSize <= (4 * (self sizeof: #double))
	 and: [argSpecSize <= 5]) "header plus up to four fields" ifFalse:
		[^false].
	typeOfFirstField := self atomicTypeOf: (firstField := argSpec at: 1).
	(typeOfFirstField ~= FFITypeSingleFloat and: [typeOfFirstField ~= FFITypeDoubleFloat]) ifTrue:
		[^false].
	2 to: argSpecSize - 1 do:
		[:idx|
		firstField ~= (argSpec at: idx) ifTrue:
			[^false]].
	^true! !

!ThreadedARM64FFIPlugin methodsFor: 'simulation support' stamp: 'KenD 12/23/2018 01:43'!
wordSize

	^ 8  "arm64/aarch64"! !

!ThreadedFFIPlugin class methodsFor: 'translation' stamp: 'eem 5/31/2020 15:41'!
ancilliaryClasses
	^{ self calloutStateClass. AlignmentStructType }! !

!ThreadedFFIPlugin class methodsFor: 'translation' stamp: 'eem 11/7/2009 09:51'!
calloutStateClass
	^ThreadedFFICalloutState! !

!ThreadedFFIPlugin class methodsFor: 'translation' stamp: 'eem 10/25/2018 12:14'!
declareCVarsIn: cg
	super declareCVarsIn: cg.
	cg removeConstant: #COGMTVM. "this should be defined at compile time"
	cg removeVariable: #allocationMap "this is simulation only"! !

!ThreadedFFIPlugin class methodsFor: 'translation' stamp: 'eem 2/16/2016 20:41'!
excludingPredefinedMacros
	"Answer the predefined macros that disqualify the platforms a subclass handles, if any.
	 These are anded together and with includingPredefinedMacros, whereas
	 identifyingPredefinedMacros are ored together.
	 This can be used to differentiate e.g. x64 Sys V from x64 Win64."
	^nil! !

!ThreadedFFIPlugin class methodsFor: 'translation' stamp: 'eem 2/23/2016 09:29'!
generateCodeStringForPrimitives
	"Output a skeletal SqueakFFIPrims.c that includes the relevant FooFFIPlugin.c for
	 each subclass based on the identifyingPredefinedMacros the subclass defines."
	 
	^String streamContents:
		[:s|
		 s nextPutAll: '/* Automatically generated by\	' withCRs.
		 s nextPutAll: (CCodeGenerator monticelloDescriptionFor: self).
		 s cr; nextPutAll: ' */'.
		 s cr; cr; nextPut: $#.
		 (self allSubclasses sort: [:a :b| a name < b name]) do:
			[:class |
			class identifyingPredefinedMacros ifNotNil:
				[:predefinedMacros|
				 s nextPutAll: 'if '.
				class includingPredefinedMacros ifNotNil:
					[:includingMacros|
					 includingMacros do:
						[:predefinedMacro| s nextPutAll: 'defined('; nextPutAll: predefinedMacro; nextPutAll: ') && '].
					 s nextPut: $(].
				class excludingPredefinedMacros ifNotNil:
					[:excludingMacros|
					 excludingMacros do:
						[:predefinedMacro| s nextPutAll: '!!defined('; nextPutAll: predefinedMacro; nextPutAll: ') && '].
					 s nextPut: $(].
				 predefinedMacros
					do: [:predefinedMacro| s nextPutAll: 'defined('; nextPutAll: predefinedMacro; nextPut: $)]
					separatedBy: [s nextPutAll: ' || '].
				class excludingPredefinedMacros ifNotNil:
					[s nextPut: $)].
				class includingPredefinedMacros ifNotNil:
					[s nextPut: $)].
				 s cr; cr; nextPutAll: '# define '; nextPutAll: class moduleName; nextPutAll: '_exports ';
												nextPutAll: self moduleName; nextPutAll: '_exports'.
				 s cr; nextPutAll: '# include "'; nextPutAll: class moduleName; nextPutAll: '.c"'.
				 s cr; cr; nextPutAll: '#el']].
		 s nextPutAll: 'se'.
		 #(	'As yet no FFI implementation appears to exist for your platform.'
			'Consider implementing it, starting by adding a subclass of ThreadedFFIPlugin.') do:
			[:msg| s cr; nextPutAll: '# error '; nextPutAll: msg].
		 s cr; nextPutAll: '#endif'; cr]! !

!ThreadedFFIPlugin class methodsFor: 'translation' stamp: 'eem 2/16/2016 20:42'!
identifyingPredefinedMacros
	"Answer the predefined macros that identify the platforms a subclass handles, if any.
	 These are ored together.
	 If the subclass isn't yet ready for production (a work in progress) simply answer nil."
	^nil! !

!ThreadedFFIPlugin class methodsFor: 'translation' stamp: 'eem 2/16/2016 20:41'!
includingPredefinedMacros
	"Answer the predefined macros that qualify the platforms a subclass handles, if any.
	 These are anded together and with excludingPredefinedMacros, whereas
	 identifyingPredefinedMacros are ored together.
	 This can be used to differentiate e.g. x64 Sys V from x64 Win64."
	^nil! !

!ThreadedFFIPlugin class methodsFor: 'class initialization' stamp: 'eem 6/15/2020 16:36'!
initialize
	"c.f. ExternalFunction allInstVarNames
		old: #('handle' 'flags' 'argTypes')
		new: #('handle' 'flags' 'argTypes' 'stackSize')"
	ExternalFunctionAddressIndex := 0.
	ExternalFunctionFlagsIndex := 1.
	ExternalFunctionArgTypesIndex := 2.
	ExternalFunctionStackSizeIndex := 3.

	"c.f. e.g. CoInterpreter class initializeMiscConstants"
	MaxNumArgs := 15.

	DefaultMaxStackSize := 1024 * 16.

	PluginVersionInfo := CCodeGenerator shortMonticelloDescriptionForClass: self.
	PluginVersionInfo := PluginVersionInfo allButFirst: (PluginVersionInfo indexOf: Character space) - 1! !

!ThreadedFFIPlugin class methodsFor: 'translation' stamp: 'eem 9/18/2020 11:01'!
internalModuleName
	"To make the inclusion of the platform-specific plugin to work it can't use
	 its moduleName in the exports but must use the proper moduleName."

	^ThreadedFFIPlugin moduleName! !

!ThreadedFFIPlugin class methodsFor: 'accessing' stamp: 'eem 11/9/2009 14:52'!
maxNumArgs
	^MaxNumArgs! !

!ThreadedFFIPlugin class methodsFor: 'translation' stamp: 'eem 4/28/2014 09:56'!
moduleName
	"IMPORTANT: IF YOU CHANGE THE NAME OF THIS PLUGIN YOU MUST CHANGE
		Interpreter>>primitiveCalloutToFFI *and* ThreadedFFIPlugin>>getModuleName
	TO REFLECT THE CHANGE."
	^#('SqueakFFIPrims' 'FFIPlugin') first! !

!ThreadedFFIPlugin class methodsFor: 'translation' stamp: 'eem 8/29/2020 10:59'!
preambleCCode
	"For a source of builtin defines grep for builtin_define in a gcc release config directory.
	 See platforms/Cross/vm/sqCogStackAlignment.h for per-platform definitions for
	 STACK_ALIGN_BYTES MUST_ALIGN_STACK, getsp, et al."
	^'
#include "sqAssert.h" /* for assert */
#define ThreadedFFIPlugin 1 /* to filter-out unwanted declarations from sqFFI.h */
#include "sqFFI.h" /* for logging and surface functions */
#include "sqCogStackAlignment.h" /* for STACK_ALIGN_BYTES and getsp() */

#ifdef _MSC_VER
# define alloca _alloca
#endif
#if !!defined(setsp) && defined(__GNUC__)
# if defined(_X86_) || defined(i386) || defined(__i386) || defined(__i386__)
#	define setsp(spval) asm volatile ("movl %0,%%esp" : : "m"(spval))
# elif defined(__amd64__) || defined(__x86_64__) ||  defined(__amd64) || defined(__x86_64)
#	define setsp(spval) asm volatile ("movq %0,%%rsp" : : "m"(spval))
# elif defined(__arm64__) || defined(__aarch64__) || defined(ARM64)
        /* https://gcc.gnu.org/onlinedocs/gcc/Extended-Asm.html#Extended-Asm
         * http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.den0024a/index.html
         */
#	define setsp(spval) asm volatile ("mov sp, %0"  : : "r"(spval))
# elif defined(__arm__)
#	define setsp(spval) asm volatile ("ldr %%sp, %0" : : "m"(spval))
# endif
#endif
#if !!defined(getsp)
# define getsp() 0
#endif 
#if !!defined(setsp)
# define setsp(ignored) 0
#endif 

#if !!defined(STACK_ALIGN_BYTES)
#  define STACK_ALIGN_BYTES 0
#endif /* !!defined(STACK_ALIGN_BYTES) */

/* For ABI that require stack alignment greater than natural word size */
#define MUST_ALIGN_STACK (STACK_ALIGN_BYTES > sizeof(void*))

#if defined(_X86_) || defined(i386) || defined(__i386) || defined(__i386__)
/* Both Mac OS X x86 and Win32 x86 return structs of a power of two in size
 * less than or equal to eight bytes in length in registers. Linux never does so.
 */
# if __linux__
#	define WIN32_X86_STRUCT_RETURN 0
# else
#	define WIN32_X86_STRUCT_RETURN 1
# endif
# if _WIN32
#	define PLATFORM_API_USES_CALLEE_POPS_CONVENTION 1
# endif
#elif defined(__amd64__) || defined(__x86_64__) ||  defined(__amd64) || defined(__x86_64)
# if _WIN32 | _WIN64
#	define PLATFORM_API_USES_CALLEE_POPS_CONVENTION 1
# endif
#endif /* defined(_X86_) || defined(i386) || defined(__i386) || defined(__i386__) */

#if !!defined(ALLOCA_LIES_SO_SETSP_BEFORE_CALL)
# if defined(__MINGW32__) && !!defined(__clang__) && (__GNUC__ >= 3) && (defined(_X86_) || defined(i386) || defined(__i386) || defined(__i386__))
    /*
     * cygwin -mno-cygwin (MinGW) gcc 3.4.x''s alloca is a library routine that answers
     * %esp + xx, so the outgoing stack is offset by one or more word if uncorrected.
     * Grab the actual stack pointer to correct.
     */
#	define ALLOCA_LIES_SO_SETSP_BEFORE_CALL 1
# else
#	define ALLOCA_LIES_SO_SETSP_BEFORE_CALL 0
# endif
#endif /* !!defined(ALLOCA_LIES_SO_SETSP_BEFORE_CALL) */

#if !!defined(PLATFORM_API_USES_CALLEE_POPS_CONVENTION)
# define PLATFORM_API_USES_CALLEE_POPS_CONVENTION 0
#endif

/* The dispatchOn:in:with:with: generates an unwanted call on error.  Just squash it. */
#define error(foo) 0

/* sanitize */
#ifdef SQUEAK_BUILTIN_PLUGIN
# define EXTERN 
#else
# define EXTERN extern
#endif
'! !

!ThreadedFFIPlugin class methodsFor: 'simulation' stamp: 'eem 5/31/2020 15:28'!
primitiveAlignmentOf: atomicTypeCode
	"Answer the platform C compiler's alignment in a structure of an element of the
	 C type corresponding to atomicTypeCode, where atomicTypeCode is in the range
	 FFITypeUnsignedByte (2) to FFITypeDoubleFloat (13), or is FFIFlagStructure (65536)."
	<primitive: #primitiveStructureElementAlignment module: #SqueakFFIPrims>
	self primitiveFailed

	"(FFITypeUnsignedByte to: FFITypeDoubleFloat), {FFIFlagStructure} collect: [:n| self primitiveAlignmentOf: n]"! !

!ThreadedFFIPlugin class methodsFor: 'simulation' stamp: 'eem 7/16/2020 18:16'!
primitiveCDataModel
	"Answer the name for the platform C compiler's data model, ILP32, LP64, et al."
	<primitive: #primitiveCDataModel module: #SqueakFFIPrims error: ec>
	self primitiveFailed

	"self primitiveCDataModel"! !

!ThreadedFFIPlugin class methodsFor: 'simulation' stamp: 'eem 7/16/2020 18:26'!
primitiveSizesInto: aByteArrayOfSizeNine
	"Answer the platform C compiler's sizes for char, short, int, long, long long, wchar_t, float, double, void *."
	<primitive: #primitiveCDataModel module: #SqueakFFIPrims error: ec>
	self primitiveFailed

	"(#(char short int long #'long long' wchar_t float double #'void *')
		with: (self primitiveSizesInto: (ByteArray new: 9))
		collect: [:n :s| {n. s}]) flatten"! !

!ThreadedFFIPlugin class methodsFor: 'translation' stamp: 'dtl 5/29/2014 19:42'!
translateInDirectory: directory doInlining: inlineFlag
	"As a hack to allow ThreadedFFIPlugin to generate a skeletal SqueakFFIPrims.c that
	 simply includes the fleshed-out SqueakIA32FFIPrims.c, SqueakARMFFIPrims.c et al,
	 intercept for ThreadedFFIPlugin and output generateCodeStringForPrimitives."
	| ffiPluginRootClass |
	ffiPluginRootClass := thisContext method methodClass soleInstance.
	^self == ffiPluginRootClass
		ifTrue:
			[self
				storeString: self generateCodeStringForPrimitives
				onFileNamed: (directory fullNameFor: self moduleName, '.c').
			#()]
		ifFalse:
			[super
				translateInDirectory: (directory containingDirectory directoryNamed: ffiPluginRootClass moduleName) assureExistence
				doInlining: inlineFlag]! !

!ThreadedFFIPlugin methodsFor: 'marshalling-struct' stamp: 'eem 8/28/2020 13:10'!
alignmentOfStructSpec: specs OfLength: specSize StartingAt: indexPtr
	"Answer with the alignment requirement for a structure/union.
	Note that indexPtr is a pointer so as to be changed on return.
	On input, the index points to the structure header (the one with FFIFlagStructure + structSize).
	On output, the index points the the structure trailer (the FFIFlagStructure)."
	| spec byteAlignment thisAlignment |
	<var: #specs type: #'unsigned int *'>
	<var: #indexPtr type: #'sqInt *'>
	<inline: false>
	spec := specs at: (indexPtr at: 0).
	self assert: (spec bitAnd: FFIFlagPointer + FFIFlagAtomic + FFIFlagStructure) = FFIFlagStructure.
	byteAlignment := 1.
	[indexPtr at: 0 put: (indexPtr at: 0) + 1.
	(indexPtr at: 0) < specSize]
		whileTrue:
			[spec := specs at: (indexPtr at: 0).
			spec = FFIFlagStructure
				ifTrue: [^byteAlignment].
			thisAlignment := (spec anyMask: FFIFlagPointer)
				ifTrue: [BytesPerWord]
				ifFalse: [(spec anyMask: FFIFlagStructure)
					ifTrue: [self alignmentOfStructSpec: specs OfLength: specSize StartingAt: indexPtr]
					ifFalse: [spec bitAnd: FFIStructSizeMask]].
			byteAlignment := byteAlignment max: thisAlignment].
	self assert: false. "should not reach here - because only ever called for sub-struct"
	^byteAlignment! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'nice 12/13/2016 22:08'!
allocaLiesSoSetSpBeforeCall
	"At least one alloca implementation does not answer the actual top of stack.
	 If so we need to reset the actual stack pointer just before the call.
	Answer whether this is necessary."
	<cmacro: '() ALLOCA_LIES_SO_SETSP_BEFORE_CALL'>
	^false! !

!ThreadedFFIPlugin methodsFor: 'primitive support' stamp: 'eem 11/7/2009 17:07'!
atomicTypeOf: typeSpec
	^(typeSpec bitAnd: FFIAtomicTypeMask) >> FFIAtomicTypeShift! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 11/3/2018 13:27'!
cStackAlignment
	"Many ABIs mandate a particular stack alignment greater than the natural word
	 size. If so, this macro will answer that alignment.  If not, this macro will answer 0.
	 See platforms/Cross/vm/sqCogStackAlignment.h and class-side preambleCCode."
	<cmacro: '() STACK_ALIGN_BYTES'>
	^0! !

!ThreadedFFIPlugin methodsFor: 'marshalling-struct' stamp: 'nice 1/30/2020 21:49'!
canReturnInRegistersStructOfSize: returnStructSize
	"Answer if a struct result of a given size can be returned via registers or not.
	Size is a necessary condition, but it might not be a sufficient condition.
	For example, SysV X64 also require that struct fields be properly aligned."
	^self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling-struct' stamp: 'eem 8/28/2020 13:16'!
checkAlignmentOfStructSpec: specs OfLength: specSize StartingAt: startIndex
	"Check the alignment of a structure and return true if correctly aligned.
	If computed size = declared size, then the struct is assumed correctly aligned."
	| index spec computedSize fieldAlignment fieldSize declaredSize maxAlignment |
	<var: #specs type: #'unsigned int *'>
	<inline: false>
	index := startIndex.
	spec := specs at: index.
	self assert: (spec bitAnd: FFIFlagPointer + FFIFlagAtomic + FFIFlagStructure) = FFIFlagStructure.
	(self isUnionSpec: specs OfLength: specSize StartingAt: index) ifTrue:
		[^self checkAlignmentOfUnionSpec: specs OfLength: specSize StartingAt: startIndex].
	declaredSize := spec bitAnd: FFIStructSizeMask.
	computedSize := 0.
	maxAlignment := 1.
	[(index := index + 1) < specSize] whileTrue:
		[spec := specs at: index.
		spec = FFIFlagStructure ifTrue:
			[^(computedSize - 1 bitOr: maxAlignment - 1) + 1 = declaredSize].
		(spec anyMask: FFIFlagPointer)
			ifTrue:
				[fieldSize := BytesPerWord.
				fieldAlignment := fieldSize]
			ifFalse:
				[fieldSize := spec bitAnd: FFIStructSizeMask.
				(spec anyMask: FFIFlagStructure)
					ifTrue:
						[(self checkAlignmentOfStructSpec: specs OfLength: specSize StartingAt: index) ifFalse:
							[^false].
						 fieldAlignment := self alignmentOfStructSpec: specs OfLength: specSize StartingAt: (self addressOf: index put: [:v| index := v])]
					ifFalse: [fieldAlignment := fieldSize]].
		"round to fieldAlignment"
		maxAlignment := maxAlignment max: fieldAlignment.
		computedSize := (computedSize - 1 bitOr: fieldAlignment - 1) + 1.
		computedSize := computedSize + fieldSize].
	^(computedSize - 1 bitOr: maxAlignment - 1) + 1 = declaredSize! !

!ThreadedFFIPlugin methodsFor: 'marshalling-struct' stamp: 'eem 5/31/2020 14:31'!
checkAlignmentOfUnionSpec: specs OfLength: specSize StartingAt: startIndex
	"Check the alignment of a union and return true if correctly aligned.
	Union are correctly aligned, but a sub-structure might not."
	| index spec |
	<var: #specs type: #'unsigned int*'>
	<inline: false>
	index := startIndex.
	spec := specs at: index.
	[index := index + 1.
	index < specSize]
		whileTrue:
			[spec := specs at: index.
			spec = FFIFlagStructure
				ifTrue: [^true].
			(spec anyMask: FFIFlagPointer)
				ifFalse:
					[(spec anyMask: FFIFlagStructure)
						ifTrue:
							[(self checkAlignmentOfStructSpec: specs OfLength: specSize StartingAt: index)
								ifFalse: [^false]]]].
	^true! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 11/11/2009 12:16'!
cleanupCalloutState: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<returnTypeC: #void>
	"Free any temporary arg strings."
	<inline: true>
	[calloutState stringArgIndex > 0] whileTrue:
		[self free: (calloutState stringArgs at: (calloutState stringArgIndex: calloutState stringArgIndex - 1))]! !

!ThreadedFFIPlugin methodsFor: 'simulation' stamp: 'eem 10/16/2018 09:40'!
destroyManualSurface: surfaceID
	<doNotGenerate>
	"Temporarily hack the simulation.  Soon enough we could defer top the actual SurfacePlugin to simulate surfaces properly."
	^1! !

!ThreadedFFIPlugin methodsFor: 'primitive support' stamp: 'eem 8/23/2018 13:40'!
disownFlagsFor: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	^self cppIf: COGMTVM
		ifTrue: [(calloutState callFlags anyMask: FFICallFlagThreaded)
					ifTrue: [DisownVMForFFICall+DisownVMForThreading]
					ifFalse: [DisownVMForFFICall]]
		ifFalse: [DisownVMForFFICall]! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 8/23/2018 14:01'!
dispatchFunctionPointer: aFunctionPointer
	"In C aFunctionPointer is void (*aFunctionPointer)()"
	<cmacro: '(aFunctionPointer) (aFunctionPointer)()'>
	"To write the FFI call failure code we simulate invoking the production VM's
	 fatal exception handlers (sigsegv on Unix, squeakExceptionHandler on WIN32, et al)."
	^[self perform: aFunctionPointer]
		on: Error
		do: [:ex|
			interpreterProxy
				primitiveFailForFFIException: PrimErrFFIException
				at: aFunctionPointer asInteger.
			ex pass "NOTREACHED if VM is handling FFI exceptions"]! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 8/23/2018 14:18'!
dispatchFunctionPointer: aFunctionPointer with: int1 with: int2 with: int3 with: int4
	"In C aFunctionPointer is void (*aFunctionPointer)(int, int, int, int)"
	<cmacro: '(aFunctionPointer, int1, int2, int3, int4) (aFunctionPointer)(int1, int2, int3, int4)'>
	"To write the FFI call failure code we simulate invoking the production VM's
	 fatal exception handlers (sigsegv on Unix, squeakExceptionHandler on WIN32, et al)."
	^[self perform: aFunctionPointer
			with: int1
			with: int2
			with: int3
			with: int4]
		on: Error
		do: [:ex|
			interpreterProxy
				primitiveFailForFFIException: PrimErrFFIException
				at: aFunctionPointer asInteger.
			ex pass "NOTREACHED if VM is handling FFI exceptions"]! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 8/23/2018 14:18'!
dispatchFunctionPointer: aFunctionPointer with: int1 with: int2 with: int3 with: int4 with: int5 with: int6
	"In C aFunctionPointer is void (*aFunctionPointer)(int, int, int, int, int, int)"
	<cmacro: '(aFunctionPointer, int1, int2, int3, int4, int5, int6) (aFunctionPointer)(int1, int2, int3, int4, int5, int6)'>
	"To write the FFI call failure code we simulate invoking the production VM's
	 fatal exception handlers (sigsegv on Unix, squeakExceptionHandler on WIN32, et al)."
	^[self perform: aFunctionPointer
			with: int1
			with: int2
			with: int3
			with: int4
			with: int5
			with: int6]
		on: Error
		do: [:ex|
			interpreterProxy
				primitiveFailForFFIException: PrimErrFFIException
				at: aFunctionPointer asInteger.
			ex pass "NOTREACHED if VM is handling FFI exceptions"]! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 1/6/2019 16:53'!
dispatchFunctionPointer: aFunctionPointer with: int1 with: int2 with: int3 with: int4 with: int5 with: int6 with: int7 with: int8
	"In C aFunctionPointer is void (*aFunctionPointer)(int, int, int, int, int, int)"
	<cmacro: '(aFunctionPointer, int1, int2, int3, int4, int5, int6, int7, int8) (aFunctionPointer)(int1, int2, int3, int4, int5, int6, int7, int8)'>
	"To write the FFI call failure code we simulate invoking the production VM's
	 fatal exception handlers (sigsegv on Unix, squeakExceptionHandler on WIN32, et al)."
	^[self perform: aFunctionPointer
			with: int1
			with: int2
			with: int3
			with: int4
			with: int5
			with: int6
			with: int7
			with: int8]
		on: Error
		do: [:ex|
			interpreterProxy
				primitiveFailForFFIException: PrimErrFFIException
				at: aFunctionPointer asInteger.
			ex pass "NOTREACHED if VM is handling FFI exceptions"]! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 7/27/2021 16:46'!
encodeStructReturnTypeIn: calloutState
	"Set a variable encoding how the struct is to be returned.
	It is an OS dependent encoding, leaved to subclass responsibility."
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self subclassResponsibility
	! !

!ThreadedFFIPlugin methodsFor: 'symbol loading' stamp: 'eem 8/28/2020 20:32'!
externalFunctionHasStackSizeSlot
	<inline: #always>
	^externalFunctionInstSize > ExternalFunctionStackSizeIndex! !

!ThreadedFFIPlugin methodsFor: 'primitive support' stamp: 'eem 7/26/2021 09:12'!
ffiAddressOf: rcvr startingAt: byteOffset size: byteSize
	"Answer a long of the address of the byteSize slot (byte, short, int, whatever) at byteOffset in rcvr.
	 Nominally intended for use with ExternalAddress objects, this code will work (for obscure historical
	 reasons) with plain Byte or Word Arrays as well. Answer 0 on error."
	| rcvrClass rcvrSize addr |
	byteOffset > 0 ifFalse: [^nil].
	rcvrClass := interpreterProxy fetchClassOf: rcvr.
	rcvrSize := interpreterProxy byteSizeOf: rcvr.
	rcvrClass = interpreterProxy classExternalAddress
		ifTrue:
			[rcvrSize = BytesPerWord ifFalse: [^nil].
			addr := interpreterProxy fetchPointer: 0 ofObject: rcvr. "Hack!!!!"
			"don't you dare to read from object memory (unless is pinned)!!"
			(addr = 0 "or: [(interpreterProxy isInMemory: addr) or: [(interpreterProxy isPinned: rcvr) not]]") ifTrue:
				[^nil]]
		ifFalse:
			[byteOffset + byteSize - 1 <= rcvrSize ifFalse:
				[^nil].
			addr := self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'sqIntptr_t'].
	^(addr + byteOffset - 1) asVoidPointer! !

!ThreadedFFIPlugin methodsFor: 'primitive support' stamp: 'eem 10/25/2018 12:25'!
ffiAlloc: bytes
	"Default to malloc/free.  If a platform needs a different allocator define
	 something in the preamble and redefine this to take account of that."
	<cmacro: '(bytes) (usqInt)malloc(bytes)'>
	"Simulate external allocation.  This would be straight-forward were it not for simulating a
	 32-bit system above a 64-bit system, where the result of an allocation won't fit in a 32-bit
	 pointer.  So in this case maintain a map from external allocations to fake addresses."
	| externalAllocation key |
	(self evaluateIfFailed: [externalAllocation := ExternalAddress allocate: bytes]) ifTrue: [^0].
	(self sizeof: #sqIntptr_t) >= Smalltalk wordSize ifTrue:
		[^externalAllocation asInteger].
	allocationMap ifNil:
		[allocationMap := Dictionary new].
	allocationMap at: (key := allocationMap size * (4 * 1024 * 1024) + 16r8E000000) put: externalAllocation asInteger.
	^key! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'mt 8/13/2021 15:30'!
ffiArgByValue: oop in: calloutState
	"Support for generic callout. Prepare an argument by value for a callout."
	<var: #calloutState type: #'CalloutState *'>
	| atomicType intValue floatValue |
	<inline: true>
	<var: #floatValue type: #double>
	atomicType := self atomicTypeOf: calloutState ffiArgHeader.
	"check if the range is valid"
	(atomicType < 0 or:[atomicType > FFITypeDoubleFloat])
		ifTrue:[^FFIErrorBadAtomicType].
	atomicType < FFITypeSingleFloat ifTrue:["integer types"
		(atomicType >> 1) = (FFITypeSignedInt64 >> 1)
			ifTrue:[intValue := oop] "ffi support code must coerce longlong"
			ifFalse:[intValue := self ffiIntegerValueOf: oop]. "does all the coercions"
		interpreterProxy failed ifTrue:[^FFIErrorCoercionFailed].
		^self dispatchOn: atomicType
			in: #(
				ffiPushVoid:in:
				ffiPushUnsignedInt:in:
				ffiPushUnsignedByte:in:
				ffiPushSignedByte:in:
				ffiPushUnsignedShort:in:
				ffiPushSignedShort:in:
				ffiPushUnsignedInt:in:
				ffiPushSignedInt:in:
				ffiPushUnsignedLongLongOop:in:
				ffiPushSignedLongLongOop:in:
				ffiPushUnsignedChar:in:
				ffiPushSignedChar:in:)
			with: intValue
			with: calloutState].
	"either float or double"
	floatValue := self ffiFloatValueOf: oop.
	interpreterProxy failed ifTrue:
		[^FFIErrorCoercionFailed].
	atomicType = FFITypeSingleFloat
		ifTrue: [^self ffiPushSingleFloat: floatValue in: calloutState]
		ifFalse:[^self ffiPushDoubleFloat: floatValue in: calloutState]! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 8/29/2020 13:13'!
ffiArgument: oop Spec: argSpec Class: argClass in: calloutState
	"Callout support. Prepare the given oop as argument.
	argSpec defines the compiled spec for the argument.
	argClass (if non-nil) defines the required (super)class for the argument."
	<var: #calloutState type: #'CalloutState *'>
	| valueOop oopClass isStruct nilOop |
	<inline: false>
	oopClass := interpreterProxy fetchClassOf: oop. "Prefetch class (we'll need it)"
	nilOop :=  interpreterProxy nilObject.
	"Do the necessary type checks"
	argClass = nilOop ifFalse:
		"Type check 1: 
		Is the required class of the argument a general instance of ExternalStructure?"
		[(interpreterProxy
			includesBehavior: argClass 
			ThatOf: interpreterProxy classExternalStructure) ifFalse:
				[^FFIErrorWrongType].
		"Type check 2:
		Is the class of the argument a general instance of the required class?"
		(nilOop = oop or: [interpreterProxy includesBehavior: oopClass ThatOf: argClass]) ifFalse:
			[^FFIErrorCoercionFailed]].
	"Okay, we've passed the type check (so far)"

	"Check if oopClass is a general instance of ExternalStructure.
	 If this is the case we'll work on its handle and not the actual oop."
	isStruct := false.
	(oop ~= nilOop
	 and: [interpreterProxy isPointers: oop]) ifTrue: "#isPointers: will fail if oop is immediate so don't even attempt to use it"
		[isStruct := interpreterProxy
						includesBehavior: oopClass 
						ThatOf: interpreterProxy classExternalStructure.
		 (argClass = nilOop or: [isStruct]) ifFalse:
			[^FFIErrorCoercionFailed]].
	"note: the test for #isPointers: above should speed up execution since no pointer type
	 ST objects are allowed in external calls and thus if #isPointers: is true then the arg must
	 be ExternalStructure to work. If it isn't then the code fails anyways so speed isn't an issue."

	"Determine valueOop (e.g., the actual oop to pass as argument)"
	valueOop := isStruct
					ifTrue: [interpreterProxy fetchPointer: 0 ofObject: oop]
					ifFalse: [oop].

	"Fetch and check the contents of the compiled spec"
	(interpreterProxy isWords: argSpec) ifFalse:
		[^FFIErrorWrongType].
	calloutState ffiArgSpecSize: (interpreterProxy byteSizeOf: argSpec) / (self sizeof: #'unsigned int').
	calloutState ffiArgSpecSize = 0 ifTrue:
		[^FFIErrorWrongType].
	calloutState ffiArgSpec: (self cCoerce: (interpreterProxy firstIndexableField: argSpec) to: #'unsigned int *').
	calloutState ffiArgHeader: (calloutState ffiArgSpec at: 0).

	"Do the actual preparation of the argument"
	"Note: Order is important since FFIFlagStructure + FFIFlagPointer is used to represent 'typedef void* VoidPointer' and VoidPointer really is *struct* not pointer."

	(calloutState ffiArgHeader anyMask: FFIFlagStructure) ifTrue: "argument must be ExternalStructure"
		[isStruct ifFalse:
			[^FFIErrorCoercionFailed].
		(calloutState ffiArgHeader anyMask: FFIFlagAtomic) ifTrue:
			[^FFIErrorWrongType]. "bad combination"
		^self ffiPushStructureContentsOf: valueOop in: calloutState].

	(calloutState ffiArgHeader anyMask: FFIFlagPointer) ifTrue: "no integers (or characters) for pointers please"
		[(interpreterProxy isImmediate: oop) ifTrue:
			[^FFIErrorIntAsPointer].

		"but allow passing nil pointer for any pointer type"
		oop = nilOop ifTrue:
			[^self ffiPushPointer: nil in: calloutState].

		"argument is reference to either atomic or structure type"
		(calloutState ffiArgHeader anyMask: FFIFlagAtomic) ifTrue:
			[isStruct ifTrue:"e.g., ExternalData"
				[^self ffiAtomicStructByReference: oop Class: oopClass in: calloutState].
			^self ffiAtomicArgByReference: oop Class: oopClass in: calloutState].
			"********* NOTE: The above uses 'oop' not 'valueOop' (for ExternalData) ******"

		"Needs to be external structure here"
		isStruct ifTrue:
			[^self ffiPushPointerContentsOf: valueOop in: calloutState].
		^FFIErrorCoercionFailed].

	(calloutState ffiArgHeader anyMask: FFIFlagAtomic) ifTrue: "argument is atomic value"
		[^self ffiArgByValue: valueOop in: calloutState].

	"None of the above - bad spec"
	^FFIErrorWrongType! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'mt 8/16/2021 09:32'!
ffiAtomicArgByReference: oop Class: oopClass in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	"Support for generic callout. Prepare a pointer reference to an atomic type for callout.
	 Note: 
		for type 'void*' we allow ByteArray/String/Symbol, wordVariableSubclass, Alien or ExternalAddress.
		for other types we allow ByteArray, wordVariableSubclass, Alien or ExternalAddress."
	| atomicType isString argIsAlien |
	<inline: #always>
	atomicType := self atomicTypeOf: calloutState ffiArgHeader.
	(atomicType = FFITypeBool) ifTrue: "No bools on input"
		[^FFIErrorCoercionFailed].
	argIsAlien := (isString := interpreterProxy 
								includesBehavior: oopClass 
								ThatOf: interpreterProxy classString)
					ifTrue: [false]
					ifFalse:
						[interpreterProxy 
							includesBehavior: oopClass 
							ThatOf: interpreterProxy classAlien].
	((atomicType >> 1) = (FFITypeSignedChar8 >> 1)) ifTrue:"string value (char*)"
		"note: the only types allowed for passing into char* types are
		ByteArray, String, Symbol, Alien and *no* other byte indexed objects
		(e.g., CompiledMethod, LargeInteger). We only check for strings
		here and fall through to the byte* check otherwise."
		[isString ifTrue:"String/Symbol"
			"Strings must be allocated by the ffi support code"
			[^self ffiPushString: (interpreterProxy firstIndexableField: oop)
				OfLength: (interpreterProxy byteSizeOf: oop)
				in: calloutState].
		"Fall through to byte* test"
		atomicType := FFITypeUnsignedInt8].

	self cppIf: COGMTVM ifTrue:
	["Since all the following pass the address of the first indexable field we need to fail
	 the call if it is threaded and the object is young, since it may move during the call."
	((calloutState callFlags anyMask: FFICallFlagThreaded)
	and: [(argIsAlien not or: [self isDirectAlien: oop])
	and: [interpreterProxy isYoung: oop]]) ifTrue:
		[^PrimErrObjectMayMove negated]].

	(atomicType = FFITypeVoid or:[(atomicType >> 1) = (FFITypeSignedInt8 >> 1)]) ifTrue:
		"byte* -- see comment on string above"
		[(isString or: [oopClass = interpreterProxy classByteArray]) ifTrue: "String/Symbol/ByteArray"
			[^self ffiPushPointer: (interpreterProxy firstIndexableField: oop) in: calloutState].
		(oopClass = interpreterProxy classExternalAddress) ifTrue: 
			[^self ffiPushPointer: (self longAt: oop + interpreterProxy baseHeaderSize) in: calloutState].
		argIsAlien ifTrue:
			[^self ffiPushPointer: (self pointerForOop: (self startOfData: oop)) in: calloutState].
		atomicType = FFITypeVoid ifFalse:
			[^FFIErrorCoercionFailed]].
		"note: type void falls through"

	"I can push pointers to any type (take for instance calls who receive int* output arguments, etc.)
	 but I need to store them into a ByteArray, ExternalAddress or Alien"
	(atomicType <= FFITypeDoubleFloat) ifTrue:
		[((interpreterProxy isWords: oop) or: [oopClass = interpreterProxy classByteArray]) ifTrue:
			[^self ffiPushPointer: (interpreterProxy firstIndexableField: oop) in: calloutState].
		(oopClass = interpreterProxy classExternalAddress) ifTrue: 
			[^self ffiPushPointer: (self longAt: oop + interpreterProxy baseHeaderSize) in: calloutState].
		argIsAlien ifTrue:
			[^self ffiPushPointer: (self pointerForOop: (self startOfData: oop)) in: calloutState]].

	^FFIErrorCoercionFailed! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 7/27/2021 16:48'!
ffiAtomicStructByReference: oop Class: oopClass in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	"Support for generic callout. Prepare an external pointer reference to an atomic type for callout."
	| atomicType err valueOop |
	<inline: #always>
	"must be external data to pass pointers to atomic type"
	oopClass = interpreterProxy classExternalData 
		ifFalse:[^FFIErrorCoercionFailed].
	atomicType := self atomicTypeOf: calloutState ffiArgHeader.
	"no type checks for void pointers"
	atomicType ~= FFITypeVoid ifTrue:[
		err := self ffiValidateExternalData: oop AtomicType: atomicType.
		err ~= 0 ifTrue:[^err].
	].
	"and push pointer contents"
	valueOop := interpreterProxy fetchPointer: 0 ofObject: oop.
	^self ffiPushPointerContentsOf: valueOop in: calloutState! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'nice 1/30/2020 21:50'!
ffiCall: externalFunction ArgArrayOrNil: argArrayOrNil NumArgs: nArgs
	"Generic callout. Does the actual work.  If argArrayOrNil is nil it takes args from the stack
	 and the spec from the method.  If argArrayOrNil is not nil takes args from argArrayOrNil
	 and the spec from the receiver."
	| flags argTypeArray address argType oop argSpec argClass err theCalloutState calloutState requiredStackSize stackSize allocation result primNumArgs |
	<inline: true>
	<var: #theCalloutState type: #'CalloutState'>
	<var: #calloutState type: #'CalloutState *'>
	<var: #allocation type: #'char *'>

	primNumArgs := interpreterProxy methodArgumentCount.
	(interpreterProxy is: externalFunction KindOfClass: interpreterProxy classExternalFunction) ifFalse:
		[^self ffiFail: FFIErrorNotFunction].
	"Load and check the values in the externalFunction before we call out"
	flags := interpreterProxy fetchInteger: ExternalFunctionFlagsIndex ofObject: externalFunction.
	interpreterProxy failed ifTrue:
		[^self ffiFail: FFIErrorBadArgs].

	"This must come early for compatibility with the old FFIPlugin.  Image-level code
	 may assume the function pointer is loaded eagerly.  Thanks to Nicolas Cellier."
	address := self ffiLoadCalloutAddress: externalFunction.
	interpreterProxy failed ifTrue:
		[^0 "error code already set by ffiLoadCalloutAddress:"].
	
	argTypeArray := interpreterProxy fetchPointer: ExternalFunctionArgTypesIndex ofObject: externalFunction.
	"must be array of arg types"
	((interpreterProxy isArray: argTypeArray)
	and: [(interpreterProxy slotSizeOf: argTypeArray) = (nArgs + 1)]) ifFalse:
		[^self ffiFail: FFIErrorBadArgs].
	"check if the calling convention is supported"
	self cppIf: COGMTVM
		ifTrue:
			[(self ffiSupportsCallingConvention: (flags bitAnd: FFICallTypesMask)) ifFalse:
				[^self ffiFail: FFIErrorCallType]]
		ifFalse: "not masking causes threaded calls to fail, which is as they should if the plugin is not threaded."
			[(self ffiSupportsCallingConvention: flags) ifFalse:
				[^self ffiFail: FFIErrorCallType]].
		
	requiredStackSize := self externalFunctionHasStackSizeSlot
							ifTrue: [interpreterProxy
										fetchInteger: ExternalFunctionStackSizeIndex
										ofObject: externalFunction]
							ifFalse: [-1].
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: (argArrayOrNil isNil
												ifTrue: [PrimErrBadMethod]
												ifFalse: [PrimErrBadReceiver])].
	stackSize := requiredStackSize < 0 ifTrue: [DefaultMaxStackSize] ifFalse: [requiredStackSize].
	self cCode: [] inSmalltalk: [theCalloutState := self class calloutStateClass new].
	calloutState := self addressOf: theCalloutState.
	self cCode: [self memset: calloutState _: 0 _: (self sizeof: #CalloutState)].
	calloutState callFlags: flags.
	"Fetch return type and args"
	argType := interpreterProxy fetchPointer: 0 ofObject: argTypeArray.
	argSpec := interpreterProxy fetchPointer: 0 ofObject: argType.
	argClass := interpreterProxy fetchPointer: 1 ofObject: argType.
	(err := self ffiCheckReturn: argSpec With: argClass in: calloutState) ~= 0 ifTrue:
		[^self ffiFail: err]. "cannot return"
	"alloca the outgoing stack frame, leaving room for marshalling args, and including space for the return struct, if any.
	Additional space reserved for saving register args like mandated by Win64 X64 or PPC ABI, will be managed by the call itself"
	allocation := self alloca: stackSize + calloutState structReturnSize + self cStackAlignment.
	self mustAlignStack ifTrue:
		[allocation := self cCoerce: (allocation asUnsignedIntegerPtr bitClear: self cStackAlignment - 1) to: #'char *'].
	calloutState
		argVector: allocation;
		currentArg: allocation;
		limit: allocation + stackSize.
	(calloutState structReturnSize > 0
	 and: [self nonRegisterStructReturnIsViaImplicitFirstArgument
	 and: [(self returnStructInRegisters: calloutState) not]]) ifTrue:
		[err := self ffiPushPointer: calloutState limit in: calloutState.
		 err ~= 0 ifTrue:
			[self cleanupCalloutState: calloutState.
			 self cppIf: COGMTVM ifTrue:
			 [err = PrimErrObjectMayMove negated ifTrue:
				[^PrimErrObjectMayMove]]. "N.B. Do not fail if object may move because caller will GC and retry."
			 ^self ffiFail: err]].
	1 to: nArgs do:
		[:i|
		argType := interpreterProxy fetchPointer: i ofObject: argTypeArray.
		argSpec := interpreterProxy fetchPointer: 0 ofObject: argType.
		argClass := interpreterProxy fetchPointer: 1 ofObject: argType.
		oop := argArrayOrNil isNil
				ifTrue: [interpreterProxy stackValue: nArgs - i]
				ifFalse: [interpreterProxy fetchPointer: i - 1 ofObject: argArrayOrNil].
		err := self ffiArgument: oop Spec: argSpec Class: argClass in: calloutState.
		err ~= 0 ifTrue:
			[self cleanupCalloutState: calloutState.
			 self cppIf: COGMTVM ifTrue:
			 [err = PrimErrObjectMayMove negated ifTrue:
				[^PrimErrObjectMayMove]]. "N.B. Do not fail if object may move because caller will GC and retry."
			 ^self ffiFail: err]]. "coercion failed or out of stack space"
	"Failures must be reported back from ffiArgument:Spec:Class:in:.
	 Should not fail from here on in."
	self assert: interpreterProxy failed not.
	self ffiLogCallout: externalFunction.
	(requiredStackSize < 0
	 and: [self externalFunctionHasStackSizeSlot]) ifTrue:
		[stackSize := calloutState currentArg - calloutState argVector.
		 interpreterProxy storeInteger: ExternalFunctionStackSizeIndex ofObject: externalFunction withValue: stackSize].
	"Go out and call this guy"
	result := self ffiCalloutTo: address SpecOnStack: argArrayOrNil notNil in: calloutState.
	self cleanupCalloutState: calloutState.
	"Can not safely use argumentCount (via e.g. methodReturnValue:) since it may have been changed by a callback."
	interpreterProxy pop: primNumArgs + 1 thenPush: result. 
	^result! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 8/29/2020 13:20'!
ffiCalloutTo: procAddr SpecOnStack: specOnStack in: calloutState
	<var: #procAddr type: #'void *'>
	<arg: #calloutState type: #'CalloutState *'>
	"Perform the callout, collect the result and and create the return value.
	 If a threaded call, disown and own VM around the call.  If there are floating-point
	 arguments that are passed in registers then call a dummy function to load them.
	 This *must* be inlined because of the alloca of the outgoing stack frame in
	 ffiCall:SpecOnStack:Flags:NumArgs:Args:AndTypes:"
	<inline: #always>
	self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'nice 1/30/2020 21:14'!
ffiCheckReturn: retSpec With: retClass in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	"Make sure we can return an object of the given type"
	<inline: true>
	retClass = interpreterProxy nilObject ifFalse:
		[(interpreterProxy
				includesBehavior: retClass 
				ThatOf: interpreterProxy classExternalStructure) ifFalse:
			[^FFIErrorBadReturn]].

	((interpreterProxy isWords: retSpec)
	 and: [(interpreterProxy slotSizeOf: retSpec) > 0]) ifFalse:
		[^FFIErrorWrongType].

	calloutState ffiRetSpec: retSpec.
	calloutState ffiRetHeader: (interpreterProxy fetchLong32: 0 ofObject: retSpec).
	(self isAtomicType: calloutState ffiRetHeader) ifFalse:
		[retClass = interpreterProxy nilObject ifTrue:
			[^FFIErrorBadReturn]].
	(calloutState ffiRetHeader bitAnd: (FFIFlagPointer bitOr: FFIFlagStructure)) = FFIFlagStructure ifTrue:
		[calloutState structReturnSize: (calloutState ffiRetHeader bitAnd: FFIStructSizeMask).
		self encodeStructReturnTypeIn: calloutState].
	^0! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 7/27/2021 16:42'!
ffiContentsOfHandle: oop errCode: errCode
	"Make sure that the given oop is a valid external handle"
	<inline: #always>
	((interpreterProxy isBytes: oop)
	 and: [(interpreterProxy byteSizeOf: oop) = (self sizeof: #sqInt)]) ifFalse:
		[^self ffiFail: errCode].
	^interpreterProxy fetchPointer: 0 ofObject: oop! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'mt 8/13/2021 15:28'!
ffiCreateIntegralResultOop: retVal ofAtomicType: atomicType in: calloutState
	<inline: #always>
	<var: #calloutState type: #'CalloutState *'>
	<var: #retVal type: #usqLong>
	"Callout support. Return the appropriate oop for the given atomic type"
	| shift value mask byteSize |
	<var: 'value' type: #usqLong>
	<var: 'mask' type: #usqLong>
	self assert: atomicType < FFITypeSingleFloat.

	atomicType = FFITypeBool ifTrue:
		["Make sure bool honors the byte size requested"
		 byteSize := calloutState ffiRetHeader bitAnd: FFIStructSizeMask.
		 value := byteSize = (self sizeof: retVal)
					ifTrue:[retVal]
					ifFalse:[retVal bitAnd: 1 asUnsignedLongLong << (byteSize * 8) - 1].
		 ^value = 0
			ifTrue:[interpreterProxy falseObject]
			ifFalse:[interpreterProxy trueObject]].
	atomicType <= FFITypeSignedInt32 ifTrue:
		["these are all generall integer returns"
		atomicType <= (BytesPerWord = 8 ifTrue: [FFITypeSignedInt32] ifFalse: [FFITypeSignedInt16]) ifTrue:
			["byte/short. first extract partial word, then sign extend"
			shift := (BytesPerWord = 8 and: [atomicType >= FFITypeUnsignedInt32])
						ifTrue: [32]
						ifFalse: [(atomicType >> 1) * 8]. "# of significant bits"
			value := retVal bitAnd: (1 asUnsignedLongLong << shift - 1). 
			(atomicType anyMask: 1) ifTrue:
				["make the guy signed"
				mask := 1 asUnsignedLongLong << (shift-1).
				value := (value bitAnd: mask-1) - (value bitAnd: mask)].
			^interpreterProxy integerObjectOf: value].
		"Word sized integer return"
		^(atomicType anyMask: 1)
			ifTrue:[interpreterProxy signedMachineIntegerFor: retVal] "signed return"
			ifFalse:[interpreterProxy positiveMachineIntegerFor: retVal]]. "unsigned return"

	"longlong, char"
	(atomicType >> 1) = (FFITypeSignedInt64 >> 1) ifTrue:
		[^(atomicType anyMask: 1)
			ifTrue: [interpreterProxy signed64BitIntegerFor: retVal] "signed return"
			ifFalse: [interpreterProxy positive64BitIntegerFor: retVal]].
	self flag: 'we need a solution for wide characters; spur supports upto 30 bit characters'.
	^interpreterProxy characterObjectOf: (retVal bitAnd: 16rFF)! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 7/13/2010 14:32'!
ffiFail: reason
	<inline: false>
	"Map the FFI error code into a primitive error code.  If reason is negative it encodes one of the
	 standard PrimErr... codes, negated to distinguish it from the FFIError codes.  If it is an FFIError...
	 code then add the size of the primitive error table + 2 to disambiguate it from the PrimErr... codes.
	 For historic reasons the FFIError codes range from -1 on up hence adding size + 2 maps them to
	 size of table + 1 on up.  This OFFSET IS undone by ExternalFunction class>>externalCallFailedWith:.
	 Thus we can communicate back both PrimErr.. and FFIError codes.  Complex but necessary."
	ffiLastError := reason.
	^interpreterProxy primitiveFailFor:
		(reason >= FFINoCalloutAvailable
			ifTrue: [reason + 2 + (interpreterProxy slotSizeOf: interpreterProxy primitiveErrorTable)]
			ifFalse: [reason negated])! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 7/27/2021 16:43'!
ffiFloatValueOf: oop
	"Support for generic callout. Return a float value that is coerced as C would do."
	<returnTypeC: #double>
	<inline: #always>
	(interpreterProxy isFloatObject: oop) ifTrue:
		[^interpreterProxy floatValueOf: oop].
	"otherwise try the integer coercions and return its float value"
	^(self ffiIntegerValueOf: oop) asFloat! !

!ThreadedFFIPlugin methodsFor: 'primitive support' stamp: 'eem 11/11/2009 13:34'!
ffiFree: pointer
	"Default to malloc/free.  If a platform needs a different allocator define
	 something in the preamble and redefine this to take account of that."
	<cmacro: '(pointer) free((void *)(pointer))'>
	^self! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'nice 4/19/2016 22:06'!
ffiIntegerValueOf: oop
	"Support for generic callout. Answer an integer value that is coerced as C would do."
	<inline: true>
	"Cheat with a tag test"
	(oop anyMask: BytesPerWord - 1)
		ifTrue:
			[(interpreterProxy isIntegerObject: oop) ifTrue:
				[^interpreterProxy integerValueOf: oop].
			self cppIf: SPURVM
				ifTrue:
					[(interpreterProxy isCharacterObject: oop) ifTrue: "Immediate in Spur"
						[^interpreterProxy characterValueOf: oop].
					 (interpreterProxy isFloatObject: oop) ifTrue: "Immediate in 64-bit Spur"
						[^interpreterProxy floatValueOf: oop]]]
		ifFalse:
			[self cppIf: SPURVM
				ifTrue: "No non-immediate characters in Spur"
					[]
				ifFalse:
					[(interpreterProxy isCharacterObject: oop) ifTrue:
						[^interpreterProxy characterValueOf: oop]].
			 (interpreterProxy isFloatObject: oop) ifTrue:
				[^interpreterProxy floatValueOf: oop].
			 oop = interpreterProxy nilObject ifTrue: [^0]. "@@: should we really allow this????"
			 oop = interpreterProxy falseObject ifTrue: [^0].
			 oop = interpreterProxy trueObject ifTrue: [^1].
			 (interpreterProxy isLargePositiveIntegerObject: oop) ifTrue:
				[self cppIf: BytesPerWord = 8 "Use cppIf: to get the return type of the function right.  Should be sqInt on 32-bits."
					ifTrue: [^interpreterProxy positive64BitValueOf: oop]
					ifFalse: [^interpreterProxy positive32BitValueOf: oop]]].
	^interpreterProxy signedMachineIntegerValueOf: oop "<- will fail if not integer"! !

!ThreadedFFIPlugin methodsFor: 'symbol loading' stamp: 'eem 7/27/2021 16:50'!
ffiLoadCalloutAddress: lit
	"Load the address of the foreign function from the given object"
	<inline: #always>
	| addressPtr address ptr |
	<var: #ptr type: #'sqIntptr_t *'>
	"Lookup the address"
	addressPtr := interpreterProxy fetchPointer: 0 ofObject: lit.
	"Make sure it's an external handle"
	address := self ffiContentsOfHandle: addressPtr errCode: FFIErrorBadAddress.
	interpreterProxy failed ifTrue:
		[^0].
	address = 0 ifTrue:"Go look it up in the module"
		[self externalFunctionHasStackSizeSlot ifTrue:
			[interpreterProxy storeInteger: ExternalFunctionStackSizeIndex ofObject: lit withValue: -1].
		(interpreterProxy slotSizeOf: lit) < 5 ifTrue:
			[^self ffiFail: FFIErrorNoModule].
		address := self ffiLoadCalloutAddressFrom: lit.
		interpreterProxy failed ifTrue:
			[^0].
		"Store back the address"
		ptr := interpreterProxy firstIndexableField: addressPtr.
		ptr at: 0 put: address].
	^address! !

!ThreadedFFIPlugin methodsFor: 'symbol loading' stamp: 'eem 11/24/2014 13:02'!
ffiLoadCalloutAddressFrom: oop
	"Load the function address for a call out to an external function"
	| module moduleHandle functionName address |
	<inline: false>
	"First find and load the module"
	module := interpreterProxy fetchPointer: externalFunctionInstSize + 1 ofObject: oop.
	moduleHandle := self ffiLoadCalloutModule: module.
	interpreterProxy failed ifTrue:
		[^0]. "failed"
	"fetch the function name"
	functionName := interpreterProxy fetchPointer: externalFunctionInstSize ofObject: oop.
	(interpreterProxy isBytes: functionName) ifFalse:
		[^self ffiFail: FFIErrorBadExternalFunction].
	address := (interpreterProxy
					ioLoadSymbol: (interpreterProxy firstIndexableField: functionName) asInteger
					OfLength: (interpreterProxy byteSizeOf: functionName) 
					FromModule: moduleHandle) asInteger.
	(interpreterProxy failed or: [address = 0]) ifTrue:
		[^self ffiFail: FFIErrorAddressNotFound].
	^address! !

!ThreadedFFIPlugin methodsFor: 'symbol loading' stamp: 'nice 1/24/2017 22:55'!
ffiLoadCalloutModule: module
	"Load the given module and return its handle"
	| moduleHandlePtr moduleHandle ffiModuleName moduleLength rcvr ptr |
	<var: #ptr type:'sqInt *'>
	(interpreterProxy isBytes: module) ifTrue:[
		"plain module name"
		ffiModuleName := module.
		moduleLength := interpreterProxy byteSizeOf: ffiModuleName.
		moduleHandle := (interpreterProxy
							ioLoadModule: (interpreterProxy firstIndexableField: ffiModuleName) asInteger
							OfLength: moduleLength) asInteger.
		(interpreterProxy failed
		 or: [moduleHandle = 0]) ifTrue:
			[^self ffiFail: FFIErrorModuleNotFound]. "failed"
		^moduleHandle].
	"Check if the external method is defined in an external library"
	rcvr := interpreterProxy stackValue: interpreterProxy methodArgumentCount.
	(interpreterProxy is: rcvr KindOfClass: interpreterProxy classExternalLibrary) ifFalse:
		[^self ffiFail: FFIErrorNoModule].
	"external library"
	moduleHandlePtr := interpreterProxy fetchPointer: 0 ofObject: rcvr.
	moduleHandle := self ffiContentsOfHandle: moduleHandlePtr errCode: FFIErrorBadExternalLibrary.
	interpreterProxy failed ifTrue:[^0].
	moduleHandle = 0 ifTrue:["need to reload module"
		ffiModuleName := interpreterProxy fetchPointer: 1 ofObject: rcvr.
		(interpreterProxy isBytes: ffiModuleName) ifFalse:[^self ffiFail: FFIErrorBadExternalLibrary].
		moduleLength := interpreterProxy byteSizeOf: ffiModuleName.
		moduleHandle := (interpreterProxy
								ioLoadModule: (interpreterProxy firstIndexableField: ffiModuleName) asInteger
								OfLength: moduleLength) asInteger.
		(interpreterProxy failed
		 or: [moduleHandle = 0]) ifTrue:
			[^self ffiFail: FFIErrorModuleNotFound]. "failed"
		"and store back"
		ptr := interpreterProxy firstIndexableField: moduleHandlePtr.
		ptr at: 0 put: moduleHandle].
	^moduleHandle! !

!ThreadedFFIPlugin methodsFor: 'symbol loading' stamp: 'eem 11/18/2009 16:13'!
ffiLogCallout: lit
	<returnTypeC: #void>
	"fetch the function name"
	| functionName |
	ffiLogEnabled ifTrue:[
		functionName := interpreterProxy fetchPointer: externalFunctionInstSize ofObject: lit.
		(interpreterProxy isBytes: functionName) ifFalse:[^nil].
		self ffiLogCall: (interpreterProxy firstIndexableField: functionName)
			OfLength: (interpreterProxy byteSizeOf: functionName).
	].! !

!ThreadedFFIPlugin methodsFor: 'initialize' stamp: 'ar (auto pragmas 12/08) 7/27/2007 11:18'!
ffiLogCallsTo: fileName
	"This is a special entry point exposed such that client code can 
	enable and disable logging of FFI calls."
	| ok |
	<export: true>
	<var: #fileName type: 'char*'>
	fileName == nil ifTrue:[ "disable logging"
		ok := self ffiLogFileName: nil OfLength: 0.
		ok ifFalse:[^false].
		ffiLogEnabled := false.
	] ifFalse:[ "enable logging"
		ok := self ffiLogFileName: fileName OfLength: (self strlen: fileName).
		ok ifFalse:[^false].
		ffiLogEnabled := true.
	].
	^true! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushDoubleFloat: value in: calloutState
	<var: #calloutState type: #double>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	^self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushPointer: pointer in: calloutState
	<var: #pointer type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 8/28/2020 16:27'!
ffiPushPointerContentsOf: oop in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	"Push the contents of the given external structure"
	| ptrClass ptrAddress |
	<inline: false>
	ptrClass := interpreterProxy fetchClassOf: oop.
	ptrClass = interpreterProxy classExternalAddress ifTrue:
		[ptrAddress := (interpreterProxy fetchPointer: 0 ofObject: oop) asVoidPointer.
		"Don't you dare to pass pointers into object memory"
		(interpreterProxy isInMemory: ptrAddress asUnsignedIntegerPtr) ifTrue:
			[^FFIErrorInvalidPointer].
		^self ffiPushPointer: ptrAddress in: calloutState].

	ptrClass = interpreterProxy classByteArray ifTrue:
		["Since this involves passing the address of the first indexable field we need to fail
		  the call if it is threaded and the object is young, since it may move during the call."
		 self cppIf: COGMTVM ifTrue:
		 [((calloutState callFlags anyMask: FFICallFlagThreaded)
		   and: [interpreterProxy isYoung: oop]) ifTrue:
			[^PrimErrObjectMayMove negated]].

		ptrAddress := interpreterProxy firstIndexableField: oop.
		^self ffiPushPointer: ptrAddress in: calloutState].

	(interpreterProxy includesBehavior: ptrClass ThatOf: interpreterProxy classAlien) ifTrue:
		[self cppIf: COGMTVM ifTrue:
		 [((calloutState callFlags anyMask: FFICallFlagThreaded)
		   and: [(self isDirectAlien: oop)
		   and: [interpreterProxy isYoung: oop]]) ifTrue:
			[^PrimErrObjectMayMove negated]].

		ptrAddress := self cCoerce: (self startOfData: oop) to: #'void *'.
		^self ffiPushPointer: ptrAddress in: calloutState].

	^FFIErrorBadArg! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 7/27/2021 16:54'!
ffiPushSignedByte: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedChar: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedInt: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedLongLong: value in: calloutState
	<var: #value type: #sqLong>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	^self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedLongLongOop: oop in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	"Push a longlong type (e.g., a 64bit integer).
	Note: Coercions from float are *not* supported."
	| value |
	<var: #value type: #sqLong>
	(oop = interpreterProxy nilObject
	 or: [oop = interpreterProxy falseObject])
		ifTrue:[value := 0] ifFalse:
	[oop = interpreterProxy trueObject
		ifTrue:[value := 1] ifFalse:
	[value := interpreterProxy signed64BitValueOf: oop.
	 interpreterProxy failed ifTrue:
		[^FFIErrorCoercionFailed]]].
	^self ffiPushSignedLongLong: value in: calloutState! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedShort: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSingleFloat: value in: calloutState
	<var: #value type: #float>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushString: pointer OfLength: length in: calloutState
	<var: #pointer type: #'char *'>
	<var: #calloutState type: #'CalloutState *'>
	| copy |
	<var: #copy type: #'char *'>
	<inline: #always>
	calloutState stringArgIndex >= MaxNumArgs ifTrue:
		[^PrimErrBadNumArgs negated].
	copy := self malloc: length + 1.
	copy isNil ifTrue:
		[^PrimErrNoCMemory negated].
	self memcpy: copy _: pointer _: length.
	copy at: length put: 0.
	calloutState stringArgs at: calloutState stringArgIndex put: copy.
	calloutState stringArgIndex: calloutState stringArgIndex + 1.
	^self ffiPushPointer: copy in: calloutState! !

!ThreadedFFIPlugin methodsFor: 'marshalling-struct' stamp: 'eem 11/11/2009 12:41'!
ffiPushStructure: pointer ofSize: structSize typeSpec: argSpec ofLength: argSpecSize in: calloutState
	<var: #pointer type: #'void *'>
	<var: #argSpec type: #'sqInt *'>
	<var: #calloutState type: #'CalloutState *'>
	<inline: true>
	self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling-struct' stamp: 'eem 8/28/2020 16:11'!
ffiPushStructureContentsOf: oop in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	"Push the contents of the given external structure"
	| ptrClass ptrAddress |
	<inline: true>
	ptrClass := interpreterProxy fetchClassOf: oop.
	ptrClass = interpreterProxy classExternalAddress ifTrue: "ExternalAddress is bytes"
		[ptrAddress := (interpreterProxy fetchPointer: 0 ofObject: oop) asVoidPointer.
		"There is no way we can make sure the structure is valid.
		But we can at least check for attempts to pass pointers to ST memory."
		(interpreterProxy isInMemory: ptrAddress asUnsignedIntegerPtr) ifTrue:
			[^FFIErrorInvalidPointer].
		^self ffiPushStructure: ptrAddress
			ofSize: (calloutState ffiArgHeader bitAnd: FFIStructSizeMask)
			typeSpec: calloutState ffiArgSpec
			ofLength: calloutState ffiArgSpecSize
			in: calloutState].
	ptrClass = interpreterProxy classByteArray ifTrue:
		["The following is a somewhat pessimistic test but I like being sure..."
		(interpreterProxy byteSizeOf: oop) = (calloutState ffiArgHeader bitAnd: FFIStructSizeMask)
			ifFalse:[^FFIErrorStructSize].
		ptrAddress := interpreterProxy firstIndexableField: oop.
		(calloutState ffiArgHeader anyMask: FFIFlagPointer) ifFalse:
			"Since this involves passing the address of the first indexable field we need to fail
			  the call if it is threaded and the object is young, since it may move during the call."
			[self cppIf: COGMTVM ifTrue:
			 [((calloutState callFlags anyMask: FFICallFlagThreaded)
			 and: [interpreterProxy isYoung: oop]) ifTrue:
				[^PrimErrObjectMayMove negated]].
			^self ffiPushStructure: ptrAddress
				ofSize: (calloutState ffiArgHeader bitAnd: FFIStructSizeMask)
				typeSpec: calloutState ffiArgSpec
				ofLength: calloutState ffiArgSpecSize
				in: calloutState].
		"If FFIFlagPointer + FFIFlagStructure is set use ffiPushPointer on the contents"
		(calloutState ffiArgHeader bitAnd: FFIStructSizeMask) = BytesPerWord ifFalse:
			[^FFIErrorStructSize].
		ptrAddress := (interpreterProxy fetchPointer: 0 ofObject: oop) asVoidPointer.
		(interpreterProxy isInMemory: ptrAddress asUnsignedIntegerPtr) ifTrue:
			[^FFIErrorInvalidPointer].
		^self ffiPushPointer: ptrAddress in: calloutState].
	^FFIErrorBadArg! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedByte: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedChar: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedInt: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 11/6/2009 14:29'!
ffiPushUnsignedLongLong: value in: calloutState
	<var: #value type: #usqLong>
	^self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedLongLongOop: oop in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	"Push an unsigned longlong type (e.g., a 64bit integer).
	Note: Coercions from float are *not* supported."
	<inline: #always>
	| value |
	<var: #value type: #usqLong>
	(oop = interpreterProxy nilObject
	 or: [oop = interpreterProxy falseObject])
		ifTrue:[value := 0] ifFalse:
	[oop = interpreterProxy trueObject
		ifTrue:[value := 1] ifFalse:
	[value := interpreterProxy positive64BitValueOf: oop.
	 interpreterProxy failed ifTrue:
		[^FFIErrorCoercionFailed]]].
	^self ffiPushUnsignedLongLong: value in: calloutState! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedShort: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushVoid: ignored in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	"This is a fallback in case somebody tries to pass a 'void' value.
	We could simply ignore the argument but I think it's better to let
	the caller know what he did"
	^FFIErrorAttemptToPassVoid! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 12/14/2016 12:55'!
ffiReturnCStringFrom: cPointer
	"Create a Smalltalk string from a zero terminated C string"
	| strLen strOop cString strPtr |
	<var: #cString type: #'char *'>
	<var: #strPtr type: #'char *'>
	cPointer ifNil: [^interpreterProxy nilObject]. "nil always returns as nil"
	cString := self cCoerce: cPointer to: #'char *'.
	strLen := 0.
	[(cString at: strLen) = 0] whileFalse:[strLen := strLen+1].
	strOop := interpreterProxy 
				instantiateClass: interpreterProxy classString 
				indexableSize: strLen.
	strPtr := interpreterProxy firstIndexableField: strOop.
	0 to: strLen-1 do: [:i| strPtr at: i put: (cString at: i)].
	^strOop! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'mt 8/16/2021 09:32'!
ffiReturnPointer: retVal ofType: retType in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<var: #retVal type: #usqLong>
	"Generic callout support. Create a pointer return value from an external function call"
	| retClass atomicType retOop oop ptr classOop |
	<var: #ptr type: #'sqInt *'>
	retClass := interpreterProxy fetchPointer: 1 ofObject: retType.
	retClass = interpreterProxy nilObject ifTrue:
		["Create ExternalData upon return"
		atomicType := self atomicTypeOf: calloutState ffiRetHeader.
		(atomicType >> 1) = (FFITypeSignedChar8 >> 1) ifTrue: "String return"
			[retOop := self ffiReturnCStringFrom: (self cCoerceSimple: retVal to: #usqInt).
			 ^retOop].
		"generate external data"
		self remapOop: retType in:
			[oop := interpreterProxy
						instantiateClass: interpreterProxy classExternalAddress 
						indexableSize: BytesPerWord.
			ptr := interpreterProxy firstIndexableField: oop.
			ptr at: 0 put: (self cCoerceSimple: retVal to: #sqInt).
			self remapOop: oop in:
				[retOop := interpreterProxy 
								instantiateClass: interpreterProxy classExternalData 
								indexableSize: 0].
			interpreterProxy storePointer: 0 ofObject: retOop withValue: oop].
		interpreterProxy storePointer: 1 ofObject: retOop withValue: retType.
		^retOop].
	"non-atomic pointer return"
	classOop := (calloutState ffiRetHeader anyMask: FFIFlagStructure)
					ifTrue:[interpreterProxy classByteArray]
					ifFalse:[interpreterProxy classExternalAddress].
	self remapOop: retClass in:
		[oop := interpreterProxy 
					instantiateClass: classOop
					indexableSize: BytesPerWord].
	ptr := interpreterProxy firstIndexableField: oop.
	ptr at: 0 put: (self cCoerceSimple: retVal to: #sqInt).
	self remapOop: oop in:
		[retOop := interpreterProxy instantiateClass: retClass indexableSize: 0].
	interpreterProxy storePointer: 0 ofObject: retOop withValue: oop.
	^retOop! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'nice 1/25/2020 23:20'!
ffiReturnStruct: longLongRetPtr ofType: ffiRetClass in: calloutState
	<var: #longLongRetPtr type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	"Create a structure return value from an external function call.  The value as been stored in
	 alloca'ed space pointed to by the calloutState."
	<inline: true>
	self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'symbol loading' stamp: 'eem 12/16/2009 11:08'!
ffiReturnType: specOnStack
	"Answer the return type object for the current invocation"
	| specLiteral argTypes |
	specLiteral := specOnStack
					ifTrue: [interpreterProxy stackValue: 1]
					ifFalse: [interpreterProxy literal: 0 ofMethod: interpreterProxy primitiveMethod].
	argTypes := interpreterProxy fetchPointer: ExternalFunctionArgTypesIndex ofObject: specLiteral.
	^interpreterProxy fetchPointer: 0 ofObject: argTypes! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 7/21/2011 14:33'!
ffiSupportsCallingConvention: aCallingConvention
	"Check that the calling convention is valid.  This test also filters out attempts
	 to do a threaded call in the non-threaded VM/plugin combinatioin."

	<inline: true>
	^aCallingConvention = FFICallTypeCDecl or: [aCallingConvention = FFICallTypeApi]! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'mt 8/16/2021 09:33'!
ffiValidateExternalData: oop AtomicType: atomicType
	"Validate if the given oop (an instance of ExternalData) can be passed as a pointer to the given atomic type."
	| ptrType specOop spec specType |
	<inline: true>
	ptrType := interpreterProxy fetchPointer: 1 ofObject: oop.
	((interpreterProxy isPointers: ptrType)
	and: [(interpreterProxy slotSizeOf: ptrType) >= 2]) ifFalse:
		[^FFIErrorWrongType].
	specOop := interpreterProxy fetchPointer: 0 ofObject: ptrType.
	((interpreterProxy isWords: specOop)
	and: [(interpreterProxy slotSizeOf: specOop) > 0]) ifFalse:
		[^FFIErrorWrongType].
	spec := interpreterProxy fetchPointer: 0 ofObject: specOop.
	(self isAtomicType: spec) ifFalse:
		[^FFIErrorWrongType].
	specType := self atomicTypeOf: spec.
	specType ~= atomicType ifTrue:
		"Allow for signed/unsigned conversion but nothing else.
		 See FFIConstants class>>#initializeTypeConstants"
		[(atomicType >= FFITypeUnsignedInt8
		  and: [atomicType <= FFITypeSignedChar32
		  and: [(atomicType >> 1) = (specType >> 1)]]) ifFalse:
			[^FFIErrorCoercionFailed]].
	^0! !

!ThreadedFFIPlugin methodsFor: 'initialize' stamp: 'eem 9/27/2020 14:13'!
getModuleName
	"Note: This is hardcoded so it can be run from Squeak.
	The module name is used for validating a module *after*
	it is loaded to check if it does really contain the module
	we're thinking it contains. This is important!!"
	<returnTypeC: #'const char *'>
	<export: true>
	^'SqueakFFIPrims', PluginVersionInfo! !

!ThreadedFFIPlugin methodsFor: 'simulation' stamp: 'eem 12/21/2017 15:12'!
initSurfacePluginFunctionPointers
	"This is a simulation-only stub.  The real code is in
		platforms/Cross/plugins/SqueakFFIPrims/sqManualSurface.c"
	<doNotGenerate>! !

!ThreadedFFIPlugin methodsFor: 'initialize' stamp: 'eem 8/23/2018 09:42'!
initialiseModule
	<export: true>
	ffiLastError := 0.
	"By default, disable logging"
	ffiLogEnabled := false.
	"Get the instSize of ExternalFunction to know whether it contains a cache of the stackSize,
	 and what the offset of ExternalLibraryFunction's functionName and moduleName slots are."
	externalFunctionInstSize := interpreterProxy instanceSizeOf: interpreterProxy classExternalFunction.
	self initSurfacePluginFunctionPointers.
	^true! !

!ThreadedFFIPlugin methodsFor: 'primitive support' stamp: 'ar 12/2/1999 21:04'!
isAtomicType: typeSpec
	^typeSpec anyMask: FFIFlagAtomic! !

!ThreadedFFIPlugin methodsFor: 'callout support' stamp: 'eem 11/11/2009 12:55'!
isCalleePopsConvention: callType
	<cmacro: '(callType) (PLATFORM_API_USES_CALLEE_POPS_CONVENTION && (callType) == FFICallTypeApi)'>
	^callType == FFICallTypeApi! !

!ThreadedFFIPlugin methodsFor: 'primitive support' stamp: 'eem 10/6/2010 17:39'!
isDirectAlien: oop
	self assert: (self isAlien: oop).
	^(self sizeField: oop) > 0! !

!ThreadedFFIPlugin methodsFor: 'marshalling-struct' stamp: 'eem 8/28/2020 13:16'!
isUnionSpec: specs OfLength: specSize StartingAt: startIndex
	"We can't easily distinguish union from structures with available flags.
	But we have a trick: a union should have one field size equal to its own size."
	| index spec unionSize thisSize |
	<var: #specs type: #'unsigned int*'>
	<inline: false>
	index := startIndex.
	spec := specs at: index.
	self assert: (spec bitAnd: FFIFlagPointer + FFIFlagAtomic + FFIFlagStructure) = FFIFlagStructure.
	unionSize := spec bitAnd: FFIStructSizeMask.
	[index := index + 1.
	index < specSize]
		whileTrue:
			[spec := specs at: index.
			spec = FFIFlagStructure
				ifTrue: [^false].
			thisSize := spec bitAnd: FFIStructSizeMask.
			thisSize = unionSize ifTrue: [^true].
			((spec bitAnd: FFIFlagPointer + FFIFlagStructure) = FFIFlagStructure)
				ifTrue:
					["Asking for alignment is a trick for skipping this sub structure/union"
					self alignmentOfStructSpec: specs OfLength: specSize StartingAt: (self addressOf: index put: [:v| index := v])]].
	^false! !

!ThreadedFFIPlugin methodsFor: 'simulation' stamp: 'eem 4/27/2020 13:40'!
morphIntoConcreteSubclass: aCoInterpreter
	<doNotGenerate>
	| concreteClass |
	concreteClass :=
		aCoInterpreter ISA caseOf: {
			[#X64]		->	[(Smalltalk platformName beginsWith: 'Win')
								ifTrue: [ThreadedX64Win64FFIPlugin]
								ifFalse: [ThreadedX64SysVFFIPlugin]].
			[#IA32]	->	[ThreadedIA32FFIPlugin].
			[#ARMv5]	->	[ThreadedARM32FFIPlugin].
			[#ARMv8]	->	[ThreadedARM64FFIPlugin] }
			otherwise: [self error: 'simulation not set up for this ISA'].
	"If the concreteClass has an initialize method, other than ThreadedFFIPlugin class>>initialize
	 then it needs to be run."
	((concreteClass class whichClassIncludesSelector: #initialize) inheritsFrom: self class class) ifTrue:
		[concreteClass initialize].
	concreteClass adoptInstance: self! !

!ThreadedFFIPlugin methodsFor: 'debugging' stamp: 'eem 4/28/2014 10:06'!
msg: s
	<var: #s type: 'char *'>
	self cCode: 'fprintf(stderr, "\n%s: %s", getModuleName(), s)'
		inSmalltalk: [super msg: s].
	^0! !

!ThreadedFFIPlugin methodsFor: 'marshalling' stamp: 'eem 11/3/2018 13:24'!
mustAlignStack
	"Many ABIs mandate a particular stack alignment greater than the natural word size.
	 If so, this macro will answer true.  See platforms/Cross/vm/sqCogStackAlignment.h
	 and class-side preambleCCode."
	<cmacro: '() MUST_ALIGN_STACK'>
	^false! !

!ThreadedFFIPlugin methodsFor: 'marshalling-struct' stamp: 'eem 11/5/2009 15:25'!
nonRegisterStructReturnIsViaImplicitFirstArgument
	"Answer if a struct returned in memory is returned to the
	 referent of a pointer passed as an implciit first argument.
	 It almost always is.  Subclasses can override if not."
	^true! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 08:26'!
primitiveCDataModel
	"Two forms of C Data Model infomation.
	 With 0 arguments answer the string naming the C data model, LP32, LP64, LLP64, etc.
	 WIth 1 argument, which must be a ByteArray of at least 9 elements, answer the sizes of
	 char, short, int, long, long long, wchar_t, float, double, void *."
	<export: true flags: #FastCPrimitiveFlag>
	| errorCode model |
	interpreterProxy methodArgumentCount = 1 ifTrue:
		[| sizes |
		sizes := interpreterProxy stackValue: 0.
		((interpreterProxy isBytes: sizes)
		 and: [(interpreterProxy slotSizeOf: sizes) = 9]) ifFalse:
			[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
		 (self cCoerceSimple: (interpreterProxy firstIndexableField: sizes) to: #'char *')
			at: 0 put: (self sizeof: #char);
			at: 1 put: (self sizeof: #short);
			at: 2 put: (self sizeof: #int);
			at: 3 put: (self sizeof: #long);
			at: 4 put: (self sizeof: #'long long');
			at: 5 put: (self sizeof: #wchar_t);
			at: 6 put: (self sizeof: #float);
			at: 7 put: (self sizeof: #double);
			at: 8 put: (self sizeof: #'void *').
		^interpreterProxy methodReturnValue: sizes].

	interpreterProxy methodArgumentCount = 0 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].

	"Attempt to identify the programming model:
					   LP32    ILP32    LLP64   LP64    ILP64  SILP64(unidentified)
		char			 8		 8		  8		 8		 8		 8

		short			16		16		16		16		16		64

		int				16		32		32		32		64		64

		long			32		32		32 		64		64		64

		long long		64		64		64		64		64		64

		pointer			32		32		64		64		64		64"

	errorCode := 0. "Set bit 0 if char is wrong, bit 1 if short is wrong, 2 for int, 3 for long, 4 for long long, 5 for void *"
	(self sizeof: #char) ~= 1 ifTrue:
		[errorCode := errorCode + 1].
	(self sizeof: #short) ~= 2 ifTrue: "N.B. SILP64 exists on Cray supercomputers; we don't care..."
		[errorCode := errorCode + 2].
	(self sizeof: #'long long') ~= 8 ifTrue:
		[errorCode := errorCode + 16].

	(self sizeof: #'void *') = 8 ifTrue: "LP64 LLP64 ILP64"
		[(self sizeof: #int) = 8 ifTrue: "ILP64"
			[(self sizeof: #long) = 8
				ifTrue: [model := 'ILP64']
				ifFalse: [errorCode := errorCode + 8]].
		 (self sizeof: #int) = 4 ifTrue: "LP64 or LLP64"
			[(self sizeof: #long) = 8 ifTrue: "LP64"
				[model := 'LP64'].
			 (self sizeof: #long) = 4 ifTrue: "LLP64"
				[model := 'LLP64'].
			 ((self sizeof: #long) ~= 8 and: [(self sizeof: #long) ~= 4]) ifTrue:
				[errorCode := errorCode + 8]].
		 ((self sizeof: #int) ~= 8 and: [(self sizeof: #int) ~= 4]) ifTrue:
			[errorCode := errorCode + 4]].

	(self sizeof: #'void *') = 4 ifTrue: "LP32 ILP32"
		[(self sizeof: #long) ~= 4 ifTrue:
			[errorCode := errorCode + 8].
		 (self sizeof: #int) = 4 ifTrue: "ILP32"
			[model := 'ILP32'].
		 (self sizeof: #int) = 2 ifTrue: "LP32"
			[model := 'LP32'].
		 ((self sizeof: #int) ~= 4 and: [(self sizeof: #int) ~= 2]) ifTrue:
			[errorCode := errorCode + 4]].

	((self sizeof: #'void *') ~= 8 and: [(self sizeof: #'void *') ~= 4]) ifTrue:
		[errorCode := errorCode + 32].

	errorCode ~= 0 ifTrue:
		[^interpreterProxy primitiveFailForOSError: errorCode].
	model ifNil:
		[^interpreterProxy primitiveFailFor: PrimErrNotFound].
	interpreterProxy methodReturnString: model

"Screed for testing
	| proxy plugin |
	proxy := InterpreterProxy new.
	plugin := ThreadedFFIPluginPartialSimulator new.
	plugin sizes: (Dictionary newFromPairs: #(char 1 short 2 int 4 long 4 #'long long' 8 #'void *' 8  #'void *' 4 float 4 double 8 wchar_t 4)).
	plugin sizes: (Dictionary newFromPairs: #(char 1 short 2 int 2 long 4 #'long long' 8 #'void *' 4 float 4 double 8 wchar_t 4)).
	plugin instVarNamed: 'interpreterProxy' put: proxy.
	proxy synthesizeStackFor: plugin with: (Array with: (ByteArray new: 9)).
	plugin primitiveCDataModel.
	^proxy stackValue: 0"! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 11/18/2012 19:42'!
primitiveCallout
	"IMPORTANT: IF YOU CHANGE THE NAME OF THIS METHOD YOU MUST CHANGE
		Interpreter>>primitiveCalloutToFFI
	TO REFLECT THE CHANGE."

	"Perform a function call to a foreign function.
	Only invoked from method containing explicit external call spec."
 
	<returnTypeC: #void>
	<export: true>
	| meth externalFunction |
	meth := interpreterProxy primitiveMethod.
	(interpreterProxy literalCountOf: meth) > 0 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadMethod].
	externalFunction := interpreterProxy literal: 0 ofMethod: meth.
	self cppIf: COGMTVM
		ifTrue:
			[ | nArgs retryCount result |
			nArgs := interpreterProxy methodArgumentCount.
			retryCount := 0.
			[result := self ffiCall: externalFunction ArgArrayOrNil: nil NumArgs: nArgs.
			 result = PrimErrObjectMayMove
			 and: [(retryCount := retryCount + 1) <= (nArgs + 1)]] whileTrue:
				[interpreterProxy tenuringIncrementalGC]]
		ifFalse:
			[self ffiCall: externalFunction ArgArrayOrNil: nil NumArgs: interpreterProxy methodArgumentCount].
	^0! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 11/18/2012 19:42'!
primitiveCalloutWithArgs
	"Perform a function call to a foreign function.
	 Only invoked from ExternalFunction>>invokeWithArguments:"

	<returnTypeC: #void>
	<export: true>
	| externalFunction argArray nArgs |
	interpreterProxy methodArgumentCount = 1 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	self cppIf: COGMTVM
		ifTrue:
			[| retryCount result |
			retryCount := 0.
			[externalFunction := interpreterProxy stackValue: 1.
			 argArray := interpreterProxy stackValue: 0.
			 (interpreterProxy isArray: argArray) ifFalse:
				[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
			 nArgs := interpreterProxy slotSizeOf: argArray.
			 result := self ffiCall: externalFunction ArgArrayOrNil: argArray NumArgs: nArgs.
			 result = PrimErrObjectMayMove
			  and: [(retryCount := retryCount + 1) <= (nArgs + 1)]] whileTrue:
				[interpreterProxy tenuringIncrementalGC]]
		ifFalse:
			[externalFunction := interpreterProxy stackValue: 1.
			argArray := interpreterProxy stackValue: 0.
			(interpreterProxy isArray: argArray) ifFalse:
				[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
			nArgs := interpreterProxy slotSizeOf: argArray.
			self ffiCall: externalFunction ArgArrayOrNil: argArray NumArgs: nArgs].
	^0! !

!ThreadedFFIPlugin methodsFor: 'primitives - surfaces' stamp: 'nice 11/12/2016 02:53'!
primitiveCreateManualSurface
	"arguments: name(type, stack offset)
		width(Integer, 4)
		height(Integer, 3)
		rowPitch(Integer, 2)
		depth(Integer, 1)
		isMSB(Boolean, 0)"
	| width height rowPitch depth isMSB result |
	<export: true>
	
	interpreterProxy methodArgumentCount = 5 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].
	width := interpreterProxy stackIntegerValue: 4.
	height := interpreterProxy stackIntegerValue: 3.
	rowPitch := interpreterProxy stackIntegerValue: 2.
	depth := interpreterProxy stackIntegerValue: 1.
	isMSB := interpreterProxy stackObjectValue: 0.
	isMSB := interpreterProxy booleanValueOf: isMSB. 
	interpreterProxy failed ifTrue: [^nil].
	
	result := self cCode: 'createManualSurface(width, height, rowPitch, depth, isMSB)'
				inSmalltalk: [self create: width Man: height ual: rowPitch Surf: depth ace: isMSB].
	result < 0 ifTrue:
		[^interpreterProxy primitiveFail].
	result := interpreterProxy signed32BitIntegerFor: result.
	^interpreterProxy pop: 6 thenPush: result
	! !

!ThreadedFFIPlugin methodsFor: 'primitives - surfaces' stamp: 'eem 10/16/2018 09:15'!
primitiveDestroyManualSurface
	| surfaceID |
	<export: true>
	interpreterProxy methodArgumentCount = 1 ifFalse:
		[^interpreterProxy primitiveFail].
	surfaceID := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifFalse:
		[(self destroyManualSurface: surfaceID) = 0
			ifTrue: [interpreterProxy primitiveFail]
			ifFalse: [interpreterProxy pop: 1]]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 08:54'!
primitiveExternalAddressAsInteger
	"Answer the address of a 4 byte or 8 byte ExternalAddress."
	<export: true flags: #FastCPrimitiveFlag>
	| rcvr size |
	rcvr := interpreterProxy stackValue: 0.
	size := interpreterProxy byteSizeOf: rcvr.
	size = 8 ifTrue:
		[^interpreterProxy methodReturnValue:
			(interpreterProxy positive64BitIntegerFor:
				((self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'usqLong *') at: 0))].
	size = 4 ifTrue:
		[^interpreterProxy methodReturnValue:
			(interpreterProxy positive32BitIntegerFor:
				((self cCoerce: (interpreterProxy firstIndexableField: rcvr) to: #'unsigned int *') at: 0))].
	interpreterProxy primitiveFailFor: PrimErrBadReceiver! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 8/13/2021 12:40'!
primitiveExternalAddressFromInteger
	"Answer a 4 byte or 8 byte ExternalAddress with value of the argument."
	<export: true flags: #FastCPrimitiveFlag>
	| value |
	value := interpreterProxy positiveMachineIntegerValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(interpreterProxy instantiateClass: interpreterProxy classExternalAddress indexableSize: BytesPerWord)
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrNoMemory]
		ifNotNil:
			[:address|
			(self cCoerce: (interpreterProxy firstIndexableField: address) to: #'usqIntptr_t *') at: 0 put: value.
			^interpreterProxy methodReturnValue: address]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'nice 7/13/2016 18:35'!
primitiveFFIAllocate
	"Primitive. Allocate an object on the external heap."
	| byteSize addr oop ptr |
	<export: true>
	<inline: false>
	<var: #ptr type: #'sqIntptr_t *'>
	byteSize := interpreterProxy stackIntegerValue: 0.
	interpreterProxy failed ifTrue:
		[^nil].
	addr := self ffiAlloc: byteSize.
	addr = 0 ifTrue:
		[^interpreterProxy primitiveFail].
	oop := interpreterProxy 
			instantiateClass: interpreterProxy classExternalAddress 
			indexableSize: (self sizeof: #'sqIntptr_t').
	ptr := interpreterProxy firstIndexableField: oop.
	ptr at: 0 put: addr.
	^interpreterProxy pop: 2 thenPush: oop! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 8/1/2021 11:02'!
primitiveFFIDoubleAt
	"Answer a 64-bit IEEE double the given byte offset."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| byteOffset rcvr floatValue |
	<var: #floatValue type: #double>
	byteOffset := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isIntegerObject: byteOffset) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #double))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			self memcpy: (self addressOf: floatValue) _: addr _: (self sizeof: #double).
			^interpreterProxy methodReturnFloat: floatValue]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 8/1/2021 12:38'!
primitiveFFIDoubleAtPut
	"Store a 64-bit IEEE double the given byte offset."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| byteOffset rcvr floatValue valueOop |
	<var: #floatValue type: #double>
	valueOop := interpreterProxy stackValue: 0.
	(interpreterProxy isFloatObject: valueOop)
		ifTrue: [floatValue := self cCoerce: (interpreterProxy floatValueOf: valueOop) to: #double]
		ifFalse:
			[(interpreterProxy isIntegerObject: valueOop)
				ifTrue: [floatValue := self cCoerce: (interpreterProxy integerValueOf: valueOop) to: #double]
				ifFalse: [^interpreterProxy primitiveFailFor: PrimErrBadArgument]].
	byteOffset := interpreterProxy stackValue: 1.
	(interpreterProxy isIntegerObject: byteOffset) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	rcvr := interpreterProxy stackValue: 2.
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #double))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			self memcpy: addr _: (self addressOf: floatValue put: [:v| floatValue := Float fromIEEE64Bit: v]) _: (self sizeof: #double).
			^interpreterProxy methodReturnValue: valueOop]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 8/1/2021 12:32'!
primitiveFFIFloatAt
	"Answer a 32-bit IEEE float the given byte offset."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| byteOffset rcvr floatValue |
	<var: #floatValue type: #float>
	byteOffset := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isIntegerObject: byteOffset) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #float))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			self memcpy: (self addressOf: floatValue put: [:v| floatValue := Float fromIEEE32Bit: v]) _: addr _: (self sizeof: #float).
			^interpreterProxy methodReturnFloat: floatValue]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 8/1/2021 11:03'!
primitiveFFIFloatAtPut
	"Store a 32-bit IEEE float the given byte offset."
	<export: true flags: #(FastCPrimitiveFlag FastCPrimitiveAlignForFloatsFlag)>
	| byteOffset rcvr floatValue valueOop |
	<var: #floatValue type: #float>
	valueOop := interpreterProxy stackValue: 0.
	(interpreterProxy isFloatObject: valueOop)
		ifTrue: [floatValue := self cCoerce: (interpreterProxy floatValueOf: valueOop) to: #float]
		ifFalse:
			[(interpreterProxy isIntegerObject: valueOop)
				ifTrue: [floatValue := self cCoerce: (interpreterProxy integerValueOf: valueOop) to: #float]
				ifFalse: [^interpreterProxy primitiveFailFor: PrimErrBadArgument]].
	byteOffset := interpreterProxy stackValue: 1.
	(interpreterProxy isIntegerObject: byteOffset) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	rcvr := interpreterProxy stackValue: 2.
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #float))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			self memcpy: addr _: (self addressOf: floatValue) _: (self sizeof: floatValue).
			^interpreterProxy methodReturnValue: valueOop]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 8/28/2020 10:54'!
primitiveFFIFree
	"Primitive. Free the object pointed to on the external heap."
	| addr oop ptr |
	<export: true>
	<inline: false>
	<var: #ptr type: #'sqIntptr_t *'>
	oop := interpreterProxy stackObjectValue: 0.
	((interpreterProxy fetchClassOf: oop) = interpreterProxy classExternalAddress
	 and: [(interpreterProxy byteSizeOf: oop) = (self sizeof: #'sqIntptr_t')]) ifFalse:
		[^interpreterProxy primitiveFail].
	ptr := interpreterProxy firstIndexableField: oop.
	addr := ptr at: 0.
	"Don't you dare to free Squeak's memory!!"
	(addr = 0
	 or: [(addr asUnsignedIntegerPtr bitAnd: (self sizeof: #'sqIntptr_t') - 1) ~= 0
	 or: [interpreterProxy isInMemory: addr asUnsignedIntegerPtr]]) ifTrue:
		[^interpreterProxy primitiveFail].
	self ffiFree: addr.
	^ptr at: 0 put: 0 "cleanup"! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 08:44'!
primitiveFFIGetLastError
	"Primitive. Return the error code from a failed call to the foreign function interface.
	 This is for backwards-compatibility.  Thread-safe access to the error code is via the
	 primitive error code."
	<export: true flags: #FastCPrimitiveFlag>
	interpreterProxy methodReturnInteger: ffiLastError! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 09:23'!
primitiveFFIIntegerAt
	"Answer a (signed or unsigned) n byte integer from the given byte offset
	 in the receiver, using the platform's endianness."
	<export: true flags: #FastCPrimitiveFlag>
	| isSigned byteSize byteOffset rcvr value mask valueOop |
	<var: 'value' type: #usqLong>
	<var: 'mask' type: #usqLong>
	<export: true>
	<inline: false>
	isSigned := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
	byteSize := interpreterProxy stackIntegerValue: 1.
	byteOffset := interpreterProxy stackIntegerValue: 2.
	rcvr := interpreterProxy stackObjectValue: 3.
	interpreterProxy failed ifTrue:[^0].
	(byteOffset > 0
	 and: [(byteSize between: 1 and: 8)
	 and: [(byteSize bitAnd: byteSize - 1) = 0 "a.k.a. isPowerOfTwo"]]) ifFalse:
		[^interpreterProxy primitiveFail].
	(self ffiAddressOf: rcvr startingAt: byteOffset size: byteSize)
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			byteSize <= 2
				ifTrue:
					[byteSize = 1
						ifTrue: [value := self cCoerceSimple: (interpreterProxy byteAt: addr) to: #'unsigned char']
						ifFalse: [value := self cCoerceSimple: (interpreterProxy unalignedShortAt: addr) to: #'unsigned short']]
				ifFalse:
					[byteSize = 4
						ifTrue: [value := self cCoerceSimple: (interpreterProxy unalignedLong32At: addr) to: #'unsigned int']
						ifFalse: [value := interpreterProxy unalignedLong64At: addr]].
			byteSize < BytesPerWord
				ifTrue:
					[isSigned ifTrue: "sign extend value"
						[mask := 1 asUnsignedLongLong << (byteSize * 8 - 1).
						value := (value bitAnd: mask-1) - (value bitAnd: mask)].
					 "note: byte/short (&long if BytesPerWord=8) never exceed SmallInteger range"
					 valueOop := interpreterProxy integerObjectOf: value]
				ifFalse: "general 64 bit integer; note these never fail"
					[isSigned
						ifTrue:
							[byteSize < 8 ifTrue: "sign extend value"
								[mask := 1 asUnsignedLongLong << (byteSize * 8 - 1).
								value := (value bitAnd: mask-1) - (value bitAnd: mask)].
							 self cCode: [] inSmalltalk:
								[(byteSize = 8 and: [(value bitShift: -56) >= 128]) ifTrue:
									[value := value - (1 bitShift: 64)]].
							 valueOop := interpreterProxy signed64BitIntegerFor: value]
						ifFalse:[valueOop := interpreterProxy positive64BitIntegerFor: value]].
			^interpreterProxy methodReturnValue: valueOop]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 09:22'!
primitiveFFIIntegerAtPut
	"Store a (signed or unsigned) n byte integer at the given byte offset
	 in the receiver, using the platform's endianness."
	<export: true flags: #FastCPrimitiveFlag>
	| isSigned byteSize byteOffset rcvr value max valueOop |
	<var: 'value' type: #sqLong>
	<var: 'max' type: #sqLong>
	<export: true>
	<inline: false>
	isSigned := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
	byteSize := interpreterProxy stackIntegerValue: 1.
	valueOop := interpreterProxy stackValue: 2.
	byteOffset := interpreterProxy stackIntegerValue: 3.
	rcvr := interpreterProxy stackObjectValue: 4.
	interpreterProxy failed ifTrue:[^0].
	(byteOffset > 0
	 and: [(byteSize between: 1 and: 8)
	 and: [(byteSize bitAnd: byteSize - 1) = 0 "a.k.a. isPowerOfTwo"]]) ifFalse:
		[^interpreterProxy primitiveFail].
	(self ffiAddressOf: rcvr startingAt: byteOffset size: byteSize)
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			isSigned 
				ifTrue:[value := interpreterProxy signed64BitValueOf: valueOop]
				ifFalse:[value := interpreterProxy positive64BitValueOf: valueOop].
			interpreterProxy failed ifTrue:[^0].
			byteSize < 8 ifTrue:
				[isSigned
					ifTrue:
						[max := 1 asUnsignedLongLong << (8 * byteSize - 1).
						(value >= (0 - max) and: [value < max]) ifFalse: [^interpreterProxy primitiveFail]]
					ifFalse:
						[value asUnsignedLongLong < (1 asUnsignedLongLong << (8 * byteSize)) ifFalse: [^interpreterProxy primitiveFail]]].
			byteSize <= 2
				ifTrue:
					[byteSize = 1
						ifTrue: [interpreterProxy byteAt: addr put: value]
						ifFalse: [interpreterProxy unalignedShortAt: addr put: value]]
				ifFalse:
					[byteSize = 4
						ifTrue: [interpreterProxy unalignedLong32At: addr put: value]
						ifFalse: [interpreterProxy unalignedLong64At: addr put: value]].
			^interpreterProxy methodReturnValue: valueOop]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 11/24/2014 10:59'!
primitiveForceLoad
	"Primitive. Force loading the receiver (an instance of ExternalLibrary)."
	| rcvr moduleHandlePtr moduleHandle ffiModuleName ptr |
	<export: true>
	<inline: false>
	<var: #ptr type: #'sqInt *'>
	interpreterProxy methodArgumentCount = 0 ifFalse:
		[^interpreterProxy primitiveFail].
	rcvr := interpreterProxy stackValue: 0.
	(interpreterProxy is: rcvr KindOfClass: interpreterProxy classExternalLibrary) ifFalse:
		[^self ffiFail: FFIErrorBadExternalLibrary].
	moduleHandlePtr := interpreterProxy fetchPointer: 0 ofObject: rcvr.
	moduleHandle := self ffiContentsOfHandle: moduleHandlePtr errCode: FFIErrorBadExternalLibrary.
	interpreterProxy failed ifTrue:
		[^0].
	ffiModuleName := interpreterProxy fetchPointer: 1 ofObject: rcvr.
	(interpreterProxy isBytes: ffiModuleName) ifFalse:
		[^self ffiFail: FFIErrorBadExternalLibrary].
	moduleHandle := (interpreterProxy
						ioLoadModule: (interpreterProxy firstIndexableField: ffiModuleName) asInteger
						OfLength: (interpreterProxy byteSizeOf: ffiModuleName)) asInteger.
	interpreterProxy failed ifTrue:
		[^self ffiFail: FFIErrorModuleNotFound]. "failed"
	"and store back"
	ptr := interpreterProxy firstIndexableField: moduleHandlePtr.
	ptr at: 0 put: moduleHandle.
	^0 "done"! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 12/14/2016 12:57'!
primitiveLoadSymbolFromModule
	"Attempt to find the address of a symbol in a loaded library.
	loadSymbol: aSymbol fromModule: moduleName
		<primitive: 'primitiveLoadSymbolFromModule' error: errorCode module: 'SqueakFFIPrims'>
	"
	<export: true>

	| symbol module moduleHandle address oop ptr |

	<var: #address type: #'void *'>
	<var: #ptr type: #'void **'>
	
	interpreterProxy methodArgumentCount = 2 ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadNumArgs].	

	module := interpreterProxy stackValue: 0.
	symbol := interpreterProxy stackValue: 1.

	moduleHandle := module ~= interpreterProxy nilObject ifTrue:
						[self ffiLoadCalloutModule: module].
	interpreterProxy failed ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNotFound].
	address := interpreterProxy
					ioLoadSymbol: (self cCoerce: (interpreterProxy firstIndexableField: symbol) to: #sqInt)
					OfLength: (interpreterProxy byteSizeOf: symbol)
					FromModule: moduleHandle.
	(interpreterProxy failed
	 or: [address = 0]) ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNotFound].
	
	oop := interpreterProxy 
			instantiateClass: interpreterProxy classExternalAddress 
			indexableSize: (self sizeof: #'void *').
	ptr := interpreterProxy firstIndexableField: oop.
	ptr at: 0 put: address.
	
	^interpreterProxy methodReturnValue: oop! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 6/28/2016 12:57'!
primitiveLogCallsTo
	"Enable logging of FFI calls by providing it with a log file name."
	| logFile ok |
	<export: true>
	interpreterProxy methodArgumentCount = 1 
		ifFalse:[^interpreterProxy primitiveFail].
	logFile := interpreterProxy stackValue: 0.
	logFile = interpreterProxy nilObject ifTrue:[ "disable logging"
		ok := self ffiLogFileName: nil OfLength: 0.
		ok ifFalse:[^interpreterProxy primitiveFail].
		ffiLogEnabled := false.
	] ifFalse:[ "enable logging"
		(interpreterProxy isBytes: logFile) ifFalse:[^interpreterProxy primitiveFail].
		ok := self ffiLogFileName: (interpreterProxy firstIndexableField: logFile)
					OfLength: (interpreterProxy byteSizeOf: logFile).
		ok ifFalse:[^interpreterProxy primitiveFail].
		ffiLogEnabled := true.
	].
	^interpreterProxy pop: 1. "pop arg; return rcvr"
! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'mt 8/16/2021 10:50'!
primitivePluginVersion
	"Answer the plugins current version to ensure compatibility with data structures shared between plugin an image code such as:
		- Type codes in FFIConstants
		- Known classes in the special-objects array"
	<export: true flags: #FastCPrimitiveFlag>
	^ interpreterProxy methodReturnInteger: 1! !

!ThreadedFFIPlugin methodsFor: 'primitives - surfaces' stamp: 'nice 7/8/2016 02:05'!
primitiveSetManualSurfacePointer
	"Create a 'manual surface' data-structure.  See the ExternalForm class in the FFI package for example usage."
	"arguments: name(type, stack offset)
		surfaceID(Integer, 1)
		ptr(uint32/uint64, 0)"
	| surfaceID ptr result |
	<export: true>
	<var: #ptr type: #'usqIntptr_t'>
	
	interpreterProxy methodArgumentCount = 2 ifFalse: [^interpreterProxy primitiveFail].
	surfaceID := interpreterProxy stackIntegerValue: 1.
	ptr := interpreterProxy positiveMachineIntegerValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifTrue: [^nil].

	self touch: surfaceID; touch: ptr.
	
	result := self setManualSurface: surfaceID Pointer: ptr asVoidPointer.
	result = 0 ifTrue: [^interpreterProxy primitiveFail].
	^interpreterProxy pop: 2
	! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 09:44'!
primitiveSignedInt16At
	"Answer the signed 16-bit integer starting at the given byte offset (native endian)."
	<export: true flags: #FastCPrimitiveFlag>
	| byteOffset rcvr |

	byteOffset := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isIntegerObject: byteOffset) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #short))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			^interpreterProxy methodReturnInteger: (self cCoerce: (interpreterProxy unalignedShortAt: addr) to: #'signed short')]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 11:00'!
primitiveSignedInt16AtPut
	"Store the signed 16-bit integer starting at the given byte offset (native endian)."
	<export: true flags: #FastCPrimitiveFlag>
	| valueArg value byteOffset rcvr |
	valueArg := interpreterProxy stackValue: 0.
	byteOffset := interpreterProxy stackValue: 1.
	rcvr := interpreterProxy stackValue: 2.
	((interpreterProxy isIntegerObject: valueArg)
	and: [((value := interpreterProxy integerValueOf: valueArg) between: -32768 and: 32767)
	and: [interpreterProxy isIntegerObject: byteOffset]]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #short))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			interpreterProxy unalignedShortAt: addr put: value.
			^interpreterProxy methodReturnValue: valueArg]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 09:44'!
primitiveSignedInt32At
	"Answer the signed 16-bit integer starting at the given byte offset (native endian)."
	<export: true flags: #FastCPrimitiveFlag>
	| byteOffset rcvr |

	byteOffset := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isIntegerObject: byteOffset) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #int))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			^interpreterProxy methodReturnValue:
				(interpreterProxy signed32BitIntegerFor: (self cCoerce: (interpreterProxy unalignedLong32At: addr) to: #'signed int'))]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 10:59'!
primitiveSignedInt32AtPut
	"Store the signed 32-bit integer starting at the given byte offset (native endian)."
	<export: true flags: #FastCPrimitiveFlag>
	| valueArg value byteOffset rcvr |
	valueArg := interpreterProxy stackValue: 0.
	byteOffset := interpreterProxy stackValue: 1.
	rcvr := interpreterProxy stackValue: 2.
	value := interpreterProxy signed32BitValueOf: valueArg.
	(interpreterProxy failed not
	and: [interpreterProxy isIntegerObject: byteOffset]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #int))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			interpreterProxy unalignedLong32At: addr put: value.
			^interpreterProxy methodReturnValue: valueArg]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 09:44'!
primitiveSignedInt64At
	"Answer the signed 64-bit integer starting at the given byte offset (native endian)."
	<export: true flags: #FastCPrimitiveFlag>
	| byteOffset rcvr |

	byteOffset := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isIntegerObject: byteOffset) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #sqLong))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			^interpreterProxy methodReturnValue:
				(interpreterProxy signed64BitIntegerFor: (self cCoerce: (interpreterProxy unalignedLong64At: addr) to: #sqLong))]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 11:00'!
primitiveSignedInt64AtPut
	"Store the signed 64-bit integer starting at the given byte offset (native endian)."
	<export: true flags: #FastCPrimitiveFlag>
	| valueArg value byteOffset rcvr |
	valueArg := interpreterProxy stackValue: 0.
	byteOffset := interpreterProxy stackValue: 1.
	rcvr := interpreterProxy stackValue: 2.
	value := interpreterProxy signed64BitValueOf: valueArg.
	(interpreterProxy failed not
	and: [interpreterProxy isIntegerObject: byteOffset]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #sqLong))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			interpreterProxy unalignedLong64At: addr put: value.
			^interpreterProxy methodReturnValue: valueArg]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 09:44'!
primitiveSignedInt8At
	"Answer the signed 8-bit integer starting at the given byte offset (native endian)."
	<export: true flags: #FastCPrimitiveFlag>
	| byteOffset rcvr |

	byteOffset := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isIntegerObject: byteOffset) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #char))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			^interpreterProxy methodReturnInteger: (self cCoerce: (interpreterProxy byteAt: addr) to: #'signed char')]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 10:54'!
primitiveSignedInt8AtPut
	"Store the signed 8-bit integer starting at the given byte offset."
	<export: true flags: #FastCPrimitiveFlag>
	| valueArg value byteOffset rcvr |
	valueArg := interpreterProxy stackValue: 0.
	byteOffset := interpreterProxy stackValue: 1.
	rcvr := interpreterProxy stackValue: 2.
	((interpreterProxy isIntegerObject: valueArg)
	and: [((value := interpreterProxy integerValueOf: valueArg) between: -128 and: 127)
	and: [interpreterProxy isIntegerObject: byteOffset]]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #char))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			interpreterProxy byteAt: addr put: value.
			^interpreterProxy methodReturnValue: valueArg]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'mt 8/13/2021 14:56'!
primitiveStructureElementAlignment
	"Answer the alignment of an element of an atomic type, or a structure, within a structure on the current platform."
	<export: true flags: #FastCPrimitiveFlag>
	| typeCode alignment |
	<var: 'alignment' type: #'void *'>
	typeCode := interpreterProxy stackValue: 0.
	((interpreterProxy isIntegerObject: typeCode)
	 and: [((typeCode := interpreterProxy integerValueOf: typeCode) between: FFITypeUnsignedInt8 and: FFITypeDoubleFloat)
		or: [typeCode = FFIFlagStructure]]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].

	alignment := typeCode
					caseOf: {
						[FFITypeUnsignedInt8]			-> [self structOffsetOf: 'structByte *' atomicTypeCode: FFITypeUnsignedInt8].
						[FFITypeSignedInt8]			-> [self structOffsetOf: 'structByte *' atomicTypeCode: FFITypeSignedInt8].
						[FFITypeUnsignedInt16]		-> [self structOffsetOf: 'structShort *' atomicTypeCode: FFITypeUnsignedInt16].
						[FFITypeSignedInt16]			-> [self structOffsetOf: 'structShort *' atomicTypeCode: FFITypeSignedInt16].
						[FFITypeUnsignedInt32]			-> [self structOffsetOf: 'structInt *' atomicTypeCode: FFITypeUnsignedInt32].
						[FFITypeSignedInt32]				-> [self structOffsetOf: 'structInt *' atomicTypeCode: FFITypeSignedInt32].
						[FFITypeUnsignedInt64]	-> [self structOffsetOf: 'structLongLong *' atomicTypeCode: FFITypeUnsignedInt64].
						[FFITypeSignedInt64]		-> [self structOffsetOf: 'structLongLong *' atomicTypeCode: FFITypeSignedInt64].
						[FFITypeSingleFloat]			-> [self structOffsetOf: 'structFloat *' atomicTypeCode: FFITypeSingleFloat].
						[FFITypeDoubleFloat]			-> [self structOffsetOf: 'structDouble *' atomicTypeCode: FFITypeDoubleFloat].
					}
					otherwise: [self structOffsetOf: 'structStruct *' atomicTypeCode: FFIFlagStructure].
	^interpreterProxy methodReturnInteger: alignment asUnsignedIntegerPtr! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 09:44'!
primitiveUnsignedInt16At
	"Answer the unsigned 16-bit integer starting at the given byte offset (native endian)."
	<export: true flags: #FastCPrimitiveFlag>
	| byteOffset rcvr |

	byteOffset := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isIntegerObject: byteOffset) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #short))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			^interpreterProxy methodReturnInteger: (self cCoerce: (interpreterProxy unalignedShortAt: addr) to: #'unsigned short')]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 11:02'!
primitiveUnsignedInt16AtPut
	"Store the signed 16-bit integer starting at the given byte offset (native endian)."
	<export: true flags: #FastCPrimitiveFlag>
	| valueArg value byteOffset rcvr |
	valueArg := interpreterProxy stackValue: 0.
	byteOffset := interpreterProxy stackValue: 1.
	rcvr := interpreterProxy stackValue: 2.
	((interpreterProxy isIntegerObject: valueArg)
	and: [((value := interpreterProxy integerValueOf: valueArg) between: 0 and: 65535)
	and: [interpreterProxy isIntegerObject: byteOffset]]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #short))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			interpreterProxy unalignedShortAt: addr put: value.
			^interpreterProxy methodReturnValue: valueArg]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 09:44'!
primitiveUnsignedInt32At
	"Answer the unsigned 32-bit integer starting at the given byte offset (native endian)."
	<export: true flags: #FastCPrimitiveFlag>
	| byteOffset rcvr |

	byteOffset := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isIntegerObject: byteOffset) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #int))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			^interpreterProxy methodReturnValue:
				(interpreterProxy positive32BitIntegerFor: (self cCoerce: (interpreterProxy unalignedLong32At: addr) to: #'unsigned int'))]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 11:02'!
primitiveUnsignedInt32AtPut
	"Store the unsigned 32-bit integer starting at the given byte offset (native endian)."
	<export: true flags: #FastCPrimitiveFlag>
	| valueArg value byteOffset rcvr |
	valueArg := interpreterProxy stackValue: 0.
	byteOffset := interpreterProxy stackValue: 1.
	rcvr := interpreterProxy stackValue: 2.
	value := interpreterProxy positive32BitValueOf: valueArg.
	(interpreterProxy failed not
	and: [interpreterProxy isIntegerObject: byteOffset]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #int))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			interpreterProxy unalignedLong32At: addr put: value.
			^interpreterProxy methodReturnValue: valueArg]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 09:44'!
primitiveUnsignedInt64At
	"Answer the unsigned 64-bit integer starting at the given byte offset (native endian)."
	<export: true flags: #FastCPrimitiveFlag>
	| byteOffset rcvr |

	byteOffset := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isIntegerObject: byteOffset) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #usqLong))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			^interpreterProxy methodReturnValue:
				(interpreterProxy positive64BitIntegerFor: (self cCoerce: (interpreterProxy unalignedLong64At: addr) to: #usqLong))]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 11:01'!
primitiveUnsignedInt64AtPut
	"Store the unsigned 64-bit integer starting at the given byte offset (native endian)."
	<export: true flags: #FastCPrimitiveFlag>
	| valueArg value byteOffset rcvr |
	valueArg := interpreterProxy stackValue: 0.
	byteOffset := interpreterProxy stackValue: 1.
	rcvr := interpreterProxy stackValue: 2.
	value := interpreterProxy positive64BitValueOf: valueArg.
	(interpreterProxy failed not
	and: [interpreterProxy isIntegerObject: byteOffset]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #sqLong))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			interpreterProxy unalignedLong64At: addr put: value.
			^interpreterProxy methodReturnValue: valueArg]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 09:44'!
primitiveUnsignedInt8At
	"Answer the unsigned 8-bit integer starting at the given byte offset (native endian)."
	<export: true flags: #FastCPrimitiveFlag>
	| byteOffset rcvr |

	byteOffset := interpreterProxy stackValue: 0.
	rcvr := interpreterProxy stackValue: 1.
	(interpreterProxy isIntegerObject: byteOffset) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #char))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			^interpreterProxy methodReturnInteger: (self cCoerce: (interpreterProxy byteAt: addr) to: #'unsigned char')]! !

!ThreadedFFIPlugin methodsFor: 'primitives' stamp: 'eem 7/26/2021 10:53'!
primitiveUnsignedInt8AtPut
	"Store the unsigned 8-bit integer starting at the given byte offset."
	<export: true flags: #FastCPrimitiveFlag>
	| valueArg value byteOffset rcvr |
	valueArg := interpreterProxy stackValue: 0.
	byteOffset := interpreterProxy stackValue: 1.
	rcvr := interpreterProxy stackValue: 2.
	((interpreterProxy isIntegerObject: valueArg)
	and: [((value := interpreterProxy integerValueOf: valueArg) between: 0 and: 255)
	and: [interpreterProxy isIntegerObject: byteOffset]]) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	(self ffiAddressOf: rcvr startingAt: (interpreterProxy integerValueOf: byteOffset) size: (self sizeof: #char))
		ifNil: [^interpreterProxy primitiveFailFor: PrimErrBadIndex]
		ifNotNil:
			[:addr|
			interpreterProxy byteAt: addr put: value.
			^interpreterProxy methodReturnValue: valueArg]! !

!ThreadedFFIPlugin methodsFor: 'marshalling-struct' stamp: 'nice 1/30/2020 21:43'!
returnStructInRegisters: calloutState
	"Answer if struct result is returned in registers or not.
	Use the OS specific encoding stored in structReturnType.
	Since it is OS dependent, leave the responsibility to subclass"
	<var: #calloutState type: #'CalloutState *'>
	^self subclassResponsibility! !

!ThreadedFFIPlugin methodsFor: 'simulation' stamp: 'eem 12/21/2017 14:55'!
setInterpreter: anInterpreter
	"Initialization of the plugin in the simulator.
	 The real routine is in the superclass."
	self cCode: []
		inSmalltalk: [self class == thisContext method methodClass ifTrue:
						[self morphIntoConcreteSubclass: anInterpreter]].
	^super setInterpreter: anInterpreter! !

!ThreadedFFIPlugin methodsFor: 'primitive support' stamp: 'eem 5/30/2020 18:34'!
structOffsetOf: structPointerType atomicTypeCode: atomicTypeCode
	<inline: #always>
	^self cCode: [self addressOf: (self cCoerceSimple: 0 to: structPointerType) element]
		inSmalltalk: [self class primitiveAlignmentOf: atomicTypeCode]! !

!ThreadedFFIPlugin methodsFor: 'primitive support' stamp: 'eem 8/29/2020 11:01'!
unalignedLong32At: index
	"This alignment stuff is a hack for integerAt:put:size:signed:/primitiveFFIIntegerAt[Put].
	 The assumption right now is that all processors support unaligned access.  That only holds true
	 for x86, x86-64 & ARMv6 & later.  But this keeps us going until we can address it properly."
	<cmacro: '(index) long32At(index)'>
	^interpreterProxy long32At: index! !

!ThreadedFFIPlugin methodsFor: 'primitive support' stamp: 'eem 8/29/2020 11:03'!
unalignedLong32At: index put: value
	"This alignment stuff is a hack for integerAt:put:size:signed:/primitiveFFIIntegerAt[Put].
	 The assumption right now is that all processors support unaligned access.  That only holds true
	 for x86, x86-64 & ARMv6 & later.  But this keeps us going until we can address it properly."
	<cmacro: '(index,value) long32Atput(index,value)'>
	^interpreterProxy long32At: index put: value! !

!ThreadedFFIPlugin methodsFor: 'primitive support' stamp: 'eem 8/29/2020 11:02'!
unalignedLong64At: index
	"This alignment stuff is a hack for integerAt:put:size:signed:/primitiveFFIIntegerAt[Put].
	 The assumption right now is that all processors support unaligned access.  That only holds true
	 for x86, x86-64 & ARMv6 & later.  But this keeps us going until we can address it properly."
	<cmacro: '(index) long64At(index)'>
	^interpreterProxy long64At: index! !

!ThreadedFFIPlugin methodsFor: 'primitive support' stamp: 'eem 8/29/2020 11:03'!
unalignedLong64At: index put: value
	"This alignment stuff is a hack for integerAt:put:size:signed:/primitiveFFIIntegerAt[Put].
	 The assumption right now is that all processors support unaligned access.  That only holds true
	 for x86, x86-64 & ARMv6 & later.  But this keeps us going until we can address it properly."
	<cmacro: '(index,value) long64Atput(index,value)'>
	^interpreterProxy long64At: index put: value! !

!ThreadedFFIPlugin methodsFor: 'primitive support' stamp: 'eem 8/29/2020 11:01'!
unalignedShortAt: index
	"This alignment stuff is a hack for integerAt:put:size:signed:/primitiveFFIIntegerAt[Put].
	 The assumption right now is that all processors support unaligned access.  That only holds true
	 for x86, x86-64 & ARMv6 & later.  But this keeps us going until we can address it properly."
	<cmacro: '(index) shortAt(index)'>
	^interpreterProxy shortAt: index! !

!ThreadedFFIPlugin methodsFor: 'primitive support' stamp: 'eem 8/29/2020 11:03'!
unalignedShortAt: index put: value
	"This alignment stuff is a hack for integerAt:put:size:signed:/primitiveFFIIntegerAt[Put].
	 The assumption right now is that all processors support unaligned access.  That only holds true
	 for x86, x86-64 & ARMv6 & later.  But this keeps us going until we can address it properly."
	<cmacro: '(index,value) shortAtput(index,value)'>
	^interpreterProxy shortAt: index put: value! !

ThreadedFFIPlugin subclass: #ThreadedFFIPluginPartialSimulator
	instanceVariableNames: 'sizes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedFFIPluginPartialSimulator commentStamp: 'eem 7/16/2020 12:22' prior: 0!
A ThreadedFFIPluginPartialSimulator exists to test a few primitives such as primitiveCProgrammingModel.

Instance Variables
	sizes	a Dictionary of sizes for sizeof:!

!ThreadedFFIPluginPartialSimulator class methodsFor: 'translation' stamp: 'eem 7/27/2021 13:08'!
shouldBeTranslated
	^false! !

!ThreadedFFIPluginPartialSimulator methodsFor: 'simulation support' stamp: 'eem 7/16/2020 12:23'!
sizeof: aType
	^sizes
		ifNil: [super sizeof: aType]
		ifNotNil: [sizes at: aType]! !

!ThreadedFFIPluginPartialSimulator methodsFor: 'accessing' stamp: 'eem 7/16/2020 12:23'!
sizes

	^ sizes! !

!ThreadedFFIPluginPartialSimulator methodsFor: 'accessing' stamp: 'eem 7/16/2020 12:23'!
sizes: anObject

	sizes := anObject.! !

ThreadedFFIPlugin subclass: #ThreadedIA32FFIPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedIA32FFIPlugin commentStamp: '<historical>' prior: 0!
This subclass is for the IA32/x86 ABI.!

!ThreadedIA32FFIPlugin class methodsFor: 'translation' stamp: 'eem 6/12/2019 18:39'!
identifyingPredefinedMacros
	^#('_M_I386' '_M_IX86' '_X86_' 'i386' 'i486' 'i586' 'i686' '__i386__' '__386__' 'X86' 'I386')! !

!ThreadedIA32FFIPlugin class methodsFor: 'translation' stamp: 'eem 4/21/2014 11:05'!
moduleName
	^'IA32FFIPlugin'! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 5/16/2017 09:21'!
canReturnInRegistersStructOfSize: returnStructSize
	"Answer if a struct result of a given size is returned in registers or not."
	<cmacro: '(sz) (WIN32_X86_STRUCT_RETURN && (sz) <= 8 && !!((sz)&((sz)-1)))'>
	^returnStructSize <= 8 and: [returnStructSize isPowerOfTwo]! !

!ThreadedIA32FFIPlugin methodsFor: 'callout support' stamp: 'eem 7/27/2021 16:46'!
encodeStructReturnTypeIn: calloutState
	"Set the return type to true if returning the struct via register"
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>

	calloutState structReturnType: (self canReturnInRegistersStructOfSize: calloutState structReturnSize)! !

!ThreadedIA32FFIPlugin methodsFor: 'callout support' stamp: 'eem 8/29/2020 13:20'!
ffiCalloutTo: procAddr SpecOnStack: specOnStack in: calloutState
	<var: #procAddr type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	"Go out, call this guy and create the return value.  This *must* be inlined because of
	 the alloca of the outgoing stack frame in ffiCall:WithFlags:NumArgs:Args:AndTypes:"
	| myThreadIndex atomicType floatRet intRet |
	<var: #floatRet type: #double>
	<var: #intRet type: #usqLong>
	<inline: #always>
	myThreadIndex := interpreterProxy disownVM: (self disownFlagsFor: calloutState).

	(self allocaLiesSoSetSpBeforeCall or: [self mustAlignStack]) ifTrue:
		[self setsp: calloutState argVector].

	atomicType := self atomicTypeOf: calloutState ffiRetHeader.
	(atomicType >> 1) = (FFITypeSingleFloat >> 1) ifTrue:
		[floatRet := self dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'double (*)()').

		 "undo any callee argument pops because it may confuse stack management with the alloca."
		 (self isCalleePopsConvention: calloutState callFlags) ifTrue:
			[self setsp: calloutState argVector].
		 interpreterProxy ownVM: myThreadIndex.

		 ^interpreterProxy floatObjectOf: floatRet].

	intRet := self dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'usqLong (*)()').

	"undo any callee argument pops because it may confuse stack management with the alloca."
	(self isCalleePopsConvention: calloutState callFlags) ifTrue:
		[self setsp: calloutState argVector].
	interpreterProxy ownVM: myThreadIndex.

	(calloutState ffiRetHeader anyMask: FFIFlagPointer+FFIFlagStructure) ifTrue:
		["Note: Order is important here since FFIFlagPointer + FFIFlagStructure is used to represent
		 'typedef void* VoidPointer' and VoidPointer must be returned as pointer *not* as struct."
		 (calloutState ffiRetHeader anyMask: FFIFlagPointer) ifTrue:
			[^self ffiReturnPointer: intRet ofType: (self ffiReturnType: specOnStack) in: calloutState].
		 ^self ffiReturnStruct: (self addressOf: intRet) ofType: (self ffiReturnType: specOnStack) in: calloutState].
	
	^self ffiCreateIntegralResultOop: intRet ofAtomicType: atomicType in: calloutState! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 11/3/2018 13:21'!
ffiPushDoubleFloat: value in: calloutState
	<var: #value type: #double>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState currentArg + 8 > calloutState limit ifTrue:
		[^FFIErrorCallFrameTooBig].
	interpreterProxy storeFloatAtPointer: calloutState currentArg from: value.
	calloutState currentArg: calloutState currentArg + 8.
	^0! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushPointer: pointer in: calloutState
	<var: #pointer type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState currentArg + 4 > calloutState limit ifTrue:
		[^FFIErrorCallFrameTooBig].
	interpreterProxy longAt: calloutState currentArg put: pointer.
	calloutState currentArg: calloutState currentArg + 4.
	^0! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 7/27/2021 16:54'!
ffiPushSignedByte: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState currentArg + 4 > calloutState limit ifTrue:
		[^FFIErrorCallFrameTooBig].
	interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'signed char').
	calloutState currentArg: calloutState currentArg + 4.
	^0! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedChar: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState currentArg + 4 > calloutState limit ifTrue:
		[^FFIErrorCallFrameTooBig].
	interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'signed char').
	calloutState currentArg: calloutState currentArg + 4.
	^0! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedInt: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState currentArg + 4 > calloutState limit ifTrue:
		[^FFIErrorCallFrameTooBig].
	interpreterProxy longAt: calloutState currentArg put: value.
	calloutState currentArg: calloutState currentArg + 4.
	^0! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedLongLong: value in: calloutState
	<var: #value type: #usqLong>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState currentArg + 8 > calloutState limit ifTrue:
		[^FFIErrorCallFrameTooBig].
	interpreterProxy
		longAt: calloutState currentArg put: (self cCoerceSimple: value to: #usqInt);
		longAt: calloutState currentArg + 4 put: (self cCoerceSimple: value >> 32 to: #usqInt).
	calloutState currentArg: calloutState currentArg + 8.
	^0! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedShort: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState currentArg + 4 > calloutState limit ifTrue:
		[^FFIErrorCallFrameTooBig].
	interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'signed short').
	calloutState currentArg: calloutState currentArg + 4.
	^0! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 11/3/2018 13:19'!
ffiPushSingleFloat: value in: calloutState
	<var: #value type: #float>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState currentArg + 4 > calloutState limit ifTrue:
		[^FFIErrorCallFrameTooBig].
	interpreterProxy storeSingleFloatAtPointer: calloutState currentArg from: value.
	calloutState currentArg: calloutState currentArg + 4.
	^0! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushStructure: pointer ofSize: structSize typeSpec: argSpec ofLength: argSpecSize in: calloutState
	<var: #pointer type: #'void *'>
	<var: #argSpec type: #'sqInt *'>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	| roundedSize |
	roundedSize := structSize + 3 bitClear: 3.
	calloutState currentArg + roundedSize > calloutState limit ifTrue:
		[^FFIErrorCallFrameTooBig].
	self memcpy: calloutState currentArg _: pointer _: structSize.
	calloutState currentArg: calloutState currentArg + roundedSize.
	^0! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedByte: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState currentArg + 4 > calloutState limit ifTrue:
		[^FFIErrorCallFrameTooBig].
	interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'unsigned char').
	calloutState currentArg: calloutState currentArg + 4.
	^0! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedChar: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState currentArg + 4 > calloutState limit ifTrue:
		[^FFIErrorCallFrameTooBig].
	interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'unsigned char').
	calloutState currentArg: calloutState currentArg + 4.
	^0! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedInt: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState currentArg + 4 > calloutState limit ifTrue:
		[^FFIErrorCallFrameTooBig].
	interpreterProxy longAt: calloutState currentArg put: value.
	calloutState currentArg: calloutState currentArg + 4.
	^0! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedLongLong: value in: calloutState
	<var: #value type: #usqLong>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState currentArg + 8 > calloutState limit ifTrue:
		[^FFIErrorCallFrameTooBig].
	interpreterProxy
		longAt: calloutState currentArg put: (self cCoerceSimple: value to: #usqInt);
		longAt: calloutState currentArg + 4 put: (self cCoerceSimple: value >> 32 to: #usqInt).
	calloutState currentArg: calloutState currentArg + 8.
	^0! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedShort: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState currentArg + 4 > calloutState limit ifTrue:
		[^FFIErrorCallFrameTooBig].
	interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'unsigned short').
	calloutState currentArg: calloutState currentArg + 4.
	^0! !

!ThreadedIA32FFIPlugin methodsFor: 'callout support' stamp: 'nice 1/30/2020 21:50'!
ffiReturnStruct: longLongRetPtr ofType: ffiRetType in: calloutState
	<var: #longLongRetPtr type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	"Create a structure return value from an external function call.  The value as been stored in
	 alloca'ed space pointed to by the calloutState or in the return value passed by pointer."
	| retOop retClass oop |
	<inline: true>
	retClass := interpreterProxy fetchPointer: 1 ofObject: ffiRetType.
	retOop := interpreterProxy instantiateClass: retClass indexableSize: 0.
	self remapOop: retOop
		in: [oop := interpreterProxy 
					instantiateClass: interpreterProxy classByteArray 
					indexableSize: calloutState structReturnSize].
	self memcpy: (interpreterProxy firstIndexableField: oop)
		_: ((self returnStructInRegisters: calloutState)
				ifTrue: [longLongRetPtr]
				ifFalse: [calloutState limit])
		 _: calloutState structReturnSize.
	interpreterProxy storePointer: 0 ofObject: retOop withValue: oop.
	^retOop! !

!ThreadedIA32FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
returnStructInRegisters: calloutState
	"Return through register if structReturnType is true"
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	^calloutState structReturnType! !

ThreadedFFIPlugin subclass: #ThreadedPPCBEFFIPlugin
	instanceVariableNames: ''
	classVariableNames: 'NumRegArgs'
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedPPCBEFFIPlugin commentStamp: '<historical>' prior: 0!
This subclass is for the PowerPC ABI in BigEndian mode.  It typically uses 8 integer registers and 8 single-precision floating-point registers that shadow the top 8 words on the stack.!

!ThreadedPPCBEFFIPlugin class methodsFor: 'translation' stamp: 'eem 11/7/2009 09:51'!
calloutStateClass
	^ThreadedFFICalloutStateForPPC! !

!ThreadedPPCBEFFIPlugin class methodsFor: 'class initialization' stamp: 'eem 11/9/2009 17:38'!
initialize
	NumRegArgs := 8! !

!ThreadedPPCBEFFIPlugin class methodsFor: 'translation' stamp: 'eem 4/21/2014 11:05'!
moduleName
	^'PPC32FFIPlugin'! !

!ThreadedPPCBEFFIPlugin class methodsFor: 'accessing' stamp: 'eem 11/9/2009 17:39'!
numRegArgs
	^NumRegArgs! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
canReturnInRegistersStructOfSize: returnStructSize
	"Answer if a struct result of a given size is returned in registers or not.
	 The ABI spec defines return in registers, but some linux gcc versions implemented an
	 erroneous draft which does not return any struct in memory.  Implement the SysV ABI."
	<inline: #always>
	^returnStructSize <= 8! !

!ThreadedPPCBEFFIPlugin methodsFor: 'callout support' stamp: 'eem 7/27/2021 16:46'!
encodeStructReturnTypeIn: calloutState
	"Set the return type to true if returning the struct via register"
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>

	calloutState structReturnType: (self canReturnInRegistersStructOfSize: calloutState structReturnSize)! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushDoubleFloat: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self shouldBeImplemented.
	^0! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushPointer: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self shouldBeImplemented.
	^0! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 7/27/2021 16:54'!
ffiPushSignedByte: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self shouldBeImplemented.
	^0! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedChar: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self shouldBeImplemented.
	^0! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedInt: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self shouldBeImplemented.
	^0! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedLongLong: value in: calloutState
	<var: #value type: #sqLong>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self shouldBeImplemented.
	^0! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedShort: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self shouldBeImplemented.
	^0! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSingleFloat: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self shouldBeImplemented.
	^0! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedByte: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self shouldBeImplemented.
	^0! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedChar: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self shouldBeImplemented.
	^0! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedInt: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self shouldBeImplemented.
	^0! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedLongLong: value in: calloutState
	<var: #value type: #usqLong>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self shouldBeImplemented.
	^0! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedShort: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	self shouldBeImplemented.
	^0! !

!ThreadedPPCBEFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
returnStructInRegisters: calloutState
	"Return through register if structReturnType is true"
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	^calloutState structReturnType! !

ThreadedFFIPlugin subclass: #ThreadedX64FFIPlugin
	instanceVariableNames: ''
	classVariableNames: 'NumFloatRegArgs NumIntRegArgs WordSize'
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedX64FFIPlugin commentStamp: 'eem 2/19/2016 15:03' prior: 0!
This is an abstract superclass for the System V and WIN64 x86-64 ABIs that share a lot of code other than for structure passing.!

!ThreadedX64FFIPlugin class methodsFor: 'translation' stamp: 'eem 2/19/2016 16:38'!
calloutStateClass
	^ThreadedFFICalloutStateForX64! !

!ThreadedX64FFIPlugin class methodsFor: 'class initialization' stamp: 'eem 2/19/2016 15:01'!
initialize
	WordSize := 8.
	NumIntRegArgs := 6.
	NumFloatRegArgs := 8! !

!ThreadedX64FFIPlugin class methodsFor: 'translation' stamp: 'eem 2/19/2016 15:01'!
moduleName
	^'X64SysVFFIPlugin'! !

!ThreadedX64FFIPlugin class methodsFor: 'accessing' stamp: 'eem 2/19/2016 15:01'!
numFloatRegArgs
	^NumFloatRegArgs! !

!ThreadedX64FFIPlugin class methodsFor: 'accessing' stamp: 'eem 2/19/2016 15:01'!
numRegArgs
	^NumIntRegArgs! !

!ThreadedX64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 11/3/2018 13:21'!
ffiPushDoubleFloat: value in: calloutState
	<var: #value type: #double>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>

	calloutState floatRegisterIndex < NumFloatRegArgs
		ifTrue:
			[calloutState floatRegisters at: calloutState floatRegisterIndex put: value.
			 calloutState incrementFloatRegisterIndex]
		ifFalse:
			[calloutState currentArg + WordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy storeFloatAtPointer: calloutState currentArg from: value.
			 calloutState currentArg: calloutState currentArg + WordSize].
	^0! !

!ThreadedX64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushPointer: pointer in: calloutState
	<var: #pointer type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: pointer asInteger.
			 calloutState incrementIntegerRegisterIndex]
		ifFalse:
			[calloutState currentArg + WordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: pointer.
			 calloutState currentArg: calloutState currentArg + WordSize].
	^0! !

!ThreadedX64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 7/27/2021 16:53'!
ffiPushSignedByte: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #'signed char').
			 calloutState incrementIntegerRegisterIndex]
		ifFalse:
			[calloutState currentArg + WordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'signed char').
			 calloutState currentArg: calloutState currentArg + WordSize].
	^0! !

!ThreadedX64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedChar: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #'signed char').
			 calloutState incrementIntegerRegisterIndex]
		ifFalse:
			[calloutState currentArg + WordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'signed char').
			 calloutState currentArg: calloutState currentArg + WordSize].
	^0! !

!ThreadedX64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedInt: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: value.
			 calloutState incrementIntegerRegisterIndex]
		ifFalse:
			[calloutState currentArg + WordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: value.
			 calloutState currentArg: calloutState currentArg + WordSize].
	^0! !

!ThreadedX64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedLongLong: value in: calloutState
	<var: #value type: #sqLong>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #usqInt).
			 calloutState incrementIntegerRegisterIndex]
		ifFalse:
			[calloutState currentArg + WordSize > calloutState limit ifTrue: [^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: value.
			 calloutState currentArg: calloutState currentArg + WordSize].
	^0! !

!ThreadedX64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushSignedShort: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #'signed short').
			 calloutState incrementIntegerRegisterIndex]
		ifFalse:
			[calloutState currentArg + WordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'signed short').
			 calloutState currentArg: calloutState currentArg + WordSize].
	^0! !

!ThreadedX64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 11/5/2018 10:32'!
ffiPushSingleFloat: value in: calloutState
	<var: #value type: #float>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>

	calloutState floatRegisterIndex < NumFloatRegArgs
		ifTrue:
			[(self cCoerce: calloutState floatRegisters + calloutState floatRegisterIndex to: #'float *') at: 0 put: value.
			 calloutState incrementFloatRegisterIndex]
		ifFalse:
			[calloutState currentArg + WordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 "Float passed in least significant word, but good to zero the high bits for clarity."
			 interpreterProxy
				long64AtPointer: calloutState currentArg put: 0;
				storeSingleFloatAtPointer: calloutState currentArg from: value.
			 calloutState currentArg: calloutState currentArg + WordSize].
	^0! !

!ThreadedX64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedByte: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #'unsigned char').
			 calloutState incrementIntegerRegisterIndex]
		ifFalse:
			[calloutState currentArg + WordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'unsigned char').
			 calloutState currentArg: calloutState currentArg + WordSize].
	^0! !

!ThreadedX64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedChar: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #'unsigned char').
			 calloutState incrementIntegerRegisterIndex]
		ifFalse:
			[calloutState currentArg + WordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'unsigned char').
			 calloutState currentArg: calloutState currentArg + WordSize].
	^0! !

!ThreadedX64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedInt: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: value.
			 calloutState incrementIntegerRegisterIndex]
		ifFalse:
			[calloutState currentArg + WordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: value.
			 calloutState currentArg: calloutState currentArg + WordSize].
	^0! !

!ThreadedX64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedLongLong: value in: calloutState
	<var: #value type: #usqLong>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: value.
			 calloutState incrementIntegerRegisterIndex]
		ifFalse:
			[calloutState currentArg + WordSize > calloutState limit ifTrue: [^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: value.
			 calloutState currentArg: calloutState currentArg + WordSize].
	^0! !

!ThreadedX64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushUnsignedShort: value in: calloutState
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	calloutState integerRegisterIndex < NumIntRegArgs
		ifTrue:
			[calloutState integerRegisters at: calloutState integerRegisterIndex put: (self cCoerceSimple: value to: #'unsigned short').
			 calloutState incrementIntegerRegisterIndex]
		ifFalse:
			[calloutState currentArg + WordSize > calloutState limit ifTrue:
				[^FFIErrorCallFrameTooBig].
			 interpreterProxy longAt: calloutState currentArg put: (self cCoerceSimple: value to: #'unsigned short').
			 calloutState currentArg: calloutState currentArg + WordSize].
	^0! !

ThreadedX64FFIPlugin subclass: #ThreadedX64SysVFFIPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedX64SysVFFIPlugin commentStamp: 'eem 2/17/2016 20:03' prior: 0!
This subclass is for the System V x86-64 ABI.  The System V ABI uses 6 integer registers and 8 double-precision floating-point registers.  See http://www.x86-64.org/documentation/abi.pdf.  It seems that Mac OS X and linux use draft version 0.90, http://people.freebsd.org/~obrien/amd64-elf-abi.pdf.

Note that the System V x86-64 ABI decomposes structs of 2 eightbytes or smaller passed by value across available parameter registers.  double fields will be passed in an xmm register.!

!ThreadedX64SysVFFIPlugin class methodsFor: 'translation' stamp: 'eem 5/31/2020 15:42'!
ancilliaryClasses
	^super ancilliaryClasses,
	  { ThreadedFFIX64SixteenByteReturnDD.
		ThreadedFFIX64SixteenByteReturnDI.
		ThreadedFFIX64SixteenByteReturnID.
		ThreadedFFIX64SixteenByteReturnII }! !

!ThreadedX64SysVFFIPlugin class methodsFor: 'translation' stamp: 'eem 5/30/2017 19:24'!
calloutStateClass
	^ThreadedFFICalloutStateForX64SysV! !

!ThreadedX64SysVFFIPlugin class methodsFor: 'translation' stamp: 'eem 5/16/2017 09:16'!
excludingPredefinedMacros
	"Answer the predefined macros that disqualify the platforms a subclass handles, if any.
	 This can be used to differentiate e.g. x64 Sys V from x64 Win64."
	^#('_WIN64')! !

!ThreadedX64SysVFFIPlugin class methodsFor: 'translation' stamp: 'eem 2/16/2016 20:37'!
identifyingPredefinedMacros
	"Answer the predefined macros that identify the platforms a subclass handles, if any.
	 If the subclass isn't yet ready for production (a work in progress) simply answer nil."
	^#('x86_64' '__amd64' '__x86_64' '__amd64__' '__x86_64__')! !

!ThreadedX64SysVFFIPlugin class methodsFor: 'class initialization' stamp: 'eem 2/16/2016 19:33'!
initialize
	WordSize := 8.
	NumIntRegArgs := 6.
	NumFloatRegArgs := 8! !

!ThreadedX64SysVFFIPlugin class methodsFor: 'translation' stamp: 'eem 2/16/2016 19:33'!
moduleName
	^'X64SysVFFIPlugin'! !

!ThreadedX64SysVFFIPlugin class methodsFor: 'accessing' stamp: 'eem 2/16/2016 19:07'!
numFloatRegArgs
	^NumFloatRegArgs! !

!ThreadedX64SysVFFIPlugin class methodsFor: 'accessing' stamp: 'eem 2/16/2016 19:23'!
numRegArgs
	^NumIntRegArgs! !

!ThreadedX64SysVFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
canReturnInRegistersStructOfSize: returnStructSize
	"Answer if a struct result of a given size is returned in registers or not."
	<inline: #always>
	^returnStructSize <= (WordSize * 2)! !

!ThreadedX64SysVFFIPlugin methodsFor: 'callout support' stamp: 'eem 7/27/2021 16:46'!
encodeStructReturnTypeIn: calloutState
	"Set the return type to an integer encoding the type of registers used to return the struct
	* 2r00 for float float (XMM0 XMM1)
	* 2r01 for int float (RAX XMM0)
	* 2r10 for float int (XMM0 RAX)
	* 2r11 for int int (RAX RDX)
	* 2r100 for float (XMM0)
	* 2r101 for int (RAX)
	* 2r110 For return thru memory (struct field not aligned or struct too big)"
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>

	| registerType |
	registerType := (self canReturnInRegistersStructOfSize: calloutState structReturnSize)
		ifTrue:
			[ self
				registerTypeForStructSpecs: (interpreterProxy firstIndexableField: calloutState ffiRetSpec)
				OfLength: (interpreterProxy slotSizeOf: calloutState ffiRetSpec)]
		ifFalse:
			[ "We cannot return via register, struct is too big"
			2r110 ].
	calloutState structReturnType: registerType! !

!ThreadedX64SysVFFIPlugin methodsFor: 'callout support' stamp: 'eem 8/29/2020 13:20'!
ffiCalloutTo: procAddr SpecOnStack: specOnStack in: calloutState
	<var: #procAddr type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	<var: #loadFloatRegs declareC: 'extern void loadFloatRegs(double, double, double, double, double, double, double, double)'>
	"Go out, call this guy and create the return value.  This *must* be inlined because of
	 the alloca of the outgoing stack frame in ffiCall:WithFlags:NumArgs:Args:AndTypes:"
	| myThreadIndex atomicType floatRet intRet sddRet sdiRet sidRet siiRet returnStructByValue registerType sRetPtr |
	<var: #floatRet type: #double>
	<var: #intRet type: #sqInt>
	<var: #siiRet type: #SixteenByteReturnII>
	<var: #sidRet type: #SixteenByteReturnID>
	<var: #sdiRet type: #SixteenByteReturnDI>
	<var: #sddRet type: #SixteenByteReturnDD>
	<var: #sRetPtr type: #'void *'>
	<inline: #always>
	
	returnStructByValue := (calloutState ffiRetHeader bitAnd: FFIFlagStructure + FFIFlagPointer + FFIFlagAtomic) = FFIFlagStructure.
	
	myThreadIndex := interpreterProxy disownVM: (self disownFlagsFor: calloutState).

	calloutState floatRegisterIndex > 0 ifTrue:
		[self 
			load: (calloutState floatRegisters at: 0)
			Flo: (calloutState floatRegisters at: 1)
			a: (calloutState floatRegisters at: 2)
			t: (calloutState floatRegisters at: 3)
			R: (calloutState floatRegisters at: 4)
			e: (calloutState floatRegisters at: 5)
			g: (calloutState floatRegisters at: 6)
			s: (calloutState floatRegisters at: 7)].

	(self allocaLiesSoSetSpBeforeCall or: [self mustAlignStack]) ifTrue:
		[self setsp: calloutState argVector].

	atomicType := self atomicTypeOf: calloutState ffiRetHeader.
	(atomicType >> 1) = (FFITypeSingleFloat >> 1) ifTrue:
		[atomicType = FFITypeSingleFloat
			ifTrue:
				[floatRet := self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'float (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)
					with: (calloutState integerRegisters at: 4)
					with: (calloutState integerRegisters at: 5)]
			ifFalse: "atomicType = FFITypeDoubleFloat"
				[floatRet := self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'double (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)
					with: (calloutState integerRegisters at: 4)
					with: (calloutState integerRegisters at: 5)].

		 interpreterProxy ownVM: myThreadIndex.

		 ^interpreterProxy floatObjectOf: floatRet].

	returnStructByValue  ifFalse:
		[intRet := self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'sqInt (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)
					with: (calloutState integerRegisters at: 4)
					with: (calloutState integerRegisters at: 5).
		interpreterProxy ownVM: myThreadIndex.
		(calloutState ffiRetHeader anyMask: FFIFlagPointer) ifTrue:
			[^self ffiReturnPointer: intRet ofType: (self ffiReturnType: specOnStack) in: calloutState].
		^self ffiCreateIntegralResultOop: intRet ofAtomicType: atomicType in: calloutState].

	registerType := calloutState structReturnType.
	registerType
		caseOf:
			{[2r00] ->
				[sddRet := self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'SixteenByteReturnDD (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)
					with: (calloutState integerRegisters at: 4)
					with: (calloutState integerRegisters at: 5).
				sRetPtr := (self addressOf: sddRet) asVoidPointer].
			 [2r01] ->
				[sidRet := self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'SixteenByteReturnID (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)
					with: (calloutState integerRegisters at: 4)
					with: (calloutState integerRegisters at: 5).
				sRetPtr := (self addressOf: sidRet) asVoidPointer].
			 [2r10] ->
				[sdiRet := self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'SixteenByteReturnDI (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)
					with: (calloutState integerRegisters at: 4)
					with: (calloutState integerRegisters at: 5).
				sRetPtr := (self addressOf: sdiRet) asVoidPointer].
			 [2r11] ->
				[siiRet := self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'SixteenByteReturnII (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)
					with: (calloutState integerRegisters at: 4)
					with: (calloutState integerRegisters at: 5).
				sRetPtr := (self addressOf: siiRet) asVoidPointer].
			 [2r100] ->
				[floatRet := self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'double (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)
					with: (calloutState integerRegisters at: 4)
					with: (calloutState integerRegisters at: 5).
				sRetPtr := (self addressOf: floatRet) asVoidPointer].
			 [2r101] ->
				[intRet := self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'sqInt (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)
					with: (calloutState integerRegisters at: 4)
					with: (calloutState integerRegisters at: 5).
				sRetPtr := (self addressOf: intRet) asVoidPointer].
			 [2r110] ->
				["return a pointer to alloca'd memory"
				intRet := self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'sqInt (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)')
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)
					with: (calloutState integerRegisters at: 4)
					with: (calloutState integerRegisters at: 5).
				sRetPtr := intRet asVoidPointer "address of struct is returned in RAX, which also is calloutState limit"]}
		otherwise:
			[interpreterProxy ownVM: myThreadIndex.
			self ffiFail: FFIErrorWrongType. ^nil].

	interpreterProxy ownVM: myThreadIndex.
	^self ffiReturnStruct: sRetPtr ofType: (self ffiReturnType: specOnStack) in: calloutState! !

!ThreadedX64SysVFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushStructure: pointer ofSize: structSize typeSpec: argSpec ofLength: argSpecSize in: calloutState
	<var: #pointer type: #'void *'>
	<var: #argSpec type: #'sqInt *'>
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	| roundedSize registerType numDoubleRegisters numIntegerRegisters passField0InXmmReg passField1InXmmReg |
	structSize <= 16 ifTrue:
		["See sec 3.2.3 of http://people.freebsd.org/~obrien/amd64-elf-abi.pdf. (dravft version 0.90).
		  All of the folowing are passed in registers:
			typedef struct { long a; } s0;
			typedef struct { double a; } s1;
			typedef struct { long a; double b; } s2;
			typedef struct { int a; int b; double c; } s2a;
			typedef struct { short a; short b; short c; short d; double e; } s2b;
			typedef struct { long a; float b; } s2f;
			typedef struct { long a; float b; float c; } s2g;
			typedef struct { int a; float b; int c; float d; } s2h;"
		 registerType := self registerTypeForStructSpecs: (self cCoerce: argSpec to: #'unsigned int *') OfLength: argSpecSize.
		 registerType = 2r110 "check case of invalid alignment => pass by memory"
			ifFalse: 
				 [passField0InXmmReg := (registerType bitAnd: 1) = 0.
				  structSize <= 8
					ifTrue:
						[numIntegerRegisters := registerType bitAnd: 1.
						 numDoubleRegisters := 1 - numIntegerRegisters]
					ifFalse:
						[passField1InXmmReg := (registerType bitAnd: 2) = 0.
						 numIntegerRegisters := (registerType bitAnd: 2) >> 1 + (registerType bitAnd: 1).
						 numDoubleRegisters := 2 - numIntegerRegisters].
				 (calloutState floatRegisterIndex + numDoubleRegisters <= NumFloatRegArgs
				  and: [calloutState integerRegisterIndex + numIntegerRegisters <= NumIntRegArgs]) ifTrue:
					[passField0InXmmReg
						ifTrue: [self ffiPushDoubleFloat: ((self cCoerceSimple: pointer to: #'double *') at: 0) in: calloutState]
						ifFalse: [self ffiPushSignedLongLong: ((self cCoerceSimple: pointer to: #'long long *') at: 0) in: calloutState].
					 structSize > 8 ifTrue:
						[passField1InXmmReg
							ifTrue: [self ffiPushDoubleFloat: ((self cCoerceSimple: pointer to: #'double *') at: 1) in: calloutState]
							ifFalse: [self ffiPushSignedLongLong: ((self cCoerceSimple: pointer to: #'long long *') at: 1) in: calloutState]].
					 ^0]]].

	roundedSize := structSize + 7 bitClear: 7.
	calloutState currentArg + roundedSize > calloutState limit ifTrue:
		 [^FFIErrorCallFrameTooBig].
	self memcpy: calloutState currentArg _: (self cCoerceSimple: pointer to: 'char *') _: structSize.
	calloutState currentArg: calloutState currentArg + roundedSize.
	^0! !

!ThreadedX64SysVFFIPlugin methodsFor: 'callout support' stamp: 'nice 1/29/2020 22:50'!
ffiReturnStruct: structRetPtr ofType: ffiRetType in: calloutState
	<var: #structRetPtr type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	"Create a structure return value from an external function call.  The value has been stored in
	 alloca'ed space pointed to by the calloutState or in the return value passed by pointer."
	| retOop retClass oop |
	<inline: true>
	retClass := interpreterProxy fetchPointer: 1 ofObject: ffiRetType.
	retOop := interpreterProxy instantiateClass: retClass indexableSize: 0.
	self remapOop: retOop
		in: [oop := interpreterProxy 
					instantiateClass: interpreterProxy classByteArray 
					indexableSize: calloutState structReturnSize].
	self memcpy: (interpreterProxy firstIndexableField: oop)
		 _: structRetPtr
		 _: calloutState structReturnSize.
	interpreterProxy storePointer: 0 ofObject: retOop withValue: oop.
	^retOop! !

!ThreadedX64SysVFFIPlugin methodsFor: 'marshalling' stamp: 'eem 8/28/2020 13:21'!
registerType: initialRegisterType ForStructSpecs: specs OfLength: specSize StartingAt: indexPtr ByteOffset: initialByteOffset EightbyteOffset: initialEightbyteOffset
	"Answer with a number characterizing the register type for passing a struct of size <= 16 bytes.
	On input, the index points to the structure header (the one with FFIFlagStructure + structSize)
	On output, the index points to the structure trailer (the FFIFlagStructure)."

	<var: #specs type: #'unsigned int *'>
	<var: #indexPtr type: #'sqInt *'>
	<inline: false>
	| registerType eightbyteOffset byteOffset spec fieldSize alignment atomic subIndex isInt recurse subLevel |
	registerType := initialRegisterType.
	byteOffset := initialByteOffset.
	eightbyteOffset := initialEightbyteOffset.
	[indexPtr at: 0 put: (indexPtr at: 0) + 1.
	subLevel := 0.
	(indexPtr at: 0) < specSize]
		whileTrue:
			[spec := specs at: (indexPtr at: 0).
			isInt := false.
			recurse := false.
			spec = FFIFlagStructure "this marks end of structure/union"
				ifTrue:
					[subLevel = 0 ifTrue: [^registerType].
					subLevel := subLevel - 1]
				ifFalse:
					[(spec anyMask: FFIFlagPointer)
						ifTrue:
							[fieldSize := BytesPerWord.
							alignment := fieldSize.
							isInt := true]
						ifFalse:
							[(spec bitAnd: FFIFlagStructure + FFIFlagAtomic)
								caseOf:
									{[FFIFlagStructure] ->
										[fieldSize := 0.
										subIndex := indexPtr at: 0.
										alignment := self alignmentOfStructSpec: specs OfLength: specSize StartingAt: (self addressOf: subIndex put: [:v| subIndex := v]).
										recurse := self isUnionSpec: specs OfLength: specSize StartingAt: (indexPtr at: 0).
										recurse
											ifTrue: [fieldSize := spec bitAnd: FFIStructSizeMask]
											ifFalse: [subLevel := subLevel + 1]].
									[FFIFlagAtomic] ->
										[fieldSize := spec bitAnd: FFIStructSizeMask.
										alignment := fieldSize.
										atomic := self atomicTypeOf: spec.
										isInt := (atomic >> 1) ~= (FFITypeSingleFloat >> 1)]}
								otherwise: ["invalid spec" ^-1]].
					(byteOffset bitAnd: alignment - 1) = 0
						ifFalse:
							["this field requires alignment"
							byteOffset := (byteOffset bitClear: alignment - 1) + alignment].
					byteOffset + fieldSize > 8
						ifTrue:
							["Not enough room on current Eightbyte for this field, skip to next one"
							eightbyteOffset := eightbyteOffset + 1.
							byteOffset := 0].
					isInt
						ifTrue:
							["If this eightbyte contains an int field, then we must use an int register"
							registerType := registerType bitOr: 1 << eightbyteOffset].
					recurse ifTrue:
						["union in structs require a recursive form, because we handle byteOffset/eightbyteOffset differently"
						registerType := self
								registerType: registerType
								ForUnionSpecs: specs
								OfLength: specSize
								StartingAt: indexPtr
								ByteOffset: byteOffset
								EightbyteOffset: eightbyteOffset].
					"where to put the next field?"
					byteOffset := byteOffset + fieldSize.
					byteOffset >= 8
						ifTrue:
							["This eightbyte is full, skip to next one"
							eightbyteOffset := eightbyteOffset + 1.
							byteOffset := 0]]].
	self assert: subLevel = 0.
	^registerType! !

!ThreadedX64SysVFFIPlugin methodsFor: 'marshalling' stamp: 'eem 8/28/2020 13:13'!
registerType: initialRegisterType ForUnionSpecs: specs OfLength: specSize StartingAt: indexPtr ByteOffset: byteOffset EightbyteOffset: eightbyteOffset
	"Answer with a number characterizing the register type for passing a union of size <= 16 bytes.
	On input, the index points to the structure header (the one with FFIFlagStructure + structSize)
	On output, the index points to the structure trailer (the FFIFlagStructure)."

	<var: #specs type: #'unsigned int *'>
	<var: #indexPtr type: #'sqInt *'>
	<inline: false>
	| registerType spec atomic isInt recurse subLevel |
	registerType := initialRegisterType.
	[indexPtr at: 0 put: (indexPtr at: 0) + 1.
	subLevel := 0.
	(indexPtr at: 0) < specSize]
		whileTrue:
			[spec := specs at: (indexPtr at: 0).
			isInt := false.
			recurse := false.
			spec = FFIFlagStructure "this marks end of structure/union"
				ifTrue:
					[subLevel = 0 ifTrue: [^registerType].
					subLevel := subLevel - 1]
				ifFalse:
					[(spec anyMask: FFIFlagPointer)
						ifTrue:
							[isInt := true]
						ifFalse:
							[(spec bitAnd: FFIFlagStructure + FFIFlagAtomic)
								caseOf:
									{[FFIFlagStructure] ->
										[recurse := (self isUnionSpec: specs OfLength: specSize StartingAt: (indexPtr at: 0))not.
										recurse ifFalse: [subLevel := subLevel + 1]].
									[FFIFlagAtomic] ->
										[atomic := self atomicTypeOf: spec.
										isInt := (atomic >> 1) ~= (FFITypeSingleFloat >> 1)]}
								otherwise: ["invalid spec" ^-1]].
					isInt
						ifTrue:
							["If this eightbyte contains an int field, then we must use an int register"
							registerType := registerType bitOr: 1 << eightbyteOffset].
					recurse ifTrue:
						["struct in union require a recursive form, because we handle byteOffset/eightbyteOffset differently"
						registerType := self
								registerType: registerType
								ForStructSpecs: specs
								OfLength: specSize
								StartingAt: indexPtr
								ByteOffset: byteOffset
								EightbyteOffset: eightbyteOffset]]].
	self assert: subLevel = 0.
	^registerType! !

!ThreadedX64SysVFFIPlugin methodsFor: 'marshalling' stamp: 'eem 8/28/2020 13:14'!
registerTypeForStructSpecs: specs OfLength: specSize
	"Answer with a number characterizing the register type for passing a struct of size <= 16 bytes.
	The bit at offset i of registerType is set to 1 if eightbyte at offset i is a int register (RAX ...)
	The bit at offset 2 indicates if there is a single eightbyte (struct size <= 8)
	* 2r00 for float float (XMM0 XMM1)
	* 2r01 for int float (RAX XMM0)
	* 2r10 for float int (XMM0 RAX)
	* 2r11 for int int (RAX RDX)
	* 2r100 for float (XMM0)
	* 2r101 for int (RAX)
	* 2r110 INVALID (not aligned)
	Beware, the bits must be read from right to left for decoding register type.
	Note: this method reconstructs the struct layout according to X64 alignment rules.
	Therefore, it will not work for packed struct or other exotic alignment."

	<var: #specs type: #'unsigned int *'>
	<inline: false>
	| index byteSize registerType |
	index := 0.
	byteSize := (specs at: index) bitAnd: FFIStructSizeMask.
	byteSize > 16 ifTrue: [^2r110].
	(self checkAlignmentOfStructSpec: specs OfLength: specSize StartingAt: index)
		ifFalse: [^2r110].
	registerType := byteSize <= 8 ifTrue: [2r100] ifFalse: [0].
	^(self isUnionSpec: specs OfLength: specSize StartingAt: 0)
		ifTrue: [ self 
			registerType: registerType
			ForUnionSpecs: specs
			OfLength: specSize
			StartingAt: (self addressOf: index put: [:v| index := v])
			ByteOffset: 0
			EightbyteOffset: 0 ]
		ifFalse: [ self 
			registerType: registerType
			ForStructSpecs: specs
			OfLength: specSize
			StartingAt: (self addressOf: index put: [:v| index := v])
			ByteOffset: 0
			EightbyteOffset: 0 ]! !

!ThreadedX64SysVFFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
returnStructInRegisters: calloutState
	"Use the register type encoding stored in structReturnType
	2r110 means impossible, pass thru memory.
	Anything smaller encodes the type of register used, thus means true."
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	^calloutState structReturnType < 2r110! !

ThreadedX64FFIPlugin subclass: #ThreadedX64Win64FFIPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedX64Win64FFIPlugin commentStamp: 'nice 8/10/2016 19:23' prior: 0!
This subclass is for the Win64 x86-64 ABI.  The Win64 ABI uses 4 integer registers or 4 double-precision floating-point registers or a mix of the two.  See w.g. https://msdn.microsoft.com/en-us/library/ms235286.aspx, or google for "Overview of x64 Calling Conventions - MSDN - Microsoft".

Note that unlike the System V x86-64 ABI, the Win64 ABI does /not/ decompose structs passed by value across available parameter registers.!

!ThreadedX64Win64FFIPlugin class methodsFor: 'translation' stamp: 'nice 8/11/2016 03:00'!
calloutStateClass
	^ThreadedFFICalloutStateForX64Win64! !

!ThreadedX64Win64FFIPlugin class methodsFor: 'translation' stamp: 'eem 2/16/2016 20:37'!
identifyingPredefinedMacros
	"Answer the predefined macros that identify the platforms a subclass handles, if any.
	 If the subclass isn't yet ready for production (a work in progress) simply answer nil."
	^#('x86_64' '__amd64' '__x86_64' '__amd64__' '__x86_64__' '_M_AMD64' '_M_X64')! !

!ThreadedX64Win64FFIPlugin class methodsFor: 'translation' stamp: 'eem 5/16/2017 09:16'!
includingPredefinedMacros
	"Answer the predefined macros that qualify the platforms a subclass handles, if any.
	 These are anded together and with excludingPredefinedMacros, whereas
	 identifyingPredefinedMacros are ored together.
	 This can be used to differentiate e.g. x64 Sys V from x64 Win64."
	^#('_WIN64')! !

!ThreadedX64Win64FFIPlugin class methodsFor: 'class initialization' stamp: 'eem 2/19/2016 16:52'!
initialize
	WordSize := 8.
	NumIntRegArgs := 4.
	NumFloatRegArgs := 4! !

!ThreadedX64Win64FFIPlugin class methodsFor: 'translation' stamp: 'eem 2/16/2016 20:43'!
moduleName
	^'X64Win64FFIPlugin'! !

!ThreadedX64Win64FFIPlugin class methodsFor: 'accessing' stamp: 'eem 2/16/2016 19:35'!
numFloatRegArgs
	^NumFloatRegArgs! !

!ThreadedX64Win64FFIPlugin class methodsFor: 'accessing' stamp: 'eem 2/16/2016 19:35'!
numRegArgs
	^NumIntRegArgs! !

!ThreadedX64Win64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
canReturnInRegistersStructOfSize: returnStructSize
	"Answer if a struct result of a given size is returned in registers or not."
	<inline: #always>
	^returnStructSize <= WordSize and: ["returnStructSize isPowerOfTwo" (returnStructSize bitAnd: returnStructSize-1) = 0]! !

!ThreadedX64Win64FFIPlugin methodsFor: 'callout support' stamp: 'eem 7/27/2021 16:46'!
encodeStructReturnTypeIn: calloutState
	"Set the return type to true if returning the struct via register"
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>

	calloutState structReturnType: (self canReturnInRegistersStructOfSize: calloutState structReturnSize)! !

!ThreadedX64Win64FFIPlugin methodsFor: 'callout support' stamp: 'eem 8/29/2020 13:19'!
ffiCalloutTo: procAddr SpecOnStack: specOnStack in: calloutState
	<var: #procAddr type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	<var: #loadFloatRegs declareC: 'extern void loadFloatRegs(double, double, double, double)'>
	"Go out, call this guy and create the return value.  This *must* be inlined because of
	 the alloca of the outgoing stack frame in ffiCall:WithFlags:NumArgs:Args:AndTypes:"
	| myThreadIndex atomicType floatRet intRet |
	<var: #floatRet type: #double>
	<var: #intRet type: #usqLong>
	<inline: #always>
	myThreadIndex := interpreterProxy disownVM: (self disownFlagsFor: calloutState).

	calloutState floatRegisterSignature > 0 ifTrue:
		[self 
			load: (calloutState floatRegisters at: 0)
			Flo: (calloutState floatRegisters at: 1)
			atR: (calloutState floatRegisters at: 2)
			egs: (calloutState floatRegisters at: 3)].

	(self allocaLiesSoSetSpBeforeCall or: [self mustAlignStack]) ifTrue:
		[self setsp: calloutState argVector].

	atomicType := self atomicTypeOf: calloutState ffiRetHeader.
	(atomicType >> 1) = (FFITypeSingleFloat >> 1) ifTrue:
		[atomicType = FFITypeSingleFloat
			ifTrue:
				[floatRet := self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'float (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)]
			ifFalse: "atomicType = FFITypeDoubleFloat"
				[floatRet := self 
					dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'double (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
					with: (calloutState integerRegisters at: 0)
					with: (calloutState integerRegisters at: 1)
					with: (calloutState integerRegisters at: 2)
					with: (calloutState integerRegisters at: 3)].

		 "undo any callee argument pops because it may confuse stack management with the alloca."
		 (self isCalleePopsConvention: calloutState callFlags) ifTrue:
			[self setsp: calloutState argVector].
		 interpreterProxy ownVM: myThreadIndex.

		 ^interpreterProxy floatObjectOf: floatRet].

	intRet := self 
				dispatchFunctionPointer: (self cCoerceSimple: procAddr to: 'usqIntptr_t (*)(sqIntptr_t, sqIntptr_t, sqIntptr_t, sqIntptr_t)') 
				with: (calloutState integerRegisters at: 0)
				with: (calloutState integerRegisters at: 1)
				with: (calloutState integerRegisters at: 2)
				with: (calloutState integerRegisters at: 3).

	"undo any callee argument pops because it may confuse stack management with the alloca."
	(self isCalleePopsConvention: calloutState callFlags) ifTrue:
		[self setsp: calloutState argVector].
	interpreterProxy ownVM: myThreadIndex.

	(calloutState ffiRetHeader anyMask: FFIFlagPointer+FFIFlagStructure) ifTrue:
		["Note: Order is important here since FFIFlagPointer + FFIFlagStructure is used to represent
		 'typedef void* VoidPointer' and VoidPointer must be returned as pointer *not* as struct."
		 (calloutState ffiRetHeader anyMask: FFIFlagPointer) ifTrue:
			[^self ffiReturnPointer: intRet ofType: (self ffiReturnType: specOnStack) in: calloutState].
		 ^self ffiReturnStruct: (self addressOf: intRet) ofType: (self ffiReturnType: specOnStack) in: calloutState].
	
	^self ffiCreateIntegralResultOop: intRet ofAtomicType: atomicType in: calloutState! !

!ThreadedX64Win64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
ffiPushStructure: pointer ofSize: structSize typeSpec: argSpec ofLength: argSpecSize in: calloutState
	<var: #pointer type: #'void *'>
	<var: #argSpec type: #'sqInt *'>
	<var: #calloutState type: #'CalloutState *'>
	<var: #arg type: #usqLong>
	<inline: #always>
	structSize <= 0 ifTrue:
		[^FFIErrorStructSize].

	"See https://docs.microsoft.com/en-us/cpp/build/x64-calling-convention?view=vs-2019"
	(structSize <= WordSize
	 and: [(structSize bitAnd: structSize - 1) = 0 "a.k.a. structSize isPowerOfTwo"]) ifTrue:
		[| arg |
		self memcpy: (self addressOf: arg) _: pointer _: structSize.
		^self ffiPushUnsignedLongLong: arg in: calloutState].


	"BUG!!!! This memory should be 16-byte aligned; Spur guarantees only 8-byte alignment."
	self flag: #bug.
	^self ffiPushPointer: pointer in: calloutState! !

!ThreadedX64Win64FFIPlugin methodsFor: 'callout support' stamp: 'nice 1/30/2020 21:52'!
ffiReturnStruct: intRetPtr ofType: ffiRetType in: calloutState
	<var: #intRetPtr type: #'void *'>
	<var: #calloutState type: #'CalloutState *'>
	"Create a structure return value from an external function call.  The value has been stored in
	 alloca'ed space pointed to by the calloutState or in the return value passed by pointer."
	| retOop retClass oop |
	<inline: true>
	retClass := interpreterProxy fetchPointer: 1 ofObject: ffiRetType.
	retOop := interpreterProxy instantiateClass: retClass indexableSize: 0.
	self remapOop: retOop
		in: [oop := interpreterProxy 
					instantiateClass: interpreterProxy classByteArray 
					indexableSize: calloutState structReturnSize].
	self memcpy: (interpreterProxy firstIndexableField: oop)
		_: ((self returnStructInRegisters: calloutState)
				ifTrue: [intRetPtr]
				ifFalse: [calloutState limit])
		 _: calloutState structReturnSize.
	interpreterProxy storePointer: 0 ofObject: retOop withValue: oop.
	^retOop! !

!ThreadedX64Win64FFIPlugin methodsFor: 'marshalling' stamp: 'eem 07/27/2021 17:00'!
returnStructInRegisters: calloutState
	"Return through register if structReturnType is true"
	<var: #calloutState type: #'CalloutState *'>
	<inline: #always>
	^calloutState structReturnType! !

InterpreterPlugin subclass: #UnicodePlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!UnicodePlugin commentStamp: '<historical>' prior: 0!
This plugin measures and renders Unicode (UTF8) strings.
!

!UnicodePlugin class methodsFor: 'translation' stamp: 'ikp 7/30/2012 14:36'!
hasHeaderFile

	^true! !

!UnicodePlugin methodsFor: 'utility' stamp: 'eem 1/30/2014 14:36'!
asCString: stringOop
	"Return a C char * pointer into the given Squeak string object."
	"Warning: A Squeak string is not necessarily null-terminated."
	"Warning: the resulting pointer may become invalid after the next garbage collection and should only be using during the current primitive call."

	<returnTypeC: #'char *'>

	interpreterProxy success: (interpreterProxy isBytes: stringOop).

	^interpreterProxy failed
		ifTrue: [0]
		ifFalse: [self cCoerce: (interpreterProxy firstIndexableField: stringOop) to: #'char *']! !

!UnicodePlugin methodsFor: 'utility' stamp: 'eem 1/30/2014 14:34'!
cWordsPtr: oop minSize: minSize
	"Return a C pointer to the first indexable field of oop, which must be a words object of at least the given size."
	"Warning: the resulting pointer may become invalid after the next garbage collection and should only be using during the current primitive call."

	<returnTypeC: #'void *'>

	interpreterProxy success:
		((interpreterProxy isWords: oop)
		 and: [(interpreterProxy stSizeOf: oop) >= minSize]).
	^interpreterProxy failed
		ifTrue: [0]
		ifFalse: [self cCoerce: (interpreterProxy firstIndexableField: oop) to: #'void *']! !

!UnicodePlugin methodsFor: 'utility' stamp: 'eem 1/30/2014 14:32'!
copyString: stringOop into: stringPtr max: maxChars
	"Copy the Squeak string into a temporary buffer and add a terminating null byte. Fail if there is not sufficent space in the buffer."

	| srcPtr count |
	<inline: false>
	<var: 'stringPtr' type: #'char *'>
	<var: 'srcPtr' type: #'char *'>

	((interpreterProxy isBytes: stringOop)
	and: [(count := interpreterProxy stSizeOf: stringOop) < maxChars]) ifFalse:
		[interpreterProxy success: false.
		^ 0].

	srcPtr := self cCoerce: (interpreterProxy firstIndexableField: stringOop) to: #'char *'.
	1 to: count do: [:i | self cCode: '*stringPtr++ = *srcPtr++'].
	self cCode: '*stringPtr = 0'.
	^ 0! !

!UnicodePlugin methodsFor: 'primitives' stamp: 'eem 3/16/2018 14:40'!
primitiveClipboardGet
	"Read the clipboard into the given UTF16 string.."

	| utf16Oop utf16 utf16Length count |
	<export: true>
	<var: 'utf16' type: #'unsigned short *'>

	utf16Oop := interpreterProxy stackValue: 0.

	interpreterProxy success: (interpreterProxy isWords: utf16Oop).

	interpreterProxy failed ifTrue: [^ 0].

	utf16 := self cCoerce: (interpreterProxy firstIndexableField: utf16Oop) to: #'unsigned short *'.
	utf16Length := 2 * (interpreterProxy stSizeOf: utf16Oop).

	count := self unicodeClipboard: utf16 Get: utf16Length.

	^interpreterProxy methodReturnInteger: count! !

!UnicodePlugin methodsFor: 'primitives' stamp: 'eem 1/30/2014 14:25'!
primitiveClipboardPut
	"Set the clipboard to a UTF16 string.."

	| strOop count utf16 utf16Length |
	<export: true>
	<var: 'utf16' type: #'unsigned short *'>

	strOop := interpreterProxy stackValue: 1.
	count := interpreterProxy stackIntegerValue: 0.

	interpreterProxy success: (interpreterProxy isWords: strOop).

	interpreterProxy failed ifTrue: [^ 0].

	utf16 := self cCoerce: (interpreterProxy firstIndexableField: strOop) to: #'unsigned short *'.
	utf16Length := 2 * (interpreterProxy stSizeOf: strOop).
	((count >= 0) & (count < utf16Length)) ifTrue: [utf16Length := count].

	self unicodeClipboard: utf16 Put: utf16Length.

	interpreterProxy pop: 2.  "pop args, leave rcvr on stack"
	^ 0! !

!UnicodePlugin methodsFor: 'primitives' stamp: 'eem 3/16/2018 14:41'!
primitiveClipboardSize

	<export: true>
	^interpreterProxy methodReturnInteger: self unicodeClipboardSize! !

!UnicodePlugin methodsFor: 'primitives' stamp: 'jm (auto pragmas dtl 2014-01-04) 6/17/2008 10:51'!
primitiveDrawString

	| utf8Oop utf8 w h bitmapOop bitmapPtr utf8Length result |
	<export: true>
	<var: 'utf8' declareC: 'char *utf8'>
	<var: 'bitmapPtr' declareC: 'void *bitmapPtr'>

	utf8Oop := interpreterProxy stackValue: 3.
	utf8 := self asCString: utf8Oop.
	w := interpreterProxy stackIntegerValue: 2.
	h := interpreterProxy stackIntegerValue: 1.
	bitmapOop := interpreterProxy stackValue: 0.
	bitmapPtr := self cWordsPtr: bitmapOop minSize: w * h.
	interpreterProxy failed ifTrue: [^ nil].

	utf8Length := interpreterProxy stSizeOf: utf8Oop.
	self cCode: 'unicodeDrawString(utf8, utf8Length, &w, &h, bitmapPtr)'.

	result := interpreterProxy makePointwithxValue: w yValue: h.
	interpreterProxy pop: 5 thenPush: result.
	^ 0
! !

!UnicodePlugin methodsFor: 'primitives' stamp: 'eem 3/16/2018 14:43'!
primitiveGetFontList


	| strOop str strLength |
	<export: true>
	<var: 'str' declareC: 'char *str'>

	strOop := interpreterProxy stackValue: 0.
	str := self asCString: strOop.
	interpreterProxy failed ifFalse:
		[strLength := interpreterProxy stSizeOf: strOop.
		 interpreterProxy methodReturnInteger: (self unicodeGetFont: str List: strLength)]! !

!UnicodePlugin methodsFor: 'primitives' stamp: 'eem 3/16/2018 14:44'!
primitiveGetXRanges

	| utf8Oop utf8 resultOop resultPtr utf8Length count resultLength |
	<export: true>
	<var: 'utf8' declareC: 'char *utf8'>
	<var: 'resultPtr' declareC: 'int *resultPtr'>

	utf8Oop := interpreterProxy stackValue: 1.
	utf8 := self asCString: utf8Oop.

	resultOop := interpreterProxy stackValue: 0.
	resultPtr := self cWordsPtr: resultOop minSize: 0.
	interpreterProxy failed ifTrue: [^ nil].

	utf8Length := interpreterProxy stSizeOf: utf8Oop.
	resultLength := interpreterProxy stSizeOf: resultOop.
	count := self unicode: utf8 Get: utf8Length X: resultPtr Ranges: resultLength.

	^interpreterProxy methodReturnInteger: count! !

!UnicodePlugin methodsFor: 'primitives' stamp: 'jm (auto pragmas dtl 2014-01-04) 6/17/2008 10:51'!
primitiveMeasureString

	| utf8Oop utf8 utf8Length w h result |
	<export: true>
	<var: 'utf8' declareC: 'char *utf8'>

	utf8Oop := interpreterProxy stackValue: 0.
	utf8 := self asCString: utf8Oop.
	interpreterProxy failed ifTrue: [^ nil].

	w := h := 0.
	utf8Length := interpreterProxy stSizeOf: utf8Oop.
	self cCode: 'unicodeMeasureString(utf8, utf8Length, &w, &h)'.

	result := interpreterProxy makePointwithxValue: w yValue: h.
	interpreterProxy pop: 2 thenPush: result.
	^ 0
! !

!UnicodePlugin methodsFor: 'primitives' stamp: 'jm (auto pragmas dtl 2014-01-04) 1/24/2008 16:29'!
primitiveSetColors

	| fgRed fgGreen fgBlue bgRed bgGreen bgBlue mapBGToTransparent |
	<export: true>

	fgRed := interpreterProxy stackIntegerValue: 6.
	fgGreen := interpreterProxy stackIntegerValue: 5.
	fgBlue := interpreterProxy stackIntegerValue: 4.
	bgRed := interpreterProxy stackIntegerValue: 3.
	bgGreen := interpreterProxy stackIntegerValue: 2.
	bgBlue := interpreterProxy stackIntegerValue: 1.
	mapBGToTransparent := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifTrue: [^ nil].

	self cCode: 'unicodeSetColors(fgRed, fgGreen, fgBlue, bgRed, bgGreen, bgBlue, mapBGToTransparent)'.

	interpreterProxy pop: 7.
	^ 0
! !

!UnicodePlugin methodsFor: 'primitives' stamp: 'jm (auto pragmas dtl 2014-01-04) 1/13/2008 11:24'!
primitiveSetFont

	| fontName fontSize boldFlag italicFlag antiAliasFlag |
	<export: true>
	<var: 'fontName' declareC: 'char fontName[200]'>

	self copyString: (interpreterProxy stackValue: 4) into: fontName max: 200.
	fontSize := interpreterProxy stackIntegerValue: 3.
	boldFlag := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 2).
	italicFlag := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 1).
	antiAliasFlag := interpreterProxy booleanValueOf: (interpreterProxy stackValue: 0).
	interpreterProxy failed ifTrue: [^ nil].

	self cCode: 'unicodeSetFont(fontName, fontSize, boldFlag, italicFlag, antiAliasFlag)'.

	interpreterProxy pop: 5.
	^ 0
! !

InterpreterPlugin subclass: #VMProfileLinuxSupportPlugin
	instanceVariableNames: 'numModules primErr'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!VMProfileLinuxSupportPlugin commentStamp: 'eem 6/5/2013 12:10' prior: 0!
This class provides support for the VMProfiler on Linux (at least linux versions that provide dl_iterate_phdr).  This support is for reading symbols from the executable.  We can use the OS's nm command to list symbols in the VM executable and loaded libraries.  To do this we need to know what libraries are loaded, not simply which libraries are linked against, since some libraries are loaded dynamically, and for each loaded library what the vm address relocation, if any, is for the loaded library.  

Further, we need to know the actual addresses in memory of symbols in the program and in memory.  Since the OS may be using address randomization we cannot assume that the addresses answered by nm for symbols in the program and/or libraries match the addresses of the same symbols in memory.  Instead we must correlate.  dlsym can be used to look up symbols in loaded dynamic load libraries, but it does _not_ answer the addresses of symbols in the main program.  Back in the day nlist could be used to do this, but it is no more.  Instead this module contains a reference to interpret and answers its address ia a primitive (alas this means the plugin must be internal, but it's extremely small, and the VM links against libdl.so anyway).  Any address space shift will therefore be the difference between nm's output for interpret and the primitive's value.  We can similarly compute the address shift for libraries by using dlsym to lookup a symbol in a library and comparing it to nm's output for the library.


The primitive primitiveExecutableModules returns the names of the executable and the loaded libraries.!

!VMProfileLinuxSupportPlugin class methodsFor: 'translation' stamp: 'topa 6/16/2018 21:05'!
declareHeaderFilesIn: cg

	cg
		addHeaderFile: '<limits.h>';
		addHeaderFile: '#if defined(HAVE_DLFCN_H)\# include <dlfcn.h>\#endif' withCRs;
		addHeaderFile: '<link.h>';
		addHeaderFile: '#ifndef RTLD_NODELETE\# define RTLD_NODELETE 0\#endif' withCRs! !

!VMProfileLinuxSupportPlugin methodsFor: 'iteration callbacks' stamp: 'eem 12/11/2014 17:42'!
count: info num: size modules: ignored
	<returnTypeC: #int>
	<var: #info type: #'struct dl_phdr_info *'>
	<var: #size type: #'size_t'>
	<var: #ignored type: #'void *'>
	numModules := numModules + 1.
	^0! !

!VMProfileLinuxSupportPlugin methodsFor: 'primitives' stamp: 'eem 9/19/2018 18:02'!
primitiveDLSymInLibrary
	"Answer the address of the symbol whose name is the first argument
	 in the library whose name is the second argument, or nil if none."
	| nameObj symName libName lib sz addr ok |
	<export: true>
	<var: #symName type: #'char *'>
	<var: #libName type: #'char *'>
	<var: #lib type: #'void *'>
	<var: #addr type: #'void *'>
	nameObj := interpreterProxy stackValue: 0.
	(interpreterProxy isBytes: nameObj) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	sz := interpreterProxy byteSizeOf: nameObj.
	libName := self malloc: sz+1.
	self strncpy: libName _: (interpreterProxy firstIndexableField: nameObj) _: sz.
	libName at: sz put: 0.
	nameObj := interpreterProxy stackValue: 1.
	(interpreterProxy isBytes: nameObj) ifFalse:
		[self free: libName.
		 ^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	sz := interpreterProxy byteSizeOf: nameObj.
	symName := self malloc: sz+1.
	self strncpy: symName _: (interpreterProxy firstIndexableField: nameObj) _: sz.
	symName at: sz put: 0.
	lib := self dl: libName open: (#'RTLD_LAZY' bitOr: #'RTLD_NODELETE').
	lib ifNil:
		[self free: libName; free: symName.
		 ^interpreterProxy primitiveFailFor: PrimErrInappropriate].
	self dlerror. "clear dlerror"
	addr := self dl: lib sym: symName.
	ok := self dlerror isNil.
	self free: symName.
	self free: libName.
	self dlclose: lib.
	ok ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrNotFound].
	^interpreterProxy methodReturnValue: (interpreterProxy positiveMachineIntegerFor: addr asUnsignedIntegerPtr)! !

!VMProfileLinuxSupportPlugin methodsFor: 'primitives' stamp: 'eem 11/2/2013 00:26'!
primitiveExecutableModules
	"Answer an Array of pairs of strings for executable modules (the VM executable and loaded libraries).
	 The first element in each pair is the filename of the module.  The second element is either nil or
	 the symlink's target, if the filename is a symlink."
	<export: true>
	| resultObj |
	numModules := 0.
	self cCode: 'dl_iterate_phdr(countnummodules,0)' inSmalltalk: [0].
	resultObj := interpreterProxy
					instantiateClass: interpreterProxy classArray
					indexableSize: numModules - 1 * 2. "skip the fake linux-gate.so.1"
	resultObj = 0 ifTrue:
		[^interpreterProxy primitiveFailFor: PrimErrNoMemory].
	interpreterProxy pushRemappableOop: resultObj.
	primErr := numModules := 0.
	self cCode: 'dl_iterate_phdr(reapmodulesymlinks,0)' inSmalltalk: [0].
	resultObj := interpreterProxy popRemappableOop.
	primErr ~= 0 ifTrue:
		[^interpreterProxy primitiveFailFor: primErr].
	^interpreterProxy methodReturnValue: resultObj! !

!VMProfileLinuxSupportPlugin methodsFor: 'primitives' stamp: 'nice 7/8/2016 02:43'!
primitiveInterpretAddress
	"Answer the address of the interpret routine."
	<export: true>
	| interpret |
	<var: #interpret declareC: 'extern void interpret()'>
	^interpreterProxy methodReturnValue: (interpreterProxy positiveMachineIntegerFor: interpret asUnsignedIntegerPtr)! !

!VMProfileLinuxSupportPlugin methodsFor: 'iteration callbacks' stamp: 'eem 9/19/2018 18:02'!
reap: info module: size names: ignored
	<returnTypeC: #int>
	| elfModuleName len moduleNameObj GetAttributeString |
	<var: #info type: #'struct dl_phdr_info *'>
	<var: 'elfModuleName' type: #'const char *'>
	<var: #GetAttributeString declareC: 'extern char *GetAttributeString(sqInt)'>
	<var: #size type: #'size_t'>
	<var: #ignored type: #'void *'>
	self touch: GetAttributeString.
	elfModuleName := self cCode: 'numModules ? info->dlpi_name : GetAttributeString(0)'.
	(elfModuleName isNil
	 or: [(len := self strlen: elfModuleName) = 0]) ifTrue:
		[^0]. "skip the fake linux-gate.so.1"
	moduleNameObj := interpreterProxy
							instantiateClass: interpreterProxy classString
							indexableSize: len.
	moduleNameObj = 0 ifTrue:
		[primErr := PrimErrNoMemory.
		 ^1]. "stop iteration"
	self strncpy: (interpreterProxy arrayValueOf: moduleNameObj)
		_: elfModuleName
		_: len. "(char *)strncpy()"
	interpreterProxy
		storePointer: numModules
		ofObject: interpreterProxy topRemappableOop
		withValue: moduleNameObj.
	numModules := numModules + 1.
	^0! !

!VMProfileLinuxSupportPlugin methodsFor: 'iteration callbacks' stamp: 'eem 9/19/2018 18:02'!
reap: info module: size symlinks: ignored
	"like reap:module:names:, but follows symlinks"
	<returnTypeC: #int>
	| elfModuleName len moduleNameObj GetAttributeString symLinkBuf |
	<var: #info type: #'struct dl_phdr_info *'>
	<var: 'elfModuleName' type: #'const char *'>
	<var: #GetAttributeString declareC: 'extern char *GetAttributeString(sqInt)'>
	<var: #symLinkBuf declareC: 'char symLinkBuf[PATH_MAX]'>
	<var: #size type: #'size_t'>
	<var: #ignored type: #'void *'>
	self touch: GetAttributeString.
	elfModuleName := self cCode: 'numModules ? info->dlpi_name : GetAttributeString(0)'.
	(elfModuleName isNil
	 or: [(len := self strlen: elfModuleName) = 0]) ifTrue:
		[^0]. "skip the fake linux-gate.so.1"
	moduleNameObj := interpreterProxy
							instantiateClass: interpreterProxy classString
							indexableSize: len.
	moduleNameObj = 0 ifTrue:
		[primErr := PrimErrNoMemory.
		 ^1]. "stop iteration"
	self strncpy: (interpreterProxy arrayValueOf: moduleNameObj)
		_: elfModuleName
		_: len. "(char *)strncpy()"
	interpreterProxy
		storePointer: numModules
		ofObject: interpreterProxy topRemappableOop
		withValue: moduleNameObj.
	"now dereference the symlink, if it exists"
	self str: symLinkBuf cpy: elfModuleName.
	(len := self read: elfModuleName li: symLinkBuf nk: #'PATH_MAX') > 0
		ifTrue:
			[moduleNameObj := interpreterProxy
									instantiateClass: interpreterProxy classString
									indexableSize: len.
			 moduleNameObj = 0 ifTrue:
				[primErr := PrimErrNoMemory.
				 ^1]. "stop iteration"
			 self strncpy: (interpreterProxy arrayValueOf: moduleNameObj)
				_: symLinkBuf
				_: len. "(char *)strncpy()"
			 interpreterProxy
				storePointer: numModules + 1
				ofObject: interpreterProxy topRemappableOop
				withValue: moduleNameObj]
		ifFalse:
			[interpreterProxy
				storePointer: numModules + 1
				ofObject: interpreterProxy topRemappableOop
				withValue: interpreterProxy nilObject].
	numModules := numModules + 2.
	^0! !

InterpreterPlugin subclass: #VMProfileMacSupportPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!VMProfileMacSupportPlugin commentStamp: '<historical>' prior: 0!
This class provides support for the VMProfiler on MacOS X.  This support is for reading symbols from the executable.  We can use the OS's nm command to list symbols in the VM executable and loaded libraries.  To do this we need to know what libraries are loaded, not simply which libraries are linked against, since some libraries are loaded dynamically, and for each loaded library what the vm address relocation, if any, is for the loaded library.  The primitive primitiveExecutableModulesAndOffsets returns these as an array of pairs, module name to relocation.!

!VMProfileMacSupportPlugin class methodsFor: 'translation' stamp: 'eem 12/14/2012 11:21'!
declareCVarsIn: aCCodeGen
	aCCodeGen
		addHeaderFile: '<dlfcn.h>';
		addHeaderFile: '<mach-o/dyld.h>';
		addHeaderFile: '<mach-o/getsect.h>'! !

!VMProfileMacSupportPlugin methodsFor: 'primitives' stamp: 'nice 7/8/2016 02:43'!
primitiveDLSym
	"Answer the address of the argument in the current process or nil if none."
	| nameObj name namePtr sz addr |
	<export: true>
	<var: #name type: #'char *'>
	<var: #namePtr type: #'char *'>
	<var: #addr type: #'void *'>
	nameObj := interpreterProxy stackValue: 0.
	(interpreterProxy isBytes: nameObj) ifFalse:
		[^interpreterProxy primitiveFailFor: PrimErrBadArgument].
	sz := interpreterProxy byteSizeOf: nameObj.
	name := self malloc: sz+1.
	namePtr := interpreterProxy firstIndexableField: nameObj.
	0 to: sz-1 do:[:i| name at: i put: (namePtr at: i)].
	name at: sz put: 0.
	addr := self cCode: 'dlsym(RTLD_SELF,name)' inSmalltalk: [0].
	self free: name.
	^interpreterProxy methodReturnValue: (addr = 0
												ifTrue: [interpreterProxy nilObject]
												ifFalse: [interpreterProxy positiveMachineIntegerFor: addr asUnsignedIntegerPtr])! !

!VMProfileMacSupportPlugin methodsFor: 'primitives' stamp: 'eem 9/19/2018 17:31'!
primitiveExecutableModulesAndOffsets
	"Answer an Array of quads for executable modules (the VM executable
	 and loaded libraries).  Each quad is the module's name, its vm address
	 relocation in memory, the (unrelocated) start address, and the size."
	| nimages resultObj name valueObj nameObjData slide start size |
	<export: true>
	<var: #name type: 'const char *'>
	<var: #nameObjData type: #'char *'>
	<var: #h type: 'const struct mach_header *'>
	<var: #h64 type: 'const struct mach_header_64 *'>
	<var: #s64 type: 'const struct section_64 *'>
	<var: #s type: 'const struct section *'>
	<var: #start type: 'usqIntptr_t'>
	<var: #slide type: 'usqIntptr_t'>
	<var: #size type: 'usqIntptr_t'>
	self cppIf: #'MAC_OS_X_VERSION_MIN_REQUIRED' <= #'MAC_OS_X_VERSION_10_4'
		ifTrue: "_dyld_present was deprecated in 10.5"
			[(self cCode: '_dyld_present()' inSmalltalk: false) ifFalse:
				[^interpreterProxy primitiveFail]].
	nimages := self cCode: '_dyld_image_count()' inSmalltalk: 0.
	resultObj := interpreterProxy instantiateClass: interpreterProxy classArray indexableSize: nimages * 4.
	resultObj = 0 ifTrue:
		[^interpreterProxy primitiveFail].

	interpreterProxy pushRemappableOop: resultObj.
	0 to: nimages - 1 do:
		[:i|
		start := size := -1. "impossible start & size"
		name := self cCode: '_dyld_get_image_name(i)' inSmalltalk: 0.
		slide   := self cCode: '_dyld_get_image_vmaddr_slide(i)' inSmalltalk: 0.
		self cppIf: #'__x86_64__'
			ifTrue:
				[(self cCode: '(const struct mach_header_64 *)_dyld_get_image_header(i)' inSmalltalk: nil) ifNotNil:
					[:h64|
					 (self cCode: 'getsectbynamefromheader_64(h64,SEG_TEXT,SECT_TEXT)' inSmalltalk: nil) ifNotNil:
						[:s64|
						 start := self cCode: 's64->addr' inSmalltalk: 0.
						 size := self cCode: 's64->size' inSmalltalk: 0]]]
			ifFalse:
				[(self cCode: '_dyld_get_image_header(i)' inSmalltalk: nil) ifNotNil:
					[:h|
					 (self cCode: 'getsectbynamefromheader(h,SEG_TEXT,SECT_TEXT)' inSmalltalk: nil) ifNotNil:
						[:s|
						 start := self cCode: 's->addr' inSmalltalk: 0.
						 size := self cCode: 's->size' inSmalltalk: 0]]].

		valueObj := interpreterProxy
						instantiateClass: interpreterProxy classString
						indexableSize: (self strlen: name).
		interpreterProxy failed ifTrue:
			[interpreterProxy popRemappableOop.
			 ^interpreterProxy primitiveFail].
		interpreterProxy storePointer: i * 4 ofObject: interpreterProxy topRemappableOop withValue: valueObj.
		nameObjData := interpreterProxy arrayValueOf: valueObj.
		self memcpy: nameObjData _: name _: (self strlen: name).

		valueObj := interpreterProxy signedMachineIntegerFor: slide.
		interpreterProxy failed ifTrue:
			[interpreterProxy popRemappableOop.
			 ^interpreterProxy primitiveFail].
		interpreterProxy storePointer: i * 4 + 1 ofObject: interpreterProxy topRemappableOop withValue: valueObj.

		valueObj := interpreterProxy positiveMachineIntegerFor: start.
		interpreterProxy failed ifTrue:
			[interpreterProxy popRemappableOop.
			 ^interpreterProxy primitiveFail].
		interpreterProxy storePointer: i * 4 + 2 ofObject: interpreterProxy topRemappableOop withValue: valueObj.

		valueObj := interpreterProxy positiveMachineIntegerFor: size.
		interpreterProxy failed ifTrue:
			[interpreterProxy popRemappableOop.
			 ^interpreterProxy primitiveFail].
		interpreterProxy storePointer: i * 4 + 3 ofObject: interpreterProxy topRemappableOop withValue: valueObj].

	resultObj := interpreterProxy popRemappableOop.
	^interpreterProxy pop: 1 thenPush: resultObj! !

InterpreterPlugin subclass: #WeDoPlugin
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins'!

!WeDoPlugin commentStamp: 'jm 12/2/2008 14:22' prior: 0!
Low level interface to the Lego WeDo.
!

!WeDoPlugin class methodsFor: 'translation' stamp: 'ikp 7/30/2012 14:37'!
hasHeaderFile

	^true! !

!WeDoPlugin methodsFor: 'translated prims' stamp: 'jm (auto pragmas dtl 2014-01-04) 5/12/2009 15:14'!
primClosePort
	"Close the WeDo port."

	<export: true>
	interpreterProxy success: (self cCode: 'WeDoClosePort()').
	^ 0
! !

!WeDoPlugin methodsFor: 'translated prims' stamp: 'jm (auto pragmas dtl 2014-01-04) 12/3/2008 09:32'!
primOpenPort
	"Open the WeDo port."

	<export: true>
	interpreterProxy success: (self cCode: 'WeDoOpenPort()').
	^ 0
! !

!WeDoPlugin methodsFor: 'translated prims' stamp: 'eem 1/30/2014 12:41'!
primRead
	"Read data from the WeDo port into the given buffer (a ByteArray or String). Answer the number of bytes read."

	| bufOop bufPtr bufSize byteCount |
	<export: true>
	<var: 'bufPtr' declareC: 'char *bufPtr'>

	bufOop := interpreterProxy stackValue: 0.
	(interpreterProxy isBytes: bufOop) ifFalse: [
		interpreterProxy success: false.
		^ 0].
	bufPtr := self cCoerce: (interpreterProxy firstIndexableField: bufOop) to: 'char *'.
	bufSize := interpreterProxy stSizeOf: bufOop.
	interpreterProxy failed ifTrue: [^ 0].

	byteCount := self cCode: 'WeDoRead(bufPtr, bufSize)'.
	byteCount < 0 ifTrue: [
		interpreterProxy success: false.
		^ 0].

	interpreterProxy pop: 2.  					"pop args and rcvr"
	interpreterProxy pushInteger: byteCount.	"push result"
	^ 0
! !

!WeDoPlugin methodsFor: 'translated prims' stamp: 'eem 1/30/2014 12:41'!
primWrite
	"Write data to the WeDo port from the given buffer (a ByteArray or String). Answer the number of bytes written."

	| bufOop bufPtr bufSize byteCount |
	<export: true>
	<var: 'bufPtr' declareC: 'char *bufPtr'>

	bufOop := interpreterProxy stackValue: 0.
	(interpreterProxy isBytes: bufOop) ifFalse: [
		interpreterProxy success: false.
		^ 0].
	bufPtr := self cCoerce: (interpreterProxy firstIndexableField: bufOop) to: 'char *'.
	bufSize := interpreterProxy stSizeOf: bufOop.
	interpreterProxy failed ifTrue: [^ 0].

	byteCount := self cCode: 'WeDoWrite(bufPtr, bufSize)'.
	byteCount < 0 ifTrue: [
		interpreterProxy success: false.
		^ 0].

	interpreterProxy pop: 2.  					"pop args and rcvr"
	interpreterProxy pushInteger: byteCount.	"push result"
	^ 0
! !

VMClass subclass: #InterpreterPrimitives
	instanceVariableNames: 'objectMemory messageSelector argumentCount newMethod primFailCode osErrorCode exceptionPC inFFIFlags profileMethod profileProcess profileSemaphore nextProfileTick preemptionYields newFinalization sHEAFn ffiExceptionResponse eventTraceMask'
	classVariableNames: 'CrossedX EndOfRun MillisecondClockMask'
	poolDictionaries: 'VMBasicConstants VMBytecodeConstants VMMethodCacheConstants VMObjectIndices VMSqueakClassIndices VMStackFrameOffsets'
	category: 'VMMaker-Interpreter'!

!InterpreterPrimitives commentStamp: 'eem 8/24/2018 11:05' prior: 0!
InterpreterPrimitives implements most of the VM's core primitives.  It is the root of the interpreter hierarchy so as to share the core primitives amongst the varioius interpreters.

Instance Variables
	argumentCount			<Integer>
	ffiExceptionResponse		<Integer>
	inFFIFlags					<Integer>
	messageSelector			<Integer>
	newMethod					<Integer>
	nextProfileTick				<Integer>
	objectMemory				<ObjectMemory> (simulation only)
	preemptionYields			<Boolean>
	primFailCode				<Integer>
	osErrorCode				<Integer>
	profileMethod				<Integer>
	profileProcess				<Integer>
	profileSemaphore			<Integer>
	secHasEnvironmentAccess <Integer>

argumentCount
	- the number of arguments of the current message

ffiExceptionResponse
	- controls system response to exceptions during FFI calls.  See primitiveFailForFFIException:at:

inFFIFlags
	- flags recording currently only whether the system is in an FFI call

messageSelector
	- the oop of the selector of the current message

newMethod
	- the oop of the result of looking up the current message

nextProfileTick
	- the millisecond clock value of the next profile tick (if profiling is in effect)

objectMemory
	- the memory manager and garbage collector that manages the heap

preemptionYields
	- a boolean controlling the process primitives.  If true (old, incorrect, blue-book semantics) a preempted process is sent to the back of its run-queue.  If false, a process preempted by a higher-priority process is put back at the head of its run queue, hence preserving cooperative scheduling within priorities.

primFailCode
	- primitive success/failure flag, 0 for success, otherwise the reason code for failure

osErrorCode
	- a 64-bit value settable by external primitives conveying arbitrary error codes from the operating system and/or system libraries

profileMethod
	- the oop of the method at the time nextProfileTick was reached

profileProcess
	- the oop of the activeProcess at the time nextProfileTick was reached

profileSemaphore
	- the oop of the semaphore to signal when nextProfileTick is reached

secHasEnvironmentAccess
	- the function to call to check if access to the envronment should be granted to primitiveGetenv
!

!InterpreterPrimitives class methodsFor: 'C translation' stamp: 'eem 8/24/2018 12:41'!
declareCVarsIn: aCCodeGen
	aCCodeGen
		var: 'osErrorCode' type: #sqLong;
		var: 'exceptionPC' type: #usqInt;
		var: 'sHEAFn' declareC: 'int (*sHEAFn)() = 0' "the hasEnvironmentAccess function"! !

!InterpreterPrimitives class methodsFor: 'C translation' stamp: 'eem 4/28/2016 21:09'!
defineAtCompileTime: anObject
	"Override to define at translation time those variables that need to
	 be defined at compile time only in plugins, but not in the main VM,
	 because the VM generated is specific to these varables."
	anObject isSymbol ifFalse:
		[^false].
	(#(STACKVM COGVM COGMTVM SPURVM) includes: anObject) ifTrue:
		[^false].
	^VMBasicConstants namesDefinedAtCompileTime includes: anObject! !

!InterpreterPrimitives class methodsFor: 'class initialization' stamp: 'eem 4/18/2017 11:27'!
initializeMiscConstants
	"Initialize the hashMultiply constants."
	super initializeMiscConstants.
	HashMultiplyConstant := 1664525.
	HashMultiplyMask := 16rFFFFFFF "(2 raisedTo: 28) - 1"! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'nice 11/6/2016 13:55'!
asUnsigned: anInteger
	<inline: true>
	^self cCode: [anInteger asUnsignedInteger] inSmalltalk: [anInteger bitAnd: objectMemory maxCInteger]! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 10/23/2018 10:45'!
cStringOrNullFor: oop
	"Answer either a malloced string with the null-terminated contents of oop if oop is a string,
	 or the null pointer if oop is nil, or fail.  It is the client's responsibility to free the string later."
	<api>
	<returnTypeC: #'char *'>
	<inline: false>
	| isString len cString |
	<var: 'cString' type: #'char *'>
	isString := self isInstanceOfClassByteString: oop.
	isString ifFalse:
		[oop ~= objectMemory nilObject ifTrue:
			[self primitiveFailFor: PrimErrBadArgument].
		 ^0].
	len := objectMemory lengthOf: oop.
	len = 0 ifTrue:
		[^0].
	cString := self malloc: len + 1.
	cString ifNil:
		[self primitiveFailFor: PrimErrNoCMemory.
		 ^0].
	self memcpy: cString _: (objectMemory firstIndexableField: oop) _: len.
	cString at: len put: 0.
	^cString! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 7/25/2020 18:57'!
canBeImmutable: oop
	<option: #IMMUTABILITY>
	| scheduler processLists |
	
	self assert: (objectMemory isNonImmediate: oop).
	
	"For now we fail the primitive for contexts to we ensure there are no immutable contexts.
	Later we can consider having immutable contexts and send cannotReturn callback
	when returning to an immutable context. That would mean that setting a context 
	to immutable would require a divorce and returns to immutable context are 
	necessarily across stack pages"
	(objectMemory isContext: oop) ifTrue: [^ false].
	
	"Weak structures can't be immutable"
	(objectMemory isEphemeron: oop) ifTrue: [^ false].
	(objectMemory isWeakNonImm: oop) ifTrue: [^ false].
	
	"No clue what is going on for semaphores so they can't be immutable"
	(objectMemory isSemaphoreObj: oop) ifTrue: [^ false].
	
	"Simple version of process management: we forbid Process and LinkedList instances to be immutable 
	 as well as the Processor and the array of activeProcess"
	scheduler := objectMemory fetchPointer: ValueIndex ofObject: (objectMemory splObj: SchedulerAssociation).
	processLists := objectMemory fetchPointer: ProcessListsIndex ofObject: scheduler.
	oop = scheduler ifTrue: [^ false].
	oop = processLists ifTrue: [^ false].
	"Is it a linkedList ?"
	(objectMemory classIndexOf: (objectMemory fetchPointer: 1 ofObject: processLists)) = (objectMemory classIndexOf: oop) ifTrue: [^ false].
	"is it a Process ?"
	(objectMemory classIndexOf: (objectMemory fetchPointer: ActiveProcessIndex ofObject: scheduler)) =  (objectMemory classIndexOf: oop) ifTrue: [^ false].
	
	"The rest of the code is relative to process management: the Processor (the active 
	process scheduler) can't be immutable, as well as all the objects relative to Process management "
	"scheduler := self fetchPointer: ValueIndex ofObject: (self splObj: SchedulerAssociation).
	processLists := objectMemory fetchPointer: ProcessListsIndex ofObject: scheduler.
	((objectMemory formatOf: oop) = objectMemory nonIndexablePointerFormat)
		ifFalse: 
			[ (objectMemory isArrayNonImm: oop) ifFalse: [^ true].
			  ^ (oop = processLists) not ].
	(objectMemory numSlotsOf: oop) >= 2 ifFalse: [^ true].
	""is the oop the scheduler itself ?""
	oop = scheduler ifTrue: [^ false].
	1 to: (objectMemory numSlotsOf: processLists) do: [ :i |
		""is the oop one of the linked lists ?""
		(list := processLists at: i) = oop ifTrue: [^ false].
		""is the oop one of the runnable process ?""
		first := objectMemory fetchPointer: FirstLinkIndex ofObject: list.
		first = objectMemory nilObject ifFalse: 
			[ last := objectMemory fetchPointer: LastLinkIndex ofObject: list.
			  link := first.
			  [ link = last ] whileFalse: 
				[ link = oop ifTrue: [^ false]. 
				  link := objectMemory fetchPointer: NextLinkIndex ofObject: link. ] ] ]."
	^ true! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 6/3/2016 11:32'!
failed
	<api>
	<inline: true>
	"In C, non-zero is true, so avoid computation by simply answering primFailCode in the C version."
	^self cCode: [primFailCode] inSmalltalk: [primFailCode ~= 0]! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 2/12/2015 11:15'!
floatExponentBits
	"Answer the number of bits in the double-precision exponent.  This is an 11-bit field."
	<inline: true>
	^11! !

!InterpreterPrimitives methodsFor: 'simulation support' stamp: 'akg 10/23/2018 22:30'!
getenv: aByteStringOrByteArray
	<doNotGenerate>
	"The primitiveGetenv: primitive answers nil for undefined variables.
	 The primitiveGetenv implementation is written to expect getenv: to
	 answer 0, not nil,  for undefined variables.  Map nil to 0 for simulation."

	"aByteStringOrByteArray is probably null terminated, convert to non-null-terminated"
	^(self primitiveGetenv: (self asString: aByteStringOrByteArray)) ifNil: [0]! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 5/7/2020 19:30'!
identityHashOf: anOop
	"Answer the identity hash of an object, assigning a hash if it doesn't have one.
	 On Spur refuse to assign a hash to something that looks like a behavior."
	<api>
	(objectMemory isImmediate: anOop) ifTrue:
		[^self
			immediateAsInteger: anOop
			ifFail: [-1]].
	(objectMemory hasSpurMemoryManagerAPI
	 and: [(objectMemory rawHashBitsOf: anOop) = 0
	 and: [objectMemory objCouldBeClassObj: anOop]]) ifTrue:
		[^-1].
	^objectMemory hashBitsOf: anOop! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 5/7/2020 19:30'!
immediateAsInteger: anOop ifFail: aBlock
	<inline: #always>
	(objectMemory isIntegerObject: anOop) ifTrue:
		[^objectMemory integerValueOf: anOop].
	(objectMemory isCharacterObject: anOop) ifTrue:
		[^objectMemory characterValueOf: anOop].
	(objectMemory isImmediateFloat: anOop) ifTrue:
		[^objectMemory rotatedFloatBitsOf: anOop].
	^aBlock value! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 9/12/2010 06:35'!
initPrimCall
	"Set the failure code/success flag in preparation for calling a primitve.
	 If primFailCode is non-zero a primitive has failed.  If primFailCode is
	 greater than one then its value indicates the reason for failure."
	<inline: true>
	primFailCode := 0! !

!InterpreterPrimitives methodsFor: 'initialization' stamp: 'eem 12/22/2020 14:35'!
initialize
	"Here we can initialize the variables C initializes to zero.  #initialize methods do /not/ get translated."
	argumentCount := primFailCode := nextProfileTick := osErrorCode := exceptionPC := inFFIFlags := ffiExceptionResponse := eventTraceMask := 0.
	newFinalization := false! !

!InterpreterPrimitives methodsFor: 'initialization' stamp: 'eem 3/17/2017 11:24'!
initializeInterpreter: bytesToShift
	sHEAFn := self ioLoadFunction: 'secHasEnvironmentAccess' From: 'SecurityPlugin'! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 12/1/2014 14:14'!
isAppropriateForCopyObject: oop
	^objectMemory isPointersNonImm: oop! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 10/27/2020 21:38'!
isDirectAlien: oop
	^(objectMemory sizeFieldOfAlien: oop) > 0! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 2/11/2015 17:25'!
isFinite: aDouble
	<var: #aDouble type: #double>
	<inline: true>
	^aDouble - aDouble = 0.0! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 10/27/2020 21:38'!
isIndirectAlien: oop
	^(objectMemory sizeFieldOfAlien: oop) < 0! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 12/20/2012 09:56'!
isInstanceOfClassArray: oop
	<inline: true>
	"N.B.  Because Slang always inlines is:instanceOf:compactClassIndex:
	 (because is:instanceOf:compactClassIndex: has an inline: pragma) the
	 phrase (objectMemory splObj: ClassArray) is expanded in-place and
	 is _not_ evaluated if oop has a non-zero CompactClassIndex."
	^objectMemory
		is: oop
		instanceOf: (objectMemory splObj: ClassArray) 
		compactClassIndex: ClassArrayCompactIndex! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 12/20/2012 10:00'!
isInstanceOfClassBlockClosure: oop
	<inline: true>
	"N.B.  Because Slang always inlines is:instanceOf:compactClassIndex:
	 (because is:instanceOf:compactClassIndex: has an inline: pragma) the
	 phrase (objectMemory splObj: ClassBlockClosure) is expanded in-place
	 and is _not_ evaluated if oop has a non-zero CompactClassIndex."
	^objectMemory
		is: oop
		instanceOf: (objectMemory splObj: ClassBlockClosure) 
		compactClassIndex: ClassBlockClosureCompactIndex! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 1/2/2013 17:37'!
isInstanceOfClassByteString: oop
	<inline: true>
	"N.B.  Because Slang always inlines is:instanceOf:compactClassIndex:
	 (because is:instanceOf:compactClassIndex: has an inline: pragma) the
	 phrase (objectMemory splObj: ClassByteString) is expanded in-place and
	 is _not_ evaluated if oop has a non-zero CompactClassIndex."
	^objectMemory
		is: oop
		instanceOf: (objectMemory splObj: ClassByteString) 
		compactClassIndex: ClassByteStringCompactIndex! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 7/2/2021 06:43'!
isInstanceOfClassFullBlockClosure: oop
	<inline: true>
	"N.B.  Because Slang always inlines is:instanceOf:compactClassIndex:
	 (because is:instanceOf:compactClassIndex: has an inline: pragma) the
	 phrase (objectMemory splObj: ClassFullBlockClosure) is expanded in-place
	 and is _not_ evaluated if oop has a non-zero CompactClassIndex."
	^objectMemory
		is: oop
		instanceOf: (objectMemory splObj: ClassFullBlockClosure) 
		compactClassIndex: ClassFullBlockClosureCompactIndex! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 10/31/2014 11:40'!
isNegativeIntegerValueOf: oop
	"Answer true if integer object is negative.
	Fail if object pointed by oop i not an integer."
	| ok smallInt |

	(objectMemory isIntegerObject: oop) ifTrue:
		[smallInt := objectMemory integerValueOf: oop.
		^smallInt < 0].
	
	(objectMemory isNonIntegerNonImmediate: oop) ifTrue:
		[ok := objectMemory isClassOfNonImm: oop
						equalTo: (objectMemory splObj: ClassLargePositiveInteger)
						compactClassIndex: ClassLargePositiveIntegerCompactIndex.
		 ok ifTrue: [^false].
			
		 ok := objectMemory isClassOfNonImm: oop
								equalTo: (objectMemory splObj: ClassLargeNegativeInteger)
								compactClassIndex: ClassLargeNegativeIntegerCompactIndex.
		 ok ifTrue: [^true]].
	self primitiveFail.
	^false! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 10/27/2020 21:38'!
isPointerAlien: oop
	^(objectMemory sizeFieldOfAlien: oop) = 0! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 12/18/2018 17:39'!
isPositiveMachineIntegerObject: oop
	"Answer if oop is a value of an integer in address range, i.e up to the size of a machine word.
	The object may be either a positive SmallInteger or a LargePositiveInteger of size <= word size."
	| ok |
	(objectMemory isIntegerObject: oop) ifTrue:
		[^(objectMemory integerValueOf: oop) >= 0].

	(objectMemory isNonIntegerImmediate: oop) ifTrue:
		[^false].

	ok := objectMemory
			isClassOfNonImm: oop
			equalTo: (objectMemory splObj: ClassLargePositiveInteger)
			compactClassIndex: ClassLargePositiveIntegerCompactIndex.
	^ok and: [(objectMemory numBytesOfBytes: oop) <= (self sizeof: #'usqIntptr_t')]! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 6/29/2018 16:41'!
isSignedInteger: integer inRangeForBits: nBits
	"Answer if integer will fit within a variable of nBits, where nBits is 8, 16, 32 or 64.
	 Signed shift right by nBits - 1 to map in-range values to either 0 or -1.
	 Add one to map in-range values to 0 or 1.
	 Perform an unsigned comparison for greater than 1 to eliminate values out of range."
	<inline: true>
	^self cCode: [(self asUnsigned: (integer signedBitShift: 1 - nBits) + 1) <= 1]
		inSmalltalk: [((integer bitShift: 1 - nBits) + 1 bitAnd: objectMemory maxCInteger) <= 1]! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 4/28/2016 20:50'!
magnitude64BitIntegerFor: magnitude neg: isNegative
	"Return a Large Integer object for the given integer magnitude and sign"
	| newLargeInteger largeClass highWord sz isSmall smallVal |
	<var: 'magnitude' type: #usqLong>
	<var: 'highWord' type: #usqInt>

	isSmall := isNegative
				ifTrue: [magnitude <= (objectMemory maxSmallInteger + 1)]
				ifFalse: [magnitude <= objectMemory maxSmallInteger].
	isSmall ifTrue:
		[smallVal := self cCoerceSimple: magnitude to: #sqInt.
		 isNegative ifTrue: [smallVal := 0 - smallVal].
		 ^objectMemory integerObjectOf: smallVal].

	largeClass := isNegative
					ifTrue: [objectMemory classLargeNegativeInteger]
					ifFalse: [objectMemory classLargePositiveInteger].
	objectMemory wordSize = 8
		ifTrue: [sz := 8]
		ifFalse:
			[(highWord := magnitude >> 32) = 0
				ifTrue: [sz := 4] 
				ifFalse:
					[sz := 5.
					 (highWord := highWord >> 8) = 0 ifFalse:
						[sz := sz + 1.
						 (highWord := highWord >> 8) = 0 ifFalse:
							[sz := sz + 1.
							 (highWord := highWord >> 8) = 0 ifFalse: [sz := sz + 1]]]]].
	newLargeInteger := objectMemory instantiateClass: largeClass indexableSize:  sz.
	SPURVM
		ifTrue:
			["Memory is eight byte aligned in SPUR, so we are sure to have room for 64bits word whatever allocated sz"
			objectMemory storeLong64: 0 ofObject: newLargeInteger withValue: (objectMemory byteSwapped64IfBigEndian: magnitude)]
		ifFalse:
			[sz > 4
				ifTrue: [objectMemory storeLong64: 0 ofObject: newLargeInteger withValue: (objectMemory byteSwapped64IfBigEndian: magnitude)]
				ifFalse: [objectMemory storeLong32: 0 ofObject: newLargeInteger withValue: (objectMemory byteSwapped32IfBigEndian: magnitude)]].

	^newLargeInteger! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 8/20/2020 10:45'!
magnitude64BitValueOf: oop
	"Convert the given object into an integer value.
	The object may be either a positive SmallInteger or up to an eight-byte LargeInteger."
	<returnTypeC: #usqLong>
	| sz ok smallIntValue |

	(objectMemory isIntegerObject: oop) ifTrue:
		[smallIntValue := (objectMemory integerValueOf: oop).
		 smallIntValue < 0 ifTrue: [smallIntValue := 0 - smallIntValue].
		^smallIntValue].

	(objectMemory isNonIntegerImmediate: oop) ifTrue:
		[self primitiveFail.
		 ^0].

	ok := objectMemory
			isClassOfNonImm: oop
			equalTo: (objectMemory splObj: ClassLargePositiveInteger)
			compactClassIndex: ClassLargePositiveIntegerCompactIndex.
	ok ifFalse:
		[ok := objectMemory isClassOfNonImm: oop
						equalTo: (objectMemory splObj: ClassLargeNegativeInteger)
						compactClassIndex: ClassLargeNegativeIntegerCompactIndex.
		ok ifFalse:
			[self primitiveFail.
			 ^0]].
	sz := objectMemory numBytesOfBytes: oop.
	sz > (self sizeof: #sqLong) ifTrue:
		[self primitiveFail.
		 ^0].

	sz > 4 ifTrue:
		[^objectMemory byteSwapped64IfBigEndian: (objectMemory fetchLong64: 0 ofObject: oop)].
	^self cCoerceSimple: (objectMemory byteSwapped32IfBigEndian: (objectMemory fetchLong32: 0 ofObject: oop))
		to: #'unsigned int'! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 8/20/2020 10:32'!
maybeInlinePositive32BitValueOf: oop
	"Convert the given object into an integer value.
	The object may be either a positive SmallInteger or a four-byte LargePositiveInteger."
	<notOption: #Spur64BitMemoryManager>
	<returnTypeC: #'unsigned int'>
	| value ok sz |
	(objectMemory isIntegerObject: oop) ifTrue:
		[value := objectMemory integerValueOf: oop.
		 (value < 0) ifTrue:
			[self primitiveFail. value := 0].
		 ^value].

	(objectMemory isNonIntegerImmediate: oop) ifTrue:
		[self primitiveFail.
		 ^0].
	ok := objectMemory
				isClassOfNonImm: oop
				equalTo: (objectMemory splObj: ClassLargePositiveInteger)
				compactClassIndex: ClassLargePositiveIntegerCompactIndex.
	ok ifFalse:
		[self primitiveFail.
		 ^0].
	sz := objectMemory numBytesOfBytes: oop.
	sz > 4 ifTrue:
		[self primitiveFail.
		 ^0].
	^self cCoerceSimple: (objectMemory byteSwapped32IfBigEndian: (objectMemory fetchLong32: 0 ofObject: oop))
		to: #'unsigned int'! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 12/9/2016 10:03'!
noInlineSigned32BitValueGutsOf: oop
	"Convert the given object into an integer value.
	The object may be a four-byte LargeInteger."
	| value negative ok magnitude |
	<notOption: #Spur64BitMemoryManager>
	<inline: false>
	<returnTypeC: #int>
	<var: #value type: #int>
	<var: #magnitude type: #'unsigned int'>
	self deny: objectMemory hasSixtyFourBitImmediates.
	self deny: (objectMemory isIntegerObject: oop).

	(objectMemory isNonIntegerImmediate: oop) ifTrue:
		[self primitiveFail.
		 ^0].

	ok := objectMemory
			isClassOfNonImm: oop
			equalTo: (objectMemory splObj: ClassLargePositiveInteger)
			compactClassIndex: ClassLargePositiveIntegerCompactIndex.
	ok
		ifTrue: [negative := false]
		ifFalse:
			[negative := true.
			 ok := objectMemory isClassOfNonImm: oop
							equalTo: (objectMemory splObj: ClassLargeNegativeInteger)
							compactClassIndex: ClassLargeNegativeIntegerCompactIndex.
			 ok ifFalse:
				[self primitiveFail.
				 ^0]].
	(objectMemory numBytesOfBytes: oop) > 4 ifTrue:
		[^self primitiveFail].

	magnitude := self cCoerceSimple: (objectMemory byteSwapped32IfBigEndian: (objectMemory fetchLong32: 0 ofObject: oop)) to: #'unsigned int'.

	(negative
		ifTrue: [magnitude > 16r80000000]
		ifFalse: [magnitude >= 16r80000000])
			ifTrue:
				[self primitiveFail.
				^0].
	negative
		ifTrue: [value := 0 - magnitude]
		ifFalse: [value := magnitude].
	^value! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'nice 8/8/2016 12:09'!
positive32BitValueOf: oop
	"Convert the given object into an integer value.
	The object may be either a positive SmallInteger or a four-byte LargePositiveInteger."
	<returnTypeC: #'unsigned int'>
	objectMemory hasSixtyFourBitImmediates
		ifTrue:
			[(objectMemory isIntegerObject: oop) ifTrue:
				[| value64 |
				value64 := objectMemory integerValueOf: oop.
				 (value64 < 0
		 			 or: [self cCode: [(self cCoerceSimple: value64 to: #'unsigned int') ~= value64]
							inSmalltalk: [value64 >> 32 ~= 0]]) ifTrue:
						[self primitiveFail. value64 := 0].
				 ^value64].
			self primitiveFail.
			^0]
		ifFalse:
			[^self maybeInlinePositive32BitValueOf: oop]! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 8/20/2020 10:45'!
positive64BitValueOf: oop
	"Convert the given object into an integer value.
	The object may be either a positive SmallInteger or up to an eight-byte LargePositiveInteger."

	<returnTypeC: #usqLong>
	| sz ok smallIntValue |
	(objectMemory isIntegerObject: oop) ifTrue:
		[smallIntValue := objectMemory integerValueOf: oop.
		 smallIntValue < 0 ifTrue:
			[^self primitiveFail].
		 ^smallIntValue].

	(objectMemory isNonIntegerImmediate: oop) ifTrue:
		[self primitiveFail.
		 ^0].

	ok := objectMemory
			isClassOfNonImm: oop
			equalTo: (objectMemory splObj: ClassLargePositiveInteger)
			compactClassIndex: ClassLargePositiveIntegerCompactIndex.
	ok ifFalse:
		[self primitiveFail.
		 ^0].
	sz := objectMemory numBytesOfBytes: oop.
	sz > (self sizeof: #sqLong) ifTrue:
		[self primitiveFail.
		 ^0].

	sz > 4 ifTrue:
		[^objectMemory byteSwapped64IfBigEndian: (objectMemory fetchLong64: 0 ofObject: oop)].
	^self
		cCoerceSimple: (objectMemory byteSwapped32IfBigEndian: (objectMemory fetchLong32: 0 ofObject: oop))
		to: #'unsigned int'! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 8/20/2020 10:25'!
positiveMachineIntegerValueOf: oop
	"Answer a value of an integer in address range, i.e up to the size of a machine word.
	The object may be either a positive SmallInteger or a LargePositiveInteger of size <= word size."
	<returnTypeC: #'usqIntptr_t'>
	<inline: true> "only two callers & one is primitiveNewWithArg"
	| value bs ok |
	(objectMemory isIntegerObject: oop) ifTrue:
		[value := objectMemory integerValueOf: oop.
		 value < 0 ifTrue: [^self primitiveFail].
		^value].

	(objectMemory isNonIntegerImmediate: oop) ifTrue:
		[self primitiveFail.
		 ^0].

	ok := objectMemory
			isClassOfNonImm: oop
			equalTo: (objectMemory splObj: ClassLargePositiveInteger)
			compactClassIndex: ClassLargePositiveIntegerCompactIndex.
	ok ifFalse:
		[self primitiveFail.
		 ^0].
	bs := objectMemory numBytesOfBytes: oop.
	bs > (self sizeof: #'usqIntptr_t') ifTrue:
		[self primitiveFail.
		 ^0].

	((self sizeof: #'usqIntptr_t') = 8
	 and: [bs > 4]) ifTrue:
		[^objectMemory byteSwapped64IfBigEndian: (objectMemory fetchLong64: 0 ofObject: oop)].
	^self cCoerceSimple: (objectMemory byteSwapped32IfBigEndian: (objectMemory fetchLong32: 0 ofObject: oop)) to: #'unsigned int'! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'di 11/30/1998 10:13'!
primitiveAdd

	self pop2AndPushIntegerIfOK: (self stackIntegerValue: 1) + (self stackIntegerValue: 0)! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'dtl 9/2/2012 10:05'!
primitiveAddLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| a b result oopResult aIsNegative bIsNegative resultIsNegative oopArg oopRcvr |
	<export: true>
	<var: 'a' type: 'usqLong'>
	<var: 'b' type: 'usqLong'>
	<var: 'result' type: 'usqLong'>

	oopArg := self stackValue: 0.
	oopRcvr := self stackValue: 1.
	aIsNegative := self isNegativeIntegerValueOf: oopRcvr.
	bIsNegative := self isNegativeIntegerValueOf: oopArg.
	a := self magnitude64BitValueOf: oopRcvr.
	b := self magnitude64BitValueOf: oopArg.
	self successful ifFalse:[^nil].
	(aIsNegative = bIsNegative)
		ifTrue:
			["Protect against overflow"
			a > (16rFFFFFFFFFFFFFFFF - b) ifTrue: [self primitiveFail. ^nil].
			result := a + b.
			resultIsNegative := aIsNegative]
		ifFalse:
			[(a >= b)
				ifTrue:
					[result := a - b.
					resultIsNegative := aIsNegative]
				ifFalse:
					[result := b - a.
					resultIsNegative := bIsNegative]].
	oopResult := self magnitude64BitIntegerFor: result neg: resultIsNegative.
	self successful ifTrue:[self pop: 2 thenPush: oopResult].! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 10/11/2018 15:55'!
primitiveAdoptInstance
	"Primitive. Change the class of the argument to make it an instance of the receiver
	 given that the format of the receiver matches the format of the argument's class.
	 Fail if receiver or argument are SmallIntegers, or the receiver is an instance of a
	 compact class and the argument isn't, or when the argument's class is compact and
	 the receiver isn't, or when the format of the receiver is different from the format of
	 the argument's class, or when the arguments class is fixed and the receiver's size
	 differs from the size that an instance of the argument's class should have."
	| rcvr arg err |

	arg := self stackTop.
	rcvr := self stackValue: 1.
	((objectMemory isImmediate: arg)
	 or: [argumentCount > 1
		and: [(objectMemory isImmediate: rcvr)
			or: [(self objCouldBeClassObj: rcvr) not]]]) ifTrue:
		[^self primitiveFailFor: PrimErrBadArgument].

	err := objectMemory changeClassOf: arg to: rcvr.
	err = 0
		ifTrue:
			["Flush at cache because rcvr's class has changed."
			 self flushAtCache.
			 self pop: self methodArgumentCount]
		ifFalse:
			["changeClassOf:to: answers errors as if rcvr (the class) is an argument..."
			 err = PrimErrBadReceiver
				ifTrue:
					[err := PrimErrBadArgument]
				ifFalse:
					[err = PrimErrBadArgument ifTrue:
						[err := PrimErrBadReceiver]].
			 self primitiveFailFor: err].
	^nil! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 7/18/2015 08:06'!
primitiveAllInstances
	"Answer an array of all instances of the receiver that exist
	 when the primitive is called, excluding any that may be
	 garbage collected as a side effect of allocating the result array."

	<export: true>
	| result |
	NewspeakVM ifTrue: "For the mirror prims check that the class obj is actually a valid class."
		[(argumentCount < 1
		  or: [(objectMemory isNonImmediate: self stackTop)
			  and: [self objCouldBeClassObj: self stackTop]]) ifFalse:
			[^self primitiveFailFor: PrimErrBadArgument]].
	result := objectMemory allInstancesOf: self stackTop.
	(objectMemory isIntegerObject: result) ifTrue:
		[objectMemory growToAccomodateContainerWithNumSlots: (objectMemory integerValueOf: result).
		 result := objectMemory allInstancesOf: self stackTop.
		 (objectMemory isIntegerObject: result) ifTrue:
			[^self primitiveFailFor: PrimErrNoMemory]].
	self pop: argumentCount+1 thenPush: result! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 3/11/2014 12:57'!
primitiveAllObjects
	"Answer an array of all objects that exist when the primitive
	 is called, excluding those that may be garbage collected as
	 a side effect of allocating the result array."

	<export: true>
	| result |
	result := objectMemory allObjects.
	(objectMemory isIntegerObject: result) ifTrue:
		[objectMemory growToAccomodateContainerWithNumSlots: (objectMemory integerValueOf: result).
		 result := objectMemory allObjects.
		 (objectMemory isIntegerObject: result) ifTrue:
			[^self primitiveFailFor: PrimErrNoMemory]].
	self pop: argumentCount+1 thenPush: result! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2017 14:47'!
primitiveArctan
	"N.B. IMO we should be able to assume the receiver is a float because this primitive is specific to floats. eem 2/13/2017"
	| rcvr |
	<var: #rcvr type: #double>
	rcvr := self stackFloatValue: 0.
	self successful ifTrue:
		[self stackTopPut: (objectMemory floatObjectOf:
								(self cCode: [rcvr atan]
									inSmalltalk: [rcvr = rcvr
													ifTrue: [rcvr arcTan]
													ifFalse: [Float nan]]))]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 3/18/2014 18:52'!
primitiveArrayBecome
	"We must flush the method cache here, to eliminate stale references
	to mutated classes and/or selectors."

	| arg rcvr ec |
	arg := self stackTop.
	rcvr := self stackValue: 1.
	ec := objectMemory become: rcvr with: arg twoWay: true copyHash: false.
	ec = PrimNoErr
		ifTrue: [self pop: 1]
		ifFalse: [self primitiveFailFor: ec]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
primitiveArrayBecomeOneWay
	"We must flush the method cache here, to eliminate stale references
	to mutated classes and/or selectors."

	| arg rcvr ec |
	arg := self stackTop.
	rcvr := self stackValue: 1.
	ec := objectMemory become: rcvr with: arg twoWay: false copyHash: true.
	ec = PrimNoErr
		ifTrue: [self pop: 1]
		ifFalse: [self primitiveFailFor: ec]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 4/15/2020 16:22'!
primitiveArrayBecomeOneWayCopyHashArg
	"Similar to primitiveArrayBecomeOneWay but accepts a third argument deciding whether to
	 copy the receiver's elements identity hashes over the argument's elements identity hashes."

	| copyHashFlag ec |
	self stackTop = objectMemory trueObject
		ifTrue: [copyHashFlag := true]
		ifFalse:
			[self stackTop = objectMemory falseObject
				ifTrue: [copyHashFlag := false]
				ifFalse:
					[self primitiveFailFor: PrimErrBadArgument.
					 ^nil]].
	ec := objectMemory
			become: (self stackValue: 2)
			with: (self stackValue: 1)
			twoWay: false
			copyHash: copyHashFlag.
	ec = PrimNoErr
		ifTrue: [self pop: argumentCount]
		ifFalse: [self primitiveFailFor: ec]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 7/19/2017 13:55'!
primitiveArrayBecomeOneWayNoCopyHash
	"Similar to primitiveArrayBecomeOneWay but does /not/ copy the receiver's
	 elements identity hashes over the argument's elements identity hashes."

	| arg rcvr ec |
	arg := self stackTop.
	rcvr := self stackValue: 1.
	ec := objectMemory become: rcvr with: arg twoWay: false copyHash: false.
	ec = PrimNoErr
		ifTrue: [self pop: 1]
		ifFalse: [self primitiveFailFor: ec]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 9/5/2013 16:35'!
primitiveAsCharacter
	| characterCode characterObject |
	characterCode := self stackTop.
	((objectMemory isIntegerObject: characterCode)
	and: [characterCode := objectMemory integerValueOf: characterCode.
		objectMemory isInRangeCharacterCode: characterCode]) ifFalse:
		[^self primitiveFailFor: (argumentCount = 0
									ifTrue: [PrimErrBadReceiver]
									ifFalse: [PrimErrBadArgument])].
	characterObject := self characterForAscii: characterCode.
	self pop: argumentCount + 1 thenPush: characterObject! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 3/7/2015 17:47'!
primitiveAsFloat
	"N.B. This will answer inexact results for integers with > 53 bits of magnitude."
	| rcvr |
	rcvr := self stackTop.
	self assert: (objectMemory isIntegerObject: rcvr).
	self pop: 1 thenPushFloat: (objectMemory integerValueOf: rcvr) asFloat! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 1/15/2015 12:49'!
primitiveAt
	<accessorDepth: 0>
	self commonAt: false! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 1/15/2015 12:48'!
primitiveAtPut
	<accessorDepth: 0>
	self commonAtPut: false! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'nice 11/18/2016 04:13'!
primitiveBeCursor
	"Set the cursor to the given shape. The Mac only supports 16x16 pixel cursors. Cursor offsets are handled by Smalltalk."

	| cursorObj maskBitsIndex maskObj bitsObj extentX extentY depth offsetObj offsetX offsetY cursorBitsIndex ourCursor |

	argumentCount = 0 ifTrue: [
		cursorObj := self stackTop.
		maskBitsIndex := nil].
	argumentCount = 1 ifTrue: [
		cursorObj := self stackValue: 1.
		maskObj := self stackTop].
	self success: argumentCount < 2.

	self successful ifTrue: [
		self success: ((objectMemory isPointers: cursorObj) and: [(objectMemory lengthOf: cursorObj) >= 5])].
	self successful ifTrue: [
		bitsObj := objectMemory fetchPointer: 0 ofObject: cursorObj.
		extentX := self fetchInteger: 1 ofObject: cursorObj.
		extentY := self fetchInteger: 2 ofObject: cursorObj.
		depth := self fetchInteger: 3 ofObject: cursorObj.
		offsetObj := objectMemory fetchPointer: 4 ofObject: cursorObj.
		self success: ((objectMemory isPointers: offsetObj) and: [(objectMemory lengthOf: offsetObj) >= 2])].

	self successful ifTrue: [
		offsetX := self fetchInteger: 0 ofObject: offsetObj.
		offsetY := self fetchInteger: 1 ofObject: offsetObj.
		(argumentCount = 0 and: [depth = 32])
			ifTrue: [
				"Support arbitrary-sized 32 bit ARGB forms --bf 3/1/2007 23:51"
				self success: ((extentX > 0) and: [extentY > 0]).
				self success: ((offsetX >= (extentX * -1)) and: [offsetX <= 0]).
				self success: ((offsetY >= (extentY * -1)) and: [offsetY <= 0]).
				self success: ((objectMemory isWords: bitsObj) and: [(objectMemory lengthOf: bitsObj) = (extentX * extentY)]).
				cursorBitsIndex := bitsObj + objectMemory baseHeaderSize.
				self cCode: '' inSmalltalk:
					[ourCursor := Cursor
						extent: extentX @ extentY
						depth: 32
						fromArray: ((1 to: extentX * extentY) collect: [:i |
							objectMemory fetchLong32: i-1 ofObject: bitsObj])
						offset: offsetX  @ offsetY]]
			ifFalse: [
				self success: ((extentX = 16) and: [extentY = 16 and: [depth = 1]]).
				self success: ((offsetX >= -16) and: [offsetX <= 0]).
				self success: ((offsetY >= -16) and: [offsetY <= 0]).
				self success: ((objectMemory isWords: bitsObj) and: [(objectMemory lengthOf: bitsObj) = 16]).
				cursorBitsIndex := bitsObj + objectMemory baseHeaderSize.
				self cCode: '' inSmalltalk:
					[ourCursor := Cursor
						extent: extentX @ extentY
						fromArray: ((1 to: 16) collect: [:i |
							((objectMemory fetchLong32: i-1 ofObject: bitsObj) >> (objectMemory wordSize*8 - 16)) bitAnd: 16rFFFF])
						offset: offsetX  @ offsetY]]].

	argumentCount = 1 ifTrue: [
		self success: ((objectMemory isPointers: maskObj) and: [(objectMemory lengthOf: maskObj) >= 5]).
		self successful ifTrue: [
			bitsObj := objectMemory fetchPointer: 0 ofObject: maskObj.
			extentX := self fetchInteger: 1 ofObject: maskObj.
			extentY := self fetchInteger: 2 ofObject: maskObj.
			depth := self fetchInteger: 3 ofObject: maskObj].

		self successful ifTrue: [
			self success: ((extentX = 16) and: [extentY = 16 and: [depth = 1]]).
			self success: ((objectMemory isWords: bitsObj) and: [(objectMemory lengthOf: bitsObj) = 16]).
			maskBitsIndex := bitsObj + objectMemory baseHeaderSize]].

	self successful ifTrue: [
		argumentCount = 0
			ifTrue: [
				depth = 32
					ifTrue: [(self cCode: 'ioSetCursorARGB(cursorBitsIndex, extentX, extentY, offsetX, offsetY)'
						inSmalltalk: [ourCursor show. Cursor currentCursor == ourCursor])	
							ifFalse: [^self success: false]]
					ifFalse: [self cCode: 'ioSetCursor(cursorBitsIndex, offsetX, offsetY)'
						inSmalltalk: [ourCursor show]]]
			ifFalse: [self cCode: 'ioSetCursorWithMask(cursorBitsIndex, maskBitsIndex, offsetX, offsetY)'
						inSmalltalk: [cursorBitsIndex == maskBitsIndex. "placate compiler"
									ourCursor show]].
		self pop: argumentCount]! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 5/30/2018 18:37'!
primitiveBeDisplay
	"Record the system Display object in the specialObjectsTable,
	 and if possible pin the display bitmap.  Further, invoke ioBeDisplay
	 to alow the VM to record the location, width heigth & depth of the bitmap."
	| rcvr bitsOop depthOop heightOop widthOop |
	rcvr := self stackTop.
	((objectMemory isPointers: rcvr)
	and: [(objectMemory lengthOf: rcvr) >= 4
	and: [bitsOop := objectMemory fetchPointer: 0 ofObject: rcvr.
		((objectMemory isWordsOrBytes: bitsOop)
		or: [objectMemory isIntegerObject: bitsOop]) "for surface plugin handles"
	and: [(objectMemory isIntegerObject: (widthOop := objectMemory fetchPointer: 1 ofObject: rcvr))
	and: [(objectMemory isIntegerObject: (heightOop := objectMemory fetchPointer: 2 ofObject: rcvr))
	and: [(objectMemory isIntegerObject: (depthOop := objectMemory fetchPointer: 3 ofObject: rcvr))]]]]]) ifFalse:
		[^self primitiveFailFor: PrimErrBadReceiver].
	objectMemory splObj: TheDisplay put: rcvr.
	(objectMemory hasSpurMemoryManagerAPI
	 and: [(objectMemory isNonImmediate: bitsOop)
	 and: [(objectMemory isPinned: bitsOop) not]]) ifTrue:
		[rcvr := objectMemory pinObject: bitsOop. "Answers 0 if memory required to pin bit not enough memory available."
		 rcvr ~= 0 ifTrue: [bitsOop := rcvr]].
	self ioBeDisplay: ((objectMemory isNonImmediate: bitsOop)
						ifTrue: [objectMemory firstIndexableField: bitsOop]
						ifFalse: [bitsOop asVoidPointer])
		width: (objectMemory integerValueOf: widthOop)
		height: (objectMemory integerValueOf: heightOop)
		depth: (objectMemory integerValueOf: depthOop)! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'tpr 3/15/2004 12:13'!
primitiveBeep
"make the basic beep noise"
	self ioBeep.! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 7/18/2015 08:02'!
primitiveBehaviorHash
	| hashOrError |
	NewspeakVM ifTrue: "For the mirror prims check that the class obj is actually a valid class."
		[argumentCount > 0 ifTrue:
			[((objectMemory isNonImmediate: self stackTop)
			  and: [self objCouldBeClassObj: self stackTop]) ifFalse:
				[^self primitiveFailFor: PrimErrBadArgument]]].
	self assert: ((objectMemory isNonImmediate: self stackTop)
				  and: [self addressCouldBeClassObj: self stackTop]).
	hashOrError := objectMemory ensureBehaviorHash: self stackTop.
	hashOrError >= 0
		ifTrue: [self pop: argumentCount + 1 thenPushInteger: hashOrError]
		ifFalse: [self primitiveFailFor: hashOrError negated]! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 3/23/2016 12:54'!
primitiveBitAnd
	<inline: false>
	<var: 'integerArgumentValue' type: #usqInt>
	<var: 'integerReceiverValue' type: #usqInt>
	| integerArgumentValue integerReceiverValue |
	"Note no short-cut for SmallIntegers.  Either the inline interpreter bytecode or the JIT primitive will handle this case."
	integerArgumentValue := self positiveMachineIntegerValueOf: self stackTop.
	integerReceiverValue := self positiveMachineIntegerValueOf: (self stackValue: 1).
	self successful ifTrue:
		[self pop: 2 thenPush: (self positiveMachineIntegerFor: (integerArgumentValue bitAnd: integerReceiverValue))]! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'nice 1/16/2013 22:41'!
primitiveBitAndLargeIntegers
	"Primitive logical operations for large integers in 64 bit range"
	| integerRcvr integerArg oopResult |
	<export: true>
	<var: 'integerRcvr' type: 'usqLong'>
	<var: 'integerArg' type: 'usqLong'>

	integerArg := self positive64BitValueOf: (self stackValue: 0).
	integerRcvr := self positive64BitValueOf: (self stackValue: 1).
	self successful ifFalse:[^nil].

	oopResult := self positive64BitIntegerFor: (integerRcvr bitAnd: integerArg).
	self successful ifTrue:[self pop: 2 thenPush: oopResult]! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 3/23/2016 12:54'!
primitiveBitOr
	<inline: false>
	<var: 'integerArgumentValue' type: #usqInt>
	<var: 'integerReceiverValue' type: #usqInt>
	| integerArgumentValue integerReceiverValue |
	"Note no short-cut for SmallIntegers.  Either the inline interpreter bytecode or the JIT primitive will handle this case."
	integerArgumentValue := self positiveMachineIntegerValueOf: self stackTop.
	integerReceiverValue := self positiveMachineIntegerValueOf: (self stackValue: 1).
	self successful ifTrue:
		[self pop: 2 thenPush: (self positiveMachineIntegerFor: (integerArgumentValue bitOr: integerReceiverValue))]! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'nice 1/16/2013 22:41'!
primitiveBitOrLargeIntegers
	"Primitive logical operations for large integers in 64 bit range"
	| integerRcvr integerArg oopResult |
	<export: true>
	<var: 'integerRcvr' type: 'usqLong'>
	<var: 'integerArg' type: 'usqLong'>

	integerArg := self positive64BitValueOf: (self stackValue: 0).
	integerRcvr := self positive64BitValueOf: (self stackValue: 1).
	self successful ifFalse:[^nil].

	oopResult := self positive64BitIntegerFor: (integerRcvr bitOr: integerArg).
	self successful ifTrue:[self pop: 2 thenPush: oopResult]! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 12/9/2015 19:32'!
primitiveBitShift
	"Perform a bitShift.  In 32-bits deal only with non-negative 32-bit integers.
	  In 64-bits deal with signed 64-bit quantities (max (2^63)-1)."
	| integerReceiver integerArgument shifted |
	<var: #integerReceiver type: #sqInt>
	integerArgument := self stackTop.
	(objectMemory isIntegerObject: integerArgument) ifFalse:
		[^self primitiveFail].
	integerReceiver := self stackValue: 1.
	objectMemory wordSize = 4
		ifTrue:
			[integerReceiver := self positive32BitValueOf: integerReceiver]
		ifFalse:
			[integerReceiver := self signed64BitValueOf: integerReceiver].
	self successful ifTrue:
		[(integerArgument := objectMemory integerValueOf: integerArgument) >= 0
			ifTrue: "Left shift -- must fail bits would be lost"
				[integerArgument <= objectMemory numSmallIntegerBits ifFalse:
					[^self primitiveFail].
				shifted := integerReceiver << integerArgument.
				self cCode: '' inSmalltalk: [shifted := objectMemory wordSize = 4
									ifTrue: [shifted signedIntFromLong]
									ifFalse: [shifted signedIntFromLong64]].
				integerReceiver = (objectMemory wordSize = 4
									ifTrue: [shifted >> integerArgument]
									ifFalse: [shifted >>> integerArgument]) ifFalse:
					[^self primitiveFail]]
			ifFalse: "Right shift -- OK to lose bits"
				[integerArgument >= objectMemory numSmallIntegerBits negated ifFalse:
					[^self primitiveFail].
				 shifted := objectMemory wordSize = 4
								ifTrue: [integerReceiver >> (0 - integerArgument)]
								ifFalse: [integerReceiver >>> (0 - integerArgument)]].
		shifted := objectMemory wordSize = 4
					ifTrue: [self positive32BitIntegerFor: shifted]
					ifFalse:
						[(objectMemory isIntegerValue: shifted)
							ifTrue: [objectMemory integerObjectOf: shifted]
							ifFalse: [self signed64BitIntegerFor: shifted]].
		self pop: 2 thenPush: shifted]! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'dtl 9/2/2012 10:06'!
primitiveBitShiftLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| a shift result oopResult aIsNegative oopRcvr |
	<export: true>
	<var: 'a' type: 'usqLong'>
	<var: 'result' type: 'usqLong'>

	shift := self stackIntegerValue: 0.
	oopRcvr := self stackValue: 1.
	aIsNegative := self isNegativeIntegerValueOf: oopRcvr.
	a := self magnitude64BitValueOf: oopRcvr.
	self successful ifFalse:[^nil].
	(shift >= 0)
		ifTrue:
			["Protect against overflow"
			result := 16rFFFFFFFFFFFFFFFF. "This is to avoid undue (usqInt) cast"
			(shift >= 64 or: [a > (result >> shift)]) ifTrue: [self primitiveFail. ^nil].
			result := a << shift]
		ifFalse:
			[shift := 0 - shift.
			shift >= 64
				ifTrue: [result := 0]
				ifFalse: [result := a >> shift].
			"Fake 2 complement for negative values"
			(aIsNegative and: [result << shift ~= a]) ifTrue: [result := result + 1]].
	oopResult := self magnitude64BitIntegerFor: result neg: aIsNegative.
	self successful ifTrue:[self pop: 2 thenPush: oopResult].! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 3/23/2016 12:55'!
primitiveBitXor
	<inline: false>
	<var: 'integerArgumentValue' type: #usqInt>
	<var: 'integerReceiverValue' type: #usqInt>
	| integerReceiver integerArgument integerArgumentValue integerReceiverValue |
	integerArgument := self stackTop.
	integerReceiver := self stackValue: 1.
	((objectMemory isIntegerObject: integerArgument)
	 and: [objectMemory isIntegerObject: integerReceiver])
		ifTrue: "xoring will leave the tag bits zero, whether the tag is 1 or zero, so add it back in."
			[self pop: 2 thenPush: (integerArgument bitXor: integerReceiver) + objectMemory smallIntegerTag]
		ifFalse:
			[integerArgumentValue := self positiveMachineIntegerValueOf: integerArgument.
			integerReceiverValue := self positiveMachineIntegerValueOf: integerReceiver.
			 self successful ifTrue:
				[self pop: 2 thenPush: (self positiveMachineIntegerFor: (integerArgumentValue bitXor: integerReceiverValue))]]! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'nice 1/16/2013 22:41'!
primitiveBitXorLargeIntegers
	"Primitive logical operations for large integers in 64 bit range"
	| integerRcvr integerArg oopResult |
	<export: true>
	<var: 'integerRcvr' type: 'usqLong'>
	<var: 'integerArg' type: 'usqLong'>

	integerArg := self positive64BitValueOf: (self stackValue: 0).
	integerRcvr := self positive64BitValueOf: (self stackValue: 1).
	self successful ifFalse:[^nil].

	oopResult := self positive64BitIntegerFor: (integerRcvr bitXor: integerArg).
	self successful ifTrue:[self pop: 2 thenPush: oopResult]! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 6/27/2017 18:21'!
primitiveBytesLeft
	"Answer bytes available at this moment. For more meaningful 
	 results, calls to this primitive should be precedeed by a full 
	 or incremental garbage collection."
	| aBool |
	self methodArgumentCount = 0 ifTrue: "old behavior - just return the size of free memory"
		[^self pop: 1 thenPushInteger: objectMemory freeSize].
	self methodArgumentCount = 1 ifTrue:
		["Spur behavior; if argument is nil answer size of largest free chunk in oldSpace."
		(objectMemory hasSpurMemoryManagerAPI
		 and: [self stackTop = objectMemory nilObject]) ifTrue:
			[^self pop: 2 thenPushInteger: objectMemory sizeOfLargestFreeChunk].
		"new behaviour -including or excluding swap space depending on aBool"
		aBool := self booleanValueOf: self stackTop.
		self successful ifTrue:
			[^self pop: 2 thenPushInteger: (objectMemory bytesLeft: aBool)]].
	^self primitiveFail! !

!InterpreterPrimitives methodsFor: 'plugin primitives' stamp: 'eem 7/27/2021 13:16'!
primitiveCalloutToFFI
	"Perform a function call to a foreign function.
	Only invoked from method containing explicit external call spec.
	Due to this we use the pluggable prim mechanism explicitly here
	(the first literal of any FFI spec'ed method is an ExternalFunction
	and not an array as used in the pluggable primitive mechanism)."

	<accessorDepth: 2> "Manually copied from primitiveCalloutAccessorDepth in the ThreadedFFIPlugins..."
	<var: #primitiveCallout declareC: 'void (*primitiveCallout)(void)'>
	self functionForPrimitiveCallout
		ifNil: [self primitiveFail]
		ifNotNil: [:primitiveCallout| self dispatchFunctionPointer: primitiveCallout]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 9/4/2013 16:58'!
primitiveChangeClass
	"Primitive.  Change the class of the receiver into the class of the argument given that
	 the format of the receiver matches the format of the argument's class.  Fail if the
	 receiver or argument are SmallIntegers, or the receiver is an instance of a compact
	 class and the argument isn't, or when the argument's class is compact and the receiver
	 isn't, or when the format of the receiver is different from the format of the argument's
	 class, or when the arguments class is fixed and the receiver's size differs from the size
	 that an instance of the argument's class should have."
	| arg rcvr argClass err |
	arg := self stackObjectValue: 0.
	rcvr := self stackObjectValue: 1.
	self successful ifFalse:[^nil].
	argClass := objectMemory fetchClassOfNonImm: arg.
	err := objectMemory changeClassOf: rcvr to: argClass.
	err = 0
		ifTrue: ["Flush at cache because rcvr's class has changed."
				self flushAtCache.
				self pop: self methodArgumentCount]
		ifFalse: [self primitiveFailFor: err].
	^nil! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 6/20/2015 10:34'!
primitiveClass
	| instance |
	instance := self stackTop.
	(argumentCount > 0
	 and: [objectMemory isOopForwarded: instance])
		ifTrue:
			[self primitiveFail]
		ifFalse:
			[self pop: argumentCount + 1 thenPush: (objectMemory fetchClassOf: instance)]! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 8/21/2008 11:02'!
primitiveClearVMProfile
	"Primitive. Void the VM profile histograms."
	self cCode: 'ioClearProfile()'.
	self pop: argumentCount! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 3/6/2015 13:45'!
primitiveClipboardText
	"When called with a single string argument, post the string to 
	the clipboard. When called with zero arguments, return a 
	string containing the current clipboard contents."
	| s sz |
	argumentCount = 1
		ifTrue:
			[s := self stackTop.
			 (objectMemory isBytes: s) ifFalse: [^ self primitiveFail].
			 self successful ifTrue:
				[sz := objectMemory numBytesOfBytes: s.
				 self clipboardWrite: sz From: s + objectMemory baseHeaderSize At: 0.
				 self pop: 1]]
		ifFalse:
			[sz := self clipboardSize.
			 objectMemory hasSpurMemoryManagerAPI
				ifTrue:
					[s := objectMemory allocateBytes: sz classIndex: ClassByteStringCompactIndex.
					 s ifNil: [^self primitiveFail]]
				ifFalse:
					[(objectMemory sufficientSpaceToAllocate: sz) ifFalse: [^self primitiveFail].
					 s := objectMemory instantiateClass: (objectMemory splObj: ClassByteString) indexableSize: sz].
			 self clipboardRead: sz Into: s + objectMemory baseHeaderSize At: 0.
			 self pop: 1 thenPush: s]! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'nice 7/29/2016 01:28'!
primitiveClockLogAddresses
	"Take a boolean which if true turns or keeps clock logging on.  Answer an array supplying
	 the size of the clock logs, the address of the usecs log, the index in it, the address of the
	 msecs log, and the index into it."
	<export: true>
	| result runInNOut usecs uidx msecs midx v1 v2 |
	<var: #usecs type: #'void *'>
	<var: #msecs type: #'void *'>
	<var: #runInNOut type: #sqInt> "bypass type inference which would deduce int"
	argumentCount ~= 1 ifTrue:
		[^self primitiveFail].
	runInNOut := (self stackValue: 0) == objectMemory trueObject.
	self ioGetClockLogSize: (self addressOf: runInNOut)
		Usecs: (self addressOf: usecs) Idx: (self addressOf: uidx)
		Msecs: (self addressOf: msecs) Idx: (self addressOf: midx).
	result := objectMemory instantiateClass: (objectMemory splObj: ClassArray) indexableSize: 5.
	result = 0 ifTrue:
		[^self primitiveFail].
	objectMemory pushRemappableOop: result.
	objectMemory storePointerUnchecked: 0 ofObject: objectMemory topRemappableOop withValue: (objectMemory integerObjectOf: runInNOut).
	v1 := self positive32BitIntegerFor: usecs asUnsignedInteger.
	v2 := self positive32BitIntegerFor: msecs asUnsignedInteger.
	self successful ifFalse:
		[objectMemory popRemappableOop.
		 ^self primitiveFail].
	objectMemory storePointer: 1 ofObject: objectMemory topRemappableOop withValue: v1.
	objectMemory storePointerUnchecked: 2 ofObject: objectMemory topRemappableOop withValue: (objectMemory integerObjectOf: uidx).
	objectMemory storePointer: 3 ofObject: objectMemory topRemappableOop withValue: v2.
	objectMemory storePointerUnchecked: 4 ofObject: objectMemory topRemappableOop withValue: (objectMemory integerObjectOf: midx).
	self pop: 2 thenPush: objectMemory popRemappableOop
! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 01/07/2021 10:40'!
primitiveClone
	"Return a shallow copy of the receiver."

	| rcvr newCopy |
	rcvr := self stackTop.
	(objectMemory isImmediate: rcvr)
		ifTrue:
			[newCopy := rcvr]
		ifFalse:
			[(argumentCount = 0
			  or: [(objectMemory isForwarded: rcvr) not])
				ifTrue: [newCopy := objectMemory cloneObject: rcvr]
				ifFalse: [newCopy := 0].
			 newCopy = 0 ifTrue: "not enough memory most likely"
				[^self primitiveFail]].
	self pop: argumentCount + 1 thenPush: newCopy! !

!InterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 4/23/2017 11:35'!
primitiveClosureValue
	| blockClosure numArgs closureMethod outerContext |
	blockClosure := self stackValue: argumentCount.
	numArgs := self argumentCountOfClosure: blockClosure.
	argumentCount = numArgs ifFalse:
		[^self primitiveFail].

	"Somewhat paranoiac checks we need while debugging that we may be able to discard
	 in a robust system."
	outerContext := objectMemory fetchPointer: ClosureOuterContextIndex ofObject: blockClosure.
	(objectMemory isContext: outerContext) ifFalse:
		[^self primitiveFail].

	closureMethod := objectMemory fetchPointer: MethodIndex ofObject: outerContext.
	"Check if the closure's method is actually a CompiledMethod."
	(objectMemory isOopCompiledMethod: closureMethod) ifFalse:
		[^self primitiveFail].

	self activateNewClosure: blockClosure outer: outerContext method: closureMethod numArgs: numArgs mayContextSwitch: true! !

!InterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 4/23/2017 11:35'!
primitiveClosureValueNoContextSwitch
	"An exact clone of primitiveClosureValue except that this version will not
	 check for interrupts on stack overflow.  It may invoke the garbage collector
	 but will not switch processes.  See checkForInterruptsMayContextSwitch:"
	<api>
	| blockClosure numArgs closureMethod outerContext |
	blockClosure := self stackValue: argumentCount.
	numArgs := self argumentCountOfClosure: blockClosure.
	argumentCount = numArgs ifFalse:
		[^self primitiveFail].

	"Somewhat paranoiac checks we need while debugging that we may be able to discard
	 in a robust system."
	outerContext := objectMemory fetchPointer: ClosureOuterContextIndex ofObject: blockClosure.
	(objectMemory isContext: outerContext) ifFalse:
		[^self primitiveFail].

	closureMethod := objectMemory fetchPointer: MethodIndex ofObject: outerContext.
	"Check if the closure's method is actually a CompiledMethod."
	(objectMemory isOopCompiledMethod: closureMethod) ifFalse:
		[^self primitiveFail].

	self activateNewClosure: blockClosure outer: outerContext method: closureMethod numArgs: numArgs mayContextSwitch: false! !

!InterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 4/23/2017 11:35'!
primitiveClosureValueWithArgs
	| argumentArray arraySize blockClosure numArgs closureMethod index outerContext |
	argumentArray := self stackTop.
	(objectMemory isArray: argumentArray) ifFalse:
		[^self primitiveFail].

	"Check for enough space in thisContext to push all args"
	arraySize := objectMemory numSlotsOf: argumentArray.
	(self roomToPushNArgs: arraySize) ifFalse:
		[^self primitiveFail].

	blockClosure := self stackValue: argumentCount.
	numArgs := self argumentCountOfClosure: blockClosure.
	arraySize = numArgs ifFalse:
		[^self primitiveFail].

	"Somewhat paranoiac checks we need while debugging that we may be able to discard
	 in a robust system."
	outerContext := objectMemory fetchPointer: ClosureOuterContextIndex ofObject: blockClosure.
	(objectMemory isContext: outerContext) ifFalse:
		[^self primitiveFail].
	closureMethod := objectMemory fetchPointer: MethodIndex ofObject: outerContext.
	"Check if the closure's method is actually a CompiledMethod."
	(objectMemory isOopCompiledMethod: closureMethod) ifFalse:
		[^self primitiveFail].

	self popStack.

	"Copy the arguments to the stack, and activate"
	index := 1.
	[index <= numArgs] whileTrue:
		[self push: (objectMemory fetchPointer: index - 1 ofObject: argumentArray).
		index := index + 1].

	self activateNewClosure: blockClosure outer: outerContext method: closureMethod numArgs: numArgs mayContextSwitch: true! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 5/26/2017 15:09'!
primitiveCoarseLocalMicrosecondClock
	"Return the value of the microsecond clock in the local timezone, as updated by the heartbeat, as an integer.
	 This is the number of microseconds since the Smalltalk epoch, 1901/1/1 12:00am.
	 The microsecond clock is at least 60 bits wide which means it'll get to around August
	 38435 before it wraps around.  Be sure to put it on your calendar.  The coarse clock is
	 updated by the heartbeat thread and as such is much cheaper than
	 primitiveUTCMicrosecondClock, which always entails a system call."

	self pop: 1 thenPush: (self positive64BitIntegerFor: self ioLocalMicroseconds)! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 5/26/2017 15:11'!
primitiveCoarseUTCMicrosecondClock
	"Return the value of the microsecond clock as updated by the heartbeat as an integer.
	 This is the number of microseconds since the Smalltalk epoch, 1901/1/1 12:00am.
	 The microsecond clock is at least 60 bits wide which means it'll get to around August
	 38435 before it wraps around.  Be sure to put it on your calendar.  The coarse clock is
	 updated by the heartbeat thread and as such is much cheaper than
	 primitiveUTCMicrosecondClock, which always entails a system call."

	self pop: 1 thenPush: (self positive64BitIntegerFor: self ioUTCMicroseconds)! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 8/13/2021 11:56'!
primitiveCompareBytes
	"Primitive. Compare two byte-indexed objects for equality"
	<export: true flags: #FastCPrimitiveFlag>
	| arg1 arg2 len |
	arg1 := self stackValue: 1.
	arg2 := self stackValue: 0.
	"Quick identity test"
	arg1 = arg2 ifTrue:
		[^self methodReturnBool: true].
	((objectMemory isBytes: arg1) and:[objectMemory isBytes: arg2]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	len := objectMemory numBytesOfBytes: arg1.
	len = (objectMemory numBytesOfBytes: arg2) ifFalse:
		[^self methodReturnBool: false].
	0 to: len - 1 do:
		[:i|
		(objectMemory fetchByte: i ofObject: arg1) = (objectMemory fetchByte: i ofObject: arg2) ifFalse:
			[^self methodReturnBool: false]].
	^self methodReturnBool: true! !

!InterpreterPrimitives methodsFor: 'string primitives' stamp: 'eem 4/20/2018 09:37'!
primitiveCompareWith
	"<string1> primitiveCompareWith: string2 [collated: order] "
	<export: true>
	
	| string1 string2 order strLength1 strLength2 result |

	"1 - fetch the parameters from the stack"	
	argumentCount = 1 ifFalse:
		[argumentCount ~= 2 ifTrue:
			[^self primitiveFailFor: PrimErrBadNumArgs].
			 order := self stackTop.
			 ((objectMemory isBytes: order)
			 and: [(objectMemory numBytesOfBytes: order) = 256]) ifFalse:
				[^self primitiveFailFor: PrimErrBadArgument]].
	string1 := self stackValue: argumentCount.
	string2 := self stackValue: argumentCount - 1. 
			
	"2 - check their types - all parameters are ByteObject"
	((objectMemory isBytes: string1)
	 and: [objectMemory isBytes: string2]) ifFalse: 
		[^self primitiveFailFor: PrimErrBadArgument].
	
	"3 - compare the strings"	
	strLength1 := objectMemory numBytesOfBytes: string1.
	strLength2 := objectMemory numBytesOfBytes: string2.
	result := order 
		ifNil: [self rawCompare: string1 length: strLength1 with: string2 length: strLength2 accessBlock:
				[:str :index | objectMemory fetchByte: index ofObject: str ]]
		ifNotNil: 
			[self rawCompare: string1 length: strLength1 with: string2 length: strLength2 accessBlock:
				[:str :index | objectMemory fetchByte: (objectMemory fetchByte: index ofObject: str) + 1 ofObject: order ]].

	self methodReturnInteger: result

	
	
	! !

!InterpreterPrimitives methodsFor: 'sound primitives' stamp: 'eem 10/12/2016 17:24'!
primitiveConstantFill
	"Fill the receiver, which must be an indexable non-pointer
	 object, with the given integer value."
	objectMemory hasSpurMemoryManagerAPI
		ifTrue: [self primitiveConstantFillSpur]
		ifFalse: [self primitiveConstantFillV3]! !

!InterpreterPrimitives methodsFor: 'sound primitives' stamp: 'eem 5/31/2020 15:35'!
primitiveConstantFillSpur
	"Fill the receiver, which must be an indexable non-pointer object, with the given integer value."
	<inline: true>
	| fillValue rcvr format end i oddBytes |
	<var: #fillValue type: #usqLong>
	<var: #end type: #usqInt>
	<var: #i type: #usqInt>
	argumentCount ~= 1 ifTrue:
		[^self primitiveFailFor: PrimErrBadNumArgs].
	rcvr := self stackValue: 1.
	((objectMemory isNonImmediate: rcvr) 
	 and: [(format := objectMemory formatOf: rcvr) >= objectMemory sixtyFourBitIndexableFormat]) ifFalse:
		[^self primitiveFailFor: PrimErrBadReceiver].
	(objectMemory isObjImmutable: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].
	fillValue := self positive64BitValueOf: self stackTop.
	self successful ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	format >= objectMemory firstShortFormat
		ifTrue:
			[format >= objectMemory firstByteFormat
				ifTrue:
					[(fillValue > 16rFF or: [format >= objectMemory firstCompiledMethodFormat]) ifTrue:
						[^self primitiveFailFor: (fillValue > 16rFF ifTrue: [PrimErrBadArgument] ifFalse: [PrimErrBadReceiver])].
					 fillValue := fillValue + (fillValue << 8) + (fillValue << 16) + (fillValue << 24).
					 oddBytes := format bitAnd: 7]
				ifFalse:
					[fillValue > 16rFFFF ifTrue:
						[^self primitiveFailFor: PrimErrBadArgument].
					 fillValue := fillValue + (fillValue << 16).
					 oddBytes := (format bitAnd: 3) << 1].
			 fillValue := fillValue + (fillValue << 32)]
		ifFalse:
			[format = objectMemory sixtyFourBitIndexableFormat
				ifTrue:
					[oddBytes := 0]
				ifFalse:
					[fillValue > 16rFFFFFFFF ifTrue:
						[^self primitiveFailFor: PrimErrBadArgument].
					 fillValue := fillValue + (fillValue << 32).
					 oddBytes := (format bitAnd: 1) << 2]].
	end := objectMemory addressAfter: rcvr.
	i := rcvr + objectMemory baseHeaderSize.
	[i < end] whileTrue:
		[objectMemory long64At: i put: fillValue.
		 i := i + 8].
	"now ensure trailing bytes are zero"
	oddBytes > 0 ifTrue:
		[self flag: #endianness.
		 fillValue := fillValue >> (8 * oddBytes).
		 objectMemory long64At: i - 8 put: fillValue].
	self pop: 1! !

!InterpreterPrimitives methodsFor: 'sound primitives' stamp: 'eem 10/12/2016 17:56'!
primitiveConstantFillV3
	"Fill the receiver, which must be an indexable bytes or words 
	 object, with the given integer value."
	<inline: true>
	| fillValue rcvr end i |
	<var: #end type: #usqInt>
	<var: #i type: #usqInt>
	fillValue := self positive32BitValueOf: self stackTop.
	rcvr := self stackValue: 1.
	self success: (objectMemory isWordsOrBytes: rcvr).
	(objectMemory isBytes: rcvr) ifTrue:
		[self success: (fillValue >= 0 and: [fillValue <= 255]).
		 fillValue := fillValue + (fillValue << 8) + (fillValue << 16) + (fillValue << 24)].
	self successful ifTrue:
		[end := rcvr + (objectMemory sizeBitsOf: rcvr).
		 i := rcvr + objectMemory baseHeaderSize.
		 [i < end] whileTrue:
			[objectMemory long32At: i put: fillValue.
			 i := i + 4].
		 self pop: 1]! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'cb 5/30/2018 17:19'!
primitiveControlVMProfiling
	"Primitive. Start or stop the VM profiler.  The first argument is a boolean
	 to switch profiling on or off.  The second argument is an integer or nil.
	 If an integer it determines the maximum number of samples in the VM's
	 sample buffer. Answer the current number of samples in the buffer."
	| onOffBar bufferSize numSamples |
	argumentCount ~= 2 ifTrue:
		[^self primitiveFail].
	(onOffBar := self stackValue: 1) = objectMemory trueObject
		ifTrue: [onOffBar := 1]
		ifFalse:
			[onOffBar = objectMemory falseObject
				ifTrue: [onOffBar := 0]
				ifFalse: [^self primitiveFail]].
	(bufferSize := self stackTop) = objectMemory nilObject
		ifTrue: [bufferSize := 0]
		ifFalse:
			[((objectMemory isIntegerObject: bufferSize)
			  and: [(bufferSize := objectMemory integerValueOf: bufferSize) > 0]) ifFalse:
				[^self primitiveFail]].
	numSamples := self cCode: 'ioControlNewProfile(onOffBar,bufferSize)' inSmalltalk: [1667].
	self pop: 3 thenPushInteger: numSamples! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 9/19/2018 17:31'!
primitiveCopyObject
	"Primitive. Copy the state of the receiver from the argument. 
		Fail if receiver and argument are of a different class.
		Fail if the receiver or argument are contexts (because of context-to-stack mapping).
		Fail if receiver and argument have different lengths (for indexable objects).
		Fail if the objects are not in a fit state to be copied (e.g. married contexts and Cogged methods)"
	| rcvr arg length |
	self methodArgumentCount >= 1 ifFalse:
		[^self primitiveFailFor: PrimErrBadNumArgs].
	arg := self stackTop.
	rcvr := self stackValue: 1.
	(objectMemory isImmediate: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	(objectMemory isImmediate: arg) ifTrue:
		[^self primitiveFailFor: PrimErrBadArgument].

	(objectMemory fetchClassTagOfNonImm: rcvr)
		~= (objectMemory fetchClassTagOfNonImm: arg) ifTrue:
		[^self primitiveFailFor: PrimErrBadArgument].

	(objectMemory isWordsOrBytesNonImm: rcvr)
		ifTrue:
			[length := objectMemory numBytesOf: rcvr.
			((objectMemory formatOf: rcvr) = (objectMemory formatOf: arg)
			  and: [length = (objectMemory numBytesOf: arg)]) ifFalse:
				[^self primitiveFailFor: PrimErrBadArgument].
			 self memcpy: (rcvr + objectMemory baseHeaderSize) asVoidPointer
				_: (arg + objectMemory baseHeaderSize) asVoidPointer
				_: length]
		ifFalse:
			[(self isAppropriateForCopyObject: rcvr) ifFalse:
				[^self primitiveFailFor: PrimErrBadReceiver].
			 length := objectMemory numSlotsOf: rcvr.
			 ((self isAppropriateForCopyObject: arg)
			  and: [length = (objectMemory lengthOf: arg)]) ifFalse:
				[^self primitiveFailFor: PrimErrBadArgument].
			 0 to: length - 1 do:
				[:i|
				objectMemory
					storePointer: i
					ofObject: rcvr
					withValue: (objectMemory fetchPointer: i ofObject: arg)]].

	"Note: The above could be faster for young receivers but I don't think it'll matter"
	self pop: self methodArgumentCount "pop arg; answer receiver"! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 9/21/2011 14:36'!
primitiveCrashVM
	"Crash the VM by indirecting through a null pointer.  If the sole argument
	 is true crash in this thread, and if it is false crash in a new thread.  If the
	 argument is an integer use the method that implies.
		bit 0 = thread to crash in; 1 => this thread
		bit 1 = crash method; 0 => indirect through null pointer; 1 => call exit"

	| crashInThisThread |
	<export: true>
	(objectMemory isIntegerObject: self stackTop)
		ifTrue: [crashInThisThread := objectMemory integerValueOf: self stackTop]
		ifFalse: [crashInThisThread := self booleanValueOf: self stackTop].
	(self failed
	 or: [argumentCount ~= 1]) ifTrue:
		[^self primitiveFailFor: PrimErrBadNumArgs].
	self crashInThisOrAnotherThread: crashInThisThread.
	self pop: 1! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem (auto pragmas 12/08) 11/25/2008 12:50'!
primitiveDisablePowerManager
	"Pass in a non-negative value to disable the architectures powermanager if any, zero to enable. This is a named (not numbered) primitive in the null module (ie the VM)"

	| integer |
	<export: true>
	integer := self stackIntegerValue: 0.
	self successful ifTrue: [
		self ioDisablePowerManager: integer.
		self pop: 1].  "integer; leave rcvr on stack"! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'tpr 3/15/2004 20:24'!
primitiveDiv
	| quotient |
	quotient := self doPrimitiveDiv: (self stackValue: 1) by: (self stackTop).
	self pop2AndPushIntegerIfOK: quotient! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'dtl 9/2/2012 10:07'!
primitiveDivLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| result oopResult a aIsNegative b bIsNegative oopArg oopRcvr rem |
	<export: true>
	<var: 'a' type: 'usqLong'>
	<var: 'b' type: 'usqLong'>
	<var: 'result' type: 'usqLong'>
	<var: 'rem' type: 'usqLong'>

	oopArg := self stackValue: 0.
	oopRcvr := self stackValue: 1.
	aIsNegative := self isNegativeIntegerValueOf: oopRcvr.
	bIsNegative := self isNegativeIntegerValueOf: oopArg.
	a := self magnitude64BitValueOf: oopRcvr.
	b := self magnitude64BitValueOf: oopArg.
	b = 0 ifTrue:[self primitiveFail].
	self successful ifFalse:[^nil].

	result := a // b.
	
	a = 0
		ifFalse: [bIsNegative = aIsNegative
			ifFalse:
				["Round toward negative infinity"
				rem := a \\ b.
				rem = 0 ifFalse:
					["This can not overflow, because b > 1, otherwise rem = 0"
					result := result + 1]]].

	oopResult := self magnitude64BitIntegerFor: result neg: bIsNegative ~= aIsNegative.
	self successful ifTrue: [self pop: 2 thenPush: oopResult].! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'di 11/30/1998 10:25'!
primitiveDivide
	| integerReceiver integerArgument |
	integerReceiver := self stackIntegerValue: 1.
	integerArgument := self stackIntegerValue: 0.
	(integerArgument ~= 0 and: [integerReceiver \\ integerArgument = 0])
		ifTrue: [self pop2AndPushIntegerIfOK: integerReceiver // integerArgument]
		ifFalse: [self primitiveFail]! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'dtl 9/2/2012 10:07'!
primitiveDivideLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| a b result oopResult aIsNegative bIsNegative oopArg oopRcvr |
	<export: true>
	<var: 'a' type: 'usqLong'>
	<var: 'b' type: 'usqLong'>
	<var: 'result' type: 'usqLong'>

	oopArg := self stackValue: 0.
	oopRcvr := self stackValue: 1.
	aIsNegative := self isNegativeIntegerValueOf: oopRcvr.
	bIsNegative := self isNegativeIntegerValueOf: oopArg.
	a := self magnitude64BitValueOf: oopRcvr.
	b := self magnitude64BitValueOf: oopArg.
	self successful ifFalse:[^nil].
	
	"check for exact division"
	(b ~= 0 and:[a \\ b = 0]) ifFalse:[self primitiveFail. ^nil].

	result := a // b.
	oopResult := self magnitude64BitIntegerFor: result neg: aIsNegative ~= bIsNegative.

	self successful ifTrue:[self pop: 2 thenPush: oopResult].
! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 10/21/2010 21:58'!
primitiveEnterCriticalSection
	"Attempt to enter a CriticalSection/Mutex.  If not owned, set the owner to the current
	 process and answer false. If owned by the current process  answer true.   Otherwise
	 suspend the process.  Answer if the receiver is owned by the current process.
	 For simulation if there is an argument it is taken to be the effective activeProcess
	 (see Process>>effectiveProcess)."
	| criticalSection owningProcessIndex owningProcess activeProc |
	argumentCount > 0
		ifTrue:
			[criticalSection := self stackValue: 1.  "rcvr"
			 activeProc := self stackTop]
		ifFalse:
			[criticalSection := self stackTop.  "rcvr"
			 activeProc := self activeProcess].
	owningProcessIndex := ExcessSignalsIndex. "CriticalSections are laid out like Semaphores"
	owningProcess := objectMemory fetchPointer: owningProcessIndex ofObject: criticalSection.
	owningProcess = objectMemory nilObject ifTrue:
		[objectMemory storePointer: owningProcessIndex
			ofObject: criticalSection
			withValue: activeProc.
		 ^self pop: argumentCount + 1 thenPush: objectMemory falseObject].
	owningProcess = activeProc ifTrue:
		[^self pop: argumentCount + 1 thenPush: objectMemory trueObject].
	"Arrange to answer false (unowned) when the process is resumed."
	self pop: argumentCount + 1 thenPush: objectMemory falseObject.
	self addLastLink: activeProc toList: criticalSection.
	self transferTo: self wakeHighestPriority! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 3/6/2015 10:07'!
primitiveEqual
	| integerReceiver integerArgument result |
	integerArgument := self stackTop.
	integerReceiver := self stackValue: 1.
	(objectMemory areIntegers: integerReceiver and: integerArgument)
		ifTrue: [self pop: 2 thenPushBool: integerReceiver = integerArgument]
		ifFalse:
			[result := objectMemory hasSixtyFourBitImmediates
						ifTrue:
							[(self signed64BitValueOf: integerReceiver)
								= (self signed64BitValueOf: integerArgument)]
						ifFalse:
							[(self positiveMachineIntegerValueOf: integerReceiver)
								= (self positiveMachineIntegerValueOf: integerArgument)].
			 self successful ifTrue:
				[self pop: 2 thenPushBool: result]]! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'eem (auto pragmas 12/08) 11/25/2008 12:50'!
primitiveEqualLargeIntegers
	"Primitive comparison operations for large integers in 64 bit range"
	| integerRcvr integerArg |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).

	self successful ifTrue:
		[self pop: 2 thenPushBool: integerRcvr = integerArg]! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 10/21/2010 16:49'!
primitiveExitCriticalSection
	"Exit the critical section.
	 This may change the active process as a result."
	| criticalSection owningProcessIndex owningProcess |
	criticalSection := self stackTop.  "rcvr"
	owningProcessIndex := ExcessSignalsIndex. "CriticalSections are laid out like Semaphores"
	(self isEmptyList: criticalSection)
		ifTrue:
			[objectMemory storePointerUnchecked: owningProcessIndex
				ofObject: criticalSection
				withValue: objectMemory nilObject]
		ifFalse:
			[owningProcess := self removeFirstLinkOfList: criticalSection.
			 "store check unnecessary because criticalSection referred to owningProcess
			  via its FirstLinkIndex slot before owningProcess was removed."
			 objectMemory storePointerUnchecked: owningProcessIndex
				ofObject: criticalSection
				withValue: owningProcess.
			 "Note that resume: isn't fair; it won't suspend the active process.
			  For fairness we must do the equivalent of a primitiveYield, but that
			  may break old code, so we stick with unfair resume:."
			self resume: owningProcess preemptedYieldingIf: preemptionYields]! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: ''!
primitiveExitToDebugger

	self error: 'Exit to debugger at user request'.! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2017 14:47'!
primitiveExp
	"Computes E raised to the receiver power.
	 N.B. IMO we should be able to assume the receiver is a float because this primitive is specific to floats. eem 2/13/2017"
	| rcvr |
	<var: #rcvr type: #double>
	rcvr := self stackFloatValue: 0.
	self successful ifTrue:
		[self stackTopPut: (objectMemory floatObjectOf:
								(self cCode: [rcvr exp]
									inSmalltalk: [rcvr = rcvr
													ifTrue: [rcvr exp]
													ifFalse: [Float nan]]))]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2017 14:49'!
primitiveExponent
	"Exponent part of this float.
	 N.B. IMO we should be able to assume the receiver is a float because this primitive is specific to floats. eem 2/13/2017"
	| rcvr pwr |
	<var: #rcvr type: #double>
	<var: #pwr type: #int>
	rcvr := self stackFloatValue: 0.
	self successful ifTrue:
		["rcvr = frac * 2^pwr, where frac is in [0.5..1.0)"
		 self cCode: [self fr: rcvr exp: (self addressOf: pwr)]
			inSmalltalk: [pwr := rcvr exponent].
		 self stackTopPut: (objectMemory integerObjectOf: pwr - 1)]! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 7/18/2011 15:50'!
primitiveFail
	"Set general (unspecified) primitive failure.  Don't overwrite an error code that has already been set."

	"Use returnTypeC: #sqInt because that's the way it is defined in sq.h.
	 Use no explicit return so that Slang doesn't fail an inlining type-check when
	 a primitive with return type void uses ^self primitiveFail to exit."
	<api>
	<returnTypeC: #sqInt>
	<inline: true>
	self successful ifTrue:
		[primFailCode := 1]! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 11/25/2009 10:43'!
primitiveFailFor: reasonCode
	"Set specific primitive failure.
	 N.B. primitiveFailFor: PrimNoErr is expected to clear the primFailCode."
	<api>
	^primFailCode := reasonCode! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 8/24/2018 15:39'!
primitiveFailForFFIException: exceptionCode at: pc
	<var: 'exceptionCode' type: #usqLong>
	<var: 'pc' type: #usqInt>
	"Set PrimErrFFIException primitive failure and associated exceptionCode (a.k.a. 
	 osErrorCode) and exceptionPC. Under control of the ffiExceptionResponse flag,
	 if in a primitive with an error code and the inFFIFlags indicate we're in an FFI call,
	 then fail the primitive.
	 ffiExceptionResponse < 0 never fail
	 ffiExceptionResponse = 0 fail if method has a primitive error code (default)
	 ffiExceptionResponse > 0 always fail"
	<api>
	((inFFIFlags noMask: DisownVMForFFICall)	"i.e. not in an FFI call"
	 or: [ffiExceptionResponse < 0]) ifTrue:		"i.e. never fail"
		[^self].
	osErrorCode := self cCoerceSimple: exceptionCode to: #sqLong.
	exceptionPC := pc.
	primFailCode := PrimErrFFIException.
	(ffiExceptionResponse > 0					"always fail..."
	 or: [(objectMemory isOopCompiledMethod: newMethod)
		 and: [self methodUsesPrimitiveErrorCode: newMethod]]) ifTrue:
		[self ownVM: DisownVMForFFICall. "To take ownership but importantly to reset inFFIFlags"
		 self activateFailingPrimitiveMethod]! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 12/7/2017 19:29'!
primitiveFailForOSError: osError
	<var: 'osError' type: #sqLong>
	"Set PrimErrOSError primitive failure and associated osErrorCode."
	<api>
	osErrorCode := osError.
	^primFailCode := PrimErrOSError! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 11/25/2009 10:40'!
primitiveFailureCode
	<api>
	^primFailCode! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 8/31/2015 10:44'!
primitiveFetchNextMourner
	<option: #SpurObjectMemory>
	objectMemory dequeueMourner
		ifNil: [self primitiveFailFor: PrimErrNotFound]
		ifNotNil: [:mourner| self pop: 1 thenPush: mourner]! !

!InterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 2/14/2013 10:37'!
primitiveFindHandlerContext
	"Primitive. Search up the context stack for the next method context marked
	 for exception handling starting at the receiver. Return nil if none found"
	self subclassResponsibility! !

!InterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 2/14/2013 10:37'!
primitiveFindNextUnwindContext
	"Primitive. Search up the context stack for the next method context marked for unwind
	 handling from the receiver up to but not including the argument. Return nil if none found."
	self subclassResponsibility! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'di 11/27/1998 11:45'!
primitiveFloatAdd
	^ self primitiveFloatAdd: (self stackValue: 1) toArg: self stackTop! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 1/18/2020 17:59'!
primitiveFloatArrayAt
	"Index the receiver, which must be an indexable non-pointer
	 object, and yield a float."
	objectMemory hasSpurMemoryManagerAPI
		ifTrue: [self primitiveSpurFloatArrayAt]
		ifFalse: [self primitiveFailFor: PrimErrUnsupported]! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 1/18/2020 17:59'!
primitiveFloatArrayAtPut
	"Index the receiver, which must be an indexable non-pointer
	 object, and store a float."
	objectMemory hasSpurMemoryManagerAPI
		ifTrue: [self primitiveSpurFloatArrayAtPut]
		ifFalse: [self primitiveFailFor: PrimErrUnsupported]! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 12/26/2014 15:51'!
primitiveFloatAt
	"Provide platform-independent access to 32-bit words comprising
	 a Float.  Map index 1 onto the most significant word and index 2
	 onto the least significant word."
	| rcvr index result |
	<var: #result type: #usqInt>
	rcvr := self stackValue: 1.
	index := self stackTop.
	index = ConstOne ifTrue:
		[result := self positive32BitIntegerFor:
					(objectMemory
						fetchLong32: (VMBIGENDIAN ifTrue: [0] ifFalse: [1])
						ofFloatObject: rcvr).
		^self pop: 2 thenPush: result].
	index = ConstTwo ifTrue:
		[result := self positive32BitIntegerFor:
					(objectMemory
						fetchLong32: (VMBIGENDIAN ifTrue: [1] ifFalse: [0])
						ofFloatObject: rcvr).
		^self pop: 2 thenPush: result].
	self primitiveFailFor: ((objectMemory isIntegerObject: index)
							ifTrue: [PrimErrBadIndex]
							ifFalse: [PrimErrBadArgument])! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 1/20/2016 14:29'!
primitiveFloatAtPut
	"Provide platform-independent access to 32-bit words comprising
	 a Float.  Map index 1 onto the most significant word and index 2
	 onto the least significant word."
	| rcvr index oopToStore valueToStore |
	<var: #valueToStore type: #usqInt>
	oopToStore := self stackTop.
	valueToStore := self positive32BitValueOf: oopToStore.
	self successful ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	rcvr := self stackValue: 2.
	index := self stackValue: 1.
	(objectMemory isImmediateFloat: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	(objectMemory isObjImmutable: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].
	index = ConstOne ifTrue:
		[objectMemory storeLong32: (VMBIGENDIAN ifTrue: [0] ifFalse: [1])
			ofObject: rcvr
			withValue: valueToStore.
		^self pop: 3 thenPush: oopToStore].
	index = ConstTwo ifTrue:
		[objectMemory storeLong32: (VMBIGENDIAN ifTrue: [1] ifFalse: [0])
			ofObject: rcvr
			withValue: valueToStore.
		^self pop: 3 thenPush: oopToStore].
	self primitiveFailFor: ((objectMemory isIntegerObject: index)
							ifTrue: [PrimErrBadIndex]
							ifFalse: [PrimErrBadArgument])! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 8/11/2008 10:40'!
primitiveFloatDivide
	self primitiveFloatDivide: (self stackValue: 1) byArg: self stackTop! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 11/25/2008 12:50'!
primitiveFloatEqual
	| aBool |
	aBool := self primitiveFloatEqual: (self stackValue: 1) toArg: self stackTop.
	self successful ifTrue:
		[self pop: 2 thenPushBool: aBool]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'nice (eem) 1/6/2009 09:58'!
primitiveFloatGreaterOrEqual
	| aBool |
	aBool := self primitiveFloatGreaterOrEqual: (self stackValue: 1) toArg: self stackTop.
	self successful ifTrue: [self pop: 2 thenPushBool: aBool]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 11/25/2008 12:50'!
primitiveFloatGreaterThan
	| aBool |
	aBool := self primitiveFloatGreater: (self stackValue: 1) thanArg: self stackTop.
	self successful ifTrue:
		[self pop: 2 thenPushBool: aBool]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'nice (eem) 1/6/2009 09:58'!
primitiveFloatLessOrEqual
	| aBool |
	aBool := self primitiveFloatLessOrEqual: (self stackValue: 1) toArg: self stackTop.
	self successful ifTrue: [self pop: 2 thenPushBool: aBool]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 11/25/2008 12:50'!
primitiveFloatLessThan
	| aBool |
	aBool := self primitiveFloatLess: (self stackValue: 1) thanArg: self stackTop.
	self successful ifTrue:
		[self pop: 2 thenPushBool: aBool]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 8/27/2010 10:26'!
primitiveFloatMultiply
	self primitiveFloatMultiply: (self stackValue: 1) byArg: self stackTop! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 11/25/2008 12:50'!
primitiveFloatNotEqual
	| aBool |
	aBool := self primitiveFloatEqual: (self stackValue: 1) toArg: self stackTop.
	self successful ifTrue:
		[self pop: 2 thenPushBool: aBool not]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'di 11/27/1998 11:46'!
primitiveFloatSubtract
	^ self primitiveFloatSubtract: (self stackValue: 1) fromArg: self stackTop! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 2/10/2015 18:35'!
primitiveFlushCache
	"Clear the method lookup cache. This must be done after every programming change."

	self flushMethodCache! !

!InterpreterPrimitives methodsFor: 'plugin primitives' stamp: 'eem 8/11/2008 10:38'!
primitiveFlushExternalPrimitives
	"Primitive. Flush all the existing external primitives in the image thus forcing a reload on next invokation."
	self flushExternalPrimitives! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'jm 5/17/1998 07:06'!
primitiveForceDisplayUpdate
	"On some platforms, this primitive forces enqueued display updates to be processed immediately. On others, it does nothing."

	self ioForceDisplayUpdate.
! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 10/15/2013 10:13'!
primitiveForceTenure
	"Set force tenure flag to true, this forces a tenure operation on the next incremental GC"

	<export: true>
	<option: #SqueakV3ObjectMemory>
	objectMemory forceTenureFlag: 1! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 11/21/2014 17:13'!
primitiveFormPrint
	"On platforms that support it, this primitive prints the receiver, assumed to be a Form, to the default printer."

	| landscapeFlag vScale hScale rcvr bitsArray w h
	 depth pixelsPerWord wordsPerLine bitsArraySize ok |

	<var: #vScale type: #double>
	<var: #hScale type: #double>
	landscapeFlag := self booleanValueOf: self stackTop.
	vScale := objectMemory floatValueOf: (self stackValue: 1).
	hScale := objectMemory floatValueOf: (self stackValue: 2).
	rcvr := self stackValue: 3.
	((objectMemory isPointers: rcvr)
	 and: [(objectMemory lengthOf: rcvr) >= 4]) ifFalse:
		[self success: false].
	self successful ifTrue:
		[bitsArray := objectMemory fetchPointer: 0 ofObject: rcvr.
		w := self fetchInteger: 1 ofObject: rcvr.
		h := self fetchInteger: 2 ofObject: rcvr.
		depth := self fetchInteger: 3 ofObject: rcvr.
		(w > 0 and: [h > 0]) ifFalse: [self success: false].
		pixelsPerWord := 32 // depth.
		wordsPerLine := (w + (pixelsPerWord - 1)) // pixelsPerWord.
		(objectMemory isWordsOrBytes: bitsArray)
			ifTrue:
				[bitsArraySize := objectMemory numBytesOf: bitsArray.
				self success: (bitsArraySize = (wordsPerLine * h * 4))]
			ifFalse: [self success: false]].	
	self successful ifTrue:
		[ok := self cCode: 'ioFormPrint(bitsArray + BaseHeaderSize, w, h, depth, hScale, vScale, landscapeFlag)'.
		self success: ok].
	self successful ifTrue:
		[self pop: 3]	"pop hScale, vScale, and landscapeFlag; leave rcvr on stack"! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2017 14:51'!
primitiveFractionalPart
	"Fractional part of this float.
	 N.B. IMO we should be able to assume the receiver is a float because this primitive is specific to floats. eem 2/13/2017"
	| rcvr trunc |
	<var: #rcvr type: #double>
	<var: #trunc type: #double>
	rcvr := self stackFloatValue: 0.
	self successful ifTrue:
		[self stackTopPut: (objectMemory floatObjectOf:
							(self cCode: [self mod: rcvr f: (self addressOf: trunc)]
								inSmalltalk: [rcvr fractionPart]))]! !

!InterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 4/23/2017 11:40'!
primitiveFullClosureValue
	<option: #SistaV1BytecodeSet>
	| blockClosure numArgs closureMethod |
	blockClosure := self stackValue: argumentCount.
	numArgs := self argumentCountOfClosure: blockClosure.
	argumentCount = numArgs ifFalse:
		[^self primitiveFail].

	closureMethod := objectMemory fetchPointer: FullClosureCompiledBlockIndex ofObject: blockClosure.
	(objectMemory isOopCompiledMethod: closureMethod) ifFalse:
		[^self primitiveFail].

	self activateNewFullClosure: blockClosure method: closureMethod numArgs: numArgs mayContextSwitch: true! !

!InterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 4/23/2017 11:41'!
primitiveFullClosureValueNoContextSwitch
	"An exact clone of primitiveFullClosureValue except that this version will not
	 check for interrupts on stack overflow.  It may invoke the garbage collector
	 but will not switch processes.  See checkForInterruptsMayContextSwitch:"
	<api>
	<option: #SistaV1BytecodeSet>
	| blockClosure numArgs closureMethod |
	blockClosure := self stackValue: argumentCount.
	numArgs := self argumentCountOfClosure: blockClosure.
	argumentCount = numArgs ifFalse:
		[^self primitiveFail].

	closureMethod := objectMemory fetchPointer: FullClosureCompiledBlockIndex ofObject: blockClosure.
	(objectMemory isOopCompiledMethod: closureMethod) ifFalse:
		[^self primitiveFail].

	self activateNewFullClosure: blockClosure method: closureMethod numArgs: numArgs mayContextSwitch: false! !

!InterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 4/23/2017 11:41'!
primitiveFullClosureValueWithArgs
	<option: #SistaV1BytecodeSet>
	| argumentArray arraySize blockClosure numArgs closureMethod index |
	argumentArray := self stackTop.
	(objectMemory isArray: argumentArray) ifFalse:
		[^self primitiveFail].

	"Check for enough space in thisContext to push all args"
	arraySize := objectMemory numSlotsOf: argumentArray.
	(self roomToPushNArgs: arraySize) ifFalse:
		[^self primitiveFail].

	blockClosure := self stackValue: argumentCount.
	numArgs := self argumentCountOfClosure: blockClosure.
	arraySize = numArgs ifFalse:
		[^self primitiveFail].

	closureMethod := objectMemory fetchPointer: FullClosureCompiledBlockIndex ofObject: blockClosure.
	(objectMemory isOopCompiledMethod: closureMethod) ifFalse:
		[^self primitiveFail].

	self popStack.

	"Copy the arguments to the stack, and activate"
	index := 1.
	[index <= numArgs] whileTrue:
		[self push: (objectMemory fetchPointer: index - 1 ofObject: argumentArray).
		index := index + 1].

	self activateNewFullClosure: blockClosure method: closureMethod numArgs: numArgs mayContextSwitch: true! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 11/9/2020 19:17'!
primitiveFullGC
	"Do a full garbage collection.  In SqueakV3ObjectMemory, answer the number
	 of bytes available (including swap space if dynamic memory management is
	 supported).  In Spur, answer the size of the largest free chunk."

	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[self methodReturnInteger: objectMemory fullGC.
		 ^self].
	objectMemory incrementalGC.  "maximimize space for forwarding table"
	objectMemory fullGC.
	self methodReturnInteger: (objectMemory bytesLeft: true)! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
primitiveGetAttribute
	"Fetch the system attribute with the given integer ID. The 
	result is a string, which will be empty if the attribute is not 
	defined."
	| attr sz s |
	attr := self stackIntegerValue: 0.
	self successful
		ifTrue: [sz := self attributeSize: attr].
	self successful
		ifTrue: [s := objectMemory
						instantiateClass: (objectMemory splObj: ClassByteString)
						indexableSize: sz.
			self
				getAttribute: attr
				Into: s + objectMemory baseHeaderSize
				Length: sz.
			self pop: 2 thenPush: s]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'cb 4/19/2016 09:35'!
primitiveGetImmutability
	<option: #IMMUTABILITY>
	| rcvr |
	rcvr := self stackValue: 0.
	self pop: argumentCount + 1 thenPushBool: (objectMemory isOopImmutable: rcvr)! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 7/19/2017 10:52'!
primitiveGetLogDirectory
	"Primitive. Answer the VM's current log directory"
	| ptr sz stringOop |
	<var: 'ptr' type: 'char*'>
	<export: true>
	ptr := self ioGetLogDirectory.
	ptr == nil ifTrue:[^self success: false].
	sz := self strlen: ptr.
	stringOop := objectMemory instantiateClass: objectMemory classString indexableSize: sz.
	0 to: sz-1 do:[:i| objectMemory storeByte: i ofObject: stringOop withValue: (ptr at: i)].
	self pop: argumentCount+1 thenPush: stringOop! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 12/22/2020 16:47'!
primitiveGetNextEvent
	"Primitive. Return the next input event from the VM event queue."
	| evtBuf arg value eventTypeIs |
	<var: #evtBuf declareC:'sqIntptr_t evtBuf[8] = { 0, 0, 0, 0, 0, 0, 0, 0 }'>
	self cCode: [] inSmalltalk: [evtBuf := objectMemory newInputEventAccessorOfSize: 8].
	arg := self stackTop.
	((objectMemory isArray: arg) and:[(objectMemory slotSizeOf: arg) = 8]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].

	self ioGetNextEvent: (self cCoerce: evtBuf to: 'sqInputEvent*').
	self successful ifFalse:
		[^nil].

	(eventTraceMask ~= 0
	 and: [eventTraceMask anyMask: (1 bitShift: (evtBuf at: 0))]) ifTrue:
		[self traceInputEvent: evtBuf].

	"Event type"
	eventTypeIs := evtBuf at: 0.
	self storeInteger: 0 ofObject: arg withValue: (evtBuf at: 0).
	self successful ifFalse:
		[^nil].

	eventTypeIs = 6 
		ifTrue: "Event is Complex, assume evtBuf is populated correctly and return"
			[1 to: 7 do: [:i |
				value := evtBuf at: i.
				self storePointer: i ofObject: arg withValue: value]]
		ifFalse:
			["Event time stamp"
			self storeInteger: 1 ofObject: arg withValue: ((evtBuf at: 1) bitAnd: MillisecondClockMask).
			self successful ifFalse:
				[^nil].	

			"Event arguments"
			2 to: 7 do:[:i|
				value := evtBuf at: i.
				(objectMemory isIntegerValue: value)
					ifTrue:[self storeInteger: i ofObject: arg withValue: value]
					ifFalse:
						[value := self positiveMachineIntegerFor: value.
						objectMemory storePointer: i ofObject: arg withValue: value]]].

	self successful ifTrue: [self pop: 1]! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 9/6/2020 16:43'!
primitiveGetWindowLabel
	"Primitive. Answer the OS window's label"
	<export: true>
	self methodReturnString: self ioGetWindowLabel! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'ar (auto pragmas 12/08) 2/27/2008 10:54'!
primitiveGetWindowSize
	"Primitive. Answer the OS window's size (extent)"
	| w h |
	<export: true>
	w := self ioGetWindowWidth.
	h := self ioGetWindowHeight.
	self pop: self methodArgumentCount+1.
	self push: (self makePointwithxValue: w yValue: h).! !

!InterpreterPrimitives methodsFor: 'other primitives' stamp: 'akg 10/23/2018 18:51'!
primitiveGetenv
	"Access to environment variables via getenv.  No putenv or setenv as yet."
	| key var result |
	<export: true>
	<var: #key type: #'char *'>
	<var: #var type: #'char *'>
	sHEAFn ~= 0 ifTrue: "secHasEnvironmentAccess"
		[self sHEAFn ifFalse: [^self primitiveFailFor: PrimErrInappropriate]].
	key := self cStringOrNullFor: self stackTop.
	key = 0 ifTrue:
		[self successful ifTrue:
			[^self primitiveFailFor: PrimErrBadArgument].
		 ^self primitiveFailFor: primFailCode].
	var := self getenv: key.
	self free: key.
	var ~= 0 ifTrue:
		[result := objectMemory stringForCString: var.
		 result ifNil:
			[^self primitiveFailFor: PrimErrNoMemory]].
	self assert: primFailCode = 0.
	self pop: 2 thenPush: (var = 0 ifTrue: [objectMemory nilObject] ifFalse: [result])! !

!InterpreterPrimitives methodsFor: 'simulation support' stamp: 'eem 10/15/2018 17:22'!
primitiveGetenv: aByteStringOrByteArray
	<doNotGenerate>
	<primitive: 'primitiveGetenv' module: '' error: ec>
	ec == #'bad argument' ifTrue:
		[aByteStringOrByteArray isString ifFalse:
			[^self getenv: aByteStringOrByteArray asString]].
	self primitiveFail! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 3/11/2015 08:55'!
primitiveGreaterOrEqual
	| integerReceiver integerArgument |
	integerArgument := self stackTop.
	integerReceiver := self stackValue: 1.
	(objectMemory areIntegers: integerReceiver and: integerArgument)
		ifTrue: [self cCode: '' inSmalltalk:
					[integerReceiver := objectMemory integerValueOf: integerReceiver.
					 integerArgument := objectMemory integerValueOf: integerArgument].
				self pop: 2 thenPushBool: integerReceiver >= integerArgument]
		ifFalse: [self primitiveFail]! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'eem (auto pragmas 12/08) 11/25/2008 12:50'!
primitiveGreaterOrEqualLargeIntegers
	"Primitive comparison operations for large integers in 64 bit range"
	| integerRcvr integerArg |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).

	self successful ifTrue:
		[self pop: 2 thenPushBool: integerRcvr >= integerArg]! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 3/11/2015 08:55'!
primitiveGreaterThan
	| integerReceiver integerArgument |
	integerArgument := self stackTop.
	integerReceiver := self stackValue: 1.
	(objectMemory areIntegers: integerReceiver and: integerArgument)
		ifTrue: [self cCode: '' inSmalltalk:
					[integerReceiver := objectMemory integerValueOf: integerReceiver.
					 integerArgument := objectMemory integerValueOf: integerArgument].
				self pop: 2 thenPushBool: integerReceiver > integerArgument]
		ifFalse: [self primitiveFail]! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'eem (auto pragmas 12/08) 11/25/2008 12:50'!
primitiveGreaterThanLargeIntegers
	"Primitive comparison operations for large integers in 64 bit range"
	| integerRcvr integerArg |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).

	self successful ifTrue:
		[self pop: 2 thenPushBool: integerRcvr > integerArg]! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 10/16/2013 16:00'!
primitiveGrowMemoryByAtLeast
	<option: #SpurObjectMemory>
	| ammount |
	ammount := self stackTop.
	(objectMemory isIntegerObject: ammount) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory growOldSpaceByAtLeast: (objectMemory integerValueOf: ammount))
		ifNil: [self primitiveFailFor: PrimErrNoMemory]
		ifNotNil: [:segSize| self pop: 2 thenPushInteger: segSize]! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'nice 1/13/2020 22:16'!
primitiveHashMultiply
	"Implement 28-bit hashMultiply for SmallInteger and LargePositiveInteger receivers."
	| value |
	value := self stackTop.
	(objectMemory isIntegerObject: value)
		ifTrue: [value := objectMemory integerValueOf: value]
		ifFalse:
			[| ok |
			 ok := objectMemory is: value instanceOf: (objectMemory splObj: ClassLargePositiveInteger) compactClassIndex: ClassLargePositiveIntegerCompactIndex.
			 ok ifFalse:
				[^self primitiveFailFor: PrimErrBadReceiver].
			 value := objectMemory byteSwapped32IfBigEndian: (objectMemory fetchLong32: 0 ofObject: value)].
	self pop: 1
		thenPush: (objectMemory integerObjectOf: ((self cCoerce: value to: #usqInt) * HashMultiplyConstant bitAnd: 16rFFFFFFF))! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
primitiveHeartbeatFrequency
	"Answer the heartbeat frequency in beats per second.  If the argument is true, reset the frequency measure."
	<export: true>
	| reset |
	reset := argumentCount = 1 and: [self stackTop = objectMemory trueObject].
	self pop: argumentCount + 1
		thenPush: (self positive32BitIntegerFor: (self ioHeartbeatFrequency: reset))! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 1/14/2021 21:45'!
primitiveHighBit
	| integerReceiverOop leadingZeroCount highestBitZeroBased |
	integerReceiverOop := self stackTop.
	"Convert the receiver Oop to use a single tag bit"
	objectMemory numSmallIntegerTagBits > 1 ifTrue:
		[integerReceiverOop := (integerReceiverOop >>> (objectMemory numSmallIntegerTagBits-1) bitOr: 1)].
	self cppIf: #'__GNUC__' defined
		ifTrue:
			["Note: in gcc, result is undefined if input is zero (for compatibility with BSR fallback when no CLZ instruction available).
			but input is never zero because we pass the oop with tag bits set, so we are safe"
			objectMemory wordSize = 4
				ifTrue: [leadingZeroCount := self __builtin_clz: integerReceiverOop]
				ifFalse: [leadingZeroCount := self __builtin_clzll: integerReceiverOop].
			leadingZeroCount = 0 ifTrue: "highBit is not defined for negative Integer"
				[^self primitiveFail].
			"Nice bit trick: 1-based high-bit is (32 - clz) - 1 to account for tag bit.
			 This is like two-complement - clz - 1 on 5 bits, or in other words a bit-invert operation clz ^16r1F"
			self pop: 1 thenPushInteger: (leadingZeroCount bitXor: (BytesPerWord * 8 - 1))]
		cppIf: #'_MSC_VER' defined | #'__ICC' defined
		ifTrue:
			["In MSVC, _lzcnt and _lzcnt64 builtins do not fallback to BSR when not supported by CPU
			Instead of messing with __cpuid() we always use the BSR intrinsic"

			"Trick: we test the oop sign rather than the integerValue. Assume oop are signed (so far, they are, sqInt are signed)"
			integerReceiverOop < 0 ifTrue:
				[^self primitiveFail].
			"Setting this variable is useless, but VMMaker will generate an automatic initialization at a worse place if this isn't initialized explicitly."
			highestBitZeroBased := 0.
			"We do not even test the return value, because integerReceiverOop is never zero"
			objectMemory wordSize = 4
					ifTrue: [self _BitScanReverse: (self addressOf: highestBitZeroBased) _: integerReceiverOop]
					ifFalse: [self _BitScanReverse64: (self addressOf: highestBitZeroBased) _: integerReceiverOop].
			"thanks to the tag bit, the +1 operation for getting 1-based rank is not necessary"
			self pop: 1 thenPushInteger: highestBitZeroBased]
		ifFalse:
			["not gcc/clang, nor MSVC/ICC, you have to implement if your compiler provides useful builtins"
			self cCode:
					[self primitiveFail]
				inSmalltalk: "Simulate so that the simulator is closer to the actual VM"
					[integerReceiverOop < 0 ifTrue:
						[^self primitiveFail].
					 self pop: 1 thenPushInteger: integerReceiverOop highBit - 1]]! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 8/4/2021 15:01'!
primitiveHighResClock
	"Return the value of the high resolution clock if this system has any.
	 The exact frequency of the high res clock is undefined specifically so that we can use
	 processor dependent instructions (like RDTSC). The only use for the high res clock is for
	 profiling where we can allocate time based on sub-msec resolution of the high res clock.
	 If no high-resolution counter is available, the platform should return zero. ar 6/22/2007"
	<export: true flags: #FastCPrimitiveFlag>
	self methodReturnValue: (self positive64BitIntegerFor: self ioHighResClock)! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 3/16/2021 12:37'!
primitiveIdentical
	"is the receiver/first argument the same object as the (last) argument?.
	 pop argumentCount because this can be used as a mirror primitive."
	| thisObject otherObject |
	thisObject := self stackValue: 1.
	otherObject := self stackTop.
	((objectMemory isOopForwarded: otherObject)
	 or: [argumentCount > 1
		 and: [objectMemory isOopForwarded: thisObject]])
		ifTrue:
			[self primitiveFailFor: PrimErrBadArgument]
		ifFalse:
			[self pop: argumentCount + 1 thenPushBool: thisObject = otherObject]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 6/20/2015 15:25'!
primitiveIdentityHash
	| thisReceiver |
	thisReceiver := self stackTop.
	((objectMemory isImmediate: thisReceiver)
	 or: [argumentCount > 0
		 and: [objectMemory isForwarded: thisReceiver]])
		ifTrue: [self primitiveFail]
		ifFalse: [self pop: argumentCount + 1
					thenPushInteger: (objectMemory hashBitsOf: thisReceiver)]! !

!InterpreterPrimitives methodsFor: 'other primitives' stamp: 'dtl (auto pragmas dtl 2010-09-26) 4/10/2010 12:32'!
primitiveImageFormatVersion
	"Answer an integer identifying the type of image. The image version number may
	identify the format of the image (e.g. 32 or 64-bit word size) or specific requirements
	of the image (e.g. block closure support required).
	
	This is a named (not numbered) primitive in the null module (ie the VM)"

	<export: true>
	self pop: 1 thenPush: (self positive32BitIntegerFor: self imageFormatVersion)
! !

!InterpreterPrimitives methodsFor: 'other primitives' stamp: 'eem 8/16/2020 16:19'!
primitiveImageName
	"When called with a single string argument, record the string as the current image file name.
	 When called with zero arguments, return a string containing the current image file name."

	| s isString sCRIfn okToRename sz |
	<var: #sCRIfn type: #'void *'>
	argumentCount = 1 ifTrue:
		[s := self stackTop.
		 isString := self isInstanceOfClassByteString: s.
		 isString ifFalse:
			[^self primitiveFailFor: PrimErrBadArgument].
		"If the security plugin can be loaded, use it to check for rename permission.
		 If not, assume it's ok"
		sCRIfn := self ioLoadFunction: 'secCanRenameImage' From: 'SecurityPlugin'.
		sCRIfn ~= 0 ifTrue:
			[okToRename := self cCode: '((sqInt (*)(void))sCRIfn)()'
								inSmalltalk: [self dispatchMappedPluginEntry: sCRIfn].
			 okToRename ifFalse:
				[^self primitiveFailFor: PrimErrUnsupported]].
		self imageNamePut: (s + objectMemory baseHeaderSize) Length: (objectMemory numBytesOf: s).
		^self pop: 1].  "pop s, leave rcvr on stack"

	"A char *ioImageName(void) style interface would be less cumbersome."
	sz := self imageNameSize.
	s := objectMemory instantiateClass: (objectMemory splObj: ClassByteString) indexableSize: sz.
	self imageNameGet: (s + objectMemory baseHeaderSize) Length: sz.
	self methodReturnValue: s! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 5/7/2020 19:08'!
primitiveImmediateAsInteger
	"For a Smalllnteger, answer itself.
	 For a Character, answer its code as an unsigned integer.
	 For a SmallFloat, answer the signed, but unadjusted bit pattern (so as to keep the result a SmallInteger).
	 This is a good value for an immediate's hash."
	<option: #SpurObjectMemory>
	| oop value |
	oop := self stackTop.
	value := self
				immediateAsInteger: oop
				ifFail: [^self primitiveFailFor: PrimErrBadReceiver].
	self pop: argumentCount + 1 thenPushInteger: value! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 6/3/2016 11:42'!
primitiveIncrementalGC
	"Do a quick, incremental garbage collection and return the number of bytes immediately available.
	 (Note: more space may be made available by doing a full garbage collection."

	objectMemory hasSpurMemoryManagerAPI
		ifTrue: [objectMemory scavengingGC]
		ifFalse: [objectMemory incrementalGC].
	self pop: 1 thenPushInteger: (objectMemory bytesLeft: false)! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 6/29/2014 11:42'!
primitiveInputSemaphore
	"Register the input semaphore. If the argument is not a 
	Semaphore, unregister the current input semaphore."
	| arg |
	arg := self stackTop.
	(objectMemory isIntegerObject: arg) ifTrue:
		["If arg is integer, then use it as an index 
		  into the external objects array and install it 
		  as the new event semaphore"
		 self ioSetInputSemaphore: (objectMemory integerValueOf: arg).
		 self successful ifTrue:
			[self pop: 1].
		 ^nil].

	"old code for compatibility"
	TheInputSemaphore = nil ifTrue:
		[^self primitiveFail].
	(arg = objectMemory nilObject
	 or: [objectMemory isSemaphoreOop: arg])
		ifTrue:
			[objectMemory splObj: TheInputSemaphore put: arg.
			 self pop: 1]
		ifFalse:
			[self primitiveFailFor: PrimErrBadArgument]! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'tpr 5/13/2005 10:17'!
primitiveInputWord
	"Return an integer indicating the reason for the most recent input interrupt."

	self pop: 1 thenPushInteger: 0.	"noop for now"! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 6/20/2015 10:46'!
primitiveInstVarAt
	| index rcvr hdr fmt totalLength fixedFields value |
	index := self stackTop.
	rcvr := self stackValue: 1.
	((objectMemory isNonIntegerObject: index)
	 or: [argumentCount > 1 "e.g. object:instVarAt:"
		and: [objectMemory isOopForwarded: rcvr]]) ifTrue:
		[^self primitiveFailFor: PrimErrBadArgument].
	index := objectMemory integerValueOf: index.
	hdr := self baseHeader: rcvr.
	fmt := objectMemory formatOfHeader: hdr.
	totalLength := self lengthOf: rcvr baseHeader: hdr format: fmt.
	fixedFields := self fixedFieldsOf: rcvr format: fmt length: totalLength.
	(index >= 1 and: [index <= fixedFields]) ifFalse:
		[^self primitiveFailFor: PrimErrBadIndex].
	value := self subscript: rcvr with: index format: fmt.
	self pop: argumentCount + 1 thenPush: value! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 2/26/2018 12:36'!
primitiveInstVarAtPut
	| newValue index rcvr hdr fmt totalLength fixedFields |
	newValue := self stackTop.
	index := self stackValue: 1.
	rcvr := self stackValue: 2.
	((objectMemory isNonIntegerObject: index)
	 or: [argumentCount > 2 "e.g. object:instVarAt:put:"
		and: [objectMemory isOopForwarded: rcvr]]) ifTrue:
		[^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory isImmediate: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	(objectMemory isObjImmutable: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].
	index := objectMemory integerValueOf: index.
	hdr := objectMemory baseHeader: rcvr.
	fmt := objectMemory formatOfHeader: hdr.
	totalLength := objectMemory lengthOf: rcvr baseHeader: hdr format: fmt.
	fixedFields := objectMemory fixedFieldsOf: rcvr format: fmt length: totalLength.
	(index >= 1 and: [index <= fixedFields]) ifFalse:
		[^self primitiveFailFor: PrimErrBadIndex].
	self subscript: rcvr with: index storing: newValue format: fmt.
	self pop: argumentCount + 1 thenPush: newValue! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 6/21/2018 17:08'!
primitiveIntegerAt
	SPURVM
		ifTrue: [self primitiveSpurIntegerAt] "Answer the signed integer element of a pure bits receiver"
		ifFalse: [self primitiveV3IntegerAt]    "Answer the 32 bit signed integer contents of a words receiver"! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 6/29/2018 11:16'!
primitiveIntegerAtPut
	SPURVM
		ifTrue: [self primitiveSpurIntegerAtPut] "Assign an indexable variable of a pure bits receiver with a signed integer."
		ifFalse: [self primitiveV3IntegerAtPut]    "Assign an indexable variable of a words receiver with a 32 bit signed integer."! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
primitiveInterruptChecksPerMSec
	"Primitive. Answer the number of interrupt checks per milliseconds that we execute
	on this machine. This can be used to adjust the sub-msecs profiler to check (roughly) 
	n times per millisecond."
	<export: true>
	self methodArgumentCount = 0 ifFalse:[^self success: false].
	self pop: 1 thenPush: (objectMemory integerObjectOf: self ioHeartbeatMilliseconds).! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 6/29/2014 11:41'!
primitiveInterruptSemaphore
	"Register the user interrupt semaphore. If the argument is 
	not a Semaphore, unregister the current interrupt 
	semaphore. "
	| arg |
	arg := self stackTop.
	(arg = objectMemory nilObject
	 or: [objectMemory isSemaphoreOop: arg])
		ifTrue:
			[objectMemory splObj: TheInterruptSemaphore put: arg.
			 self pop: 1]
		ifFalse:
			[self primitiveFailFor: PrimErrBadArgument]! !

!InterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 10/14/2013 13:39'!
primitiveInvokeObjectAsMethod
	"Primitive. 'Invoke' an object like a function, sending the special message 
		run: originalSelector with: arguments in: aReceiver.
	"
	<returnTypeC: #void>
	| runReceiver runArgs lookupClassTag |
	runArgs := objectMemory hasSpurMemoryManagerAPI
					ifTrue: [objectMemory
								eeInstantiateClassIndex: ClassArrayCompactIndex
								format: objectMemory arrayFormat
								numSlots: argumentCount]
					ifFalse: [objectMemory
								eeInstantiateClass: (objectMemory splObj: ClassArray)
								indexableSize: argumentCount].
	argumentCount - 1 to: 0 by: -1  do:
		[:i| objectMemory storePointerUnchecked: i ofObject: runArgs withValue: self popStack].

	runReceiver := self popStack.
	"setup send of newMethod run: originalSelector with: runArgs in: runReceiver"
	self push: newMethod. "newReceiver"
	self push: messageSelector "original selector".
	self push: runArgs.
	self push: runReceiver.

	"stack is clean here"

	messageSelector := objectMemory splObj: SelectorRunWithIn.
	argumentCount := 3.
	lookupClassTag := objectMemory fetchClassTagOf: newMethod.
	self findNewMethodInClassTag: lookupClassTag.
	self executeNewMethod.  "Recursive xeq affects successFlag"
	self initPrimCall! !

!InterpreterPrimitives methodsFor: 'other primitives' stamp: 'eem 4/1/2015 11:34'!
primitiveIsBigEnder
	"Answer if running on a big endian machine."
	<export: true>
	self pop: 1 thenPushBool: VMBIGENDIAN! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 5/1/2016 20:51'!
primitiveIsPinned
	"Answer if the receiver is pinned, i.e. immobile."
	<option: #SpurObjectMemory>
	| obj |
	obj := self stackTop.
	((objectMemory isImmediate: obj)
	 or: [objectMemory isForwarded: obj]) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	self pop: argumentCount + 1
		thenPushBool: (objectMemory isPinned: obj)! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 10/14/2013 11:09'!
primitiveIsRoot
	"Primitive. Answer whether the argument to the primitive is a root for young space"
	| oop |
	<export: true>
	<option: #SqueakV3ObjectMemory>
	oop := self stackObjectValue: 0.
	self successful ifTrue:
		[self
			pop: argumentCount + 1
			thenPushBool: (objectMemory isYoungRoot: oop)]! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'ar 1/6/2009 16:40'!
primitiveIsWindowObscured
	"Primitive. Answer whether the OS window is currently partially or fully obscured."
	<export: true>
	self pop: self methodArgumentCount+1.
	self pushBool: self ioIsWindowObscured.
! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 10/14/2013 20:17'!
primitiveIsYoung
	"Primitive. Answer whether the argument to the primitive resides in young space."
	| oop |
	<export: true>
	<option: #SqueakV3ObjectMemory> "for now..."
	oop := self stackObjectValue: 0.
	self successful ifTrue:
		[self pop: argumentCount + 1 thenPushBool: (self oop: oop isGreaterThanOrEqualTo: objectMemory youngStart)]! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 3/5/2021 12:27'!
primitiveKbdNext
	"Obsolete on virtually all platforms; old style input polling code.
	Return the next keycode and remove it from the input buffer. The low byte is the 8-bit ISO character. The next four bits are the Smalltalk modifier bits <cmd><option><ctrl><shift>."
	<option: #Legacy>
	| keystrokeWord |
	self pop: 1.
	keystrokeWord := self ioGetKeystroke.
	keystrokeWord >= 0
		ifTrue: [self pushInteger: keystrokeWord]
		ifFalse: [self push: objectMemory nilObject].! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 3/5/2021 12:27'!
primitiveKbdPeek
	"Obsolete on virtually all platforms; old style input polling code.
	Return the next keycode and without removing it from the input buffer. The low byte is the 8-bit ISO character. The next four bits are the Smalltalk modifier bits <cmd><option><ctrl><shift>."
	<option: #Legacy>
	| keystrokeWord |
	self pop: 1.
	keystrokeWord := self ioPeekKeystroke.
	keystrokeWord >= 0
		ifTrue: [self pushInteger: keystrokeWord]
		ifFalse: [self push: objectMemory nilObject].! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 3/11/2015 08:56'!
primitiveLessOrEqual
	| integerReceiver integerArgument |
	integerArgument := self stackTop.
	integerReceiver := self stackValue: 1.
	(objectMemory areIntegers: integerReceiver and: integerArgument)
		ifTrue: [self cCode: '' inSmalltalk:
					[integerReceiver := objectMemory integerValueOf: integerReceiver.
					 integerArgument := objectMemory integerValueOf: integerArgument].
				self pop: 2 thenPushBool: integerReceiver <= integerArgument]
		ifFalse: [self primitiveFail]! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'eem (auto pragmas 12/08) 11/25/2008 12:50'!
primitiveLessOrEqualLargeIntegers
	"Primitive comparison operations for large integers in 64 bit range"
	| integerRcvr integerArg |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).

	self successful ifTrue:
		[self pop: 2 thenPushBool: integerRcvr <= integerArg]! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 3/11/2015 08:55'!
primitiveLessThan
	| integerReceiver integerArgument |
	integerArgument := self stackTop.
	integerReceiver := self stackValue: 1.
	(objectMemory areIntegers: integerReceiver and: integerArgument)
		ifTrue: [self cCode: '' inSmalltalk:
					[integerReceiver := objectMemory integerValueOf: integerReceiver.
					 integerArgument := objectMemory integerValueOf: integerArgument].
				self pop: 2 thenPushBool: integerReceiver < integerArgument]
		ifFalse: [self primitiveFail]! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'eem (auto pragmas 12/08) 11/25/2008 12:50'!
primitiveLessThanLargeIntegers
	"Primitive comparison operations for large integers in 64 bit range"
	| integerRcvr integerArg |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).

	self successful ifTrue:
		[self pop: 2 thenPushBool: integerRcvr < integerArg]! !

!InterpreterPrimitives methodsFor: 'plugin primitives' stamp: 'eem 7/25/2021 15:40'!
primitiveListBuiltinModule
	"Primitive. Return the n-th builtin module name."
	| index |
	self methodArgumentCount = 1 ifFalse:
		[^self primitiveFail].
	index := self stackIntegerValue: 0.
	(self failed or: [index <= 0]) ifTrue:
		[^self primitiveFail].
	(self ioListBuiltinModule: index)
		ifNil: [self methodReturnValue: objectMemory nilObject]
		ifNotNil: [:moduleName| self methodReturnString: moduleName]! !

!InterpreterPrimitives methodsFor: 'plugin primitives' stamp: 'eem 7/25/2021 15:30'!
primitiveListExternalModule
	"Primitive. Answer the n-th loaded external module name."
	| index |
	self methodArgumentCount = 1 ifFalse:
		[^self primitiveFail].
	index := self stackIntegerValue: 0.
	(self failed or: [index <= 0]) ifTrue:
		[^self primitiveFail].
	(self ioListLoadedModule: index)
		ifNil: [self methodReturnValue: objectMemory nilObject]
		ifNotNil: [:moduleName| self methodReturnString: moduleName]! !

!InterpreterPrimitives methodsFor: 'image segment in/out' stamp: 'eem 05/07/2020 16:13'!
primitiveLoadImageSegment
	"This primitive is called from Smalltalk as...
		<imageSegment> loadSegmentFrom: aWordArray outPointers: anArray.

	 This primitive will load a binary image segment created by primitiveStoreImageSegment.
	 It expects the outPointer array to be of the proper size, and the wordArray to be well
	 formed.  It will return as its value the original array of roots, and the erstwhile
	 segmentWordArray will have been truncated to a size of one word, i.e. retaining the version
	 stamp.  If this primitive should fail, the segmentWordArray will, sadly, have been reduced to
	 an unrecognizable and unusable jumble.  But what more could you have done with it anyway?
	 [How about saving it so the system functions as primitives are intended?  eem 5/9/2017 16:31]

	 In Spur, if the primitive succeeds, the segmentWordArray is also becomed into the array of loaded
	 objects, to allow fixing up of loaded objects directly without nextObject, which Spur doesn't support."

	| outPointerArray segmentWordArray result |

	outPointerArray := self stackTop.
	segmentWordArray := self stackValue: 1.

	"Essential type checks"
	((objectMemory isArray: outPointerArray)		"Must be indexable pointers"
	 and: [objectMemory isWords: segmentWordArray])	"Must be indexable words"
		ifFalse: [^self primitiveFail].

	"the engine returns the roots array which was first in the segment, or an error code on failure."
	result := objectMemory loadImageSegmentFrom: segmentWordArray outPointers: outPointerArray.
	(self oop: result isGreaterThan: segmentWordArray)
		ifTrue: [self pop: 3 thenPush: result]
		ifFalse: [self primitiveFailFor: result]! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 5/26/2017 15:10'!
primitiveLocalMicrosecondClock
	"Return the value of the microsecond clock in the local timezone as an integer.
	 This is the number of microseconds since the Smalltalk epoch, 1901/1/1 12:00am.
	 The microsecond clock is at least 60 bits wide which means it'll get to around August
	 38435 before it wraps around.  Be sure to put it on your calendar.  This primitive
	 accesses the time as answered by the OS."

	self pop: 1 thenPush: (self positive64BitIntegerFor: self ioLocalMicrosecondsNow)! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2017 14:47'!
primitiveLogN
	"Natural log.
	 N.B. IMO we should be able to assume the receiver is a float because this primitive is specific to floats. eem 2/13/2017"
	| rcvr |
	<var: #rcvr type: #double>
	rcvr := self stackFloatValue: 0.
	self successful ifTrue:
		[self stackTopPut: (objectMemory floatObjectOf:
								(self cCode: [rcvr log]
									inSmalltalk: [rcvr = rcvr
													ifTrue: [rcvr ln]
													ifFalse: [Float nan]]))]! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 6/29/2014 11:41'!
primitiveLowSpaceSemaphore
	"Register the low-space semaphore. If the argument is not a 
	 Semaphore, unregister the current low-space Semaphore."
	| arg |
	arg := self stackTop.
	(arg = objectMemory nilObject
	 or: [objectMemory isSemaphoreOop: arg])
		ifTrue:
			[objectMemory splObj: TheLowSpaceSemaphore put: arg.
			 self pop: 1]
		ifFalse:
			[self primitiveFailFor: PrimErrBadArgument]! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 6/5/2017 10:35'!
primitiveMakePoint
	<inline: false>
	| rcvr arg pt |
	rcvr := self stackValue: 1.
	arg := self stackTop.
	(objectMemory isFloatOrInt: arg) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	pt := objectMemory eeInstantiateSmallClass: (objectMemory splObj: ClassPoint) numSlots: YIndex + 1.
	objectMemory "No need to check since new object is always new."
		storePointerUnchecked: XIndex ofObject: pt withValue: rcvr;
		storePointerUnchecked: YIndex ofObject: pt withValue: arg.
	self pop: 2 thenPush: pt! !

!InterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 8/27/2010 10:27'!
primitiveMarkHandlerMethod
	"Primitive. Mark the method for exception handling. The primitive must fail after marking the context so that the regular code is run."
	<inline: false>
	self primitiveFail! !

!InterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 1/25/2009 19:50'!
primitiveMarkUnwindMethod
	"Primitive. Mark the method for exception unwinding. The primitive must fail after marking the context so that the regular code is run.  It must also *not* allow a context switch."
	<inline: false>
	^self primitiveFail! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 7/19/2010 12:48'!
primitiveMaxIdentityHash
	self pop: 1 thenPushInteger: objectMemory maxIdentityHash! !

!InterpreterPrimitives methodsFor: 'plugin primitive support' stamp: 'ar 11/28/1999 17:42'!
primitiveMethod
	"Return the method an external primitive was defined in"
	^newMethod! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
primitiveMillisecondClock
	"Return the value of the millisecond clock as an integer. Note that the millisecond clock wraps around periodically. On some platforms it can wrap daily. The range is limited to SmallInteger maxVal / 2 to allow delays of up to that length without overflowing a SmallInteger."

	self pop: 1 thenPush: (objectMemory integerObjectOf: (self ioMSecs bitAnd: MillisecondClockMask)).
! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 2/28/2013 15:10'!
primitiveMillisecondClockMask
	"Provide access to the millisecond clock mask to support calculation
	of durations based on the millisecond clock value."

	<export: true>
	self pop: 1 thenPush: (objectMemory integerObjectOf: MillisecondClockMask)
! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'tpr 3/15/2004 20:27'!
primitiveMod
	| mod |
	mod := self doPrimitiveMod: (self stackValue: 1) by: (self stackTop).
	self pop2AndPushIntegerIfOK: mod! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'dtl 9/2/2012 10:07'!
primitiveModLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| result oopResult a aIsNegative b bIsNegative oopArg oopRcvr |
	<export: true>
	<var: 'a' type: 'usqLong'>
	<var: 'b' type: 'usqLong'>
	<var: 'result' type: 'usqLong'>

	oopArg := self stackValue: 0.
	oopRcvr := self stackValue: 1.
	aIsNegative := self isNegativeIntegerValueOf: oopRcvr.
	bIsNegative := self isNegativeIntegerValueOf: oopArg.
	a := self magnitude64BitValueOf: oopRcvr.
	b := self magnitude64BitValueOf: oopArg.
	b = 0 ifTrue:[self primitiveFail].
	self successful ifFalse:[^nil].

	result := a \\ b.

	"Handle remainder of same sign as argument"
	result = 0
		ifFalse: [bIsNegative = aIsNegative
			ifFalse: [result := b - result]].

	oopResult := self magnitude64BitIntegerFor: result neg: bIsNegative.
	self successful ifTrue: [self pop: 2 thenPush: oopResult].
! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 3/5/2021 12:26'!
primitiveMouseButtons
	"Obsolete on virtually all platforms; old style input polling code.
	Return the mouse button state. The low three bits encode the state of the <red><yellow><blue> mouse buttons. The next four bits encode the Smalltalk modifier bits <cmd><option><ctrl><shift>."
	<option: #Legacy>
	| buttonWord |
	self pop: 1.
	buttonWord := self ioGetButtonState.
	self pushInteger: buttonWord.! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 3/5/2021 12:27'!
primitiveMousePoint
	"Obsolete on virtually all platforms; old style input polling code.
	Return a Point indicating current position of the mouse. Note that mouse coordinates may be negative if the mouse moves above or to the left of the top-left corner of the Smalltalk window."
	<option: #Legacy>
	| pointWord x y |
	pointWord := self ioMousePoint.
	x := self signExtend16: ((pointWord >> 16) bitAnd: 16rFFFF).
	y := self signExtend16: (pointWord bitAnd: 16rFFFF).
	self pop: 1 thenPush: (self makePointwithxValue: x  yValue: y)! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 4/26/2016 08:44'!
primitiveMultiply
	| integerRcvr integerArg integerResult overflow |
	integerRcvr := self stackIntegerValue: 1.
	integerArg := self stackIntegerValue: 0.
	self successful ifTrue:
		[overflow := integerRcvr > 0
					ifTrue:  [integerArg > 0
						ifTrue: [integerRcvr > (objectMemory maxSmallInteger / integerArg)]
						ifFalse: [integerArg < (objectMemory minSmallInteger / integerRcvr)]]
					ifFalse: [integerArg > 0
						ifTrue: [integerRcvr < (objectMemory minSmallInteger / integerArg)]
						ifFalse: [(integerRcvr < 0) and: [integerArg < (objectMemory maxSmallInteger / integerRcvr)]]].
		overflow
			ifTrue: [self primitiveFail]
			ifFalse:
				[integerResult := integerRcvr * integerArg.
				self pop: 2 thenPush: (objectMemory integerObjectOf: integerResult)]]! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'dtl 9/2/2012 10:07'!
primitiveMultiplyLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| a b result oopResult aIsNegative bIsNegative oopArg oopRcvr |
	<export: true>
	<var: 'a' type: 'usqLong'>
	<var: 'b' type: 'usqLong'>
	<var: 'result' type: 'usqLong'>

	oopArg := self stackValue: 0.
	oopRcvr := self stackValue: 1.
	aIsNegative := self isNegativeIntegerValueOf: oopRcvr.
	bIsNegative := self isNegativeIntegerValueOf: oopArg.
	a := self magnitude64BitValueOf: oopRcvr.
	b := self magnitude64BitValueOf: oopArg.
	self successful ifFalse:[^nil].
	
	"check for overflow"
	(a > 1 and: [b > 1 and: [a > (16rFFFFFFFFFFFFFFFF / b)]])
		ifTrue: [self primitiveFail. ^nil].

	result := a * b.
	oopResult := self magnitude64BitIntegerFor: result neg: aIsNegative ~= bIsNegative.

	self successful ifTrue:[self pop: 2 thenPush: oopResult].
! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 7/18/2015 08:05'!
primitiveNew
	NewspeakVM ifTrue: "For the mirror prims check that the class obj is actually a valid class."
		[(argumentCount < 1
		  or: [self objCouldBeClassObj: self stackTop]) ifFalse:
			[^self primitiveFailFor: PrimErrBadArgument]].
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			["Allocate a new fixed-size instance.  Fail if the allocation would leave
			  less than lowSpaceThreshold bytes free. This *will not* cause a GC :-)"
			(objectMemory instantiateClass: self stackTop)
				ifNotNil: [:obj| self pop: argumentCount + 1 thenPush: obj]
				ifNil: [self primitiveFailFor: ((objectMemory isFixedSizePointerFormat: (objectMemory instSpecOfClass: self stackTop))
											ifTrue: [PrimErrNoMemory]
											ifFalse: [PrimErrBadReceiver])]]
		ifFalse:
			["Allocate a new fixed-size instance. Fail if the allocation would leave
			  less than lowSpaceThreshold bytes free. May cause a GC."
			| spaceOkay |
			"The following may cause GC!! Use var for result to permit inlining."
			spaceOkay := objectMemory
								sufficientSpaceToInstantiate: self stackTop
								indexableSize: 0.
			spaceOkay
				ifTrue:
					[self
						pop: argumentCount + 1
						thenPush: (objectMemory
									instantiateClass: self stackTop
									indexableSize: 0)]
				ifFalse: [self primitiveFailFor: PrimErrNoMemory]]! !

!InterpreterPrimitives methodsFor: 'compiled methods' stamp: 'eem 8/29/2020 19:18'!
primitiveNewMethod
	| header bytecodeCount class size theMethod literalCount |
	header := self stackTop.
	bytecodeCount := self stackValue: 1.
	((objectMemory isIntegerObject: header)
	 and: [(objectMemory isIntegerObject: bytecodeCount)
	 and: [(bytecodeCount := objectMemory integerValueOf: bytecodeCount) >= 0]]) ifFalse:
		[self primitiveFailFor: PrimErrBadArgument.
		 ^self].
	class := self stackValue: 2.
	literalCount := objectMemory literalCountOfMethodHeader: header.
	size := literalCount + LiteralStart * objectMemory bytesPerOop + bytecodeCount.
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[theMethod := objectMemory instantiateCompiledMethodClass: class indexableSize: size.
			 theMethod ifNil:
				[self primitiveFailFor: ((objectMemory isCompiledMethodFormat: (objectMemory instSpecOfClass: class))
										ifTrue: [PrimErrNoMemory]
										ifFalse: [PrimErrBadReceiver]).
				 ^self]]
		ifFalse:
			[theMethod := objectMemory instantiateClass: class indexableSize: size].
	objectMemory storePointerUnchecked: HeaderIndex ofObject: theMethod withValue: header.
	1 to: literalCount do:
		[:i | objectMemory storePointerUnchecked: i ofObject: theMethod withValue: objectMemory nilObject].
	self pop: 3 thenPush: theMethod! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 7/18/2015 08:05'!
primitiveNewWithArg
	"Allocate a new indexable instance. Fail if the allocation would leave less than lowSpaceThreshold bytes free. May cause a GC."
	| size spaceOkay instSpec |
	NewspeakVM ifTrue: "For the mirror prims check that the class obj is actually a valid class."
		[(argumentCount < 2
		  or: [self addressCouldBeClassObj: (self stackValue: 1)]) ifFalse:
			[^self primitiveFailFor: PrimErrBadArgument]].
	size := self positiveMachineIntegerValueOf: self stackTop.
	self successful ifFalse:"positiveMachineIntegerValueOf: succeeds only for non-negative integers."
		[^self primitiveFailFor: PrimErrBadArgument].
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[(objectMemory instantiateClass: (self stackValue: 1) indexableSize: size)
				ifNotNil: [:obj| self pop: argumentCount + 1 thenPush: obj]
				ifNil: [instSpec := objectMemory instSpecOfClass: (self stackValue: 1).
					  self primitiveFailFor: (((objectMemory isIndexableFormat: instSpec)
											and: [(objectMemory isCompiledMethodFormat: instSpec) not])
												ifTrue: [PrimErrNoMemory]
												ifFalse: [PrimErrBadReceiver])]]
		ifFalse:
			[spaceOkay := objectMemory sufficientSpaceToInstantiate: (self stackValue: 1) indexableSize: size.
			 spaceOkay
				ifTrue:
					[self
						pop: argumentCount + 1
						thenPush: (objectMemory instantiateClass: (self stackValue: 1) indexableSize: size)]
				ifFalse:
					[self primitiveFailFor: PrimErrNoMemory]]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 9/4/2013 17:20'!
primitiveNextInstance
	| object subsequentObject |
	object := self stackTop.
	(objectMemory isImmediate: object) ifFalse:
		[subsequentObject := objectMemory instanceAfter: object.
		 subsequentObject ifNotNil:
			[^self pop: argumentCount+1 thenPush: subsequentObject]].
	self primitiveFail! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 7/19/2019 19:16'!
primitiveNextObject
	"Return the object following the receiver in the heap. Return the SmallInteger zero when there are no more objects."

	(objectMemory hasSpurMemoryManagerAPI
	 and: [objectMemory isImmediate: self stackTop]) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	(objectMemory accessibleObjectAfter: self stackTop)
		ifNil: [self pop: argumentCount+1 thenPushInteger: 0]
		ifNotNil: [:instance|
			self assert: (objectMemory isInMemory: instance).
			self pop: argumentCount+1 thenPush: instance]! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: ''!
primitiveNoop
	"A placeholder for primitives that haven't been implemented or are being withdrawn gradually. Just absorbs any arguments and returns the receiver."

	self pop: argumentCount.  "pop args, leave rcvr on stack"! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 3/6/2015 02:05'!
primitiveNotEqual
	| integerReceiver integerArgument result |
	integerArgument := self stackTop.
	integerReceiver := self stackValue: 1.
	(objectMemory areIntegers: integerReceiver and: integerArgument)
		ifTrue: [self pop: 2 thenPushBool: integerReceiver ~= integerArgument]
		ifFalse:
			[result := objectMemory hasSixtyFourBitImmediates
						ifTrue:
							[(self signedMachineIntegerValueOf: integerReceiver)
								~= (self signedMachineIntegerValueOf: integerArgument)]
						ifFalse:
							[(self positiveMachineIntegerValueOf: integerReceiver)
								~= (self positiveMachineIntegerValueOf: integerArgument)].
			 self successful ifTrue:
				[self pop: 2 thenPushBool: result]]! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'eem (auto pragmas 12/08) 11/25/2008 12:50'!
primitiveNotEqualLargeIntegers
	"Primitive comparison operations for large integers in 64 bit range"
	| integerRcvr integerArg |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).

	self successful ifTrue:
		[self pop: 2 thenPushBool: integerRcvr ~= integerArg]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 3/16/2021 13:24'!
primitiveNotIdentical
	"is the receiver/first argument not the same object as the (last) argument?.
	 pop argumentCount because this can be used as a mirror primitive."
	"is the receiver/first argument the same object as the (last) argument?.
	 pop argumentCount because this can be used as a mirror primitive."
	| thisObject otherObject |
	thisObject := self stackValue: 1.
	otherObject := self stackTop.
	((objectMemory isOopForwarded: otherObject)
	 or: [argumentCount > 1
		 and: [objectMemory isOopForwarded: thisObject]])
		ifTrue:
			[self primitiveFailFor: PrimErrBadArgument]
		ifFalse:
			[self pop: argumentCount + 1 thenPushBool: thisObject ~= otherObject]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 08/08/2014 10:01'!
primitiveObjectAt
"Defined for CompiledMethods only"
	| thisReceiver index |
	index  := self stackIntegerValue: 0.
	self successful ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	thisReceiver := self stackValue: 1.
	(index > 0 and: [index <= ((objectMemory literalCountOf: thisReceiver) + LiteralStart)]) ifFalse:
		[^self primitiveFailFor: PrimErrBadIndex].
	self pop: 2 thenPush: (objectMemory fetchPointer: index - 1 ofObject: thisReceiver)! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 1/20/2016 14:26'!
primitiveObjectAtPut
	"Store a literal into a CompiledMethod at the given index. Defined for CompiledMethods only."
	| thisReceiver index newValue |
	newValue := self stackValue: 0.
	index := self stackValue: 1.
	((objectMemory isNonIntegerObject: index)
	 or: [index = ConstOne and: [(objectMemory isNonIntegerObject: newValue)]]) ifTrue:
		[^self primitiveFailFor: PrimErrBadArgument].
	index := objectMemory integerValueOf: index.
	thisReceiver := self stackValue: 2.
	(objectMemory isObjImmutable: thisReceiver) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].
	(index > 0 and: [index <= ((objectMemory literalCountOf: thisReceiver) + LiteralStart)]) ifFalse:
		[^self primitiveFailFor: PrimErrBadIndex].
	objectMemory storePointer: index - 1 ofObject: thisReceiver withValue: newValue.
	self pop: 3 thenPush: newValue! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 1/4/2017 08:48'!
primitiveObjectPointsTo
	| rcvr thang lastField |
	thang := self stackTop.
	rcvr := self stackValue: 1.
	(objectMemory isIntegerObject: rcvr) ifTrue:
		[^self pop: 2 thenPushBool: false].

	lastField := self lastPointerOf: rcvr.
	objectMemory baseHeaderSize to: lastField by: objectMemory bytesPerOop do:
		[:i |
		(objectMemory longAt: rcvr + i) = thang ifTrue:
			[^self pop: 2 thenPushBool: true]].
	self pop: 2 thenPushBool: false! !

!InterpreterPrimitives methodsFor: 'other primitives' stamp: 'eem 3/27/2014 15:49'!
primitivePathToUsing
	"primitivePathTo: anObject using: stack <Array> followWeak: boolean
	 Answer a path to anObject from the root that does not pass through
	 the current context"
	| err path |
	<export: true>
	self externalWriteBackHeadFramePointers.
	argumentCount >= 2 ifFalse:
		[^self primitiveFailFor: PrimErrBadNumArgs].
	(self stackTop = objectMemory trueObject
	 or: [self stackTop = objectMemory falseObject]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	err := self pathTo: (self stackValue: 2)
				using: (self stackValue: 1)
				followWeak: self stackTop = objectMemory trueObject.
	err ~= 0 ifTrue:
		[^self primitiveFailFor: err].
	path := self self stackValue: 1.
	self pop: argumentCount + 1 thenPush: path! !

!InterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 6/20/2015 10:44'!
primitivePerformInSuperclass
	| lookupClass rcvr currentClass |
	rcvr := self stackValue: 3.
	lookupClass := self stackTop.
	(argumentCount > 3 "e.g. object:perform:withArguments:inClass:"
	 and: [objectMemory isOopForwarded: rcvr]) ifTrue:
		[^self primitiveFail].
	currentClass := objectMemory fetchClassOf: rcvr.
	[currentClass ~= lookupClass] whileTrue:
		[currentClass := self superclassOf: currentClass.
		 currentClass = objectMemory nilObject ifTrue:
			[^self primitiveFailFor: PrimErrBadArgument]].

	self primitiveObject: rcvr "a.k.a. self stackValue: 3"
		perform: (self stackValue: 2)
		withArguments: (self stackValue: 1)
		lookedUpIn: lookupClass "a.k.a. self stackTop"! !

!InterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 2/4/2015 15:01'!
primitivePerformWithArgs
	self primitiveObject: (self stackValue: 2)
		perform: (self stackValue: 1)
		withArguments: self stackTop
		lookedUpIn: nil! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 12/9/2015 12:49'!
primitivePin
	"Pin or unpin the receiver, i.e. make it immobile or mobile, based on the argument.
	 Answer whether the object was already pinned. N.B. pinning does *not* prevent
	 an object from being garbage collected."
	<option: #SpurObjectMemory>
	| obj boolean wasPinned |

	obj := self stackValue: 1.
	((objectMemory isImmediate: obj)
	 or: [objectMemory isForwarded: obj]) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].

	boolean := self stackTop.
	(boolean = objectMemory falseObject
	 or: [boolean = objectMemory trueObject]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].

	(objectMemory isPinned: obj)
		ifTrue:
			[wasPinned := objectMemory trueObject.
			 boolean ~= wasPinned ifTrue:
				[objectMemory setIsPinnedOf: obj to: false]]
		ifFalse:
			[wasPinned := objectMemory falseObject.
			 (boolean ~= wasPinned
			  and: [(objectMemory pinObject: obj) = 0]) ifTrue:
				[^self primitiveFailFor: PrimErrNoMemory]].
	
	self pop: argumentCount + 1 thenPush: wasPinned! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 11/21/2010 11:27'!
primitiveProcessBindToThreadId
	"Simulation only.  Fail."
	<doNotGenerate>
	self primitiveFail! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 11/21/2010 11:27'!
primitiveProcessBoundThreadId
	"Simulation only.  Fail."
	<doNotGenerate>
	self primitiveFail! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 8/3/2021 21:27'!
primitiveProfilePrimitive
	"Primitive. Answer the last primitive method sampled by the profiler."
	<export: true flags: #FastCPrimitiveFlag>
	self methodReturnValue: profileMethod.
	profileMethod := objectMemory nilObject! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 8/4/2021 15:03'!
primitiveProfileSample
	"Primitive. Answer the last sample taken by the profiler, or nil if the profiler isn't active.
	See also primitiveProfileStart."
	<export: true flags: #FastCPrimitiveFlag>
	self methodReturnValue: profileProcess.
	profileProcess := objectMemory nilObject! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 6/29/2014 11:37'!
primitiveProfileSemaphore
	"Primitive. Install the semaphore to be used for profiling, 
	or nil if no semaphore should be used.
	See also primitiveProfileStart."
	| sema |
	<export: true>
	self methodArgumentCount = 1 ifFalse:
		[^self primitiveFailFor: PrimErrBadNumArgs].
	sema := self stackValue: 0.
	(sema = objectMemory nilObject
	 or: [objectMemory isSemaphoreOop: sema]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	profileSemaphore := sema.
	profileProcess := profileMethod := objectMemory nilObject.
	self pop: 1! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 8/4/2021 12:13'!
primitiveProfileStart
	"Primitive. Begin profiling execution every by using the interrupt check-counter instead of a time-based process (which is limited to timing resolution and triggers off the same signal that many of the processes being profiled trigger off leading to consistently wrong results).
	The argument is the number of interrupt checks (method activations) to let go by before taking a sample. The sample is being stored in the profileSample iVar which can be retrieved by executing primitiveProfileSample. When a sample is taken, it signals the semaphore specified in primitiveProfileSemaphore.
	If the argument is less or equal to zero, it disables profiling."
	<export: true flags: #FastCPrimitiveFlag>
	| deltaTicks |
	deltaTicks := self stackValue: 0.
	(objectMemory isIntegerObject: deltaTicks) ifTrue:
		[nextProfileTick := self ioHighResClock + (objectMemory integerValueOf: deltaTicks).
		 ^self methodReturnReceiver].
	^self primitiveFailFor: PrimErrBadArgument! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 6/22/2010 17:04'!
primitiveQuit

	self ioExitWithErrorCode: (argumentCount = 1 ifTrue: [objectMemory integerValueOf: self stackTop] ifFalse: [0])! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 8/7/2014 06:45'!
primitiveQuo
	"Rounds negative results towards zero."
	| integerRcvr integerArg integerResult |
	integerRcvr := self stackIntegerValue: 1.
	integerArg := self stackIntegerValue: 0.
	self success: integerArg ~= 0.
	self successful ifTrue:
		[integerResult := self quot: integerRcvr ient: integerArg].
	self pop2AndPushIntegerIfOK: integerResult! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'dtl 9/2/2012 10:07'!
primitiveQuoLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| result oopResult a aIsNegative b bIsNegative oopArg oopRcvr |
	<export: true>
	<var: 'a' type: 'usqLong'>
	<var: 'b' type: 'usqLong'>
	<var: 'result' type: 'usqLong'>

	oopArg := self stackValue: 0.
	oopRcvr := self stackValue: 1.
	aIsNegative := self isNegativeIntegerValueOf: oopRcvr.
	bIsNegative := self isNegativeIntegerValueOf: oopArg.
	a := self magnitude64BitValueOf: oopRcvr.
	b := self magnitude64BitValueOf: oopArg.
	b = 0 ifTrue:[self primitiveFail].
	self successful ifFalse:[^nil].

	result := a // b.

	oopResult := self magnitude64BitIntegerFor: result neg: bIsNegative ~= aIsNegative.
	self successful ifTrue: [self pop: 2 thenPush: oopResult].! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 10/23/2020 13:07'!
primitiveRelinquishProcessor
	"Relinquish the processor for up to the given number of microseconds. The exact behavior of this primitive is platform dependent."

	| microSecs |
	microSecs := self stackIntegerValue: 0.
	self successful ifTrue:
		"DO NOT allow relinquishing the processor while we are profiling since this
		may skew the time base for our measures (it may reduce processor speed etc).
		Instead we go full speed, therefore measuring the precise time we spend in the
		inner idle loop as a busy loop."
		[nextProfileTick = 0 ifTrue:
			[self ioRelinquishProcessorForMicroseconds: microSecs.
			 "In simulation we allow ioRelinquishProcessorForMicroseconds: to fail so that
			  we can arrange that the simulator responds to input events promply.  This
			  *DOES NOT HAPPEN* in the real vm."
			 self cCode: [] inSmalltalk: [primFailCode ~= 0 ifTrue: [^self]]].
		self pop: 1]  "microSecs; leave rcvr on stack"! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'nice 1/9/2013 02:16'!
primitiveRemLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| result oopResult a aIsNegative b oopArg oopRcvr |
	<export: true>
	<var: 'a' type: 'usqLong'>
	<var: 'b' type: 'usqLong'>
	<var: 'result' type: 'usqLong'>

	oopArg := self stackValue: 0.
	oopRcvr := self stackValue: 1.
	aIsNegative := self isNegativeIntegerValueOf: oopRcvr.
	a := self magnitude64BitValueOf: oopRcvr.
	b := self magnitude64BitValueOf: oopArg.
	b = 0 ifTrue:[self primitiveFail].
	self successful ifFalse:[^nil].

	result := a \\ b.

	oopResult := self magnitude64BitIntegerFor: result neg: aIsNegative.
	self successful ifTrue: [self pop: 2 thenPush: oopResult].! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 5/20/2021 15:53'!
primitiveResume
	"Put this process on the scheduler's lists thus allowing it to proceed next time there is
	 a chance for processes of it's priority level.  It must go to the back of its run queue so
	 as not to preempt any already running processes at this level.  If the process's priority
	 is higher than the current process, preempt the current process."
	| proc ctxt |
	proc := self stackTop.  "rcvr"
	ctxt := objectMemory followField: SuspendedContextIndex ofObject: proc. "written this way to get better Slang inlining"
	(objectMemory isContext: ctxt) ifFalse:
		[^self primitiveFailFor: PrimErrBadReceiver].
	self resume: proc preemptedYieldingIf: preemptionYields

	"Personally I would like to check MyList, which should not be one of the elements of the scheduler lists.
	 But there are awful race conditions in things like should:notTakeMoreThan: that mean we can't.
	 eem 9/27/2010 23:08, updated eem 5/20/2021 15:36. e.g.

	| proc field |
	proc := self stackTop.  ''rcvr''
	''We only have to check for myList being nil.  If it is nil then this is either the active process or
	 a process suspended with primitiveSuspend (and if it is the activeProcess suspendedContext will
	 be nil and the isContext: test will fail).  If it is not nil then either the process is waiting on some
	 semaphore-like list or on one of the scheduler's lists. If it is on some semaphore-like list it should
	 not resume.  If it is on one of the scheduler's lists it is runnable (already resumed).''
	field := objectMemory followField: MyListIndex ofObject: proc.
	objectMemory nilObject = field ifFalse:
		[^self primitiveFailFor: PrimErrInappropriate].
	field := objectMemory followField: SuspendedContextIndex ofObject: proc.
	(objectMemory isContext: field) ifFalse:
		[^self primitiveFailFor: PrimErrBadReceiver].
	self resume: proc preemptedYieldingIf: preemptionYields"! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 10/13/2013 16:18'!
primitiveRootTable
	"Primitive. Answer a copy (snapshot) element of the root table.
	The primitive can cause GC itself and if so the return value may
	be inaccurate - in this case one should guard the read operation
	by looking at the gc counter statistics."
	<option: #SqueakV3ObjectMemory>
	self pop: argumentCount + 1 thenPush: objectMemory rootTableObject! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 10/13/2013 16:18'!
primitiveRootTableAt
	"Primitive. Answer the nth element of the root table.
	This primitive avoids the creation of an extra array;
	it is intended for enumerations of the form:
		index := 1.
		[root := Smalltalk rootTableAt: index.
		root == nil] whileFalse:[index := index + 1].
	"
	<export: true>
	<option: #SqueakV3ObjectMemory>
	| index |
	index := self stackIntegerValue: 0.
	self success: (index > 0 and:[index <= objectMemory rootTableCount]).
	self successful ifTrue:[
		self pop: argumentCount + 1.
		self push: (objectMemory rootTable at: index).
	]! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 7/19/2017 10:54'!
primitiveScanCharacters
	"The character scanner primitive."
	| kernDelta stops sourceString scanStopIndex scanStartIndex rcvr scanDestX scanLastIndex scanXTable
	  scanMap maxGlyph ascii stopReason glyphIndex sourceX sourceX2 nextDestX scanRightX nilOop |

	self methodArgumentCount = 6
		ifFalse: [^ self primitiveFail].

	"Load the receiver and arguments"
	kernDelta := self stackIntegerValue: 0.
	stops := self stackValue: 1.
	scanRightX := self stackIntegerValue: 2.
	sourceString := self stackValue: 3.
	scanStopIndex := self stackIntegerValue: 4.
	scanStartIndex := self stackIntegerValue: 5.
	rcvr := self stackValue: 6.
	self successful ifFalse: [^ nil].
	
	"check argument type and range and rcvr"
	((objectMemory isArray: stops)
	 and: [(objectMemory slotSizeOf: stops) >= 258
	 and: [(objectMemory isBytes: sourceString)
	 and: [scanStartIndex > 0
	 and: [scanStopIndex > 0
	 and: [scanStopIndex <= (objectMemory byteSizeOf: sourceString)
	 and: [(objectMemory isPointers: rcvr)
	 and: [(objectMemory slotSizeOf: rcvr) >= 4]]]]]]])
		ifFalse: [^ self primitiveFail].

	"Check required rcvr instVars"
	scanDestX := self fetchInteger: 0 ofObject: rcvr.
	scanLastIndex := self fetchInteger: 1 ofObject: rcvr.
	scanXTable := objectMemory fetchPointer: 2 ofObject: rcvr.
	scanMap := objectMemory fetchPointer: 3 ofObject: rcvr.
	((objectMemory isArray: scanXTable)
	 and: [(objectMemory isArray: scanMap)
	 and: [(objectMemory slotSizeOf: scanMap) = 256
	 and: [self successful "for the fetchInteger:ofObject:'s above"]]]) ifFalse:
		[^ self primitiveFail].
	maxGlyph := (objectMemory slotSizeOf: scanXTable) - 2.

	"Okay, here we go. We have eliminated nearly all failure 
	conditions, to optimize the inner fetches."
	scanLastIndex := scanStartIndex.
	nilOop := objectMemory nilObject.
	[scanLastIndex <= scanStopIndex]
		whileTrue: [
			"Known to be okay since scanStartIndex > 0 and scanStopIndex <= sourceString size"
			ascii := objectMemory fetchByte: scanLastIndex - 1 ofObject: sourceString.
			"Known to be okay since stops size >= 258"
			(stopReason := objectMemory fetchPointer: ascii ofObject: stops) = nilOop
				ifFalse: ["Store everything back and get out of here since some stop conditionn needs to be checked"
					(objectMemory isIntegerValue: scanDestX) ifFalse: [^ self primitiveFail].
					self storeInteger: 0 ofObject: rcvr withValue: scanDestX.
					self storeInteger: 1 ofObject: rcvr withValue: scanLastIndex.
					self pop: 7 "args+rcvr" thenPush: stopReason.
					^nil].
			"Known to be okay since scanMap size = 256"
			glyphIndex := self fetchInteger: ascii ofObject: scanMap.
			"fail if the glyphIndex is out of range"
			(self failed or: [glyphIndex < 0 	or: [glyphIndex > maxGlyph]]) ifTrue: [^ self primitiveFail].
			sourceX := self fetchInteger: glyphIndex ofObject: scanXTable.
			sourceX2 := self fetchInteger: glyphIndex + 1 ofObject: scanXTable.
			"Above may fail if non-integer entries in scanXTable"
			self failed ifTrue: [^ nil].
			nextDestX := scanDestX + sourceX2 - sourceX.
			nextDestX > scanRightX ifTrue:
				["Store everything back and get out of here since we got to the right edge"
				(objectMemory isIntegerValue: scanDestX) ifFalse: [^ self primitiveFail].
				self storeInteger: 0 ofObject: rcvr withValue: scanDestX.
				self storeInteger: 1 ofObject: rcvr withValue: scanLastIndex.
				self pop: 7 "args+rcvr" thenPush: (objectMemory fetchPointer: CrossedX - 1 ofObject: stops).
				^nil].
			scanDestX := nextDestX + kernDelta.
			scanLastIndex := scanLastIndex + 1].
	(objectMemory isIntegerValue: scanDestX) ifFalse: [^ self primitiveFail].
	self storeInteger: 0 ofObject: rcvr withValue: scanDestX.
	self storeInteger: 1 ofObject: rcvr withValue: scanStopIndex.
	self pop: 7 "args+rcvr" thenPush: (objectMemory fetchPointer: EndOfRun - 1 ofObject: stops)! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'tpr (auto pragmas 12/08) 5/13/2005 10:20'!
primitiveScreenDepth
	"Return a SmallInteger indicating the current depth of the OS screen. Negative values are used to imply LSB type pixel format an there is some support in the VM for handling either MSB or LSB"
	| depth |
	<export: true>
	depth := self ioScreenDepth.
	self failed ifTrue:[^self primitiveFail].
	self pop: 1 thenPushInteger: depth.! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'topa 7/10/2016 23:44'!
primitiveScreenScaleFactor
	"Answer a float indicating the current scale factor for pixels of the Smalltalk window."
	| factor |
	<export: true>
	<var: #factor type: #double>
	factor := self ioScreenScaleFactor.
	self failed ifTrue:[^self primitiveFail].
	self pop: 1; pushFloat: factor.! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 2/23/2015 18:08'!
primitiveScreenSize
	"Answer a point indicating the current size of the Smalltalk window.
	 Currently there is a limit of 65535 in each direction because the
	 point is encoded into a single 32bit value in the image header.
	 This might well become a problem one day"
	| pointWord |
	pointWord := self ioScreenSize.
	self pop: 1
		thenPush: (self makePointwithxValue: (pointWord >> 16 bitAnd: 65535)
						yValue: (pointWord bitAnd: 65535))! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 3/11/2014 11:02'!
primitiveSecondsClock
	"Return the number of seconds since January 1, 1901 as an integer."

	self pop: 1 thenPush: (self positive32BitIntegerFor: self ioSecondsNow)! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 10/7/2020 18:15'!
primitiveSetDisplayMode
	"Ask the GUI to set the requested display mode.
	 See DisplayScreen class depth:width:height:fullscreen:"
	| fsFlag h w d okay |
	fsFlag := self booleanValueOf: self stackTop.
	h := self stackIntegerValue: 1.
	w := self stackIntegerValue: 2.
	d := self stackIntegerValue: 3.
	self successful ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	okay := self ioSetDisplayMode: w _: h _: d _: fsFlag.
	self successful ifFalse:
		[^self primitiveFailFor: PrimErrOperationFailed].
	self methodReturnBool: okay! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 10/18/2020 15:56'!
primitiveSetFullScreen
	"On platforms that support it, set full-screen mode to the value of the boolean argument."

	| argOop |
	argOop := self stackTop.
	argOop = objectMemory trueObject
		ifTrue:
			[self ioSetFullScreen: true]
		ifFalse:
			[argOop = objectMemory falseObject
				ifTrue: [self ioSetFullScreen: false]
				ifFalse: [self primitiveFail]].
	self successful ifTrue:
		[self setFullScreenFlag: argOop = objectMemory trueObject.
		 self methodReturnReceiver]! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 10/14/2013 20:19'!
primitiveSetGCBiasToGrow
	"Primitive. Indicate if the GC logic should have bias to grow"
	| flag |
	<export: true>
	<option: #SqueakV3ObjectMemory> "for now..."
	flag := self stackIntegerValue: 0.
	self successful ifTrue:[
		objectMemory gcBiasToGrow: flag.
		self pop: argumentCount.
	]! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 10/14/2013 20:17'!
primitiveSetGCBiasToGrowGCLimit
	"Primitive. If the GC logic has  bias to grow, set growth limit"
	| value |
	<export: true>
	<option: #SqueakV3ObjectMemory> "for now..."
	value := self stackIntegerValue: 0.
	self successful ifTrue:
		[objectMemory gcBiasToGrowGCLimit: value.
		 objectMemory gcBiasToGrowThreshold: objectMemory youngStart - objectMemory startOfMemory asInteger.
		 self pop: argumentCount]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 7/21/2010 17:11'!
primitiveSetIdentityHash
	| hash oldHash thisReceiver |
	hash := self stackIntegerValue: 0.
	thisReceiver := self stackObjectValue: 1.
	self successful ifTrue:
		[oldHash := objectMemory hashBitsOf: thisReceiver.
		 objectMemory setHashBitsOf: thisReceiver to: hash.
		 self pop: argumentCount + 1 thenPushInteger: oldHash]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'cb 4/19/2016 09:35'!
primitiveSetImmutability
	<option: #IMMUTABILITY>
	| rcvr wasImmutable |
	 rcvr := self stackValue: 1.
	 (objectMemory isImmediate: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	 wasImmutable := objectMemory isObjImmutable: rcvr.
	 self stackTop = objectMemory trueObject
		ifTrue:
			[(self canBeImmutable: rcvr) ifFalse:
				[^self primitiveFailFor: PrimErrInappropriate].
			  objectMemory setIsImmutableOf: rcvr to: true]
		ifFalse: [
	self stackTop = objectMemory falseObject
		ifTrue: [objectMemory setIsImmutableOf: rcvr to: false]
	 	ifFalse:
			[^self primitiveFailFor: PrimErrBadArgument]].
	 self pop: argumentCount + 1 thenPushBool: wasImmutable! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 9/29/2010 18:38'!
primitiveSetInterruptKey
	"Set the user interrupt keycode. The keycode is an integer whose encoding is described in the comment for primitiveKbdNext."

	| keycode |
	keycode := self stackIntegerValue: 0.
	self successful ifTrue:
		[self setInterruptKeycode: keycode.
		 self pop: argumentCount]! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 6/10/2010 16:13'!
primitiveSetLogDirectory
	"Primitive. Set the VM's log directory"
	| stringOop sz |
	<export: true>
	stringOop := self stackTop.
	(objectMemory isBytes: stringOop) ifFalse:[^self success: false].
	sz := objectMemory byteSizeOf: stringOop.
	self ioSetLogDirectory: (objectMemory firstIndexableField: stringOop) OfSize: sz.
	self successful ifTrue:[self pop: self methodArgumentCount]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 9/5/2021 18:33'!
primitiveSetOrHasIdentityHash
	| hash oldHash thisReceiver isReceiverAClass |
	argumentCount = 0 ifTrue:
		[| hasHash |
		 hasHash := (objectMemory isNonImmediate: self stackTop)
						and: [objectMemory hasIdentityHash: self stackTop].
		 ^self methodReturnBool: hasHash].
	isReceiverAClass := false.
	argumentCount = 2 
		ifTrue: 
			[| lastArg |
			 lastArg := self stackTop.
			 (objectMemory isIntegerObject: lastArg)
				ifTrue: "e.g. Symbol primitiveSetIdentityHashOf: aSymbol to: hash"
					[hash := lastArg.
					 thisReceiver := self stackValue: 1]
				ifFalse: "anObject primitiveSetIdentityHash: hashValue isBehavior: boolean"
					[thisReceiver := self stackValue: 2.
					 hash := self stackValue: 1.
					 lastArg = objectMemory trueObject
						ifTrue: [isReceiverAClass := true]
						ifFalse:
							[lastArg = objectMemory falseObject ifFalse:
								[^self primitiveFailFor: PrimErrBadArgument]]]]
		ifFalse: "anObject primitiveSetIdentityHashTo: hash"
			[thisReceiver := self stackValue: 1.
			 hash := self stackTop].
	((objectMemory isIntegerObject: hash)
	and: [((hash := objectMemory integerValueOf: hash) bitAnd: objectMemory maxIdentityHash) = hash]) ifFalse:
		 [^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory isNonImmediate: thisReceiver) ifFalse:
		[^self primitiveFailFor: (thisReceiver = (self stackValue: argumentCount)
									ifTrue: [PrimErrBadReceiver]
									ifFalse: [PrimErrBadArgument])].
	oldHash := objectMemory hashBitsOf: thisReceiver.
	objectMemory setHashBitsOf: thisReceiver to: hash.
	(isReceiverAClass and: [objectMemory hasSpurMemoryManagerAPI]) ifTrue:
		[objectMemory classAtIndex: hash put: thisReceiver.
		"this figures out if the index is ambiguous and fixes all the instances if needed"
		objectMemory allInstancesOf: thisReceiver].
	 self methodReturnInteger: oldHash! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 6/10/2010 16:13'!
primitiveSetWindowLabel
	"Primitive. Set the OS window's label"
	| labelOop sz |
	<export: true>
	labelOop := self stackTop.
	(objectMemory isBytes: labelOop) ifFalse:[^self success: false].
	sz := objectMemory byteSizeOf: labelOop.
	self ioSetWindowLabel: (objectMemory firstIndexableField: labelOop) OfSize: sz.
	self successful ifTrue:[self pop: self methodArgumentCount]! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'bgf 3/14/2009 23:16'!
primitiveSetWindowSize
	"Primitive. Answer the OS window's size (extent)"
	| w h |
	<export: true>
	h := self stackIntegerValue: 0.
	w := self stackIntegerValue: 1.
	self successful ifTrue:[
		self ioSetWindowWidth: w Height: h.
		self pop: self methodArgumentCount.
	]! !

!InterpreterPrimitives methodsFor: 'sound primitives' stamp: 'eem 4/28/2015 14:46'!
primitiveShortAt
	"Treat the receiver, which can be indexible by either bytes or words, as
	 an array of signed 16-bit values. Answer the contents of the given index.
	 Note that the index specifies the i-th 16-bit entry, not the i-th byte or word."

	| index rcvr value |
	index := self stackTop.
	(objectMemory isIntegerObject: index) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	rcvr := self stackValue: 1.
	(objectMemory isWordsOrBytes: rcvr) ifFalse:
		[^self primitiveFailFor: PrimErrInappropriate].
	index := objectMemory integerValueOf: index.
	((index >= 1) and: [index <= (objectMemory num16BitUnitsOf: rcvr)]) ifFalse:
		[^self primitiveFailFor: PrimErrBadIndex].
	value := objectMemory fetchShort16: index - 1 ofObject: rcvr.
	self cCode: []
		inSmalltalk: [value > 32767 ifTrue: [value := value - 65536]].
	self pop: 2 thenPushInteger: value! !

!InterpreterPrimitives methodsFor: 'sound primitives' stamp: 'eem 1/20/2016 14:30'!
primitiveShortAtPut
	"Treat the receiver, which can be indexible by either bytes or words, as an array
	 of signed 16-bit values. Set the contents of the given index to the given value.
	 Note that the index specifies the i-th 16-bit entry, not the i-th byte or word."

	| index rcvr value |
	value := self stackTop.
	index := self stackValue: 1.
	((objectMemory isIntegerObject: value)
	 and: [(objectMemory isIntegerObject: index)
	 and: [value := objectMemory integerValueOf: value.
		  (value >= -32768) and: [value <= 32767]]]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	rcvr := self stackValue: 2.
	(objectMemory isWordsOrBytes: rcvr) ifFalse:
		[^self primitiveFailFor: PrimErrInappropriate].
	(objectMemory isObjImmutable: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].
	index := objectMemory integerValueOf: index.
	(index >= 1 and: [index <= (objectMemory num16BitUnitsOf: rcvr)]) ifFalse:
		[^self primitiveFailFor: PrimErrBadIndex].
	objectMemory storeShort16: index - 1 ofObject: rcvr withValue: value.
	self pop: 3 thenPush: (objectMemory integerObjectOf: value)! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 5/12/2017 11:08'!
primitiveShowDisplayRect
	"Force the given rectangular section of the Display to be 
	copied to the screen."
	| bottom top right left |
	bottom := self stackIntegerValue: 0.
	top := self stackIntegerValue: 1.
	right := self stackIntegerValue: 2.
	left := self stackIntegerValue: 3.
	self updateDisplayLeft: left Top: top Right: right Bottom: bottom.
	self successful ifTrue:
		[self ioForceDisplayUpdate.
		 self pop: 4]! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 5/26/2009 10:48'!
primitiveSignal
	"Synchronously signal the semaphore.
	 This may change the active process as a result."
	self synchronousSignal: self stackTop! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 3/5/2015 16:08'!
primitiveSignalAtBytesLeft
	"Set the low-water mark for free space. When the free space falls
	 below this level, the new and new: primitives fail and system attempts
	 to allocate space (e.g., to create a method context) cause the low-space
	 semaphore (if one is registered) to be signalled."
	| bytes |
	bytes := self stackTop.
	((objectMemory isIntegerObject: bytes)
	 and: [(bytes := objectMemory integerValueOf: bytes) >= 0])
		ifTrue: [objectMemory lowSpaceThreshold: bytes. self pop: 1]
		ifFalse: [self primitiveFailFor: PrimErrBadArgument]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2017 14:52'!
primitiveSine
	"N.B. IMO we should be able to assume the receiver is a float because this primitive is specific to floats. eem 2/13/2017"
	| rcvr |
	<var: #rcvr type: #double>
	rcvr := self stackFloatValue: 0.
	self successful ifTrue:
		[self stackTopPut: (objectMemory floatObjectOf:
								(self cCode: [rcvr sin]
									inSmalltalk: [rcvr = rcvr
													ifTrue: [rcvr sin]
													ifFalse: [Float nan]]))]! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 5/4/2021 14:36'!
primitiveSize
	<accessorDepth: 0>
	| rcvr hdr fmt fixedFields totalLength |
	rcvr := self stackTop.
	((objectMemory isImmediate: rcvr) "Integers are not indexable"
	 or: [hdr := objectMemory baseHeader: rcvr.
		(fmt := objectMemory formatOfHeader: hdr) < objectMemory arrayFormat "This is not an indexable object"
	 or: [objectMemory isForwarderFormat: fmt]]) "And if Spur, this is a forwarder."
		ifTrue:
			[^self primitiveFailFor: PrimErrBadReceiver].
	(fmt = objectMemory indexablePointersFormat
	 and: [objectMemory isContextHeader: hdr]) ifTrue:
		[^self primitiveContextSize].
	totalLength := objectMemory lengthOf: rcvr baseHeader: hdr format: fmt.
	fixedFields := objectMemory fixedFieldsOf: rcvr format: fmt length: totalLength.
	self methodReturnInteger: totalLength - fixedFields! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 6/27/2014 15:59'!
primitiveSizeInBytes
	<option: #SpurObjectMemory>
	| byteSize |
	byteSize := objectMemory totalByteSizeOf: self stackTop.
	self pop: argumentCount + 1
		 thenPush: (self positive64BitIntegerFor: byteSize)! !

!InterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 7/18/2015 07:36'!
primitiveSizeInBytesOfInstance
	"Answer the byte size of an instance of the receiver.  If num args > 0
	 then the last argument is a variable size and the size answered is the
	 size of an instance of the receiver with that many indexable elements."
	<option: #SpurObjectMemory>
	| byteSize err |
	NewspeakVM
		ifTrue: "Support VMMirror>>byteSizeOfInstanceOf:WithIndexableVariables:"
			[argumentCount > 2 ifTrue:
				[^self primitiveFailFor: PrimErrBadNumArgs]]
		ifFalse:
			[argumentCount > 1 ifTrue:
				[^self primitiveFailFor: PrimErrBadNumArgs]].
	err := -1.
	argumentCount >= 1 ifTrue:
		[(objectMemory isIntegerObject: self stackTop) ifFalse:
			[^self primitiveFailFor: PrimErrBadArgument].
		 byteSize := objectMemory
						byteSizeOfInstanceOf: (self stackValue: 1)
						withIndexableSlots: (objectMemory integerValueOf: self stackTop)
						errInto: [:code| err := code].
		 err >= 0 ifTrue:
			[^self primitiveFailFor: err].
		 ^self pop: argumentCount + 1 thenPush: (self positive64BitIntegerFor: byteSize)].
	byteSize := objectMemory
						byteSizeOfInstanceOf: (self stackValue: 0)
						errInto: [:code| err := code].
	err >= 0 ifTrue:
		[^self primitiveFailFor: err].
	self pop: 1 thenPushInteger: byteSize! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'nice 11/14/2016 22:34'!
primitiveSlotAt
	"Answer a slot in an object.  This numbers all slots from 1, ignoring the distinction between
	 named and indexed inst vars.  In objects with both named and indexed inst vars, the named
	 inst vars precede the indexed ones.  In non-object indexed objects (objects that contain
	 bits, not object references) this primitive answers the raw integral value at each slot. 
	 e.g. for Strings it answers the character code, not the Character object at each slot."
	| index rcvr fmt numSlots |
	index := self stackTop.
	rcvr := self stackValue: 1.
	(objectMemory isIntegerObject: index) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory isImmediate: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	fmt := objectMemory formatOf: rcvr.
	index := (objectMemory integerValueOf: index) - 1.

	fmt <= objectMemory lastPointerFormat ifTrue:
		[numSlots := objectMemory numSlotsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[self pop: argumentCount + 1 thenPush: (objectMemory fetchPointer: index ofObject: rcvr).
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	fmt >= objectMemory firstByteFormat ifTrue:
		[fmt >= objectMemory firstCompiledMethodFormat ifTrue:
			[^self primitiveFailFor: PrimErrUnsupported].
		 numSlots := objectMemory numBytesOfBytes: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[self pop: argumentCount + 1 thenPushInteger: (objectMemory fetchByte: index ofObject: rcvr).
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	(objectMemory hasSpurMemoryManagerAPI
	 and: [fmt >= objectMemory firstShortFormat]) ifTrue:
		[numSlots := objectMemory num16BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[self pop: argumentCount + 1 thenPushInteger: (objectMemory fetchUnsignedShort16: index ofObject: rcvr).
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	fmt = objectMemory sixtyFourBitIndexableFormat ifTrue:
		[numSlots := objectMemory num64BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[self pop: argumentCount + 1
				thenPush: (self positive64BitIntegerFor: (objectMemory fetchLong64: index ofObject: rcvr)).
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	fmt >= objectMemory firstLongFormat ifTrue:
		[numSlots := objectMemory num32BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[self pop: argumentCount + 1
				thenPush: (objectMemory bytesPerOop = 8
							ifTrue: [objectMemory integerObjectOf: (objectMemory fetchLong32: index ofObject: rcvr)]
							ifFalse: [self positive32BitIntegerFor: (objectMemory fetchLong32: index ofObject: rcvr)]).
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	^self primitiveFailFor: PrimErrBadReceiver! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 2/26/2018 12:32'!
primitiveSlotAtPut
	"Assign a slot in an object.  This numbers all slots from 1, ignoring the distinction between
	 named and indexed inst vars.  In objects with both named and indexed inst vars, the named
	 inst vars precede the indexed ones.  In non-object indexed objects (objects that contain
	 bits, not object references) this primitive assigns a raw integral value at each slot."
	| newValue index rcvr fmt numSlots value |
	newValue := self stackTop.
	index := self stackValue: 1.
	rcvr := self stackValue: 2.
	(objectMemory isIntegerObject: index) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory isImmediate: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	(objectMemory isObjImmutable: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].
	fmt := objectMemory formatOf: rcvr.
	index := (objectMemory integerValueOf: index) - 1.

	fmt <= objectMemory lastPointerFormat ifTrue:
		[numSlots := objectMemory numSlotsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storePointer: index ofObject: rcvr withValue: newValue.
			 self pop: argumentCount + 1 thenPush: newValue.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	value := self positiveMachineIntegerValueOf: newValue.
	self failed ifTrue:
		[primFailCode := PrimErrBadArgument.
		^0].

	fmt >= objectMemory firstByteFormat ifTrue:
		[fmt >= objectMemory firstCompiledMethodFormat ifTrue:
			[^self primitiveFailFor: PrimErrUnsupported].
		 (self asUnsigned: value) > 16rFF ifTrue:
			[^self primitiveFailFor: PrimErrBadArgument].
		 numSlots := objectMemory numBytesOfBytes: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storeByte: index ofObject: rcvr withValue: value.
			 self pop: argumentCount + 1 thenPush: newValue.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	(objectMemory hasSpurMemoryManagerAPI
	 and: [fmt >= objectMemory firstShortFormat]) ifTrue:
		[(self asUnsigned: value) > 16rFFFF ifTrue:
			[^self primitiveFailFor: PrimErrBadArgument].
		 numSlots := objectMemory num16BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storeShort16: index ofObject: rcvr withValue: value.
			 self pop: argumentCount + 1 thenPush: newValue.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	(objectMemory bytesPerOop = 8
	 and: [fmt = objectMemory sixtyFourBitIndexableFormat]) ifTrue:
		[numSlots := objectMemory num64BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storeLong64: index ofObject: rcvr withValue: value.
			 self pop: argumentCount + 1 thenPush: newValue.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	fmt >= objectMemory firstLongFormat ifTrue:
		[(objectMemory wordSize > 4
		  and: [(self asUnsigned: value) > 16rFFFFFFFF]) ifTrue:
			[^self primitiveFailFor: PrimErrBadArgument].
		 numSlots := objectMemory num32BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storeLong32: index ofObject: rcvr withValue: value.
			 self pop: argumentCount + 1 thenPush: newValue.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	^self primitiveFailFor: PrimErrBadReceiver! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 12/18/2014 13:23'!
primitiveSmallFloatAdd
	<option: #Spur64BitMemoryManager>
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory smallFloatValueOf: (self stackValue: 1).
	arg := objectMemory loadFloatOrIntFrom: self stackTop.
	self successful ifTrue:
		[self pop: 2 thenPushFloat: rcvr + arg]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2017 14:55'!
primitiveSmallFloatArctan
	<option: #Spur64BitMemoryManager>
	| rcvr |
	<var: #rcvr type: #double>
	rcvr := objectMemory smallFloatValueOf: self stackTop.
	self stackTopPut: (objectMemory floatObjectOf: (self cCode: [rcvr atan]
														inSmalltalk: [rcvr arcTan]))! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/12/2015 16:40'!
primitiveSmallFloatDivide
	<option: #Spur64BitMemoryManager>
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory smallFloatValueOf: (self stackValue: 1).
	arg := objectMemory loadFloatOrIntFrom: self stackTop.
	arg = 0.0 ifTrue:
		[self primitiveFail].
	self successful ifTrue:
		[self pop: 2 thenPushFloat: rcvr / arg]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 9/4/2019 16:59'!
primitiveSmallFloatEqual
	<option: #Spur64BitMemoryManager>
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory smallFloatValueOf: (self stackValue: 1).
	arg := objectMemory loadFloatOrIntFrom: self stackTop.
	self successful ifTrue:
		[(rcvr = arg
		  and: [objectMemory isIntegerObject: self stackTop])
			ifTrue: "Resolve case of ambiguity so as to have comparison of exact values"
				[| intArg |
				intArg := objectMemory integerValueOf: self stackTop.
				self pop: 2 thenPushBool: rcvr asInteger = intArg]
			ifFalse: [self pop: 2 thenPushBool: rcvr = arg]]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2017 14:56'!
primitiveSmallFloatExp
	"Computes E raised to the receiver power.
	 Since SmallFloats cannot represent NaNs there's no need to special case."
	<option: #Spur64BitMemoryManager>

	self stackTopPut: (objectMemory floatObjectOf: (objectMemory smallFloatValueOf: self stackTop) exp)! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2017 14:57'!
primitiveSmallFloatExponent
	"Answer the exponent part of this float."
	<option: #Spur64BitMemoryManager>

	self stackTopPut: (objectMemory integerObjectOf: (objectMemory exponentOfSmallFloat: self stackTop) - 1)! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2017 14:57'!
primitiveSmallFloatFractionalPart
	<option: #Spur64BitMemoryManager>
	| rcvr frac trunc |
	<var: #rcvr type: #double>
	<var: #frac type: #double>
	<var: #trunc type: #double>
	rcvr := objectMemory smallFloatValueOf: self stackTop.
	frac := self cCode: [self mod: rcvr f: (self addressOf: trunc)]
				inSmalltalk: [rcvr fractionPart].
	self stackTopPut: (objectMemory floatObjectOf: frac)! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 9/4/2019 16:59'!
primitiveSmallFloatGreaterOrEqual
	<option: #Spur64BitMemoryManager>
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory smallFloatValueOf: (self stackValue: 1).
	arg := objectMemory loadFloatOrIntFrom: self stackTop.
	self successful ifTrue:
		[(rcvr = arg
		  and: [objectMemory isIntegerObject: self stackTop])
			ifTrue: "Resolve case of ambiguity so as to have comparison of exact values"
				[| intArg |
				intArg := objectMemory integerValueOf: self stackTop.
				self pop: 2 thenPushBool: rcvr asInteger >= intArg]
			ifFalse: [self pop: 2 thenPushBool: rcvr >= arg]]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 9/4/2019 16:59'!
primitiveSmallFloatGreaterThan
	<option: #Spur64BitMemoryManager>
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory smallFloatValueOf: (self stackValue: 1).
	arg := objectMemory loadFloatOrIntFrom: self stackTop.
	self successful ifTrue:
		[(rcvr = arg
		  and: [objectMemory isIntegerObject: self stackTop])
			ifTrue: "Resolve case of ambiguity so as to have comparison of exact values"
				[| intArg |
				intArg := objectMemory integerValueOf: self stackTop.
				self pop: 2 thenPushBool: rcvr asInteger > intArg]
			ifFalse: [self pop: 2 thenPushBool: rcvr > arg]]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 12/19/2014 15:15'!
primitiveSmallFloatHash
	"Answer the ieee 754 double precision floating point bits of the SmallFloat receiver."
	<option: #Spur64BitMemoryManager>

	self pop: 1 thenPush: (objectMemory positive64BitIntegerFor: (objectMemory smallFloatBitsOf: self stackTop))! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 9/4/2019 16:59'!
primitiveSmallFloatLessOrEqual
	<option: #Spur64BitMemoryManager>
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory smallFloatValueOf: (self stackValue: 1).
	arg := objectMemory loadFloatOrIntFrom: self stackTop.
	self successful ifTrue:
		[(rcvr = arg
		  and: [objectMemory isIntegerObject: self stackTop])
			ifTrue: "Resolve case of ambiguity so as to have comparison of exact values"
				[| intArg |
				intArg := objectMemory integerValueOf: self stackTop.
				self pop: 2 thenPushBool: rcvr asInteger <= intArg]
			ifFalse: [self pop: 2 thenPushBool: rcvr <= arg]]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 9/4/2019 16:59'!
primitiveSmallFloatLessThan
	<option: #Spur64BitMemoryManager>
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory smallFloatValueOf: (self stackValue: 1).
	arg := objectMemory loadFloatOrIntFrom: self stackTop.
	self successful ifTrue:
		[(rcvr = arg
		  and: [objectMemory isIntegerObject: self stackTop])
			ifTrue: "Resolve case of ambiguity so as to have comparison of exact values"
				[| intArg |
				intArg := objectMemory integerValueOf: self stackTop.
				self pop: 2 thenPushBool: rcvr asInteger < intArg]
			ifFalse: [self pop: 2 thenPushBool: rcvr < arg]]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2017 14:58'!
primitiveSmallFloatLogN
	"Natural log."
	<option: #Spur64BitMemoryManager>
	| rcvr |
	<var: #rcvr type: #double>
	rcvr := objectMemory smallFloatValueOf: self stackTop.
	self stackTopPut: (objectMemory floatObjectOf: (self cCode: [rcvr log] inSmalltalk: [rcvr ln]))! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 12/18/2014 13:23'!
primitiveSmallFloatMultiply
	<option: #Spur64BitMemoryManager>
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory smallFloatValueOf: (self stackValue: 1).
	arg := objectMemory loadFloatOrIntFrom: self stackTop.
	self successful ifTrue:
		[self pop: 2 thenPushFloat: rcvr * arg]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 9/4/2019 16:59'!
primitiveSmallFloatNotEqual
	<option: #Spur64BitMemoryManager>
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory smallFloatValueOf: (self stackValue: 1).
	arg := objectMemory loadFloatOrIntFrom: self stackTop.
	self successful ifTrue:
		[(rcvr = arg
		  and: [objectMemory isIntegerObject: self stackTop])
			ifTrue: "Resolve case of ambiguity so as to have comparison of exact values"
				[| intArg |
				intArg := objectMemory integerValueOf: self stackTop.
				self pop: 2 thenPushBool: (rcvr asInteger = intArg) not]
			ifFalse: [self pop: 2 thenPushBool: (rcvr = arg) not]]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2017 14:59'!
primitiveSmallFloatSine
	<option: #Spur64BitMemoryManager>
	| rcvr |
	<var: #rcvr type: #double>
	rcvr := objectMemory smallFloatValueOf: self stackTop.
	self stackTopPut: (objectMemory floatObjectOf: rcvr sin)! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/14/2017 09:52'!
primitiveSmallFloatSquareRoot
	<option: #Spur64BitMemoryManager>
	<var: #rcvr type: #double>
	| rcvr |
	rcvr := objectMemory smallFloatValueOf: self stackTop.
	rcvr >= 0.0
		ifTrue: [self stackTopPut: (objectMemory floatObjectOf: rcvr sqrt)]
		ifFalse: [self primitiveFail]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 12/18/2014 13:24'!
primitiveSmallFloatSubtract
	<option: #Spur64BitMemoryManager>
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory smallFloatValueOf: (self stackValue: 1).
	arg := objectMemory loadFloatOrIntFrom: self stackTop.
	self successful ifTrue:
		[self pop: 2 thenPushFloat: rcvr - arg]! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/12/2015 10:05'!
primitiveSmallFloatTimesTwoPower
	"Multiply the receiver by the power of the argument."
	<option: #Spur64BitMemoryManager>
	| rcvr result arg twiceMaxExponent |
	arg := self stackTop.
	(objectMemory isIntegerObject: arg) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	rcvr := self stackValue: 1.
	"N.B. SmallFloats are finite.  NaN and Infinity overflow into boxed floats.
	 This is doing range checking work that is done in ldexp, but we include
	 it explicitly to exemplify bit manipulation of SmallFloats."
	(objectMemory isSmallFloatZero: rcvr)
		ifTrue:
			[result := rcvr]
		ifFalse:
			[arg := objectMemory integerValueOf: arg.
			 twiceMaxExponent := 2 * (1 << self floatExponentBits).
			 arg < twiceMaxExponent negated
				ifTrue:
					[result := objectMemory mapSignedSmallFloatToSignedSmallFloatZero: rcvr]
				ifFalse:
					["clip arg to at most int range; ldexp's last arg is of type int"
					 arg > twiceMaxExponent ifTrue: [arg := twiceMaxExponent].
					 result := objectMemory floatObjectOf: (self cCode: [self ld: (objectMemory smallFloatValueOf: rcvr)
																				exp: (self cCoerceSimple: arg to: #int)]
																inSmalltalk: [(objectMemory smallFloatValueOf: rcvr) timesTwoPower: arg])]].
	self pop: 2 thenPush: result! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 3/12/2019 15:15'!
primitiveSmallFloatTruncated
	<option: #Spur64BitMemoryManager>
	| rcvr trunc |
	<var: #rcvr type: #double>
	<var: #trunc type: #double>
	rcvr := objectMemory smallFloatValueOf: self stackTop.
	self cCode: [self mod: rcvr f: (self addressOf: trunc)]
		inSmalltalk: [trunc := rcvr truncated].
	((trunc between: objectMemory minSmallInteger asFloat and: objectMemory maxSmallInteger asFloat)
	 and: [objectMemory isIntegerValue: trunc asInteger])
		ifTrue: [self stackTopPut: (objectMemory integerObjectOf: trunc asInteger)]
		ifFalse: [self primitiveFail]! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem (auto pragmas 12/08) 8/11/2008 10:37'!
primitiveSnapshot
	"save a normal snapshot under the same name as it was loaded unless it has been renamed by the last primitiveImageName"
	<inline: false>
	self snapshot: false
! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem (auto pragmas 12/08) 8/11/2008 10:37'!
primitiveSnapshotEmbedded
	"save an embedded snapshot"
	<inline: false>
	self snapshot: true! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 7/18/2015 08:05'!
primitiveSomeInstance
	| class instance |
	class := self stackTop.
	NewspeakVM ifTrue: "For the mirror prims check that the class obj is actually a valid class."
		[(argumentCount < 1
		  or: [(objectMemory isNonImmediate: class)
			  and: [self objCouldBeClassObj: class]]) ifFalse:
			[^self primitiveFailFor: PrimErrBadArgument]].
	instance := objectMemory initialInstanceOf: class.
	instance
		ifNil: [self primitiveFail]
		ifNotNil: [self pop: argumentCount+1 thenPush: instance]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
primitiveSomeObject
	"Return the first object in the heap."

	self pop: argumentCount+1.
	self push: objectMemory firstAccessibleObject.! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
primitiveSpecialObjectsOop
	"Return the oop of the SpecialObjectsArray."

	self pop: 1 thenPush: objectMemory specialObjectsOop.! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 1/18/2020 18:10'!
primitiveSpurFloatArrayAt
	"Index the receiver, which must be an indexable non-pointer object,
	 and yield a float. In Spur, if the receiver is a WordArray the float is
	 interpreted as IEEE single precision, and if a DoubleWordArray as
	 IEEE double precision."

	<inline: true>
	| index rcvr fmt numSlots aFloat doubleBits |
	index := self stackTop.
	rcvr := self stackValue: 1.
	(objectMemory isIntegerObject: index) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory isImmediate: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	fmt := objectMemory formatOf: rcvr.
	index := (objectMemory integerValueOf: index) - 1.

	fmt = objectMemory sixtyFourBitIndexableFormat ifTrue:
		[numSlots := objectMemory num64BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[doubleBits := objectMemory fetchLong64: index ofObject: rcvr.
			 self methodReturnValue: (objectMemory floatObjectOfBits: doubleBits).
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	(fmt >= objectMemory firstLongFormat
	 and: [fmt <= (objectMemory firstLongFormat + 1)]) ifTrue:
		[numSlots := objectMemory num32BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[aFloat := objectMemory fetchFloat32: index ofObject: rcvr.
			 self methodReturnValue: (objectMemory floatObjectOf: aFloat).
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	^self primitiveFailFor: PrimErrBadReceiver! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 1/18/2020 17:58'!
primitiveSpurFloatArrayAtPut
	"Index the receiver, which must be an indexable non-pointer
	 object, and store a float. In Spur, if the receiver is a WordArray the float is
	 stored in IEEE single precision (if possible), and if a DoubleWordArray in
	 IEEE double precision."

	<inline: true>
	| index rcvr valueOop fmt numSlots |
	valueOop := self stackValue: 0.
	index := self stackValue: 1.
	rcvr := self stackValue: 2.
	((objectMemory isFloatInstance: valueOop)
	 and: [objectMemory isIntegerObject: index]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory isImmediate: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	(objectMemory isObjImmutable: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].
	fmt := objectMemory formatOf: rcvr.
	index := (objectMemory integerValueOf: index) - 1.

	fmt = objectMemory sixtyFourBitIndexableFormat ifTrue:
		[numSlots := objectMemory num64BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storeLong64: index ofObject: rcvr withValue: (objectMemory floatValueBitsOf: valueOop).
			 self methodReturnValue: valueOop.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	"N.B. Currently we simply truncate to 32-bits, which matches the behavior of the FloatArrayPlugin.
	 Maybe we should validate and range check."
	(fmt >= objectMemory firstLongFormat
	 and: [fmt <= (objectMemory firstLongFormat + 1)]) ifTrue:
		[numSlots := objectMemory num32BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storeFloat32: index ofObject: rcvr withValue: (objectMemory floatValueOf: valueOop).
			 self methodReturnValue: valueOop.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	^self primitiveFailFor: PrimErrBadReceiver! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 6/29/2018 11:49'!
primitiveSpurIntegerAt
	"Answer the signed integer element of a pure bits receiver.
	 If the receiver is indexable pointers simply function like at:.
	 Favour bits access over pointer access (normal at: being available)."

	<inline: true>
	| index rcvr fmt numSlots value |
	index := self stackTop.
	rcvr := self stackValue: 1.
	(objectMemory isIntegerObject: index) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory isImmediate: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	fmt := objectMemory formatOf: rcvr.
	index := (objectMemory integerValueOf: index) - 1.

	fmt >= objectMemory firstByteFormat ifTrue:
		[fmt >= objectMemory firstCompiledMethodFormat ifTrue:
			[^self primitiveFailFor: PrimErrUnsupported].
		 numSlots := objectMemory numBytesOfBytes: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[value := objectMemory fetchByte: index ofObject: rcvr.
			 value > 127 ifTrue: [value := value - 256].
			 self methodReturnInteger: value.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	fmt >= objectMemory firstShortFormat ifTrue:
		[numSlots := objectMemory num16BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[value := objectMemory fetchShort16: index ofObject: rcvr.
			 value > 32767 ifTrue: [value := value - 65536].
			 self methodReturnInteger: value.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	fmt = objectMemory sixtyFourBitIndexableFormat ifTrue:
		[numSlots := objectMemory num64BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[self
				cCode:
					[self methodReturnValue: (self signed64BitIntegerFor: (objectMemory fetchLong64: index ofObject: rcvr))]
				inSmalltalk:
					[value := objectMemory fetchLong64: index ofObject: rcvr.
					 value > ((2 raisedTo: 63) - 1) ifTrue:
						[value := value - (2 raisedTo: 64)].
					 self methodReturnValue: (self signed64BitIntegerFor: value)].
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	fmt >= objectMemory firstLongFormat ifTrue:
		[numSlots := objectMemory num32BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory bytesPerOop = 8
				ifTrue:
					[value := objectMemory fetchLong32: index ofObject: rcvr.
					 value > 2147483647 ifTrue: [value := value - 4294967296].
					 self methodReturnInteger: value]
				ifFalse: [self methodReturnValue: (self signed32BitIntegerFor: (objectMemory fetchLong32: index ofObject: rcvr))].
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	(fmt <= objectMemory lastPointerFormat
	 and: [objectMemory isIndexableFormat: fmt]) ifTrue:
		[| numFixed |
		 numSlots := objectMemory numSlotsOf: rcvr.
		 fmt = objectMemory arrayFormat ifTrue:
			[(self asUnsigned: index) < numSlots ifTrue:
				[self methodReturnValue: (objectMemory fetchPointer: index ofObject: rcvr).
				 ^0]].
		 numFixed := self numFixedSlotsOf: rcvr.
		 (index + 1 between: numFixed and: numSlots) ifTrue:
			[self methodReturnValue: (objectMemory fetchPointer: index + numFixed ofObject: rcvr).
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	^self primitiveFailFor: PrimErrBadReceiver! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 7/3/2018 18:14'!
primitiveSpurIntegerAtPut
	"Assign an indexable variable of a pure bits receiver with a signed integer.
	 If the receiver is indexable pointers simply function like at:put: primitive 61.
	 Favour bits access over pointer access (normal at:put: being available)."

	<inline: true>
	| index rcvr value valueOop fmt numSlots |
	valueOop := self stackValue: 0.
	index := self stackValue: 1.
	rcvr := self stackValue: 2.
	value := self signed64BitValueOf: valueOop.
	(self successful
	 and: [objectMemory isIntegerObject: index]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory isImmediate: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	(objectMemory isObjImmutable: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].
	fmt := objectMemory formatOf: rcvr.
	index := (objectMemory integerValueOf: index) - 1.

	fmt >= objectMemory firstByteFormat ifTrue:
		[fmt >= objectMemory firstCompiledMethodFormat ifTrue:
			[^self primitiveFailFor: PrimErrUnsupported].
		 (self isSignedInteger: value inRangeForBits: 8) ifFalse:
			[^self primitiveFailFor: PrimErrBadArgument].
		 numSlots := objectMemory numBytesOfBytes: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storeByte: index ofObject: rcvr withValue: (self cCode: [value] inSmalltalk: [value bitAnd: 16rFF]).
			 self methodReturnValue: valueOop.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	fmt >= objectMemory firstShortFormat ifTrue:
		[(self isSignedInteger: value inRangeForBits: 16) ifFalse:
			[^self primitiveFailFor: PrimErrBadArgument].
		 numSlots := objectMemory num16BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storeShort16: index ofObject: rcvr withValue: (self cCode: [value] inSmalltalk: [value bitAnd: 16rFFFF]).
			 self methodReturnValue: valueOop.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	fmt = objectMemory sixtyFourBitIndexableFormat ifTrue:
		["No range check on value in this case because signed64BitValueOf: performed it above."
		 numSlots := objectMemory num64BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storeLong64: index ofObject: rcvr withValue: (self cCode: [value] inSmalltalk: [value bitAnd: 16rFFFFFFFFFFFFFFFF]).
			 self methodReturnValue: valueOop.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	fmt >= objectMemory firstLongFormat ifTrue:
		[(self isSignedInteger: value inRangeForBits: 32) ifFalse:
			[^self primitiveFailFor: PrimErrBadArgument].
		 numSlots := objectMemory num32BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storeLong32: index ofObject: rcvr withValue: (self cCode: [value] inSmalltalk: [value bitAnd: 16rFFFFFFFF]).
			 self methodReturnValue: valueOop.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	(fmt <= objectMemory lastPointerFormat
	 and: [objectMemory isIndexableFormat: fmt]) ifTrue:
		[| numFixed |
		 numSlots := objectMemory numSlotsOf: rcvr.
		 fmt = objectMemory arrayFormat ifTrue:
			[(self asUnsigned: index) < numSlots ifTrue:
				[self storePointer: index ofObject: rcvr withValue: valueOop.
				 self methodReturnValue: valueOop.
				 ^0]].
		 numFixed := self numFixedSlotsOf: rcvr.
		 (index + 1 between: numFixed and: numSlots) ifTrue:
			[self storePointer: index + numFixed ofObject: rcvr withValue: valueOop.
			 self methodReturnValue: valueOop.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	^self primitiveFailFor: PrimErrBadReceiver! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 1/10/2017 17:18'!
primitiveSpurStringReplace
	"<array> primReplaceFrom: start to: stop with: replacement startingAt: repStart  
		<primitive: 105>"
	<inline: true>
	| array start stop repl replStart arrayFmt arrayLength arrayInstSize replFmt replLength replInstSize srcDelta |
	array := self stackValue: 4.
	start := self stackValue: 3.
	stop := self stackValue: 2.
	repl := self stackValue: 1.
	replStart := self stackValue: 0.

	((objectMemory isNonIntegerObject: start)
	 or: [(objectMemory isNonIntegerObject: stop)
	 or: [(objectMemory isNonIntegerObject: replStart)
	 or: [objectMemory isImmediate: repl]]]) ifTrue: "can happen in LgInt copy"
		[^self primitiveFailFor: PrimErrBadArgument].

	start := objectMemory integerValueOf: start.
	stop := objectMemory integerValueOf: stop.
	replStart := objectMemory integerValueOf: replStart.

	(stop >= start and: [objectMemory isObjImmutable: array]) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].

	arrayFmt := objectMemory formatOf: array.
	arrayLength := objectMemory lengthOf: array format: arrayFmt.
	arrayFmt := objectMemory classFormatFromInstFormat: arrayFmt.
	replFmt := objectMemory formatOf: repl.
	replLength := objectMemory lengthOf: repl format: replFmt.
	replFmt := objectMemory classFormatFromInstFormat: replFmt.

	"Array formats (without oddFields bits) must be the same"
	(arrayFmt = replFmt
	 and: [arrayFmt < objectMemory firstCompiledMethodFormat]) ifFalse:
		[^self primitiveFailFor: PrimErrInappropriate].

	"N.B. In the below start - 1 to: stop - 1 do:, Slang is intelligent enough to use < instead of <= so avoiding the stop - 1."
	arrayFmt <= objectMemory lastPointerFormat
		ifTrue:
			[arrayInstSize := objectMemory fixedFieldsOf: array format: arrayFmt length: arrayLength.
			 replInstSize := objectMemory fixedFieldsOf: repl format: replFmt length: replLength.
			 (start >= 1 and: [start - 1 <= stop and: [stop + arrayInstSize <= arrayLength
			  and: [replStart >= 1 and: [stop - start + replStart + replInstSize <= replLength]]]]) ifFalse:
				[^self primitiveFailFor: PrimErrBadIndex].
			 start := start + arrayInstSize.
			 stop := stop + arrayInstSize.
			 srcDelta := (replStart + replInstSize) - start.
			 (objectMemory isOldObject: array)
				ifTrue:
					[| mustRemember oop |
					 mustRemember := false.
					 start - 1 to: stop - 1 do:
						[:i |
						oop := objectMemory fetchPointer: srcDelta + i ofObject: repl.
						(objectMemory isYoung: oop) ifTrue:
							[mustRemember := true].
						objectMemory storePointerUnchecked: i ofObject: array withValue: oop].
					 mustRemember ifTrue:
						[objectMemory possibleRootStoreInto: array]]
				ifFalse:
					[start - 1 to: stop - 1 do:
						[:i | objectMemory storePointerUnchecked: i ofObject: array withValue: (objectMemory fetchPointer: srcDelta + i ofObject: repl)]]]
		ifFalse:
			[(start >= 1 and: [start - 1 <= stop and: [stop <= arrayLength
			  and: [replStart >= 1 and: [stop - start + replStart <= replLength]]]]) ifFalse:
				[^self primitiveFailFor: PrimErrBadIndex].
			 srcDelta := replStart - start.
			 arrayFmt >= objectMemory firstShortFormat
				ifTrue: "8 & 16-bit word type objects"
					[arrayFmt >= objectMemory firstByteFormat
						ifTrue: "byte-type objects"
							[start - 1 to: stop - 1 do:
								[:i | objectMemory storeByte: i ofObject: array withValue: (objectMemory fetchByte: srcDelta + i ofObject: repl)]]
						ifFalse: "short type objects"
							[start - 1 to: stop - 1 do:
								[:i | objectMemory storeShort16: i ofObject: array withValue: (objectMemory fetchShort16: srcDelta + i ofObject: repl)]]]
				ifFalse: "32 & 64-bit word type objects"
					[arrayFmt >= objectMemory firstLongFormat
						ifTrue: "word-type objects"
							[start - 1 to: stop - 1 do:
								[:i | objectMemory storeLong32: i ofObject: array withValue: (objectMemory fetchLong32: srcDelta + i ofObject: repl)]]
						ifFalse: "long type objects"
							[start - 1 to: stop - 1 do:
								[:i | objectMemory storeLong64: i ofObject: array withValue: (objectMemory fetchLong64: srcDelta + i ofObject: repl)]]]].
	"We might consider comparing stop - start to some value here and using forceInterruptCheck"

	self pop: argumentCount "leave rcvr on stack"! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2017 14:54'!
primitiveSquareRoot
	"N.B. IMO we should be able to assume the receiver is a float because this primitive is specific to floats. eem 2/13/2017"
	| rcvr |
	<var: #rcvr type: #double>
	rcvr := self stackFloatValue: 0.
	(self successful and: [rcvr >= 0.0])
		ifTrue: [self stackTopPut: (objectMemory floatObjectOf: rcvr sqrt)]
		ifFalse: [self primitiveFail]! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 8/21/2008 10:58'!
primitiveStartVMProfiling
	"Primitive. Start the VM profiler."
	self cCode: 'ioControlProfile(1,0,0,0,0)'.
	self pop: argumentCount! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 07/28/2014 07:54'!
primitiveStopVMProfiling
	"Primitive. Stop the VM profiler and either copy the histogram data into the
	 supplied arguments, if they're non-nil.  Fail if the arguments are not of the right type or size."
	| vmHistArrayOrNil vmHist vmBins easHistArrayOrNil easHist easBins |
	<var: #vmHist type: #'long *'>
	<var: #vmBins type: #long>
	<var: #easHist type: #'long *'>
	<var: #easBins type: #long>
	self success: argumentCount = 2.
	vmHistArrayOrNil := self stackObjectValue: 1.
	easHistArrayOrNil := self stackObjectValue: 0.
	self successful ifFalse:
		[^nil].
	"Both args must be either nil or arrays.  If they're arrays and the wrong size we incorrectly stop profiling."
	((vmHistArrayOrNil = objectMemory nilObject or: [(objectMemory fetchClassOfNonImm: vmHistArrayOrNil) = (objectMemory splObj: ClassArray)])
	 and: [(objectMemory fetchClassOfNonImm: vmHistArrayOrNil) = (objectMemory fetchClassOfNonImm: easHistArrayOrNil)]) ifFalse:
		[^self primitiveFail].
	self cCode: 'ioControlProfile(0,&vmHist,&vmBins,&easHist,&easBins)'
		inSmalltalk: [vmHist := vmBins := easHist := easBins := 0].
	vmHistArrayOrNil ~= objectMemory nilObject ifTrue:
		[((objectMemory numSlotsOf: vmHistArrayOrNil) = vmBins
		  and: [(objectMemory numSlotsOf: easHistArrayOrNil) = easBins]) ifFalse:
			[^self primitiveFail].
		0 to: vmBins - 1 do:
			[:i|
			objectMemory storePointerUnchecked: i
				ofObject: vmHistArrayOrNil
				withValue: (objectMemory integerObjectOf: (vmHist at: i))].
		0 to: easBins - 1 do:
			[:i|
			objectMemory storePointerUnchecked: i
				ofObject: easHistArrayOrNil
				withValue: (objectMemory integerObjectOf: (easHist at: i))]].
	self pop: argumentCount! !

!InterpreterPrimitives methodsFor: 'image segment in/out' stamp: 'eem 2/28/2019 11:06'!
primitiveStoreImageSegment
	"This primitive is called from Squeak as...
		<imageSegment> storeSegmentFor: arrayOfRoots into: aWordArray outPointers: anArray."

"This primitive will store a binary image segment (in the same format as the Squeak image file) of the receiver and every object in its proper tree of subParts (ie, that is not refered to from anywhere else outside the tree).  All pointers from within the tree to objects outside the tree will be copied into the array of outpointers.  In their place in the image segment will be an oop equal to the offset in the outPointer array (the first would be 4). but with the high bit set."

"The primitive expects the array and wordArray to be more than adequately long.  In this case it returns normally, and truncates the two arrays to exactly the right size.  To simplify truncation, both incoming arrays are required to be 256 bytes or more long (ie with 3-word headers).  If either array is too small, the primitive will fail, but in no other case.

During operation of the primitive, it is necessary to convert from both internal and external oops to their mapped values.  To make this fast, the headers of the original objects in question are replaced by the mapped values (and this is noted by adding the forbidden XX header type).  Tables are kept of both kinds of oops, as well as of the original headers for restoration.

To be specific, there are two similar two-part tables, the outpointer array, and one in the upper fifth of the segmentWordArray.  Each grows oops from the bottom up, and preserved headers from halfway up.

In case of either success or failure, the headers must be restored.  In the event of primitive failure, the table of outpointers must also be nilled out (since the garbage in the high half will not have been discarded."

	| outPointerArray segmentWordArray arrayOfRoots ecode |

	outPointerArray := self stackTop.
	segmentWordArray := self stackValue: 1.
	arrayOfRoots := self stackValue: 2.

	"Essential type checks"
	((objectMemory isArray: arrayOfRoots)				"Must be indexable pointers"
	and: [(objectMemory isArray: outPointerArray)		"Must be indexable pointers"
	and: [objectMemory isWords: segmentWordArray]])	"Must be indexable words"
		ifFalse: [^self primitiveFail].

	ecode := objectMemory storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots.
	(objectMemory hasSpurMemoryManagerAPI
	 and: [ecode = PrimErrNeedCompaction]) ifTrue:
		[objectMemory fullGC.
		 outPointerArray := self stackTop.
		 segmentWordArray := self stackValue: 1.
		 arrayOfRoots := self stackValue: 2.
		 ecode := objectMemory storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots].
	ecode = PrimNoErr
		ifTrue: [self pop: 3]  "...leaving the receiver on the stack as return value"
		ifFalse: [self primitiveFailFor: ecode]! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: ''!
primitiveStringAt

	self commonAt: true.! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: ''!
primitiveStringAtPut

	self commonAtPut: true.! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 1/10/2017 16:37'!
primitiveStringReplace
	"<array> primReplaceFrom: start to: stop with: replacement startingAt: repStart  
		<primitive: 105>"
	objectMemory hasSpurMemoryManagerAPI
		ifTrue: [self primitiveSpurStringReplace]
		ifFalse: [self primitiveV3StringReplace]! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'di 11/30/1998 10:13'!
primitiveSubtract

	self pop2AndPushIntegerIfOK: (self stackIntegerValue: 1) - (self stackIntegerValue: 0)! !

!InterpreterPrimitives methodsFor: 'arithmetic largeint primitives' stamp: 'dtl 9/2/2012 10:08'!
primitiveSubtractLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| a b result oopResult aIsNegative bIsNegative resultIsNegative oopArg oopRcvr |
	<export: true>
	<var: 'a' type: 'usqLong'>
	<var: 'b' type: 'usqLong'>
	<var: 'result' type: 'usqLong'>

	oopArg := self stackValue: 0.
	oopRcvr := self stackValue: 1.
	aIsNegative := self isNegativeIntegerValueOf: oopRcvr.
	bIsNegative := self isNegativeIntegerValueOf: oopArg.
	a := self magnitude64BitValueOf: oopRcvr.
	b := self magnitude64BitValueOf: oopArg.
	self successful ifFalse:[^nil].
	(aIsNegative ~= bIsNegative)
		ifTrue:
			["Protect against overflow"
			a > (16rFFFFFFFFFFFFFFFF - b) ifTrue: [self primitiveFail. ^nil].
			result := a + b.
			resultIsNegative := aIsNegative]
		ifFalse:
			[(a >= b)
				ifTrue:
					[result := a - b.
					resultIsNegative := aIsNegative]
				ifFalse:
					[result := b - a.
					resultIsNegative := aIsNegative not]].
	oopResult := self magnitude64BitIntegerFor: result neg: resultIsNegative.
	self successful ifTrue:[self pop: 2 thenPush: oopResult].
! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 8/29/2020 19:18'!
primitiveSuspend
	"Primitive. Suspend the receiver, aProcess such that it can be executed again
	by sending #resume. If the given process is not currently running, take it off
	its corresponding list. The primitive returns the list the receiver was previously on."
	| process myList |
	process := self stackTop.
	process = self activeProcess ifTrue:
		[self pop: 1 thenPush: objectMemory nilObject.
		 ^self transferTo: self wakeHighestPriority].
	myList := objectMemory fetchPointer: MyListIndex ofObject: process.
	"XXXX Fixme. We should really check whether myList is a kind of LinkedList or not
	but we can't easily so just do a quick check for nil which is the most common case."
	myList = objectMemory nilObject ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	"Alas in Spur we need a read barrier"
	(objectMemory isForwarded: myList) ifTrue:
		[myList := objectMemory followForwarded: myList.
		 objectMemory storePointer: MyListIndex ofObject: process withValue: myList].
	self removeProcess: process fromList: myList.
	self successful ifTrue:
		[objectMemory storePointerUnchecked: MyListIndex ofObject: process withValue: objectMemory nilObject.
		 self pop: 1 thenPush: myList]! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 10/21/2010 21:59'!
primitiveTestAndSetOwnershipOfCriticalSection
	"Attempt to test-and-set the ownership of the critical section.  If not owned,
	 set the owner to the current process and answer false. If owned by the
	 current process answer true.  If owned by some other process answer nil.
	 For simulation if there is an argument it is taken to be the effective activeProcess
	 (see Process>>effectiveProcess)."
	| criticalSection owningProcessIndex owningProcess activeProc |
	argumentCount > 0
		ifTrue:
			[criticalSection := self stackValue: 1.  "rcvr"
			 activeProc := self stackTop]
		ifFalse:
			[criticalSection := self stackTop.  "rcvr"
			 activeProc := self activeProcess].
	owningProcessIndex := ExcessSignalsIndex. "CriticalSections are laid out like Semaphores"
	owningProcess := objectMemory fetchPointer: owningProcessIndex ofObject: criticalSection.
	owningProcess = objectMemory nilObject ifTrue:
		[objectMemory storePointer: owningProcessIndex
			ofObject: criticalSection
			withValue: activeProc.
		 ^self pop: argumentCount + 1 thenPush: objectMemory falseObject].
	owningProcess = activeProc ifTrue:
		[^self pop: argumentCount + 1 thenPush: objectMemory trueObject].
	self pop: argumentCount + 1 thenPush: objectMemory nilObject! !

!InterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 11/25/2008 12:50'!
primitiveTestDisplayDepth
	"Return true if the host OS does support the given display depth."
	| bitsPerPixel okay|
	bitsPerPixel := self stackIntegerValue: 0.
	self successful ifTrue: [okay := self ioHasDisplayDepth: bitsPerPixel].
	self successful ifTrue: [self pop: 2 thenPushBool: okay"Pop arg+rcvr"]! !

!InterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 6/21/2021 12:08'!
primitiveTestShortenIndexableSize
	"Given an object with indexable pointer fields, reduce the size of the indexable fields
	to the requested size. Answer the number of bytes freed, or zero if the object cannot
	be shortened."
	<option: #TestingPrimitives>
	<export: true>
	| array newSizeOop |
	newSizeOop := self stackTop.
	array := self stackValue: 1.
	((objectMemory isIntegerObject: newSizeOop)
	 and: [(objectMemory integerValueOf: newSizeOop) between: 0 and: (self stSizeOf: array) - 1]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory canShorten: array) ifFalse:
		[^self primitiveFailFor: PrimErrBadReceiver].
	self methodReturnInteger: (objectMemory
									shorten: array
									toIndexableSize: (objectMemory integerValueOf: newSizeOop))! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2015 10:34'!
primitiveTimesTwoPower
	"Multiply the receiver by the power of the argument."
	| rcvr result arg |
	<var: #rcvr type: #double>
	<var: #result type: #double>
	arg := self stackTop.
	(objectMemory isIntegerObject: arg) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	arg := objectMemory integerValueOf: arg.
	objectMemory bytesPerOop > 4 ifTrue:
		[| twiceMaxExponent | "clip arg to at most int range; ldexp's last arg is of type int"
		 twiceMaxExponent := 2 * (1 << self floatExponentBits).
	 	 arg < twiceMaxExponent negated
			ifTrue: [arg := twiceMaxExponent negated]
			ifFalse: [arg > twiceMaxExponent ifTrue:
						[arg := twiceMaxExponent]]].
	rcvr := objectMemory floatValueOf: (self stackValue: 1).
	result := self cCode: [self ld: rcvr exp: (self cCoerceSimple: arg to: #int)]
					inSmalltalk: [rcvr timesTwoPower: arg].
	self pop: 2 thenPushFloat: result! !

!InterpreterPrimitives methodsFor: 'arithmetic float primitives' stamp: 'eem 2/13/2017 15:05'!
primitiveTruncated
	"N.B. IMO we should be able to assume the receiver is a float because this primitive is specific to floats. eem 2/13/2017"
	| rcvr trunc |
	<var: #rcvr type: #double>
	<var: #trunc type: #double>
	rcvr := self stackFloatValue: 0.
	self successful ifTrue:
		[self cCode: [self mod: rcvr f: (self addressOf: trunc)]
			inSmalltalk: [trunc := rcvr = rcvr
									ifTrue: [rcvr truncated]
									ifFalse: [Float nan]].
		 (trunc between: objectMemory minSmallInteger asFloat and: objectMemory maxSmallInteger asFloat)
			ifTrue: [self stackTopPut: (objectMemory integerObjectOf: trunc asInteger)]
			ifFalse: [self primitiveFail]]! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 5/26/2017 15:11'!
primitiveUTCMicrosecondClock
	"Return the value of the microsecond clock in UTC as an integer.
	 This is the number of microseconds since the Smalltalk epoch, 1901/1/1 12:00am.
	 The microsecond clock is at least 60 bits wide which means it'll get to around August
	 38435 before it wraps around.  Be sure to put it on your calendar.  This primitive
	 accesses the time as answered by the OS."

	self pop: 1 thenPush: (self positive64BitIntegerFor: self ioUTCMicrosecondsNow)! !

!InterpreterPrimitives methodsFor: 'plugin primitives' stamp: 'eem 2/20/2019 17:44'!
primitiveUnloadModule
	"Primitive. Unload the module with the given name."
	"Reloading of the module will happen *later* automatically, when a 
	 function from it is called. This is ensured by invalidating current sessionID."
	| moduleName |
	self methodArgumentCount = 1 ifFalse:[^self primitiveFail].
	moduleName := self stackTop.
	(objectMemory isBytes: moduleName) ifFalse:[^self primitiveFail].
	(self ioUnloadModule: (self oopForPointer: (objectMemory firstIndexableField: moduleName))
		OfLength: (objectMemory byteSizeOf: moduleName)) ifFalse:[^self primitiveFail].
	self flushExternalPrimitives.
	self forceInterruptCheck.
	self pop: 1 "pop moduleName; return receiver"! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 8/24/2010 09:32'!
primitiveUpdateTimezone
	"Update the VMs notion of the current timezone.  The VM sets its notion
	 of the timezone once at start-up.  If one wants the VM to keep its notion
	 up-to-date arrange to invoke this primitive periodically."
	self ioUpdateVMTimezone! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 4/22/2016 12:58'!
primitiveUtcAndTimezoneOffset
	"Answer an array with UTC microseconds since the Smalltalk epoch and the
	current seconds offset from GMT in the local time zone. Any pointer object
	 with two or more slots) may be supplied as a parameter."
	| resultArray |
	argumentCount > 0
		ifTrue:
			[argumentCount > 1 ifTrue: [^self primitiveFailFor: PrimErrBadNumArgs].
			 resultArray := self stackTop.
			 ((objectMemory isPointers: resultArray)
			  and: [(objectMemory lengthOf: resultArray) >= 2]) ifFalse:
				[^self primitiveFailFor: PrimErrBadArgument]]
		ifFalse: [resultArray := objectMemory instantiateClass: objectMemory classArray indexableSize: 2].
	"N.B. No pushRemappableOop:/popRemappableOop in Cog because positive64BitIntegerFor: et al use
	 eeInstantiate... allocators which are guaranteed not to do a GC."
	objectMemory
		storePointerUnchecked: 1 ofObject: resultArray withValue: (objectMemory integerObjectOf: self ioLocalSecondsOffset);
		storePointer: 0 ofObject: resultArray withValue: (self positive64BitIntegerFor: self ioUTCMicroseconds).
	self pop: argumentCount + 1 thenPush: resultArray! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'dtl 5/9/2020 19:01'!
primitiveUtcWithOffset
	"Answer an array with UTC microseconds since the Posix epoch and the
	current seconds offset from GMT in the local time zone. An empty two
	element array (or any object with two or more slots) may be supplied
	as a parameter.
	This is a named (not numbered) primitive in the null module (ie the VM)"
	| epochDelta resultArray |
	<export: true>
	<var: #epochDelta declareC: 'static usqLong epochDelta = 2177452800000000ULL'>
	self cCode: '' inSmalltalk: [epochDelta := 2177452800000000].
	argumentCount > 0
		ifTrue:
			[argumentCount > 1 ifTrue: [^self primitiveFailFor: PrimErrBadNumArgs].
			 resultArray := self stackTop.
			 ((objectMemory isPointers: resultArray)
			  and: [(objectMemory lengthOf: resultArray) >= 2]) ifFalse:
				[^self primitiveFailFor: PrimErrBadArgument]]
		ifFalse: [resultArray := objectMemory instantiateClass: objectMemory classArray indexableSize: 2].
	"N.B. No pushRemappableOop:/popRemappableOop in Cog because positive64BitIntegerFor: et al use
	 eeInstantiate... allocators which are guaranteed not to do a GC."
	objectMemory
		storePointerUnchecked: 1 ofObject: resultArray withValue: (objectMemory integerObjectOf: self ioLocalSecondsOffset);
		storePointer: 0 ofObject: resultArray withValue: (self positive64BitIntegerFor: self ioUTCMicrosecondsNow - epochDelta).
	self pop: argumentCount + 1 thenPush: resultArray! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 6/21/2018 16:49'!
primitiveV3IntegerAt
	"Answer the 32bit signed integer contents of a words receiver"
	<inline: true>
	| index rcvr sz addr intValue result |
	<var: #intValue type: #int>
	index := self stackValue: 0.
	(objectMemory isIntegerObject: index) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	rcvr := self stackValue: 1.
	(objectMemory isWords: rcvr) ifFalse:
		[^self primitiveFailFor: PrimErrInappropriate].
	index := objectMemory integerValueOf: index.
	sz := objectMemory lengthOf: rcvr.  "number of fields"
	(index >= 1 and: [index <= sz]) ifFalse:
		[^self primitiveFailFor: PrimErrBadIndex].
	"4 = 32 bits / 8"
	addr := rcvr + objectMemory baseHeaderSize + (index - 1 * 4). "for zero indexing"
	intValue := objectMemory intAt: addr.
	result := self signed32BitIntegerFor: intValue.
	self pop: 2 thenPush: result! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 6/29/2018 11:17'!
primitiveV3IntegerAtPut
	"Assign an indexable variable of a words receiver with a 32 bit signed integer."
	| index rcvr sz addr value valueOop |
	<var: 'value' type: #int>
	valueOop := self stackValue: 0.
	index := self stackIntegerValue: 1.
	value := self signed32BitValueOf: valueOop.
	self successful ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	rcvr := self stackValue: 2.
	(objectMemory isWords: rcvr) ifFalse:
		[^self primitiveFailFor: PrimErrInappropriate].
	(objectMemory isObjImmutable: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].
	sz := objectMemory lengthOf: rcvr.  "number of fields"
	(index >= 1 and: [index <= sz]) ifFalse:
		[^self primitiveFailFor: PrimErrBadIndex].
	"4 = 32 bits / 8"
	addr := rcvr + objectMemory baseHeaderSize + (index - 1 * 4). "for zero indexing"
	value := objectMemory intAt: addr put: value.
	self pop: 3 thenPush: valueOop "pop all; return value"! !

!InterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 1/10/2017 16:38'!
primitiveV3StringReplace
	"<array> primReplaceFrom: start to: stop with: replacement startingAt: repStart  
		<primitive: 105>"
	<inline: true>
	| array start stop repl replStart hdr arrayFmt totalLength arrayInstSize replFmt replInstSize srcIndex |
	array := self stackValue: 4.
	start := self stackIntegerValue: 3.
	stop := self stackIntegerValue: 2.
	repl := self stackValue: 1.
	replStart := self stackIntegerValue: 0.

	self successful ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument]. 
	(objectMemory isImmediate: repl) ifTrue: "can happen in LgInt copy"
		[^self primitiveFailFor: PrimErrBadArgument].
	(stop >= start and: [objectMemory isObjImmutable: array]) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].

	hdr := objectMemory baseHeader: array.
	arrayFmt := objectMemory formatOfHeader: hdr.
	totalLength := objectMemory lengthOf: array baseHeader: hdr format: arrayFmt.
	arrayInstSize := objectMemory fixedFieldsOf: array format: arrayFmt length: totalLength.
	(start >= 1 and: [start - 1 <= stop and: [stop + arrayInstSize <= totalLength]]) ifFalse:
		[^self primitiveFailFor: PrimErrBadIndex].

	hdr := objectMemory baseHeader: repl.
	replFmt := objectMemory formatOfHeader: hdr.
	totalLength := objectMemory lengthOf: repl baseHeader: hdr format: replFmt.
	replInstSize := objectMemory fixedFieldsOf: repl format: replFmt length: totalLength.
	(replStart >= 1 and: [stop - start + replStart + replInstSize <= totalLength]) ifFalse:
		[^self primitiveFailFor: PrimErrBadIndex].

	"Array formats (without byteSize bits, if bytes array) must be the same"
	arrayFmt < objectMemory firstByteFormat
		ifTrue: [arrayFmt = replFmt ifFalse:
					[^self primitiveFailFor: PrimErrInappropriate]]
		ifFalse: [(arrayFmt bitAnd: objectMemory byteFormatMask) = (replFmt bitAnd: objectMemory byteFormatMask) ifFalse:
					[^self primitiveFailFor: PrimErrInappropriate]].

	srcIndex := replStart + replInstSize - 1.
	"- 1 for 0-based access"

	arrayFmt <= objectMemory lastPointerFormat
		ifTrue:
			[start + arrayInstSize - 1 to: stop + arrayInstSize - 1 do:
				[:i |
				objectMemory storePointer: i ofObject: array withValue: (objectMemory fetchPointer: srcIndex ofObject: repl).
				srcIndex := srcIndex + 1]]
		ifFalse:
			[arrayFmt < objectMemory firstByteFormat
				ifTrue: "32-bit-word type objects"
					[start + arrayInstSize - 1 to: stop + arrayInstSize - 1 do:
						[:i |
						objectMemory storeLong32: i ofObject: array withValue: (objectMemory fetchLong32: srcIndex ofObject: repl).
						srcIndex := srcIndex + 1]]
				ifFalse: "byte-type objects"
					[start + arrayInstSize - 1 to: stop + arrayInstSize - 1 do:
						[:i |
						objectMemory storeByte: i ofObject: array withValue: (objectMemory fetchByte: srcIndex ofObject: repl).
						srcIndex := srcIndex + 1]]].
	"We might consider comparing stop - start to some value here and using forceInterruptCheck"

	self pop: argumentCount "leave rcvr on stack"! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 11/21/2010 11:27'!
primitiveVMCurrentThreadId
	"Simulation only.  Fail."
	<doNotGenerate>
	self primitiveFail! !

!InterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
primitiveVMPath
	"Return a string containing the path name of VM's directory."

	| s sz |
	sz := self vmPathSize.
	s := objectMemory instantiateClass: (objectMemory splObj: ClassByteString) indexableSize: sz.
	self vmPathGet: (s + objectMemory baseHeaderSize) Length: sz.
	self pop: 1 thenPush: s.
! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 10/25/2016 13:32'!
primitiveVMProfileSamplesInto
	"Primitive.
	 0 args: Answer whether the VM Profiler is running or not.
	 1 arg:	Copy the sample data into the supplied argument, which must be a Bitmap
			of suitable size. Answer the number of samples copied into the buffer."
	| sampleBuffer running bufferSize numSamples |
	<var: #bufferSize type: #long>
	self ioNewProfile: (self addressOf: running put: [:v| running := v])
		Status: (self addressOf: bufferSize put: [:v| bufferSize := v]).
	argumentCount = 0 ifTrue:
		[^self pop: 1 thenPushBool: running].
	argumentCount = 1 ifFalse:
		[^self primitiveFailFor: PrimErrBadNumArgs].
	sampleBuffer := self stackValue: 0.
	((objectMemory isNonImmediate: sampleBuffer)
	 and: [(objectMemory isPureBitsNonImm: sampleBuffer)
	 and: [(objectMemory numBytesOf: sampleBuffer) >= (bufferSize * objectMemory wordSize)]]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	numSamples := self ioNewProfileSamplesInto: (objectMemory firstFixedField: sampleBuffer).
	self pop: argumentCount + 1 thenPushInteger: numSamples! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 8/6/2009 18:16'!
primitiveWait
	| sema excessSignals activeProc |
	sema := self stackTop.  "rcvr"
	excessSignals := self fetchInteger: ExcessSignalsIndex ofObject: sema.
	excessSignals > 0
		ifTrue:
			[self storeInteger: ExcessSignalsIndex ofObject: sema withValue: excessSignals - 1]
		ifFalse:
			[activeProc := self activeProcess.
			 self addLastLink: activeProc toList: sema.
			 self transferTo: self wakeHighestPriority]! !

!InterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 3/2/-4654 00:02'!
primitiveYield
	"Primitively do the equivalent of Process>yield, avoiding the overhead of a fork and a wait in the standard implementation."
	| scheduler activeProc priority processLists processList |
	scheduler := self schedulerPointer.
	activeProc := objectMemory fetchPointer: ActiveProcessIndex ofObject: scheduler.
	priority := self quickFetchInteger: PriorityIndex ofObject: activeProc.
	processLists := objectMemory fetchPointer: ProcessListsIndex ofObject: scheduler.
	processList := objectMemory fetchPointer: priority - 1 ofObject: processLists.

	(self isEmptyList: processList) ifFalse:
		[self addLastLink: activeProc toList: processList.
		 self transferTo: self wakeHighestPriority]! !

!InterpreterPrimitives methodsFor: 'arithmetic integer primitives' stamp: 'eem 8/7/2014 06:43'!
quot: integerRcvr ient: integerArg
	"See Number>>#quo:.  Rounds results towards zero."
	<inline: true>
	^integerRcvr > 0
		ifTrue:
			[integerArg > 0
				ifTrue: [integerRcvr // integerArg]
				ifFalse: [0 - (integerRcvr // (0 - integerArg))]]
		ifFalse:
			[integerArg > 0
				ifTrue: [0 - ((0 - integerRcvr) // integerArg)]
				ifFalse: [(0 - integerRcvr) // (0 - integerArg)]]! !

!InterpreterPrimitives methodsFor: 'string primitives' stamp: 'sk 2/15/2018 23:09'!
rawCompare: string1 length: strLength1 with: string2 length: strLength2 accessBlock: accessBlock
	| c1 c2 min |
	<inline: true> "needs to be forced else slang does not inline it by default"
	min := strLength1 min: strLength2.
	0 to: min-1 do: 
		[:i | c1 := accessBlock value: string1 value: i.
			c2 := accessBlock value: string2 value: i.
			c1 = c2 ifFalse: [^c1 - c2]].
	^strLength1 - strLength2


	
	
	! !

!InterpreterPrimitives methodsFor: 'simulation support' stamp: 'eem 3/17/2017 11:55'!
sHEAFn
	<doNotGenerate>
	^true! !

!InterpreterPrimitives methodsFor: 'plugin primitive support' stamp: 'eem 10/18/2020 15:57'!
setFullScreenFlag: value
	"stub for overriding by the interpeter"! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 5/27/2011 16:21'!
signalNoResume: aSemaphore
	"Signal the given semaphore from within the interpreter.  Used to serialize callbacks."
	| empty |
	<api>
	<inline: false>
	empty := self isEmptyList: aSemaphore. 
	empty ifFalse:
		[self putToSleep: (self removeFirstLinkOfList: aSemaphore) yieldingIf: true].
	^empty! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'nice 11/18/2016 22:31'!
signed32BitValueOf: oop
	"Convert the given object into an integer value.
	The object may be either a SmallInteger or a four-byte LargeInteger."
	<returnTypeC: #int>
	
	objectMemory hasSixtyFourBitImmediates
		ifTrue:
			[(objectMemory isIntegerObject: oop) ifTrue:
				[| value64 |
				 value64 := objectMemory integerValueOf: oop.
				 (self cCode: [(self cCoerceSimple: value64 to: #int) ~= value64]
						inSmalltalk: [value64 >> 31 ~= 0 and: [value64 >> 31 ~= -1]]) ifTrue:
					[self primitiveFail. value64 := 0].
				 ^value64].
			self primitiveFail.
		 	^0]
		ifFalse:
			[(objectMemory isIntegerObject: oop) ifTrue:
				[^objectMemory integerValueOf: oop].
			^self noInlineSigned32BitValueGutsOf: oop]! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'nice 4/3/2016 09:29'!
signed64BitValueOf: oop
	"Convert the given object into an integer value.
	 The object may be either a positive SmallInteger or a eight-byte LargeInteger."
	| sz value negative ok magnitude |
	<inline: false>
	<returnTypeC: #sqLong>
	<var: #value type: #sqLong>
	<var: #magnitude type: #usqLong>
	(objectMemory isIntegerObject: oop) ifTrue:
		[^self cCoerce: (objectMemory integerValueOf: oop) to: #sqLong].

	(objectMemory isNonIntegerImmediate: oop) ifTrue:
		[self primitiveFail.
		 ^0].

	ok := objectMemory isClassOfNonImm: oop
					equalTo: (objectMemory splObj: ClassLargePositiveInteger)
					compactClassIndex: ClassLargePositiveIntegerCompactIndex.
	ok
		ifTrue: [negative := false]
		ifFalse:
			[negative := true.
			 ok := objectMemory isClassOfNonImm: oop
							equalTo: (objectMemory splObj: ClassLargeNegativeInteger)
							compactClassIndex: ClassLargeNegativeIntegerCompactIndex.
			ok ifFalse:
				[self primitiveFail.
				 ^0]].
	sz := objectMemory numBytesOfBytes: oop.
	sz > (self sizeof: #sqLong) ifTrue:
		[self primitiveFail.
		 ^0].

	"self cppIf: SPURVM
		ifTrue:
			[""Memory is 8 byte aligned in Spur and oversized bytes are set to zero, so we can safely fetch 8 bytes""
			magnitude := objectMemory byteSwapped64IfBigEndian: (objectMemory fetchLong64: 0 ofObject: oop)]
		ifFalse:
			["sz > 4
				ifTrue: [magnitude := objectMemory byteSwapped64IfBigEndian: (objectMemory fetchLong64: 0 ofObject: oop)]
				ifFalse: [magnitude := self cCoerceSimple: (objectMemory byteSwapped32IfBigEndian: (objectMemory fetchLong32: 0 ofObject: oop)) to: #'unsigned int']"]".

	(negative
		ifTrue: [magnitude > 16r8000000000000000]
		ifFalse: [magnitude >= 16r8000000000000000])
			ifTrue: [self primitiveFail.
				^0].
	negative
		ifTrue: [value := 0 - magnitude]
		ifFalse: [value := magnitude].
	^value! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'nice 7/8/2016 01:58'!
signedMachineIntegerValueOf: oop
	"Answer a signed value of an integer up to the size of a machine word.
	The object may be either a positive SmallInteger or a LargeInteger of size <= word size."
	<returnTypeC: #'sqIntptr_t'>
	| negative ok bs value limit magnitude |
	<var: #value type: #sqInt>
	<var: #magnitude type: #'usqIntptr_t'>
	<var: #limit type: #'usqIntptr_t'>
	(objectMemory isIntegerObject: oop) ifTrue:
		[^objectMemory integerValueOf: oop].

	(objectMemory isNonIntegerImmediate: oop) ifTrue:
		[^self primitiveFail].

	ok := objectMemory isClassOfNonImm: oop
					equalTo: (objectMemory splObj: ClassLargePositiveInteger)
					compactClassIndex: ClassLargePositiveIntegerCompactIndex.
	ok
		ifTrue: [negative := false]
		ifFalse:
			[negative := true.
			 ok := objectMemory isClassOfNonImm: oop
							equalTo: (objectMemory splObj: ClassLargeNegativeInteger)
							compactClassIndex: ClassLargeNegativeIntegerCompactIndex.
			ok ifFalse: [^self primitiveFail]].
	bs := objectMemory numBytesOf: oop.
	bs > (self sizeof: #'usqIntptr_t') ifTrue:
		[^self primitiveFail].

	"self cppIf: SPURVM
		ifTrue:
			[""Memory is 8 byte aligned in Spur and oversized bytes are set to zero, so we can safely fetch 8 bytes""
			magnitude := objectMemory byteSwapped64IfBigEndian: (objectMemory fetchLong64: 0 ofObject: oop)]
		ifFalse:
			["((self sizeof: #'sqIntptr_t') = 8
			and: [bs > 4])
				ifTrue:
					[magnitude := objectMemory byteSwapped64IfBigEndian: (objectMemory fetchLong64: 0 ofObject: oop)]
				ifFalse:
					[magnitude := self cCoerceSimple: (objectMemory byteSwapped32IfBigEndian: (objectMemory fetchLong32: 0 ofObject: oop)) to: #'unsigned int']"]".

	limit := 1 asUnsignedIntegerPtr << ((self sizeof: #'sqIntptr_t') * 8 - 1).
	(negative
		ifTrue: [magnitude > limit]
		ifFalse: [magnitude >= limit])
			ifTrue: [self primitiveFail.
				^0].
	negative
		ifTrue: [value := 0 - magnitude]
		ifFalse: [value := magnitude].
	^value! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 10/27/2020 21:38'!
sizeOfAlienData: oop
	"Answer the start of the Alien's data or fail if oop is not an Alien."
	<api>
	<returnTypeC: #usqInt>
	(self is: oop KindOfClass: (objectMemory splObj: ClassAlien)) ifFalse:
		[self primitiveFailFor: PrimErrBadArgument.
		 ^0].
	^(objectMemory sizeFieldOfAlien: oop) abs! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 1/4/2017 08:48'!
startOfAlienData: oop
	"Answer the start of the Alien's data or fail if oop is not an Alien."
	<api>
	<returnTypeC: #'void *'>
	(self is: oop KindOfClass: (objectMemory splObj: ClassAlien)) ifFalse:
		[self primitiveFailFor: PrimErrBadArgument.
		 ^0].
	^self cCoerceSimple: ((self isDirectAlien: oop)
						 	ifTrue: [oop + objectMemory baseHeaderSize + objectMemory bytesPerOop]
							ifFalse: [objectMemory longAt: oop + objectMemory baseHeaderSize + objectMemory bytesPerOop])
			to: #'void *'! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 5/24/2011 09:12'!
success: successBoolean
	"Set the state of the primitive failure code/success flag, iff successBoolean
	 is false. If primFailCode is non-zero a primitive has failed.  If primFailCode
	 is greater than one then its value indicates the reason for failure."

	"Use returnTypeC: #sqInt because that's the way it is defined in sq.h.
	 Use no explicit return so that Slang doesn't fail an inlining type-check when
	 a primitive with return type void uses ^self success: false to exit."
	<returnTypeC: #sqInt>
	<inline: true>
	successBoolean ifFalse:
		["Don't overwrite an error code that has already been set."
		 self successful ifTrue:
			[primFailCode := 1]]! !

!InterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 5/24/2011 09:14'!
successful
	"Answer the state of the primitive failure code/success flag.  If
	 primFailCode is non-zero a primitive has failed.  If primFailCode
	 is greater than one then its value indicates the reason for failure."
	<inline: true>
	
	"In C, non-zero is true, so avoid computation by using not the C version."
	^self cCode: [primFailCode not] inSmalltalk: [primFailCode = 0]! !

!InterpreterPrimitives methodsFor: 'I/O primitive support' stamp: 'eem 2/10/2021 23:28'!
traceInputEvent: evtBuf
	<var: #evtBuf declareC: 'sqIntptr_t evtBuf[8]'>
	| eventTypeNames |
	eventTypeNames := self
							cCoerce: #('None' 'Mouse' 'Keyboard' 'DragDropFiles' 'Menu' 'Window' 'Complex' 'MouseWheel' 'Plugin')
							to: #'char **'.
	'Event%s/%d @ %u\t\t%d/%x %d/%x\n\t%d/%x %d/%x\t %d/%x %p\n'
		f: #stderr
		printf: {	((evtBuf at: 0) between: 0 and: 8)
						ifTrue: [eventTypeNames at: (evtBuf at: 0)]
						ifFalse: ['?'].
					evtBuf at: 0.
					evtBuf at: 1. "timestamp"
					evtBuf at: 2. evtBuf at: 2. evtBuf at: 3. evtBuf at: 3.
					evtBuf at: 4. evtBuf at: 4. evtBuf at: 5. evtBuf at: 5.
					evtBuf at: 6. evtBuf at: 6.
					(evtBuf at: 7) asVoidPointer } "windowIndex"! !

InterpreterPrimitives subclass: #StackInterpreter
	instanceVariableNames: 'currentBytecode bytecodeSetSelector localFP localIP localSP stackLimit stackPage stackPages method instructionPointer stackPointer framePointer localReturnValue localAbsentReceiver localAbsentReceiverOrZero extA extB numExtB primitiveFunctionPointer methodCache nsMethodCache atCache lkupClassTag lkupClass methodDictLinearSearchLimit highestRunnableProcessPriority reenterInterpreter nextWakeupUsecs nextPollUsecs inIOProcessEvents interruptKeycode interruptPending savedWindowSize imageHeaderFlags fullScreenFlag sendWheelEvents deferDisplayUpdates pendingFinalizationSignals extraVMMemory interpreterProxy showSurfaceFn primitiveTable primitiveAccessorDepthTable externalPrimitiveTable externalPrimitiveTableFirstFreeIndex overflowedPage extraFramesToMoveOnOverflow globalSessionID jmpBuf jmpDepth suspendedCallbacks suspendedMethods numStackPages desiredNumStackPages desiredEdenBytes classNameIndex thisClassIndex metaclassNumSlots interruptCheckChain suppressHeartbeatFlag breakSelector breakSelectorLength breakLookupClassTag tempOop tempOop2 metaAccessorDepth theUnknownShort the2ndUnknownShort imageFloatsBigEndian maxExtSemTabSizeSet lastMethodCacheProbeWrite gcSemaphoreIndex classByteArrayCompactIndex checkedPluginName nativeSP nativeStackPointer lowcodeCalloutState shadowCallStackPointer displayBits displayWidth displayHeight displayDepth statForceInterruptCheck statStackOverflow statStackPageDivorce statCheckForEvents statProcessSwitch statIOProcessEvents statPendingFinalizationSignals statIdleUsecs debugCallbackPath debugCallbackReturns debugCallbackInvokes primitiveDoMixedArithmetic'
	classVariableNames: 'AccessModifierPrivate AccessModifierProtected AccessModifierPublic AltBytecodeEncoderClassName AltLongStoreBytecode AlternateHeaderHasPrimFlag AlternateHeaderIsOptimizedFlag AlternateHeaderNumLiteralsMask AtCacheFixedFields AtCacheFmt AtCacheMask AtCacheOop AtCacheSize AtCacheTotalSize AtPutBase BytecodeEncoderClassName BytecodeTable CacheProbeMax DirBadPath DirEntryFound DirNoMoreEntries DumpStackOnLowSpace EnclosingMixinIndex EnclosingObjectIndex EnforceAccessControl ExternalCallLiteralFlagsIndex ExternalCallLiteralFunctionNameIndex ExternalCallLiteralModuleNameIndex ExternalCallLiteralTargetFunctionIndex FailImbalancedPrimitives LongStoreBytecode MaxExternalPrimitiveTableSize MaxJumpBuf MaxPrimitiveIndex MaxQuickPrimitiveIndex MethodHeaderArgCountShift MethodHeaderFlagBitPosition MethodHeaderTempCountShift MixinIndex PrimitiveTable ReturnToInterpreter StackPageReachedButUntraced StackPageTraceInvalid StackPageTraced StackPageUnreached V3PrimitiveBitsMask'
	poolDictionaries: 'VMBasicConstants VMBytecodeConstants VMMethodCacheConstants VMObjectIndices VMSpurObjectRepresentationConstants VMSqueakClassIndices VMStackFrameOffsets'
	category: 'VMMaker-Interpreter'!

!StackInterpreter commentStamp: 'eem 12/7/2017 11:01' prior: 0!
This class is a complete implementation of the Smalltalk-80 virtual machine, derived originally from the Blue Book specification but quite different in some areas.  This VM supports Closures but *not* old-style BlockContexts.

It has been modernized with 32-bit (and indeed 64-bit) pointers, better management of Contexts (see next item), and attention to variable use that allows the CCodeGenerator (qv) to keep, eg, the instruction pointer and stack pointer in registers as well as keeping most simple variables in a global array that seems to improve performance for most platforms.

The VM does not use Contexts directly.  Instead Contexts serve as proxies for a more conventional stack format that is invisible to the image.  There is considerable explanation at http://www.mirandabanda.org/cogblog/2009/01/14/under-cover-contexts-and-the-big-frame-up.  The VM maintains a fixed-size stack zone divided into pages, each page being capable of holding several method/block activations.  A send establishes a new frame in the current stack page, a return returns to the previous frame.  This eliminates allocation/deallocation of contexts and the moving of receiver and arguments from caller to callee on each send/return.  Contexts are created lazily when an activation needs a context (creating a block, explicit use of thisContext, access to sender when sender is a frame, or linking of stack pages together).  Contexts are either conventional and heap-resident ("single") or "married" and serve as proxies for their corresponding frame or "widowed", meaning that their spouse frame has been returned from (died).  A married context is specially marked (more details in the code) and refers to its frame.  Likewise a married frame is specially marked and refers to its context.

In addition to SmallInteger arithmetic and Floats, the VM supports logic on 32-bit (and 64-bit) PositiveLargeIntegers, thus allowing it to simulate itself much more efficiently than would otherwise be the case.

StackInterpreter and subclasses support multiple memory managers.  Currently there are two.  NewMemoryManager is a slightly refined version of ObjectMemory, and is the memory manager and garbage collector for the original Squeak object representation as described in "Back to the Future The Story of Squeak, A Practical Smalltalk Written in Itself", see http://ftp.squeak.org/docs/OOPSLA.Squeak.html.  Spur is a faster, more regular object representation that is designed for more performance and functionality, and to have a common header format for both 32-bit and 64-bit versions.  You can read about it in SpurMemoryManager's class comment.  There is also a video of a presentation at ESUG 2014 (https://www.youtube.com/watch?v=k0nBNS1aHZ4), along with slides (http://www.slideshare.net/esug/spur-a-new-object-representation-for-cog?related=1).

StackInterpreter is designed toi be subclassed by CoInterpreter et al.  CoInterpreter cooperates with a bytecode to machine-code JIT compiler, for much higher performance

Instance Variables

atCache
	- a small array holding tuples of oop => format,fixedFields,size to speed up at:[put:].  Unused in CoInterpreter

breakLookupClassTag
	- if non-nil, the class tag to break on sends, breaking (calling warning) if there's a match

breakSelector
	- if breakSelectorLength ~= 0, a pointer to a string to compare messageSelector against on send, breaking (calling warning) if there's a match

breakSelectorLength
	- if > 0, breakSelector is checked on each send.  If < 0, breakSelector is checked on each MNU

bytecodeSetSelector
	- if multiple bytecode sets are in use this is the offset (0 or 256) for the current bytecode set

checkAllocFiller
	- a variable used to fill eden to allow checking for overwriting the end of an object

checkedPluginName
	- the oop of a symbol which, when gcmode is GCCheckPrimCall, can check for heap memory leaks after a specific primitive has been called

classByteArrayCompactIndex
	- the compact class index of the class ByteArray, if appropriate

classNameIndex
	- the slot index of name in a class object, used or debug printing, derived from class Array at start-up

currentBytecode
	- the value of the currentBytecode

debugCallbackInvokes
	- a counter used to debug the old (Python) callback support

debugCallbackPath
	- a variable used to hold the path through the new callback support

debugCallbackReturns
	- a counter used to debug the old (Python) callback support; it should match debugCallbackInvokes

deferDisplayUpdates
	- the variable that implements primitive 126, primitiveDeferDisplayUpdates

desiredEdenBytes
	- the desired size of eden, used at image load and snapshot time

desiredNumStackPages
	- the desired number of stack pages, used at image load and snapshot time

displayBits
	- either the first indexable field of the current display object, or the handle of the current display surface

displayDepth
	- pixel depth of the current display object or surface

displayHeight
	- height in pixels of the current display object or surface

displayWidth
	- width in pixels of the current display object or surface

extA
	- if a prefix btecode set, such as SstaV1, is in use, the accumulator for extension A

extB
	- if a prefix btecode set, such as SstaV1, is in use, the accumulator for extension B

externalPrimitiveTable
	- xxxxx

externalPrimitiveTableFirstFreeIndex
	- a small array cacheing the target function pointer for external calls (primitiveExternalCall)

extraFramesToMoveOnOverflow
	- a variable used to prevent thrashing on sends and returns at the end of a page (see overflowedPage)

extraVMMemory
	- in Spur, free heap to be allocated at startup.  In V3, C memory to reserve for malloc at startup (i.e. malloc extraVMMemory, alloc heap, free extraVMMemory)

framePointer
	- the pointer to the current frame when primitives (or in the CoInterpreter, when machine-code <-> interpreter transitions) are invoked (c.f. localFP)

fullScreenFlag
	- true if the system is in full-screen mode

gcSemaphoreIndex
	- if non-zero, the index in the semaphore table of a semaphore to signal after any GC

globalSessionID
	- a non-zero number (derived from the time) identifying the current session

highestRunnableProcessPriority
	- a tide mark in the runnableProcesses array, used to reduce the time spent scanning the array looking for the highest priority runnable process.  This variable is set high when the priority of the highest process is unknown, and set to the highest found when the array is scanned.  Hence most of the time the nused entries in the array are not scanned.

imageFloatsBigEndian
	- true if the order of data in floats is big-endian

imageHeaderFlags
	- the flags stored in the image at load time or to be stored in the image at snapshto time.  These flags identify various options, sich as whether floats are in platfrm order.

inIOProcessEvents
	- true if the system is receiving events in ioProcessEvents, used to avoid recursion when receiuving events via callbacks

instructionPointer
	- the pointer to the current bytecode when primitives (or in the CoInterpreter, when machine-code <-> interpreter transitions) are invoked (c.f. localIP)

interpreterProxy
	- the table of functions used to export interpreter and memory manager functions to plugin primitives

interruptCheckChain
	- if non nil a funciton pointer to be called when a user interrupt event is received, used by plugin primitives to abort long-running calls on interrupt

interruptKeycode
	- the keycode of the event signifying a user interrupt

interruptPending
	- a variable set by ioProcessEvents if a user interrupt is pending

jmpBuf
	- a small array of jmpbufs (for setjmp/longjmp) used to manage returning in the iold (Python) callback support (see suspendedCallbacks et al)

jmpDepth
	- the index of the next free entry in jmpBuf

lastMethodCacheProbeWrite
	- the index of the last write into the method lookup cache, used by primitiveExternalCall to rewrite the method cache entry to call the target function, avoiding going through primitiveExternalCall on every plugin primitive dispatch

lkupClass
	- the implicit class argument to the class hierarchy looup routines such as lookupOrdinarySend

lkupClassTag
	- the implicit class tag argument to message send lookup

localAbsentReceiver
	- Newspeak, the absent receiver in any Newspeak absent receiver send

localAbsentReceiverOrZero
	- Newspeak, if non-zero, the absent receiver in any Newspeak outer absent receiver send

localFP
	- the frame pointer in the fully-inlined interpret routine, hopfully allowing a C compiler to assign localFP to a register variable (c.f. framePointer)

localIP
	- the instruction pointer in the fully-inlined interpret routine, hopfully allowing a C compiler to assign localFP to a register variable (c.f. instructionPointer)

localReturnValue
	- the value being returned in return bytecodes

localSP
	- the stack pointer in the fully-inlined interpret routine, hopfully allowing a C compiler to assign localFP to a register variable (c.f. framePointer)

longRunningPrimitiveCheckMethod
	- a variable used to monitor long-running primtiives via primitiveLongRunningPrimitive

longRunningPrimitiveCheckSemaphore
	- a variable used to monitor long-running primtiives via primitiveLongRunningPrimitive

longRunningPrimitiveCheckSequenceNumber
	- a variable used to monitor long-running primtiives via primitiveLongRunningPrimitive

longRunningPrimitiveGCUsecs
	- a variable used to monitor long-running primtiives via primitiveLongRunningPrimitive

longRunningPrimitiveSignalUndelivered
	- a variable used to monitor long-running primtiives via primitiveLongRunningPrimitive

longRunningPrimitiveStartUsecs
	- a variable used to monitor long-running primtiives via primitiveLongRunningPrimitive

longRunningPrimitiveStopUsecs
	- a variable used to monitor long-running primtiives via primitiveLongRunningPrimitive

lowcodeCalloutState
	- Lowcode, the state used to marshall an outgoing Lowcode call

maxExtSemTabSizeSet
	- a flag stating whether the maximum size of the external semaphore table has been set or not

metaAccessorDepth
	- Spur, the accessor depth of a primitive being invoked by one of the meta primitive primitives, primitiveDoPrimitiveWithArgs and primitiveDoNamedPrimitiveWithArgs

metaclassNumSlots
	- teh size of a Metaclass instanc, set at startup, used for debug printing

method
	- the oop of the current method

methodCache
	- a small array holding tuples of selector x class => method,functionPointer to speed up message send (avoid class hierarchy search on send)

methodDictLinearSearchLimit
	- a variable used to control searching of method dictionaries.  Dictionaries of size within the limit are searched linearly instead of hashed.  This can be used either to speed up search in small dictionaries or to turn off hashed lookup wen hashing needs to be changed

nativeSP
	- Lowcode, the native (C) stack pointer (?; then why nativeStackPointer?)

nativeStackPointer
	- Lowcode, the native (C) stack pointer (?; then why nativeSP?)

nextPollUsecs
	- if non-zero, the micosecond clock value at which to next poll for events, used to control event polling rate

nextWakeupUsecs
	- if non-zero, the micosecond clock value at which the next delay expires

nsMethodCache
	- Newspeak, a small array hilding tuples of selector x class tag x calling method x privacy code => target method,primitive function,actual receiver, used to speed up Newspeak absent receiver sends

numExtB
	- the number of bytes in extB, used to correctly maintain extB, given that it is unsigned

numStackPages
	- the number of stack pages being used

overflowedPage
	- the stack page that a send last overflowed, used to avoid thrashing on send and return at the end of a stack page (see extraFramesToMoveOnOverflow)

pendingFinalizationSignals
	- a variable used to control signalling the finalizationSemahpore

primitiveAccessorDepthTable
	- Spur, the accessor depths for the iundexed primitives, used to determine how much state to scan loking for forwarders on primitive failure

primitiveFunctionPointer
	- the function pointer (in simulation, the selector symbol) of the current primitive, nil for newMethod if there is no primitive

primitiveTable
	- the table holding the functions (in simulation the selector symbols) for the indexed primitives

savedWindowSize
	- the display size to write into a snapshot.  Avoids writing a zero window size into a snapshot when an image is loaded headless.  

shadowCallStackPointer
	- Lowcode, the shadow stack pointer, used when marshalling Lowcode external calls

showSurfaceFn
	- if non-zero, the function to call to update the display when the current display is a surface

stackLimit
	- the variable marking the end of the current stack page, also used to break out of interpretation to check for events.  When marking the end of a page, it enables a send to create an activation on a new page,  When set so that any send would fail the stack limit check, causes a check for events on stack page overflow

stackPage
	- the current stack page

stackPages
	- the set of stack pages

stackPointer
	- the pointer to the top of stack in the current frame when primitives (or in the CoInterpreter, when machine-code <-> interpreter transitions) are invoked (c.f. localSP)

statCheckForEvents
	- the count of checkForEventsMaybeContextSwitch: calls

statForceInterruptCheck
	- the count of forceInterruptCheck calls

statIOProcessEvents
	- the count of ioProcessEvents calls

statIdleUsecs
	- the count of microseconds at which the system is at idle (a bad idea; this should be replaced by its inverse; asking the OS how long the VM process has been active)

statPendingFinalizationSignals
	- the count of pendingFinalizationSignals

statProcessSwitch
	- the count of process switches

statStackOverflow
	- the count of stack overflows

statStackPageDivorce
	- the cound of stack page divorces (a stack page divorce occurs when a stack page must be freed for use)

suppressHeartbeatFlag
	- a flag controlling whether the heatbeat is in effect; the heartbeat is used to set stackLimit periodically to cause break out to chedck for events (see stackLimit)

suspendedCallbacks
	- a small array holding the processes suspended when the old callback machinery is in use (see jmpBuf et al)

suspendedMethods
	- a small array holding the methods of the processes suspended when the old callback machinery is in use

tempOop
	- if non-zero, an oop that may move during primitive execution

tempOop2
	- if non-zero, an oop that may move during primitive execution

the2ndUnknownShort
	- the value of a field in the image header that is unused in the StackInterpreter but used in the CoInterpreter (and hence to be preserved in snapshots)

theUnknownShort
	- the value of a field in the image header that is unused in the StackInterpreter but used in the CoInterpreter (and hence to be preserved in snapshots)

thisClassIndex
	- the index in a Metaclass of thisClass (the instance of the metaclass) used in debug printing!

StackInterpreter subclass: #ImageLeakChecker
	instanceVariableNames: 'verbose'
	classVariableNames: ''
	poolDictionaries: 'VMObjectIndices VMSpurObjectRepresentationConstants VMSqueakClassIndices'
	category: 'VMMaker-Utilities'!

!ImageLeakChecker commentStamp: 'eem 11/15/2019 15:11' prior: 0!
An ImageLeakChecker loads an image and runs the leak checker, answering true if there are no leaks.  When used stand alone it takes one argument, --verbose, which if set causes the ImageLeakChecker to print information on each leak.

Instance Variables
verbose <Boolean> whether to print details, or simply to answer an exist status.!

!ImageLeakChecker class methodsFor: 'translation' stamp: 'eem 11/14/2019 12:08'!
additionalHeadersDo: aBinaryBlock
	"Nothing to do here. These are not the droids you are looking for..."! !

!ImageLeakChecker class methodsFor: 'translation' stamp: 'eem 11/28/2019 12:08'!
ancilliaryClasses
	"Answer any extra classes to be included in the translation."
	^super ancilliaryClasses, { SpurLeakCheckingSegmentManager }! !

!ImageLeakChecker class methodsFor: 'translation' stamp: 'eem 11/15/2019 09:27'!
declareCVarsIn: aCCodeGenerator
	self declareInterpreterVersionIn: aCCodeGenerator defaultName: 'ImageChecker'.
	aCCodeGenerator
		var: #interpreterProxy  type: #'struct VirtualMachine*'.
	"Do this to avoid redefining interp.h when we generate the image checker.  i.e. agree with StackInterpereter class's version above."
	aCCodeGenerator addConstantForBinding: (self bindingOf: #MULTIPLEBYTECODESETS).
	#('checkAllocFiller' 'primitiveTable' 'primitiveAccessorDepthTable' 'stackPages') do:
		[:surplus| aCCodeGenerator removeVariable: surplus]! !

!ImageLeakChecker class methodsFor: 'translation' stamp: 'eem 11/14/2019 14:24'!
doInlining
	^#asSpecifiedOrQuick! !

!ImageLeakChecker class methodsFor: 'translation' stamp: 'eem 11/28/2019 12:08'!
isAcceptableAncilliaryClass: aClass
	^aClass == CogStackPage
	 or: [aClass == SpurLeakCheckingSegmentManager
	 or: [self objectMemoryClass ancilliaryClasses includes: aClass]]! !

!ImageLeakChecker class methodsFor: 'translation' stamp: 'eem 11/14/2019 12:31'!
namesOfVariablesToLocalize
	^#()! !

!ImageLeakChecker class methodsFor: 'translation' stamp: 'eem 11/15/2019 12:36'!
preGenerationHook: aCCodeGen
	"Attempt to nuke everything we don't need in our hierarchy.  We will leave the
	 pruning of most of methods in the memory manager hierarchy to the code generator."
	| interpreterSelectors toBeRemoved |
	toBeRemoved := Set withAll: (interpreterSelectors := StackInterpreter selectors, InterpreterPrimitives selectors).

	"Now make sure we *don't* remove any of the following.  These we need."
	{self selectors.
	  #(addressCouldBeClassObj: checkAssumedCompactClasses fetchStackPointerOf:
		highBit: is:KindOfClass: includesBehavior:ThatOf: initializeExtraClassInstVarIndices initializeInterpreter:
		literalCountOfAlternateHeader: maybeSelectorOfMethod: maybeFixClonedCompiledMethod:
		objCouldBeClassObj: primitiveFail primitiveFailFor: successful superclassOf:).
	  (StackInterpreter methodsInCategory: #'debug printing') reject:
		[:s1| #('Context' 'Frame' 'Proc' "Procs and Process" 'Stack') anySatisfy: [:s2| s1 includesSubstring: s2]].
	  StackInterpreter methodsInCategory: #'compiled methods'.
	  StackInterpreter methodsInCategory: #'frame access'.
	  StackInterpreter methodsInCategory: #'image save/restore'.
	 interpreterSelectors select: [:s| s includesSubstring: 'Alien'].
	 "the ^self methods are usually no-op stubs overridden in CoInterpreter; we want these."
	  interpreterSelectors select: [:s| ((StackInterpreter whichClassIncludesSelector: s) >> s) isReturnSelf].
	  "And include all the isFoo methods..."
	  interpreterSelectors select: [:s| s size > 2 and: [(s beginsWith: 'is') and: [s third isUppercase]]]} do:
		[:selectors|
		 toBeRemoved removeAll: (selectors intersection: toBeRemoved)].

	"These we explicitly prune; they're used by InterpreterProxy and/or the Cogit and/or pieces of machinery we don't want"
	toBeRemoved addAll: #(indexOf:in: instantiateClass:indexableSize: integerValueOf:
							maybeMethodClassOf:seemsToBeInstantiating: remapObj:).

	aCCodeGen pruneMethods: toBeRemoved.
	aCCodeGen sortedExportMethods do:
		[:tm| tm export: false]! !

!ImageLeakChecker class methodsFor: 'translation' stamp: 'eem 11/15/2019 12:20'!
preambleCCode

	"readImageFromFile:HeapSizeStartingAt is declared public in sq.h, so it has to be exported,
	 but some weird bug with clang clang-1000.11.45.5 makes it complian that _m is undecared in this:

		static char _m[] = "";
		void* vm_exports[][3] = {
		    {(void*)_m, ''readImageFromFileHeapSizeStartingAt'', (void*)readImageFromFileHeapSizeStartingAt},
		    {NULL, NULL, NULL}
		};

	 so we hack around it by redefining readImageFromFileHeapSizeStartingAt after sq.h is included and keep
	 our version of it private, and hence have no exports, avoiding generating the offending code above."
	
	^super preambleCCode,
		('\#define readImageFromFileHeapSizeStartingAt readImageFromFile\',
		 '\void error(char *msg) {\  warning(msg);\  abort();\}\\') withCRs! !

!ImageLeakChecker class methodsFor: 'translation' stamp: 'eem 3/9/2020 16:16'!
prepareToBeAddedToCodeGenerator: aCodeGen
	aCodeGen
		removeMethodForSelector: #become:with:;
		removeMethodForSelector: #fullGC;
		removeMethodForSelector: #growOldSpaceByAtLeast:;
		removeMethodForSelector: #runLeakCheckerFor:;
		removeMethodForSelector: #runLeakCheckerForFreeSpace:ignoring:;
		removeMethodForSelector: #runLeakCheckerFor:excludeUnmarkedObjs:classIndicesShouldBeValid:;
		removeMethodForSelector: #inLineRunLeakCheckerFor:excludeUnmarkedObjs:classIndicesShouldBeValid:;
		removeMethodForSelector: #scheduleScavenge;
		removeMethodForSelector: #tenuringIncrementalGC! !

!ImageLeakChecker class methodsFor: 'translation' stamp: 'eem 11/14/2019 12:10'!
requiredMethodNames: options
	"Answer the list of method names that should be retained for export or other support reasons"
	^#(#'main:_:')! !

!ImageLeakChecker class methodsFor: 'translation' stamp: 'eem 11/14/2019 12:00'!
sourceFileName
	"Answer the filename for the core interpreter"

	^'validImage.c'! !

!ImageLeakChecker methodsFor: 'implementation' stamp: 'eem 11/14/2019 10:42'!
asCommandLineArgument: aString
	<inline: #always>
	^PharoVM ifTrue: ['--', aString] ifFalse: ['-', aString]! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 10:14'!
couldBeFramePointer: fp
	<var: #pointer type: #'void *'>
	<inline: #always>
	^false! !

!ImageLeakChecker methodsFor: 'printing' stamp: 'eem 11/15/2019 12:45'!
cr
	verbose ifTrue: [super cr]! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 10:23'!
defaultNumStackPages
	^0! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 11:23'!
dummyReferToProxy! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/14/2019 21:07'!
fireEphemeron: ephemeron! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 08:33'!
fireFinalization: thing! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/14/2019 20:58'!
freeUntracedStackPages! !

!ImageLeakChecker methodsFor: 'implementation' stamp: 'eem 11/15/2019 12:43'!
growOldSpaceByAtLeast: minAmmount
	self error: 'growOldSpaceByAtLeast: was invoked; should not happen!!'.
	^false! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 08:32'!
initStackPageGC! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/14/2019 20:44'!
initialCleanup! !

!ImageLeakChecker methodsFor: 'initialization' stamp: 'eem 11/15/2019 08:52'!
initialize
	verbose := false! !

!ImageLeakChecker methodsFor: 'initialization' stamp: 'eem 11/28/2019 12:04'!
initializeInterpreter: bytesToShift
	objectMemory segmentManager noteValidHeapExtent: bytesToShift.
	super initializeInterpreter: bytesToShift! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 08:33'!
interpreterAllocationReserveBytes
	^0! !

!ImageLeakChecker methodsFor: 'implementation' stamp: 'eem 11/15/2019 12:22'!
ioExitWithErrorCode: ec
	<inline: true>
	self exit: ec! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 12:05'!
ioLoadFunction: functionName From: moduleName
	<api>
	<returnTypeC: #'void *'>
	<var: #functionName type: #'char *'>
	<var: #moduleName type: #'char *'>
	^nil! !

!ImageLeakChecker methodsFor: 'implementation' stamp: 'eem 11/15/2019 12:42'!
ioMSecs
	^1! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 10:45'!
isFree: thePage
	<var: 'thePage' type: #'StackPage *'>
	^true! !

!ImageLeakChecker methodsFor: 'api' stamp: 'eem 11/28/2019 12:18'!
main: argc _: argv
	<var: 'argc' type: #int>
	<var: 'argv' declareC: 'char *argv[]'>
	<returnTypeC: #int>
	<api>
	| forVersion argi ifh ok okSwizzle |
	<var: 'ifh' type: #'FILE *'>
	argi := 1.
	forVersion := false.
	[argi < argc
	 and: [((argv at: argi) at: 0) = $-]] whileTrue:
		[(self strcmp: (self asCommandLineArgument: 'version') _: (argv at: argi)) = 0 ifTrue:
			['%s\n' printf: (self cCode: 'interpreterVersion' inSmalltalk: ['don''t ask silly questions']).
			 forVersion := true].
		 (self strcmp: (self asCommandLineArgument: 'verbose') _: (argv at: argi)) = 0 ifTrue:
			[verbose := true].
		 argi := argi + 1].
	(argc = argi and: [forVersion]) ifTrue:
		[^0].

	(argc ~= (argi + 1)
	 or: [(ifh := self sqImageFileOpen: (argv at: argi) _: 'rb') = 0]) ifTrue:
		['Usage: %s [%s] [%s] imageFileName\n' printf: {(argv at: 0). self asCommandLineArgument: 'verbose'. self asCommandLineArgument: 'version'}.
		 ^self cCode: [1] inSmalltalk: [false]].

	self readImageFromFile: ifh HeapSize: 0 StartingAt: 0.
	self sqImageFileClose: ifh.
	okSwizzle := objectMemory segmentManager detectedInvalidOop not.
	ok := objectMemory runLeakCheckerForResult: GCModeFull.
	(verbose and: [okSwizzle and: [ok]]) ifTrue:
		['Image %s is free of leaks\n' printf: {(argv at: argi)}].
	^self cCode: [ok ifTrue: [0] ifFalse: [2]] inSmalltalk: [ok]! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 08:28'!
mapInterpreterOops! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/14/2019 21:09'!
markAndTraceAndMaybeFreeStackPages: fullGCFlag! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/14/2019 20:44'!
markAndTraceInterpreterOops: fullGCFlag! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/14/2019 21:09'!
markAndTraceUntracedReachableStackPages! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 10:27'!
mostRecentlyUsedPage
	<returnTypeC: #'StackPage *'>
	<inline: #always>
	^nil! !

!ImageLeakChecker methodsFor: 'implementation' stamp: 'eem 11/15/2019 12:12'!
plugInNotifyUser: msg
	"This is used in macros in sq.h, sigh..."
	<api>
	<returnTypeC: #int>
	<var: 'msg' type: #'char *'>
	self error: msg! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/14/2019 21:06'!
postBecomeAction: theBecomeEffectsFlags! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/14/2019 21:07'!
postGCAction: gcModeArg! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/14/2019 21:06'!
preBecomeAction! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/14/2019 21:07'!
preGCAction: gcModeArg! !

!ImageLeakChecker methodsFor: 'printing' stamp: 'eem 11/14/2019 12:02'!
print: s
	verbose ifTrue: [super print: s]! !

!ImageLeakChecker methodsFor: 'printing' stamp: 'eem 11/14/2019 12:03'!
printHex: n
	verbose ifTrue: [super printHex: n]! !

!ImageLeakChecker methodsFor: 'printing' stamp: 'eem 11/14/2019 12:04'!
printNum: n
	verbose ifTrue: [super printNum: n]! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 10:30'!
runLeakCheckerFor: gcModes! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 10:53'!
runLeakCheckerFor: gcModes excludeUnmarkedObjs: excludeUnmarkedObjs classIndicesShouldBeValid: classIndicesShouldBeValid! !

!ImageLeakChecker methodsFor: 'implementation' stamp: 'eem 3/9/2020 16:16'!
runLeakCheckerForFreeSpace: gcModes ignoring: anOop! !

!ImageLeakChecker methodsFor: 'implementation' stamp: 'eem 11/15/2019 16:55'!
runLeakCheckerForResult: gcModes
	objectMemory clearLeakMapAndMapAccessibleObjects.
	^objectMemory checkHeapIntegrity: false classIndicesShouldBeValid: true! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 11:18'!
scheduleScavenge! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 11:23'!
setMaxExtSemSizeTo: maxExtSemTabSize! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 08:27'!
setTraceFlagOnContextsFramesPageIfNeeded: aContext! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 12:00'!
sqGetInterpreterProxy
	<returnTypeC: 'struct VirtualMachine *'>
	^nil! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 10:37'!
stackLimitOffset
	^0! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 10:19'!
stackPageAt: index
	<returnTypeC: #'StackPage *'>
	<inline: #always>
	^nil! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 10:45'!
stackPageFor: pointer
	<returnTypeC: #'StackPage *'>
	<var: #pointer type: #'void *'>
	<inline: #always>
	^nil! !

!ImageLeakChecker methodsFor: 'accessing' stamp: 'eem 11/28/2019 12:19'!
verbose
	^verbose! !

!ImageLeakChecker methodsFor: 'no-op overrides' stamp: 'eem 11/15/2019 10:09'!
whereIsMaybeStackThing: thing
	<returnTypeC: #'char *'>
	^nil! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 3/16/2017 10:06'!
additionalHeadersDo: aBinaryBlock
	"Evaluate aBinaryBlock with the names and contents of
	 any additional header files that need to be generated."
	aBinaryBlock
		value: 'vmCallback.h'
		value: self vmCallbackHeader! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 2/10/2009 08:39'!
additionalSelectorTables
	^{ self primitiveTable }! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 06/07/2018 15:29'!
ancilliaryClasses
	"Answer any extra classes to be included in the translation."
	^{	self objectMemoryClass.
		VMCallbackContext.
		CogStackPages.
		InterpreterStackPages.
		CogStackPage }! !

!StackInterpreter class methodsFor: 'constants' stamp: ''!
bytecodeTable

	^ BytecodeTable! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 10/19/2020 10:22'!
declareCVarsIn: aCCodeGenerator
	| vmClass |
	self class == thisContext methodClass ifFalse: [^self]. "Don't duplicate decls in subclasses"
	vmClass := aCCodeGenerator vmClass. "Generate primitiveTable etc based on vmClass, not just StackInterpreter"
	aCCodeGenerator
		addHeaderFile: '<stdio.h> /* for printf */';
		addHeaderFile: '<stdlib.h> /* for e.g. alloca */';
		addHeaderFile: '<setjmp.h>';
		addHeaderFile: '<wchar.h> /* for wint_t */';
		addHeaderFile: '"vmCallback.h"';
		addHeaderFile: '"sqMemoryFence.h"';
		addHeaderFile: '"sqImageFileAccess.h"';
		addHeaderFile: '"sqSetjmpShim.h"';
		addHeaderFile: '"dispdbg.h"'.
	LowcodeVM ifTrue:
		[aCCodeGenerator addHeaderFile: '"sqLowcodeFFI.h"'].

	vmClass declareInterpreterVersionIn: aCCodeGenerator defaultName: 'Stack'.
	aCCodeGenerator
		var: #interpreterProxy  type: #'struct VirtualMachine*'.
	aCCodeGenerator
		declareVar: #sendTrace type: 'volatile int';
		declareVar: #byteCount type: #usqLong. "see dispdbg.h"
	"These need to be pointers or unsigned."
	self declareC: #(instructionPointer method newMethod)
		as: #usqInt
		in: aCCodeGenerator.
	"These are all pointers; char * because Slang has no support for C pointer arithmetic."
	self declareC: #(localIP localSP localFP stackPointer framePointer stackLimit breakSelector)
		as: #'char *'
		in: aCCodeGenerator.
	aCCodeGenerator
		var: #breakSelectorLength
		declareC: 'sqInt breakSelectorLength = MinSmallInteger'.
	self declareC: #(stackPage overflowedPage)
		as: #'StackPage *'
		in: aCCodeGenerator.
	aCCodeGenerator removeVariable: 'stackPages'.  "this is an implicit receiver in the translated code."
	"This defines bytecodeSetSelector as 0 if MULTIPLEBYTECODESETS
	 is not defined, for the benefit of the interpreter on slow machines."
	aCCodeGenerator addConstantForBinding: (self bindingOf: #MULTIPLEBYTECODESETS).
	MULTIPLEBYTECODESETS == false ifTrue:
		[aCCodeGenerator
			removeVariable: 'bytecodeSetSelector'].
	BytecodeSetHasExtensions == false ifTrue:
		[aCCodeGenerator
			removeVariable: 'extA';
			removeVariable: 'extB'].
	aCCodeGenerator
		var: #methodCache
		declareC: 'sqIntptr_t methodCache[MethodCacheSize + 1 /* ', (MethodCacheSize + 1) printString, ' */]'.
	NewspeakVM
		ifTrue:
			[aCCodeGenerator
				var: #nsMethodCache
				declareC: 'sqIntptr_t nsMethodCache[NSMethodCacheSize + 1 /* ', (NSMethodCacheSize + 1) printString, ' */]']
		ifFalse:
			[aCCodeGenerator
				removeVariable: #nsMethodCache;
				removeVariable: 'localAbsentReceiver';
				removeVariable: 'localAbsentReceiverOrZero'].
	AtCacheTotalSize isInteger ifTrue:
		[aCCodeGenerator
			var: #atCache
			declareC: 'sqInt atCache[AtCacheTotalSize + 1 /* ', (AtCacheTotalSize + 1) printString, ' */]'].
	aCCodeGenerator
		var: #primitiveTable
		declareC: 'void (*primitiveTable[MaxPrimitiveIndex + 2 /* ', (MaxPrimitiveIndex + 2) printString, ' */])(void) = ', vmClass primitiveTableString.
	vmClass primitiveTable do:
		[:symbolOrNot|
		(symbolOrNot isSymbol
		 and: [symbolOrNot ~~ #primitiveFail]) ifTrue:
			[(aCCodeGenerator methodNamed: symbolOrNot) ifNotNil:
				[:tMethod| tMethod returnType: #void]]].
	vmClass objectMemoryClass hasSpurMemoryManagerAPI
		ifTrue:
			[aCCodeGenerator
				var: #primitiveAccessorDepthTable
				type: 'signed char'
				sizeString: 'MaxPrimitiveIndex + 2 /* ', (MaxPrimitiveIndex + 2) printString, ' */'
				array: vmClass primitiveAccessorDepthTable]
		ifFalse:
			[aCCodeGenerator removeVariable: #primitiveAccessorDepthTable].
	aCCodeGenerator
		var: #displayBits type: #'void *'.
	self declareC: #(displayWidth displayHeight displayDepth) as: #int in: aCCodeGenerator.
	aCCodeGenerator
		var: #primitiveFunctionPointer
			declareC: 'void (*primitiveFunctionPointer)()';
		var: #externalPrimitiveTable
			declareC: 'void (*externalPrimitiveTable[MaxExternalPrimitiveTableSize + 1 /* ', (MaxExternalPrimitiveTableSize + 1) printString, ' */])(void)';
		var: #interruptCheckChain
			declareC: 'void (*interruptCheckChain)(void) = 0';
		var: #showSurfaceFn
			declareC: 'int (*showSurfaceFn)(sqIntptr_t, int, int, int, int)';
		var: #jmpBuf
			declareC: 'jmp_buf jmpBuf[MaxJumpBuf + 1 /* ', (MaxJumpBuf + 1) printString, ' */]';
		var: #suspendedCallbacks
			declareC: 'usqInt suspendedCallbacks[MaxJumpBuf + 1 /* ', (MaxJumpBuf + 1) printString, ' */]';
		var: #suspendedMethods
			declareC: 'usqInt suspendedMethods[MaxJumpBuf + 1 /* ', (MaxJumpBuf + 1) printString, ' */]'.

	self declareCAsUSqLong: #(nextPollUsecs nextWakeupUsecs longRunningPrimitiveGCUsecs
								longRunningPrimitiveStartUsecs longRunningPrimitiveStopUsecs
								"these are high-frequency enough that they're overflowing quite quickly on modern hardware"
								statProcessSwitch statIOProcessEvents statForceInterruptCheck
								statCheckForEvents statStackOverflow statStackPageDivorce
								statIdleUsecs)
		in: aCCodeGenerator.
	aCCodeGenerator var: #nextProfileTick type: #sqLong.
	aCCodeGenerator var: #reenterInterpreter type: 'jmp_buf'.
	LowcodeVM
		ifTrue:
			[aCCodeGenerator
				var: #lowcodeCalloutState type: #'sqLowcodeCalloutState*'.
			 self declareC: #(nativeSP nativeStackPointer shadowCallStackPointer)
				as: #'char *'
				in: aCCodeGenerator]
		ifFalse:
			[#(lowcodeCalloutState nativeSP nativeStackPointer shadowCallStackPointer) do:
				[:var| aCCodeGenerator removeVariable: var]].
		
	aCCodeGenerator
		var: #primitiveDoMixedArithmetic
		declareC: 'char primitiveDoMixedArithmetic = 1'.! !

!StackInterpreter class methodsFor: 'accessing class hierarchy' stamp: 'eem 9/30/2013 13:17'!
defaultObjectMemoryClass
	^NewObjectMemory! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 3/3/2009 10:22'!
implicitReturnTypeFor: aSelector
	"Answer the return type for methods that don't have an explicit return."
	^#void! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 4/12/2012 09:22'!
initializeAssociationIndex
	KeyIndex := 0.
	ValueIndex := 1! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 6/7/2018 15:21'!
initializeBytecodeTable
	"StackInterpreter initializeBytecodeTable"

	VMBytecodeConstants falsifyBytecodeSetFlags: InitializationOptions.
	BytecodeSetHasDirectedSuperSend := false.

	(InitializationOptions at: #bytecodeTableInitializer ifAbsent: nil) ifNotNil:
		[:initalizer| ^self perform: initalizer].

	NewspeakVM ifTrue:
		[^MULTIPLEBYTECODESETS
			ifTrue: [self initializeBytecodeTableForSqueakV3PlusClosuresNewspeakV4Hybrid]
			ifFalse: [self initializeBytecodeTableForNewspeakV4]].

	^MULTIPLEBYTECODESETS
		ifTrue: [self initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid]
		ifFalse: [self initializeBytecodeTableForSqueakV3PlusClosures]! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'nice 11/9/2019 10:17'!
initializeBytecodeTableForNewspeakV4
	"StackInterpreter initializeBytecodeTableForNewspeakV4"
	"Note: This table will be used to generate a C switch statement."

	InitializationOptions at: #NewsqueakV4BytecodeSet put: (NewsqueakV4BytecodeSet := true).
	BytecodeTable := Array new: 256.
	BytecodeEncoderClassName := #EncoderForNewsqueakV4.
	BytecodeSetHasDirectedSuperSend := true.
	BytecodeSetHasExtensions := true.
	LongStoreBytecode := 234.
	self table: BytecodeTable from:
	#(	"1 byte bytecodes"
		(   0  15 pushReceiverVariableBytecode)
		( 16  31 pushLiteralVariable16CasesBytecode)
		( 32  63 pushLiteralConstantBytecode)
		( 64  75 pushTemporaryVariableBytecode)
		( 76	 pushReceiverBytecode)
		( 77	 extPushPseudoVariableOrOuterBytecode)
		( 78	 pushConstantZeroBytecode)
		( 79	 pushConstantOneBytecode)

		( 80	 bytecodePrimAdd)
		( 81	 bytecodePrimSubtract)
		( 82	 bytecodePrimLessThanV4) "for booleanCheatV4:"
		( 83	 bytecodePrimGreaterThanV4) "for booleanCheatV4:"
		( 84	 bytecodePrimLessOrEqualV4) "for booleanCheatV4:"
		( 85	 bytecodePrimGreaterOrEqualV4) "for booleanCheatV4:"
		( 86	 bytecodePrimEqualV4) "for booleanCheatV4:"
		( 87	 bytecodePrimNotEqualV4) "for booleanCheatV4:"
		( 88	 bytecodePrimMultiply)
		( 89	 bytecodePrimDivide)
		( 90	 bytecodePrimMod)
		( 91	 bytecodePrimMakePoint)
		( 92	 bytecodePrimBitShift)
		( 93	 bytecodePrimDiv)
		( 94	 bytecodePrimBitAnd)
		( 95	 bytecodePrimBitOr)

		( 96	 bytecodePrimAt)
		( 97	 bytecodePrimAtPut)
		( 98	 bytecodePrimSize)
		( 99	 bytecodePrimNext)
		(100	 bytecodePrimNextPut)
		(101	 bytecodePrimAtEnd)
		(102	 bytecodePrimIdenticalV4) "for booleanCheatV4:"
		(103	 bytecodePrimClass)
		(104	 bytecodePrimNotIdenticalV4) "was blockCopy:"
		(105	 bytecodePrimValue)
		(106	 bytecodePrimValueWithArg)
		(107	 bytecodePrimDo)
		(108	 bytecodePrimNew)
		(109	 bytecodePrimNewWithArg)
		(110	 bytecodePrimPointX)
		(111	 bytecodePrimPointY)

		(112 127	sendLiteralSelector0ArgsBytecode)
		(128 143	sendLiteralSelector1ArgBytecode)
		(144 159	sendLiteralSelector2ArgsBytecode)
		(160 175	sendAbsentImplicit0ArgsBytecode)

		(176 183	storeAndPopReceiverVariableBytecode)
		(184 191	storeAndPopTemporaryVariableBytecode)

		(192 199	shortUnconditionalJump)
		(200 207	shortConditionalJumpTrue)
		(208 215	shortConditionalJumpFalse)

		(216		returnReceiver)
		(217		returnTopFromMethod)
		(218		extReturnTopFromBlock)

		(219		duplicateTopBytecode)
		(220		popStackBytecode)
		(221		extNopBytecode)
		(222 223	unknownBytecode)

		"2 byte bytecodes"
		(224		extABytecode)
		(225		extBBytecode)

		(226		extPushReceiverVariableBytecode)
		(227		extPushLiteralVariableBytecode)
		(228		extPushLiteralBytecode)
		(229		extPushIntegerBytecode)
		(230		longPushTemporaryVariableBytecode)
		(231		pushNewArrayBytecode)
		(232		extStoreReceiverVariableBytecode)
		(233		extStoreLiteralVariableBytecode)
		(234		longStoreTemporaryVariableBytecode)
		(235		extStoreAndPopReceiverVariableBytecode)
		(236		extStoreAndPopLiteralVariableBytecode)
		(237		longStoreAndPopTemporaryVariableBytecode)

		(238		extSendBytecode)
		(239		extSendSuperBytecode)
		(240		extSendAbsentImplicitBytecode)
		(241		extSendAbsentDynamicSuperBytecode)

		(242		extUnconditionalJump)
		(243		extJumpIfTrue)
		(244		extJumpIfFalse)

		(245		extSendAbsentSelfBytecode)

		(246 248	unknownBytecode)

		"3 byte bytecodes"
		(249		callPrimitiveBytecode)

		(250		pushRemoteTempLongBytecode)
		(251		storeRemoteTempLongBytecode)
		(252		storeAndPopRemoteTempLongBytecode)
		(253		extPushClosureBytecode)
		(254		extSendAbsentOuterBytecode)

		(255		unknownBytecode)
	)! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 10/29/2020 10:36'!
initializeBytecodeTableForSistaV1
	"See e.g. the cass comment for EncoderForSistaV1"
	"StackInterpreter initializeBytecodeTableForSistaV1"
	"Note: This table will be used to generate a C switch statement."

	InitializationOptions at: #SistaV1BytecodeSet put: (SistaV1BytecodeSet := true).

	BytecodeTable := Array new: 256.
	BytecodeEncoderClassName := #EncoderForSistaV1.
	BytecodeSetHasDirectedSuperSend := true.
	BytecodeSetHasExtensions := true.
	LongStoreBytecode := 245.
	self table: BytecodeTable from:
	#(	"1 byte bytecodes"
		(   0  15 pushReceiverVariableBytecode)
		( 16  31 pushLiteralVariable16CasesBytecode)
		( 32  63 pushLiteralConstantBytecode)
		( 64  75 pushTemporaryVariableBytecode)
		( 76	 pushReceiverBytecode)
		( 77	 pushConstantTrueBytecode)
		( 78	 pushConstantFalseBytecode)
		( 79	 pushConstantNilBytecode)
		( 80	 pushConstantZeroBytecode)
		( 81	 pushConstantOneBytecode)
		( 82	 extPushPseudoVariable)
		( 83	 duplicateTopBytecode)
	
		( 84 87	unknownBytecode)
		( 88	returnReceiver)
		( 89	returnTrue)
		( 90	returnFalse)
		( 91	returnNil)
		( 92	returnTopFromMethod)
		( 93	returnNilFromBlock)
		( 94	returnTopFromBlock)
		( 95	extNopBytecode)

		( 96	 bytecodePrimAdd)
		( 97	 bytecodePrimSubtract)
		( 98	 bytecodePrimLessThanSistaV1) 		"for booleanCheatSistaV1:"
		( 99	 bytecodePrimGreaterThanSistaV1) 	"for booleanCheatSistaV1:"
		(100	 bytecodePrimLessOrEqualSistaV1) 	"for booleanCheatSistaV1:"
		(101	 bytecodePrimGreaterOrEqualSistaV1) 	"for booleanCheatSistaV1:"
		(102	 bytecodePrimEqualSistaV1) 			"for booleanCheatSistaV1:"
		(103	 bytecodePrimNotEqualSistaV1) 		"for booleanCheatSistaV1:"
		(104	 bytecodePrimMultiply)
		(105	 bytecodePrimDivide)
		(106	 bytecodePrimMod)
		(107	 bytecodePrimMakePoint)
		(108	 bytecodePrimBitShift)
		(109	 bytecodePrimDiv)
		(110	 bytecodePrimBitAnd)
		(111	 bytecodePrimBitOr)

		(112	 bytecodePrimAt)
		(113	 bytecodePrimAtPut)
		(114	 bytecodePrimSize)
		(115	 bytecodePrimNext)		 "i.e. a 0 arg special selector"
		(116	 bytecodePrimNextPut)		 "i.e. a 1 arg special selector"
		(117	 bytecodePrimAtEnd)
		(118	 bytecodePrimIdenticalSistaV1) "for booleanCheatSistaV1:"
		(119	 bytecodePrimClass)
		(120	 bytecodePrimNotIdenticalSistaV1) "was blockCopy:"
		(121	 bytecodePrimValue)
		(122	 bytecodePrimValueWithArg)
		(123	 bytecodePrimDo)			"i.e. a 1 arg special selector"
		(124	 bytecodePrimNew)			"i.e. a 0 arg special selector"
		(125	 bytecodePrimNewWithArg)	"i.e. a 1 arg special selector"
		(126	 bytecodePrimPointX)		"i.e. a 0 arg special selector"
		(127	 bytecodePrimPointY)		"i.e. a 0 arg special selector"

		(128 143	sendLiteralSelector0ArgsBytecode)
		(144 159	sendLiteralSelector1ArgBytecode)
		(160 175	sendLiteralSelector2ArgsBytecode)

		(176 183	shortUnconditionalJump)
		(184 191	shortConditionalJumpTrue)
		(192 199	shortConditionalJumpFalse)
	
		(200 207	storeAndPopReceiverVariableBytecode)
		(208 215	storeAndPopTemporaryVariableBytecode)
		(216		popStackBytecode)
		(217		unconditionalTrapBytecode)

		(218 223	unknownBytecode)

		"2 byte bytecodes"
		(224		extABytecode)
		(225		extBBytecode)

		(226		extPushReceiverVariableBytecode)
		(227		extPushLiteralVariableBytecode)
		(228		extPushLiteralBytecode)
		(229		longPushTemporaryVariableBytecode)
		(230		unknownBytecode)
		(231		pushNewArrayBytecode)
		(232		extPushIntegerBytecode)
		(233		extPushCharacterBytecode)

		(234		extSendBytecode)
		(235		extSendSuperBytecode)

		(236		callMappedInlinedPrimitive)

		(237		extUnconditionalJump)
		(238		extJumpIfTrue)
		(239		extJumpIfFalse)

		(240		extStoreAndPopReceiverVariableBytecode)
		(241		extStoreAndPopLiteralVariableBytecode)
		(242		longStoreAndPopTemporaryVariableBytecode)

		(243		extStoreReceiverVariableBytecode)
		(244		extStoreLiteralVariableBytecode)
		(245		longStoreTemporaryVariableBytecode)

		(246 247	unknownBytecode)

		"3 byte bytecodes"
		(248		callPrimitiveBytecode)
		(249		extPushFullClosureBytecode)

		(250		extPushClosureBytecode)
		(251		pushRemoteTempLongBytecode)
		(252		storeRemoteTempLongBytecode)
		(253		storeAndPopRemoteTempLongBytecode)
				
		(254 255	unknownBytecode)
	)! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 6/7/2018 15:22'!
initializeBytecodeTableForSqueakV3PlusClosures
	"StackInterpreter initializeBytecodeTableForSqueakV3PlusClosures"
	"Note: This table will be used to generate a C switch statement."

	InitializationOptions at: #SqueakV3PlusClosuresBytecodeSet put: (SqueakV3PlusClosuresBytecodeSet := true).

	BytecodeTable := Array new: 256.
	BytecodeEncoderClassName := #EncoderForV3PlusClosures.
	LongStoreBytecode := 129.
	BytecodeSetHasExtensions := false.
	self table: BytecodeTable from:
	#(
		(  0  15 pushReceiverVariableBytecode)
		( 16  31 pushTemporaryVariableBytecode)
		( 32  63 pushLiteralConstantBytecode)
		( 64  95 pushLiteralVariableBytecode)
		( 96 103 storeAndPopReceiverVariableBytecode)
		(104 111 storeAndPopTemporaryVariableBytecode)
		(112 pushReceiverBytecode)
		(113 pushConstantTrueBytecode)
		(114 pushConstantFalseBytecode)
		(115 pushConstantNilBytecode)
		(116 pushConstantMinusOneBytecode)
		(117 pushConstantZeroBytecode)
		(118 pushConstantOneBytecode)
		(119 pushConstantTwoBytecode)
		(120 returnReceiver)
		(121 returnTrue)
		(122 returnFalse)
		(123 returnNil)
		(124 returnTopFromMethod)
		(125 returnTopFromBlock)

		(126 127 unknownBytecode)

		(128 extendedPushBytecode)
		(129 extendedStoreBytecode)
		(130 extendedStoreAndPopBytecode)
		(131 singleExtendedSendBytecode)
		(132 doubleExtendedDoAnythingBytecode)
		(133 singleExtendedSuperBytecode)
		(134 secondExtendedSendBytecode)
		(135 popStackBytecode)
		(136 duplicateTopBytecode)

		(137 pushActiveContextBytecode)
		(138 pushNewArrayBytecode)),

	((InitializationOptions at: #SpurObjectMemory ifAbsent: [false])
		ifTrue: [#((139 callPrimitiveBytecode))]	"V3PlusClosures on Spur"
		ifFalse: [#((139 unknownBytecode))]),	"V3PlusClosures on V3"

	  #(
		(140 pushRemoteTempLongBytecode)
		(141 storeRemoteTempLongBytecode)
		(142 storeAndPopRemoteTempLongBytecode)
		(143 pushClosureCopyCopiedValuesBytecode)

		(144 151 shortUnconditionalJump)
		(152 159 shortConditionalJumpFalse)
		(160 167 longUnconditionalJump)
		(168 171 longJumpIfTrue)
		(172 175 longJumpIfFalse)

		"176-191 were sendArithmeticSelectorBytecode"
		(176 bytecodePrimAdd)
		(177 bytecodePrimSubtract)
		(178 bytecodePrimLessThan)
		(179 bytecodePrimGreaterThan)
		(180 bytecodePrimLessOrEqual)
		(181 bytecodePrimGreaterOrEqual)
		(182 bytecodePrimEqual)
		(183 bytecodePrimNotEqual)
		(184 bytecodePrimMultiply)
		(185 bytecodePrimDivide)
		(186 bytecodePrimMod)
		(187 bytecodePrimMakePoint)
		(188 bytecodePrimBitShift)
		(189 bytecodePrimDiv)
		(190 bytecodePrimBitAnd)
		(191 bytecodePrimBitOr)

		"192-207 were sendCommonSelectorBytecode"
		(192 bytecodePrimAt)
		(193 bytecodePrimAtPut)
		(194 bytecodePrimSize)
		(195 bytecodePrimNext)
		(196 bytecodePrimNextPut)
		(197 bytecodePrimAtEnd)
		(198 bytecodePrimIdentical)
		(199 bytecodePrimClass)
		(200 bytecodePrimNotIdentical) "was bytecodePrimSpecialSelector24 / blockCopy"
		(201 bytecodePrimValue)
		(202 bytecodePrimValueWithArg)
		(203 bytecodePrimDo)
		(204 bytecodePrimNew)
		(205 bytecodePrimNewWithArg)
		(206 bytecodePrimPointX)
		(207 bytecodePrimPointY)

		(208 223 sendLiteralSelector0ArgsBytecode)
		(224 239 sendLiteralSelector1ArgBytecode)
		(240 255 sendLiteralSelector2ArgsBytecode)
	)! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 1/19/2017 14:40'!
initializeBytecodeTableForSqueakV3PlusClosuresNewspeakV4Hybrid
	"StackInterpreter initializeBytecodeTableForSqueakV3PlusClosuresNewspeakV4Hybrid"

	| v3Table v4Table |
	self initializeBytecodeTableForNewspeakV4.
	v4Table := BytecodeTable.
	AltBytecodeEncoderClassName := BytecodeEncoderClassName.
	AltLongStoreBytecode := LongStoreBytecode.
	self initializeBytecodeTableForSqueakV3PlusClosures.
	BytecodeSetHasExtensions := true.
	v3Table := BytecodeTable.
	BytecodeTable := v3Table, v4Table! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 1/19/2017 14:40'!
initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid
	"StackInterpreter initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid"

	| v3Table v1Table |
	self initializeBytecodeTableForSistaV1.
	v1Table := BytecodeTable.
	AltBytecodeEncoderClassName := BytecodeEncoderClassName.
	AltLongStoreBytecode := LongStoreBytecode.
	self initializeBytecodeTableForSqueakV3PlusClosures.
	BytecodeSetHasExtensions := true.
	v3Table := BytecodeTable.
	BytecodeTable := v3Table, v1Table! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 1/24/2016 11:28'!
initializeCaches

	| atCacheEntries atCacheEntrySize |
	NewspeakVM
		ifTrue: [MethodCacheEntries := 512]
		ifFalse: [MethodCacheEntries := 1024].
	MethodCacheSelector := 1.
	MethodCacheClass := 2.
	MethodCacheMethod := 3.
	MethodCachePrimFunction := 4.
	MethodCacheEntrySize := 4.  "Must be power of two for masking scheme."
	MethodCacheMask := (MethodCacheEntries - 1) * MethodCacheEntrySize.
	MethodCacheSize := MethodCacheEntries * MethodCacheEntrySize.
	CacheProbeMax := 3.

	NSMethodCacheEntries := 512.
	NSMethodCacheSelector := 1.
	NSMethodCacheClassTag := 2.
	NSMethodCacheCallingMethod := 3.
	NSMethodCacheDepthOrLookupRule := 4.
	NSMethodCacheTargetMethod := 5.
	NSMethodCachePrimFunction := 6.
	NSMethodCacheActualReceiver := 7.
	NSMethodCacheEntrySize := 8.	"Must be power of two for masking scheme."
	NSMethodCacheMask := NSMethodCacheEntries - 1 * NSMethodCacheEntrySize.
	NSMethodCacheSize := NSMethodCacheEntries * NSMethodCacheEntrySize.

	AtCacheOop := 1.
	AtCacheSize := 2.
	AtCacheFmt := 3.
	AtCacheFixedFields := 4.
	atCacheEntries := 8.
	atCacheEntrySize := 4.  "Must be power of two for masking scheme."
	AtCacheMask := (atCacheEntries-1) * atCacheEntrySize.
	AtPutBase := atCacheEntries * atCacheEntrySize.
	AtCacheTotalSize := atCacheEntries * atCacheEntrySize * 2.

	"LookupRuleOuter is [0, 255], with the value being the lexical depth. Note that an
	 outer send to lexical depth 0 is equivalent to a self send. Implicit receiver and
	 outer sends are encoded as adjacent values to allow a quick range check to
	 determine whether the absent receiver might differ from the method receiver.
	 Note also Smalltalk super sends use ordinary send lookup rules."
	LookupRuleSelf := 0.
	LookupRuleImplicit := 256.
	LookupRuleDynamicSuper := 257.
	LookupRuleOrdinary := 258.
	LookupRuleMNU := 259.
! !

!StackInterpreter class methodsFor: 'initialization' stamp: ''!
initializeCharacterIndex
	CharacterValueIndex := 0! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'tpr 4/29/2003 12:10'!
initializeCharacterScannerIndices
	CrossedX := 258.
	EndOfRun := 257
! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'tfel 2/22/2016 09:25'!
initializeClassIndices
	"Class Class"
	SuperclassIndex := 0.
	MethodDictionaryIndex := 1.
	InstanceSpecificationIndex := 2.
	"Fields of a message dictionary"
	MethodArrayIndex := 1. 
	SelectorStart := 2.

	"Newspeak extensions"
	"Fields of a Mixin (also defined by Behavior)"
	EnclosingMixinIndex := NewspeakVM ifTrue: [3].  "(Class instVarIndexFor: 'mixinSlot') - 1"
	"Fields of a MixinApplication (also defined by Behavior)"
	MixinIndex := NewspeakVM ifTrue: [3]. "(Class instVarIndexFor: 'mixinSlot') - 1"
	EnclosingObjectIndex := NewspeakVM ifTrue: [4]. "(Class instVarIndexFor: 'enclosingObjectSlot') - 1"

! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'cb 4/5/2016 18:07'!
initializeContextIndices
	"Class MethodContext"
	SenderIndex := 0.
	InstructionPointerIndex := 1.
	StackPointerIndex := 2.
	MethodIndex := 3.
	ClosureIndex := 4. "N.B. Called receiverMap in old images, closureOrNil in newer images."
	ReceiverIndex := 5.
	CtxtTempFrameStart := 6.

	SmallContextSlots := CtxtTempFrameStart + 16.  "16 indexable fields"
	"Large contexts have 56 indexable fields.  Max with single header word of ObjectMemory [but not SpurMemoryManager ;-)]."
	LargeContextSlots := CtxtTempFrameStart + 56.
	
	"Including the header size in these sizes is problematic for multiple memory managers,
	 so we don't use them.  Set to #bogus for error checking."
	SmallContextSize := #bogus.
	LargeContextSize := #bogus.

	"Class BlockClosure"
	ClosureOuterContextIndex := 0.
	ClosureStartPCIndex := 1.
	ClosureNumArgsIndex := 2.
	ClosureFirstCopiedValueIndex := 3.
	
	"Class FullBlockClosure, outercontext and numArgs index are shared with the BlockClosure class."	
	FullClosureCompiledBlockIndex := 1.
	FullClosureReceiverIndex := 3.
	FullClosureFirstCopiedValueIndex := 4.! !

!StackInterpreter class methodsFor: 'initialization' stamp: ''!
initializeDirectoryLookupResultCodes

	DirEntryFound := 0.
	DirNoMoreEntries := 1.
	DirBadPath := 2.! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 1/24/2016 11:38'!
initializeFrameIndices
	"Format of a stack frame.  Word-sized indices relative to the frame pointer.
	 Terminology
		Frames are either single (have no context) or married (have a context).
		Contexts are either single (exist on the heap), married (have a context) or widowed (had a frame that has exited).
	 Stacks grow down:

			receiver for method activations/closure for block activations
			arg0
			...
			argN
			caller's method ip/base frame's sender context
	fp->	saved fp
			method
			frame flags
			context (uninitialized)
			receiver
			first temp
			...
	sp->	Nth temp

	frame flags holds the number of arguments (since argument temporaries are above the frame)
	the flag for a block activation
	and the flag indicating if the context field is valid (whether the frame is married).

	The first frame in a stack page is the baseFrame and is marked as such by a null saved fp,
	in which case the saved method ip is actually the context (possibly hybrid) beneath the base frame"

	| fxCallerSavedIP fxSavedFP fxMethod fxFrameFlags fxThisContext fxReceiver |
	fxCallerSavedIP := 1.
	fxSavedFP := 0.
	fxMethod := -1.
	fxFrameFlags := -2.	"Can find numArgs, needed for fast temp access. args are above fxCallerSavedIP.
						 Can find ``is block'' bit
						 Can find ``has context'' bit"
	fxThisContext := -3.
	fxReceiver := -4.

	FrameSlots := fxCallerSavedIP - fxReceiver + 1.

	FoxCallerSavedIP := fxCallerSavedIP * BytesPerWord.
	"In base frames the caller saved ip field holds the caller context."
	FoxCallerContext := FoxCallerSavedIP.
	FoxSavedFP := fxSavedFP * BytesPerWord.
	FoxMethod := fxMethod * BytesPerWord.
	FoxFrameFlags := fxFrameFlags * BytesPerWord.
	FoxThisContext := fxThisContext * BytesPerWord.
	FoxReceiver := fxReceiver * BytesPerWord.

	"Mark the CoInterpreter-specific offsets as #undefined to
	 avoid including them accidentally in StackInterpreter code."
	IFrameSlots := #undefined.
	MFrameSlots := #undefined.
	FoxIFrameFlags := #undefined.
	FoxIFSavedIP := #undefined.
	FoxIFReceiver := #undefined.
	FoxMFReceiver := #undefined! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'di 3/25/1999 22:09'!
initializeMessageIndices
	MessageSelectorIndex := 0.
	MessageArgumentsIndex := 1.
	MessageLookupClassIndex := 2.! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'rmacnak 4/11/2015 15:51'!
initializeMethodIndices
	| tagBits |
	"Class CompiledMethod"
	HeaderIndex := 0.
	LiteralStart := 1.

	tagBits := self objectMemoryClass numSmallIntegerTagBits.
	LargeContextBit := 16r20000 << tagBits.  "This bit set in method headers if large context is needed."
	MethodHeaderTempCountShift := 18 + tagBits.
	MethodHeaderArgCountShift := 24 + tagBits.
	V3PrimitiveBitsMask := 16r100001FF << tagBits.
	AlternateHeaderHasPrimFlag := 16r10000 << tagBits.
	AlternateHeaderIsOptimizedFlag := 16r8000 << tagBits.
	AlternateHeaderNumLiteralsMask := 16r7FFF. "N.B.  *not* shifted"
	"The position of the unused flag bit in the method header, not including tag bit(s).
	 Bits 28 & 29 are either used as an accessModifer code in Newspeak (public, private, protected)
	 or are free for use as flag bits."
	MethodHeaderFlagBitPosition := 28 + tagBits.
	AccessModifierPublic := 2r00.
	AccessModifierPrivate := 2r01.
	AccessModifierProtected := 2r10.! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 6/21/2021 12:05'!
initializeMiscConstants

	super initializeMiscConstants.
	STACKVM := true.

	"These flags identify a GC operation (& hence a reason to leak check),
	 or just operations the leak checker can be run for."
	GCModeFull := 1.				"stop-the-world global GC"
	GCModeNewSpace := 2.		"Spur's scavenge, or V3's incremental"
	GCModeIncremental := 4.		"incremental global gc (Dijkstra tri-colour marking); as yet unimplemented"
	GCModeBecome := 8.			"v3 post-become sweeping/Spur forwarding"
	GCCheckImageSegment := 16.	"just a flag for leak checking image segments"
	GCCheckFreeSpace := 32.		"just a flag for leak checking free space; Spur only"
	GCCheckShorten := 64.		"just a flag for leak checking object shortening operations; Spur only"
	GCCheckPrimCall := 128.		"just a flag for leak checking external primitive calls"

	StackPageTraceInvalid := -1.
	StackPageUnreached := 0.
	StackPageReachedButUntraced := 1.
	StackPageTraced := 2.

	DumpStackOnLowSpace := 0.
	MillisecondClockMask := 16r1FFFFFFF.
	"Note: The external primitive table should actually be dynamically sized but for the sake of inferior platforms (e.g., Mac :-) who cannot allocate memory in any reasonable way, we keep it static (and cross our fingers...)"
	MaxExternalPrimitiveTableSize := 4096. "entries"

	MaxJumpBuf := 32. "max. callback depth"
	FailImbalancedPrimitives := InitializationOptions at: #FailImbalancedPrimitives ifAbsentPut: [true].
	EnforceAccessControl := InitializationOptions at: #EnforceAccessControl ifAbsent: [true].

	ReturnToInterpreter := 1. "setjmp/longjmp code."

	"Because of a hack with callbacks in the non-threaded VM they must not conflct with the VM's tag bits."
	DisownVMForFFICall := 16.
	DisownVMForThreading := 32
! !

!StackInterpreter class methodsFor: 'initialization' stamp: ''!
initializePointIndices
	XIndex := 0.
	YIndex := 1! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 9/4/2021 17:27'!
initializePrimitiveTable
	"This table generates a C function address table used in slowPrimitiveResponse along with dispatchFunctionPointerOn:in:"
	"self initializePrimitiveTable"

	"NOTE: The real limit here is 2047 because of the old method header layout but there is no point in going over the needed size"
	MaxPrimitiveIndex := 576.	
	MaxQuickPrimitiveIndex := 519.

	PrimNumberInstVarAt := 73.
	PrimNumberExternalCall := 117.
		ExternalCallLiteralModuleNameIndex := 0.
		ExternalCallLiteralFunctionNameIndex := 1.
		ExternalCallLiteralFlagsIndex := 2.
		ExternalCallLiteralTargetFunctionIndex := 3.
	PrimNumberDoPrimitive := 118.
	PrimNumberFFICall := 120.
	PrimNumberNoContextSwitchMarker := 123.
	PrimNumberShallowCopy := 148.
	PrimNumberSlotAt := 173.
	PrimNumberUnwindMarker := 198.
	PrimNumberHandlerMarker := 199.
	PrimNumberDoExternalCall := 218.
	PrimNumberFlushExternalPrimitives := 570.
	PrimNumberUnloadModule := 571.

	PrimitiveTable := Array new: MaxPrimitiveIndex + 1.
	self primitivesClass table: PrimitiveTable from: 
	#(	"Integer Primitives (0-19)"
		(0 0 "a.k.a. primitiveFail but faster because primitiveFail won't even be called")
		(1 primitiveAdd)
		(2 primitiveSubtract)
		(3 primitiveLessThan)
		(4 primitiveGreaterThan)
		(5 primitiveLessOrEqual)
		(6 primitiveGreaterOrEqual)
		(7 primitiveEqual)
		(8 primitiveNotEqual)
		(9 primitiveMultiply)
		(10 primitiveDivide)
		(11 primitiveMod)
		(12 primitiveDiv)
		(13 primitiveQuo)
		(14 primitiveBitAnd)
		(15 primitiveBitOr)
		(16 primitiveBitXor)
		(17 primitiveBitShift)
		(18 primitiveMakePoint)
		(19 primitiveFail)					"Guard primitive for simulation -- *must* fail"

		"LargeInteger Primitives (20-37)"
		(20 primitiveRemLargeIntegers)
		(21 primitiveAddLargeIntegers)
		(22 primitiveSubtractLargeIntegers)
		(23 primitiveLessThanLargeIntegers)
		(24 primitiveGreaterThanLargeIntegers)
		(25 primitiveLessOrEqualLargeIntegers)
		(26 primitiveGreaterOrEqualLargeIntegers)
		(27 primitiveEqualLargeIntegers)
		(28 primitiveNotEqualLargeIntegers)
		(29 primitiveMultiplyLargeIntegers)
		(30 primitiveDivideLargeIntegers)
		(31 primitiveModLargeIntegers)
		(32 primitiveDivLargeIntegers)
		(33 primitiveQuoLargeIntegers)
		(34 primitiveBitAndLargeIntegers)
		(35 primitiveBitOrLargeIntegers)
		(36 primitiveBitXorLargeIntegers)
		(37 primitiveBitShiftLargeIntegers)

		"Float Primitives (38-59)"
		(38 primitiveFloatAt)
		(39 primitiveFloatAtPut)
		(40 primitiveAsFloat)
		(41 primitiveFloatAdd)
		(42 primitiveFloatSubtract)
		(43 primitiveFloatLessThan)
		(44 primitiveFloatGreaterThan)
		(45 primitiveFloatLessOrEqual)
		(46 primitiveFloatGreaterOrEqual)
		(47 primitiveFloatEqual)
		(48 primitiveFloatNotEqual)
		(49 primitiveFloatMultiply)
		(50 primitiveFloatDivide)
		(51 primitiveTruncated)
		(52 primitiveFractionalPart)
		(53 primitiveExponent)
		(54 primitiveTimesTwoPower)
		(55 primitiveSquareRoot)
		(56 primitiveSine)
		(57 primitiveArctan)
		(58 primitiveLogN)
		(59 primitiveExp)

		"Subscript and Stream Primitives (60-67)"
		(60 primitiveAt)
		(61 primitiveAtPut)
		(62 primitiveSize)
		(63 primitiveStringAt)
		(64 primitiveStringAtPut)
		"The stream primitives no longer pay their way; normal Smalltalk code is faster."
		(65 0 "a.k.a. primitiveFail but faster because primitiveFail won't even be called")"was primitiveNext"
		(66 0 "a.k.a. primitiveFail but faster because primitiveFail won't even be called") "was primitiveNextPut"
		(67 0 "a.k.a. primitiveFail but faster because primitiveFail won't even be called") "was primitiveAtEnd"

		"StorageManagement Primitives (68-79)"
		(68 primitiveObjectAt)
		(69 primitiveObjectAtPut)
		(70 primitiveNew)
		(71 primitiveNewWithArg)
		(72 primitiveArrayBecomeOneWay)	"Blue Book: primitiveBecome"
		(73 primitiveInstVarAt)
		(74 primitiveInstVarAtPut)
		(75 primitiveIdentityHash)
		(76 primitiveStoreStackp)			"Blue Book: primitiveAsObject"
		(77 primitiveSomeInstance)
		(78 primitiveNextInstance)
		(79 primitiveNewMethod)

		"Control Primitives (80-89)"
		(80 primitiveFail)					"Blue Book: primitiveBlockCopy"
		(81 primitiveFail)					"Blue Book: primitiveValue"
		(82 primitiveFail)					"Blue Book: primitiveValueWithArgs"
		(83 primitivePerform)
		(84 primitivePerformWithArgs)
		(85 primitiveSignal)
		(86 primitiveWait)
		(87 primitiveResume)
		(88 primitiveSuspend)
		(89 primitiveFlushCache)			"See 116 & 119 below"

		"Input/Output Primitives (90-109)"
		(90 primitiveFail)						"Blue Book, Squeak 3.x: primitiveMousePoint)"
		(91 primitiveTestDisplayDepth)			"Blue Book: primitiveCursorLocPut"
		(92 primitiveSetDisplayMode)				"Blue Book: primitiveCursorLink"
		(93 primitiveInputSemaphore)
		(94 primitiveGetNextEvent)				"Blue Book: primitiveSampleInterval"
		(95 primitiveInputWord)
		(96 primitiveFail)	"primitiveCopyBits"
		(97 primitiveSnapshot)
		(98 primitiveStoreImageSegment)
		(99 primitiveLoadImageSegment)
		(100 primitivePerformInSuperclass)		"Blue Book: primitiveSignalAtTick"
		(101 primitiveBeCursor)
		(102 primitiveBeDisplay)
		(103 primitiveScanCharacters)
		(104 primitiveFail)	"primitiveDrawLoop"
		(105 primitiveStringReplace)
		(106 primitiveScreenSize)
		(107 primitiveFail)						"Blue Book, Squeak 3.x: primitiveMouseButtons"
		(108 primitiveFail)						"Blue Book, Squeak 3.x: primitiveKbdNext"
		(109 primitiveFail)						"Blue Book, Squeak 3.x: primitiveKbdPeek"

		"System Primitives (110-119)"
		(110 primitiveIdentical)
		(111 primitiveClass)
		(112 primitiveBytesLeft)
		(113 primitiveQuit)
		(114 primitiveExitToDebugger)
		(115 primitiveChangeClass)					"Blue Book: primitiveOopsLeft"
		(116 primitiveFlushCacheByMethod)
		(117 primitiveExternalCall)
		(118 primitiveDoPrimitiveWithArgs)
		(119 primitiveFlushCacheBySelector)
			"Squeak 2.2 and earlier use 119.  Squeak 2.3 and later use 116.
			 Spur (Squeak 5 and later) use 119.  All use 89 above for coarse-grain flushing.
			 Both 116 and 119 are supported as both have their uses."

		"Miscellaneous Primitives (120-127)"
		(120 primitiveCalloutToFFI)
		(121 primitiveImageName)
		(122 primitiveNoop)					"Blue Book: primitiveImageVolume"
		(123 0)									"PrimNumberNoContextSwitch marker; must fail; was primitiveValueUninterruptably"
		(124 primitiveLowSpaceSemaphore)
		(125 primitiveSignalAtBytesLeft)

		"Squeak Primitives Start Here"

		"Squeak Miscellaneous Primitives (128-149)"
		(126 primitiveDeferDisplayUpdates)
		(127 primitiveShowDisplayRect)
		(128 primitiveArrayBecome)
		(129 primitiveSpecialObjectsOop)
		(130 primitiveFullGC)
		(131 primitiveIncrementalGC)
		(132 primitiveObjectPointsTo)
		(133 primitiveSetInterruptKey)
		(134 primitiveInterruptSemaphore)
		(135 primitiveMillisecondClock)
		(136 primitiveSignalAtMilliseconds)
		(137 primitiveSecondsClock)
		(138 primitiveSomeObject)
		(139 primitiveNextObject)
		(140 primitiveBeep)
		(141 primitiveClipboardText)
		(142 primitiveVMPath)
		(143 primitiveShortAt)
		(144 primitiveShortAtPut)
		(145 primitiveConstantFill)
		"NOTE: When removing the obsolete indexed primitives,
		the following two should go become #primitiveIntegerAt / atPut"
		(146 primitiveFail)	"primitiveReadJoystick"
		(147 primitiveFail)	"primitiveWarpBits"
		(148 primitiveClone)
		(149 primitiveGetAttribute)

		"Were File Primitives (150-169) - NO LONGER INDEXED"
		(150 157 primitiveFail)
		(158 primitiveCompareWith)
		(159 primitiveHashMultiply)
		(160 primitiveAdoptInstance)
		(161 primitiveSetOrHasIdentityHash)
		(162 primitiveFail)			"reserved for immutablity primitives"
		(163 primitiveGetImmutability)
		(164 primitiveSetImmutability)	
		(165 primitiveIntegerAt)		"hacked in here for now"
		(166 primitiveIntegerAtPut)
		(167 primitiveYield)
		(168 primitiveCopyObject)
		(169 primitiveNotIdentical)

		"SpurMemoryManager primitives"
		(170 primitiveAsCharacter)
		(171 primitiveImmediateAsInteger)
		(172 primitiveFetchNextMourner)
		(173 primitiveSlotAt)		"c.f. (73 primitiveInstVarAt)"
		(174 primitiveSlotAtPut)	"c.f. (74 primitiveInstVarAtPut)"
		(175 primitiveBehaviorHash)
		(176 primitiveMaxIdentityHash)
		(177 primitiveAllInstances)
		(178 primitiveAllObjects)
		(179 primitiveFail) "reserved for primitiveAllInstancesOfAny"
		(180 primitiveGrowMemoryByAtLeast)
		(181 primitiveSizeInBytesOfInstance)
		(182 primitiveSizeInBytes)
		(183 primitiveIsPinned)
		(184 primitivePin)

		"CriticalSection primitives"
		(185 primitiveExitCriticalSection) "similar to signal hence index = signal + 100"
		(186 primitiveEnterCriticalSection) "similar to wait hence index = wait + 100. was primitiveClosureValue"
		(187 primitiveTestAndSetOwnershipOfCriticalSection) "was primitiveClosureValueWithArgs"

		"Perform method directly"
		(188 primitiveExecuteMethodArgsArray)
		(189 primitiveExecuteMethod)

		"Sound Primitives (continued) - NO LONGER INDEXED"
		(190 194 primitiveFail)

		"Unwind primitives"
		(195 primitiveFindNextUnwindContext)
		(196 primitiveTerminateTo)
		(197 primitiveFindHandlerContext)
		(198 0 "PrimNumberUnwindMarker; must fail; a.k.a. primitiveMarkUnwindMethod")
		(199 0 "PrimNumberHandlerMarker; must fail; a.k.a. primitiveMarkHandlerMethod")

		"new closure primitives (were Networking primitives)"
		(200 primitiveClosureCopyWithCopiedValues)
		(201 primitiveClosureValue) "value"
		(202 primitiveClosureValue) "value:"
		(203 primitiveClosureValue) "value:value:"
		(204 primitiveClosureValue) "value:value:value:"
		(205 primitiveClosureValue) "value:value:value:value:"
		(206 primitiveClosureValueWithArgs) "valueWithArguments:"

		(207 primitiveFullClosureValue) "value[:value:value:value:] et al"
		(208 primitiveFullClosureValueWithArgs) "valueWithArguments:"
		(209 primitiveFullClosureValueNoContextSwitch) "valueNoContextSwitch[:value:] et al"

		(210 primitiveContextAt)
		(211 primitiveContextAtPut)
		(212 primitiveContextSize)
		(213 primitiveContextXray)
		(214 primitiveVoidVMState)
		(215 primitiveFlushCacheByMethod) "primitiveVoidVMStateForMethod in the CoInterpreter"
		(216 217 primitiveFail)	"reserved for Cog primitives"
		(218 primitiveDoNamedPrimitiveWithArgs)
		(219 primitiveFail)		"reserved for Cog primitives"

		(220 primitiveFail)		"reserved for Cog primitives"

		(221 primitiveClosureValueNoContextSwitch) "valueNoContextSwitch"
		(222 primitiveClosureValueNoContextSwitch) "valueNoContextSwitch:"

		(223 229 primitiveFail)	"reserved for Cog primitives (see 228 & 229 in CoInterpreterMT"

		(230 primitiveRelinquishProcessor)
		(231 primitiveForceDisplayUpdate)
		(232 primitiveFormPrint)
		(233 primitiveSetFullScreen)
		(234 primitiveFail) "primBitmapdecompressfromByteArrayat"
		(235 primitiveFail) "primStringcomparewithcollated"
		(236 primitiveFail) "primSampledSoundconvert8bitSignedFromto16Bit"
		(237 primitiveFail) "primBitmapcompresstoByteArray"
		(238 primitiveFloatArrayAt)					"was serial port primitive"
		(239 primitiveFloatArrayAtPut)				"was serial port primitive"
		(240 primitiveUTCMicrosecondClock)		"was serial port primitive"
		(241 primitiveLocalMicrosecondClock)		"was serial port primitive"
		(242 primitiveSignalAtUTCMicroseconds)
		(243 primitiveUpdateTimezone) "primStringtranslatefromtotable"
		(244 primitiveUtcAndTimezoneOffset)
		(245 primitiveCoarseUTCMicrosecondClock) "primStringindexOfAsciiinStringstartingAt"
		(246 primitiveCoarseLocalMicrosecondClock) "primStringfindSubstringinstartingAtmatchTable"
		(247 primitiveSnapshotEmbedded)
		(248 primitiveArrayBecomeOneWayNoCopyHash) "was primitiveInvokeObjectAsMethod"
		(249 primitiveArrayBecomeOneWayCopyHashArg)

		"VM Implementor Primitives (250-255)"
		(250 primitiveClearVMProfile)
		(251 primitiveControlVMProfiling "primitiveStartVMProfiling")
		(252 primitiveVMProfileSamplesInto "primitiveStopVMProfiling")
		(253 primitiveFail "N.B. primitiveCollectCogCodeConstituents in CoInterpreter below")
		(254 primitiveVMParameter)
		(255 primitiveFail) "RESERVED FOR THE R/Squeak VM. Was primitiveInstVarsPutFromStack. Never used except in Disney tests."

		"Quick Push Const Methods"
		(256 nil) "primitivePushSelf"
		(257 nil) "primitivePushTrue"
		(258 nil) "primitivePushFalse"
		(259 nil) "primitivePushNil"
		(260 nil) "primitivePushMinusOne"
		(261 nil) "primitivePushZero"
		(262 nil) "primitivePushOne"
		(263 nil) "primitivePushTwo"

		"Quick Push Inst Var Methods"
		(264 519 nil) "primitiveLoadInstVar"

		(520 primitiveFail)
		(521 529 primitiveFail) "Reserved for Spur primitives"
		(530 539 primitiveFail)

		(540 primitiveFail)
		(541 primitiveSmallFloatAdd)
		(542 primitiveSmallFloatSubtract)
		(543 primitiveSmallFloatLessThan)
		(544 primitiveSmallFloatGreaterThan)
		(545 primitiveSmallFloatLessOrEqual)
		(546 primitiveSmallFloatGreaterOrEqual)
		(547 primitiveSmallFloatEqual)
		(548 primitiveSmallFloatNotEqual)
		(549 primitiveSmallFloatMultiply)
		(550 primitiveSmallFloatDivide)
		(551 primitiveSmallFloatTruncated)
		(552 primitiveSmallFloatFractionalPart)
		(553 primitiveSmallFloatExponent)
		(554 primitiveSmallFloatTimesTwoPower)
		(555 primitiveSmallFloatSquareRoot)
		(556 primitiveSmallFloatSine)
		(557 primitiveSmallFloatArctan)
		(558 primitiveSmallFloatLogN)
		(559 primitiveSmallFloatExp)

		(560 569 primitiveFail)

		"External primitive support primitives"
		(570 primitiveFlushExternalPrimitives)
		(571 primitiveUnloadModule)
		(572 primitiveListBuiltinModule)
		(573 primitiveListExternalModule)
		(574 primitiveFail) "reserved for addl. external support prims"

		"Unassigned Primitives"
		(575 primitiveHighBit)
		(576 primitiveFail))! !

!StackInterpreter class methodsFor: 'initialization' stamp: ''!
initializeSchedulerIndices
	"Class ProcessorScheduler"
	ProcessListsIndex := 0.
	ActiveProcessIndex := 1.
	"Class LinkedList"
	FirstLinkIndex := 0.
	LastLinkIndex := 1.
	"Class Semaphore"
	ExcessSignalsIndex := 2.
	"Class Link"
	NextLinkIndex := 0.
	"Class Process"
	SuspendedContextIndex := 1.
	PriorityIndex := 2.
	MyListIndex := 3! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 6/9/2010 14:31'!
initializeSmallIntegers
	"SmallIntegers"
	ConstMinusOne := self objectMemoryClass basicNew integerObjectOf: -1.
	ConstZero := self objectMemoryClass basicNew integerObjectOf: 0.
	ConstOne := self objectMemoryClass basicNew integerObjectOf: 1.
	ConstTwo := self objectMemoryClass basicNew integerObjectOf: 2! !

!StackInterpreter class methodsFor: 'initialization' stamp: ''!
initializeStreamIndices
	StreamArrayIndex := 0.
	StreamIndexIndex := 1.
	StreamReadLimitIndex := 2.
	StreamWriteLimitIndex := 3.! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 6/7/2018 15:03'!
initializeWithOptions: optionsDictionary
	"StackInterpreter initializeWithOptions: Dictionary new"

	super initializeWithOptions: optionsDictionary.
	self initializeMiscConstants. "must precede other initialization."
	self initializeAssociationIndex.
	self initializeBytecodeTable.
	self initializeCaches.
	self initializeCharacterIndex.
	self initializeCharacterScannerIndices.
	self initializeClassIndices.
	self initializeContextIndices.
	self initializeDirectoryLookupResultCodes.
	self initializeFrameIndices.
	self initializeMessageIndices.
	self initializeMethodIndices.
	self initializePointIndices.
	self initializePrimitiveTable.
	self initializeSchedulerIndices.
	self initializeSmallIntegers.
	self initializeStreamIndices! !

!StackInterpreter class methodsFor: 'accessing' stamp: 'EstebanLorenzano 10/23/2014 14:20'!
interpreterVersion 
	^ 'Stack'! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 2/10/2009 16:02'!
isInterpreterClass
	^true! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 10/12/2013 11:00'!
isNonArgumentImplicitReceiverVariableName: aString
	^(#('self' 'stackPages' 'interpreter' 'coInterpreter' 'objectMemory') includes: aString)
	  or: [self objectMemoryClass isNonArgumentImplicitReceiverVariableName: aString]! !

!StackInterpreter class methodsFor: 'spur compilation support' stamp: 'eem 12/28/2013 13:03'!
isObjectAccessor: selector
	"Answer if selector is one of fetchPointer:ofObject: storePointer:ofObject:withValue:
	 et al."
	^(InterpreterProxy whichCategoryIncludesSelector: selector) = #'object access'
	 or: [(SpurMemoryManager whichCategoryIncludesSelector: selector) = #'object access']

	"This for checking.  The above two protocols are somewhat disjoint."
	"(InterpreterProxy allMethodsInCategory: #'object access') copyWithoutAll: (SpurMemoryManager allMethodsInCategory: #'object access')"
	"(SpurMemoryManager allMethodsInCategory: #'object access') copyWithoutAll: (InterpreterProxy allMethodsInCategory: #'object access')"! !

!StackInterpreter class methodsFor: 'spur compilation support' stamp: 'eem 8/17/2014 12:28'!
isStackAccessor: selector
	^(StackInterpreter whichCategoryIncludesSelector: selector) = #'stack access'! !

!StackInterpreter class methodsFor: 'benchmarks' stamp: 'eem 6/21/2011 08:46'!
macroBenchmark  "StackInterpreter macroBenchmark"
	| dir |
	dir := 'benchmark2.dir'.
	(FileDirectory default directoryExists: dir)
		ifTrue: [(FileDirectory default directoryNamed: dir) recursiveDeleteContents]
		ifFalse: [(FileDirectory default directoryNamed: dir) assureExistence].
	([VMMaker
			makerFor: StackInterpreter
			and: nil
			with: #()
			to: (FileDirectory default pathFromURI: dir)
			platformDir: 'none'
			excluding:  (InterpreterPlugin withAllSubclasses collect: [:ea| ea name])]
		on: VMMakerException "suppress bleats about non-existent platforms dir"
		do: [:ex| ex resume: nil])
			generateInterpreterFile.
	(FileDirectory default directoryNamed: dir) recursiveDeleteContents; recursiveDelete! !

!StackInterpreter class methodsFor: 'spur compilation support' stamp: 'eem 7/22/2021 18:59'!
metadataFlagsForPrimitive: aPrimitiveMethod
	"We allow methods to decorate themselves with 8 flags (only one specified so far: FastCPrimitiveFlag)
	 using the export:flags: pragma.  But this is not relevant to the StackInterpreter."
	^0! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 12/22/2020 14:22'!
mustBeGlobal: var
	"Answer if a variable must be global and exported.  Used for inst vars that are accessed from VM support code,
	 and for variables that are initialized to some value (e.g. primitiveDoMixedArithmetic)."

	^(super mustBeGlobal: var)
	   or: [(self objectMemoryClass mustBeGlobal: var)
	   or: [(#('interpreterProxy' 'interpreterVersion' 'inIOProcessEvents' 'sendWheelEvents'
			'deferDisplayUpdates' 'eventTraceMask' 'extraVMMemory'
			'showSurfaceFn' 'displayBits' 'displayWidth' 'displayHeight' 'displayDepth'
			'desiredNumStackPages' 'desiredEdenBytes'
			'primitiveDoMixedArithmetic'
			'breakLookupClassTag' 'breakSelector' 'breakSelectorLength' 'sendTrace' 'checkAllocFiller' 'checkedPluginName'
			"'reenterInterpreter'" 'suppressHeartbeatFlag' 'ffiExceptionResponse'
			'debugCallbackInvokes' 'debugCallbackPath' 'debugCallbackReturns') includes: var)
	   or: [ "This allows slow machines to define bytecodeSetSelector as 0
			to avoid the interpretation overhead."
			MULTIPLEBYTECODESETS not and: [var = 'bytecodeSetSelector']]]]! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 6/18/2008 10:46'!
namesOfVariablesToLocalize
	^#(currentBytecode localFP localIP localSP localReturnValue)! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 2/9/2009 13:23'!
needsCogit
	^false! !

!StackInterpreter class methodsFor: 'translation' stamp: 'sw 5/23/2001 14:34'!
patchInterpGCCPPC: fileName
	"Interpreter patchInterpGCCPPC: 'Squeak copy 1'"
	"This will patch out the unneccesary range check (a compare
	 and branch) in the inner interpreter dispatch loop. for the PPC version of the GCC compiled
	version of Squeak under MPW"
	"NOTE: You must edit in the Interpeter file name"

	| delta f code len remnant i |
	delta := 7.
	f := FileStream fileNamed: fileName.
	f binary.
	code := Bitmap new: (len := f size) // 4.
	f nextInto: code.
	remnant := f next: len - (code size * 4).
	i := 0.
	["Look for a BCTR instruction"
	(i := code indexOf: 16r4E800420 startingAt: i + 1 ifAbsent: [0]) > 0] whileTrue:
		["Look for a CMPLWI cr1,rxx,FF, 7 instrs back"
	       ((code at: i - delta) bitAnd: 16rFFE0FFFF) = 16r288000FF ifTrue:
	       	["Copy dispatch instrs back over the compare"
			self inform: 'Patching at ', i hex.
			0 to: delta - 2 do: [ :j |
				code at: (i - delta) + j put: (code at: (i - delta) + j + 2)]]].
	f position: 0; nextPutAll: code; nextPutAll: remnant.
	f close! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 12/26/2016 20:42'!
preGenerationHook: aCCodeGen
	"Perform any last-minute changes to the code generator immediately
	 before it performs code analysis and generation.  In this case, make
	 all non-exported methods private."
	| publicMethodNames |
	self primitiveTable do:
		[:s|
		(s isSymbol and: [s ~~ #primitiveFail]) ifTrue:
			[(aCCodeGen methodNamed: s) returnType: #void]].
	publicMethodNames := (self requiredMethodNames: aCCodeGen options)
								copyWithoutAll: (self primitiveTable
														copyWithout: #primitiveFail).
	aCCodeGen selectorsAndMethodsDo:
		[:s :m|
		(m export or: [m isAPIMethod or: [publicMethodNames includes: s]]) ifTrue:
			[m static: false]]! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 9/7/2020 13:35'!
preambleCCode
	^	
'/* Disable Intel compiler inlining of warning which is used for breakpoints */
#pragma auto_inline(off)
sqInt warnpid, erroronwarn;
EXPORT(void)
warning(const char *s) { /* Print an error message but don''t necessarily exit. */
	if (erroronwarn) error(s);
	if (warnpid)
		printf("\n%s pid %ld\n", s, (long)warnpid);
	else
		printf("\n%s\n", s);
}
EXPORT(void)
warningat(const char *s, int l) { /* ditto with line number. */
	/* use alloca to call warning so one does not have to remember to set two breakpoints... */
	char *sl = alloca(strlen(s) + 16);
	sprintf(sl, "%s %d", s, l);
	warning(sl);
}
#pragma auto_inline(on)
'! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 5/1/2014 08:02'!
prepareToBeAddedToCodeGenerator: aCodeGen
	"It is either this or scan cmacro methods for selectors."
	aCodeGen retainMethods: #(enterSmalltalkExecutiveImplementation)! !

!StackInterpreter class methodsFor: 'constants' stamp: 'eem 1/15/2015 12:40'!
primitiveAccessorDepthTable
	| cg |
	cg := CCodeGenerator new.
	cg vmClass: StackInterpreter.
	^self primitiveTable collect:
		[:thing| | implementingClass method |
		(thing isInteger "quick prims, 0 for fast primitve fail"
		 or: [thing == #primitiveFail
		 or: [(implementingClass := self primitivesClass whichClassIncludesSelector: thing) isNil]])
			ifTrue: [-1]
			ifFalse:
				[method := cg compileToTMethodSelector: thing in: implementingClass.
				 cg accessorDepthForMethod: method]]! !

!StackInterpreter class methodsFor: 'constants' stamp: 'eem 9/6/2014 11:39'!
primitiveTable
	| cg |
	cg := CCodeGenerator new.
	cg vmClass: self.
	^PrimitiveTable collect:
		[:thing|
		(thing isInteger "quick prims, 0 for fast primitve fail"
		 or: [thing == #primitiveFail])
			ifTrue: [thing]
			ifFalse:
				[(self primitivesClass whichClassIncludesSelector: thing)
					ifNil: [#primitiveFail]
					ifNotNil:
						[:class|
						 (cg shouldIncludeMethodFor: class selector: thing)
							ifTrue: [thing]
							ifFalse: [#primitiveFail]]]]! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 6/24/2012 17:46'!
primitiveTableString
	"StackInterpreter initializePrimitiveTable primitiveTableString"
	^self primitiveTableStringFor: self primitiveTable! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 6/24/2012 17:45'!
primitiveTableStringFor: table
	"StackInterpreter initializePrimitiveTable primitiveTableString"
	^String streamContents:
		[:s | 
		s nextPut: ${.
		table withIndexDo:
			[:primSpec :index |
			s
				crtab;
				nextPutAll: '/* ';
				print: index - 1;
				nextPutAll: ' */ ';
				nextPutAll: (primSpec isString
								ifTrue: [primSpec == #primitiveFail
											ifTrue: ['(void (*)(void))0']
											ifFalse: [primSpec]]
								ifFalse: ['(void (*)(void))', primSpec printString]);
				nextPut: $,].
			s cr; nextPutAll: ' 0 }']! !

!StackInterpreter class methodsFor: 'accessing class hierarchy' stamp: 'eem 6/30/2014 05:56'!
primitivesClass
	^StackInterpreterPrimitives! !

!StackInterpreter class methodsFor: 'miscellaneous' stamp: 'eem 04/27/2017 12:55'!
reorganizeAsISeeFit
	"StackInterpreter reorganizeAsISeeFit"
	"CoInterpreter reorganizeAsISeeFit"
	| initializations bytecodes privates remainder |
	initializations := OrderedCollection new.
	bytecodes := OrderedCollection new.
	privates := OrderedCollection new.
	remainder := OrderedCollection new.
	self organization categories do:
		[:cat|
		((cat includesSubstring: 'initializ')
			ifTrue: [initializations]
			ifFalse:
				[((cat endsWith: 'bytecodes') or: [cat includesSubstring: 'selector sends'])
					ifTrue: [bytecodes]
					ifFalse:
						[(cat includesSubstring: 'private')
							ifTrue: [privates]
							ifFalse: [remainder]]])
			add: cat].
	self organization categories: initializations sort, bytecodes sort, remainder sort, privates sort! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 4/13/2021 17:36'!
requiredMethodNames: options
	"Answer the list of method names that should be retained for export or other support reasons"
	| requiredList |
	"A number of methods required by VM support code, specific platforms, etc"
	requiredList := #(
		assertValidExecutionPointe:r:s:
		characterForAscii:
		findClassOfMethod:forReceiver: findSelectorOfMethod:
			forceInterruptCheck forceInterruptCheckFromHeartbeat fullDisplayUpdate
		getCurrentBytecode getFullScreenFlag getInterruptKeycode getInterruptPending
			getSavedWindowSize getThisSessionID
		interpret
		loadInitialContext
		primitiveFail primitiveFailFor: primitiveFlushExternalPrimitives printAllStacks printCallStack printContext:
			printExternalHeadFrame printFramesInPage: printFrame: printMemory printOop:
				printStackPages printStackPageList printStackPagesInUse printStackPageListInUse
		readImageFromFile:HeapSize:StartingAt:
		setFullScreenFlag: setInterruptKeycode: setInterruptPending: setInterruptCheckChain:
			setSavedWindowSize: success:
		validInstructionPointer:inMethod:framePointer:) asSet.

	"Nice to actually have all the primitives available"
	requiredList addAll: (self primitiveTable select: [:each| each isSymbol]).

	"InterpreterProxy is the internal analogue of sqVirtualMachine.c, so make sure to keep all those"
	InterpreterProxy organization categories do:
		[:cat |
		((cat ~= 'initialize') and: [cat ~= 'private']) ifTrue:
			[requiredList addAll: (InterpreterProxy organization listAtCategoryNamed: cat)]].

	^requiredList! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 2/9/2009 12:41'!
sourceFileName
	^'interp.c'! !

!StackInterpreter class methodsFor: 'initialization' stamp: 'eem 01/19/2021 20:40'!
table: anArray from: specArray 
	"SpecArray is an array of one of (index selector) or (index1 
	 index2 selector) or (index nil) or (index1 index2 nil).  If selector
	 then the entry is the selector, but if nil the entry is the index."
	| contiguous |
	contiguous := 0.
	specArray do:
		[:spec | 
		(spec at: 1) = contiguous ifFalse:
			[self error: 'Non-contiguous table entry'].
		spec size = 2
			ifTrue:
				[anArray
					at: (spec at: 1) + 1
					put: ((spec at: 2) ifNil: [spec at: 1] ifNotNil: [:sym| sym]).
				 contiguous := contiguous + 1]
			ifFalse:
				[(spec at: 1) to: (spec at: 2) do:
					[:i | anArray at: i + 1 put: ((spec at: 3) ifNil: [i] ifNotNil: [:sym| sym])].
				 contiguous := contiguous + ((spec at: 2) - (spec at: 1)) + 1]].
	anArray withIndexDo:
		[:entry :index|
		entry isSymbol ifTrue:
			[(self shouldIncludeMethodForSelector: entry) ifFalse:
				[anArray at: index put: 0]]]! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 6/11/2010 12:20'!
translationClass
	"Return te class to use as the interpreterCLass when translating.  For the all-in-one
	 VMs that inherit from ObjectMemory this is the receiver.  But for the separate VMs
	 where most primitives are in a subclass it will be the subclass with the primitives."
	^self subclasses detect: [:s| s name = (self name, 'Primitives')] ifNone: [self]! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 12/25/2010 22:38'!
vmCallbackHeader
	^String streamContents:
		[:s|
		s nextPutAll: '#define VM_CALLBACK_INC 1'; cr; cr.
		VMCallbackContext printTypedefOn: s.
		s cr]! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 10/13/2008 14:28'!
wantsLabels
	"Only label the VM, not plugins (internal or external).  This to cut down the scope
	 of problems with labels being duplicated by C compiler optimizer inlining and
	 loop unrolling.  We use the asmLabel: directive to control labelling in the
	 interpreter proper. But it is too much work doing that for plugins too."
	^true! !

!StackInterpreter class methodsFor: 'translation' stamp: 'eem 11/9/2020 18:50'!
writeVMHeaderTo: aStream bytesPerWord: bytesPerWord generator: aCCodeGenerator
	super writeVMHeaderTo: aStream bytesPerWord: bytesPerWord generator: aCCodeGenerator.
	SistaVM ifTrue:
		[aCCodeGenerator putDefineOf: #SistaVM as: 1 on: aStream].
	NewspeakVM ifTrue:
		[aCCodeGenerator putDefineOf: #NewspeakVM as: 1 on: aStream].
	MULTIPLEBYTECODESETS ifTrue:
		[aCCodeGenerator putDefineOf: #MULTIPLEBYTECODESETS as: 1 on: aStream].
	IMMUTABILITY ifTrue:
		[aCCodeGenerator
			putConditionalDefineOf: #IMMUTABILITY
			as: 1
			comment: 'Allow this to be overridden on the compiler command line'
			on: aStream].
	SistaVM | NewspeakVM | MULTIPLEBYTECODESETS | IMMUTABILITY ifTrue:
		[aStream cr].
	aCCodeGenerator putDefineOf: #STACKVM as: 1 on: aStream.
	(InitializationOptions at: #SpurObjectMemory ifAbsent: false) ifTrue:
		[aCCodeGenerator putDefineOf: #SPURVM as: 1 on: aStream].

	aCCodeGenerator
		putDefineOf: #DisownVMForFFICall as: DisownVMForFFICall on: aStream;
		putDefineOf: #DisownVMForThreading as: DisownVMForThreading on: aStream! !

!StackInterpreter methodsFor: 'simulation' stamp: 'eem 12/21/2017 14:59'!
ISA
	<doNotGenerate>
	^self class initializationOptions
		at: #ISA
		ifAbsent: [Smalltalk wordSize = 8
					ifTrue: [#X64]
					ifFalse: [#IA32]]! !

!StackInterpreter methodsFor: 'simulation' stamp: 'eem 9/1/2020 14:22'!
_longjmp: aJumpBuf _: returnValue
	"Hack simulation of _setjmp/_longjmp, intended to invoke the most minimal setjmp/longjmp
	 pair available on the platform; no saving/restoring signal masks, no stack unwinding, etc.
	 Signal the exception that simulates a longjmp back to the interpreter." 
	<doNotGenerate>
	aJumpBuf == reenterInterpreter ifTrue:
		[self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer imbar: true line: nil].
	aJumpBuf returnValue: returnValue; signal! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 10/30/2020 10:24'!
_setjmp: aJumpBuf
	"Hack simulation of _setjmp/_longjmp, intended to invoke the most minimal setjmp/longjmp
	 pair available on the platform; no saving/restoring signal masks, no stack unwinding, etc." 
	<doNotGenerate>
	self assert: (#(ReenterInterpreter ReenterThreadSchedulingLoop) includes: aJumpBuf class name).
	^aJumpBuf returnValue ifNil: [0]! !

!StackInterpreter methodsFor: 'interpreter shell' stamp: 'eem 10/29/2020 10:13'!
aboutToDispatchBytecode
	"This is a hook for the simulator; null in production"
	<inline: #always>! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 11/16/2012 11:08'!
abstractDetailedSymbolicMethod: aMethod
	<doNotGenerate>
	| ts prim |
	(ts := self transcript) ensureCr.
	(prim := self primitiveIndexOf: aMethod) > 0 ifTrue:
		[ts nextPutAll: '<primitive: '; print: prim; nextPut: $>.
		(self isQuickPrimitiveIndex: prim) ifTrue:
			[ts nextPutAll: ' quick method'; cr; flush.
			 ^self].
		ts cr].
	(RelativeDetailedInstructionPrinter
			on: (VMCompiledMethodProxy new
					for: aMethod
					coInterpreter: self
					objectMemory: objectMemory))
		indent: 0;
		printInstructionsOn: ts.
	ts flush! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 6/26/2015 14:53'!
accessModifierOfMethod: methodObj
	<option: #NewspeakVM>
	<inline: true>
	^self accessModifierOfMethodHeader: (objectMemory methodHeaderOf: methodObj)! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 6/26/2015 14:40'!
accessModifierOfMethodHeader: header
	<option: #NewspeakVM>
	<inline: true>
	"accessModifier bits:
		 00 public
		 01 private
		 10 protected
		 11 undefined"
	^EnforceAccessControl
		ifTrue: [header >> MethodHeaderFlagBitPosition bitAnd: 3]
		ifFalse: [AccessModifierPublic]! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 9/1/2020 14:26'!
activateFailingPrimitiveMethod
	"Assuming the primFailCode (and any other relevant failure state) has been set,
	 switch the VM to the interpreter if necessary (if in the CoInterpreter executing machine code),
	 and activate the newMethod (which is expected to have a primitive)."
	self assert: primFailCode ~= 0.
	self assert: (objectMemory addressCouldBeObj: newMethod).
	self assert: (objectMemory isCompiledMethod: newMethod).
	self assert: (self primitiveIndexOf: newMethod) ~= 0.
	self justActivateNewMethod: true. "Frame must be interpreted"
	self _longjmp: reenterInterpreter _: ReturnToInterpreter.
	"NOTREACHED"
	^nil! !

!StackInterpreter methodsFor: 'control primitives' stamp: 'eem 4/23/2017 11:34'!
activateNewClosure: blockClosure outer: outerContext method: theMethod  numArgs: numArgs mayContextSwitch: mayContextSwitch
	"Similar to activateNewMethod but for Closure and newMethod."
	| numCopied closureIP |
	<inline: true>
	self assert: (objectMemory isContext: outerContext).
	numCopied := self copiedValueCountOfClosure: blockClosure.
	self assert: theMethod = (objectMemory fetchPointer: MethodIndex ofObject: outerContext).
	self assert: (objectMemory isOopCompiledMethod: theMethod).

	self push: instructionPointer.
	self push: framePointer.
	framePointer := stackPointer.
	self push: theMethod.
	self push: (self encodeFrameFieldHasContext: false isBlock: true numArgs: numArgs).
	self push: objectMemory nilObject. "FxThisContext field"
	"Because inst var access is not checked, we must follow the receiver in Spur to ensure it is valid."
	self push: (objectMemory followField: ReceiverIndex ofObject: outerContext).

	"Copy the copied values..."
	0 to: numCopied - 1 do:
		[:i|
		self push: (objectMemory
					fetchPointer: i + ClosureFirstCopiedValueIndex
					ofObject: blockClosure)].

	self assert: (self frameIsBlockActivation: framePointer).
	self assert: (self frameHasContext: framePointer) not.

	"The initial instructions in the block nil-out remaining temps."

	"the instruction pointer is a pointer variable equal to 
	method oop + ip + BaseHeaderSize 
	-1 for 0-based addressing of fetchByte 
	-1 because it gets incremented BEFORE fetching currentByte"
	closureIP := self quickFetchInteger: ClosureStartPCIndex ofObject: blockClosure.
	instructionPointer := theMethod + closureIP + objectMemory baseHeaderSize - 2.
	self setMethod: theMethod.

	"Now check for stack overflow or an event (interrupt, must scavenge, etc)"
	stackPointer < stackLimit ifTrue:
		[self handleStackOverflowOrEventAllowContextSwitch: mayContextSwitch]! !

!StackInterpreter methodsFor: 'control primitives' stamp: 'eem 4/23/2017 11:41'!
activateNewFullClosure: blockClosure method: theMethod numArgs: numArgs mayContextSwitch: mayContextSwitch
	"Similar to activateNewMethod but for Closure and newMethod."
	| numCopied methodHeader numTemps |
	<inline: true>
	self assert: theMethod = (objectMemory fetchPointer: FullClosureCompiledBlockIndex ofObject: blockClosure).
	numCopied := self copiedValueCountOfFullClosure: blockClosure.
	self push: instructionPointer.
	self push: framePointer.
	framePointer := stackPointer.
	self push: theMethod.
	self push: (self encodeFrameFieldHasContext: false isBlock: true numArgs: numArgs).
	self push: objectMemory nilObject. "FxThisContext field"
	"Because inst var access is not checked, we must follow the receiver in Spur to ensure it is valid."
	self push: (objectMemory followField: FullClosureReceiverIndex ofObject: blockClosure).

	"Copy the copied values..."
	0 to: numCopied - 1 do:
		[:i|
		self push: (objectMemory
					fetchPointer: i + FullClosureFirstCopiedValueIndex
					ofObject: blockClosure)].

	self assert: (self frameIsBlockActivation: framePointer).
	self assert: (self frameHasContext: framePointer) not.

	methodHeader := objectMemory methodHeaderOf: theMethod.
	numTemps := self temporaryCountOfMethodHeader: methodHeader.

	numArgs + numCopied + 1 to: numTemps do: [ :i | self push: objectMemory nilObject].

	instructionPointer := (self initialIPForHeader: methodHeader method: theMethod) - 1.
	
	self setMethod: theMethod.

	"Now check for stack overflow or an event (interrupt, must scavenge, etc)"
	stackPointer < stackLimit ifTrue:
		[self handleStackOverflowOrEventAllowContextSwitch: mayContextSwitch]! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 8/23/2018 11:14'!
activateNewMethod
	| methodHeader |
	methodHeader := self justActivateNewMethod: false. "either interpreted or machine code"

	"Now check for stack overflow or an event (interrupt, must scavenge, etc)."
	stackPointer < stackLimit ifTrue:
		[self handleStackOverflowOrEventAllowContextSwitch: (self canContextSwitchIfActivating: newMethod header: methodHeader)]! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 4/13/2021 17:39'!
activeProcess
	"Answer the current activeProcess."
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	^objectMemory fetchPointer: ActiveProcessIndex ofObject: self schedulerPointer! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 6/29/2014 19:32'!
addFirstLink: proc toList: aList 
	"Add the given process to the front of the given linked list
	 and set the backpointer of process to its new list."
	| firstLink |
	self assert: (objectMemory isForwarded: aList) not.
	self assert: (objectMemory fetchPointer: NextLinkIndex ofObject: proc) = objectMemory nilObject.
	firstLink := objectMemory fetchPointer: FirstLinkIndex ofObject: aList.
	self assert: firstLink ~= proc.
	objectMemory storePointer: FirstLinkIndex ofObject: aList withValue: proc.
	firstLink = objectMemory nilObject "a.k.a. (self isEmptyList: aList)"
		ifTrue: [objectMemory storePointer: LastLinkIndex ofObject: aList withValue: proc]
		ifFalse: [objectMemory storePointer: NextLinkIndex ofObject: proc withValue: firstLink].
	objectMemory storePointer: MyListIndex ofObject: proc withValue: aList! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 5/25/2017 16:07'!
addIdleUsecs: idleUsecs
	"The various poll/select calls in the VM should attempt to tally the ammount
	 of time spent at idle here, so as to render the uptime value meaningful."
	<api>
	statIdleUsecs := statIdleUsecs + idleUsecs! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 7/6/2015 10:49'!
addLastLink: proc toList: aList 
	"Add the given process to the end of the given linked list
	 and set the backpointer of process to its new list."
	| lastLink |
	self deny: (objectMemory isForwarded: proc).
	self deny: (objectMemory isForwarded: aList).
	self assert: (objectMemory fetchPointer: NextLinkIndex ofObject: proc) = objectMemory nilObject.
	(self isEmptyList: aList)
		ifTrue: [objectMemory storePointer: FirstLinkIndex ofObject: aList withValue: proc]
		ifFalse:
			[lastLink := objectMemory fetchPointer: LastLinkIndex ofObject: aList.
			 self assert: lastLink ~= proc.
			 objectMemory storePointer: NextLinkIndex ofObject: lastLink withValue: proc].
	objectMemory storePointer: LastLinkIndex ofObject: aList withValue: proc.
	objectMemory storePointer: MyListIndex ofObject: proc withValue: aList! !

!StackInterpreter methodsFor: 'method lookup cache' stamp: 'nice 11/6/2016 14:42'!
addNewMethodToCache: classObj
	"Add the given entry to the method cache.
	The policy is as follows:
		Look for an empty entry anywhere in the reprobe chain.
		If found, install the new entry there.
		If not found, then install the new entry at the first probe position
			and delete the entries in the rest of the reprobe chain.
		This has two useful purposes:
			If there is active contention over the first slot, the second
				or third will likely be free for reentry after ejection.
			Also, flushing is good when reprobe chains are getting full."
	| probe hash primitiveIndex |
	<inline: false>
	hash := objectMemory methodCacheHashOf: messageSelector with: (objectMemory classTagForClass: classObj).  "shift drops low-order zeros from addresses"
	(objectMemory isOopCompiledMethod: newMethod)
		ifTrue:
			[primitiveIndex := self primitiveIndexOf: newMethod.
			 primitiveFunctionPointer := self functionPointerFor: primitiveIndex inClass: classObj]
		ifFalse:
			[self assert: ((objectMemory isNonImmediate: newMethod)
						  and: [objectMemory isForwarded: newMethod]) not.
			 primitiveFunctionPointer := #primitiveInvokeObjectAsMethod].

	0 to: CacheProbeMax-1 do:
		[:p | probe := (hash >> p) bitAnd: MethodCacheMask.
		(methodCache at: probe + MethodCacheSelector) = 0 ifTrue:
			["Found an empty entry -- use it"
			methodCache at: probe + MethodCacheSelector put: messageSelector.
			methodCache at: probe + MethodCacheClass put: (objectMemory classTagForClass: classObj).
			methodCache at: probe + MethodCacheMethod put: newMethod.
			methodCache at: probe + MethodCachePrimFunction put: (self cCoerce: primitiveFunctionPointer to: #'sqIntptr_t').
			lastMethodCacheProbeWrite := probe. "this for primitiveExternalMethod"
			^self]].

	"OK, we failed to find an entry -- install at the first slot..."
	probe := hash bitAnd: MethodCacheMask.  "first probe"
	methodCache at: probe + MethodCacheSelector put: messageSelector.
	methodCache at: probe + MethodCacheClass put: (objectMemory classTagForClass: classObj).
	methodCache at: probe + MethodCacheMethod put: newMethod.
	methodCache at: probe + MethodCachePrimFunction put: (self cCoerce: primitiveFunctionPointer to: #'sqIntptr_t').
	lastMethodCacheProbeWrite := probe. "this for primitiveExternalMethod"

	"...and zap the following entries"
	1 to: CacheProbeMax-1 do:
		[:p | probe := (hash >> p) bitAnd: MethodCacheMask.
		methodCache at: probe + MethodCacheSelector put: 0]! !

!StackInterpreter methodsFor: 'method lookup cache' stamp: 'nice 11/6/2016 14:42'!
addNewMethodToNSCache: rule
	<option: #NewspeakVM>
	<inline: false>
	| classObj probe hash primitiveIndex |
	classObj := lkupClass.
	hash := (objectMemory methodCacheHashOf: messageSelector with: lkupClassTag) bitXor: (method bitXor: rule).
	self deny: rule = LookupRuleOrdinary.

	(objectMemory isOopCompiledMethod: newMethod)
		ifTrue:
			[primitiveIndex := self primitiveIndexOf: newMethod.
			 primitiveFunctionPointer := self functionPointerFor: primitiveIndex inClass: classObj]
		ifFalse:
			[self assert: ((objectMemory isNonImmediate: newMethod)
						  and: [objectMemory isForwarded: newMethod]) not.
			 primitiveFunctionPointer := #primitiveInvokeObjectAsMethod].

	0 to: CacheProbeMax-1 do:
		[:p | probe := (hash >> p) bitAnd: NSMethodCacheMask.
		(nsMethodCache at: probe + NSMethodCacheSelector) = 0 ifTrue:
			["Found an empty entry -- use it"
			nsMethodCache at: probe + NSMethodCacheSelector put: messageSelector.
			nsMethodCache at: probe + NSMethodCacheClassTag put: lkupClassTag. "(objectMemory classTagForClass: classObj)."
			nsMethodCache at: probe + NSMethodCacheCallingMethod put: method.
			nsMethodCache at: probe + NSMethodCacheDepthOrLookupRule put: rule.
			nsMethodCache at: probe + NSMethodCacheTargetMethod put: newMethod.
			nsMethodCache at: probe + NSMethodCachePrimFunction put: (self cCoerce: primitiveFunctionPointer to: #'sqIntptr_t').
			nsMethodCache at: probe + NSMethodCacheActualReceiver put: localAbsentReceiverOrZero.
			"lastMethodCacheProbeWrite := probe." "this for primitiveExternalMethod"
			^self]].

	"OK, we failed to find an entry -- install at the first slot..."
	probe := hash bitAnd: NSMethodCacheMask.  "first probe"
	nsMethodCache at: probe + NSMethodCacheSelector put: messageSelector.
	nsMethodCache at: probe + NSMethodCacheClassTag put: lkupClassTag. "(objectMemory classTagForClass: classObj)."
	nsMethodCache at: probe + NSMethodCacheCallingMethod put: method.
	nsMethodCache at: probe + NSMethodCacheDepthOrLookupRule put: rule.
	nsMethodCache at: probe + NSMethodCacheTargetMethod put: newMethod.
	nsMethodCache at: probe + NSMethodCachePrimFunction put: (self cCoerce: primitiveFunctionPointer to: #'sqIntptr_t').
	nsMethodCache at: probe + NSMethodCacheActualReceiver put: localAbsentReceiverOrZero.
	"lastMethodCacheProbeWrite := probe. ""this for primitiveExternalMethod"

	"...and zap the following entries"
	1 to: CacheProbeMax-1 do:
		[:p | probe := (hash >> p) bitAnd: NSMethodCacheMask.
		nsMethodCache at: probe + NSMethodCacheSelector put: 0]! !

!StackInterpreter methodsFor: 'plugin support' stamp: 'eem (auto pragmas 12/08) 7/14/2008 12:30'!
addToExternalPrimitiveTable: functionAddress
	"Add the given function address to the external primitive table and return the index where it's stored. This function doesn't need to be fast since it is only called when an external primitive has been looked up (which takes quite a bit of time itself). So there's nothing specifically complicated here.
	Note: Return index will be one-based (ST convention)"

	<var: #functionAddress declareC: 'void *functionAddress'>

	externalPrimitiveTableFirstFreeIndex to: MaxExternalPrimitiveTableSize-1 do:
		[ :i |
		(externalPrimitiveTable at: i) = 0 ifTrue:
			[externalPrimitiveTable at: i put: functionAddress.
			^externalPrimitiveTableFirstFreeIndex := i+1]].
	"if no space left, return zero so it'll looked up again"
	^0! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 9/13/2013 11:38'!
addressCouldBeClassObj: maybeClassObj
	"Answer if maybeClassObj looks like a class object"
	<inline: false>
	^(objectMemory addressCouldBeObj: maybeClassObj)
	  and: [self objCouldBeClassObj: maybeClassObj]! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 2/4/2014 11:13'!
allAccessibleObjectsOkay
	"Ensure that all accessible objects in the heap are okay."
	| ok |
	ok := true.
	objectMemory allObjectsDoSafely:
		[:oop|
		ok := ok & (self okayFields: oop)].
	^ok! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 12/15/2015 09:43'!
allOnesAsCharStar
	<inline: true>
	^(self
		cCoerceSimple: (objectMemory wordSize = 8
								ifTrue: [-1 signedIntToLong64]
								ifFalse: [-1 signedIntToLong])
		to: #'char *')! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'rsf 8/9/2016 23:42'!
allocateLowcodeCalloutState
	<option: #LowcodeVM>
	| calloutStateSize |
	calloutStateSize := self sizeof: #'sqLowcodeCalloutState'.
	shadowCallStackPointer := self cCoerce: ((self cCoerce: shadowCallStackPointer - calloutStateSize to: 'size_t') bitAnd: -16) to: 'char*'.
	lowcodeCalloutState := self cCoerce: shadowCallStackPointer to: #'sqLowcodeCalloutState*'.
	! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 11/14/2012 16:28'!
allocateMemory: heapSize minimum: minimumMemory imageFile: fileStream headerSize: headerSize
	"Translate to C function call with (case sensitive) camelCase. The purpose of this
	 method is to document the translation.
	 The default implementation is sqAllocateMemory(minimumMemory, heapSize). This may
	 be redefined to make use of the image file and header size parameters for efficient
	 implementation with mmap().
	 See CCodeGenerator>>writeDefaultMacrosOn: which specifies a default implementation."

	<inline: true>
	<returnTypeC: #'char *'>
	<var: #fileStream type: #sqImageFile>
	^self pointerForOop: (self
							allocateMemory: heapSize
							Minimum: minimumMemory
							ImageFile: fileStream
							HeaderSize: headerSize)! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 11/18/2014 18:26'!
alternateHeaderHasPrimitiveFlag: methodHeader
	<inline: true>
	^methodHeader anyMask: AlternateHeaderHasPrimFlag! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 12/12/2015 14:14'!
alternateHeaderNumLiteralsMask
	<api>
	<cmacro>
	^AlternateHeaderNumLiteralsMask! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 08/08/2014 10:42'!
argumentCountOf: methodPointer
	<api>
	^self argumentCountOfMethodHeader: (objectMemory methodHeaderOf: methodPointer)! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 9/10/2013 13:08'!
argumentCountOfClosure: closurePointer
	<api> "for Cogit"
	<inline: true>
	^self quickFetchInteger: ClosureNumArgsIndex ofObject: closurePointer! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 11/18/2014 17:54'!
argumentCountOfMethodHeader: header
	<api>
	^header >> MethodHeaderArgCountShift bitAnd: 16rF! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
arrayValueOf: arrayOop
	"Return the address of first indexable field of resulting array object, or fail if
	 the instance variable does not contain an indexable bytes or words object."
	"Note: May be called by translated primitive code."

	<returnTypeC: #'void *'>
	(objectMemory isWordsOrBytes: arrayOop) ifTrue:
		[^self cCoerceSimple: (self pointerForOop: arrayOop + objectMemory baseHeaderSize) to: #'void *'].
	self primitiveFail! !

!StackInterpreter methodsFor: 'indexing primitive support' stamp: 'eem 9/12/2013 03:16'!
asciiOfCharacter: characterObj  "Returns an integer object"

	<inline: false>
	(objectMemory isCharacterObject: characterObj) ifTrue:
		[^CharacterTable
			ifNil: [objectMemory integerObjectOfCharacterObject: characterObj]
			ifNotNil: [objectMemory fetchPointer: CharacterValueIndex ofObject: characterObj]].
	self primitiveFailFor: PrimErrBadArgument.
	^ConstZero  "in case some code needs an int"! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 7/16/2013 14:14'!
assertValidExecutionPointe: lip r: lifp s: lisp
	<inline: true>
	<var: #lip type: #'usqInt'>
	<var: #lifp type: #'char *'>
	<var: #lisp type: #'char *'>
	self assertValidExecutionPointe: lip r: lifp s: lisp imbar: (self isMachineCodeFrame: lifp) not line: #'__LINE__'! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 1/15/2015 17:31'!
assertValidExecutionPointe: lip r: lfp s: lsp imbar: inInterpreter line: ln
	<var: #lip type: #usqInt>
	<var: #lfp type: #'char *'>
	<var: #lsp type: #'char *'>
	self assert: inInterpreter l: ln.
	self assert: stackPage = stackPages mostRecentlyUsedPage l: ln.
	self assertValidStackLimits: ln.
	self assert: (stackPage addressIsInPage: lfp) l: ln.
	self assert: lsp < lfp l: ln.
	self assert: lfp > lsp l: ln.
	self assert: lsp >= (stackPage realStackLimit - self stackLimitOffset) l: ln.
	self assert: (lfp - lsp) / objectMemory bytesPerOop < LargeContextSlots l: ln.
	self assert: (self validInstructionPointer: lip inFrame: lfp) l: ln.
	self assert: ((self frameIsBlockActivation: lfp)
				or: [(self pushedReceiverOrClosureOfFrame: lfp) = (self frameReceiver: lfp)])
		l: ln.
	self assert: method = (self frameMethod: lfp) l: ln.
	self cppIf: MULTIPLEBYTECODESETS
		ifTrue: [self assert: (self methodUsesAlternateBytecodeSet: method) = (bytecodeSetSelector = 256)  l: ln].! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 7/16/2013 14:34'!
assertValidExecutionPointers
	<doNotGenerate> "simulator only"
	self assertValidExecutionPointe: localIP r: localFP s: localSP! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 12/15/2015 09:43'!
assertValidStackLimits: ln
	<returnTypeC: #void>
	"Order in the stackLimit checks is important because stackLimit is smashed by
	 interrupts. So always check for unsmashed value first to avoid race condition."
	self assert: (stackLimit = stackPage realStackLimit
				or: [stackLimit = self allOnesAsCharStar])
		l: ln.
	self assert: (stackPage stackLimit = stackPage realStackLimit
				or: [stackPage stackLimit = self allOnesAsCharStar])
		l: ln! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem 2/27/2013 13:50'!
baseFrameReturn
	"Return from a baseFrame (the bottom frame in a stackPage).  The context to
	 return to (which may be married) is stored in the saved instruction pointer slot."
	<inline: true>
	| contextToReturnTo isAContext theFP theSP thePage frameAbove |
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<var: #frameAbove type: #'char *'>
	contextToReturnTo := self frameCallerContext: localFP.
	isAContext := objectMemory isContext: contextToReturnTo.
	(isAContext
	 and: [self isStillMarriedContext: contextToReturnTo])
		ifTrue:
			[theFP := self frameOfMarriedContext: contextToReturnTo.
			 thePage := stackPages stackPageFor: theFP.
			 theFP = thePage headFP
				ifTrue:
					[theSP := thePage headSP.
					 stackPages freeStackPageNoAssert: stackPage "for a short time invariant is violated; assert follows"]
				ifFalse:
					["Returning to some interior frame, presumably because of a sender assignment.
					  Move the frames above to another page (they may be in use, e.g. via coroutining).
					  Make the interior frame the top frame."
					 frameAbove := self findFrameAbove: theFP inPage: thePage.
					 "Reuse the page we're exiting, which avoids allocating a new page and
					  manipulating the page list to mark the page we're entering as least recently
					  used (to avoid it being deallocated when allocating a new page)."
					 self moveFramesIn: thePage through: frameAbove toPage: stackPage.
					 theFP := thePage headFP.
					 theSP := thePage headSP]]
		ifFalse:
			[(isAContext
			  and: [objectMemory isIntegerObject: (objectMemory fetchPointer: InstructionPointerIndex ofObject: contextToReturnTo)]) ifFalse:
				[^self internalCannotReturn: localReturnValue].
			 thePage := self makeBaseFrameFor: contextToReturnTo.
			 theFP := thePage headFP.
			 theSP := thePage headSP.
			 stackPages freeStackPageNoAssert: stackPage "for a short time invariant is violated; assert follows"].
	self setStackPageAndLimit: thePage.
	self assert: (stackPages stackPageFor: theFP) = stackPage.
	localSP := theSP.
	localFP := theFP.
	self setMethod: (self frameMethod: localFP).
	localIP := self pointerForOop: self internalStackTop.
	self internalStackTopPut: localReturnValue.
	self assert: (self checkIsStillMarriedContext: contextToReturnTo currentFP: localFP).
	^self fetchNextBytecode! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 4/27/2016 07:51'!
bereaveAllMarriedContextsForSnapshotFlushingExternalPrimitivesIf: flushExtPrims
	"Enumerate all contexts and convert married contexts to widowed contexts so
	 that the snapshot contains only single contexts.  This allows the test for being
	 married to avoid checking for a context's frame pointer being in bounds.  If
	 flushExtPrims is true, flush references to external primitives in methods."
	objectMemory allObjectsDo:
		[:obj| | fmt |
		fmt := objectMemory formatOf: obj.
		(fmt = objectMemory indexablePointersFormat
		  and: [objectMemory isContextNonImm: obj]) ifTrue:
			[self makeContextSnapshotSafe: obj].
		 "Clean out external functions from compiled methods"
		 (flushExtPrims
		  and: [objectMemory isCompiledMethodFormat: fmt]) ifTrue:
			[self flushExternalPrimitiveOf: obj]]! !

!StackInterpreter methodsFor: 'sista inline primitives - binary' stamp: 'cb 12/13/2017 15:19'!
binaryAtInlinePrimitive: primIndex
	<option: #SistaVM>
	| result rec argIntAdjusted top |
	rec := self internalStackValue: 1.
	top := self internalStackTop.
	self deny: ((objectMemory isOopForwarded: rec) or: [(objectMemory isImmediate: rec)]).
	self assert: (objectMemory isIntegerObject: top).
	argIntAdjusted := (objectMemory integerValueOf: top) - 1.
	self assert: argIntAdjusted >= 0.
	primIndex caseOf: {
		"2064	pointerAt:
		Pointer object (Fixed sized or not) and not a context, Smi =>  (1-based, optimised if arg1 is a constant)"
		[64]	->	[self assert: (objectMemory isPointers: rec).
					 self assert: argIntAdjusted < (objectMemory numSlotsOfAny: rec).
					 result := objectMemory fetchPointer: argIntAdjusted ofObject: rec.].
		"2065	maybeContextPointerAt:
		 Pointer object (Fixed sized or not), Smi =>  (1-based, optimised if arg1 is a constant)"
		[65]	->	[ ((objectMemory isContextNonImm: rec) 
						 and: [self isMarriedOrWidowedContext: rec])
							ifTrue:
								[self externalizeIPandSP.
								 result := self externalInstVar: argIntAdjusted ofContext: rec.
								 self internalizeIPandSP]
					ifFalse: [result := objectMemory fetchPointer: argIntAdjusted ofObject: rec]
			].
		"2066	byteAt:
		 byte object, Smi => 8 bits unsigned Smi (1-based, optimised if arg1 is a constant)"
		[66]	->	[self assert: (objectMemory isBytes: rec).
					 self assert: argIntAdjusted < (objectMemory numBytesOf: rec).
					 result := objectMemory integerObjectOf: (objectMemory fetchByte: argIntAdjusted ofObject: rec)].
		"2067	shortAt:
		short object, Smi => 16 bits unsigned Smi (1-based, optimised if arg1 is a constant)"
		[67]	->	[self assert: (objectMemory isShorts: rec).
					 self assert: argIntAdjusted < (objectMemory num16BitUnitsOf: rec).
					 result := objectMemory integerObjectOf: (objectMemory fetchShort16: argIntAdjusted ofObject: rec)].
		"2068	wordAt:
		 word object, Smi => 32 bits unsigned Smi (1-based, optimised if arg1 is a constant)."
		[68]	->	[self assert: (objectMemory isWords: rec).
					 self assert: argIntAdjusted < (objectMemory num32BitUnitsOf: rec).
					 result := self positive32BitIntegerFor: (objectMemory fetchLong32: argIntAdjusted ofObject: rec)].
		"2069	doubleWordAt:
		 double word object, Smi => 64 bits unsigned Smi or LargePositiveInteger (1-based, optimised if arg1 is a constant)"
		[69]	->	[self assert: (objectMemory isLong64s: rec).
					 self assert: argIntAdjusted < (objectMemory num64BitUnitsOf: rec).
					 result := self positive64BitIntegerFor: (objectMemory fetchLong64: argIntAdjusted ofObject: rec)].}.
		self internalPop: 1; internalStackTopPut: result! !

!StackInterpreter methodsFor: 'sista inline primitives - binary' stamp: 'cb 10/24/2017 10:42'!
binaryCompInlinePrimitive: primIndex
	<option: #SistaVM>
	| result |
	self assert: (primIndex between: 38 and: 39).
	primIndex caseOf: {
		"2038	rawEqualsEquals:
		 not a forwarder, not a forwarder => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)"
		[38]	->	[result := ((self internalStackValue: 1) = self internalStackTop).].
		"2039	rawNotEqualsEquals:
		 not a forwarder, not a forwarder => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)"
		[39]	->	[result := ((self internalStackValue: 1) ~= self internalStackTop)].
	}.
	self internalPop: 1; internalStackTopPut: (objectMemory booleanObjectOf: result)! !

!StackInterpreter methodsFor: 'sista inline primitives - binary' stamp: 'cb 10/24/2017 11:28'!
binaryInlinePrimitive: primIndex
	<option: #SistaVM>
	"Bulk comments: each sub-method has its own comment with the specific case.
	2000	+
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2001	-
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2002	*
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2003	/
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2004	//
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2005	\
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2006	quo:
	Smi, Smi => Smi (no overflow, optimised if one operand is a constant)
	2016	bitAnd:
	Smi, Smi => Smi (optimised if one operand is a constant)
	2017	bitOr:
	Smi, Smi => Smi (optimised if one operand is a constant)
	2018	bitXor:
	Smi, Smi => Smi (optimised if one operand is a constant)
	2019	bitShiftLeft:
	Smi greater or equal to 0, Smi greater or equal to 0 => Smi (no overflow, optimised if arg1 is a constant)
	2020	bitShiftRight:
	Smi, Smi greater or equal to 0 => Smi (optimised if arg1 is a constant)
	2032	>
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2033	<
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2034	>=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2035	<=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2036	=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2037	~=
	Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2038	rawEqualsEquals:
	not a forwarder, not a forwarder => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2039	rawNotEqualsEquals:
	not a forwarder, not a forwarder => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)
	2048	rawNew:
	literal which is a fixed-sized behavior, Smi => instance of receiver, fields nilled out (optimised if arg1 is a constant)
	2049	rawNewNoInit:
	literal which is a fixed-sized behavior, Smi => instance of receiver (Fields of returned value contain undefined data, optimised if arg1 is a constant)
	2064	pointerAt:
	Pointer object (Fixed sized or not) and not a context, Smi =>  (1-based, optimised if arg1 is a constant)
	2065	maybeContextPointerAt:
	Pointer object (Fixed sized or not), Smi =>  (1-based, optimised if arg1 is a constant)
	2066	byteAt:
	byte object, Smi => 8 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2067	shortAt:
	short object, Smi => 16 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2068	wordAt:
	word object, Smi => 32 bits unsigned Smi (1-based, optimised if arg1 is a constant)
	2069	doubleWordAt:
	double word object, Smi => 64 bits unsigned Smi or LargePositiveInteger (1-based, optimised if arg1 is a constant)"
	primIndex <= 6 ifTrue: [^self binarySmiArithmeticInlinePrimitive: primIndex].
	primIndex < 16 ifTrue: [^self unknownInlinePrimitive].
	primIndex <= 20 ifTrue: [^self binarySmiBitInlinePrimitive: primIndex].
	primIndex < 32 ifTrue: [^self unknownInlinePrimitive].
	primIndex <= 37 ifTrue: [^self binarySmiCompInlinePrimitive: primIndex].
	primIndex <= 39 ifTrue: [^self binaryCompInlinePrimitive: primIndex].
	primIndex < 48 ifTrue: [^self unknownInlinePrimitive].
	primIndex <= 49 ifTrue: [^self binaryNewInlinePrimitive: primIndex].
	primIndex < 64 ifTrue: [^self unknownInlinePrimitive].
	primIndex <= 69 ifTrue: [^self binaryAtInlinePrimitive: primIndex].
	self unknownInlinePrimitive! !

!StackInterpreter methodsFor: 'sista inline primitives - binary' stamp: 'cb 12/11/2017 14:20'!
binaryNewInlinePrimitive: primIndex
	"2048	rawNew:
	literal which is a variable-sized behavior, Smi => instance of receiver, fields nilled/zeroed out (optimised if arg1 is a constant)
	2049	rawNewNoInit:
	literal which is a variable-sized behavior, Smi => instance of receiver (Fields of returned value contain undefined data, optimised if arg1 is a constant)
	
	WARNING: In the interpreter version, fields are always initialized."
	| classObj result size top |
	self assert: (primIndex between: 48 and: 49).
	top := self internalStackTop.
	classObj := self internalStackValue: 1.
	self assert: ((objectMemory isNonImmediate: classObj) and: [self objCouldBeClassObj: classObj]).
	self assert: (objectMemory isIntegerObject: top).
	size := self positiveMachineIntegerValueOf: top.
	result := objectMemory instantiateClass: classObj indexableSize: size.
	self internalPop: 1; internalStackTopPut: result! !

!StackInterpreter methodsFor: 'sista inline primitives - binary' stamp: 'cb 12/11/2017 14:19'!
binarySmiArithmeticInlinePrimitive: primIndex
	<option: #SistaVM>
	| result recInt argInt rcvr top |
	rcvr := self internalStackValue: 1.
	top := self internalStackTop.
	self assert: primIndex <= 6.
	self assert: (objectMemory isIntegerObject: rcvr).
	self assert: (objectMemory isIntegerObject: top).
	recInt := objectMemory integerValueOf: rcvr.
	argInt := objectMemory integerValueOf: top.
	primIndex caseOf: {
		"2000	+
		 Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[0]	->	[result := recInt + argInt].
		"2001	-
		 Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[1]	->	[result := recInt - argInt].
		"2002	*
		 Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[2]	->	[result := recInt * argInt].
		"2003	/
		 Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[3]	->	[result := recInt / argInt].
		"2004	//
		 Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[4]	->	[result := recInt // argInt].
		"2005	\
		 Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[5]	->	[result := recInt \\ argInt].
		"2006	quo:
		 Smi, Smi => Smi (no overflow, optimised if one operand is a constant)"
		[6]	->	[result := self quot: recInt ient: argInt].
	}.
	self internalPop: 1; internalStackTopPutIntegerObjectOf: result! !

!StackInterpreter methodsFor: 'sista inline primitives - binary' stamp: 'cb 12/12/2017 10:31'!
binarySmiBitInlinePrimitive: primIndex
	<option: #SistaVM>
	| result recInt argInt rcvr top |
	rcvr := self internalStackValue: 1.
	top := self internalStackTop.
	self assert: (primIndex between: 16 and: 20).
	self assert: (objectMemory isIntegerObject: rcvr).
	self assert: (objectMemory isIntegerObject: top).
	recInt := objectMemory integerValueOf: rcvr.
	argInt := objectMemory integerValueOf: top.
	primIndex caseOf: {
		"2016	bitAnd:
		 Smi, Smi => Smi (optimised if one operand is a constant)"
		[16]	->	[result := recInt bitAnd: argInt].
		"2017	bitOr:
		 Smi, Smi => Smi (optimised if one operand is a constant)"
		[17]	->	[result := recInt bitOr: argInt].
		"2018	bitXor:
		 Smi, Smi => Smi (optimised if one operand is a constant)"
		[18]	->	[result := recInt bitXor: argInt].
		"2019	bitShiftLeft:
		 Smi greater or equal to 0, Smi greater or equal to 0 => Smi (no overflow, optimised if arg1 is a constant)"
		[19]	->	[self assert: recInt >= 0.
					 self assert: argInt >= 0.
					 result := recInt << argInt].
		"2020	bitShiftRight:
		 Smi, Smi greater or equal to 0 => Smi (optimised if arg1 is a constant)"
		[20]	->	[self assert: argInt >= 0.
					 result := recInt >> argInt].
	}.
	self internalPop: 1; internalStackTopPutIntegerObjectOf: result! !

!StackInterpreter methodsFor: 'sista inline primitives - binary' stamp: 'cb 12/11/2017 14:18'!
binarySmiCompInlinePrimitive: primIndex
	<option: #SistaVM>
	| result top rcvr |
	self assert: (primIndex between: 32 and: 37).
	rcvr := self internalStackValue: 1.
	top := self internalStackTop.
	self assert: (objectMemory isIntegerObject: rcvr).
	self assert: (objectMemory isIntegerObject: top).
	primIndex caseOf: {
		"2032	>
		Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)"
		[32]	->	[result := rcvr > top].
		"2033	<
		Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)"
		[33]	->	[result := rcvr < top].
		"2034	>=
		Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)"
		[34]	->	[result := rcvr >= top].
		"2035	<=
		Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)"
		[35]	->	[result :=rcvr <= top].
		"2036	=
		Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)"
		[36]	->	[result := (rcvr = top).].
		"2037	~=
		Smi, Smi => Boolean (optimised if one operand is a constant, Pipelined with ifTrue:ifFalse:)"
		[37]	->	[result := (rcvr ~= top)].
	}.
	self internalPop: 1; internalStackTopPut: (objectMemory booleanObjectOf: result)! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 9/19/2010 00:17'!
booleanCheat: cond
	"cheat the interpreter out of the pleasure of handling the next bytecode IFF it is a jump-on-boolean. Which it is, often enough when the current bytecode is something like bytecodePrimEqual"
	<inline: true>

	cond
		ifTrue: [self booleanCheatTrue]
		ifFalse: [self booleanCheatFalse]! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 7/26/2014 06:56'!
booleanCheatFalse
	"cheat the interpreter out of the pleasure of handling the next bytecode IFF it is a jump-on-boolean. Which it is, often enough when the current bytecode is something like bytecodePrimEqual"
	<sharedCodeInCase: #bytecodePrimGreaterThan>
	| bytecode offset |

	bytecode := self fetchByte.  "assume next bytecode is jumpIfFalse (99%)"
	self internalPop: 2.
	(bytecode < 160 and: [bytecode > 151]) ifTrue:  "short jumpIfFalse"
		[^self jump: bytecode - 151].

	bytecode = 172 ifTrue:  "long jumpIfFalse"
		[offset := self fetchByte.
		^self jump: offset].

	"not followed by a jumpIfFalse; (un)do instruction fetch and push boolean result"
	self cppIf: MULTIPLEBYTECODESETS
		ifTrue: [currentBytecode := bytecode + bytecodeSetSelector]
		ifFalse: [currentBytecode := bytecode].
	self internalPush: objectMemory falseObject! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 7/26/2014 06:56'!
booleanCheatFalseSistaV1
	"cheat the interpreter out of the pleasure of handling the next bytecode IFF it is a jump-on-boolean. Which it is, often enough when the current bytecode is something like bytecodePrimEqual"
	<sharedCodeInCase: #bytecodePrimGreaterThanSistaV1>
	| bytecode offset |

	bytecode := self fetchByte.  "assume next bytecode is jumpIfFalse (99%)"
	self internalPop: 2.
	(bytecode < 199 and: [bytecode > 191]) ifTrue:  "short jumpIfFalse"
		[^self jump: bytecode - 191].

	bytecode = 239 ifTrue:  "long jumpIfFalse"
		[offset := self fetchByte.
		^self jump: offset].

	"not followed by a jumpIfFalse; (un)do instruction fetch and push boolean result"
	self cppIf: MULTIPLEBYTECODESETS
		ifTrue: [currentBytecode := bytecode + bytecodeSetSelector]
		ifFalse: [currentBytecode := bytecode].
	self internalPush: objectMemory falseObject! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 7/26/2014 06:56'!
booleanCheatFalseV4
	"cheat the interpreter out of the pleasure of handling the next bytecode IFF it is a jump-on-boolean. Which it is, often enough when the current bytecode is something like bytecodePrimEqual"
	<sharedCodeInCase: #bytecodePrimGreaterThanV4>
	| bytecode offset |

	bytecode := self fetchByte.  "assume next bytecode is jumpIfFalse (99%)"
	self internalPop: 2.
	(bytecode < 216 and: [bytecode > 207]) ifTrue:  "short jumpIfFalse"
		[^self jump: bytecode - 207].

	bytecode = 244 ifTrue:  "long jumpIfFalse"
		[offset := self fetchByte.
		^self jump: offset].

	"not followed by a jumpIfFalse; (un)do instruction fetch and push boolean result"
	self cppIf: MULTIPLEBYTECODESETS
		ifTrue: [currentBytecode := bytecode + bytecodeSetSelector]
		ifFalse: [currentBytecode := bytecode].
	self internalPush: objectMemory falseObject! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 07/25/2014 06:49'!
booleanCheatSistaV1: cond
	"cheat the interpreter out of the pleasure of handling the next bytecode IFF it is a jump-on-boolean. Which it is, often enough when the current bytecode is something like bytecodePrimEqual"
	<inline: true>

	cond
		ifTrue: [self booleanCheatTrueSistaV1]
		ifFalse: [self booleanCheatFalseSistaV1]! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 7/26/2014 06:56'!
booleanCheatTrue
	"cheat the interpreter out of the pleasure of handling the next bytecode IFF it is a jump-on-boolean. Which it is, often enough when the current bytecode is something like bytecodePrimEqual"
	<sharedCodeInCase: #bytecodePrimLessThan>
	| bytecode offset |

	bytecode := self fetchByte.  "assume next bytecode is jumpIfFalse (99%)"
	self internalPop: 2.
	(bytecode < 173 and: [bytecode > 151]) ifTrue:
		[bytecode < 160 ifTrue: "short jumpIfFalse 152 - 159"
			[^self fetchNextBytecode].
		bytecode = 172 ifTrue: "long jumpIfFalse"
			[self fetchByte.
			^self fetchNextBytecode].
		bytecode > 167 ifTrue: "long jumpIfTrue 168 - 171"
			[offset := bytecode - 168 << 8 + self fetchByte.
			^self jump: offset]].

	"not followed by a jumpIfFalse; (un)do instruction fetch and push boolean result"
	self cppIf: MULTIPLEBYTECODESETS
		ifTrue: [currentBytecode := bytecode + bytecodeSetSelector]
		ifFalse: [currentBytecode := bytecode].
	self internalPush: objectMemory trueObject! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 7/26/2014 06:55'!
booleanCheatTrueSistaV1
	"cheat the interpreter out of the pleasure of handling the next bytecode IFF it is a jump-on-boolean. Which it is, often enough when the current bytecode is something like bytecodePrimEqual"
	<sharedCodeInCase: #bytecodePrimLessThanSistaV1>
	| bytecode offset |

	bytecode := self fetchByte.  "assume next bytecode is jumpIfFalse (99%)"
	self internalPop: 2.
	bytecode >= 192 ifTrue:
		[bytecode <= 199 ifTrue: "short jumpIfFalse 192 - 199"
			[^self fetchNextBytecode].
		bytecode = 239 ifTrue: "long jumpIfFalse"
			[self fetchByte.
			^self fetchNextBytecode].
		bytecode = 238 ifTrue: "long jumpIfTrue 238"
			[offset := self fetchByte.
			^self jump: offset]].

	"not followed by a jumpIfFalse; (un)do instruction fetch and push boolean result"
	self cppIf: MULTIPLEBYTECODESETS
		ifTrue: [currentBytecode := bytecode + bytecodeSetSelector]
		ifFalse: [currentBytecode := bytecode].
	self internalPush: objectMemory trueObject! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 7/26/2014 06:55'!
booleanCheatTrueV4
	"cheat the interpreter out of the pleasure of handling the next bytecode IFF it is a jump-on-boolean. Which it is, often enough when the current bytecode is something like bytecodePrimEqual"
	<sharedCodeInCase: #bytecodePrimLessThanV4>
	| bytecode offset |

	bytecode := self fetchByte.  "assume next bytecode is jumpIfFalse (99%)"
	self internalPop: 2.
	bytecode >= 208 ifTrue:
		[bytecode <= 215 ifTrue: "short jumpIfFalse 208 - 215"
			[^self fetchNextBytecode].
		bytecode = 244 ifTrue: "long jumpIfFalse"
			[self fetchByte.
			^self fetchNextBytecode].
		bytecode = 243 ifTrue: "long jumpIfTrue 243"
			[offset := self fetchByte.
			^self jump: offset]].

	"not followed by a jumpIfFalse; (un)do instruction fetch and push boolean result"
	self cppIf: MULTIPLEBYTECODESETS
		ifTrue: [currentBytecode := bytecode + bytecodeSetSelector]
		ifFalse: [currentBytecode := bytecode].
	self internalPush: objectMemory trueObject! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 11/12/2012 13:50'!
booleanCheatV4: cond
	"cheat the interpreter out of the pleasure of handling the next bytecode IFF it is a jump-on-boolean. Which it is, often enough when the current bytecode is something like bytecodePrimEqual"
	<inline: true>

	cond
		ifTrue: [self booleanCheatTrueV4]
		ifFalse: [self booleanCheatFalseV4]! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
booleanValueOf: obj
"convert true and false (Smalltalk) to true or false(C)"
	obj = objectMemory trueObject ifTrue: [ ^ true ].
	obj = objectMemory falseObject ifTrue: [ ^ false ].
	self success: false.
	^ nil! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimAdd
	| rcvr arg result |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg)
		ifTrue: [result := (objectMemory integerValueOf: rcvr) + (objectMemory integerValueOf: arg).
				(objectMemory isIntegerValue: result) ifTrue:
					[self internalPop: 2 thenPush: (objectMemory integerObjectOf: result).
					^ self fetchNextBytecode "success"]]
		ifFalse: [self initPrimCall.
				self externalizeIPandSP.
				self primitiveFloatAdd: rcvr toArg: arg.
				self internalizeIPandSP.
				self successful ifTrue: [^ self fetchNextBytecode "success"]].

	messageSelector := self specialSelector: 0.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'rmacnak 5/19/2015 20:27'!
bytecodePrimAt
	"BytecodePrimAt will only succeed if the receiver is in the atCache.
	 Otherwise it will fail so that the more general primitiveAt will put it in the
	 cache after validating that message lookup results in a primitive response.
	 Override to insert in the at: cache here.  This is necessary since once there
	 is a compiled at: primitive method (which doesn't use the at: cache) the only
	 way something can get installed in the atCache is here."
	| index rcvr result atIx |
	index := self internalStackTop.
	rcvr := self internalStackValue: 1.
	((objectMemory isNonImmediate: rcvr)
	 and: [objectMemory isIntegerObject: index]) ifTrue:
		[atIx := rcvr bitAnd: AtCacheMask.  "Index into atCache = 4N, for N = 0 ... 7"
		(atCache at: atIx+AtCacheOop) ~= rcvr ifTrue:
			[lkupClassTag := objectMemory fetchClassTagOfNonImm: rcvr.
			 messageSelector := self specialSelector: 16.
			 (self lookupInMethodCacheSel: messageSelector classTag: lkupClassTag) ifFalse:
				[argumentCount := 1.
				 ^self commonSendOrdinary].
			 primitiveFunctionPointer == #primitiveAt
				ifTrue: [self install: rcvr inAtCache: atCache at: atIx string: false]
				ifFalse:
					[primitiveFunctionPointer == #primitiveStringAt
						ifTrue: [self install: rcvr inAtCache: atCache at: atIx string: true]
						ifFalse:
							[argumentCount := 1.
							 ^self commonSendOrdinary]]].
		 self successful ifTrue:
			[result := self commonVariable: rcvr at: (objectMemory integerValueOf: index) cacheIndex: atIx].
		 self successful ifTrue:
			[self fetchNextBytecode.
			 ^self internalPop: 2 thenPush: result].
		 self initPrimCall].

	messageSelector := self specialSelector: 16.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 10/5/2012 11:19'!
bytecodePrimAtEnd
	messageSelector := self specialSelector: 21.
	argumentCount := self specialSelectorNumArgs: 21.
	self normalSend.! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'cb 11/20/2015 14:20'!
bytecodePrimAtPut
	"BytecodePrimAtPut will only succeed if the receiver is in the atCache.
	Otherwise it will fail so that the more general primitiveAtPut will put it in the
	cache after validating that message lookup results in a primitive response.
	 Override to insert in the atCache here.  This is necessary since once there
	 is a compiled at:[put:] primitive method (which doesn't use the at: cache) the
	 only way something can get installed in the atCache is here."
	| index rcvr atIx value correctRcvr |
	value := self internalStackTop.
	index := self internalStackValue: 1.
	rcvr := self internalStackValue: 2.
	self cppIf: IMMUTABILITY 
		ifTrue: [ correctRcvr := objectMemory isOopMutable: rcvr ]
		ifFalse: [ correctRcvr := objectMemory isNonImmediate: rcvr ].
	(correctRcvr
	 and: [objectMemory isIntegerObject: index]) ifTrue:
		[atIx := (rcvr bitAnd: AtCacheMask) + AtPutBase.  "Index into atPutCache"
		 (atCache at: atIx+AtCacheOop) ~= rcvr ifTrue:
			[lkupClassTag := objectMemory fetchClassTagOfNonImm: rcvr.
			 messageSelector := self specialSelector: 17.
			 (self lookupInMethodCacheSel: messageSelector classTag: lkupClassTag) ifFalse:
				[argumentCount := 2.
				 ^self commonSendOrdinary].
			 primitiveFunctionPointer == #primitiveAtPut
				ifTrue: [self install: rcvr inAtCache: atCache at: atIx string: false]
				ifFalse:
					[primitiveFunctionPointer == #primitiveStringAtPut
						ifTrue: [self install: rcvr inAtCache: atCache at: atIx string: true]
						ifFalse:
							[argumentCount := 2.
							 ^self commonSendOrdinary]]].
		 self successful ifTrue:
			[self commonVariable: rcvr at: (objectMemory integerValueOf: index) put: value cacheIndex: atIx].
		 self successful ifTrue:
			[self fetchNextBytecode.
			 ^self internalPop: 3 thenPush: value].
		 self initPrimCall].

	messageSelector := self specialSelector: 17.
	argumentCount := 2.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 2/11/2015 12:34'!
bytecodePrimBitAnd
	| rcvr arg |
	arg := self internalStackTop.
	rcvr := self internalStackValue: 1.
	((objectMemory isIntegerObject: arg)
	 and: [objectMemory isIntegerObject: rcvr]) ifTrue:
		[self internalPop: 2 thenPush: (arg bitAnd: rcvr).
		 ^self fetchNextBytecode "success"].

	self initPrimCall.
	self externalizeIPandSP.
	self primitiveBitAnd.
	self internalizeIPandSP.
	self successful ifTrue:
		[^self fetchNextBytecode "success"].

	messageSelector := self specialSelector: 14.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 2/11/2015 12:36'!
bytecodePrimBitOr
	| rcvr arg |
	arg := self internalStackTop.
	rcvr := self internalStackValue: 1.
	((objectMemory isIntegerObject: arg)
	 and: [objectMemory isIntegerObject: rcvr]) ifTrue:
		[self internalPop: 2 thenPush: (arg bitOr: rcvr).
		 ^self fetchNextBytecode "success"].

	self initPrimCall.
	self externalizeIPandSP.
	self primitiveBitOr.
	self internalizeIPandSP.
	self successful ifTrue:
		[^self fetchNextBytecode "success"].

	messageSelector := self specialSelector: 15.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/25/2008 12:50'!
bytecodePrimBitShift

	self initPrimCall.
	self externalizeIPandSP.
	self primitiveBitShift.
	self internalizeIPandSP.
	self successful ifTrue: [^ self fetchNextBytecode "success"].

	messageSelector := self specialSelector: 12.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 10/3/2013 13:40'!
bytecodePrimClass
	| rcvr |
	rcvr := self internalStackTop.
	(objectMemory isOopForwarded: rcvr) ifTrue:
		[rcvr := self handleSpecialSelectorSendFaultFor: rcvr].
	self internalStackTopPut: (objectMemory fetchClassOf: rcvr).
	self fetchNextBytecode! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
bytecodePrimDiv
	| quotient |
	self initPrimCall.
	quotient := self doPrimitiveDiv: (self internalStackValue: 1) by: (self internalStackValue: 0).
	self successful ifTrue: [self internalPop: 2 thenPush: (objectMemory integerObjectOf: quotient).
		^ self fetchNextBytecode "success"].

	messageSelector := self specialSelector: 13.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimDivide
	| rcvr arg result |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg)
		ifTrue: [rcvr := objectMemory integerValueOf: rcvr.
			arg := objectMemory integerValueOf: arg.
			(arg ~= 0 and: [rcvr \\ arg = 0])
				ifTrue: [result := rcvr // arg.
					"generates C / operation"
					(objectMemory isIntegerValue: result)
						ifTrue: [self internalPop: 2 thenPush: (objectMemory integerObjectOf: result).
							^ self fetchNextBytecode"success"]]]
		ifFalse: [self initPrimCall.
			self externalizeIPandSP.
			self primitiveFloatDivide: rcvr byArg: arg.
			self internalizeIPandSP.
			self successful ifTrue: [^ self fetchNextBytecode"success"]].

	messageSelector := self specialSelector: 9.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 10/5/2012 11:19'!
bytecodePrimDo

	messageSelector := self specialSelector: 27.
	argumentCount := self specialSelectorNumArgs: 27.
	self normalSend.
! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimEqual
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue: [^self booleanCheat: rcvr = arg].

	self initPrimCall.
	aBool := self primitiveFloatEqual: rcvr toArg: arg.
	self successful ifTrue: [^self booleanCheat: aBool].

	messageSelector := self specialSelector: 6.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimEqualSistaV1
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue: [^self booleanCheatSistaV1: rcvr = arg].

	self initPrimCall.
	aBool := self primitiveFloatEqual: rcvr toArg: arg.
	self successful ifTrue: [^self booleanCheatSistaV1: aBool].

	messageSelector := self specialSelector: 6.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimEqualV4
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue: [^self booleanCheatV4: rcvr = arg].

	self initPrimCall.
	aBool := self primitiveFloatEqual: rcvr toArg: arg.
	self successful ifTrue: [^self booleanCheatV4: aBool].

	messageSelector := self specialSelector: 6.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimGreaterOrEqual
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue:
		["The C code can avoid detagging since tagged integers are still signed.
		 But this means the simulator must override to do detagging."
		^self cCode: [self booleanCheat: rcvr >= arg]
			inSmalltalk: [self booleanCheat: (objectMemory integerValueOf: rcvr) >= (objectMemory integerValueOf: arg)]].

	self initPrimCall.
	aBool := self primitiveFloatGreaterOrEqual: rcvr toArg: arg.
	self successful ifTrue: [^self booleanCheat: aBool].

	messageSelector := self specialSelector: 5.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimGreaterOrEqualSistaV1
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue:
		["The C code can avoid detagging since tagged integers are still signed.
		 But this means the simulator must override to do detagging."
		^self cCode: [self booleanCheatSistaV1: rcvr >= arg]
			inSmalltalk: [self booleanCheatSistaV1: (objectMemory integerValueOf: rcvr) >= (objectMemory integerValueOf: arg)]].

	self initPrimCall.
	aBool := self primitiveFloatGreaterOrEqual: rcvr toArg: arg.
	self successful ifTrue: [^self booleanCheatSistaV1: aBool].

	messageSelector := self specialSelector: 5.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimGreaterOrEqualV4
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue:
		["The C code can avoid detagging since tagged integers are still signed.
		 But this means the simulator must override to do detagging."
		^self cCode: [self booleanCheatV4: rcvr >= arg]
			inSmalltalk: [self booleanCheatV4: (objectMemory integerValueOf: rcvr) >= (objectMemory integerValueOf: arg)]].

	self initPrimCall.
	aBool := self primitiveFloatGreaterOrEqual: rcvr toArg: arg.
	self successful ifTrue: [^self booleanCheatV4: aBool].

	messageSelector := self specialSelector: 5.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimGreaterThan
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue:
		["The C code can avoid detagging since tagged integers are still signed.
		 But this means the simulator must override to do detagging."
		^self cCode: [self booleanCheat: rcvr > arg]
			inSmalltalk: [self booleanCheat: (objectMemory integerValueOf: rcvr) > (objectMemory integerValueOf: arg)]].

	self initPrimCall.
	aBool := self primitiveFloatGreater: rcvr thanArg: arg.
	self successful ifTrue: [^self booleanCheat: aBool].

	messageSelector := self specialSelector: 3.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimGreaterThanSistaV1
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue:
		["The C code can avoid detagging since tagged integers are still signed.
		 But this means the simulator must override to do detagging."
		^self cCode: [self booleanCheatSistaV1: rcvr > arg]
			inSmalltalk: [self booleanCheatSistaV1: (objectMemory integerValueOf: rcvr) > (objectMemory integerValueOf: arg)]].

	self initPrimCall.
	aBool := self primitiveFloatGreater: rcvr thanArg: arg.
	self successful ifTrue: [^self booleanCheatSistaV1: aBool].

	messageSelector := self specialSelector: 3.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimGreaterThanV4
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue:
		["The C code can avoid detagging since tagged integers are still signed.
		 But this means the simulator must override to do detagging."
		^self cCode: [self booleanCheatV4: rcvr > arg]
			inSmalltalk: [self booleanCheatV4: (objectMemory integerValueOf: rcvr) > (objectMemory integerValueOf: arg)]].

	self initPrimCall.
	aBool := self primitiveFloatGreater: rcvr thanArg: arg.
	self successful ifTrue: [^self booleanCheatV4: aBool].

	messageSelector := self specialSelector: 3.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 10/3/2013 14:15'!
bytecodePrimIdentical
	| rcvr arg |
	rcvr := self internalStackValue: 1.
	(objectMemory isOopForwarded: rcvr) ifTrue:
		[rcvr := self handleSpecialSelectorSendFaultFor: rcvr].
	arg := self internalStackValue: 0.
	(objectMemory isOopForwarded: arg) ifTrue:
		[arg := self handleSpecialSelectorSendFaultFor: arg].
	self booleanCheat: rcvr = arg! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 07/25/2014 06:49'!
bytecodePrimIdenticalSistaV1
	| rcvr arg |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory isOopForwarded: rcvr) ifTrue:
		[rcvr := self handleSpecialSelectorSendFaultFor: rcvr].
	(objectMemory isOopForwarded: arg) ifTrue:
		[arg := self handleSpecialSelectorSendFaultFor: arg].
	self booleanCheatSistaV1: rcvr = arg! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 10/3/2013 14:15'!
bytecodePrimIdenticalV4
	| rcvr arg |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory isOopForwarded: rcvr) ifTrue:
		[rcvr := self handleSpecialSelectorSendFaultFor: rcvr].
	(objectMemory isOopForwarded: arg) ifTrue:
		[arg := self handleSpecialSelectorSendFaultFor: arg].
	self booleanCheatV4: rcvr = arg! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimLessOrEqual
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue:
		["The C code can avoid detagging since tagged integers are still signed.
		 But this means the simulator must override to do detagging."
		^self cCode: [self booleanCheat: rcvr <= arg]
			inSmalltalk: [self booleanCheat: (objectMemory integerValueOf: rcvr) <= (objectMemory integerValueOf: arg)]].

	self initPrimCall.
	aBool := self primitiveFloatLessOrEqual: rcvr toArg: arg.
	self successful ifTrue: [^self booleanCheat: aBool].

	messageSelector := self specialSelector: 4.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimLessOrEqualSistaV1
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue:
		["The C code can avoid detagging since tagged integers are still signed.
		 But this means the simulator must override to do detagging."
		^self cCode: [self booleanCheatSistaV1: rcvr <= arg]
			inSmalltalk: [self booleanCheatSistaV1: (objectMemory integerValueOf: rcvr) <= (objectMemory integerValueOf: arg)]].

	self initPrimCall.
	aBool := self primitiveFloatLessOrEqual: rcvr toArg: arg.
	self successful ifTrue: [^self booleanCheatSistaV1: aBool].

	messageSelector := self specialSelector: 4.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimLessOrEqualV4
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue:
		["The C code can avoid detagging since tagged integers are still signed.
		 But this means the simulator must override to do detagging."
		^self cCode: [self booleanCheatV4: rcvr <= arg]
			inSmalltalk: [self booleanCheatV4: (objectMemory integerValueOf: rcvr) <= (objectMemory integerValueOf: arg)]].

	self initPrimCall.
	aBool := self primitiveFloatLessOrEqual: rcvr toArg: arg.
	self successful ifTrue: [^self booleanCheatV4: aBool].

	messageSelector := self specialSelector: 4.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimLessThan
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue:
		["The C code can avoid detagging since tagged integers are still signed.
		 But this means the simulator must override to do detagging."
		^self cCode: [self booleanCheat: rcvr < arg]
			inSmalltalk: [self booleanCheat: (objectMemory integerValueOf: rcvr) < (objectMemory integerValueOf: arg)]].

	self initPrimCall.
	aBool := self primitiveFloatLess: rcvr thanArg: arg.
	self successful ifTrue: [^ self booleanCheat: aBool].

	messageSelector := self specialSelector: 2.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimLessThanSistaV1
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue:
		["The C code can avoid detagging since tagged integers are still signed.
		 But this means the simulator must override to do detagging."
		^self cCode: [self booleanCheatSistaV1: rcvr < arg]
			inSmalltalk: [self booleanCheatSistaV1: (objectMemory integerValueOf: rcvr) < (objectMemory integerValueOf: arg)]].

	self initPrimCall.
	aBool := self primitiveFloatLess: rcvr thanArg: arg.
	self successful ifTrue: [^ self booleanCheatSistaV1: aBool].

	messageSelector := self specialSelector: 2.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimLessThanV4
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue:
		["The C code can avoid detagging since tagged integers are still signed.
		 But this means the simulator must override to do detagging."
		^self cCode: [self booleanCheatV4: rcvr < arg]
			inSmalltalk: [self booleanCheatV4: (objectMemory integerValueOf: rcvr) < (objectMemory integerValueOf: arg)]].

	self initPrimCall.
	aBool := self primitiveFloatLess: rcvr thanArg: arg.
	self successful ifTrue: [^ self booleanCheatV4: aBool].

	messageSelector := self specialSelector: 2.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 6/5/2017 10:40'!
bytecodePrimMakePoint
	"Inline primitiveMakePoint for the benefit of the interpreter. Externalizing the
	 sp & fp, then internalizing and testign for primitive failure add much overhead.
	 So simply inline the relatively small ammount of code directly."
	| rcvr arg pt |
	rcvr := self internalStackValue: 1.
	arg := self internalStackTop.
	((objectMemory isFloatOrInt: rcvr)
	 and: [objectMemory isFloatOrInt: arg]) ifTrue:
		[pt := objectMemory eeInstantiateSmallClass: (objectMemory splObj: ClassPoint) numSlots: YIndex + 1.
		 objectMemory "No need to check since new object is always new."
			storePointerUnchecked: XIndex ofObject: pt withValue: rcvr;
			storePointerUnchecked: YIndex ofObject: pt withValue: arg.
		 self internalPop: 2 thenPush: pt.
		 ^self fetchNextBytecode "success"].

	messageSelector := self specialSelector: 11.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
bytecodePrimMod
	| mod |
	self initPrimCall.
	mod := self doPrimitiveMod: (self internalStackValue: 1) by: (self internalStackValue: 0).
	self successful ifTrue:
		[self internalPop: 2 thenPush: (objectMemory integerObjectOf: mod).
		^ self fetchNextBytecode "success"].

	messageSelector := self specialSelector: 10.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 4/26/2016 08:44'!
bytecodePrimMultiply
	| rcvr arg result overflow oop |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg)
		ifTrue: [rcvr := objectMemory integerValueOf: rcvr.
				arg := objectMemory integerValueOf: arg.
				overflow := rcvr > 0
							ifTrue:  [arg > 0
								ifTrue: [rcvr > (objectMemory maxSmallInteger / arg)]
								ifFalse: [arg < (objectMemory minSmallInteger / rcvr)]]
							ifFalse: [arg > 0
								ifTrue: [rcvr < (objectMemory minSmallInteger / arg)]
								ifFalse: [(rcvr < 0) and: [arg < (objectMemory maxSmallInteger / rcvr)]]].
				overflow
					ifFalse:
						[result := rcvr * arg.
						oop := objectMemory integerObjectOf: result.
      					self internalPop: 2 thenPush: oop.
						^self fetchNextBytecode "success"]]
		ifFalse: [self initPrimCall.
				self externalizeIPandSP.
				self primitiveFloatMultiply: rcvr byArg: arg.
				self internalizeIPandSP.
				self successful ifTrue: [^ self fetchNextBytecode "success"]].

	messageSelector := self specialSelector: 8.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: ''!
bytecodePrimNew

	messageSelector := self specialSelector: 28.
	argumentCount := 0.
	self normalSend.
! !

!StackInterpreter methodsFor: 'common selector sends' stamp: ''!
bytecodePrimNewWithArg

	messageSelector := self specialSelector: 29.
	argumentCount := 1.
	self normalSend.
! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 10/5/2012 11:21'!
bytecodePrimNext
	messageSelector := self specialSelector: 19.
	argumentCount := self specialSelectorNumArgs: 19.
	self normalSend.! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 10/5/2012 11:21'!
bytecodePrimNextPut
	messageSelector := self specialSelector: 20.
	argumentCount := self specialSelectorNumArgs: 20.
	self normalSend.! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimNotEqual
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue: [^self booleanCheat: rcvr ~= arg].

	self initPrimCall.
	aBool := self primitiveFloatEqual: rcvr toArg: arg.
	self successful ifTrue: [^self booleanCheat: aBool not].

	messageSelector := self specialSelector: 7.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimNotEqualSistaV1
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue: [^self booleanCheatSistaV1: rcvr ~= arg].

	self initPrimCall.
	aBool := self primitiveFloatEqual: rcvr toArg: arg.
	self successful ifTrue: [^self booleanCheatSistaV1: aBool not].

	messageSelector := self specialSelector: 7.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimNotEqualV4
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg) ifTrue: [^self booleanCheatV4: rcvr ~= arg].

	self initPrimCall.
	aBool := self primitiveFloatEqual: rcvr toArg: arg.
	self successful ifTrue: [^self booleanCheatV4: aBool not].

	messageSelector := self specialSelector: 7.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'cb 11/24/2016 10:35'!
bytecodePrimNotIdentical
	| rcvr arg |
	rcvr := self internalStackValue: 1.
	(objectMemory isOopForwarded: rcvr) ifTrue:
		[rcvr := self handleSpecialSelectorSendFaultFor: rcvr].
	arg := self internalStackValue: 0.
	(objectMemory isOopForwarded: arg) ifTrue:
		[arg := self handleSpecialSelectorSendFaultFor: arg].
	self booleanCheat: rcvr ~= arg! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'cb 11/24/2016 10:36'!
bytecodePrimNotIdenticalSistaV1
	| rcvr arg |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory isOopForwarded: rcvr) ifTrue:
		[rcvr := self handleSpecialSelectorSendFaultFor: rcvr].
	(objectMemory isOopForwarded: arg) ifTrue:
		[arg := self handleSpecialSelectorSendFaultFor: arg].
	self booleanCheatSistaV1: rcvr ~= arg! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'cb 11/24/2016 10:36'!
bytecodePrimNotIdenticalV4
	| rcvr arg |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory isOopForwarded: rcvr) ifTrue:
		[rcvr := self handleSpecialSelectorSendFaultFor: rcvr].
	(objectMemory isOopForwarded: arg) ifTrue:
		[arg := self handleSpecialSelectorSendFaultFor: arg].
	self booleanCheatV4: rcvr ~= arg! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 8/16/2020 15:39'!
bytecodePrimPointX

	| rcvr ok |
	self initPrimCall.
	rcvr := self internalStackTop.
	(objectMemory isNonImmediate: rcvr) ifTrue:
		[ok := objectMemory isClassOfNonImm: rcvr equalTo: (objectMemory splObj: ClassPoint).
		 ok ifTrue:
			[self internalStackTopPut: (objectMemory fetchPointer: XIndex ofObject: rcvr).
			^self fetchNextBytecode]].

	messageSelector := self specialSelector: 30.
	argumentCount := 0.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 8/16/2020 15:39'!
bytecodePrimPointY

	| rcvr ok |
	self initPrimCall.
	rcvr := self internalStackTop.
	(objectMemory isNonImmediate: rcvr) ifTrue:
		[ok := objectMemory isClassOfNonImm: rcvr equalTo: (objectMemory splObj: ClassPoint).
		 ok ifTrue:
			[self internalStackTopPut: (objectMemory fetchPointer: YIndex ofObject: rcvr).
			^self fetchNextBytecode]].

	messageSelector := self specialSelector: 31.
	argumentCount := 0.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 6/26/2012 15:27'!
bytecodePrimSize
	| rcvr sz isString isArray |
	self initPrimCall.
	rcvr := self internalStackTop.

	"Shortcut the mega-lookup for ByteString and Array, the two big consumers of cycles
	 here. Both of these have compact class indices and neither has any added fields."
      isString := self isInstanceOfClassByteString: rcvr.
	isString ifTrue:
		[sz := objectMemory lengthOf: rcvr.
		 self internalStackTopPut: (objectMemory integerObjectOf: sz).
		^self fetchNextBytecode].

      isArray := self isInstanceOfClassArray: rcvr.
	isArray ifTrue:
		[sz := objectMemory lengthOf: rcvr.
		 self internalStackTopPut: (objectMemory integerObjectOf: sz).
		^self fetchNextBytecode].

	messageSelector := self specialSelector: 18.
	argumentCount := 0.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 10/27/2012 15:09'!
bytecodePrimSpecialSelector24

	messageSelector := self specialSelector: 24.
	argumentCount := self specialSelectorNumArgs: 24.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 11/19/2014 13:59'!
bytecodePrimSubtract
	| rcvr arg result |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(objectMemory areIntegers: rcvr and: arg)
		ifTrue: [result := (objectMemory integerValueOf: rcvr) - (objectMemory integerValueOf: arg).
				(objectMemory isIntegerValue: result) ifTrue:
					[self internalPop: 2 thenPush: (objectMemory integerObjectOf: result).
					^self fetchNextBytecode "success"]]
		ifFalse: [self initPrimCall.
				self externalizeIPandSP.
				self primitiveFloatSubtract: rcvr fromArg: arg.
				self internalizeIPandSP.
				self successful ifTrue: [^self fetchNextBytecode "success"]].

	messageSelector := self specialSelector: 1.
	argumentCount := 1.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 12/20/2012 10:00'!
bytecodePrimValue
	| rcvr isBlock |
	rcvr := self internalStackTop.
	argumentCount := 0.
	isBlock := self isInstanceOfClassBlockClosure: rcvr.
	isBlock ifTrue:
		[self externalizeIPandSP.
		self initPrimCall.
		self primitiveClosureValue.
		self internalizeIPandSP.
		self successful ifTrue:
			[^self fetchNextBytecode].
		primFailCode := 0].
	messageSelector := self specialSelector: 25.
	self normalSend! !

!StackInterpreter methodsFor: 'common selector sends' stamp: 'eem 12/20/2012 10:00'!
bytecodePrimValueWithArg
	| rcvr isBlock |
	rcvr := self internalStackValue: 1.
	argumentCount := 1.
	isBlock := self isInstanceOfClassBlockClosure: rcvr.
	isBlock ifTrue:
		[self externalizeIPandSP.
		self initPrimCall.
		self primitiveClosureValue.
		self internalizeIPandSP.
		self successful ifTrue:
			[^self fetchNextBytecode].
		primFailCode := 0].
	messageSelector := self specialSelector: 26.
	self normalSend! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 3/18/2020 18:29'!
callExternalPrimitive: functionID
	"Call the external plugin function identified. In the VM this is an address;
	 see StackInterpreterSimulator for its version."

	<var: #functionID declareC: 'void (*functionID)()'>
	"Spur needs the primitiveFunctionPointer to be set correctly
	 for accurate following of forwarders on primitive failure."
	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[primitiveFunctionPointer := functionID].
	self dispatchFunctionPointer: functionID! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 11/23/2017 11:34'!
callMappedInlinedPrimitive
	"SistaV1:	236		11101100	iiiiiiii		callMappedInlinedPrimitive"
	| primIndex |
	SistaVM
		ifTrue:
			[primIndex := self fetchByte.
			self fetchNextBytecode.
			self sistaMappedInlinePrimitive: primIndex]
		ifFalse: 
			[localIP := localIP - 2.
			self respondToUnknownBytecode].! !

!StackInterpreter methodsFor: 'miscellaneous bytecodes' stamp: 'eem 5/14/2020 13:35'!
callPrimitiveBytecode
	"SistaV1:	**	248	(2)	11111000 	iiiiiiii		mssjjjjj		Call Primitive #iiiiiiii + (jjjjj * 256) 
								m=1 means inlined primitive, no hard return after execution. 
								ss defines the unsafe operation set used to encode the operations. 
								(ss = 0 means sista unsafe operations, ss = 01 means lowcode operations, other numbers are as yet used)
	 V3/Spur:		139		10001011	i i i i i i i i	jjjjjjjj		Call Primitive #iiiiiiii + (jjjjjjjj * 256)
	 V4:			249		11111001	i i i i i i i i	jjjjjjjj		Call Primitive #iiiiiiii + (jjjjjjjj * 256)"

	"Note that we simply skip a callPrimitiveBytecode at the start of a method
	 that contains a primitive.  This because methods like Context(Part)>>reset
	 have to be updated to skip the callPrimtiive bytecode otherwise."
	SistaVM
		ifTrue:
			[| byte1 byte2 prim primSet header |
			 byte1 := self fetchByte.
			 byte2 := self fetchByte.
			 self fetchNextBytecode.
			 byte2 < 128 ifTrue:
				[header := objectMemory methodHeaderOf: method.
				 ((self methodHeaderHasPrimitive: header)
				  and: [localIP asUnsignedInteger
						= (self initialIPForHeader: header method: method) + (self sizeOfCallPrimitiveBytecode: header)]) ifTrue:
					[^self].
				 localIP := localIP - 3.
				 ^self respondToUnknownBytecode].

			 prim := byte2 - 128 << 8 + byte1.
			 primSet := prim >> 13 bitAnd: 3.
			 prim := prim bitAnd: 8191.
			 primSet = 0 ifTrue: [^self sistaInlinePrimitive: prim].
		
			LowcodeVM ifTrue:
				[primSet = 1 ifTrue:
					[prim < 1000 ifTrue:
						[^self lowcodeNullaryInlinePrimitive: prim].

					prim < 2000 ifTrue:
						[^self lowcodeUnaryInlinePrimitive: prim - 1000].
				
					prim < 3000 ifTrue:
						[^self lowcodeBinaryInlinePrimitive: prim - 2000].

					prim < 4000 ifTrue:
						[^self lowcodeTrinaryInlinePrimitive: prim - 3000]]].
		
			localIP := localIP - 3]
		ifFalse:
			[| header |
			 header := objectMemory methodHeaderOf: method.
			 ((self methodHeaderHasPrimitive: header)
			  and: [localIP asInteger = (self initialIPForHeader: header method: method)]) ifTrue:
				[localIP := localIP + (self sizeOfCallPrimitiveBytecode: header) - 1.
				 ^self fetchNextBytecode]].

	^self respondToUnknownBytecode! !

!StackInterpreter methodsFor: 'callback support' stamp: 'eem 9/1/2020 14:27'!
callbackEnter: callbackID
	"Re-enter the interpreter to execute a (non-ALien,non-FFI) callback (as used by the Python bridge)."
	<volatile>
	<export: true>
	<var: #callbackID type: #'sqInt *'>
	| savedReenterInterpreter |
	<var: #savedReenterInterpreter type: #'jmp_buf'>

	"For now, do not allow a callback unless we're in a primitiveResponse"
	(self asserta: primitiveFunctionPointer ~= 0) ifFalse:
		[^false].

	self assert: primFailCode = 0.

	"Check if we've exceeded the callback depth"
	(self asserta: jmpDepth < MaxJumpBuf) ifFalse:
		[^false].
	jmpDepth := jmpDepth + 1.

	"Suspend the currently active process"
	suspendedCallbacks at: jmpDepth put: self activeProcess.
	"We need to preserve newMethod explicitly since it is not activated yet
	and therefore no context has been created for it. If the caller primitive
	for any reason decides to fail we need to make sure we execute the correct
	method and not the one 'last used' in the call back"
	suspendedMethods at: jmpDepth put: newMethod.
	"Signal external semaphores since a signalSemaphoreWithIndex: request may
	 have been issued immediately prior to this callback before the VM has any
	 chance to do a signalExternalSemaphores in checkForEventsMayContextSwitch:"
	self signalExternalSemaphores.
	"If no process is awakened by signalExternalSemaphores then transfer
	 to the highest priority runnable one."
	(suspendedCallbacks at: jmpDepth) = self activeProcess ifTrue:
		[self transferTo: self wakeHighestPriority].

	"Typically, invoking the callback means that some semaphore has been 
	signaled to indicate the callback. Force an interrupt check as soon as possible."
	self forceInterruptCheck.

	"Save the previous interpreter entry jmp_buf."
	self memcpy: savedReenterInterpreter asVoidPointer
		_: reenterInterpreter
		_: (self sizeof: #'jmp_buf').
	(self _setjmp: (jmpBuf at: jmpDepth)) = 0 ifTrue: "Fill in callbackID"
		[callbackID at: 0 put: jmpDepth.
		 self enterSmalltalkExecutive.
		 self assert: false "NOTREACHED"].

	"Restore the previous interpreter entry jmp_buf."
	self memcpy: reenterInterpreter
		_: (self cCoerceSimple: savedReenterInterpreter to: #'void *')
		_: (self sizeof: #'jmp_buf').

	"Transfer back to the previous process so that caller can push result"
	self putToSleep: self activeProcess yieldingIf: preemptionYields.
	self transferTo: (suspendedCallbacks at: jmpDepth).
	newMethod := suspendedMethods at: jmpDepth.	"see comment above"
	argumentCount := self argumentCountOf: newMethod.
	self assert: primFailCode = 0.
	jmpDepth := jmpDepth - 1.
	^true! !

!StackInterpreter methodsFor: 'callback support' stamp: 'eem 9/1/2020 14:25'!
callbackLeave: cbID
	"Leave from a previous callback"
	<export: true>

	"For now, do not allow a callback return unless we're in a primitiveResponse"
	(self asserta: primitiveFunctionPointer ~= 0) ifFalse:
		[^false].

	"Check if this is the top-level callback"
	cbID = jmpDepth ifFalse:[^false].
	cbID < 1 ifTrue:[^false].
	"This is ugly but necessary, or otherwise the Mac will not build"
	self _longjmp: (jmpBuf at: jmpDepth) _: 1.
	"NOTREACHED"
	^nil! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 1/26/2021 17:35'!
canContextSwitchIfActivating: theMethod header: methodHeader
	"Context switch should not be allowed on every method activation.  In particular the
	 implementation of ensure: and ifCurtailed: depends on there being no suspension point
	 on failing primitive 198 (primitiveMarkUnwindMethod, primitiveMarkUnwindMethod).
	 slowPrimitiveResponse used to state
		``N.B.  This means there is no suspension point on primitive failure
		    which methods such as ensure: and ifCurtailed: rely on.''
	 Rather than prevent context switch on all primitives but the ones we really need
	 to be suspension points (primitiveSignal et al) we choose to allow context switch
	 for all but primitiveMarkUnwindMethod and PrimNumberNoContextSwitch."
	| primitiveIndex |
	<api>
	<inline: true>
	primitiveIndex := self primitiveIndexOfMethod: theMethod header: methodHeader.
	^self cppIf: true
		ifTrue:
			[primitiveIndex ~= PrimNumberUnwindMarker "198 primitiveMarkUnwindMethod"
			 and: [primitiveIndex ~= PrimNumberNoContextSwitchMarker "123"]] 
		ifFalse:
			[primitiveIndex = 0
			  or: [(primitiveIndex between: 85 and: 88) "primitiveSignal primitiveWait primitiveResume primitiveSuspend"
			  or: [primitiveIndex = 167 "primitiveYield"
			  or: [primitiveIndex between: 185 and: 186 "primitiveExitCriticalSection primitiveEnterCriticalSection"]]]]! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 4/25/2020 18:39'!
cannotAssign: resultObj to: targetObj withIndex: index
	<option: #IMMUTABILITY>
	<inline: true> "because of use of normalSend..."
	self internalPush: targetObj.
	self internalPush: resultObj.
	self internalPush: (objectMemory integerObjectOf: index + 1).
	messageSelector := objectMemory splObj: SelectorAttemptToAssign.
	argumentCount := 2.
	^ self normalSend! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 9/8/2009 11:16'!
capturePendingFinalizationSignals
	statPendingFinalizationSignals := pendingFinalizationSignals.
! !

!StackInterpreter methodsFor: 'indexing primitive support' stamp: 'eem 9/5/2013 16:30'!
characterForAscii: ascii  "Above ObjectMemory, arg must lie in range 0-255!!"
	<inline: true>
	^CharacterTable
		ifNil: [objectMemory characterObjectOf: ascii]
		ifNotNil: [objectMemory fetchPointer: ascii ofObject: objectMemory characterTable]! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 1/5/2018 18:27'!
charsOfLong: long
	"N.B. This will need to be overridden in big-endian simulator subclasses."
	<doNotGenerate>
	^ (1 to: objectMemory wordSize) collect:
		[:i | | d |
		(d := long digitAt: i) >= 14
			ifTrue: [d asCharacter]
			ifFalse: [$?]]! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 4/13/2021 17:39'!
checkAllAccessibleObjectsOkay
	"Ensure that all accessible objects in the heap are okay."
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| ok |
	ok := true.
	objectMemory allObjectsDoSafely:
		[:oop| ok := ok & (self checkOkayFields: oop)].
	^ok! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 11/30/2010 15:48'!
checkAssumedCompactClasses
	"Check that the classes the VM assumes are compact have the right indices."
	objectMemory checkCompactIndex: ClassArrayCompactIndex isClass: ClassArray named: 'Array'.
	objectMemory checkCompactIndex: ClassLargeNegativeIntegerCompactIndex isClass: ClassLargeNegativeInteger named: 'LargeNegativeInteger'.
	objectMemory checkCompactIndex: ClassLargePositiveIntegerCompactIndex isClass: ClassLargePositiveInteger named: 'LargePositiveInteger'.
	objectMemory checkCompactIndex: ClassFloatCompactIndex isClass: ClassFloat named: 'Float'.
	objectMemory checkCompactIndex: ClassBlockClosureCompactIndex isClass: ClassBlockClosure named: 'BlockClosure'.
	objectMemory checkCompactIndex: ClassMethodContextCompactIndex isClass: ClassMethodContext named: 'MethodContext'.
	classByteArrayCompactIndex := objectMemory compactIndexOfClass: (objectMemory splObj: ClassByteArray)! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 6/8/2009 10:38'!
checkCodeIntegrity: fullGCFlag
	"This is a no-op in the StackVM"
	^true! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 2/28/2009 12:45'!
checkCogCompiledCodeCompactionCalledFor
	"This is a no-op in the StackVM"! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 8/3/2021 13:20'!
checkForAndFollowForwardedPrimitiveState
	"In Spur a primitive may fail due to encountering a forwarder. On failure, check the accessorDepth for the
	 primitive and if non-negative scan the args to the depth, following any forwarders.  Answer if any are found
	 so the prim can be retried.  The primitive index is derived from newMethod.

	 See http://www.mirandabanda.org/cogblog/2014/02/08/primitives-and-the-partial-read-barrier/
	 and SpurMemoryManager's class comment."

	<option: #SpurObjectMemory>
	| primIndex accessorDepth found |
	self assert: self failed.
	found := false.
	primIndex := self primitiveIndexOf: newMethod.
	self assert: (argumentCount = (self argumentCountOf: newMethod) or: [self isMetaPrimitiveIndex: primIndex]).
	"First things first; make sure the metadata has been followed before it is accessed to derive accessorDepth..."
	(((self isCalloutPrimitiveIndex: primIndex) or: [primIndex = PrimNumberDoExternalCall])
	 and: [self unfollowFirstLiteralOfMaybeCalloutMethod: newMethod primitiveIndex: primIndex]) ifTrue:
		[found := true].
	"If the primitive is one of the meta primitives PrimNumberDoPrimitive or PrimNumberDoExternalCall, then
	 metaAccessorDepth will have been set to nil at the start of the primitive, and to the accessor depth of the
	 called primitive (or external call) immediately before dispatch.  Hence if primIndex is that of a meta primitive
	 then if metaAccessorDepth is -2, the accessor depth is that of the meta primitive, and if > -2, then
	 metaAccessorDepth is the accessor depth of the primitive (or external call).  Similarly, if the primitive is
	 primitiveExternalCall then the accessor depth is that of primitiveExternalCall until primitiveFunctionPointer
	 is assigned, at which point the accessor depth is taken from the slot in newMethod's first literal."
	accessorDepth := ((self isMetaPrimitiveIndex: primIndex)
						 and: [metaAccessorDepth > -2])
							ifTrue: [metaAccessorDepth]
							ifFalse:
								[(primIndex = PrimNumberExternalCall
								  and: [primitiveFunctionPointer ~~ #primitiveExternalCall])
									ifTrue: [self primitiveAccessorDepthForExternalPrimitiveMethod: newMethod]
									ifFalse: [primitiveAccessorDepthTable at: primIndex]].
	self assert: (self saneFunctionPointerForFailureOfPrimIndex: primIndex).
	self assert: (accessorDepth between: -1 and: 5).
	accessorDepth >= 0 ifTrue:
		[| scannedStackFrame |
		 scannedStackFrame := false.
		 0 to: argumentCount do:
			[:index| | oop |
			oop := self stackValue: index.
			(objectMemory isNonImmediate: oop) ifTrue:
				[(objectMemory isForwarded: oop) ifTrue:
					[self assert: index < argumentCount. "receiver should have been caught at send time."
					 found := true.
					 oop := objectMemory followForwarded: oop.
					 self stackValue: index put: oop.
					 scannedStackFrame ifFalse:
						[scannedStackFrame := true.
						 self	"Avoid repeated primitive failures by following all state in the current stack frame."
							followForwardedFrameContents: framePointer
							stackPointer: stackPointer + (argumentCount + 1 * objectMemory wordSize)]].
				(accessorDepth > 0
			 	 and: [(objectMemory hasPointerFields: oop)
				 and: [objectMemory followForwardedObjectFields: oop toDepth: accessorDepth]]) ifTrue:
					[found := true]]]].
	^found! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 8/6/2021 12:45'!
checkForEventsMayContextSwitch: mayContextSwitch
	"Check for possible interrupts and handle one if necessary.
	 Answer if a context switch has occurred."
	| switched sema now |
	<inline: false>
	<var: #now type: #usqLong>
	statCheckForEvents := statCheckForEvents + 1.

	"restore the stackLimit if it has been smashed."
	self restoreStackLimit.
	self externalWriteBackHeadFramePointers.
	self assert: stackPage = stackPages mostRecentlyUsedPage.

	"Allow the platform to do anything it needs to do synchronously."
	self ioSynchronousCheckForEvents.

	self checkCogCompiledCodeCompactionCalledFor.

	objectMemory needGCFlag ifTrue:
		["sufficientSpaceAfterGC: runs the incremental GC and
		 then, if not enough space is available, the fullGC."
		 (objectMemory sufficientSpaceAfterGC: 0) ifFalse:
			[self setSignalLowSpaceFlagAndSaveProcess]].

	mayContextSwitch ifFalse: [^false].

	switched := false.

	(profileProcess ~= objectMemory nilObject
	 or: [nextProfileTick > 0 and:[self ioHighResClock >= nextProfileTick]]) ifTrue:
		[self zeroNextProfileTick.
		 "Take a sample (if not already done so) for the profiler if it is active.  This
		  must be done before any of the synchronousSignals below or else we will
		  attribute a pause in ioRelinquishProcessor to the newly activated process."
		 profileProcess = objectMemory nilObject ifTrue:
			[profileProcess := self activeProcess.
			 profileMethod := objectMemory nilObject].
		 "and signal the profiler semaphore if it is present"
		 (profileSemaphore ~= objectMemory nilObject
		  and: [self synchronousSignal: profileSemaphore]) ifTrue:
			[switched := true]].

	objectMemory signalLowSpace ifTrue:
		[objectMemory signalLowSpace: false. "reset flag"
		 sema := objectMemory splObj: TheLowSpaceSemaphore.
		 (sema ~= objectMemory nilObject
		  and: [self synchronousSignal: sema]) ifTrue:
			[switched := true]].

	"inIOProcessEvents prevents reentrancy into ioProcessEvents and allows disabling
	 ioProcessEvents e.g. for native GUIs.  We would like to manage that here but can't
	 since the platform code may choose to call ioProcessEvents itself in various places."
	false
		ifTrue:
			[((now := self ioUTCMicroseconds) >= nextPollUsecs
			 and: [inIOProcessEvents = 0]) ifTrue:
				[statIOProcessEvents := statIOProcessEvents + 1.
				 inIOProcessEvents := inIOProcessEvents + 1.
				 self ioProcessEvents. "sets interruptPending if interrupt key pressed; may callback"
				 inIOProcessEvents > 0 ifTrue:
					[inIOProcessEvents := inIOProcessEvents - 1].
				 nextPollUsecs := now + 20000
				 "msecs to wait before next call to ioProcessEvents.  Note that strictly
				  speaking we might need to update 'now' at this point since
				  ioProcessEvents could take a very long time on some platforms"]]
		ifFalse:
			[(now := self ioUTCMicroseconds) >= nextPollUsecs ifTrue:
				[statIOProcessEvents := statIOProcessEvents + 1.
				 self ioProcessEvents. "sets interruptPending if interrupt key pressed; may callback"
				 nextPollUsecs := now + 20000
				 "msecs to wait before next call to ioProcessEvents.  Note that strictly
				  speaking we might need to update 'now' at this point since
				  ioProcessEvents could take a very long time on some platforms"]].

	interruptPending ifTrue:
		[interruptPending := false.
		 "reset interrupt flag"
		 sema := objectMemory splObj: TheInterruptSemaphore.
		 (sema ~= objectMemory nilObject
		  and: [self synchronousSignal: sema]) ifTrue:
			[switched := true]].

	nextWakeupUsecs ~= 0 ifTrue:
		[now >= nextWakeupUsecs ifTrue:
			[nextWakeupUsecs := 0.
			 "set timer interrupt to 0 for 'no timer'"
			 sema := objectMemory splObj: TheTimerSemaphore.
			 (sema ~= objectMemory nilObject
			  and: [self synchronousSignal: sema]) ifTrue:
				[switched := true]]].

	"signal any pending finalizations"
	pendingFinalizationSignals > 0 ifTrue:
		[pendingFinalizationSignals := 0.
		 sema := objectMemory splObj: TheFinalizationSemaphore.
		 (sema ~= objectMemory nilObject
		  and: [self synchronousSignal: sema]) ifTrue:
			[switched := true]].

	"signal all semaphores in semaphoresToSignal"
	self signalExternalSemaphores ifTrue:
		[switched := true].

	^switched! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 2/17/2017 16:28'!
checkForStackOverflow
	"Check for stack overflow, moving frames to another stack if so.  Since event checking
	 piggy-backs off the stackLimit by setting it to all ones, the check for overflow must be
	 against the real stack limit to find out if overflow has actually occurred."
	<inline: true>
	self externalWriteBackHeadFramePointers.
	stackPointer < stackPage realStackLimit ifTrue:
		[self handleStackOverflow]! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 11/18/2014 13:26'!
checkImageVersionFrom: f startingAt: imageOffset
	"Read and verify the image file version number and return true if the the given image file needs to be byte-swapped. As a side effect, position the file stream just after the version number of the image header. This code prints a warning and does a hard-exit if it cannot find a valid version number."
	"This code is based on C code by Ian Piumarta."

	<inline: false>
	| version firstVersion |
	<var: #f type: #sqImageFile>
	<var: #imageOffset type: #squeakFileOffsetType>

	"check the version number"
	self sqImageFile: f Seek: imageOffset.
	version := firstVersion := self getWord32FromFile: f swap: false.
	(self readableFormat: version) ifTrue: [^ false].

	"try with bytes reversed"
	self sqImageFile: f Seek: imageOffset.
	version := self getWord32FromFile: f swap: true.
	(self readableFormat: version) ifTrue: [^ true].

	"Note: The following is only meaningful if not reading an embedded image"
	imageOffset = 0 ifTrue:[
		"try skipping the first 512 bytes (prepended by certain Mac file transfer utilities)"
		self sqImageFile: f Seek: 512.
		version := self getWord32FromFile: f swap: false.
		(self readableFormat: version) ifTrue: [^ false].

		"try skipping the first 512 bytes with bytes reversed"
		self sqImageFile: f Seek: 512.
		version := self getWord32FromFile: f swap: true.
		(self readableFormat: version) ifTrue: [^ true]].

	"hard failure; abort"
	self print: 'This interpreter (vers. '.
	self printNum: self imageFormatVersion.
	self print: ') cannot read image file (vers. '.
	self printNum: firstVersion.
	self print: ').'.
	self cr.
	self print: 'Press CR to quit...'.
	self getchar.
	self ioExitWithErrorCode: 1.
	^false! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 12/7/2017 09:33'!
checkInterpreterIntegrity
	"Perform an integrity/leak check using the heapMap.  Assume
	 clearLeakMapAndMapAccessibleObjects has set a bit at each
	 object's header.  Check that all oops in the interpreter's state
	 points to a header.  Answer 0 if all checks pass."
	| flags |
	flags := 0.
	(objectMemory checkOopIntegrity: objectMemory specialObjectsOop named: 'specialObjectsOop') ifFalse:
		[flags := 1].
	"No longer check messageSelector; it is ephemeral, not living beyond message lookup.
	(objectMemory isNonImmediate: messageSelector) ifTrue:
		[(objectMemory checkOopIntegrity: messageSelector named: 'messageSelector') ifFalse:
			[flags := flags + N]]."
	(objectMemory checkOopIntegrity: newMethod named: 'newMethod') ifFalse:
		[flags := flags + 2].
	"No longer check lkupClass; it is ephemeral, not living beyond message lookup.
	(objectMemory checkOopIntegrity: lkupClass named: 'lkupClass') ifFalse:
		[flags := flags + N]."
	(objectMemory checkOopIntegrity: profileProcess named: 'profileProcess') ifFalse:
		[flags := flags + 4].
	(objectMemory checkOopIntegrity: profileMethod named: 'profileMethod') ifFalse:
		[flags := flags + 8].
	(objectMemory checkOopIntegrity: profileSemaphore named: 'profileSemaphore') ifFalse:
		[flags := flags + 16].
	tempOop = 0 ifFalse:
		[(objectMemory checkOopIntegrity: tempOop named: 'tempOop') ifFalse:
			[flags := flags + 32]].
	tempOop2 = 0 ifFalse:
		[(objectMemory checkOopIntegrity: tempOop2 named: 'tempOop2') ifFalse:
			[flags := flags + 64]].

	"Callback support - check suspended callback list"
	1 to: jmpDepth do:
		[:i|
		(objectMemory checkOopIntegrity: (suspendedCallbacks at: i) named: 'suspendedCallbacks' index: i) ifFalse:
			[flags := flags + 128].
		(objectMemory checkOopIntegrity: (suspendedMethods at: i) named: 'suspendedMethods' index: i) ifFalse:
			[flags := flags + 256]].

	self checkLogIntegrity ifFalse:
		[flags := flags + 512].

	^flags! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 10/1/2016 14:14'!
checkIsStillMarriedContext: aContext currentFP: currentFP
	"Another version of isWidowedContext: for debugging.
	 This will not bereave a widowed context."
	| thePage maybeFP limitFP maybeFrameCtxt |
	<inline: false>
	<var: #currentFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<var: #maybeFP type: #'char *'>
	<var: #limitFP type: #'char *'>
	((objectMemory isContext: aContext)
	 and: [self isMarriedOrWidowedContext: aContext]) ifFalse:
		[^false].
	maybeFP := self frameOfMarriedContext: aContext.
	thePage := stackPages stackPageFor: maybeFP.
	limitFP := (thePage = stackPage and: [currentFP notNil])
				ifTrue: [currentFP]
				ifFalse: [thePage headFP].
	(maybeFP >= limitFP
	 and: [(objectMemory isNonImmediate: (self frameCallerFP: maybeFP) asInteger)
	 and: [(self withSmallIntegerTags: (self frameCallerFP: maybeFP))
			= (objectMemory fetchPointer: InstructionPointerIndex ofObject: aContext)
	 and: [self frameHasContext: maybeFP]]]) ifFalse:
		[^false].
	maybeFrameCtxt := self frameContext: maybeFP.
	"On Spur we need to follow the context to check for a match, but since the VM is
	 only speculating about maybeFrame being a frame, and only speculating about
	 maybeContext being a context, we need to be sure before we can safely follow."
	(objectMemory hasSpurMemoryManagerAPI
	 and: [(stackPages isFree: thePage) not
	 and: [(self isFrame: maybeFP onPage: thePage)
	 and: [objectMemory isForwarded: maybeFrameCtxt]]]) ifTrue:
		[maybeFrameCtxt := objectMemory followForwarded: maybeFrameCtxt].
	^maybeFrameCtxt = aContext! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 5/28/2010 19:10'!
checkLogIntegrity
	"This is a no-op in the StackInterpreter"
	^true! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 08/08/2014 10:01'!
checkOkayFields: oop
	"Check if the argument is an ok object.
	 If this is a pointers object, check that its fields are all okay oops."

	| hasYoung i fieldOop |
	(oop = nil or: [oop = 0]) ifTrue: [ ^true ]. "?? eem 1/16/2013"
	(objectMemory isIntegerObject: oop) ifTrue: [ ^true ].
	(objectMemory checkOkayOop: oop) ifFalse: [ ^false ].
	(objectMemory checkOopHasOkayClass: oop) ifFalse: [ ^false ].
	((objectMemory isPointersNonImm: oop) or: [objectMemory isCompiledMethod: oop]) ifFalse: [ ^true ].
	hasYoung := objectMemory hasSpurMemoryManagerAPI not
				  and: [objectMemory isYoung: (objectMemory fetchClassOfNonImm: oop)].
	(objectMemory isCompiledMethod: oop)
		ifTrue:
			[i := (objectMemory literalCountOf: oop) + LiteralStart - 1]
		ifFalse:
			[(objectMemory isContext: oop)
				ifTrue: [i := CtxtTempFrameStart + (self fetchStackPointerOf: oop) - 1]
				ifFalse: [i := (objectMemory lengthOf: oop) - 1]].
	[i >= 0] whileTrue:
		[fieldOop := objectMemory fetchPointer: i ofObject: oop.
		(objectMemory isIntegerObject: fieldOop) ifFalse:
			[hasYoung := hasYoung or: [objectMemory isYoung: fieldOop].
			(objectMemory checkOkayOop: fieldOop) ifFalse: [ ^false ].
			(self checkOopHasOkayClass: fieldOop) ifFalse: [ ^false ]].
		i := i - 1].
	hasYoung ifTrue:
		[^objectMemory checkOkayYoungReferrer: oop].
	^true! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 4/13/2021 17:40'!
checkOkayInterpreterObjects: writeBack
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| ok oopOrZero oop |
	ok := true.
	ok := ok & (self checkOkayFields: objectMemory nilObject).
	ok := ok & (self checkOkayFields: objectMemory falseObject).
	ok := ok & (self checkOkayFields: objectMemory trueObject).
	ok := ok & (self checkOkayFields: objectMemory specialObjectsOop).
	ok := ok & (self checkOkayFields: messageSelector).
	ok := ok & (self checkOkayFields: newMethod).
	ok := ok & (self checkOkayFields: lkupClass).
	0 to: MethodCacheEntries - 1 by: MethodCacheEntrySize do:
		[ :i |
		oopOrZero := methodCache at: i + MethodCacheSelector.
		oopOrZero = 0 ifFalse:
			[ok := ok & (self checkOkayFields: (methodCache at: i + MethodCacheSelector)).
			objectMemory hasSpurMemoryManagerAPI ifFalse:
				[ok := ok & (self checkOkayFields: (methodCache at: i + MethodCacheClass))].
			ok := ok & (self checkOkayFields: (methodCache at: i + MethodCacheMethod))]].
	1 to: objectMemory remapBufferCount do:
		[ :i |
		oop := objectMemory remapBuffer at: i.
		(objectMemory isImmediate: oop) ifFalse:
			[ok := ok & (self checkOkayFields: oop)]].
	ok := ok & (self checkOkayStackZone: writeBack).
	^ok! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 4/2/2015 10:57'!
checkOkayStackPage: thePage
	| theSP theFP ok frameRcvrOffset callerFP oop |
	<var: #thePage type: #'StackPage *'>
	<var: #theSP type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #frameRcvrOffset type: #'char *'>
	<var: #callerFP type: #'char *'>
	<inline: false>
	theSP := thePage headSP.
	theFP := thePage  headFP.
	ok := true.
	"Skip the instruction pointer on top of stack of inactive pages."
	thePage = stackPage ifFalse:
		[theSP := theSP + objectMemory wordSize].
	[frameRcvrOffset := self frameReceiverLocation: theFP.
	 [theSP <= frameRcvrOffset] whileTrue:
		[oop := stackPages longAt: theSP.
		 (objectMemory isIntegerObject: oop) ifFalse:
			[ok := ok & (self checkOkayFields: oop)].
		 theSP := theSP + objectMemory wordSize].
	(self frameHasContext: theFP) ifTrue:
		[self assert: (objectMemory isContext: (self frameContext: theFP)).
		 ok := ok & (self checkOkayFields: (self frameContext: theFP))].
	ok := ok & (self checkOkayFields: (self frameMethodObject: theFP)).
	(callerFP := self frameCallerFP: theFP) ~= 0] whileTrue:
		[theSP := theFP + FoxCallerSavedIP + objectMemory wordSize.
		 theFP := callerFP].
	theSP := self isCog
				ifTrue: [theFP + FoxCallerSavedIP + objectMemory wordSize] "caller ip is ceBaseReturnPC"
				ifFalse: [theFP + FoxCallerSavedIP]. "caller ip is frameCallerContext in a base frame"
	[theSP <= thePage baseAddress] whileTrue:
		[oop := stackPages longAt: theSP.
		 (objectMemory isIntegerObject: oop) ifFalse:
			[ok := ok & (self checkOkayFields: oop)].
		 theSP := theSP + objectMemory wordSize].
	^ok! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 1/17/2013 12:36'!
checkOkayStackZone: writeBack
	"Check that all objects in the stack zone are okay"
	| ok thePage |
	<var: #thePage type: #'StackPage *'>
	<inline: false>
	writeBack ifTrue:
		[self externalWriteBackHeadFramePointers].
	ok := true.

	0 to: numStackPages - 1 do:
		[:i|
		thePage := stackPages stackPageAt: i.
		(stackPages isFree: thePage) ifFalse:
			[ok := ok & (self checkOkayStackPage: thePage)]].

	^ok! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 8/6/2021 12:46'!
checkProfileTick: aPrimitiveMethod
	"Check if the profile timer has expired and if so take a sample.
	 If the primitive has failed sample the profileMethod as nil."
	<inline: false>
	self assert: nextProfileTick ~= 0.
	self ioHighResClock >= nextProfileTick ifTrue:
		[profileProcess := self activeProcess.
		 profileMethod := self successful ifTrue: [aPrimitiveMethod] ifFalse: [objectMemory nilObject].
		 self forceInterruptCheck.
		 self zeroNextProfileTick]! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
checkStackIntegrity
	"Perform an integrity/leak check using the heapMap.  Assume
	 clearLeakMapAndMapAccesibleObjects has set a bit at each
	 object's header.  Scan all objects accessible from the stack
	 checking that every pointer points to a header.  Answer if no
	 dangling pointers were detected."
	| ok |
	<inline: false>
	<var: #thePage type: #'StackPage *'>
	<var: #theSP type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #callerFP type: #'char *'>
	ok := true.
	0 to: numStackPages - 1 do:
		[:i| | thePage theSP theFP callerFP oop |
		thePage := stackPages stackPageAt: i.
		(stackPages isFree: thePage) ifFalse:
			[thePage = stackPage
				ifTrue:
					[theSP := stackPointer.
					 theFP := framePointer]
				ifFalse:
					[theSP := thePage headSP.
					 theFP := thePage  headFP].
			 "Skip the instruction pointer on top of stack of inactive pages."
			 thePage = stackPage ifFalse:
				[theSP := theSP + objectMemory wordSize].
			 [[theSP <= (theFP + FoxReceiver)] whileTrue:
				[oop := stackPages longAt: theSP.
				 ((objectMemory isNonImmediate: oop) 
				   and: [(objectMemory heapMapAtWord: (self pointerForOop: oop)) = 0]) ifTrue:
					[self printFrameThing: 'object leak in frame temp' andFrame: theFP at: theSP.
					 ok := false].
				 theSP := theSP + objectMemory wordSize].
			 (self frameHasContext: theFP) ifTrue:
				[oop := self frameContext: theFP.
				 ((objectMemory isImmediate: oop) 
				   or: [(objectMemory heapMapAtWord: (self pointerForOop: oop)) = 0]) ifTrue:
					[self printFrameThing: 'object leak in frame ctxt' andFrame: theFP at: theFP + FoxThisContext.
					 ok := false].
				 (objectMemory isContext: oop) ifFalse:
					[self printFrameThing: 'frame ctxt should be context' andFrame: theFP at: theFP + FoxThisContext.
					 ok := false].
				 ((objectMemory isContext: oop) and: [self isMarriedOrWidowedContext: oop]) ifFalse:
					[self printFrameThing: 'frame ctxt should be married' andFrame: theFP at: theFP + FoxThisContext.
					 ok := false].
				 ((objectMemory isContext: oop)
				  and: [(self isMarriedOrWidowedContext: oop)
				  and: [(self frameOfMarriedContext: oop) = theFP]]) ifFalse:
					[self printFrameThing: 'frame ctxt should be married to this frame ' andFrame: theFP at: theFP + FoxThisContext.
					 ok := false]].
			 oop := self frameMethod: theFP.
			 ((objectMemory isImmediate: oop) 
			   or: [(objectMemory heapMapAtWord: (self pointerForOop: oop)) = 0]) ifTrue:
				[self printFrameThing: 'object leak in frame mthd' andFrame: theFP at: theFP + FoxMethod.
				 ok := false].
			 (callerFP := self frameCallerFP: theFP) ~= 0] whileTrue:
				[theSP := theFP + FoxCallerSavedIP + objectMemory wordSize.
				 theFP := callerFP].
			 theSP := theFP + FoxCallerSavedIP + objectMemory wordSize.
			 [theSP <= thePage baseAddress] whileTrue:
				[oop := stackPages longAt: theSP.
				 ((objectMemory isNonImmediate: oop) 
				   and: [(objectMemory heapMapAtWord: (self pointerForOop: oop)) = 0]) ifTrue:
					[self printFrameThing: 'object leak in frame arg' andFrame: theFP at: theSP.
					 ok := false].
				 theSP := theSP + objectMemory wordSize]]].
	^ok! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 7/5/2016 15:34'!
checkStackPointerForMaybeMarriedContext: aContext
	"Version of stackPointerForMaybeMarriedContext: with no side-effects (does not widow).
	 Used for assertion checking. Safe only in external primitives (framePointer valid).
	 Answer the stackPointer of a Context."
	| sp |
	<inline: true>
	(self checkIsStillMarriedContext: aContext currentFP: framePointer) ifTrue:
		[sp := self checkStackPointerIndexForFrame: (self frameOfMarriedContext: aContext).
		 self assert: ReceiverIndex + (objectMemory integerValueOf: sp) < (objectMemory lengthOf: aContext).
		 ^sp].
	(self isMarriedOrWidowedContext: aContext) ifTrue:
		[^self argumentCountOf: (objectMemory fetchPointer: MethodIndex ofObject: aContext)].
	^self fetchStackPointerOf: aContext! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 6/12/2017 12:53'!
checkStackPointerIndexForFrame: theFP
	"Version of stackPointerIndexForFrame: that does not depend on writing back head frame pointers.
	 Used for assertion checking. Safe only in external primitives (framePointer valid).
	 Answer the 0-based index rel to the given frame.
	 (This is what stackPointer used to be before conversion to pointer)"
	"In the StackInterpreter stacks grow down."
	| thePage theSP |
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<var: #theSP type: #'char *'>
	theFP = framePointer ifTrue:
		[^self stackPointerIndexForFrame: theFP WithSP: stackPointer].
	thePage := stackPages stackPageFor: theFP.
	theSP := self findSPOrNilOf: theFP
				on: thePage
				startingFrom: (thePage = stackPage ifTrue: [framePointer] ifFalse: [thePage headFP]).
	theSP ifNil: [^-1].
	^self stackPointerIndexForFrame: theFP WithSP: theSP! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
checkedIntegerValueOf: intOop
	"Note: May be called by translated primitive code."

	(objectMemory isIntegerObject: intOop)
		ifTrue: [ ^ objectMemory integerValueOf: intOop ]
		ifFalse: [ self primitiveFail. ^ 0 ]! !

!StackInterpreter methodsFor: 'simulation' stamp: 'eem 9/6/2013 21:18'!
classNameIndex
	<doNotGenerate>
	^classNameIndex! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 8/2/2021 21:56'!
classNameOf: aClass Is: className 
	"Check if aClass's name is className"
	| srcName name length |
	<var: #className type: #'char *'>
	<var: #srcName type: #'char *'>
	(objectMemory numSlotsOf: aClass) <= classNameIndex ifTrue:
		[^ false]. "Not a class but might be behavior"
	name := objectMemory fetchPointer: classNameIndex ofObject: aClass.
	(objectMemory isBytes: name) ifFalse:
		[^false].
	length := self numBytesOfBytes: name.
	srcName := self cCoerce: (self arrayValueOf: name) to: 'char *'.
	0 to: length - 1 do: [:i | (srcName at: i) = (className at: i) ifFalse: [^false]].
	"Check if className really ends at this point"
	^(className at: length) = 0! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 11/24/2019 19:22'!
cloneOSErrorObj: errObj numSlots: numSlots
	"If errObj is a pointer object with at least two slots, then answer a clone
	  of the error object with the second slot set to the value of osErrorCode,
	  and if an PrimErrFFIException, then the third slow with the exceptionPC."
	| clone |
	<inline: true>
	clone := objectMemory hasSpurMemoryManagerAPI
				ifTrue: [objectMemory
							eeInstantiateAnySmallClassIndex: (objectMemory classIndexOf: errObj)
							format: objectMemory nonIndexablePointerFormat
							numSlots: numSlots]
				ifFalse: [objectMemory
							eeInstantiateSmallClass: (objectMemory fetchClassOfNonImm: errObj)
							numSlots: numSlots].
	0 to: numSlots - 1 do:
		[:i| objectMemory
				storePointerUnchecked: i
				ofObject: clone
				withValue: (objectMemory fetchPointer: i ofObject: errObj)].
	(numSlots > 2
	 and: [primFailCode = PrimErrFFIException])
		ifTrue:
			[objectMemory
				storePointerUnchecked: 1
					ofObject: clone
						withValue: (self positive64BitIntegerFor: (self cCoerceSimple: osErrorCode to: #usqLong));
				storePointerUnchecked: 2
					ofObject: clone
						withValue: (self positiveMachineIntegerFor: exceptionPC)]
		ifFalse:
			[objectMemory
				storePointerUnchecked: 1
				ofObject: clone
				withValue: (self signed64BitIntegerFor: osErrorCode)].
	^clone! !

!StackInterpreter methodsFor: 'control primitives' stamp: 'eem 07/28/2014 07:54'!
closureIn: context numArgs: numArgs instructionPointer: initialIP copiedValues: copiedValues
	| newClosure numCopied |
	<inline: true>
	"numCopied should be zero for nil"
	numCopied := objectMemory numSlotsOf: copiedValues.
	ClassBlockClosureCompactIndex ~= 0
		ifTrue:
			[newClosure := objectMemory
								eeInstantiateSmallClassIndex: ClassBlockClosureCompactIndex
								format: objectMemory indexablePointersFormat
								numSlots: ClosureFirstCopiedValueIndex + numCopied]
		ifFalse:
			[newClosure := objectMemory
								eeInstantiateSmallClass: (objectMemory splObj: ClassBlockClosure)
								numSlots: ClosureFirstCopiedValueIndex + numCopied].
	"Assume: have just allocated a new closure; it must be young. Thus, can use unchecked stores."
	objectMemory
		storePointerUnchecked: ClosureOuterContextIndex ofObject: newClosure withValue: context;
		storePointerUnchecked: ClosureStartPCIndex ofObject: newClosure withValue: (objectMemory integerObjectOf: initialIP);
		storePointerUnchecked: ClosureNumArgsIndex ofObject: newClosure withValue: (objectMemory integerObjectOf: numArgs).
	0 to: numCopied - 1 do:
		[:i|
		objectMemory storePointerUnchecked: i + ClosureFirstCopiedValueIndex
			ofObject: newClosure
			withValue: (objectMemory fetchPointer: i ofObject: copiedValues)].
	^newClosure! !

!StackInterpreter methodsFor: 'control primitives' stamp: 'eem 3/4/2014 17:26'!
closureIn: context numArgs: numArgs instructionPointer: initialIP numCopiedValues: numCopied
	| newClosure |
	<inline: true>
	ClassBlockClosureCompactIndex ~= 0
		ifTrue:
			[newClosure := objectMemory
								eeInstantiateSmallClassIndex: ClassBlockClosureCompactIndex
								format: objectMemory indexablePointersFormat
								numSlots: ClosureFirstCopiedValueIndex + numCopied]
		ifFalse:
			[newClosure := objectMemory
								eeInstantiateSmallClass: (objectMemory splObj: ClassBlockClosure)
								numSlots: ClosureFirstCopiedValueIndex + numCopied].
	"Assume: have just allocated a new closure; it must be young. Thus, can use unchecked stores."
	objectMemory storePointerUnchecked: ClosureOuterContextIndex ofObject: newClosure withValue: context.
	objectMemory storePointerUnchecked: ClosureStartPCIndex ofObject: newClosure withValue: (objectMemory integerObjectOf: initialIP).
	objectMemory storePointerUnchecked: ClosureNumArgsIndex ofObject: newClosure withValue: (objectMemory integerObjectOf: numArgs).
	^newClosure! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 2/9/2018 09:59'!
codeGeneratorToComputeAccessorDepth
	^(VMMaker new
		buildCodeGeneratorForInterpreter: self class primitivesClass
		includeAPIMethods: false
		initializeClasses: false)
			logger: self transcript;
			yourself! !

!StackInterpreter methodsFor: 'indexing primitive support' stamp: 'eem 1/1/1970 15:10'!
commonAt: stringy
	"This code is called if the receiver responds primitively to at:.
	 N.B. this does *not* use the at cache, instead inlining stObject:at:.
	 Using the at cache here would require that callers set messageSelector
	 and lkupClass and that is onerous and error-prone, and in any case,
	 inlining produces much better performance than using the at cache here."
	| index rcvr result |
	<inline: true> "to get it inlined in primitiveAt and primitiveStringAt"
	self initPrimCall.
	rcvr := self stackValue: 1.
	index := self stackTop.
	(objectMemory isImmediate: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrInappropriate].
	"No need to test for large positive integers here.  No object has 1g elements"
	((objectMemory isNonIntegerObject: index)
	 or: [argumentCount > 1 "e.g. object:basicAt:"
		 and: [objectMemory isForwarded: rcvr]]) ifTrue:
		[^self primitiveFailFor: PrimErrBadArgument].
	index := objectMemory integerValueOf: index.
	result := self stObject: rcvr at: index.
	self successful ifTrue:
		[stringy ifTrue:
			[(objectMemory isInRangeCharacterCode: result) ifFalse:
				[^self primitiveFailFor: PrimErrBadReceiver].
			 result := self characterForAscii: (objectMemory integerValueOf: result)].
		 self pop: argumentCount+1 thenPush: result]! !

!StackInterpreter methodsFor: 'indexing primitive support' stamp: 'eem 1/20/2016 14:32'!
commonAtPut: stringy
	"This code is called if the receiver responds primitively to at:Put:.
	 N.B. this does *not* use the at cache, instead inlining stObject:at:put:.
	 Using the at cache here would require that callers set messageSelector
	 and lkupClass and that is onerous and error-prone, and in any case,
	 inlining produces much better performance than using the at cache here."
	| value index rcvr |
	<inline: true> "to get it inlined in primitiveAtPut and primitiveStringAtPut"
	self initPrimCall.
	rcvr := self stackValue: 2.
	index := self stackValue: 1.
	value := self stackTop.
	(objectMemory isImmediate: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrInappropriate].
	"No need to test for large positive integers here.  No object has 1g elements"
	((objectMemory isNonIntegerObject: index)
	 or: [argumentCount > 2 "e.g. object:basicAt:put:"
		 and: [objectMemory isForwarded: rcvr]]) ifTrue:
		[^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory isObjImmutable: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].
	index := objectMemory integerValueOf: index.
	stringy
		ifTrue: [self stObject: rcvr at: index put: (self asciiOfCharacter: value)]
		ifFalse: [self stObject: rcvr at: index put: value].
	self successful ifTrue:
		[self pop: argumentCount+1 thenPush: value]! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem 12/8/2017 12:49'!
commonCallerReturn
	"Return to the previous context/frame (sender for method activations, caller for block activations)."
	<sharedCodeInCase: #returnTopFromBlock>
	| callersFPOrNull doWeHaveANativeFrame |
	<var: #callersFPOrNull type: #'char *'>

	"TODO: Store/restore the nativeSP more properly, when it exists"
	LowcodeVM ifTrue: [ 
		doWeHaveANativeFrame := self frameHasNativeFrame: localFP.
		doWeHaveANativeFrame ifTrue: [
			nativeStackPointer := (self nativePreviousStackPointerIn: localFP) - 1.
			nativeSP := 0.
			self setFrameHasNotNativeFrame: localFP.
		].
	].

	callersFPOrNull := self frameCallerFP: localFP.
	callersFPOrNull = 0 "baseFrame" ifTrue:
		[self assert: localFP = stackPage baseFP.
		 ^self baseFrameReturn].

	localIP := self frameCallerSavedIP: localFP.
	localSP := localFP + (self frameStackedReceiverOffset: localFP).
	localFP := callersFPOrNull.
	self setMethod: (self frameMethod: localFP).
	self fetchNextBytecode.
	self internalStackTopPut: localReturnValue! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'akg 11/29/2018 17:21'!
commonReturn
	"Do an ^-return (return from method), checking for unwinds if this is a block activation.
	 Note: Assumed to be inlined into the dispatch loop."

	<sharedCodeInCase: #returnReceiver>
	| closure home unwindContextOrNilOrZero frameToReturnTo contextToReturnTo theFP callerFP newPage |
	<var: #frameToReturnTo type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #callerFP type: #'char *'>
	<var: #newPage type: #'StackPage *'>
	<var: #thePage type: #'StackPage *'>

	"If this is a method simply return to the  sender/caller."
	(self iframeIsBlockActivation: localFP) ifFalse:
		[^self commonCallerReturn].

	"Update the current page's headFrame pointers to enable the search for unwind protects below
	 to identify widowed contexts correctly."
	self writeBackHeadFramePointers.
	self externalizeIPandSP.

	"Since this is a block activation the closure is on the stack above any args and the frame."
	closure := self pushedReceiverOrClosureOfFrame: localFP.
	home := nil. "avoid compiler warning"
	"Walk the closure's lexical chain to find the context or frame to return from (home).
	 If home is missing (Sista closures) then throw cannotReturn rather than crash."
	[closure ~= objectMemory nilObject] whileTrue:
		[home := objectMemory followField: ClosureOuterContextIndex ofObject: closure.
		 (objectMemory isContext: home) ifFalse:
			[^self internalCannotReturn: localReturnValue].
		 closure := objectMemory followField: ClosureIndex ofObject: home].
	"home is to be returned from provided there is no unwind-protect activation between
	 this frame and home's sender.  Search for an unwind.  findUnwindThroughContext:
	 will answer either the context for an unwind-protect activation or nilObj if the sender
	 cannot be found or 0 if no unwind is found but the sender is."
	unwindContextOrNilOrZero := self findUnwindThroughContext: home.
	unwindContextOrNilOrZero = objectMemory nilObject ifTrue:
		["error: can't find home on chain; cannot return"
		 ^self internalCannotReturn: localReturnValue].
	unwindContextOrNilOrZero ~= 0 ifTrue:
		[^self internalAboutToReturn: localReturnValue through: unwindContextOrNilOrZero].

	"Now we know home is on the sender chain.
	 We could be returning to either a context or a frame.  Find out which."
	contextToReturnTo := nil.
	(self isMarriedOrWidowedContext: home)
		ifTrue:
			[self assert: (self checkIsStillMarriedContext: home currentFP: localFP).
			 theFP := self frameOfMarriedContext: home.
			 (self isBaseFrame: theFP)
				ifTrue:
					[contextToReturnTo := self frameCallerContext: theFP]
				ifFalse:
					[frameToReturnTo := self frameCallerFP: theFP]]
		ifFalse:
			[contextToReturnTo := objectMemory fetchPointer: SenderIndex ofObject: home.
			 ((objectMemory isContext: contextToReturnTo)
			  and: [self isMarriedOrWidowedContext: contextToReturnTo]) ifTrue:
				[self assert: (self checkIsStillMarriedContext: contextToReturnTo currentFP: localFP).
			 	 frameToReturnTo := self frameOfMarriedContext: contextToReturnTo.
				 contextToReturnTo := nil]].

	"If returning to a context we must make a frame for it unless it is dead."
	contextToReturnTo ~= nil ifTrue:
		[frameToReturnTo := self establishFrameForContextToReturnTo: contextToReturnTo.
		 frameToReturnTo = 0 ifTrue: "error: home's sender is dead; cannot return"
			[^self internalCannotReturn: localReturnValue]].

	"Now we have a frame to return to.  If it is on a different page we must free intervening pages and
	 nil out intervening contexts.  We must free intervening stack pages because if we leave the pages
	 to be divorced then their contexts will be divorced with intact senders and instruction pointers.  This
	 code is similar to primitiveTerminateTo.  We must move any frames on intervening pages above the
	 frame linked to because these may be in use, e.g. via co-routining (see baseFrameReturn)."
	self assert: stackPages pageListIsWellFormed.
	newPage := stackPages stackPageFor: frameToReturnTo.
	newPage ~~ stackPage ifTrue:
		[| currentCtx thePage nextCntx |
		 currentCtx := self frameCallerContext: stackPage baseFP.
		 stackPages freeStackPage: stackPage.
		 [self assert: (objectMemory isContext: currentCtx).
		  (self isMarriedOrWidowedContext: currentCtx)
		   and: [(stackPages stackPageFor: (theFP := self frameOfMarriedContext: currentCtx)) = newPage]] whileFalse:
			[(self isMarriedOrWidowedContext: currentCtx)
				ifTrue:
					[thePage := stackPages stackPageFor: theFP.
					 theFP ~= thePage headFP ifTrue:
						["Since we've just deallocated a page we know that newStackPage won't deallocate an existing one."
						 self moveFramesIn: thePage through: (self findFrameAbove: theFP inPage: thePage) toPage: stackPages newStackPage].
					 currentCtx := self frameCallerContext: thePage baseFP.
					 stackPages freeStackPage: thePage]
				ifFalse:
					[nextCntx := objectMemory fetchPointer: SenderIndex ofObject: currentCtx.
					 self markContextAsDead: currentCtx.
					 currentCtx := nextCntx]].
		 self setStackPageAndLimit: newPage.
		 localSP := stackPage headSP.
		 localFP := stackPage headFP].

	"Two cases.  Returning to the top frame on a new page or an interior frame on the current page.
	 The top frame has its instruction pointer on top of stack. An interior frame has its instruction pointer
	 in the caller frame. We need to peel back any frames on the page until we get to the correct frame."

	localFP = frameToReturnTo
		ifTrue: "pop the saved IP, push the return value and continue."
			[localIP := self pointerForOop: self internalStackTop]
		ifFalse:
			[[callerFP := localFP.
			  localFP := self frameCallerFP: localFP.
			  localFP ~~ frameToReturnTo] whileTrue.
			localIP := self frameCallerSavedIP: callerFP.
			localSP := (self frameCallerSP: callerFP) - objectMemory wordSize].
	self maybeReturnToMachineCodeFrame.
	self setMethod: (self frameMethod: localFP).
	self fetchNextBytecode.
	self internalStackTopPut: localReturnValue! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'rmacnak 5/18/2015 21:21'!
commonSendDynamicSuper
	"Assume: messageSelector and argumentCount have been set, and that 
	the receiver and arguments have been pushed onto the stack,"
	"Note: This method is inlined into the interpreter dispatch loop."
	<sharedCodeInCase: #extSendAbsentDynamicSuperBytecode>
	self sendBreakpoint: messageSelector receiver: (self internalStackValue: argumentCount).
	self printSends ifTrue:
		[self printActivationNameForSelector: messageSelector startClass: (objectMemory classForClassTag: lkupClassTag); cr].
	self internalFindNewMethodDynamicSuper.
	self shuffleArgumentsAndStoreAbsentReceiver: localAbsentReceiver.
	self internalExecuteNewMethod.
	self fetchNextBytecode! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'rmacnak 5/18/2015 21:19'!
commonSendImplicitReceiver
	"Assume: messageSelector and argumentCount have been set, and that 
	the receiver and arguments have been pushed onto the stack,"
	"Note: This method is inlined into the interpreter dispatch loop."
	<sharedCodeInCase: #extSendAbsentImplicitBytecode>
	self sendBreakpoint: messageSelector receiver: (self internalStackValue: argumentCount).
	self printSends ifTrue:
		[self printActivationNameForSelector: messageSelector startClass: (objectMemory classForClassTag: lkupClassTag); cr].
	self internalFindNewMethodImplicitReceiver.
	self shuffleArgumentsAndStoreAbsentReceiver: localAbsentReceiver.
	self internalExecuteNewMethod.
	self fetchNextBytecode! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'rmacnak 5/19/2015 20:26'!
commonSendOrdinary
	"Send a message, starting lookup with the receiver's class."
	"Assume: messageSelector and argumentCount have been set, and that 
	the receiver and arguments have been pushed onto the stack,"
	"Note: This method is inlined into the interpreter dispatch loop."
	<sharedCodeInCase: #singleExtendedSendBytecode>
	self sendBreakpoint: messageSelector receiver: (self internalStackValue: argumentCount).
	self printSends ifTrue:
		[self printActivationNameForSelector: messageSelector startClass: (objectMemory classForClassTag: lkupClassTag); cr].
	self internalFindNewMethodOrdinary.
	self internalExecuteNewMethod.
	self fetchNextBytecode! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'rmacnak 5/18/2015 21:50'!
commonSendOuter: depth
	"Assume: messageSelector and argumentCount have been set, and that 
	the receiver and arguments have been pushed onto the stack,"
	"Note: This method is inlined into the interpreter dispatch loop."
	"<sharedCodeInCase: #extSendAbsentOuterBytecode>"
	self sendBreakpoint: messageSelector receiver: (self internalStackValue: argumentCount).
	self printSends ifTrue:
		[self printActivationNameForSelector: messageSelector startClass: (objectMemory classForClassTag: lkupClassTag); cr].
	self internalFindNewMethodOuter: depth.
	self shuffleArgumentsAndStoreAbsentReceiver: localAbsentReceiver.
	self internalExecuteNewMethod.
	self fetchNextBytecode! !

!StackInterpreter methodsFor: 'indexing primitive support' stamp: 'nice 11/14/2016 23:01'!
commonVariable: rcvr at: index cacheIndex: atIx 
	"This code assumes the receiver has been identified at location atIx in the atCache."
	| stSize fmt fixedFields result |
	<inline: true>
	stSize := atCache at: atIx+AtCacheSize.
	((self oop: index isGreaterThanOrEqualTo: 1)
	 and: [self oop: index isLessThanOrEqualTo: stSize]) ifTrue:
		[fmt := atCache at: atIx+AtCacheFmt.
		 fmt <= objectMemory weakArrayFormat ifTrue:
			[self assert: (objectMemory isContextNonImm: rcvr) not.
			 fixedFields := atCache at: atIx+AtCacheFixedFields.
			 ^objectMemory fetchPointer: index + fixedFields - 1 ofObject: rcvr].
		 fmt < objectMemory firstByteFormat ifTrue: "64, 32, & 16 bits"
			[objectMemory hasSpurMemoryManagerAPI ifTrue:
				[fmt >= objectMemory firstShortFormat ifTrue:
					[^objectMemory integerObjectOf:
						(objectMemory fetchUnsignedShort16: index - 1 ofObject: rcvr)].
				 fmt = objectMemory sixtyFourBitIndexableFormat ifTrue:
					[^self positive64BitIntegerFor:
						(objectMemory fetchLong64: index - 1 ofObject: rcvr)]].
			 result := objectMemory fetchLong32: index - 1 ofObject: rcvr.
			 ^self positive32BitIntegerFor: result].
		 fmt >= objectMemory firstStringyFakeFormat  "Note fmt >= firstStringyFormat is an artificial flag for strings"
			ifTrue: "String"
				["Spur supports the String at:[put:] primitives on WideString and DoubleByteString"
				 result := (objectMemory hasSpurMemoryManagerAPI
							and: [fmt < (objectMemory firstByteFormat + objectMemory firstStringyFakeFormat)])
								ifTrue:
									[fmt < (objectMemory firstShortFormat + objectMemory firstStringyFakeFormat)
										ifTrue: [objectMemory fetchLong32: index - 1 ofObject: rcvr]
										ifFalse: [objectMemory fetchUnsignedShort16: index - 1 ofObject: rcvr]]
								ifFalse: [objectMemory fetchByte: index - 1 ofObject: rcvr].
				^self characterForAscii: result]
			ifFalse:
				[(fmt < objectMemory firstCompiledMethodFormat "ByteArray"
				  or: [index >= (self firstByteIndexOfMethod: rcvr) "CompiledMethod"]) ifTrue:
					[^objectMemory integerObjectOf: (objectMemory fetchByte: index - 1 ofObject: rcvr)]]].

	^self primitiveFailFor: ((objectMemory isIndexable: rcvr)
								ifFalse: [PrimErrBadReceiver]
								ifTrue: [PrimErrBadIndex])! !

!StackInterpreter methodsFor: 'indexing primitive support' stamp: 'eem 1/25/2017 08:29'!
commonVariable: rcvr at: index put: value cacheIndex: atIx
	"This code assumes the receiver has been identified at location atIx in the atCache."
	<returnTypeC: #void>
	| stSize fmt fixedFields valToPut isCharacter |
	<inline: true>
	stSize := atCache at: atIx+AtCacheSize.
	((self oop: index isGreaterThanOrEqualTo: 1)
	  and: [self oop: index isLessThanOrEqualTo: stSize]) ifTrue:
		[fmt := atCache at: atIx+AtCacheFmt.
		fmt <= objectMemory weakArrayFormat ifTrue:
			[self assert: (objectMemory isContextNonImm: rcvr) not.
			 fixedFields := atCache at: atIx+AtCacheFixedFields.
			 ^objectMemory storePointer: index + fixedFields - 1 ofObject: rcvr withValue: value].
		fmt < objectMemory firstByteFormat ifTrue:  "64, 32, & 16 bits"
			[objectMemory hasSpurMemoryManagerAPI ifTrue:
				[fmt >= objectMemory firstShortFormat ifTrue:
					[valToPut := (objectMemory isIntegerObject: value)
									ifTrue: [objectMemory integerValueOf: value]
									ifFalse: [-1].
					(valToPut >= 0 and: [valToPut <= 65535]) ifFalse: [^self primitiveFailFor: PrimErrBadArgument].
					^objectMemory storeShort16: index - 1 ofObject: rcvr withValue: valToPut].
				 fmt = objectMemory sixtyFourBitIndexableFormat ifTrue:
					[| val64ToPut |
					 val64ToPut := self positive64BitValueOf: value.
					 self successful ifTrue:
						[^objectMemory storeLong64: index - 1 ofObject: rcvr withValue: val64ToPut].
					 ^self primitiveFailFor: PrimErrBadArgument]].
			 valToPut := self positive32BitValueOf: value.
			 self successful ifTrue:
				[^objectMemory storeLong32: index - 1 ofObject: rcvr withValue: valToPut].
			 ^self primitiveFailFor: PrimErrBadArgument].
		fmt >= objectMemory firstStringyFakeFormat  "Note fmt >= firstStringyFormat is an artificial flag for strings"
			ifTrue:
				[isCharacter := objectMemory isCharacterObject: value.
				 isCharacter ifFalse:
					[^self primitiveFailFor: PrimErrBadArgument].
				 objectMemory hasSpurMemoryManagerAPI
					ifTrue: [valToPut := objectMemory characterValueOf: value]
					ifFalse:
						[valToPut := objectMemory fetchPointer: CharacterValueIndex ofObject: value.
						 valToPut := (objectMemory isIntegerObject: valToPut)
										ifTrue: [objectMemory integerValueOf: valToPut]
										ifFalse: [-1]].
				 objectMemory hasSpurMemoryManagerAPI ifTrue:
				 	[fmt < (objectMemory firstByteFormat + objectMemory firstStringyFakeFormat) ifTrue:
						[fmt < (objectMemory firstShortFormat + objectMemory firstStringyFakeFormat)
							ifTrue:
								[self assert: fmt ~= (objectMemory sixtyFourBitIndexableFormat + objectMemory firstStringyFakeFormat).
								 ^objectMemory storeLong32: index - 1 ofObject: rcvr withValue: valToPut]
							ifFalse:
								[(valToPut >= 0 and: [valToPut <= 65535]) ifFalse: [^self primitiveFailFor: PrimErrBadArgument].
								 ^objectMemory storeShort16: index - 1 ofObject: rcvr withValue: valToPut]]]]
			ifFalse:
				[(fmt >= objectMemory firstCompiledMethodFormat
				  and: [index < (self firstByteIndexOfMethod: rcvr)]) ifTrue:
					[^self primitiveFailFor: PrimErrBadIndex].
				valToPut := (objectMemory isIntegerObject: value)
								ifTrue: [objectMemory integerValueOf: value]
								ifFalse: [-1]].
		(valToPut >= 0 and: [valToPut <= 255]) ifFalse: [^self primitiveFailFor: PrimErrBadArgument].
		^objectMemory storeByte: index - 1 ofObject: rcvr withValue: valToPut].

	^self primitiveFailFor: ((objectMemory isIndexable: rcvr)
								ifFalse: [PrimErrBadReceiver]
								ifTrue: [PrimErrBadIndex])! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 8/14/2015 12:07'!
computeStackZoneSize
	self cCode: [] inSmalltalk:
		[stackPages ifNil:
			[stackPages := self stackPagesClass new setInterpreter: self]].
	^numStackPages * ((self sizeof: CogStackPage) + self stackPageByteSize)
	 + stackPages extraStackBytes! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 12/8/2017 12:50'!
context: thisCntx hasSender: aContext 
	"Does thisCntx have aContext in its sender chain?
	 Cheapo implementation above extant machinery."
	| handlerOrNilOrZero |
	<inline: true>
	handlerOrNilOrZero := self
							findMethodWithPrimitive: -1
							FromContext: thisCntx
							UpToContext: aContext.
	^handlerOrNilOrZero = 0! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 8/8/2014 13:59'!
context: aContext hasValidInversePCMappingOf: theIP in: theFP
	"For asserts.  Check that theIP maps back correctly to the context's pc.
	 The CallPrimitive bytecode presents a complication."
	| pc encodedip |
	<var: #theFP type: #'char *'>
	pc := objectMemory fetchPointer: InstructionPointerIndex ofObject: aContext.
	encodedip := self contextInstructionPointer: theIP frame: theFP.
	^(objectMemory hasSpurMemoryManagerAPI or: [MULTIPLEBYTECODESETS])
		ifTrue:
			[pc = encodedip
			or: [| methodHeader |
				methodHeader := objectMemory methodHeaderOf: (objectMemory fetchPointer: MethodIndex ofObject: aContext).
				(self methodHeaderHasPrimitive: methodHeader)
				and: [(objectMemory integerValueOf: encodedip) - (objectMemory integerValueOf: pc)
					= (self sizeOfCallPrimitiveBytecode: methodHeader)]]]
		ifFalse: [pc = encodedip]! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
contextInstructionPointer: theIP frame: theFP
	<var: #theFP type: #'char *'>
	self assert: (self validInstructionPointer: theIP + 1 inFrame: theFP).
	^objectMemory integerObjectOf: theIP - (self iframeMethod: theFP) - objectMemory baseHeaderSize + 2! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 11/19/2014 22:05'!
convertFloatsToPlatformOrder
	"Byte-swap the words of all bytes objects in a range of the 
	 image, including Strings, ByteArrays, and CompiledMethods.
	 This returns these objects to their original byte ordering 
	 after blindly byte-swapping the entire image. For compiled 
	 methods, byte-swap only their bytecodes part.
	 Ensure floats are in platform-order."
	objectMemory vmEndianness = imageFloatsBigEndian ifTrue:
		[^nil].
	self assert: ClassFloatCompactIndex ~= 0.
	objectMemory allObjectsDo:
		[:obj| | temp |
		(objectMemory compactClassIndexOf: obj) = ClassFloatCompactIndex ifTrue:
			[temp := objectMemory long32At: obj + objectMemory baseHeaderSize.
			 objectMemory long32At: obj + objectMemory baseHeaderSize put: (objectMemory long32At: obj + objectMemory baseHeaderSize + 4).
			 objectMemory long32At: obj + objectMemory baseHeaderSize + 4 put: temp]]! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 07/28/2014 07:54'!
copiedValueCountOfClosure: closurePointer
	<api> "for Cogit"
	^(objectMemory numSlotsOf: closurePointer) - ClosureFirstCopiedValueIndex! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'cb 4/6/2016 10:17'!
copiedValueCountOfFullClosure: closurePointer
	<api> "for Cogit"
	^(objectMemory numSlotsOf: closurePointer) - FullClosureFirstCopiedValueIndex! !

!StackInterpreter methodsFor: 'bitblt support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:23'!
copyBits
	"This entry point needs to be implemented for the interpreter proxy.
	Since BitBlt is now a plugin we need to look up BitBltPlugin:=copyBits
	and call it. This entire mechanism should eventually go away and be
	replaced with a dynamic lookup from BitBltPlugin itself but for backward
	compatibility this stub is provided"

	| fn |
	<var: #fn type: 'void *'>
	fn := self ioLoadFunction: 'copyBits' From: 'BitBltPlugin'.
	fn = 0 ifTrue: [^self primitiveFail].
	^self cCode: '((sqInt (*)(void))fn)()'! !

!StackInterpreter methodsFor: 'bitblt support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:24'!
copyBitsFrom: x0 to: x1 at: y
	"This entry point needs to be implemented for the interpreter proxy.
	Since BitBlt is now a plugin we need to look up BitBltPlugin:=copyBitsFrom:to:at:
	and call it. This entire mechanism should eventually go away and be
	replaced with a dynamic lookup from BitBltPlugin itself but for backward
	compatibility this stub is provided"

	| fn |
	<var: #fn type: 'void *'>
	fn := self ioLoadFunction: 'copyBitsFromtoat' From: 'BitBltPlugin'.
	fn = 0 ifTrue: [^self primitiveFail].
	^self cCode: '((sqInt (*)(sqInt, sqInt, sqInt))fn)(x0, x1, y)'! !

!StackInterpreter methodsFor: 'simulation support' stamp: 'eem 2/24/2020 22:08'!
couldBeContext: anAddress
	^(objectMemory addressCouldBeObj: anAddress)
	 and: [objectMemory isContextNonImm: anAddress]! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 1/27/2015 13:33'!
couldBeProcess: oop
	<inline: #never>
	^(objectMemory addressCouldBeObj: oop)
	  and: [(objectMemory isPointersNonImm: oop)
	  and: [(objectMemory isContext: oop) not
	  and: [(objectMemory lengthOf: oop) > MyListIndex
	  and: [objectMemory isContext: (objectMemory
										fetchPointer: SuspendedContextIndex
										ofObject: oop)]]]]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem (auto pragmas 12/08) 7/8/2008 20:29'!
cr
	"For testing in Smalltalk, this method should be overridden in a subclass."
	<inline: true>
	self printf: '\n'! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 1/4/2017 09:14'!
createActualMessageTo: lookupClass 
	"Bundle up the selector, arguments and lookupClass into a Message object. 
	 In the process it pops the arguments off the stack, and pushes the message object. 
	 This can then be presented as the argument of e.g. #doesNotUnderstand:"
	| argumentArray message |
	<inline: false> "This is a useful break-point"
	self assert: ((objectMemory isImmediate: messageSelector) or: [objectMemory addressCouldBeObj: messageSelector]).
	self mnuBreakpoint: messageSelector receiver: nil.
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[argumentArray := objectMemory
								eeInstantiateSmallClassIndex: ClassArrayCompactIndex
								format: objectMemory arrayFormat
								numSlots: argumentCount.
			 message := objectMemory
								eeInstantiateSmallClassIndex: ClassMessageCompactIndex
								format: objectMemory nonIndexablePointerFormat
								numSlots: MessageLookupClassIndex + 1]
		ifFalse:
			[argumentArray := objectMemory
								eeInstantiateSmallClass: (objectMemory splObj: ClassArray)
								numSlots: argumentCount.
			 message := objectMemory
								eeInstantiateSmallClass: (objectMemory splObj: ClassMessage)
								numSlots: MessageLookupClassIndex + 1].

	"Since the array is new can use unchecked stores."
	(argumentCount - 1) * objectMemory bytesPerOop to: 0 by: objectMemory bytesPerOop negated do:
		[:i|
		objectMemory longAt:  argumentArray + objectMemory baseHeaderSize + i put: self popStack].
	"Since message is new can use unchecked stores."
	objectMemory
		storePointerUnchecked: MessageSelectorIndex ofObject: message withValue: messageSelector;
		storePointerUnchecked: MessageArgumentsIndex ofObject: message withValue: argumentArray;
		storePointerUnchecked: MessageLookupClassIndex ofObject: message withValue: lookupClass.

	self push: message.

	argumentCount := 1! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'RonieSalgado 4/27/2016 14:45'!
defaultNativeStackFrameSize
	"default native stack frame size"
	^ 256! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 10/28/2016 12:12'!
defaultNumStackPages
	"Return the default number of stack pages allocate at startup.
	 This V3 default suits Qwaq Forums (specifically general rendering).
	 The Spur default reflects tuning for GC performance ast Cadence.
	 It is probably a bit high for normal use but QF is profligate with processes.
	 The actual value can be set via vmParameterAt: and/or a preference in the ini file."
	<inline: false>
	^objectMemory hasSpurMemoryManagerAPI
		ifTrue: [50]
		ifFalse: [160]! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 1/24/2017 09:24'!
detailedSymbolicMethod: aMethod
	<doNotGenerate>
	 self transcript
		ensureCr;
		nextPutAll:
			((String streamContents:
				[:ts| | prim proxy |
				(prim := self primitiveIndexOf: aMethod) > 0 ifTrue:
					[ts nextPutAll: '<primitive: '; print: prim; nextPut: $>.
					(self isQuickPrimitiveIndex: prim) ifTrue:
						[ts nextPutAll: ' quick method'; cr; flush.
						 ^self].
					ts cr].
				proxy := VMCompiledMethodProxy new
								for: aMethod
								coInterpreter: self
								objectMemory: objectMemory.
				(DetailedInstructionPrinter on: proxy)
					stackHeightComputer: (StackDepthFinder on: proxy);
					indent: 0;
					printInstructionsOn: ts]) copyReplaceAll: 'a VMObjectProxy for ' with: '');
		flush! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'cb 8/29/2016 10:54'!
directedSuperclassSend
	"Send a message to self, starting lookup with the superclass of the class on top of stack."
	"Assume: messageSelector and argumentCount have been set, and that
	 the receiver and arguments have been pushed onto the stack,"
	"Note: This method is inlined into the interpreter dispatch loop."
	<sharedCodeInCase: #extSendSuperBytecode>
	<option: #BytecodeSetHasDirectedSuperSend>
	| class superclass |
	class := self internalPopStack.
	(objectMemory isForwarded: class) ifTrue:
		[class := objectMemory followForwarded: class].
	superclass := self superclassOf: class.
	"classTagForClass: uses ensureBehaviorHash:"
	lkupClassTag := objectMemory classTagForClass: superclass.
	"To maintain the invariant that all receivers are unforwarded we need an explicit
	 read barrier in the super send cases.  Even though we always follow receivers
	 on become  e.g. super doSomethingWith: (self become: other) forwards the receiver
	 self pushed on the stack."
	self ensureReceiverUnforwarded.
	self assert: lkupClassTag ~= objectMemory nilObject.
	self commonSendOrdinary! !

!StackInterpreter methodsFor: 'vm scheduling' stamp: 'eem 11/9/2020 18:50'!
disownVM: flags
	<api>
	<inline: false>
	"Release the VM to other threads and answer the current thread's index.
	 Currently valid flags for the non-threaded VM are:
		DisownVMLockOutFullGC	- prevent fullGCs while this thread disowns the VM
		DisownVMForFFICall			- informs the VM that it is entering an FFI call

	 This is the entry-point for plugins and primitives that wish to release the VM while
	 performing some operation that may potentially block, and for callbacks returning
	 back to some blocking operation.  While this exists for the threaded FFI VM we use
	 it to reset newMethod and the argumentCount after a callback."
	self assert: primFailCode = 0.

	"Hack encodings of client state.  We use non-immediates (bottom three bits clear)
	 for FFI/Plugin doing
		save := self disownVM: FLAGS. ... callout ... self ownVM: save.
	 We use immediate integer (bottom bit 1) for callbacks doing
		save := self ownVM: 0. ... callback ... self disownVM: save. return to C"
	self assert: ((objectMemory isImmediate: flags)
				ifFalse:
					[flags = (flags bitAnd: DisownVMForFFICall+DisownVMForThreading)
					 and: [flags anyMask: DisownVMForFFICall]]
				ifTrue:
					[(objectMemory isIntegerObject: flags)
					 and: [(objectMemory integerValueOf: flags)
							between: 0
							and: (self argumentCountOfMethodHeader: -1)]]).

	"If DisownVMForFFICall this is from the FFI plugin and we're making a callout; remember the fact."
	(((objectMemory isImmediate: flags)) not
	 and: [flags anyMask: DisownVMForFFICall]) ifTrue:
		[self assert: ((objectMemory isOopCompiledMethod: newMethod)
				and: [(self argumentCountOf: newMethod) = argumentCount]).
		 inFFIFlags := DisownVMForFFICall.
		 ^flags].

	self assert: ((objectMemory isIntegerObject: flags)
				and: [(objectMemory integerValueOf: flags)
						between: 0
						and: (self argumentCountOfMethodHeader: -1)]).

	"Otherwise this is a callback return; restore argumentCount and newMethod as per the ownVM: on callback."
	argumentCount := objectMemory integerValueOf: flags.
	newMethod := self popStack.
	self assert: ((objectMemory isOopCompiledMethod: newMethod)
				and: [(self argumentCountOf: newMethod) = argumentCount]).
	^0! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 6/16/2014 12:51'!
dispatchFunctionPointer: aFunctionPointer
	"In C aFunctionPointer is void (*aFunctionPointer)()"
	<cmacro: '(aFunctionPointer) (aFunctionPointer)()'>
	"In Smalltalk aFunctionPointer is a message selector symbol, except for
	 external primitives which are funkily encoded as integers >= 1000."
	(aFunctionPointer isInteger
	 and: [aFunctionPointer >= 1000])
		ifTrue: [self callExternalPrimitive: aFunctionPointer]
		ifFalse: [self perform: aFunctionPointer]! !

!StackInterpreter methodsFor: 'simulation' stamp: 'eem 1/15/2018 12:56'!
displayBits
	<doNotGenerate>
	^displayBits! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
divorceAllFrames
	| activeContext |
	<inline: false>
	<var: #aPage type: #'StackPage *'>
	stackPage ~= 0 ifTrue:
		[self externalWriteBackHeadFramePointers].
	activeContext := self
						ensureFrameIsMarried: framePointer
						SP: stackPointer + objectMemory wordSize.
	0 to: numStackPages - 1 do:
		[:i| | aPage |
		aPage := stackPages stackPageAt: i.
		(stackPages isFree: aPage) ifFalse:
			[self divorceFramesIn: aPage]].
	self zeroStackPage.
	^activeContext! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
divorceFramesIn: aStackPage
	| theFP calleeFP theSP theIP calleeContext theContext |
	<inline: false>
	<var: #aStackPage type: #'StackPage *'>
	<var: #theFP type: #'char *'>
	<var: #calleeFP type: #'char *'>
	<var: #theSP type: #'char *'>

	statStackPageDivorce := statStackPageDivorce + 1.

	theFP := aStackPage headFP.
	theSP := aStackPage headSP.
	theIP := stackPages longAt: theSP.
	theSP := theSP + objectMemory wordSize. "theSP points at hottest item on frame's stack"
	calleeContext := nil.

	[theContext := self ensureFrameIsMarried: theFP SP: theSP.
	 self updateStateOfSpouseContextForFrame: theFP WithSP: theSP.
	 objectMemory storePointerUnchecked: InstructionPointerIndex
		ofObject: theContext
		withValue: (self contextInstructionPointer: theIP frame: theFP).
	 self assert: (self frameReceiver: theFP)
				= (objectMemory fetchPointer: ReceiverIndex ofObject: theContext).
	 calleeContext ~~ nil ifTrue:
		[objectMemory storePointer: SenderIndex
			ofObject: calleeContext
			withValue: theContext].
	 calleeContext := theContext.
	 calleeFP := theFP.
	 theIP := (self frameCallerSavedIP: theFP) asInteger.
	 theFP := self frameCallerFP: theFP.
	 theFP ~= 0] whileTrue:
		["theSP points at stacked hottest item on frame's stack"
		 theSP := self frameCallerSP: calleeFP].

	objectMemory storePointer: SenderIndex
		ofObject: theContext
		withValue: (self frameCallerContext: calleeFP).

	"The page is now free; mark it so."
	aStackPage baseFP: 0! !

!StackInterpreter methodsFor: 'simulation' stamp: 'eem 10/23/2015 19:09:20'!
doItContext
	"This is for the bottom-right-hand pane in the simulator window."
	<doNotGenerate>
	^nil! !

!StackInterpreter methodsFor: 'simulation' stamp: 'eem 10/23/2015 19:08:56'!
doItReceiver
	"This is for the bottom-right-hand pane in the simulator window."
	<doNotGenerate>
	^self! !

!StackInterpreter methodsFor: 'arithmetic primitive support' stamp: 'eem 11/19/2014 13:59'!
doPrimitiveDiv: rcvr by: arg
	"Rounds negative results towards negative infinity, rather than zero."
	| result posArg posRcvr integerRcvr integerArg |
	(objectMemory areIntegers: rcvr and: arg)
		ifTrue: [integerRcvr := objectMemory integerValueOf: rcvr.
				integerArg := objectMemory integerValueOf: arg.
				self success: integerArg ~= 0]
		ifFalse: [self primitiveFail].
	self successful ifFalse: [^ 1 "fail"].

	integerRcvr > 0
		ifTrue: [integerArg > 0
					ifTrue: [result := integerRcvr // integerArg]
					ifFalse: ["round negative result toward negative infinity"
							posArg := 0 - integerArg.
							result := 0 - ((integerRcvr + (posArg - 1)) // posArg)]]
		ifFalse: [posRcvr := 0 - integerRcvr.
				integerArg > 0
					ifTrue: ["round negative result toward negative infinity"
							result := 0 - ((posRcvr + (integerArg - 1)) // integerArg)]
					ifFalse: [posArg := 0 - integerArg.
							result := posRcvr // posArg]].
	self success: (objectMemory isIntegerValue: result).
	^ result! !

!StackInterpreter methodsFor: 'arithmetic primitive support' stamp: 'eem 11/19/2014 13:59'!
doPrimitiveMod: rcvr by: arg
	| integerResult integerRcvr integerArg |
	(objectMemory areIntegers: rcvr and: arg)
		ifTrue: [integerRcvr := objectMemory integerValueOf: rcvr.
				integerArg := objectMemory integerValueOf: arg.
				self success: integerArg ~= 0]
		ifFalse: [self primitiveFail].
	self successful ifFalse: [^ 1 "fail"].

	integerResult := integerRcvr \\ integerArg.

	"ensure that the result has the same sign as the integerArg"
	integerArg < 0
		ifTrue: [integerResult > 0
			ifTrue: [integerResult := integerResult + integerArg]]
		ifFalse: [integerResult < 0
			ifTrue: [integerResult := integerResult + integerArg]].
	self success: (objectMemory isIntegerValue: integerResult).
	^ integerResult! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 7/13/2015 15:41'!
doSignalSemaphoreWithIndex: index
	"Signal the external semaphore with the given index.  Answer if a context switch
	 occurs as a result.  Do not bounds check.  This has been done in the caller."
	<api>
	| xArray sema |
	xArray := objectMemory splObj: ExternalObjectsArray.
	self assert: (objectMemory isArray: xArray).
	sema := objectMemory fetchPointer: index - 1 ofObject: xArray. "Note: semaphore indices are 1-based"
	self assert: (objectMemory isOopForwarded: sema) not.
	^(objectMemory isSemaphoreOop: sema)
	  and: [self synchronousSignal: sema]! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'cb 1/18/2016 18:06'!
doubleExtendedDoAnythingBytecode
	"Replaces the Blue Book double-extended send [132], in which the first byte was wasted on 8 bits of argument count. 
	Here we use 3 bits for the operation sub-type (opType),  and the remaining 5 bits for argument count where needed. 
	The last byte give access to 256 instVars or literals. 
	See also secondExtendedSendBytecode"
	| byte2 byte3 opType top |
	byte2 := self fetchByte.
	byte3 := self fetchByte.
	opType := byte2 >> 5.
	opType = 0 ifTrue:
		[messageSelector := self literal: byte3.
		 argumentCount := byte2 bitAnd: 31.
		 ^self normalSend].
	opType = 1 ifTrue:
		[messageSelector := self literal: byte3.
		 argumentCount := byte2 bitAnd: 31.
		 ^self superclassSend].
	opType = 2 ifTrue: [self fetchNextBytecode. ^self pushMaybeContextReceiverVariable: byte3].
	opType = 3 ifTrue: [self fetchNextBytecode. ^self pushLiteralConstant: byte3].
	opType = 4 ifTrue: [self fetchNextBytecode. ^self pushLiteralVariable: byte3].
	top := self internalStackTop.
	opType = 7 ifTrue:
		[self storeLiteralVariable: byte3 withValue: top.
		 ^self fetchNextBytecode].
	"opType = 5 is store; opType = 6 is storePop"
	opType = 6 ifTrue:
		[self internalPop: 1].
	self storeMaybeContextReceiverVariable: byte3 withValue: top.
	self fetchNextBytecode! !

!StackInterpreter methodsFor: 'initialization' stamp: 'JMM (auto pragmas 12/08) 12/28/2002 22:07'!
dummyReferToProxy
	<inline: false>
	interpreterProxy := interpreterProxy! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 6/30/2021 18:40'!
dumpImage: fileName
	"Dump the entire image out to the given file. Intended for debugging only.  Doesn't work for Spur."
	<notOption: #SpurObjectMemory>
	<export: true>
	<var: #f type: #sqImageFile>

	(self sqImageFile: (self pointerForOop: fileName) Open: 'wb')
		ifNil: [^-1]
		ifNotNil:
			[:f| | result |
			result := self sq: (self pointerForOop: objectMemory startOfMemory)
						Image: (self sizeof: #'unsigned char')
						File: objectMemory endOfMemory - objectMemory startOfMemory
						Write: f.
			self sqImageFileClose: f.
			^result]! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 12/13/2010 11:08'!
dumpPrimTraceLog
	"Dummy definition to allow the StackInterpreter to link against the Cog run-time."

	<api>! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:42'!
duplicateTopBytecode

	self fetchNextBytecode.
	self internalPush: self internalStackTop.
! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 9/5/2021 19:27'!
eekcr
	"For marking the end of a leak check print message"
	<api>
	<inline: #never>
	self printf: '\n'.
	self cCode: '' inSmalltalk: [self halt]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 7/11/2008 15:11'!
elementsPerPrintOopLine
	^5! !

!StackInterpreter methodsFor: 'newspeak bytecode support' stamp: 'eem 4/16/2015 11:29'!
enclosingObjectAt: depth withObject: methodReceiver withMixin: methodMixin 
	"This is used to implement outer sends and outer expressions in Newspeak."
	| enclosingObject mixinApplication targetMixin count |
	
	enclosingObject := methodReceiver.
	targetMixin := methodMixin.
	count := 0.
	[count < depth] whileTrue:
		[count := count + 1.
		self deny: (targetMixin = objectMemory nilObject).
		mixinApplication := self
							findApplicationOfTargetMixin: targetMixin
							startingAtBehavior: (objectMemory fetchClassOf: enclosingObject).
		self deny: (mixinApplication = objectMemory nilObject).
		enclosingObject := objectMemory followObjField: EnclosingObjectIndex ofObject: mixinApplication.	
		targetMixin := objectMemory followObjField: EnclosingMixinIndex ofObject: targetMixin].
	
	^enclosingObject! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
encodeFrameFieldHasContext: hasContext "<Boolean>" isBlock: isBlock "<Boolean>" numArgs: numArgs
	"For ``fast'' temporary access (ok, to mitigate slower temp access) we need
	 fast access to a method's numArgs.  Could have a variable set on save and return.
	 We'll investigate this.  For the moment we just use a byte in the frameFlags
	 field.  This is endian dependent.  Store numArgs in byte at FoxFrameFields + 1.
	 Store hasContext flag in top bit (allows for 64-bit tags) of byte at FoxFrameFields.
	 Make frameFields look like a SmallInteger for the benefit of gc (dubious)."
	"bitsPerWord := BytesPerWord * 8"
	<inline: true>
	^VMBIGENDIAN
		ifTrue: [1
				+ (numArgs << ((objectMemory wordSize * 8) - 8))
				+ (hasContext ifTrue: [1 << ((objectMemory wordSize * 8) - 16)] ifFalse: [0])
				+  (isBlock ifTrue: [1 << ((objectMemory wordSize * 8) - 24)] ifFalse: [0])]
		ifFalse: [1
				+ (numArgs << 8)
				+  (hasContext ifTrue: [1 << 16] ifFalse: [0])
				+  (isBlock ifTrue: [1 << 24] ifFalse: [0])]! !

!StackInterpreter methodsFor: 'simulation' stamp: 'eem 3/9/2018 18:37'!
encoderClassForHeader: headerInteger
	<doNotGenerate>
	^Smalltalk classNamed: ((headerInteger < 0 or: [objectMemory headerIndicatesAlternateBytecodeSet: headerInteger])
								ifTrue: [AltBytecodeEncoderClassName]
								ifFalse: [BytecodeEncoderClassName])! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 9/5/2021 20:09'!
ensureCallerContext: theFP
	"Answer the caller context for a frame.  If the frame has a caller
	 frame that doesn't have a context, then marry the caller frame."
	| callerFP |
	<inline: true>
	<var: #theFP type: #'char *'>
	<var: #callerFP type: #'char *'>
	callerFP := self frameCallerFP: theFP.
	callerFP = 0 ifTrue: "base frame, context in saved ip slot (or base of stack in Cog)"
		[^self frameCallerContext: theFP].
	^self ensureFrameIsMarried: callerFP SP: (self frameCallerStackPointer: theFP)! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 12/20/2013 11:01'!
ensureContextHasBytecodePC: aContext
	"Make sure the context has a byetcode pc.  Can only be used on single contexts.
	 This is a nop in the StackInterpreter."! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 12/3/2010 07:10'!
ensureContextIsExecutionSafeAfterAssignToStackPointer: aContext
	"Safety to give the JIT lattitude in calling convention.  A noop in the interpreter.
	 See subclass implementation(s) for explanation."! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 2/27/2013 13:50'!
ensureFrameIsMarried: theFP SP: theSP
	<inline: true>
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	(self frameHasContext: theFP) ifTrue:
		[self assert: (objectMemory isContext: (self frameContext: theFP)).
		^self frameContext: theFP].

	^self marryFrame: theFP SP: theSP! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 10/9/2013 14:32'!
ensureImageFormatIsUpToDate: swapBytes
	"Ensure the image data has been updated to suit the current VM."
	<inline: false>
	swapBytes
		ifTrue: [self reverseBytesInImage]
		ifFalse: [self convertFloatsToPlatformOrder]! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'eem 2/1/2015 16:24'!
ensureReceiverUnforwarded
	"To maintain the invariant that all receivers are unforwarded we need an explicit
	 read barrier in the super send cases."
	(objectMemory isOopForwarded: (self internalStackValue: argumentCount)) ifTrue:
		[self internalStackValue: argumentCount
			put: (objectMemory followForwarded: (self internalStackValue: argumentCount))]! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 11/13/2020 14:05'!
enterSmalltalkExecutive
	"Main entry-point into the interpreter at each execution level, where an
	 execution level is either the start of execution or reentry for a callback."
	<cmacro: '() enterSmalltalkExecutiveImplementation()'>
	"Simulation of the setjmp in enterSmalltalkExecutiveImplementation for reentry into interpreter."
	[[self enterSmalltalkExecutiveImplementation]
		on: ReenterInterpreter
		do: [:ex| self assert: ex returnValue = ReturnToInterpreter. ex return: true]] whileTrue! !

!StackInterpreter methodsFor: 'callback support' stamp: 'eem 11/11/2010 12:14'!
enterSmalltalkExecutiveFromCallback
	<inline: true>
	self enterSmalltalkExecutive! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 10/29/2020 10:16'!
enterSmalltalkExecutiveImplementation
	"Main entry-point into the interpreter at each execution level, where an execution
	 level is either the start of execution or reentry for a callback.
	 This is the actual implementation, separated from enterSmalltalkExecutive so the
	 simulator can wrap it in an exception handler and hence simulate the setjmp/longjmp."
	<inline: false>
	"Setjmp for reentry into interpreter from elsewhere, e.g. FFI exception primitive failure."
	self _setjmp: reenterInterpreter.
	self interpret.
	^0! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 2/27/2013 13:50'!
establishFrameForContextToReturnTo: contextToReturnTo
	| thePage |
	<var: #thePage type: #'StackPage *'>
	<returnTypeC: 'char *'>
	(objectMemory isContext: contextToReturnTo) ifFalse:
		[^0].
	(self isMarriedOrWidowedContext: contextToReturnTo) ifTrue:
		[(self isWidowedContext: contextToReturnTo) ifTrue:
			["error: home's sender is dead; cannot return"
			 ^0].
		 ^self frameOfMarriedContext: contextToReturnTo].
	(objectMemory isIntegerObject: (objectMemory fetchPointer: InstructionPointerIndex ofObject: contextToReturnTo)) ifFalse:
		[^0].
	thePage := self makeBaseFrameFor: contextToReturnTo.
	stackPages markStackPageMostRecentlyUsed: thePage.
	^thePage baseFP! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 6/9/2009 10:32'!
executeNewMethod
	"Execute newMethod - either primitiveFunctionPointer must be set directly
	 (i.e. from primitiveExecuteMethod et al), or it would have been set probing
	 the method cache (i.e. primitivePerform et al)."
	primitiveFunctionPointer ~= 0 ifTrue:
		[self isPrimitiveFunctionPointerAnIndex ifTrue:
			[self externalQuickPrimitiveResponse.
			 ^nil].
		 self slowPrimitiveResponse.
		 self successful ifTrue: [^nil]].
	"if not primitive, or primitive failed, activate the method"
	self activateNewMethod! !

!StackInterpreter methodsFor: 'translation support' stamp: 'eem (auto pragmas 12/08) 7/8/2008 19:55'!
expandCases
	"For translation only; noop when running in Smalltalk.
	 Must not be inlined otherwise the directive will disappear!!"
	<inline: false>! !

!StackInterpreter methodsFor: 'miscellaneous bytecodes' stamp: 'eem 10/30/2012 17:01'!
extABytecode
	"224		11100000	aaaaaaaa	Extend A (Ext A = Ext A prev * 256 + Ext A)"
	extA := (extA bitShift: 8) + self fetchByte.
	self fetchNextBytecode! !

!StackInterpreter methodsFor: 'miscellaneous bytecodes' stamp: 'rsf 1/5/2017 15:32'!
extBBytecode
	"225		11100001	sbbbbbbb	Extend B (Ext B = Ext B prev * 256 + Ext B)"
	| byte |
	byte := self fetchByte.
	self fetchNextBytecode.
	extB := (numExtB = 0 and: [byte > 127])
				ifTrue: [byte - 256]
				ifFalse: [(extB bitShift: 8) + byte].
	numExtB := numExtB + 1! !

!StackInterpreter methodsFor: 'jump bytecodes' stamp: 'rsf 1/5/2017 15:33'!
extJumpIfFalse
	"244		11110100	i i i i i i i i	Pop and Jump 0n False i i i i i i i i (+ Extend B * 256, where Extend B >= 0)"
	| byte offset |
	byte := self fetchByte.
	offset := byte + (extB << 8).
	numExtB := extB := extA := 0.
	self jumplfFalseBy: offset! !

!StackInterpreter methodsFor: 'jump bytecodes' stamp: 'rsf 1/5/2017 15:34'!
extJumpIfTrue
	"243		11110011	i i i i i i i i	Pop and Jump 0n True i i i i i i i i (+ Extend B * 256, where Extend B >= 0)"
	| byte offset |
	byte := self fetchByte.
	offset := byte + (extB << 8).
	numExtB := extB := extA := 0.
	self jumplfTrueBy: offset! !

!StackInterpreter methodsFor: 'miscellaneous bytecodes' stamp: 'rsf 1/5/2017 15:34'!
extNopBytecode
	"SistaV1		94		01011111		Nop"
	"NewspeakV4: 221		11011101		Nop"
	self fetchNextBytecode.
	numExtB := extA := extB := 0! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'rsf 1/5/2017 15:34'!
extPushCharacterBytecode
	"SistaV1:	*	233		11101001	iiiiiiii		Push Character #iiiiiiii (+ Extend B * 256)"
	| value |
	value := self fetchByte + (extB << 8).
	self fetchNextBytecode.
	self internalPush: (objectMemory characterObjectOf: value).
	numExtB := extB := 0! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'rsf 1/5/2017 15:34'!
extPushClosureBytecode
	"253		11111101 eei i i kkk	jjjjjjjj		Push Closure Num Copied iii (+ Ext A // 16 * 8) Num Args kkk (+ Ext A \\ 16 * 8) BlockSize jjjjjjjj (+ Ext B * 256). ee = num extensions.
	 The compiler has pushed the values to be copied, if any.  Find numArgs and numCopied in the byte following.
	 Create a Closure with space for the copiedValues and pop numCopied values off the stack into the closure.
	 Set numArgs as specified, and set startpc to the pc following the block size and jump over that code."
	| byte numArgs numCopied blockSize |
	byte := self fetchByte.
	numArgs := (byte bitAnd: 7) + (extA \\ 16 * 8).
	numCopied := ((byte >> 3) bitAnd: 7) + (extA // 16 * 8).
	extA := 0.
	blockSize := self fetchByte + (extB << 8).
	numExtB := extB := 0.
	self pushClosureNumArgs: numArgs copiedValues: numCopied blockSize: blockSize! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 4/8/2016 09:26'!
extPushFullClosureBytecode
	"255		11111111	xxxxxxxx	siyyyyyy	
		push Closure Compiled block literal index xxxxxxxx (+ Extend A * 256) 
		numCopied yyyyyy 
		receiverOnStack: s = 1 
		ignoreOuterContext: i = 1
	 The compiler has pushed the values to be copied, if any. The receiver has been pushed on stack before if specified. 
	 Create a Closure with space for the copiedValues and pop numCopied values off the stack into the closure.
	 Sets outerContext, compiledBlock, numArgs and receiver as specified.."
	| compiledBlockLiteralIndex compiledBlock byte numArgs numCopied receiverIsOnStack ignoreContext |
	compiledBlockLiteralIndex := self fetchByte + (extA << 8).
	extA := 0.
	compiledBlock := self literal: compiledBlockLiteralIndex.
	self assert: (objectMemory isOopCompiledMethod: compiledBlock).
	numArgs := self argumentCountOf: compiledBlock.
	byte := self fetchByte.
	numCopied := byte bitAnd: 1<< 6 - 1.
	receiverIsOnStack := byte anyMask: 1 << 7.
	ignoreContext := byte anyMask: 1 << 6.
	self pushFullClosureNumArgs: numArgs copiedValues: numCopied compiledBlock: compiledBlock receiverIsOnStack: receiverIsOnStack ignoreContext: ignoreContext! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'rsf 1/5/2017 15:39'!
extPushIntegerBytecode
	"229		11100101	i i i i i i i i	Push Integer #iiiiiiii (+ Extend B * 256, where bbbbbbbb = sddddddd, e.g. -32768 = i=0, a=0, s=1)"
	| value |
	value := self fetchByte + (extB << 8).
	self fetchNextBytecode.
	extB := 0.
	numExtB := 0.
	self internalPush: (objectMemory integerObjectOf: value)! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 10/30/2012 16:59'!
extPushLiteralBytecode
	"228		11100100	i i i i i i i i	Push Literal #iiiiiiii (+ Extend A * 256)"
	| index |
	index := self fetchByte + (extA << 8).
	self fetchNextBytecode.
	extA := 0.
	self pushLiteralConstant: index! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 10/30/2012 16:59'!
extPushLiteralVariableBytecode
	"227		11100011	i i i i i i i i	Push Literal Variable #iiiiiiii (+ Extend A * 256)"
	| index |
	index := self fetchByte + (extA << 8).
	self fetchNextBytecode.
	extA := 0.
	self pushLiteralVariable: index! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'rsf 1/5/2017 15:39'!
extPushPseudoVariable
	"SistaV1:	*	82			01010010			Push thisContext, (then e.g. Extend B 1 = push thisProcess)"
	| theThingToPush |
	extB
		caseOf: {
			[0]	->	[theThingToPush := self ensureFrameIsMarried: localFP SP: localSP].
			[1]	->	[theThingToPush := self activeProcess] }
		otherwise:
			[self respondToUnknownBytecode].
	self fetchNextBytecode.
	self internalPush: theThingToPush.
	extB := 0.
	numExtB := 0.! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'rsf 1/5/2017 15:39'!
extPushPseudoVariableOrOuterBytecode
	"77			01001101		Push false [* 1:true, 2:nil, 3:thisContext, ..., -N: pushExplicitOuter: N, N = Extend B]"
	| thing |
	self fetchNextBytecode.
	thing := extB
				caseOf: {
					[0]	->	[^self internalPush: objectMemory falseObject].
					[1]	->	[objectMemory trueObject].
					[2]	->	[objectMemory nilObject].
					[3]	->	[| context |
							 context := self ensureFrameIsMarried: localFP SP: localSP.
							 context]
				}
				otherwise:
					[extB < 0
						ifTrue:
							[self 
								enclosingObjectAt: 0 - extB 
								withObject: self receiver 
								withMixin: (self methodClassOf: method)]
						ifFalse:
							[self error: 'undefined extension for extPushPseudoVariableOrOuter'.
							 objectMemory nilObject]].
	extB := 0.
	numExtB := 0.
	self internalPush: thing! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 10/30/2012 16:58'!
extPushReceiverVariableBytecode
	"226		11100010	i i i i i i i i	Push Receiver Variable #iiiiiiii (+ Extend A * 256)"
	| index |
	index := self fetchByte + (extA << 8).
	self fetchNextBytecode.
	extA := 0.
	self pushMaybeContextReceiverVariable: index! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem 10/30/2012 17:00'!
extReturnTopFromBlock
	"218		11011010		Return Stack Top From Block [* return from enclosing block N, N = Extend A]
	 If extA is zero, return to the caller of the current block activation.
	 If extA is non-zero return to the caller of the Nth enclosing block activation."
	extA = 0 ifTrue:
		[localReturnValue := self internalStackTop.
		 ^self commonCallerReturn].
	self shouldBeImplemented.
	extA := 0! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'rsf 1/5/2017 15:40'!
extSendAbsentDynamicSuperBytecode
	"241		11110001	i i i i i j j j	Send To Dynamic Superclass Literal Selector #iiiii (+ Extend A * 32) with jjj (+ Extend B * 8) Arguments"
	| byte |
	byte := self fetchByte.
	messageSelector := self literal: (byte >> 3) + (extA << 5).
	extA := 0.
	argumentCount := (byte bitAnd: 7) + (extB << 3).
	extB := 0.
	numExtB := 0.
	self commonSendDynamicSuper! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'rsf 1/5/2017 15:40'!
extSendAbsentImplicitBytecode
	"240		11110000	i i i i i j j j	Send To Absent Implicit Receiver Literal Selector #iiiii (+ Extend A * 32) with jjj (+ Extend B * 8) Arguments"
	| byte |
	byte := self fetchByte.
	messageSelector := self literal: (byte >> 3) + (extA << 5).
	extA := 0.
	argumentCount := (byte bitAnd: 7) + (extB << 3).
	extB := 0.
	numExtB := 0.
	self commonSendImplicitReceiver.! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'rsf 1/5/2017 15:40'!
extSendAbsentOuterBytecode
	"254		  11111110 	i i i i i j j j	kkkkkkkk Send To Enclosing Object at Depth kkkkkkkk Literal Selector #iiiii (+ Extend A * 32) with jjj (+ Extend B * 8) Arguments"
	| byte depth |
	byte := self fetchByte.
	messageSelector := self literal: (byte >> 3) + (extA << 5).
	extA := 0.
	argumentCount := (byte bitAnd: 7) + (extB << 3).
	extB := 0.
	numExtB := 0.
	depth := self fetchByte.
	self commonSendOuter: depth! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'rsf 1/5/2017 15:40'!
extSendAbsentSelfBytecode
	"245		 11110101 	i i i i i j j j	Send To Self Literal Selector #iiiii (+ Extend A * 32) with jjj (+ Extend B * 8) Arguments"
	| byte |
	byte := self fetchByte.
	messageSelector := self literal: (byte >> 3) + (extA << 5).
	extA := 0.
	argumentCount := (byte bitAnd: 7) + (extB << 3).
	extB := 0.
	numExtB := 0.
	self commonSendOuter: 0! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'rsf 1/5/2017 15:40'!
extSendBytecode
	"238		11101110	i i i i i j j j	Send Literal Selector #iiiii (+ Extend A * 32) with jjj (+ Extend B * 8) Arguments"
	| byte rcvr |
	byte := self fetchByte.
	messageSelector := self literal: (byte >> 3) + (extA << 5).
	extA := 0.
	argumentCount := (byte bitAnd: 7) + (extB << 3).
	extB := 0.
	numExtB := 0.
	rcvr := self internalStackValue: argumentCount.
	lkupClassTag := objectMemory fetchClassTagOf: rcvr.
	self commonSendOrdinary! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'cb 3/13/2017 18:08'!
extSendSuperBytecode
	"239		11101111	i i i i i j j j
		ExtendB < 64
			ifTrue: [Send To Superclass Literal Selector #iiiii (+ Extend A * 32) with jjj (+ Extend B * 8) Arguments]
			ifFalse: [Send To Superclass of Stacked Class Literal Selector #iiiii (+ Extend A * 32) with jjj (+ (Extend B bitAnd: 63) * 8) Arguments]"
	| byte |
	byte := self fetchByte.
	messageSelector := self literal: (byte >> 3) + (extA << 5).
	extA := 0.
	BytecodeSetHasDirectedSuperSend ifTrue:
		[extB >= 64 ifTrue:
			[argumentCount := (byte bitAnd: 7) + (extB - 64 << 3).
			 extB := 0.
			 numExtB := 0.
			 ^self directedSuperclassSend]].
	argumentCount := (byte bitAnd: 7) + (extB << 3).
	extB := 0.
	numExtB := 0.
	self superclassSend! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 1/18/2016 18:03'!
extStoreAndPopLiteralVariableBytecode
	"236		11101100	i i i i i i i i	Pop and Store Literal Variable #iiiiiiii (+ Extend A * 256)"
	| variableIndex value |
	variableIndex := self fetchByte + (extA << 8).
	value := self internalStackTop.
	self internalPop: 1.
	extA := 0.
	self storeLiteralVariable: variableIndex withValue: value.
	self fetchNextBytecode.! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 1/18/2016 18:05'!
extStoreAndPopReceiverVariableBytecode
	"235		11101011	i i i i i i i i	Pop and Store Receiver Variable #iiiiiii (+ Extend A * 256)"
	| variableIndex value |
	variableIndex := self fetchByte + (extA << 8).
	extA := 0.
	value := self internalStackTop.
	self internalPop: 1.
	self storeMaybeContextReceiverVariable: variableIndex withValue: value.
	self fetchNextBytecode.! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 1/18/2016 18:03'!
extStoreLiteralVariableBytecode
	"233		11101001	i i i i i i i i	Store Literal Variable #iiiiiiii (+ Extend A * 256)"
	| variableIndex |
	variableIndex := self fetchByte + (extA << 8).
	extA := 0.
	self storeLiteralVariable: variableIndex withValue: self internalStackTop.
	self fetchNextBytecode.! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 1/18/2016 18:05'!
extStoreReceiverVariableBytecode
	"232		11101000	i i i i i i i i	Store Receiver Variable #iiiiiii (+ Extend A * 256)"
	| variableIndex |
	variableIndex := self fetchByte + (extA << 8).
	extA := 0.
	self storeMaybeContextReceiverVariable: variableIndex withValue: self internalStackTop.
	self fetchNextBytecode.! !

!StackInterpreter methodsFor: 'jump bytecodes' stamp: 'eem 3/24/2017 15:39'!
extUnconditionalJump
	"242		11110010	i i i i i i i i	Jump i i i i i i i i (+ Extend B * 256, where bbbbbbbb = sddddddd, e.g. -32768 = i=0, a=0, s=1)"
	| byte offset bcpcDelta |
	byte := self fetchByte.
	offset := byte + (extB << 8).
	bcpcDelta := offset < 0 ifTrue: [numExtB * 2] ifFalse: [0].
	extB := 0.
	numExtB := 0.
	localIP := localIP + offset.
	self ifBackwardsCheckForEvents: offset + bcpcDelta.
	self fetchNextBytecode! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:46'!
extendedPushBytecode

	| descriptor variableType variableIndex |
	descriptor := self fetchByte.
	self fetchNextBytecode.
	variableType := (descriptor >> 6) bitAnd: 16r3.
	variableIndex := descriptor bitAnd: 16r3F.
	variableType=0 ifTrue: [^self pushReceiverVariable: variableIndex].
	variableType=1 ifTrue: [^self pushTemporaryVariable: variableIndex].
	variableType=2 ifTrue: [^self pushLiteralConstant: variableIndex].
	variableType=3 ifTrue: [^self pushLiteralVariable: variableIndex].
! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 1/6/2016 14:41'!
extendedStoreAndPopBytecode
	<inline: true>
	self extendedStoreBytecodePop: true
	"may not be reached (immutable receiver)"! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 1/6/2016 14:41'!
extendedStoreBytecode
	<inline: true>
	self extendedStoreBytecodePop: false
	"may not be reached (immutable receiver)"! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 8/20/2020 10:20'!
extendedStoreBytecodePop: popBoolean
	| descriptor variableIndex value |
	<inline: true>
	descriptor := self fetchByte.
	variableIndex := descriptor bitAnd: 63.
	value := self internalStackTop.
	popBoolean ifTrue: [self internalPop: 1].
	(descriptor >> 6 bitAnd: 3) caseOf: {
		[0] ->	[objectMemory storePointerImmutabilityCheck: variableIndex ofObject: self receiver withValue: value].
		[1] ->	[self temporary: variableIndex in: localFP put: value].
		[2] ->	[self error: 'illegal store'].
		[3] ->	[self storeLiteralVariable: variableIndex withValue: value] }.
	self fetchNextBytecode! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 12/8/2017 12:50'!
externalDivorceFrame: theFP andContext: ctxt
	"Divorce a single frame and its context.  If it is not the top frame of a stack this means splitting its stack."
	| thePage onCurrent theSP callerCtx newPage frameAbove callerFP callerSP callerIP theIP |
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<var: #theSP type: #'char *'>
	<var: #newPage type: #'StackPage *'>
	<var: #frameAbove type: #'char *'>
	<var: #callerFP type: #'char *'>
	<var: #callerSP type: #'char *'>
	"stackPage needs to have current head pointers to avoid confusion."
	self assert: (stackPage = 0 or: [stackPage = stackPages mostRecentlyUsedPage]).
	thePage := stackPages stackPageFor: theFP.
	(onCurrent := thePage = stackPage) ifFalse:
		[stackPages markStackPageNextMostRecentlyUsed: thePage].
	theSP := self findSPOf: theFP on: thePage.
	self updateStateOfSpouseContextForFrame: theFP WithSP: theSP.
	callerCtx := self ensureCallerContext: theFP.
	(frameAbove := self findFrameAbove: theFP inPage: thePage) = 0
		ifTrue: "If we're divorcing the top frame we can simply peel it off."
			[theIP := stackPages longAt: thePage headSP]
		ifFalse: "othewise move all frames above to a new stack and then peel the frame off."
			[newPage := stackPages newStackPage.
			 theIP := self oopForPointer: (self frameCallerSavedIP: frameAbove).
			 frameAbove := self moveFramesIn: thePage through: frameAbove toPage: newPage.
			 onCurrent
				ifTrue:
					[self setStackPageAndLimit: newPage.
					 self setStackPointersFromPage: newPage]
				ifFalse:
					[stackPages markStackPageMostRecentlyUsed: newPage].
			 self assert: (self frameCallerContext: frameAbove) = ctxt].
	objectMemory storePointerUnchecked: InstructionPointerIndex
		ofObject: ctxt
		withValue: (self contextInstructionPointer: theIP frame: theFP).
	objectMemory storePointer: SenderIndex
		ofObject: ctxt
		withValue: callerCtx.
	callerFP := self frameCallerFP: theFP.
	callerFP = 0 "theFP is a base frame; it is now alone; free the entire page"
		ifTrue: [stackPages freeStackPage: thePage]
		ifFalse:
			[callerIP := self oopForPointer: (self frameCallerSavedIP: theFP).
			 callerSP := (self frameCallerSP: theFP) - objectMemory wordSize.
			 stackPages longAt: callerSP put: callerIP.
			 self setHeadFP: callerFP andSP: callerSP inPage: thePage]
	! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 3/10/2017 12:47'!
externalEnsureIsBaseFrame: aFramePtr
	"Ensure aFramePtr is a base frame.  Then we can assign its sender.
	 Answer the possibly moved location of the frame."
	| theFP thePage onCurrent |
	<var: #aFramePtr type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<returnTypeC: 'char *'>
	(self isBaseFrame: aFramePtr) ifTrue:
		[^aFramePtr].
	theFP := aFramePtr.
	thePage := stackPages stackPageFor: theFP.
	onCurrent := thePage = stackPage.
	"Storing the frame's sender with its caller's context
	 has the side effect of making theFP a base frame."
	theFP := self
				storeSenderOfFrame: theFP
				withValue: (self ensureCallerContext: theFP).
	onCurrent
		ifTrue:
			[self assert: stackPage ~~ thePage. "stackPage has moved to a new page"
			 self setStackPointersFromPage: stackPage]
		ifFalse:
			[stackPages markStackPageMostRecentlyUsed: stackPage].
	self assert: stackPages pageListIsWellFormed.
	self assert: stackPage = stackPages mostRecentlyUsedPage.
	^theFP! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 9/7/2016 14:24'!
externalInstVar: offset ofContext: aContext
	"Fetch an instance variable from a maybe married context.
	 If the context is still married compute the value of the
	 relevant inst var from the spouse frame's state."
	<inline: false>

	self assert: (objectMemory isContext: aContext).
	self assert: offset <= (ReceiverIndex + (self checkStackPointerForMaybeMarriedContext: aContext)).
	"method, closureOrNil & receiver need no special handling; only
	 sender, pc & stackp have to be computed for married contexts."
	(self isReadMediatedContextInstVarIndex: offset) ifTrue:
		[self externalWriteBackHeadFramePointers.
		 (self isStillMarriedContext: aContext) ifTrue:
			[^self fetchPointer: offset ofMarriedContext: aContext]].
	
	^objectMemory fetchPointer: offset ofObject: aContext! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 3/21/2017 14:58'!
externalInstVar: index ofContext: maybeMarriedContext put: anOop
	| theFP thePage onCurrentPage |
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	self assert: (objectMemory isContext: maybeMarriedContext).
	self externalWriteBackHeadFramePointers.
	"Assign the field of a married context."
	self deny: (objectMemory isObjImmutable: maybeMarriedContext).
	(self isStillMarriedContext: maybeMarriedContext) ifFalse:
		[objectMemory storePointer: index ofObject: maybeMarriedContext withValue: anOop.
		 index = StackPointerIndex ifTrue:
			[self ensureContextIsExecutionSafeAfterAssignToStackPointer: maybeMarriedContext].
		 ^nil].
	theFP := self frameOfMarriedContext: maybeMarriedContext.
	thePage := stackPages stackPageFor: theFP.
	self assert: stackPage = stackPages mostRecentlyUsedPage.
	onCurrentPage := thePage = stackPage.
	index = SenderIndex
		ifTrue:
			[self storeSenderOfFrame: theFP withValue: anOop]
		ifFalse:
			[self externalDivorceFrame: theFP andContext: maybeMarriedContext.
			 (self is: index methodAssignmentToContextWithMachineCodePC: maybeMarriedContext) ifTrue:
				[self ensureContextHasBytecodePC: maybeMarriedContext].
			 objectMemory storePointer: index ofObject: maybeMarriedContext withValue: anOop.
			 index = StackPointerIndex ifTrue:
				[self ensureContextIsExecutionSafeAfterAssignToStackPointer: maybeMarriedContext]].
	onCurrentPage
		ifTrue:
			[self setStackPointersFromPage: stackPage]
		ifFalse:
			[stackPages markStackPageMostRecentlyUsed: stackPage].
	stackPages assert: stackPage = stackPages mostRecentlyUsedPage.
	stackPages assert: stackPages pageListIsWellFormed.
	stackPages assert: self validStackPageBaseFrames! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 6/26/2012 17:53'!
externalQuickPrimitiveResponse
	"Invoke a quick primitive.
	 Called under the assumption that primFunctionPtr has been preloaded"
	
	| localPrimIndex |
	self assert: self isPrimitiveFunctionPointerAnIndex.
	localPrimIndex := self cCoerceSimple: primitiveFunctionPointer to: #sqInt.
	self assert: (localPrimIndex > 255 and: [localPrimIndex < 520]).
	"Quick return inst vars"
	localPrimIndex >= 264 ifTrue:
		[self pop: 1 thenPush: (objectMemory fetchPointer: localPrimIndex - 264 ofObject: self stackTop).
		 ^true].
	"Quick return constants"
	localPrimIndex = 256 ifTrue: [^true "return self"].
	localPrimIndex = 257 ifTrue: [self pop: 1 thenPush: objectMemory trueObject. ^true].
	localPrimIndex = 258 ifTrue: [self pop: 1 thenPush: objectMemory falseObject. ^true].
	localPrimIndex = 259 ifTrue: [self pop: 1 thenPush: objectMemory nilObject. ^true].
	self pop: 1 thenPush: (objectMemory integerObjectOf: localPrimIndex - 261).
	^true! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 3/10/2017 12:53'!
externalSetStackPageAndPointersForSuspendedContextOfProcess: aProcess
	"Set stackPage, instructionPointer, framePointer and stackPointer for the suspendedContext of
	 aProcess, marrying the context if necessary, and niling the suspendedContext slot.  This is used
	 on process switch to ensure a context has a stack frame and so can continue execution."
	| newContext theFrame thePage newPage |
	<inline: true>
	<var: #theFrame type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<var: #newPage type: #'StackPage *'>
	
	newContext := objectMemory fetchPointer: SuspendedContextIndex ofObject: aProcess.
	self assert: (objectMemory isContext: newContext).
	(self isMarriedOrWidowedContext: newContext) ifTrue:
		[self assert: (self checkIsStillMarriedContext: newContext currentFP: framePointer)].
	objectMemory
		storePointerUnchecked: SuspendedContextIndex
		ofObject: aProcess
		withValue: objectMemory nilObject.
	(self isStillMarriedContext: newContext)
		ifTrue:
			[theFrame := self frameOfMarriedContext: newContext.
			 thePage := stackPages stackPageFor: theFrame.
			 theFrame ~= thePage headFP ifTrue:
				["explicit assignment of suspendedContext can cause switch to interior frame."
				 newPage := stackPages newStackPage.
				 self moveFramesIn: thePage
					through: (self findFrameAbove: theFrame inPage: thePage)
					toPage: newPage.
				  stackPages markStackPageLeastMostRecentlyUsed: newPage].
			 self assert: thePage headFP = theFrame]
		ifFalse:
			[thePage := self makeBaseFrameFor: newContext.
			 theFrame := thePage baseFP].
	self setStackPageAndLimit: thePage.
	self setStackPointersFromPage: thePage.
	(self isMachineCodeFrame: framePointer) ifFalse:
		[self setMethod: (self iframeMethod: framePointer)].
	instructionPointer := self popStack.
	self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 5/30/2018 12:28'!
externalWriteBackHeadFramePointers
	self assert:  (framePointer - stackPointer) < (LargeContextSlots * objectMemory bytesPerOop).
	self assert: stackPage = stackPages mostRecentlyUsedPage.
	self deny: stackPage isFree.
	self setHeadFP: framePointer andSP: stackPointer inPage: stackPage.
	self assert: stackPages pageListIsWellFormed! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 11/13/2014 13:16'!
externalizeFPandSP
	"Copy the frame and stack pointers to global variables for use in primitives and other functions outside the interpret loop."
	self assert: (localSP < stackPage baseAddress
				and: [localSP > (stackPage realStackLimit - (LargeContextSlots * objectMemory bytesPerOop))]).
	stackPointer := localSP.
	framePointer := localFP! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 6/17/2008 14:09'!
externalizeIPandSP
	"Copy the local instruction, stack and frame pointers to global variables for use in primitives and other functions outside the interpret loop."

	instructionPointer := self oopForPointer: localIP.
	stackPointer := localSP.
	framePointer := localFP! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 1/28/2009 15:35'!
failUnbalancedPrimitive
	"not inlined for breakpoint value..."
	<inline: false>
	self primitiveFailFor: PrimErrBadNumArgs! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
fetchArray: fieldIndex ofObject: objectPointer
	"Fetch the instance variable at the given index of the given object. Return the address of first indexable field of resulting array object, or fail if the instance variable does not contain an indexable bytes or words object."
	"Note: May be called by translated primitive code."

	| arrayOop |
	<returnTypeC: 'void *'>
	arrayOop := objectMemory fetchPointer: fieldIndex ofObject: objectPointer.
	^ self arrayValueOf: arrayOop
! !

!StackInterpreter methodsFor: 'interpreter shell' stamp: 'eem 6/25/2011 12:21'!
fetchByte
	"This method uses the preIncrement builtin function which has no Smalltalk equivalent. Thus, it must be overridden in the simulator."

	^objectMemory byteAtPointer: localIP preIncrement! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 11/21/2014 17:14'!
fetchFloat: fieldIndex ofObject: objectPointer
	"Fetch the instance variable at the given index of the given object. Return the C double precision floating point value of that instance variable, or fail if it is not a Float."
	"Note: May be called by translated primitive code."

	| floatOop |
	<returnTypeC: #double>
	floatOop := objectMemory fetchPointer: fieldIndex ofObject: objectPointer.
	^objectMemory floatValueOf: floatOop! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
fetchInteger: fieldIndex ofObject: objectPointer
	"Note: May be called by translated primitive code."

	| intOop |
	<inline: false>
	intOop := objectMemory fetchPointer: fieldIndex ofObject: objectPointer.
	^self checkedIntegerValueOf: intOop! !

!StackInterpreter methodsFor: 'interpreter shell' stamp: 'eem 4/28/2016 17:24'!
fetchNextBytecode
	"This method fetches the next instruction (bytecode). Each bytecode method is responsible for fetching the next bytecode, preferably as early as possible to allow the memory system time to process the request before the next dispatch."

	MULTIPLEBYTECODESETS
		ifTrue: [currentBytecode := self fetchByte + bytecodeSetSelector]
		ifFalse: [currentBytecode := self fetchByte]! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 10/29/2014 15:24'!
fetchPointer: offset ofMarriedContext: aContext
	"Fetch a slot from a married context.  Compute the value
	 of the relevant inst var from the spouse frame's state.

	 This method assumes frame pointers have been written back."
	| spouseFP |
	<inline: false>
	<var: #spouseFP type: #'char *'>

	self assert: (objectMemory isContext: aContext).
	self assert: (stackPage headFP = framePointer and: [stackPage headSP = stackPointer]).
	self assert: (self checkIsStillMarriedContext: aContext currentFP: framePointer).

	"method, closureOrNil & receiver need no special handling; only
	 sender, pc & stackp have to be computed for married contexts."
	offset <= ReceiverIndex ifTrue:
		[(self isReadMediatedContextInstVarIndex: offset) ifFalse:
			[^objectMemory fetchPointer: offset ofObject: aContext].

		 spouseFP := self frameOfMarriedContext: aContext.
		 offset = SenderIndex ifTrue:
			[^self ensureCallerContext: spouseFP].
		 offset = StackPointerIndex ifTrue:
			[^objectMemory integerObjectOf: (self stackPointerIndexForFrame: spouseFP)].
		 offset = InstructionPointerIndex ifTrue:
			[^self instructionPointerForFrame: spouseFP currentFP: framePointer currentIP: instructionPointer]].
	
	spouseFP := self frameOfMarriedContext: aContext.
	^(offset - ReceiverIndex between: 1 and: (self stackPointerIndexForFrame: spouseFP))
		ifTrue: [self temporary: offset - (ReceiverIndex + 1) in: spouseFP]
		ifFalse: [objectMemory nilObject]! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
fetchStackPointerOf: aContext
	"Return the stackPointer of a Context or BlockContext.
	 Does not deal with married contexts.  Use only for debug
	 printing or object tracing functions.  To obtain an accurate
	 stack pointer use stackPointerForMaybeMarriedContext:"
	| sp |
	<inline: true>
	sp := objectMemory fetchPointer: StackPointerIndex ofObject: aContext.
	(objectMemory isIntegerObject: sp) ifFalse: [^0].
	self assert: ReceiverIndex + (objectMemory integerValueOf: sp) < (objectMemory lengthOf: aContext).
	^objectMemory integerValueOf: sp! !

!StackInterpreter methodsFor: 'plugin support' stamp: 'eem 11/8/2020 22:54'!
fileTimesInUTC
	"Answer if file times, including those from the FilePlugin and FileAttributesPlugin,
	 should be answered in UTC seconds from the epoch rather than local seconds."
	^imageHeaderFlags anyMask: 512! !

!StackInterpreter methodsFor: 'newspeak bytecode support' stamp: 'rmacnak 5/20/2015 21:26'!
findApplicationOfTargetMixin: targetMixin startingAtBehavior: aBehavior
	"This is used to implement implicit receiver and enclosing object lookup
	 for Newspeak. Find the mixin applcation of which aClass is a subclass that
	 is an application of targetMixin. This is an implementation derived from

	<ContextPart> findApplicationOf: targetMixin startingAt: aBehavior
	"
	| mixinApplication mixin |
	mixinApplication := aBehavior.
	[mixinApplication = objectMemory nilObject
	 or: [mixinApplication = targetMixin
	 or: [(mixin := objectMemory followObjField: MixinIndex ofObject: mixinApplication) = targetMixin]]] whileFalse:
		[mixinApplication := objectMemory followObjField: SuperclassIndex ofObject: mixinApplication].

	mixinApplication = objectMemory nilObject ifTrue: [
		self print: 'looking for '; cr.
		self longPrintOop: targetMixin; cr.
		self print: ' in behavior '; cr.
		self longPrintOop: aBehavior; cr.
	].

	^mixinApplication! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 10/30/2014 10:09'!
findClassContainingMethod: meth startingAt: classObj
	| currClass classDict classDictSize methodArray i |
	(objectMemory isOopForwarded: classObj)
		ifTrue: [currClass := objectMemory followForwarded: classObj]
		ifFalse: [currClass := classObj].
	[self assert: (objectMemory isForwarded: currClass) not.
	 (self addressCouldBeClassObj: currClass) ifFalse:
		[^objectMemory nilObject].
	 classDict := objectMemory noFixupFollowField: MethodDictionaryIndex ofObject: currClass.
	 self assert: (objectMemory isForwarded: classDict) not.
	 classDictSize := objectMemory numSlotsOf: classDict.
	 classDictSize > MethodArrayIndex ifTrue:
		[methodArray := objectMemory noFixupFollowField: MethodArrayIndex ofObject: classDict.
		 self assert: (objectMemory isForwarded: methodArray) not.
		 i := 0.
		 [i < (classDictSize - SelectorStart)] whileTrue:
			[meth = (objectMemory noFixupFollowField: i ofObject: methodArray) ifTrue:
				[^currClass].
			 i := i + 1]].
	 currClass := self noFixupSuperclassOf: currClass.
	 currClass = objectMemory nilObject] whileFalse.
	^currClass		"method not found in superclass chain"! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 07/28/2014 07:54'!
findClassForSelector: aSelector lookupClass: startClass do: unaryBlock
	"Search startClass' class hierarchy looking for aSelector and if found, evaluate unaryBlock
	 with the class where the selector is found.  Otherwise evaluate unaryBlock with nil."
	| currClass classDict classDictSize i |
	currClass := startClass.
	[classDict := objectMemory fetchPointer: MethodDictionaryIndex ofObject: currClass.
	 classDictSize := objectMemory numSlotsOf: classDict.
	 i := SelectorStart.
	 [i < classDictSize] whileTrue:
		[aSelector = (objectMemory fetchPointer: i ofObject: classDict) ifTrue:
			[^unaryBlock value: currClass].
			i := i + 1].
	 currClass := self superclassOf: currClass.
	 currClass = objectMemory nilObject] whileFalse.
	^unaryBlock value: nil    "selector not found in superclass chain"
		! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 10/30/2014 10:11'!
findClassOfMethod: meth forReceiver: rcvr
	((objectMemory addressCouldBeOop: rcvr)
	and: [(objectMemory isOopForwarded: rcvr) not]) ifTrue:
		[| rclass |
		 rclass := self
					findClassContainingMethod: meth
					startingAt: (objectMemory fetchClassOf: rcvr).
		 rclass ~= objectMemory nilObject ifTrue:
			[^rclass]].
	((objectMemory addressCouldBeObj: meth)
	 and: [objectMemory isCompiledMethod: meth]) ifFalse:
		[^objectMemory nilObject].
	^self
		findClassContainingMethod: meth
		startingAt: (self safeMethodClassOf: meth)! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 9/17/2009 09:07'!
findFrameAbove: theFP inPage: thePage
	"Answer the frame above theFP (adjacent frame nearest head end).
	 If theFP is the head frame answer 0."
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	| fp callerFP |
	<var: #fp type: #'char *'>
	<var: #callerFP type: #'char *'>
	<returnTypeC: #'char *'>
	fp := thePage headFP.
	fp = theFP ifTrue:
		[^0].
	[(callerFP := self frameCallerFP: fp) ~= 0] whileTrue:
		[callerFP = theFP ifTrue:
			[^fp].
		 fp := callerFP].
	self error: 'did not find theFP in stack page'.
	^0! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 9/5/2016 13:09'!
findHomeForContext: aContext
	| closureOrNil |
	<inline: false>
	(objectMemory isContext: aContext) ifFalse:
		[^nil].
	closureOrNil := objectMemory fetchPointer: ClosureIndex ofObject: aContext.
	closureOrNil = objectMemory nilObject ifTrue:
		[^aContext].
	ClassBlockClosureCompactIndex ~= 0
		ifTrue:
			[((objectMemory compactClassIndexOf: closureOrNil) ~= ClassBlockClosureCompactIndex
			  and: [(objectMemory compactClassIndexOf: closureOrNil) ~= ClassFullBlockClosureCompactIndex]) ifTrue:
				[^nil]]
		ifFalse:
			[(objectMemory fetchClassOf: closureOrNil) ~= (objectMemory splObj: ClassBlockClosure) ifTrue:
				[(ClassFullBlockClosureCompactIndex ~= 0
				  and: [(objectMemory fetchClassOf: closureOrNil) = (objectMemory splObj: ClassFullBlockClosure)]) ifFalse:
					[^nil]]].
	^self findHomeForContext: (objectMemory fetchPointer: ClosureOuterContextIndex ofObject: closureOrNil)! !

!StackInterpreter methodsFor: 'handler search' stamp: 'eem 1/23/2016 17:17'!
findMethodWithPrimitive: primitive FromContext: senderContext UpToContext: homeContext
	"See findUnwindThroughContext:.  Alas this is mutually recursive with
	 findMethodWithPrimitive:FromFP:SP:ThroughContext: instead of iterative.
	 We're doing the simplest thing that could possibly work.  Niceties can wait."
	<inline: false>
	<returnTypeC: #sqInt> "Being mutually-recursive with findMethodWithPrimitive:FromFP:UpToContext: gives the author's type inference algorithm headaches.  Wimp out by declaring the return type."
	| theContext theMethod |
	self assert: (senderContext = objectMemory nilObject or: [objectMemory isContext: senderContext]).
	self assert: (homeContext = objectMemory nilObject or: [objectMemory isContext: homeContext]).
	theContext := senderContext.
	[theContext = objectMemory nilObject ifTrue:
		[^theContext].
	 self isMarriedOrWidowedContext: theContext] whileFalse:
		[theContext = homeContext ifTrue: [^0].
		 (primitive = 0
		  or: [(objectMemory fetchPointer: ClosureIndex ofObject: theContext) ~= objectMemory nilObject]) ifFalse:
		 	[theMethod := objectMemory fetchPointer: MethodIndex ofObject: theContext.
			 (self primitiveIndexOf: theMethod) = primitive ifTrue:
				[^theContext]].
		 theContext := objectMemory fetchPointer: SenderIndex ofObject: theContext].
	(self isWidowedContext: theContext) ifTrue:
		[^objectMemory nilObject].
	^self
		findMethodWithPrimitive: primitive
		FromFP: (self frameOfMarriedContext: theContext)
		UpToContext: homeContext! !

!StackInterpreter methodsFor: 'handler search' stamp: 'eem 12/28/2014 17:43'!
findMethodWithPrimitive: primitive FromFP: startFP UpToContext: homeContext
	"See findUnwindThroughContext:.  Alas this is mutually recursive with
	 findMethodWithPrimitive:FromContext:ThroughContext: instead of iterative.
	 We're doing the simplest thing that could possibly work.  Niceties can wait."
	<inline: true>
	| theFP theFPAbove theSP theMethod senderContext |
	<var: #startFP type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #theFPAbove type: #'char *'>
	<var: #theSP type: #'char *'>
	theFP := startFP.
	theFPAbove := startFP.
	[((self frameHasContext: theFP)
	  and: [homeContext = (self frameContext: theFP)]) ifTrue:
		[^0].
	 (primitive = 0
	  or: [self frameIsBlockActivation: theFP]) ifFalse:
	 	[theMethod := self frameMethodObject: theFP.
		 (self primitiveIndexOf: theMethod) = primitive ifTrue:
			[theFP = theFPAbove
						ifTrue: [theSP := self findSPOf: theFP on: (stackPages stackPageFor: theFP)]
						ifFalse: [theSP := self frameCallerStackPointer: theFPAbove].
			 ^self ensureFrameIsMarried: theFP SP: theSP]].
	 theFPAbove := theFP.
	 theFP := self frameCallerFP: theFP.
	 theFP ~= 0] whileTrue.
	senderContext := self frameCallerContext: theFPAbove.
	(objectMemory isContext: senderContext) ifFalse:
		[^objectMemory nilObject].
	^self
		findMethodWithPrimitive: primitive
		FromContext: senderContext
		UpToContext: homeContext! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 2/12/2016 14:00'!
findNewMethodInClassTag: classTagArg
	"Find the compiled method to be run when the current 
	messageSelector is sent to the given class, setting the values 
	of 'newMethod' and 'primitiveIndex'."
	| ok classTag |
	<inline: false>
	ok := self lookupInMethodCacheSel: messageSelector classTag: classTagArg.
	ok ifFalse: "entry was not found in the cache; look it up the hard way "
		[classTag := classTagArg.
		 ((objectMemory isOopForwarded: messageSelector)
		  or: [objectMemory isForwardedClassTag: classTag]) ifTrue:
			[(objectMemory isOopForwarded: messageSelector) ifTrue:
				[messageSelector := self handleForwardedSelectorFaultFor: messageSelector].
			 (objectMemory isForwardedClassTag: classTag) ifTrue:
				[classTag := self handleForwardedSendFaultForTag: classTag].
			ok := self lookupInMethodCacheSel: messageSelector classTag: classTag.
			ok ifTrue:
				[^nil]].
 		 lkupClass := objectMemory classForClassTag: classTag.
		 NewspeakVM
				ifTrue: [self lookupOrdinarySend]
				ifFalse: [self lookupMethodInClass: lkupClass].
		 self addNewMethodToCache: lkupClass]! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 6/12/2017 14:11'!
findSPOf: theFP on: thePage
	"Search for the stack pointer for theFP.  This points to the hottest item on the frame's stack.
	 DO NOT CALL THIS WITH theFP == localFP OR theFP == framePointer!!"
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<returnTypeC: #'char *'>
	| theSP |
	<inline: false>
	<var: #theSP type: #'char *'>
	self deny: (stackPages isFree: thePage).
	theSP := self findSPOrNilOf: theFP on: thePage startingFrom: thePage headFP.
	theSP ifNotNil:
		[^theSP].
	self error: 'did not find theFP in stack page'.
	^0! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 6/12/2017 12:47'!
findSPOrNilOf: theFP on: thePage startingFrom: startFrame
	"Search for the stack pointer for theFP.  This points to the hottest item on the frame's stack.
	 DO NOT CALL THIS WITH theFP == localFP OR theFP == framePointer!!"
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<var: #startFrame type: #'char *'>
	<returnTypeC: #'char *'>
	| aFrame prevFrame |
	<inline: true>
	<var: #aFrame type: #'char *'>
	<var: #prevFrame type: #'char *'>
	startFrame = theFP ifTrue:
		[thePage headSP >= startFrame ifTrue:
			["If the SP is invalid return the pointer to the receiver field."
			 ^self frameReceiverLocation: theFP].
		 "Skip the instruction pointer on top of stack of inactive pages."
		^thePage = stackPage
			ifTrue: [thePage headSP]
			ifFalse: [thePage headSP + objectMemory wordSize]].
	aFrame := startFrame.
	[prevFrame := aFrame.
	 aFrame := self frameCallerFP: aFrame.
	 aFrame ~= 0] whileTrue:
		[theFP = aFrame ifTrue:
			[^self frameCallerSP: prevFrame]].
	^nil! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 10/21/2014 16:57'!
findSelectorAndClassForMethod: meth lookupClass: startClass do: binaryBlock
	"Search startClass' class hierarchy searching for method and if found, evaluate aBinaryBlock
	 with the selector and class where the method is found.  Otherwise evaluate aBinaryBlock
	 with doesNotUnderstand: and nil."
	| currClass classDict classDictSize methodArray i |
	currClass := startClass.
	[classDict := objectMemory fetchPointer: MethodDictionaryIndex ofObject: currClass.
	 classDictSize := objectMemory numSlotsOf: classDict.
	 classDictSize > MethodArrayIndex ifTrue:
		[methodArray := objectMemory fetchPointer: MethodArrayIndex ofObject: classDict.
		 i := 0.
		 [i <= (classDictSize - SelectorStart)] whileTrue:
			[meth = (objectMemory fetchPointer: i ofObject: methodArray) ifTrue:
				[^binaryBlock
					value: (objectMemory fetchPointer: i + SelectorStart ofObject: classDict)
					value: currClass].
				i := i + 1]].
	 currClass := self superclassOf: currClass.
	 currClass = objectMemory nilObject] whileFalse.
	^binaryBlock    "method not found in superclass chain"
		value: (objectMemory splObj: SelectorDoesNotUnderstand)
		value: nil! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 9/6/2016 09:56'!
findSelectorOfMethod: meth
	| classObj classDict classDictSize methodArray i |
	(objectMemory addressCouldBeObj: meth) ifFalse:
		[^objectMemory nilObject].
	(objectMemory isForwarded: meth) ifTrue:
		[^self findSelectorOfMethod: (objectMemory followForwarded: meth)].
	 (objectMemory isOopCompiledMethod: meth) ifFalse:
		[^objectMemory nilObject].
	(self maybeSelectorOfMethod: meth) ifNotNil:
		[:selector| ^selector].
	classObj := self safeMethodClassOf: meth.
	(self addressCouldBeClassObj: classObj) ifTrue:
		[classDict := objectMemory fetchPointer: MethodDictionaryIndex ofObject: classObj.
		 classDictSize := objectMemory numSlotsOf: classDict.
		 classDictSize > MethodArrayIndex ifTrue:
			[methodArray := objectMemory fetchPointer: MethodArrayIndex ofObject: classDict.
			 i := 0.
			 [i < (classDictSize - SelectorStart)] whileTrue:
				[meth = (objectMemory fetchPointer: i ofObject: methodArray) ifTrue:
					[^(objectMemory fetchPointer: i + SelectorStart ofObject: classDict)].
					 i := i + 1]]].
	^objectMemory nilObject! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem 1/26/2021 16:30'!
findUnwindThroughContext: homeContext
	"Search for either an unwind-protect (activation of method with primitive 198, PrimNumberUnwindMarker)
	 or homeContext along the sender chain, which ever is found first.  Return values:
		0			home context was found on sender chain with no intervening unwind-protects
		nilObj		home context could not be found => cannotReturn
		context		the context of an intervening unwind-protect implies home context was found"
	| onSamePage ctxtOrNilOrZero theMethod |
	"Almost always (98%) the home is on the same page, in which case we know it will be found."
	onSamePage := (self isStillMarriedContext: homeContext)
					and: [(stackPages pageIndexFor: framePointer) = (stackPages pageIndexFor: (self frameOfMarriedContext: homeContext))].

	"Since nothing changes we don't need to internalize."
	ctxtOrNilOrZero := self findMethodWithPrimitive: PrimNumberUnwindMarker  FromFP: framePointer UpToContext: homeContext.
	self deny: (onSamePage and: [ctxtOrNilOrZero = objectMemory nilObject]).

	ctxtOrNilOrZero = 0 ifTrue:
		[theMethod := objectMemory fetchPointer: MethodIndex ofObject: homeContext.
		 (self primitiveIndexOf: theMethod) = PrimNumberUnwindMarker ifTrue:
			[^homeContext].
		 ^0].

	"If an unwind was found, can the home context be found also?  No need to look if on the same page.
	 No need to look if cannot return (ctxtOrNilOrZero = objectMemory nilObject)"
	(onSamePage
	 or: [ctxtOrNilOrZero = objectMemory nilObject]) ifFalse:
		[(self findMethodWithPrimitive: 0 FromContext: ctxtOrNilOrZero UpToContext: homeContext)
		  = objectMemory nilObject ifTrue:
			[^objectMemory nilObject]].
	^ctxtOrNilOrZero! !

!StackInterpreter methodsFor: 'finalization' stamp: 'eem 5/25/2016 22:04'!
fireEphemeron: ephemeron
	<option: #SpurObjectMemory>
	objectMemory
		queueMourner: ephemeron;
		setFormatOf: ephemeron to: objectMemory nonIndexablePointerFormat.
	self signalFinalization: ephemeron! !

!StackInterpreter methodsFor: 'finalization' stamp: 'eem 8/31/2015 11:10'!
fireFinalization: weakling
	<option: #SpurObjectMemory>
	newFinalization ifTrue:
		[objectMemory queueMourner: weakling].
	self signalFinalization: weakling! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
firstByteIndexOfMethod: methodObj
	"Answer the one-relative index of the first bytecode in methodObj.
	 Used for safer bounds-checking on methods."
	^(objectMemory literalCountOf: methodObj) + LiteralStart * objectMemory bytesPerOop + 1! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 8/8/2014 14:30'!
firstBytecodeOfAlternateHeader: methodHeader method: theMethod
	^theMethod
	 + ((LiteralStart + (self literalCountOfAlternateHeader: methodHeader)) * objectMemory bytesPerOop)
	 + objectMemory baseHeaderSize! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 3/16/2018 15:32'!
floatArg: index
	"Like #stackFloatValue: but access method arguments left-to-right"
	<returnTypeC: #double>
	<doNotGenerate> "Obsolete; was never used; replaced with methodReturnString: in the VirtualMachine struct."
	| oop |
	self deprecated.
	oop := self methodArg: index.
	oop = 0 ifTrue:[^0.0]. "methodArg: failed"
	^objectMemory floatValueOf: oop! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 3/2/2016 17:14'!
flush
	<api>
	<cmacro: '() fflush(stdout)'>! !

!StackInterpreter methodsFor: 'method lookup cache' stamp: 'eem 11/25/2008 15:15'!
flushAtCache
	"Flush the at cache. The method cache is flushed on every programming change and garbage collect."

	1 to: AtCacheTotalSize do: [ :i | atCache at: i put: 0 ]! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 9/23/2014 17:20'!
flushBecommedClassesInMethodCache
	"Flush any entries in the cache which refer to a forwarded (becommed) class."
	0 to: MethodCacheSize - 1 by: MethodCacheEntrySize do:
		[:i | | c s |
		c := methodCache at: i + MethodCacheClass.
		s := methodCache at: i + MethodCacheSelector.
		(c ~= 0 and: [s ~= 0
		 and: [objectMemory isForwarded: (objectMemory classOrNilAtIndex: c)]]) ifTrue:
			[methodCache
				at: i + MethodCacheClass put: 0;
				at: i + MethodCacheSelector put: 0]].
	self flushAtCache! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 9/23/2014 12:42'!
flushBecommedClassesInMethodZone
	"This is just a stub for the CoInterpreter"! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 4/17/2017 11:22'!
flushExternalPrimitiveOf: methodObj
	"methodObj is a CompiledMethod. If it contains an external primitive,
	 flush the function address and session ID of the CM.  Answer the prim
	 index for the benefit of subclass overrides."
	<inline: false>
	| header primIdx lit |
	header := objectMemory methodHeaderOf: methodObj.
	primIdx := self primitiveIndexOfMethod: methodObj header: header.
	(primIdx = PrimNumberExternalCall
	 and: [(objectMemory literalCountOfMethodHeader: header) > 0]) ifTrue: "If not, something's broken"
		[lit := self literal: 0 ofMethod: methodObj.
		((objectMemory isArray: lit) and: [(objectMemory numSlotsOf: lit) = 4]) ifTrue: "If not, something's broken"
			[objectMemory
				storePointerUnchecked: 2 ofObject: lit withValue: ConstZero;
				storePointerUnchecked: 3 ofObject: lit withValue: ConstZero]].
	^primIdx! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 7/14/2008 12:29'!
flushExternalPrimitiveTable
	"Flush the external primitive table"
	0 to: MaxExternalPrimitiveTableSize-1 do:[:i|
		externalPrimitiveTable at: i put: 0].
	externalPrimitiveTableFirstFreeIndex := 0! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 4/27/2016 07:51'!
flushExternalPrimitives
	"Flush the references to external functions from plugin primitives.
	 This will force a reload of those primitives when accessed next. 
	 Note: We must flush the method cache here also, so that any failed
	 primitives are looked up again."
	objectMemory allObjectsDo:
		[:oop|
		(objectMemory isCompiledMethod: oop) ifTrue: "This is a compiled method"
			[self flushExternalPrimitiveOf: oop]].
	self flushMethodCache.
	self flushExternalPrimitiveTable! !

!StackInterpreter methodsFor: 'method lookup cache' stamp: 'eem 7/18/2015 08:06'!
flushMethodCache
	"Flush the method cache. The method cache is flushed on every programming change and garbage collect."

	1 to: MethodCacheSize do: [ :i | methodCache at: i put: 0 ].
	NewspeakVM ifTrue:
		[1 to: NSMethodCacheSize do: [ :i | nsMethodCache at: i put: 0 ]].
	lastMethodCacheProbeWrite := 0. "this for primitiveExternalMethod"
	self flushAtCache! !

!StackInterpreter methodsFor: 'method lookup cache' stamp: 'eem 6/21/2017 15:16'!
flushMethodCacheForMethod: oldMethod
	"Flush the method caches of references to oldMethod."
	| probe |
	probe := 0.
	1 to: MethodCacheEntries do:
		[:i |
		(methodCache at: probe + MethodCacheMethod) = oldMethod ifTrue:
			[methodCache at: probe + MethodCacheSelector put: 0].
		probe := probe + MethodCacheEntrySize].
	self flushExternalPrimitiveOf: oldMethod.
	self flushAtCache! !

!StackInterpreter methodsFor: 'method lookup cache' stamp: 'eem 7/18/2015 08:10'!
flushMethodCacheFrom: memStart to: memEnd 
	"Flush entries in the method cache only if the oop address is within the given memory range. 
	This reduces over-aggressive cache clearing. Note the AtCache is fully flushed, 70% of the time 
	cache entries live in newspace, new objects die young"
	| probe |
	probe := 0.
	1 to: MethodCacheEntries do:
		[:i | 
		(methodCache at: probe + MethodCacheSelector) = 0 ifFalse:
			[((((self oop: (methodCache at: probe + MethodCacheSelector) isGreaterThanOrEqualTo: memStart)
				 and: [self oop: (methodCache at: probe + MethodCacheSelector) isLessThan: memEnd])
			 or: [(self oop: (methodCache at: probe + MethodCacheClass) isGreaterThanOrEqualTo: memStart)
				 and: [self oop: (methodCache at: probe + MethodCacheClass) isLessThan: memEnd]])
			 or: [(self oop: (methodCache at: probe + MethodCacheMethod) isGreaterThanOrEqualTo: memStart)
				 and: [self oop: (methodCache at: probe + MethodCacheMethod) isLessThan: memEnd]]) ifTrue:
				[methodCache at: probe + MethodCacheSelector put: 0]].
			probe := probe + MethodCacheEntrySize].
	NewspeakVM ifTrue:
		[1 to: NSMethodCacheSize do: [ :i | nsMethodCache at: i put: 0]].
	self flushAtCache! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 9/6/2019 12:49'!
flushMethodsWithMachineCodePrimitivesAndContinueAnswering: result
	"In the StackInterpreter this is simply a no op"! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 7/15/2014 11:12'!
followForwardedFieldsInCurrentMethod
	<inline: true>
	objectMemory
		followForwardedObjectFields: method
		toDepth: 0! !

!StackInterpreter methodsFor: 'lazy become' stamp: 'eem 1/16/2016 06:03'!
followForwardedFrameContents: theFP stackPointer: theSP
	"follow pointers in the current stack frame up to theSP."
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	<inline: false>
	 | oop |
	<var: #ptr type: #'char *'>
	theSP
		to: (self frameReceiverLocation: theFP)
		by: objectMemory wordSize
		do: [:ptr|
			oop := stackPages longAt: ptr.
			((objectMemory isNonImmediate: oop)
			 and: [objectMemory isForwarded: oop]) ifTrue:
				[stackPages longAt: ptr put: (objectMemory followForwarded: oop)]].
	theFP + FoxCallerSavedIP + objectMemory wordSize
		to: theFP + (self frameStackedReceiverOffset: theFP)
		by: objectMemory wordSize
		do: [:ptr|
			oop := stackPages longAt: ptr.
			((objectMemory isNonImmediate: oop)
			 and: [objectMemory isForwarded: oop]) ifTrue:
				[oop := objectMemory followForwarded: oop.
				 stackPages longAt: ptr put: oop]].
	"If we're in a block then also follow fields in the stacked closure as it may be activated again."
	(self frameIsBlockActivation: theFP) ifTrue:
		[self assert: oop = (stackPages longAt: theFP + (self frameStackedReceiverOffset: theFP)).
		 objectMemory followForwardedObjectFields: oop toDepth: 0].
	self assert: (objectMemory isForwarded: (self frameMethodObject: theFP)) not.
	(self frameHasContext: theFP) ifTrue:
		[self assert: (objectMemory isForwarded: (self frameContext: theFP)) not]! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 12/6/2013 16:58'!
followForwardedMethodsInMethodCache
	0 to: MethodCacheSize - 1 by: MethodCacheEntrySize do:
		[:i | | c s m |
		c := methodCache at: i + MethodCacheClass.
		s := methodCache at: i + MethodCacheSelector.
		m := methodCache at: i + MethodCacheMethod.
		(c ~= 0 and: [s ~= 0 and: [m ~= 0
		 and: [objectMemory isOopForwarded: m]]]) ifTrue:
			[m := objectMemory followForwarded: m.
			 methodCache at: i + MethodCacheMethod put: m]]! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 1/14/2014 15:32'!
followForwardedMethodsInMethodZone
	"This is just a stub for the CoInterpreter"! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 6/29/2014 06:58'!
followForwardingPointersInProfileState
	(objectMemory isForwarded: profileProcess) ifTrue:
		[profileProcess := objectMemory followForwarded: profileProcess].
	(objectMemory isForwarded: profileMethod) ifTrue:
		[profileMethod := objectMemory followForwarded: profileMethod].
	(objectMemory isForwarded: profileProcess) ifTrue:
		[profileSemaphore := objectMemory followForwarded: profileSemaphore].! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'rmacnak 12/12/2015 17:04'!
followForwardingPointersInScheduler
	| schedAssoc sched procLists |
	schedAssoc := objectMemory splObj: SchedulerAssociation.
	"the GC follows pointers in the special objects array for us."
	self assert: (objectMemory isForwarded: schedAssoc) not.

	"Make sure the active process has been followed."
	objectMemory followForwardedObjectFields: schedAssoc toDepth: 1.

	sched := objectMemory fetchPointer: ValueIndex ofObject: schedAssoc.

	procLists := objectMemory followObjField: ProcessListsIndex ofObject: sched.

	"Follow all links in the process list to ensure the lists are valid."
	0 to: (objectMemory numSlotsOf: procLists) - 1 do:
		[:i| | list first last next |
		list := objectMemory followObjField: i ofObject: procLists.
		first := objectMemory followObjField: FirstLinkIndex ofObject: list.
		last := objectMemory followObjField: LastLinkIndex ofObject: list.
		[first ~= last] whileTrue:
			[next := objectMemory followObjField: NextLinkIndex ofObject: first.
			 first := next]]
! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 6/29/2014 07:17'!
followForwardingPointersInSpecialObjectsArray
	"Various semaphores in the specialObjectsArray are signalled in checkForEventsMayContextSwitch:.
	 These must be followed post become to avoid a read barrier in checkForEventsMayContextSwitch:,
	 or worse still in synchronousSignal."
	| xArray |
	self followSemaphoreIn: objectMemory specialObjectsOop
		at: TheLowSpaceSemaphore.
	self followSemaphoreIn: objectMemory specialObjectsOop
		at: TheInterruptSemaphore.
	self followSemaphoreIn: objectMemory specialObjectsOop
		at: TheTimerSemaphore.
	self followSemaphoreIn: objectMemory specialObjectsOop
		at: TheFinalizationSemaphore.
	xArray := objectMemory splObj: ExternalObjectsArray.
	(objectMemory isForwarded: xArray) ifTrue:
		[xArray := objectMemory followForwarded: xArray.
		 objectMemory splObj: ExternalObjectsArray put: xArray].
	0 to: (objectMemory numSlotsOf: xArray) - 1 do:
		[:i|
		self followSemaphoreIn: xArray at: i]! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 11/15/2020 23:02'!
followForwardingPointersInStackZone: theBecomeEffectsFlags
	"Spur's become: is lazy, turning the becommed object into a forwarding object to the other.
	 The read-barrier is minimised by arranging that forwarding pointers will fail a method cache
	 probe, since notionally objects' internals are accessed only via sending messages to them,
	 the exception is primitives that access the internals of the non-receiver argument(s).

	 To avoid a read barrier on bytecode, literal and inst var fetch and non-local return, we scan
	 the receivers (including the stacked receiver for non-local return) and method references
	 in the stack zone and follow any forwarded ones.  This is of course way cheaper than
	 scanning all of memory as in the old become."
	| theIPPtr |
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #theIPPtr type: #usqInt>
	<var: #callerFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>

	stackPage = 0 ifTrue: "the system must be snapshotting; nothing to do..."
		[self assert: (stackPages mostRecentlyUsedPage isNil or: [stackPages mostRecentlyUsedPage isFree]).
		 self cCode: [] inSmalltalk: [self assert: stackPages allPagesFree].
		 ^self].

	self externalWriteBackHeadFramePointers.

	(theBecomeEffectsFlags anyMask: BecameCompiledMethodFlag) ifTrue:
		[(objectMemory isForwarded: method) ifTrue:
			[theIPPtr := instructionPointer - method.
			 method := objectMemory followForwarded: method.
			 instructionPointer := method + theIPPtr].
		(objectMemory isOopForwarded: newMethod) ifTrue:
			[newMethod := objectMemory followForwarded: newMethod]].

	self assert: stackPage ~= 0.
	0 to: numStackPages - 1 do:
		[:i| | thePage theFP callerFP offset oop |
		thePage := stackPages stackPageAt: i.
		thePage isFree ifFalse:
			[self assert: (self ifCurrentStackPageHasValidHeadPointers: thePage).
			 theFP := thePage  headFP.
			 "Skip the instruction pointer on top of stack of inactive pages."
			 theIPPtr := thePage = stackPage ifTrue: [0] ifFalse: [thePage headSP asUnsignedInteger].
			 [self assert: (thePage addressIsInPage: theFP).
			  self assert: (theIPPtr = 0 or: [thePage addressIsInPage: theIPPtr asVoidPointer]).
			  oop := stackPages longAt: theFP + FoxReceiver.
			  (objectMemory isOopForwarded: oop) ifTrue:
				[stackPages
					longAt: theFP + FoxReceiver
					put: (objectMemory followForwarded: oop)].
			  ((self frameHasContext: theFP)
			   and: [(objectMemory isForwarded: (self frameContext: theFP))]) ifTrue:
				[stackPages
					longAt: theFP + FoxThisContext
					put: (objectMemory followForwarded: (self frameContext: theFP))].
			  oop := self frameMethod: theFP.
			  (objectMemory isForwarded: oop) ifTrue:
				[| newOop delta |
				 newOop := objectMemory followForwarded: oop.
				 theIPPtr ~= 0 ifTrue:
					[self assert: (stackPages longAt: theIPPtr) > (self frameMethod: theFP).
					 delta := newOop - oop.
					 stackPages
						longAt: theIPPtr
						put: (stackPages longAt: theIPPtr) + delta].
				stackPages
					longAt: theFP + FoxMethod
					put: (oop := newOop)].
			  offset := self frameStackedReceiverOffset: theFP.
			  oop := stackPages longAt: theFP + offset.
			  (objectMemory isOopForwarded: oop) ifTrue:
				[stackPages
					longAt: theFP + offset
					put: (objectMemory followForwarded: oop)].
			  (callerFP := self frameCallerFP: theFP) ~= 0] whileTrue:
				[theIPPtr := (theFP + FoxCallerSavedIP) asUnsignedInteger.
				 theFP := callerFP].
			 "And finally follow the caller context."
			 self assert: theFP = thePage baseFP.
			 oop := self frameCallerContext: theFP.
			 (objectMemory isForwarded: oop) ifTrue:
				[self frameCallerContext: theFP put: (objectMemory followForwarded: oop)]]]! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 1/20/2015 15:14'!
followLiteral: offset ofMethod: methodPointer
	^objectMemory followField: offset + LiteralStart ofObject: methodPointer
! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 6/29/2014 07:38'!
followSemaphoreIn: anArray at: index
	"Only follow the reference to the semaphore.  synchronousSignal
	 checks the chain to the suspendedContext."
	 
	| obj |
	obj := objectMemory fetchPointer: index ofObject: anArray.
	(objectMemory isForwarded: obj) ifTrue:
		[obj := objectMemory followForwarded: obj.
		 objectMemory storePointer: index ofObject: anArray withValue: obj]! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 12/15/2015 09:43'!
forceInterruptCheck
	"Force an interrupt check ASAP.
	 Must set the stack page's limit before stackLimit to avoid
	 a race condition if this is called from an interrupt handler."
	| thePage iccFunc |
	<inline: false>
	<var: #thePage type: #'StackPage *'>
	<var: #iccFunc declareC: 'void (*iccFunc)()'>
	"Do _not_ set stackLimit until the stack system has been initialized.
	 stackLimit is the initialization flag for the stack system."
	stackLimit = 0 ifTrue:
		[^nil].
	thePage := stackPage.
	(thePage notNil and: [thePage ~= 0]) ifTrue:
		[thePage stackLimit: self allOnesAsCharStar].
	stackLimit := self allOnesAsCharStar.
	self sqLowLevelMFence.
	"There is a race condition if we test the function and then dereference
	 it a second time to call it.  This is called from interrupt code but at the
	 same time other code could be clearing the interruptCheckChain via
	 setInterruptCheckChain:."
	(iccFunc := interruptCheckChain) notNil ifTrue:
		[self perform: iccFunc].
	statForceInterruptCheck := statForceInterruptCheck + 1! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 8/4/2021 11:13'!
forceInterruptCheckFromHeartbeat
	"Force an interrupt check ASAP. This version is the
	 entry-point to forceInterruptCheck for the heartbeat
	 timer to allow for repeatable debugging."
	suppressHeartbeatFlag ifFalse:
		[self forceInterruptCheck]! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 2/3/2009 11:36'!
frameCallerContext: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	^stackPages longAt: theFP + FoxCallerContext "a.k.a. FoxCallerSavedIP"! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 1/25/2018 17:26'!
frameCallerContext: theFP put: aValue
	<var: #theFP type: #'char *'>
	<inline: true>
	self assert: (aValue = objectMemory nilObject or: [objectMemory isContext: aValue]).
	^stackPages
		longAt: theFP + FoxCallerContext "a.k.a. FoxCallerSavedIP"
		put: aValue! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 11/14/2020 15:22'!
frameCallerFP: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	<returnTypeC: #'char *'>
	^self pointerForOop: (stackPages longAt: theFP + FoxSavedFP)! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 11/10/2014 12:49'!
frameCallerSP: theFP
	"Answer the SP of the caller provided theFP is not a base frame.
	 This points to the hottest item on the caller frame's stack."
	<var: #theFP type: #'char *'>
	<returnTypeC: 'char *'>
	self assert: (self isBaseFrame: theFP) not.
	^theFP + (self frameStackedReceiverOffset: theFP) + objectMemory wordSize! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 1/2/2011 14:09'!
frameCallerSavedIP: theFP
	<inline: true>
	<returnTypeC: #'char *'>
	<var: #theFP type: #'char *'>
	^self pointerForOop: (stackPages longAt: theFP + FoxCallerSavedIP) "a.k.a. FoxCallerSavedIP"! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 11/10/2014 12:49'!
frameCallerStackPointer: theFP
	"Answer the stack pointer of the caller frame."
	<var: #theFP type: #'char *'>
	<returnTypeC: #'char *'>
	<inline: true>
	self assert: (self isBaseFrame: theFP) not.
	^theFP + (self frameStackedReceiverOffset: theFP) + objectMemory wordSize! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (auto pragmas 12/08) 7/2/2008 20:57'!
frameContext: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	^stackPages longAt: theFP + FoxThisContext! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (auto pragmas 12/08) 7/4/2008 08:32'!
frameHasContext: theFP
	"See encodeFrameFieldHasContext:numArgs:"
	<inline: true>
	<var: #theFP type: #'char *'>
	^(stackPages byteAt: theFP + FoxFrameFlags + 2) ~= 0! !

!StackInterpreter methodsFor: 'frame access' stamp: 'rsf 8/9/2016 23:36'!
frameHasNativeFrame: theFP
	"See encodeFrameFieldHasContext:numArgs:"
	<inline: true>
	<option: #LowcodeVM>
	<var: #theFP type: #'char *'>
	| markObject |
	markObject := self temporary: (self frameNumArgs: theFP) in: theFP.
	
	^ markObject ~= objectMemory nilObject and: [ markObject = (objectMemory splObj: LowcodeContextMark)]! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (auto pragmas 12/08) 7/4/2008 08:33'!
frameIsBlockActivation: theFP "<Integer>"
	<inline: true>
	<var: #theFP type: #'char *'>
	^(stackPages byteAt: theFP + FoxFrameFlags + 3) ~= 0! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 12/9/2016 10:50'!
frameLastArgumentOffset
	"Answer the offset in bytes from the frame pointer to the last argument,
	 which is simply the address of the word above the saved ip.  See the
	 diagram in StackInterpreter class>>initializeFrameIndices."
	<inline: true>
	^FoxCallerSavedIP + objectMemory wordSize! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (auto pragmas 12/08) 7/2/2008 20:47'!
frameMethod: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	^stackPages longAt: theFP + FoxMethod! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 1/27/2009 18:37'!
frameMethodObject: theFP
	"Homonym of frameMethod: for compatibility with CoInterpreter"
	<inline: true>
	<var: #theFP type: #'char *'>
	^stackPages longAt: theFP + FoxMethod! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (auto pragmas 12/08) 7/4/2008 08:32'!
frameNumArgs: theFP
	"See encodeFrameFieldHasContext:numArgs:"
	<inline: true>
	<var: #theFP type: #'char *'>
	^stackPages byteAt: theFP + FoxFrameFlags + 1! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 7/3/2014 06:29'!
frameOfMarriedContext: aContext 
	| senderOop |
	<inline: true>
	<returnTypeC: 'char *'>
	senderOop := objectMemory fetchPointer: SenderIndex ofObject: aContext.
	^self withoutSmallIntegerTags: senderOop! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 9/5/2016 16:18'!
framePointerOfNativeArgument: offset in: theFP
	<inline: false>
	<option: #LowcodeVM>
	<returnTypeC: 'char*'>
	<var: #theFP type: #'char*'>
	^ (self nativePreviousStackPointerIn: theFP) + offset - 1! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:38'!
framePointerOfNativeLocal: offset in: theFP
	<inline: false>
	<option: #LowcodeVM>
	<returnTypeC: 'char*'>
	<var: #theFP type: #'char*'>
	^ (self nativeFramePointerIn: theFP) + offset - 1! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (auto pragmas 12/08) 7/2/2008 20:57'!
frameReceiver: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	^stackPages longAt: theFP + FoxReceiver! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 4/2/2015 10:56'!
frameReceiverLocation: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	^theFP + FoxReceiver! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 3/3/2012 11:35'!
frameStackedReceiver: theFP numArgs: numArgs
	"Answer the stacked receiver given the frame's argument count.
	 The receiver of a message send or the closure of a block activation is
	 always on the stack above any arguments and the frame itself.  See the
	 diagram in StackInterpreter class>>initializeFrameIndices."
	<inline: true>
	<var: #theFP type: #'char *'>
	^stackPages longAt: theFP + (self frameStackedReceiverOffsetNumArgs: numArgs)! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 3/3/2012 11:33'!
frameStackedReceiverOffset: theFP
	"Answer the offset in bytes from the the frame pointer to its stacked receiver.
	 The receiver of a message send or the closure of a block activation is
	 always on the stack above any arguments and the frame itself.  See the
	 diagram in StackInterpreter class>>initializeFrameIndices."
	<inline: true>
	<var: #theFP type: #'char *'>
	^self frameStackedReceiverOffsetNumArgs: (self frameNumArgs: theFP)! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:03'!
frameStackedReceiverOffsetNumArgs: numArgs
	"Answer the offset in bytes from the a frame pointer to its stacked receiver,
	 given the argument count.  The receiver of a message send or the closure of
	 a block activation is always on the stack above any arguments and the frame
	 itself.  See the diagram in StackInterpreter class>>initializeFrameIndices."
	<inline: true>
	^FoxCallerSavedIP + objectMemory wordSize + (numArgs << objectMemory shiftForWord)! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 3/13/2014 17:17'!
freeUnmarkedMachineCode
	"This is a no-op in the StackInterpreter"! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 11/7/2013 11:48'!
freeUntracedStackPages
	"Free any untraced stack pages."
	<var: #thePage type: #'StackPage *'>
	<inline: false>

	0 to: numStackPages - 1 do:
		[:i| | thePage |
		thePage := stackPages stackPageAt: i.
		((stackPages isFree: thePage) not
		 and: [thePage trace = StackPageUnreached]) ifTrue:
			[self assert: (self noMarkedContextsOnPage: thePage).
			 stackPages freeStackPage: thePage].
		self assert: (thePage trace: StackPageTraceInvalid) ~= 0] "Invalidate the trace state for assertion checks"! !

!StackInterpreter methodsFor: 'control primitives' stamp: 'cb 4/11/2016 10:32'!
fullClosureIn: context numArgs: numArgs numCopiedValues: numCopied compiledBlock: compiledBlock 
	| newClosure |
	<inline: true>
	ClassFullBlockClosureCompactIndex ~= 0
		ifTrue:
			[newClosure := objectMemory
								eeInstantiateSmallClassIndex: ClassFullBlockClosureCompactIndex
								format: objectMemory indexablePointersFormat
								numSlots: FullClosureFirstCopiedValueIndex + numCopied]
		ifFalse:
			[newClosure := objectMemory
								eeInstantiateSmallClass: (objectMemory splObj: ClassFullBlockClosure)
								numSlots: FullClosureFirstCopiedValueIndex + numCopied].
	"Assume: have just allocated a new closure; it must be young. Thus, can use unchecked stores."
	objectMemory
		storePointerUnchecked: ClosureOuterContextIndex ofObject: newClosure withValue: context;
		storePointerUnchecked: FullClosureCompiledBlockIndex ofObject: newClosure withValue: compiledBlock;
		storePointerUnchecked: ClosureNumArgsIndex ofObject: newClosure withValue: (objectMemory integerObjectOf: numArgs).
	^newClosure! !

!StackInterpreter methodsFor: 'I/O primitive support' stamp: 'eem 5/12/2017 17:45'!
fullDisplayUpdate
	"Repaint the entire smalltalk screen, ignoring the affected rectangle. Used in some platform's code when the Smalltalk window is brought to the front or uncovered."
	<returnTypeC: #sqInt>
	self updateDisplayLeft: 0 Top: 0 Right: displayWidth Bottom: displayHeight.
	self ioForceDisplayUpdate! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 12/11/2010 07:39'!
functionForPrimitiveCallout
	"Answer the function primitiveCallout from the FFI plugin or nil if it can't
	 be found.  Cache it for performance.  We use this circumlocution so that
	 Squeak can be deployed without the FFI plugin for security reasons."

	<returnTypeC: 'void (*functionForPrimitiveCallout())(void)'>
	| function |
	<var: #function declareC: 'static void *function = (void *)-1'>
	self cCode: '' inSmalltalk: [function := -1].
	function asInteger = -1 ifTrue:
		[function := self ioLoadFunction: 'primitiveCallout' From: 'SqueakFFIPrims'].
	^self cCoerceSimple: function to: #'void (*)(void)'! !

!StackInterpreter methodsFor: 'method lookup cache' stamp: 'eem 8/21/2013 12:54'!
functionPointerFor: primIdx inClass: theClass
	"Find an actual function pointer for this primitiveIndex.  This is an
	opportunity to specialise the prim for the relevant class (format for
	example).  Default for now is simply the entry in the base primitiveTable."
	<api>
	<returnTypeC: 'void (*functionPointerForinClass(sqInt primIdx,sqInt theClass))(void)'>
	^primIdx > MaxPrimitiveIndex ifTrue: [0] ifFalse: [primitiveTable at: primIdx]! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
getCodeCompactionCount
	"This is nil in the StackVM"
	^objectMemory nilObject! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
getCodeCompactionMSecs
	"This is nil in the StackVM"
	^objectMemory nilObject! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
getCogCodeSize
	"This is nil in the StackVM"
	^objectMemory nilObject! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 7/19/2010 09:45'!
getCogMethodCount
	"This is 0 in the StackVM"
	^ConstZero! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 4/23/2020 12:46'!
getCogVMFeatureFlags
	"Answer an array of flags indicating various optional features of the Cog VM.
	 Bit 0: supports two bytecode sets (MULTIPLEBYTECODESETS)
	 Bit 1: supports immutablity (IMMUTABILITY)
	 Bit 2: suffers from a UNIX setitimer signal-based heartbeat
	 Bit 3: the VM provides cross-platform bit-identical floating point"
	^objectMemory integerObjectOf: (MULTIPLEBYTECODESETS ifTrue: [1] ifFalse: [0])
									+ (IMMUTABILITY ifTrue: [2] ifFalse: [0])
									+ (self cppIf: #'ITIMER_HEARTBEAT' ifTrue: [4] ifFalse: [0])
									+ (self cppIf: #'BIT_IDENTICAL_FLOATING_POINT' ifTrue: [8] ifFalse: [0])! !

!StackInterpreter methodsFor: 'interpreter shell' stamp: 'eem 6/25/2011 12:32'!
getCurrentBytecode
	"currentBytecode will be private to the main dispatch loop in the generated code. This method allows the currentBytecode to be retrieved from global variables."

	^objectMemory byteAt: instructionPointer! !

!StackInterpreter methodsFor: 'simulation support' stamp: 'eem 3/14/2014 16:44'!
getDeferDisplayUpdates
	<doNotGenerate>
	^deferDisplayUpdates! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
getDesiredCogCodeSize
	"This is nil in the StackVM"
	^objectMemory nilObject! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 8/22/2018 12:33'!
getErrorObjectFromPrimFailCode
	"Answer the errorCode object to supply to a failing primitive method that accepts one.
	 If there is a primitive error table and the primFailCode is a valid index there-in answer
	 the corresponding entry in the table, otherwise simply answer the code as an integer."
	| table errObj numSlots |
	primFailCode > 0 ifTrue:
		[table := objectMemory splObj: PrimErrTableIndex.
		 primFailCode <= (objectMemory numSlotsOf: table) ifTrue:
			[errObj := objectMemory followField: primFailCode - 1 ofObject: table.
			 "If this is a PrimErrOSError/PrimErrFFIException and there's a clonable object in the table at that index,
			 answer a clone of the error object with the second slot set to the value of osErrorCode."
			 ((primFailCode = PrimErrOSError or: [primFailCode = PrimErrFFIException])
			  and: [(objectMemory formatOf: errObj) = objectMemory nonIndexablePointerFormat
			  and: [(numSlots := objectMemory numSlotsOf: errObj) >= 2]]) ifTrue:
				[errObj := self cloneOSErrorObj: errObj numSlots: numSlots].
			 ^errObj]].
	^objectMemory integerObjectOf: primFailCode! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/17/2002 12:02'!
getFullScreenFlag
	^fullScreenFlag! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 5/7/2014 08:33'!
getGCMode
	"This is a no-op in the StackVM"
	^0! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 10/18/2020 17:22'!
getImageHeaderFlags
	"Answer the flags that are contained in the 7th long of the image header."
	^fullScreenFlag "0 or 1"
	+ (VMBIGENDIAN ifTrue: [0] ifFalse: [2]) "this is the imageFloatsLittleEndian flag"
	+ (preemptionYields ifTrue: [0] ifFalse: [16r10])
	+ (newFinalization ifTrue: [16r40] ifFalse: [0])
	+ (sendWheelEvents ifTrue: [16r80] ifFalse: [0])
	+ (primitiveDoMixedArithmetic ifTrue: [0] ifFalse: [16r100])
	+ (imageHeaderFlags bitClear: 1+2+16r10+16r40+16r80+16r100) "these are any flags we do not recognize"! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 10/18/2020 17:27'!
getImageHeaderFlagsParameter
	"Answer the flags that are contained in the 7th long of the image header
	 as reported to the image through the vm parameter primitives # 48."
	^objectMemory integerObjectOf: self getImageHeaderFlags >> 2! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/17/2002 12:02'!
getInterruptKeycode
	^interruptKeycode! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/17/2002 12:02'!
getInterruptPending
	^interruptPending! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 12/5/2014 12:04'!
getLongFromFile: aFile swap: swapFlag
	"Answer the next 32 or 64 bit word read from aFile, byte-swapped according to the swapFlag."
	<var: #aFile type: #sqImageFile>
	| w |
	w := 0.
	self cCode: [self
					sq: (self addressOf: w)
					Image: (self sizeof: w)
					File: 1
					Read: aFile]
		inSmalltalk: [w := objectMemory nextLongFrom: aFile].
	^swapFlag
		ifTrue: [objectMemory byteSwapped: w]
		ifFalse: [w]! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 8/21/2009 12:36'!
getNextWakeupUsecs
	<api>
	<returnTypeC: #usqLong>
	^nextWakeupUsecs! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/17/2002 12:03'!
getSavedWindowSize
	^savedWindowSize! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'nice 4/3/2016 00:40'!
getShortFromFile: aFile swap: swapFlag
	"Answer the next 16 bit word read from aFile, byte-swapped according to the swapFlag."

	<var: #aFile type: #sqImageFile>
	| w |
	<var: #w type: #'unsigned short'>
	w := 0.
	self cCode: [self
					sq: (self addressOf: w)
					Image: (self sizeof: #'unsigned short')
					File: 1
					Read: aFile]
		inSmalltalk: [w := objectMemory nextShortFrom: aFile].
	^swapFlag
		ifTrue: [((w >> 8) bitAnd: 16rFF) bitOr: ((w bitAnd: 16rFF) << 8)]
		ifFalse: [w]! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 8/22/2017 17:47'!
getSnapshotScreenSize
	"Answer the screen size to write to the snapshot.
	 If the actual screen size is zero then write the savedWindowSize instead."
	| screenSize |
	screenSize := self ioScreenSize.
	((screenSize bitAnd: 16rFFFF) = 0
	 or: [screenSize >> 16 = 0]) ifTrue:
		[screenSize := savedWindowSize].
	^screenSize! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 10/26/2018 20:43'!
getStackPointer
	"For Alien FFI"
	<api>
	<returnTypeC: #'sqInt *'>
	^self cCoerceSimple: stackPointer to: #'sqInt *'! !

!StackInterpreter methodsFor: 'plugin support' stamp: 'tpr (auto pragmas 12/08) 6/17/2005 17:52'!
getThisSessionID
	"return the global session ID value"
	<inline: false>
	^globalSessionID! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'nice 4/3/2016 00:10'!
getWord32FromFile: aFile swap: swapFlag
	"Answer the next 32 bit word read from aFile, byte-swapped according to the swapFlag."

	<var: #aFile type: #sqImageFile>
	| w |
	<var: #w type: #int>
	w := 0.
	self cCode: [self
					sq: (self addressOf: w)
					Image: (self sizeof: #int)
					File: 1
					Read: aFile]
		inSmalltalk: [w := objectMemory nextWord32From: aFile].
	^swapFlag
		ifTrue: [w byteSwap32]
		ifFalse: [w]! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 12/7/2013 11:57'!
handleForwardedSelectorFaultFor: selectorOop
	"Handle a send fault that is due to a send using a forwarded selector.
	 Unforward the selector and follow the current method and special
	 selectors array to unforward the source of the forwarded selector."
	<option: #SpurObjectMemory>
	<inline: false>
	self assert: (objectMemory isOopForwarded: selectorOop).
	self followForwardedFieldsInCurrentMethod.
	objectMemory
		followForwardedObjectFields: (objectMemory splObj: SpecialSelectors)
		toDepth: 0.
	^objectMemory followForwarded: selectorOop! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
handleForwardedSendFaultForTag: classTag
	"Handle a send fault that may be due to a send to a forwarded object.
	 Unforward the receiver on the stack and answer its actual class."
	<option: #SpurObjectMemory>
	| rcvr |
	<inline: false>
	self assert: (objectMemory isForwardedClassTag: classTag).

	rcvr := self stackValue: argumentCount.
	"should *not* be a super send, so the receiver should be forwarded."
	self assert: (objectMemory isOopForwarded: rcvr).
	rcvr := objectMemory followForwarded: rcvr.
	self stackValue: argumentCount put: rcvr.
	self followForwardedFrameContents: framePointer
		stackPointer: stackPointer + (argumentCount + 1 * objectMemory wordSize). "don't repeat effort"
	(objectMemory isPointers: (self frameReceiver: framePointer)) ifTrue:
		[objectMemory
			followForwardedObjectFields: (self frameReceiver: framePointer)
			toDepth: 0].
	^objectMemory fetchClassTagOf: rcvr! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 12/10/2013 17:16'!
handleSpecialSelectorSendFaultFor: obj
	<inline: true>
	<option: #SpurObjectMemory>
	^self handleSpecialSelectorSendFaultFor: obj fp: localFP sp: localSP! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 12/3/2014 15:12'!
handleSpecialSelectorSendFaultFor: obj fp: theFP sp: theSP
	"Handle a special send fault that may be due to a special selector
	 send accessing a forwarded object.  obj is forwarded.
	 Unforward stack contents and and inst vars and answer obj's target."
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	| rcvr |
	self assert: (objectMemory isOopForwarded: obj).
	self followForwardedFrameContents: theFP stackPointer: theSP.
	rcvr := self frameReceiver: theFP.
	(objectMemory isPointers: rcvr) ifTrue:
		[objectMemory
			followForwardedObjectFields: rcvr
			toDepth: 0].
	^objectMemory followForwarded: obj! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 3/10/2017 13:00'!
handleStackOverflow
	"Check for stack overflow, moving frames to another stack if so.
	 This should *only* be sent from checkForStackOverflow."
	<inline: #never>
	| newPage theFP callerFP overflowLimitAddress overflowCount |
	<var: #newPage type: #'StackPage *'>
	<var: #theFP type: #'char *'>
	<var: #callerFP type: #'char *'>
	<var: #overflowLimitAddress type: #'char *'>

	self assert: stackPointer < stackPage realStackLimit.

	self maybeTraceStackOverflow.
	statStackOverflow := statStackOverflow + 1.

	"The stack has overflowed this page.  If the system is executing some recursive algorithm,
	 e.g. fibonacci, then the system could thrash overflowing the stack if the call soon returns
	 back to the current page.  To avoid thrashing, since overflow is quite slow, we can move
	 more than one frame.  The idea is to record which page has overflowed, and the first
	 time it overflows move one frame, the second time two frames, and so on.  We move no
	 more frames than would leave the page half occupied."
	theFP := framePointer.
	stackPage = overflowedPage
		ifTrue:
			[overflowLimitAddress := stackPage baseAddress - stackPages overflowLimit.
			 overflowCount := extraFramesToMoveOnOverflow := extraFramesToMoveOnOverflow + 1.
			 [(overflowCount := overflowCount - 1) >= 0
			   and: [(callerFP := self frameCallerFP: theFP) < overflowLimitAddress
			   and: [(self isBaseFrame: callerFP) not]]] whileTrue:
				[theFP := callerFP]]
		ifFalse:
			[overflowedPage := stackPage.
			 extraFramesToMoveOnOverflow := 0].

	self ensureCallerContext: theFP.
	newPage := stackPages newStackPage.
	self moveFramesIn: stackPage through: theFP toPage: newPage.
	self setStackPageAndLimit: newPage.
	self setStackPointersFromPage: newPage.
	self isCog
		ifFalse: "To overflow the stack this must be a new frame, but in Cog base frames are married."
			[self assert: (self frameHasContext: framePointer) not.
			 self assert: (self validInstructionPointer: instructionPointer + 1
							inMethod: method
							framePointer: framePointer)]
		ifTrue:
			[self assert: (self validInstructionPointer: instructionPointer + 1
							inFrame: framePointer).
			 self assert: ((self frameHasContext: framePointer) not
						or: [objectMemory isContext: (self frameContext: framePointer)])]! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 2/17/2017 16:26'!
handleStackOverflowOrEventAllowContextSwitch: mayContextSwitch
	"The stackPointer is below the stackLimit.  This is either because of a
	 stack overflow or the setting of stackLimit to indicate a possible interrupt.
	 Check for stackOverflow and  interrupts anddeal with each appropriately.
	 Answer if a context switch occurred."
	<inline: false>
	self checkForStackOverflow.
	"If the stackLimit differs from the realStackLimit then the stackLimit
	 has been set to indicate an event or interrupt that needs servicing."
	stackLimit = stackPage realStackLimit ifTrue:
		[^false].
	^self checkForEventsMayContextSwitch: mayContextSwitch! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
highBit: anUnsignedValue 
	"This is a C implementation needed by ioSetMaxExtSemTableSize
	 and e.g. stackPageByteSize."
	| shifted bitNo |
	<api>
	<highBit> "so it shows up in senders..."
	<var: #anUnsignedValue type: #usqInt>
	<var: #shifted type: #usqInt>
	shifted := anUnsignedValue.
	bitNo := 0.
	self cppIf: objectMemory wordSize > 4
		ifTrue:
			[shifted < (1 << 32) ifFalse:
				[shifted := shifted >> 32.
				 bitNo := bitNo + 32]].
	shifted < (1 << 16) ifFalse:
		[shifted := shifted >> 16.
		 bitNo := bitNo + 16].
	shifted < (1 << 8) ifFalse:
		[shifted := shifted >> 8.
		 bitNo := bitNo + 8].
	shifted < (1 << 4) ifFalse:
		[shifted := shifted >> 4.
		 bitNo := bitNo + 4].
	shifted < (1 << 2) ifFalse:
		[shifted := shifted >> 2.
		 bitNo := bitNo + 2].
	shifted < (1 << 1) ifFalse:
		[shifted := shifted >> 1.
		 bitNo := bitNo + 1].
	"shifted 0 or 1 now"
	^bitNo + shifted! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 07/28/2014 07:54'!
highestPriorityProcess
	"Answer the highest priority process that is ready to run, but
	 unlike wakeHighestPriority do not remove it from the list.
	 To save time looking at many empty lists before finding a
	 runnable process the VM maintains a variable holding the
	 highest priority runnable process.  If this variable is 0 then the
	 VM does not know the highest priority and must search all lists.
	 Note: It is a fatal VM error if there is no runnable process."
	| schedLists p processList processOrNil |
	<inline: false>
	schedLists := objectMemory fetchPointer: ProcessListsIndex ofObject: self schedulerPointer.
	p := highestRunnableProcessPriority = 0
			ifTrue: [objectMemory numSlotsOf: schedLists]
			ifFalse: [highestRunnableProcessPriority].
	p := p - 1.
	"index of last indexable field"
	[processList := objectMemory fetchPointer: p ofObject: schedLists.
	 processOrNil := objectMemory fetchPointer: FirstLinkIndex ofObject: processList.
	 processOrNil = objectMemory nilObject] whileTrue:
		[(p := p - 1) < 0 ifTrue:
			[^nil]].
	highestRunnableProcessPriority := p + 1.
	^processOrNil! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 1/30/2011 17:10'!
ifAppropriateCompileToNativeCode: aMethodObj selector: selector
	"Convenient  noop in the StackInterpreter e.g. for sendInvokeCallbackContext:"! !

!StackInterpreter methodsFor: 'jump bytecodes' stamp: 'eem 11/15/2012 11:46'!
ifBackwardsCheckForEvents: offset
	"Backward jump means we're in a loop; check for possible interrupts."
	<inline: true>
	(offset < 0
	 and: [localSP < stackLimit]) ifTrue:
		[self externalizeIPandSP.
		 self checkForEventsMayContextSwitch: true.
		 self browserPluginReturnIfNeeded.
		 self internalizeIPandSP]! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 5/6/2015 17:04'!
ifCurrentStackPageHasValidHeadPointers: thePage
	"If thePage is the stackPage and the stackPointer and/or the framePointer are pointing within it,
	 answer if thePage's heapSP and headFP are equal to the stackPointer and framePointer respectively."
	<var: #thePage type: #'StackPage *'>
	thePage = stackPage ifTrue:
		[(thePage addressIsInPage: framePointer) ifTrue:
			[thePage headFP ~= framePointer ifTrue:
				[^false]].
		(thePage addressIsInPage: stackPointer) ifTrue:
			[thePage headSP ~= stackPointer ifTrue:
				[^false]]].
	^true! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 12/8/2016 10:45'!
ifSoAssertValidIPAssign: index ofContext: maybeMarriedContext with: anOop
	self assert: (index ~= InstructionPointerIndex
				or: [(objectMemory isContextNonImm: maybeMarriedContext) not
				or: [| mo |
					mo := self fetchPointer: MethodIndex ofObject: maybeMarriedContext.
					(anOop = objectMemory nilObject
					 or: [(objectMemory isIntegerObject: anOop)
						  and: [(objectMemory integerValueOf: anOop)
								between: (LiteralStart + (objectMemory literalCountOf: mo)) * objectMemory bytesPerOop
								and: (objectMemory numBytesOf: mo)]])]])! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 2/22/2011 15:17'!
ifValidWriteBackStack: theCFP Pointers: theCSP Save: savedFPP To: savedSPP
	"This is for low-level error reporting.  If either of the C stack pointers are
	 pointing into the stack zone then write them back to framePointer and/or
	 stackPointer so that the stack backtrace will be up to date.  Write their
	 original values through savedFPP & savedSPP if non-null.
	 This is a noop in the stack VM since the C stack pointers are always
	 elsewhere (e.g., in some C function running the interpreter)."
	<api>
	<var: #theCFP type: #'void *'>
	<var: #theCSP type: #'void *'>
	<var: #savedFPP type: #'char **'>
	<var: #savedSPP type: #'char **'>
	<returnTypeC: #void>! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
iframeInstructionPointerForIndex: ip method: aMethod
	"Answer the instruction pointer for use in an interpreter frame (a pointer to a bytecode)."
	self assert: (ip between: (((LiteralStart + (objectMemory literalCountOf: aMethod)) * objectMemory bytesPerOop)) + 1
					and: (objectMemory lengthOf: aMethod)).
	^aMethod + ip + objectMemory baseHeaderSize - 2! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 12/31/2014 15:25'!
iframeIsBlockActivation: theFP "<Integer>"
	^self frameIsBlockActivation: theFP! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 10/26/2012 16:38'!
iframeMethod: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	<returnTypeC: #usqInt>
	^stackPages longAt: theFP + FoxMethod! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 12/8/2017 12:50'!
imageFormatCompatibilityVersion
	"This VM is backward-compatible with the immediately preceding version."

	^objectMemory wordSize = 4 ifTrue: [6504] ifFalse: [68002]! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 3/10/2016 12:46'!
imageFormatVersion
	"Return a magic constant that changes when the image format changes.
	 Since the image reading code uses this to detect byte ordering, one
	 must avoid version numbers that are invariant under byte reversal."
	<doNotGenerate>
	self assert: (objectMemory imageFormatVersion anyMask: 16) = objectMemory hasSpurMemoryManagerAPI.
	^objectMemory imageFormatVersion! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 3/29/2011 19:07'!
includesBehavior: aClass ThatOf: aSuperclass
	"Return the equivalent of 
		aClass includesBehavior: aSuperclass.
	Note: written for efficiency and better inlining (only 1 temp)"
	| theClass |
	<inline: true>
	aSuperclass = objectMemory nilObject ifTrue:
		[^false].
	theClass := aClass.
	[theClass = aSuperclass ifTrue:
		[^true].
	 theClass ~= objectMemory nilObject] whileTrue:
		[theClass := self superclassOf: theClass].
	^false! !

!StackInterpreter methodsFor: 'simulation support' stamp: 'rsf 1/5/2017 15:44'!
initExtensions
	<inline: true>
	BytecodeSetHasExtensions ifTrue: [extA := numExtB := extB := 0]! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 11/7/2013 12:31'!
initStackPageGC
	"GC of pages.  Throwing away all stack pages on full GC is simple but dangerous
	 because it causes us to allocate lots of contexts immediately before a GC.
	 Reclaiming pages whose top context is not referenced is poor because it would
	 take N incrementalGCs to reclaim N unused pages.  Only the page whose top
	 context is not referred to by the bottom context of any other page would be
	 reclaimed.  Not until the next GC would the page whose top contect is the
	 previously reclaimed page's base frame's bottom context be reclaimed.

	 Better is to not mark stack pages until their contexts are encountered.  We can
	 eagerly trace the active page and the page reachable from its bottom context
	 if any, and so on.  Other pages can be marked when we encounter a married
	 context."
	<var: #thePage type: #'StackPage *'>
	<inline: true>
	"Need to write back the frame pointers unless all pages are free (as in snapshot)"
	stackPage ~= 0 ifTrue:
		[self externalWriteBackHeadFramePointers].

	0 to: numStackPages - 1 do:
		[:i| | thePage |
		thePage := stackPages stackPageAt: i.
		thePage trace: StackPageUnreached]! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 11/25/2019 16:17'!
initStackPages
	"Initialize the stackPages.  This version is only for simulation
	 because Slang refuses to inline it, which makes the alloca invalid."
	| stackPageBytes stackPagesBytes theStackMemory |
	stackPageBytes := self stackPageByteSize.
	stackPagesBytes := self computeStackZoneSize.
	theStackMemory := self
							cCode: [self alloca: stackPagesBytes]
							inSmalltalk: [stackPages initializeWithByteSize: stackPagesBytes for: self].
	self cCode: [self memset: theStackMemory _: 0 _: stackPagesBytes].
	stackPages
		initializeStack: theStackMemory
		numSlots: stackPagesBytes / objectMemory wordSize
		pageSize: stackPageBytes / objectMemory wordSize! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 11/25/2019 16:17'!
initStackPagesAndInterpret
	"Initialize the stack pages and enter interpret. Use alloca'ed memory so that when
	 we have a JIT its stack pointer will be on the native stack since alloca allocates
	 memory on the stack. Certain thread systems use the native stack pointer as the
	 frame ID so putting the stack anywhere else can confuse the thread system."

	"This should be in its own initStackPages method but Slang can't inline
	 C code strings."
	| stackPageBytes stackPagesBytes theStackMemory |
	<var: #theStackMemory type: #'void *'>
	stackPageBytes := self stackPageByteSize.
	stackPagesBytes := self computeStackZoneSize.
	theStackMemory := self
							cCode: [self alloca: stackPagesBytes]
							inSmalltalk: [stackPages initializeWithByteSize: stackPagesBytes for: self].
	self cCode: [self memset: theStackMemory _: 0 _: stackPagesBytes].
	stackPages
		initializeStack: theStackMemory
		numSlots: stackPagesBytes / objectMemory wordSize
		pageSize: stackPageBytes / objectMemory wordSize.

	"Once the stack pages are initialized we can continue to bootstrap the system."
	self loadInitialContext.
	"We're ready for the heartbeat (poll interrupt)"
	self ioInitHeartbeat.
	self initialEnterSmalltalkExecutive.
	^nil! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 6/25/2015 13:28'!
initialCleanup
	"This used to cope with issues with images written by VMs earlier than 3.6/3.7.
	 Since we won't be loading such images (being a closure only VM) we only have to
	 deal with external primitives.  Since references to external plugins in methods are
	 cleaned up in snapshotCleanUp only initialize the tables, not visit each method."
	<inline: true>
	self flushMethodCache.
	self flushExternalPrimitiveTable! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 1/2/2021 12:31'!
initialEnterSmalltalkExecutive
	"Main entry-point into the interpreter at system start-up.
	 In the non-threaded VM this is identical to enterSmalltalkExecutive

	 N.B. It also provides the simulator's implementation of ceReturnToInterpreter/ceInvokeInterpreter, which
	 via a simulation trap raise the ReenterInterpreter signal in handleCallOrJumpSimulationTrap:/reenterInterpreter.
	 So when ReenterInterpreter is caught this metod invokes interpret directly. "
	<cmacro: '() enterSmalltalkExecutiveImplementation()'>
	"Simulation of the setjmp in enterSmalltalkExecutiveImplementation for reentry into interpreter."
	| caught |
	caught := false.
	[([caught
			ifFalse: [self enterSmalltalkExecutiveImplementation]
			ifTrue: [self interpret]]
		on: ReenterInterpreter
		do: [:ex|
			caught := true.
			ex return: ex returnValue]) = ReturnToInterpreter] whileTrue! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 12/8/2016 14:10'!
initialIPForHeader: methodHeader method: theMethod
	"Answer a pointer to the initial byte for a method; used only in methods that build a frame."
	<inline: true>
	^theMethod
	+ ((LiteralStart + (objectMemory literalCountOfMethodHeader: methodHeader)) * objectMemory bytesPerOop)
	+ objectMemory baseHeaderSize! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 8/4/2021 11:14'!
initialize
	"Here we can initialize the variables C initializes to zero.  #initialize methods do /not/ get translated."
	super initialize.
	primitiveDoMixedArithmetic := true. "whether we authorize primitives to perform mixed arithmetic or not"
	stackLimit := 0. "This is also the initialization flag for the stack system."
	stackPage := overflowedPage := 0.
	extraFramesToMoveOnOverflow := 0.
	bytecodeSetSelector := 0.
	highestRunnableProcessPriority := 0.
	nextPollUsecs := 0.
	nextWakeupUsecs := 0.
	tempOop := tempOop2 := theUnknownShort := 0.
	interruptPending := false.
	inIOProcessEvents := 0.
	fullScreenFlag := 0.
	sendWheelEvents := deferDisplayUpdates := false.
	displayBits := displayWidth := displayHeight := displayDepth := 0.
	pendingFinalizationSignals := statPendingFinalizationSignals := 0.
	globalSessionID := 0.
	jmpDepth := 0.
	maxExtSemTabSizeSet := false.
	debugCallbackInvokes := debugCallbackPath := debugCallbackReturns := 0.
	statForceInterruptCheck := statStackOverflow := statCheckForEvents :=
	statProcessSwitch := statIOProcessEvents := statStackPageDivorce :=
	statIdleUsecs := 0! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 8/8/2021 07:53'!
initializeExtraClassInstVarIndices
	"Initialize metaclassNumSlots and thisClassIndex which are used in debug printing, and
	 classNameIndex which is used not only for debug printing but for is:KindOf: & is:MemberOf:
	 via classNameOf:is: (evil but a reality we have to accept)."
	| classArrayObj classArrayClass |
	classArrayObj := objectMemory splObj: ClassArray.
	classArrayClass := objectMemory fetchClassOfNonImm: classArrayObj.
	metaclassNumSlots := objectMemory numSlotsOf: classArrayClass.	"determine actual Metaclass instSize"
	thisClassIndex := 5. "default"
	InstanceSpecificationIndex + 1 to: (objectMemory numSlotsOf: classArrayClass) do:
		[:i|
		(objectMemory fetchPointer: i - 1 ofObject: classArrayClass) = classArrayObj ifTrue:
			[thisClassIndex := i - 1]].
	classNameIndex := 6. "default"
	InstanceSpecificationIndex + 1 to: (objectMemory numSlotsOf: classArrayObj) do:
		[:i| | oop |
		oop := objectMemory fetchPointer: i - 1 ofObject: classArrayObj.
		(self object: oop equalsString: 'Array') ifTrue:
			[classNameIndex := i - 1]]! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 8/5/2021 12:04'!
initializeInterpreter: bytesToShift
	"Initialize Interpreter state before starting execution of a new image."
	interpreterProxy := self sqGetInterpreterProxy.
	self dummyReferToProxy.
	objectMemory initializeObjectMemory: bytesToShift.
	self checkAssumedCompactClasses.
	self initializeExtraClassInstVarIndices.
	method := newMethod := objectMemory nilObject.
	self cCode: '' inSmalltalk:
		[breakSelectorLength ifNil:
			[breakSelectorLength := objectMemory minSmallInteger].
		 breakLookupClassTag ifNil: [breakLookupClassTag := -1].
		 reenterInterpreter := ReenterInterpreter new].
	methodDictLinearSearchLimit := 8.
	self initialCleanup.
	LowcodeVM ifTrue: [ self setupNativeStack ].
	profileSemaphore := profileProcess := profileMethod := objectMemory nilObject.
	self cCode: '' inSmalltalk:
		[InitializationOptions at: #profiling ifPresent:
			[:profiling| "hack turn on profiling, for testing in the simulator."
			 profiling ifTrue:
				[profileSemaphore := objectMemory cloneObject: (objectMemory splObj: TheInterruptSemaphore).
				 objectMemory
					storePointerUnchecked: FirstLinkIndex ofObject: profileSemaphore withValue: objectMemory nilObject;
					storePointerUnchecked: NextLinkIndex ofObject: profileSemaphore withValue: objectMemory nilObject;
					storePointerUnchecked: ExcessSignalsIndex ofObject: profileSemaphore withValue: (objectMemory integerObjectOf: 0)]]].
	interruptKeycode := 2094. "cmd-. as used for Mac but no other OS"
	[globalSessionID = 0] whileTrue:
		[globalSessionID := self
								cCode: [((self time: #NULL) + self ioMSecs) bitAnd: 16r7FFFFFFF]
								inSmalltalk: [(Random new next * (SmallInteger maxVal min: 16r7FFFFFFF)) asInteger]].
	metaAccessorDepth := -2.
	super initializeInterpreter: bytesToShift! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 11/7/2013 11:47'!
initializePageTraceToInvalid: aPage
	<var: #aPage type: #'StackPage *'>
	aPage trace: StackPageTraceInvalid "for assert checking of the page tracing flags"! !

!StackInterpreter methodsFor: 'method lookup cache' stamp: 'eem 8/4/2016 11:41'!
inlineLookupInMethodCacheSel: selector classTag: classTag
	"This method implements a simple method lookup cache.  If an entry for the given selector and classTag is
	 found in the cache, set the values of 'newMethod' and 'primitiveFunctionPointer' and answer true. Otherwise,
	 answer false."
	"About the re-probe scheme: The hash is the low bits of the XOR of two large addresses, minus their useless
	 lowest two bits. If a probe doesn't get a hit, the hash is shifted right one bit to compute the next probe,
	 introducing a new randomish bit. The cache is probed CacheProbeMax times before giving up."
	"WARNING: Since the hash computation is based on the object addresses of the class and selector, we must
	 rehash or flush when compacting storage. We've chosen to flush, since that also saves the trouble of updating
	 the addresses of the objects in the cache."
	"classTag is either a class object, if using NewObjectMemory, or a classIndex, if using SpurMemoryManager."

	| hash probe |
	<inline: true>
	hash := objectMemory methodCacheHashOf: selector with: classTag.  "shift drops two low-order zeros from addresses"

	probe := hash bitAnd: MethodCacheMask.  "first probe"
	(((methodCache at: probe + MethodCacheSelector) = selector) and:
		 [(methodCache at: probe + MethodCacheClass) = classTag]) ifTrue:
			[newMethod := methodCache at: probe + MethodCacheMethod.
			primitiveFunctionPointer := self cCoerceSimple: (methodCache at: probe + MethodCachePrimFunction)
											to: #'void (*)()'.
			^true	"found entry in cache; done"].

	probe := (hash >> 1) bitAnd: MethodCacheMask.  "second probe"
	(((methodCache at: probe + MethodCacheSelector) = selector) and:
		 [(methodCache at: probe + MethodCacheClass) = classTag]) ifTrue:
			[newMethod := methodCache at: probe + MethodCacheMethod.
			primitiveFunctionPointer := self cCoerceSimple: (methodCache at: probe + MethodCachePrimFunction)
											to: #'void (*)()'.
			^true	"found entry in cache; done"].

	probe := (hash >> 2) bitAnd: MethodCacheMask.
	(((methodCache at: probe + MethodCacheSelector) = selector) and:
		 [(methodCache at: probe + MethodCacheClass) = classTag]) ifTrue:
			[newMethod := methodCache at: probe + MethodCacheMethod.
			primitiveFunctionPointer := self cCoerceSimple: (methodCache at: probe + MethodCachePrimFunction)
											to: #'void (*)()'.
			^true	"found entry in cache; done"].

	^false! !

!StackInterpreter methodsFor: 'method lookup cache' stamp: 'eem 8/4/2016 11:41'!
inlineLookupInNSMethodCacheSel: selector classTag: classTag method: callingMethod lookupRule: lookupRule
	"Like inlineLookupInMethodCacheSel:classTag:, but the cache is additionally key'd by the calling method and lookupRule/depth and additionally answers localAbsentReceiverOrZero."

	| hash probe |
	<inline: true>
	hash := (objectMemory methodCacheHashOf: selector with: classTag) bitXor: (callingMethod bitXor: lookupRule).

	probe := hash bitAnd: NSMethodCacheMask.  "first probe"
	(((((nsMethodCache at: probe + NSMethodCacheSelector) = selector) and:
		[(nsMethodCache at: probe + NSMethodCacheClassTag) = classTag]) and:
		[(nsMethodCache at: probe + NSMethodCacheCallingMethod) = callingMethod]) and:
		[(nsMethodCache at: probe + NSMethodCacheDepthOrLookupRule) = lookupRule]) ifTrue:
			[newMethod := nsMethodCache at: probe + NSMethodCacheTargetMethod.
			primitiveFunctionPointer := self cCoerceSimple: (nsMethodCache at: probe + NSMethodCachePrimFunction)
											to: #'void (*)()'.
			localAbsentReceiverOrZero := nsMethodCache at: probe + NSMethodCacheActualReceiver.
			^true	"found entry in cache; done"].

	probe := (hash >> 1) bitAnd: NSMethodCacheMask.  "second probe"
	(((((nsMethodCache at: probe + NSMethodCacheSelector) = selector) and:
		[(nsMethodCache at: probe + NSMethodCacheClassTag) = classTag]) and:
		[(nsMethodCache at: probe + NSMethodCacheCallingMethod) = callingMethod]) and:
		[(nsMethodCache at: probe + NSMethodCacheDepthOrLookupRule) = lookupRule]) ifTrue:
			[newMethod := nsMethodCache at: probe + NSMethodCacheTargetMethod.
			primitiveFunctionPointer := self cCoerceSimple: (nsMethodCache at: probe + NSMethodCachePrimFunction)
											to: #'void (*)()'.
			localAbsentReceiverOrZero := nsMethodCache at: probe + NSMethodCacheActualReceiver.
			^true	"found entry in cache; done"].

	probe := (hash >> 2) bitAnd: NSMethodCacheMask.
	(((((nsMethodCache at: probe + NSMethodCacheSelector) = selector) and:
		[(nsMethodCache at: probe + NSMethodCacheClassTag) = classTag]) and:
		[(nsMethodCache at: probe + NSMethodCacheCallingMethod) = callingMethod]) and:
		[(nsMethodCache at: probe + NSMethodCacheDepthOrLookupRule) = lookupRule]) ifTrue:
			[newMethod := nsMethodCache at: probe + NSMethodCacheTargetMethod.
			primitiveFunctionPointer := self cCoerceSimple: (nsMethodCache at: probe + NSMethodCachePrimFunction)
											to: #'void (*)()'.
			localAbsentReceiverOrZero := nsMethodCache at: probe + NSMethodCacheActualReceiver.
			^true	"found entry in cache; done"].

	^false! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 11/3/2014 10:05'!
instVar: offset ofContext: aContext
	"Fetch an instance variable from a maybe married context.
	 If the context is still married compute the value of the
	 relevant inst var from the spouse frame's state."
	| spouseFP |
	<var: #spouseFP type: #'char *'>
	<inline: true>
	self assert: offset < MethodIndex.
	self assert: (objectMemory isContext: aContext).
	(self isMarriedOrWidowedContext: aContext) ifFalse:
		[^objectMemory fetchPointer: offset ofObject: aContext].

	self writeBackHeadFramePointers.
	(self isWidowedContext: aContext) ifTrue:
		[^objectMemory fetchPointer: offset ofObject: aContext].

	spouseFP := self frameOfMarriedContext: aContext.
	offset = SenderIndex ifTrue:
		[^self ensureCallerContext: spouseFP].
	offset = StackPointerIndex ifTrue:
		[self assert: ReceiverIndex + (self stackPointerIndexForFrame: spouseFP) < (objectMemory lengthOf: aContext).
		^objectMemory integerObjectOf: (self stackPointerIndexForFrame: spouseFP)].
	offset = InstructionPointerIndex ifTrue:
		[^self instructionPointerForFrame: spouseFP currentFP: localFP currentIP: (self oopForPointer: localIP)].
	self error: 'bad index'.
	^0! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 3/21/2017 14:58'!
instVar: index ofContext: aMarriedContext put: anOop
	| theFP |
	"Assign the field of a married context.  The important case to optimize is
	 assigning the sender.  We could also consider optimizing assigning the IP but
	 typically that is followed by an assignment to the stack pointer and we can't
	 efficiently assign the stack pointer because it involves moving frames around."
	<inline: true>
	self assert: (self isMarriedOrWidowedContext: aMarriedContext).
	self deny: (objectMemory isObjImmutable: aMarriedContext).
	self writeBackHeadFramePointers.
	(self isStillMarriedContext: aMarriedContext) ifFalse:
		[objectMemory storePointer: index ofObject: aMarriedContext withValue: anOop.
		 index = StackPointerIndex ifTrue:
			[self ensureContextIsExecutionSafeAfterAssignToStackPointer: aMarriedContext].
		 ^nil].
	theFP := self frameOfMarriedContext: aMarriedContext.
	index = SenderIndex ifTrue:
		[| thePage onCurrentPage |
		 thePage := stackPages stackPageFor: theFP.
		 self assert: stackPage = stackPages mostRecentlyUsedPage.
		 onCurrentPage := thePage = stackPage.
		 self storeSenderOfFrame: theFP withValue: anOop.
		 onCurrentPage
			ifTrue:
				[localFP := stackPage headFP.
				 localSP := stackPage headSP]
			ifFalse:
				[stackPages markStackPageMostRecentlyUsed: stackPage].
		 ^nil].
	self externalizeIPandSP.
	self externalDivorceFrame: theFP andContext: aMarriedContext.
	(self is: index methodAssignmentToContextWithMachineCodePC: aMarriedContext) ifTrue:
		[self ensureContextHasBytecodePC: aMarriedContext].
	objectMemory storePointer: index ofObject: aMarriedContext withValue: anOop.
	index = StackPointerIndex ifTrue:
		[self ensureContextIsExecutionSafeAfterAssignToStackPointer: aMarriedContext].
	self internalizeIPandSP.
	"Assigning various fields can force a divorce which can change the stackPage."
	stackPages markStackPageMostRecentlyUsed: stackPage.
	self assertValidExecutionPointe: localIP asUnsignedInteger r: localFP s: localSP imbar: true line: #'__LINE__'! !

!StackInterpreter methodsFor: 'indexing primitive support' stamp: 'eem 1/25/2016 20:12'!
install: rcvr inAtCache: cache at: atIx string: stringy
	"Attempt to install the oop of this object in the given cache (at or atPut),
	 along with its size, format and fixedSize. Answer if this was successful."
	| hdr fmt totalLength fixedFields |
	<inline: true>
	<var: #cache type: 'sqInt *'>
	self assert: (objectMemory isContext: rcvr) not.
	hdr := objectMemory baseHeader: rcvr.
	fmt := objectMemory formatOfHeader: hdr.
	"For now the AtCache code doesn't support 16-bit indexability."
	(objectMemory hasSpurMemoryManagerAPI
	and: [fmt between: objectMemory firstShortFormat and: objectMemory firstByteFormat - 1]) ifTrue:
		[^self primitiveFail].
	stringy
		ifTrue:
			[totalLength := objectMemory lengthOf: rcvr baseHeader: hdr format: fmt.
			 fixedFields := 0.
			 fmt := fmt + objectMemory firstStringyFakeFormat]  "special flag for strings"
		ifFalse:
			[(fmt = objectMemory indexablePointersFormat and: [objectMemory isContextHeader: hdr]) ifTrue:
				["Contexts must not be put in the atCache, since their size is not constant"
				self primitiveFailFor: PrimErrBadReceiver.
				^false].
			 totalLength := objectMemory lengthOf: rcvr baseHeader: hdr format: fmt.
			 fixedFields := objectMemory fixedFieldsOf: rcvr format: fmt length: totalLength].

	cache at: atIx+AtCacheOop put: rcvr.
	cache at: atIx+AtCacheFmt put: fmt.
	cache at: atIx+AtCacheFixedFields put: fixedFields.
	cache at: atIx+AtCacheSize put: totalLength - fixedFields.
	^true! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 12/8/2017 12:51'!
instructionPointerForFrame: spouseFP currentFP: currentFP currentIP: instrPtr
	"Answer the bytecode pc object (i.e. SmallInteger) for an active frame.  The bytecode
	 pc is derived from the frame's pc.  If the frame is the top frame on the current stack
	 the frame pc is whatever the current instruction pointer is.  If the frame is the top
	 frame on some other stack the frame pc is the value on top of stack.  Otherwise the
	 frame pc is the saved pc of the frame above.  Once the frame pc is found it must be
	 mapped to a bytecode pc."
	<var: #spouseFP type: #'char *'>
	<var: #currentFP type: #'char *'>
	| theIP thePage theFPAbove |
	<var: #thePage type: #'StackPage *'>
	<var: #theFPAbove type: #'char *'>
	spouseFP = currentFP
		ifTrue: [theIP := self oopForPointer: instrPtr]
		ifFalse:
			[thePage := stackPages stackPageFor: spouseFP.
			 theFPAbove := self findFrameAbove: spouseFP inPage: thePage.
			 theIP := theFPAbove = 0
						ifTrue: [stackPages longAt: thePage headSP]
						ifFalse:[self oopForPointer: (self frameCallerSavedIP: theFPAbove)]].
	^self contextInstructionPointer: theIP frame: spouseFP! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 3/16/2018 15:31'!
integerArg: index
	"Like #stackIntegerValue: but access method arguments left-to-right"
	<doNotGenerate> "Obsolete; was never used; replaced with methodReturnInteger: in the VirtualMachine struct."
	| oop |
	self deprecated.
	oop := self methodArg: index.
	oop = 0 ifTrue:[^0]. "methodArg: failed"
	^self checkedIntegerValueOf: oop! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
internalAboutToReturn: resultOop through: aContext
	| ourContext |
	<inline: true>
	ourContext := self ensureFrameIsMarried: localFP SP: localSP.
	self internalPush: ourContext.
	self internalPush: resultOop.
	self internalPush: aContext.
	messageSelector := objectMemory splObj: SelectorAboutToReturn.
	argumentCount := 2.
	^self normalSend! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 10/26/2018 12:56'!
internalActivateNewMethod
	| methodHeader numTemps rcvr |
	<inline: true>

	methodHeader := objectMemory methodHeaderOf: newMethod.
	numTemps := self temporaryCountOfMethodHeader: methodHeader.
	self assert: argumentCount = (self argumentCountOfMethodHeader: methodHeader).
	rcvr := self internalStackValue: argumentCount. "could new rcvr be set at point of send?"
	self assert: (objectMemory isOopForwarded: rcvr) not.

	self internalPush: localIP.
	self internalPush: localFP.
	localFP := localSP.
	self internalPush: newMethod.
	self setMethod: newMethod methodHeader: methodHeader.
	self internalPush: (self
						encodeFrameFieldHasContext: false
						isBlock: false
						numArgs: (self argumentCountOfMethodHeader: methodHeader)).
	self internalPush: objectMemory nilObject. "FxThisContext field"
	self internalPush: rcvr.

	"Initialize temps..."
	argumentCount + 1 to: numTemps do:
		[:i | self internalPush: objectMemory nilObject].

	"-1 to account for pre-increment in fetchNextBytecode"
	localIP := self pointerForOop: (self initialIPForHeader: methodHeader method: newMethod) - 1.

	(self methodHeaderHasPrimitive: methodHeader) ifTrue:
		["Skip the CallPrimitive bytecode, if it's there, and store the error code if the method starts
		  with a long store temp.  Strictly no need to skip the store because it's effectively a noop."
		 localIP := localIP + (self sizeOfCallPrimitiveBytecode: methodHeader).
		 primFailCode ~= 0 ifTrue:
			[self reapAndResetErrorCodeTo: localSP header: methodHeader]].

	self assert: (self frameNumArgs: localFP) = argumentCount.
	self assert: (self frameIsBlockActivation: localFP) not.
	self assert: (self frameHasContext: localFP) not.

	"Now check for stack overflow or an event (interrupt, must scavenge, etc)."
	localSP < stackLimit ifTrue:
		[self externalizeIPandSP.
		 self handleStackOverflowOrEventAllowContextSwitch: (self canContextSwitchIfActivating: newMethod header: methodHeader).
		 self internalizeIPandSP]! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
internalCannotReturn: resultOop
	| ourContext |
	<inline: true>
	ourContext := self ensureFrameIsMarried: localFP SP: localSP.
	self internalPush: ourContext.
	self internalPush: resultOop.
	messageSelector := objectMemory splObj: SelectorCannotReturn.
	argumentCount := 1.
	^self normalSend! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 10/31/2012 11:54'!
internalExecuteNewMethod
	| succeeded |
	<inline: true>
	primitiveFunctionPointer ~= 0 ifTrue:
		[self isPrimitiveFunctionPointerAnIndex ifTrue:
			[^self internalQuickPrimitiveResponse].
		 self externalizeIPandSP.
		 succeeded := self slowPrimitiveResponse.
		 self internalizeIPandSP.
		 succeeded ifTrue:
			[self browserPluginReturnIfNeeded.
			^nil]].
	"if not primitive, or primitive failed, activate the method"
	^self internalActivateNewMethod! !

!StackInterpreter methodsFor: 'message sending' stamp: 'rmacnak 5/23/2015 16:17'!
internalFindNewMethodDynamicSuper
	"Find the compiled method to be run when the current messageSelector is sent to the superclass of the class where the current method was looked up, setting the values of 'newMethod' and 'primitiveIndex'."
	<inline: true>
	localAbsentReceiver := self receiver.
	self deny: (objectMemory isOopForwarded: localAbsentReceiver).
	lkupClassTag := objectMemory fetchClassTagOf: localAbsentReceiver.

	(self inlineLookupInNSMethodCacheSel: messageSelector classTag: lkupClassTag method: method lookupRule: LookupRuleDynamicSuper)
	 ifFalse:
		[ | actualLookupRule |
		"entry was not found in the cache; look it up the hard way"
		 self externalizeIPandSP.
		 ((objectMemory isOopForwarded: messageSelector)
		  or: [objectMemory isForwardedClassTag: lkupClassTag]) ifTrue:
			[(objectMemory isOopForwarded: messageSelector) ifTrue:
				[messageSelector := self handleForwardedSelectorFaultFor: messageSelector].
			 (objectMemory isForwardedClassTag: lkupClassTag) ifTrue:
				[lkupClassTag := self handleForwardedSendFaultForTag: lkupClassTag].
			"(self lookupInMethodCacheSel: messageSelector classTag: lkupClassTag) ifTrue:
				[^nil]"].
		 lkupClass := objectMemory classForClassTag: lkupClassTag.
		 actualLookupRule := self lookupDynamicSuperSend.
		 self internalizeIPandSP.
		 self addNewMethodToNSCache: actualLookupRule].
! !

!StackInterpreter methodsFor: 'message sending' stamp: 'rmacnak 5/23/2015 16:21'!
internalFindNewMethodImplicitReceiver
	"Find the compiled method to be run when the for an implicit receiver send of messageSelector, setting the values of 'newMethod' and 'primitiveIndex'."
	<inline: true>
	localAbsentReceiver := self receiver.
	self deny: (objectMemory isOopForwarded: localAbsentReceiver).
	lkupClassTag := objectMemory fetchClassTagOf: localAbsentReceiver.

	(self inlineLookupInNSMethodCacheSel: messageSelector classTag: lkupClassTag method: method lookupRule: LookupRuleImplicit)
	 ifTrue:
		[localAbsentReceiverOrZero = 0 ifFalse: [localAbsentReceiver := localAbsentReceiverOrZero.]]
	 ifFalse:
		[ | actualLookupRule |
		"entry was not found in the cache; look it up the hard way"
		 self externalizeIPandSP.
		 ((objectMemory isOopForwarded: messageSelector)
		  or: [objectMemory isForwardedClassTag: lkupClassTag]) ifTrue:
			[(objectMemory isOopForwarded: messageSelector) ifTrue:
				[messageSelector := self handleForwardedSelectorFaultFor: messageSelector].
			 (objectMemory isForwardedClassTag: lkupClassTag) ifTrue:
				[lkupClassTag := self handleForwardedSendFaultForTag: lkupClassTag].
			"(self lookupInMethodCacheSel: messageSelector classTag: lkupClassTag) ifTrue:
				[^nil]"].
		 lkupClass := objectMemory classForClassTag: lkupClassTag.
		 actualLookupRule := self lookupImplicitReceiverSend.
		 self internalizeIPandSP.
		 self addNewMethodToNSCache: actualLookupRule].
! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 7/18/2015 07:35'!
internalFindNewMethodOrdinary
	"Find the compiled method to be run when the current messageSelector is sent to the class 'lkupClass', setting the values of 'newMethod' and 'primitiveIndex'."
	<inline: true>
	(self inlineLookupInMethodCacheSel: messageSelector classTag: lkupClassTag) ifFalse:
		["entry was not found in the cache; look it up the hard way"
		 self externalizeIPandSP.
		 ((objectMemory isOopForwarded: messageSelector)
		  or: [objectMemory isForwardedClassTag: lkupClassTag]) ifTrue:
			[(objectMemory isOopForwarded: messageSelector) ifTrue:
				[messageSelector := self handleForwardedSelectorFaultFor: messageSelector].
			 (objectMemory isForwardedClassTag: lkupClassTag) ifTrue:
				[lkupClassTag := self handleForwardedSendFaultForTag: lkupClassTag].
			(self lookupInMethodCacheSel: messageSelector classTag: lkupClassTag) ifTrue:
				[^nil]].
		 lkupClass := objectMemory classForClassTag: lkupClassTag.
		 NewspeakVM
			ifTrue: [self lookupOrdinarySend]
			ifFalse: [self lookupMethodInClass: lkupClass].
		 self internalizeIPandSP.
		 self addNewMethodToCache: lkupClass].
! !

!StackInterpreter methodsFor: 'message sending' stamp: 'rmacnak 5/23/2015 16:20'!
internalFindNewMethodOuter: depth
	"Find the compiled method to be run when the current messageSelector is sent to the depth'th enclosing object w/r/t the mixin of the current method, setting the values of 'newMethod' and 'primitiveIndex'."
	<inline: true>
	localAbsentReceiver := self receiver.
	self deny: (objectMemory isOopForwarded: localAbsentReceiver).
	lkupClassTag := objectMemory fetchClassTagOf: localAbsentReceiver.

	(self inlineLookupInNSMethodCacheSel: messageSelector classTag: lkupClassTag method: method lookupRule: depth)
	 ifTrue:
		[localAbsentReceiverOrZero = 0 ifFalse: [localAbsentReceiver := localAbsentReceiverOrZero.]]
	 ifFalse:
		[ | actualLookupRule |
		"entry was not found in the cache; look it up the hard way"
		 self externalizeIPandSP.
		 ((objectMemory isOopForwarded: messageSelector)
		  or: [objectMemory isForwardedClassTag: lkupClassTag]) ifTrue:
			[(objectMemory isOopForwarded: messageSelector) ifTrue:
				[messageSelector := self handleForwardedSelectorFaultFor: messageSelector].
			 (objectMemory isForwardedClassTag: lkupClassTag) ifTrue:
				[lkupClassTag := self handleForwardedSendFaultForTag: lkupClassTag].
			"(self lookupInMethodCacheSel: messageSelector classTag: lkupClassTag) ifTrue:
				[^nil]"].
		 lkupClass := objectMemory classForClassTag: lkupClassTag.
		 actualLookupRule := self lookupOuterSend: depth.
		 self internalizeIPandSP.
		 self addNewMethodToNSCache: actualLookupRule].
! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
internalMustBeBoolean
	<inline: true>
	messageSelector := objectMemory splObj: SelectorMustBeBoolean.
	argumentCount := 0.
	self normalSend! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
internalPop: nItems
	"In the StackInterpreter stacks grow down."
	localSP := localSP + (nItems * objectMemory bytesPerOop)! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
internalPop: nItems thenPush: oop
	"In the StackInterpreter stacks grow down."
	stackPages longAtPointer: (localSP := localSP + ((nItems - 1) * objectMemory bytesPerOop)) put: oop! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
internalPopStack
	"In the StackInterpreter stacks grow down."
	| top |
	top := stackPages longAt: localSP.
	localSP := localSP + objectMemory bytesPerOop.
	^top! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 1/15/2017 20:12'!
internalPopStackFloat32
	<option: #LowcodeVM>
	<returnTypeC: #float>
	<var: #topSingle type: #float>
	"In the StackInterpreter stacks grow down."
	| topSingle  |
	nativeSP := self nativeStackPointerIn: localFP.
	topSingle := stackPages singleFloatAtPointer: nativeSP - 1.
	nativeSP := self nativeStackPointerIn: localFP put: nativeSP + objectMemory bytesPerOop.
	^topSingle! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:40'!
internalPopStackFloat64
	<option: #LowcodeVM>
	<returnTypeC: #double>
	<var: #topDouble type: #double>
	"In the StackInterpreter stacks grow down."
	| topDouble |
	nativeSP := self nativeStackPointerIn: localFP.
	topDouble := stackPages floatAtPointer: nativeSP - 1.
	nativeSP := self nativeStackPointerIn: localFP put: nativeSP + 8.
	^topDouble! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 1/15/2017 20:12'!
internalPopStackInt32
	<option: #LowcodeVM>
	"In the StackInterpreter stacks grow down."
	| topInt32 |
	nativeSP := self nativeStackPointerIn: localFP.
	topInt32 := stackPages int32AtPointer: nativeSP - 1.
	nativeSP := self nativeStackPointerIn: localFP put: nativeSP + objectMemory bytesPerOop.
	^topInt32! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:40'!
internalPopStackInt64
	<option: #LowcodeVM>
	<returnTypeC: #sqLong>
	<var: #topInt64 type: #sqLong>
	"In the StackInterpreter stacks grow down."
	| topInt64 |
	nativeSP := self nativeStackPointerIn: localFP.
	topInt64 := stackPages long64AtPointer: nativeSP - 1.
	nativeSP := self nativeStackPointerIn: localFP put: nativeSP + 8.
	^topInt64! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 9/8/2016 16:48'!
internalPopStackNativeSize: popSize
	<option: #LowcodeVM>
	"In the StackInterpreter stacks grow down."
	nativeSP := self nativeStackPointerIn: localFP.
	nativeSP := self nativeStackPointerIn: localFP put: nativeSP + popSize.! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 1/15/2017 20:12'!
internalPopStackPointer
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #topPointer type: #'char*'>
	"In the StackInterpreter stacks grow down."
	| topPointer |
	nativeSP := self nativeStackPointerIn: localFP.
	topPointer := stackPages pointerAtPointer: nativeSP - 1.
	nativeSP := self nativeStackPointerIn: localFP put: nativeSP + objectMemory bytesPerOop.
	^topPointer! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
internalPush: object
	"In the StackInterpreter stacks grow down."
	stackPages longAtPointer: (localSP := localSP - objectMemory bytesPerOop) put: object! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 1/15/2017 20:10'!
internalPushFloat32: singleFloatValue
	<option: #LowcodeVM>
	<var: #singleFloatValue type: #float >
	"In the StackInterpreter stacks grow down."
	nativeSP := (self nativeStackPointerIn: localFP) - objectMemory bytesPerOop.
	self nativeStackPointerIn: localFP put: nativeSP.

	"In the StackInterpreter stacks grow down."
	stackPages singleFloatAtPointer: (nativeSP - 1) put: singleFloatValue! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:40'!
internalPushFloat64: floatValue
	<option: #LowcodeVM>
	<arg: #floatValue type: #double >
	"In the StackInterpreter stacks grow down."
	nativeSP := (self nativeStackPointerIn: localFP) - 8.
	self nativeStackPointerIn: localFP put: nativeSP.

	"In the StackInterpreter stacks grow down."
	stackPages floatAtPointer: (nativeSP - 1) put: floatValue
! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 1/15/2017 20:10'!
internalPushInt32: int32Value
	<option: #LowcodeVM>
	"In the StackInterpreter stacks grow down."
	nativeSP := (self nativeStackPointerIn: localFP) - objectMemory bytesPerOop.
	self nativeStackPointerIn: localFP put: nativeSP.

	"In the StackInterpreter stacks grow down."
	stackPages int32AtPointer: (nativeSP - 1) put: int32Value
! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:41'!
internalPushInt64: int64Value
	<option: #LowcodeVM>
	<arg: #int64Value type: #sqLong >
	"In the StackInterpreter stacks grow down."
	nativeSP := (self nativeStackPointerIn: localFP) - 8.
	self nativeStackPointerIn: localFP put: nativeSP.

	"In the StackInterpreter stacks grow down."
	stackPages long64AtPointer: (nativeSP - 1) put: int64Value
! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 1/15/2017 20:11'!
internalPushPointer: pointerValue
	<option: #LowcodeVM>
	<var: #pointerValue type: #'char*' >
	"In the StackInterpreter stacks grow down."
	nativeSP := (self nativeStackPointerIn: localFP) - objectMemory bytesPerOop.
	self nativeStackPointerIn: localFP put: nativeSP.

	"In the StackInterpreter stacks grow down."
	stackPages pointerAtPointer: (nativeSP - 1) put: pointerValue
! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 1/15/2017 20:11'!
internalPushShadowCallStackFloat32: singleFloatValue
	<option: #LowcodeVM>
	<var: #singleFloatValue type: #float >
	"In the StackInterpreter stacks grow down."
	shadowCallStackPointer := shadowCallStackPointer - objectMemory bytesPerOop.

	"In the StackInterpreter stacks grow down."
	stackPages singleFloatAtPointer: shadowCallStackPointer put: singleFloatValue! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:41'!
internalPushShadowCallStackFloat64: floatValue
	<option: #LowcodeVM>
	<var: #floatValue type: #double >
	"In the StackInterpreter stacks grow down."
	shadowCallStackPointer := shadowCallStackPointer - 8.

	"In the StackInterpreter stacks grow down."
	stackPages singleFloatAtPointer: shadowCallStackPointer put: floatValue! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 1/15/2017 20:11'!
internalPushShadowCallStackInt32: int32Value
	<option: #LowcodeVM>
	"In the StackInterpreter stacks grow down."
	shadowCallStackPointer := shadowCallStackPointer - objectMemory bytesPerOop.

	"In the StackInterpreter stacks grow down."
	stackPages int32AtPointer: shadowCallStackPointer put: int32Value! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:41'!
internalPushShadowCallStackInt64: int64Value
	<option: #LowcodeVM>
	<arg: #int64Value type: #sqLong >
	
	"In the StackInterpreter stacks grow down."
	shadowCallStackPointer := shadowCallStackPointer - 8.

	"In the StackInterpreter stacks grow down."
	stackPages int64AtPointer: shadowCallStackPointer put: int64Value! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 2/11/2020 10:16'!
internalPushShadowCallStackPointer: pointerValue
	<option: #LowcodeVM>
	<var: #pointerValue type: #'char*' >
	"In the StackInterpreter stacks grow down."
	shadowCallStackPointer := shadowCallStackPointer - objectMemory bytesPerOop.

	"In the StackInterpreter stacks grow down."
	stackPages pointerAtPointer: shadowCallStackPointer put: pointerValue! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 9/5/2016 16:16'!
internalPushShadowCallStackSpace: spaceSize
	<option: #LowcodeVM>
	"In the StackInterpreter stacks grow down."
	shadowCallStackPointer := shadowCallStackPointer - spaceSize.! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 1/11/2017 03:49'!
internalPushShadowCallStackStructure: structurePointer size: size
	<option: #LowcodeVM>
	shadowCallStackPointer := shadowCallStackPointer - size.
	self lowcode_mem: shadowCallStackPointer cp: structurePointer y: size! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 6/26/2012 17:53'!
internalQuickPrimitiveResponse
	"Invoke a quick primitive.
	 Called under the assumption that primFunctionPtr has been preloaded"
	
	| localPrimIndex |
	self assert: self isPrimitiveFunctionPointerAnIndex.
	localPrimIndex := self cCoerceSimple: primitiveFunctionPointer to: #sqInt.
	self assert: (localPrimIndex > 255 and: [localPrimIndex < 520]).
	"Quick return inst vars"
	localPrimIndex >= 264 ifTrue:
		[self internalStackTopPut: (objectMemory fetchPointer: localPrimIndex - 264 ofObject: self internalStackTop).
		 ^true].
	"Quick return constants"
	localPrimIndex = 256 ifTrue: [^true "return self"].
	localPrimIndex = 257 ifTrue: [self internalStackTopPut: objectMemory trueObject. ^true].
	localPrimIndex = 258 ifTrue: [self internalStackTopPut: objectMemory falseObject. ^true].
	localPrimIndex = 259 ifTrue: [self internalStackTopPut: objectMemory nilObject. ^true].
	self internalStackTopPut: (objectMemory integerObjectOf: localPrimIndex - 261).
	^true! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 6/20/2008 14:17'!
internalStackTop

	^stackPages longAtPointer: localSP! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 10/16/2009 11:13'!
internalStackTopPut: aValue

	^stackPages longAtPointer: localSP put: aValue! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'cb 10/24/2017 09:43'!
internalStackTopPutIntegerObjectOf: aValue
	
	^self internalStackTopPut: (objectMemory integerObjectOf: aValue)! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
internalStackValue: offset
	"In the StackInterpreter stacks grow down."
	^stackPages longAtPointer: localSP + (offset * objectMemory bytesPerOop)! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
internalStackValue: offset put: aValue
	"In the StackInterpreter stacks grow down."
	^stackPages longAtPointer: localSP + (offset * objectMemory bytesPerOop) put: aValue! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 09/27/2016 11:20'!
internalizeIPandSP
	"Copy the instruction, stack and frame pointers to local variables for rapid access within the interpret loop."

	localIP := self pointerForOop: instructionPointer.
	localSP := self pointerForOop: stackPointer.
	localFP := self pointerForOop: framePointer.
	LowcodeVM ifTrue: [ nativeSP := 0. ]! !

!StackInterpreter methodsFor: 'interpreter shell' stamp: 'eem 10/29/2020 10:17'!
interpret
	"This is the main interpreter loop. It normally loops forever, fetching and executing bytecodes. When running in the context of a browser plugin VM, however, it must return control to the browser periodically. This should done only when the state of the currently running Squeak thread is safely stored in the object heap. Since this is the case at the moment that a check for interrupts is performed, that is when we return to the browser if it is time to do so. Interrupt checks happen quite frequently."

	<inline: false>
	"If stacklimit is zero then the stack pages have not been initialized."
	stackLimit = 0 ifTrue:
		[^self initStackPagesAndInterpret].
	"record entry time when running as a browser plug-in"
	self browserPluginInitialiseIfNeeded.
	self setMethod: (self frameMethod: framePointer).
	self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer imbar: true line: #'__LINE__'.
	self internalizeIPandSP.
	self initExtensions.
	self fetchNextBytecode.
	[true] whileTrue:
		[self aboutToDispatchBytecode.
		 self dispatchOn: currentBytecode in: BytecodeTable].
	localIP := localIP - 1.  "undo the pre-increment of IP before returning"
	self externalizeIPandSP.
	^nil! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 10/27/2016 15:54'!
interpreterAllocationReserveBytes
	"At a rough approximation we may need to allocate up to a couple
	 of page's worth of contexts when switching stack pages, assigning
	 to senders, etc.  But the snapshot primitive voids all stack pages.
	 So a safe margin is the size of a large context times the maximum
	 number of frames per page times the number of pages."
	<inline: #never>
	| maxUsedBytesPerPage maxFramesPerPage |
	maxUsedBytesPerPage := self stackPageFrameBytes + self stackLimitOffset.
	maxFramesPerPage := maxUsedBytesPerPage / objectMemory wordSize // FrameSlots.
	^maxFramesPerPage * LargeContextSlots * objectMemory bytesPerOop * numStackPages! !

!StackInterpreter methodsFor: 'I/O primitive support' stamp: 'eem 11/10/2020 14:22'!
ioBeDisplay: bitsOrHandle width: width height: height depth: depth
	"Record the position and dimensions of the display bitmap, and inform
	 the display subsystem of the installation of a new display bitmap."
	<var: #bitsOrHandle type: #'void *'>
	displayBits := bitsOrHandle.
	displayWidth := width.
	displayHeight := height.
	displayDepth := depth.
	self ioNoteDisplayChanged: bitsOrHandle width: width height: height depth: depth! !

!StackInterpreter methodsFor: 'plugin support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:43'!
ioFilename: aCharBuffer fromString: aFilenameString ofLength: filenameLength resolveAliases: aBoolean
"the vm has to convert aFilenameString via any canonicalization and char-mapping and put the result in aCharBuffer.
Note the resolveAliases flag - this is an awful artefact of OSX and Apples demented alias handling. When opening a file, the flag must be  true, when closing or renaming it must be false. Sigh."
	<var: #aCharBuffer type: #'char *'>
	<var: #aFilenameString type: #'char *'>
	self cCode:'sqGetFilenameFromString(aCharBuffer, aFilenameString, filenameLength, aBoolean)'
		inSmalltalk:["this doesn't translate well in Smalltalk since we know how long strings are rather than considering them terminated by a 0 char. Do the best we can"
			aCharBuffer replaceFrom:1 to: filenameLength with: aFilenameString]! !

!StackInterpreter methodsFor: 'simulation' stamp: 'eem 1/16/2018 11:16'!
ioForceDisplayUpdate
	"Make sure the display is updated.  Simulation only"
	<doNotGenerate>
	Project current restore! !

!StackInterpreter methodsFor: 'I/O primitive support' stamp: 'eem 10/23/2020 13:42'!
ioGetNextEvent: evtBuf
	<doNotGenerate>
	"SimulatorMorphicModel browse"
	^self eventQueue
		ifNil: [self primitiveFail]
		ifNotNil:
			[:eventQueue|
			 eventQueue nextOrNil ifNotNil:
				[:evt|
				 1 to: evt size do:
					[:i|
					(evt at: i) ifNotNil:
						[:val|
						evtBuf at: i - 1 put: val]]]]! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 1/5/2014 18:41'!
ioLoadExternalFunction: functionName OfLength: functionLength FromModule: moduleName OfLength: moduleLength
	"Load and return the requested function from a module.
	 N.B. The actual code lives in platforms/Cross/vm/sqNamedPrims.h"
	<doNotGenerate>
	| pluginString functionString |
	pluginString := String new: moduleLength.
	1 to: moduleLength do:[:i| pluginString byteAt: i put: (objectMemory byteAt: moduleName+i-1)].
	functionString := String new: functionLength.
	1 to: functionLength do:[:i| functionString byteAt: i put: (objectMemory byteAt: functionName+i-1)].
	^self ioLoadFunction: functionString From: pluginString! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 7/22/2021 17:44'!
ioLoadExternalFunction: functionName OfLength: functionLength FromModule: moduleName OfLength: moduleLength MetadataInto: metadataPtr
	"Load and return the requested function from a module.  Assign the accessor depth and flags through metadataPtr.
	 N.B. The actual code lives in platforms/Cross/vm/sqNamedPrims.h"
	<doNotGenerate>
	| pluginString functionString |
	pluginString := String new: moduleLength.
	(1 to: moduleLength) do:[:i| pluginString byteAt: i put: (objectMemory byteAt: moduleName+i-1)].
	functionString := String new: functionLength.
	(1 to: functionLength) do:[:i| functionString byteAt: i put: (objectMemory byteAt: functionName+i-1)].
	"We used to ignore loads of the SqueakFFIPrims plugin, but that means doing without integerAt:[put:]size:signed:
	 which is too much of a limitation (not that these simulate unaligned accesses yet)."
	^self ioLoadFunction: functionString From: pluginString MetadataInto: metadataPtr! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 10/3/2013 08:51'!
ioLocalMicroseconds
	<doNotGenerate>
	^self ioUTCMicroseconds + (1000000 * DateAndTime localOffset asSeconds)! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 3/11/2014 10:43'!
ioLocalMicrosecondsNow
	<doNotGenerate>
	^self ioUTCMicrosecondsNow + (1000000 * DateAndTime localOffset asSeconds)! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 3/11/2014 11:08'!
ioLocalSecondsOffset
	"simulation-only implementation"
	<doNotGenerate>
	^DateAndTime localOffset asSeconds! !

!StackInterpreter methodsFor: 'I/O primitive support' stamp: 'eem 5/12/2017 17:50'!
ioNoteDisplayChanged: bitsOrHandle width: width height: height depth: depth
	"The real platform may want to note the location of the display and respond to the fact of
	 its installation."
	<doNotGenerate>! !

!StackInterpreter methodsFor: 'I/O primitive support' stamp: 'eem 10/23/2020 14:39'!
ioProcessEvents
	<doNotGenerate>! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 3/11/2014 11:10'!
ioSeconds
	"simulation-only implementation"
	<doNotGenerate>
	^self ioLocalMicroseconds // 1000000! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 3/11/2014 11:10'!
ioSecondsNow
	"simulation-only implementation"
	<doNotGenerate>
	^self ioLocalMicrosecondsNow // 1000000! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 3/11/2014 11:11'!
ioSetHeartbeatMilliseconds: ignored
	"simulation-only implementation"
	<doNotGenerate>! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 3/11/2014 11:12'!
ioUTCMicrosecondsNow
	"simulation-only implementation"
	<doNotGenerate>
	^self ioUTCMicroseconds! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 2/20/2019 17:17'!
ioUnloadModule: moduleName OfLength: moduleNameLength
	"Fake forgiving stub.  The actual code is in platforms/Cross/vm/sqNamedPrims.c"
	<doNotGenerate>
	^true! !

!StackInterpreter methodsFor: 'system control primitives' stamp: 'eem 9/28/2009 10:50'!
ioUpdateVMTimezone
	"A noop in the simulator"
	<doNotGenerate>! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
is: oop KindOf: className
	"Support for external primitives."
	| oopClass |
	<var: #className type:#'char *'>
	oopClass := objectMemory fetchClassOf: oop.
	[oopClass = objectMemory nilObject] whileFalse:[
		(self classNameOf: oopClass Is: className) ifTrue:[^true].
		oopClass := self superclassOf: oopClass].
	^false! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 5/10/2011 15:37'!
is: oop KindOfClass: aClass
	"Support for external primitives."
	<api>
	| oopClass |
	oopClass := objectMemory fetchClassOf: oop.
	[oopClass = objectMemory nilObject] whileFalse:
		[oopClass = aClass ifTrue: [^true].
		 oopClass := self superclassOf: oopClass].
	^false! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
is: oop MemberOf: className
	"Support for external primitives"
	| oopClass |
	<var: #className type:#'char *'>
	oopClass := objectMemory fetchClassOf: oop.
	^(self classNameOf: oopClass Is: className)! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 3/21/2017 14:43'!
is: fieldIndex methodAssignmentToContextWithMachineCodePC: anOop
	"This is a hook for the CoInterpreter.  It is a noop in the StackInterpreter."
	<inline: true>
	^false! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 12/8/2017 12:51'!
isBaseFrame: theFP
	"A base frame (first frame in a stack page) is so marked by having a null saved fp."
	<inline: true>
	<var: #theFP type: #'char *'>
	^(stackPages longAt: theFP + FoxSavedFP) = 0! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 12/18/2018 17:45'!
isBooleanObject: oop
	^oop = objectMemory trueObject or: [oop = objectMemory falseObject]! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 5/7/2020 17:17'!
isCalloutPrimitiveIndex: primIndex
	"This virtual machine provides two primitives that call external code,
	 primitiveExternalCall for plugin primitives, and primitiveCalloutToFFI
	 for FFI calls."
	<inline: true>
	self cCode: [] inSmalltalk: [#(primitiveExternalCall primitiveCalloutToFFI)]. "For senders..."
	^primIndex = PrimNumberExternalCall		"#primitiveExternalCall"
	 or: [primIndex = PrimNumberFFICall]		"#primitiveCalloutToFFI"! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 2/2/2009 16:19'!
isCog
	^false! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 6/29/2014 19:33'!
isEmptyList: aLinkedList
	self assert: (objectMemory isForwarded: aLinkedList) not.
	^ (objectMemory fetchPointer: FirstLinkIndex ofObject: aLinkedList) = objectMemory nilObject! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 05/07/2020 16:13'!
isExternalPrimitiveCall: aMethodObj
	"Answer if the method is an external primitive call (prim 117)."
	<inline: true>
	^(self primitiveIndexOf: aMethodObj) = PrimNumberExternalCall! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 12/1/2014 16:40'!
isFloatObject: oop
	<api>
	<inline: true>
	^objectMemory isFloatInstance: oop! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 10/24/2014 15:30'!
isFrame: aFrame onPage: aPage
	<var: #aFrame type: #'char *'>
	<var: #aPage type: #'StackPage *'>
	| theFP prevFP |
	<var: #theFP type: #'char *'>
	<var: #prevFP type: #'char *'>
	self deny: (stackPages isFree: aPage).
	"Walk the static chain making sure progress is being made,
	 and in the right direction, looking for a match with aFrame."
	theFP := aPage headFP.
	prevFP := theFP - objectMemory wordSize.
	[theFP = aFrame ifTrue: [^true].
	 theFP > prevFP
	 and: [theFP < aPage baseFP]] whileTrue:
		[prevFP := theFP.
		 theFP := self frameCallerFP: theFP].
	^false! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'nice 4/20/2016 00:02'!
isKindOfInteger: oop
	"Answer true if the oop is kind of Integer (Small or Large)."
	<api>
	<inline: true>
	^(objectMemory isIntegerObject: oop)
		or: [objectMemory isLargeIntegerInstance: oop]! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'nice 4/18/2016 23:36'!
isLargeIntegerObject: oop
	<api>
	<inline: true>
	^objectMemory isLargeIntegerInstance: oop! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'nice 4/19/2016 21:45'!
isLargeNegativeIntegerObject: oop
	<api>
	<inline: true>
	^objectMemory isInstanceOfClassLargeNegativeInteger: oop! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'nice 4/19/2016 21:44'!
isLargePositiveIntegerObject: oop
	<api>
	<inline: true>
	^objectMemory isInstanceOfClassLargePositiveInteger: oop! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 1/14/2015 10:36'!
isLiveContext: oop
	"Answer if the argument, which can be any object, is a live context."
	self deny: (objectMemory isOopForwarded: oop).
	(objectMemory isContext: oop) ifFalse:
		[^false].
	(self isSingleContext: oop) ifTrue:
		[^objectMemory isIntegerObject: (objectMemory fetchPointer: InstructionPointerIndex ofObject: oop)].
	^(self isWidowedContext: oop) not! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 1/27/2009 16:45'!
isMachineCodeFrame: theFP
	"For compatibility with CoInterpreter.  Needed to avoid slowPrimitiveResponse
	 failing within ceSend:to:numArgs: et al with an unbalanced stack."
	<var: #theFP type: #'char *'>
	<inline: true>
	^false! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
isMarriedOrWidowedContext: aContext
	^objectMemory isIntegerObject: (objectMemory fetchPointer: SenderIndex ofObject: aContext)! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 5/7/2020 17:19'!
isMetaPrimitiveIndex: primIndex
	"This virtual machine provides two primitives that executes arbitrary primitives, one
	 for indexed primitivces and one for named primitives.  These meta primitives are used
	 in the debugger to execute primitives while simulating execution.  Spur needs to know
	 the accessor depth for a primitive so that failures due to forwarders can be fixed up
	 and retried.  This method identifies such meta primitives so that metaAccessorDepth
	 can be substituted when appropriate."
	<inline: true>
	self cCode: [] inSmalltalk: [#(primitiveDoPrimitiveWithArgs primitiveDoNamedPrimitiveWithArgs)]. "For senders..."
	^primIndex = PrimNumberDoPrimitive
	  or: [primIndex = PrimNumberDoExternalCall]! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 05/07/2020 16:13'!
isNullExternalPrimitiveCall: aMethodObj
	"Answer if the method is an external primitive call (prim 117) with a null external primitive.
	 This is just for an assert in the CoInterpreter."
	| lit |
	((self isExternalPrimitiveCall: aMethodObj)
	and: [(objectMemory literalCountOf: aMethodObj) > 0]) ifFalse:
		[^false].

	lit := self literal: 0 ofMethod: aMethodObj.
	^(objectMemory isArray: lit)
	  and: [(objectMemory numSlotsOf: lit) = 4
	  and: [(objectMemory fetchPointer: 3 ofObject: lit) = ConstZero
			or: [(objectMemory fetchPointer: 3 ofObject: lit) = ConstMinusOne]]]! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 1/22/2015 15:32'!
isOptimizedMethod: methodObj
	<api>
	<option: #SistaVM>
	^self isOptimizedMethodHeader: (objectMemory methodHeaderOf: methodObj)! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 1/22/2015 15:32'!
isOptimizedMethodHeader: header
	<option: #SistaVM>
	^header anyMask: AlternateHeaderIsOptimizedFlag! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'nice 7/8/2016 03:16'!
isPrimitiveFunctionPointerAnIndex
	"We save slots in the method cache by using the primitiveFunctionPointer
	 to hold either a function pointer or the index of a quick primitive. Since
	 quick primitive indices are small they can't be confused with function
	 addresses. "
	^(self cCoerce: primitiveFunctionPointer to: #'usqIntptr_t') <= MaxQuickPrimitiveIndex! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 6/26/2015 15:01'!
isPrivateMethod: methodObj
	<option: #NewspeakVM>
	<inline: true>
	^(self accessModifierOfMethod: methodObj) = AccessModifierPrivate! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 6/26/2015 15:01'!
isProtectedMethod: methodObj
	<option: #NewspeakVM>
	<inline: true>
	^(self accessModifierOfMethod: methodObj) = AccessModifierProtected! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 6/26/2015 15:01'!
isPublicMethod: methodObj
	<option: #NewspeakVM>
	<inline: true>
	^(self accessModifierOfMethod: methodObj) = AccessModifierPublic! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 2/12/2009 20:45'!
isQuickPrimitiveIndex: anInteger
	<api>
	^anInteger between: 256 and: 519! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 7/12/2013 15:27'!
isReadMediatedContextInstVarIndex: index
	"Reading the sender, instructionPointer and stackPointer inst vars of a context must take
	 account of potentially married contexts and fetch the state from the frame. method,
	 closureOrNil and receiver can safely be fetched from the context without checking."
	<api>
	<inline: true>
	^index <= StackPointerIndex! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 9/3/2013 14:58'!
isSingleContext: aContext
	^objectMemory isNonImmediate: (objectMemory fetchPointer: SenderIndex ofObject: aContext)! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 6/28/2008 18:24'!
isStillMarriedContext: aContext
	"Answer if aContext is married or widowed and still married.
	 If a context is widowed then turn it into a single dead context."
	^(self isMarriedOrWidowedContext: aContext)
	    and: [(self isWidowedContext: aContext) not]! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 12/31/2014 13:51'!
isStillMarriedContextDuringGC: aContext
	"Answer if aContext is married or widowed and still married.
	 If a context is widowed then turn it into a single dead context.
	This version is for use during scavenging when stack references may be forwarded.
	 Following what appear to be references to forwarded objects on the stack is dangerous;
	 an instruction ponter may be correctly aligned and may point to bytes that just happen
	 to look like a forwarder. So it is only safe to follow fields that we know are frameContext
	 fields; hence the stack page is walked to check that aOnceMarriedContext is pointing to
	 a live frame.  This only has to happen during scavenging because after a become: all
	 frameContext fields have been followed and so there is no need to follow forwarders."
	^(self isMarriedOrWidowedContext: aContext)
	    and: [(self isWidowedContextDuringGC: aContext) not]! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 4/14/2016 10:31'!
isVanillaBlockClosure: aClosure
	"Answer if aClosure is a vanilla BlockClosure as in the first Cog release, BlockClosure laid out as 
	 'outerContext, startpc, numArgs.  FullBlockClosure is laid out as  outerContext, method, numArgs, receiver.
	 So either answer true if we're not supporting FullBlockClosure, or test the startpc/method field."
	<inline: true>
	^SistaV1BytecodeSet
		ifTrue: [objectMemory isIntegerObject: (objectMemory fetchPointer: ClosureStartPCIndex ofObject: aClosure)]
		ifFalse: [true]! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 12/31/2014 13:58'!
isWidowedContext: aOnceMarriedContext
	"See if the argument is married to a live frame or not.
	 If it is not, turn it into a bereaved single context. This version is safe for use
	 only when no frameContext fields may be forwarded (as maybe the case
	 when scavenging).  Post become: all frameContext fields are followed, and
	 hence nrmally no following of frameCOtext fields is necessary.  But during
	 a scavenge one must use isWidowedContextDuringGC:."
	| theFrame thePage shouldBeFrameCallerField |
	<var: #theFrame type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<var: #shouldBeFrameCallerField type: #'char *'>
	self assert: ((objectMemory isContext: aOnceMarriedContext)
				  and: [self isMarriedOrWidowedContext: aOnceMarriedContext]).
	theFrame := self frameOfMarriedContext: aOnceMarriedContext.
	thePage := stackPages stackPageFor: theFrame.
	((stackPages isFree: thePage)
	 or: [theFrame < thePage headFP]) ifFalse:
		["The frame pointer is within the bounds of a live page.
		   Now check if it matches a frame."
		 shouldBeFrameCallerField := self withoutSmallIntegerTags:
											(objectMemory
												fetchPointer: InstructionPointerIndex
												ofObject: aOnceMarriedContext).
		 ((self frameCallerFP: theFrame) = shouldBeFrameCallerField
		  and: [self frameHasContext: theFrame]) ifTrue:
			[self deny: (((self isFrame: theFrame onPage: thePage))
						and: [objectMemory isForwarded: (self frameContext: theFrame)]).
			 (self frameContext: theFrame) = aOnceMarriedContext ifTrue: "It is still married!!"
				[^false]]].
	"It is out of range or doesn't match the frame's context.
	 It is widowed. Time to wear black."
	self markContextAsDead: aOnceMarriedContext.
	^true! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 12/31/2014 13:49'!
isWidowedContextDuringGC: aOnceMarriedContext
	"See if the argument is married to a live frame or not.  i.e. see if there is a matching
	 frame whose frameContext field is aOnceMarriedContext, or a forwarder to it.
	 If aOnceMarriedContext is not married to a live frame, turn it into a bereaved single context.
	 This version is for use during scavenging when stack references may be forwarded.
	 Following what appear to be references to forwarded objects on the stack is dangerous;
	 an instruction ponter may be correctly aligned and may point to bytes that just happen
	 to look like a forwarder. So it is only safe to follow fields that we know are frameContext
	 fields; hence the stack page is walked to check that aOnceMarriedContext is pointing to
	 a live frame.  This only has to happen during scavenging because after a become: all
	 frameContext fields have been followed and so there is no need to follow forwarders."
	| maybeFrame thePage shouldBeFrameCallerField maybeFrameCtxt |
	<var: #maybeFrame type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<var: #shouldBeFrameCallerField type: #'char *'>
	self assert: ((objectMemory isContext: aOnceMarriedContext)
				  and: [self isMarriedOrWidowedContext: aOnceMarriedContext]).
	maybeFrame := self frameOfMarriedContext: aOnceMarriedContext.
	thePage := stackPages stackPageFor: maybeFrame.
	((stackPages isFree: thePage)
	 or: [maybeFrame < thePage headFP]) ifFalse:
		["The frame pointer is within the bounds of a live page.
		   Now check if it matches a frame."
		 shouldBeFrameCallerField := self withoutSmallIntegerTags:
											(objectMemory
												fetchPointer: InstructionPointerIndex
												ofObject: aOnceMarriedContext).
		 ((self frameCallerFP: maybeFrame) = shouldBeFrameCallerField
		  and: [self frameHasContext: maybeFrame]) ifTrue:
			[maybeFrameCtxt := self frameContext: maybeFrame.
			 "On Spur we need to follow the context to check for a match, but since the VM is
			  only speculating about maybeFrame being a frame, and only speculating about
			  maybeContext being a context, we need to be sure before we can safely follow."
			 (objectMemory hasSpurMemoryManagerAPI
			  and: [(self isFrame: maybeFrame onPage: thePage)
			  and: [objectMemory isForwarded: maybeFrameCtxt]]) ifTrue:
				[maybeFrameCtxt := objectMemory followForwarded: maybeFrameCtxt.
				 self setFrameContext: maybeFrame to: maybeFrameCtxt].
			 maybeFrameCtxt = aOnceMarriedContext ifTrue: "It is still married!!"
				[^false]]].
	"It is out of range or doesn't match the frame's context.
	 It is widowed. Time to wear black."
	self markContextAsDead: aOnceMarriedContext.
	^true! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 7/12/2013 15:28'!
isWriteMediatedContextInstVarIndex: index
	"Wrining any inst vars of a context must take account of potentially married contexts
	 and set the state in the frame. Inst vars in subclasses don't need mediation; subclasses
	 can't marry."
	<api>
	<inline: true>
	^index <= ReceiverIndex! !

!StackInterpreter methodsFor: 'jump bytecodes' stamp: 'eem 11/12/2012 12:20'!
jump: offset
	localIP := localIP + offset + 1.
	self cppIf: MULTIPLEBYTECODESETS
		ifTrue: [currentBytecode := (objectMemory byteAtPointer: localIP) + bytecodeSetSelector]
		ifFalse: [currentBytecode := objectMemory byteAtPointer: localIP]! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 12/13/2017 17:00'!
jumpBinaryInlinePrimitive: primIndex
	<option: #SistaVM>
	| test offset top |
	top := self internalStackTop.
	self assert: (objectMemory isIntegerObject: top).
	offset := objectMemory integerValueOf: top.
	test := self internalStackValue: 1.
	self assert: offset >= 0.
	primIndex caseOf: {
		"7016	jumpIfWritable:
		 Not a forwarder, literal which is a Smi"
		[16]	->	[self deny: (objectMemory isOopForwarded: test).
					 self deny: (objectMemory isImmediate: test).
					 (objectMemory isImmutable: test) ifFalse: [localIP := localIP + offset]].
		"7017	jumpIfReadOnly:
		 Not a forwarder, literal which is a Smi"
		[17]	->	[self deny: (objectMemory isOopForwarded: test).
					 self deny: (objectMemory isImmediate: test).
					 (objectMemory isImmutable: test) ifTrue: [localIP := localIP + offset]].
		"7018	jumpIfYoung:
		 Not a forwarder, literal which is a Smi"
		[18]	->	[self deny: (objectMemory isOopForwarded: test).
					 self deny: (objectMemory isImmediate: test).
					 (objectMemory isYoungObject: test) ifTrue: [localIP := localIP + offset]].
		"7019	jumpIfOld:
		 Not a forwarder, literal which is a Smi"
		[19]	->	[self deny: (objectMemory isOopForwarded: test).
					 self deny: (objectMemory isImmediate: test).
					 (objectMemory isYoungObject: test) ifFalse: [localIP := localIP + offset]].
	}.
	localIP := localIP - 1. "we've already fetched, but we may have incorrectly fetched if jump"
	self fetchNextBytecode.
	self internalPop: 2.
	! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 12/13/2017 17:12'!
jumpTrinaryInlinePrimitive: primIndex
	<option: #SistaVM>
	"Note: those tests work with forwarders (wrong class index)"
	| test classObj offset classTag top |
	top := self internalStackTop.
	self assert: (objectMemory isIntegerObject: top).
	offset := objectMemory integerValueOf: top.
	self assert: offset >= 0.
	test := self internalStackValue: 2.
	classObj := self internalStackValue: 1.
	self assert: (objectMemory isNonImmediate: classObj).
	primIndex caseOf: {
	[0] -> 	["8000	jumpIfInstanceOf:distance:
			 Anything, literal which is a Behavior, literal which is a Smi"
			 self assert: (self objCouldBeClassObj: classObj).
			 classTag := objectMemory fetchClassTagOf: test.
			 classTag = (objectMemory rawClassTagForClass: classObj) ifTrue: 
				[localIP := localIP + offset]].
	[1] -> 	["8001	jumpIfNotInstanceOf:distance:
			 Anything, literal which is a Behavior, literal which is a Smi"
			 self assert: (self objCouldBeClassObj: classObj).
			 classTag := objectMemory fetchClassTagOf: test.
			 classTag = (objectMemory rawClassTagForClass: classObj) ifFalse: 
				[localIP := localIP + offset]].
	[2] -> 	["8002	jumpIfInstanceOfOneOf:distance:
			 Anything, Array of behaviors, literal which is a Smi"
			 self assert: (objectMemory isArrayNonImm: classObj).
			 classTag := objectMemory fetchClassTagOf: test.
			0 to: (objectMemory numSlotsOf: classObj) asInteger - 1 do:
			[:i |
			 classTag = (objectMemory rawClassTagForClass: (objectMemory fetchPointer: i ofObject: classObj)) ifTrue:
				[localIP := localIP + offset]]].
	[3] -> 	["8003	jumpIfNotInstanceOfOneOf:distance:
			  Anything, Array of behaviors, literal which is a Smi"
			 self assert: (objectMemory isArrayNonImm: classObj).
			 classTag := objectMemory fetchClassTagOf: test.
			0 to: (objectMemory numSlotsOf: classObj) asInteger - 1 do:
			[:i |
			 classTag = (objectMemory rawClassTagForClass: (objectMemory fetchPointer: i ofObject: classObj)) ifTrue:
				[localIP := localIP - 1.
				 self fetchNextBytecode.
				 ^self internalPop: 3]].
			localIP := localIP + offset].
	}.
	localIP := localIP - 1. "we've already fetched, but we may have incorrectly fetched if jump"
	self fetchNextBytecode.
	self internalPop: 3
	! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 1/3/2018 16:51'!
jumpUnaryInlinePrimitive: primIndex
	<option: #SistaVM>
	"6000	backjumpNoInterrupt
	 literal which is a Smi"
	| top |
	top := self internalStackTop.
	primIndex = 0 ifTrue: 
		[self assert: (objectMemory isIntegerObject: top). 
		 self assert: (objectMemory integerValueOf: top) < 0.
		"We've already fetched next bytecode, so we add -1"
		 localIP := localIP + (objectMemory integerValueOf: top) - 1. 
		 self fetchNextBytecode.
		 ^self internalPop: 1].
	^ self unknownInlinePrimitive	! !

!StackInterpreter methodsFor: 'jump bytecodes' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
jumplfFalseBy: offset 
	| boolean |
	boolean := self internalStackTop.
	boolean = objectMemory falseObject
		ifTrue: [self jump: offset]
		ifFalse:
			[boolean = objectMemory trueObject ifFalse:
				[^self internalMustBeBoolean].
			self fetchNextBytecode].
	self internalPop: 1! !

!StackInterpreter methodsFor: 'jump bytecodes' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
jumplfTrueBy: offset 
	| boolean |
	boolean := self internalStackTop.
	boolean = objectMemory trueObject
		ifTrue: [self jump: offset]
		ifFalse:
			[boolean = objectMemory falseObject ifFalse:
				[^self internalMustBeBoolean].
			self fetchNextBytecode].
	self internalPop: 1! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 8/23/2018 09:53'!
justActivateNewMethod: mustBeInterpreterFrame
	| methodHeader numArgs numTemps rcvr |
	<inline: true>
	methodHeader := objectMemory methodHeaderOf: newMethod.
	numTemps := self temporaryCountOfMethodHeader: methodHeader.
	numArgs := self argumentCountOfMethodHeader: methodHeader.

	rcvr := self stackValue: numArgs. "could new rcvr be set at point of send?"
	self assert: (objectMemory isOopForwarded: rcvr) not.

	self push: instructionPointer.
	self push: framePointer.
	framePointer := stackPointer.
	self push: newMethod.
	self setMethod: newMethod methodHeader: methodHeader.
	self push: (self encodeFrameFieldHasContext: false isBlock: false numArgs: numArgs).
	self push: objectMemory nilObject. "FxThisContext field"
	self push: rcvr.

	"clear remaining temps to nil"
	numArgs+1 to: numTemps do:
		[:i | self push: objectMemory nilObject].

	instructionPointer := (self initialIPForHeader: methodHeader method: newMethod) - 1.

	(self methodHeaderHasPrimitive: methodHeader) ifTrue:
		["Skip the CallPrimitive bytecode, if it's there, and store the error code if the method starts
		  with a long store temp.  Strictly no need to skip the store because it's effectively a noop."
		 instructionPointer := instructionPointer + (self sizeOfCallPrimitiveBytecode: methodHeader).
		 primFailCode ~= 0 ifTrue:
			[self reapAndResetErrorCodeTo: stackPointer header: methodHeader]].

	^methodHeader! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 10/18/2013 14:15'!
lengthOfNameOfClass: classOop
	<inline: false>
	| numSlots |
	numSlots := objectMemory numSlotsOf: classOop.
	numSlots = metaclassNumSlots ifTrue:
		[^self lengthOfNameOfClass: (objectMemory fetchPointer: thisClassIndex ofObject: classOop)].
	numSlots <= classNameIndex ifTrue:
		[^0].
	^objectMemory lengthOf: (objectMemory fetchPointer: classNameIndex ofObject: classOop)! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 2/27/2009 18:21'!
literal: offset
	self assert: method = (self iframeMethod: localFP).
	^self literal: offset ofMethod: method! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 8/21/2013 12:56'!
literal: offset ofMethod: methodPointer
	<api>
	^objectMemory fetchPointer: offset + LiteralStart ofObject: methodPointer
! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 4/20/2020 10:42'!
literal: offset ofMethod: methodPointer put: oop
	<option: #SpurObjectMemory>
	<inline: true>
	objectMemory storePointer: offset + LiteralStart ofObject: methodPointer withValue: oop
! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 1/22/2015 15:34'!
literalCountOfAlternateHeader: headerPointer
	<inline: true>
	^(objectMemory integerValueOf: headerPointer) bitAnd: AlternateHeaderNumLiteralsMask! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 10/26/2012 15:39'!
literalCountOfOriginalHeader: headerPointer
	<inline: true>
	^(headerPointer >> 10) bitAnd: 16rFF! !

!StackInterpreter methodsFor: 'bitblt support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:26'!
loadBitBltFrom: bb
	"This entry point needs to be implemented for the interpreter proxy.
	Since BitBlt is now a plugin we need to look up BitBltPlugin:=loadBitBltFrom
	and call it. This entire mechanism should eventually go away and be
	replaced with a dynamic lookup from BitBltPlugin itself but for backward
	compatibility this stub is provided"
	| fn |
	<var: #fn type: 'void *'>
	fn := self ioLoadFunction: 'loadBitBltFrom' From: 'BitBltPlugin'.
	fn = 0 ifTrue: [^self primitiveFail].
	^self cCode: '((sqInt (*)(sqInt))fn)(bb)'! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 2/7/2015 10:20'!
loadInitialContext
	<inline: false>
	| activeProc activeContext |
	self cCode: [] inSmalltalk: [self initExtensions].
	objectMemory runLeakCheckerFor: GCModeFull.
	activeProc := self activeProcess.
	activeContext := objectMemory fetchPointer: SuspendedContextIndex ofObject: activeProc.
	self marryContextInNewStackPageAndInitializeInterpreterRegisters: activeContext! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 2/12/2009 14:54'!
log10: n
	"compat with C library."
	<doNotGenerate>
	^n log: 10! !

!StackInterpreter methodsFor: 'jump bytecodes' stamp: 'tpr 3/24/2004 18:37'!
longJumpIfFalse

	self jumplfFalseBy: ((currentBytecode bitAnd: 3) * 256) + self fetchByte.! !

!StackInterpreter methodsFor: 'jump bytecodes' stamp: 'tpr 3/24/2004 18:37'!
longJumpIfTrue

	self jumplfTrueBy: ((currentBytecode bitAnd: 3) * 256) + self fetchByte.! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:49'!
longPrintOop: oop
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| fmt lastIndex startIP bytecodesPerLine column |
	<var: 'field16' type: #'unsigned short'>
	<var: 'field32' type: #'unsigned int'>
	<var: 'field64' type: #usqLong>
	((objectMemory isImmediate: oop)
	 or: [(objectMemory addressCouldBeObj: oop) not
	 or: [(oop bitAnd: objectMemory allocationUnit - 1) ~= 0
	 or: [(objectMemory isFreeObject: oop)
	 or: [objectMemory isForwarded: oop]]]]) ifTrue:
		[self printOop: oop.
		 ^self].
	self printHex: oop.
	(objectMemory fetchClassOfNonImm: oop)
		ifNil: [self print: ' has a nil class!!!!']
		ifNotNil: [:class|
			self print: ': a(n) '; printNameOfClass: class count: 5;
				print: ' ('.
			objectMemory hasSpurMemoryManagerAPI ifTrue:
				[self printHexnp: (objectMemory compactClassIndexOf: oop); print: '=>'].
			self printHexnp: class; print: ')'].
	fmt := objectMemory formatOf: oop.
	self print: ' format '; printHexnp: fmt.
	fmt > objectMemory lastPointerFormat
		ifTrue: [self print: ' nbytes '; printNum: (objectMemory numBytesOf: oop)]
		ifFalse: [(objectMemory isIndexableFormat: fmt) ifTrue:
					[| len |
					len := objectMemory lengthOf: oop.
					self print: ' size '; printNum: len - (objectMemory fixedFieldsOf: oop format: fmt length: len)]].
	objectMemory printHeaderTypeOf: oop.
	self print: ' hash '; printHexnp: (objectMemory rawHashBitsOf: oop).
	self cr.
	(fmt between: objectMemory firstByteFormat and: objectMemory firstCompiledMethodFormat - 1) ifTrue:
		[^self printStringOf: oop; cr].
	(fmt between: objectMemory firstLongFormat and: objectMemory firstByteFormat - 1) ifTrue:
		[0 to: ((objectMemory num32BitUnitsOf: oop) min: 256) - 1 do:
			[:i| | field32 |
			field32 := objectMemory fetchLong32: i ofObject: oop.
			self space; printNum: i; space; printHex: field32; space; cr].
		 ^self].
	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[fmt = objectMemory sixtyFourBitIndexableFormat ifTrue:
			[0 to: ((objectMemory num64BitUnitsOf: oop) min: 256) - 1 do:
				[:i| | field64 |
				field64 := objectMemory fetchLong64: i ofObject: oop.
				self space; printNum: i; space; printHex: field64; space; cr].
			 ^self].
		 (fmt between: objectMemory firstShortFormat and: objectMemory firstShortFormat + 1) ifTrue:
			[0 to: ((objectMemory num16BitUnitsOf: oop) min: 256) - 1 do:
				[:i| | field16 |
				field16 := objectMemory fetchShort16: i ofObject: oop.
				self space; printNum: i; space; printHex: field16; space; cr].
			 ^self]].
	"this is nonsense.  apologies."
	startIP := (objectMemory lastPointerOf: oop) + objectMemory bytesPerOop - objectMemory baseHeaderSize / objectMemory bytesPerOop.
	lastIndex := 256 min: startIP.
	lastIndex > 0 ifTrue:
		[1 to: lastIndex do:
			[:i| | fieldOop |
			fieldOop := objectMemory fetchPointer: i - 1 ofObject: oop.
			self space; printNum: i - 1; space; printHex: fieldOop; space.
			(i = 1 and: [objectMemory isCompiledMethod: oop])
				ifTrue: [self printMethodHeaderOop: fieldOop]
				ifFalse: [self cCode: [self printOopShort: fieldOop]
							inSmalltalk: [self print: (self shortPrint: fieldOop)]].
			self cr]].
	(objectMemory isCompiledMethod: oop)
		ifFalse:
			[startIP > lastIndex ifTrue: [self print: '...'; cr]]
		ifTrue:
			[startIP := startIP * objectMemory wordSize + 1.
			 lastIndex := objectMemory lengthOf: oop.
			 lastIndex - startIP > 100 ifTrue:
				[lastIndex := startIP + 100].
			 bytecodesPerLine := 8.
			 column := 1.
			 startIP to: lastIndex do:
				[:index| | byte |
				column = 1 ifTrue:
					[self cCode: 'printf("0x%08" PRIxSQPTR ": ", (usqIntptr_t)(oop+BaseHeaderSize+index-1))'
						inSmalltalk: [self print: (oop+objectMemory baseHeaderSize+index-1) hex; print: ': ']].
				byte := objectMemory fetchByte: index - 1 ofObject: oop.
				self cCode: 'printf(" %02x/%-3d", (int)byte,(int)byte)'
					inSmalltalk: [self space; print: (byte radix: 16); printChar: $/; printNum: byte].
				column := column + 1.
				column > bytecodesPerLine ifTrue:
					[column := 1. self cr]].
			column = 1 ifFalse:
				[self cr]]! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 10/30/2012 17:34'!
longPushTemporaryVariableBytecode
	"230		11100110	i i i i i i i i	Push Temporary Variable #iiiiiiii"
	| index |
	index := self fetchByte.
	self fetchNextBytecode.
	self internalPush: (self temporary: index in: localFP)! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 10/30/2012 16:56'!
longStoreAndPopTemporaryVariableBytecode
	"237		11101101	i i i i i i i i	Pop and Store Temporary Variable #iiiiiiii"
	self longStoreTemporaryVariableBytecode.
	self internalPop: 1! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 6/7/2016 14:45'!
longStoreBytecodeForHeader: methodHeader
	"Answer the relevant long store temp bytecode, which indicates it has a primitive error code."
	"234		11101010	i i i i i i i i	Store Temporary Variable #iiiiiiii"
	"129 		10000001 jjkkkkkk 	Store (Receiver Variable, Temporary Location, Illegal, Literal Variable) [jj] #kkkkkk"
	<api>
	<inline: true>
	^self cppIf: MULTIPLEBYTECODESETS
		ifTrue: [(objectMemory headerIndicatesAlternateBytecodeSet: methodHeader)
					ifTrue: [AltLongStoreBytecode]
					ifFalse: [LongStoreBytecode]]
		ifFalse: [LongStoreBytecode]! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 10/30/2012 17:34'!
longStoreTemporaryVariableBytecode
	"234		11101010	i i i i i i i i	Store Temporary Variable #iiiiiiii"
	| index |
	index := self fetchByte.
	self fetchNextBytecode.
	self temporary: index in: localFP put: self internalStackTop! !

!StackInterpreter methodsFor: 'jump bytecodes' stamp: 'eem 11/15/2012 11:44'!
longUnconditionalJump
	| offset |
	offset := (((currentBytecode bitAnd: 7) - 4) * 256) + self fetchByte.
	localIP := localIP + offset.
	self ifBackwardsCheckForEvents: offset.
	self fetchNextBytecode! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 2/24/2020 13:55'!
lookupBreakFor: lookupClass
	<inline: true>
	(breakSelectorLength <= 0
	 and: [objectMemory shouldBreakForLookupIn: lookupClass given: breakLookupClassTag]) ifTrue:
		[self
			cCode: [self warning: 'lookup class send break (heartbeat suppressed)']
			inSmalltalk: [self halt: 'Lookup in class ', lookupClass hex]]! !

!StackInterpreter methodsFor: 'message sending' stamp: 'rmacnak 8/25/2015 19:46'!
lookupDynamicSuperSend
	"Do the full lookup for a Newspeak super send.
	IN: messageSelector
	IN: argumentCount
	OUT: localAbsentReceiver
	OUT: newMethod
	OUT: primitiveIndex
	RESULT: LookupRuleDynamicSuper or LookupRuleMNU"

	| methodMixin methodMixinApplication |
	localAbsentReceiver := self receiver.
	localAbsentReceiverOrZero := 0.
	methodMixin := self methodClassOf: method.
	methodMixinApplication := self
		findApplicationOfTargetMixin: methodMixin
		startingAtBehavior: (objectMemory fetchClassOf: localAbsentReceiver).
	lkupClass := self superclassOf: methodMixinApplication. "MNU lookup starts here."
	^self lookupProtected: messageSelector startingAt: lkupClass rule: LookupRuleDynamicSuper! !

!StackInterpreter methodsFor: 'message sending' stamp: 'rmacnak 8/25/2015 19:51'!
lookupImplicitReceiverSend
	"Do the full lookup for an implicit receiver send.
	IN: messageSelector
	IN: argumentCount
	OUT: localAbsentReceiver
	OUT: localAbsentReceiverOrZero
	OUT: newMethod
	OUT: primitiveIndex
	RESULT: LookupRuleImplicit or LookupRuleMNU"

	| methodReceiver candidateReceiver candidateMixin candidateMixinApplication dictionary found |
	messageSelector := objectMemory followMaybeForwarded: messageSelector.
	methodReceiver := self receiver.
	candidateReceiver := methodReceiver.
	self deny: (objectMemory isForwarded: method).
	candidateMixin := self methodClassOf: method.
	localAbsentReceiverOrZero := 0.
	[self deny: (objectMemory isForwarded: candidateMixin).
	self deny: (objectMemory isForwarded: candidateReceiver).
	candidateMixinApplication := self
		findApplicationOfTargetMixin: candidateMixin
		startingAtBehavior: (objectMemory fetchClassOf: candidateReceiver).
	self deny: (candidateMixinApplication = 0).
	self deny: (candidateMixinApplication = objectMemory nilObject).
	self deny: (objectMemory isForwarded: candidateMixinApplication).
	self assert: (self addressCouldBeClassObj: candidateMixinApplication).
	dictionary := objectMemory followObjField: MethodDictionaryIndex ofObject: candidateMixinApplication.
	found := self lookupMethodInDictionary: dictionary.
	found ifTrue:
		[localAbsentReceiver := candidateReceiver.
		^self lookupLexical: messageSelector from: candidateMixin rule: LookupRuleImplicit].
	candidateMixin := objectMemory followObjField: EnclosingMixinIndex ofObject: candidateMixin.
	self deny: (objectMemory isForwarded: candidateMixin).
	candidateMixin = objectMemory nilObject]
		whileFalse:
			[localAbsentReceiverOrZero := candidateReceiver := objectMemory followObjField: EnclosingObjectIndex ofObject: candidateMixinApplication].
	"There is no lexically visible method, so the implicit receiver is the method receiver."
	localAbsentReceiverOrZero := 0.
	localAbsentReceiver := methodReceiver.
	lkupClass := objectMemory fetchClassOf: methodReceiver. "MNU lookup starts here."
	^self lookupProtected: messageSelector startingAt: lkupClass rule: LookupRuleImplicit.! !

!StackInterpreter methodsFor: 'method lookup cache' stamp: 'eem 7/21/2014 06:32'!
lookupInMethodCacheSel: selector classTag: classTag
	"This method implements a simple method lookup cache.  If an entry for the given
	 selector and classTag is found in the cache, set the values of 'newMethod' and
	 'primitiveFunctionPointer' and answer true. Otherwise, answer false."
	<inline: false>
	^self inlineLookupInMethodCacheSel: selector classTag: classTag! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 6/5/2017 11:05'!
lookupLexical: selector from: mixin rule: rule
	"A shared part of the lookup for implicit receiver sends that found a lexically visible
	method, and self and outer sends."
	| receiverClass mixinApplication dictionary found |
	receiverClass := objectMemory fetchClassOf: localAbsentReceiver.
	self lookupBreakFor: receiverClass.
	lkupClass := receiverClass. "MNU lookup starts here."
	mixinApplication := self findApplicationOfTargetMixin: mixin startingAtBehavior: receiverClass.
	dictionary := objectMemory followObjField: MethodDictionaryIndex ofObject: mixinApplication.
	found := self lookupMethodInDictionary: dictionary.
	(found and: [(self isPrivateMethod: newMethod)]) ifTrue:
		[^rule].
	^self lookupProtected: selector startingAt: receiverClass rule: rule
! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 6/5/2017 11:04'!
lookupMNU
	"A send lookup failed. Replace the arguments on the stack with a Message and lookup
	 #doesNotUndestand: starting at lkupClass. Note that MNU lookup ignores access modifiers.
	 This makes it different from an ordinary send of #doesNotUnderstand:, which must only
	 find public methods.
	IN: lkupClass
	IN: messageSelector
	IN: argumentCount
	OUT: newMethod
	OUT: primitiveIndex
	RESULT: LookupRuleMNU"

	| currentClass dictionary found |
	self lookupBreakFor: lkupClass.
	self createActualMessageTo: lkupClass.
	messageSelector := objectMemory splObj: SelectorDoesNotUnderstand.
	currentClass := lkupClass.
	[currentClass ~= objectMemory nilObject] whileTrue:
		[dictionary := objectMemory followObjField: MethodDictionaryIndex ofObject: currentClass.
		found := self lookupMethodInDictionary: dictionary.
		found ifTrue: [^LookupRuleMNU].
		currentClass := self superclassOf: currentClass].

	self error: 'Recursive not understood error encountered'
! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 6/5/2017 11:04'!
lookupMNUInClass: class
	"Lookup messageSelector in class.  Answer 0 on success. Answer the splObj: index
	 for the error selector to use on failure rather than performing MNU processing etc."
	| currentClass dictionary found |
	<inline: false>
	self lookupBreakFor: class.
	currentClass := class.
	[currentClass ~= objectMemory nilObject] whileTrue:
		[dictionary := objectMemory followObjField: MethodDictionaryIndex ofObject: currentClass.
		 dictionary = objectMemory nilObject ifTrue:
			[lkupClass := self superclassOf: currentClass.
			 ^SelectorCannotInterpret].
		found := self lookupMethodInDictionary: dictionary.
		NewspeakVM
			ifTrue: [found ifTrue: [lkupClass := class. self addNewMethodToNSCache: LookupRuleMNU. ^0]]
			ifFalse: [found ifTrue: [self addNewMethodToCache: class. ^0]].
		currentClass := self superclassOf: currentClass].
	lkupClass := class.
	^SelectorDoesNotUnderstand! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 11/10/2014 12:49'!
lookupMethodFor: selector InDictionary: dictionary
	"Lookup the argument selector in aDictionary and answer either the
	 method or nil, if not found.
	This method lookup tolerates integers as Dictionary keys to support
	 execution of images in which Symbols have been compacted out."
	| length index mask wrapAround nextSelector methodArray |
	<inline: true>
	length := objectMemory numSlotsOf: dictionary.
	mask := length - SelectorStart - 1.
	index := SelectorStart + (objectMemory methodDictionaryHash: selector mask: mask).

	"It is assumed that there are some nils in this dictionary, and search will 
	 stop when one is encountered. However, if there are no nils, then wrapAround 
	 will be detected the second time the loop gets to the end of the table."
	wrapAround := false.
	[true] whileTrue:
		[nextSelector := objectMemory fetchPointer: index ofObject: dictionary.
		 nextSelector = objectMemory nilObject ifTrue:
			[^nil].
		 (objectMemory isOopForwarded: nextSelector) ifTrue:
			[nextSelector := objectMemory
								fixFollowedField: index + SelectorStart
								ofObject: dictionary
								withInitialValue: nextSelector].
		 nextSelector = selector ifTrue:
			[methodArray := objectMemory followObjField: MethodArrayIndex ofObject: dictionary.
			 ^objectMemory followField: index - SelectorStart ofObject: methodArray].
		 index := index + 1.
		 index = length ifTrue:
			[wrapAround ifTrue: [^nil].
			 wrapAround := true.
			 index := SelectorStart]].
	^nil "for Slang"! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 2/24/2020 14:10'!
lookupMethodInClass: class
	| currentClass dictionary found |
	<inline: false>
	self assert: (self addressCouldBeClassObj: class).
	self lookupBreakFor: class.
	currentClass := class.
	[currentClass ~= objectMemory nilObject] whileTrue:
		[dictionary := objectMemory followObjField: MethodDictionaryIndex ofObject: currentClass.
		dictionary = objectMemory nilObject ifTrue:
			["MethodDict pointer is nil (hopefully due a swapped out stub)
				-- raise exception #cannotInterpret:."
			self createActualMessageTo: class.
			messageSelector := objectMemory splObj: SelectorCannotInterpret.
			self sendBreakpoint: messageSelector receiver: nil.
			^self lookupMethodInClass: (self superclassOf: currentClass)].
		found := self lookupMethodInDictionary: dictionary.
		found ifTrue: [^currentClass].
		currentClass := self superclassOf: currentClass].

	"Could not find #doesNotUnderstand: -- unrecoverable error."
	messageSelector = (objectMemory splObj: SelectorDoesNotUnderstand) ifTrue:
		[self error: 'Recursive not understood error encountered'].

	"Cound not find a normal message -- raise exception #doesNotUnderstand:"
	self createActualMessageTo: class.
	messageSelector := objectMemory splObj: SelectorDoesNotUnderstand.
	self sendBreak: messageSelector + objectMemory baseHeaderSize
		point: (objectMemory lengthOf: messageSelector)
		classTag: (objectMemory fetchClassTagOfClass: class).
	^self lookupMethodInClass: class! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 11/10/2014 12:49'!
lookupMethodInDictionary: dictionary 
	"This method lookup tolerates integers as Dictionary keys to support
	 execution of images in which Symbols have been compacted out."
	| length index mask wrapAround nextSelector methodArray |
	<inline: true>
	length := objectMemory numSlotsOf: dictionary.
	mask := length - SelectorStart - 1.
	"Use linear search on small dictionaries; its cheaper.
	 Also the limit can be set to force linear search of all dictionaries, which supports the
	 booting of images that need rehashing (e.g. because a tracer has generated an image
	 with different hashes but hasn't rehashed it yet.)"
	mask <= methodDictLinearSearchLimit ifTrue:
		[index := 0.
		 [index <= mask] whileTrue:
			[nextSelector := objectMemory fetchPointer: index + SelectorStart ofObject: dictionary.
			 (objectMemory isOopForwarded: nextSelector) ifTrue:
				[nextSelector := objectMemory
									fixFollowedField: index + SelectorStart
									ofObject: dictionary
									withInitialValue: nextSelector].
			 nextSelector = messageSelector ifTrue:
				[methodArray := objectMemory followObjField: MethodArrayIndex ofObject: dictionary.
				 newMethod := objectMemory followField: index ofObject: methodArray.
				^true].
		 index := index + 1].
		 ^false].
	index := SelectorStart + (objectMemory methodDictionaryHash: messageSelector mask: mask).

	"It is assumed that there are some nils in this dictionary, and search will 
	 stop when one is encountered. However, if there are no nils, then wrapAround 
	 will be detected the second time the loop gets to the end of the table."
	wrapAround := false.
	[true] whileTrue:
		[nextSelector := objectMemory fetchPointer: index ofObject: dictionary.
		 nextSelector = objectMemory nilObject ifTrue: [^false].
		 (objectMemory isOopForwarded: nextSelector) ifTrue:
			[nextSelector := objectMemory
								fixFollowedField: index + SelectorStart
								ofObject: dictionary
								withInitialValue: nextSelector].
		 nextSelector = messageSelector ifTrue:
			[methodArray := objectMemory followObjField: MethodArrayIndex ofObject: dictionary.
			 newMethod := objectMemory followField: index - SelectorStart ofObject: methodArray.
			^true].
		 index := index + 1.
		 index = length ifTrue:
			[wrapAround ifTrue: [^false].
			 wrapAround := true.
			 index := SelectorStart]].
	
	^false "for Slang"! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 6/5/2017 11:03'!
lookupOrdinaryNoMNUEtcInClass: class
	"Lookup messageSelector in class.  Answer 0 on success. Answer the splObj: index
	 for the error selector to use on failure rather than performing MNU processing etc."
	| currentClass dictionary found |
	<inline: false>
	self lookupBreakFor: class.
	currentClass := class.
	[currentClass ~= objectMemory nilObject] whileTrue:
		[dictionary := objectMemory followObjField: MethodDictionaryIndex ofObject: currentClass.
		 dictionary = objectMemory nilObject ifTrue:
			[lkupClass := self superclassOf: currentClass.
			 ^SelectorCannotInterpret].
		found := self lookupMethodInDictionary: dictionary.
		NewspeakVM
			ifTrue:
				[found ifTrue:
					[(self isPublicMethod: newMethod) ifTrue:
						[self addNewMethodToCache: class. ^0].
					(self isProtectedMethod: newMethod) ifTrue:
						[lkupClass := class. ^SelectorDoesNotUnderstand]]]
			ifFalse:
				[found ifTrue:
					[self addNewMethodToCache: class. ^0]].
		currentClass := self superclassOf: currentClass].
	lkupClass := class.
	^SelectorDoesNotUnderstand! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 6/5/2017 11:03'!
lookupOrdinarySend
	"Do the full lookup for an ordinary send (i.e., a Newspeak or Smalltalk ordinary send or a Smalltalk super send).
	IN: lkupClass
	IN: messageSelector
	IN: argumentCount
	OUT: newMethod
	OUT: primitiveIndex
	RESULT: LookupRuleOrdinary or LookupRuleMNU"
	<option: #NewspeakVM>
	| currentClass dictionary found |
	self assert: (self addressCouldBeClassObj: lkupClass).
	self lookupBreakFor: lkupClass.
	currentClass := lkupClass.
	[currentClass ~= objectMemory nilObject] whileTrue:
		[dictionary := objectMemory followObjField: MethodDictionaryIndex ofObject: currentClass.
		found := self lookupMethodInDictionary: dictionary.
		found ifTrue:
			[(self isPublicMethod: newMethod) ifTrue:
				[^LookupRuleOrdinary].
			(self isProtectedMethod: newMethod) ifTrue:
				[^self lookupMNU]].
		currentClass := self superclassOf: currentClass].
	^self lookupMNU! !

!StackInterpreter methodsFor: 'message sending' stamp: 'rmacnak 5/23/2015 16:40'!
lookupOuterSend: depth
	"Do the full lookup for a self or outer send.
	IN: messageSelector
	IN: argumentCount
	OUT: localAbsentReceiver
	OUT: localAbsentReceiverOrZero
	OUT: newMethod
	OUT: primitiveIndex
	RESULT: [depth] or LookupRuleMNU"

	| targetMixin count mixinApplication |
	localAbsentReceiver := self receiver.
	localAbsentReceiverOrZero := 0.
	targetMixin := self methodClassOf: method.
	count := 0.
	[count < depth] whileTrue:
		[count := count + 1.
		mixinApplication := self
			findApplicationOfTargetMixin: targetMixin
			startingAtBehavior: (objectMemory fetchClassOf: localAbsentReceiver).
		localAbsentReceiverOrZero := localAbsentReceiver := objectMemory followObjField: EnclosingObjectIndex ofObject: mixinApplication.
		targetMixin := objectMemory followObjField: EnclosingMixinIndex ofObject: targetMixin].
	^self lookupLexical: messageSelector from: targetMixin rule: depth! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 6/5/2017 11:02'!
lookupProtected: selector startingAt: mixinApplication rule: rule
	"A shared part of the lookup for self, outer or implicit receiver sends that did not find a
	private lexically visible method, and (Newspeak) super sends."
	| currentClass dictionary found |
	self lookupBreakFor: mixinApplication.
	currentClass := mixinApplication.
	[currentClass = objectMemory nilObject] whileFalse:
		[dictionary := objectMemory followObjField: MethodDictionaryIndex ofObject: currentClass.
		found := self lookupMethodInDictionary: dictionary.
		(found and: [(self isPrivateMethod: newMethod) not]) ifTrue:
			[^rule].
		currentClass := self superclassOf: currentClass].
	^self lookupMNU! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 2/23/2017 17:15'!
lookupSelector: selector inClass: class
	"Lookup selector in class.  Answer the method or nil.  This is a debugging routine.
	 It does /not/ side-effect lookupClass or newMethod."
	| currentClass dictionary |
	<api>

	currentClass := class.
	[currentClass ~= objectMemory nilObject] whileTrue:
		[dictionary := objectMemory followObjField: MethodDictionaryIndex ofObject: currentClass.
		 dictionary = objectMemory nilObject ifTrue:
			[^nil].
		 (self lookupMethodFor: selector InDictionary: dictionary) ifNotNil:
			[:meth| ^meth].
		currentClass := self superclassOf: currentClass].
	^nil! !

!StackInterpreter methodsFor: 'inline primitive dispatch generated code' stamp: ''!
lowcodeBinaryInlinePrimitive: prim
	<option: #LowcodeVM>	"Lowcode instruction interpreter dispatch"
	prim
		caseOf: {
			[0]	-> [ ^ self lowcodePrimitiveByteSizeOf ].
			[1]	-> [ ^ self lowcodePrimitiveFirstFieldPointer ].
			[2]	-> [ ^ self lowcodePrimitiveFirstIndexableFieldPointer ].
			[3]	-> [ ^ self lowcodePrimitiveIsBytes ].
			[4]	-> [ ^ self lowcodePrimitiveIsFloatObject ].
			[5]	-> [ ^ self lowcodePrimitiveIsIndexable ].
			[6]	-> [ ^ self lowcodePrimitiveIsIntegerObject ].
			[7]	-> [ ^ self lowcodePrimitiveIsPointers ].
			[8]	-> [ ^ self lowcodePrimitiveIsWords ].
			[9]	-> [ ^ self lowcodePrimitiveIsWordsOrBytes ].
			[10]	-> [ ^ self lowcodePrimitiveOopSmallIntegerToInt32 ].
			[11]	-> [ ^ self lowcodePrimitiveOopSmallIntegerToInt64 ].
			[12]	-> [ ^ self lowcodePrimitiveOopToBoolean32 ].
			[13]	-> [ ^ self lowcodePrimitiveOopToBoolean64 ].
			[14]	-> [ ^ self lowcodePrimitiveOopToFloat32 ].
			[15]	-> [ ^ self lowcodePrimitiveOopToFloat64 ].
			[16]	-> [ ^ self lowcodePrimitiveOopToInt32 ].
			[17]	-> [ ^ self lowcodePrimitiveOopToInt64 ].
			[18]	-> [ ^ self lowcodePrimitiveOopToPointer ].
			[19]	-> [ ^ self lowcodePrimitiveOopToPointerReinterpret ].
			[20]	-> [ ^ self lowcodePrimitiveOopToUInt32 ].
			[21]	-> [ ^ self lowcodePrimitiveOopToUInt64 ].
			[22]	-> [ ^ self lowcodePrimitivePin ].
			[23]	-> [ ^ self lowcodePrimitiveUnpin ].
		}
		otherwise: [
			localIP := localIP - 3.
			^self respondToUnknownBytecode
		].
	

! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'rsf 8/9/2016 23:42'!
lowcodeCalloutState: calloutState callFunction: functionPointer
	<option: #LowcodeVM>
	self lowcodeCalloutState: calloutState stackPointer: shadowCallStackPointer stackSize: (self cCoerce: lowcodeCalloutState to: #'char*') - shadowCallStackPointer callFunction: functionPointer
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/27/2016 16:53'!
lowcodeCalloutState: calloutState float32Register: registerId
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/27/2016 16:52'!
lowcodeCalloutState: calloutState float32Register: registerId value: value
	"This is implemented as C macros in platform specific headers"
	<var: #value type: #'float' >
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/27/2016 16:54'!
lowcodeCalloutState: calloutState float64Register: registerId
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/27/2016 16:53'!
lowcodeCalloutState: calloutState float64Register: registerId value: value
	"This is implemented as C macros in platform specific headers"
	<var: #value type: #'double' >
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/27/2016 16:51'!
lowcodeCalloutState: calloutState int32Register: registerId
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/27/2016 16:53'!
lowcodeCalloutState: calloutState int32Register: registerId value: value
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/27/2016 16:53'!
lowcodeCalloutState: calloutState int64Register: registerId
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/27/2016 16:53'!
lowcodeCalloutState: calloutState int64Register: registerId value: value
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/27/2016 20:23'!
lowcodeCalloutState: calloutState pointerRegister: registerId
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/27/2016 20:23'!
lowcodeCalloutState: calloutState pointerRegister: registerId value: value
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/28/2016 11:45'!
lowcodeCalloutState: calloutState stackPointer: calloutStackPointer stackSize: calloutStackSize callFunction: functionPointer
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/28/2016 11:14'!
lowcodeCalloutStateFetchResultFloat32: calloutState
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/28/2016 11:14'!
lowcodeCalloutStateFetchResultFloat64: calloutState
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/28/2016 11:14'!
lowcodeCalloutStateFetchResultInt32: calloutState
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/28/2016 11:14'!
lowcodeCalloutStateFetchResultInt64: calloutState
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/28/2016 11:14'!
lowcodeCalloutStateFetchResultPointer: calloutState
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'RonieSalgado 4/28/2016 11:42'!
lowcodeCalloutStateFetchResultStructure: calloutState
	"This is implemented as C macros in platform specific headers"
	<doNotGenerate>
! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'rsf 2/12/2017 01:38'!
lowcodeDoCallout: functionPointer
	<option: #LowcodeVM>
	<var: #functionPointer type: #'char*' >
	self undoFetchNextBytecode.
	self externalizeIPandSP.
	self lowcodeCalloutState: lowcodeCalloutState callFunction: functionPointer.
	self internalizeIPandSP.
	self reloadLowcodeStateAfterCallout.
	self fetchNextBytecode.
	^ self lowcodeCalloutStateFetchResultFloat32: lowcodeCalloutState! !

!StackInterpreter methodsFor: 'inline primitive dispatch generated code' stamp: ''!
lowcodeNullaryInlinePrimitive: prim
	<option: #LowcodeVM>	"Lowcode instruction interpreter dispatch"
	prim
		caseOf: {
			[0]	-> [ ^ self lowcodePrimitiveBoolean32ToOop ].
			[1]	-> [ ^ self lowcodePrimitiveBoolean64ToOop ].
			[2]	-> [ ^ self lowcodePrimitiveFloat32ToOop ].
			[3]	-> [ ^ self lowcodePrimitiveFloat64ToOop ].
			[4]	-> [ ^ self lowcodePrimitiveInt32ToOop ].
			[5]	-> [ ^ self lowcodePrimitiveInt64ToOop ].
			[6]	-> [ ^ self lowcodePrimitivePointerToOop ].
			[7]	-> [ ^ self lowcodePrimitivePointerToOopReinterprer ].
			[8]	-> [ ^ self lowcodePrimitiveSmallInt32ToOop ].
			[9]	-> [ ^ self lowcodePrimitiveUint32ToOop ].
			[10]	-> [ ^ self lowcodePrimitiveUint64ToOop ].
		}
		otherwise: [
			localIP := localIP - 3.
			^self respondToUnknownBytecode
		].
	

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveAdd32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqInt' >
	<var: #result type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	result := first + second.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveAdd64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqLong' >
	<var: #result type: #'sqLong' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	result := first + second.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveAlloca32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer size |
	<var: #pointer type: #'char*' >
	<var: #size type: #'sqInt' >
	size := self internalPopStackInt32.

	nativeStackPointer := self cCoerce:
	((self cCoerce: nativeStackPointer - size to: 'size_t') bitAnd: -16)
	to: 'char*'.
	pointer := nativeStackPointer.

	self internalPushPointer: pointer.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveAlloca64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer size |
	<var: #pointer type: #'char*' >
	<var: #size type: #'sqLong' >
	size := self internalPopStackInt64.

	nativeStackPointer := nativeStackPointer - size.
	pointer := nativeStackPointer.

	self internalPushPointer: pointer.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveAnd32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqInt' >
	<var: #result type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	result := first bitAnd: second.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveAnd64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqLong' >
	<var: #result type: #'sqLong' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	result := first bitAnd: second.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveArithmeticRightShift32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value shiftAmount result |
	<var: #value type: #'sqInt' >
	<var: #shiftAmount type: #'sqInt' >
	<var: #result type: #'sqInt' >
	shiftAmount := self internalPopStackInt32.
	value := self internalPopStackInt32.

	result := value >> shiftAmount.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveArithmeticRightShift64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value shiftAmount result |
	<var: #value type: #'sqLong' >
	<var: #shiftAmount type: #'sqLong' >
	<var: #result type: #'sqLong' >
	shiftAmount := self internalPopStackInt64.
	value := self internalPopStackInt64.

	result := value >> shiftAmount.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveBeginCall
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| alignment |
	alignment := extA.

	"Store the shadow stack pointer"
	self shadowCallStackPointerIn: localFP put: shadowCallStackPointer + 1.
	"Allocate the callout state"
	self allocateLowcodeCalloutState.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveBoolean32ToOop
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	value := self internalPopStackInt32.

	object := value ~= 0 ifTrue: [ objectMemory trueObject ] ifFalse: [objectMemory falseObject].

	self internalPush: object.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveBoolean64ToOop
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqLong' >
	value := self internalPopStackInt64.

	object := value ~= 0 ifTrue: [ objectMemory trueObject ] ifFalse: [objectMemory falseObject].

	self internalPush: object.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveByteSizeOf
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	object := self internalPopStack.

	value := self byteSizeOf: object.

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveCallArgumentFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| argumentValue |
	<var: #argumentValue type: #'float' >
	argumentValue := self internalPopStackFloat32.

	self internalPushShadowCallStackFloat32: argumentValue.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveCallArgumentFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| argumentValue |
	<var: #argumentValue type: #'double' >
	argumentValue := self internalPopStackFloat64.

	self internalPushShadowCallStackFloat64: argumentValue.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveCallArgumentInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value |
	<var: #value type: #'sqInt' >
	value := self internalPopStackInt32.

	self internalPushShadowCallStackInt32: value.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveCallArgumentInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value |
	<var: #value type: #'sqLong' >
	value := self internalPopStackInt64.

	self internalPushShadowCallStackInt64: value.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveCallArgumentPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointerValue |
	<var: #pointerValue type: #'char*' >
	pointerValue := self internalPopStackPointer.

	self internalPushShadowCallStackPointer: pointerValue.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveCallArgumentSpace
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| spaceSize |
	spaceSize := extA.

	self internalPushShadowCallStackSpace: spaceSize.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveCallArgumentStructure
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| structurePointer structureSize |
	<var: #structurePointer type: #'char*' >
	structureSize := extA.
	structurePointer := self internalPopStackPointer.

	self internalPushShadowCallStackStructure: structurePointer size: structureSize.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveCallInstruction
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| function |
	function := extA.

	self abort.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveCallPhysical
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| registerID |
	registerID := extA.

	self abort.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveCheckSessionIdentifier
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| expectedSession value |
	<var: #value type: #'sqInt' >
	expectedSession := extA.

	value := (expectedSession = self getThisSessionID) ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveCompareAndSwap32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| newValue check oldValue value |
	<var: #newValue type: #'sqInt' >
	<var: #check type: #'char*' >
	<var: #oldValue type: #'sqInt' >
	<var: #value type: #'sqInt' >
	newValue := self internalPopStackInt32.
	oldValue := self internalPopStackInt32.
	check := self internalPopStackPointer.

	self abort.

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveDiv32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqInt' >
	<var: #result type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	result := first // second.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveDiv64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqLong' >
	<var: #result type: #'sqLong' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	result := first // second.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveDuplicateFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value dup1 dup2 |
	<var: #value type: #'float' >
	<var: #dup1 type: #'float' >
	<var: #dup2 type: #'float' >
	value := self internalPopStackFloat32.

	dup1 := value.
	dup2 := value.

	self internalPushFloat32: dup1.
	self internalPushFloat32: dup2.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveDuplicateFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value dup1 dup2 |
	<var: #value type: #'double' >
	<var: #dup1 type: #'double' >
	<var: #dup2 type: #'double' >
	value := self internalPopStackFloat64.

	dup1 := value.
	dup2 := value.

	self internalPushFloat64: dup1.
	self internalPushFloat64: dup2.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveDuplicateInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value dup1 dup2 |
	<var: #value type: #'sqInt' >
	<var: #dup1 type: #'sqInt' >
	<var: #dup2 type: #'sqInt' >
	value := self internalPopStackInt32.

	dup1 := value.
	dup2 := value.

	self internalPushInt32: dup1.
	self internalPushInt32: dup2.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveDuplicateInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value dup1 dup2 |
	<var: #value type: #'sqLong' >
	<var: #dup1 type: #'sqLong' >
	<var: #dup2 type: #'sqLong' >
	value := self internalPopStackInt64.

	dup1 := value.
	dup2 := value.

	self internalPushInt64: dup1.
	self internalPushInt64: dup2.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveDuplicatePointer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| dup2 dup1 pointerValue |
	<var: #dup2 type: #'char*' >
	<var: #dup1 type: #'char*' >
	<var: #pointerValue type: #'char*' >
	pointerValue := self internalPopStackPointer.

	dup1 := pointerValue.
	dup2 := pointerValue.

	self internalPushPointer: dup1.
	self internalPushPointer: dup2.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveEffectiveAddress32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result offset base scale index |
	<var: #result type: #'char*' >
	<var: #offset type: #'sqInt' >
	<var: #base type: #'char*' >
	<var: #scale type: #'sqInt' >
	<var: #index type: #'sqInt' >
	offset := self internalPopStackInt32.
	scale := self internalPopStackInt32.
	index := self internalPopStackInt32.
	base := self internalPopStackPointer.

	result := base + (index*scale) + offset.

	self internalPushPointer: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveEffectiveAddress64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result offset base scale index |
	<var: #result type: #'char*' >
	<var: #offset type: #'sqLong' >
	<var: #base type: #'char*' >
	<var: #scale type: #'sqLong' >
	<var: #index type: #'sqLong' >
	offset := self internalPopStackInt64.
	scale := self internalPopStackInt64.
	index := self internalPopStackInt64.
	base := self internalPopStackPointer.

	result := base + (index*scale) + offset.

	self internalPushPointer: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveEndCall
	<option: #LowcodeVM>	"Lowcode instruction generator"

	"Restore the shadow stack pointer"
	shadowCallStackPointer := (self shadowCallStackPointerIn: localFP) - 1.
	self shadowCallStackPointerIn: localFP put: (self cCoerce: 1 to: #'char*').


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveEndCallNoCleanup
	<option: #LowcodeVM>	"Lowcode instruction generator"

	"Restore the shadow stack pointer"
	shadowCallStackPointer := (self shadowCallStackPointerIn: localFP) - 1.
	self shadowCallStackPointerIn: localFP put: (self cCoerce: 1 to: #'char*').


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFirstFieldPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object pointer |
	<var: #object type: #'sqInt' >
	<var: #pointer type: #'char*' >
	object := self internalPopStack.

	pointer := objectMemory firstFixedField: object.

	self internalPushPointer: pointer.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFirstIndexableFieldPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object pointer |
	<var: #object type: #'sqInt' >
	<var: #pointer type: #'char*' >
	object := self internalPopStack.

	pointer := objectMemory firstIndexableField: object.

	self internalPushPointer: pointer.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32Add
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'float' >
	<var: #result type: #'float' >
	<var: #first type: #'float' >
	second := self internalPopStackFloat32.
	first := self internalPopStackFloat32.

	result := first + second.

	self internalPushFloat32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32Div
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'float' >
	<var: #result type: #'float' >
	<var: #first type: #'float' >
	second := self internalPopStackFloat32.
	first := self internalPopStackFloat32.

	result := first / second.

	self internalPushFloat32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32Equal
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'float' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'float' >
	second := self internalPopStackFloat32.
	first := self internalPopStackFloat32.

	value := first = second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32Great
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'float' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'float' >
	second := self internalPopStackFloat32.
	first := self internalPopStackFloat32.

	value := first > second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32GreatEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'float' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'float' >
	second := self internalPopStackFloat32.
	first := self internalPopStackFloat32.

	value := first >= second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32Less
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'float' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'float' >
	second := self internalPopStackFloat32.
	first := self internalPopStackFloat32.

	value := first < second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32LessEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'float' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'float' >
	second := self internalPopStackFloat32.
	first := self internalPopStackFloat32.

	value := first <= second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32Mul
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'float' >
	<var: #result type: #'float' >
	<var: #first type: #'float' >
	second := self internalPopStackFloat32.
	first := self internalPopStackFloat32.

	result := first * second.

	self internalPushFloat32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32Neg
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'float' >
	<var: #result type: #'float' >
	value := self internalPopStackFloat32.

	result := value negated.

	self internalPushFloat32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32NotEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'float' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'float' >
	second := self internalPopStackFloat32.
	first := self internalPopStackFloat32.

	value := first ~= second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32Sqrt
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'float' >
	<var: #result type: #'float' >
	value := self internalPopStackFloat32.

	result := self sqrt: value.

	self internalPushFloat32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32Sub
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'float' >
	<var: #result type: #'float' >
	<var: #first type: #'float' >
	second := self internalPopStackFloat32.
	first := self internalPopStackFloat32.

	result := first - second.

	self internalPushFloat32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32ToFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| singleFloatValue doubleResult |
	<var: #singleFloatValue type: #'float' >
	<var: #doubleResult type: #'double' >
	singleFloatValue := self internalPopStackFloat32.

	doubleResult := self cCoerce: singleFloatValue to: 'double'.

	self internalPushFloat64: doubleResult.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32ToInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'float' >
	<var: #result type: #'sqInt' >
	value := self internalPopStackFloat32.

	result := self cCoerce: value to: 'sqInt'.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32ToInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'float' >
	<var: #result type: #'sqLong' >
	value := self internalPopStackFloat32.

	result := self cCoerce: value to: 'sqLong'.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32ToOop
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| singleFloatValue object |
	<var: #singleFloatValue type: #'float' >
	<var: #object type: #'sqInt' >
	singleFloatValue := self internalPopStackFloat32.

	object := objectMemory floatObjectOf: singleFloatValue.

	self internalPush: object.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32ToUInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'float' >
	<var: #result type: #'sqInt' >
	value := self internalPopStackFloat32.

	result := self cCoerce: value to: 'uint32_t'.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat32ToUInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'float' >
	<var: #result type: #'sqLong' >
	value := self internalPopStackFloat32.

	result := self cCoerce: value to: 'uint64_t'.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64Add
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'double' >
	<var: #result type: #'double' >
	<var: #first type: #'double' >
	second := self internalPopStackFloat64.
	first := self internalPopStackFloat64.

	result := first + second.

	self internalPushFloat64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64Div
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'double' >
	<var: #result type: #'double' >
	<var: #first type: #'double' >
	second := self internalPopStackFloat64.
	first := self internalPopStackFloat64.

	result := first / second.

	self internalPushFloat64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64Equal
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'double' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'double' >
	second := self internalPopStackFloat64.
	first := self internalPopStackFloat64.

	value := first = second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64Great
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'double' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'double' >
	second := self internalPopStackFloat64.
	first := self internalPopStackFloat64.

	value := first > second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64GreatEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'double' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'double' >
	second := self internalPopStackFloat64.
	first := self internalPopStackFloat64.

	value := first >= second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64Less
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'double' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'double' >
	second := self internalPopStackFloat64.
	first := self internalPopStackFloat64.

	value := first < second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64LessEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'double' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'double' >
	second := self internalPopStackFloat64.
	first := self internalPopStackFloat64.

	value := first <= second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64Mul
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'double' >
	<var: #result type: #'double' >
	<var: #first type: #'double' >
	second := self internalPopStackFloat64.
	first := self internalPopStackFloat64.

	result := first * second.

	self internalPushFloat64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64Neg
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'double' >
	<var: #result type: #'double' >
	value := self internalPopStackFloat64.

	result := value negated.

	self internalPushFloat64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64NotEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'double' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'double' >
	second := self internalPopStackFloat64.
	first := self internalPopStackFloat64.

	value := first ~= second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64Sqrt
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'double' >
	<var: #result type: #'double' >
	value := self internalPopStackFloat64.

	result := self sqrt: value.

	self internalPushFloat64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64Sub
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'double' >
	<var: #result type: #'double' >
	<var: #first type: #'double' >
	second := self internalPopStackFloat64.
	first := self internalPopStackFloat64.

	result := first - second.

	self internalPushFloat64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64ToFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| floatValue singleFloatResult |
	<var: #floatValue type: #'double' >
	<var: #singleFloatResult type: #'float' >
	floatValue := self internalPopStackFloat64.

	singleFloatResult := self cCoerce: floatValue to: 'float'.

	self internalPushFloat32: singleFloatResult.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64ToInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| floatValue int32Result |
	<var: #floatValue type: #'double' >
	<var: #int32Result type: #'sqInt' >
	floatValue := self internalPopStackFloat64.

	int32Result := self cCoerce: floatValue to: 'sqInt'.

	self internalPushInt32: int32Result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64ToInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| floatValue int64Result |
	<var: #floatValue type: #'double' >
	<var: #int64Result type: #'sqLong' >
	floatValue := self internalPopStackFloat64.

	int64Result := self cCoerce: floatValue to: 'sqLong'.

	self internalPushInt64: int64Result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64ToOop
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| floatValue object |
	<var: #floatValue type: #'double' >
	<var: #object type: #'sqInt' >
	floatValue := self internalPopStackFloat64.

	object := objectMemory floatObjectOf: floatValue.

	self internalPush: object.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64ToUInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| floatValue int64Result |
	<var: #floatValue type: #'double' >
	<var: #int64Result type: #'sqInt' >
	floatValue := self internalPopStackFloat64.

	int64Result := self cCoerce: floatValue to: 'uint32_t'.

	self internalPushInt32: int64Result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFloat64ToUInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| floatValue int64Result |
	<var: #floatValue type: #'double' >
	<var: #int64Result type: #'sqLong' >
	floatValue := self internalPopStackFloat64.

	int64Result := self cCoerce: floatValue to: 'uint64_t'.

	self internalPushInt64: int64Result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveFree
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer |
	<var: #pointer type: #'char*' >
	pointer := self internalPopStackPointer.

	self free: pointer.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInstantiateIndexable32Oop
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| indexableSize object classOop |
	<var: #indexableSize type: #'sqInt' >
	<var: #object type: #'sqInt' >
	<var: #classOop type: #'sqInt' >
	indexableSize := self internalPopStackInt32.
	classOop := self internalPopStack.

	object := objectMemory instantiateClass: classOop indexableSize: indexableSize.

	self internalPush: object.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInstantiateIndexableOop
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| indexableSize object classOop |
	<var: #object type: #'sqInt' >
	<var: #classOop type: #'sqInt' >
	indexableSize := extA.
	classOop := self internalPopStack.

	object := objectMemory instantiateClass: classOop indexableSize: indexableSize.

	self internalPush: object.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInstantiateOop
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object classOop |
	<var: #object type: #'sqInt' >
	<var: #classOop type: #'sqInt' >
	classOop := self internalPopStack.

	object := objectMemory instantiateClass: classOop indexableSize: 0.

	self internalPush: object.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt32Equal
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqInt' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	value := first = second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt32Great
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqInt' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	value := first > second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt32GreatEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqInt' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	value := first >= second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt32Less
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqInt' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	value := first < second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt32LessEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqInt' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	value := first <= second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt32NotEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqInt' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	value := first ~= second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt32ToFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'float' >
	value := self internalPopStackInt32.

	result := self cCoerce: value to: 'float'.

	self internalPushFloat32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt32ToFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'double' >
	value := self internalPopStackInt32.

	result := self cCoerce: value to: 'double'.

	self internalPushFloat64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt32ToOop
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	value := self internalPopStackInt32.

	object := objectMemory signed32BitIntegerFor: value.

	self internalPush: object.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: 'eem 11/11/2020 18:36'!
lowcodePrimitiveInt32ToPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'char*' >
	value := self internalPopStackInt32.

	result := self cCoerce: (self cCoerce: value to: 'usqIntptr_t') to: 'char*'.

	self internalPushPointer: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt64Equal
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqLong' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	value := first = second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt64Great
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqLong' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	value := first > second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt64GreatEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqLong' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	value := first >= second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt64Less
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqLong' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	value := first < second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt64LessEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqLong' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	value := first <= second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt64NotEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqLong' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	value := first ~= second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt64ToFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqLong' >
	<var: #result type: #'double' >
	value := self internalPopStackInt64.

	result := self cCoerce: value to: 'float'.

	self internalPushFloat64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt64ToFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqLong' >
	<var: #result type: #'double' >
	value := self internalPopStackInt64.

	result := self cCoerce: value to: 'double'.

	self internalPushFloat64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt64ToOop
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqLong' >
	value := self internalPopStackInt64.

	object := objectMemory signed64BitIntegerFor: value.

	self internalPush: object.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveInt64ToPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqLong' >
	<var: #result type: #'char*' >
	value := self internalPopStackInt64.

	result := self cCoerce:
	(self cCoerce: value to: 'intptr_t')
	to: 'char*'.

	self internalPushPointer: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveIsBytes
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	object := self internalPopStack.

	value := (objectMemory isBytes: object) ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveIsFloatObject
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	object := self internalPopStack.

	value := (objectMemory isFloatObject: object) ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveIsIndexable
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	object := self internalPopStack.

	value := (objectMemory isIndexable: object) ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveIsIntegerObject
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	object := self internalPopStack.

	value := (objectMemory isIntegerObject: object) ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveIsPointers
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	object := self internalPopStack.

	value := (objectMemory isPointers: object) ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveIsWords
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	object := self internalPopStack.

	value := (objectMemory isWords: object) ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveIsWordsOrBytes
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	object := self internalPopStack.

	value := (objectMemory isWordsOrBytes: object) ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLeftShift32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value shiftAmount result |
	<var: #value type: #'sqInt' >
	<var: #shiftAmount type: #'sqInt' >
	<var: #result type: #'sqInt' >
	shiftAmount := self internalPopStackInt32.
	value := self internalPopStackInt32.

	result := value << shiftAmount.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLeftShift64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value shiftAmount result |
	<var: #value type: #'sqLong' >
	<var: #shiftAmount type: #'sqLong' >
	<var: #result type: #'sqLong' >
	shiftAmount := self internalPopStackInt64.
	value := self internalPopStackInt64.

	result := value << shiftAmount.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadArgumentAddress
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer baseOffset |
	<var: #pointer type: #'char*' >
	baseOffset := extA.

	pointer := self framePointerOfNativeArgument: baseOffset in: localFP.

	self internalPushPointer: pointer.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadArgumentFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| floatValue baseOffset |
	<var: #floatValue type: #'float' >
	baseOffset := extA.

	floatValue := stackPages singleFloatAtPointer: (self framePointerOfNativeArgument: baseOffset in: localFP).

	self internalPushFloat32: floatValue.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadArgumentFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| doubleValue baseOffset |
	<var: #doubleValue type: #'double' >
	baseOffset := extA.

	doubleValue := stackPages floatAtPointer: (self framePointerOfNativeArgument: baseOffset in: localFP).

	self internalPushFloat64: doubleValue.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadArgumentInt16
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.

	value := stackPages int16AtPointer: (self framePointerOfNativeArgument: baseOffset in: localFP).

	self internalPushInt32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadArgumentInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.

	value := stackPages int32AtPointer: (self framePointerOfNativeArgument: baseOffset in: localFP).

	self internalPushInt32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadArgumentInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqLong' >
	baseOffset := extA.

	value := stackPages int64AtPointer: (self framePointerOfNativeArgument: baseOffset in: localFP).

	self internalPushInt64: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadArgumentInt8
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.

	value := stackPages int8AtPointer: (self framePointerOfNativeArgument: baseOffset in: localFP).

	self internalPushInt32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadArgumentPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointerResult baseOffset |
	<var: #pointerResult type: #'char*' >
	baseOffset := extA.

	pointerResult := stackPages pointerAtPointer: (self framePointerOfNativeArgument: baseOffset in: localFP).

	self internalPushPointer: pointerResult.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadArgumentUInt16
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.

	value := stackPages uint16AtPointer: (self framePointerOfNativeArgument: baseOffset in: localFP).

	self internalPushInt32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadArgumentUInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.

	value := stackPages uint32AtPointer: (self framePointerOfNativeArgument: baseOffset in: localFP).

	self internalPushInt32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadArgumentUInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqLong' >
	baseOffset := extA.

	value := stackPages uint64AtPointer: (self framePointerOfNativeArgument: baseOffset in: localFP).

	self internalPushInt64: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadArgumentUInt8
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.

	value := stackPages uint8AtPointer: (self framePointerOfNativeArgument: baseOffset in: localFP).

	self internalPushInt32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadFloat32FromMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer value |
	<var: #pointer type: #'char*' >
	<var: #value type: #'float' >
	pointer := self internalPopStackPointer.

	value := self singleFloatAtPointer: pointer.

	self internalPushFloat32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadFloat64FromMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer value |
	<var: #pointer type: #'char*' >
	<var: #value type: #'double' >
	pointer := self internalPopStackPointer.

	value := self floatAtPointer: pointer.

	self internalPushFloat64: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadInt16FromMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer value |
	<var: #pointer type: #'char*' >
	<var: #value type: #'sqInt' >
	pointer := self internalPopStackPointer.

	value := self int16AtPointer: pointer.

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadInt32FromMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer value |
	<var: #pointer type: #'char*' >
	<var: #value type: #'sqInt' >
	pointer := self internalPopStackPointer.

	value := self int32AtPointer: pointer.

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadInt64FromMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer value |
	<var: #pointer type: #'char*' >
	<var: #value type: #'sqLong' >
	pointer := self internalPopStackPointer.

	value := self int64AtPointer: pointer.

	self internalPushInt64: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadInt8FromMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer value |
	<var: #pointer type: #'char*' >
	<var: #value type: #'sqInt' >
	pointer := self internalPopStackPointer.

	value := self int8AtPointer: pointer.

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadLocalAddress
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer baseOffset |
	<var: #pointer type: #'char*' >
	baseOffset := extA.

	pointer := self framePointerOfNativeLocal: baseOffset in: localFP.

	self internalPushPointer: pointer.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadLocalFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| floatValue baseOffset |
	<var: #floatValue type: #'float' >
	baseOffset := extA.

	floatValue := stackPages singleFloatAtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP).

	self internalPushFloat32: floatValue.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadLocalFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| doubleValue baseOffset |
	<var: #doubleValue type: #'double' >
	baseOffset := extA.

	doubleValue := stackPages floatAtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP).

	self internalPushFloat64: doubleValue.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadLocalInt16
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.

	value := stackPages int16AtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP).

	self internalPushInt32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadLocalInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.

	value := stackPages int32AtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP).

	self internalPushInt32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadLocalInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqLong' >
	baseOffset := extA.

	value := stackPages int64AtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP).

	self internalPushInt64: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadLocalInt8
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.

	value := stackPages int8AtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP).

	self internalPushInt32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadLocalPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointerResult baseOffset |
	<var: #pointerResult type: #'char*' >
	baseOffset := extA.

	pointerResult := stackPages pointerAtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP).

	self internalPushPointer: pointerResult.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadLocalUInt16
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.

	value := stackPages uint16AtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP).

	self internalPushInt32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadLocalUInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.

	value := stackPages uint32AtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP).

	self internalPushInt32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadLocalUInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqLong' >
	baseOffset := extA.

	value := stackPages uint64AtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP).

	self internalPushInt64: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadLocalUInt8
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.

	value := stackPages uint8AtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP).

	self internalPushInt32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadObjectAt
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| fieldIndex object fieldValue |
	<var: #fieldIndex type: #'sqInt' >
	<var: #object type: #'sqInt' >
	<var: #fieldValue type: #'sqInt' >
	fieldIndex := self internalPopStackInt32.
	object := self internalPopStack.

	fieldValue := objectMemory fetchPointer: fieldIndex ofObject: object.

	self internalPush: fieldValue.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadObjectField
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| fieldIndex object fieldValue |
	<var: #object type: #'sqInt' >
	<var: #fieldValue type: #'sqInt' >
	fieldIndex := extA.
	object := self internalPopStack.

	fieldValue := self fetchPointer: fieldIndex ofObject: object.

	self internalPush: fieldValue.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadPointerFromMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointerResult pointer |
	<var: #pointerResult type: #'char*' >
	<var: #pointer type: #'char*' >
	pointer := self internalPopStackPointer.

	pointerResult := self pointerAtPointer: pointer.

	self internalPushPointer: pointerResult.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadUInt16FromMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer value |
	<var: #pointer type: #'char*' >
	<var: #value type: #'sqInt' >
	pointer := self internalPopStackPointer.

	value := self uint16AtPointer: pointer.

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadUInt32FromMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer value |
	<var: #pointer type: #'char*' >
	<var: #value type: #'sqInt' >
	pointer := self internalPopStackPointer.

	value := self uint32AtPointer: pointer.

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadUInt64FromMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer value |
	<var: #pointer type: #'char*' >
	<var: #value type: #'sqLong' >
	pointer := self internalPopStackPointer.

	value := self uint64AtPointer: pointer.

	self internalPushInt64: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLoadUInt8FromMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer value |
	<var: #pointer type: #'char*' >
	<var: #value type: #'sqInt' >
	pointer := self internalPopStackPointer.

	value := self uint8AtPointer: pointer.

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLocalFrameSize
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| alignedSize size |
	size := extA.

	"Align the size to 16 bytes."
	alignedSize := size + 15 bitAnd: -16.
	"Mark the frame"
	self setFrameHasNativeFrame: localFP.
	"Store the previous stack pointer"
	self nativePreviousStackPointerIn: localFP put: nativeStackPointer + 1.
	"Make the frame pointer"
	nativeStackPointer := nativeStackPointer - alignedSize.
	self nativeFramePointerIn: localFP put: nativeStackPointer + 1.
	"Set the stack pointer"
	nativeSP := nativeStackPointer + 1.
	self nativeStackPointerIn: localFP put: nativeStackPointer + 1.
	"Reserve space for the native stack"
	nativeStackPointer := nativeStackPointer - self defaultNativeStackFrameSize.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLockRegisters
	<option: #LowcodeVM>	"Lowcode instruction generator"

	"Nop in the interpreter"


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveLockVM
	<option: #LowcodeVM>	"Lowcode instruction generator"

	self abort.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveMalloc32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer size |
	<var: #pointer type: #'char*' >
	<var: #size type: #'sqInt' >
	size := self internalPopStackInt32.

	pointer := self malloc: size.

	self internalPushPointer: pointer.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveMalloc64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer size |
	<var: #pointer type: #'char*' >
	<var: #size type: #'sqLong' >
	size := self internalPopStackInt64.

	pointer := self malloc: size.

	self internalPushPointer: pointer.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveMemcpy32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| source dest size |
	<var: #source type: #'char*' >
	<var: #dest type: #'char*' >
	<var: #size type: #'sqInt' >
	size := self internalPopStackInt32.
	source := self internalPopStackPointer.
	dest := self internalPopStackPointer.

	self lowcode_mem: dest cp: source y: size.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveMemcpy64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| source dest size |
	<var: #source type: #'char*' >
	<var: #dest type: #'char*' >
	<var: #size type: #'sqLong' >
	size := self internalPopStackInt64.
	source := self internalPopStackPointer.
	dest := self internalPopStackPointer.

	self lowcode_mem: dest cp: source y: size.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveMemcpyFixed
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| source size dest |
	<var: #source type: #'char*' >
	<var: #dest type: #'char*' >
	size := extA.
	source := self internalPopStackPointer.
	dest := self internalPopStackPointer.

	self lowcode_mem: dest cp: source y: size.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveMoveFloat32ToPhysical
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| registerID value |
	<var: #value type: #'float' >
	registerID := extA.
	value := self internalPopStackFloat32.

	self lowcodeCalloutState: lowcodeCalloutState float32Register: registerID value: value.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveMoveFloat64ToPhysical
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| registerID value |
	<var: #value type: #'double' >
	registerID := extA.
	value := self internalPopStackFloat64.

	self lowcodeCalloutState: lowcodeCalloutState float64Register: registerID value: value.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveMoveInt32ToPhysical
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| registerID value |
	<var: #value type: #'sqInt' >
	registerID := extA.
	value := self internalPopStackInt32.

	self lowcodeCalloutState: lowcodeCalloutState int32Register: registerID value: value.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveMoveInt64ToPhysical
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| registerID value |
	<var: #value type: #'sqLong' >
	registerID := extA.
	value := self internalPopStackInt64.

	self lowcodeCalloutState: lowcodeCalloutState int64Register: registerID value: value.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveMovePointerToPhysical
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| registerID pointerValue |
	<var: #pointerValue type: #'char*' >
	registerID := extA.
	pointerValue := self internalPopStackPointer.

	self lowcodeCalloutState: lowcodeCalloutState pointerRegister: registerID value: pointerValue.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveMul32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqInt' >
	<var: #result type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	result := first * second.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveMul64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqLong' >
	<var: #result type: #'sqLong' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	result := first * second.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveNeg32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'sqInt' >
	value := self internalPopStackInt32.

	result := value negated.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveNeg64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqLong' >
	<var: #result type: #'sqLong' >
	value := self internalPopStackInt64.

	result := value negated.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveNot32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'sqInt' >
	value := self internalPopStackInt32.

	result := value bitXor: -1.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveNot64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqLong' >
	<var: #result type: #'sqLong' >
	value := self internalPopStackInt64.

	result := value bitXor: -1.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOopEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqInt' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStack.
	first := self internalPopStack.

	value := first = second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOopNotEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqInt' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStack.
	first := self internalPopStack.

	value := first ~= second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOopSmallIntegerToInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	object := self internalPopStack.

	value := objectMemory integerValueOf: object.

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOopSmallIntegerToInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqLong' >
	object := self internalPopStack.

	value := objectMemory integerValueOf: object.

	self internalPushInt64: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOopToBoolean32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	object := self internalPopStack.

	value := self booleanValueOf: object.

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOopToBoolean64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqLong' >
	object := self internalPopStack.

	value := self booleanValueOf: object.

	self internalPushInt64: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOopToFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'float' >
	object := self internalPopStack.

	value := objectMemory floatValueOf: object.

	self internalPushFloat32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOopToFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'double' >
	object := self internalPopStack.

	value := objectMemory floatValueOf: object.

	self internalPushFloat64: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOopToInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	object := self internalPopStack.

	value := self signed32BitValueOf: object.

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOopToInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqLong' >
	object := self internalPopStack.

	value := self signed64BitValueOf: object.

	self internalPushInt64: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOopToPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object pointer |
	<var: #object type: #'sqInt' >
	<var: #pointer type: #'char*' >
	object := self internalPopStack.

	pointer := self pointerAtPointer: (objectMemory firstIndexableField: object).

	self internalPushPointer: pointer.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOopToPointerReinterpret
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object pointer |
	<var: #object type: #'sqInt' >
	<var: #pointer type: #'char*' >
	object := self internalPopStack.

	pointer := self cCoerce: object to: 'char*'.

	self internalPushPointer: pointer.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOopToUInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	object := self internalPopStack.

	value := self positive32BitValueOf: object.

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOopToUInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqLong' >
	object := self internalPopStack.

	value := self positive64BitValueOf: object.

	self internalPushInt64: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOr32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqInt' >
	<var: #result type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	result := first bitOr: second.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveOr64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqLong' >
	<var: #result type: #'sqLong' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	result := first bitOr: second.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePerformCallout
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| function |
	function := extA.

	self lowcodeDoCallout: (self cCoerce: function to: #'char*').

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePerformCalloutIndirect
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| function |
	<var: #function type: #'char*' >
	function := self internalPopStackPointer.

	self lowcodeDoCallout: function.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePin
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object |
	<var: #object type: #'sqInt' >
	object := self internalPopStack.

	objectMemory pinObject: object.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePlaftormCode
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| code |
	<var: #code type: #'sqInt' >

	self abort.

	self internalPushInt32: code.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePointerAddConstantOffset
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| base offset result |
	<var: #base type: #'char*' >
	<var: #result type: #'char*' >
	offset := extB.
	base := self internalPopStackPointer.

	result := base + offset.

	self internalPushPointer: result.
	extB := 0.
	numExtB := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePointerAddOffset32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| base offset result |
	<var: #base type: #'char*' >
	<var: #offset type: #'sqInt' >
	<var: #result type: #'char*' >
	offset := self internalPopStackInt32.
	base := self internalPopStackPointer.

	result := base + offset.

	self internalPushPointer: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePointerAddOffset64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| base offset result |
	<var: #base type: #'char*' >
	<var: #offset type: #'sqLong' >
	<var: #result type: #'char*' >
	offset := self internalPopStackInt64.
	base := self internalPopStackPointer.

	result := base + offset.

	self internalPushPointer: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePointerEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'char*' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'char*' >
	second := self internalPopStackPointer.
	first := self internalPopStackPointer.

	value := first = second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePointerNotEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'char*' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'char*' >
	second := self internalPopStackPointer.
	first := self internalPopStackPointer.

	value := first ~= second ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: 'eem 11/11/2020 18:36'!
lowcodePrimitivePointerToInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer result |
	<var: #pointer type: #'char*' >
	<var: #result type: #'sqInt' >
	pointer := self internalPopStackPointer.

	result := self cCoerce: pointer to: 'usqIntptr_t'.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: 'eem 11/11/2020 18:37'!
lowcodePrimitivePointerToInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer result |
	<var: #pointer type: #'char*' >
	<var: #result type: #'sqLong' >
	pointer := self internalPopStackPointer.

	result := self cCoerce: pointer to: 'usqIntptr_t'.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePointerToOop
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer object pointerClassLiteral |
	<var: #pointer type: #'char*' >
	<var: #object type: #'sqInt' >
	pointerClassLiteral := self literal: extA.
	pointer := self internalPopStackPointer.

	object := objectMemory instantiateClass: pointerClassLiteral indexableSize: BytesPerWord.
	self pointerAtPointer: (objectMemory firstIndexableField: object) put: pointer.

	self internalPush: object.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePointerToOopReinterprer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer object |
	<var: #pointer type: #'char*' >
	<var: #object type: #'sqInt' >
	pointer := self internalPopStackPointer.

	object := self cCoerce: pointer to: 'sqInt'.

	self internalPush: object.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePopFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value |
	<var: #value type: #'float' >
	value := self internalPopStackFloat32.



! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePopFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value |
	<var: #value type: #'double' >
	value := self internalPopStackFloat64.



! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePopInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value |
	<var: #value type: #'sqInt' >
	value := self internalPopStackInt32.



! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePopInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value |
	<var: #value type: #'sqLong' >
	value := self internalPopStackInt64.



! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePopMultipleNative
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| popSize |
	popSize := extA.

	self internalPopStackNativeSize: popSize.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePopPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointerValue |
	<var: #pointerValue type: #'char*' >
	pointerValue := self internalPopStackPointer.



! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushCalloutResultFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result |
	<var: #result type: #'float' >

	result := self lowcodeCalloutStateFetchResultFloat32: lowcodeCalloutState.

	self internalPushFloat32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushCalloutResultFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result |
	<var: #result type: #'double' >

	result := self lowcodeCalloutStateFetchResultFloat64: lowcodeCalloutState.

	self internalPushFloat64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushCalloutResultInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result |
	<var: #result type: #'sqInt' >

	result := self lowcodeCalloutStateFetchResultInt32: lowcodeCalloutState.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushCalloutResultInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result |
	<var: #result type: #'sqLong' >

	result := self lowcodeCalloutStateFetchResultInt64: lowcodeCalloutState.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushCalloutResultPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result |
	<var: #result type: #'char*' >

	result := self lowcodeCalloutStateFetchResultPointer: lowcodeCalloutState.

	self internalPushPointer: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushConstantUInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result constant |
	<var: #result type: #'sqInt' >
	constant := extA.

	result := constant.

	self internalPushInt32: result.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushConstantUInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result constant |
	<var: #result type: #'sqLong' >
	constant := extA.

	result := constant.

	self internalPushInt64: result.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushNullPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result |
	<var: #result type: #'char*' >

	result := 0.

	self internalPushPointer: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushOne32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result |
	<var: #result type: #'sqInt' >

	result := 1.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushOne64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result |
	<var: #result type: #'sqLong' >

	result := 1.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushOneFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result |
	<var: #result type: #'float' >

	result := 1.0.

	self internalPushFloat32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushOneFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result |
	<var: #result type: #'double' >

	result := 1.0.

	self internalPushFloat64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushPhysicalFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| registerID value |
	<var: #value type: #'float' >
	registerID := extA.

	value := self lowcodeCalloutState: lowcodeCalloutState float32Register: registerID.

	self internalPushFloat32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushPhysicalFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| registerID value |
	<var: #value type: #'double' >
	registerID := extA.

	value := self lowcodeCalloutState: lowcodeCalloutState float64Register: registerID.

	self internalPushFloat64: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushPhysicalInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| registerID value |
	<var: #value type: #'sqInt' >
	registerID := extA.

	value := self lowcodeCalloutState: lowcodeCalloutState int32Register: registerID.

	self internalPushInt32: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushPhysicalInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| registerID value |
	<var: #value type: #'sqLong' >
	registerID := extA.

	value := self lowcodeCalloutState: lowcodeCalloutState int64Register: registerID.

	self internalPushInt64: value.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushPhysicalPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| registerID pointerValue |
	<var: #pointerValue type: #'char*' >
	registerID := extA.

	pointerValue := self lowcodeCalloutState: lowcodeCalloutState pointerRegister: registerID.

	self internalPushPointer: pointerValue.
	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushSessionIdentifier
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value |
	<var: #value type: #'sqInt' >

	value := self getThisSessionID.

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushZero32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result |
	<var: #result type: #'sqInt' >

	result := 0.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushZero64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result |
	<var: #result type: #'sqLong' >

	result := 0.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushZeroFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result |
	<var: #result type: #'float' >

	result := 0.0.

	self internalPushFloat32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitivePushZeroFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| result |
	<var: #result type: #'double' >

	result := 0.0.

	self internalPushFloat64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveRem32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqInt' >
	<var: #result type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	result := first \\ second.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveRem64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqLong' >
	<var: #result type: #'sqLong' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	result := first \\ second.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveRightShift32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value shiftAmount result |
	<var: #value type: #'sqInt' >
	<var: #shiftAmount type: #'sqInt' >
	<var: #result type: #'sqInt' >
	shiftAmount := self internalPopStackInt32.
	value := self internalPopStackInt32.

	result := value >> shiftAmount.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveRightShift64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value shiftAmount result |
	<var: #value type: #'sqLong' >
	<var: #shiftAmount type: #'sqLong' >
	<var: #result type: #'sqLong' >
	shiftAmount := self internalPopStackInt64.
	value := self internalPopStackInt64.

	result := value >> shiftAmount.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveSignExtend32From16
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'sqInt' >
	value := self internalPopStackInt32.

	result := self cCoerce: value to: 'signed short'.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveSignExtend32From8
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'sqInt' >
	value := self internalPopStackInt32.

	result := self cCoerce: value to: 'signed char'.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveSignExtend64From16
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqLong' >
	<var: #result type: #'sqLong' >
	value := self internalPopStackInt64.

	result := self cCoerce: value to: 'int16_t'.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveSignExtend64From32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'sqLong' >
	value := self internalPopStackInt32.

	result := self cCoerce: value to: 'int32_t'.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveSignExtend64From8
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqLong' >
	<var: #result type: #'sqLong' >
	value := self internalPopStackInt64.

	result := self cCoerce: value to: 'signed char'.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveSmallInt32ToOop
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	value := self internalPopStackInt32.

	object := objectMemory integerObjectOf: value.

	self internalPush: object.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreFloat32ToMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| floatValue pointer |
	<var: #floatValue type: #'float' >
	<var: #pointer type: #'char*' >
	pointer := self internalPopStackPointer.
	floatValue := self internalPopStackFloat32.

	self singleFloatAtPointer: pointer put: floatValue.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreFloat64ToMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| doubleValue pointer |
	<var: #doubleValue type: #'double' >
	<var: #pointer type: #'char*' >
	pointer := self internalPopStackPointer.
	doubleValue := self internalPopStackFloat64.

	self floatAtPointer: pointer put: doubleValue.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreInt16ToMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer value |
	<var: #pointer type: #'char*' >
	<var: #value type: #'sqInt' >
	pointer := self internalPopStackPointer.
	value := self internalPopStackInt32.

	self int16AtPointer: pointer put: value.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreInt32ToMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer value |
	<var: #pointer type: #'char*' >
	<var: #value type: #'sqInt' >
	pointer := self internalPopStackPointer.
	value := self internalPopStackInt32.

	self int32AtPointer: pointer put: value.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreInt64ToMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer value |
	<var: #pointer type: #'char*' >
	<var: #value type: #'sqLong' >
	pointer := self internalPopStackPointer.
	value := self internalPopStackInt64.

	self int64AtPointer: pointer put: value.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreInt8ToMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointer value |
	<var: #pointer type: #'char*' >
	<var: #value type: #'sqInt' >
	pointer := self internalPopStackPointer.
	value := self internalPopStackInt32.

	self int8AtPointer: pointer put: value.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreLocalFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'float' >
	baseOffset := extA.
	value := self internalPopStackFloat32.

	stackPages singleFloatAtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP) put: value.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreLocalFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'double' >
	baseOffset := extA.
	value := self internalPopStackFloat64.

	stackPages floatAtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP) put: value.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreLocalInt16
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.
	value := self internalPopStackInt32.

	stackPages int16AtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP) put: value.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreLocalInt32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.
	value := self internalPopStackInt32.

	stackPages int32AtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP) put: value.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreLocalInt64
	<option: #LowcodeVM>	"Lowcode instruction generator"

	<var: #valueInt64 type: #'sqLong'>
	|valueInt64|
	BytesPerWord = 4 ifTrue: [
	self lowcodeStoreLocalInt64Workaround: extA in: localFP sp: localSP.
	] ifFalse: [
	valueInt64 := self internalPopStackInt64.
	stackPages int64AtPointer: (self framePointerOfNativeLocal: extA in: localFP) put: valueInt64.
	].
	extA := 0.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreLocalInt8
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value baseOffset |
	<var: #value type: #'sqInt' >
	baseOffset := extA.
	value := self internalPopStackInt32.

	stackPages int8AtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP) put: value.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreLocalPointer
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| pointerValue baseOffset |
	<var: #pointerValue type: #'char*' >
	baseOffset := extA.
	pointerValue := self internalPopStackPointer.

	stackPages pointerAtPointer: (self framePointerOfNativeLocal: baseOffset in: localFP) put: pointerValue.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreObjectField
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| fieldIndex object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	fieldIndex := extA.
	value := self internalPopStack.
	object := self internalPopStack.

	objectMemory storePointer: fieldIndex ofObject: object withValue: value.

	extA := 0.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStoreObjectFieldAt
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| fieldIndex object value |
	<var: #fieldIndex type: #'sqInt' >
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	value := self internalPopStack.
	fieldIndex := self internalPopStackInt32.
	object := self internalPopStack.

	objectMemory storePointer: fieldIndex ofObject: object withValue: value.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveStorePointerToMemory
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| memoryPointer pointerValue |
	<var: #memoryPointer type: #'char*' >
	<var: #pointerValue type: #'char*' >
	memoryPointer := self internalPopStackPointer.
	pointerValue := self internalPopStackPointer.

	self pointerAtPointer: memoryPointer put: pointerValue.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveSub32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqInt' >
	<var: #result type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	result := first - second.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveSub64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqLong' >
	<var: #result type: #'sqLong' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	result := first - second.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveTruncate32To16
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'sqInt' >
	value := self internalPopStackInt32.

	result := value bitAnd: 16rFFFF.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveTruncate32To8
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'sqInt' >
	value := self internalPopStackInt32.

	result := value bitAnd: 16rFF.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveTruncate64To16
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqLong' >
	<var: #result type: #'sqInt' >
	value := self internalPopStackInt64.

	result := value bitAnd: 16rFFFF.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveTruncate64To32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqLong' >
	<var: #result type: #'sqInt' >
	value := self internalPopStackInt64.

	result := value bitAnd: 16rFFFFFFFF.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveTruncate64To8
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqLong' >
	<var: #result type: #'sqInt' >
	value := self internalPopStackInt64.

	result := value bitAnd: 16rFF.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUdiv32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqInt' >
	<var: #result type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	result := (self cCoerce: first to: 'unsigned int') //
	(self cCoerce: second to: 'unsigned int').

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUdiv64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqLong' >
	<var: #result type: #'sqLong' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	result := (self cCoerce: first to: 'uint64_t') //
	(self cCoerce: second to: 'uint64_t').

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUint32Great
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqInt' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	value := (self cCoerce: first to: 'unsigned int') >
	(self cCoerce: second to: 'unsigned int') ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUint32GreatEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqInt' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	value := (self cCoerce: first to: 'unsigned int') >=
	(self cCoerce: second to: 'unsigned int') ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUint32Less
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqInt' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	value := (self cCoerce: first to: 'unsigned int') <
	(self cCoerce: second to: 'unsigned int') ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUint32LessEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqInt' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	value := (self cCoerce: first to: 'unsigned int') <=
	(self cCoerce: second to: 'unsigned int') ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUint32ToFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'float' >
	value := self internalPopStackInt32.

	result := self cCoerce: (self cCoerce: value to: 'unsigned int') to: 'float'.

	self internalPushFloat32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUint32ToFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'double' >
	value := self internalPopStackInt32.

	result := self cCoerce: (self cCoerce: value to: 'unsigned int') to: 'double'.

	self internalPushFloat64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUint32ToOop
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqInt' >
	value := self internalPopStackInt32.

	object := objectMemory positive32BitIntegerFor: value.

	self internalPush: object.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUint64Great
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqLong' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	value := (self cCoerce: first to: 'uint64_t') >
	(self cCoerce: second to: 'uint64_t') ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUint64GreatEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqLong' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	value := (self cCoerce: first to: 'uint64_t') >=
	(self cCoerce: second to: 'uint64_t') ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUint64Less
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqLong' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	value := (self cCoerce: first to: 'uint64_t') <
	(self cCoerce: second to: 'uint64_t') ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUint64LessEqual
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second value first |
	<var: #second type: #'sqLong' >
	<var: #value type: #'sqInt' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	value := (self cCoerce: first to: 'uint64_t') <=
	(self cCoerce: second to: 'uint64_t') ifTrue: [1] ifFalse: [0].

	self internalPushInt32: value.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUint64ToFloat32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqLong' >
	<var: #result type: #'float' >
	value := self internalPopStackInt64.

	result := self cCoerce: (self cCoerce: value to: 'uint64_t') to: 'float'.

	self internalPushFloat32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUint64ToFloat64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqLong' >
	<var: #result type: #'double' >
	value := self internalPopStackInt64.

	result := self cCoerce: (self cCoerce: value to: 'uint64_t') to: 'double'.

	self internalPushFloat64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUint64ToOop
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object value |
	<var: #object type: #'sqInt' >
	<var: #value type: #'sqLong' >
	value := self internalPopStackInt64.

	object := self positive64BitIntegerFor: value.

	self internalPush: object.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUmul32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqInt' >
	<var: #result type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	result := (self cCoerce: first to: 'unsigned int') *
	(self cCoerce: second to: 'unsigned int').

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUmul64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqLong' >
	<var: #result type: #'sqLong' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	result := (self cCoerce: first to: 'uint64_t') *
	(self cCoerce: second to: 'uint64_t').

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUnlockRegisters
	<option: #LowcodeVM>	"Lowcode instruction generator"

	"Nop in the interpreter"


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUnlockVM
	<option: #LowcodeVM>	"Lowcode instruction generator"

	self abort.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUnpin
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| object |
	<var: #object type: #'sqInt' >
	object := self internalPopStack.

	objectMemory unpinObject: object.


! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUrem32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqInt' >
	<var: #result type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	result := (self cCoerce: first to: 'unsigned int') \\
	(self cCoerce: second to: 'unsigned int').

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveUrem64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqLong' >
	<var: #result type: #'sqLong' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	result := (self cCoerce: first to: 'unsigned int') \\
	(self cCoerce: second to: 'unsigned int').

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveXor32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqInt' >
	<var: #result type: #'sqInt' >
	<var: #first type: #'sqInt' >
	second := self internalPopStackInt32.
	first := self internalPopStackInt32.

	result := first bitXor: second.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveXor64
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| second result first |
	<var: #second type: #'sqLong' >
	<var: #result type: #'sqLong' >
	<var: #first type: #'sqLong' >
	second := self internalPopStackInt64.
	first := self internalPopStackInt64.

	result := first bitXor: second.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveZeroExtend32From16
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'sqInt' >
	value := self internalPopStackInt32.

	result := self cCoerce: value to: 'uint16_t'.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveZeroExtend32From8
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'sqInt' >
	value := self internalPopStackInt32.

	result := self cCoerce: value to: 'uint8_t'.

	self internalPushInt32: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveZeroExtend64From16
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqLong' >
	<var: #result type: #'sqLong' >
	value := self internalPopStackInt64.

	result := self cCoerce: value to: 'uint16_t'.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveZeroExtend64From32
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqInt' >
	<var: #result type: #'sqLong' >
	value := self internalPopStackInt32.

	result := self cCoerce: value to: 'uint32_t'.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive generated code' stamp: ''!
lowcodePrimitiveZeroExtend64From8
	<option: #LowcodeVM>	"Lowcode instruction generator"
	| value result |
	<var: #value type: #'sqLong' >
	<var: #result type: #'sqLong' >
	value := self internalPopStackInt64.

	result := self cCoerce: value to: 'uint8_t'.

	self internalPushInt64: result.

! !

!StackInterpreter methodsFor: 'inline primitive support' stamp: 'rsf 8/9/2016 23:37'!
lowcodeStoreLocalInt64Workaround: baseOffset in: theFP sp: theSP
	"This method is a workaround a GCC bug"
	<inline: #never>
	<option: #LowcodeVM>
	<var: #theFP type: #'char*'>
	<var: #theSP type: #'char*'>
	<var: #value type: #'sqLong'>
	| value workNativeSP|

	"Fetch the value from the native tack"
	workNativeSP := (self nativeStackPointerIn: theFP).
	value := stackPages long64AtPointer: (workNativeSP - 1).
	self nativeStackPointerIn: theFP put: workNativeSP + 8.

	"Store the local"
 	self int64AtPointer: (self framePointerOfNativeLocal: baseOffset in: theFP) put: value.! !

!StackInterpreter methodsFor: 'inline primitive dispatch generated code' stamp: ''!
lowcodeTrinaryInlinePrimitive: prim
	<option: #LowcodeVM>	"Lowcode instruction interpreter dispatch"
	prim
		caseOf: {
			[0]	-> [ ^ self lowcodePrimitiveOopEqual ].
			[1]	-> [ ^ self lowcodePrimitiveOopNotEqual ].
			[2]	-> [ ^ self lowcodePrimitiveStoreObjectField ].
			[3]	-> [ ^ self lowcodePrimitiveStoreObjectFieldAt ].
		}
		otherwise: [
			localIP := localIP - 3.
			^self respondToUnknownBytecode
		].
	

! !

!StackInterpreter methodsFor: 'inline primitive dispatch generated code' stamp: ''!
lowcodeUnaryInlinePrimitive2: prim
	<option: #LowcodeVM>	"Lowcode instruction interpreter dispatch"
	prim
		caseOf: {
			[60]	-> [ ^ self lowcodePrimitiveFloat64ToFloat32 ].
			[61]	-> [ ^ self lowcodePrimitiveFloat64ToInt32 ].
			[62]	-> [ ^ self lowcodePrimitiveFloat64ToInt64 ].
			[63]	-> [ ^ self lowcodePrimitiveFloat64ToUInt32 ].
			[64]	-> [ ^ self lowcodePrimitiveFloat64ToUInt64 ].
			[65]	-> [ ^ self lowcodePrimitiveFree ].
			[66]	-> [ ^ self lowcodePrimitiveInstantiateIndexable32Oop ].
			[67]	-> [ ^ self lowcodePrimitiveInstantiateIndexableOop ].
			[68]	-> [ ^ self lowcodePrimitiveInstantiateOop ].
			[69]	-> [ ^ self lowcodePrimitiveInt32Equal ].
			[70]	-> [ ^ self lowcodePrimitiveInt32Great ].
			[71]	-> [ ^ self lowcodePrimitiveInt32GreatEqual ].
			[72]	-> [ ^ self lowcodePrimitiveInt32Less ].
			[73]	-> [ ^ self lowcodePrimitiveInt32LessEqual ].
			[74]	-> [ ^ self lowcodePrimitiveInt32NotEqual ].
			[75]	-> [ ^ self lowcodePrimitiveInt32ToFloat32 ].
			[76]	-> [ ^ self lowcodePrimitiveInt32ToFloat64 ].
			[77]	-> [ ^ self lowcodePrimitiveInt32ToPointer ].
			[78]	-> [ ^ self lowcodePrimitiveInt64Equal ].
			[79]	-> [ ^ self lowcodePrimitiveInt64Great ].
			[80]	-> [ ^ self lowcodePrimitiveInt64GreatEqual ].
			[81]	-> [ ^ self lowcodePrimitiveInt64Less ].
			[82]	-> [ ^ self lowcodePrimitiveInt64LessEqual ].
			[83]	-> [ ^ self lowcodePrimitiveInt64NotEqual ].
			[84]	-> [ ^ self lowcodePrimitiveInt64ToFloat32 ].
			[85]	-> [ ^ self lowcodePrimitiveInt64ToFloat64 ].
			[86]	-> [ ^ self lowcodePrimitiveInt64ToPointer ].
			[87]	-> [ ^ self lowcodePrimitiveLeftShift32 ].
			[88]	-> [ ^ self lowcodePrimitiveLeftShift64 ].
			[89]	-> [ ^ self lowcodePrimitiveLoadArgumentAddress ].
			[90]	-> [ ^ self lowcodePrimitiveLoadArgumentFloat32 ].
			[91]	-> [ ^ self lowcodePrimitiveLoadArgumentFloat64 ].
			[92]	-> [ ^ self lowcodePrimitiveLoadArgumentInt16 ].
			[93]	-> [ ^ self lowcodePrimitiveLoadArgumentInt32 ].
			[94]	-> [ ^ self lowcodePrimitiveLoadArgumentInt64 ].
			[95]	-> [ ^ self lowcodePrimitiveLoadArgumentInt8 ].
			[96]	-> [ ^ self lowcodePrimitiveLoadArgumentPointer ].
			[97]	-> [ ^ self lowcodePrimitiveLoadArgumentUInt16 ].
			[98]	-> [ ^ self lowcodePrimitiveLoadArgumentUInt32 ].
			[99]	-> [ ^ self lowcodePrimitiveLoadArgumentUInt64 ].
			[100]	-> [ ^ self lowcodePrimitiveLoadArgumentUInt8 ].
			[101]	-> [ ^ self lowcodePrimitiveLoadFloat32FromMemory ].
			[102]	-> [ ^ self lowcodePrimitiveLoadFloat64FromMemory ].
			[103]	-> [ ^ self lowcodePrimitiveLoadInt16FromMemory ].
			[104]	-> [ ^ self lowcodePrimitiveLoadInt32FromMemory ].
			[105]	-> [ ^ self lowcodePrimitiveLoadInt64FromMemory ].
			[106]	-> [ ^ self lowcodePrimitiveLoadInt8FromMemory ].
			[107]	-> [ ^ self lowcodePrimitiveLoadLocalAddress ].
			[108]	-> [ ^ self lowcodePrimitiveLoadLocalFloat32 ].
			[109]	-> [ ^ self lowcodePrimitiveLoadLocalFloat64 ].
			[110]	-> [ ^ self lowcodePrimitiveLoadLocalInt16 ].
			[111]	-> [ ^ self lowcodePrimitiveLoadLocalInt32 ].
			[112]	-> [ ^ self lowcodePrimitiveLoadLocalInt64 ].
			[113]	-> [ ^ self lowcodePrimitiveLoadLocalInt8 ].
			[114]	-> [ ^ self lowcodePrimitiveLoadLocalPointer ].
			[115]	-> [ ^ self lowcodePrimitiveLoadLocalUInt16 ].
			[116]	-> [ ^ self lowcodePrimitiveLoadLocalUInt32 ].
			[117]	-> [ ^ self lowcodePrimitiveLoadLocalUInt64 ].
			[118]	-> [ ^ self lowcodePrimitiveLoadLocalUInt8 ].
			[119]	-> [ ^ self lowcodePrimitiveLoadObjectAt ].
		}
		otherwise: [ ^ self lowcodeUnaryInlinePrimitive3: prim ].
	

! !

!StackInterpreter methodsFor: 'inline primitive dispatch generated code' stamp: ''!
lowcodeUnaryInlinePrimitive3: prim
	<option: #LowcodeVM>	"Lowcode instruction interpreter dispatch"
	prim
		caseOf: {
			[120]	-> [ ^ self lowcodePrimitiveLoadObjectField ].
			[121]	-> [ ^ self lowcodePrimitiveLoadPointerFromMemory ].
			[122]	-> [ ^ self lowcodePrimitiveLoadUInt16FromMemory ].
			[123]	-> [ ^ self lowcodePrimitiveLoadUInt32FromMemory ].
			[124]	-> [ ^ self lowcodePrimitiveLoadUInt64FromMemory ].
			[125]	-> [ ^ self lowcodePrimitiveLoadUInt8FromMemory ].
			[126]	-> [ ^ self lowcodePrimitiveLocalFrameSize ].
			[127]	-> [ ^ self lowcodePrimitiveLockRegisters ].
			[128]	-> [ ^ self lowcodePrimitiveLockVM ].
			[129]	-> [ ^ self lowcodePrimitiveMalloc32 ].
			[130]	-> [ ^ self lowcodePrimitiveMalloc64 ].
			[131]	-> [ ^ self lowcodePrimitiveMemcpy32 ].
			[132]	-> [ ^ self lowcodePrimitiveMemcpy64 ].
			[133]	-> [ ^ self lowcodePrimitiveMemcpyFixed ].
			[134]	-> [ ^ self lowcodePrimitiveMoveFloat32ToPhysical ].
			[135]	-> [ ^ self lowcodePrimitiveMoveFloat64ToPhysical ].
			[136]	-> [ ^ self lowcodePrimitiveMoveInt32ToPhysical ].
			[137]	-> [ ^ self lowcodePrimitiveMoveInt64ToPhysical ].
			[138]	-> [ ^ self lowcodePrimitiveMovePointerToPhysical ].
			[139]	-> [ ^ self lowcodePrimitiveMul32 ].
			[140]	-> [ ^ self lowcodePrimitiveMul64 ].
			[141]	-> [ ^ self lowcodePrimitiveNeg32 ].
			[142]	-> [ ^ self lowcodePrimitiveNeg64 ].
			[143]	-> [ ^ self lowcodePrimitiveNot32 ].
			[144]	-> [ ^ self lowcodePrimitiveNot64 ].
			[145]	-> [ ^ self lowcodePrimitiveOr32 ].
			[146]	-> [ ^ self lowcodePrimitiveOr64 ].
			[147]	-> [ ^ self lowcodePrimitivePerformCallout ].
			[148]	-> [ ^ self lowcodePrimitivePerformCalloutIndirect ].
			[149]	-> [ ^ self lowcodePrimitivePushCalloutResultFloat32 ].
			[150]	-> [ ^ self lowcodePrimitivePushCalloutResultFloat64 ].
			[151]	-> [ ^ self lowcodePrimitivePushCalloutResultInt32 ].
			[152]	-> [ ^ self lowcodePrimitivePushCalloutResultInt64 ].
			[153]	-> [ ^ self lowcodePrimitivePushCalloutResultPointer ].
			[161]	-> [ ^ self lowcodePrimitivePlaftormCode ].
			[162]	-> [ ^ self lowcodePrimitivePointerAddConstantOffset ].
			[163]	-> [ ^ self lowcodePrimitivePointerAddOffset32 ].
			[164]	-> [ ^ self lowcodePrimitivePointerAddOffset64 ].
			[165]	-> [ ^ self lowcodePrimitivePointerEqual ].
			[166]	-> [ ^ self lowcodePrimitivePointerNotEqual ].
			[167]	-> [ ^ self lowcodePrimitivePointerToInt32 ].
			[168]	-> [ ^ self lowcodePrimitivePointerToInt64 ].
			[169]	-> [ ^ self lowcodePrimitivePopFloat32 ].
			[170]	-> [ ^ self lowcodePrimitivePopFloat64 ].
			[171]	-> [ ^ self lowcodePrimitivePopInt32 ].
			[172]	-> [ ^ self lowcodePrimitivePopInt64 ].
			[173]	-> [ ^ self lowcodePrimitivePopMultipleNative ].
			[174]	-> [ ^ self lowcodePrimitivePopPointer ].
			[175]	-> [ ^ self lowcodePrimitivePushConstantUInt32 ].
			[176]	-> [ ^ self lowcodePrimitivePushConstantUInt64 ].
			[177]	-> [ ^ self lowcodePrimitivePushNullPointer ].
			[178]	-> [ ^ self lowcodePrimitivePushOne32 ].
			[179]	-> [ ^ self lowcodePrimitivePushOne64 ].
			[180]	-> [ ^ self lowcodePrimitivePushOneFloat32 ].
			[181]	-> [ ^ self lowcodePrimitivePushOneFloat64 ].
			[182]	-> [ ^ self lowcodePrimitivePushPhysicalFloat32 ].
			[183]	-> [ ^ self lowcodePrimitivePushPhysicalFloat64 ].
			[184]	-> [ ^ self lowcodePrimitivePushPhysicalInt32 ].
			[185]	-> [ ^ self lowcodePrimitivePushPhysicalInt64 ].
			[186]	-> [ ^ self lowcodePrimitivePushPhysicalPointer ].
		}
		otherwise: [ ^ self lowcodeUnaryInlinePrimitive4: prim ].
	

! !

!StackInterpreter methodsFor: 'inline primitive dispatch generated code' stamp: ''!
lowcodeUnaryInlinePrimitive4: prim
	<option: #LowcodeVM>	"Lowcode instruction interpreter dispatch"
	prim
		caseOf: {
			[187]	-> [ ^ self lowcodePrimitivePushSessionIdentifier ].
			[188]	-> [ ^ self lowcodePrimitivePushZero32 ].
			[189]	-> [ ^ self lowcodePrimitivePushZero64 ].
			[190]	-> [ ^ self lowcodePrimitivePushZeroFloat32 ].
			[191]	-> [ ^ self lowcodePrimitivePushZeroFloat64 ].
			[192]	-> [ ^ self lowcodePrimitiveRem32 ].
			[193]	-> [ ^ self lowcodePrimitiveRem64 ].
			[194]	-> [ ^ self lowcodePrimitiveRightShift32 ].
			[195]	-> [ ^ self lowcodePrimitiveRightShift64 ].
			[196]	-> [ ^ self lowcodePrimitiveSignExtend32From16 ].
			[197]	-> [ ^ self lowcodePrimitiveSignExtend32From8 ].
			[198]	-> [ ^ self lowcodePrimitiveSignExtend64From16 ].
			[199]	-> [ ^ self lowcodePrimitiveSignExtend64From32 ].
			[200]	-> [ ^ self lowcodePrimitiveSignExtend64From8 ].
			[201]	-> [ ^ self lowcodePrimitiveStoreFloat32ToMemory ].
			[202]	-> [ ^ self lowcodePrimitiveStoreFloat64ToMemory ].
			[203]	-> [ ^ self lowcodePrimitiveStoreInt16ToMemory ].
			[204]	-> [ ^ self lowcodePrimitiveStoreInt32ToMemory ].
			[205]	-> [ ^ self lowcodePrimitiveStoreInt64ToMemory ].
			[206]	-> [ ^ self lowcodePrimitiveStoreInt8ToMemory ].
			[207]	-> [ ^ self lowcodePrimitiveStoreLocalFloat32 ].
			[208]	-> [ ^ self lowcodePrimitiveStoreLocalFloat64 ].
			[209]	-> [ ^ self lowcodePrimitiveStoreLocalInt16 ].
			[210]	-> [ ^ self lowcodePrimitiveStoreLocalInt32 ].
			[211]	-> [ ^ self lowcodePrimitiveStoreLocalInt64 ].
			[212]	-> [ ^ self lowcodePrimitiveStoreLocalInt8 ].
			[213]	-> [ ^ self lowcodePrimitiveStoreLocalPointer ].
			[214]	-> [ ^ self lowcodePrimitiveStorePointerToMemory ].
			[215]	-> [ ^ self lowcodePrimitiveSub32 ].
			[216]	-> [ ^ self lowcodePrimitiveSub64 ].
			[217]	-> [ ^ self lowcodePrimitiveTruncate32To16 ].
			[218]	-> [ ^ self lowcodePrimitiveTruncate32To8 ].
			[219]	-> [ ^ self lowcodePrimitiveTruncate64To16 ].
			[220]	-> [ ^ self lowcodePrimitiveTruncate64To32 ].
			[221]	-> [ ^ self lowcodePrimitiveTruncate64To8 ].
			[222]	-> [ ^ self lowcodePrimitiveUdiv32 ].
			[223]	-> [ ^ self lowcodePrimitiveUdiv64 ].
			[224]	-> [ ^ self lowcodePrimitiveUint32Great ].
			[225]	-> [ ^ self lowcodePrimitiveUint32GreatEqual ].
			[226]	-> [ ^ self lowcodePrimitiveUint32Less ].
			[227]	-> [ ^ self lowcodePrimitiveUint32LessEqual ].
			[228]	-> [ ^ self lowcodePrimitiveUint32ToFloat32 ].
			[229]	-> [ ^ self lowcodePrimitiveUint32ToFloat64 ].
			[230]	-> [ ^ self lowcodePrimitiveUint64Great ].
			[231]	-> [ ^ self lowcodePrimitiveUint64GreatEqual ].
			[232]	-> [ ^ self lowcodePrimitiveUint64Less ].
			[233]	-> [ ^ self lowcodePrimitiveUint64LessEqual ].
			[234]	-> [ ^ self lowcodePrimitiveUint64ToFloat32 ].
			[235]	-> [ ^ self lowcodePrimitiveUint64ToFloat64 ].
			[236]	-> [ ^ self lowcodePrimitiveUmul32 ].
			[237]	-> [ ^ self lowcodePrimitiveUmul64 ].
			[238]	-> [ ^ self lowcodePrimitiveUnlockRegisters ].
			[239]	-> [ ^ self lowcodePrimitiveUnlockVM ].
			[240]	-> [ ^ self lowcodePrimitiveUrem32 ].
			[241]	-> [ ^ self lowcodePrimitiveUrem64 ].
			[242]	-> [ ^ self lowcodePrimitiveXor32 ].
			[243]	-> [ ^ self lowcodePrimitiveXor64 ].
			[244]	-> [ ^ self lowcodePrimitiveZeroExtend32From16 ].
			[245]	-> [ ^ self lowcodePrimitiveZeroExtend32From8 ].
			[246]	-> [ ^ self lowcodePrimitiveZeroExtend64From16 ].
		}
		otherwise: [ ^ self lowcodeUnaryInlinePrimitive5: prim ].
	

! !

!StackInterpreter methodsFor: 'inline primitive dispatch generated code' stamp: ''!
lowcodeUnaryInlinePrimitive5: prim
	<option: #LowcodeVM>	"Lowcode instruction interpreter dispatch"
	prim
		caseOf: {
			[247]	-> [ ^ self lowcodePrimitiveZeroExtend64From32 ].
			[248]	-> [ ^ self lowcodePrimitiveZeroExtend64From8 ].
		}
		otherwise: [
			localIP := localIP - 3.
			^self respondToUnknownBytecode
		].
	

! !

!StackInterpreter methodsFor: 'inline primitive dispatch generated code' stamp: ''!
lowcodeUnaryInlinePrimitive: prim
	<option: #LowcodeVM>	"Lowcode instruction interpreter dispatch"
	prim
		caseOf: {
			[0]	-> [ ^ self lowcodePrimitiveAdd32 ].
			[1]	-> [ ^ self lowcodePrimitiveAdd64 ].
			[2]	-> [ ^ self lowcodePrimitiveAlloca32 ].
			[3]	-> [ ^ self lowcodePrimitiveAlloca64 ].
			[4]	-> [ ^ self lowcodePrimitiveAnd32 ].
			[5]	-> [ ^ self lowcodePrimitiveAnd64 ].
			[6]	-> [ ^ self lowcodePrimitiveArithmeticRightShift32 ].
			[7]	-> [ ^ self lowcodePrimitiveArithmeticRightShift64 ].
			[8]	-> [ ^ self lowcodePrimitiveBeginCall ].
			[9]	-> [ ^ self lowcodePrimitiveCallArgumentFloat32 ].
			[10]	-> [ ^ self lowcodePrimitiveCallArgumentFloat64 ].
			[11]	-> [ ^ self lowcodePrimitiveCallArgumentInt32 ].
			[12]	-> [ ^ self lowcodePrimitiveCallArgumentInt64 ].
			[13]	-> [ ^ self lowcodePrimitiveCallArgumentPointer ].
			[14]	-> [ ^ self lowcodePrimitiveCallArgumentSpace ].
			[15]	-> [ ^ self lowcodePrimitiveCallArgumentStructure ].
			[16]	-> [ ^ self lowcodePrimitiveCallInstruction ].
			[17]	-> [ ^ self lowcodePrimitiveCallPhysical ].
			[18]	-> [ ^ self lowcodePrimitiveCheckSessionIdentifier ].
			[19]	-> [ ^ self lowcodePrimitiveCompareAndSwap32 ].
			[20]	-> [ ^ self lowcodePrimitiveDiv32 ].
			[21]	-> [ ^ self lowcodePrimitiveDiv64 ].
			[22]	-> [ ^ self lowcodePrimitiveDuplicateFloat32 ].
			[23]	-> [ ^ self lowcodePrimitiveDuplicateFloat64 ].
			[24]	-> [ ^ self lowcodePrimitiveDuplicateInt32 ].
			[25]	-> [ ^ self lowcodePrimitiveDuplicateInt64 ].
			[26]	-> [ ^ self lowcodePrimitiveDuplicatePointer ].
			[27]	-> [ ^ self lowcodePrimitiveEffectiveAddress32 ].
			[28]	-> [ ^ self lowcodePrimitiveEffectiveAddress64 ].
			[29]	-> [ ^ self lowcodePrimitiveEndCall ].
			[30]	-> [ ^ self lowcodePrimitiveEndCallNoCleanup ].
			[31]	-> [ ^ self lowcodePrimitiveFloat32Add ].
			[32]	-> [ ^ self lowcodePrimitiveFloat32Div ].
			[33]	-> [ ^ self lowcodePrimitiveFloat32Equal ].
			[34]	-> [ ^ self lowcodePrimitiveFloat32Great ].
			[35]	-> [ ^ self lowcodePrimitiveFloat32GreatEqual ].
			[36]	-> [ ^ self lowcodePrimitiveFloat32Less ].
			[37]	-> [ ^ self lowcodePrimitiveFloat32LessEqual ].
			[38]	-> [ ^ self lowcodePrimitiveFloat32Mul ].
			[39]	-> [ ^ self lowcodePrimitiveFloat32Neg ].
			[40]	-> [ ^ self lowcodePrimitiveFloat32NotEqual ].
			[41]	-> [ ^ self lowcodePrimitiveFloat32Sqrt ].
			[42]	-> [ ^ self lowcodePrimitiveFloat32Sub ].
			[43]	-> [ ^ self lowcodePrimitiveFloat32ToFloat64 ].
			[44]	-> [ ^ self lowcodePrimitiveFloat32ToInt32 ].
			[45]	-> [ ^ self lowcodePrimitiveFloat32ToInt64 ].
			[46]	-> [ ^ self lowcodePrimitiveFloat32ToUInt32 ].
			[47]	-> [ ^ self lowcodePrimitiveFloat32ToUInt64 ].
			[48]	-> [ ^ self lowcodePrimitiveFloat64Add ].
			[49]	-> [ ^ self lowcodePrimitiveFloat64Div ].
			[50]	-> [ ^ self lowcodePrimitiveFloat64Equal ].
			[51]	-> [ ^ self lowcodePrimitiveFloat64Great ].
			[52]	-> [ ^ self lowcodePrimitiveFloat64GreatEqual ].
			[53]	-> [ ^ self lowcodePrimitiveFloat64Less ].
			[54]	-> [ ^ self lowcodePrimitiveFloat64LessEqual ].
			[55]	-> [ ^ self lowcodePrimitiveFloat64Mul ].
			[56]	-> [ ^ self lowcodePrimitiveFloat64Neg ].
			[57]	-> [ ^ self lowcodePrimitiveFloat64NotEqual ].
			[58]	-> [ ^ self lowcodePrimitiveFloat64Sqrt ].
			[59]	-> [ ^ self lowcodePrimitiveFloat64Sub ].
		}
		otherwise: [ ^ self lowcodeUnaryInlinePrimitive2: prim ].
	

! !

!StackInterpreter methodsFor: 'inline primitive support' stamp: 'eem 9/19/2018 17:30'!
lowcode_mem: destAddress cp: sourceAddress y: bytes
	"This method is a workaround a GCC bug.
	In Windows memcpy is putting too much register pressure on GCC when used by Lowcode instructions"
	<inline: #never>
	<option: #LowcodeVM>
	<var: #destAddress type: #'void*'>
	<var: #sourceAddress type: #'void*'>
	<var: #bytes type: #'sqInt'>
	
	"Using memmove instead of memcpy to avoid crashing GCC in Windows."
	self memmove: destAddress _: sourceAddress _: bytes! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 11/30/2018 15:56'!
makeBaseFrameFor: aContext "<Integer>"
	"Marry aContext with the base frame of a new stack page.  Build the base
	 frame to reflect the context's state.  Answer the new page."
	<returnTypeC: #'StackPage *'>
	| page pointer theMethod theIP numArgs stackPtrIndex maybeClosure rcvr |
	<inline: false>
	<var: #page type: #'StackPage *'>
	<var: #pointer type: #'char *'>
	self assert: (objectMemory isContext: aContext).
	self assert: (self isSingleContext: aContext).
	self assert: (objectMemory goodContextSize: aContext).
	page := stackPages newStackPage.
	pointer := page baseAddress.
	theIP := objectMemory fetchPointer: InstructionPointerIndex ofObject: aContext.
	theMethod := objectMemory followObjField: MethodIndex ofObject: aContext.
	(objectMemory isIntegerObject: theIP) ifFalse:
		[self error: 'context is not resumable'].
	theIP := objectMemory integerValueOf: theIP.
	rcvr := objectMemory followField: ReceiverIndex ofObject: aContext.
	"If the frame is a closure activation then the closure should be on the stack in
	 the pushed receiver position (closures receive the value[:value:] messages).
	 Otherwise it should be the receiver proper."
	maybeClosure := objectMemory fetchPointer: ClosureIndex ofObject: aContext.
	maybeClosure ~= objectMemory nilObject
		ifTrue:
			[(objectMemory isForwarded: maybeClosure) ifTrue:
				[maybeClosure := objectMemory fixFollowedField: ClosureIndex ofObject: aContext withInitialValue: maybeClosure].
			 numArgs := self argumentCountOfClosure: maybeClosure.
			 stackPages longAt: pointer put: maybeClosure]
		ifFalse:
			[| header |
			 header := objectMemory methodHeaderOf: theMethod.
			 numArgs := self argumentCountOfMethodHeader: header.
			 "If this is a synthetic context its IP could be pointing at the CallPrimitive opcode.  If so, skip it."
			 ((self methodHeaderHasPrimitive: header)
			  and: [theIP = (1 + (self startPCOfMethodHeader: header))]) ifTrue:
				[theIP := theIP + (self sizeOfCallPrimitiveBytecode: header)].
			 stackPages longAt: pointer put: rcvr].
	"Put the arguments on the stack"
	1 to: numArgs do:
		[:i|
		stackPages
			longAt: (pointer := pointer - objectMemory wordSize)
			put: (objectMemory fetchPointer: ReceiverIndex + i ofObject: aContext).
		"nil the slot in the context so that it doesn't inadvertently hang onto some collectable object.
		 Thanks to Ryan Macnak for identifying this bug"
		objectMemory storePointerUnchecked: ReceiverIndex + i ofObject: aContext withValue: objectMemory nilObject].
	"saved caller ip is sender context in base frame"
	stackPages
		longAt: (pointer := pointer - objectMemory wordSize)
		put: (objectMemory followObjField: SenderIndex ofObject: aContext).
	"base frame's saved fp is null"
	stackPages
		longAt: (pointer := pointer - objectMemory wordSize)
		put: 0.
	page baseFP: pointer; headFP: pointer.
	stackPages
		longAt: (pointer := pointer - objectMemory wordSize)
		put: theMethod.
	stackPages
		longAt: (pointer := pointer - objectMemory wordSize)
		put: (self encodeFrameFieldHasContext: true isBlock: maybeClosure ~= objectMemory nilObject numArgs: numArgs).
	self assert: (self frameHasContext: page baseFP).
	self assert: (self frameNumArgs: page baseFP) == numArgs.
	stackPages
		longAt: (pointer := pointer - objectMemory wordSize)
		put: aContext.
	stackPages
		longAt: (pointer := pointer - objectMemory wordSize)
		put: rcvr.
	stackPtrIndex := self quickFetchInteger: StackPointerIndex ofObject: aContext.
	self assert: ReceiverIndex + stackPtrIndex < (objectMemory lengthOf: aContext).
	numArgs + 1 to: stackPtrIndex do:
		[:i|
		stackPages
			longAt: (pointer := pointer - objectMemory wordSize)
			put: (objectMemory fetchPointer: ReceiverIndex + i ofObject: aContext)].
	"top of stack is the instruction pointer"
	theIP := self iframeInstructionPointerForIndex: theIP method: theMethod.
	stackPages longAt: (pointer := pointer - objectMemory wordSize) put: theIP.
	page headSP: pointer.
	self assert: (self context: aContext hasValidInversePCMappingOf: theIP in: page baseFP).

	"Mark context as married by setting its sender to the frame pointer plus SmallInteger
	 tags and the InstructionPointer to the saved fp (which ensures correct alignment
	 w.r.t. the frame when we check for validity) plus SmallInteger tags."
	objectMemory storePointerUnchecked: SenderIndex
		ofObject: aContext
		withValue: (self withSmallIntegerTags: page baseFP).
	objectMemory storePointerUnchecked: InstructionPointerIndex
		ofObject: aContext
		withValue: (self withSmallIntegerTags: 0).
	self assert: (objectMemory isIntegerObject: (objectMemory fetchPointer: SenderIndex ofObject: aContext)).
	self assert: (self frameOfMarriedContext: aContext) = page baseFP.
	self assert: (self validStackPageBaseFrame: page).
	^page! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 4/27/2016 07:37'!
makeContextSnapshotSafe: ctxt
	"Convert married contexts to widowed contexts, and in the CoInterpreter,
	 map any machine code pcs to bytecode pcs, and then nil all slots beyond
	 top of stack, so that the context can be resumed on any VM."
	<inline: false>
	self assert: (objectMemory isContext: ctxt).
	"The stack pages have already been discarded.  Any remaining married contexts are actually widows."
	(self isMarriedOrWidowedContext: ctxt)
		ifTrue: [self markContextAsDead: ctxt]
		ifFalse: [self ensureContextHasBytecodePC: ctxt].
	 "Fill slots beyond top of stack with nil"
	 (self fetchStackPointerOf: ctxt) + CtxtTempFrameStart
		to: (objectMemory numSlotsOf: ctxt) - 1
		do: [:i |
			 objectMemory
				storePointerUnchecked: i
				ofObject: ctxt
				withValue: objectMemory nilObject]! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 9/19/2013 14:29'!
makePointwithxValue: xValue yValue: yValue
	"make a Point xValue@yValue.
	 We know both will be integers so no value nor root checking is needed"
	| pointResult |
	pointResult := objectMemory eeInstantiateSmallClass: (objectMemory splObj: ClassPoint) numSlots: YIndex + 1.
	objectMemory storePointerUnchecked: XIndex ofObject: pointResult withValue: (objectMemory integerObjectOf: xValue).
	objectMemory storePointerUnchecked: YIndex ofObject: pointResult withValue: (objectMemory integerObjectOf: yValue).
	^pointResult! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 12/7/2017 09:31'!
mapInterpreterOops
	"Map all oops in the interpreter's state to their new values 
	 during garbage collection or a become: operation."
	"Assume: All traced variables contain valid oops."
	<inline: false>
	self mapStackPages.
	self mapMachineCode: self getGCMode.
	self mapTraceLogs.
	self mapVMRegisters.
	self mapProfileState.
	self remapCallbackState.
	(tempOop ~= 0
	 and: [objectMemory shouldRemapOop: tempOop]) ifTrue:
		[tempOop := objectMemory remapObj: tempOop].
	(tempOop2 ~= 0
	 and: [objectMemory shouldRemapOop: tempOop2]) ifTrue:
		[tempOop2 := objectMemory remapObj: tempOop2]! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 5/7/2014 08:35'!
mapMachineCode: theGCMode
	<inline: true>
	"This is a no-op in the StackVM"! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 8/4/2021 11:13'!
mapProfileState
	(objectMemory shouldRemapObj: profileProcess) ifTrue:
		[profileProcess := objectMemory remapObj: profileProcess].
	(objectMemory shouldRemapObj: profileMethod) ifTrue:
		[profileMethod := objectMemory remapObj: profileMethod].
	(objectMemory shouldRemapObj: profileSemaphore) ifTrue:
		[profileSemaphore := objectMemory remapObj: profileSemaphore]! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 12/25/2016 10:30'!
mapStackPages
	<inline: #never>
	<var: #thePage type: #'StackPage *'>
	<var: #theSP type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #callerFP type: #'char *'>
	<var: #theIPPtr type: #'char *'>
	| numLivePages |
	numLivePages := 0.
	0 to: numStackPages - 1 do:
		[:i| | thePage theSP theFP callerFP theIPPtr theIP oop |
		thePage := stackPages stackPageAt: i.
		thePage isFree ifFalse:
			[self assert: (self ifCurrentStackPageHasValidHeadPointers: thePage).
			 numLivePages := numLivePages + 1.
			 theSP := thePage headSP.
			 theFP := thePage  headFP.
			 "Skip the instruction pointer on top of stack of inactive pages."
			 thePage = stackPage
				ifTrue: [theIPPtr := 0]
				ifFalse:
					[theIPPtr := theSP.
					 theSP := theSP + objectMemory wordSize].
			[self assert: (thePage addressIsInPage: theFP).
			 self assert: (thePage addressIsInPage: theSP).
			 self assert: (theIPPtr = 0 or: [thePage addressIsInPage: theIPPtr]).
			 [theSP <= (theFP + FoxReceiver)] whileTrue:
				[oop := stackPages longAt: theSP.
				 (objectMemory shouldRemapOop: oop) ifTrue:
					[stackPages longAt: theSP put: (objectMemory remapObj: oop)].
				 theSP := theSP + objectMemory wordSize].
			 (self frameHasContext: theFP) ifTrue:
				[(objectMemory shouldRemapObj: (self frameContext: theFP)) ifTrue:
					[stackPages
						longAt: theFP + FoxThisContext
						put: (objectMemory remapObj: (self frameContext: theFP))].
				 "With SqueakV3 objectMemory or SpurPlanningCompactor can't assert since object body is yet to move."
				 (objectMemory hasSpurMemoryManagerAPI
				  and: [objectMemory slidingCompactionInProgress not]) ifTrue:
					[self assert: ((self isMarriedOrWidowedContext: (self frameContext: theFP))
								  and: [(self frameOfMarriedContext: (self frameContext: theFP)) = theFP])]].
			 (objectMemory shouldRemapObj: (self frameMethod: theFP)) ifTrue:
				[theIPPtr ~= 0 ifTrue:
					[self assert: (stackPages longAt: theIPPtr) > (self frameMethod: theFP).
					 theIP := (stackPages longAt: theIPPtr) - (self frameMethod: theFP)].
				 stackPages
					longAt: theFP + FoxMethod
					put: (objectMemory remapObj: (self frameMethod: theFP)).
				 theIPPtr ~= 0 ifTrue:
					[stackPages longAt: theIPPtr put: theIP + (self frameMethod: theFP)]].
			 (callerFP := self frameCallerFP: theFP) ~= 0] whileTrue:
				[theSP := (theIPPtr := theFP + FoxCallerSavedIP) + objectMemory wordSize.
				 theFP := callerFP].
			 theSP := theFP + FoxCallerContext. "a.k.a. FoxCallerSavedIP"
			 [theSP <= thePage baseAddress] whileTrue:
				[oop := stackPages longAt: theSP.
				 (objectMemory shouldRemapOop: oop) ifTrue:
					[stackPages longAt: theSP put: (objectMemory remapObj: oop)].
				 theSP := theSP + objectMemory wordSize]]].
	stackPages recordLivePagesOnMapping: numLivePages! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 4/28/2010 12:31'!
mapTraceLogs
	"This is a no-op in the StackVM"! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 9/14/2013 11:24'!
mapVMRegisters
	"Map the oops in the interpreter's vm ``registers'' to their new values 
	during garbage collection or a become: operation."
	"Assume: All traced variables contain valid oops.
	 N.B. Don't trace messageSelector and lkupClass; these are ephemeral, live
	 only during message lookup and because createActualMessageTo will not
	 cause a GC these cannot change during message lookup."
	(objectMemory shouldRemapObj: method) ifTrue:
		[instructionPointer := instructionPointer - method. "*rel to method"
		 method := objectMemory remapObj: method.
		 instructionPointer := instructionPointer + method]. "*rel to method"
	(objectMemory shouldRemapOop: newMethod) ifTrue: "maybe oop due to object-as-method"
		[newMethod := objectMemory remapObj: newMethod]! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 12/13/2017 18:22'!
mappedBackjumpAlwaysInterrupt
	| top offset |
	top := self internalStackTop.
	self assert: (objectMemory isIntegerObject: top).
	offset := objectMemory integerObjectOf: top.
	localIP := localIP - offset.
	self internalPop: 1.
	"+1 since this instr is 3 bytes not 2"
	self ifBackwardsCheckForEvents: 0 - offset + 1. 
	localIP := localIP - 1.
	self fetchNextBytecode! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 12/11/2017 14:34'!
mappedDirectCall
	"250	directCall
	literal index of the method to call on top of stack =>  (variable number of parameters)"
	|methodHeader localPrimIndex methodIndex|
	methodIndex := self internalPopStack. "Can't write this inside next line for Slang inliner"
	newMethod := self literal: (objectMemory integerValueOf: methodIndex).
	self assert: (objectMemory isCompiledMethod: newMethod).
	methodHeader := objectMemory methodHeaderOf: newMethod.
	localPrimIndex := self primitiveIndexOfMethod: newMethod header: methodHeader.
	argumentCount := self argumentCountOfMethodHeader: methodHeader.
	"The primitive function pointer is not cached in the interpreter, but it's called quickly in the JIT"
	primitiveFunctionPointer := self functionPointerFor: localPrimIndex inClass: objectMemory nilObject..
	self internalActivateNewMethod! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 12/11/2017 14:34'!
mappedEnsureEnoughWords
	"50	EnsureEnoughWords
	 literal which is a Smi => ret value is receiver"
	 | slots ok top |
	 top := self internalStackTop.
	 self assert: (objectMemory isIntegerObject: top).
	 slots := objectMemory integerValueOf: top.
	 self assert: slots >= 0.
	 ok := objectMemory checkForAvailableSlots: slots.
	 ok ifFalse:
		[self externalizeIPandSP.
		 self checkForEventsMayContextSwitch: true.
		 self browserPluginReturnIfNeeded.
		 self internalizeIPandSP]! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 12/11/2017 14:35'!
mappedImmcheckDataAtPut: primIndex
	<option: #SistaVM>
	| argIntAdjusted rec result arg1 |
	self assert: (primIndex between: 154 and: 157).
	arg1 := self internalStackValue: 1.
	rec := self internalStackValue: 2.
	self deny: ((objectMemory isOopForwarded: rec) or: [(objectMemory isImmediate: rec)]).
	self assert: (objectMemory isIntegerObject: arg1).
	argIntAdjusted := (objectMemory integerValueOf: arg1) - 1.
	self assert: argIntAdjusted >= 0.
	result := self internalStackTop.
	self internalPop: 2; internalStackTopPut: result.
	(objectMemory isImmutable: rec) ifTrue: [^self cannotAssign: result to: rec withIndex: argIntAdjusted].
	
	primIndex caseOf: {
		"154	immCheckByteAt:put:
		byte object, Smi, 8 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)"
		[154]	->	[self assert: (objectMemory isBytes: rec).
				 self assert: argIntAdjusted < (objectMemory numBytesOf: rec).
				 self assert: (objectMemory isIntegerObject: result).
				 objectMemory
					storeByte: argIntAdjusted
					ofObject: rec
					withValue: (objectMemory integerValueOf: result)].
		"155	immCheckShortAt:put:
		short object, Smi, 16 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)"
		[155]	->	[self assert: (objectMemory isShorts: rec).
				 self assert: argIntAdjusted < (objectMemory num16BitUnitsOf: rec).
				 self assert: (objectMemory isIntegerObject: result).
				 objectMemory
					storeShort16: argIntAdjusted
					ofObject: rec
					withValue: (objectMemory integerValueOf: result)].
		"156	immCheckWordAt:put:
		word object, Smi, 32 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)"
		[156]	->	[self assert: (objectMemory isWords: rec).
				 self assert: argIntAdjusted < (objectMemory num32BitUnitsOf: rec).
				 objectMemory
					storeLong32: argIntAdjusted
					ofObject: rec
					withValue: (objectMemory positive32BitValueOf: result)].
		"157	immCheckDoubleWordAt:put:
		double word object, Smi, 64 bits unsigned Smi or LargePositiveInteger => arg2 (1-based, optimised if arg1 is a constant)"
		[157]	->	[self assert: (objectMemory isLong64s: rec).
				 self assert: argIntAdjusted < (objectMemory num64BitUnitsOf: rec).
				 objectMemory
					storeLong64: argIntAdjusted
					ofObject: rec
					withValue: (objectMemory positive64BitValueOf: result)]}.
		! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 12/11/2017 14:35'!
mappedImmcheckMaybeContextStoreCheckPointerAtPut
	"153	immCheckMaybeContextStoreCheckPointerAt:put:
	 pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
	|rec argIntAdjusted result arg1|
	rec := self internalStackValue: 2.
	arg1 := self internalStackValue: 1.
	self assert: (objectMemory isIntegerObject: arg1).
	argIntAdjusted := (objectMemory integerValueOf: arg1) - 1.
	result := self internalStackTop.
	self internalPop: 3.
	self internalPush: result.
	(self isWriteMediatedContextInstVarIndex: argIntAdjusted) 
						ifFalse: [objectMemory storePointerImmutabilityCheck: argIntAdjusted ofObject: rec withValue: result]
						ifTrue: [self externalizeIPandSP.
								self externalInstVar: argIntAdjusted ofContext: rec  put: result.
								self internalizeIPandSP]! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 12/11/2017 14:36'!
mappedImmcheckStoreCheckPointerAtPut
	"151	immCheckStoreCheckPointerAt:put:
	 pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
	|rec argIntAdjusted result arg1|
	rec := self internalStackValue: 2.
	arg1 := self internalStackValue: 1.
	self assert: (objectMemory isPointers: rec).
	self assert: (objectMemory isIntegerObject: arg1).
	argIntAdjusted := (objectMemory integerValueOf: arg1) - 1.
	self assert: argIntAdjusted >= 0.
	self assert: argIntAdjusted < (objectMemory numSlotsOfAny: rec).
	result := self internalStackTop.
	self internalPop: 3.
	self internalPush: result.
	objectMemory
		storePointerImmutabilityCheck: argIntAdjusted
		ofObject: rec
		withValue: result! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 10/29/2013 18:25'!
markAndTraceAndMaybeFreeStackPages: fullGCFlag
	"Read markAndTraceStackPages:'s comment.  Finish tracing to-be-traced pages.
	 Then free any untraced pages."
	<var: #thePage type: #'StackPage *'>
	<inline: false>
	fullGCFlag ifFalse:
		[0 to: numStackPages - 1 do:
			[:i| | thePage |
			thePage := stackPages stackPageAt: i.
			self assert: (thePage trace: -1) ~= 0]. "Invalidate the trace state for assertion checks"
		^nil].

	self markAndTraceUntracedReachableStackPages.
	self freeUntracedStackPages! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 12/7/2017 09:31'!
markAndTraceInterpreterOops: fullGCFlag
	"Mark and trace all oops in the interpreter's state."
	"Assume: All traced variables contain valid oops.
	 N.B. Don't trace messageSelector and lkupClass; these are ephemeral, live
	 only during message lookup and because createActualMessageTo will not
	 cause a GC these cannot change during message lookup."
	| oop |
	"Must mark stack pages first to initialize the per-page trace
	 flags for full garbage collect before any subsequent tracing."
	self markAndTraceStackPages: fullGCFlag.
	self markAndTraceTraceLog.
	self markAndTracePrimTraceLog.
	objectMemory markAndTrace: objectMemory specialObjectsOop. "also covers nilObj, trueObj, falseObj, and compact classes"
	(objectMemory isImmediate: newMethod) ifFalse:
		[objectMemory markAndTrace: newMethod].
	self traceProfileState.
	tempOop = 0 ifFalse: [objectMemory markAndTrace: tempOop].
	tempOop2 = 0 ifFalse: [objectMemory markAndTrace: tempOop2].

	"V3 memory manager support"
	1 to: objectMemory remapBufferCount do:
		[:i | 
		oop := objectMemory remapBuffer at: i.
		(objectMemory isImmediate: oop) ifFalse: [objectMemory markAndTrace: oop]].

	"Old callback support - trace suspended callback list"
	1 to: jmpDepth do:
		[:i|
		oop := suspendedCallbacks at: i.
		(objectMemory isImmediate: oop) ifFalse:[objectMemory markAndTrace: oop].
		oop := suspendedMethods at: i.
		(objectMemory isImmediate: oop) ifFalse:[objectMemory markAndTrace: oop]]! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 3/13/2014 16:48'!
markAndTraceMachineCodeOfMarkedMethods
	"This is a no-op in the StackVM"! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 3/23/2009 11:56'!
markAndTraceOrFreeMachineCode: fullGCFlag
	"This is a no-op in the StackVM"! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 6/26/2010 13:27'!
markAndTracePrimTraceLog
	"This is a no-op in the StackVM"! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 5/6/2015 17:06'!
markAndTraceStackPage: thePage
	| theSP theFP frameRcvrOffset callerFP oop |
	<var: #thePage type: #'StackPage *'>
	<var: #theSP type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #frameRcvrOffset type: #'char *'>
	<var: #callerFP type: #'char *'>
	<inline: false>

	self assert: (stackPages isFree: thePage) not.
	self assert: (self ifCurrentStackPageHasValidHeadPointers: thePage).
	self assert: thePage trace ~= StackPageTraced.
	thePage trace: StackPageTraced.

	theSP := thePage headSP.
	theFP := thePage  headFP.
	"Skip the instruction pointer on top of stack of inactive pages."
	thePage = stackPage ifFalse:
		[theSP := theSP + objectMemory wordSize].
	[frameRcvrOffset := self frameReceiverLocation: theFP.
	 [theSP <= frameRcvrOffset] whileTrue:
		[oop := stackPages longAt: theSP.
		 (objectMemory isOopForwarded: oop) ifTrue:
			[oop := objectMemory followForwarded: oop.
			 stackPages longAt: theSP put: oop].
		 (objectMemory isImmediate: oop) ifFalse:
			[objectMemory markAndTrace: oop].
		 theSP := theSP + objectMemory wordSize].
	(self frameHasContext: theFP) ifTrue:
		[self assert: (objectMemory isContext: (self frameContext: theFP)).
		 objectMemory markAndTrace: (self frameContext: theFP)].
	objectMemory markAndTrace: (self iframeMethod: theFP).
	(callerFP := self frameCallerFP: theFP) ~= 0] whileTrue:
		[theSP := theFP + FoxCallerSavedIP + objectMemory wordSize.
		 theFP := callerFP].
	theSP := theFP + FoxCallerSavedIP. "caller ip is frameCallerContext in a base frame"
	[theSP <= thePage baseAddress] whileTrue:
		[oop := stackPages longAt: theSP.
		 (objectMemory isOopForwarded: oop) ifTrue:
			[oop := objectMemory followForwarded: oop.
			 stackPages longAt: theSP put: oop].
		 (objectMemory isImmediate: oop) ifFalse:
			[objectMemory markAndTrace: oop].
		 theSP := theSP + objectMemory wordSize]! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 11/7/2013 11:51'!
markAndTraceStackPages: fullGCFlag
	"GC of pages.  Throwing away all stack pages on full GC is simple but dangerous
	 because it causes us to allocate lots of contexts immediately before a GC.
	 Reclaiming pages whose top context is not referenced is poor because it would
	 take N incrementalGCs to reclaim N unused pages.  Only the page whose top
	 context is not referred to by the bottom context of any other page would be
	 reclaimed.  Not until the next GC would the page whose top contect is the
	 previously reclaimed page's base frame's bottom context be reclaimed.

	 Better is to not mark stack pages until their contexts are encountered.  We can
	 eagerly trace the active page and the page reachable from its bottom context
	 if any, and so on.  Other pages can be marked when we encounter a married
	 context."
	| thePage context |
	<var: #thePage type: #'StackPage *'>
	<inline: false>
	objectMemory hasSpurMemoryManagerAPI ifFalse:
		[self initStackPageGC].

	"On an incremental GC simply consider all non-free stack pages to be roots."
	fullGCFlag ifFalse:
		[0 to: numStackPages - 1 do:
			[:i|
			thePage := stackPages stackPageAt: i.
			(stackPages isFree: thePage) ifFalse:
				[self markAndTraceStackPage: thePage]].
		^nil].

	"On a full GC only eagerly trace pages referenced from
	 the base of the active page, i.e. on the active stack."
	stackPage = 0 ifTrue: [^nil].
	thePage := stackPage.
	[self markAndTraceStackPage: thePage.
	 context := self frameCallerContext: thePage baseFP.
	 ((objectMemory isContext: context)
	  and: [(self isMarriedOrWidowedContext: context)
	  and: [self isStillMarriedContext: context]]) ifTrue:
		[thePage := stackPages stackPageFor:  (self frameOfMarriedContext: context).
		 self assert: (stackPages isFree: thePage) not].
	 thePage trace < StackPageTraced] whileTrue! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 2/26/2009 20:38'!
markAndTraceTraceLog
	"This is a no-op in the StackVM"! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 11/7/2013 11:51'!
markAndTraceUntracedReachableStackPages
	"Trace any untraced pages"
	| thePage foundToBeTracedPage |
	<var: #thePage type: #'StackPage *'>
	<inline: false>

	[foundToBeTracedPage := false.
	0 to: numStackPages - 1 do:
		[:i|
		thePage := stackPages stackPageAt: i.
		((stackPages isFree: thePage) not
		 and: [thePage trace = StackPageReachedButUntraced]) ifTrue:
			[foundToBeTracedPage := true.
			 self markAndTraceStackPage: thePage]].
	foundToBeTracedPage] whileTrue! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 2/27/2013 13:50'!
markContextAsDead: oop
	"Mark the argument, which must be a context, married, widowed or single, as dead.
	 For married or widowed contexts this breaks any link to the spouse and makes the context single.
	 For all contexts, marks the context as inactive/having been returned from."
	<inline: true>
	self assert: (objectMemory isContext: oop).
	objectMemory
		storePointerUnchecked: SenderIndex ofObject: oop withValue: objectMemory nilObject;
		storePointerUnchecked: InstructionPointerIndex ofObject: oop withValue: objectMemory nilObject! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 4/2/2015 10:58'!
marriedContext: spouseContext pointsTo: anOop stackDeltaForCurrentFrame: stackDeltaForCurrentFrame
	"This is a helper for primitiveObjectPointsTo so it *does not* check the frameContext field because that is an implicit self-reference not present in the state ."
	| theFP thePage theSP rcvrOffset |
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<var: #theSP type: #'char *'>
	<var: #rcvrOffset type: #'char *'>
	theFP := self frameOfMarriedContext: spouseContext.
	theFP = framePointer
		ifTrue: [theSP := stackPointer + (stackDeltaForCurrentFrame * objectMemory wordSize)]
		ifFalse:
			[thePage := stackPages stackPageFor: theFP.
			theSP := self findSPOf: theFP on: thePage].
	(objectMemory isIntegerObject: anOop)
		ifTrue: "Check stack and instruction pointer fields."
			[(anOop = (objectMemory integerObjectOf: (self stackPointerIndexForFrame: theFP WithSP: theSP))
			or: [anOop = (self externalInstVar: InstructionPointerIndex ofContext: spouseContext)]) ifTrue:
				[^true]]
		ifFalse: "Check method and sender fields, avoiding unnecessarily reifying sender context."
			[anOop = (self frameMethodObject: theFP) ifTrue:
				[^true].
			 (self isBaseFrame: theFP)
				ifTrue: [anOop = (self frameCallerContext: theFP) ifTrue:
							[^true]]
				ifFalse: [((self frameHasContext: (self frameCallerFP: theFP))
						and: [anOop = (self frameContext: (self frameCallerFP: theFP))]) ifTrue:
							[^true]]].
	"Now check receiver, temps and stack contents"
	rcvrOffset := self frameReceiverLocation: theFP.
	 [theSP <= rcvrOffset] whileTrue:
		[anOop = (stackPages longAt: theSP) ifTrue:
			[^true].
		 theSP := theSP + objectMemory wordSize].
	"Finally check stacked receiver (closure field or duplicate of receiver) and arguments"
	theSP := theFP + FoxCallerSavedIP + objectMemory wordSize.
	rcvrOffset := theFP + (self frameStackedReceiverOffset: theFP).
	 [theSP <= rcvrOffset] whileTrue:
		[anOop = (stackPages longAt: theSP) ifTrue:
			[^true].
		 theSP := theSP + objectMemory wordSize].
	^false! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 3/10/2017 12:55'!
marryContextInNewStackPageAndInitializeInterpreterRegisters: aContext
	"Establish aContext at the base of a new stackPage, make the stackPage the
	 active one and set-up the interreter registers.  This is used to boot the system
	 and bring it back after a snapshot."
	<inline: false>
	| newPage |
	<var: #newPage type: #'StackPage *'>
	self assert: stackPage = 0.
	newPage := self makeBaseFrameFor: aContext.
	self setStackPageAndLimit: newPage.
	self setStackPointersFromPage: newPage.
	self setMethod: (self iframeMethod: stackPage headFP).
	instructionPointer := self popStack! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 8/14/2012 08:59'!
marryFrame: theFP SP: theSP
	"Marry an unmarried frame.  This means creating a spouse context
	 initialized with a subset of the frame's state (state through the last argument)
	 that references the frame."
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	<inline: false>
	^self marryFrame: theFP SP: theSP copyTemps: false! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
marryFrame: theFP SP: theSP copyTemps: copyTemps
	"Marry an unmarried frame.  This means creating a spouse context
	 initialized with a subset of the frame's state that references the frame.
	 For the default closure implementation we do not need to copy temps.
	 Different closure implementations may require temps to be copied."
	| theContext methodHeader numSlots numArgs numStack closureOrNil numTemps |
	<inline: true>
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	self assert: (self frameHasContext: theFP) not.

	"The SP is expected to be pointing at the last oop on the stack, not at the pc"
	self assert: (objectMemory addressCouldBeOop: (stackPages longAt: theSP)).

	methodHeader := objectMemory methodHeaderOf: (self frameMethod: theFP).
	"Decide how much of the stack to preserve in widowed contexts.  Preserving too much
	 state will potentially hold onto garbage.  Holding onto too little may mean that a dead
	 context isn't informative enough in a debugging situation.  If copyTemps is false (as it
	 is in the default closure implementation) compromise, retaining only the arguments with
	 no temporaries.  Note that we still set the stack pointer to its current value, but stack
	 contents other than the arguments are nil."
	numArgs := self frameNumArgs: theFP.
	numStack := self stackPointerIndexForFrame: theFP WithSP: theSP.

	closureOrNil := (self frameIsBlockActivation: theFP)
						ifTrue: [self pushedReceiverOrClosureOfFrame: theFP]
						ifFalse: [objectMemory nilObject].

	numSlots := (self methodHeaderIndicatesLargeFrame: methodHeader)
					ifTrue: [LargeContextSlots]
					ifFalse: [SmallContextSlots].
	theContext := objectMemory eeInstantiateMethodContextSlots: numSlots.
	self assert: numStack + ReceiverIndex <= numSlots. 
	"Mark context as married by setting its sender to the frame pointer plus SmallInteger
	 tags and the InstructionPointer to the saved fp (which ensures correct alignment
	 w.r.t. the frame when we check for validity)"
	objectMemory storePointerUnchecked: SenderIndex
		ofObject: theContext
		withValue: (self withSmallIntegerTags: theFP).
	objectMemory storePointerUnchecked: InstructionPointerIndex
		ofObject: theContext
		withValue: (self withSmallIntegerTags: (self frameCallerFP: theFP)).
	objectMemory storePointerUnchecked: StackPointerIndex
		ofObject: theContext
		withValue: (objectMemory integerObjectOf: numStack).
	objectMemory storePointerUnchecked: MethodIndex
		ofObject: theContext
		withValue: (self frameMethod: theFP).
	objectMemory storePointerUnchecked: ClosureIndex ofObject: theContext withValue: closureOrNil.
	objectMemory storePointerUnchecked: ReceiverIndex
		ofObject: theContext
		withValue: (self frameReceiver: theFP).
	"If copyTemps is false, store just the arguments.  If the frame is divorced the context
	 will have valid arguments but all temporaries will be nil."
	1 to: numArgs do:
		[:i|
		objectMemory storePointerUnchecked: ReceiverIndex + i
			ofObject: theContext "inline self temporary: i - 1 in:theFP" 
			withValue: (stackPages longAt: theFP
										+ FoxCallerSavedIP
										+ ((numArgs - i + 1) * objectMemory wordSize))].
	copyTemps ifTrue:
		[numTemps := self frameNumTemps: theFP.
		 1 to: numTemps do:
			[:i|
			objectMemory storePointerUnchecked: ReceiverIndex + i + numArgs
				ofObject: theContext
				withValue: (self temporary: i - 1 in: theFP)].
		 numArgs := numArgs + numTemps].

	numArgs + 1 to: numStack do:
		[:i|
		objectMemory storePointerUnchecked: ReceiverIndex + i
			ofObject: theContext
			withValue: objectMemory nilObject].

	self setFrameContext: theFP to: theContext.
	self setFrameHasContext: theFP.

	self assert: (self frameHasContext: theFP).
	self assert: (self frameOfMarriedContext: theContext) = theFP.
	self assert: numStack + ReceiverIndex < (objectMemory lengthOf: theContext).

	^theContext
! !

!StackInterpreter methodsFor: 'cog jit support' stamp: 'eem 7/14/2014 18:28'!
maybeFixClonedCompiledMethod: objOop
	"This is a noop in the Stack VM"
	<inline: true>! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 1/30/2011 17:13'!
maybeFlagMethodAsInterpreted: aMethod
	"Convenient  noop in the StackInterpreter e.g. for sendInvokeCallbackContext:"! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 4/20/2020 15:24'!
maybeInlinePositive32BitIntegerFor: integerValue
	"N.B. will *not* cause a GC.
	 integerValue is interpreted as POSITIVE, e.g. as the result of Bitmap>at:."
	<notOption: #Spur64BitMemoryManager>
	<inline: #always>
	<var: 'integerValue' type: #'unsigned int'>
	| newLargeInteger |
	self deny: objectMemory hasSixtyFourBitImmediates.
       "force coercion because slang inliner sometimes incorrectly pass a signed int without converting to unsigned"
       (self cCode: [self cCoerceSimple: integerValue to: #'unsigned int']
			inSmalltalk: [integerValue bitAnd: 1 << 32 - 1]) <= objectMemory maxSmallInteger ifTrue:
		[^objectMemory integerObjectOf: integerValue].
	newLargeInteger := objectMemory
							eeInstantiateSmallClassIndex: ClassLargePositiveIntegerCompactIndex
							format: (objectMemory byteFormatForNumBytes: 4)
							numSlots: 1.
	SPURVM
		ifTrue:
			["Memory is 8 byte aligned in Spur, make sure that oversized bytes are set to zero" "eem 4/28/2016 questionable; they should never be read"
			objectMemory storeLong32: 0 ofObject: newLargeInteger withValue: (objectMemory byteSwapped32IfBigEndian: integerValue).
			objectMemory storeLong32: 1 ofObject: newLargeInteger withValue: 0]
		ifFalse: 
			[objectMemory storeLong32: 0 ofObject: newLargeInteger withValue: (objectMemory byteSwapped32IfBigEndian: integerValue)].
	^newLargeInteger! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 5/12/2016 11:58'!
maybeLeakCheckExternalPrimCall: aMethodObj
	"Assert-only check for leaks after external prim calls if checkForLeaks includes the GCCheckPrimCall flag.
	 This is ionly really useful from a low-level C debugger, hence no accessors for checkedPluginName."
	((objectMemory checkForLeaks anyMask: GCCheckPrimCall)
	 and: [(self isExternalPrimitiveCall: aMethodObj)
	 and: [checkedPluginName = (objectMemory fetchPointer: 0 ofObject: (self literal: 0 ofMethod: aMethodObj))]]) ifTrue:
		[objectMemory runLeakCheckerFor: GCCheckPrimCall].
	^true! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 6/6/2016 12:12'!
maybeMapPrimitiveFunctionPointerBackToSomethingEvaluable
	"In the real VM primitiveFunctionPointer is either an index (for quick primitives)
	 or a proper function pointer to a primitive.  In the simulator it may be a small
	 index (corresponding to a quick primitive index), a symbol (corresponding to
	 a function pointer) or an index into the externalPrimitiveTable, or an invalid
	 address that references an evaluable in the simulatedTrampolines dictionary
	 of the Cogit.  The simulator expects dispatchFunctionPointer to be called with
	 primitiveFunctionPointer being a symbol only for internal primitives.  External
	 primitives must have their funciton pointer mapped back to an index.  This
	 method does the reverse mapping."
	<doNotGenerate>
	(self isExternalPrimitiveCall: newMethod) ifTrue: "External prims must be evaluated by the right plugin..."
		[| pfp index externalIndex |
		 pfp := primitiveFunctionPointer.
		 index := self mappedPluginEntries findFirst: [:entry| entry second == primitiveFunctionPointer].
		 self assert: index ~= 0.
		 externalIndex := 1000 + (externalPrimitiveTable object
												indexOf: index
												ifAbsent: [self error: 'entry not found']).
		 self assert: ((self pluginEntryFor: externalIndex) notNil
					   and: [(self pluginEntryFor: externalIndex) second == primitiveFunctionPointer]).
		 primitiveFunctionPointer := externalIndex]! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 05/07/2020 16:13'!
maybeRetryPrimitiveOnFailure
	"In Spur two cases of pirmitive failure are handled specially.  A primitive may fail due to validation
	 encountering a forwarder. On failure, check the accessorDepth for the primitive and if non-negative
	 scan the args to the depth, following any forwarders.  Retry the primitive if any are found.  Hence
	 lazily and transparently following forwarders on primitive failue.  Additionally a prmitive might fail
	 due to an allocation failing.  Retry if primitives have failed with PrimErrNoMemory after running
	 first the scavenger and then on a subsequent failure, the global mark-sweep collector.  Hence lazily
	 and transparently GC on memory exhaustion."
	<inline: true>
	(objectMemory hasSpurMemoryManagerAPI and: [self failed]) ifTrue:
		[self retryPrimitiveOnFailure]! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem 12/28/2014 15:03'!
maybeReturnToMachineCodeFrame
	"Hook for the CoInterpreter to return to machine-code or
	 load the saved IP.  This is a nop in the StackInterpreter."! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 3/6/2018 18:43'!
maybeSelectorOfMethod: methodObj
	"Answer the selector of a method, assuming its penultimate literal is either
	 a symbol or a pointer object whose first slot references the method and
	 whose second slot is a symbol (i.e. an AdditionalMethodState).  If a Symbol
	 can't be found answer nil.  This isn't satisfactory, as it puts a lot of information
	 into the VM, but it is needed for adequate crash debugging at Cadence.
	 With full blocks as of 9/2016 the last literal of a CompiledBlock is a back pointer
	 to the enclosing block or compiled method."
	<api>
	| ultimateLiteral penultimateLiteral maybeSelector |
	self assert: (objectMemory isOopCompiledMethod: methodObj).
	ultimateLiteral := self ultimateLiteralOf: methodObj.
	(objectMemory isOopCompiledMethod: ultimateLiteral) ifTrue:
		[^self maybeSelectorOfMethod: ultimateLiteral].
	penultimateLiteral := self penultimateLiteralOf: methodObj.
	(objectMemory isWordsOrBytes: penultimateLiteral) ifTrue:
		[^(objectMemory fetchClassTagOfNonImm: penultimateLiteral)
			= (objectMemory fetchClassTagOfNonImm: (objectMemory splObj: SelectorDoesNotUnderstand)) ifTrue:
			[penultimateLiteral]].
	^((objectMemory isPointers: penultimateLiteral)
	 and: [(objectMemory numSlotsOf: penultimateLiteral) >= 2
	 and: [(objectMemory fetchPointer: 0 ofObject: penultimateLiteral) = methodObj
	 and: [maybeSelector := objectMemory fetchPointer: 1 ofObject: penultimateLiteral.
		(objectMemory isWordsOrBytes: maybeSelector)
	 and: [(objectMemory fetchClassTagOfNonImm: maybeSelector)
			= (objectMemory fetchClassTagOfNonImm: (objectMemory splObj: SelectorDoesNotUnderstand))]]]]) ifTrue:
		[maybeSelector]! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 11/27/2012 14:53'!
maybeTraceStackOverflow
	"nop in the stack interpreter"! !

!StackInterpreter methodsFor: 'simulation' stamp: 'eem 8/9/2014 17:49'!
metaclassNumSlots
	<doNotGenerate>
	^metaclassNumSlots! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 7/22/2021 19:18'!
metadataFlagsFor: selector using: aCCodeGenerator
	"Compute a primitive's metadata.  This is the accessorDepth which is derived from parse tree analysis,
	 and some optional flags. Try and locate the flags parameter in an export:flags: pragma in the primitive method."
	<doNotGenerate>
		
	^(((aCCodeGenerator accessorDepthForSelector: selector) ifNil: [0]) bitShift: 8)
	+ (self class metadataFlagsForPrimitive: ((aCCodeGenerator methodNamed: selector) ifNotNil: [:m| m definingClass >> m smalltalkSelector]))! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 3/16/2018 15:35'!
methodArg: index
	"Like #stackValue: but access method arguments left-to-right"
	<doNotGenerate> "Obsolete; was never used; replaced with methodReturnBool: in the VirtualMachine struct."
	self deprecated.
	index > argumentCount + 1 ifTrue:
		[self cCode: 'fprintf(stderr,"[VM]: Attempt to access method args beyond range\n")'.
		self printCallStack.
		self primitiveFail.
		^0].
	^self stackValue: argumentCount - index! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'ar 10/7/1998 18:38'!
methodArgumentCount
	^argumentCount! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 08/08/2014 10:01'!
methodClassAssociationOf: methodPointer
	<api>
	^self literal: (objectMemory literalCountOf: methodPointer) - 1 ofMethod: methodPointer! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'cb 3/14/2017 13:50'!
methodClassOf: methodPointer
	<api>
	"Answer the method class of a method which is the value of an Association in the last literal,
	 or answer nil if there isn't one.
	 Using a read barrier here simplifies the become implementation and costs very little
	 because the class index and ValueIndex of the association almost certainly share a cache line."
	| literal |
	literal := self followLiteral: (objectMemory literalCountOf: methodPointer) - 1 ofMethod: methodPointer.
	^ (literal ~= objectMemory nilObject 
		and: [objectMemory isPointers: literal]) "filters out fullBlock last literal"
			ifTrue: 
				[self assert: (objectMemory numSlotsOf: literal) > ValueIndex.
			 	objectMemory followField: ValueIndex ofObject: literal]
			ifFalse: [objectMemory nilObject]! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 6/7/2016 14:45'!
methodHeaderHasPrimitive: methodHeader
	"Note: We now have 10 bits of primitive index, but they are in two places
	 for temporary backward compatibility.  The time to unpack is negligible,
	 since the derived primitive function pointer is stored in the method cache."
	^objectMemory hasSpurMemoryManagerAPI
		ifTrue: [self alternateHeaderHasPrimitiveFlag: methodHeader]
		ifFalse:
			[MULTIPLEBYTECODESETS
				ifTrue:
					[(objectMemory headerIndicatesAlternateBytecodeSet: methodHeader)
						ifTrue: [self alternateHeaderHasPrimitiveFlag: methodHeader]
						ifFalse: [methodHeader anyMask: V3PrimitiveBitsMask]]
				ifFalse:
					[methodHeader anyMask: V3PrimitiveBitsMask]]! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 8/8/2014 09:23'!
methodHeaderIndicatesLargeFrame: methodHeader
	<inline: true>
	^(methodHeader bitAnd: LargeContextBit) ~= 0! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
methodPrimitiveIndex
	<api>
	((objectMemory addressCouldBeObj: newMethod)
	 and: [objectMemory isCompiledMethod: newMethod]) ifFalse:
		[^-1].
	^self primitiveIndexOf: newMethod! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 6/22/2020 22:16'!
methodReturnBool: boolean
	"Sets the return value for a method.  In the CoInterpreter we replace the cumbersome
	 primResult machinery."
	self deny: self failed.
	self pop: argumentCount+1 thenPushBool: boolean.
	^0! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 6/22/2020 22:16'!
methodReturnFloat: aFloat
	"Sets the return value for a method."
	<var: 'aFloat' type: #double>
	self deny: self failed.
	self pop: argumentCount+1 thenPushFloat: aFloat.
	^0! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 6/22/2020 22:17'!
methodReturnInteger: integer
	"Sets the return value for a method.  In the CoInterpreter we replace the cumbersome
	 primResult machinery."
	self deny: self failed.
	self pop: argumentCount+1 thenPushInteger: integer.
	^0! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 6/22/2020 22:10'!
methodReturnReceiver
	"Sets the return value for a method"
	self deny: self failed.
	self pop: argumentCount.
	^0! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 9/7/2020 13:54'!
methodReturnString: aCString
	"Attempt to answer a ByteString for a given C string as the result of a primitive."
	<var: 'aCString' type: #'char *'>
	<inline: #never>
	self deny: self failed.
	aCString
		ifNil: [primFailCode := PrimErrOperationFailed]
		ifNotNil:
			[(objectMemory stringForCString: aCString)
				ifNil: [primFailCode := PrimErrNoMemory]
				ifNotNil: [:result| self pop: argumentCount+1 thenPush: result]].
	^0! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 6/22/2020 22:17'!
methodReturnValue: oop
	"Sets the return value for a method.  In the CoInterpreter we replace the cumbersome
	 primResult machinery."
	self deny: self failed.
	self pop: argumentCount+1 thenPush: oop.
	^0! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 6/7/2016 14:46'!
methodUsesAlternateBytecodeSet: aMethodObj
	<api>
	<inline: true>
	"A negative header selects the alternate bytecode set."
	^objectMemory headerIndicatesAlternateBytecodeSet: (objectMemory methodHeaderOf: aMethodObj)! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 8/23/2018 14:30'!
methodUsesPrimitiveErrorCode: aMethodObj
	"Answer if aMethodObj contains a primitive and uses the primitive error code."
	<inline: true>
	| methodHeader |
	methodHeader := objectMemory methodHeaderOf: aMethodObj.
	^(self primitiveIndexOfMethod: aMethodObj header: methodHeader) > 0
	  and: [(self longStoreBytecodeForHeader: methodHeader)
			= (objectMemory
				fetchByte: (self startPCOfMethod: aMethodObj) + (self sizeOfCallPrimitiveBytecode: methodHeader)
				ofObject: aMethodObj)]! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 2/24/2009 15:54'!
mframeCogMethod: theFP
	self shouldNotImplement! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 12/10/2009 13:15'!
microsecondsToMilliseconds: microseconds
	<var: #microseconds type: #usqLong>
	^microseconds // 1000 bitAnd: MillisecondClockMask! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 12/5/2014 11:50'!
mnuBreak: selectorString point: selectorLength receiver: receiverOrNil
	<doNotGenerate> "C version is in platforms/Cross/vm/dispdbg.h"
	"self shortPrintFrameAndCallers: localFP"
	| i |
	breakSelectorLength negated = selectorLength ifTrue:
		[i := breakSelectorLength negated.
		 [i > 0] whileTrue:
			[(objectMemory byteAt: selectorString + i - 1) = (breakSelector at: i) asInteger
				ifTrue: [(i := i - 1) = 0 ifTrue:
							[self changed: #byteCountText.
							 self halt: 'MNU of '
									, breakSelector,
									(receiverOrNil
										ifNotNil: [' to ', (self shortPrint: receiverOrNil)]
										ifNil: [''])]]
				ifFalse: [i := 0]]]! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 12/5/2014 11:51'!
mnuBreakpoint: selector receiver: rcvr
	<inline: true>
	self mnuBreak: (objectMemory firstFixedFieldOfMaybeImmediate: selector)
		point: (objectMemory lengthOfMaybeImmediate: selector)
		receiver: rcvr! !

!StackInterpreter methodsFor: 'initialization' stamp: 'tpr (auto pragmas 12/08) 3/24/2004 21:22'!
moduleUnloaded: aModuleName 
	"The module with the given name was just unloaded. 
	Make sure we have no dangling references."
	<export: true>
	<var: #aModuleName type: 'char *'>
	(aModuleName strcmp: 'SurfacePlugin') = 0
		ifTrue: ["Surface plugin went away. Should never happen. But  then, who knows"
			showSurfaceFn := 0]! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 11/13/2014 13:17'!
moveFramesIn: oldPage through: theFP toPage: newPage
	"Move frames from the hot end of oldPage through to theFP to newPage.
	 This has the effect of making theFP a base frame which can be stored into.
	 Answer theFP's new location."
	| newSP newFP stackedReceiverOffset delta callerFP callerIP fpInNewPage offsetCallerFP theContext |
	<inline: false>
	<var: #oldPage type: #'StackPage *'>
	<var: #theFP type: #'char *'>
	<var: #newPage type: #'StackPage *'>
	<var: #newSP type: #'char *'>
	<var: #newFP type: #'char *'>
	<var: #callerFP type: #'char *'>
	<var: #fpInNewPage type: #'char *'>
	<var: #offsetCallerFP type: #'char *'>
	<var: #source type: #'char *'>
	<returnTypeC: 'char *'>
	newSP := newPage baseAddress + objectMemory wordSize.
	stackedReceiverOffset := self frameStackedReceiverOffset: theFP.
	"First move the data.  We will fix up frame pointers later."
	theFP + stackedReceiverOffset
		to: oldPage headSP
		by: objectMemory wordSize negated
		do: [:source|
			newSP := newSP - objectMemory wordSize.
			stackPages longAt: newSP put: (stackPages longAt: source)].
	"newSP = oldSP + delta => delta = newSP - oldSP"
	delta := newSP - oldPage headSP.
	newFP := newPage baseAddress - stackedReceiverOffset.
	self setHeadFP: oldPage headFP + delta andSP: newSP inPage: newPage.
	newPage baseFP: newFP.
	callerFP := self frameCallerFP: theFP.
	self assert: (self isBaseFrame: theFP) not.
	self assert: (self frameHasContext: callerFP).
	callerIP := self oopForPointer: (self frameCallerSavedIP: theFP).
	stackPages longAt: theFP + stackedReceiverOffset put: callerIP.
	self assert: (callerFP < oldPage baseAddress
				and: [callerFP > (oldPage realStackLimit - (LargeContextSlots * objectMemory bytesPerOop / 2))]).
	oldPage
		headFP: callerFP;
		headSP: theFP + stackedReceiverOffset.
	"Mark the new base frame in the new page (FoxCallerContext a.k.a. FoxCallerSavedIP)"
	stackPages longAt: newFP + FoxCallerContext put:  (self frameContext: callerFP).
	stackPages longAt: newFP + FoxSavedFP put: 0.
	"Now relocate frame pointers, updating married contexts to refer to their moved spouse frames."
	fpInNewPage := newPage headFP.
	[offsetCallerFP := self frameCallerFP: fpInNewPage.
	 offsetCallerFP ~= 0 ifTrue:
		[offsetCallerFP := offsetCallerFP + delta].
	 stackPages longAt: fpInNewPage + FoxSavedFP put: (self oopForPointer: offsetCallerFP).
	 (self frameHasContext: fpInNewPage) ifTrue:
		[theContext := self frameContext: fpInNewPage.
		 objectMemory storePointerUnchecked: SenderIndex
			ofObject: theContext
			withValue: (self withSmallIntegerTags: fpInNewPage).
		 objectMemory storePointerUnchecked: InstructionPointerIndex
			ofObject: theContext
			withValue: (self withSmallIntegerTags: offsetCallerFP)].
	 fpInNewPage := offsetCallerFP.
	 fpInNewPage ~= 0] whileTrue.
	^newFP! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
nacFetchStackPointerOf: aContext
	"A version of fetchStackPointerOf: for use when objects may be forwarded.
	 Does not do an assert-check of the stack pointer being in bounds."
	| sp |
	<inline: true>
	sp := objectMemory fetchPointer: StackPointerIndex ofObject: aContext.
	(objectMemory isIntegerObject: sp) ifFalse: [^0].
	^objectMemory integerValueOf: sp! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 8/21/2014 07:19'!
nameOfClass: classOop
	"Brain-damaged nameOfClass: for C VM.  Does *not* answer Foo class for metaclasses.
	 Use e.g. classIsMeta: to avoid being fooled."
	<inline: false>
	<returnTypeC: #'char *'>
	| numSlots maybeNameOop maybeThisClassOop |
	numSlots := objectMemory numSlotsOf: classOop.
	numSlots = metaclassNumSlots ifTrue:
		[maybeThisClassOop := objectMemory fetchPointer: thisClassIndex ofObject: classOop.
		(self addressCouldBeClassObj: maybeThisClassOop) ifTrue:
			[^self nameOfClass: (objectMemory fetchPointer: thisClassIndex ofObject: classOop)].
		 ^'bad class'].
	(numSlots > classNameIndex
	 and: [maybeNameOop := objectMemory fetchPointer: classNameIndex ofObject: classOop.
		objectMemory isBytes: maybeNameOop]) ifFalse:
		[^'bad class'].
	^objectMemory firstIndexableField: maybeNameOop! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:38'!
nativeFramePointerIn: theFP
	"See StackInterpreter class>>initializeFrameIndices"
	<inline: false>
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	^ stackPages pointerAtPointer: theFP + FoxReceiver - (BytesPerWord*3)! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 9/7/2016 14:07'!
nativeFramePointerIn: theFP put: valueOopPointer
	"See StackInterpreter class>>initializeFrameIndices"
	<inline: false>
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	<var: #valueOopPointer type: #'char*'>
	^ stackPages pointerAtPointer: theFP + FoxReceiver - (BytesPerWord*3) put: valueOopPointer! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:39'!
nativeFrameSignalObjectIn: theFP
	"See StackInterpreter class>>initializeFrameIndices"
	<inline: false>
	<option: #LowcodeVM>
	<var: #theFP type: #'char *'>
	^ stackPages longAt: theFP + FoxReceiver - (BytesPerWord*3)! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:39'!
nativeFrameSignalObjectIn: theFP put: oop
	"See StackInterpreter class>>initializeFrameIndices"
	<inline: false>
	<option: #LowcodeVM>
	<var: #theFP type: #'char *'>
	^ stackPages longAt: theFP + FoxReceiver - BytesPerWord put: oop! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:36'!
nativePreviousStackPointerIn: theFP
	"See StackInterpreter class>>initializeFrameIndices"
	<inline: false>
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	^ stackPages pointerAtPointer: theFP + FoxReceiver - (BytesPerWord*2)! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 9/7/2016 14:06'!
nativePreviousStackPointerIn: theFP put: valueOopPointer
	"See StackInterpreter class>>initializeFrameIndices"
	<inline: false>
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	<var: #valueOopPointer type: #'char*'>
	^ stackPages pointerAtPointer: theFP + FoxReceiver - (BytesPerWord*2) put: valueOopPointer! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:39'!
nativeStackPointerIn: theFP
	"See StackInterpreter class>>initializeFrameIndices"
	<inline: false>
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	^ stackPages pointerAtPointer: theFP + FoxReceiver - (BytesPerWord*4)! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 9/7/2016 13:52'!
nativeStackPointerIn: theFP put: valueOopPointer
	"See StackInterpreter class>>initializeFrameIndices"
	<inline: false>
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	<var: #valueOopPointer type: #'char *'>
	^ stackPages pointerAtPointer: theFP + FoxReceiver - (BytesPerWord*4) put: valueOopPointer! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'RonieSalgado 4/27/2016 14:02'!
nativeStackSize
	"4 MB native stack size"
	^ 4 *1024 * 1024! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 11/28/2012 11:04'!
noAssertHeaderOf: methodPointer
	^objectMemory fetchPointer: HeaderIndex ofObject: methodPointer! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 8/23/2014 11:43'!
noFixupSuperclassOf: classPointer
	"No fixup version of superclassOf: that should not create heisenbugs"
	<inline: true>
	^objectMemory noFixupFollowField: SuperclassIndex ofObject: classPointer! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 11/16/2016 18:40'!
noInlineSigned32BitIntegerGutsFor: integerValue
	"Answer a full 32 bit integer object for the given integer value which
	 is known not to be a SmallInteger value.  This serves to share the
	 code for creating a four byte LargeInteger in one place."
	<notOption: #Spur64BitMemoryManager>
	| newLargeInteger magnitude largeClass |
	<inline: #never>
	<var: #magnitude type: #'unsigned int'>
	self deny: objectMemory hasSixtyFourBitImmediates.
	self deny: (objectMemory isIntegerValue: integerValue).
	 integerValue < 0
		ifTrue: [largeClass := ClassLargeNegativeIntegerCompactIndex.
				magnitude := 0 asUnsignedInteger - integerValue]
		ifFalse: [largeClass := ClassLargePositiveIntegerCompactIndex.
				magnitude := integerValue].
	newLargeInteger := objectMemory
							eeInstantiateSmallClassIndex: largeClass
							format: (objectMemory byteFormatForNumBytes: 4)
							numSlots: 1.
	objectMemory storeLong32: 0 ofObject: newLargeInteger withValue: (objectMemory byteSwapped32IfBigEndian: magnitude).
	SPURVM ifTrue: "Memory is 8 byte aligned in Spur, make sure that oversized bytes are set to zero" "eem 4/28/2016 questionable; they should never be read"
		[objectMemory storeLong32: 1 ofObject: newLargeInteger withValue: 0].
	^newLargeInteger! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 8/19/2011 13:44'!
noInlineTemporary: offset in: theFP
	<var: #theFP type: #'char *'>
	<inline: false>
	^self temporary: offset in: theFP! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 12/7/2012 13:54'!
noInlineTemporary: offset in: theFP put: valueOop
	<var: #theFP type: #'char *'>
	<inline: false>
	^self temporary: offset in: theFP put: valueOop! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 2/27/2013 13:50'!
noMarkedContextsOnPage: thePage
	"Answer true if there are no marked contexts on thePage."
	| theFP |
	<var: #thePage type: #'StackPage *'>
	<var: #theFP type: #'char *'>
	<inline: false>
	self assert: (stackPages isFree: thePage) not.
	theFP := thePage  headFP.
	[(self frameHasContext: theFP) ifTrue:
		[self assert: (objectMemory isContext: (self frameContext: theFP)).
		 (objectMemory isMarked: (self frameContext:  theFP)) ifTrue:
			[^false]].
	(theFP := self frameCallerFP: theFP) ~= 0] whileTrue.
	^true! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'rmacnak 5/19/2015 20:27'!
normalSend
	"Send a message, starting lookup with the receiver's class."
	"Assume: messageSelector and argumentCount have been set, and that 
	the receiver and arguments have been pushed onto the stack,"
	"Note: This method is inlined into the interpreter dispatch loop."
	<sharedCodeInCase: #singleExtendedSendBytecode>
	| rcvr |
	rcvr := self internalStackValue: argumentCount.
	lkupClassTag := objectMemory fetchClassTagOf: rcvr.
	self assert: lkupClassTag ~= objectMemory nilObject.
	self commonSendOrdinary! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 10/24/2017 11:28'!
nullaryInlinePrimitive: primIndex
	<option: #SistaVM>
	self unknownInlinePrimitive! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 10/26/2014 17:11'!
numSlotsOfMarriedContext: aContext
	"Answer the number of effective pointer fields in the given context."
	<inline: true>
	| contextSize |
	contextSize := self stackPointerIndexForFrame: (self frameOfMarriedContext: aContext).
	^CtxtTempFrameStart + contextSize! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 8/14/2015 18:12'!
numStkPages
	<inline: true>
	^numStackPages! !

!StackInterpreter methodsFor: 'debug support' stamp: 'cb 8/31/2016 10:36'!
objCouldBeClassObj: objOop
	"Answer if objOop looks like a class object.  WIth Spur be lenient if the object doesn't
	 yet have a hash (i.e. is not yet in the classTable), and accept forwarding pointers."
	<api>
	<inline: false>
	| fieldOop |
	^(objectMemory isPointersNonImm: objOop)
	  and: [(objectMemory numSlotsOfAny: objOop) > InstanceSpecificationIndex
	  and: [fieldOop := objectMemory fetchPointer: SuperclassIndex ofObject: objOop.
			((objectMemory addressCouldBeObj: fieldOop)
			and:[ (objectMemory isPointersNonImm: fieldOop)
				or: [(objectMemory isOopForwarded: fieldOop)
					and: [objectMemory isPointers: (objectMemory followForwarded: fieldOop)]]])
	  and: [fieldOop := objectMemory fetchPointer: MethodDictionaryIndex ofObject: objOop.
			((objectMemory addressCouldBeObj: fieldOop)
			and:[ (objectMemory isPointersNonImm: fieldOop)
				or: [(objectMemory isOopForwarded: fieldOop)
					and: [objectMemory isPointers: (objectMemory followForwarded: fieldOop)]]])
	  and: [(objectMemory isIntegerObject: (objectMemory fetchPointer: InstanceSpecificationIndex ofObject: objOop))]]]]! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 8/8/2021 07:52'!
object: anOop equalsString: aCString
	^self object: anOop equalsString: aCString ofSize: (self strlen: aCString)! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem 8/9/2021 10:30'!
object: anOop equalsString: aCString ofSize: aCStringStrlen
	| size |
	^(objectMemory isBytes: anOop)
	 and: [(size := objectMemory numBytesOfBytes: anOop) = aCStringStrlen
	 and: [(self strncmp: aCString _: (objectMemory firstIndexableField: anOop)  _: aCStringStrlen) = 0]]! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 3/16/2018 15:33'!
objectArg: index
	"Like #stackObjectValue: but access method arguments left-to-right"
	<doNotGenerate> "Obsolete; was never used; replaced with methodReturnFloat: in the VirtualMachine struct."
	| oop |
	self deprecated.
	oop := self methodArg: index.
	oop = 0 ifTrue:[^0]. "methodArg: failed"
	(objectMemory isImmediate: oop) ifTrue: [self primitiveFail. ^ nil].
	^oop! !

!StackInterpreter methodsFor: 'simulation support' stamp: 'eem 10/9/2009 03:07'!
objectMemory
	<doNotGenerate>
	^objectMemory! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 10/11/2013 16:34'!
objectMemory: anObjectMemory
	<doNotGenerate>
	objectMemory ifNotNil: [self halt].
	objectMemory := anObjectMemory! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 08/08/2014 10:01'!
okayFields: oop
	"Check if the argument is an ok object.
	 If this is a pointers object, check that its fields are all okay oops."

	| i fieldOop |
	(oop = nil or: [oop = 0]) ifTrue: [ ^true ].
	(objectMemory isIntegerObject: oop) ifTrue: [ ^true ].
	(objectMemory okayOop: oop) ifFalse: [ ^false ].
	(objectMemory oopHasOkayClass: oop) ifFalse: [ ^false ].
	((objectMemory isPointersNonImm: oop) or: [objectMemory isCompiledMethod: oop]) ifFalse: [ ^true ].
	(objectMemory isCompiledMethod: oop)
		ifTrue:
			[i := (objectMemory literalCountOf: oop) + LiteralStart - 1]
		ifFalse:
			[(objectMemory isContext: oop)
				ifTrue: [i := CtxtTempFrameStart + (self fetchStackPointerOf: oop) - 1]
				ifFalse: [i := (objectMemory lengthOf: oop) - 1]].
	[i >= 0] whileTrue: [
		fieldOop := objectMemory fetchPointer: i ofObject: oop.
		(objectMemory isIntegerObject: fieldOop) ifFalse: [
			(objectMemory okayOop: fieldOop) ifFalse: [ ^false ].
			(self oopHasOkayClass: fieldOop) ifFalse: [ ^false ].
		].
		i := i - 1.
	].
	^true! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 10/31/2014 11:45'!
okayInterpreterObjects

	| oopOrZero oop |
	self okayFields: objectMemory nilObject.
	self okayFields: objectMemory falseObject.
	self okayFields: objectMemory trueObject.
	self okayFields: objectMemory specialObjectsOop.
	self okayFields: messageSelector.
	self okayFields: newMethod.
	self okayFields: lkupClass.
	0 to: MethodCacheEntries - 1 by: MethodCacheEntrySize do: [ :i |
		oopOrZero := methodCache at: i + MethodCacheSelector.
		oopOrZero = 0 ifFalse: [
			self okayFields: (methodCache at: i + MethodCacheSelector).
			self okayFields: (methodCache at: i + MethodCacheClass).
			self okayFields: (methodCache at: i + MethodCacheMethod).
		].
	].
	1 to: objectMemory remapBufferCount do: [ :i |
		oop := objectMemory remapBuffer at: i.
		(objectMemory isImmediate: oop) ifFalse: [
			self okayFields: oop.
		].
	].
	self okayStackZone.! !

!StackInterpreter methodsFor: 'simulation support' stamp: 'eem 10/16/2018 09:05'!
openImageFileNamed: fileName
	"Attempt to open fileName or fileName, '.image'"
	<doNotGenerate>
	| f |
	f := [FileStream readOnlyFileNamed: fileName]
			on: FileDoesNotExistException
			do: [:ex|
				 ((fileName endsWith: '.image') not
				  and: [FileDirectory default fileExists: fileName, '.image']) ifFalse:
					[ex pass].
				 FileStream readOnlyFileNamed: fileName, '.image'].
	^f ifNil: [self error: 'no image found'. nil].! !

!StackInterpreter methodsFor: 'simulation' stamp: 'eem 3/23/2015 11:42'!
openOn: fileName
	"(StackInterpreterSimulator new openOn: 'clonex.image') openAsMorph; run"
	<doNotGenerate>
	self openOn: fileName extraMemory: 2500000! !

!StackInterpreter methodsFor: 'vm scheduling' stamp: 'eem 8/24/2018 16:43'!
ownVM: threadIndexAndFlags
	<api>
	<inline: false>
	"This is the entry-point for plugins and primitives that wish to reacquire the VM after having
	 released it via disownVM or callbacks that want to acquire it without knowing their ownership
	 status.  While this exists for the threaded FFI VM we use it to reset newMethod and the
	 argumentCount after a callback.

	 Answer -1 if the current thread is unknown to the VM and fails to take ownership."
	<var: 'amInVMThread' declareC: 'extern sqInt amInVMThread(void)'>
	self cppIf: COGMTVM
		ifTrue:
			[self amInVMThread ifFalse:
				[^-1]].

	self assert: ((objectMemory isOopCompiledMethod: newMethod)
				and: [(self argumentCountOf: newMethod) = argumentCount]).

	"Hack encodings of client state.  We use non-immediates (bottom three bits clear)
	 for FFI/Plugin doing
		save := self disownVM: FLAGS. ... callout ... self ownVM: save.
	 We use immediate integer (bottom bit 1) for callbacks doing
		save := self ownVM: 0. ... callback ... self disownVM: save. return to C"

	"If DisownVMForFFICall this is from the FFI plugin and we're returning from a callout."
	(threadIndexAndFlags anyMask: DisownVMForFFICall) ifTrue:
		[inFFIFlags := 0.
		 ^threadIndexAndFlags].

	"Otherwise this is a callback; stash newMethod on the stack and encode
	 argumentCount in the flags retrieved when the calback calls disownVM:."
	self assert: primFailCode = 0.
	self push: newMethod.
	^objectMemory integerObjectOf: argumentCount! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 08/08/2014 10:01'!
penultimateLiteralOf: aMethodOop
	<api>
	self assert: (objectMemory isOopCompiledMethod: aMethodOop).
	^self literal: (objectMemory literalCountOf: aMethodOop) - 2 ofMethod: aMethodOop! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
pop2AndPushIntegerIfOK: integerResult

	self successful ifTrue:
		[(objectMemory isIntegerValue: integerResult)
			ifTrue: [self pop: 2 thenPush: (objectMemory integerObjectOf: integerResult)]
			ifFalse: [self success: false]]! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
pop: nItems
	<inline: true>
	"In the StackInterpreter stacks grow down."
	stackPointer := stackPointer + (nItems*objectMemory wordSize).
	^nil! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
pop: nItems thenPush: oop
	"In the StackInterpreter stacks grow down."
	| sp |
	<inline: true>
	<var: #sp type: #'char *'>
	stackPages longAt: (sp := stackPointer + ((nItems - 1) * objectMemory wordSize)) put: oop.
	stackPointer := sp! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
pop: nItems thenPushBool: trueOrFalse
	"In the StackInterpreter stacks grow down."
	| sp |
	<inline: true>
	<var: #sp type: #'char *'>
	stackPages
		longAt: (sp := stackPointer + ((nItems - 1) * objectMemory wordSize))
		put: (objectMemory booleanObjectOf: trueOrFalse).
	stackPointer := sp! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 11/21/2014 16:20'!
pop: nItems thenPushFloat: f
	"In the StackInterpreter stacks grow down."
	| sp |
	<inline: true>
	<var: #f type: #double>
	<var: #sp type: #'char *'>
	stackPages
		longAt: (sp := stackPointer + ((nItems - 1) * objectMemory wordSize))
		put: (objectMemory floatObjectOf: f).
	stackPointer := sp! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
pop: nItems thenPushInteger: integerVal
	"lots of places pop a few items off the stack and then push an integer. Make it convenient.
	 In the StackInterpreter stacks grow down."
	| sp |
	<var: #sp type: #'char *'>
	stackPages
		longAt: (sp := stackPointer + ((nItems - 1) * objectMemory wordSize))
		put: (objectMemory integerObjectOf: integerVal).
	stackPointer := sp! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 11/21/2014 17:14'!
popFloat
	<returnTypeC: #double>
	^objectMemory floatValueOf: self popStack! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: ''!
popPos32BitInteger
	"May set successFlag, and return false if not valid"

	| top |
	top := self popStack.
	^ self positive32BitValueOf: top! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
popStack
	"In the StackInterpreter stacks grow down."
	<api>
	| top |
	<inline: true>
	top := stackPages longAt: stackPointer.
	stackPointer := stackPointer + objectMemory wordSize.
	^top! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:47'!
popStackBytecode

	self fetchNextBytecode.
	self internalPop: 1.
! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'nice 7/8/2016 02:41'!
positive32BitIntegerFor: integerValue
	"integerValue is interpreted as POSITIVE, e.g. as the result of Bitmap>at:.
	 N.B.  Returning in each arm separately enables Slang inlining.
	 /Don't/ return the ifTrue:ifFalse: unless Slang inlining of conditionals is fixed."
	<inline: true>
	<var: 'integerValue' type: #'unsigned int'>
	objectMemory hasSixtyFourBitImmediates
		ifTrue:
			[^objectMemory integerObjectOf: (integerValue asUnsignedInteger bitAnd: 16rFFFFFFFF)]
		ifFalse:
			[^self maybeInlinePositive32BitIntegerFor: integerValue]! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'nice 4/3/2016 09:18'!
positive64BitIntegerFor: integerValue
	<api>
	<var: 'integerValue' type: #usqLong>
	<var: 'highWord' type: #'unsigned int'>
	"Answer a Large Positive Integer object for the given integer value.  N.B. will *not* cause a GC."
	| newLargeInteger highWord sz |
	objectMemory hasSixtyFourBitImmediates
		ifTrue:
			[(self cCode: [integerValue] inSmalltalk: [integerValue bitAnd: 1 << 64 - 1]) <= objectMemory maxSmallInteger ifTrue:
				[^objectMemory integerObjectOf: integerValue].
			 sz := 8]
		ifFalse:
			[(highWord := integerValue >> 32) = 0 ifTrue:
				[^self positive32BitIntegerFor: integerValue].
			 sz := 5.
			 (highWord := highWord >> 8) = 0 ifFalse:
				[sz := sz + 1.
				 (highWord := highWord >> 8) = 0 ifFalse:
					[sz := sz + 1.
					 (highWord := highWord >> 8) = 0 ifFalse:[sz := sz + 1]]]].
	newLargeInteger := objectMemory
							eeInstantiateSmallClassIndex: ClassLargePositiveIntegerCompactIndex
							format: (objectMemory byteFormatForNumBytes: sz)
							numSlots: 8 / objectMemory bytesPerOop.
	objectMemory storeLong64: 0 ofObject: newLargeInteger withValue: (objectMemory byteSwapped64IfBigEndian: integerValue).
	^newLargeInteger
! !

!StackInterpreter methodsFor: 'callback support' stamp: 'nice 11/1/2016 21:36'!
positiveMachineIntegerFor: value
	<var: #value type: #'usqIntptr_t'>
	<inline: false>
	| resultObj |
	objectMemory wordSize = 8
		ifTrue: [resultObj := self positive64BitIntegerFor: value]
		ifFalse: [resultObj := self positive32BitIntegerFor: value].
	^resultObj! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 1/26/2015 11:49'!
postBecomeAction: theBecomeEffectsFlags
	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[self spurPostBecomeAction: theBecomeEffectsFlags]! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 1/15/2018 13:08'!
postGCAction: gcModeArg
	"Signal the gc semaphore, and inform the display subsystem if the display bitmap has moved."

	self signalSemaphoreWithIndex: gcSemaphoreIndex.
	self postGCUpdateDisplayBits! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 6/18/2021 16:23'!
postGCUpdateDisplayBits
	"Update the displayBits after a GC may have moved it.
	 Answer if the displayBits appear valid.  The wrinkle here is that the displayBits could be a surface handle."
	<inline: false>
	| displayObj bitsOop bitsNow |
	displayObj := objectMemory splObj: TheDisplay.
	((objectMemory isPointers: displayObj)
	 and: [(objectMemory lengthOf: displayObj) >= 4]) ifFalse:
		[^false].
	
	bitsOop := objectMemory fetchPointer: 0 ofObject: displayObj.
	(bitsOop = objectMemory nilObject "it ain't yet set"
	 or: [objectMemory isIntegerObject: bitsOop]) ifTrue: "It's a surface; our work here is done..."
		[^true].

	self assert: ((objectMemory addressCouldBeObj: bitsOop)
				 and: [objectMemory isWordsOrBytes: bitsOop]).

	(objectMemory hasSpurMemoryManagerAPI
	 and: [objectMemory isPinned: bitsOop]) ifFalse:
		[bitsNow := self cCode: [objectMemory firstIndexableField: bitsOop]
					inSmalltalk: [(objectMemory firstIndexableField: bitsOop) asInteger].
		 displayBits ~= bitsNow ifTrue:
			[displayBits := bitsNow.
			 self ioNoteDisplayChanged: displayBits width: displayWidth height: displayHeight depth: displayDepth].
		 (objectMemory hasSpurMemoryManagerAPI
		  and: [stackPage ~= 0]) ifTrue: "If stackPage is zero we're snapshotting and now is not the time to pin."
			[objectMemory pinObject: bitsOop]].
	^true! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 8/5/2011 13:13'!
preBecomeAction
	"Need to write back the frame pointers unless all pages are free (as in snapshot)"
	stackPage ~= 0 ifTrue:
		[self externalWriteBackHeadFramePointers]! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 11/10/2014 12:49'!
preGCAction: gcModeArg
	"Need to write back the frame pointers unless all pages are free (as in snapshot)"
	stackPage ~= 0 ifTrue:
		[self externalWriteBackHeadFramePointers].
	(gcModeArg = GCModeFull
	 and: [objectMemory hasSpurMemoryManagerAPI]) ifTrue:
		[self flushMethodCache]! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 8/2/2021 22:30'!
primitiveAccessorDepthForExternalPrimitiveMethod: methodObj
	^(objectMemory integerValueOf:
		(objectMemory
			fetchPointer: ExternalCallLiteralFlagsIndex
			ofObject: (self literal: 0 ofMethod: methodObj))) >>> SpurPrimitiveAccessorDepthShift! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 9/6/2019 11:57'!
primitiveDoMixedArithmetic
	"If primitiveDoMixedArithmetic is true, then primitive can handle the conversions:
	SmallInteger arithmeticOp: Float (Small or Boxed)
	SmallInteger compareOp: Float (Small or Boxed)
	Else, the primitive fail in case of mixed arithmetic, and conversion should be performed in the image."
	<api>
	<cmacro: '() primitiveDoMixedArithmetic'>
	^primitiveDoMixedArithmetic! !

!StackInterpreter methodsFor: 'I/O primitives' stamp: 'eem 11/1/2013 10:19'!
primitiveEventProcessingControl
	"With 0 args answers whether ioProcessEvents is enabled and being called.
	 With 1 arg expects a boolean which will enable ioProcessEvents if true and
	 disable it if false, answering its previous state."
	<export: true>
	| enabled |
	enabled := inIOProcessEvents >= 0.
	argumentCount = 0 ifTrue:
		[^self pop: 1 thenPushBool: enabled].
	argumentCount = 1 ifTrue:
		[self stackTop = objectMemory trueObject
			ifTrue: [inIOProcessEvents < 0 ifTrue:
					[inIOProcessEvents := 0]]
			ifFalse:
				[self stackTop = objectMemory falseObject
					ifTrue: [inIOProcessEvents := -1]
					ifFalse: [self primitiveFailFor: PrimErrBadArgument.
							^self]].
		 ^self pop: 2 thenPushBool: enabled].
	self primitiveFailFor: PrimErrBadNumArgs! !

!StackInterpreter methodsFor: 'arithmetic float primitives' stamp: 'eem 12/18/2014 13:24'!
primitiveFloatAdd: rcvrOop toArg: argOop
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory loadFloatOrIntFrom: rcvrOop.
	arg := objectMemory loadFloatOrIntFrom: argOop.
	self successful ifTrue:
		[self pop: 2 thenPushFloat: rcvr + arg]! !

!StackInterpreter methodsFor: 'arithmetic float primitives' stamp: 'eem 12/18/2014 13:24'!
primitiveFloatDivide: rcvrOop byArg: argOop
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory loadFloatOrIntFrom: rcvrOop.
	arg := objectMemory loadFloatOrIntFrom: argOop.
	self success: arg ~= 0.0.
	self successful ifTrue:
		[self pop: 2 thenPushFloat: rcvr / arg]! !

!StackInterpreter methodsFor: 'comparison float primitives' stamp: 'eem 9/5/2019 09:59'!
primitiveFloatEqual: rcvrOop toArg: argOop
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory loadFloatOrIntFrom: rcvrOop.
	arg := objectMemory loadFloatOrIntFrom: argOop.
	(objectMemory wordSize > 4
	 and: [rcvr = arg]) ifTrue:
		[(objectMemory isIntegerObject: argOop) ifTrue: "Resolve case of ambiguity so as to have comparison of exact values"
			[^rcvr asInteger = (objectMemory integerValueOf: argOop)].
		 (objectMemory isIntegerObject: rcvrOop) ifTrue:
			["Same when used from bytecodePrim...
			  note that rcvr and arg cannot be both integer (case is already handled)"
			^(objectMemory integerValueOf: rcvrOop) = arg asInteger]].
	^rcvr = arg! !

!StackInterpreter methodsFor: 'comparison float primitives' stamp: 'eem 9/5/2019 09:58'!
primitiveFloatGreater: rcvrOop thanArg: argOop
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory loadFloatOrIntFrom: rcvrOop.
	arg := objectMemory loadFloatOrIntFrom: argOop.
	(objectMemory wordSize > 4
	 and: [rcvr = arg]) ifTrue:
		[(objectMemory isIntegerObject: argOop) ifTrue: "Resolve case of ambiguity so as to have comparison of exact values"
			[^rcvr asInteger > (objectMemory integerValueOf: argOop)].
		 (objectMemory isIntegerObject: rcvrOop) ifTrue:
			["Same when used from bytecodePrim...
			  note that rcvr and arg cannot be both integer (case is already handled)"
			^(objectMemory integerValueOf: rcvrOop) > arg asInteger]].
	^rcvr > arg! !

!StackInterpreter methodsFor: 'comparison float primitives' stamp: 'eem 9/5/2019 09:59'!
primitiveFloatGreaterOrEqual: rcvrOop toArg: argOop
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory loadFloatOrIntFrom: rcvrOop.
	arg := objectMemory loadFloatOrIntFrom: argOop.
	(objectMemory wordSize > 4
	 and: [rcvr = arg]) ifTrue:
		[(objectMemory isIntegerObject: argOop) ifTrue: "Resolve case of ambiguity so as to have comparison of exact values"
			[^rcvr asInteger >= (objectMemory integerValueOf: argOop)].
		 (objectMemory isIntegerObject: rcvrOop) ifTrue:
			["Same when used from bytecodePrim...
			  note that rcvr and arg cannot be both integer (case is already handled)"
			^(objectMemory integerValueOf: rcvrOop) >= arg asInteger]].
	^rcvr >= arg! !

!StackInterpreter methodsFor: 'comparison float primitives' stamp: 'eem 9/5/2019 09:59'!
primitiveFloatLess: rcvrOop thanArg: argOop
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory loadFloatOrIntFrom: rcvrOop.
	arg := objectMemory loadFloatOrIntFrom: argOop.
	(objectMemory wordSize > 4
	 and: [rcvr = arg]) ifTrue:
		[(objectMemory isIntegerObject: argOop) ifTrue: "Resolve case of ambiguity so as to have comparison of exact values"
			[^rcvr asInteger < (objectMemory integerValueOf: argOop)].
		 (objectMemory isIntegerObject: rcvrOop) ifTrue:
			["Same when used from bytecodePrim...
			  note that rcvr and arg cannot be both integer (case is already handled)"
			^(objectMemory integerValueOf: rcvrOop) < arg asInteger]].
	^rcvr < arg! !

!StackInterpreter methodsFor: 'comparison float primitives' stamp: 'eem 9/5/2019 09:59'!
primitiveFloatLessOrEqual: rcvrOop toArg: argOop
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory loadFloatOrIntFrom: rcvrOop.
	arg := objectMemory loadFloatOrIntFrom: argOop.
	(objectMemory wordSize > 4
	 and: [rcvr = arg]) ifTrue:
		[(objectMemory isIntegerObject: argOop) ifTrue: "Resolve case of ambiguity so as to have comparison of exact values"
			[^rcvr asInteger <= (objectMemory integerValueOf: argOop)].
		 (objectMemory isIntegerObject: rcvrOop) ifTrue:
			["Same when used from bytecodePrim...
			  note that rcvr and arg cannot be both integer (case is already handled)"
			^(objectMemory integerValueOf: rcvrOop) <= arg asInteger]].
	^rcvr <= arg! !

!StackInterpreter methodsFor: 'arithmetic float primitives' stamp: 'eem 12/18/2014 13:24'!
primitiveFloatMultiply: rcvrOop byArg: argOop
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory loadFloatOrIntFrom: rcvrOop.
	arg := objectMemory loadFloatOrIntFrom: argOop.
	self successful ifTrue:
		[self pop: 2 thenPushFloat: rcvr * arg]! !

!StackInterpreter methodsFor: 'arithmetic float primitives' stamp: 'eem 12/18/2014 13:25'!
primitiveFloatSubtract: rcvrOop fromArg: argOop
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := objectMemory loadFloatOrIntFrom: rcvrOop.
	arg := objectMemory loadFloatOrIntFrom: argOop.
	self successful ifTrue:
		[self pop: 2 thenPushFloat: rcvr - arg]! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 08/08/2014 10:42'!
primitiveIndexOf: methodPointer
	<api>
	^self primitiveIndexOfMethod: methodPointer header: (objectMemory methodHeaderOf: methodPointer)! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 6/7/2016 14:57'!
primitiveIndexOfMethod: theMethod header: methodHeader
	"Note: With the Squeak V3 format we now have 10 bits of primitive index, but they are
	 in two places for temporary backward compatibility.  The time to unpack is negligible,
	 since the derived primitive function pointer is stored in the method cache.  With the
	 Spur format we assume a 3-byte CallPrimitive with a little-endian 16-bit primitive index."
	<api>
	<inline: true>
	| firstBytecode |
	^objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[(self alternateHeaderHasPrimitiveFlag: methodHeader)
				ifTrue:
					[firstBytecode := self firstBytecodeOfAlternateHeader: methodHeader method: theMethod.
					 (objectMemory byteAt: firstBytecode + 1) + ((objectMemory byteAt: firstBytecode + 2) << 8)]
				ifFalse:
					[0]]
		ifFalse:
			[MULTIPLEBYTECODESETS
				ifTrue:
					[(objectMemory headerIndicatesAlternateBytecodeSet: methodHeader)
						ifTrue:
							[(self alternateHeaderHasPrimitiveFlag: methodHeader)
								ifTrue:
									[firstBytecode := self firstBytecodeOfAlternateHeader: methodHeader method: theMethod.
									 (objectMemory byteAt: firstBytecode + 1) + ((objectMemory byteAt: firstBytecode + 2) << 8)]
								ifFalse:
									[0]]
						ifFalse:
							[| primBits |
							 primBits := objectMemory integerValueOf: methodHeader.
							 (primBits bitAnd: 16r1FF) + (primBits >> 19 bitAnd: 16r200)]]
				ifFalse:
					[| primBits |
					 primBits := objectMemory integerValueOf: methodHeader.
					 (primBits bitAnd: 16r1FF) + (primBits >> 19 bitAnd: 16r200)]]! !

!StackInterpreter methodsFor: 'control primitives' stamp: 'eem 2/17/2017 09:41'!
primitiveObject: actualReceiver perform: selector withArguments: argumentArray lookedUpIn: lookupClassOrNil
	"Common routine used by perform:withArgs:, perform:withArgs:inSuperclass:,
	 object:perform:withArgs:inClass: et al.  Answer nil on success.

	 NOTE:  The case of doesNotUnderstand: is not a failure to perform.
	 The only failures are arg types and consistency of argumentCount.

	 Since we're in the stack VM we can assume there is space to push the arguments
	 provided they are within limits (max argument count is 15).  We can therefore deal
	 with the arbitrary amount of state to remove from the stack (lookup class, selector,
	 mirror receiver) and arbitrary argument orders by deferring popping anything until
	 we know whether the send has succeeded.  So on failure we merely have to remove
	 the actual receiver and arguments pushed, and on success we have to slide the actual
	 receiver and arguments down to replace the original ones."
	<inline: true>
	| arraySize performArgCount delta |
	(objectMemory isArray: argumentArray) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].

	"Check if number of arguments is reasonable; MaxNumArgs isn't available
	 so just use LargeContextSize"
	arraySize := objectMemory numSlotsOf: argumentArray.
	arraySize > (LargeContextSlots - CtxtTempFrameStart) ifTrue:
		[^self primitiveFailFor: PrimErrBadNumArgs].

	performArgCount := argumentCount.
	"Push newMethod to save it in case of failure,
	 then push the actual receiver and the args in the array."
	self push: newMethod.
	self push: actualReceiver.
	"Copy the arguments to the stack, in case of MNU, and lookup"
	1 to: arraySize do:
		[:index| self push: (objectMemory fetchPointer: index - 1 ofObject: argumentArray)].
	argumentCount := arraySize.
	messageSelector := selector.
	self sendBreakpoint: messageSelector receiver: actualReceiver.
	self printSends ifTrue:
		[self printActivationNameForSelector: messageSelector
				startClass: (lookupClassOrNil isNil
								ifTrue: [objectMemory fetchClassOf: actualReceiver]
								ifFalse: [lookupClassOrNil]);
			cr].
	self findNewMethodInClassTag: (lookupClassOrNil isNil
										ifTrue: [objectMemory fetchClassTagOf: actualReceiver]
										ifFalse: [objectMemory classTagForClass: lookupClassOrNil]).

	"Only test CompiledMethods for argument count - any other objects playacting as CMs will have to take their chances"
	((objectMemory isOopCompiledMethod: newMethod)
	  and: [(self argumentCountOf: newMethod) ~= argumentCount]) ifTrue:
		["Restore the state by popping the array entries, the actual receiver and the saved
		  newMethod, leaving the selector and array, and fail.  N.B.  If an MNU has happened
		  then argumentCount will match newMethod, so this code will not be reached."
		 "These asserts check that an MNU has not occurred if the argumentCount doesn't match the newMethod."
		 self assert: (self stackTop = (arraySize = 0
											ifTrue: [actualReceiver]
											ifFalse: [(objectMemory fetchPointer: arraySize - 1 ofObject: argumentArray)])).
		 self assert:  argumentCount = arraySize.
		 self pop: arraySize + 1.
		 newMethod := self popStack.
		 "Must reset primitiveFunctionPointer for checkForAndFollowForwardedPrimitiveState"
		 objectMemory hasSpurMemoryManagerAPI ifTrue:
			[argumentCount := performArgCount.
			 primitiveFunctionPointer := lookupClassOrNil
												ifNil: [#primitivePerformWithArgs]
												ifNotNil: [#primitivePerformInSuperclass]].
		 ^self primitiveFailFor: PrimErrBadNumArgs].

	"Cannot fail this primitive from here-on.  Slide the actual receiver and arguments down
	 to replace the perform arguments and saved newMethod and then execute the new
	 method. Use argumentCount not arraySize because an MNU may have changed it."
	delta := objectMemory wordSize * (performArgCount + 2). "+2 = receiver + saved newMethod"
	argumentCount * objectMemory wordSize to: 0 by: objectMemory wordSize negated do:
		[:offset|
		stackPages
			longAt: stackPointer + offset + delta
			put: (stackPages longAt: stackPointer + offset)].
	self pop: performArgCount + 2.
	self executeNewMethod.
	self initPrimCall.  "Recursive xeq affects primErrorCode"
	^nil! !

!StackInterpreter methodsFor: 'simulation support' stamp: 'eem 8/15/2021 17:38'!
primitiveTable
	<doNotGenerate>
	^primitiveTable! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 1/17/2013 12:42'!
print: s
	"For testing in Smalltalk, this method should be overridden in a subclass."
	<api>
	<var: #s type: #'char *'>
	self cCode: 'fputs(s, stdout)'! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 12/29/2019 14:06'!
printActivationNameFor: aMethod receiver: anObject isBlock: isBlock firstTemporary: maybeMessage
	| methClass methodSel classObj |
	<inline: false>
	isBlock ifTrue:
		[self print: '[] in '].
	methClass := self findClassOfMethod: aMethod forReceiver: anObject.
	methodSel := self findSelectorOfMethod: aMethod.
	((objectMemory addressCouldBeOop: anObject)
	 and: [(objectMemory isOopForwarded: anObject) not
	 and: [self addressCouldBeClassObj: (classObj := objectMemory fetchClassOf: anObject)]])
		ifTrue:
			[(classObj = methClass or: [methClass isNil or: [methClass = objectMemory nilObject] "i.e. doits"])
				ifTrue: [self printNameOfClass: classObj count: 5]
				ifFalse:
					[self printNameOfClass: classObj count: 5.
					 self print: '('.
					 self printNameOfClass: methClass count: 5.
					 self print: ')']]
		ifFalse:
			[self print: 'INVALID RECEIVER'].
	self print: '>'.
	(objectMemory addressCouldBeOop: methodSel)
		ifTrue:
			[methodSel = objectMemory nilObject
				ifTrue: [self print: '(nil)']
				ifFalse: [self printStringOf: methodSel]]
		ifFalse: [self print: 'INVALID SELECTOR'].
	(methodSel = (objectMemory splObj: SelectorDoesNotUnderstand)
	and: [(objectMemory addressCouldBeObj: maybeMessage)
	and: [(objectMemory fetchClassOfNonImm: maybeMessage) = (objectMemory splObj: ClassMessage)]]) ifTrue:
		["print arg message selector"
		methodSel := objectMemory fetchPointer: MessageSelectorIndex ofObject: maybeMessage.
		self print: ' '.
		self printStringOf: methodSel]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 10/4/2013 14:21'!
printActivationNameForSelector: aSelector startClass: startClass
	| methClass |
	<inline: false>
	(objectMemory addressCouldBeObj: startClass)
		ifTrue:
			[self findClassForSelector: aSelector
				lookupClass: startClass
				do: [:class| methClass := class].
			(methClass isNil or: [startClass = methClass])
				ifTrue:
					[self printNameOfClass: methClass count: 5.
					 self printChar: $>.
					 methClass ifNil:
						[self printStringOf: (objectMemory splObj: SelectorDoesNotUnderstand).
						 self print: ' ']]
				ifFalse:
					[self printNameOfClass: startClass count: 5.
					 self printChar: $(.
					 self printNameOfClass: methClass count: 5.
					 self printChar: $).
					 self printChar: $>]]
		ifFalse: [self print: 'INVALID CLASS'].
	(objectMemory addressCouldBeOop: aSelector)
		ifTrue:
			[(objectMemory isBytes: aSelector)
				ifTrue: [self printStringOf: aSelector]
				ifFalse: [self printOopShort: aSelector]]
		ifFalse: [self print: 'INVALID SELECTOR']! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:48'!
printAllStacks
	"Print all the stacks of all running processes, including those that are currently suspended."
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| proc semaphoreClass mutexClass schedLists p processList |
	<inline: false>
	proc := self activeProcess.
	self printNameOfClass: (objectMemory fetchClassOf: proc) count: 5; space; printHex: proc.
	self print: ' priority '; printNum: (self quickFetchInteger: PriorityIndex ofObject: proc); cr.
	self printCallStackFP: framePointer. "first the current activation"
	schedLists := objectMemory fetchPointer: ProcessListsIndex ofObject: self schedulerPointer.
	"then the runnable processes"
	p := highestRunnableProcessPriority = 0
			ifTrue: [objectMemory numSlotsOf: schedLists]
			ifFalse: [highestRunnableProcessPriority].
	p - 1 to: 0 by: -1 do:
		[:pri|
		processList := objectMemory fetchPointer: pri ofObject: schedLists.
		(self isEmptyList: processList) ifFalse:
			[self cr; print: 'processes at priority '; printNum: pri + 1.
			 self printProcsOnList: processList]].
	self cr; print: 'suspended processes'.
	semaphoreClass := objectMemory classSemaphore.
	mutexClass := objectMemory classMutex.
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[semaphoreClass := objectMemory compactIndexOfClass: semaphoreClass.
			 mutexClass := objectMemory compactIndexOfClass: mutexClass.
			 objectMemory allHeapEntitiesDo:
				[:obj| | classIdx |
				 classIdx := objectMemory classIndexOf: obj.
				 (classIdx = semaphoreClass
				  or: [classIdx = mutexClass]) ifTrue:
					[self printProcsOnList: obj]]]
		ifFalse:
			[objectMemory allObjectsDoSafely:
				[:obj| | classObj |
				 classObj := objectMemory fetchClassOfNonImm: obj.
				 (classObj = semaphoreClass
				  or: [classObj = mutexClass]) ifTrue:
					[self printProcsOnList: obj]]]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 10/2/2013 11:27'!
printAtCache
	0 to: AtCacheTotalSize - 1 by: 4 do:
		[:i | | obj sz fmt fixed |
		obj := atCache at: i + AtCacheOop.
		sz := atCache at: i + AtCacheSize.
		fmt := atCache at: i + AtCacheFmt.
		fixed := atCache at: i + AtCacheFixedFields.
		(objectMemory addressCouldBeObj: obj) ifTrue:
			[self transcript ensureCr.
			 self print: i; tab; print: (i < AtPutBase ifTrue: ['at   '] ifFalse: ['put ']);
				tab; printNum: sz; tab; printNum: fmt; tab; printNum: fixed; tab;
				shortPrintOop: obj]]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:48'!
printCallStack
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	<inline: false>
	framePointer = nil
		ifTrue: [self printCallStackOf: (objectMemory fetchPointer: SuspendedContextIndex ofObject: self activeProcess)]
		ifFalse: [self printCallStackFP: framePointer]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
printCallStackFP: theFP
	| context |
	<inline: false>
	<var: #theFP type: #'char *'>
	context := self shortReversePrintFrameAndCallers: theFP.
	[context = objectMemory nilObject] whileFalse:
		[(self isMarriedOrWidowedContext: context)
			ifTrue:
				[(self checkIsStillMarriedContext: context currentFP: framePointer) ifFalse:
					[self shortPrintContext: context.
					 ^nil].
				 context := self shortReversePrintFrameAndCallers: (self frameOfMarriedContext: context)]
			ifFalse:
				[context := self printContextCallStackOf: context]]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:42'!
printCallStackOf: aContextOrProcessOrFrame
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| context |
	<inline: false>
	(stackPages couldBeFramePointer: aContextOrProcessOrFrame) ifTrue:
		[^self printCallStackFP: (self cCoerceSimple: aContextOrProcessOrFrame to: #'char *')].
	aContextOrProcessOrFrame = self activeProcess ifTrue:
		[^self printCallStackOf: (self cCode: [framePointer asInteger] inSmalltalk: [self headFramePointer])].
	(self couldBeProcess: aContextOrProcessOrFrame) ifTrue:
		[^self printCallStackOf: (objectMemory
									fetchPointer: SuspendedContextIndex
									ofObject: aContextOrProcessOrFrame)].
	context := aContextOrProcessOrFrame.
	[context = objectMemory nilObject] whileFalse:
		[(self isMarriedOrWidowedContext: context)
			ifTrue:
				[(self checkIsStillMarriedContext: context currentFP: framePointer) ifFalse:
					[self shortPrintContext: context.
					 ^nil].
				 context := self shortReversePrintFrameAndCallers: (self frameOfMarriedContext: context)]
			ifFalse:
				[context := self printContextCallStackOf: context]]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 5/4/2014 17:07'!
printCallStackOf: aContext currentFP: currFP
	| ctxt theFP thePage |
	<inline: false>
	<var: #currFP type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	ctxt := aContext.
	[ctxt = objectMemory nilObject] whileFalse:
		[(self isMarriedOrWidowedContext: ctxt)
			ifFalse:
				[self shortPrintContext: ctxt.
				 ctxt := objectMemory fetchPointer: SenderIndex ofObject: ctxt]
			ifTrue:
				[theFP := self frameOfMarriedContext: ctxt.
				 (self checkIsStillMarriedContext: ctxt currentFP: currFP)
					ifTrue:
						[thePage := stackPages stackPageFor: theFP.
						 (stackPages isFree: thePage) ifTrue:
							[self printHexPtr: theFP; print: ' is on a free page?!!'; cr.
							 ^nil].
						 self shortPrintFrameAndCallers: theFP.
						 theFP := thePage baseFP.
						 ctxt := self frameCallerContext: theFP.
						 (objectMemory isForwarded: ctxt) ifTrue:
							[ctxt := objectMemory followForwarded: ctxt]]
					ifFalse: [self print: 'widowed caller frame '; printHexPtr: theFP; cr.
							^nil]]]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 3/26/2009 13:00'!
printChar: aByte
	<api>
	"For testing in Smalltalk, this method should be overridden in a subclass."
	self putchar: aByte.! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:48'!
printContext: aContext
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| sender ip sp |
	<inline: false>
	self shortPrintContext: aContext.
	sender := objectMemory fetchPointer: SenderIndex ofObject: aContext.
	ip := objectMemory fetchPointer: InstructionPointerIndex ofObject: aContext.
	(objectMemory isIntegerObject: sender)
		ifTrue:
			[(self checkIsStillMarriedContext: aContext currentFP: framePointer)
				ifTrue: [self print: 'married (assuming framePointer valid)'; cr]
				ifFalse: [self print: 'widowed (assuming framePointer valid)'; cr].
			self print: 'sender   '; printNum: sender; print: ' (';
				printHexPtr: (self withoutSmallIntegerTags: sender); printChar: $); cr.
			 self print: 'ip       '; printNum: ip; print: ' (';
				printHexPtr: (self withoutSmallIntegerTags: ip); printChar: $); cr]
		ifFalse:
			[self print: 'sender   '; shortPrintOop: sender.
			 self print: 'ip       '.
			 ip = objectMemory nilObject
				ifTrue: [self shortPrintOop: ip]
				ifFalse: [self printNum: ip; print: ' ('; printNum: (objectMemory integerValueOf: ip); space; printHex: (objectMemory integerValueOf: ip); printChar: $); cr]].
	sp := objectMemory fetchPointer: StackPointerIndex ofObject: aContext.
	self print: 'sp       '; printNum: (objectMemory integerValueOf: sp); print: ' ('; printHex: sp; printChar: $); cr.
	self print: 'method   '; printMethodFieldForPrintContext: aContext.
	self print: 'closure  '; shortPrintOop: (objectMemory fetchPointer: ClosureIndex ofObject: aContext).
	self print: 'receiver '; shortPrintOop: (objectMemory fetchPointer: ReceiverIndex ofObject: aContext).
	sp := objectMemory integerValueOf: sp.
	sp := sp min: (objectMemory lengthOf: aContext) - ReceiverIndex.
	1 to: sp do:
		[:i|
		self print: '       '; printNum: i; space; shortPrintOop: (objectMemory fetchPointer: ReceiverIndex + i ofObject: aContext)]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
printContextCallStackOf: aContext
	"Print the call stack of aContext until it links to a frame."
	| ctxt |
	<inline: false>
	ctxt := aContext.
	[ctxt = objectMemory nilObject or: [self isMarriedOrWidowedContext: ctxt]] whileFalse:
		[self shortPrintContext: ctxt.
		 ctxt := objectMemory fetchPointer: SenderIndex ofObject: ctxt].
	^ctxt! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 12/7/2017 12:17'!
printDecodeMethodHeaderOop: methodHeaderOop
	self printOopShort: methodHeaderOop.
	(self methodHeaderHasPrimitive: methodHeaderOop) ifTrue:
		[self print: ' hasPrim'].
	(self methodHeaderIndicatesLargeFrame: methodHeaderOop) ifTrue:
		[self print: ' largeFrame'].
	(SistaVM and: [self isOptimizedMethodHeader: methodHeaderOop]) ifTrue:
		[self print: ' optimized'].
	(MULTIPLEBYTECODESETS and: [(objectMemory integerValueOf: methodHeaderOop) < 0]) ifTrue:
		[self print: ' altSet'].
	NewspeakVM ifTrue:
		[| s |
		s := (self accessModifierOfMethodHeader: methodHeaderOop) caseOf: {
						[0] -> [' public'].
						[1] -> [' private'].
						[2] -> [' protected'].
						[3] -> [' access undefined'] }.
		 self print: s].
	self print: ' nLits '; printNum: (objectMemory literalCountOfMethodHeader: methodHeaderOop);
		print: ' nArgs '; printNum: (self argumentCountOfMethodHeader: methodHeaderOop);
		print: ' nTemps '; printNum: (self temporaryCountOfMethodHeader: methodHeaderOop)! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:48'!
printExternalHeadFrame
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	<inline: false>
	self printFrame: framePointer WithSP: stackPointer! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/21/2009 11:03'!
printFloat: f
	"For testing in Smalltalk, this method should be overridden in a subclass."
	<cmacro: '(f) printf("%g", f)'>
	self print: f! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 1/9/2020 22:18'!
printForwarder: oop
	<inline: false>
	self
		print: ' is a forwarded object to '; printHex: (objectMemory followForwarded: oop);
		print: ' of slot size '; printNum: (objectMemory numSlotsOfAny: oop).
	objectMemory printHeaderTypeOf: oop.
	self cr.
	^0! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:49'!
printFrame: theFP
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| thePage frameAbove theSP |
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	<var: #frameAbove type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	(stackPages couldBeFramePointer: theFP) ifFalse:
		[((objectMemory addressCouldBeObj: theFP asInteger)
		  and: [(objectMemory isInMemory: theFP asInteger)
		  and: [(objectMemory isContextNonImm: theFP asInteger)
		  and: [(self checkIsStillMarriedContext: theFP asInteger currentFP: framePointer)]]]) ifTrue:
			[^self printFrame: (self frameOfMarriedContext: theFP asInteger)].
		self printHexPtr: theFP; print: ' is not in the stack zone?!!'; cr.
		 ^nil].
	frameAbove := nil.
	theFP = framePointer
		ifTrue: [theSP := stackPointer]
		ifFalse:
			[thePage := stackPages stackPageFor: theFP.
			 (stackPages isFree: thePage) ifTrue:
				[self printHexPtr: theFP; print: ' is on a free page?!!'; cr.
				 ^nil].
			 (thePage ~= stackPage
			  and: [theFP = thePage headFP])
				ifTrue: [theSP := thePage headSP]
				ifFalse:
					[frameAbove := self safeFindFrameAbove: theFP
										on: thePage
										startingFrom: ((thePage = stackPage
														and: [framePointer
																between: thePage realStackLimit
																and: thePage baseAddress])
														ifTrue: [framePointer]
														ifFalse: [thePage headFP]).
					 theSP := frameAbove ifNotNil:
								[self frameCallerSP: frameAbove]]].
	theSP ifNil:
		[self print: 'could not find sp; using bogus value'; cr.
		 theSP := self frameReceiverLocation: theFP].
	self printFrame: theFP WithSP: theSP.
	frameAbove ifNotNil:
		[self printFrameThing: 'frame pc' at: frameAbove + FoxCallerSavedIP]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:42'!
printFrame: theFP WithSP: theSP
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| theMethod numArgs topThing |
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	<var: #addr type: #'char *'>
	self cCode: '' inSmalltalk: [self transcript ensureCr].
	(stackPages couldBeFramePointer: theFP) ifFalse:
		[self printHexPtr: theFP; print: ' is not in the stack zone?!!'; cr.
		 ^nil].
	theMethod := self frameMethod: theFP.
	numArgs := self frameNumArgs: theFP.
	self shortPrintFrame: theFP.
	self printFrameOop: 'rcvr/clsr'
		at: theFP + FoxCallerSavedIP + ((numArgs + 1) * objectMemory wordSize).
	numArgs to: 1 by: -1 do:
		[:i| self printFrameOop: 'arg' at: theFP + FoxCallerSavedIP + (i * objectMemory wordSize)].
	self printFrameThing: 'cllr ip/ctxt' at: theFP + FoxCallerSavedIP.
	self printFrameThing: 'saved fp' at: theFP + FoxSavedFP.
	self printFrameOop: 'method' at: theFP + FoxMethod.
	self printFrameFlagsForFP: theFP.
	self printFrameThing: 'context' at: theFP + FoxThisContext.
	self printFrameOop: 'receiver' at: theFP + FoxReceiver.
	topThing := stackPages longAt: theSP.
	(topThing >= theMethod
	 and: [topThing <= (theMethod + (objectMemory sizeBitsOfSafe: theMethod))])
		ifTrue:
			[theFP + FoxReceiver - objectMemory wordSize to: theSP + objectMemory wordSize by: objectMemory wordSize negated do:
				[:addr|
				self printFrameOop: 'temp/stck' at: addr].
			self printFrameThing: 'frame ip' at: theSP]
		ifFalse:
			[theFP + FoxReceiver - objectMemory wordSize to: theSP by: objectMemory wordSize negated do:
				[:addr|
				self printFrameOop: 'temp/stck' at: addr]]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 3/3/2012 10:08'!
printFrameAndCallers: theFP SP: theSP
	self printFrameAndCallers: theFP SP: theSP short: false! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 2/14/2013 10:51'!
printFrameAndCallers: theFP SP: theSP short: printShort
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	(stackPages couldBeFramePointer: theFP) ifFalse: [^nil].
	(self isBaseFrame: theFP) ifFalse:
		[self printFrameAndCallers: (self frameCallerFP: theFP)
			SP: (self frameCallerSP: theFP)
			short: printShort].
	printShort ifTrue:
		[self shortPrintFrame: theFP.
		 ^nil].
	self cr.
	self printFrame: theFP WithSP: theSP! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 6/26/2012 11:18'!
printFrameFlagsForFP: theFP
	| address it |
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #address type: #'char *'>
	address := theFP + FoxFrameFlags.
	it := stackPages longAt: address.
	self printHexPtr: address;
		print: ':       flags: ';
		printHex: it.
	it ~= 0 ifTrue:
		[self printChar: $=; printNum: it].
	self print: '  numArgs: '; printNum: (self frameNumArgs: theFP);
		print: ((self frameHasContext: theFP) ifTrue: [' hasContext'] ifFalse: [' noContext']);
		print: ((self frameIsBlockActivation: theFP) ifTrue: [' isBlock'] ifFalse: [' notBlock']);
		cr! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 8/11/2011 15:27'!
printFrameOop: name at: address
	| it |
	<inline: false>
	<var: #name type: #'char *'>
	<var: #address type: #'char *'>
	it := stackPages longAt: address.
	self printHexPtr: address;
		printChar: $:.
	1 to: 12 - (self strlen: name) do: [:i| self printChar: $ ].
	self print: name;
		print: ': ';
		printHex: it;
		tab;
		printChar: $=;
		printOopShort: it;
		cr! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 8/11/2011 15:27'!
printFrameOop: name index: idx at: address
	| it |
	<inline: false>
	<var: #name type: #'char *'>
	<var: #address type: #'char *'>
	it := stackPages longAt: address.
	self printHexPtr: address;
		printChar: $:.
	1	to: 11 - (self strlen: name) - (self log10: (idx max: 1)) floor
		do: [:i| self printChar: $ ].
	self print: name;
		printNum: idx;
		print: ': ';
		printHex: it;
		tab;
		printChar: $=;
		printOopShort: it;
		cr! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 11/13/2013 12:52'!
printFrameThing: name andFrame: theFP at: address
	<var: #theFP type: #'char *'>
	| it len |
	<inline: false>
	<var: #name type: #'char *'>
	<var: #address type: #'char *'>
	it := stackPages longAt: address.
	self printHexPtr: address;
		printChar: $:.
	len := self strlen: name.
	1 to: 12 - len do: [:i| self space].
	self print: name;
		print: ': ';
		printHex: it.
	it ~= 0 ifTrue:
		[self printChar: $=.
		 it = objectMemory nilObject
			ifTrue: [self print: 'nil']
			ifFalse:
				[self printNum: it]].
	self print: ' frame: '; printHexPtr: theFP; cr! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
printFrameThing: name at: address
	| it len |
	<inline: false>
	<var: #name type: #'char *'>
	<var: #address type: #'char *'>
	it := stackPages longAt: address.
	self printHexPtr: address;
		printChar: $:.
	len := self strlen: name.
	1 to: 12 - len do: [:i| self space].
	self print: name;
		print: ': ';
		printHex: it.
	it ~= 0 ifTrue:
		[self printChar: $=.
		 it = objectMemory nilObject
			ifTrue: [self print: 'nil']
			ifFalse:
				[self printNum: it]].
	self cr! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 3/3/2012 10:19'!
printFramesInPage: thePage
	<export: true> "use export: not api, so it won't be written to cointerp.h. cogit.c is unaware of StackPage"
	<inline: false>
	<var: #thePage type: #'StackPage *'>
	self printFrameAndCallers: thePage headFP SP: thePage headSP short: false! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 11/8/2013 14:57'!
printFramesOnStackPageListInUse
	<export: true>
	| page |
	<inline: false>
	<var: #page type: #'StackPage *'>
	page := stackPages mostRecentlyUsedPage.
	[(stackPages isFree: page) ifFalse:
		[self print: 'page '; printHexPtrnp: (self cCode: [page] inSmalltalk: [page baseAddress]); cr.
		 self printFramesInPage: page.
		 self cr].
	 (page := page prevPage) ~= stackPages mostRecentlyUsedPage] whileTrue! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem (auto pragmas 12/08) 7/1/2008 21:07'!
printHeadFrame
	<inline: false>
	self printFrame: localFP WithSP: localSP! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 5/2/2018 16:25'!
printHex: n
	"Print n in hex,  in the form '    0x1234', padded to a width of 10 characters
	 in 32-bits ('0x' + 8 nibbles) or 18 characters in 64-bits ('0x' + 16 nibbles)"
	<api>
	<var: #n type: #usqInt>
	| len buf |
	<var: #buf declareC: 'char buf[37]'> "large enough for a 64-bit value in hex plus the null plus 16 spaces"
	self cCode: 'memset(buf,'' '',36)' inSmalltalk: [buf := 'doh!!'].
	len := self cCode: 'sprintf(buf + 2 + 2 * BytesPerWord, "0x%" PRIxSQPTR, (usqIntptr_t)(n))'.
	self cCode: 'printf("%s", buf + len)'.
	len touch: buf! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 9/24/2013 14:42'!
printHexPtr: p
	"Print p in hex, padded to 10 characters in the form '    0x1234'"
	<inline: true>
	<var: #p type: #'void *'>
	self printHex: (self oopForPointer: p)! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 9/24/2013 14:43'!
printHexPtrnp: p
	"Print p in hex, unpadded, in the form '0x1234'"
	<inline: true>
	<var: #p type: #'void *'>
	self printHexnp: (self oopForPointer: p)! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'nice 12/25/2018 23:28'!
printHexnp: n
	<api>
	<var: #n type: #usqInt>
	"Print n in hex,  in the form '0x1234', unpadded"
	^'0x%lx' printf: n! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'nice 12/25/2018 23:28'!
printHexnpnp: n
	<var: #n type: #usqInt>
	"Print n in hex, in the form '1234', unpadded"
	^'%lx' printf: n! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:42'!
printLikelyImplementorsOfSelector: selector
	"Print all methods whose penultimate literal is either selector,
	 or an object whose first inst var is the method and whose
	 second is selector (e.g. an AdditionalMethodState)."
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	objectMemory allObjectsDo:
		[:obj| | methodClassAssociation |
		((objectMemory isCompiledMethod: obj)
		 and: [(self maybeSelectorOfMethod: obj) = selector]) ifTrue:
			["try and print the key of the method class association (the name of the implementing class)"
			 methodClassAssociation := self methodClassAssociationOf: obj.
			 self printHexnp: obj;
				space;
				printOopShortInner: (((objectMemory isPointers: methodClassAssociation)
									  and: [(objectMemory numSlotsOf: methodClassAssociation) >= 2])
										ifTrue: [objectMemory fetchPointer: 0 ofObject: methodClassAssociation]
										ifFalse: [methodClassAssociation]);
				cr]]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:43'!
printMethodCache
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	self printMethodCacheFor: -1! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:43'!
printMethodCacheFor: thing
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| n |
	n := 0.
	0 to: MethodCacheSize - 1 by: MethodCacheEntrySize do:
		[:i | | s c m p |
		s := methodCache at: i + MethodCacheSelector.
		c := methodCache at: i + MethodCacheClass.
		m := methodCache at: i + MethodCacheMethod.
		p := methodCache at: i + MethodCachePrimFunction.
		((thing = -1 or: [s = thing or: [c = thing or: [p = thing or: [m = thing]]]])
		 and: [(objectMemory addressCouldBeOop: s)
		 and: [c ~= 0
		 and: [(self addressCouldBeClassObj: c)
			or: [self addressCouldBeClassObj: (objectMemory classForClassTag: c)]]]]) ifTrue:
			[self cCode: [] inSmalltalk: [self transcript ensureCr].
			 self printNum: i; space; printHexnp: i; cr; tab.
			 (objectMemory isBytesNonImm: s)
				ifTrue: [self cCode: 'printf("%" PRIxSQPTR " %.*s\n", s, (int)(numBytesOf(s)), (char *)firstIndexableField(s))'
						inSmalltalk: [self printHex: s; space; print: (self stringOf: s); cr]]
				ifFalse: [self shortPrintOop: s].
			 self tab.
			 (self addressCouldBeClassObj: c)
				ifTrue: [self shortPrintOop: c]
				ifFalse: [self printNum: c; space; shortPrintOop: (objectMemory classForClassTag: c)].
			self tab; shortPrintOop: m; tab.
			self cCode:
					[p > 1024
						ifTrue: [self printHexnp: p]
						ifFalse: [self printNum: p]]
				inSmalltalk:
					[p isSymbol ifTrue: [self print: p] ifFalse: [self printNum: p]].
			self cr]].
	n > 1 ifTrue:
		[self printNum: n; cr]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:43'!
printMethodDictionary: dictionary
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| methodArray |
	methodArray := objectMemory fetchPointer: MethodArrayIndex ofObject: dictionary.
	SelectorStart to: (objectMemory numSlotsOf: dictionary) - 1 do:
		[:index | | selector meth |
		 selector := objectMemory fetchPointer: index ofObject: dictionary.
		 selector ~= objectMemory nilObject ifTrue:
			[meth := objectMemory fetchPointer: index - SelectorStart ofObject: methodArray.
			 self
				printOopShort: selector;
				print: ' => ';
				printOopShort: meth;
				print: ' (';
				printHex: selector;
				print: ' => ';
				printHex: meth;
				putchar: $);
				cr]]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:43'!
printMethodDictionaryOf: behavior
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	self printMethodDictionary: (objectMemory fetchPointer: MethodDictionaryIndex ofObject: behavior)! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 8/15/2016 11:13'!
printMethodFieldForPrintContext: aContext
	<inline: true>
	self shortPrintOop: (objectMemory fetchPointer: MethodIndex ofObject: aContext)! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 1/20/2017 15:57'!
printMethodHeaderOop: anOop
	"Override hook for CoInterpreter"
	<inline: true>
	^self printDecodeMethodHeaderOop: anOop! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 10/30/2013 00:27'!
printNameOfClass: classOop count: cnt
	"Details: The count argument is used to avoid a possible infinite recursion if classOop is a corrupted object."
	<inline: false>
	| numSlots |
	classNameIndex ifNil:
		[self print: '??nil cnidx??'.
		 ^self].
	(classOop isNil or: [classOop = 0 or: [cnt <= 0]]) ifTrue:
		[self print: 'bad class'.
		 ^self].
	numSlots := objectMemory numSlotsOf: classOop.
	(numSlots = metaclassNumSlots
	 and: [metaclassNumSlots > thisClassIndex])
		ifTrue: [self printNameOfClass: (objectMemory fetchPointer: thisClassIndex ofObject: classOop) count: cnt - 1.
				self print: ' class']
		ifFalse:
			[numSlots <= classNameIndex
				ifTrue: [self print: 'bad class']
				ifFalse:
					[self printStringOf: (objectMemory fetchPointer: classNameIndex ofObject: classOop)]]! !

!StackInterpreter methodsFor: 'debug printing' stamp: ''!
printNum: n
	"For testing in Smalltalk, this method should be overridden in a subclass."

	self cCode: 'printf("%ld", (long) n)'.! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:30'!
printOop: oop
	<export: true> "use export: not api, so it won't be written to cointerp.h"
	| cls fmt lastIndex startIP bytecodesPerLine column |
	<inline: false>
	(objectMemory isImmediate: oop) ifTrue:
		[^self shortPrintOop: oop].
	self printHex: oop.
	(objectMemory addressCouldBeObj: oop) ifFalse:
		[(oop bitAnd: objectMemory allocationUnit - 1) ~= 0 ifTrue: [^self print: ' is misaligned'; cr].
		 ((objectMemory isInNewSpace: oop)
		  and: [objectMemory isForwarded: oop]) ifTrue:
			[self printForwarder: oop].
		 ^self print: (self whereIs: oop); cr].
	(objectMemory isFreeObject: oop) ifTrue:
		[self print: ' is a free chunk of size '; printNum: (objectMemory sizeOfFree: oop).
		 objectMemory hasSpurMemoryManagerAPI ifTrue:
			[self print: ' 0th: '; printHex: (objectMemory fetchPointer: 0 ofFreeChunk: oop).
			 objectMemory printHeaderTypeOf: oop].
		 ^self cr].
	(objectMemory isForwarded: oop) ifTrue:
		[^self printForwarder: oop].
	self print: ': a(n) '.
	self printNameOfClass: (cls := objectMemory fetchClassOfNonImm: oop) count: 5.
	cls = (objectMemory splObj: ClassFloat) ifTrue:
		[^self cr; printFloat: (objectMemory dbgFloatValueOf: oop); cr].
	fmt := objectMemory formatOf: oop.
	fmt > objectMemory lastPointerFormat ifTrue:
		[self print: ' nbytes '; printNum: (objectMemory numBytesOf: oop)].
	self cr.
	(fmt between: objectMemory firstLongFormat and: objectMemory firstCompiledMethodFormat - 1) ifTrue:
		["This will answer false if splObj: ClassAlien is nilObject"
		 (self is: oop KindOfClass: (objectMemory splObj: ClassAlien)) ifTrue:
			[self print: ' datasize '; printNum: (objectMemory sizeFieldOfAlien: oop).
			self print: ((self isIndirectAlien: oop)
							ifTrue: [' indirect @ ']
							ifFalse:
								[(self isPointerAlien: oop)
									ifTrue: [' pointer @ ']
									ifFalse: [' direct @ ']]).
			 ^self printHex: (self startOfAlienData: oop) asUnsignedInteger; cr].
		 (objectMemory isWordsNonImm: oop) ifTrue:
			[lastIndex := 64 min: ((objectMemory numBytesOf: oop) / objectMemory wordSize).
			 lastIndex > 0 ifTrue:
				[1 to: lastIndex do:
					[:index|
					self space; printHex: (self cCoerceSimple: (objectMemory fetchLong32: index - 1 ofObject: oop)
											to: #'unsigned int').
					(index \\ self elementsPerPrintOopLine) = 0 ifTrue:
						[self cr]].
				(lastIndex \\ self elementsPerPrintOopLine) = 0 ifFalse:
					[self cr]].
			^self].
		^self printStringOf: oop; cr].
	"this is nonsense.  apologies."
	startIP := (objectMemory lastPointerOf: oop) + objectMemory bytesPerOop - objectMemory baseHeaderSize / objectMemory bytesPerOop.
	lastIndex := 256 min: startIP.
	lastIndex > 0 ifTrue:
		[1 to: lastIndex do:
			[:index|
			self cCode: [self printHex: (objectMemory fetchPointer: index - 1 ofObject: oop); space]
				inSmalltalk: [self space; printHex: (objectMemory fetchPointer: index - 1 ofObject: oop); space.
							 self print: (self shortPrint: (objectMemory fetchPointer: index - 1 ofObject: oop))].
			(index \\ self elementsPerPrintOopLine) = 0 ifTrue:
				[self cr]].
		(lastIndex \\ self elementsPerPrintOopLine) = 0 ifFalse:
			[self cr]].
	(objectMemory isCompiledMethod: oop)
		ifFalse:
			[startIP > 64 ifTrue: [self print: '...'; cr]]
		ifTrue:
			[startIP := startIP * objectMemory wordSize + 1.
			 lastIndex := objectMemory lengthOf: oop.
			 lastIndex - startIP > 100 ifTrue:
				[lastIndex := startIP + 100].
			 bytecodesPerLine := 8.
			 column := 1.
			 startIP to: lastIndex do:
				[:index| | byte |
				column = 1 ifTrue:
					[self cCode: 'printf("0x%08" PRIxSQPTR ": ", (usqIntptr_t)(oop+BaseHeaderSize+index-1))'
						inSmalltalk: [self print: (oop+objectMemory baseHeaderSize+index-1) hex; print: ': ']].
				byte := objectMemory fetchByte: index - 1 ofObject: oop.
				self cCode: 'printf(" %02x/%-3d", (int)byte,(int)byte)'
					inSmalltalk: [self space; print: (byte radix: 16); printChar: $/; printNum: byte].
				column := column + 1.
				column > bytecodesPerLine ifTrue:
					[column := 1. self cr]].
			column = 1 ifFalse:
				[self cr]]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 11/4/2010 17:05'!
printOopShort: oop
	<inline: false>
	self printOopShortInner: oop.
	self flush! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 09/19/2018 17:19'!
printOopShortInner: oop
	| classOop name nameLen |
	<var: #name type: #'char *'>
	<inline: true>
	(objectMemory isImmediate: oop) ifTrue:
		[(objectMemory isImmediateCharacter: oop) ifTrue:
			[^self
				printChar: $$;
				printChar: (objectMemory characterValueOf: oop);
				printChar: $(;
				printHexnp: (objectMemory characterValueOf: oop);
				printChar: $)].
		 (objectMemory isIntegerObject: oop) ifTrue:
			[^self
				printNum: (objectMemory integerValueOf: oop);
				printChar: $(;
				printHexnp: (objectMemory integerValueOf: oop);
				printChar: $)].
		 (objectMemory isImmediateFloat: oop) ifTrue:
			[^self
				printFloat: (objectMemory dbgFloatValueOf: oop);
				printChar: $(;
				printHexnp: oop;
				printChar: $)].
		 ^self print: 'unknown immediate '; printHexnp: oop].
	(objectMemory addressCouldBeObj: oop) ifFalse:
		[^self print: ((oop bitAnd: objectMemory allocationUnit - 1) ~= 0
						ifTrue: [' is misaligned']
						ifFalse: [self whereIs: oop])].
	(objectMemory isFreeObject: oop) ifTrue:
		[^self print: ' is a free chunk'].
	(objectMemory isForwarded: oop) ifTrue:
		[^self print: ' is a forwarder to '; printHexnp: (objectMemory followForwarded: oop)].
	(self isFloatObject: oop) ifTrue:
		[^self printFloat: (objectMemory dbgFloatValueOf: oop)].
	classOop := objectMemory fetchClassOfNonImm: oop.
	(objectMemory addressCouldBeObj: classOop) ifFalse:
		[^self print: 'a ??'].
	(objectMemory numSlotsOf: classOop) = metaclassNumSlots ifTrue:
		[^self printNameOfClass: oop count: 5].
	oop = objectMemory nilObject ifTrue: [^self print: 'nil'].
	oop = objectMemory trueObject ifTrue: [^self print: 'true'].
	oop = objectMemory falseObject ifTrue: [^self print: 'false'].
	nameLen := self lengthOfNameOfClass: classOop.
	nameLen = 0 ifTrue: [^self print: 'a ??'].
	name := self nameOfClass: classOop.
	nameLen = 10 ifTrue:
		[(self strncmp: name _: 'ByteString' _: 10) = 0 "strncmp is weird" ifTrue:
			[^self printChar: $'; printStringOf: oop; printChar: $'].
		 (self strncmp: name _: 'ByteSymbol' _: 10) = 0 "strncmp is weird" ifTrue:
			[self printChar: $#; printStringOf: oop. ^self]].
	(nameLen = 9 and: [(self strncmp: name _: 'Character' _: 9) = 0]) ifTrue:
		[^self printChar: $$; printChar: (objectMemory integerValueOf: (objectMemory fetchPointer: 0 ofObject: oop))].
	self print: 'a(n) '.
	self
		cCode: [0 to: nameLen - 1 do: [:i| self printChar: (name at: i)]]
		inSmalltalk:
			[name isString
				ifTrue: [self print: name]
				ifFalse: [0 to: nameLen - 1 do: [:i| self printChar: (name at: i)]]].
	"Try to spot association-like things; they're all subclasses of LookupKey"
	((objectMemory isPointersNonImm: oop)
	 and: [(objectMemory instanceSizeOf: classOop) = (ValueIndex + 1)
	 and: [(objectMemory isBytes: (objectMemory fetchPointer: KeyIndex ofObject: oop))]]) ifTrue:
		[| classLookupKey |
		 classLookupKey := objectMemory fetchClassOfNonImm: (objectMemory splObj: SchedulerAssociation).
		 [classLookupKey = objectMemory nilObject ifTrue:
			[^self].
		  (objectMemory instanceSizeOf: classLookupKey) = (KeyIndex + 1)] whileFalse:
			[classLookupKey := self superclassOf: classLookupKey].
		 (self includesBehavior: classOop ThatOf: classLookupKey) ifTrue:
			[self space;
				printOopShort: (objectMemory fetchPointer: KeyIndex ofObject: oop);
				print: ' -> ';
				printHexnp: (objectMemory fetchPointer: ValueIndex ofObject: oop)]]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 1/14/2009 17:57'!
printPageHeadFrame
	<inline: false>
	self printFrame: stackPage headFP WithSP: stackPage headSP! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:43'!
printProcessStack: aProcess
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	<inline: false>
	| ctx |
	self cr; printNameOfClass: (objectMemory fetchClassOf: aProcess) count: 5; space; printHex: aProcess.
	self print: ' priority '; printNum: (self quickFetchInteger: PriorityIndex ofObject: aProcess); cr.
	ctx := objectMemory followField: SuspendedContextIndex ofObject: aProcess.
	ctx = objectMemory nilObject ifFalse:
		[self printCallStackOf: ctx currentFP: framePointer]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:43'!
printProcsOnList: procList
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	<inline: false>
	| proc firstProc |
	proc := firstProc := objectMemory followField: FirstLinkIndex ofObject: procList.
	[proc = objectMemory nilObject] whileFalse:
		[self printProcessStack: proc.
		 proc := objectMemory followField: NextLinkIndex ofObject: proc.
		 proc = firstProc ifTrue:
			[self warning: 'circular process list!!!!'.
			 ^nil]]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 10/11/2011 15:25'!
printSends
	^false! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 3/18/2015 18:51'!
printStackCallStack
	<doNotGenerate>
	self printStackCallStackOf: (localFP ifNil: [framePointer])! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:48'!
printStackCallStackOf: aContextOrProcessOrFrame
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| theFP context |
	<var: #theFP type: #'char *'>
	(self cCode: [false] "In the stack simulator, frame pointers are negative which upsets addressCouldBeObj:"
		inSmalltalk: [stackPages couldBeFramePointer: aContextOrProcessOrFrame]) ifFalse:
		[(objectMemory addressCouldBeObj: aContextOrProcessOrFrame) ifTrue:
			[((objectMemory isContext: aContextOrProcessOrFrame)
			  and: [self checkIsStillMarriedContext: aContextOrProcessOrFrame currentFP: nil]) ifTrue:
				[^self printStackCallStackOf: (self frameOfMarriedContext: aContextOrProcessOrFrame) asInteger].
			 aContextOrProcessOrFrame = self activeProcess ifTrue:
				[^self printStackCallStackOf: (self cCode: [framePointer asInteger] inSmalltalk: [self headFramePointer])].
			 (self couldBeProcess: aContextOrProcessOrFrame) ifTrue:
				[^self printCallStackOf: (objectMemory
											fetchPointer: SuspendedContextIndex
											ofObject: aContextOrProcessOrFrame)].
			 ^nil]].

	theFP := aContextOrProcessOrFrame asVoidPointer.
	[context := self shortReversePrintFrameAndCallers: theFP.
	 ((self isMarriedOrWidowedContext: context)
	  and:
		[theFP := self frameOfMarriedContext: context.
		 self checkIsStillMarriedContext: context currentFP: theFP]) ifFalse:
			[^nil]] repeat! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 8/11/2015 11:58'!
printStackPage: page
	<inline: true>
	<var: #page type: #'StackPage *'>
	self printStackPage: page useCount: -1! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 8/3/2015 16:03'!
printStackPage: page useCount: n
	<inline: false>
	<var: #page type: #'StackPage *'>
	self print: 'page '; printHexPtr: (self cCode: [page] inSmalltalk: [page baseAddress]);
		print: ' ('; printNum: (stackPages pageIndexFor: page realStackLimit).
	n >= 0 ifTrue:
		[self print: ','; printNum: n].
	self print: ')  (trace: '; printNum: page trace; printChar: $).
	(stackPages isFree: page) ifTrue:
		[self print: ' (free)'].
	page = stackPages mostRecentlyUsedPage ifTrue:
		[self print: ' (MRU)'].
	page prevPage = stackPages mostRecentlyUsedPage ifTrue:
		[self print: ' (LRU)'].
	self cr; tab; print: 'ba: ';
		printHexPtr: page baseAddress; print: ' - sl: ';
		printHexPtr: page realStackLimit; print: ' - sl-so: ';
		printHexPtr: page realStackLimit - self stackLimitOffset; print: ' - la:';
		printHexPtr: page lastAddress.
	(stackPages isFree: page) ifFalse:
		[self cr; tab; print: 'baseFP '; printHexPtr: page baseFP.
		 self "cr;" tab; print: 'headFP '; printHexPtr: page headFP.
		 self "cr;" tab; print: 'headSP '; printHexPtr: page headSP].
	self cr; tab; print: 'prev '; printHexPtr: (self cCode: 'page->prevPage' inSmalltalk: [page prevPage baseAddress]);
		print: ' ('; printNum: (stackPages pageIndexFor: page prevPage realStackLimit); printChar: $).
	self tab; print: 'next '; printHexPtr: (self cCode: 'page->nextPage' inSmalltalk: [page nextPage baseAddress]);
		print: ' ('; printNum: (stackPages pageIndexFor: page nextPage realStackLimit); printChar: $).
	self cr! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:49'!
printStackPageList
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| page |
	<inline: false>
	<var: #page type: #'StackPage *'>
	page := stackPages mostRecentlyUsedPage.
	[self printStackPage: page.
	 self cr.
	 (page := page prevPage) ~= stackPages mostRecentlyUsedPage] whileTrue! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:49'!
printStackPageListInUse
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| page n |
	<inline: false>
	<var: #page type: #'StackPage *'>
	page := stackPages mostRecentlyUsedPage.
	n := 0.
	[(stackPages isFree: page) ifFalse:
		[self printStackPage: page useCount: (n := n + 1); cr].
	 (page := page prevPage) ~= stackPages mostRecentlyUsedPage] whileTrue! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:48'!
printStackPages
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	0 to: numStackPages - 1 do:
		[:i|
		self printStackPage: (stackPages stackPageAt: i).
		self cr]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:49'!
printStackPagesInUse
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| n |
	n := 0.
	0 to: numStackPages - 1 do:
		[:i|
		(stackPages isFree: (stackPages stackPageAt: i)) ifFalse:
			[self printStackPage: (stackPages stackPageAt: i) useCount: (n := n + 1); cr]]! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 4/13/2021 17:46'!
printStackReferencesTo: oop
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	0 to: numStackPages - 1 do:
		[:i| | thePage theSP theFP callerFP |
		thePage := stackPages stackPageAt: i.
		thePage isFree ifFalse:
			[theSP := thePage headSP.
			 theFP := thePage headFP.
			 "Skip the instruction pointer on top of stack of inactive pages."
			 thePage = stackPage ifFalse:
				[theSP := theSP + objectMemory wordSize].
			 [[theSP <= (self frameReceiverLocation: theFP)] whileTrue:
				[oop = (stackPages longAt: theSP) ifTrue:
					[self print: 'FP: '; printHexnp: theFP; print: ' @ '; printHexnp: theSP; cr].
				 theSP := theSP + objectMemory wordSize].
			  (self frameHasContext: theFP) ifTrue:
				[oop = (self frameContext: theFP) ifTrue:
					[self print: 'FP: '; printHexnp: theFP; print: ' CTXT'; cr]].
			  oop = (self frameMethod: theFP) ifTrue:
				[self print: 'FP: '; printHexnp: theFP; print: ' MTHD'; cr].
			  (callerFP := self frameCallerFP: theFP) ~= 0]
				whileTrue:
					[theSP := (theFP + FoxCallerSavedIP) + objectMemory wordSize.
					 theFP := callerFP].
			 theSP := theFP + FoxCallerSavedIP. "a.k.a. FoxCallerContext"
			 [theSP <= thePage baseAddress] whileTrue:
				[oop = (stackPages longAt: theSP) ifTrue:
					[self print: 'FP: '; printHexnp: theFP; print: ' @ '; printHexnp: theSP; cr].
				 theSP := theSP + objectMemory wordSize]]]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 8/5/2021 15:52'!
printStringOf: oop
	| fmt len cnt max i |
	<inline: false>
	(objectMemory isImmediate: oop) ifTrue:
		[^self].
	(objectMemory addressCouldBeObj: oop) ifFalse:
		[^self].
	fmt := objectMemory formatOf: oop.
	fmt < objectMemory firstByteFormat ifTrue: [^self].

	cnt := (max := 128) min: (len := objectMemory lengthOf: oop).
	i := 0.

	((objectMemory is: oop
		  instanceOf: (objectMemory splObj: ClassByteArray)
		  compactClassIndex: classByteArrayCompactIndex)
	or: [(objectMemory isLargeIntegerInstance: oop)])
		ifTrue:
			[[i < cnt] whileTrue:
				[self printHex: (objectMemory fetchByte: i ofObject: oop).
				 i := i + 1]]
		ifFalse:
			[[i < cnt] whileTrue:
				[| code |
				code := objectMemory fetchByte: i ofObject: oop.
				code
					caseOf: {
						[10 "Character lf asInteger"] -> [self print: '<LF>'].
						[13 "Character cr asInteger"] -> [self print: '<CR>'] }
					otherwise: [self printChar: code].
				 i := i + 1]].
	len > max ifTrue:
		[self print: '...'].
	self flush! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem (auto pragmas 12/08) 11/25/2008 12:50'!
printUnbalancedStack: primIdx
	<inline: false>
	self print: 'Stack unbalanced after '.
	self successful 
		ifTrue:[self print:'successful primitive '] 
		ifFalse:[self print: 'failed primitive '].
	self printNum: primIdx.
	self cr.
		! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
push: object
	"In the StackInterpreter stacks grow down."
	| sp |
	<inline: true>
	<var: #sp type: #'char *'>
	stackPages longAt: (sp := stackPointer - objectMemory wordSize) put: object.
	stackPointer := sp! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 10/8/2009 23:24'!
pushActiveContextBytecode
	| ourContext |
	ourContext := self ensureFrameIsMarried: localFP SP: localSP.
	self fetchNextBytecode.
	self internalPush: ourContext! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 10/26/2012 17:03'!
pushBool: trueOrFalse
	<inline: true>
	self push: (objectMemory booleanObjectOf: trueOrFalse)! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 10/30/2012 17:24'!
pushClosureCopyCopiedValuesBytecode
	"The compiler has pushed the values to be copied, if any.  Find numArgs and numCopied in the byte following.
	 Create a Closure with space for the copiedValues and pop numCopied values off the stack into the closure.
	 Set numArgs as specified, and set startpc to the pc following the block size and jump over that code."
	| numArgsNumCopied numArgs numCopied blockSize |
	numArgsNumCopied := self fetchByte.
	numArgs := numArgsNumCopied bitAnd: 16rF.
	numCopied := numArgsNumCopied bitShift: -4.
	"Split blockSize := (self fetchByte * 256) + self fetchByte. into two because evaluation order in C is undefined."
	blockSize := self fetchByte << 8.
	blockSize := blockSize + self fetchByte.
	self pushClosureNumArgs: numArgs copiedValues: numCopied blockSize: blockSize! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
pushClosureNumArgs: numArgs copiedValues: numCopied blockSize: blockSize
	"The compiler has pushed the values to be copied, if any.
	 Create a Closure with space for the copiedValues and pop numCopied values off the stack into the closure.
	 Set numArgs as specified, and set startpc to the pc following the block size and jump over that code."
	<inline: true>
	| newClosure context |
	"No need to record the pushed copied values in the outerContext."
	context := self ensureFrameIsMarried: localFP SP: localSP + (numCopied * objectMemory bytesPerOop).
	newClosure := self
					closureIn: context
					numArgs: numArgs
					instructionPointer: (self oopForPointer: localIP) + 2 - (method+objectMemory baseHeaderSize)
					numCopiedValues: numCopied.
	numCopied > 0 ifTrue:
		[0 to: numCopied - 1 do:
			[:i|
			"Assume: have just allocated a new BlockClosure; it must be young.
			 Thus, can use unchecked stores."
			 objectMemory storePointerUnchecked: i + ClosureFirstCopiedValueIndex
				ofObject: newClosure
				withValue: (self internalStackValue: numCopied - i - 1)].
		 self internalPop: numCopied].
	localIP := localIP + blockSize.
	self fetchNextBytecode.
	self internalPush: newClosure! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
pushConstantFalseBytecode

	self fetchNextBytecode.
	self internalPush: objectMemory falseObject.
! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:48'!
pushConstantMinusOneBytecode

	self fetchNextBytecode.
	self internalPush: ConstMinusOne.
! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
pushConstantNilBytecode

	self fetchNextBytecode.
	self internalPush: objectMemory nilObject.
! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:48'!
pushConstantOneBytecode

	self fetchNextBytecode.
	self internalPush: ConstOne.
! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
pushConstantTrueBytecode

	self fetchNextBytecode.
	self internalPush: objectMemory trueObject.
! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:48'!
pushConstantTwoBytecode

	self fetchNextBytecode.
	self internalPush: ConstTwo.
! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:48'!
pushConstantZeroBytecode

	self fetchNextBytecode.
	self internalPush: ConstZero.
! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 10/7/2014 11:12'!
pushEnclosingObjectBytecode
	"Find the enclosing object at level N"
	| litIndex  n anIntOop |
	<inline: true>
	litIndex := self fetchByte.
	anIntOop := self literal: litIndex.
	n := (objectMemory isIntegerObject: anIntOop)
			ifTrue: [objectMemory integerValueOf: anIntOop]
			ifFalse: [0].
	self fetchNextBytecode.
	self internalPush:(self 
						enclosingObjectAt: n 
						withObject: self receiver 
						withMixin: (self methodClassOf: method))! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 11/21/2014 16:20'!
pushFloat: f

	<var: #f type: #double>
	self push: (objectMemory floatObjectOf: f).! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 4/8/2016 09:25'!
pushFullClosureNumArgs: numArgs copiedValues: numCopiedArg compiledBlock: compiledBlock receiverIsOnStack: receiverIsOnStack ignoreContext: ignoreContext
	"The compiler has pushed the values to be copied, if any. The receiver has been pushed on stack before if specified. 
	 Create a Closure with space for the copiedValues and pop numCopied values off the stack into the closure.
	 Sets outerContext, compiledBlock, numArgs and receiver as specified.."
	<inline: true>
	| numCopied newClosure context startIndex |
	"No need to record the pushed copied values in the outerContext."
	context := ignoreContext
		ifTrue: [objectMemory nilObject ]
		ifFalse: [self ensureFrameIsMarried: localFP SP: localSP + (numCopiedArg * objectMemory bytesPerOop)].
	newClosure := self
					fullClosureIn: context 
					numArgs: numArgs 
					numCopiedValues: numCopiedArg 
					compiledBlock: compiledBlock.
	receiverIsOnStack
		ifFalse: 
			[ startIndex := FullClosureFirstCopiedValueIndex.
			   objectMemory storePointerUnchecked: FullClosureReceiverIndex
				ofObject: newClosure
				withValue: self receiver.
			numCopied := numCopiedArg ]
		ifTrue:
			[ startIndex := FullClosureReceiverIndex.
			numCopied := numCopiedArg + 1 ].
	numCopied > 0 ifTrue:
		[0 to: numCopied - 1 do:
			[ :i |
			"Assume: have just allocated a new BlockClosure; it must be young.
			 Thus, can use unchecked stores."
			 objectMemory storePointerUnchecked: i + startIndex
				ofObject: newClosure
				withValue: (self internalStackValue: numCopied - i - 1)].
		 self internalPop: numCopied].
	self fetchNextBytecode.
	self internalPush: newClosure! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
pushInteger: integerValue
	self push: (objectMemory integerObjectOf: integerValue).
	^nil! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: ''!
pushLiteralConstant: literalIndex

	self internalPush: (self literal: literalIndex).! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 11/13/2012 10:40'!
pushLiteralConstantBytecode
	<expandCases>
	self
		cCode: "this bytecode will be expanded so that refs to currentBytecode below will be constant"
			[self fetchNextBytecode.
			 self pushLiteralConstant: (currentBytecode bitAnd: 16r1F)]
		inSmalltalk: "Interpreter version has fetchNextBytecode out of order"
			[self pushLiteralConstant: (currentBytecode bitAnd: 16r1F).
			 self fetchNextBytecode]! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 11/13/2012 10:34'!
pushLiteralVariable16CasesBytecode
	<expandCases>
	self
		cCode: "this bytecode will be expanded so that refs to currentBytecode below will be constant"
			[self fetchNextBytecode.
			 self pushLiteralVariable: (currentBytecode bitAnd: 16rF)]
		inSmalltalk: "Interpreter version has fetchNextBytecode out of order"
			[self pushLiteralVariable: (currentBytecode bitAnd: 16rF).
			 self fetchNextBytecode]! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 5/25/2016 11:24'!
pushLiteralVariable: literalIndex
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[| litVar |
			 "push/store/popLiteralVariable all fetch a literal, and either read or write the literal's value field.
			  The fetch of the literal needs an explicit check (otherwise we would have to scan all literals in
			  all methods in the stack zone, and the entire method on return, and global variables are relatively
			  rare; in my work image 8.7% of literals are globals)."
			 litVar := self literal: literalIndex.
			 (objectMemory isForwarded: litVar) ifTrue:
				[litVar := self unfollow: litVar atIndex: literalIndex].
			 self internalPush:
				(objectMemory fetchPointer: ValueIndex ofObject: litVar)]
		ifFalse:
			[self internalPush:
				(objectMemory fetchPointer: ValueIndex ofObject: (self literal: literalIndex))]! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 11/13/2012 10:43'!
pushLiteralVariableBytecode
	<expandCases>
	self
		cCode: "this bytecode will be expanded so that refs to currentBytecode below will be constant"
			[self fetchNextBytecode.
			 self pushLiteralVariable: (currentBytecode bitAnd: 16r1F)]
		inSmalltalk: "Interpreter version has fetchNextBytecode out of order"
			[self pushLiteralVariable: (currentBytecode bitAnd: 16r1F).
			 self fetchNextBytecode]! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 4/5/2016 17:26'!
pushMaybeContext: obj receiverVariable: fieldIndex
	"Must trap accesses to married and widowed contexts.
	 But don't want to check on all inst var accesses.  This
	 method is only used by the long-form bytecodes, evading
	 the cost. Note that the method, closure and receiver fields
	 of married contexts are correctly initialized so they don't
	 need special treatment on read.  Only sender, instruction
	 pointer and stack pointer need to be intercepted on reads."
	<inline: true>
	((self isReadMediatedContextInstVarIndex: fieldIndex)
	and: [objectMemory isContextNonImm: obj])
		ifTrue:
			[self internalPush: (self instVar: fieldIndex ofContext: obj)]
		ifFalse:
			[self internalPush: (objectMemory fetchPointer: fieldIndex ofObject: obj)]! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 4/5/2016 17:26'!
pushMaybeContextReceiverVariable: fieldIndex
	"Must trap accesses to married and widowed contexts.
	 But don't want to check on all inst var accesses.  This
	 method is only used by the long-form bytecodes, evading
	 the cost. Note that the method, closure and receiver fields
	 of married contexts are correctly initialized so they don't
	 need special treatment on read.  Only sender, instruction
	 pointer and stack pointer need to be intercepted on reads."
	<inline: true>
	self pushMaybeContext: self receiver receiverVariable: fieldIndex! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 3/4/2014 17:55'!
pushNewArrayBytecode
	| size popValues array |
	size := self fetchByte.
	popValues := size > 127.
	size := size bitAnd: 127.
	self fetchNextBytecode.
	self externalizeIPandSP. "in case of abort"
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[array := objectMemory
						eeInstantiateSmallClassIndex: ClassArrayCompactIndex
						format: objectMemory arrayFormat
						numSlots: size]
		ifFalse:
			[array := objectMemory
						eeInstantiateClassIndex: ClassArrayCompactIndex
						format: objectMemory arrayFormat
						numSlots: size].
	popValues
		ifTrue:
			[0 to: size - 1 do:
				[:i|
				"Assume: have just allocated a new Array; it must be young. Thus, can use unchecked stores."
				objectMemory storePointerUnchecked: i ofObject: array withValue: (self internalStackValue: size - i - 1)].
			 self internalPop: size]
		ifFalse:
			[0 to: size - 1 do:
				[:i|
				objectMemory storePointerUnchecked: i ofObject: array withValue: objectMemory nilObject]].
	self internalPush: array! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 6/17/2008 10:22'!
pushReceiverBytecode

	self fetchNextBytecode.
	self internalPush: self receiver.
! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
pushReceiverVariable: fieldIndex

	self internalPush: (objectMemory fetchPointer: fieldIndex ofObject: self receiver).! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 11/13/2012 10:44'!
pushReceiverVariableBytecode
	<expandCases>
	self
		cCode: "this bytecode will be expanded so that refs to currentBytecode below will be constant"
			[self fetchNextBytecode.
			 self pushReceiverVariable: (currentBytecode bitAnd: 16rF)]
		inSmalltalk: "Interpreter version has fetchNextBytecode out of order"
			[self pushReceiverVariable: (currentBytecode bitAnd: 16rF).
			 self fetchNextBytecode]! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 5/30/2018 13:41'!
pushRemoteTemp: index inVectorAt: tempVectorIndex
	| tempVector |
	tempVector := self temporary: tempVectorIndex in: localFP.
	TempVectReadBarrier
		ifTrue: 
			[(objectMemory isForwarded: tempVector) ifTrue:
				[tempVector := self unfollowTempVector: tempVector atIndex: tempVectorIndex in: localFP]].
	self internalPush: (objectMemory fetchPointer: index ofObject: tempVector)! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 5/27/2008 13:28'!
pushRemoteTempLongBytecode
	| remoteTempIndex tempVectorIndex |
	remoteTempIndex := self fetchByte.
	tempVectorIndex := self fetchByte.
	self fetchNextBytecode.
	self pushRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 6/23/2008 20:06'!
pushTemporaryVariable: temporaryIndex

	self internalPush: (self temporary: temporaryIndex in: localFP).! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 11/13/2012 10:46'!
pushTemporaryVariableBytecode
	<expandCases>
	self
		cCode: "this bytecode will be expanded so that refs to currentBytecode below will be constant"
			[self fetchNextBytecode.
			 self pushTemporaryVariable: (currentBytecode bitAnd: 16rF)]
		inSmalltalk: "Interpreter version has fetchNextBytecode out of order"
			[self pushTemporaryVariable: (currentBytecode bitAnd: 16rF).
			 self fetchNextBytecode]! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (auto pragmas 12/08) 7/10/2008 10:20'!
pushedReceiverOrClosureOfFrame: theFP
	"The receiver of a message send or the closure of a block activation is
	 always on the stack above any arguments and the frame itself.  See the
	 diagram in StackInterpreter class>>initializeFrameIndices."
	<inline: true>
	<var: #theFP type: #'char *'>
	^stackPages longAt: theFP + (self frameStackedReceiverOffset: theFP)! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'nice 7/13/2016 18:32'!
putLong: aLong toFile: aFile
	"Append aLong to aFile in this platform's 'natural' byte order.  aLong is either 32 or 64 bits,
	 depending on ObjectMemory.  (Bytes will be swapped, if necessary, when the image is read
	 on a different platform.) Set successFlag to false if the write fails."

	<var: #aLong type: #sqInt>
	<var: #aFile type: #sqImageFile>
	<inline: false>
	| objectsWritten |
	objectsWritten := self
						cCode: [self sq: (self addressOf: aLong) Image: (self sizeof: #sqInt) File: 1 Write: aFile]
						inSmalltalk:
							[| value |
							 value := aLong.
							 objectMemory wordSize timesRepeat:
								[aFile nextPut: (value bitAnd: 16rFF).
								 value := value >> 8].
							 1].
	self success: objectsWritten = 1! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 11/18/2014 14:27'!
putShort: aShort toFile: aFile
	"Append the 16-bit aShort to aFile in this platform's 'natural' byte order.
	 (Bytes will be swapped, if necessary, when the image is read on a
	 different platform.) Set successFlag to false if the write fails."

	<var: #aShort type: #short>
	<var: #aFile type: #sqImageFile>
	<inline: false>
	| objectsWritten |
	objectsWritten := self
						cCode: [self sq: (self addressOf: aShort) Image: (self sizeof: #short) File: 1 Write: aFile]
						inSmalltalk:
							[aFile
								nextPut: (aShort bitAnd: 16rFF);
								nextPut: (aShort >> 8 bitAnd: 16rFF).
							 1].
	self success: objectsWritten = 1! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 11/13/2014 13:17'!
putToSleep: aProcess yieldingIf: yieldImplicitly
	"Save the given process on the scheduler process list for its priority,
	 adding to the back if yieldImplicitly or to the front if not yieldImplicitly."

	| priority processLists processList |
	self assert: (framePointer - stackPointer) < (LargeContextSlots * objectMemory bytesPerOop).
	priority := self quickFetchInteger: PriorityIndex ofObject: aProcess.
	(highestRunnableProcessPriority ~= 0
	 and: [priority > highestRunnableProcessPriority]) ifTrue:
		[highestRunnableProcessPriority := priority].
	processLists := objectMemory fetchPointer: ProcessListsIndex ofObject: self schedulerPointer.
	processList := objectMemory fetchPointer: priority - 1 ofObject: processLists.
	yieldImplicitly
		ifTrue: [self addLastLink: aProcess toList: processList]
		ifFalse: [self addFirstLink: aProcess toList: processList]! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 11/18/2014 14:27'!
putWord32: aWord32 toFile: aFile
	"Append aWord32 to aFile in this platform's 'natural' byte order.  aWord32 is 32 bits,
	 depending on ObjectMemory.  (Bytes will be swapped, if necessary, when the image is read
	 on a different platform.) Set successFlag to false if the write fails."

	<var: #aWord32 type: #int>
	<var: #aFile type: #sqImageFile>
	<inline: false>
	| objectsWritten |
	objectsWritten := self
						cCode: [self sq: (self addressOf: aWord32) Image: 4 File: 1 Write: aFile]
						inSmalltalk:
							[| value |
							 value := aWord32.
							 4 timesRepeat:
								[aFile nextPut: (value bitAnd: 16rFF).
								 value := value >> 8].
							 1].
	self success: objectsWritten = 1! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 10/24/2017 15:53'!
quaternaryInlinePrimitive: primIndex
	<option: #SistaVM>
	self unknownInlinePrimitive! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
quickFetchInteger: fieldIndex ofObject: objectPointer
	"Return the integer value of the field without verifying that it is an integer value!! For use in time-critical places where the integer-ness of the field can be guaranteed."

	| oop |
	oop := objectMemory fetchPointer: fieldIndex ofObject: objectPointer.
	self assert: (objectMemory isIntegerObject: oop).
	^objectMemory integerValueOf: oop! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 10/24/2017 15:53'!
quinaryInlinePrimitive: primIndex
	| src srcIndex dest destIndex destLimit oop |
	<option: #SistaVM>
	"0 is deprecated, now this can be handled at Scorch level"
	primIndex = 0 ifTrue: "Array copy, pointer variable object with no inst vars"
		[dest := self internalStackValue: 4.
		destIndex := (objectMemory integerValueOf: (self internalStackValue: 3)) - 1.
		destLimit := (objectMemory integerValueOf: (self internalStackValue: 2)) - 1.
		src := self internalStackValue: 1.
		srcIndex := (objectMemory integerValueOf: (self internalStackValue: 0)) - 1.
		self internalPop: 4.
		destLimit < destIndex ifTrue: [^self].
		(objectMemory isYoung: dest) ifFalse: [objectMemory possibleRootStoreInto: dest].
		0 to: destLimit - destIndex do: 
			[:i| oop := objectMemory fetchPointer: srcIndex + i ofObject: src.
			    objectMemory storePointerUnchecked: destIndex + i ofObject: dest withValue: oop].
		^ self].
	self unknownInlinePrimitive! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 7/1/2021 15:02'!
readImageFromFile: f HeapSize: desiredHeapSize StartingAt: imageOffset
	"Read an image from the given file stream, allocating an amount of memory to its object heap.
	
	 V3: desiredHeapSize is the total size of the heap.  Fail if the image has an unknown format or
	 requires more than the specified amount of memory.

	 Spur: desiredHeapSize is ignored; this routine will attempt to provide at least extraVMMemory's
	 ammount of free space after the image is loaded, taking any free space in teh image into account.
	 extraVMMemory is stored in the image header and is accessible as vmParameterAt: 23.  If
	 extraVMMemory is 0, the value defaults to the default grow headroom.  Fail if the image has an
	 unknown format or if sufficient memory cannot be allocated.

	 Details: This method detects when the image was stored on a machine with the opposite byte
	 ordering from this machine and swaps the bytes automatically. Furthermore, it allows the header
	 information to start 512 bytes into the file, since some file transfer programs for the Macintosh
	 apparently prepend a Mac-specific header of this size. Note that this same 512 bytes of prefix
	 area could also be used to store an exec command on Unix systems, allowing one to launch
	 Smalltalk by invoking the image name as a command."

	| headerStart headerSize headerFlags dataSize oldBaseAddr swapBytes
	  minimumMemory bytesRead bytesToShift heapSize firstSegSize
	  hdrEdenBytes hdrMaxExtSemTabSize hdrNumStackPages allocationReserve |
	<var: #f type: #sqImageFile>
	<var: #heapSize type: #usqInt>
	<var: #dataSize type: #'size_t'>
	<var: #minimumMemory type: #usqInt>
	<var: #desiredHeapSize type: #usqInt>
	<var: #allocationReserve type: #usqInt>
	<var: #headerStart type: #squeakFileOffsetType>
	<var: #imageOffset type: #squeakFileOffsetType>

	metaclassNumSlots := 6.	"guess Metaclass instSize"
	classNameIndex := 6.		"guess (Class instVarIndexFor: 'name' ifAbsent: []) - 1"
	swapBytes := self checkImageVersionFrom: f startingAt: imageOffset.
	headerStart := (self sqImageFilePosition: f) - 4.  "record header start position"

	headerSize			:= self getWord32FromFile: f swap: swapBytes.
	dataSize			:= self getLongFromFile: f swap: swapBytes.
	oldBaseAddr		:= self getLongFromFile: f swap: swapBytes.
	objectMemory specialObjectsOop: (self getLongFromFile: f swap: swapBytes).
	objectMemory lastHash: (self getLongFromFile: f swap: swapBytes). "N.B.  not used."
	savedWindowSize	:= self getLongFromFile: f swap: swapBytes.
	headerFlags		:= self getLongFromFile: f swap: swapBytes.
	self setImageHeaderFlagsFrom: headerFlags.
	extraVMMemory	:= self getWord32FromFile: f swap: swapBytes.
	hdrNumStackPages	:= self getShortFromFile: f swap: swapBytes.
	"4 stack pages is small.  Should be able to run with as few as
	 three. 4 should be comfortable but slow.  8 is a reasonable
	 default.  Can be changed via vmParameterAt: 43 put: n.
	 Can be set as a preference (Info.plist, VM.ini, command line etc).
	 If desiredNumStackPages is already non-zero then it has been
	 set as a preference.  Ignore (but preserve) the header's default."
	numStackPages := desiredNumStackPages ~= 0
						ifTrue: [desiredNumStackPages]
						ifFalse: [hdrNumStackPages = 0
									ifTrue: [self defaultNumStackPages]
									ifFalse: [hdrNumStackPages]].
	desiredNumStackPages := hdrNumStackPages.
	"pad to word boundary.  This slot can be used for anything else that will fit in 16 bits.
	 It is used for the cog code size in Cog.  Preserve it to be polite to other VMs."
	theUnknownShort	:= self getShortFromFile: f swap: swapBytes.
	hdrEdenBytes		:= self getWord32FromFile: f swap: swapBytes.
	objectMemory edenBytes: (desiredEdenBytes ~= 0
						ifTrue: [desiredEdenBytes]
						ifFalse:
							[hdrEdenBytes = 0
									ifTrue: [objectMemory defaultEdenBytes]
									ifFalse: [hdrEdenBytes]]).
	desiredEdenBytes := hdrEdenBytes.
	hdrMaxExtSemTabSize := self getShortFromFile: f swap: swapBytes.
	hdrMaxExtSemTabSize ~= 0 ifTrue:
		[self setMaxExtSemSizeTo: hdrMaxExtSemTabSize].
	"pad to word boundary.  This slot can be used for anything else that will fit in 16 bits.
	 Preserve it to be polite to other VMs."
	the2ndUnknownShort	:= self getShortFromFile: f swap: swapBytes.
	firstSegSize := self getLongFromFile: f swap: swapBytes.
	objectMemory firstSegmentSize: firstSegSize.
	"compare memory requirements with availability"
	allocationReserve := self interpreterAllocationReserveBytes.
	minimumMemory := dataSize
						+ objectMemory newSpaceBytes
						+ allocationReserve.
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[| freeOldSpaceInImage headroom |
			 freeOldSpaceInImage := self getLongFromFile: f swap: swapBytes.
			 headroom := objectMemory
							initialHeadroom: extraVMMemory
							givenFreeOldSpaceInImage: freeOldSpaceInImage.
			 heapSize := objectMemory roundUpHeapSize:
						   dataSize
						+ headroom
						+ objectMemory newSpaceBytes
						+ (headroom > allocationReserve
							ifTrue: [0]
							ifFalse: [allocationReserve])]
		ifFalse:
			[heapSize :=  desiredHeapSize
						+ objectMemory newSpaceBytes
						+ (desiredHeapSize - dataSize > allocationReserve
							ifTrue: [0]
							ifFalse: [allocationReserve]).
			 heapSize < minimumMemory ifTrue:
				[self insufficientMemorySpecifiedError]].

	"allocate a contiguous block of memory for the Squeak heap"
	(self
			allocateMemory: heapSize
			minimum: minimumMemory
			imageFile: f
			headerSize: headerSize) asUnsignedInteger
		ifNil: [self insufficientMemoryAvailableError]
		ifNotNil:
			[:mem|
			objectMemory
				setHeapBase: mem
				memoryLimit: mem + heapSize
				endOfMemory: mem + dataSize].

	"position file after the header"
	self sqImageFile: f Seek: headerStart + headerSize.

	"read in the image in bulk, then swap the bytes if necessary"
	bytesRead := objectMemory readHeapFromImageFile: f dataBytes: dataSize.
	bytesRead ~= dataSize ifTrue: [self unableToReadImageError].

	self ensureImageFormatIsUpToDate: swapBytes.

	"compute difference between old and new memory base addresses"
	bytesToShift := objectMemory memoryBaseForImageRead - oldBaseAddr.
	self initializeInterpreter: bytesToShift.  "adjusts all oops to new location"
	^dataSize! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 11/14/2019 14:27'!
readableFormat: imageVersion
	"Anwer true if images of the given format are readable by this interpreter.
	 Allows a virtual machine to accept selected older image formats."
	<api>
	^imageVersion = self imageFormatVersion "Float words in platform-order"
	   or: [objectMemory hasSpurMemoryManagerAPI not "No compatibility version for Spur as yet"
			and: [imageVersion = self imageFormatCompatibilityVersion]] "Float words in BigEndian order"! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 8/24/2018 19:48'!
reapAndResetErrorCodeTo: theSP header: methodHeader
	"Assuming the primFailCode is non-zero, check if the method consumes the error code
	 and if so, assign it through theSP.  Then zero the primFailCode.  This is infrequent code,
	 so keep it out of the common path."
	<var: 'theSP' type: #'char *'>
	<inline: #never>
	| initialPC |
	self assert: primFailCode ~= 0.
	initialPC := (self initialIPForHeader: methodHeader method: newMethod) + (self sizeOfCallPrimitiveBytecode: methodHeader).
	(objectMemory byteAt: initialPC) = (self longStoreBytecodeForHeader: methodHeader) ifTrue:
		[stackPages longAtPointer: theSP put: self getErrorObjectFromPrimFailCode].
	 primFailCode := 0! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (auto pragmas 12/08) 7/4/2008 08:46'!
receiver
	<inline: true>
	^stackPages longAt: localFP + FoxReceiver! !

!StackInterpreter methodsFor: 'callback support' stamp: 'eem 3/10/2017 12:59'!
reestablishContextPriorToCallback: callbackContext
	"callbackContext is an activation of invokeCallback:[stack:registers:jmpbuf:].
	 Its sender is the VM's state prior to the callback.  Reestablish that state,
	 and mark calloutContext as dead."
	| calloutContext theFP thePage |
	<export: true>
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	self flag: #obsolete.
	(self isLiveContext: callbackContext) ifFalse:
		[^false].
	calloutContext := self externalInstVar: SenderIndex ofContext: callbackContext.
	(self isLiveContext: calloutContext) ifFalse:
		[^false].
	"We're about to leave this stack page; must save the current frame's instructionPointer."
	self push: instructionPointer.
	self externalWriteBackHeadFramePointers.
	"Mark callbackContext as dead; the common case is that it is the current frame.
	 We go the extra mile for the debugger."
	(self isSingleContext: callbackContext)
		ifTrue: [self markContextAsDead: callbackContext]
		ifFalse:
			[theFP := self frameOfMarriedContext: callbackContext.
			 framePointer = theFP "common case"
				ifTrue:
					[(self isBaseFrame: theFP)
						ifTrue: [stackPages freeStackPage: stackPage]
						ifFalse: "calloutContext is immediately below on the same page.  Make it current."
							[instructionPointer := (self frameCallerSavedIP: framePointer) asUnsignedInteger.
							 stackPointer := framePointer + (self frameStackedReceiverOffset: framePointer) + objectMemory wordSize.
							 framePointer := self frameCallerFP: framePointer.
							 ^true]]
				ifFalse:
					[self externalDivorceFrame: theFP andContext: callbackContext.
					 self markContextAsDead: callbackContext]].
	"Make the calloutContext the active frame.  The case where calloutContext
	 is immediately below callbackContext on the same page is handled above."
	(self isStillMarriedContext: calloutContext)
		ifTrue:
			[theFP := self frameOfMarriedContext: calloutContext.
			 thePage := stackPages stackPageFor: theFP.
			 "findSPOf:on: points to the word beneath the instructionPointer, but
			  there is no instructionPointer on the top frame of the current page."
			 self assert: thePage ~= stackPage.
			 stackPointer := (self findSPOf: theFP on: thePage) - objectMemory wordSize.
			 framePointer := theFP.
			 self assert: stackPointer < framePointer]
		ifFalse:
			[thePage := self makeBaseFrameFor: calloutContext.
			 self setStackPointersFromPage: thePage].
	instructionPointer := self popStack.
	self setStackPageAndLimit: thePage.
	^true! !

!StackInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'rsf 9/5/2016 17:33'!
reloadLowcodeStateAfterCallout
	<option: #LowcodeVM>
	<var: #initialShadowCallStackPointer type: #'char*' >
	
	| calloutStateSize initialShadowCallStackPointer |
	calloutStateSize := self sizeof: #'sqLowcodeCalloutState'.
	initialShadowCallStackPointer := (self shadowCallStackPointerIn: localFP) - 1.
	
	initialShadowCallStackPointer := self cCoerce: ((self cCoerce: initialShadowCallStackPointer - calloutStateSize to: 'size_t') bitAnd: -16) to: 'char*'.
	lowcodeCalloutState := self cCoerce: initialShadowCallStackPointer to: #'sqLowcodeCalloutState*'.! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 9/14/2013 11:13'!
remapCallbackState
	"Callback support - trace suspended callback list"
	1 to: jmpDepth do:
		[:i| | oop |
		oop := suspendedCallbacks at: i.
		(objectMemory shouldRemapOop: oop) ifTrue:
			[suspendedCallbacks at: i put: (objectMemory remapObj: oop)].
		oop := suspendedMethods at: i.
		(objectMemory shouldRemapObj: oop) ifTrue:
			[suspendedMethods at: i put: (objectMemory remapObj: oop)]]! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 4/5/2016 17:28'!
remoteIsInstVarAccess
	"If this byte is set in the second byte (zero-based) of the instruction, 
	the remote temp instruction is in fact a remote inst var access"
	<api>
	<cmacro>
	^ 128 "1 << 7"! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 7/6/2015 11:00'!
removeFirstLinkOfList: aList 
	"Remove the first process from the given linked list."
	| first last next |
	self deny: (objectMemory isForwarded: aList).
	first := objectMemory fetchPointer: FirstLinkIndex ofObject: aList.
	last := objectMemory fetchPointer: LastLinkIndex ofObject: aList.
	self deny: (objectMemory isForwarded: first).
	self deny: (objectMemory isForwarded: last).
	first = last
		ifTrue:
			[objectMemory
				storePointerUnchecked: FirstLinkIndex ofObject: aList withValue: objectMemory nilObject;
				storePointerUnchecked: LastLinkIndex ofObject: aList withValue: objectMemory nilObject]
		ifFalse:
			[next := objectMemory fetchPointer: NextLinkIndex ofObject: first.
			 objectMemory storePointer: FirstLinkIndex ofObject: aList withValue: next].
	objectMemory storePointerUnchecked: NextLinkIndex ofObject: first withValue: objectMemory nilObject.
	^first! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 8/29/2020 19:17'!
removeProcess: aProcess fromList: aList 
	"Remove a given process from a linked list. May fail if aProcess is not on the list."
	| firstLink lastLink nextLink tempLink |
	self deny: (objectMemory isForwarded: aProcess).
	self deny: (objectMemory isForwarded: aList).
	firstLink := objectMemory fetchPointer: FirstLinkIndex ofObject: aList.
	lastLink := objectMemory fetchPointer: LastLinkIndex ofObject: aList.
	self deny: (objectMemory isForwarded: firstLink).
	self deny: (objectMemory isForwarded: lastLink).
	aProcess  = firstLink
		ifTrue:
			[nextLink := objectMemory fetchPointer: NextLinkIndex ofObject: aProcess.
			 self deny: (objectMemory isForwarded: nextLink).
			 objectMemory storePointer: FirstLinkIndex ofObject: aList withValue: nextLink.
			 aProcess  = lastLink ifTrue:
				[objectMemory storePointerUnchecked: LastLinkIndex ofObject: aList withValue: objectMemory nilObject]]
		ifFalse:
			[tempLink := firstLink.
			 [self deny: (objectMemory isForwarded: tempLink).
			  tempLink = objectMemory nilObject ifTrue:
				[self primitiveFail. ^self].
			  nextLink := objectMemory fetchPointer: NextLinkIndex ofObject: tempLink.
			  nextLink = aProcess] whileFalse:
				[tempLink := objectMemory fetchPointer: NextLinkIndex ofObject: tempLink].
			 nextLink := objectMemory fetchPointer: NextLinkIndex ofObject: aProcess.
			 objectMemory storePointer: NextLinkIndex ofObject: tempLink withValue: nextLink.
			 aProcess  = lastLink ifTrue:
				[objectMemory storePointer: LastLinkIndex ofObject: aList withValue: tempLink]].
	objectMemory storePointerUnchecked: NextLinkIndex ofObject: aProcess withValue: objectMemory nilObject! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'eem 10/29/2020 10:36'!
respondToSistaTrap
	| ourContext |
	<sharedCodeInCase: #unconditionalTrapBytecode>
	messageSelector := objectMemory splObj: SelectorSistaTrap.
	ourContext := self ensureFrameIsMarried: localFP SP: localSP.
	self internalPush: ourContext.
	argumentCount := 0.
	self normalSend! !

!StackInterpreter methodsFor: 'miscellaneous bytecodes' stamp: 'eem 7/26/2014 06:54'!
respondToUnknownBytecode
	"If an error selector is available then send it to the activeContext, otherwise abort."
	<sharedCodeInCase: #unknownBytecode>
	| ourContext |
	messageSelector := objectMemory maybeSplObj: SelectorUnknownBytecode.
	(messageSelector isNil
	or: [messageSelector = objectMemory nilObject]) ifTrue:
		[self error: 'Unknown bytecode'].
	ourContext := self ensureFrameIsMarried: localFP SP: localSP.
	"undo fetch of bytecode so that context's pc is pointing to the unknown bytecode."
	localIP := localIP - 1.
	self internalPush: ourContext.
	argumentCount := 0.
	self normalSend! !

!StackInterpreter methodsFor: 'callback support' stamp: 'eem 9/19/2018 17:31'!
restoreCStackStateForCallbackContext: vmCallbackContext
	<var: #vmCallbackContext type: #'VMCallbackContext *'>
	self memcpy: reenterInterpreter
		_: vmCallbackContext savedReenterInterpreter asVoidPointer
		_: (self sizeof: #'jmp_buf')! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 2/8/2009 11:19'!
restoreStackLimit
	"restore the stackLimit if it has been smashed."
	<inline: true>
	stackPage stackLimit: stackPage realStackLimit.
	stackLimit := stackPage stackLimit! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 11/15/2012 11:11'!
resume: aProcess
	"Replaced by resume:preemptedYieldingIf:"
	"Make aProcess runnable and if its priority is higher than
	 that of the current process, preempt the current process.
	 Answer if the current process was preempted.  Override
	 to add tracing info (see resume:from:)."
	<doNotGenerate>
	self shouldNotImplement! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'qwaq 4/23/2010 11:39'!
resume: aProcess preemptedYieldingIf: yieldImplicitly
	"Make aProcess runnable and if its priority is higher than  that of the
	 current process, preempt the current process.   Answer if the current
	 process was preempted.  If the current process was preempted then if
	 yieldImplicitly add the current process to the back of its run queue,
	 causing an implicit yiled to other processes on the run queue,  otherwise
	 add the current process to the front of its run queue, hence not yielding.
	 Blue book behaviour is to yield implicitly but is arguably incorrect."
	| activeProc activePriority newPriority |
	<inline: false>
	activeProc := self activeProcess.
	activePriority := self quickFetchInteger: PriorityIndex ofObject: activeProc.
	newPriority := self quickFetchInteger: PriorityIndex ofObject: aProcess.
	newPriority <= activePriority ifTrue:
		[self putToSleep: aProcess yieldingIf: true.
		 ^false].
	self putToSleep: activeProc yieldingIf: yieldImplicitly.
	self transferTo: aProcess.
	^true! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 5/7/2020 16:17'!
retryPrimitiveOnFailure
	"In Spur two cases of primitive failure are handled specially.  A primitive may fail due to validation
	 encountering a forwarder. On failure, check the accessorDepth for the primitive and if non-negative
	 scan the args to the depth, following any forwarders.  Retry the primitive if any are found.  Hence
	 lazily and transparently following forwarders on primitive failure.  Additionally a primitive might fail
	 due to an allocation failing.  Retry if external primitives have failed with PrimErrNoMemory after running
	 first the scavenger and then on a subsequent failure, the global mark-sweep collector.  Hence lazily
	 and transparently GC on memory exhaustion."
	<option: #SpurObjectMemory>
	<inline: false>
	| gcDone followDone canRetry retry retried |
	gcDone := 0.
	followDone := canRetry := retried := false.
	[retry := false.
	 primFailCode = PrimErrNoMemory
		ifTrue:
			[(gcDone := gcDone + 1) = 1 ifTrue:
				[canRetry := self isExternalPrimitiveCall: newMethod].
			 canRetry ifTrue:
				 [gcDone = 1 ifTrue:
					[objectMemory scavengingGC].
				 gcDone = 2 ifTrue:
					[objectMemory fullGC].
				 retry := gcDone <= 2]]
		 ifFalse:
			[followDone ifFalse:
				[followDone := true.
				 retry := self checkForAndFollowForwardedPrimitiveState]].
	 retry] whileTrue:
		[self assert: primFailCode ~= 0.
		 retried := true.
		 self initPrimCall.
		 self cCode: [] inSmalltalk:
			[self maybeMapPrimitiveFunctionPointerBackToSomethingEvaluable].
		 self dispatchFunctionPointer: primitiveFunctionPointer].
	^retried! !

!StackInterpreter methodsFor: 'callback support' stamp: 'eem 9/1/2020 14:26'!
returnAs: returnTypeOop ThroughCallback: vmCallbackContext Context: callbackMethodContext
	"callbackMethodContext is an activation of invokeCallback:[stack:registers:jmpbuf:].
	 Its sender is the VM's state prior to the callback.  Reestablish that state (via longjmp),
	 and mark callbackMethodContext as dead."
	<export: true>
	<var: #vmCallbackContext type: #'VMCallbackContext *'>
	| calloutMethodContext theFP thePage |
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	self assert: primFailCode = 0.
	self assert: (objectMemory isIntegerObject: returnTypeOop).
	self assert: (objectMemory isImmediate: vmCallbackContext asInteger) not.
	self assert: ((objectMemory addressCouldBeObj: callbackMethodContext)
				and: [objectMemory isContext: callbackMethodContext]).
	self assert: (debugCallbackPath := 0) = 0.
	((objectMemory isIntegerObject: returnTypeOop)
	 and: [self isLiveContext: callbackMethodContext]) ifFalse:
		[self assert: (debugCallbackPath := 1) = 1.
		 ^false].
	calloutMethodContext := self externalInstVar: SenderIndex ofContext: callbackMethodContext.
	(self isLiveContext: calloutMethodContext) ifFalse:
		[self assert: (debugCallbackPath := 2) = 2.
		 ^false].
	self assert: (debugCallbackReturns := debugCallbackReturns + 1) > 0.
	"self assert: debugCallbackReturns < 3802."
	"We're about to leave this stack page; must save the current frame's instructionPointer."
	self push: instructionPointer.
	self externalWriteBackHeadFramePointers.
	"Mark callbackMethodContext as dead; the common case is that it is the current frame.
	 We go the extra mile for the debugger."
	(self isSingleContext: callbackMethodContext)
		ifTrue:
			[self assert: (debugCallbackPath := debugCallbackPath bitOr: 4) > 0.
		 	 self markContextAsDead: callbackMethodContext]
		ifFalse:
			[self assert: (debugCallbackPath := debugCallbackPath bitOr: 8) > 0.
		 	 theFP := self frameOfMarriedContext: callbackMethodContext.
			 self assert: (self frameReceiver: theFP) = (objectMemory splObj: ClassAlien).
			 framePointer = theFP "common case"
				ifTrue:
					[self assert: (debugCallbackPath := debugCallbackPath bitOr: 16) > 0.
		 			 (self isBaseFrame: theFP) ifFalse: "calloutMethodContext is immediately below on the same page.  Make it current."
						[self assert: (debugCallbackPath := debugCallbackPath bitOr: 32) > 0.
		 				 instructionPointer := (self frameCallerSavedIP: theFP) asUnsignedInteger.
						 stackPointer := theFP + (self frameStackedReceiverOffset: theFP) + objectMemory wordSize.
						 framePointer := self frameCallerFP: theFP.
						 self setMethod: (self frameMethodObject: framePointer).
						 self restoreCStackStateForCallbackContext: vmCallbackContext.
						 self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer.
						 "N.B. siglongjmp is defines as _longjmp on non-win32 platforms.
						  This matches the use of _setjmp in ia32abicc.c."
						 self _longjmp: vmCallbackContext trampoline _: (self integerValueOf: returnTypeOop).
						 ^true].
					 stackPages freeStackPage: stackPage]
				ifFalse:
					[self assert: (debugCallbackPath := debugCallbackPath bitOr: 64) > 0.
		 			 self externalDivorceFrame: theFP andContext: callbackMethodContext.
					 self markContextAsDead: callbackMethodContext]].
	"Make the calloutMethodContext the active frame.  The case where calloutMethodContext
	 is immediately below callbackMethodContext on the same page is handled above."
	(self isStillMarriedContext: calloutMethodContext)
		ifTrue:
			[self assert: (debugCallbackPath := debugCallbackPath bitOr: 128) > 0.
		 	 theFP := self frameOfMarriedContext: calloutMethodContext.
			 thePage := stackPages stackPageFor: theFP.
			 "findSPOf:on: points to the word beneath the instructionPointer, but
			  there is no instructionPointer on the top frame of the current page."
			 self assert: thePage ~= stackPage.
			 stackPointer := thePage headFP = theFP
								ifTrue: [thePage headSP]
								ifFalse: [(self findSPOf: theFP on: thePage) - objectMemory wordSize].
			 framePointer := theFP.
			 self assert: stackPointer < framePointer]
		ifFalse:
			[self assert: (debugCallbackPath := debugCallbackPath bitOr: 256) > 0.
		 	 thePage := self makeBaseFrameFor: calloutMethodContext.
			 self setStackPointersFromPage: thePage].
	instructionPointer := self popStack.
	self setMethod: (objectMemory fetchPointer: MethodIndex ofObject: calloutMethodContext).
	self setStackPageAndLimit: thePage.
	self restoreCStackStateForCallbackContext: vmCallbackContext.
	primitiveFunctionPointer := vmCallbackContext savedPrimFunctionPointer.
	"N.B. siglongjmp is defined as _longjmp on non-win32 platforms.
	  This matches the use of _setjmp in ia32abicc.c."
	self _longjmp: vmCallbackContext trampoline _: (self integerValueOf: returnTypeOop).
	"NOTREACHED"
	^true! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
returnFalse
	localReturnValue := objectMemory falseObject.
	self commonReturn! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
returnNil
	localReturnValue := objectMemory nilObject.
	self commonReturn! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem 7/25/2014 07:07'!
returnNilFromBlock
	"Return nil to the caller of the current block activation."
	localReturnValue := objectMemory nilObject.
	self commonCallerReturn! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem 6/17/2008 11:35'!
returnReceiver
	localReturnValue := self receiver.
	self commonReturn! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem 7/25/2014 07:07'!
returnTopFromBlock
	"Return top-of-stack to the caller of the current block activation."
	localReturnValue := self internalStackTop.
	self commonCallerReturn! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem 6/17/2008 11:35'!
returnTopFromMethod
	localReturnValue := self internalStackTop.
	self commonReturn! !

!StackInterpreter methodsFor: 'return bytecodes' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
returnTrue
	localReturnValue := objectMemory trueObject.
	self commonReturn! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 10/9/2013 14:32'!
reverseBytesInImage
	"Byte-swap all words in memory after reading in the entire image file with bulk read. Contributed by Tim Rowledge."

	"First, byte-swap every word in the image. This fixes objects headers."
	objectMemory reverseBytesInMemory.

	"Second, return the bytes of bytes-type objects to their
	 orginal order, and perform any other format conversions."
	self updateObjectsPostByteSwap! !

!StackInterpreter methodsFor: 'I/O primitive support' stamp: 'eem 1/15/2018 13:08'!
reverseDisplayFrom: startIndex to: endIndex 
	"Reverse the given range of Display pixels, rounded to whole word boundary.
	Used to give feedback during VM activities such as garbage collection when debugging.
	 It is assumed that the given word range falls entirely within the first line of the Display."
	
	| wordStartIndex wordEndIndex primFailCodeValue |
	self postGCUpdateDisplayBits ifFalse:
		[^self].
	(displayBits = 0 or: [(objectMemory isImmediate: displayBits asInteger) or: [displayDepth <= 0]]) ifTrue: [^nil].
	wordStartIndex := (startIndex max: 0) * displayDepth // 32.
	wordEndIndex := (endIndex min: displayWidth) * displayDepth // 32.
	displayBits asInteger + (wordStartIndex * 4) to: displayBits asInteger + (wordEndIndex * 4) by: 4 do:
		[:ptr | | reversed |
		reversed := (objectMemory long32At: ptr) bitXor: 16rFFFFFFFF.
		objectMemory long32At: ptr put: reversed].
	primFailCodeValue := primFailCode.
	self initPrimCall.
	self updateDisplayLeft: 0 Top: 0 Right: displayWidth Bottom: 1.
	self ioForceDisplayUpdate.
	primFailCode := primFailCodeValue! !

!StackInterpreter methodsFor: 'method lookup cache' stamp: 'nice 11/2/2016 21:11'!
rewriteMethodCacheEntryForExternalPrimitiveToFunction: localPrimAddress
	"Rewrite an existing entry in the method cache with a new primitive function address.
	 Used by primitiveExternalCall to make direct calls to found external prims, or quickly
	 fail not found external prims."
	<inline: false>
	<var: #localPrimAddress declareC: 'void (*localPrimAddress)(void)'>
	(methodCache at: lastMethodCacheProbeWrite + MethodCacheMethod) = newMethod ifTrue:
		[methodCache
			at: lastMethodCacheProbeWrite + MethodCachePrimFunction
			put: (self cCoerce: localPrimAddress to: #'sqIntptr_t')]! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 08/08/2014 10:42'!
roomToPushNArgs: n
	"Answer if there is room to push n arguments onto the current stack.  We assume
	 this is called by primitives that check there is enough room in any new context, and
	 won't actually push the arguments in the current context if the primitive fails.  With
	 this assumption it is safe to answer based on the maximum argument count, /not/
	 the ammount of space in the current frame were it converted to a context.."
	false
		ifTrue: "old code that checked size of context..."
			[| cntxSize |
			 self assert: method = (stackPages longAt: framePointer + FoxMethod).
			 cntxSize := (self methodHeaderIndicatesLargeFrame: (objectMemory methodHeaderOf: method))
							ifTrue: [LargeContextSlots - CtxtTempFrameStart]
							ifFalse: [SmallContextSlots - CtxtTempFrameStart].
			 ^self stackPointerIndex + n <= cntxSize]
		ifFalse: "simpler code that simply insists args are <= max arg count"
			[^n <= (LargeContextSlots - CtxtTempFrameStart)]! !

!StackInterpreter methodsFor: 'simulation' stamp: 'eem 9/30/2015 12:24'!
runLeakChecker
	<doNotGenerate>
	| oldCheckForLeaks |
	oldCheckForLeaks := objectMemory checkForLeaks.
	objectMemory setCheckForLeaks: -1.
	[objectMemory runLeakCheckerFor: GCModeIncremental]
		ensure:
			[objectMemory setCheckForLeaks: oldCheckForLeaks]! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 2/6/2015 16:26'!
safeFindFrameAbove: theFP on: thePage startingFrom: startFrame
	"Search for the previous frame to theFP (the frame that theFP calls).
	 DO NOT CALL THIS WITH theFP == localFP OR theFP == framePointer!!"
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<var: #startFrame type: #'char *'>
	<returnTypeC: #'char *'>
	| aFrame prevFrame |
	<inline: true>
	<var: #aFrame type: #'char *'>
	<var: #prevFrame type: #'char *'>
	self assert: (stackPages isFree: thePage) not.
	startFrame = theFP ifTrue:
		[^nil].
	aFrame := startFrame.
	[prevFrame := aFrame.
	 aFrame := self frameCallerFP: aFrame.
	 aFrame ~= 0] whileTrue:
		[theFP = aFrame ifTrue:
			[^prevFrame]].
	^nil! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 8/23/2014 11:21'!
safeMethodClassOf: methodPointer
	"Safe version of methodClassOf: that deals with malformed compiled methods,
	 etc, and does not fixup forwarding pointers.."
	| literal maybeClass |
	literal := self literal: (objectMemory literalCountOf: methodPointer) - 1 ofMethod: methodPointer.
	(objectMemory isOopForwarded: literal) ifTrue:
		[literal := objectMemory followForwarded: literal].
	((objectMemory isPointers: literal) and: [(objectMemory numSlotsOf: literal) > ValueIndex]) ifFalse:
		[^objectMemory nilObject].
	maybeClass := objectMemory fetchPointer: ValueIndex ofObject: literal.
	(objectMemory isOopForwarded: maybeClass) ifTrue:
		[maybeClass := objectMemory followForwarded: maybeClass].
	^maybeClass! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 5/7/2020 17:18'!
saneFunctionPointerForFailureOfPrimIndex: primIndex
	"This is an assert function used to ensure consistency between the primitiveFunctionPointer
	 and the primitive index when a primitive fails in Spur.  Since Spur automagically retries
	 primitives that fail and are found to have a forwarder within the primitive's accessor depth
	 we want to know that the primitiveFuncitonPointer is actually valid.  This isn't always
	 possible for the `indirect'' primitives (calling a plugin primitive, calling the ffi, using the
	 receiver:tryPrimitive:withArgs: primitive evaluator in the simulator. We do the best we can."
	| basePrimitive |
	<var: 'basePrimitive' declareC: 'void (*basePrimitive)(void)'>
	basePrimitive := self functionPointerFor: primIndex inClass: objectMemory nilObject.
	^primitiveFunctionPointer = basePrimitive
	  or: [((self isCalloutPrimitiveIndex: primIndex) and: [self isPrimitiveFunctionPointerAnIndex not])
	  or: [(self isMetaPrimitiveIndex: primIndex) and: [metaAccessorDepth > -2]]]! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 8/23/2014 16:48'!
saneMethodClassAssociationIn: methodObj numLiterals: numLiterals
	| shouldBeAssoc |
	shouldBeAssoc := self literal: numLiterals - 1 ofMethod: methodObj.
	(objectMemory isForwarded: shouldBeAssoc) ifTrue:
		[shouldBeAssoc := objectMemory followForwarded: shouldBeAssoc].
	^(objectMemory isPointers: shouldBeAssoc)
	 and: [(objectMemory numSlotsOf: shouldBeAssoc) > ValueIndex
	 and: [self addressCouldBeClassObj: (objectMemory noFixupFollowField: ValueIndex ofObject: shouldBeAssoc)]]! !

!StackInterpreter methodsFor: 'callback support' stamp: 'eem 9/19/2018 17:31'!
saveCStackStateForCallbackContext: vmCallbackContext
	<var: #vmCallbackContext type: #'VMCallbackContext *'>
	self memcpy: vmCallbackContext savedReenterInterpreter asVoidPointer
		_: reenterInterpreter
		_: (self sizeof: #'jmp_buf')! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
schedulerPointer

	^ objectMemory fetchPointer: ValueIndex ofObject: (objectMemory splObj: SchedulerAssociation)! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: ''!
secondExtendedSendBytecode
	"This replaces the Blue Book double-extended super-send [134],
	which is subsumed by the new double-extended do-anything [132].
	It offers a 2-byte send of 0-3 args for up to 63 literals, for which 
	the Blue Book opcode set requires a 3-byte instruction."

	| descriptor |
	descriptor := self fetchByte.
	messageSelector := self literal: (descriptor bitAnd: 16r3F).
	argumentCount := descriptor >> 6.
	self normalSend.
! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/3/2013 10:29'!
selectorOfContext: aContext
	(objectMemory isContext: aContext) ifFalse:
		[^nil].
	^self findSelectorOfMethod: (objectMemory fetchPointer: MethodIndex ofObject: aContext)! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'rmacnak 5/18/2015 21:13'!
sendAbsentImplicit0ArgsBytecode
	"160-175	1010 i i i i		Send To Absent Implicit Receiver Literal Selector #iiii With 0 Arguments"
	messageSelector := self literal: (currentBytecode bitAnd: 16rF).
	argumentCount := 0.
	self commonSendImplicitReceiver! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 2/24/2020 13:59'!
sendBreak: selectorString point: selectorLength classTag: classTag
	<doNotGenerate> "C version is in platforms/Cross/vm/dispdbg.h"
	"self shortPrintFrameAndCallers: localFP"
	| i |
	breakSelectorLength = selectorLength ifTrue:
		[i := breakSelectorLength.
		 [i > 0] whileTrue:
			[(objectMemory byteAt: selectorString + i - 1) = (breakSelector at: i) asInteger
				ifTrue:
					[((i := i - 1) = 0
					  and: [breakLookupClassTag < 0 or: [breakLookupClassTag = classTag]]) ifTrue:
						[self changed: #byteCountText.
						 self halt: 'Send of ', breakSelector]]
				ifFalse: [i := 0]]]! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 2/24/2020 14:09'!
sendBreakpoint: selector receiver: rcvr
	<inline: true>
	self sendBreak: (objectMemory firstFixedFieldOfMaybeImmediate: selector)
		point: (objectMemory lengthOfMaybeImmediate: selector)
		classTag: (objectMemory fetchClassTagOf: rcvr)! !

!StackInterpreter methodsFor: 'callback support' stamp: 'eem 4/25/2020 18:43'!
sendInvokeCallback: thunkPtr Stack: stackPtr Registers: regsPtr Jmpbuf: jmpBufPtr
	"Send the 4 argument callback message invokeCallback:stack:registers:jmpbuf:
	 to Alien class with the supplied args.  The arguments are raw C addresses
	 and are converted to integer objects on the way."
	<export: true>
	| classTag |
	classTag := self fetchClassTagOfNonImm: (objectMemory splObj: ClassAlien).
	messageSelector := self splObj: SelectorInvokeCallback.
	argumentCount := 4.
	(self lookupInMethodCacheSel: messageSelector classTag: classTag) ifFalse:
	 	[(self lookupOrdinaryNoMNUEtcInClass: (objectMemory classForClassTag: classTag)) ~= 0 ifTrue:
			[^false]].
	((self argumentCountOf: newMethod) = 4
	and: [primitiveFunctionPointer = 0]) ifFalse:
		[^false].
	self push: (self splObj: ClassAlien). "receiver"
	self push: (self positiveMachineIntegerFor: thunkPtr).
	self push: (self positiveMachineIntegerFor: stackPtr).
	self push: (self positiveMachineIntegerFor: regsPtr).
	self push: (self positiveMachineIntegerFor: jmpBufPtr).
	self ifAppropriateCompileToNativeCode: newMethod selector: messageSelector.
	self justActivateNewMethod: false. "either interpreted or machine code"
	(self isMachineCodeFrame: framePointer) ifFalse:
		[self maybeFlagMethodAsInterpreted: newMethod].
	self checkForStackOverflow.
	self enterSmalltalkExecutiveFromCallback.
	"not reached"
	^true! !

!StackInterpreter methodsFor: 'callback support' stamp: 'eem 8/23/2018 11:14'!
sendInvokeCallbackContext: vmCallbackContext
	"Send the calllback message to Alien class with the supplied arg(s).  Use either the 1 arg
	 invokeCallbackContext: or the 4 arg invokeCallback:stack:registers:jmpbuf: message,
	 depending on what selector is installed in the specialObjectsArray. Note that if invoking the
	 legacy invokeCallback:stack:registers:jmpbuf: we pass the vmCallbackContext as the jmpbuf
	 argument (see reestablishContextPriorToCallback:). The arguments are raw C addresses and
	 are converted to integer objects on the way. sendInvokeCallbackContext: &
	 returnAs:ThroughCallback:Context: along with ownVM: and disownVM: conspire to save and
	 restore newMethod, argumentCount and primitiveFunctionPointer around a callback.
	 The VM depends on argumentCount being correct to cut-back the correct number of
	 arguments on primitive return.  If a primitive that invokes a callback fails after invoking a
	 callback (a bad idea, but s**t happens during development) then newMethod is required to
	 activate the right faling method, and Spur expects primitiveFunctionPointer to be valid, so
	 asserts will fail misleadingly if not."
	<export: true>
	<var: #vmCallbackContext type: #'VMCallbackContext *'>
	| classTag |
	vmCallbackContext savedPrimFunctionPointer: primitiveFunctionPointer.
	classTag := self fetchClassTagOfNonImm: (objectMemory splObj: ClassAlien).
	messageSelector := self splObj: SelectorInvokeCallback.
	(self lookupInMethodCacheSel: messageSelector classTag: classTag) ifFalse:
	 	[(self lookupOrdinaryNoMNUEtcInClass: (objectMemory classForClassTag: classTag)) ~= 0 ifTrue:
			[^false]].
	primitiveFunctionPointer ~= 0 ifTrue:
		[primitiveFunctionPointer := vmCallbackContext savedPrimFunctionPointer.
		 ^false].
	self assert: (debugCallbackInvokes := debugCallbackInvokes + 1) > 0.
	"self assert: debugCallbackInvokes < 3802."
	self saveCStackStateForCallbackContext: vmCallbackContext.
	self push: (objectMemory splObj: ClassAlien). "receiver"
	(self argumentCountOf: newMethod) = 4 ifTrue:
		[self push: (self positiveMachineIntegerFor: vmCallbackContext thunkp asUnsignedInteger).
		 self push: (self positiveMachineIntegerFor: vmCallbackContext stackp asUnsignedInteger).
		 self push: (self positiveMachineIntegerFor: vmCallbackContext intregargsp asUnsignedInteger)].
	self push: (self positiveMachineIntegerFor: vmCallbackContext asUnsignedInteger).
	self ifAppropriateCompileToNativeCode: newMethod selector: messageSelector.
	self justActivateNewMethod: false. "either interpreted or machine code"
	(self isMachineCodeFrame: framePointer) ifFalse:
		[self maybeFlagMethodAsInterpreted: newMethod].
	self checkForStackOverflow.
	self assert: (self frameReceiver: framePointer) = (objectMemory splObj: ClassAlien).
	self enterSmalltalkExecutiveFromCallback.
	"not reached"
	^true! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'rmacnak 5/19/2015 20:27'!
sendLiteralSelector0ArgsBytecode
	"Can use any of the first 16 literals for the selector."
	| rcvr |
	messageSelector := self literal: (currentBytecode bitAnd: 16rF).
	argumentCount := 0.
	rcvr := self internalStackValue: 0.
	lkupClassTag := objectMemory fetchClassTagOf: rcvr.
	self assert: lkupClassTag ~= objectMemory nilObject.
	self commonSendOrdinary! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'rmacnak 5/19/2015 20:27'!
sendLiteralSelector1ArgBytecode
	"Can use any of the first 16 literals for the selector."
	| rcvr |
	messageSelector := self literal: (currentBytecode bitAnd: 16rF).
	argumentCount := 1.
	rcvr := self internalStackValue: 1.
	lkupClassTag := objectMemory fetchClassTagOf: rcvr.
	self assert: lkupClassTag ~= objectMemory nilObject.
	self commonSendOrdinary! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'rmacnak 5/19/2015 20:27'!
sendLiteralSelector2ArgsBytecode
	"Can use any of the first 16 literals for the selector."
	| rcvr |
	messageSelector := self literal: (currentBytecode bitAnd: 16rF).
	argumentCount := 2.
	rcvr := self internalStackValue: 2.
	lkupClassTag := objectMemory fetchClassTagOf: rcvr.
	self assert: lkupClassTag ~= objectMemory nilObject.
	self commonSendOrdinary! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 2/24/2020 14:12'!
setBreakLookupClassTag: aClassTag
	breakLookupClassTag := aClassTag! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 4/13/2021 17:46'!
setBreakMNUSelector: aString
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	<var: #aString type: #'char *'>
	(breakSelector := aString)
		ifNil: [breakSelectorLength := objectMemory minSmallInteger "nil's effective length is zero"]
		ifNotNil: [breakSelectorLength := (self strlen: aString) negated]! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 4/13/2021 17:46'!
setBreakSelector: aString
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	<var: #aString type: #'char *'>
	(breakSelector := aString)
		ifNil: [breakSelectorLength := objectMemory minSmallInteger "nil's effective length is zero"]
		ifNotNil: [breakSelectorLength := self strlen: aString]! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 1/10/2017 11:11'!
setDesiredCogCodeSize: dccs
	<inline: true>
	"This should perhaps be a no-op in the StackVM, but we implement it for the convenience of bootstrap scripts etc."
	theUnknownShort := dccs! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 6/21/2012 13:08'!
setFrameContext: theFP to: aContext
	<inline: true>
	<var: #theFP type: #'char *'>
	stackPages longAt: theFP + FoxThisContext put: aContext! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 6/21/2012 13:05'!
setFrameHasContext: theFP
	"See encodeFrameFieldHasContext:numArgs:"
	<inline: true>
	<var: #theFP type: #'char *'>
	stackPages byteAt: theFP + FoxFrameFlags + 2 put: 1! !

!StackInterpreter methodsFor: 'frame access' stamp: 'RonieSalgado 5/31/2016 19:07'!
setFrameHasNativeFrame: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	self temporary: (self frameNumArgs: theFP) in: theFP put: (objectMemory splObj: LowcodeContextMark)! !

!StackInterpreter methodsFor: 'frame access' stamp: 'rsf 8/9/2016 23:36'!
setFrameHasNotNativeFrame: theFP
	<inline: true>
	<option: #LowcodeVM>
	<var: #theFP type: #'char *'>
	self temporary: (self frameNumArgs: theFP) in: theFP put: (objectMemory nilObject)! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/16/2002 13:44'!
setFullScreenFlag: value
	fullScreenFlag := value! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'cb 5/28/2018 15:39'!
setGCMode: mode
	"This is a no-op in the StackVM"
	^0! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 11/13/2014 13:17'!
setHeadFP: theFP andSP: theSP inPage: thePage
	<inline: true>
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	self assert: theSP < theFP.
	self assert: (theSP < thePage baseAddress
				and: [theSP > (thePage realStackLimit - (LargeContextSlots * objectMemory bytesPerOop))]).
	self assert: (theFP < thePage baseAddress
				and: [theFP > (thePage realStackLimit - (LargeContextSlots * objectMemory bytesPerOop / 2))]).
	thePage headFP: theFP; headSP: theSP! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 11/8/2020 22:53'!
setImageHeaderFlags: flags
	"Set an array of flags indicating various properties of the saved image, responded to on image load.
	 These are the same as the image header flags shifted right two bits, omitting the fullScreenFlag and float byte order flag.
	 Bit 0: if set, implies the image's Process class has threadId as its 3rd inst var (zero relative) (meaningful to the MT VM only)
	 Bit 1: if set, methods that are interpreted will have the flag bit set in their header (meaningful to the Cog VM only)
	 Bit 2: if set, implies preempting a process does not put it to the back of its run queue
	 Bit 3: if set, implies a threaded VM will not disown the VM if owned by the GUI thread (meaningful to the MT VM only)
	 Bit 4: if set, implies the new finalization scheme where WeakArrays are queued
	 Bit 5: if set, implies wheel events will be delivered as such and not mapped to arrow key events
	 Bit 6: if set, implies arithmetic primitives will fail if given arguments of different types (float vs int)
	 Bit 7: if set, implies file primitives (FilePlugin, FileAttributesPlugin) will answer file times in UTC not local times"
	flags asUnsignedInteger > 255 ifTrue:
		[^self primitiveFailFor: PrimErrUnsupported].
	"processHasThreadId := flags anyMask: 1. specific to CoInterpreterMT"
	imageHeaderFlags := (flags anyMask: 1)
							ifTrue: [imageHeaderFlags bitOr: 4]
							ifFalse: [imageHeaderFlags bitClear: 4].
	"flagInterpretedMethods := flags anyMask: 2. specific to CoInterpreter"
	imageHeaderFlags := (flags anyMask: 2)
							ifTrue: [imageHeaderFlags bitOr: 8]
							ifFalse: [imageHeaderFlags bitClear: 8].
	preemptionYields := flags noMask: 4.
	"noThreadingOfGUIThread := flags anyMask: 8.. specific to CoInterpreterMT"
	imageHeaderFlags := (flags anyMask: 8)
							ifTrue: [imageHeaderFlags bitOr: 32]
							ifFalse: [imageHeaderFlags bitClear: 32].
	newFinalization := flags anyMask: 16.
	sendWheelEvents := flags anyMask: 32.
	primitiveDoMixedArithmetic := flags noMask: 64.
	imageHeaderFlags := (flags anyMask: 128)
							ifTrue: [imageHeaderFlags bitOr: 512]
							ifFalse: [imageHeaderFlags bitClear: 512]! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 9/6/2019 14:18'!
setImageHeaderFlagsFrom: headerFlags
	"Set the flags that are contained in the 7th long of the image header."
	imageHeaderFlags := headerFlags. "so as to preserve unrecognised flags."
	fullScreenFlag := headerFlags bitAnd: 1.
	imageFloatsBigEndian := (headerFlags noMask: 2) ifTrue: [1] ifFalse: [0].
	"processHasThreadId := headerFlags anyMask: 4. specific to CoInterpreterMT"
	"flagInterpretedMethods := headerFlags anyMask: 8. specific to CoInterpreter"
	preemptionYields := headerFlags noMask: 16.
	"noThreadingOfGUIThread := headerFlags anyMask: 32. specific to CoInterpreterMT"
	newFinalization := headerFlags anyMask: 64.
	sendWheelEvents := headerFlags anyMask: 128.
	primitiveDoMixedArithmetic := headerFlags noMask: 256! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 4/10/2020 15:29'!
setInterruptCheckChain: aFunction
	<returnTypeC: #'void *'>
	<var: #aFunction declareC: 'void (*aFunction)()'>
	<var: #prevFunction declareC: 'void (*prevFunction)()'>
	<export: true>
	| prevFunction |
	prevFunction := interruptCheckChain.
	interruptCheckChain := aFunction.
	^prevFunction! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/16/2002 13:46'!
setInterruptKeycode: value
	interruptKeycode := value! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 8/22/2008 10:50'!
setInterruptPending: value
	self forceInterruptCheck.
	interruptPending := value! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 2/3/2010 13:02'!
setMaxExtSemSizeTo: maxExtSemTabSize
	maxExtSemTabSizeSet := true.
	self ioSetMaxExtSemTableSize: maxExtSemTabSize! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 2/20/2012 11:37'!
setMethod: aMethodObj
	"Set the method and determine the bytecode set based on the method header's sign.
	 If MULTIPLEBYTECODESETS then a negative header selects the alternate bytecode set.
	 Conditionalizing the code on MULTIPLEBYTECODESETS allows the header sign bit to be
	 used for other experiments."
	<inline: true>
	method := aMethodObj.
	self assert: (objectMemory isOopCompiledMethod: method).
	self cppIf: MULTIPLEBYTECODESETS
		ifTrue: [bytecodeSetSelector := (self methodUsesAlternateBytecodeSet: method)
											ifTrue: [256]
											ifFalse: [0]]! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 6/7/2016 14:46'!
setMethod: aMethodObj methodHeader: methodHeader
	"Set the method and determine the bytecode set based on the method header's sign.
	 If MULTIPLEBYTECODESETS then a negative header selects the alternate bytecode set.
	 Conditionalizing the code on MULTIPLEBYTECODESETS allows the header sign bit to be
	 used for other experiments."
	<inline: true>
	method := aMethodObj.
	self assert: (objectMemory isOopCompiledMethod: method).
	self assert: (objectMemory methodHeaderOf: method) = methodHeader.
	self cppIf: MULTIPLEBYTECODESETS
		ifTrue: [bytecodeSetSelector := (objectMemory headerIndicatesAlternateBytecodeSet: methodHeader)
											ifTrue: [256]
											ifFalse: [0]]! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 08/08/2014 10:01'!
setMethodClassAssociationOf: methodPointer to: anObject
	objectMemory
		storePointer: (objectMemory literalCountOf: methodPointer) + LiteralStart - 1
		ofObject: methodPointer
		withValue: anObject! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 8/21/2009 12:36'!
setNextWakeupUsecs: value
	<api>
	<var: #value type: #usqLong>
	nextWakeupUsecs := value! !

!StackInterpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/16/2002 13:47'!
setSavedWindowSize: value
	savedWindowSize := value! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 8/26/2010 15:52'!
setSignalLowSpaceFlagAndSaveProcess
	"The low space semaphore is about to be signaled. Set the signalLowSpace flag,
	 and force an interrupt check.  Save the currently active process in the special
	 objects array so that the low space handler will be able to determine the process
	 that first triggered a low space condition. The image's low space handler is expected
	 to nil out the special objects array slot when it handles the low space condition."

	| lastSavedProcess activeProc |
	<inline: false>
	DumpStackOnLowSpace ~= 0 ifTrue:
		[self printCallStack.
		 self printAllStacks].
	objectMemory signalLowSpace: true.
	objectMemory lowSpaceThreshold: 0. "disable additional interrupts until lowSpaceThreshold is reset by image"
	lastSavedProcess := objectMemory splObj: ProcessSignalingLowSpace.
	lastSavedProcess = objectMemory nilObject ifTrue:
		[activeProc := self activeProcess.
		objectMemory splObj: ProcessSignalingLowSpace put: activeProc].
	self forceInterruptCheck! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 12/15/2015 09:41'!
setStackPageAndLimit: thePage
	"Set stackPage to a different page.  Set stackLimit unless it has
	 been smashed.  Make the stackPage the most recently used"
	<inline: true>
	<var: #thePage type: #'StackPage *'>
	self assert: thePage ~= 0.
	stackPage := thePage.
	stackLimit ~= self allOnesAsCharStar ifTrue:
		[stackLimit := stackPage stackLimit].
	stackPages markStackPageMostRecentlyUsed: thePage! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 3/10/2017 12:45'!
setStackPointersFromPage: thePage
	<var: #thePage type: #'StackPage *'>
	<inline: true>
	stackPointer := thePage headSP.
	framePointer := thePage headFP! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 12/31/2014 13:51'!
setTraceFlagOnContextsFramesPageIfNeeded: aContext
	| thePage |
	<inline: false>
	<var: #thePage type: #'StackPage *'>
	(self isStillMarriedContextDuringGC: aContext) ifTrue:
		[thePage := stackPages stackPageFor: (self frameOfMarriedContext: aContext).
		 self assert: (thePage trace between: StackPageUnreached and: StackPageTraced).
		 thePage trace = StackPageUnreached ifTrue:
			[thePage trace: StackPageReachedButUntraced]]! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 10/23/2018 10:48'!
setupNativeStack
	<option: #LowcodeVM>
	<var: #theNativeStackMemory type: #'char*'>
	<var: #theShadowCallStackMemory type: #'char*'>
	<inline: false>
	"This initializes an alternate stack that is used by the Lowcode instructions"
	| nativeStackMemorySize theNativeStackMemory shadowCallStackMemorySize theShadowCallStackMemory|
	nativeStackMemorySize := self nativeStackSize.
	theNativeStackMemory := self malloc: nativeStackMemorySize.
	nativeStackPointer := theNativeStackMemory + nativeStackMemorySize.
	
	shadowCallStackMemorySize := self shadowCallStackSize.
	theShadowCallStackMemory := self malloc: shadowCallStackMemorySize.
	shadowCallStackPointer := theShadowCallStackMemory + shadowCallStackMemorySize! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 9/7/2016 14:03'!
shadowCallStackPointerIn: theFP
	"See StackInterpreter class>>initializeFrameIndices"
	<option: #LowcodeVM>
	<inline: false>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	^ stackPages pointerAtPointer: theFP + FoxReceiver - (BytesPerWord*5)! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 9/7/2016 13:56'!
shadowCallStackPointerIn: theFP put: valueOopPointer
	"See StackInterpreter class>>initializeFrameIndices"
	<option: #LowcodeVM>
	<inline: false>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	<var: #valueOopPointer type: #'char*'>
	^ stackPages pointerAtPointer: theFP + FoxReceiver - (BytesPerWord*5) put: valueOopPointer! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'RonieSalgado 4/27/2016 20:42'!
shadowCallStackSize
	"1 MB shadow call stack size"
	^ 1 *1024 * 1024! !

!StackInterpreter methodsFor: 'jump bytecodes' stamp: 'eem 10/27/2012 15:17'!
shortConditionalJumpFalse

	self jumplfFalseBy: (currentBytecode bitAnd: 7) + 1! !

!StackInterpreter methodsFor: 'jump bytecodes' stamp: 'eem 10/27/2012 15:17'!
shortConditionalJumpTrue

	self jumplfTrueBy: (currentBytecode bitAnd: 7) + 1! !

!StackInterpreter methodsFor: 'simulation' stamp: 'eem 2/12/2017 16:55'!
shortPrint: oop
	<doNotGenerate>
	| name classOop key |
	(objectMemory isImmediate: oop) ifTrue:
		[(objectMemory isImmediateCharacter: oop) ifTrue:
			[^(objectMemory characterValueOf: oop) < 256
				ifTrue:
					['=$', (objectMemory characterValueOf: oop) printString,
					' ($', (String with: (Character value: (objectMemory characterValueOf: oop))), ')']
				ifFalse:
					['=$', (objectMemory characterValueOf: oop) printString, '($???)']].
		(objectMemory isIntegerObject: oop) ifTrue:
			[^'=', (objectMemory integerValueOf: oop) printString,
			' (', (objectMemory integerValueOf: oop) hex, ')'].
		(objectMemory isImmediateFloat: oop) ifTrue:
			[^ '=', (objectMemory floatValueOf: oop) printString, ' (', oop hex, ')'].
		^'= UNKNOWN IMMEDIATE', ' (', (objectMemory integerValueOf: oop) hex, ')'].
	(objectMemory addressCouldBeObj: oop) ifFalse:
		[^(oop bitAnd: objectMemory allocationUnit - 1) ~= 0
			ifTrue: [' is misaligned']
			ifFalse: [self whereIs: oop]].
	(objectMemory isFreeObject: oop) ifTrue:
		[^' is a free chunk of size ', (objectMemory sizeOfFree: oop) printString,
			(objectMemory hasSpurMemoryManagerAPI
				ifTrue: [' 0th: ', (objectMemory fetchPointer: 0 ofFreeChunk: oop) hex]
				ifFalse: [''])].
	(objectMemory isForwarded: oop) ifTrue:
		[^' is a forwarded object to ', (objectMemory followForwarded: oop) hex,
			' of slot size ', (objectMemory numSlotsOfAny: oop) printString].
	(objectMemory isFloatInstance: oop) ifTrue:
		[^'=', (objectMemory dbgFloatValueOf: oop) printString].
	oop = objectMemory nilObject ifTrue:
		[^'nil'].
	oop = objectMemory falseObject ifTrue:
		[^'false'].
	oop = objectMemory trueObject ifTrue:
		[^'true'].

	classOop := objectMemory fetchClassOfNonImm: oop.
	((self objCouldBeClassObj: oop)
	 and: [(objectMemory numSlotsOf: classOop) = metaclassNumSlots]) ifTrue:
		[^'class ', (self nameOfClass: oop)].
	name := self nameOfClass: classOop.
	name size = 0 ifTrue: [name := '??'].
	(#('String'  'ByteString') includes: name) ifTrue:
		[^(self stringOf: oop) printString].
	(#('Symbol'  'ByteSymbol') includes: name) ifTrue:
		[^'#', (self stringOf: oop)].
	name = 'Character' ifTrue: "SpurMemoryManager has immediate Characters (see above); ObjectMemory does not"
		[^'=', (Character value: (objectMemory integerValueOf: 
				(objectMemory fetchPointer: 0 ofObject: oop))) printString].

	"Try to spot association-like things; they're all subclasses of LookupKey"
	((objectMemory isPointersNonImm: oop)
	 and: [classOop ~= objectMemory nilObject
	 and: [((objectMemory instanceSizeOf: classOop) between: ValueIndex + 1 and: ValueIndex + 2)
	 and: [(objectMemory addressCouldBeObj: (key := objectMemory fetchPointer: KeyIndex ofObject: oop))
	 and: [(key = objectMemory nilObject and: [self addressCouldBeClassObj: (objectMemory fetchPointer: ValueIndex ofObject: oop)])
		or: [objectMemory isBytesNonImm: key]]]]]) ifTrue:
		[| classLookupKey |
		 classLookupKey := objectMemory fetchClassOfNonImm: (objectMemory splObj: SchedulerAssociation).
		 [classLookupKey = objectMemory nilObject ifTrue:
			[^(('AEIOU' includes: name first) ifTrue: ['an '] ifFalse: ['a ']), name].
		  (objectMemory instanceSizeOf: classLookupKey) = (KeyIndex + 1)] whileFalse:
			[classLookupKey := self superclassOf: classLookupKey].
		[(objectMemory instanceSizeOf: (self superclassOf: classLookupKey)) = (KeyIndex + 1)] whileTrue:
			[classLookupKey := self superclassOf: classLookupKey].
		 (self includesBehavior: classOop ThatOf: classLookupKey) ifTrue:
			[^(('AEIOU' includes: name first) ifTrue: ['an '] ifFalse: ['a ']), name,
				' ', (self shortPrint: (objectMemory fetchPointer: KeyIndex ofObject: oop)),
				' -> ',
				(objectMemory fetchPointer: ValueIndex ofObject: oop) hex8]].

	^(('AEIOU' includes: name first) ifTrue: ['an '] ifFalse: ['a ']), name! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'nice 11/11/2016 18:04'!
shortPrintContext: aContext
	| theFP |
	<inline: false>
	<var: #theFP type: #'char *'>
	(objectMemory isContext: aContext) ifFalse:
		[self printHex: aContext; print: ' is not a context'; cr.
		^nil].
	self printHex: aContext.
	(self isMarriedOrWidowedContext: aContext)
		ifTrue: [(self checkIsStillMarriedContext: aContext currentFP: framePointer)
					ifTrue:
						[theFP := self frameOfMarriedContext: aContext.
						(self isMachineCodeFrame: theFP)
							ifTrue: [self print: ' M (']
							ifFalse: [self print: ' I ('].
						 self printHex: theFP asUnsignedIntegerPtr; print: ') ']
					ifFalse:
						[self print: ' w ']]
		ifFalse: [self print: ' s '].
	(self findHomeForContext: aContext)
		ifNil: [self print: ' BOGUS CONTEXT (can''t determine home)']
		ifNotNil:
			[:home|
			 self printActivationNameFor: (objectMemory
											fetchPointer: MethodIndex
											ofObject: (home ifNil: [aContext]))
				receiver: (home
							ifNil: [objectMemory nilObject]
							ifNotNil: [objectMemory fetchPointer: ReceiverIndex ofObject: home])
				isBlock: home ~= aContext
				firstTemporary: (objectMemory fetchPointer: 0 + CtxtTempFrameStart ofObject: home)].
	self cr! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 11/14/2011 17:47'!
shortPrintFrame: theFP
	<inline: false>
	<var: #theFP type: #'char *'>
	| rcvr |
	(stackPages couldBeFramePointer: theFP) ifFalse:
		[self print: 'invalid frame pointer'; cr.
		 ^nil].
	rcvr := self frameReceiver: theFP.
	self printHexPtr: theFP.
	self space.
	self printActivationNameFor: (self frameMethod: theFP)
		receiver: rcvr
		isBlock: (self frameIsBlockActivation: theFP)
		firstTemporary: (self temporary: 0 in: theFP).
	self space.
	self shortPrintOop: rcvr "shortPrintOop: adds a cr"! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:46'!
shortPrintFrame: theFP AndNCallers: n
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	<inline: false>
	<var: #theFP type: #'char *'>
	(n ~= 0 and: [stackPages couldBeFramePointer: theFP]) ifTrue:
		[self shortPrintFrame: theFP.
		 self shortPrintFrame: (self frameCallerFP: theFP) AndNCallers: n - 1]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 9/12/2013 03:34'!
shortPrintFrameAndCallers: theFP
	<inline: false>
	<var: #theFP type: #'char *'>
	(stackPages couldBeFramePointer: theFP) ifFalse: [^nil].
	self shortPrintFrame: theFP.
	(self isBaseFrame: theFP) ifFalse:
		[self shortPrintFrameAndCallers: (self frameCallerFP: theFP)]! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem (auto pragmas 12/08) 7/4/2008 14:51'!
shortPrintFramesInCurrentPage
	<inline: false>
	self shortPrintFrameAndCallers: localFP! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 3/3/2012 10:19'!
shortPrintFramesInPage: thePage
	<export: true> "use export: not api, so it won't be written to cointerp.h. cogit.c is unaware of StackPage"
	<inline: false>
	<var: #thePage type: #'StackPage *'>
	self printFrameAndCallers: thePage headFP SP: thePage headSP short: true! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 11/8/2013 14:58'!
shortPrintFramesOnStackPageListInUse
	<export: true>
	| page |
	<inline: false>
	<var: #page type: #'StackPage *'>
	page := stackPages mostRecentlyUsedPage.
	[(stackPages isFree: page) ifFalse:
		[self print: 'page '; printHexPtrnp: (self cCode: [page] inSmalltalk: [page baseAddress]); cr.
		 self shortPrintFramesInPage: page.
		 self cr].
	 (page := page prevPage) ~= stackPages mostRecentlyUsedPage] whileTrue! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 11/15/2019 09:53'!
shortPrintOop: oop
	<inline: false>
	self printHexnp: oop.
	(objectMemory isImmediate: oop) ifTrue:
		[(objectMemory isIntegerObject: oop) ifTrue:
			[self
				cCode: ['=%ld\n' printf: {(objectMemory integerValueOf: oop) asLong}]
				inSmalltalk: [self print: (self shortPrint: oop); cr]].
		 (objectMemory isImmediateCharacter: oop) ifTrue:
			[self
				cCode: ['=$%ld ($%lc)\n' printf: {(objectMemory characterValueOf: oop) asLong.
													self cCoerce: (objectMemory characterValueOf: oop) to: 'wint_t'}]
				inSmalltalk: [self print: (self shortPrint: oop); cr]].
		 (objectMemory isImmediateFloat: oop) ifTrue:
			[self
				cCode: ['=%g\n' printf: (objectMemory floatValueOf: oop)]
				inSmalltalk: [self print: '='; printFloat: (objectMemory floatValueOf: oop); cr]].
		 ^self].
	(objectMemory addressCouldBeObj: oop) ifFalse:
		[^self print: ((oop bitAnd: objectMemory allocationUnit - 1) ~= 0
						ifTrue: [' is misaligned']
						ifFalse: [self whereIs: oop]); cr].
	((objectMemory isFreeObject: oop)
	 or: [objectMemory isForwarded: oop]) ifTrue:
		[^self printOop: oop].
	self print: ': a(n) '.
	self printNameOfClass: (objectMemory fetchClassOfNonImm: oop) count: 5.
	self cr! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 5/4/2014 16:02'!
shortReversePrintFrameAndCallers: aFramePointer
	| theFP callerFP caller |
	<inline: false>
	<var: #aFramePointer type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #callerFP type: #'char *'>
	aFramePointer = 0 ifTrue:
		[^objectMemory nilObject].
	theFP := aFramePointer.
	[self shortPrintFrame: theFP.
	 callerFP := self frameCallerFP: theFP.
	 callerFP ~= 0] whileTrue:
		[theFP := callerFP].
	caller := self frameCallerContext: theFP.
	(objectMemory isForwarded: caller) ifTrue:
		[caller := objectMemory followForwarded: caller].
	^caller! !

!StackInterpreter methodsFor: 'jump bytecodes' stamp: 'eem (auto pragmas) 7/8/2008 19:49'!
shortUnconditionalJump
	<expandCases>
	self jump: (currentBytecode bitAnd: 7) + 1.! !

!StackInterpreter methodsFor: 'I/O primitive support' stamp: 'eem 5/12/2017 11:03'!
show: surfaceHandle Sur: left fa: top ce: width Fn: height
	"Simulate the showSurfaceFn function call as a failure to load the surface."
	<doNotGenerate>
	^0! !

!StackInterpreter methodsFor: 'I/O primitive support' stamp: 'eem 5/12/2017 11:05'!
showDisplayBits: aForm Left: l Top: t Right: r Bottom: b
	"Repaint the portion of the Smalltalk screen bounded by the affected rectangle. Used to synchronize the screen after a Bitblt to the Smalltalk Display object."
	(deferDisplayUpdates
	 or: [aForm ~= (objectMemory splObj: TheDisplay)]) ifTrue: [^nil].
	self updateDisplayLeft: l Top: t Right: r Bottom: b! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
shuffleArgumentsAndStoreAbsentReceiver: theReceiver
	"For the absent receiver sends move the arguments up the stack and store the supplied receiver."
	<inline: true>
	localSP := localSP - objectMemory bytesPerOop. "a.k.a. self internalPush: anything"
	1 to: argumentCount do:
		[:i| | oop |
		oop := self internalStackValue: i.
		self internalStackValue: i - 1 put: oop].
	self internalStackValue: argumentCount put: theReceiver! !

!StackInterpreter methodsFor: 'utilities' stamp: ''!
signExtend16: int16
	"Convert a signed 16-bit integer into a signed 32-bit integer value. The integer bit is not added here."

	(int16 bitAnd: 16r8000) = 0
		ifTrue: [ ^ int16 ]
		ifFalse: [ ^ int16 - 16r10000 ].! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 6/29/2014 07:18'!
signalExternalSemaphores
	"Signal all requested semaphores.  Answer if a context switch has occurred."
	| xArray |
	xArray := objectMemory splObj: ExternalObjectsArray.
	^self doSignalExternalSemaphores: (objectMemory numSlotsOf: xArray)! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 11/4/2010 13:40'!
signalFinalization: weakReferenceOop
	"Record the fact that there is at least one object to finalize
	 and cause the VM to check for interrupts at the earliest
	 opportunity. The finalization semaphore will actually be
	 signalled once in checkForEventsMayContextSwitch:"

	self forceInterruptCheck.
	pendingFinalizationSignals := pendingFinalizationSignals + 1! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 11/16/2016 18:41'!
signed32BitIntegerFor: integerValue
	"Answer a full 32 bit integer object for the given integer value.
	 N.B.  Returning in each arm separately enables Slang inlining.
	 /Don't/ return the ifTrue:ifFalse: unless Slang inlining of conditionals is fixed."
	<inline: true>
	objectMemory hasSixtyFourBitImmediates
		ifTrue:
			[^objectMemory integerObjectOf: 
				(self cCode: [self cCoerceSimple: integerValue to: #int]
					inSmalltalk: [(integerValue bitAnd: 16r7FFFFFFF)
								- ((integerValue >> 31 anyMask: 1)
									ifTrue: [-16r100000000]
									ifFalse: [0])])]
		ifFalse:
			[(objectMemory isIntegerValue: integerValue) ifTrue:
				[^objectMemory integerObjectOf: integerValue].
			 ^self noInlineSigned32BitIntegerGutsFor: integerValue]! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 4/28/2016 20:50'!
signed64BitIntegerFor: integerValue
	<var: 'integerValue' type: #sqLong>
	"Answer a Large Integer object for the given integer value.  N.B. will *not* cause a GC."
	| newLargeInteger magnitude largeClass highWord sz |
	<inline: false>
	<var: 'magnitude' type: #usqLong>
	<var: 'highWord' type: #usqInt>

	integerValue < 0
		ifTrue:[	integerValue >= objectMemory minSmallInteger ifTrue: [^objectMemory integerObjectOf: integerValue asInteger].
				largeClass := ClassLargeNegativeIntegerCompactIndex.
				magnitude := 0 - (self cCoerceSimple: integerValue to: #usqLong)]
		ifFalse:[	integerValue <= objectMemory maxSmallInteger ifTrue: [^objectMemory integerObjectOf: integerValue asInteger].
				largeClass := ClassLargePositiveIntegerCompactIndex.
				magnitude := integerValue].

	objectMemory wordSize = 8
		ifTrue: [sz := 8]
		ifFalse: [
		 (highWord := magnitude >> 32) = 0 
			ifTrue: [sz := 4] 
			ifFalse:
				[sz := 5.
				 (highWord := highWord >> 8) = 0 ifFalse:
					[sz := sz + 1.
					 (highWord := highWord >> 8) = 0 ifFalse:
						[sz := sz + 1.
						 (highWord := highWord >> 8) = 0 ifFalse:
							[sz := sz + 1]]]]].

	newLargeInteger := objectMemory
							eeInstantiateSmallClassIndex: largeClass
							format: (objectMemory byteFormatForNumBytes: sz)
							numSlots: sz + 3 // objectMemory bytesPerOop.
	SPURVM
		ifTrue:
			["Memory is eight byte aligned in SPUR, so we are sure to have room for 64bits word whatever allocated sz"
			objectMemory storeLong64: 0 ofObject: newLargeInteger withValue: (objectMemory byteSwapped64IfBigEndian: magnitude)]
		ifFalse:
			[sz > 4
				ifTrue: [objectMemory storeLong64: 0 ofObject: newLargeInteger withValue: (objectMemory byteSwapped64IfBigEndian: magnitude)]
				ifFalse: [objectMemory storeLong32: 0 ofObject: newLargeInteger withValue: (objectMemory byteSwapped32IfBigEndian: magnitude)]].
	^newLargeInteger! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: ''!
singleExtendedSendBytecode
	"Can use any of the first 32 literals for the selector and pass up to 7 arguments."

	| descriptor |
	descriptor := self fetchByte.
	messageSelector := self literal: (descriptor bitAnd: 16r1F).
	argumentCount := descriptor >> 5.
	self normalSend.! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: ''!
singleExtendedSuperBytecode
	"Can use any of the first 32 literals for the selector and pass up to 7 arguments."

	| descriptor |
	descriptor := self fetchByte.
	messageSelector := self literal: (descriptor bitAnd: 16r1F).
	argumentCount := descriptor >> 5.
	self superclassSend.
! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 10/24/2017 15:54'!
sistaInlinePrimitive: prim
	"SistaV1:	248		11111000 	iiiiiiii		mjjjjjjj		Call Primitive #iiiiiiii + (jjjjjjj * 256) m=1 means inlined primitive, no hard return after execution. This is the dispatch for unmapped sista inlined primitives."
	<option: #SistaVM>
	
	 prim < 1000 ifTrue:
		[^self nullaryInlinePrimitive: prim].

	 prim < 2000 ifTrue:
		[^self unaryInlinePrimitive: prim - 1000].

	 prim < 3000 ifTrue:
		[^self binaryInlinePrimitive: prim - 2000].

	 prim < 4000 ifTrue:
		[^self trinaryInlinePrimitive: prim - 3000].

	 prim < 5000 ifTrue:
		[^self quaternaryInlinePrimitive: prim - 4000].

	 prim < 6000 ifTrue:
		[^self quinaryInlinePrimitive: prim - 5000].
	
	 prim < 7000 ifTrue:
		[^self jumpUnaryInlinePrimitive: prim - 6000].
	
 	prim < 8000 ifTrue:
		[^self jumpBinaryInlinePrimitive: prim - 7000].
	
	^ self jumpTrinaryInlinePrimitive: prim - 8000.! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 12/11/2017 14:27'!
sistaMappedInlinePrimitive: primIndex
	"SistaV1:	236		11101100	iiiiiiii		callMappedInlinedPrimitive"
	"Number of arguments:
	 0-49 nullary
	 50-99 unary
	 100-149  binary
	 150-199 trinary
	 200-249 variable
	 250-255 mapped jumps"
	"Specification:
	50	EnsureEnoughWords
	literal which is a Smi => ret value is receiver
	150	immCheckPointerAt:put:
	pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	151	immCheckStoreCheckPointerAt:put:
	pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	152	immCheckMaybeContextPointerAt:put:
	pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	153	immCheckMaybeContextStoreCheckPointerAt:put:
	pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	154	immcheckByteAt:put:
	byte object, Smi, 8 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	155	immCheckShortAt:put:
	short object, Smi, 16 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	156	immCheckWordAt:put:
	word object, Smi, 32 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	157	immCheckDoubleWordAt:put:
	double word object, Smi, 64 bits unsigned Smi or LargePositiveInteger => arg2 (1-based, optimised if arg1 is a constant)
	200	directCall
	method to call on top of stack =>  (variable number of parameters)
	250 	backjumpAlwaysInterrupt
	literal which is a Smi"
	<option: #SistaVM>
	(primIndex between: 154 and: 157) ifTrue: [^self mappedImmcheckDataAtPut: primIndex].
	primIndex caseOf: {
		[50]	->	[self mappedEnsureEnoughWords].
		[150]	->	["150	immCheckPointerAt:put:
		 			 pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
					"We could ignore the store check but we don't care in the interpreter"
					self mappedImmcheckStoreCheckPointerAtPut].
		[151]	->	[self mappedImmcheckStoreCheckPointerAtPut].
		[152]	->	["152	immCheckMaybeContextPointerAt:put:
					 pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
					 "We could ignore the store check but we don't care in the interpreter"
					 self mappedImmcheckMaybeContextStoreCheckPointerAtPut].
		[153]	->	[self mappedImmcheckMaybeContextStoreCheckPointerAtPut].
		[200] 	->	[self mappedDirectCall].
		[250] 	->	[self mappedBackjumpAlwaysInterrupt].
	} otherwise: [self unknownInlinePrimitive]
	
! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 12/7/2017 16:31'!
sizeOfCallPrimitiveBytecode: methodHeader
	"Answer the size of the CallPrimitive bytecode that may be used to store a method's primitive."
	"NewsqueakV4:	249	11111001	iiiiiiii	jjjjjjjj	Call Primitive #iiiiiiii + (jjjjjjjj * 256)"
	"SistaV1:		248	11111000 	iiiiiiii	mssjjjjj	Call Primitive #iiiiiiii + (  jjjjj * 256)"
	"V3+Closures:	139	11101111	iiiiiiii	jjjjjjjj	Call Primitive #iiiiiiii + (jjjjjjjj * 256)"
	<api>
	<inline: true>
	^objectMemory hasSpurMemoryManagerAPI
		ifTrue: [3]
		ifFalse:
			[MULTIPLEBYTECODESETS
				ifTrue: [(objectMemory headerIndicatesAlternateBytecodeSet: methodHeader)
							ifTrue: [3]
							ifFalse: [0]]
				ifFalse: [0]]! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 11/14/2012 14:49'!
sizeOfLongStoreTempBytecode: methodHeader
	"234		11101010	i i i i i i i i	Store Temporary Variable #iiiiiiii"
	"129 		10000001 jjkkkkkk 	Store (Receiver Variable, Temporary Location, Illegal, Literal Variable) [jj] #kkkkkk"
	<api>
	^2! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
sizeOfSTArrayFromCPrimitive: cPtr
	"Return the number of indexable fields of the given object. This method is to be called from an automatically generated C primitive. The argument is assumed to be a pointer to the first indexable field of a words or bytes object; the object header starts 4 bytes before that."
	"Note: Only called by translated primitive code."

	| oop |
	<var: #cPtr type: 'void *'>
	oop := (self oopForPointer: cPtr) - objectMemory baseHeaderSize.
	(objectMemory isWordsOrBytes: oop) ifFalse: [
		self primitiveFail.
		^0].
	^objectMemory lengthOf: oop
! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 3/18/2020 18:29'!
slowPrimitiveResponse
	"Invoke a normal (non-quick) primitive.
	 Called under the assumption that primFunctionPointer has been preloaded."
	| nArgs savedFramePointer savedStackPointer |
	<inline: true>
	<var: #savedFramePointer type: #'char *'>
	<var: #savedStackPointer type: #'char *'>
	self assert: (objectMemory isOopForwarded: (self stackValue: argumentCount)) not.
	self assert: objectMemory remapBufferCount = 0.
	FailImbalancedPrimitives ifTrue:
		[nArgs := argumentCount.
		 savedStackPointer := stackPointer.
		 savedFramePointer := framePointer].
	self initPrimCall.
	self dispatchFunctionPointer: primitiveFunctionPointer.
	self assert: (self maybeLeakCheckExternalPrimCall: newMethod).
	self maybeRetryPrimitiveOnFailure.
	(FailImbalancedPrimitives
	and: [self successful
	and: [framePointer = savedFramePointer
	and: [(self isMachineCodeFrame: framePointer) not]]]) ifTrue:"Don't fail if primitive has done something radical, e.g. perform:"
		[stackPointer ~= (savedStackPointer + (nArgs * objectMemory wordSize)) ifTrue:
			[self flag: 'Would be nice to make this a message send of e.g. unbalancedPrimitive to the current process or context'.
			 "This is necessary but insufficient; the result may still have been written to the stack.
			   At least we'll know something is wrong."
			 self failUnbalancedPrimitive.
			 stackPointer := savedStackPointer]].
	"If we are profiling, take accurate primitive measures"
	nextProfileTick > 0 ifTrue:
		[self checkProfileTick: newMethod].
	^self successful! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 8/23/2018 11:15'!
snapshot: embedded 
	"update state of active context"
	| activeContext activeProc rcvr setMacType stackIndex |
	<var: #setMacType type: #'void *'>

	"For now the stack munging below doesn't deal with more than one argument.
	 It can, and should."
	argumentCount ~= 0 ifTrue:
		[^self primitiveFailFor: PrimErrBadNumArgs].

	"Need to convert all frames into contexts since the snapshot file only holds objects."
	self push: instructionPointer.
	activeContext := self voidVMStateForSnapshotFlushingExternalPrimitivesIf: true.

	"update state of active process"
	activeProc := self activeProcess.
	objectMemory
		storePointer: SuspendedContextIndex
		ofObject: activeProc
		withValue: activeContext.

	tempOop := activeContext.
	objectMemory garbageCollectForSnapshot.
	"Nothing moves from here on so it is safe to grab the activeContext again."
	activeContext := tempOop.
	tempOop := 0.

	self successful ifTrue:
		["Without contexts or stacks simulate
			rcvr := self popStack. ''pop rcvr''
			self push: trueObj.
		  to arrange that the snapshot resumes with true.  N.B. stackIndex is one-relative."
		stackIndex := self quickFetchInteger: StackPointerIndex ofObject: activeContext.
		rcvr := objectMemory fetchPointer: stackIndex + CtxtTempFrameStart - 1 ofObject: activeContext.
		objectMemory
			storePointerUnchecked: stackIndex + CtxtTempFrameStart - 1
			ofObject: activeContext
			withValue: objectMemory trueObject.
		"now attempt to write the snapshot file"
		self writeImageFileIO.
		(self successful and: [embedded not]) ifTrue:
			["set Mac file type and creator; this is a noop on other platforms"
			setMacType := self ioLoadFunction: 'setMacFileTypeAndCreator' From: 'FilePlugin'.
			setMacType = 0 ifFalse:
				[self cCode: '((sqInt (*)(char *, char *, char *))setMacType)(imageName, "STim", "FAST")']].
		"Without contexts or stacks simulate
			self pop: 1"
		objectMemory
			storePointerUnchecked: StackPointerIndex
			ofObject: activeContext
			withValue: (objectMemory integerObjectOf: stackIndex - 1)].

	objectMemory postSnapshot.
	self marryContextInNewStackPageAndInitializeInterpreterRegisters: activeContext.
	self successful
		ifTrue: [self push: objectMemory falseObject]
		ifFalse:
			[self push: rcvr.
			 self justActivateNewMethod: true] "no point checking for machine code; there isn't any at this point"! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem (auto pragmas 12/08) 7/8/2008 20:29'!
space
	<inline: true>
	self printChar: $ ! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 8/21/2013 12:57'!
specialSelector: index
	<api>
	^objectMemory fetchPointer: (index * 2) ofObject: (objectMemory splObj: SpecialSelectors)! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 10/5/2012 11:12'!
specialSelectorNumArgs: index
	^objectMemory integerValueOf: (objectMemory fetchPointer: (index * 2) + 1 ofObject: (objectMemory splObj: SpecialSelectors))! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 09/28/2015 18:49'!
spurPostBecomeAction: theBecomeEffectsFlags
	"Insulate the stack zone from the effects of a become.
	 All receivers must be unfollowed for two reasons:
		1. inst var access is direct with no read barrier
		2. super sends (always to the receiver) have no class check and so don't trap
		   for forwarded receivers.  This is an issue for primitives that assume their receiver
		   is valid and don't validate.
	 Super sends require an explicit check to ensure receivers in super sends are unforwarded.
	 e.g. super doSomethingWith: (self become: other) forwards the receiver self pushed on the
	 stack.  So we could avoid following non-pointer receivers.  But this is too tricky,  Instead, we
	 always follow receivers.
	 Methods must be unfollowed since bytecode access is direct with no read barrier.
	 But this only needs to be done if the becomeEffectsFlags indicate that a
	 CompiledMethod was becommed.
	 The scheduler state must be followed, but only if the becomeEffectsFlags indicate
	 that a pointer object was becommed."
	<option: #SpurObjectMemory>
	<inline: false> "For VM profiling"
	self flushAtCache.
	theBecomeEffectsFlags ~= 0 ifTrue:
		[(theBecomeEffectsFlags anyMask: BecameCompiledMethodFlag) ifTrue:
			[self followForwardedMethodsInMethodCache.
			 self followForwardedMethodsInMethodZone]. "for CoInterpreter; must precede followForwardingPointersInStackZone:"
		 (theBecomeEffectsFlags anyMask: BecameActiveClassFlag) ifTrue:
			[self flushBecommedClassesInMethodCache.
			 self flushBecommedClassesInMethodZone]. "for CoInterpreter"
		 (theBecomeEffectsFlags anyMask: BecamePointerObjectFlag) ifTrue:
			[self followForwardingPointersInScheduler.
			 self followForwardingPointersInSpecialObjectsArray].
		 (theBecomeEffectsFlags anyMask: BecamePointerObjectFlag + BecameCompiledMethodFlag) ifTrue:
			[self followForwardingPointersInProfileState]].
	self followForwardingPointersInStackZone: theBecomeEffectsFlags! !

!StackInterpreter methodsFor: 'indexing primitive support' stamp: 'eem 9/11/2013 23:18'!
stObject: array at: index
	"Return what ST would return for <obj> at: index."

	| hdr fmt totalLength fixedFields stSize |
	<inline: true>
	hdr := objectMemory baseHeader: array.
	fmt := objectMemory formatOfHeader: hdr.
	totalLength := objectMemory lengthOf: array baseHeader: hdr format: fmt.
	fixedFields := objectMemory fixedFieldsOf: array format: fmt length: totalLength.
	(fmt = objectMemory indexablePointersFormat
	 and: [objectMemory isContextHeader: hdr])
		ifTrue:
			[stSize := self stackPointerForMaybeMarriedContext: array.
			((self oop: index isGreaterThanOrEqualTo: 1)
			 and: [(self oop: index isLessThanOrEqualTo: stSize)
			 and: [self isStillMarriedContext: array]]) ifTrue:
				[^self noInlineTemporary: index - 1 in: (self frameOfMarriedContext: array)]]
		ifFalse: [stSize := totalLength - fixedFields].
	((self oop: index isGreaterThanOrEqualTo: (objectMemory firstValidIndexOfIndexableObject: array withFormat: fmt))
	 and: [self oop: index isLessThanOrEqualTo: stSize]) ifTrue:
		[^self subscript: array with: (index + fixedFields) format: fmt].
	self primitiveFailFor: (fmt <= 1 ifTrue: [PrimErrBadReceiver] ifFalse: [PrimErrBadIndex]).
	^0! !

!StackInterpreter methodsFor: 'indexing primitive support' stamp: 'eem 9/11/2013 23:20'!
stObject: array at: index put: value
	"Do what ST would return for <obj> at: index put: value."
	| hdr fmt totalLength fixedFields stSize |
	<inline: true>
	hdr := objectMemory baseHeader: array.
	fmt := objectMemory formatOfHeader: hdr.
	totalLength := objectMemory lengthOf: array baseHeader: hdr format: fmt.
	fixedFields := objectMemory fixedFieldsOf: array format: fmt length: totalLength.
	(fmt = objectMemory indexablePointersFormat
	 and: [objectMemory isContextHeader: hdr])
		ifTrue:
			[stSize := self stackPointerForMaybeMarriedContext: array.
			((self oop: index isGreaterThanOrEqualTo: 1)
			 and: [(self oop: index isLessThanOrEqualTo: stSize)
			 and: [self isStillMarriedContext: array]]) ifTrue:
				[^self noInlineTemporary: index - 1 in: (self frameOfMarriedContext: array) put: value]]
		ifFalse: [stSize := totalLength - fixedFields].
	((self oop: index isGreaterThanOrEqualTo: (objectMemory firstValidIndexOfIndexableObject: array withFormat: fmt))
	 and: [self oop: index isLessThanOrEqualTo: stSize])
		ifTrue: [self subscript: array with: (index + fixedFields) storing: value format: fmt]
		ifFalse: [self primitiveFailFor: (fmt <= 1 ifTrue: [PrimErrBadReceiver] ifFalse: [PrimErrBadIndex])].
	^value! !

!StackInterpreter methodsFor: 'indexing primitive support' stamp: 'eem 5/27/2021 20:11'!
stSizeOf: oop
	"Answer the number of indexable fields in the given object. (i.e., what Smalltalk would return for <obj> size)."
	"Note: Assume oop is not an immediate!!"

	| hdr fmt totalLength fixedFields |
	<inline: false>
	hdr := objectMemory baseHeader: oop.
	fmt := objectMemory formatOfHeader: hdr.
	totalLength := objectMemory lengthOf: oop baseHeader: hdr format: fmt.
	(objectMemory isPureBitsFormat: fmt) ifTrue:
		[^totalLength].
	fixedFields := objectMemory fixedFieldsOf: oop format: fmt length: totalLength.
	fmt = objectMemory indexablePointersFormat ifTrue:
		[self assert: (objectMemory isContextHeader: hdr) not].
	^totalLength - fixedFields! !

!StackInterpreter methodsFor: 'stack access' stamp: 'eem 11/21/2014 17:15'!
stackFloatValue: offset
	"In the StackInterpreter stacks grow down."
	<returnTypeC: #double>
	^objectMemory floatValueOf: (stackPages longAt: stackPointer + (offset*objectMemory wordSize))! !

!StackInterpreter methodsFor: 'stack access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
stackIntegerValue: offset
	"In the StackInterpreter stacks grow down."
	| integerPointer |
	integerPointer := stackPages longAt: stackPointer + (offset*objectMemory wordSize).
	^self checkedIntegerValueOf: integerPointer! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 1/16/2013 07:15'!
stackLimitBytes
	"Answer the actual stackLimit offset in a page.  Since stackPageByteSize may have chosen to shrink a page
	 this may be less than stackPageFrameBytes, but it should be no more than stackPageFrameBytes."
	^self stackPageFrameBytes min: self stackPageByteSize - self stackLimitOffset - self stackPageHeadroom.! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
stackLimitOffset
	"Answer the amount of slots needed to fit a new frame at the point the stack
	 limit is checked.  A frame looks like this at the point the stack limit is checked:
			stacked receiver/closure
			arg0
			...
			argN
			caller's method ip/base frame's sender context
	fp->	saved fp
			method
			method header fields
			context (uninitialized)
			receiver
			first temp
			...
	sp->	Nth temp
	So the amount of headroom is
		the maximum number of arguments + 1 (for stacked receiver and arguments)
		+ the frame size
		+ the max number of temps.
	 Since a method's number of temps includes its arguments the actual offset is:"
	^(FrameSlots + 64) * objectMemory wordSize! !

!StackInterpreter methodsFor: 'stack access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
stackObjectValue: offset
	"Ensures that the given object is a real object, not a SmallInteger."
	"In the StackInterpreter stacks grow down."
	| oop |
	oop := stackPages longAt: stackPointer + (offset * objectMemory wordSize).
	(objectMemory isImmediate: oop) ifTrue:
		[self primitiveFail. ^ nil].
	^oop! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 5/15/2015 23:48'!
stackPageByteSize
	"Answer a page size that is a power-of-two and contains a useful number of frames.
	 Room for 256 slots for frames gives around 40 frames a page which is a
	 good compromise between overflow rate and latency in divorcing a page."
	<inline: false>
	| pageBytes largeSize smallSize |
	pageBytes := self stackPageFrameBytes + self stackLimitOffset + self stackPageHeadroom.
	(pageBytes bitAnd: pageBytes - 1) = 0 ifTrue: "= 0 => a power of two"
		[^pageBytes].
	"round up or round down; that is the question.  If rounding down reduces
	 the size by no more than 1/8th round down, otherwise roundup."
	largeSize := 1 << pageBytes highBit.
	smallSize := 1 << (pageBytes highBit - 1).
	self assert: (largeSize > pageBytes and: [pageBytes > smallSize]).
	^(pageBytes - smallSize) <= (smallSize / 8)
		ifTrue: [smallSize]
		ifFalse: [largeSize]! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 5/16/2015 00:03'!
stackPageFrameBytes
	"Answer a byte size that accomodates a useful number of frames.  The minimum frame size is
	 7 slots in the StackInterpreter, and 6 slots in the CoInterpreter, and the maximum size is 56 + 7
	 slots in the StackInterpreter and 56 + 8 slots in the CoInterpreter. 256 slots gives from 4 to 36
	 frames in the StackInterpreter and from 4 to 42 in the CoInterpreter. Hence 2048 bytes in 32-bits
	 and 4096 bytes in 64-bits; a compromise between overflow rate and latency in divorcing a page."
	"Defining as a macro simplifies hand editing the C for experiments..."
	<cmacro: '() (256 * BytesPerWord)'>
	^256 * objectMemory wordSize! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 2/14/2009 15:38'!
stackPageHeadroom
	"Return a minimum amount of headroom for each stack page (in bytes).
	 In a JIT the stack has to have room for interrupt handlers which will run on the
	 stack.  In the interpreter we don't actually need any headroom."
	^0! !

!StackInterpreter methodsFor: 'initialization' stamp: 'eem 12/29/2008 09:54'!
stackPagesClass
	<doNotGenerate>
	^VMBIGENDIAN
		ifTrue: [InterpreterStackPagesMSB]
		ifFalse: [InterpreterStackPagesLSB]! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 12/24/2019 21:36'!
stackPointerForFramePointer: theFP
	| thePage frameAbove |
	"c.f. the code in printFrame:"
	<doNotGenerate>
	frameAbove := nil.
	^theFP = self headFramePointer
		ifTrue: [self headStackPointer]
		ifFalse:
			[thePage := stackPages stackPageFor: theFP.
			 (stackPages isFree: thePage) ifTrue:
				[self printHexPtr: theFP; print: ' is on a free page?!!'; cr.
				 ^nil].
			 (thePage ~= stackPage
			  and: [theFP = thePage headFP])
				ifTrue: [thePage headSP]
				ifFalse:
					[frameAbove := self safeFindFrameAbove: theFP
										on: thePage
										startingFrom: ((thePage = stackPage
														and: [self headFramePointer
																between: thePage realStackLimit
																and: thePage baseAddress])
														ifTrue: [self headFramePointer]
														ifFalse: [thePage headFP]).
					 frameAbove ifNotNil:
						[self frameCallerSP: frameAbove]]]
! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
stackPointerForMaybeMarriedContext: aContext
	"Return the stackPointer of a Context or BlockContext."
	| sp |
	<inline: true>
	(self isStillMarriedContext: aContext) ifTrue:
		[sp := self stackPointerIndexForFrame: (self frameOfMarriedContext: aContext).
		 self assert: ReceiverIndex + (objectMemory integerValueOf: sp) < (objectMemory lengthOf: aContext).
		 ^sp].
	^self fetchStackPointerOf: aContext! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 1/30/2009 05:54'!
stackPointerIndex
	"Return the 1-based value of the stack pointer for the current frame.
	 This is what the value of the stackp slot would be in a context object."
	^self stackPointerIndexForFrame: framePointer WithSP: stackPointer! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (auto pragmas 12/08) 7/4/2008 11:30'!
stackPointerIndexForFrame: theFP
	"Return the 0-based index rel to the given frame.
	 (This is what stackPointer used to be before conversion to pointer)"
	"In the StackInterpreter stacks grow down."
	| thePage theSP |
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<var: #theSP type: #'char *'>
	thePage := stackPages stackPageFor: theFP.
	theSP := self findSPOf: theFP on: thePage.
	^self stackPointerIndexForFrame: theFP WithSP: theSP! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:03'!
stackPointerIndexForFrame: theFP WithSP: theSP
	"Return the 1-based index rel to the given frame"
	"In the StackInterpreter stacks grow down."
	^(((theFP + FoxReceiver) - theSP) >> objectMemory shiftForWord) + (self frameNumArgs: theFP)! !

!StackInterpreter methodsFor: 'stack access' stamp: 'nice 7/13/2016 09:58'!
stackPositiveMachineIntegerValue: offset
	<api>
	<returnTypeC: #'usqIntptr_t'>
	"In the StackInterpreter stacks grow down."
	| integerPointer |
	integerPointer := stackPages longAt: stackPointer + (offset*objectMemory wordSize).
	^self positiveMachineIntegerValueOf: integerPointer! !

!StackInterpreter methodsFor: 'stack access' stamp: 'nice 7/22/2016 00:50'!
stackSignedMachineIntegerValue: offset
	<api>
	<returnTypeC: #'sqIntptr_t'>
	"In the StackInterpreter stacks grow down."
	| integerPointer |
	integerPointer := stackPages longAt: stackPointer + (offset*objectMemory wordSize).
	^self signedMachineIntegerValueOf: integerPointer! !

!StackInterpreter methodsFor: 'stack access' stamp: 'eem 8/21/2013 12:57'!
stackTop
	<api>
	^stackPages longAt: stackPointer! !

!StackInterpreter methodsFor: 'stack access' stamp: 'eem 2/3/2009 14:27'!
stackTopPut: aValue

	^stackPages longAtPointer: stackPointer put: aValue! !

!StackInterpreter methodsFor: 'stack access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
stackValue: offset
	<api>
	"In the StackInterpreter stacks grow down."
	^stackPages longAt: stackPointer + (offset*objectMemory wordSize)! !

!StackInterpreter methodsFor: 'stack access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
stackValue: offset put: oop
	"In the StackInterpreter stacks grow down."
	^stackPages
		longAt: stackPointer + (offset*objectMemory wordSize)
		put: oop! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 08/08/2014 10:01'!
startPCOfMethod: aCompiledMethod
	<api>
	"Zero-relative version of CompiledMethod>>startpc."
	^(objectMemory literalCountOf: aCompiledMethod) + LiteralStart * objectMemory bytesPerOop! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 12/8/2016 14:25'!
startPCOfMethodHeader: methodHeader
	"Answer the zero-relative index to the initial byte for a method.
	 Zero-relative version of CompiledMethod>>startpc."
	^(objectMemory literalCountOfMethodHeader: methodHeader) + LiteralStart * objectMemory bytesPerOop! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 1/18/2016 17:51'!
storeAndPopReceiverVariableBytecode
	| rcvr top instVarIndex |
	rcvr := self receiver.
	top := self internalStackTop.
	instVarIndex := currentBytecode bitAnd: 7.
	self internalPop: 1.
	objectMemory
		storePointerImmutabilityCheck: instVarIndex
		ofObject: rcvr
		withValue: top.
	self fetchNextBytecode.! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 5/24/2008 11:01'!
storeAndPopRemoteTempLongBytecode
	self storeRemoteTempLongBytecode.
	self internalPop: 1! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 11/13/2012 10:30'!
storeAndPopTemporaryVariableBytecode
	<expandCases>
	self
		cCode: "this bytecode will be expanded so that refs to currentBytecode below will be constant"
			[self fetchNextBytecode.
			 self temporary: (currentBytecode bitAnd: 7) in: localFP put: self internalStackTop.
			 self internalPop: 1]
		inSmalltalk: "Interpreter version has fetchNextBytecode out of order"
			[self temporary: (currentBytecode bitAnd: 7) in: localFP put: self internalStackTop.
			 self fetchNextBytecode.
			 self internalPop: 1]! !

!StackInterpreter methodsFor: 'utilities' stamp: 'eem (objmem refactor) 6/14/2010 10:15'!
storeInteger: fieldIndex ofObject: objectPointer withValue: integerValue 
	"Note: May be called by translated primitive code."
	(objectMemory isIntegerValue: integerValue)
		ifTrue: [objectMemory storePointerUnchecked: fieldIndex ofObject: objectPointer
					withValue: (objectMemory integerObjectOf: integerValue)]
		ifFalse: [self primitiveFail].
	^nil! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 5/29/2018 08:53'!
storeLiteralVariable: literalIndex withValue: anObject
	| litVar |
	litVar := self literal: literalIndex.
	"push/store/popLiteralVariable all fetch a literal, and either read or write the literal's value field.
	 The fetch of the literal needs an explicit check (otherwise we would have to scan all literals in
	 all methods in the stack zone, and the entire method on return, and global variables are relatively
	 rare; in my work image 8.7% of literals are globals)."

	(objectMemory isForwarded: litVar) ifTrue:
		[litVar := self unfollow: litVar atIndex: literalIndex].
	objectMemory storePointerImmutabilityCheck: ValueIndex ofObject: litVar withValue: anObject! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 3/21/2017 14:41'!
storeMaybeContext: obj receiverVariable: fieldIndex withValue: anOop
	"Must trap accesses to married and widowed contexts.
	 But don't want to check on all inst var accesses.  This
	 method is only used by the long-form bytecodes, evading the cost."
	<inline: true>
	((self isWriteMediatedContextInstVarIndex: fieldIndex)
	and: [(objectMemory isContextNonImm: obj)
	and: [(self isMarriedOrWidowedContext: obj)
		or: [self is: fieldIndex methodAssignmentToContextWithMachineCodePC: anOop]]])
		ifTrue:
			[self instVar: fieldIndex ofContext: obj put: anOop]
		ifFalse:
			[objectMemory storePointerImmutabilityCheck: fieldIndex ofObject: obj withValue: anOop]! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 4/5/2016 17:23'!
storeMaybeContextReceiverVariable: fieldIndex withValue: anObject
	"Must trap accesses to married and widowed contexts.
	 But don't want to check on all inst var accesses.  This
	 method is only used by the long-form bytecodes, evading the cost."
	self storeMaybeContext: self receiver receiverVariable: fieldIndex withValue: anObject
! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 5/30/2018 13:41'!
storeRemoteTemp: index inVectorAt: tempVectorIndex
	| tempVector |
	tempVector := self temporary: tempVectorIndex in: localFP.
	TempVectReadBarrier
		ifTrue: 
			[(objectMemory isForwarded: tempVector) ifTrue:
				[tempVector := self unfollowTempVector: tempVector atIndex: tempVectorIndex in: localFP]].
	objectMemory storePointer: index ofObject: tempVector withValue: self internalStackTop.! !

!StackInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 5/27/2008 13:29'!
storeRemoteTempLongBytecode
	| remoteTempIndex tempVectorIndex |
	remoteTempIndex := self fetchByte.
	tempVectorIndex := self fetchByte.
	self fetchNextBytecode.
	self storeRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 8/14/2015 12:17'!
storeSenderOfFrame: theFP withValue: anOop
	"Set the sender of a frame.  If the frame is a base frame then this is trivial;
	 merely store into the FoxCallerSavedIP/FoxCallerContext field.  If not, then
	 split the stack at the frame, moving the frame and those hotter than it to a
	 new stack page.  In the new stack page the frame will be the base frame
	 and storing trivial.  Answer the possibly changed location of theFP."
	| thePage onCurrentPage newPage theMovedFP |
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<var: #newPage type: #'StackPage *'>
	<var: #theMovedFP type: #'char *'>
	<returnTypeC: #'char *'>
	(self isBaseFrame: theFP) ifTrue:
		[self frameCallerContext: theFP put: anOop.
		 ^theFP].
	self ensureCallerContext: theFP.
	thePage := stackPages stackPageFor: theFP.
	self assert: stackPage = stackPages mostRecentlyUsedPage.
	onCurrentPage := thePage = stackPage.
	onCurrentPage ifFalse:
		["Make sure the frame's page isn't divorced when a new page is allocated."
		 stackPages markStackPageNextMostRecentlyUsed: thePage].
	newPage := stackPages newStackPage.
	theMovedFP := self moveFramesIn: thePage through: theFP toPage: newPage.
	onCurrentPage
		ifTrue: [self setStackPageAndLimit: newPage]
		ifFalse: [stackPages markStackPageMostRecentlyUsed: newPage].
	self assert: (self isBaseFrame: theMovedFP).
	self frameCallerContext: theMovedFP put: anOop.
	^theMovedFP! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (objmem refactor) 6/14/2010 10:15'!
storeStackPointerValue: value inContext: aContext
	"Assume: value is an integerValue"
	self assert: ReceiverIndex + value < (objectMemory lengthOf: aContext).
	objectMemory storePointerUnchecked: StackPointerIndex
		ofObject: aContext
		withValue: (objectMemory integerObjectOf: value)! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 1/4/2017 08:51'!
stringOf: oop
	<doNotGenerate>
	| size long nLongs chars |
	^ String streamContents:
		[:strm |
		size := 128 min: (self stSizeOf: oop).
		nLongs := size-1//objectMemory wordSize+1.
		1 to: nLongs do:
			[:i | long := objectMemory longAt: oop + objectMemory baseHeaderSize + (i-1*objectMemory wordSize).
			chars := self charsOfLong: long.
			strm nextPutAll: (i=nLongs
							ifTrue: [chars copyFrom: 1 to: size-1\\objectMemory wordSize+1]
							ifFalse: [chars])]]! !

!StackInterpreter methodsFor: 'indexing primitive support' stamp: 'nice 11/14/2016 22:33'!
subscript: array with: index format: fmt
	"Note: This method assumes that the index is within bounds!!"

	<inline: true>
	fmt <= objectMemory lastPointerFormat ifTrue:
		[^objectMemory fetchPointer: index - 1 ofObject: array].
	fmt >= objectMemory firstByteFormat ifTrue:
		[^objectMemory integerObjectOf:
			(objectMemory fetchByte: index - 1 ofObject: array)].
	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[fmt >= objectMemory firstShortFormat ifTrue:
			[^objectMemory integerObjectOf:
				(objectMemory fetchUnsignedShort16: index - 1 ofObject: array)].
		 fmt = objectMemory sixtyFourBitIndexableFormat ifTrue:
			[^self positive64BitIntegerFor:
				(objectMemory fetchLong64: index - 1 ofObject: array)]].
	"32bit-word type objects; for now assume no 64-bit indexable objects"
	^self positive32BitIntegerFor:
			(objectMemory fetchLong32: index - 1 ofObject: array)! !

!StackInterpreter methodsFor: 'indexing primitive support' stamp: 'eem 10/25/2016 16:43'!
subscript: array with: index storing: oopToStore format: fmt 
	"Note: This method assumes that the index is within bounds!!"
	| signedValueToStore unsignedValueToStore |
	<inline: true>
	fmt <= objectMemory lastPointerFormat ifTrue:
		[objectMemory storePointer: index - 1 ofObject: array withValue: oopToStore.
		 ^self].
	fmt >= objectMemory firstByteFormat ifTrue:
		[(objectMemory isIntegerObject: oopToStore) ifFalse:
			[primFailCode := PrimErrBadArgument.
			 ^self].
		 signedValueToStore := objectMemory integerValueOf: oopToStore.
		 (signedValueToStore >= 0 and: [signedValueToStore <= 255]) ifFalse:
			[primFailCode := PrimErrBadArgument.
			 ^self].
		 objectMemory storeByte: index - 1 ofObject: array withValue: signedValueToStore.
		 ^self].
	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[fmt >= objectMemory firstShortFormat ifTrue:
			[(objectMemory isIntegerObject: oopToStore) ifFalse:
				[primFailCode := PrimErrBadArgument.
				 ^self].
			 signedValueToStore := objectMemory integerValueOf: oopToStore.
			 (signedValueToStore >= 0 and: [signedValueToStore <= 65535]) ifFalse:
				[primFailCode := PrimErrBadArgument.
				 ^self].
			 objectMemory storeShort16: index - 1 ofObject: array withValue: signedValueToStore.
			 ^self].
		 fmt = objectMemory sixtyFourBitIndexableFormat ifTrue:
			[| unsigned64BitValueToStore |
			 unsigned64BitValueToStore := self positive64BitValueOf: oopToStore.
			 self successful ifTrue:
				[objectMemory storeLong64: index - 1 ofObject: array withValue: unsigned64BitValueToStore].
			 ^self]].
	"32bit-word type objects"
	unsignedValueToStore := self positive32BitValueOf: oopToStore.
	self successful ifTrue:
		[objectMemory storeLong32: index - 1 ofObject: array withValue: unsignedValueToStore]! !

!StackInterpreter methodsFor: 'message sending' stamp: 'eem 7/11/2014 12:58'!
superclassOf: classPointer
	"Using a read barrier here simplifies the become implementation and costs very
	 little because the class index and superclass almost certainly share a cache line."
	<inline: true>
	^objectMemory followField: SuperclassIndex ofObject: classPointer! !

!StackInterpreter methodsFor: 'send bytecodes' stamp: 'eem 4/26/2016 16:53'!
superclassSend
	"Send a message to self, starting lookup with the superclass of the class
	 containing the currently executing method."
	"Assume: messageSelector and argumentCount have been set, and that
	 the receiver and arguments have been pushed onto the stack,"
	"Note: This method is inlined into the interpreter dispatch loop."
	<sharedCodeInCase: #singleExtendedSuperBytecode>
	| superclass |
	superclass := self superclassOf: (self methodClassOf: method).
	"classTagForClass: uses ensureBehaviorHash:"
	lkupClassTag := objectMemory classTagForClass: superclass.
	"To maintain the invariant that all receivers are unforwarded we need an explicit
	 read barrier in the super send cases.  Even though we always follow receivers
	 on become  e.g. super doSomethingWith: (self become: other) forwards the receiver
	 self pushed on the stack."
	self ensureReceiverUnforwarded.
	self assert: lkupClassTag ~= objectMemory nilObject.
	self commonSendOrdinary! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 1/24/2017 09:23'!
symbolicMethod: aMethod
	<doNotGenerate>
	 self transcript
		ensureCr;
		nextPutAll:
			((String streamContents:
				[:ts| | prim |
				(prim := self primitiveIndexOf: aMethod) > 0 ifTrue:
					[ts nextPutAll: '<primitive: '; print: prim; nextPut: $>.
					(self isQuickPrimitiveIndex: prim) ifTrue:
						[ts nextPutAll: ' quick method'; cr; flush.
						 ^self].
					ts cr].
				(InstructionPrinter
						on: (VMCompiledMethodProxy new
								for: aMethod
								coInterpreter: self
								objectMemory: objectMemory))
					indent: 0;
					printInstructionsOn: ts]) copyReplaceAll: 'a VMObjectProxy for ' with: '');
		flush! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 1/17/2015 13:54'!
synchronousSignal: aSemaphore 
	"Signal the given semaphore from within the interpreter.
	 Answer if the current process was preempted."
	| excessSignals |
	<inline: false>
	(self isEmptyList: aSemaphore) ifTrue:
		["no process is waiting on this semaphore"
		 excessSignals := self fetchInteger: ExcessSignalsIndex ofObject: aSemaphore.
		 self storeInteger: ExcessSignalsIndex
			ofObject: aSemaphore
			withValue: excessSignals + 1.
		 ^false].

	objectMemory ensureSemaphoreUnforwardedThroughContext: aSemaphore.

	^self resume: (self removeFirstLinkOfList: aSemaphore)
		preemptedYieldingIf: preemptionYields! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem (auto pragmas 12/08) 7/8/2008 20:29'!
tab
	<inline: true>
	self printChar: $	"<-Character tab"! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 08/08/2014 10:42'!
tempCountOf: methodPointer
	<api>
	^self temporaryCountOfMethodHeader: (objectMemory methodHeaderOf: methodPointer)! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
temporary: offset in: theFP
	"See StackInterpreter class>>initializeFrameIndices"
	| frameNumArgs |
	<inline: true>
	<var: #theFP type: #'char *'>
	^offset < (frameNumArgs := self frameNumArgs: theFP)
		ifTrue: [stackPages longAt: theFP + FoxCallerSavedIP + ((frameNumArgs - offset) * objectMemory wordSize)]
		ifFalse: [stackPages longAt: theFP + FoxReceiver - objectMemory wordSize + ((frameNumArgs - offset) * objectMemory wordSize)]! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
temporary: offset in: theFP put: valueOop
	"See StackInterpreter class>>initializeFrameIndices"
	| frameNumArgs |
	<inline: true>
	<var: #theFP type: #'char *'>
	^offset < (frameNumArgs := self frameNumArgs: theFP)
		ifTrue: [stackPages longAt: theFP + FoxCallerSavedIP + ((frameNumArgs - offset) * objectMemory wordSize) put: valueOop]
		ifFalse: [stackPages longAt: theFP + FoxReceiver - objectMemory wordSize + ((frameNumArgs - offset) * objectMemory wordSize) put: valueOop]! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 11/18/2014 17:54'!
temporaryCountOfMethodHeader: header
	<api>
	<inline: true>
	^header >> MethodHeaderTempCountShift bitAnd: 16r3F! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 11/10/2014 12:49'!
temporaryLocation: offset in: theFP numArgs: numArgs
	"Answer the pointer to a given temporary (for debug frame printing in odd circumstances)"
	<var: #theFP type: #'char *'>
	<returnTypeC: #'char *'>
	^offset < numArgs
		ifTrue: [theFP + FoxCallerSavedIP + ((numArgs - offset) * objectMemory wordSize)]
		ifFalse: [theFP + FoxReceiver - objectMemory wordSize + ((numArgs - offset) * objectMemory wordSize)]! !

!StackInterpreter methodsFor: 'simulation' stamp: 'eem 7/14/2011 12:40'!
thisClassIndex
	<doNotGenerate>
	^thisClassIndex! !

!StackInterpreter methodsFor: 'object memory support' stamp: 'eem 8/4/2021 11:13'!
traceProfileState
	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[self followForwardingPointersInProfileState].
	objectMemory markAndTrace: profileProcess.
	objectMemory markAndTrace: profileMethod.
	objectMemory markAndTrace: profileSemaphore! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
transferTo: newProc 
	"Record a process to be awoken on the next interpreter cycle."
	| activeContext sched oldProc |
	<inline: false>
	statProcessSwitch := statProcessSwitch + 1.
	self push: instructionPointer.
	self externalWriteBackHeadFramePointers.
	self assertValidExecutionPointe: instructionPointer + 1 r: framePointer s: stackPointer.
	sched := self schedulerPointer.
	oldProc := objectMemory fetchPointer: ActiveProcessIndex ofObject: sched.
	activeContext := self ensureFrameIsMarried: framePointer SP: stackPointer + objectMemory wordSize.
	objectMemory storePointer: SuspendedContextIndex ofObject: oldProc withValue: activeContext.
	objectMemory storePointer: ActiveProcessIndex ofObject: sched withValue: newProc.
	objectMemory storePointerUnchecked: MyListIndex ofObject: newProc withValue: objectMemory nilObject.
	self externalSetStackPageAndPointersForSuspendedContextOfProcess: newProc! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 12/13/2017 15:19'!
trinaryAtPutInlinePrimitive: primIndex
	<option: #SistaVM>
	| result rec argIntAdjusted arg1|
	arg1 := self internalStackValue: 1.
	rec := self internalStackValue: 2.
	self deny: ((objectMemory isOopForwarded: rec) or: [(objectMemory isImmediate: rec)]).
	self assert: (objectMemory isIntegerObject: arg1).
	argIntAdjusted := (objectMemory integerValueOf: arg1) - 1.
	self assert: argIntAdjusted >= 0.
	result := self internalStackTop.
	
	primIndex caseOf: {
		"3000	pointerAt:put:
		 Mutable pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
		[0]	->	[self assert: (objectMemory isPointers: rec).
				 self assert: argIntAdjusted < (objectMemory numSlotsOfAny: rec).
				 objectMemory
					storePointerUnchecked: argIntAdjusted
					ofObject: rec
					withValue: result.].
		"3001	storeCheckPointerAt:put:
		 Mutable pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
		[1]	->	[self assert: (objectMemory isPointers: rec).
				 self assert: argIntAdjusted < (objectMemory numSlotsOfAny: rec).
				 objectMemory
					storePointer: argIntAdjusted
					ofObject: rec
					withValue: result].
		"3002	maybeContextPointerAt:put:
		 Mutable pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
		[2]	->	[((objectMemory isContextNonImm: rec) 
				 and: [self isMarriedOrWidowedContext: rec])
					ifTrue:
						[self externalizeIPandSP.
						 self externalInstVar: argIntAdjusted ofContext: rec  put: result.
						 self internalizeIPandSP]
					ifFalse: [objectMemory storePointer: argIntAdjusted ofObject: rec withValue: result]].
		"3003	maybeContextStoreCheckPointerAt:put:
		Mutable pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)"
		[3]	->	[((objectMemory isContextNonImm: rec) 
				 and: [self isMarriedOrWidowedContext: rec])
					ifTrue:
						[self externalizeIPandSP.
						 self externalInstVar: argIntAdjusted ofContext: rec  put: result.
						 self internalizeIPandSP]
					ifFalse: [objectMemory storePointer: argIntAdjusted ofObject: rec withValue: result]].
		"3004	byteAt:put:
		Mutable byte object, Smi, 8 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)"
		[4]	->	[self assert: (objectMemory isBytes: rec).
				 self assert: argIntAdjusted < (objectMemory numBytesOf: rec).
				 self assert: (objectMemory isIntegerObject: result).
				 objectMemory
					storeByte: argIntAdjusted
					ofObject: rec
					withValue: (objectMemory integerValueOf: result).].
		"3005	shortAt:put:
		 Mutable short object, Smi, 16 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)"
		[5]	->	[self assert: (objectMemory isShorts: rec).
				 self assert: argIntAdjusted < (objectMemory num16BitUnitsOf: rec).
				 self assert: (objectMemory isIntegerObject: result).
				 objectMemory
					storeShort16: argIntAdjusted
					ofObject: rec
					withValue: (objectMemory integerValueOf: result).].
		"3006	wordAt:put:
		 Mutable word object, Smi, 32 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)"
		[6]	->	[self assert: (objectMemory isWords: rec).
				 self assert: argIntAdjusted < (objectMemory num32BitUnitsOf: rec).
				 objectMemory
					storeLong32: argIntAdjusted
					ofObject: rec
					withValue: (objectMemory positive32BitValueOf: result).].
		"3007	doubleWordAt:put:
		 Mutable double word object, Smi, 64 bits unsigned Smi or LargePositiveInteger => arg2 (1-based, optimised if arg1 is a constant)"
		[7]	->	[self assert: (objectMemory isLong64s: rec).
				 self assert: argIntAdjusted < (objectMemory num64BitUnitsOf: rec).
				 objectMemory
					storeLong64: argIntAdjusted
					ofObject: rec
					withValue: (objectMemory positive64BitValueOf: result).].
	}.
	 self internalPop: 2; internalStackTopPut: result! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 11/10/2017 11:12'!
trinaryInlinePrimitive: primIndex
	<option: #SistaVM>
	"Bulk comment, each sub method has its own comment
	3000	pointerAt:put:
	Mutable pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	3001	storeCheckPointerAt:put:
	Mutable pointer object (Fixed sized or not) and not a context, Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	3002	maybeContextPointerAt:put:
	Mutable pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	3003	maybeContextStoreCheckPointerAt:put:
	Mutable pointer object (Fixed sized or not), Smi, Anything => arg2 (1-based, optimised if arg1 is a constant)
	3004	byteAt:put:
	Mutable byte object, Smi, 8 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	3005	shortAt:put:
	Mutable short object, Smi, 16 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	3006	wordAt:put:
	Mutable word object, Smi, 32 bits unsigned Smi => arg2 (1-based, optimised if arg1 is a constant)
	3007	doubleWordAt:put:
	Mutable double word object, Smi, 64 bits unsigned Smi or LargePositiveInteger => arg2 (1-based, optimised if arg1 is a constant)
	3021 is deprecated."
	primIndex <= 7 ifTrue: [^self trinaryAtPutInlinePrimitive: primIndex].
	"21 is deprecated, now this can be handled at Scorch level"
	primIndex = 21 ifTrue: 
		[ | str1 str2 word1 word2 len |
		  len := objectMemory integerValueOf: self internalStackTop.
		  len = 0 ifTrue: [^self internalPop: 2; internalStackTopPut: objectMemory trueObject].
		  str1 := self internalStackValue: 2.
  					  str2 := self internalStackValue: 1.
		  0 to: len - 1 >> objectMemory shiftForWord do: [:i |
			word1 := objectMemory fetchPointer: i ofObject: str1.
			word2 := objectMemory fetchPointer: i ofObject: str2. 
			word1 = word2 ifFalse: [^self internalPop: 2; internalStackTopPut: objectMemory falseObject] ].
		 ^self internalPop: 2; internalStackTopPut: objectMemory trueObject ].
	self unknownInlinePrimitive! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 7/22/2021 19:08'!
tryLoadNewPlugin: pluginString pluginEntries: pluginEntries
	"Load the plugin and if on Spur, populate pluginEntries with the prmitives in the plugin."
	<doNotGenerate>
	| plugin realPluginClass plugins |
	self transcript cr; show: 'Looking for module ', pluginString.
	pluginString isEmpty
		ifTrue:
			[plugin := self]
		ifFalse:
			[plugins := InterpreterPlugin allSubclasses select:
							[:psc|
							 psc moduleName asString = pluginString asString
							 and: [psc shouldBeTranslated]].
			 plugins isEmpty ifTrue: [self transcript show: ' ... not found'. ^nil].
			 plugins size > 1 ifTrue: [^self error: 'This won''t work...'].
			 "plugins size > 1 ifTrue:
				[self transcript show: '...multiple plugin classes; choosing ', plugins last name]."
			 realPluginClass := plugins anyOne. "hopefully lowest in the hierarchy..."
			 plugin := realPluginClass simulatorForInterpreterInterface: objectMemory.
			 plugin ifNil: [self transcript show: ' ... no simulator class; cannot simulate'. ^nil].
			 (plugin respondsTo: #initialiseModule) ifTrue:
				[plugin initialiseModule ifFalse:
					[self transcript show: ' ... initialiser failed'. ^nil]]]. "module initialiser failed"
	self transcript show: ' ... loaded'.
	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[| realPlugin cg |
		 self transcript show: '...computing accessor depths'.
		 plugin class isPluginClass
			ifTrue:
				[realPlugin := (plugin isSmartSyntaxPluginSimulator
									ifTrue: [realPluginClass]
									ifFalse: [plugin class])
								 withAllSuperclasses detect: [:class| class shouldBeTranslated].
				 cg := realPlugin buildCodeGenerator]
			ifFalse:
				[cg := self codeGeneratorToComputeAccessorDepth.
				 primitiveTable withIndexDo:
					[:prim :index| | depth |
					 prim isSymbol ifTrue:
						[depth := cg accessorDepthForSelector: prim.
						 self assert: (depth isInteger or: [depth isNil and: [(plugin class whichClassIncludesSelector: prim) isNil]]).
						 primitiveAccessorDepthTable at: index - 1 put: depth]]].
		 cg exportedPrimitiveNames do:
			[:primName| | fnSymbol |
			 fnSymbol := primName asSymbol.
			 pluginEntries addLast: {plugin.
									fnSymbol.
									[plugin perform: fnSymbol. self].
									self metadataFlagsFor: fnSymbol using: cg}].
		 self transcript show: '...done'].
	^pluginString asString -> plugin! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 9/6/2016 09:59'!
ultimateLiteralOf: aMethodOop
	<api>
	self assert: (objectMemory isOopCompiledMethod: aMethodOop).
	^self literal: (objectMemory literalCountOf: aMethodOop) - 1 ofMethod: aMethodOop! !

!StackInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
unPop: nItems
	"In the StackInterpreter stacks grow down."
	stackPointer := stackPointer - (nItems*objectMemory wordSize)! !

!StackInterpreter methodsFor: 'sista inline primitives - unary' stamp: 'cb 12/11/2017 14:11'!
unaryClassPrimitive
	| result top |
	"1000	rawClass
	 not a forwarder => Behavior (Same as class special send, but receiver is not a forwarder)"
	 top := self internalStackTop.
	 self deny: (objectMemory isOopForwarded: top).
	 result := objectMemory fetchClassOf: top.
	 self internalStackTopPut: result! !

!StackInterpreter methodsFor: 'sista inline primitives - unary' stamp: 'cb 12/11/2017 14:12'!
unaryConvertInlinePrimitive: primIndex
	<option: #SistaVM>
	| result top |
	self assert: (primIndex between: 30 and: 32).
	top := self internalStackTop.
	primIndex caseOf: {
		"1030	characterAsInteger
		 Character => 22 bits strictly positive Smi (Unicode)"
		[30] -> [self assert: (objectMemory isImmediateCharacter: top).
				result := objectMemory characterValueOf: top.
				self assert: (result between: 1 and: "1 << 22 - 1" 4194303).
				self internalStackTopPutIntegerObjectOf: result].
		"1031	smallFloatAsInteger
		 SmallFloat => Smi"
		[31] -> [objectMemory hasSixtyFourBitImmediates
					ifTrue: "Needs to protect rotatedFloatBitsOf:"
						[self assert: (objectMemory isImmediateFloat: top).
						result := objectMemory rotatedFloatBitsOf: top.
						self assert: (objectMemory isIntegerObject: result).
						self internalStackTopPutIntegerObjectOf: result]
					ifFalse: [self unknownInlinePrimitive]].
		"1032	smiAsFloat
		 Smi => SmallFloat"
		[32] -> [self assert: (objectMemory isIntegerObject: top).
				self pop: 1 thenPushFloat: (objectMemory integerValueOf: top) asFloat].
		 }! !

!StackInterpreter methodsFor: 'sista inline primitives - unary' stamp: 'cb 12/12/2017 10:30'!
unaryHashInlinePrimitive: primIndex
	<option: #SistaVM>
	| result top |
	self assert: (primIndex between: 20 and: 23).
	top := self internalStackTop.
	primIndex caseOf: {
		"1020	objectIdentityHash
		 non-immediate and non-behavior => 22 bits strictly positive Smi"
		[20] ->	[self deny: ((objectMemory isOopForwarded: top)
						or: [(objectMemory isImmediate: top)
							or: [objectMemory isInClassTable: top]]).
				 result := objectMemory hashBitsOf: top.
				self assert: (result between: 1 and: "1 << 22 - 1" 4194303).
				self internalStackTopPutIntegerObjectOf: result].
		"1021	smiIdentityHash
		 Smi => Smi"
		[21] -> [self assert: (objectMemory isIntegerObject: top).
				"Don't do anything, Smi hash is the object itself"].		
		"1022	charIdentityHash
		 Character => 22 bits strictly positive Smi"
		[22] -> [self assert: (objectMemory isImmediateCharacter: top).
				result := objectMemory integerObjectOfCharacterObject: top.
				self assert: (result between: 1 and: "1 << 22 - 1" 4194303).
				self internalStackTopPut: result].
		"1023	smallfloatIdentityHash
		 SmallFloat => Smi"
		[23] -> [objectMemory hasSixtyFourBitImmediates
					ifTrue: "Needs to protect rotatedFloatBitsOf:"
						[self assert: (objectMemory isImmediateFloat: top).
						result := objectMemory rotatedFloatBitsOf: top.
						self assert: (objectMemory isIntegerObject: result).
						self internalStackTopPutIntegerObjectOf: result]
					ifFalse: [self unknownInlinePrimitive]].
		"1024	behaviorIdentityHash
		 Behavior => 22 bits strictly positive Smi"
		[24] -> [self assert: ((objectMemory isNonImmediate: top) 
					and: [objectMemory objCouldBeClassObj: top]).
				result := objectMemory ensureBehaviorHash: top.
				self assert: (result between: 1 and: "1 << 22 - 1" 4194303).
				self internalStackTopPutIntegerObjectOf: result].	
		 }! !

!StackInterpreter methodsFor: 'sista inline primitives - unary' stamp: 'cb 12/13/2017 14:20'!
unaryInlinePrimitive: primIndex
	<option: #SistaVM>
	"Bulk comments: each sub-method has its own comment with the specific case.
	1000	rawClass
	not a forwarder => Behavior (Same as class special send, but receiver is not a forwarder)
	1001	numSlots
	pointer object => Smi between 0 and SmallInteger maxVal // 4 - 1 (Answers total size in pointer-sized slots)
	1002	numBytes
	byte object => Smi between 0 and SmallInteger maxVal - 9 (Includes compiled code)
	1003	numShorts
	short object => Smi between 0 and SmallInteger maxVal - 9
	1004	numWords
	word object => Smi between 0 and SmallInteger maxVal - 9
	1005	numDoubleWords
	double word object => Smi between 0 and SmallInteger maxVal - 9
	1011	RawNew
	literal which is a fixed-sized behavior => instance of the receiver with fields nilled out
	1012	RawNewNoInit
	literal which is a fixed-sized behavior => instance of the receiver (Fields of returned value contain undefined data)
	1020	objectIdentityHash
	non-immediate and non-behavior => 22 bits strictly positive Smi
	1021	smiIdentityHash
	Smi => Smi
	1022	charIdentityHash
	Character => 22 bits strictly positive Smi
	1023	smallfloatIdentityHash
	SmallFloat => Smi
	1024	behaviorIdentityHash
	Behavior => 22 bits strictly positive Smi
	1030	characterAsInteger
	Character => 22 bits strictly positive Smi (Unicode)
	1031	smallFloatAsInteger
	SmallFloat => Smi
	1032	smiAsFloat
	Smi => SmallFloat
	1039	unforwardNonImmediate
	non immediate => Not a forwarder
	1040	unforward
	Anything => Not a forwarder
	1041	possibleRoot
	non-immediate, not a forwarder => receiver is returned (should be effect-only) (If old, becomes gray and remembered to allow many unchecked stores in a row afterwards)"
	primIndex = 0 ifTrue: [^self unaryClassPrimitive].
	primIndex <= 6 ifTrue: [^self unarySizeInlinePrimitive: primIndex].
	primIndex < 11 ifTrue: [^self unknownInlinePrimitive].
	primIndex <= 12 ifTrue: [^self unaryNewInlinePrimitive: primIndex].
	primIndex < 20 ifTrue: [^self unknownInlinePrimitive].
	primIndex <= 24 ifTrue: [^self unaryHashInlinePrimitive: primIndex].
	primIndex < 30 ifTrue: [^self unknownInlinePrimitive].
	primIndex <= 32 ifTrue: [^self unaryConvertInlinePrimitive: primIndex].
	primIndex < 39 ifTrue: [^ self unknownInlinePrimitive].
	primIndex <= 40 ifTrue: [^self unaryUnforwardInlinePrimitive: primIndex].
	primIndex = 41 ifTrue: [^self unaryPossibleRootInlinePrimitive].
	self unknownInlinePrimitive! !

!StackInterpreter methodsFor: 'sista inline primitives - unary' stamp: 'cb 3/1/2018 18:02'!
unaryNewInlinePrimitive: primIndex
	"1011	RawNew
	 literal which is a fixed-sized behavior => instance of the receiver with fields nilled out
	1012	RawNewNoInit
	literal which is a fixed-sized behavior => instance of the receiver (Fields of returned value contain undefined data)
	
	WARNING: In the interpreter version, fields are always initialized."
	| classObj result |
	self assert: (primIndex between: 11 and: 12).
	classObj := self internalStackTop.
	self assert: ((objectMemory isNonImmediate: classObj) and: [self objCouldBeClassObj: classObj]).
	result := objectMemory instantiateClass: classObj.
	self internalStackTopPut: result! !

!StackInterpreter methodsFor: 'sista inline primitives - unary' stamp: 'cb 1/21/2018 21:18'!
unaryPossibleRootInlinePrimitive
	<option: #SistaVM>
	"1041	possibleRoot
	 non-immediate, not a forwarder => receiver is returned (should be effect-only) (If old, becomes gray and remembered to allow many unchecked stores in a row afterwards)"
	| top |
	top := self internalStackTop.
	self deny: (objectMemory isImmediate: top).
	self deny: (objectMemory isOopForwarded: top).
	(objectMemory isYoungObject: top)
		ifFalse: [objectMemory possibleRootStoreInto: top]! !

!StackInterpreter methodsFor: 'sista inline primitives - unary' stamp: 'cb 12/11/2017 14:14'!
unarySizeInlinePrimitive: primIndex
	<option: #SistaVM>
	| result top |
	self assert: (primIndex between: 1 and: 6).
	top := self internalStackTop.
	primIndex caseOf: {
		"1001	numSlots
		 pointer object => Smi between 0 and SmallInteger maxVal // 4 - 1 (Answers total size in pointer-sized slots)"
		[1]	->	[self assert: (objectMemory isPointers: top).
				 result := objectMemory numSlotsOfAny: top.
				 self assert: (result between: 0 and: objectMemory maxSmallInteger // 4 - 1)].
		"1002	numBytes
		 byte object => Smi between 0 and SmallInteger maxVal - 9 (Includes compiled code)"
		[2]	->	[self assert: (objectMemory isBytes: top).
				 result := objectMemory numBytesOf: top].
		"1003	numShorts
		 short object => Smi between 0 and SmallInteger maxVal - 9"
		[3]	->	[self assert: (objectMemory isShorts: top).
				 result := objectMemory num16BitUnitsOf: top].
		"1004	numWords
		 word object => Smi between 0 and SmallInteger maxVal - 9"
		[5]	->	[self assert: (objectMemory isWords: top).
				 result := objectMemory num32BitUnitsOf: top].
		"1005	numDoubleWords 
		 double word object => Smi between 0 and SmallInteger maxVal - 9"
		[6]	->	[self assert: (objectMemory isLong64s: top).
				 result := objectMemory num64BitUnitsOf: top].
	}.
	self assert: (result between: 0 and: objectMemory maxSmallInteger - 9).
	self internalStackTopPut: (objectMemory integerObjectOf: result)
	! !

!StackInterpreter methodsFor: 'sista inline primitives - unary' stamp: 'cb 1/21/2018 21:17'!
unaryUnforwardInlinePrimitive: primIndex
	<option: #SistaVM>
	"1039	unforwardNonImmediate
	 non immediate => Not a forwarder
	 1040	unforward
	 Anything => Not a forwarder"
	"Only perf in jitted code matters, don't remove the immediate check"
	 | top |
	 top := self internalStackTop.
	self assert: (primIndex = 40 or: [objectMemory isNonImmediate: top]).
	 (objectMemory isOopForwarded: top) ifTrue:
		[self internalStackTopPut: (objectMemory followForwarded: top)]! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'eem 10/29/2020 10:36'!
unconditionalTrapBytecode
	"SistaV1: *	217	Trap"
	SistaVM 
		ifTrue: [^self respondToSistaTrap]
		ifFalse: [^self respondToUnknownBytecode]! !

!StackInterpreter methodsFor: 'interpreter shell' stamp: 'eem 6/20/2011 17:33'!
undoFetchNextBytecode
	"Backup the ip when it has been incremented to fetch the next bytecode."

	localIP := localIP - 1! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'eem 4/20/2020 10:42'!
unfollow: litVar atIndex: literalIndex
	<option: #SpurObjectMemory>
	<inline: #never> "So rare it mustn't bulk up the common path"
	| followed |
	followed := objectMemory followForwarded: litVar.
	self literal: literalIndex ofMethod: method put: followed.
	^followed! !

!StackInterpreter methodsFor: 'primitive support' stamp: 'eem 8/3/2021 12:15'!
unfollowFirstLiteralOfMaybeCalloutMethod: methodObj primitiveIndex: primIndex
	"Follow the first literal of either a primitiveCallout or primitiveExternalCall primitive method. This
	 will be an ExternalFunction for primitiveCallout or a four element Array for primitiveExternalCall.
	 This is here to avoid following all the literals in a method, which would be slow.  Remember
	 forwarders are unlikely, so we only want to follow what is necessary, and for an FFI call or
	 external primitive only the first literal is salient."
	<option: #SpurObjectMemory>
	| found firstLiteral |
	found := false.
	"inlined self literal: 0 ofMethod: methodObj for clarity..."
	firstLiteral := objectMemory fetchPointer: LiteralStart ofObject: methodObj.
	(objectMemory isOopForwarded: firstLiteral) ifTrue:
		[found := true.
		 firstLiteral := objectMemory fixFollowedField: LiteralStart ofObject: methodObj withInitialValue: firstLiteral].
	(objectMemory
		followForwardedObjectFields: firstLiteral
		toDepth: (primIndex = PrimNumberFFICall
					ifTrue: [(primitiveAccessorDepthTable at: primIndex) - 1]
					ifFalse: [0])) ifTrue:
		[found := true].
	^found! !

!StackInterpreter methodsFor: 'compiled methods' stamp: 'cb 5/29/2018 09:08'!
unfollowTempVector: tempVector atIndex: tempVectorIndex in: theFP
	<option: #TempVectReadBarrier>
	<inline: #never> "So rare it mustn't bulk up the common path"
	| followed |
	followed := objectMemory followForwarded: tempVector.
	self temporary: tempVectorIndex in: theFP put: followed.
	^followed! !

!StackInterpreter methodsFor: 'miscellaneous bytecodes' stamp: 'eem 2/21/2013 10:39'!
unknownBytecode
	"If an error selector is available then send it to the activeContext, otherwise abort."
	self respondToUnknownBytecode! !

!StackInterpreter methodsFor: 'sista bytecodes' stamp: 'cb 10/24/2017 15:58'!
unknownInlinePrimitive
	<inline: true> "Should be inlined everywhere to access localIP"
	localIP := localIP - 3.
	self respondToUnknownBytecode! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 4/1/2009 10:44'!
unknownShortOrCodeSizeInKs
	"preserve whatever this value was (for images run on Cog)"
	^theUnknownShort! !

!StackInterpreter methodsFor: 'I/O primitive support' stamp: 'eem 5/12/2017 18:27'!
updateDisplayLeft: l Top: t Right: r Bottom: b
	"Repaint the portion of the Smalltalk screen bounded by the affected rectangle.
	 Used to synchronize the screen after a Bitblt to the Smalltalk Display object."
	| left right top bottom surfaceHandle |
	left := l max: 0.
	right := r min: displayWidth.
	top := t max: 0.
	bottom := b min: displayHeight.
	(left <= right and: [top <= bottom]) ifFalse: [^nil].
	(objectMemory isIntegerObject: displayBits asInteger)
		ifTrue:
			[surfaceHandle := objectMemory integerValueOf: displayBits asInteger.
			 showSurfaceFn = 0 ifTrue:
				[showSurfaceFn := self ioLoadFunction: 'ioShowSurface' From: 'SurfacePlugin'.
				 showSurfaceFn = 0 ifTrue: [^self primitiveFailFor: PrimErrNotFound]].
			 self show: surfaceHandle Sur: left fa: top ce: right - left Fn: bottom - top]
		ifFalse:
			[self assert: (objectMemory isNonImmediate: displayBits asInteger).
			 self ioShow: displayBits asInteger
				D: displayWidth i: displayHeight s: displayDepth
				p: left l: right a: top y: bottom]! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 1/4/2017 08:51'!
updateObjectsPostByteSwap
	"Byte-swap the words of all bytes objects in the image, including Strings, ByteArrays,
	 and CompiledMethods. This returns these objects to their original byte ordering
	 after blindly byte-swapping the entire image. For compiled  methods, byte-swap
	 only their bytecodes part. Ensure floats are in platform-order."
	| swapFloatWords |
	swapFloatWords := objectMemory vmEndianness ~= imageFloatsBigEndian.
	self assert: ClassFloatCompactIndex ~= 0.
	objectMemory allObjectsDo:
		[:oop| | fmt wordAddr methodHeader temp |
		fmt := objectMemory formatOf: oop.
		 fmt >= self firstByteFormat ifTrue: "oop contains bytes"
			[wordAddr := oop + objectMemory baseHeaderSize.
			fmt >= self firstCompiledMethodFormat ifTrue: "compiled method; start after methodHeader and literals"
				[methodHeader := objectMemory longAt: oop + objectMemory baseHeaderSize.
				 wordAddr := wordAddr + (((objectMemory literalCountOfMethodHeader: methodHeader) + LiteralStart) * objectMemory bytesPerOop)].
			objectMemory reverseBytesFrom: wordAddr to: oop + (objectMemory sizeBitsOf: oop)].
		 fmt = self firstLongFormat ifTrue: "Bitmap, Float etc"
			[(swapFloatWords
			  and: [(objectMemory compactClassIndexOf: oop) = ClassFloatCompactIndex])
				ifTrue:
					[temp := objectMemory longAt: oop + objectMemory baseHeaderSize.
					 objectMemory longAt: oop + objectMemory baseHeaderSize put: (self longAt: oop + objectMemory baseHeaderSize + 4).
					 objectMemory longAt: oop + objectMemory baseHeaderSize + 4 put: temp]
				ifFalse:
					[(objectMemory hasSpurMemoryManagerAPI not
					  and: [objectMemory wordSize = 8]) ifTrue: "Object contains 32-bit half-words packed into 64-bit machine words."
						[wordAddr := oop + objectMemory baseHeaderSize.
						 objectMemory reverseWordsFrom: wordAddr to: oop + (objectMemory sizeBitsOf: oop)]]]]! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
updateStateOfSpouseContextForFrame: theFP WithSP: theSP
	"Update the frame's spouse context with the frame's current state except for the
	 sender and instruction pointer, which are used to mark the context as married."
	| theContext tempIndex pointer |
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	<var: #pointer type: #'char *'>
	self assert: (self frameHasContext: theFP).
	theContext := self frameContext: theFP.
	self assert: (self frameReceiver: theFP)
				= (objectMemory fetchPointer: ReceiverIndex ofObject: theContext).
	tempIndex := self frameNumArgs: theFP.
	"update the arguments. this would appear not to be strictly necessary, but is for two reasons.
	 First, the fact that arguments are read-only is only as convention in the Smalltalk compiler;
	 other languages may choose to modify arguments.
	 Second, the Squeak runUntilErrorOrReturnFrom: nightmare pops the stack top, which may, in
	 certain circumstances, be the last argument, and hence the last argument may not have been
	 stored into the context."
	pointer := theFP + (self frameStackedReceiverOffsetNumArgs: tempIndex).
	1 to: tempIndex do:
		[:i|
		pointer := pointer - objectMemory wordSize.
		self assert: (objectMemory addressCouldBeOop: (stackPages longAt: pointer)).
		 objectMemory storePointer: ReceiverIndex + i
			ofObject: theContext
			withValue: (stackPages longAt: pointer)].
	"now update the non-argument stack contents."
	pointer := theFP + FoxReceiver - objectMemory wordSize.
	[pointer >= theSP] whileTrue:
		[self assert: (objectMemory addressCouldBeOop: (stackPages longAt: pointer)).
		 tempIndex := tempIndex + 1.
		 objectMemory storePointer: ReceiverIndex + tempIndex
			ofObject: theContext
			withValue: (stackPages longAt: pointer).
		 pointer := pointer - objectMemory wordSize].
	self assert: ReceiverIndex + tempIndex < (objectMemory lengthOf: theContext).
	objectMemory storePointerUnchecked: StackPointerIndex
		ofObject: theContext
		withValue: (objectMemory integerObjectOf: tempIndex)! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 3/24/2017 17:26'!
validBCPC: thePC inMethod: aMethod
	<var: #aMethod type: #usqInt>
	"Note that we accept anInstrPointer pointing to a callPrimitiveBytecode
	 at the start of a method that contains a primitive.  This because methods like
	 Context(Part)>>reset have to be updated to skip the callPrimtiive bytecode otherwise."
	^thePC >= (self startPCOfMethod: aMethod)
	  and: [thePC < (objectMemory numBytesOfBytes: aMethod)]! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 9/29/2014 13:53'!
validInstructionPointer: anInstrPointer inFrame: fp
	<var: #anInstrPointer type: #usqInt>
	<var: #fp type: #'char *'>
	<inline: false>
	"Note that we accept anInstrPointer pointing to a callPrimitiveBytecode
	 at the start of a method that contains a primitive.  This because methods like
	 Context(Part)>>reset have to be updated to skip the callPrimtiive bytecode otherwise."
	^self validInstructionPointer: anInstrPointer inMethod: (self frameMethodObject: fp) framePointer: fp! !

!StackInterpreter methodsFor: 'debug support' stamp: 'eem 3/6/2015 13:47'!
validInstructionPointer: theInstrPointer inMethod: aMethod framePointer: fp
	<var: #theInstrPointer type: #usqInt>
	<var: #aMethod type: #usqInt>
	<var: #fp type: #'char *'>
	"Note that we accept anInstrPointer pointing to a callPrimitiveBytecode
	 at the start of a method that contains a primitive.  This because methods like
	 Context(Part)>>reset have to be updated to skip the callPrimtiive bytecode otherwise."
	"-1 for pre-increment in fetchNextBytecode"
	^theInstrPointer >= (aMethod + (objectMemory lastPointerOf: aMethod) + objectMemory bytesPerOop - 1)
	  and: [theInstrPointer < (aMethod + (objectMemory numBytesOfBytes: aMethod) + objectMemory baseHeaderSize - 1)]! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 12/4/2013 16:44'!
validStackPageBaseFrame: aPage
	"Check that the base frame in the stack page has a valid sender context."
	<var: #aPage type: #'StackPage *'>
	<inline: false>
	| senderContextOrNil |
	(self asserta: (self isBaseFrame: aPage baseFP)) ifFalse:
		[^false].
	senderContextOrNil := self frameCallerContext: aPage baseFP.
	(self asserta: (objectMemory addressCouldBeObj: senderContextOrNil)) ifFalse:
		[^false].
	(self asserta: (senderContextOrNil = objectMemory nilObject or: [objectMemory isContext: senderContextOrNil])) ifFalse:
		[^false].
	^true! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 12/4/2013 16:42'!
validStackPageBaseFrames
	"Check that the base frames in all in-use stack pages have a sender and a saved context."
	<var: #aPage type: #'StackPage *'>
	0 to: numStackPages - 1 do:
		[:i| | aPage senderContextOrNil savedThisContext |
		aPage := stackPages stackPageAt: i.
		(stackPages isFree: aPage) ifFalse:
			[(self validStackPageBaseFrame: aPage) ifFalse:
				[^false]]].
	^true! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 12/11/2013 10:05'!
voidVMStateForSnapshotFlushingExternalPrimitivesIf: flushExtPrims
	"Make sure that all VM state that affects the heap contents is voided so that the heap is
	 ready to be snapshotted.  If flushExtPrims is true, flush references to external
	 primitives in methods.  Answer the activeContext that should be stored in the snapshot."
	| activeContext |
	<inline: false>
	activeContext := self divorceAllFrames.
	self bereaveAllMarriedContextsForSnapshotFlushingExternalPrimitivesIf: flushExtPrims.
	^activeContext! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'cb 7/31/2018 14:24'!
wakeHighestPriority
	"Return the highest priority process that is ready to run.
	 To save time looking at many empty lists before finding a
	 runnable process the VM maintains a variable holding the
	 highest priority runnable process.  If this variable is 0 then the
	 VM does not know the highest priority and must search all lists.
	 Note: It is a fatal VM error if there is no runnable process."
	| schedLists p processList proc ctxt |
	self externalWriteBackHeadFramePointers.
	schedLists := objectMemory fetchPointer: ProcessListsIndex ofObject: self schedulerPointer.
	p := highestRunnableProcessPriority = 0
			ifTrue: [objectMemory numSlotsOf: schedLists]
			ifFalse: [highestRunnableProcessPriority].
	[(p := p - 1) >= 0] whileTrue:
		[processList := objectMemory fetchPointer: p ofObject: schedLists.
	 	 [self isEmptyList: processList] whileFalse:
			["Only answer processes with a runnable suspendedContext.
			  Discard those that aren't; the VM would crash otherwise."
			 proc := self removeFirstLinkOfList: processList.
			 ctxt := objectMemory fetchPointer: SuspendedContextIndex ofObject: proc.
			 (self isLiveContext: ctxt) ifTrue:
				[highestRunnableProcessPriority := p + 1.
				^proc].
			  self cppIf: SPURVM ifTrue: 
				["This is uncommon, so we can deal with forwarders here instead of assuming there isn't."
				 (self isOopForwarded: ctxt) ifTrue:
					[ctxt := self fixFollowedField: SuspendedContextIndex ofObject: proc withInitialValue: ctxt].
				  (self isLiveContext: ctxt) ifTrue:
					[highestRunnableProcessPriority := p + 1.
					^proc].].
			 self warning: 'evicted zombie process from run queue']].
	self error: 'scheduler could not find a runnable process'.
	^nil! !

!StackInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 17:47'!
whereIs: anOop
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	<returnTypeC: #'char *'>
	<inline: false>
	<var: 'where' type: #'char *'>
	(objectMemory whereIsMaybeHeapThing: anOop) ifNotNil: [:where| ^where].
	(stackPages whereIsMaybeStackThing: anOop) ifNotNil: [:where| ^where].
	^' is no where obvious'! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem 1/8/2015 14:43'!
withSmallIntegerTags: value
	<inline: true>
	<var: #value type: #'char *'>
	self assert: ((self oopForPointer: value) bitAnd: objectMemory wordSize - 1) = 0.
	^(self
		cCode: [self oopForPointer: value]
		inSmalltalk: [value bitAnd: objectMemory maxCInteger]) + objectMemory smallIntegerTag! !

!StackInterpreter methodsFor: 'frame access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
withoutSmallIntegerTags: anInteger
	<inline: true>
	<returnTypeC: #'char *'>
	self assert: (objectMemory isIntegerObject: anInteger).
	^self pointerForOop: (anInteger - 1)! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'nice 4/3/2016 00:33'!
wordSwapped: w
	"Return the given 64-bit integer with its halves in the reverse order."

	objectMemory wordSize = 8 ifFalse: [self error: 'This cannot happen.'].
	^   ((w >> 32) bitAnd: 16r00000000FFFFFFFF) bitOr:
	     ((w << 32) bitAnd: 16rFFFFFFFF00000000)
! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 4/23/2009 17:11'!
writeBackHeadFramePointers
	self assert: stackPage = stackPages mostRecentlyUsedPage.
	self setHeadFP: localFP andSP: localSP inPage: stackPage.
	self assert: stackPages pageListIsWellFormed! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 8/22/2017 18:38'!
writeImageFileIO
	"Write the image header and heap contents to imageFile for snapshot. c.f. writeImageFileIOSimulation.
	 The game below is to maintain 64-bit alignment for all putLong:toFile: occurrences."
	<inline: #never>
	| imageName headerStart headerSize f imageBytes bytesWritten sCWIfn okToWrite |
	<var: #f type: #sqImageFile>
	<var: #headerStart type: #squeakFileOffsetType>
	<var: #sCWIfn type: #'void *'>
	<var: #imageName declareC: 'extern char imageName[]'>

	self cCode: [] inSmalltalk: [imageName := 'sooth compiler'. ^self writeImageFileIOSimulation].

	"If the security plugin can be loaded, use it to check for write permission.
	 If not, assume it's ok"
	sCWIfn := self ioLoadFunction: 'secCanWriteImage' From: 'SecurityPlugin'.
	sCWIfn ~= 0 ifTrue:
		[okToWrite := self cCode: '((sqInt (*)(void))sCWIfn)()'.
		 okToWrite ifFalse:[^self primitiveFail]].
	
	"local constants"
	headerStart := 0.  
	headerSize := objectMemory wordSize = 4 ifTrue: [64] ifFalse: [128].  "header size in bytes; do not change!!"

	f := self sqImageFile: imageName Open: 'wb'.
	f = nil ifTrue: "could not open the image file for writing"
		[^self primitiveFail].

	imageBytes := objectMemory imageSizeToWrite.
	headerStart := self sqImage: f File: imageName StartLocation: headerSize + imageBytes.
	self cCode: '/* Note: on Unix systems one could put an exec command here, padded to 512 bytes */'.
	"position file to start of header"
	self sqImageFile: f Seek: headerStart.

	self putWord32: self imageFormatVersion toFile: f.
	self putWord32: headerSize toFile: f.
	self putLong: imageBytes toFile: f.
	self putLong: objectMemory baseAddressOfImage toFile: f.
	self putLong: objectMemory specialObjectsOop toFile: f.
	self putLong: objectMemory newObjectHash toFile: f.
	self putLong: self getSnapshotScreenSize toFile: f.
	self putLong: self getImageHeaderFlags toFile: f.
	self putWord32: extraVMMemory toFile: f.
	self putShort: desiredNumStackPages toFile: f.
	self putShort: self unknownShortOrCodeSizeInKs toFile: f.
	self putWord32: desiredEdenBytes toFile: f.
	self putShort: (maxExtSemTabSizeSet ifTrue: [self ioGetMaxExtSemTableSize] ifFalse: [0]) toFile: f.
	self putShort: the2ndUnknownShort toFile: f.
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[self putLong: objectMemory firstSegmentBytes toFile: f.
			 self putLong: objectMemory bytesLeftInOldSpace toFile: f.
			 2 timesRepeat: [self putLong: 0 toFile: f]	"Pad the rest of the header."]
		ifFalse:
			[4 timesRepeat: [self putLong: 0 toFile: f]].  "Pad the rest of the header."

	 objectMemory wordSize = 8 ifTrue:
		[3 timesRepeat: [self putLong: 0 toFile: f]]. "Pad the rest of the header."

	self assert: headerStart + headerSize = (self sqImageFilePosition: f).
	"position file after the header"
	self sqImageFile: f Seek: headerStart + headerSize.

	self successful ifFalse: "file write or seek failure"
		[self sqImageFileClose: f.
		 ^nil].

	"write the image data"
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[bytesWritten := objectMemory writeImageSegmentsToFile: f]
		ifFalse:
			[bytesWritten := self sq: (self pointerForOop: objectMemory baseAddressOfImage)
								Image: (self sizeof: #char)
								File: imageBytes
								Write: f].
	self success: bytesWritten = imageBytes.
	self sqImageFileClose: f! !

!StackInterpreter methodsFor: 'image save/restore' stamp: 'eem 8/14/2021 07:04'!
writeImageFileIOSimulation
	"Write the image header and heap contents to imageFile for snapshot.
	 c.f. writeImageFileIO.  The game below is to maintain 64-bit alignment
	 for all putLong:toFile: occurrences."
	<doNotGenerate>
	| headerSize file |
	headerSize := objectMemory wordSize = 4 ifTrue: [64] ifFalse: [128].

	(file := FileStream fileNamed: self imageName) ifNil:
		[self primitiveFail.
		 ^nil].
	[file binary.
	 self putWord32: self imageFormatVersion toFile: file.
	 self putWord32: headerSize toFile: file.
	 {
		objectMemory imageSizeToWrite.
		objectMemory baseAddressOfImage.
		objectMemory specialObjectsOop.
		objectMemory lastHash.
		self ioScreenSize.
		self getImageHeaderFlags
	 }
		do: [:long | self putLong: long toFile: file].

	 self putWord32: (extraVMMemory ifNil: [0]) toFile: file.
	 {	desiredNumStackPages. self unknownShortOrCodeSizeInKs } do:
		[:short| self putShort: short toFile: file].

	 self putWord32: desiredEdenBytes toFile: file.

	 {	maxExtSemTabSizeSet ifTrue: [self ioGetMaxExtSemTableSize] ifFalse: [0]. 0 } do:
		[:short| self putShort: short toFile: file].

	 objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[| bytesWritten |
			 self putLong: objectMemory firstSegmentBytes toFile: file.
			 self putLong: objectMemory bytesLeftInOldSpace toFile: file.
			 2 timesRepeat: [self putLong: 0 toFile: file] "Pad the rest of the header.".
			 objectMemory wordSize = 8 ifTrue:
				[3 timesRepeat: [self putLong: 0 toFile: file]].

			 self assert: file position = headerSize.
			"Position the file after the header."
			file position: headerSize.
			bytesWritten := objectMemory writeImageSegmentsToFile: file.
			self assert: bytesWritten = objectMemory imageSizeToWrite]
		ifFalse:
			["Pad the rest of the header."
			 4 timesRepeat: [self putLong: 0 toFile: file].
			 objectMemory wordSize = 8 ifTrue:
				[3 timesRepeat: [self putLong: 0 toFile: file]].

			 self assert: file position = headerSize.
			 "Position the file after the header."
			 file position: headerSize.

			 "Write the object memory."
			 objectMemory baseAddressOfImage // 4 + 1
				to: objectMemory baseAddressOfImage + objectMemory imageSizeToWrite // 4
				do: [:index |
					self
						putLong: (objectMemory memory at: index)
						toFile: file]].
	 file truncate: file position.
	 self success: true]
		 ensure: [file ifNotNil: [file close]]! !

!StackInterpreter methodsFor: 'process primitive support' stamp: 'eem 8/6/2021 12:45'!
zeroNextProfileTick
	"this is a hook for the simulator..."
	<inline: #always>
	nextProfileTick := 0! !

!StackInterpreter methodsFor: 'stack pages' stamp: 'eem 9/18/2013 18:12'!
zeroStackPage
	"In its own method as a debugging hook.
	 Frame pointers should have been written back already."
	<inline: true>
	self assert: (stackPage = 0
				or: [stackPage headFP = framePointer
					and: [stackPage headSP = stackPointer]]).
	stackPage := 0! !

StackInterpreter subclass: #StackInterpreterPrimitives
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Interpreter'!

StackInterpreterPrimitives subclass: #CoInterpreter
	instanceVariableNames: 'cogit cogMethodZone gcMode cogCodeSize desiredCogCodeSize heapBase lastCoggableInterpretedBlockMethod deferSmash deferredSmash primTraceLog primTraceLogIndex traceLog traceLogIndex traceSources cogCompiledCodeCompactionCalledFor statCodeCompactionCount statCodeCompactionUsecs lastUncoggableInterpretedBlockMethod flagInterpretedMethods maxLiteralCountForCompile minBackwardJumpCountForCompile CFramePointer CStackPointer CReturnAddress primTracePluginName'
	classVariableNames: 'CSCallbackEnter CSCallbackLeave CSCheckEvents CSEnterCriticalSection CSExitCriticalSection CSOwnVM CSResume CSSignal CSSuspend CSSwitchIfNeccessary CSThreadBind CSThreadSchedulingLoop CSWait CSYield HasBeenReturnedFromMCPC HasBeenReturnedFromMCPCOop MFMethodFlagFrameIsMarkedFlag MinBackwardJumpCountForCompile PrimNumberCopyObject PrimNumberObjectAtPut PrimTraceLogSize PrimitiveMetadataFlagsShift RumpCStackSize TraceBlockActivation TraceBlockCreation TraceBufferSize TraceCodeCompaction TraceContextSwitch TraceDisownVM TraceFullGC TraceIncrementalGC TraceIsFromInterpreter TraceIsFromMachineCode TraceOwnVM TracePreemptDisowningThread TracePrimitiveFailure TracePrimitiveRetry TraceSources TraceStackOverflow TraceThreadSwitch TraceVMCallback TraceVMCallbackReturn'
	poolDictionaries: 'CogMethodConstants VMStackFrameOffsets'
	category: 'VMMaker-JIT'!

!CoInterpreter commentStamp: 'eem 3/31/2020 18:56' prior: 0!
I am a variant of the StackInterpreter that can co-exist with the Cog JIT.  I interpret unjitted methods, either because they have been found for the first time or because they are judged to be too big to JIT.  See CogMethod class's comment for method interoperability.

cogCodeSize
	- the current size of the machine code zone

cogCompiledCodeCompactionCalledFor
	- a variable set when the machine code zone runs out of space, causing a machine code zone compaction at the next available opportunity

cogMethodZone
	- the manager for the machine code zone (instance of CogMethodZone)

cogit
	- the JIT (co-jit) (instance of SimpleStackBasedCogit, StackToRegisterMappoingCogit, etc)

deferSmash
	- a flag causing deferral of smashes of the stackLimit around the call of functionSymbol (for assert checks)

deferredSmash
	- a flag noting deferral of smashes of the stackLimit around the call of functionSymbol (for assert checks)

desiredCogCodeSize
	- the desred size of the machine code zone, set at startup or via primitiveVMParameter to be written at snapshot time

flagInterpretedMethods
	- true if methods that are interpreted shoudl have their flag bit set (used to identity methods that are interpreted because they're unjittable for some reason)

gcMode
	- the variable holding the gcMode, used to inform the cogit of how to scan the machine code zone for oops on GC

heapBase
	- the address in memory of the base of the objectMemory's heap, which is immediately above the machine code zone

lastCoggableInterpretedBlockMethod
	- a variable used to invoke the cogit for a block mehtod being invoked repeatedly in the interpreter

lastUncoggableInterpretedBlockMethod
	- a variable used to avoid invoking the cogit for an unjittable method encountered on block evaluation

maxLiteralCountForCompile
	- the variable controlling which methods to jit.  methods with a literal count above this value will not be jitted (on the grounds that large methods are typically used for initialization, and take up a lot of space in the code zone)

minBackwardJumpCountForCompile
	- the variable controlling when to attempt to jit a method being interpreted.  If as many backward jumps as this occur, the current method will be jitted

primTraceLog
	- a small array implementing a crcular buffer logging the last N primitive invocations, GCs, code compactions, etc used for crash reporting

primTraceLogIndex
	- the index into primTraceLog of the next entry

reenterInterpreter
	- the jmpbuf used to jmp back into the interpreter when transitioning from machine code to the interpreter

statCodeCompactionCount
	- the count of machine code zone compactions

statCodeCompactionUsecs
	- the total microseconds spent in machine code zone compactions

traceLog
	- a log of various events, used in debugging

traceLogIndex
	- the index into traceLog of the next entry

traceSources
	- the names associated with the codes of events in traceLog

CFramePointer
	- if in use, the value of the C frame pointer on most recent entry to the interpreter after start-up or a callback.  Used to establish the C stack when calling the run-time from generated machine code.

CStackPointer
	- the value of the C stack pointer on most recent entry to the interpreter after start-up or a callback.  Used to establish the C stack when calling the run-time from generated machine code.

CReturnAddress
	- the return address for the function call which invoked the interpreter at start-up.  Using this as teh return address when entering the interpreter via ceInvokeInterpeter maintains a valid stack.  Since this is effevtively a constant it does not need to be saved and restored once set.!

!CoInterpreter class methodsFor: 'translation' stamp: 'eem 06/07/2018 15:29'!
ancilliaryClasses
	"Answer any extra classes to be included in the translation."
	^(super ancilliaryClasses copyWithout: InterpreterStackPages),
	   {	CoInterpreterStackPages.
		CogBlockMethod. NSSendCache },
	((Cogit ancilliaryClasses) select: [:class| class inheritsFrom: CogBlockMethod])! !

!CoInterpreter class methodsFor: 'translation' stamp: 'eem 2/10/2009 16:33'!
apiExportHeaderName
	^'cointerp.h'! !

!CoInterpreter class methodsFor: 'translation' stamp: 'eem 7/8/2021 11:08'!
clusteredVariableNames
	"Insist that these variables are present early in the list of variables, and in this order,
	 so that e.g. they are conveniently accessed via the VarBaseReg if it is available."
	^#(stackLimitFromMachineCode "ensures zero offset in simulation" stackLimit "stackLimit is e.g. lowest using the clang toolchain on MacOS X"
		stackPointer framePointer CStackPointer CFramePointer CReturnAddress
		scavengeThreshold freeStart needGCFlag specialObjectsOop
		primFailCode primTraceLogIndex primitiveFunctionPointer
		newMethod instructionPointer argumentCount nextProfileTick
		nativeSP nativeStackPointer shadowCallStackPointer
		primTraceLog "should be last because it is large") "reject: [:n| self allInstVarNames includes: n]"! !

!CoInterpreter class methodsFor: 'translation' stamp: 'eem 7/24/2021 14:39'!
declareCVarsIn: aCCodeGenerator
	"Override to avoid repeating StackInterpreter's declarations and add our own extensions"
	self class == thisContext methodClass ifFalse: [^self]. "Don't duplicate decls in subclasses"
	aCCodeGenerator
		addHeaderFile:'"sqCogStackAlignment.h"';
		addHeaderFile:'"cogmethod.h"'.
	NewspeakVM ifTrue:
		[aCCodeGenerator addHeaderFile:'"nssendcache.h"'].
	aCCodeGenerator
		addHeaderFile: (aCCodeGenerator vmClass isThreadedVM 
			ifTrue: ['"cointerpmt.h"'] 
			ifFalse: ['"cointerp.h"']);
		addHeaderFile:'"cogit.h"'.
	aCCodeGenerator vmClass
		declareInterpreterVersionIn: aCCodeGenerator
		defaultName: aCCodeGenerator interpreterVersion.
	aCCodeGenerator
		var: #cogCodeSize type: #usqInt;
		var: #heapBase type: #usqInt;
		var: #statCodeCompactionUsecs type: #usqLong;
		var: #maxLiteralCountForCompile
			declareC: 'sqInt maxLiteralCountForCompile = MaxLiteralCountForCompile /* ', MaxLiteralCountForCompile printString, ' */';
		var: #minBackwardJumpCountForCompile
			declareC: 'sqInt minBackwardJumpCountForCompile = MinBackwardJumpCountForCompile /* ', MinBackwardJumpCountForCompile printString, ' */'.
	aCCodeGenerator
		removeVariable: 'atCache'; "Way too much trouble than it's worth in the Cog VM"
		removeVariable: 'reenterInterpreter'. "We can use the JIT and CFrame/StrackPointer for a lighter-weight solution."
	aCCodeGenerator
		var: #primTraceLogIndex type: #'unsigned char';
		var: #primTraceLog declareC: 'sqInt primTraceLog[256]';
		var: #traceLog
		declareC: 'sqInt traceLog[TraceBufferSize /* ', TraceBufferSize printString, ' */]';
		var: 'primTracePluginName' type: #'char *';
		var: #traceSources type: #'char *' array: TraceSources.
	aCCodeGenerator
		var: #CFramePointer type: #'volatile usqIntptr_t';
		var: #CStackPointer type: #'volatile usqIntptr_t';
		var: #CReturnAddress type: #'volatile usqIntptr_t'! !

!CoInterpreter class methodsFor: 'accessing class hierarchy' stamp: 'eem 9/30/2013 13:18'!
defaultObjectMemoryClass
	^NewCoObjectMemory! !

!CoInterpreter class methodsFor: 'accessing class hierarchy' stamp: 'eem 3/17/2015 13:45'!
hasCogit
	^true! !

!CoInterpreter class methodsFor: 'initialization' stamp: 'eem 4/8/2016 12:16'!
initializeCaches
	"Eliminate the AtCache"
	super initializeCaches.
	AtCacheTotalSize := AtCacheSize := AtCacheMask := AtCacheFixedFields := AtCacheFmt := AtCacheOop := #undefined! !

!CoInterpreter class methodsFor: 'initialization' stamp: 'eem 12/14/2015 12:37'!
initializeContextIndices
	super initializeContextIndices.

	HasBeenReturnedFromMCPC := -1.
	HasBeenReturnedFromMCPCOop := self objectMemoryClass basicNew integerObjectOf: HasBeenReturnedFromMCPC! !

!CoInterpreter class methodsFor: 'initialization' stamp: 'eem 1/24/2016 11:38'!
initializeFrameIndices
	"Format of a stack frame.  Word-sized indices relative to the frame pointer.
	 Terminology
		Frames are either single (have no context) or married (have a context).
		Contexts are either single (exist on the heap), married (have a context) or widowed (had a frame that has exited).
	 Stacks grow down:

			receiver for method activations/closure for block activations
			arg0
			...
			argN
			caller's saved ip/this stackPage (for a base frame)
	fp->	saved fp
			method
			context (initialized to nil)
			frame flags (interpreter only)
			saved method ip (initialized to 0; interpreter only)
			receiver
			first temp
			...
	sp->	Nth temp

	In an interpreter frame
		frame flags holds
			the backward jump count (see ifBackwardsCheckForEvents)
			the number of arguments (since argument temporaries are above the frame)
			the flag for a block activation
			and the flag indicating if the context field is valid (whether the frame is married).
		saved method ip holds the saved method ip when the callee frame is a machine code frame.
		This is because the saved method ip is actually the ceReturnToInterpreterTrampoline address.
	In a machine code frame
		the flag indicating if the context is valid is the least significant bit of the method pointer
		the flag for a block activation is the next most significant bit of the method pointer

	Interpreter frames are distinguished from method frames by the method field which will
	be a pointer into the heap for an interpreter frame and a pointer into the method zone for
	a machine code frame.

	The first frame in a stack page is the baseFrame and is marked as such by a saved fp being its stackPage,
	in which case the first word on the stack is the caller context (possibly hybrid) beneath the base frame."

	| fxCallerSavedIP fxSavedFP fxMethod fxIFrameFlags fxThisContext fxIFReceiver fxMFReceiver fxIFSavedIP |
	fxCallerSavedIP := 1.
	fxSavedFP := 0.
	fxMethod := -1.
	fxThisContext := -2.
	fxIFrameFlags := -3.	"Can find numArgs, needed for fast temp access. args are above fxCallerSavedIP.
							 Can find ``is block'' bit
							 Can find ``has context'' bit"
	fxIFSavedIP := -4.
	fxIFReceiver := -5.
	fxMFReceiver := -3.

	"For debugging undefine values that differ in the StackInterpreter."
	FrameSlots := #undefined.
	IFrameSlots := fxCallerSavedIP - fxIFReceiver + 1.
	MFrameSlots := fxCallerSavedIP - fxMFReceiver + 1.

	FoxCallerSavedIP := fxCallerSavedIP * BytesPerWord.
	"In Cog a base frame's caller context is stored on the first word of the stack page."
	FoxCallerContext := #undefined.
	FoxSavedFP := fxSavedFP * BytesPerWord.
	FoxMethod := fxMethod * BytesPerWord.
	FoxThisContext := fxThisContext * BytesPerWord.
	FoxFrameFlags := #undefined.
	FoxIFrameFlags := fxIFrameFlags * BytesPerWord.
	FoxIFSavedIP := fxIFSavedIP * BytesPerWord.
	FoxReceiver := #undefined.
	FoxIFReceiver := fxIFReceiver * BytesPerWord.
	FoxMFReceiver := fxMFReceiver * BytesPerWord.

	"N.B.  There is room for one more flag given the current 8 byte alignment of methods (which
	 is at least needed to distinguish the checked and uncecked entry points by their alignment."
	MFMethodFlagHasContextFlag := 1.
	MFMethodFlagIsBlockFlag := 2.
	MFMethodFlagFrameIsMarkedFlag := 4. "for pathTo:using:followWeak:"
	MFMethodFlagsMask := MFMethodFlagHasContextFlag + MFMethodFlagIsBlockFlag + MFMethodFlagFrameIsMarkedFlag.
	MFMethodMask := (MFMethodFlagsMask + 1) negated! !

!CoInterpreter class methodsFor: 'initialization' stamp: 'eem 8/20/2021 07:51'!
initializeMiscConstants

	super initializeMiscConstants.
	COGVM := true.

	MinBackwardJumpCountForCompile := 40.

	MaxNumArgs := 15.
	PrimCallNeedsNewMethod := 1.
	PrimCallNeedsPrimitiveFunction := 2.
	PrimCallMayEndureCodeCompaction := 4.
	PrimCallCollectsProfileSamples := 8.
	PrimCallDoNotJIT := 16.
	PrimCallIsExternalCall := 32.
	"CheckAllocationFillerAfterPrimCall := 32. this has never been successfully used in all the years we've had it; nuking it"
	PrimCallOnSmalltalkStack := 64. "Speed up simple external prims by avoiding stack switch"
	PrimCallOnSmalltalkStackAlign2x := 128. "Align stack to a 2 x word size boundary, e.g. for MMX instructions etc"

	"Flags for use in export:flags:, shifted to overlap with the PrimCallXXX flags above"
	FastCPrimitiveFlag := 1.				"a.k.a. PrimCallOnSmalltalkStack"
	FastCPrimitiveAlignForFloatsFlag := 2.	"a.k.a. PrimCallOnSmalltalkStackAlign2x"
	PrimitiveMetadataFlagsShift := PrimCallOnSmalltalkStack highBit - FastCPrimitiveFlag highBit.

	"And to shift away the flags, to compute the accessor depth, use...
	 c.f. NullSpurMetadata in sq.h"
	SpurPrimitiveAccessorDepthShift := 8.
	SpurPrimitiveFlagsMask := 1 << SpurPrimitiveAccessorDepthShift - 1.

	"the primitive trace log; a record of the last 256 named/external primitives or significant events invoked."
	PrimTraceLogSize := 256. "Room for 256 selectors.  Must be 256 because we use a byte to hold the index"
	TraceBufferSize := 256 * 3. "Room for 256 events"
	TraceContextSwitch := self objectMemoryClass basicNew integerObjectOf: 1.
	TraceBlockActivation := self objectMemoryClass basicNew integerObjectOf: 2.
	TraceBlockCreation := self objectMemoryClass basicNew integerObjectOf: 3.
	TraceIncrementalGC := self objectMemoryClass basicNew integerObjectOf: 4.
	TraceFullGC := self objectMemoryClass basicNew integerObjectOf: 5.
	TraceCodeCompaction := self objectMemoryClass basicNew integerObjectOf: 6.
	TraceOwnVM := self objectMemoryClass basicNew integerObjectOf: 7.
	TraceDisownVM := self objectMemoryClass basicNew integerObjectOf: 8.
	TraceThreadSwitch := self objectMemoryClass basicNew integerObjectOf: 9.
	TracePreemptDisowningThread := self objectMemoryClass basicNew integerObjectOf: 10.
	TraceVMCallback := self objectMemoryClass basicNew integerObjectOf: 11.
	TraceVMCallbackReturn := self objectMemoryClass basicNew integerObjectOf: 12.
	TraceStackOverflow := self objectMemoryClass basicNew integerObjectOf: 13.
	TracePrimitiveFailure := self objectMemoryClass basicNew integerObjectOf: 14.
	TracePrimitiveRetry := self objectMemoryClass basicNew integerObjectOf: 15.

	TraceIsFromMachineCode := 1.
	TraceIsFromInterpreter := 2.
	CSCallbackEnter := 3.
	CSCallbackLeave := 4.
	CSEnterCriticalSection := 5.
	CSExitCriticalSection := 6.
	CSResume := 7.
	CSSignal := 8.
	CSSuspend := 9.
	CSWait := 10.
	CSYield := 11.
	CSCheckEvents := 12.
	CSThreadSchedulingLoop := 13.
	CSOwnVM := 14.
	CSThreadBind := 15.
	CSSwitchIfNeccessary := 16.

	TraceSources := CArrayAccessor on: #('?' 'm' 'i' 'callbackEnter' 'callbackLeave' 'enterCritical' 'exitCritical' 'resume' 'signal'  'suspend' 'wait' 'yield' 'eventcheck' 'threadsched' 'ownVM' 'bindToThread' 'switchIfNecessary').

	"this is simulation only"
	RumpCStackSize := 4096! !

!CoInterpreter class methodsFor: 'initialization' stamp: 'eem 9/4/2021 17:34'!
initializePrimitiveTable
	super initializePrimitiveTable.

	"These two are run on the Smalltalk stack under Spur.  See primitivePropertyFlagsForSpur:"
	PrimNumberObjectAtPut := 69.
	self assert: (PrimitiveTable at: PrimNumberObjectAtPut + 1) = #primitiveObjectAtPut.
	PrimNumberCopyObject := 168.
	self assert: (PrimitiveTable at: PrimNumberCopyObject + 1) = #primitiveCopyObject.

	#(216 253) do:
		[:pidx| self assert: (PrimitiveTable at: pidx + 1) = #primitiveFail].
	self assert: (PrimitiveTable at: 215 + 1) = #primitiveFlushCacheByMethod.
	PrimitiveTable
		at: 253 + 1 put: #primitiveCollectCogCodeConstituents;
		at: 215 + 1 put: #primitiveVoidVMStateForMethod;
		at: 216 + 1 put: #primitiveMethodXray! !

!CoInterpreter class methodsFor: 'documentation' stamp: 'eem 05/07/2020 16:13'!
interpreterMachineCodeTransitions
	"The CoInterpreter only asks the Cog compiler to generate machine-code methods
	 when a bytecoded method has been found in the cache, or block value has tried to
	 invoke a block in the method two times consecutively.  This prevents the compiler
	 being asked to compile an infrequenttly used method.

	I would like the following to be true, but it isn't.  The interpreter *does* invoke
	machine-code primitives that may context switch.

	 The CoInterpreter will only activate a Cog method that doesn't have a primitive
	 (this does not mean it won't invoke a Cog block method; it just does so through the
	 interpreted block value primitives).  This is to avoid serious complications with the
	 process switch primitives.  The CoInterpreter needs to know if it should push the
	 instructionPointer or save it in frameSavedIP and substitute ceReturtnToInterpreterPC
	 as the pushed instruction pointer.  The process switch primitives need to know if
	 they were called from the interpreter or from machine-code to know how to continue.

	 If a process switch primitive has been invoked from the interpreter and switches to
	 a process suspended in an interpreted method it can return to the interpreter.  In both
	 cases switching to a process in machine-code the primitive can continue via the
	 ceEnterCogCodeXXX enilopmart(s).  But if in machine-code and switching to a process
	 in the interpreter it must longjmp to the interpreter.  So the process-switch primitives
	 need to know whether they werer invoked from the interpreter or not.

	 If the process-switch primitives are allowed to be invoked from the interpreter via a
	 machine-code method then, in the immortal words of Robert Fripp, ``affairs stand a
	 good chance of getting severely out of hand...'' (The Guitar Handbook, Ralph Denyer,
	 p 114, Pan Books).  The VM now has to longjmp not only if invoked from machine code
	 and switching to the interpreter but if invoked from the interpreter via machine code
	 and switching to the interpreter.  The issue is that it is difficult to discover from within
	 a primitive whether the primitive call is from machine code, as it should be; it isn't a
	 concern of the primitive.  Hence KISS says ``no machine-code invocation of primitives
	 from the interpreter''."! !

!CoInterpreter class methodsFor: 'accessing' stamp: 'EstebanLorenzano 10/23/2014 14:20'!
interpreterVersion 
	^ 'Cog'! !

!CoInterpreter class methodsFor: 'translation' stamp: 'eem 10/25/2013 14:40'!
isAcceptableAncilliaryClass: aClass
	^aClass ~~ InterpreterStackPages! !

!CoInterpreter class methodsFor: 'translation' stamp: 'eem 11/16/2013 13:51'!
isNonArgumentImplicitReceiverVariableName: aString
	^(#('self' 'stackPages' 'cogit' 'coInterpreter' 'cogMethodZone' 'interpreter' 'objectMemory') includes: aString)
	  or: [self objectMemoryClass isNonArgumentImplicitReceiverVariableName: aString]! !

!CoInterpreter class methodsFor: 'testing' stamp: 'eem 8/27/2009 11:20'!
isThreadedVM
	^false! !

!CoInterpreter class methodsFor: 'spur compilation support' stamp: 'eem 8/1/2021 11:08'!
metadataFlagsForPrimitive: aPrimitiveMethodOrNil
	"We allow methods to decorate themselves with 8 flags (only one specified so far: FastCPrimitiveFlag)
	 using the export:flags: pragma."
	aPrimitiveMethodOrNil ifNil: [^0].
	^(aPrimitiveMethodOrNil pragmaAt: #export:flags:)
		ifNil: [0]
		ifNotNil: [:pragma| | flags |
			(flags := pragma arguments second) isInteger
				ifTrue: [flags]
				ifFalse: [self valueOfMetadataFlag: flags]]! !

!CoInterpreter class methodsFor: 'translation' stamp: 'eem 6/20/2021 16:21'!
mustBeGlobal: var
	"Answer if a variable must be global and exported.  Used for inst vars that are accessed from VM support code."

	^(super mustBeGlobal: var)
	   or: [#('desiredCogCodeSize' 'heapBase'
			'maxLiteralCountForCompile' 'minBackwardJumpCountForCompile'
			'primTracePluginName') includes: var]! !

!CoInterpreter class methodsFor: 'translation' stamp: 'eem 2/9/2009 13:26'!
needsCogit
	^true! !

!CoInterpreter class methodsFor: 'translation' stamp: 'eem 7/7/2014 07:01'!
preGenerationHook: aCCodeGenerator
	"Override to undo the hiding of primitiveClosureValueNoContextSwitch"
	super preGenerationHook: aCCodeGenerator.
	(aCCodeGenerator methodNamed: #primitiveClosureValueNoContextSwitch) static: false.

	"horrible hack to declare primErrorCode and argumentCount as bytes in the same word, and
	 hence save an instruction by initializing two birds^H^H^H^H^Hbytes with one word write.
	 Stalled awaiting MoveAbR and MoveRAb support in the Cogit"
	false ifTrue:
		[aCCodeGenerator
			var: #argumentCount declareC: '#define argumentCount acpfc.ac\#define primFailCode acpfc.pfc' withCRs;
			var: #primFailCode declareC: '#if VMBIGENDIAN\struct { short pad; unsigned char pfc; unsigned char ac; } acpfc;\#else /* argumentCount & primFailCode */\struct { unsigned char ac; unsigned char pfc; } acpfc;\#endif' withCRs]! !

!CoInterpreter class methodsFor: 'accessing class hierarchy' stamp: 'eem 6/30/2014 05:56'!
primitivesClass
	^CoInterpreterPrimitives! !

!CoInterpreter class methodsFor: 'translation' stamp: 'eem 10/12/2013 12:28'!
shouldGenerateTypedefFor: aStructClass
	"Hack to work-around multiple definitions.  Sometimes a type has been defined in an include."
	^(super shouldGenerateTypedefFor: aStructClass)
	  and: [Cogit shouldGenerateTypedefFor: aStructClass]! !

!CoInterpreter class methodsFor: 'translation' stamp: 'eem 2/9/2009 12:25'!
sourceFileName
	"Answer the filename for the core interpreter"

	^'cointerp.c'! !

!CoInterpreter class methodsFor: 'translation' stamp: 'eem 7/6/2010 15:32'!
specialValueForConstant: constantName default: defaultValue
	constantName = 'DoAssertionChecks' ifTrue:
		[^'(!!PRODUCTION)'].
	constantName = 'AllocationCheckFiller' ifTrue:
		[^('#if !!defined(AllocationCheckFiller)\# define AllocationCheckFiller ', defaultValue, '\#endif') withCRs].
	^super specialValueForConstant: constantName default: defaultValue! !

!CoInterpreter class methodsFor: 'spur compilation support' stamp: 'eem 8/1/2021 11:08'!
valueOfMetadataFlag: symbolOrArray
	"We allow methods to decorate themselves with 8 flags (only one specified so far: FastCPrimitiveFlag)
	 using the export:flags: pragma.  A single flag is supplied as a symbol.  Multiple flags are supplied as
	 an array of symbols (pragmas take only literal arguments)."
	^symbolOrArray isArray
		ifTrue:
			[symbolOrArray inject: 0 into: [:flags :flagName| flags + (self valueOfMetadataFlag: flagName)]]
		ifFalse:
			[(self bindingOf: symbolOrArray)
				ifNil: [self error: 'could not find primitive flag'. 0]
				ifNotNil:
					[:binding| | flagValue |
					((flagValue := binding value) isInteger and: [flagValue between: 1 and: 255])
						ifTrue: [flagValue]
						ifFalse: [self error: 'integer flag required'. 0]]]! !

!CoInterpreter class methodsFor: 'translation' stamp: 'eem 8/23/2018 13:26'!
writeVMHeaderTo: aStream bytesPerWord: bytesPerWord generator: aCCodeGenerator
	super writeVMHeaderTo: aStream bytesPerWord: bytesPerWord generator: aCCodeGenerator.
	aCCodeGenerator
		putDefineOf: #COGVM as: 1 on: aStream;
		putConditionalDefineOf: #COGMTVM as: 0 comment: nil on: aStream.
	aStream cr! !

!CoInterpreter methodsFor: 'simulation' stamp: 'eem 12/21/2017 17:14'!
ISA
	<doNotGenerate>
	^cogit backEnd class ISA! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 10/30/2020 07:41'!
_longjmp: aJumpBuf _: returnValue
	"Hack simulation of _setjmp/_longjmp, intended to invoke the most minimal setjmp/longjmp
	 pair available on the platform; no saving/restoring signal masks, no stack unwinding, etc.
	 Signal the exception that simulates a longjmp back to the interpreter." 
	<doNotGenerate>
	self halt: 'This should not be encountered now we use ceInvokeInterpreter!!!!'! !

!CoInterpreter methodsFor: 'interpreter shell' stamp: 'eem 10/29/2020 10:10'!
aboutToDispatchBytecode
	"This is a hook for the simulator; null in production"
	<inline: #always>! !

!CoInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 9/5/2021 06:48'!
accessorDepthForExternalPrimitiveMethod: methodObj
	<option: #SpurObjectMemory>
	| flags lit |
	self assert: (self isLinkedExternalPrimitive: methodObj).
	lit := self literal: 0 ofMethod: methodObj.
	 flags := objectMemory fetchPointer: ExternalCallLiteralFlagsIndex ofObject: lit.
 	 ^(objectMemory integerValueOf: flags) >>> SpurPrimitiveAccessorDepthShift! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 6/30/2014 20:25'!
accessorDepthForPrimitiveIndex: primIndex
	<api>
	<option: #SpurObjectMemory>
	^primitiveAccessorDepthTable at: primIndex! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 8/19/2021 16:23'!
accessorDepthForPrimitiveMethod: aMethodObj
	<api>
	<option: #SpurObjectMemory>
	| primIndex |
	primIndex := self primitiveIndexOf: aMethodObj.
	^primIndex = PrimNumberExternalCall
		ifTrue: [self accessorDepthForExternalPrimitiveMethod: aMethodObj]
		ifFalse: [self accessorDepthForPrimitiveIndex: primIndex]! !

!CoInterpreter methodsFor: 'primitive support' stamp: 'eem 3/29/2020 19:11'!
activateFailingPrimitiveMethod
	"Assuming the primFailCode (and any other relevant failure state) has been set,
	 switch the VM to the interpreter if necessary (if in the CoInterpreter executing machine code),
	 and activate the newMethod (which is expected to have a primitive)."
	self assert: primFailCode ~= 0.
	self assert: (objectMemory addressCouldBeObj: newMethod).
	self assert: (objectMemory isCompiledMethod: newMethod).
	self assert: (self primitiveIndexOf: newMethod) ~= 0.
	self justActivateNewMethod: true. "Frame must be interpreted"
	cogit ceInvokeInterpret.
	"NOTREACHED"
	^nil! !

!CoInterpreter methodsFor: 'control primitives' stamp: 'eem 4/23/2017 11:33'!
activateNewClosure: blockClosure outer: outerContext method: theMethod  numArgs: numArgs mayContextSwitch: mayContextSwitch
	"Similar to activateNewMethod but for Closure and newMethod.
	 Override to handle the various interpreter/machine code transitions
	 and to create an appropriate frame layout."
	| numCopied methodHeader inInterpreter closureIP switched |
	<inline: true>
	self assert: (objectMemory isContext: outerContext).
	numCopied := self copiedValueCountOfClosure: blockClosure.
	self assert: theMethod = (objectMemory fetchPointer: MethodIndex ofObject: outerContext).
	self assert: (objectMemory isOopCompiledMethod: theMethod).
	methodHeader := self rawHeaderOf: theMethod.
	(self isCogMethodReference: methodHeader) ifTrue:
		[^self executeCogBlock: (self cogMethodOf: theMethod)
			closure: blockClosure
			mayContextSwitch: mayContextSwitch].
	"How do we know when to compile a block method?
	 One simple criterion is to check if the block is running within its inner context,
	 i.e. if the outerContext is married.
	 Even simpler is to remember the previous block entered via the interpreter and
	 compile if this is the same one.  But we can thrash trying to compile an uncoggable
	 method unless we try and remember which ones can't be cogged.  So also record
	 the last block method we failed to compile and avoid recompiling it."
	(self methodWithHeaderShouldBeCogged: methodHeader)
		ifTrue:
			[(instructionPointer < objectMemory startOfMemory "If from machine code (via value primitive) attempt jitting"
			  or: [theMethod = lastCoggableInterpretedBlockMethod]) "If from interpreter and repeat block, attempt jitting"
				ifTrue:
					[theMethod ~= lastUncoggableInterpretedBlockMethod ifTrue:
						[cogit cog: theMethod selector: objectMemory nilObject.
						 (self methodHasCogMethod: theMethod) ifTrue:
							[^self executeCogBlock: (self cogMethodOf: theMethod)
								closure: blockClosure
								mayContextSwitch: mayContextSwitch].
						 cogCompiledCodeCompactionCalledFor ifFalse:
							[lastUncoggableInterpretedBlockMethod := theMethod]]]
				ifFalse:
					[lastCoggableInterpretedBlockMethod := theMethod]]
		ifFalse:
			[self maybeFlagMethodAsInterpreted: theMethod].

	self assert: (self methodHasCogMethod: theMethod) not.
	"Because this is an uncogged method we need to continue via the interpreter.
	 We could have been reached either from the interpreter, in which case we
	 should simply return, or from a machine code frame or from a compiled
	 primitive.  In these latter two cases we must longjmp back to the interpreter.
	 The instructionPointer tells us which path we took.
	 If the sender was an interpreter frame but called through a (failing) primitive
	 then make sure we restore the saved instruction pointer and avoid pushing
	 ceReturnToInterpreterPC which is only valid between an interpreter caller
	 frame and a machine code callee frame."
	(inInterpreter := instructionPointer >= objectMemory startOfMemory) ifFalse:
		[instructionPointer = cogit ceReturnToInterpreterPC ifTrue:
			[instructionPointer := self iframeSavedIP: framePointer]].
	self push: instructionPointer.
	self push: framePointer.
	framePointer := stackPointer.
	self push: theMethod.
	self push: objectMemory nilObject. "FxThisContext field"
	self push: (self encodeFrameFieldHasContext: false isBlock: true numArgs: numArgs).
	self push: 0. "FoxIFSavedIP"
	"Because inst var access is not checked, we must follow the receiver in Spur to ensure it is valid."
	self push: (objectMemory followField: ReceiverIndex ofObject: outerContext).

	"Copy the copied values..."
	0 to: numCopied - 1 do:
		[:i|
		self push: (objectMemory
					fetchPointer: i + ClosureFirstCopiedValueIndex
					ofObject: blockClosure)].

	self assert: (self frameIsBlockActivation: framePointer).
	self assert: (self frameHasContext: framePointer) not.

	"The initial instructions in the block nil-out remaining temps."

	"the instruction pointer is a pointer variable equal to 
	method oop + ip + BaseHeaderSize 
	-1 for 0-based addressing of fetchByte 
	-1 because it gets incremented BEFORE fetching currentByte"
	closureIP := self quickFetchInteger: ClosureStartPCIndex ofObject: blockClosure.
	instructionPointer := theMethod + closureIP + objectMemory baseHeaderSize - 2.
	self setMethod: theMethod methodHeader: methodHeader.

	"Now check for stack overflow or an event (interrupt, must scavenge, etc)"
	switched := false.
	stackPointer < stackLimit ifTrue:
		[switched := self handleStackOverflowOrEventAllowContextSwitch: mayContextSwitch].
	self returnToExecutive: inInterpreter postContextSwitch: switched! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem 1/9/2020 10:42'!
activateNewCogMethod: cogMethod inInterpreter: inInterpreter
	<var: #cogMethod type: #'CogMethod *'>
	"Activate newMethod when newMethod has been cogged, i.e. create a machine-code frame and (re)enter machine-code."
	| methodHeader rcvr numTemps switched |
	self assert: (self rawHeaderOf: newMethod) = cogMethod asInteger.
	methodHeader := cogMethod methodHeader.
	rcvr := self stackValue: cogMethod cmNumArgs. "could new rcvr be set at point of send?"
	self push: instructionPointer.
	cogMethod stackCheckOffset = 0 ifTrue:
		["frameless method; nothing to activate..."
		 cogit numRegArgs > 0 ifTrue: "dont use and: so as to get Slang to inline cogit numRegArgs > 0"
			[cogMethod cmNumArgs <= cogit numRegArgs ifTrue:
				[self callRegisterArgCogMethod: cogMethod at: cogit noCheckEntryOffset receiver: rcvr]].
		 self push: cogMethod asInteger + cogit noCheckEntryOffset.
		 self push: rcvr.
		 cogit ceCallCogCodePopReceiverReg.
		 self error: 'should not be reached'].
	self push: framePointer.
	framePointer := stackPointer.
	self push: cogMethod asInteger.
	self push: objectMemory nilObject. "FxThisContext field"
	self push: rcvr.

	"clear remaining temps to nil"
	numTemps := self temporaryCountOfMethodHeader: methodHeader.
	cogMethod cmNumArgs + 1 to: numTemps do:
		[:i | self push: objectMemory nilObject].

	((self methodHeaderHasPrimitive: methodHeader)
	 and: [primFailCode ~= 0]) ifTrue:
		[self reapAndResetErrorCodeTo: stackPointer header: methodHeader].

	"Now check for stack overflow or an event (interrupt, must scavenge, etc)."
	stackPointer >= stackLimit ifTrue:
		[self assert: cogMethod stackCheckOffset > cogit noCheckEntryOffset.
		 self push: cogMethod asInteger + cogMethod stackCheckOffset.
		 self push: rcvr.
		 cogit ceEnterCogCodePopReceiverReg.
		 self error: 'should not be reached'].
	instructionPointer := cogMethod asInteger + cogMethod stackCheckOffset.
	switched := self handleStackOverflowOrEventAllowContextSwitch: (self canContextSwitchIfActivating: newMethod header: methodHeader).
	self returnToExecutive: inInterpreter postContextSwitch: switched! !

!CoInterpreter methodsFor: 'control primitives' stamp: 'eem 4/23/2017 11:39'!
activateNewFullClosure: blockClosure method: theMethod numArgs: numArgs mayContextSwitch: mayContextSwitch
	"Similar to activateNewMethod but for Closure and newMethod."
	| numCopied methodHeader numTemps inInterpreter switched |
	<inline: true>
	self assert: theMethod = (objectMemory fetchPointer: FullClosureCompiledBlockIndex ofObject: blockClosure).
	methodHeader := self rawHeaderOf: theMethod.
	(self isCogMethodReference: methodHeader) ifTrue:
		[^self
			executeFullCogBlock: (self cogMethodOf: theMethod)
			closure: blockClosure
			mayContextSwitch: mayContextSwitch].
	numCopied := self copiedValueCountOfFullClosure: blockClosure.
	"How do we know when to compile a block method?
	 One simple criterion is to check if the block is running within its inner context,
	 i.e. if the outerContext is married.
	 Even simpler is to remember the previous block entered via the interpreter and
	 compile if this is the same one.  But we can thrash trying to compile an uncoggable
	 method unless we try and remember which ones can't be cogged.  So also record
	 the last block method we failed to compile and avoid recompiling it."
	(self methodWithHeaderShouldBeCogged: methodHeader)
		ifTrue:
			[(instructionPointer < objectMemory startOfMemory "If from machine code (via value primitive) attempt jitting"
			  or: [theMethod = lastCoggableInterpretedBlockMethod]) "If from interpreter and repeat block, attempt jitting"
				ifTrue:
					[theMethod ~= lastUncoggableInterpretedBlockMethod ifTrue:
						[cogit cogFullBlockMethod: theMethod numCopied: numCopied.
						 (self methodHasCogMethod: theMethod) ifTrue:
							[^self executeFullCogBlock: (self cogMethodOf: theMethod)
								closure: blockClosure
								mayContextSwitch: mayContextSwitch].
						 cogCompiledCodeCompactionCalledFor ifFalse:
							[lastUncoggableInterpretedBlockMethod := theMethod]]]
				ifFalse:
					[lastCoggableInterpretedBlockMethod := theMethod]]
		ifFalse:
			[self maybeFlagMethodAsInterpreted: theMethod].

	self assert: (self methodHasCogMethod: theMethod) not.
	"Because this is an uncogged method we need to continue via the interpreter.
	 We could have been reached either from the interpreter, in which case we
	 should simply return, or from a machine code frame or from a compiled
	 primitive.  In these latter two cases we must longjmp back to the interpreter.
	 The instructionPointer tells us which path we took.
	 If the sender was an interpreter frame but called through a (failing) primitive
	 then make sure we restore the saved instruction pointer and avoid pushing
	 ceReturnToInterpreterPC which is only valid between an interpreter caller
	 frame and a machine code callee frame."
	(inInterpreter := instructionPointer >= objectMemory startOfMemory) ifFalse:
		[instructionPointer = cogit ceReturnToInterpreterPC ifTrue:
			[instructionPointer := self iframeSavedIP: framePointer]].

	self push: instructionPointer.
	self push: framePointer.
	framePointer := stackPointer.
	self push: theMethod.
	self push: objectMemory nilObject. "FxThisContext field"
	self push: (self encodeFrameFieldHasContext: false isBlock: true numArgs: numArgs).
	self push: 0. "FoxIFSavedIP"
	"Because inst var access is not checked, we must follow the receiver in Spur to ensure it is valid."
	self push: (objectMemory followField: FullClosureReceiverIndex ofObject: blockClosure).

	"Copy the copied values..."
	0 to: numCopied - 1 do:
		[:i|
		self push: (objectMemory
					fetchPointer: i + FullClosureFirstCopiedValueIndex
					ofObject: blockClosure)].

	self assert: (self frameIsBlockActivation: framePointer).
	self assert: (self frameHasContext: framePointer) not.

	methodHeader := objectMemory methodHeaderOf: theMethod.
	numTemps := self temporaryCountOfMethodHeader: methodHeader.

	numArgs + numCopied + 1 to: numTemps do: [ :i | self push: objectMemory nilObject].

	instructionPointer := (self initialIPForHeader: methodHeader method: theMethod) - 1.
	
	self setMethod: theMethod.

	"Now check for stack overflow or an event (interrupt, must scavenge, etc)"
	switched := false.
	stackPointer < stackLimit ifTrue:
		[switched := self handleStackOverflowOrEventAllowContextSwitch: mayContextSwitch].
	self returnToExecutive: inInterpreter postContextSwitch: switched! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem 12/7/2017 16:32'!
activateNewMethod
	| methodHeader numArgs numTemps rcvr inInterpreter switched |

	methodHeader := objectMemory methodHeaderOf: newMethod.
	numTemps := self temporaryCountOfMethodHeader: methodHeader.
	numArgs := self argumentCountOfMethodHeader: methodHeader.

	rcvr := self stackValue: numArgs. "could new rcvr be set at point of send?"
	self assert: (objectMemory isOopForwarded: rcvr) not.

	"Because this is an uncogged method we need to continue via the interpreter.
	 We could have been reached either from the interpreter, in which case we
	 should simply return, or from a machine code frame or from a compiled
	 primitive.  In these latter two cases we must longjmp back to the interpreter.
	 The instructionPointer tells us which path we took.
	 If the sender was an interpreter frame but called through a (failing) primitive
	 then make sure we restore the saved instruction pointer and avoid pushing
	 ceReturnToInterpreterPC which is only valid between an interpreter caller
	 frame and a machine code callee frame."
	(inInterpreter := instructionPointer >= objectMemory startOfMemory) ifFalse:
		[instructionPointer = cogit ceReturnToInterpreterPC ifTrue:
			[instructionPointer := self iframeSavedIP: framePointer]].
	self push: instructionPointer.
	self push: framePointer.
	framePointer := stackPointer.
	self push: newMethod.
	self setMethod: newMethod methodHeader: methodHeader.
	self push: objectMemory nilObject. "FxThisContext field"
	self push: (self encodeFrameFieldHasContext: false isBlock: false numArgs: numArgs).
	self push: 0. "FoxIFSavedIP"
	self push: rcvr.

	"clear remaining temps to nil"
	numArgs+1 to: numTemps do:
		[:i | self push: objectMemory nilObject].

	instructionPointer := (self initialIPForHeader: methodHeader method: newMethod) - 1.

	(self methodHeaderHasPrimitive: methodHeader) ifTrue:
		["Skip the CallPrimitive bytecode, if it's there, and store the error code if the method starts
		  with a long store temp.  Strictly no need to skip the store because it's effectively a noop."
		 instructionPointer := instructionPointer + (self sizeOfCallPrimitiveBytecode: methodHeader).
		 primFailCode ~= 0 ifTrue:
			[self reapAndResetErrorCodeTo: stackPointer header: methodHeader]].

	"Now check for stack overflow or an event (interrupt, must scavenge, etc)."
	switched := true.
	stackPointer < stackLimit ifTrue:
		[switched := self handleStackOverflowOrEventAllowContextSwitch:
							(self canContextSwitchIfActivating: newMethod header: methodHeader)].
	self returnToExecutive: inInterpreter postContextSwitch: switched! !

!CoInterpreter methodsFor: 'method lookup cache' stamp: 'eem 10/9/2013 16:01'!
addNewMethodToCache: classObj
	"Override to refuse to cache other than compiled methods.
	 This protects open PICs against having to test for compiled methods."
	(objectMemory isOopCompiledMethod: newMethod) ifFalse:
		[primitiveFunctionPointer := #primitiveInvokeObjectAsMethod.
		^self].
	super addNewMethodToCache: classObj! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 1/27/2009 12:57'!
argumentCount
	<doNotGenerate>
	^argumentCount! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 1/3/2009 19:42'!
argumentCount: numArgs
	<doNotGenerate>
	argumentCount := numArgs! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'eem 5/6/2010 16:00'!
argumentCountAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: argumentCount) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #argumentCount in: self]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 1/9/2011 11:46'!
asCogHomeMethod: aCogMethod
	"Coerce either a CMMethod or a CMBlock to the home CMMethod"
	<var: #aCogMethod type: #'CogBlockMethod *'>
	<returnTypeC: #'CogMethod *'>
	^aCogMethod cmType = CMMethod
		ifTrue: [self cCoerceSimple: aCogMethod to: #'CogMethod *']
		ifFalse: [aCogMethod cmHomeMethod]! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 3/6/2015 13:42'!
assertValidExecutionPointe: lip r: lifp s: lisp imbar: inInterpreter line: ln
	<var: #lip type: #usqInt>
	<var: #lifp type: #'char *'>
	<var: #lisp type: #'char *'>
	| methodField cogMethod theIP  |
	<var: #cogMethod type: #'CogMethod *'>
	self assert: stackPage = stackPages mostRecentlyUsedPage l: ln.
	self assert: (stackPage addressIsInPage: lifp) l: ln.
	self assert: (self deferStackLimitSmashAround: #assertValidStackLimits: asSymbol with: ln).
	self assert: lisp < lifp l: ln.
	self assert: lifp > lisp l: ln.
	self assert: lisp >= (stackPage realStackLimit - self stackLimitOffset) l: ln.
	self assert: (lifp - lisp) / objectMemory bytesPerOop < LargeContextSlots l: ln.
	methodField := self frameMethodField: lifp.
	inInterpreter
		ifTrue:
			[self assert: (self isMachineCodeFrame: lifp) not l: ln.
			 self assert: method = methodField l: ln.
			 self cppIf: MULTIPLEBYTECODESETS
				ifTrue: [self assert: (self methodUsesAlternateBytecodeSet: method) = (bytecodeSetSelector = 256) l: ln].
			 (self asserta: (objectMemory cheapAddressCouldBeInHeap: methodField) l: ln) ifTrue:
				[theIP := lip = cogit ceReturnToInterpreterPC
							ifTrue: [self iframeSavedIP: lifp]
							ifFalse: [lip].
				 self assert: (theIP >= (methodField + (objectMemory lastPointerOf: methodField))
							  and: [theIP < (methodField + (objectMemory numBytesOfBytes: methodField) + objectMemory baseHeaderSize - 1)])
					l: ln].
			 self assert: ((self iframeIsBlockActivation: lifp)
					or: [(self pushedReceiverOrClosureOfFrame: lifp) = (self iframeReceiver: lifp)])
				l: ln]
		ifFalse:
			[self assert: (self isMachineCodeFrame: lifp) l: ln.
			 ((self asserta: methodField asUnsignedInteger >= cogit minCogMethodAddress l: ln)
			  and: [self asserta: methodField asUnsignedInteger < cogit maxCogMethodAddress l: ln]) ifTrue:
				[cogMethod := self mframeHomeMethod: lifp.
				 self assert: (lip > (methodField + ((self mframeIsBlockActivation: lifp)
													ifTrue: [self sizeof: CogBlockMethod]
													ifFalse: [self sizeof: CogMethod]))
						and: [lip < (methodField + cogMethod blockSize)])
					l: ln].
			 self assert: ((self mframeIsBlockActivation: lifp)
					or: [(self pushedReceiverOrClosureOfFrame: lifp) = (self mframeReceiver: lifp)])
				l: ln].
	(self isBaseFrame: lifp) ifTrue:
		[self assert: (self frameHasContext: lifp) l: ln.
		 self assert: (self frameContext: lifp) = (stackPages longAt: stackPage baseAddress - objectMemory wordSize) l: ln]! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 7/25/2021 12:01'!
assertValidExternalStackPointers
	<doNotGenerate>
	"For use *ONLY* by routines coming in to the VM,
	 i.e. handleCallOrJumpSimulationTrap:.  This is because it nils localFP as a side-effect,
	 and it does so so that the head frame can be determined reliably."
	self assert: framePointer < stackPage baseAddress.
	self assert: stackPointer < framePointer.
	self assert: framePointer > stackPointer.
	self assert: stackPointer >= (stackPage realStackLimit - self stackLimitOffset).
	self nilLocalFP! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 11/24/2010 14:23'!
assertValidMachineCodeFrame: instrPtr
	<api>
	| cogMethod homeMethod |
	<var: #cogMethod type: #'CogBlockMethod *'>
	<var: #homeMethod type: #'CogMethod *'>
	self assert: (self isMachineCodeFrame: framePointer).
	cogMethod := self mframeCogMethod: framePointer.
	homeMethod := self asCogHomeMethod: cogMethod.
	self assert: (cogMethodZone methodFor: cogMethod) = homeMethod.
	self assert: (instrPtr > cogMethod asInteger
				and: [instrPtr < (homeMethod asInteger + homeMethod blockSize)])! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 1/5/2009 14:16'!
assertValidStackPageHeadPointers
	self assert: stackPage headFP < stackPage baseAddress.
	self assert: stackPage headSP < stackPage headFP.
	self assert: stackPage headFP > stackPage headSP.
	self assert: stackPage headSP >= (stackPage realStackLimit - self stackLimitOffset)! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 7/23/2013 15:44'!
assertValidStackedInstructionPointers: ln
	"Check that the stacked instruction pointers in all pages are correct.
	 Checks the interpreter sender/machine code callee contract.
	 Written so it will be optimized away if not in an assert VM."
	| thePage |
	<inline: false>
	<var: #thePage type: #'StackPage *'>
	0 to: numStackPages - 1 do:
		[:i|
		thePage := stackPages stackPageAt: i.
		(stackPages isFree: thePage) ifFalse:
			[self assert: (self assertValidStackedInstructionPointersIn: thePage line: ln) l: ln]]! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 3/6/2015 13:42'!
assertValidStackedInstructionPointersIn: aStackPage line: ln
	"Check that the stacked instruction pointers in the given page are correct.
	 Checks the interpreter sender/machine code callee contract."
	<var: #aStackPage type: #'StackPage *'>
	<var: #theFP type: #'char *'>
	<var: #callerFP type: #'char *'>
	<var: #theIP type: #usqInt>
	<var: #theMethod type: #'CogMethod *'>
	<inline: false>
	| prevFrameWasCogged theFP callerFP theMethod theIP methodObj |
	(self asserta: (stackPages isFree: aStackPage) not l: ln) ifFalse:
		[^false].
	prevFrameWasCogged := false.
	"The top of stack of an inactive page is always the instructionPointer.
	 The top of stack of the active page may be the instructionPointer if it has been pushed,
	 which is indicated by a 0 instructionPointer."
	(stackPage = aStackPage and: [instructionPointer ~= 0])
		ifTrue:
			[theIP := instructionPointer.
			theFP := framePointer]
		ifFalse:
			[theIP := (stackPages longAt: aStackPage headSP) asUnsignedInteger.
			 theFP := aStackPage headFP.
			 stackPage = aStackPage ifTrue:
				[self assert: framePointer = theFP l: ln]].
	[(self isMachineCodeFrame: theFP)
		ifTrue:
			[theMethod := self mframeHomeMethod: theFP.
			 self assert: (theIP = cogit ceCannotResumePC
						  or: [theIP >= theMethod asUnsignedInteger
							   and: [theIP < (theMethod asUnsignedInteger + theMethod blockSize)]])
					l: ln.
			prevFrameWasCogged := true]
		ifFalse: "assert-check the interpreter frame."
			[methodObj := self iframeMethod: theFP.
			 prevFrameWasCogged ifTrue:
				[self assert: theIP = cogit ceReturnToInterpreterPC l: ln].
			 theIP = cogit ceReturnToInterpreterPC ifTrue:
				[theIP := self iframeSavedIP: theFP].
			 self assert: (theIP >= (methodObj + (objectMemory lastPointerOf: methodObj))
						  and: [theIP < (methodObj + (objectMemory numBytesOfBytes: methodObj) + objectMemory baseHeaderSize - 1)])
				l: ln.
			 prevFrameWasCogged := false].
	 theIP := (stackPages longAt: theFP + FoxCallerSavedIP) asUnsignedInteger.
	 (callerFP := self frameCallerFP: theFP) ~= 0] whileTrue:
		[theFP := callerFP].
	self assert: theIP = cogit ceBaseFrameReturnPC l: ln.
	^true! !

!CoInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 8/2/2021 21:49'!
attemptToLinkExternalPrimitive: methodObj
	| header primIdx firstLiteral targetFunctionIndex |
	header := objectMemory methodHeaderOf: methodObj.
	primIdx := self primitiveIndexOfMethod: methodObj header: header.
	firstLiteral := self literal: 0 ofMethod: methodObj.
	self assert: (primIdx = PrimNumberExternalCall
				 and: [(objectMemory literalCountOfMethodHeader: header) > 0
				 and: [(objectMemory isArray: firstLiteral)
				 and: [(objectMemory numSlotsOf: firstLiteral) = 4]]]).
	 targetFunctionIndex := objectMemory fetchPointer: ExternalCallLiteralTargetFunctionIndex ofObject: firstLiteral.
	 self assert: (objectMemory isIntegerObject: targetFunctionIndex).
	 (objectMemory integerValueOf: targetFunctionIndex) = 0 ifTrue:
		[self linkExternalCall: firstLiteral errInto: nil].
	^firstLiteral! !

!CoInterpreter methodsFor: 'jump bytecodes' stamp: 'cb 9/1/2016 17:18'!
attemptToSwitchToMachineCode: bcpc
	"Attempt to convert the current interpreted activation into a machine code
	 activation, and if this is popssible, jump into machine code.  bcpc is the
	 0-relative pc of the backward branch bytecode (not any preceding extension)."
	| cogMethod pc cls |
	<inline: false>
	<var: #cogMethod type: #'CogMethod *'>
	(self methodHasCogMethod: method)
		ifFalse: 
			[SistaV1BytecodeSet
				ifTrue: 
					[ ((self iframeIsBlockActivation: framePointer) 
					   and: [(self isVanillaBlockClosure: (cls := self frameStackedReceiver: framePointer numArgs: (self frameNumArgs: framePointer))) not ])
						ifTrue: 
							["Compiled block / full closure"
							 cogit cogFullBlockMethod: method numCopied: (objectMemory numPointerSlotsOf: cls) - FullClosureFirstCopiedValueIndex]
						ifFalse: 
							["Compiled method"
							cogit cog: method selector: objectMemory nilObject] ]
				ifFalse: 
					[cogit cog: method selector: objectMemory nilObject] ].
	(self methodHasCogMethod: method) ifTrue:
		[cogMethod := self cogMethodOf: method.
		 pc := self convertToMachineCodeFrame: cogMethod bcpc: bcpc.
		 self assertValidMachineCodeFrame: pc.
		 self push: pc.
		 self push: objectMemory nilObject.
		 cogit ceEnterCogCodePopReceiverReg]! !

!CoInterpreter methodsFor: 'return bytecodes' stamp: 'eem 8/14/2015 12:15'!
baseFrameReturn
	"Return from a baseFrame (the bottom frame in a stackPage).  The context to
	 return to (which may be married) is stored in the first word of the stack."
	<inline: true>
	| contextToReturnTo retToContext theFP theSP thePage newPage frameAbove |
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<var: #newPage type: #'StackPage *'>
	<var: #frameAbove type: #'char *'>
	contextToReturnTo := self frameCallerContext: localFP.

	"The stack page is effectively free now, so free it.  We must free it to be
	 correct in determining if contextToReturnTo is still married, and in case
	 makeBaseFrameFor: cogs a method, which may cause a code compaction,
	 in which case the frame must be free to avoid the relocation machinery
	 tracing the dead frame.  Since freeing now temporarily violates the page-list
	 ordering invariant, use the assert-free version."
	stackPages freeStackPageNoAssert: stackPage.
	retToContext := objectMemory isContext: contextToReturnTo.
	(retToContext
	 and: [self isStillMarriedContext: contextToReturnTo])
		ifTrue:
			[theFP := self frameOfMarriedContext: contextToReturnTo.
			 thePage := stackPages stackPageFor: theFP.
			 theFP = thePage headFP
				ifTrue:
					[theSP := thePage headSP]
				ifFalse:
					["Returning to some interior frame, presumably because of a sender assignment.
					  Move the frames above to another page (they may be in use, e.g. via coroutining).
					  Make the interior frame the top frame."
					 frameAbove := self findFrameAbove: theFP inPage: thePage.
					 "Since we've just deallocated a page we know that newStackPage won't deallocate an existing one."
					 newPage := stackPages newStackPage.
					 self assert: newPage = stackPage.
					 self moveFramesIn: thePage through: frameAbove toPage: newPage.
					 stackPages markStackPageMostRecentlyUsed: newPage.
					 theFP := thePage headFP.
					 theSP := thePage headSP]]
		ifFalse:
			[(retToContext
			  and: [objectMemory isIntegerObject: (objectMemory fetchPointer: InstructionPointerIndex ofObject: contextToReturnTo)]) ifFalse:
				[| contextToReturnFrom |
				 contextToReturnFrom := stackPages longAt: stackPage baseAddress - objectMemory wordSize.
				 self tearDownAndRebuildFrameForCannotReturnBaseFrameReturnFrom: contextToReturnFrom
					to: contextToReturnTo
					returnValue: localReturnValue.
				^self externalCannotReturn: localReturnValue from: contextToReturnFrom].
			 "We must void the instructionPointer to stop it being updated if makeBaseFrameFor:
			  cogs a method, which may cause a code compaction."
			 instructionPointer := 0.
			 thePage := self makeBaseFrameFor: contextToReturnTo.
			 theFP := thePage headFP.
			 theSP := thePage headSP].
	self setStackPageAndLimit: thePage.
	self assert: (stackPages stackPageFor: theFP) = stackPage.
	localSP := theSP.
	localFP := theFP.
	localIP := self pointerForOop: self internalStackTop.
	localIP asUnsignedInteger < objectMemory startOfMemory ifTrue:
		[localIP asUnsignedInteger ~= cogit ceReturnToInterpreterPC ifTrue:
			["localIP in the cog method zone indicates a return to machine code."
			 ^self returnToMachineCodeFrame].
		 localIP := self pointerForOop: (self iframeSavedIP: localFP)].
	self assert: (self checkIsStillMarriedContext: contextToReturnTo currentFP: localFP).
	self setMethod: (self iframeMethod: localFP).
	self internalStackTopPut: localReturnValue.
	^self fetchNextBytecode! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 6/30/2014 06:03'!
bytecodePCFor: theIP cogMethod: cogMethod startBcpc: startBcpc
	"Answer the mapping of the native pc theIP to a zero-relative bytecode pc.
	 See contextInstructionPointer:frame: for the explanation."
	<var: #cogMethod type: #'CogMethod *'>
	| cogMethodForIP mcpc |
	<inline: true>
	<var: #cogMethodForIP type: #'CogBlockMethod *'>
	self assert: theIP < 0.
	(theIP signedBitShift: -16) < -1 "See contextInstructionPointer:frame:"
		ifTrue:
			[cogMethodForIP := self cCoerceSimple: cogMethod asInteger - ((theIP signedBitShift: -16) * cogit blockAlignment)
									to: #'CogBlockMethod *'.
			 self assert: cogMethodForIP cmType = CMBlock.
			 self assert: cogMethodForIP cmHomeMethod = cogMethod.
			 mcpc := cogMethodForIP asInteger - theIP signedIntFromShort]
		ifFalse:
			[cogMethodForIP := self cCoerceSimple: cogMethod to: #'CogBlockMethod *'.
			 self assert: cogMethodForIP cmType = CMMethod.
			 mcpc := cogMethod asInteger - theIP.
			 "map any pcs in primitive code (i.e. return addresses for interpreter primitive calls) to the initial pc"
			 mcpc asUnsignedInteger < cogMethod stackCheckOffset ifTrue:
				[^startBcpc]].
	self assert: (mcpc between: cogMethod asInteger and: cogMethod asInteger + cogMethod blockSize).
	^cogit bytecodePCFor: mcpc startBcpc: startBcpc in: cogMethodForIP! !

!CoInterpreter methodsFor: 'common selector sends' stamp: 'eem 1/24/2016 11:31'!
bytecodePrimAt
	"Override to eliminate the atCache, something of little benefit to the JIT."
	messageSelector := self specialSelector: 16.
	argumentCount := 1.
	self normalSend! !

!CoInterpreter methodsFor: 'common selector sends' stamp: 'eem 1/24/2016 11:31'!
bytecodePrimAtPut
	"Override to eliminate the atCache, something of little benefit to the JIT."
	messageSelector := self specialSelector: 17.
	argumentCount := 2.
	self normalSend! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'eem 2/2/2020 18:56'!
cFramePointerAddress
	<api>
	<returnTypeC: #usqInt>
	^self
		cCode: [(self addressOf: CFramePointer) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #getCFramePointer in: self]! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'eem 3/31/2020 18:53'!
cReturnAddressAddress
	<api>
	<returnTypeC: #usqInt>
	^self
		cCode: [(self addressOf: CReturnAddress) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #getCReturnAddress in: self]! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'eem 2/2/2020 18:56'!
cStackPointerAddress
	<api>
	<returnTypeC: #usqInt>
	^self
		cCode: [(self addressOf: CStackPointer) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #getCStackPointer in: self]! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 3/6/2009 19:48'!
callForCogCompiledCodeCompaction
	<api>
	cogCompiledCodeCompactionCalledFor := true.
	self forceInterruptCheck! !

!CoInterpreter methodsFor: 'enilopmarts' stamp: 'eem 10/7/2014 10:37'!
callRegisterArgCogMethod: cogMethod at: entryOffset receiver: rcvr
	"convert
	 		rcvr	base
			arg(s)
			retpc	<- sp
	 to
			retpc	base
			entrypc
			rcvr
			arg(s)	<- sp
	 and then enter at either the checked or the unchecked entry-point."
	<option: #StackToRegisterMappingCogit>
	<var: #cogMethod type: #'CogMethod *'>
	self assert: (cogit numRegArgs > 0 and: [cogit numRegArgs <= 2 and: [cogMethod cmNumArgs <= cogit numRegArgs]]).
	cogMethod cmNumArgs = 2 ifTrue:
		[self stackValue: 3 put: self stackTop. "retpc"
		 self push: (self stackValue: 1). "last arg"
		 self stackValue: 1 put: (self stackValue: 3). "first arg"
		 self stackValue: 2 put: rcvr.
		 self stackValue: 3 put: cogMethod asInteger + entryOffset.
		 cogit ceCallCogCodePopReceiverArg1Arg0Regs
		"NOTREACHED"].
	cogMethod cmNumArgs = 1 ifTrue:
		[self stackValue: 2 put: self stackTop. "retpc"
		 self push: (self stackValue: 1). "arg"
		 self stackValue: 1 put: rcvr.
		 self stackValue: 2 put: cogMethod asInteger + entryOffset.
		 cogit ceCallCogCodePopReceiverArg0Regs
		"NOTREACHED"].
	self assert: cogMethod cmNumArgs = 0.
	self stackValue: 1 put: self stackTop. "retpc"
	self stackValue: 0 put: cogMethod asInteger + entryOffset.
	self push: rcvr.
	cogit ceCallCogCodePopReceiverReg
	"NOTREACHED"! !

!CoInterpreter methodsFor: 'callback support' stamp: 'eem 9/1/2020 14:27'!
callbackEnter: callbackID
	"Re-enter the interpreter for executing a callback"
	| currentCStackPointer currentCFramePointer wasInMachineCode calledFromMachineCode |
	<volatile>
	<export: true>
	<var: #currentCStackPointer type: #usqIntptr_t>
	<var: #currentCFramePointer type: #usqIntptr_t>
	<var: #callbackID type: #'sqInt *'>

	"For now, do not allow a callback unless we're in a primitiveResponse"
	(self asserta: primitiveFunctionPointer ~= 0) ifFalse:
		[^false].

	self assert: primFailCode = 0.

	"Check if we've exceeded the callback depth"
	(self asserta: jmpDepth < MaxJumpBuf) ifFalse:
		[^false].
	jmpDepth := jmpDepth + 1.

	wasInMachineCode := self isMachineCodeFrame: framePointer.
	calledFromMachineCode := instructionPointer <= objectMemory startOfMemory.

	"Suspend the currently active process"
	suspendedCallbacks at: jmpDepth put: self activeProcess.
	"We need to preserve newMethod explicitly since it is not activated yet
	and therefore no context has been created for it. If the caller primitive
	for any reason decides to fail we need to make sure we execute the correct
	method and not the one 'last used' in the call back"
	suspendedMethods at: jmpDepth put: newMethod.
	self flag: 'need to debug this properly.  Conceptually it is the right thing to do but it crashes in practice'.
	false
		ifTrue:
			["Signal external semaphores since a signalSemaphoreWithIndex: request may
			  have been issued immediately prior to this callback before the VM has any
			  chance to do a signalExternalSemaphores in checkForEventsMayContextSwitch:"
			 self signalExternalSemaphores.
			 "If no process is awakened by signalExternalSemaphores then transfer
			  to the highest priority runnable one."
			 (suspendedCallbacks at: jmpDepth) = self activeProcess ifTrue:
				[self transferTo: self wakeHighestPriority from: CSCallbackLeave]]
		ifFalse:
			[self transferTo: self wakeHighestPriority from: CSCallbackLeave].

	"Typically, invoking the callback means that some semaphore has been 
	signaled to indicate the callback. Force an interrupt check as soon as possible."
	self forceInterruptCheck.

	"Save the previous CStackPointers..."
	currentCStackPointer := CStackPointer.
	currentCFramePointer := CFramePointer.
	cogit assertCStackWellAligned.
	(self _setjmp: (jmpBuf at: jmpDepth)) = 0 ifTrue: "Fill in callbackID"
		[callbackID at: 0 put: jmpDepth.
		 self enterSmalltalkExecutive.
		 self assert: false "NOTREACHED"].

	"Restore the previous CStackPointers..."
	self setCFramePointer: currentCFramePointer setCStackPointer: currentCStackPointer.

	"Transfer back to the previous process so that caller can push result"
	self putToSleep: self activeProcess yieldingIf: preemptionYields.
	self transferTo: (suspendedCallbacks at: jmpDepth) from: CSCallbackLeave.
	newMethod := suspendedMethods at: jmpDepth.	"see comment above"
	argumentCount := self argumentCountOf: newMethod.
	self assert: wasInMachineCode = (self isMachineCodeFrame: framePointer).
	calledFromMachineCode
		ifTrue:
			[instructionPointer asUnsignedInteger >= objectMemory startOfMemory ifTrue:
				[self iframeSavedIP: framePointer put: instructionPointer.
				 instructionPointer := cogit ceReturnToInterpreterPC]]
		ifFalse:
			["Even if the context was flushed to the heap and rebuilt in transferTo:from:
			  above it will remain an interpreted frame because the context's pc would
			  remain a bytecode pc.  So the instructionPointer must also be a bytecode pc."
			 self assert: (self isMachineCodeFrame: framePointer) not.
			 self assert: instructionPointer > objectMemory startOfMemory].
	self assert: primFailCode = 0.
	jmpDepth := jmpDepth-1.
	^true! !

!CoInterpreter methodsFor: 'enilopmarts' stamp: 'eem 8/8/2021 10:33'!
ceActivateFailingPrimitiveMethod: aPrimitiveMethod
	"An external call or FFI primitive has failed.  Build the frame and
	 activate as appropriate.  Enter either the interpreter or machine
	 code depending on whether aPrimitiveMethod has been or is still
	 cogged.  Note that we could always interpret but want the efficiency
	 of executing machine code if it is available."
	<api>
	| methodHeader result |
	self assert: primFailCode ~= 0.
	self assert: newMethod = aPrimitiveMethod.
	"If we're on Spur, retry the primitive, if appropriate,
	 returning if successful after retry."
	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[self retryPrimitiveOnFailure.
		 self successful ifTrue:
			[(profileSemaphore ~= objectMemory nilObject
			  and: [self ioHighResClock >= nextProfileTick]) ifTrue:
				[self ceTakeProfileSample: nil].
			 result := self stackTop.
			 self stackTopPut: instructionPointer.
			 self push: result.
			 cogit ceEnterCogCodePopReceiverReg]].
	methodHeader := self rawHeaderOf: aPrimitiveMethod.
	(self isCogMethodReference: methodHeader)
		ifTrue: [self activateNewCogMethod: (self cCoerceSimple: methodHeader to: #'CogMethod *') inInterpreter: false]
		ifFalse: [self activateNewMethod]! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 1/20/2015 15:08'!
ceActiveContext
	<api>
	<option: #SqueakV3ObjectMemory>
	"Since the trampoline checks for marriage we should only be here for a single frame."
	self assert: (self isMachineCodeFrame: framePointer).
	self assert: (self frameHasContext: framePointer) not.
	"Do *not* include the return pc in the stack contents; hence + BytesPerWord"
	^self marryFrame: framePointer SP: stackPointer + objectMemory wordSize! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 4/1/2020 14:20'!
ceBaseFrameReturn: returnValue
	"Return across a page boundary.  The context to return to (which may be married)
	 is stored in the first word of the stack.  We get here when a return instruction jumps
	 to the ceBaseFrameReturn: address that is the return pc for base frames.  A consequence
	 of this is that the current frame is no longer valid since an interrupt may have overwritten
	 its state as soon as the stack pointer has been cut-back beyond the return pc.  So to have
	 a context to send the cannotReturn: message to we also store the base frame's context
	 in the second word of the stack page."
	<api>
	| contextToReturnTo contextToReturnFrom isAContext thePage newPage frameAbove |
	<var: #thePage type: #'StackPage *'>
	<var: #newPage type: #'StackPage *'>
	<var: #frameAbove type: #'char *'>
	self assert: (stackPages stackPageFor: stackPointer) = stackPage.
	self assert: stackPages mostRecentlyUsedPage = stackPage.
	cogit assertCStackWellAligned.
	self assert: framePointer = 0.
	self assert: stackPointer <= (stackPage baseAddress - objectMemory wordSize).
	self assert: stackPage baseFP + (2 * objectMemory wordSize) < stackPage baseAddress.
	"We would like to use the following assert but we can't since the stack pointer will be above the
	 base frame pointer in the base frame return and hence the 0 a base frame pointer points at could
	 be overwritten which will cause the isBaseFrame assert in frameCallerContext: to fail."
	"self assert: (self frameCallerContext: stackPage baseFP) = (stackPages longAt: stackPage baseAddress)."
	self assert: ((objectMemory addressCouldBeObj: (stackPages longAt: stackPage baseAddress - objectMemory wordSize))
				and: [objectMemory isContext: (stackPages longAt: stackPage baseAddress - objectMemory wordSize)]).
	contextToReturnTo := stackPages longAt: stackPage baseAddress.
	self assert: (objectMemory addressCouldBeObj: contextToReturnTo).

	"The stack page is effectively free now, so free it.  We must free it to be
	 correct in determining if contextToReturnTo is still married, and in case
	 makeBaseFrameFor: cogs a method, which may cause a code compaction,
	 in which case the frame must be free to avoid the relocation machinery
	 tracing the dead frame.  Since freeing now temporarily violates the page-list
	 ordering invariant, use the assert-free version."
	stackPages freeStackPageNoAssert: stackPage.
	isAContext := objectMemory isContext: contextToReturnTo.
	(isAContext
	 and: [self isStillMarriedContext: contextToReturnTo])
		ifTrue:
			[framePointer := self frameOfMarriedContext: contextToReturnTo.
			 thePage := stackPages stackPageFor: framePointer.
			 framePointer = thePage headFP
				ifTrue:
					[stackPointer := thePage headSP]
				ifFalse:
					["Returning to some interior frame, presumably because of a sender assignment.
					  Move the frames above to another page (they may be in use, e.g. via coroutining).
					  Make the interior frame the top frame."
					 frameAbove := self findFrameAbove: framePointer inPage: thePage.
					 "Since we've just deallocated a page we know that newStackPage won't deallocate an existing one."
					 newPage := stackPages newStackPage.
					 self assert: newPage = stackPage.
					 self moveFramesIn: thePage through: frameAbove toPage: newPage.
					 stackPages markStackPageMostRecentlyUsed: newPage.
					 self setStackPointersFromPage: thePage]]
		ifFalse:
			[(isAContext
			  and: [objectMemory isIntegerObject: (objectMemory fetchPointer: InstructionPointerIndex ofObject: contextToReturnTo)]) ifFalse:
				[contextToReturnFrom := stackPages longAt: stackPage baseAddress - objectMemory wordSize.
				 self tearDownAndRebuildFrameForCannotReturnBaseFrameReturnFrom: contextToReturnFrom
					to: contextToReturnTo
					returnValue: returnValue.
				^self externalCannotReturn: returnValue from: contextToReturnFrom].
			 "void the instructionPointer to stop it being incorrectly updated in a code
			 compaction in makeBaseFrameFor:."
			 instructionPointer := 0.
			 thePage := self makeBaseFrameFor: contextToReturnTo.
			 self setStackPointersFromPage: thePage].
	self setStackPageAndLimit: thePage.
	self assert: (stackPages stackPageFor: framePointer) = stackPage.
	(self isMachineCodeFrame: framePointer) ifTrue:
		[self push: returnValue.
		 cogit ceEnterCogCodePopReceiverReg.
		 "NOTREACHED"].
	instructionPointer := self stackTop.
	instructionPointer = cogit ceReturnToInterpreterPC ifTrue:
		[instructionPointer := self iframeSavedIP: framePointer].
	self setMethod: (self iframeMethod: framePointer).
	self stackTopPut: returnValue. "a.k.a. pop saved ip then push result"
	self assert: (self checkIsStillMarriedContext: contextToReturnTo currentFP: framePointer).
	cogit ceInvokeInterpret.
	"NOTREACHED"
	^nil! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'cb 1/31/2017 14:40'!
ceCannotAssignTo: immutableObject withIndex: index valueToAssign: valueToAssign
	"index is unboxed and 0-based. The call-back expects 1-based value (to perform the operation with instVarAt:put:"
	<api>
	<option: #IMMUTABILITY>
	instructionPointer := self popStack.
	self push: immutableObject.
	self push: valueToAssign.
	self push: (objectMemory integerObjectOf: index + 1).
	self push: instructionPointer.
	^ self
		ceSendAbort: (objectMemory splObj: SelectorAttemptToAssign)
		to: immutableObject
		numArgs: 2! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 3/11/2011 10:47'!
ceCannotResume
	<api>
	"A context that has been returned from, or otherwise has an invalid pc has been reentered.
	 Until we have a cannotResume: selector, simply resend cannotReturn:."
	| resultOop |
	self assert: (self isMachineCodeFrame: framePointer).
	self assert: (self frameHasContext: framePointer).
	resultOop := self stackTop.
	self push: (self frameContext: framePointer).
	self push: resultOop.
	"Make sure the happy couple remain returned from."
	self push: cogit ceCannotResumePC.
	^self
		ceSendAbort: (objectMemory splObj: SelectorCannotReturn)
		to: (self frameContext: framePointer)
		numArgs: 1! !

!CoInterpreter methodsFor: 'primitive support' stamp: 'eem 6/20/2021 19:38'!
ceCheckAndMaybeRetryPrimitive: primIndex
	"Log failure and then retry if there's an accessorDepth or failure due to no memory."
	<api>
	<option: #SpurObjectMemory>
	| retried |
	(cogit recordPrimTrace and: [primTracePluginName isNil]) ifTrue:
		[self fastLogPrim: TracePrimitiveFailure].
	retried := self retryPrimitiveOnFailure.
	(retried and: [cogit recordPrimTrace]) ifTrue:
		[self fastLogPrim: TracePrimitiveRetry]! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 12/1/2020 17:56'!
ceCheckForInterrupt
	<api>
	| switched |
	self cCode: [] inSmalltalk:
		[self maybeCheckStackDepth: 0 sp: stackPointer pc: instructionPointer].
	switched := self checkForEventsMayContextSwitch: true.
	self returnToExecutive: false postContextSwitch: switched! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 9/10/2013 22:49'!
ceContext: maybeContext instVar: slotIndex
	<api>
	| result |
	(objectMemory isContextNonImm: maybeContext)
		ifTrue:
			[instructionPointer := self popStack.
			 result := self externalInstVar: slotIndex ofContext: maybeContext.
			 self push: instructionPointer]
		ifFalse: [result := objectMemory fetchPointer: slotIndex ofObject: maybeContext].
	^result! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'cb 8/31/2016 13:42'!
ceContext: maybeMarriedContext instVar: slotIndex value: anOop
	<api>
	"genStorePop:MaybeContextReceiverVariable: filters out unmarried contexts
	 but not arbitrary objects in subclasses.  It answers maybeMarriedContext so
	 that the StackToRegisterMappingCogit can keep ReceiverResultReg live."
	((objectMemory isContextNonImm: maybeMarriedContext) 
	 and: [self isMarriedOrWidowedContext: maybeMarriedContext])
		ifTrue:
			[instructionPointer := self popStack.
			 self externalInstVar: slotIndex ofContext: maybeMarriedContext put: anOop.
			 self push: instructionPointer]
		ifFalse:
			[objectMemory storePointer: slotIndex ofObject: maybeMarriedContext withValue: anOop].
	^maybeMarriedContext! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 3/16/2017 16:01'!
ceCounterTripped: condition
	"Two things are going on here.  The main one is catching a counter trip and attempting
	 to send the SelectorCounterTripped selector.  In this case we would like to back-up
	 the pc to the return address of the send that yields the boolean to be tested, so that
	 after potential optimization, computation proceeds by retrying the jump.  But we cannot,
	 since there may be no send, just a pop (as in and: [] and or: [] chains).  In this case we also
	 want to prevent further callbacks until optimization is complete.  So we nil-out the
	 SelectorCounterTripped entry in the specialSelectorArray.

	 The minor case is that there is an unlikely  possibility that the cointer tripped but condition
	 is not a boolean, in which case a mustBeBoolean response should occur."
	<api>
	<option: #SistaCogit>
	"Send e.g. thisContext conditionalBranchCounterTrippedOn: boolean."
	| context counterTrippedSelector classTag classObj |
	(condition = objectMemory falseObject
	or: [condition = objectMemory trueObject]) ifFalse:
		[^self ceSendMustBeBoolean: condition].

	counterTrippedSelector := objectMemory maybeSplObj: SelectorCounterTripped.
	(counterTrippedSelector isNil
	or: [counterTrippedSelector = objectMemory nilObject]) ifTrue:
		[cogit resetCountersIn: (self mframeHomeMethod: framePointer).
		 ^condition].

	classTag := objectMemory
					classTagForSpecialObjectsIndex: ClassMethodContext
					compactClassIndex: ClassMethodContextCompactIndex.
	(self lookupInMethodCacheSel: counterTrippedSelector classTag: classTag) ifFalse:
	 	[messageSelector := counterTrippedSelector.
		 classObj := objectMemory classForClassTag: classTag.
		 (self lookupOrdinaryNoMNUEtcInClass: classObj) ~= 0 ifTrue:
			[cogit resetCountersIn: (self mframeHomeMethod: framePointer).
			 ^condition]].

	(primitiveFunctionPointer ~= 0
	or: [(self argumentCountOf: newMethod) ~= 1]) ifTrue:
		[cogit resetCountersIn: (self mframeHomeMethod: framePointer).
		 ^condition].

	cogit setCogCodeZoneThreshold: 1.0.
	objectMemory splObj: SelectorCounterTripped put: objectMemory nilObject.
	instructionPointer := self popStack.
	context := self ensureFrameIsMarried: framePointer SP: stackPointer.
	self push: context.
	self push: condition.
	self ifAppropriateCompileToNativeCode: newMethod selector: counterTrippedSelector.
	self activateNewMethod.
	"not reached"
	^true! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'rmacnak 8/24/2015 22:45'!
ceDynamicSuperSend: cacheAddress receiver: methodReceiver
	"A dynamic super send cache missed."
	| nsSendCache methodReceiverClassTag cogMethod errSelIdx |
	<api>
	<option: #NewspeakVM>
	<inline: false>
	<var: #nsSendCache type: #'NSSendCache *'>
	<var: #cogMethod type: #'CogMethod *'>

	cogit assertCStackWellAligned.
	self assert: (objectMemory addressCouldBeOop: methodReceiver).
	self deny: (objectMemory isOopForwarded: methodReceiver).

	nsSendCache := self cCoerceSimple: cacheAddress to: #'NSSendCache *'.
	messageSelector := nsSendCache selector.
	argumentCount := nsSendCache numArgs.
	method := (self mframeHomeMethod: framePointer) methodObject.

	self assert: (self stackValue: argumentCount + 1 "ret addr") = methodReceiver.

	methodReceiverClassTag := objectMemory fetchClassTagOf: methodReceiver.

	(self
		inlineLookupInNSMethodCacheSel: messageSelector
		classTag: methodReceiverClassTag
		method: method
		lookupRule: LookupRuleDynamicSuper)
			ifTrue:
				[self assert: localAbsentReceiverOrZero = 0.
				"check for coggability because method is in the cache"
				self ifAppropriateCompileToNativeCode: newMethod selector: messageSelector]
			ifFalse:
				[self deny: (objectMemory isOopForwarded: messageSelector).
				self deny: (objectMemory isForwardedClassTag: methodReceiverClassTag).
				errSelIdx := self lookupDynamicSuperSendNoMNU: methodReceiver.
				errSelIdx ~= 0 ifTrue:
					[self assert: errSelIdx == SelectorDoesNotUnderstand.
					self assert: localAbsentReceiver == methodReceiver.
					self handleMNU: errSelIdx 
						InMachineCodeTo: methodReceiver
						classForMessage: lkupClass.
					self unreachable].
				lkupClassTag := methodReceiverClassTag.
				self addNewMethodToNSCache: LookupRuleDynamicSuper].

	(self maybeMethodHasCogMethod: newMethod) 
		ifTrue: [
			cogMethod := self cogMethodOf: newMethod.
			cogMethod selector = objectMemory nilObject
				ifTrue: [cogit setSelectorOf: cogMethod to: messageSelector]
				ifFalse: ["Deal with anonymous accessors, e.g. in Newspeak.
					The cogMethod may not have the
					correct selector. If not, try and compile a new method
					with the correct selector."
					cogMethod selector ~= messageSelector ifTrue: [
							(cogit cog: newMethod selector: messageSelector)
								ifNotNil: [:newCogMethod | cogMethod := newCogMethod]]].
			cogMethod selector = messageSelector
				ifTrue:
					[cogit
						linkNSSendCache: nsSendCache 
						classTag: methodReceiverClassTag
						enclosingObject: 0
						target: cogMethod
						caller: self mframeHomeMethodExport]
				ifFalse: ["Out of code memory. Fall through to interpret."].
			instructionPointer := self popStack.
			self executeNewMethod.
			self unreachable].
	instructionPointer := self popStack.
	self interpretMethodFromMachineCode.
	self unreachable.! !

!CoInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 10/7/2014 11:01'!
ceEnclosingObjectAt: level
	<api>
	<option: #NewspeakVM>
	^self 
		enclosingObjectAt: level
		withObject: (self mframeReceiver: framePointer)
		withMixin: self mMethodClass! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'rmacnak 8/24/2015 22:45'!
ceImplicitReceiverSend: cacheAddress receiver: methodReceiver
	"An implicit receiver send cache missed."
	| nsSendCache methodReceiverClassTag cogMethod errSelIdx |
	<api>
	<option: #NewspeakVM>
	<inline: false>
	<var: #nsSendCache type: #'NSSendCache *'>
	<var: #cogMethod type: #'CogMethod *'>

	cogit assertCStackWellAligned.
	self assert: (objectMemory addressCouldBeOop: methodReceiver).
	self deny: (objectMemory isOopForwarded: methodReceiver).

	nsSendCache := self cCoerceSimple: cacheAddress to: #'NSSendCache *'.
	messageSelector := nsSendCache selector.
	argumentCount := nsSendCache numArgs.
	method := (self mframeHomeMethod: framePointer) methodObject.

	self assert: (self stackValue: argumentCount + 1 "ret addr") = methodReceiver.

	methodReceiverClassTag := objectMemory fetchClassTagOf: methodReceiver.

	(self
		inlineLookupInNSMethodCacheSel: messageSelector
		classTag: methodReceiverClassTag
		method: method
		lookupRule: LookupRuleImplicit)
			ifTrue:
				[localAbsentReceiverOrZero = 0
					ifTrue: [localAbsentReceiver := methodReceiver]
					ifFalse: [localAbsentReceiver := localAbsentReceiverOrZero].
				"check for coggability because method is in the cache"
				self ifAppropriateCompileToNativeCode: newMethod selector: messageSelector]
			ifFalse:
				[self deny: (objectMemory isOopForwarded: messageSelector).
				self deny: (objectMemory isForwardedClassTag: methodReceiverClassTag).
				errSelIdx := self lookupImplicitReceiverSendNoMNU: methodReceiver.
				errSelIdx ~= 0 ifTrue:
					[self assert: errSelIdx == SelectorDoesNotUnderstand.
					self assert: localAbsentReceiver == methodReceiver.
					self handleMNU: errSelIdx 
						InMachineCodeTo: methodReceiver
						classForMessage: lkupClass.
					self unreachable].
				lkupClassTag := methodReceiverClassTag.
				self addNewMethodToNSCache: LookupRuleImplicit].

	"Fix stacked receiver."
	self stackValue: argumentCount + 1 "ret addr" put: localAbsentReceiver.

	(self maybeMethodHasCogMethod: newMethod) 
		ifTrue: [
			cogMethod := self cogMethodOf: newMethod.
			cogMethod selector = objectMemory nilObject
				ifTrue: [cogit setSelectorOf: cogMethod to: messageSelector]
				ifFalse: ["Deal with anonymous accessors, e.g. in Newspeak.
					The cogMethod may not have the
					correct selector. If not, try and compile a new method
					with the correct selector."
					cogMethod selector ~= messageSelector ifTrue: [
							(cogit cog: newMethod selector: messageSelector)
								ifNotNil: [:newCogMethod | cogMethod := newCogMethod]]].
			cogMethod selector = messageSelector
				ifTrue:
					[cogit
						linkNSSendCache: nsSendCache 
						classTag: methodReceiverClassTag
						enclosingObject: localAbsentReceiverOrZero
						target: cogMethod
						caller: self mframeHomeMethodExport]
				ifFalse: ["Out of code memory. Fall through to interpret."].
			instructionPointer := self popStack.
			self executeNewMethod.
			self unreachable].
	instructionPointer := self popStack.
	self interpretMethodFromMachineCode.
	self unreachable.! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 10/24/2014 13:35'!
ceInterpretMethodFromPIC: aMethodObj receiver: rcvr
	<api>
	| pic primitiveIndex |
	<var: #pic type: #'CogMethod *'>
	"pop off inner return and locate open or closed PIC"
	pic := self cCoerceSimple: self popStack - cogit interpretOffset to: #'CogMethod *'.
	self assert: (pic cmType = CMOpenPIC or: [pic cmType = CMClosedPIC]).
	"If found from an open PIC then it must be an uncogged method and, since it's been found
	 in the method cache, should be cogged if possible.  If found from a closed PIC then at the
	 time the closed PIC was created the method was uncoggable, either because there was
	 no space, it had too many literals or it contained an illegal bytecode).  So don't try and cog
	 it, but subsequently it may have been cogged via another path.  If the method is, or ends up
	 cogged, jump to machine code, otherwise interpret."
	pic cmType = CMOpenPIC ifTrue:
		[self assert: (self methodHasCogMethod: aMethodObj) not.
		 (self methodShouldBeCogged: aMethodObj) ifTrue:
			[cogit cog: aMethodObj selector: pic selector]].
	(self methodHasCogMethod: aMethodObj) ifTrue:
		[self executeCogMethod: (self cogMethodOf: aMethodObj)
			fromUnlinkedSendWithReceiver: rcvr
		 "NOTREACHED"].
	messageSelector := pic selector.
	newMethod := aMethodObj.
	primitiveIndex := self primitiveIndexOf: aMethodObj.
	primitiveFunctionPointer := self functionPointerFor: primitiveIndex inClass: objectMemory nilObject.
	argumentCount := pic cmNumArgs.
	instructionPointer := self popStack.
	^self interpretMethodFromMachineCode
	"NOTREACHED"! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 2/24/2020 14:31'!
ceMNUFromPICMNUMethod: aMethodObj receiver: rcvr
	<api>
	| cPIC primitiveIndex |
	<var: #cPIC type: #'CogMethod *'>
	self assert: (objectMemory addressCouldBeOop: rcvr).
	self assert: (aMethodObj = 0
				or: [(objectMemory addressCouldBeObj: aMethodObj)
					and: [objectMemory isOopCompiledMethod: aMethodObj]]).
	self sendBreakpoint: (objectMemory splObj: SelectorDoesNotUnderstand) receiver: rcvr.
	cPIC := self cCoerceSimple: self popStack - cogit mnuOffset to: #'CogMethod *'.
	self assert: (cPIC cmType = CMClosedPIC or: [cPIC cmType = CMOpenPIC]).
	argumentCount := cPIC cmNumArgs.
	messageSelector := cPIC selector.
	aMethodObj ~= 0 ifTrue:
		[instructionPointer := self popStack.
		self createActualMessageTo: (objectMemory fetchClassOf: rcvr).
		(self maybeMethodHasCogMethod: aMethodObj) ifTrue:
			[self push: instructionPointer.
			 self executeCogMethod: (self cogMethodOf: aMethodObj)
				 fromUnlinkedSendWithReceiver: rcvr.
			 "NOTREACHED"
			 self assert: false].
		newMethod := aMethodObj.
		primitiveIndex := self primitiveIndexOf: aMethodObj.
		primitiveFunctionPointer := self functionPointerFor: primitiveIndex inClass: objectMemory nilObject.
		^self interpretMethodFromMachineCode].
	"handleMNU:InMachineCodeTo:classForMessage: assumes lkupClass is set, since every other use is
	 after a lookupMethodNoMNUEtcInClass: call, which sets lkupClass.  Here we must set it manually.
	 Global variables.  Bah!!"
	self handleMNU: SelectorDoesNotUnderstand
		InMachineCodeTo: rcvr
		classForMessage: (lkupClass := objectMemory fetchClassOf: rcvr).
	"NOTREACHED"
	self assert: false! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 1/20/2015 15:09'!
ceNewArraySlotSize: slotSize
	<api>
	<option: #SqueakV3ObjectMemory>
	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[| obj |
		 obj := objectMemory
					eeInstantiateSmallClassIndex: ClassArrayCompactIndex
					format: objectMemory arrayFormat
					numSlots: slotSize.
		objectMemory fillObj: obj numSlots: slotSize with: objectMemory nilObject.
		^obj].
	^objectMemory
		eeInstantiateAndInitializeClass: (objectMemory splObj: ClassArray)
		indexableSize: slotSize! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 3/16/2017 18:22'!
ceNewHashOf: anObject
	<api>
	<option: #SpurObjectMemory>
	"We know anObject has not a hash yet (or this trampoline would not be called.
	 Sets the hash, then answers it as a smallinteger"
	self assert: ((objectMemory isNonImmediate: anObject)
				and: [(objectMemory rawHashBitsOf: anObject) = 0]).
	^objectMemory integerObjectOf: (objectMemory newHashBitsOf: anObject)! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'akg 11/29/2018 17:21'!
ceNonLocalReturn: returnValue
	<api>
	| closure home unwindContextOrNilOrZero ourContext frameToReturnTo contextToReturnTo theFP callerFP newPage |
	<var: #frameToReturnTo type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #callerFP type: #'char *'>
	<var: #newPage type: #'StackPage *'>
	<var: #thePage type: #'StackPage *'>

	"self shortPrintFrameAndCallers: framePointer.
	self printOop: returnValue.
	self halt."

	self assert: (self isMachineCodeFrame: framePointer).
	self assert: (self frameIsBlockActivation: framePointer).

	"Update the current page's headFrame pointers to enable the search for unwind protects below
	 to identify widowed contexts correctly."
	self externalWriteBackHeadFramePointers.

	"Since this is a block activation the closure is on the stack above any args and the frame."
	closure := self pushedReceiverOrClosureOfFrame: framePointer.
	home := nil. "avoid compiler warning"
	"Walk the closure's lexical chain to find the context or frame to return from (home)."
	[closure ~= objectMemory nilObject] whileTrue:
		[home := objectMemory followField: ClosureOuterContextIndex ofObject: closure.
		 (objectMemory isContext: home) ifFalse:
			["error: can't find home on chain; cannot return"
			 ourContext := self ensureFrameIsMarried: framePointer SP: stackPointer.
			 ^self externalCannotReturn: returnValue from: ourContext].
		 closure := objectMemory followField: ClosureIndex ofObject: home].
	"home is to be returned from provided there is no unwind-protect activation between
	 this frame and home's sender.  Search for an unwind.  findUnwindThroughContext:
	 will answer either the context for an unwind-protect activation or nilObj if the sender
	 cannot be found or 0 if no unwind is found but the sender is."
	unwindContextOrNilOrZero := self findUnwindThroughContext: home.
	unwindContextOrNilOrZero = objectMemory nilObject ifTrue:
		["error: can't find home on chain; cannot return"
		 ourContext := self ensureFrameIsMarried: framePointer SP: stackPointer.
		 ^self externalCannotReturn: returnValue from: ourContext].
	unwindContextOrNilOrZero ~= 0 ifTrue:
		[^self externalAboutToReturn: returnValue through: unwindContextOrNilOrZero].

	"Now we know home is on the sender chain.
	 We could be returning to either a context or a frame.  Find out which."
	contextToReturnTo := nil.
	(self isMarriedOrWidowedContext: home)
		ifTrue:
			[self assert: (self checkIsStillMarriedContext: home currentFP: framePointer).
			 theFP := self frameOfMarriedContext: home.
			 (self isBaseFrame: theFP)
				ifTrue:
					[contextToReturnTo := self frameCallerContext: theFP]
				ifFalse:
					[frameToReturnTo := self frameCallerFP: theFP]]
		ifFalse:
			[contextToReturnTo := objectMemory fetchPointer: SenderIndex ofObject: home.
			 ((objectMemory isContext: contextToReturnTo)
			  and: [self isMarriedOrWidowedContext: contextToReturnTo]) ifTrue:
				[self assert: (self checkIsStillMarriedContext: contextToReturnTo currentFP: framePointer).
			 	 frameToReturnTo := self frameOfMarriedContext: contextToReturnTo.
				 contextToReturnTo := nil]].

	"If returning to a context we must make a frame for it unless it is dead."
	contextToReturnTo ~= nil ifTrue:
		[frameToReturnTo := self establishFrameForContextToReturnTo: contextToReturnTo.
		 frameToReturnTo = 0 ifTrue:
			["error: home's sender is dead; cannot return"
			 ourContext := self ensureFrameIsMarried: framePointer SP: stackPointer.
			 ^self externalCannotReturn: returnValue from: ourContext]].

	"Now we have a frame to return to.  If it is on a different page we must
	 free intervening pages and nil out intervening contexts.  We must free
	 intervening stack pages because if we leave the pages to be divorced
	 then their contexts will be divorced with intact senders and instruction
	 pointers.  This code is similar to primitiveTerminateTo."
	self assert: stackPages pageListIsWellFormed.
	newPage := stackPages stackPageFor: frameToReturnTo.
	newPage ~~ stackPage ifTrue:
		[| currentCtx thePage nextCntx |
		 currentCtx := self frameCallerContext: stackPage baseFP.
		 self assert: (objectMemory isContext: currentCtx).
		 stackPages freeStackPage: stackPage.
		 [self assert: (objectMemory isContext: currentCtx).
		  (self isMarriedOrWidowedContext: currentCtx)
		   and: [(stackPages stackPageFor: (theFP := self frameOfMarriedContext: currentCtx)) = newPage]] whileFalse:
			[(self isMarriedOrWidowedContext: currentCtx)
				ifTrue:
					[thePage := stackPages stackPageFor: theFP.
					 currentCtx := self frameCallerContext: thePage baseFP.
					 stackPages freeStackPage: thePage]
				ifFalse:
					[nextCntx := objectMemory fetchPointer: SenderIndex ofObject: currentCtx.
					 self markContextAsDead: currentCtx.
					 currentCtx := nextCntx]].
		 self setStackPageAndLimit: newPage.
		 self setStackPointersFromPage: newPage].

	"Two cases.  Returning to the top frame or an interior frame.  The
	 top frame has its instruction pointer on top of stack.  An interior
	 frame has its instruction pointer in the caller frame. We need to
	 peel back any frames on the page until we get to the correct frame."
	framePointer = frameToReturnTo
		ifTrue:
			[instructionPointer := self popStack]
		ifFalse:
			[[callerFP := framePointer.
			  framePointer := self frameCallerFP: framePointer.
			  framePointer ~~ frameToReturnTo] whileTrue.
			 instructionPointer := (self frameCallerSavedIP: callerFP) asUnsignedInteger.
			 stackPointer := (self frameCallerSP: callerFP)].
	^self return: returnValue toExecutive: false! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'rmacnak 8/24/2015 22:41'!
ceOuterSend: cacheAddress receiver: methodReceiver
	"An outer send cache missed."
	| nsSendCache depth methodReceiverClassTag cogMethod errSelIdx |
	<api>
	<option: #NewspeakVM>
	<inline: false>
	<var: #nsSendCache type: #'NSSendCache *'>
	<var: #cogMethod type: #'CogMethod *'>

	cogit assertCStackWellAligned.
	self assert: (objectMemory addressCouldBeOop: methodReceiver).
	self deny: (objectMemory isOopForwarded: methodReceiver).

	nsSendCache := self cCoerceSimple: cacheAddress to: #'NSSendCache *'.
	messageSelector := nsSendCache selector.
	argumentCount := nsSendCache numArgs.
	depth := nsSendCache depth.
	method := (self mframeHomeMethod: framePointer) methodObject.

	self assert: (self stackValue: argumentCount + 1 "ret addr") = methodReceiver.

	methodReceiverClassTag := objectMemory fetchClassTagOf: methodReceiver.

	(self
		inlineLookupInNSMethodCacheSel: messageSelector
		classTag: methodReceiverClassTag
		method: method
		lookupRule: depth)
			ifTrue:
				[localAbsentReceiverOrZero = 0
					ifTrue: [localAbsentReceiver := methodReceiver]
					ifFalse: [localAbsentReceiver := localAbsentReceiverOrZero].
				"check for coggability because method is in the cache"
				self ifAppropriateCompileToNativeCode: newMethod selector: messageSelector]
			ifFalse:
				[self deny: (objectMemory isOopForwarded: messageSelector).
				self deny: (objectMemory isForwardedClassTag: methodReceiverClassTag).
				errSelIdx := self lookupOuterSendNoMNU: methodReceiver depth: depth.
				errSelIdx ~= 0 ifTrue:
					[self assert: errSelIdx == SelectorDoesNotUnderstand.
					"Fix stacked receiver."
					self stackValue: argumentCount + 1 "ret addr" put: localAbsentReceiver.
					self handleMNU: errSelIdx 
						InMachineCodeTo: localAbsentReceiver
						classForMessage: lkupClass.
					self unreachable].
				lkupClassTag := methodReceiverClassTag. "Successful lookup cached under the sender's class"
				self addNewMethodToNSCache: depth].

	"Fix stacked receiver."
	self stackValue: argumentCount + 1 "ret addr" put: localAbsentReceiver.

	(self maybeMethodHasCogMethod: newMethod) 
		ifTrue: [
			cogMethod := self cogMethodOf: newMethod.
			cogMethod selector = objectMemory nilObject
				ifTrue: [cogit setSelectorOf: cogMethod to: messageSelector]
				ifFalse: ["Deal with anonymous accessors, e.g. in Newspeak.
					The cogMethod may not have the
					correct selector. If not, try and compile a new method
					with the correct selector."
					cogMethod selector ~= messageSelector ifTrue: [
							(cogit cog: newMethod selector: messageSelector)
								ifNotNil: [:newCogMethod | cogMethod := newCogMethod]]].
			cogMethod selector = messageSelector
				ifTrue:
					[cogit
						linkNSSendCache: nsSendCache 
						classTag: methodReceiverClassTag
						enclosingObject: localAbsentReceiverOrZero
						target: cogMethod
						caller: self mframeHomeMethodExport]
				ifFalse: ["Out of code memory. Fall through to interpret."].
			instructionPointer := self popStack.
			self executeNewMethod.
			self unreachable].
	instructionPointer := self popStack.
	self interpretMethodFromMachineCode.
	self unreachable.! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 12/8/2017 14:44'!
ceReapAndResetErrorCodeFor: cogMethod
	<api>
	<var: #cogMethod type: #'CogMethod *'>
	self assert: primFailCode ~= 0.
	newMethod := cogMethod methodObject.
	self reapAndResetErrorCodeTo: stackPointer + objectMemory wordSize
		header: cogMethod methodHeader! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'rmacnak 8/24/2015 22:45'!
ceSelfSend: cacheAddress receiver: methodReceiver
	"A self send cache missed."
	| nsSendCache methodReceiverClassTag cogMethod errSelIdx |
	<api>
	<option: #NewspeakVM>
	<inline: false>
	<var: #nsSendCache type: #'NSSendCache *'>
	<var: #cogMethod type: #'CogMethod *'>

	cogit assertCStackWellAligned.
	self assert: (objectMemory addressCouldBeOop: methodReceiver).
	self deny: (objectMemory isOopForwarded: methodReceiver).

	nsSendCache := self cCoerceSimple: cacheAddress to: #'NSSendCache *'.
	messageSelector := nsSendCache selector.
	argumentCount := nsSendCache numArgs.
	method := (self mframeHomeMethod: framePointer) methodObject.

	self assert: (self stackValue: argumentCount + 1 "ret addr") = methodReceiver.

	methodReceiverClassTag := objectMemory fetchClassTagOf: methodReceiver.

	(self
		inlineLookupInNSMethodCacheSel: messageSelector
		classTag: methodReceiverClassTag
		method: method
		lookupRule: LookupRuleSelf)
			ifTrue:
				[self assert: localAbsentReceiverOrZero = 0.
				"check for coggability because method is in the cache"
				self ifAppropriateCompileToNativeCode: newMethod selector: messageSelector]
			ifFalse:
				[self deny: (objectMemory isOopForwarded: messageSelector).
				self deny: (objectMemory isForwardedClassTag: methodReceiverClassTag).
				errSelIdx := self lookupOuterSendNoMNU: methodReceiver depth: 0.
				errSelIdx ~= 0 ifTrue:
					[self assert: errSelIdx == SelectorDoesNotUnderstand.
					self assert: localAbsentReceiver == methodReceiver.
					self handleMNU: errSelIdx 
						InMachineCodeTo: methodReceiver
						classForMessage: lkupClass.
					self unreachable].
				lkupClassTag := methodReceiverClassTag.
				self addNewMethodToNSCache: LookupRuleSelf].

	(self maybeMethodHasCogMethod: newMethod) 
		ifTrue: [
			cogMethod := self cogMethodOf: newMethod.
			cogMethod selector = objectMemory nilObject
				ifTrue: [cogit setSelectorOf: cogMethod to: messageSelector]
				ifFalse: ["Deal with anonymous accessors, e.g. in Newspeak.
					The cogMethod may not have the
					correct selector. If not, try and compile a new method
					with the correct selector."
					cogMethod selector ~= messageSelector ifTrue: [
							(cogit cog: newMethod selector: messageSelector)
								ifNotNil: [:newCogMethod | cogMethod := newCogMethod]]].
			cogMethod selector = messageSelector
				ifTrue:
					[cogit
						linkNSSendCache: nsSendCache 
						classTag: methodReceiverClassTag
						enclosingObject: 0
						target: cogMethod
						caller: self mframeHomeMethodExport]
				ifFalse: ["Out of code memory. Fall through to interpret."].
			instructionPointer := self popStack.
			self executeNewMethod.
			self unreachable].
	instructionPointer := self popStack.
	self interpretMethodFromMachineCode.
	self unreachable.! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 1/11/2018 16:30'!
ceSend: selector above: methodClass to: rcvr numArgs: numArgs
	"Entry-point for an unlinked directed super send in a CogMethod.  Smalltalk stack looks like
					receiver
					args
		head sp ->	sender return pc
	methodClass is the class above which to start the lookup.

	If an MNU then defer to handleMNUInMachineCodeTo:... which will dispatch the MNU and
	may choose to allocate a closed PIC with a fast MNU dispatch for this send.  Otherwise
	attempt to link the send site as efficiently as possible.  All link attempts may fail; e.g.
	because we're out of code memory.

	Continue execution via either executeMethod or interpretMethodFromMachineCode:
	depending on whether the target method is cogged or not."
	<api>
	<option: #BytecodeSetHasDirectedSuperSend>
	| classTag classObj errSelIdx cogMethod |
	<inline: false>
	<var: #cogMethod type: #'CogMethod *'>
	<var: #newCogMethod type: #'CogMethod *'>
	"self printExternalHeadFrame"
	"self printStringOf: selector"
	cogit assertCStackWellAligned.
	self assert: (objectMemory addressCouldBeOop: rcvr).
	self sendBreakpoint: selector receiver: rcvr.
	classTag := objectMemory classTagForClass: (self superclassOf: (objectMemory followMaybeForwarded: methodClass)).
	argumentCount := numArgs.
	(self lookupInMethodCacheSel: selector classTag: classTag)
		ifTrue:"check for coggability because method is in the cache"
			[self
				ifAppropriateCompileToNativeCode: newMethod
				selector: selector]
		ifFalse:
			[self deny: (objectMemory isForwardedClassTag: classTag).
			 (objectMemory isOopForwarded: selector) ifTrue:
				[^self
					ceSend: (self handleForwardedSelectorFaultFor: selector)
					above: methodClass
					to: rcvr
					numArgs: numArgs].
			 messageSelector := selector.
			 classObj := objectMemory classForClassTag: classTag.
			 (errSelIdx := self lookupOrdinaryNoMNUEtcInClass: classObj) ~= 0 ifTrue:
				[(errSelIdx = SelectorDoesNotUnderstand
				  and: [(cogMethod := cogit cogMNUPICSelector: messageSelector
											receiver: rcvr
											methodOperand: (self mnuMethodOrNilFor: rcvr)
											numArgs: argumentCount) asUnsignedInteger
						> cogit minCogMethodAddress]) ifTrue:
						[cogit
							linkSendAt: (stackPages longAt: stackPointer)
							in: (self mframeHomeMethod: framePointer)
							to: cogMethod
							offset: cogit noCheckEntryOffset
							receiver: rcvr].
				self handleMNU: errSelIdx
					InMachineCodeTo: rcvr
					classForMessage: classObj.
				self assert: false "NOTREACHED"]].
	"Method found and has a cog method.  Attempt to link to it.  The receiver's class may be young.
	 We must not link to an Open PIC since they perform normal sends."
	(self maybeMethodHasCogMethod: newMethod) ifTrue:
		[cogMethod := self cogMethodOf: newMethod.
		 cogMethod selector = objectMemory nilObject
			ifTrue: [cogit setSelectorOf: cogMethod to: selector]
			ifFalse:
				["Deal with anonymous accessors, e.g. in Newspeak.  The cogMethod may not have the
				  correct selector.  If not, try and compile a new method with the correct selector."
				 cogMethod selector ~= selector ifTrue:
					[(cogit cog: newMethod selector: selector) ifNotNil:
						[:newCogMethod| cogMethod := newCogMethod]]].
		 cogMethod selector = selector ifTrue:
			[cogit
				linkSendAt: (stackPages longAt: stackPointer)
				in: (self mframeHomeMethod: framePointer)
				to: cogMethod
				offset: cogit noCheckEntryOffset
				receiver: rcvr].
		 instructionPointer := self popStack.
		 self executeNewMethod.
		 self assert: false "NOTREACHED"].
	instructionPointer := self popStack.
	^self interpretMethodFromMachineCode
	"NOTREACHED"! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 3/10/2018 13:04'!
ceSend: selector aboveClassBinding: methodClassBinding to: rcvr numArgs: numArgs
	"Entry-point for an unlinked directed super send in a CogMethod.  Smalltalk stack looks like
					receiver
					args
		head sp ->	sender return pc
	methodClassBinding is an association whose value is the class above which to start the lookup."
	<api>
	<option: #BytecodeSetHasDirectedSuperSend>
	self ceSend: selector
		above: (self fetchPointer: ValueIndex
					ofObject: (objectMemory followMaybeForwarded: methodClassBinding))
		to: rcvr
		numArgs: numArgs! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 1/11/2018 13:19'!
ceSend: selector super: superNormalBar to: rcvr numArgs: numArgs
	"Entry-point for an unlinked send in a CogMethod.  Smalltalk stack looks like
					receiver
					args
		head sp ->	sender return pc
		
	If an MNU then defer to handleMNUInMachineCodeTo:... which will dispatch the MNU and
	may choose to allocate a closed PIC with a fast MNU dispatch for this send.  Otherwise
	attempt to link the send site as efficiently as possible.  All link attempts may fail; e.g.
	because we're out of code memory.

	Continue execution via either executeMethod or interpretMethodFromMachineCode:
	depending on whether the target method is cogged or not."
	<api>
	| classTag classObj errSelIdx cogMethod |
	<inline: false>
	<var: #cogMethod type: #'CogMethod *'>
	<var: #newCogMethod type: #'CogMethod *'>
	"self printExternalHeadFrame"
	"self printStringOf: selector"
	cogit assertCStackWellAligned.
	self assert: (objectMemory addressCouldBeOop: rcvr).
	self sendBreakpoint: selector receiver: rcvr.
	superNormalBar = 0
		ifTrue: [classTag := objectMemory fetchClassTagOf: rcvr]
		ifFalse: [classTag := objectMemory classTagForClass: (self superclassOf: (self methodClassOf: (self frameMethodObject: framePointer)))].
	argumentCount := numArgs.
	(self lookupInMethodCacheSel: selector classTag: classTag)
		ifTrue:"check for coggability because method is in the cache"
			[self
				ifAppropriateCompileToNativeCode: newMethod
				selector: selector]
		ifFalse:
			[(objectMemory isOopForwarded: selector) ifTrue:
				[^self
					ceSend: (self handleForwardedSelectorFaultFor: selector)
					super: superNormalBar
					to: rcvr
					numArgs: numArgs].
			 (objectMemory isForwardedClassTag: classTag) ifTrue:
				[self assert: superNormalBar = 0.
				^self
					ceSend: selector
					super: superNormalBar
					to: (self handleForwardedSendFaultForReceiver: rcvr stackDelta: 1 "skip return pc")
					numArgs: numArgs].
			 messageSelector := selector.
			 classObj := objectMemory classForClassTag: classTag.
			 (errSelIdx := self lookupOrdinaryNoMNUEtcInClass: classObj) ~= 0 ifTrue:
				[(errSelIdx = SelectorDoesNotUnderstand
				  and: [(cogMethod := cogit cogMNUPICSelector: messageSelector
											receiver: rcvr
											methodOperand: (self mnuMethodOrNilFor: rcvr)
											numArgs: argumentCount) asUnsignedInteger
						> cogit minCogMethodAddress]) ifTrue:
						[cogit
							linkSendAt: (stackPages longAt: stackPointer)
							in: (self mframeHomeMethod: framePointer)
							to: cogMethod
							offset: (superNormalBar = 0
									ifTrue: [cogit entryOffset]
									ifFalse: [cogit noCheckEntryOffset])
							receiver: rcvr].
				self handleMNU: errSelIdx InMachineCodeTo: rcvr classForMessage: classObj.
				self assert: false "NOTREACHED"]].
	"Method found and has a cog method.  Attempt to link to it.  The receiver's class may be young.
	 If the Cogit can't store young classes in inline caches we can link to an open PIC instead."
	(self maybeMethodHasCogMethod: newMethod) ifTrue:
		[cogMethod := self cogMethodOf: newMethod.
		 cogMethod selector = objectMemory nilObject
			ifTrue: [cogit setSelectorOf: cogMethod to: selector]
			ifFalse:
				["Deal with anonymous accessors, e.g. in Newspeak.  The cogMethod may not have the
				  correct selector.  If not, try and compile a new method with the correct selector."
				 cogMethod selector ~= selector ifTrue:
					[(cogit cog: newMethod selector: selector) ifNotNil:
						[:newCogMethod| cogMethod := newCogMethod]]].
		 cogMethod selector = selector
			ifTrue:
				[cogit
					linkSendAt: (stackPages longAt: stackPointer)
					in: (self mframeHomeMethod: framePointer)
					to: cogMethod
					offset: (superNormalBar = 0
								ifTrue: [cogit entryOffset]
								ifFalse: [cogit noCheckEntryOffset])
					receiver: rcvr]
			ifFalse: "If patchToOpenPICFor:.. returns we're out of code memory"
				[superNormalBar = 0 ifTrue: "Open PICs perform normal sends. Can't patch if this is a super send."
					[cogit
						patchToOpenPICFor: selector
						numArgs: numArgs
						receiver: rcvr]].
		 instructionPointer := self popStack.
		 self executeNewMethod.
		 self assert: false "NOTREACHED"].
	instructionPointer := self popStack.
	^self interpretMethodFromMachineCode
	"NOTREACHED"! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 3/12/2017 15:13'!
ceSendAbort: selector to: rcvr numArgs: numArgs
	"Entry-point for an abort send in a CogMethod (aboutToReturn:through:, cannotReturn: et al).
	 Try and dispatch the send, but the send may turn into an MNU in which case defer to
	 handleMNUInMachineCodeTo:... which will dispatch the MNU.

	 Continue execution via either executeMethod or interpretMethodFromMachineCode:
	 depending on whether the target method is cogged or not."
	<api>
	| classTag classObj errSelIdx |
	<inline: false>
	"self printExternalHeadFrame"
	"self printStringOf: selector"
	cogit assertCStackWellAligned.
	self assert: (objectMemory addressCouldBeOop: rcvr).
	self sendBreakpoint: selector receiver: rcvr.
	argumentCount := numArgs.
	classTag := objectMemory fetchClassTagOf: rcvr.
	(self lookupInMethodCacheSel: selector classTag: classTag)
		ifTrue:"check for coggability because method is in the cache"
			[self
				ifAppropriateCompileToNativeCode: newMethod
				selector: selector]
		ifFalse:
			[messageSelector := selector.
			 classObj := objectMemory classForClassTag: classTag.
			 (errSelIdx := self lookupOrdinaryNoMNUEtcInClass: classObj) ~= 0 ifTrue:
				[self handleMNU: errSelIdx InMachineCodeTo: rcvr classForMessage: classObj.
				"NOTREACHED"
				self assert: false]].
	instructionPointer := self popStack.
	(self maybeMethodHasCogMethod: newMethod) ifTrue:
		[self executeNewMethod.
		 self assert: false
		 "NOTREACHED"].
	^self interpretMethodFromMachineCode
	"NOTREACHED"! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 3/12/2017 15:13'!
ceSendFromInLineCacheMiss: cogMethodOrPIC
	"Send from an Open PIC when the first-level method lookup probe has failed,
	 or to continue when PIC creation has failed (e.g. because we're out of code space),
	 or when a send has failed due to a forwarded receiver."
	<api>
	<var: #cogMethodOrPIC type: #'CogMethod *'>
	| numArgs rcvr classTag classObj errSelIdx |
	"self printFrame: stackPage headFP WithSP: stackPage headSP"
	"self printStringOf: selector"
	numArgs := cogMethodOrPIC cmNumArgs.
	rcvr := self stackValue: numArgs + 1. "skip return pc"
	self assert: (objectMemory addressCouldBeOop: rcvr).
	classTag := objectMemory fetchClassTagOf: rcvr.
	argumentCount := numArgs.
	false ifTrue: "would like to assert this but must also allow for an interpretable method in the cache."
		[self deny: (cogMethodOrPIC cmType = CMOpenPIC
					and: [self newMethodInLookupCacheAt: cogMethodOrPIC selector and: classTag])].
	(self lookupInMethodCacheSel: cogMethodOrPIC selector classTag: classTag)
		ifTrue:"check for coggability because method is in the cache"
			[self
				ifAppropriateCompileToNativeCode: newMethod
				selector: cogMethodOrPIC selector]
		ifFalse:
			[(objectMemory isOopForwarded: cogMethodOrPIC selector) ifTrue:
				[self handleForwardedSelectorFaultFor: cogMethodOrPIC selector.
				 ^self ceSendFromInLineCacheMiss: cogMethodOrPIC].
			 (objectMemory isForwardedClassTag: classTag) ifTrue:
				[self handleForwardedSendFaultForReceiver: rcvr stackDelta: 1 "skip return pc".
				 ^self ceSendFromInLineCacheMiss: cogMethodOrPIC].
			 messageSelector := cogMethodOrPIC selector.
			 classObj := objectMemory classForClassTag: classTag.
			 (errSelIdx := self lookupOrdinaryNoMNUEtcInClass: classObj) ~= 0 ifTrue:
				[self handleMNU: errSelIdx InMachineCodeTo: rcvr classForMessage: classObj.
				"NOTREACHED"
				self assert: false]].
	instructionPointer := self popStack.
	(self maybeMethodHasCogMethod: newMethod) ifTrue:
		[self executeNewMethod.
		 self assert: false
		 "NOTREACHED"].
	^self interpretMethodFromMachineCode
	"NOTREACHED"! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 3/11/2011 10:47'!
ceSendMustBeBoolean: anObject
	<api>
	instructionPointer := self popStack.
	self push: anObject.
	self push: instructionPointer.
	^self
		ceSendAbort: (objectMemory splObj: SelectorMustBeBoolean)
		to: anObject
		numArgs: 0! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 9/26/2020 13:08'!
ceSendMustBeBooleanTo: aNonBooleanObject interpretingAtDelta: jumpSize
	"For RegisterAllocatingCogit we want the pc following a conditional branch not to be reachable, so
	 we don't have to generate code to reload registers.  But notionally the pc following a conditional
	 branch is reached when continuing from a mustBeBoolean error.  Instead of supporting this in the
	 JIT, simply convert to an interpreter frame, backup the pc to the branch, reenter the interpreter
	 and hence retry the mustBeBoolean send therein.  N.B. We could do this for immutability violations
	 too, but immutability is used in actual applications and so should be performant, whereas
	 mustBeBoolean errors are extremely rare and so we choose brevity over performance in this case."
	<api>
	self assert: (objectMemory addressCouldBeOop: aNonBooleanObject).
	instructionPointer := self popStack.
	self convertToInterpreterFrame: jumpSize.
	self push: aNonBooleanObject.

	"and now reenter the interpreter..."
	cogit ceInvokeInterpret.
	"NOTREACHED"
	^nil! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'cb 5/5/2015 16:23'!
ceSistaTrap
	"When we arrive here, the value that trapped is pushed on stack"
	<api>
	<option: #SistaVM>
	| context |
	instructionPointer := self popStack.
	context := self ensureFrameIsMarried: framePointer SP: stackPointer.
	self push: context.
	self push: instructionPointer.
	^self
		ceSendAbort: (objectMemory splObj: SelectorSistaTrap)
		to: context
		numArgs: 0! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 9/12/2013 10:07'!
ceStackOverflow: contextSwitchIfNotNil
	"If contextSwitchIfNotNil is nil we can't context switch.
	 contextSwitchIfNotNil is set to nil by
		- the special primitiveClosureValueNoContextSwitch entry-point in block dispatch
		- the stack check in methods with primitive 198.
	 In a normal method contextSwitchIfNotNil will be the method (see e.g.
	 SimpleStackBasedCogit>>compileFrameBuild).  In a block it will be the
	 closure (see e.g. SimpleStackBasedCogit>>compileMethodBody)."
	<api>
	| cogMethod switched cesoRetAddr |
	<var: #cogMethod type: #'CogBlockMethod *'>
	cesoRetAddr := self popStack. "discard the ceStackOverflow call return address."
	cogMethod := self mframeCogMethod: framePointer.
	self assert: cesoRetAddr - cogit abortOffset = (self asCogHomeMethod: cogMethod) asInteger.
	instructionPointer := cogMethod asInteger + cogMethod stackCheckOffset.
	self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer imbar: false line: #'__LINE__'.
	method := newMethod := messageSelector := objectMemory nilObject.
	switched := self handleStackOverflowOrEventAllowContextSwitch: contextSwitchIfNotNil ~= 0.
	self returnToExecutive: false postContextSwitch: switched.
	self error: 'should not be reached'
! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 8/8/2021 10:52'!
ceTakeProfileSample: aCogMethodOrNil
	"A primitive has succeeded and the nextProfileTick has been reached (all done in machine code).
	 If aCogMethodOrNil is not nil then it is the cog method containing the primitive call.
	 If aCogMethodOrNil is nil then this has been called from primReturnEnterCogCode and newMethod
	 Now take a sample. c.f. checkProfileTick:"
	<api>
	<var: 'aCogMethodOrNil' type: #'CogMethod *'>
	<inline: false> "Slang type inferrence can't deal with self ceTakeProfileSample: nil..."
	self cCode: '' inSmalltalk:
		[aCogMethodOrNil isInteger ifTrue:
			[^self ceTakeProfileSample: (aCogMethodOrNil = 0 ifFalse:
											[cogit cCoerceSimple: aCogMethodOrNil to: #'CogMethod *'])]].
	profileProcess := self activeProcess.
	profileMethod := aCogMethodOrNil
						ifNil: [newMethod]
						ifNotNil: [aCogMethodOrNil methodObject].
	self forceInterruptCheck.
	self zeroNextProfileTick! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 11/1/2013 14:10'!
ceTraceBlockActivation
	<api>
	cogit recordBlockTrace ifTrue:
		[self recordTrace: TraceBlockActivation
			thing: (self mframeHomeMethod: framePointer) methodObject
			source: TraceIsFromMachineCode.
		 cogit printOnTrace ifTrue:
			[self printActivationNameFor: (self mframeHomeMethod: framePointer) methodObject
				receiver: (self frameReceiver: framePointer)
				isBlock: true
				firstTemporary: nil.
			 self cr]]! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 3/19/2015 16:09'!
ceTraceLinkedSend: theReceiver
	| cogMethod |
	<api>
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := self cCoerceSimple: (self stackTop - cogit traceLinkedSendOffset)
						to: #'CogMethod *'.
	self cCode: [] inSmalltalk:
		[cogit checkStackDepthOnSend ifTrue:
			[self maybeCheckStackDepth: (cogMethod cmNumArgs > cogit numRegArgs
											ifTrue: [cogMethod cmNumArgs + 1]
											ifFalse: [0])
				sp: stackPointer + objectMemory wordSize
				pc: (self stackValue: 1)]].
	"cogit recordSendTrace ifTrue: is implicit; wouldn't compile the call otherwise."
	self recordTrace: (objectMemory fetchClassOf: theReceiver)
		thing: cogMethod selector
		source: TraceIsFromMachineCode.
	cogit printOnTrace ifTrue:
		[self printActivationNameFor: cogMethod methodObject
			receiver: theReceiver
			isBlock: false
			firstTemporary: (self cCode: [nil] inSmalltalk: [0]);
			cr].
	self sendBreakpoint: cogMethod selector receiver: theReceiver! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 8/2/2016 10:17'!
ceTraceStoreOf: aValue into: anObject
	<api>
	"For assertion checking."
	self assert: ((objectMemory isImmediate: aValue) or: [objectMemory addressCouldBeObj: aValue]).
	self assert: (objectMemory addressCouldBeObj: anObject)! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 6/1/2009 17:26'!
checkAssertsEnabledInCoInterpreter
	<api>
	| assertsAreEnabledInCoInterpreter |
	assertsAreEnabledInCoInterpreter := false.
	self assert: assertsAreEnabledInCoInterpreter! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 2/7/2015 10:12'!
checkCodeIntegrity: gcModes
	"Perform an integrity/leak check using the heapMap.  Assume
	 clearLeakMapAndMapAccessibleObjects has set a bit at each
	 object's header.  Check that all object references in machine
	 code are valid.  Answer if all checks pass."
	^cogit checkIntegrityOfObjectReferencesInCode: gcModes! !

!CoInterpreter methodsFor: 'process primitive support' stamp: 'eem 6/18/2009 14:16'!
checkCogCompiledCodeCompactionCalledFor
	cogCompiledCodeCompactionCalledFor ifTrue:
		[self commenceCogCompiledCodeCompaction]! !

!CoInterpreter methodsFor: 'primitive support' stamp: 'eem 7/25/2021 17:27'!
checkForAndFollowForwardedPrimitiveState
	"Override to log and export to Cogit"
	<api>
	<option: #SpurObjectMemory>
	| found |
	(cogit recordPrimTrace and: [primTracePluginName isNil]) ifTrue:
		[self fastLogPrim: TracePrimitiveFailure].
	found := super checkForAndFollowForwardedPrimitiveState.
	(found and: [cogit recordPrimTrace]) ifTrue:
		[self fastLogPrim: TracePrimitiveRetry].
	^found! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 10/10/2019 09:10'!
checkIfCFramePointerInUse
	<api>
	"Necessary because we very much want CStackPointer and CFramePointer to be static/private and grouped
	 with other interpreter variables which will hence be accessed via VarBaseReg on platforms trhat have one."
	^self isCFramePointerInUse: (self addressOf: CFramePointer) _: (self addressOf: CStackPointer)! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 10/3/2013 09:46'!
checkLogIntegrity
	"Check the log for leaks.  The trace log is a circular buffer of pairs of entries.
	 If there is an entry at traceLogIndex - 3 \\ TraceBufferSize it has entries.  If
	 there is something at traceLogIndex it has wrapped."
	| limit ok |
	limit := self safe: traceLogIndex - 3 mod: TraceBufferSize.
	(traceLog at: limit) = 0 ifTrue: [^true].
	(traceLog at: traceLogIndex) ~= 0 ifTrue:
		[limit := TraceBufferSize - 3].
	ok := true.
	0 to: limit by: 3 do:
		[:i| | oop |
		oop := traceLog at: i.
		(objectMemory isImmediate: oop) ifFalse:
			[(objectMemory checkOopIntegrity: oop named: 'traceLog' index: i) ifFalse:
				[ok := false]].
		oop := traceLog at: i + 1.
		(objectMemory isImmediate: oop) ifFalse:
			[(objectMemory checkOopIntegrity: oop named: 'traceLog' index: i + 1) ifFalse:
				[ok := false]]].
	^ok! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 2/24/2020 08:16'!
checkOkayFields: oop
	"Check if the argument is an ok object.
	 If this is a pointers object, check that its fields are all okay oops."

	| hasYoung i fieldOop |
	(oop = nil or: [oop = 0]) ifTrue: [ ^true ]. "?? eem 1/16/2013"
	(objectMemory isIntegerObject: oop) ifTrue: [ ^true ].
	(objectMemory checkOkayOop: oop) ifFalse: [ ^false ].
	(objectMemory checkOopHasOkayClass: oop) ifFalse: [ ^false ].
	((objectMemory isPointersNonImm: oop) or: [objectMemory isCompiledMethod: oop]) ifFalse: [ ^true ].
	hasYoung := objectMemory hasSpurMemoryManagerAPI not
				  and: [objectMemory isYoungObject: (objectMemory fetchClassOfNonImm: oop)].
	(objectMemory isCompiledMethod: oop)
		ifTrue:
			[i := (objectMemory literalCountOf: oop) + LiteralStart - 1]
		ifFalse:
			[(objectMemory isContext: oop)
				ifTrue: [i := CtxtTempFrameStart + (self fetchStackPointerOf: oop) - 1]
				ifFalse: [i := (objectMemory lengthOf: oop) - 1]].
	[i >= 0] whileTrue:
		[fieldOop := objectMemory fetchPointer: i ofObject: oop.
		(objectMemory isNonIntegerObject: fieldOop) ifTrue:
			[(i = 0 and: [objectMemory isCompiledMethod: oop])
				ifTrue:
					[(cogMethodZone methodFor: (self pointerForOop: fieldOop)) ifNil:
						[self print: 'method '; printHex: oop; print: ' has an invalid cog method reference'.
						^false]]
				ifFalse:
					[hasYoung := hasYoung or: [objectMemory isYoung: fieldOop].
					(objectMemory checkOkayOop: fieldOop) ifFalse: [ ^false ].
					(self checkOopHasOkayClass: fieldOop) ifFalse: [ ^false ]]].
		i := i - 1].
	hasYoung ifTrue:
		[^objectMemory checkOkayYoungReferrer: oop].
	^true! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 4/2/2015 10:57'!
checkStackIntegrity
	"Perform an integrity/leak check using the heapMap.  Assume
	 clearLeakMapAndMapAccesibleObjects has set a bit at each
	 object's header.  Scan all objects accessible from the stack
	 checking that every pointer points to a header.  Answer if no
	 dangling pointers were detected."
	| ok |
	<inline: false>
	<var: #thePage type: #'StackPage *'>
	<var: #theSP type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #callerFP type: #'char *'>
	<var: #frameRcvrOffset type: #'char *'>
	<var: #cogMethod type: #'CogMethod *'>
	ok := true.
	0 to: numStackPages - 1 do:
		[:i| | thePage theSP theFP frameRcvrOffset callerFP oop |
		thePage := stackPages stackPageAt: i.
		(stackPages isFree: thePage) ifFalse:
			[thePage = stackPage
				ifTrue:
					[theSP := stackPointer.
					 theFP := framePointer]
				ifFalse:
					[theSP := thePage headSP.
					 theFP := thePage  headFP].
			 "Skip the instruction pointer on top of stack of inactive pages."
			 thePage = stackPage ifFalse:
				[theSP := theSP + objectMemory wordSize].
			 [frameRcvrOffset := self frameReceiverLocation: theFP.
			  [theSP <= frameRcvrOffset] whileTrue:
				[oop := stackPages longAt: theSP.
				 ((objectMemory isNonImmediate: oop) 
				   and: [(objectMemory heapMapAtWord: (self pointerForOop: oop)) = 0]) ifTrue:
					[self printFrameThing: 'object leak in frame temp' andFrame: theFP at: theSP.
					 ok := false].
				 theSP := theSP + objectMemory wordSize].
			 (self frameHasContext: theFP) ifTrue:
				[oop := self frameContext: theFP.
				 ((objectMemory isImmediate: oop) 
				   or: [(objectMemory heapMapAtWord: (self pointerForOop: oop)) = 0]) ifTrue:
					[self printFrameThing: 'object leak in frame ctxt' andFrame: theFP at: theFP + FoxThisContext.
					 ok := false].
				 (objectMemory isContext: oop) ifFalse:
					[self printFrameThing: 'frame ctxt should be context' andFrame: theFP at: theFP + FoxThisContext.
					 ok := false].
				 ((objectMemory isContext: oop) and: [self isMarriedOrWidowedContext: oop]) ifFalse:
					[self printFrameThing: 'frame ctxt should be married' andFrame: theFP at: theFP + FoxThisContext.
					 ok := false].
				 ((objectMemory isContext: oop) and: [(self frameOfMarriedContext: oop) = theFP]) ifFalse:
					[self printFrameThing: 'frame ctxt should be married to this frame ' andFrame: theFP at: theFP + FoxThisContext.
					 ok := false]].
			 (self isMachineCodeFrame: theFP)
				ifTrue:
					[| cogMethod |
					 cogMethod := self mframeHomeMethod: theFP.
					 (objectMemory heapMapAtWord: (self pointerForOop: cogMethod)) = 0 ifTrue:
						[self printFrameThing: 'object leak in mframe mthd' andFrame: theFP at: theFP + FoxMethod.
						 ok := false]]
				ifFalse:
					[oop := self iframeMethod: theFP.
					 ((objectMemory isImmediate: oop) 
					   or: [(objectMemory heapMapAtWord: (self pointerForOop: oop)) = 0]) ifTrue:
						[self printFrameThing: 'object leak in iframe mthd' andFrame: theFP at: theFP + FoxMethod.
						 ok := false]].
			 (callerFP := self frameCallerFP: theFP) ~= 0] whileTrue:
				[theSP := theFP + FoxCallerSavedIP + objectMemory wordSize.
				 theFP := callerFP].
			 theSP := theFP + FoxCallerSavedIP + objectMemory wordSize.
			 [theSP <= thePage baseAddress] whileTrue:
				[oop := stackPages longAt: theSP.
				 ((objectMemory isNonImmediate: oop) 
				   and: [(objectMemory heapMapAtWord: (self pointerForOop: oop)) = 0]) ifTrue:
					[self printFrameThing: 'object leak in frame arg' andFrame: theFP at: theSP.
					 ok := false].
				 theSP := theSP + objectMemory wordSize]]].
	^ok! !

!CoInterpreter methodsFor: 'process primitive support' stamp: 'eem 1/8/2011 09:41'!
clearCogCompiledCodeCompactionCalledFor
	"For in-image tests"
	cogCompiledCodeCompactionCalledFor := false! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 6/23/2010 22:07'!
clearTraceLog
	<api>
	traceLogIndex := 0.
	0 to: TraceBufferSize - 1 do:
		[:i|
		traceLog at: i put: 0]! !

!CoInterpreter methodsFor: 'accessing' stamp: 'eem 10/9/2013 13:18'!
cogCodeSize
	^cogCodeSize! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem 10/30/2013 15:32'!
cogMethodOf: aMethodOop
	<api>
	| methodHeader |
	methodHeader := self rawHeaderOf: aMethodOop.
	self assert: ((objectMemory isNonImmediate: methodHeader)
				and: [methodHeader asUnsignedInteger < objectMemory startOfMemory]).
	^self cCoerceSimple: methodHeader to: #'CogMethod *'! !

!CoInterpreter methodsFor: 'accessing' stamp: 'eem 1/8/2011 13:07'!
cogit: aCogit
	<doNotGenerate>
	"for in-image tests"
	cogit := aCogit! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 8/2/2016 15:22'!
commenceCogCompiledCodeCompaction
	| startTime |
	<var: #startTime type: #usqLong>
	cogCompiledCodeCompactionCalledFor := false.
	cogit recordEventTrace ifTrue:
		[self recordTrace: TraceCodeCompaction thing: TraceCodeCompaction source: 0].
	cogit recordPrimTrace ifTrue:
		[self fastLogPrim: TraceCodeCompaction].
	startTime := self ioUTCMicrosecondsNow.

	"This can be called in a number of circumstances.  The instructionPointer
	 may contain a native pc that must be relocated.  There may already be a
	 pushed instructionPointer on stack.  Clients ensure that instructionPointer
	 is 0 if it should not be pushed and/or relocated.  Pushing twice is a mistake
	 because only the top one will be relocated."
	instructionPointer ~= 0 ifTrue:
		["better not have already been pushed"
		 self assert: self stackTop asUnsignedInteger ~= instructionPointer.
		 self push: instructionPointer.
		 self externalWriteBackHeadStackPointer].
	self assertValidStackedInstructionPointers: #'__LINE__'.
	cogit compactCogCompiledCode.
	self nilUncoggableMethods.
	instructionPointer ~= 0 ifTrue:
		[instructionPointer := self popStack.
		 self externalWriteBackHeadStackPointer].
	self assertValidStackedInstructionPointers: #'__LINE__'.

	statCodeCompactionCount := statCodeCompactionCount + 1.
	statCodeCompactionUsecs := statCodeCompactionUsecs + (self ioUTCMicrosecondsNow - startTime).

	objectMemory checkForLeaks ~= 0 ifTrue:
		[objectMemory clearLeakMapAndMapAccessibleObjects.
		 self asserta: (self checkCodeIntegrity: false)]! !

!CoInterpreter methodsFor: 'return bytecodes' stamp: 'eem 12/8/2017 12:49'!
commonCallerReturn
	"Return to the previous context/frame (sender for method activations, caller for block activations)."
	<sharedCodeNamed: 'commonCallerReturn' inCase: #returnTopFromBlock>
	| callersFPOrNull doWeHaveANativeFrame |
	<var: #callersFPOrNull type: #'char *'>

	"TODO: Store/restore the nativeSP more properly, when it exists"
	LowcodeVM ifTrue: [ 
		doWeHaveANativeFrame := self frameHasNativeFrame: localFP.
		doWeHaveANativeFrame ifTrue: [
			nativeStackPointer := (self nativePreviousStackPointerIn: localFP) - 1.
			nativeSP := 0.
			self setFrameHasNotNativeFrame: localFP.
		].
	].

	callersFPOrNull := self frameCallerFP: localFP.
	callersFPOrNull = 0 "baseFrame" ifTrue:
		[self assert: localFP = stackPage baseFP.
		 ^self baseFrameReturn].

	localIP := self frameCallerSavedIP: localFP.
	localSP := localFP + (self frameStackedReceiverOffset: localFP).
	localFP := callersFPOrNull.
	localIP asUnsignedInteger < objectMemory startOfMemory ifTrue:
		[localIP asUnsignedInteger ~= cogit ceReturnToInterpreterPC ifTrue:
			["localIP in the cog method zone indicates a return to machine code."
			 ^self returnToMachineCodeFrame].
		 localIP := self pointerForOop: (self iframeSavedIP: localFP)].
	self setMethod: (self iframeMethod: localFP).
	self fetchNextBytecode.
	self internalStackTopPut: localReturnValue! !

!CoInterpreter methodsFor: 'send bytecodes' stamp: 'rmacnak 5/19/2015 21:37'!
commonSendOrdinary
	"Send a message, starting lookup with the receiver's class."
	"Assume: messageSelector and argumentCount have been set, and that 
	the receiver and arguments have been pushed onto the stack,"
	"Note: This method is inlined into the interpreter dispatch loop."
	<sharedCodeNamed: 'commonSendOrdinary' inCase: #singleExtendedSendBytecode>
	self sendBreakpoint: messageSelector receiver: (self internalStackValue: argumentCount).
	cogit recordSendTrace ifTrue:
		[self recordTrace: (objectMemory classForClassTag: lkupClassTag)
			thing: messageSelector
			source: TraceIsFromInterpreter.
		cogit printOnTrace ifTrue:
			[self printActivationNameForSelector: messageSelector
				startClass: (objectMemory classForClassTag: lkupClassTag); cr]].
	self internalFindNewMethodOrdinary.
	self internalExecuteNewMethod.
	self fetchNextBytecode! !

!CoInterpreter methodsFor: 'indexing primitive support' stamp: 'eem 1/24/2016 11:48'!
commonVariable: rcvr at: index cacheIndex: atIx
	"There is no atCache in the CoInterpreter."
	self shouldNotImplement! !

!CoInterpreter methodsFor: 'indexing primitive support' stamp: 'eem 1/24/2016 11:48'!
commonVariable: rcvr at: index put: value cacheIndex: atIx
	"There is no atCache in the CoInterpreter."
	self shouldNotImplement! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 3/21/2020 19:28'!
compilationBreakpoint: selectorOop classTag: classTag isMNUCase: isMNUCase
	<api>
	<cmacro: '(sel, classTag, isMNU) do { \
	if (numBytesOf(sel) == (isMNU ? -breakSelectorLength : breakSelectorLength) \
	 && !!strncmp((char *)((sel) + BaseHeaderSize), breakSelector, (isMNU ? -breakSelectorLength : breakSelectorLength))) { \
		suppressHeartbeatFlag = 1; \
		compilationBreakpointFor(sel); \
	} \
} while (0)'>
	| bsl i |
	isMNUCase
		ifTrue:
			[(breakSelectorLength = 18 "doesNotUnderstand: size"
			  and: [(self strncmp: 'doesNotUnderstand:' _: breakSelector _: 18) == 0]) ifTrue:
				[(breakLookupClassTag < 0 or: [breakLookupClassTag = classTag]) ifTrue:
					[^self compilationBreakpointFor: selectorOop]].
			bsl := breakSelectorLength negated]
		ifFalse: [bsl := breakSelectorLength].
	bsl = (objectMemory numBytesOf: selectorOop) ifTrue:
		[i := bsl.
		 [i > 0] whileTrue:
			[(objectMemory byteAt: selectorOop + i + objectMemory baseHeaderSize - 1) = (breakSelector at: i) asInteger
				ifTrue:
					[((i := i - 1) = 0
					   and: [breakLookupClassTag < 0 or: [breakLookupClassTag = classTag]]) ifTrue:
							[self compilationBreakpointFor: selectorOop]]
				ifFalse: [i := 0]]]! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 3/21/2020 19:28'!
compilationBreakpoint: selectorOop isMNUCase: isMNUCase
	<api>
	<cmacro: '(sel, isMNU) do { \
	if (numBytesOf(sel) == (isMNU ? -breakSelectorLength : breakSelectorLength) \
	 && !!strncmp((char *)((sel) + BaseHeaderSize), breakSelector, (isMNU ? -breakSelectorLength : breakSelectorLength))) { \
		suppressHeartbeatFlag = 1; \
		compilationBreakpointFor(sel); \
	} \
} while (0)'>
	| bsl i |
	bsl := isMNUCase ifTrue: [breakSelectorLength negated] ifFalse: [breakSelectorLength].
	bsl = (objectMemory numBytesOf: selectorOop) ifTrue:
		[i := bsl.
		 [i > 0] whileTrue:
			[(objectMemory byteAt: selectorOop + i + objectMemory baseHeaderSize - 1) = (breakSelector at: i) asInteger
				ifTrue: [(i := i - 1) = 0 ifTrue:
							[self compilationBreakpointFor: selectorOop]]
				ifFalse: [i := 0]]]! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 10/22/2009 15:56'!
compilationBreakpointFor: selectorOop
	<api>
	suppressHeartbeatFlag := true.
	self
		cCode: [self warning: 'compilation send break (heartbeat suppressed)']
		inSmalltalk: [self halt: 'Compilation of ', breakSelector]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
contextInstructionPointer: theIP frame: theFP
	"Answer a value to store in the InstructionPointer index of a context object for theIP and theFP.
	 Mapping native pcs to bytecode pcs is quite expensive, requiring a search through the method
	 map.  We mitigate this cost by deferring mapping until we really have to, which is when a context's
	 instruction pointer is accessed by Smalltalk code (either direct inst var access or through the
	 instVarAt: primitive).  But to defer mapping we have to be able to distinguish machine code from
	 bytecode pcs, which we do by using negative values for machine code pcs.  So if the frame is a
	 machine code one answer the negation of the offset in the cog method.

	 As a whorish performance hack we also include the block method offset in the pc of a block.
	 The least significant 16 bits are the native pc and the most significant 14 bits are the block
	 start, in block alignment units.  So when mapping back we can find the start of the block.

	 See mustMapMachineCodePC:context: for the code that does the actual mapping."
	<var: #theFP type: #'char *'>
	<inline: false>
	self assert: (self validInstructionPointer: theIP inFrame: theFP).
	(self isMachineCodeFrame: theFP) ifTrue:
		[^self encodedNativePCOf: theIP cogMethod: (self mframeCogMethod: theFP)].
	^objectMemory integerObjectOf: (theIP = cogit ceReturnToInterpreterPC
							ifTrue: [self iframeSavedIP: theFP]
							ifFalse: [theIP])
						- (self iframeMethod: theFP)
						- objectMemory baseHeaderSize
						+ 2! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 11/14/2020 20:48'!
convertToInterpreterFrame: pcDelta
	"Convert the top machine code frame to an interpreter frame.  Support for
	 mustBeBoolean in the RegisterAllocatingCogit and for cloneContext: in shallowCopy
	 when a code compaction is caused by machine code to bytecode pc mapping."

	|  cogMethod methodHeader methodObj startBcpc |
	<var: 'cogMethod' type: #'CogBlockMethod *'>
	<var: 'p' type: #'char *'>

	self assert: (self isMachineCodeFrame: framePointer).

	cogMethod := self mframeCogMethod: framePointer.
	((self mframeIsBlockActivation: framePointer)
	 and: [cogMethod cmIsFullBlock not])
		ifTrue:
			[methodHeader := (self cCoerceSimple: cogMethod cmHomeMethod to: #'CogMethod *') methodHeader.
			 methodObj := (self cCoerceSimple: cogMethod cmHomeMethod to: #'CogMethod *') methodObject.
			 startBcpc := cogMethod startpc]
		ifFalse:
			[methodHeader := (self cCoerceSimple: cogMethod to: #'CogMethod *') methodHeader.
			 methodObj := (self cCoerceSimple: cogMethod to: #'CogMethod *') methodObject.
			 startBcpc := self startPCOfMethod: methodObj].

	"Map the machine code instructionPointer to the interpreter instructionPointer of the branch."
	instructionPointer := cogit bytecodePCFor: instructionPointer startBcpc: startBcpc in: cogMethod.
	instructionPointer := methodObj + objectMemory baseHeaderSize + instructionPointer - pcDelta - 1. "pre-decrement"
	 self validInstructionPointer: instructionPointer inMethod: methodObj framePointer: framePointer.

	"Make space for the two extra fields in an interpreter frame"
	stackPointer to: framePointer + FoxMFReceiver by: objectMemory wordSize do:
		[:p| | oop |
		 oop := objectMemory longAt: p.
		 objectMemory
			longAt: p - objectMemory wordSize - objectMemory wordSize
			put: (objectMemory longAt: p)].
	stackPointer := stackPointer - objectMemory wordSize - objectMemory wordSize.
	"Fill in the fields"
	objectMemory
		longAt: framePointer + FoxIFrameFlags
			put: (self
					encodeFrameFieldHasContext: (self mframeHasContext: framePointer)
					isBlock: (self mframeIsBlockActivation: framePointer)
					numArgs: cogMethod cmNumArgs);
		longAt: framePointer + FoxIFSavedIP
			put: instructionPointer;
		longAt: framePointer + FoxMethod
			put: methodObj.

	self setMethod: methodObj methodHeader: methodHeader! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 4/11/2016 18:10'!
convertToMachineCodeFrame: cogHomeMethod bcpc: bcpc
	<var: #cogHomeMethod type: #'CogMethod *'>
	<returnTypeC: #usqInt>
	"Convert the current interpreter frame into a machine code frame
	 and answer the machine code pc matching bcpc."
	| startBcpc methodField closure cogMethod pc |
	<var: #cogMethod type: #'CogBlockMethod *'>
	<var: #p type: #'char *'>
	self assert: (self isMachineCodeFrame: framePointer) not.
	"Update the return pc, perhaps saving it in the caller's iframeSavedIP."
	(self isBaseFrame: framePointer)
		ifTrue:
			[stackPages
				longAt: framePointer + FoxCallerSavedIP
				put: cogit ceBaseFrameReturnPC]
		ifFalse:
			[(self isMachineCodeFrame: (self frameCallerFP: framePointer)) ifFalse:
				[self iframeSavedIP: (self frameCallerFP: framePointer)
					put: (self frameCallerSavedIP: framePointer) asInteger.
				 stackPages
					longAt: framePointer + FoxCallerSavedIP
					put: cogit ceReturnToInterpreterPC]].
	"Compute the cog method field"
	(self iframeIsBlockActivation: framePointer)
		ifTrue:
			[closure := self pushedReceiverOrClosureOfFrame: framePointer.
			 (self isVanillaBlockClosure: closure)
				ifTrue:
					[startBcpc := self startPCOfClosure: closure.
					 cogMethod := cogit
										findMethodForStartBcpc: startBcpc
										inHomeMethod: cogHomeMethod]
				ifFalse:
					[startBcpc := self startPCOfMethodHeader: cogHomeMethod methodHeader.
					 cogMethod := self cCoerceSimple: cogHomeMethod to: #'CogBlockMethod *'].
			 methodField := cogMethod asInteger + MFMethodFlagIsBlockFlag]
		ifFalse:
			[startBcpc := self startPCOfMethodHeader: cogHomeMethod methodHeader.
			 cogMethod := self cCoerceSimple: cogHomeMethod to: #'CogBlockMethod *'.
			 methodField := cogHomeMethod asInteger].
	"compute the pc before converting the frame to help with debugging."
	pc := cogit mcPCForBackwardBranch: bcpc startBcpc: startBcpc in: cogMethod.
	self assert: pc > (cogMethod asUnsignedInteger + cogit noCheckEntryOffset).
	self assert: bcpc = (cogit bytecodePCFor: pc startBcpc: startBcpc in: cogMethod).
	"now convert to a machine code frame"
	stackPages
		longAt: framePointer + FoxMethod
		put: methodField
			+ ((self iframeHasContext: framePointer)
				ifTrue: [MFMethodFlagHasContextFlag]
				ifFalse: [0]).
	framePointer + FoxIFReceiver to: stackPointer by: objectMemory wordSize negated do:
		[:p|
		stackPages longAt: p + FoxMFReceiver - FoxIFReceiver put: (stackPages longAt: p)].
	stackPointer := stackPointer + FoxMFReceiver - FoxIFReceiver.
	^pc! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 1/20/2015 15:11'!
createClosureNumArgs: numArgs numCopied: numCopied startpc: initialIP
	<api>
	<option: #SqueakV3ObjectMemory>
	| context newClosure |
	self assert: (self isMachineCodeFrame: framePointer).
	"Do *not* include the return pc or copied values in the stack contents;
	 hence + ((1 + numCopied) * BytesPerWord)"
	context := self ensureFrameIsMarried: framePointer
					SP: stackPointer + ((1 + numCopied) * objectMemory wordSize).
	newClosure := self
					closureIn: context
					numArgs: numArgs
					instructionPointer: initialIP
					numCopiedValues: numCopied.
	cogit recordSendTrace ifTrue:
		[self recordTrace: TraceBlockCreation thing: newClosure source: TraceIsFromMachineCode].
	numCopied > 0 ifTrue:
		["N.B. the expression ((numCopied - i) * BytesPerWord)) skips the return address"
		 0 to: numCopied - 1 do:
			[:i|
			"Assume: have just allocated a new BlockClosure; it must be young.
			 Thus, can use unchecked stores."
			 objectMemory storePointerUnchecked: i + ClosureFirstCopiedValueIndex
				ofObject: newClosure
				withValue: (stackPages longAt: stackPointer + ((numCopied - i) * objectMemory wordSize))]].
	"Assume caller will pop stack"
	^newClosure! !

!CoInterpreter methodsFor: 'stack pages' stamp: 'RonieSalgado 4/29/2016 15:16'!
defaultNativeStackFrameSize
	<api>
	^ super defaultNativeStackFrameSize! !

!CoInterpreter methodsFor: 'process primitive support' stamp: 'eem 7/17/2013 12:12'!
deferStackLimitSmashAround: functionSymbol
	"Defer smashes of the stackLimit around the call of functionSymbol (for assert checks)"
	<var: #functionSymbol declareC: 'void (*functionSymbol)(void)'>
	deferSmash := true.
	self perform: functionSymbol.
	deferSmash := false.
	deferredSmash ifTrue:
		[deferredSmash := false.
		 self forceInterruptCheck].
	^true "called from assert"! !

!CoInterpreter methodsFor: 'process primitive support' stamp: 'eem 7/17/2013 12:11'!
deferStackLimitSmashAround: functionSymbol with: arg
	"Defer smashes of the stackLimit around the call of functionSymbol (for assert checks)"
	<var: #functionSymbol declareC: 'void (*functionSymbol)(sqInt)'>
	deferSmash := true.
	self sqLowLevelMFence.
	self perform: functionSymbol with: arg.
	deferSmash := false.
	self sqLowLevelMFence.
	deferredSmash ifTrue:
		[deferredSmash := false.
		 self sqLowLevelMFence.
		 self forceInterruptCheck].
	^true "called from assert"! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 7/24/2021 14:58'!
divorceAFrameIf: criterion in: aStackPage
	"Divorce at most one frame in the current page (since the divorce may cause the page to be split)
	 and answer whether a frame was divorced."
	<var: #criterion declareC: 'sqInt (*criterion)(char *fp)'>
	<var: #aStackPage type: #'StackPage *'>
	| theFP calleeFP theSP theContext |

	theFP := aStackPage headFP.
	theSP := aStackPage headSP.
	theSP := theSP + objectMemory wordSize. "theSP points at hottest item on frame's stack"

	[(self perform: criterion with: theFP) ifTrue:
		[theContext := self ensureFrameIsMarried: theFP SP: theSP.
		 self externalDivorceFrame: theFP andContext: theContext.
		 ^true].
	 calleeFP := theFP.
	 theFP := self frameCallerFP: theFP.
	 theFP ~= 0] whileTrue:
		["theSP points at stacked hottest item on frame's stack"
		 theSP := self frameCallerSP: calleeFP].

	^false! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
divorceAMachineCodeFrameWithCogMethod: cogMethod in: aStackPage
	"Divorce at most one frame in the current page (since the divorce may cause the page to be split)
	 and answer whether a frame was divorced."
	<var: #cogMethod type: #'CogMethod *'>
	<var: #aStackPage type: #'StackPage *'>
	| theFP calleeFP theSP theContext |
	<var: #aStackPage type: #'StackPage *'>
	<var: #theFP type: #'char *'>
	<var: #calleeFP type: #'char *'>
	<var: #theSP type: #'char *'>

	theFP := aStackPage headFP.
	theSP := aStackPage headSP.
	theSP := theSP + objectMemory wordSize. "theSP points at hottest item on frame's stack"

	[((self isMachineCodeFrame: theFP)
	  and: [cogMethod = (self mframeHomeMethod: theFP)]) ifTrue:
		[theContext := self ensureFrameIsMarried: theFP SP: theSP.
		 self externalDivorceFrame: theFP andContext: theContext.
		 ^true].
	 calleeFP := theFP.
	 theFP := self frameCallerFP: theFP.
	 theFP ~= 0] whileTrue:
		["theSP points at stacked hottest item on frame's stack"
		 theSP := self frameCallerSP: calleeFP].

	^false! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 7/24/2021 14:43'!
divorceAllFramesSuchThat: criterion
	"Divorce all frames that satisfy criterion nd answer the current activeContext."
	<var: #criterion declareC: 'sqInt (*criterion)(char *fp)'>
	| activeContext divorcedSome |
	activeContext := self ensureFrameIsMarried: framePointer SP: stackPointer.
	self ensurePushedInstructionPointer.
	self externalWriteBackHeadFramePointers.
	[stackPage ~= 0 ifTrue: "This is needed for the assert in externalDivorceFrame:andContext:"
		[stackPages markStackPageMostRecentlyUsed: stackPage].
	 "Slang can't currently cope with the lack of the variable here.
	  Something to do with the preceding statement.  Take it out
	  and the code is good.  leave it in and we get do { ... } while(l1:)"
	 divorcedSome := self divorceSomeFramesIf: criterion.
	 divorcedSome] whileTrue.
	^activeContext! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 09/28/2015 18:48'!
divorceMachineCodeFramesWithMethod: methodObj
	| cogMethod divorcedSome |
	<var: #cogMethod type: #'CogMethod *'>
	cogMethod := self cogMethodOf: methodObj.
	[stackPage ~= 0 ifTrue: "This is needed for the assert in externalDivorceFrame:andContext:"
		[stackPages markStackPageMostRecentlyUsed: stackPage].
	 "Slang can't currently cope with the lack of the variable here.
	  Something to do with the preceding statement.  Take it out
	  and the code is good.  leave it in and we get do { ... } while(l1:)"
	 divorcedSome := self divorceSomeMachineCodeFramesWithMethod: cogMethod.
	 divorcedSome] whileTrue! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 7/24/2021 14:59'!
divorceSomeFramesIf: criterion
	"Divorce at most one frame (since the divorce may cause the containing
	 page to be split) and answer whether a frame was divorced."
	<var: #criterion declareC: 'sqInt (*criterion)(char *fp)'>
	| divorcedSome |
	divorcedSome := false.
	0 to: numStackPages - 1 do:
		[:i| | aPage |
		aPage := stackPages stackPageAt: i.
		(stackPages isFree: aPage) ifFalse:
			["this to avoid assert in externalDivorceFrame:andContext:"
			 stackPages markStackPageMostRecentlyUsed: stackPage.
			 (self divorceAFrameIf: criterion in: aPage) ifTrue:
				[divorcedSome := true]]].
	^divorcedSome! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 4/10/2015 13:02'!
divorceSomeMachineCodeFramesWithMethod: cogMethod
	"Divorce at most one frame (since the divorce may cause the containing
	 page to be split) and answer whether a frame was divorced."
	<var: #cogMethod type: #'CogMethod *'>
	| divorcedSome |
	<var: #aPage type: #'StackPage *'>
	divorcedSome := false.
	0 to: numStackPages - 1 do:
		[:i| | aPage |
		aPage := stackPages stackPageAt: i.
		(stackPages isFree: aPage) ifFalse:
			["this to avoid assert in externalDivorceFrame:andContext:"
			 stackPages markStackPageMostRecentlyUsed: stackPage.
			 (self divorceAMachineCodeFrameWithCogMethod: cogMethod in: aPage) ifTrue:
				[divorcedSome := true]]].
	^divorcedSome! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 1/10/2014 11:20'!
dumpPrimTraceLog
	"The prim trace log is a circular buffer of entries. If there is
	 an entry at primTraceLogIndex \\ PrimTraceLogSize it has entries.
	 If there is something at primTraceLogIndex it has wrapped."

	<api>
	<inline: false>
	(primTraceLog at: (self safe: primTraceLogIndex - 1 mod: PrimTraceLogSize)) = 0 ifTrue: [^self].
	(primTraceLog at: primTraceLogIndex) ~= 0 ifTrue:
		[primTraceLogIndex to: PrimTraceLogSize - 1 do:
			[:i | self printPrimLogEntryAt: i; cr]].
	0 to: primTraceLogIndex - 1 do:
		[:i | self printPrimLogEntryAt: i; cr]! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 1/10/2014 11:22'!
dumpTraceLog
	<api>
	"The trace log is a circular buffer of pairs of entries. If there is
	 an entry at traceLogIndex - 3 \\ TraceBufferSize it has entries.
	 If there is something at traceLogIndex it has wrapped."
	<inline: false>
	(traceLog at: (self safe: traceLogIndex - 3 mod: TraceBufferSize)) = 0 ifTrue: [^self].
	(traceLog at: traceLogIndex) ~= 0 ifTrue:
		[traceLogIndex to: TraceBufferSize - 3 by: 3 do:
			[:i| self printLogEntryAt: i]].

	0 to: traceLogIndex - 3 by: 3 do:
		[:i| self printLogEntryAt: i]! !

!CoInterpreter methodsFor: 'accessing' stamp: 'eem 1/30/2020 13:07'!
effectiveCogCodeSize
	"With the single-mapped regime the effectiveCogCodeSize is cogCodeSize.
	 With the dual-mapped regime then in production effectiveCogCodeSize is
	 cogCodeSize and in simulation it is two times cogCodeSize (meaning we
	 simulate a dual zone by having two copies of the zone at the start of memory)."
	^cogCodeSize! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 12/14/2015 12:42'!
encodedNativePCOf: mcpc cogMethod: cogMethod
	"Encode the mcpc in cogMethod as a value that can be stashed in a context.
	 Mapping native pcs to bytecode pcs is quite expensive, requiring a search
	 through the method map.  We mitigate this cost by deferring mapping until
	 we really have to, which is when a context's instruction pointer is accessed
	 by Smalltalk code.  But to defer mapping we have to be able to distinguish
	 machine code from bytecode pcs, which we do by using negative values for
	 machine code pcs.

	 As a whorish performance hack we also include the block method offset in
	 the pc of a block. The least significant 16 bits are the native pc and the most
	 significant 15 bits are the block start, in block alignment units.  So when
	 mapping back we can find the start of the block.

	 See mustMapMachineCodePC:context: for the code that does the actual mapping."
	<var: #cogMethod type: #'CogBlockMethod *'>
	| homeMethod blockOffset |
	<var: #homeMethod type: #'CogMethod *'>
	mcpc = cogit ceCannotResumePC ifTrue:
		[^HasBeenReturnedFromMCPCOop].
	cogMethod cmType = CMMethod ifTrue:
		[^objectMemory integerObjectOf: cogMethod asInteger - mcpc].
	homeMethod := cogMethod cmHomeMethod.
	blockOffset := homeMethod asInteger - cogMethod asInteger / cogit blockAlignment.
	^objectMemory integerObjectOf: ((blockOffset bitShift: 16) bitOr: (cogMethod asInteger - mcpc bitAnd: 16rFFFF))! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 4/27/2016 08:11'!
ensureAllContextsWithMethodHaveBytecodePCs: methodObj
	"Map all native pcs to bytecoded pcs in all contexts on methodObj.
	 Used to implement primitiveVoidVMStateForMethod."
	objectMemory allObjectsDo:
		[:oop|
		 ((objectMemory isContextNonImm: oop)
		  and: [(objectMemory fetchPointer: MethodIndex ofObject: oop) = methodObj]) ifTrue:
			[self widowOrForceToBytecodePC: oop]]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 9/6/2019 11:48'!
ensureAllContextsWithMethodMachineCodePrimitiveMethodHaveBytecodePCs
	"Map all native pcs to bytecoded pcs in all contexts that have a method with a cog method with a machine code primitive.
	 Used to implement flushMethodsWithMachineCodePrimitivesAndContinueAnswering:/vmParameterAt: 75 put: aBool."
	<inline: true>
	objectMemory allObjectsDo:
		[:oop| | methodHeader |
		 (objectMemory isContextNonImm: oop) ifTrue:
			[methodHeader := self rawHeaderOf: (objectMemory fetchPointer: MethodIndex ofObject: oop).
			 ((self isCogMethodReference: methodHeader)
			  and: [cogit cogMethodHasMachineCodePrim: (self cCoerceSimple: methodHeader to: #'CogMethod *')]) ifTrue:
				[self widowOrForceToBytecodePC: oop]]]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 3/24/2017 15:47'!
ensureContextHasBytecodePC: aContext
	"Make sure the context has a byetcode pc.  Can only be used on single contexts."
	| pc |
	self assert: (self isMarriedOrWidowedContext: aContext) not.
	pc := objectMemory fetchPointer: InstructionPointerIndex ofObject: aContext.
	((objectMemory isIntegerObject: pc)
	 and: [(pc := objectMemory integerValueOf: pc) < 0]) ifTrue:
		[pc := self mustMapMachineCodePC: pc context: aContext.
		 self assert: (self validBCPC: (objectMemory integerValueOf: pc) inMethod: (objectMemory fetchPointer: MethodIndex ofObject: aContext)).
		 objectMemory storePointerUnchecked: InstructionPointerIndex ofObject: aContext withValue: pc]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 4/12/2012 16:43'!
ensureContextIsExecutionSafeAfterAssignToStackPointer: aContext
	"Safety to give the JIT lattitude in calling convention.  Conceptually, returning
	 a value to a context involves pushing that value onto the stack.  This is used
	 in Squeak methods such as ContextPart>>jump
		jump
			| top |
			thisContext sender push: nil.
			stackp = 0 ifTrue: [self stepToSendOrReturn].
			stackp = 0 ifTrue: [self push: nil].
			top := self pop.
			thisContext privSender: self.
			^top
	 Here jump may pop the value of a temporary variable off the stack which will,
	 conceptually and, in the interpreter, actually, get pushed back on return.  But
	 if the JIT is mapping the stack to registers disaster may ensue since the value
	 may not get pushed to the stack and code may access an invalid value (e.g. a pc).

	 The solution is to fall back on the interpreter.  If the stack pointer is changed we
	 also ensure the pc is a bytecode pc (+ive) which will cause makeBaseFrameFor:
	 to create an interpreter frame if the context is executed again."
	<inline: false>
	self ensureContextHasBytecodePC: aContext! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 12/21/2016 10:59'!
ensureMethodIsCogged: methodObj maybeClosure: maybeClosure
	"Ensure that methodObj has been cogged.  It may be a FullBlockMethod if maybeClosure is a FullBlockClosure."
	<returnTypeC: #'CogMethod *'>
	| rawHeader cogMethod yetToCompact |
	<inline: true>
	<var: #cogMethod type: #'CogMethod *'>
	rawHeader := self rawHeaderOf: methodObj.
	(self isCogMethodReference: rawHeader) ifTrue:
		[^self cCoerceSimple: rawHeader to: #'CogMethod *'].
	yetToCompact := true.
	[(maybeClosure ~= objectMemory nilObject and: [(self isVanillaBlockClosure: maybeClosure) not])
		ifTrue: [cogMethod := cogit cogFullBlockMethod: methodObj numCopied: (self copiedValueCountOfFullClosure: maybeClosure)]
		ifFalse: [cogMethod := cogit cog: methodObj selector: objectMemory nilObject].
	 cogMethod == nil
	 and: [cogCompiledCodeCompactionCalledFor
	 and: [yetToCompact]]] whileTrue:
		[yetToCompact := false.
		 self commenceCogCompiledCodeCompaction].
	(self asserta: cogMethod ~~ nil) ifFalse:
		[self error: 'could not compile method that should have been compiled'].
	^cogMethod! !

!CoInterpreter methodsFor: 'enilopmarts' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
ensurePushedInstructionPointer
	"We're about to make some transition to a machine code method which
	 requires the instructionPointer must be on the stack.  We could have come
	 from the interpreter, either directly or via a machine code primitive.  We
	 could have come from machine code.  The instructionPointer tells us where
	 from.  Make sure the instruction pointer is pushed and/or saved."
	instructionPointer asUnsignedInteger >= objectMemory startOfMemory
		ifTrue:
			"invoked directly from the interpreter"
			[self iframeSavedIP: framePointer put: instructionPointer.
			 self push: cogit ceReturnToInterpreterPC]
		ifFalse:
			["instructionPointer == cogit ceReturnToInterpreterPC
				ifTrue: [invoked from the interpreter via a machine code primitive]
				ifFalse: [invoked from machine code].
			 If in the first case the bytecode instructionPointer has already been
			 saved in iframeSavedIP so all we need to do is push the instructionPointer."
			 self push: instructionPointer]! !

!CoInterpreter methodsFor: 'initialization' stamp: 'eem 10/29/2020 10:16'!
enterSmalltalkExecutiveImplementation
	"Main entry-point into the interpreter at each execution level, where an execution
	 level is either the start of execution or reentry for a callback.  Capture the C stack
	 pointers so that calls from machine-code into the C run-time occur at this level.
	 This is the actual implementation, separated from enterSmalltalkExecutive so the
	 simulator can wrap it in an exception handler and hence simulate the setjmp/longjmp."
	<inline: false>
	cogit assertCStackWellAligned.
	cogit ceCaptureCStackPointers.
	(self isMachineCodeFrame: framePointer) ifTrue:
		[self returnToExecutive: false postContextSwitch: true
		 "NOTREACHED"].
	self interpret.
	^0! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 9/7/2020 13:38'!
error: aString
	<api: 'extern void error(const char *s)'>
	<doNotGenerate>
	super error: aString! !

!CoInterpreter methodsFor: 'enilopmarts' stamp: 'eem 3/25/2015 16:27'!
executeCogBlock: cogMethod closure: closure mayContextSwitch: mayContextSwitch
	"Execute a block within a CogMethod.  The caller has already pushed the block and
	 any arguments and the return pc.  First push the return-to-interpreter trampoline,
	 then the entry-point and finally the receiver.  We /do not/ push any register
	 argument(s) to reduce complications in block dispatch; effectively there are no
	 register arguments to blocks. Instead, the machine code block value primitives
	 push the reg args if necessary before dispatching to the block.  Hence here, only
	 the receiver gets pushed. See genPrimitiveClosureValue"
	<var: #cogMethod type: #'CogMethod *'>
	cogit assertCStackWellAligned.
	self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer.
	self ensurePushedInstructionPointer.
	self push: cogMethod asInteger
			+ (mayContextSwitch
				ifTrue: [cogMethod blockEntryOffset]
				ifFalse: [cogMethod blockEntryOffset - cogit noContextSwitchBlockEntryOffset]).
	self push: closure.
	cogit ceCallCogCodePopReceiverReg
	"NOTREACHED"! !

!CoInterpreter methodsFor: 'enilopmarts' stamp: 'eem 7/3/2014 17:12'!
executeCogMethod: cogMethod fromLinkedSendWithReceiver: rcvr
	<api>
	"Execute a CogMethod from a linked send.  The receiver,
	 arguments and return address are on the Smalltalk stack.  First
	 push the entry-point and finally the register argument(s).  Then write
	 back the frame pointers and call the routine that will pop off the register
	 argument(s) and jump to the entry by executing a return instruction.

	 In the simple jit only the receiver gets passed in registers, so only the
	 receiver gets pushed."
	<var: #cogMethod type: #'CogMethod *'>
	cogit assertCStackWellAligned.
	self assert: (self isMachineCodeFrame: framePointer).
	self assertValidExecutionPointe: self stackTop r: framePointer s: stackPointer imbar: false line: #'__LINE__'.
	cogit numRegArgs > 0 ifTrue: "dont use and: so as to get Slang to inline cogit numRegArgs > 0"
		[cogMethod cmNumArgs <= cogit numRegArgs ifTrue:
			[self callRegisterArgCogMethod: cogMethod at: cogit entryOffset receiver: rcvr]].
	self
		push: cogMethod asInteger + cogit entryOffset;
		push: rcvr.
	cogit ceCallCogCodePopReceiverReg
	"NOTREACHED"! !

!CoInterpreter methodsFor: 'enilopmarts' stamp: 'eem 7/3/2014 17:12'!
executeCogMethod: cogMethod fromUnlinkedSendWithReceiver: rcvr
	"Execute a CogMethod from an unlinked send.  The receiver,
	 arguments and return address are on the Smalltalk stack.  First
	 push the entry-point and finally the register argument(s).  Then write
	 back the frame pointers and call the routine that will pop off the register
	 argument(s) and jump to the entry by executing a return instruction.

	 In the simple jit only the receiver gets passed in registers, so only the
	 receiver gets pushed."
	<var: #cogMethod type: #'CogMethod *'>
	cogit assertCStackWellAligned.
	self assert: (self isMachineCodeFrame: framePointer).
	self assertValidExecutionPointe: self stackTop r: framePointer s: stackPointer imbar: false line: #'__LINE__'.
	cogit numRegArgs > 0 ifTrue: "dont use and: so as to get Slang to inline cogit numRegArgs > 0"
		[cogMethod cmNumArgs <= cogit numRegArgs ifTrue:
			[self callRegisterArgCogMethod: cogMethod at: cogit noCheckEntryOffset receiver: rcvr]].
	self
		push: cogMethod asInteger + cogit noCheckEntryOffset;
		push: rcvr.
	cogit ceCallCogCodePopReceiverReg
	"NOTREACHED"! !

!CoInterpreter methodsFor: 'enilopmarts' stamp: 'eem 3/27/2015 15:19'!
executeCogPIC: cogPIC fromLinkedSendWithReceiver: rcvr andCacheTag: cacheTag
	<api>
	"Execute a closed PIC from a linked send, to redispatch based on the rcvr.
	 The receiver, arguments and return address are on the Smalltalk stack.
	 First push the entry-point and finally the register argument(s).  Then write
	 back the frame pointers and call the routine that will pop off the register
	 argument(s) and jump to the entry by executing a return instruction.

	 In the simple jit only the receiver gets passed in registers, so only the
	 receiver gets pushed."
	<var: #cogPIC type: #'CogMethod *'>
	cogit assertCStackWellAligned.
	self assert: (self isMachineCodeFrame: framePointer).
	self assertValidExecutionPointe: self stackTop r: framePointer s: stackPointer imbar: false line: #'__LINE__'.
	self push: cogPIC asInteger + cogit entryOffset.
	cogit numRegArgs > 0 ifTrue:"dont use and: so as to get Slang to inline cogit numRegArgs > 0"
		[cogPIC cmNumArgs <= cogit numRegArgs ifTrue:
			[self push: cacheTag.
			 cogPIC cmNumArgs caseOf: {
				[0]	->	[cogit ceCall0ArgsPIC].
				[1]	->	[cogit ceCall1ArgsPIC].
				[2]	->	[cogit ceCall2ArgsPIC]
			 	}
				otherwise: [].
			 self error: 'not reached']].
	self
		push: rcvr;
		push: cacheTag.
	cogit ceCallCogCodePopReceiverAndClassRegs
	"NOTREACHED"! !

!CoInterpreter methodsFor: 'enilopmarts' stamp: 'eem 4/7/2016 16:39'!
executeFullCogBlock: cogMethod closure: closure mayContextSwitch: mayContextSwitch
	"Execute a FullBlockClosure with a CogMethod.  The caller has already pushed the block and
	 any arguments and the return pc.  First push the return-to-interpreter trampoline,
	 then the entry-point and finally the receiver.  We /do not/ push any register
	 argument(s) to reduce complications in block dispatch; effectively there are no
	 register arguments to blocks. Instead, the machine code block value primitives
	 push the reg args if necessary before dispatching to the block.  Hence here, only
	 the receiver gets pushed. See genPrimitiveClosureValue"
	<var: #cogMethod type: #'CogMethod *'>
	cogit assertCStackWellAligned.
	self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer.
	self ensurePushedInstructionPointer.
	self push: cogMethod asInteger 
		+ (mayContextSwitch
				ifTrue: [cogit fullBlockEntryOffset]
				ifFalse: [cogit fullBlockNoContextSwitchEntryOffset]).
	self push: closure.
	cogit ceCallCogCodePopReceiverReg
	"NOTREACHED"! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem 1/9/2020 10:46'!
executeNewMethod
	"Execute newMethod - either primitiveFunctionPointer must be set directly
	 (i.e. from primitiveExecuteMethod et al), or it would have been set probing
	 the method cache (i.e. primitivePerform et al)."
	| methodHeader inInterpreter |
	inInterpreter := instructionPointer >= objectMemory startOfMemory.
	primitiveFunctionPointer ~= 0 ifTrue:
		[self isPrimitiveFunctionPointerAnIndex ifTrue:
			[self externalQuickPrimitiveResponse.
			 self return: self popStack toExecutive: inInterpreter.
			 ^nil].
		 "slowPrimitiveResponse may of course context-switch.  If so we must reenter the
		  new process appopriately, returning only if we've reached here directly from the
		  interpreter and have found an interpreter frame.  The instructionPointer tells us
		  from whence we came."
		 self slowPrimitiveResponse ifTrue:
			[self return: self popStack toExecutive: inInterpreter.
			 ^nil]].
	methodHeader := self rawHeaderOf: newMethod.
	"if not primitive, or primitive failed, activate the method"
	(self isCogMethodReference: methodHeader)
		ifTrue:
			[instructionPointer asUnsignedInteger >= objectMemory startOfMemory ifTrue:
				[self iframeSavedIP: framePointer put: instructionPointer asInteger.
				 instructionPointer := cogit ceReturnToInterpreterPC].
			self activateNewCogMethod: (self cCoerceSimple: methodHeader to: #'CogMethod *') inInterpreter: inInterpreter]
		ifFalse:
			[self activateNewMethod]! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem 1/9/2020 10:43'!
executeNewMethodJitting
	"Execute newMethod - either primitiveFunctionPointer must be set directly
	 (i.e. from primitiveExecuteMethod et al), or it would have been set probing
	 the method cache (i.e. primitivePerform et al).
	 Eagerly compile it if appropriate so that doits are fast."
	| methodHeader inInterpreter |
	inInterpreter := instructionPointer >= objectMemory startOfMemory.
	primitiveFunctionPointer ~= 0 ifTrue:
		[self isPrimitiveFunctionPointerAnIndex ifTrue:
			[self externalQuickPrimitiveResponse.
			 self return: self popStack toExecutive: inInterpreter.
			 ^nil].
		 "slowPrimitiveResponse may of course context-switch.  If so we must reenter the
		  new process appopriately, returning only if we've reached here directly from the
		  interpreter and have found an interpreter frame.  The instructionPointer tells us
		  from whence we came."
		 self slowPrimitiveResponse ifTrue:
			[self return: self popStack toExecutive: inInterpreter.
			 ^nil]].
	"Eagerly compile it if appropriate so that doits are fast."
	methodHeader := self rawHeaderOf: newMethod.
	(self isCogMethodReference: methodHeader) ifFalse:
		[(self methodWithHeaderShouldBeCogged: methodHeader)
			ifTrue:
				[cogit cog: newMethod selector: objectMemory nilObject.
				 methodHeader := self rawHeaderOf: newMethod]
			ifFalse: [self maybeFlagMethodAsInterpreted: newMethod]].
	"if not primitive, or primitive failed, activate the method"
	(self isCogMethodReference: methodHeader)
		ifTrue:
			[instructionPointer asUnsignedInteger >= objectMemory startOfMemory ifTrue:
				[self iframeSavedIP: framePointer put: instructionPointer asInteger.
				 instructionPointer := cogit ceReturnToInterpreterPC].
			self activateNewCogMethod: (self cCoerceSimple: methodHeader to: #'CogMethod *') inInterpreter: inInterpreter]
		ifFalse:
			[self activateNewMethod]! !

!CoInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 8/20/2020 10:19'!
extendedStoreBytecodePop: popBoolean
	"Override to use itemporary:in:put:"
	| descriptor variableIndex value |
	<inline: true>
	descriptor := self fetchByte.
	variableIndex := descriptor bitAnd: 63.
	value := self internalStackTop.
	popBoolean ifTrue: [self internalPop: 1].
	(descriptor >> 6 bitAnd: 3) caseOf: {
		[0] ->	[objectMemory storePointerImmutabilityCheck: variableIndex ofObject: self receiver withValue: value].
		[1] ->	[self itemporary: variableIndex in: localFP put: value].
		[2] ->	[self error: 'illegal store'].
		[3] ->	[self storeLiteralVariable: variableIndex withValue: value] }.
	self fetchNextBytecode! !

!CoInterpreter methodsFor: 'return bytecodes' stamp: 'eem 3/11/2011 10:47'!
externalAboutToReturn: resultOop through: aContext
	| ourContext |
	<inline: true>
	ourContext := self ensureFrameIsMarried: framePointer SP: stackPointer.
	self push: ourContext.
	self push: resultOop.
	self push: aContext.
	"The ceNonLocalReturnTrampoline pops its caller's return pc into instructionPointer.
	 In this uncommon case restore it, since a send's call pushes the instructionPointer (after the arguments)."
	self push: instructionPointer.
	^self
		ceSendAbort: (objectMemory splObj: SelectorAboutToReturn)
		to: ourContext
		numArgs: 2! !

!CoInterpreter methodsFor: 'return bytecodes' stamp: 'eem 3/11/2011 10:48'!
externalCannotReturn: resultOop from: aContext
	<inline: true>
	self push: aContext.
	self push: resultOop.
	"Both ceBaseFrameReturnTrampoline & ceNonLocalReturnTrampoline pop
	 their caller's return pc into instructionPointer.  In this uncommon case restore
	 it, since a send's call pushes the instructionPointer (after the arguments)."
	self push: instructionPointer.
	^self
		ceSendAbort: (objectMemory splObj: SelectorCannotReturn)
		to: aContext
		numArgs: 1! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 9/30/2020 00:02'!
externalInstVar: offset ofContext: aContext
	"Fetch an instance variable from a maybe married context.
	 If the context is still married compute the value of the
	 relevant inst var from the spouse frame's state.

	 If the context is single but has a negative instruction pointer
	 recognise that the instruction pointer is actually into machine
	 code and convert it to the corresponding bytecode pc."
	<inline: false>
	| value |

	self assert: (objectMemory isContext: aContext).
	self assert: offset <= (ReceiverIndex + (self checkStackPointerForMaybeMarriedContext: aContext)).
	"method, closureOrNil & receiver need no special handling; only
	 sender, pc & stackp have to be computed for married contexts."
	(self isReadMediatedContextInstVarIndex: offset) ifFalse:
		[^objectMemory fetchPointer: offset ofObject: aContext].

	self externalWriteBackHeadFramePointers.
	(self isStillMarriedContext: aContext) ifTrue:
		[^self fetchPointer: offset ofMarriedContext: aContext].
	
	value := objectMemory fetchPointer: offset ofObject: aContext.
	"Why not update the slot to refer to the mapped pc?  We're damned if we do, and damned if
	 we don't.  If it is and the context is live and is returned to, then the pc will be mapped back
	 to machine code.  If we don't then the mapping will be repeated on every access.  What we
	 need are stats that show whether or not a read here predicts one or other outcome with any
	 certainty.  For now, we keep it as it has been for a long time and do not update the slot."
	(offset = InstructionPointerIndex
 	 and: [(objectMemory isIntegerObject: value)
 	 and: [value signedIntFromLong < 0]]) ifTrue:
		[^self mustMapMachineCodePC: (objectMemory integerValueOf: value) context: aContext].
	^value! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 11/13/2014 13:11'!
externalWriteBackHeadStackPointer
	self assert: (stackPointer < stackPage baseAddress
				and: [stackPointer > (stackPage realStackLimit - (LargeContextSlots * objectMemory bytesPerOop))]).
	stackPage headSP: stackPointer! !

!CoInterpreter methodsFor: 'utilities' stamp: 'eem 5/8/2009 16:36'!
externalizeIPandSP
	"Copy the local instruction, stack and frame pointers to global variables for use in primitives and other functions outside the interpret loop."

	self assert: localIP asUnsignedInteger ~= cogit ceReturnToInterpreterPC.
	instructionPointer := self oopForPointer: localIP.
	stackPointer := localSP.
	framePointer := localFP! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 5/15/2015 17:10'!
fastLogPrim: aSelectorOrImmediate
	"Fast tracing of named primitives.  primTraceLogIndex is a byte variable.
	 aSelectorOrImmediate is a selector oop or one of TraceCodeCompaction et al.
	 primTraceLog has 256 entries.  In C the + 1 below is hence implicitly modulo 256."
	<inline: true>
	primTraceLog at: primTraceLogIndex put: aSelectorOrImmediate.
	self primTraceLogIndex: primTraceLogIndex + 1! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem 2/12/2016 13:57'!
findNewMethodInClassTag: classTagArg
	"Find the compiled method to be run when the current messageSelector is
	 sent to the given classTag, setting the values of newMethod and primitiveIndex."
	| ok classTag |
	<inline: false>
	ok := self lookupInMethodCacheSel: messageSelector classTag: classTagArg.
	ok	ifTrue:
			[self ifAppropriateCompileToNativeCode: newMethod selector: messageSelector]
		ifFalse:
			["entry was not found in the cache; perhaps soemthing was forwarded."
			 classTag := classTagArg.
			 ((objectMemory isOopForwarded: messageSelector)
			  or: [objectMemory isForwardedClassTag: classTag]) ifTrue:
				[(objectMemory isOopForwarded: messageSelector) ifTrue:
					[messageSelector := self handleForwardedSelectorFaultFor: messageSelector].
				 (objectMemory isForwardedClassTag: classTag) ifTrue:
					[classTag := self handleForwardedSendFaultForTag: classTag].
				ok := self lookupInMethodCacheSel: messageSelector classTag: classTag.
				ok ifTrue:
					[^self ifAppropriateCompileToNativeCode: newMethod selector: messageSelector]].
			 "entry was not found in the cache; look it up the hard way "
			 lkupClass := objectMemory classForClassTag: classTag.
			 NewspeakVM
				ifTrue: [self lookupOrdinarySend]
				ifFalse: [self lookupMethodInClass: lkupClass].
			 self addNewMethodToCache: lkupClass]! !

!CoInterpreter methodsFor: 'method lookup cache' stamp: 'eem 1/24/2016 11:46'!
flushAtCache
	"There is no atCache in the CoInterpreter."
	<inline: true>! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 4/9/2015 10:21'!
flushBecommedClassesInMethodZone
	<inline: true>
	cogit unlinkSendsLinkedForInvalidClasses! !

!CoInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 7/25/2021 17:07'!
flushExternalPrimitives
	"Flush the references to external functions from plugin primitives.
	 Then continue execution answering self.
	 This will force a reload of those primitives when accessed next. 
	 Note: We must flush the method cache here also, so that any failed
	 primitives are looked up again.
	 Override to ensure that any and all activations of an external method
	 have a bytecode pc so that if code generation changes (e.g. a primitive
	 method is used, unloaded, and the reloaded primitive is marked with
	 the FastCPrimitiveFlag) stale machine code pcs have been eliminated.
	 THIS MUST BE INVOKED IN THE CONTEXT OF A PRIMITIVE."
	| activeContext theFrame thePage |
	activeContext := self divorceAllFramesSuchThat: #isMachineCodeFrameForExternalPrimitiveMethod:.
	objectMemory allObjectsDo:
		[:oop|
		(objectMemory isCompiledMethod: oop)
			ifTrue:
				[self flushExternalPrimitiveOf: oop]
			ifFalse:
				[(objectMemory isContext: oop) ifTrue:
					[self mapToBytecodePCIfActivationOfExternalMethod: oop]]].
	cogit unlinkSendsToMethodsSuchThat: #cogMethodHasExternalPrim: AndFreeIf: true.
	self flushMethodCache.
	self flushExternalPrimitiveTable.
	self cCode: '' inSmalltalk:
		[self assert: (cogit methodZone cogMethodsSelect: [:cogMethod| cogMethod cmType > CMFree and: [cogit cogMethodHasExternalPrim: cogMethod]]) isEmpty].
	"If flushing led to divorce continue in the interpreter."
	(self isStillMarriedContext: activeContext) ifFalse:
		[self zeroStackPage. "to avoid assert in marryContextInNewStackPageAndInitializeInterpreterRegisters:"
		 self marryContextInNewStackPageAndInitializeInterpreterRegisters: activeContext.
		 self popStack. "pop pushed instructionPointer"
		 self pop: argumentCount.
		 cogit ceInvokeInterpret
		 "NOTREACHED"].
	"If not, work out where we are and continue"
	theFrame := self frameOfMarriedContext: activeContext.
	thePage := stackPages stackPageFor: theFrame.
	self assert: thePage headFP = theFrame.
	self setStackPageAndLimit: thePage.
	self setStackPointersFromPage: thePage.
	instructionPointer := self popStack.
	self pop: argumentCount! !

!CoInterpreter methodsFor: 'method lookup cache' stamp: 'eem 10/31/2013 11:07'!
flushMethodCache
	"Flush the method cache. The method cache is flushed on every programming change and garbage collect."

	super flushMethodCache.
	cogit unlinkAllSends! !

!CoInterpreter methodsFor: 'primitive support' stamp: 'eem 7/25/2021 10:28'!
flushMethodsWithMachineCodePrimitivesAndContinueAnswering: result
	"Arrange that any and all cog methods with machine code primitives can be and are discarded.
	 Hence scan contexts and map their PCs to bytecode PCs if required, and scan frames, divorcing
	 the frames of activationsif required.  Then continue execution answering result.  THIS MUST BE
	 INVOKED IN THE CONTEXT OF A PRIMITIVE.  It exists to support vmParameterAt:put:."
	| activeContext theFrame thePage |
	activeContext := self divorceAllFramesSuchThat: #isMachineCodeFrameWithCogMethod:.
	self ensureAllContextsWithMethodMachineCodePrimitiveMethodHaveBytecodePCs.
	cogit unlinkSendsToMethodsSuchThat: #cogMethodHasMachineCodePrim: AndFreeIf: true.

	"If flushing led to divorce continue in the interpreter."
	(self isStillMarriedContext: activeContext) ifFalse:
		[self zeroStackPage. "to avoid assert in marryContextInNewStackPageAndInitializeInterpreterRegisters:"
		 self marryContextInNewStackPageAndInitializeInterpreterRegisters: activeContext.
		"pop bogus machine-code instructionPointer, arguments and receiver"
		 self pop: argumentCount + 2 thenPush: result.
		 cogit ceInvokeInterpret
		 "NOTREACHED"].
	"If not, work out where we are and continue"
	theFrame := self frameOfMarriedContext: activeContext.
	thePage := stackPages stackPageFor: theFrame.
	self assert: thePage headFP = theFrame.
	self setStackPageAndLimit: thePage.
	self setStackPointersFromPage: thePage.
	instructionPointer := self popStack.
	 self pop: argumentCount + 1 thenPush: result! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem 9/9/2019 15:55'!
followForwardedFieldsInCurrentMethod
	| cogMethod wasInYoungReferrers |
	<var: #cogMethod type: #'CogMethod *'>
	<inline: false>
	(self isMachineCodeFrame: framePointer)
		ifTrue:
			[cogMethod := self mframeHomeMethod: framePointer.
			 objectMemory
				followForwardedObjectFields: cogMethod methodObject
				toDepth: 0.
			 wasInYoungReferrers := cogMethod cmRefersToYoung.
			 cogit followForwardedLiteralsIn: cogMethod.
			 (wasInYoungReferrers and: [cogMethod cmRefersToYoung not]) ifTrue:
				[cogit pruneYoungReferrers]]
		ifFalse:
			[objectMemory
				followForwardedObjectFields: method
				toDepth: 0]! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 1/14/2014 15:33'!
followForwardedMethodsInMethodZone
	<inline: true>
	cogit followForwardedMethods! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 2/20/2016 19:30'!
followForwardingPointersInStackZone: theBecomeEffectsFlags
	"Spur's become: is lazy, turning the becommed object into a forwarding object to the other.
	 The read-barrier is minimised by arranging that forwarding pointers will fail a method cache
	 probe, since notionally objects' internals are accessed only via sending messages to them,
	 the exception is primitives that access the internals of the non-receiver argument(s).

	 To avoid a read barrier on bytecode, literal and inst var fetch and non-local return, we scan
	 the receivers (including the stacked receiver for non-local return) and method references
	 in the stack zone and follow any forwarded ones.  This is of course way cheaper than
	 scanning all of memory as in the old become.

	 Override to handle machine code frames"
	| theIPPtr |
	<inline: false>
	<var: #theSP type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #theIPPtr type: #usqInt>
	<var: #callerFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>

	self externalWriteBackHeadFramePointers.

	(theBecomeEffectsFlags anyMask: BecameCompiledMethodFlag) ifTrue:
		[(objectMemory isForwarded: method) ifTrue:
			[theIPPtr := instructionPointer - method.
			 method := objectMemory followForwarded: method.
			 instructionPointer := method + theIPPtr].
		(objectMemory isOopForwarded: newMethod) ifTrue:
			[newMethod := objectMemory followForwarded: newMethod]].

	self assert: stackPage ~= 0.
	0 to: numStackPages - 1 do:
		[:i| | thePage theSP theFP callerFP oop offset |
		thePage := stackPages stackPageAt: i.
		thePage isFree ifFalse:
			[self assert: (self ifCurrentStackPageHasValidHeadPointers: thePage).
			 theFP := thePage headFP.
			 "Skip the instruction pointer on top of stack of inactive pages."
			 theIPPtr := thePage = stackPage ifTrue: [0] ifFalse: [thePage headSP asUnsignedInteger].
			 [self assert: (thePage addressIsInPage: theFP).
			  self assert: (theIPPtr = 0 or: [thePage addressIsInPage: theIPPtr asVoidPointer]).
			  (self isMachineCodeFrame: theFP)
				ifTrue:
					[oop := stackPages longAt: theFP + FoxMFReceiver.
					 (objectMemory isOopForwarded: oop) ifTrue:
						[stackPages
							longAt: theFP + FoxMFReceiver
							put: (objectMemory followForwarded: oop)].
					 self assert: (objectMemory isForwarded: (self mframeHomeMethod: theFP) methodObject) not]
				ifFalse:
					[oop := stackPages longAt: theFP + FoxIFReceiver.
					 (objectMemory isOopForwarded: oop) ifTrue:
						[stackPages
							longAt: theFP + FoxIFReceiver
							put: (objectMemory followForwarded: oop)].
					 oop := self iframeMethod: theFP.
					 (objectMemory isForwarded: oop) ifTrue:
						[| newOop |
						 newOop := objectMemory followForwarded: oop.
						 offset := newOop - oop.
						 (theIPPtr ~= 0
						  and: [(stackPages longAt: theIPPtr) > oop]) ifTrue:
							[stackPages
								longAt: theIPPtr
								put: (stackPages longAt: theIPPtr) + offset].
						stackPages
							longAt: theFP + FoxIFSavedIP
							put: (stackPages longAt: theFP + FoxIFSavedIP) + offset.
						stackPages
							longAt: theFP + FoxMethod
							put: (oop := newOop)]].
			  ((self frameHasContext: theFP)
			   and: [(objectMemory isForwarded: (self frameContext: theFP))]) ifTrue:
				[stackPages
					longAt: theFP + FoxThisContext
					put: (objectMemory followForwarded: (self frameContext: theFP))].
			  offset := self frameStackedReceiverOffset: theFP.
			  oop := stackPages longAt: theFP + offset.
			  (objectMemory isOopForwarded: oop) ifTrue:
				[stackPages
					longAt: theFP + offset
					put: (objectMemory followForwarded: oop)].
			  (callerFP := self frameCallerFP: theFP) ~= 0] whileTrue:
				[theIPPtr := (theFP + FoxCallerSavedIP) asUnsignedInteger.
				 theFP := callerFP].
			 "And finally follow the saved context and the caller context."
			 theSP := thePage baseAddress - objectMemory wordSize.
			 [theSP <= thePage baseAddress] whileTrue:
				[oop := stackPages longAt: theSP.
				 (objectMemory isForwarded: oop) ifTrue:
					[stackPages longAt: theSP put: (objectMemory followForwarded: oop)].
				 theSP := theSP + objectMemory wordSize]]]! !

!CoInterpreter methodsFor: 'process primitive support' stamp: 'eem 5/7/2009 11:28'!
forProcessPrimitiveReturnToExecutivePostContextSwitch: inInterpreter
	"Do a returnToExecutive: inInterpreter postContextSwitch: true for a process primtive
	 being sure to sample the profile clock before making the switch."
	<inline: true>
	"If we are profiling, take accurate primitive measures"
	nextProfileTick > 0 ifTrue:
		[self checkProfileTick: newMethod].
	^self returnToExecutive: inInterpreter postContextSwitch: true! !

!CoInterpreter methodsFor: 'process primitive support' stamp: 'eem 8/4/2021 11:12'!
forceInterruptCheckFromHeartbeat
	"Force an interrupt check ASAP. This version is the
	 entry-point to forceInterruptCheck for the heartbeat
	 timer to allow for repeatable debugging."
	suppressHeartbeatFlag ifFalse:
		[self sqLowLevelMFence.
		 deferSmash
			ifTrue:
				[deferredSmash := true.
				 self sqLowLevelMFence]
			ifFalse:
				[self forceInterruptCheck]]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 2/23/2020 23:44'!
frameCallerContext: theFP
	"In the StackInterpreter the saved ip field of a base frame holds the
	 base frame's caller context. But in the Cog VM the first word on the
	 stack holds the base frame's caller context, which is immediately
	 above the stacked receiver."
	<var: #theFP type: #'char *'>
	| thePage callerContextOrNil |
	<var: #thePage type: #'StackPage *'>
	self assert: (self isBaseFrame: theFP).
	thePage := stackPages stackPageFor: theFP.
	callerContextOrNil := stackPages longAt: thePage baseAddress.
	self assert: (objectMemory addressCouldBeObj: callerContextOrNil).
	self assert: (callerContextOrNil = objectMemory nilObject or: [objectMemory isContext: (objectMemory followMaybeForwarded: callerContextOrNil)]).
	^callerContextOrNil! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 1/25/2018 17:26'!
frameCallerContext: theFP put: aValue
	"In the StackInterpreter the saved ip field of a base frame holds the
	 base frame's caller context. But in the Cog VM the first word on the
	 stack holds the base frame's caller context, which is immediately
	 above the stacked receiver."
	<var: #theFP type: #'char *'>
	<inline: true>
	self assert: (aValue = objectMemory nilObject or: [objectMemory isContext: aValue]).
	self assert: (self isBaseFrame: theFP).
	self assert: theFP + (self frameStackedReceiverOffset: theFP) + (2 * objectMemory wordSize) = (stackPages stackPageFor: theFP) baseAddress.
	self assert: (stackPages longAt: theFP + (self frameStackedReceiverOffset: theFP) + objectMemory wordSize) = (self frameContext: theFP).
	^stackPages
		longAt: theFP + (self frameStackedReceiverOffset: theFP) + (2 * objectMemory wordSize)
		put: aValue! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 1/18/2009 15:28'!
frameHasContext: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	^(self isMachineCodeFrame: theFP)
		ifTrue: [self mframeHasContext: theFP]
		ifFalse: [self iframeHasContext: theFP]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 1/18/2009 15:26'!
frameIsBlockActivation: theFP "<Integer>"
	<inline: true>
	<var: #theFP type: #'char *'>
	^(self isMachineCodeFrame: theFP)
		ifTrue: [self mframeIsBlockActivation: theFP]
		ifFalse: [self iframeIsBlockActivation: theFP]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 1/22/2009 15:06'!
frameMethodField: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	^stackPages longAt: theFP + FoxMethod! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 2/5/2009 14:39'!
frameMethodObject: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	^(self isMachineCodeFrame: theFP)
		ifTrue: [(self mframeHomeMethod: theFP) methodObject]
		ifFalse: [self iframeMethod: theFP]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 1/30/2009 11:32'!
frameNumArgs: theFP
	"See encodeFrameFieldHasContext:numArgs:"
	<inline: true>
	<var: #theFP type: #'char *'>
	^(self isMachineCodeFrame: theFP)
		ifTrue: [(self mframeCogMethod: theFP) cmNumArgs]
		ifFalse: [stackPages byteAt: theFP + FoxIFrameFlags + 1]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 6/23/2012 13:40'!
frameNumTemps: theFP
	"For subclasses to redefine to implement different closure semantics."
	<var: #theFP type: #'char *'>
	^0! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 1/1/2009 23:00'!
framePointer: theFP
	"Simulation only"
	<doNotGenerate>
	framePointer := theFP! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'eem 5/6/2010 16:00'!
framePointerAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: framePointer) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #framePointer in: self]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 1/29/2009 11:31'!
frameReceiver: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	^(self isMachineCodeFrame: theFP)
		ifTrue: [self mframeReceiver: theFP]
		ifFalse: [self iframeReceiver: theFP]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 4/2/2015 10:56'!
frameReceiverLocation: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	^(self isMachineCodeFrame: theFP)
		ifTrue: [theFP + FoxMFReceiver]
		ifFalse: [theFP + FoxIFReceiver]! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 3/13/2014 17:16'!
freeUnmarkedMachineCode
	"Free machine-code methods whose compiled methods are unmarked
	 (or open PICs whose selectors are not marked).
	 The stack pages have already been traced so any methods
	 of live stack activations have already been marked and traced."
	<doNotGenerate>
	cogit freeUnmarkedMachineCode! !

!CoInterpreter methodsFor: 'plugin primitives' stamp: 'eem 7/24/2021 19:33'!
functionForPrimitiveExternalCall: methodObj
	"Arrange to call the external primitive directly.  The complication is arranging
	 that the call can be flushed, given that it is embedded in machine code."
	<returnTypeC: 'void (*functionForPrimitiveExternalCall(sqInt methodObj))(void)'>
	| lit index functionPointer |
	<var: #functionPointer declareC: #'void (*functionPointer)(void)'>
	(objectMemory literalCountOf: methodObj) > 0 ifFalse:
		[^#primitiveExternalCall].
	lit := self literal: 0 ofMethod: methodObj. 
	"Check if it's an array of length 4"
	((objectMemory isArray: lit) and: [(objectMemory lengthOf: lit) = 4]) ifFalse:
		[^#primitiveExternalCall].
	index := objectMemory fetchPointer: ExternalCallLiteralTargetFunctionIndex ofObject: lit.
	((objectMemory isIntegerObject: index)
	and: [(index := objectMemory integerValueOf: index) > 0
	and: [index <= MaxExternalPrimitiveTableSize]]) ifFalse:
		[^#primitiveExternalCall].
	functionPointer := externalPrimitiveTable at: index - 1.
	functionPointer = 0 ifTrue:
		[^#primitiveExternalCall].
	^functionPointer! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 8/9/2021 11:08'!
functionPointerForCompiledMethod: methodObj primitiveIndex: primitiveIndex primitivePropertyFlagsInto: flagsPtr
	<api>
	<returnTypeC: 'void (*functionPointerForCompiledMethodprimitiveIndexprimitivePropertyFlagsInto(sqInt methodObj, sqInt primitiveIndex, sqInt *flagsPtr))(void)'>
	| functionPointer flags |
	<var: #functionPointer declareC: #'void (*functionPointer)(void)'>
	flagsPtr ifNotNil:
		[flagsPtr at: 0 put: (flags := self primitivePropertyFlags: primitiveIndex)].
	functionPointer := self functionPointerFor: primitiveIndex inClass: nil.
	functionPointer == #primitiveCalloutToFFI ifTrue:
		[^self functionForPrimitiveCallout].
	functionPointer == #primitiveExternalCall ifTrue:
		[| lit |
		 lit := self attemptToLinkExternalPrimitive: methodObj.
		 "N.B. We only support the FastCPrimitiveFlag on Spur because Spur
		  will *not* run a GC to satisfy an allocation in a primitive. The V3
		  ObjectMemory will and hence the depth of stack needed in a V3
		  primitive is probably too large to safely execute on a stack page."
		  objectMemory hasSpurMemoryManagerAPI ifTrue:
			[| metadataFlags shiftedMetadataFlags |
			 metadataFlags := objectMemory fetchPointer: ExternalCallLiteralFlagsIndex ofObject: lit.
		 	 (objectMemory isIntegerObject: metadataFlags) ifTrue:
				[shiftedMetadataFlags := ((objectMemory integerValueOf: metadataFlags)
											bitAnd: SpurPrimitiveFlagsMask)
												bitShift: PrimitiveMetadataFlagsShift.
				 shiftedMetadataFlags > 0 ifTrue:
					["Intentionally clear all other flags if there are Spur metadata flags..."
					 flags := shiftedMetadataFlags]]].
		 (self object: (objectMemory fetchPointer: ExternalCallLiteralTargetFunctionIndex ofObject: lit)
				equalsString: #primitiveProfileSemaphore) ifTrue:
			[flags := flags bitOr: PrimCallMayEndureCodeCompaction].
		 profileSemaphore ~= objectMemory nilObject ifTrue:
			[flags := flags bitOr: PrimCallCollectsProfileSamples].
		 flagsPtr at: 0 put: flags.
		 ^self functionForPrimitiveExternalCall: methodObj].
	^functionPointer! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 10/25/2013 16:24'!
gcMode
	^gcMode! !

!CoInterpreter methodsFor: 'callback support' stamp: 'eem 10/10/2019 09:56'!
getCFramePointer
	<doNotGenerate>
	^CFramePointer! !

!CoInterpreter methodsFor: 'simulation only' stamp: 'eem 4/1/2020 14:00'!
getCReturnAddress
	<doNotGenerate>
	self assert: CReturnAddress isSymbol.
	^cogit simulatedAddressFor: CReturnAddress! !

!CoInterpreter methodsFor: 'callback support' stamp: 'eem 10/10/2019 09:56'!
getCStackPointer
	<doNotGenerate>
	^CStackPointer! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 8/21/2021 19:13'!
getCodeCompactionCount
	^objectMemory integerObjectOf: statCodeCompactionCount! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 8/21/2021 19:13'!
getCodeCompactionMSecs
	^objectMemory integerObjectOf: statCodeCompactionUsecs + 500 // 1000! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 8/21/2021 19:11'!
getCogCodeSize
	^objectMemory integerObjectOf: cogCodeSize! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 3/16/2017 14:52'!
getCogCodeZoneThreshold
	<doNotGenerate>
	<returnTypeC: #double>
	^cogit getCogCodeZoneThreshold! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 7/19/2010 09:47'!
getCogMethodCount
	^objectMemory integerObjectOf: (cogMethodZone numMethodsOfType: CMMethod)! !

!CoInterpreter methodsFor: 'interpreter shell' stamp: 'eem 6/25/2011 12:29'!
getCurrentBytecode
	"currentBytecode will be private to the main dispatch loop in the generated code.
	 This method allows the currentBytecode to be retrieved from global variables.
	 Override to answer -1 if we're not in an interpreter frame."

	^((stackPages couldBeFramePointer: framePointer)
	   and: [(self isMachineCodeFrame: framePointer) not])
		ifTrue: [objectMemory byteAt: instructionPointer]
		ifFalse: [-1]! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
getDesiredCogCodeSize
	<cmacro: '() integerObjectOf(desiredCogCodeSize)'>
	^objectMemory integerObjectOf: desiredCogCodeSize! !

!CoInterpreter methodsFor: 'accessing' stamp: 'eem 4/6/2015 22:40'!
getFramePointer
	<api>
	<option: #NewspeakVM>
	^framePointer! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 5/7/2014 08:33'!
getGCMode
	^gcMode! !

!CoInterpreter methodsFor: 'image save/restore' stamp: 'eem 10/18/2020 17:23'!
getImageHeaderFlags
	"Answer the flags that are contained in the 7th long of the image header."
	^fullScreenFlag "0 or 1"
	+ (VMBIGENDIAN ifTrue: [0] ifFalse: [2]) "this is the imageFloatsLittleEndian flag"
	+ (flagInterpretedMethods ifTrue: [8] ifFalse: [0])
	+ (preemptionYields ifTrue: [0] ifFalse: [16r10])
	+ (newFinalization ifTrue: [16r40] ifFalse: [0])
	+ (sendWheelEvents ifTrue: [16r80] ifFalse: [0])
	+ (primitiveDoMixedArithmetic ifTrue: [0] ifFalse: [16r100])
	+ (imageHeaderFlags bitClear: 1+2+8+16r10+16r40+16r80+16r100) "these are any flags we do not recognize"! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
handleForwardedSendFaultForReceiver: forwardedReceiver stackDelta: stackDelta
	"Handle a send fault that may be due to a send to a forwarded object.
	 Unforward the receiver on the stack and answer it."
	<option: #SpurObjectMemory>
	| rcvrStackIndex rcvr |
	<inline: false>
	"should *not* be a super send, so the receiver should be forwarded."
	self assert: (objectMemory isOopForwarded: forwardedReceiver).
	rcvrStackIndex := argumentCount + stackDelta.
	self assert: (self stackValue: rcvrStackIndex) = forwardedReceiver.
	rcvr := objectMemory followForwarded: forwardedReceiver.
	self stackValue: rcvrStackIndex put: rcvr.
	self followForwardedFrameContents: framePointer
		stackPointer: stackPointer + (rcvrStackIndex + 1 * objectMemory wordSize). "don't repeat effort"
	(objectMemory isPointers: (self frameReceiver: framePointer)) ifTrue:
		[objectMemory
			followForwardedObjectFields: (self frameReceiver: framePointer)
			toDepth: 0].
	self followForwardedFieldsInCurrentMethod.
	^rcvr! !

!CoInterpreter methodsFor: 'message sending' stamp: 'rmacnak 6/24/2015 21:19'!
handleMNU: selectorIndex InMachineCodeTo: rcvr classForMessage: classForMessage
	"A message send from either an open PIC or an unlinked send has not  been
	 understood.  Create a message and execute the relevant resulting MNU method.
	 messageSelector is an implicit argument (yuck)."
	| errSelIdx classForThisMessage |
	self assert: (objectMemory addressCouldBeOop: rcvr).
	instructionPointer := self popStack.
	self createActualMessageTo: classForMessage.
	messageSelector := objectMemory splObj: selectorIndex.
	(self lookupInMethodCacheSel: messageSelector classTag: (objectMemory classTagForClass: lkupClass))
		ifTrue:"check for coggability because method is in the cache"
			[self
				ifAppropriateCompileToNativeCode: newMethod
				selector: messageSelector]
		ifFalse:
			[errSelIdx := self lookupMNUInClass: (classForThisMessage := lkupClass).
			 errSelIdx ~= 0 ifTrue:
				[selectorIndex = SelectorDoesNotUnderstand ifTrue:
					[self error: 'Recursive not understood error encountered'].
				 self push: instructionPointer.
				 ^self handleMNU: errSelIdx InMachineCodeTo: rcvr classForMessage: classForThisMessage]].
	(self maybeMethodHasCogMethod: newMethod) ifTrue:
		[self push: instructionPointer.
		 self executeCogMethod: (self cogMethodOf: newMethod)
			 fromUnlinkedSendWithReceiver: rcvr.
		 "NOTREACHED"
		 self assert: false].
	^self interpretMethodFromMachineCode
	"NOTREACHED"! !

!CoInterpreter methodsFor: 'accessing' stamp: 'eem 8/12/2014 15:41'!
heapBase
	<cmacro: '() heapBase'>
	^heapBase! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
ifAppropriateCompileToNativeCode: aMethodObj selector: selector
	| methodHeader cogMethod |
	<inline: true>
	<var: #cogMethod type: #'CogMethod *'>
	methodHeader := self rawHeaderOf: aMethodObj.
	(self isCogMethodReference: methodHeader)
		ifTrue: "makeBaseFrame: can create cog methods with nil selectors."
			[cogMethod := self cCoerceSimple: methodHeader to: #'CogMethod *'.
			 cogMethod selector = objectMemory nilObject ifTrue:
				[cogit setSelectorOf: cogMethod to: selector]]
		ifFalse:
			[(self methodWithHeaderShouldBeCogged: methodHeader)
				ifTrue: [cogit cog: aMethodObj selector: selector]
				ifFalse: [self maybeFlagMethodAsInterpreted: aMethodObj]]! !

!CoInterpreter methodsFor: 'jump bytecodes' stamp: 'eem 3/24/2017 15:14'!
ifBackwardsCheckForEvents: offsetToJumpBytecode
	"Backward jump means we're in a loop.
		- check for possible interrupts.
		- check for long-running loops and JIT if appropriate."
	| switched backwardJumpCountByte |
	<inline: true>
	offsetToJumpBytecode >= 0 ifTrue:
		[^self].

	localSP < stackLimit ifTrue:
		[self externalizeIPandSP.
		 switched := self checkForEventsMayContextSwitch: true.
		 self returnToExecutive: true postContextSwitch: switched.
		 self browserPluginReturnIfNeeded.
		 self internalizeIPandSP.
		 switched ifTrue:
			[^self]].

	"We use the least significant byte of the flags word (which is marked as an immediate) and
	 subtract two each time to avoid disturbing the least significant tag bit.  Since the byte is
	 initialized to 1 (on frame build), on first decrement it will become -1.  Trip when it reaches 1 again."
	backwardJumpCountByte := self iframeBackwardBranchByte: localFP.
	(backwardJumpCountByte := backwardJumpCountByte - 2) = 1
		ifTrue:
			[(self methodWithHeaderShouldBeCogged: (objectMemory methodHeaderOf: method)) ifTrue:
				[self externalizeIPandSP.
				 self attemptToSwitchToMachineCode: (self oopForPointer: localIP) - offsetToJumpBytecode - method - objectMemory baseHeaderSize - 1
				 "If attemptToSwitchToMachineCode: returns the method could not be cogged, hence..."].
			 "can't cog method; avoid asking to cog it again for the longest possible time."
			 backwardJumpCountByte := 16r7F]
		ifFalse:
			[backwardJumpCountByte = -1 ifTrue: "initialize the count"
				[self assert: minBackwardJumpCountForCompile <= 128.
				 backwardJumpCountByte := minBackwardJumpCountForCompile - 1 << 1 + 1]].
	self iframeBackwardBranchByte: localFP put: backwardJumpCountByte! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 2/21/2011 14:15'!
ifValidWriteBackStack: theCFP Pointers: theCSP Save: savedFPP To: savedSPP
	"This is for low-level error reporting.  If either of the C stack pointers are
	 pointing into the stack zone then write them back to framePointer and/or
	 stackPointer so that the stack backtrace will be up to date.  Write their
	 original values through savedFPP & savedSPP if non-null."
	<api>
	<var: #theCFP type: #'void *'>
	<var: #theCSP type: #'void *'>
	<var: #savedFPP type: #'char **'>
	<var: #savedSPP type: #'char **'>
	<returnTypeC: #void>
	savedFPP ~= 0 ifTrue:
		[savedFPP at: 0 put: framePointer].
	savedSPP ~= 0 ifTrue:
		[savedSPP at: 0 put: stackPointer].
	(self couldBeFramePointer: theCFP) ifTrue:
		[framePointer := theCFP].
	(self couldBeFramePointer: theCSP) ifTrue:
		[stackPointer := theCSP]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
iframeBackwardBranchByte: theFP
	"See encodeFrameFieldHasContext:numArgs: and ifBackwardsCheckForEvents:"
	<inline: true>
	<var: #theFP type: #'char *'>
	^stackPages byteAt: theFP + (VMBIGENDIAN ifTrue: [FoxIFrameFlags + objectMemory wordSize - 1] ifFalse: [FoxIFrameFlags])! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
iframeBackwardBranchByte: theFP put: aByte
	"See encodeFrameFieldHasContext:numArgs: and ifBackwardsCheckForEvents:"
	<inline: true>
	<var: #theFP type: #'char *'>
	stackPages
		byteAt: theFP + (VMBIGENDIAN ifTrue: [FoxIFrameFlags + objectMemory wordSize - 1] ifFalse: [FoxIFrameFlags])
		put: aByte! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 1/18/2009 15:27'!
iframeHasContext: theFP
	"See encodeFrameFieldHasContext:numArgs:"
	<inline: true>
	<var: #theFP type: #'char *'>
	^(stackPages byteAt: theFP + FoxIFrameFlags + 2) ~= 0! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 1/18/2009 16:00'!
iframeIsBlockActivation: theFP "<Integer>"
	<inline: true>
	<var: #theFP type: #'char *'>
	^(stackPages byteAt: theFP + FoxIFrameFlags + 3) ~= 0! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 12/31/2008 12:08'!
iframeNumArgs: theFP
	"See encodeFrameFieldHasContext:numArgs:"
	<inline: true>
	<var: #theFP type: #'char *'>
	^stackPages byteAt: theFP + FoxIFrameFlags + 1! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 12/31/2008 11:46'!
iframeReceiver: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	^stackPages longAt: theFP + FoxIFReceiver! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 11/14/2020 21:31'!
iframeSavedIP: theFP
	<var: #theFP type: #'char *'>
	^(stackPages longAt: theFP + FoxIFSavedIP) asUnsignedInteger! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 10/9/2010 23:03'!
iframeSavedIP: theFP put: savedIP
	<var: #theFP type: #'char *'>
	self assert: (self isMachineCodeFrame: theFP) not.
	stackPages longAt: theFP + FoxIFSavedIP put: savedIP! !

!CoInterpreter methodsFor: 'initialization' stamp: 'eem 7/1/2021 11:51'!
initStackPagesAndInterpret
	"Initialize the stack pages and enter interpret. Use alloca'ed memory so that when
	 we have a JIT its stack pointer will be on the native stack since alloca allocates
	 memory on the stack. Certain thread systems use the native stack pointer as the
	 frame ID so putting the stack anywhere else can confuse the thread system."

	"Override to establish the setjmp/longjmp handler for reentering the interpreter
	 from machine code, and disable executablity on the heap and stack pages."

	"This should be in its own initStackPages method but Slang can't inline
	 C code strings."
	| stackPageBytes stackPagesBytes theStackMemory |
	<var: #theStackMemory type: #'char *'>
	stackPageBytes := self stackPageByteSize.
	stackPagesBytes := self computeStackZoneSize.
	theStackMemory := self
							cCode: [self alloca: stackPagesBytes]
							inSmalltalk: [stackPages initializeWithByteSize: stackPagesBytes for: self].
	self cCode: [self memset: theStackMemory _: 0 _: stackPagesBytes].
	stackPages
		initializeStack: theStackMemory
		numSlots: stackPagesBytes / objectMemory wordSize
		pageSize: stackPageBytes / objectMemory wordSize.
	self assert: self minimumUnusedHeadroom = stackPageBytes.

	"Once the stack pages are initialized we can continue to bootstrap the system."
	self loadInitialContext.
	"We're ready for the heartbeat (poll interrupt)"
	self ioInitHeartbeat.
	self initialEnterSmalltalkExecutive.
	^nil! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 11/3/2014 10:05'!
instVar: offset ofContext: aContext
	"Fetch an instance avriable from a maybe married context.
	 If the context is still married compute the value of the
	 relevant inst var from the spouse frame's state.

	 If the context is single but has a negative instruction pointer
	 recognise that the instruction pointer is actually into machine
	 code and convert it to the corresponding bytecode pc."
	| value spouseFP |
	<var: #spouseFP type: #'char *'>
	<inline: true>
	self assert: offset < MethodIndex.
	self assert: (objectMemory isContext: aContext).
	self writeBackHeadFramePointers.
	(self isMarriedOrWidowedContext: aContext) ifFalse:
		[value := objectMemory fetchPointer: offset ofObject: aContext.
		 (offset = InstructionPointerIndex
		  and: [(objectMemory isIntegerObject: value)
		  and: [value signedIntFromLong < 0]]) ifTrue:
			[value := self internalMustMapMachineCodePC: (objectMemory integerValueOf: value)
						context: aContext].
		 ^value].

	(self isWidowedContext: aContext) ifTrue:
		[^objectMemory fetchPointer: offset ofObject: aContext].

	spouseFP := self frameOfMarriedContext: aContext.
	offset = SenderIndex ifTrue:
		[^self ensureCallerContext: spouseFP].
	offset = StackPointerIndex ifTrue:
		[self assert: ReceiverIndex + (self stackPointerIndexForFrame: spouseFP) < (objectMemory lengthOf: aContext).
		^objectMemory integerObjectOf: (self stackPointerIndexForFrame: spouseFP)].
	offset = InstructionPointerIndex ifTrue:
		[^self instructionPointerForFrame: spouseFP currentFP: localFP currentIP: (self oopForPointer: localIP)].
	self error: 'bad index'.
	^0! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 2/2/2009 11:57'!
instructionPointer
	<doNotGenerate>
	^instructionPointer! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 2/12/2009 13:20'!
instructionPointer: aValue
	<doNotGenerate>
	instructionPointer := aValue! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'eem 5/6/2010 16:00'!
instructionPointerAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: instructionPointer) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #instructionPointer in: self]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 12/8/2017 12:49'!
instructionPointerForFrame: spouseFP currentFP: currentFP currentIP: instrPtr
	"Answer the bytecode pc object (i.e. SmallInteger) for an active frame.  The bytecode
	 pc is derived from the frame's pc.  If the frame is the top frame on the current stack
	 the frame pc is whatever the current instruction pointer is.  If the frame is the top
	 frame on some other stack the frame pc is the value on top of stack.  Otherwise the
	 frame pc is the saved pc of the frame above.  Once the frame pc is found it must be
	 mapped to a bytecode pc."
	<var: #spouseFP type: #'char *'>
	<var: #currentFP type: #'char *'>
	| value theIP thePage theFPAbove |
	<var: #thePage type: #'StackPage *'>
	<var: #theFPAbove type: #'char *'>
	spouseFP = currentFP
		ifTrue: [theIP := self oopForPointer: instrPtr]
		ifFalse:
			[thePage := stackPages stackPageFor: spouseFP.
			 theFPAbove := self findFrameAbove: spouseFP inPage: thePage.
			 theIP := theFPAbove = 0
						ifTrue: [stackPages longAt: thePage headSP]
						ifFalse:[self oopForPointer: (self frameCallerSavedIP: theFPAbove)]].
	value := self contextInstructionPointer: theIP frame: spouseFP.
	^value signedIntFromLong < 0
		ifTrue: [self mustMapMachineCodePC: (objectMemory integerValueOf: value)
					context: (self frameContext: spouseFP)]
		ifFalse: [value]! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem 10/26/2018 12:56'!
internalActivateNewMethod
	| methodHeader numTemps rcvr switched |
	<inline: true>

	methodHeader := self rawHeaderOf: newMethod.
	self assert: (self isCogMethodReference: methodHeader) not.
	numTemps := self temporaryCountOfMethodHeader: methodHeader.
	self assert: argumentCount = (self argumentCountOfMethodHeader: methodHeader).
	rcvr := self internalStackValue: argumentCount. "could new rcvr be set at point of send?"
	self assert: (objectMemory isOopForwarded: rcvr) not.

	self internalPush: localIP.
	self internalPush: localFP.
	localFP := localSP.
	self internalPush: newMethod.
	self setMethod: newMethod methodHeader: methodHeader.
	self internalPush: objectMemory nilObject. "FxThisContext field"
	self internalPush: (self
						encodeFrameFieldHasContext: false
						isBlock: false
						numArgs: (self argumentCountOfMethodHeader: methodHeader)).
	self internalPush: 0. "FoxIFSavedIP"
	self internalPush: rcvr.

	"Initialize temps..."
	argumentCount + 1 to: numTemps do:
		[:i | self internalPush: objectMemory nilObject].

	"-1 to account for pre-increment in fetchNextBytecode"
	localIP := self pointerForOop: (self initialIPForHeader: methodHeader method: newMethod) - 1.

	(self methodHeaderHasPrimitive: methodHeader) ifTrue:
		["Skip the CallPrimitive bytecode, if it's there, and store the error code if the method starts
		  with a long store temp.  Strictly no need to skip the store because it's effectively a noop."
		 localIP := localIP + (self sizeOfCallPrimitiveBytecode: methodHeader).
		 primFailCode ~= 0 ifTrue:
			[self reapAndResetErrorCodeTo: localSP header: methodHeader]].

	self assert: (self frameNumArgs: localFP) = argumentCount.
	self assert: (self frameIsBlockActivation: localFP) not.
	self assert: (self frameHasContext: localFP) not.

	"Now check for stack overflow or an event (interrupt, must scavenge, etc)."
	localSP < stackLimit ifTrue:
		[self externalizeIPandSP.
		 switched := self handleStackOverflowOrEventAllowContextSwitch:
						(self canContextSwitchIfActivating: newMethod header: methodHeader).
		 self returnToExecutive: true postContextSwitch: switched.
		 self internalizeIPandSP]! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem 5/19/2020 16:18'!
internalExecuteNewMethod
	<inline: true>
	"For interpreter performance and to ease the objectAsMethod implementation eagerly
	 evaluate the primitive, i.e. if the method is cogged and has a primitive /do not/ evaluate
	 the machine code primitive, just evaluate primitiveFunctionPointer directly."
	| succeeded methodHeader |
	primitiveFunctionPointer ~= 0 ifTrue:
		[self isPrimitiveFunctionPointerAnIndex ifTrue:
			[^self internalQuickPrimitiveResponse].
		 "slowPrimitiveResponse may of course context-switch.  If so we must reenter the
		  new process appropriately, returning only if we've found an interpreter frame."
		 self externalizeIPandSP.
		 succeeded := self slowPrimitiveResponse.
		 instructionPointer = cogit ceReturnToInterpreterPC ifTrue:
			[instructionPointer := self iframeSavedIP: framePointer].
		 self internalizeIPandSP.
		 succeeded ifTrue:
			[self return: self popStack toExecutive: true.
			 self browserPluginReturnIfNeeded.
			^nil]].
	methodHeader := self rawHeaderOf: newMethod.
	"if not primitive, or primitive failed, activate the method"
	(self isCogMethodReference: methodHeader) ifFalse:
		[^self internalActivateNewMethod].
	self iframeSavedIP: localFP put: localIP asInteger.
	instructionPointer := cogit ceReturnToInterpreterPC.
	self externalizeFPandSP.
	"This may context switch and hence return..."
	self activateNewCogMethod: (self cCoerceSimple: methodHeader to: #'CogMethod *') inInterpreter: true.
	"Hence this si reachable..."
	self internalizeIPandSP! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem 2/20/2017 13:16'!
internalFindNewMethodOrdinary
	"Find the compiled method to be run when the current messageSelector is
	 sent to the given class, setting the values of newMethod and primitiveIndex."
	| ok |
	<inline: true>
	ok := self inlineLookupInMethodCacheSel: messageSelector classTag: lkupClassTag.
	ok	ifTrue:
			[self ifAppropriateCompileToNativeCode: newMethod selector: messageSelector]
		ifFalse:
			[self externalizeIPandSP.
			 ((objectMemory isOopForwarded: messageSelector)
			  or: [objectMemory isForwardedClassTag: lkupClassTag]) ifTrue:
				[(objectMemory isOopForwarded: messageSelector) ifTrue:
					[messageSelector := self handleForwardedSelectorFaultFor: messageSelector].
				 (objectMemory isForwardedClassTag: lkupClassTag) ifTrue:
					[lkupClassTag := self handleForwardedSendFaultForTag: lkupClassTag].
				(self lookupInMethodCacheSel: messageSelector classTag: lkupClassTag) ifTrue:
					[^self ifAppropriateCompileToNativeCode: newMethod selector: messageSelector]].
			lkupClass := objectMemory classForClassTag: lkupClassTag.
			self assert: (lkupClass notNil and: [self addressCouldBeClassObj: lkupClass]).
			NewspeakVM
				ifTrue: [self lookupOrdinarySend]
				ifFalse: [self lookupMethodInClass: lkupClass].
			self internalizeIPandSP.
			self addNewMethodToCache: lkupClass]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 6/18/2009 16:59'!
internalMustMapMachineCodePC: theIP context: aOnceMarriedContext
	"Must externalize before calling mustMapMachineCodePC:context:
	 because it may cause a code compaction."
	| result |
	self externalizeIPandSP.
	result := self mustMapMachineCodePC: theIP context: aOnceMarriedContext.
	self internalizeIPandSP.
	^result! !

!CoInterpreter methodsFor: 'utilities' stamp: 'eem 09/27/2016 11:20'!
internalizeIPandSP
	"Copy the instruction, stack and frame pointers to local variables for rapid access within the interpret loop."

	self assert: instructionPointer ~= cogit ceReturnToInterpreterPC.
	localIP := self pointerForOop: instructionPointer.
	localSP := self pointerForOop: stackPointer.
	localFP := self pointerForOop: framePointer.
	LowcodeVM ifTrue: [ nativeSP := 0. ]! !

!CoInterpreter methodsFor: 'interpreter shell' stamp: 'eem 1/5/2021 15:48'!
interpret
	"This is the main interpreter loop.
	 In a pure interpreter it loops forever, fetching and executing bytecodes.
	 With the Cogit JIT executing code as well, the interpreter is reentered from machine code
	 whenever the machine code wants to interpret a method instead of executing its machine
	 code.  Entry into the interpreter is done via a ''jump call'' in machine code that uses
	 CFramePointer and CStackPointer to find the base of the C stack (set in CoInterpreter>>
	 enterSmalltalkExecutiveImplementation) and substitutes CReturnAddress as the return
	 address in the code so it always appears that interpret has been called from
	 CoInterpreter>>enterSmalltalkExecutiveImplementation, which may be important to,
	 for example, C exception handling inside the VM.

	 When running in the context of a browser plugin VM the interpreter must return control
	 to the browser periodically. This should done only when the state of the currently running
	 Squeak thread is safely stored in the object heap. Since this is the case at the moment
	 that a check for interrupts is performed, that is when we return to the browser if it is time
	 to do so. Interrupt checks happen quite frequently."

	<api>
	<inline: false>
	"If stacklimit is zero then the stack pages have not been initialized."
	stackLimit = 0 ifTrue:
		[^self initStackPagesAndInterpret].
	"An unchecked write is probably faster, so instead of
	 CReturnAddress ifNil:
		[CReturnAddress := self cCoerceSimple: self getReturnAddress to: #usqIntptr_t]
	 we have simply"
	self assert: (CReturnAddress isNil or: [CReturnAddress = (self cCoerceSimple: self getReturnAddress to: #usqIntptr_t)]).
	CReturnAddress := self cCoerceSimple: self getReturnAddress to: #usqIntptr_t.

	self useCogitBreakBlockIfNone.
	"record entry time when running as a browser plug-in"
	self browserPluginInitialiseIfNeeded.
	self setMethod: (self iframeMethod: framePointer).
	self deny: instructionPointer = cogit ceReturnToInterpreterPC.
	self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer imbar: true line: #'__LINE__'.
	self internalizeIPandSP.
	self initExtensions.
	self fetchNextBytecode.
	[true] whileTrue:
		[self aboutToDispatchBytecode.
		 self dispatchOn: currentBytecode in: BytecodeTable].
	localIP := localIP - 1.  "undo the pre-increment of IP before returning"
	self externalizeIPandSP.
	^nil! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'eem 12/12/2012 11:34'!
interpretAddress
	"This is used for asserts that check that inline cache editing results in valid addresses.
	 In the C VM interpret is presumed to come before any primitives and so it constitutes
	 the lowest address in C code that machine code should be linked.  In the simulator
	 we just answer something not low."
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: #interpret) asUnsignedInteger]
		inSmalltalk: [heapBase]! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem 3/29/2020 17:26'!
interpretMethodFromMachineCode
	"Execute a method interpretively from machine code.  We assume (require) that newMethod
	 messageSelector, primitiveFunctionPointer and argumentCount have been set in the caller.
	 Once evaluated either continue in the interpreter via a jongjmp or in machine code via an
	 enilopmart (a form of longjmp - a stinking rose by any other name)."
	<inline: false>
	cogit assertCStackWellAligned.
	self assert: (self validInstructionPointer: instructionPointer inFrame: framePointer).
	primitiveFunctionPointer ~= 0
		ifTrue:
			[primitiveFunctionPointer = #primitiveInvokeObjectAsMethod
				ifTrue: [self assert: (objectMemory isOopCompiledMethod: newMethod) not]
				ifFalse: [self assert: ((objectMemory isOopCompiledMethod: newMethod)
									  and: [(self primitiveIndexOf: newMethod) ~= 0])].
			 "Invoke an interpreter primitive (because the method is to be interpreted or has not yet been
			  compiled).  This is very similar to invoking an interpreter primitive from a compiled primitive
			  (see e.g. SimpleStackBasedCogit>>compileInterpreterPrimitive:).  Cut back the stack pointer
			  (done above) to skip the return address and invoke the function.  On return if it has succeeded
			  simply continue otherwise restore the stackPointer, collect the pc and interpret.  Note that
			  frame building primitives such as primitiveClosureValue, primitiveEvaluateMethod et al will not
			  return but will instead jump into either machine code or longjmp back to the interpreter."
			"Assign stackPage headFP so we can tell if the primitive built a frame.  We can't simply save
			 the framePointer since e.g. assignment to contexts (via primitiveInstVarAt:put:) can change the
			 framePointer.  But context assignments will change both the framePointer and stackPage headFP."
			
			 self assert: (framePointer < stackPage baseAddress
						and: [framePointer > (stackPage realStackLimit - (LargeContextSlots * objectMemory bytesPerOop / 2))]).
			 stackPage headFP: framePointer.
			 self isPrimitiveFunctionPointerAnIndex
				ifTrue:
					[self externalQuickPrimitiveResponse.
					 primFailCode := 0]
				ifFalse:
					[self slowPrimitiveResponse].
			self successful ifTrue:
				[self return: self popStack toExecutive: false
				 "NOTREACHED"]]
		ifFalse:
			[self assert: ((objectMemory isOopCompiledMethod: newMethod)
						   and: [(self primitiveIndexOf: newMethod) = 0
								or: [(self functionPointerFor: (self primitiveIndexOf: newMethod) inClass: objectMemory nilObject) = 0
								or: [self isNullExternalPrimitiveCall: newMethod]]])].
	"if not primitive, or primitive failed, activate the method and reenter the interpreter"
	self activateNewMethod.
	cogit ceInvokeInterpret.
	"NOTREACHED"
	^nil! !

!CoInterpreter methodsFor: 'stack pages' stamp: 'eem 10/27/2016 15:54'!
interpreterAllocationReserveBytes
	"At a rough approximation we may need to allocate up to a couple
	 of page's worth of contexts when switching stack pages, assigning
	 to senders, etc.  But the snapshot primitive voids all stack pages.
	 So a safe margin is the size of a large context times the maximum
	 number of frames per page times the number of pages."
	<inline: #never>
	| maxUsedBytesPerPage maxFramesPerPage |
	maxUsedBytesPerPage := self stackPageFrameBytes + self stackLimitOffset.
	maxFramesPerPage := maxUsedBytesPerPage / objectMemory wordSize // MFrameSlots.
	^maxFramesPerPage * LargeContextSlots * objectMemory bytesPerOop * numStackPages! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 3/21/2017 15:05'!
is: fieldIndex methodAssignmentToContextWithMachineCodePC: anOop
	"If the method is assigned, any machine code pc must be mapped to a bytecode one
	 before the method is changed."
	<inline: true>
	| thePC |
	^fieldIndex = MethodIndex
	  and: [(thePC := objectMemory fetchPointer: InstructionPointerIndex ofObject: anOop) signedIntFromLong < 0
	  and: [objectMemory isIntegerObject: thePC]]! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 11/11/2020 18:38'!
isCFramePointerInUse: cFrmPtrPtr _: cStkPtrPtr
	<var: 'cFrmPtrPtr' type: #'usqIntptr_t *'>
	<var: 'cStkPtrPtr' type: #'usqIntptr_t *'>
	<doNotGenerate>
	"This must be implemented externally, e.g. in sqPlatMain.c."
	^self class initializationOptions at: #isCFramePointerInUse ifAbsent: [true]! !

!CoInterpreter methodsFor: 'primitive support' stamp: 'eem 9/4/2021 17:27'!
isCodeCompactingPrimitiveIndex: primIndex
	"If instVarAt:, slotAt: or shallowCopy operate on a Context then they compute a
	 bytecode pc and hence may provoke a code compaction. Hence primitive invocation
	 from these primitives must use a static return address (cePrimReturnEnterCogCode:).
	 Note that the process switch primitives may also provoke a code compaction, which
	 happens when switching to a process whose top context has a machine code pc but
	 the method is no longer in the code cache.  However, in this case they are switching
	 process and don't go through the normal return. So we don't include them here."
	<inline: true>
	self cCode: [] inSmalltalk: [#primitiveClone. #primitiveInstVarAt. #primitiveSlotAt. #primitiveFlushExternalPrimitives. #primitiveUnloadModule]. "For senders..."
	^primIndex = PrimNumberInstVarAt
	or: [primIndex = PrimNumberShallowCopy
	or: [primIndex = PrimNumberSlotAt
	or: [primIndex = PrimNumberFlushExternalPrimitives
	or: [primIndex = PrimNumberUnloadModule]]]]! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 2/2/2009 16:19'!
isCog
	^true! !

!CoInterpreter methodsFor: 'process primitive support' stamp: 'eem 1/8/2011 09:38'!
isCogCompiledCodeCompactionCalledFor
	"For in-image tests"
	^cogCompiledCodeCompactionCalledFor! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem 7/5/2011 16:25'!
isCogMethodReference: methodHeader
	<api>
	self assert: ((objectMemory isIntegerObject: methodHeader)
				or: [methodHeader asUnsignedInteger < objectMemory startOfMemory
					and: [methodHeader asUnsignedInteger >= cogit minCogMethodAddress]]).
	^objectMemory isNonIntegerObject: methodHeader! !

!CoInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 7/25/2021 14:39'!
isExternalMethodInPlugin: methodObj
	| header primIdx literal |
	header := objectMemory methodHeaderOf: methodObj.
	primIdx := self primitiveIndexOfMethod: methodObj header: header.
	^primIdx = PrimNumberExternalCall
	 and: [(objectMemory literalCountOfMethodHeader: header) > 0
	 and: [(objectMemory isArray: (literal := self literal: 0 ofMethod: methodObj))
	 and: [(objectMemory numSlotsOf: literal) = 4
	 and: [(objectMemory isBytes: (objectMemory fetchPointer: ExternalCallLiteralModuleNameIndex ofObject: literal))
	 and: [(objectMemory numBytesOf: (objectMemory fetchPointer: ExternalCallLiteralModuleNameIndex ofObject: literal)) > 0]]]]] "A 0 byte module name implies a primitive in the main VM; these can't change"! !

!CoInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 7/24/2021 17:07'!
isLinkedExternalPrimitive: methodObj
	| header primIdx literal targetFunctionIndex |
	header := objectMemory methodHeaderOf: methodObj.
	primIdx := self primitiveIndexOfMethod: methodObj header: header.
	^primIdx = PrimNumberExternalCall
	 and: [(objectMemory literalCountOfMethodHeader: header) > 0
	 and: [(objectMemory isArray: (literal := self literal: 0 ofMethod: methodObj))
	 and: [(objectMemory numSlotsOf: literal) = 4
	 and: [targetFunctionIndex := objectMemory fetchPointer: ExternalCallLiteralTargetFunctionIndex ofObject: literal.
		(objectMemory isIntegerObject: targetFunctionIndex)
	 and: [(objectMemory integerValueOf: targetFunctionIndex) > 0]]]]] "A 0 byte module name implies a primitive in the main VM; these can't change"! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
isMachineCodeFrame: theFP 
	<var: #theFP type: #'char *'>
	^(stackPages longAt: theFP + FoxMethod) asUnsignedInteger < objectMemory startOfMemory! !

!CoInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 7/24/2021 14:44'!
isMachineCodeFrameForExternalPrimitiveMethod: theFP
	<var: #theFP type: #'char *'>
	| homeMethod |
	^(self isMachineCodeFrame: theFP)
	  and: [(cogit cogMethodHasMachineCodePrim: (homeMethod := self mframeHomeMethod: theFP))
	  and: [self isExternalMethodInPlugin: homeMethod methodObject]]! !

!CoInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 7/24/2021 14:27'!
isMachineCodeFrameWithCogMethod: theFP
	<var: #theFP type: #'char *'>
	^(self isMachineCodeFrame: theFP)
	  and: [cogit cogMethodHasMachineCodePrim: (self mframeHomeMethod: theFP)]! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 11/28/2012 10:35'!
isMachineCodeIP: anInstrPointer
	^anInstrPointer < objectMemory startOfMemory! !

!CoInterpreter methodsFor: 'simulation' stamp: 'eem 12/25/2010 14:22'!
isThreadedVM
	<doNotGenerate>
	^false! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
itemporary: offset in: theFP
	"Temporary access for an interpreter frame only."
	"See StackInterpreter class>>initializeFrameIndices"
	| frameNumArgs |
	<inline: true>
	<var: #theFP type: #'char *'>
	^offset < (frameNumArgs := self iframeNumArgs: theFP)
		ifTrue: [stackPages longAt: theFP + FoxCallerSavedIP + ((frameNumArgs - offset) * objectMemory wordSize)]
		ifFalse: [stackPages longAt: theFP + FoxIFReceiver - objectMemory wordSize + ((frameNumArgs - offset) * objectMemory wordSize)]! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
itemporary: offset in: theFP put: valueOop
	"Temporary access for an interpreter frame only."
	"See StackInterpreter class>>initializeFrameIndices"
	| frameNumArgs |
	<inline: true>
	<var: #theFP type: #'char *'>
	^offset < (frameNumArgs := self iframeNumArgs: theFP)
		ifTrue: [stackPages longAt: theFP + FoxCallerSavedIP + ((frameNumArgs - offset) * objectMemory wordSize) put: valueOop]
		ifFalse: [stackPages longAt: theFP + FoxIFReceiver - objectMemory wordSize + ((frameNumArgs - offset) * objectMemory wordSize) put: valueOop]! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem 8/23/2018 11:55'!
justActivateNewMethod: mustBeInterpreterFrame
	| methodHeader cogMethod numArgs numTemps rcvr initialIP |
	<var: #cogMethod type: #'CogMethod *'>
	<var: #initialIP type: #usqInt>
	<inline: true>
	methodHeader := self rawHeaderOf: newMethod.
	(mustBeInterpreterFrame not
	 and: [self isCogMethodReference: methodHeader]) ifTrue:
		[cogMethod := self cCoerceSimple: methodHeader to: #'CogMethod *'.
		 methodHeader := cogMethod methodHeader].
	numTemps := self temporaryCountOfMethodHeader: methodHeader.
	numArgs := self argumentCountOfMethodHeader: methodHeader.

	rcvr := self stackValue: numArgs. "could new rcvr be set at point of send?"
	self assert: (objectMemory isOopForwarded: rcvr) not.

	(cogMethod notNil
	and: [instructionPointer asUnsignedInteger >= objectMemory startOfMemory]) ifTrue:
		[self iframeSavedIP: framePointer put: instructionPointer.
		 instructionPointer := cogit ceReturnToInterpreterPC].
	self push: instructionPointer.
	self push: framePointer.
	framePointer := stackPointer.
	initialIP := self initialIPForHeader: methodHeader method: newMethod.
	cogMethod
		ifNotNil:
			[self push: cogMethod asUnsignedInteger.
			 self push: objectMemory nilObject. "FoxThisContext field"
			 instructionPointer := cogMethod asUnsignedInteger + cogMethod stackCheckOffset]
		ifNil:
			[self push: newMethod.
			 self setMethod: newMethod methodHeader: methodHeader.
			 self push: objectMemory nilObject. "FoxThisContext field"
			 self push: (self encodeFrameFieldHasContext: false isBlock: false numArgs: numArgs).
			 self push: 0. "FoxIFSavedIP"
			 instructionPointer := initialIP - 1].
	self push: rcvr.

	"clear remaining temps to nil"
	numArgs+1 to: numTemps do:
		[:i | self push: objectMemory nilObject].

	(self methodHeaderHasPrimitive: methodHeader) ifTrue:
		["Skip the CallPrimitive bytecode, if it's there, and store the error code if the method starts
		  with a long store temp.  Strictly no need to skip the store because it's effectively a noop."
		 cogMethod ifNil:
			[instructionPointer := instructionPointer + (self sizeOfCallPrimitiveBytecode: methodHeader)].
		 primFailCode ~= 0 ifTrue:
			[self reapAndResetErrorCodeTo: stackPointer header: methodHeader]].

	^methodHeader! !

!CoInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 10/30/2012 17:34'!
longPushTemporaryVariableBytecode
	"230		11100110	i i i i i i i i	Push Temporary Variable #iiiiiiii"
	| index |
	index := self fetchByte.
	self fetchNextBytecode.
	self internalPush: (self itemporary: index in: localFP)! !

!CoInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 10/30/2012 17:35'!
longStoreTemporaryVariableBytecode
	"234		11101010	i i i i i i i i	Store Temporary Variable #iiiiiiii"
	| index |
	index := self fetchByte.
	self fetchNextBytecode.
	self itemporary: index in: localFP put: self internalStackTop! !

!CoInterpreter methodsFor: 'simulation' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
lookupAddress: address
	"If address appears to be that of a Symbol or a few well-known objects (such as classes) answer it, otherwise answer nil.
	 For code disassembly"
	<doNotGenerate>
	(objectMemory lookupAddress: address) ifNotNil:
		[:lookup| ^lookup].
	address / objectMemory wordSize = primTraceLog offset ifTrue: [^'primTraceLog'].
	^nil! !

!CoInterpreter methodsFor: 'message sending' stamp: 'rmacnak 8/25/2015 19:46'!
lookupDynamicSuperSendNoMNU: methodReceiver
	"Do the full lookup for a Newspeak super send.
	IN: messageSelector
	IN: argumentCount
	OUT: localAbsentReceiver
	OUT: newMethod
	OUT: primitiveIndex
	RESULT: 0 or SelectorDoesNotUnderstand"

	| methodMixin methodMixinApplication |
	localAbsentReceiver := methodReceiver.
	localAbsentReceiverOrZero := 0.
	methodMixin := self methodClassOf: method.
	methodMixinApplication := self
		findApplicationOfTargetMixin: methodMixin
		startingAtBehavior: (objectMemory fetchClassOf: localAbsentReceiver).
	lkupClass := self superclassOf: methodMixinApplication. "MNU lookup starts here."
	^self lookupProtectedNoMNU: messageSelector startingAt: lkupClass rule: LookupRuleDynamicSuper! !

!CoInterpreter methodsFor: 'message sending' stamp: 'rmacnak 8/25/2015 19:51'!
lookupImplicitReceiverSendNoMNU: methodReceiver
	"Do the full lookup for an implicit receiver send.
	IN: messageSelector
	IN: argumentCount
	OUT: localAbsentReceiver
	OUT: localAbsentReceiverOrZero
	OUT: newMethod
	OUT: primitiveIndex
	RESULT: 0 or SelectorDoesNotUnderstand"

	| candidateReceiver candidateMixin candidateMixinApplication dictionary found |
	messageSelector := objectMemory followMaybeForwarded: messageSelector.
	candidateReceiver := methodReceiver.
	self deny: (objectMemory isForwarded: method).
	candidateMixin := self methodClassOf: method.
	localAbsentReceiverOrZero := 0.
	[self deny: (objectMemory isForwarded: candidateMixin).
	self deny: (objectMemory isForwarded: candidateReceiver).
	candidateMixinApplication := self
		findApplicationOfTargetMixin: candidateMixin
		startingAtBehavior: (objectMemory fetchClassOf: candidateReceiver).
	self deny: (candidateMixinApplication = 0).
	self deny: (candidateMixinApplication = objectMemory nilObject).
	self deny: (objectMemory isForwarded: candidateMixinApplication).
	self assert: (self addressCouldBeClassObj: candidateMixinApplication).
	dictionary := objectMemory followObjField: MethodDictionaryIndex ofObject: candidateMixinApplication.
	found := self lookupMethodInDictionary: dictionary.
	found ifTrue:
		[localAbsentReceiver := candidateReceiver.
		^self lookupLexicalNoMNU: messageSelector from: candidateMixin rule: LookupRuleImplicit].
	candidateMixin := objectMemory followObjField: EnclosingMixinIndex ofObject: candidateMixin.
	self deny: (objectMemory isForwarded: candidateMixin).
	candidateMixin = objectMemory nilObject]
		whileFalse:
			[localAbsentReceiverOrZero := candidateReceiver := objectMemory followObjField: EnclosingObjectIndex ofObject: candidateMixinApplication].
	"There is no lexically visible method, so the implicit receiver is the method receiver."
	localAbsentReceiverOrZero := 0.
	localAbsentReceiver := methodReceiver.
	lkupClass := objectMemory fetchClassOf: methodReceiver. "MNU lookup starts here."
	^self lookupProtectedNoMNU: messageSelector startingAt: lkupClass rule: LookupRuleImplicit.! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem 6/5/2017 11:09'!
lookupLexicalNoMNU: selector from: mixin rule: rule
	"A shared part of the lookup for implicit receiver sends that found a lexically visible
	method, and self and outer sends."
	| receiverClass mixinApplication dictionary found |
	receiverClass := objectMemory fetchClassOf: localAbsentReceiver.
	lkupClass := receiverClass. "MNU lookup starts here."
	self lookupBreakFor: lkupClass.
	mixinApplication := self findApplicationOfTargetMixin: mixin startingAtBehavior: receiverClass.
	dictionary := objectMemory followObjField: MethodDictionaryIndex ofObject: mixinApplication.
	found := self lookupMethodInDictionary: dictionary.
	(found and: [(self isPrivateMethod: newMethod)]) ifTrue:
		[^0].
	^self lookupProtectedNoMNU: selector startingAt: receiverClass rule: rule! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 2/12/2016 13:59'!
lookupMNU: selector receiver: rcvr
	<api>
	"Lookup selector in rcvr, without doing MNU processing, and answer either a
	 method or an error code if the message was not understood.  Used to populate closed PICs."
	| classTag inCache erridx |
	"self printFrame: stackPage headFP WithSP: stackPage headSP"
	"self printStringOf: selector"
	classTag := objectMemory fetchClassTagOf: rcvr.
	NewspeakVM
	  	ifTrue:
			[inCache := self
							inlineLookupInNSMethodCacheSel: selector
							classTag: classTag
							method: 0
							lookupRule: LookupRuleMNU]
		ifFalse:
			[inCache := self lookupInMethodCacheSel: selector classTag: classTag].
	inCache ifFalse:
		[messageSelector := selector.
		 erridx := self lookupMNUInClass: (objectMemory classForClassTag: classTag).
		 erridx ~= 0 ifTrue:
			[self assert: erridx <= self maxLookupNoMNUErrorCode.
			 ^erridx]].
	^newMethod! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'rmacnak 6/24/2015 21:16'!
lookupOrdinary: selector receiver: rcvr
	<api>
	"Lookup selector in rcvr, without doing MNU processing, and answer either a
	 method or an error code if the message was not understood.  Used to populate closed PICs."
	| classTag erridx |
	"self printFrame: stackPage headFP WithSP: stackPage headSP"
	"self printStringOf: selector"
	classTag := objectMemory fetchClassTagOf: rcvr.
	(self lookupInMethodCacheSel: selector classTag: classTag) ifFalse:
		[messageSelector := selector.
		 (erridx := self lookupOrdinaryNoMNUEtcInClass: (objectMemory classForClassTag: classTag)) ~= 0 ifTrue:
			[self assert: erridx <= self maxLookupNoMNUErrorCode.
			 ^erridx]].
	^newMethod! !

!CoInterpreter methodsFor: 'message sending' stamp: 'rmacnak 6/22/2015 22:00'!
lookupOuterSendNoMNU: methodReceiver depth: depth
	"Do the full lookup for a self or outer send.
	IN: messageSelector
	IN: argumentCount
	OUT: localAbsentReceiver
	OUT: localAbsentReceiverOrZero
	OUT: newMethod
	OUT: primitiveIndex
	RESULT: 0 or SelectorDoesNotUnderstand"

	| targetMixin count mixinApplication |
	localAbsentReceiver := methodReceiver.
	localAbsentReceiverOrZero := 0.
	targetMixin := self methodClassOf: method.
	count := 0.
	[count < depth] whileTrue:
		[count := count + 1.
		mixinApplication := self
			findApplicationOfTargetMixin: targetMixin
			startingAtBehavior: (objectMemory fetchClassOf: localAbsentReceiver).
		localAbsentReceiverOrZero := localAbsentReceiver := objectMemory followObjField: EnclosingObjectIndex ofObject: mixinApplication.
		targetMixin := objectMemory followObjField: EnclosingMixinIndex ofObject: targetMixin].
	^self lookupLexicalNoMNU: messageSelector from: targetMixin rule: depth! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem 6/5/2017 11:09'!
lookupProtectedNoMNU: selector startingAt: mixinApplication rule: rule
	"A shared part of the lookup for self, outer or implicit receiver sends that did not find a
	private lexically visible method, and (Newspeak) super sends."
	| currentClass dictionary found |
	self lookupBreakFor: mixinApplication.
	currentClass := mixinApplication.
	[currentClass = objectMemory nilObject] whileFalse:
		[dictionary := objectMemory followObjField: MethodDictionaryIndex ofObject: currentClass.
		found := self lookupMethodInDictionary: dictionary.
		(found and: [(self isPrivateMethod: newMethod) not]) ifTrue:
			[^0].
		currentClass := self superclassOf: currentClass].
	^SelectorDoesNotUnderstand! !

!CoInterpreter methodsFor: 'inline primitive ffi abi' stamp: 'rsf 2/12/2017 01:37'!
lowcodeDoCallout: functionPointer
	<option: #LowcodeVM>
	<var: #functionPointer type: #'char*' >
	self undoFetchNextBytecode.
	self externalizeIPandSP.
	self lowcodeCalloutState: lowcodeCalloutState callFunction: functionPointer.
	instructionPointer = cogit ceReturnToInterpreterPC ifTrue: [
		instructionPointer := self iframeSavedIP: framePointer.
	].
	self internalizeIPandSP.
	self reloadLowcodeStateAfterCallout.
	self fetchNextBytecode.
	^ self lowcodeCalloutStateFetchResultFloat32: lowcodeCalloutState! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 6/27/2011 17:05'!
mMethodClass
	<api>
	^self methodClassOf: (self mframeHomeMethod: framePointer) methodObject! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 11/30/2018 15:56'!
makeBaseFrameFor: aContext "<Integer>"
	"Marry aContext with the base frame of a new stack page.  Build the base
	 frame to reflect the context's state.  Answer the new page.  Override to
	 hold the caller context in a different place,  In the StackInterpreter we use
	 the caller saved ip, but in the Cog VM caller saved ip is the ceBaseReturn:
	 trampoline.  Simply hold the caller context in the first word of the stack."
	<returnTypeC: #'StackPage *'>
	| page pointer theMethod theIP numArgs stackPtrIndex maybeClosure rcvr |
	<inline: false>
	<var: #page type: #'StackPage *'>
	<var: #pointer type: #'char *'>
	<var: #cogMethod type: #'CogMethod *'>
	"theIP must be typed as signed because it is assigned ceCannotResumePC and so maybe implicitly typed as unsigned."
	<var: #theIP type: #sqInt>
	self assert: (objectMemory isContext: aContext).
	self assert: (self isSingleContext: aContext).
	self assert: (objectMemory goodContextSize: aContext).
	theIP := objectMemory fetchPointer: InstructionPointerIndex ofObject: aContext.
	self assert: HasBeenReturnedFromMCPC < 0.
	theIP := (objectMemory isIntegerObject: theIP)
				ifTrue: [objectMemory integerValueOf: theIP]
				ifFalse: [HasBeenReturnedFromMCPC].
	theMethod := objectMemory followObjField: MethodIndex ofObject: aContext.
	page := stackPages newStackPage.
	"first word on stack is caller context of base frame"
	stackPages
		longAt: (pointer := page baseAddress)
		put: (objectMemory followObjField: SenderIndex ofObject: aContext).
	"second word is the context itself; needed for cannotReturn processing; see ceBaseReturn:."
	stackPages
		longAt: (pointer := pointer - objectMemory wordSize)
		put: aContext.
	rcvr := objectMemory followField: ReceiverIndex ofObject: aContext.
	"If the frame is a closure activation then the closure should be on the stack in
	 the pushed receiver position (closures receive the value[:value:] messages).
	 Otherwise it should be the receiver proper."
	maybeClosure := objectMemory fetchPointer: ClosureIndex ofObject: aContext.
	maybeClosure ~= objectMemory nilObject
		ifTrue:
			[(objectMemory isForwarded: maybeClosure) ifTrue:
				[maybeClosure := objectMemory fixFollowedField: ClosureIndex ofObject: aContext withInitialValue: maybeClosure].
			 numArgs := self argumentCountOfClosure: maybeClosure.
			 stackPages
				longAt: (pointer := pointer - objectMemory wordSize)
				put: maybeClosure]
		ifFalse:
			[| header |
			 header := objectMemory methodHeaderOf: theMethod.
			 numArgs := self argumentCountOfMethodHeader: header.
			 "If this is a synthetic context its IP could be pointing at the CallPrimitive opcode.  If so, skip it."
			 ((self methodHeaderHasPrimitive: header)
			  and: [theIP = (1 + (self startPCOfMethodHeader: header))]) ifTrue:
				[theIP := theIP + (self sizeOfCallPrimitiveBytecode: header)].
			 stackPages
				longAt: (pointer := pointer - objectMemory wordSize)
				put: rcvr].
	"Put the arguments on the stack"
	1 to: numArgs do:
		[:i|
		stackPages
			longAt: (pointer := pointer - objectMemory wordSize)
			put: (objectMemory fetchPointer: ReceiverIndex + i ofObject: aContext)].
	"saved caller ip is base return trampoline"
	stackPages
		longAt: (pointer := pointer - objectMemory wordSize)
		put: cogit ceBaseFrameReturnPC.
	"base frame's saved fp is null"
	stackPages
		longAt: (pointer := pointer - objectMemory wordSize)
		put: 0.
	"N.B.  Don't set the baseFP, which marks the page as in use, until after
	 ensureMethodIsCogged: and/or instructionPointer:forContext:frame:. These
	 can cause a compiled code compaction which, if marked as in use, will
	 examine this partially initialized page and crash."
	page headFP: pointer.
	"Create either a machine code frame or an interpreter frame based on the pc.  If the pc is -ve
	 it is a machine code pc and so we produce a machine code frame.  If +ve an interpreter frame.
	 N.B. Do *not* change this to try and map from a bytecode pc to a machine code frame under
	 any circumstances.  See ensureContextIsExecutionSafeAfterAssignToStackPointer:"
	theIP < 0
		ifTrue:
			[| cogMethod |
			 "Since we would have to generate a machine-code method to be able to map
			  the native pc anyway we should create a native method and native frame."
			 cogMethod := self ensureMethodIsCogged: theMethod maybeClosure: maybeClosure.
			 theMethod := cogMethod asInteger.
			 maybeClosure ~= objectMemory nilObject
				ifTrue:
					[(self isVanillaBlockClosure: maybeClosure)
						ifTrue:
							["If the pc is the special HasBeenReturnedFromMCPC pc set the pc
							  appropriately so that the frame stays in the cannotReturn: state."
							 theIP = HasBeenReturnedFromMCPC
								ifTrue:
									[theMethod := (cogit findMethodForStartBcpc: (self startPCOfClosure: maybeClosure)
														inHomeMethod: (self cCoerceSimple: theMethod
																			to: #'CogMethod *')) asInteger.
									 theMethod = 0 ifTrue:
										[self error: 'cannot find machine code block matching closure''s startpc'].
									 theIP := cogit ceCannotResumePC]
								ifFalse:
									[self assert: (theIP signedBitShift: -16) < -1. "See contextInstructionPointer:frame:"
									 theMethod := theMethod - ((theIP signedBitShift: -16) * cogit blockAlignment).
									 theIP := theMethod - theIP signedIntFromShort]]
						ifFalse:
							[self assert: (theIP signedBitShift: -16) >= -1.
							 "If the pc is the special HasBeenReturnedFromMCPC pc set the pc
							  appropriately so that the frame stays in the cannotReturn: state."
							 theIP := theIP = HasBeenReturnedFromMCPC
										ifTrue: [cogit ceCannotResumePC]
										ifFalse: [theMethod asInteger - theIP]].
					 stackPages
						longAt: (pointer := pointer - objectMemory wordSize)
						put: theMethod + MFMethodFlagHasContextFlag + MFMethodFlagIsBlockFlag]
				ifFalse:
					[self assert: (theIP signedBitShift: -16) >= -1.
					 "If the pc is the special HasBeenReturnedFromMCPC pc set the pc
					  appropriately so that the frame stays in the cannotReturn: state."
					 theIP := theIP = HasBeenReturnedFromMCPC
								ifTrue: [cogit ceCannotResumePC]
								ifFalse: [theMethod asInteger - theIP].
					 stackPages
						longAt: (pointer := pointer - objectMemory wordSize)
						put: theMethod + MFMethodFlagHasContextFlag].
			 stackPages
				longAt: (pointer := pointer - objectMemory wordSize)
				put: aContext]
		ifFalse:
			[stackPages
				longAt: (pointer := pointer - objectMemory wordSize)
				put: theMethod.
			stackPages
				longAt: (pointer := pointer - objectMemory wordSize)
				put: aContext.
			stackPages
				longAt: (pointer := pointer - objectMemory wordSize)
				put: (self encodeFrameFieldHasContext: true isBlock: maybeClosure ~= objectMemory nilObject numArgs: numArgs).
			stackPages
				longAt: (pointer := pointer - objectMemory wordSize)
				put: 0. "FoxIFSavedIP"
			theIP := self iframeInstructionPointerForIndex: theIP method: theMethod].
	page baseFP: page headFP.
	self assert: (self frameHasContext: page baseFP).
	self assert: (self frameNumArgs: page baseFP) == numArgs.
	stackPages
		longAt: (pointer := pointer - objectMemory wordSize)
		put: rcvr.
	stackPtrIndex := self quickFetchInteger: StackPointerIndex ofObject: aContext.
	self assert: ReceiverIndex + stackPtrIndex < (objectMemory lengthOf: aContext).
	numArgs + 1 to: stackPtrIndex do:
		[:i|
		stackPages
			longAt: (pointer := pointer - objectMemory wordSize)
			put: (objectMemory fetchPointer: ReceiverIndex + i ofObject: aContext).
		"nil the slot in the context so that it doesn't inadvertently hang onto some collectable object.
		 Thanks to Ryan Macnak for identifying this bug"
		objectMemory storePointerUnchecked: ReceiverIndex + i ofObject: aContext withValue: objectMemory nilObject].
	"top of stack is the instruction pointer"
	stackPages longAt: (pointer := pointer - objectMemory wordSize) put: theIP.
	page headSP: pointer.
	self assert: (self context: aContext hasValidInversePCMappingOf: theIP in: page baseFP).

	"Mark context as married by setting its sender to the frame pointer plus SmallInteger
	 tags and the InstructionPointer to the saved fp (which ensures correct alignment
	 w.r.t. the frame when we check for validity) plus SmallInteger tags."
	objectMemory storePointerUnchecked: SenderIndex
		ofObject: aContext
		withValue: (self withSmallIntegerTags: page baseFP).
	objectMemory storePointerUnchecked: InstructionPointerIndex
		ofObject: aContext
		withValue: (self withSmallIntegerTags: 0).
	self assert: (objectMemory isIntegerObject: (objectMemory fetchPointer: SenderIndex ofObject: aContext)).
	self assert: (self frameOfMarriedContext: aContext) = page baseFP.
	self assert: (self validStackPageBaseFrame: page).
	^page! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 5/7/2014 08:35'!
mapMachineCode: theGCMode
	<inline: true>
	"Update all references to objects in machine code."
	cogit mapObjectReferencesInMachineCode: theGCMode! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 4/8/2014 15:17'!
mapPrimTraceLog
	"The prim trace log is a circular buffer of selectors. If there is
	 an entry at primTraceLogIndex - 1 \\ PrimTraceBufferSize it has entries.
	 If there is something at primTraceLogIndex it has wrapped."
	<inline: false>
	(primTraceLog at: (self safe: primTraceLogIndex - 1 mod: PrimTraceLogSize)) = 0 ifTrue:
		[^self].
	(primTraceLog at: primTraceLogIndex) ~= 0 ifTrue:
		[primTraceLogIndex to: PrimTraceLogSize - 1 do:
			[:i| | selector |
			 selector := primTraceLog at: i.
			 (selector ~= 0
			  and: [objectMemory shouldRemapOop: selector]) ifTrue:
				[primTraceLog at: i put: (objectMemory remapObj: selector)]]].
	0 to: primTraceLogIndex - 1 do:
		[:i| | selector |
		 selector := primTraceLog at: i.
		 (selector ~= 0
		  and: [objectMemory shouldRemapOop: selector]) ifTrue:
			[primTraceLog at: i put: (objectMemory remapObj: selector)]]! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 2/23/2020 23:48'!
mapStackPages
	<inline: #never>
	<var: #thePage type: #'StackPage *'>
	<var: #theSP type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #frameRcvrOffset type: #'char *'>
	<var: #callerFP type: #'char *'>
	<var: #theIPPtr type: #'char *'>
	| numLivePages |
	numLivePages := 0.
	0 to: numStackPages - 1 do:
		[:i| | thePage theSP theFP frameRcvrOffset callerFP theIPPtr theIP oop |
		thePage := stackPages stackPageAt: i.
		thePage isFree ifFalse:
			[self assert: (self ifCurrentStackPageHasValidHeadPointers: thePage).
			 numLivePages := numLivePages + 1.
			 theSP := thePage headSP.
			 theFP := thePage  headFP.
			 "Skip the instruction pointer on top of stack of inactive pages."
			 thePage = stackPage
				ifTrue: [theIPPtr := ((self isMachineCodeFrame: theFP)
									or: [(self iframeSavedIP: theFP) = 0])
										ifTrue: [0]
										ifFalse: [theFP + FoxIFSavedIP]]
				ifFalse:
					[theIPPtr := theSP.
					 theSP := theSP + objectMemory wordSize].
			[self assert: (thePage addressIsInPage: theFP).
			 self assert: (thePage addressIsInPage: theSP).
			 self assert: (theIPPtr = 0 or: [thePage addressIsInPage: theIPPtr]).
			 frameRcvrOffset := self frameReceiverLocation: theFP.
	 		  [theSP <= frameRcvrOffset] whileTrue:
				[oop := stackPages longAt: theSP.
				 (objectMemory shouldRemapOop: oop) ifTrue:
					[stackPages longAt: theSP put: (objectMemory remapObj: oop)].
				 theSP := theSP + objectMemory wordSize].
			 (self frameHasContext: theFP) ifTrue:
				[(objectMemory shouldRemapObj: (self frameContext: theFP)) ifTrue:
					[stackPages
						longAt: theFP + FoxThisContext
						put: (objectMemory remapObj: (self frameContext: theFP))].
				 "With SqueakV3 objectMemory or SpurPlanningCompactor can't assert since object body is yet to move."
				 (objectMemory hasSpurMemoryManagerAPI
				  and: [(objectMemory slidingCompactionInProgress or: [objectMemory scavengeInProgress]) not]) ifTrue:
					[self assert: ((self isMarriedOrWidowedContext: (self frameContext: theFP))
								and: [(self frameOfMarriedContext: (self frameContext: theFP)) = theFP])]].
			(self isMachineCodeFrame: theFP) ifFalse:
				[(objectMemory shouldRemapObj: (self iframeMethod: theFP)) ifTrue:
					[theIPPtr ~= 0 ifTrue:
						[theIP := stackPages longAt: theIPPtr.
						 theIP = cogit ceReturnToInterpreterPC
							ifTrue:
								[self assert: (self iframeSavedIP: theFP) > (self iframeMethod: theFP).
								 theIPPtr := theFP + FoxIFSavedIP.
								 theIP := stackPages longAt: theIPPtr]
							ifFalse:
								[self assert: theIP > (self iframeMethod: theFP)].
						 theIP := theIP - (self iframeMethod: theFP)].
					 stackPages
						longAt: theFP + FoxMethod
						put: (objectMemory remapObj: (self iframeMethod: theFP)).
					 theIPPtr ~= 0 ifTrue:
						[stackPages longAt: theIPPtr put: theIP + (self iframeMethod: theFP)]]].
			 (callerFP := self frameCallerFP: theFP) ~= 0] whileTrue:
				[theSP := (theIPPtr := theFP + FoxCallerSavedIP) + objectMemory wordSize.
				 theFP := callerFP].
			 theSP := theFP + FoxCallerSavedIP + objectMemory wordSize.
			 [theSP <= thePage baseAddress] whileTrue:
				[oop := stackPages longAt: theSP.
				 (objectMemory shouldRemapOop: oop) ifTrue:
					[stackPages longAt: theSP put: (objectMemory remapObj: oop)].
				 theSP := theSP + objectMemory wordSize]]].
	stackPages recordLivePagesOnMapping: numLivePages! !

!CoInterpreter methodsFor: 'plugin primitive support' stamp: 'eem 8/2/2021 20:35'!
mapToBytecodePCIfActivationOfExternalMethod: ctxtObj
	(self isExternalMethodInPlugin: (objectMemory fetchPointer: MethodIndex ofObject: ctxtObj)) ifTrue:
		[(self isMarriedOrWidowedContext: ctxtObj) ifTrue:
			[| theFP |
			 (self isWidowedContext: ctxtObj) ifTrue:
				[^self].
			 theFP := self frameOfMarriedContext: ctxtObj.
			 (self isMachineCodeFrame: theFP) ifFalse:
				[^self]].
		 self ensureContextHasBytecodePC: ctxtObj]! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 1/10/2014 11:23'!
mapTraceLog
	"The trace log is a circular buffer of pairs of entries. If there is
	 an entry at traceLogIndex - 3 \\ TraceBufferSize it has entries.
	 If there is something at traceLogIndex it has wrapped."
	<inline: false>
	| limit |
	limit := self safe: traceLogIndex - 3 mod: TraceBufferSize.
	(traceLog at: limit) = 0 ifTrue: [^self].
	(traceLog at: traceLogIndex) ~= 0 ifTrue:
		[limit := TraceBufferSize - 3].
	0 to: limit by: 3 do:
		[:i| | intOrClass selectorOrMethod |
		intOrClass := traceLog at: i.
		(objectMemory shouldRemapOop: intOrClass) ifTrue:
			[traceLog at: i put: (objectMemory remapObj: intOrClass)].
		selectorOrMethod := traceLog at: i + 1.
		(objectMemory shouldRemapOop: selectorOrMethod) ifTrue:
			[traceLog at: i + 1 put: (objectMemory remapObj: selectorOrMethod)]]! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 4/28/2010 12:31'!
mapTraceLogs
	self mapTraceLog.
	self mapPrimTraceLog! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 10/1/2013 17:27'!
mapVMRegisters
	"Map the oops in the interpreter's vm ``registers'' to their new values 
	 during garbage collection or a become: operation."
	"Assume: All traced variables contain valid oops.
	 N.B. Don't trace messageSelector and lkupClass; these are ephemeral, live
	 only during message lookup and because createActualMessageTo will not
	 cause a GC these cannot change during message lookup."
	| mapInstructionPointer |
	(objectMemory shouldRemapObj: method) ifTrue:
		["i.e. interpreter instructionPointer in method as opposed to machine code?"
		(mapInstructionPointer := instructionPointer > method) ifTrue:
			[instructionPointer := instructionPointer - method]. "*rel to method"
		method := objectMemory remapObj: method.
		mapInstructionPointer ifTrue:
			[instructionPointer := instructionPointer + method]]. "*rel to method"
	(objectMemory shouldRemapOop: newMethod) ifTrue: "maybe oop due to object-as-method"
		[newMethod := objectMemory remapObj: newMethod]! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 11/25/2020 23:41'!
markActiveMethodsAndReferents
	<api>
	"If instructionPointer is referring to machine code, as it will be if a primitive is in progress
	 (see isCodeCompactingPrimitiveIndex:) it may refer to a method, and if so that method
	 must be retained."
	 instructionPointer ~= 0 ifTrue:
		[(cogit cogMethodContaining: instructionPointer) ifNotNil:
			[:primCogMethod|
			 cogit markMethodAndReferents: (self cCoerceSimple: primCogMethod to: #'CogBlockMethod *')]].

	0 to: numStackPages - 1 do:
		[:i| | thePage |
		thePage := stackPages stackPageAt: i.
		(stackPages isFree: thePage) ifFalse:
			[self markCogMethodsAndReferentsOnPage: thePage]]! !

!CoInterpreter methodsFor: 'gc -- mark and sweep' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
markAndTraceMachineCodeMethod: aCogMethod
	<var: #aCogMethod type: #'CogBlockMethod *'>
	| homeMethod |
	<var: #homeMethod type: #'CogMethod *'>
	homeMethod := self asCogHomeMethod: aCogMethod.
	objectMemory markAndTrace: homeMethod methodObject! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 3/13/2014 16:51'!
markAndTraceMachineCodeOfMarkedMethods
	"Deal with a fulGC's effects on machine code.  Mark and
	 trace oops in marked machine code methods.  The stack
	 pages have already been traced so any methods of live
	 stack activations have already been marked and traced."
	<doNotGenerate>
	cogit markAndTraceMachineCodeOfMarkedMethods! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 3/13/2014 16:49'!
markAndTraceOrFreeMachineCode: fullGCFlag
	"Deal with a fulGC's effects on machine code.  Either mark and trace
	 oops in machine code or free machine-code methods that refer to freed
	 oops.  The stack pages have already been traced so any methods
	 of live stack activations have already been marked and traced."
	cogit markAndTraceObjectsOrFreeMachineCode: fullGCFlag! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 4/8/2014 15:19'!
markAndTracePrimTraceLog
	"The prim trace log is a circular buffer of selectors. If there is
	 an entry at primTraceLogIndex - 1 \\ PrimTraceBufferSize it has entries.
	 If there is something at primTraceLogIndex it has wrapped."
	<inline: false>
	(primTraceLog at: (self safe: primTraceLogIndex - 1 mod: PrimTraceLogSize)) = 0 ifTrue:
		[^self].
	(primTraceLog at: primTraceLogIndex) ~= 0 ifTrue:
		[primTraceLogIndex to: PrimTraceLogSize - 1 do:
			[:i| | selector |
			 selector := primTraceLog at: i.
			 (selector ~= 0
			  and: [objectMemory isNonImmediate: selector]) ifTrue:
				[objectMemory markAndTrace: selector]]].
	0 to: primTraceLogIndex - 1 do:
		[:i| | selector |
		selector := primTraceLog at: i.
		(selector ~= 0
		  and: [objectMemory isNonImmediate: selector]) ifTrue:
			[objectMemory markAndTrace: selector]]! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 5/6/2015 17:06'!
markAndTraceStackPage: thePage
	| theSP theFP frameRcvrOffset callerFP oop |
	<var: #thePage type: #'StackPage *'>
	<var: #theSP type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #frameRcvrOffset type: #'char *'>
	<var: #callerFP type: #'char *'>
	<inline: false>

	self assert: (stackPages isFree: thePage) not.
	self assert: (self ifCurrentStackPageHasValidHeadPointers: thePage).
	self assert: thePage trace ~= StackPageTraced.
	thePage trace: StackPageTraced.

	theSP := thePage headSP.
	theFP := thePage  headFP.
	"Skip the instruction pointer on top of stack of inactive pages."
	thePage = stackPage ifFalse:
		[theSP := theSP + objectMemory wordSize].
	[frameRcvrOffset := self frameReceiverLocation: theFP.
	 [theSP <= frameRcvrOffset] whileTrue:
		[oop := stackPages longAt: theSP.
		 (objectMemory isOopForwarded: oop) ifTrue:
			[oop := objectMemory followForwarded: oop.
			 stackPages longAt: theSP put: oop].
		 (objectMemory isImmediate: oop) ifFalse:
			[objectMemory markAndTrace: oop].
		 theSP := theSP + objectMemory wordSize].
	(self frameHasContext: theFP) ifTrue:
		[self assert: (objectMemory isContext: (self frameContext: theFP)).
		 objectMemory markAndTrace: (self frameContext: theFP)].
	(self isMachineCodeFrame: theFP)
		ifTrue: [self markAndTraceMachineCodeMethod: (self mframeCogMethod: theFP)]
		ifFalse: [objectMemory markAndTrace: (self iframeMethod: theFP)].
	(callerFP := self frameCallerFP: theFP) ~= 0] whileTrue:
		[theSP := theFP + FoxCallerSavedIP + objectMemory wordSize.
		 theFP := callerFP].
	theSP := theFP + FoxCallerSavedIP + objectMemory wordSize. "caller ip is ceBaseReturnPC"
	[theSP <= thePage baseAddress] whileTrue:
		[oop := stackPages longAt: theSP.
		 (objectMemory isOopForwarded: oop) ifTrue:
			[oop := objectMemory followForwarded: oop.
			 stackPages longAt: theSP put: oop].
		 (objectMemory isImmediate: oop) ifFalse:
			[objectMemory markAndTrace: oop].
		 theSP := theSP + objectMemory wordSize]! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 1/10/2014 11:23'!
markAndTraceTraceLog
	"The trace log is a circular buffer of pairs of entries. If there is an entry at
	 traceLogIndex - 3 \\ TraceBufferSize it has entries.  If there is something at
	 traceLogIndex it has wrapped."
	<inline: false>
	| limit |
	limit := self safe: traceLogIndex - 3 mod: TraceBufferSize.
	(traceLog at: limit) = 0 ifTrue: [^self].
	(traceLog at: traceLogIndex) ~= 0 ifTrue:
		[limit := TraceBufferSize - 3].
	0 to: limit by: 3 do:
		[:i| | oop |
		oop := traceLog at: i.
		(objectMemory isImmediate: oop) ifFalse:
			[objectMemory markAndTrace: oop].
		oop := traceLog at: i + 1.
		(objectMemory isImmediate: oop) ifFalse:
			[objectMemory markAndTrace: oop]]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 11/20/2020 08:48'!
markCogMethodsAndReferentsOnPage: thePage
	<var: #thePage type: #'StackPage *'>
	| theFP callerFP |
	<inline: false>
	self assert: (stackPages isFree: thePage) not.
	self assert: (self ifCurrentStackPageHasValidHeadPointers: thePage).
	theFP := thePage headFP.
	[(self isMachineCodeFrame: theFP) ifTrue:
		[cogit markMethodAndReferents: (self mframeCogMethod: theFP)].
	(callerFP := self frameCallerFP: theFP) ~= 0] whileTrue:
		[theFP := callerFP]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 9/7/2016 13:29'!
marryFrame: theFP SP: theSP copyTemps: copyTemps
	"Marry an unmarried frame.  This means creating a spouse context
	 initialized with a subset of the frame's state that references the frame.
	 For the default closure implementation we do not need to copy temps.
	 Different closure implementations may require temps to be copied.

	 This method is important enough for performance to be worth streamlining.

	Override to set the ``has context'' flag appropriately for both machine code and interpreter frames
	and to streamline the machine code/interpreter differences.."
	| theContext methodFieldOrObj closureOrNil rcvr numSlots numArgs numStack numTemps |
	<inline: true>
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	<var: #cogMethod type: #'CogMethod *'>
	self assert: (self frameHasContext: theFP) not.
	self assert: (self isBaseFrame: theFP) not. "base frames must aready be married for cannotReturn: processing"

	"The SP is expected to be pointing at the last oop on the stack, not at the pc"
	self assert: (objectMemory addressCouldBeOop: (objectMemory longAt: theSP)).

	"Decide how much of the stack to preserve in widowed contexts.  Preserving too much
	 state will potentially hold onto garbage.  Holding onto too little may mean that a dead
	 context isn't informative enough in a debugging situation.  If copyTemps is false (as it
	 is in the default closure implementation) compromise, retaining only the arguments with
	 no temporaries.  Note that we still set the stack pointer to its current value, but stack
	 contents other than the arguments are nil."
	methodFieldOrObj := self frameMethodField: theFP.
	methodFieldOrObj asUnsignedInteger < objectMemory startOfMemory "inline (self isMachineCodeFrame: theFP)"
		ifTrue:
			[| cogMethod |
			 stackPages
				longAt: theFP + FoxMethod
				put: methodFieldOrObj + MFMethodFlagHasContextFlag.
			 cogMethod := self cCoerceSimple: (methodFieldOrObj bitAnd: MFMethodMask) to: #'CogMethod *'.
			 numArgs := cogMethod cmNumArgs.
			 cogMethod cmType = CMMethod
				ifTrue:
					[closureOrNil := cogMethod cmIsFullBlock
										ifTrue: [self frameStackedReceiver: theFP numArgs: numArgs]
										ifFalse: [objectMemory nilObject]]
				ifFalse:
					[cogMethod := (self cCoerceSimple: cogMethod to: #'CogBlockMethod *') cmHomeMethod.
					 closureOrNil := self frameStackedReceiver: theFP numArgs: numArgs].
			 numSlots := (self methodHeaderIndicatesLargeFrame: cogMethod methodHeader)
							ifTrue: [LargeContextSlots]
							ifFalse: [SmallContextSlots].
			 methodFieldOrObj := cogMethod methodObject.
			 rcvr := self mframeReceiver: theFP.
			 numStack := self stackPointerIndexForMFrame: theFP WithSP: theSP numArgs: numArgs]
		ifFalse:
			[self setIFrameHasContext: theFP.
			 numArgs := self iframeNumArgs: theFP.
			 numSlots := (self methodHeaderIndicatesLargeFrame: (objectMemory methodHeaderOf: methodFieldOrObj))
							ifTrue: [LargeContextSlots]
							ifFalse: [SmallContextSlots].
			 closureOrNil := (self iframeIsBlockActivation: theFP)
								ifTrue: [self frameStackedReceiver: theFP numArgs: numArgs]
								ifFalse: [objectMemory nilObject].
			 rcvr := self iframeReceiver: theFP.
			 numStack := self stackPointerIndexForIFrame: theFP WithSP: theSP numArgs: numArgs].
	theContext := objectMemory eeInstantiateMethodContextSlots: numSlots.
	self setFrameContext: theFP to: theContext.
	"Mark context as married by setting its sender to the frame pointer plus SmallInteger
	 tags and the InstructionPointer to the saved fp (which ensures correct alignment
	 w.r.t. the frame when we check for validity)"
	objectMemory storePointerUnchecked: SenderIndex
		ofObject: theContext
		withValue: (self withSmallIntegerTags: theFP).
	objectMemory storePointerUnchecked: InstructionPointerIndex
		ofObject: theContext
		withValue: (self withSmallIntegerTags: (self frameCallerFP: theFP)).
	objectMemory storePointerUnchecked: StackPointerIndex
		ofObject: theContext
		withValue: (objectMemory integerObjectOf: numStack).
	objectMemory storePointerUnchecked: MethodIndex
		ofObject: theContext
		withValue: methodFieldOrObj.
	objectMemory storePointerUnchecked: ClosureIndex ofObject: theContext withValue: closureOrNil.
	objectMemory storePointerUnchecked: ReceiverIndex
		ofObject: theContext
		withValue: rcvr.
	1 to: numArgs do:
		[:i|
		objectMemory storePointerUnchecked: ReceiverIndex + i
			ofObject: theContext
			withValue: (self temporary: i - 1 in: theFP)].
	copyTemps ifTrue:
		[numTemps := self frameNumTemps: theFP.
		 1 to: numTemps do:
			[:i|
			objectMemory storePointerUnchecked: ReceiverIndex + i + numArgs
				ofObject: theContext
				withValue: (self temporary: i - 1 in: theFP)].
		 numArgs := numArgs + numTemps].

	numArgs + 1 to: numStack do:
		[:i|
		objectMemory storePointerUnchecked: ReceiverIndex + i
			ofObject: theContext
			withValue: objectMemory nilObject].

	self assert: (self frameHasContext: theFP).
	self assert: (self frameOfMarriedContext: theContext) = theFP.
	self assert: numStack + ReceiverIndex < (objectMemory lengthOf: theContext).

	^theContext! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 4/15/2015 18:19'!
marryFrameCopiesTemps
	"Answer whether marryFrame:SP: copies non-argument temporaries."
	<api>
	^false! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 12/9/2013 13:09'!
maxLookupNoMNUErrorCode
	<api>
	^SelectorCannotInterpret max: SelectorDoesNotUnderstand! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 7/14/2014 18:29'!
maybeFixClonedCompiledMethod: objOop
	"Make sure a cloned method doesn't reference its originals Cog method, if any."
	| rawHeader |
	self assert: (objectMemory isOopCompiledMethod: objOop).
	rawHeader := self rawHeaderOf: objOop.
	(self isCogMethodReference: rawHeader) ifTrue:
		[self
			rawHeaderOf: objOop
			put: (self cCoerceSimple: rawHeader to: #'CogMethod *') methodHeader]! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'cb 8/29/2016 10:30'!
maybeFlagMethodAsInterpreted: aMethod
	"The flag bit can be used to flag methods that are interpreted, if it has been requested
	 from the image header flags."
	flagInterpretedMethods ifTrue:
		[| rawHeader realHeader |
		 rawHeader := self rawHeaderOf: aMethod.
		 realHeader := (self isCogMethodReference: rawHeader)
						ifTrue: [(self cCoerceSimple: rawHeader to: #'CogMethod *') methodHeader]
						ifFalse: [rawHeader].
		 realHeader := realHeader bitOr: (objectMemory integerObjectOf: 1 << MethodHeaderFlagBitPosition).
		 (self isCogMethodReference: rawHeader)
			ifTrue: [(self cCoerceSimple: rawHeader to: #'CogMethod *') methodHeader: realHeader]
			ifFalse: [objectMemory storePointerUnchecked: 0 ofObject: aMethod withValue: realHeader]]! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem 10/1/2013 14:36'!
maybeMethodHasCogMethod: anOop
	^(objectMemory isNonImmediate: anOop)
	  and: [(objectMemory isCompiledMethod: anOop)
	  and: [self isCogMethodReference: (self rawHeaderOf: anOop)]]! !

!CoInterpreter methodsFor: 'return bytecodes' stamp: 'eem 12/28/2014 15:02'!
maybeReturnToMachineCodeFrame
	"If the frame we're returning to is a machine code one, then return to it.
	 Otherwise, if it's an interpreter frame, load the saved ip."
	<inline: true>
	localIP asUnsignedInteger < objectMemory startOfMemory ifTrue:
		[localIP asUnsignedInteger ~= cogit ceReturnToInterpreterPC ifTrue:
			["localIP in the cog method zone indicates a return to machine code."
			 ^self returnToMachineCodeFrame].
		 localIP := self pointerForOop: (self iframeSavedIP: localFP)]! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 6/20/2021 19:36'!
maybeTraceStackOverflow
	cogit recordOverflowTrace ifTrue:
		[self recordTrace: TraceStackOverflow
			thing: TraceStackOverflow
			source: ((self isMachineCodeFrame: framePointer)
						ifTrue: [TraceIsFromMachineCode]
						ifFalse: [TraceIsFromInterpreter])].
	(cogit recordPrimTrace and: [primTracePluginName isNil]) ifTrue:
		[self fastLogPrim: TraceStackOverflow]! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 10/17/2015 11:16'!
methodCacheAddress
	<api>
	<returnTypeC: #'void *'>
	^self cCode: [methodCache] inSmalltalk: [methodCache address]! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem 9/30/2013 16:17'!
methodHasCogMethod: aMethodOop
	<api>
	self assert: (objectMemory isNonImmediate: aMethodOop).
	^self isCogMethodReference: (self rawHeaderOf: aMethodOop)! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem 8/9/2021 13:29'!
methodHasPrimitiveInPrimTracePlugin: aMethodObj
	<inline: #never>
	| lit pluginName ok |
	((self isExternalPrimitiveCall: aMethodObj)
	 and: [(objectMemory literalCountOf: aMethodObj) > 0]) ifFalse:
		[^false].
	lit := self literal: 0 ofMethod: aMethodObj.
	((objectMemory isArray: lit)
	 and: [(objectMemory numSlotsOf: lit) = 4]) ifFalse:
		[^false].
	self assert: primTracePluginName notNil.
	pluginName := objectMemory fetchPointer: 0 ofObject: lit.
	(objectMemory isBytes: pluginName) ifFalse:
		[^false].
	ok := self object: pluginName equalsString: primTracePluginName.
	^ok! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 08/08/2014 10:42'!
methodNeedsLargeContext: methodObj
	<api>
	^self methodHeaderIndicatesLargeFrame: (objectMemory methodHeaderOf: methodObj)! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem 08/08/2014 10:42'!
methodShouldBeCogged: aMethodObj
	<api>
	(self methodWithHeaderShouldBeCogged: (objectMemory methodHeaderOf: aMethodObj)) ifTrue:
		[^true].
	self maybeFlagMethodAsInterpreted: aMethodObj.
	^false! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem 1/20/2017 17:13'!
methodWithHeaderShouldBeCogged: methodHeader
	"At the moment jit any method with less than N literals, where N defaults to 60.
	 See e.g. SimpleStackBasedCogit class>>initialize.
	 In my dev image eem 2/22/2009 13:39
		(30 to: 100 by: 5) collect:
			[:n| n -> (SystemNavigation default allSelect: [:m| m numLiterals > n]) size]
		#(30->1681 35->1150 40->765 45->523 50->389 55->289 60->206
		    65->151 70->124 75->99 80->73 85->63 90->54 95->42 100->38).
	 And running the CogVMSimulator with flagging of interpreted methods turned on reveals
	 the following sizes of interpreted methods.
		| sizes |
		sizes := Bag new.
		SystemNavigation default allSelect: [:m| m flag ifTrue: [sizes add: m numLiterals]. false].
		sizes sortedElements asArray
			#(	40->4 41->1 42->2 44->1 45->3 46->1 47->2 48->1
				50->2 51->1 53->1 55->1 56->1
				87->1 108->1 171->1)
	 literalCountOfHeader: does not include the header word."
	^SistaVM
		ifTrue: [(self isOptimizedMethodHeader: methodHeader)
				or: [(objectMemory literalCountOfMethodHeader: methodHeader) <= maxLiteralCountForCompile]]
		ifFalse: [(objectMemory literalCountOfMethodHeader: methodHeader) <= maxLiteralCountForCompile]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 10/30/2013 15:37'!
mframeCogMethod: theFP
	"Answer the Cog method for a machine code frame.  This may be
	 either a full CogMethod or merely a CogBlockMethod rump header."
	<var: #theFP type: #'char *'>
	^self cCoerceSimple: (self mframeMethod: theFP) to: #'CogBlockMethod *'! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 1/22/2009 15:05'!
mframeHasContext: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	^((self frameMethodField: theFP) bitAnd: MFMethodFlagHasContextFlag) ~= 0! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 7/26/2011 14:06'!
mframeHomeMethod: theFP
	"Answer the home method for a machine code frame.  From a block frame we find
	 the home method through the block's homeOffset field which is the delta to it.
	 In both cases we need to strip the isBlock and isContext flags from the method field."
	<api>
	<returnTypeC: #'CogMethod *'>
	<var: #theFP type: #'char *'>
	| methodField |
	methodField := self frameMethodField: theFP.
	(methodField bitAnd: MFMethodFlagIsBlockFlag) ~= 0 ifTrue:
		[^(self cCoerceSimple: (methodField bitAnd: MFMethodMask) to: #'CogBlockMethod *') cmHomeMethod].
	^self cCoerceSimple: (methodField bitAnd: MFMethodMask) to: #'CogMethod *'! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 12/12/2015 09:42'!
mframeHomeMethodExport
	<api>
	<returnTypeC: #'CogMethod *'>
	^self mframeHomeMethod: framePointer! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 1/22/2009 15:09'!
mframeIsBlockActivation: theFP "<Integer>"
	<inline: true>
	<var: #theFP type: #'char *'>
	^((self frameMethodField: theFP) bitAnd: MFMethodFlagIsBlockFlag) ~= 0! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 2/21/2009 16:13'!
mframeMethod: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	^(self frameMethodField: theFP) bitAnd: MFMethodMask! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 10/30/2013 16:33'!
mframeNumArgs: theFP
	<returnTypeC: #sqInt>
	^(self mframeCogMethod: theFP) cmNumArgs! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 1/29/2009 11:31'!
mframeReceiver: theFP
	<inline: true>
	<var: #theFP type: #'char *'>
	^stackPages longAt: theFP + FoxMFReceiver! !

!CoInterpreter methodsFor: 'stack pages' stamp: 'eem 7/29/2021 11:25'!
minimumUnusedHeadroom
	"Traverse all stack pages looking for non-zero bytes in the headroom part of each page.
	 Answer the minimum size of unused headroom (zero bytes) in the pages.  This is for
	 checking that there is enough headroom allocated in stack pages."
	| minUnused page |
	<inline: #never>
	<var: #p type: #'char *'>
	minUnused := (stackPages stackPageAt: 0) baseAddress - (stackPages stackPageAt: 0) lastAddress.
	0 to: numStackPages - 1 do:
		[:i| | p unused |
		page := stackPages stackPageAt: i.
		p := page lastAddress.
		[p := p + objectMemory wordSize.
		(self longAtPointer: p) = 0
		 and: [p <= page baseAddress]] whileTrue.
		unused := p - objectMemory wordSize - page lastAddress.
		unused < minUnused ifTrue:
			[minUnused := unused]].
	^minUnused! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 12/5/2014 14:48'!
mnuCompilationBreak: selectorOop point: selectorLength
	<api>
	<cmacro: '(sel, len) do { \
	if ((len) == -breakSelectorLength \
	 && !!strncmp((char *)((sel) + BaseHeaderSize), breakSelector, -breakSelectorLength)) { \
		suppressHeartbeatFlag = 1; \
		compilationBreakpointFor(sel); \
	} \
} while (0)'>
	| i |
	breakSelectorLength negated = selectorLength ifTrue:
		[i := breakSelectorLength negated.
		 [i > 0] whileTrue:
			[(objectMemory byteAt: selectorOop + i + objectMemory baseHeaderSize - 1) = (breakSelector at: i) asInteger
				ifTrue: [(i := i - 1) = 0 ifTrue:
							[self mnuCompilationBreakpointFor: selectorOop]]
				ifFalse: [i := 0]]]! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 12/5/2014 14:48'!
mnuCompilationBreakpointFor: selectorOop
	<api>
	suppressHeartbeatFlag := true.
	self
		cCode: [self warning: 'compilation MNU break (heartbeat suppressed)']
		inSmalltalk: [self halt: 'Compilation for MNU of ', breakSelector]! !

!CoInterpreter methodsFor: 'message sending' stamp: 'eem 5/4/2015 20:33'!
mnuMethodOrNilFor: rcvr
	"Lookup the doesNotUnderstand: selector in the class of the argument rcvr.
	 Answer either the matching method (cogged if appropriate), or nil, if not found."
	| currentClass mnuSelector dictionary mnuMethod methodHeader |
	self deny: (objectMemory isOopForwarded: rcvr).
	currentClass := objectMemory fetchClassOf: rcvr.
	mnuSelector := objectMemory splObj: SelectorDoesNotUnderstand.
	[currentClass ~= objectMemory nilObject] whileTrue:
		[dictionary := objectMemory fetchPointer: MethodDictionaryIndex ofObject: currentClass.
		 dictionary = objectMemory nilObject ifTrue:
			[^nil].
		 mnuMethod := self lookupMethodFor: mnuSelector InDictionary: dictionary.
		 mnuMethod ifNotNil:
			[methodHeader := self rawHeaderOf: mnuMethod.
			 ((self isCogMethodReference: methodHeader) not
			  and: [self methodWithHeaderShouldBeCogged: methodHeader]) ifTrue:
				[cogit cog: mnuMethod selector: mnuSelector].
			^mnuMethod].
		currentClass := self superclassOf: currentClass].
	^nil! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 11/13/2014 13:12'!
moveFramesIn: oldPage through: theFP toPage: newPage
	"Move frames from the hot end of oldPage through to theFP to newPage.
	 This has the effect of making theFP a base frame which can be stored into.
	 Answer theFP's new location."
	| newSP newFP stackedReceiverOffset delta callerFP callerIP fpInNewPage offsetCallerFP theContext |
	<inline: false>
	<var: #oldPage type: #'StackPage *'>
	<var: #theFP type: #'char *'>
	<var: #newPage type: #'StackPage *'>
	<var: #newSP type: #'char *'>
	<var: #newFP type: #'char *'>
	<var: #callerFP type: #'char *'>
	<var: #fpInNewPage type: #'char *'>
	<var: #offsetCallerFP type: #'char *'>
	<var: #source type: #'char *'>
	<returnTypeC: #'char *'>
	"A base frame must have a context for cannotReturn: processing."
	self assert: (self isBaseFrame: theFP) not.
	self assert: self validStackPageBaseFrames.
	callerFP := self frameCallerFP: theFP.
	self assert: (self frameHasContext: callerFP).
	self assert: (objectMemory isContext: (self frameContext: callerFP)).
	theContext := self ensureFrameIsMarried: theFP
					SP: theFP + ((self isMachineCodeFrame: theFP) ifTrue: [FoxMFReceiver] ifFalse: [FoxIFReceiver]).
	stackPages
		longAt: (newSP := newPage baseAddress) put: (self frameContext: callerFP);
		longAt: (newSP := newSP - objectMemory wordSize) put:  theContext.
	stackedReceiverOffset := self frameStackedReceiverOffset: theFP.
	"First move the data, leaving room for the caller and base frame contexts.  We will fix up frame pointers later."
	theFP + stackedReceiverOffset
		to: oldPage headSP
		by: objectMemory wordSize negated
		do: [:source|
			newSP := newSP - objectMemory wordSize.
			stackPages longAt: newSP put: (stackPages longAt: source)].
	"newSP = oldSP + delta => delta = newSP - oldSP"
	delta := newSP - oldPage headSP.
	newFP := newPage baseAddress - stackedReceiverOffset - (2 * objectMemory wordSize).
	self setHeadFP: oldPage headFP + delta andSP: newSP inPage: newPage.
	newPage baseFP: newFP.
	callerIP := self oopForPointer: (self frameCallerSavedIP: theFP).
	callerIP asUnsignedInteger >= objectMemory startOfMemory ifTrue:
		[self iframeSavedIP: callerFP put: callerIP.
		 callerIP := cogit ceReturnToInterpreterPC].
	stackPages longAt: theFP + stackedReceiverOffset put: callerIP.
	self assert: (callerFP < oldPage baseAddress
				and: [callerFP > (oldPage realStackLimit - (LargeContextSlots * objectMemory bytesPerOop / 2))]).
	oldPage
		headFP: callerFP;
		headSP: theFP + stackedReceiverOffset.
	"Mark the new base frame in the new page"
	stackPages
		longAt: newFP + FoxCallerSavedIP put: cogit ceBaseFrameReturnPC;
		longAt: newFP + FoxSavedFP put: 0.
	"Now relocate frame pointers, updating married contexts to refer to their moved spouse frames."
	fpInNewPage := newPage headFP.
	[offsetCallerFP := self frameCallerFP: fpInNewPage.
	 offsetCallerFP ~= 0 ifTrue:
		[offsetCallerFP := offsetCallerFP + delta].
	 stackPages longAt: fpInNewPage + FoxSavedFP put: (self oopForPointer: offsetCallerFP).
	 (self frameHasContext: fpInNewPage) ifTrue:
		[theContext := self frameContext: fpInNewPage.
		 objectMemory storePointerUnchecked: SenderIndex
			ofObject: theContext
			withValue: (self withSmallIntegerTags: fpInNewPage).
		 objectMemory storePointerUnchecked: InstructionPointerIndex
			ofObject: theContext
			withValue: (self withSmallIntegerTags: offsetCallerFP)].
	 fpInNewPage := offsetCallerFP.
	 fpInNewPage ~= 0] whileTrue.
	self assert: self validStackPageBaseFrames.
	^newFP! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
mtemporary: offset in: theFP put: valueOop
	"Temporary access for a machine code frame only."
	"See StackInterpreter class>>initializeFrameIndices"
	| frameNumArgs |
	<inline: true>
	<var: #theFP type: #'char *'>
	^stackPages
		longAt: (offset < (frameNumArgs := self mframeNumArgs: theFP)
					ifTrue: [theFP + FoxCallerSavedIP + ((frameNumArgs - offset) * objectMemory wordSize)]
					ifFalse: [theFP + FoxMFReceiver - objectMemory wordSize + ((frameNumArgs - offset) * objectMemory wordSize)])
		put: valueOop! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 12/21/2016 10:54'!
mustMapMachineCodePC: theIP context: aOnceMarriedContext
	"Map the native pc theIP into a bytecode pc integer object and answer it.
	 See contextInstructionPointer:frame: for the explanation."
	| maybeClosure methodObj cogMethod startBcpc bcpc |
	<inline: false>
	<var: #cogMethod type: #'CogMethod *'>
	theIP = HasBeenReturnedFromMCPC ifTrue:
		[^objectMemory nilObject].
	maybeClosure := objectMemory fetchPointer: ClosureIndex ofObject: aOnceMarriedContext.
	methodObj := objectMemory fetchPointer: MethodIndex ofObject: aOnceMarriedContext.
	(maybeClosure ~= objectMemory nilObject
	and: [self isVanillaBlockClosure: maybeClosure])
		ifTrue: [self assert: (theIP signedBitShift: -16) < -1.
				startBcpc := self startPCOfClosure: maybeClosure]
		ifFalse: [self assert: (theIP signedBitShift: -16) = -1.
				startBcpc := self startPCOfMethod: methodObj].
	cogMethod := self ensureMethodIsCogged: methodObj maybeClosure: maybeClosure.
	bcpc := self bytecodePCFor: theIP cogMethod: cogMethod startBcpc: startBcpc.
	self assert: bcpc >= (self startPCOfMethod: methodObj).
	"If there's a CallPrimitive we need to skip it."
	(bcpc = startBcpc
	 and: [maybeClosure = objectMemory nilObject
	 and: [self methodHeaderHasPrimitive: cogMethod methodHeader]]) ifTrue:
		[bcpc := bcpc + (self sizeOfCallPrimitiveBytecode: cogMethod methodHeader)].
	^objectMemory integerObjectOf: bcpc + 1! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:39'!
nativeFramePointerIn: theFP
	"See StackInterpreter class>>initializeFrameIndices"
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	^(self isMachineCodeFrame: theFP)
		ifTrue:
			[stackPages pointerAtPointer: theFP + FoxMFReceiver - (BytesPerWord*3) ]
		ifFalse:
			[stackPages pointerAtPointer: theFP + FoxIFReceiver - (BytesPerWord*3) ]! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:39'!
nativeFramePointerIn: theFP put: valueOopPointer
	"See StackInterpreter class>>initializeFrameIndices"
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	<var: #valueOopPointer type: #'char*'>
	^(self isMachineCodeFrame: theFP)
		ifTrue:
			[stackPages pointerAtPointer: theFP + FoxMFReceiver - (BytesPerWord*3) put: valueOopPointer]
		ifFalse:
			[stackPages pointerAtPointer: theFP + FoxIFReceiver - (BytesPerWord*3) put: valueOopPointer]! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:39'!
nativePreviousStackPointerIn: theFP
	"See StackInterpreter class>>initializeFrameIndices"
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	^(self isMachineCodeFrame: theFP)
		ifTrue:
			[stackPages pointerAtPointer: theFP + FoxMFReceiver - (BytesPerWord*2)]
		ifFalse:
			[stackPages pointerAtPointer: theFP + FoxIFReceiver - (BytesPerWord*2)]! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:39'!
nativePreviousStackPointerIn: theFP put: valueOopPointer
	"See StackInterpreter class>>initializeFrameIndices"
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	<var: #valueOopPointer type: #'char *'>
	^(self isMachineCodeFrame: theFP)
		ifTrue:
			[stackPages pointerAtPointer: theFP + FoxMFReceiver - (BytesPerWord*2) put: valueOopPointer]
		ifFalse:
			[stackPages pointerAtPointer: theFP + FoxIFReceiver - (BytesPerWord*2) put: valueOopPointer]! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'rsf 8/10/2016 00:11'!
nativeStackPointerAddress
	<option: #LowcodeVM>
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: nativeStackPointer) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #nativeStackPointer in: self]! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:39'!
nativeStackPointerIn: theFP
	"See StackInterpreter class>>initializeFrameIndices"
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	^(self isMachineCodeFrame: theFP)
		ifTrue:
			[stackPages pointerAtPointer: theFP + FoxMFReceiver - (BytesPerWord*4)]
		ifFalse:
			[stackPages pointerAtPointer: theFP + FoxIFReceiver - (BytesPerWord*4)]! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:40'!
nativeStackPointerIn: theFP put: valueOopPointer
	"See StackInterpreter class>>initializeFrameIndices"
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	<var: #valueOopPointer type: #'char *'>
	^(self isMachineCodeFrame: theFP)
		ifTrue:
			[stackPages pointerAtPointer: theFP + FoxMFReceiver - (BytesPerWord*4) put: valueOopPointer]
		ifFalse:
			[stackPages pointerAtPointer: theFP + FoxIFReceiver - (BytesPerWord*4) put: valueOopPointer]! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 5/6/2009 13:11'!
newMethod
	<doNotGenerate>
	^newMethod! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 1/30/2009 17:55'!
newMethod: oop
	<doNotGenerate>
	newMethod := oop! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'eem 5/6/2010 16:00'!
newMethodAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: newMethod) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #newMethod in: self]! !

!CoInterpreter methodsFor: 'method lookup cache' stamp: 'eem 8/4/2016 13:10'!
newMethodInLookupCacheAt: selector and: classTag
	"Answer if classTag x messageSelector => newMethod is in the lookup cache.
	 This is for assert checking to check that open PICs find entries."
	| probe hash |
	<inline: false>
	hash := objectMemory methodCacheHashOf: selector with: classTag.

	0 to: CacheProbeMax-1 do:
		[:p |
		probe := (hash >> p) bitAnd: MethodCacheMask.
		((methodCache at: probe + MethodCacheSelector) = selector
		 and: [(methodCache at: probe + MethodCacheClass) = classTag
		 and: [(methodCache at: probe + MethodCacheMethod) = newMethod]]) ifTrue:
			[^true]].

	^false! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 5/5/2009 11:48'!
nextProfileTick
	<doNotGenerate>
	^nextProfileTick! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'eem 8/5/2021 10:10'!
nextProfileTickAddress
	<api>
	<returnTypeC: #usqInt>
	"N.B. nextProfileTick is 64-bits"
	^self cCode: [(self addressOf: nextProfileTick) asUnsignedInteger]
		inSmalltalk:
			[objectMemory wordSize = 8
				ifTrue:
					[cogit simulatedReadWriteVariableAddress: #nextProfileTick in: self]
				ifFalse:
					["This looks weird, but the cogit doles out fake addresses counting down from
					  16r80000000, so we have to ask for the halves backwards to get them in the
					  right order, and ask for both now so that botehr get defined at teh same time
					  and are hence contiguous in memory."
					 VMBIGENDIAN
						ifTrue:
							[cogit simulatedReadWriteVariableAddress: #nextProfileTickHigh in: self.
							 cogit simulatedReadWriteVariableAddress: #nextProfileTickLow in: self.
							 cogit simulatedReadWriteVariableAddress: #nextProfileTickHigh in: self]
						ifFalse:
							[cogit simulatedReadWriteVariableAddress: #nextProfileTickLow in: self.
							 cogit simulatedReadWriteVariableAddress: #nextProfileTickHigh in: self.
							 cogit simulatedReadWriteVariableAddress: #nextProfileTickLow in: self]]]! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 2/3/2011 17:55'!
nextProfileTickHigh
	<doNotGenerate>
	^nextProfileTick bitShift: -32! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 2/3/2011 17:55'!
nextProfileTickLow
	<doNotGenerate>
	^nextProfileTick bitAnd: 16rFFFFFFFF! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 7/25/2021 12:01'!
nilLocalFP
	<doNotGenerate>
	"For use *ONLY* by handleCallOrJumpSimulationTrap:. Nil
	 localFP so that the head frame can be determined reliably."
	localFP := nil! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 8/2/2016 15:21'!
nilUncoggableMethods
	<inline: true>
	lastCoggableInterpretedBlockMethod := lastUncoggableInterpretedBlockMethod := nil! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem 8/17/2011 12:24'!
noAssertHeaderOf: methodPointer
	<api>
	| methodHeader |
	methodHeader := self rawHeaderOf: methodPointer.
	^(self isCogMethodReference: methodHeader)
		ifTrue: [(self cCoerceSimple: methodHeader to: #'CogMethod *') methodHeader]
		ifFalse: [methodHeader]! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'eem 7/10/2014 20:54'!
positive32BitIntegerFor: integerValue
	<api>
	^super positive32BitIntegerFor: integerValue! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'RonieSalgado 4/29/2016 15:19'!
positive32BitValueOf: oop
	<api>
	^super positive32BitValueOf: oop! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'RonieSalgado 4/29/2016 15:19'!
positive64BitIntegerFor: integerValue
	<api>
	^super positive64BitIntegerFor: integerValue! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'RonieSalgado 4/29/2016 15:19'!
positive64BitValueOf: oop
	<api>
	^super positive64BitValueOf: oop! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 9/10/2019 18:59'!
postBecomeAction: theBecomeEffectsFlags
	"Clear the gcMode var and let the Cogit do its post GC checks."
	super postBecomeAction: theBecomeEffectsFlags.

	(objectMemory hasSpurMemoryManagerAPI)
		ifTrue: [cogit followMovableLiteralsAndUpdateYoungReferrers]
		ifFalse: [cogit cogitPostGCAction: gcMode].
	self nilUncoggableMethods.
	self assert: cogit kosherYoungReferrers.
	gcMode := 0! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 8/2/2016 15:22'!
postGCAction: gcModeArg
	"Attempt to shrink free memory, signal the gc semaphore and let the Cogit do its post GC thang"
	<inline: false>
	self assert: gcModeArg = gcMode.
	super postGCAction: gcModeArg.
	cogit cogitPostGCAction: gcModeArg.
	self nilUncoggableMethods.
	gcMode := 0! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 8/5/2011 13:13'!
preBecomeAction
	"Need to set gcMode var (to avoid passing the flag through a lot of the updating code)"
	super preBecomeAction.
	gcMode := GCModeBecome! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'eem 6/1/2014 16:51'!
preGCAction: gcModeArg
	<inline: false>
	"Need to write back the frame pointers unless all pages are free (as in snapshot).
	 Need to set gcMode var (to avoid passing the flag through a lot of the updating code)"
	super preGCAction: gcModeArg.

	gcMode := gcModeArg.

	cogit recordEventTrace ifTrue:
		[| traceType |
		traceType := gcModeArg == GCModeFull ifTrue: [TraceFullGC] ifFalse: [TraceIncrementalGC].
		self recordTrace: traceType thing: traceType source: 0].

	cogit recordPrimTrace ifTrue:
		[| traceType |
		traceType := gcModeArg == GCModeFull ifTrue: [TraceFullGC] ifFalse: [TraceIncrementalGC].
		self fastLogPrim: traceType]! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
primErrTable
	<api>
	^objectMemory splObj: PrimErrTableIndex! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 1/2/2009 12:21'!
primFailCode
	<doNotGenerate>
	^primFailCode! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 1/2/2009 12:28'!
primFailCode: anInteger
	<doNotGenerate>
	primFailCode := anInteger! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'eem 5/6/2010 16:00'!
primFailCodeAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: primFailCode) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #primFailCode in: self]! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem 05/07/2020 16:13'!
primNumberExternalCall
	"Answer if the method is an external primitive call (prim 117)."
	<api>
	<cmacro>
	^PrimNumberExternalCall! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
primTraceLogAddress
	<api>
	<returnTypeC: #'void *'>
	^self cCode: [primTraceLog] inSmalltalk: [primTraceLog offset * objectMemory wordSize]! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 4/26/2010 15:52'!
primTraceLogIndex
	<doNotGenerate>
	^primTraceLogIndex! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 5/7/2010 12:34'!
primTraceLogIndex: aValue
	<cmacro: '(aValue) (GIV(primTraceLogIndex) = (aValue))'>
	"N.B. primTraceLogIndex is 8-bits"
	^primTraceLogIndex := aValue bitAnd: 16rFF! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 5/10/2010 17:58'!
primTraceLogIndexAddress
	<api>
	<returnTypeC: #usqInt>
	"N.B. primTraceLogIndex is 8-bits"
	^self cCode: [(self addressOf: primTraceLogIndex) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #primTraceLogIndex in: self]! !

!CoInterpreter methodsFor: 'control primitives' stamp: 'eem 1/9/2020 10:33'!
primitiveExecuteMethod
	"receiver, args, then method are on top of stack. Execute method against receiver and args.
	 Set primitiveFunctionPointer because no cache lookup has been done for the method, and
	 hence primitiveFunctionPointer is stale."
	| methodArgument primitiveIndex |
	methodArgument := self stackTop.
	(objectMemory isOopCompiledMethod: methodArgument) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	argumentCount - 1 = (self argumentCountOf: methodArgument) ifFalse:
		[^self primitiveFailFor: PrimErrBadNumArgs].
	newMethod := self popStack.
	primitiveIndex := self primitiveIndexOf: newMethod.
	primitiveFunctionPointer := self functionPointerFor: primitiveIndex inClass: nil.
	argumentCount := argumentCount - 1.
	self executeNewMethodJitting.
	"Recursive xeq affects primErrorCode"
	self initPrimCall! !

!CoInterpreter methodsFor: 'control primitives' stamp: 'eem 1/9/2020 10:33'!
primitiveExecuteMethodArgsArray
	"receiver, argsArray, then method are on top of stack.  Execute method against
	 receiver and args.  Allow for up to two extra arguments (e.g. for mirror primitives).
	 Set primitiveFunctionPointer because no cache lookup has been done for the
	 method, and hence primitiveFunctionPointer is stale."
	| methodArgument argCnt rcvr argumentArray primitiveIndex |
	methodArgument := self stackTop.
	argumentArray := self stackValue: 1.
	((objectMemory isOopCompiledMethod: methodArgument)
	 and: [objectMemory isArray: argumentArray]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	argCnt := self argumentCountOf: methodArgument.
	argCnt = (objectMemory numSlotsOf: argumentArray) ifFalse:
		[^self primitiveFailFor: PrimErrBadNumArgs].
	argumentCount > 2 ifTrue: "CompiledMethod class>>receiver:withArguments:executeMethod:
								SqueakObjectPrimitives class>>receiver:withArguments:apply:
								VMMirror>>ifFail:object:with:executeMethod: et al"
		[rcvr := self stackValue: 2.
		 (argumentCount > 4
		  or: [objectMemory isOopForwarded: rcvr]) ifTrue:
			[^self primitiveFailFor: PrimErrUnsupported].
		self stackValue: argumentCount put: rcvr]. "replace actual receiver with desired receiver"
	"and push the actual arguments"
	self pop: argumentCount.
	0 to: argCnt - 1 do:
		[:i|
		self push: (objectMemory fetchPointer: i ofObject: argumentArray)].
	newMethod := methodArgument.
	primitiveIndex := self primitiveIndexOf: newMethod.
	primitiveFunctionPointer := self functionPointerFor: primitiveIndex inClass: nil.
	argumentCount := argCnt.
	self executeNewMethodJitting.
	"Recursive xeq affects primErrorCode"
	self initPrimCall! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'eem 12/12/2012 11:35'!
primitiveFailAddress
	"This is used for asserts that check that inline cache editing results in valid addresses.
	 In the C VM interpret is presumed to come before any primitives and so it constitutes
	 the lowest address in C code that machine code should be linked, but optimizing
	 compilers change things around.  In the simulator we just answer something not low."
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: #primitiveFail) asUnsignedInteger]
		inSmalltalk: [heapBase]! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 1/21/2011 09:18'!
primitiveFunctionPointer: oop
	"Apparently not sent but is used in the simulator."
	<doNotGenerate>
	primitiveFunctionPointer := oop! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 10/30/2014 13:04'!
primitivePropertyFlags: primIndex
	<api>
	"Answer any special requirements of the given primitive"
	objectMemory hasSpurMemoryManagerAPI
		ifTrue: [^self primitivePropertyFlagsForSpur: primIndex]
		ifFalse: [^self primitivePropertyFlagsForV3: primIndex]! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 8/4/2021 12:51'!
primitivePropertyFlagsFor: externalBytecodedMethod
	<doNotGenerate>
	"Support for in-image compilation.  This is intended to do for methods in the image what
	 functionPointerForCompiledMethod:primitiveIndex:primitivePropertyFlagsInto: does for methods in the simulation."
	| flags metadata |
	flags := self primitivePropertyFlags: externalBytecodedMethod primitive.
	externalBytecodedMethod primitive = PrimNumberExternalCall ifTrue:
		 [metadata := self class metadataFlagsForPrimitive: externalBytecodedMethod.
		  flags := flags bitOr: ((metadata bitAnd: SpurPrimitiveFlagsMask) bitShift: PrimitiveMetadataFlagsShift).
		(InitializationOptions at: #profiling ifAbsent: [false]) ifTrue:
			[flags := flags bitOr: PrimCallCollectsProfileSamples]].
	^flags! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 8/4/2021 12:57'!
primitivePropertyFlagsFor: externalBytecodedMethod primitiveIndex: primIndex
	<doNotGenerate>
	"Support for in-image compilation.  This is intended to do for methods in the image what
	 functionPointerForCompiledMethod:primitiveIndex:primitivePropertyFlagsInto: does for methods in the simulation."
	| flags metadata |
	flags := self primitivePropertyFlags: primIndex.
	primIndex = PrimNumberExternalCall ifTrue:
		[metadata := self class metadataFlagsForPrimitive: externalBytecodedMethod.
		 "nuke base flags..."
		 flags := (metadata bitAnd: SpurPrimitiveFlagsMask) bitShift: PrimitiveMetadataFlagsShift.
		 (InitializationOptions at: #profiling ifAbsent: [false]) ifTrue:
			[flags := flags bitOr: PrimCallCollectsProfileSamples]].
	^flags! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 9/4/2021 17:33'!
primitivePropertyFlagsForSpur: primIndex
	<inline: true>
	"Answer any special requirements of the given primitive.  Spur always needs to set
	 primitiveFunctionPointer and newMethod so primitives can retry on failure due to forwarders."
	| baseFlags |
	self cCode: [] inSmalltalk: [#(primitiveObjectAtPut primitiveExternalCall primitiveCalloutToFFI)]. "For senders..."
	baseFlags := profileSemaphore = objectMemory nilObject
					ifTrue: [0]
					ifFalse: [PrimCallCollectsProfileSamples].
	(primIndex = PrimNumberObjectAtPut
	 or: [primIndex = PrimNumberCopyObject]) ifTrue:
		[^baseFlags + PrimCallOnSmalltalkStack].
	baseFlags := baseFlags + PrimCallNeedsPrimitiveFunction + PrimCallNeedsNewMethod.

	(self isCalloutPrimitiveIndex: primIndex) ifTrue: "For callbacks & module unloading"
		[^baseFlags + PrimCallMayEndureCodeCompaction + PrimCallIsExternalCall].
	(self isCodeCompactingPrimitiveIndex: primIndex) ifTrue: "For code reclamations"
		[^baseFlags bitOr: PrimCallMayEndureCodeCompaction].

	^baseFlags! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 8/4/2021 11:18'!
primitivePropertyFlagsForV3: primIndex
	<inline: true>
	"Answer any special requirements of the given primitive"
	| baseFlags |
	baseFlags := profileSemaphore ~= objectMemory nilObject
					ifTrue: [PrimCallNeedsNewMethod + PrimCallCollectsProfileSamples]
					ifFalse: [0].

	(self isCalloutPrimitiveIndex: primIndex) ifTrue: "For callbacks & module unloading"
		[baseFlags := baseFlags bitOr: PrimCallNeedsNewMethod + PrimCallNeedsPrimitiveFunction + PrimCallMayEndureCodeCompaction + PrimCallIsExternalCall].
	(self isCodeCompactingPrimitiveIndex: primIndex) ifTrue:
		[baseFlags := baseFlags bitOr: PrimCallNeedsNewMethod + PrimCallMayEndureCodeCompaction].

	^baseFlags! !

!CoInterpreter methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:48'!
printCogMethod: cogMethod
	<export: true> "useful for VM debugging; use export: so it will be accessible on win32"
	<api>
	<var: #cogMethod type: #'CogMethod *'>
	| address primitive |
	self cCode: ''
		inSmalltalk:
			[self transcript ensureCr.
			 cogMethod isInteger ifTrue:
				[^self printCogMethod: (self cCoerceSimple: cogMethod to: #'CogMethod *')]].
	address := cogMethod asInteger.
	self printHex: address;
		print: ' <-> ';
		printHex: address + cogMethod blockSize.
	cogMethod cmType = CMMethod ifTrue:
		[self print: ': method: ';
			printHex: cogMethod methodObject.
		 primitive := self primitiveIndexOfMethod: cogMethod methodObject
							header: cogMethod methodHeader.
		 primitive ~= 0 ifTrue:
			[self print: ' prim '; printNum: primitive].
		 ((objectMemory addressCouldBeObj: cogMethod methodObject)
		 and: [objectMemory addressCouldBeObj: (self methodClassOf: cogMethod methodObject)]) ifTrue:
			[self space; printNameOfClass: (self methodClassOf: cogMethod methodObject) count: 2]].
	cogMethod cmType = CMBlock ifTrue:
		[self print: ': block home: ';
			printHex: (self cCoerceSimple: cogMethod to: #'CogBlockMethod *') cmHomeMethod asUnsignedInteger].
	cogMethod cmType = CMClosedPIC ifTrue:
		[self print: ': Closed PIC N: ';
			printHex: cogMethod cPICNumCases].
	cogMethod cmType = CMOpenPIC ifTrue:
		[self print: ': Open PIC '].
	self print: ' selector: '; printHex: cogMethod selector.
	cogMethod selector = objectMemory nilObject
		ifTrue: [| s |
			(cogMethod cmType = CMMethod
			 and: [(s := self maybeSelectorOfMethod: cogMethod methodObject) notNil])
				ifTrue: [self print: ' (nil: '; printStringOf: s; print: ')']
				ifFalse: [self print: ' (nil)']]
		ifFalse: [self space; printStringOf: cogMethod selector].
	self cr! !

!CoInterpreter methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printFrame: theFP WithSP: theSP
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| theMethod theMethodEnd numArgs numTemps rcvrAddress topThing |
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	<var: #addr type: #'char *'>
	<var: #rcvrAddress type: #'char *'>
	<var: #cogMethod type: #'CogBlockMethod *'>
	<var: #homeMethod type: #'CogMethod *'>
	self cCode: '' inSmalltalk: [self transcript ensureCr].
	(stackPages couldBeFramePointer: theFP) ifNil:
		[self printHexPtr: theFP; print: ' is not in the stack zone?!!'; cr.
		 ^nil].
	(self isMachineCodeFrame: theFP)
		ifTrue:
			[| cogMethod homeMethod |
			 cogMethod := self mframeCogMethod: theFP.
			 homeMethod := self mframeHomeMethod: theFP.
			 theMethod := homeMethod asInteger.
			 theMethodEnd := homeMethod asInteger + homeMethod blockSize.
			 numArgs := cogMethod cmNumArgs.
			 numTemps := self temporaryCountOfMethodHeader: homeMethod methodHeader]
		ifFalse:
			[theMethod := self frameMethodObject: theFP.
			 theMethodEnd := theMethod + (objectMemory sizeBitsOfSafe: theMethod).
			 numArgs := self iframeNumArgs: theFP.
			 numTemps := self tempCountOf: theMethod].
	(self frameIsBlockActivation: theFP) ifTrue:
		[| rcvrOrClosure |
		 "No BlockLocalTempCounter in the Cogit's C code, so quick hack is to use numCopied + numArgs"
		 rcvrOrClosure := self pushedReceiverOrClosureOfFrame: theFP.
		 ((objectMemory isNonImmediate: rcvrOrClosure)
		 and: [(objectMemory addressCouldBeObj: rcvrOrClosure)
		 and: [(objectMemory fetchClassOfNonImm: rcvrOrClosure) = (objectMemory splObj: ClassBlockClosure)]])
			ifTrue: [numTemps := numArgs + (self stSizeOf: rcvrOrClosure)]
			ifFalse: [numTemps := numArgs]].
	self shortPrintFrame: theFP.
	(self isBaseFrame: theFP) ifTrue:
		[self printFrameOop: '(caller ctxt'
			at: theFP + (self frameStackedReceiverOffset: theFP) + (2 * objectMemory wordSize).
		 self printFrameOop: '(saved ctxt'
			at: theFP + (self frameStackedReceiverOffset: theFP) + (1 * objectMemory wordSize)].
	self printFrameOop: 'rcvr/clsr'
		at: theFP + FoxCallerSavedIP + ((numArgs + 1) * objectMemory wordSize).
	numArgs to: 1 by: -1 do:
		[:i|
		self printFrameOop: 'arg' index: numArgs - i at: theFP + FoxCallerSavedIP + (i * objectMemory wordSize)].
	self printFrameThing: 'caller ip'
		at: theFP + FoxCallerSavedIP
		extraString: ((stackPages longAt: theFP + FoxCallerSavedIP) = cogit ceReturnToInterpreterPC ifTrue:
						['ceReturnToInterpreter']).
	self printFrameThing: 'saved fp' at: theFP + FoxSavedFP.
	self printFrameMethodFor: theFP.
	(self isMachineCodeFrame: theFP) ifTrue:
		[self printFrameFlagsForFP: theFP].
	self printFrameOop: 'context' at: theFP + FoxThisContext.
	(self isMachineCodeFrame: theFP) ifFalse:
		[self printFrameFlagsForFP: theFP].
	(self isMachineCodeFrame: theFP)
		ifTrue: [rcvrAddress := theFP + FoxMFReceiver]
		ifFalse:
			[self printFrameThing: 'saved ip'
				at: theFP + FoxIFSavedIP
				extra: ((self iframeSavedIP: theFP) = 0
							ifTrue: [0]
							ifFalse: [(self iframeSavedIP: theFP) - theMethod + 2 - objectMemory baseHeaderSize]).
			 rcvrAddress := theFP + FoxIFReceiver].
	self printFrameOop: 'receiver' at: rcvrAddress.
	topThing := stackPages longAt: theSP.
	(self oop: topThing isGreaterThanOrEqualTo: theMethod andLessThan: theMethodEnd)
		ifTrue:
			[rcvrAddress - objectMemory wordSize to: theSP + objectMemory wordSize by: objectMemory wordSize negated do:
				[:addr| | index |
				index := rcvrAddress - addr / objectMemory wordSize + numArgs.
				index <= numTemps
					ifTrue: [self printFrameOop: 'temp' index: index - 1 at: addr]
					ifFalse: [self printFrameOop: ((self frameIsBlockActivation: theFP)
													ifTrue: ['temp/stck']
													ifFalse: ['stck'])
								at: addr]].
			self printFrameThing: 'frame ip'
				at: theSP
				extra: ((self isMachineCodeFrame: theFP)
						ifTrue: [topThing - theMethod]
						ifFalse: [topThing - theMethod + 2 - objectMemory baseHeaderSize])]
		ifFalse:
			[rcvrAddress - objectMemory wordSize to: theSP by: objectMemory wordSize negated do:
				[:addr| | index |
				index := rcvrAddress - addr / objectMemory wordSize + numArgs.
				index <= numTemps
					ifTrue: [self printFrameOop: 'temp' index: index - 1 at: addr]
					ifFalse: [self printFrameOop: ((self frameIsBlockActivation: theFP)
													ifTrue: ['temp/stck']
													ifFalse: ['stck'])
								at: addr]]]! !

!CoInterpreter methodsFor: 'debug printing' stamp: 'eem 6/26/2012 11:18'!
printFrameFlagsForFP: theFP
	| address it |
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #address type: #'char *'>
	(self isMachineCodeFrame: theFP)
		ifTrue:
			[address := theFP + FoxMethod.
			it := (stackPages longAt: address) bitAnd: 16r7]
		ifFalse:
			[address := theFP + FoxIFrameFlags.
			 it := stackPages longAt: address].
	self printHexPtr: address;
		print: ((self isMachineCodeFrame: theFP)
				ifTrue: [': mcfrm flags: ']
				ifFalse: [':intfrm flags: ']);
		printHex: it.
	it ~= 0 ifTrue:
		[self printChar: $=; printNum: it].
	self print: '  numArgs: '; printNum: (self frameNumArgs: theFP);
		print: ((self frameHasContext: theFP) ifTrue: [' hasContext'] ifFalse: [' noContext']);
		print: ((self frameIsBlockActivation: theFP) ifTrue: [' isBlock'] ifFalse: [' notBlock']);
		cr! !

!CoInterpreter methodsFor: 'debug printing' stamp: 'eem 11/27/2016 11:49'!
printFrameMethodFor: theFP
	<inline: false>
	| address it homeMethod obj |
	<var: #theFP type: #'char *'>
	<var: #address type: #'char *'>
	<var: #homeMethod type: #'CogMethod *'>

	address := theFP + FoxMethod.
	it := stackPages longAt: address.
	self printHex: address asInteger;
		printChar: $:.
	self print: '      method: ';
		printHex: it.
	self tab.
	((self isMachineCodeFrame: theFP)
	 and: [self mframeIsBlockActivation: theFP]) ifTrue:
		[homeMethod := self mframeHomeMethod: theFP.
		 self print: 'hm: '; printHex: homeMethod asInteger; tab].
	obj := self frameMethodObject: theFP.
	self shortPrintOop: obj! !

!CoInterpreter methodsFor: 'debug printing' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
printFrameThing: name at: address extra: extraValue
	| it len |
	<inline: false>
	<var: #name type: #'char *'>
	<var: #address type: #'char *'>
	it := stackPages longAt: address.
	self printHexPtr: address;
		printChar: $:.
	len := self strlen: name.
	1 to: 12 - len do: [:i| self space].
	self print: name;
		print: ': ';
		printHex: it.
	it ~= 0 ifTrue:
		[self printChar: $=.
		 it = objectMemory nilObject
			ifTrue: [self print: 'nil']
			ifFalse:
				[self printNum: it]].
	self space; printNum: extraValue; cr! !

!CoInterpreter methodsFor: 'debug printing' stamp: 'eem 12/4/2013 13:27'!
printFrameThing: name at: address extraString: extraStringOrNil
	| it len |
	<inline: false>
	<var: #name type: #'char *'>
	<var: #address type: #'char *'>
	<var: #extraStringOrNil type: #'char *'>
	it := stackPages longAt: address.
	self printHexPtr: address;
		printChar: $:.
	len := self strlen: name.
	1 to: 12 - len do: [:i| self space].
	self print: name;
		print: ': ';
		printHex: it.
	it ~= 0 ifTrue:
		[self printChar: $=.
		 it = objectMemory nilObject
			ifTrue: [self print: 'nil']
			ifFalse:
				[self printNum: it]].
	extraStringOrNil ifNotNil: [self space; print: extraStringOrNil].
	self cr! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 11/27/2012 14:57'!
printLogEntryAt: i
	<inline: false>
	| intOrClass selectorMethodOrProcess source |
	intOrClass := traceLog at: i.
	selectorMethodOrProcess := traceLog at: i + 1.
	source := traceLog at: i + 2.
	source <= TraceIsFromInterpreter ifTrue:
		[self print: (traceSources at: source); space].
	(objectMemory isIntegerObject: intOrClass)
		ifTrue:
			[intOrClass = TraceStackOverflow ifTrue:
				[self print: 'stack overflow'].
			 intOrClass = TraceContextSwitch ifTrue:
				[self print: 'context switch from '; printHex: selectorMethodOrProcess].
			 intOrClass = TraceBlockActivation ifTrue:
				[self print: ' [] in '; printHex: selectorMethodOrProcess].
			 intOrClass = TraceBlockCreation ifTrue:
				[self print: 'create [] '; printHex: selectorMethodOrProcess].
			 intOrClass = TraceIncrementalGC ifTrue:
				[self print: 'incrementalGC'].
			 intOrClass = TraceFullGC ifTrue:
				[self print: 'fullGC'].
			 intOrClass = TraceCodeCompaction ifTrue:
				[self print: 'compactCode'].
			 intOrClass = TraceVMCallback ifTrue:
				[self print: 'callback'].
			 intOrClass = TraceVMCallbackReturn ifTrue:
				[self print: 'return from callback']]
		ifFalse:
			[self space; printNameOfClass: intOrClass count: 5; print: '>>'; printStringOf: selectorMethodOrProcess].
	source > TraceIsFromInterpreter ifTrue:
		[self space; print: (traceSources at: source)].
	self cr! !

!CoInterpreter methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printMethodCacheFor: thing
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| n |
	n := 0.
	0 to: MethodCacheSize - 1 by: MethodCacheEntrySize do:
		[:i | | s c m p |
		s := methodCache at: i + MethodCacheSelector.
		c := methodCache at: i + MethodCacheClass.
		m := methodCache at: i + MethodCacheMethod.
		p := methodCache at: i + MethodCachePrimFunction.
		((thing = -1 or: [s = thing or: [c = thing or: [p = thing or: [m = thing
			or: [(objectMemory addressCouldBeObj: m)
				and: [(self maybeMethodHasCogMethod: m)
				and: [(self cogMethodOf: m) asInteger = thing]]]]]]])
		 and: [(objectMemory addressCouldBeOop: s)
		 and: [c ~= 0
		 and: [(self addressCouldBeClassObj: c)
			or: [self addressCouldBeClassObj: (objectMemory classForClassTag: c)]]]]) ifTrue:
			[n := n + 1.
			 self cCode: [] inSmalltalk: [self transcript ensureCr].
			 self printNum: i; space; printHexnp: i; cr; tab.
			 (objectMemory isBytesNonImm: s)
				ifTrue: [self cCode: 'printf("%" PRIxSQPTR " %.*s\n", s, (int)(numBytesOf(s)), (char *)firstIndexableField(s))'
						inSmalltalk: [self printHex: s; space; print: (self stringOf: s); cr]]
				ifFalse: [self shortPrintOop: s].
			 self tab.
			 (self addressCouldBeClassObj: c)
				ifTrue: [self shortPrintOop: c]
				ifFalse: [self printNum: c; space; printHexnp: c; space; shortPrintOop: (objectMemory classForClassTag: c)].
			self tab; shortPrintOop: m; tab.
			self cCode:
					[p > 1024
						ifTrue: [self printHexnp: p]
						ifFalse: [self printNum: p]]
				inSmalltalk:
					[p isSymbol ifTrue: [self print: p] ifFalse: [self printNum: p]].
			self cr]].
	n > 1 ifTrue:
		[self printNum: n; cr]! !

!CoInterpreter methodsFor: 'debug printing' stamp: 'eem 2/23/2020 23:41'!
printMethodFieldForPrintContext: aContext
	<inline: true>
	| meth |
	meth := objectMemory fetchPointer: MethodIndex ofObject: aContext.
	self printOopShortInner: meth.
	(self methodHasCogMethod: meth) ifTrue:
		[self space; print: '('; printHexnp: (self cogMethodOf: meth); print: ')'].
	self cr! !

!CoInterpreter methodsFor: 'debug printing' stamp: 'eem 2/24/2020 08:16'!
printMethodHeaderOop: anOop
	"Print the CogMethod and its header if this is a CogMethod reference."
	<var: #cogMethod type: #'CogMethod *'>
	(self isCogMethodReference: anOop) ifTrue:
		[(cogMethodZone methodFor: (self pointerForOop: anOop)) ifNotNil:
			[:cogMethod| ^self printHex: anOop; space; printDecodeMethodHeaderOop: cogMethod methodHeader]].
	^self printDecodeMethodHeaderOop: anOop! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 5/15/2015 17:13'!
printPrimLogEntryAt: i
	<inline: false>
	| intOrSelector |
	intOrSelector := primTraceLog at: i.
	(objectMemory isImmediate: intOrSelector)
		ifTrue:
			[intOrSelector = TraceIncrementalGC ifTrue:
				[self print: '**IncrementalGC**'. ^nil].
			 intOrSelector = TraceFullGC ifTrue:
				[self print: '**FullGC**'. ^nil].
			 intOrSelector = TraceCodeCompaction ifTrue:
				[self print: '**CompactCode**'. ^nil].
			 intOrSelector = TraceStackOverflow ifTrue:
				[self print: '**StackOverflow**'. ^nil].
			 intOrSelector = TracePrimitiveFailure ifTrue:
				[self print: '**PrimitiveFailure**'. ^nil].
			 intOrSelector = TracePrimitiveRetry ifTrue:
				[self print: '**PrimitiveRetry**'. ^nil].
			 self print: '???']
		ifFalse:
			[intOrSelector = 0
				ifTrue: [self printNum: i; print: '!!!!!!']
				ifFalse: [objectMemory safePrintStringOf: intOrSelector]]! !

!CoInterpreter methodsFor: 'debug printing' stamp: 'eem 10/11/2011 15:26'!
printSends
	<inline: true>
	^cogit printOnTrace! !

!CoInterpreter methodsFor: 'stack bytecodes' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
pushClosureNumArgs: numArgs copiedValues: numCopied blockSize: blockSize
	"The compiler has pushed the values to be copied, if any.
	 Create a Closure with space for the copiedValues and pop numCopied values off the stack into the closure.
	 Set numArgs as specified, and set startpc to the pc following the block size and jump over that code.

	Override only to add debug tracing as of 4/26/2009"
	<inline: true>
	| newClosure context |
	"No need to record the pushed copied values in the outerContext."
	context := self ensureFrameIsMarried: localFP SP: localSP + (numCopied * objectMemory bytesPerOop).
	newClosure := self
					closureIn: context
					numArgs: numArgs
					instructionPointer: (self oopForPointer: localIP) + 2 - (method+objectMemory baseHeaderSize)
					numCopiedValues: numCopied.
	cogit recordSendTrace ifTrue:
		[self recordTrace: TraceBlockCreation thing: newClosure source: TraceIsFromInterpreter].
	numCopied > 0 ifTrue:
		[0 to: numCopied - 1 do:
			[:i|
			"Assume: have just allocated a new BlockClosure; it must be young.
			 Thus, can use unchecked stores."
			 objectMemory storePointerUnchecked: i + ClosureFirstCopiedValueIndex
				ofObject: newClosure
				withValue: (self internalStackValue: numCopied - i - 1)].
		 self internalPop: numCopied].
	localIP := localIP + blockSize.
	self fetchNextBytecode.
	self internalPush: newClosure! !

!CoInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 4/8/2016 09:25'!
pushFullClosureNumArgs: numArgs copiedValues: numCopiedArg compiledBlock: compiledBlock receiverIsOnStack: receiverIsOnStack ignoreContext: ignoreContext
	"The compiler has pushed the values to be copied, if any. The receiver has been pushed on stack before if specified. 
	 Create a Closure with space for the copiedValues and pop numCopied values off the stack into the closure.
	 Sets outerContext, compiledBlock, numArgs and receiver as specified.."
	<inline: true>
	| numCopied newClosure context startIndex |
	"No need to record the pushed copied values in the outerContext."
	context := ignoreContext
		ifTrue: [objectMemory nilObject ]
		ifFalse: [self ensureFrameIsMarried: localFP SP: localSP + (numCopiedArg * objectMemory bytesPerOop)].
	newClosure := self
					fullClosureIn: context 
					numArgs: numArgs 
					numCopiedValues: numCopiedArg 
					compiledBlock: compiledBlock.
	cogit recordSendTrace ifTrue:
		[self recordTrace: TraceBlockCreation thing: newClosure source: TraceIsFromInterpreter].
	receiverIsOnStack
		ifFalse: 
			[ startIndex := FullClosureFirstCopiedValueIndex.
			   objectMemory storePointerUnchecked: FullClosureReceiverIndex
				ofObject: newClosure
				withValue: self receiver.
			numCopied := numCopiedArg ]
		ifTrue:
			[ startIndex := FullClosureReceiverIndex.
			numCopied := numCopiedArg + 1 ].
	numCopied > 0 ifTrue:
		[0 to: numCopied - 1 do:
			[ :i |
			"Assume: have just allocated a new BlockClosure; it must be young.
			 Thus, can use unchecked stores."
			 objectMemory storePointerUnchecked: i + startIndex
				ofObject: newClosure
				withValue: (self internalStackValue: numCopied - i - 1)].
		 self internalPop: numCopied].
	self fetchNextBytecode.
	self internalPush: newClosure! !

!CoInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 5/29/2018 09:08'!
pushRemoteTemp: index inVectorAt: tempVectorIndex
	"Override to use itemporary:in:put:"
	| tempVector |
	tempVector := self itemporary: tempVectorIndex in: localFP.
	TempVectReadBarrier
		ifTrue: 
			[(objectMemory isForwarded: tempVector) ifTrue:
				[tempVector := self unfollowTempVector: tempVector atIndex: tempVectorIndex in: localFP]].
	self internalPush: (objectMemory fetchPointer: index ofObject: tempVector)! !

!CoInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 1/20/2009 05:47'!
pushTemporaryVariable: temporaryIndex
	"Override to use itemporary:in:put:"
	self internalPush: (self itemporary: temporaryIndex in: localFP)! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
quickPrimitiveConstantFor: aQuickPrimitiveIndex
	<api>
	^aQuickPrimitiveIndex caseOf: {
		[257] -> [objectMemory trueObject].
		[258] -> [objectMemory falseObject].
		[259] -> [objectMemory nilObject].
		[260] -> [ConstMinusOne].
		[261] -> [ConstZero].
		[262] -> [ConstOne].
		[263] -> [ConstTwo] }! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 12/13/2015 12:57'!
quickPrimitiveGeneratorFor: aQuickPrimitiveIndex
	<api>
	<returnTypeC: 'sqInt (*quickPrimitiveGeneratorFor(sqInt aQuickPrimitiveIndex))(void)'>
	^aQuickPrimitiveIndex
		caseOf: {
			[256] -> [#genQuickReturnSelf].
			[257] -> [#genQuickReturnConst].
			[258] -> [#genQuickReturnConst].
			[259] -> [#genQuickReturnConst].
			[260] -> [#genQuickReturnConst].
			[261] -> [#genQuickReturnConst].
			[262] -> [#genQuickReturnConst].
			[263] -> [#genQuickReturnConst] }
		otherwise: [#genQuickReturnInstVar]! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 2/12/2009 21:20'!
quickPrimitiveInstVarIndexFor: primIndex
	<api>
	^primIndex - 264! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
rawHeaderOf: methodPointer
	<api>
	^objectMemory fetchPointer: HeaderIndex ofObject: methodPointer! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem 2/20/2013 15:33'!
rawHeaderOf: methodOop put: cogMethodOrMethodHeader
	<api>
	"Since methods may be updated while forwarding during become, make the assert accomodate this."
	self assert: (objectMemory isCompiledMethodHeader: (objectMemory headerWhileForwardingOf: methodOop)).
	objectMemory
		storePointerUnchecked: HeaderIndex
		ofObject: methodOop
		withValue: cogMethodOrMethodHeader! !

!CoInterpreter methodsFor: 'image save/restore' stamp: 'eem 7/1/2021 14:59'!
readImageFromFile: f HeapSize: desiredHeapSize StartingAt: imageOffset
	"Read an image from the given file stream, allocating an amount of memory to its object heap.
	
	 V3: desiredHeapSize is the total size of the heap.  Fail if the image has an unknown format or
	 requires more than the specified amount of memory.

	 Spur: desiredHeapSize is ignored; this routine will attempt to provide at least extraVMMemory's
	 ammount of free space after the image is loaded, taking any free space in the image into account.
	 extraVMMemory is stored in the image header and is accessible as vmParameterAt: 23.  If
	 extraVMMemory is 0, the value defaults to the default grow headroom.  Fail if the image has an
	 unknown format or if sufficient memory cannot be allocated.

	 Details: This method detects when the image was stored on a machine with the opposite byte
	 ordering from this machine and swaps the bytes automatically. Furthermore, it allows the header
	 information to start 512 bytes into the file, since some file transfer programs for the Macintosh
	 apparently prepend a Mac-specific header of this size. Note that this same 512 bytes of prefix
	 area could also be used to store an exec command on Unix systems, allowing one to launch
	 Smalltalk by invoking the image name as a command."

	| swapBytes headerStart headerSize headerFlags dataSize bytesRead bytesToShift heapSize
	  oldBaseAddr minimumMemory allocationReserve cogCodeBase
	  firstSegSize hdrNumStackPages hdrEdenBytes hdrCogCodeSize hdrMaxExtSemTabSize |
	<var: #f type: #sqImageFile>
	<var: #heapSize type: #usqInt>
	<var: #dataSize type: #'size_t'>
	<var: #minimumMemory type: #usqInt>
	<var: #desiredHeapSize type: #usqInt>
	<var: #allocationReserve type: #usqInt>
	<var: #headerStart type: #squeakFileOffsetType>
	<var: #imageOffset type: #squeakFileOffsetType>

	metaclassNumSlots := 6.	"guess Metaclass instSize"
	classNameIndex := 6.		"guess (Class instVarIndexFor: 'name' ifAbsent: []) - 1"
	swapBytes := self checkImageVersionFrom: f startingAt: imageOffset.
	headerStart := (self sqImageFilePosition: f) - 4.  "record header start position"

	headerSize			:= self getWord32FromFile: f swap: swapBytes.
	dataSize			:= self getLongFromFile: f swap: swapBytes.
	oldBaseAddr		:= self getLongFromFile: f swap: swapBytes.
	objectMemory specialObjectsOop: (self getLongFromFile: f swap: swapBytes).
	objectMemory lastHash: (self getLongFromFile: f swap: swapBytes). "N.B.  not used."
	savedWindowSize	:= self getLongFromFile: f swap: swapBytes.
	headerFlags		:= self getLongFromFile: f swap: swapBytes.
	self setImageHeaderFlagsFrom: headerFlags.
	extraVMMemory	:= self getWord32FromFile: f swap: swapBytes. "N.B.  ignored in V3."
	hdrNumStackPages	:= self getShortFromFile: f swap: swapBytes.
	"4 stack pages is small.  Should be able to run with as few as
	 three. 4 should be comfortable but slow.  8 is a reasonable
	 default.  Can be changed via vmParameterAt: 43 put: n.
	 Can be set as a preference (Info.plist, VM.ini, command line etc).
	 If desiredNumStackPages is already non-zero then it has been
	 set as a preference.  Ignore (but preserve) the header's default."
	numStackPages := desiredNumStackPages ~= 0
						ifTrue: [desiredNumStackPages]
						ifFalse: [hdrNumStackPages = 0
									ifTrue: [self defaultNumStackPages]
									ifFalse: [hdrNumStackPages]].
	desiredNumStackPages := hdrNumStackPages.
	"This slot holds the size of the native method zone in 1k units. (pad to word boundary)."
	hdrCogCodeSize := (self getShortFromFile: f swap: swapBytes) * 1024.
	cogCodeSize := desiredCogCodeSize ~= 0
						ifTrue: [desiredCogCodeSize]
						ifFalse:
							[hdrCogCodeSize = 0
									ifTrue: [cogit defaultCogCodeSize]
									ifFalse: [desiredCogCodeSize := hdrCogCodeSize]]. "set for vmParameter 47"
	cogCodeSize > cogit maxCogCodeSize ifTrue:
		[cogCodeSize := cogit maxCogCodeSize].
	hdrEdenBytes		:= self getWord32FromFile: f swap: swapBytes.
	objectMemory edenBytes: (desiredEdenBytes ~= 0
						ifTrue: [desiredEdenBytes]
						ifFalse:
							[hdrEdenBytes = 0
									ifTrue: [objectMemory defaultEdenBytes]
									ifFalse: [hdrEdenBytes]]).
	desiredEdenBytes := hdrEdenBytes.
	hdrMaxExtSemTabSize := self getShortFromFile: f swap: swapBytes.
	hdrMaxExtSemTabSize ~= 0 ifTrue:
		[self setMaxExtSemSizeTo: hdrMaxExtSemTabSize].
	"pad to word boundary.  This slot can be used for anything else that will fit in 16 bits.
	 Preserve it to be polite to other VMs."
	the2ndUnknownShort	:= self getShortFromFile: f swap: swapBytes.
	firstSegSize := self getLongFromFile: f swap: swapBytes.
	objectMemory firstSegmentSize: firstSegSize.

	"compare memory requirements with availability"
	allocationReserve := self interpreterAllocationReserveBytes.
	minimumMemory := cogCodeSize "no need to include the stackZone; this is alloca'ed"
						+ dataSize
						+ objectMemory newSpaceBytes
						+ allocationReserve.
	"Compute how much space is needed for the initial heap allocation.
	 no need to include the stackZone; this is alloca'ed.
	 no need to include the JIT code zone size; this is allocated separately."
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[| freeOldSpaceInImage headroom |
			 freeOldSpaceInImage := self getLongFromFile: f swap: swapBytes.
			 headroom := objectMemory
							initialHeadroom: extraVMMemory
							givenFreeOldSpaceInImage: freeOldSpaceInImage.
			 heapSize := objectMemory roundUpHeapSize:
						  dataSize
						+ headroom
						+ objectMemory newSpaceBytes
						+ (headroom > allocationReserve
							ifTrue: [0]
							ifFalse: [allocationReserve])]
		ifFalse:
			[heapSize :=  desiredHeapSize
						+ objectMemory newSpaceBytes
						+ (desiredHeapSize - dataSize > allocationReserve
							ifTrue: [0]
							ifFalse: [allocationReserve]).
			 heapSize < minimumMemory ifTrue:
				[self insufficientMemorySpecifiedError]].

	"allocateJITMemory will assign the actual size allocated, which is rounded up to a page boundary."
	cogCodeBase := self allocateJITMemory: (self addressOf: cogCodeSize).

	"allocate a contiguous block of memory for the Squeak heap and ancilliary data structures"
	(self
			allocateMemory: heapSize
			minimum: minimumMemory
			imageFile: f
			headerSize: headerSize) asUnsignedInteger
		ifNil: [self insufficientMemoryAvailableError]
		ifNotNil:
			[:mem| "cannot clash with the variable memory still in use in NewCoObjectMemory and superclasses"
			objectMemory
				setHeapBase: (heapBase := mem)
				memoryLimit: mem + heapSize
				endOfMemory: mem + dataSize].

	"position file after the header"
	self sqImageFile: f Seek: headerStart + headerSize.

	"read in the image in bulk, then swap the bytes if necessary"
	bytesRead := objectMemory readHeapFromImageFile: f dataBytes: dataSize.
	bytesRead ~= dataSize ifTrue: [self unableToReadImageError].

	self ensureImageFormatIsUpToDate: swapBytes.

	"compute difference between old and new memory base addresses"
	bytesToShift := objectMemory memoryBaseForImageRead - oldBaseAddr.
	self initializeInterpreter: bytesToShift.  "adjusts all oops to new location"
	cogit initializeCodeZoneFrom: cogCodeBase upTo: cogCodeBase + cogCodeSize.
	^dataSize! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 12/28/2008 10:27'!
receiver
	<inline: true>
	^stackPages longAt: localFP + FoxIFReceiver! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 6/20/2021 19:37'!
recordContextSwitchFrom: aProcess in: sourceCode
	(cogit recordEventTrace and: [primTracePluginName isNil]) ifTrue:
		[self recordTrace: TraceContextSwitch thing: aProcess source: sourceCode]! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem 8/3/2021 19:19'!
recordFastCCallPrimTraceForMethod: aMethodObj
	"This is a little elaborate.  The primTraceLog is only useful if it is not full of noise.
	 To reduce noise when debugging a specific plugin we allow a plugin name to be
	 specified and will only generate the primTraceLog code for primitives in that plugin."
	<api>
	^cogit recordFastCCallPrimTrace
	  and: [primTracePluginName
				ifNil: [true]
				ifNotNil: [self methodHasPrimitiveInPrimTracePlugin: aMethodObj]]! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem 8/3/2021 19:19'!
recordPrimTraceForMethod: aMethodObj
	"This is a little elaborate.  The primTraceLog is only useful if it is not full of noise.
	 To reduce noise when debugging a specific plugin we allow a plugin name to be
	 specified and will only generate the primTraceLog code for primitives in that plugin."
	<api>
	^cogit recordPrimTrace
	  and: [primTracePluginName
				ifNil: [true]
				ifNotNil: [self methodHasPrimitiveInPrimTracePlugin: aMethodObj]]! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 4/2/2009 11:42'!
recordTrace: classOrInteger thing: selector source: source
	traceLog at: traceLogIndex put: classOrInteger.
	traceLog at: traceLogIndex + 1 put: selector.
	traceLog at: traceLogIndex + 2 put: source.
	traceLogIndex := traceLogIndex + 3 \\ TraceBufferSize! !

!CoInterpreter methodsFor: 'debug support' stamp: 'nice 11/6/2016 09:57'!
reportMinimumUnusedHeadroom
	"Report the stack page size and minimum unused headroom to stdout."
	<api>
	self cCode:
			[self pri: 'stack page bytes %lld available headroom %lld minimum unused headroom %lld\n'
				n: self stackPageByteSize asUnsignedLongLong
				t: (self stackPageByteSize - self stackLimitBytes - self stackLimitOffset) asUnsignedLongLong
				f: self minimumUnusedHeadroom asUnsignedLongLong]
		inSmalltalk:
			["CogVMSimulator new initStackPagesForTests reportMinimumUnusedHeadroom"
			 self print: 'stack page bytes '; printNum: self stackPageByteSize;
				print: ' available headroom '; printNum: self stackPageByteSize - self stackLimitBytes - self stackLimitOffset;
				print: ' minimum unused headroom '; printNum: self minimumUnusedHeadroom;
				cr]! !

!CoInterpreter methodsFor: 'callback support' stamp: 'eem 3/29/2020 18:45'!
restoreCStackStateForCallbackContext: vmCallbackContext
	<var: #vmCallbackContext type: #'VMCallbackContext *'>
	<inline: #always>
	self setCFramePointer: vmCallbackContext savedCFramePointer setCStackPointer: vmCallbackContext savedCStackPointer! !

!CoInterpreter methodsFor: 'process primitive support' stamp: 'eem 11/15/2012 11:10'!
resume: aProcess
	"Replaced by resume:preemptedYieldingIf:from:"
	"Make aProcess runnable and if its priority is higher than
	 that of the current process, preempt the current process.
	 Answer if the current process was preempted.  Override
	 to add tracing info (see resume:from:)."
	<doNotGenerate>
	self shouldNotImplement! !

!CoInterpreter methodsFor: 'process primitive support' stamp: 'eem 11/15/2012 11:10'!
resume: aProcess preemptedYieldingIf: yieldImplicitly
	"Replaced by resume:preemptedYieldingIf:from:"
	"Make aProcess runnable and if its priority is higher than  that of the
	 current process, preempt the current process.   Answer if the current
	 process was preempted.  If the current process was preempted then if
	 yieldImplicitly add the current process to the back of its run queue,
	 causing an implicit yiled to other processes on the run queue,  otherwise
	 add the current process to the front of its run queue, hence not yielding.
	 Blue book behaviour is to yield implicitly but is arguably incorrect."
	<doNotGenerate>
	self shouldNotImplement! !

!CoInterpreter methodsFor: 'process primitive support' stamp: 'eem 11/4/2010 18:38'!
resume: aProcess preemptedYieldingIf: yieldImplicitly from: sourceCode
	"Make aProcess runnable and if its priority is higher than  that of the
	 current process, preempt the current process.   Answer if the current
	 process was preempted.  If the current process was preempted then if
	 yieldImplicitly add the current process to the back of its run queue,
	 causing an implicit yeild to other processes on the run queue,  otherwise
	 add the current process to the front of its run queue, hence not yielding.
	 Blue book behaviour is to yield implicitly but is arguably incorrect.
	 Override to add tracing info."
	| activeProc activePriority newPriority |
	<inline: false>
	activeProc := self activeProcess.
	activePriority := self quickFetchInteger: PriorityIndex ofObject: activeProc.
	newPriority := self quickFetchInteger: PriorityIndex ofObject: aProcess.
	newPriority <= activePriority ifTrue:
		[self putToSleep: aProcess yieldingIf: true.
		 ^false].
	self putToSleep: activeProc yieldingIf: yieldImplicitly.
	self transferTo: aProcess from: sourceCode.
	^true! !

!CoInterpreter methodsFor: 'enilopmarts' stamp: 'eem 3/29/2020 17:25'!
return: returnValue toExecutive: inInterpreter
	"We have made a context switch, either when interpreting or from machine code.
	 Effectively return to the current frame, either by entering machine code, or
	 longjmp-ing back to the interpreter or simply returning, depending on where we are."

	cogit assertCStackWellAligned.
	(self isMachineCodeFrame: framePointer) ifTrue:
		[self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer imbar: false line: #'__LINE__'.
		 self push: instructionPointer.
		 self push: returnValue.
		 cogit ceEnterCogCodePopReceiverReg
		 "NOTREACHED"].
	self push: returnValue.
	self setMethod: (self iframeMethod: framePointer).
	self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer imbar: true line: #'__LINE__'.
	instructionPointer = cogit ceReturnToInterpreterPC ifTrue:
		[instructionPointer := self iframeSavedIP: framePointer].
	inInterpreter ifTrue:
		[^nil].
	cogit ceInvokeInterpret.
	"NOTREACHED"
	^nil! !

!CoInterpreter methodsFor: 'enilopmarts' stamp: 'eem 11/14/2020 21:31'!
returnToExecutive: inInterpreter postContextSwitch: switchedContext
	"Return to the current frame, either by entering machine code, or longjmp-ing back to the
	 interpreter or simply returning, depending on where we are. To know whether to return or
	 enter machine code we have to know from whence we came.  We could have come from
	 the interpreter, either directly or via a machine code primitive.  We could have come from
	 machine code.  The instructionPointer tells us where from.  If it is above startOfMemory we're
	 in the interpreter.  If it is below, then we are in machine-code unless it is ceReturnToInterpreterPC,
	 in which case we're in a machine-code primitive called from the interpreter."
	<inline: false>
	| cogMethod retValue fullyInInterpreter |
	<var: #cogMethod type: #'CogBlockMethod *'>

	cogit assertCStackWellAligned.
	(self isMachineCodeFrame: framePointer) ifTrue:
		[self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer imbar: false line: #'__LINE__'.
		 "If returning after a context switch then a result may have to be popped from the stack.
		  If the process is suspended at a send then the result of the primitive in which the
		  process was suspended is still on the stack and must be popped into ReceiverResultReg.
		  If not, nothing should be popped and ReceiverResultReg gets the receiver."
		 switchedContext
			ifTrue:
				[cogMethod := self mframeCogMethod: framePointer.
				self assert: (instructionPointer > cogit minCogMethodAddress 
							and: [instructionPointer < cogit maxCogMethodAddress]).
				 (instructionPointer ~= (cogMethod asInteger + cogMethod stackCheckOffset)
				  and: [cogit isSendReturnPC: instructionPointer])
					ifTrue:
						[self assert: (objectMemory addressCouldBeOop: self stackTop).
						 retValue := self popStack]
					ifFalse:
						[retValue := self mframeReceiver: framePointer]]
			ifFalse: [retValue := self mframeReceiver: framePointer].
		 self push: instructionPointer.
		 self push: retValue.
		 cogit ceEnterCogCodePopReceiverReg
		 "NOTREACHED"].
	self setMethod: (self iframeMethod: framePointer).
	fullyInInterpreter := inInterpreter.
	instructionPointer = cogit ceReturnToInterpreterPC ifTrue:
		[instructionPointer := self iframeSavedIP: framePointer.
		 fullyInInterpreter := false].
	self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer imbar: true line: #'__LINE__'.
	fullyInInterpreter ifFalse:
		[cogit ceInvokeInterpret
		 "NOTREACHED"].
	^nil! !

!CoInterpreter methodsFor: 'return bytecodes' stamp: 'eem 7/16/2013 14:10'!
returnToMachineCodeFrame
	"Return to the previous context/frame after assigning localIP, localSP and localFP."
	<inline: true>
	cogit assertCStackWellAligned.
	self assert: localIP asUnsignedInteger < objectMemory startOfMemory.
	self assert: (self isMachineCodeFrame: localFP).
	self assertValidExecutionPointe: localIP asUnsignedInteger r: localFP s: localSP imbar: false line: #'__LINE__'.
	self internalStackTopPut: localIP.
	self internalPush: localReturnValue.
	self externalizeFPandSP.
	self cCode: '' inSmalltalk:
		[self maybeCheckStackDepth: 1 sp: stackPointer pc: localIP].
	cogit ceEnterCogCodePopReceiverReg
	"NOTREACHED"! !

!CoInterpreter methodsFor: 'primitive support' stamp: 'eem 08/08/2014 10:42'!
roomToPushNArgs: n
	"Answer if there is room to push n arguments onto the current stack.  We assume
	 this is called by primitives that check there is enough room in any new context, and
	 won't actually push the arguments in the current context if the primitive fails.  With
	 this assumption it is safe to answer based on the maximum argument count, /not/
	 the ammount of space in the current frame were it converted to a context.."
	false
		ifTrue: "old code that checked size of context..."
			[| methodHeader cntxSize |
			(self isMachineCodeFrame: framePointer)
				ifTrue: [methodHeader := (self mframeHomeMethod: framePointer) methodHeader]
				ifFalse: [methodHeader := objectMemory methodHeaderOf: (self iframeMethod: framePointer)].
			cntxSize := (self methodHeaderIndicatesLargeFrame: methodHeader)
							ifTrue: [LargeContextSlots - CtxtTempFrameStart]
							ifFalse: [SmallContextSlots - CtxtTempFrameStart].
			^self stackPointerIndex + n <= cntxSize]
		ifFalse: "simpler code that simply insists args are <= max arg count"
			[^n <= (LargeContextSlots - CtxtTempFrameStart)]! !

!CoInterpreter methodsFor: 'callback support' stamp: 'eem 3/29/2020 19:02'!
saveCStackStateForCallbackContext: vmCallbackContext
	<var: #vmCallbackContext type: #'VMCallbackContext *'>
	<inline: #always>
	vmCallbackContext
		savedCStackPointer: CStackPointer asVoidPointer;
		savedCFramePointer: CFramePointer asVoidPointer! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
scavengeThreshold
	<doNotGenerate>
	^objectMemory scavengeThreshold! !

!CoInterpreter methodsFor: 'callback support' stamp: 'eem 4/25/2020 18:40'!
sendInvokeCallback: thunkPtr Stack: stackPtr Registers: regsPtr Jmpbuf: jmpBufPtr
	"Override to log and check stack alignment.  Since this is an implicit send we need to
	 log it explicitly. The return side is done via a primitive so that gets logged normally."
	cogit assertCStackWellAligned.
	cogit recordPrimTrace ifTrue:
		[self fastLogPrim: (objectMemory splObj: SelectorInvokeCallback)].
	^super sendInvokeCallback: thunkPtr Stack: stackPtr Registers: regsPtr Jmpbuf: jmpBufPtr! !

!CoInterpreter methodsFor: 'callback support' stamp: 'eem 4/25/2020 18:40'!
sendInvokeCallbackContext: vmCallbackContext
	"Override to log and check stack alignment.  Since this is an implicit send we need to
	 log it explicitly. The return side is done via a primitive so that gets logged normally."
	cogit assertCStackWellAligned.
	cogit recordPrimTrace ifTrue:
		[self fastLogPrim: (objectMemory splObj: SelectorInvokeCallback)].
	^super sendInvokeCallbackContext: vmCallbackContext! !

!CoInterpreter methodsFor: 'simulation' stamp: 'eem 2/1/2020 15:57'!
setCFramePointer: cfp
	"Set the CFramePointer, either on initialization (see initializeProcessorStack:) or from
	 machine code in ceCaptureCStackPointers.  Since the CFramePointer should not change,
	 (except during callbacks, which we don't yet simulate) unless we are initializing,
	 ceCaptureCFramePointers should yield the same value every time."
	<doNotGenerate>
	CFramePointer ifNotNil: [self assert: CFramePointer = cfp].
	^CFramePointer := cfp! !

!CoInterpreter methodsFor: 'callback support' stamp: 'eem 11/24/2020 22:50'!
setCFramePointer: cFramePointer setCStackPointer: cStackPointer
	<inline: #always>
	self deny: (cFramePointer isNil or: [cStackPointer isNil]).
	CStackPointer := cStackPointer.
	CFramePointer := cFramePointer! !

!CoInterpreter methodsFor: 'simulation' stamp: 'eem 2/1/2020 15:58'!
setCStackPointer: csp
	"Set the CStackPointer, either on initialization (see initializeProcessorStack:) or from
	 machine code in ceCaptureCStackPointers.  Since the CStackPointer should not change,
	 (except during callbacks, which we don't yet simulate) unless we are initializing,
	 ceCaptureCStackPointers should yield the same value every time."
	<doNotGenerate>
	CStackPointer ifNotNil: [self assert: CStackPointer = csp].
	^CStackPointer := csp! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 3/16/2017 14:52'!
setCogCodeZoneThreshold: threshold
	<doNotGenerate>
	<var: 'threshold' type: #double>
	^cogit setCogCodeZoneThreshold: threshold! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 1/10/2017 11:12'!
setDesiredCogCodeSize: dccs
	<inline: true>
	desiredCogCodeSize := dccs! !

!CoInterpreter methodsFor: 'object memory support' stamp: 'cb 5/28/2018 15:39'!
setGCMode: mode
	gcMode := mode! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 6/21/2012 12:37'!
setIFrameHasContext: theFP
	"See encodeFrameFieldHasContext:numArgs:"
	<inline: true>
	<var: #theFP type: #'char *'>
	stackPages byteAt: theFP + FoxIFrameFlags + 2 put: 1! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 11/8/2020 22:53'!
setImageHeaderFlags: flags
	"Set an array of flags indicating various properties of the saved image, responded to on image load.
	 These are the same as the image header flags shifted right two bits, omitting the fullScreenFlag and float byte order flag.
	 Bit 0: if set, implies the image's Process class has threadId as its 3rd inst var (zero relative) (meaningful to the MT VM only)
	 Bit 1: if set, methods that are interpreted will have the flag bit set in their header
	 Bit 2: if set, implies preempting a process does not put it to the back of its run queue
	 Bit 3: if set, implies a threaded VM will not disown the VM if owned by the GUI thread (meaningful to the MT VM only)
	 Bit 4: if set, implies the new finalization scheme where WeakArrays are queued
	 Bit 5: if set, implies wheel events will be delivered as such and not mapped to arrow key events
	 Bit 6: if set, implies arithmetic primitives will fail if given arguments of different types (float vs int)
	 Bit 7: if set, implies file primitives (FilePlugin, FileAttributesPlugin) will answer file times in UTC not local times"
	flags asUnsignedInteger > 255 ifTrue:
		[^self primitiveFailFor: PrimErrUnsupported].
	"processHasThreadId := flags anyMask: 1. specific to CoInterpreterMT"
	imageHeaderFlags := (flags anyMask: 1)
							ifTrue: [imageHeaderFlags bitOr: 4]
							ifFalse: [imageHeaderFlags bitClear: 4].
	flagInterpretedMethods := flags anyMask: 2.
	preemptionYields := flags noMask: 4.
	"noThreadingOfGUIThread := flags anyMask: 8.. specific to CoInterpreterMT"
	imageHeaderFlags := (flags anyMask: 8)
							ifTrue: [imageHeaderFlags bitOr: 32]
							ifFalse: [imageHeaderFlags bitClear: 32].
	newFinalization := flags anyMask: 16.
	sendWheelEvents := flags anyMask: 32.
	primitiveDoMixedArithmetic := flags noMask: 64.
	imageHeaderFlags := (flags anyMask: 128)
							ifTrue: [imageHeaderFlags bitOr: 512]
							ifFalse: [imageHeaderFlags bitClear: 512]! !

!CoInterpreter methodsFor: 'image save/restore' stamp: 'eem 9/6/2019 14:19'!
setImageHeaderFlagsFrom: headerFlags
	"Set the flags that are contained in the 7th long of the image header."
	imageHeaderFlags := headerFlags. "so as to preserve unrecognised flags."
	fullScreenFlag := headerFlags bitAnd: 1.
	imageFloatsBigEndian := (headerFlags noMask: 2) ifTrue: [1] ifFalse: [0].
	"processHasThreadId := headerFlags anyMask: 4. specific to CoInterpreterMT"
	flagInterpretedMethods := headerFlags anyMask: 8.
	preemptionYields := headerFlags noMask: 16.
	"noThreadingOfGUIThread := headerFlags anyMask: 32. specific to CoInterpreterMT"
	newFinalization := headerFlags anyMask: 64.
	sendWheelEvents := headerFlags anyMask: 128.
	primitiveDoMixedArithmetic := headerFlags noMask: 256! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 2/19/2012 10:06'!
setMethod: aMethodObj
	self assert: aMethodObj asUnsignedInteger >= objectMemory startOfMemory.
	super setMethod: aMethodObj! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 3/9/2016 17:48'!
setUpForUseByFacade: aCurrentImageCoInterpreterFacade
	"Set up variables with default values so that other initializations work.
	 numStackPages needs to be initialized so that interpreterAllocationReserveBytes
	 can be computed."
	<doNotGenerate>
	numStackPages := 0! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'rsf 8/10/2016 00:12'!
shadowCallStackAddress
	<option: #LowcodeVM>
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: shadowCallStackPointer) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #shadowCallStackPointer in: self]! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:40'!
shadowCallStackPointerIn: theFP
	"See StackInterpreter class>>initializeFrameIndices"
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	^(self isMachineCodeFrame: theFP)
		ifTrue:
			[stackPages pointerAtPointer: theFP + FoxMFReceiver - (BytesPerWord*5)]
		ifFalse:
			[stackPages pointerAtPointer: theFP + FoxIFReceiver - (BytesPerWord*5)]! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'rsf 8/9/2016 23:40'!
shadowCallStackPointerIn: theFP put: valueOopPointer
	"See StackInterpreter class>>initializeFrameIndices"
	<option: #LowcodeVM>
	<returnTypeC: #'char*'>
	<var: #theFP type: #'char *'>
	<var: #valueOopPointer type: #'char *'>
	^(self isMachineCodeFrame: theFP)
		ifTrue:
			[stackPages pointerAtPointer: theFP + FoxMFReceiver - (BytesPerWord*5) put: valueOopPointer]
		ifFalse:
			[stackPages pointerAtPointer: theFP + FoxIFReceiver - (BytesPerWord*5) put: valueOopPointer]! !

!CoInterpreter methodsFor: 'debug printing' stamp: 'eem 11/25/2009 13:20'!
shortPrintFrame: theFP
	<inline: false>
	<var: #theFP type: #'char *'>
	| rcvr mthd |
	(stackPages couldBeFramePointer: theFP) ifFalse:
		[self print: 'invalid frame pointer'; cr.
		 ^nil].
	rcvr := self frameReceiver: theFP.
	mthd := self frameMethodObject: theFP.
	self printHexPtr: theFP.
	self space.
	self printChar: ((self isMachineCodeFrame: theFP) ifTrue: [$M] ifFalse: [$I]).
	self space.
	self printActivationNameFor: mthd
		receiver: rcvr
		isBlock: (self frameIsBlockActivation: theFP)
		firstTemporary: (self temporary: 0 in: theFP).
	self space.
	self shortPrintOop: rcvr "shortPrintOop: adds a cr"! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'RonieSalgado 4/29/2016 15:19'!
signed32BitIntegerFor: integerValue
	<api>
	^super signed32BitIntegerFor: integerValue! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'RonieSalgado 4/29/2016 15:19'!
signed32BitValueOf: oop
	<api>
	^super signed32BitValueOf: oop! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'RonieSalgado 4/29/2016 15:19'!
signed64BitIntegerFor: integerValue
	<api>
	^super signed64BitIntegerFor: integerValue! !

!CoInterpreter methodsFor: 'trampolines' stamp: 'RonieSalgado 4/29/2016 15:19'!
signed64BitValueOf: oop
	<api>
	^super signed64BitValueOf: oop! !

!CoInterpreter methodsFor: 'primitive support' stamp: 'eem 6/20/2021 16:43'!
slowPrimitiveResponse
	"Invoke a normal (non-quick) primitive.
	 Called under the assumption that primFunctionPointer has been preloaded.
	 Override to log primitive."
	(self recordPrimTraceForMethod: newMethod) ifTrue:
		[self fastLogPrim: messageSelector].
	^super slowPrimitiveResponse! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
specialSelectorNumArgs: index "<SmallInteger>"
	<api>
	^objectMemory integerValueOf: (objectMemory fetchPointer: (index * 2) + 1
							ofObject: (objectMemory splObj: SpecialSelectors))! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'eem 2/2/2020 18:59'!
stackLimitAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: stackLimit) asUnsignedInteger]
		  inSmalltalk: [cogit simulatedVariableAddress: #stackLimitFromMachineCode in: self]! !

!CoInterpreter methodsFor: 'stack pages' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
stackLimitOffset
	"Answer the amount of slots needed to fit a new frame at the point the stack
	 limit is checked.  A frame looks like this at the point the stack limit is checked:
			stacked receiver/closure
			arg0
			...
			argN
			caller's method ip/base frame's sender context
	fp->	saved fp
			method
			context (uninitialized?)
			method header fields (interpreter only)
			saved method ip (uninitialized?; interpreter only)
			receiver
			first temp
			...
	sp->	Nth temp
	So the amount of headroom is
		the maximum number of arguments + 1 (for stacked receiver and arguments)
		+ the frame size
		+ the max number of temps.
	 Since a method's number of temps includes its arguments the actual offset is:"
	^(IFrameSlots + 64) * objectMemory wordSize! !

!CoInterpreter methodsFor: 'stack pages' stamp: 'eem 1/29/2013 11:53'!
stackPageHeadroom
	"Return a minimum amount of headroom for each stack page (in bytes).
	 In the interpreter we don't actually need any headroom.  In a JIT the stack
	 has to have room for interrupt handlers which will run on the stack.
	 Defer to the platform for this one."
	<inline: true>
	^self osCogStackPageHeadroom! !

!CoInterpreter methodsFor: 'initialization' stamp: 'eem 12/29/2008 12:11'!
stackPagesClass
	<doNotGenerate>
	^VMBIGENDIAN
		ifTrue: [CoInterpreterStackPagesMSB]
		ifFalse: [CoInterpreterStackPagesLSB]! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 1/4/2009 13:48'!
stackPointer: theSP
	"Simulation only"
	<doNotGenerate>
	stackPointer := theSP! !

!CoInterpreter methodsFor: 'trampoline support' stamp: 'eem 5/6/2010 16:00'!
stackPointerAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: stackPointer) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #stackPointer in: self]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:03'!
stackPointerIndexForFrame: theFP WithSP: theSP
	"Return the 1-based index rel to the given frame"
	"In the StackInterpreter stacks grow down."
	^(self isMachineCodeFrame: theFP)
		ifTrue: [(((theFP + FoxMFReceiver) - theSP) >> objectMemory shiftForWord) + (self mframeNumArgs: theFP)]
		ifFalse: [(((theFP + FoxIFReceiver) - theSP) >> objectMemory shiftForWord) + (self iframeNumArgs: theFP)]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:03'!
stackPointerIndexForIFrame: theFP WithSP: theSP numArgs: numArgs
	"Return the 1-based index rel to the given frame"
	"In the StackInterpreter stacks grow down."
	^(((theFP + FoxIFReceiver) - theSP) >> objectMemory shiftForWord) + numArgs! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:03'!
stackPointerIndexForMFrame: theFP WithSP: theSP numArgs: numArgs
	"Return the 1-based index rel to the given machine code frame"
	"In the StackInterpreter stacks grow down."
	^(((theFP + FoxMFReceiver) - theSP) >> objectMemory shiftForWord) + numArgs! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
startPCOfClosure: aBlockClosure
	"Zero-relative version of BlockClosure>>startpc."
	^(objectMemory integerValueOf: (objectMemory fetchPointer: ClosureStartPCIndex ofObject: aBlockClosure)) - 1! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'eem 08/08/2014 10:05'!
startPCOfMethodHeader: aCompiledMethodHeader
	<api>
	"Zero-relative version of CompiledMethod>>startpc."
	^(objectMemory literalCountOfMethodHeader: aCompiledMethodHeader) + LiteralStart * objectMemory bytesPerOop! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'cb 4/5/2016 18:06'!
startPCOrNilOfLiteral: lit in: aMethodObj
	"Answer the startPC of lit if it is a (clean) block in aMethodObj, otherwise answer nil."
	<api>
	| outerContext |
	((objectMemory isNonImmediate: lit)
	 and: [(objectMemory formatOf: lit) = objectMemory indexablePointersFormat
	 and: [(objectMemory numSlotsOf: lit) >= ClosureFirstCopiedValueIndex]]) ifFalse:
		[^nil].
	outerContext := objectMemory fetchPointer: ClosureOuterContextIndex ofObject: lit.
	(objectMemory isContext: outerContext) ifFalse:
		[^nil].
	aMethodObj ~~ (objectMemory fetchPointer: MethodIndex ofObject: outerContext) ifTrue:
		[^nil].
	^self quickFetchInteger: ClosureStartPCIndex ofObject: lit! !

!CoInterpreter methodsFor: 'stack bytecodes' stamp: 'eem 11/13/2012 10:27'!
storeAndPopTemporaryVariableBytecode
	<expandCases>
	self
		cCode: "this bytecode will be expanded so that refs to currentBytecode below will be constant"
			[self fetchNextBytecode.
			 self itemporary: (currentBytecode bitAnd: 7) in: localFP put: self internalStackTop.
			 self internalPop: 1]
		inSmalltalk: "Interpreter version has fetchNextBytecode out of order"
			[self itemporary: (currentBytecode bitAnd: 7) in: localFP put: self internalStackTop.
			 self fetchNextBytecode.
			 self internalPop: 1]! !

!CoInterpreter methodsFor: 'stack bytecodes' stamp: 'cb 5/29/2018 09:08'!
storeRemoteTemp: index inVectorAt: tempVectorIndex
	"Override to use itemporary:in:put:"
	| tempVector |
	tempVector := self itemporary: tempVectorIndex in: localFP.
	TempVectReadBarrier
		ifTrue: 
			[(objectMemory isForwarded: tempVector) ifTrue:
				[tempVector := self unfollowTempVector: tempVector atIndex: tempVectorIndex in: localFP]].
	objectMemory storePointer: index ofObject: tempVector withValue: self internalStackTop! !

!CoInterpreter methodsFor: 'process primitive support' stamp: 'eem 1/17/2015 13:54'!
synchronousSignal: aSemaphore 
	"Signal the given semaphore from within the interpreter.
	 Answer if the current process was preempted.
	 Override to add tracing info."
	| excessSignals |
	<inline: false>
	(self isEmptyList: aSemaphore) ifTrue:
		["no process is waiting on this semaphore"
		 excessSignals := self fetchInteger: ExcessSignalsIndex ofObject: aSemaphore.
		 self storeInteger: ExcessSignalsIndex
			ofObject: aSemaphore
			withValue: excessSignals + 1.
		 ^false].

	objectMemory ensureSemaphoreUnforwardedThroughContext: aSemaphore.

	^self resume: (self removeFirstLinkOfList: aSemaphore)
		preemptedYieldingIf: preemptionYields
		from: CSSignal! !

!CoInterpreter methodsFor: 'return bytecodes' stamp: 'eem 3/10/2017 13:01'!
tearDownAndRebuildFrameForCannotReturnBaseFrameReturnFrom: contextToReturnFrom to: contextToReturnTo returnValue: returnValue
	"Handle the cannot return response for a base frame return to an invalid context.
	 Build a new base frame for the context in the cannot resume state ready for the
	 send of cannotReturn:.

	 Since we have returned from the base frame of the page the context is effectively widowed.
	 But its sender needs to be contextToReturnTo, and its pc needs to be the HasBeenReturnedFromMCPC
	 marker.  So bereave it (as a side-effect of isWidowedContext:), assign contextToReturnTo to
	 sender, and rebuild its frame, which will have the ceCannotResumePC as its pc.  Finally push
	 returnValue and set instructionPointer to ceCannotResumePC in preparation for the send."
	| newPage |
	<inline: false>
	<var: #newPage type: #'StackPage *'>
	self assert: (stackPage ~= 0 and: [stackPage isFree]).
	self isWidowedContext: contextToReturnFrom.
	self assert: (self isMarriedOrWidowedContext: contextToReturnFrom) not.
	objectMemory
		storePointer: SenderIndex ofObject: contextToReturnFrom withValue: contextToReturnTo;
		storePointer: InstructionPointerIndex ofObject: contextToReturnFrom withValue: HasBeenReturnedFromMCPCOop.
	"void the instructionPointer to stop it being incorrectly updated in a code
	 compaction in makeBaseFrameFor:."
	instructionPointer := 0.
	newPage := self makeBaseFrameFor: contextToReturnFrom.
	self assert: stackPage = newPage.
	self setStackPageAndLimit: newPage.
	self setStackPointersFromPage: newPage.
	self assert: self stackTop = cogit ceCannotResumePC.
	"overwrite the ceSendCannotResumePC on the stack.  If ever re-executed
	 the returnValue will be taken from top-of-stack by ceCannotResume."
	self stackTopPut: returnValue.
	"Assign it to instructionPointer as externalCannotReturn:from: pushes it."
	instructionPointer := cogit ceCannotResumePC! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
temporary: offset in: theFP
	"See StackInterpreter class>>initializeFrameIndices"
	| frameNumArgs |
	<inline: true>
	<var: #theFP type: #'char *'>
	^(self isMachineCodeFrame: theFP)
		ifTrue:
			[offset < (frameNumArgs := self mframeNumArgs: theFP)
				ifTrue: [stackPages longAt: theFP + FoxCallerSavedIP + ((frameNumArgs - offset) * objectMemory wordSize)]
				ifFalse: [stackPages longAt: theFP + FoxMFReceiver - objectMemory wordSize + ((frameNumArgs - offset) * objectMemory wordSize)]]
		ifFalse:
			[self itemporary: offset in: theFP]! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 12/7/2012 13:54'!
temporary: offset in: theFP put: valueOop
	<inline: true>
	^(self isMachineCodeFrame: theFP)
		ifTrue: [self mtemporary: offset in: theFP put: valueOop]
		ifFalse: [self itemporary: offset in: theFP put: valueOop]! !

!CoInterpreter methodsFor: 'internal interpreter access' stamp: 'eem 11/10/2014 12:49'!
temporaryLocation: offset in: theFP numArgs: numArgs
	"Answer the pointer to a given temporary (for debug frame printing in odd circumstances)"
	<var: #theFP type: #'char *'>
	<returnTypeC: #'char *'>
	^offset < numArgs
		ifTrue: [theFP + FoxCallerSavedIP + ((numArgs - offset) * objectMemory wordSize)]
		ifFalse: [theFP
			+ ((self isMachineCodeFrame: theFP)
					ifTrue: [FoxMFReceiver - objectMemory wordSize]
					ifFalse: [FoxIFReceiver - objectMemory wordSize])
			+ ((numArgs - offset) * objectMemory wordSize)]! !

!CoInterpreter methodsFor: 'simulation' stamp: 'eem 9/18/2009 11:21'!
threadManager
	<doNotGenerate>
	^nil! !

!CoInterpreter methodsFor: 'simulation' stamp: 'eem 10/12/2009 01:05'!
transcript
	<doNotGenerate>
	^Transcript! !

!CoInterpreter methodsFor: 'process primitive support' stamp: 'eem 4/8/2009 12:08'!
transferTo: newProc
	"replaced by transferTo:from: for better tracing (for debugging)"
	<doNotGenerate>
	self shouldNotImplement! !

!CoInterpreter methodsFor: 'process primitive support' stamp: 'eem 3/2/-4654 00:06'!
transferTo: newProc from: sourceCode
	"Record a process to be awoken on the next interpreter cycle.
	 Reimplement to record the source of the switch for debugging,
	 and to cope with possible code compaction in makeBaseFrameFor:."
	| activeContext sched oldProc |
	<inline: false>
	statProcessSwitch := statProcessSwitch + 1.
	self push: instructionPointer.
	self externalWriteBackHeadFramePointers.
	self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer.
	"ensureMethodIsCogged: in makeBaseFrameFor: in
	 externalSetStackPageAndPointersForSuspendedContextOfProcess:
	 below may do a code compaction. Nil instructionPointer to avoid it getting pushed twice."
	instructionPointer := 0.
	sched := self schedulerPointer.
	oldProc := objectMemory fetchPointer: ActiveProcessIndex ofObject: sched.
	self recordContextSwitchFrom: oldProc in: sourceCode.
	activeContext := self ensureFrameIsMarried: framePointer SP: stackPointer + objectMemory wordSize.
	objectMemory storePointer: SuspendedContextIndex ofObject: oldProc withValue: activeContext.
	objectMemory storePointer: ActiveProcessIndex ofObject: sched withValue: newProc.
	objectMemory storePointerUnchecked: MyListIndex ofObject: newProc withValue: objectMemory nilObject.
	self externalSetStackPageAndPointersForSuspendedContextOfProcess: newProc! !

!CoInterpreter methodsFor: 'compiled methods' stamp: 'cb 5/29/2018 09:08'!
unfollowTempVector: tempVector atIndex: tempVectorIndex in: theFP
	"override for itemporary"
	<option: #TempVectReadBarrier>
	<inline: #never> "So rare it mustn't bulk up the common path"
	| followed |
	followed := objectMemory followForwarded: tempVector.
	self itemporary: tempVectorIndex in: theFP put: followed.
	^followed! !

!CoInterpreter methodsFor: 'image save/restore' stamp: 'eem 4/1/2009 10:45'!
unknownShortOrCodeSizeInKs
	^desiredCogCodeSize + 1023 // 1024! !

!CoInterpreter methodsFor: 'code compaction' stamp: 'eem 11/20/2020 09:10'!
updateStackZoneReferencesToCompiledCodePreCompaction
	<api>
	"Go through all frames in the stack zone and mark their methods
	 so that compaction does not free any methods that are in use."
	<var: 'primCogMethod' type: #'CogMethod *'>

	"If instructionPointer is referring to machine code, as it will be if a primitive is in progress
	 (see isCodeCompactingPrimitiveIndex:) it must be updated if it is referring to a moved
	 method."
	 instructionPointer ~= 0 ifTrue:
		[(cogit cogMethodContaining: instructionPointer) ifNotNil:
			[:primCogMethod|
			 instructionPointer := instructionPointer + primCogMethod objectHeader]].

	0 to: numStackPages - 1 do:
		[:i| | thePage |
		thePage := stackPages stackPageAt: i.
		(stackPages isFree: thePage) ifFalse:
			[self updateStackZoneReferencesToCompiledCodePreCompactionOnPage: thePage]]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 11/20/2020 08:29'!
updateStackZoneReferencesToCompiledCodePreCompactionOnPage: thePage
	<var: #thePage type: #'StackPage *'>
	<inline: true>
	| theFP callerFP theIPPtr theIP theMethodField theFlags theMethod |
	theIPPtr := thePage headSP.
	theFP := thePage  headFP.
	[(self isMachineCodeFrame: theFP) ifTrue:
		[theMethodField := self frameMethodField: theFP.
		 theFlags := theMethodField bitAnd: MFMethodFlagsMask.
		 theMethod := self cCoerceSimple: theMethodField - theFlags to: #'CogMethod *'.
		 theMethod cmType = CMBlock ifTrue:
			[theMethod := (self cCoerceSimple: theMethodField - theFlags to: #'CogBlockMethod *') cmHomeMethod].
		 theIP := (stackPages longAt: theIPPtr) asUnsignedInteger.
		 (theIP ~= cogit ceCannotResumePC
		  and: [self asserta: (theIP >= theMethod asUnsignedInteger
					   and: [theIP < (theMethod asUnsignedInteger + theMethod blockSize)])]) ifTrue:
			[stackPages
				longAt: theIPPtr
				put: theIP + theMethod objectHeader].
		 stackPages
			longAt: theFP + FoxMethod
			put: theMethodField + theMethod objectHeader].
	 (callerFP := self frameCallerFP: theFP) ~= 0] whileTrue:
		[theIPPtr := theFP + FoxCallerSavedIP.
		 theFP := callerFP]! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
updateStateOfSpouseContextForFrame: theFP WithSP: theSP
	"Update the frame's spouse context with the frame's current state except for the
	 sender and instruction pointer, which are used to mark the context as married."
	| theContext tempIndex pointer argsPointer |
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #theSP type: #'char *'>
	<var: #pointer type: #'char *'>
	<var: #argsPointer type: #'char *'>
	self assert: (self frameHasContext: theFP).
	theContext := self frameContext: theFP.
	self assert: (objectMemory isContext: theContext).
	self assert: (self frameReceiver: theFP)
				= (objectMemory noFixupFollowField: ReceiverIndex ofObject: theContext).
	(self isMachineCodeFrame: theFP)
		ifTrue:
			[tempIndex := self mframeNumArgs: theFP.
			 pointer := theFP + FoxMFReceiver - objectMemory wordSize]
		ifFalse:
			[tempIndex := self iframeNumArgs: theFP.
			 pointer := theFP + FoxIFReceiver - objectMemory wordSize].
	"update the arguments. this would appear not to be strictly necessary, but is for two reasons.
	 First, the fact that arguments are read-only is only as convention in the Smalltalk compiler;
	 other languages may choose to modify arguments.
	 Second, the Squeak runUntilErrorOrReturnFrom: nightmare pops the stack top, which may, in
	 certain circumstances, be the last argument, and hence the last argument may not have been
	 stored into the context."
	argsPointer := theFP + (self frameStackedReceiverOffsetNumArgs: tempIndex).
	1 to: tempIndex do:
		[:i|
		argsPointer := argsPointer - objectMemory wordSize.
		self assert: (objectMemory addressCouldBeOop: (stackPages longAt: argsPointer)).
		 objectMemory storePointer: ReceiverIndex + i
			ofObject: theContext
			withValue: (stackPages longAt: argsPointer)].
	"now update the non-argument stack contents."
	[pointer >= theSP] whileTrue:
		[self assert: (objectMemory addressCouldBeOop: (stackPages longAt: pointer)).
		 tempIndex := tempIndex + 1.
		 objectMemory storePointer: ReceiverIndex + tempIndex
			ofObject: theContext
			withValue: (stackPages longAt: pointer).
		 pointer := pointer - objectMemory wordSize].
	self assert: ReceiverIndex + tempIndex < (objectMemory lengthOf: theContext).
	objectMemory storePointerUnchecked: StackPointerIndex
		ofObject: theContext
		withValue: (objectMemory integerObjectOf: tempIndex)! !

!CoInterpreter methodsFor: 'interpreter shell' stamp: 'eem 10/29/2020 10:09'!
useCogitBreakBlockIfNone
	"This is a hook for the simulator; null in production"
	<inline: #always>! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 7/25/2021 11:48'!
validExternalStackPointers
	<doNotGenerate>
	^framePointer < stackPage baseAddress
	 and: [stackPointer < framePointer
	 and: [framePointer > stackPointer
	 and: [stackPointer >= (stackPage realStackLimit - self stackLimitOffset)]]]! !

!CoInterpreter methodsFor: 'debug support' stamp: 'eem 7/18/2015 07:36'!
validInstructionPointer: instrPointer inMethod: aMethod framePointer: fp
	<var: #instrPointer type: #usqInt>
	<var: #aMethod type: #usqInt>
	<var: #fp type: #'char *'>
	| theInstrPointer cogMethod |
	<var: #theInstrPointer type: #usqInt>
	<var: #cogMethod type: #'CogMethod *'>
	instrPointer = cogit ceCannotResumePC ifTrue:
		[^self isMachineCodeFrame: fp].
	instrPointer = cogit ceReturnToInterpreterPC
		ifTrue:
			[(self isMachineCodeFrame: fp) ifTrue:
				[^false].
			 theInstrPointer := self iframeSavedIP: fp]
		ifFalse:
			[theInstrPointer := instrPointer.
			 NewspeakVM
				ifTrue:
					[(self isMachineCodeFrame: fp) ifTrue:
						[cogMethod := self mframeHomeMethod: fp.
						 ^theInstrPointer >= (cogMethod asUnsignedInteger + (cogit sizeof: CogMethod))
						   and: [theInstrPointer < (cogMethod asUnsignedInteger + cogMethod blockSize)]]]
				ifFalse:
					[| header |
					 header := self rawHeaderOf: aMethod.
					 ((self isCogMethodReference: header)
					   and: [theInstrPointer < objectMemory startOfMemory]) ifTrue:
					 	[cogMethod := self cCoerceSimple: header to: #'CogMethod *'.
					 	 ^theInstrPointer >= (header + (cogit sizeof: CogMethod))
					 	 and: [theInstrPointer < (header + cogMethod blockSize)]]]].
	^super validInstructionPointer: theInstrPointer inMethod: aMethod framePointer: fp! !

!CoInterpreter methodsFor: 'stack pages' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
validStackPageBaseFrame: aPage
	"Check that the base frame in the stack page has a valid sender and saved context."
	<var: #aPage type: #'StackPage *'>
	<inline: false>
	| savedThisContext senderContextOrNil |
	senderContextOrNil := stackPages longAt: aPage baseAddress.
	savedThisContext := stackPages longAt: aPage baseAddress - objectMemory wordSize.
	(self asserta: aPage baseFP + (self frameStackedReceiverOffset: aPage baseFP) + (2 * objectMemory wordSize) = aPage baseAddress) ifFalse:
		[^false].
	(self asserta: (objectMemory addressCouldBeObj: senderContextOrNil)) ifFalse:
		[^false].
	(self asserta: (objectMemory addressCouldBeObj: savedThisContext)) ifFalse:
		[^false].
	(self asserta: (senderContextOrNil = objectMemory nilObject or: [objectMemory isContext: senderContextOrNil])) ifFalse:
		[^false].
	(self asserta: (objectMemory isContext: savedThisContext)) ifFalse:
		[^false].
	(self asserta: (self frameCallerContext: aPage baseFP) = senderContextOrNil) ifFalse:
		[^false].
	(self asserta: (self frameContext: aPage baseFP) = savedThisContext) ifFalse:
		[^false].
	^true! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 12/11/2013 10:04'!
voidVMStateForSnapshotFlushingExternalPrimitivesIf: flushExtPrims
	"Make sure that all VM state that affects the heap contents is voided so that the heap is
	 ready to be snapshotted.  If flushExtPrims is true, flush references to external
	 primitives in methods.  Answer the activeContext that should be stored in the snapshot."
	<inline: false>
	| activeContext |
	instructionPointer := 0. "in case of code compactions."
	activeContext := super voidVMStateForSnapshotFlushingExternalPrimitivesIf: flushExtPrims.
	cogit voidCogCompiledCode.
	^activeContext! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 9/7/2020 13:34'!
warning: aString
	<api: 'extern void warning(const char *s)'>
	<doNotGenerate>
	self transcript cr; nextPutAll: aString; flush! !

!CoInterpreter methodsFor: 'cog jit support' stamp: 'eem 9/7/2020 13:36'!
warning: aString at: lineNumber
	<api: 'extern void warningat(const char *s, int l)'>
	<doNotGenerate>
	self transcript cr; nextPutAll: aString; space; print: lineNumber; flush! !

!CoInterpreter methodsFor: 'debug printing' stamp: 'eem 3/19/2014 16:31'!
whereIs: anOop
	<var: 'somewhere' type: #'char *'>
	(cogit whereIsMaybeCodeThing: anOop) ifNotNil: [:somewhere| ^somewhere].
	^super whereIs: anOop! !

!CoInterpreter methodsFor: 'frame access' stamp: 'eem 4/27/2016 08:10'!
widowOrForceToBytecodePC: ctxt
	"Either widow the context or map its pc to a bytecode one.
	 Used to implement primitiveVoidVMStateForMethod."
	<inline: #never> "for debugging & saving space"
	(self isMarriedOrWidowedContext: ctxt)
		ifTrue:
			"Since any machine-code frame activations of the method have been divorced
			 there should only be interpreted activations of marriecd contexts."
			[(self isWidowedContext: ctxt) ifFalse:
				[self deny: (self isMachineCodeFrame: (self frameOfMarriedContext: ctxt))]]
		ifFalse:
			[self ensureContextHasBytecodePC: ctxt]! !

CoInterpreter subclass: #CoInterpreterPrimitives
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

CoInterpreterPrimitives subclass: #CoInterpreterMT
	instanceVariableNames: 'cogThreadManager checkThreadActivation maxWaitingPriority foreignCallbackPriority deferThreadSwitch disowningVMThread disownCount foreignCallbackProcessSlot willNotThreadWarnCount activeProcessAffined relinquishing processHasThreadId noThreadingOfGUIThread reenterThreadSchedulingLoop'
	classVariableNames: 'DisownFlagsShift DisownVMForProcessorRelinquish LockGUIThreadFlag LockGUIThreadShift OwnVMForeignThreadFlag PrimNumberRelinquishProcessor ProcessUnaffinedOnDisown ReturnToThreadSchedulingLoop VMAlreadyOwnedHenceDoNotDisown'
	poolDictionaries: 'VMThreadingConstants'
	category: 'VMMaker-Multithreading'!

!CoInterpreterMT class methodsFor: 'translation' stamp: 'eem 6/7/2018 15:34'!
ancilliaryClasses
	"Answer any extra classes to be included in the translation."
	^super ancilliaryClasses, { CogThreadManager. CogVMThread }! !

!CoInterpreterMT class methodsFor: 'translation' stamp: 'eem 9/4/2009 10:45'!
apiExportHeaderName
	^'cointerpmt.h'! !

!CoInterpreterMT class methodsFor: 'translation' stamp: 'eem 7/8/2021 11:09'!
clusteredVariableNames
	"Insist that these variables are present early in the list of variables, and in this order,
	 so that e.g. they are conveniently accessed via the VarBaseReg if it is available."
	^#(vmOwner), super clusteredVariableNames "reject: [:n| self allInstVarNames includes: n]"! !

!CoInterpreterMT class methodsFor: 'translation' stamp: 'eem 8/9/2021 12:44'!
declareCVarsIn: aCCodeGenerator
	self class == thisContext methodClass ifFalse: [^self]. "Don't duplicate decls in subclasses"
	aCCodeGenerator
		addHeaderFile:'"sqAtomicOps.h"'. "For THRLOG"
	aCCodeGenerator vmClass
		declareInterpreterVersionIn: aCCodeGenerator
		defaultName: 'Cog MT'.
	aCCodeGenerator
		var: #disowningVMThread type: #'CogVMThread *'.
	aCCodeGenerator var: #reenterThreadSchedulingLoop type: 'jmp_buf'.! !

!CoInterpreterMT class methodsFor: 'initialization' stamp: 'eem 8/24/2018 16:19'!
initializeMiscConstants

	super initializeMiscConstants.

	"N.B. some of these DisownFlags are replicated in platforms/Cross/vm/sqVirtualMachine.h.
	 Hence they should always be initialized."
	DisownVMForProcessorRelinquish := 64.

	(InitializationOptions at: #COGMTVM ifAbsent: [false]) == false ifTrue:
		[^self].

	COGMTVM := true.

	ReturnToThreadSchedulingLoop := 2 "setjmp/longjmp code."! !

!CoInterpreterMT class methodsFor: 'initialization' stamp: 'eem 11/19/2020 17:13'!
initializePrimitiveTable
	super initializePrimitiveTable.
	PrimNumberRelinquishProcessor := 230.
	COGMTVM ifTrue:
		[(227 to: 229) do:
			[:pidx| self assert: (PrimitiveTable at: pidx + 1) = #primitiveFail].
		PrimitiveTable
			at: 227 + 1 put: #primitiveVMCurrentThreadId;
			at: 228 + 1 put: #primitiveProcessBoundThreadId;
			at: 229 + 1 put: #primitiveProcessBindToThreadId]! !

!CoInterpreterMT class methodsFor: 'initialization' stamp: 'eem 11/3/2010 12:08'!
initializeSchedulerIndices
	super initializeSchedulerIndices.
	"Class Process"
	ThreadIdIndex := 4.

	"disown result/own argument flags & max number of threads"
	LockGUIThreadShift := 16.
	LockGUIThreadFlag := 1 << LockGUIThreadShift.
	OwnVMForeignThreadFlag := 1 << (LockGUIThreadShift + 1).
	VMAlreadyOwnedHenceDoNotDisown := 1 << (LockGUIThreadShift + 2).
	ProcessUnaffinedOnDisown := 1 << (LockGUIThreadShift + 3).
	DisownFlagsShift := LockGUIThreadShift + 4.
	ThreadIdMask := (1 << LockGUIThreadShift) - 1 "results in 64k thread indices"
! !

!CoInterpreterMT class methodsFor: 'accessing' stamp: 'EstebanLorenzano 10/23/2014 14:21'!
interpreterVersion 
	^ 'Cog MT'! !

!CoInterpreterMT class methodsFor: 'translation' stamp: 'eem 7/16/2009 13:43'!
isNonArgumentImplicitReceiverVariableName: aString
	^'cogThreadManager' = aString
	   or: [super isNonArgumentImplicitReceiverVariableName: aString]! !

!CoInterpreterMT class methodsFor: 'testing' stamp: 'eem 8/27/2009 11:20'!
isThreadedVM
	^true! !

!CoInterpreterMT class methodsFor: 'translation' stamp: 'eem 9/28/2010 22:59'!
mustBeGlobal: var
	"Make disownCount global so that debugging/instrumentation code can use it to check
	 if a threaded FFI call is in progress (i.e. this isn't necessary for production)"

	^(super mustBeGlobal: var)
	   or: [ #('disownCount' 'willNotThreadWarnCount') includes: var]! !

!CoInterpreterMT class methodsFor: 'accessing class hierarchy' stamp: 'eem 9/6/2014 11:27'!
primitivesClass
	^CoInterpreterMT! !

!CoInterpreterMT class methodsFor: 'translation' stamp: 'eem 9/4/2009 10:45'!
sourceFileName
	"Answer the filename for the core interpreter"

	^'cointerpmt.c'! !

!CoInterpreterMT methodsFor: 'simulation' stamp: 'eem 10/30/2020 07:42'!
_longjmp: aJumpBuf _: returnValue
	"Hack simulation of _setjmp/_longjmp, intended to invoke the most minimal setjmp/longjmp
	 pair available on the platform; no saving/restoring signal masks, no stack unwinding, etc.
	 Signal the exception that simulates a longjmp back to the interpreter." 
	<doNotGenerate>
	self assert: aJumpBuf == reenterThreadSchedulingLoop.
	aJumpBuf returnValue: returnValue; signal! !

!CoInterpreterMT methodsFor: 'debug support' stamp: 'eem 10/31/2020 18:04'!
assertCStackAssignedToCurrentThread
	"void in production; see the simulator subclass..."
	<inline: #always>! !

!CoInterpreterMT methodsFor: 'simulation' stamp: 'eem 11/22/2020 21:02'!
assertCStackPointersBelongToCurrentThread
	<cmacro: '(ignored) 0'> "simulation only"
	| ownerIndex range |
	self assert: (ownerIndex := cogThreadManager getVMOwner) > 0.
	range := self cStackRangeForThreadIndex: ownerIndex.
	self assert: ((range includes: CFramePointer) and: [range includes: CStackPointer])! !

!CoInterpreterMT methodsFor: 'debug support' stamp: 'eem 11/9/2020 19:23'!
assertCStackPointersBelongToDisowningThread
	<cmacro: '() 0'> "simulation only"
	| range |
	self assert: disowningVMThread notNil.
	self assert: ((range := self cStackRangeForThreadIndex: disowningVMThread index) includes: CFramePointer).
	self assert: (range includes: CStackPointer)! !

!CoInterpreterMT methodsFor: 'simulation' stamp: 'eem 11/22/2020 21:03'!
assertProcessorStackPointersBelongToCurrentThread
	<cmacro: '(ignored) 0'> "simulation only"
	| ownerIndex range |
	self assert: (ownerIndex := cogThreadManager getVMOwner) > 0.
	range := self cStackRangeForThreadIndex: ownerIndex.
	self assert: ((range includes: cogit processor fp) and: [range includes: cogit processor sp])! !

!CoInterpreterMT methodsFor: 'debug support' stamp: 'eem 10/30/2020 16:00'!
assertSaneThreadAndProcess
	<inline: true>
	self assert: cogThreadManager vmIsOwned.
	self assert: cogThreadManager currentVMThread state = CTMAssignableOrInVM.
	self assert: (objectMemory fetchPointer: MyListIndex ofObject: self activeProcess) = objectMemory nilObject! !

!CoInterpreterMT methodsFor: 'debug support' stamp: 'eem 11/24/2020 22:33'!
assertValidExternalStackPointers
	<doNotGenerate>
	"For use *ONLY* by routines coming in to the VM,
	 i.e. handleCallOrJumpSimulationTrap:.  This is because it nils localFP as a side-effect,
	 and it does so so that the head frame can be determined reliably."
	super assertValidExternalStackPointers.
	self assertCStackPointersBelongToCurrentThread.
	self assertProcessorStackPointersBelongToCurrentThread! !

!CoInterpreterMT methodsFor: 'simulation' stamp: 'eem 11/20/2020 11:39'!
assertValidNewMethodPropertyFlags
	<cmacro: '() 0'> "simulation only"
	self assert: (objectMemory addressCouldBeObj: newMethod).
	self assert: (objectMemory isCompiledMethod: newMethod).
	self assert: (self primitiveIndexOf: newMethod) > 0.
	self assert: ((self primitivePropertyFlags: (self primitiveIndexOf: newMethod)) anyMask: PrimCallMayEndureCodeCompaction)! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 7/23/2010 19:12'!
bindProcess: aProcess toId: newId
	"Change a Process's thread binding and answer 0, otherwise answer a suitable error code.
	 Cases:
		process is unbound & unaffined
			id 0 nothing to do
			id non-zero ensure thread and bind
		process is affined (temporarily bound to a thread for the duration of a surrender of ownership)
			id = affined index nothing to do
			id = 0 nothing to do
			id ~= 0 && id ~= affined index fail
		process is bound (permanently bound to a thread)
			id = bound index nothing to do
			id ~= bound index set bound index"
	| threadId ownerIndex newThreadIndex |
	processHasThreadId ifFalse:
		[^PrimErrUnsupported].

	threadId := self threadIdFieldOf: aProcess.
	newThreadIndex := self ownerIndexOfThreadId: newId.
	ownerIndex := self ownerIndexOfThreadId: threadId.

	(self isAffinedThreadId: threadId) ifTrue:
		[(ownerIndex = newId
		  or: [newId = 0]) ifTrue:
			[^0].
		^PrimErrInappropriate].

	newThreadIndex > 0 ifTrue:
		[(self startThreadForThreadIndex: newThreadIndex) ifFalse:
			[^PrimErrLimitExceeded]].

	self setOwnerIndexOfProcess: aProcess to: newId bind: true.
	^0! !

!CoInterpreterMT methodsFor: 'simulation' stamp: 'eem 9/30/2009 18:53'!
cStackRangeForCurrentThread
	<doNotGenerate>
	^self cStackRangeForThreadIndex: cogThreadManager ioGetThreadLocalThreadIndex! !

!CoInterpreterMT methodsFor: 'simulation' stamp: 'eem 10/27/2020 17:12'!
cStackRangeForThreadIndex: threadIndex
	"Each simulated processor thread gets 4k of the rump C stack.
	 The top-most section is reserved for in-memory variables such as vmOwnerLock."
	<doNotGenerate>
	| top |
	^(threadIndex between: 1 and: cogThreadManager maxNumThreads) ifTrue:
		[top := self rumpCStackAddress - (threadIndex - 1 * RumpCStackSize).
		 top - RumpCStackSize + 1 to: top]! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 11/5/2010 09:25'!
cedeToHigherPriorityThreads
	"Invoked from checkForEventsMayContextSwitch: to switch threads if a thread
	 wanting to acquire the VM has higher priority than the active process."
	| activeProc ownerIndex activeContext activePriority activeThread vmThread |
	<var: #activeThread type: #'CogVMThread *'>
	<var: #vmThread type: #'CogVMThread *'>
	<inline: false>
	activeProc := self activeProcess.
	activePriority := self quickFetchInteger: PriorityIndex ofObject: activeProc.
	ownerIndex := self ownerIndexOfProcess: activeProc.
	ownerIndex = 0
		ifTrue: [activeThread := cogThreadManager currentVMThread]
		ifFalse: [activeThread := cogThreadManager vmThreadAt: ownerIndex].
	activeThread priority: activePriority.
	vmThread := cogThreadManager
					highestPriorityThreadIfHigherThan: activePriority
					expectedMax: maxWaitingPriority.
	(vmThread isNil					"no waiting thread of sufficiently high priority.  Do not switch."
	or: [vmThread = activeThread]) "The activeProcess needs to run on a different thread.  Leave this to
									  threadSwitchIfNecessary:from: in checkForEventsMayContextSwitch:"
		ifTrue:
			[maxWaitingPriority > activePriority ifTrue:
				[maxWaitingPriority := activePriority].
			 ^self].

	self assert: vmThread priority > activePriority.
	self assert: vmThread ~= cogThreadManager currentVMThread.
	self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer.

	maxWaitingPriority > vmThread priority ifTrue:
		[maxWaitingPriority := vmThread priority].
	statProcessSwitch := statProcessSwitch + 1.
	activeContext := self ensureFrameIsMarried: framePointer SP: stackPointer.
	objectMemory storePointer: SuspendedContextIndex ofObject: activeProc withValue: activeContext.
	self ensurePushedInstructionPointer.
	self externalWriteBackHeadFramePointers.
	self putToSleep: activeProc yieldingIf: preemptionYields.
	"Transcript cr; print: #cedeToHighestPriorityThreadIfHigherThan:; cr.
	  self printExternalHeadFrame.
	  self print: 'ip: '; printHex: self instructionPointer. Transcript cr; flush."
	self returnToSchedulingLoopAndReleaseVMOrWakeThread: vmThread source: CSCheckEvents! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 8/6/2021 12:45'!
checkForEventsMayContextSwitch: mayContextSwitch
	"Check for possible interrupts and handle one if necessary.
	 Answer if a context switch has occurred."
	| switched sema now |
	<inline: false>
	<var: #now type: #usqLong>
	self assertSaneThreadAndProcess.
	cogit assertCStackWellAligned.
	statCheckForEvents := statCheckForEvents + 1.

	"restore the stackLimit if it has been smashed."
	self restoreStackLimit.
	self externalWriteBackHeadFramePointers.
	self assert: stackPage = stackPages mostRecentlyUsedPage.

	"Allow the platform to do anything it needs to do synchronously."
	self ioSynchronousCheckForEvents.

	self checkCogCompiledCodeCompactionCalledFor.

	objectMemory needGCFlag ifTrue:
		["sufficientSpaceAfterGC: runs the incremental GC and
		 then, if not enough space is available, the fullGC."
		 (objectMemory sufficientSpaceAfterGC: 0) ifFalse:
			[self setSignalLowSpaceFlagAndSaveProcess]].

	mayContextSwitch ifFalse: [^false].

	switched := false.
	self assert: deferThreadSwitch not.
	deferThreadSwitch := true.

	(profileProcess ~= objectMemory nilObject
	 or: [nextProfileTick > 0 and:[self ioHighResClock >= nextProfileTick]]) ifTrue:
		[self zeroNextProfileTick.
		 "Take a sample (if not already done so) for the profiler if it is active.  This
		  must be done before any of the synchronousSignals below or else we will
		  attribute a pause in ioRelinquishProcessor to the newly activated process."
		 profileProcess = objectMemory nilObject ifTrue:
			[profileProcess := self activeProcess.
			 profileMethod := objectMemory nilObject].
		 "and signal the profiler semaphore if it is present"
		 (profileSemaphore ~= objectMemory nilObject
		  and: [self synchronousSignal: profileSemaphore]) ifTrue:
			[switched := true]].

	objectMemory signalLowSpace ifTrue:
		[objectMemory signalLowSpace: false. "reset flag"
		 sema := objectMemory splObj: TheLowSpaceSemaphore.
		 (sema ~= objectMemory nilObject
		  and: [self synchronousSignal: sema]) ifTrue:
			[switched := true]].

	"inIOProcessEvents prevents reentrancy into ioProcessEvents and allows disabling
	 ioProcessEvents e.g. for native GUIs.  We would like to manage that here but can't
	 since the platform code may choose to call ioProcessEvents itself in various places."
	false
		ifTrue:
			[((now := self ioUTCMicroseconds) >= nextPollUsecs
			 and: [inIOProcessEvents = 0]) ifTrue:
				[statIOProcessEvents := statIOProcessEvents + 1.
				 inIOProcessEvents := inIOProcessEvents + 1.
				 self ioProcessEvents. "sets interruptPending if interrupt key pressed; may callback"
				 inIOProcessEvents > 0 ifTrue:
					[inIOProcessEvents := inIOProcessEvents - 1].
				 nextPollUsecs := now + 20000
				 "msecs to wait before next call to ioProcessEvents.  Note that strictly
				  speaking we might need to update 'now' at this point since
				  ioProcessEvents could take a very long time on some platforms"]]
		ifFalse:
			[(now := self ioUTCMicroseconds) >= nextPollUsecs ifTrue:
				[statIOProcessEvents := statIOProcessEvents + 1.
				 self ioProcessEvents. "sets interruptPending if interrupt key pressed; may callback"
				 nextPollUsecs := now + 20000
				 "msecs to wait before next call to ioProcessEvents.  Note that strictly
				  speaking we might need to update 'now' at this point since
				  ioProcessEvents could take a very long time on some platforms"]].

	interruptPending ifTrue:
		[interruptPending := false.
		 "reset interrupt flag"
		 sema := objectMemory splObj: TheInterruptSemaphore.
		 (sema ~= objectMemory nilObject
		  and: [self synchronousSignal: sema]) ifTrue:
			[switched := true]].

	nextWakeupUsecs ~= 0 ifTrue:
		[now >= nextWakeupUsecs ifTrue:
			[nextWakeupUsecs := 0.
			 "set timer interrupt to 0 for 'no timer'"
			 sema := objectMemory splObj: TheTimerSemaphore.
			 (sema ~= objectMemory nilObject
			  and: [self synchronousSignal: sema]) ifTrue:
				[switched := true]]].

	"signal any pending finalizations"
	pendingFinalizationSignals > 0 ifTrue:
		[pendingFinalizationSignals := 0.
		 sema := objectMemory splObj: TheFinalizationSemaphore.
		 (sema ~= objectMemory nilObject
		  and: [self synchronousSignal: sema]) ifTrue:
			[switched := true]].

	"signal all semaphores in semaphoresToSignal"
	self signalExternalSemaphores ifTrue:
		[switched := true].

	deferThreadSwitch := false.
	checkThreadActivation ifTrue:
		[checkThreadActivation := false.
		 self cedeToHigherPriorityThreads]. "N.B.  This may not return if we do switch."

	self threadSwitchIfNecessary: self activeProcess from: CSCheckEvents.
	^switched! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 10/30/2020 13:51'!
checkVMOwnershipFromHeartbeat
	"Check whether the VM is unowned and needs to set a thread running to try and own it.
	 Do not attempt this if the image doesn't have a threadId inst var in Process; the VM
	 can't thread these images."
	<inline: false>
	self sqLowLevelMFence.
	(processHasThreadId
	 and: [cogThreadManager vmIsUnowned]) ifTrue:
		[cogThreadManager ensureRunningVMThread: relinquishing]! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 7/17/2013 12:12'!
deferStackLimitSmashAround: functionSymbol
	"Defer smashes of the stackLimit around the call of functionSymbol (for assert checks).	

	 N.B. SYNCHRONIZE WITH forceInterruptCheckFromHeartbeat"

	<var: #functionSymbol declareC: 'void (*functionSymbol)(void)'>
	deferSmash := true.
	self perform: functionSymbol.
	deferSmash := false.
	deferredSmash ifTrue:
		[deferredSmash := false.
		 self forceInterruptCheck.
		 self checkVMOwnershipFromHeartbeat].
	^true "called from assert"! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 7/17/2013 12:10'!
deferStackLimitSmashAround: functionSymbol with: arg
	"Defer smashes of the stackLimit around the call of functionSymbol (for assert checks).	

	 N.B. SYNCHRONIZE WITH forceInterruptCheckFromHeartbeat"

	<var: #functionSymbol declareC: 'void (*functionSymbol)(sqInt)'>
	deferSmash := true.
	self sqLowLevelMFence.
	self perform: functionSymbol with: arg.
	deferSmash := false.
	self sqLowLevelMFence.
	deferredSmash ifTrue:
		[deferredSmash := false.
		 self sqLowLevelMFence.
		 self forceInterruptCheck.
		 self checkVMOwnershipFromHeartbeat].
	^true "called from assert"! !

!CoInterpreterMT methodsFor: 'vm scheduling' stamp: 'eem 11/19/2020 17:07'!
disownVM: flags
	"Release the VM to other threads and answer the current thread's index.
	 Currently valid flags:
		DisownVMForFFICall			- informs the VM that it is entering an FFI call
		DisownVMForThreading		- informs the VM that it is entering code during which threading should be permitted
		OwnVMForeignThreadFlag	- indicates lowest-level entry from a foreign thread
									- not to be used explicitly by clients
									- only set by ownVMFromUnidentifiedThread
		VMAlreadyOwnedHenceDoNotDisown
									- indicates an ownVM from a callback was made when
									  the vm was still owned.
									- not to be used explicitly by clients
									- only set by ownVMFromUnidentifiedThread

	 This is the entry-point for plugins and primitives that wish to release the VM while
	 performing some operation that may potentially block, and for callbacks returning
	 back to some blocking operation.  If this thread does not reclaim the VM before-
	 hand then when the next heartbeat occurs the thread manager will schedule a
	 thread to acquire the VM which may start running the VM in place of this thread.

	 N.B. Most of the state needed to resume after preemption is set in preemptDisowningThread."
	<api>
	<inline: false>
	| vmThread result |
	<var: #vmThread type: #'CogVMThread *'>
	self assert: self successful.
	cogit recordEventTrace ifTrue:
		[self recordTrace: TraceDisownVM thing: (objectMemory integerObjectOf: flags) source: 0].
	processHasThreadId ifFalse:
		[willNotThreadWarnCount < 10 ifTrue:
			[self print: 'warning: VM parameter 48 indicates Process doesn''t have threadId; VM will not thread'; cr.
			 willNotThreadWarnCount := willNotThreadWarnCount + 1]].
	vmThread := cogThreadManager currentVMThread.
	(flags anyMask: VMAlreadyOwnedHenceDoNotDisown) ifTrue:
		[disowningVMThread := vmThread.
		 vmThread state: CTMUnavailable.
		 ^0].
	self assertCStackPointersBelongToCurrentThread.
	self assertValidNewMethodPropertyFlags.
	(flags anyMask: DisownVMForProcessorRelinquish) ifTrue:
		[| proc |
		 (proc := objectMemory splObj: foreignCallbackProcessSlot) ~= objectMemory nilObject ifTrue:
			[foreignCallbackPriority := self quickFetchInteger: PriorityIndex ofObject: proc].
		 relinquishing := true.
		 self sqLowLevelMFence].
	(noThreadingOfGUIThread and: [self inGUIThread]) ifTrue:
		[^vmThread index
		 + LockGUIThreadFlag
		 + (activeProcessAffined ifTrue: [0] ifFalse: [ProcessUnaffinedOnDisown])
		 + (flags << DisownFlagsShift)].
	disownCount := disownCount + 1.
	disowningVMThread := vmThread.
	"self cr; cr; print: 'disownVM  Csp: '; printHex: vmThread cStackPointer; cr.
	(0 to: 16 by: 4) do:
		[:offset|
		self print: ' *(esp+'; printNum: offset; print: ': '; printHex: (stackPages longAt: cogit processor sp + offset); cr].
	cogit processor printIntegerRegistersOn: Transcript."

	"OwnVMForeignThreadFlag indicates lowest-level of entry by a foreign
	 thread. If that's where we are then release the vmThread.  Otherwise
	 indicate the vmThread is off doing something outside of the VM."
	(flags anyMask: OwnVMForeignThreadFlag)
		ifTrue:
			["I don't think this is quite right.  Josh's use case is creating some foreign thread and then registering
			 it with the VM. That's not the same as binding a process to a foreign thread given that the foreign
			 callback process is about to terminate anyway (it is returning from a callback here).  So do we need
			 an additional concept, that of a vmThread being either of the set known to the VM or floating?"
			self flag: 'issue with registering foreign threads with the VM'.
			(self isBoundProcess: self activeProcess) ifFalse:
				[cogThreadManager unregisterVMThread: vmThread]]
		ifFalse: [vmThread state: CTMUnavailable].
	result := vmThread index
				+ (activeProcessAffined ifTrue: [0] ifFalse: [ProcessUnaffinedOnDisown])
				+ (flags << DisownFlagsShift).
	cogThreadManager releaseVM.
	^result! !

!CoInterpreterMT methodsFor: 'initialization' stamp: 'eem 1/2/2021 21:51'!
enterSmalltalkExecutive
	"Main entry-point into the interpreter at each execution level, where an
	 execution level is either the start of execution or reentry for a callback.

	 N.B. It also provides the simulator's implementation of ceReturnToInterpreter/ceInvokeInterpreter, which
	 via a simulation trap raise the ReenterInterpreter signal in handleCallOrJumpSimulationTrap:/reenterInterpreter.
	 So when ReenterInterpreter is caught this method invokes interpret directly."
	<cmacro: '() enterSmalltalkExecutiveImplementation()'>
	"Simulation of the setjmp in enterSmalltalkExecutiveImplementation for reentry into interpreter."
	| caught |
	caught := false.
	[[caught
			ifFalse: [self enterSmalltalkExecutiveImplementation]
			ifTrue: [self interpret]]
		on: ReenterInterpreter
		do: [:ex|
			caught := true.
			self assert: ex returnValue = ReturnToInterpreter.
			ex return: true]] whileTrue! !

!CoInterpreterMT methodsFor: 'initialization' stamp: 'eem 11/22/2020 21:03'!
enterSmalltalkExecutiveImplementation
	"Main entry-point into the interpreter at each execution level, where an execution
	 level is either the start of execution or reentry for a callback.  Capture the C stack
	 pointers so that calls from machine-code into the C run-time occur at this level.
	 This is the actual implementation, separated from enterSmalltalkExecutive so the
	 simulator can wrap it in an exception handler and hence simulate the setjmp/longjmp.

	 Override to return if a longjmp to reenterInterpreter passes a parameter greater than 1.
	 This causes a return to threadSchedulingLoop:startingVM: and is used to surrender
	 control to another thread."
	<inline: false>
	self assertSaneThreadAndProcess.
	self assertProcessorStackPointersBelongToCurrentThread. "At this point CFramePointer/CStackPointer have yet to be assigned".
	^super enterSmalltalkExecutiveImplementation! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 8/4/2021 11:12'!
forceInterruptCheckFromHeartbeat
	"Force an interrupt check ASAP. This version is the
	 entry-point to forceInterruptCheck for the heartbeat
	 timer to allow for repeatable debugging.

	 N.B. SYNCHRONIZE WITH deferStackLimitSmashAround:"
	suppressHeartbeatFlag ifFalse:
		[self sqLowLevelMFence.
		 deferSmash
			ifTrue:
				[deferredSmash := true.
				self sqLowLevelMFence]
			ifFalse:
				[self forceInterruptCheck.
				 self checkVMOwnershipFromHeartbeat]]! !

!CoInterpreterMT methodsFor: 'image save/restore' stamp: 'eem 9/6/2019 14:53'!
getImageHeaderFlags
	"Answer the flags that are contained in the 7th long of the image header."
	^fullScreenFlag "0 or 1"
	+ (VMBIGENDIAN ifTrue: [0] ifFalse: [2]) "this is the imageFloatsLittleEndian flag"
	+ (processHasThreadId ifTrue: [4] ifFalse: [0])
	+ (flagInterpretedMethods ifTrue: [8] ifFalse: [0])
	+ (preemptionYields ifTrue: [0] ifFalse: [16r10])
	+ (noThreadingOfGUIThread ifTrue: [16r20] ifFalse: [0])
	+ (newFinalization ifTrue: [16r40] ifFalse: [0])
	+ (sendWheelEvents ifTrue: [16r80] ifFalse: [0])
	+ (primitiveDoMixedArithmetic ifTrue: [0] ifFalse: [16r100])
	+ (imageHeaderFlags bitClear: 16r1FF) "these are any flags we do not recognize"! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 8/13/2009 16:42'!
getMaxWaitingPriority
	<cmacro: '() GIV(maxWaitingPriority)'>
	^maxWaitingPriority! !

!CoInterpreterMT methodsFor: 'vm scheduling' stamp: 'eem 10/30/2020 13:44'!
inGUIThread
	"The first thread is assumed to be the GUI thread, the VM thread that expects to receive
	 window events, etc.  This might appear to invite race conditions but it is only to be used
	 to decide whether to not give up the VM from the GUI thread (see disownVM:)."
	self assert: noThreadingOfGUIThread.
	^cogThreadManager vmOwnerIs: 1! !

!CoInterpreterMT methodsFor: 'initialization' stamp: 'eem 11/9/2020 19:31'!
initialEnterSmalltalkExecutive
	"Main entry-point into the interpreter at system start-up."
	"Ensure that the myList of the activeProcess is nil.  Needed to load
	 old images which don't nil myList in transferTo:{from:}"
	objectMemory storePointerUnchecked: MyListIndex ofObject: self activeProcess withValue: objectMemory nilObject.
	cogThreadManager startThreadSubsystem.
	self threadSchedulingLoop: (cogThreadManager vmThreadAt: 1).
	"NOTREACHED.  The following is to fool Slang"
	self threadSchedulingLoopImplementation: cogThreadManager currentVMThread! !

!CoInterpreterMT methodsFor: 'initialization' stamp: 'eem 2/2/2018 09:51'!
initialize
	super initialize.
	relinquishing := checkThreadActivation := deferThreadSwitch := false.
	foreignCallbackPriority := maxWaitingPriority := disownCount := willNotThreadWarnCount := 0! !

!CoInterpreterMT methodsFor: 'initialization' stamp: 'eem 10/30/2020 07:39'!
initializeInterpreter: bytesToShift 
	super initializeInterpreter: bytesToShift.
	foreignCallbackProcessSlot := (objectMemory lengthOf: objectMemory specialObjectsOop) > ForeignCallbackProcess
										ifTrue: [ForeignCallbackProcess]
										ifFalse: [NilObject].
	self cCode: '' inSmalltalk:
		[reenterThreadSchedulingLoop := ReenterThreadSchedulingLoop new].
! !

!CoInterpreterMT methodsFor: 'simulation' stamp: 'eem 10/27/2020 16:15'!
initializeProcessorForThreadIndex: threadIndex
	"Each simulated processor thread gets 4k of the rump C stack."
	<doNotGenerate>
	threadIndex > 1 ifTrue:
		[cogit initializeProcessorStack: (self cStackRangeForThreadIndex: threadIndex) last]! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 7/23/2010 17:49'!
isAffinedProcess: aProcess
	^self isAffinedThreadId: (self ownerIndexOfProcess: aProcess)! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 7/23/2010 17:49'!
isAffinedThreadId: threadId
	^(objectMemory isIntegerObject: threadId)
	  and: [((objectMemory integerValueOf: threadId) bitAnd: 1) = 0]! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 8/30/2010 10:02'!
isBoundProcess: aProcess
	^self isBoundThreadId: (self ownerIndexOfProcess: aProcess)! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 7/23/2010 17:48'!
isBoundThreadId: threadId
	^(objectMemory isIntegerObject: threadId)
	  and: [((objectMemory integerValueOf: threadId) bitAnd: 1) = 1]! !

!CoInterpreterMT methodsFor: 'simulation' stamp: 'eem 12/25/2010 14:22'!
isThreadedVM
	<doNotGenerate>
	^true! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 7/23/2010 17:48'!
isUnboundThreadId: threadId
	"Answer if the threadId is neither affined nor bound.  Not the same as bound not."
	^threadId = objectMemory nilObject
	  or: [((objectMemory integerValueOf: threadId) bitAnd: (ThreadIdMask << 1) + 1) = 0]! !

!CoInterpreterMT methodsFor: 'initialization' stamp: 'eem 8/17/2011 13:12'!
loadInitialContext
	| activeProc |
	super loadInitialContext.
	activeProc := self activeProcess.
	self assert: (self ownerIndexOfProcess: activeProc) = 0.
	activeProcessAffined := (self ownerIndexOfProcess: activeProc) ~= 0! !

!CoInterpreterMT methodsFor: 'object memory support' stamp: 'eem 10/30/2020 07:08'!
mapInterpreterOops
	"Map all oops in the interpreter's state to their new values 
	 during garbage collection or a become: operation."
	"Assume: All traced variables contain valid oops."
	<var: #vmThread type: #'CogVMThread *'>
	super mapInterpreterOops.

	"Per-thread state; trace each thread's own newMethod and stack of awol processes."
	1 to: cogThreadManager getNumThreads do:
		[:i| | vmThread |
		vmThread := cogThreadManager vmThreadAt: i.
		vmThread state ifNotNil:
			[(vmThread newMethodOrNull notNil
			 and: [objectMemory shouldRemapOop: vmThread newMethodOrNull]) ifTrue:
				[vmThread newMethodOrNull: (objectMemory remapObj: vmThread newMethodOrNull)].
			 0 to: vmThread awolProcIndex - 1 do:
				[:j|
				(objectMemory shouldRemapOop: (vmThread awolProcesses at: j)) ifTrue:
					[vmThread awolProcesses at: j put: (objectMemory remapObj: (vmThread awolProcesses at: j))]]]]! !

!CoInterpreterMT methodsFor: 'object memory support' stamp: 'eem 10/30/2020 10:31'!
markAndTraceInterpreterOops: fullGCFlag
	"Override to mark the awolProcesses"
	<var: #vmThread type: #'CogVMThread *'>

	super markAndTraceInterpreterOops: fullGCFlag.

	"Per-thread state; trace each thread's own newMethod and stack of awol processes."
	1 to: cogThreadManager getNumThreads do:
		[:i| | vmThread |
		vmThread := cogThreadManager vmThreadAt: i.
		vmThread state ifNotNil:
			[vmThread newMethodOrNull ifNotNil:
				[objectMemory markAndTrace: vmThread newMethodOrNull].
			 0 to: vmThread awolProcIndex - 1 do:
				[:j| objectMemory markAndTrace: (vmThread awolProcesses at: j)]]]! !

!CoInterpreterMT methodsFor: 'vm scheduling' stamp: 'eem 11/24/2020 23:04'!
ownVM: threadIndexAndFlags
	<api>
	<inline: false>
	"This is the entry-point for plugins and primitives that wish to reacquire the VM after having
	 released it via disownVM or callbacks that want to acquire it without knowing their ownership
	 status.  This call will block until the VM is owned by the current thread or an error occurs.
	 The argument should be the value answered by disownVM, or 0 for callbacks that don't know
	 if they have disowned or not.  This is both an optimization to avoid having to query thread-
	 local storage for the current thread's index (since it can easily keep it in some local variable),
	 and a record of when an unbound process becomes affined to a thread for the dynamic
	 extent of some operation.

	 Answer 0 if the current thread is known to the VM (and on return owns the VM).
	 Answer 1 if the current thread is unknown to the VM and takes ownership.
	 Answer -1 if the current thread is unknown to the VM and fails to take ownership."
	| threadIndex flags vmThread |
	threadIndexAndFlags = 0 ifTrue:
		[^self ownVMFromUnidentifiedThread].

	threadIndex := threadIndexAndFlags bitAnd: ThreadIdMask.
	flags := threadIndexAndFlags >> DisownFlagsShift.

	(flags anyMask: DisownVMForProcessorRelinquish) ifTrue:
		["Presumably we have nothing to do; this primitive is typically called from the
		  background process. So we should /not/ try and activate any threads in the
		  pool; they will waste cycles finding there is no runnable process, and will
		  cause a VM abort if no runnable process is found.  But we /do/ want to allow
		  FFI calls that have completed, or callbacks a chance to get into the VM; they
		  do have something to do.  DisownVMForProcessorRelinquish indicates this."
		 relinquishing := false.
		 self sqLowLevelMFence].

	(threadIndexAndFlags anyMask: LockGUIThreadFlag) ifTrue:
		[self assert: (noThreadingOfGUIThread and: [self inGUIThread]).
		 self assert: disowningVMThread isNil.
		 cogit recordEventTrace ifTrue:
			[self recordTrace: TraceOwnVM thing: ConstZero source: 0].
		 ^0].

	vmThread := cogThreadManager acquireVMFor: threadIndex.
	disownCount := disownCount - 1.

	disowningVMThread ifNotNil:
		[vmThread = disowningVMThread ifTrue:
			[self assert: (vmThread cFramePointer isNil
						or: [CFramePointer = vmThread cFramePointer and: [CStackPointer = vmThread cStackPointer]]).
			 self assert: self successful.
			 self assert: (objectMemory fetchPointer: MyListIndex ofObject: self activeProcess) = objectMemory nilObject.
			 disowningVMThread := nil.
			 cogit recordEventTrace ifTrue:
				[self recordTrace: TraceOwnVM thing: ConstOne source: 0].
			 ^0].  "if not preempted we're done."
		self preemptDisowningThread].

	"We've been preempted; we must restore state and update the threadId
	 in our process, and may have to put the active process to sleep."
	self restoreVMStateFor: vmThread threadIndexAndFlags: threadIndexAndFlags.

	cogit recordEventTrace ifTrue:
		[self recordTrace: TraceOwnVM thing: ConstTwo source: 0].
	^threadIndexAndFlags bitAnd: OwnVMForeignThreadFlag! !

!CoInterpreterMT methodsFor: 'vm scheduling' stamp: 'eem 11/19/2020 19:37'!
ownVMFromUnidentifiedThread
	"Attempt to take ownership from a thread that as yet doesn't know its index.
	 This supports callbacks where the callback could originate from any thread.
	
	 Answer 0 if the owning thread is known to the VM.
	 Answer 1 if the owning thread is unknown to the VM and now owns the VM.
	 Answer -1 if the owning thread is unknown to the VM and fails to own the VM.
	 Answer -2 if the owning thread is unknown to the VM and there is no foreign callback process installed."
	| count threadIndex vmThread |
	<var: #vmThread type: #'CogVMThread *'>
	<inline: false>
	(threadIndex := cogThreadManager ioGetThreadLocalThreadIndex) ~= 0 ifTrue:
		[ "this is a callback from a known thread"
		 (cogThreadManager vmOwnerIs: threadIndex) ifTrue: "the VM has not been disowned"
			[self assert: (disowningVMThread isNil or: [disowningVMThread = self currentVMThread]).
			 disowningVMThread := nil.
			 self currentVMThread state: CTMAssignableOrInVM.
			 ^VMAlreadyOwnedHenceDoNotDisown].
		 ^self ownVM: threadIndex].
	foreignCallbackPriority = 0 ifTrue:
		[^-2].
	count := 0.
	"If the current thread doesn't have an index it's new to the vm
	 and we need to allocate a new threadInfo, failing if we can't.
	 We also need a process in the foreignCallbackProcessSlot upon
	 which to run the thread's eventual callback."
	[[cogThreadManager tryLockVMOwnerTo: cogThreadManager ioCurrentOSThread asUnsignedInteger] whileFalse:
		[self waitingPriorityIsAtLeast: foreignCallbackPriority. 
		cogThreadManager ioTransferTimeslice].
	 (objectMemory splObj: foreignCallbackProcessSlot) ~= objectMemory nilObject] whileFalse:
		[cogThreadManager releaseVM.
		 (count := count + 1) > 1000 ifTrue:
			[^-2].
		 cogThreadManager ioMilliSleep: 1].

	vmThread := cogThreadManager unusedThreadInfo.
	"N.B.  Keep the VM locked anonymously so that we reserve the non-nil ForeignCallbackProcess
	 for this thread, avoiding the race between competing foreign callbacks.  The acquireVMFor: in
	 ownVM: will set the vmOwner to the actual index.  So only unlock on failure."
	vmThread ifNil:
		[cogThreadManager releaseVM.
		^-1].
	cogThreadManager setVMOwner: vmThread index.
	vmThread
		state: CTMWantingOwnership;
		priority: foreignCallbackPriority.
	cogThreadManager registerVMThread: vmThread.
	^self ownVM: vmThread index + OwnVMForeignThreadFlag! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 7/23/2010 17:43'!
ownerIndexOfProcess: aProcess
	^self ownerIndexOfThreadId: (self threadIdFieldOf: aProcess)! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 7/23/2010 19:02'!
ownerIndexOfThreadId: threadId
	^(objectMemory isIntegerObject: threadId)
		ifTrue: [(objectMemory integerValueOf: threadId) >> 1 bitAnd: ThreadIdMask]
		ifFalse: [0]! !

!CoInterpreterMT methodsFor: 'vm scheduling' stamp: 'eem 11/30/2020 11:15'!
preemptDisowningThread
	"Set the relevant state for disowningVMThread so that it can resume after
	 being preempted and set disowningVMThread to nil to indicate preemption.

	 N.B.  This should only be sent from checkPreemptionOfDisowningThread.

	 There are essentially four things to do.
	 a)	save the VM's notion of the current C stack pointers; these are pointers
		into a thread's stack and must be saved and restored in thread switch.
	 b)	save the VM's notion of the current Smalltalk execution point.  This is
		simply the suspend half of a process switch that saves the current context
		in the current process.
	 c)	add the process to the thread's set of AWOL processes so that the scheduler
		won't try to run the process while the thread has disowned the VM.
	 d)	save the in-primitive VM state, newMethod and argumentCount

	 ownVM: will restore the VM context as of disownVM: from the above when it
	 finds it has been preempted."

	| activeProc activeContext preemptedThread |
	<var: #preemptedThread type: #'CogVMThread *'>
	<inline: false>
	self assert: disowningVMThread notNil.
	self assert: (disowningVMThread state = CTMUnavailable
				or: [disowningVMThread state = CTMWantingOwnership]).
	self assertCStackPointersBelongToDisowningThread.
	cogit recordEventTrace ifTrue:
		[self recordTrace: TracePreemptDisowningThread
			thing: (objectMemory integerObjectOf: disowningVMThread index)
			source: 0].
	disowningVMThread cStackPointer: CStackPointer.
	disowningVMThread cFramePointer: CFramePointer.
	activeProc := self activeProcess.
	self assert: (objectMemory fetchPointer: MyListIndex ofObject: activeProc) = objectMemory nilObject.
	objectMemory
		storePointer: MyListIndex
		ofObject: activeProc
		withValue: (objectMemory splObj: ProcessInExternalCodeTag).
	activeContext := self ensureFrameIsMarried: framePointer SP: stackPointer.
	objectMemory
		storePointer: SuspendedContextIndex
		ofObject: activeProc
		withValue: activeContext.
	"The instructionPointer must be pushed because the convention for inactive stack pages is that the
	 instructionPointer is top of stack.  We need to know if this primitive is called from machine code
	 because the invariant that the return pc of an interpreter callee calling a machine code caller is
	 ceReturnToInterpreterPC must be maintained."
	self push: instructionPointer.
	self externalWriteBackHeadFramePointers.
	"Since pushing the awol process may realloc disowningVMThread we need to reassign.
	 But since we're going to nil disowningVMThread anyway we can assign to a local."
	preemptedThread := cogThreadManager pushAWOLProcess: activeProc on: disowningVMThread.
	disowningVMThread := nil.
	preemptedThread priority: (self quickFetchInteger: PriorityIndex ofObject: activeProc).
	(self ownerIndexOfProcess: activeProc) = 0
		ifTrue: [self setOwnerIndexOfProcess: activeProc to: preemptedThread index bind: false]
		ifFalse: [self assert: (self ownerIndexOfProcess: activeProc) = preemptedThread index].
	preemptedThread
		newMethodOrNull: newMethod;
		argumentCount: argumentCount;
		primitiveFunctionPointer: primitiveFunctionPointer;
		inMachineCode: instructionPointer asUnsignedInteger <= objectMemory startOfMemory! !

!CoInterpreterMT methodsFor: 'process primitives' stamp: 'eem 10/30/2020 13:40'!
primitiveProcessBindToThreadId
	"Attempt to bind the receiver to the thread with the id of the argument or nil, where the receiver is a Process.
	 If successful the VM will ensure that there are at least id many threads active."
	| aProcess id ec |
	<export: true>
	self cCode: [] inSmalltalk: [cogThreadManager isNil ifTrue: [^self primitiveFail]].
	id := self stackTop.
	aProcess := self stackValue: 1.
	((id = objectMemory nilObject or: [(objectMemory isIntegerObject: id)
										and: [(objectMemory integerValueOf: id) >= 0]])
	and: [(objectMemory isPointers: aProcess)
	and: [(objectMemory slotSizeOf: aProcess) >= (ThreadIdIndex + 1)]]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	id := id = objectMemory nilObject ifTrue: [0] ifFalse: [objectMemory integerValueOf: id].
	id >= cogThreadManager maxNumThreads ifTrue:
		[^self primitiveFailFor: PrimErrLimitExceeded].
	(ec := self bindProcess: aProcess toId: id) ~= 0 ifTrue:
		[^self primitiveFailFor: ec].
	(aProcess = self activeProcess
	and: [(activeProcessAffined := id ~= 0)
	and: [(cogThreadManager vmOwnerIs: id) not]]) ifTrue:
		[(self quickFetchInteger: PriorityIndex ofObject: aProcess) < maxWaitingPriority ifTrue:
			[maxWaitingPriority = self quickFetchInteger: PriorityIndex ofObject: aProcess].
		 checkThreadActivation := true.
		 self forceInterruptCheck].
	self methodReturnReceiver! !

!CoInterpreterMT methodsFor: 'process primitives' stamp: 'eem 9/21/2010 18:41'!
primitiveProcessBoundThreadId
	"Answer the receiver's current thread Id or nil, where the receiver is a Process."
	| aProcess id |
	<export: true>
	self cCode: [] inSmalltalk: [cogThreadManager isNil ifTrue: [^self primitiveFail]].
	aProcess := self stackTop.
	id := self ownerIndexOfProcess: aProcess.
	self pop: argumentCount + 1
		thenPush: (id = 0
						ifTrue: [objectMemory nilObject]
						ifFalse: [objectMemory integerObjectOf: id])! !

!CoInterpreterMT methodsFor: 'cog jit support' stamp: 'eem 11/20/2020 11:39'!
primitivePropertyFlagsForSpur: primIndex
	<inline: true>
	"Answer any special requirements of the given primitive.  Spur always needs to set
	 primitiveFunctionPointer and newMethod so primitives can retry on failure due to forwarders."
	self cCode: [] inSmalltalk: [#(primitiveRelinquishProcessor)]. "For senders..."
	primIndex = PrimNumberRelinquishProcessor ifTrue:
		[^profileSemaphore ~= objectMemory nilObject
			ifTrue: [PrimCallNeedsPrimitiveFunction + PrimCallNeedsNewMethod + PrimCallMayEndureCodeCompaction + PrimCallCollectsProfileSamples]
			ifFalse: [PrimCallNeedsPrimitiveFunction + PrimCallNeedsNewMethod + PrimCallMayEndureCodeCompaction]].
	^super primitivePropertyFlagsForSpur: primIndex! !

!CoInterpreterMT methodsFor: 'I/O primitives' stamp: 'eem 11/26/2020 00:16'!
primitiveRelinquishProcessor
	"Relinquish the processor for up to the given number of microseconds.
	 The exact behavior of this primitive is platform dependent.
	 Override to check for waiting threads."

	| microSecs threadIndexAndFlags currentCStackPointer currentCFramePointer |
	<var: #currentCStackPointer type: #'volatile usqIntptr_t'>
	<var: #currentCFramePointer type: #'volatile usqIntptr_t'>
	microSecs := self stackTop.
	(objectMemory isIntegerObject: microSecs) ifFalse:
		[^self primitiveFail].
	self assert: (objectMemory fetchPointer: MyListIndex ofObject: self activeProcess) = objectMemory nilObject.
	self assert: relinquishing not.
	"DO NOT allow relinquishing the processor while we are profiling since this
	 may skew the time base for our measures (it may reduce processor speed etc).
	 Instead we go full speed, therefore measuring the precise time we spend in the
	 inner idle loop as a busy loop."
	nextProfileTick = 0 ifTrue:
		"Presumably we have nothing to do; this primitive is typically called from the
		 background process. So we should /not/ try and activate any threads in the
		 pool; they will waste cycles finding there is no runnable process, and will
		 cause a VM abort if no runnable process is found.  But we /do/ want to allow
		 FFI calls that have completed, or callbacks a chance to get into the VM; they
		 do have something to do.  DisownVMForProcessorRelinquish indicates this."
		[currentCStackPointer := CStackPointer.
		 currentCFramePointer := CFramePointer.
		 threadIndexAndFlags := self disownVM: DisownVMForProcessorRelinquish.
		 self assert: relinquishing.
		 self ioRelinquishProcessorForMicroseconds: (objectMemory integerValueOf: microSecs).
		 self assert: relinquishing.
		 self ownVM: threadIndexAndFlags.
		 self assert: relinquishing not.
		 self assert: cogThreadManager currentVMThread state = CTMAssignableOrInVM.
		 self assert: currentCStackPointer = CStackPointer.
		 self assert: currentCFramePointer = CFramePointer.
		 "In simulation we allow ioRelinquishProcessorForMicroseconds: to fail so that
		  we can arrange that the simulator responds to input events promptly.  This
		  *DOES NOT HAPPEN* in the real vm."
		 self cCode: [] inSmalltalk: [primFailCode ~= 0 ifTrue: [^self]]].
	self assert: (objectMemory fetchPointer: MyListIndex ofObject: self activeProcess) = objectMemory nilObject.
	self pop: 1  "microSecs; leave rcvr on stack"! !

!CoInterpreterMT methodsFor: 'process primitives' stamp: 'eem 10/30/2020 13:39'!
primitiveVMCurrentThreadId
	<export: true>
	"Answer the VM's current thread's Id"
	self cCode: [] inSmalltalk: [cogThreadManager isNil ifTrue: [^self primitiveFail]].
	self methodReturnInteger: cogThreadManager getVMOwner! !

!CoInterpreterMT methodsFor: 'debug support' stamp: 'eem 11/9/2010 08:16'!
printLogEntryAt: i
	<inline: false>
	| intOrClass selectorOrMethod source |
	intOrClass := traceLog at: i.
	selectorOrMethod := traceLog at: i + 1.
	self printNum: ((traceLog at: i + 2) bitShift: -16); space.
	source := (traceLog at: i + 2) bitAnd: 16rFFFF.
	source <= TraceIsFromInterpreter ifTrue:
		[self print: (traceSources at: source); space].
	(objectMemory isIntegerObject: intOrClass)
		ifTrue:
			[| value |
			value := objectMemory integerValueOf: selectorOrMethod.
			intOrClass = TraceContextSwitch ifTrue:
				[self print: 'context switch'].
			 intOrClass = TraceBlockActivation ifTrue:
				[self print: ' [] in '; printHex: selectorOrMethod].
			 intOrClass = TraceBlockCreation ifTrue:
				[self print: 'create [] '; printHex: selectorOrMethod].
			 intOrClass = TraceIncrementalGC ifTrue:
				[self print: 'incrementalGC'].
			 intOrClass = TraceFullGC ifTrue:
				[self print: 'fullGC'].
			 intOrClass = TraceCodeCompaction ifTrue:
				[self print: 'compactCode'].
			 intOrClass = TraceVMCallback ifTrue:
				[self print: 'callback'].
			 intOrClass = TraceVMCallbackReturn ifTrue:
				[self print: 'return from callback'].
			 intOrClass = TraceThreadSwitch ifTrue:
				[self print: 'thread switch '; printNum: (value bitAnd: 16rFFFF); print: '->'; printNum: (value >> 16)].
			 intOrClass = TracePreemptDisowningThread ifTrue:
				[self print: 'preempt thread '; printNum: value].
			 intOrClass = TraceOwnVM ifTrue:
				[self print: 'ownVM '; printNum: value].
			 intOrClass = TraceDisownVM ifTrue:
				[self print: 'disownVM '; printHex: value]]
		ifFalse:
			[self space; printNameOfClass: intOrClass count: 5; print: '>>'; printStringOf: selectorOrMethod].
	source > TraceIsFromInterpreter ifTrue:
		[self space; print: (traceSources at: source)].
	self cr! !

!CoInterpreterMT methodsFor: 'debug support' stamp: 'eem 11/4/2010 16:38'!
recordThreadSwitchTo: ownerIndex source: sourceCode
	cogit recordEventTrace ifTrue:
		[self recordTrace: TraceThreadSwitch
			thing: (objectMemory integerObjectOf: (ownerIndex << 16) + cogThreadManager getVMOwner)
			source: sourceCode]! !

!CoInterpreterMT methodsFor: 'debug support' stamp: 'eem 11/9/2010 08:13'!
recordTrace: classOrInteger thing: selector source: source
	traceLog at: traceLogIndex put: classOrInteger.
	traceLog at: traceLogIndex + 1 put: selector.
	traceLog at: traceLogIndex + 2 put: source + (cogThreadManager getVMOwner bitShift: 16).
	traceLogIndex := traceLogIndex + 3 \\ TraceBufferSize! !

!CoInterpreterMT methodsFor: 'vm scheduling' stamp: 'eem 11/9/2020 16:53'!
restoreVMStateFor: vmThread threadIndexAndFlags: threadIndexAndFlags
	"We've been preempted; we must restore state and update the threadId
	 in our process, and may have to put the active process to sleep."
	| sched activeProc myProc |
	sched := self schedulerPointer.
	activeProc := objectMemory fetchPointer: ActiveProcessIndex ofObject: sched.
	(threadIndexAndFlags anyMask: OwnVMForeignThreadFlag)
		ifTrue:
			[self assert: foreignCallbackProcessSlot == ForeignCallbackProcess.
			 myProc := objectMemory splObj: foreignCallbackProcessSlot.
			self assert: myProc ~= objectMemory nilObject.
			objectMemory splObj: foreignCallbackProcessSlot put: objectMemory nilObject]
		ifFalse: [myProc := cogThreadManager popAWOLProcess: vmThread].
	self assert: activeProc ~= myProc.
	(activeProc ~= objectMemory nilObject
	 and: [(objectMemory fetchPointer: MyListIndex ofObject: activeProc) = objectMemory nilObject]) ifTrue:
		[self putToSleep: activeProc yieldingIf: preemptionYields].
	self assert: (objectMemory fetchPointer: MyListIndex ofObject: myProc) = (objectMemory splObj: ProcessInExternalCodeTag).
	objectMemory
		storePointer: ActiveProcessIndex ofObject: sched withValue: myProc;
		storePointerUnchecked: MyListIndex ofObject: myProc withValue: objectMemory nilObject.
	"Only unaffine if the process was affined at this level and did not become bound in the interim."
	((threadIndexAndFlags anyMask: ProcessUnaffinedOnDisown)
	 and: [(self isBoundProcess: myProc) not]) ifTrue:
		[self setOwnerIndexOfProcess: myProc to: 0 bind: false].
	self initPrimCall.
	self cCode:
			[self externalSetStackPageAndPointersForSuspendedContextOfProcess: myProc]
		inSmalltalk:
			["Bypass the no-offset stack depth check in the simulator's externalSetStackPageAndPointersForSuspendedContextOfProcess:"
			 super externalSetStackPageAndPointersForSuspendedContextOfProcess: myProc.
			 "We're in ownVM:, hence in a primitive, hence need to include the argument count"
			 (self isMachineCodeFrame: framePointer) ifTrue:
				[self maybeCheckStackDepth: vmThread argumentCount
					sp: stackPointer
					pc: instructionPointer]].
	"If this primitive is called from machine code maintain the invariant that the return pc
	 of an interpreter callee calling a machine code caller is ceReturnToInterpreterPC."
	(vmThread inMachineCode
	 and: [instructionPointer >= objectMemory startOfMemory]) ifTrue:
		[self iframeSavedIP: framePointer put: instructionPointer.
		 instructionPointer := cogit ceReturnToInterpreterPC].
	newMethod := vmThread newMethodOrNull.
	argumentCount := vmThread argumentCount.
	primitiveFunctionPointer := vmThread primitiveFunctionPointer.
	vmThread newMethodOrNull: nil.
	self cCode: '' inSmalltalk:
		[| range |
		 range := self cStackRangeForThreadIndex: vmThread index.
		 self assert: ((range includes: vmThread cStackPointer) and: [range includes: vmThread cFramePointer])].
	self setCFramePointer: vmThread cFramePointer setCStackPointer: vmThread cStackPointer.
	self assert: newMethod notNil
! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 11/20/2020 16:21'!
returnToSchedulingLoopAndReleaseVMOrWakeThread: vmThread source: source
	<var: #vmThread type: #'CogVMThread *'>
	<inline: false>
	self recordThreadSwitchTo: (vmThread ifNotNil: [vmThread index] ifNil: [0]) source: source.
	vmThread
		ifNotNil: [cogThreadManager wakeVMThreadFor: vmThread index]
		ifNil: [cogThreadManager releaseVM].
	"I am not frightened of flying.
	 Any value will do.  I don't mind.
	 Why should I be frightened of flying?
	 There's no reason for it."
	self _longjmp: reenterThreadSchedulingLoop _: 1 ! !

!CoInterpreterMT methodsFor: 'vm scheduling' stamp: 'eem 10/31/2020 19:36'!
saveVMStateFor: vmThread threadIndexAndFlags: threadIndexAndFlags
	"Save the VM state for the disowning thread."

	vmThread
		newMethodOrNull: newMethod;
		argumentCount: argumentCount;
		primitiveFunctionPointer: primitiveFunctionPointer;
		inMachineCode: instructionPointer asUnsignedInteger <= objectMemory startOfMemory! !

!CoInterpreterMT methodsFor: 'callback support' stamp: 'eem 11/6/2010 19:11'!
sendInvokeCallbackContext: vmCallbackContext
	"Override to add sanity assertions."
	self assertSaneThreadAndProcess.
	^super sendInvokeCallbackContext: vmCallbackContext! !

!CoInterpreterMT methodsFor: 'simulation' stamp: 'eem 10/31/2020 17:59'!
setCFramePointer: cfp
	"Set the CFramePointer, either on initialization (see initializeProcessorStack:) or from
	 machine code in ceCaptureCStackPointers.  Unlike the superclass, CFramePointer
	 will be different for each thread using teh VM."
	<doNotGenerate>
	^CFramePointer := cfp! !

!CoInterpreterMT methodsFor: 'simulation' stamp: 'eem 10/31/2020 18:00'!
setCStackPointer: csp
	"Set the CStackPointer, either on initialization (see initializeProcessorStack:) or from
	 machine code in ceCaptureCStackPointers.  Unlike the superclass, CStackPointer
	 will be different for each thread using the VM."
	<doNotGenerate>
	^CStackPointer := csp! !

!CoInterpreterMT methodsFor: 'internal interpreter access' stamp: 'eem 11/8/2020 22:53'!
setImageHeaderFlags: flags
	"Set an array of flags indicating various properties of the saved image, responded to on image load.
	 These are the same as the image header flags shifted right two bits, omitting the fullScreenFlag and float byte order flag.
	 Bit 0: if set, implies the image's Process class has threadId as its 3rd inst var (zero relative)
	 Bit 1: if set, methods that are interpreted will have the flag bit set in their header
	 Bit 2: if set, implies preempting a process does not put it to the back of its run queue
	 Bit 3: if set, implies a threaded VM will not disown the VM if owned by the GUI thread
	 Bit 4: if set, implies the new finalization scheme where WeakArrays are queued
	 Bit 5: if set, implies wheel events will be delivered as such and not mapped to arrow key events
	 Bit 6: if set, implies arithmetic primitives will fail if given arguments of different types (float vs int)
	 Bit 7: if set, implies file primitives (FilePlugin, FileAttributesPlugin) will answer file times in UTC not local times"
	flags asUnsignedInteger > 255 ifTrue:
		[^self primitiveFailFor: PrimErrUnsupported].
	processHasThreadId := flags anyMask: 1.
	flagInterpretedMethods := flags anyMask: 2.
	preemptionYields := flags noMask: 4.
	noThreadingOfGUIThread := flags anyMask: 8.
	newFinalization := flags anyMask: 16.
	sendWheelEvents := flags anyMask: 32.
	primitiveDoMixedArithmetic := flags noMask: 64.
	imageHeaderFlags := (flags anyMask: 128)
							ifTrue: [imageHeaderFlags bitOr: 512]
							ifFalse: [imageHeaderFlags bitClear: 512]! !

!CoInterpreterMT methodsFor: 'image save/restore' stamp: 'eem 9/6/2019 14:19'!
setImageHeaderFlagsFrom: headerFlags
	"Set the flags that are contained in the 7th long of the image header."
	imageHeaderFlags := headerFlags. "so as to preserve unrecognised flags."
	fullScreenFlag := headerFlags bitAnd: 1.
	imageFloatsBigEndian := (headerFlags noMask: 2) ifTrue: [1] ifFalse: [0].
	processHasThreadId := headerFlags anyMask: 4.
	flagInterpretedMethods := headerFlags anyMask: 8.
	preemptionYields := headerFlags noMask: 16.
	noThreadingOfGUIThread := headerFlags anyMask: 32.
	newFinalization := headerFlags anyMask: 64.
	sendWheelEvents := headerFlags anyMask: 128.
	primitiveDoMixedArithmetic := headerFlags noMask: 256.

	processHasThreadId ifFalse:
		[self print: 'warning, processHasThreadId flag is unset; cannot function as a threaded VM if so.'; cr]! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 8/13/2009 18:26'!
setMaxWaitingPriorityTo: minPriority
	maxWaitingPriority := minPriority! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem (objmem refactor) 6/14/2010 10:15'!
setOwnerIndexOfProcess: aProcess to: anIndex bind: bind
	| threadId |
	threadId := anIndex = 0
				ifTrue: [objectMemory nilObject]
				ifFalse: [objectMemory integerObjectOf: (anIndex << 1) + (bind ifTrue: [1] ifFalse: [0])].
	objectMemory storePointerUnchecked: ThreadIdIndex ofObject: aProcess withValue: threadId! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 7/23/2010 19:00'!
threadIdFieldOf: aProcess
	^processHasThreadId
		ifTrue: [objectMemory fetchPointer: ThreadIdIndex ofObject: aProcess]
		ifFalse: [objectMemory nilObject]! !

!CoInterpreterMT methodsFor: 'accessing' stamp: 'eem 9/16/2009 11:10'!
threadManager
	<doNotGenerate>
	^cogThreadManager! !

!CoInterpreterMT methodsFor: 'vm scheduling' stamp: 'eem 11/22/2020 20:52'!
threadSchedulingLoop: vmThread
	"Enter a loop attempting to run the VM with the highest priority process and
	 blocking on the thread's OS semaphore when unable to run that process.
	 This version is for simulation only, simulating the longjmp back to the real
	 threadSchedulingLoopImplementation: through exception handling."

	<cmacro: '(vmThread) threadSchedulingLoopImplementation(vmThread)'>
	cogit initializeProcessorStack: (self cStackRangeForThreadIndex: vmThread index) last.
	[[self threadSchedulingLoopImplementation: vmThread]
		on: ReenterThreadSchedulingLoop
		do: [:ex| self assert: ex returnValue = 1. ex return: true]] whileTrue! !

!CoInterpreterMT methodsFor: 'vm scheduling' stamp: 'eem 11/19/2020 19:40'!
threadSchedulingLoopImplementation: vmThread
	"Enter a loop attempting to run the VM with the highest priority process and
	 blocking on the thread's OS semaphore when unable to run that process.
	 We will return to this via threadSwitchIfNecessary:from: which is called in the
	 middle of transferTo:from: once the active process has been stored in the scheduler."
	<var: #vmThread type: #'CogVMThread *'>
	<inline: false>
	self _setjmp: reenterThreadSchedulingLoop.
	[self assert: vmThread state = CTMAssignableOrInVM.
	 (cogThreadManager tryLockVMOwnerTo: vmThread index) ifTrue:
		["If relinquishing is true, then primitiveRelinquishProcessor has disowned the
		  VM and only a returning call or callback should take ownership in that case."
		 relinquishing
			ifTrue: [cogThreadManager releaseVM]
			ifFalse: [self tryToExecuteSmalltalk: vmThread]].
	 (cogThreadManager vmOwnerIs: vmThread index) ifFalse:
		[cogThreadManager waitForWork: vmThread].
	 true] whileTrue! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 10/30/2020 16:12'!
threadSwitchIfNecessary: newProc from: sourceCode
	"Invoked from transferTo:from: to switch threads if the new process is bound or affined to some other thread."
	| newProcThreadId vmThread activeContext |
	self assert: (cogThreadManager vmOwnerIs: cogThreadManager ioGetThreadLocalThreadIndex).
	deferThreadSwitch ifTrue: [^self].

	"If the current process is unaffined or it is affined to the current thread we're
	 ok to run, but we should yield asap if a higher-priority t6hread wants the VM."
	newProcThreadId := self ownerIndexOfProcess: newProc.
	((activeProcessAffined := newProcThreadId ~= 0)
	 and: [(cogThreadManager vmOwnerIs: newProcThreadId) not]) ifFalse:
		[(self quickFetchInteger: PriorityIndex ofObject: newProc) < maxWaitingPriority ifTrue:
			[checkThreadActivation := true.
			 self forceInterruptCheck].
		^self].

	"The current process is affined to a thread, but not to the current owner.  So switch to that owner."
	self cCode: '' inSmalltalk:
		[self transcript ensureCr; nextPutAll: #threadSwitchIfNecessary:from:; space; print: newProc;
						space; print: cogThreadManager getVMOwner; nextPutAll: '->'; print: newProcThreadId; cr; flush].
	 "If the activeProcess doesn't have a context yet, it needs one from which the new thread can resume execution."
	 (objectMemory fetchPointer: SuspendedContextIndex ofObject: newProc) = objectMemory nilObject ifTrue:
		[self assert: newProc = self activeProcess.
		 self push: instructionPointer.
		 self externalWriteBackHeadFramePointers.
		 activeContext := self ensureFrameIsMarried: framePointer SP: stackPointer.
		 objectMemory storePointer: SuspendedContextIndex ofObject: newProc withValue: activeContext].

	 vmThread := cogThreadManager vmThreadAt: newProcThreadId.
	 vmThread priority: (self quickFetchInteger: PriorityIndex ofObject: newProc).
	 vmThread state = CTMUnavailable ifTrue:
		[vmThread state: CTMWantingOwnership].
	 self returnToSchedulingLoopAndReleaseVMOrWakeThread: vmThread source: CSSwitchIfNeccessary! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 10/30/2020 22:54'!
transferTo: newProc from: sourceCode
	"Record a process to be awoken on the next interpreter cycle.  Override to
	 potentially switch threads either if the new process is bound to another thread,
	 or if there is no runnable process but there is a waiting thread. Note that the
	 abort on no runnable process has beeen moved here from wakeHighestPriority."
	| sched oldProc activeContext |
	<inline: false>
	statProcessSwitch := statProcessSwitch + 1.
	self push: instructionPointer.
	self externalWriteBackHeadFramePointers.
	self assertValidExecutionPointe: instructionPointer r: framePointer s: stackPointer.
	"ensureMethodIsCogged: in makeBaseFrameFor: in
	 externalSetStackPageAndPointersForSuspendedContextOfProcess:
	 below may do a code compaction. Nil instructionPointer to avoid it getting pushed twice."
	instructionPointer := 0.
	sched := self schedulerPointer.
	oldProc := objectMemory fetchPointer: ActiveProcessIndex ofObject: sched.
	self recordContextSwitchFrom: oldProc in: sourceCode.
	activeContext := self ensureFrameIsMarried: framePointer SP: stackPointer + objectMemory wordSize.
	objectMemory storePointer: SuspendedContextIndex ofObject: oldProc withValue: activeContext.

	newProc ifNil:
		["Two possibilities.  One, there is at least one thread waiting to own the VM in which
		  case it should be activated.  Two, there are no processes to run and so abort."
		 self willingVMThread ifNotNil:
			 [:vmThread|
			  vmThread state = CTMWantingOwnership ifTrue:
				[self returnToSchedulingLoopAndReleaseVMOrWakeThread: vmThread source: sourceCode]].
		 self error: 'scheduler could not find a runnable process'].

	objectMemory
		storePointer: ActiveProcessIndex ofObject: sched withValue: newProc;
		storePointerUnchecked: MyListIndex ofObject: newProc withValue: objectMemory nilObject.

	self threadSwitchIfNecessary: newProc from: sourceCode.

	self externalSetStackPageAndPointersForSuspendedContextOfProcess: newProc! !

!CoInterpreterMT methodsFor: 'vm scheduling' stamp: 'eem 11/14/2020 21:27'!
tryToExecuteSmalltalk: vmThread
	"Attempt to run the current process, if it exists, on the given vmThread."
	<var: #vmThread type: #'CogVMThread *'>
	| dvmt activeProc ownerIndex |
	<var: #dvmt type: #'CogVMThread *'>
	self assert: (cogThreadManager vmOwnerIs: vmThread index).
	self assert: cogThreadManager ioGetThreadLocalThreadIndex = vmThread index.
	dvmt := disowningVMThread.
	disowningVMThread
		ifNil: [activeProc := self activeProcess]
		ifNotNil:
			[self preemptDisowningThread.
			 activeProc := self wakeHighestPriority.
			 activeProc
				ifNil: [activeProc := objectMemory nilObject]
				ifNotNil: [objectMemory
							storePointerUnchecked: MyListIndex
							ofObject: activeProc
							withValue: objectMemory nilObject].
			 objectMemory
				storePointer: ActiveProcessIndex
				ofObject: self schedulerPointer
				withValue: activeProc].
	activeProc = objectMemory nilObject ifTrue:
		[cogThreadManager releaseVM.
		 ^nil].
	ownerIndex := self ownerIndexOfProcess: activeProc.
	(ownerIndex = 0 or: [cogThreadManager vmOwnerIs: ownerIndex]) ifTrue:
		[self assert: (objectMemory fetchPointer: MyListIndex ofObject: self activeProcess) = objectMemory nilObject.
		 (objectMemory fetchPointer: SuspendedContextIndex ofObject: activeProc) ~= objectMemory nilObject ifTrue:
			[self externalSetStackPageAndPointersForSuspendedContextOfProcess: activeProc].
		 instructionPointer = cogit ceReturnToInterpreterPC ifTrue:
			[self deny: (self isMachineCodeFrame: framePointer).
			 instructionPointer := self iframeSavedIP: framePointer].
		 self enterSmalltalkExecutive.
		 "When we return here we should have already given up
		  the VM and so we cannot touch any interpreter state."
		"NOTREACHED"].
	cogThreadManager wakeVMThreadFor: ownerIndex! !

!CoInterpreterMT methodsFor: 'simulation' stamp: 'eem 10/28/2020 23:30'!
vmOwnerAddress
	<doNotGenerate>
	^cogThreadManager vmOwnerAddress! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 8/15/2009 23:11'!
waitingPriorityIsAtLeast: minPriority
	"Set the maxWaitingPriority to at least minPriority on behalf
	 of a thread wanting to acquire the VM.  If maxWaitingPriority
	 is increased, schedule a thread activation check asap."
	maxWaitingPriority < minPriority ifTrue:
		[maxWaitingPriority := minPriority.
		 checkThreadActivation := true.
		 self forceInterruptCheck]! !

!CoInterpreterMT methodsFor: 'process primitive support' stamp: 'eem 10/27/2020 19:41'!
wakeHighestPriority
	"Return the highest priority process that is ready to run.
	 To save time looking at many empty lists before finding a
	 runnable process the VM maintains a variable holding the
	 highest priority runnable process.  If this variable is 0 then the
	 VM does not know the highest priority and must search all lists.

	 Override to answer nil when there is no runnable process instead of
	 aborting.  In the threaded VM the abort test is done in transferTo:from:
	 becaue there may be some thread waiting to own the VM.  The transfer
	 to the thread shouldn't be done here because not all clients call this in
	 the right context (allowing a longjmp back to the threadSchedulingLoop)."
	| schedLists p processList proc ctxt |
	self externalWriteBackHeadFramePointers.
	schedLists := objectMemory fetchPointer: ProcessListsIndex ofObject: self schedulerPointer.
	p := highestRunnableProcessPriority = 0
			ifTrue: [objectMemory numSlotsOf: schedLists]
			ifFalse: [highestRunnableProcessPriority].
	[(p := p - 1) >= 0] whileTrue:
		[processList := objectMemory fetchPointer: p ofObject: schedLists.
	 	 [self isEmptyList: processList] whileFalse:
			["Only answer processes with a runnable suspendedContext.
			  Discard those that aren't; the VM would crash otherwise."
			 proc := self removeFirstLinkOfList: processList.
			 ctxt := objectMemory fetchPointer: SuspendedContextIndex ofObject: proc.
			 (self isLiveContext: ctxt) ifTrue:
				[highestRunnableProcessPriority := p + 1.
				^proc].
			  self cppIf: SPURVM ifTrue: 
				["This is uncommon, so we can deal with forwarders here instead of assuming there isn't."
				 (self isOopForwarded: ctxt) ifTrue:
					[ctxt := self fixFollowedField: SuspendedContextIndex ofObject: proc withInitialValue: ctxt].
				  (self isLiveContext: ctxt) ifTrue:
					[highestRunnableProcessPriority := p + 1.
					^proc].].
			 self warning: 'evicted zombie process from run queue']].
	^nil! !

CoInterpreterMT subclass: #CogVMSimulator
	instanceVariableNames: 'parent enableCog byteCount lastPollCount lastExtPC sendCount lookupCount printSends traceOn myBitBlt displayForm fakeForm imageName pluginList mappedPluginEntries quitBlock transcript displayView eventTransformer printFrameAtEachStep printBytecodeAtEachStep systemAttributes uniqueIndices uniqueIndex breakCount atEachStepBlock startMicroseconds lastYieldMicroseconds externalSemaphoreSignalRequests externalSemaphoreSignalResponses extSemTabSize debugStackDepthDictionary performFilters eventQueue effectiveCogCodeSize expectedSends expecting inputSemaphoreIndex zeroNextProfileTickCount perMethodProfile'
	classVariableNames: 'ByteCountsPerMicrosecond ExpectedSends NLRFailures NLRSuccesses StackAlteringPrimitives'
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CogVMSimulator commentStamp: 'eem 9/3/2013 11:16' prior: 0!
This class defines basic memory access and primitive simulation so that the CoInterpreter can run simulated in the Squeak environment.  It also defines a number of handy object viewing methods to facilitate pawing around in the object memory.  Remember that you can test the Cogit using its class-side in-image compilation facilities.

To see the thing actually run, you could (after backing up this image and changes), execute

	(CogVMSimulator new openOn: Smalltalk imageName) test

and be patient both to wait for things to happen, and to accept various things that may go wrong depending on how large or unusual your image may be.  We usually do this with a small and simple benchmark image.

Here's an example to launch the simulator in a window.  The bottom-right window has a menu packed with useful stuff:

(CogVMSimulator newWithOptions: #(Cogit StackToRegisterMappingCogit))
	desiredNumStackPages: 8;
	openOn: '/Users/eliot/Cog/startreader.image';
	openAsMorph;
	run

Here's a hairier example that I (Eliot) actually use in daily development with some of the breakpoint facilities commented out.

| cos proc opts |
CoInterpreter initializeWithOptions: (opts := Dictionary newFromPairs: #(Cogit StackToRegisterMappingCogit)).
CogVMSimulator chooseAndInitCogitClassWithOpts: opts.
cos := CogVMSimulator new.
"cos initializeThreadSupport." "to test the multi-threaded VM"
cos desiredNumStackPages: 8. "to set the size of the stack zone"
"cos desiredCogCodeSize: 8 * 1024 * 1024." "to set the size of the Cogit's code zone"
cos openOn: '/Users/eliot/Squeak/Squeak4.4/trunk44.image'. "choose your favourite image"
"cos setBreakSelector: 'r:degrees:'." "set a breakpoint at a specific selector"
proc := cos cogit processor.
"cos cogit sendTrace: 7." "turn on tracing"
"set a complex breakpoint at a specific point in machine code"
"cos cogit singleStep: true; breakPC: 16r56af; breakBlock: [:cg|  cos framePointer > 16r101F3C and: [(cos longAt: cos framePointer - 4) = 16r2479A and: [(cos longAt: 16r101F30) = (cos longAt: 16r101F3C) or: [(cos longAt: 16r101F2C) = (cos longAt: 16r101F3C)]]]]; sendTrace: 1".
"[cos cogit compilationTrace: -1] on: MessageNotUnderstood do: [:ex|]." "turn on compilation tracing in the StackToRegisterMappingCogit"
"cos cogit setBreakMethod: 16rB38880."
cos
	openAsMorph;
	"toggleTranscript;" "toggleTranscript will send output to the Transcript instead of the morph's rather small window"
	halt;
	run!

!CogVMSimulator class methodsFor: 'debug support' stamp: 'eem 7/12/2016 13:10'!
expectedSends
	^ExpectedSends! !

!CogVMSimulator class methodsFor: 'class initialization' stamp: 'eem 9/26/2020 12:43'!
initialize
	"These are primitives that alter the state of the stack.  They are here simply for assert checking.
	 After invocation the Cogit should not check for the expected stack delta when these primitives
	 succeed, because the stack will usually have been modified."
	StackAlteringPrimitives := #(	primitiveClosureValue primitiveClosureValueWithArgs primitiveClosureValueNoContextSwitch
									primitiveClone primitiveInstVarAt primitiveSlotAt "because these can cause code compactions..."
									primitiveEnterCriticalSection primitiveExitCriticalSection
									primitiveFullClosureValue primitiveFullClosureValueWithArgs primitiveFullClosureValueNoContextSwitch
									primitiveSignal primitiveWait primitiveResume primitiveSuspend primitiveYield
									primitiveExecuteMethodArgsArray primitiveExecuteMethod
									primitivePerform primitivePerformWithArgs primitivePerformInSuperclass
									primitiveTerminateTo primitiveStoreStackp primitiveDoPrimitiveWithArgs) asIdentitySet! !

!CogVMSimulator class methodsFor: 'initialization' stamp: 'eem 6/7/2018 15:16'!
initializeWithOptions: optionsDictionary
	super initializeWithOptions: optionsDictionary.

	ByteCountsPerMicrosecond := InitializationOptions
										at: #ByteCountsPerMicrosecond
										ifAbsent: [100]! !

!CogVMSimulator class methodsFor: 'class initialization' stamp: 'eem 6/7/2018 15:03'!
initializeWithOptions: optionsDictionaryOrArray objectMemoryClass: objectMemoryClassOrNil
	"The relevant ObjectMemory, Interpreter and Cogit classes must be initialized in order.
	 This happens notionally every time we start the simulator,
	 but in fact happens when ever we instantiate a simulator."
	InitializationOptions := optionsDictionaryOrArray isArray
							ifTrue: [Dictionary newFromPairs: optionsDictionaryOrArray]
							ifFalse: [optionsDictionaryOrArray].

	(objectMemoryClassOrNil ifNil: [self objectMemoryClass])
		initializeWithOptions: InitializationOptions.

	self initializeWithOptions: InitializationOptions.

	(self cogitClass withAllSuperclasses copyUpTo: Cogit) reverseDo:
		[:c| c initializeWithOptions: InitializationOptions]! !

!CogVMSimulator class methodsFor: 'instance creation' stamp: 'eem 9/3/2013 11:01'!
new
	^self onObjectMemory: nil cogit: nil options: #()! !

!CogVMSimulator class methodsFor: 'instance creation' stamp: 'eem 9/3/2013 11:11'!
newWithOptions: optionsDictionaryOrArray
	^self onObjectMemory: nil cogit: nil options: optionsDictionaryOrArray! !

!CogVMSimulator class methodsFor: 'instance creation' stamp: 'eem 9/3/2013 11:01'!
onObjectMemory: anObjectMemory
	^self onObjectMemory: anObjectMemory cogit: nil options: #()! !

!CogVMSimulator class methodsFor: 'instance creation' stamp: 'eem 9/3/2013 11:01'!
onObjectMemory: anObjectMemory cogit: aCogit
	^self onObjectMemory: anObjectMemory cogit: aCogit options: #()! !

!CogVMSimulator class methodsFor: 'instance creation' stamp: 'eem 10/11/2013 16:37'!
onObjectMemory: anObjectMemory cogit: aCogit options: optionsDictionaryOrArray
	| simulatorClass |
	^self == CogVMSimulator
		ifTrue:
			[simulatorClass := SmalltalkImage current endianness == #big
				ifTrue: [self notYetImplemented]
				ifFalse: [CogVMSimulatorLSB].
			simulatorClass
				initializeWithOptions: optionsDictionaryOrArray
				objectMemoryClass: (anObjectMemory ifNotNil: [anObjectMemory class]).
			 simulatorClass
				onObjectMemory: (anObjectMemory ifNil:
										[self objectMemoryClass simulatorClass new])
				cogit: aCogit
				options: optionsDictionaryOrArray]
		ifFalse:
			[| sim |
			sim := self basicNew.
			sim objectMemory: anObjectMemory.
			sim cogit: aCogit.
			sim initialize.
			COGMTVM ifTrue: "Set via options"
				[sim initializeThreadSupport; initialize].
			sim]! !

!CogVMSimulator class methodsFor: 'accessing' stamp: 'eem 4/25/2020 17:58'!
stackAlteringPrimitives
	^StackAlteringPrimitives! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 10/30/2020 07:49'!
_longjmp: aJumpBuf _: returnValue
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #'_longjmp:_:'
		withArguments: {aJumpBuf. returnValue}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'interpreter shell' stamp: 'eem 8/20/2021 15:01'!
aboutToDispatchBytecode
	perMethodProfile ifNotNil:
		[:pmp| pmp at: method put: (pmp at: method ifAbsent: 0) + 1].
	self incrementByteCount.
	self assertValidExecutionPointers.
	atEachStepBlock value "N.B. may be nil"! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 2/4/2014 11:15'!
allObjectsSelect: objBlock
	"self allObjectsSelect: [:oop | (self baseHeader: oop) = 1234]"

	| selected |
	selected := OrderedCollection new.
	objectMemory allObjectsDoSafely:
		[:obj|
		(objBlock value: obj) ifTrue: [selected addLast: obj]].
	^selected! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
assertSaneThreadAndProcess
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #assertSaneThreadAndProcess
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/24/2020 22:33'!
assertValidExternalStackPointers
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #assertValidExternalStackPointers
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'accessing' stamp: 'eem 1/23/2017 09:22'!
atEachStepBlock
	^atEachStepBlock! !

!CogVMSimulator methodsFor: 'accessing' stamp: 'eem 1/23/2017 09:22'!
atEachStepBlock: aBlock
	atEachStepBlock := aBlock! !

!CogVMSimulator methodsFor: 'jump bytecodes' stamp: 'eem 12/5/2010 10:34'!
attemptToSwitchToMachineCode: bcpc
	"method = 16r96A294 ifTrue: [self halt]."
	^super attemptToSwitchToMachineCode: bcpc! !

!CogVMSimulator methodsFor: 'return bytecodes' stamp: 'eem 4/16/2014 09:53'!
baseFrameReturn
	"| contextToReturnTo |
	contextToReturnTo := self frameCallerContext: localFP.
	(self stringOf: (self penultimateLiteralOf: (self fetchPointer: MethodIndex ofObject: contextToReturnTo))) = #indexOf:startingAt:ifAbsent: ifTrue:
		[self halt]."
	^super baseFrameReturn! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
bindProcess: aProcess toId: newId
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #bindProcess:toId:
		withArguments: {aProcess. newId}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'accessing' stamp: 'eem 1/5/2018 15:20'!
breakCount
	^breakCount! !

!CogVMSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:07'!
browserPluginInitialiseIfNeeded
"do nothing - its a macro in C code to support Mac browser plugin strangeness"! !

!CogVMSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:03'!
browserPluginReturnIfNeeded
"do nothing - its a macro in C code to support Mac browser plugin strangeness"! !

!CogVMSimulator methodsFor: 'memory access' stamp: 'eem 10/16/2018 09:09'!
byteAt: byteAddress
	"This is really only for the C library simulations memcpy:_:_: et al in VMClass.
	 Use objectMemory byteAt: directly where possible."
	^objectMemory byteAt: byteAddress! !

!CogVMSimulator methodsFor: 'memory access' stamp: 'eem 6/25/2011 12:22'!
byteAt: byteAddress put: byte
	^objectMemory byteAt: byteAddress put: byte! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'di 6/21/2004 15:05'!
byteCount
	"So you can call this from temp debug statements in, eg, Interpreter, such as
	self byteCount = 12661 ifTrue: [self halt].
	"

	^ byteCount! !

!CogVMSimulator methodsFor: 'UI' stamp: 'eem 1/16/2018 15:38'!
byteCountText
	^(String streamContents: [:s| s nextPutAll: byteCount asStringWithCommas; nextPut: $/; nextPutAll: sendCount asStringWithCommas]) asText! !

!CogVMSimulator methodsFor: 'translation support' stamp: 'eem 1/1/1970 11:18'!
cCoerceSimple: value to: cTypeString
	"Type coercion for translation and simulation.
	 For simulation answer a suitable surrogate for the struct types"
	^cTypeString
		caseOf:
		   {	[#'CogMethod *']		->	[cogit cogMethodSurrogateAt: value asUnsignedInteger].
			[#'CogBlockMethod *']	->	[cogit cogBlockMethodSurrogateAt: value asUnsignedInteger].
			[#'NSSendCache *']		->	[cogit nsSendCacheSurrogateAt: value asUnsignedInteger] }
		otherwise: [value]! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
cStackRangeForCurrentThread
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #cStackRangeForCurrentThread
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
cStackRangeForThreadIndex: threadIndex
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #cStackRangeForThreadIndex:
		withArguments: {threadIndex}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'plugin support' stamp: 'eem 10/30/2014 12:49'!
callExternalPrimitive: mapIndex
	| entry |
	entry := self pluginEntryFor: mapIndex.
	"Spur needs the primitiveFunctionPointer to be set correctly
	 for accurate following of forwarders on primitive failure."
	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[primitiveFunctionPointer := entry at: 2].
	^(entry at: 1) perform: (entry at: 2)! !

!CogVMSimulator methodsFor: 'enilopmarts' stamp: 'eem 5/6/2009 16:05'!
ceActivateFailingPrimitiveMethod: aPrimitiveMethod
	"self halt."
	^super ceActivateFailingPrimitiveMethod: aPrimitiveMethod! !

!CogVMSimulator methodsFor: 'trampolines' stamp: 'eem 4/16/2014 09:53'!
ceBaseFrameReturn: returnValue
	"self printCallStackOf: (stackPages longAt: stackPage baseAddress) currentFP: stackPage baseFP.
	Transcript print: byteCount; tab; print: thisContext; cr.
	(self confirm: 'continue?') ifFalse: [self returnValue hex]."
	"returnValue = 16r01934F78 ifTrue: [self halt]."
	"| contextToReturnTo |
	contextToReturnTo := stackPages longAt: stackPage baseAddress.
	(self stringOf: (self penultimateLiteralOf: (self fetchPointer: MethodIndex ofObject: contextToReturnTo))) = #indexOf:startingAt:ifAbsent: ifTrue:
		[self halt]."
	^super ceBaseFrameReturn: returnValue! !

!CogVMSimulator methodsFor: 'trampolines' stamp: 'eem 12/2/2010 16:57'!
ceContext: aOnceMarriedContext instVar: slotIndex
	"aOnceMarriedContext = 26431360 ifTrue: [self halt]."
	cogit assertCStackWellAligned.
	^super ceContext: aOnceMarriedContext instVar: slotIndex! !

!CogVMSimulator methodsFor: 'trampolines' stamp: 'eem 12/2/2010 18:29'!
ceContext: aOnceMarriedContext instVar: slotIndex value: anOop
	"aOnceMarriedContext = 26431360 ifTrue:
		[transcript cr; cr.
		 self printContext: aOnceMarriedContext.
		 transcript cr.
		 (self isContext: anOop)
			ifTrue: [self printContext: anOop]
			ifFalse: [self printOop: anOop].
		 self halt]."
	cogit assertCStackWellAligned.
	^super ceContext: aOnceMarriedContext instVar: slotIndex value: anOop! !

!CogVMSimulator methodsFor: 'cog jit support' stamp: 'eem 4/14/2014 17:31'!
ceCounterTripped: condition
	| counterTrippedSelector |
	self transcript cr; nextPutAll: 'counter tripped in '.
	self shortPrintFrame: framePointer.
	counterTrippedSelector := objectMemory maybeSplObj: SelectorCounterTripped.
	(counterTrippedSelector isNil
	or: [counterTrippedSelector = objectMemory nilObject]) ifFalse:
		[self halt: 'counter tripped'].
	^super ceCounterTripped: condition! !

!CogVMSimulator methodsFor: 'trampolines' stamp: 'eem 5/15/2015 12:42'!
ceMNUFromPICMNUMethod: aMethodObj receiver: rcvr
	| cPIC |
	cPIC := self cCoerceSimple: self stackTop - cogit mnuOffset to: #'CogMethod *'.
	self assert: (cPIC cmType = CMClosedPIC or: [cPIC cmType = CMOpenPIC]).
	self mnuBreakpoint: cPIC selector receiver: nil.
	^super ceMNUFromPICMNUMethod: aMethodObj receiver: rcvr! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 2/20/2009 21:58'!
ceNonLocalReturn: returnValue
	"self halt."
	^super ceNonLocalReturn: returnValue! !

!CogVMSimulator methodsFor: 'trampolines' stamp: 'eem 12/8/2017 14:45'!
ceReapAndResetErrorCodeFor: cogMethod
	"Override to map the address into a CogMethodSurrogate"
	| surrogate |
	surrogate := cogMethod isInteger
					ifTrue: [cogit cogMethodSurrogateAt: cogMethod]
					ifFalse: [cogMethod].
	^super ceReapAndResetErrorCodeFor: surrogate! !

!CogVMSimulator methodsFor: 'trampolines' stamp: 'eem 10/2/2013 17:40'!
ceSend: selector super: superNormalBar to: rcvr numArgs: numArgs
	"self stringOf: selector"
	"self printOop: rcvr"
	"(superNormalBar ~= 0 and: [(self stringOf: selector) = #bitShift:]) ifTrue:
		[self halt]."
	self logSend: selector.
	cogit assertCStackWellAligned.
	self maybeCheckStackDepth: numArgs + 1 sp: stackPointer pc: (stackPages longAt: stackPointer).
	^super ceSend: selector super: superNormalBar to: rcvr numArgs: numArgs! !

!CogVMSimulator methodsFor: 'trampolines' stamp: 'eem 1/4/2017 10:06'!
ceSendAbort: selector to: rcvr numArgs: numArgs
	"self stringOf: selector"
	"self printOop: rcvr"
	self logSend: selector.
	cogit assertCStackWellAligned.
	self maybeCheckStackDepth: numArgs + 1
		sp: stackPointer
		pc: (stackPages longAt: stackPointer).
	^super ceSendAbort: selector to: rcvr numArgs: numArgs! !

!CogVMSimulator methodsFor: 'trampolines' stamp: 'eem 10/17/2015 07:42'!
ceSendFromInLineCacheMiss: oPIC
	"Override to map the address into a CogMethodSurrogate"
	| surrogate |
	surrogate := oPIC isInteger
					ifTrue: [cogit cogMethodSurrogateAt: oPIC]
					ifFalse: [oPIC].
	self logSend: surrogate selector.
	(surrogate cmNumArgs = 0
	 and: [(self stackValue: 1) = 16r8169D0
	 and: [self stackTop = 16r53EA7]]) ifTrue:
		[self halt].
	^super ceSendFromInLineCacheMiss: surrogate! !

!CogVMSimulator methodsFor: 'trampolines' stamp: 'eem 6/15/2009 19:40'!
ceSendMustBeBoolean: anObject
	self halt.
	^super ceSendMustBeBoolean: anObject! !

!CogVMSimulator methodsFor: 'trampolines' stamp: 'eem 1/18/2018 11:25'!
ceStackOverflow: contextSwitchIfNotNil
	"Override to bump up the byteCount from which the microsecond clock is derived."
	byteCount := byteCount + 1000.
	self doOrDefer: [self changed: #byteCountText; changed: #composeAll].
	^super ceStackOverflow: contextSwitchIfNotNil! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 3/18/2015 15:34'!
ceTraceBlockActivation
	cogit printOnTrace ifTrue:
		[transcript print: byteCount; nextPut: $/; print: (sendCount := sendCount + 1); space].
	cogit assertCStackWellAligned.
	super ceTraceBlockActivation.
	^#continue! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 7/8/2021 17:09'!
ceTraceLinkedSend: theReceiver
	(sendCount := sendCount + 1) \\ 500 = 0 ifTrue:
		[self changed: #byteCountText].
	cogit printOnTrace ifTrue:
		[transcript print: byteCount; nextPut: $/; print: sendCount; space].
	super ceTraceLinkedSend: theReceiver.
	^#continue! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
cedeToHigherPriorityThreads
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #cedeToHigherPriorityThreads
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
checkForEventsMayContextSwitch: mayContextSwitch
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #checkForEventsMayContextSwitch:
		withArguments: {mayContextSwitch}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'object memory support' stamp: 'eem 9/30/2013 22:35'!
checkStackIntegrity
	"Override to deal with incomplete initialization."
	stackPages ifNil: [^true].
	stackPages pages ifNil: [^true].
	^super checkStackIntegrity! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
checkVMOwnershipFromHeartbeat
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #checkVMOwnershipFromHeartbeat
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 07/28/2014 07:54'!
classAndSelectorOfMethod: meth forReceiver: rcvr
	| mClass dict length methodArray |
	mClass := objectMemory fetchClassOf: rcvr.
	[dict := objectMemory fetchPointer: MethodDictionaryIndex ofObject: mClass.
	length := objectMemory numSlotsOf: dict.
	methodArray := objectMemory fetchPointer: MethodArrayIndex ofObject: dict.
	0 to: length-SelectorStart-1 do: 
		[:index | 
		meth = (objectMemory fetchPointer: index ofObject: methodArray) 
			ifTrue: [^ Array
				with: mClass
				with: (objectMemory fetchPointer: index + SelectorStart ofObject: dict)]].
	mClass := objectMemory fetchPointer: SuperclassIndex ofObject: mClass.
	mClass = objectMemory nilObject]
		whileFalse: [].
	^ Array
		with: (objectMemory fetchClassOf: rcvr)
		with: (objectMemory splObj: SelectorDoesNotUnderstand)! !

!CogVMSimulator methodsFor: 'plugin support' stamp: 'eem 8/2/2021 21:57'!
classNameOf: aClass Is: className
	"Check if aClass' name is className"
	| name |
	(objectMemory numSlotsOf: aClass) <= classNameIndex ifTrue:
		[^false]. "Not a class but maybe behavior" 
	name := objectMemory fetchPointer: classNameIndex ofObject: aClass.
	(objectMemory isBytes: name) ifFalse:
		[^false].
	^ className = (self stringOf: name)! !

!CogVMSimulator methodsFor: 'I/O primitives' stamp: 'eem 1/4/2017 08:40'!
clipboardRead: sz Into: actualAddress At: zeroBaseIndex
	| str |
	str := Clipboard clipboardText.
	1 to: sz do:
		[:i | objectMemory byteAt: actualAddress + zeroBaseIndex + i - 1 put: (str at: i) asciiValue]! !

!CogVMSimulator methodsFor: 'I/O primitives' stamp: 'di 9/23/2001 12:18'!
clipboardSize

	^ Clipboard clipboardText size! !

!CogVMSimulator methodsFor: 'I/O primitives' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
clipboardWrite: sz From: actualDataAddress At: ignored

	Clipboard clipboardText: (self stringOf: actualDataAddress - objectMemory baseHeaderSize)! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 6/21/2021 14:26'!
cloneSimulation
	| savedDisplayView savedDisplayForm savedQuitBlock savedTranscript |
	savedDisplayView := displayView. displayView := nil.
	savedDisplayForm := displayForm. displayForm = nil.
	savedQuitBlock := quitBlock. quitBlock := nil.
	savedTranscript := transcript. transcript := nil.

	^[| theClone |
	 self changed: #byteCountText.
	 Smalltalk garbageCollect.
	 theClone := self veryDeepCopy.
	 theClone parent: self; transcript: Transcript.
	 theClone objectMemory parent: objectMemory.
	 theClone]
		ensure:
			[displayView := savedDisplayView.
			 displayForm = savedDisplayForm.
			 quitBlock := savedQuitBlock.
			 transcript := savedTranscript]! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 5/25/2014 20:06'!
cloneSimulationWindow
	| label |
	label := 'Clone of ', (displayView containingWindow label allButFirst: 'Simulation of ' size).
	^self cloneSimulation openAsMorph
		setLabel: label;
		yourself! !

!CogVMSimulator methodsFor: 'initialize-release' stamp: 'eem 7/26/2020 12:14'!
close  "close any files that ST may have opened, etc"
	pluginList do: [:assoc| | plugin | plugin := assoc value. plugin ~~ self ifTrue: [plugin close]].
	"Ugh; at least some of this code belongs in the UI..."
	displayView ifNotNil:
		[displayView activeHand removeEventListener: self].
	ActiveHand removeEventListener: self.
	World submorphs do:
		[:submorph|
		(submorph model isVMObjectInspector
		 and: [submorph model coInterpreter == self]) ifTrue:
			[submorph delete].
		(submorph model isDebugger
		 and: [(submorph model interruptedProcess suspendedContext ifNotNil:
				[:sctxt|
				 sctxt findContextSuchThat:
					[:ctxt|
					(ctxt receiver == cogit
					 and: [ctxt selector == #simulateCogCodeAt:])
					or: [ctxt receiver == self
					 and: [ctxt selector == #interpret]]]]) notNil]) ifTrue:
			[submorph model windowIsClosing.
			 submorph delete]]! !

!CogVMSimulator methodsFor: 'primitive support' stamp: 'eem 10/23/2020 19:20'!
codeGeneratorToComputeAccessorDepth
	^(VMMaker new
		buildCodeGeneratorForInterpreter: (COGMTVM ifTrue: [CoInterpreterMT] ifFalse: [CoInterpreterPrimitives])
		includeAPIMethods: false
		initializeClasses: false)
			logger: self transcript;
			yourself! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 12/29/2008 12:30'!
cogit
	^cogit! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 7/12/2016 12:59'!
collectSends
	expectedSends := WriteStream on: (Array new: 65536).
	expecting := false! !

!CogVMSimulator methodsFor: 'cog jit support' stamp: 'eem 1/4/2017 13:53'!
commenceCogCompiledCodeCompaction
	"self halt."
	^super commenceCogCompiledCodeCompaction! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem 8/8/2014 15:07'!
convertToMachineCodeFrame: cogHomeMethod bcpc: bcpc
	"bytecodeSetSelector ~= 0 ifTrue: [self halt]".
	^super convertToMachineCodeFrame: cogHomeMethod bcpc: bcpc! !

!CogVMSimulator methodsFor: 'I/O primitives support' stamp: ''!
copyBits

	^ myBitBlt copyBits! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 6/20/2008 15:46'!
cr

	traceOn ifTrue: [ transcript cr; flush ].! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 5/4/2015 20:55'!
createActualMessageTo: class

	class = objectMemory nilObject ifTrue: [self halt].

	^super createActualMessageTo: class! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 5/14/2014 11:35'!
debugStackPointersFor: aMethod
	^CArrayAccessor on:
		(StackDepthFinder on: (VMCompiledMethodProxy new
									for: aMethod
									coInterpreter: self
									objectMemory: objectMemory))
			stackPointers! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
deferStackLimitSmashAround: functionSymbol
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #deferStackLimitSmashAround:
		withArguments: {functionSymbol}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 7/18/2013 11:28'!
deferStackLimitSmashAround: functionSymbol with: arg
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #deferStackLimitSmashAround:with:
		withArguments: {functionSymbol. arg}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'initialize-release' stamp: 'eem 3/6/2009 19:14'!
desiredCogCodeSize: anInteger
	desiredCogCodeSize := anInteger! !

!CogVMSimulator methodsFor: 'UI' stamp: 'eem 5/11/2011 16:58'!
desiredDisplayExtent
	^(savedWindowSize
		ifNil: [640@480]
		ifNotNil: [savedWindowSize >> 16 @ (savedWindowSize bitAnd: 16rFFFF)])
			min: Display extent * 2 // 3! !

!CogVMSimulator methodsFor: 'initialize-release' stamp: 'eem 3/6/2009 19:17'!
desiredEdenBytes: anInteger
	desiredEdenBytes := anInteger! !

!CogVMSimulator methodsFor: 'initialize-release' stamp: 'eem 3/6/2009 19:14'!
desiredNumStackPages: anInteger
	desiredNumStackPages := anInteger! !

!CogVMSimulator methodsFor: 'vm scheduling' stamp: 'eem 11/9/2020 16:37'!
disownVM: flags
	"Override to insert a stack depth check."
	self assert: self successful.
	(self isMachineCodeFrame: framePointer) ifTrue:
		[self maybeCheckStackDepth: argumentCount
			sp: stackPointer
			pc: instructionPointer].
	^self perform: #disownVM:
		withArguments: {flags}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'plugin primitive support' stamp: 'eem 10/9/2013 08:48'!
dispatchMappedPluginEntry: n 
	^(mappedPluginEntries at: n) first
		perform: (mappedPluginEntries at: n) second! !

!CogVMSimulator methodsFor: 'interpreter shell' stamp: 'eem 12/5/2015 10:46'!
dispatchOn: anInteger in: selectorArray
	"Simulate a case statement via selector table lookup.
	The given integer must be between 0 and selectorArray size-1, inclusive.
	For speed, no range test is done, since it is done by the at: operation.
	Note that, unlike many other arrays used in the Interpreter, this method expect NO CArrayAccessor wrapping - it would duplicate the +1. Maybe this would be better updated to make it all uniform"
	cogit breakPC ifNil:
		[cogit breakBlock ifNotNil:
			[:bb| (bb value: cogit) ifTrue: [self halt: 'breakpoint reached']]].
	self perform: (selectorArray at: (anInteger + 1)).! !

!CogVMSimulator methodsFor: 'I/O primitives support' stamp: ''!
displayLocation

	^ Display extent - displayForm extent - (10@10)! !

!CogVMSimulator methodsFor: 'accessing' stamp: 'tty 2/5/2014 12:45'!
displayView
	"The VM SImulator draws directly on an imageMorph named displayView
	  displayView is housed on a Form named displayForm
	  Since this is housed directly on the Morph, I have this tight compling from here->model->morph. 
	"
	self flag: 'tty'.	
	^displayView! !

!CogVMSimulator methodsFor: 'process primitive support' stamp: 'eem 12/29/2016 15:19'!
doSignalExternalSemaphores: minTableSize
	"This is a non-thread-safe simulation.  See platforms/Cross/vm/sqExternalSemaphores.c
	 for the real code.  For the benefit of the SocketPluginSimulator, do a yield every 100
	 virtual microseconds."
	| now switched |
	now := self ioUTCMicroseconds.
	now - lastYieldMicroseconds >= 100 ifTrue:
		[lastYieldMicroseconds := now.
		 Processor yield].

	switched := false.
	1 to: (minTableSize min: externalSemaphoreSignalRequests size) do:
		[:i| | responses |
		responses := externalSemaphoreSignalResponses at: i.
		[responses < (externalSemaphoreSignalRequests at: i)] whileTrue:
			[(self doSignalSemaphoreWithIndex: i) ifTrue:
				[switched := true].
			 externalSemaphoreSignalResponses at: i put: (responses := responses + 1)]].
	^switched! !

!CogVMSimulator methodsFor: 'error handling' stamp: 'eem 12/15/2014 10:05'!
doesNotUnderstand: aMessage
	"If this is a doit and the objectMemory understands, pass it on."
	(thisContext findContextSuchThat: [:ctxt| ctxt selector == #evaluate:in:to:notifying:ifFail:logged:]) ifNotNil:
		[(objectMemory class whichClassIncludesSelector: aMessage selector) ifNotNil:
			[:implementingClass|
			(implementingClass inheritsFrom: Object) ifTrue: "i.e. VMClass and below"
				[thisContext sender selector ~~ #DoIt ifTrue:
					[Transcript nextPutAll: 'warning: redirecting ', aMessage selector, ' in ', thisContext sender printString, ' to objectMemory'; cr; flush].
				 aMessage lookupClass: nil.
				^aMessage sentTo: objectMemory]].
		(cogit class whichClassIncludesSelector: aMessage selector) ifNotNil:
			[:implementingClass|
			(implementingClass inheritsFrom: Object) ifTrue: "i.e. VMClass and below"
				[thisContext sender selector ~~ #DoIt ifTrue:
					[Transcript nextPutAll: 'warning: redirecting ', aMessage selector, ' in ', thisContext sender printString, ' to cogit'; cr; flush].
				 aMessage lookupClass: nil.
				^aMessage sentTo: cogit]]].
	^super doesNotUnderstand: aMessage! !

!CogVMSimulator methodsFor: 'I/O primitives support' stamp: ''!
drawLoopX: xDelta Y: yDelta

	^ myBitBlt drawLoopX: xDelta Y: yDelta! !

!CogVMSimulator methodsFor: 'debug support' stamp: ''!
dumpMethodHeader: hdr
	^ String streamContents:
		[:strm |
		strm nextPutAll: '<nArgs=', ((hdr >> 25) bitAnd: 16r1F) printString , '>'.
		strm nextPutAll: '<nTemps=', ((hdr >> 19) bitAnd: 16r3F) printString , '>'.
		strm nextPutAll: '<lgCtxt=', ((hdr >> 18) bitAnd: 16r1) printString , '>'.
		strm nextPutAll: '<nLits=', ((hdr >> 10) bitAnd: 16rFF) printString , '>'.
		strm nextPutAll: '<prim=', ((hdr >> 1) bitAnd: 16r1FF) printString , '>'.
		]! !

!CogVMSimulator methodsFor: 'accessing' stamp: 'eem 1/30/2020 13:07'!
effectiveCogCodeSize
	"With the single-mapped regime the effectiveCogCodeSize is cogCodeSize.
	 With the dual-mapped regime then in production effectiveCogCodeSize is
	 cogCodeSize and in simulation it is two times cogCodeSize (meaning we
	 simulate a dual zone by having two copies of the zone at the start of memory)."
	^effectiveCogCodeSize! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 6/20/2008 15:25'!
elementsPerPrintOopLine
	^4! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 2/7/2009 08:39'!
enableCog: aBoolean
	enableCog := aBoolean! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 8/20/2021 15:53'!
enablePerMethodProfiling
	perMethodProfile := Dictionary new.
	cogit enablePerMethodProfiling! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem 12/3/2010 08:34'!
ensureContextIsExecutionSafeAfterAssignToStackPointer: aContext
	"16r1934F80 = aContext ifTrue:
		[self halt]."
	^super ensureContextIsExecutionSafeAfterAssignToStackPointer: aContext! !

!CogVMSimulator methodsFor: 'testing' stamp: 'eem 9/27/2010 16:49'!
ensureDebugAtEachStepBlock
	atEachStepBlock := [printFrameAtEachStep ifTrue:
							[self printFrame: localFP WithSP: localSP].
						 printBytecodeAtEachStep ifTrue:
							[self printCurrentBytecodeOn: transcript.
							 transcript cr; flush].
						 byteCount = breakCount ifTrue:
							["printFrameAtEachStep :=" printBytecodeAtEachStep := true]]! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem 12/21/2016 10:52'!
ensureMethodIsCogged: methodObj maybeClosure: maybeClosure
	"Uncomment this to compact frequently and hence test if clients are ready for the shock."
	"[self commenceCogCompiledCodeCompaction]
		on: Halt
		do: [:ex| ex resume: nil]."
	^super ensureMethodIsCogged: methodObj maybeClosure: maybeClosure! !

!CogVMSimulator methodsFor: 'initialize-release' stamp: 'cb 4/13/2016 09:59'!
ensureMultiThreadingOverridesAreUpToDate
	"Make sure the CoInterpreterMT switch methods are implemented.  These methods select
	 between CoInterpreterMT's implementation or CoInterpreter's implementation depending
	 on cogThreadManager being non-nil or nil respectively.  i.e. they allow us to use this one
	 simulator class to simulate for both CoInterpreterMT and CoInterpreter."
	| thisClass me mtInterpreterClass |
	self cppIf: COGMTVM ifTrue: [  ] ifFalse: [ ^self ].
	thisClass := thisContext methodClass.
	mtInterpreterClass := thisClass superclass.
	me := thisClass name, '>>', thisContext method selector.
	"We want override switches for everything implemented by CoInterpreter
	 and CoInterpreterMT that is either not implemented by CogVMSimulator
	 or already implemented by CogVMSimulator as an override switch."
	(mtInterpreterClass selectors select:
		[:sel|
		(mtInterpreterClass superclass whichClassIncludesSelector: sel) notNil
		and: [(thisClass organization categoryOfElement: sel)
				ifNil: [true]
				ifNotNil: [:cat| cat == #'multi-threading simulation switch']]])
		do: [:sel| | argNames desiredSource |
			argNames := Parser new
							initPattern: (mtInterpreterClass sourceCodeAt: sel)
							return: [:pattern| pattern second].
			desiredSource := String streamContents:
								[:str|
								argNames isEmpty
									ifTrue: [str nextPutAll: sel]
									ifFalse:
										[sel keywords with: argNames do:
											[:kw :arg| str nextPutAll: kw; space; nextPutAll: arg; space].
										 str skip: -1].
								str
									crtab;
									nextPutAll: '"This method includes or excludes ', mtInterpreterClass name, ' methods as required.';
									crtab;
									nextPutAll: ' Auto-generated by ', me, '"';
									cr;
									crtab;
									nextPutAll: '^self perform: ';
									store: sel;
									crtab: 2;
									nextPutAll: 'withArguments: {'.
								argNames
									do: [:arg| str nextPutAll: arg]
									separatedBy: [str nextPut: $.; space].
								str
									nextPut: $};
									crtab: 2;
									nextPutAll: 'inSuperclass: (cogThreadManager ifNil: [';
									print: mtInterpreterClass superclass;
									nextPutAll: '] ifNotNil: [';
									print: mtInterpreterClass;
									nextPutAll: '])'].
			desiredSource ~= (thisClass sourceCodeAt: sel ifAbsent: ['']) asString ifTrue:
				[((thisClass includesSelector: sel)
				  and: [(thisClass compiledMethodAt: sel) messages includesAnyOf: #(halt halt:)])
					ifTrue: [self transcript cr; nextPutAll: 'WARNING, because of halts, not generating '; nextPutAll: desiredSource; cr; flush]
					ifFalse: [thisClass compile: desiredSource classified: #'multi-threading simulation switch']]].
	"Make sure obsolete CoInterpreterMT switch methods are deleted."
	((thisContext methodClass organization listAtCategoryNamed: #'multi-threading simulation switch') select:
		[:sel| (mtInterpreterClass whichClassIncludesSelector: sel) isNil]) do:
			[:sel| thisClass removeSelector: sel]! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
enterSmalltalkExecutive
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #enterSmalltalkExecutive
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
enterSmalltalkExecutiveImplementation
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #enterSmalltalkExecutiveImplementation
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'enilopmarts' stamp: 'eem 1/3/2011 17:30'!
executeCogBlock: cogMethod closure: closure mayContextSwitch: mayContextSwitch
	"self shortPrintFrameAndCallers: framePointer"
	"(#(16r56E90 16r133D0) includes: cogMethod asInteger)  ifTrue: [self halt]."
	"self halt."
	^super executeCogBlock: cogMethod closure: closure mayContextSwitch: mayContextSwitch! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 7/12/2016 12:59'!
expectSends: anArray
	expectedSends := ReadStream on: anArray.
	expecting := true! !

!CogVMSimulator methodsFor: 'miscellaneous bytecodes' stamp: 'eem 11/13/2012 16:40'!
extABytecode
	"224		11100000	aaaaaaaa	Extend A (Ext A = Ext A prev * 256 + Ext A)"
	(extA = 0 and: [extB = 0]) ifTrue:
		[lastExtPC := localIP].
	^super extABytecode! !

!CogVMSimulator methodsFor: 'miscellaneous bytecodes' stamp: 'eem 11/2/2016 15:18'!
extBBytecode
	"225		11100001	sbbbbbbb	Extend B (Ext B = Ext B prev * 256 + Ext B)"
	(extA = 0 and: [extB = 0]) ifTrue:
		[lastExtPC := localIP].
	^super extBBytecode! !

!CogVMSimulator methodsFor: 'jump bytecodes' stamp: 'eem 3/24/2017 15:49'!
extUnconditionalJump
	"242		11110010	i i i i i i i i	Jump i i i i i i i i (+ Extend B * 256, where bbbbbbbb = sddddddd, e.g. -32768 = i=0, a=0, s=1)"
	"| byte offset |
	byte := objectMemory byteAt: localIP + 1.
	offset := byte + (extB << 8).
	(offset < 0 and: [(self iframeBackwardBranchByte: localFP) - 2 = 1]) ifTrue: [self halt]."
	^super extUnconditionalJump! !

!CogVMSimulator methodsFor: 'return bytecodes' stamp: 'eem 1/18/2020 12:02'!
externalAboutToReturn: resultOop through: aContext
	| cogMethod |
	cogMethod := self mframeMethod: framePointer.
	"(cogMethod = 16rF0328 and:[resultOop ~= 16r3]) ifTrue: [self halt: resultOop hex]."
	NLRSuccesses ifNotNil:
		[NLRSuccesses at: cogMethod put: 1 + (NLRSuccesses at: cogMethod ifAbsent: 0)].
	[super externalAboutToReturn: resultOop through: aContext]
		on: AssertionFailure
		do: [:ex|
			"Block in SystemProgressMorph>>nextSlotFor:"
			"cogMethod = 16rF0328 ifTrue: [self halt: resultOop hex]."
			NLRSuccesses ifNotNil:
				[NLRSuccesses at: cogMethod put: (NLRSuccesses at: cogMethod) - 1.
				 NLRFailures at: cogMethod put: 1 + (NLRFailures at: cogMethod ifAbsent: 0)].
			ex resume]

	"NLRFailures := Dictionary new. NLRSuccesses := Dictionary new"
	"{ NLRFailures. NLRSuccesses }"! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 2/12/2009 16:51'!
externalCannotReturn: resultOop from: aContext
	self halt.
	^super externalCannotReturn: resultOop from: aContext! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 2/7/2009 10:56'!
externalDivorceFrame: theFP andContext: ctxt
	"((#(16r100570 16r101BC8) includes: theFP) or: [#(16r17159A4 16r1715948) includes: ctxt]) ifTrue:
		[self halt]."
	^super externalDivorceFrame: theFP andContext: ctxt! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 4/17/2014 15:30'!
externalInstVar: offset ofContext: aContext

	"offset = InstructionPointerIndex ifTrue:
		[transcript nextPutAll: '==================='; cr.
		 self printContext: 16r1715630.
		 self printCallStackOf: aContext currentFP: framePointer.
		 transcript nextPutAll: '==================='; cr.
		 self halt]."

	| result |
	"self shortPrintFrameAndCallers: framePointer.
	transcript print: byteCount; tab; print: thisContext; cr.
	self print: offset; cr.
	self printContext: aContext.
	(self confirm: 'continue?') ifFalse: [self halt]."
	result := super externalInstVar: offset ofContext: aContext.
	"(offset = InstructionPointerIndex
	and: [(self stringOf: (self penultimateLiteralOf: (self fetchPointer: MethodIndex ofObject: aContext))) = #indexOf:startingAt:ifAbsent:]) ifTrue:
		[transcript space; nextPutAll: '^pc '; nextPutAll: result hex.
		 (objectMemory isIntegerObject: result) ifTrue:
			[transcript space; print: (objectMemory integerValueOf: result)].
		 transcript tab.
		 self shortPrintContext: aContext]."
	"offset = StackPointerIndex ifTrue:
		[transcript nextPutAll: '^stackp ', (self integerValueOf: result) printString; tab.
		 self shortPrintContext: aContext.
		 (#(24205456 24205732) includes: aContext) ifTrue:
		 	[(self checkIsStillMarriedContext: aContext currentFP: framePointer)
				ifTrue: [self printFrame: (self frameOfMarriedContext: aContext) WithSP: (self frameOfMarriedContext: aContext) - 48]
				ifFalse: [self printContext: aContext]]]."
	^result! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem 4/16/2014 09:54'!
externalInstVar: index ofContext: aContext put: anOop
	"self shortPrintFrameAndCallers: framePointer.
	Transcript print: byteCount; tab; print: thisContext; cr.
	self print: index; cr.
	self printContext: aContext.
	self shortPrintOop: anOop.
	(self confirm: 'continue?') ifFalse: [self halt]."
	"(self stringOf: (self penultimateLiteralOf: (self fetchPointer: MethodIndex ofObject: aContext))) = #indexOf:startingAt:ifAbsent: ifTrue:
		[self halt]."
	^super externalInstVar: index ofContext: aContext put: anOop! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem 1/4/2017 11:29'!
externalSetStackPageAndPointersForSuspendedContextOfProcess: aProcess
	"Override to check stack depth (hence in the simulator only)."
	super externalSetStackPageAndPointersForSuspendedContextOfProcess:  aProcess.
	(self isMachineCodeFrame: framePointer) ifTrue:
		[self maybeCheckStackDepth: 0
			sp: stackPointer
			pc: instructionPointer]! !

!CogVMSimulator methodsFor: 'interpreter shell' stamp: 'eem 6/25/2011 12:30'!
fetchByte
	^objectMemory byteAt: (localIP := localIP + 1)! !

!CogVMSimulator methodsFor: 'interpreter access' stamp: 'eem 11/22/2013 14:54'!
fetchPointer: fieldIndex ofObject: oop
	"index by word size, and return a pointer as long as the word size"
	self assert: (objectMemory cheapIsInMemory: oop).
	^objectMemory fetchPointer: fieldIndex ofObject: oop! !

!CogVMSimulator methodsFor: 'internal interpreter access' stamp: 'eem 1/9/2020 15:36'!
fetchStackPointerOf: aContext
	"Return the stackPointer of a Context or BlockContext.
	 Does not deal with married contexts.  Use only for debug
	 printing or object tracing functions.  To obtain an accurate
	 stack pointer use stackPointerForMaybeMarriedContext:"
	| sp |
	sp := super fetchStackPointerOf: aContext.
	self assert: sp >= 0.
	^sp! !

!CogVMSimulator methodsFor: 'control primitives' stamp: 'eem 1/11/2013 18:38'!
filterPerformOf: selector to: receiver
	"This is to allow simulator to filter start-up items to avoid as-yet unsimulatable plugins."
	performFilters ifNil: [^false].
	(performFilters at: (self shortPrint: receiver) ifAbsent: []) ifNotNil:
		[:messages|
		^messages includes: (self stringOf: selector)].
	^false! !

!CogVMSimulator methodsFor: 'testing' stamp: 'eem 12/8/2016 11:26'!
findNewMethodInClassTag: classTag
"
	| cName |
	traceOn ifTrue:
		[cName := (self sizeBitsOf: class) = 16r20
			ifTrue: ['class ' , (self nameOfClass: (self fetchPointer: 6 ofObject: class))]
			ifFalse: [(self nameOfClass: class)].
		self cr; print: cName , '>>' , (self stringOf: messageSelector)].
"
	messageSelector = (objectMemory splObj: SelectorDoesNotUnderstand) ifTrue: [self halt].

	self logSend: messageSelector.
"
	(sendCount > 1000 and: [sendCount\\10 = 0]) ifTrue:
		[Transcript print: sendCount; space.
		self validate].
"
"
	(sendCount > 100150) ifTrue:
		[self qvalidate.
		messageQueue == nil ifTrue: [messageQueue := OrderedCollection new].
		messageQueue addLast: (self stringOf: messageSelector)].
"
	^super findNewMethodInClassTag: classTag! !

!CogVMSimulator methodsFor: 'memory access' stamp: 'eem 6/11/2010 11:37'!
firstIndexableField: oop
	"This is in ObjectMemory and overridden in the obj mem simulators"
	self shouldNotImplement! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 1/3/2021 17:33'!
flush
	traceOn ifTrue:
		[transcript flush.
		 "We *always* want to make output visible on flush"
		 TranscriptStream forceUpdate ifFalse:
			[transcript changed: #appendEntry]]! !

!CogVMSimulator methodsFor: 'control primitives' stamp: 'eem 1/11/2013 18:33'!
forShortPrintString: shortPrintString filterPerformMessages: aCollection
	performFilters ifNil:
		[performFilters := Dictionary new].
	performFilters at: shortPrintString put: aCollection! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
forceInterruptCheckFromHeartbeat
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #forceInterruptCheckFromHeartbeat
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
frameCallerContext: theFP
	"In the StackInterpreter the saved ip field of a base frame holds the base
	 frame's caller context. But in the Cog VM the first word on the stack holds
	 the base frame's caller context, which is immediately above the stacked
	 receiver.  The asserts using frameStackedReceiverOffset: are simulation
	 only since they depend on being able to access numArgs and frameContext
	 from the frame's method and in a base return the frame state, being below
	 the stack pointer, may have already been smashed by an interrupt."
	| thePage |
	self assert: (self isBaseFrame: theFP).
	thePage := stackPages stackPageFor: theFP.
	self assert: theFP + (self frameStackedReceiverOffset: theFP) + (2 * objectMemory wordSize) = thePage baseAddress.
	self assert: (stackPages longAt: thePage baseAddress - objectMemory wordSize) = (self frameContext: theFP).
	^super frameCallerContext: theFP! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
frameOfMarriedContext: aContext 
	| senderOop |
	senderOop := objectMemory fetchPointer: SenderIndex ofObject: aContext.
	self assert: (objectMemory isIntegerObject: senderOop).
	^self withoutSmallIntegerTags: senderOop! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 1/8/2009 16:39'!
framePointer
	^framePointer! !

!CogVMSimulator methodsFor: 'I/O primitives' stamp: 'eem 5/12/2017 11:06'!
fullDisplay
	| primFailCodeValue |
	displayForm ifNil: [^self].
	primFailCodeValue := primFailCode.
	self initPrimCall.
	self updateDisplayLeft: 0 Top: 0 Right: displayWidth Bottom: displayHeight.
	primFailCode := primFailCodeValue! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 11/25/2008 13:32'!
fullDisplayUpdate
	"Preserve self successful when call asynchronously from Simulator"
	| primFailCodeValue |
	primFailCodeValue := primFailCode.
	self initPrimCall.
	super fullDisplayUpdate.
	primFailCode := primFailCodeValue! !

!CogVMSimulator methodsFor: 'plugin primitives' stamp: 'eem 4/8/2009 10:43'!
functionForPrimitiveExternalCall: methodObj
	"Override to jump through the plugin simulation hoops."
	| symbolOrIndex |
	symbolOrIndex := super functionForPrimitiveExternalCall: methodObj.
	^symbolOrIndex isSymbol
		ifTrue: [symbolOrIndex]
		ifFalse: [(self pluginEntryFor: symbolOrIndex) at: 3]! !

!CogVMSimulator methodsFor: 'interpreter shell' stamp: 'eem 6/24/2008 16:00'!
functionPointerFor: primIndex inClass: lookupClass
	"Override Interpreter to handle the external primitives caching.  See also
	 internalExecuteNewMethod."

	^(primIndex between: 1 and: MaxPrimitiveIndex)
		ifTrue: [primitiveTable at: primIndex + 1]
		ifFalse: [0]! !

!CogVMSimulator methodsFor: 'cog jit support' stamp: 'eem 7/24/2021 16:51'!
functionPointerForCompiledMethod: methodObj primitiveIndex: primIndex primitivePropertyFlagsInto: flagsPtr
	^self mapFunctionToAddress: (super functionPointerForCompiledMethod: methodObj primitiveIndex: primIndex primitivePropertyFlagsInto: flagsPtr)! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 10/08/2020 11:20'!
getImageHeaderFlags
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #getImageHeaderFlags
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
getMaxWaitingPriority
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #getMaxWaitingPriority
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 1/2/2021 21:56'!
getReturnAddress
	"In the real VM this answers the return address for its caller, i.e. for interpret.
	 In the simulator we're playing fast and loose with initialEnterSmalltalkExecutive
	 and enterSmalltalkExecutiveImplementation and need them to look and act the same."
	| selector |
	selector := thisContext getReturnAddress.
	^(selector == #initialEnterSmalltalkExecutive
	  or: [selector == #enterSmalltalkExecutive])
		ifTrue: [#enterSmalltalkExecutiveImplementation]
		ifFalse: [selector]! !

!CogVMSimulator methodsFor: 'image save/restore' stamp: 'eem 10/10/2013 06:23'!
getShortFromFile: aFile swap: swapFlag
	| aShort |
	aShort := self nextShortFrom: aFile.
	^swapFlag 
		ifTrue: [(aShort bitShift: -8) + ((aShort bitAnd: 16rFF) bitShift: 8)]
		ifFalse: [aShort]! !

!CogVMSimulator methodsFor: 'memory access' stamp: 'di 7/16/2004 14:56'!
halfWordHighInLong32: long32
	^self subclassResponsibility! !

!CogVMSimulator methodsFor: 'memory access' stamp: 'di 7/16/2004 14:57'!
halfWordLowInLong32: long32
	^self subclassResponsibility! !

!CogVMSimulator methodsFor: 'message sending' stamp: 'eem 8/15/2011 10:21'!
handleMNU: selectorIndex InMachineCodeTo: rcvr classForMessage: classForMessage
	selectorIndex = SelectorCannotInterpret ifTrue:
		[self halt].
	^super handleMNU: selectorIndex InMachineCodeTo: rcvr classForMessage: classForMessage! !

!CogVMSimulator methodsFor: 'plugin primitive support' stamp: 'eem 8/20/2021 23:28'!
hasFastCLinkage: methodObj
	"Simulation only!!!! the attempt to look up is entirely inappropriate for the real VM!!!!"
	| header primIdx literal metadata |
	objectMemory hasSpurMemoryManagerAPI ifFalse:
		[^false].
	header := objectMemory methodHeaderOf: methodObj.
	primIdx := self primitiveIndexOfMethod: methodObj header: header.
	(primIdx = PrimNumberExternalCall
	 and: [(objectMemory literalCountOfMethodHeader: header) > 0
	 and: [(objectMemory isArray: (literal := self literal: 0 ofMethod: methodObj))
	 and: [(objectMemory numSlotsOf: literal) = 4]]]) ifFalse:
		[^primIdx = PrimNumberObjectAtPut].
	(objectMemory fetchPointer: ExternalCallLiteralTargetFunctionIndex ofObject: literal) ~= (objectMemory integerObjectOf: 0) ifTrue:
		[^FastCPrimitiveFlag anyMask: (objectMemory integerValueOf: (objectMemory fetchPointer: ExternalCallLiteralFlagsIndex ofObject: literal))].
	^(self ioLoadFunction: (self stringOf: (objectMemory fetchPointer: ExternalCallLiteralFunctionNameIndex ofObject: literal))
		From: (self stringOf: (objectMemory fetchPointer: ExternalCallLiteralModuleNameIndex ofObject: literal))
		MetadataInto: (self addressOf: metadata put: [:v| metadata := v])) ~= 0
	 and: [FastCPrimitiveFlag anyMask: metadata]! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
headerStart: oop

	^ (objectMemory extraHeaderBytes: oop) negated! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 10/1/2013 17:01'!
heapMapAtWord: address Put: aBit
	^objectMemory heapMap heapMapAtWord: address asUnsignedInteger Put: aBit! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 2/7/2009 08:40'!
ifAppropriateCompileToNativeCode: aMethodObject selector: selector
	enableCog ifTrue:
		[super ifAppropriateCompileToNativeCode: aMethodObject selector: selector]! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 8/6/2021 13:27'!
ifTestProfilingAdvanceNextProfileTick
	"bump nextProfileTick along if simulating; but not by much so the mechanism is stressed a lot"
	(InitializationOptions at: #profiling ifAbsent: [false]) ifTrue:
		[nextProfileTick := self ioHighResClock + ((InitializationOptions at: #profileIncrement ifAbsent: [10000])).
		 self assert: nextProfileTick > (self ioHighResClock + 5000)]! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem 12/31/2008 12:06'!
iframeIsBlockActivation: theFP
	self assert: (self isMachineCodeFrame: theFP) not.
	^super iframeIsBlockActivation: theFP! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem 12/31/2008 12:06'!
iframeMethod: theFP
	self assert: (self isMachineCodeFrame: theFP) not.
	^super iframeMethod: theFP! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem 12/31/2008 12:06'!
iframeNumArgs: theFP
	self assert: (self isMachineCodeFrame: theFP) not.
	^super iframeNumArgs: theFP! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem 12/31/2008 12:06'!
iframeSavedIP: theFP
	self assert: (self isMachineCodeFrame: theFP) not.
	^super iframeSavedIP: theFP! !

!CogVMSimulator methodsFor: 'image save/restore' stamp: 'eem 3/9/2014 08:38'!
imageName
	^imageName! !

!CogVMSimulator methodsFor: 'spur bootstrap' stamp: 'eem 9/30/2013 16:15'!
imageName: aString
	imageName := aString! !

!CogVMSimulator methodsFor: 'file primitives' stamp: 'eem 10/8/2013 17:00'!
imageNameGet: p Length: sz
	1 to: sz  do:
		[:i |
		objectMemory
			byteAt:  p + i - 1
			put: (imageName at: i) asInteger]! !

!CogVMSimulator methodsFor: 'file primitives' stamp: 'eem 9/7/2018 18:47'!
imageNamePut: p Length: sz
	| newName window |
	newName := ByteString new: sz.
	1 to: sz  do:
		[:i |
		newName
			at: i
			put: (Character value: (objectMemory byteAt: p + i - 1))].
	imageName := newName.
	(displayView notNil
	 and: [(window := displayView containingWindow) notNil]) ifTrue:
		[window setLabel: (window label
							copyReplaceFrom: (window label lastIndexOf: Character space) + 1
							to: window label size
							with: (self localNameFor: imageName))]! !

!CogVMSimulator methodsFor: 'file primitives' stamp: 'eem 10/8/2013 16:36'!
imageNameSize
	^imageName size! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
inGUIThread
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #inGUIThread
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'rump c stack' stamp: 'eem 2/8/2009 11:39'!
inMemoryStackLimitAddress
	^self rumpCStackAddress - 24! !

!CogVMSimulator methodsFor: 'interpreter shell' stamp: 'eem 1/5/2018 15:22'!
incrementByteCount
	(byteCount := byteCount + 1) = breakCount ifTrue:
		[self doOrDefer: [self changed: #byteCountText; changed: #composeAll].
		 self halt: 'breakCount reached'].
	byteCount \\ 1000 = 0 ifTrue:
		[self doOrDefer: [self changed: #byteCountText; changed: #composeAll].
		 self forceInterruptCheck]! !

!CogVMSimulator methodsFor: 'test support' stamp: 'eem 1/25/2013 06:33'!
initStackPagesForTests
	numStackPages := 8.
	stackPages := self stackPagesClass new.
	objectMemory allocateMemoryOfSize: 1024 * 1024.
	cogCodeSize := 0.
	heapBase := self methodCacheSize
				+ self primTraceLogSize
				+ self rumpCStackSize
				+ self computeStackZoneSize.
	self initStackPages! !

!CogVMSimulator methodsFor: 'initialize-release' stamp: 'eem 10/30/2020 18:14'!
initialEnterSmalltalkExecutive
	"Main entry-point into the interpreter at system start-up.
	 Override to choose between the threaded and non-threaded versions and if threaded
	 to ensure that the switch method overrides are up-to-date."
	self ensureMultiThreadingOverridesAreUpToDate.
	self assert: (cogit processor fp = CFramePointer and: [cogit processor sp = CStackPointer]).
	^self perform: #initialEnterSmalltalkExecutive
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'initialize-release' stamp: 'eem 8/6/2021 13:24'!
initialize
	"Initialize the CogVMSimulator when running the interpreter inside Smalltalk.  The
	 primary responsibility of this method is to allocate Smalltalk Arrays for variables
	 that will be declared as statically-allocated global arrays in the translated code."
	super initialize.

	transcript := Transcript.

	objectMemory ifNil:
		[objectMemory := self class objectMemoryClass simulatorClass new].
	cogit ifNil:
		[cogit := self class cogitClass new setInterpreter: self].
	objectMemory coInterpreter: self cogit: cogit.

	(cogit numRegArgs > 0
	 and: [VMClass initializationOptions at: #CheckStackDepth ifAbsent: [true]]) ifTrue:
		[debugStackDepthDictionary := Dictionary new].

	cogThreadManager ifNotNil:
		[super initialize].

	self assert: ConstMinusOne = (objectMemory integerObjectOf: -1).

	cogMethodZone := cogit methodZone. "Because Slang can't remove intermediate implicit receivers (cogit methodZone foo doesn't reduce to foo())"
	enableCog := true.

	methodCache := Array new: MethodCacheSize.
	nsMethodCache := Array new: NSMethodCacheSize.
	atCache := nil.
	self flushMethodCache.
	cogCompiledCodeCompactionCalledFor := false.
	gcSemaphoreIndex := 0.
	externalSemaphoreSignalRequests := externalSemaphoreSignalResponses := #().
	externalPrimitiveTable := CArrayAccessor on: (Array new: MaxExternalPrimitiveTableSize).
	externalPrimitiveTableFirstFreeIndex := 0.
	primitiveTable := self class primitiveTable copy.
	self initializePluginEntries.
	desiredNumStackPages := InitializationOptions at: #desiredNumStackPages ifAbsent: [0].
	desiredEdenBytes := InitializationOptions at: #desiredEdenBytes ifAbsent: [0].
	desiredCogCodeSize  := InitializationOptions at: #desiredCogCodeSize ifAbsent: [0].
	"This is initialized on loading the image, but convenient for testing stack page values..."
	numStackPages := self defaultNumStackPages. 
	startMicroseconds := lastYieldMicroseconds := self ioUTCStartMicroseconds.
	maxLiteralCountForCompile := MaxLiteralCountForCompile.
	minBackwardJumpCountForCompile := MinBackwardJumpCountForCompile.
	flagInterpretedMethods := false.

	"initialize InterpreterSimulator variables used for debugging"
	byteCount := lastPollCount := sendCount := lookupCount := 0.
	quitBlock := [^self close].
	traceOn := true.
	printSends := printFrameAtEachStep := printBytecodeAtEachStep := false.
	myBitBlt := BitBltSimulator new setInterpreter: self.
	displayForm := fakeForm := 'Display has not yet been installed' asDisplayText form.
	suppressHeartbeatFlag := deferSmash := deferredSmash := false.
	systemAttributes := Dictionary new.
	primTraceLog := CArrayAccessor on: (Array new: 256 withAll: 0).
	primTraceLogIndex := 0.
	traceLog := CArrayAccessor on: (Array new: TraceBufferSize withAll: 0).
	traceLogIndex := 0.
	traceSources := TraceSources.
	statCodeCompactionCount := 0.
	statCodeCompactionUsecs := 0.
	extSemTabSize := 256.
	zeroNextProfileTickCount := 0! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
initializeInterpreter: bytesToShift
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #initializeInterpreter:
		withArguments: {bytesToShift}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'plugin support' stamp: 'eem 7/25/2021 16:11'!
initializePluginEntries
	mappedPluginEntries := OrderedCollection new.
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[primitiveAccessorDepthTable := Array new: primitiveTable size.
			 pluginList := {}.
			 self loadNewPlugin: '']
		ifFalse:
			[pluginList := {'' -> self }]! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
initializeProcessorForThreadIndex: threadIndex
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #initializeProcessorForThreadIndex:
		withArguments: {threadIndex}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'initialize-release' stamp: 'eem 1/3/2021 09:12'!
initializeThreadSupport
	"Do this post new if you want to simulate with thread support."
	cogThreadManager := CogThreadManager new setInterpreter: self cogit: cogit! !

!CogVMSimulator methodsFor: 'I/O primitive support' stamp: 'eem 10/23/2020 14:30'!
inputSemaphoreIndex
	^inputSemaphoreIndex! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 4/17/2014 15:30'!
instVar: offset ofContext: aContext

	"offset = InstructionPointerIndex ifTrue:
		[Transcript nextPutAll: '==================='; cr.
		  self printContext: 16r1715630.
		 self printCallStackOf: aContext currentFP: framePointer.
		 Transcript nextPutAll: '==================='; cr.
		 self halt]."

	| result |
	"self shortPrintFrameAndCallers: localFP.
	Transcript print: byteCount; tab; print: thisContext; cr.
	self print: offset; cr.
	self printContext: aContext.
	(self confirm: 'continue?') ifFalse: [self halt]."
	result := super instVar: offset ofContext: aContext.
	"(offset = InstructionPointerIndex
	and: [(self stringOf: (self penultimateLiteralOf: (self fetchPointer: MethodIndex ofObject: aContext))) = #indexOf:startingAt:ifAbsent:]) ifTrue:
		[transcript space; nextPutAll: '^pc '; nextPutAll: result hex.
		 (objectMemory isIntegerObject: result) ifTrue:
			[transcript space; print: (objectMemory integerValueOf: result)].
		 transcript tab.
		 self shortPrintContext: aContext]."
	"offset = StackPointerIndex ifTrue:
		[Transcript nextPutAll: '^stackp ', (self integerValueOf: result) printString; tab.
		 self shortPrintContext: aContext.
		 (#(24205456 24205732) includes: aContext) ifTrue:
		 	[(self checkIsStillMarriedContext: aContext currentFP: localFP)
				ifTrue: [self printFrame: (self frameOfMarriedContext: aContext) WithSP: (self frameOfMarriedContext: aContext) - 48]
				ifFalse: [self printContext: aContext]]]."
	^result! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem 4/16/2014 09:56'!
instVar: index ofContext: aMarriedContext put: anOop
	"self shortPrintFrameAndCallers: localFP.
	Transcript print: byteCount; tab; print: thisContext; cr.
	self print: index; cr.
	self printContext: aMarriedContext.
	self shortPrintOop: anOop.
	(self confirm: 'continue?') ifFalse: [self halt]."
	"(self stringOf: (self penultimateLiteralOf: (self fetchPointer: MethodIndex ofObject: aMarriedContext))) = #indexOf:startingAt:ifAbsent: ifTrue:
		[self halt]."
	^super instVar: index ofContext: aMarriedContext put: anOop! !

!CogVMSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:01'!
insufficientMemoryAvailableError
	self error: 'Failed to allocate memory for the heap'! !

!CogVMSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:01'!
insufficientMemorySpecifiedError
	self error: 'Insufficient memory for this image'! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 6/27/2008 14:13'!
internalCannotReturn: resultOop
	self halt.
	^super internalCannotReturn: resultOop! !

!CogVMSimulator methodsFor: 'testing' stamp: 'rmacnak 5/18/2015 21:15'!
internalFindNewMethodOrdinary
"
	| cName |
	traceOn ifTrue:
		[cName := (self sizeBitsOf: class) = 16r20
			ifTrue: ['class ' , (self nameOfClass: (self fetchPointer: 6 ofObject: class))]
			ifFalse: [(self nameOfClass: class)].
		self cr; print: cName , '>>' , (self stringOf: messageSelector)].
"
	messageSelector = (objectMemory splObj: SelectorDoesNotUnderstand) ifTrue: [self halt].

	self logSend: messageSelector.
"
	(sendCount > 1000 and: [sendCount\\10 = 0]) ifTrue:
		[Transcript print: sendCount; space.
		self validate].
"
"
	(sendCount > 100150) ifTrue:
		[self qvalidate.
		messageQueue == nil ifTrue: [messageQueue := OrderedCollection new].
		messageQueue addLast: (self stringOf: messageSelector)].
"
	^super internalFindNewMethodOrdinary! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 6/27/2008 14:13'!
internalMustBeBoolean
	self halt.
	^super internalMustBeBoolean! !

!CogVMSimulator methodsFor: 'I/O primitives' stamp: 'eem 5/12/2017 17:24'!
ioBeDisplay: bitsOrHandleCArray width: width height: height depth: depth
	"The real platform may want to note the location of the display and respond to the fact of
	 its installation.  Here merely create a scratch form the same size as Smalltalk displayObj"
	super ioBeDisplay: bitsOrHandleCArray asInteger width: width height: height depth: depth.
	displayForm := Form extent: width @ height depth: depth.
	displayView ifNotNil: [displayView image: displayForm]! !

!CogVMSimulator methodsFor: 'security' stamp: 'ar 2/5/2001 18:33'!
ioCanRenameImage
	^true! !

!CogVMSimulator methodsFor: 'security' stamp: 'ar 2/5/2001 20:42'!
ioCanWriteImage
	^true! !

!CogVMSimulator methodsFor: 'primitive support' stamp: 'eem 11/10/2020 12:00'!
ioExit
	self ioExitWithErrorCode: 0! !

!CogVMSimulator methodsFor: 'primitive support' stamp: 'eem 11/10/2020 12:24'!
ioExitWithErrorCode: ec
	self ioShutdownAllModules.
	self threadManager ifNotNil:
		[:threadManager| threadManager shutdownModule].
	quitBlock value  "Cause return from #test"! !

!CogVMSimulator methodsFor: 'process primitive support' stamp: 'eem 2/3/2010 12:56'!
ioGetMaxExtSemTableSize
	^extSemTabSize! !

!CogVMSimulator methodsFor: 'I/O primitives' stamp: 'ar 5/1/1999 09:43'!
ioHasDisplayDepth: depth
	^Display supportsDisplayDepth: depth! !

!CogVMSimulator methodsFor: 'I/O primitives support' stamp: 'eem 8/22/2008 11:18'!
ioHeartbeatMilliseconds
	^1! !

!CogVMSimulator methodsFor: 'I/O primitives support' stamp: 'eem 8/6/2021 11:03'!
ioHighResClock
	^cogit backEnd has64BitPerformanceCounter
		ifTrue: [cogit ioHighResClock]
		ifFalse: [self ioUTCMicroseconds]! !

!CogVMSimulator methodsFor: 'initialize-release' stamp: 'eem 8/3/2008 21:48'!
ioInitHeartbeat
	"No-op in the simulator.  We cause a poll every 1000 bytecodes instead."! !

!CogVMSimulator methodsFor: 'simulation support' stamp: 'eem 7/25/2021 15:38'!
ioListBuiltinModule: index
	^index > 0 ifTrue:
		[(InitializationOptions
			at: #builtinModules
			ifAbsent: [#('BitBltPlugin' 'FloatArrayPlugin' 'FloatMathPlugin' 'LargeIntegers' 'MiscPrimitivePlugin')]) at: index ifAbsent: nil]! !

!CogVMSimulator methodsFor: 'simulation support' stamp: 'eem 7/25/2021 15:21'!
ioListLoadedModule: index
	^index > 0 ifTrue: [(pluginList at: index + 1 ifAbsent: [^nil]) key]! !

!CogVMSimulator methodsFor: 'plugin support' stamp: 'eem 10/29/2018 15:39'!
ioLoadFunction: functionString From: pluginString
	"Load and return the requested function from a module"
	| firstTime plugin fnSymbol |
	firstTime := false.
	fnSymbol := functionString asSymbol.
	transcript
		cr;
		show: '(', byteCount printString, ') Looking for ', functionString, ' in ',
				(pluginString isEmpty ifTrue: ['vm'] ifFalse: [pluginString]).
	(breakSelector notNil
	 and: [(pluginString size = breakSelector size
			and: [(self strncmp: pluginString _: breakSelector _: pluginString size) = 0])
		or: [functionString size = breakSelector size
			and: [(self strncmp: functionString _: breakSelector _: functionString size) = 0]]]) ifTrue:
		[self halt: pluginString, '>>', functionString].
	plugin := pluginList 
				detect: [:any| any key = pluginString asString]
				ifNone:
					[firstTime := true.
					 self loadNewPlugin: pluginString].
	plugin ifNil:
		[firstTime ifTrue: [transcript cr; show: '... FAILED; no plugin found'].
		 ^0].
	plugin := plugin value.
	mappedPluginEntries withIndexDo:
		[:pluginAndName :index|
		((pluginAndName at: 1) == plugin 
		and:[(pluginAndName at: 2) == fnSymbol]) ifTrue:
			[^index]].
	(plugin respondsTo: fnSymbol) ifFalse:
		[firstTime ifTrue: [transcript cr; show: '... FAILED; primitive not in plugin'].
		 ^0].
	mappedPluginEntries addLast: { plugin. fnSymbol. [plugin perform: fnSymbol. self]}.
	transcript show: ' ... okay'.
	^mappedPluginEntries size! !

!CogVMSimulator methodsFor: 'plugin support' stamp: 'eem 7/22/2021 17:40'!
ioLoadFunction: functionString From: pluginString MetadataInto: metadataPtr
	"Load and return the requested function from a module.
	 Assign the accessor depth and flags through metadataPtr.
	 N.B. The actual code lives in platforms/Cross/vm/sqNamedPrims.h"
	| firstTime plugin fnSymbol |
	firstTime := false.
	fnSymbol := functionString asSymbol.
	transcript
		cr;
		show: '(', byteCount printString, ') Looking for ', functionString, ' in ',
				(pluginString isEmpty ifTrue: ['vm'] ifFalse: [pluginString]).
	(breakSelector notNil
	 and: [(pluginString size = breakSelector size
			and: [(self strncmp: pluginString _: breakSelector _: pluginString size) = 0])
		or: [functionString size = breakSelector size
			and: [(self strncmp: functionString _: breakSelector _: functionString size) = 0]]]) ifTrue:
		[self halt: pluginString, '>>', functionString].
	plugin := pluginList 
				detect: [:any| any key = pluginString asString]
				ifNone:
					[firstTime := true.
					 self loadNewPlugin: pluginString].
	plugin ifNil:
		[firstTime ifTrue: [transcript cr; show: '... FAILED; primitive not in plugin'].
		 ^0].
	plugin := plugin value.
	mappedPluginEntries withIndexDo:
		[:pluginAndName :index|
		 ((pluginAndName at: 1) == plugin 
		  and:[(pluginAndName at: 2) == fnSymbol]) ifTrue:
			[firstTime ifTrue: [transcript show: ' ... okay'].
			 metadataPtr at: 0 put: (pluginAndName at: 4).
			 ^index]].
	firstTime ifTrue: [transcript cr; show: '... FAILED; primitive not in plugin'].
	^0! !

!CogVMSimulator methodsFor: 'I/O primitives support' stamp: 'eem 10/3/2013 04:33'!
ioMSecs
	"Return the value of the millisecond clock."
	"NOT.  Actually, we want something a lot slower and, for exact debugging,
	something more repeatable than real time.  I have an idea: use the byteCount... (di 7/1/2004 13:55)"

	^self microsecondsToMilliseconds: self ioUTCMicroseconds
	
"At 20k bytecodes per second, this gives us about 200 ticks per second, or about 1/5 of what you'd expect for the real time clock.  This should still service events at one or two per second"! !

!CogVMSimulator methodsFor: 'I/O primitives support' stamp: 'jm 3/9/1999 10:00'!
ioMicroMSecs
	"Answer the value of the high-resolution millisecond clock."

	^ Time millisecondClockValue
! !

!CogVMSimulator methodsFor: 'I/O primitives' stamp: 'eem 3/4/2015 17:05'!
ioMousePoint
	| relPt |
	^displayForm
		ifNil: [99 << 16 + 66]
		ifNotNil: [relPt := Sensor cursorPoint - self displayLocation.
				relPt x << 16 + relPt y]! !

!CogVMSimulator methodsFor: 'I/O primitives support' stamp: 'eem 10/23/2020 13:51'!
ioRelinquishProcessorForMicroseconds: microseconds
	"In the simulator give an indication that we're idling and check for input.
	 If there's an input event, fail to give the system a chance to respond to it in a timely manner.
	 If called from machine code then increment the byte count since the clock
	 is derived from it and the clock will not advance otherwise.
	 If we're simulating threading we're in difficulties.  We need a UI process
	 (to run activities such as fill-in-the-blanks) but we also need an independent
	 thread of control to run this VM thread.  So we need to fork a new UI process."
	Display reverse: ((displayView
						ifNil: [0@0]
						ifNotNil: [displayView bounds origin]) extent: 16@16).
	Processor activeProcess == Project uiProcess ifTrue:
		[World doOneCycle].
	(stackLimit = self allOnesAsCharStar
	 and: [nextPollUsecs <= self ioUTCMicroseconds])
		ifTrue:
			["Not only do we need to fail, we also need to push time faster
			 since the damn Morphic eventTickler process waits on a delay."
			"ioUTCMicroseconds"
			byteCount := byteCount + (1000 * ByteCountsPerMicrosecond).
			self primitiveFail]
		ifFalse:
			[microseconds >= 1000
				ifTrue: [self isThreadedVM ifTrue:
							[self forceInterruptCheckFromHeartbeat].
						(Delay forMilliseconds: microseconds + 999 // 1000) wait]
				ifFalse: [Processor yield]].
	byteCount := byteCount + (microseconds * ByteCountsPerMicrosecond) - 1.
	self incrementByteCount! !

!CogVMSimulator methodsFor: 'I/O primitives' stamp: 'ar 2/5/2001 17:24'!
ioScreenDepth
	^DisplayScreen actualScreenDepth.! !

!CogVMSimulator methodsFor: 'I/O primitives support' stamp: 'topa 7/10/2016 23:14'!
ioScreenScaleFactor
	^ 1.0! !

!CogVMSimulator methodsFor: 'I/O primitives support' stamp: 'eem 3/11/2015 19:27'!
ioScreenSize
	"Answer the screen extent packed into 32 bits.  In the simulator,
	 displayForm gets initialized with a fake form; don't be deceived."
	| extent |
	extent := (displayForm notNil and: [displayForm ~~ fakeForm])
				ifTrue: [displayForm extent]
				ifFalse: [self desiredDisplayExtent].
	^extent x << 16 + extent y! !

!CogVMSimulator methodsFor: 'I/O primitives' stamp: 'eem 10/23/2020 12:49'!
ioSetInputSemaphore: index
	inputSemaphoreIndex := index! !

!CogVMSimulator methodsFor: 'process primitive support' stamp: 'eem 2/3/2010 12:57'!
ioSetMaxExtSemTableSize: anInteger
	"Inform the VM of the maximum size the image will grow the exernal semaphore table (specialObjectsArray at: 39) to.
	 This allows the VM to allocate storage such that external semaphores can be signalled without locking."
	extSemTabSize := anInteger! !

!CogVMSimulator methodsFor: 'I/O primitives' stamp: 'eem 5/12/2017 17:13'!
ioShow: destBits D: w i: h s: d p: left l: right a: top y: bottom
	"This is the simulator's implementation of ioShowDisplay."
	| raster pixPerWord simDisp realDisp rect |
	displayForm ifNil: [^self].
	displayBits = 0 ifTrue: [^self].
	pixPerWord := 32 // d.
	raster := displayForm width + (pixPerWord - 1) // pixPerWord.
	simDisp := Form new hackBits: objectMemory memory.
	displayForm unhibernate.
	realDisp := Form new hackBits: displayForm bits.
	realDisp
		copy: (0 @ (top * raster) extent: 4 @ (bottom - top * raster))
		from: 0 @ (destBits // 4 + (top * raster))
		in: simDisp
		rule: Form over.
	displayView ifNotNil: [^displayView changed].
	
	"If running without a view, just blat the bits onto the screen..."
	rect := 0 @ top corner: displayForm width @ bottom.
	Display
		copy: (rect translateBy: self displayLocation)
		from: rect topLeft
		in: displayForm
		rule: Form over! !

!CogVMSimulator methodsFor: 'initialize-release' stamp: 'eem 11/10/2020 12:24'!
ioShutdownAllModules
	pluginList do:
		[:assoc| | pluginOrSelf |
		 ((pluginOrSelf := assoc value) ~~ self
		  and: [pluginOrSelf respondsTo: #shutdownModule]) ifTrue:
			[pluginOrSelf shutdownModule]]! !

!CogVMSimulator methodsFor: 'process primitive support' stamp: 'eem 2/4/2010 09:37'!
ioSynchronousCheckForEvents
	"Hook allowing the platform to do anything it needs to do synchronously."! !

!CogVMSimulator methodsFor: 'I/O primitives support' stamp: 'eem 5/27/2021 10:15'!
ioUTCMicroseconds
	"Return the value of the microsecond clock."
	"NOT.  Actually, we want something a lot slower and, for exact debugging,
	 something more repeatable than real time.  Dan had an idea: use the byteCount...
	 We increment byteCount in stackLimitFromMachineCode and a real machine
	 can easily run e.g. nfib at 6e7 / second, which this would be 1 usec ~= 60 byteCounts.
	 Use 100 byteCounts per usec by default; see CogVMSimulator class>>initializeWithOptions:"

	^byteCount // ByteCountsPerMicrosecond + startMicroseconds
	
	"Dan:
	 At 20k bytecodes per second, this gives us about 200 ticks per second, or about 1/5
	 of what you'd expect for the real time clock.  This should still service events at one or
	 two per second"! !

!CogVMSimulator methodsFor: 'I/O primitives support' stamp: 'eem 1/25/2016 18:08'!
ioUTCStartMicroseconds
	"Answer the value of the microsecond clock at startup."

	^startMicroseconds ifNil:
		[self class initializationOptions
			at: #startMicroseconds
			ifAbsent: [[Time utcMicrosecondClock] on: Error do: [Time totalSeconds * 1000000]]]! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
isAffinedProcess: aProcess
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #isAffinedProcess:
		withArguments: {aProcess}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
isAffinedThreadId: threadId
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #isAffinedThreadId:
		withArguments: {threadId}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
isBoundProcess: aProcess
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #isBoundProcess:
		withArguments: {aProcess}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
isBoundThreadId: threadId
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #isBoundThreadId:
		withArguments: {threadId}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'rump c stack' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
isOnRumpCStack: address
	^address between: heapBase - self rumpCStackSize and: heapBase - objectMemory wordSize! !

!CogVMSimulator methodsFor: 'primitive support' stamp: 'eem 9/11/2010 12:57'!
isPrimitiveFunctionPointerAnIndex
	"We save slots in the method cache by using the primitiveFunctionPointer
	 to hold either a function pointer or the index of a quick primitive. Since
	 quick primitive indices are small they can't be confused with function
	 addresses.  But since we use 1001 and up for external primitives that
	 would be functions in the C VM but are indices under simulation we treat
	 values above 1000 as if they were pointers (actually indices into the
	 externalPrimitiveTable)"

	^primitiveFunctionPointer isInteger
	  and: [primitiveFunctionPointer >= 0
	  and: [primitiveFunctionPointer <= MaxQuickPrimitiveIndex]]! !

!CogVMSimulator methodsFor: 'testing' stamp: 'eem 12/25/2010 14:21'!
isThreadedVM
	^self threadManager notNil! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
isUnboundThreadId: threadId
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #isUnboundThreadId:
		withArguments: {threadId}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem 9/30/2010 13:29'!
isWidowedContext: aOnceMarriedContext
	"See if the argument is connected with a live frame or not.
	 If it is not, turn it into a bereaved single context."
	(stackPages isFree: stackPage) ifFalse: "in baseReturn the active page has been freed."
		[((stackPages somePageHasHeadFrameFP: framePointer)
		  or: [(stackPages somePageHasHeadFrameFP: localFP)
		  or: [stackPages allPagesFree]]) ifFalse:
			[(thisContext sender sender method = (CoInterpreter >> #baseFrameReturn)
			  or: [thisContext sender sender method = (CoInterpreter >> #ceBaseFrameReturn:)]) ifFalse:
				[self halt: 'currentFP may not be written back to stack page']]].
	^super isWidowedContext: aOnceMarriedContext! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
loadInitialContext
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #loadInitialContext
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'plugin support' stamp: 'eem 8/20/2021 16:11'!
loadNewPlugin: pluginString
	(breakSelector notNil
	 and: [pluginString size = breakSelector size
	 and: [(self strncmp: pluginString _: breakSelector _: pluginString size) = 0]]) ifTrue:
		[self halt: pluginString].
	^(self tryLoadNewPlugin: pluginString pluginEntries: mappedPluginEntries) ifNotNil:
		[:entry|
		 pluginList := pluginList copyWith: entry.
		 entry]! !

!CogVMSimulator methodsFor: 'accessing' stamp: 'eem 1/23/2017 09:28'!
localIP
	^localIP! !

!CogVMSimulator methodsFor: 'testing' stamp: 'eem 12/29/2016 08:28'!
logOfBytesVerify: nBytes fromFileNamed: fileName fromStart: loggingStart
	"Verify a questionable interpreter against a successful run"
	"self logOfBytesVerify: 10000 fromFileNamed: 'clone32Bytecodes.log' "
	
	| logFile rightWord prevCtxt |
	logFile := (FileStream readOnlyFileNamed: fileName) binary.
	transcript clear.
	byteCount := 0.
	quitBlock := [^self close].
	self initStackPages.
	self loadInitialContext.
	self internalizeIPandSP.
	self fetchNextBytecode.
	prevCtxt := 0.  prevCtxt := prevCtxt.
	[byteCount < nBytes] whileTrue:
		[
"
byteCount > 14560 ifTrue:
[self externalizeIPandSP.
prevCtxt = activeContext ifFalse:
 [prevCtxt := activeContext.
 transcript cr; nextPutAll: (self printTop: 2); endEntry].
transcript cr; print: byteCount; nextPutAll: ': ' , (activeContext hex); space;
 print: (instructionPointer - method - (BaseHeaderSize - 2));
 nextPutAll: ': <' , (self byteAt: localIP) hex , '>'; space;
 nextPutAll: (self symbolic: currentBytecode at: localIP inMethod: method); space;
 print: (self stackPointerIndex - TempFrameStart + 1); endEntry.
byteCount = 14590 ifTrue: [self halt]].
"
		loggingStart >= byteCount ifTrue:
			[rightWord := logFile nextWord.
			 currentBytecode = rightWord ifFalse:
				[self halt: 'halt at ', byteCount printString]].
		self dispatchOn: currentBytecode in: BytecodeTable.
		self incrementByteCount].
	self externalizeIPandSP.
	logFile close.
	self inform: nBytes printString , ' bytecodes verfied.'! !

!CogVMSimulator methodsFor: 'testing' stamp: 'eem 12/29/2016 08:28'!
logOfBytesWrite: nBytes toFileNamed: fileName fromStart: loggingStart
	"Write a log file for testing a flaky interpreter on the same image"
	"self logOfBytesWrite: 10000 toFileNamed: 'clone32Bytecodes.log' "
	
	| logFile |
	logFile := (FileStream newFileNamed: fileName) binary.
	transcript clear.
	byteCount := 0.
	quitBlock := [^self close].
	self initStackPages.
	self loadInitialContext.
	self internalizeIPandSP.
	self fetchNextBytecode.
	[byteCount < nBytes] whileTrue:
		[byteCount >= loggingStart ifTrue:
			[logFile nextWordPut: currentBytecode].
		self dispatchOn: currentBytecode in: BytecodeTable.
		self incrementByteCount].
	self externalizeIPandSP.
	logFile close! !

!CogVMSimulator methodsFor: 'testing' stamp: 'eem 12/29/2016 08:27'!
logOfSendsVerify: nSends fromFileNamed: fileName fromStart: loggingStart
	"Write a log file for testing a flaky interpreter on the same image"
	"self logOfSendsWrite: 10000 toFileNamed: 'clone32Messages.log' "
	
	| logFile priorFrame rightSelector prevCtxt |
	logFile := FileStream readOnlyFileNamed: fileName.
	transcript clear.
	byteCount := 0.
	sendCount := 0.
	priorFrame := localFP.
	quitBlock := [^self close].
	self initStackPages.
	self loadInitialContext.
	self internalizeIPandSP.
	self fetchNextBytecode.
	prevCtxt := 0.  prevCtxt := prevCtxt.
	[sendCount < nSends] whileTrue:
		[
"
byteCount>500 ifTrue:
[byteCount>550 ifTrue: [self halt].
self externalizeIPandSP.
prevCtxt = localFP ifFalse:
 [prevCtxt := localFP.
 transcript cr; nextPutAll: (self printTop: 2); endEntry].
transcript cr; print: byteCount; nextPutAll: ': ' , (localFP hex); space;
 print: (instructionPointer - method - (BaseHeaderSize - 2));
 nextPutAll: ': <' , (self byteAt: localIP) hex , '>'; space;
 nextPutAll: (self symbolic: currentBytecode at: localIP inMethod: method); space;
 print: (self stackPointerIndex - TempFrameStart + 1); endEntry.
].
"
		self dispatchOn: currentBytecode in: BytecodeTable.
		localFP = priorFrame ifFalse:
			[sendCount := sendCount + 1.
			 loggingStart >= sendCount ifTrue:
				[rightSelector := logFile nextLine.
				 (self stringOf: messageSelector) = rightSelector ifFalse:
					[self halt: 'halt at ', sendCount printString]].
			priorFrame := localFP].
		self incrementByteCount].
	self externalizeIPandSP.
	logFile close.
	self inform: nSends printString , ' sends verfied.'! !

!CogVMSimulator methodsFor: 'testing' stamp: 'eem 12/29/2016 08:28'!
logOfSendsWrite: nSends toFileNamed: fileName fromStart: loggingStart
	"Write a log file for testing a flaky interpreter on the same image"
	"self logOfSendsWrite: 10000 toFileNamed: 'clone32Messages.log' fromStart: 2500"
	
	| logFile priorFrame |
	logFile := FileStream newFileNamed: fileName.
	transcript clear.
	byteCount := 0.
	sendCount := 0.
	priorFrame := localFP.
	quitBlock := [^self close].
	self initStackPages.
	self loadInitialContext.
	self internalizeIPandSP.
	self fetchNextBytecode.
	[sendCount < nSends] whileTrue:
		[self dispatchOn: currentBytecode in: BytecodeTable.
		localFP == priorFrame ifFalse:
			[sendCount >= loggingStart ifTrue:
				[sendCount := sendCount + 1.
				 logFile nextPutAll: (self stringOf: messageSelector); cr].
			priorFrame := localFP].
		self incrementByteCount].
	self externalizeIPandSP.
	logFile close! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 12/8/2016 11:24'!
logSend: oop
	sendCount := sendCount + 1.
	(printSends or: [printBytecodeAtEachStep]) ifTrue:
		[transcript print: byteCount; nextPut: $/; print: sendCount; space.
		 self printStringOf: oop.
		 transcript cr; flush]! !

!CogVMSimulator methodsFor: 'memory access' stamp: 'eem 6/25/2011 12:18'!
long32At: byteAddress
	"Return the 32-bit word at byteAddress which must be 0 mod 4."
	^objectMemory long32At: byteAddress! !

!CogVMSimulator methodsFor: 'memory access' stamp: 'eem 6/25/2011 12:18'!
longAt: byteAddress
	^objectMemory longAt: byteAddress! !

!CogVMSimulator methodsFor: 'memory access' stamp: 'eem 6/25/2011 12:12'!
longAt: byteAddress put: a32BitValue
	^objectMemory longAt: byteAddress put: a32BitValue! !

!CogVMSimulator methodsFor: 'message sending' stamp: 'eem 8/17/2014 18:22'!
lookupMethodInClass: class
	lookupCount := lookupCount + 1.
	^super lookupMethodInClass: class! !

!CogVMSimulator methodsFor: 'callback support' stamp: 'rmacnak 6/24/2015 21:17'!
lookupOrdinaryNoMNUEtcInClass: class
	lookupCount := lookupCount + 1.
	^super lookupOrdinaryNoMNUEtcInClass: class! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem 12/8/2016 14:15'!
makeBaseFrameFor: aContext "<Integer>"
	"aContext =  26431360 ifTrue: [self halt]."
	"(objectMemory fetchPointer: MethodIndex ofObject: aContext) = 16rD4C178 ifTrue:
		[self halt]."
	^super makeBaseFrameFor: aContext! !

!CogVMSimulator methodsFor: 'cog jit support' stamp: 'eem 12/10/2015 08:32'!
mapFunctionToAddress: aSymbolOrIndexOrBlock
	"Hackery to deal with the plugin primitive simulation hoops.
	 aSymbolOrIndex is either a Symbol (#primitiveExternalCall) or an
	 index above 1001 (an index + 1000 into the externalPrimitiveTable)."
	| sobui symbolOrBlock |
	aSymbolOrIndexOrBlock = 0 ifTrue: [^0].
	self assert: (aSymbolOrIndexOrBlock isSymbol
				or: [aSymbolOrIndexOrBlock isBlock
				or: [aSymbolOrIndexOrBlock isInteger
					and: [aSymbolOrIndexOrBlock between: 1001 and: 2000]]]).
	symbolOrBlock := aSymbolOrIndexOrBlock isInteger
						ifTrue: [(self pluginEntryFor: aSymbolOrIndexOrBlock) at: 3]
						ifFalse: [aSymbolOrIndexOrBlock].
	uniqueIndices ifNil:
		[uniqueIndices := Dictionary new.
		 uniqueIndex := 65535].
	sobui := uniqueIndices at: symbolOrBlock ifAbsentPut: [uniqueIndex := uniqueIndex + 1].
	^cogit
		mapPrimitive: symbolOrBlock
		withIndexToUniqueAddress: sobui! !

!CogVMSimulator methodsFor: 'object memory support' stamp: 'eem 8/20/2021 15:15'!
mapInterpreterOops
	"Override both to elide the CoInterpreterMT version if not running MT,
	 and to update perMethodProfile if in use."
	self perform: #mapInterpreterOops
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT]).

	perMethodProfile ifNotNil:
		[:pmp| | map |
		 "The tricky thing here is that a method may get remapped to another method already in the profile, etc..."
		 map := Dictionary new.
		 perMethodProfile keysAndValuesDo:
			[:methodOop :count|
			(objectMemory shouldRemapOop: methodOop) ifTrue:
				[map at: methodOop put: {objectMemory remapObj: methodOop. count}]].
		 map isEmpty ifFalse:
			[map keysAndValuesDo:
				[:newOop :tuple|
				 [:oldOop :count|
				  (map includesKey: oldOop) ifFalse:
					[perMethodProfile removeKey: oldOop].
				  perMethodProfile at: newOop put: count] valueWithArguments: tuple]]]! !

!CogVMSimulator methodsFor: 'plugin support' stamp: 'eem 6/3/2016 16:22'!
mappedPluginEntries
	^mappedPluginEntries! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
markAndTraceInterpreterOops: fullGCFlag
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #markAndTraceInterpreterOops:
		withArguments: {fullGCFlag}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 1/16/2020 17:38'!
maybeCheckStackDepth: delta sp: sp pc: mcpc
	| asp bcpc startbcpc cogHomeMethod cogBlockMethod csp debugStackPointers |
	debugStackDepthDictionary ifNil: [^self].
	(self isMachineCodeFrame: framePointer) ifFalse: [^self].
	cogBlockMethod := self mframeCogMethod: framePointer.
	cogHomeMethod := self asCogHomeMethod: cogBlockMethod.
	debugStackPointers := debugStackDepthDictionary
								at: cogHomeMethod methodObject
								ifAbsentPut: [self debugStackPointersFor: cogHomeMethod methodObject].
	startbcpc := cogHomeMethod = cogBlockMethod
					ifTrue: [self startPCOfMethod: cogHomeMethod methodObject]
					ifFalse: [self startPCOfClosure: (self pushedReceiverOrClosureOfFrame: framePointer)].
	bcpc := cogit bytecodePCFor: mcpc startBcpc: startbcpc in: cogBlockMethod.
	self assert: bcpc ~= 0.
	((cogBlockMethod ~= cogHomeMethod or: [cogBlockMethod cmIsFullBlock])
	 and: [cogit isNonLocalReturnPC: mcpc]) ifTrue:
		[| lastbcpc |
		 "Method returns within a block (within an unwind-protect) must check the stack depth at the
		  return, not the bytecode following, but the pc mapping maps to the bytecode following the
		  return. lastBytecodePCForBlockAt:in: catches method returns at the end of a block, modifying
		  the bcpc to that of the return.  isNonLocalReturnPC: catches method returns not at the end.
		  Assumes method return bytecodes are 1 bytecode long;a  dodgy assumption, but good enough."
		 lastbcpc := cogBlockMethod cmIsFullBlock
						ifTrue: [cogit endPCOf: cogHomeMethod methodObject]
						ifFalse: [cogit lastBytecodePCForBlockAt: startbcpc in: cogHomeMethod methodObject].
		 bcpc > lastbcpc ifTrue: [bcpc := lastbcpc]].
	asp := self stackPointerIndexForFrame: framePointer WithSP: sp + objectMemory wordSize.
	csp := debugStackPointers at: bcpc ifAbsent: [-1].
	"Compensate for some edge cases"
	asp - delta = csp ifTrue:
		["Compensate for the implicit context receiver push in a trap bytecode with the absence of a continuation.
		  Assumes trap bytecodes are 1 byte bytecodes."
		 (SistaVM
		  and: [cogit isTrapAt: mcpc]) ifTrue:
			[csp := csp + 1].
		"Compensate lazily for absent receiver sends (cuz mapping is slow, even though incrememting csp is a dodgy idea)."
		(NewspeakVM
		 and: [cogit isAbsentReceiverSendAt: mcpc in: cogHomeMethod]) ifTrue:
			[csp := debugStackPointers at: bcpc put: csp + 1]].
	self assert: asp - delta + 1 = csp! !

!CogVMSimulator methodsFor: 'primitive support' stamp: 'eem 6/6/2016 12:09'!
maybeMapPrimitiveFunctionPointerBackToSomethingEvaluable
	"In the real VM primitiveFunctionPointer is either an index (for quick primitives)
	 or a proper function pointer to a primitive.  In the simulator it may be a small
	 index (corresponding to a quick primitive index), a symbol (corresponding to
	 a function pointer) or an index into the externalPrimitiveTable, or an invalid
	 address that references an evaluable in the simulatedTrampolines dictionary
	 of the Cogit.  The simulator expects dispatchFunctionPointer to be called with
	 primitiveFunctionPointer being a symbol only for internal primitives.  External
	 primitives must have their funciton pointer mapped back to an index.  This
	 method does the mapping back from fake addresses."
	<doNotGenerate>
	(primitiveFunctionPointer isInteger
	 and: [self isExternalPrimitiveCall: newMethod]) ifTrue: "External prims must be evaluated by the right plugin..."
		[(cogit simulatedTrampolines at: primitiveFunctionPointer ifAbsent: nil) ifNotNil:
			[:evaluable| | pfp index externalIndex |
						"primitiveFunctionPointer := pfp"
						"(1 to: self mappedPluginEntries size) select: [:index| (self mappedPluginEntries at: index) third == evaluable]"
			 pfp := primitiveFunctionPointer.
			 index := self mappedPluginEntries findFirst: [:entry| entry third == evaluable].
			 self assert: index ~= 0.
			 externalIndex := 1000 + (externalPrimitiveTable object
										indexOf: index
										ifAbsent: [self error: 'entry not found']).
			 self assert: ((self pluginEntryFor: externalIndex) notNil
						   and: [(self pluginEntryFor: externalIndex) third == evaluable]).
			 primitiveFunctionPointer := externalIndex.
			 ^self]].
	^super maybeMapPrimitiveFunctionPointerBackToSomethingEvaluable! !

!CogVMSimulator methodsFor: 'accessing' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
memory
	^objectMemory memory! !

!CogVMSimulator methodsFor: 'spur bootstrap' stamp: 'eem 10/2/2013 15:19'!
methodCache
	^methodCache! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
methodCacheSize
	^MethodCacheSize * objectMemory wordSize! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 2/27/2013 13:50'!
methodForContext: aContextOop
	self assert: (objectMemory isContext: aContextOop).
	^objectMemory fetchPointer: MethodIndex ofObject: aContextOop! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 4/11/2014 16:50'!
minBackwardJumpCountForCompile
	<doNotGenerate>
	^minBackwardJumpCountForCompile! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 4/11/2014 16:50'!
minBackwardJumpCountForCompile: n
	<doNotGenerate>
	minBackwardJumpCountForCompile := n! !

!CogVMSimulator methodsFor: 'initialize-release' stamp: 'eem 10/7/2015 16:13'!
moveMethodCacheToMemoryAt: address
	| oldMethodCache |
	oldMethodCache := methodCache.
	"In the VM the methodCache is written as a normal array with 1-relative addressing.
	 In C this works by allocating an extra element in the methodCache array (see
	 class-side declareCVarsIn:).  In simulation simply position the start of the methodCache
	 one word lower, achieving the same effect.  -1 because CArrayAccessor is 0-relative
	 and adds 1 on accesses itself."
	methodCache := CMethodCacheAccessor new
						objectMemory: objectMemory
						at: address
						array: oldMethodCache
						functionPointerIndex: MethodCachePrimFunction
						entrySize: MethodCacheEntrySize.
	self assert: address - objectMemory wordSize = self methodCacheAddress.
	1 to: MethodCacheSize do:
		[:i|
		self assert: (methodCache at: i) = 0].
	methodCache at: 1 put: 16rC4EC4.
	self assert: (objectMemory longAt: address) = 16rC4EC4.
	1 to: MethodCacheSize do:
		[:i|
		methodCache at: i put: (oldMethodCache at: i)]! !

!CogVMSimulator methodsFor: 'initialize-release' stamp: 'eem 10/17/2015 11:11'!
movePrimTraceLogToMemoryAt: address
	| oldTraceLog |
	oldTraceLog := primTraceLog.
	primTraceLog := CArrayOfLongsAccessor new
						objectMemory: objectMemory at: address.
	self assert: address = self primTraceLogAddress.
	0 to: PrimTraceLogSize - 1 do:
		[:i|
		self assert: (primTraceLog at: i) = 0].
	primTraceLog at: 0 put: 16rC4EC4.
	self assert: (objectMemory longAt: address) = 16rC4EC4.
	0 to: PrimTraceLogSize - 1 do:
		[:i|
		primTraceLog at: i put: (oldTraceLog at: i)]! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 8/21/2014 07:17'!
nameOfClass: classOop
	| numSlots maybeThisClassOop maybeNameOop name |
	classNameIndex ifNil: [^'??nil cnidx??'].
	numSlots := objectMemory numSlotsOf: classOop.
	numSlots = metaclassNumSlots ifTrue:
		[maybeThisClassOop := objectMemory fetchPointer: thisClassIndex ofObject: classOop.
		(self addressCouldBeClassObj: maybeThisClassOop) ifTrue:
			[name := self nameOfClass: maybeThisClassOop.
			 (name beginsWith: 'bad ') ifFalse:
				[^name, ' class']].
		 ^'bad class'].
	(numSlots > classNameIndex
	 and: [maybeNameOop := objectMemory fetchPointer: classNameIndex ofObject: classOop.
		objectMemory isBytes: maybeNameOop]) ifFalse:
		[^'bad class'].
	^self stringOf: maybeNameOop! !

!CogVMSimulator methodsFor: 'initialize-release' stamp: 'eem 10/9/2008 16:37'!
nextShortFrom: aStream
	"Read a 16-bit quantity from the given (binary) stream."
	^self subclassResponsibility! !

!CogVMSimulator methodsFor: 'test support' stamp: 'eem 1/25/2013 06:28'!
numStackPages
	^numStackPages! !

!CogVMSimulator methodsFor: 'initialize-release' stamp: 'eem 6/30/2021 18:55'!
openOn: fileName extraMemory: extraBytes
	"CogVMSimulator new openOn: 'clone.im' extraMemory: 100000"

	| f version headerSize dataSize count oldBaseAddr bytesToShift swapBytes
	  headerFlags firstSegSize heapSize
	  hdrNumStackPages hdrEdenBytes hdrMaxExtSemTabSize hdrCogCodeSize
	  stackZoneSize methodCacheSize primTraceLogSize allocationReserve |
	"open image file and read the header"

	(f := self openImageFileNamed: fileName) ifNil: [^self].

	"Set the image name and the first argument; there are
	 no arguments during simulation unless set explicitly."
	systemAttributes at: 1 put: fileName.

	["begin ensure block..."
	imageName := f fullName.
	f binary.

	version := self getWord32FromFile: f swap: false.  "current version: 16r1968 (=6504) vive la revolucion!!"
	(self readableFormat: version)
		ifTrue: [swapBytes := false]
		ifFalse: [(version := version byteSwap32) = self imageFormatVersion
					ifTrue: [swapBytes := true]
					ifFalse: [self error: 'incomaptible image format']].
	headerSize := self getWord32FromFile: f swap: swapBytes.
	dataSize := self getLongFromFile: f swap: swapBytes.  "length of heap in file"
	oldBaseAddr := self getLongFromFile: f swap: swapBytes.  "object memory base address of image"
	objectMemory specialObjectsOop: (self getLongFromFile: f swap: swapBytes).
	objectMemory lastHash: (self getLongFromFile: f swap: swapBytes).  "Should be loaded from, and saved to the image header"

	savedWindowSize	:= self getLongFromFile: f swap: swapBytes.
	headerFlags		:= self getLongFromFile: f swap: swapBytes.
	self setImageHeaderFlagsFrom: headerFlags.
	extraVMMemory	:= self getWord32FromFile: f swap: swapBytes.
	hdrNumStackPages	:= self getShortFromFile: f swap: swapBytes.
	"4 stack pages is small.  Should be able to run with as few as
	 three. 4 should be comfortable but slow.  8 is a reasonable
	 default. Can be changed via vmParameterAt: 43 put: n"
	numStackPages := desiredNumStackPages ~= 0
						ifTrue: [desiredNumStackPages]
						ifFalse: [hdrNumStackPages = 0
									ifTrue: [self defaultNumStackPages]
									ifFalse: [hdrNumStackPages]].
	desiredNumStackPages := hdrNumStackPages.
	stackZoneSize := self computeStackZoneSize.
	"This slot holds the size of the native method zone in 1k units. (pad to word boundary)."
	hdrCogCodeSize := (self getShortFromFile: f swap: swapBytes) * 1024.
	cogCodeSize := desiredCogCodeSize ~= 0
						ifTrue: [desiredCogCodeSize]
						ifFalse:
							[hdrCogCodeSize = 0
									ifTrue: [cogit defaultCogCodeSize]
									ifFalse: [hdrCogCodeSize]].
	desiredCogCodeSize := hdrCogCodeSize.
	self assert: f position = (objectMemory wordSize = 4 ifTrue: [40] ifFalse: [64]).
	hdrEdenBytes	:= self getWord32FromFile: f swap: swapBytes.
	objectMemory edenBytes: (desiredEdenBytes ~= 0
						ifTrue: [desiredEdenBytes]
						ifFalse:
							[hdrEdenBytes = 0
									ifTrue: [objectMemory defaultEdenBytes]
									ifFalse: [hdrEdenBytes]]).
	desiredEdenBytes := hdrEdenBytes.
	hdrMaxExtSemTabSize := self getShortFromFile: f swap: swapBytes.
	hdrMaxExtSemTabSize ~= 0 ifTrue:
		[self setMaxExtSemSizeTo: hdrMaxExtSemTabSize].
	"pad to word boundary.  This slot can be used for anything else that will fit in 16 bits.
	 Preserve it to be polite to other VMs."
	the2ndUnknownShort	:= self getShortFromFile: f swap: swapBytes.
	self assert: f position = (objectMemory wordSize = 4 ifTrue: [48] ifFalse: [72]).
	firstSegSize := self getLongFromFile: f swap: swapBytes.
	objectMemory firstSegmentSize: firstSegSize.
	"For Open PICs to be able to probe the method cache during
	 simulation the methodCache must be relocated to memory."
	methodCacheSize := methodCache size * objectMemory wordSize.
	primTraceLogSize := primTraceLog size * objectMemory wordSize.

	"To cope with modern OSs that disallow executing code in writable memory we dual-map
	 the code zone, one mapping with read/write permissions and the other with read/execute
	 permissions. In simulation all we can do is use memory, so if we're simulating dual mapping
	 we use double the memory and simulate the memory sharing in the Cogit's backEnd."
	effectiveCogCodeSize := (InitializationOptions at: #DUAL_MAPPED_CODE_ZONE ifAbsent: [false])
								ifTrue: [cogCodeSize * 2]
								ifFalse: [cogCodeSize].

	"allocate interpreter memory. This list is in address order, low to high.
	 In the actual VM the stack zone exists on the C stack."
	heapBase := (Cogit guardPageSize
				+ effectiveCogCodeSize
				+ stackZoneSize
				+ methodCacheSize
				+ primTraceLogSize
				+ self rumpCStackSize) roundUpTo: objectMemory allocationUnit.
	"compare memory requirements with availability"
	allocationReserve := self interpreterAllocationReserveBytes.
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[| freeOldSpaceInImage headroom |
			 freeOldSpaceInImage := self getLongFromFile: f swap: swapBytes.
			 headroom := objectMemory
							initialHeadroom: extraVMMemory
							givenFreeOldSpaceInImage: freeOldSpaceInImage.
			 heapSize := objectMemory roundUpHeapSize:
						   dataSize
						+ headroom
						+ objectMemory newSpaceBytes
						+ (headroom > allocationReserve
							ifTrue: [0]
							ifFalse: [allocationReserve])]
		ifFalse:
			[heapSize :=  dataSize
						+ extraBytes
						+ objectMemory newSpaceBytes
						+ (extraBytes > allocationReserve
							ifTrue: [0]
							ifFalse: [allocationReserve])].
	"allocate interpreter memory"
	heapBase := objectMemory
					setHeapBase: heapBase
					memoryLimit: heapBase + heapSize
					endOfMemory: heapBase + dataSize. "bogus for Spur"

	self assert: cogCodeSize \\ 4 = 0.
	self assert: objectMemory memoryLimit \\ 4 = 0.
	self assert: self rumpCStackSize \\ 4 = 0.
	objectMemory allocateMemoryOfSize: objectMemory memoryLimit.
	"read in the image in bulk, then swap the bytes if necessary"
	f position: headerSize.
	count := objectMemory readHeapFromImageFile: f dataBytes: dataSize.
	count ~= dataSize ifTrue: [self halt]]
		ensure: [f close].
	self moveMethodCacheToMemoryAt: objectMemory cogCodeBase + effectiveCogCodeSize + stackZoneSize.
	self movePrimTraceLogToMemoryAt: objectMemory cogCodeBase + effectiveCogCodeSize + stackZoneSize + methodCacheSize.

	self ensureImageFormatIsUpToDate: swapBytes.

	bytesToShift := objectMemory memoryBaseForImageRead - oldBaseAddr.  "adjust pointers for zero base address"
	UIManager default
		informUser: 'Relocating object pointers...'
		during: [self initializeInterpreter: bytesToShift].
	cogit
		initializeCodeZoneFrom: Cogit guardPageSize
		upTo: Cogit guardPageSize + cogCodeSize! !

!CogVMSimulator methodsFor: 'stack pages' stamp: 'eem 1/29/2013 11:50'!
osCogStackPageHeadroom
	"Notional headroom for the simulator.  The platform provides this in the real VM."
	^1024! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 11/22/2020 21:00'!
ownVM: threadIndexAndFlags
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"
	"(threadIndexAndFlags anyMask: DisownVMForProcessorRelinquish << DisownFlagsShift) ifFalse:
		[self break]."
	^self perform: #ownVM:
		withArguments: {threadIndexAndFlags}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
ownVMFromUnidentifiedThread
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #ownVMFromUnidentifiedThread
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
ownerIndexOfProcess: aProcess
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #ownerIndexOfProcess:
		withArguments: {aProcess}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
ownerIndexOfThreadId: threadId
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #ownerIndexOfThreadId:
		withArguments: {threadId}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'accessing' stamp: 'eem 10/11/2014 14:53'!
parent
	^ parent! !

!CogVMSimulator methodsFor: 'accessing' stamp: 'eem 10/11/2014 14:54'!
parent: anObject

	parent := anObject! !

!CogVMSimulator methodsFor: 'plugin support' stamp: 'eem 4/8/2009 10:09'!
pluginEntryFor: mapIndex
	^mappedPluginEntries at: (mapIndex > 1000
								ifTrue: [externalPrimitiveTable at: mapIndex - 1001]
								ifFalse: [mapIndex])! !

!CogVMSimulator methodsFor: 'object memory support' stamp: 'eem 3/17/2014 17:14'!
preGCAction: gcModeArg
	"Override to void debugStackPointers on any GC/remap"
	debugStackDepthDictionary ifNotNil:
		[debugStackDepthDictionary := Dictionary new].
	^super preGCAction: gcModeArg! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
preemptDisowningThread
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #preemptDisowningThread
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'other primitives' stamp: 'di 6/29/1998 22:17'!
primStringcomparewithcollated
	^ self primitiveFail! !

!CogVMSimulator methodsFor: 'other primitives' stamp: 'ikp 12/15/1998 23:30'!
primStringfindSubstringinstartingAtmatchTable
	^self primitiveFail! !

!CogVMSimulator methodsFor: 'other primitives' stamp: 'di 11/8/1998 13:04'!
primStringindexOfAsciiinStringstartingAt
	^ self primitiveFail! !

!CogVMSimulator methodsFor: 'other primitives' stamp: 'yo 1/29/1999 21:37'!
primStringtranslatefromtotable
	^ self primitiveFail! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
primTraceLogSize
	^PrimTraceLogSize * objectMemory wordSize! !

!CogVMSimulator methodsFor: 'other primitives' stamp: ''!
primitiveBeep

	Beeper beep.! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 2/21/2009 17:25'!
primitiveClosureValueWithArgs
	"Transcript clear.
	 self shortPrintFrameAndCallers: framePointer.
	 self halt."
	^super primitiveClosureValueWithArgs! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 2/7/2009 10:58'!
primitiveContextAtPut
	"| aContext |
	aContext := self stackValue: 2.
	(#(24205456 24205732) includes: aContext) ifTrue:
		[(self checkIsStillMarriedContext: aContext currentFP: framePointer)
			ifTrue: [self printFrame: (self frameOfMarriedContext: aContext)
						WithSP: (self frameOfMarriedContext: aContext) - 48]
			ifFalse: [self printContext: aContext]]."
	^super primitiveContextAtPut! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 12/17/2018 10:39'!
primitiveDoPrimitiveWithArgs
	| primIndex |
	primIndex := objectMemory integerValueOf: (self stackValue: 1).
	NewspeakVM ifFalse:
		[transcript nextPutAll: 'DO PRIMITIVE: '; print: (self functionPointerFor: primIndex inClass: nil); cr; flush].
	(#(76 "primitiveStoreStackp" 188 189 "eval method") includes: primIndex) ifTrue:
		[self halt].
	^super primitiveDoPrimitiveWithArgs! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 7/25/2020 21:38'!
primitiveExecuteMethod
	(InitializationOptions at: #haltOnExecuteMethod ifAbsent: [true]) ifTrue:
		[self halt: thisContext selector].
	^super primitiveExecuteMethod

	"InitializationOptions at: #haltOnExecuteMethod put: false"! !

!CogVMSimulator methodsFor: 'control primitives' stamp: 'eem 7/25/2020 19:11'!
primitiveExecuteMethodArgsArray
	(InitializationOptions at: #haltOnExecuteMethod ifAbsent: [true]) ifTrue:
		[self halt: thisContext selector].
	^super primitiveExecuteMethodArgsArray

	"InitializationOptions at: #haltOnExecuteMethod put: false"! !

!CogVMSimulator methodsFor: 'other primitives' stamp: 'eem 2/13/2014 09:35'!
primitiveGetAttribute
	"Fetch the system attribute with the given integer ID. The result is a string, which will be empty if the attribute is not defined."

	| index s attribute |
	index := self stackIntegerValue: 0.
	self successful ifTrue: [
		attribute := systemAttributes at: index ifAbsent: [Smalltalk vm getSystemAttribute: index].
		attribute ifNil: [ ^self primitiveFail ].
		s := objectMemory instantiateClass: (objectMemory splObj: ClassByteString) indexableSize: attribute size.
		1 to: attribute size do: [ :i |
			objectMemory storeByte: i-1 ofObject: s withValue: (attribute at: i) asciiValue].
		self pop: 2 "rcvr, attr" thenPush: s]! !

!CogVMSimulator methodsFor: 'I/O primitives' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
primitiveKbdNext

	self pop: 1.
	Sensor keyboardPressed
		ifTrue: [self pushInteger: Sensor primKbdNext]
		ifFalse: [self push: objectMemory nilObject]! !

!CogVMSimulator methodsFor: 'I/O primitives' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
primitiveKbdPeek

	self pop: 1.
	Sensor keyboardPressed
		ifTrue: [self pushInteger: Sensor primKbdPeek]
		ifFalse: [self push: objectMemory nilObject]! !

!CogVMSimulator methodsFor: 'I/O primitives' stamp: 'di 9/23/97 12:54'!
primitiveMouseButtons
	| buttons |
	self pop: 1.
	buttons := Sensor primMouseButtons.
	self pushInteger: buttons! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 10/3/2013 12:52'!
primitiveNewWithArg
	"(objectMemory hasSpurMemoryManagerAPI
	 and: [self classNameOf: (self stackValue: 1) Is: 'Bitmap']) ifTrue:
		[self printExternalHeadFrame.
		 self halt]."
	^super primitiveNewWithArg! !

!CogVMSimulator methodsFor: 'control primitives' stamp: 'eem 1/11/2013 18:38'!
primitiveObject: actualReceiver perform: selector withArguments: argumentArray lookedUpIn: lookupClass
	"Override to allow simulator to filter start-up items to avoid as-yet unsimulatable plugins."

	(self filterPerformOf: selector to: actualReceiver) ifTrue:
		[^self pop: argumentCount + 1 thenPush: actualReceiver].
	^super primitiveObject: actualReceiver perform: selector withArguments: argumentArray lookedUpIn: lookupClass! !

!CogVMSimulator methodsFor: 'object access primitives' stamp: 'eem 12/6/2015 15:55'!
primitiveObjectAt
	"self transcript
		nextPutAll: (self nameOfClass: (self methodClassOf: (self stackValue: 1)));
		nextPutAll: '>>#';
		nextPutAll: (self stringOf: (self maybeSelectorOfMethod: (self stackValue: 1)));
		nextPutAll: ' objectAt: '; print: (objectMemory integerValueOf: self stackTop); cr; flush.
	self halt."
	^super primitiveObjectAt! !

!CogVMSimulator methodsFor: 'object access primitives' stamp: 'eem 11/14/2012 23:37'!
primitiveObjectAtPut
	| newValue |
	newValue := self stackValue: 0.
	((self stackValue: 1) = ConstOne
	 and: [(objectMemory isIntegerObject: newValue) not
		   or: [MULTIPLEBYTECODESETS not
			and: [(objectMemory integerValueOf: newValue) < 0]]]) ifTrue:
		[self halt].
	^super primitiveObjectAtPut! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 12/5/2010 10:51'!
primitiveObjectPointsTo
	"self halt."
	^super primitiveObjectPointsTo! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 2/28/2017 10:46'!
primitivePerform
	| receiver selector |
	"If called from the machine code perform primitive, it should not have been found,
	 except that the cacheing for V3 has a mismatch between that used ror the first-level
	 method cache and inline caches."
	receiver := (self stackValue: argumentCount).
	selector := self stackValue: argumentCount - 1.
	((self methodHasCogMethod: newMethod)
	 and: [(objectMemory isCompactInstance: receiver) not]) ifTrue:
		[self deny: (self newMethodInLookupCacheAt: selector and: (objectMemory fetchClassTagOf: receiver))].
	self sendBreakpoint: selector receiver: receiver.
	(self filterPerformOf: selector to: receiver) ifTrue:
		[^self pop: argumentCount].
	^super primitivePerform! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
primitiveProcessBindToThreadId
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #primitiveProcessBindToThreadId
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
primitiveProcessBoundThreadId
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #primitiveProcessBoundThreadId
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/19/2020 17:26'!
primitivePropertyFlagsForSpur: primIndex
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #primitivePropertyFlagsForSpur:
		withArguments: {primIndex}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
primitiveRelinquishProcessor
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #primitiveRelinquishProcessor
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'system control primitives' stamp: 'eem 11/26/2013 16:42'!
primitiveSignalAtMilliseconds
	super primitiveSignalAtMilliseconds.
	"self successful ifTrue:
		[Transcript
			cr; nextPutAll: thisContext selector;
			nextPutAll: ' now '; nextPutAll: self ioUTCMicroseconds hex;
			nextPutAll: ' wakeup '; nextPutAll: nextWakeupUsecs hex;
			nextPutAll: ' wakeup - now '; print: nextWakeupUsecs - self ioUTCMicroseconds; flush]"! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 12/2/2010 18:39'!
primitiveStoreStackp
	"self printContext: (self stackValue: 1).
	self halt."
	"(self stackValue: 1) = 16r1934F80 ifTrue: [self halt]."
	super primitiveStoreStackp! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
primitiveVMCurrentThreadId
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #primitiveVMCurrentThreadId
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'system control primitives' stamp: 'eem 6/7/2017 13:34'!
primitiveVMParameter
	(argumentCount = 2
	 and: [(self stackValue: 1) = (objectMemory integerObjectOf: 34)]) ifTrue:
		[self halt].
	^super primitiveVMParameter! !

!CogVMSimulator methodsFor: 'system control primitives' stamp: 'eem 3/24/2017 11:55'!
primitiveVoidVMState
	self halt.
	^super primitiveVoidVMState! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 6/20/2008 17:39'!
print: it

	traceOn ifTrue:
		[it isString ifTrue: [transcript nextPutAll: it] ifFalse: [it printOn: transcript]]! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 8/4/2016 14:25'!
printActivationNameForSelector: aSelector startClass: startClass
	super printActivationNameForSelector: aSelector startClass: startClass.
	"(self transcript dependents detect: [:d| d isTextView] ifNone: []) ifNotNil:
		[:textView| | text size |
		text := textView text asString.
		((size := text size) >= 30
		and: [(text copyFrom: size - 29 to: size) = 'Object(ProtoObject)>initialize']) ifTrue:
			[self doOrDefer: [self changed: #byteCountText; changed: #composeAll].
			 self halt]"! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 12/27/2019 16:45'!
printCallStack
	<inline: false>
	cogit headFramePointer
		ifNil: [self printCallStackOf: (objectMemory fetchPointer: SuspendedContextIndex ofObject: self activeProcess)]
		ifNotNil: [:fp| self printCallStackFP: fp]! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 12/27/2019 10:03'!
printChar: aByte

	traceOn ifTrue:
		[(aByte between: 0 and: 255)
			ifTrue: [transcript nextPut: aByte asCharacter]
			ifFalse: [transcript nextPutAll: 'BAD CHARACTER '.
					aByte printOn: transcript base: 16]]! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 2/24/2020 21:54'!
printCogMethod: cogMethod on: aStream
	<doNotGenerate>
	| oldTranscript |
	oldTranscript := transcript.
	transcript := aStream.
	[self printCogMethod: cogMethod] ensure:
		[transcript := oldTranscript]! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 6/26/2012 09:15'!
printCurrentBytecodeOn: aStream
	| code |
	code := currentBytecode radix: 16.
	aStream ensureCr; print: localIP - method - 3; tab.
	bytecodeSetSelector > 0 ifTrue:
		[aStream nextPutAll: 'ALT '].
	aStream
		nextPut: (code size < 2
					ifTrue: [$0]
					ifFalse: [code at: 1]);
		nextPut: code last; space;
		nextPutAll: (BytecodeTable at: currentBytecode + 1);
		space;
		nextPut: $(; print: byteCount + 1; nextPut: $)! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 7/4/2008 15:02'!
printECS
	self printCallStackFP: framePointer! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 7/7/2008 16:46'!
printFloat: f

	traceOn ifTrue: [ transcript print: f ]! !

!CogVMSimulator methodsFor: 'testing' stamp: 'eem 3/18/2015 15:21'!
printFrameAtEachStep
	^printFrameAtEachStep! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 12/30/2008 08:17'!
printFrameOop: name at: address
	<inline: false>
	self printFrameOop: name obj: (stackPages longAt: address) at: address! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 3/12/2009 15:38'!
printFrameOop: name obj: obj at: address
	| it len |
	<inline: false>
	<var: #name type: #'char *'>
	<var: #address type: #'char *'>
	it := stackPages longAt: address.
	self printHex: address;
		printChar: $:.
	len := self strlen: name.
	1 to: 12 - len do: [:i| self printChar: $ ].
	self print: name;
		print: ': ';
		printHex: it.
	self tab.
	it = obj
		ifTrue: [self tab]
		ifFalse: [self printHex: obj; space].
	self
		print: (self shortPrint: obj);
		cr! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 12/4/2013 09:10'!
printFrameThing: name andFrame: theFP at: address
	<var: #theFP type: #'char *'>
	| it |
	<inline: false>
	<var: #name type: #'char *'>
	<var: #address type: #'char *'>
	it := stackPages longAt: address.
	self printHex: address;
		printChar: $:.
	1 to: 12 - (self strlen: name) do: [:i| self printChar: $ ].
	self print: name;
		print: ': ';
		printHex: it.
	it ~= 0 ifTrue:
		[self printChar: $=; printNum: it.
		 (objectMemory isInMemory: it) ifFalse:
			[(cogit lookupAddress: it) ifNotNil:
				[:label| self space; printChar: $(; print: label; printChar: $)]]].
	self print: ' frame: '; printHex: theFP; cr! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 12/4/2013 09:09'!
printFrameThing: name at: address
	| it |
	<inline: false>
	<var: #name type: #'char *'>
	<var: #address type: #'char *'>
	it := stackPages longAt: address.
	self printHex: address;
		printChar: $:.
	1 to: 12 - (self strlen: name) do: [:i| self printChar: $ ].
	self print: name;
		print: ': ';
		printHex: it.
	it ~= 0 ifTrue:
		[self printChar: $=; printNum: it.
		 (objectMemory isInMemory: it) ifFalse:
			[(cogit lookupAddress: it) ifNotNil:
				[:label| self space; printChar: $(; print: label; printChar: $)]]].
	self cr! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 1/7/2020 16:40'!
printHeadFrame
	self printFrame: cogit headFramePointer WithSP: cogit headStackPointer! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 12/25/2019 17:35'!
printHex: anInteger
	traceOn ifTrue:
		[self printHex: anInteger on: transcript]! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 12/25/2016 10:51'!
printHexnp: anInteger

	traceOn ifTrue:
		[transcript nextPutAll: ((anInteger ifNil: [0]) asInteger storeStringBase: 16)]! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 1/2/2017 17:39'!
printHexnpnp: anInteger
	"Print n in hex, in the form '1234', unpadded"
	traceOn ifTrue:
		[transcript nextPutAll: ((anInteger ifNil: [0]) printStringBase: 16)]! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 7/4/2008 15:02'!
printLCS
	self printCallStackFP: localFP! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
printLogEntryAt: i
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #printLogEntryAt:
		withArguments: {i}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 7/7/2008 16:45'!
printNum: anInteger

	traceOn ifTrue: [ transcript print: anInteger ].! !

!CogVMSimulator methodsFor: 'rump c stack' stamp: 'eem 11/20/2020 13:54'!
printRumpCStack
	cogThreadManager ifNil:
		[^self printRumpCStackTo: heapBase - objectMemory wordSize].
	cogThreadManager threadsDo:
		[:thread| self printRumpCStackForThread: thread]! !

!CogVMSimulator methodsFor: 'rump c stack' stamp: 'eem 11/24/2020 22:58'!
printRumpCStackForThread: thread
	| range start coldTop |
	range := self cStackRangeForThreadIndex: thread index.
	start := range first bitClear: objectMemory tagMask.
	coldTop := range last.
	[start < coldTop and: [(objectMemory longAt: start) = 0]] whileTrue:
		[start := start + objectMemory wordSize].
	(start = coldTop and: [(objectMemory longAt: start) = 0]) ifTrue:
		[^self].
	start := start - (2 * objectMemory wordSize) max: range start.
	transcript nextPutAll: 'thread '; print: thread index.
	thread index = cogThreadManager getVMOwner ifTrue: [transcript nextPut: $*].
	transcript cr.
	self printRumpCStackFrom: start to: coldTop cfp: thread cFramePointer csp: thread cStackPointer.
	range first < start ifTrue:
		[self print: 'zeros...'; cr]! !

!CogVMSimulator methodsFor: 'rump c stack' stamp: 'eem 11/24/2020 23:12'!
printRumpCStackFrom: start to: address cfp: cfp csp: csp
	address
		to: start
		by: objectMemory wordSize negated
		do: [:addr| | ptrName value |
			self printHex: addr.
			addr = cogit processor sp ifTrue: [ptrName := ' sp'].
			addr = cogit processor fp ifTrue: [ptrName := (ptrName ifNil: [''] ifNotNil: [ptrName, ',']), 'fp'].
			addr = csp ifTrue: [ptrName := (ptrName ifNil: [''] ifNotNil: [ptrName, ',']), 'CSP'].
			addr = cfp ifTrue: [ptrName := (ptrName ifNil: [''] ifNotNil: [ptrName, ',']), 'CFP'].
			ptrName ifNil: [self tab] ifNotNil: [self print: ' ', ptrName, '->'].
			self tab; printHex: (value := objectMemory longAt: addr).
			(cogit lookupAddress: value) ifNotNil:
				[:label| self print: ' ', label].
			self cr]! !

!CogVMSimulator methodsFor: 'rump c stack' stamp: 'eem 11/20/2020 13:58'!
printRumpCStackTo: address
	| start |
	self assert: (self isOnRumpCStack: address).
	start := heapBase - self rumpCStackSize.
	[start < address and: [(objectMemory longAt: start) = 0]] whileTrue:
		[start := start + objectMemory wordSize].
	start := start - (3 * objectMemory wordSize) max: heapBase - self rumpCStackSize.
	self printRumpCStackFrom: start to: address cfp: CFramePointer csp: CStackPointer.
	heapBase - self rumpCStackSize < start ifTrue:
		[self print: 'zeros...'; cr]! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 12/27/2019 16:46'!
printStackCallStack
	<doNotGenerate>
	self printStackCallStackOf: cogit headFramePointer! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 7/1/2008 09:24'!
printStringForCurrentBytecode
	^String streamContents: [:str| self printCurrentBytecodeOn: str]! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
recordThreadSwitchTo: ownerIndex source: sourceCode
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #recordThreadSwitchTo:source:
		withArguments: {ownerIndex. sourceCode}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 7/12/2016 12:58'!
recordTrace: classOrInteger thing: selector source: source
	"This method includes or excludes CoInterpreterMT methods as required.
	 It used to be auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate.
	 Now checks for or collects expectedSends."
	(expectedSends == nil
	 or: [objectMemory isIntegerObject: classOrInteger]) ifFalse:
		[expecting
			ifTrue:
				[expectedSends peek ~~ selector ifTrue: [self halt].
				 expectedSends next]
			ifFalse:
				[expectedSends nextPut: selector]].
	
	^self perform: #recordTrace:thing:source:
		withArguments: {classOrInteger. selector. source}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 4/1/2020 14:11'!
reenterInterpreter
	ReenterInterpreter new
		returnValue: ReturnToInterpreter;
		signal! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 8/23/2021 11:15'!
reportPerMethodProfilingOn: aStream
	"Generate a sorted report of bytecodes per method, and instructions per trampoline/cog method/pic.
	 Answer an Array with the four totals."
	^self reportPerMethodProfilingOn: aStream cutoff: nil! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 8/23/2021 11:13'!
reportPerMethodProfilingOn: aStream cutoff: cutoffOrNilOrPair
	"Generate a sorted report of bytecodes per method, and instructions per trampoline/cog method/pic.
	 Answer an Array with the four totals. cutoffOrNilOrPair is either nil, or an integer, or a pair of integers,
	 { interpreter cutoff, cogit cutoff }"
	| total cutoff |
	perMethodProfile ifNil:
		[aStream nextPutAll: 'not collecting profile'.
		 ^self].
	cutoff := cutoffOrNilOrPair isArray ifTrue: [cutoffOrNilOrPair first] ifFalse: [cutoffOrNilOrPair].
	total := 0.
	aStream nextPutAll: 'bytecodes dispatched per method'; cr.
	(perMethodProfile associations sorted: [:a1 :a2| a1 value >= a2 value]) do:
		[:assoc| 
 		total := total + assoc value.
		(cutoff isNil or: [assoc value >= cutoff]) ifTrue:
			[self reportSingleMethodProfile: assoc on: aStream]].
	aStream flush.
	^{total}, (cogit
					reportPerMethodProfilingOn: aStream
					cutoff: (cutoffOrNilOrPair isArray ifTrue: [cutoffOrNilOrPair last] ifFalse: [cutoffOrNilOrPair]))! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 8/23/2021 11:09'!
reportSingleMethodProfile: assoc on: aStream
	"Generate a sorted report of bytecodes per method, and instructions per trampoline/cog method/pic.
	 Answer an Array with the four totals."
	| primIndex literal |
	aStream tab; print: assoc value; nextPut: $:; tab.
	self printNameOfClass: (self methodClassOf: assoc key) count: 2 on: aStream.
	aStream next: 2 put: $>.
	self printStringOf: (self findSelectorOfMethod: assoc key) on: aStream.
	primIndex := self primitiveIndexOf: assoc key.
	(primIndex = 0
	or: [(self isQuickPrimitiveIndex: primIndex)
	or: [(primitiveTable at: primIndex + 1) isInteger]]) ifFalse:
		[aStream space; nextPutAll: (primitiveTable at: primIndex + 1).
		 (cogit primitiveGeneratorForPrimitiveIndex: primIndex) ifNotNil:
			[:descriptor|
			 descriptor primitiveGenerator ifNotNil:
				[:generator| aStream space; nextPutAll: generator]].
		 primIndex = PrimNumberExternalCall ifTrue:
			[literal := self literal: 0 ofMethod: assoc key.
			 aStream
				space;
				nextPutAll: (self stringOf: (objectMemory fetchPointer: ExternalCallLiteralModuleNameIndex ofObject: literal));
				space;
				nextPutAll: (self stringOf: (objectMemory fetchPointer: ExternalCallLiteralFunctionNameIndex ofObject: literal))]].
	aStream cr! !

!CogVMSimulator methodsFor: 'enilopmarts' stamp: 'eem 12/25/2019 13:50'!
returnToExecutive: inInterpreter postContextSwitch: switchedContext
	"Return to the current frame, either by entering machine code, or longjmp-ing back to the
	 interpreter or simply returning, depending on where we are. To know whether to return or
	 enter machine code we have to know from whence we came.  We could have come from
	 the interpreter, either directly or via a machine code primitive.  We could have come from
	 machine code.  The instructionPointer tells us where from.  If it is above startOfMemory we're
	 in the interpreter.  If it is below, then we are in machine-code unless it is ceReturnToInterpreterPC,
	 in which case we're in a machine-code primitive called from the interpreter.

	 Override to nil localFP.  This mimics the production VM's housing of localFP (& localIP & localSP)
	 in a local (hopefully register) variable in interpret, and hence outside of interpret it is invalid.
	 In the simulator it is also used as the flag as to whether the head frame is to be derived from
	 localFP & localSP (when localFP notNl) or from framePointer & stackPointer (when localFP is nil)."
	localFP := nil.
	^super returnToExecutive: inInterpreter postContextSwitch: switchedContext! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
returnToSchedulingLoopAndReleaseVMOrWakeThread: vmThread source: source
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #returnToSchedulingLoopAndReleaseVMOrWakeThread:source:
		withArguments: {vmThread. source}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'image save/restore' stamp: 'eem 8/13/2018 15:21'!
reverseBytesInImage
	UIManager default
		informUser: 'Swapping bytes of foreign image...'
		during: [super reverseBytesInImage]! !

!CogVMSimulator methodsFor: 'method lookup cache' stamp: 'nice 11/6/2016 13:33'!
rewriteMethodCacheEntryForExternalPrimitiveToFunction: localPrimAddress
	super rewriteMethodCacheEntryForExternalPrimitiveToFunction:
				(self mapFunctionToAddress: (localPrimAddress = 0
												ifTrue: [#primitiveFail]
												ifFalse: [localPrimAddress])).
	"Hack; the super call will rewrite the entry to the address of the function.
	 So (when simulating) undo the damage and put back the functionPointer"
	(methodCache at: lastMethodCacheProbeWrite + MethodCacheMethod) = newMethod ifTrue:
		[methodCache
			at: lastMethodCacheProbeWrite + MethodCachePrimFunction
			put: (self cCoerce: localPrimAddress to: #'sqIntptr_t')]! !

!CogVMSimulator methodsFor: 'rump c stack' stamp: 'eem 9/30/2013 16:33'!
rumpCStackAddress
	| alignment alignedHeapBase |
	alignment := 64. "byte alignment for stack frames; IA32 requires 16 bytes."
	alignedHeapBase := heapBase bitAnd: alignment negated.
	^alignedHeapBase = heapBase
		ifTrue: [heapBase - alignment]
		ifFalse: [alignedHeapBase]! !

!CogVMSimulator methodsFor: 'rump c stack' stamp: 'eem 12/25/2010 14:25'!
rumpCStackSize
	"Allocate a rump C stack for simulation.  This stack is used for C calls, i.e. calls
	 made into the simulated C run-time, typically of routines in CoInterpreter.
	 We also use it to hold in-memory copies of the cStackPointer and cFramePointer
	 variables for simulation speed.  Alas we don't use an in-memory copy of stackLimit,
	 because we need some activity to count in order to implement the ioMSecs
	 routine and the stackLimit check is the ideal one."
	^self threadManager
		ifNil: [RumpCStackSize]
		ifNotNil: [:threadManager| RumpCStackSize * (threadManager maxNumThreads + 1)]! !

!CogVMSimulator methodsFor: 'testing' stamp: 'eem 12/29/2016 08:27'!
run
	"Just run"
	quitBlock := [displayView ifNotNil:
				   [displayView containingWindow ifNotNil:
					[:topWindow|
					((World submorphs includes: topWindow)
					 and: [UIManager default confirm: 'close?']) ifTrue:
						[topWindow delete]]].
				  ^self close].
	self initStackPages.
	self loadInitialContext.
	self initialEnterSmalltalkExecutive! !

!CogVMSimulator methodsFor: 'primitive support' stamp: 'eem 12/7/2017 17:56'!
saneFunctionPointerForFailureOfPrimIndex: primIndex
	"For simulation override to do the check specially when the
	 primitiveFunctionPointer is an invalid address proxy for a primitive."
	| basePrimitive |
	(instructionPointer < objectMemory nilObject asUnsignedInteger
	and: [primitiveFunctionPointer isInteger
	and: [self isPrimitiveFunctionPointerAnIndex not
	and: [primIndex ~= PrimNumberExternalCall
	and: [(self isMetaPrimitiveIndex: primIndex) not]]]]) ifTrue:
		[basePrimitive := self functionPointerFor: primIndex inClass: objectMemory nilObject.
		 ^(cogit lookupAddress: primitiveFunctionPointer) endsWith: basePrimitive].

	^super saneFunctionPointerForFailureOfPrimIndex: primIndex! !

!CogVMSimulator methodsFor: 'I/O primitives' stamp: 'eem 10/11/2013 09:56'!
savedWindowSize
	^savedWindowSize ifNil: [0]! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
sendInvokeCallbackContext: vmCallbackContext
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #sendInvokeCallbackContext:
		withArguments: {vmCallbackContext}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'UI' stamp: 'eem 3/18/2020 16:35'!
setBreakBlockFromString: aString for: coInterpreterOrCogit
	| bString block |
	bString := aString withBlanksTrimmed.
	bString first = $- ifTrue:
		[^cogit breakBlock: nil].
	bString first ~= $[ ifTrue:
		[bString := '[:_address|', bString, ']'].
	block := [Compiler evaluate: bString for: coInterpreterOrCogit logged: false]
				on: Error
				do: [:ex|
					UIManager default warn: ex messageText.
					^self].
	cogit breakBlock: block.
	(atEachStepBlock notNil and: [atEachStepBlock home selector == #useCogitBreakBlockIfNone]) ifTrue:
		[atEachStepBlock := nil]! !

!CogVMSimulator methodsFor: 'accessing' stamp: 'eem 1/5/2018 15:19'!
setBreakCount: anInteger
	breakCount := anInteger! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 10/31/2020 17:59'!
setCFramePointer: cfp
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #setCFramePointer:
		withArguments: {cfp}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 10/31/2020 18:00'!
setCStackPointer: csp
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #setCStackPointer:
		withArguments: {csp}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 10/08/2020 11:20'!
setImageHeaderFlags: flags
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #setImageHeaderFlags:
		withArguments: {flags}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
setImageHeaderFlagsFrom: headerFlags
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #setImageHeaderFlagsFrom:
		withArguments: {headerFlags}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
setMaxWaitingPriorityTo: minPriority
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #setMaxWaitingPriorityTo:
		withArguments: {minPriority}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
setOwnerIndexOfProcess: aProcess to: anIndex bind: bind
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #setOwnerIndexOfProcess:to:bind:
		withArguments: {aProcess. anIndex. bind}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 12/3/2010 07:23'!
shortPrintContext: aContext
	transcript ensureCr.
	^super shortPrintContext: aContext! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 10/11/2013 12:37'!
shortPrintFrame: theFP
	self transcript ensureCr.
	^super shortPrintFrame: theFP! !

!CogVMSimulator methodsFor: 'process primitive support' stamp: 'eem 10/23/2020 12:53'!
signalSemaphoreWithIndex: index
	"Record the given semaphore index in the double buffer semaphores array to be signaled
	 at the next convenient moment. Force a real interrupt check as soon as possible.
	 This is a simulation.  See platforms/Cross/vm/sqExternalSemaphores.c for the real code.
	 Thanks to Levente Uzoni for making this version almost thread-safe (in Smalltalk)"
	<doNotGenerate>
	| originalResponses newRequests newResponses |
	index <= 0 ifTrue: [^false].
	index > externalSemaphoreSignalRequests size ifTrue:
		[newRequests := Array new: 1 << index highBit withAll: 0.
		 newResponses := newRequests shallowCopy].
	"This is a lock-free thread-safe grow...; thanks Levente"
	originalResponses := externalSemaphoreSignalResponses.
	[index > externalSemaphoreSignalRequests size] whileTrue:
		[newRequests
			replaceFrom: 1
			to: externalSemaphoreSignalRequests size
			with: externalSemaphoreSignalRequests
			startingAt: 1.
		 newResponses
			replaceFrom: 1
			to: externalSemaphoreSignalResponses size
			with: externalSemaphoreSignalResponses
			startingAt: 1.
		 externalSemaphoreSignalResponses == originalResponses "This should always be true."
			ifTrue:
				[externalSemaphoreSignalRequests := newRequests.
				 externalSemaphoreSignalResponses := newResponses]
			ifFalse:
				[originalResponses := externalSemaphoreSignalResponses]].
	"This is not thread-safe however..."
	externalSemaphoreSignalRequests
		at: index
		put: (externalSemaphoreSignalRequests at: index) + 1.
	^true! !

!CogVMSimulator methodsFor: 'primitive support' stamp: 'eem 12/25/2019 13:52'!
slowPrimitiveResponse
	"Invoke a normal (non-quick) primitive.
	 Called under the assumption that primFunctionPointer has been preloaded.

	 Override to nil localFP.  This mimics the production VM's housing of localFP (& localIP & localSP)
	 in a local (hopefully register) variable in interpret, and hence outside of interpret it is invalid.
	 In the simulator it is also used as the flag as to whether the head frame is to be derived from
	 localFP & localSP (when localFP notNl) or from framePointer & stackPointer (when localFP is nil)."
	localFP := nil.
	^super slowPrimitiveResponse! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 6/20/2008 15:21'!
space

	traceOn ifTrue: [ transcript space ]! !

!CogVMSimulator methodsFor: 'plugin support' stamp: 'ar 5/11/2000 21:47'!
sqGetInterpreterProxy
	"I am basically my own proxy..."
	^self! !

!CogVMSimulator methodsFor: 'memory access' stamp: 'ar 2/25/2001 17:16'!
sqMemoryExtraBytesLeft: includingSwap
	^0! !

!CogVMSimulator methodsFor: 'memory access' stamp: 'di 5/11/2004 18:29'!
sqShrinkMemory: oldLimit By: delta
	transcript show: 'shrink memory from ', oldLimit printString, ' by ', delta printString, ' remember it doesn''t actually shrink in simulation'; cr.

	^ oldLimit! !

!CogVMSimulator methodsFor: 'I/O primitives support' stamp: 'eem 1/17/2018 20:04'!
stackLimitFromMachineCode
	"Intercept accesses to the stackLimit from machine code to increment byteCount so that
	 ioMSecs/ioMicroseconds does something reasonable when we're purely in machine code.
	 Force an interrupt check every 2 ms in simulated time (2ms = the default heartbeat), or if
	 the profile tick has expired.."

	(byteCount := byteCount + 1) - lastPollCount >= (2000 * ByteCountsPerMicrosecond) ifTrue:
		[lastPollCount := byteCount.
		 self doOrDefer: [self changed: #byteCountText; changed: #composeAll]].
	(lastPollCount = byteCount
	 or: [nextProfileTick > 0 and: [nextProfileTick <= self ioUTCMicroseconds]]) ifTrue:
		[suppressHeartbeatFlag "gets set by selector breakpoints"
			ifTrue: [self forceInterruptCheck]
			ifFalse: [self forceInterruptCheckFromHeartbeat]].
	^stackLimit! !

!CogVMSimulator methodsFor: 'stack pages' stamp: 'eem 12/28/2008 18:08'!
stackPage
	^stackPage! !

!CogVMSimulator methodsFor: 'accessing' stamp: 'eem 12/29/2008 14:53'!
stackPages
	^stackPages! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 1/8/2009 16:39'!
stackPointer
	^stackPointer! !

!CogVMSimulator methodsFor: 'stack pages' stamp: 'eem 1/26/2020 11:40'!
stackZoneBase
	"In the simulator the stack zone starts immediately after the code zone."
	^cogit cogCodeBase + self effectiveCogCodeSize! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 9/6/2020 14:28'!
startInContextSuchThat: aBlock
	"Change the active process's suspendedContext to its sender, which short-cuts the
	 initialization of the system.  This can be a short-cut to running code, e.g. when doing
		Smalltalk saveAs.
		Compiler recompileAll
	 via e.g.
		vm startInContextSuchThat: [:ctxt| (vm stringOf: (vm penultimateLiteralOf: (vm methodForContext: ctxt))) = 'DoIt']"
	<doNotGenerate>
	| context activeProc |
	activeProc := self activeProcess.
	context := objectMemory fetchPointer: SuspendedContextIndex ofObject: activeProc.
	[context = objectMemory nilObject ifTrue:
		[^self error: 'no context found'].
	 aBlock value: context] whileFalse:
		[context := objectMemory fetchPointer: SenderIndex ofObject: context].
	objectMemory storePointer: SuspendedContextIndex ofObject: activeProc withValue: context.
	"Now push a dummy return value."
	objectMemory
		storePointerUnchecked: (self fetchStackPointerOf: context) + CtxtTempFrameStart
		ofObject: context
		withValue: objectMemory nilObject.
	self storeInteger: StackPointerIndex
		ofObject: context
		withValue: (self fetchStackPointerOf: context) + 1! !

!CogVMSimulator methodsFor: 'debugging traps' stamp: 'eem 6/9/2010 22:14'!
startOfMemory
	self shouldNotImplement! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 2/22/2009 19:22'!
systemAttributes
	^systemAttributes! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 6/20/2008 17:05'!
tab

	traceOn ifTrue: [ transcript tab ].! !

!CogVMSimulator methodsFor: 'return bytecodes' stamp: 'eem 6/22/2010 11:21'!
tearDownAndRebuildFrameForCannotReturnBaseFrameReturnFrom: contextToReturnFrom to: contextToReturnTo returnValue: returnValue
	self halt.
	^super tearDownAndRebuildFrameForCannotReturnBaseFrameReturnFrom: contextToReturnFrom to: contextToReturnTo returnValue: returnValue! !

!CogVMSimulator methodsFor: 'testing' stamp: 'eem 3/6/2009 18:05'!
test
	self testBreakCount: -1 printSends: true printFrames: false printBytecodes: false! !

!CogVMSimulator methodsFor: 'testing' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
testBecome
	"Become some young things.  AA testBecome    "
	| array list1 list2 p1 p2 p3 p4 |
	array := objectMemory splObj: ClassArray.
	list1 := objectMemory instantiateClass: array indexableSize: 2.
	list2 := objectMemory instantiateClass: array indexableSize: 2.
	p1 := objectMemory instantiateClass: (objectMemory splObj: ClassPoint) indexableSize: 0.
	self push: p1.
	objectMemory storePointer: 0 ofObject: list1 withValue: p1.
	p2 := objectMemory instantiateClass: (objectMemory splObj: ClassPoint) indexableSize: 0.
	self push: p2.
	objectMemory storePointer: 1 ofObject: list1 withValue: p2.
	p3 := objectMemory instantiateClass: (objectMemory splObj: ClassMessage) indexableSize: 0.
	self push: p3.
	objectMemory storePointer: 0 ofObject: list2 withValue: p3.
	p4 := objectMemory instantiateClass: (objectMemory splObj: ClassMessage) indexableSize: 0.
	self push: p4.
	objectMemory storePointer: 1 ofObject: list2 withValue: p4.
	(objectMemory become: list1 with: list2 twoWay: true copyHash: true) ifFalse: [self error: 'failed'].
	self popStack = p2 ifFalse: [self halt].
	self popStack = p1 ifFalse: [self halt].
	self popStack = p4 ifFalse: [self halt].
	self popStack = p3 ifFalse: [self halt].
	(objectMemory fetchPointer: 0 ofObject: list1) = p3 ifFalse: [self halt].
	(objectMemory fetchPointer: 1 ofObject: list1) = p4 ifFalse: [self halt].
	(objectMemory fetchPointer: 0 ofObject: list2) = p1 ifFalse: [self halt].
	(objectMemory fetchPointer: 1 ofObject: list2) = p2 ifFalse: [self halt].! !

!CogVMSimulator methodsFor: 'testing' stamp: 'eem 12/29/2016 08:28'!
testBreakCount: breakCount printSends: shouldPrintSends printFrames: shouldPrintFrames printBytecodes: shouldPrintBytecodes
	self initStackPages.
	self loadInitialContext.
	transcript clear.
	quitBlock := [^self close].
	printSends := true & shouldPrintSends. "true & foo allows evaluating printFoo := true in the debugger"
	printFrameAtEachStep := true & shouldPrintFrames.
	printBytecodeAtEachStep := true & shouldPrintBytecodes.
	self ensureDebugAtEachStepBlock.
	self initialEnterSmalltalkExecutive! !

!CogVMSimulator methodsFor: 'testing' stamp: 'eem 9/5/2013 14:54'!
testPCMapping
	objectMemory allObjectsDo:
		[:o|
		((objectMemory isCompiledMethod: o)
		 and: [self methodShouldBeCogged: o]) ifTrue:
			[(self methodHasCogMethod: o) ifFalse:
				[[([cogit cog: o selector: objectMemory nilObject]
					on: Error
					do: [:ex|
						ex messageText = 'This won''t work...' ifTrue:
							[ex resumeUnchecked: nil].
						ex pass]) isNil
				   and: [cogCompiledCodeCompactionCalledFor]] whileTrue:
					[cogMethodZone clearCogCompiledCode.
					 cogCompiledCodeCompactionCalledFor := false]].
			 (self methodHasCogMethod: o)
				ifTrue:
					[transcript nextPut: $.; flush.
					 cogit testMcToBcPcMappingForMethod: (self cogMethodOf: o)]
				ifFalse:
					[transcript nextPutAll: 'failed to compile method '; print: o; cr; flush]]]! !

!CogVMSimulator methodsFor: 'testing' stamp: 'eem 3/6/2009 18:06'!
testWithBytecodePrint
	self testBreakCount: -1 printSends: true printFrames: false printBytecodes: true! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
threadIdFieldOf: aProcess
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #threadIdFieldOf:
		withArguments: {aProcess}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
threadManager
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #threadManager
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
threadSchedulingLoop: vmThread
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #threadSchedulingLoop:
		withArguments: {vmThread}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
threadSwitchIfNecessary: newProc from: sourceCode
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #threadSwitchIfNecessary:from:
		withArguments: {newProc. sourceCode}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 10/12/2009 01:05'!
transcript
	^transcript! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 7/18/2010 18:28'!
transcript: aTranscript
	transcript := aTranscript! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
transferTo: newProc from: sourceCode
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #transferTo:from:
		withArguments: {newProc. sourceCode}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
tryToExecuteSmalltalk: vmThread
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #tryToExecuteSmalltalk:
		withArguments: {vmThread}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:00'!
unableToReadImageError
	self error:  'Read failed or premature end of image file'! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem 4/16/2014 09:58'!
updateStateOfSpouseContextForFrame: theFP WithSP: theSP
	 "26431360 = (self frameContext: theFP) ifTrue:
		[self halt]."
	"((self stringOf: (self penultimateLiteralOf: (self frameMethodObject: theFP))) = #indexOf:startingAt:ifAbsent:) ifTrue:
		[self halt]."
	^super updateStateOfSpouseContextForFrame: theFP WithSP: theSP! !

!CogVMSimulator methodsFor: 'interpreter shell' stamp: 'eem 1/9/2020 15:20'!
useCogitBreakBlockIfNone
	atEachStepBlock ifNil:
		[cogit breakBlock ifNotNil:
			[: bb| atEachStepBlock := [(bb value: 0) ifTrue: [self halt: 'Cogit breakpoint reached']]]]! !

!CogVMSimulator methodsFor: 'debug support' stamp: 'eem 12/5/2015 14:57'!
veryDeepCopyWith: deepCopier
	"Override to short-circuit the copying of any VMPluginCodeGenerators referenced from mappedPluginEntries and
	 uniqueIndices. These can in turn hold onto Monticello state, resulting in a huge ammount of unnecessary copying."
	deepCopier references
		at: mappedPluginEntries ifAbsentPut: [mappedPluginEntries];
		at: uniqueIndices ifAbsentPut: [uniqueIndices].
	mappedPluginEntries do:
		[:tuple| | block |
		block := tuple third.
		deepCopier references at: block ifAbsentPut: [block]].
	uniqueIndices keysDo:
		[:block|
		deepCopier references at: block ifAbsentPut: [block]].
	^super veryDeepCopyWith: deepCopier! !

!CogVMSimulator methodsFor: 'file primitives' stamp: 'eem 9/7/2018 19:42'!
vmPathGet: stringBase Length: stringSize
	| pathName stringOop |
	pathName := self vmPath.
	stringOop := stringBase - objectMemory baseHeaderSize. "Due to C call in Interp"
	1 to: stringSize do:
		[:i | objectMemory storeByte: i-1 ofObject: stringOop
			withValue: (pathName at: i) asciiValue].
! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 9/30/2010 13:21'!
waitingPriorityIsAtLeast: minPriority
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #waitingPriorityIsAtLeast:
		withArguments: {minPriority}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'multi-threading simulation switch' stamp: 'eem 11/20/2010 21:51'!
wakeHighestPriority
	"This method includes or excludes CoInterpreterMT methods as required.
	 Auto-generated by CogVMSimulator>>ensureMultiThreadingOverridesAreUpToDate"

	^self perform: #wakeHighestPriority
		withArguments: {}
		inSuperclass: (cogThreadManager ifNil: [CoInterpreterPrimitives] ifNotNil: [CoInterpreterMT])! !

!CogVMSimulator methodsFor: 'I/O primitives support' stamp: 'yo 6/3/1999 16:10'!
warpBits

	^ myBitBlt warpBits! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 12/22/2019 18:25'!
whereIs: anOop
	(self isOnRumpCStack: anOop) ifTrue: [^' is on rump C stack'].
	^super whereIs: anOop! !

!CogVMSimulator methodsFor: 'primitive support' stamp: 'eem 11/20/2020 07:30'!
windowIsClosing
	self threadManager ifNotNil:
		[:threadManager| threadManager shutdownModule].
	quitBlock ifNotNil:
		[:effectiveQuitBlock|
		quitBlock := nil. "stop recursion on explicit window close."
		[effectiveQuitBlock value]
			on: BlockCannotReturn
			do: [:ex|]]	"Cause return from #test, et al"! !

!CogVMSimulator methodsFor: 'debug printing' stamp: 'eem 11/18/2016 17:03'!
withMemoryProtectionDo: aBlock
	objectMemory writeProtectMemory.
	stackPages writeProtectMemory.
	^aBlock ensure:
		[objectMemory writeEnableMemory.
		 stackPages writeEnableMemory]! !

!CogVMSimulator methodsFor: 'frame access' stamp: 'eem 6/23/2008 13:57'!
withoutSmallIntegerTags: anInteger
	self assert: (anInteger >= 0 and: [anInteger highBit <= 32]).
	^(anInteger bitAnd: 16r80000000) ~= 0
		ifTrue: ["negative"
				(anInteger bitAnd: 16r7FFFFFFE) - 16r80000000]
		ifFalse: ["positive"
				anInteger - 1]! !

!CogVMSimulator methodsFor: 'simulation only' stamp: 'eem 8/6/2021 13:22'!
zeroNextProfileTick
	zeroNextProfileTickCount := zeroNextProfileTickCount + 1.
	super zeroNextProfileTick.
	self ifTestProfilingAdvanceNextProfileTick! !

CogVMSimulator subclass: #CogVMSimulatorLSB
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CogVMSimulatorLSB commentStamp: '<historical>' prior: 0!
This class overrides a few methods in CogVMSimulator required for simulation to work on little-endian architectures (such as the x86 family of processors).  To start it up simply use CogVMSimulatorLSB instead of CogVMSimulator (see the class comment there for more details).  For example:

	(CogVMSimulatorLSB new openOn: Smalltalk imageName) test

Note that the image must have been saved at least once on the local architecture, since the compiled VM performs some byte swapping that the simulator cannot cope with.!

!CogVMSimulatorLSB methodsFor: 'memory access' stamp: 'eem 6/25/2011 12:16'!
byteAt: byteAddress put: byte
	^objectMemory byteAt: byteAddress put: byte! !

!CogVMSimulatorLSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:59'!
halfWordHighInLong32: long32
	"Used by Balloon"

	^ long32 bitAnd: 16rFFFF! !

!CogVMSimulatorLSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:59'!
halfWordLowInLong32: long32
	"Used by Balloon"

	^ long32 bitShift: -16! !

!CogVMSimulatorLSB methodsFor: 'initialization' stamp: 'eem 10/9/2008 16:39'!
nextShortFrom: aStream
	"Read a 16-bit quantity from the given (binary) stream."
	^aStream nextLittleEndianNumber: 2! !

!CogVMSimulatorLSB methodsFor: 'memory access' stamp: 'tpr 12/22/2005 17:46'!
vmEndianness
	"return 0 for little endian, 1 for big endian"
	^0! !

!CoInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 3/29/2013 14:52'!
frameIsMarked: theFPInt
	| methodField |
	methodField := stackPages longAt: theFPInt + FoxMethod.
	^methodField asUnsignedInteger < objectMemory startOfMemory
		ifTrue: [(methodField bitAnd: 4) ~= 0]
		ifFalse: [((stackPages longAt: theFPInt + FoxIFrameFlags) bitAnd: 2) ~= 0]! !

!CoInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 3/29/2013 14:54'!
markFrame: theFPInt
	| methodField |
	methodField := stackPages longAt: theFPInt + FoxMethod.
	methodField asUnsignedInteger < objectMemory startOfMemory
		ifTrue:
			[stackPages
				longAt: theFPInt + FoxMethod
				put: (methodField bitOr: 4)]
		ifFalse:
			[stackPages
				longAt: theFPInt + FoxIFrameFlags
				put: ((stackPages longAt: theFPInt + FoxIFrameFlags) bitOr: 2)]! !

!CoInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'nice 4/23/2016 16:23'!
pathTo: goal using: stack followWeak: followWeak
	"Trace objects and frames from the root, marking visited objects, pushing the current path on stack, until goal is found.
	 If found, unmark, leaving path in stack, and answer 0.  Otherwise answer an error:
		PrimErrBadArgument if stack is not an Array
		PrimErrBadIndex if search overflows stack
		PrimErrNotFound if goal cannot be found"
	| current index next stackSize stackp freeStartAtStart |
	<var: #index type: #sqInt> "Force the sign because typeInference does not seem to work"
	(objectMemory isArray: stack) ifFalse:
		[^PrimErrBadArgument].
	self assert: objectMemory allObjectsUnmarked.
	freeStartAtStart := objectMemory freeStart. "check no allocations during search"
	objectMemory beRootIfOld: stack. "so no store checks are necessary on stack"
	stackSize := objectMemory lengthOf: stack.
	objectMemory mark: stack.
	"no need. the current context is not reachable from the active process (suspendedContext is nil)"
	"objectMemory mark: self activeProcess."
	current := objectMemory specialObjectsOop.
	objectMemory mark: current.
	index := objectMemory lengthOf: current.
	stackp := 0.
	[[(index := index - 1) >= -1] whileTrue:
		[(stackPages couldBeFramePointer: current)
			ifTrue:
				[next := index >= 0
							ifTrue: [self field: index ofFrame: (self cCoerceSimple: current to: #'char *')]
							ifFalse: [objectMemory nilObject]]
			ifFalse:
				[index >= 0
					ifTrue:
						[next := (objectMemory isContextNonImm: current)
									ifTrue: [self fieldOrSenderFP: index ofContext: current]
									ifFalse: [objectMemory fetchPointer: index ofObject: current]]
					ifFalse:
						[next := objectMemory fetchClassOfNonImm: current]].
		 (stackPages couldBeFramePointer: next)
			ifTrue: [self assert: (self isFrame: (self cCoerceSimple: next to: #'char *')
										onPage: (stackPages stackPageFor: (self cCoerceSimple: next to: #'char *')))]
			ifFalse:
				[next >= heapBase ifTrue: "exclude Cog methods"
					[self assert: (self checkOkayOop: next)]].
		 next = goal ifTrue:
			[self assert: freeStartAtStart = objectMemory freeStart.
			 self unmarkAfterPathTo.
			 objectMemory storePointer: stackp ofObject: stack withValue: current.
			 self pruneStack: stack stackp: stackp.
			 ^0].
		 ((objectMemory isNonIntegerObject: next)
		  and: [(stackPages couldBeFramePointer: next)
				ifTrue: [(self frameIsMarked: next) not]
				ifFalse:
					[next >= heapBase "exclude Cog methods"
					  and: [(objectMemory isMarked: next) not
					  and: [((objectMemory isPointers: next) or: [objectMemory isCompiledMethod: next])
					  and: [followWeak or: [(objectMemory isWeakNonImm: next) not]]]]]])
			ifTrue:
				[stackp + 2 > stackSize ifTrue:
					[self assert: freeStartAtStart = objectMemory freeStart.
					 self unmarkAfterPathTo.
					 objectMemory nilFieldsOf: stack.
					 ^PrimErrBadIndex]. "PrimErrNoMemory ?"
				 objectMemory
					storePointerUnchecked: stackp ofObject: stack withValue: current;
					storePointerUnchecked: stackp + 1 ofObject: stack withValue: (objectMemory integerObjectOf: index).
				 stackp := stackp + 2.
				 (stackPages couldBeFramePointer: (self cCoerceSimple: next to: #'char *'))
					ifTrue:
						[self markFrame: next.
						index := self fieldsInFrame: (self cCoerceSimple: next to: #'char *')]
					ifFalse:
						[objectMemory mark: next.
						 (objectMemory isCompiledMethod: next)
							ifTrue: [index := (objectMemory literalCountOf: next) + LiteralStart]
							ifFalse: [index := objectMemory lengthOf: next]].
				 current := next]].
		 current = objectMemory specialObjectsOop ifTrue:
			[self assert: freeStartAtStart = objectMemory freeStart.
			 self unmarkAfterPathTo.
			 objectMemory nilFieldsOf: stack.
			^PrimErrNotFound].
		 index := objectMemory integerValueOf: (objectMemory fetchPointer: stackp - 1 ofObject: stack).
		 current := objectMemory fetchPointer: stackp - 2 ofObject: stack.
		 stackp := stackp - 2] repeat! !

!CoInterpreterPrimitives methodsFor: 'method introspection support' stamp: 'eem 4/19/2016 16:22'!
pcDataFor: cogMethod
	<var: #cogMethod type: #'CogMethod *'>
	| cm nSlots nEntries data |
	cm := cogMethod methodObject.
	nSlots := (objectMemory byteSizeOf: cm) - (self startPCOfMethod: cm) * 2 + objectMemory minSlotsForShortening.
	data := objectMemory instantiateClass: (objectMemory splObj: ClassArray) indexableSize: nSlots.
	data ifNil: [^-1].
	nEntries := cogit mapPCDataFor: cogMethod into: data.
	nEntries = 0 ifTrue:
		[^0].
	nEntries < nSlots ifTrue:
		[objectMemory shorten: data toIndexableSize: nEntries].
	^data! !

!CoInterpreterPrimitives methodsFor: 'method introspection support' stamp: 'eem 4/19/2016 16:10'!
picDataFor: cogMethod
	<var: #cogMethod type: #'CogMethod *'>
	| cm nSlots nEntries data |
	cm := cogMethod methodObject.
	nSlots := (objectMemory byteSizeOf: cm) - (self startPCOfMethod: cm) + objectMemory minSlotsForShortening.
	data := objectMemory instantiateClass: (objectMemory splObj: ClassArray) indexableSize: nSlots.
	data ifNil: [^-1].
	nEntries := cogit picDataFor: cogMethod into: data.
	nEntries = 0 ifTrue:
		[^0].
	nEntries < nSlots ifTrue:
		[objectMemory shorten: data toIndexableSize: nEntries].
	^data! !

!CoInterpreterPrimitives methodsFor: 'method introspection primitives' stamp: 'cb 4/26/2016 09:31'!
primitiveAllMethodsCompiledToMachineCode
	<export: true>
	<option: #SistaCogit>
	| arrayObj nEntries maxNumEntries |
	argumentCount ~= 0 ifTrue:
		[^self primitiveFailFor: PrimErrBadNumArgs].
	maxNumEntries := cogit numMethods + objectMemory minSlotsForShortening.
	arrayObj := objectMemory 
		instantiateClass: (objectMemory splObj: ClassArray) 
		indexableSize: maxNumEntries.
	arrayObj ifNil: [^-1].
	nEntries := cogit methodsCompiledToMachineCodeInto: arrayObj.
	(nEntries < maxNumEntries) ifTrue:
		[objectMemory shorten: arrayObj toIndexableSize: nEntries].
	self pop: 1 thenPush: arrayObj! !

!CoInterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 2/19/2009 20:12'!
primitiveClosureCopyWithCopiedValues
	"This is optional old obsolete stuff tedious to implement in Cog because of
	 the need to map from machine code pc to bytecode pc.  So for now just fail."
	self primitiveFail! !

!CoInterpreterPrimitives methodsFor: 'process primitives' stamp: 'cb 6/9/2017 20:43'!
primitiveCollectCogCodeConstituents
	"Answer the contents of the code zone as an array of pair-wise element, address in ascending
	 address order. Answer a string for a runtime routine or abstract label (beginning, end, etc),
	 a CompiledMethod for a CMMethod, or a selector (presumably a Symbol) for a PIC.
	 If there is an argument and it is true, then collect inner information about the CogMethod."
	| constituents withDetails |
	argumentCount = 0
		ifTrue: [withDetails := false]
		ifFalse:
			[withDetails := self stackTop.
			 (withDetails = objectMemory trueObject
			  or: [withDetails = objectMemory falseObject]) ifFalse:
				[^self primitiveFailFor: PrimErrBadArgument].
			  withDetails := withDetails = objectMemory trueObject].
	constituents := cogit cogCodeConstituents: withDetails.
	constituents ifNil:
		[^self primitiveFailFor: PrimErrNoMemory].
	self pop: argumentCount + 1 thenPush: constituents! !

!CoInterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 9/24/2010 14:11'!
primitiveContextXray
	"Lift the veil from a context and answer an integer describing its interior state.
	 Used for e.g. VM tests so they can verify they're testing what they think they're testing.
	 0 implies a vanilla heap context.
	 Bit 0 = is or was married to a frame
	 Bit 1 = is still married to a frame
	 Bit 2 = frame is executing machine code
	 Bit 3 = has machine code pc (as opposed to nil or a bytecode pc)
	 Bit 4 = method is currently compiled to machine code"
	| context pc flags theFP theMethod |
	<var: #theFP type: #'char *'>
	context := self stackTop.
	pc := objectMemory fetchPointer: InstructionPointerIndex ofObject: context.
	(self isMarriedOrWidowedContext: context)
		ifTrue:
			[(self checkIsStillMarriedContext: context currentFP: framePointer)
				ifTrue: [theFP := self frameOfMarriedContext: context.
						(self isMachineCodeFrame: theFP)
							ifTrue: [flags := 7]
							ifFalse: [flags := 3]]
				ifFalse: [flags := 1]]
		ifFalse:
			[flags := 0].
	((objectMemory isIntegerObject: pc)
	 and: [(objectMemory integerValueOf: pc) < 0]) ifTrue:
		[flags := flags bitOr: 8].
	theMethod := objectMemory fetchPointer: MethodIndex ofObject: context.
	(self maybeMethodHasCogMethod: theMethod) ifTrue:
		[flags := flags bitOr: 16].
	self pop: 1 thenPush: (objectMemory integerObjectOf: flags)! !

!CoInterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 10/21/2010 21:59'!
primitiveEnterCriticalSection
	"Attempt to enter a CriticalSection/Mutex.  If not owned, set the owner to the current
	 process and answer false. If owned by the current process  answer true.   Otherwise
	 suspend the process.  Answer if the receiver is owned by the current process."
	| criticalSection owningProcessIndex owningProcess activeProc inInterpreter |
	argumentCount > 0
		ifTrue:
			[criticalSection := self stackValue: 1.  "rcvr"
			 activeProc := self stackTop]
		ifFalse:
			[criticalSection := self stackTop.  "rcvr"
			 activeProc := self activeProcess].
	owningProcessIndex := ExcessSignalsIndex. "CriticalSections are laid out like Semaphores"
	owningProcess := objectMemory fetchPointer: owningProcessIndex ofObject: criticalSection.
	owningProcess = objectMemory nilObject ifTrue:
		[objectMemory storePointer: owningProcessIndex
			ofObject: criticalSection
			withValue: activeProc.
		 ^self pop: argumentCount + 1 thenPush: objectMemory falseObject].
	owningProcess = activeProc ifTrue:
		[^self pop: argumentCount + 1 thenPush: objectMemory trueObject].
	"Arrange to answer false (unowned) when the process is resumed."
	self pop: argumentCount + 1 thenPush: objectMemory falseObject.
	"We're going to switch process, either to an interpreted frame or a machine
	 code frame. To know whether to return or enter machine code we have to
	 know from whence we came.  We could have come from the interpreter,
	 either directly or via a machine code primitive.  We could have come from
	 machine code.  The instructionPointer tells us where from:"
	inInterpreter := instructionPointer >= objectMemory startOfMemory.
	self addLastLink: activeProc toList: criticalSection.
	self transferTo: self wakeHighestPriority from: CSEnterCriticalSection.
	self forProcessPrimitiveReturnToExecutivePostContextSwitch: inInterpreter! !

!CoInterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem (objmem refactor) 6/14/2010 10:15'!
primitiveExitCriticalSection
	"Exit the critical section.
	 This may change the active process as a result."
	| criticalSection owningProcessIndex inInterpreter owningProcess |
	criticalSection := self stackTop.  "rcvr"
	owningProcessIndex := ExcessSignalsIndex. "CriticalSections are laid out like Semaphores"
	(self isEmptyList: criticalSection)
		ifTrue:
			[objectMemory storePointerUnchecked: owningProcessIndex
				ofObject: criticalSection
				withValue: objectMemory nilObject]
		ifFalse:
			["We're going to switch process, either to an interpreted frame or a machine
			  code frame. To know whether to return or enter machine code we have to
			  know from whence we came.  We could have come from the interpreter,
			  either directly or via a machine code primitive.  We could have come from
			  machine code.  The instructionPointer tells us where from:"
			 inInterpreter := instructionPointer >= objectMemory startOfMemory.
			 owningProcess := self removeFirstLinkOfList: criticalSection.
			 "store check unnecessary because aSemaphore referred to owningProcess
			  via its FirstLinkIndex slot before owningProcess was removed."
			 objectMemory storePointerUnchecked: owningProcessIndex
				ofObject: criticalSection
				withValue: owningProcess.
			 "Note that resume: isn't fair; it won't suspend the active process.
			  For fairness we must do the equivalent of a primitiveYield, but that
			  may break old code, so we stick with unfair resume:."
			 (self resume: owningProcess
				preemptedYieldingIf: preemptionYields
				from: CSExitCriticalSection) ifTrue:
						[self forProcessPrimitiveReturnToExecutivePostContextSwitch: inInterpreter]]! !

!CoInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 4/11/2012 08:53'!
primitiveFlushCacheByMethod
	"The receiver is a compiledMethod.  Clear all entries in the method lookup cache that
	 refer to this method, presumably because it has been redefined, overridden or removed.
	 Override to flush appropriate machine code caches also."
	super primitiveFlushCacheByMethod.
	cogit unlinkSendsTo: self stackTop andFreeIf: false! !

!CoInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 8/11/2011 10:29'!
primitiveFlushCacheBySelector
	"The receiver is a message selector.  Clear all entries in the method lookup cache
	 with this selector, presumably because an associated method has been redefined.
	 Override to also flush machine code caches."
	| selector probe |
	selector := self stackTop.
	probe := 0.
	1 to: MethodCacheEntries do:
		[:i | (methodCache at: probe + MethodCacheSelector) = selector ifTrue:
			[methodCache at: probe + MethodCacheSelector put: 0].
		probe := probe + MethodCacheEntrySize].
	(selector = (self specialSelector: 16) "at:"
	 or: [selector = (self specialSelector: 17) "at:put:"]) ifTrue:
		[self flushAtCache].
	cogit
		unlinkSendsOf: selector
		isMNUSelector: (selector = (objectMemory splObj: SelectorDoesNotUnderstand))! !

!CoInterpreterPrimitives methodsFor: 'trampoline support' stamp: 'eem 5/6/2010 16:00'!
primitiveFunctionPointerAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: primitiveFunctionPointer) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #primitiveFunctionPointer in: self]! !

!CoInterpreterPrimitives methodsFor: 'method introspection primitives' stamp: 'eem 4/19/2016 16:08'!
primitiveMethodPCData
	<export: true>
	| methodReceiver data |
	argumentCount ~= 0 ifTrue:
		[^self primitiveFailFor: PrimErrBadNumArgs].
	methodReceiver := self stackTop.
	data := 0.
	(self methodHasCogMethod: methodReceiver) ifTrue:
		[data := self pcDataFor: (self cogMethodOf: methodReceiver).
		 data = -1 ifTrue:
			[^self primitiveFailFor: PrimErrNoMemory]].
	data = 0 ifTrue:
		[data := objectMemory instantiateClass: (objectMemory splObj: ClassArray) indexableSize: 0].
	self pop: 1 thenPush: data! !

!CoInterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 7/5/2021 16:36'!
primitiveMethodXray
	"Lift the veil from a method and answer an integer describing the interior state
	 of its machine code.
	 Used for e.g. VM tests so they can verify they're testing what they think they're testing.
	 0 implies a vanilla method.
	 Bit 0 = method might be compiled to machine code
	 Bit 1 = method is currently compiled to machine code
	 Bit 2 = is compiled frameless.
	 Bit 3 = method refers to young object.
	 Bit 4 = method too big to be jitted (more than 64k of code, or needs more than 1.5Mb of stack space to compile)
	 Bit 5 = method contains unknown/unjittable bytecode
	 Bit 6 = method should not be jitted because it contains a primitive not to be called from machine code (unused)"
	| alreadyCogged flags cogMethod |
	<var: #cogMethod type: #'CogMethod *'>
	(self methodWithHeaderShouldBeCogged: (objectMemory methodHeaderOf: self stackTop))
		ifTrue:
			[alreadyCogged := self maybeMethodHasCogMethod: self stackTop.
			 flags := 1.
			 alreadyCogged ifFalse:
				[cogMethod := cogit cog: self stackTop selector: objectMemory nilObject.
				 (cogMethod = nil
				  and: [cogCompiledCodeCompactionCalledFor]) ifTrue:
					[self commenceCogCompiledCodeCompaction.
					 cogMethod := cogit cog: self stackTop selector: objectMemory nilObject].
			 cogMethod asInteger
				caseOf: {
					[MethodTooBig] -> [flags := 1 + 16].
					[EncounteredUnknownBytecode] -> [flags := 1 + 32].
					[ShouldNotJIT] -> [flags := 1 + 64] }
				otherwise: [self deny: (cogMethod asInteger between: MaxNegativeErrorCode and: NotFullyInitialized)]].
			 (flags = 1
			  and: [self maybeMethodHasCogMethod: self stackTop]) ifTrue:
				[cogMethod := self cogMethodOf: self stackTop.
				 flags := cogMethod stackCheckOffset = 0 ifTrue: [7] ifFalse: [3].
				 cogMethod cmRefersToYoung ifTrue:
					[flags := flags + 8].
				 alreadyCogged ifFalse:
					[cogit freeCogMethod: cogMethod]]]
		ifFalse: [flags := 0].
	self pop: 1 thenPush: (objectMemory integerObjectOf: flags)! !

!CoInterpreterPrimitives methodsFor: 'other primitives' stamp: 'eem 1/23/2013 15:44'!
primitiveMinimumUnusedHeadroom
	<export: true>
	self methodReturnValue: (self integerObjectOf: self minimumUnusedHeadroom)! !

!CoInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 08/08/2014 10:05'!
primitiveObjectAt
"Defined for CompiledMethods only"
	| thisReceiver rawHeader realHeader index |
	index  := self stackIntegerValue: 0.
	self successful ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	thisReceiver := self stackValue: 1.
	rawHeader := self rawHeaderOf: thisReceiver.
	realHeader := (self isCogMethodReference: rawHeader)
					ifTrue: [(self cCoerceSimple: rawHeader to: #'CogMethod *') methodHeader]
					ifFalse: [rawHeader].
	(index > 0
	 and: [index <= ((objectMemory literalCountOfMethodHeader: realHeader) + LiteralStart)]) ifFalse:
		[^self primitiveFailFor: PrimErrBadIndex].
	self pop: 2
		thenPush: (index = 1
					ifTrue: [realHeader]
					ifFalse: [objectMemory fetchPointer: index - 1 ofObject: thisReceiver])! !

!CoInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 9/9/2019 09:17'!
primitiveObjectAtPut
	"Store a literal into a CompiledMethod at the given index. Defined for CompiledMethods only.
	 We assume that if the user is using this on active code then they will use primitiveVoidVMStateForMethod
	 to discard the machine code as required."
	| thisReceiver rawHeader realHeader index newValue |
	newValue := self stackValue: 0.
	index := self stackValue: 1.
	(objectMemory isNonIntegerObject: index) ifTrue:
		[^self primitiveFailFor: PrimErrBadArgument].
	index := objectMemory integerValueOf: index.
	thisReceiver := self stackValue: 2.
	(objectMemory isObjImmutable: thisReceiver) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].
	rawHeader := self rawHeaderOf: thisReceiver.
	realHeader := (self isCogMethodReference: rawHeader)
					ifTrue: [(self cCoerceSimple: rawHeader to: #'CogMethod *') methodHeader]
					ifFalse: [rawHeader].
	(index > 0
	 and: [index <= ((objectMemory literalCountOfMethodHeader: realHeader) + LiteralStart)]) ifFalse:
		[^self primitiveFailFor: PrimErrBadIndex].
	index = 1
		ifTrue:
			[((objectMemory isNonIntegerObject: newValue)
			 or: [(objectMemory literalCountOfMethodHeader: newValue) ~= (objectMemory literalCountOfMethodHeader: realHeader)]) ifTrue:
				[^self primitiveFailFor: PrimErrBadArgument].
			 (self isCogMethodReference: rawHeader)
				ifTrue: [(self cCoerceSimple: rawHeader to: #'CogMethod *') methodHeader: newValue]
				ifFalse: [objectMemory storePointerUnchecked: 0 ofObject: thisReceiver withValue: newValue]]
		ifFalse:
			[objectMemory storePointer: index - 1 ofObject: thisReceiver withValue: newValue].
	self pop: 3 thenPush: newValue! !

!CoInterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 8/4/2021 15:31'!
primitiveProfileSemaphore
	"Primitive. Install the semaphore to be used for profiling, 
	or nil if no semaphore should be used.
	See also primitiveProfileStart."
	| sema flushState |
	<export: true>
	sema := self stackValue: 0.
	sema = objectMemory nilObject
		ifTrue:
			[flushState := profileSemaphore ~= objectMemory nilObject]
		ifFalse:
			[flushState := profileSemaphore = objectMemory nilObject.
			 (objectMemory isSemaphoreOop: sema) ifFalse:
				[^self primitiveFailFor: PrimErrBadArgument]].
	profileSemaphore := sema.
	profileProcess := profileMethod := objectMemory nilObject.
	"If we've switched profiling on or off we must void machine code
	 (and machine code pcs in contexts) since we will start or stop
	 testing the profile clock in machine code primitive invocations,
	 and so generate slightly different code from here on in."
	flushState
		ifTrue: [self flushExternalPrimitives]
		ifFalse: [self methodReturnReceiver]! !

!CoInterpreterPrimitives methodsFor: 'method introspection primitives' stamp: 'eem 4/19/2016 14:23'!
primitiveResetCountersInMethod
	<export: true>
	<option: #SistaCogit>
	| methodReceiver |
	argumentCount ~= 0 ifTrue:
		[^self primitiveFailFor: PrimErrBadNumArgs].
	methodReceiver := self stackTop.
	(self methodHasCogMethod: methodReceiver) ifTrue:
		[cogit resetCountersIn: (self cogMethodOf: methodReceiver)]! !

!CoInterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 5/20/2021 15:53'!
primitiveResume
	"Put this process on the scheduler's lists thus allowing it to proceed next time there is
	 a chance for processes of it's priority level.  It must go to the back of its run queue so
	 as not to preempt any already running processes at this level.  If the process's priority
	 is higher than the current process, preempt the current process."
	| proc ctxt inInterpreter |
	proc := self stackTop.  "rcvr"
	ctxt := objectMemory followField: SuspendedContextIndex ofObject: proc. "written this way to get better Slang inlining"
	(objectMemory isContext: ctxt) ifFalse:
		[^self primitiveFailFor: PrimErrBadReceiver].
	"We're about to switch process, either to an interpreted frame or a
	 machine code frame. To know whether to return or enter machine code
	 we have to know from whence we came.  We could have come from the
	 interpreter, either directly or via a machine code primitive.  We could have
	 come from machine code.  The instructionPointer tells us where from:"
	inInterpreter := instructionPointer >= objectMemory startOfMemory.
	(self resume: proc preemptedYieldingIf: preemptionYields from: CSResume) ifTrue:
		[self forProcessPrimitiveReturnToExecutivePostContextSwitch: inInterpreter]

	"Personally I would like to check MyList, which should not be one of the elements of the scheduler lists.
	 But there are awful race conditions in things like should:notTakeMoreThan: that mean we can't.
	 eem 9/27/2010 23:08, updated eem 5/20/2021 15:36. e.g.

	| proc field |
	proc := self stackTop.  ''rcvr''
	''We only have to check for myList being nil.  If it is nil then this is either the active process or
	 a process suspended with primitiveSuspend (and if it is the activeProcess suspendedContext will
	 be nil and the isContext: test will fail).  If it is not nil then either the process is waiting on some
	 semaphore-like list or on one of the scheduler's lists. If it is on some semaphore-like list it should
	 not resume.  If it is on one of the scheduler's lists it is runnable (already resumed).''
	field := objectMemory followField: MyListIndex ofObject: proc.
	objectMemory nilObject = field ifFalse:
		[^self primitiveFailFor: PrimErrInappropriate].
	field := objectMemory followField: SuspendedContextIndex ofObject: proc.
	(objectMemory isContext: field) ifFalse:
		[^self primitiveFailFor: PrimErrBadReceiver].
	''We're about to switch process, either to an interpreted frame or a
	 machine code frame. To know whether to return or enter machine code
	 we have to know from whence we came.  We could have come from the
	 interpreter, either directly or via a machine code primitive.  We could have
	 come from machine code.  The instructionPointer tells us where from:''
	inInterpreter := instructionPointer >= objectMemory startOfMemory.
	(self resume: proc preemptedYieldingIf: preemptionYields from: CSResume) ifTrue:
		[self forProcessPrimitiveReturnToExecutivePostContextSwitch: inInterpreter]"! !

!CoInterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
primitiveSignal
	"Synchronously signal the semaphore.
	 This may change the active process as a result."
	| inInterpreter |
	"We may be about to switch process, either to an interpreted frame or a
	 machine code frame. To know whether to return or enter machine code
	 we have to know from whence we came.  We could have come from the
	 interpreter, either directly or via a machine code primitive.  We could have
	 come from machine code.  The instructionPointer tells us where from:"
	inInterpreter := instructionPointer >= objectMemory startOfMemory.
	(self synchronousSignal: self stackTop) ifTrue:
		[self forProcessPrimitiveReturnToExecutivePostContextSwitch: inInterpreter]! !

!CoInterpreterPrimitives methodsFor: 'method introspection primitives' stamp: 'eem 4/19/2016 14:23'!
primitiveSistaMethodPICAndCounterData
	<export: true>
	<option: #SistaCogit>
	| methodReceiver data |
	argumentCount ~= 0 ifTrue:
		[^self primitiveFailFor: PrimErrBadNumArgs].
	methodReceiver := self stackTop.
	data := 0.
	(self methodHasCogMethod: methodReceiver) ifTrue:
		[data := self picDataFor: (self cogMethodOf: methodReceiver).
		 data = -1 ifTrue:
			[^self primitiveFailFor: PrimErrNoMemory]].
	data = 0 ifTrue:
		[data := objectMemory instantiateClass: (objectMemory splObj: ClassArray) indexableSize: 0].
	self pop: 1 thenPush: data! !

!CoInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 3/29/2020 17:24'!
primitiveSnapshot
	"Save a normal snapshot under the same name as it was loaded
	 unless it has been renamed by the last primitiveImageName.

	 Override to jump to the interpreter because the machine code zone is now void."
	<inline: false>
	self snapshot: false.
	(self iframeMethod: framePointer) = newMethod ifTrue:
		["snapshot: has reached the end and built a frame.
		 In the JIT we need to back-up the pc before reentering the interpreter."
		instructionPointer := instructionPointer - 1].
	cogit ceInvokeInterpret
	"NOTREACHED"! !

!CoInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 3/29/2020 17:24'!
primitiveSnapshotEmbedded
	"Save an embedded snapshot.

	 Override to jump to the interpreter because the machine code zone is now void."
	<inline: false>
	self snapshot: true.
	(self iframeMethod: framePointer) = newMethod ifTrue:
		["snapshot: has reached the end and built a frame.
		 In the JIT we need to back-up the pc before reentering the interpreter."
		instructionPointer := instructionPointer - 1].
	cogit ceInvokeInterpret
	"NOTREACHED"! !

!CoInterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 8/29/2020 19:18'!
primitiveSuspend
	"Primitive. Suspend the receiver, aProcess such that it can be executed again
	by sending #resume. If the given process is not currently running, take it off
	its corresponding list. The primitive returns the list the receiver was previously on."
	| process myList |
	process := self stackTop.
	process = self activeProcess ifTrue:
		[| inInterpreter |
		"We're going to switch process, either to an interpreted frame or a machine
		 code frame. To know whether to return or enter machine code we have to
		 know from whence we came.  We could have come from the interpreter,
		 either directly or via a machine code primitive.  We could have come from
		 machine code.  The instructionPointer tells us where from:"
		self pop: 1 thenPush: objectMemory nilObject.
		inInterpreter := instructionPointer >= objectMemory startOfMemory.
		self transferTo: self wakeHighestPriority from: CSSuspend.
		^self forProcessPrimitiveReturnToExecutivePostContextSwitch: inInterpreter].
	myList := objectMemory fetchPointer: MyListIndex ofObject: process.
	"XXXX Fixme. We should really check whether myList is a kind of LinkedList or not
	but we can't easily so just do a quick check for nil which is the most common case."
	myList = objectMemory nilObject ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	"Alas in Spur we need a read barrier"
	(objectMemory isForwarded: myList) ifTrue:
		[myList := objectMemory followForwarded: myList.
		 objectMemory storePointer: MyListIndex ofObject: process withValue: myList].
	self removeProcess: process fromList: myList.
	self successful ifTrue:
		[objectMemory storePointerUnchecked: MyListIndex ofObject: process withValue: objectMemory nilObject.
		 self pop: 1 thenPush: myList]! !

!CoInterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem (const to msg rewrite) 11/5/2014 18:08'!
primitiveTerminateTo
	"Primitive. Terminate up the context stack from the receiver up to but not including
	 the argument, if previousContext is on my Context stack. Make previousContext my
	 sender. This prim has to shadow the code in ContextPart>terminateTo: to be correct.

	 Override to ensure the caller's saved ip is correct, i.e. if an interpreter frame it may
	 have to move to iframeSavedIP."
	| thisCtx currentCtx aContextOrNil contextsFP contextsSP contextsIP nextCntx stackedReceiverOffset 
	  theFP newFP newSP pageToStopOn thePage frameAbove |
	<var: #contextsFP type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #newFP type: #'char *'>
	<var: #newSP type: #'char *'>
	<var: #contextsIP type: #usqInt>
	<var: #frameAbove type: #'char *'>
	<var: #contextsSP type: #'char *'>
	<var: #source type: #'char *'>
	<var: #pageToStopOn type: #'StackPage *'>
	<var: #thePage type: #'StackPage *'>

	aContextOrNil := self stackTop.
	(aContextOrNil = objectMemory nilObject or: [objectMemory isContext: aContextOrNil]) ifFalse:
		[^self primitiveFail].
	thisCtx := self stackValue: 1.
	thisCtx = aContextOrNil ifTrue:
		[^self primitiveFail].		

	"All stackPages need to have current head pointers to avoid confusion."
	self externalWriteBackHeadFramePointers.

	"If we're searching for aContextOrNil it might be on a stack page.  Helps to know
	 if we can free a whole page or not, or if we can short-cut the termination."
	(aContextOrNil ~= objectMemory nilObject and: [self isStillMarriedContext: aContextOrNil])
		ifTrue: [contextsFP := self frameOfMarriedContext: aContextOrNil.
				pageToStopOn := stackPages stackPageFor: contextsFP]
		ifFalse: [pageToStopOn := 0].

	"if thisCtx is married ensure it is a base frame.  Then we can assign its sender."
	(self isStillMarriedContext: thisCtx)
		ifTrue:
			[theFP := self frameOfMarriedContext: thisCtx.
			 "Optimize terminating thisContext.  Move its frame down to be next to
			  aContextOrNil's frame. Common in the exception system and so helps to be fast."
			 (theFP = framePointer
			  and: [pageToStopOn = stackPage]) ifTrue:
				[self assertValidStackedInstructionPointersIn: stackPage line: #'__LINE__'.
				 (self frameCallerFP: theFP) ~= contextsFP ifTrue:
					[stackedReceiverOffset := self frameStackedReceiverOffset: theFP.
					 frameAbove := self findFrameAbove: contextsFP inPage: pageToStopOn.
					 contextsIP := (self frameCallerSavedIP: frameAbove) asUnsignedInteger.
					 self assert: ((contextsIP asUnsignedInteger >= objectMemory startOfMemory)
								or: [contextsIP = cogit ceReturnToInterpreterPC]) == (self isMachineCodeFrame: contextsFP) not.
					 newSP := self frameCallerSP: frameAbove.
					 newFP := newSP - stackedReceiverOffset - objectMemory wordSize.
					 theFP + stackedReceiverOffset
						to: stackPointer
						by: objectMemory wordSize negated
						do: [:source|
							newSP := newSP - objectMemory wordSize.
							stackPages longAt: newSP put: (stackPages longAt: source)].
					 stackPages longAt: newFP + FoxSavedFP put: contextsFP.
					"Ensure contract between machine-code callee and interpreter caller frames is preserved.
					 Return pc needs to be ceReturnToInterpreterPC."
					 ((self isMachineCodeFrame: newFP)
					  and: [contextsIP >= objectMemory startOfMemory]) ifTrue:
						[self iframeSavedIP: contextsFP put: contextsIP.
						 contextsIP := cogit ceReturnToInterpreterPC].
					 stackPages longAt: newFP + FoxCallerSavedIP put: contextsIP.
					 self assert: (objectMemory isContext: thisCtx).
					 objectMemory storePointerUnchecked: SenderIndex
						ofObject: thisCtx
						withValue: (self withSmallIntegerTags: newFP).
					 objectMemory storePointerUnchecked: InstructionPointerIndex
						ofObject: thisCtx
						withValue: (self withSmallIntegerTags: contextsFP).
					 framePointer := newFP.
					 stackPointer := newSP].
				 self pop: 1.
				 self assertValidStackedInstructionPointersIn: stackPage line: #'__LINE__'.
				 self assert: stackPage = stackPages mostRecentlyUsedPage.
				 ^nil].
			 self assertValidStackedInstructionPointers: #'__LINE__'.
			 theFP := self externalEnsureIsBaseFrame: theFP. "May cause a GC!!!!"
			 currentCtx := self frameCallerContext: theFP.
			 "May also reclaim aContextOrNil's page, hence..."
			 (aContextOrNil ~= objectMemory nilObject and: [self isStillMarriedContext: aContextOrNil])
				ifTrue: [contextsFP := self frameOfMarriedContext: aContextOrNil.
						pageToStopOn := stackPages stackPageFor: contextsFP]
				ifFalse: [pageToStopOn := 0]]
		ifFalse:
			[currentCtx := objectMemory fetchPointer: SenderIndex ofObject: thisCtx].

	self assertValidStackedInstructionPointers: #'__LINE__'.
	(self context: thisCtx hasSender: aContextOrNil) ifTrue:
		["Need to walk the stack freeing stack pages and nilling contexts."
		[currentCtx = aContextOrNil
		 or: [currentCtx = objectMemory nilObject]] whileFalse:
			[self assert: (objectMemory isContext: currentCtx).
			 (self isMarriedOrWidowedContext: currentCtx)
				ifTrue:
					[theFP := self frameOfMarriedContext: currentCtx.
					thePage := stackPages stackPageFor: theFP.
					"If externalEnsureIsBaseFrame: above has moved thisContext to its own stack
					 then we will always terminate to a frame on a different page.  But if we are
					 terminating some other context to a context somewhere on the current page
					 we must save the active frames above that context.  Things will look e.g. like this:
		thisCtx			499383332 s MethodContext(ContextPart)>resume:
						499380484 s BlockClosure>ensure:
						499377320 s MethodContext(ContextPart)>handleSignal:
						499373760 s MethodContext(ContextPart)>handleSignal:
						499372772 s MessageNotUnderstood(Exception)>signal
						499369068 s CodeSimulationTests(Object)>doesNotUnderstand: absentMethod
						499368708 s [] in CodeSimulationTests>testDNU
							(sender is 0xbffc2480 I CodeSimulationTests>runSimulated:)
						------------
		framePointer	0xbffc234c M MethodContext(ContextPart)>doPrimitive:method:receiver:args:
						0xbffc2378 M MethodContext(ContextPart)>tryPrimitiveFor:receiver:args:
						0xbffc23ac M MethodContext(ContextPart)>send:to:with:super:
						0xbffc23e4 M MethodContext(ContextPart)>send:super:numArgs:
						0xbffc2418 M MethodContext(InstructionStream)>interpretNextInstructionFor:
						0xbffc2434 M MethodContext(ContextPart)>step
						0xbffc2458 I MethodContext(ContextPart)>runSimulated:contextAtEachStep:
						------------
(499368708's sender)	0xbffc2480 I CodeSimulationTests>runSimulated:
						0xbffc249c M CodeSimulationTests>testDNU
						0xbffc24bc I CodeSimulationTests(TestCase)>performTest
						0xbffc24dc I [] in CodeSimulationTests(TestCase)>runCase
		aContextOrNil	0xbffc24fc M BlockClosure>ensure:
						0xbffc2520 I CodeSimulationTests(TestCase)>runCase
						0xbffc253c M [] in TestResult>runCase:
					When we find this case we move the frames above to a new page by making the
					frame above currentCtx a base frame, i.e. making 0xbffc2458 in the above example
					a base frame.  But in this iteration of the loop we don't move down a frame i.e. currentCtx
					doesn't change on this iteration."
					thePage = stackPage
						ifTrue:
							[frameAbove := self findFrameAbove: theFP inPage: thePage.
							self assert: frameAbove ~= 0.
							frameAbove := self externalEnsureIsBaseFrame: frameAbove. "May cause a GC!!!! May also reclaim aContextOrNil's page, hence..."
							(aContextOrNil ~= objectMemory nilObject and: [self isStillMarriedContext: aContextOrNil])
								ifTrue: [contextsFP := self frameOfMarriedContext: aContextOrNil.
										pageToStopOn := stackPages stackPageFor: contextsFP]
								ifFalse: [pageToStopOn := 0]]
						ifFalse:
							[thePage = pageToStopOn
								ifTrue:
									["We're here.  Cut back the stack to aContextOrNil's frame,
									  push its instructionPointer if it's not already a head frame,
									  and we're done."
									 frameAbove := self findFrameAbove: contextsFP inPage: thePage.
									 frameAbove ~= 0 ifTrue:
										[contextsSP := (self frameCallerSP: frameAbove) - objectMemory wordSize.
										 stackPages longAt: contextsSP put: (self frameCallerSavedIP: frameAbove).
										 self setHeadFP: contextsFP andSP: contextsSP inPage: thePage].
									 currentCtx := aContextOrNil]
								ifFalse:
									["We can free the entire page without further ado."
									 currentCtx := self frameCallerContext: thePage baseFP.
									 "for a short time invariant is violated; assert follows"
									 stackPages freeStackPageNoAssert: thePage]]]
				ifFalse:
					[nextCntx := objectMemory fetchPointer: SenderIndex ofObject: currentCtx.
					 self markContextAsDead: currentCtx.
					 currentCtx := nextCntx]]].
	self assert: stackPages pageListIsWellFormed.
	(self isMarriedOrWidowedContext: thisCtx)
		ifTrue:
			[self assert: (self checkIsStillMarriedContext: thisCtx currentFP: framePointer).
			 self assert: (self isBaseFrame: (self frameOfMarriedContext: thisCtx)).
			 theFP := self frameOfMarriedContext: thisCtx.
			 self frameCallerContext: theFP put: aContextOrNil]
		ifFalse:
			[objectMemory storePointer: SenderIndex ofObject: thisCtx withValue: aContextOrNil].
	self pop: 1.
	self assertValidStackedInstructionPointers: #'__LINE__'.
	self assert: stackPage = stackPages mostRecentlyUsedPage! !

!CoInterpreterPrimitives methodsFor: 'plugin primitives' stamp: 'eem 7/25/2021 20:17'!
primitiveUnloadModule
	"Primitive. Unload the module with the given name."
	"Reloading of the module will happen *later* automatically, when a 
	 function from it is called. This is ensured by invalidating current sessionID."
	| moduleName |
	self methodArgumentCount = 1 ifFalse:[^self primitiveFail].
	moduleName := self stackTop.
	(objectMemory isBytes: moduleName) ifFalse:[^self primitiveFail].
	(self ioUnloadModule: (self oopForPointer: (objectMemory firstIndexableField: moduleName))
		OfLength: (objectMemory byteSizeOf: moduleName)) ifFalse:[^self primitiveFail].
	"N.B. flushExternalPrimitives continues.  Do *not* do anything after flushExternalPrimitives in the CoInterpreter"
	self forceInterruptCheck.
	self flushExternalPrimitives
	"NOTREACHED"! !

!CoInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 3/29/2020 17:25'!
primitiveVoidVMState
	"Void all internal VM state in the stack and machine code zones

	 Override to jump to the interpreter because the machine code zone is now void."
	| activeContext |
	self push: instructionPointer.
	activeContext := self voidVMStateForSnapshotFlushingExternalPrimitivesIf: false.
	self marryContextInNewStackPageAndInitializeInterpreterRegisters: activeContext.
	cogit ceInvokeInterpret
	"NOTREACHED"! !

!CoInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 3/29/2020 17:25'!
primitiveVoidVMStateForMethod
	"The receiver (or first argument) must be a compiledMethod.  The optional (or second) argument must be a
	 boolean. Clear all VM state associated with the method, including any machine code, or machine code pcs
	 in context objects.  If the optional boolean argument is false do not scan the heap looking for contexts."
	| activeContext methodObj scanHeapForContexts hasCogMethod theFrame thePage |
	<var: #theFrame type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	scanHeapForContexts := true. "See comment ''One might think...'' below"
	"In Smalltalk allow both aMethod voidCogVMState and aMethod voidCogVMStateScanningContextsIf: aBoolean"
	argumentCount = 0
		ifTrue:
			[methodObj := self stackTop]
		ifFalse:
			[methodObj := self stackValue: 1.
			 scanHeapForContexts := self booleanValueOf: self stackTop.
			 self failed ifTrue:
				[^self primitiveFailFor: PrimErrBadArgument]].
	NewspeakVM
		ifFalse:
			[argumentCount > 1 ifTrue:
				[^self primitiveFailFor: PrimErrBadNumArgs]]
		 ifTrue: "In the NewspeakVM we allow VMMirror voidStateFor: method scanningIf: aBoolean as well as the Smalltalk forms."
			[argumentCount >= 2 ifTrue:
				[argumentCount > 2 ifTrue:
					[^self primitiveFailFor: PrimErrBadNumArgs].
				 (objectMemory isOopCompiledMethod: methodObj) ifFalse:
					[^self primitiveFailFor: PrimErrBadArgument]]].
	self flushMethodCacheForMethod: methodObj.
	activeContext := self ensureFrameIsMarried: framePointer SP: stackPointer.
	self ensurePushedInstructionPointer.
	self externalWriteBackHeadFramePointers.
	(hasCogMethod := self methodHasCogMethod: methodObj) ifTrue:
		[self divorceMachineCodeFramesWithMethod: methodObj].
	"One might think (as this author did) that the heap scan is unnecessary if the method does not
	 have a cog method.  But it could be the case that the code zone has recently been reclaimed
	 and so not having a cog method is no indication that it didn't have a cog method some time in
	 the recent past, and that there are indeed still contexts with machine code pcs out there.  The
	 only steps that can be avoided are divorcing frames in the stack zone, and scanning to unlink and
	 free if there isn't a cog method, unless we are told otherwise."
	scanHeapForContexts ifTrue:
		[self ensureAllContextsWithMethodHaveBytecodePCs: methodObj].
	hasCogMethod ifTrue:
		[cogit unlinkSendsTo: methodObj andFreeIf: true].

	"If flushing led to divorce continue in the interpreter."
	(self isStillMarriedContext: activeContext) ifFalse:
		[self zeroStackPage. "to avoid assert in marryContextInNewStackPageAndInitializeInterpreterRegisters:"
		 self marryContextInNewStackPageAndInitializeInterpreterRegisters: activeContext.
		 self popStack. "pop bogus machine-code instructionPointer"
		 self assert: (methodObj = self stackTop or: [argumentCount > 0 and: [methodObj = (self stackValue: 1)]]).
		 self pop: argumentCount.
		 cogit ceInvokeInterpret
		 "NOTREACHED"].
	"If not, work out where we are and continue"
	theFrame := self frameOfMarriedContext: activeContext.
	thePage := stackPages stackPageFor: theFrame.
	self assert: thePage headFP = theFrame.
	self setStackPageAndLimit: thePage.
	self setStackPointersFromPage: thePage.
	instructionPointer := self popStack.
	self assert: (methodObj = self stackTop or: [argumentCount > 0 and: [methodObj = (self stackValue: 1)]]).
	self pop: argumentCount! !

!CoInterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
primitiveWait
	| sema excessSignals activeProc inInterpreter |
	sema := self stackTop.  "rcvr"
	excessSignals := self fetchInteger: ExcessSignalsIndex ofObject: sema.
	excessSignals > 0
		ifTrue:
			[self storeInteger: ExcessSignalsIndex
				ofObject: sema
				withValue: excessSignals - 1]
		ifFalse:
			["We're going to switch process, either to an interpreted frame or a machine
			  code frame. To know whether to return or enter machine code we have to
			  know from whence we came.  We could have come from the interpreter,
			  either directly or via a machine code primitive.  We could have come from
			  machine code.  The instructionPointer tells us where from:"
			inInterpreter := instructionPointer >= objectMemory startOfMemory.
			activeProc := self activeProcess.
			self addLastLink: activeProc toList: sema.
			self transferTo: self wakeHighestPriority from: CSWait.
			self forProcessPrimitiveReturnToExecutivePostContextSwitch: inInterpreter]! !

!CoInterpreterPrimitives methodsFor: 'process primitives' stamp: 'eem 3/2/-4654 00:12'!
primitiveYield
"primitively do the equivalent of Process>yield"
	| scheduler activeProc priority processLists processList inInterpreter |
	scheduler := self schedulerPointer.
	activeProc := objectMemory fetchPointer: ActiveProcessIndex ofObject: scheduler.
	priority := self quickFetchInteger: PriorityIndex ofObject: activeProc.
	processLists := objectMemory fetchPointer: ProcessListsIndex ofObject: scheduler.
	processList := objectMemory fetchPointer: priority - 1 ofObject: processLists.

	(self isEmptyList: processList) ifTrue:
		[^nil].
	"We're going to switch process, either to an interpreted frame or a machine
	 code frame. To know whether to return or enter machine code we have to
	 know from whence we came.  We could have come from the interpreter,
	 either directly or via a machine code primitive.  We could have come from
	 machine code.  The instructionPointer tells us where from:"
	inInterpreter := instructionPointer >= objectMemory startOfMemory.
	self addLastLink: activeProc toList: processList.
	self transferTo: self wakeHighestPriority from: CSYield.
	self forProcessPrimitiveReturnToExecutivePostContextSwitch: inInterpreter! !

!CoInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 1/4/2017 08:45'!
unmarkAllFrames
	| thePage theFP methodField flags |
	<var: #thePage type: #'StackPage *'>
	<var: #theFP type: #'char *'>
	<inline: false>
	0 to: numStackPages - 1 do:
		[:i|
		thePage := stackPages stackPageAt: i.
		(stackPages isFree: thePage) ifFalse:
			[theFP := thePage  headFP.
			 [methodField := stackPages longAt: theFP + FoxMethod.
			 methodField asUnsignedInteger < objectMemory startOfMemory
				ifTrue:
					[(methodField bitAnd: 4) ~= 0 ifTrue:
						[self longAt: theFP + FoxMethod put: methodField - 4]]
				ifFalse:
					[flags := stackPages longAt: theFP + FoxIFrameFlags.
					  (flags bitAnd: 2) ~= 0 ifTrue:
						[stackPages longAt: theFP + FoxIFrameFlags put: flags - 2]].
			  (theFP := self frameCallerFP: theFP) ~= 0] whileTrue]]! !

!StackInterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 12/12/2017 09:54'!
cloneContext: aContext 
	| sz cloned spouseFP sp |
	<var: #spouseFP type: #'char *'>
	sz := objectMemory numSlotsOf: aContext.
	cloned := objectMemory eeInstantiateMethodContextSlots: sz.
	cloned ~= 0 ifTrue:
		[0 to: StackPointerIndex do:
			[:i|
			objectMemory
				storePointerUnchecked: i
				ofObject: cloned
				withValue: (self externalInstVar: i ofContext: aContext)].
		MethodIndex to: ReceiverIndex do:
			[:i|
			objectMemory
				storePointerUnchecked: i
				ofObject: cloned
				withValue: (objectMemory fetchPointer: i ofObject: aContext)].
		(self isStillMarriedContext: aContext)
			ifTrue:
				[spouseFP := self frameOfMarriedContext: aContext.
				 sp := (self stackPointerIndexForFrame: spouseFP) - 1.
				 0 to: sp do:
					[:i|
					objectMemory
						storePointerUnchecked: i + CtxtTempFrameStart
						ofObject: cloned
						withValue: (self temporary: i in: spouseFP)]]
			ifFalse:
				[sp := (self fetchStackPointerOf: aContext) - 1.
				 0 to: sp do:
					[:i|
					objectMemory
						storePointerUnchecked: i + CtxtTempFrameStart
						ofObject: cloned
						withValue: (objectMemory fetchPointer: i + CtxtTempFrameStart ofObject: aContext)]]].
	^cloned! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 7/16/2013 15:19'!
field: index ofFrame: theFP
	"Arrange to answer naked frame pointers for unmarried
	 senders to avoid reifying contexts in the search."
	<var: #theFP type: #'char *'>
	<inline: false>
	| callerFP |
	<var: #callerFP type: #'char *'>
	^index caseOf:
		{[SenderIndex] ->	[callerFP := self frameCallerFP: theFP.
							 callerFP = 0
								ifTrue: [self frameCallerContext: theFP]
								ifFalse: [(self frameHasContext: callerFP)
											ifTrue: [self assert: (self checkIsStillMarriedContext: (self frameContext: callerFP) currentFP: nil).
													self frameContext: callerFP]
											ifFalse: [callerFP asInteger]]].
		[StackPointerIndex]			->	[ConstZero].
		[InstructionPointerIndex]	->	[ConstZero].
		[MethodIndex]				->	[self frameMethodObject: theFP].
		[ClosureIndex]				->	[(self frameIsBlockActivation: theFP)
											ifTrue: [self frameStackedReceiver: theFP
														numArgs: (self frameNumArgs: theFP)]
											ifFalse: [objectMemory nilObject]].
		[ReceiverIndex]				->	[self frameReceiver: theFP] }
		otherwise:
			[self assert: (index - CtxtTempFrameStart between: 0 and: (self stackPointerIndexForFrame: theFP)).
			 self temporary: index - CtxtTempFrameStart in: theFP]! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 3/31/2013 15:46'!
fieldOrSenderFP: index ofContext: contextObj
	"Arrange to answer naked frame pointers for unmarried
	 senders to avoid reifying contexts in the search."
	<inline: false>
	| tempIndex spouseFP |
	<var: #spouseFP type: #'char *'>
	tempIndex := index - CtxtTempFrameStart.
	(self isStillMarriedContext: contextObj) ifFalse:
		[^tempIndex >= (self fetchStackPointerOf: contextObj)
			ifTrue: [objectMemory nilObject]
			ifFalse: [self fetchPointer: index ofObject: contextObj]].
	spouseFP := self frameOfMarriedContext: contextObj.
	tempIndex >= (self stackPointerIndexForFrame: spouseFP) ifTrue:
		[^objectMemory nilObject].
	^self field: index ofFrame: spouseFP! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 3/31/2013 15:55'!
fieldsInFrame: theFP
	<var: #theFP type: #'char *'>
	^CtxtTempFrameStart + (self stackPointerIndexForFrame: theFP)! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 3/29/2013 14:50'!
frameIsMarked: theFPInt
	^((stackPages longAt: theFPInt + FoxFrameFlags) bitAnd: 2) ~= 0! !

!StackInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 7/28/2020 12:57'!
getVMTickeeCallCount
	^self cppIf: #VM_TICKER
		ifTrue: [self positiveMachineIntegerFor: self ioVMTickeeCallCount]
		ifFalse: [ConstZero]! !

!StackInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 7/28/2020 10:03'!
getVMTickerCount
	^self cppIf: #VM_TICKER
		ifTrue: [self positiveMachineIntegerFor: self ioVMTickerCount]
		ifFalse: [ConstZero]! !

!StackInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 7/28/2020 10:02'!
getVMTickerStartUSecs
	^self cppIf: #VM_TICKER
		ifTrue: [self positive64BitIntegerFor: self ioVMTickerStartUSecs]
		ifFalse: [ConstZero]! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 9/11/2020 09:40'!
isAppropriateForCopyObject: oop
	(objectMemory isPointersNonImm: oop) ifFalse:
		[^false].
	(objectMemory isContext: oop) ifTrue:
		[^(self isStillMarriedContext: oop) not].
	"Note there is no version in CoInterpreterPrimitives such as
		(objectMemory isCompiledMethod: oop) ifTrue:
			[^(self methodHasCogMethod: oop) not].
	 because isPointersNonImm: excludes compiled methods and the
	 copy loop in primitiveCopyObject cannot handle compiled methods."
	^true! !

!StackInterpreterPrimitives methodsFor: 'plugin primitive support' stamp: 'eem 8/3/2021 00:51'!
linkExternalCall: externalCallLiteral errInto: failPtr
	<returnTypeC: 'void (*linkExternalCallerrInto(sqInt externalCallLiteral, sqInt *failPtr))()'>
	<var: 'failPtr' type: #'sqInt *'>
	"The function has not been loaded yet. Fetch module and function name."
	| addr functionLength functionName index metadata moduleLength moduleName |
	<inline: false>
	<var: #addr declareC: 'void (*addr)()'>
	moduleName := objectMemory fetchPointer: ExternalCallLiteralModuleNameIndex ofObject: externalCallLiteral.
	moduleName = objectMemory nilObject
		ifTrue: [moduleLength := 0]
		ifFalse: [(objectMemory isBytes: moduleName) ifFalse:
					[failPtr ifNotNil: [failPtr at: 0 put: PrimErrBadMethod].
					 ^0].
				moduleLength := objectMemory numBytesOfBytes: moduleName].
	functionName := objectMemory fetchPointer: ExternalCallLiteralFunctionNameIndex ofObject: externalCallLiteral.
	(objectMemory isBytes: functionName) ifFalse:
		[failPtr ifNotNil: [failPtr at: 0 put: PrimErrBadMethod].
		 ^0].
	functionLength := objectMemory numBytesOfBytes: functionName.

	"Spur needs metadata for the primitive, which is stored in the third slot of the literal."
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[addr := self ioLoadExternalFunction: functionName + objectMemory baseHeaderSize
						OfLength: functionLength
						FromModule: moduleName + objectMemory baseHeaderSize
						OfLength: moduleLength
						MetadataInto: (self addressOf: metadata
												 put: [:val| metadata := val]).
			 addr = 0
				ifTrue: [index := -1]
				ifFalse: "add the function to the external primitive table"
					[self assert: (metadata >>> SpurPrimitiveAccessorDepthShift between: -1 and: 5).
					 index := self addToExternalPrimitiveTable: addr.
					 objectMemory
						storePointerUnchecked: ExternalCallLiteralFlagsIndex
						ofObject: externalCallLiteral
						withValue: (objectMemory integerObjectOf: metadata)]]
		ifFalse:
			[addr := self ioLoadExternalFunction: functionName + objectMemory baseHeaderSize
						OfLength: functionLength
						FromModule: moduleName + objectMemory baseHeaderSize
						OfLength: moduleLength.
			 addr = 0
				ifTrue: [index := -1]
				ifFalse: "add the function to the external primitive table"
					[index := self addToExternalPrimitiveTable: addr]].

	"Store the index (or -1 if failure) back in the literal"
	objectMemory
		storePointerUnchecked: ExternalCallLiteralTargetFunctionIndex
		ofObject: externalCallLiteral
		withValue: (objectMemory integerObjectOf: index).

	"If the function has been successfully loaded cache it"
	self rewriteMethodCacheEntryForExternalPrimitiveToFunction: (index >= 0
																		ifTrue: [self cCode: [addr] inSmalltalk: [1000 + index]]
																		ifFalse: [0]).
	^addr! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 3/29/2013 14:53'!
markFrame: theFPInt
	stackPages
		longAt: theFPInt + FoxFrameFlags
		put: ((stackPages longAt: theFPInt + FoxFrameFlags) bitOr: 2)! !

!StackInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 7/29/2021 11:23'!
minimumUnusedHeadroom
	<inline: #always>
	^-1! !

!StackInterpreterPrimitives methodsFor: 'primitive support' stamp: 'eem 6/15/2017 14:30'!
noInlineLoadFloatOrIntFrom: floatOrInt
	<inline: #never>
	^objectMemory loadFloatOrIntFrom: floatOrInt! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'nice 4/23/2016 23:55'!
pathTo: goal using: stack followWeak: followWeak
	"Trace objects and frames from the root, marking visited objects, pushing the current path on stack, until goal is found.
	 If found, unmark, leaving path in stack, and answer 0.  Otherwise answer an error:
		PrimErrBadArgument if stack is not an Array
		PrimErrBadIndex if search overflows stack
		PrimErrNotFound if goal cannot be found"
	<var: #index type: #sqInt> "beware, must be signed"
	| current index next stackSize stackp freeStartAtStart |
	(objectMemory isArray: stack) ifFalse:
		[^PrimErrBadArgument].
	self assert: objectMemory allObjectsUnmarked.
	freeStartAtStart := objectMemory freeStart. "check no allocations during search"
	objectMemory beRootIfOld: stack. "so no store checks are necessary on stack"
	stackSize := objectMemory lengthOf: stack.
	objectMemory mark: stack.
	"no need. the current context is not reachable from the active process (suspendedContext is nil)"
	"objectMemory mark: self activeProcess."
	current := objectMemory specialObjectsOop.
	objectMemory mark: current.
	index := objectMemory lengthOf: current.
	stackp := 0.
	[[(index := index - 1) >= -1] whileTrue:
		[(stackPages couldBeFramePointer: current)
			ifTrue:
				[next := index >= 0
							ifTrue: [self field: index ofFrame: (self cCoerceSimple: current to: #'char *')]
							ifFalse: [objectMemory nilObject]]
			ifFalse:
				[index >= 0
					ifTrue:
						[next := (objectMemory isContextNonImm: current)
									ifTrue: [self fieldOrSenderFP: index ofContext: current]
									ifFalse: [objectMemory fetchPointer: index ofObject: current]]
					ifFalse:
						[next := objectMemory fetchClassOfNonImm: current]].
		 (stackPages couldBeFramePointer: next)
			ifTrue: [self assert: (self isFrame: (self cCoerceSimple: next to: #'char *')
									onPage: (stackPages stackPageFor: (self cCoerceSimple: next to: #'char *')))]
			ifFalse: [self assert: (self checkOkayOop: next)].
		 next = goal ifTrue:
			[self assert: freeStartAtStart = objectMemory freeStart.
			 self unmarkAfterPathTo.
			 objectMemory storePointer: stackp ofObject: stack withValue: current.
			 self pruneStack: stack stackp: stackp.
			 ^0].
		 ((objectMemory isNonIntegerObject: next)
		  and: [(stackPages couldBeFramePointer: next)
				ifTrue: [(self frameIsMarked: next) not]
				ifFalse:
					[(objectMemory isMarked: next) not
					  and: [((objectMemory isPointers: next) or: [objectMemory isCompiledMethod: next])
					  and: [followWeak or: [(objectMemory isWeakNonImm: next) not]]]]])
			ifTrue:
				[stackp + 2 > stackSize ifTrue:
					[self assert: freeStartAtStart = objectMemory freeStart.
					 self unmarkAfterPathTo.
					 objectMemory nilFieldsOf: stack.
					 ^PrimErrBadIndex]. "PrimErrNoMemory ?"
				 objectMemory
					storePointerUnchecked: stackp ofObject: stack withValue: current;
					storePointerUnchecked: stackp + 1 ofObject: stack withValue: (objectMemory integerObjectOf: index).
				 stackp := stackp + 2.
				 (stackPages couldBeFramePointer: (self cCoerceSimple: next to: #'char *'))
					ifTrue:
						[self markFrame: next.
						index := self fieldsInFrame: (self cCoerceSimple: next to: #'char *')]
					ifFalse:
						[objectMemory mark: next.
						 (objectMemory isCompiledMethod: next)
							ifTrue: [index := (objectMemory literalCountOf: next) + LiteralStart]
							ifFalse: [index := objectMemory lengthOf: next]].
				 current := next]].
		 current = objectMemory specialObjectsOop ifTrue:
			[self assert: freeStartAtStart = objectMemory freeStart.
			 self unmarkAfterPathTo.
			 objectMemory nilFieldsOf: stack.
			^PrimErrNotFound].
		 index := objectMemory integerValueOf: (objectMemory fetchPointer: stackp - 1 ofObject: stack).
		 current := objectMemory fetchPointer: stackp - 2 ofObject: stack.
		 stackp := stackp - 2] repeat! !

!StackInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 7/29/2021 11:23'!
primitiveAllVMParameters: paramsArraySize
	"See primitiveVMParameter method comment.
	 N.B. use storePointerUnchecked:and finally beRootIfOld: to save size and time"

	| result |
	result := objectMemory instantiateClass: (objectMemory splObj: ClassArray) indexableSize: paramsArraySize.
	objectMemory storePointerUnchecked: 0	ofObject: result withValue: (self positiveMachineIntegerFor: objectMemory oldSpaceSize).
	objectMemory storePointerUnchecked: 1	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory newSpaceSize).
	objectMemory storePointerUnchecked: 2	ofObject: result withValue: (self positiveMachineIntegerFor: objectMemory totalMemorySize).
	"objectMemory storePointerUnchecked: 3	ofObject: result withValue: objectMemory nilObject was allocationCount".
	"objectMemory storePointerUnchecked: 4	ofObject: result withValue: objectMemory nilObject allocationsBetweenGCs".
	objectMemory storePointerUnchecked: 5	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory tenuringThreshold).
	objectMemory storePointerUnchecked: 6	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statFullGCs).
	objectMemory storePointerUnchecked: 7	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statFullGCUsecs + 500 // 1000).
	objectMemory
		storePointerUnchecked: 8
		ofObject: result
		withValue: (objectMemory integerObjectOf: (objectMemory hasSpurMemoryManagerAPI
														ifTrue: [objectMemory statScavenges]
														ifFalse: [objectMemory statIncrGCs])).
	objectMemory
		storePointerUnchecked: 9
		ofObject: result
		withValue: (objectMemory integerObjectOf: (objectMemory hasSpurMemoryManagerAPI
														ifTrue: [objectMemory statScavengeGCUsecs]
														ifFalse: [objectMemory statIncrGCUsecs]) + 500 // 1000).
	objectMemory storePointerUnchecked: 10 ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statTenures).
	objectMemory storePointerUnchecked: 11 ofObject: result withValue: (objectMemory integerObjectOf: eventTraceMask).
	objectMemory storePointerUnchecked: 12 ofObject: result withValue: self getVMTickerStartUSecs.
	objectMemory storePointerUnchecked: 13 ofObject: result withValue: self getVMTickerCount.
	objectMemory storePointerUnchecked: 14 ofObject: result withValue: self getVMTickeeCallCount.
	objectMemory storePointerUnchecked: 15 ofObject: result withValue: (objectMemory positive64BitIntegerFor: statIdleUsecs).
	objectMemory storePointerUnchecked: 16 ofObject: result withValue: ((SistaVM and: [self isCog])
																			ifTrue: [objectMemory floatObjectOf: self getCogCodeZoneThreshold]
																			ifFalse: [ConstZero]).
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[objectMemory
				storePointerUnchecked: 17	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statCompactionUsecs + 500 // 1000);
				storePointerUnchecked: 18	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory scavengeThresholdAsExtent)]
		ifFalse:
			[objectMemory
				storePointerUnchecked: 17	ofObject: result withValue: ConstZero;
				storePointerUnchecked: 18	ofObject: result withValue: ConstZero].
	objectMemory storePointerUnchecked: 19	ofObject: result withValue: (objectMemory positive64BitIntegerFor: self ioUTCStartMicroseconds).
	objectMemory storePointerUnchecked: 20	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory rootTableCount).
	objectMemory storePointerUnchecked: 21	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statRootTableOverflows).
	objectMemory storePointerUnchecked: 22	ofObject: result withValue: (objectMemory integerObjectOf: extraVMMemory).
	objectMemory storePointerUnchecked: 23	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory shrinkThreshold).
	objectMemory storePointerUnchecked: 24	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory growHeadroom).
	objectMemory storePointerUnchecked: 25	ofObject: result withValue: (objectMemory integerObjectOf: self ioHeartbeatMilliseconds).
	objectMemory storePointerUnchecked: 26	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statMarkCount).
	objectMemory storePointerUnchecked: 27	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statSweepCount).
	objectMemory storePointerUnchecked: 28	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statMkFwdCount).
	objectMemory storePointerUnchecked: 29	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statCompMoveCount).
	objectMemory storePointerUnchecked: 30	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statGrowMemory).
	objectMemory storePointerUnchecked: 31	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statShrinkMemory).
	objectMemory storePointerUnchecked: 32	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statRootTableCount).
	objectMemory hasSpurMemoryManagerAPI ifTrue: "was statAllocationCount"
		[objectMemory storePointerUnchecked: 33	ofObject: result withValue: (objectMemory positive64BitIntegerFor: objectMemory currentAllocatedBytes)].
	objectMemory storePointerUnchecked: 34	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statSurvivorCount).
	objectMemory storePointerUnchecked: 35	ofObject: result withValue: (objectMemory integerObjectOf: (self microsecondsToMilliseconds: objectMemory statGCEndUsecs)).
	objectMemory storePointerUnchecked: 36	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statSpecialMarkCount).
	objectMemory storePointerUnchecked: 37	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statIGCDeltaUsecs + 500 // 1000).
	objectMemory storePointerUnchecked: 38	ofObject: result withValue: (objectMemory integerObjectOf: statPendingFinalizationSignals).
	objectMemory storePointerUnchecked: 39	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory wordSize).
	objectMemory storePointerUnchecked: 40	ofObject: result withValue: (objectMemory integerObjectOf: self imageFormatVersion).
	objectMemory storePointerUnchecked: 41	ofObject: result withValue: (objectMemory integerObjectOf: numStackPages).
	objectMemory storePointerUnchecked: 42	ofObject: result withValue: (objectMemory integerObjectOf: desiredNumStackPages).
	objectMemory storePointerUnchecked: 43	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory edenBytes).
	objectMemory storePointerUnchecked: 44	ofObject: result withValue: (objectMemory integerObjectOf: desiredEdenBytes).
	objectMemory storePointerUnchecked: 45	ofObject: result withValue: self getCogCodeSize.
	objectMemory storePointerUnchecked: 46	ofObject: result withValue: self getDesiredCogCodeSize.
	objectMemory storePointerUnchecked: 47	ofObject: result withValue: self getImageHeaderFlagsParameter.
	objectMemory storePointerUnchecked: 48	ofObject: result withValue: (objectMemory integerObjectOf: self ioGetMaxExtSemTableSize).
	"50 & 51 (49 & 50) reserved for parameters that persist in the image"
	objectMemory storePointerUnchecked: 51	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory rootTableCapacity).
	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[objectMemory
			storePointerUnchecked: 52 ofObject: result withValue: (objectMemory integerObjectOf: objectMemory numSegments);
			storePointerUnchecked: 53 ofObject: result withValue: (objectMemory integerObjectOf: objectMemory freeSize);
			storePointerUnchecked: 54 ofObject: result withValue: (objectMemory floatObjectOf: objectMemory getHeapGrowthToSizeGCRatio)].
	objectMemory storePointerUnchecked: 55 ofObject: result withValue: (self positive64BitIntegerFor: statProcessSwitch).
	objectMemory storePointerUnchecked: 56 ofObject: result withValue: (self positive64BitIntegerFor: statIOProcessEvents).
	objectMemory storePointerUnchecked: 57 ofObject: result withValue: (self positive64BitIntegerFor: statForceInterruptCheck).
	objectMemory storePointerUnchecked: 58 ofObject: result withValue: (self positive64BitIntegerFor: statCheckForEvents).
	objectMemory storePointerUnchecked: 59 ofObject: result withValue: (self positive64BitIntegerFor: statStackOverflow).
	objectMemory storePointerUnchecked: 60 ofObject: result withValue: (self positive64BitIntegerFor: statStackPageDivorce).
	objectMemory storePointerUnchecked: 61 ofObject: result withValue: self getCodeCompactionCount.
	objectMemory storePointerUnchecked: 62 ofObject: result withValue: self getCodeCompactionMSecs.
	objectMemory storePointerUnchecked: 63 ofObject: result withValue: self getCogMethodCount.
	objectMemory storePointerUnchecked: 64 ofObject: result withValue: self getCogVMFeatureFlags.
	objectMemory storePointerUnchecked: 65 ofObject: result withValue: (objectMemory integerObjectOf: self stackPageByteSize).
	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[objectMemory
			storePointerUnchecked: 66 ofObject: result withValue: (self positiveMachineIntegerFor: objectMemory maxOldSpaceSize)].
	objectMemory storePointerUnchecked: 67 ofObject: result withValue: (objectMemory floatObjectOf: stackPages statAverageLivePagesWhenMapping).
	objectMemory storePointerUnchecked: 68 ofObject: result withValue: (objectMemory integerObjectOf: stackPages statMaxPageCountWhenMapping).
	objectMemory storePointerUnchecked: 69 ofObject: result withValue: (objectMemory integerObjectOf: self vmProxyMajorVersion).
	objectMemory storePointerUnchecked: 70 ofObject: result withValue: (objectMemory integerObjectOf: self vmProxyMinorVersion).	
	objectMemory storePointerUnchecked: 71 ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statMarkUsecs + 500 // 1000).
	objectMemory storePointerUnchecked: 72 ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statSweepUsecs + 500 // 1000).
	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[objectMemory
			storePointerUnchecked: 73	ofObject: result withValue: (objectMemory integerObjectOf: objectMemory statMaxAllocSegmentTime + 500 // 1000)].
	objectMemory storePointerUnchecked: 74 ofObject: result withValue: (objectMemory booleanObjectOf: self primitiveDoMixedArithmetic).
	objectMemory storePointerUnchecked: 75 ofObject: result withValue: (objectMemory integerObjectOf: self minimumUnusedHeadroom).

	objectMemory beRootIfOld: result.
	self methodReturnValue: result! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 01/07/2021 10:40'!
primitiveClone
	"Return a shallow copy of the receiver.
	 Special-case non-single contexts (because of context-to-stack mapping).
	 Can't fail for contexts cuz of image context instantiation code (sigh)."

	| rcvr newCopy |
	rcvr := self stackTop.
	(objectMemory isImmediate: rcvr)
		ifTrue:
			[newCopy := rcvr]
		ifFalse:
			[(objectMemory isContextNonImm: rcvr)
				ifTrue:
					[newCopy := self cloneContext: rcvr]
				ifFalse:
					[(argumentCount = 0
					  or: [(objectMemory isForwarded: rcvr) not])
						ifTrue: [newCopy := objectMemory cloneObject: rcvr]
						ifFalse: [newCopy := 0]].
			newCopy = 0 ifTrue:
				[^self primitiveFailFor: PrimErrNoMemory]].
	self pop: argumentCount + 1 thenPush: newCopy! !

!StackInterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
primitiveClosureCopyWithCopiedValues
	| newClosure numArgs |
	numArgs := self stackIntegerValue: 1.
	self successful ifFalse:
		[^self primitiveFail].

	newClosure := self
					closureIn: (self stackValue: 2)
					numArgs: numArgs
							"greater by 1 due to preIncrement of localIP"
					instructionPointer: instructionPointer + 2 - (method+objectMemory baseHeaderSize)
					copiedValues: self stackTop.
	self pop: 3 thenPush: newClosure! !

!StackInterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 9/13/2014 16:38'!
primitiveContextAt
	"Special version of primitiveAt for accessing contexts.
	 Written to be varargs for use from mirror primitives."
	| index value aContext spouseFP hdr fmt totalLength fixedFields stSize |
	<inline: false>
	<var: #spouseFP type: #'char *'>
	index := self stackTop.
	(objectMemory isIntegerObject: index) ifFalse:
		[self primitiveFailFor: PrimErrBadArgument.
		 ^self].
	index := objectMemory integerValueOf: index.
	aContext := self stackValue: 1.
	"Duplicating much of stObject:at:put: here allows stObject:at:put: to omit tests for contexts."
	hdr := objectMemory baseHeader: aContext.
	(objectMemory isContextHeader: hdr) ifFalse: "might be an instance of a subclass"
		[value := self stObject: aContext at: index.
		 self successful ifTrue:
			[self pop: argumentCount + 1 thenPush: value].
		 ^self].
	self externalWriteBackHeadFramePointers.
	(self isStillMarriedContext: aContext) ifFalse:
		[fmt := objectMemory formatOfHeader: hdr.
		 totalLength := objectMemory lengthOf: aContext baseHeader: hdr format: fmt.
		 fixedFields := objectMemory fixedFieldsOf: aContext format: fmt length: totalLength.
		 stSize := self fetchStackPointerOf: aContext.
		 (index between: 1 and: stSize) ifFalse:
			[self primitiveFailFor: PrimErrBadIndex.
			 ^self].			
		value := self subscript: aContext with: (index + fixedFields) format: fmt.
		self pop: argumentCount + 1 thenPush: value.
		^self].
	spouseFP := self frameOfMarriedContext: aContext.
	(index between: 1 and: (self stackPointerIndexForFrame: spouseFP)) ifFalse:
		[self primitiveFailFor: PrimErrBadIndex.
		 ^self].
	value := self temporary: index - 1 in: spouseFP.
	self pop: argumentCount + 1 thenPush: value! !

!StackInterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 9/13/2014 16:38'!
primitiveContextAtPut
	"Special version of primitiveAtPut for accessing contexts.
	 Written to be varargs for use from mirror primitives."
	| index value aContext spouseFP hdr fmt totalLength fixedFields stSize |
	<inline: false>
	<var: #spouseFP type: #'char *'>
	value := self stackTop.
	index := self stackValue: 1.
	aContext := self stackValue: 2.
	(objectMemory isIntegerObject: index) ifFalse:
		[self primitiveFailFor: PrimErrBadArgument.
		 ^self].
	"Duplicating much of stObject:at:put: here allows stObject:at:put: to omit tests for contexts."
	hdr := objectMemory baseHeader: aContext.
	index := objectMemory integerValueOf: index.
	(objectMemory isContextHeader: hdr) ifFalse: "might be an instance of a subclass"
		[self stObject: aContext at: index put: value.
		 self successful ifTrue:
			[self pop: argumentCount + 1 thenPush: value].
		 ^self].
	self externalWriteBackHeadFramePointers.
	(self isStillMarriedContext: aContext) ifFalse:
		[fmt := objectMemory formatOfHeader: hdr.
		 totalLength := objectMemory lengthOf: aContext baseHeader: hdr format: fmt.
		 fixedFields := objectMemory fixedFieldsOf: aContext format: fmt length: totalLength.
		 stSize := self fetchStackPointerOf: aContext.
		 (index between: 1 and: stSize) ifFalse:
			[self primitiveFailFor: PrimErrBadIndex.
			 ^self].
		self subscript: aContext with: (index + fixedFields) storing: value format: fmt.
		self pop: argumentCount + 1 thenPush: value.
		^self].
	spouseFP := self frameOfMarriedContext: aContext.
	(index between: 1 and: (self stackPointerIndexForFrame: spouseFP)) ifFalse:
		[self primitiveFailFor: PrimErrBadIndex.
		 ^self].
	self temporary: index - 1 in: spouseFP put: value.
	self pop: argumentCount + 1 thenPush: value! !

!StackInterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem 5/4/2021 14:51'!
primitiveContextSize
	"Special version of primitiveSize for accessing contexts.
	 Written to be varargs for use from mirror primitives."
	| rcvr sz hdr fmt totalLength fixedFields |
	<inline: false>
	rcvr := self stackTop.
	hdr := objectMemory baseHeader: rcvr.
	fmt := objectMemory formatOfHeader: hdr.
	totalLength := objectMemory lengthOf: rcvr baseHeader: hdr format: fmt.
	fixedFields := objectMemory fixedFieldsOf: rcvr format: fmt length: totalLength.
	(objectMemory isContextHeader: hdr)
		ifTrue:
			[self externalWriteBackHeadFramePointers.
			sz := self stackPointerForMaybeMarriedContext: rcvr]
		ifFalse: [sz := totalLength - fixedFields].
	self methodReturnInteger: sz! !

!StackInterpreterPrimitives methodsFor: 'indexing primitives' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
primitiveContextXray
	"Lift the veil from a context and answer an integer describing its interior state.
	 Used for e.g. VM tests so they can verify they're testing what they think they're testing.
	 0 implies a vanilla heap context.
	 Bit 0 = is or was married to a frame
	 Bit 1 = is still married to a frame
	 Bit 2 = frame is executing machine code
	 Bit 3 = has machine code pc (as opposed to nil or a bytecode pc)
	 Bit 4 = method is currently compiled to machine code"
	| context flags |
	context := self stackTop.
	(self isMarriedOrWidowedContext: context)
		ifTrue:
			[(self checkIsStillMarriedContext: context currentFP: framePointer)
				ifTrue: [flags := 3]
				ifFalse: [flags := 1]]
		ifFalse:
			[flags := 0].
	self pop: 1 thenPush: (objectMemory integerObjectOf: flags)! !

!StackInterpreterPrimitives methodsFor: 'I/O primitives' stamp: 'eem 1/15/2018 10:15'!
primitiveDeferDisplayUpdates
	"Set or clear the flag that controls whether modifications of 
	 the Display object are propagated to the underlying 
	 platform's screen."
	| flag |
	flag := self stackTop.
	flag = objectMemory trueObject
		ifTrue: [deferDisplayUpdates := true]
		ifFalse:
			[flag = objectMemory falseObject
				ifTrue: [deferDisplayUpdates := false]
				ifFalse: [^self primitiveFail]].
	self cCode: [] inSmalltalk: [self fullDisplayUpdate].
	self pop: 1! !

!StackInterpreterPrimitives methodsFor: 'plugin primitives' stamp: 'eem 8/2/2021 21:55'!
primitiveDoNamedPrimitiveWithArgs
	"Simulate an primitiveExternalCall invocation (e.g. for the Debugger).  Do not cache anything.
	 e.g. ContextPart>>tryNamedPrimitiveIn: aCompiledMethod for: aReceiver withArgs: arguments"
	| argumentArray arraySize methodArg methodHeader
	  moduleName functionName moduleLength functionLength
	  spec addr primRcvr isArray |
	<var: #addr declareC: 'void (*addr)()'>
	objectMemory hasSpurMemoryManagerAPI ifTrue: "See checkForAndFollowForwardedPrimitiveState"
		[metaAccessorDepth := -2].
	argumentArray := self stackTop.
	methodArg := self stackValue: 2.
	((objectMemory isArray: argumentArray)
	 and: [objectMemory isOopCompiledMethod: methodArg]) ifFalse:
		[^self primitiveFailFor: -2]. "invalid args"
	arraySize := objectMemory numSlotsOf: argumentArray.
	(self roomToPushNArgs: arraySize) ifFalse:
		[^self primitiveFailFor: -2]. "invalid args"

	methodHeader := objectMemory methodHeaderOf: methodArg.
	(objectMemory literalCountOfMethodHeader: methodHeader) > 2 ifFalse:
		[^self primitiveFailFor: -3]. "invalid methodArg state"
	spec := objectMemory fetchPointer: 1 "first literal" ofObject: methodArg.
	isArray := self isInstanceOfClassArray: spec.
	(isArray
	and: [(objectMemory numSlotsOf: spec) = 4
	and: [(self primitiveIndexOfMethod: methodArg header: methodHeader) = PrimNumberExternalCall]]) ifFalse:
		[^self primitiveFailFor: -3]. "invalid methodArg state"

	(self argumentCountOfMethodHeader: methodHeader) = arraySize ifFalse:
		[^self primitiveFailFor: -2]. "invalid args (Array args wrong size)"

	"The function has not been loaded yet. Fetch module and function name."
	moduleName := objectMemory fetchPointer: 0 ofObject: spec.
	moduleName = objectMemory nilObject
		ifTrue: [moduleLength := 0]
		ifFalse: [self success: (objectMemory isBytes: moduleName).
				moduleLength := objectMemory numBytesOfBytes: moduleName].
	functionName := objectMemory fetchPointer: 1 ofObject: spec.
	self success: (objectMemory isBytes: functionName).
	functionLength := objectMemory numBytesOfBytes: functionName.
	self successful ifFalse: [^self primitiveFailFor: -3]. "invalid methodArg state"

	"Spur needs to know the primitive's accessorDepth."
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[addr := self ioLoadExternalFunction: functionName + objectMemory baseHeaderSize
						OfLength: functionLength
						FromModule: moduleName + objectMemory baseHeaderSize
						OfLength: moduleLength
						MetadataInto: (self addressOf: metaAccessorDepth
												 put: [:val| metaAccessorDepth := val]).
			"N.B. the accessor depth is the second byte of the primitive's metadata;
			 the first byte is various flags (currently l.s.b. = use fast C linkage)."
			 metaAccessorDepth := addr = 0 ifTrue: [-2] ifFalse: [metaAccessorDepth bitShift: -8]]
		ifFalse:
			[addr := self ioLoadExternalFunction: functionName + objectMemory baseHeaderSize
						OfLength: functionLength
						FromModule: moduleName + objectMemory baseHeaderSize
						OfLength: moduleLength].
	addr = 0 ifTrue:
		[^self primitiveFailFor: -1]. "could not find function; answer generic failure (see below)"

	"Cannot fail this primitive from now on.  Can only fail the external primitive."
	tempOop := objectMemory
						eeInstantiateClassIndex: ClassArrayCompactIndex
						format: objectMemory arrayFormat
						numSlots: 4.
	objectMemory
		storePointerUnchecked: 0 ofObject: tempOop withValue: (argumentArray := self popStack);
		storePointerUnchecked: 1 ofObject: tempOop withValue: (primRcvr := self popStack);
		storePointerUnchecked: 2 ofObject: tempOop withValue: self popStack; "the method"
		storePointerUnchecked: 3 ofObject: tempOop withValue: self popStack. "the context receiver"
	self push: primRcvr. "replace context receiver with actual receiver"
	argumentCount := arraySize.
	1 to: arraySize do:
		[:index| self push: (objectMemory fetchPointer: index - 1 ofObject: argumentArray)].
	self callExternalPrimitive: addr.
	self successful ifFalse: "If primitive failed, then restore state for failure code"
		[self pop: arraySize + 1.
		 self push: (objectMemory fetchPointer: 3 ofObject: tempOop).
		 self push: (objectMemory fetchPointer: 2 ofObject: tempOop).
		 self push: (objectMemory fetchPointer: 1 ofObject: tempOop).
		 self push: (objectMemory fetchPointer: 0 ofObject: tempOop).
		 argumentCount := 3.
		 "Must reset primitiveFunctionPointer for checkForAndFollowForwardedPrimitiveState"
		 objectMemory hasSpurMemoryManagerAPI ifTrue:
			[primitiveFunctionPointer := #primitiveDoNamedPrimitiveWithArgs].
		 "Hack.  A nil prim error code (primErrorCode = 1) is interpreted by the image
		  as meaning this primitive is not implemented.  So to pass back nil as an error
		  code we use -1 to indicate generic failure."
		 primFailCode = 1 ifTrue:
			[primFailCode := -1]]! !

!StackInterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 12/7/2017 11:27'!
primitiveDoPrimitiveWithArgs
	"Implement either ProtoObject>>tryPrimitive: primIndex withArgs: argArray
	 or Context>>receiver: anObject tryPrimitive: primIndex withArgs: argArray.
	 If this primitive fails, arrange that its error code is a negative integer, to
	 distinguish between this failing and the primitive it invokes failing."
	| argumentArray arraySize index primIdx savedNumArgs rcvr |
	objectMemory hasSpurMemoryManagerAPI ifTrue: "See checkForAndFollowForwardedPrimitiveState"
		[metaAccessorDepth := -2].
	(argumentCount between: 2 and: 3) ifFalse:
		[^self primitiveFailFor: PrimErrUnsupported negated].
	argumentArray := self stackTop.
	primIdx := self stackValue: 1.
	((objectMemory isArray: argumentArray)
	 and: [objectMemory isIntegerObject: primIdx]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument negated].
	arraySize := objectMemory numSlotsOf: argumentArray.
	(self roomToPushNArgs: arraySize) ifFalse:
		[^self primitiveFailFor: PrimErrLimitExceeded negated].

	primIdx := objectMemory integerValueOf: primIdx.
	primitiveFunctionPointer := self functionPointerFor: primIdx inClass: nil.
	primitiveFunctionPointer = 0 ifTrue:
		[primitiveFunctionPointer := #primitiveDoPrimitiveWithArgs.
		 ^self primitiveFailFor: PrimErrBadIndex negated].

	"Pop primIndex and argArray, then push args in place..."
	(savedNumArgs := argumentCount) = 3
		ifTrue: "...and receiver if the three arg form"
			[tempOop2 := self stackValue: 4. "actual receiver"
			 rcvr := self stackValue: 3. "receiver for primitive"
			 (objectMemory isOopForwarded: rcvr) ifTrue:
				[rcvr := objectMemory followForwarded: rcvr].
			 self pop: 4; push: rcvr] "use first arg as receiver"
		ifFalse:
			[self pop: 2].
	argumentCount := arraySize.
	index := 1.
	[index <= arraySize] whileTrue:
		[self push: (objectMemory fetchPointer: index - 1 ofObject: argumentArray).
		 index := index + 1].

	self isPrimitiveFunctionPointerAnIndex ifTrue:
		[self externalQuickPrimitiveResponse.
		 tempOop2 := 0.
		^nil].
	"We use tempOop instead of pushRemappableOop:/popRemappableOop here because in
	 the Cogit primitiveEnterCriticalSection, primitiveSignal, primitiveResume et al longjmp back
	 to either the interpreter or machine code, depending on the process activated.  So if we're
	 executing one of these primitives, control won't actually return here and the matching
	 popRemappableOop: wouldn't occur, potentially overflowing the remap buffer.
	 Note that while recursion could occur (nil tryPrimitive: 118 withArgs: #(118 #(110 #())))
	 it counts as shooting oneself in the foot."
	tempOop := argumentArray. "prim might alloc/gc"

	"Run the primitive (sets primFailCode)"
	objectMemory hasSpurMemoryManagerAPI ifTrue: "See checkForAndFollowForwardedPrimitiveState"
		[metaAccessorDepth := primitiveAccessorDepthTable at: primIdx].
	self slowPrimitiveResponse.

	self successful ifFalse: "If primitive failed, then restore state for failure code"
		[self pop: arraySize.
		 savedNumArgs = 3 ifTrue:
			[rcvr := self stackTop.
			 self stackTopPut: tempOop2.
			 self push: rcvr].
		 self pushInteger: primIdx.
		 self push: tempOop.
		 primitiveFunctionPointer := #primitiveDoPrimitiveWithArgs.
		 argumentCount := savedNumArgs].
	tempOop := tempOop2 := 0! !

!StackInterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 9/12/2013 09:47'!
primitiveExecuteMethod
	"receiver, args, then method are on top of stack. Execute method against receiver and args.
	 Set primitiveFunctionPointer because no cache lookup has been done for the method, and
	 hence primitiveFunctionPointer is stale."
	| methodArgument primitiveIndex |
	methodArgument := self stackTop.
	(objectMemory isOopCompiledMethod: methodArgument) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	argumentCount - 1 = (self argumentCountOf: methodArgument) ifFalse:
		[^self primitiveFailFor: PrimErrBadNumArgs].
	newMethod := self popStack.
	primitiveIndex := self primitiveIndexOf: newMethod.
	primitiveFunctionPointer := self functionPointerFor: primitiveIndex inClass: nil.
	argumentCount := argumentCount - 1.
	self executeNewMethod.
	"Recursive xeq affects primErrorCode"
	self initPrimCall! !

!StackInterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 6/20/2015 11:32'!
primitiveExecuteMethodArgsArray
	"receiver, argsArray, then method are on top of stack.  Execute method against
	 receiver and args.  Allow for up to two extra arguments (e.g. for mirror primitives).
	 Set primitiveFunctionPointer because no cache lookup has been done for the
	 method, and hence primitiveFunctionPointer is stale."
	| methodArgument argCnt rcvr argumentArray primitiveIndex |
	methodArgument := self stackTop.
	argumentArray := self stackValue: 1.
	((objectMemory isOopCompiledMethod: methodArgument)
	 and: [objectMemory isArray: argumentArray]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	argCnt := self argumentCountOf: methodArgument.
	argCnt = (objectMemory numSlotsOf: argumentArray) ifFalse:
		[^self primitiveFailFor: PrimErrBadNumArgs].
	argumentCount > 2 ifTrue: "CompiledMethod class>>receiver:withArguments:executeMethod:
								SqueakObjectPrimitives class>>receiver:withArguments:apply:
								VMMirror>>ifFail:object:with:executeMethod: et al"
		[rcvr := self stackValue: 2.
		 (argumentCount > 4
		  or: [objectMemory isOopForwarded: rcvr]) ifTrue:
			[^self primitiveFailFor: PrimErrUnsupported].
		self stackValue: argumentCount put: rcvr]. "replace actual receiver with desired receiver"
	"and push the actual arguments"
	self pop: argumentCount.
	0 to: argCnt - 1 do:
		[:i|
		self push: (objectMemory fetchPointer: i ofObject: argumentArray)].
	newMethod := methodArgument.
	primitiveIndex := self primitiveIndexOf: newMethod.
	primitiveFunctionPointer := self functionPointerFor: primitiveIndex inClass: nil.
	argumentCount := argCnt.
	self executeNewMethod.
	"Recursive xeq affects primErrorCode"
	self initPrimCall! !

!StackInterpreterPrimitives methodsFor: 'plugin primitives' stamp: 'eem 8/2/2021 21:50'!
primitiveExternalCall
	"Call an external primitive. External primitive methods first literals are an array of
		* The module name (String | Symbol)
		* The function name (String | Symbol)
		* The session ID (SmallInteger) [OBSOLETE], or in Spur, the metadata (accessorDepth and flags; Integer))
		* The function index (Integer) in the externalPrimitiveTable
	For fast interpreter dispatch in subsequent invocations the primitiveFunctionPointer
	in the method cache is rewritten, either to the function itself, or to zero if the external
	function is not found.   This allows for fast responses as long as the method stays in
	the cache. The cache rewrite relies on lastMethodCacheProbeWrite which is set in
	addNewMethodToCache:.
	Now that the VM flushes function addresses from its tables, the session ID is obsolete,
	but it is kept for backward compatibility. Also, a failed lookup is reported specially. If a
	method has been  looked up and not been found, the function address is stored as -1
	(i.e., the SmallInteger -1 to distinguish from 16rFFFFFFFF which may be returned from
	lookup), and the primitive fails with PrimErrNotFound."
	| lit addr index |
	
	"Check for it being a method for primitiveDoPrimitiveWithArgs.
	 Fetch the first literal of the method; check its an Array of length 4.
	 Look at the function index in case it has been loaded before"
	((objectMemory isOopCompiledMethod: newMethod)
	 and: [(objectMemory literalCountOf: newMethod) > 0
	 and: [lit := self literal: 0 ofMethod: newMethod.
		(objectMemory isArray: lit)
	 and: [(objectMemory numSlotsOf: lit) = 4
	 and: [index := objectMemory fetchPointer: ExternalCallLiteralTargetFunctionIndex ofObject: lit.
		objectMemory isIntegerObject: index]]]]) ifFalse:
		[^self primitiveFailFor: PrimErrBadMethod].

	index := objectMemory integerValueOf: index.
	"Check if we have already looked up the function and failed."
	index < 0 ifTrue:
		["Function address was not found in this session, 
		  Void the primitive function."
		 self rewriteMethodCacheEntryForExternalPrimitiveToFunction: 0.
		 ^self primitiveFailFor: PrimErrNotFound].

	"Try to call the function directly"
	(index > 0 and: [index <= MaxExternalPrimitiveTableSize]) ifTrue:
		[addr := externalPrimitiveTable at: index - 1.
		 addr ~= 0 ifTrue:
			[self rewriteMethodCacheEntryForExternalPrimitiveToFunction: (self cCode: 'addr' inSmalltalk: [1000 + index]).
			 self callExternalPrimitive: addr. "On Spur, sets primitiveFunctionPointer"
			 self maybeRetryPrimitiveOnFailure.
			 ^nil].
		"if we get here, then an index to the external prim was 
		kept on the ST side although the underlying prim 
		table was already flushed"
		^self primitiveFailFor: PrimErrNamedInternal].

	"Clean up session id/metadata and external primitive index"
	objectMemory storePointerUnchecked: ExternalCallLiteralFlagsIndex ofObject: lit withValue: ConstZero.
	objectMemory storePointerUnchecked: ExternalCallLiteralTargetFunctionIndex ofObject: lit withValue: ConstZero.

	"The function has not been loaded yet. Attempt to link it, cache it, and call it."
	addr := self linkExternalCall: lit errInto: (self addressOf: primFailCode put: [:v| primFailCode := v]).
	addr = 0 ifTrue:
		[self assert: (objectMemory fetchPointer: ExternalCallLiteralFlagsIndex ofObject: lit) = ConstZero.
		 ^self primitiveFailFor: (primFailCode = 0 ifTrue: [PrimErrNotFound] ifFalse: [primFailCode])].

	self callExternalPrimitive: addr.
	self maybeRetryPrimitiveOnFailure	! !

!StackInterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 1/26/2021 16:18'!
primitiveFindHandlerContext
	"Primitive. Search up the context stack for the next method context marked
	 for exception handling starting at the receiver. Return nil if none found"
	| handlerOrNilOrZero |
	self externalWriteBackHeadFramePointers.
	handlerOrNilOrZero := self
							findMethodWithPrimitive: PrimNumberHandlerMarker
							FromContext: self stackTop
							UpToContext: objectMemory nilObject.
	handlerOrNilOrZero = 0 ifTrue:
		[handlerOrNilOrZero := objectMemory nilObject].
	self pop: 1 thenPush: handlerOrNilOrZero! !

!StackInterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 1/26/2021 16:29'!
primitiveFindNextUnwindContext
	"Primitive. Search up the context stack for the next method context marked for unwind
	 handling from the receiver up to but not including the argument. Return nil if none found."
	| stopContext calleeContext handlerOrNilOrZero |
	<var: #theFP type: #'char *'>
	stopContext := self stackTop.
	calleeContext := self stackValue: 1.
	(stopContext = objectMemory nilObject or: [objectMemory isContext: stopContext]) ifFalse:
		[^self primitiveFail].
	"The following should never be true, but developing full blocks, early in September
	 2016 we were seeing invalid invocations of this primitive..  Hence the assert:"
	self assert: stopContext ~= calleeContext.
	self externalWriteBackHeadFramePointers.
	(self isStillMarriedContext: calleeContext)
		ifTrue:
			[| theFP |
			 theFP := self frameOfMarriedContext: calleeContext.
			 (self isBaseFrame: theFP)
				ifTrue:
					[handlerOrNilOrZero := self findMethodWithPrimitive: PrimNumberUnwindMarker
												FromContext: (self frameCallerContext: theFP)
												UpToContext: stopContext]
				ifFalse:
					[handlerOrNilOrZero :=  self findMethodWithPrimitive: PrimNumberUnwindMarker
												FromFP: (self frameCallerFP: theFP)
												UpToContext: stopContext]]
		ifFalse:
			[| startContext |
			 startContext := objectMemory fetchPointer: SenderIndex ofObject: calleeContext.
			 (objectMemory isContext: startContext)
				ifTrue:
					[handlerOrNilOrZero := self findMethodWithPrimitive: PrimNumberUnwindMarker
												FromContext: startContext
												UpToContext: stopContext]
				ifFalse:
					[handlerOrNilOrZero := 0]].
	handlerOrNilOrZero = 0 ifTrue:
		[handlerOrNilOrZero := objectMemory nilObject].
	self pop: 2 thenPush: handlerOrNilOrZero! !

!StackInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 6/21/2017 15:18'!
primitiveFlushCacheByMethod
	"The receiver is a compiledMethod.  Clear all entries in the method lookup cache that refer to this method, presumably because it has been redefined, overridden or removed."
	self flushMethodCacheForMethod: self stackTop! !

!StackInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 4/9/2009 14:35'!
primitiveFlushCacheBySelector
	"The receiver is a message selector.  Clear all entries in the method lookup cache
	 with this selector, presumably because an associated method has been redefined."
	| selector probe |
	selector := self stackTop.
	probe := 0.
	1 to: MethodCacheEntries do:
		[:i | (methodCache at: probe + MethodCacheSelector) = selector ifTrue:
			[methodCache at: probe + MethodCacheSelector put: 0].
		probe := probe + MethodCacheEntrySize].
	(selector = (self specialSelector: 16) "at:"
	 or: [selector = (self specialSelector: 17) "at:put:"]) ifTrue:
		[self flushAtCache]! !

!StackInterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 11/25/2013 11:28'!
primitiveFullGC
	"Do a full garbage collection.  In SqueakV3ObjectMemory, answer the number
	 of bytes available (including swap space if dynamic memory management is
	 supported).  In Spur, answer the size of the largest free chunk."

	self externalWriteBackHeadFramePointers.
	super primitiveFullGC! !

!StackInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 7/29/2021 11:24'!
primitiveGetVMParameter: arg 
	"See primitiveVMParameter method comment.
	 N.B. written as a returning case to avoid branch limits in the V3 bytecode set."
	arg caseOf: {
			[1]  ->	[^self positiveMachineIntegerFor: objectMemory oldSpaceSize].
			[2]  ->	[^objectMemory integerObjectOf: objectMemory newSpaceSize].
			[3]  ->	[^self positiveMachineIntegerFor: objectMemory totalMemorySize].
			[6]  ->	[^objectMemory integerObjectOf: objectMemory tenuringThreshold].
			[7]  ->	[^objectMemory integerObjectOf: objectMemory statFullGCs].
			[8]  ->	[^objectMemory integerObjectOf: objectMemory statFullGCUsecs + 500 // 1000].
			[9]  ->	[^objectMemory integerObjectOf: (objectMemory hasSpurMemoryManagerAPI
														ifTrue: [objectMemory statScavenges]
														ifFalse: [objectMemory statIncrGCs])].
			[10] ->	[^objectMemory integerObjectOf: (objectMemory hasSpurMemoryManagerAPI
														ifTrue: [objectMemory statScavengeGCUsecs]
														ifFalse: [objectMemory statIncrGCUsecs]) + 500 // 1000].
			[11] ->	[^objectMemory integerObjectOf: objectMemory statTenures].
			[12] ->	[^objectMemory integerObjectOf: eventTraceMask].
			[13] ->	[^self getVMTickerStartUSecs].
			[14] ->	[^self getVMTickerCount].
			[15] ->	[^self getVMTickeeCallCount].
			[16] ->	[^self positive64BitIntegerFor: statIdleUsecs].
			[17] ->	[^(SistaVM and: [self isCog])
						ifTrue: [objectMemory floatObjectOf: self getCogCodeZoneThreshold]
						ifFalse: [ConstZero]].
			[18] ->	[^objectMemory hasSpurMemoryManagerAPI
						ifTrue: [objectMemory integerObjectOf: objectMemory statCompactionUsecs + 500 // 1000]
						ifFalse: [ConstZero]].
			[19] ->	[^objectMemory hasSpurMemoryManagerAPI
						ifTrue: [objectMemory integerObjectOf: objectMemory scavengeThresholdAsExtent]
						ifFalse: [ConstZero]].
			[20] ->	[^objectMemory positive64BitIntegerFor: self ioUTCStartMicroseconds].
			[21] ->	[^objectMemory integerObjectOf: objectMemory rootTableCount].
			[22] ->	[^objectMemory integerObjectOf: objectMemory statRootTableOverflows].
			[23] ->	[^objectMemory integerObjectOf: extraVMMemory].
			[24] ->	[^objectMemory integerObjectOf: objectMemory shrinkThreshold].
			[25] ->	[^objectMemory integerObjectOf: objectMemory growHeadroom].
			[26] ->	[^objectMemory integerObjectOf: self ioHeartbeatMilliseconds].
			[27] ->	[^objectMemory integerObjectOf: objectMemory statMarkCount].
			[28] ->	[^objectMemory integerObjectOf: objectMemory statSweepCount].
			[29] ->	[^objectMemory integerObjectOf: objectMemory statMkFwdCount].
			[30] ->	[^objectMemory integerObjectOf: objectMemory statCompMoveCount].
			[31] ->	[^objectMemory integerObjectOf: objectMemory statGrowMemory].
			[32] ->	[^objectMemory integerObjectOf: objectMemory statShrinkMemory].
			[33] ->	[^objectMemory integerObjectOf: objectMemory statRootTableCount].
			[34] ->	[^objectMemory hasSpurMemoryManagerAPI ifTrue:"was statAllocationCount"
						[objectMemory positive64BitIntegerFor: objectMemory currentAllocatedBytes]].
			[35] ->	[^objectMemory integerObjectOf: objectMemory statSurvivorCount].
			[36] ->	[^objectMemory integerObjectOf: (self microsecondsToMilliseconds: objectMemory statGCEndUsecs)].
			[37] ->	[^objectMemory integerObjectOf: objectMemory statSpecialMarkCount].
			[38] ->	[^objectMemory integerObjectOf: objectMemory statIGCDeltaUsecs + 500 // 1000].
			[39] ->	[^objectMemory integerObjectOf: statPendingFinalizationSignals].
			[40] ->	[^objectMemory integerObjectOf: objectMemory wordSize].
			[41] ->	[^objectMemory integerObjectOf: self imageFormatVersion].
			[42] ->	[^objectMemory integerObjectOf: numStackPages].
			[43] ->	[^objectMemory integerObjectOf: desiredNumStackPages].
			[44] ->	[^objectMemory integerObjectOf: objectMemory edenBytes].
			[45] ->	[^objectMemory integerObjectOf: desiredEdenBytes].
			[46] ->	[^self getCogCodeSize].
			[47] ->	[^self getDesiredCogCodeSize].
			[48] ->	[^self getImageHeaderFlagsParameter].
			[49] ->	[^objectMemory integerObjectOf: self ioGetMaxExtSemTableSize].
			[52] ->	[^objectMemory integerObjectOf: objectMemory rootTableCapacity].
			[53] ->	[^objectMemory hasSpurMemoryManagerAPI ifTrue:
						[objectMemory integerObjectOf: objectMemory numSegments]].
			[54] ->	[^objectMemory hasSpurMemoryManagerAPI ifTrue:
						[objectMemory integerObjectOf: objectMemory freeSize]].
			[55] ->	[^objectMemory hasSpurMemoryManagerAPI ifTrue:
						[objectMemory floatObjectOf: objectMemory getHeapGrowthToSizeGCRatio]].
			[56] ->	[^self positive64BitIntegerFor: statProcessSwitch].
			[57] ->	[^self positive64BitIntegerFor: statIOProcessEvents].
			[58] ->	[^self positive64BitIntegerFor: statForceInterruptCheck].
			[59] ->	[^self positive64BitIntegerFor: statCheckForEvents].
			[60] ->	[^self positive64BitIntegerFor: statStackOverflow].
			[61] ->	[^self positive64BitIntegerFor: statStackPageDivorce].
			[62] ->	[^self getCodeCompactionCount].
			[63] ->	[^self getCodeCompactionMSecs].
			[64] ->	[^self getCogMethodCount].
			[65] ->	[^self getCogVMFeatureFlags].
			[66] ->	[^objectMemory integerObjectOf: self stackPageByteSize].
			[67] ->	[^objectMemory hasSpurMemoryManagerAPI ifTrue:
						[self positiveMachineIntegerFor: objectMemory maxOldSpaceSize]].
			[68] ->	[^objectMemory floatObjectOf: stackPages statAverageLivePagesWhenMapping].
			[69] ->	[^objectMemory integerObjectOf: stackPages statMaxPageCountWhenMapping].
			[70] ->	[^objectMemory integerObjectOf: self vmProxyMajorVersion].
			[71] ->	[^objectMemory integerObjectOf: self vmProxyMinorVersion].
			[72] ->	[^objectMemory integerObjectOf: objectMemory statMarkUsecs + 500 // 1000].
			[73] ->	[^objectMemory integerObjectOf: objectMemory statSweepUsecs + 500 // 1000].
			[74] ->	[^objectMemory hasSpurMemoryManagerAPI ifTrue:
						[objectMemory integerObjectOf: objectMemory statMaxAllocSegmentTime + 500 // 1000]].
			[75] ->	[^objectMemory booleanObjectOf: self primitiveDoMixedArithmetic].
			[76] ->	[^objectMemory integerObjectOf: self minimumUnusedHeadroom] }
		otherwise: [^nil]! !

!StackInterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 6/6/2016 12:22'!
primitiveIncrementalGC
	"Do a quick, incremental garbage collection and return the number of bytes immediately available.
	 (Note: more space may be made available by doing a full garbage collection."

	self externalWriteBackHeadFramePointers.
	super primitiveIncrementalGC! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'cb 11/26/2015 17:37'!
primitiveInstVarAt
	| index rcvr hdr fmt totalLength fixedFields value |
	index := self stackTop.
	rcvr := self stackValue: 1.
	((objectMemory isNonIntegerObject: index)
	 or: [argumentCount > 1 "e.g. object:instVarAt:"
		and: [objectMemory isOopForwarded: rcvr]]) ifTrue:
		[^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory isImmediate: rcvr) ifTrue: [^self primitiveFailFor: PrimErrInappropriate].
	index := objectMemory integerValueOf: index.
	hdr := objectMemory baseHeader: rcvr.
	fmt := objectMemory formatOfHeader: hdr.
	totalLength := objectMemory lengthOf: rcvr baseHeader: hdr format: fmt.
	fixedFields := objectMemory fixedFieldsOf: rcvr format: fmt length: totalLength.
	(index >= 1 and: [index <= fixedFields]) ifFalse:
		[^self primitiveFailFor: PrimErrBadIndex].
	(fmt = objectMemory indexablePointersFormat
	 and: [objectMemory isContextHeader: hdr])
		ifTrue: [value := self externalInstVar: index - 1 ofContext: rcvr]
		ifFalse: [value := self subscript: rcvr with: index format: fmt].
	self pop: argumentCount + 1 thenPush: value! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 1/20/2016 14:35'!
primitiveInstVarAtPut
	| newValue index rcvr hdr fmt totalLength fixedFields |
	newValue := self stackTop.
	index := self stackValue: 1.
	rcvr := self stackValue: 2.
	((objectMemory isNonIntegerObject: index)
	 or: [argumentCount > 2 "e.g. object:instVarAt:put:"
		and: [objectMemory isOopForwarded: rcvr]]) ifTrue:
		[^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory isImmediate: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	(objectMemory isObjImmutable: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].
	index := objectMemory integerValueOf: index.
	hdr := objectMemory baseHeader: rcvr.
	fmt := objectMemory formatOfHeader: hdr.
	totalLength := objectMemory lengthOf: rcvr baseHeader: hdr format: fmt.
	fixedFields := objectMemory fixedFieldsOf: rcvr format: fmt length: totalLength.
	(index >= 1 and: [index <= fixedFields]) ifFalse:
		[^self primitiveFailFor: PrimErrBadIndex].
	(fmt = objectMemory indexablePointersFormat
	 and: [objectMemory isContextHeader: hdr])
		ifTrue: [self externalInstVar: index - 1 ofContext: rcvr put: newValue]
		ifFalse: [self subscript: rcvr with: index storing: newValue format: fmt].
	self pop: argumentCount + 1 thenPush: newValue! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 1/4/2017 08:51'!
primitiveObjectPointsTo
	"This primitive is assumed to be fast (see e.g. MethodDictionary>>includesKey:) so make it so.
	 N.B.  Works forrectly for cogged methods too."
	| rcvr thang header fmt numSlots methodHeader |
	thang := self stackTop.
	rcvr := self stackValue: 1.
	(objectMemory isImmediate: rcvr) ifTrue:
		[^self pop: 2 thenPushBool: false].

	"Inlined version of lastPointerOf: for speed in determining if rcvr is a context."
	header := objectMemory baseHeader: rcvr.
	fmt := objectMemory formatOfHeader: header.
	(objectMemory isPointersFormat: fmt)
		ifTrue:
			[(fmt = objectMemory indexablePointersFormat
			  and: [objectMemory isContextHeader: header]) 
				ifTrue:
	 				[(self isMarriedOrWidowedContext: rcvr) ifTrue:
						[self externalWriteBackHeadFramePointers.
						 (self isStillMarriedContext: rcvr) ifTrue:
							[^self pop: 2
									thenPushBool: (self marriedContext: rcvr
														pointsTo: thang
														stackDeltaForCurrentFrame: 2)]].
					"contexts end at the stack pointer"
					numSlots := CtxtTempFrameStart + (self fetchStackPointerOf: rcvr)]
				ifFalse:
					[numSlots := objectMemory numSlotsOf: rcvr]]
		ifFalse:
			[fmt < objectMemory firstCompiledMethodFormat "no pointers" ifTrue:
				[^self pop: 2 thenPushBool: false].
			"CompiledMethod: contains both pointers and bytes:"
			methodHeader := objectMemory methodHeaderOf: rcvr.
			methodHeader = thang ifTrue: [^self pop: 2 thenPushBool: true].
			numSlots := (objectMemory literalCountOfMethodHeader: methodHeader) + LiteralStart].

	self assert: numSlots - 1 * objectMemory bytesPerOop + objectMemory baseHeaderSize = (objectMemory lastPointerOf: rcvr).
	objectMemory baseHeaderSize
		to: numSlots - 1 * objectMemory bytesPerOop + objectMemory baseHeaderSize
		by: objectMemory bytesPerOop
		do: [:i|
			(objectMemory longAt: rcvr + i) = thang ifTrue:
				[^self pop: 2 thenPushBool: true]].
	self pop: 2 thenPushBool: false! !

!StackInterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem 2/17/2017 09:20'!
primitivePerform
	<returnTypeC: #void>
	| newReceiver lookupClassTag performMethod |
	performMethod := newMethod.
	messageSelector := self stackValue: argumentCount - 1.
	newReceiver := self stackValue: argumentCount.

	"NOTE: the following lookup may fail and be converted to #doesNotUnderstand:,
	 so we must adjust argumentCount and slide args now, so that will work."

	"Slide arguments down over selector"
	argumentCount := argumentCount - 1.
	argumentCount to: 1 by: -1 do:
		[:i|
		stackPages
			longAt: stackPointer + (i * objectMemory wordSize)
			put: (stackPages longAt: stackPointer + ((i - 1) * objectMemory wordSize))].
	self pop: 1.
	lookupClassTag := objectMemory fetchClassTagOf: newReceiver.
	self sendBreakpoint: messageSelector receiver: newReceiver.
	self printSends ifTrue:
		[self printActivationNameForSelector: messageSelector
			startClass: (objectMemory classForClassTag: lookupClassTag); cr].
	self findNewMethodInClassTag: lookupClassTag.

	"Only test CompiledMethods for argument count - other objects will have to take their chances"
	((objectMemory isOopCompiledMethod: newMethod)
	  and: [(self argumentCountOf: newMethod) = argumentCount]) ifFalse:
		["Slide the args back up (sigh) and re-insert the selector."
		self unPop: 1.
		1 to: argumentCount by: 1 do:
			[:i |
			stackPages longAt: stackPointer + ((i - 1) * objectMemory wordSize)
				put: (stackPages longAt: stackPointer + (i * objectMemory wordSize))].
		stackPages longAt: stackPointer + (argumentCount * objectMemory wordSize) put: messageSelector.
		argumentCount := argumentCount + 1.
		newMethod := performMethod.
		 "Must reset primitiveFunctionPointer for checkForAndFollowForwardedPrimitiveState"
		 objectMemory hasSpurMemoryManagerAPI ifTrue:
			[primitiveFunctionPointer := #primitivePerform].
		^self primitiveFailFor: PrimErrBadNumArgs].

	self executeNewMethod.
	"Recursive xeq affects primErrorCode"
	self initPrimCall! !

!StackInterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem 1/3/2015 13:08'!
primitivePin
	"Pin or unpin the receiver, i.e. make it immobile or mobile, based on the argument.
	 Answer whether the object was already pinned. N.B. pinning does *not* prevent
	 an object from being garbage collected."
	| obj boolean wasPinned |
	objectMemory hasSpurMemoryManagerAPI ifFalse:
		[^self primitiveFailFor: PrimErrUnsupported].

	obj := self stackValue: 1.
	((objectMemory isImmediate: obj)
	 or: [objectMemory isForwarded: obj]) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].

	boolean := self stackTop.
	(boolean = objectMemory falseObject
	 or: [boolean = objectMemory trueObject]) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].

	(objectMemory isPinned: obj)
		ifTrue:
			[wasPinned := objectMemory trueObject.
			 boolean ~= wasPinned ifTrue:
				[objectMemory setIsPinnedOf: obj to: false]]
		ifFalse:
			[wasPinned := objectMemory falseObject.
			 boolean = objectMemory trueObject ifTrue:
				[((objectMemory isContext: obj)
				  and: [self isStillMarriedContext: obj]) ifTrue:
					[^self primitiveFailFor: PrimErrBadReceiver].
				 (objectMemory pinObject: obj) = 0 ifTrue:
					[^self primitiveFailFor: PrimErrNoMemory]]].
	
	self pop: argumentCount + 1 thenPush: wasPinned! !

!StackInterpreterPrimitives methodsFor: 'memory space primitives' stamp: 'eem (auto pragmas 12/08) 11/25/2008 12:50'!
primitiveSetGCSemaphore
	"Primitive. Indicate the semaphore to be signalled for upon garbage collection"
	| index |
	<export: true>
	index := self stackIntegerValue: 0.
	self successful ifTrue:[
		gcSemaphoreIndex := index.
		self pop: argumentCount.
	]! !

!StackInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 12/22/2020 14:38'!
primitiveSetVMParameter: index arg: argOop
	"See primitiveVMParameter method comment"
	| arg result |

	"argOop read & checks; in most cases this is an integer parameter. Handle the exceptions."
	index
		caseOf: {
		[17]	->	[((objectMemory isFloatInstance: argOop)
				 	 or: [objectMemory isIntegerObject: argOop]) ifFalse:
						[primFailCode := PrimErrBadArgument]].
		[55]	->	[((objectMemory isFloatInstance: argOop)
				 	 or: [objectMemory isIntegerObject: argOop]) ifFalse:
						[primFailCode := PrimErrBadArgument]].
		[68]	->	[((objectMemory isFloatInstance: argOop)
				 	 or: [objectMemory isIntegerObject: argOop]) ifFalse:
						[primFailCode := PrimErrBadArgument]].
		[67]	->	[arg := self positiveMachineIntegerValueOf: argOop].
		[75]	->	[arg := objectMemory booleanValueOf: argOop] }
		otherwise: [arg := objectMemory integerValueOf: argOop].
	self failed ifTrue:
		[^self primitiveFailFor: PrimErrBadArgument].

	"assume failure, then set success for handled indices"
	self primitiveFailFor: PrimErrBadArgument.
	index caseOf: {
		[5] ->	[objectMemory hasSpurMemoryManagerAPI ifFalse:
					["Was:
							result := allocationsBetweenGCs.
							allocationsBetweenGCs := arg."
						"Ignore for now, because old images won't start up otherwise.
						 See 45 for eden size setting."
					 result := objectMemory nilObject.
					 self initPrimCall]].
		[6] ->	[result := objectMemory integerObjectOf: objectMemory tenuringThreshold.
				 primFailCode := objectMemory tenuringThreshold: arg].
		[12] ->	[result := objectMemory integerObjectOf: eventTraceMask.
				 eventTraceMask := arg.
				 self initPrimCall].
		[11] ->	[arg >= 0 ifTrue:
					[result := objectMemory integerObjectOf: objectMemory statTenures.
					 objectMemory statTenures: arg.
					 self initPrimCall]].
		[17] ->	[(SistaVM and: [self isCog]) ifTrue:
					[result := objectMemory floatObjectOf: self getCogCodeZoneThreshold.
					 primFailCode := self setCogCodeZoneThreshold: (self noInlineLoadFloatOrIntFrom: argOop)]].
		[23] ->	[result := objectMemory integerObjectOf: extraVMMemory.
				 extraVMMemory := arg.
				 self initPrimCall].
		[24] ->	[arg > 0 ifTrue:
					[result := objectMemory integerObjectOf: objectMemory shrinkThreshold.
					 objectMemory shrinkThreshold: arg.
					 self initPrimCall]].
		[25] ->	[arg > 0 ifTrue:
					[result := objectMemory integerObjectOf: objectMemory growHeadroom.
					 objectMemory growHeadroom: arg.
					 self initPrimCall]].
		[26] ->	[arg >= 0 ifTrue: "0 turns off the heartbeat"
					[result := objectMemory integerObjectOf: self ioHeartbeatMilliseconds.
					 self ioSetHeartbeatMilliseconds: arg.
					 self initPrimCall]].
		[34] ->	[(objectMemory hasSpurMemoryManagerAPI "was statAllocationCount; now statAllocatedBytes"
				  and: [arg >= 0]) ifTrue:
					[result := objectMemory positive64BitIntegerFor: objectMemory currentAllocatedBytes.
					 objectMemory setCurrentAllocatedBytesTo: arg.
					 self initPrimCall]].
		[43] ->	[(arg between: 0 and: 65535) ifTrue:
					[result := objectMemory integerObjectOf: desiredNumStackPages.
					 desiredNumStackPages := arg.
					 self initPrimCall]].
		[45] ->	[arg >= 0 ifTrue:
					[result := objectMemory integerObjectOf: desiredEdenBytes.
					 desiredEdenBytes := arg.
					 self initPrimCall]].
		[47] ->	[(self isCog
				  and: [arg between: 0 and: self maxCogCodeSize]) ifTrue:
					[result := objectMemory integerObjectOf: self getDesiredCogCodeSize.
					 self setDesiredCogCodeSize: arg.
					 self initPrimCall]].
		[48] ->	[arg >= 0 ifTrue:
					[| oldPrimitiveDoMixedArithmetic |
					 oldPrimitiveDoMixedArithmetic := primitiveDoMixedArithmetic.
					 result := self getImageHeaderFlagsParameter.
					 self initPrimCall. "i.e. setImageHeaderFlags: can fail"
					 self setImageHeaderFlags: arg.
					 (primFailCode = 0
					  and: [oldPrimitiveDoMixedArithmetic ~= primitiveDoMixedArithmetic]) ifTrue:
						[self flushMethodCache.
						 self flushMethodsWithMachineCodePrimitivesAndContinueAnswering: result
						 "NOT REACHED (in CoInterpreter)"]]].
		[49] ->	[(arg between: 0 and: 65535) ifTrue:
					[result := objectMemory integerObjectOf: self ioGetMaxExtSemTableSize.
					 self initPrimCall. "i.e. ioSetMaxExtSemTableSize: is allowed to fail"
					 self setMaxExtSemSizeTo: arg]].
		[55] ->	[objectMemory hasSpurMemoryManagerAPI ifTrue:
					[result := objectMemory floatObjectOf: objectMemory getHeapGrowthToSizeGCRatio.
					 primFailCode := objectMemory setHeapGrowthToSizeGCRatio: (self noInlineLoadFloatOrIntFrom: argOop)]].
		[67] ->	[(arg >= 0
				  and: [objectMemory hasSpurMemoryManagerAPI]) ifTrue:
					[result := self positiveMachineIntegerFor: objectMemory maxOldSpaceSize.
					 primFailCode := objectMemory setMaxOldSpaceSize: arg]].
		[68] ->	[result := objectMemory floatObjectOf: stackPages statAverageLivePagesWhenMapping.
				 self initPrimCall. "i.e. statAverageLivePagesWhenMapping: is allowed to fail"
				 stackPages statAverageLivePagesWhenMapping: (self noInlineLoadFloatOrIntFrom: argOop)].
		[69] ->	[arg >= 0 ifTrue:
					[result := objectMemory integerObjectOf: stackPages statMaxPageCountWhenMapping.
					 stackPages statMaxPageCountWhenMapping: arg.
					 self initPrimCall]].
		[74] ->	[(arg >= 0
				  and: [objectMemory hasSpurMemoryManagerAPI]) ifTrue:
					[result := objectMemory integerObjectOf: objectMemory statMaxAllocSegmentTime + 500 // 1000.
					 stackPages statMaxAllocSegmentTime: arg. "usually 0"
					 self initPrimCall]].
		[75] ->	[| mustFlush |
				 result := objectMemory booleanObjectOf: self primitiveDoMixedArithmetic.
				 self initPrimCall.
				 mustFlush := primitiveDoMixedArithmetic ~= arg.
				 primitiveDoMixedArithmetic := arg.
				 mustFlush ifTrue:
					[self flushMethodCache.
					 self flushMethodsWithMachineCodePrimitivesAndContinueAnswering: result
					 "NOT REACHED (in CoInterpreter)"]] }
		otherwise: [].

	self successful
		ifTrue: [self methodReturnValue: result]  "return old value"
		ifFalse: [self primitiveFailFor: PrimErrInappropriate] "attempting to write a read-only or non-existent parameter"! !

!StackInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 8/29/2020 19:17'!
primitiveSignalAtMilliseconds
	"Cause the time semaphore, if one has been registered, to be
	 signalled when the microsecond clock is greater than or equal to
	 the given tick value. A tick value of zero turns off timer interrupts."
	| msecsObj msecs deltaMsecs sema limit |
	<var: #msecs type: #sqLong>
	<var: #deltaMsecs type: #sqLong>
	<var: #limit type: #sqLong>
	msecsObj := self stackTop.
	sema := self stackValue: 1.
	msecs := self positive32BitValueOf: msecsObj.
	
	self successful ifTrue:
		[(objectMemory isSemaphoreOop: sema) ifTrue:
			[objectMemory splObj: TheTimerSemaphore put: sema.
			 deltaMsecs := msecs - (self ioMSecs bitAnd: MillisecondClockMask).
			 limit := MillisecondClockMask >> 1.
			 "Handle a roll-over that could happen in between image invocation of ioMSecs and this invocation.
			  This will limit the maximum relative duration to MillisecondClockMask/2, about 3 days currently.
			  Every delay longer than that limit may lead to undefined behavior (shorten delay, or no delay at all)"
			 deltaMsecs > limit ifTrue: [deltaMsecs := deltaMsecs - MillisecondClockMask].
			 nextWakeupUsecs := deltaMsecs > 0
				ifTrue: [self ioUTCMicroseconds + (deltaMsecs * 1000)]
				ifFalse: [self ioUTCMicroseconds].
			 ^self pop: 2].
		 sema = objectMemory nilObject ifTrue:
			[objectMemory
				storePointerUnchecked: TheTimerSemaphore
				ofObject: objectMemory specialObjectsOop
				withValue: objectMemory nilObject.
			 nextWakeupUsecs := 0.
			 ^self pop: 2]].
	self primitiveFailFor: PrimErrBadArgument! !

!StackInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 8/29/2020 19:17'!
primitiveSignalAtUTCMicroseconds
	"Cause the time semaphore, if one has been registered, to be
	 signalled when the microsecond clock is greater than or equal to
	 the given tick value. A tick value of zero turns off timer interrupts."
	| usecsObj sema usecs |
	<var: #usecs type: #usqLong>
	usecsObj := self stackTop.
	sema := self stackValue: 1.
	usecs := self positive64BitValueOf: usecsObj.
	self successful ifTrue:
		[(objectMemory isSemaphoreOop: sema) ifTrue:
			[objectMemory splObj: TheTimerSemaphore put: sema.
			 nextWakeupUsecs := usecs.
			 ^self pop: 2].
		 sema = objectMemory nilObject ifTrue:
			[objectMemory
				storePointerUnchecked: TheTimerSemaphore
				ofObject: objectMemory specialObjectsOop
				withValue: objectMemory nilObject.
			 nextWakeupUsecs := 0.
			 ^self pop: 2]].
	self primitiveFailFor: PrimErrBadArgument! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'nice 11/14/2016 23:01'!
primitiveSlotAt
	"Answer a slot in an object.  This numbers all slots from 1, ignoring the distinction between
	 named and indexed inst vars.  In objects with both named and indexed inst vars, the named
	 inst vars precede the indexed ones.  In non-object indexed objects (objects that contain
	 bits, not object references) this primitive answers the raw integral value at each slot. 
	 e.g. for Strings it answers the character code, not the Character object at each slot."
	| index rcvr fmt numSlots |
	index := self stackTop.
	rcvr := self stackValue: 1.
	(objectMemory isIntegerObject: index) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory isImmediate: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	fmt := objectMemory formatOf: rcvr.
	index := (objectMemory integerValueOf: index) - 1.

	fmt <= objectMemory lastPointerFormat ifTrue:
		[numSlots := objectMemory numSlotsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[| value numLiveSlots |
			 (objectMemory isContextNonImm: rcvr)
				ifTrue:
					[self externalWriteBackHeadFramePointers.
					 numLiveSlots := (self stackPointerForMaybeMarriedContext: rcvr) + CtxtTempFrameStart.
					 value := (self asUnsigned: index) < numLiveSlots
								ifTrue: [self externalInstVar: index ofContext: rcvr]
								ifFalse: [objectMemory nilObject]]
				ifFalse:
					[value := objectMemory fetchPointer: index ofObject: rcvr].
			 self pop: argumentCount + 1 thenPush: value.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	fmt >= objectMemory firstByteFormat ifTrue:
		[fmt >= objectMemory firstCompiledMethodFormat ifTrue:
			[^self primitiveFailFor: PrimErrUnsupported].
		 numSlots := objectMemory numBytesOfBytes: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[self pop: argumentCount + 1 thenPushInteger: (objectMemory fetchByte: index ofObject: rcvr).
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	(objectMemory hasSpurMemoryManagerAPI
	 and: [fmt >= objectMemory firstShortFormat]) ifTrue:
		[numSlots := objectMemory num16BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[self pop: argumentCount + 1 thenPushInteger: (objectMemory fetchUnsignedShort16: index ofObject: rcvr).
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	fmt = objectMemory sixtyFourBitIndexableFormat ifTrue:
		[numSlots := objectMemory num64BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[self pop: argumentCount + 1
				thenPush: (self positive64BitIntegerFor: (objectMemory fetchLong64: index ofObject: rcvr)).
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	fmt >= objectMemory firstLongFormat ifTrue:
		[numSlots := objectMemory num32BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[self pop: argumentCount + 1
				thenPush: (self positive32BitIntegerFor: (objectMemory fetchLong32: index ofObject: rcvr)).
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	^self primitiveFailFor: PrimErrBadReceiver! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 2/26/2018 12:32'!
primitiveSlotAtPut
	"Assign a slot in an object.  This numbers all slots from 1, ignoring the distinction between
	 named and indexed inst vars.  In objects with both named and indexed inst vars, the named
	 inst vars precede the indexed ones.  In non-object indexed objects (objects that contain
	 bits, not object references) this primitive assigns a raw integral value at each slot."
	| newValue index rcvr fmt numSlots value |
	newValue := self stackTop.
	index := self stackValue: 1.
	rcvr := self stackValue: 2.
	(objectMemory isIntegerObject: index) ifFalse:
		[^self primitiveFailFor: PrimErrBadArgument].
	(objectMemory isImmediate: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrBadReceiver].
	(objectMemory isObjImmutable: rcvr) ifTrue:
		[^self primitiveFailFor: PrimErrNoModification].

	fmt := objectMemory formatOf: rcvr.
	index := (objectMemory integerValueOf: index) - 1.

	fmt <= objectMemory lastPointerFormat ifTrue:
		[numSlots := objectMemory numSlotsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[(objectMemory isContextNonImm: rcvr)
				ifTrue: [self externalInstVar: index ofContext: rcvr put: newValue]
				ifFalse: [objectMemory storePointer: index ofObject: rcvr withValue: newValue].
			 self pop: argumentCount + 1 thenPush: newValue.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	value := self positiveMachineIntegerValueOf: newValue.
	self failed ifTrue:
		[primFailCode := PrimErrBadArgument.
		^0].

	fmt >= objectMemory firstByteFormat ifTrue:
		[fmt >= objectMemory firstCompiledMethodFormat ifTrue:
			[^self primitiveFailFor: PrimErrUnsupported].
		 (self asUnsigned: value) > 16rFF ifTrue:
			[^self primitiveFailFor: PrimErrBadArgument].
		 numSlots := objectMemory numBytesOfBytes: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storeByte: index ofObject: rcvr withValue: value.
			 self pop: argumentCount + 1 thenPush: newValue.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	(objectMemory hasSpurMemoryManagerAPI
	 and: [fmt >= objectMemory firstShortFormat]) ifTrue:
		[(self asUnsigned: value) > 16rFFFF ifTrue:
			[^self primitiveFailFor: PrimErrBadArgument].
		 numSlots := objectMemory num16BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storeShort16: index ofObject: rcvr withValue: value.
			 self pop: argumentCount + 1 thenPush: newValue.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	(objectMemory bytesPerOop = 8
	 and: [fmt = objectMemory sixtyFourBitIndexableFormat]) ifTrue:
		[numSlots := objectMemory num64BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storeLong64: index ofObject: rcvr withValue: value.
			 self pop: argumentCount + 1 thenPush: newValue.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	fmt >= objectMemory firstLongFormat ifTrue:
		[(objectMemory wordSize > 4
		  and: [(self asUnsigned: value) > 16rFFFFFFFF]) ifTrue:
			[^self primitiveFailFor: PrimErrBadArgument].
		 numSlots := objectMemory num32BitUnitsOf: rcvr.
		 (self asUnsigned: index) < numSlots ifTrue:
			[objectMemory storeLong32: index ofObject: rcvr withValue: value.
			 self pop: argumentCount + 1 thenPush: newValue.
			 ^0].
		 ^self primitiveFailFor: PrimErrBadIndex].

	^self primitiveFailFor: PrimErrBadReceiver! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 3/10/2017 13:09'!
primitiveStoreStackp
	"Atomic store into context stackPointer. 
	Also ensures that any newly accessible cells are initialized to nil "
	| ctxt newStackp theFP thePage onCurrentPage stackp |
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	ctxt := self stackValue: 1.
	newStackp := self stackIntegerValue: 0.
	(self successful
	 and: [newStackp between: 0 and: (objectMemory numSlotsOf: ctxt) - CtxtTempFrameStart]) ifFalse:
		[^self primitiveFail].
	self externalWriteBackHeadFramePointers.
	(self isStillMarriedContext: ctxt) ifTrue:
		[theFP := self frameOfMarriedContext: ctxt.
		 thePage := stackPages stackPageFor: theFP.
		 ((onCurrentPage := thePage = stackPage)
		 and: [theFP = framePointer]) ifTrue:
			[^self primitiveFail]. "Probably easy to do this right here right now (just move stackPointer).  But fail for now."
		 self externalDivorceFrame: theFP andContext: ctxt.
		 onCurrentPage
			ifTrue:
				[self setStackPointersFromPage: stackPage]
			ifFalse:
				[self assert: stackPage = (stackPages stackPageFor: framePointer).
				 stackPages markStackPageMostRecentlyUsed: stackPage]].
	stackp := self fetchStackPointerOf: ctxt.
	"Nil any newly accessible cells"
	stackp + 1 to: newStackp do:
		[:i | objectMemory storePointerUnchecked: i + CtxtTempFrameStart - 1 ofObject: ctxt withValue: objectMemory nilObject].
	self storeStackPointerValue: newStackp inContext: ctxt.
	self ensureContextIsExecutionSafeAfterAssignToStackPointer: ctxt.
	self pop: 1! !

!StackInterpreterPrimitives methodsFor: 'control primitives' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
primitiveTerminateTo
	"Primitive. Terminate up the context stack from the receiver up to but not including
	 the argument, if previousContext is on my Context stack. Make previousContext my
	 sender. This prim has to shadow the code in ContextPart>terminateTo: to be correct."
	| thisCtx currentCtx aContextOrNil contextsFP contextsSP contextsIP nextCntx stackedReceiverOffset 
	  theFP newFP newSP pageToStopOn thePage frameAbove |
	<var: #contextsFP type: #'char *'>
	<var: #theFP type: #'char *'>
	<var: #newFP type: #'char *'>
	<var: #newSP type: #'char *'>
	<var: #contextsIP type: #usqInt>
	<var: #frameAbove type: #'char *'>
	<var: #contextsSP type: #'char *'>
	<var: #source type: #'char *'>
	<var: #pageToStopOn type: #'StackPage *'>
	<var: #thePage type: #'StackPage *'>

	aContextOrNil := self stackTop.
	(aContextOrNil = objectMemory nilObject or: [objectMemory isContext: aContextOrNil]) ifFalse:
		[^self primitiveFail].
	thisCtx := self stackValue: 1.
	thisCtx = aContextOrNil ifTrue:
		[^self primitiveFail].		

	"All stackPages need to have current head pointers to avoid confusion."
	self externalWriteBackHeadFramePointers.

	"If we're searching for aContextOrNil it might be on a stack page.  Helps to know
	 if we can free a whole page or not, or if we can short-cut the termination."
	(aContextOrNil ~= objectMemory nilObject and: [self isStillMarriedContext: aContextOrNil])
		ifTrue: [contextsFP := self frameOfMarriedContext: aContextOrNil.
				pageToStopOn := stackPages stackPageFor: contextsFP]
		ifFalse: [pageToStopOn := 0].

	"if thisCtx is married ensure it is a base frame.  Then we can assign its sender."
	(self isStillMarriedContext: thisCtx)
		ifTrue:
			[theFP := self frameOfMarriedContext: thisCtx.
			 "Optimize terminating thisContext.  Move its frame down to be next to
			  aContextOrNil's frame. Common in the exception system and so helps to be fast."
			 (theFP = framePointer
			  and: [pageToStopOn = stackPage]) ifTrue:
				[(self frameCallerFP: theFP) ~= contextsFP ifTrue:
					[stackedReceiverOffset := self frameStackedReceiverOffset: theFP.
					 frameAbove := self findFrameAbove: contextsFP inPage: pageToStopOn.
					 contextsIP := (self frameCallerSavedIP: frameAbove) asUnsignedInteger.
					 newSP := self frameCallerSP: frameAbove.
					 newFP := newSP - stackedReceiverOffset - objectMemory wordSize.
					 theFP + stackedReceiverOffset
						to: stackPointer
						by: objectMemory wordSize negated
						do: [:source|
							newSP := newSP - objectMemory wordSize.
							stackPages longAt: newSP put: (stackPages longAt: source)].
					 stackPages longAt: newFP + FoxSavedFP put: contextsFP.
					 stackPages longAt: newFP + FoxCallerSavedIP put: contextsIP.
					 self assert: (objectMemory isContext: thisCtx).
					 objectMemory storePointerUnchecked: SenderIndex
						ofObject: thisCtx
						withValue: (self withSmallIntegerTags: newFP).
					 objectMemory storePointerUnchecked: InstructionPointerIndex
						ofObject: thisCtx
						withValue: (self withSmallIntegerTags: contextsFP).
					 framePointer := newFP.
					 stackPointer := newSP].
				self pop: 1.
				self assert: stackPage = stackPages mostRecentlyUsedPage.
				^nil].
			 theFP := self externalEnsureIsBaseFrame: theFP. "May cause a GC!!!!"
			 currentCtx := self frameCallerContext: theFP.
			 "May also reclaim aContextOrNil's page, hence..."
			 (aContextOrNil ~= objectMemory nilObject and: [self isStillMarriedContext: aContextOrNil])
				ifTrue: [contextsFP := self frameOfMarriedContext: aContextOrNil.
						pageToStopOn := stackPages stackPageFor: contextsFP]
				ifFalse: [pageToStopOn := 0]]
		ifFalse:
			[currentCtx := objectMemory fetchPointer: SenderIndex ofObject: thisCtx].

	(self context: thisCtx hasSender: aContextOrNil) ifTrue:
		["Need to walk the stack freeing stack pages and nilling contexts."
		[currentCtx = aContextOrNil
		 or: [currentCtx = objectMemory nilObject]] whileFalse:
			[self assert: (objectMemory isContext: currentCtx).
			 (self isMarriedOrWidowedContext: currentCtx)
				ifTrue:
					[theFP := self frameOfMarriedContext: currentCtx.
					thePage := stackPages stackPageFor: theFP.
					"If externalEnsureIsBaseFrame: above has moved thisContext to its own stack
					 then we will always terminate to a frame on a different page.  But if we are
					 terminating some other context to a context somewhere on the current page
					 we must save the active frames above that context.  Things will look e.g. like this:
		thisCtx			499383332 s MethodContext(ContextPart)>resume:
						499380484 s BlockClosure>ensure:
						499377320 s MethodContext(ContextPart)>handleSignal:
						499373760 s MethodContext(ContextPart)>handleSignal:
						499372772 s MessageNotUnderstood(Exception)>signal
						499369068 s CodeSimulationTests(Object)>doesNotUnderstand: absentMethod
						499368708 s [] in CodeSimulationTests>testDNU
							(sender is 0xbffc2480 I CodeSimulationTests>runSimulated:)
						------------
		framePointer	0xbffc234c M MethodContext(ContextPart)>doPrimitive:method:receiver:args:
						0xbffc2378 M MethodContext(ContextPart)>tryPrimitiveFor:receiver:args:
						0xbffc23ac M MethodContext(ContextPart)>send:to:with:super:
						0xbffc23e4 M MethodContext(ContextPart)>send:super:numArgs:
						0xbffc2418 M MethodContext(InstructionStream)>interpretNextInstructionFor:
						0xbffc2434 M MethodContext(ContextPart)>step
						0xbffc2458 I MethodContext(ContextPart)>runSimulated:contextAtEachStep:
						------------
(499368708's sender)	0xbffc2480 I CodeSimulationTests>runSimulated:
						0xbffc249c M CodeSimulationTests>testDNU
						0xbffc24bc I CodeSimulationTests(TestCase)>performTest
						0xbffc24dc I [] in CodeSimulationTests(TestCase)>runCase
		aContextOrNil	0xbffc24fc M BlockClosure>ensure:
						0xbffc2520 I CodeSimulationTests(TestCase)>runCase
						0xbffc253c M [] in TestResult>runCase:
					When we find this case we move the frames above to a new page by making the
					frame above currentCtx a base frame, i.e. making 0xbffc2458 in the above example
					a base frame.  But in this iteration of the loop we don't move down a frame i.e. currentCtx
					doesn't change on this iteration."
					thePage = stackPage
						ifTrue:
							[frameAbove := self findFrameAbove: theFP inPage: thePage.
							self assert: frameAbove ~= 0.
							frameAbove := self externalEnsureIsBaseFrame: frameAbove. "May cause a GC!!!! May also reclaim aContextOrNil's page, hence..."
							(aContextOrNil ~= objectMemory nilObject and: [self isStillMarriedContext: aContextOrNil])
								ifTrue: [contextsFP := self frameOfMarriedContext: aContextOrNil.
										pageToStopOn := stackPages stackPageFor: contextsFP]
								ifFalse: [pageToStopOn := 0]]
						ifFalse:
							[thePage = pageToStopOn
								ifTrue:
									["We're here.  Cut back the stack to aContextOrNil's frame,
									  push its instructionPointer if it's not already a head frame,
									  and we're done."
									 frameAbove := self findFrameAbove: contextsFP inPage: thePage.
									 frameAbove ~= 0 ifTrue:
										[contextsSP := (self frameCallerSP: frameAbove) - objectMemory wordSize.
										 stackPages longAt: contextsSP put: (self frameCallerSavedIP: frameAbove).
										 self setHeadFP: contextsFP andSP: contextsSP inPage: thePage].
									 currentCtx := aContextOrNil]
								ifFalse:
									["We can free the entire page without further ado."
									 currentCtx := self frameCallerContext: thePage baseFP.
									 "for a short time invariant is violated; assert follows"
									 stackPages freeStackPageNoAssert: thePage]]]
				ifFalse:
					[nextCntx := objectMemory fetchPointer: SenderIndex ofObject: currentCtx.
					 self markContextAsDead: currentCtx.
					 currentCtx := nextCntx]]].
	self assert: stackPages pageListIsWellFormed.
	(self isMarriedOrWidowedContext: thisCtx)
		ifTrue:
			[self assert: (self checkIsStillMarriedContext: thisCtx currentFP: framePointer).
			 self assert: (self isBaseFrame: (self frameOfMarriedContext: thisCtx)).
			 theFP := self frameOfMarriedContext: thisCtx.
			 self frameCallerContext: theFP put: aContextOrNil]
		ifFalse: [objectMemory storePointer: SenderIndex ofObject: thisCtx withValue: aContextOrNil].
	self pop: 1.
	self assert: stackPage = stackPages mostRecentlyUsedPage! !

!StackInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 7/29/2021 11:22'!
primitiveVMParameter
	"Behaviour depends on argument count:
		0 args:	return an Array of VM parameter values;
		1 arg:	return the indicated VM parameter;
		2 args:	set the VM indicated parameter.
	VM parameters are numbered as follows:
		1	end (v3)/size(Spur) of old-space (0-based, read-only)
		2	end (v3)/size(Spur) of young/new-space (read-only)
		3	end (v3)/size(Spur) of heap (read-only)
		4	nil (was allocationCount (read-only))
		5	nil (was allocations between GCs (read-write)
		6	survivor count tenuring threshold (read-write)
		7	full GCs since startup (read-only)
		8	total milliseconds in full GCs since startup (read-only)
		9	incremental GCs (SqueakV3) or scavenges (Spur) since startup (read-only)
		10	total milliseconds in incremental GCs (SqueakV3) or scavenges (Spur) since startup (read-only)
		11	tenures of surving objects since startup or reset (read-write)
		12-20 were specific to ikp's JITTER VM, now 12 16 open for use
		13	if started, the start time in utc microseconds of the high-priority ticker
		14	if started, the number of checkHighPriorityTickees calls
		15	if started, the number of tickee calls from checkHighPriorityTickees
		16	total microseconds at idle since start-up (if non-zero)
		17	fraction of the code zone to use (Sista only; used to control code zone use to preserve sendAndBranchData on counter tripped callback)
		18	total milliseconds in compaction phase of full GC since start-up (Spur only)
		19	scavenge threshold, the effective size of eden.  When eden fills to the threshold a scavenge is scheduled. Newer Spur VMs only.
		20	utc microseconds at VM start-up (actually at time initialization, which precedes image load).
		21	root/remembered table size (occupancy) (read-only)
		22	root table overflows since startup (read-only)
		23	bytes of extra memory to reserve for VM buffers, plugins, etc (stored in image file header).
		24	memory threshold above which shrinking object memory (rw)
		25	memory headroom when growing object memory (rw)
		26	interruptChecksEveryNms - force an ioProcessEvents every N milliseconds (rw)
		27	number of times mark loop iterated for current IGC/FGC (read-only) includes ALL marking
		28	number of times sweep loop iterated for current IGC/FGC (read-only)
		29	number of times make forward loop iterated for current IGC/FGC (read-only)
		30	number of times compact move loop iterated for current IGC/FGC (read-only)
		31	number of grow memory requests (read-only)
		32	number of shrink memory requests (read-only)
		33	number of root table entries used for current IGC/FGC (read-only)
		34	Spur: bytes allocated in total since start-up or reset (read-write) (Used to be number of allocations done before current IGC/FGC (read-only))
		35	number of survivor objects after current IGC/FGC (read-only)
		36	millisecond clock when current IGC/FGC completed (read-only)
		37	number of marked objects for Roots of the world, not including Root Table entries for current IGC/FGC (read-only)
		38	milliseconds taken by current IGC (read-only)
		39	Number of finalization signals for Weak Objects pending when current IGC/FGC completed (read-only)
		40	BytesPerOop for this image
		41	imageFormatVersion for the VM
		42	number of stack pages in use
		43	desired number of stack pages (stored in image file header, max 65535)
		44	size of eden, in bytes
		45	desired size of eden, in bytes (stored in image file header)
		46	machine code zone size, in bytes (Cog only; otherwise nil)
		47	desired machine code zone size (stored in image file header; Cog only; otherwise nil)
		48	various header flags.  See getImageHeaderFlags.
		49	max size the image promises to grow the external semaphore table to (0 sets to default, which is 256 as of writing)
		50-51 nil; reserved for VM parameters that persist in the image (such as eden above)
		52	root/remembered table capacity
		53	number of segments (Spur only; otherwise nil)
		54	total size of free old space (Spur only, otherwise nil)
		55	ratio of growth and image size at or above which a GC will be performed post scavenge
		56	number of process switches since startup (read-only)
		57	number of ioProcessEvents calls since startup (read-only)
		58	number of ForceInterruptCheck calls since startup (read-only)
		59	number of check event calls since startup (read-only)
		60	number of stack page overflows since startup (read-only)
		61	number of stack page divorces since startup (read-only)
		62	compiled code compactions since startup (read-only; Cog only; otherwise nil)
		63	total milliseconds in compiled code compactions since startup (read-only; Cog only; otherwise nil)
		64	the number of methods that currently have jitted machine-code
		65	various VM feature flags; see getCogVMFeatureFlags
		66	the byte size of a stack page
		67	the max allowed size of old space (Spur only; nil otherwise; 0 implies no limit except that of the underlying platform)
		68	the average number of live stack pages when scanned by GC (at scavenge/gc/become et al) (read-write)
		69	the maximum number of live stack pages when scanned by GC (at scavenge/gc/become et al) (read-write)
		70	the vmProxyMajorVersion (the interpreterProxy VM_MAJOR_VERSION)
		71	the vmProxyMinorVersion (the interpreterProxy VM_MINOR_VERSION)
		72 total milliseconds in full GCs Mark phase since startup (read-only)
		73 total milliseconds in full GCs Sweep phase since startup (read-only, can be 0 depending on compactors)
		74 maximum pause time due to segment allocation
		75 whether the arithmetic primitives perform conversion in case of mixed SmallInteger/Float (true) or fail (false)
		76 the minimum unused headroom in all stack pages; Cog VMs only
		
	Note: Thanks to Ian Piumarta for this primitive."

	| paramsArraySize index |
	paramsArraySize := 76.
	argumentCount = 0 ifTrue: [^self primitiveAllVMParameters: paramsArraySize].
	argumentCount > 2 ifTrue: [^self primitiveFailFor: PrimErrBadNumArgs].
	
	"index read & checks"
	index := self stackValue: (argumentCount = 1 ifTrue: [0] ifFalse: [1]).
	(objectMemory isIntegerObject: index) ifFalse: [^self primitiveFailFor: PrimErrBadArgument].
	index := objectMemory integerValueOf: index.
	(index < 1 or: [index > paramsArraySize]) ifTrue: [^self primitiveFailFor: PrimErrBadIndex].
	
	argumentCount = 1 ifTrue:	 "read VM parameter; written this way to avoid branch limits in V3 bytecode set"
		[| result |
		 result := self primitiveGetVMParameter: index.
		 ^self methodReturnValue: (result ifNil: [objectMemory nilObject])].

	"write a VM parameter"
	self primitiveSetVMParameter: index arg: self stackTop! !

!StackInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 4/2/2015 10:59'!
primitiveVoidReceiver
	"Potentially crash the VM by voiding the receiver.  A subsequent inst var
	 access in the caller's frame should indirect through a null pointer."
	<export: true>
	stackPages longAtPointer: (self frameReceiverLocation: framePointer) put: 0! !

!StackInterpreterPrimitives methodsFor: 'system control primitives' stamp: 'eem 7/25/2021 10:03'!
primitiveVoidVMState
	"Void all internal VM state in the stack and machine code zones"
	| activeContext |
	self push: instructionPointer.
	activeContext := self voidVMStateForSnapshotFlushingExternalPrimitivesIf: false.
	self flushMethodCache.
	self marryContextInNewStackPageAndInitializeInterpreterRegisters: activeContext! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 4/1/2013 11:06'!
pruneStack: stack stackp: stackp
	"Prune the stack to contain only the path, removing stacked indices
	 and mapping frame pointers to contexts  The  issue here is that a
	 GC can occur during ensureFrameIsMarried:SP:, but frame pointers
	 are not valid objects.  So first prune back to objects and framePointers
	 as integers, and then replace frame pointers as integers by contexts."
	
	<inline: false>
	| objOrFP theStack finger |
	<var: #theFP type: #'char *'>
	<var: #thePage type: #'StackPage *'>
	<var: #theFPAbove type: #'char *'>
	finger := 1.
	2 to: stackp - 1 by: 2 do:
		[:i|
		objOrFP := objectMemory fetchPointer: i ofObject: stack.
		(stackPages couldBeFramePointer: (self cCoerceSimple: objOrFP to: #'char *')) ifTrue:
			[objOrFP := self withSmallIntegerTags: objOrFP].
		objectMemory
			storePointerUnchecked: finger
			ofObject: stack
			withValue: objOrFP.
		finger := finger + 1].
	finger to: (objectMemory lengthOf: stack) - 1 do:
		[:i|
		objectMemory
			storePointerUnchecked: i
			ofObject: stack
			withValue: objectMemory nilObject].
	objectMemory pushRemappableOop: (theStack := stack).
	1 to: finger - 1 do:
		[:i| | thePage theFP theFPAbove |
		objOrFP := objectMemory fetchPointer: i ofObject: theStack.
		(self isIntegerObject: objOrFP) ifTrue:
			[theFP := self withoutSmallIntegerTags: objOrFP.
			 thePage := stackPages stackPageFor: theFP.
			 theFPAbove := self findFrameAbove: theFP inPage: thePage.
			 objOrFP := self ensureFrameIsMarried: theFP SP: (self frameCallerSP: theFPAbove).
			 theStack := objectMemory topRemappableOop.
			 objectMemory "after a GC stack may no longer be a root."
				storePointer: finger
				ofObject: theStack
				withValue: objOrFP]].
	objectMemory popRemappableOop! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 3/29/2013 15:48'!
unmarkAfterPathTo
	<inline: false>
	self unmarkAllFrames.
	objectMemory unmarkAllObjects! !

!StackInterpreterPrimitives methodsFor: 'object access primitives' stamp: 'eem 1/4/2017 08:52'!
unmarkAllFrames
	| thePage theFP flags |
	<var: #thePage type: #'StackPage *'>
	<var: #theFP type: #'char *'>
	<inline: false>
	0 to: numStackPages - 1 do:
		[:i|
		thePage := stackPages stackPageAt: i.
		(stackPages isFree: thePage) ifFalse:
			[theFP := thePage  headFP.
			 [flags := objectMemory longAt: theFP + FoxFrameFlags.
			  (flags bitAnd: 2) ~= 0 ifTrue:
				[objectMemory longAt: theFP + FoxFrameFlags put: flags - 2].
			  (theFP := self frameCallerFP: theFP) ~= 0] whileTrue]]! !

StackInterpreterPrimitives subclass: #StackInterpreterSimulator
	instanceVariableNames: 'parent bootstrapping byteCount breakCount sendCount lookupCount printSends printReturns traceOn myBitBlt displayForm fakeForm filesOpen imageName pluginList mappedPluginEntries quitBlock transcript displayView eventTransformer printFrameAtEachStep printBytecodeAtEachStep systemAttributes startMicroseconds lastYieldMicroseconds externalSemaphoreSignalRequests externalSemaphoreSignalResponses extSemTabSize atEachStepBlock disableBooleanCheat performFilters eventQueue assertVEPAES primTraceLog breakBlock inputSemaphoreIndex perMethodProfile'
	classVariableNames: 'ByteCountsPerMicrosecond'
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!StackInterpreterSimulator commentStamp: 'eem 9/3/2013 11:05' prior: 0!
This class defines basic memory access and primitive simulation so that the StackInterpreter can run simulated in the Squeak environment.  It also defines a number of handy object viewing methods to facilitate pawing around in the object memory.

To see the thing actually run, you could (after backing up this image and changes), execute

	(StackInterpreterSimulator new openOn: Smalltalk imageName) test

	((StackInterpreterSimulator newWithOptions: #(NewspeakVM true MULTIPLEBYTECODESETS true))
		openOn: 'ns101.image') test

and be patient both to wait for things to happen, and to accept various things that may go wrong depending on how large or unusual your image may be.  We usually do this with a small and simple benchmark image.

Here's an example of what Eliot uses to launch the simulator in a window.  The bottom-right window has a menu packed with useful stuff:

| vm |
vm := StackInterpreterSimulator newWithOptions: #().
vm openOn: '/Users/eliot/Squeak/Squeak4.4/trunk44.image'.
vm setBreakSelector: #&.
vm openAsMorph; run!

!StackInterpreterSimulator class methodsFor: 'class initialization' stamp: 'eem 5/27/2021 10:12'!
initializeWithOptions: optionsDictionary
	super initializeWithOptions: optionsDictionary.

	ByteCountsPerMicrosecond := InitializationOptions
										at: #ByteCountsPerMicrosecond
										ifAbsent: [50]! !

!StackInterpreterSimulator class methodsFor: 'class initialization' stamp: 'eem 6/7/2018 15:04'!
initializeWithOptions: optionsDictionaryOrArray objectMemoryClass: objectMemoryClassOrNil
	"The relevant ObjectMemory and Interpreter classes must be initialized in order.
	 This happens notionally every time we start the simulator,
	 but in fact happens when ever we instantiate a simulator."
	InitializationOptions := optionsDictionaryOrArray isArray
							ifTrue: [Dictionary newFromPairs: optionsDictionaryOrArray]
							ifFalse: [optionsDictionaryOrArray].

	(objectMemoryClassOrNil ifNil: [self objectMemoryClass])
		initializeWithOptions: InitializationOptions.

	self initializeWithOptions: InitializationOptions! !

!StackInterpreterSimulator class methodsFor: 'instance creation' stamp: 'eem 9/3/2013 11:00'!
new
	^self onObjectMemory: nil options: #()! !

!StackInterpreterSimulator class methodsFor: 'instance creation' stamp: 'eem 9/3/2013 11:01'!
newWithOptions: optionsDictionaryOrArray
	^self onObjectMemory: nil options: optionsDictionaryOrArray! !

!StackInterpreterSimulator class methodsFor: 'instance creation' stamp: 'eem 9/3/2013 11:01'!
onObjectMemory: anObjectMemory
	^self onObjectMemory: anObjectMemory options: #()! !

!StackInterpreterSimulator class methodsFor: 'instance creation' stamp: 'eem 10/9/2013 13:05'!
onObjectMemory: anObjectMemory options: optionsDictionaryOrArray
	| simulatorClass |
	^self == StackInterpreterSimulator
		ifTrue:
			[simulatorClass := SmalltalkImage current endianness == #big
								ifTrue: [self notYetImplemented"StackInterpreterSimulatorMSB"]
								ifFalse: [StackInterpreterSimulatorLSB].
			simulatorClass initializeWithOptions: optionsDictionaryOrArray
				objectMemoryClass: (anObjectMemory ifNotNil: [anObjectMemory class]).
			 simulatorClass
				onObjectMemory: (anObjectMemory ifNil:
										[self objectMemoryClass simulatorClass new])
				options: optionsDictionaryOrArray]
		ifFalse: [super basicNew objectMemory: anObjectMemory; initialize]! !

!StackInterpreterSimulator methodsFor: 'interpreter shell' stamp: 'eem 8/20/2021 15:01'!
aboutToDispatchBytecode
	perMethodProfile ifNotNil:
		[:pmp| pmp at: method put: (pmp at: method ifAbsent: 0) + 1].
	self incrementByteCount.
	self assertValidExecutionPointers.
	atEachStepBlock value "N.B. may be nil"! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'eem 2/4/2014 11:15'!
allObjectsSelect: objBlock
	"self allObjectsSelect: [:oop | (self baseHeader: oop) = 1234]"

	| selected |
	selected := OrderedCollection new.
	objectMemory allObjectsDoSafely:
		[:obj|
		(objBlock value: obj) ifTrue: [selected addLast: obj]].
	^selected! !

!StackInterpreterSimulator methodsFor: 'spur bootstrap' stamp: 'eem 9/6/2013 13:31'!
argumentCount: n
	argumentCount := n! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'eem 7/31/2014 21:38'!
assertValidExecutionPointers
	assertVEPAES ifTrue:
		[super assertValidExecutionPointers]! !

!StackInterpreterSimulator methodsFor: 'accessing' stamp: 'eem 5/28/2014 11:13'!
assertValidExecutionPointersAtEachStep
	^assertVEPAES! !

!StackInterpreterSimulator methodsFor: 'accessing' stamp: 'eem 5/28/2014 11:13'!
assertValidExecutionPointersAtEachStep: aBoolean
	assertVEPAES := aBoolean! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 9/10/2014 19:14'!
atEachStepBlock: aBlock
	atEachStepBlock := aBlock! !

!StackInterpreterSimulator methodsFor: 'return bytecodes' stamp: 'eem 2/27/2013 13:50'!
baseFrameReturn
	| contextToReturnTo |
	contextToReturnTo := self frameCallerContext: localFP.
	((objectMemory isContext: contextToReturnTo)
	 and: [self isMarriedOrWidowedContext: contextToReturnTo]) ifTrue:
		[(self checkIsStillMarriedContext: contextToReturnTo currentFP: nil) ifFalse:
			[self halt]].
	^super baseFrameReturn! !

!StackInterpreterSimulator methodsFor: 'utilities' stamp: 'eem 10/4/2010 15:57'!
booleanCheat: cond
	disableBooleanCheat
		ifTrue: [self internalPop: 2
					thenPush: (cond
								ifTrue: [objectMemory trueObject]
								ifFalse: [objectMemory falseObject]).
				self fetchNextBytecode]
		ifFalse: [super booleanCheat: cond]! !

!StackInterpreterSimulator methodsFor: 'spur bootstrap' stamp: 'eem 11/14/2013 16:20'!
bootstrapping
	^bootstrapping! !

!StackInterpreterSimulator methodsFor: 'spur bootstrap' stamp: 'eem 11/14/2013 15:27'!
bootstrapping: aBoolean
	bootstrapping := aBoolean.
	objectMemory ifNotNil:
		[objectMemory bootstrapping: aBoolean]! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'eem 2/11/2020 10:47'!
breakBlock
	^breakBlock! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'eem 2/11/2020 10:46'!
breakBlock: aBlockOrNil
	breakBlock := aBlockOrNil! !

!StackInterpreterSimulator methodsFor: 'accessing' stamp: 'eem 1/5/2018 15:20'!
breakCount
	^breakCount! !

!StackInterpreterSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:07'!
browserPluginInitialiseIfNeeded
"do nothing - its a macro in C code to support Mac browser plugin strangeness"! !

!StackInterpreterSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:03'!
browserPluginReturnIfNeeded
"do nothing - its a macro in C code to support Mac browser plugin strangeness"! !

!StackInterpreterSimulator methodsFor: 'memory access' stamp: 'eem 10/16/2018 09:09'!
byteAt: byteAddress
	"This is really only for the C library simulations memcpy:_:_: et al in VMClass.
	 Use objectMemory byteAt: directly where possible."
	^objectMemory byteAt: byteAddress! !

!StackInterpreterSimulator methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:44'!
byteAt: byteAddress put: byte
	self deprecated.
	^objectMemory byteAt: byteAddress put: byte! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'di 6/21/2004 15:05'!
byteCount
	"So you can call this from temp debug statements in, eg, Interpreter, such as
	self byteCount = 12661 ifTrue: [self halt].
	"

	^ byteCount! !

!StackInterpreterSimulator methodsFor: 'UI' stamp: 'eem 1/16/2018 15:38'!
byteCountText
	^ byteCount asStringWithCommas asText! !

!StackInterpreterSimulator methodsFor: 'indexing primitives' stamp: 'eem 2/7/2014 11:03'!
bytecodePrimAtPut
	"self halt."
	^super bytecodePrimAtPut! !

!StackInterpreterSimulator methodsFor: 'translation support' stamp: 'eem 6/24/2008 15:38'!
cCoerceSimple: value to: cTypeString
	"Type coercion for translation only; just return the value when running in Smalltalk."

	^value! !

!StackInterpreterSimulator methodsFor: 'plugin support' stamp: 'eem 10/30/2014 12:50'!
callExternalPrimitive: mapIndex
	| entry |
	entry := self pluginEntryFor: mapIndex.
	"Spur needs the primitiveFunctionPointer to be set correctly
	 for accurate following of forwarders on primitive failure."
	objectMemory hasSpurMemoryManagerAPI ifTrue:
		[primitiveFunctionPointer := entry at: 2].
	^(entry at: 1) perform: (entry at: 2)! !

!StackInterpreterSimulator methodsFor: 'object memory support' stamp: 'eem 9/20/2013 13:55'!
checkStackIntegrity
	"Override to deal with incomplete initialization."
	stackPages ifNil: [^true].
	^super checkStackIntegrity! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'eem 07/28/2014 07:54'!
classAndSelectorOfMethod: meth forReceiver: rcvr
	| mClass dict length methodArray |
	mClass := objectMemory fetchClassOf: rcvr.
	[dict := objectMemory fetchPointer: MethodDictionaryIndex ofObject: mClass.
	length := objectMemory numSlotsOf: dict.
	methodArray := objectMemory fetchPointer: MethodArrayIndex ofObject: dict.
	0 to: length-SelectorStart-1 do: 
		[:index | 
		meth = (objectMemory fetchPointer: index ofObject: methodArray) 
			ifTrue: [^ Array
				with: mClass
				with: (objectMemory fetchPointer: index + SelectorStart ofObject: dict)]].
	mClass := self superclassOf: mClass.
	mClass = objectMemory nilObject]
		whileFalse.
	^ Array
		with: (objectMemory fetchClassOf: rcvr)
		with: (objectMemory splObj: SelectorDoesNotUnderstand)! !

!StackInterpreterSimulator methodsFor: 'plugin support' stamp: 'eem 8/2/2021 21:55'!
classNameOf: aClass Is: className
	"Check if aClass' name is className"
	| name |
	(objectMemory numSlotsOf: aClass) <= classNameIndex ifTrue:
		[^false]. "Not a class but maybe behavior" 
	name := objectMemory fetchPointer: classNameIndex ofObject: aClass.
	(objectMemory isBytes: name) ifFalse:
		[^false].
	^ className = (self stringOf: name)! !

!StackInterpreterSimulator methodsFor: 'I/O primitives' stamp: 'eem 1/4/2017 08:41'!
clipboardRead: sz Into: actualAddress At: zeroBaseIndex
	| str |
	str := Clipboard clipboardText.
	1 to: sz do:
		[:i | objectMemory byteAt: actualAddress + zeroBaseIndex + i - 1 put: (str at: i) asciiValue]! !

!StackInterpreterSimulator methodsFor: 'I/O primitives' stamp: 'di 9/23/2001 12:18'!
clipboardSize

	^ Clipboard clipboardText size! !

!StackInterpreterSimulator methodsFor: 'I/O primitives' stamp: 'eem (const to msg rewrite) 11/5/2014 18:04'!
clipboardWrite: sz From: actualDataAddress At: ignored

	Clipboard clipboardText: (self stringOf: actualDataAddress - objectMemory baseHeaderSize)! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'eem 6/21/2021 15:11'!
cloneSimulation
	| savedAtEachStepBlock savedDisplayView savedDisplayForm savedQuitBlock savedTranscript |
	savedAtEachStepBlock := atEachStepBlock. atEachStepBlock := nil.
	savedDisplayView := displayView. displayView := nil.
	savedDisplayForm := displayForm. displayForm = nil.
	savedQuitBlock := quitBlock. quitBlock := nil.
	savedTranscript := transcript. transcript := nil.

	^[| theClone |
	 self changed: #byteCountText.
	 Smalltalk garbageCollect.
	 theClone := self veryDeepCopy.
	 theClone parent: self; transcript: Transcript.
	 theClone objectMemory parent: objectMemory.
	 theClone]
		ensure:
			[atEachStepBlock := savedAtEachStepBlock.
			 displayView := savedDisplayView.
			 displayForm = savedDisplayForm.
			 quitBlock := savedQuitBlock.
			 transcript := savedTranscript]! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'eem 5/25/2014 20:06'!
cloneSimulationWindow
	| label |
	label := 'Clone of ', (displayView containingWindow label allButFirst: 'Simulation of ' size).
	^self cloneSimulation openAsMorph
		setLabel: label;
		yourself! !

!StackInterpreterSimulator methodsFor: 'initialize-release' stamp: 'eem 7/26/2020 12:14'!
close  "close any files that ST may have opened, etc"
	pluginList do: [:assoc| | plugin | plugin := assoc value. plugin ~~ self ifTrue: [plugin close]].
	"Ugh; at least some of this code belongs in the UI..."
	displayView ifNotNil:
		[displayView activeHand removeEventListener: self].
	ActiveHand removeEventListener: self.
	World submorphs do:
		[:submorph|
		(submorph model isVMObjectInspector
		 and: [submorph model coInterpreter == self]) ifTrue:
			[submorph delete].
		(submorph model isDebugger
		 and: [(submorph model interruptedProcess suspendedContext ifNotNil:
				[:sctxt|
				 sctxt findContextSuchThat:
					[:ctxt|
					 ctxt receiver == self
					 and: [ctxt selector == #run]]]) notNil]) ifTrue:
			[submorph model windowIsClosing.
			 submorph delete]]! !

!StackInterpreterSimulator methodsFor: 'simulation only' stamp: 'eem 1/22/2016 09:16'!
cogit
	"We don't have a cogit; try and get by on our own devices."
	^self! !

!StackInterpreterSimulator methodsFor: 'I/O primitives support' stamp: ''!
copyBits

	^ myBitBlt copyBits! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 6/20/2008 15:46'!
cr

	traceOn ifTrue: [ transcript cr; flush ].! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 8/8/2014 11:32'!
createActualMessageTo: class
	objectMemory bootstrapping ifTrue:
		[self halt].
	"false
		ifTrue:
			[(self stringOf: messageSelector) = 'run:with:in:' ifTrue:
				[self halt]]
		ifFalse:
			[self halt: (self stringOf: messageSelector)]."

	^super createActualMessageTo: class! !

!StackInterpreterSimulator methodsFor: 'UI' stamp: 'eem 5/11/2011 16:58'!
desiredDisplayExtent
	^(savedWindowSize
		ifNil: [640@480]
		ifNotNil: [savedWindowSize >> 16 @ (savedWindowSize bitAnd: 16rFFFF)])
			min: Display extent * 2 // 3! !

!StackInterpreterSimulator methodsFor: 'initialize-release' stamp: 'eem 3/10/2009 08:59'!
desiredEdenBytes: anInteger
	desiredEdenBytes := anInteger! !

!StackInterpreterSimulator methodsFor: 'initialize-release' stamp: 'eem 3/10/2009 08:58'!
desiredNumStackPages: anInteger
	desiredNumStackPages := anInteger! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 10/4/2010 15:54'!
disableBooleanCheat: aBoolean
	disableBooleanCheat := aBoolean! !

!StackInterpreterSimulator methodsFor: 'plugin primitive support' stamp: 'eem 10/8/2013 16:44'!
dispatchMappedPluginEntry: n 
	^(mappedPluginEntries at: n) first
		perform: (mappedPluginEntries at: n) second! !

!StackInterpreterSimulator methodsFor: 'interpreter shell' stamp: 'eem 11/12/2012 22:37'!
dispatchOn: anInteger in: selectorArray
	"Simulate a case statement via selector table lookup.
	The given integer must be between 0 and selectorArray size-1, inclusive.
	For speed, no range test is done, since it is done by the at: operation.
	Note that, unlike many other arrays used in the Interpreter, this method expect NO CArrayAccessor wrapping - it would duplicate the +1. Maybe this would be better updated to make it all uniform"
	selectorArray == BytecodeTable ifTrue:
		[self assert: (anInteger between: bytecodeSetSelector and: bytecodeSetSelector + 255)].
	self perform: (selectorArray at: (anInteger + 1)).! !

!StackInterpreterSimulator methodsFor: 'accessing' stamp: 'tty 2/5/2014 12:44'!
displayForm
	"The VM SImulator draws directly on an imageMorph named displayView
	  displayView is housed on a Form named displayForm
	  displayForm is initialized in my initialize method. 
	  it 'needs' to be initialized there to support Eliot's legacy openAsMorph functionality
	"
	^displayForm! !

!StackInterpreterSimulator methodsFor: 'I/O primitives support' stamp: ''!
displayLocation

	^ Display extent - displayForm extent - (10@10)! !

!StackInterpreterSimulator methodsFor: 'accessing' stamp: 'tty 2/5/2014 12:45'!
displayView
	"The VM SImulator draws directly on an imageMorph named displayView
	  displayView is housed on a Form named displayForm
	  Since this is housed directly on the Morph, I have this tight compling from here->model->morph. 
	"
	self flag: 'tty'.	
	^displayView! !

!StackInterpreterSimulator methodsFor: 'accessing' stamp: 'tty 2/5/2014 12:45'!
displayView: anImageMorph
	"The VM SImulator draws directly on an imageMorph named displayView
	  displayView is housed on a Form named displayForm
	  Since this is housed directly on the Morph, I have this tight compling from here->model->morph. 
	"
	self flag: 'tty'.
	displayView := anImageMorph! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 6/28/2008 17:06'!
divorceFramesIn: aStackPage
	"| thisPage |
	(self checkIsStillMarriedContext: 22189568 currentFP: framePointer) ifTrue:
		[thisPage := stackPages stackPageFor: (self frameOfMarriedContext: 22189568).
		 aStackPage == thisPage ifTrue:
			[self halt]]."
	^super divorceFramesIn: aStackPage! !

!StackInterpreterSimulator methodsFor: 'process primitive support' stamp: 'eem 12/29/2016 15:19'!
doSignalExternalSemaphores: minTableSize
	"This is a non-thread-safe simulation.  See platforms/Cross/vm/sqExternalSemaphores.c
	 for the real code.  For the benefit of the SocketPluginSimulator, do a yield every 100
	 virtual microseconds."
	| now switched |
	now := self ioUTCMicroseconds.
	now - lastYieldMicroseconds >= 100 ifTrue:
		[lastYieldMicroseconds := now.
		 Processor yield].

	switched := false.
	1 to: (minTableSize min: externalSemaphoreSignalRequests size) do:
		[:i| | responses |
		responses := externalSemaphoreSignalResponses at: i.
		[responses < (externalSemaphoreSignalRequests at: i)] whileTrue:
			[(self doSignalSemaphoreWithIndex: i) ifTrue:
				[switched := true].
			 externalSemaphoreSignalResponses at: i put: (responses := responses + 1)]].
	^switched! !

!StackInterpreterSimulator methodsFor: 'error handling' stamp: 'eem 12/15/2014 10:06'!
doesNotUnderstand: aMessage
	"If this is a doit and the objectMemory understands, pass it on."
	(thisContext findContextSuchThat: [:ctxt| ctxt selector == #evaluate:in:to:notifying:ifFail:logged:]) ifNotNil:
		[(objectMemory class whichClassIncludesSelector: aMessage selector) ifNotNil:
			[:implementingClass|
			(implementingClass inheritsFrom: Object) ifTrue: "i.e. VMClass and below"
				[thisContext sender selector ~~ #DoIt ifTrue:
					[Transcript nextPutAll: 'warning: redirecting ', aMessage selector, ' in ', thisContext sender printString, ' to objectMemory'; cr; flush].
				 aMessage lookupClass: nil.
				^aMessage sentTo: objectMemory]]].
	^super doesNotUnderstand: aMessage! !

!StackInterpreterSimulator methodsFor: 'I/O primitives support' stamp: ''!
drawLoopX: xDelta Y: yDelta

	^ myBitBlt drawLoopX: xDelta Y: yDelta! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: ''!
dumpMethodHeader: hdr
	^ String streamContents:
		[:strm |
		strm nextPutAll: '<nArgs=', ((hdr >> 25) bitAnd: 16r1F) printString , '>'.
		strm nextPutAll: '<nTemps=', ((hdr >> 19) bitAnd: 16r3F) printString , '>'.
		strm nextPutAll: '<lgCtxt=', ((hdr >> 18) bitAnd: 16r1) printString , '>'.
		strm nextPutAll: '<nLits=', ((hdr >> 10) bitAnd: 16rFF) printString , '>'.
		strm nextPutAll: '<prim=', ((hdr >> 1) bitAnd: 16r1FF) printString , '>'.
		]! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 6/20/2008 15:25'!
elementsPerPrintOopLine
	^4! !

!StackInterpreterSimulator methodsFor: 'simulation only' stamp: 'eem 8/20/2021 15:54'!
enablePerMethodProfiling
	perMethodProfile := Dictionary new! !

!StackInterpreterSimulator methodsFor: 'compiled methods' stamp: 'eem 2/11/2020 10:14'!
endPCOf: aMethod
	"Determine the endPC of a method in the heap using interpretation that looks for returns and uses branches to skip intervening bytecodes."
	| pc end farthestContinuation prim encoderClass inst is |
	(prim := self primitiveIndexOf: aMethod) > 0 ifTrue:
		[(self isQuickPrimitiveIndex: prim) ifTrue:
			[^(self startPCOfMethod: aMethod) - 1]].
	encoderClass := self encoderClassForHeader: (objectMemory methodHeaderOf: aMethod).
	is := (InstructionStream
			on: (VMCompiledMethodProxy new
					for: aMethod
					coInterpreter: self
					objectMemory: objectMemory)).
	pc := farthestContinuation := self startPCOfMethod: aMethod.
	end := objectMemory numBytesOf: aMethod.
	is pc: pc + 1.
	[pc <= end] whileTrue:
		[inst := encoderClass interpretNextInstructionFor: MessageCatcher new in: is.
		 pc := is pc - 1.
		 inst selector
			caseOf: {
				 [#pushClosureCopyNumCopiedValues:numArgs:blockSize:]	
											->	[is pc: is pc + inst arguments last.
												 farthestContinuation := farthestContinuation max: pc].
				 [#jump:]					->	[farthestContinuation := farthestContinuation max: pc + inst arguments first].
				 [#jump:if:]					->	[farthestContinuation := farthestContinuation max: pc + inst arguments first].
				 [#methodReturnConstant:]	->	[pc > farthestContinuation ifTrue: [end := pc]].
				 [#methodReturnReceiver]	->	[pc > farthestContinuation ifTrue: [end := pc]].
				 [#methodReturnTop]		->	[pc > farthestContinuation ifTrue: [end := pc]].
				"This is for CompiledBlock/FullBlockClosure.  Since the response to pushClosure... above
				 skips over all block bytecoes, we will only see a blockReturnTop if it is at the top level,
				 and so it must be a blockReturnTop in a CompiledBlock for a FullBlockClosure."
				 [#blockReturnTop]			->	[pc > farthestContinuation ifTrue: [end := pc]].
				 [#branchIfInstanceOf:distance:]
											->	[farthestContinuation := farthestContinuation max: pc + inst arguments last].
				 [#branchIfNotInstanceOf:distance:]
											->	[farthestContinuation := farthestContinuation max: pc + inst arguments last] }
			otherwise: []].
	^end - 1! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 2/11/2020 10:45'!
ensureDebugAtEachStepBlock
	atEachStepBlock := [printFrameAtEachStep ifTrue:
							[self printFrame: localFP WithSP: localSP].
						 printBytecodeAtEachStep ifTrue:
							[self printCurrentBytecodeOn: transcript.
							 transcript cr; flush].
						 byteCount = breakCount ifTrue:
							["printFrameAtEachStep :=" printBytecodeAtEachStep := true].
						 breakBlock ifNotNil:
							[breakBlock value ifTrue: [self halt]]]! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 6/28/2008 17:06'!
externalDivorceFrame: theFP andContext: ctxt
	"(theFP = -208 or: [ctxt = 22189568]) ifTrue:
		[self halt]."
	^super externalDivorceFrame: theFP andContext: ctxt! !

!StackInterpreterSimulator methodsFor: 'frame access' stamp: 'eem 2/27/2013 13:50'!
externalInstVar: index ofContext: aMarriedContext put: anOop
	| imMarried shesMarried result |
	index == SenderIndex ifTrue:
		[imMarried := self checkIsStillMarriedContext: aMarriedContext currentFP: localFP.
		(objectMemory isContext: anOop) ifTrue:
			[shesMarried := self checkIsStillMarriedContext: anOop currentFP: localFP.
			 "self shortPrintContext: aMarriedContext.
			  self shortPrintContext: anOop"]].
	result := super externalInstVar: index ofContext: aMarriedContext put: anOop.
	imMarried ifNotNil:
		[self assert: imMarried == (self checkIsStillMarriedContext: aMarriedContext currentFP: nil).
		 shesMarried ifNotNil:
			[self assert: shesMarried == (self checkIsStillMarriedContext: anOop currentFP: nil)]].
	^result! !

!StackInterpreterSimulator methodsFor: 'interpreter shell' stamp: 'eem 1/3/2017 10:25'!
fetchByte
	^objectMemory byteAt: (localIP := localIP + 1).! !

!StackInterpreterSimulator methodsFor: 'control primitives' stamp: 'eem 1/11/2013 18:38'!
filterPerformOf: selector to: receiver
	"This is to allow simulator to filter start-up items to avoid as-yet unsimulatable plugins."
	performFilters ifNil: [^false].
	(performFilters at: (self shortPrint: receiver) ifAbsent: []) ifNotNil:
		[:messages|
		^messages includes: (self stringOf: selector)].
	^false! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 12/8/2016 11:25'!
findNewMethodInClassTag: classTag
"
	| cName |
	traceOn ifTrue:
		[cName := (self sizeBitsOf: class) = 16r20
			ifTrue: ['class ' , (self nameOfClass: (self fetchPointer: 6 ofObject: class))]
			ifFalse: [(self nameOfClass: class)].
		self cr; print: cName , '>>' , (self stringOf: messageSelector)].
"
	messageSelector = (objectMemory splObj: SelectorDoesNotUnderstand) ifTrue: [self halt].

	sendCount := sendCount + 1.

	(printSends or: [printBytecodeAtEachStep]) ifTrue:
		[self cr; print: byteCount; space; printStringOf: messageSelector; cr].
"
	(sendCount > 1000 and: [sendCount\\10 = 0]) ifTrue:
		[Transcript print: sendCount; space.
		self validate].
"
"
	(sendCount > 100150) ifTrue:
		[self qvalidate.
		messageQueue == nil ifTrue: [messageQueue := OrderedCollection new].
		messageQueue addLast: (self stringOf: messageSelector)].
"
	^super findNewMethodInClassTag: classTag! !

!StackInterpreterSimulator methodsFor: 'memory access' stamp: 'eem 6/11/2010 11:36'!
firstIndexableField: oop
	"This is in ObjectMemory and overridden in the obj mem simulators"
	self shouldNotImplement! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 1/3/2021 17:33'!
flush
	traceOn ifTrue:
		[transcript flush.
		 "We *always* want to make output visible on flush"
		 TranscriptStream forceUpdate ifFalse:
			[transcript changed: #appendEntry]]! !

!StackInterpreterSimulator methodsFor: 'control primitives' stamp: 'eem 1/11/2013 18:33'!
forShortPrintString: shortPrintString filterPerformMessages: aCollection
	performFilters ifNil:
		[performFilters := Dictionary new].
	performFilters at: shortPrintString put: aCollection! !

!StackInterpreterSimulator methodsFor: 'frame access' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
frameOfMarriedContext: aContext 
	| senderOop |
	senderOop := objectMemory fetchPointer: SenderIndex ofObject: aContext.
	self assert: (objectMemory isIntegerObject: senderOop).
	^self withoutSmallIntegerTags: senderOop! !

!StackInterpreterSimulator methodsFor: 'spur bootstrap' stamp: 'eem 9/18/2013 11:32'!
framePointer
	^framePointer! !

!StackInterpreterSimulator methodsFor: 'I/O primitives' stamp: 'eem 5/12/2017 11:07'!
fullDisplay
	| primFailCodeValue |
	displayForm ifNil: [^self].
	primFailCodeValue := primFailCode.
	self initPrimCall.
	self updateDisplayLeft: 0 Top: 0 Right: displayWidth Bottom: displayHeight.
	primFailCode := primFailCodeValue! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'eem 11/25/2008 13:32'!
fullDisplayUpdate
	"Preserve self successful when call asynchronously from Simulator"
	| primFailCodeValue |
	primFailCodeValue := primFailCode.
	self initPrimCall.
	super fullDisplayUpdate.
	primFailCode := primFailCodeValue! !

!StackInterpreterSimulator methodsFor: 'interpreter shell' stamp: 'eem 6/24/2008 16:00'!
functionPointerFor: primIndex inClass: lookupClass
	"Override Interpreter to handle the external primitives caching.  See also
	 internalExecuteNewMethod."

	^(primIndex between: 1 and: MaxPrimitiveIndex)
		ifTrue: [primitiveTable at: primIndex + 1]
		ifFalse: [0]! !

!StackInterpreterSimulator methodsFor: 'image save/restore' stamp: 'eem 10/10/2013 06:23'!
getShortFromFile: aFile swap: swapFlag
	| aShort |
	aShort := self nextShortFrom: aFile.
	^swapFlag 
		ifTrue: [(aShort bitShift: -8) + ((aShort bitAnd: 16rFF) bitShift: 8)]
		ifFalse: [aShort]! !

!StackInterpreterSimulator methodsFor: 'memory access' stamp: 'di 7/16/2004 14:56'!
halfWordHighInLong32: long32
	^self subclassResponsibility! !

!StackInterpreterSimulator methodsFor: 'memory access' stamp: 'di 7/16/2004 14:57'!
halfWordLowInLong32: long32
	^self subclassResponsibility! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
headerStart: oop

	^ (objectMemory extraHeaderBytes: oop) negated! !

!StackInterpreterSimulator methodsFor: 'image save/restore' stamp: 'eem 9/19/2013 15:46'!
imageName
	^imageName! !

!StackInterpreterSimulator methodsFor: 'spur bootstrap' stamp: 'eem 9/19/2013 15:47'!
imageName: aString
	imageName := aString! !

!StackInterpreterSimulator methodsFor: 'file primitives' stamp: 'eem 10/8/2013 16:37'!
imageNameGet: p Length: sz
	1 to: sz  do:
		[:i |
		objectMemory
			byteAt:  p + i - 1
			put: (imageName at: i) asInteger]! !

!StackInterpreterSimulator methodsFor: 'file primitives' stamp: 'eem 9/7/2018 18:48'!
imageNamePut: p Length: sz
	| newName window |
	newName := ByteString new: sz.
	1 to: sz  do:
		[:i |
		newName
			at: i
			put: (Character value: (objectMemory byteAt: p + i - 1))].
	imageName := newName.
	(displayView notNil
	 and: [(window := displayView containingWindow) notNil]) ifTrue:
		[window setLabel: (window label
							copyReplaceFrom: (window label lastIndexOf: Character space) + 1
							to: window label size
							with: (self localNameFor: imageName))]! !

!StackInterpreterSimulator methodsFor: 'file primitives' stamp: 'eem 10/8/2013 16:36'!
imageNameSize
	^imageName size! !

!StackInterpreterSimulator methodsFor: 'interpreter shell' stamp: 'eem 1/5/2018 15:22'!
incrementByteCount
	(byteCount := byteCount + 1) = breakCount ifTrue:
		[self doOrDefer: [self changed: #byteCountText; changed: #composeAll].
		 self halt: 'breakCount reached'].
	byteCount \\ 1000 = 0 ifTrue:
		[self doOrDefer: [self changed: #byteCountText; changed: #composeAll].
		 self forceInterruptCheck]! !

!StackInterpreterSimulator methodsFor: 'initialize-release' stamp: 'eem 10/23/2020 11:57'!
initialize
	"Initialize the StackInterpreterSimulator when running the interpreter
	 inside Smalltalk. The primary responsibility of this method is to allocate
	 Smalltalk Arrays for variables that will be declared as statically-allocated
	 global arrays in the translated code."
	super initialize.

	bootstrapping := false.
	transcript := Transcript.

	objectMemory ifNil:
		[objectMemory := self class objectMemoryClass simulatorClass new].
	objectMemory coInterpreter: self.

	self assert: ConstMinusOne = (objectMemory integerObjectOf: -1).

	methodCache := Array new: MethodCacheSize.
	nsMethodCache := Array new: NSMethodCacheSize.
	atCache := Array new: AtCacheTotalSize.
	self flushMethodCache.
	gcSemaphoreIndex := 0.
	externalSemaphoreSignalRequests := externalSemaphoreSignalResponses := #().
	externalPrimitiveTable := CArrayAccessor on: (Array new: MaxExternalPrimitiveTableSize).
	externalPrimitiveTableFirstFreeIndex := 0.
	primitiveTable := self class primitiveTable copy.
	self initializePluginEntries.
	desiredNumStackPages := desiredEdenBytes := 0.
	"This is initialized on loading the image, but convenient for testing stack page values..."
	numStackPages := self defaultNumStackPages. 
	startMicroseconds := lastYieldMicroseconds := self ioUTCStartMicroseconds.

	"initialize InterpreterSimulator variables used for debugging"
	byteCount := sendCount := lookupCount := 0.
	quitBlock := [^self close].
	traceOn := true.
	printSends := printReturns := printFrameAtEachStep := printBytecodeAtEachStep := false.
	myBitBlt := BitBltSimulator new setInterpreter: self.
	displayForm := fakeForm := 'Display has not yet been installed' asDisplayText form.
	suppressHeartbeatFlag := false.
	systemAttributes := Dictionary new.
	extSemTabSize := 256.
	disableBooleanCheat := false.
	assertVEPAES := false. "a flag so the assertValidExecutionPointers can be disabled for simulation speed and enabled when necessary."! !

!StackInterpreterSimulator methodsFor: 'plugin support' stamp: 'eem 7/25/2021 16:11'!
initializePluginEntries
	mappedPluginEntries := OrderedCollection new.
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[primitiveAccessorDepthTable := Array new: primitiveTable size.
			 pluginList := {}.
			 self loadNewPlugin: '']
		ifFalse:
			[pluginList := {'' -> self }]! !

!StackInterpreterSimulator methodsFor: 'I/O primitive support' stamp: 'eem 10/23/2020 14:30'!
inputSemaphoreIndex
	^inputSemaphoreIndex! !

!StackInterpreterSimulator methodsFor: 'frame access' stamp: 'eem 2/27/2013 13:50'!
instVar: index ofContext: aMarriedContext put: anOop
	| imMarried shesMarried result |
	index == SenderIndex ifTrue:
		[imMarried := self checkIsStillMarriedContext: aMarriedContext currentFP: localFP.
		(objectMemory isContext: anOop) ifTrue:
			[shesMarried := self checkIsStillMarriedContext: anOop currentFP: localFP.
			"self cr.
			self shortPrintContext: aMarriedContext.
			self shortPrintContext: anOop.
			(#('yield:' 'nextPut:') includesAnyOf: {self stringOf: (self selectorOfContext: aMarriedContext). self stringOf: (self selectorOfContext: anOop)}) ifTrue:
				[self halt]"]].
	result := super instVar: index ofContext: aMarriedContext put: anOop.
	imMarried ifNotNil:
		[self assert: imMarried == (self checkIsStillMarriedContext: aMarriedContext currentFP: nil).
		 shesMarried ifNotNil:
			[self assert: shesMarried == (self checkIsStillMarriedContext: anOop currentFP: nil)]].
	^result! !

!StackInterpreterSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:01'!
insufficientMemoryAvailableError
	self error: 'Failed to allocate memory for the heap'! !

!StackInterpreterSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:01'!
insufficientMemorySpecifiedError
	self error: 'Insufficient memory for this image'! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 6/27/2008 14:13'!
internalCannotReturn: resultOop
	self halt.
	^super internalCannotReturn: resultOop! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 10/16/2018 14:49'!
internalFindNewMethodOrdinary
"
	| cName |
	traceOn ifTrue:
		[cName := (self sizeBitsOf: class) = 16r20
			ifTrue: ['class ' , (self nameOfClass: (self fetchPointer: 6 ofObject: class))]
			ifFalse: [(self nameOfClass: class)].
		self cr; print: cName , '>>' , (self stringOf: messageSelector)].
"
	messageSelector = (objectMemory splObj: SelectorDoesNotUnderstand) ifTrue: [self halt].

	sendCount := sendCount + 1.

"	printSends ifTrue:
		[self print: byteCount; space; printStringOf: messageSelector; cr]."
"
	(sendCount > 1000 and: [sendCount\\10 = 0]) ifTrue:
		[Transcript print: sendCount; space.
		self validate].
"
"
	(sendCount > 100150) ifTrue:
		[self qvalidate.
		messageQueue == nil ifTrue: [messageQueue := OrderedCollection new].
		messageQueue addLast: (self stringOf: messageSelector)].
"
	super internalFindNewMethodOrdinary! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 6/27/2008 14:13'!
internalMustBeBoolean
	self halt.
	^super internalMustBeBoolean! !

!StackInterpreterSimulator methodsFor: 'stack pages' stamp: 'eem 11/14/2013 15:01'!
interpreterAllocationReserveBytes
	^bootstrapping
		ifTrue: [0]
		ifFalse: [super interpreterAllocationReserveBytes]! !

!StackInterpreterSimulator methodsFor: 'I/O primitives' stamp: 'eem 5/12/2017 17:24'!
ioBeDisplay: bitsOrHandleCArray width: width height: height depth: depth
	"The real platform may want to note the location of the display and respond to the fact of
	 its installation.  Here merely create a scratch form the same size as Smalltalk displayObj"
	super ioBeDisplay: bitsOrHandleCArray asInteger width: width height: height depth: depth.
	displayForm := Form extent: width @ height depth: depth.
	displayView ifNotNil: [displayView image: displayForm]! !

!StackInterpreterSimulator methodsFor: 'security' stamp: 'ar 2/5/2001 18:33'!
ioCanRenameImage
	^true! !

!StackInterpreterSimulator methodsFor: 'security' stamp: 'ar 2/5/2001 20:42'!
ioCanWriteImage
	^true! !

!StackInterpreterSimulator methodsFor: 'primitive support' stamp: 'eem 11/10/2020 11:59'!
ioExit
	self ioExitWithErrorCode: 0! !

!StackInterpreterSimulator methodsFor: 'primitive support' stamp: 'eem 11/10/2020 12:24'!
ioExitWithErrorCode: ec
	self ioShutdownAllModules.
	quitBlock value  "Cause return from #test"! !

!StackInterpreterSimulator methodsFor: 'process primitive support' stamp: 'eem 2/3/2010 12:56'!
ioGetMaxExtSemTableSize
	^extSemTabSize! !

!StackInterpreterSimulator methodsFor: 'I/O primitives' stamp: 'ar 5/1/1999 09:43'!
ioHasDisplayDepth: depth
	^Display supportsDisplayDepth: depth! !

!StackInterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'eem 8/22/2008 11:18'!
ioHeartbeatMilliseconds
	^1! !

!StackInterpreterSimulator methodsFor: 'initialize-release' stamp: 'eem 8/3/2008 21:48'!
ioInitHeartbeat
	"No-op in the simulator.  We cause a poll every 1000 bytecodes instead."! !

!StackInterpreterSimulator methodsFor: 'simulation support' stamp: 'eem 7/25/2021 15:39'!
ioListBuiltinModule: index
	^index > 0 ifTrue:
		[(InitializationOptions
			at: #builtinModules
			ifAbsent: [#('BitBltPlugin' 'FloatArrayPlugin' 'FloatMathPlugin' 'LargeIntegers' 'MiscPrimitivePlugin')]) at: index ifAbsent: nil]! !

!StackInterpreterSimulator methodsFor: 'simulation support' stamp: 'eem 7/25/2021 15:21'!
ioListLoadedModule: index
	^index > 0 ifTrue: [(pluginList at: index + 1 ifAbsent: [^nil]) key]! !

!StackInterpreterSimulator methodsFor: 'plugin support' stamp: 'eem 10/29/2018 15:40'!
ioLoadFunction: functionString From: pluginString
	"Load and return the requested function from a module"
	| firstTime plugin fnSymbol |
	firstTime := false.
	fnSymbol := functionString asSymbol.
	transcript
		cr;
		show: '(', byteCount printString, ') Looking for ', functionString, ' in ',
				(pluginString isEmpty ifTrue: ['vm'] ifFalse: [pluginString]).
	(breakSelector notNil
	 and: [(pluginString size = breakSelector size
			and: [(self strncmp: pluginString _: breakSelector _: pluginString size) = 0])
		or: [functionString size = breakSelector size
			and: [(self strncmp: functionString _: breakSelector _: functionString size) = 0]]]) ifTrue:
		[self halt: pluginString, '>>', functionString].
	plugin := pluginList 
				detect: [:any| any key = pluginString asString]
				ifNone:
					[firstTime := true.
					 self loadNewPlugin: pluginString].
	plugin ifNil:
		[firstTime ifTrue: [transcript cr; show: '... FAILED; no plugin found'].
		 ^0].
	plugin := plugin value.
	mappedPluginEntries withIndexDo:
		[:pluginAndName :index|
		((pluginAndName at: 1) == plugin 
		and:[(pluginAndName at: 2) == fnSymbol]) ifTrue:
			[^index]].
	(plugin respondsTo: fnSymbol) ifFalse:
		[firstTime ifTrue: [transcript cr; show: '... FAILED; primitive not in plugin'].
		 ^0].
	mappedPluginEntries addLast: { plugin. fnSymbol }.
	transcript show: ' ... okay'.
	^mappedPluginEntries size! !

!StackInterpreterSimulator methodsFor: 'plugin support' stamp: 'eem 7/22/2021 17:43'!
ioLoadFunction: functionString From: pluginString MetadataInto: metadataPtr
	"Load and return the requested function from a module.
	 Assign the accessor depth and flags through metadataPtr.
	 N.B. The actual code lives in platforms/Cross/vm/sqNamedPrims.h"
	| firstTime plugin fnSymbol |
	firstTime := false.
	fnSymbol := functionString asSymbol.
	transcript
		cr;
		show: '(', byteCount printString, ') Looking for ', functionString, ' in ',
				(pluginString isEmpty ifTrue: ['vm'] ifFalse: [pluginString]).
	(breakSelector notNil
	 and: [(pluginString size = breakSelector size
			and: [(self strncmp: pluginString _: breakSelector _: pluginString size) = 0])
		or: [functionString size = breakSelector size
			and: [(self strncmp: functionString _: breakSelector _: functionString size) = 0]]]) ifTrue:
		[self halt: pluginString, '>>', functionString].
	plugin := pluginList 
				detect: [:any| any key = pluginString asString]
				ifNone:
					[firstTime := true.
					 self loadNewPlugin: pluginString].
	plugin ifNil:
		[firstTime ifTrue: [transcript show: '... FAILED; no plugin found'].
		 ^0].
	plugin := plugin value.
	mappedPluginEntries withIndexDo:
		[:pluginAndName :index|
		 ((pluginAndName at: 1) == plugin 
		  and:[(pluginAndName at: 2) == fnSymbol]) ifTrue:
			[firstTime ifTrue: [transcript show: ' ... okay'].
			 metadataPtr at: 0 put: (pluginAndName at: 4).
			 ^index]].
	firstTime ifTrue: [transcript show: '... FAILED; primitive not in plugin'].
	^0! !

!StackInterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'eem 12/10/2009 14:53'!
ioMSecs
	"Return the value of the millisecond clock."
	"NOT.  Actually, we want something a lot slower and, for exact debugging,
	something more repeatable than real time.  I have an idea: use the byteCount... (di 7/1/2004 13:55)"

	^self microsecondsToMilliseconds: self ioUTCMicroseconds
	
"At 20k bytecodes per second, this gives us about 200 ticks per second, or about 1/5 of what you'd expect for the real time clock.  This should still service events at one or two per second"! !

!StackInterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'jm 3/9/1999 10:00'!
ioMicroMSecs
	"Answer the value of the high-resolution millisecond clock."

	^ Time millisecondClockValue
! !

!StackInterpreterSimulator methodsFor: 'I/O primitives' stamp: 'eem 3/4/2015 17:05'!
ioMousePoint
	| relPt |
	^displayForm
		ifNil: [99 << 16 + 66]
		ifNotNil: [relPt := Sensor cursorPoint - self displayLocation.
				relPt x << 16 + relPt y]! !

!StackInterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'eem 5/27/2021 10:20'!
ioRelinquishProcessorForMicroseconds: microseconds
	"In the simulator give an indication that we're idling and check for input.
	 If there's an input event, fail to give the system a chance to respond to it in a timely manner."
	Display reverse: ((displayView
						ifNil: [0@0]
						ifNotNil: [displayView bounds origin]) extent: 16@16).
	Processor activeProcess == Project uiProcess ifTrue:
		[World doOneCycle].
	(stackLimit = self allOnesAsCharStar
	 and: [nextPollUsecs <= self ioUTCMicroseconds])
		ifTrue:
			["Not only do we need to fail, we also need to push time faster
			 since the damn Morphic eventTickler process waits on a delay."
			"ioUTCMicroseconds"
			byteCount := byteCount + (1000 * ByteCountsPerMicrosecond).
			self primitiveFail]
		ifFalse:
			[microseconds >= 1000
				ifTrue:
					[stackLimit = self allOnesAsCharStar ifFalse:
						[(Delay forMilliseconds: microseconds + 999 // 1000) wait.
						 self forceInterruptCheck]]
				ifFalse: [Processor yield]].
	"And increase the byteCount form which the microsecond clock is derived..."
	byteCount := byteCount + (microseconds * ByteCountsPerMicrosecond) - 1.
	self incrementByteCount! !

!StackInterpreterSimulator methodsFor: 'I/O primitives' stamp: 'ar 2/5/2001 17:24'!
ioScreenDepth
	^DisplayScreen actualScreenDepth.! !

!StackInterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'topa 7/10/2016 23:14'!
ioScreenScaleFactor
	^ 1.0! !

!StackInterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'eem 3/11/2015 19:25'!
ioScreenSize
	"Answer the screen extent packed into 32 bits.  In the simulator,
	 displayForm gets initialized with a fake form; don't be deceived."
	| extent |
	extent := (displayForm notNil and: [displayForm ~~ fakeForm])
				ifTrue: [displayForm extent]
				ifFalse: [self desiredDisplayExtent].
	^extent x << 16 + extent y! !

!StackInterpreterSimulator methodsFor: 'I/O primitives' stamp: 'eem 10/23/2020 12:49'!
ioSetInputSemaphore: index
	inputSemaphoreIndex := index! !

!StackInterpreterSimulator methodsFor: 'process primitive support' stamp: 'eem 2/3/2010 12:57'!
ioSetMaxExtSemTableSize: anInteger
	"Inform the VM of the maximum size the image will grow the exernal semaphore table (specialObjectsArray at: 39) to.
	 This allows the VM to allocate storage such that external semaphores can be signalled without locking."
	extSemTabSize := anInteger! !

!StackInterpreterSimulator methodsFor: 'I/O primitives' stamp: 'eem 5/12/2017 17:13'!
ioShow: destBits D: w i: h s: d p: left l: right a: top y: bottom
	"This is the simulator's implementation of ioShowDisplay."
	| raster pixPerWord simDisp realDisp rect |
	displayForm ifNil: [^self].
	displayBits = 0 ifTrue: [^self].
	pixPerWord := 32 // d.
	raster := displayForm width + (pixPerWord - 1) // pixPerWord.
	simDisp := Form new hackBits: objectMemory memory.
	displayForm unhibernate.
	realDisp := Form new hackBits: displayForm bits.
	realDisp
		copy: (0 @ (top * raster) extent: 4 @ (bottom - top * raster))
		from: 0 @ (destBits // 4 + (top * raster))
		in: simDisp
		rule: Form over.
	displayView ifNotNil: [^displayView changed].
	
	"If running without a view, just blat the bits onto the screen..."
	rect := 0 @ top corner: displayForm width @ bottom.
	Display
		copy: (rect translateBy: self displayLocation)
		from: rect topLeft
		in: displayForm
		rule: Form over! !

!StackInterpreterSimulator methodsFor: 'initialize-release' stamp: 'eem 11/10/2020 12:24'!
ioShutdownAllModules
	pluginList do:
		[:assoc| | pluginOrSelf |
		 ((pluginOrSelf := assoc value) ~~ self
		  and: [pluginOrSelf respondsTo: #shutdownModule]) ifTrue:
			[pluginOrSelf shutdownModule]]! !

!StackInterpreterSimulator methodsFor: 'process primitive support' stamp: 'eem 2/4/2010 09:37'!
ioSynchronousCheckForEvents
	"Hook allowing the platform to do anything it needs to do synchronously."! !

!StackInterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'eem 5/27/2021 10:15'!
ioUTCMicroseconds
	"Return the value of the microsecond clock."
	"NOT.  Actually, we want something a lot slower and, for exact debugging,
	something more repeatable than real time.  Dan had an idea: use the byteCount...
	 Use 50 byteCounts per usec by default; see StackInterpreterSimulator class>>initializeWithOptions:"

	^byteCount // ByteCountsPerMicrosecond + startMicroseconds
	
	"Dan:
	 At 20k bytecodes per second, this gives us about 200 ticks per second, or about 1/5
	 of what you'd expect for the real time clock.  This should still service events at one or
	 two per second"! !

!StackInterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'eem 1/25/2016 18:08'!
ioUTCStartMicroseconds
	"Answer the value of the microsecond clock at startup."

	^startMicroseconds ifNil:
		[self class initializationOptions
			at: #startMicroseconds
			ifAbsent: [[Time utcMicrosecondClock] on: Error do: [Time totalSeconds * 1000000]]]! !

!StackInterpreterSimulator methodsFor: 'primitive support' stamp: 'eem 7/9/2008 18:58'!
isPrimitiveFunctionPointerAnIndex
	"We save slots in the method cache by using the primitiveFunctionPointer
	 to hold either a function pointer or the index of a quick primitive. Since
	 quick primitive indices are small they can't be confused with function
	 addresses.  But since we use 1001 and up for external primitives that
	 would be functions in the C VM but are indices under simulation we treat
	 values above 1000 as if they were pointers (actually indices into the
	 externalPrimitiveTable)"

	^primitiveFunctionPointer isInteger
	  and: [primitiveFunctionPointer ~= 0
	  and: [primitiveFunctionPointer <= MaxQuickPrimitiveIndex]]! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 10/12/2009 01:18'!
isThreadedVM
	^false! !

!StackInterpreterSimulator methodsFor: 'frame access' stamp: 'eem 9/30/2010 13:37'!
isWidowedContext: aOnceMarriedContext
	"See if the argument is connected with a live frame or not.
	 If it is not, turn it into a bereaved single context."
	(stackPages isFree: stackPage) ifFalse: "in baseReturn the active page has been freed."
		[((stackPages somePageHasHeadFrameFP: framePointer)
		  or: [(stackPages somePageHasHeadFrameFP: localFP)
		  or: [stackPages allPagesFree]]) ifFalse:
			[thisContext sender sender method = (StackInterpreter >> #baseFrameReturn) ifFalse:
				[self halt: 'currentFP may not be written back to stack page']]].
	^super isWidowedContext: aOnceMarriedContext! !

!StackInterpreterSimulator methodsFor: 'compiled methods' stamp: 'eem 11/5/2013 17:53'!
literal: offset
	"trap pushes of forwarded literals to help debug following forwarded primitive args.
	 it is not an error to push a forwarded literal, but we'd like to step through any resulting
	 primtive failure code"
	| lit |
	lit := super literal: offset.
	(objectMemory isOopForwarded: lit) ifTrue:
		[self halt: 'forwarded literal in ', thisContext selector].
	^lit! !

!StackInterpreterSimulator methodsFor: 'plugin support' stamp: 'eem 8/20/2021 17:07'!
loadNewPlugin: pluginString
	(breakSelector notNil
	 and: [pluginString size = breakSelector size
	 and: [(self strncmp: pluginString _: breakSelector _: pluginString size) = 0]]) ifTrue:
		[self halt: pluginString].
	^(self tryLoadNewPlugin: pluginString pluginEntries: mappedPluginEntries) ifNotNil:
		[:entry|
		 pluginList := pluginList copyWith: entry.
		 entry]! !

!StackInterpreterSimulator methodsFor: 'spur bootstrap' stamp: 'eem 9/6/2013 13:30'!
localFP
	^localFP! !

!StackInterpreterSimulator methodsFor: 'spur bootstrap' stamp: 'eem 9/6/2013 13:51'!
localIP
	^localIP! !

!StackInterpreterSimulator methodsFor: 'spur bootstrap' stamp: 'eem 9/6/2013 14:04'!
localIP: pc
	localIP := pc! !

!StackInterpreterSimulator methodsFor: 'spur bootstrap' stamp: 'eem 9/16/2013 01:07'!
localSP
	^localSP! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 12/29/2016 08:25'!
logOfBytesVerify: nBytes fromFileNamed: fileName fromStart: loggingStart
	"Verify a questionable interpreter against a successful run"
	"self logOfBytesVerify: 10000 fromFileNamed: 'clone32Bytecodes.log' "
	
	| logFile rightWord prevCtxt |
	logFile := (FileStream readOnlyFileNamed: fileName) binary.
	transcript clear.
	byteCount := 0.
	quitBlock := [^self close].
	self initStackPages.
	self loadInitialContext.
	self internalizeIPandSP.
	self fetchNextBytecode.
	prevCtxt := 0.  prevCtxt := prevCtxt.
	[byteCount < nBytes] whileTrue:
		[
"
byteCount > 14560 ifTrue:
[self externalizeIPandSP.
prevCtxt = activeContext ifFalse:
 [prevCtxt := activeContext.
 transcript cr; nextPutAll: (self printTop: 2); endEntry].
transcript cr; print: byteCount; nextPutAll: ': ' , (activeContext hex); space;
 print: (instructionPointer - method - (BaseHeaderSize - 2));
 nextPutAll: ': <' , (self byteAt: localIP) hex , '>'; space;
 nextPutAll: (self symbolic: currentBytecode at: localIP inMethod: method); space;
 print: (self stackPointerIndex - TempFrameStart + 1); endEntry.
byteCount = 14590 ifTrue: [self halt]].
"
		loggingStart >= byteCount ifTrue:
			[rightWord := logFile nextWord.
			 currentBytecode = rightWord ifFalse:
				[self halt: 'halt at ', byteCount printString]].
		self dispatchOn: currentBytecode in: BytecodeTable.
		self incrementByteCount].
	self externalizeIPandSP.
	logFile close.
	self inform: nBytes printString , ' bytecodes verfied.'! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 12/29/2016 08:25'!
logOfBytesWrite: nBytes toFileNamed: fileName fromStart: loggingStart
	"Write a log file for testing a flaky interpreter on the same image"
	"self logOfBytesWrite: 10000 toFileNamed: 'clone32Bytecodes.log' "
	
	| logFile |
	logFile := (FileStream newFileNamed: fileName) binary.
	transcript clear.
	byteCount := 0.
	quitBlock := [^self close].
	self initStackPages.
	self loadInitialContext.
	self internalizeIPandSP.
	self fetchNextBytecode.
	[byteCount < nBytes] whileTrue:
		[byteCount >= loggingStart ifTrue:
			[logFile nextWordPut: currentBytecode].
		self dispatchOn: currentBytecode in: BytecodeTable.
		self incrementByteCount].
	self externalizeIPandSP.
	logFile close! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 12/29/2016 08:24'!
logOfSendsVerify: nSends fromFileNamed: fileName fromStart: loggingStart
	"Write a log file for testing a flaky interpreter on the same image"
	"self logOfSendsWrite: 10000 toFileNamed: 'clone32Messages.log' "
	
	| logFile priorFrame rightSelector prevCtxt |
	logFile := FileStream readOnlyFileNamed: fileName.
	transcript clear.
	byteCount := 0.
	sendCount := 0.
	priorFrame := localFP.
	quitBlock := [^self close].
	self initStackPages.
	self loadInitialContext.
	self internalizeIPandSP.
	self fetchNextBytecode.
	prevCtxt := 0.  prevCtxt := prevCtxt.
	[sendCount < nSends] whileTrue:
		[
"
byteCount>500 ifTrue:
[byteCount>550 ifTrue: [self halt].
self externalizeIPandSP.
prevCtxt = localFP ifFalse:
 [prevCtxt := localFP.
 transcript cr; nextPutAll: (self printTop: 2); endEntry].
transcript cr; print: byteCount; nextPutAll: ': ' , (localFP hex); space;
 print: (instructionPointer - method - (BaseHeaderSize - 2));
 nextPutAll: ': <' , (self byteAt: localIP) hex , '>'; space;
 nextPutAll: (self symbolic: currentBytecode at: localIP inMethod: method); space;
 print: (self stackPointerIndex - TempFrameStart + 1); endEntry.
].
"
		self dispatchOn: currentBytecode in: BytecodeTable.
		localFP = priorFrame ifFalse:
			[sendCount := sendCount + 1.
			 loggingStart >= sendCount ifTrue:
				[rightSelector := logFile nextLine.
				 (self stringOf: messageSelector) = rightSelector ifFalse:
					[self halt: 'halt at ', sendCount printString]].
			priorFrame := localFP].
		self incrementByteCount].
	self externalizeIPandSP.
	logFile close.
	self inform: nSends printString , ' sends verfied.'! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 12/29/2016 08:25'!
logOfSendsWrite: nSends toFileNamed: fileName fromStart: loggingStart
	"Write a log file for testing a flaky interpreter on the same image"
	"self logOfSendsWrite: 10000 toFileNamed: 'clone32Messages.log' "
	
	| logFile priorFrame |
	logFile := FileStream newFileNamed: fileName.
	transcript clear.
	byteCount := 0.
	sendCount := 0.
	priorFrame := localFP.
	quitBlock := [^self close].
	self initStackPages.
	self loadInitialContext.
	self internalizeIPandSP.
	self fetchNextBytecode.
	[sendCount < nSends] whileTrue:
		[self dispatchOn: currentBytecode in: BytecodeTable.
		localFP = priorFrame ifFalse:
			[sendCount >= loggingStart ifTrue:
				[sendCount := sendCount + 1.
				logFile nextPutAll: (self stringOf: messageSelector); cr].
			priorFrame := localFP].
		self incrementByteCount].
	self externalizeIPandSP.
	logFile close! !

!StackInterpreterSimulator methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:44'!
longAt: byteAddress
	self deprecated.
	^objectMemory longAt: byteAddress! !

!StackInterpreterSimulator methodsFor: 'memory access' stamp: 'eem 11/19/2016 09:44'!
longAt: byteAddress put: a32BitValue
	self deprecated.
	^objectMemory longAt: byteAddress put: a32BitValue! !

!StackInterpreterSimulator methodsFor: 'message sending' stamp: 'eem 8/17/2014 18:22'!
lookupMethodInClass: class
	lookupCount := lookupCount + 1.
	^super lookupMethodInClass: class! !

!StackInterpreterSimulator methodsFor: 'callback support' stamp: 'rmacnak 6/24/2015 21:17'!
lookupOrdinaryNoMNUEtcInClass: class
	lookupCount := lookupCount + 1.
	^super lookupOrdinaryNoMNUEtcInClass: class! !

!StackInterpreterSimulator methodsFor: 'object memory support' stamp: 'eem 8/20/2021 15:14'!
mapInterpreterOops
	super mapInterpreterOops.
	perMethodProfile ifNotNil:
		[:pmp| | map |
		 "The tricky thing here is that a method may get remapped to another method already in the profile, etc..."
		 map := Dictionary new.
		 perMethodProfile keysAndValuesDo:
			[:methodOop :count|
			(objectMemory shouldRemapOop: methodOop) ifTrue:
				[map at: methodOop put: {objectMemory remapObj: methodOop. count}]].
		 map isEmpty ifFalse:
			[map keysAndValuesDo:
				[:newOop :tuple|
				 [:oldOop :count|
				  (map includesKey: oldOop) ifFalse:
					[perMethodProfile removeKey: oldOop].
				  perMethodProfile at: newOop put: count] valueWithArguments: tuple]]]! !

!StackInterpreterSimulator methodsFor: 'plugin support' stamp: 'eem 6/3/2016 15:04'!
mappedPluginEntries
	^mappedPluginEntries! !

!StackInterpreterSimulator methodsFor: 'frame access' stamp: 'eem 11/8/2013 17:00'!
markContextAsDead: oop
	"(self withoutSmallIntegerTags: (objectMemory fetchPointer: SenderIndex ofObject: oop)) = -16r26824 ifTrue:
		[self halt]."
	^super markContextAsDead: oop! !

!StackInterpreterSimulator methodsFor: 'spur bootstrap' stamp: 'eem 9/12/2013 10:18'!
messageSelector: s
	"For e.g. SpurBootstrap"
	messageSelector := s! !

!StackInterpreterSimulator methodsFor: 'simulation only' stamp: 'eem 2/27/2013 13:50'!
methodForContext: aContextOop
	self assert: (objectMemory isContext: aContextOop).
	^objectMemory fetchPointer: MethodIndex ofObject: aContextOop! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'eem 8/21/2014 07:17'!
nameOfClass: classOop
	| numSlots maybeThisClassOop maybeNameOop name |
	classNameIndex ifNil: [^'??nil cnidx??'].
	numSlots := objectMemory numSlotsOf: classOop.
	numSlots = metaclassNumSlots ifTrue:
		[maybeThisClassOop := objectMemory fetchPointer: thisClassIndex ofObject: classOop.
		(self addressCouldBeClassObj: maybeThisClassOop) ifTrue:
			[name := self nameOfClass: maybeThisClassOop.
			 (name beginsWith: 'bad ') ifFalse:
				[^name, ' class']].
		 ^'bad class'].
	(numSlots > classNameIndex
	 and: [maybeNameOop := objectMemory fetchPointer: classNameIndex ofObject: classOop.
		objectMemory isBytes: maybeNameOop]) ifFalse:
		[^'bad class'].
	^self stringOf: maybeNameOop! !

!StackInterpreterSimulator methodsFor: 'initialize-release' stamp: 'eem 10/9/2008 16:37'!
nextShortFrom: aStream
	"Read a 16-bit quantity from the given (binary) stream."
	^self subclassResponsibility! !

!StackInterpreterSimulator methodsFor: 'initialize-release' stamp: 'eem 6/30/2021 18:53'!
openOn: fileName extraMemory: extraBytes
	"StackInterpreterSimulator new openOn: 'clone.im' extraMemory: 100000"

	| f version headerSize dataSize count oldBaseAddr bytesToShift swapBytes
	  headerFlags heapBase firstSegSize heapSize
	  hdrNumStackPages hdrEdenBytes hdrMaxExtSemTabSize allocationReserve |
	"open image file and read the header"

	(f := self openImageFileNamed: fileName) ifNil: [^self].

	"Set the image name and the first argument; there are
	 no arguments during simulation unless set explicitly."
	systemAttributes at: 1 put: fileName.

	["begin ensure block..."
	imageName := f fullName.
	f binary.

	version := self getWord32FromFile: f swap: false.  "current version: 16r1968 (=6504) vive la revolucion!!"
	(self readableFormat: version)
		ifTrue: [swapBytes := false]
		ifFalse: [(version := version byteSwap32) = self imageFormatVersion
					ifTrue: [swapBytes := true]
					ifFalse: [self error: 'incomaptible image format']].
	headerSize := self getWord32FromFile: f swap: swapBytes.
	dataSize := self getLongFromFile: f swap: swapBytes.  "length of heap in file"
	oldBaseAddr := self getLongFromFile: f swap: swapBytes.  "object memory base address of image"
	objectMemory specialObjectsOop: (self getLongFromFile: f swap: swapBytes).
	objectMemory lastHash: (self getLongFromFile: f swap: swapBytes).  "Should be loaded from, and saved to the image header"

	savedWindowSize	:= self getLongFromFile: f swap: swapBytes.
	headerFlags		:= self getLongFromFile: f swap: swapBytes.
	self setImageHeaderFlagsFrom: headerFlags.
	extraVMMemory	:= self getWord32FromFile: f swap: swapBytes.
	hdrNumStackPages	:= self getShortFromFile: f swap: swapBytes.
	"4 stack pages is small.  Should be able to run with as few as
	 three. 4 should be comfortable but slow.  8 is a reasonable
	 default. Can be changed via vmParameterAt: 43 put: n"
	numStackPages := desiredNumStackPages ~= 0
						ifTrue: [desiredNumStackPages]
						ifFalse: [hdrNumStackPages = 0
									ifTrue: [self defaultNumStackPages]
									ifFalse: [hdrNumStackPages]].
	desiredNumStackPages := hdrNumStackPages.
	"pad to word boundary.  This slot can be used for anything else that will fit in 16 bits.
	 It is used for the cog code size in Cog.  Preserve it to be polite to other VMs."
	theUnknownShort	:= self getShortFromFile: f swap: swapBytes.
	self assert: f position = (objectMemory wordSize = 4 ifTrue: [40] ifFalse: [64]).
	hdrEdenBytes		:= self getWord32FromFile: f swap: swapBytes.
	objectMemory edenBytes: (hdrEdenBytes = 0
							ifTrue: [objectMemory defaultEdenBytes]
							ifFalse: [hdrEdenBytes]).
	desiredEdenBytes := hdrEdenBytes.
	hdrMaxExtSemTabSize := self getShortFromFile: f swap: swapBytes.
	hdrMaxExtSemTabSize ~= 0 ifTrue:
		[self setMaxExtSemSizeTo: hdrMaxExtSemTabSize].
	"pad to word boundary.  This slot can be used for anything else that will fit in 16 bits.
	 Preserve it to be polite to other VMs."
	the2ndUnknownShort	:= self getShortFromFile: f swap: swapBytes.
	self assert: f position = (objectMemory wordSize = 4 ifTrue: [48] ifFalse: [72]).
	firstSegSize := self getLongFromFile: f swap: swapBytes.
	objectMemory firstSegmentSize: firstSegSize.
	"compare memory requirements with availability"
	allocationReserve := self interpreterAllocationReserveBytes.
	objectMemory hasSpurMemoryManagerAPI
		ifTrue:
			[| freeOldSpaceInImage headroom |
			 freeOldSpaceInImage := self getLongFromFile: f swap: swapBytes.
			 headroom := objectMemory
							initialHeadroom: extraVMMemory
							givenFreeOldSpaceInImage: freeOldSpaceInImage.
			 heapSize := objectMemory roundUpHeapSize:
						   dataSize
						+ headroom
						+ objectMemory newSpaceBytes
						+ (headroom > allocationReserve
							ifTrue: [0]
							ifFalse: [allocationReserve])]
		ifFalse:
			[heapSize :=  dataSize
						+ extraBytes
						+ objectMemory newSpaceBytes
						+ (extraBytes > allocationReserve
							ifTrue: [0]
							ifFalse: [allocationReserve])].
	"allocate interpreter memory"
	heapBase := objectMemory
					setHeapBase: objectMemory startOfMemory
					memoryLimit: objectMemory startOfMemory + heapSize
					endOfMemory: objectMemory startOfMemory + dataSize. "bogus for Spur"
	objectMemory allocateMemoryOfSize: objectMemory memoryLimit.
	"read in the image in bulk, then swap the bytes if necessary"
	f position: headerSize.
	count := objectMemory readHeapFromImageFile: f dataBytes: dataSize.
	count ~= dataSize ifTrue: [self halt]]
		ensure: [f close].

	self ensureImageFormatIsUpToDate: swapBytes.

	bytesToShift := objectMemory memoryBaseForImageRead - oldBaseAddr.  "adjust pointers for zero base address"
	UIManager default
		informUser: 'Relocating object pointers...'
		during: [self initializeInterpreter: bytesToShift]! !

!StackInterpreterSimulator methodsFor: 'accessing' stamp: 'eem 10/11/2014 14:53'!
parent
	^ parent! !

!StackInterpreterSimulator methodsFor: 'accessing' stamp: 'eem 5/24/2014 23:38'!
parent: anObject

	parent := anObject! !

!StackInterpreterSimulator methodsFor: 'plugin support' stamp: 'eem 10/30/2014 12:32'!
pluginEntryFor: mapIndex
	^mappedPluginEntries at: (mapIndex > 1000
								ifTrue: [externalPrimitiveTable at: mapIndex - 1001]
								ifFalse: [mapIndex])! !

!StackInterpreterSimulator methodsFor: 'object memory support' stamp: 'eem 1/17/2014 19:03'!
preBecomeAction
	"((objectMemory instVarNamed: 'becomeEffectsFlags') anyMask: BecameCompiledMethodFlag) ifTrue:
		[self halt]."
	super preBecomeAction! !

!StackInterpreterSimulator methodsFor: 'other primitives' stamp: 'di 6/29/1998 22:17'!
primStringcomparewithcollated
	^ self primitiveFail! !

!StackInterpreterSimulator methodsFor: 'other primitives' stamp: 'ikp 12/15/1998 23:30'!
primStringfindSubstringinstartingAtmatchTable
	^self primitiveFail! !

!StackInterpreterSimulator methodsFor: 'other primitives' stamp: 'di 11/8/1998 13:04'!
primStringindexOfAsciiinStringstartingAt
	^ self primitiveFail! !

!StackInterpreterSimulator methodsFor: 'other primitives' stamp: 'yo 1/29/1999 21:37'!
primStringtranslatefromtotable
	^ self primitiveFail! !

!StackInterpreterSimulator methodsFor: 'indexing primitives' stamp: 'eem 9/13/2013 16:45'!
primitiveAtPut
	"16r1510B8 = (self stackValue: 2) ifTrue:
		[self halt]."
	^super primitiveAtPut! !

!StackInterpreterSimulator methodsFor: 'other primitives' stamp: ''!
primitiveBeep

	Beeper beep.! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 6/27/2008 16:27'!
primitiveContextAt
	"self halt."
	^super primitiveContextAt! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 3/16/2021 12:24'!
primitiveDoPrimitiveWithArgs
	| primIndex |
	primIndex := objectMemory integerValueOf: (self stackValue: 1).
	NewspeakVM ifFalse:
		[transcript nextPutAll: 'DO PRIMITIVE: '; print: (self functionPointerFor: primIndex inClass: nil); cr; flush].
	(#(110) includes: primIndex) ifTrue:
		[self halt].
	^super primitiveDoPrimitiveWithArgs! !

!StackInterpreterSimulator methodsFor: 'control primitives' stamp: 'eem 7/25/2020 21:39'!
primitiveExecuteMethod
	(InitializationOptions at: #haltOnExecuteMethod ifAbsent: [true]) ifTrue:
		[self halt: thisContext selector].
	^super primitiveExecuteMethod

	"InitializationOptions at: #haltOnExecuteMethod put: false"! !

!StackInterpreterSimulator methodsFor: 'control primitives' stamp: 'eem 7/25/2020 19:11'!
primitiveExecuteMethodArgsArray
	(InitializationOptions at: #haltOnExecuteMethod ifAbsent: [true]) ifTrue:
		[self halt: thisContext selector].
	^super primitiveExecuteMethodArgsArray

	"InitializationOptions at: #haltOnExecuteMethod put: false"! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 12/4/2008 12:58'!
primitiveFailFor: reasonCode
	"self halt."
	^super primitiveFailFor: reasonCode! !

!StackInterpreterSimulator methodsFor: 'other primitives' stamp: 'eem 2/13/2014 09:35'!
primitiveGetAttribute
	"Fetch the system attribute with the given integer ID. The result is a string, which will be empty if the attribute is not defined."

	| index s attribute |
	index := self stackIntegerValue: 0.
	self successful ifTrue: [
		attribute := systemAttributes at: index ifAbsent: [Smalltalk vm getSystemAttribute: index].
		attribute ifNil: [ ^self primitiveFail ].
		s := objectMemory instantiateClass: (objectMemory splObj: ClassByteString) indexableSize: attribute size.
		1 to: attribute size do: [ :i |
			objectMemory storeByte: i-1 ofObject: s withValue: (attribute at: i) asciiValue].
		self pop: 2 "rcvr, attr" thenPush: s]! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 9/22/2013 07:56'!
primitiveIdentityHash
	"| oop |
	oop := self stackTop.
	((objectMemory isBytes: oop)
	and: [(objectMemory lengthOf: oop) = 'smallSelect' size
	and: [(self stringOf: oop) = 'smallSelect']]) ifTrue:
		[self halt]."
	^super primitiveIdentityHash! !

!StackInterpreterSimulator methodsFor: 'I/O primitives' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
primitiveKbdNext

	self pop: 1.
	Sensor keyboardPressed
		ifTrue: [self pushInteger: Sensor primKbdNext]
		ifFalse: [self push: objectMemory nilObject]! !

!StackInterpreterSimulator methodsFor: 'I/O primitives' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
primitiveKbdPeek

	self pop: 1.
	Sensor keyboardPressed
		ifTrue: [self pushInteger: Sensor primKbdPeek]
		ifFalse: [self push: objectMemory nilObject]! !

!StackInterpreterSimulator methodsFor: 'I/O primitives' stamp: 'di 9/23/97 12:54'!
primitiveMouseButtons
	| buttons |
	self pop: 1.
	buttons := Sensor primMouseButtons.
	self pushInteger: buttons! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 2/11/2014 17:05'!
primitiveNewWithArg
	"(objectMemory hasSpurMemoryManagerAPI
	 and: [self classNameOf: (self stackValue: 1) Is: 'MethodDictionary']) ifTrue:
		[self halt]."
	"| hash |
	hash := objectMemory rawHashBitsOf: (self stackValue: 1)."
	"| format |
	format := objectMemory instSpecOfClass: (self stackValue: 1)."
	"(objectMemory numSlotsOf: (self stackValue: 1)) = 3 ifTrue:
		[self halt]."
	super primitiveNewWithArg.
	"self successful ifTrue:
		[(Smalltalk at: #Counts ifAbsentPut: [Bag new]) add: format]."
	"(self successful and: [objectMemory hasSpurMemoryManagerAPI]) ifTrue:
		[(Smalltalk at: #Counts ifAbsentPut: [Bag new]) add: hash]"
	"Smalltalk removeKey: #Counts"
	"Counts sortedCounts collect: [:assoc|
		assoc key -> ((SpurMemoryManager organization listAtCategoryNamed: #'header formats') detect:
										[:f| (SpurMemoryManager basicNew perform: f) = assoc value])]
		{3251->#arrayFormat.
		 1685->#firstByteFormat.
		 1533->#firstLongFormat.
		 110->#weakArrayFormat.
		 35->#indexablePointersFormat.
		 5->#nonIndexablePointerFormat}"
	"Counts sortedCounts collect: [:assoc|
		assoc value = 0
			ifTrue: [assoc]
			ifFalse: [assoc key -> {(self nameOfClass: (objectMemory classAtIndex: assoc value)).
									(SpurMemoryManager organization listAtCategoryNamed: #'header formats') detect:
										[:f| (objectMemory perform: f)
											= (objectMemory instSpecOfClass: (objectMemory classAtIndex: assoc value))]}]]
	{1062->#('Array' #arrayFormat).
	 777->#('Bitmap' #firstLongFormat).
	 395->#('Float' #firstLongFormat).
	 345->#('ByteString' #firstByteFormat).
	 237->#('MatrixTransform2x3' #firstLongFormat).
	 233->#('LargePositiveInteger' #firstByteFormat).
	 103->#('WordArray' #firstLongFormat).
	 58->#('WeakArray' #weakArrayFormat).
	 52->#('WeakMessageSend' #weakArrayFormat).
	 9->#('MethodContext' #indexablePointersFormat).
	 4->#('DirectoryEntry' #nonIndexablePointerFormat).
	 3->#('BalloonBuffer' #firstLongFormat).
	 1->#('ByteArray' #firstByteFormat).
	 1->0}"! !

!StackInterpreterSimulator methodsFor: 'control primitives' stamp: 'eem 1/11/2013 18:38'!
primitiveObject: actualReceiver perform: selector withArguments: argumentArray lookedUpIn: lookupClass
	"Override to allow simulator to filter start-up items to avoid as-yet unsimulatable plugins."

	(self filterPerformOf: selector to: actualReceiver) ifTrue:
		[^self pop: argumentCount + 1 thenPush: actualReceiver].
	^super primitiveObject: actualReceiver perform: selector withArguments: argumentArray lookedUpIn: lookupClass! !

!StackInterpreterSimulator methodsFor: 'object access primitives' stamp: 'eem 2/5/2014 18:31'!
primitiveObjectPointsTo
	"self halt."
	^super primitiveObjectPointsTo! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 9/17/2013 12:49'!
primitivePerform
	| selector |
	selector := self stackValue: argumentCount - 1.
	self sendBreakpoint: selector receiver: (self stackValue: argumentCount).
	(self filterPerformOf: selector to: (self stackValue: argumentCount)) ifTrue:
		[^self pop: argumentCount].
	^super primitivePerform! !

!StackInterpreterSimulator methodsFor: 'object access primitives' stamp: 'eem 6/22/2010 11:06'!
primitiveStoreStackp
	"self printContext: (self stackValue: 1).
	self halt."
	super primitiveStoreStackp.
	"self printContext: self stackTop"! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 6/25/2008 09:25'!
primitiveSuspend
	"Catch errors before we start the whole morphic error process"

	"byteCount > 1000000 ifTrue: [self halt]."  "Ignore early process activity"
	^ super primitiveSuspend! !

!StackInterpreterSimulator methodsFor: 'system control primitives' stamp: 'eem 5/5/2015 19:15'!
primitiveVMParameter
	(self stackTop = (objectMemory integerObjectOf: 9)
	 or: [self stackTop = (objectMemory integerObjectOf: 52)]) ifTrue:
		[self halt].
	^super primitiveVMParameter! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 6/25/2008 09:24'!
primitiveWait
	"Catch errors before we start the whole morphic error process"

	"byteCount > 1000000 ifTrue: [self halt]."  "Ignore early process activity"
	^ super primitiveWait! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 6/20/2008 17:39'!
print: it

	traceOn ifTrue:
		[it isString ifTrue: [transcript nextPutAll: it] ifFalse: [it printOn: transcript]]! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'di 5/11/2004 18:28'!
printChar: aByte

	traceOn ifTrue: [ transcript nextPut: aByte asCharacter ].! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 6/26/2012 09:20'!
printCurrentBytecodeOn: aStream
	| code |
	code := currentBytecode radix: 16.
	aStream ensureCr; print: localIP - method - 3; tab.
	bytecodeSetSelector > 0 ifTrue:
		[aStream nextPutAll: 'ALT '].
	aStream
		nextPut: (code size < 2
					ifTrue: [$0]
					ifFalse: [code at: 1]);
		nextPut: code last; space;
		nextPutAll: (BytecodeTable at: currentBytecode + 1);
		space;
		nextPut: $(; print: byteCount + 1; nextPut: $)! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 7/4/2008 15:02'!
printECS
	self printCallStackFP: framePointer! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 6/22/2010 14:46'!
printFrameFlagsForFP: theFP
	| address it |
	<inline: false>
	<var: #theFP type: #'char *'>
	<var: #address type: #'char *'>
	address := theFP + FoxFrameFlags.
	it := stackPages longAt: address.
	self printHex: address;
		printChar: $/;
		printNum: (stackPages memIndexFor: address);
		print: ':       flags: ';
		printHex: it.
	it ~= 0 ifTrue:
		[self printChar: $=; printNum: it].
	self print: '  numArgs: '; printNum: (self frameNumArgs: theFP);
		print: '  hasContext: '; printNum: (self frameHasContext: theFP);
		print: '  isBlock: '; printNum: (self frameIsBlockActivation: theFP);
		cr! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 6/22/2010 14:42'!
printFrameOop: name at: address
	| it |
	<inline: false>
	<var: #name type: #'char *'>
	<var: #address type: #'char *'>
	it := stackPages longAt: address.
	self printHex: address;
		printChar: $/;
		printNum: (stackPages memIndexFor: address);
		printChar: $:.
	1 to: 12 - (self strlen: name) do: [:i| self printChar: $ ].
	self print: name;
		print: ': ';
		printHex: it.
	self tab;
		print: (self shortPrint: it);
		cr! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 10/29/2013 16:05'!
printFrameThing: name andFrame: theFP at: address
	<var: #theFP type: #'char *'>
	| it |
	<inline: false>
	<var: #name type: #'char *'>
	<var: #address type: #'char *'>
	it := stackPages longAt: address.
	self printHex: address;
		printChar: $/;
		printNum: (stackPages memIndexFor: address);
		printChar: $:.
	1 to: 12 - (self strlen: name) do: [:i| self printChar: $ ].
	self print: name;
		print: ': ';
		printHex: it.
	it ~= 0 ifTrue:
		[self printChar: $=; printNum: it].
	self print: ' frame: '; printHex: theFP; cr! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 6/22/2010 14:44'!
printFrameThing: name at: address
	| it |
	<inline: false>
	<var: #name type: #'char *'>
	<var: #address type: #'char *'>
	it := stackPages longAt: address.
	self printHex: address;
		printChar: $/;
		printNum: (stackPages memIndexFor: address);
		printChar: $:.
	1 to: 12 - (self strlen: name) do: [:i| self printChar: $ ].
	self print: name;
		print: ': ';
		printHex: it.
	it ~= 0 ifTrue:
		[self printChar: $=; printNum: it].
	self cr! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 12/25/2019 19:19'!
printHex: anInteger

	traceOn ifTrue:
		[transcript
			next: 8 - (anInteger digitLength * 2) put: Character space;
			nextPutAll: (anInteger storeStringBase: 16)]! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 12/25/2016 11:37'!
printHexnp: anInteger

	traceOn ifTrue:
		[transcript nextPutAll: ((anInteger ifNil: [0]) storeStringBase: 16)]! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 1/2/2017 17:39'!
printHexnpnp: anInteger
	"Print n in hex, in the form '1234', unpadded"
	traceOn ifTrue:
		[transcript nextPutAll: ((anInteger ifNil: [0]) printStringBase: 16)]! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 7/4/2008 15:02'!
printLCS
	self printCallStackFP: localFP! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 7/7/2008 16:45'!
printNum: anInteger

	traceOn ifTrue: [ transcript print: anInteger ].! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 1/15/2018 13:20'!
printSends
	^printSends or: [printBytecodeAtEachStep]! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 12/18/2015 10:27'!
printSends: aBoolean
	printSends := aBoolean! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 7/1/2008 09:24'!
printStringForCurrentBytecode
	^String streamContents: [:str| self printCurrentBytecodeOn: str]! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'di 5/11/2004 18:28'!
profile: nBytecodes
	"(InterpreterSimulator new openOn: 'clonex.image') profile: 60000"
	transcript clear.
	byteCount := 0.
	MessageTally spyOn: [self runForNBytes: nBytecodes].
	self close! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'di 11/29/2000 10:25'!
profileSends: nBytecodes
	"(InterpreterSimulator new openOn: 'clonex.image') profileSends: 5000"
	MessageTally tallySendsTo: self
		inBlock: [self runForNBytes: nBytecodes]
		showTree: true.
	self close! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 6/30/2008 15:04'!
pushMaybeContextReceiverVariable: index
	"(index = SenderIndex or: [index = ClosureIndex]) ifTrue: [self halt]."
	^super pushMaybeContextReceiverVariable: index! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'EstebanLorenzano 1/22/2015 15:12'!
redirectTranscriptToHost
	"Sometimes you need the capability to see the transcript output of the image into the 
	 local transcript (for debugging when bootstrapping, for example). "
	transcript := Transcript! !

!StackInterpreterSimulator methodsFor: 'simulation only' stamp: 'eem 8/23/2021 11:15'!
reportPerMethodProfilingOn: aStream
	"Generate a sorted report of bytecodes per method, and instructions per trampoline/cog method/pic.
	 Answer the total number of bytecodes executed."
	^self reportPerMethodProfilingOn: aStream cutoff: nil! !

!StackInterpreterSimulator methodsFor: 'simulation only' stamp: 'eem 8/23/2021 11:15'!
reportPerMethodProfilingOn: aStream cutoff: cutoffOrNil
	"Generate a sorted report of bytecodes per method, and instructions per trampoline/cog method/pic.
	 Answer the total number of bytecodes executed."
	| total |
	perMethodProfile ifNil:
		[aStream nextPutAll: 'not collecting profile'.
		 ^self].
	total := 0.
	aStream nextPutAll: 'bytecodes dispatched per method'; cr.
	(perMethodProfile associations sorted: [:a1 :a2| a1 value >= a2 value]) do:
		[:assoc|
 		total := total + assoc value.
		(cutoffOrNil isNil
		 or: [assoc value > cutoffOrNil]) ifTrue:
			[aStream tab; print: assoc value; nextPut: $:; tab.
			self printNameOfClass: (self methodClassOf: assoc key) count: 2 on: aStream.
			aStream next: 2 put: $>.
			self printStringOf: (self findSelectorOfMethod: assoc key) on: aStream.
			aStream cr]].
	aStream flush.
	^total! !

!StackInterpreterSimulator methodsFor: 'simulation only' stamp: 'eem 8/23/2021 10:47'!
reportSingleMethodProfile: assoc on: aStream
	"Generate a sorted report of bytecodes per method, and instructions per trampoline/cog method/pic.
	 Answer an Array with the four totals."
	| primIndex literal |
	aStream tab; print: assoc value; nextPut: $:; tab.
	self printNameOfClass: (self methodClassOf: assoc key) count: 2 on: aStream.
	aStream next: 2 put: $>.
	self printStringOf: (self findSelectorOfMethod: assoc key) on: aStream.
	primIndex := self primitiveIndexOf: assoc key.
	((self isQuickPrimitiveIndex: primIndex)
	or: [(primitiveTable at: primIndex) isInteger]) ifFalse:
		[aStream space; nextPutAll: (primitiveTable at: primIndex).
		 primIndex = PrimNumberExternalCall ifTrue:
			[literal := self literal: 0 ofMethod: assoc key.
			 aStream
				space;
				nextPutAll: (self stringOf: (objectMemory fetchPointer: ExternalCallLiteralModuleNameIndex ofObject: literal));
				space;
				nextPutAll: (self stringOf: (objectMemory fetchPointer: ExternalCallLiteralFunctionNameIndex ofObject: literal))]].
	aStream cr! !

!StackInterpreterSimulator methodsFor: 'image save/restore' stamp: 'eem 8/13/2018 15:22'!
reverseBytesInImage
	UIManager default
		informUser: 'Swapping bytes of foreign image...'
		during: [super reverseBytesInImage]! !

!StackInterpreterSimulator methodsFor: 'method lookup cache' stamp: 'eem 6/20/2009 11:33'!
rewriteMethodCacheEntryForExternalPrimitiveToFunction: localPrimAddress
	self assert: (localPrimAddress isSymbol
			or: [localPrimAddress isInteger
				and: [localPrimAddress = 0
					or: [(localPrimAddress between: 256 and: 519)
					or: [localPrimAddress > 1000]]]]).
	^super rewriteMethodCacheEntryForExternalPrimitiveToFunction: localPrimAddress! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 8/23/2018 09:38'!
run
	"Just run"
	quitBlock := [displayView ifNotNil:
				   [displayView containingWindow ifNotNil:
					[:topWindow|
					((World submorphs includes: topWindow)
					 and: [UIManager default confirm: 'close?']) ifTrue:
						[topWindow delete]]].
				  ^self close].
	self initStackPages.
	self loadInitialContext.
	self initialEnterSmalltalkExecutive.
	localIP := localIP - 1.
	"undo the pre-increment of IP before returning"
	self externalizeIPandSP! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 8/3/2008 21:52'!
runAtEachStep: aBlock
	self initStackPages.
	self loadInitialContext.
	self internalizeIPandSP.
	self fetchNextBytecode.
	[true] whileTrue:
		[self assertValidExecutionPointers.
		 aBlock value: currentBytecode.
		 self dispatchOn: currentBytecode in: BytecodeTable.
		 self incrementByteCount].
	localIP := localIP - 1.
	"undo the pre-increment of IP before returning"
	self externalizeIPandSP! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 8/3/2008 21:52'!
runAtEachStep: aBlock breakCount: breakCount
	self initStackPages.
	self loadInitialContext.
	self internalizeIPandSP.
	self fetchNextBytecode.
	[true] whileTrue:
		[self assertValidExecutionPointers.
		 aBlock value: currentBytecode.
		 self dispatchOn: currentBytecode in: BytecodeTable.
		 self incrementByteCount.
		 byteCount = breakCount ifTrue:
			[self halt]].
	localIP := localIP - 1.
	"undo the pre-increment of IP before returning"
	self externalizeIPandSP! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 8/3/2008 21:54'!
runForNBytes: nBytecodes 
	"Do nByteCodes more bytecode dispatches.
	Keep byteCount up to date.
	This can be run repeatedly."
	| endCount |
	self initStackPages.
	self loadInitialContext.
	endCount := byteCount + nBytecodes.
	self internalizeIPandSP.
	self fetchNextBytecode.
	[byteCount < endCount] whileTrue:
		[self dispatchOn: currentBytecode in: BytecodeTable.
		 self incrementByteCount].
	localIP := localIP - 1.
	"undo the pre-increment of IP before returning"
	self externalizeIPandSP! !

!StackInterpreterSimulator methodsFor: 'I/O primitives' stamp: 'eem 10/11/2013 09:56'!
savedWindowSize
	^savedWindowSize ifNil: [0]! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'eem 2/11/2020 10:50'!
setBreakBlockFromString: aString
	| bString block |
	bString := aString withBlanksTrimmed.
	bString first = $- ifTrue:
		[^self breakBlock: nil].
	bString first ~= $[ ifTrue:
		[bString := '[:_address|', bString, ']'].
	block := [Compiler evaluate: bString for: self logged: false]
				on: Error
				do: [:ex|
					UIManager default warn: ex messageText.
					^self].
	self breakBlock: block! !

!StackInterpreterSimulator methodsFor: 'accessing' stamp: 'eem 1/5/2018 15:19'!
setBreakCount: anInteger
	breakCount := anInteger! !

!StackInterpreterSimulator methodsFor: 'spur bootstrap' stamp: 'eem 10/11/2013 10:06'!
setDisplayForm: aForm
	displayForm := aForm! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 12/3/2010 07:23'!
shortPrintContext: aContext
	transcript ensureCr.
	^super shortPrintContext: aContext! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 10/11/2013 12:37'!
shortPrintFrame: theFP
	self transcript ensureCr.
	^super shortPrintFrame: theFP! !

!StackInterpreterSimulator methodsFor: 'process primitive support' stamp: 'eem 10/23/2020 12:53'!
signalSemaphoreWithIndex: index
	"Record the given semaphore index in the double buffer semaphores array to be signaled
	 at the next convenient moment. Force a real interrupt check as soon as possible.
	 This is a simulation.  See platforms/Cross/vm/sqExternalSemaphores.c for the real code.
	 Thanks to Levente Uzoni for making this version almost thread-safe (in Smalltalk)"
	<doNotGenerate>
	| originalResponses newRequests newResponses |
	index <= 0 ifTrue: [^false].
	index > externalSemaphoreSignalRequests size ifTrue:
		[newRequests := Array new: 1 << index highBit withAll: 0.
		 newResponses := newRequests shallowCopy].
	"This is a lock-free thread-safe grow...; thanks Levente"
	originalResponses := externalSemaphoreSignalResponses.
	[index > externalSemaphoreSignalRequests size] whileTrue:
		[newRequests
			replaceFrom: 1
			to: externalSemaphoreSignalRequests size
			with: externalSemaphoreSignalRequests
			startingAt: 1.
		 newResponses
			replaceFrom: 1
			to: externalSemaphoreSignalResponses size
			with: externalSemaphoreSignalResponses
			startingAt: 1.
		 externalSemaphoreSignalResponses == originalResponses "This should always be true."
			ifTrue:
				[externalSemaphoreSignalRequests := newRequests.
				 externalSemaphoreSignalResponses := newResponses]
			ifFalse:
				[originalResponses := externalSemaphoreSignalResponses]].
	"This is not thread-safe however..."
	externalSemaphoreSignalRequests
		at: index
		put: (externalSemaphoreSignalRequests at: index) + 1.
	^true! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 9/18/2013 12:37'!
singleStep
	self assertValidExecutionPointers.
	atEachStepBlock value. "N.B. may be nil"
	self dispatchOn: currentBytecode in: BytecodeTable.
	self incrementByteCount! !

!StackInterpreterSimulator methodsFor: 'primitive support' stamp: 'eem 12/22/2017 10:56'!
slowPrimitiveResponse
	primTraceLog ifNotNil:
		[primTraceLog size > 127 ifTrue:
			[primTraceLog removeFirst].
		 primTraceLog addLast: primitiveFunctionPointer].
	^super slowPrimitiveResponse! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 6/20/2008 15:21'!
space

	traceOn ifTrue: [ transcript space ]! !

!StackInterpreterSimulator methodsFor: 'plugin support' stamp: 'ar 5/11/2000 21:47'!
sqGetInterpreterProxy
	"I am basically my own proxy..."
	^self! !

!StackInterpreterSimulator methodsFor: 'memory access' stamp: 'ar 2/25/2001 17:16'!
sqMemoryExtraBytesLeft: includingSwap
	^0! !

!StackInterpreterSimulator methodsFor: 'memory access' stamp: 'di 5/11/2004 18:29'!
sqShrinkMemory: oldLimit By: delta
	transcript show: 'shrink memory from ', oldLimit printString, ' by ', delta printString, ' remember it doesn''t actually shrink in simulation'; cr.

	^ oldLimit! !

!StackInterpreterSimulator methodsFor: 'spur bootstrap' stamp: 'eem 9/17/2013 15:16'!
stackPage
	^stackPage! !

!StackInterpreterSimulator methodsFor: 'simulation only' stamp: 'eem 11/2/2013 01:12'!
stackPages
	^stackPages! !

!StackInterpreterSimulator methodsFor: 'simulation only' stamp: 'eem 9/6/2020 14:22'!
startInContextSuchThat: aBlock
	"Change the active process's suspendedContext to its sender, which short-cuts the
	 initialization of the system.  This can be a short-cut to running code, e.g. when doing
		Smalltalk saveAs.
		Compiler recompileAll
	 via e.g.
		vm startInContextSuchThat: [:ctxt| (vm stringOf: (vm penultimateLiteralOf: (vm methodForContext: ctxt))) = 'DoIt']"
	<doNotGenerate>
	| context activeProc |
	activeProc := self activeProcess.
	context := objectMemory fetchPointer: SuspendedContextIndex ofObject: activeProc.
	[context = objectMemory nilObject ifTrue:
		[self error: 'no context found'].
	 aBlock value: context] whileFalse:
		[context := objectMemory fetchPointer: SenderIndex ofObject: context].
	objectMemory storePointer: SuspendedContextIndex ofObject: activeProc withValue: context.
	"Now push a dummy return value."
	objectMemory
		storePointerUnchecked: (self fetchStackPointerOf: context) + CtxtTempFrameStart
		ofObject: context
		withValue: objectMemory nilObject.
	self storeInteger: StackPointerIndex
		ofObject: context
		withValue: (self fetchStackPointerOf: context) + 1! !

!StackInterpreterSimulator methodsFor: 'simulation only' stamp: 'eem 9/6/2020 14:29'!
startInSender
	"Change the active process's suspendedContext to its sender, which short-cuts the
	 initialization of the system.  This can be a short-cut to running code, e.g. when doing
		Smalltalk saveAs.
		Compiler recompileAll"
	
	| activeContext activeProc senderContext |
	activeProc := self activeProcess.
	activeContext := objectMemory fetchPointer: SuspendedContextIndex ofObject: activeProc.
	senderContext := objectMemory fetchPointer: SenderIndex ofObject: activeContext.
	objectMemory storePointer: SuspendedContextIndex ofObject: activeProc withValue: senderContext.
	"Now push a dummy return value."
	objectMemory
		storePointerUnchecked: (self fetchStackPointerOf: senderContext) + CtxtTempFrameStart
		ofObject: senderContext
		withValue: objectMemory nilObject.
	self storeInteger: StackPointerIndex
		ofObject: senderContext
		withValue: (self fetchStackPointerOf: senderContext) + 1! !

!StackInterpreterSimulator methodsFor: 'initialize-release' stamp: 'eem 10/9/2013 14:39'!
startOfMemory
	self shouldNotImplement! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 12/4/2008 12:57'!
success: successBoolean
	"successBoolean ifFalse:
		[self halt]."
	^super success: successBoolean! !

!StackInterpreterSimulator methodsFor: 'simulation only' stamp: 'eem 1/16/2014 11:40'!
systemAttributes
	^systemAttributes! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 6/20/2008 17:05'!
tab

	traceOn ifTrue: [ transcript tab ].! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 12/29/2016 08:23'!
test
	self initStackPages.
	self loadInitialContext.
	transcript clear.
	byteCount := 0.
	breakCount := -1.
	quitBlock := [^self close].
	printSends := printReturns := true.
	self internalizeIPandSP.
	self fetchNextBytecode.
	[true] whileTrue:
		[self assertValidExecutionPointers.
		 printFrameAtEachStep ifTrue:
			[self printFrame: localFP WithSP: localSP].
		 printBytecodeAtEachStep ifTrue:
			[self printCurrentBytecodeOn: Transcript.
			 Transcript cr; flush].
		 self dispatchOn: currentBytecode in: BytecodeTable.
		 self incrementByteCount.
		 byteCount = breakCount ifTrue:
			["printFrameAtEachStep :=" printBytecodeAtEachStep := true.
			 self halt: 'hit breakCount break-point']].
	self externalizeIPandSP! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 12/29/2016 08:23'!
test1
	self initStackPages.
	self loadInitialContext.
	transcript clear.
	byteCount := 0.
	breakCount := -1.
	self setBreakSelector: #blockCopy:.
	quitBlock := [^self close].
	printSends := printReturns := true.
	self internalizeIPandSP.
	self fetchNextBytecode.
	[true] whileTrue:
		[self assertValidExecutionPointers.
		 "byteCount >= 22283 ifTrue:
			[(self checkIsStillMarriedContext: 22186072 currentFP: localFP) ifFalse:
				[self halt]]."
		 (printBytecodeAtEachStep
		  "and: [self isMarriedOrWidowedContext: 22189568]") ifTrue:
			["| thePage |
			 thePage := stackPages stackPageFor: (self frameOfMarriedContext: 22189568).
			 thePage == stackPage
				ifTrue: [self shortPrintFrameAndCallers: localFP SP: localSP]
				ifFalse: [self shortPrintFrameAndCallers: thePage headFrameFP SP: thePage headFrameSP]."
			 self printCurrentBytecodeOn: Transcript.
			 Transcript cr; flush].

		 self dispatchOn: currentBytecode in: BytecodeTable.
		 self incrementByteCount.
		 byteCount = breakCount ifTrue:
			["printFrameAtEachStep := true."
			 printSends := printBytecodeAtEachStep := true.
			 self halt: 'hit breakCount break-point']].
	self externalizeIPandSP! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem (objmem refactor) 6/8/2010 17:12'!
testBecome
	"Become some young things.  AA testBecome    "
	| array list1 list2 p1 p2 p3 p4 |
	array := objectMemory splObj: ClassArray.
	list1 := objectMemory instantiateClass: array indexableSize: 2.
	list2 := objectMemory instantiateClass: array indexableSize: 2.
	p1 := objectMemory instantiateClass: (objectMemory splObj: ClassPoint) indexableSize: 0.
	self push: p1.
	objectMemory storePointer: 0 ofObject: list1 withValue: p1.
	p2 := objectMemory instantiateClass: (objectMemory splObj: ClassPoint) indexableSize: 0.
	self push: p2.
	objectMemory storePointer: 1 ofObject: list1 withValue: p2.
	p3 := objectMemory instantiateClass: (objectMemory splObj: ClassMessage) indexableSize: 0.
	self push: p3.
	objectMemory storePointer: 0 ofObject: list2 withValue: p3.
	p4 := objectMemory instantiateClass: (objectMemory splObj: ClassMessage) indexableSize: 0.
	self push: p4.
	objectMemory storePointer: 1 ofObject: list2 withValue: p4.
	(objectMemory become: list1 with: list2 twoWay: true copyHash: true) ifFalse: [self error: 'failed'].
	self popStack = p2 ifFalse: [self halt].
	self popStack = p1 ifFalse: [self halt].
	self popStack = p4 ifFalse: [self halt].
	self popStack = p3 ifFalse: [self halt].
	(objectMemory fetchPointer: 0 ofObject: list1) = p3 ifFalse: [self halt].
	(objectMemory fetchPointer: 1 ofObject: list1) = p4 ifFalse: [self halt].
	(objectMemory fetchPointer: 0 ofObject: list2) = p1 ifFalse: [self halt].
	(objectMemory fetchPointer: 1 ofObject: list2) = p2 ifFalse: [self halt].! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 12/29/2016 08:25'!
testBreakCount: breakCount printSends: shouldPrintSends printFrames: shouldPrintFrames printBytecodes: shouldPrintBytecodes
	self initStackPages.
	self loadInitialContext.
	transcript clear.
	byteCount := 0.
	quitBlock := [^self close].
	printSends := true & shouldPrintSends. "true & foo allows evaluating printFoo := true in the debugger"
	printFrameAtEachStep := true & shouldPrintFrames.
	printBytecodeAtEachStep := true & shouldPrintBytecodes.
	self internalizeIPandSP.
	self fetchNextBytecode.
	[true] whileTrue:
		[self assertValidExecutionPointers.
		 printFrameAtEachStep ifTrue:
			[self printFrame: localFP WithSP: localSP].
		 printBytecodeAtEachStep ifTrue:
			[self printCurrentBytecodeOn: Transcript.
			 Transcript cr; flush].
		 self dispatchOn: currentBytecode in: BytecodeTable.
		 self incrementByteCount.
		 byteCount = breakCount ifTrue:
			["printFrameAtEachStep :=" printBytecodeAtEachStep := true.
			 self halt: 'hit breakCount break-point']].
	self externalizeIPandSP! !

!StackInterpreterSimulator methodsFor: 'testing' stamp: 'eem 6/25/2008 12:15'!
testWithFramePrint
	printFrameAtEachStep := printBytecodeAtEachStep := true.
	self test! !

!StackInterpreterSimulator methodsFor: 'accessing' stamp: 'eem 10/12/2009 01:05'!
transcript
	^transcript! !

!StackInterpreterSimulator methodsFor: 'accessing' stamp: 'eem 7/18/2010 18:28'!
transcript: aTranscript
	transcript := aTranscript! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 7/1/2008 13:27'!
transferTo: aProc
	"self halt."
	^super transferTo: aProc! !

!StackInterpreterSimulator methodsFor: 'accessing' stamp: 'eem 12/22/2017 10:57'!
turnOnPrimTraceLog
	primTraceLog ifNil:
		[primTraceLog := OrderedCollection new: 512]! !

!StackInterpreterSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:00'!
unableToReadImageError
	self error:  'Read failed or premature end of image file'! !

!StackInterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 7/6/2008 22:36'!
updateStateOfSpouseContextForFrame: theFP WithSP: theSP
	"(self frameContext: theFP) = 22163268 ifTrue:
		[self halt]."
	^super updateStateOfSpouseContextForFrame: theFP WithSP: theSP! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'EstebanLorenzano 12/7/2015 12:49'!
veryDeepCopyWith: deepCopier
	"Override to short-circuit the copying of any VMPluginCodeGenerators referenced from mappedPluginEntries.
	 These can in turn hold onto Monticello state, resulting in a huge ammount of unnecessary copying."
	self objectMemory hasSpurMemoryManagerAPI ifTrue: 
		[deepCopier references
			at: mappedPluginEntries ifAbsentPut: [mappedPluginEntries].
		mappedPluginEntries do:
			[:tuple| | block |
			block := tuple third.
			deepCopier references at: block ifAbsentPut: [block]]].
	^super veryDeepCopyWith: deepCopier! !

!StackInterpreterSimulator methodsFor: 'file primitives' stamp: 'eem 9/7/2018 19:43'!
vmPathGet: stringBase Length: stringSize
	| pathName stringOop |
	pathName := self vmPath.
	stringOop := stringBase - objectMemory baseHeaderSize. "Due to C call in Interp"
	1 to: stringSize do:
		[:i | objectMemory storeByte: i-1 ofObject: stringOop
			withValue: (pathName at: i) asciiValue].
! !

!StackInterpreterSimulator methodsFor: 'debug support' stamp: 'eem 10/12/2009 01:03'!
warning: aString
	transcript cr; nextPutAll: aString; flush! !

!StackInterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'yo 6/3/1999 16:10'!
warpBits

	^ myBitBlt warpBits! !

!StackInterpreterSimulator methodsFor: 'primitive support' stamp: 'eem 12/29/2016 16:45'!
windowIsClosing
	quitBlock ifNotNil:
		[:effectiveQuitBlock|
		quitBlock := nil. "stop recursion on explicit window close."
		[effectiveQuitBlock value]
			on: BlockCannotReturn
			do: [:ex|]]	"Cause return from #test, et al"! !

!StackInterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 11/18/2016 17:03'!
withMemoryProtectionDo: aBlock
	objectMemory writeProtectMemory.
	stackPages writeProtectMemory.
	^aBlock ensure:
		[objectMemory writeEnableMemory.
		 stackPages writeEnableMemory]! !

!StackInterpreterSimulator methodsFor: 'frame access' stamp: 'nice 3/15/2016 23:38'!
withoutSmallIntegerTags: anInteger
	| signBit |
	self assert: (anInteger >= 0 and: [anInteger highBit <= (objectMemory wordSize * 8)]).
	signBit := 1 << (objectMemory wordSize * 8 - 1).
	^(anInteger bitAnd: signBit) ~= 0
		ifTrue: ["negative"
				(anInteger bitClear: signBit + 1) - signBit]
		ifFalse: ["positive"
				anInteger - 1]! !

StackInterpreterSimulator subclass: #StackInterpreterSimulatorLSB
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!StackInterpreterSimulatorLSB commentStamp: '<historical>' prior: 0!
This class overrides a few methods in StackInterpreterSimulator required for simulation to work on little-endian architectures (such as the x86 family of processors).  To start it up simply use StackInterpreterSimulatorLSB instead of StackInterpreterSimulator (see the class comment there for more details).  For example:

	(StackInterpreterSimulatorLSB new openOn: Smalltalk imageName) test

Note that the image must have been saved at least once on the local architecture, since the compiled VM performs some byte swapping that the simulator cannot cope with.!

!StackInterpreterSimulatorLSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:59'!
halfWordHighInLong32: long32
	"Used by Balloon"

	^ long32 bitAnd: 16rFFFF! !

!StackInterpreterSimulatorLSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:59'!
halfWordLowInLong32: long32
	"Used by Balloon"

	^ long32 bitShift: -16! !

!StackInterpreterSimulatorLSB methodsFor: 'initialization' stamp: 'eem 10/9/2008 16:39'!
nextShortFrom: aStream
	"Read a 16-bit quantity from the given (binary) stream."
	^aStream nextLittleEndianNumber: 2! !

!StackInterpreterSimulatorLSB methodsFor: 'memory access' stamp: 'tpr 12/22/2005 17:46'!
vmEndianness
	"return 0 for little endian, 1 for big endian"
	^0! !

VMClass subclass: #NSSendCacheSurrogate
	instanceVariableNames: 'address memory cogit'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!NSSendCacheSurrogate methodsFor: 'as yet unclassified' stamp: 'rmacnak 3/27/2015 19:40'!
at: anAddress objectMemory: objectMemory cogit: aCogit
	address := anAddress.
	memory := objectMemory memory.
	cogit := aCogit! !

NSSendCacheSurrogate subclass: #NSSendCacheSurrogate32
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!NSSendCacheSurrogate32 class methodsFor: 'accessing' stamp: 'rmacnak 4/4/2015 15:27'!
alignedByteSize
	^24! !

!NSSendCacheSurrogate32 methodsFor: 'accessing generated' stamp: 'rmacnak 4/4/2015 15:27'!
classTag
	^memory unsignedLongAt: address + 1! !

!NSSendCacheSurrogate32 methodsFor: 'accessing generated' stamp: 'rmacnak 4/4/2015 15:27'!
classTag: aValue
	^memory
		unsignedLongAt: address + 1
		put: aValue! !

!NSSendCacheSurrogate32 methodsFor: 'accessing generated' stamp: 'rmacnak 4/4/2015 15:27'!
depth
	^memory unsignedLongAt: address + 21! !

!NSSendCacheSurrogate32 methodsFor: 'accessing generated' stamp: 'rmacnak 4/4/2015 15:27'!
depth: aValue
	^memory
		unsignedLongAt: address + 21
		put: aValue! !

!NSSendCacheSurrogate32 methodsFor: 'accessing generated' stamp: 'rmacnak 4/4/2015 15:27'!
enclosingObject
	^memory unsignedLongAt: address + 5! !

!NSSendCacheSurrogate32 methodsFor: 'accessing generated' stamp: 'rmacnak 4/4/2015 15:27'!
enclosingObject: aValue
	^memory
		unsignedLongAt: address + 5
		put: aValue! !

!NSSendCacheSurrogate32 methodsFor: 'accessing generated' stamp: 'rmacnak 4/4/2015 15:27'!
numArgs
	^memory unsignedLongAt: address + 17! !

!NSSendCacheSurrogate32 methodsFor: 'accessing generated' stamp: 'rmacnak 4/4/2015 15:27'!
numArgs: aValue
	^memory
		unsignedLongAt: address + 17
		put: aValue! !

!NSSendCacheSurrogate32 methodsFor: 'accessing generated' stamp: 'rmacnak 4/4/2015 15:27'!
selector
	^memory unsignedLongAt: address + 13! !

!NSSendCacheSurrogate32 methodsFor: 'accessing generated' stamp: 'rmacnak 4/4/2015 15:27'!
selector: aValue
	^memory
		unsignedLongAt: address + 13
		put: aValue! !

!NSSendCacheSurrogate32 methodsFor: 'accessing generated' stamp: 'rmacnak 4/4/2015 15:27'!
target
	^memory unsignedLongAt: address + 9! !

!NSSendCacheSurrogate32 methodsFor: 'accessing generated' stamp: 'rmacnak 4/4/2015 15:27'!
target: aValue
	^memory
		unsignedLongAt: address + 9
		put: aValue! !

NSSendCacheSurrogate subclass: #NSSendCacheSurrogate64
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!NSSendCacheSurrogate64 class methodsFor: 'accessing' stamp: 'rmacnak 4/4/2015 15:27'!
alignedByteSize
	^48! !

!NSSendCacheSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
classTag
	^memory unsignedLong64At: address + 1! !

!NSSendCacheSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
classTag: aValue
	^memory
		unsignedLong64At: address + 1
		put: aValue! !

!NSSendCacheSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
depth
	^memory unsignedLong64At: address + 41! !

!NSSendCacheSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
depth: aValue
	^memory
		unsignedLong64At: address + 41
		put: aValue! !

!NSSendCacheSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
enclosingObject
	^memory unsignedLong64At: address + 9! !

!NSSendCacheSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
enclosingObject: aValue
	^memory
		unsignedLong64At: address + 9
		put: aValue! !

!NSSendCacheSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
numArgs
	^memory unsignedLong64At: address + 33! !

!NSSendCacheSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
numArgs: aValue
	^memory
		unsignedLong64At: address + 33
		put: aValue! !

!NSSendCacheSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
selector
	^memory unsignedLong64At: address + 25! !

!NSSendCacheSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
selector: aValue
	^memory
		unsignedLong64At: address + 25
		put: aValue! !

!NSSendCacheSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
target
	^memory unsignedLong64At: address + 17! !

!NSSendCacheSurrogate64 methodsFor: 'accessing generated' stamp: 'eem 9/28/2015 21:00'!
target: aValue
	^memory
		unsignedLong64At: address + 17
		put: aValue! !

VMClass subclass: #ObjectMemory
	instanceVariableNames: 'memory youngStart endOfMemory memoryLimit nilObj falseObj trueObj specialObjectsOop rootTable rootTableCount rootTableOverflowed extraRoots extraRootCount weakRoots weakRootCount child field parentField freeBlock lastHash allocationCount lowSpaceThreshold signalLowSpace compStart compEnd fwdTableNext fwdTableLast remapBuffer remapBufferCount allocationsBetweenGCs tenuringThreshold gcBiasToGrow gcBiasToGrowGCLimit gcBiasToGrowThreshold statFullGCs statIncrGCs statFullGCUsecs statIncrGCUsecs statGCEndTime statIGCDeltaUsecs statTenures statRootTableOverflows freeContexts freeLargeContexts interruptCheckCounter totalObjectCount shrinkThreshold growHeadroom headerTypeBytes youngStartLocal statMarkCount statMarkCountLocal statSweepCount statMkFwdCount statCompMoveCount statGrowMemory statShrinkMemory statRootTableCount statAllocationCount statSurvivorCount statSpecialMarkCount forceTenureFlag gcStartUsecs oldSpaceStart'
	classVariableNames: 'AllButHashBits AllButImmutabilityBit AllButMarkBit AllButMarkBitAndTypeMask AllButRootBit Done ExtraRootSize GCTopMarker HashBits HeaderTypeClass HeaderTypeExtraBytes HeaderTypeFree HeaderTypeGC LongSizeNumBits NilContext RemapBufferSize RootTableRedZone RootTableSize StartField StartObj Upward WeakRootTableSize WordMask'
	poolDictionaries: 'VMBasicConstants VMBytecodeConstants VMObjectIndices VMSqueakClassIndices VMSqueakV3ObjectRepresentationConstants'
	category: 'VMMaker-Interpreter'!

!ObjectMemory commentStamp: '<historical>' prior: 0!
This class describes a 32-bit direct-pointer object memory for Smalltalk.  The model is very simple in principle:  a pointer is either a SmallInteger or a 32-bit direct object pointer.

SmallIntegers are tagged with a low-order bit equal to 1, and an immediate 31-bit 2s-complement signed value in the rest of the word.

All object pointers point to a header, which may be followed by a number of data fields.  This object memory achieves considerable compactness by using a variable header size (the one complexity of the design).  The format of the 0th header word is as follows:

	3 bits	reserved for gc (mark, root, unused)
	12 bits	object hash (for HashSets)
	5 bits	compact class index
	4 bits	object format
	6 bits	object size in 32-bit words
	2 bits	header type (0: 3-word, 1: 2-word, 2: forbidden, 3: 1-word)

If a class is in the compact class table, then this is the only header information needed.  If it is not, then it will have another header word at offset -4 bytes with its class in the high 30 bits, and the header type repeated in its low 2 bits.  It the objects size is greater than 255 bytes, then it will have yet another header word at offset -8 bytes with its full word size in the high 30 bits and its header type repeated in the low two bits.

The object format field provides the remaining information as given in the formatOf: method (including isPointers, isVariable, isBytes, and the low 2 size bits of byte-sized objects).

This implementation includes incremental (2-generation) and full garbage collection, each with compaction and rectification of direct pointers.  It also supports a bulk-become (exchange object identity) feature that allows many objects to be becomed at once, as when all instances of a class must be grown or shrunk.

There is now a simple 64-bit version of the object memory.  It is the simplest possible change that could work.  It merely sign-extends all integer oops, and extends all object headers and oops by adding 32 zeroes in the high bits.  The format of the base header word is changed in one minor, not especially elegant, way.  Consider the old 32-bit header:
	ggghhhhhhhhhhhhcccccffffsssssstt
The 64-bit header is almost identical, except that the size field (now being in units of 8 bytes, has a zero in its low-order bit.  At the same time, the byte-size residue bits for byte objects, which are in the low order bits of formats 8-11 and 12-15, are now in need of another bit of residue.  So, the change is as follows:
	ggghhhhhhhhhhhhcccccffffsssssrtt
where bit r supplies the 4's bit of the byte size residue for byte objects.  Oh, yes, this is also needed now for 'variableWord' objects, since their size in 32-bit words requires a low-order bit.

See the comment in formatOf: for the change allowing for 64-bit wide bitmaps, now dubbed 'variableLong'.!

ObjectMemory subclass: #Interpreter
	instanceVariableNames: 'activeContext theHomeContext method receiver instructionPointer stackPointer localIP localSP localHomeContext localReturnContext localReturnValue messageSelector argumentCount newMethod currentBytecode successFlag primitiveIndex primitiveFunctionPointer methodCache atCache lkupClass reclaimableContextCount nextPollTick nextWakeupTick lastTick interruptKeycode interruptPending semaphoresToSignalA semaphoresUseBufferA semaphoresToSignalCountA semaphoresToSignalB semaphoresToSignalCountB savedWindowSize fullScreenFlag deferDisplayUpdates pendingFinalizationSignals compilerInitialized extraVMMemory receiverClass interpreterProxy showSurfaceFn interruptCheckCounterFeedBackReset interruptChecksEveryNms externalPrimitiveTable primitiveTable globalSessionID jmpBuf jmpDepth jmpMax suspendedCallbacks suspendedMethods profileProcess profileMethod profileSemaphore nextProfileTick metaclassSizeBits statIOProcessEvents statCheckForEvents statQuickCheckForEvents statProcessSwitch statPendingFinalizationSignals gcSemaphoreIndex'
	classVariableNames: 'AtCacheEntries AtCacheFixedFields AtCacheFmt AtCacheMask AtCacheOop AtCacheSize AtCacheTotalSize AtPutBase BlockArgumentCountIndex BlockMethodIndex BytecodeTable CacheProbeMax CallerIndex ClosureMethodIndex CompilerHooksSize CrossedX DirBadPath DirEntryFound DirNoMoreEntries DoBalanceChecks EndOfRun HomeIndex InitialIPIndex JitterTable MaxExternalPrimitiveTableSize MaxJumpBuf MaxPrimitiveIndex MillisecondClockMask PrimitiveExternalCallIndex PrimitiveTable SemaphoresToSignalSize TempFrameStart'
	poolDictionaries: 'VMBasicConstants VMMethodCacheConstants VMObjectIndices'
	category: 'VMMaker-Interpreter'!

!Interpreter commentStamp: '<historical>' prior: 0!
This class is a complete implementation of the Smalltalk-80 virtual machine, derived originally from the Blue Book specification but quite different in some areas.

It has been modernized with 32-bit pointers, better management of Contexts, and attention to variable use that allows the CCodeGenerator (qv) to keep, eg, the instruction pointer and stack pointer in registers as well as keeping most simple variables in a global array that seems to improve performance for most platforms.

In addition to SmallInteger arithmetic and Floats, it supports logic on 32-bit PositiveLargeIntegers, thus allowing it to simulate itself much more effectively than would otherwise be the case.

NOTE:  Here follows a list of things to be borne in mind when working on this code, or when making changes for the future.

1.  There are a number of things that should be done the next time we plan to release a copletely incompatible image format.  These include unifying the instanceSize field of the class format word -- see instantiateClass:indexableSize:, and unifying the bits of the method primitive index (if we decide we need more than 512, after all) -- see primitiveIndexOf:.  Also, contexts should be given a special format code (see next item).

2.  There are several fast checks for contexts (see isContextHeader: and isMethodContextHeader:) which will fail if the compact class indices of BlockContext or MethodContext change.  This is necessary because the oops may change during a compaction when the oops are being adjusted.  It's important to be aware of this when writing a new image using the systemTracer.  A better solution would be to reserve one of the format codes for Contexts only.

3.  We have made normal files tolerant to size and positions up to 32 bits.  This has not been done for async files, since they are still experimental.  The code in size, at: and at:put: should work with sizes and indices up to 31 bits, although I have not tested it (di 12/98); it might or might not work with 32-bit sizes.

4.  Note that 0 is used in a couple of places as an impossible oop.  This should be changed to a constant that really is impossible (or perhaps there is code somewhere that guarantees it --if so it should be put in this comment).  The places include the method cache and the at cache. !

!Interpreter class methodsFor: 'translation' stamp: 'eem 2/10/2009 08:39'!
additionalSelectorTables
	^{ self primitiveTable }! !

!Interpreter class methodsFor: 'constants' stamp: ''!
bytecodeTable

	^ BytecodeTable! !

!Interpreter class methodsFor: 'translation' stamp: 'nice 7/12/2016 14:03'!
declareCVarsIn: aCCodeGenerator
	aCCodeGenerator addHeaderFile:'<setjmp.h>'.
	self declareInterpreterVersionIn: aCCodeGenerator
		defaultName: 'Context'.
	aCCodeGenerator 
		var: #interpreterProxy 
		type: #'struct VirtualMachine*'.
	aCCodeGenerator
		var: #primitiveTable
		declareC: 'void *primitiveTable[MaxPrimitiveIndex + 2 /* ', (MaxPrimitiveIndex +2) printString, ' */] = ',	self primitiveTableString.
	aCCodeGenerator
		var: #primitiveFunctionPointer
		declareC: 'void *primitiveFunctionPointer'.	"xxxx FIX THIS STUPIDITY xxxx - ikp. What he means is use a better type than void *, apparently - tpr"
	aCCodeGenerator
		var: #methodCache
		declareC: 'sqIntptr_t methodCache[MethodCacheSize + 1 /* ', (MethodCacheSize + 1) printString, ' */]'.
	aCCodeGenerator
		var: #atCache
		declareC: 'sqInt atCache[AtCacheTotalSize + 1 /* ', (AtCacheTotalSize + 1) printString, ' */]'.
	aCCodeGenerator var: #localIP type: #'char*'.
	aCCodeGenerator var: #localSP type: #'char*'.
	aCCodeGenerator var: #showSurfaceFn type: #'void*'.
	aCCodeGenerator var: 'semaphoresToSignalA'
		declareC: 'sqInt semaphoresToSignalA[SemaphoresToSignalSize + 1 /* ', (SemaphoresToSignalSize + 1) printString, ' */]'.
	aCCodeGenerator var: 'semaphoresToSignalB'
		declareC: 'sqInt semaphoresToSignalB[SemaphoresToSignalSize + 1 /* ', (SemaphoresToSignalSize + 1) printString, ' */]'.
	aCCodeGenerator
		var: #compilerHooks
		declareC: 'sqInt (*compilerHooks[CompilerHooksSize + 1 /* ', (CompilerHooksSize + 1) printString, ' */])()'.
	aCCodeGenerator
		var: #externalPrimitiveTable
		declareC: 'void *externalPrimitiveTable[MaxExternalPrimitiveTableSize + 1 /* ', (MaxExternalPrimitiveTableSize + 1) printString, ' */]'.

	aCCodeGenerator
		var: #jmpBuf
		declareC: 'jmp_buf jmpBuf[MaxJumpBuf + 1 /* ', (MaxJumpBuf + 1) printString, ' */]'.
	aCCodeGenerator
		var: #suspendedCallbacks
		declareC: 'usqInt suspendedCallbacks[MaxJumpBuf + 1 /* ', (MaxJumpBuf + 1) printString, ' */]'.
	aCCodeGenerator
		var: #suspendedMethods
		declareC: 'usqInt suspendedMethods[MaxJumpBuf + 1 /* ', (MaxJumpBuf + 1) printString, ' */]'.

	self declareCAsOop: {
		#instructionPointer. 
		#method. 
		#newMethod. 
		#activeContext. 
		#theHomeContext. 
		#stackPointer
	} in: aCCodeGenerator.

	aCCodeGenerator var: #nextProfileTick type: #sqLong.
! !

!Interpreter class methodsFor: 'initialization' stamp: ''!
initializeAssociationIndex
	ValueIndex := 1! !

!Interpreter class methodsFor: 'initialization' stamp: 'eem 6/16/2008 10:07'!
initializeBytecodeTable
	"Interpreter initializeBytecodeTable"
	"Note: This table will be used to generate a C switch statement."

	BytecodeTable := Array new: 256.
	self table: BytecodeTable from:
	#(
		(  0  15 pushReceiverVariableBytecode)
		( 16  31 pushTemporaryVariableBytecode)
		( 32  63 pushLiteralConstantBytecode)
		( 64  95 pushLiteralVariableBytecode)
		( 96 103 storeAndPopReceiverVariableBytecode)
		(104 111 storeAndPopTemporaryVariableBytecode)
		(112 pushReceiverBytecode)
		(113 pushConstantTrueBytecode)
		(114 pushConstantFalseBytecode)
		(115 pushConstantNilBytecode)
		(116 pushConstantMinusOneBytecode)
		(117 pushConstantZeroBytecode)
		(118 pushConstantOneBytecode)
		(119 pushConstantTwoBytecode)
		(120 returnReceiver)
		(121 returnTrue)
		(122 returnFalse)
		(123 returnNil)
		(124 returnTopFromMethod)
		(125 returnTopFromBlock)

		(126 127 unknownBytecode)

		(128 extendedPushBytecode)
		(129 extendedStoreBytecode)
		(130 extendedStoreAndPopBytecode)
		(131 singleExtendedSendBytecode)
		(132 doubleExtendedDoAnythingBytecode)
		(133 singleExtendedSuperBytecode)
		(134 secondExtendedSendBytecode)
		(135 popStackBytecode)
		(136 duplicateTopBytecode)

		(137 pushActiveContextBytecode)
		(138 pushNewArrayBytecode)
		(139 unknownBytecode)
		(140 pushRemoteTempLongBytecode)
		(141 storeRemoteTempLongBytecode)
		(142 storeAndPopRemoteTempLongBytecode)
		(143 pushClosureCopyCopiedValuesBytecode)

		(144 151 shortUnconditionalJump)
		(152 159 shortConditionalJump)
		(160 167 longUnconditionalJump)
		(168 171 longJumpIfTrue)
		(172 175 longJumpIfFalse)

		"176-191 were sendArithmeticSelectorBytecode"
		(176 bytecodePrimAdd)
		(177 bytecodePrimSubtract)
		(178 bytecodePrimLessThan)
		(179 bytecodePrimGreaterThan)
		(180 bytecodePrimLessOrEqual)
		(181 bytecodePrimGreaterOrEqual)
		(182 bytecodePrimEqual)
		(183 bytecodePrimNotEqual)
		(184 bytecodePrimMultiply)
		(185 bytecodePrimDivide)
		(186 bytecodePrimMod)
		(187 bytecodePrimMakePoint)
		(188 bytecodePrimBitShift)
		(189 bytecodePrimDiv)
		(190 bytecodePrimBitAnd)
		(191 bytecodePrimBitOr)	

		"192-207 were sendCommonSelectorBytecode"
		(192 bytecodePrimAt)
		(193 bytecodePrimAtPut)
		(194 bytecodePrimSize)
		(195 bytecodePrimNext)
		(196 bytecodePrimNextPut)
		(197 bytecodePrimAtEnd)
		(198 bytecodePrimEquivalent)
		(199 bytecodePrimClass)
		(200 bytecodePrimBlockCopy)
		(201 bytecodePrimValue)
		(202 bytecodePrimValueWithArg)
		(203 bytecodePrimDo)
		(204 bytecodePrimNew)
		(205 bytecodePrimNewWithArg)
		(206 bytecodePrimPointX)
		(207 bytecodePrimPointY)

		(208 255 sendLiteralSelectorBytecode)
	).! !

!Interpreter class methodsFor: 'initialization' stamp: 'eem 5/27/2008 17:54'!
initializeCaches

	| atCacheEntrySize |
	MethodCacheEntries := 512. 
	MethodCacheSelector := 1.
	MethodCacheClass := 2.
	MethodCacheMethod := 3.
	MethodCachePrim := 4.
	MethodCachePrimFunction := 5.
	MethodCacheEntrySize := 8.  "Must be power of two for masking scheme."
	MethodCacheMask := (MethodCacheEntries - 1) * MethodCacheEntrySize.
	MethodCacheSize := MethodCacheEntries * MethodCacheEntrySize.
	CacheProbeMax := 3.

	AtCacheEntries := 8.  "Must be a power of two"
	AtCacheOop := 1.
	AtCacheSize := 2.
	AtCacheFmt := 3.
	AtCacheFixedFields := 4.
	atCacheEntrySize := 4.  "Must be power of two for masking scheme."
	AtCacheMask := (AtCacheEntries-1) * atCacheEntrySize.
	AtPutBase := AtCacheEntries * atCacheEntrySize.
	AtCacheTotalSize := AtCacheEntries * atCacheEntrySize * 2.
! !

!Interpreter class methodsFor: 'initialization' stamp: ''!
initializeCharacterIndex
	CharacterValueIndex := 0! !

!Interpreter class methodsFor: 'initialization' stamp: 'tpr 4/29/2003 12:10'!
initializeCharacterScannerIndices
	CrossedX := 258.
	EndOfRun := 257
! !

!Interpreter class methodsFor: 'initialization' stamp: ''!
initializeClassIndices
	"Class Class"
	SuperclassIndex := 0.
	MethodDictionaryIndex := 1.
	InstanceSpecificationIndex := 2.
	"Fields of a message dictionary"
	MethodArrayIndex := 1.
	SelectorStart := 2! !

!Interpreter class methodsFor: 'initialization' stamp: 'ikp 10/27/2000 14:47'!
initializeCompilerHooks
	"Interpreter initializeCompilerHooks"

	"compilerHooks[] indices:
	1	void compilerTranslateMethodHook(void)
	2	void compilerFlushCacheHook(CompiledMethod *oldMethod)
	3	void compilerPreGCHook(int fullGCFlag)
	4	void compilerMapHook(int memStart, int memEnd)
	5	void compilerPostGCHook(void)
	6	void compilerProcessChangeHook(void)
	7	void compilerPreSnapshotHook(void)
	8	void compilerPostSnapshotHook(void)
	9	void compilerMarkHook(void)
	10	void compilerActivateMethodHook(void)
	11	void compilerNewActiveContextHook(int sendFlag)
	12	void compilerGetInstructionPointerHook(void)
	13	void compilerSetInstructionPointerHook(void)
	14	void compilerCreateActualMessageHook(void)"

	CompilerHooksSize := 15.! !

!Interpreter class methodsFor: 'initialization' stamp: 'eem 10/10/2009 22:25'!
initializeContextIndices
	"Class MethodContext"
	| contextFixedSizePlusHeader |
	SenderIndex := 0.
	InstructionPointerIndex := 1.
	StackPointerIndex := 2.
	MethodIndex := 3.
	ClosureIndex := 4. "N.B. Called receiverMap in the image."
	ReceiverIndex := 5.
	TempFrameStart := 6.  "Note this is in two places!!"

	"Class BlockContext"
	CallerIndex := 0.
	BlockArgumentCountIndex := 3.
	InitialIPIndex := 4.
	HomeIndex := 5.

	"Class BlockClosure"
	ClosureOuterContextIndex := 0.
	ClosureStartPCIndex := 1.
	ClosureNumArgsIndex := 2.
	ClosureFirstCopiedValueIndex := 3.

	CtxtTempFrameStart := 6.  "Copy of TempFrameStart in Interp"
	contextFixedSizePlusHeader := CtxtTempFrameStart + 1.
	SmallContextSize := contextFixedSizePlusHeader + 16 * BytesPerWord.  "16 indexable fields"
	"Large contexts have 56 indexable fileds.  Max with single header word."
	"However note that in 64 bits, for now, large contexts have 3-word headers"
	LargeContextSize := contextFixedSizePlusHeader + 56 * BytesPerWord! !

!Interpreter class methodsFor: 'initialization' stamp: ''!
initializeDirectoryLookupResultCodes

	DirEntryFound := 0.
	DirNoMoreEntries := 1.
	DirBadPath := 2.! !

!Interpreter class methodsFor: 'initialization' stamp: 'di 3/25/1999 22:09'!
initializeMessageIndices
	MessageSelectorIndex := 0.
	MessageArgumentsIndex := 1.
	MessageLookupClassIndex := 2.! !

!Interpreter class methodsFor: 'initialization' stamp: 'eem 10/10/2009 22:23'!
initializeMethodIndices
	"Class CompiledMethod"
	HeaderIndex := 0.
	LiteralStart := 1.

	LargeContextBit := 16r40000.  "This bit set in method headers if large context is needed."! !

!Interpreter class methodsFor: 'initialization' stamp: 'eem 6/7/2018 15:19'!
initializeMiscConstants

	super initializeMiscConstants.

	DoBalanceChecks := InitializationOptions at: #DoBalanceChecks ifAbsent: [false]. "generate stack balance checks"
	SemaphoresToSignalSize := 500.
	PrimitiveExternalCallIndex := 117. "Primitive index for #primitiveExternalCall"
	MillisecondClockMask := 16r1FFFFFFF.
	"Note: The external primitive table should actually be dynamically sized but for the sake of inferior platforms (e.g., Mac :-) who cannot allocate memory in any reasonable way, we keep it static (and cross our fingers...)"
	MaxExternalPrimitiveTableSize := 4096. "entries"

	MaxJumpBuf := 32. "max. callback depth"! !

!Interpreter class methodsFor: 'initialization' stamp: ''!
initializePointIndices
	XIndex := 0.
	YIndex := 1! !

!Interpreter class methodsFor: 'initialization' stamp: 'eem 4/15/2020 16:22'!
initializePrimitiveTable 
	"This table generates a C function address table use in primitiveResponse along with dispatchFunctionPointerOn:in:"

	"NOTE: The real limit here is 2047 because of the method header layout but there is no point in going over the needed size"
	MaxPrimitiveIndex := 575.
	PrimitiveTable := Array new: MaxPrimitiveIndex + 1.
	self table: PrimitiveTable from: 
	#(	"Integer Primitives (0-19)"
		(0 primitiveFail)
		(1 primitiveAdd)
		(2 primitiveSubtract)
		(3 primitiveLessThan)
		(4 primitiveGreaterThan)
		(5 primitiveLessOrEqual)
		(6 primitiveGreaterOrEqual)
		(7 primitiveEqual)
		(8 primitiveNotEqual)
		(9 primitiveMultiply)
		(10 primitiveDivide)
		(11 primitiveMod)
		(12 primitiveDiv)
		(13 primitiveQuo)
		(14 primitiveBitAnd)
		(15 primitiveBitOr)
		(16 primitiveBitXor)
		(17 primitiveBitShift)
		(18 primitiveMakePoint)
		(19 primitiveFail)					"Guard primitive for simulation -- *must* fail"

		"LargeInteger Primitives (20-37)"
		(20 primitiveRemLargeIntegers)
		(21 primitiveAddLargeIntegers)
		(22 primitiveSubtractLargeIntegers)
		(23 primitiveLessThanLargeIntegers)
		(24 primitiveGreaterThanLargeIntegers)
		(25 primitiveLessOrEqualLargeIntegers)
		(26 primitiveGreaterOrEqualLargeIntegers)
		(27 primitiveEqualLargeIntegers)
		(28 primitiveNotEqualLargeIntegers)
		(29 primitiveMultiplyLargeIntegers)
		(30 primitiveDivideLargeIntegers)
		(31 primitiveModLargeIntegers)
		(32 primitiveDivLargeIntegers)
		(33 primitiveQuoLargeIntegers)
		(34 primitiveBitAndLargeIntegers)
		(35 primitiveBitOrLargeIntegers)
		(36 primitiveBitXorLargeIntegers)
		(37 primitiveBitShiftLargeIntegers)

		"Float Primitives (38-59)"
		(38 primitiveFloatAt)
		(39 primitiveFloatAtPut)
		(40 primitiveAsFloat)
		(41 primitiveFloatAdd)
		(42 primitiveFloatSubtract)
		(43 primitiveFloatLessThan)
		(44 primitiveFloatGreaterThan)
		(45 primitiveFloatLessOrEqual)
		(46 primitiveFloatGreaterOrEqual)
		(47 primitiveFloatEqual)
		(48 primitiveFloatNotEqual)
		(49 primitiveFloatMultiply)
		(50 primitiveFloatDivide)
		(51 primitiveTruncated)
		(52 primitiveFractionalPart)
		(53 primitiveExponent)
		(54 primitiveTimesTwoPower)
		(55 primitiveSquareRoot)
		(56 primitiveSine)
		(57 primitiveArctan)
		(58 primitiveLogN)
		(59 primitiveExp)

		"Subscript and Stream Primitives (60-67)"
		(60 primitiveAt)
		(61 primitiveAtPut)
		(62 primitiveSize)
		(63 primitiveStringAt)
		(64 primitiveStringAtPut)
		(65 primitiveFail) "was primitiveNext which no longer pays its way (normal Smalltalk code is faster)"
		(66 primitiveFail) "was primitiveNextPut which no longer pays its way (normal Smalltalk code is faster)"
		(67 primitiveFail) "was primitiveAtEnd which no longer pays its way (normal Smalltalk code is faster)"

		"StorageManagement Primitives (68-79)"
		(68 primitiveObjectAt)
		(69 primitiveObjectAtPut)
		(70 primitiveNew)
		(71 primitiveNewWithArg)
		(72 primitiveArrayBecomeOneWay)	"Blue Book: primitiveBecome"
		(73 primitiveInstVarAt)
		(74 primitiveInstVarAtPut)
		(75 primitiveAsOop)
		(76 primitiveStoreStackp)					"Blue Book: primitiveAsObject"
		(77 primitiveSomeInstance)
		(78 primitiveNextInstance)
		(79 primitiveNewMethod)

		"Control Primitives (80-89)"
		(80 primitiveBlockCopy)
		(81 primitiveValue)
		(82 primitiveValueWithArgs)
		(83 primitivePerform)
		(84 primitivePerformWithArgs)
		(85 primitiveSignal)
		(86 primitiveWait)
		(87 primitiveResume)
		(88 primitiveSuspend)
		(89 primitiveFlushCache)

		"Input/Output Primitives (90-109)"
		(90 primitiveMousePoint)
		(91 primitiveTestDisplayDepth)			"Blue Book: primitiveCursorLocPut"
		(92 primitiveSetDisplayMode)				"Blue Book: primitiveCursorLink"
		(93 primitiveInputSemaphore)
		(94 primitiveGetNextEvent)				"Blue Book: primitiveSampleInterval"
		(95 primitiveInputWord)
		(96 primitiveFail)	"primitiveCopyBits"
		(97 primitiveSnapshot)
		(98 primitiveStoreImageSegment)
		(99 primitiveLoadImageSegment)
		(100 primitivePerformInSuperclass)		"Blue Book: primitiveSignalAtTick"
		(101 primitiveBeCursor)
		(102 primitiveBeDisplay)
		(103 primitiveScanCharacters)
		(104 primitiveFail)	"primitiveDrawLoop"
		(105 primitiveStringReplace)
		(106 primitiveScreenSize)
		(107 primitiveMouseButtons)
		(108 primitiveKbdNext)
		(109 primitiveKbdPeek)

		"System Primitives (110-119)"
		(110 primitiveIdentical)
		(111 primitiveClass)
		(112 primitiveBytesLeft)
		(113 primitiveQuit)
		(114 primitiveExitToDebugger)
		(115 primitiveChangeClass)					"Blue Book: primitiveOopsLeft"
		(116 primitiveFlushCacheByMethod)
		(117 primitiveExternalCall)
		(118 primitiveDoPrimitiveWithArgs)
		(119 primitiveFlushCacheSelective)
			"Squeak 2.2 and earlier use 119.  Squeak 2.3 and later use 116.
			Both are supported for backward compatibility."

		"Miscellaneous Primitives (120-127)"
		(120 primitiveCalloutToFFI)
		(121 primitiveImageName)
		(122 primitiveNoop)					"Blue Book: primitiveImageVolume"
		(123 primitiveValueUninterruptably)	"@@@: Remove this when all VMs have support"
		(124 primitiveLowSpaceSemaphore)
		(125 primitiveSignalAtBytesLeft)

		"Squeak Primitives Start Here"

		"Squeak Miscellaneous Primitives (128-149)"
		(126 primitiveDeferDisplayUpdates)
		(127 primitiveShowDisplayRect)
		(128 primitiveArrayBecome)
		(129 primitiveSpecialObjectsOop)
		(130 primitiveFullGC)
		(131 primitiveIncrementalGC)
		(132 primitiveObjectPointsTo)
		(133 primitiveSetInterruptKey)
		(134 primitiveInterruptSemaphore)
		(135 primitiveMillisecondClock)
		(136 primitiveSignalAtMilliseconds)
		(137 primitiveSecondsClock)
		(138 primitiveSomeObject)
		(139 primitiveNextObject)
		(140 primitiveBeep)
		(141 primitiveClipboardText)
		(142 primitiveVMPath)
		(143 primitiveShortAt)
		(144 primitiveShortAtPut)
		(145 primitiveConstantFill)
		"NOTE: When removing the obsolete indexed primitives,
		the following two should go become #primitiveIntegerAt / atPut"
		(146 primitiveFail)	"primitiveReadJoystick"
		(147 primitiveFail)	"primitiveWarpBits"
		(148 primitiveClone)
		(149 primitiveGetAttribute)

		"File Primitives (150-169) - NO LONGER INDEXED"
		(150 157 primitiveFail)
		(158 primitiveCompareWith)
		(159 primitiveFail)
		(160 primitiveAdoptInstance)
		(161 primitiveSetIdentityHash) "CogMemoryManager primitives"
		(162 164 primitiveFail)
		(165 primitiveIntegerAt)		"hacked in here for now"
		(166 primitiveIntegerAtPut)
		(167 primitiveYield)
		(168 primitiveCopyObject)
		(169 primitiveNotIdentical)

		"Sound Primitives (170-199) - NO LONGER INDEXED"
		(170 174 primitiveFail)

		"CogMemoryManager primitives"
		(175 primitiveBehaviorHash)
		(176 primitiveMaxIdentityHash)
		(177 185 primitiveFail)

		"Old closure primitives"
		(186 primitiveFail) "was primitiveClosureValue"
		(187 primitiveFail) "was primitiveClosureValueWithArgs"

		"Perform method directly"
		(188 primitiveExecuteMethodArgsArray)
		(189 primitiveExecuteMethod)

		"Sound Primitives (continued) - NO LONGER INDEXED"
		(190 194 primitiveFail)

		"Unwind primitives"
		(195 primitiveFindNextUnwindContext)
		(196 primitiveTerminateTo)
		(197 primitiveFindHandlerContext)
		(198 primitiveMarkUnwindMethod)
		(199 primitiveMarkHandlerMethod)

		"new closure primitives (were Networking primitives)"
		(200 primitiveClosureCopyWithCopiedValues)
		(201 primitiveClosureValue) "value"
		(202 primitiveClosureValue) "value:"
		(203 primitiveClosureValue) "value:value:"
		(204 primitiveClosureValue) "value:value:value:"
		(205 primitiveClosureValue) "value:value:value:value:"
		(206 primitiveClosureValueWithArgs) "valueWithArguments:"

		(207 209 primitiveFail) "reserved for Cog primitives"

		(210 primitiveAt)		"Compatibility with Cog StackInterpreter Context primitives"
		(211 primitiveAtPut)	"Compatibility with Cog StackInterpreter Context primitives"
		(212 primitiveSize)	"Compatibility with Cog StackInterpreter Context primitives"
		(213 219 primitiveFail) "reserved for Cog primitives"

		(220 primitiveFail)		"reserved for Cog primitives"

		(221 primitiveClosureValueNoContextSwitch) "valueNoContextSwitch"
		(222 primitiveClosureValueNoContextSwitch) "valueNoContextSwitch:"

		(223 229 primitiveFail)	"reserved for Cog primitives"

		(230 primitiveRelinquishProcessor)
		(231 primitiveForceDisplayUpdate)
		(232 primitiveFormPrint)
		(233 primitiveSetFullScreen)
		(234 primitiveFail) "primBitmapdecompressfromByteArrayat"
		(235 primitiveFail) "primStringcomparewithcollated"
		(236 primitiveFail) "primSampledSoundconvert8bitSignedFromto16Bit"
		(237 primitiveFail) "primBitmapcompresstoByteArray"
		(238 239 primitiveFail) "serial port primitives"
		(240 primitiveUTCMicrosecondClock)		"was serial port primitive"
		(241 primitiveLocalMicrosecondClock)		"was serial port primitive"
		(242 primitiveFail)
		(243 primitiveFail) "primStringtranslatefromtotable"
		(244 primitiveFail) "primStringfindFirstInStringinSetstartingAt"
		(245 primitiveFail) "primStringindexOfAsciiinStringstartingAt"
		(246 primitiveFail) "primStringfindSubstringinstartingAtmatchTable"
		(247 primitiveSnapshotEmbedded)
		(248 primitiveInvokeObjectAsMethod)
		(249 primitiveArrayBecomeOneWayCopyHashArg)

		"VM Implementor Primitives (250-255)"
		(250 primitiveClearVMProfile)
		(251 primitiveControlVMProfiling "primitiveStartVMProfiling")
		(252 primitiveVMProfileSamplesInto "primitiveStopVMProfiling")
		(253 primitiveFail "primitiveVMProfileInfoInto")
		(254 primitiveVMParameter)
		(255 primitiveFail) "primitiveInstVarsPutFromStack. Never used except in Disney tests.  Remove after 2.3 release."

		"Quick Push Const Methods"
		(256 primitivePushSelf)
		(257 primitivePushTrue)
		(258 primitivePushFalse)
		(259 primitivePushNil)
		(260 primitivePushMinusOne)
		(261 primitivePushZero)
		(262 primitivePushOne)
		(263 primitivePushTwo)

		"Quick Push Const Methods"
		(264 519 primitiveLoadInstVar)

		"These ranges used to be used by obsiolete indexed primitives."
		(520 529 primitiveFail)
		(530 539 primitiveFail)
		(540 549 primitiveFail)
		(550 559 primitiveFail)
		(560 569 primitiveFail)

		"External primitive support primitives"
		(570 primitiveFlushExternalPrimitives)
		(571 primitiveUnloadModule)
		(572 primitiveListBuiltinModule)
		(573 primitiveListExternalModule)
		(574 primitiveFail) "reserved for addl. external support prims"

		"Unassigned Primitives"
		(575 primitiveHighBit)
		(576 primitiveFail)).
! !

!Interpreter class methodsFor: 'initialization' stamp: ''!
initializeSchedulerIndices
	"Class ProcessorScheduler"
	ProcessListsIndex := 0.
	ActiveProcessIndex := 1.
	"Class LinkedList"
	FirstLinkIndex := 0.
	LastLinkIndex := 1.
	"Class Semaphore"
	ExcessSignalsIndex := 2.
	"Class Link"
	NextLinkIndex := 0.
	"Class Process"
	SuspendedContextIndex := 1.
	PriorityIndex := 2.
	MyListIndex := 3! !

!Interpreter class methodsFor: 'initialization' stamp: ''!
initializeSmallIntegers
	"SmallIntegers"
	ConstMinusOne := Interpreter new integerObjectOf: -1.
	ConstZero := Interpreter new integerObjectOf: 0.
	ConstOne := Interpreter new integerObjectOf: 1.
	ConstTwo := Interpreter new integerObjectOf: 2! !

!Interpreter class methodsFor: 'initialization' stamp: ''!
initializeStreamIndices
	StreamArrayIndex := 0.
	StreamIndexIndex := 1.
	StreamReadLimitIndex := 2.
	StreamWriteLimitIndex := 3.! !

!Interpreter class methodsFor: 'initialization' stamp: 'eem 09/28/2015 18:49'!
initializeWithOptions: optionsDictionary
	"Interpreter initializeWithOptions: Dictionary new"

	super initializeWithOptions: optionsDictionary.  "initialize ObjectMemory constants"
	self initializeMiscConstants. "must precede other initialization."
	self initializeAssociationIndex.
	self initializeBytecodeTable.
	self initializeCaches.
	self initializeCharacterIndex.
	self initializeCharacterScannerIndices.
	self initializeClassIndices.
	self initializeCompilerHooks.
	self initializeContextIndices.
	self initializeDirectoryLookupResultCodes.
	self initializeMessageIndices.
	self initializeMethodIndices.
	self initializePointIndices.
	self initializePrimitiveTable.
	self initializeSchedulerIndices.
	self initializeSmallIntegers.
	self initializeStreamIndices! !

!Interpreter class methodsFor: 'translation' stamp: 'eem 2/10/2009 16:02'!
isInterpreterClass
	^true! !

!Interpreter class methodsFor: 'translation' stamp: 'eem 6/19/2008 22:16'!
isNonArgumentImplicitReceiverVariableName: aString
	aString = 'interpreterProxy' ifTrue: [self halt].
	^'self' = aString! !

!Interpreter class methodsFor: 'benchmarks' stamp: 'eem 6/21/2011 08:52'!
macroBenchmark  "Interpreter macroBenchmark"
	"Copied from Interpreter class>>translate:doInlining:forBrowserPlugin:"
	| cg fileName |
	fileName := 'benchmark2.out'.
	Interpreter initialize.
	ObjectMemory initialize.
	cg := CCodeGenerator new initialize.
	Interpreter initialize.
	ObjectMemory initializeWithOptions: (Dictionary new
											at: #BytesPerWord put: 4;
											yourself).
	cg addClass: Interpreter.
	cg addClass: ObjectMemory.
	Interpreter declareCVarsIn: cg.
	ObjectMemory declareCVarsIn: cg.
	FileDirectory default deleteFileNamed: fileName.
	cg storeCodeOnFile: fileName doInlining: true.
	FileDirectory default deleteFileNamed: fileName.! !

!Interpreter class methodsFor: 'translation' stamp: 'eem 2/10/2009 15:37'!
mustBeGlobal: var
	"Answer if a variable must be global and exported.  Used for inst vars that are accessed from VM support code."

	^(super mustBeGlobal: var)
	   or: [ #('deferDisplayUpdates' 'extraVMMemory' 'interpreterProxy' 'interpreterVersion' 'showSurfaceFn') includes: var]! !

!Interpreter class methodsFor: 'translation' stamp: 'eem 6/18/2008 10:02'!
namesOfVariablesToLocalize
	^#(currentBytecode localIP localSP localHomeContext localReturnContext localReturnValue)! !

!Interpreter class methodsFor: 'translation' stamp: 'eem 2/9/2009 13:15'!
needsCogit
	^false! !

!Interpreter class methodsFor: 'accessing class hierarchy' stamp: 'eem 11/15/2019 16:16'!
objectMemoryClass
	^ObjectMemory! !

!Interpreter class methodsFor: 'translation' stamp: 'sw 5/23/2001 14:33'!
patchInterp: fileName
	"Interpreter patchInterp: 'Squeak VM PPC'"
	"This will patch out the unneccesary range check (a compare
	 and branch) in the inner interpreter dispatch loop."
	"NOTE: You must edit in the Interpeter file name, and the
	 number of instructions (delta) to count back to find the compare
	 and branch that we want to get rid of."

	| delta f code len remnant i |
	delta := 6.
	f := FileStream fileNamed: fileName.
	f binary.
	code := Bitmap new: (len := f size) // 4.
	f nextInto: code.
	remnant := f next: len - (code size * 4).
	i := 0.
	["Look for a BCTR instruction"
	(i := code indexOf: 16r4E800420 startingAt: i + 1 ifAbsent: [0]) > 0] whileTrue:
		["Look for a CMPLWI FF, 6 instrs back"
	       ((code at: i - delta) bitAnd: 16rFFE0FFFF) = 16r280000FF ifTrue:
			["Copy dispatch instrs back over the compare"
			self inform: 'Patching at ', i hex.
			0 to: delta - 2 do: [ :j |
				code at: (i - delta) + j put: (code at: (i - delta) + j + 2)]]].
	f position: 0; nextPutAll: code; nextPutAll: remnant.
	f close.
! !

!Interpreter class methodsFor: 'translation' stamp: 'sw 5/23/2001 14:34'!
patchInterpGCCPPC: fileName
	"Interpreter patchInterpGCCPPC: 'Squeak copy 1'"
	"This will patch out the unneccesary range check (a compare
	 and branch) in the inner interpreter dispatch loop. for the PPC version of the GCC compiled
	version of Squeak under MPW"
	"NOTE: You must edit in the Interpeter file name"

	| delta f code len remnant i |
	delta := 7.
	f := FileStream fileNamed: fileName.
	f binary.
	code := Bitmap new: (len := f size) // 4.
	f nextInto: code.
	remnant := f next: len - (code size * 4).
	i := 0.
	["Look for a BCTR instruction"
	(i := code indexOf: 16r4E800420 startingAt: i + 1 ifAbsent: [0]) > 0] whileTrue:
		["Look for a CMPLWI cr1,rxx,FF, 7 instrs back"
	       ((code at: i - delta) bitAnd: 16rFFE0FFFF) = 16r288000FF ifTrue:
	       	["Copy dispatch instrs back over the compare"
			self inform: 'Patching at ', i hex.
			0 to: delta - 2 do: [ :j |
				code at: (i - delta) + j put: (code at: (i - delta) + j + 2)]]].
	f position: 0; nextPutAll: code; nextPutAll: remnant.
	f close! !

!Interpreter class methodsFor: 'translation' stamp: 'eem 8/21/2013 12:46'!
preGenerationHook: aCCodeGenerator
	"Perform any last-minute changes to the code generator immediately
	 before it performs code analysis and generation.  In this case, make
	 all non-exported methods private."
	| requiredMethodNames |
	requiredMethodNames := self requiredMethodNames: aCCodeGenerator options.
	aCCodeGenerator selectorsAndMethodsDo:
		[:s :m|
		(m export or: [requiredMethodNames includes: s]) ifTrue:
			[m static: false]]! !

!Interpreter class methodsFor: 'translation' stamp: 'topa 7/25/2016 15:53'!
preambleCCode
	^	
'void printCallStack(void);

/* Disable Intel compiler inlining of error & warning which are used for breakpoints */
#pragma auto_inline(off)

void
warning(char *s) { /* Print an error message but don''t exit. */
	static sqInt printingStack = false;

	printf("\n%s\n", s);
}
#pragma auto_inline(on)

void
invalidCompactClassError(char *s) { /* Print a compact class index error message and exit. */
	static sqInt printingStack = true; /* not running at this point */

	printf("\nClass %s does not have the required compact class index\n", s);
	exit(-1);
}
'! !

!Interpreter class methodsFor: 'constants' stamp: ''!
primitiveTable

	^ PrimitiveTable! !

!Interpreter class methodsFor: 'initialization' stamp: 'tpr 3/17/2005 11:32'!
primitiveTableString
	"Interpreter initializePrimitiveTable primitiveTableString"
	| table |
	table := self primitiveTable.
	^ String
		streamContents: [:s | 
			s nextPut: ${.
			table
				withIndexDo: [:primSpec :index | s cr; tab;
					nextPutAll: '/* ';
					nextPutAll: (index - 1) printString;
					nextPutAll: '*/ ';
					nextPutAll: '(void *)';
					nextPutAll: primSpec;
					nextPut: $,].
			s cr; nextPutAll: ' 0 }']! !

!Interpreter class methodsFor: 'translation' stamp: 'eem 11/14/2019 14:27'!
requiredMethodNames: options
	"return the list of method names that should be retained for export or other support reasons"
	| requiredList |
	requiredList := Set new: 400.
	"A number of methods required by VM support code, jitter, specific platforms etc"
	requiredList addAll: #(fullDisplayUpdate interpret printCallStack printAllStacks printOop: readImageFromFile:HeapSize:StartingAt: success: getCurrentBytecode characterForAscii: findClassOfMethod:forReceiver: findSelectorOfMethod:forReceiver: loadInitialContext nullCompilerHook primitiveFlushExternalPrimitives setCompilerInitialized: getFullScreenFlag getInterruptCheckCounter getInterruptKeycode getInterruptPending getNextWakeupTick getSavedWindowSize setFullScreenFlag: setInterruptCheckCounter: setInterruptKeycode: setInterruptPending: setNextWakeupTick: setSavedWindowSize: forceInterruptCheck getThisSessionID).

	"Nice to actually have all the primitives available"
	requiredList addAll: self primitiveTable.

	"InterpreterProxy is the internal analogue of sqVirtualMachine.c, so make sure to keep all those"
	InterpreterProxy organization categories do: [:cat |
		((cat ~= 'initialize') and: [cat ~= 'private']) ifTrue: [
			requiredList addAll: (InterpreterProxy organization listAtCategoryNamed: cat)]].
	
	^requiredList! !

!Interpreter class methodsFor: 'translation' stamp: 'eem 2/9/2009 12:41'!
sourceFileName
	^'interp.c'! !

!Interpreter class methodsFor: 'initialization' stamp: 'tpr 3/24/2004 21:29'!
table: anArray from: specArray 
	"SpecArray is an array of either (index selector) or (index1 
	index2 selector)."
	| contiguous |
	contiguous := 0.
	specArray do: [:spec | 
			(spec at: 1) = contiguous
				ifFalse: [self error: 'Non-contiguous table entry'].
			spec size = 2
				ifTrue: [anArray at: (spec at: 1) + 1
						put: (spec at: 2).
					contiguous := contiguous + 1]
				ifFalse: [(spec at: 1) to: (spec at: 2) do: [:i | anArray at: i + 1 put: (spec at: 3)].
					contiguous := contiguous + ((spec at: 2) - (spec at: 1)) + 1]]! !

!Interpreter class methodsFor: 'translation' stamp: 'eem 10/13/2008 14:28'!
wantsLabels
	"Only label the VM, not plugins (internal or external).  This to cut down the scope
	 of problems with labels being duplicated by C compiler optimizer inlining and
	 loop unrolling.  We use the asmLabel: directive to control labelling in the
	 interpreter proper. But it is too much work doing that for plugins too."
	^true! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
activateNewClosureMethod: blockClosure
	"Similar to activateNewMethod but for Closure and newMethod."
	| theBlockClosure closureMethod newContext methodHeader numCopied where outerContext |

	DoAssertionChecks ifTrue:
		[self okayOop: blockClosure].
	outerContext := self fetchPointer: ClosureOuterContextIndex ofObject: blockClosure.
	DoAssertionChecks ifTrue:
		[self okayOop: outerContext].
	closureMethod := self fetchPointer: MethodIndex ofObject: outerContext.
	methodHeader := self methodHeaderOf: closureMethod.
	self pushRemappableOop: blockClosure.
	newContext := self allocateOrRecycleContext: (methodHeader bitAnd: LargeContextBit). "All for one, and one for all!!"

	"allocateOrRecycleContext: may cause a GC; restore blockClosure and refetch outerContext et al"
	theBlockClosure := self popRemappableOop.
	outerContext := self fetchPointer: ClosureOuterContextIndex ofObject: theBlockClosure.
	numCopied := (self numSlotsOf: theBlockClosure) - ClosureFirstCopiedValueIndex.

	"Assume: newContext will be recorded as a root if necessary by the
	 call to newActiveContext: below, so we can use unchecked stores."
	where :=  newContext + self baseHeaderSize.
	self longAt: where + (SenderIndex << self shiftForWord)
		put: activeContext.
	self longAt: where + (InstructionPointerIndex << self shiftForWord)
		put: (self fetchPointer: ClosureStartPCIndex ofObject: theBlockClosure).
	self longAt: where + (StackPointerIndex << self shiftForWord)
		put: (self integerObjectOf: argumentCount + numCopied).
	self longAt: where + (MethodIndex << self shiftForWord)
		put: (self fetchPointer: MethodIndex ofObject: outerContext).
	self longAt: where + (ClosureIndex << self shiftForWord)
		put: theBlockClosure.
	self longAt: where + (ReceiverIndex << self shiftForWord)
		put: (self fetchPointer: ReceiverIndex ofObject: outerContext).

	"Copy the arguments..."
	1 to: argumentCount do:
		[:i | self longAt: where + ((ReceiverIndex+i) << self shiftForWord)
				put: (self stackValue: argumentCount-i)].

	"Copy the copied values..."
	where := newContext + self baseHeaderSize + ((ReceiverIndex + 1 + argumentCount) << self shiftForWord).
	0 to: numCopied - 1 do:
		[:i| self longAt: where + (i << self shiftForWord)
				put: (self fetchPointer: i + ClosureFirstCopiedValueIndex
						  ofObject: theBlockClosure)].

	"The initial instructions in the block nil-out remaining temps."

	self pop: argumentCount + 1.
	self newActiveContext: newContext! !

!Interpreter methodsFor: 'message sending' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
activateNewMethod
	| newContext methodHeader initialIP tempCount nilOop where |

	methodHeader := self methodHeaderOf: newMethod.
	newContext := self allocateOrRecycleContext: (methodHeader bitAnd: LargeContextBit).

	initialIP := ((LiteralStart + (self literalCountOfMethodHeader: methodHeader)) * self wordSize) + 1.
	tempCount := (methodHeader >> 19) bitAnd: 16r3F.

	"Assume: newContext will be recorded as a root if necessary by the
	 call to newActiveContext: below, so we can use unchecked stores."

	where :=  newContext  + self baseHeaderSize.
	self longAt: where + (SenderIndex << self shiftForWord) put: activeContext.
	self longAt: where + (InstructionPointerIndex << self shiftForWord) put: (self integerObjectOf: initialIP).
	self longAt: where + (StackPointerIndex << self shiftForWord) put: (self integerObjectOf: tempCount).
	self longAt: where + (MethodIndex << self shiftForWord) put: newMethod.
	self longAt: where + (ClosureIndex << self shiftForWord) put: nilObj.

	"Copy the receiver and arguments..."
	0 to: argumentCount do:
		[:i | self longAt: where + ((ReceiverIndex+i) << self shiftForWord) put: (self stackValue: argumentCount-i)].

	"clear remaining temps to nil in case it has been recycled"
	nilOop := nilObj.
	argumentCount+1+ReceiverIndex to: tempCount+ReceiverIndex do:
		[:i | self longAt: where + (i << self shiftForWord) put: nilOop].

	self pop: argumentCount + 1.
	reclaimableContextCount := reclaimableContextCount + 1.
	self newActiveContext: newContext.! !

!Interpreter methodsFor: 'process primitive support' stamp: 'tpr 3/24/2004 20:52'!
addLastLink: proc toList: aList 
	"Add the given process to the given linked list and set the 
	backpointer of process to its new list."
	| lastLink |
	(self isEmptyList: aList)
		ifTrue: [self storePointer: FirstLinkIndex ofObject: aList withValue: proc]
		ifFalse: [lastLink := self fetchPointer: LastLinkIndex ofObject: aList.
			self storePointer: NextLinkIndex ofObject: lastLink withValue: proc].
	self storePointer: LastLinkIndex ofObject: aList withValue: proc.
	self storePointer: MyListIndex ofObject: proc withValue: aList! !

!Interpreter methodsFor: 'method lookup cache' stamp: 'nice 7/13/2016 18:51'!
addNewMethodToCache
	"Add the given entry to the method cache.
	The policy is as follows:
		Look for an empty entry anywhere in the reprobe chain.
		If found, install the new entry there.
		If not found, then install the new entry at the first probe position
			and delete the entries in the rest of the reprobe chain.
		This has two useful purposes:
			If there is active contention over the first slot, the second
				or third will likely be free for reentry after ejection.
			Also, flushing is good when reprobe chains are getting full."
	| probe hash |
	<inline: false>
	hash := messageSelector bitXor: lkupClass.  "drop low-order zeros from addresses"
	(self isOopCompiledMethod: newMethod)
		ifTrue:
			[primitiveIndex := self primitiveIndexOf: newMethod.
			 primitiveFunctionPointer := self functionPointerFor: primitiveIndex inClass: lkupClass]
		ifFalse:
			[primitiveFunctionPointer := #primitiveInvokeObjectAsMethod].

	primitiveFunctionPointer := self functionPointerFor: primitiveIndex inClass: lkupClass.
	
	0 to: CacheProbeMax-1 do:
		[:p | probe := (hash >> p) bitAnd: MethodCacheMask.
		(methodCache at: probe + MethodCacheSelector) = 0 ifTrue:
			["Found an empty entry -- use it"
			methodCache at: probe + MethodCacheSelector put: messageSelector.
			methodCache at: probe + MethodCacheClass put: lkupClass.
			methodCache at: probe + MethodCacheMethod put: newMethod.
			methodCache at: probe + MethodCachePrim put: primitiveIndex.
			methodCache at: probe + MethodCachePrimFunction put: (self cCoerce: primitiveFunctionPointer to: #'sqIntptr_t').
			^ nil]].

	"OK, we failed to find an entry -- install at the first slot..."
	probe := hash bitAnd: MethodCacheMask.  "first probe"
	methodCache at: probe + MethodCacheSelector put: messageSelector.
	methodCache at: probe + MethodCacheClass put: lkupClass.
	methodCache at: probe + MethodCacheMethod put: newMethod.
	methodCache at: probe + MethodCachePrim put: primitiveIndex.
	methodCache at: probe + MethodCachePrimFunction put: (self cCoerce: primitiveFunctionPointer to: #'sqIntptr_t').

	"...and zap the following entries"
	1 to: CacheProbeMax-1 do:
		[:p | probe := (hash >> p) bitAnd: MethodCacheMask.
		methodCache at: probe + MethodCacheSelector put: 0].
! !

!Interpreter methodsFor: 'plugin support' stamp: 'ikp (auto pragmas 12/08) 6/10/2004 12:26'!
addToExternalPrimitiveTable: functionAddress
	"Add the given function address to the external primitive table and return the index where it's stored. This function doesn't need to be fast since it is only called when an external primitive has been looked up (which takes quite a bit of time itself). So there's nothing specifically complicated here.
	Note: Return index will be one-based (ST convention)"

	<var: #functionAddress declareC: 'void *functionAddress'>

	0 to: MaxExternalPrimitiveTableSize-1 do: [ :i |
		(externalPrimitiveTable at: i) = 0 ifTrue: [
			externalPrimitiveTable at: i put: functionAddress.
			^i+1]].
	"if no space left, return zero so it'll looked up again"
	^0! !

!Interpreter methodsFor: 'process primitive support' stamp: 'ar 2/27/2008 10:36'!
adjustInterruptCheckCounterFeedback
	"Answer true if we should adjust the interrupt counter feedback.
	Only do this if the current interrupt wasn't forced and we're not
	profiling execution"
	^self interruptCheckForced = false and:[nextProfileTick = 0].! !

!Interpreter methodsFor: 'debug support' stamp: 'eem 12/19/2012 15:35'!
allAccessibleObjectsOkay
	"Ensure that all accessible objects in the heap are okay."

	| oop ok |
	ok := true.
	oop := self firstAccessibleObject.
	[oop = nil] whileFalse:
		[ok := ok & (self okayFields: oop).
		 oop := self accessibleObjectAfter: oop].
	^ok! !

!Interpreter methodsFor: 'contexts' stamp: 'di 10/22/1999 11:52'!
allocateOrRecycleContext: needsLarge
	"Return a recycled context or a newly allocated one if none is available for recycling."
	| cntxt |
	needsLarge = 0
	ifTrue: [freeContexts ~= NilContext ifTrue:
				[cntxt := freeContexts.
				freeContexts := self fetchPointer: 0 ofObject: cntxt.
				^ cntxt]]
	ifFalse: [freeLargeContexts ~= NilContext ifTrue:
				[cntxt := freeLargeContexts.
				freeLargeContexts := self fetchPointer: 0 ofObject: cntxt.
				^ cntxt]].
	
	needsLarge = 0
		ifTrue: [cntxt := self instantiateContext: (self splObj: ClassMethodContext)
				sizeInBytes: SmallContextSize]
		ifFalse: [cntxt := self instantiateContext: (self splObj: ClassMethodContext)
				sizeInBytes: LargeContextSize].
	"Required init -- above does not fill w/nil.  All others get written."
	self storePointerUnchecked: 4 "InitialIPIndex" ofObject: cntxt
					withValue: nilObj.
	^ cntxt
! !

!Interpreter methodsFor: 'compiled methods' stamp: 'eem 08/08/2014 10:47'!
argumentCountOf: methodPointer
	^((self methodHeaderOf: methodPointer) >> 25) bitAnd: 16r0F! !

!Interpreter methodsFor: 'contexts' stamp: 'tpr 3/24/2004 20:59'!
argumentCountOfBlock: blockPointer

	| localArgCount |
	localArgCount := self fetchPointer: BlockArgumentCountIndex ofObject: blockPointer.
	^self checkedIntegerValueOf: localArgCount! !

!Interpreter methodsFor: 'contexts' stamp: 'eem 6/20/2008 21:49'!
argumentCountOfClosure: closurePointer

	^self quickFetchInteger: ClosureNumArgsIndex ofObject: closurePointer! !

!Interpreter methodsFor: 'utilities' stamp: 'eem 11/5/2014 18:23'!
arrayValueOf: arrayOop
	"Return the address of first indexable field of resulting array object, or fail if
	 the instance variable does not contain an indexable bytes or words object."
	"Note: May be called by translated primitive code."

	<returnTypeC: #'void *'>
	(self isWordsOrBytes: arrayOop) ifTrue:
		[^self cCoerceSimple: (self pointerForOop: arrayOop + self baseHeaderSize) to: #'void *'].
	self primitiveFail! !

!Interpreter methodsFor: 'array primitive support' stamp: '(auto pragmas 12/08) '!
asciiOfCharacter: characterObj  "Returns an integer object"

	<inline: false>
	self assertClassOf: characterObj is: (self splObj: ClassCharacter).
	successFlag
		ifTrue: [^ self fetchPointer: CharacterValueIndex ofObject: characterObj]
		ifFalse: [^ ConstZero]  "in case some code needs an int"! !

!Interpreter methodsFor: 'utilities' stamp: 'eem 11/10/2014 12:49'!
assertClassOf: oop is: classOop
	"Succeed if oop is an instance of the given class. Fail if the object is an integer."

	<inline: true>
	self success: (self isClassOfNonImm: oop equalTo: classOop)! !

!Interpreter methodsFor: 'debug support' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
balancedStack: delta afterPrimitive: primIdx withArgs: nArgs
	"Return true if the stack is still balanced after executing primitive primIndex with nArgs args. Delta is 'stackPointer - activeContext' which is a relative measure for the stack pointer (so we don't have to relocate it during the primitive)"
	(primIdx >= 81 and:[primIdx <= 88]) ifTrue:[^true].
	"81-88 are control primitives after which the stack may look unbalanced"
	successFlag ifTrue:[
		"Successful prim, stack must have exactly nArgs arguments popped off"
		^(stackPointer - activeContext + (nArgs * self wordSize)) = delta
	].
	"Failed prim must leave stack intact"
	^(stackPointer - activeContext) = delta
! !

!Interpreter methodsFor: 'utilities' stamp: 'tpr (auto pragmas 12/08) 3/15/2004 19:44'!
booleanCheat: cond
"cheat the interpreter out of the pleasure of handling the next bytecode IFF it is a jump-on-boolean. Which it is, often enough when the current bytecode is something like bytecodePrimEqual"
	| bytecode offset |
	<inline: true>

	bytecode := self fetchByte.  "assume next bytecode is jumpIfFalse (99%)"
	self internalPop: 2.
	(bytecode < 160 and: [bytecode > 151]) ifTrue: [  "short jumpIfFalse"
		cond
			ifTrue: [^ self fetchNextBytecode]
			ifFalse: [^ self jump: bytecode - 151]].

	bytecode = 172 ifTrue: [  "long jumpIfFalse"
		offset := self fetchByte.
		cond
			ifTrue: [^ self fetchNextBytecode]
			ifFalse: [^ self jump: offset]].

	"not followed by a jumpIfFalse; undo instruction fetch and push boolean result"
	localIP := localIP - 1.
	self fetchNextBytecode.
	cond
		ifTrue: [self internalPush: trueObj]
		ifFalse: [self internalPush: falseObj].
! !

!Interpreter methodsFor: 'utilities' stamp: 'tpr 3/15/2004 19:46'!
booleanValueOf: obj
"convert true and false (Smalltalk) to true or false(C)"
	obj = trueObj ifTrue: [ ^ true ].
	obj = falseObj ifTrue: [ ^ false ].
	successFlag := false.
	^ nil! !

!Interpreter methodsFor: 'image save/restore' stamp: 'di 10/18/1999 16:58'!
byteSwapByteObjects
	"Byte-swap the words of all bytes objects in the image. This returns these objects to their original byte ordering after blindly byte-swapping the entire image."

	self byteSwapByteObjectsFrom: self firstObject to: endOfMemory! !

!Interpreter methodsFor: 'image save/restore' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
byteSwapByteObjectsFrom: startOop to: stopAddr 
	"Byte-swap the words of all bytes objects in a range of the 
	image, including Strings, ByteArrays, and CompiledMethods. 
	This returns these objects to their original byte ordering 
	after blindly byte-swapping the entire image. For compiled 
	methods, byte-swap only their bytecodes part."
	| oop fmt wordAddr methodHeader |
	oop := startOop.
	[self oop: oop isLessThan: stopAddr]
		whileTrue: [(self isFreeObject: oop)
				ifFalse: [fmt := self formatOf: oop.
					fmt >= 8
						ifTrue: ["oop contains bytes"
							wordAddr := oop + self baseHeaderSize.
							fmt >= 12
								ifTrue: ["compiled method; start after methodHeader and literals"
									methodHeader := self longAt: oop + self baseHeaderSize.
									wordAddr := wordAddr + self wordSize + ((methodHeader >> 10 bitAnd: 255) * self wordSize)].
							self reverseBytesFrom: wordAddr to: oop + (self sizeBitsOf: oop)].
					(fmt = 6 and: [self wordSize = 8])
						ifTrue: ["Object contains 32-bit half-words packed into 64-bit machine words."
							wordAddr := oop + self baseHeaderSize.
							self reverseWordsFrom: wordAddr to: oop + (self sizeBitsOf: oop)]].
			oop := self objectAfter: oop]! !

!Interpreter methodsFor: 'common selector sends' stamp: 'tpr 3/23/2004 18:23'!
bytecodePrimAdd
	| rcvr arg result |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(self areIntegers: rcvr and: arg)
		ifTrue: [result := (self integerValueOf: rcvr) + (self integerValueOf: arg).
				(self isIntegerValue: result) ifTrue:
					[self internalPop: 2 thenPush: (self integerObjectOf: result).
					^ self fetchNextBytecode "success"]]
		ifFalse: [successFlag := true.
				self externalizeIPandSP.
				self primitiveFloatAdd: rcvr toArg: arg.
				self internalizeIPandSP.
				successFlag ifTrue: [^ self fetchNextBytecode "success"]].

	messageSelector := self specialSelector: 0.
	argumentCount := 1.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'tpr 3/23/2004 18:24'!
bytecodePrimAt
	"BytecodePrimAt will only succeed if the receiver is in the atCache.
	Otherwise it will fail so that the more general primitiveAt will put it in the
	cache after validating that message lookup results in a primitive response."
	| index rcvr result atIx |
	index := self internalStackTop.
	rcvr := self internalStackValue: 1.
	successFlag := (self isIntegerObject: rcvr) not and: [self isIntegerObject: index].
	successFlag ifTrue:
		[atIx := rcvr bitAnd: AtCacheMask.  "Index into atCache = 4N, for N = 0 ... 7"
		(atCache at: atIx+AtCacheOop) = rcvr
		ifTrue: [result := self commonVariableInternal: rcvr at: (self integerValueOf: index) cacheIndex: atIx.
			successFlag ifTrue:
				[self fetchNextBytecode.
				^self internalPop: 2 thenPush: result]]].

	messageSelector := self specialSelector: 16.
	argumentCount := 1.
	self normalSend.
! !

!Interpreter methodsFor: 'common selector sends' stamp: 'di 6/21/97 08:58'!
bytecodePrimAtEnd
	messageSelector := self specialSelector: 21.
	argumentCount := 0.
	self normalSend.! !

!Interpreter methodsFor: 'common selector sends' stamp: 'tpr 3/23/2004 18:25'!
bytecodePrimAtPut
	"BytecodePrimAtPut will only succeed if the receiver is in the atCache.
	Otherwise it will fail so that the more general primitiveAtPut will put it in the
	cache after validating that message lookup results in a primitive response."
	| index rcvr atIx value |
	value := self internalStackTop.
	index := self internalStackValue: 1.
	rcvr := self internalStackValue: 2.
	successFlag := (self isIntegerObject: rcvr) not and: [self isIntegerObject: index].
	successFlag
		ifTrue: [atIx := (rcvr bitAnd: AtCacheMask) + AtPutBase.  "Index into atPutCache"
			(atCache at: atIx+AtCacheOop) = rcvr
				ifTrue: [self commonVariable: rcvr at: (self integerValueOf: index) put: value cacheIndex: atIx.
					successFlag ifTrue: [self fetchNextBytecode.
						^self internalPop: 3 thenPush: value]]].

	messageSelector := self specialSelector: 17.
	argumentCount := 2.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'jm 12/10/1998 17:29'!
bytecodePrimBitAnd

	successFlag := true.
	self externalizeIPandSP.
	self primitiveBitAnd.
	self internalizeIPandSP.
	successFlag ifTrue: [^ self fetchNextBytecode "success"].

	messageSelector := self specialSelector: 14.
	argumentCount := 1.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'jm 12/10/1998 17:29'!
bytecodePrimBitOr

	successFlag := true.
	self externalizeIPandSP.
	self primitiveBitOr.
	self internalizeIPandSP.
	successFlag ifTrue: [^ self fetchNextBytecode "success"].

	messageSelector := self specialSelector: 15.
	argumentCount := 1.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'jm 12/10/1998 17:29'!
bytecodePrimBitShift

	successFlag := true.
	self externalizeIPandSP.
	self primitiveBitShift.
	self internalizeIPandSP.
	successFlag ifTrue: [^ self fetchNextBytecode "success"].

	messageSelector := self specialSelector: 12.
	argumentCount := 1.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'tpr 3/23/2004 18:26'!
bytecodePrimBlockCopy

	| rcvr hdr |
	rcvr := self internalStackValue: 1.
	successFlag := true.
	hdr := self baseHeader: rcvr.
	self success: (self isContextHeader: hdr).
	successFlag ifTrue: [self externalizeIPandSP.
		self primitiveBlockCopy.
		self internalizeIPandSP].
	successFlag ifFalse: [messageSelector := self specialSelector: 24.
		argumentCount := 1.
		^ self normalSend].
	self fetchNextBytecode.
! !

!Interpreter methodsFor: 'common selector sends' stamp: 'di 1/11/1999 00:09'!
bytecodePrimClass
	| rcvr |
	rcvr := self internalStackTop.
	self internalPop: 1 thenPush: (self fetchClassOf: rcvr).
	self fetchNextBytecode.
! !

!Interpreter methodsFor: 'common selector sends' stamp: 'tpr 3/24/2004 18:26'!
bytecodePrimDiv
	| quotient |
	successFlag := true.
	quotient := self doPrimitiveDiv: (self internalStackValue: 1) by: (self internalStackValue: 0).
	successFlag ifTrue: [self internalPop: 2 thenPush: (self integerObjectOf: quotient).
		^ self fetchNextBytecode "success"].

	messageSelector := self specialSelector: 13.
	argumentCount := 1.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'tpr 3/24/2004 18:27'!
bytecodePrimDivide
	| rcvr arg result |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(self areIntegers: rcvr and: arg)
		ifTrue: [rcvr := self integerValueOf: rcvr.
			arg := self integerValueOf: arg.
			(arg ~= 0 and: [rcvr \\ arg = 0])
				ifTrue: [result := rcvr // arg.
					"generates C / operation"
					(self isIntegerValue: result)
						ifTrue: [self internalPop: 2 thenPush: (self integerObjectOf: result).
							^ self fetchNextBytecode"success"]]]
		ifFalse: [successFlag := true.
			self externalizeIPandSP.
			self primitiveFloatDivide: rcvr byArg: arg.
			self internalizeIPandSP.
			successFlag ifTrue: [^ self fetchNextBytecode"success"]].

	messageSelector := self specialSelector: 9.
	argumentCount := 1.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: ''!
bytecodePrimDo

	messageSelector := self specialSelector: 27.
	argumentCount := 1.
	self normalSend.
! !

!Interpreter methodsFor: 'common selector sends' stamp: 'tpr 3/24/2004 18:28'!
bytecodePrimEqual
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(self areIntegers: rcvr and: arg) ifTrue: [^self booleanCheat: rcvr = arg].

	successFlag := true.
	aBool := self primitiveFloatEqual: rcvr toArg: arg.
	successFlag ifTrue: [^self booleanCheat: aBool].

	messageSelector := self specialSelector: 6.
	argumentCount := 1.
	self normalSend
! !

!Interpreter methodsFor: 'common selector sends' stamp: ''!
bytecodePrimEquivalent

	| rcvr arg |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	self booleanCheat: rcvr = arg.! !

!Interpreter methodsFor: 'common selector sends' stamp: 'eem 9/18/2010 11:41'!
bytecodePrimGreaterOrEqual
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(self areIntegers: rcvr and: arg) ifTrue:
		[self cCode: '' inSmalltalk: [^self booleanCheat: (self integerValueOf: rcvr) >= (self integerValueOf: arg)].
		^self booleanCheat: rcvr >= arg].

	successFlag := true.
	aBool := self primitiveFloatGreaterOrEqual: rcvr toArg: arg.
	successFlag ifTrue: [^self booleanCheat: aBool].

	messageSelector := self specialSelector: 5.
	argumentCount := 1.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'eem 1/6/2009 10:04'!
bytecodePrimGreaterThan
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(self areIntegers: rcvr and: arg) ifTrue:
		["The C code can avoid detagging since tagged integers are still signed.
		 But this means the simulator must override to do detagging."
		self cCode: '' inSmalltalk: [^self booleanCheat: (self integerValueOf: rcvr) > (self integerValueOf: arg)].
		^self booleanCheat: rcvr > arg].

	successFlag := true.
	aBool := self primitiveFloatGreater: rcvr thanArg: arg.
	successFlag ifTrue: [^self booleanCheat: aBool].

	messageSelector := self specialSelector: 3.
	argumentCount := 1.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'eem 9/18/2010 11:41'!
bytecodePrimLessOrEqual
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(self areIntegers: rcvr and: arg) ifTrue:
		[self cCode: '' inSmalltalk: [^self booleanCheat: (self integerValueOf: rcvr) <= (self integerValueOf: arg)].
		^ self booleanCheat: rcvr <= arg].

	successFlag := true.
	aBool := self primitiveFloatLessOrEqual: rcvr toArg: arg.
	successFlag ifTrue: [^self booleanCheat: aBool].

	messageSelector := self specialSelector: 4.
	argumentCount := 1.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'eem 1/6/2009 10:06'!
bytecodePrimLessThan
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(self areIntegers: rcvr and: arg) ifTrue:
		["The C code can avoid detagging since tagged integers are still signed.
		 But this means the simulator must override to do detagging."
		self cCode: '' inSmalltalk: [^self booleanCheat: (self integerValueOf: rcvr) < (self integerValueOf: arg)].
		^ self booleanCheat: rcvr < arg].

	successFlag := true.
	aBool := self primitiveFloatLess: rcvr thanArg: arg.
	successFlag ifTrue: [^ self booleanCheat: aBool].

	messageSelector := self specialSelector: 2.
	argumentCount := 1.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'jm 12/10/1998 17:31'!
bytecodePrimMakePoint

	successFlag := true.
	self externalizeIPandSP.
	self primitiveMakePoint.
	self internalizeIPandSP.
	successFlag ifTrue: [^ self fetchNextBytecode "success"].

	messageSelector := self specialSelector: 11.
	argumentCount := 1.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'tpr 3/24/2004 18:31'!
bytecodePrimMod
	| mod |
	successFlag := true.
	mod := self doPrimitiveMod: (self internalStackValue: 1) by: (self internalStackValue: 0).
	successFlag ifTrue:
		[self internalPop: 2 thenPush: (self integerObjectOf: mod).
		^ self fetchNextBytecode "success"].

	messageSelector := self specialSelector: 10.
	argumentCount := 1.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'eem 10/4/2011 12:18'!
bytecodePrimMultiply
	| rcvr arg result |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(self areIntegers: rcvr and: arg)
		ifTrue: [rcvr := self integerValueOf: rcvr.
				arg := self integerValueOf: arg.
				result := rcvr * arg.
				(arg = 0
				 or: [(result // arg) = rcvr and: [self isIntegerValue: result]]) ifTrue:
					[self internalPop: 2 thenPush: (self integerObjectOf: result).
					 ^self fetchNextBytecode "success"]]
		ifFalse: [successFlag := true.
				self externalizeIPandSP.
				self primitiveFloatMultiply: rcvr byArg: arg.
				self internalizeIPandSP.
				successFlag ifTrue: [^ self fetchNextBytecode "success"]].

	messageSelector := self specialSelector: 8.
	argumentCount := 1.
	self normalSend.
! !

!Interpreter methodsFor: 'common selector sends' stamp: ''!
bytecodePrimNew

	messageSelector := self specialSelector: 28.
	argumentCount := 0.
	self normalSend.
! !

!Interpreter methodsFor: 'common selector sends' stamp: ''!
bytecodePrimNewWithArg

	messageSelector := self specialSelector: 29.
	argumentCount := 1.
	self normalSend.
! !

!Interpreter methodsFor: 'common selector sends' stamp: 'di 6/21/97 08:58'!
bytecodePrimNext
	messageSelector := self specialSelector: 19.
	argumentCount := 0.
	self normalSend.! !

!Interpreter methodsFor: 'common selector sends' stamp: 'di 6/21/97 10:12'!
bytecodePrimNextPut
	messageSelector := self specialSelector: 20.
	argumentCount := 1.
	self normalSend.! !

!Interpreter methodsFor: 'common selector sends' stamp: 'tpr 3/24/2004 18:32'!
bytecodePrimNotEqual
	| rcvr arg aBool |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(self areIntegers: rcvr and: arg) ifTrue: [^self booleanCheat: rcvr ~= arg].

	successFlag := true.
	aBool := self primitiveFloatEqual: rcvr toArg: arg.
	successFlag ifTrue: [^self booleanCheat: aBool not].

	messageSelector := self specialSelector: 7.
	argumentCount := 1.
	self normalSend
! !

!Interpreter methodsFor: 'common selector sends' stamp: 'tpr 2/11/2004 20:33'!
bytecodePrimPointX

	| rcvr |
	successFlag := true.
	rcvr := self internalStackTop.
	self assertClassOf: rcvr is: (self splObj: ClassPoint).
	successFlag
		ifTrue: [self internalPop: 1 thenPush: (self fetchPointer: XIndex ofObject: rcvr).
			^ self fetchNextBytecode "success"].

	messageSelector := self specialSelector: 30.
	argumentCount := 0.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'tpr 2/11/2004 20:34'!
bytecodePrimPointY

	| rcvr |
	successFlag := true.
	rcvr := self internalStackTop.
	self assertClassOf: rcvr is: (self splObj: ClassPoint).
	successFlag
		ifTrue: [self internalPop: 1 thenPush: (self fetchPointer: YIndex ofObject: rcvr).
			^ self fetchNextBytecode "success"].

	messageSelector := self specialSelector: 31.
	argumentCount := 0.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'ar 3/5/2008 15:16'!
bytecodePrimSize
	| rcvr sz |
	successFlag := true.
	rcvr := self internalStackTop.
	self assertClassOf: rcvr is: (self splObj: ClassArray).
	successFlag ifTrue:[
		sz := self lengthOf: rcvr.
		self internalPop: 1 thenPush: (self integerObjectOf: sz).
		^self fetchNextBytecode].
	messageSelector := self specialSelector: 18.
	argumentCount := 0.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'tpr 3/24/2004 18:32'!
bytecodePrimSubtract
	| rcvr arg result |
	rcvr := self internalStackValue: 1.
	arg := self internalStackValue: 0.
	(self areIntegers: rcvr and: arg)
		ifTrue: [result := (self integerValueOf: rcvr) - (self integerValueOf: arg).
				(self isIntegerValue: result) ifTrue:
					[self internalPop: 2 thenPush: (self integerObjectOf: result).
					^self fetchNextBytecode "success"]]
		ifFalse: [successFlag := true.
				self externalizeIPandSP.
				self primitiveFloatSubtract: rcvr fromArg: arg.
				self internalizeIPandSP.
				successFlag ifTrue: [^self fetchNextBytecode "success"]].

	messageSelector := self specialSelector: 1.
	argumentCount := 1.
	self normalSend! !

!Interpreter methodsFor: 'common selector sends' stamp: 'eem 9/4/2013 17:16'!
bytecodePrimValue
	"In-line value for BlockClosure and BlockContext"
	| maybeBlock rcvrClass |
	maybeBlock := self internalStackTop.
	argumentCount := 0.
	successFlag := true.
	(self isNonIntegerObject: maybeBlock) ifTrue:
		[rcvrClass := self fetchClassOfNonImm: maybeBlock.
		 rcvrClass = (self splObj: ClassBlockClosure)
			ifTrue:
				[self externalizeIPandSP.
				 self primitiveClosureValue.
				 self internalizeIPandSP]
			ifFalse:
				[rcvrClass = (self splObj: ClassBlockContext)
					ifTrue:
						[self externalizeIPandSP.
						 self primitiveValue.
						 self internalizeIPandSP]
					ifFalse:
						[successFlag := false]]].
	successFlag ifFalse:
		[messageSelector := self specialSelector: 25.
		 ^self normalSend].
	self fetchNextBytecode! !

!Interpreter methodsFor: 'common selector sends' stamp: 'eem 9/4/2013 17:16'!
bytecodePrimValueWithArg
	"In-line value: for BlockClosure and BlockContext"
	| maybeBlock rcvrClass |
	maybeBlock := self internalStackValue: 1.
	argumentCount := 1.
	successFlag := true.
	(self isNonIntegerObject: maybeBlock) ifTrue:
		[rcvrClass := self fetchClassOfNonImm: maybeBlock.
		 rcvrClass = (self splObj: ClassBlockClosure)
			ifTrue:
				[self externalizeIPandSP.
				 self primitiveClosureValue.
				 self internalizeIPandSP]
			ifFalse:
				[rcvrClass = (self splObj: ClassBlockContext)
					ifTrue:
						[self externalizeIPandSP.
						 self primitiveValue.
						 self internalizeIPandSP]
					ifFalse:
						[successFlag := false]]].
	successFlag ifFalse:
		[messageSelector := self specialSelector: 26.
		 ^self normalSend].
	self fetchNextBytecode! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'tpr (auto pragmas 12/08) 5/12/2005 22:48'!
callExternalPrimitive: functionID
	"Call the external plugin function identified. In the VM this is an address, see 	InterpreterSimulator for it's version. "

	<var: #functionID type: 'void *'>
	self dispatchFunctionPointer: functionID! !

!Interpreter methodsFor: 'callback support' stamp: 'eem 9/1/2020 14:27'!
callbackEnter: callbackID
	"Re-enter the interpreter for executing a callback"
	| result activeProc |
	<export: true>
	<var: #callbackID declareC: 'sqInt *callbackID'>

	"For now, do not allow a callback unless we're in a primitiveResponse"
	primitiveIndex = 0 ifTrue:[^false].

	"Check if we've exceeded the callback depth"
	jmpDepth >= jmpMax ifTrue:[^false].
	jmpDepth := jmpDepth + 1.

	"Suspend the currently active process"
	activeProc := self fetchPointer: ActiveProcessIndex
						 ofObject: self schedulerPointer.
	suspendedCallbacks at: jmpDepth put: activeProc.
	"We need to preserve newMethod explicitly since it is not activated yet
	and therefore no context has been created for it. If the caller primitive
	for any reason decides to fail we need to make sure we execute the correct
	method and not the one 'last used' in the call back"
	suspendedMethods at: jmpDepth put: newMethod.
	self transferTo: self wakeHighestPriority.

	"Typically, invoking the callback means that some semaphore has been 
	signaled to indicate the callback. Force an interrupt check right away."
	self forceInterruptCheck.

	result := self _setjmp: (jmpBuf at: jmpDepth).
	result == 0 ifTrue:["Fill in callbackID"
		callbackID at: 0 put: jmpDepth.
		"This is ugly but the inliner treats interpret() in very special and strange ways and calling any kind of 'self interpret' either directly or even via cCode:inSmalltalk: will cause this entire method to vanish."
		self cCode: 'interpret()'.
	].

	"Transfer back to the previous process so that caller can push result"
	activeProc := self fetchPointer: ActiveProcessIndex
						 ofObject: self schedulerPointer.
	self putToSleep: activeProc.
	activeProc := suspendedCallbacks at: jmpDepth.
	newMethod := suspendedMethods at: jmpDepth.	"see comment above"
	self transferTo: activeProc.
	jmpDepth := jmpDepth-1.
	^true! !

!Interpreter methodsFor: 'callback support' stamp: 'eem 9/1/2020 14:24'!
callbackLeave: cbID
	"Leave from a previous callback"
	<export: true>

	"For now, do not allow a callback unless we're in a primitiveResponse"
	primitiveIndex = 0 ifTrue:[^false].

	"Check if this is the top-level callback"
	cbID = jmpDepth ifFalse:[^false].
	cbID < 1 ifTrue:[^false].

	"Pop the arguments of the return primitive"
	self pop: argumentCount.

	"This is ugly but necessary, or otherwise the Mac will not build"
	self _longjmp: (jmpBuf at: jmpDepth) _: 1.
	"NOTREACHED"
	^nil! !

!Interpreter methodsFor: 'contexts' stamp: 'tpr 3/24/2004 20:59'!
caller
	^self fetchPointer: CallerIndex ofObject: activeContext! !

!Interpreter methodsFor: 'debug support' stamp: 'eem 9/8/2009 11:16'!
capturePendingFinalizationSignals
	statPendingFinalizationSignals := pendingFinalizationSignals.
! !

!Interpreter methodsFor: 'array primitive support' stamp: 'di (auto pragmas 12/08) 12/10/1998 14:53'!
characterForAscii: ascii  "Arg must lie in range 0-255!!"
	<inline: true>
	^ self fetchPointer: ascii ofObject: (self splObj: CharacterTable)! !

!Interpreter methodsFor: 'initialization' stamp: 'eem 12/16/2008 18:12'!
checkAssumedCompactClasses
	"Check that the classes the VM assumes are compact have the right indices."

	self checkCompactIndex: ClassBlockContextCompactIndex isClass: ClassBlockContext named: 'BlockContext'.
	self checkCompactIndex: ClassMethodContextCompactIndex isClass: ClassMethodContext named: 'MethodContext'! !

!Interpreter methodsFor: 'arithmetic primitive support' stamp: 'di 11/27/1998 15:38'!
checkBooleanResult: result
	successFlag
		ifTrue: [self pushBool: result]
		ifFalse: [self unPop: 2]! !

!Interpreter methodsFor: 'process primitive support' stamp: 'eem (auto pragmas 12/08) 11/3/2008 13:41'!
checkForInterrupts
	"Check for possible interrupts and handle one if necessary."
	| sema now |
	<inline: false>

	statCheckForEvents := statCheckForEvents + 1.

	(nextProfileTick > 0 and:[self ioHighResClock >= nextProfileTick]) ifTrue:[
		"Take a sample for the profiler if it is active. Note that this must
		be done before any of the synchronousSignals below or else we will
		attribute a pause in ioRelinquishProcessor to the newly activated process."
		profileProcess := self fetchPointer: ActiveProcessIndex 
									ofObject: self schedulerPointer.
		"and signal the profiler semaphore if it is present"
		profileSemaphore == nilObj 
			ifFalse:[self synchronousSignal: profileSemaphore].
		nextProfileTick := 0.
	] ifFalse:[
		"Reset process and method to avoid creating memory leaks"
		profileProcess := nilObj.
		profileMethod := nilObj.
	].

	"Mask so same wrapping as primitiveMillisecondClock"
	now := self ioMSecs bitAnd: MillisecondClockMask.

	self adjustInterruptCheckCounterFeedback ifTrue: [
		"don't play with the feedback if we forced a check. It only makes life difficult"
		now - lastTick < interruptChecksEveryNms
			ifTrue: ["wrapping is not a concern, it'll get caught quickly  
				enough. This clause is trying to keep a reasonable  
				guess of how many times per 	interruptChecksEveryNms we are calling  
				quickCheckForInterrupts. Not sure how effective it really is."
				interruptCheckCounterFeedBackReset := interruptCheckCounterFeedBackReset + 10]
			ifFalse: [interruptCheckCounterFeedBackReset <= 1000
					ifTrue: [interruptCheckCounterFeedBackReset := 1000]
					ifFalse: [interruptCheckCounterFeedBackReset := interruptCheckCounterFeedBackReset - 12]]].

	"reset the interrupt check counter"
	interruptCheckCounter := interruptCheckCounterFeedBackReset.

	signalLowSpace
		ifTrue: [signalLowSpace := false. "reset flag"
			sema := self splObj: TheLowSpaceSemaphore.
			sema = nilObj ifFalse: [self synchronousSignal: sema]].

	now < lastTick
		ifTrue: ["millisecond clock wrapped so correct the nextPollTick"
			nextPollTick := nextPollTick - MillisecondClockMask - 1].
	now >= nextPollTick
		ifTrue: [statIOProcessEvents := statIOProcessEvents + 1.
		 	self ioProcessEvents.
			"sets interruptPending if interrupt key pressed"
			nextPollTick := now + 200
			"msecs to wait before next call to ioProcessEvents.  
			Note that strictly speaking we might need to update  
			'now' at this point since ioProcessEvents could take a  
			very long time on some platforms"].
	interruptPending
		ifTrue: [interruptPending := false.
			"reset interrupt flag"
			sema := self splObj: TheInterruptSemaphore.
			sema = nilObj
				ifFalse: [self synchronousSignal: sema]].

	nextWakeupTick ~= 0
		ifTrue: [now < lastTick
				ifTrue: ["the clock has wrapped. Subtract the wrap  
					interval from nextWakeupTick - this might just  
					possibly result in 0. Since this is used as a flag  
					value for 'no timer' we do the 0 check above"
					nextWakeupTick := nextWakeupTick - MillisecondClockMask - 1].
			now >= nextWakeupTick
				ifTrue: [nextWakeupTick := 0.
					"set timer interrupt to 0 for 'no timer'"
					sema := self splObj: TheTimerSemaphore.
					sema = nilObj ifFalse: [self synchronousSignal: sema]]].

	"signal any pending finalizations"
	pendingFinalizationSignals > 0
		ifTrue: [sema := self splObj: TheFinalizationSemaphore.
			(self fetchClassOf: sema) = (self splObj: ClassSemaphore)
				ifTrue: [self synchronousSignal: sema].
			pendingFinalizationSignals := 0].

	"signal all semaphores in semaphoresToSignal"
	(semaphoresToSignalCountA > 0 or: [semaphoresToSignalCountB > 0])
		ifTrue: [self signalExternalSemaphores].

	"update the tracking value"
	lastTick := now! !

!Interpreter methodsFor: 'image save/restore' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:23'!
checkImageVersionFrom: f startingAt: imageOffset
	"Read and verify the image file version number and return true if the the given image file needs to be byte-swapped. As a side effect, position the file stream just after the version number of the image header. This code prints a warning and does a hard-exit if it cannot find a valid version number."
	"This code is based on C code by Ian Piumarta."

	| version firstVersion |
	<var: #f type: 'sqImageFile '>
	<var: #imageOffset type: 'squeakFileOffsetType '>

	"check the version number"
	self sqImageFile: f Seek: imageOffset.
	version := firstVersion := self getLongFromFile: f swap: false.
	(self readableFormat: version) ifTrue: [^ false].

	"try with bytes reversed"
	self sqImageFile: f Seek: imageOffset.
	version := self getLongFromFile: f swap: true.
	(self readableFormat: version) ifTrue: [^ true].

	"Note: The following is only meaningful if not reading an embedded image"
	imageOffset = 0 ifTrue:[
		"try skipping the first 512 bytes (prepended by certain Mac file transfer utilities)"
		self sqImageFile: f Seek: 512.
		version := self getLongFromFile: f swap: false.
		(self readableFormat: version) ifTrue: [^ false].

		"try skipping the first 512 bytes with bytes reversed"
		self sqImageFile: f Seek: 512.
		version := self getLongFromFile: f swap: true.
		(self readableFormat: version) ifTrue: [^ true]].

	"hard failure; abort"
	self print: 'This interpreter (vers. '.
	self printNum: self imageFormatVersion.
	self print: ') cannot read image file (vers. '.
	self printNum: firstVersion.
	self print: ').'.
	self cr.
	self print: 'Press CR to quit...'.
	self getchar.
	self ioExit.
! !

!Interpreter methodsFor: 'arithmetic primitive support' stamp: 'di 11/27/1998 15:26'!
checkIntegerResult: integerResult
	(successFlag and: [self isIntegerValue: integerResult])
		ifTrue: [self pushInteger: integerResult]
		ifFalse: [self unPop: 2]! !

!Interpreter methodsFor: 'object memory support' stamp: 'eem 1/5/2017 13:19'!
checkInterpreterIntegrity
	"Perform an integrity/leak check using the heapMap.  Assume
	 clearLeakMapAndMapAccessibleObjects has set a bit at each
	 object's header.  Check that all oops in the interpreter's state
	 points to a header.  Answer 0 if all checks pass."
	| flags |
	flags := 0.
	(self checkOopIntegrity: specialObjectsOop named: 'specialObjectsOop')ifFalse:
		[flags := 1].
	compilerInitialized
		ifTrue:
			[(self checkOopIntegrity: receiver named: 'receiver')ifFalse:
				[flags := 2].
			(self checkOopIntegrity: method named: 'method')ifFalse:
				[flags := 4]]
		ifFalse:
			[(self checkOopIntegrity: activeContext named: 'activeContext') ifFalse:
				[flags := 8]].
	(self checkOopIntegrity: messageSelector named: 'messageSelector')ifFalse:
		[flags := 16].
	(self checkOopIntegrity: newMethod named: 'newMethod')ifFalse:
		[flags := 32].
	(self checkOopIntegrity: lkupClass named: 'lkupClass')ifFalse:
		[flags := 64].
	(self checkOopIntegrity: receiverClass named: 'receiverClass')ifFalse:
		[flags := 128].
	(self checkOopIntegrity: profileProcess named: 'profileProcess')ifFalse:
		[flags := 256].
	(self checkOopIntegrity: profileMethod named: 'profileMethod')ifFalse:
		[flags := 512].
	(self checkOopIntegrity: profileSemaphore named: 'profileSemaphore')ifFalse:
		[flags := 1024].

	"Callback support - check suspended callback list"
	1 to: jmpDepth do:
		[:i|
		(self checkOopIntegrity: (suspendedCallbacks at: i) named: 'suspendedCallbacks' index: i) ifFalse:
			[flags := 2048].
		(self checkOopIntegrity: (suspendedMethods at: i) named: 'suspendedMethods' index: i) ifFalse:
			[flags := 4096]].
	^flags! !

!Interpreter methodsFor: 'utilities' stamp: ''!
checkedIntegerValueOf: intOop
	"Note: May be called by translated primitive code."

	(self isIntegerObject: intOop)
		ifTrue: [ ^ self integerValueOf: intOop ]
		ifFalse: [ self primitiveFail. ^ 0 ]! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:23'!
classNameOf: aClass Is: className 
	"Check if aClass's name is className"
	| srcName name length |
	<var: #className type: 'char *'>
	<var: #srcName type: 'char *'>
	(self lengthOf: aClass) <= 6 ifTrue: [^ false].

	"Not a class but might be behavior"
	name := self fetchPointer: 6 ofObject: aClass.
	(self isBytes: name) ifFalse: [^ false].
	length := self stSizeOf: name.
	srcName := self cCoerce: (self arrayValueOf: name) to: 'char *'.
	0 to: length - 1 do: [:i | (srcName at: i) = (className at: i) ifFalse: [^ false]].
	"Check if className really ends at this point"
	^ (className at: length) = 0! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
closureNumArgs: numArgs instructionPointer: initialIP numCopiedValues: numCopied
	| newClosure |
	<inline: true>
	newClosure := self
					instantiateSmallClass: (self splObj: ClassBlockClosure)
					sizeInBytes: (self wordSize * (ClosureFirstCopiedValueIndex + numCopied)) + self baseHeaderSize.
	"Assume: have just allocated a new closure; it must be young. Thus, can use unchecked stores."
	self storePointerUnchecked: ClosureStartPCIndex ofObject: newClosure withValue: (self integerObjectOf: initialIP).
	self storePointerUnchecked: ClosureNumArgsIndex ofObject: newClosure withValue: (self integerObjectOf: numArgs).
	"It is up to the caller to store the outer context and copiedValues."
	^newClosure! !

!Interpreter methodsFor: 'array primitive support' stamp: 'eem 9/4/2013 17:16'!
commonAt: stringy
	"This code is called if the receiver responds primitively to at:.
	If this is so, it will be installed in the atCache so that subsequent calls of at:
	or next may be handled immediately in bytecode primitive routines."
	| index rcvr atIx result |
	index := self positive32BitValueOf: (self stackTop).  "Sets successFlag"
	rcvr := self stackValue: 1.
	successFlag & (self isIntegerObject: rcvr) not
		ifFalse: [^ self primitiveFail].

	"NOTE:  The at-cache, since it is specific to the non-super response to #at:.
	Therefore we must determine that the message is #at: (not, eg, #basicAt:),
	and that the send is not a super-send, before using the at-cache."
	(messageSelector = (self specialSelector: 16)
		and: [lkupClass = (self fetchClassOfNonImm: rcvr)])
		ifTrue:
		["OK -- look in the at-cache"
		atIx := rcvr bitAnd: AtCacheMask.  "Index into atCache = 4N, for N = 0 ... 7"
		(atCache at: atIx+AtCacheOop) = rcvr ifFalse:
			["Rcvr not in cache.  Install it..."
			self install: rcvr inAtCache: atCache at: atIx string: stringy].
		successFlag ifTrue:
			[result := self commonVariable: rcvr at: index cacheIndex: atIx].
		successFlag ifTrue:
			[^ self pop: argumentCount+1 thenPush: result]].

	"The slow but sure way..."
	successFlag := true.
	result := self stObject: rcvr at: index.
	successFlag ifTrue:
		[stringy ifTrue: [result := self characterForAscii: (self integerValueOf: result)].
		^ self pop: argumentCount+1 thenPush: result]! !

!Interpreter methodsFor: 'array primitive support' stamp: 'eem 9/4/2013 17:16'!
commonAtPut: stringy
	"This code is called if the receiver responds primitively to at:Put:.
	If this is so, it will be installed in the atPutCache so that subsequent calls of at:
	or  next may be handled immediately in bytecode primitive routines."
	| value index rcvr atIx |
	value := self stackTop.
	index := self positive32BitValueOf: (self stackValue: 1).  "Sets successFlag"
	rcvr := self stackValue: 2.
	successFlag & (self isIntegerObject: rcvr) not
		ifFalse: [^ self primitiveFail].

	"NOTE:  The atPut-cache, since it is specific to the non-super response to #at:Put:.
	Therefore we must determine that the message is #at:Put: (not, eg, #basicAt:Put:),
	and that the send is not a super-send, before using the at-cache."
	(messageSelector = (self specialSelector: 17)
		and: [lkupClass = (self fetchClassOfNonImm: rcvr)])
		ifTrue:
		["OK -- look in the at-cache"
		atIx := (rcvr bitAnd: AtCacheMask) + AtPutBase.  "Index into atPutCache"
		(atCache at: atIx+AtCacheOop) = rcvr ifFalse:
			["Rcvr not in cache.  Install it..."
			self install: rcvr inAtCache: atCache at: atIx string: stringy].
		successFlag ifTrue:
			[self commonVariable: rcvr at: index put: value cacheIndex: atIx].
		successFlag ifTrue:
			[^ self pop: argumentCount+1 thenPush: value]].

	"The slow but sure way..."
	successFlag := true.
	stringy ifTrue: [self stObject: rcvr at: index put: (self asciiOfCharacter: value)]
			ifFalse: [self stObject: rcvr at: index put: value].
	successFlag ifTrue: [^ self pop: argumentCount+1 thenPush: value].
! !

!Interpreter methodsFor: 'return bytecodes' stamp: 'eem 11/12/2012 15:02'!
commonReturn
	"Note: Assumed to be inlined into the dispatch loop."

	<sharedCodeNamed: 'commonReturn' inCase: #returnReceiver>
	| nilOop thisCntx contextOfCaller localCntx localVal unwindMarked |

	nilOop := nilObj. "keep in a register"
	thisCntx := activeContext.
	localCntx := localReturnContext.
	localVal := localReturnValue.

	"make sure we can return to the given context"
	((localCntx = nilOop) or:
	 [(self fetchPointer: InstructionPointerIndex ofObject: localCntx) = nilOop]) ifTrue: [
		"error: sender's instruction pointer or context is nil; cannot return"
		^self internalCannotReturn: localVal].

	"If this return is not to our immediate predecessor (i.e. from a method to its sender, or from a block to its caller), scan the stack for the first unwind marked context and inform this context and let it deal with it. This provides a chance for ensure unwinding to occur."
	thisCntx := self fetchPointer: SenderIndex ofObject: activeContext.

	"Just possibly a faster test would be to compare the homeContext and activeContext - they are of course different for blocks. Thus we might be able to optimise a touch by having a different returnTo for the blockreteurn (since we know that must return to caller) and then if active ~= home we must be doing a non-local return. I think. Maybe."
	[thisCntx = localCntx] whileFalse:
		[thisCntx = nilOop ifTrue:[
			"error: sender's instruction pointer or context is nil; cannot return"
			^self internalCannotReturn: localVal].
		"Climb up stack towards localCntx. Break out to a send of #aboutToReturn:through: if an unwind marked context is found"
		unwindMarked := self isUnwindMarked: thisCntx.
		unwindMarked ifTrue:[
			"context is marked; break out"
			^self internalAboutToReturn: localVal through: thisCntx].
		thisCntx := self fetchPointer: SenderIndex ofObject: thisCntx].

	"If we get here there is no unwind to worry about. Simply terminate the stack up to the localCntx - often just the sender of the method"
	thisCntx := activeContext.
	[thisCntx = localCntx]
		whileFalse:
		["climb up stack to localCntx"
		contextOfCaller := self fetchPointer: SenderIndex ofObject: thisCntx.

		"zap exited contexts so any future attempted use will be caught"
		self storePointerUnchecked: SenderIndex ofObject: thisCntx withValue: nilOop.
		self storePointerUnchecked: InstructionPointerIndex ofObject: thisCntx withValue: nilOop.
		reclaimableContextCount > 0 ifTrue:
			["try to recycle this context"
			reclaimableContextCount := reclaimableContextCount - 1.
			self recycleContextIfPossible: thisCntx].
		thisCntx := contextOfCaller].

	activeContext := thisCntx.
	(self oop: thisCntx isLessThan: youngStart) ifTrue: [ self beRootIfOld: thisCntx ].

	self internalFetchContextRegisters: thisCntx.  "updates local IP and SP"
	self fetchNextBytecode.
	self internalPush: localVal.
! !

!Interpreter methodsFor: 'message sending' stamp: 'eem 11/12/2012 15:10'!
commonSend
	"Send a message, starting lookup with the receiver's class."
	"Assume: messageSelector and argumentCount have been set, and that 
	the receiver and arguments have been pushed onto the stack,"
	"Note: This method is inlined into the interpreter dispatch loop."
	<sharedCodeNamed: 'commonSend' inCase: #singleExtendedSendBytecode>
	self internalFindNewMethod.
	self internalExecuteNewMethod.
	self fetchNextBytecode! !

!Interpreter methodsFor: 'array primitive support' stamp: 'eem 7/18/2009 17:52'!
commonVariable: rcvr at: index cacheIndex: atIx 
	"This code assumes the receiver has been identified at location atIx in the atCache."
	| stSize fmt fixedFields result |

	stSize := atCache at: atIx+AtCacheSize.
	((self oop: index isGreaterThanOrEqualTo: 1)
		and: [self oop: index isLessThanOrEqualTo: stSize])
	ifTrue:
		[fmt := atCache at: atIx+AtCacheFmt.
		fmt <= 4 ifTrue:
			[fixedFields := atCache at: atIx+AtCacheFixedFields.
			^ self fetchPointer: index + fixedFields - 1 ofObject: rcvr].
		fmt < 8 ifTrue:  "Bitmap"
			[result := self fetchLong32: index - 1 ofObject: rcvr.
			result := self positive32BitIntegerFor: result.
			^ result].
		fmt >= 16  "Note fmt >= 16 is an artificial flag for strings"
			ifTrue: "String"
			[^ self characterForAscii: (self fetchByte: index - 1 ofObject: rcvr)]
			ifFalse: "ByteArray"
			[^ self integerObjectOf: (self fetchByte: index - 1 ofObject: rcvr)]].

	self primitiveFail! !

!Interpreter methodsFor: 'array primitive support' stamp: 'eem 7/18/2009 17:51'!
commonVariable: rcvr at: index put: value cacheIndex: atIx
	"This code assumes the receiver has been identified at location atIx in the atCache."
	| stSize fmt fixedFields valToPut |
	<inline: true>

	stSize := atCache at: atIx+AtCacheSize.
	((self oop: index isGreaterThanOrEqualTo: 1)
		and: [self oop: index isLessThanOrEqualTo: stSize])
	ifTrue:
		[fmt := atCache at: atIx+AtCacheFmt.
		fmt <= 4 ifTrue:
			[fixedFields := atCache at: atIx+AtCacheFixedFields.
			^ self storePointer: index + fixedFields - 1 ofObject: rcvr withValue: value].
		fmt < 8 ifTrue:  "Bitmap"
			[valToPut := self positive32BitValueOf: value.
			successFlag ifTrue: [self storeLong32: index - 1 ofObject: rcvr withValue: valToPut].
			^ nil].
		fmt >= 16  "Note fmt >= 16 is an artificial flag for strings"
			ifTrue: [valToPut := self asciiOfCharacter: value.
					successFlag ifFalse: [^ nil]]
			ifFalse: [valToPut := value].
		(self isIntegerObject: valToPut) ifTrue:
			[valToPut := self integerValueOf: valToPut.
			((valToPut >= 0) and: [valToPut <= 255]) ifFalse: [^ self primitiveFail].
			^ self storeByte: index - 1 ofObject: rcvr withValue: valToPut]].

	self primitiveFail! !

!Interpreter methodsFor: 'array primitive support' stamp: 'eem 7/18/2009 17:52'!
commonVariableInternal: rcvr at: index cacheIndex: atIx 
	"This code assumes the receiver has been identified at location atIx in the atCache."
	| stSize fmt fixedFields result |
	<inline: true>

	stSize := atCache at: atIx+AtCacheSize.
	((self oop: index isGreaterThanOrEqualTo: 1)
		and: [self oop: index isLessThanOrEqualTo: stSize])
	ifTrue:
		[fmt := atCache at: atIx+AtCacheFmt.
		fmt <= 4 ifTrue:
			[fixedFields := atCache at: atIx+AtCacheFixedFields.
			^ self fetchPointer: index + fixedFields - 1 ofObject: rcvr].
		fmt < 8 ifTrue:  "Bitmap"
			[result := self fetchLong32: index - 1 ofObject: rcvr.
			self externalizeIPandSP.
			result := self positive32BitIntegerFor: result.
			self internalizeIPandSP.
			^ result].
		fmt >= 16  "Note fmt >= 16 is an artificial flag for strings"
			ifTrue: "String"
			[^ self characterForAscii: (self fetchByte: index - 1 ofObject: rcvr)]
			ifFalse: "ByteArray"
			[^ self integerObjectOf: (self fetchByte: index - 1 ofObject: rcvr)]].

	self primitiveFail! !

!Interpreter methodsFor: 'arithmetic primitive support' stamp: ''!
compare31or32Bits: obj1 equal: obj2
	"May set success to false"

	"First compare two ST integers..."
	((self isIntegerObject: obj1)
		and: [self isIntegerObject: obj2])
		ifTrue: [^ obj1 = obj2].

	"Now compare, assuming positive integers, but setting fail if not"
	^ (self positive32BitValueOf: obj1) = (self positive32BitValueOf: obj2)! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 11/20/1999 05:05'!
compilerCreateActualMessage: aMessage storingArgs: argArray
	^self cCode: 'compilerHooks[14](aMessage, argArray)'! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 1/3/1999 18:00'!
compilerFlushCache: aCompiledMethod
	^self cCode: 'compilerHooks[2](aCompiledMethod)'! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp (auto pragmas 12/08) 1/3/1999 17:59'!
compilerFlushCacheHook: aCompiledMethod
	<inline: true>
	compilerInitialized ifTrue: [self compilerFlushCache: aCompiledMethod]! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 12/12/1998 17:00'!
compilerMapFrom: memStart to: memEnd
	^self cCode: 'compilerHooks[4](memStart, memEnd)'! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp (auto pragmas 12/08) 12/12/1998 17:03'!
compilerMapHookFrom: memStart to: memEnd
	<inline: true>
	compilerInitialized ifTrue: [self compilerMapFrom: memStart to: memEnd]! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 12/19/1998 17:08'!
compilerMark
	^self cCode: 'compilerHooks[9]()'! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 12/12/1998 17:00'!
compilerPostGC
	^self cCode: 'compilerHooks[5]()'! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 12/15/1998 12:43'!
compilerPostSnapshot
	^self cCode: 'compilerHooks[8]()'! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 12/12/1998 17:01'!
compilerPreGC: fullGCFlag
	^self cCode: 'compilerHooks[3](fullGCFlag)'! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp (auto pragmas 12/08) 12/12/1998 17:03'!
compilerPreGCHook: fullGCFlag
	<inline: true>
	compilerInitialized ifTrue: [self compilerPreGC: fullGCFlag]! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 12/15/1998 13:10'!
compilerPreSnapshot
	^self cCode: 'compilerHooks[7]()'! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 12/12/1998 17:01'!
compilerProcessChange
	^self cCode: 'compilerHooks[6]()'! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 11/24/1999 07:09'!
compilerProcessChange: oldProc to: newProc
	^self cCode: 'compilerHooks[6](oldProc, newProc)'! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 10/22/1999 17:01'!
compilerTranslateMethod
	^self cCode: 'compilerHooks[1]()'! !

!Interpreter methodsFor: 'contexts' stamp: 'tpr (auto pragmas 12/08) 3/24/2004 21:00'!
context: thisCntx hasSender: aContext 
	"Does thisCntx have aContext in its sender chain?"
	| s nilOop |
	<inline: true>
	thisCntx == aContext ifTrue: [^false].
	nilOop := nilObj.
	s := self fetchPointer: SenderIndex ofObject: thisCntx.
	[s == nilOop]
		whileFalse: [s == aContext ifTrue: [^true].
			s := self fetchPointer: SenderIndex ofObject: s].
	^false! !

!Interpreter methodsFor: 'bitblt support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:23'!
copyBits
	"This entry point needs to be implemented for the interpreter proxy.
	Since BitBlt is now a plugin we need to look up BitBltPlugin:=copyBits
	and call it. This entire mechanism should eventually go away and be
	replaced with a dynamic lookup from BitBltPlugin itself but for backward
	compatibility this stub is provided"

	| fn |
	<var: #fn type: 'void *'>
	fn := self ioLoadFunction: 'copyBits' From: 'BitBltPlugin'.
	fn = 0 ifTrue: [^self primitiveFail].
	^self cCode: '((sqInt (*)(void))fn)()'! !

!Interpreter methodsFor: 'bitblt support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:24'!
copyBitsFrom: x0 to: x1 at: y
	"This entry point needs to be implemented for the interpreter proxy.
	Since BitBlt is now a plugin we need to look up BitBltPlugin:=copyBitsFrom:to:at:
	and call it. This entire mechanism should eventually go away and be
	replaced with a dynamic lookup from BitBltPlugin itself but for backward
	compatibility this stub is provided"

	| fn |
	<var: #fn type: 'void *'>
	fn := self ioLoadFunction: 'copyBitsFromtoat' From: 'BitBltPlugin'.
	fn = 0 ifTrue: [^self primitiveFail].
	^self cCode: '((sqInt (*)(sqInt, sqInt, sqInt))fn)(x0, x1, y)'! !

!Interpreter methodsFor: 'debug printing' stamp: ''!
cr
	"For testing in Smalltalk, this method should be overridden in a subclass."

	self printf: '\n'.! !

!Interpreter methodsFor: 'message sending' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
createActualMessageTo: aClass 
	"Bundle up the selector, arguments and lookupClass into a Message object. 
	In the process it pops the arguments off the stack, and pushes the message object. 
	This can then be presented as the argument of e.g. #doesNotUnderstand:. 
	ikp 11/20/1999 03:59 -- added hook for external runtime compilers."
	"remap lookupClass in case GC happens during allocation"
	| argumentArray message lookupClass |
	<inline: false> "This is a useful break-point"
	self pushRemappableOop: aClass.
	argumentArray := self instantiateClass: (self splObj: ClassArray) indexableSize: argumentCount.
	"remap argumentArray in case GC happens during allocation"
	self pushRemappableOop: argumentArray.
	message := self instantiateClass: (self splObj: ClassMessage) indexableSize: 0.
	argumentArray := self popRemappableOop.
	lookupClass := self popRemappableOop.
	self beRootIfOld: argumentArray.

	compilerInitialized
		ifTrue: [self compilerCreateActualMessage: message storingArgs: argumentArray]
		ifFalse: [self transfer: argumentCount from: stackPointer - (argumentCount - 1 * self wordSize) to: argumentArray + self baseHeaderSize.
			self pop: argumentCount thenPush: message].

	argumentCount := 1.
	self storePointer: MessageSelectorIndex ofObject: message withValue: messageSelector.
	self storePointer: MessageArgumentsIndex ofObject: message withValue: argumentArray.
	(self lastPointerOf: message) >= (MessageLookupClassIndex * self wordSize + self baseHeaderSize)
		ifTrue: ["Only store lookupClass if message has 3 fields (old images don't)"
			self storePointer: MessageLookupClassIndex ofObject: message withValue: lookupClass]! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 12/12/1998 21:44'!
disableCompiler
	compilerInitialized := false! !

!Interpreter methodsFor: 'message sending' stamp: 'ikp (auto pragmas 12/08) 8/2/2004 18:08'!
dispatchFunctionPointer: aFunctionPointer

	<var: #aFunctionPointer type: 'void *'>
	self cCode: '((void (*)(void))aFunctionPointer)()'
			inSmalltalk: [self error: 'my simulator should simulate me']! !

!Interpreter methodsFor: 'I/O primitives' stamp: 'nice 11/13/2016 13:39'!
displayBitsOf: aForm Left: l Top: t Right: r Bottom: b
	"Repaint the portion of the Smalltalk screen bounded by the affected rectangle. Used to synchronize the screen after a Bitblt to the Smalltalk Display object."

	| displayObj dispBits w h dispBitsIndex d left right top bottom surfaceHandle |
	displayObj := self splObj: TheDisplay.
	aForm = displayObj ifFalse: [^ nil].
	self success: ((self isPointers: displayObj) and: [(self lengthOf: displayObj) >= 4]).
	successFlag ifTrue: [
		dispBits := self fetchPointer: 0 ofObject: displayObj.
		w := self fetchInteger: 1 ofObject: displayObj.
		h := self fetchInteger: 2 ofObject: displayObj.
		d := self fetchInteger: 3 ofObject: displayObj.
	].
	l < 0 ifTrue:[left := 0] ifFalse: [left := l].
	r > w ifTrue: [right := w] ifFalse: [right := r].
	t < 0 ifTrue: [top := 0] ifFalse: [top := t].
	b > h ifTrue: [bottom := h] ifFalse: [bottom := b].
	((left <= right) and: [top <= bottom]) ifFalse: [^nil].
	successFlag ifTrue: [
		(self isIntegerObject: dispBits) ifTrue: [
			surfaceHandle := self integerValueOf: dispBits.
			showSurfaceFn = 0 ifTrue: [
				showSurfaceFn := self ioLoadFunction: 'ioShowSurface' From: 'SurfacePlugin'.
				showSurfaceFn = 0 ifTrue: [^self success: false]].
			self cCode:'((int (*)(sqIntptr_t, int, int, int, int))showSurfaceFn)(surfaceHandle, left, top, right-left, bottom-top)'.
		] ifFalse: [
			dispBitsIndex := dispBits + self baseHeaderSize.  "index in memory byte array"
			self cCode: 'ioShowDisplay(dispBitsIndex, w, h, d, left, right, top, bottom)'
				inSmalltalk: [self showDisplayBits: dispBitsIndex 
								w: w h: h d: d
								left: left right: right top: top bottom: bottom]
		].
	].! !

!Interpreter methodsFor: 'arithmetic primitive support' stamp: 'di 11/29/1998 11:22'!
doPrimitiveDiv: rcvr by: arg
	"Rounds negative results towards negative infinity, rather than zero."
	| result posArg posRcvr integerRcvr integerArg |
	(self areIntegers: rcvr and: arg)
		ifTrue: [integerRcvr := self integerValueOf: rcvr.
				integerArg := self integerValueOf: arg.
				self success: integerArg ~= 0]
		ifFalse: [self primitiveFail].
	successFlag ifFalse: [^ 1 "fail"].

	integerRcvr > 0
		ifTrue: [integerArg > 0
					ifTrue: [result := integerRcvr // integerArg]
					ifFalse: ["round negative result toward negative infinity"
							posArg := 0 - integerArg.
							result := 0 - ((integerRcvr + (posArg - 1)) // posArg)]]
		ifFalse: [posRcvr := 0 - integerRcvr.
				integerArg > 0
					ifTrue: ["round negative result toward negative infinity"
							result := 0 - ((posRcvr + (integerArg - 1)) // integerArg)]
					ifFalse: [posArg := 0 - integerArg.
							result := posRcvr // posArg]].
	self success: (self isIntegerValue: result).
	^ result! !

!Interpreter methodsFor: 'arithmetic primitive support' stamp: 'di 11/29/1998 10:02'!
doPrimitiveMod: rcvr by: arg
	| integerResult integerRcvr integerArg |
	(self areIntegers: rcvr and: arg)
		ifTrue: [integerRcvr := self integerValueOf: rcvr.
				integerArg := self integerValueOf: arg.
				self success: integerArg ~= 0]
		ifFalse: [self primitiveFail].
	successFlag ifFalse: [^ 1 "fail"].

	integerResult := integerRcvr \\ integerArg.

	"ensure that the result has the same sign as the integerArg"
	integerArg < 0
		ifTrue: [integerResult > 0
			ifTrue: [integerResult := integerResult + integerArg]]
		ifFalse: [integerResult < 0
			ifTrue: [integerResult := integerResult + integerArg]].
	self success: (self isIntegerValue: integerResult).
	^ integerResult
! !

!Interpreter methodsFor: 'send bytecodes' stamp: 'tpr 3/24/2004 18:35'!
doubleExtendedDoAnythingBytecode
	"Replaces the Blue Book double-extended send [132], in which the first byte was wasted on 8 bits of argument count. 
	Here we use 3 bits for the operation sub-type (opType),  and the remaining 5 bits for argument count where needed. 
	The last byte give access to 256 instVars or literals. 
	See also secondExtendedSendBytecode"
	| byte2 byte3 opType top |
	byte2 := self fetchByte.
	byte3 := self fetchByte.
	opType := byte2 >> 5.
	opType = 0 ifTrue: [messageSelector := self literal: byte3.
			argumentCount := byte2 bitAnd: 31.
			^ self normalSend].
	opType = 1 ifTrue: [messageSelector := self literal: byte3.
			argumentCount := byte2 bitAnd: 31.
			^ self superclassSend].
	self fetchNextBytecode.
	opType = 2 ifTrue: [^ self pushReceiverVariable: byte3].
	opType = 3 ifTrue: [^ self pushLiteralConstant: byte3].
	opType = 4 ifTrue: [^ self pushLiteralVariable: byte3].
	opType = 5 ifTrue: [top := self internalStackTop.
			^ self storePointer: byte3 ofObject: receiver withValue: top].
	opType = 6
		ifTrue: [top := self internalStackTop.
			self internalPop: 1.
			^ self storePointer: byte3 ofObject: receiver withValue: top].
	opType = 7
		ifTrue: [top := self internalStackTop.
			^ self storePointer: ValueIndex ofObject: (self literal: byte3) withValue: top]! !

!Interpreter methodsFor: 'initialization' stamp: 'JMM (auto pragmas 12/08) 12/28/2002 22:07'!
dummyReferToProxy
	<inline: false>
	interpreterProxy := interpreterProxy! !

!Interpreter methodsFor: 'image save/restore' stamp: 'tpr 12/18/2013 17:45'!
dumpImage: fileName
	"Dump the entire image out to the given file. Intended for debugging only."
	| f dataSize result |
	<export: true>
	<var: #f type: 'sqImageFile'>

	f := self cCode: 'sqImageFileOpen(fileName, "wb")'.
	f = nil ifTrue: [^-1].
	dataSize := endOfMemory - self startOfMemory.
	result := self cCode: 'sqImageFileWrite(pointerForOop(memory()), sizeof(unsigned char), dataSize, f)'.
	self cCode: 'sqImageFileClose(f)'.
	^result
! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:42'!
duplicateTopBytecode

	self fetchNextBytecode.
	self internalPush: self internalStackTop.
! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 12/12/1998 21:45'!
enableCompiler
	"Calling this before loading the compiler will provoke a nullCompilerHook error"

	compilerInitialized := true! !

!Interpreter methodsFor: 'message sending' stamp: 'eem 6/9/2009 10:32'!
executeNewMethod
	"Execute newMethod - either primitiveIndex must be set directly
	 (i.e. from primitiveExecuteMethod et al), or it would have been
	 set probing the method cache (i.e. primitivePerform et al)."
	primitiveIndex > 0
		ifTrue: [self primitiveResponse.
			successFlag ifTrue: [^ nil]].
	"if not primitive, or primitive failed, activate the method"
	self activateNewMethod.
	"check for possible interrupts at each real send"
	self quickCheckForInterrupts! !

!Interpreter methodsFor: 'translation support' stamp: 'eem (auto pragmas 12/08) 7/8/2008 19:55'!
expandCases
	"For translation only; noop when running in Smalltalk.
	 Must not be inlined otherwise the directive will disappear!!"
	<inline: false>! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:46'!
extendedPushBytecode

	| descriptor variableType variableIndex |
	descriptor := self fetchByte.
	self fetchNextBytecode.
	variableType := (descriptor >> 6) bitAnd: 16r3.
	variableIndex := descriptor bitAnd: 16r3F.
	variableType=0 ifTrue: [^self pushReceiverVariable: variableIndex].
	variableType=1 ifTrue: [^self pushTemporaryVariable: variableIndex].
	variableType=2 ifTrue: [^self pushLiteralConstant: variableIndex].
	variableType=3 ifTrue: [^self pushLiteralVariable: variableIndex].
! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:47'!
extendedStoreAndPopBytecode

	self extendedStoreBytecode.
	self internalPop: 1.
! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'tpr (auto pragmas 12/08) 3/24/2004 19:06'!
extendedStoreBytecode
	| descriptor variableType variableIndex association |
	<inline: true>
	descriptor := self fetchByte.
	self fetchNextBytecode.
	variableType := descriptor >> 6 bitAnd: 3.
	variableIndex := descriptor bitAnd: 63.
	variableType = 0
		ifTrue: [^ self storePointer: variableIndex ofObject: receiver withValue: self internalStackTop].
	variableType = 1
		ifTrue: [^ self storePointerUnchecked: variableIndex + TempFrameStart ofObject: localHomeContext withValue: self internalStackTop].
	variableType = 2
		ifTrue: [self error: 'illegal store'].
	variableType = 3
		ifTrue: [association := self literal: variableIndex.
			^ self storePointer: ValueIndex ofObject: association withValue: self internalStackTop]! !

!Interpreter methodsFor: 'utilities' stamp: 'ikp 6/10/2004 11:08'!
externalizeIPandSP
	"Copy the local instruction and stack pointer to global variables for use in primitives and other functions outside the interpret loop."

	instructionPointer := self oopForPointer: localIP.
	stackPointer := self oopForPointer: localSP.
	theHomeContext := localHomeContext.
! !

!Interpreter methodsFor: 'primitive support' stamp: ''!
failed

	^successFlag not! !

!Interpreter methodsFor: 'utilities' stamp: 'jm (auto pragmas 12/08) 2/15/98 17:11'!
fetchArray: fieldIndex ofObject: objectPointer
	"Fetch the instance variable at the given index of the given object. Return the address of first indexable field of resulting array object, or fail if the instance variable does not contain an indexable bytes or words object."
	"Note: May be called by translated primitive code."

	| arrayOop |
	<returnTypeC: 'void *'>
	arrayOop := self fetchPointer: fieldIndex ofObject: objectPointer.
	^ self arrayValueOf: arrayOop
! !

!Interpreter methodsFor: 'interpreter shell' stamp: 'ikp 6/10/2004 11:01'!
fetchByte
	"This method uses the preIncrement builtin function which has no Smalltalk equivalent. Thus, it must be overridden in the simulator."

	^ self byteAtPointer: localIP preIncrement! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
fetchContextRegisters: activeCntx 
	"Note: internalFetchContextRegisters: should track changes  to this method."
	| tmp |
	<inline: true>
	tmp := self fetchPointer: MethodIndex ofObject: activeCntx.
	(self isIntegerObject: tmp)
		ifTrue: ["if the MethodIndex field is an integer, activeCntx is a block context"
			tmp := self fetchPointer: HomeIndex ofObject: activeCntx.
			(self oop: tmp isLessThan: youngStart) ifTrue: [self beRootIfOld: tmp]]
		ifFalse: ["otherwise, it is a method context and is its own home context "
			tmp := activeCntx].
	theHomeContext := tmp.
	receiver := self fetchPointer: ReceiverIndex ofObject: tmp.
	method := self fetchPointer: MethodIndex ofObject: tmp.

	"the instruction pointer is a pointer variable equal to 
	method oop + ip + BaseHeaderSize 
	-1 for 0-based addressing of fetchByte 
	-1 because it gets incremented BEFORE fetching currentByte "
	tmp := self quickFetchInteger: InstructionPointerIndex ofObject: activeCntx.
	instructionPointer := method + tmp + self baseHeaderSize - 2.

	"the stack pointer is a pointer variable also..."
	tmp := self quickFetchInteger: StackPointerIndex ofObject: activeCntx.
	stackPointer := activeCntx + self baseHeaderSize + (TempFrameStart + tmp - 1 * self wordSize)! !

!Interpreter methodsFor: 'utilities' stamp: '(auto pragmas 12/08) '!
fetchFloat: fieldIndex ofObject: objectPointer
	"Fetch the instance variable at the given index of the given object. Return the C double precision floating point value of that instance variable, or fail if it is not a Float."
	"Note: May be called by translated primitive code."

	| floatOop |
	<returnTypeC: 'double'>
	floatOop := self fetchPointer: fieldIndex ofObject: objectPointer.
	^ self floatValueOf: floatOop! !

!Interpreter methodsFor: 'utilities' stamp: 'tpr (auto pragmas 12/08) 3/15/2004 19:52'!
fetchInteger: fieldIndex ofObject: objectPointer
	"Note: May be called by translated primitive code."

	| intOop |
	<inline: false>
	intOop := self fetchPointer: fieldIndex ofObject: objectPointer.
	^self checkedIntegerValueOf: intOop! !

!Interpreter methodsFor: 'utilities' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
fetchIntegerOrTruncFloat: fieldIndex ofObject: objectPointer
	"Return the integer value of the given field of the given object. If the field contains a Float, truncate it and return its integral part. Fail if the given field does not contain a small integer or Float, or if the truncated Float is out of the range of small integers."
	"Note: May be called by translated primitive code."

	| intOrFloat floatVal frac trunc |
	<inline: false>
	<var: #floatVal type: 'double '>
	<var: #frac type: 'double '>
	<var: #trunc type: 'double '>

	intOrFloat := self fetchPointer: fieldIndex ofObject: objectPointer.
	(self isIntegerObject: intOrFloat) ifTrue: [^ self integerValueOf: intOrFloat].
	self assertClassOf: intOrFloat is: (self splObj: ClassFloat).
	successFlag ifTrue: [
		self cCode: '' inSmalltalk: [floatVal := Float new: 2].
		self fetchFloatAt: intOrFloat + self baseHeaderSize into: floatVal.
		self cCode: 'frac = modf(floatVal, &trunc)'.
		"the following range check is for C ints, with range -2^31..2^31-1"
		self flag: #Dan.		"The ranges are INCORRECT if SmallIntegers are wider than 31 bits."
		self cCode: 'success((-2147483648.0 <= trunc) && (trunc <= 2147483647.0))'.].
	successFlag
		ifTrue: [^ self cCode: '((sqInt) trunc)' inSmalltalk: [floatVal truncated]]
		ifFalse: [^ 0].
! !

!Interpreter methodsFor: 'interpreter shell' stamp: 'jm 12/10/1998 16:44'!
fetchNextBytecode
	"This method fetches the next instruction (bytecode). Each bytecode method is responsible for fetching the next bytecode, preferably as early as possible to allow the memory system time to process the request before the next dispatch."

	currentBytecode := self fetchByte.
! !

!Interpreter methodsFor: 'contexts' stamp: 'eem 6/2/2009 14:31'!
fetchStackPointerOf: aContext
	"Return the stackPointer of a Context or BlockContext."
	| sp |
	<inline: true>
	sp := self fetchPointer: StackPointerIndex ofObject: aContext.
	(self isIntegerObject: sp) ifFalse: [^0].
	self assert: ReceiverIndex + (self integerValueOf: sp) < (self lengthOf: aContext).
	^self integerValueOf: sp! !

!Interpreter methodsFor: 'debug support' stamp: ''!
findClassOfMethod: meth forReceiver: rcvr

	| currClass classDict classDictSize methodArray i done |
	currClass := self fetchClassOf: rcvr.
	done := false.
	[done] whileFalse: [
		classDict := self fetchPointer: MethodDictionaryIndex ofObject: currClass.
		classDictSize := self numSlotsOf: classDict.
		methodArray := self fetchPointer: MethodArrayIndex ofObject: classDict.
		i := 0.
		[i < (classDictSize - SelectorStart)] whileTrue: [
			meth = (self fetchPointer: i ofObject: methodArray) ifTrue: [ ^currClass ].
			i := i + 1.
		].
		currClass := self fetchPointer: SuperclassIndex ofObject: currClass.
		done := currClass = nilObj.
	].
	^self fetchClassOf: rcvr    "method not found in superclass chain"! !

!Interpreter methodsFor: 'message sending' stamp: 'tpr (auto pragmas 12/08) 3/24/2004 21:13'!
findNewMethodInClass: class 
	"Find the compiled method to be run when the current 
	messageSelector is sent to the given class, setting the values 
	of 'newMethod' and 'primitiveIndex'."
	| ok |
	<inline: false>
	ok := self lookupInMethodCacheSel: messageSelector class: class.
	ok
		ifFalse: ["entry was not found in the cache; look it up the hard way "
			self lookupMethodInClass: class.
			lkupClass := class.
			self addNewMethodToCache]! !

!Interpreter methodsFor: 'debug support' stamp: 'eem 07/28/2014 07:54'!
findSelectorOfMethod: meth forReceiver: rcvr

	| currClass done classDict classDictSize methodArray i |
	currClass := self fetchClassOf: rcvr.
	done := false.
	[done] whileFalse: [
		classDict := self fetchPointer: MethodDictionaryIndex ofObject: currClass.
		classDictSize := self numSlotsOf: classDict.
		methodArray := self fetchPointer: MethodArrayIndex ofObject: classDict.
		i := 0.
		[i <= (classDictSize - SelectorStart)] whileTrue: [
			meth = (self fetchPointer: i ofObject: methodArray) ifTrue: [
				^(self fetchPointer: i + SelectorStart ofObject: classDict)
			].
			i := i + 1.
		].
		currClass := self fetchPointer: SuperclassIndex ofObject: currClass.
		done := currClass = nilObj.
	].
	^ nilObj    "method not found in superclass chain"! !

!Interpreter methodsFor: 'method lookup cache' stamp: 'eem 11/25/2008 15:15'!
flushAtCache
	"Flush the at cache. The method cache is flushed on every programming change and garbage collect."

	1 to: AtCacheTotalSize do: [ :i | atCache at: i put: 0 ]! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'tpr 5/12/2005 22:43'!
flushExternalPrimitiveOf: methodPtr
	"methodPtr is a CompiledMethod containing an external primitive. Flush the function address and session ID of the CM"
	| lit |
	(self literalCountOf: methodPtr) > 0 ifFalse:[^nil]. "Something's broken"
	lit := self literal: 0 ofMethod: methodPtr.
	((self isArray: lit) and:[(self lengthOf: lit) = 4])
		ifFalse:[^nil]. "Something's broken"
	"ConstZero is a known SmallInt so no root check needed"
	self storePointerUnchecked: 2 ofObject: lit withValue: ConstZero.
	self storePointerUnchecked: 3 ofObject: lit withValue: ConstZero.
! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'tpr 3/15/2004 20:46'!
flushExternalPrimitiveTable
	"Flush the external primitive table"
	0 to: MaxExternalPrimitiveTableSize-1 do:[:i|
		externalPrimitiveTable at: i put: 0].
! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'eem 6/23/2008 16:16'!
flushExternalPrimitives
	"Flush the references to external functions from plugin 
	primitives. This will force a reload of those primitives when 
	accessed next. 
	Note: We must flush the method cache here so that any 
	failed primitives are looked up again."
	| oop primIdx |
	oop := self firstObject.
	[self oop: oop isLessThan: endOfMemory]
		whileTrue: [(self isFreeObject: oop)
				ifFalse: [(self isCompiledMethod: oop)
						ifTrue: ["This is a compiled method"
							primIdx := self primitiveIndexOf: oop.
							primIdx = PrimitiveExternalCallIndex
								ifTrue: ["It's primitiveExternalCall"
									self flushExternalPrimitiveOf: oop]]].
			oop := self objectAfter: oop].
	self flushMethodCache.
	self flushExternalPrimitiveTable! !

!Interpreter methodsFor: 'method lookup cache' stamp: 'di 12/12/1998 23:37'!
flushMethodCache
	"Flush the method cache. The method cache is flushed on every programming change and garbage collect."

	1 to: MethodCacheSize do: [ :i | methodCache at: i put: 0 ].
	1 to: AtCacheTotalSize do: [ :i | atCache at: i put: 0 ].
! !

!Interpreter methodsFor: 'method lookup cache' stamp: 'eem 5/27/2008 17:56'!
flushMethodCacheFrom: memStart to: memEnd 
	"Flush entries in the method cache only if the oop address is within the given memory range. 
	This reduces overagressive cache clearing. Note the AtCache is fully flushed, 70% of the time 
	cache entries live in newspace, new objects die young"
	| probe |
	probe := 0.
	1 to: MethodCacheEntries do: [:i | 
			(methodCache at: probe + MethodCacheSelector) = 0
				ifFalse: [((((self oop: (methodCache at: probe + MethodCacheSelector) isGreaterThanOrEqualTo: memStart)
										and: [self oop: (methodCache at: probe + MethodCacheSelector) isLessThan: memEnd])
									or: [(self oop: (methodCache at: probe + MethodCacheClass) isGreaterThanOrEqualTo: memStart)
											and: [self oop: (methodCache at: probe + MethodCacheClass) isLessThan: memEnd]])
								or: [(self oop: (methodCache at: probe + MethodCacheMethod) isGreaterThanOrEqualTo: memStart)
										and: [self oop: (methodCache at: probe + MethodCacheMethod) isLessThan: memEnd]])
						ifTrue: [methodCache at: probe + MethodCacheSelector put: 0]].
			probe := probe + MethodCacheEntrySize].
	1 to: AtCacheTotalSize do: [:i | atCache at: i put: 0]! !

!Interpreter methodsFor: 'process primitive support' stamp: 'ikp 4/14/2006 16:30'!
forceInterruptCheck
	"force an interrupt check ASAP"
	interruptCheckCounter := -1000.
	nextPollTick := 0! !

!Interpreter methodsFor: 'I/O primitive support' stamp: 'eem 3/4/2009 14:34'!
fullDisplayUpdate
	"Repaint the entire smalltalk screen, ignoring the affected rectangle. Used in some platform's code when the Smalltalk window is brought to the front or uncovered."

	| displayObj w h |
	displayObj := self splObj: TheDisplay.
	((self isPointers: displayObj) and: [(self lengthOf: displayObj) >= 4]) ifTrue: [
		w := self fetchInteger: 1 ofObject: displayObj.
		h := self fetchInteger: 2 ofObject: displayObj.
		self displayBitsOf: displayObj Left: 0 Top: 0 Right: w Bottom: h.
		self ioForceDisplayUpdate].
	^nil! !

!Interpreter methodsFor: 'method lookup cache' stamp: 'ikp (auto pragmas 12/08) 8/2/2004 17:54'!
functionPointerFor: primIdx inClass: theClass
	"Find an actual function pointer for this primitiveIndex.  This is an
	opportunity to specialise the prim for the relevant class (format for
	example).  Default for now is simply the entry in the base primitiveTable."

	<returnTypeC: 'void *'>
	^primitiveTable at: primIdx! !

!Interpreter methodsFor: 'interpreter shell' stamp: ''!
getCurrentBytecode
	"currentBytecode will be private to the main dispatch loop in the generated code. This method allows the currentBytecode to be retrieved from global variables."

	^ self byteAt: instructionPointer! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/17/2002 12:02'!
getFullScreenFlag
	^fullScreenFlag! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/17/2002 12:02'!
getInterruptCheckCounter
	^interruptCheckCounter! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/17/2002 12:02'!
getInterruptKeycode
	^interruptKeycode! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/17/2002 12:02'!
getInterruptPending
	^interruptPending! !

!Interpreter methodsFor: 'image save/restore' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:25'!
getLongFromFile: aFile swap: swapFlag
	"Answer the next word read from aFile, byte-swapped according to the swapFlag."

	| w |
	<var: #aFile type: 'sqImageFile '>
	w := 0.
	self cCode: 'sqImageFileRead(&w, sizeof(w), 1, aFile)'.
	swapFlag
		ifTrue: [^ self byteSwapped: w]
		ifFalse: [^ w].
! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/17/2002 12:03'!
getNextWakeupTick
	^nextWakeupTick! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/17/2002 12:03'!
getSavedWindowSize
	^savedWindowSize! !

!Interpreter methodsFor: 'plugin support' stamp: 'tpr (auto pragmas 12/08) 6/17/2005 17:52'!
getThisSessionID
	"return the global session ID value"
	<inline: false>
	^globalSessionID! !

!Interpreter methodsFor: 'image save/restore' stamp: 'eem 09/28/2015 18:48'!
imageFormatCompatibilityVersion
	"This VM is backward-compatible with the immediately preceding non-closure version."

	self wordSize == 4
		ifTrue: [^6502]
		ifFalse: [^68000]! !

!Interpreter methodsFor: 'image save/restore' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
imageFormatVersion
	"Return a magic constant that changes when the image format changes. Since the image reading code uses this to detect byte ordering, one must avoid version numbers that are invariant under byte reversal."

	self wordSize == 4
		ifTrue: [^6504]
		ifFalse: [^68002]! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'eem 3/29/2011 19:14'!
includesBehavior: aClass ThatOf: aSuperclass
	"Return the equivalent of 
		aClass includesBehavior: aSuperclass.
	Note: written for efficiency and better inlining (only 1 temp)"
	| theClass |
	<inline: true>
	aSuperclass = nilObj ifTrue:
		[^false].
	theClass := aClass.
	[theClass = aSuperclass ifTrue:
		[^true].
	 theClass ~= nilObj] whileTrue:
		[theClass := self superclassOf: theClass].
	^false! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 11/20/1999 04:53'!
initCompilerHooks
	"Initialize hooks for the 'null compiler'"

	self cCode: 'compilerHooks[1]= nullCompilerHook'.
	self cCode: 'compilerHooks[2]= nullCompilerHook'.
	self cCode: 'compilerHooks[3]= nullCompilerHook'.
	self cCode: 'compilerHooks[4]= nullCompilerHook'.
	self cCode: 'compilerHooks[5]= nullCompilerHook'.
	self cCode: 'compilerHooks[6]= nullCompilerHook'.
	self cCode: 'compilerHooks[7]= nullCompilerHook'.
	self cCode: 'compilerHooks[8]= nullCompilerHook'.
	self cCode: 'compilerHooks[9]= nullCompilerHook'.
	self cCode: 'compilerHooks[10]= nullCompilerHook'.
	self cCode: 'compilerHooks[11]= nullCompilerHook'.
	self cCode: 'compilerHooks[12]= nullCompilerHook'.
	self cCode: 'compilerHooks[13]= nullCompilerHook'.
	self cCode: 'compilerHooks[14]= nullCompilerHook'.

	compilerInitialized := false! !

!Interpreter methodsFor: 'initialization' stamp: 'ar 8/6/2003 23:32'!
initialCleanup
	"Images written by VMs earlier than 3.6/3.7 will wrongly have the root bit set on the active context. Besides clearing the root bit, we treat this as a marker that these images also lack a cleanup of external primitives (which has been introduced at the same time when the root bit problem was fixed). In this case, we merely flush them from here."

	((self longAt: activeContext) bitAnd: RootBit) = 0 ifTrue:[^nil]. "root bit is clean"
	"Clean root bit of activeContext"
	self longAt: activeContext put: ((self longAt: activeContext) bitAnd: AllButRootBit).
	"Clean external primitives"
	self flushExternalPrimitives.! !

!Interpreter methodsFor: 'as yet unclassified' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
initializeContextIndices
	"Class MethodContext"
	| contextFixedSizePlusHeader |
	SenderIndex := 0.
	InstructionPointerIndex := 1.
	StackPointerIndex := 2.
	MethodIndex := 3.
	ClosureIndex := 4. "N.B. Called receiverMap in the image."
	ReceiverIndex := 5.
	TempFrameStart := 6.  "Note this is in two places!!"

	"Class BlockContext"
	CallerIndex := 0.
	BlockArgumentCountIndex := 3.
	InitialIPIndex := 4.
	HomeIndex := 5.

	"Class BlockClosure"
	ClosureOuterContextIndex := 0.
	ClosureStartPCIndex := 1.
	ClosureNumArgsIndex := 2.
	ClosureFirstCopiedValueIndex := 3.

	CtxtTempFrameStart := 6.  "Copy of TempFrameStart in Interp"
	contextFixedSizePlusHeader := CtxtTempFrameStart + 1.
	SmallContextSize := contextFixedSizePlusHeader + 16 * self wordSize.  "16 indexable fields"
	"Large contexts have 56 indexable fileds.  Max with single header word."
	"However note that in 64 bits, for now, large contexts have 3-word headers"
	LargeContextSize := contextFixedSizePlusHeader + 56 * self wordSize! !

!Interpreter methodsFor: 'initialization' stamp: 'eem 9/4/2013 17:17'!
initializeInterpreter: bytesToShift 
	"Initialize Interpreter state before starting execution of a new image."
	interpreterProxy := self sqGetInterpreterProxy.
	self dummyReferToProxy.
	self initializeObjectMemory: bytesToShift.
	self initCompilerHooks.
	self checkAssumedCompactClasses.
	metaclassSizeBits := self sizeBitsOf: (self fetchClassOfNonImm: (self splObj: ClassArray)).	"determine actual (Metaclass instSize * 4)"
	activeContext := nilObj.
	theHomeContext := nilObj.
	method := nilObj.
	receiver := nilObj.
	messageSelector := nilObj.
	newMethod := nilObj.
	lkupClass := nilObj.
	receiverClass := nilObj.
	self flushMethodCache.
	self loadInitialContext.
	self initialCleanup.
	interruptCheckCounter := 0.
	interruptCheckCounterFeedBackReset := 1000.
	interruptChecksEveryNms := 1.
	nextProfileTick := 0.
	profileSemaphore := nilObj.
	profileProcess := nilObj.
	profileMethod := nilObj.
	nextPollTick := 0.
	nextWakeupTick := 0.
	lastTick := 0.
	interruptKeycode := 2094. "cmd-. as used for Mac but no other OS"
	interruptPending := false.
	semaphoresUseBufferA := true.
	semaphoresToSignalCountA := 0.
	semaphoresToSignalCountB := 0.
	deferDisplayUpdates := false.
	pendingFinalizationSignals := statPendingFinalizationSignals := 0.
	globalSessionID := 0.
	[globalSessionID = 0]
		whileTrue: [globalSessionID := self
						cCode: 'time(NULL) + ioMSecs()'
						inSmalltalk: [(Random new next * SmallInteger maxVal) asInteger]].
	jmpDepth := 0.
	jmpMax := MaxJumpBuf. "xxxx: Must match the definition of jmpBuf and suspendedCallbacks"
	statQuickCheckForEvents := 0.
	statCheckForEvents := 0.
	statProcessSwitch := 0.
	statIOProcessEvents := 0
! !

!Interpreter methodsFor: 'array primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:26'!
install: rcvr inAtCache: cache at: atIx string: stringy
	"Install the oop of this object in the given cache (at or atPut), along with
	its size, format and fixedSize"
	| hdr fmt totalLength fixedFields |
	<var: #cache type: 'sqInt *'>

	hdr := self baseHeader: rcvr.
	fmt := (hdr >> 8) bitAnd: 16rF.
	(fmt = 3 and: [self isContextHeader: hdr]) ifTrue:
		["Contexts must not be put in the atCache, since their size is not constant"
		^ self primitiveFail].
	totalLength := self lengthOf: rcvr baseHeader: hdr format: fmt.
	fixedFields := self fixedFieldsOf: rcvr format: fmt length: totalLength.

	cache at: atIx+AtCacheOop put: rcvr.
	stringy ifTrue: [cache at: atIx+AtCacheFmt put: fmt + 16]  "special flag for strings"
			ifFalse: [cache at: atIx+AtCacheFmt put: fmt].
	cache at: atIx+AtCacheFixedFields put: fixedFields.
	cache at: atIx+AtCacheSize put: totalLength - fixedFields.
! !

!Interpreter methodsFor: 'return bytecodes' stamp: 'ar (auto pragmas 12/08) 3/6/2001 15:21'!
internalAboutToReturn: resultObj through: aContext
	<inline: true>
	self internalPush: activeContext.
	self internalPush: resultObj.
	self internalPush: aContext.
	messageSelector := self splObj: SelectorAboutToReturn.
	argumentCount := 2.
	^self normalSend! !

!Interpreter methodsFor: 'message sending' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
internalActivateNewMethod
	| methodHeader newContext tempCount argCount2 needsLarge where |
	<inline: true>

	methodHeader := self methodHeaderOf: newMethod.
	needsLarge := methodHeader bitAnd: LargeContextBit.
	(needsLarge = 0 and: [freeContexts ~= NilContext])
		ifTrue: [newContext := freeContexts.
				freeContexts := self fetchPointer: 0 ofObject: newContext]
		ifFalse: ["Slower call for large contexts or empty free list"
				self externalizeIPandSP.
				newContext := self allocateOrRecycleContext: needsLarge.
				self internalizeIPandSP].
	tempCount := (methodHeader >> 19) bitAnd: 16r3F.

	"Assume: newContext will be recorded as a root if necessary by the
	 call to newActiveContext: below, so we can use unchecked stores."
	where :=   newContext + self baseHeaderSize.
	self longAt: where + (SenderIndex << self shiftForWord) put: activeContext.
	self longAt: where + (InstructionPointerIndex << self shiftForWord)
		put: (self integerObjectOf: (((LiteralStart + (self literalCountOfMethodHeader: methodHeader)) * self wordSize) + 1)).
	self longAt: where + (StackPointerIndex << self shiftForWord) put: (self integerObjectOf: tempCount).
	self longAt: where + (MethodIndex << self shiftForWord) put: newMethod.
	self longAt: where + (ClosureIndex << self shiftForWord) put: nilObj.

	"Copy the receiver and arguments..."
	argCount2 := argumentCount.
	0 to: argCount2 do:
		[:i | self longAt: where + ((ReceiverIndex+i) << self shiftForWord) put: (self internalStackValue: argCount2-i)].

	"clear remaining temps to nil in case it has been recycled"
	methodHeader := nilObj.  "methodHeader here used just as faster (register?) temp"
	argCount2+1+ReceiverIndex to: tempCount+ReceiverIndex do:
		[:i | self longAt: where + (i << self shiftForWord) put: methodHeader].

	self internalPop: argCount2 + 1.
	reclaimableContextCount := reclaimableContextCount + 1.
	self internalNewActiveContext: newContext.
 ! !

!Interpreter methodsFor: 'return bytecodes' stamp: 'ar (auto pragmas 12/08) 3/6/2001 15:21'!
internalCannotReturn: resultObj
	<inline: true>
	self internalPush: activeContext.
	self internalPush: resultObj.
	messageSelector := self splObj: SelectorCannotReturn.
	argumentCount := 1.
	^ self normalSend! !

!Interpreter methodsFor: 'message sending' stamp: 'ar (auto pragmas 12/08) 6/22/2007 14:25'!
internalExecuteNewMethod
	| localPrimIndex |
	<inline: true>
	localPrimIndex := primitiveIndex.
	localPrimIndex > 0
		ifTrue: [(localPrimIndex > 255
					and: [localPrimIndex < 520])
				ifTrue: ["Internal return instvars"
					localPrimIndex >= 264
						ifTrue: [^ self internalPop: 1 thenPush: (self fetchPointer: localPrimIndex - 264 ofObject: self internalStackTop)]
						ifFalse: ["Internal return constants"
							localPrimIndex = 256 ifTrue: [^ nil].
							localPrimIndex = 257 ifTrue: [^ self internalPop: 1 thenPush: trueObj].
							localPrimIndex = 258 ifTrue: [^ self internalPop: 1 thenPush: falseObj].
							localPrimIndex = 259 ifTrue: [^ self internalPop: 1 thenPush: nilObj].
							^ self internalPop: 1 thenPush: (self integerObjectOf: localPrimIndex - 261)]]
				ifFalse: [
					self externalizeIPandSP.
					self internalPrimitiveResponse.
					self internalizeIPandSP.
					successFlag
						ifTrue: [self browserPluginReturnIfNeeded.
							^ nil]]].
	"if not primitive, or primitive failed, activate the method"
	self internalActivateNewMethod.
	"check for possible interrupts at each real send"
	self internalQuickCheckForInterrupts! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
internalFetchContextRegisters: activeCntx
	"Inlined into return bytecodes. The only difference between this method and fetchContextRegisters: is that this method sets the local IP and SP."

	| tmp |
	<inline: true>
	tmp := self fetchPointer: MethodIndex ofObject: activeCntx.
	(self isIntegerObject: tmp) ifTrue: [
		"if the MethodIndex field is an integer, activeCntx is a block context"
		tmp := self fetchPointer: HomeIndex ofObject: activeCntx.
		(self oop: tmp isLessThan: youngStart) ifTrue: [ self beRootIfOld: tmp ].
	] ifFalse: [
		"otherwise, it is a method context and is its own home context"
		tmp := activeCntx.
	].
	localHomeContext := tmp.
	receiver := self fetchPointer: ReceiverIndex ofObject: tmp.
	method := self fetchPointer: MethodIndex ofObject: tmp.

	"the instruction pointer is a pointer variable equal to
		method oop + ip + BaseHeaderSize
		  -1 for 0-based addressing of fetchByte
		  -1 because it gets incremented BEFORE fetching currentByte"
	tmp := self quickFetchInteger: InstructionPointerIndex ofObject: activeCntx.
	localIP := self pointerForOop: method + tmp + self baseHeaderSize - 2.

	"the stack pointer is a pointer variable also..."
	tmp := self quickFetchInteger: StackPointerIndex ofObject: activeCntx.
	localSP := self pointerForOop: activeCntx + self baseHeaderSize + ((TempFrameStart + tmp - 1) * self wordSize)! !

!Interpreter methodsFor: 'message sending' stamp: 'ar (auto pragmas 12/08) 7/6/2003 23:55'!
internalFindNewMethod
	"Find the compiled method to be run when the current messageSelector is sent to the class 'lkupClass', setting the values of 'newMethod' and 'primitiveIndex'."
	| ok | 
	<inline: true>
	ok := self lookupInMethodCacheSel: messageSelector class: lkupClass.
	ok ifFalse: [
		"entry was not found in the cache; look it up the hard way"
		self externalizeIPandSP.
		self lookupMethodInClass: lkupClass.
		self internalizeIPandSP.
		self addNewMethodToCache].
! !

!Interpreter methodsFor: 'contexts' stamp: 'dtl (auto pragmas 12/08) 4/22/2007 23:27'!
internalNewActiveContext: aContext
	"The only difference between this method and newActiveContext: is that this method uses internal context registers."
	<inline: true>

	self internalStoreContextRegisters: activeContext.
	(self oop: aContext isLessThan: youngStart) ifTrue: [ self beRootIfOld: aContext ].
	activeContext := aContext.
	self internalFetchContextRegisters: aContext.! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
internalPop: nItems

	localSP := localSP - (nItems * self wordSize).! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
internalPop: nItems thenPush: oop

	self longAtPointer: (localSP := localSP - ((nItems - 1) * self wordSize)) put: oop.
! !

!Interpreter methodsFor: 'primitive support' stamp: 'ar (auto pragmas 12/08) 2/27/2008 09:32'!
internalPrimitiveResponse
	"Called under the assumption that primFunctionPtr has been preloaded"
	| delta nArgs |
	<inline: true>
	DoBalanceChecks ifTrue:["check stack balance"
		nArgs := argumentCount.
		delta := stackPointer - activeContext.
	].
	successFlag := true.
	self dispatchFunctionPointer: primitiveFunctionPointer.
	DoBalanceChecks ifTrue:[
		(self balancedStack: delta afterPrimitive: primitiveIndex withArgs: nArgs) 
			ifFalse:[self printUnbalancedStack: primitiveIndex].
	].
	"If we are profiling, take accurate primitive measures"
	nextProfileTick > 0 ifTrue:[
		self ioHighResClock >= nextProfileTick ifTrue:[
			"Take a profiler sample. Note that we sample the primitive method
			here because we don't want to attribute a failed primitive improperly."
			successFlag
				ifTrue: ["process the interrupt now"
						profileMethod := newMethod.
						self checkForInterrupts]
				ifFalse: ["process the interrupt in primtive failure code"
						profileMethod := nilObj.
						interruptCheckCounter := 0].
		].
	].
	primitiveIndex := 0. "clear out primIndex so VM knows we're no longer in primitive"
	^ successFlag
! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem 11/12/2012 15:15'!
internalPrimitiveValue
	<sharedCodeNamed: 'commonPrimitiveValue' inCase: #bytecodePrimValue>
	| newContext blockArgumentCount initialIP |
	successFlag := true.
	newContext := self internalStackValue: argumentCount.
	self assertClassOf: newContext is: (self splObj: ClassBlockContext).
	blockArgumentCount := self argumentCountOfBlock: newContext.

	self success: (argumentCount = blockArgumentCount and: [(self fetchPointer: CallerIndex ofObject: newContext) = nilObj]).

	successFlag
		ifTrue: ["This code assumes argCount can only = 0 or 1"
			argumentCount = 1
				ifTrue: [self storePointer: TempFrameStart ofObject: newContext withValue: self internalStackTop].
			self internalPop: argumentCount + 1.
			"copy the initialIP value to the ip slot"
			initialIP := self fetchPointer: InitialIPIndex ofObject: newContext.
			self storePointerUnchecked: InstructionPointerIndex ofObject: newContext withValue: initialIP.
			self storeStackPointerValue: argumentCount inContext: newContext.
			self storePointerUnchecked: CallerIndex ofObject: newContext withValue: activeContext.
			self internalNewActiveContext: newContext]
		ifFalse: [messageSelector := self specialSelector: 25 + argumentCount.
			self normalSend]! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
internalPush: object

	self longAtPointer: (localSP := localSP + self wordSize) put: object.! !

!Interpreter methodsFor: 'process primitive support' stamp: 'eem 11/10/2014 12:49'!
internalQuickCheckForInterrupts
	"Internal version of quickCheckForInterrupts for use within jumps."

	<inline: true>
	<asmLabel: true> 
	statQuickCheckForEvents := statQuickCheckForEvents + 1.
	((interruptCheckCounter := interruptCheckCounter - 1) <= 0) ifTrue: [
		self externalizeIPandSP.
		self checkForInterrupts.

		self browserPluginReturnIfNeeded.

		self internalizeIPandSP]! !

!Interpreter methodsFor: 'contexts' stamp: 'ikp 6/10/2004 11:16'!
internalStackTop

	^ self longAtPointer: localSP! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
internalStackValue: offset

	^ self longAtPointer: localSP - (offset * self wordSize)! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
internalStoreContextRegisters: activeCntx
	"The only difference between this method and fetchContextRegisters: is that this method stores from the local IP and SP."

	"InstructionPointer is a pointer variable equal to
	method oop + ip + BaseHeaderSize
		-1 for 0-based addressing of fetchByte
		-1 because it gets incremented BEFORE fetching currentByte"

	<inline: true>
	self storePointerUnchecked: InstructionPointerIndex ofObject: activeCntx
		withValue: (self integerObjectOf: 
			((self oopForPointer: localIP) + 2 - (method + self baseHeaderSize))).
	self storePointerUnchecked: StackPointerIndex		  ofObject: activeCntx
		withValue: (self integerObjectOf:
			((((self oopForPointer: localSP) - (activeCntx + self baseHeaderSize)) >> self shiftForWord) - TempFrameStart + 1)).
! !

!Interpreter methodsFor: 'utilities' stamp: 'ikp 6/10/2004 11:08'!
internalizeIPandSP
	"Copy the local instruction and stack pointer to local variables for rapid access within the interpret loop."

	localIP := self pointerForOop: instructionPointer.
	localSP := self pointerForOop: stackPointer.
	localHomeContext := theHomeContext.
! !

!Interpreter methodsFor: 'interpreter shell' stamp: 'eem 3/4/2009 14:33'!
interpret
	"This is the main interpreter loop. It normally loops forever, fetching and executing
	 bytecodes.  When running in the context of a browser plugin VM, however, it must
	 return control to the browser periodically.  This should done only when the state of
	 the currently running Squeak thread is safely stored in the object heap.  Since this
	 is the case at the moment that a check for interrupts is performed, that is when we
	 return to the browser if it is time to do so.  Interrupt checks happen quite frequently."

	<inline: false> "should *not* be inlined into sendInvokeCallback:Stack:Registers:Jmpbuf:"
	"record entry time when running as a browser plug-in"
	self browserPluginInitialiseIfNeeded.
	self internalizeIPandSP.
	self fetchNextBytecode.
	[true] whileTrue: [self dispatchOn: currentBytecode in: BytecodeTable].
	localIP := localIP - 1.  "undo the pre-increment of IP before returning"
	self externalizeIPandSP.
	^nil
! !

!Interpreter methodsFor: 'process primitive support' stamp: 'tpr 3/22/2004 14:20'!
interruptCheckForced
	"was this interrupt check forced by outside code?"
	^interruptCheckCounter < -100! !

!Interpreter methodsFor: 'plugin support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 17:43'!
ioFilename: aCharBuffer fromString: aFilenameString ofLength: filenameLength resolveAliases: aBoolean
"the vm has to convert aFilenameString via any canonicalization and char-mapping and put the result in aCharBuffer.
Note the resolveAliases flag - this is an awful artefact of OSX and Apples demented alias handling. When opening a file, the flag must be  true, when closing or renaming it must be false. Sigh."
	<var: #aCharBuffer type: 'char *'>
	<var: #aFilenameString type: 'char *'>
	self cCode:'sqGetFilenameFromString(aCharBuffer, aFilenameString, filenameLength, aBoolean)'
		inSmalltalk:["this doesn't translate well in Smalltalk since we know how long strings are rather than considering them terminated by a 0 char. Do the best we can"
			aCharBuffer replaceFrom:1 to: filenameLength with: aFilenameString]! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:26'!
is: oop KindOf: className
	"Support for external primitives."
	| oopClass |
	<var: #className type:'char *'>
	oopClass := self fetchClassOf: oop.
	[oopClass == nilObj] whileFalse:[
		(self classNameOf: oopClass Is: className) ifTrue:[^true].
		oopClass := self superclassOf: oopClass].
	^false! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'eem 5/10/2011 15:37'!
is: oop KindOfClass: aClass
	"Support for external primitives."
	<api>
	| oopClass |
	oopClass := self fetchClassOf: oop.
	[oopClass = nilObj] whileFalse:
		[oopClass = aClass ifTrue: [^true].
		 oopClass := self superclassOf: oopClass].
	^false! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:26'!
is: oop MemberOf: className
	"Support for external primitives"
	| oopClass |
	<var: #className type:'char *'>
	oopClass := self fetchClassOf: oop.
	^(self classNameOf: oopClass Is: className)! !

!Interpreter methodsFor: 'process primitive support' stamp: ''!
isEmptyList: aLinkedList

	^ (self fetchPointer: FirstLinkIndex ofObject: aLinkedList) = nilObj! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'eem 9/4/2013 17:17'!
isFloatObject: oop
	^(self isNonIntegerObject: oop)
	  and: [(self fetchClassOfNonImm: oop) = self classFloat]! !

!Interpreter methodsFor: 'compiled methods' stamp: 'eem 1/26/2021 16:15'!
isHandlerMarked: aContext
	"Is this a MethodContext whose meth has a primitive number of 199?"
	| header meth pIndex |
	"NB: the use of a primitive number for marking the method is pretty grungy, but it is simple to use for a test sytem, not too expensive and we don't actually have the two spare method header bits we need. We can probably obtain them when the method format is changed.
	NB 2: actually, the jitter will probably implement the prim to actually mark the volatile frame by changing the return function pointer."
	<inline: true>
	header := self baseHeader: aContext.
	(self isMethodContextHeader: header) ifFalse: [^false].
	meth := self fetchPointer: MethodIndex ofObject: aContext.
	pIndex := self primitiveIndexOf: meth.
	^pIndex == PrimNumberHandlerMarker
! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'nice 4/18/2016 23:26'!
isKindOfInteger: oop
	"Answer true if the oop is kind of Integer (Small or Large)."
	<api>
	<inline: true>
	^(self isIntegerObject: oop)
		or: [self isLargeIntegerInstance: oop]! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'nice 4/18/2016 23:38'!
isLargeIntegerObject: oop
	^(self isLargeIntegerInstance: oop)! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'nice 4/19/2016 21:45'!
isLargeNegativeIntegerObject: oop
	^(self isInstanceOfClassLargeNegativeInteger: oop)! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'nice 4/19/2016 21:44'!
isLargePositiveIntegerObject: oop
	^(self isInstanceOfClassLargePositiveInteger: oop)! !

!Interpreter methodsFor: 'compiled methods' stamp: 'ar (auto pragmas 12/08) 3/6/2001 15:04'!
isUnwindMarked: aContext
	"Is this a MethodContext whose meth has a primitive number of 198?"
	| header meth pIndex |
	"NB: the use of a primitive number for marking the method is pretty grungy, but it is simple to use for a test sytem, not too expensive and we don't actually have the two spare method header bits we need. We can probably obtain them when the method format is changed
	NB 2: actually, the jitter will probably implement the prim to actually mark the volatile frame by changing the return function pointer."
	<inline: true>
	header := self baseHeader: aContext.
	(self isMethodContextHeader: header) ifFalse: [^false].
	meth := self fetchPointer: MethodIndex ofObject: aContext.
	pIndex := self primitiveIndexOf: meth.
	^pIndex == 198
! !

!Interpreter methodsFor: 'jump bytecodes' stamp: 'ikp 6/10/2004 11:01'!
jump: offset

	localIP := localIP + offset + 1.
	currentBytecode := self byteAtPointer: localIP.
! !

!Interpreter methodsFor: 'jump bytecodes' stamp: 'tpr 3/24/2004 18:36'!
jumplfFalseBy: offset 
	| boolean |
	boolean := self internalStackTop.
	boolean = falseObj
		ifTrue: [self jump: offset]
		ifFalse: [boolean = trueObj
				ifFalse: [messageSelector := self splObj: SelectorMustBeBoolean.
					argumentCount := 0.
					^ self normalSend].
			self fetchNextBytecode].
	self internalPop: 1! !

!Interpreter methodsFor: 'jump bytecodes' stamp: 'tpr 3/24/2004 18:36'!
jumplfTrueBy: offset 
	| boolean |
	boolean := self internalStackTop.
	boolean = trueObj
		ifTrue: [self jump: offset]
		ifFalse: [boolean = falseObj
				ifFalse: [messageSelector := self splObj: SelectorMustBeBoolean.
					argumentCount := 0.
					^ self normalSend].
			self fetchNextBytecode].
	self internalPop: 1! !

!Interpreter methodsFor: 'callback support' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
justActivateNewMethod
	"Activate the new method but *do not* copy receiver or arguments from activeContext."
	| methodHeader initialIP newContext tempCount needsLarge where |
	<inline: true>

	methodHeader := self methodHeaderOf: newMethod.
	needsLarge := methodHeader bitAnd: LargeContextBit.
	(needsLarge = 0 and: [freeContexts ~= NilContext])
		ifTrue: [newContext := freeContexts.
				freeContexts := self fetchPointer: 0 ofObject: newContext]
		ifFalse: ["Slower call for large contexts or empty free list"
				newContext := self allocateOrRecycleContext: needsLarge].
	initialIP := ((LiteralStart + (self literalCountOfMethodHeader: methodHeader)) * self wordSize) + 1.
	tempCount := (methodHeader >> 19) bitAnd: 16r3F.

	"Assume: newContext will be recorded as a root if necessary by the
	 call to newActiveContext: below, so we can use unchecked stores."
	where := newContext + self baseHeaderSize.
	self longAt: where + (SenderIndex << self shiftForWord) put: activeContext.
	self longAt: where + (InstructionPointerIndex << self shiftForWord) put: (self integerObjectOf: initialIP).
	self longAt: where + (StackPointerIndex << self shiftForWord) put: (self integerObjectOf: tempCount).
	self longAt: where + (MethodIndex << self shiftForWord) put: newMethod.

	"Set the receiver..."
	self longAt: where + (ReceiverIndex << self shiftForWord) put: receiver.

	"clear all args and temps to nil in case it has been recycled"
	needsLarge := nilObj.  "needsLarge here used just as faster (register?) temp"
	ReceiverIndex + 1 to: tempCount + ReceiverIndex do:
		[:i | self longAt: where + (i << self shiftForWord) put: needsLarge].
	reclaimableContextCount := reclaimableContextCount + 1.

	activeContext := newContext.
	(self oop: newContext isLessThan: youngStart) ifTrue:
		[self beRootIfOld: newContext].
	self fetchContextRegisters: activeContext! !

!Interpreter methodsFor: 'compiled methods' stamp: 'tpr 3/24/2004 21:07'!
literal: offset
	^self literal: offset ofMethod: method! !

!Interpreter methodsFor: 'compiled methods' stamp: ''!
literal: offset ofMethod: methodPointer

	^ self fetchPointer: offset + LiteralStart ofObject: methodPointer
! !

!Interpreter methodsFor: 'bitblt support' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:26'!
loadBitBltFrom: bb
	"This entry point needs to be implemented for the interpreter proxy.
	Since BitBlt is now a plugin we need to look up BitBltPlugin:=loadBitBltFrom
	and call it. This entire mechanism should eventually go away and be
	replaced with a dynamic lookup from BitBltPlugin itself but for backward
	compatibility this stub is provided"
	| fn |
	<var: #fn type: 'void *'>
	fn := self ioLoadFunction: 'loadBitBltFrom' From: 'BitBltPlugin'.
	fn = 0 ifTrue: [^self primitiveFail].
	^self cCode: '((sqInt (*)(sqInt))fn)(bb)'! !

!Interpreter methodsFor: 'initialization' stamp: 'dtl 4/22/2007 23:29'!
loadInitialContext

	| sched proc |
	sched := self fetchPointer: ValueIndex ofObject: (self splObj: SchedulerAssociation).
	proc := self fetchPointer: ActiveProcessIndex ofObject: sched.
	activeContext := self fetchPointer: SuspendedContextIndex ofObject: proc.
	(self oop: activeContext isLessThan: youngStart) ifTrue: [ self beRootIfOld: activeContext ].
	self fetchContextRegisters: activeContext.
	reclaimableContextCount := 0.! !

!Interpreter methodsFor: 'jump bytecodes' stamp: 'tpr 3/24/2004 18:37'!
longJumpIfFalse

	self jumplfFalseBy: ((currentBytecode bitAnd: 3) * 256) + self fetchByte.! !

!Interpreter methodsFor: 'jump bytecodes' stamp: 'tpr 3/24/2004 18:37'!
longJumpIfTrue

	self jumplfTrueBy: ((currentBytecode bitAnd: 3) * 256) + self fetchByte.! !

!Interpreter methodsFor: 'jump bytecodes' stamp: 'tpr 3/24/2004 18:37'!
longUnconditionalJump

	| offset |
	offset := (((currentBytecode bitAnd: 7) - 4) * 256) + self fetchByte.
	localIP := localIP + offset.
	offset < 0 ifTrue: [
		"backward jump means we're in a loop; check for possible interrupts"
		self internalQuickCheckForInterrupts.
	].
	self fetchNextBytecode
! !

!Interpreter methodsFor: 'method lookup cache' stamp: 'eem (auto pragmas 12/08) 5/27/2008 17:53'!
lookupInMethodCacheSel: selector class: class
	"This method implements a simple method lookup cache. If an entry for the given selector and class is found in the cache, set the values of 'newMethod' and 'primitiveIndex' and return true. Otherwise, return false."
	"About the re-probe scheme: The hash is the low bits of the XOR of two large addresses, minus their useless lowest two bits. If a probe doesn't get a hit, the hash is shifted right one bit to compute the next probe, introducing a new randomish bit. The cache is probed CacheProbeMax times before giving up."
	"WARNING: Since the hash computation is based on the object addresses of the class and selector, we must rehash or flush when compacting storage. We've chosen to flush, since that also saves the trouble of updating the addresses of the objects in the cache."

	| hash probe |
	<inline: true>
	hash := selector bitXor: class.  "shift drops two low-order zeros from addresses"

	probe := hash bitAnd: MethodCacheMask.  "first probe"
	(((methodCache at: probe + MethodCacheSelector) = selector) and:
		 [(methodCache at: probe + MethodCacheClass) = class]) ifTrue:
			[newMethod := methodCache at: probe + MethodCacheMethod.
			primitiveIndex := methodCache at: probe + MethodCachePrim.
			primitiveFunctionPointer := self cCoerce: (methodCache at: probe + MethodCachePrimFunction) to: 'void *'.
			^ true	"found entry in cache; done"].

	probe := (hash >> 1) bitAnd: MethodCacheMask.  "second probe"
	(((methodCache at: probe + MethodCacheSelector) = selector) and:
		 [(methodCache at: probe + MethodCacheClass) = class]) ifTrue:
			[newMethod := methodCache at: probe + MethodCacheMethod.
			primitiveIndex := methodCache at: probe + MethodCachePrim.
			primitiveFunctionPointer := self cCoerce: (methodCache at: probe + MethodCachePrimFunction) to: 'void *'.
			^ true	"found entry in cache; done"].

	probe := (hash >> 2) bitAnd: MethodCacheMask.
	(((methodCache at: probe + MethodCacheSelector) = selector) and:
		 [(methodCache at: probe + MethodCacheClass) = class]) ifTrue:
			[newMethod := methodCache at: probe + MethodCacheMethod.
			primitiveIndex := methodCache at: probe + MethodCachePrim.
			primitiveFunctionPointer := self cCoerce: (methodCache at: probe + MethodCachePrimFunction) to: 'void *'.
			^ true	"found entry in cache; done"].

	^ false
! !

!Interpreter methodsFor: 'message sending' stamp: 'eem 9/29/2010 17:44'!
lookupMethodInClass: class
	| currentClass dictionary found rclass |
	<inline: false>

	currentClass := class.
	[currentClass ~= nilObj]
		whileTrue:
		[dictionary := self fetchPointer: MethodDictionaryIndex ofObject: currentClass.
		dictionary = nilObj ifTrue:
			["MethodDict pointer is nil (hopefully due a swapped out stub)
				-- raise exception #cannotInterpret:."
			self pushRemappableOop: currentClass.  "may cause GC!!"
			self createActualMessageTo: class.
			currentClass := self popRemappableOop.
			messageSelector := self splObj: SelectorCannotInterpret.
			^ self lookupMethodInClass: (self superclassOf: currentClass)].
		found := self lookupMethodInDictionary: dictionary.
		found ifTrue: [^currentClass].
		currentClass := self superclassOf: currentClass].

	"Could not find #doesNotUnderstand: -- unrecoverable error."
	messageSelector = (self splObj: SelectorDoesNotUnderstand) ifTrue:
		[self error: 'Recursive not understood error encountered'].

	"Cound not find a normal message -- raise exception #doesNotUnderstand:"
	self pushRemappableOop: class.  "may cause GC!!"
	self createActualMessageTo: class.
	rclass := self popRemappableOop.
	messageSelector := self splObj: SelectorDoesNotUnderstand.
	^ self lookupMethodInClass: rclass! !

!Interpreter methodsFor: 'message sending' stamp: 'eem 07/28/2014 07:54'!
lookupMethodInDictionary: dictionary 
	"This method lookup tolerates integers as Dictionary keys to 
	support execution of images in which Symbols have been 
	compacted out"
	| length index mask wrapAround nextSelector methodArray |
	<inline: true>
	length := self numSlotsOf: dictionary.
	mask := length - SelectorStart - 1.
	(self isIntegerObject: messageSelector)
		ifTrue: [index := (mask bitAnd: (self integerValueOf: messageSelector)) + SelectorStart]
		ifFalse: [index := (mask bitAnd: (self hashBitsOf: messageSelector)) + SelectorStart].

	"It is assumed that there are some nils in this dictionary, and search will 
	stop when one is encountered. However, if there are no nils, then wrapAround 
	will be detected the second time the loop gets to the end of the table."
	wrapAround := false.
	[true]
		whileTrue: [nextSelector := self fetchPointer: index ofObject: dictionary.
			nextSelector = nilObj ifTrue: [^ false].
			nextSelector = messageSelector
				ifTrue: [methodArray := self fetchPointer: MethodArrayIndex ofObject: dictionary.
					newMethod := self fetchPointer: index - SelectorStart ofObject: methodArray.
					"Check if newMethod is a CompiledMethod."
					(self isCompiledMethod: newMethod)
						ifTrue: [primitiveIndex := self primitiveIndexOf: newMethod.
							primitiveIndex > MaxPrimitiveIndex
								ifTrue: ["If primitiveIndex is out of range, set to zero before putting in 
									cache. This is equiv to primFail, and avoids the need to check on 
									every send."
									primitiveIndex := 0]]
						ifFalse: ["indicate that this is no compiled method - use primitiveInvokeObjectAsMethod"
							primitiveIndex := 248].
					^ true].
			index := index + 1.
			index = length
				ifTrue: [wrapAround
						ifTrue: [^ false].
					wrapAround := true.
					index := SelectorStart]]! !

!Interpreter methodsFor: 'callback support' stamp: 'eem 9/29/2010 17:44'!
lookupMethodNoMNUEtcInClass: class
	"Lookup.  Answer false on failure father than performing MNU processing etc."
	| currentClass dictionary |
	<inline: true>

	currentClass := class.
	[currentClass ~= nilObj] whileTrue:
		[dictionary := self fetchPointer: MethodDictionaryIndex ofObject: currentClass.
		(dictionary ~= nilObj
		 and: [self lookupMethodInDictionary: dictionary]) ifTrue:
			[^true].
		currentClass := self superclassOf: currentClass].

	^false! !

!Interpreter methodsFor: 'utilities' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
makePointwithxValue: xValue yValue: yValue
"make a Point xValue@yValue.
We know both will be integers so no value nor root checking is needed"
	| pointResult |
	pointResult := self instantiateSmallClass: (self splObj: ClassPoint) sizeInBytes: 3*self wordSize.
	self storePointerUnchecked: XIndex ofObject: pointResult withValue: (self integerObjectOf: xValue).
	self storePointerUnchecked: YIndex ofObject: pointResult withValue: (self integerObjectOf: yValue).
	^ pointResult! !

!Interpreter methodsFor: 'object memory support' stamp: 'eem 9/14/2013 11:17'!
mapInterpreterOops
	"Map all oops in the interpreter's state to their new values 
	during garbage collection or a become: operation."
	"Assume: All traced variables contain valid oops."
	| oop |
	compilerInitialized ifFalse:
		[stackPointer := stackPointer - activeContext. "*rel to active"
		 activeContext := self remap: activeContext.
		 stackPointer := stackPointer + activeContext. "*rel to active"
		 theHomeContext := self remap: theHomeContext].
	instructionPointer := instructionPointer - method. "*rel to method"
	method := self remap: method.
	instructionPointer := instructionPointer + method. "*rel to method"
	receiver := self remap: receiver.
	messageSelector := self remap: messageSelector.
	newMethod := self remap: newMethod.
	lkupClass := self remap: lkupClass.
	receiverClass := self remap: receiverClass.
	profileProcess := self remap: profileProcess.
	profileMethod := self remap: profileMethod.
	profileSemaphore := self remap: profileSemaphore.

	"Callback support - trace suspended callback list"
	1 to: jmpDepth do:[:i|
		oop := suspendedCallbacks at: i.
		(self isIntegerObject: oop) 
			ifFalse:[suspendedCallbacks at: i put: (self remap: oop)].
		oop := suspendedMethods at: i.
		(self isIntegerObject: oop) 
			ifFalse:[suspendedMethods at: i put: (self remap: oop)].
	].
! !

!Interpreter methodsFor: 'object memory support' stamp: 'eem 5/6/2011 15:08'!
markAndTraceInterpreterOops
	"Mark and trace all oops in the interpreter's state."
	"Assume: All traced variables contain valid oops."
	| oop |
	self markAndTrace: specialObjectsOop. "also covers nilObj, trueObj, falseObj, and compact classes"
	self markAndTrace: activeContext.
	self markAndTrace: messageSelector.
	self markAndTrace: newMethod.
	self markAndTrace: lkupClass.
	self markAndTrace: receiverClass.
	self markAndTrace: profileProcess.
	self markAndTrace: profileMethod.
	self markAndTrace: profileSemaphore.

	1 to: remapBufferCount do: [:i | 
			oop := remapBuffer at: i.
			(self isIntegerObject: oop) ifFalse: [self markAndTrace: oop]].

	"Callback support - trace suspended callback list"
	1 to: jmpDepth do:[:i|
		oop := suspendedCallbacks at: i.
		(self isIntegerObject: oop) ifFalse:[self markAndTrace: oop].
		oop := suspendedMethods at: i.
		(self isIntegerObject: oop) ifFalse:[self markAndTrace: oop].
	].
! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'ar 10/7/1998 18:38'!
methodArgumentCount
	^argumentCount! !

!Interpreter methodsFor: 'compiled methods' stamp: 'tpr 3/24/2004 21:08'!
methodClassOf: methodPointer

	^ self fetchPointer: ValueIndex ofObject: (self literal: (self literalCountOf: methodPointer) - 1 ofMethod: methodPointer)! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'ar 10/7/1998 18:39'!
methodPrimitiveIndex
	^primitiveIndex! !

!Interpreter methodsFor: 'initialization' stamp: 'tpr (auto pragmas 12/08) 3/24/2004 21:22'!
moduleUnloaded: aModuleName 
	"The module with the given name was just unloaded. 
	Make sure we have no dangling references."
	<export: true>
	<var: #aModuleName type: 'char *'>
	(aModuleName strcmp: 'SurfacePlugin') = 0
		ifTrue: ["Surface plugin went away. Should never happen. But  then, who knows"
			showSurfaceFn := 0]! !

!Interpreter methodsFor: 'contexts' stamp: 'eem 6/2/2009 14:30'!
nacFetchStackPointerOf: aContext
	"A version of fetchStackPointerOf: for use when objects may be forwarded.
	 Does not do an assert-check of the stack pointer being in bounds."
	| sp |
	<inline: true>
	sp := self fetchPointer: StackPointerIndex ofObject: aContext.
	(self isIntegerObject: sp) ifFalse: [^0].
	^self integerValueOf: sp! !

!Interpreter methodsFor: 'contexts' stamp: 'dtl 4/22/2007 23:29'!
newActiveContext: aContext
	"Note: internalNewActiveContext: should track changes to this method."

	self storeContextRegisters: activeContext.
	(self oop: aContext isLessThan: youngStart) ifTrue: [ self beRootIfOld: aContext ].
	activeContext := aContext.
	self fetchContextRegisters: aContext.! !

!Interpreter methodsFor: 'message sending' stamp: 'eem 11/12/2012 15:28'!
normalSend
	"Send a message, starting lookup with the receiver's class."
	"Assume: messageSelector and argumentCount have been set, and that 
	the receiver and arguments have been pushed onto the stack,"
	"Note: This method is inlined into the interpreter dispatch loop."
	<sharedCodeNamed: 'normalSend' inCase: #singleExtendedSendBytecode>
	| rcvr |
	rcvr := self internalStackValue: argumentCount.
	lkupClass := self fetchClassOf: rcvr.
	receiverClass := lkupClass.
	self commonSend.! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 12/12/1998 17:08'!
nullCompilerHook
	"This should never be called: either the compiler is uninitialised (in which case the hooks should never be reached) or the compiler initialisation should have replaced all the hook with their external implementations."

	self error: 'uninitialised compiler hook called'.
	^false! !

!Interpreter methodsFor: 'debug support' stamp: ''!
okayActiveProcessStack

	| cntxt |
	cntxt := activeContext.	
	[cntxt = nilObj] whileFalse: [
		self okayFields: cntxt.
		cntxt := (self fetchPointer: SenderIndex ofObject: cntxt).
	].! !

!Interpreter methodsFor: 'debug support' stamp: 'eem 8/1/2013 11:32'!
okayFields: oop
	"Check if the argument is an ok object.
	 If this is a pointers object, check that its fields are all okay oops."

	| i fieldOop |
	(oop = nil or: [oop = 0]) ifTrue: [ ^true ].
	(self isIntegerObject: oop) ifTrue: [ ^true ].
	(self okayOop: oop) ifFalse: [ ^false ].
	(self oopHasOkayClass: oop) ifFalse: [ ^false ].
	((self isPointers: oop) or: [self isCompiledMethod: oop]) ifFalse: [ ^true ].
	(self isCompiledMethod: oop)
		ifTrue:
			[i := (self literalCountOf: oop) + LiteralStart - 1]
		ifFalse:
			[(self isContext: oop)
				ifTrue: [i := CtxtTempFrameStart + (self fetchStackPointerOf: oop) - 1]
				ifFalse: [i := (self lengthOf: oop) - 1]].
	[i >= 0] whileTrue: [
		fieldOop := self fetchPointer: i ofObject: oop.
		(self isIntegerObject: fieldOop) ifFalse: [
			(self okayOop: fieldOop) ifFalse: [ ^false ].
			(self oopHasOkayClass: fieldOop) ifFalse: [ ^false ].
		].
		i := i - 1.
	].
	^true! !

!Interpreter methodsFor: 'debug support' stamp: ''!
okayInterpreterObjects

	| oopOrZero oop |
	self okayFields: nilObj.
	self okayFields: falseObj.
	self okayFields: trueObj.
	self okayFields: specialObjectsOop.
	self okayFields: activeContext.
	self okayFields: method.
	self okayFields: receiver.
	self okayFields: theHomeContext.
	self okayFields: messageSelector.
	self okayFields: newMethod.
	self okayFields: lkupClass.
	0 to: MethodCacheEntries - 1 by: MethodCacheEntrySize do: [ :i |
		oopOrZero := methodCache at: i + MethodCacheSelector.
		oopOrZero = 0 ifFalse: [
			self okayFields: (methodCache at: i + MethodCacheSelector).
			self okayFields: (methodCache at: i + MethodCacheClass).
			self okayFields: (methodCache at: i + MethodCacheMethod).
		].
	].
	1 to: remapBufferCount do: [ :i |
		oop := remapBuffer at: i.
		(self isIntegerObject: oop) ifFalse: [
			self okayFields: oop.
		].
	].
	self okayActiveProcessStack.! !

!Interpreter methodsFor: 'debug support' stamp: 'eem (auto pragmas 12/08) 6/6/2008 06:56'!
oopHasOkayClass: signedOop
	"Attempt to verify that the given oop has a reasonable behavior. The class must be a valid, non-integer oop and must not be nilObj. It must be a pointers object with three or more fields. Finally, the instance specification field of the behavior must match that of the instance."

	| oop oopClass formatMask behaviorFormatBits oopFormatBits |
	<var: #oop type: 'usqInt'>
	<var: #oopClass type: 'usqInt'>

	oop := self cCoerce: signedOop to: 'usqInt'.
	self okayOop: oop.
	oopClass := self cCoerce: (self fetchClassOf: oop) to: 'usqInt'.

	(self isIntegerObject: oopClass)
		ifTrue: [ self error: 'a SmallInteger is not a valid class or behavior'. ^false ].
	(self okayOop: oopClass)
		ifFalse: [ self error: 'class oop is not ok'. ^false ].
	((self isPointers: oopClass) and: [(self lengthOf: oopClass) >= 3])
		ifFalse: [ self error: 'a class (behavior) must be a pointers object of size >= 3'. ^false ].
	(self isBytes: oop)
		ifTrue: [ formatMask := 16rC00 ]  "ignore extra bytes size bits"
		ifFalse: [ formatMask := 16rF00 ].

	behaviorFormatBits := (self formatOfClass: oopClass) bitAnd: formatMask.
	oopFormatBits := (self baseHeader: oop) bitAnd: formatMask.
	behaviorFormatBits = oopFormatBits
		ifFalse: [ self error: 'object and its class (behavior) formats differ'. ^false ].
	^true! !

!Interpreter methodsFor: 'contexts' stamp: 'di 11/30/1998 12:31'!
pop2AndPushIntegerIfOK: integerResult

	successFlag ifTrue:
		[(self isIntegerValue: integerResult)
			ifTrue: [self pop: 2 thenPush: (self integerObjectOf: integerResult)]
			ifFalse: [successFlag := false]]! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
pop: nItems
	"Note: May be called by translated primitive code."

	stackPointer := stackPointer - (nItems*self wordSize).
	^nil! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
pop: nItems thenPush: oop

	| sp |
	self longAt: (sp := stackPointer - ((nItems - 1) * self wordSize)) put: oop.
	stackPointer := sp.
! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
pop: nItems thenPushBool: trueOrFalse
	"A few places pop a few items off the stack and then push a boolean. Make it convenient"
	| sp |
	self longAt: (sp := stackPointer - ((nItems - 1) * self wordSize))
		put:(trueOrFalse ifTrue: [trueObj] ifFalse: [falseObj]).
	stackPointer := sp! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
pop: nItems thenPushInteger: integerVal
"lots of places pop a few items off the stack and then push an integer. MAke it convenient"
	| sp |
	self longAt: (sp := stackPointer - ((nItems - 1) * self wordSize)) put:(self integerObjectOf: integerVal).
	stackPointer := sp.
! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'eem 9/9/2011 09:05'!
popFloat
	<returnTypeC: #double>
	^self floatValueOf: self popStack! !

!Interpreter methodsFor: 'contexts' stamp: 'tpr 3/15/2004 20:00'!
popInteger
"returns 0 if the stackTop was not an integer value, plus sets successFlag false"
	| integerPointer |
	integerPointer := self popStack.
	^self checkedIntegerValueOf: integerPointer! !

!Interpreter methodsFor: 'contexts' stamp: ''!
popPos32BitInteger
	"May set successFlag, and return false if not valid"

	| top |
	top := self popStack.
	^ self positive32BitValueOf: top! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
popStack

	| top |
	top := self longAt: stackPointer.
	stackPointer := stackPointer - self wordSize.
	^ top! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:47'!
popStackBytecode

	self fetchNextBytecode.
	self internalPop: 1.
! !

!Interpreter methodsFor: 'primitive support' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
positive32BitIntegerFor: integerValue

	| newLargeInteger |
	"Note - integerValue is interpreted as POSITIVE, eg, as the result of
		Bitmap>at:, or integer>bitAnd:."
	integerValue >= 0
		ifTrue: [(self isIntegerValue: integerValue)
					ifTrue: [^ self integerObjectOf: integerValue]].

	self wordSize = 4
	ifTrue: ["Faster instantiateSmallClass: currently only works with integral word size."
			newLargeInteger := self instantiateSmallClass: (self splObj: ClassLargePositiveInteger)
					sizeInBytes: self baseHeaderSize + 4]
	ifFalse: ["Cant use instantiateSmallClass: due to integral word requirement."
			newLargeInteger := self instantiateClass: (self splObj: ClassLargePositiveInteger)
					indexableSize: 4].
	self storeByte: 3 ofObject: newLargeInteger withValue: ((integerValue >> 24) bitAnd: 16rFF).
	self storeByte: 2 ofObject: newLargeInteger withValue: ((integerValue >> 16) bitAnd: 16rFF).
	self storeByte: 1 ofObject: newLargeInteger withValue: ((integerValue >> 8) bitAnd: 16rFF).
	self storeByte: 0 ofObject: newLargeInteger withValue: (integerValue bitAnd: 16rFF).
	^ newLargeInteger! !

!Interpreter methodsFor: 'primitive support' stamp: ''!
positive32BitValueOf: oop
	"Convert the given object into an integer value.
	The object may be either a positive ST integer or a four-byte LargePositiveInteger."

	| sz value |
	(self isIntegerObject: oop) ifTrue: [
		value := self integerValueOf: oop.
		value < 0 ifTrue: [^ self primitiveFail].
		^ value].

	self assertClassOf: oop is: (self splObj: ClassLargePositiveInteger).
	successFlag ifTrue: [
		sz := self lengthOf: oop.
		sz = 4 ifFalse: [^ self primitiveFail]].
	successFlag ifTrue: [
		^ (self fetchByte: 0 ofObject: oop) +
		  ((self fetchByte: 1 ofObject: oop) <<  8) +
		  ((self fetchByte: 2 ofObject: oop) << 16) +
		  ((self fetchByte: 3 ofObject: oop) << 24) ].! !

!Interpreter methodsFor: 'primitive support' stamp: 'ar (auto pragmas 12/08) 3/21/2008 20:40'!
positive64BitIntegerFor: integerValue

	| newLargeInteger value highWord sz |
	"Note - integerValue is interpreted as POSITIVE, eg, as the result of
		Bitmap>at:, or integer>bitAnd:."
	<var: 'integerValue' type: 'sqLong'>
 
	(self sizeof: integerValue) = 4 ifTrue: [^self positive32BitIntegerFor: integerValue].


	highWord := self cCode: 'integerValue >> 32'. "shift is coerced to usqInt otherwise"
	highWord = 0 ifTrue:[^self positive32BitIntegerFor: integerValue].
	sz := 5.
	(highWord := highWord >> 8) = 0 ifFalse:[sz := sz + 1].
	(highWord := highWord >> 8) = 0 ifFalse:[sz := sz + 1].
	(highWord := highWord >> 8) = 0 ifFalse:[sz := sz + 1].
	newLargeInteger := self instantiateClass: (self splObj: ClassLargePositiveInteger) indexableSize:  sz.
	0 to: sz-1 do: [:i |
		value := self cCode: '(integerValue >> (i * 8)) & 255'.
		self storeByte: i ofObject: newLargeInteger withValue: value].
	^ newLargeInteger
! !

!Interpreter methodsFor: 'primitive support' stamp: 'eem 12/12/2012 11:44'!
positive64BitValueOf: oop
	"Convert the given object into an integer value.
	The object may be either a positive SmallInteger or an eight-byte LargePositiveInteger."

	| sz szsqLong value  |
	<returnTypeC: #sqLong>
	<var: #value type: #sqLong>
	(self isIntegerObject: oop) ifTrue:
		[value := self integerValueOf: oop.
		value < 0 ifTrue: [^self primitiveFail].
		^ value].

	(self
			isClassOfNonImm: oop
			equalTo: (self splObj: ClassLargePositiveInteger)
			compactClassIndex: ClassLargePositiveIntegerCompactIndex) ifFalse:
		[^self primitiveFail].
	szsqLong := self sizeof: #sqLong.
	sz := self lengthOf: oop.
	sz > szsqLong ifTrue:
		[^self primitiveFail].
	value := 0.
	0 to: sz - 1 do: [:i |
		value := value + ((self cCoerce: (self fetchByte: i ofObject: oop) to: #sqLong) <<  (i*8))].
	^value! !

!Interpreter methodsFor: 'object memory support' stamp: 'eem 10/25/2013 16:21'!
postGCAction: gcModeArg
	"Mark the active and home contexts as roots if old. This 
	allows the interpreter to use storePointerUnchecked to 
	store into them."

	compilerInitialized
		ifTrue: [self compilerPostGC]
		ifFalse: [(self oop: activeContext isLessThan: youngStart)
				ifTrue: [self beRootIfOld: activeContext].
			(self oop: theHomeContext isLessThan: youngStart)
				ifTrue: [self beRootIfOld: theHomeContext]].
	(self sizeOfFree: freeBlock) > shrinkThreshold
		ifTrue: ["Attempt to shrink memory after successfully 
			reclaiming lots of memory"
			self shrinkObjectMemory: (self sizeOfFree: freeBlock) - growHeadroom].
	
	self signalSemaphoreWithIndex: gcSemaphoreIndex.
! !

!Interpreter methodsFor: 'object memory support' stamp: 'eem 8/4/2011 17:18'!
preGCAction: gcMode

	compilerInitialized
		ifTrue: [self compilerPreGC: gcMode = GCModeFull]
		ifFalse: [self storeContextRegisters: activeContext].! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/19/2008 21:37'!
primitiveAddLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| integerRcvr integerArg result oopResult |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>
	<var: 'result' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).
	successFlag ifFalse:[^nil].

	"Compute the preliminary result (which may overflow)"
	result := integerRcvr + integerArg.

	"Now check overflow conditions. First is whether rcvr and arg are of the same sign.
	If they are we need to check for overflow more carefully."
	(integerRcvr bitXor: integerArg) < 0 ifFalse:[
		"Second is whether rcvr and result are of the same sign. If not, we have an overflow."
		(integerRcvr bitXor: result) < 0 ifTrue:[self primitiveFail]].
			
	successFlag ifFalse:[^nil].

	oopResult := self signed64BitIntegerFor: result.
	successFlag ifTrue:[self pop: 2 thenPush: oopResult].
! !

!Interpreter methodsFor: 'object access primitives' stamp: 'eem 7/16/2010 19:57'!
primitiveAdoptInstance
	"Primitive. Change the class of the argument to make it an instance of the receiver
	 given that the format of the receiver matches the format of the argument's class.
	 Fail if receiver or argument are SmallIntegers, or the receiver is an instance of a
	 compact class and the argument isn't, or when the argument's class is compact and
	 the receiver isn't, or when the format of the receiver is different from the format of
	 the argument's class, or when the arguments class is fixed and the receiver's size
	 differs from the size that an instance of the argument's class should have."
	| rcvr arg err |

	arg := self stackObjectValue: 0.
	rcvr := self stackObjectValue: 1.

	err := self changeClassOf: arg to: rcvr.
	err = 0
		ifTrue: ["Flush at cache because rcvr's class has changed."
				self flushAtCache.
				self pop: self methodArgumentCount]
		ifFalse: [self primitiveFail].
	^nil! !

!Interpreter methodsFor: 'object access primitives' stamp: 'tpr 5/13/2005 10:17'!
primitiveAsOop
	| thisReceiver |
	thisReceiver := self stackTop.
	self success: (self isIntegerObject: thisReceiver) not.
	successFlag
		ifTrue: [self pop:1 thenPushInteger: (self hashBitsOf: thisReceiver)]! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/19/2008 21:37'!
primitiveBitAndLargeIntegers
	"Primitive logical operations for large integers in 64 bit range"
	| integerRcvr integerArg oopResult |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self positive64BitValueOf: (self stackValue: 0).
	integerRcvr := self positive64BitValueOf: (self stackValue: 1).
	successFlag ifFalse:[^nil].

	oopResult := self positive64BitIntegerFor: (integerRcvr bitAnd: integerArg).
	successFlag ifTrue:[self pop: 2 thenPush: oopResult].
! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/19/2008 21:37'!
primitiveBitOrLargeIntegers
	"Primitive logical operations for large integers in 64 bit range"
	| integerRcvr integerArg oopResult |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self positive64BitValueOf: (self stackValue: 0).
	integerRcvr := self positive64BitValueOf: (self stackValue: 1).
	successFlag ifFalse:[^nil].

	oopResult := self positive64BitIntegerFor: (integerRcvr bitOr: integerArg).
	successFlag ifTrue:[self pop: 2 thenPush: oopResult].
! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/21/2008 20:41'!
primitiveBitShiftLargeIntegers
	"Primitive logical operations for large integers in 64 bit range"
	| shifted integerArg integerRcvr oopResult |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>
	<var: 'shifted' type: 'sqLong'>

	integerArg := self stackIntegerValue: 0.
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).

	successFlag ifTrue: [
		integerArg >= 0 ifTrue: [
			"Left shift -- must fail if we lose bits beyond 64"
			self success: integerArg < 64.
			shifted := integerRcvr << integerArg.
			self success: (self cCode: 'shifted >> integerArg') = integerRcvr.
		] ifFalse: [
			"Right shift -- OK to lose bits"
			self success: integerArg > -64.
			shifted := self cCode: 'integerRcvr >> (0 - integerArg)'. "right shift coerces to usqInt"
		].
	].
	successFlag ifFalse:[^nil].
	oopResult := self signed64BitIntegerFor: shifted.
	successFlag ifTrue:[self pop: 2 thenPush: oopResult].
! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/19/2008 21:38'!
primitiveBitXorLargeIntegers
	"Primitive logical operations for large integers in 64 bit range"
	| integerRcvr integerArg oopResult |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self positive64BitValueOf: (self stackValue: 0).
	integerRcvr := self positive64BitValueOf: (self stackValue: 1).
	successFlag ifFalse:[^nil].

	oopResult := self positive64BitIntegerFor: (integerRcvr bitXor: integerArg).
	successFlag ifTrue:[self pop: 2 thenPush: oopResult].
! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
primitiveBlockCopy

	| context methodContext contextSize newContext initialIP |
	context := self stackValue: 1.
	(self isIntegerObject: (self fetchPointer: MethodIndex ofObject: context))
		ifTrue: ["context is a block; get the context of its enclosing method"
				methodContext := self fetchPointer: HomeIndex ofObject: context]
		ifFalse: [methodContext := context].
	contextSize := self sizeBitsOf: methodContext.  "in bytes, including header"
	context := nil.  "context is no longer needed and is not preserved across allocation"

	"remap methodContext in case GC happens during allocation"
	self pushRemappableOop: methodContext.
	newContext := self instantiateContext: (self splObj: ClassBlockContext) sizeInBytes: contextSize.
	methodContext := self popRemappableOop.

	initialIP := self integerObjectOf: (instructionPointer+1+3) - (method+self baseHeaderSize).
	"Was instructionPointer + 3, but now it's greater by 1 due to preIncrement"

	"Assume: have just allocated a new context; it must be young.
	 Thus, can use uncheck stores. See the comment in fetchContextRegisters."

	self storePointerUnchecked: InitialIPIndex ofObject: newContext withValue: initialIP.
	self storePointerUnchecked: InstructionPointerIndex ofObject: newContext withValue: initialIP.
	self storeStackPointerValue: 0 inContext: newContext.
	self storePointerUnchecked: BlockArgumentCountIndex ofObject: newContext withValue: (self stackValue: 0).
	self storePointerUnchecked: HomeIndex ofObject: newContext withValue: methodContext.
	self storePointerUnchecked: SenderIndex ofObject: newContext withValue: nilObj.

	self pop: 2 thenPush: newContext.! !

!Interpreter methodsFor: 'process primitives' stamp: 'eem 8/21/2008 11:02'!
primitiveClearVMProfile
	"Primitive. Void the VM profile histograms."
	self cCode: 'ioClearProfile()'.
	self pop: argumentCount! !

!Interpreter methodsFor: 'object access primitives' stamp: 'eem 01/07/2021 10:40'!
primitiveClone
	"Return a shallow copy of the receiver."

	| newCopy |
	newCopy := self cloneObject: self stackTop.
	newCopy = 0 ifTrue: "not enough memory most likely"
		[^self primitiveFail].
	self pop: argumentCount + 1 thenPush: newCopy! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
primitiveClosureCopyWithCopiedValues
	| newClosure copiedValues numCopiedValues numArgs |
	numArgs := self stackIntegerValue: 1.
	copiedValues := self stackTop.
	self success: (self fetchClassOf: copiedValues) = (self splObj: ClassArray).
	successFlag ifFalse:
		[^self primitiveFail].
	numCopiedValues := self numSlotsOf: copiedValues.
	newClosure := self
					closureNumArgs: numArgs
									"greater by 1 due to preIncrement of localIP"
					instructionPointer: instructionPointer + 2 - (method+self baseHeaderSize)
					numCopiedValues: numCopiedValues.
	"Assume: have just allocated a new closure; it must be young. Thus, can use unchecked stores."
	self storePointerUnchecked: ClosureOuterContextIndex ofObject: newClosure withValue: (self stackValue: 2).
	numCopiedValues > 0 ifTrue:
		["Allocation may have done a GC and copiedValues may have moved."
		 copiedValues := self stackTop.
		 0 to: numCopiedValues - 1 do:
			[:i|
			"Assume: have just allocated a new BlockClosure; it must be young.
			 Thus, can use unchecked stores."
			 self storePointerUnchecked: i + ClosureFirstCopiedValueIndex
				ofObject: newClosure
				withValue: (self fetchPointer: i ofObject: copiedValues)]].
	self pop: 3 thenPush: newClosure! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem 9/23/2010 21:00'!
primitiveClosureValue
	| blockClosure blockArgumentCount closureMethod outerContext |
	blockClosure := self stackValue: argumentCount.
	blockArgumentCount := self argumentCountOfClosure: blockClosure.
	argumentCount = blockArgumentCount ifFalse:
		[^self primitiveFail].

	"Somewhat paranoiac checks we need while debugging that we may be able to discard
	 in a robust system."
	outerContext := self fetchPointer: ClosureOuterContextIndex ofObject: blockClosure.
	(self isContext: outerContext) ifFalse:
		[^self primitiveFail].
	closureMethod := self fetchPointer: MethodIndex ofObject: outerContext.
	"Check if the closure's method is actually a CompiledMethod."
	(self isOopCompiledMethod: closureMethod) ifFalse:
		[^self primitiveFail].

	self activateNewClosureMethod: blockClosure.
	self quickCheckForInterrupts! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem 9/23/2010 21:00'!
primitiveClosureValueNoContextSwitch
	"An exact clone of primitiveClosureValue except that this version will not
	 check for interrupts on stack overflow."
	| blockClosure blockArgumentCount closureMethod outerContext |
	blockClosure := self stackValue: argumentCount.
	blockArgumentCount := self argumentCountOfClosure: blockClosure.
	argumentCount = blockArgumentCount ifFalse:
		[^self primitiveFail].

	"Somewhat paranoiac checks we need while debugging that we may be able to discard
	 in a robust system."
	outerContext := self fetchPointer: ClosureOuterContextIndex ofObject: blockClosure.
	(self isContext: outerContext) ifFalse:
		[^self primitiveFail].
	closureMethod := self fetchPointer: MethodIndex ofObject: outerContext.
	"Check if the closure's method is actually a CompiledMethod."
	(self isOopCompiledMethod: closureMethod) ifFalse:
		[^self primitiveFail].

	self activateNewClosureMethod: blockClosure! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem 07/28/2014 07:54'!
primitiveClosureValueWithArgs
	| argumentArray arraySize cntxSize blockClosure blockArgumentCount closureMethod index outerContext |
	argumentArray := self stackTop.
	(self isArray: argumentArray) ifFalse:
		[^self primitiveFail].

	"Check for enough space in thisContext to push all args"
	arraySize := self numSlotsOf: argumentArray.
	cntxSize := self numSlotsOf: activeContext.
	(self stackPointerIndex + arraySize) < cntxSize ifFalse:
		[^self primitiveFail].

	blockClosure := self stackValue: argumentCount.
	blockArgumentCount := self argumentCountOfClosure: blockClosure.
	arraySize = blockArgumentCount ifFalse:
		[^self primitiveFail].

	"Somewhat paranoiac checks we need while debugging that we may be able to discard
	 in a robust system."
	outerContext := self fetchPointer: ClosureOuterContextIndex ofObject: blockClosure.
	(self isContext: outerContext) ifFalse:
		[^self primitiveFail].
	closureMethod := self fetchPointer: MethodIndex ofObject: outerContext.
	"Check if the closure's method is actually a CompiledMethod."
	(self isOopCompiledMethod: closureMethod) ifFalse:
		[^self primitiveFail].

	self popStack.

	"Copy the arguments to the stack, and activate"
	index := 1.
	[index <= arraySize]
		whileTrue:
		[self push: (self fetchPointer: index - 1 ofObject: argumentArray).
		index := index + 1].

	argumentCount := arraySize.
	self activateNewClosureMethod: blockClosure.
	self quickCheckForInterrupts! !

!Interpreter methodsFor: 'array primitives' stamp: 'ar (auto pragmas 12/08) 6/3/2007 17:23'!
primitiveCompareBytes
	"Primitive. Compare two byte-indexed objects for equality"
	| arg1 arg2 len1 len2 |
	<export: true>
	argumentCount = 1 ifFalse:[^self primitiveFail].
	arg1 := self stackValue: 1.
	arg2 := self stackValue: 0.
	((self isBytes: arg1) and:[self isBytes: arg2]) 
		ifFalse:[^self primitiveFail].
	"Quick identity test"
	(arg1 == arg2) ifTrue:[^self pop: 2 thenPush: trueObj].
	len1 := self byteSizeOf: arg1.
	len2 := self byteSizeOf: arg2.
	len1 = len2 ifFalse:[^self pop: 2 thenPush: falseObj].
	0 to: len1-1 do:[:i|
		(self fetchByte: i ofObject: arg1) = (self fetchByte: i ofObject: arg2) 
			ifFalse:[^self pop: 2 thenPush: falseObj]].
	self pop: 2 thenPush: trueObj.
! !

!Interpreter methodsFor: 'process primitives' stamp: 'eem 11/2/2008 10:38'!
primitiveControlVMProfiling
	"Primitive. Start or stop the VM profiler.  The first argument is a boolean
	 to switch profiling on or off.  The second argument is an integer or nil.
	 If an integer it determines the maximum number of samples in the VM's
	 sample buffer. Answer the current number of samples in the buffer."
	| onOffBar bufferSize numSamples |
	argumentCount ~= 2 ifTrue:
		[^self primitiveFail].
	(onOffBar := self stackValue: 1) = trueObj
		ifTrue: [onOffBar := 1]
		ifFalse:
			[onOffBar = falseObj
				ifTrue: [onOffBar := 0]
				ifFalse: [^self primitiveFail]].
	(bufferSize := self stackTop) = nilObj
		ifTrue: [bufferSize := 0]
		ifFalse:
			[((self isIntegerObject: bufferSize)
			  and: [(bufferSize := self integerValueOf: bufferSize) > 0]) ifFalse:
				[^self primitiveFail]].
	numSamples := self cCode: 'ioControlNewProfile(onOffBar,bufferSize)'.
	self pop: 3 thenPushInteger: numSamples! !

!Interpreter methodsFor: 'object access primitives' stamp: 'tpr 3/23/2004 17:40'!
primitiveCopyObject
	"Primitive. Copy the state of the receiver from the argument. 
		Fail if receiver and argument are of a different class. 
		Fail if the receiver or argument are non-pointer objects.
		Fail if receiver and argument have different lengths (for indexable objects).
	"
	| rcvr arg length |
	self methodArgumentCount = 1 ifFalse:[^self primitiveFail].
	arg := self stackObjectValue: 0.
	rcvr := self stackObjectValue: 1.

	self failed ifTrue:[^nil].
	(self isPointers: rcvr) ifFalse:[^self primitiveFail].
	(self fetchClassOf: rcvr) = (self fetchClassOf: arg) ifFalse:[^self primitiveFail].
	length := self lengthOf: rcvr.
	length = (self lengthOf: arg) ifFalse:[^self primitiveFail].
	
	"Now copy the elements"
	0 to: length-1 do:[:i|
		self storePointer: i ofObject: rcvr withValue: (self fetchPointer: i ofObject: arg)].

	"Note: The above could be faster for young receivers but I don't think it'll matter"
	self pop: 1. "pop arg; answer receiver"
! !

!Interpreter methodsFor: 'I/O primitives' stamp: 'tpr 3/15/2004 12:16'!
primitiveDeferDisplayUpdates
	"Set or clear the flag that controls whether modifications of 
	the Display object are propagated to the underlying 
	platform's screen."
	| flag |
	flag := self stackTop.
	flag = trueObj
		ifTrue: [deferDisplayUpdates := true]
		ifFalse: [flag = falseObj
				ifTrue: [deferDisplayUpdates := false]
				ifFalse: [self primitiveFail]].
	successFlag
		ifTrue: [self pop: 1]! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/20/2008 22:52'!
primitiveDivLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| integerRcvr integerArg result posArg posRcvr oopResult |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>
	<var: 'posRcvr' type: 'sqLong'>
	<var: 'posArg' type: 'sqLong'>
	<var: 'result' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).
	integerArg = 0 ifTrue:[self primitiveFail].
	successFlag ifFalse:[^nil].

	integerRcvr > 0
		ifTrue: [integerArg > 0
					ifTrue: [result := integerRcvr // integerArg]
					ifFalse: ["round negative result toward negative infinity"
							posArg := 0 - integerArg.
							posRcvr := integerRcvr + (posArg - 1). "can overflow!!"
							posRcvr < 0 ifTrue:[self primitiveFail].
							result := 0 - (posRcvr // posArg)]]
		ifFalse: [posRcvr := 0 - integerRcvr.
				integerArg > 0
					ifTrue: ["round negative result toward negative infinity"
							posRcvr := posRcvr + (integerArg - 1). "can overflow!!"
							posRcvr < 0 ifTrue:[self primitiveFail].
							result := 0 - (posRcvr // integerArg)]
					ifFalse: [posArg := 0 - integerArg.
							result := posRcvr // posArg]].
	successFlag ifTrue:[oopResult := self signed64BitIntegerFor: result].
	successFlag ifTrue:[self pop: 2 thenPush: oopResult].
! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/19/2008 21:38'!
primitiveDivideLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| integerRcvr integerArg result oopResult |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>
	<var: 'result' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).
	(integerArg ~= 0 and:[integerRcvr \\ integerArg = 0]) ifFalse:[self primitiveFail].
	successFlag ifFalse:[^nil].
	result := integerRcvr // integerArg.
	oopResult := self signed64BitIntegerFor: result.
	successFlag ifTrue:[self pop: 2 thenPush: oopResult].
! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem 07/28/2014 07:54'!
primitiveDoPrimitiveWithArgs
	| argumentArray arraySize index cntxSize primIdx |
	argumentArray := self stackTop.
	arraySize := self numSlotsOf: argumentArray.
	cntxSize := self numSlotsOf: activeContext.
	self success: self stackPointerIndex + arraySize < cntxSize.
	(self isArray: argumentArray) ifFalse: [^ self primitiveFail].

	primIdx := self stackIntegerValue: 1.
	successFlag ifFalse: [^ self primitiveFail]. "invalid args"

	"Pop primIndex and argArray, then push args in place..."
	self pop: 2.
	primitiveIndex := primIdx.
	argumentCount := arraySize.
	index := 1.
	[index <= argumentCount]
		whileTrue: [self push: (self fetchPointer: index - 1 ofObject: argumentArray).
			index := index + 1].

	"Run the primitive (sets successFlag)"
	self pushRemappableOop: argumentArray. "prim might alloc/gc"
	lkupClass := nilObj.
	self primitiveResponse.
	argumentArray := self popRemappableOop.
	successFlag
		ifFalse: ["If primitive failed, then restore state for failure code"
			self pop: arraySize.
			self pushInteger: primIdx.
			self push: argumentArray.
			argumentCount := 2]! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/19/2008 21:38'!
primitiveEqualLargeIntegers
	"Primitive comparison operations for large integers in 64 bit range"
	| integerRcvr integerArg |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).

	successFlag ifTrue:[
		self pop: 2.
		self pushBool: integerRcvr = integerArg
	].
! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem 10/26/2008 18:49'!
primitiveExecuteMethod
	"receiver, args, then method are on top of stack. Execute method against receiver and args"
	newMethod := self popStack.
	primitiveIndex := self primitiveIndexOf: newMethod.
	self success: argumentCount - 1 = (self argumentCountOf: newMethod).
	successFlag
		ifTrue: [argumentCount := argumentCount - 1.
			self executeNewMethod.
			"Recursive xeq affects successFlag"
			successFlag := true]
		ifFalse: [self unPop: 1]! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem 07/28/2014 07:54'!
primitiveExecuteMethodArgsArray
	"receiver, argsArray, then method are on top of stack.  Execute method against
	 receiver and args.  Allow for up to two extra arguments (e.g. for mirror primitives).
	 Set primitiveFunctionPointer because no cache lookup has been done for the
	 method, and hence primitiveFunctionPointer is stale."
	| methodArgument argCnt argumentArray |
	methodArgument := self stackTop.
	argumentArray := self stackValue: 1.
	((self isOopCompiledMethod: methodArgument)
	 and: [self isArray: argumentArray]) ifFalse:
		[^self primitiveFail].
	argCnt := self argumentCountOf: methodArgument.
	argCnt = (self numSlotsOf: argumentArray) ifFalse:
		[^self primitiveFail].
	argumentCount > 2 ifTrue: "CompiledMethod class>>receiver:withArguments:executeMethod:
								SqueakObjectPrimitives class >> receiver:withArguments:apply:
								VMMirror>>ifFail:object:with:executeMethod: et al"
		[argumentCount > 4 ifTrue:
			[^self primitiveFail].
		self stackValue: argumentCount put: (self stackValue: 2)]. "replace actual receiver with desired receiver"
	"and push the actual arguments"
	self pop: argumentCount.
	0 to: argCnt - 1 do:
		[:i|
		self push: (self fetchPointer: i ofObject: argumentArray)].
	newMethod := methodArgument.
	primitiveIndex := self primitiveIndexOf: newMethod.
	primitiveFunctionPointer := self functionPointerFor: primitiveIndex inClass: nil.
	argumentCount := argCnt.
	"We set the messageSelector for executeMethod below since things
	 like the at cache read messageSelector and so it cannot be left stale."
	messageSelector := self nilObject.
	self executeNewMethod.
	"Recursive xeq affects successFlag"
	successFlag := true! !

!Interpreter methodsFor: 'plugin primitives' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
primitiveExternalCall
	"Call an external primitive. The external primitive methods 
	contain as first literal an array consisting of: 
	* The module name (String | Symbol) 
	* The function name (String | Symbol) 
	* The session ID (SmallInteger) [OBSOLETE] 
	* The function index (Integer) in the externalPrimitiveTable 
	For fast failures the primitive index of any method where the 
	external prim is not found is rewritten in the method cache 
	with zero. This allows for ultra fast responses as long as the 
	method stays in the cache. 
	The fast failure response relies on lkupClass being properly 
	set. This is done in 
	#addToMethodCacheSel:class:method:primIndex: to 
	compensate for execution of methods that are looked up in a 
	superclass (such as in primitivePerformAt). 
	With the latest modifications (e.g., actually flushing the 
	function addresses from the VM), the session ID is obsolete. 
	But for backward compatibility it is still kept around. Also, a 
	failed lookup is reported specially. If a method has been 
	looked up and not been found, the function address is stored 
	as -1 (e.g., the SmallInteger -1 to distinguish from 
	16rFFFFFFFF which may be returned from the lookup). 
	It is absolutely okay to remove the rewrite if we run into any 
	problems later on. It has an approximate speed difference of 
	30% per failed primitive call which may be noticable but if, 
	for any reasons, we run into problems (like with J3) we can 
	always remove the rewrite. 
	"
	| lit addr moduleName functionName moduleLength functionLength index |
	<var: #addr type: 'void *'>
	
	"Fetch the first literal of the method"
	self success: (self literalCountOf: newMethod) > 0. "@@: Could this be omitted for speed?!!"
	successFlag ifFalse: [^ nil].

	lit := self literal: 0 ofMethod: newMethod. 
	"Check if it's an array of length 4"
	self success: ((self isArray: lit) and: [(self lengthOf: lit) = 4]).
	successFlag ifFalse: [^ nil].

	"Look at the function index in case it has been loaded before"
	index := self fetchPointer: 3 ofObject: lit.
	index := self checkedIntegerValueOf: index.
	successFlag ifFalse: [^ nil].
	"Check if we have already looked up the function and failed."
	index < 0
		ifTrue: ["Function address was not found in this session, 
			Rewrite the mcache entry with a zero primitive index."
			self
				rewriteMethodCacheSel: messageSelector
				class: lkupClass
				primIndex: 0.
			^ self success: false].

	"Try to call the function directly"
	(index > 0 and: [index <= MaxExternalPrimitiveTableSize])
		ifTrue: [addr := externalPrimitiveTable at: index - 1.
			addr ~= 0
				ifTrue: [self rewriteMethodCacheSel: messageSelector class: lkupClass primIndex: (1000 + index) primFunction: addr.
					self callExternalPrimitive: addr.
					^ nil].
			"if we get here, then an index to the external prim was 
			kept on the ST side although the underlying prim 
			table was already flushed"
			^ self primitiveFail].

	"Clean up session id and external primitive index"
	self storePointerUnchecked: 2 ofObject: lit withValue: ConstZero.
	self storePointerUnchecked: 3 ofObject: lit withValue: ConstZero.

	"The function has not been loaded yet. Fetch module and function name."
	moduleName := self fetchPointer: 0 ofObject: lit.
	moduleName = nilObj
		ifTrue: [moduleLength := 0]
		ifFalse: [self success: (self isBytes: moduleName).
				moduleLength := self lengthOf: moduleName.
				self cCode: '' inSmalltalk:
					[ (#('FloatArrayPlugin' 'Matrix2x3Plugin') includes: (self stringOf: moduleName))
						ifTrue: [moduleLength := 0  "Cause all of these to fail"]]].
	functionName := self fetchPointer: 1 ofObject: lit.
	self success: (self isBytes: functionName).
	functionLength := self lengthOf: functionName.
	successFlag ifFalse: [^ nil].

	addr := self ioLoadExternalFunction: functionName + self baseHeaderSize
				OfLength: functionLength
				FromModule: moduleName + self baseHeaderSize
				OfLength: moduleLength.
	addr = 0
		ifTrue: [index := -1]
		ifFalse: ["add the function to the external primitive table"
			index := self addToExternalPrimitiveTable: addr].
	self success: index >= 0.
	"Store the index (or -1 if failure) back in the literal"
	self storePointerUnchecked: 3 ofObject: lit withValue: (self integerObjectOf: index).

	"If the function has been successfully loaded process it"
	(successFlag and: [addr ~= 0])
		ifTrue: [self rewriteMethodCacheSel: messageSelector class: lkupClass primIndex: (1000 + index) primFunction: addr.
				self callExternalPrimitive: addr]
		ifFalse: ["Otherwise rewrite the primitive index"
			self
				rewriteMethodCacheSel: messageSelector
				class: lkupClass
				primIndex: 0]! !

!Interpreter methodsFor: 'arithmetic float primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:29'!
primitiveFloatAdd: rcvrOop toArg: argOop
	| rcvr arg |
	<var: #rcvr type: 'double '>
	<var: #arg type: 'double '>

	rcvr := self loadFloatOrIntFrom: rcvrOop.
	arg := self loadFloatOrIntFrom: argOop.
	successFlag ifTrue: [
		self pop: 2.
		self pushFloat: rcvr + arg].! !

!Interpreter methodsFor: 'arithmetic float primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:29'!
primitiveFloatDivide: rcvrOop byArg: argOop
	| rcvr arg |
	<var: #rcvr type: 'double '>
	<var: #arg type: 'double '>

	rcvr := self loadFloatOrIntFrom: rcvrOop.
	arg := self loadFloatOrIntFrom: argOop.
	successFlag ifTrue: [
		self success: arg ~= 0.0.
		successFlag ifTrue: [
			self pop: 2.
			self pushFloat: (self cCode: 'rcvr / arg' inSmalltalk: [rcvr / arg])]].! !

!Interpreter methodsFor: 'arithmetic float primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:29'!
primitiveFloatEqual: rcvrOop toArg: argOop
	| rcvr arg |
	<var: #rcvr type: 'double '>
	<var: #arg type: 'double '>

	rcvr := self loadFloatOrIntFrom: rcvrOop.
	arg := self loadFloatOrIntFrom: argOop.
	successFlag ifTrue: [^ rcvr = arg]! !

!Interpreter methodsFor: 'arithmetic float primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:29'!
primitiveFloatGreater: rcvrOop thanArg: argOop
	| rcvr arg |
	<var: #rcvr type: 'double '>
	<var: #arg type: 'double '>

	rcvr := self loadFloatOrIntFrom: rcvrOop.
	arg := self loadFloatOrIntFrom: argOop.
	successFlag ifTrue: [^ rcvr > arg].
! !

!Interpreter methodsFor: 'float primitives' stamp: 'eem 1/6/2009 10:49'!
primitiveFloatGreaterOrEqual: rcvrOop toArg: argOop
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := self loadFloatOrIntFrom: rcvrOop.
	arg := self loadFloatOrIntFrom: argOop.
	successFlag ifTrue: [^ rcvr >= arg].! !

!Interpreter methodsFor: 'arithmetic float primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:29'!
primitiveFloatLess: rcvrOop thanArg: argOop
	| rcvr arg |
	<var: #rcvr type: 'double '>
	<var: #arg type: 'double '>

	rcvr := self loadFloatOrIntFrom: rcvrOop.
	arg := self loadFloatOrIntFrom: argOop.
	successFlag ifTrue: [^ rcvr < arg].
! !

!Interpreter methodsFor: 'float primitives' stamp: 'nice (eem) 1/4/2009 21:23'!
primitiveFloatLessOrEqual: rcvrOop toArg: argOop
	| rcvr arg |
	<var: #rcvr type: #double>
	<var: #arg type: #double>

	rcvr := self loadFloatOrIntFrom: rcvrOop.
	arg := self loadFloatOrIntFrom: argOop.
	successFlag ifTrue: [^ rcvr <= arg].
! !

!Interpreter methodsFor: 'arithmetic float primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:29'!
primitiveFloatMultiply: rcvrOop byArg: argOop
	| rcvr arg |
	<var: #rcvr type: 'double '>
	<var: #arg type: 'double '>

	rcvr := self loadFloatOrIntFrom: rcvrOop.
	arg := self loadFloatOrIntFrom: argOop.
	successFlag ifTrue: [
		self pop: 2.
		self pushFloat: rcvr * arg].! !

!Interpreter methodsFor: 'arithmetic float primitives' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:30'!
primitiveFloatSubtract: rcvrOop fromArg: argOop
	| rcvr arg |
	<var: #rcvr type: 'double '>
	<var: #arg type: 'double '>

	rcvr := self loadFloatOrIntFrom: rcvrOop.
	arg := self loadFloatOrIntFrom: argOop.
	successFlag ifTrue: [
		self pop: 2.
		self pushFloat: rcvr - arg].! !

!Interpreter methodsFor: 'system control primitives' stamp: 'ikp 1/4/1999 11:31'!
primitiveFlushCacheByMethod
	"The receiver is a compiledMethod.  Clear all entries in the method lookup cache that refer to this method, presumably because it has been redefined, overridden or removed."
	| probe oldMethod |
	oldMethod := self stackTop.
	probe := 0.
	1 to: MethodCacheEntries do:
		[:i | (methodCache at: probe + MethodCacheMethod) = oldMethod ifTrue:
			[methodCache at: probe + MethodCacheSelector put: 0].
		probe := probe + MethodCacheEntrySize].
	self compilerFlushCacheHook: oldMethod.		"Flush the dynamic compiler's inline caches."! !

!Interpreter methodsFor: 'I/O primitives' stamp: 'ar (auto pragmas 12/08) 10/12/2007 16:20'!
primitiveGetLogDirectory
	"Primitive. Answer the VM's current log directory"
	| ptr sz stringOop |
	<var: 'ptr' type: 'char*'>
	<export: true>
	ptr := self ioGetLogDirectory.
	ptr == nil ifTrue:[^self success: false].
	sz := self strlen: ptr.
	stringOop := self instantiateClass: self classString indexableSize: sz.
	0 to: sz-1 do:[:i| self storeByte: i ofObject: stringOop withValue: (ptr at: i)].
	self pop: self methodArgumentCount+1.
	self push: stringOop.! !

!Interpreter methodsFor: 'I/O primitives' stamp: 'ar (auto pragmas 12/08) 3/8/2007 14:09'!
primitiveGetWindowLabel
	"Primitive. Answer the OS window's label"
	| ptr sz labelOop |
	<var: 'ptr' type: 'char*'>
	<export: true>
	ptr := self ioGetWindowLabel.
	ptr == nil ifTrue:[^self success: false].
	sz := self strlen: ptr.
	labelOop := self instantiateClass: self classString indexableSize: sz.
	0 to: sz-1 do:[:i| self storeByte: i ofObject: labelOop withValue: (ptr at: i)].
	self pop: self methodArgumentCount+1.
	self push: labelOop.! !

!Interpreter methodsFor: 'I/O primitives' stamp: 'ar (auto pragmas 12/08) 2/27/2008 10:54'!
primitiveGetWindowSize
	"Primitive. Answer the OS window's size (extent)"
	| w h |
	<export: true>
	w := self ioGetWindowWidth.
	h := self ioGetWindowHeight.
	self pop: self methodArgumentCount+1.
	self push: (self makePointwithxValue: w yValue: h).! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/19/2008 21:38'!
primitiveGreaterOrEqualLargeIntegers
	"Primitive comparison operations for large integers in 64 bit range"
	| integerRcvr integerArg |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).

	successFlag ifTrue:[
		self pop: 2.
		self pushBool: integerRcvr >= integerArg
	].
! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/19/2008 21:38'!
primitiveGreaterThanLargeIntegers
	"Primitive comparison operations for large integers in 64 bit range"
	| integerRcvr integerArg |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).

	successFlag ifTrue:[
		self pop: 2.
		self pushBool: integerRcvr > integerArg
	].
! !

!Interpreter methodsFor: 'system control primitives' stamp: 'ar (auto pragmas 12/08) 6/22/2007 13:01'!
primitiveHighResClock
	"Return the value of the high resolution clock if this system has any. The exact frequency of the high res clock is undefined specifically so that we can use processor dependent instructions (like RDTSC). The only use for the high res clock is for profiling where we can allocate time based on sub-msec resolution of the high res clock. If no high-resolution counter is available, the platform should return zero."
	<export: true>
	self pop: 1.
	self push: (self positive64BitIntegerFor: self ioHighResClock).! !

!Interpreter methodsFor: 'compiled methods' stamp: 'eem 08/08/2014 10:47'!
primitiveIndexOf: methodPointer
	"Note: We now have 10 bits of primitive index, but they are in two places
	for temporary backward compatibility.  The time to unpack is negligible,
	since the reconstituted full index is stored in the method cache."
	| primBits |
	primBits := ((self methodHeaderOf: methodPointer) >> 1) bitAnd: 16r100001FF.
	
	^(primBits bitAnd: 16r1FF) + (primBits >> 19)! !

!Interpreter methodsFor: 'process primitives' stamp: 'ar (auto pragmas 12/08) 6/11/2007 23:10'!
primitiveInterruptChecksPerMSec
	"Primitive. Answer the number of interrupt checks per milliseconds that we execute
	on this machine. This can be used to adjust the sub-msecs profiler to check (roughly) 
	n times per millisecond."
	<export: true>
	self methodArgumentCount = 0 ifFalse:[^self success: false].
	self pop: 1 thenPush: (self integerObjectOf: interruptCheckCounterFeedBackReset).! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
primitiveInvokeObjectAsMethod
	"Primitive. 'Invoke' an object like a function, sending the special message 
		run: originalSelector with: arguments in: aReceiver.
	"
	| runSelector runReceiver runArgs newReceiver lookupClass |
	runArgs := self instantiateClass: (self splObj: ClassArray) indexableSize: argumentCount.
	self beRootIfOld: runArgs. "do we really need this?"
	self transfer: argumentCount from: stackPointer - ((argumentCount - 1) * self wordSize) to: runArgs + self baseHeaderSize.

	runSelector := messageSelector.
	runReceiver := self stackValue: argumentCount.
	self pop: argumentCount+1.

	"stack is clean here"

	newReceiver := newMethod.
	messageSelector := self splObj: SelectorRunWithIn.
	argumentCount := 3.

	self push: newReceiver.
	self push: runSelector.
	self push: runArgs.
	self push: runReceiver.

	lookupClass := self fetchClassOf: newReceiver.
	self findNewMethodInClass: lookupClass.
	self executeNewMethod.  "Recursive xeq affects successFlag"
	successFlag := true.
! !

!Interpreter methodsFor: 'I/O primitives' stamp: 'ar 1/6/2009 16:39'!
primitiveIsWindowObscured
	"Primitive. Answer whether the OS window is currently partially or fully obscured."
	<export: true>
	self pop: self methodArgumentCount+1.
	self pushBool: self ioIsWindowObscured.
! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/19/2008 21:38'!
primitiveLessOrEqualLargeIntegers
	"Primitive comparison operations for large integers in 64 bit range"
	| integerRcvr integerArg |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).

	successFlag ifTrue:[
		self pop: 2.
		self pushBool: integerRcvr <= integerArg
	].
! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/19/2008 21:38'!
primitiveLessThanLargeIntegers
	"Primitive comparison operations for large integers in 64 bit range"
	| integerRcvr integerArg |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).

	successFlag ifTrue:[
		self pop: 2.
		self pushBool: integerRcvr < integerArg
	].
! !

!Interpreter methodsFor: 'quick primitives' stamp: 'jm 9/18/97 21:06'!
primitiveLoadInstVar
	| thisReceiver |
	thisReceiver := self popStack.
	self push: (self fetchPointer: primitiveIndex-264 ofObject: thisReceiver)! !

!Interpreter methodsFor: 'system control primitives' stamp: 'dtl 8/13/2012 22:20'!
primitiveLocalMicrosecondClock
	"Answer the local microseconds since the Smalltalk epoch. The value is
	derived from the Posix epoch with a constant offset corresponding to
	elapsed microseconds between the two epochs according to RFC 868,
	and with an offset duration corresponding to the current offset of local
	time from UTC."
	
	| clock offset offsetMillis epochDelta uSecs |

	<export: true>
	<var: #clock type: 'usqLong'>
	<var: #offset type: 'int'>
	<var: #offsetMillis type: 'usqLong'>
	<var: #epochDelta declareC: 'static usqLong epochDelta= 2177452800000000ULL'>
	(self cCode: 'ioUtcWithOffset(&clock, &offset)' inSmalltalk: [-1]) = -1
		ifTrue: [^ self primitiveFail].
	clock := clock + epochDelta. "adjust for nominal Smalltalk epoch"
	offsetMillis := offset.
	offsetMillis := offsetMillis * 1000000.
	clock := clock + offsetMillis. "adjust for local time offset"
	uSecs := self positive64BitIntegerFor: clock.
	self pop: 1 thenPush: uSecs.
! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/19/2008 21:38'!
primitiveModLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| integerRcvr integerArg result oopResult |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>
	<var: 'result' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).
	integerArg = 0 ifTrue:[self primitiveFail].
	successFlag ifFalse:[^nil].

	result := integerRcvr \\ integerArg.

	"ensure that the result has the same sign as the integerArg"
	integerArg < 0
		ifTrue: [result > 0
			ifTrue: [result := result + integerArg]]
		ifFalse: [result < 0
			ifTrue: [result := result + integerArg]].

	oopResult := self signed64BitIntegerFor: result.
	successFlag ifTrue:[self pop: 2 thenPush: oopResult].
! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/19/2008 21:38'!
primitiveMultiplyLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| integerRcvr integerArg result oopResult |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>
	<var: 'result' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).
	successFlag ifFalse:[^nil].

	result := integerRcvr * integerArg.
	"check for C overflow by seeing if computation is reversible"
	((integerArg = 0) or: [(result // integerArg) = integerRcvr])
		ifTrue:[oopResult := self signed64BitIntegerFor: result]
		ifFalse: [self primitiveFail].

	successFlag ifTrue:[self pop: 2 thenPush: oopResult].
! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/19/2008 21:39'!
primitiveNotEqualLargeIntegers
	"Primitive comparison operations for large integers in 64 bit range"
	| integerRcvr integerArg |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).

	successFlag ifTrue:[
		self pop: 2.
		self pushBool: integerRcvr ~= integerArg
	].
! !

!Interpreter methodsFor: 'object access primitives' stamp: 'eem 10/12/2011 11:08'!
primitiveNotIdentical
	"is the receiver/first argument not the same object as the (last) argument?.
	 pop argumentCount because this can be used as a mirror primitive."
	| thisObject otherObject |
	otherObject := self stackValue: 1.
	thisObject := self stackTop.
	self pop: argumentCount + 1 thenPushBool: thisObject ~= otherObject! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem 9/25/2010 09:40'!
primitivePerform
	| performSelector newReceiver selectorIndex lookupClass performMethod |
	performSelector := messageSelector.
	performMethod := newMethod.
	messageSelector := self stackValue: argumentCount - 1.
	newReceiver := self stackValue: argumentCount.

	"NOTE: the following lookup may fail and be converted to #doesNotUnderstand:, so we must adjust argumentCount and slide args now, so that would work."

	"Slide arguments down over selector"
	argumentCount := argumentCount - 1.
	selectorIndex := self stackPointerIndex - argumentCount.
	self
		transfer: argumentCount
		fromIndex: selectorIndex + 1
		ofObject: activeContext
		toIndex: selectorIndex
		ofObject: activeContext.
	self pop: 1.
	lookupClass := self fetchClassOf: newReceiver.
	self findNewMethodInClass: lookupClass.

	"Only test CompiledMethods for argument count - other objects will have to take their chances"
	(self isOopCompiledMethod: newMethod)
		ifTrue: [self success: (self argumentCountOf: newMethod) = argumentCount].

	successFlag
		ifTrue: [self executeNewMethod.
			"Recursive xeq affects successFlag"
			successFlag := true]
		ifFalse: ["Slide the args back up (sigh) and re-insert the 
			selector. "
			1 to: argumentCount do: [:i | self
						storePointer: argumentCount - i + 1 + selectorIndex
						ofObject: activeContext
						withValue: (self fetchPointer: argumentCount - i + selectorIndex ofObject: activeContext)].
			self unPop: 1.
			self storePointer: selectorIndex
				ofObject: activeContext
				withValue: messageSelector.
			argumentCount := argumentCount + 1.
			newMethod := performMethod.
			messageSelector := performSelector]! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem 07/28/2014 07:54'!
primitivePerformAt: lookupClass
	"Common routine used by perform:withArgs: and perform:withArgs:inSuperclass:"

	"NOTE:  The case of doesNotUnderstand: is not a failure to perform.
	The only failures are arg types and consistency of argumentCount."

	| performSelector argumentArray arraySize index cntxSize performMethod performArgCount |
	argumentArray := self stackTop.
	(self isArray: argumentArray) ifFalse:[^self primitiveFail].

	successFlag ifTrue:
		["Check for enough space in thisContext to push all args"
		arraySize := self numSlotsOf: argumentArray.
		cntxSize := self numSlotsOf: activeContext.
		self success: (self stackPointerIndex + arraySize) < cntxSize].
	successFlag ifFalse: [^nil].

	performSelector := messageSelector.
	performMethod := newMethod.
	performArgCount := argumentCount.
	"pop the arg array and the selector, then push the args out of the array, as if they were on the stack"
	self popStack.
	messageSelector := self popStack.

	"Copy the arguments to the stack, and execute"
	index := 1.
	[index <= arraySize]
		whileTrue:
		[self push: (self fetchPointer: index - 1 ofObject: argumentArray).
		index := index + 1].
	argumentCount := arraySize.

	self findNewMethodInClass: lookupClass.

	"Only test CompiledMethods for argument count - any other objects playacting as CMs will have to take their chances"
	(self isOopCompiledMethod: newMethod)
		ifTrue: [self success: (self argumentCountOf: newMethod) = argumentCount].

	successFlag
		ifTrue: [self executeNewMethod.  "Recursive xeq affects successFlag"
				successFlag := true]
		ifFalse: ["Restore the state by popping all those array entries and pushing back the selector and array, and fail"
				self pop: argumentCount.
				self push: messageSelector.
				self push: argumentArray.
				messageSelector := performSelector.
				newMethod := performMethod.
				argumentCount := performArgCount]
! !

!Interpreter methodsFor: 'control primitives' stamp: 'di 4/12/1999 10:57'!
primitivePerformInSuperclass
	| lookupClass rcvr currentClass |
	lookupClass := self stackTop.
	rcvr := self stackValue: argumentCount.
	currentClass := self fetchClassOf: rcvr.
	[currentClass ~= lookupClass]
		whileTrue:
		[currentClass := self superclassOf: currentClass.
		currentClass = nilObj ifTrue: [^ self primitiveFail]].

	self popStack.
	self primitivePerformAt: lookupClass.
	successFlag ifFalse:
		[self push: lookupClass]! !

!Interpreter methodsFor: 'control primitives' stamp: 'di 4/12/1999 10:55'!
primitivePerformWithArgs

	| lookupClass rcvr |
	rcvr := self stackValue: argumentCount.
	lookupClass := self fetchClassOf: rcvr.
	self primitivePerformAt: lookupClass.
! !

!Interpreter methodsFor: 'process primitives' stamp: 'ar (auto pragmas 12/08) 2/27/2008 09:30'!
primitiveProfilePrimitive
	"Primitive. Answer the last primitive method sampled by the profiler."
	<export: true>
	self methodArgumentCount = 0 ifFalse:[^self success: false].
	self pop: 1.
	self push: profileMethod.! !

!Interpreter methodsFor: 'process primitives' stamp: 'ar (auto pragmas 12/08) 2/27/2008 09:29'!
primitiveProfileSample
	"Primitive. Answer the last sample taken by the profiler, or nil if the profiler isn't active.
	See also primitiveProfileStart."
	<export: true>
	self methodArgumentCount = 0 ifFalse:[^self success: false].
	self pop: 1 thenPush: profileProcess.! !

!Interpreter methodsFor: 'process primitives' stamp: 'bgf 3/2/2009 01:58'!
primitiveProfileSemaphore
	"Primitive. Install the semaphore to be used for profiling, 
	or nil if no semaphore should be used.
	See also primitiveProfileStart."
	| sema |
	<export: true>
	self methodArgumentCount = 1 ifFalse:[^self success: false].
	sema := self stackValue: 0.
	sema == nilObj ifFalse:[
		(self fetchClassOf: sema) == (self splObj: ClassSemaphore) 
			ifFalse:[^self success: false]].
	profileSemaphore := sema.
	self pop: 1.! !

!Interpreter methodsFor: 'process primitives' stamp: 'ar (auto pragmas 12/08) 2/26/2008 15:50'!
primitiveProfileStart
	"Primitive. Begin profiling execution every by using the interrupt check-counter instead of a time-based process (which is limited to timing resolution and triggers off the same signal that many of the processes being profiled trigger off leading to consistently wrong results).
	The argument is the number of interrupt checks (method activations) to let go by before taking a sample. The sample is being stored in the profileSample iVar which can be retrieved by executing primitiveProfileSample. When a sample is taken, it signals the semaphore specified in primitiveProfileSemaphore.
	If the argument is less or equal to zero, it disables profiling."
	| deltaTicks |
	<export: true>
	self methodArgumentCount = 1 ifFalse:[^self success: false].
	deltaTicks := self stackIntegerValue: 0.
	successFlag ifTrue:[
		nextProfileTick := self ioHighResClock + deltaTicks.
		self pop: 1.
	].! !

!Interpreter methodsFor: 'quick primitives' stamp: 'jm 9/18/97 21:06'!
primitivePushFalse
	self popStack.
	self push: falseObj! !

!Interpreter methodsFor: 'quick primitives' stamp: 'jm 9/18/97 21:06'!
primitivePushMinusOne
	self popStack.
	self push: ConstMinusOne! !

!Interpreter methodsFor: 'quick primitives' stamp: 'jm 9/18/97 21:06'!
primitivePushNil
	self popStack.
	self push: nilObj! !

!Interpreter methodsFor: 'quick primitives' stamp: 'jm 9/18/97 21:06'!
primitivePushOne
	self popStack.
	self push: ConstOne! !

!Interpreter methodsFor: 'quick primitives' stamp: 'jm 9/18/97 21:06'!
primitivePushSelf
"	no-op, really...
	thisReceiver := self popStack.
	self push: thisReceiver
"! !

!Interpreter methodsFor: 'quick primitives' stamp: 'jm 9/18/97 21:06'!
primitivePushTrue
	self popStack.
	self push: trueObj! !

!Interpreter methodsFor: 'quick primitives' stamp: 'jm 9/18/97 21:06'!
primitivePushTwo
	self popStack.
	self push: ConstTwo! !

!Interpreter methodsFor: 'quick primitives' stamp: 'jm 9/18/97 21:06'!
primitivePushZero
	self popStack.
	self push: ConstZero! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'ar (auto pragmas 12/08) 3/19/2008 21:39'!
primitiveQuoLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| integerRcvr integerArg result oopResult |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>
	<var: 'result' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).
	integerArg = 0 ifTrue:[self primitiveFail].
	successFlag ifFalse:[^nil].

	integerRcvr > 0 ifTrue: [
		integerArg > 0 
			ifTrue: [result := integerRcvr // integerArg]
			ifFalse: [result := 0 - (integerRcvr // (0 - integerArg))].
	] ifFalse: [
		integerArg > 0 
			ifTrue: [result := 0 - ((0 - integerRcvr) // integerArg)]
			ifFalse: [result := (0 - integerRcvr) // (0 - integerArg)].
	].

	oopResult := self signed64BitIntegerFor: result.
	successFlag ifTrue:[self pop: 2 thenPush: oopResult].
! !

!Interpreter methodsFor: 'primitive support' stamp: 'eem 6/30/2008 10:24'!
primitiveResponse
	"Called under the assumption that primitiveFunctionPointer is not preloaded"
	primitiveFunctionPointer := self functionPointerFor: primitiveIndex inClass: nil.
	^self internalPrimitiveResponse! !

!Interpreter methodsFor: 'process primitives' stamp: 'tpr 3/15/2004 14:32'!
primitiveResume
"put this process on the scheduler's lists thus allowing it to proceed next time there is a chance for processes of it's priority level"
	| proc |
	proc := self stackTop.  "rcvr"
	"self success: ((self fetchClassOf: proc) = (self splObj: ClassProcess))."
	successFlag ifTrue: [ self resume: proc ].! !

!Interpreter methodsFor: 'memory space primitives' stamp: 'ar (auto pragmas 12/08) 1/18/2005 15:42'!
primitiveSetGCSemaphore
	"Primitive. Indicate the semaphore to be signalled for upon garbage collection"
	| index |
	<export: true>
	index := self stackIntegerValue: 0.
	successFlag ifTrue:[
		gcSemaphoreIndex := index.
		self pop: argumentCount.
	].! !

!Interpreter methodsFor: 'I/O primitives' stamp: 'ar (auto pragmas 12/08) 10/12/2007 16:21'!
primitiveSetLogDirectory
	"Primitive. Set the VM's log directory"
	| stringOop sz |
	<export: true>
	stringOop := self stackTop.
	(self isBytes: stringOop) ifFalse:[^self success: false].
	sz := self byteSizeOf: stringOop.
	self ioSetLogDirectory: (self firstIndexableField: stringOop) OfSize: sz.
	successFlag ifTrue:[self pop: self methodArgumentCount]! !

!Interpreter methodsFor: 'I/O primitives' stamp: 'ar (auto pragmas 12/08) 3/8/2007 14:10'!
primitiveSetWindowLabel
	"Primitive. Set the OS window's label"
	| labelOop sz |
	<export: true>
	labelOop := self stackTop.
	(self isBytes: labelOop) ifFalse:[^self success: false].
	sz := self byteSizeOf: labelOop.
	self ioSetWindowLabel: (self firstIndexableField: labelOop) OfSize: sz.
	successFlag ifTrue:[self pop: self methodArgumentCount]! !

!Interpreter methodsFor: 'I/O primitives' stamp: 'ar (auto pragmas 12/08) 2/27/2008 10:54'!
primitiveSetWindowSize
	"Primitive. Answer the OS window's size (extent)"
	| w h |
	<export: true>
	h := self stackIntegerValue: 0.
	w := self self stackIntegerValue: 1.
	successFlag ifTrue:[
		self ioSetWindowWidth: w Height: h.
		self pop: self methodArgumentCount.
	].! !

!Interpreter methodsFor: 'system control primitives' stamp: 'tpr 5/9/2005 18:58'!
primitiveSignalAtMilliseconds
	"Cause the time semaphore, if one has been registered, to
	be signalled when the millisecond clock is greater than or
	equal to the given tick value. A tick value of zero turns off
	timer interrupts."
	| tick sema |
	tick := self popInteger.
	sema := self popStack.
	successFlag
		ifTrue: [(self fetchClassOf: sema) = (self splObj: ClassSemaphore)
				ifTrue: [self
						storePointer: TheTimerSemaphore
						ofObject: specialObjectsOop
						withValue: sema.
					nextWakeupTick := tick]
				ifFalse: [self
						storePointer: TheTimerSemaphore
						ofObject: specialObjectsOop
						withValue: nilObj.
					nextWakeupTick := 0]]
		ifFalse: [self unPop: 2]! !

!Interpreter methodsFor: 'process primitives' stamp: 'eem 8/21/2008 10:58'!
primitiveStartVMProfiling
	"Primitive. Start the VM profiler."
	self cCode: 'ioControlProfile(1,0,0,0,0)'.
	self pop: argumentCount! !

!Interpreter methodsFor: 'process primitives' stamp: 'eem 07/28/2014 07:54'!
primitiveStopVMProfiling
	"Primitive. Stop the VM profiler and either copy the histogram data into the
	 supplied arguments, if they're non-nil.  Fail if the arguments are not of the right type or size."
	| vmHistArrayOrNil vmHist vmBins easHistArrayOrNil easHist easBins |
	<var: #vmHist type: #'long *'>
	<var: #vmBins type: #long>
	<var: #easHist type: #'long *'>
	<var: #easBins type: #long>
	self success: argumentCount = 2.
	vmHistArrayOrNil := self stackObjectValue: 1.
	easHistArrayOrNil := self stackObjectValue: 0.
	successFlag ifFalse:
		[^nil].
	"Both args must be either nil or arrays.  If they're arrays and the wrong size we incorrectly stop profiling."
	((vmHistArrayOrNil = nilObj or: [(self fetchClassOfNonImm: vmHistArrayOrNil) = (self splObj: ClassArray)])
	 and: [(self fetchClassOfNonImm: vmHistArrayOrNil) = (self fetchClassOfNonImm: easHistArrayOrNil)]) ifFalse:
		[^self primitiveFail].
	self cCode: 'ioControlProfile(0,&vmHist,&vmBins,&easHist,&easBins)'
		inSmalltalk: [vmHist := vmBins := easHist := easBins := 0].
	vmHistArrayOrNil ~= nilObj ifTrue:
		[((self numSlotsOf: vmHistArrayOrNil) = vmBins
		  and: [(self numSlotsOf: easHistArrayOrNil) = easBins]) ifFalse:
			[^self primitiveFail].
		0 to: vmBins - 1 do:
			[:i|
			self storePointerUnchecked: i
				ofObject: vmHistArrayOrNil
				withValue: (self integerObjectOf: (vmHist at: i))].
		0 to: easBins - 1 do:
			[:i|
			self storePointerUnchecked: i
				ofObject: easHistArrayOrNil
				withValue: (self integerObjectOf: (easHist at: i))]].
	self pop: argumentCount! !

!Interpreter methodsFor: 'object access primitives' stamp: 'eem 11/13/2014 13:34'!
primitiveStoreStackp
	"Atomic store into context stackPointer. 
	Also ensures that any newly accessible cells are initialized to nil "
	| ctxt newStackp stackp |
	ctxt := self stackValue: 1.
	newStackp := self stackIntegerValue: 0.
	self success: newStackp >= 0.
	self success: newStackp <= (LargeContextSlots - CtxtTempFrameStart).
	successFlag ifFalse: [^ self primitiveFail].
	stackp := self fetchStackPointerOf: ctxt.
	"Nil any newly accessible cells"
	stackp + 1 to: newStackp do:
		[:i | self storePointerUnchecked: i + CtxtTempFrameStart - 1 ofObject: ctxt withValue: nilObj].
	self storeStackPointerValue: newStackp inContext: ctxt.
	self pop: 1! !

!Interpreter methodsFor: 'arithmetic largeint primitives' stamp: 'eem 1/13/2009 15:12'!
primitiveSubtractLargeIntegers
	"Primitive arithmetic operations for large integers in 64 bit range"
	| integerRcvr integerArg integerArgNegated result oopResult |
	<export: true>
	<var: 'integerRcvr' type: 'sqLong'>
	<var: 'integerArg' type: 'sqLong'>
	<var: 'integerArgNegated' type: 'sqLong'>
	<var: 'result' type: 'sqLong'>

	integerArg := self signed64BitValueOf: (self stackValue: 0).
	integerRcvr := self signed64BitValueOf: (self stackValue: 1).
	successFlag ifFalse:
		[^nil].

	"In the test for the argument being most -ve we would
	 have to check for zero if we don't do so here.
	 So do so here and short-circuit the whole subtraction."
	integerArg = 0
		ifTrue: [result := integerRcvr]
		ifFalse:
			["Reverse the argument so that we can recycle the overflow code from addition.
			 But the most -ve 64-bit value can overflow, so check; 0 - most -ve = most -ve"
			integerArgNegated := 0 - integerArg.
			integerArgNegated = integerArg ifTrue:
				[^self primitiveFail].

			"Compute the preliminary result (which may overflow)"
			result := integerRcvr + integerArgNegated.

			"Now check overflow conditions. First is whether rcvr and arg are of the same sign.
			 If they are we need to check for overflow more carefully."
			(integerRcvr bitXor: integerArgNegated) < 0 ifFalse:
				"Second is whether rcvr and result are of the same sign. If not, we have an overflow."
				[(integerRcvr bitXor: result) < 0 ifTrue:
					[^self primitiveFail]]].

	oopResult := self signed64BitIntegerFor: result.
	successFlag ifTrue:
		[self pop: 2 thenPush: oopResult]! !

!Interpreter methodsFor: 'process primitives' stamp: 'ajh 1/28/2003 12:58'!
primitiveTerminateTo
	"Primitive. Terminate up the context stack from the receiver up to but not including the argument, if previousContext is on my Context stack. Make previousContext my sender. This prim has to shadow the code in ContextPart>terminateTo: to be correct"
	| thisCntx currentCntx aContext nextCntx nilOop |
	aContext := self popStack.
	thisCntx := self popStack.

	"make sure that aContext is in my chain"
	(self context: thisCntx hasSender: aContext) ifTrue:[
		nilOop := nilObj.
		currentCntx := self fetchPointer: SenderIndex ofObject: thisCntx.
		[currentCntx = aContext] whileFalse: [
			nextCntx := self fetchPointer: SenderIndex ofObject: currentCntx.
			self storePointer: SenderIndex ofObject: currentCntx withValue: nilOop.
			self storePointer: InstructionPointerIndex ofObject: currentCntx withValue: nilOop.
			currentCntx := nextCntx]].

	self storePointer: SenderIndex ofObject: thisCntx withValue: aContext.
	^self push: thisCntx! !

!Interpreter methodsFor: 'system control primitives' stamp: 'dtl 8/13/2012 22:19'!
primitiveUTCMicrosecondClock
	"Answer the UTC microseconds since the Smalltalk epoch. The value is
	derived from the Posix epoch with a constant offset corresponding to
	elapsed microseconds between the two epochs according to RFC 868."
	| clock epochDelta uSecs |

	<export: true>
	<var: #clock type: 'usqLong'>
	<var: #offset type: 'int'>
	<var: #epochDelta declareC: 'static usqLong epochDelta= 2177452800000000ULL'>
	(self cCode: 'ioUtcWithOffset(&clock, &offset)' inSmalltalk: [-1]) = -1
		ifTrue: [^ self primitiveFail].
	clock := clock + epochDelta.
	uSecs := self positive64BitIntegerFor: clock.
	self pop: 1 thenPush: uSecs.
! !

!Interpreter methodsFor: 'system control primitives' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
primitiveVMParameter
	"Behaviour depends on argument count:
		0 args:	return an Array of VM parameter values;
		1 arg:	return the indicated VM parameter;
		2 args:	set the VM indicated parameter.
	VM parameters are numbered as follows:
		1	end of old-space (0-based, read-only)
		2	end of young-space (read-only)
		3	end of memory (read-only)
		4	allocationCount (read-only)
		5	allocations between GCs (read-write)
		6	survivor count tenuring threshold (read-write)
		7	full GCs since startup (read-only)
		8	total milliseconds in full GCs since startup (read-only)
		9	incremental GCs since startup (read-only)
		10	total milliseconds in incremental GCs since startup (read-only)
		11	tenures of surving objects since startup (read-only)
		12-20 specific to the translating VM
		21	root table size (read-only)
		22	root table overflows since startup (read-only)
		23	bytes of extra memory to reserve for VM buffers, plugins, etc.
		24	memory threshold above which shrinking object memory (rw)
		25	memory headroom when growing object memory (rw)
		26  interruptChecksEveryNms - force an ioProcessEvents every N milliseconds, in case the image  is not calling getNextEvent often (rw)
		27	number of times mark loop iterated for current IGC/FGC (read-only) includes ALL marking
		28	number of times sweep loop iterated  for current IGC/FGC (read-only)
		29	number of times make forward loop iterated for current IGC/FGC (read-only)
		30	number of times compact move loop iterated for current IGC/FGC (read-only)
		31	number of grow memory requests (read-only)
		32	number of shrink memory requests (read-only)
		33	number of root table entries used for current IGC/FGC (read-only)
		34	number of allocations done before current IGC/FGC (read-only)
		35	number of survivor objects after current IGC/FGC (read-only)
		36	millisecond clock when current IGC/FGC completed (read-only)
		37	number of marked objects for Roots of the world, not including Root Table entries for current IGC/FGC (read-only)
		38  milliseconds taken by current IGC  (read-only)
		39  Number of finalization signals for Weak Objects pending when current IGC/FGC completed (read-only)
		40	BytesPerWord for this image
		41  imageFormatVersion for the VM
		42	nil (number of stack pages in use in Stack VM)
		43	nil (desired number of stack pages in Stack VM)
		44	nil (size of eden, in bytes in Stack VM)
		45	nil (desired size of eden in Stack VM)
		46-55 nil; reserved for VM parameters that persist in the image (such as eden above)
		56	number of process switches since startup (read-only)
		57	number of ioProcessEvents calls since startup (read-only)
		58	number of ForceInterruptCheck calls since startup (read-only)
		59	number of check event calls since startup (read-only)
		
	Note: Thanks to Ian Piumarta for this primitive."

	| mem paramsArraySize result arg index |
	<var: #mem type: #usqInt>
	mem := self startOfMemory.
	paramsArraySize := 59.
	argumentCount = 0 ifTrue: [
		result := self instantiateClass: (self splObj: ClassArray) indexableSize: paramsArraySize.
		self storePointerUnchecked: 0	ofObject: result withValue: (self integerObjectOf: youngStart - mem).
		self storePointerUnchecked: 1	ofObject: result withValue: (self integerObjectOf: freeBlock - mem).
		self storePointerUnchecked: 2	ofObject: result withValue: (self integerObjectOf: endOfMemory - mem).
		self storePointerUnchecked: 3	ofObject: result withValue: (self integerObjectOf: allocationCount).
		self storePointerUnchecked: 4	ofObject: result withValue: (self integerObjectOf: allocationsBetweenGCs).
		self storePointerUnchecked: 5	ofObject: result withValue: (self integerObjectOf: tenuringThreshold).
		self storePointerUnchecked: 6	ofObject: result withValue: (self integerObjectOf: statFullGCs).
		self storePointerUnchecked: 7	ofObject: result withValue: (self integerObjectOf: statFullGCUsecs + 500 // 1000).
		self storePointerUnchecked: 8	ofObject: result withValue: (self integerObjectOf: statIncrGCs).
		self storePointerUnchecked: 9	ofObject: result withValue: (self integerObjectOf: statIncrGCUsecs + 500 // 1000).
		self storePointerUnchecked: 10	ofObject: result withValue: (self integerObjectOf: statTenures).
		11 to: 19 do:
			[:i | self storePointerUnchecked: i ofObject: result withValue: ConstZero].
		self storePointerUnchecked: 20	ofObject: result withValue: (self integerObjectOf: rootTableCount).
		self storePointerUnchecked: 21	ofObject: result withValue: (self integerObjectOf: statRootTableOverflows).
		self storePointerUnchecked: 22	ofObject: result withValue: (self integerObjectOf: extraVMMemory).
		self storePointerUnchecked: 23	ofObject: result withValue: (self integerObjectOf: shrinkThreshold).
		self storePointerUnchecked: 24	ofObject: result withValue: (self integerObjectOf: growHeadroom).
		self storePointerUnchecked: 25	ofObject: result withValue: (self integerObjectOf: interruptChecksEveryNms).
		self storePointerUnchecked: 26	ofObject: result withValue: (self integerObjectOf: statMarkCount).
		self storePointerUnchecked: 27	ofObject: result withValue: (self integerObjectOf: statSweepCount).
		self storePointerUnchecked: 28	ofObject: result withValue: (self integerObjectOf: statMkFwdCount).
		self storePointerUnchecked: 29	ofObject: result withValue: (self integerObjectOf: statCompMoveCount).
		self storePointerUnchecked: 30	ofObject: result withValue: (self integerObjectOf: statGrowMemory).
		self storePointerUnchecked: 31	ofObject: result withValue: (self integerObjectOf: statShrinkMemory).
		self storePointerUnchecked: 32	ofObject: result withValue: (self integerObjectOf: statRootTableCount).
		self storePointerUnchecked: 33	ofObject: result withValue: (self integerObjectOf: statAllocationCount).
		self storePointerUnchecked: 34	ofObject: result withValue: (self integerObjectOf: statSurvivorCount).
		self storePointerUnchecked: 35	ofObject: result withValue: (self integerObjectOf: statGCEndTime).
		self storePointerUnchecked: 36	ofObject: result withValue: (self integerObjectOf: statSpecialMarkCount).
		self storePointerUnchecked: 37	ofObject: result withValue: (self integerObjectOf: statIGCDeltaUsecs + 500 // 1000).
		self storePointerUnchecked: 38	ofObject: result withValue: (self integerObjectOf: statPendingFinalizationSignals).
		self storePointerUnchecked: 39	ofObject: result withValue: (self integerObjectOf: self wordSize).
		self storePointerUnchecked: 40	ofObject: result withValue: (self integerObjectOf: self imageFormatVersion).
		41 to: 54 do:
			[:i | self storePointerUnchecked: i ofObject: result withValue: nilObj].
		self storePointerUnchecked: 55	ofObject: result withValue: (self integerObjectOf: statProcessSwitch).
		self storePointerUnchecked: 56	ofObject: result withValue: (self integerObjectOf: statIOProcessEvents).
		self storePointerUnchecked: 57	ofObject: result withValue: (self integerObjectOf: statQuickCheckForEvents).
		self storePointerUnchecked: 58	ofObject: result withValue: (self integerObjectOf: statCheckForEvents).
		self pop: 1 thenPush: result.
		^nil].

	arg := self stackTop.
	(self isIntegerObject: arg) ifFalse: [^self primitiveFail].
	arg := self integerValueOf: arg.
	argumentCount = 1 ifTrue: [	 "read VM parameter"
		(arg < 1 or: [arg > paramsArraySize]) ifTrue: [^self primitiveFail].
		arg = 1		ifTrue: [result := youngStart - mem].
		arg = 2		ifTrue: [result := freeBlock - mem].
		arg = 3		ifTrue: [result := endOfMemory - mem].
		arg = 4		ifTrue: [result := allocationCount].
		arg = 5		ifTrue: [result := allocationsBetweenGCs].
		arg = 6		ifTrue: [result := tenuringThreshold].
		arg = 7		ifTrue: [result := statFullGCs].
		arg = 8		ifTrue: [result := statFullGCUsecs + 500 // 1000].
		arg = 9		ifTrue: [result := statIncrGCs].
		arg = 10	ifTrue: [result := statIncrGCUsecs + 500 // 1000].
		arg = 11	ifTrue: [result := statTenures].
		(arg >= 12
		 and: [arg <= 20]) ifTrue: [result := 0].
		arg = 21	ifTrue: [result := rootTableCount].
		arg = 22	ifTrue: [result := statRootTableOverflows].
		arg = 23	ifTrue: [result := extraVMMemory].
		arg = 24	ifTrue: [result := shrinkThreshold].
		arg = 25	ifTrue: [result := growHeadroom].
		arg = 26	ifTrue: [result := interruptChecksEveryNms]. 
		arg = 27	ifTrue: [result := statMarkCount]. 
		arg = 28	ifTrue: [result := statSweepCount]. 
		arg = 29	ifTrue: [result := statMkFwdCount]. 
		arg = 30	ifTrue: [result := statCompMoveCount]. 
		arg = 31	ifTrue: [result := statGrowMemory]. 
		arg = 32	ifTrue: [result := statShrinkMemory]. 
		arg = 33	ifTrue: [result := statRootTableCount]. 
		arg = 34	ifTrue: [result := statAllocationCount]. 
		arg = 35	ifTrue: [result := statSurvivorCount]. 
		arg = 36  	ifTrue: [result := statGCEndTime]. 
		arg = 37  	ifTrue: [result := statSpecialMarkCount]. 
		arg = 38  	ifTrue: [result := statIGCDeltaUsecs + 500 // 1000]. 
		arg = 39  	ifTrue: [result := statPendingFinalizationSignals]. 
		arg = 40  	ifTrue: [result := self wordSize].
		arg = 41  	ifTrue: [result := self imageFormatVersion].
		(arg >= 42
		 and: [arg <= 55]) ifTrue: [result := nilObj].
		arg = 56  	ifTrue: [result := statProcessSwitch].
		arg = 57  	ifTrue: [result := statIOProcessEvents].
		arg = 58  	ifTrue: [result := statQuickCheckForEvents].
		arg = 59  	ifTrue: [result := statCheckForEvents].
		self pop: 2 thenPush: (self integerObjectOf: result).
		^nil].

	"write a VM parameter"
	argumentCount = 2 ifFalse: [^self primitiveFail].
	index := self stackValue: 1.
	(self isIntegerObject: index) ifFalse: [^self primitiveFail].
	index := self integerValueOf: index.
	index <= 0 ifTrue: [^self primitiveFail].
	successFlag := false.
	index = 5 ifTrue: [
		result := allocationsBetweenGCs.
		allocationsBetweenGCs := arg.
		successFlag := true].
	index = 6 ifTrue: [
		result := tenuringThreshold.
		tenuringThreshold := arg.
		successFlag := true].
	index = 23 ifTrue: [
		result := extraVMMemory.
		extraVMMemory := arg.
		successFlag := true].
	index = 24 ifTrue: [
		result := shrinkThreshold.
		arg > 0 ifTrue:[
			shrinkThreshold := arg.
			successFlag := true]].
	index = 25 ifTrue: [
		result := growHeadroom.
		arg > 0 ifTrue:[
			growHeadroom := arg.
			successFlag := true]].
	index = 26 ifTrue: [
		arg > 1 ifTrue:[
			result := interruptChecksEveryNms.
			interruptChecksEveryNms := arg.
			successFlag := true]]. 

	successFlag ifTrue: [
		self pop: 3 thenPush: (self integerObjectOf: result).  "return old value"
		^ nil].

	self primitiveFail.  "attempting to write a read-only parameter"



! !

!Interpreter methodsFor: 'process primitives' stamp: 'eem 07/28/2014 07:54'!
primitiveVMProfileInfoInto
	"Primitive. Answer whether the profiler is running or not.
	 If the argument is an Array of suitable size fill it with the following information:
		1. the addresses of the first element of the VM histogram (the first address in the executable)
		2. the address following the last element (the last address in the executable, excluding dynamically linked libraries)
		3. the size of the VM histogram in bins (each bin is a 4 byte unsigned long)
		4. the size of the VM histogram in bins (each bin is a 4 byte unsigned long)"
	| info running exeStart exeLimit vmBins easBins |
	<var: #exeStart type: #'char *'>
	<var: #exeLimit type: #'char *'>
	<var: #vmBins type: #long>
	<var: #easBins type: #long>
	self success: argumentCount = 1.
	successFlag ifTrue:
		[info := self stackObjectValue: 0.
		 info ~= nilObj ifTrue:
			[self assertClassOf: info is: (self splObj: ClassArray).
			 self success: (self numSlotsOf: info) >= 4]].
	successFlag ifFalse:
		[^nil].
	
	self cCode: 'ioProfileStatus(&running,&exeStart,&exeLimit,0,&vmBins,0,&easBins)'
		inSmalltalk: [running := exeStart := exeLimit := vmBins := easBins := 0].
	info ~= nilObj ifTrue:
		[self storePointerUnchecked: 0
			ofObject: info
			withValue: (self integerObjectOf: (self oopForPointer: exeStart)).
		self storePointerUnchecked: 1
			ofObject: info
			withValue: (self integerObjectOf: (self oopForPointer: exeLimit)).
		self storePointerUnchecked: 2
			ofObject: info
			withValue: (self integerObjectOf: vmBins).
		self storePointerUnchecked: 3
			ofObject: info
			withValue: (self integerObjectOf: easBins)].
	self pop: 2 thenPushBool: running! !

!Interpreter methodsFor: 'process primitives' stamp: 'eem 07/28/2014 07:54'!
primitiveVMProfileSamplesInto
	"Primitive.
	 0 args: Answer whether the VM Profiler is running or not.
	 1 arg:	Copy the sample data into the supplied argument, which must be a Bitmap
			of suitable size. Answer the number of samples copied into the buffer."
	| sampleBuffer sampleBufferAddress running bufferSize numSamples |
	<var: #bufferSize type: #long>
	<var: #sampleBufferAddress type: #'unsigned long *'>
	self cCode: 'ioNewProfileStatus(&running,&bufferSize)'
		inSmalltalk: [running := false. bufferSize := 0].
	argumentCount = 0 ifTrue:
		[^self pop: 1 thenPushBool: running].
	self success: argumentCount = 1.
	successFlag ifTrue:
		[sampleBuffer := self stackObjectValue: 0.
		 self assertClassOf: sampleBuffer is: (self splObj: ClassBitmap).
		 self success: (self numSlotsOf: sampleBuffer) >= bufferSize].
	successFlag ifFalse:
		[^nil].
	sampleBufferAddress := self firstFixedField: sampleBuffer.
	numSamples := self cCode: 'ioNewProfileSamplesInto(sampleBufferAddress)'
						inSmalltalk: [sampleBufferAddress := sampleBufferAddress].
	self pop: argumentCount + 1 thenPushInteger: numSamples! !

!Interpreter methodsFor: 'control primitives' stamp: 'tpr 3/23/2004 18:08'!
primitiveValue
	| blockContext blockArgumentCount initialIP |
	blockContext := self stackValue: argumentCount.
	blockArgumentCount := self argumentCountOfBlock: blockContext.
	self success: (argumentCount = blockArgumentCount
			and: [(self fetchPointer: CallerIndex ofObject: blockContext) = nilObj]).
	successFlag
		ifTrue: [self transfer: argumentCount
				fromIndex: self stackPointerIndex - argumentCount + 1
				ofObject: activeContext
				toIndex: TempFrameStart
				ofObject: blockContext.

			"Assume: The call to transfer:... makes blockContext a root if necessary,
			 allowing use to use unchecked stored in the following code."
			self pop: argumentCount + 1.
			initialIP := self fetchPointer: InitialIPIndex	ofObject: blockContext.
			self storePointerUnchecked: InstructionPointerIndex ofObject: blockContext withValue: initialIP.
			self storeStackPointerValue: argumentCount inContext: blockContext.
			self storePointerUnchecked: CallerIndex ofObject: blockContext withValue: activeContext.
			self newActiveContext: blockContext]! !

!Interpreter methodsFor: 'control primitives' stamp: 'tpr (auto pragmas 12/08) 2/20/2004 19:46'!
primitiveValueUninterruptably
	"The only purpose of this primitive is to indicate that the new EH mechanisms are supported."
	<inline: false>
	^self primitiveValue! !

!Interpreter methodsFor: 'control primitives' stamp: 'eem 07/28/2014 07:54'!
primitiveValueWithArgs
	| argumentArray blockContext blockArgumentCount arrayArgumentCount initialIP |
	argumentArray := self popStack.
	blockContext := self popStack.
	blockArgumentCount := self argumentCountOfBlock: blockContext.
	"If the argArray isnt actually an Array we ahve to unpop the above two"
	(self isArray: argumentArray) ifFalse: [self unPop:2. ^self primitiveFail].

	successFlag ifTrue: [arrayArgumentCount := self numSlotsOf: argumentArray.
			self success: (arrayArgumentCount = blockArgumentCount
						and: [(self fetchPointer: CallerIndex ofObject: blockContext) = nilObj])].
	successFlag
		ifTrue: [self
				transfer: arrayArgumentCount
				fromIndex: 0
				ofObject: argumentArray
				toIndex: TempFrameStart
				ofObject: blockContext.
			"Assume: The call to transfer:... makes blockContext a root if necessary, 
			allowing use to use unchecked stored in the following code. "
			initialIP := self fetchPointer: InitialIPIndex ofObject: blockContext.
			self
				storePointerUnchecked: InstructionPointerIndex
				ofObject: blockContext
				withValue: initialIP.
			self storeStackPointerValue: arrayArgumentCount inContext: blockContext.
			self
				storePointerUnchecked: CallerIndex
				ofObject: blockContext
				withValue: activeContext.
			self newActiveContext: blockContext]
		ifFalse: [self unPop: 2]! !

!Interpreter methodsFor: 'debug printing' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:32'!
print: s
	"For testing in Smalltalk, this method should be overridden in a subclass."

	<var: #s type: 'char *'>
	self cCode: 'printf("%s", s)'.! !

!Interpreter methodsFor: 'debug printing' stamp: 'ar 6/15/2007 15:59'!
printAllStacks
	"Print all the stacks of all running processes, including those that are currently suspended."
	| oop proc ctx |
	proc := self fetchPointer: ActiveProcessIndex ofObject: self schedulerPointer.
	self printNameOfClass: (self fetchClassOf: proc) count: 5.
	self cr.
	self printCallStackOf: activeContext. "first the active context"
	oop := self firstObject.
	[self oop: oop isLessThan: endOfMemory] whileTrue:[
		(self fetchClassOf: oop) == self classSemaphore ifTrue:[
			proc := self fetchPointer: FirstLinkIndex ofObject: oop.
			[proc == self nilObject] whileFalse:[
				self cr.
				self printNameOfClass: (self fetchClassOf: proc) count: 5.
				self cr.
				ctx := self fetchPointer: SuspendedContextIndex ofObject: proc.
				ctx == self nilObject ifFalse:[self printCallStackOf: ctx].
				proc := self fetchPointer: NextLinkIndex ofObject: proc].
		].
		oop := self objectAfter: oop.
	].! !

!Interpreter methodsFor: 'debug printing' stamp: 'eem 3/4/2009 14:37'!
printCallStack
	self printCallStackOf: activeContext! !

!Interpreter methodsFor: 'debug printing' stamp: 'ajh (auto pragmas 12/08) 3/16/2003 13:13'!
printCallStackOf: aContext

	| ctxt home methClass methodSel message |
	<inline: false>
	ctxt := aContext.
	[ctxt = nilObj] whileFalse: [
		(self fetchClassOf: ctxt) = (self splObj: ClassBlockContext)
			ifTrue: [ home := self fetchPointer: HomeIndex ofObject: ctxt ]
			ifFalse: [ home := ctxt ].
		methClass :=
			self findClassOfMethod: (self fetchPointer: MethodIndex ofObject: home)
					   forReceiver: (self fetchPointer: ReceiverIndex ofObject: home).
		methodSel :=
			self findSelectorOfMethod: (self fetchPointer: MethodIndex ofObject: home)
						 forReceiver: (self fetchPointer: ReceiverIndex ofObject: home).
		self printNum: ctxt.
		self print: ' '.
		ctxt = home ifFalse: [ self print: '[] in ' ].
		self printNameOfClass: methClass count: 5.
		self print: '>'.
		methodSel = nilObj
			ifTrue: [self print: '?']
			ifFalse: [self printStringOf: methodSel].
		methodSel = (self splObj: SelectorDoesNotUnderstand) ifTrue: [
			"print arg message selector"
			message := self fetchPointer: 0 + TempFrameStart ofObject: home.
			methodSel := self fetchPointer: MessageSelectorIndex ofObject: message.
			self print: ' '.
			self printStringOf: methodSel.
		].
		self cr.

		ctxt := (self fetchPointer: SenderIndex ofObject: ctxt).
	].! !

!Interpreter methodsFor: 'debug printing' stamp: ''!
printChar: aByte
	"For testing in Smalltalk, this method should be overridden in a subclass."

	self putchar: aByte.! !

!Interpreter methodsFor: 'debug printing' stamp: 'eem 2/22/2016 17:03'!
printHex: n
	"Print n in hex,  in the form '    0x1234', padded to a width of 10 characters
	 in 32-bits ('0x' + 8 nibbles) or 18 characters in 64-bits ('0x' + 16 nibbles)"
	<api>
	| len buf |
	<var: #buf declareC: 'char buf[37]'> "large enough for a 64-bit value in hex plus the null plus 16 spaces"
	self cCode: 'memset(buf,'' '',36)' inSmalltalk: [buf := 'doh!!'].
	len := self cCode: 'sprintf(buf + 2 + 2 * BytesPerWord, "0x%lx", (unsigned long)(n))'.
	self cCode: 'printf("%s", buf + len)'.
	len touch: buf! !

!Interpreter methodsFor: 'debug printing' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
printNameOfClass: classOop count: cnt
	"Details: The count argument is used to avoid a possible infinite recursion if classOop is a corrupted object."

	cnt <= 0 ifTrue: [^self print: 'bad class'].
	((self sizeBitsOf: classOop) = metaclassSizeBits
	  and: [metaclassSizeBits >= (6 * self wordSize)])	"(Metaclass instSize * 4)"
		ifTrue: [self printNameOfClass: (self fetchPointer: 5 "thisClass" ofObject: classOop) count: cnt - 1.
				self print: ' class']
		ifFalse: [self printStringOf: (self fetchPointer: 6 "name" ofObject: classOop)]! !

!Interpreter methodsFor: 'debug printing' stamp: ''!
printNum: n
	"For testing in Smalltalk, this method should be overridden in a subclass."

	self cCode: 'printf("%ld", (long) n)'.! !

!Interpreter methodsFor: 'debug printing' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
printOop: oop

	| fmt lastIndex |
	<inline: false>
	self printNum: oop.
	(self isIntegerObject: oop) ifTrue:
		[^self cCode: 'printf("=%ld\n", (long)integerValueOf(oop))' inSmalltalk: [self shortPrint: oop]].
	self print: ': a(n) '.
	self printNameOfClass: (self fetchClassOf: oop) count: 5.
	self cr.
	fmt := self formatOf: oop.
	(fmt > 4 and: [fmt < 12]) ifTrue:
		[^self printStringOf: oop].
	lastIndex := 64 min: ((self lastPointerOf: oop) / self wordSize).
	lastIndex > 0 ifTrue:
		[1 to: lastIndex do:
			[:index|
			self cCode: 'printf(" %ld", fetchPointerofObject(index - 1, oop))'
				inSmalltalk: [self space; print: (self fetchPointer: index - 1 ofObject: oop) printString; space.
							 self print: (self shortPrint: (self fetchPointer: index - 1 ofObject: oop))].
			(index \\ 8) = 0 ifTrue:
				[self cr]].
		(lastIndex \\ 8) = 0 ifFalse:
			[self cr]]! !

!Interpreter methodsFor: 'debug printing' stamp: 'eem 10/24/2008 14:25'!
printStringOf: oop

	| fmt cnt i |
	(self isIntegerObject: oop) ifTrue:[^nil].
	fmt := self formatOf: oop.
	fmt < 8 ifTrue: [ ^nil ].

	cnt := 100 min: (self lengthOf: oop).
	i := 0.
	[i < cnt] whileTrue: [
		self printChar: (self fetchByte: i ofObject: oop).
		i := i + 1.
	].
	self cCode: 'fflush(stdout)' inSmalltalk: []! !

!Interpreter methodsFor: 'debug printing' stamp: 'ar (auto pragmas 12/08) 2/20/2001 01:00'!
printUnbalancedStack: primIdx
	<inline: false>
	self print: 'Stack unbalanced after '.
	successFlag 
		ifTrue:[self print:'successful primitive '] 
		ifFalse:[self print: 'failed primitive '].
	self printNum: primIdx.
	self cr.
		! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
push: object

	| sp |
	self longAt: (sp := stackPointer + self wordSize) put: object.
	stackPointer := sp.! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:48'!
pushActiveContextBytecode
	"Puts reclaimability of this context in question."

	self fetchNextBytecode.
	reclaimableContextCount := 0.
	self internalPush: activeContext.
! !

!Interpreter methodsFor: 'contexts' stamp: ''!
pushBool: trueOrFalse

	trueOrFalse
		ifTrue: [ self push: trueObj ]
		ifFalse: [ self push: falseObj ].! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
pushClosureCopyCopiedValuesBytecode
	"The compiler has pushed the values to be copied, if any.  Find numArgs and numCopied in the byte following.
	 Create a Closure with space for the copiedValues and pop numCopied values off the stack into the closure.
	 Set numArgs as specified, and set startpc to the pc following the block size and jump over that code."
	| newClosure numArgsNumCopied numArgs numCopied blockSize |
	numArgsNumCopied := self fetchByte.
	numArgs := numArgsNumCopied bitAnd: 16rF.
	numCopied := numArgsNumCopied bitShift: -4.
	"Split blockSize := (self fetchByte * 256) + self fetchByte. into two because evaluation order in C is undefined."
	blockSize := self fetchByte << 8.
	blockSize := blockSize + self fetchByte.
	self externalizeIPandSP. "This is a pain."
	newClosure := self
					closureNumArgs: numArgs
					instructionPointer: ((self cCoerce: localIP to: 'sqInt') + 2 - (method+self baseHeaderSize))
					numCopiedValues: numCopied.
	self internalizeIPandSP.
	"Assume: have just allocated a new closure; it must be young. Thus, can use unchecked stores."
	self storePointerUnchecked: ClosureOuterContextIndex ofObject: newClosure withValue: activeContext.
	reclaimableContextCount := 0. "The closure refers to thisContext so it can't be reclaimed."
	numCopied > 0 ifTrue:
		[0 to: numCopied - 1 do:
			[:i|
			"Assume: have just allocated a new BlockClosure; it must be young.
			 Thus, can use unchecked stores."
			 self storePointerUnchecked: i + ClosureFirstCopiedValueIndex
				ofObject: newClosure
				withValue: (self internalStackValue: numCopied - i - 1)].
		 self internalPop: numCopied].
	localIP := localIP + blockSize.
	self fetchNextBytecode.
	self internalPush: newClosure! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:48'!
pushConstantFalseBytecode

	self fetchNextBytecode.
	self internalPush: falseObj.
! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:48'!
pushConstantMinusOneBytecode

	self fetchNextBytecode.
	self internalPush: ConstMinusOne.
! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:48'!
pushConstantNilBytecode

	self fetchNextBytecode.
	self internalPush: nilObj.
! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:48'!
pushConstantOneBytecode

	self fetchNextBytecode.
	self internalPush: ConstOne.
! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:48'!
pushConstantTrueBytecode

	self fetchNextBytecode.
	self internalPush: trueObj.
! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:48'!
pushConstantTwoBytecode

	self fetchNextBytecode.
	self internalPush: ConstTwo.
! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:48'!
pushConstantZeroBytecode

	self fetchNextBytecode.
	self internalPush: ConstZero.
! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:32'!
pushFloat: f

	<var: #f type: 'double '>
	self push: (self floatObjectOf: f).! !

!Interpreter methodsFor: 'contexts' stamp: 'eem 3/4/2009 14:39'!
pushInteger: integerValue
	self push: (self integerObjectOf: integerValue).
	^nil! !

!Interpreter methodsFor: 'stack bytecodes' stamp: ''!
pushLiteralConstant: literalIndex

	self internalPush: (self literal: literalIndex).! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'eem 11/13/2012 10:40'!
pushLiteralConstantBytecode
	<expandCases>
	self
		cCode: "this bytecode will be expanded so that refs to currentBytecode below will be constant"
			[self fetchNextBytecode.
			 self pushLiteralConstant: (currentBytecode bitAnd: 16r1F)]
		inSmalltalk: "Interpreter version has fetchNextBytecode out of order"
			[self pushLiteralConstant: (currentBytecode bitAnd: 16r1F).
			 self fetchNextBytecode]! !

!Interpreter methodsFor: 'stack bytecodes' stamp: ''!
pushLiteralVariable: literalIndex

	self internalPush:
		(self fetchPointer: ValueIndex ofObject: (self literal: literalIndex)).! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'eem 11/13/2012 10:43'!
pushLiteralVariableBytecode
	<expandCases>
	self
		cCode: "this bytecode will be expanded so that refs to currentBytecode below will be constant"
			[self fetchNextBytecode.
			 self pushLiteralVariable: (currentBytecode bitAnd: 16r1F)]
		inSmalltalk: "Interpreter version has fetchNextBytecode out of order"
			[self pushLiteralVariable: (currentBytecode bitAnd: 16r1F).
			 self fetchNextBytecode]! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'eem 7/22/2008 16:46'!
pushNewArrayBytecode
	| size popValues array |
	size := self fetchByte.
	popValues := size > 127.
	size := size bitAnd: 127.
	self fetchNextBytecode.
	self externalizeIPandSP.
	array := self instantiateClass: (self splObj: ClassArray) indexableSize: size.
	self internalizeIPandSP.
	popValues ifTrue:
		[0 to: size - 1 do:
			[:i|
			"Assume: have just allocated a new Array; it must be young. Thus, can use unchecked stores."
			self storePointerUnchecked: i ofObject: array withValue: (self internalStackValue: size - i - 1)].
		 self internalPop: size].
	self internalPush: array! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'jm 12/10/1998 16:48'!
pushReceiverBytecode

	self fetchNextBytecode.
	self internalPush: receiver.
! !

!Interpreter methodsFor: 'stack bytecodes' stamp: ''!
pushReceiverVariable: fieldIndex

	self internalPush:
		(self fetchPointer: fieldIndex ofObject: receiver).! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'eem 11/13/2012 10:44'!
pushReceiverVariableBytecode
	<expandCases>
	self
		cCode: "this bytecode will be expanded so that refs to currentBytecode below will be constant"
			[self fetchNextBytecode.
			 self pushReceiverVariable: (currentBytecode bitAnd: 16rF)]
		inSmalltalk: "Interpreter version has fetchNextBytecode out of order"
			[self pushReceiverVariable: (currentBytecode bitAnd: 16rF).
			 self fetchNextBytecode]! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'eem 5/27/2008 15:12'!
pushRemoteTemp: index inVectorAt: tempVectorIndex
	| tempVector |
	tempVector := self temporary: tempVectorIndex.
	self internalPush: (self fetchPointer: index ofObject: tempVector)! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'eem 5/27/2008 13:28'!
pushRemoteTempLongBytecode
	| remoteTempIndex tempVectorIndex |
	remoteTempIndex := self fetchByte.
	tempVectorIndex := self fetchByte.
	self fetchNextBytecode.
	self pushRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!Interpreter methodsFor: 'stack bytecodes' stamp: ''!
pushTemporaryVariable: temporaryIndex

	self internalPush: (self temporary: temporaryIndex).! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'eem 11/13/2012 10:45'!
pushTemporaryVariableBytecode
	<expandCases>
	self
		cCode: "this bytecode will be expanded so that refs to currentBytecode below will be constant"
			[self fetchNextBytecode.
			 self pushTemporaryVariable: (currentBytecode bitAnd: 16rF)]
		inSmalltalk: "Interpreter version has fetchNextBytecode out of order"
			[self pushTemporaryVariable: (currentBytecode bitAnd: 16rF).
			 self fetchNextBytecode]! !

!Interpreter methodsFor: 'image save/restore' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:32'!
putLong: aWord toFile: aFile
	"Append aWord to aFile in this platforms 'natural' byte order.  (Bytes will be swapped, if
	necessary, when the image is read on a different platform.) Set successFlag to false if
	the write fails."

	| objectsWritten |
	<var: #aFile type: 'sqImageFile '>

	objectsWritten := self cCode: 'sqImageFileWrite(&aWord, sizeof(aWord), 1, aFile)'.
	self success: objectsWritten = 1.
! !

!Interpreter methodsFor: 'process primitive support' stamp: ''!
putToSleep: aProcess
	"Save the given process on the scheduler process list for its priority."

	| priority processLists processList |
	priority := self quickFetchInteger: PriorityIndex ofObject: aProcess.
	processLists := self fetchPointer: ProcessListsIndex ofObject: self schedulerPointer.
	processList := self fetchPointer: priority - 1 ofObject: processLists.
	self addLastLink: aProcess toList: processList.! !

!Interpreter methodsFor: 'process primitive support' stamp: 'eem 11/3/2008 13:48'!
quickCheckForInterrupts
	"Quick check for possible user or timer interrupts. Decrement a counter and only do a real check when counter reaches zero or when a low space or user interrupt is pending."
	"Note: Clients that trigger interrupts should set use forceInterruptCheck to set interruptCheckCounter to zero and get immediate results."
	"Note: Requires that instructionPointer and stackPointer be external."

	statQuickCheckForEvents := statQuickCheckForEvents + 1.
	((interruptCheckCounter := interruptCheckCounter - 1) <= 0)
		ifTrue: [self checkForInterrupts].
! !

!Interpreter methodsFor: 'utilities' stamp: 'tpr 3/15/2004 20:35'!
quickFetchInteger: fieldIndex ofObject: objectPointer
	"Return the integer value of the field without verifying that it is an integer value!! For use in time-critical places where the integer-ness of the field can be guaranteed."

	^ self integerValueOf: (self fetchPointer: fieldIndex ofObject: objectPointer).! !

!Interpreter methodsFor: 'image save/restore' stamp: 'EstebanLorenzano 5/20/2015 12:08'!
readImageFromFile: f HeapSize: desiredHeapSize StartingAt: imageOffset
	"Read an image from the given file stream, allocating the given amount of memory to its object heap. Fail if the image has an unknown format or requires more than the given amount of memory."
	"Details: This method detects when the image was stored on a machine with the opposite byte ordering from this machine and swaps the bytes automatically. Furthermore, it allows the header information to start 512 bytes into the file, since some file transfer programs for the Macintosh apparently prepend a Mac-specific header of this size. Note that this same 512 bytes of prefix area could also be used to store an exec command on Unix systems, allowing one to launch Smalltalk by invoking the image name as a command."
	"This code is based on C code by Ian Piumarta and Smalltalk code by Tim Rowledge. Many thanks to both of you!!!!"

	| swapBytes headerStart headerSize dataSize oldBaseAddr minimumMemory memStart bytesRead bytesToShift heapSize |
	<var: #f type: 'sqImageFile '>
	<var: #desiredHeapSize type: 'usqInt'>
	<var: #headerStart type: 'squeakFileOffsetType '>
	<var: #dataSize type: 'size_t '>
	<var: #imageOffset type: 'squeakFileOffsetType '>

	metaclassSizeBits := 6 * self wordSize.	"guess (Metaclass instSize * BPW)"
	swapBytes := self checkImageVersionFrom: f startingAt: imageOffset.
	headerStart := (self sqImageFilePosition: f) - self wordSize.  "record header start position"

	headerSize			:= self getLongFromFile: f swap: swapBytes.
	dataSize				:= self getLongFromFile: f swap: swapBytes.
	oldBaseAddr			:= self getLongFromFile: f swap: swapBytes.
	specialObjectsOop	:= self getLongFromFile: f swap: swapBytes.
	lastHash			:= self getLongFromFile: f swap: swapBytes.
	savedWindowSize	:= self getLongFromFile: f swap: swapBytes.
	fullScreenFlag		:= self getLongFromFile: f swap: swapBytes.
	extraVMMemory		:= self getLongFromFile: f swap: swapBytes.

	lastHash = 0 ifTrue: [
		"lastHash wasn't stored (e.g. by the cloner); use 999 as the seed"
		lastHash := 999].

	"decrease Squeak object heap to leave extra memory for the VM"
	heapSize := self cCode: 'reserveExtraCHeapBytes(desiredHeapSize, extraVMMemory)'.

	"compare memory requirements with availability".
	minimumMemory := dataSize + 100000.  "need at least 100K of breathing room"
	heapSize < minimumMemory ifTrue: [
		self insufficientMemorySpecifiedError].

	"allocate a contiguous block of memory for the Squeak heap"
	memory := self cCode: 'sqAllocateMemory(minimumMemory, heapSize)'.
	memory = nil ifTrue: [self insufficientMemoryAvailableError].

	memStart := self startOfMemory.
	self setMemoryLimit: (memStart + heapSize) - 24.  "decrease memoryLimit a tad for safety"
	self setEndOfMemory: memStart + dataSize.

	"position file after the header"
	self sqImageFile: f Seek: headerStart + headerSize.

	"read in the image in bulk, then swap the bytes if necessary"
	self 
		cppIf: PharoVM 
		ifTrue: [  
			bytesRead := self cCode: 'sqImageFileReadEntireImage(pointerForOop(memory), sizeof(unsigned char), dataSize, f)' ]
		ifFalse: [ 
			bytesRead := self cCode: 'sqImageFileRead(pointerForOop(memory), sizeof(unsigned char), dataSize, f)' ].
	bytesRead ~= dataSize ifTrue: [self unableToReadImageError].

	swapBytes ifTrue: [self reverseBytesInImage].

	"compute difference between old and new memory base addresses"
	bytesToShift := memStart - oldBaseAddr.
	self initializeInterpreter: bytesToShift.  "adjusts all oops to new location"
	^ dataSize
! !

!Interpreter methodsFor: 'image save/restore' stamp: 'eem 11/14/2019 14:27'!
readableFormat: imageVersion
	"Anwer true if images of the given format are readable by this interpreter. Allows a virtual machine to accept selected older image formats."
	<api>
	^ imageVersion = self imageFormatVersion
	or: [imageVersion = self imageFormatCompatibilityVersion]
"
	Example of multiple formats:
	^ (imageVersion = self imageFormatVersion) or: [imageVersion = 6504]
"! !

!Interpreter methodsFor: 'contexts' stamp: 'dtl (auto pragmas 12/08) 4/22/2007 20:13'!
recycleContextIfPossible: cntxOop 
	"If possible, save the given context on a list of free contexts to 
	be recycled."
	"Note: The context is not marked free, so it can be reused 
	with minimal fuss. The recycled context lists are cleared at 
	every garbage collect."
	| header |
	<inline: true>
	"only recycle young contexts (which should be most of them)"
	(self oop: cntxOop isGreaterThanOrEqualTo: youngStart)
		ifTrue: [header := self baseHeader: cntxOop.
			(self isMethodContextHeader: header)
				ifTrue: ["It's a young context, alright."
					(header bitAnd: SizeMask) = SmallContextSize
						ifTrue: ["Recycle small contexts"
							self storePointerUnchecked: 0 ofObject: cntxOop withValue: freeContexts.
							freeContexts := cntxOop].
					(header bitAnd: SizeMask) = LargeContextSize
						ifTrue: ["Recycle large contexts"
							self storePointerUnchecked: 0 ofObject: cntxOop withValue: freeLargeContexts.
							freeLargeContexts := cntxOop]]]! !

!Interpreter methodsFor: 'callback support' stamp: 'eem 6/1/2010 18:27'!
reestablishContextPriorToCallback: callbackContext
	"callbackContext is an activation of invokeCallback:stack:registers:jmpbuf:.  Its sender
	 is the interpreter's state prior to the callback.  Reestablish that state."
	| calloutContext |
	<export: true>
	(self fetchClassOf: callbackContext) ~~ (self splObj: ClassMethodContext) ifTrue:
		[^false].
	calloutContext := self fetchPointer: SenderIndex ofObject: callbackContext.
	self newActiveContext: calloutContext.
	^true! !

!Interpreter methodsFor: 'process primitive support' stamp: 'tpr 3/24/2004 20:54'!
removeFirstLinkOfList: aList 
	"Remove the first process from the given linked list."
	| first last next |
	first := self fetchPointer: FirstLinkIndex ofObject: aList.
	last := self fetchPointer: LastLinkIndex ofObject: aList.
	first = last
		ifTrue: [self storePointer: FirstLinkIndex ofObject: aList withValue: nilObj.
			self storePointer: LastLinkIndex ofObject: aList withValue: nilObj]
		ifFalse: [next := self fetchPointer: NextLinkIndex ofObject: first.
			self storePointer: FirstLinkIndex ofObject: aList withValue: next].
	self storePointer: NextLinkIndex ofObject: first withValue: nilObj.
	^ first! !

!Interpreter methodsFor: 'process primitive support' stamp: 'ar 12/4/2007 16:21'!
removeProcess: aProcess fromList: aList 
	"Remove a given process from a linked list. May fail if aProcess is not on the list."
	| firstLink lastLink nextLink tempLink |
	firstLink := self fetchPointer: FirstLinkIndex ofObject: aList.
	lastLink := self fetchPointer: LastLinkIndex ofObject: aList.
	aProcess  == firstLink ifTrue:[
		nextLink := self fetchPointer: NextLinkIndex ofObject: aProcess .
		self storePointer: FirstLinkIndex ofObject: aList withValue: nextLink.
		aProcess  == lastLink ifTrue:[
			self storePointer: LastLinkIndex ofObject: aList withValue: self nilObject.
		].
	] ifFalse:[
		tempLink := firstLink.
		[tempLink == self nilObject ifTrue:[^self success: false]. "fail"
		nextLink := self fetchPointer: NextLinkIndex ofObject: tempLink.
		nextLink == aProcess] whileFalse:[
			tempLink := self fetchPointer: NextLinkIndex ofObject: tempLink.
		].
		nextLink := self fetchPointer: NextLinkIndex ofObject: aProcess.
		self storePointer: NextLinkIndex ofObject: tempLink withValue: nextLink.
		aProcess  == lastLink ifTrue:[
			self storePointer: LastLinkIndex ofObject: aList withValue: tempLink.
		].
	].
	self storePointer: NextLinkIndex ofObject: aProcess withValue: self nilObject.
! !

!Interpreter methodsFor: 'process primitive support' stamp: 'tpr (auto pragmas 12/08) 3/24/2004 20:55'!
resume: aProcess 
	| activeProc activePriority newPriority |
	<inline: false>
	activeProc := self fetchPointer: ActiveProcessIndex ofObject: self schedulerPointer.
	activePriority := self quickFetchInteger: PriorityIndex ofObject: activeProc.
	newPriority := self quickFetchInteger: PriorityIndex ofObject: aProcess.
	newPriority > activePriority
		ifTrue: [self putToSleep: activeProc.
			self transferTo: aProcess]
		ifFalse: [self putToSleep: aProcess]! !

!Interpreter methodsFor: 'return bytecodes' stamp: 'ar 7/6/2003 21:55'!
returnFalse
	localReturnContext := self sender.
	localReturnValue := falseObj.
	self commonReturn.
! !

!Interpreter methodsFor: 'return bytecodes' stamp: 'ar 7/6/2003 21:55'!
returnNil
	localReturnContext := self sender.
	localReturnValue := nilObj.
	self commonReturn.! !

!Interpreter methodsFor: 'return bytecodes' stamp: 'ar 7/6/2003 21:55'!
returnReceiver
	localReturnContext := self sender.
	localReturnValue := receiver.
	self commonReturn.! !

!Interpreter methodsFor: 'return bytecodes' stamp: 'ar 7/6/2003 21:55'!
returnTopFromBlock
	"Return to the caller of the method containing the block."
	localReturnContext := self caller.  "Note: caller, not sender!!"
	localReturnValue := self internalStackTop.
	self commonReturn.! !

!Interpreter methodsFor: 'return bytecodes' stamp: 'ar 7/6/2003 21:55'!
returnTopFromMethod
	localReturnContext := self sender.
	localReturnValue := self internalStackTop.
	self commonReturn.! !

!Interpreter methodsFor: 'return bytecodes' stamp: 'ar 7/6/2003 21:55'!
returnTrue
	localReturnContext := self sender.
	localReturnValue := trueObj.
	self commonReturn.! !

!Interpreter methodsFor: 'image save/restore' stamp: 'di 10/2/97 00:31'!
reverseBytesInImage
	"Byte-swap all words in memory after reading in the entire image file with bulk read. Contributed by Tim Rowledge."

	"First, byte-swap every word in the image. This fixes objects headers."
	self reverseBytesFrom: self startOfMemory to: endOfMemory.

	"Second, return the bytes of bytes-type objects to their orginal order."
	self byteSwapByteObjects.! !

!Interpreter methodsFor: 'I/O primitive support' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
reverseDisplayFrom: startIndex to: endIndex 
	"Reverse the given range of Display words (at different bit 
	depths, this will reverse different numbers of pixels). Used to 
	give feedback during VM activities such as garbage 
	collection when debugging. It is assumed that the given 
	word range falls entirely within the first line of the Display."
	| displayObj dispBitsPtr w reversed |
	displayObj := self splObj: TheDisplay.
	((self isPointers: displayObj) and: [(self lengthOf: displayObj) >= 4]) ifFalse: [^ nil].
	w := self fetchInteger: 1 ofObject: displayObj.
	dispBitsPtr := self fetchPointer: 0 ofObject: displayObj.
	(self isIntegerObject: dispBitsPtr) ifTrue: [^ nil].
	dispBitsPtr := dispBitsPtr + self baseHeaderSize.
	dispBitsPtr + (startIndex * 4) to: dispBitsPtr + (endIndex * 4) by: 4
		do: [:ptr | 
			reversed := (self long32At: ptr) bitXor: 4294967295.
			self longAt: ptr put: reversed].
	successFlag := true.
	self displayBitsOf: displayObj Left: 0 Top: 0 Right: w Bottom: 1.
	self ioForceDisplayUpdate! !

!Interpreter methodsFor: 'method lookup cache' stamp: 'ikp (auto pragmas 12/08) 3/26/2005 14:24'!
rewriteMethodCacheSel: selector class: class primIndex: localPrimIndex

	"Rewrite the cache entry with the given primitive index and matching function pointer"
	| primPtr |
	<var: #primPtr type: 'void *'>
	<inline: false>
	localPrimIndex = 0
		ifTrue: [primPtr := 0]
		ifFalse: [primPtr := primitiveTable at: localPrimIndex].
	self
		rewriteMethodCacheSel: selector class: class
		primIndex: localPrimIndex primFunction: primPtr! !

!Interpreter methodsFor: 'method lookup cache' stamp: 'nice 7/8/2016 03:11'!
rewriteMethodCacheSel: selector class: class primIndex: localPrimIndex primFunction: localPrimAddress
	"Rewrite an existing entry in the method cache with a new primitive 
	index & function address. Used by primExternalCall to make direct jumps to found external prims"
	| probe hash |
	<inline: false>
	<var: #localPrimAddress type: 'void *'>
	hash := selector bitXor: class.
	0 to: CacheProbeMax - 1 do: [:p | 
			probe := hash >> p bitAnd: MethodCacheMask.
			((methodCache at: probe + MethodCacheSelector) = selector
					and: [(methodCache at: probe + MethodCacheClass) = class])
				ifTrue: [methodCache at: probe + MethodCachePrim put: localPrimIndex.
					methodCache at: probe + MethodCachePrimFunction put: (self cCoerce: localPrimAddress to: #'sqIntptr_t').
					^ nil]]! !

!Interpreter methodsFor: 'process primitive support' stamp: 'dtl 4/16/2005 01:36'!
saveProcessSignalingLowSpace
	"The low space semaphore is about to be signaled. Save the currently active
	process in the special objects array so that the low space handler will be able
	to determine the process that first triggered a low space condition. The low
	space handler (in the image) is expected to nil out the special objects array
	slot when it handles the low space condition."

	| lastSavedProcess sched currentProc |
	lastSavedProcess := self splObj: ProcessSignalingLowSpace.
	(lastSavedProcess == self nilObject) ifTrue:
		[sched := self schedulerPointer.
		currentProc := self fetchPointer: ActiveProcessIndex ofObject: sched.
		self storePointer: ProcessSignalingLowSpace ofObject: specialObjectsOop withValue: currentProc]! !

!Interpreter methodsFor: 'process primitive support' stamp: 'tpr 3/24/2004 20:55'!
schedulerPointer

	^ self fetchPointer: ValueIndex ofObject: (self splObj: SchedulerAssociation)! !

!Interpreter methodsFor: 'send bytecodes' stamp: ''!
secondExtendedSendBytecode
	"This replaces the Blue Book double-extended super-send [134],
	which is subsumed by the new double-extended do-anything [132].
	It offers a 2-byte send of 0-3 args for up to 63 literals, for which 
	the Blue Book opcode set requires a 3-byte instruction."

	| descriptor |
	descriptor := self fetchByte.
	messageSelector := self literal: (descriptor bitAnd: 16r3F).
	argumentCount := descriptor >> 6.
	self normalSend.
! !

!Interpreter methodsFor: 'callback support' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
sendInvokeCallback: thunkPtr Stack: stackPtr Registers: regsPtr Jmpbuf: jmpBufPtr
	"Send the 4 argument callback message invokeCallback:stack:registers:jmpbuf:
	 to Alien class with the supplied args.  The arguments are raw C addresses
	 and are converted to integer objects on the way."
	| where |
	<export: true>
	self pushRemappableOop: (self positive32BitIntegerFor: jmpBufPtr).
	self pushRemappableOop: (self positive32BitIntegerFor: regsPtr).
	self pushRemappableOop: (self positive32BitIntegerFor: stackPtr).
	self pushRemappableOop: (self positive32BitIntegerFor: thunkPtr).
	receiver := self splObj: ClassAlien.
	lkupClass := self fetchClassOfNonImm: receiver.
	messageSelector := self splObj: SelectorInvokeCallback.
	(self lookupInMethodCacheSel: messageSelector class: lkupClass) ifFalse:
	 	[(self lookupMethodNoMNUEtcInClass: lkupClass) ifFalse:
			[^false]].
	primitiveIndex ~= 0 ifTrue:
		[^false].
	self storeContextRegisters: activeContext.
	self justActivateNewMethod.
	where := activeContext + self baseHeaderSize + (ReceiverIndex << self shiftForWord).
	self longAt: where + (1 << self shiftForWord) put: self popRemappableOop.
	self longAt: where + (2 << self shiftForWord) put: self popRemappableOop.
	self longAt: where + (3 << self shiftForWord) put: self popRemappableOop.
	self longAt: where + (4 << self shiftForWord) put: self popRemappableOop.
	self interpret.
	"not reached"
	^true! !

!Interpreter methodsFor: 'send bytecodes' stamp: 'eem (auto pragmas) 7/8/2008 19:50'!
sendLiteralSelectorBytecode
	"Can use any of the first 16 literals for the selector and pass up to 2 arguments."
	<expandCases>
	messageSelector := self literal: (currentBytecode bitAnd: 16rF).
	argumentCount := ((currentBytecode >> 4) bitAnd: 3) - 1.
	self normalSend! !

!Interpreter methodsFor: 'contexts' stamp: 'eem 5/23/2008 14:53'!
sender

	| context closureOrNil |
	context := localHomeContext.
	[(closureOrNil := self fetchPointer: ClosureIndex ofObject: context) ~~ nilObj] whileTrue:
		[context := self fetchPointer: ClosureOuterContextIndex ofObject: closureOrNil].
	^self fetchPointer: SenderIndex ofObject: context! !

!Interpreter methodsFor: 'compiler support' stamp: 'ikp 12/16/1998 01:35'!
setCompilerInitialized: newFlag
	| oldFlag |
	oldFlag := compilerInitialized.
	compilerInitialized := newFlag.
	^oldFlag! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/16/2002 13:44'!
setFullScreenFlag: value
	fullScreenFlag := value! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/16/2002 13:45'!
setInterruptCheckCounter: value
	interruptCheckCounter := value! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/16/2002 13:46'!
setInterruptKeycode: value
	interruptKeycode := value! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/16/2002 13:46'!
setInterruptPending: value
	interruptPending := value! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/16/2002 13:48'!
setNextWakeupTick: value
	nextWakeupTick := value! !

!Interpreter methodsFor: 'plugin primitive support' stamp: 'JMM 4/16/2002 13:47'!
setSavedWindowSize: value
	savedWindowSize := value! !

!Interpreter methodsFor: 'jump bytecodes' stamp: ''!
shortConditionalJump

	self jumplfFalseBy: (currentBytecode bitAnd: 7) + 1.! !

!Interpreter methodsFor: 'jump bytecodes' stamp: 'eem (auto pragmas) 7/8/2008 19:49'!
shortUnconditionalJump
	<expandCases>
	self jump: (currentBytecode bitAnd: 7) + 1.! !

!Interpreter methodsFor: 'I/O primitive support' stamp: 'ar 2/20/2000 22:06'!
showDisplayBits: aForm Left: l Top: t Right: r Bottom: b
	"Repaint the portion of the Smalltalk screen bounded by the affected rectangle. Used to synchronize the screen after a Bitblt to the Smalltalk Display object."
	deferDisplayUpdates ifTrue: [^ nil].
	self displayBitsOf: aForm Left: l Top: t Right: r Bottom: b! !

!Interpreter methodsFor: 'utilities' stamp: ''!
signExtend16: int16
	"Convert a signed 16-bit integer into a signed 32-bit integer value. The integer bit is not added here."

	(int16 bitAnd: 16r8000) = 0
		ifTrue: [ ^ int16 ]
		ifFalse: [ ^ int16 - 16r10000 ].! !

!Interpreter methodsFor: 'process primitive support' stamp: 'tpr 3/24/2004 20:57'!
signalExternalSemaphores
	"Signal all requested semaphores"
	| xArray xSize index sema |
	semaphoresUseBufferA := semaphoresUseBufferA not.
	xArray := self splObj: ExternalObjectsArray.
	xSize := self stSizeOf: xArray.
	semaphoresUseBufferA
		ifTrue: ["use opposite buffer during read"
			1 to: semaphoresToSignalCountB do: [:i | 
					index := semaphoresToSignalB at: i.
					index <= xSize
						ifTrue: [sema := self fetchPointer: index - 1 ofObject: xArray.
							"Note: semaphore indices are 1-based"
							(self fetchClassOf: sema) = (self splObj: ClassSemaphore)
								ifTrue: [self synchronousSignal: sema]]].
			semaphoresToSignalCountB := 0]
		ifFalse: [1 to: semaphoresToSignalCountA do: [:i | 
					index := semaphoresToSignalA at: i.
					index <= xSize
						ifTrue: [sema := self fetchPointer: index - 1 ofObject: xArray.
							"Note: semaphore indices are 1-based"
							(self fetchClassOf: sema) = (self splObj: ClassSemaphore)
								ifTrue: [self synchronousSignal: sema]]].
			semaphoresToSignalCountA := 0]! !

!Interpreter methodsFor: 'process primitive support' stamp: 'tpr 3/22/2004 14:23'!
signalFinalization: weakReferenceOop
	"If it is not there already, record the given semaphore index in the list of semaphores to be signaled at the next convenient moment. Force a real interrupt check as soon as possible."

	self forceInterruptCheck.
	pendingFinalizationSignals := pendingFinalizationSignals + 1.! !

!Interpreter methodsFor: 'process primitive support' stamp: 'tpr 3/22/2004 14:25'!
signalSemaphoreWithIndex: index
	"Record the given semaphore index in the double buffer semaphores array to be signaled at the next convenient moment. Force a real interrupt check as soon as possible."

	index <= 0 ifTrue: [^ nil].  "bad index; ignore it"

	semaphoresUseBufferA
		ifTrue: [semaphoresToSignalCountA < SemaphoresToSignalSize
			ifTrue: [ semaphoresToSignalCountA := semaphoresToSignalCountA + 1.
				semaphoresToSignalA at: semaphoresToSignalCountA put: index]]
		ifFalse: [semaphoresToSignalCountB < SemaphoresToSignalSize
			ifTrue: [ semaphoresToSignalCountB := semaphoresToSignalCountB + 1.
				semaphoresToSignalB at: semaphoresToSignalCountB put: index]].
	self forceInterruptCheck
! !

!Interpreter methodsFor: 'primitive support' stamp: 'tpr (auto pragmas 12/08) 3/23/2004 18:13'!
signed32BitIntegerFor: integerValue
	"Return a full 32 bit integer object for the given integer value"
	| newLargeInteger value largeClass |
	<inline: false>
	(self isIntegerValue: integerValue)
		ifTrue: [^ self integerObjectOf: integerValue].
	integerValue < 0
		ifTrue:[	largeClass := self classLargeNegativeInteger.
				value := 0 - integerValue]
		ifFalse:[	largeClass := self classLargePositiveInteger.
				value := integerValue].
	newLargeInteger := self instantiateClass: largeClass indexableSize: 4.
	self storeByte: 3 ofObject: newLargeInteger withValue: ((value >> 24) bitAnd: 16rFF).
	self storeByte: 2 ofObject: newLargeInteger withValue: ((value >> 16) bitAnd: 16rFF).
	self storeByte: 1 ofObject: newLargeInteger withValue: ((value >> 8) bitAnd: 16rFF).
	self storeByte: 0 ofObject: newLargeInteger withValue: (value bitAnd: 16rFF).
	^ newLargeInteger! !

!Interpreter methodsFor: 'primitive support' stamp: 'eem 7/21/2011 16:42'!
signed32BitValueOf: oop
	"Convert the given object into an integer value.
	The object may be either a positive ST integer or a four-byte LargeInteger."
	| sz value largeClass negative |
	<inline: false>
	<returnTypeC: #int>
	<var: #value type: #int>
	(self isIntegerObject: oop) ifTrue: [^self integerValueOf: oop].
	largeClass := self fetchClassOf: oop.
	largeClass = self classLargePositiveInteger
		ifTrue:[negative := false]
		ifFalse:[largeClass = self classLargeNegativeInteger
					ifTrue:[negative := true]
					ifFalse:[^self primitiveFail]].
	sz := self lengthOf: oop.
	sz = 4 ifFalse: [^ self primitiveFail].
	value := (self fetchByte: 0 ofObject: oop) +
			  ((self fetchByte: 1 ofObject: oop) <<  8) +
			  ((self fetchByte: 2 ofObject: oop) << 16) +
			  ((self fetchByte: 3 ofObject: oop) << 24).
	self cCode: []
		inSmalltalk:
			[(value anyMask: 16r80000000) ifTrue:
				[value := value - 16r100000000]].
	"Filter out values out of range for the signed interpretation such as
	 16rFFFFFFFF (positive w/ bit 32 set) and -16rFFFFFFFF (negative w/ bit
	 32 set). Since the sign is implicit in the class we require that the high
	 bit of the magnitude is not set which is a simple test here.  Note that
	 we have to handle the most negative 32-bit value -2147483648 specially."
	value < 0 ifTrue:
		[self assert: (self sizeof: value) == 4.
		 "Don't fail for -16r80000000/-2147483648
		  Alas the simple (negative and: [value - 1 > 0]) isn't adequate since in C the result of signed integer
		  overflow is undefined and hence under optimization this may fail.  The shift, however, is well-defined."
		 (negative and: [0 = (self cCode: [value << 1]
									inSmalltalk: [value << 1 bitAnd: (1 << 32) - 1])]) ifTrue: 
			[^value].
		 ^self primitiveFail].
	^negative
		ifTrue: [0 - value]
		ifFalse: [value]! !

!Interpreter methodsFor: 'primitive support' stamp: 'eem 12/18/2012 18:46'!
signed64BitIntegerFor: integerValue
	"Return a Large Integer object for the given integer value"
	| newLargeInteger magnitude largeClass intValue highWord sz |
	<inline: false>
	<var: 'integerValue' type: #sqLong>
	<var: 'magnitude' type: #sqLong>
	<var: 'highWord' type: #usqInt>

	integerValue < 0
		ifTrue: [largeClass := self classLargeNegativeInteger.
				magnitude := 0 - integerValue]
		ifFalse: [largeClass := self classLargePositiveInteger.
				magnitude := integerValue].

	"Make sure to handle the most -ve value correctly.
	 0 - most -ve = most -ve and most -ve - 1 is +ve"
	(magnitude <= 16r7FFFFFFF
	 and: [integerValue >= 0 or: [integerValue - 1 < 0]]) ifTrue:
		[^self signed32BitIntegerFor: integerValue].

	highWord := magnitude >> 32.
	highWord = 0 
		ifTrue:[sz := 4] 
		ifFalse:
			[sz := 5.
			(highWord := highWord >> 8) = 0 ifFalse:[sz := sz + 1].
			(highWord := highWord >> 8) = 0 ifFalse:[sz := sz + 1].
			(highWord := highWord >> 8) = 0 ifFalse:[sz := sz + 1]].
	newLargeInteger := self instantiateClass: largeClass indexableSize:  sz.
	0 to: sz-1 do: [:i |
		intValue := (magnitude >> (i * 8)) bitAnd: 255.
		self storeByte: i ofObject: newLargeInteger withValue: intValue].
	^newLargeInteger! !

!Interpreter methodsFor: 'primitive support' stamp: 'eem 9/4/2013 17:17'!
signed64BitValueOf: oop
	"Convert the given object into an integer value.
	The object may be either a positive ST integer or a eight-byte LargeInteger."
	| sz value largeClass negative szsqLong |
	<inline: false>
	<returnTypeC: #sqLong>
	<var: #value type: #sqLong>
	(self isIntegerObject: oop) ifTrue: [^self cCoerce: (self integerValueOf: oop) to: #sqLong].
	largeClass := self fetchClassOfNonImm: oop.
	largeClass = self classLargePositiveInteger
		ifTrue:[negative := false]
		ifFalse:[largeClass = self classLargeNegativeInteger
					ifTrue:[negative := true]
					ifFalse:[^self primitiveFail]].
	szsqLong := self sizeof: #sqLong.
	sz := self lengthOf: oop.
	sz > szsqLong 
		ifTrue: [^ self primitiveFail].
	value := 0.
	0 to: sz - 1 do: [:i |
		value := value + ((self cCoerce: (self fetchByte: i ofObject: oop) to: #sqLong) <<  (i*8))].
	"Filter out values out of range for the signed interpretation such as
	16rFFFFFFFF... (positive w/ bit 64 set) and -16rFFFFFFFF... (negative w/ bit
	64 set). Since the sign is implicit in the class we require that the high bit of
	the magnitude is not set which is a simple test here.  Note that we have to
	handle the most negative 64-bit value -9223372036854775808 specially."
	self cCode: []
		inSmalltalk:
			[(value anyMask: 16r8000000000000000) ifTrue:
				[value := value - 16r10000000000000000]].
	value < 0 ifTrue:
		[self cCode:
			[self assert: (self sizeof: value) == 8.
			 self assert: (self sizeof: value << 1) == 8].
		"Don't fail for -9223372036854775808/-16r8000000000000000.
		 Alas the simple (negative and: [value - 1 > 0]) isn't adequate since in C the result of signed integer
		  overflow is undefined and hence under optimization this may fail.  The shift, however, is well-defined."
		 (negative and: [0 = (self cCode: [value << 1]
									inSmalltalk: [value << 1 bitAnd: (1 << 64) - 1])]) ifTrue: 
			[^value].
		 ^self primitiveFail].
	^negative
		ifTrue:[0 - value]
		ifFalse:[value]! !

!Interpreter methodsFor: 'send bytecodes' stamp: ''!
singleExtendedSendBytecode
	"Can use any of the first 32 literals for the selector and pass up to 7 arguments."

	| descriptor |
	descriptor := self fetchByte.
	messageSelector := self literal: (descriptor bitAnd: 16r1F).
	argumentCount := descriptor >> 5.
	self normalSend.! !

!Interpreter methodsFor: 'send bytecodes' stamp: ''!
singleExtendedSuperBytecode
	"Can use any of the first 32 literals for the selector and pass up to 7 arguments."

	| descriptor |
	descriptor := self fetchByte.
	messageSelector := self literal: (descriptor bitAnd: 16r1F).
	argumentCount := descriptor >> 5.
	self superclassSend.
! !

!Interpreter methodsFor: 'utilities' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
sizeOfSTArrayFromCPrimitive: cPtr
	"Return the number of indexable fields of the given object. This method is to be called from an automatically generated C primitive. The argument is assumed to be a pointer to the first indexable field of a words or bytes object; the object header starts 4 bytes before that."
	"Note: Only called by translated primitive code."

	| oop |
	<var: #cPtr type: 'void *'>
	oop := (self oopForPointer: cPtr) - self baseHeaderSize.
	(self isWordsOrBytes: oop) ifFalse: [
		self primitiveFail.
		^0].
	^self lengthOf: oop
! !

!Interpreter methodsFor: 'image save/restore' stamp: 'ikp (auto pragmas 12/08) 8/3/2004 18:24'!
snapshot: embedded 
	"update state of active context"
	| activeProc dataSize rcvr setMacType |
	<var: #setMacType type: 'void *'>
	compilerInitialized
		ifTrue: [self compilerPreSnapshot]
		ifFalse: [self storeContextRegisters: activeContext].

	"update state of active process"
	activeProc := self fetchPointer: ActiveProcessIndex ofObject: self schedulerPointer.
	self
		storePointer: SuspendedContextIndex
		ofObject: activeProc
		withValue: activeContext.

	"compact memory and compute the size of the memory actually in use"
	self incrementalGC.

	"maximimize space for forwarding table"
	self fullGC.
	self snapshotCleanUp.

	dataSize := freeBlock - self startOfMemory. "Assume all objects are below the start of the free block"
	successFlag
		ifTrue: [rcvr := self popStack.
			"pop rcvr"
			self push: trueObj.
			self writeImageFile: dataSize.
			embedded
				ifFalse: ["set Mac file type and creator; this is a noop on other platforms"
					setMacType := self ioLoadFunction: 'setMacFileTypeAndCreator' From: 'FilePlugin'.
					setMacType = 0
						ifFalse: [self cCode: '((sqInt (*)(char *, char *, char *))setMacType)(imageName, "STim", "FAST")']].
			self pop: 1].

	"activeContext was unmarked in #snapshotCleanUp, mark it old "
	self beRootIfOld: activeContext.
	successFlag
		ifTrue: [self push: falseObj]
		ifFalse: [self push: rcvr].
	compilerInitialized
		ifTrue: [self compilerPostSnapshot]! !

!Interpreter methodsFor: 'image save/restore' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
snapshotCleanUp
	"Clean up right before saving an image, sweeping memory and:
	* nilling out all fields of contexts above the stack pointer. 
	* flushing external primitives 
	* clearing the root bit of any object in the root table "
	| oop header fmt sz |
	oop := self firstObject.
	[self oop: oop isLessThan: endOfMemory]
		whileTrue: [(self isFreeObject: oop)
				ifFalse: [header := self longAt: oop.
					fmt := header >> 8 bitAnd: 15.
					"Clean out context"
					(fmt = 3 and: [self isContextHeader: header])
						ifTrue: [sz := self sizeBitsOf: oop.
							(self lastPointerOf: oop) + self wordSize
								to: sz - self baseHeaderSize by: self wordSize
								do: [:i | self longAt: oop + i put: nilObj]].
					"Clean out external functions"
					fmt >= 12
						ifTrue: ["This is a compiled method"
							(self primitiveIndexOf: oop) = PrimitiveExternalCallIndex
								ifTrue: ["It's primitiveExternalCall"
									self flushExternalPrimitiveOf: oop]]].
			oop := self objectAfter: oop].
	self clearRootsTable! !

!Interpreter methodsFor: 'message sending' stamp: ''!
specialSelector: index

	^ self fetchPointer: (index * 2) ofObject: (self splObj: SpecialSelectors)! !

!Interpreter methodsFor: 'array primitive support' stamp: 'dtl (auto pragmas 12/08) 4/22/2007 23:56'!
stObject: array at: index
	"Return what ST would return for <obj> at: index."

	| hdr fmt totalLength fixedFields stSize |
	<inline: false>
	hdr := self baseHeader: array.
	fmt := (hdr >> 8) bitAnd: 16rF.
	totalLength := self lengthOf: array baseHeader: hdr format: fmt.
	fixedFields := self fixedFieldsOf: array format: fmt length: totalLength.
	(fmt = 3 and: [self isContextHeader: hdr])
		ifTrue: [stSize := self fetchStackPointerOf: array]
		ifFalse: [stSize := totalLength - fixedFields].
	((self oop: index isGreaterThanOrEqualTo: 1)
			and: [self oop: index isLessThanOrEqualTo: stSize])
		ifTrue: [^ self subscript: array with: (index + fixedFields) format: fmt]
		ifFalse: [successFlag := false.  ^ 0].! !

!Interpreter methodsFor: 'array primitive support' stamp: 'eem 12/7/2012 11:25'!
stObject: array at: index put: value
	"Do what ST would return for <obj> at: index put: value."
	| hdr fmt totalLength fixedFields stSize |
	<inline: false>
	hdr := self baseHeader: array.
	fmt := (hdr >> 8) bitAnd: 16rF.
	totalLength := self lengthOf: array baseHeader: hdr format: fmt.
	fixedFields := self fixedFieldsOf: array format: fmt length: totalLength.
	(fmt = 3 and: [self isContextHeader: hdr])
		ifTrue: [stSize := self fetchStackPointerOf: array]
		ifFalse: [stSize := totalLength - fixedFields].
	((self oop: index isGreaterThanOrEqualTo: 1)
			and: [self oop: index isLessThanOrEqualTo: stSize])
		ifTrue: [self subscript: array with: (index + fixedFields) storing: value format: fmt]
		ifFalse: [successFlag := false].
	^value! !

!Interpreter methodsFor: 'array primitive support' stamp: 'tpr (auto pragmas 12/08) 3/7/2003 19:52'!
stSizeOf: oop
	"Return the number of indexable fields in the given object. (i.e., what Smalltalk would return for <obj> size)."
	"Note: Assume oop is not a SmallInteger!!"

	| hdr fmt totalLength fixedFields |
	<inline: false>
	hdr := self baseHeader: oop.
	fmt := (hdr >> 8) bitAnd: 16rF.
	totalLength := self lengthOf: oop baseHeader: hdr format: fmt.
	fixedFields := self fixedFieldsOf: oop format: fmt length: totalLength.
	(fmt = 3 and: [self isContextHeader: hdr])
		ifTrue: [^ self fetchStackPointerOf: oop]
		ifFalse: [^ totalLength - fixedFields]! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
stackFloatValue: offset
	<returnTypeC: #double>
	^self floatValueOf: (self longAt: stackPointer - (offset*self wordSize))! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
stackIntegerValue: offset
	| integerPointer |
	integerPointer := self longAt: stackPointer - (offset*self wordSize).
	^self checkedIntegerValueOf: integerPointer! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
stackObjectValue: offset
	"Ensures that the given object is a real object, not a SmallInteger."

	| oop |
	oop := self longAt: stackPointer - (offset * self wordSize).
	(self isIntegerObject: oop) ifTrue: [self primitiveFail. ^ nil].
	^ oop
! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
stackPointerIndex
	"Return the 0-based index rel to the current context.
	(This is what stackPointer used to be before conversion to pointer"
	^ (stackPointer - activeContext - self baseHeaderSize) >> self shiftForWord! !

!Interpreter methodsFor: 'contexts' stamp: ''!
stackTop
	^self longAt: stackPointer! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
stackValue: offset
	^ self longAt: stackPointer - (offset*self wordSize)! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'eem (auto pragmas) 7/8/2008 19:50'!
storeAndPopReceiverVariableBytecode
	"Note: This code uses 
	storePointerUnchecked:ofObject:withValue: and does the 
	store check explicitely in order to help the translator 
	produce better code."
	| rcvr top |
	<expandCases>
	self fetchNextBytecode.
	"this bytecode will be expanded so that refs to 
	currentBytecode below will be constant"
	rcvr := receiver.
	top := self internalStackTop.
	(self oop: rcvr isLessThan: youngStart)
		ifTrue: [self possibleRootStoreInto: rcvr value: top].
	self storePointerUnchecked: (currentBytecode bitAnd: 7) ofObject: rcvr withValue: top.
	self internalPop: 1! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'eem 5/24/2008 11:01'!
storeAndPopRemoteTempLongBytecode
	self storeRemoteTempLongBytecode.
	self internalPop: 1! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'eem 11/13/2012 10:29'!
storeAndPopTemporaryVariableBytecode
	<expandCases>
	self
		cCode: "this bytecode will be expanded so that refs to currentBytecode below will be constant"
			[self fetchNextBytecode.
			 self storePointerUnchecked: (currentBytecode bitAnd: 7) + TempFrameStart
				ofObject: localHomeContext
				withValue: self internalStackTop.
			self internalPop: 1]
		inSmalltalk: "Interpreter version has fetchNextBytecode out of order"
			[self storePointerUnchecked: (currentBytecode bitAnd: 7) + TempFrameStart
				ofObject: localHomeContext
				withValue: self internalStackTop.
			self internalPop: 1.
			self fetchNextBytecode]! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
storeContextRegisters: activeCntx
	"Note: internalStoreContextRegisters: should track changes to this method."

	"InstructionPointer is a pointer variable equal to
	method oop + ip + BaseHeaderSize
		-1 for 0-based addressing of fetchByte
		-1 because it gets incremented BEFORE fetching currentByte"

	<inline: true>
	self storePointerUnchecked: InstructionPointerIndex ofObject: activeCntx
		withValue: (self integerObjectOf: (instructionPointer - method - (self baseHeaderSize - 2))).
	self storePointerUnchecked: StackPointerIndex ofObject: activeCntx
		withValue: (self integerObjectOf: (self stackPointerIndex - TempFrameStart + 1)).
! !

!Interpreter methodsFor: 'utilities' stamp: 'tpr 5/12/2005 15:53'!
storeInteger: fieldIndex ofObject: objectPointer withValue: integerValue 
	"Note: May be called by translated primitive code."
	(self isIntegerValue: integerValue)
		ifTrue: [self storePointerUnchecked: fieldIndex ofObject: objectPointer
					withValue: (self integerObjectOf: integerValue)]
		ifFalse: [self primitiveFail]! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'eem 5/24/2008 11:00'!
storeRemoteTemp: index inVectorAt: tempVectorIndex
	| tempVector |
	tempVector := self temporary: tempVectorIndex.
	self storePointer: index ofObject: tempVector withValue: self internalStackTop.! !

!Interpreter methodsFor: 'stack bytecodes' stamp: 'eem 5/27/2008 13:29'!
storeRemoteTempLongBytecode
	| remoteTempIndex tempVectorIndex |
	remoteTempIndex := self fetchByte.
	tempVectorIndex := self fetchByte.
	self fetchNextBytecode.
	self storeRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!Interpreter methodsFor: 'contexts' stamp: 'tpr 5/13/2005 10:44'!
storeStackPointerValue: value inContext: contextPointer
	"Assume: value is an integerValue"

	self storePointerUnchecked: StackPointerIndex ofObject: contextPointer
		withValue: (self integerObjectOf: value).! !

!Interpreter methodsFor: 'array primitive support' stamp: 'di (auto pragmas 12/08) 7/4/2004 08:43'!
subscript: array with: index format: fmt
	"Note: This method assumes that the index is within bounds!!"

	<inline: true>
	fmt <= 4 ifTrue: [  "pointer type objects"
		^ self fetchPointer: index - 1 ofObject: array].
	fmt < 8 ifTrue: [  "long-word type objects"
		^ self positive32BitIntegerFor:
			(self fetchLong32: index - 1 ofObject: array)
	] ifFalse: [  "byte-type objects"
		^ self integerObjectOf:
			(self fetchByte: index - 1 ofObject: array)
	].! !

!Interpreter methodsFor: 'array primitive support' stamp: 'di (auto pragmas 12/08) 7/4/2004 09:00'!
subscript: array with: index storing: oopToStore format: fmt 
	"Note: This method assumes that the index is within bounds!!"
	| valueToStore |
	<inline: true>
	fmt <= 4
		ifTrue: ["pointer type objects"
			self storePointer: index - 1 ofObject: array
				withValue: oopToStore]
		ifFalse: [fmt < 8
				ifTrue: ["long-word type objects"
					valueToStore := self positive32BitValueOf: oopToStore.
					successFlag
						ifTrue: [self storeLong32: index - 1 ofObject: array
									withValue: valueToStore]]
				ifFalse: ["byte-type objects"
					(self isIntegerObject: oopToStore)
						ifFalse: [successFlag := false].
					valueToStore := self integerValueOf: oopToStore.
					(valueToStore >= 0
							and: [valueToStore <= 255])
						ifFalse: [successFlag := false].
					successFlag
						ifTrue: [self
								storeByte: index - 1
								ofObject: array
								withValue: valueToStore]]]! !

!Interpreter methodsFor: 'primitive support' stamp: ''!
success: successValue

	successFlag := successValue & successFlag.! !

!Interpreter methodsFor: 'message sending' stamp: ''!
superclassOf: classPointer

	^ self fetchPointer: SuperclassIndex ofObject: classPointer! !

!Interpreter methodsFor: 'message sending' stamp: 'eem 11/12/2012 15:22'!
superclassSend
	"Send a message to self, starting lookup with the superclass of the class containing the currently executing method."
	"Assume: messageSelector and argumentCount have been set, and that the receiver and arguments have been pushed onto the stack,"
	"Note: This method is inlined into the interpreter dispatch loop."
	<sharedCodeNamed: 'commonSupersend' inCase: #singleExtendedSuperBytecode>
	| rcvr |
	lkupClass := self superclassOf: (self methodClassOf: method).
	rcvr := self internalStackValue: argumentCount.
	receiverClass := self fetchClassOf: rcvr.
	self commonSend! !

!Interpreter methodsFor: 'debug support' stamp: 'eem 11/13/2012 16:26'!
symbolicMethod: aMethod
	<doNotGenerate>
	| ts prim |
	(ts := self transcript) ensureCr.
	(prim := self primitiveIndexOf: aMethod) > 0 ifTrue:
		[ts nextPutAll: '<primitive: '; print: prim; nextPut: $>.
		(self isQuickPrimitiveIndex: prim) ifTrue:
			[ts nextPutAll: ' quick method'; cr; flush.
			 ^self].
		ts cr].
	(InstructionPrinter
			on: (VMCompiledMethodProxy new
					for: method
					coInterpreter: self
					objectMemory: self))
		indent: 0;
		printInstructionsOn: ts.
	ts flush! !

!Interpreter methodsFor: 'process primitive support' stamp: 'tpr (auto pragmas 12/08) 3/24/2004 20:57'!
synchronousSignal: aSemaphore 
	"Signal the given semaphore from within the interpreter."
	| excessSignals |
	<inline: false>
	(self isEmptyList: aSemaphore)
		ifTrue: ["no process is waiting on this semaphore"
			excessSignals := self fetchInteger: ExcessSignalsIndex ofObject: aSemaphore.
			self storeInteger: ExcessSignalsIndex ofObject: aSemaphore withValue: excessSignals + 1]
		ifFalse: [self resume: (self removeFirstLinkOfList: aSemaphore)]! !

!Interpreter methodsFor: 'compiled methods' stamp: 'eem 08/08/2014 10:47'!
tempCountOf: methodPointer
	^((self methodHeaderOf: methodPointer) >> 19) bitAnd: 16r3F! !

!Interpreter methodsFor: 'contexts' stamp: 'jm 12/7/1998 07:42'!
temporary: offset

	^ self fetchPointer: offset + TempFrameStart ofObject: localHomeContext! !

!Interpreter methodsFor: 'utilities' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
transfer: count fromIndex: firstFrom ofObject: fromOop toIndex: firstTo ofObject: toOop
	"Transfer the specified fullword fields, as from calling context to called context"
	
	"Assume: beRootIfOld: will be called on toOop."
	| fromIndex toIndex lastFrom |
	<inline: true>
	self flag: #Dan.  "Need to check all senders before converting this for 64 bits"
	fromIndex := fromOop + (firstFrom * self wordSize).
	toIndex := toOop + (firstTo * self wordSize).
	lastFrom := fromIndex + (count * self wordSize).
	[self oop: fromIndex isLessThan: lastFrom]
		whileTrue: [fromIndex := fromIndex + self wordSize.
			toIndex := toIndex + self wordSize.
			self
				longAt: toIndex
				put: (self longAt: fromIndex)]! !

!Interpreter methodsFor: 'process primitive support' stamp: 'eem 8/17/2009 16:38'!
transferTo: aProc 
	"Record a process to be awoken on the next interpreter cycle. 
	ikp 11/24/1999 06:07 -- added hook for external runtime compiler.
	eem 8/17/2009 16:37 -- stored nil into myList of new activeProcess"
	| sched oldProc newProc |
	statProcessSwitch := statProcessSwitch + 1.
	newProc := aProc.
	sched := self schedulerPointer.
	oldProc := self fetchPointer: ActiveProcessIndex ofObject: sched.
	self storePointer: ActiveProcessIndex ofObject: sched withValue: newProc.
	self storePointerUnchecked: MyListIndex ofObject: newProc withValue: nilObj.
	compilerInitialized
		ifTrue: [self compilerProcessChange: oldProc to: newProc]
		ifFalse: [self storePointer: SuspendedContextIndex ofObject: oldProc withValue: activeContext.
			self newActiveContext: (self fetchPointer: SuspendedContextIndex ofObject: newProc).
			self storePointer: SuspendedContextIndex ofObject: newProc withValue: nilObj].
	reclaimableContextCount := 0! !

!Interpreter methodsFor: 'contexts' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
unPop: nItems
	stackPointer := stackPointer + (nItems*self wordSize)! !

!Interpreter methodsFor: 'interpreter shell' stamp: ''!
unknownBytecode
	"This should never get called; it means that an unimplemented bytecode appears in a CompiledMethod."

	self error: 'Unknown bytecode'.! !

!Interpreter methodsFor: 'process primitive support' stamp: 'eem 07/28/2014 07:54'!
wakeHighestPriority
	"Return the highest priority process that is ready to run."
	"Note: It is a fatal VM error if there is no runnable process."
	| schedLists p processList |
	schedLists := self fetchPointer: ProcessListsIndex ofObject: self schedulerPointer.
	p := self numSlotsOf: schedLists.
	p := p - 1.
	"index of last indexable field"
	processList := self fetchPointer: p ofObject: schedLists.
	[self isEmptyList: processList]
		whileTrue: [p := p - 1.
			p < 0 ifTrue: [self error: 'scheduler could not find a runnable process'].
			processList := self fetchPointer: p ofObject: schedLists].
	^ self removeFirstLinkOfList: processList! !

!Interpreter methodsFor: 'image save/restore' stamp: 'nice 4/3/2016 00:32'!
wordSwapped: w
	"Return the given 64-bit integer with its halves in the reverse order."

	self wordSize = 8 ifFalse: [self error: 'This cannot happen.'].
	^   ((w >> 32) bitAnd: 16r00000000FFFFFFFF) bitOr:
	     ((w << 32) bitAnd: 16rFFFFFFFF00000000)
! !

!Interpreter methodsFor: 'image save/restore' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:34'!
writeImageFile: imageBytes

	| fn |
	<var: #fn type: 'void *'>
	self writeImageFileIO: imageBytes.
	"set Mac file type and creator; this is a noop on other platforms"
	fn := self ioLoadFunction: 'setMacFileTypeAndCreator' From: 'FilePlugin'.
	fn = 0 ifFalse:[
		self cCode:'((sqInt (*)(char*, char*, char*))fn)(imageName, "STim", "FAST")'.
	].
! !

!Interpreter methodsFor: 'image save/restore' stamp: 'tpr (auto pragmas 12/08) 12/29/2005 16:33'!
writeImageFileIO: imageBytes

	| headerStart headerSize f bytesWritten sCWIfn okToWrite |
	<var: #f type: 'sqImageFile'>
	<var: #headerStart type: 'squeakFileOffsetType '>
	<var: #sCWIfn type: 'void *'>

	"If the security plugin can be loaded, use it to check for write permission.
	If not, assume it's ok"
	sCWIfn := self ioLoadFunction: 'secCanWriteImage' From: 'SecurityPlugin'.
	sCWIfn ~= 0 ifTrue:[okToWrite := self cCode: '((sqInt (*)(void))sCWIfn)()'.
		okToWrite ifFalse:[^self primitiveFail]].
	
	"local constants"
	headerStart := 0.  
	headerSize := 64.  "header size in bytes; do not change!!"

	f := self cCode: 'sqImageFileOpen(imageName, "wb")'.
	f = nil ifTrue: [
		"could not open the image file for writing"
		self success: false.
		^ nil].

	headerStart := self cCode: 'sqImageFileStartLocation(f,imageName,headerSize+imageBytes)'.
	self cCode: '/* Note: on Unix systems one could put an exec command here, padded to 512 bytes */'.
	"position file to start of header"
	self sqImageFile: f Seek: headerStart.

	self putLong: (self imageFormatVersion) toFile: f.
	self putLong: headerSize toFile: f.
	self putLong: imageBytes toFile: f.
	self putLong: (self startOfMemory) toFile: f.
	self putLong: specialObjectsOop toFile: f.
	self putLong: lastHash toFile: f.
	self putLong: (self ioScreenSize) toFile: f.
	self putLong: fullScreenFlag toFile: f.
	self putLong: extraVMMemory toFile: f.
	1 to: 7 do: [:i | self putLong: 0 toFile: f].  "fill remaining header words with zeros"
	successFlag ifFalse: [
		"file write or seek failure"
		self cCode: 'sqImageFileClose(f)'.
		^ nil].

	"position file after the header"
	self sqImageFile: f Seek: headerStart + headerSize.

	"write the image data"
	bytesWritten := self cCode: 'sqImageFileWrite(pointerForOop(memory), sizeof(unsigned char), imageBytes, f)'.
	self success: bytesWritten = imageBytes.
	self cCode: 'sqImageFileClose(f)'.

! !

Interpreter subclass: #InterpreterSimulator
	instanceVariableNames: 'byteCount sendCount breakSelector traceOn myBitBlt displayForm imageName pluginList mappedPluginEntries quitBlock transcript displayView printSends'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!InterpreterSimulator commentStamp: 'tpr 5/5/2003 12:24' prior: 0!
This class defines basic memory access and primitive simulation so that the Interpreter can run simulated in the Squeak environment.  It also defines a number of handy object viewing methods to facilitate pawing around in the object memory.

To see the thing actually run, you could (after backing up this image and changes), execute

	(InterpreterSimulator new openOn: Smalltalk imageName) test

and be patient both to wait for things to happen, and to accept various things that may go wrong depending on how large or unusual your image may be.  We usually do this with a small and simple benchmark image. You will probably have more luck using InterpreteSimulatorLSB or InterpreterSimulatorMSB as befits your machine.!

!InterpreterSimulator class methodsFor: 'instance creation' stamp: 'nk 4/5/2005 20:22'!
new
	^ self == InterpreterSimulator
		ifTrue: [SmalltalkImage current endianness == #big
				ifTrue: [InterpreterSimulatorMSB new]
				ifFalse: [InterpreterSimulatorLSB new]]
		ifFalse: [super new]! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'di 7/20/2004 12:11'!
allObjectsSelect: objBlock
	"self allObjectsSelect: [:oop | (self baseHeader: oop) = 1234]"

	| oop selected |
	oop := self firstObject.
	selected := OrderedCollection new.
	[oop < endOfMemory] whileTrue:
			[(self isFreeObject: oop)
				ifFalse: [(objBlock value: oop) ifTrue: [selected addLast: oop]].
			oop := self objectAfter: oop].
	^ selected! !

!InterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 1/23/2009 17:43'!
allocate: byteSize headerSize: hdrSize h1: baseHeader h2: classOop h3: extendedSize doFill: doFill format: format

	| newObj |
	newObj := super allocate: byteSize headerSize: hdrSize h1: baseHeader h2: classOop h3: extendedSize doFill: doFill format: format.
	"byteCount < 600000 ifTrue: [^ newObj]."
	"(self baseHeader: newObj) =  16r0FCC0600 ifTrue: [self halt]."
	^ newObj! !

!InterpreterSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:07'!
browserPluginInitialiseIfNeeded
"do nothing - its a macro in C code to support Mac browser plugin strangeness"! !

!InterpreterSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:03'!
browserPluginReturnIfNeeded
"do nothing - its a macro in C code to support Mac browser plugin strangeness"! !

!InterpreterSimulator methodsFor: 'memory access' stamp: 'th 4/15/2000 17:21'!
byteAt: byteAddress
	^self subclassResponsibility! !

!InterpreterSimulator methodsFor: 'memory access' stamp: 'th 4/15/2000 17:21'!
byteAt: byteAddress put: byte
	^self subclassResponsibility! !

!InterpreterSimulator methodsFor: 'memory access' stamp: 'di 6/23/2004 14:09'!
byteAtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	pointer is a raw address, and byte is an 8-bit quantity."

	^ self byteAt: pointer! !

!InterpreterSimulator methodsFor: 'memory access' stamp: 'di 6/23/2004 14:13'!
byteAtPointer: pointer put: byteValue
	"This gets implemented by Macros in C, where its types will also be checked.
	pointer is a raw address, and byteValue is an 8-bit quantity."

	^ self byteAt: pointer  put: byteValue! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'di 6/21/2004 15:05'!
byteCount
	"So you can call this from temp debug statements in, eg, Interpreter, such as
	self byteCount = 12661 ifTrue: [self halt].
	"

	^ byteCount! !

!InterpreterSimulator methodsFor: 'UI' stamp: 'di 4/21/2004 00:09'!
byteCountText
	^ byteCount printString asText! !

!InterpreterSimulator methodsFor: 'plugin support' stamp: 'ar 5/11/2000 22:01'!
callExternalPrimitive: mapIndex
	| entry |
	entry := (mappedPluginEntries at: mapIndex).
	^(entry at: 1) perform: (entry at: 2).! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'th 4/15/2000 17:22'!
charsOfLong: long
	^self subclassResponsibility! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'eem 10/22/2008 13:51'!
checkForInterrupts
	"Prevent interrupts so that traces are consistent during detailed debugging"

	"self halt."
	true ifTrue:
		[interruptCheckCounter := 1000.
		^self].
	^ super checkForInterrupts! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'eem 07/28/2014 07:54'!
classAndSelectorOfMethod: meth forReceiver: rcvr
	| mClass dict length methodArray |
	mClass := self fetchClassOf: rcvr.
	[dict := self fetchPointer: MethodDictionaryIndex ofObject: mClass.
	length := self numSlotsOf: dict.
	methodArray := self fetchPointer: MethodArrayIndex ofObject: dict.
	0 to: length-SelectorStart-1 do: 
		[:index | 
		meth = (self fetchPointer: index ofObject: methodArray) 
			ifTrue: [^ Array
				with: mClass
				with: (self fetchPointer: index + SelectorStart ofObject: dict)]].
	mClass := self fetchPointer: SuperclassIndex ofObject: mClass.
	mClass = nilObj]
		whileFalse: [].
	^ Array
		with: (self fetchClassOf: rcvr)
		with: (self splObj: SelectorDoesNotUnderstand)! !

!InterpreterSimulator methodsFor: 'plugin support' stamp: 'yo 2/14/2001 11:15'!
classNameOf: aClass Is: className
	"Check if aClass' name is className"
	| name |
	(self lengthOf: aClass) <= 6 ifTrue:[^false]. "Not a class but maybe behavior" 
	name := self fetchPointer: 6 ofObject: aClass.
	(self isBytes: name) ifFalse:[^false].
	^ className = (self stringOf: name).
! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'di 9/23/2001 12:26'!
clipboardRead: sz Into: actualAddress At: zeroBaseIndex
	| str |
	str := Clipboard clipboardText.
	1 to: sz do:
		[:i | self byteAt: actualAddress + zeroBaseIndex + i - 1 put: (str at: i) asciiValue]! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'di 9/23/2001 12:18'!
clipboardSize

	^ Clipboard clipboardText size! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
clipboardWrite: sz From: actualDataAddress At: ignored

	Clipboard clipboardText: (self stringOf: actualDataAddress - self baseHeaderSize)! !

!InterpreterSimulator methodsFor: 'initialization' stamp: 'eem 6/22/2011 12:06'!
close  "close any files that ST may have opened"
	(self loadNewPlugin: 'FilePlugin') ifNotNil:
		[:filePlugin| filePlugin close]! !

!InterpreterSimulator methodsFor: 'debugging traps' stamp: 'eem 7/24/2008 18:13'!
commonSend
	| i |
	printSends ifTrue:
		[self print: byteCount; space; printStringOf: messageSelector; cr].
	(i := breakSelector basicSize) = (self lengthOf: messageSelector) ifTrue:
		[[i > 0] whileTrue:
			[(self fetchByte: i - 1 ofObject: messageSelector) = (breakSelector at: i) asInteger
				ifTrue: [(i := i - 1) = 0 ifTrue: [self halt: 'Send of ', breakSelector]]
				ifFalse: [i := 0]]].
	^super commonSend! !

!InterpreterSimulator methodsFor: 'I/O primitives support' stamp: ''!
copyBits

	^ myBitBlt copyBits! !

!InterpreterSimulator methodsFor: 'debug printing' stamp: 'di 5/11/2004 18:26'!
cr

	traceOn ifTrue: [ transcript cr; endEntry ].! !

!InterpreterSimulator methodsFor: 'interpreter shell' stamp: 'tpr 4/20/2004 12:06'!
dispatchFunctionPointer: selector
"handle the primitive direct dispatch macro in simulation"
	^self perform: selector! !

!InterpreterSimulator methodsFor: 'interpreter shell' stamp: 'tpr 4/2/2004 15:07'!
dispatchOn: anInteger in: selectorArray
	"Simulate a case statement via selector table lookup.
	The given integer must be between 0 and selectorArray size-1, inclusive.
	For speed, no range test is done, since it is done by the at: operation.
	Note that, unlike many other arrays used in the Interpreter, this method expect NO CArrayAccessor wrapping - it would duplicate the +1. Maybe this would be better updated to make it all uniform"

	self perform: (selectorArray at: (anInteger + 1)).! !

!InterpreterSimulator methodsFor: 'I/O primitives support' stamp: ''!
displayLocation

	^ Display extent - displayForm extent - (10@10)! !

!InterpreterSimulator methodsFor: 'I/O primitives support' stamp: ''!
drawLoopX: xDelta Y: yDelta

	^ myBitBlt drawLoopX: xDelta Y: yDelta! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'jm 11/23/1998 22:44'!
dumpHeader: hdr
	| cc |
	^ String streamContents: [:strm |
		cc := (hdr bitAnd: CompactClassMask) >> 12.
		strm nextPutAll: '<cc=', cc hex.
		cc > 0 ifTrue:
			[strm nextPutAll: ':' , (self nameOfClass: (self compactClassAt: cc))].
		strm nextPutAll: '>'.
		strm nextPutAll: '<ft=', ((hdr bitShift: -8) bitAnd: 16rF) hex , '>'.
		strm nextPutAll: '<sz=', (hdr bitAnd: SizeMask) hex , '>'.
		strm nextPutAll: '<hdr=', (#(big class gcMark short) at: (hdr bitAnd: 3) +1) , '>']
! !

!InterpreterSimulator methodsFor: 'debug support' stamp: ''!
dumpMethodHeader: hdr
	^ String streamContents:
		[:strm |
		strm nextPutAll: '<nArgs=', ((hdr >> 25) bitAnd: 16r1F) printString , '>'.
		strm nextPutAll: '<nTemps=', ((hdr >> 19) bitAnd: 16r3F) printString , '>'.
		strm nextPutAll: '<lgCtxt=', ((hdr >> 18) bitAnd: 16r1) printString , '>'.
		strm nextPutAll: '<nLits=', ((hdr >> 10) bitAnd: 16rFF) printString , '>'.
		strm nextPutAll: '<prim=', ((hdr >> 1) bitAnd: 16r1FF) printString , '>'.
		]! !

!InterpreterSimulator methodsFor: 'interpreter shell' stamp: ''!
fetchByte

	^ self byteAt: (localIP := localIP + 1).! !

!InterpreterSimulator methodsFor: 'float primitives' stamp: 'di 6/13/2004 10:31'!
fetchFloatAt: floatBitsAddress into: aFloat

	aFloat at: 1 put: (self long32At: floatBitsAddress).
	aFloat at: 2 put: (self long32At: floatBitsAddress+4).
! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'di 12/11/1998 17:06'!
findNewMethodInClass: class
"
	| cName |
	traceOn ifTrue:
		[cName := (self sizeBitsOf: class) = 16r20
			ifTrue: ['class ' , (self nameOfClass: (self fetchPointer: 6 ofObject: class))]
			ifFalse: [(self nameOfClass: class)].
		self cr; print: cName , '>>' , (self stringOf: messageSelector)].
"

(self stringOf: messageSelector) = 'doesNotUnderstand:' ifTrue: [self halt].

	sendCount := sendCount + 1.

"
	(sendCount > 1000 and: [sendCount\\10 = 0]) ifTrue:
		[Transcript print: sendCount; space.
		self validate].
"
"
	(sendCount > 100150) ifTrue:
		[self qvalidate.
		messageQueue == nil ifTrue: [messageQueue := OrderedCollection new].
		messageQueue addLast: (self stringOf: messageSelector)].
"
	super findNewMethodInClass: class.! !

!InterpreterSimulator methodsFor: 'memory access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
firstIndexableField: oop
	"NOTE: overridden from ObjectMemory to add coercion to CArray, so please duplicate any changes"
	| hdr fmt totalLength fixedFields |
	<returnTypeC: #'void *'>
	hdr := self baseHeader: oop.
	fmt := self formatOfHeader: hdr.
	fmt <= 4 ifTrue: "<= 4 pointer"
		["pointer; may need to delve into the class format word"
		totalLength := self lengthOf: oop baseHeader: hdr format: fmt.
		fixedFields := self fixedFieldsOf: oop format: fmt length: totalLength.
		^self cCoerce: (self pointerForOop: oop + self baseHeaderSize + (fixedFields << self shiftForWord)) to: 'oop *'].
	^self
		cCoerce: (self pointerForOop: oop + self baseHeaderSize)
		to: (fmt < 8
				ifTrue: [fmt = 6
						ifTrue: ["32 bit field objects" 'int *']
						ifFalse: ["full word objects (bits)" 'oop *']]
				ifFalse: ["byte objects (including CompiledMethod" 'char *'])! !

!InterpreterSimulator methodsFor: 'plugin support' stamp: 'ar 5/11/2000 21:59'!
flushExternalPrimitives
	mappedPluginEntries := #().
	super flushExternalPrimitives.! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'di 7/9/2004 11:05'!
fullDisplay
	| t |
	displayForm == nil ifTrue: [^ self].
	t := successFlag.  successFlag := true.
	self displayBitsOf: (self splObj: TheDisplay) Left: 0 Top: 0 Right: displayForm width Bottom: displayForm height.
	successFlag := t! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'di 7/21/2004 15:58'!
fullDisplayUpdate
	"Preserve successFlag when call asynchronously from Simulator"
	| s |
	s := successFlag.
	successFlag := true.
	super fullDisplayUpdate.
	successFlag := s! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'di 5/11/2004 18:26'!
fullGC
	transcript cr; show:'<Running full GC ...'.
	super fullGC.
	transcript show: ' done>'.! !

!InterpreterSimulator methodsFor: 'interpreter shell' stamp: 'ikp 8/2/2004 17:59'!
functionPointerFor: primIndex inClass: lookupClass
	"Override Interpreter to handle the external primitives caching.  See also 	internalExecuteNewMethod."

	^(primIndex between: 1 and: MaxPrimitiveIndex)
		ifTrue: [primitiveTable at: primIndex + 1]! !

!InterpreterSimulator methodsFor: 'memory access' stamp: 'di 7/16/2004 14:56'!
halfWordHighInLong32: long32
	^self subclassResponsibility! !

!InterpreterSimulator methodsFor: 'memory access' stamp: 'di 7/16/2004 14:57'!
halfWordLowInLong32: long32
	^self subclassResponsibility! !

!InterpreterSimulator methodsFor: 'debug support' stamp: ''!
headerStart: oop

	^ (self extraHeaderBytes: oop) negated! !

!InterpreterSimulator methodsFor: 'debug support' stamp: ''!
hexDump100: oop
	| byteSize val |
	^ String streamContents:
		[:strm |
		byteSize := 256.
		(self headerStart: oop) to: byteSize by: 4 do:
			[:a | val := self longAt: oop+a.
			strm cr; nextPutAll: (oop+a) hex8; space; space; 
				nextPutAll: (a<16 ifTrue: [' ', a hex] ifFalse: [a hex]); 
				space; space; space; nextPutAll: val hex8;
				space; space.
			strm nextPutAll: (self charsOfLong: val).
			strm space; space; nextPutAll: (oop+a) printString]]! !

!InterpreterSimulator methodsFor: 'debug support' stamp: ''!
hexDump: oop
	| byteSize val |
	(self isIntegerObject: oop) ifTrue: [^ self shortPrint: oop].
	^ String streamContents:
		[:strm |
		byteSize := 256 min: (self sizeBitsOf: oop)-4.
		(self headerStart: oop) to: byteSize by: 4 do:
			[:a | val := self longAt: oop+a.
			strm cr; nextPutAll: (a<16 ifTrue: [' ', a hex] ifFalse: [a hex]); 
				space; space; space; nextPutAll: val hex8;
				space; space.
			a=0
				ifTrue: [strm nextPutAll: (self dumpHeader: val)]
				ifFalse: [strm nextPutAll: (self charsOfLong: val)]]]! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'eem 10/12/2009 00:57'!
incrementalGC
	transcript cr; nextPutAll: 'incrementalGC ('; print: byteCount; nextPut: $); flush.
	^super incrementalGC! !

!InterpreterSimulator methodsFor: 'initialization' stamp: 'eem 9/26/2013 01:02'!
initialize
	"Initialize the InterpreterSimulator when running the interpreter inside
	Smalltalk. The primary responsibility of this method is to allocate
	Smalltalk Arrays for variables that will be declared as statically-allocated
	global arrays in the translated code."

	"Note: we must initialize ConstMinusOne differently for simulation,
		due to the fact that the simulator works only with +ve 32-bit values"
	ConstMinusOne := self integerObjectOf: -1.

	methodCache := Array new: MethodCacheSize.
	atCache := Array new: AtCacheTotalSize.
	self flushMethodCache.
	rootTable := Array new: RootTableSize.
	weakRoots := Array new: RootTableSize + RemapBufferSize + 100.
	remapBuffer := Array new: RemapBufferSize.
	gcSemaphoreIndex := 0.
	semaphoresUseBufferA := true.
	semaphoresToSignalA := Array new: SemaphoresToSignalSize.
	semaphoresToSignalB := Array new: SemaphoresToSignalSize.
	externalPrimitiveTable := CArrayAccessor on: (Array new: MaxExternalPrimitiveTableSize).
	primitiveTable := self class primitiveTable.
	pluginList := {'' -> self }.
	mappedPluginEntries := #().

	"initialize InterpreterSimulator variables used for debugging"
	byteCount := 0.
	sendCount := 0.
	quitBlock := [^ self].
	traceOn := true.
	printSends := "printReturns := printBytecodeAtEachStep :=" false.
	myBitBlt := BitBltSimulator new setInterpreter: self.
	headerTypeBytes := CArrayAccessor on: HeaderTypeExtraBytes.
	transcript := Transcript.
	displayForm := 'Display has not yet been installed' asDisplayText form.
	! !

!InterpreterSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:01'!
insufficientMemoryAvailableError
	self error: 'Failed to allocate memory for the heap'! !

!InterpreterSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:01'!
insufficientMemorySpecifiedError
	self error: 'Insufficient memory for this image'! !

!InterpreterSimulator methodsFor: 'interpreter shell' stamp: 'tpr 6/23/2004 15:12'!
internalExecuteNewMethod
"Override the Interpreter version to trap cached external prims.
These have a 'prim index' of 1000 + the externalPrimitiveTableIndex normally used"
	primitiveIndex < 1000 ifTrue:[^super internalExecuteNewMethod].
	self externalizeIPandSP.
	self callExternalPrimitive: (externalPrimitiveTable at: primitiveIndex - 1001).
	self internalizeIPandSP! !

!InterpreterSimulator methodsFor: 'interpreter shell' stamp: 'eem 6/10/2010 15:47'!
interpreter
	^self! !

!InterpreterSimulator methodsFor: 'security' stamp: 'ar 2/5/2001 18:33'!
ioCanRenameImage
	^true! !

!InterpreterSimulator methodsFor: 'security' stamp: 'ar 2/5/2001 20:42'!
ioCanWriteImage
	^true! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'ajh 8/21/2002 22:41'!
ioExit

	quitBlock value  "Cause return from #test"! !

!InterpreterSimulator methodsFor: 'other primitives' stamp: 'ikp 12/15/1998 23:33'!
ioForceDisplayUpdate
	"no-op"! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'yo 2/14/2001 14:17'!
ioGetNextEvent: evtBuf

	self primitiveFail.
! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'ar 5/1/1999 09:43'!
ioHasDisplayDepth: depth
	^Display supportsDisplayDepth: depth! !

!InterpreterSimulator methodsFor: 'plugin support' stamp: 'eem 12/12/2012 11:45'!
ioLoadExternalFunction: functionName OfLength: functionLength FromModule: moduleName OfLength: moduleLength
	"Load and return the requested function from a module"
	| pluginString functionString |
	pluginString := String new: moduleLength.
	1 to: moduleLength do:[:i| pluginString byteAt: i put: (self byteAt: moduleName+i-1)].
	functionString := String new: functionLength.
	1 to: functionLength do:[:i| functionString byteAt: i put: (self byteAt: functionName+i-1)].
	^self ioLoadFunction: functionString From: pluginString! !

!InterpreterSimulator methodsFor: 'plugin support' stamp: 'eem 01/19/2021 20:40'!
ioLoadFunction: functionString From: pluginString
	"Load and return the requested function from a module"
	| plugin fnSymbol |
	fnSymbol := functionString asSymbol.
	transcript cr; show:'Looking for ', functionString, ' in '.
	pluginString isEmpty
		ifTrue:[transcript show: 'vm']
		ifFalse:[transcript show: pluginString].
	plugin := pluginList 
				detect:[:any| any key = pluginString asString]
				ifNone:[self loadNewPlugin: pluginString].
	plugin ifNil:[
		"Transcript cr; show:'Failed ... no plugin found'." ^ 0].
	plugin := plugin value.
	mappedPluginEntries withIndexDo:[:pluginAndName :index|
		((pluginAndName at: 1) == plugin 
			and:[(pluginAndName at: 2) == fnSymbol]) ifTrue:[
				"Transcript show:' ... okay'." ^ index]].
	(plugin respondsTo: fnSymbol) ifFalse:[
		"Transcript cr; show:'Failed ... primitive not in plugin'." ^ 0].
	mappedPluginEntries := mappedPluginEntries copyWith: (Array with: plugin with: fnSymbol).
	"Transcript show:' ... okay'."
	^ mappedPluginEntries size! !

!InterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'di 7/1/2004 13:55'!
ioMSecs
	"Return the value of the millisecond clock."
	"NOT.  Actually, we want something a lot slower and, for exact debugging,
	something more repeatable than real time.  IO have an idea: use the byteCount..."

	^ byteCount // 100
	
"At 20k bytecodes per second, this gives us aobut 200 ticks per second, or about 1/5 of what you'd expect for the real time clock.  This should still service events at one or two per second"! !

!InterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'jm 3/9/1999 10:00'!
ioMicroMSecs
	"Answer the value of the high-resolution millisecond clock."

	^ Time millisecondClockValue
! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: ''!
ioProcessEvents! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'ar 2/5/2001 17:24'!
ioScreenDepth
	^DisplayScreen actualScreenDepth.! !

!InterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'topa 7/10/2016 23:14'!
ioScreenScaleFactor
	^ 1.0! !

!InterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'jm 9/24/97 22:52'!
ioScreenSize
	"Return the screen extent packed into 32 bits."

	^ (displayForm width << 16) + displayForm height! !

!InterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'jm 9/24/97 22:52'!
ioSeconds
	"Return the value of the second clock."

	^ Time primSecondsClock! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'yo 2/14/2001 14:17'!
ioSetInputSemaphore: index

	self primitiveFail! !

!InterpreterSimulator methodsFor: 'plugin support' stamp: 'eem 10/26/2018 10:11'!
loadNewPlugin: pluginString
	| plugin plugins realPluginClass |
	transcript cr; show: 'Looking for module ', pluginString.
	"but *why*??"
	plugins := InterpreterPlugin allSubclasses select:
					[:psc|
					 psc moduleName asString = pluginString asString
					 and: [psc shouldBeTranslated]].
	 plugins isEmpty ifTrue: [self transcript show: ' ... not found'. ^nil].
	 plugins size > 1 ifTrue: [^self error: 'This won''t work...'].
	 "plugins size > 1 ifTrue:
		[self transcript show: '...multiple plugin classes; choosing ', plugins last name]."
	 realPluginClass := plugins anyOne. "hopefully lowest in the hierarchy..."
	 plugin := realPluginClass simulatorForInterpreterInterface: self.
	 (plugin respondsTo: #initialiseModule) ifTrue:
		[plugin initialiseModule ifFalse:
			[self transcript show: ' ... initialiser failed'. ^nil]]. "module initialiser failed"
	pluginList := pluginList copyWith: (pluginString asString -> plugin).
	transcript show: ' ... loaded'.
	^pluginList last! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'eem 10/14/2009 15:23'!
logOfBytesVerify: nBytes fromFileNamed: fileName fromStart: loggingStart
	"Verify a questionable interpreter against a successful run"
	"self logOfBytesVerify: 10000 fromFileNamed: 'clone32Bytecodes.log' "
	
	| logFile rightByte prevCtxt |
	logFile := (FileStream readOnlyFileNamed: fileName) binary.
	transcript clear.
	byteCount := 0.
	quitBlock := [^ self].
	self internalizeIPandSP.
	self fetchNextBytecode.
	prevCtxt := 0.  prevCtxt := prevCtxt.
	[byteCount < nBytes] whileTrue:
		[
"
byteCount > 14560 ifTrue:
[self externalizeIPandSP.
prevCtxt = activeContext ifFalse:
 [prevCtxt := activeContext.
 transcript cr; nextPutAll: (self printTop: 2); endEntry].
transcript cr; print: byteCount; nextPutAll: ': ' , (activeContext hex); space;
 print: (instructionPointer - method - (BaseHeaderSize - 2));
 nextPutAll: ': <' , (self byteAt: localIP) hex , '>'; space;
 nextPutAll: (self symbolic: currentBytecode at: localIP inMethod: method); space;
 print: (self stackPointerIndex - TempFrameStart + 1); endEntry.
byteCount = 14590 ifTrue: [self halt]].
"
		loggingStart >= byteCount ifTrue:
			[rightByte := logFile next.
			 currentBytecode = rightByte ifFalse:
				[self halt: 'halt at ', byteCount printString]].
		self dispatchOn: currentBytecode in: BytecodeTable.
		byteCount := byteCount + 1.
		byteCount \\ 10000 = 0 ifTrue: [self fullDisplayUpdate]].
	self externalizeIPandSP.
	logFile close.
	self inform: nBytes printString , ' bytecodes verfied.'! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'eem 10/14/2009 09:24'!
logOfBytesWrite: nBytes toFileNamed: fileName fromStart: loggingStart
	"Write a log file for testing a flaky interpreter on the same image"
	"self logOfBytesWrite: 10000 toFileNamed: 'clone32Bytecodes.log' "
	
	| logFile |
	logFile := (FileStream newFileNamed: fileName) binary.
	transcript clear.
	byteCount := 0.
	quitBlock := [^ self].
	self internalizeIPandSP.
	self fetchNextBytecode.
	[byteCount < nBytes] whileTrue:
		[byteCount >= loggingStart ifTrue:
			[logFile nextPut: currentBytecode].
		self dispatchOn: currentBytecode in: BytecodeTable.
		byteCount := byteCount + 1.
		byteCount \\ 10000 = 0 ifTrue: [self fullDisplayUpdate]].
	self externalizeIPandSP.
	logFile close! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'eem 10/14/2009 15:24'!
logOfSendsVerify: nSends fromFileNamed: fileName fromStart: loggingStart
	"Write a log file for testing a flaky interpreter on the same image"
	"self logOfSendsWrite: 10000 toFileNamed: 'clone32Messages.log' "
	
	| logFile priorContext rightSelector prevCtxt |
	logFile := FileStream readOnlyFileNamed: fileName.
	transcript clear.
	byteCount := 0.
	sendCount := 0.
	priorContext := activeContext.
	quitBlock := [^ self].
	self internalizeIPandSP.
	self fetchNextBytecode.
	prevCtxt := 0.  prevCtxt := prevCtxt.
	[sendCount < nSends] whileTrue:
		[
"
byteCount>500 ifTrue:
[byteCount>550 ifTrue: [self halt].
self externalizeIPandSP.
prevCtxt = activeContext ifFalse:
 [prevCtxt := activeContext.
 transcript cr; nextPutAll: (self printTop: 2); endEntry].
transcript cr; print: byteCount; nextPutAll: ': ' , (activeContext hex); space;
 print: (instructionPointer - method - (BaseHeaderSize - 2));
 nextPutAll: ': <' , (self byteAt: localIP) hex , '>'; space;
 nextPutAll: (self symbolic: currentBytecode at: localIP inMethod: method); space;
 print: (self stackPointerIndex - TempFrameStart + 1); endEntry.
].
"
		self dispatchOn: currentBytecode in: BytecodeTable.
		activeContext == priorContext ifFalse:
			[sendCount := sendCount + 1.
			 loggingStart >= sendCount ifTrue:
				[rightSelector := logFile nextLine.
				 (self stringOf: messageSelector) = rightSelector ifFalse:
					[self halt: 'halt at ', sendCount printString]].
			priorContext := activeContext].
		byteCount := byteCount + 1.
		byteCount \\ 10000 = 0 ifTrue: [self fullDisplayUpdate]].
	self externalizeIPandSP.
	logFile close.
	self inform: nSends printString , ' sends verfied.'! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'eem 10/14/2009 09:23'!
logOfSendsWrite: nSends toFileNamed: fileName fromStart: loggingStart
	"Write a log file for testing a flaky interpreter on the same image"
	"self logOfSendsWrite: 10000 toFileNamed: 'clone32Messages.log' "
	
	| logFile priorContext |
	logFile := FileStream newFileNamed: fileName.
	transcript clear.
	byteCount := 0.
	sendCount := 0.
	priorContext := activeContext.
	quitBlock := [^ self].
	self internalizeIPandSP.
	self fetchNextBytecode.
	[sendCount < nSends] whileTrue:
		[self dispatchOn: currentBytecode in: BytecodeTable.
		activeContext == priorContext ifFalse:
			[sendCount >= loggingStart ifTrue:
				[sendCount := sendCount + 1.
				 logFile nextPutAll: (self stringOf: messageSelector); cr].
			priorContext := activeContext].
		byteCount := byteCount + 1.
		byteCount \\ 10000 = 0 ifTrue: [self fullDisplayUpdate]].
	self externalizeIPandSP.
	logFile close! !

!InterpreterSimulator methodsFor: 'memory access' stamp: 'di 7/3/2004 10:47'!
long32At: byteAddress
	"Return the 32-bit word at byteAddress which must be 0 mod 4."

	^ self longAt: byteAddress! !

!InterpreterSimulator methodsFor: 'memory access' stamp: 'di 7/3/2004 10:47'!
long32At: byteAddress put: a32BitValue
	"Store the 32-bit value at byteAddress which must be 0 mod 4."

	^ self longAt: byteAddress put: a32BitValue! !

!InterpreterSimulator methodsFor: 'memory access' stamp: ''!
longAt: byteAddress
	"Note: Adjusted for Smalltalk's 1-based array indexing."

	^memory at: (byteAddress // 4) + 1! !

!InterpreterSimulator methodsFor: 'memory access' stamp: ''!
longAt: byteAddress put: a32BitValue
	"Note: Adjusted for Smalltalk's 1-based array indexing."

	^memory at: (byteAddress // 4) + 1 put: a32BitValue! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
longPrint: oop
	| lastPtr val lastLong hdrType prevVal |
	(self isIntegerObject: oop) ifTrue: [^ self shortPrint: oop].
	^ String streamContents:
		[:strm |
		lastPtr := 64*self wordSize min: (self lastPointerOf: oop).
		hdrType := self headerType: oop.
		hdrType = 2 ifTrue: [lastPtr := 0].
		prevVal := 0.
		(self headerStart: oop) to: lastPtr by: self wordSize do:
			[:a | val := self longAt: oop+a.
			(a > 0 and: [(val = prevVal) & (a ~= lastPtr)])
			ifTrue:
			[prevVal = (self longAt: oop+a-(self wordSize*2)) ifFalse: [strm cr; nextPutAll: '        ...etc...']]
			ifFalse:
			[strm cr; nextPutAll: (a<16 ifTrue: [' ', a hex] ifFalse: [a hex]); 
				space; space; space; nextPutAll: val hex8; space; space.
			a = (self wordSize*2) negated ifTrue:
				[strm nextPutAll: 'size = ' , (val - hdrType) hex].
			a = self wordSize negated ifTrue:
				[strm nextPutAll: '<' , (self nameOfClass: (val - hdrType)) , '>'].
			a = 0 ifTrue: [strm cr; tab; nextPutAll: (self dumpHeader: val)].
			a > 0 ifTrue: [strm nextPutAll: (self shortPrint: val)].
			a = self wordSize ifTrue:
				[(self isCompiledMethod: oop) ifTrue:
					[strm cr; tab; nextPutAll: (self dumpMethodHeader: val)]]].
			prevVal := val].
		lastLong := 256 min: (self sizeBitsOf: oop) - self baseHeaderSize.
		hdrType = 2
			ifTrue:
			["free" strm cr; nextPutAll: (oop+(self longAt: oop)-2) hex;
			space; space; nextPutAll: (oop+(self longAt: oop)-2) printString]
			ifFalse:
			[(self formatOf: oop) = 3
			ifTrue:
				[strm cr; tab; nextPutAll: '/ next 3 fields are above SP... /'.
				lastPtr+self wordSize to: lastPtr+(3*self wordSize) by: self wordSize do:
					[:a | val := self longAt: oop+a.
					strm cr; nextPutAll: a hex; 
						space; space; space; nextPutAll: val hex8; space; space.
					(self validOop: val) ifTrue: [strm nextPutAll: (self shortPrint: val)]]]
			ifFalse:
			[lastPtr+self wordSize to: lastLong by: self wordSize do:
				[:a | val := self longAt: oop+a.
				strm cr; nextPutAll: (a<16 ifTrue: [' ', a hex] ifFalse: [a hex]); 
					space; space; space.
				strm nextPutAll: val hex8; space; space;
						nextPutAll: (self charsOfLong: val)]]].
	]! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
nameOfClass: classOop
	(self sizeBitsOf: classOop) = (Metaclass instSize +1*self wordSize) ifTrue:
		[^ (self nameOfClass:
				(self fetchPointer: 5 "thisClass" ofObject: classOop)) , ' class'].
	^ self stringOf: (self fetchPointer: 6 "name" ofObject: classOop)! !

!InterpreterSimulator methodsFor: 'initialization' stamp: 'th 4/15/2000 17:19'!
nextLongFrom: aStream
	"Read a 32-bit quantity from the given (binary) stream."
	^self subclassResponsibility! !

!InterpreterSimulator methodsFor: 'initialization' stamp: 'di 9/23/97 15:51'!
nextLongFrom: aStream swap: swapFlag
	swapFlag 
		ifTrue: [^ self byteSwapped: (self nextLongFrom: aStream)]
		ifFalse: [^ self nextLongFrom: aStream]! !

!InterpreterSimulator methodsFor: 'debugging traps' stamp: 'di 7/22/2004 18:36'!
normalSend
	"Catch errors before we start the whole morphic error process"

	"(byteCount > 4000000 and: [(self stringOf: messageSelector) = 'sorts:before:'])
		ifTrue: [self halt]."
	^ super normalSend! !

!InterpreterSimulator methodsFor: 'UI' stamp: 'eem 9/7/2018 18:48'!
openAsMorph
	"Open a morphic view on this simulation."
	| window localImageName |
	localImageName := imageName
							ifNotNil: [self localNameFor: imageName]
							ifNil: [' synthetic image'].
	window := (SystemWindow labelled: 'Simulation of ' , localImageName) model: self.

	window addMorph: (displayView := ImageMorph new image: displayForm)
		frame: (0@0 corner: 1@0.8).

	transcript := TranscriptStream on: (String new: 10000).
	window addMorph: (PluggableTextMorph on: transcript text: nil accept: nil
			readSelection: nil menu: #codePaneMenu:shifted:)
		frame: (0@0.8 corner: 0.7@1).

	window addMorph: (PluggableTextMorph on: self
						text: #byteCountText accept: nil) hideScrollBarsIndefinitely
		frame: (0.7@0.8 corner: 1@1).

	window openInWorld.
	^window! !

!InterpreterSimulator methodsFor: 'initialization' stamp: 'di 7/1/2004 14:15'!
openOn: fileName
	"(InterpreterSimulator new openOn: 'clonex.image') test"

	self openOn: fileName extraMemory: 2500000.! !

!InterpreterSimulator methodsFor: 'initialization' stamp: 'eem 8/13/2018 15:33'!
openOn: fileName extraMemory: extraBytes
	"InterpreterSimulator new openOn: 'clone.im' extraMemory: 100000"

	| f version headerSize count oldBaseAddr bytesToShift swapBytes |
	"open image file and read the header"

	["begin ensure block..."
	f := FileStream readOnlyFileNamed: fileName.
	imageName := f fullName.
	f binary.
	version := self nextLongFrom: f.  "current version: 16r1968 (=6504) vive la revolucion!!"
	(self readableFormat: version)
		ifTrue: [swapBytes := false]
		ifFalse: [(version := self byteSwapped: version) = self imageFormatVersion
					ifTrue: [swapBytes := true]
					ifFalse: [self error: 'incomaptible image format']].
	headerSize := self nextLongFrom: f swap: swapBytes.
	self setEndOfMemory: (self nextLongFrom: f swap: swapBytes).  "first unused location in heap"
	oldBaseAddr := self nextLongFrom: f swap: swapBytes.  "object memory base address of image"
	specialObjectsOop := self nextLongFrom: f swap: swapBytes.
	lastHash := self nextLongFrom: f swap: swapBytes.  "Should be loaded from, and saved to the image header"
	lastHash = 0 ifTrue: [lastHash := 999].

	savedWindowSize	:= self nextLongFrom: f swap: swapBytes.
	fullScreenFlag		:= self nextLongFrom: f swap: swapBytes.
	extraVMMemory		:= self nextLongFrom: f swap: swapBytes.

	"allocate interpreter memory"
	self setMemoryLimit: endOfMemory + extraBytes.

	"read in the image in bulk, then swap the bytes if necessary"
	f position: headerSize.
	memory := Bitmap new: memoryLimit // 4.
	count := f readInto: memory startingAt: 1 count: endOfMemory // 4.
	count ~= (endOfMemory // 4) ifTrue: [self halt].
	]
		ensure: [f close].

	swapBytes ifTrue:
		[UIManager default
			informUser: 'Swapping bytes of foreign image...'
			during: [self reverseBytesInImage]].

	self initialize.
	bytesToShift := self startOfMemory - oldBaseAddr.  "adjust pointers for zero base address"
	UIManager default
		informUser: 'Relocating object pointers...'
		during: [self initializeInterpreter: bytesToShift]! !

!InterpreterSimulator methodsFor: 'other primitives' stamp: 'di 6/29/1998 22:17'!
primStringcomparewithcollated
	^ self primitiveFail! !

!InterpreterSimulator methodsFor: 'other primitives' stamp: 'ikp 12/15/1998 23:30'!
primStringfindSubstringinstartingAtmatchTable
	^self primitiveFail! !

!InterpreterSimulator methodsFor: 'other primitives' stamp: 'di 11/8/1998 13:04'!
primStringindexOfAsciiinStringstartingAt
	^ self primitiveFail! !

!InterpreterSimulator methodsFor: 'other primitives' stamp: 'yo 1/29/1999 21:37'!
primStringtranslatefromtotable
	^ self primitiveFail! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'di 4/20/2004 23:59'!
primitiveBeDisplay
	"Extended to create a scratch Form for use by showDisplayBits."

	| rcvr destWidth destHeight destDepth |
	rcvr := self stackTop.
	self success: ((self isPointers: rcvr) and: [(self lengthOf: rcvr) >= 4]).
	successFlag ifTrue: [
		destWidth := self fetchInteger: 1 ofObject: rcvr.
		destHeight := self fetchInteger: 2 ofObject: rcvr.
		destDepth := self fetchInteger: 3 ofObject: rcvr.
	].
	successFlag ifTrue: [
		"create a scratch form the same size as Smalltalk displayObj"
		displayForm := Form extent: destWidth @ destHeight
							depth: destDepth.
		displayView ifNotNil: [displayView image: displayForm].
	].
	super primitiveBeDisplay.! !

!InterpreterSimulator methodsFor: 'other primitives' stamp: ''!
primitiveBeep

	Beeper beep.! !

!InterpreterSimulator methodsFor: 'other primitives' stamp: 'eem 2/13/2014 09:36'!
primitiveGetAttribute
	"Fetch the system attribute with the given integer ID. The result is a string, which will be empty if the attribute is not defined."

	| attr s attribute |
	attr := self stackIntegerValue: 0.
	successFlag ifTrue: [
		attribute := Smalltalk vm getSystemAttribute: attr.
		attribute ifNil: [ ^self primitiveFail ].
		s := self instantiateClass: (self splObj: ClassByteString) indexableSize: attribute size.
		1 to: attribute size do: [ :i |
			self storeByte: i-1 ofObject: s withValue: (attribute at: i) asciiValue].
		self pop: 2  "rcvr, attr" thenPush: s].
! !

!InterpreterSimulator methodsFor: 'file primitives' stamp: 'eem 1/2/2013 17:37'!
primitiveImageName
	"Note: For now, this only implements getting, not setting, the image file name."
	| result imageNameSize |
	self pop: 1.
	imageNameSize := imageName size.
	result := self instantiateClass: (self splObj: ClassByteString)
				   indexableSize: imageNameSize.
	1 to: imageNameSize do:
		[:i | self storeByte: i-1 ofObject: result
			withValue: (imageName at: i) asciiValue].
	self push: result.! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'di 9/23/97 12:54'!
primitiveKbdNext

	self pop: 1.
	Sensor keyboardPressed
		ifTrue: [self pushInteger: Sensor primKbdNext]
		ifFalse: [self push: nilObj]! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'di 9/23/97 12:54'!
primitiveKbdPeek

	self pop: 1.
	Sensor keyboardPressed
		ifTrue: [self pushInteger: Sensor primKbdPeek]
		ifFalse: [self push: nilObj]! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'di 9/23/97 12:54'!
primitiveMouseButtons
	| buttons |
	self pop: 1.
	buttons := Sensor primMouseButtons.
	self pushInteger: buttons! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'di 4/21/2004 00:21'!
primitiveMousePoint

	| relPt |
	self pop: 1.
	displayForm == nil
		ifTrue: [self push: (self makePointwithxValue: 99 yValue: 66)]
		ifFalse: [relPt := Sensor cursorPoint - self displayLocation.
				self push: (self makePointwithxValue: relPt x yValue: relPt y)]! !

!InterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'di 12/1/1998 00:05'!
primitiveRelinquishProcessor
	"No-op in simulator"

	^ self pop: 1! !

!InterpreterSimulator methodsFor: 'debugging traps' stamp: 'di 7/22/2004 12:31'!
primitiveResume
	"Catch errors before we start the whole morphic error process"

	byteCount > 1000000 ifTrue: [self halt].  "Ignore early process activity"
	^ super primitiveResume! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'topa 7/10/2016 23:15'!
primitiveScreenScaleFactor "Dummied for now"

	self pop: 1.
	self pushFloat: 1.0
! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: ''!
primitiveScreenSize  "Dummied for now"

	self pop: 1.
	self push: (self makePointwithxValue: 640 yValue: 480).! !

!InterpreterSimulator methodsFor: 'debugging traps' stamp: 'di 7/22/2004 12:31'!
primitiveSuspend
	"Catch errors before we start the whole morphic error process"

	byteCount > 1000000 ifTrue: [self halt].  "Ignore early process activity"
	^ super primitiveSuspend! !

!InterpreterSimulator methodsFor: 'debug printing' stamp: 'di 5/11/2004 18:28'!
print: s

	traceOn ifTrue: [ transcript show: s ]! !

!InterpreterSimulator methodsFor: 'debug printing' stamp: 'di 5/11/2004 18:28'!
printChar: aByte

	traceOn ifTrue: [ transcript nextPut: aByte asCharacter ].! !

!InterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 7/1/2008 09:21'!
printCurrentBytecodeOn: aStream
	| code |
	code := currentBytecode radix: 16.
	aStream print: localIP - method - 3;
		tab;
		nextPut: (code size < 2
					ifTrue: [$0]
					ifFalse: [code at: 1]);
		nextPut: code last; space;
		nextPutAll: (BytecodeTable at: currentBytecode + 1);
		space;
		nextPut: $(; print: byteCount; nextPut: $)! !

!InterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 5/21/2009 10:00'!
printHex: anInteger

	traceOn ifTrue:
		[| it16 |
		 it16 := anInteger radix: 16.
		 transcript
			next: 8 - it16 size put: Character space;
			nextPutAll: (anInteger storeStringBase: 16)]! !

!InterpreterSimulator methodsFor: 'debug printing' stamp: 'di 5/11/2004 18:28'!
printNum: anInteger

	traceOn ifTrue: [ transcript show: anInteger printString ].! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'di 7/19/2004 14:48'!
printStack
	^ self printStack: false! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'di 7/19/2004 14:49'!
printStack: includeTemps
	| ctxt |
	ctxt := activeContext.
	^ String streamContents:
		[:strm |
		[self printStackFrame: ctxt onStream: strm.
		includeTemps ifTrue: [self printStackTemps: ctxt onStream: strm].
		(ctxt := (self fetchPointer: SenderIndex ofObject: ctxt)) = nilObj]
				whileFalse: [].
		]! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'di 7/19/2004 14:49'!
printStackFrame: ctxt onStream: strm
	| classAndSel home |
	home := (self fetchClassOf: ctxt) = (self splObj: ClassBlockContext)
		ifTrue: [self fetchPointer: HomeIndex ofObject: ctxt]
		ifFalse: [ctxt].
	classAndSel := self
		classAndSelectorOfMethod: (self fetchPointer: MethodIndex ofObject: home)
		forReceiver: (self fetchPointer: ReceiverIndex ofObject: home).
	strm cr; nextPutAll: ctxt hex8.
	ctxt = home ifFalse: [strm nextPutAll: ' [] in'].
	strm space; nextPutAll: (self nameOfClass: classAndSel first).
	strm nextPutAll: '>>'; nextPutAll: (self shortPrint: classAndSel last).
! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
printStackTemps: ctxt onStream: strm
	| home cMethod nArgs nTemps oop |
	home := (self fetchClassOf: ctxt) = (self splObj: ClassBlockContext)
		ifTrue: [self fetchPointer: HomeIndex ofObject: ctxt]
		ifFalse: [ctxt].
	cMethod := self fetchPointer: MethodIndex ofObject: home.
	nArgs := nTemps := 0.

	home = ctxt ifTrue:
		[strm cr; tab; nextPutAll: 'args: '.
		nArgs := self argumentCountOf: cMethod.
		1 to: nArgs do:
			[:i | oop := self fetchPointer: TempFrameStart + i-1 ofObject: ctxt.
			strm nextPutAll: oop hex; space].

		strm cr; tab; nextPutAll: 'temps: '.
		nTemps := self tempCountOf: cMethod.
		nArgs+1 to: nTemps do:
			[:i | oop := self fetchPointer: TempFrameStart + i-1 ofObject: ctxt.
			strm nextPutAll: oop hex; space]].
	
	strm cr; tab; nextPutAll: 'stack: '.
	nTemps + 1 to: (self lastPointerOf: ctxt)//self wordSize - TempFrameStart do:
		[:i | oop := self fetchPointer: TempFrameStart + i-1 ofObject: ctxt.
			strm nextPutAll: oop hex; space].
	! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'di 7/19/2004 15:28'!
printStackWithTemps
	^ self printStack: true! !

!InterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 7/1/2008 09:25'!
printStringForCurrentBytecode
	^String streamContents: [:str| self printCurrentBytecodeOn: str]! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
printTop: n
	"Print important fields of the top n contexts"
	| ctxt classAndSel home top ip sp |
	ctxt := activeContext.
	^ String streamContents:
		[:strm | 1 to: n do:
			[:i |
			home := (self fetchClassOf: ctxt) = (self splObj: ClassBlockContext)
				ifTrue: [self fetchPointer: HomeIndex ofObject: ctxt]
				ifFalse: [ctxt].
			classAndSel := self
				classAndSelectorOfMethod: (self fetchPointer: MethodIndex ofObject: home)
				forReceiver: (self fetchPointer: ReceiverIndex ofObject: home).
			strm cr; nextPutAll: ctxt hex8.
			ctxt = home ifFalse: [strm nextPutAll: ' [] in'].
			strm space; nextPutAll: (self nameOfClass: classAndSel first).
			strm nextPutAll: '>>'; nextPutAll: (self shortPrint: classAndSel last).
			ctxt = activeContext
				ifTrue: [ip := instructionPointer - method - (self baseHeaderSize - 2).
						sp := self stackPointerIndex - TempFrameStart + 1.
						top := self stackTop]
				ifFalse: [ip := self integerValueOf:
							(self fetchPointer: InstructionPointerIndex ofObject: ctxt).
						sp := self integerValueOf:
							(self fetchPointer: StackPointerIndex ofObject: ctxt).
						top := self longAt: ctxt + (self lastPointerOf: ctxt)].
			strm cr; tab; nextPutAll: 'ip = '; print: ip.
			strm cr; tab; nextPutAll: 'sp = '; print: sp.
			strm cr; tab; nextPutAll: 'top = '; nextPutAll: (self shortPrint: top).
			(ctxt := (self fetchPointer: SenderIndex ofObject: ctxt)) = nilObj
				ifTrue: [^strm contents].
			].
		]! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'di 5/11/2004 18:28'!
profile: nBytecodes
	"(InterpreterSimulator new openOn: 'clonex.image') profile: 60000"
	transcript clear.
	byteCount := 0.
	MessageTally spyOn: [self runForNBytes: nBytecodes].
	self close! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'di 11/29/2000 10:25'!
profileSends: nBytecodes
	"(InterpreterSimulator new openOn: 'clonex.image') profileSends: 5000"
	MessageTally tallySendsTo: self
		inBlock: [self runForNBytes: nBytecodes]
		showTree: true.
	self close! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'eem 7/9/2008 16:34'!
runAtEachStep: aBlock
	self internalizeIPandSP.
	self fetchNextBytecode.
	[true] whileTrue:
		[aBlock value: currentBytecode.
		 self dispatchOn: currentBytecode in: BytecodeTable.
		 byteCount := byteCount + 1].
	localIP := localIP - 1.
	"undo the pre-increment of IP before returning"
	self externalizeIPandSP! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'eem 7/9/2008 16:33'!
runAtEachStep: aBlock breakCount: breakCount
	self internalizeIPandSP.
	self fetchNextBytecode.
	[true] whileTrue:
		[aBlock value: currentBytecode.
		 self dispatchOn: currentBytecode in: BytecodeTable.
		 (byteCount := byteCount + 1) = breakCount ifTrue:
			[self halt]].
	localIP := localIP - 1.
	"undo the pre-increment of IP before returning"
	self externalizeIPandSP! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'nk 4/3/2004 11:56'!
runForNBytes: nBytecodes 
	"Do nByteCodes more bytecode dispatches.
	Keep byteCount up to date.
	This can be run repeatedly."
	| endCount |
	endCount := byteCount + nBytecodes.
	self internalizeIPandSP.
	self fetchNextBytecode.
	[byteCount < endCount]
		whileTrue: [self dispatchOn: currentBytecode in: BytecodeTable.
			byteCount := byteCount + 1].
	localIP := localIP - 1.
	"undo the pre-increment of IP before returning"
	self externalizeIPandSP! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'eem 7/6/2008 21:45'!
runUntilDivergenceFromTrace: aSequence
	self internalizeIPandSP.
	self fetchNextBytecode.
	[true] whileTrue:
		[(aSequence at: byteCount + 1) ~~ (BytecodeTable at: currentBytecode + 1) ifTrue:
			[self halt: byteCount printString, ' ', (aSequence at: byteCount + 1), ' ~~ ', (BytecodeTable at: currentBytecode + 1)].
		 self dispatchOn: currentBytecode in: BytecodeTable.
		 byteCount := byteCount + 1].
	localIP := localIP - 1.
	"undo the pre-increment of IP before returning"
	self externalizeIPandSP! !

!InterpreterSimulator methodsFor: 'memory access' stamp: 'di 6/23/2004 14:29'!
shortAt: byteAddress
    "Return the half-word at byteAddress which must be even."
	^self subclassResponsibility! !

!InterpreterSimulator methodsFor: 'memory access' stamp: 'di 6/23/2004 14:32'!
shortAt: byteAddress put: a16BitValue
	^ self subclassResponsibility! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
shortPrint: oop
	| name classOop |
	(self isIntegerObject: oop) ifTrue: [^ '=' , (self integerValueOf: oop) printString , 
		' (' , (self integerValueOf: oop) hex , ')'].
	classOop := self fetchClassOf: oop.
	(self sizeBitsOf: classOop) = (Metaclass instSize +1*self wordSize) ifTrue: [
		^ 'class ' , (self nameOfClass: oop)].
	name := self nameOfClass: classOop.
	name size = 0 ifTrue: [name := '??'].
	name = 'String' ifTrue: [^ (self stringOf: oop) printString].
	name = 'ByteString' ifTrue: [^ (self stringOf: oop) printString].
	name = 'Symbol' ifTrue: [^ '#' , (self stringOf: oop)].
	name = 'ByteSymbol' ifTrue: [^ '#' , (self stringOf: oop)].
	name = 'Character' ifTrue: [^ '=' , (Character value: (self integerValueOf: 
				(self fetchPointer: 0 ofObject: oop))) printString].
	name = 'UndefinedObject' ifTrue: [^ 'nil'].
	name = 'False' ifTrue: [^ 'false'].
	name = 'True' ifTrue: [^ 'true'].
	name = 'Float' ifTrue: [successFlag := true. ^ '=' , (self floatValueOf: oop) printString].
	name = 'Association' ifTrue: [^ '(' ,
				(self shortPrint: (self longAt: oop + self baseHeaderSize)) ,
				' -> ' ,
				(self longAt: oop + self baseHeaderSize + self wordSize) hex8 , ')'].
	('AEIOU' includes: name first)
		ifTrue: [^ 'an ' , name]
		ifFalse: [^ 'a ' , name]! !

!InterpreterSimulator methodsFor: 'I/O primitives' stamp: 'di 7/6/2004 10:48'!
showDisplayBits: destBits w: w h: h d: d left: left right: right top: top bottom: bottom
	| raster pixPerWord simDisp realDisp rect |
	pixPerWord := 32 // d.
	raster := displayForm width + (pixPerWord - 1) // pixPerWord.
	simDisp := Form new hackBits: memory.
	displayForm unhibernate.
	realDisp := Form new hackBits: displayForm bits.
	realDisp
		copy: (0 @ (top * raster) extent: 4 @ (bottom - top * raster))
		from: 0 @ (destBits // 4 + (top * raster))
		in: simDisp
		rule: Form over.
	displayView ifNotNil: [^ displayView changed].
	
	"If running without a view, just blat the bits onto the screen..."
	rect := 0 @ top corner: displayForm width @ bottom.
	Display
		copy: (rect translateBy: self displayLocation)
		from: rect topLeft
		in: displayForm
		rule: Form over! !

!InterpreterSimulator methodsFor: 'debug printing' stamp: 'eem 6/30/2008 21:06'!
space

	traceOn ifTrue: [ transcript space ]! !

!InterpreterSimulator methodsFor: 'plugin support' stamp: 'ar 5/11/2000 21:47'!
sqGetInterpreterProxy
	"I am basically my own proxy..."
	^self! !

!InterpreterSimulator methodsFor: 'memory access' stamp: 'di 5/11/2004 18:29'!
sqGrowMemory: oldLimit By: delta

	transcript show: 'grow memory from ', oldLimit printString, ' by ', delta printString; cr.
	memory := memory , (memory class new: delta // 4).
	^ memory size * 4! !

!InterpreterSimulator methodsFor: 'memory access' stamp: 'ar 2/25/2001 17:16'!
sqMemoryExtraBytesLeft: includingSwap
	^0! !

!InterpreterSimulator methodsFor: 'memory access' stamp: 'di 5/11/2004 18:29'!
sqShrinkMemory: oldLimit By: delta
	transcript show: 'shrink memory from ', oldLimit printString, ' by ', delta printString, ' remember it doesn''t actually shrink in simulation'; cr.

	^ oldLimit! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'di 12/22/1998 23:24'!
stackDepth
	| ctxt n |
	ctxt := activeContext.
	n := 0.
	[(ctxt := (self fetchPointer: SenderIndex ofObject: ctxt)) = nilObj]
		whileFalse: [n := n+1].
	^ n! !

!InterpreterSimulator methodsFor: 'initialization' stamp: ''!
startOfMemory
	"Return the start of object memory."

	^ 0! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
stats
	| oop fieldAddr fieldOop last stats v d |
	stats := Bag new.
	oop := self firstObject.

'Scanning the image...' displayProgressAt: Sensor cursorPoint
	from: oop to: endOfMemory
	during: [:bar |

	[oop < endOfMemory] whileTrue:
		[(self isFreeObject: oop) ifFalse:
			[stats add: #objects.
			fieldAddr := oop + (self lastPointerOf: oop).
			[fieldAddr > oop] whileTrue:
				[fieldOop := self longAt: fieldAddr.
				(self isIntegerObject: fieldOop)
					ifTrue: [v := self integerValueOf: fieldOop.
							(v between: -16000 and: 16000)
								ifTrue: [stats add: #ints32k]
								ifFalse: [stats add: #intsOther]]
					ifFalse: [fieldOop = nilObj ifTrue: [stats add: #nil]
							ifFalse:
							[d := fieldOop - oop.
							(d between: -16000 and: 16000)
								ifTrue: [stats add: #oops32k]
								ifFalse: [stats add: #oopsOther]]].
				fieldAddr := fieldAddr - self wordSize]].
		bar value: oop.
		last := oop.
		last := last.
		oop := self objectAfter: oop]].
	^ stats sortedElements! !

!InterpreterSimulator methodsFor: 'bytecode routines' stamp: 'di 12/12/1998 08:30'!
storeAndPopReceiverVariableBytecode
	"Note: This code uses storePointerUnchecked:ofObject:withValue: and does the store check explicitely in order to help the translator produce better code."

	"Interpreter version has fetchNextBytecode out of order"
	| rcvr top |
	rcvr := receiver.
	top := self internalStackTop.
	(rcvr < youngStart) ifTrue: [
		self possibleRootStoreInto: rcvr value: top.
	].
	self storePointerUnchecked: (currentBytecode bitAnd: 7)
		ofObject: rcvr
		withValue: top.
	self internalPop: 1.
	self fetchNextBytecode.
! !

!InterpreterSimulator methodsFor: 'float primitives' stamp: 'eem 2/9/2009 10:10'!
storeFloatAt: floatBitsAddress from: aFloat

	self long32At: floatBitsAddress put: (aFloat at: 1).
	self long32At: floatBitsAddress+4 put: (aFloat at: 2).
! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
stringOf: oop
	| size long nLongs chars |
	^ String streamContents:
		[:strm |
		size := 100 min: (self stSizeOf: oop).
		nLongs := size-1//self wordSize+1.
		1 to: nLongs do:
			[:i | long := self longAt: oop + self baseHeaderSize + (i-1*self wordSize).
			chars := self charsOfLong: long.
			strm nextPutAll: (i=nLongs
							ifTrue: [chars copyFrom: 1 to: size-1\\self wordSize+1]
							ifFalse: [chars])]]! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'eem 11/25/2009 11:09'!
tenuringIncrementalGC
	transcript cr; nextPutAll: 'tenuringIncrementalGC ('; print: byteCount; nextPut: $); flush.
	^super tenuringIncrementalGC! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'di 7/19/2004 17:30'!
test
	transcript clear.
	byteCount := 0.
	quitBlock := [^ self].
	self internalizeIPandSP.
	self fetchNextBytecode.
	[true] whileTrue:
		[self dispatchOn: currentBytecode in: BytecodeTable.
		byteCount := byteCount + 1.
		byteCount \\ 10000 = 0 ifTrue: [self fullDisplay]].
	self externalizeIPandSP.
! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'eem 7/24/2008 18:12'!
test1
	| breakCount printBytecodeAtEachStep |
	transcript clear.
	byteCount := 0.
	breakCount := -1.
	breakSelector := #openOn:context:label:contents:fullView:.
	quitBlock := [^self].
	printSends := true.
	printBytecodeAtEachStep := false.
	self internalizeIPandSP.
	self fetchNextBytecode.
	[true] whileTrue:
		[printBytecodeAtEachStep ifTrue:
			[self printCurrentBytecodeOn: Transcript.
			 Transcript cr; flush].

		 self dispatchOn: currentBytecode in: BytecodeTable.

		 byteCount := byteCount + 1.
		 byteCount = breakCount ifTrue:
			["printFrameAtEachStep :=" printBytecodeAtEachStep := true.
			 self halt: 'hit breakCount break-point'].
		 byteCount \\ 10000 = 0 ifTrue: [self fullDisplay]].
	self externalizeIPandSP! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'brp 9/19/2003 16:09'!
testBecome
	"Become some young things.  AA testBecome    "
	| array list1 list2 p1 p2 p3 p4 |
	array := self splObj: ClassArray.
	list1 := self instantiateClass: array indexableSize: 2.
	list2 := self instantiateClass: array indexableSize: 2.
	p1 := self instantiateClass: (self splObj: ClassPoint) indexableSize: 0.
	self push: p1.
	self storePointer: 0 ofObject: list1 withValue: p1.
	p2 := self instantiateClass: (self splObj: ClassPoint) indexableSize: 0.
	self push: p2.
	self storePointer: 1 ofObject: list1 withValue: p2.
	p3 := self instantiateClass: (self splObj: ClassMessage) indexableSize: 0.
	self push: p3.
	self storePointer: 0 ofObject: list2 withValue: p3.
	p4 := self instantiateClass: (self splObj: ClassMessage) indexableSize: 0.
	self push: p4.
	self storePointer: 1 ofObject: list2 withValue: p4.
	(self become: list1 with: list2 twoWay: true copyHash: true) ifFalse: [self error: 'failed'].
	self popStack = p2 ifFalse: [self halt].
	self popStack = p1 ifFalse: [self halt].
	self popStack = p4 ifFalse: [self halt].
	self popStack = p3 ifFalse: [self halt].
	(self fetchPointer: 0 ofObject: list1) = p3 ifFalse: [self halt].
	(self fetchPointer: 1 ofObject: list1) = p4 ifFalse: [self halt].
	(self fetchPointer: 0 ofObject: list2) = p1 ifFalse: [self halt].
	(self fetchPointer: 1 ofObject: list2) = p2 ifFalse: [self halt].! !

!InterpreterSimulator methodsFor: 'interpreter shell' stamp: 'tpr 2/26/2003 14:00'!
unableToReadImageError
	self error:  'Read failed or premature end of image file'! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'di 4/3/1999 15:45'!
validOop: oop
	" Return true if oop appears to be valid "
	(oop bitAnd: 1) = 1 ifTrue: [^ true].  "Integer"
	(oop bitAnd: 3) = 0 ifFalse: [^ false].  "Uneven address"
	oop >= endOfMemory ifTrue: [^ false].  "Out of range"
	"could test if within the first large freeblock"
	(self longAt: oop) = 4 ifTrue: [^ false].
	(self headerType: oop) = 2 ifTrue: [^ false].	"Free object"
	^ true! !

!InterpreterSimulator methodsFor: 'testing' stamp: ''!
validateActiveContext
	self validateOopsIn: activeContext.	"debug -- test if messed up"! !

!InterpreterSimulator methodsFor: 'testing' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
validateOopsIn: object
	| fieldPtr limit former header | 
	"for each oop in me see if it is legal"
	fieldPtr := object + self baseHeaderSize.	"first field"
	limit := object + (self lastPointerOf: object).	"a good field"
	[fieldPtr > limit] whileFalse: [
		former := self longAt: fieldPtr.
		(self validOop: former) ifFalse: [self error: 'invalid oop in pointers object'].
		fieldPtr := fieldPtr + self wordSize].
	"class"
	header := self baseHeader: object.
	(header bitAnd: CompactClassMask) = 0 ifTrue: [	
		former := (self classHeader: object) bitAnd: AllButTypeMask.
		(self validOop: former) ifFalse: [self halt]].! !

!InterpreterSimulator methodsFor: 'file primitives' stamp: 'eem 9/7/2018 19:43'!
vmPathGet: stringBase Length: stringSize
	| pathName stringOop |
	pathName := self vmPath.
	stringOop := stringBase - self baseHeaderSize. "Due to C call in Interp"
	1 to: stringSize do:
		[:i | self storeByte: i-1 ofObject: stringOop
			withValue: (pathName at: i) asciiValue].
! !

!InterpreterSimulator methodsFor: 'debug support' stamp: 'eem 6/9/2008 15:46'!
warning: aString
	Transcript cr; nextPutAll: aString; flush! !

!InterpreterSimulator methodsFor: 'I/O primitives support' stamp: 'yo 6/3/1999 16:10'!
warpBits

	^ myBitBlt warpBits! !

!InterpreterSimulator methodsFor: 'image save/restore' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
writeImageFileIO: numberOfBytesToWrite
	"Actually emit the first numberOfBytesToWrite object memory bytes onto the snapshot."

	| headerSize file |
	self wordSize = 4 ifFalse: [self error: 'Not rewritten for 64 bits yet'].
	headerSize := 64.

	[
		file := (FileStream fileNamed: imageName) binary.
		file == nil ifTrue: [^nil].
	
		{
			self imageFormatVersion.
			headerSize.
			numberOfBytesToWrite.
			self startOfMemory.
			specialObjectsOop.
			lastHash.
			self ioScreenSize.
			fullScreenFlag.
			extraVMMemory
		}
			do: [:long | self putLong: long toFile: file].
	
		"Pad the rest of the header."
		7 timesRepeat: [self putLong: 0 toFile: file].
	
		"Position the file after the header."
		file position: headerSize.
	
		"Write the object memory."
		1
			to: numberOfBytesToWrite // 4
			do: [:index |
				self
					putLong: (memory at: index)
					toFile: file].
	
		self success: true
	]
		ensure: [file close]! !

InterpreterSimulator subclass: #InterpreterSimulatorLSB
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!InterpreterSimulatorLSB commentStamp: '<historical>' prior: 0!
This class overrides a few methods in InterpreterSimulator required for simulation to work on little-endian architectures (such as the x86 family of processors).  To start it up simply use InterpreterSimulatorLSB instead of InterpreterSimulator (see the class comment there for more details).  For example:

	(InterpreterSimulatorLSB new openOn: Smalltalk imageName) test

Note that the image must have been saved at least once on the local architecture, since the compiled VM performs some byte swapping that the simulator cannot cope with.!

!InterpreterSimulatorLSB methodsFor: 'memory access' stamp: 'nk 4/3/2004 06:58'!
byteAt: byteAddress
	| lowBits long |
	lowBits := byteAddress bitAnd: 3.
	long := self longAt: byteAddress - lowBits.
	^(lowBits caseOf: {
		[0] -> [ long ].
		[1] -> [ long bitShift: -8  ].
		[2] -> [ long bitShift: -16 ].
		[3] -> [ long bitShift: -24 ]
	}) bitAnd: 16rFF
! !

!InterpreterSimulatorLSB methodsFor: 'memory access' stamp: 'eem 2/27/2009 11:11'!
byteAt: byteAddress put: byte
	| lowBits long longAddress |
	lowBits := byteAddress bitAnd: 3.
	longAddress := byteAddress - lowBits.
	long := self longAt: longAddress.
	long := (lowBits caseOf: {
		[0] -> [ (long bitAnd: 16rFFFFFF00) bitOr: byte ].
		[1] -> [ (long bitAnd: 16rFFFF00FF) bitOr: (byte bitShift: 8) ].
		[2] -> [ (long bitAnd: 16rFF00FFFF) bitOr: (byte bitShift: 16)  ].
		[3] -> [ (long bitAnd: 16r00FFFFFF) bitOr: (byte bitShift: 24)  ]
	}).

	self longAt: longAddress put: long.
	^byte! !

!InterpreterSimulatorLSB methodsFor: 'memory access' stamp: 'nice 4/3/2016 08:34'!
byteSwapped32IfBigEndian: w
	"Accelerate, we now we are not big endian"
	^w bitAnd: 16rFFFFFFFF! !

!InterpreterSimulatorLSB methodsFor: 'memory access' stamp: 'nice 4/3/2016 08:34'!
byteSwapped64IfBigEndian: w
	"Accelerate, we now we are not big endian"
	^w bitAnd: 16rFFFFFFFFFFFFFFFF! !

!InterpreterSimulatorLSB methodsFor: 'debug support' stamp: 'ikp 12/11/1998 01:35'!
charsOfLong: long
	^ (1 to: 4) collect:
		[:i | ((long digitAt: i) between: 14 and: 126)
					ifTrue: [(long digitAt: i) asCharacter]
					ifFalse: [$?]]! !

!InterpreterSimulatorLSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:59'!
halfWordHighInLong32: long32
	"Used by Balloon"

	^ long32 bitAnd: 16rFFFF! !

!InterpreterSimulatorLSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:59'!
halfWordLowInLong32: long32
	"Used by Balloon"

	^ long32 bitShift: -16! !

!InterpreterSimulatorLSB methodsFor: 'initialization' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
nextLongFrom: aStream
	"Read a 32- or 64-bit quantity from the given (binary) stream."

	^ aStream nextLittleEndianNumber: self wordSize! !

!InterpreterSimulatorLSB methodsFor: 'image save/restore' stamp: 'crl 2/28/2003 01:22'!
putLong: n toFile: f
	"Append the given 4-byte long word to the given file in my byte order. (Bytes will be swapped, if necessary, when the image is read on a different platform.) Set successFlag to false if the write fails."

	| remainingValue |

	remainingValue := n.
	4 timesRepeat: [
		f nextPut: (remainingValue bitAnd: 16rFF).
		remainingValue := remainingValue bitShift: -8].

	self success: true! !

!InterpreterSimulatorLSB methodsFor: 'memory access' stamp: 'di 6/23/2004 14:29'!
shortAt: byteAddress
    "Return the half-word at byteAddress which must be even."
	| lowBits long |
	lowBits := byteAddress bitAnd: 2.
	long := self longAt: byteAddress - lowBits.
	^ lowBits = 2
		ifTrue: [ long bitShift: -16 ]
		ifFalse: [ long bitAnd: 16rFFFF ].
! !

!InterpreterSimulatorLSB methodsFor: 'memory access' stamp: 'di 6/23/2004 14:31'!
shortAt: byteAddress put: a16BitValue
    "Return the half-word at byteAddress which must be even."
	| lowBits long longAddress |
	lowBits := byteAddress bitAnd: 2.
	lowBits = 0
		ifTrue:
		[ "storing into LS word"
		long := self longAt: byteAddress.
		self longAt: byteAddress
				put: ((long bitAnd: 16rFFFF0000) bitOr: a16BitValue)
		]
		ifFalse:
		[longAddress := byteAddress - 2.
		long := self longAt: longAddress.
		self longAt: longAddress
				put: ((long bitAnd: 16rFFFF) bitOr: (a16BitValue bitShift: 16))
		]! !

!InterpreterSimulatorLSB methodsFor: 'memory access' stamp: 'tpr 12/22/2005 17:46'!
vmEndianness
	"return 0 for little endian, 1 for big endian"
	^0! !

InterpreterSimulatorLSB subclass: #InterpreterSimulatorLSB64
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!InterpreterSimulatorLSB64 methodsFor: 'memory access' stamp: 'di 6/13/2004 10:56'!
long32At: byteAddress

	"Return the 32-bit word at byteAddress which must be 0 mod 4."
	| lowBits long |
	lowBits := byteAddress bitAnd: 4.
	long := self longAt: byteAddress - lowBits.
	^ lowBits = 4
		ifTrue: [ long bitShift: -32 ]
		ifFalse: [ long bitAnd: 16rFFFFFFFF ].
! !

!InterpreterSimulatorLSB64 methodsFor: 'memory access' stamp: 'di 6/13/2004 11:01'!
long32At: byteAddress put: a32BitValue
	"Store the 32-bit value at byteAddress which must be 0 mod 4."
	| lowBits long64 longAddress |
	lowBits := byteAddress bitAnd: 4.
	lowBits = 0
		ifTrue:
		[ "storing into LS word"
		long64 := self longAt: byteAddress.
		self longAt: byteAddress
				put: ((long64 bitAnd: 16rFFFFFFFF00000000) bitOr: a32BitValue)
		]
		ifFalse:
		[longAddress := byteAddress - 4.
		long64 := self longAt: longAddress.
		self longAt: longAddress
				put: ((long64 bitAnd: 16rFFFFFFFF) bitOr: (a32BitValue bitShift: 32))
		]! !

!InterpreterSimulatorLSB64 methodsFor: 'memory access' stamp: 'eem 9/10/2013 14:38'!
wordSize
	"overridden for 64-bit images..."

	^8! !

InterpreterSimulator subclass: #InterpreterSimulatorMSB
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!InterpreterSimulatorMSB methodsFor: 'memory access' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
byteAt: byteAddress
	| lowBits bpwMinus1 |
	bpwMinus1 := self wordSize-1.
	lowBits := byteAddress bitAnd: bpwMinus1.
	^ ((self longAt: byteAddress - lowBits)
		bitShift: (lowBits - bpwMinus1) * 8)
		bitAnd: 16rFF! !

!InterpreterSimulatorMSB methodsFor: 'memory access' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
byteAt: byteAddress put: byte
	| longWord shift lowBits bpwMinus1 longAddress |
	bpwMinus1 := self wordSize-1.
	lowBits := byteAddress bitAnd: bpwMinus1.
	longAddress := byteAddress - lowBits.
	longWord := self longAt: longAddress.
	shift := (bpwMinus1 - lowBits) * 8.
	longWord := longWord
				- (longWord bitAnd: (16rFF bitShift: shift))
				+ (byte bitShift: shift).
	self longAt: longAddress put: longWord.
	^byte! !

!InterpreterSimulatorMSB methodsFor: 'memory access' stamp: 'nice 4/3/2016 08:38'!
byteSwapped32IfBigEndian: w
	"Accelerate, we now we are big endian"
	^w byteSwap32! !

!InterpreterSimulatorMSB methodsFor: 'memory access' stamp: 'nice 4/3/2016 08:37'!
byteSwapped64IfBigEndian: w
	"Accelerate, we now we are big endian"
	^w byteSwap64! !

!InterpreterSimulatorMSB methodsFor: 'debug support' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
charsOfLong: long
	^ (self wordSize to: 1 by: -1) collect:
		[:i | ((long digitAt: i) between: 14 and: 126)
					ifTrue: [(long digitAt: i) asCharacter]
					ifFalse: [$?]]! !

!InterpreterSimulatorMSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:58'!
halfWordHighInLong32: long32
	"Used by Balloon"

	^ long32 bitShift: -16! !

!InterpreterSimulatorMSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:58'!
halfWordLowInLong32: long32
	"Used by Balloon"

	^ long32 bitAnd: 16rFFFF! !

!InterpreterSimulatorMSB methodsFor: 'initialization' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
nextLongFrom: aStream 
	"Read a 32- or 64-bit quantity from the given (binary) stream."

	^ aStream nextNumber: self wordSize! !

!InterpreterSimulatorMSB methodsFor: 'image save/restore' stamp: 'eem 10/4/2013 07:47'!
putLong: n toFile: f
	"Append the given 4-byte long word to the given file in my byte order. (Bytes will be swapped, if necessary, when the image is read on a different platform.) Set successFlag to false if the write fails."

	f
		nextPut: (n bitShift: -24);
		nextPut: ((n bitShift: -16) bitAnd: 16rFF);
		nextPut: ((n bitShift: -8) bitAnd: 16rFF);
		nextPut: (n bitAnd: 16rFF).

	self success: true! !

!InterpreterSimulatorMSB methodsFor: 'memory access' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
shortAt: byteAddress
    "Return the half-word at byteAddress which must be even."
	| lowBits bpwMinus2 |
	bpwMinus2 := self wordSize-2.
	lowBits := byteAddress bitAnd: bpwMinus2.
	^ ((self longAt: byteAddress - lowBits)
		bitShift: (lowBits - bpwMinus2) * 8)
		bitAnd: 16rFFFF
! !

!InterpreterSimulatorMSB methodsFor: 'memory access' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
shortAt: byteAddress put: a16BitValue
    "Return the half-word at byteAddress which must be even."
	| longWord shift lowBits bpwMinus2 longAddress |
	bpwMinus2 := self wordSize-2.
	lowBits := byteAddress bitAnd: bpwMinus2.
	longAddress := byteAddress - lowBits.
	longWord := self longAt: longAddress.
	shift := (bpwMinus2 - lowBits) * 8.
	longWord := longWord
				- (longWord bitAnd: (16rFFFF bitShift: shift))
				+ (a16BitValue bitShift: shift).
	self longAt: longAddress put: longWord
! !

!InterpreterSimulatorMSB methodsFor: 'memory access' stamp: 'tpr 12/22/2005 17:46'!
vmEndianness
	"return 0 for little endian, 1 for big endian"
	^1! !

InterpreterSimulatorMSB subclass: #InterpreterSimulatorMSB64
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!InterpreterSimulatorMSB64 methodsFor: 'memory access' stamp: 'nice 4/3/2016 00:08'!
byteSwapped: w
	"Return the given integer with its bytes in the reverse order."

	^w byteSwap64! !

!InterpreterSimulatorMSB64 methodsFor: 'memory access' stamp: 'di 7/3/2004 10:40'!
long32At: byteAddress
	"Return the 32-bit word at byteAddress which must be 0 mod 4."

	^ super longAt: byteAddress! !

!InterpreterSimulatorMSB64 methodsFor: 'memory access' stamp: 'di 7/3/2004 10:41'!
long32At: byteAddress put: a32BitValue
	"Store the 32-bit value at byteAddress which must be 0 mod 4."

	super longAt: byteAddress put: a32BitValue! !

!InterpreterSimulatorMSB64 methodsFor: 'memory access' stamp: 'di 6/9/2004 15:43'!
longAt: byteAddress
	"Note: Adjusted for Smalltalk's 1-based array indexing."

	^ ((super longAt: byteAddress) bitShift: 32) bitOr: (super longAt: byteAddress + 4)! !

!InterpreterSimulatorMSB64 methodsFor: 'memory access' stamp: 'di 6/9/2004 15:48'!
longAt: byteAddress put: a64BitValue
	"Note: Adjusted for Smalltalk's 1-based array indexing."

	super longAt: byteAddress put: (a64BitValue bitShift: -32).
	super longAt: byteAddress + 4 put: (a64BitValue bitAnd: 16rFFFFFFFF).
	^ a64BitValue! !

!InterpreterSimulatorMSB64 methodsFor: 'memory access' stamp: 'eem 9/10/2013 14:38'!
wordSize
	"overridden for 64-bit images..."

	^8! !

ObjectMemory subclass: #NewObjectMemory
	instanceVariableNames: 'coInterpreter freeStart reserveStart scavengeThreshold needGCFlag edenBytes checkForLeaks statGCEndUsecs heapMap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Interpreter'!

!NewObjectMemory commentStamp: '<historical>' prior: 0!
I am a refinement of ObjectMemory that eliminates the need for pushRemappableOop:/popRemappableOop in the interpreter proper.  Certain primitives that do major allocation may still want to provoke a garbage collection and hence may still need to remap private pointers.  But the interpreter subclass of this class does not have to provided it reserves sufficient space for it to make progress to the next scavenge point (send or backward branch).!

NewObjectMemory subclass: #NewCoObjectMemory
	instanceVariableNames: 'cogit'
	classVariableNames: ''
	poolDictionaries: 'CogMethodConstants'
	category: 'VMMaker-JIT'!

!NewCoObjectMemory class methodsFor: 'simulation' stamp: 'eem 12/4/2015 22:20'!
defaultISA
	"Answer the default instruction set architecture to use for simulation."
	^#IA32! !

!NewCoObjectMemory class methodsFor: 'accessing class hierarchy' stamp: 'eem 10/9/2013 17:57'!
objectRepresentationClass
	^CogObjectRepresentationForSqueakV3! !

!NewCoObjectMemory class methodsFor: 'simulation' stamp: 'eem 9/30/2010 13:29'!
simulatorClass
	^NewCoObjectMemorySimulator! !

!NewCoObjectMemory methodsFor: 'trampolines' stamp: 'eem 1/12/2013 17:29'!
ceStoreCheck: anOop
	<api>
	"Do the store check.  Answer the argument for the benefit of the code generator;
	ReceiverResultReg may be caller-saved and hence smashed by this call.  Answering
	it allows the code generator to reload ReceiverResultReg cheaply."
	self assert: (self isNonIntegerObject: anOop).
	self assert: (self oop: anOop isLessThan: self youngStart).
	self assert: (self isYoungRoot: anOop) not.
	self noteAsRoot: anOop headerLoc: anOop.
	^anOop! !

!NewCoObjectMemory methodsFor: 'debug support' stamp: 'eem 9/30/2013 22:54'!
cheapAddressCouldBeInHeap: address 
	^(address bitAnd: self wordSize - 1) = 0
	  and: [(self oop: address isGreaterThanOrEqualTo: self startOfMemory)
	  and: [self oop: address isLessThan: freeStart]]! !

!NewCoObjectMemory methodsFor: 'debug support' stamp: 'eem 2/27/2013 15:27'!
clearLeakMapAndMapAccessibleObjects
	"Perform an integrity/leak check using the heapMap.  Set a bit at each object's header.
	 Override to set a bit at each Cog method"
	super clearLeakMapAndMapAccessibleObjects.
	cogit addCogMethodsToHeapMap! !

!NewCoObjectMemory methodsFor: 'accessing' stamp: 'eem 8/25/2014 09:50'!
coInterpreter
	<doNotGenerate>
	^coInterpreter! !

!NewCoObjectMemory methodsFor: 'accessing' stamp: 'eem 8/25/2014 09:50'!
coInterpreter: aCoInterpreter
	<doNotGenerate>
	coInterpreter := aCoInterpreter! !

!NewCoObjectMemory methodsFor: 'accessing' stamp: 'eem 2/18/2011 15:40'!
cogit
	<doNotGenerate>
	^cogit! !

!NewCoObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:35'!
cogit: aCogit
	cogit := aCogit! !

!NewCoObjectMemory methodsFor: 'image segment in/out' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
copyObj: oop toSegment: segmentWordArray addr: lastSeg stopAt: stopAddr saveOopAt: oopPtr headerAt: hdrPtr
	"Copy this object into the segment beginning at lastSeg.
	Install a forwarding pointer, and save oop and header.
	Fail if out of space.  Return the next segmentAddr if successful."

	"Copy the object..."
	| extraSize bodySize hdrAddr |
	<inline: false>
	self flag: #Dan.  "None of the imageSegment stuff has been updated for 64 bits"
	extraSize := self extraHeaderBytes: oop.
	bodySize := self sizeBitsOf: oop.
	(self oop: (lastSeg + extraSize + bodySize) isGreaterThanOrEqualTo: stopAddr) ifTrue:
		[^0]. "failure"
	self transfer: extraSize + bodySize // self wordSize  "wordCount"
		from: oop - extraSize
		to: lastSeg+self wordSize.

	"Clear root and mark bits of all headers copied into the segment"
	hdrAddr := lastSeg+self wordSize + extraSize.
	self longAt: hdrAddr put: ((self longAt: hdrAddr) bitAnd: AllButRootBit - MarkBit).

	"Make sure Cogged methods have their true header field written to the segment."
	((self isCompiledMethod: oop)
	and: [coInterpreter methodHasCogMethod: oop]) ifTrue:
		[self longAt: hdrAddr+self baseHeaderSize put: (self methodHeaderOf: oop)].

	self forward: oop to: (lastSeg+self wordSize + extraSize - segmentWordArray)
		savingOopAt: oopPtr
		andHeaderAt: hdrPtr.

	"Return new end of segment"
	^lastSeg + extraSize + bodySize! !

!NewCoObjectMemory methodsFor: 'become' stamp: 'eem 2/20/2013 13:32'!
freeObject: obj header: objHeader
	(self isCompiledMethodHeader: objHeader) ifTrue:
		[(self asserta: (coInterpreter methodHasCogMethod: obj) not) ifFalse:
			[self error: 'attempt to free cogged method']].
	super freeObject: obj header: objHeader! !

!NewCoObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
freeStart: aValue
	^freeStart := aValue! !

!NewCoObjectMemory methodsFor: 'trampoline support' stamp: 'eem 5/6/2010 16:00'!
freeStartAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: freeStart) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #freeStart in: self]! !

!NewCoObjectMemory methodsFor: 'allocation' stamp: 'eem 7/1/2021 11:51'!
growObjectMemory: delta 
	"Attempt to grow the object memory by the given delta amount."
	| limit |
	statGrowMemory := statGrowMemory + 1.
	limit := self sqGrowMemory: memoryLimit By: delta.
	limit = memoryLimit ifFalse:
		[self setMemoryLimit: limit - 24. "remove a tad for safety"
		 self initializeMemoryFirstFree: freeStart]! !

!NewCoObjectMemory methodsFor: 'simulation' stamp: 'eem 8/28/2014 09:15'!
initializeFreeSpaceForFacadeFrom: base to: limit
	freeStart := base.
	endOfMemory := scavengeThreshold := limit! !

!NewCoObjectMemory methodsFor: 'debug support' stamp: 'eem 2/28/2017 10:43'!
isCompactInstance: oop
	"For assert checking"
	^(self isNonImmediate: oop) and: [((self baseHeader: oop) bitAnd: CompactClassMask) ~= 0]! !

!NewCoObjectMemory methodsFor: 'cog jit support' stamp: 'eem 10/25/2013 16:23'!
isReallyYoungObject: obj
	<api>
	"For machine code assertion checking.  Answer true if not in a fullGC and obj is young."
	^coInterpreter gcMode ~= GCModeFull
	  and: [self oop: obj isGreaterThanOrEqualTo: self youngStart]! !

!NewCoObjectMemory methodsFor: 'image save/restore' stamp: 'eem 10/10/2013 06:39'!
memoryBaseForImageRead
	"Answer the address to read the image into."
	^coInterpreter heapBase! !

!NewCoObjectMemory methodsFor: 'memory access' stamp: 'eem 4/28/2015 14:46'!
methodHeaderOf: methodObj
	"Answer the method header of a CompiledMethod object.
	 If the method has been cogged then the header is a pointer to
	 the CogMethod and the real header will be stored in the CogMethod."
	<api>
	<inline: true>
	| header |
	self assert: (self isCompiledMethod: methodObj).
	header := self fetchPointer: HeaderIndex ofObject: methodObj.
	^(self isIntegerObject: header)
		ifTrue: [header]
		ifFalse:
			[self assert: header asUnsignedInteger < coInterpreter heapBase.
			 self assert: (coInterpreter cCoerceSimple: header to: #'CogMethod *') objectHeader
						= self nullHeaderForMachineCodeMethod.
			(coInterpreter cCoerceSimple: header to: #'CogMethod *') methodHeader]! !

!NewCoObjectMemory methodsFor: 'memory access' stamp: 'eem 4/28/2015 14:46'!
noCheckMethodHeaderOf: methodObj
	"Answer the method header of a CompiledMethod object.
	 If the method has been cogged then the header is a pointer to
	 the CogMethod and the real header will be stored in the CogMethod."
	<inline: true>
	| header |
	header := self fetchPointer: HeaderIndex ofObject: methodObj.
	^(self isIntegerObject: header)
		ifTrue: [header]
		ifFalse:
			[self assert: header asUnsignedInteger < coInterpreter heapBase.
			 self assert: (coInterpreter cCoerceSimple: header to: #'CogMethod *') objectHeader
						= self nullHeaderForMachineCodeMethod.
			(coInterpreter cCoerceSimple: header to: #'CogMethod *') methodHeader]! !

!NewCoObjectMemory methodsFor: 'cog jit support' stamp: 'eem 7/7/2014 08:41'!
noShiftCompactClassIndexOf: oop
	<api>
	^(self baseHeader: oop) bitAnd: 16r1F << self compactClassFieldLSB! !

!NewCoObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 2/28/2009 05:54'!
nullHeaderForMachineCodeMethod
	<api>
	^(1 << 12 "CompactClassIndex 1") + HeaderTypeShort! !

!NewCoObjectMemory methodsFor: 'cog jit support' stamp: 'eem 2/24/2017 09:42'!
receiverTagBitsForMethod: aMethodObj
	"Answer the tag bits for the receiver based on the method's methodClass, if any."
	<api>
	^(coInterpreter methodClassOf: aMethodObj) = self classSmallInteger
		ifTrue: [self smallIntegerTag]
		ifFalse: [0]! !

!NewCoObjectMemory methodsFor: 'become' stamp: 'eem 3/29/2013 15:39'!
restoreHeaderOf: obj to: objHeader
	super restoreHeaderOf: obj to: objHeader.
	(self isCompiledMethodHeader: objHeader) ifTrue:
		[(self asserta: ((coInterpreter methodHasCogMethod: obj) not
						or: [obj = (coInterpreter cogMethodOf: obj) methodObject])) ifFalse:
			[self error: 'attempt to become cogged method']]! !

!NewCoObjectMemory methodsFor: 'trampoline support' stamp: 'eem 5/6/2010 16:00'!
scavengeThresholdAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: scavengeThreshold) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #scavengeThreshold in: self]! !

!NewCoObjectMemory methodsFor: 'cog jit support' stamp: 'eem 8/25/2014 09:19'!
smallIntegerTag
	<api>
	^1! !

!NewCoObjectMemory methodsFor: 'accessing' stamp: 'eem 12/7/2012 14:11'!
startOfMemory
	"Return the start of object memory.  This is immediately after the native code zone.
	 N.B. the stack zone is alloca'ed. Use a macro so as not to punish the debug VM."
	<cmacro: '() heapBase'>
	<returnTypeC: #usqInt>
	^coInterpreter heapBase! !

!NewCoObjectMemory methodsFor: 'simulation' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
validate: oop
	<doNotGenerate>
	| header type cc sz fmt nextChunk | 
	header := self longAt: oop.
	type := header bitAnd: 3.
	type = 2 ifFalse: [type = (self rightType: header) ifFalse: [self halt]].
	sz := (header bitAnd: SizeMask) >> 2.
	nextChunk := oop + ((self isFreeObject: oop)
							ifTrue: [self sizeOfFree: oop]
							ifFalse: [self sizeBitsOf: oop]).
	nextChunk >= freeStart
		ifTrue:
			[nextChunk ~= freeStart ifTrue: [self halt]]
		ifFalse:
			[(self headerType: nextChunk) = 0 ifTrue:
				[(self headerType: (nextChunk + (self wordSize*2))) = 0 ifFalse: [self halt]].
			(self headerType: nextChunk) = 1 ifTrue:
				[(self headerType: (nextChunk + self wordSize)) = 1 ifFalse: [self halt]]].
	type = 2 ifTrue: "free block"
		[^self].
	fmt := self formatOfHeader: header.
	cc := self compactClassIndexOfHeader: header.
	cc > 16 ifTrue: [self halt].	"up to 32 are legal, but not used"
	type = 0 ifTrue:
		["three-word header"
		((self longAt: oop-self wordSize) bitAnd: 3) = type ifFalse: [self halt].
		((self longAt: oop-(self wordSize*2)) bitAnd: 3) = type ifFalse: [self halt].
		((self longAt: oop-self wordSize) = type) ifTrue: [self halt].	"Class word is 0"
		sz = 0 ifFalse: [self halt]].
	type = 1 ifTrue:
		["two-word header"
		((self longAt: oop-self wordSize) bitAnd: 3) = type ifFalse: [self halt].
		cc > 0 ifTrue: [sz = 1 ifFalse: [self halt]].
		sz = 0 ifTrue: [self halt]].
	type = 3 ifTrue:
		["one-word header"
		cc = 0 ifTrue: [self halt]].
	fmt = 5 ifTrue: [self halt].
	fmt = 7 ifTrue: [self halt].
	fmt >= self firstCompiledMethodFormat ifTrue: "must have integer header or be cog method"
		[header := self longAt: oop + self wordSize.
		 ((self isIntegerObject: header)
		  or: [(header bitAnd: 7) = 0
			and: [header asUnsignedInteger < self startOfMemory
			and: [header asUnsignedInteger >= cogit minCogMethodAddress]]]) ifFalse: [self halt]].! !

!NewCoObjectMemory methodsFor: 'cog jit support' stamp: 'eem 11/18/2013 18:55'!
withoutForwardingOn: obj1 and: obj2 with: aBool sendToCogit: selector
	"For the purposes of become: send selector to the cogit with obj1, obj2 and aBool, and
	 answer the result. Undo forwarding for the selector, but redo forwarding after since
	 become:'s restoreHeadersAfter*Become* methods expect to be able to restore."
	<api>
	<var: #selector declareC: 'sqInt (*selector)(sqInt,sqInt,sqInt)'>
	| savedHeaderA savedHeaderB result |
	savedHeaderA := self baseHeader: obj1.
	self baseHeader: obj1 put: (self headerWhileForwardingOf: obj1).
	savedHeaderB := self baseHeader: obj2.
	self baseHeader: obj2 put: (self headerWhileForwardingOf: obj2).

	result := cogit perform: selector with: obj1 with: obj2 with: aBool.

	self baseHeader: obj1 put: savedHeaderA.
	self baseHeader: obj2 put: savedHeaderB.
	^result! !

!NewCoObjectMemory methodsFor: 'trampoline support' stamp: 'eem 5/6/2010 16:00'!
youngStartAddress
	<api>
	<returnTypeC: #usqInt>
	^self cCode: [(self addressOf: youngStart) asUnsignedInteger]
		inSmalltalk: [cogit simulatedReadWriteVariableAddress: #youngStart in: self]! !

NewCoObjectMemory subclass: #NewCoObjectMemorySimulator
	instanceVariableNames: 'parent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!NewCoObjectMemorySimulator class methodsFor: 'instance creation' stamp: 'eem 9/30/2010 13:30'!
new
	^self == NewCoObjectMemorySimulator
		ifTrue: [SmalltalkImage current endianness == #big
				ifTrue: [NewCoObjectMemorySimulatorMSB new]
				ifFalse: [NewCoObjectMemorySimulatorLSB new]]
		ifFalse: [super new]! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'eem 11/6/2014 11:07'!
baseHeaderSize
	^4! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'th 4/15/2000 17:21'!
byteAt: byteAddress
	^self subclassResponsibility! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'th 4/15/2000 17:21'!
byteAt: byteAddress put: byte
	^self subclassResponsibility! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 6/23/2004 14:09'!
byteAtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	pointer is a raw address, and byte is an 8-bit quantity."

	^ self byteAt: pointer! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 6/23/2004 14:13'!
byteAtPointer: pointer put: byteValue
	"This gets implemented by Macros in C, where its types will also be checked.
	pointer is a raw address, and byteValue is an 8-bit quantity."

	^ self byteAt: pointer  put: byteValue! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'eem 11/6/2014 11:07'!
bytesPerOop
	^4! !

!NewCoObjectMemorySimulator methodsFor: 'initialization' stamp: 'eem 10/12/2009 08:42'!
coInterpreter: aCoInterpreter cogit: aCogit
	coInterpreter := aCoInterpreter.
	cogit := aCogit! !

!NewCoObjectMemorySimulator methodsFor: 'simulation only' stamp: 'eem 11/25/2013 13:16'!
cogCodeBase
	^Cogit guardPageSize! !

!NewCoObjectMemorySimulator methodsFor: 'interpreter access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
fetchPointer: fieldIndex ofObject: oop
	"index by word size, and return a pointer as long as the word size"
	self assert: oop >= self startOfMemory.
	self assert: oop + self baseHeaderSize + (fieldIndex << self shiftForWord) < freeStart.
	^super fetchPointer: fieldIndex ofObject: oop! !

!NewCoObjectMemorySimulator methodsFor: 'simulation only' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
firstIndexableField: oop
	"NOTE: overridden from ObjectMemory to add coercion to CArray, so please duplicate any changes"
	| hdr fmt totalLength fixedFields |
	<returnTypeC: #'void *'>
	hdr := self baseHeader: oop.
	fmt := self formatOfHeader: hdr.
	fmt <= 4 ifTrue: "<= 4 pointer"
		["pointer; may need to delve into the class format word"
		totalLength := self lengthOf: oop baseHeader: hdr format: fmt.
		fixedFields := self fixedFieldsOf: oop format: fmt length: totalLength.
		^self cCoerce: (self pointerForOop: oop + self baseHeaderSize + (fixedFields << self shiftForWord)) to: 'oop *'].
	^self
		cCoerce: (self pointerForOop: oop + self baseHeaderSize)
		to: (fmt < 8
				ifTrue: [fmt = 6
						ifTrue: ["32 bit field objects" 'int *']
						ifFalse: ["full word objects (bits)" 'oop *']]
				ifFalse: ["byte objects (including CompiledMethod" 'char *'])! !

!NewCoObjectMemorySimulator methodsFor: 'gc -- mark and sweep' stamp: 'eem 5/30/2018 11:21'!
fullGC
	"If we're /not/ a clone, clone the VM and push it over the cliff.
	 If it survives, destroy the clone and continue.  We should be OK until next time."
	parent ifNil:
		[coInterpreter cr; print: 'GC number '; print: statFullGCs; tab; flush.
		 CloneOnGC ifTrue:
			[coInterpreter cloneSimulation objectMemory fullGC.
			 Smalltalk garbageCollect]].
	^super fullGC! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 7/16/2004 14:56'!
halfWordHighInLong32: long32
	^self subclassResponsibility! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 7/16/2004 14:57'!
halfWordLowInLong32: long32
	^self subclassResponsibility! !

!NewCoObjectMemorySimulator methodsFor: 'debug support' stamp: 'eem 10/11/2014 10:51'!
heapMapAtWord: address
	^heapMap heapMapAtWord: address asInteger! !

!NewCoObjectMemorySimulator methodsFor: 'gc -- mark and sweep' stamp: 'eem 5/30/2018 11:22'!
incrementalGC
	"If we're /not/ a clone, clone the VM and push it over the cliff.
	 If it survives, destroy the clone and continue.  We should be OK until next time."
	(self leakCheckNewSpaceGC
	 and: [parent isNil]) ifTrue:
		[coInterpreter cr; print: 'Incremental GC number '; print: statIncrGCs; tab; flush.
		 CloneOnScavenge ifTrue:
			[coInterpreter cloneSimulation objectMemory incrementalGC.
			 Smalltalk garbageCollect]].
	^super incrementalGC! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 7/3/2004 10:47'!
long32At: byteAddress
	"Return the 32-bit word at byteAddress which must be 0 mod 4."

	^ self longAt: byteAddress! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 7/3/2004 10:47'!
long32At: byteAddress put: a32BitValue
	"Store the 32-bit value at byteAddress which must be 0 mod 4."

	^ self longAt: byteAddress put: a32BitValue! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'eem 8/30/2013 15:53'!
longAt: byteAddress
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	byteAddress \\ 4 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 4 + 1! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'eem 8/30/2013 15:53'!
longAt: byteAddress put: a32BitValue
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	byteAddress \\ 4 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 4 + 1 put: a32BitValue! !

!NewCoObjectMemorySimulator methodsFor: 'debugging traps' stamp: 'eem 2/26/2009 11:35'!
noteAsRoot: oop headerLoc: headerLoc
	"Trap attempts to make cog methods roots."
	self assert: oop >= self startOfMemory.
	^super noteAsRoot: oop headerLoc: headerLoc! !

!NewCoObjectMemorySimulator methodsFor: 'testing' stamp: 'eem 1/16/2013 11:02'!
objectBefore: addr
	| oop prev |
	oop := self firstObject.
	[oop < freeStart] whileTrue:
		[prev := oop.  "look here if debugging prev obj overlapping this one"
		oop := self objectAfter: oop.
		oop >= addr ifTrue: [^ prev]].
	^0! !

!NewCoObjectMemorySimulator methodsFor: 'accessing' stamp: 'eem 10/11/2014 14:26'!
parent

	^ parent! !

!NewCoObjectMemorySimulator methodsFor: 'accessing' stamp: 'eem 10/11/2014 14:26'!
parent: anObject

	parent := anObject! !

!NewCoObjectMemorySimulator methodsFor: 'debug support' stamp: 'eem 3/9/2016 17:48'!
setUpForUseByFacade: aCurrentImageCoInterpreterFacade
	"This is a noop"
	<doNotGenerate>! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 6/23/2004 14:29'!
shortAt: byteAddress
    "Return the half-word at byteAddress which must be even."
	^self subclassResponsibility! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 6/23/2004 14:32'!
shortAt: byteAddress put: a16BitValue
	^ self subclassResponsibility! !

!NewCoObjectMemorySimulator methodsFor: 'interpreter access' stamp: 'eem 10/27/2020 21:44'!
sizeFieldOfAlien: alienObj
	"Answer the first field of alienObj which is assumed to be an Alien of at least 8 bytes"
	<inline: #always>
	^(super sizeFieldOfAlien: alienObj) signedIntFromLong! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'eem 9/25/2013 00:22'!
sqGrowMemory: oldLimit By: delta
	| newMemory |
	coInterpreter transcript show: 'grow memory from ', oldLimit printString, ' by ', delta printString; cr.
	memory size * 4 < (oldLimit + delta) ifTrue:
		[newMemory := (memory class new: oldLimit + delta + 3 // 4).
		 newMemory replaceFrom: 1 to: memory size with: memory startingAt: 1.
		 memory := newMemory].
	^memory size * 4! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'ar 2/25/2001 17:16'!
sqMemoryExtraBytesLeft: includingSwap
	^0! !

!NewCoObjectMemorySimulator methodsFor: 'memory access' stamp: 'eem 10/12/2009 04:27'!
sqShrinkMemory: oldLimit By: delta
	coInterpreter transcript show: 'shrink memory from ', oldLimit printString, ' by ', delta printString, ' remember it doesn''t actually shrink in simulation'; cr.

	^oldLimit! !

!NewCoObjectMemorySimulator methodsFor: 'interpreter access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
storePointer: fieldIndex ofObject: oop withValue: valuePointer
	"Override to ensure acess is within the heap, and within the object"
	| fmt hdr |
	self assert: oop >= self startOfMemory.
	hdr := self baseHeader: oop.
	fmt := self formatOfHeader: hdr.
	self assert: ((fmt <= 4 or: [fmt >= 12])
				and: [fieldIndex >= 0 and: [fieldIndex < (self lengthOf: oop baseHeader: hdr format: fmt)]]).
	self assert: oop + self baseHeaderSize + (fieldIndex << self shiftForWord) < freeStart.
	^super storePointer: fieldIndex ofObject: oop withValue: valuePointer! !

!NewCoObjectMemorySimulator methodsFor: 'interpreter access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
storePointerUnchecked: fieldIndex ofObject: oop withValue: valuePointer
	"Override to ensure acess is within the heap, and within the object"
	| fmt hdr |
	self assert: oop >= self startOfMemory.
	hdr := self baseHeader: oop.
	fmt := self formatOfHeader: hdr.
	self assert: ((fmt <= 4 or: [fmt >= 12])
				and: [fieldIndex >= 0 and: [fieldIndex < (self lengthOf: oop baseHeader: hdr format: fmt)]]).
	self assert: oop + self baseHeaderSize + (fieldIndex << self shiftForWord) < freeStart.
	^super storePointerUnchecked: fieldIndex ofObject: oop withValue: valuePointer! !

!NewCoObjectMemorySimulator methodsFor: 'debug support' stamp: 'eem 12/19/2012 14:41'!
tenuringIncrementalGC
	coInterpreter transcript cr; nextPutAll: 'tenuringIncrementalGC ('; print: coInterpreter byteCount; space; print: freeStart; nextPut: $); flush.
	^super tenuringIncrementalGC! !

!NewCoObjectMemorySimulator methodsFor: 'testing' stamp: 'eem 1/16/2013 11:02'!
validOop: oop
	" Return true if oop appears to be valid "
	(oop bitAnd: 1) = 1 ifTrue: [^ true].  "Integer"
	(oop bitAnd: 3) = 0 ifFalse: [^ false].  "Uneven address"
	oop >= freeStart ifTrue: [^ false].  "Out of range"
	"could test if within the first large freeblock"
	(self longAt: oop) = 4 ifTrue: [^ false].
	(self headerType: oop) = 2 ifTrue: [^ false].	"Free object"
	^ true! !

NewCoObjectMemorySimulator subclass: #NewCoObjectMemorySimulatorLSB
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'nk 4/3/2004 06:58'!
byteAt: byteAddress
	| lowBits long |
	lowBits := byteAddress bitAnd: 3.
	long := self longAt: byteAddress - lowBits.
	^(lowBits caseOf: {
		[0] -> [ long ].
		[1] -> [ long bitShift: -8  ].
		[2] -> [ long bitShift: -16 ].
		[3] -> [ long bitShift: -24 ]
	}) bitAnd: 16rFF
! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'eem 7/18/2013 17:36'!
byteAt: byteAddress put: byte
	| lowBits long longAddress |
	lowBits := byteAddress bitAnd: 3.
	longAddress := byteAddress - lowBits.
	long := self longAt: longAddress.
	long := (lowBits caseOf: {
		[0] -> [ (long bitAnd: 16rFFFFFF00) bitOr: byte ].
		[1] -> [ (long bitAnd: 16rFFFF00FF) bitOr: (byte bitShift: 8) ].
		[2] -> [ (long bitAnd: 16rFF00FFFF) bitOr: (byte bitShift: 16)  ].
		[3] -> [ (long bitAnd: 16r00FFFFFF) bitOr: (byte bitShift: 24)  ]
	}).
	self assert: longAddress < freeStart.
	self longAt: longAddress put: long.
	^byte! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'nice 4/3/2016 08:36'!
byteSwapped32IfBigEndian: w
	"Accelerate, we now we are not big endian"
	^w bitAnd: 16rFFFFFFFF! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'nice 4/3/2016 08:36'!
byteSwapped64IfBigEndian: w
	"Accelerate, we now we are not big endian"
	^w bitAnd: 16rFFFFFFFFFFFFFFFF! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'eem 6/25/2011 14:03'!
endianness
	^#little! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'float primitives' stamp: 'eem 4/20/2009 14:37'!
fetchFloatAt: floatBitsAddress into: aFloat
	aFloat at: 2 put: (self long32At: floatBitsAddress).
	aFloat at: 1 put: (self long32At: floatBitsAddress+4)! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:59'!
halfWordHighInLong32: long32
	"Used by Balloon"

	^ long32 bitAnd: 16rFFFF! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:59'!
halfWordLowInLong32: long32
	"Used by Balloon"

	^ long32 bitShift: -16! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'cb 4/27/2015 13:22'!
intAt: byteAddress put: a32BitValue
	^self longAt: byteAddress put: (a32BitValue bitAnd: 16rFFFFFFFF)! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'eem 4/14/2015 18:36'!
long64At: byteAddress
	"memory is a Bitmap, a 32-bit indexable array of bits"
	| hiWord loWord |
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	loWord := memory at: byteAddress // 4 + 1.
	hiWord := memory at: byteAddress // 4 + 2.
	^hiWord = 0
		ifTrue: [loWord]
		ifFalse: [(hiWord bitShift: 32) + loWord]! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'eem 4/14/2015 18:37'!
long64At: byteAddress put: a64BitValue
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	self
		longAt: byteAddress put: (a64BitValue bitAnd: 16rffffffff);
		longAt: byteAddress + 4 put: a64BitValue >> 32.
	^a64BitValue! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'initialization' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
nextLongFrom: aStream
	"Read a 32- or 64-bit quantity from the given (binary) stream."

	^ aStream nextLittleEndianNumber: self wordSize! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'initialization' stamp: 'eem 10/9/2008 16:39'!
nextShortFrom: aStream
	"Read a 16-bit quantity from the given (binary) stream."
	^aStream nextLittleEndianNumber: 2! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'initialization' stamp: 'eem 11/18/2014 13:47'!
nextWord32From: aStream
	"Read a 32-bit quantity from the given (binary) stream."

	^aStream nextLittleEndianNumber: 4! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'di 6/23/2004 14:29'!
shortAt: byteAddress
    "Return the half-word at byteAddress which must be even."
	| lowBits long |
	lowBits := byteAddress bitAnd: 2.
	long := self longAt: byteAddress - lowBits.
	^ lowBits = 2
		ifTrue: [ long bitShift: -16 ]
		ifFalse: [ long bitAnd: 16rFFFF ].
! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'di 6/23/2004 14:31'!
shortAt: byteAddress put: a16BitValue
    "Return the half-word at byteAddress which must be even."
	| lowBits long longAddress |
	lowBits := byteAddress bitAnd: 2.
	lowBits = 0
		ifTrue:
		[ "storing into LS word"
		long := self longAt: byteAddress.
		self longAt: byteAddress
				put: ((long bitAnd: 16rFFFF0000) bitOr: a16BitValue)
		]
		ifFalse:
		[longAddress := byteAddress - 2.
		long := self longAt: longAddress.
		self longAt: longAddress
				put: ((long bitAnd: 16rFFFF) bitOr: (a16BitValue bitShift: 16))
		]! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'float primitives' stamp: 'eem 4/20/2009 14:37'!
storeFloatAt: floatBitsAddress from: aFloat

	self long32At: floatBitsAddress put: (aFloat at: 2).
	self long32At: floatBitsAddress+4 put: (aFloat at: 1)! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'eem 6/25/2011 14:28'!
unalignedLongAt: byteAddress
	| rem |
	rem := byteAddress \\ 4.
	^rem = 0
		ifTrue: [self longAt: byteAddress]
		ifFalse: [((self longAt: byteAddress - rem) + ((self longAt: byteAddress - rem + 4) bitShift: 32) bitShift: rem * -8) bitAnd: 16rFFFFFFFF]! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'eem 6/25/2011 14:16'!
unalignedLongAt: byteAddress put: aLong
	| rem mask |
	rem := byteAddress \\ 4.
	^rem = 0
		ifTrue: [self longAt: byteAddress put: aLong]
		ifFalse:
			[mask := 16rFFFFFFFF bitAnd: (-1 bitShift: rem * 8).
			 self longAt: byteAddress - rem
				put: ((self longAt: byteAddress - rem) bitAnd: mask bitInvert)
					+ ((aLong bitShift: rem * 8) bitAnd: mask).
			 self longAt: byteAddress - rem + 4
				put: ((self longAt: byteAddress - rem + 4) bitAnd: mask)
					+ ((aLong bitShift: 4 - rem * -8) bitAnd: mask bitInvert).
			 aLong]! !

!NewCoObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'tpr 12/22/2005 17:46'!
vmEndianness
	"return 0 for little endian, 1 for big endian"
	^0! !

NewCoObjectMemorySimulator subclass: #NewCoObjectMemorySimulatorMSB
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!NewCoObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
byteAt: byteAddress
	| lowBits bpwMinus1 |
	bpwMinus1 := self wordSize-1.
	lowBits := byteAddress bitAnd: bpwMinus1.
	^ ((self longAt: byteAddress - lowBits)
		bitShift: (lowBits - bpwMinus1) * 8)
		bitAnd: 16rFF! !

!NewCoObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
byteAt: byteAddress put: byte
	| longWord shift lowBits bpwMinus1 longAddress |
	bpwMinus1 := self wordSize-1.
	lowBits := byteAddress bitAnd: bpwMinus1.
	longAddress := byteAddress - lowBits.
	longWord := self longAt: longAddress.
	shift := (bpwMinus1 - lowBits) * 8.
	longWord := longWord
				- (longWord bitAnd: (16rFF bitShift: shift))
				+ (byte bitShift: shift).
	self assert: longAddress < freeStart.
	self longAt: longAddress put: longWord.
	^byte! !

!NewCoObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'nice 4/3/2016 08:38'!
byteSwapped32IfBigEndian: w
	"Accelerate, we now we are big endian"
	^w byteSwap32! !

!NewCoObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'nice 4/3/2016 08:37'!
byteSwapped64IfBigEndian: w
	"Accelerate, we now we are big endian"
	^w byteSwap64! !

!NewCoObjectMemorySimulatorMSB methodsFor: 'float primitives' stamp: 'eem 10/12/2009 03:55'!
fetchFloatAt: floatBitsAddress into: aFloat
	aFloat at: 1 put: (self long32At: floatBitsAddress).
	aFloat at: 2 put: (self long32At: floatBitsAddress+4)! !

!NewCoObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:58'!
halfWordHighInLong32: long32
	"Used by Balloon"

	^ long32 bitShift: -16! !

!NewCoObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:58'!
halfWordLowInLong32: long32
	"Used by Balloon"

	^ long32 bitAnd: 16rFFFF! !

!NewCoObjectMemorySimulatorMSB methodsFor: 'image save/restore' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
nextLongFrom: aStream
	"Read a 32- or 64-bit quantity from the given (binary) stream."

	^ aStream nextNumber: self wordSize! !

!NewCoObjectMemorySimulatorMSB methodsFor: 'image save/restore' stamp: 'eem 10/12/2009 03:59'!
nextShortFrom: aStream
	"Read a 16-bit quantity from the given (binary) stream."
	^aStream nextNumber: 2! !

!NewCoObjectMemorySimulatorMSB methodsFor: 'image save/restore' stamp: 'eem 11/18/2014 13:47'!
nextWord32From: aStream
	"Read a 32-bit quantity from the given (binary) stream."

	^aStream nextNumber: 4! !

!NewCoObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
shortAt: byteAddress
    "Return the half-word at byteAddress which must be even."
	| lowBits bpwMinus2 |
	bpwMinus2 := self wordSize-2.
	lowBits := byteAddress bitAnd: bpwMinus2.
	^ ((self longAt: byteAddress - lowBits)
		bitShift: (lowBits - bpwMinus2) * 8)
		bitAnd: 16rFFFF
! !

!NewCoObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
shortAt: byteAddress put: a16BitValue
    "Return the half-word at byteAddress which must be even."
	| longWord shift lowBits bpwMinus2 longAddress |
	bpwMinus2 := self wordSize-2.
	lowBits := byteAddress bitAnd: bpwMinus2.
	longAddress := byteAddress - lowBits.
	longWord := self longAt: longAddress.
	shift := (bpwMinus2 - lowBits) * 8.
	longWord := longWord
				- (longWord bitAnd: (16rFFFF bitShift: shift))
				+ (a16BitValue bitShift: shift).
	self longAt: longAddress put: longWord
! !

!NewCoObjectMemorySimulatorMSB methodsFor: 'float primitives' stamp: 'eem 10/12/2009 04:05'!
storeFloatAt: floatBitsAddress from: aFloat

	self long32At: floatBitsAddress put: (aFloat at: 1).
	self long32At: floatBitsAddress+4 put: (aFloat at: 2)! !

!NewCoObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'tpr 12/22/2005 17:46'!
vmEndianness
	"return 0 for little endian, 1 for big endian"
	^1! !

!NewObjectMemory class methodsFor: 'translation' stamp: 'eem 12/10/2009 13:37'!
declareCVarsIn: aCCodeGenerator
	self declareCAsOop: #(freeStart reserveStart scavengeThreshold)
		in: aCCodeGenerator.
	aCCodeGenerator var: #statGCEndUsecs type: #usqLong! !

!NewObjectMemory class methodsFor: 'class initialization' stamp: 'eem 8/15/2013 17:07'!
initialize
	"This is for Monticello package loading"
	self initializeWithOptions: Dictionary new! !

!NewObjectMemory class methodsFor: 'translation' stamp: 'eem 5/20/2009 14:27'!
mustBeGlobal: var
	"Answer if a variable must be global and exported.  Used for inst vars that are accessed from VM support code."

	^(super mustBeGlobal: var)
	   or: ['checkForLeaks' = var]! !

!NewObjectMemory class methodsFor: 'accessing class hierarchy' stamp: 'nice 5/13/2016 22:11'!
objectRepresentationClass
	^CogObjectRepresentationForSqueakV3! !

!NewObjectMemory class methodsFor: 'translation' stamp: 'eem 5/1/2014 08:42'!
prepareToBeAddedToCodeGenerator: aCodeGen
	"Remove the methods and instance variables we don't use."

	self ~~ NewObjectMemory ifTrue: "No Need to repeat in NewCoObjectMemory"
		[^self].

	aCodeGen
		removeMethodForSelector: #markPhase; "we implement markPhase:"
		removeMethodForSelector: #printWronglySizedContexts. "we implement printWronglySizedContexts:"
	"This class uses freeStart in place of freeBlock.  It does
	 not maintain an allocationCount nor stats there-of.
	 Having an interpreter that uses a stack zone, it doesn't
	 need an optimized context allocator."
	aCodeGen
		removeVariable: 'freeBlock';
		removeVariable: 'allocationCount';
		removeVariable: 'allocationsBetweenGCs';
		removeVariable: 'statAllocationCount';
		removeVariable: 'freeContexts';
		removeVariable: 'freeLargeContexts';
		removeVariable: 'statGCEndTime' "replaced by statGCEndUsecs"! !

!NewObjectMemory class methodsFor: 'simulation' stamp: 'eem 9/30/2010 13:30'!
simulatorClass
	^NewObjectMemorySimulator! !

!NewObjectMemory class methodsFor: 'api chacterization' stamp: 'nice 12/26/2020 23:45'!
vmProxyMinorVersion
	"Define the  VM_PROXY_MAJOR version for this VM as used to define
	 the api in platforms/Cross/vm/sqVirtualMachine.[ch] and in interp.h."
	^17! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 12/21/2017 15:00'!
ISA
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter ISA! !

!NewObjectMemory methodsFor: 'object enumeration' stamp: 'eem 7/11/2008 15:47'!
accessibleObjectAfter: oop 
	"Return the accessible object following the given object or 
	free chunk in the heap. Return nil when heap is exhausted."
	| obj |
	obj := self objectAfter: oop.
	[self oop: obj isLessThan: freeStart] whileTrue:
		[(self isFreeObject: obj) ifFalse: [^obj].
		 obj := self objectAfter: obj].
	^nil! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem 4/27/2009 16:25'!
addressCouldBeObj: address
	<api>
	"Answer if address appears to be that of an object, which implies it is
	 safe to fetch the class and size. For code disassembly and assertions."
	^(address bitAnd: 3) = 0
	  and: [address asUnsignedInteger >= self startOfMemory
	  and: [address asUnsignedInteger < freeStart
	  and: [(self headerType: address) ~= HeaderTypeGC]]]! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem 6/1/2009 14:20'!
addressCouldBeObjWhileForwarding: address
	"Answer if address appears to be that of an object, which implies it is
	 safe to fetch the class and size. For code disassembly and assertions."
	^(address bitAnd: 3) = 0
	  and: [address asUnsignedInteger >= self startOfMemory
	  and: [address asUnsignedInteger < freeStart]]! !

!NewObjectMemory methodsFor: 'initialization' stamp: 'eem 9/12/2013 09:08'!
adjustAllOopsBy: bytesToShift 
	"Adjust all oop references by the given number of bytes. This is
	 done just after reading in an image when the new base address
	 of the object heap is different from the base address in the image."
	"di 11/18/2000 - return number of objects found"

	| oop nextOop totalObjects |
	<inline: false>
	bytesToShift = 0 ifTrue: [^300000].
	"this is probably an improvement over the previous answer of nil,
	 but maybe we should do the obejct counting loop and simply
	 guard the adjustFieldsAndClass... with a bytesToShift = 0 ifFalse: ?"
	totalObjects := 0.
	oop := self firstObject.
	[self oop: oop isLessThan: freeStart] whileTrue:
		[(self isFreeObject: oop) ifFalse:
			[totalObjects := totalObjects + 1.
			 self adjustFieldsAndClassOf: oop by: bytesToShift].
		 nextOop := self objectAfter: oop.
		 self assert: (self oop: oop isLessThan: nextOop).
		 oop := nextOop].
	^totalObjects! !

!NewObjectMemory methodsFor: 'primitive support' stamp: 'eem 3/18/2020 12:58'!
allInstancesOf: aBehavior
	"Attempt to answer all instances of aBehavior, failing if there is not enough room."
	| count container fillPointer obj byteSize afterPreAllocatedObject |
	"Allocate a large header Array of sufficient size to require a large header.
	 Reset its size later."
	container := self instantiateClass: (self splObj: ClassArray) indexableSize: self minLargeHeaderSize.
	self sizeHeader: container putBodySize: 0.
	afterPreAllocatedObject := freeStart.
	freeStart := fillPointer := (self firstFixedField: container) asInteger.
	count := 0.
	obj := self firstObject.
	[self oop: obj isLessThan: container] whileTrue:
		[(self isFreeObject: obj) ifFalse:
			[(self fetchClassOfNonImm: obj) = aBehavior ifTrue:
				[count := count + 1.
				 fillPointer < reserveStart ifTrue:
					[self longAt: fillPointer put: obj.
					 fillPointer := fillPointer + self bytesPerOop]]].
		 obj := self accessibleObjectAfter: obj].
	fillPointer >= reserveStart ifTrue: "didn't fit; answer instance count so large enough container can be allocated."
		[^self integerObjectOf: count].
	byteSize := fillPointer - (self firstFixedField: container) asInteger.
	self sizeHeader: container putBodySize: byteSize.
	freeStart := fillPointer.
	^container! !

!NewObjectMemory methodsFor: 'object enumeration' stamp: 'eem 10/13/2014 16:01'!
allObjectsDo: aBlock
	<inline: true>
	| oop |
	oop := self firstObject.
	[oop asUnsignedInteger < freeStart] whileTrue:
		[(self isFreeObject: oop) ifFalse:
			[aBlock value: oop].
		 oop := self objectAfter: oop]! !

!NewObjectMemory methodsFor: 'object enumeration' stamp: 'eem 10/13/2014 16:13'!
allObjectsDoSafely: aBlock
	<inline: true>
	| oop |
	oop := self firstObject.
	[oop asUnsignedInteger < freeStart] whileTrue:
		[(self isFreeObject: oop) ifFalse:
			[aBlock value: oop].
		 oop := self objectAfterWhileForwarding: oop]! !

!NewObjectMemory methodsFor: 'primitive support' stamp: 'eem 3/27/2014 15:34'!
allObjectsUnmarked
	self allObjectsDo:
		[:o| (self isMarked: o) ifTrue: [^false]].
	^true! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
allocate: byteSize headerSize: hdrSize h1: baseHeader h2: classOopArg h3: extendedSize doFill: doFill format: format
	"Allocate a new object of the given size and number of header words. (Note: byteSize already includes space for the base header word.) Initialize the header fields of the new object and fill the remainder of the object with a value appropriate for the format.
	May cause a GC"

	| newObj classOop |
	<inline: true>
	<var: #i type: #usqInt>
	<var: #end type: #usqInt>
	newObj := self allocateChunk: byteSize + (hdrSize - 1 * self wordSize).
	newObj = 0
		ifTrue:
			["remap classOop because GC may move the classOop"
			hdrSize > 1 ifTrue: [self pushRemappableOop: classOopArg].
			newObj := self allocateChunkAfterGC: byteSize + (hdrSize - 1 * self wordSize).
			hdrSize > 1 ifTrue: [classOop := self popRemappableOop].
			newObj = 0 ifTrue: [^newObj]]
		ifFalse: [classOop := classOopArg].

	hdrSize = 3 ifTrue:
		[self longAt: newObj put: (extendedSize bitOr: HeaderTypeSizeAndClass).
		 self longAt: newObj + self wordSize put: (classOop bitOr: HeaderTypeSizeAndClass).
		 self longAt: newObj + (self wordSize*2) put: (baseHeader bitOr: HeaderTypeSizeAndClass).
		 newObj := newObj + (self wordSize*2)].

	hdrSize = 2 ifTrue:
		[self longAt: newObj put: (classOop bitOr: HeaderTypeClass).
		 self longAt: newObj + self wordSize put: (baseHeader bitOr: HeaderTypeClass).
		 newObj := newObj + self wordSize].

	hdrSize = 1 ifTrue:
		[self longAt: newObj put: (baseHeader bitOr: HeaderTypeShort)].

	"clear new object"
	doFill ifTrue:
		[| fillWord end i |
		 fillWord := format <= self lastPointerFormat
						ifTrue: [nilObj] "if pointers, fill with nil oop"
						ifFalse: [0].
		 end := newObj + byteSize.
		 i := newObj + self wordSize. "skip header"
		 [i < end] whileTrue:
			[self longAt: i put: fillWord.
			 i := i + self wordSize].
		 self assert: i = freeStart.].
	DoExpensiveAssertionChecks ifTrue:
		[self okayOop: newObj.
		 self oopHasOkayClass: newObj.
		 (self safeObjectAfter: newObj) = freeStart ifFalse:
			[self error: 'allocate bug: did not set header of new oop correctly']].

	^newObj! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem 6/16/2014 12:30'!
allocateChunk: byteSize 
	"Allocate a chunk of the given size. Sender must be sure that the requested size
	 includes enough space for the header word(s).  This version is for normal allocations
	 and refuses to allocate beyond the interpreter's reserveStart.  If the allocation takes
	 freeStart over the scavenge threshold schedule a garbage collection.  If this returns 0
	 the client should prepare for garbage collection and retry using allocateChunkAfterGC:"
	| newChunk newFreeStart |
	<inline: true>
	<var: #newChunk type: #usqInt>
	<var: #newFreeStart type: #usqInt>
	newChunk := freeStart.
	newFreeStart := freeStart + byteSize.
	newFreeStart < scavengeThreshold ifTrue:
		[freeStart := newFreeStart.
		 ^self oopForPointer: newChunk].

	self scheduleIncrementalGC.
	freeStart <= reserveStart ifTrue:
		[freeStart := newFreeStart.
		 ^self oopForPointer: newChunk].

	^0! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem 6/6/2013 11:36'!
allocateChunkAfterGC: byteSize 
	"Garbage collect and then allocate a chunk of the given size. Sender must be sure
	 that the requested size includes enough space for the header word(s)."
	| newChunk enoughSpace |
	<inline: true>
	<var: #newChunk type: #usqInt>
	enoughSpace := self sufficientSpaceToAllocate: byteSize.
	enoughSpace ifFalse:
		["signal that space is running low, but proceed with allocation if possible"
		 self setSignalLowSpaceFlagAndSaveProcess].
	(self oop: freeStart + byteSize isGreaterThan: reserveStart) ifTrue:
		[^0 "Allocation failed.  Client should e.g. fail the primtive"].

	"if we get here, there is enough space for allocation to succeed "
	newChunk := freeStart.
	freeStart := freeStart + byteSize.
	^self oopForPointer: newChunk! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem 11/10/2014 12:49'!
allocateInterpreterChunk: byteSize 
	"Allocate a chunk of the given size. Sender must be sure that the requested size
	 includes enough space for the header word(s).  This version is for interpreter
	 allocations and will allocate beyond the interpreter's reserveStart.  If the allocation
	 takes freeStart over the scavenge threshold schedule a garbage collection."
	| newChunk newFreeStart |
	<inline: true>
	<var: #newChunk type: #usqInt>
	<var: #newFreeStart type: #usqInt>

	newChunk := freeStart.
	newFreeStart := freeStart + byteSize.
	newFreeStart < scavengeThreshold ifTrue:
		[freeStart := newFreeStart.
		 ^self oopForPointer: newChunk].

	"Don't thrash doing collections when over the scavengeThreshold.
	 Only schedule an incrementalGC if this allocation took us over the threshold."
	freeStart < scavengeThreshold ifTrue:
		[self scheduleIncrementalGC].

	newFreeStart < reserveStart ifTrue:
		[freeStart := newFreeStart.
		 ^self oopForPointer: newChunk].

	"space is low.  A scavenge may reclaim sufficient space and this may be a
	 false alarm.  We actually check for low space after the incremental collection.
	 But we really do need to do a scavenge promptly, if only to check for low
	 space.  We cannot do a garbage collect now without moving pointers under
	 the VM's feet, which is too error-prone and inefficient to contemplate."

	self scheduleIncrementalGC.

	freeStart <= endOfMemory ifTrue:
		[freeStart := newFreeStart.
		 ^self oopForPointer: newChunk].

	self error: 'out of memory'.
	^nil! !

!NewObjectMemory methodsFor: 'simulation' stamp: 'eem 11/19/2016 09:21'!
allocateMemoryOfSize: limit
	<doNotGenerate>
	super allocateMemoryOfSize: limit.
	freeStart := memoryLimit! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 12/12/2013 11:29'!
arrayValueOf: arrayOop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter arrayValueOf: arrayOop! !

!NewObjectMemory methodsFor: 'snapshot' stamp: 'eem 10/8/2013 15:44'!
baseAddressOfImage
	"Answer the base address of the image data written to a sapshot."
	^self startOfMemory! !

!NewObjectMemory methodsFor: 'become' stamp: 'eem 2/7/2015 10:28'!
become: array1 with: array2 twoWay: twoWayFlag copyHash: copyHashFlag
	"All references to each object in array1 are swapped with all references to the corresponding object in array2. That is, all pointers to one object are replaced with with pointers to the other. The arguments must be arrays of the same length. 
	Returns PrimNoErr if the primitive succeeds."
	"Implementation: Uses forwarding blocks to update references as done in compaction."
	| start |
	self runLeakCheckerFor: GCModeBecome.
	(self isArray: array1) ifFalse:
		[^PrimErrBadReceiver].
	((self isArray: array2)
	 and: [(self lastPointerOf: array1) = (self lastPointerOf: array2)]) ifFalse:
		[^PrimErrBadArgument].
	(twoWayFlag or: [copyHashFlag])
		ifTrue: [(self containOnlyOops: array1 and: array2) ifFalse: [^PrimErrInappropriate]]
		ifFalse: [(self containOnlyOops: array1) ifFalse: [^PrimErrInappropriate]].

	(self prepareForwardingTableForBecoming: array1 with: array2 twoWay: twoWayFlag) ifFalse:
		[^PrimErrNoMemory]. "fail; not enough space for forwarding table"

	(self allYoung: array1 and: array2)
		ifTrue: [start := youngStart"sweep only the young objects plus the roots"]
		ifFalse: [start := self startOfMemory"sweep all objects"].
	coInterpreter preBecomeAction.
	self mapPointersInObjectsFrom: start to: freeStart.
	twoWayFlag
		ifTrue: [self restoreHeadersAfterBecoming: array1 with: array2]
		ifFalse: [self restoreHeadersAfterForwardBecome: copyHashFlag].
	coInterpreter postBecomeAction: 0.

	self initializeMemoryFirstFree: freeStart. "re-initialize memory used for forwarding table"
	self runLeakCheckerFor: GCModeBecome.

	^PrimNoErr "success"! !

!NewObjectMemory methodsFor: 'garbage collection' stamp: 'eem 11/10/2014 12:49'!
biasToGrow
	self growObjectMemory: (growHeadroom*3/2) - self freeSize! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/19/2013 11:08'!
booleanValueOf: obj
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter booleanValueOf: obj! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem 12/15/2008 11:18'!
bytesLeft: includingSwap
	^self freeSize + (self sqMemoryExtraBytesLeft: includingSwap)! !

!NewObjectMemory methodsFor: 'primitive support' stamp: 'eem 4/11/2014 10:55'!
characterValueOf: oop
	<api>
	(self isCharacterObject: oop) ifFalse:
		[coInterpreter primitiveFailFor: PrimErrUnsupported.
		 ^-1].
	^self integerValueOf: (self fetchPointer: 0 ofObject: oop)! !

!NewObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
checkForLeaks
	^checkForLeaks! !

!NewObjectMemory methodsFor: 'memory access' stamp: 'eem 09/05/2021 19:12'!
checkHeapIntegrity
	"Perform an integrity/leak check using the heapMap.  Assume
	 clearLeakMapAndMapAccessibleObjects has set a bit at each
	 object's header.  Scan all objects in the heap checking that every
	 pointer points to a header.  Scan the rootTable, remapBuffer and
	 extraRootTable checking that every entry is a pointer to a header.
	 Check that the number of roots is correct and that all rootTable
	 entries have their rootBit set. Answer if all checks pass."
	| ok obj sz hdr fmt fi fieldOop numRootsInHeap |
	<inline: false>
	ok := true.
	numRootsInHeap := 0.
	obj := self firstObject.
	[self oop: obj isLessThan: self startOfFreeSpace] whileTrue:
		[(self isFreeObject: obj)
			ifTrue:
				[sz := self sizeOfFree: obj]
			ifFalse:
				[hdr := self baseHeader: obj.
				 (self isYoungRootHeader: hdr) ifTrue:
					[numRootsInHeap := numRootsInHeap + 1].
				 (self compactClassIndexOfHeader: hdr) = 0 ifTrue:
					[fieldOop := (self classHeader: obj) bitAnd: AllButTypeMask.
					 ((self isIntegerObject: fieldOop)
					   or: [(heapMap heapMapAtWord: (self pointerForOop: fieldOop)) = 0]) ifTrue:
						[self print: 'object leak in '; printHex: obj; print: ' class = '; printHex: fieldOop; eekcr.
						 ok := false]].
				 fmt := self formatOfHeader: hdr.
				 (fmt <= 4 "pointers" or: [fmt >= 12 "compiled method"]) ifTrue:
					[fmt >= 12
						ifTrue: [fi := (self literalCountOf: obj) + LiteralStart]
						ifFalse: [(fmt = 3 and: [self isContextHeader: hdr])
									ifTrue: [fi := CtxtTempFrameStart + (coInterpreter fetchStackPointerOf: obj)]
									ifFalse: [fi := self lengthOf: obj]].
					[(fi := fi - 1) >= 0] whileTrue:
						[fieldOop := self fetchPointer: fi ofObject: obj.
						 (self isNonIntegerObject: fieldOop) ifTrue:
							[(fieldOop bitAnd: self wordSize - 1) ~= 0
								ifTrue:
									[self print: 'misaligned oop in '; printHex: obj; print: ' @ '; printNum: fi; print: ' = '; printHex: fieldOop; eekcr.
									 ok := false]
								ifFalse:
									[(heapMap heapMapAtWord: (self pointerForOop: fieldOop)) = 0 ifTrue:
										[self print: 'object leak in '; printHex: obj; print: ' @ '; printNum: fi; print: ' = '; printHex: fieldOop; eekcr.
										 ok := false]]]]].
				 sz := self sizeBitsOf: obj].
		 obj := self oopFromChunk: obj + sz].
	numRootsInHeap ~= rootTableCount ifTrue:
		[self print: 'root count mismatch. #heap roots '; printNum: numRootsInHeap; print: '; #roots '; printNum: rootTableCount; cr.
		"But the system copes with overflow..."
		ok := rootTableOverflowed and: [needGCFlag]].
	1 to: rootTableCount do:
		[:ri|
		obj := rootTable at: ri.
		(obj bitAnd: self wordSize - 1) ~= 0
			ifTrue:
				[self print: 'misaligned oop in rootTable @ '; printNum: ri; print: ' = '; printHex: obj; eekcr.
				 ok := false]
			ifFalse:
				[(heapMap heapMapAtWord: (self pointerForOop: obj)) = 0
					ifTrue:
						[self print: 'object leak in rootTable @ '; printNum: ri; print: ' = '; printHex: obj; eekcr.
						 ok := false]
					ifFalse:
						[hdr := self baseHeader: obj.
						 (self isYoungRootHeader: hdr) ifFalse:
							[self print: 'non-root in rootTable @ '; printNum: ri; print: ' = '; printHex: obj; eekcr.
							 ok := false]]]].
	1 to: remapBufferCount do:
		[:ri|
		obj := remapBuffer at: ri.
		(obj bitAnd: self wordSize - 1) ~= 0
			ifTrue:
				[self print: 'misaligned remapRoot @ '; printNum: ri; print: ' = '; printHex: obj; eekcr.
				 ok := false]
			ifFalse:
				[(heapMap heapMapAtWord: (self pointerForOop: obj)) = 0
					ifTrue:
						[self print: 'object leak in remapRoots @ '; printNum: ri; print: ' = '; printHex: obj; eekcr.
						 ok := false]]].
	1 to: extraRootCount do:
		[:ri|
		obj := (extraRoots at: ri) at: 0.
		(obj bitAnd: self wordSize - 1) ~= 0
			ifTrue:
				[self print: 'misaligned extraRoot @ '; printNum: ri; print: ' => '; printHex: obj; eekcr.
				 ok := false]
			ifFalse:
				[(heapMap heapMapAtWord: (self pointerForOop: obj)) = 0
					ifTrue:
						[self print: 'object leak in extraRoots @ '; printNum: ri; print: ' => '; printHex: obj; eekcr.
						 ok := false]]].
	^ok! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
checkOkayOop: oop
	"Verify that the given oop is legitimate. Check address, header, and size but not class.
	 Answer true if OK.  Otherwise print reason and answer false."

	<api>
	<var: #oop type: #usqInt>
	| sz type fmt unusedBit |

	"address and size checks"
	(self isIntegerObject: oop) ifTrue: [ ^true ].
	((self oop: oop isGreaterThanOrEqualTo: self startOfMemory) and: [self oop: oop isLessThan: freeStart])
		ifFalse: [ self print: 'oop '; printHex: oop; print: ' is not a valid address'; cr. ^false ].
	((oop \\ self wordSize) = 0)
		ifFalse: [ self print: 'oop '; printHex: oop; print: ' is not a word-aligned address'; cr. ^false ].
	sz := self sizeBitsOf: oop.
	(self oop: oop + sz isLessThanOrEqualTo: freeStart)
		ifFalse: [ self print: 'oop '; printHex: oop; print: ' size would make it extend beyond the end of memory'; cr. ^false ].

	"header type checks"
	type := self headerType: oop.
	type = HeaderTypeFree
		ifTrue:  [ self print: 'oop '; printHex: oop; print: ' is a free chunk, not an object'; cr. ^false ].
	type = HeaderTypeShort ifTrue: [
		(self compactClassIndexOf: oop) = 0
			ifTrue:  [ self print: 'oop '; printHex: oop; print: ' cannot have zero compact class field in a short header'; cr. ^false ].
	].
	type = HeaderTypeClass ifTrue: [
		((oop >= self wordSize) and: [(self headerType: oop - self wordSize) = type])
			ifFalse: [ self print: 'oop '; printHex: oop; print: ' class header word has wrong type'; cr. ^false ].
	].
	type = HeaderTypeSizeAndClass ifTrue: [
		((oop >= (self wordSize*2)) and:
		 [(self headerType: oop - (self wordSize*2)) = type and:
		 [(self headerType: oop - self wordSize) = type]])
			ifFalse: [ self print: 'oop '; printHex: oop; print: ' class header word has wrong type'; cr. ^false ].
	].

	"format check"
	fmt := self formatOf: oop.
	((fmt = 5) | (fmt = 7))
		ifTrue:  [ self print: 'oop '; printHex: oop; print: ' has an unknown format type'; cr. ^false ].

	"mark and root bit checks"
	unusedBit := 16r20000000.
	self wordSize = 8
		ifTrue:
			[unusedBit := unusedBit << 16.
			 unusedBit := unusedBit << 16].
	((self longAt: oop) bitAnd: unusedBit) = 0
		ifFalse: [ self print: 'oop '; printHex: oop; print: ' unused header bit 30 is set; should be zero'; cr. ^false ].
"xxx
	((self longAt: oop) bitAnd: MarkBit) = 0
		ifFalse: [ self print: 'mark bit should not be set except during GC' ].
xxx"
	((self isYoungRoot: oop) and: [oop >= youngStart])
		ifTrue: [ self print: 'oop '; printHex: oop; print: ' root bit is set in a young object'; cr. ^false ].
	^true
! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem 10/11/2014 10:48'!
checkOopIntegrity: obj named: name
	<inline: false>
	<var: #name type: #'char *'>
	(heapMap heapMapAtWord: (self pointerForOop: obj)) ~= 0 ifTrue:
		[^true].
	coInterpreter print: name; print: ' leak '; printHex: obj; cr.
	^false! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 10/11/2013 16:02'!
checkedIntegerValueOf: intOop
	<doNotGenerate>
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	^coInterpreter checkedIntegerValueOf: intOop! !

!NewObjectMemory methodsFor: 'cog jit support' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
classFieldOffset
	<api>
	<cmacro: '() (0 - BaseHeaderSize)'>
	^0 - self baseHeaderSize! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem 9/11/2013 19:00'!
classForClassTag: classObj
	"Compatibility with SpurObjectMemory.  In ObjectMemory there is no distinction between a
	 classTag in the first-level method cache and a class itself."
	^classObj! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem 9/11/2013 18:39'!
classTagForClass: classObj
	"Compatibility with SpurObjectMemory.  In ObjectMemory there is no distinction between a
	 classTag in the first-level method cache and a class itself."
	^classObj! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem 9/11/2013 20:41'!
classTagForSpecialObjectsIndex: splObjIndex compactClassIndex: compactClassIndex
	"For compatibility with Spur.  Answer the class tag to use to lookup a method in the
	 first-level method lookup cache."
	^self splObj: splObjIndex! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem 2/27/2013 15:37'!
clearLeakMapAndMapAccessibleObjects
	"Perform an integrity/leak check using the heapMap.  Set a bit at each object's header."
	| obj sz nextHeader |
	<inline: false>
	<var: #obj type: #usqInt>
	<var: #sz type: #usqInt>
	<var: #nextHeader type: #usqInt>
	heapMap clearHeapMap.
	obj := self firstObject.
	[self oop: obj isLessThan: freeStart] whileTrue:
		[(self isFreeObject: obj)
			ifTrue:
				[sz := self sizeOfFree: obj]
			ifFalse:
				[heapMap heapMapAtWord: (self pointerForOop: obj) Put: 1.
				 sz := self sizeBitsOf: obj].
		nextHeader := obj + sz.
		obj := self oopFromChunk: nextHeader].! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem 01/07/2021 10:40'!
cloneObject: obj
	"Return a shallow copy of the given object. May cause GC.
	 Assume: Oop is a real object, not a small integer.
	 Override to assert it's not a married context and maybe fix cloned methods."
	| extraHdrBytes bytes newChunk remappedOop fromIndex toIndex lastFrom newOop header hash |
	<inline: false>
	<var: #lastFrom type: #usqInt>
	<var: #fromIndex type: #usqInt>
	self assert: ((self isContext: obj) not
				or: [(coInterpreter isMarriedOrWidowedContext: obj) not]). 

	self assert: (self isNonIntegerObject: obj).
	extraHdrBytes := self extraHeaderBytes: obj.
	bytes := self sizeBitsOf: obj.
	bytes := bytes + extraHdrBytes.

	"allocate space for the copy, remapping obj in case of a GC"
	self pushRemappableOop: obj.
	"check it is safe to allocate this much memory. Return 0 if not"
	(self sufficientSpaceToAllocate: 2500 + bytes) ifFalse:[^0].
	newChunk := self allocateChunk: bytes.
	remappedOop := self popRemappableOop.

	"copy old to new including all header words"
	toIndex := newChunk - self wordSize.  "loop below uses pre-increment"
	fromIndex := (remappedOop - extraHdrBytes) - self wordSize.
	lastFrom := fromIndex + bytes.
	[fromIndex < lastFrom] whileTrue:
		[self longAt: (toIndex := toIndex + self wordSize)
			put: (self longAt: (fromIndex := fromIndex + self wordSize))].
	newOop := newChunk + extraHdrBytes.  "convert from chunk to oop"

	"fix base header: compute new hash and clear Mark and Root bits"
	hash := self newObjectHash.
	header := (self longAt: newOop) bitAnd: 16r1FFFF.
	"use old ccIndex, format, size, and header-type fields"
	header := header bitOr: ((hash << HashBitsOffset) bitAnd: HashBits).
	self longAt: newOop put: header.
	(self isCompiledMethodHeader: header) ifTrue:
		[coInterpreter maybeFixClonedCompiledMethod: newOop].
	^newOop
! !

!NewObjectMemory methodsFor: 'cog jit support' stamp: 'eem 12/12/2015 14:12'!
compactClassFieldWidth
	<api>
	<cmacro>
	^5! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem 12/1/2014 10:56'!
dbgFloatValueOf: oop
	"Answer the C double precision floating point value of the argument,
	 or if it is not, answer 0."

	| isFloat result |
	<returnTypeC: #double>
	<var: #result type: #double>
	isFloat := self isFloatInstance: oop.
	isFloat ifTrue:
		[self cCode: '' inSmalltalk: [result := Float new: 2].
		 self fetchFloatAt: oop + self baseHeaderSize into: result.
		 ^result].
	^0.0! !

!NewObjectMemory methodsFor: 'initialization' stamp: 'eem (auto pragmas 12/08) 11/24/2008 13:43'!
defaultEdenBytes
	"Return the default amount of memory to allocate before doing a scavenge (incremental GC).
	 This default suits Qwaq Forums (specifically loading).  The actual value can be set via
	 vmParameterAt: and/or a preference in the ini file."
	<inline: false>
	^2 * 1024 * 1024! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 8/24/2018 16:36'!
disownVM: flags
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter disownVM: flags! !

!NewObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
edenBytes
	^edenBytes! !

!NewObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
edenBytes: aValue
	^edenBytes := aValue! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem 11/10/2014 12:49'!
eeAllocate: byteSize headerSize: hdrSize h1: baseHeader h2: classOop h3: extendedSize
	"Allocate a new object of the given size and number of header words. (Note: byteSize already includes
	 space for the base header word.) Initialize the header fields of the new object.
	 Does *not* initialize the objects' fields. Will *not* cause a GC.  This version is for the execution engine's use only."

	| newObj |
	<inline: true>
	newObj := self allocateInterpreterChunk: byteSize + (hdrSize - 1 * self wordSize).
	newObj = 0 ifTrue: [^newObj].
	hdrSize = 3 ifTrue:
		[self longAt: newObj put: (extendedSize bitOr: HeaderTypeSizeAndClass).
		 self longAt: newObj + self wordSize put: (classOop bitOr: HeaderTypeSizeAndClass).
		 self longAt: newObj + (self wordSize*2) put: (baseHeader bitOr: HeaderTypeSizeAndClass).
		 newObj := newObj + (self wordSize*2)].

	 hdrSize = 2 ifTrue:
		[self longAt: newObj put: (classOop bitOr: HeaderTypeClass).
		 self longAt: newObj + self wordSize put: (baseHeader bitOr: HeaderTypeClass).
		 newObj := newObj + self wordSize].

	 hdrSize = 1 ifTrue:
		[self longAt: newObj put: (baseHeader bitOr: HeaderTypeShort)].

	DoExpensiveAssertionChecks ifTrue:
		[self okayOop: newObj.
		 self oopHasOkayClass: newObj.
		 (self safeObjectAfter: newObj) = freeStart ifFalse:
			[self error: 'allocate bug: did not set header of new oop correctly']].

	^newObj! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem 11/10/2014 12:49'!
eeAllocate: byteSize headerSize: hdrSize h1: baseHeader h2: classOop h3: extendedSize doFill: doFill format: format
	"Allocate a new object of the given size and number of header words. (Note: byteSize already includes
	 space for the base header word.) Initialize the header fields of the new object and fill the remainder of
	 the object with the given value.  Will not cause a GC.  This version is for the execution engine"

	| newObj |
	<inline: true>
	<var: #i type: 'usqInt'>
	<var: #end type: 'usqInt'>
	newObj := self allocateInterpreterChunk: byteSize + (hdrSize - 1 * self wordSize).
	newObj = 0 ifTrue: [^newObj].
	hdrSize = 3 ifTrue:
		[self longAt: newObj put: (extendedSize bitOr: HeaderTypeSizeAndClass).
		 self longAt: newObj + self wordSize put: (classOop bitOr: HeaderTypeSizeAndClass).
		 self longAt: newObj + (self wordSize*2) put: (baseHeader bitOr: HeaderTypeSizeAndClass).
		 newObj := newObj + (self wordSize*2)].

	 hdrSize = 2 ifTrue:
		[self longAt: newObj put: (classOop bitOr: HeaderTypeClass).
		 self longAt: newObj + self wordSize put: (baseHeader bitOr: HeaderTypeClass).
		 newObj := newObj + self wordSize].

	 hdrSize = 1 ifTrue:
		[self longAt: newObj put: (baseHeader bitOr: HeaderTypeShort)].

	"clear new object"
	doFill ifTrue:
		[| fillWord end i |
		 fillWord := format <= self lastPointerFormat
					ifTrue: [nilObj] "if pointers, fill with nil oop"
					ifFalse: [0].
		 end := newObj + byteSize.
		 i := newObj + self wordSize.
		 [i < end] whileTrue:
			[self longAt: i put: fillWord.
			 i := i + self wordSize]].

	DoExpensiveAssertionChecks ifTrue:
		[self okayOop: newObj.
		 self oopHasOkayClass: newObj.
		 (self safeObjectAfter: newObj) = freeStart ifFalse:
			[self error: 'allocate bug: did not set header of new oop correctly']].

	^newObj! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
eeInstantiateAndInitializeClass: classPointer indexableSize: size 
	"NOTE: This method supports the backward-compatible split instSize field of the 
	 class format word. The sizeHiBits will go away and other shifts change by 2 
	 when the split fields get merged in an (incompatible) image change.
	 Will *not* cause a GC.  The instantiated object is initialized."

	| hash header1 header2 cClass byteSize format binc header3 hdrSize sizeHiBits bm1 classFormat |
	<inline: false>
	"cannot have a negative indexable field count"
	self assert: size >= 0.
	hash := self newObjectHash.
	classFormat := self formatOfClass: classPointer.
	"Low 2 bits are 0"
	header1 := (classFormat bitAnd: 16r1FF00) bitOr: (hash bitAnd: HashMaskUnshifted) << HashBitsOffset.
	header2 := classPointer.
	header3 := 0.
	sizeHiBits := (classFormat bitAnd: 16r60000) >> 9.
	cClass := header1 bitAnd: CompactClassMask. "compact class field from format word"
	byteSize := (classFormat bitAnd: SizeMask + Size4Bit) + sizeHiBits.
		"size in bytes -- low 2 bits are 0"
	"Note this byteSize comes from the format word of the class which is pre-shifted
		to 4 bytes per field.  Need another shift for 8 bytes per word..."
	byteSize := byteSize << (self shiftForWord-2).
	format := self formatOfHeader: classFormat.
	self flag: #sizeLowBits.
	format < self firstByteFormat
		ifTrue:
			[format = self firstLongFormat
				ifTrue: "long32 bitmaps"
					[bm1 := self wordSize-1.
					byteSize := byteSize + (size * 4) + bm1 bitAnd: LongSizeMask. "round up"
					binc := bm1 - ((size * 4) + bm1 bitAnd: bm1). "odd bytes"
					"extra low bit (4) for 64-bit VM goes in 4-bit (betw hdr bits and sizeBits)"
					header1 := header1 bitOr: (binc bitAnd: 4)]
				ifFalse: "Arrays and 64-bit bitmaps"
					[byteSize := byteSize + (size * self wordSize)]]
		ifFalse:
			["Strings and Methods"
			bm1 := self wordSize-1.
			byteSize := byteSize + size + bm1 bitAnd: LongSizeMask. "round up"
			binc := bm1 - (size + bm1 bitAnd: bm1). "odd bytes"
			"low bits of byte size go in format field"
			header1 := header1 bitOr: (binc bitAnd: 3) << self instFormatFieldLSB.
			"extra low bit (4) for 64-bit VM goes in 4-bit (betw hdr bits and sizeBits)"
			header1 := header1 bitOr: (binc bitAnd: 4)].
	byteSize > 255
		ifTrue: "requires size header word"
			[header3 := byteSize.
			header1 := header1]
		ifFalse: [header1 := header1 bitOr: byteSize].
	hdrSize := header3 > 0
					ifTrue: [3] "requires full header"
					ifFalse: [cClass = 0 ifTrue: [2] ifFalse: [1]].
	^self eeAllocate: byteSize headerSize: hdrSize h1: header1 h2: header2 h3: header3 doFill: true format: format! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
eeInstantiateClass: classPointer indexableSize: size 
	"NOTE: This method supports the backward-compatible split instSize field of the 
	 class format word. The sizeHiBits will go away and other shifts change by 2 
	 when the split fields get merged in an (incompatible) image change.
	 Will *not* cause a GC.
	 Note that the instantiated object IS NOT FILLED and must be completed before
	 returning it to Smalltalk. Since this call is used in routines that do just that we are
	 safe.  Break this rule and die."
	<api>
	| hash header1 header2 cClass byteSize format binc header3 hdrSize sizeHiBits bm1 classFormat |
	<inline: false>
	"cannot have a negative indexable field count"
	self assert: size >= 0.
	hash := self newObjectHash.
	classFormat := self formatOfClass: classPointer.
	"Low 2 bits are 0"
	header1 := (classFormat bitAnd: 16r1FF00) bitOr: (hash bitAnd: HashMaskUnshifted) << HashBitsOffset.
	header2 := classPointer.
	header3 := 0.
	sizeHiBits := (classFormat bitAnd: 16r60000) >> 9.
	cClass := header1 bitAnd: CompactClassMask. "compact class field from format word"
	byteSize := (classFormat bitAnd: SizeMask + Size4Bit) + sizeHiBits.
		"size in bytes -- low 2 bits are 0"
	"Note this byteSize comes from the format word of the class which is pre-shifted
		to 4 bytes per field.  Need another shift for 8 bytes per word..."
	byteSize := byteSize << (self shiftForWord-2).
	format := self formatOfHeader: classFormat.
	self flag: #sizeLowBits.
	format < self firstByteFormat
		ifTrue:
			[format = self firstLongFormat
				ifTrue: "long32 bitmaps"
					[bm1 := self wordSize-1.
					byteSize := byteSize + (size * 4) + bm1 bitAnd: LongSizeMask. "round up"
					binc := bm1 - ((size * 4) + bm1 bitAnd: bm1). "odd bytes"
					"extra low bit (4) for 64-bit VM goes in 4-bit (betw hdr bits and sizeBits)"
					header1 := header1 bitOr: (binc bitAnd: 4)]
				ifFalse: "Arrays and 64-bit bitmaps"
					[byteSize := byteSize + (size * self wordSize)]]
		ifFalse:
			["Strings and Methods"
			bm1 := self wordSize-1.
			byteSize := byteSize + size + bm1 bitAnd: LongSizeMask. "round up"
			binc := bm1 - (size + bm1 bitAnd: bm1). "odd bytes"
			"low bits of byte size go in format field"
			header1 := header1 bitOr: (binc bitAnd: 3) << self instFormatFieldLSB.
			"extra low bit (4) for 64-bit VM goes in 4-bit (betw hdr bits and sizeBits)"
			header1 := header1 bitOr: (binc bitAnd: 4)].
	byteSize > 255
		ifTrue: "requires size header word"
			[header3 := byteSize.
			header1 := header1]
		ifFalse: [header1 := header1 bitOr: byteSize].
	hdrSize := header3 > 0
					ifTrue: [3] "requires full header"
					ifFalse: [cClass = 0 ifTrue: [2] ifFalse: [1]].
	^self eeAllocate: byteSize headerSize: hdrSize h1: header1 h2: header2 h3: header3! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
eeInstantiateClassIndex: compactClassIndex format: objFormat numSlots: numSlots
	"Instantiate an instance of a compact class.  ee stands for execution engine and
	 implies that this allocation will *NOT* cause a GC.  N.B. the instantiated object
	 IS NOT FILLED and must be completed before returning it to Smalltalk. Since this
	 call is used in routines that do just that we are safe.  Break this rule and die in GC.
	 Result is guaranteed to be young."
	<api>
	| hash header1 header2 byteSize header3 hdrSize |
	<inline: false>
	"cannot have a negative indexable field count"
	self assert: (numSlots >= 0 and: [compactClassIndex ~= 0]).
	self assert: (objFormat < self firstByteFormat
					ifTrue: [objFormat]
					ifFalse: [objFormat bitAnd: self byteFormatMask])
				= (self instSpecOfClass: (self compactClassAt: compactClassIndex)).
	hash := self newObjectHash.
	"Low 2 bits are 0"
	header1 := (objFormat << self instFormatFieldLSB
					bitOr: compactClassIndex << 12)
					bitOr: (hash bitAnd: HashMaskUnshifted) << HashBitsOffset.
	self assert: "sizeHiBits" ((self formatOfClass: (self compactClassAt: compactClassIndex)) bitAnd: 16r60000) >> 9 = 0.
	self flag: #sizeLowBits.
	"size in bytes -- low 2 bits are 0; may need another shift if 64-bits.
	 strangely, size includes size of header, but only of single header.
	 why include header size at all?  gives us an extra word."
	byteSize := numSlots << (self shiftForWord + (self shiftForWord-2)) + self baseHeaderSize.
	(self wordSize = 8 "David, please check this!!!!"
	 and: [objFormat >= self firstLongFormat "32-bit longs and byte objects"
	 and: [(numSlots bitAnd: 1) ~= 0]]) ifTrue:
		["extra low bit (4) for 64-bit VM goes in 4-bit (betw hdr bits and sizeBits)"
		 header1 := header1 bitOr: 4].
	byteSize > 255 "requires size header word/full header"
		ifTrue: [header3 := byteSize. hdrSize := 3. header2 := self compactClassAt: compactClassIndex]
		ifFalse: [header1 := header1 bitOr: byteSize. hdrSize := 1].
	^self eeAllocate: byteSize headerSize: hdrSize h1: header1 h2: header2 h3: header3! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
eeInstantiateMethodContextSlots: numSlots 
	"This version of instantiateClass assumes that the total object 
	 size is under 256 bytes, the limit for objects with only one or 
	 two header words. Note that the size is specified in bytes 
	 and should include four bytes for the base header word.
	 Will *not* cause a GC. Result is guaranteed to be young."
	| sizeInBytes hash header1 |
	self assert: (numSlots = SmallContextSlots or: [numSlots = LargeContextSlots]).
	sizeInBytes := numSlots * self bytesPerOop + self baseHeaderSize.
	self assert: sizeInBytes <= SizeMask.
	hash := self newObjectHash.
	header1 := (hash bitAnd: HashMaskUnshifted) << HashBitsOffset bitOr: self formatOfMethodContextMinusSize.
	self assert: (header1 bitAnd: CompactClassMask) > 0. "contexts must be compact"
	self assert: (header1 bitAnd: SizeMask) = 0.
	"OR size into header1.  Must not do this if size > SizeMask"
	header1 := header1 + sizeInBytes.
	^self eeAllocate: sizeInBytes headerSize: 1 h1: header1 h2: nil h3: nil! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
eeInstantiateSmallClass: classPointer numSlots: numSlots
	"This version of instantiateClass assumes that the total object size is under
	 256 bytes, the limit for objects with only one or two header words. 
	 NOTE this code will only work for sizes that are an integral number of words
		(hence not a 32-bit LargeInteger in a 64-bit system).
	 Note that the created small object IS NOT FILLED and must be completed before returning it to Squeak.
	 Since this call is used in routines that do just that we are safe. Break this rule and die in GC.
	 Will *not* cause a GC. Result is guaranteed to be young."

	| sizeInBytes hash header1 header2 hdrSize |
	sizeInBytes := numSlots << self shiftForWord + self baseHeaderSize.
	self assert: sizeInBytes <= 252.
	hash := self newObjectHash.
	header1 := (hash bitAnd: HashMaskUnshifted) << HashBitsOffset bitOr: (self formatOfClass: classPointer).
	header2 := classPointer.
	hdrSize := (header1 bitAnd: CompactClassMask) > 0 "is this a compact class"
				ifTrue: [1]
				ifFalse: [2].
	header1 := header1 + (sizeInBytes - (header1 bitAnd: SizeMask+Size4Bit)).
	^self eeAllocate: sizeInBytes headerSize: hdrSize h1: header1 h2: header2 h3: 0! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
eeInstantiateSmallClassIndex: compactClassIndex format: objFormat numSlots: numSlots
	"This version of instantiateClass assumes that the total object size is under
	 256 bytes, the limit for objects with only one or two header words. 
	 NOTE this code will only work for sizes that are an integral number of words
		(hence not a 32-bit LargeInteger in a 64-bit system).
	 Note that the created small object IS NOT FILLED and must be completed before returning it to Squeak.
	 Since this call is used in routines that do just that we are safe. Break this rule and die in GC.
	 Will *not* cause a GC. Result is guaranteed to be young."

	| sizeInBytes hash header1 |
	"cannot have a negative indexable field count"
	self assert: (numSlots >= 0 and: [compactClassIndex ~= 0]).
	self assert: (objFormat < self firstByteFormat
					ifTrue: [objFormat]
					ifFalse: [objFormat bitAnd: self byteFormatMask])
				= (self instSpecOfClass: (self compactClassAt: compactClassIndex)).
	sizeInBytes := numSlots << self shiftForWord + self baseHeaderSize.
	self assert: sizeInBytes <= 252.
	hash := self newObjectHash.
	header1 := (objFormat << self instFormatFieldLSB
					bitOr: compactClassIndex << 12)
					bitOr: (hash bitAnd: HashMaskUnshifted) << HashBitsOffset.
	header1 := header1 + (sizeInBytes - (header1 bitAnd: SizeMask+Size4Bit)).
	^self eeAllocate: sizeInBytes headerSize: 1 h1: header1 h2: 0 h3: 0! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/19/2013 11:08'!
failed
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter failed! !

!NewObjectMemory methodsFor: 'accessing' stamp: 'eem 9/30/2010 12:27'!
falseObject: anOop
	"For mapInterpreterOops"
	falseObj := anOop! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem 2/24/2020 15:17'!
fetchClassTagOf: oop
	"Compatibility with SpurObjectMemory.  In ObjectMemory there is no distinction between a
	 classTag in the first-level method cache and a class itself."
	<api>
	^self fetchClassOf: oop! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem 9/11/2013 18:43'!
fetchClassTagOfNonImm: oop
	"Compatibility with SpurObjectMemory.  In ObjectMemory there is no distinction between a
	 classTag in the first-level method cache and a class itself."
	^self fetchClassOfNonImm: oop! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 15:37'!
fetchInteger: fieldIndex ofObject: objectPointer
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter fetchInteger: fieldIndex ofObject: objectPointer! !

!NewObjectMemory methodsFor: 'finalization' stamp: 'eem 9/7/2018 19:27'!
finalizeReference: oop 
	"During sweep phase we have encountered a weak reference. Check if its object
	 has gone away (or is about to) and if so, signal a semaphore.  Do *not* inline
	 this in sweepPhase - it is quite an unlikely case to run into a weak reference.
	 Override to ask the coInterpreter to signal finalization."
	| weakOop oopGone chunk numFields firstField lastField |
	<inline: false>
	<var: #oop type: #usqInt>
	<var: #weakOop type: #usqInt>
	numFields := self nonWeakFieldsOf: oop. "so nonWeakFieldsOf: may be inlined"
	firstField := self baseHeaderSize + (numFields << self shiftForWord).
	lastField := self lastPointerOf: oop.
	firstField to: lastField by: self wordSize do:
		[:i|
		weakOop := self longAt: oop + i.
		"ar 1/18/2005: Added oop < youngStart test to make sure we're not testing
		objects in non-GCable region. This could lead to a forward reference in
		old space with the oop pointed to not being marked and thus treated as free."
		(weakOop = nilObj or: [(self isIntegerObject: weakOop) or:[weakOop < youngStart]]) ifFalse:
			["Check if the object is being collected. 
			If the weak reference points  
			* backward: check if the weakOops chunk is free
			* forward: check if the weakOoop has been marked by GC"
			weakOop < oop
				ifTrue: [chunk := self chunkFromOop: weakOop.
						oopGone := ((self longAt: chunk) bitAnd: TypeMask) = HeaderTypeFree]
				ifFalse: [oopGone := (self isMarked: weakOop) not].
			oopGone ifTrue: "Store nil in the pointer and signal the  interpreter"
				[self longAt: oop + i put: nilObj.
				 self 
					cppIf: PharoVM 
					ifTrue: [numFields >= 2 ifTrue: [self weakFinalizerCheck: oop]].
				coInterpreter signalFinalization: oop]]]! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem 8/9/2021 13:34'!
findString: aCString
	"Print the oops of all string-like things that have the same characters as aCString"
	<api>
	<var: #aCString type: #'char *'>
	| aCStringStrlen obj sz |
	aCStringStrlen := self strlen: aCString.
	obj := self firstObject.
	[self oop: obj isLessThan: freeStart] whileTrue:
		[(self isFreeObject: obj)
			ifTrue:
				[sz := self sizeOfFree: obj]
			ifFalse:
				[(self object: obj equalsString: aCString ofSize: aCStringStrlen) ifTrue:
					[coInterpreter printHex: obj; space; printOopShort: obj; cr].
				 sz := self sizeBitsOf: obj].
		 obj := self oopFromChunk: obj + sz]! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem 8/9/2021 13:35'!
findStringBeginningWith: aCString
	"Print the oops of all string-like things that start with the same characters as aCString"
	<api>
	<var: #aCString type: #'char *'>
	| aCStringStrlen obj sz |
	aCStringStrlen := self strlen: aCString.
	obj := self firstObject.
	[self oop: obj isLessThan: freeStart] whileTrue:
		[(self isFreeObject: obj)
			ifTrue:
				[sz := self sizeOfFree: obj]
			ifFalse:
				[(self object: obj equalsString: aCString ofSize: aCStringStrlen) ifTrue:
					[coInterpreter printHex: obj; space; printNum: (self lengthOf: obj); space; printOopShort: obj; cr].
				 sz := self sizeBitsOf: obj].
		 obj := self oopFromChunk: obj + sz]! !

!NewObjectMemory methodsFor: 'object enumeration' stamp: 'eem 3/4/2009 23:06'!
firstAccessibleObject
	"Return the first accessible object in the heap."
	| obj |
	obj := self firstObject.
	[self oop: obj isLessThan: freeStart] whileTrue:
		[(self isFreeObject: obj) ifFalse: [^obj].
		 obj := self objectAfter: obj].
	self error: 'heap is empty'.
	^nil! !

!NewObjectMemory methodsFor: 'indexing primitive support' stamp: 'eem 9/10/2013 11:18'!
firstValidIndexOfIndexableObject: obj withFormat: fmt
	"Answer the one-relative index of the first valid index in an indexbale object
	 with the given format.  This is 1 for all objects except compiled methods
	 where the first index is beyond the last literal.
	 Used for safer bounds-checking on methods."
	^fmt >= self firstCompiledMethodFormat
		ifTrue: [coInterpreter firstByteIndexOfMethod: obj]
		ifFalse: [1]! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem 11/21/2014 16:25'!
floatObjectOf: aFloat
	| newFloatObj |
	<inline: false>
	<var: #aFloat type: #double>
	newFloatObj := self
						eeInstantiateSmallClassIndex: ClassFloatCompactIndex
						format: self firstLongFormat
						numSlots: (self sizeof: #double) / self bytesPerOop.
	self storeFloatAt: newFloatObj + self baseHeaderSize from: aFloat.
	^newFloatObj! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem 11/14/2019 13:43'!
floatValueOf: oop
	"Answer the C double precision floating point value of the argument,
	 or fail if it is not a Float, and answer 0.
	 Note: May be called by translated primitive code."
	<api>
	<returnTypeC: #double>
	| isFloat result |
	<var: #result type: #double>
	isFloat := self isFloatInstance: oop.
	isFloat ifTrue:
		[self cCode: '' inSmalltalk: [result := Float new: 2].
		 self fetchFloatAt: oop + self baseHeaderSize into: result.
		 ^result].
	coInterpreter primitiveFail.
	^0.0! !

!NewObjectMemory methodsFor: 'frame access' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
formatOfMethodContext
	"Answer the class format word for MethodContext which is used to instantiate
	 contexts without needing to fetch it from the class MethodContext itself."
	^(ClassMethodContextCompactIndex << 12)
	+ (self indexablePointersFormat << self instFormatFieldLSB) "Pointers+Variable"
	+ (CtxtTempFrameStart + (self baseHeaderSize / self wordSize) << 2)! !

!NewObjectMemory methodsFor: 'frame access' stamp: 'eem 9/10/2013 10:59'!
formatOfMethodContextMinusSize
	"Answer the class format word for MethodContext which is used to instantiate
	 contexts without needing to fetch it from the class MethodContext itself."
	^(ClassMethodContextCompactIndex << 12)
	+ (self indexablePointersFormat << self instFormatFieldLSB) "Pointers+Variable"! !

!NewObjectMemory methodsFor: 'become' stamp: 'eem 2/20/2013 13:11'!
freeObject: obj header: objHeader
	(self isContextHeader: objHeader) ifTrue:
		[(self asserta: ((coInterpreter isStillMarriedContext: obj) not)) ifFalse:
			[self error: 'attempt to free married context']].
	super freeObject: obj header: objHeader! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem 12/15/2008 10:55'!
freeSize
	"Return the instantaneous amount of free memory which is that
	 from the allocation pointer freeStart up to the reserve start."
	^(self oop: freeStart isLessThan: reserveStart)
		ifTrue: [(self cCoerce: reserveStart to: #usqInt) - (self cCoerce: freeStart to: #usqInt)]
		ifFalse: [0]! !

!NewObjectMemory methodsFor: 'accessing' stamp: 'eem 5/10/2013 13:43'!
freeStart
	"This is a horrible hack and only works because C macros are generated after Interpreter variables."
	<cmacro: '() GIV(freeStart)'>
	^freeStart! !

!NewObjectMemory methodsFor: 'garbage collection' stamp: 'eem 11/10/2014 12:49'!
fullCompaction
	"Move all accessible objects down to leave one big free chunk at the end of memory.
	 Assume:
		Incremental GC has just been done to maximimize forwarding table space.
		sweepPhaseForFullGC has already set compStart.
	 Need not and can not move objects below the first free chunk."
	| sz |
	self assert: compStart = (self lowestFreeAfter: self startOfMemory).
	compStart = freeStart ifTrue:
		["memory is already compact; only free chunk is at the end "
		 ^self initializeMemoryFirstFree: freeStart].
	(sz := self fwdTableSize: 8) < totalObjectCount ifTrue:
		["Try to grow OM to make a single pass full GC"
		 self growObjectMemory: totalObjectCount - sz + 10000 * 8].
	"work up through memory until all free space is at the end"
	[compStart < freeStart] whileTrue:
		["free chunk returned by incCompBody becomes start of next compaction"
		 compStart := self incCompBody]! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 3/14/2014 16:48'!
fullDisplayUpdate
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter fullDisplayUpdate! !

!NewObjectMemory methodsFor: 'garbage collection' stamp: 'eem 11/9/2020 19:15'!
fullGC
	"Do a mark/sweep garbage collection of the entire object memory.
	 Free inaccessible objects but do not move them."

	<inline: false>
	self runLeakCheckerFor: GCModeFull.
	coInterpreter preGCAction: GCModeFull.
	needGCFlag := false.
	gcStartUsecs := self ioUTCMicrosecondsNow.
	statSweepCount := statMarkCount := statMkFwdCount := statCompMoveCount := 0.
	self clearRootsTable.
	self initWeakTableForIncrementalGC: false.
	youngStart := self startOfMemory.  "process all of memory"
	self markPhase: true.
	"Sweep phase returns the number of survivors.
	Use the up-to-date version instead the one from startup."
	totalObjectCount := self sweepPhaseForFullGC.
	self runLeakCheckerFor: GCModeFull.
	self fullCompaction.
	statFullGCs := statFullGCs + 1.
	statGCEndUsecs := self ioUTCMicrosecondsNow.
	statFullGCUsecs := statFullGCUsecs + (statGCEndUsecs - gcStartUsecs).
	coInterpreter capturePendingFinalizationSignals.

	youngStart := freeStart.  "reset the young object boundary"
	self attemptToShrink.
	coInterpreter postGCAction: GCModeFull.
	self runLeakCheckerFor: GCModeFull! !

!NewObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
fwdTableInit: blkSize
	"Set the limits for a table of two- or three-word forwarding blocks above the last used oop. The pointer fwdTableNext moves up to fwdTableLast. Used for compaction of memory and become-ing objects. Returns the number of forwarding blocks available."

	<inline: false>
	"set endOfMemory to just after a minimum-sized free block"
	self setSizeOfFree: freeStart to: self baseHeaderSize.
	self setEndOfMemory: freeStart + self baseHeaderSize.

	"make a fake free chunk at endOfMemory for use as a sentinal in memory scans"
	self setSizeOfFree: endOfMemory to: self baseHeaderSize.

	"use all memory free between freeStart and memoryLimit for forwarding table"
	"Note: Forward blocks must be quadword aligned."
	fwdTableNext := (endOfMemory + self baseHeaderSize + 7) bitAnd: WordMask-7.
	self flag: #Dan.  "Above line does not do what it says (quadword is 16 or 32 bytes)"

	fwdTableLast := memoryLimit - blkSize.  "last forwarding table entry"

	"return the number of forwarding blocks available"
	^(fwdTableLast - fwdTableNext) // blkSize  "round down"! !

!NewObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
fwdTableSize: blkSize
	"Estimate the number of forwarding blocks available for compaction"
	| eom fwdFirst fwdLast |
	<inline: false>

	eom := freeStart + self baseHeaderSize.
	"use all memory free between freeStart and memoryLimit for forwarding table"

	"Note: Forward blocks must be quadword aligned."
	fwdFirst := (eom + self baseHeaderSize + 7) bitAnd: WordMask-7.
	self flag: #Dan.  "Above line does not do what it says (quadword is 16 or 32 bytes)"

	fwdLast := memoryLimit - blkSize.  "last forwarding table entry"

	"return the number of forwarding blocks available"
	^ (fwdLast - fwdFirst) // blkSize  "round down"! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem 1/18/2021 16:38'!
getStackPointer
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter getStackPointer! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 12/12/2013 11:30'!
getThisSessionID
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter getThisSessionID! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem 4/4/2009 20:11'!
growObjectMemory: delta 
	"Attempt to grow the object memory by the given delta amount."
	| limit |
	statGrowMemory := statGrowMemory + 1.
	limit := self sqGrowMemory: memoryLimit By: delta.
	limit = memoryLimit ifFalse:
		[self setMemoryLimit: limit - 24. "remove a tad for safety"
		 self initializeMemoryFirstFree: freeStart]! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem 10/11/2014 10:49'!
heapMap
	^heapMap! !

!NewObjectMemory methodsFor: 'image segment in/out' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
imageSegmentVersion
	| wholeWord |
	"a more complex version that tells both the word reversal and the endianness of the machine it came from.  Low half of word is 6502.  Top byte is top byte of #doesNotUnderstand: on this machine. ($d on the Mac or $s on the PC)"

	wholeWord := self longAt: (self splObj: SelectorDoesNotUnderstand) + self baseHeaderSize.
		"first data word, 'does' "
	^coInterpreter imageFormatVersion bitOr: (wholeWord bitAnd: 16rFF000000)! !

!NewObjectMemory methodsFor: 'snapshot' stamp: 'eem 10/8/2013 15:43'!
imageSizeToWrite
	"Answer the number of bytes that will be written to
	 the image file in a snapshot, excluding the header."
	^freeStart - self startOfMemory! !

!NewObjectMemory methodsFor: 'accessing' stamp: 'eem 7/7/2010 10:55'!
inGC
	^gcStartUsecs > statGCEndUsecs! !

!NewObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
incCompBody
	"Move objects to consolidate free space into one big chunk. Return the newly created free chunk."

	| bytesToBeFreed |
	<inline: false>
	"reserve memory for forwarding table"
	self fwdTableInit: self wordSize*2.  "Two-word blocks"

	"assign new oop locations, reverse their headers, and initialize forwarding blocks"
	bytesToBeFreed := self incCompMakeFwd.

	"update pointers to point at new oops"
	self mapPointersInObjectsFrom: youngStart to: freeStart.

	"move the objects and restore their original headers; return the new free chunk"
	^self incCompMove: bytesToBeFreed! !

!NewObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
incCompMakeFwd
	"Create and initialize forwarding blocks for all non-free objects  
	 following compStart. If the supply of forwarding blocks is exhausted,  
	 set compEnd to the first chunk above the area to be compacted;
	 otherwise, set it to endOfMemory. Return the number of bytes to be freed."
	| bytesToBeFreed oop fwdBlock newOop |
	<inline: false>
	bytesToBeFreed := 0.
	oop := self oopFromChunk: compStart.
	self assert: (self oop: oop isGreaterThan: self startOfMemory andLessThan: freeStart).
	[self oop: oop isLessThan: freeStart] whileTrue:
		[statMkFwdCount := statMkFwdCount + 1.
		 self assert: (self oop: oop isGreaterThan: self startOfMemory andLessThan: freeStart).
		 (self isFreeObject: oop)
			ifTrue: [bytesToBeFreed := bytesToBeFreed + (self sizeOfFree: oop)]
			ifFalse: "create a forwarding block for oop"
				[fwdBlock := self fwdBlockGet: self wordSize*2.
				 "Two-word block"
				 fwdBlock = nil ifTrue: "stop; we have used all available forwarding blocks"
					[compEnd := self chunkFromOop: oop.
					 ^bytesToBeFreed].
				newOop := oop - bytesToBeFreed.
				self assert: (self oop: newOop isGreaterThan: self startOfMemory andLessThan: freeStart).
				self initForwardBlock: fwdBlock mapping: oop to: newOop withBackPtr: false].
			oop := self objectAfterWhileForwarding: oop].
	compEnd := freeStart.
	^bytesToBeFreed! !

!NewObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
incCompMove: bytesFreed
	"Move all non-free objects between compStart and compEnd to their new  
	locations, restoring their headers in the process. Create a new free  
	block at the end of memory. Return the newly created free chunk. "
	"Note: The free block used by the allocator always must be the last free  
	block in memory. It may take several compaction passes to make all  
	free space bubble up to the end of memory."
	| oop next fwdBlock newOop header bytesToMove firstWord lastWord newFreeChunk sz target |
	<inline: false>
	<var: #firstWord type: 'usqInt'>
	<var: #lastWord type: 'usqInt'>
	<var: #w type: 'usqInt'>
	newOop := nil.
	oop := self oopFromChunk: compStart.
	[self oop: oop isLessThan: compEnd] whileTrue:
		[statCompMoveCount := statCompMoveCount + 1.
		next := self objectAfterWhileForwarding: oop.
		(self isFreeObject: oop) ifFalse:
			["a moving object; unwind its forwarding block"
			fwdBlock := self forwardingPointerOf: oop.
			self assert: (self fwdBlockValid: fwdBlock).
			newOop := self longAt: fwdBlock.
			header := self longAt: fwdBlock + self wordSize.
			self longAt: oop put: header. "restore the original header"
			bytesToMove := oop - newOop. "move the oop (including any extra header words) "
			sz := self sizeBitsOf: oop.
			firstWord := oop - (self extraHeaderBytes: oop).
			lastWord := oop + sz - self baseHeaderSize.
			target := firstWord - bytesToMove.
			firstWord to: lastWord by: self wordSize do:
				[:w | 
				self longAt: target put: (self longAt: w).
				target := target + self wordSize]].
		oop := next].
	newOop = nil
		ifTrue: ["no objects moved"
			oop := self oopFromChunk: compStart.
			((self isFreeObject: oop) and: [(self objectAfter: oop) = (self oopFromChunk: compEnd)])
				ifTrue: [newFreeChunk := oop]
				ifFalse: [newFreeChunk := freeStart]]
		ifFalse: ["initialize the newly freed memory chunk"
			"newOop is the last object moved; free chunk starts right after it"
			newFreeChunk := newOop + (self sizeBitsOf: newOop).
			self setSizeOfFree: newFreeChunk to: bytesFreed].
	next := self safeObjectAfter: newFreeChunk.
	self assert: (next = freeStart or: [next = (self oopFromChunk: compEnd)]).
	next = freeStart
		ifTrue: [self initializeMemoryFirstFree: newFreeChunk]
		ifFalse: ["newFreeChunk is not at end of memory; re-install freeStart.
				 This will be the case when a compaction needs more than one pass."
			self initializeMemoryFirstFree: freeStart].
	^newFreeChunk! !

!NewObjectMemory methodsFor: 'garbage collection' stamp: 'eem 10/6/2009 11:13'!
incrementalCompaction
	"Move objects down to make one big free chunk. Compact the 
	last N objects (where N = number of forwarding table 
	entries) of the young object area."
	"Assume: compStart was set during the sweep phase"
	compStart = freeStart
		ifTrue: ["Note: If compStart = freeStart then either the young 
			space is already compact  or there are enough forwarding table entries to do a 
			one-pass incr. compaction."
			self initializeMemoryFirstFree: freeStart]
		ifFalse: [self incCompBody]! !

!NewObjectMemory methodsFor: 'garbage collection' stamp: 'eem 9/7/2018 19:24'!
incrementalGC
	"Do a mark/sweep garbage collection of just the young object
	area of object memory (i.e., objects above youngStart), using
	the root table to identify objects containing pointers to
	young objects from the old object area."
	| survivorCount weDidGrow |
	<inline: false>

	rootTableOverflowed ifTrue:
		["root table overflow; cannot do an incremental GC because some roots are missing.
		 (this should be very rare)"
		 statRootTableOverflows := statRootTableOverflows + 1.
		 ^self fullGC].
	self runLeakCheckerFor: GCModeNewSpace.
	coInterpreter preGCAction: GCModeNewSpace.
	needGCFlag := false.
	gcStartUsecs := self ioUTCMicrosecondsNow.
	statSweepCount := statMarkCount := statMkFwdCount := statCompMoveCount := 0.
	self initWeakTableForIncrementalGC: true.
	"implicitly process memory from youngStart to freeStart"
	self markPhase: false.
	self assert: weakRootCount <= WeakRootTableSize.
	1 to: weakRootCount do:
		[:i| self finalizeReference: (weakRoots at: i)].
	survivorCount := self sweepPhase.
	self runLeakCheckerFor: GCModeNewSpace.
	self incrementalCompaction.
	statIncrGCs := statIncrGCs + 1.
	statGCEndUsecs := self ioUTCMicrosecondsNow.
	statIGCDeltaUsecs := statGCEndUsecs - gcStartUsecs.
	statIncrGCUsecs := statIncrGCUsecs + statIGCDeltaUsecs.
	coInterpreter capturePendingFinalizationSignals.
	
	statRootTableCount  := rootTableCount.
	statSurvivorCount := survivorCount.
	weDidGrow := false.
	(((survivorCount > tenuringThreshold)
	 or: [rootTableCount >= RootTableRedZone])
	 or: [forceTenureFlag == true]) ifTrue:
		["move up the young space boundary if
		  * there are too many survivors:
			this limits the number of objects that must be
			processed on future incremental GC's
		  * we're about to overflow the roots table:
			this limits the number of full GCs that may be caused
			by root table overflows in the near future"
		forceTenureFlag := false.
		statTenures := statTenures + 1.
		self clearRootsTable.
		((self freeSize < growHeadroom)
		 and: [gcBiasToGrow > 0]) ifTrue:
			[self biasToGrow.
			 weDidGrow := true].
		youngStart := freeStart].
	self attemptToShrink.
	coInterpreter postGCAction: GCModeNewSpace.
	
	self runLeakCheckerFor: GCModeNewSpace.
	weDidGrow ifTrue:
		[self biasToGrowCheckGCLimit]! !

!NewObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 11/10/2014 12:49'!
initForwardBlock: fwdBlock mapping: objOop to: newOop withBackPtr: backFlag
	"Initialize the given forwarding block to map oop to newOop, 
	and replace oop's header with a pointer to the fowarding block."
	"Details: The mark bit is used to indicate that an oop is 
	forwarded. When an oop is forwarded, its header (minus the 
	mark bit) contains the address of its forwarding block. (The 
	forwarding block address is actually shifted right by one bit 
	so that its top-most bit does not conflict with the header's 
	mark bit; since fowarding blocks are stored on word 
	boundaries, the low two bits of the address are always zero.) 
	The first word of the forwarding block is the new oop; the 
	second word is the oop's orginal header. In the case of a 
	forward become, a four-word block is used, with the third 
	field being a backpointer to the old oop (for header fixup), 
	and the fourth word is unused. The type bits of the 
	forwarding header are the same as those of the original 
	header. "
	| originalHeader originalHeaderType |
	<inline: true>
	self assert: fwdBlock ~= nil. "ran out of forwarding blocks in become"
	self deny: (self hasForwardingBlock: objOop). "'object already has a forwarding table entry"
	originalHeader := self longAt: objOop.
	originalHeaderType := originalHeader bitAnd: TypeMask.
	self longAt: fwdBlock put: newOop.
	self longAt: fwdBlock + self wordSize put: originalHeader.
	backFlag ifTrue: [self longAt: fwdBlock + (self wordSize*2) put: objOop].
	self longAt: objOop put: (fwdBlock >> 1 bitOr: (MarkBit bitOr: originalHeaderType))! !

!NewObjectMemory methodsFor: 'initialization' stamp: 'eem 11/9/2020 19:18'!
initialize
	"Initialize NewObjectMemory when simulating the VM inside Smalltalk."
	super initialize.
	checkForLeaks := 0.
	needGCFlag := false.
	heapMap := CogCheck32BitHeapMap new! !

!NewObjectMemory methodsFor: 'initialization' stamp: 'eem 3/18/2020 12:58'!
initializeMemoryFirstFree: firstFree 
	"Initialize endOfMemory to the top of oop storage space, reserving some space
	 for forwarding blocks, and set freeStart from which space is allocated."
	"Note: The amount of space reserved for forwarding blocks should be chosen to
	  ensure that incremental compactions can usually be done in a single pass.
	  However, there should be enough forwarding blocks so a full compaction can be done
	  in a reasonable number of passes, say ten. (A full compaction requires N object-moving
	  passes, where N = number of non-garbage objects / number of forwarding blocks).

	di 11/18/2000 Re totalObjectCount: Provide a margin of one byte per object to be
	 used for forwarding pointers at GC time. Since fwd blocks are 8 bytes, this means
	 an absolute worst case of 8 passes to compact memory. In most cases it will be
	 adequate to do compaction in a single pass. "
	| fwdBlockBytes totalReserve |
	"reserve space for forwarding blocks and the interpreter.  We can sacrifice
	 forwarding block space at the cost of slower compactions but we cannot
	 safely sacrifice interpreter allocation headroom."
	fwdBlockBytes := totalObjectCount bitAnd: WordMask - self wordSize + 1.
	totalReserve := fwdBlockBytes + coInterpreter interpreterAllocationReserveBytes.
	(self oop: memoryLimit - totalReserve isLessThan: firstFree + self baseHeaderSize) ifTrue:
		["reserve enough space for a minimal free block of BaseHeaderSize bytes.
		  We are apparently in an emergency situation here because we have no space
		  for reserve and forwarding blocks.  But a full GC will occur immediately in	
		  sufficientSpaceAfterGC: which will grow memory and restore the reserve."
		 fwdBlockBytes := memoryLimit - (firstFree  + self baseHeaderSize)].

	"set endOfMemory reserveStart and freeStart"
	self setEndOfMemory: memoryLimit - fwdBlockBytes.
	reserveStart := endOfMemory - coInterpreter interpreterAllocationReserveBytes.
	freeStart := firstFree. "bytes available for oops"
	scavengeThreshold := freeStart + edenBytes min: reserveStart.

	self assert: (self oop: freeStart isLessThan: reserveStart).
	"We would like to assert this but can't because in GC situations it may be false.  It is
	established by sufficientSpaceToAllocate: and sufficientSpaceAfterGC:"
	false ifTrue: [self assert: (self oop: reserveStart isLessThan: endOfMemory)].
	self assert: (self oop: endOfMemory isLessThan: memoryLimit)! !

!NewObjectMemory methodsFor: 'initialization' stamp: 'eem 1/18/2013 10:10'!
initializeObjectMemory: bytesToShift
	"Initialize object memory variables at startup time. Assume endOfMemory is initially set (by the image-reading code) to the end of the last object in the image. Initialization redefines endOfMemory to be the end of the object allocation area based on the total available memory, but reserving some space for forwarding blocks."
	"Assume: image reader initializes the following variables:
		memory
		endOfMemory
		memoryLimit
		specialObjectsOop
		lastHash
	"
	"di 11/18/2000 fix slow full GC"
	<inline: false>

	"set the start of the young object space"
	youngStart := freeStart := endOfMemory.

	"image may be at a different address; adjust oops for new location"
	totalObjectCount := self adjustAllOopsBy: bytesToShift.

	self initializeMemoryFirstFree: endOfMemory. "initializes endOfMemory, freeStart"

	specialObjectsOop := specialObjectsOop + bytesToShift.

	"heavily used special objects"
	nilObj	:= self splObj: NilObject.
	falseObj	:= self splObj: FalseObject.
	trueObj	:= self splObj: TrueObject.

	rootTableCount := 0.
	rootTableOverflowed := false.
	lowSpaceThreshold := 0.
	signalLowSpace := false.
	compStart := 0.
	compEnd := 0.
	fwdTableNext := 0.
	fwdTableLast := 0.
	remapBufferCount := 0.
	tenuringThreshold := 2000.  "tenure all suriving objects if survivor count is over this threshold"
	growHeadroom := 4*1024*1024. "four megabytes of headroom when growing"
	shrinkThreshold := 8*1024*1024. "eight megabytes of free space before shrinking"

	"garbage collection statistics"
	statFullGCs := 0.
	statFullGCUsecs := 0.
	statIncrGCs := 0.
	statIncrGCUsecs := 0.
	statTenures := 0.
	statRootTableOverflows := 0.
	statGrowMemory := 0.
	statShrinkMemory := 0.
	forceTenureFlag := 0.
	gcBiasToGrow := 0.
	gcBiasToGrowGCLimit := 0.
	extraRootCount := 0.
! !

!NewObjectMemory methodsFor: 'cog jit support' stamp: 'eem 12/12/2015 14:12'!
instFormatFieldLSB
	<api>
	<cmacro>
	^8! !

!NewObjectMemory methodsFor: 'cog jit support' stamp: 'eem 12/12/2015 14:12'!
instFormatFieldWidth
	<api>
	<cmacro>
	^4! !

!NewObjectMemory methodsFor: 'accessing' stamp: 'eem 6/10/2010 15:58'!
interpreter
	<doNotGenerate>
	^coInterpreter! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 1/9/2014 10:37'!
ioLoadFunction: functionName From: moduleName
	"hack around the CoInterpreter/ObjectMemory split refactoring.
	 provide accurate types for the VMPluginCodeGenerator."
	<doNotGenerate>
	<returnTypeC: #'void *'>
	<var: #functionName type: #'char *'>
	<var: #moduleName type: #'char *'>
	^coInterpreter ioLoadFunction: functionName From: moduleName! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 10/25/2013 20:33'!
ioUTCMicrosecondsNow
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter ioUTCMicrosecondsNow! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/19/2013 11:07'!
is: oop KindOf: classNameString
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter is: oop KindOf: classNameString! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 8/22/2018 12:06'!
is: oop KindOfClass: aClass
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter is: oop KindOfClass: aClass! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 12/18/2018 18:07'!
isBooleanObject: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter isBooleanObject: oop! !

!NewObjectMemory methodsFor: 'internal interpreter access' stamp: 'eem 8/1/2013 10:56'!
isContextHeader: aHeader
	<inline: true>
	"c.f. {BlockContext. MethodContext. PseudoContext} collect: [:class| class -> class indexIfCompact]"
	^(self compactClassIndexOfHeader: aHeader) = ClassMethodContextCompactIndex! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/24/2013 14:58'!
isFloatObject: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter isFloatObject: oop! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem 6/5/2017 10:27'!
isFloatOrInt: anOop
	"Answer if anOop is either a SmallInteger or a Float."

	<inline: true>

	^(self isIntegerObject: anOop)
	  or: [(self compactClassIndexOf: anOop) = ClassFloatCompactIndex]! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'nice 4/20/2016 01:58'!
isKindOfInteger: oop
	<doNotGenerate>
	^coInterpreter isKindOfInteger: oop! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'nice 4/20/2016 01:59'!
isLargeNegativeIntegerObject: oop
	<doNotGenerate>
	^coInterpreter isLargeNegativeIntegerObject: oop! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'nice 4/20/2016 01:59'!
isLargePositiveIntegerObject: oop
	<doNotGenerate>
	^coInterpreter isLargePositiveIntegerObject: oop! !

!NewObjectMemory methodsFor: 'image segment in/out' stamp: 'eem 10/10/2009 08:46'!
isPlatformFloatOrderVersion: imageVersion
	"Answer if floats are stored in the platform's float order for the given imageVersion.
	 Prior to this version they were always in big-endian (e.g. PowerPC) format."
	^imageVersion >= 6505! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 12/18/2018 17:43'!
isPositiveMachineIntegerObject: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter isPositiveMachineIntegerObject: oop! !

!NewObjectMemory methodsFor: 'memory access' stamp: 'eem 11/18/2013 15:18'!
isYoungObject: obj
	<api>
	"Answer if obj is young. Assume obj is non-immediate."
	self assert: (self isNonImmediate: obj).
	^self oop: obj isGreaterThanOrEqualTo: youngStart! !

!NewObjectMemory methodsFor: 'object enumeration' stamp: 'eem 11/10/2014 12:49'!
lastPointerOf: objOop 
	"Return the byte offset of the last pointer field of the given object.  
	 Can be used even when the type bits are not correct.
	 Works with CompiledMethods, as well as ordinary objects."
	<api>
	<inline: true>
	| fmt header contextSize |
	header := self baseHeader: objOop.
	fmt := self formatOfHeader: header.
	fmt <= self lastPointerFormat ifTrue:
		[(fmt = self indexablePointersFormat
		  and: [self isContextHeader: header]) ifTrue:
			["contexts end at the stack pointer"
			contextSize := coInterpreter fetchStackPointerOf: objOop.
			^CtxtTempFrameStart + contextSize * self bytesPerOop].
		^(self sizeBitsOfSafe: objOop) - self baseHeaderSize  "all pointers"].
	fmt < self firstCompiledMethodFormat ifTrue: [^0]. "no pointers"

	"CompiledMethod: contains both pointers and bytes"
	header := self methodHeaderOf: objOop.
	^self lastPointerOfMethodHeader: header! !

!NewObjectMemory methodsFor: 'object enumeration' stamp: 'eem 11/10/2014 12:49'!
lastPointerOf: objOop recordWeakRoot: recordWeakRoot "<Boolean>"
	"Return the byte offset of the last pointer field of the given object.  
	 Works with CompiledMethods, as well as ordinary objects. 
	 Can be used even when the type bits are not correct.
	 This is a version of lastPointerOf: for markAndTrace:.
	 Already overridden to trace stack pages for the StackInterpreter.
	 Override to ask coInterpreter to determine literalCount of methods."
	| fmt sz header contextSize numOops |
	<inline: true>
	header := self baseHeader: objOop.
	fmt := self formatOfHeader: header.
	fmt <= self lastPointerFormat ifTrue:
		[fmt >= self indexablePointersFormat ifTrue:
			[fmt = self lastPointerFormat ifTrue:
				[(recordWeakRoot and: [weakRootCount >= 0]) ifTrue:
					["And remember as weak root"
					 (weakRootCount := weakRootCount + 1) <= WeakRootTableSize ifFalse:
						[self error: 'weakRoots table overflow'].
					 weakRoots at: weakRootCount put: objOop].
				"Do not trace the object's indexed fields if it's a weak class"
				numOops := self nonWeakFieldsOf: objOop. "so nonWeakFieldsOf: may be inlined"
				^numOops * self wordSize].
			"So fmt is 3"
			(self isContextHeader: header) ifTrue:
				[coInterpreter setTraceFlagOnContextsFramesPageIfNeeded: objOop.
				 "contexts end at the stack pointer avoiding having to init fields beyond it"
				 contextSize := coInterpreter fetchStackPointerOf: objOop.
				 self assert: ReceiverIndex + contextSize < (self lengthOf: objOop baseHeader: header format: fmt).
				 ^CtxtTempFrameStart + contextSize * self bytesPerOop]].
		 sz := self sizeBitsOfSafe: objOop.
		 ^sz - self baseHeaderSize  "all pointers" ].
	fmt < self firstCompiledMethodFormat ifTrue: [^0]. "no pointers"

	"CompiledMethod: contains both pointers and bytes"
	header := self methodHeaderOf: objOop.
	^self lastPointerOfMethodHeader: header! !

!NewObjectMemory methodsFor: 'object enumeration' stamp: 'eem 11/10/2014 12:49'!
lastPointerOfMethodHeader: methodHeader 
	"Answer the byte offset of the last pointer field of a
	 CompiledMethod with the given header.  Use a temp to
	 allow inlining given MULTIPLEBYTECODESETS complications."
	<inline: true>
	| numLiterals |
	numLiterals := self literalCountOfMethodHeader: methodHeader.
	^numLiterals + LiteralStart - 1 * self bytesPerOop + self baseHeaderSize! !

!NewObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
lastPointerWhileForwarding: objOop 
	"The given object may have its header word in a forwarding block. Find  
	 the offset of the last pointer in the object in spite of this obstacle."
	| header fmt size contextSize |
	<inline: true>
	header := self headerWhileForwardingOf: objOop.
	fmt := self formatOfHeader: header.
	fmt <= self lastPointerFormat ifTrue:
		[(fmt = self indexablePointersFormat
		  and: [self isContextHeader: header]) ifTrue:
			["contexts end at the stack pointer"
			 contextSize := coInterpreter nacFetchStackPointerOf: objOop.
			 self assert: ReceiverIndex + contextSize < (self lengthOf: objOop baseHeader: header format: fmt).
			 ^CtxtTempFrameStart + contextSize * self bytesPerOop].
		 "do sizeBitsOf: using the header we obtained"
		 size := (header bitAnd: TypeMask) = HeaderTypeSizeAndClass
					ifTrue: [(self sizeHeader: objOop) bitAnd: AllButTypeMask]
					ifFalse: [header bitAnd: SizeMask].
		 ^size - self baseHeaderSize].
	fmt < self firstCompiledMethodFormat ifTrue: [^0]. "no pointers"
	"CompiledMethod: contains both pointers and bytes"
	self assert: (self isCompiledMethodHeader: header).
	header := self noCheckMethodHeaderOf: objOop.
	^self lastPointerOfMethodHeader: header! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem 2/7/2015 09:59'!
leakCheckBecome
	<api>
	^(checkForLeaks bitAnd: GCModeBecome) ~= 0! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem 2/7/2015 09:59'!
leakCheckFullGC
	<api>
	^(checkForLeaks bitAnd: GCModeFull) ~= 0! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem 10/11/2013 16:39'!
leakCheckNewSpaceGC
	<api>
	^(checkForLeaks bitAnd: 2) ~= 0! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem 12/29/2019 17:28'!
leakCheckRuntimeCalls
	<doNotGenerate>
	^(checkForLeaks bitAnd: 32) ~= 0! !

!NewObjectMemory methodsFor: 'memory access' stamp: 'eem 6/7/2016 14:54'!
literalCountOfMethodHeader: methodHeader
	<api>
	<inline: true>
	self assert: (self isIntegerObject: methodHeader).
	^(self headerIndicatesAlternateBytecodeSet: methodHeader)
		ifTrue: [coInterpreter literalCountOfAlternateHeader: methodHeader]
		ifFalse: [coInterpreter literalCountOfOriginalHeader: methodHeader]! !

!NewObjectMemory methodsFor: 'image segment in/out' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
loadImageSegmentFrom: segmentWordArray outPointers: outPointerArray
	"This primitive is called from Squeak as...
		<imageSegment> loadSegmentFrom: aWordArray outPointers: anArray."

"This primitive will load a binary image segment created by primitiveStoreImageSegment.  It expects the outPointer array to be of the proper size, and the wordArray to be well formed.  It will return as its value the original array of roots, and the erstwhile segmentWordArray will have been truncated to a size of one word, i.e. retaining the version stamp.  If this primitive should fail, the segmentWordArray will, sadly, have been reduced to an unrecognizable and unusable jumble.  But what more could you have done with it anyway?"

	| endSeg segOop fieldPtr fieldOop doingClass lastPtr extraSize mapOop lastOut outPtr hdrTypeBits header data |
	<inline: false>
	<var: #endSeg type: #usqInt>
	<var: #segOop type: #usqInt>
	<var: #fieldPtr type: #usqInt>
	<var: #lastOut type: #usqInt>
	<var: #outPtr type: #usqInt>
	<var: #lastPtr type: #usqInt>

	DoAssertionChecks ifTrue: [self verifyCleanHeaders].
	lastOut := outPointerArray + (self lastPointerOf: outPointerArray).
	endSeg := segmentWordArray + (self sizeBitsOf: segmentWordArray) - self baseHeaderSize.

	"Version check.  Byte order of the WordArray now"
	data := self longAt: segmentWordArray + self baseHeaderSize.
	(coInterpreter readableFormat: (data bitAnd: 16rFFFF "low 2 bytes")) ifFalse:
		"Not readable -- try again with reversed bytes..."
		[self reverseBytesFrom: segmentWordArray + self baseHeaderSize to: endSeg + self wordSize.
		data := self longAt: segmentWordArray + self baseHeaderSize.
		(coInterpreter readableFormat: (data bitAnd: 16rFFFF "low 2 bytes")) ifFalse:
			"Still NG -- put things back and fail"
			[self reverseBytesFrom: segmentWordArray + self baseHeaderSize to: endSeg + self wordSize.
			DoAssertionChecks ifTrue: [self verifyCleanHeaders].
			^PrimErrBadArgument]].
	"Reverse the Byte type objects if the data is from opposite endian machine.
	 Revese the words in Floats if from an earlier version with different Float order.
	 Test top byte.  $d on the Mac or $s on the PC.  Rest of word is equal."
	(data >> 16) = (self imageSegmentVersion >> 16)
		ifTrue:
			"Need to swap floats if the segment is being loaded into a little-endian VM from a version
			 that keeps Floats in big-endian word order as was the case prior to the 6505 image format."
			[(self isPlatformFloatOrderVersion: (data bitAnd: 16rFFFF "low 2 bytes")) ifFalse:
				[self vmEndianness ~= 1 "~= 1 => little-endian" ifTrue:
					[segOop := self oopFromChunk: segmentWordArray + self baseHeaderSize + self wordSize.
					 self wordSwapFloatsFrom: segOop to: endSeg + self wordSize]]]
		ifFalse: "Reverse the byte-type objects once"
			[segOop := self oopFromChunk: segmentWordArray + self baseHeaderSize + self wordSize.
				 "Oop of first embedded object"
			self byteSwapByteObjectsFrom: segOop
				to: endSeg + self wordSize
				flipFloatsIf: (self isPlatformFloatOrderVersion: (data bitAnd: 16rFFFF "low 2 bytes"))].

	"Proceed through the segment, remapping pointers..."
	segOop := self oopFromChunk: segmentWordArray + self baseHeaderSize + self wordSize.
	[segOop <= endSeg] whileTrue:
		[(self headerType: segOop) <= 1
			ifTrue: ["This object has a class field (type = 0 or 1) -- start with that."
					fieldPtr := segOop - self wordSize.  doingClass := true]
			ifFalse: ["No class field -- start with first data field"
					fieldPtr := segOop + self baseHeaderSize.  doingClass := false].
		lastPtr := segOop + (self lastPointerOf: segOop).	"last field"
		lastPtr > endSeg ifTrue:
			[DoAssertionChecks ifTrue: [self verifyCleanHeaders].
			^PrimErrBadIndex "out of bounds"].

		"Go through all oops, remapping them..."
		[fieldPtr > lastPtr] whileFalse:
			["Examine each pointer field"
			fieldOop := self longAt: fieldPtr.
			doingClass ifTrue:
				[hdrTypeBits := self headerType: fieldPtr.
				fieldOop := fieldOop - hdrTypeBits].
			(self isIntegerObject: fieldOop)
				ifTrue:
					["Integer -- nothing to do"
					fieldPtr := fieldPtr + self wordSize]
				ifFalse:
					[(fieldOop bitAnd: 3) = 0 ifFalse:
						[^PrimErrBadIndex "bad oop"].
					(fieldOop bitAnd: 16r80000000) = 0
						ifTrue: ["Internal pointer -- add segment offset"
								mapOop := fieldOop + segmentWordArray]
						ifFalse: ["External pointer -- look it up in outPointers"
								outPtr := outPointerArray + (fieldOop bitAnd: 16r7FFFFFFF).
								outPtr > lastOut ifTrue:
									[^PrimErrBadIndex "out of bounds"].
								mapOop := self longAt: outPtr].
					doingClass
						ifTrue: [self longAt: fieldPtr put: mapOop + hdrTypeBits.
								fieldPtr := fieldPtr + 8.
								doingClass := false]
						ifFalse: [self longAt: fieldPtr put: mapOop.
								fieldPtr := fieldPtr + self wordSize].
					segOop < youngStart ifTrue:
						[self possibleRootStoreInto: segOop value: mapOop]]].
		segOop := self objectAfter: segOop].

	"Again, proceed through the segment checking consistency..."
	segOop := self oopFromChunk: segmentWordArray + self baseHeaderSize + self wordSize.
	[segOop <= endSeg] whileTrue:
		[(self oopHasAcceptableClass: segOop) ifFalse:
			[^PrimErrInappropriate "inconsistency"].
		fieldPtr := segOop + self baseHeaderSize.		"first field"
		lastPtr := segOop + (self lastPointerOf: segOop).	"last field"
		"Go through all oops, remapping them..."
		[fieldPtr > lastPtr] whileFalse:
			["Examine each pointer field"
			fieldOop := self longAt: fieldPtr.
			(self oopHasAcceptableClass: fieldOop) ifFalse:
				[^PrimErrInappropriate "inconsistency"].
			fieldPtr := fieldPtr + self wordSize].
		segOop := self objectAfter: segOop].

	"Truncate the segment word array to size = BytesPerWord (vers stamp only)"
	extraSize := self extraHeaderBytes: segmentWordArray.
	hdrTypeBits := self headerType: segmentWordArray.
	extraSize = 8
		ifTrue: [self longAt: segmentWordArray-extraSize put: self baseHeaderSize + self wordSize + hdrTypeBits]
		ifFalse: [header := self longAt: segmentWordArray.
				self longAt: segmentWordArray
					put: header - (header bitAnd: SizeMask) + self baseHeaderSize + self wordSize].	
	"and return the roots array which was first in the segment"
	DoAssertionChecks ifTrue: [self verifyCleanHeaders].
	^self oopFromChunk: segmentWordArray + self baseHeaderSize + self wordSize! !

!NewObjectMemory methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:02'!
longPrintReferencesTo: anOop
	"Scan the heap long printing the oops of any and all objects that refer to anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| oop i prntObj |
	prntObj := false.
	oop := self firstAccessibleObject.
	[oop = nil] whileFalse:
		[((self isPointersNonImm: oop) or: [self isCompiledMethod: oop]) ifTrue:
			[(self isCompiledMethod: oop)
				ifTrue:
					[i := (self literalCountOf: oop) + LiteralStart]
				ifFalse:
					[(self isContextNonImm: oop)
						ifTrue: [i := CtxtTempFrameStart + (coInterpreter fetchStackPointerOf: oop)]
						ifFalse: [i := self lengthOf: oop]].
			[(i := i - 1) >= 0] whileTrue:
				[anOop = (self fetchPointer: i ofObject: oop) ifTrue:
					[self printHex: oop; print: ' @ '; printNum: i; cr.
					 prntObj := true.
					 i := 0]].
			prntObj ifTrue:
				[prntObj := false.
				 coInterpreter longPrintOop: oop]].
		 oop := self accessibleObjectAfter: oop]! !

!NewObjectMemory methodsFor: 'simulation' stamp: 'eem 12/28/2019 02:28'!
lookup32BitWordConstant: word
	<doNotGenerate>
	"Compatibility with SpurMemoryManager"
	^nil! !

!NewObjectMemory methodsFor: 'simulation' stamp: 'eem 9/4/2013 17:04'!
lookupAddress: address
	"If address appears to be that of a Symbol or a few well-known objects (such as classes) answer it, otherwise answer nil.
	 For code disassembly"
	<doNotGenerate>
	| fmt size string class classSize maybeThisClass classNameIndex thisClassIndex |
	(self addressCouldBeObj: address) ifFalse:
		[^nil].
	fmt := self formatOf: address.
	size := self lengthOf: address baseHeader: (self baseHeader: address) format: fmt.
	size = 0 ifTrue:
		[^address caseOf: { [nilObj] -> ['nil']. [trueObj] -> ['true']. [falseObj] -> ['false'] } otherwise: []].
	((fmt between: 8 and: 11) "indexable byte fields"
	and: [(size between: 1 and: 64)
	and: [Scanner isLiteralSymbol: (string := (0 to: size - 1) collect: [:i| Character value: (self fetchByte: i ofObject: address)])]]) ifTrue:
		[^'#', (ByteString withAll: string)].
	class := self fetchClassOfNonImm: address.
	((self addressCouldBeObj: class)
	 and: [(self headerType: class) ~= HeaderTypeShort]) ifFalse:
		[^nil].
	"address is either a class or a metaclass, or an instance of a class or invalid.  determine which."
	classNameIndex := coInterpreter classNameIndex.
	thisClassIndex := coInterpreter thisClassIndex.
	((classSize := self lengthOf: class baseHeader: (self baseHeader: address) format: fmt) <= (classNameIndex max: thisClassIndex)
	 or: [classSize > 255]) ifTrue:
		[^nil].
	"Address could be a class or a metaclass"
	(fmt = 1 and: [size >= classNameIndex]) ifTrue:
		["Is address a class? If so class's thisClass is address."
		 (self lookupAddress: (self fetchPointer: classNameIndex ofObject: address)) ifNotNil:
			[:maybeClassName|
			(self fetchPointer: thisClassIndex ofObject: class) = address ifTrue:
				[^maybeClassName allButFirst]].
		"Is address a Metaclass?  If so class's name is Metaclass and address's thisClass holds the class name"
		((self isBytes: (self fetchPointer: classNameIndex ofObject: class))
		 and: [(self lookupAddress: (self fetchPointer: classNameIndex ofObject: class)) = '#Metaclass'
		 and: [size >= thisClassIndex]]) ifTrue:
			[maybeThisClass := self fetchPointer: thisClassIndex ofObject: address.
			(self lookupAddress: (self fetchPointer: classNameIndex ofObject: maybeThisClass)) ifNotNil:
				[:maybeThisClassName| ^maybeThisClassName allButFirst, ' class']]].
	^(self lookupAddress: (self fetchPointer: classNameIndex ofObject: class)) ifNotNil:
		[:maybeClassName| 'a(n) ', maybeClassName allButFirst]! !

!NewObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 9/14/2013 11:07'!
mapPointersInObjectsFrom: memStart to: memEnd
	"Use the forwarding table to update the pointers of all non-free objects in the given range of memory.
	 Also remap pointers in root objects which may contains pointers into the given memory range, and
	 don't forget to flush the method cache based on the range."
	<inline: false>
	(self shouldRemapObj: nilObj) ifTrue:
		[nilObj := self remapObj: nilObj].
	(self shouldRemapObj: falseObj) ifTrue:
		[falseObj := self remapObj: falseObj].
	(self shouldRemapObj: trueObj) ifTrue:
		[trueObj := self remapObj: trueObj].
	(self shouldRemapObj: specialObjectsOop) ifTrue:
		[specialObjectsOop := self remapObj: specialObjectsOop].
	"update interpreter variables"
	coInterpreter
		mapInterpreterOops;
		flushMethodCacheFrom: memStart to: memEnd.
	self remapRemapBufferAndExtraRoots.
	self updatePointersInRootObjectsFrom: memStart to: memEnd.
	self updatePointersInRangeFrom: memStart to: memEnd! !

!NewObjectMemory methodsFor: 'primitive support' stamp: 'eem 3/28/2013 16:09'!
mark: obj
	<inline: true>
	self baseHeader: obj put: ((self baseHeader: obj) bitOr: MarkBit)! !

!NewObjectMemory methodsFor: 'gc -- mark and sweep' stamp: 'eem 9/7/2018 19:24'!
markPhase: fullGCFlag
	"Mark phase of the mark and sweep garbage collector. Set 
	 the mark bits of all reachable objects. Free chunks are 
	 untouched by this process."
	"Assume: All non-free objects are initially unmarked. Root 
	 objects were unmarked when they were made roots.
	 (Make sure this stays true!!!!)."
	| oop statMarkCountPriorToStackPageFreeing |
	<inline: false>
	"trace the interpreter's objects, including the active stacks
	 and special objects array"
	coInterpreter markAndTraceInterpreterOops: fullGCFlag.
	statSpecialMarkCount := statMarkCount.
	"trace the roots"
	1 to: rootTableCount do:
		[:i | 
		oop := rootTable at: i.
		self markAndTrace: oop].
	1 to: extraRootCount do:
		[:i|
		oop := (extraRoots at: i) at: 0.
		((self isIntegerObject: oop) or: [self isFreeObject: oop]) ifFalse:
			[self markAndTrace: oop]].
	statMarkCountPriorToStackPageFreeing := statMarkCount.
	"Only safe to free stack pages after all roots have been traced."
	coInterpreter markAndTraceAndMaybeFreeStackPages: fullGCFlag.
	"Only safe to free any machine code methods after all
	 stack pages have been traced."
	coInterpreter markAndTraceOrFreeMachineCode: fullGCFlag.
	statSpecialMarkCount := statSpecialMarkCount + (statMarkCount - statMarkCountPriorToStackPageFreeing)! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/19/2013 11:09'!
methodArgumentCount
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter methodArgumentCount! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem 8/4/2016 11:33'!
methodCacheHashOf: selector with: classTag
	"Since class tags are class objects, and classes are aligned on a 4 byte boundary there
	 is no need to shift the class to include the least significant bits of the class in the hash."
	<inline: true>
	^selector bitXor: classTag! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 03/16/2018 14:10'!
methodReturnBool: boolean
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter methodReturnBool: boolean! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 3/16/2018 15:22'!
methodReturnFloat: aFloat
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter methodReturnFloat: aFloat! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 03/16/2018 14:10'!
methodReturnInteger: integer
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter methodReturnInteger: integer! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'ul 4/17/2018 16:02'!
methodReturnReceiver
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter methodReturnReceiver! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 03/16/2018 14:10'!
methodReturnString: aCString
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter methodReturnString: aCString! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 1/24/2014 12:42'!
methodReturnValue: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter methodReturnValue: oop! !

!NewObjectMemory methodsFor: 'primitive support' stamp: 'eem 6/23/2012 18:02'!
minLargeHeaderSize
	"Answer the smallest byte size requiring a large header."
	^256! !

!NewObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
needGCFlag
	^needGCFlag! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
newObjectHash
	"Derive the new object hash from the allocation pointer.  This is less costly than
	 using lastHash because it avoids the read-modify-write cycle to update lastHash.
	 Since the size of eden is a power of two and larger than the hash range this provides
	 a well-distributed and fairly random set of values."
	<inline: true>
	^freeStart >> self shiftForWord! !

!NewObjectMemory methodsFor: 'accessing' stamp: 'eem 10/9/2013 20:20'!
newSpaceBytes
	^self edenBytes! !

!NewObjectMemory methodsFor: 'accessing' stamp: 'eem 10/12/2013 09:25'!
newSpaceSize
	^freeStart - self startOfMemory! !

!NewObjectMemory methodsFor: 'primitive support' stamp: 'eem 3/29/2013 11:35'!
nilFieldsOf: arrayObj 
	0 to: (self lengthOf: arrayObj) - 1 do:
		[:i|
		self storePointerUnchecked: i ofObject: arrayObj withValue: nilObj]! !

!NewObjectMemory methodsFor: 'accessing' stamp: 'eem 9/30/2010 12:27'!
nilObject: anOop
	"For mapInterpreterOops"
	nilObj := anOop! !

!NewObjectMemory methodsFor: 'garbage collection' stamp: 'eem 11/10/2014 12:49'!
noteAsRoot: oop headerLoc: headerLoc 
	"Record that the given oop in the old object area points to an object in the young area.
	 HeaderLoc is usually = oop, but may be an addr in a forwarding block."
	| header |
	<inline: true>
	header := self longAt: headerLoc.
	(self isYoungRootHeader: header) ifFalse:
		"record oop as root only if not already recorded"
		[rootTableCount < RootTableSize
			ifTrue:
				"record root if there is enough room in the roots table.
				 IMPORTANT: since clearRootsTable is the only thing that clears root bits
				 do *not* set the root bit unless an object is in the root table.  checking
				 routines will complain about the root bit being unset instead of the table
				 being full, but that's life"
				[rootTableCount := rootTableCount + 1.
				 rootTable at: rootTableCount put: oop.
				 self longAt: headerLoc put: (header bitOr: RootBit).
				 rootTableCount >= RootTableRedZone ifTrue:
					"if we're now in the red zone force an IGC ASAP"
					[self scheduleIncrementalGC]]
			ifFalse: "note overflow; will need to do a fullGC instead of an incremental."
				[rootTableOverflowed := true]]! !

!NewObjectMemory methodsFor: 'debug printing' stamp: 'eem 12/4/2013 17:26'!
numReferencesTo: anOop
	"Answer the number of objects that refer to anOop, other than anOop."
	| oop i n |
	oop := self firstAccessibleObject.
	n := 0.
	[oop = nil] whileFalse:
		[((self isPointersNonImm: oop) or: [self isCompiledMethod: oop]) ifTrue:
			[(self isCompiledMethod: oop)
				ifTrue:
					[i := (self literalCountOf: oop) + LiteralStart]
				ifFalse:
					[(self isContextNonImm: oop)
						ifTrue: [i := CtxtTempFrameStart + (coInterpreter fetchStackPointerOf: oop)]
						ifFalse: [i := self lengthOf: oop]].
			[(i := i - 1) >= 0] whileTrue:
				[anOop = (self fetchPointer: i ofObject: oop) ifTrue:
					[anOop ~= oop ifTrue:
						[n := n + 1].
					 i := 0]]].
		 oop := self accessibleObjectAfter: oop].
	^n! !

!NewObjectMemory methodsFor: 'object enumeration' stamp: 'eem 11/10/2014 12:49'!
objectAfter: oop 
	"Return the object or free chunk immediately following the 
	given object or free chunk in memory. Return endOfMemory 
	when enumeration is complete."
	| sz |
	<api>
	<inline: true>
	(self asserta: (self oop: oop isLessThan: freeStart)) ifFalse:
		[self error: 'no objects after the end of memory'].
	(self isFreeObject: oop)
		ifTrue: [sz := self sizeOfFree: oop]
		ifFalse: [sz := self sizeBitsOf: oop].
	^self oopFromChunk: oop + sz! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
okayOop: signedOop
	"Verify that the given oop is legitimate. Check address, header, and size but not class."

	| sz type fmt unusedBit oop |
	<var: #oop type: #usqInt>
	oop := self cCoerce: signedOop to: #usqInt.

	"address and size checks"
	(self isIntegerObject: oop) ifTrue: [ ^true ].
	(oop >= self startOfMemory and: [oop < freeStart])
		ifFalse: [ self error: 'oop is not a valid address'. ^false ].
	((oop \\ self wordSize) = 0)
		ifFalse: [ self error: 'oop is not a word-aligned address'. ^false ].
	sz := self sizeBitsOf: oop.
	(oop + sz) <= freeStart
		ifFalse: [ self error: 'oop size would make it extend beyond the end of memory'. ^false ].

	"header type checks"
	type := self headerType: oop.
	type = HeaderTypeFree
		ifTrue:  [ self error: 'oop is a free chunk, not an object'. ^false ].
	type = HeaderTypeShort ifTrue: [
		(self compactClassIndexOf: oop) = 0
			ifTrue:  [ self error: 'cannot have zero compact class field in a short header'. ^false ].
	].
	type = HeaderTypeClass ifTrue: [
		((oop >= self wordSize) and: [(self headerType: oop - self wordSize) = type])
			ifFalse: [ self error: 'class header word has wrong type'. ^false ].
	].
	type = HeaderTypeSizeAndClass ifTrue: [
		((oop >= (self wordSize*2)) and:
		 [(self headerType: oop - (self wordSize*2)) = type and:
		 [(self headerType: oop - self wordSize) = type]])
			ifFalse: [ self error: 'class header word has wrong type'. ^false ].
	].

	"format check"
	fmt := self formatOf: oop.
	((fmt = 5) | (fmt = 7))
		ifTrue:  [ self error: 'oop has an unknown format type'. ^false ].

	"mark and root bit checks"
	unusedBit := 16r20000000.
	self wordSize = 8
		ifTrue:
			[unusedBit := unusedBit << 16.
			 unusedBit := unusedBit << 16].
	((self longAt: oop) bitAnd: unusedBit) = 0
		ifFalse: [ self error: 'unused header bit 30 is set; should be zero'. ^false ].
"xxx
	((self longAt: oop) bitAnd: MarkBit) = 0
		ifFalse: [ self error: 'mark bit should not be set except during GC' ].
xxx"
	((self isYoungRoot: oop) and: [oop >= youngStart])
		ifTrue: [ self error: 'root bit is set in a young object'. ^false ].
	^true
! !

!NewObjectMemory methodsFor: 'image segment in/out' stamp: 'eem 12/4/2013 17:26'!
oopHasAcceptableClass: signedOop
	"Similar to oopHasOkayClass:, except that it only returns true or false."

	| oopClass formatMask behaviorFormatBits oopFormatBits oop |
	<var: #oop type: #usqInt>
	<var: #oopClass type: #usqInt>

	(self isIntegerObject: signedOop) ifTrue: [^ true].

	oop := self cCoerce: signedOop to: #usqInt.
	(self addressCouldBeObj: oop) ifFalse: [^ false].

	oopClass := self cCoerce: (self fetchClassOfNonImm: oop) to: #usqInt.
	(self addressCouldBeObj: oopClass) ifFalse: [^ false].
	(oopClass + (self sizeBitsOf: oopClass)) <= freeStart ifFalse: [^ false].

	((self isPointersNonImm: oopClass) and: [(self lengthOf: oopClass) >= 3]) ifFalse: [^ false].

	formatMask := (self isBytesNonImm: oop)
						ifTrue: [16rC00]  "ignore extra bytes size bits"
						ifFalse: [16rF00].

	behaviorFormatBits := (self formatOfClass: oopClass) bitAnd: formatMask.
	oopFormatBits := (self baseHeader: oop) bitAnd: formatMask.
	behaviorFormatBits = oopFormatBits ifFalse: [^ false].
	^ true! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 8/24/2018 16:37'!
ownVM: flags
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter ownVM: flags! !

!NewObjectMemory methodsFor: 'primitive support' stamp: 'eem 10/4/2014 12:48'!
pinObject: objOop
	"For forward-compatibility with Spur.  Fail; ObjectMemory does not support pinning."
	<api>
	coInterpreter primitiveFailFor: PrimErrUnsupported.
	^0! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/19/2013 15:58'!
pop: nItems
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter pop: nItems! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/19/2013 12:00'!
pop: nItems thenPush: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter pop: nItems thenPush: oop! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:15'!
positive32BitIntegerFor: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter positive32BitIntegerFor: integerValue! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:16'!
positive32BitValueOf: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter positive32BitValueOf: oop! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:20'!
positive64BitIntegerFor: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter positive64BitIntegerFor: integerValue! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:20'!
positive64BitValueOf: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter positive64BitValueOf: oop! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 11/19/2016 09:38'!
positiveMachineIntegerFor: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter positiveMachineIntegerFor: integerValue! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 11/19/2016 09:38'!
positiveMachineIntegerValueOf: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter positiveMachineIntegerValueOf: integerValue! !

!NewObjectMemory methodsFor: 'become' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
prepareForwardingTableForBecoming: array1 with: array2 twoWay: twoWayFlag 
	"Ensure that there are enough forwarding blocks to 
	accomodate this become, then prepare forwarding blocks for 
	the pointer swap. Return true if successful."
	"Details: Doing a GC might generate enough space for 
	forwarding blocks if we're short. However, this is an 
	uncommon enough case that it is better handled by primitive 
	fail code at the Smalltalk level."

	"Important note on multiple references to same object  - since the preparation of
	fwdBlocks is NOT idempotent we get VM crashes if the same object is referenced more
	than once in such a way as to require multiple fwdBlocks.
	oop1 forwardBecome: oop1 is ok since only a single fwdBlock is needed.
	oop1 become: oop1 would fail because the second fwdBlock woudl not have the actual object
	header but rather the mutated ref to the first fwdBlock.
	Further problems can arise with an array1 or array2 that refer multiply to the same 
	object. This would notbe expected input for programmer writen code but might arise from
	automatic usage such as in ImageSegment loading.
	To avoid the simple and rather common case of oop1 become*: oop1, we skip such pairs
	and simply avoid making fwdBlocks - it is redundant anyway"
	| entriesNeeded entriesAvailable fieldOffset oop1 oop2 fwdBlock fwdBlkSize |
	entriesNeeded := (self lastPointerOf: array1) // self wordSize. "need enough entries for all oops"
	"Note: Forward blocks must be quadword aligned - see fwdTableInit:."
	twoWayFlag
		ifTrue: ["Double the number of blocks for two-way become"
			entriesNeeded := entriesNeeded * 2.
			fwdBlkSize := self wordSize * 2]
		ifFalse: ["One-way become needs backPointers in fwd blocks."
			fwdBlkSize := self wordSize * 4].
	entriesAvailable := self fwdTableInit: fwdBlkSize.
	entriesAvailable < entriesNeeded ifTrue:
		[self initializeMemoryFirstFree: freeStart.
		 "re-initialize the free block"
		 ^false].
	fieldOffset := self lastPointerOf: array1.
	[fieldOffset >= self baseHeaderSize] whileTrue:
		[oop1 := self longAt: array1 + fieldOffset.
		 oop2 := self longAt: array2 + fieldOffset.
		 "if oop1 == oop2, no need to do any work for this pair.
		  May still be other entries in the arrays though so keep looking"
		 oop1 ~= oop2 ifTrue:
			[(self hasForwardingBlock: oop1) ifFalse: "Don't allocate multiple forwarding entries for duplicates."
				[fwdBlock := self fwdBlockGet: fwdBlkSize.
				 self
					initForwardBlock: fwdBlock
					mapping: oop1
					to: oop2
					withBackPtr: twoWayFlag not].
			 (twoWayFlag
			  and: [(self hasForwardingBlock: oop2) not]) ifTrue: "Again don't get confused by duplicates"
				["Second block maps oop2 back to oop1 for two-way become"
						fwdBlock := self fwdBlockGet: fwdBlkSize.
						self
							initForwardBlock: fwdBlock
							mapping: oop2
							to: oop1
							withBackPtr: twoWayFlag not]].
		fieldOffset := fieldOffset - self wordSize].
	^true! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 15:12'!
primitiveFail
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter primitiveFail! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 2/7/2014 09:59'!
primitiveFailFor: reasonCode
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter primitiveFailFor: reasonCode! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 8/22/2018 12:34'!
primitiveFailForFFIException: errorCode at: pc
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter primitiveFailForFFIException: errorCode at: pc! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:23'!
primitiveFailureCode
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter primitiveFailureCode! !

!NewObjectMemory methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:02'!
printActivationsOf: aMethodObj
	"Scan the heap printing the oops of any and all contexts that refer to anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| oop |
	oop := self firstAccessibleObject.
	[oop = nil] whileFalse:
		[((self isContextNonImm: oop)
		  and: [aMethodObj = (self fetchPointer: MethodIndex ofObject: oop)]) ifTrue:
			[coInterpreter
				printHex: oop; space; printOopShort: oop; print: ' pc ';
				printHex: (self fetchPointer: InstructionPointerIndex ofObject: oop); cr].
		 oop := self accessibleObjectAfter: oop]! !

!NewObjectMemory methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:02'!
printContextReferencesTo: anOop
	"Scan the heap printing the oops of any and all contexts that refer to anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| oop i |
	oop := self firstAccessibleObject.
	[oop = nil] whileFalse:
		[(self isContextNonImm: oop) ifTrue:
			[i := CtxtTempFrameStart + (coInterpreter fetchStackPointerOf: oop)].
			 [(i := i - 1) >= 0] whileTrue:
				[anOop = (self fetchPointer: i ofObject: oop) ifTrue:
					[coInterpreter
						printHex: oop; print: ' @ '; printNum: i; space; printOopShort: oop;
						print: ' pc '; printHex: (self fetchPointer: InstructionPointerIndex ofObject: oop); cr.
					 i := 0]].
		 oop := self accessibleObjectAfter: oop]! !

!NewObjectMemory methodsFor: 'debug printing' stamp: 'eem 12/20/2012 09:40'!
printHeaderTypeOf: obj
	(self headerType: obj) caseOf: {
		[HeaderTypeFree]			-> [coInterpreter print: ' HeaderTypeFree (4 bytes)'].
		[HeaderTypeShort]			-> [coInterpreter print: ' HeaderTypeShort (4 bytes)'].
		[HeaderTypeClass]			-> [coInterpreter print: ' HeaderTypeClass (8 bytes)'].
		[HeaderTypeSizeAndClass]	-> [coInterpreter print: ' HeaderTypeSizeAndClass (12 bytes)'] }! !

!NewObjectMemory methodsFor: 'debug printing' stamp: 'eem 7/3/2014 06:42'!
printMemField: memField name: name size: length
	<var: #memField type: #usqInt>
	<var: #name type: #'char *'>
	self print: name; tab; printHexPtr: memField asVoidPointer;
		printChar: $/; printNum: memField;
		print: ' sz: '; printHex: length.
	length ~= 0 ifTrue:
		[self printChar: $/; printNum: length].
	self cr! !

!NewObjectMemory methodsFor: 'debug printing' stamp: 'eem 7/8/2008 12:48'!
printMemory
	self printMemField: self startOfMemory name: 'memory      ' size: youngStart - self startOfMemory.
	self printMemField: youngStart name: 'youngStart  ' size: freeStart - youngStart.
	freeStart <= scavengeThreshold
		ifTrue:
			[self printMemField: freeStart name: 'freeStart   ' size: reserveStart - freeStart.
			 self printMemField: scavengeThreshold name: 'scavenge@   ' size: scavengeThreshold-freeStart]
		ifFalse:
			[self printMemField: scavengeThreshold name: 'scavenge@   ' size: scavengeThreshold-freeStart.
			 self printMemField: freeStart name: 'freeStart   ' size: reserveStart - freeStart].
	self printMemField: reserveStart name: 'reserveStart' size: endOfMemory - reserveStart.
	self printMemField: endOfMemory name: 'endOfMemory ' size: memoryLimit - endOfMemory.
	self printMemField: memoryLimit name: 'memoryLimit ' size: 0! !

!NewObjectMemory methodsFor: 'printing' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
printMemoryFrom: start to: end
	<doNotGenerate>
	| address |
	address := start bitAnd: (self wordSize - 1) bitInvert.
	[address < end] whileTrue:
		[coInterpreter printHex: address; printChar: $:; space; printHex: (self longAt: address); cr.
		 address := address + self wordSize]! !

!NewObjectMemory methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printObjectsFrom: startAddress to: endAddress
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| oop |
	oop := startAddress.
	[self oop: oop isLessThan: endAddress] whileTrue:
		[(self isFreeObject: oop) ifFalse:
			[coInterpreter printOop: oop].
		oop := self objectAfter: oop].! !

!NewObjectMemory methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:02'!
printReferencesTo: anOop
	"Scan the heap printing the oops of any and all objects that refer to anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| oop i |
	oop := self firstAccessibleObject.
	[oop = nil] whileFalse:
		[((self isPointersNonImm: oop) or: [self isCompiledMethod: oop]) ifTrue:
			[(self isCompiledMethod: oop)
				ifTrue:
					[i := (self literalCountOf: oop) + LiteralStart]
				ifFalse:
					[(self isContextNonImm: oop)
						ifTrue: [i := CtxtTempFrameStart + (coInterpreter fetchStackPointerOf: oop)]
						ifFalse: [i := self lengthOf: oop]].
			[(i := i - 1) >= 0] whileTrue:
				[anOop = (self fetchPointer: i ofObject: oop) ifTrue:
					[coInterpreter printHex: oop; print: ' @ '; printNum: i; space; printOopShort: oop; cr.
					 i := 0]]].
		 oop := self accessibleObjectAfter: oop]! !

!NewObjectMemory methodsFor: 'debug printing' stamp: 'eem 2/27/2013 12:14'!
printWronglySizedContexts
	"Override.  We implement printWronglySizedContexts:"
	<doNotGenerate>
	super printWronglySizedContexts! !

!NewObjectMemory methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:03'!
printWronglySizedContexts: printContexts
	"Scan the heap printing the oops of any and all contexts whose size is not either SmallContextSize or LargeContextSize"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32 and won't be written to cointerp.h"
	| oop |
	oop := self firstAccessibleObject.
	[oop = nil] whileFalse:
		[((self isContextNonImm: oop)
		   and: [self badContextSize: oop]) ifTrue:
			[self printHex: oop; space; printNum: (self numBytesOf: oop); cr.
			 printContexts ifTrue:
				[coInterpreter printContext: oop]].
		 oop := self accessibleObjectAfter: oop]! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:05'!
push: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter push: oop! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 14:59'!
pushBool: trueOrFalse
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter pushBool: trueOrFalse! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/24/2013 13:52'!
pushFloat: f
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter pushFloat: f! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:10'!
pushInteger: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter pushInteger: integerValue! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem 7/25/2014 21:47'!
rawClassTagForClass: classObj
	"Compatibility with SpurObjectMemory.  In ObjectMemory there is no distinction between a
	 classTag in the first-level method cache and a class itself."
	^classObj! !

!NewObjectMemory methodsFor: 'header access' stamp: 'eem 9/11/2013 21:52'!
rawHashBitsOf: objOop
	"Compatibility with Spur."
	^self hashBitsOf: objOop! !

!NewObjectMemory methodsFor: 'become' stamp: 'eem 2/20/2013 13:40'!
restoreHeaderOf: obj to: objHeader
	super restoreHeaderOf: obj to: objHeader.
	(self isContextHeader: objHeader) ifTrue:
		[(self asserta: ((coInterpreter isStillMarriedContext: obj) not)) ifFalse:
			[self error: 'attempt to become married context']]! !

!NewObjectMemory methodsFor: 'image segment in/out' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
restoreHeadersFrom: firstIn to: lastIn from: hdrBaseIn and: firstOut to: lastOut from: hdrBaseOut

	"Restore headers smashed by forwarding links"
	| tablePtr oop header |
	tablePtr := firstIn.
	[self oop: tablePtr isLessThanOrEqualTo: lastIn] whileTrue:
		[oop := self longAt: tablePtr.
		header := self longAt: hdrBaseIn + (tablePtr-firstIn).
		self longAt: oop put: header.
		tablePtr := tablePtr + self wordSize].
	tablePtr := firstOut.
	[self oop: tablePtr isLessThanOrEqualTo: lastOut] whileTrue:
		[oop := self longAt: tablePtr.
		header := self longAt: hdrBaseOut + (tablePtr-firstOut).
		self longAt: oop put: header.
		tablePtr := tablePtr + self wordSize].
	
	"Clear all mark bits"
	oop := self firstObject.
	[self oop: oop isLessThan: freeStart] whileTrue:
		[(self isFreeObject: oop) ifFalse:
			[self longAt: oop put: ((self longAt: oop) bitAnd: AllButMarkBit)].
		oop := self objectAfter: oop]! !

!NewObjectMemory methodsFor: 'snapshot' stamp: 'eem 10/9/2013 14:33'!
reverseBytesInMemory
	self reverseBytesFrom: self startOfMemory to: freeStart! !

!NewObjectMemory methodsFor: 'header access' stamp: 'eem 7/15/2008 16:03'!
rightType: headerWord
	"Compute the correct header type for an object based on the size and compact class fields of the given base header word, rather than its type bits. This is used during marking, when the header type bits are used to record the state of tracing."

	^(headerWord bitAnd: SizeMask) = 0  "zero size field in header word"
		ifTrue: [HeaderTypeSizeAndClass]
		ifFalse:
			[(headerWord bitAnd: CompactClassMask) = 0
				ifTrue: [HeaderTypeClass]
				ifFalse: [HeaderTypeShort]]! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem 1/5/2017 13:20'!
runLeakCheckerFor: gcModes
	<inline: false>
	(gcModes anyMask: checkForLeaks) ifTrue:
		[(gcModes anyMask: GCModeFull)
			ifTrue: [coInterpreter reverseDisplayFrom: 0 to: 7]
			ifFalse: [coInterpreter reverseDisplayFrom: 8 to: 15].
		 self clearLeakMapAndMapAccessibleObjects.
		 self assert: self checkHeapIntegrity.
		 self assert: coInterpreter checkInterpreterIntegrity = 0.
		 self assert: coInterpreter checkStackIntegrity.
		 self assert: (coInterpreter checkCodeIntegrity: gcModes).
		 self validate "simulation only"]! !

!NewObjectMemory methodsFor: 'object enumeration' stamp: 'eem 12/28/2019 13:35'!
safeLastPointerOf: objOop 
	"Return the byte offset of the last pointer field of the given object.  
	 Can be used even when the type bits are not correct.
	 Works with CompiledMethods, as well as ordinary objects."
	<api>
	<inline: true>
	| byteSize fmt header contextSize |
	byteSize := (self sizeBitsOfSafe: objOop) - self baseHeaderSize.
	header := self baseHeader: objOop.
	fmt := self formatOfHeader: header.
	fmt <= self lastPointerFormat ifTrue:
		[(fmt = self indexablePointersFormat
		  and: [self isContextHeader: header]) ifTrue:
			["contexts end at the stack pointer"
			contextSize := coInterpreter fetchStackPointerOf: objOop.
			^CtxtTempFrameStart + contextSize * self bytesPerOop min: byteSize].
		^byteSize  "all pointers"].
	fmt < self firstCompiledMethodFormat ifTrue: [^0]. "no pointers"

	"CompiledMethod: contains both pointers and bytes"
	header := self methodHeaderOf: objOop.
	^(self lastPointerOfMethodHeader: header) min: byteSize! !

!NewObjectMemory methodsFor: 'object enumeration' stamp: 'eem 12/4/2014 12:48'!
safeObjectAfter: oop 
	"Return the object or start of free space immediately following the 
	 given object or free chunk in memory. Return freeStart when
	 enumeration is complete.  This is for assertion checking only."
	| sz |
	(self isFreeObject: oop)
		ifTrue: [sz := self sizeOfFree: oop]
		ifFalse: [sz := self sizeBitsOf: oop].
	^(oop + sz) asUnsignedInteger >= freeStart
		ifTrue: [freeStart asInteger]
		ifFalse: [self oopFromChunk: oop + sz]! !

!NewObjectMemory methodsFor: 'debug printing' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
safePrintStringOf: oop
	"Version of printStringOf: that copes with forwarding during garbage collection."
	| fmt header cnt i |
	<inline: false>
	(self isIntegerObject: oop) ifTrue:
		[^nil].
	(self oop: oop isGreaterThanOrEqualTo: self startOfMemory andLessThan: freeStart) ifFalse:
		[^nil].
	(oop bitAnd: (self wordSize - 1)) ~= 0 ifTrue:
		[^nil].
	header := self headerWhileForwardingOf: oop.
	fmt := self formatOfHeader: header.
	fmt < 8 ifTrue: [ ^nil ].

	cnt := 100 min: (self lengthOf: oop baseHeader: header format: fmt).
	i := 0.

	[i < cnt] whileTrue:
		[self printChar: (self fetchByte: i ofObject: oop).
		 i := i + 1].
	coInterpreter flush.
	^oop! !

!NewObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
scavengeThreshold
	^scavengeThreshold! !

!NewObjectMemory methodsFor: 'garbage collection' stamp: 'eem 9/14/2013 08:53'!
scheduleIncrementalGC
	needGCFlag := true.
	coInterpreter forceInterruptCheck! !

!NewObjectMemory methodsFor: 'debug support' stamp: 'eem 12/29/2019 17:25'!
setCheckForLeaks: integerFlags
	" 0 = do nothing.
	  1 = check for leaks on fullGC (GCModeFull).
	  2 = check for leaks on scavenger (GCModeNewSpace).
	  4 = check for leaks on incremental (GCModeIncremental)
	  8 = check for leaks on become (GCModeBecome)
	 16 = check for leaks on image segments (GCModeImageSegment)
	 32 = check for leaks on simulated run-time call."
	checkForLeaks := integerFlags! !

!NewObjectMemory methodsFor: 'initialization' stamp: 'eem 1/17/2013 15:48'!
setEndOfMemory: newEndOfMemory
	super setEndOfMemory: newEndOfMemory.
	freeStart isNil ifTrue:
		[freeStart := newEndOfMemory]! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem 3/18/2020 12:56'!
shorten: obj toIndexableSize: nSlots
	"Reduce the number of indexable fields in obj, a pointer object, to nSlots. Convert the
	unused residual to a free chunk. Word and byte indexable objects are not changed.
	Answer the number of bytes returned to free memory, which may be zero if no change
	was possible."
	| deltaBytes desiredLength fixedFields fmt hdr totalLength
	 indexableFields |
	<api>
	(self isPointersNonImm: obj) ifFalse: [^0].
	nSlots >  0
		ifFalse: [^0]. "no change if nSlots is zero, error if nSlots is negative"
	hdr := self baseHeader: obj.
	fmt := self formatOfHeader: hdr.
	totalLength := self lengthOf: obj baseHeader: hdr format: fmt.
	fixedFields := self fixedFieldsOf: obj format: fmt length: totalLength.
	indexableFields := totalLength - fixedFields.
	nSlots >= indexableFields
		ifTrue: [^0]. "no change, or error if attempting to increase size into next chunk"
	desiredLength := fixedFields + nSlots.		
	deltaBytes := (totalLength - desiredLength) * self wordSize.
	obj + self baseHeaderSize + (totalLength * self wordSize) = freeStart
		ifTrue: "Shortening the last object.  Need to reduce freeStart."
			[freeStart := obj + self baseHeaderSize + (desiredLength * self wordSize)]
		ifFalse: "Shortening some interior object.  Need to create a free block."
			[self setSizeOfFree: obj + self baseHeaderSize + (desiredLength * self wordSize)
				to: deltaBytes].
	(self headerType: obj) caseOf:	{
		[HeaderTypeSizeAndClass] ->
			[self longAt: (obj - (self baseHeaderSize * 2)) put: (self sizeHeader: obj) - deltaBytes].
		[HeaderTypeClass] ->
			[self longAt: obj put: ((hdr bitClear: SizeMask) bitOr: (hdr bitAnd: SizeMask) - deltaBytes)].
		[HeaderTypeShort] ->
			[self longAt: obj put: ((hdr bitClear: SizeMask) bitOr: (hdr bitAnd: SizeMask) - deltaBytes)] }.
	^deltaBytes! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 15:40'!
showDisplayBits: aForm Left: l Top: t Right: r Bottom: b
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter showDisplayBits: aForm Left: l Top: t Right: r Bottom: b! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem 4/4/2009 20:11'!
shrinkObjectMemory: delta 
	"Attempt to shrink the object memory by the given delta 
	amount "
	| limit |
	statShrinkMemory := statShrinkMemory + 1. 
	limit := self sqShrinkMemory: memoryLimit By: delta.
	limit = memoryLimit ifFalse:
		[self setMemoryLimit: limit - 24. "remove a tad for safety"
		 self initializeMemoryFirstFree: freeStart]! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 12/30/2016 14:59'!
signalSemaphoreWithIndex: index
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter signalSemaphoreWithIndex: index! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 1/24/2014 12:41'!
signed32BitIntegerFor: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter signed32BitIntegerFor: integerValue! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 2/18/2014 15:05'!
signed32BitValueOf: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter signed32BitValueOf: oop! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 12/21/2017 20:39'!
signed64BitIntegerFor: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter signed64BitIntegerFor: integerValue! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 12/21/2017 15:17'!
signed64BitValueOf: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter signed64BitValueOf: oop! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 1/24/2015 16:41'!
signedMachineIntegerValueOf: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter signedMachineIntegerValueOf: oop! !

!NewObjectMemory methodsFor: 'object access' stamp: 'eem 4/30/2009 11:32'!
sizeBitsOf: oop
	"Answer the number of bytes in the given object, including its base header, rounded up to an integral number of words."
	"Note: byte indexable objects need to have low bits subtracted from this size."
	<inline: true>
	| header |
	header := self baseHeader: oop.
	^(header bitAnd: TypeMask) = HeaderTypeSizeAndClass
		ifTrue: [(self sizeHeader: oop) bitAnd: LongSizeMask]
		ifFalse: [header bitAnd: SizeMask]! !

!NewObjectMemory methodsFor: 'object access' stamp: 'eem 7/15/2008 16:01'!
sizeBitsOfSafe: oop
	"Compute the size of the given object from the cc and size fields in its header.
	 This works even if its type bits are not correct."

	| header type |
	header := self baseHeader: oop.
	type := self rightType: header.
	^type = HeaderTypeSizeAndClass
		ifTrue: [(self sizeHeader: oop) bitAnd: AllButTypeMask]
		ifFalse: [header bitAnd: SizeMask]! !

!NewObjectMemory methodsFor: 'interpreter access' stamp: 'eem 10/27/2020 21:43'!
sizeFieldOfAlien: alienObj
	"Answer the first field of alienObj which is assumed to be an Alien of at least 8 bytes"
	<inline: #always>
	^self longAt: alienObj + self baseHeaderSize! !

!NewObjectMemory methodsFor: 'primitive support' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
sizeHeader: obj putBodySize: byteSize
	"Too lazy to set the odd bits right now.  Just insist on a multiple of 4 bytes."
	self assert: byteSize \\ 4 = 0.
	self longAt: obj - (self wordSize*2)
		put: ((byteSize + self wordSize bitAnd: LongSizeMask)
				bitOr: ((self longAt: obj - (self wordSize*2)) bitClear: LongSizeMask))! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 2/21/2018 15:19'!
sizeOfSTArrayFromCPrimitive: cPtr
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter sizeOfSTArrayFromCPrimitive: cPtr! !

!NewObjectMemory methodsFor: 'cog jit support' stamp: 'eem 1/8/2015 14:36'!
smallIntegerTag
	^1! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/23/2013 14:24'!
stObject: objOop at: indexOop put: valueOop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stObject: objOop at: indexOop put: valueOop! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/24/2013 13:51'!
stackFloatValue: offset
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stackFloatValue: offset! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:07'!
stackIntegerValue: offset
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stackIntegerValue: offset! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 18:05'!
stackObjectValue: offset
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stackObjectValue: offset! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 10/27/2020 20:51'!
stackPositiveMachineIntegerValue: offset
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stackPositiveMachineIntegerValue: offset! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/19/2013 11:06'!
stackValue: offset
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stackValue: offset! !

!NewObjectMemory methodsFor: 'memory access' stamp: 'eem 4/30/2009 10:20'!
startOfFreeSpace
	<inline: true>
	^freeStart! !

!NewObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statGCEndUsecs
	^statGCEndUsecs! !

!NewObjectMemory methodsFor: 'image segment in/out' stamp: 'eem 9/7/2018 19:23'!
storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots

"This primitive will store a binary image segment (in the same format as the Squeak image file) of the receiver and every object in its proper tree of subParts (ie, that is not refered to from anywhere else outside the tree).  All pointers from within the tree to objects outside the tree will be copied into the array of outpointers.  In their place in the image segment will be an oop equal to the offset in the outPointer array (the first would be 4). but with the high bit set."

"The primitive expects the array and wordArray to be more than adequately long.  In this case it returns normally, and truncates the two arrays to exactly the right size.  To simplify truncation, both incoming arrays are required to be 256 bytes or more long (ie with 3-word headers).  If either array is too small, the primitive will fail, but in no other case.

During operation of the primitive, it is necessary to convert from both internal and external oops to their mapped values.  To make this fast, the headers of the original objects in question are replaced by the mapped values (and this is noted by adding the forbidden XX header type).  Tables are kept of both kinds of oops, as well as of the original headers for restoration.

To be specific, there are two similar two-part tables, the outpointer array, and one in the upper fifth of the segmentWordArray.  Each grows oops from the bottom up, and preserved headers from halfway up.

In case of either success or failure, the headers must be restored.  In the event of primitive failure, the table of outpointers must also be nilled out (since the garbage in the high half will not have been discarded."

	| savedYoungStart lastOut lastIn firstIn lastSeg endSeg segOop fieldPtr fieldOop mapOop
	  doingClass lastPtr extraSize hdrTypeBits hdrBaseIn hdrBaseOut header firstOut versionOffset |
	<inline: false>
	<var: #firstIn type: #usqInt>
	<var: #lastIn type: #usqInt>
	<var: #firstOut type: #usqInt>
	<var: #lastOut type: #usqInt>
	<var: #hdrBaseIn type: #usqInt>
	<var: #hdrBaseOut type: #usqInt>
	<var: #lastSeg type: #usqInt>
	<var: #endSeg type: #usqInt>
	<var: #fieldPtr type: #usqInt>
	<var: #lastPtr type: #usqInt>
	<var: #segOop type: #usqInt>
	<var: #savedYoungStart type: #usqInt>

	((self headerType: outPointerArray) = HeaderTypeSizeAndClass			"Must be 3-word header"
	and: [(self headerType: segmentWordArray) = HeaderTypeSizeAndClass])	"Must be 3-word header"
		ifFalse: [^PrimErrGenericFailure].

	DoAssertionChecks ifTrue: [self verifyCleanHeaders].
	"Use the top half of outPointers for saved headers."
	firstOut := outPointerArray + self baseHeaderSize.
	lastOut := firstOut - self wordSize.
	hdrBaseOut := outPointerArray + ((self lastPointerOf: outPointerArray) // (self wordSize*2) * self wordSize). "top half"

	lastSeg := segmentWordArray.
	endSeg := segmentWordArray + (self sizeBitsOf: segmentWordArray) - self wordSize.

	"Write a version number for byte order and version check"
	versionOffset := self wordSize.
	lastSeg := lastSeg + versionOffset.
	lastSeg > endSeg ifTrue: [^PrimErrGenericFailure].
	self longAt: lastSeg put: self imageSegmentVersion.

	"Allocate top 1/8 of segment for table of internal oops and saved headers"
	firstIn := endSeg - ((self sizeBitsOf: segmentWordArray) // (self wordSize*8) * self wordSize).  "Take 1/8 of seg"
	lastIn := firstIn - self wordSize.
	hdrBaseIn := firstIn + ((self sizeBitsOf: segmentWordArray) // (self wordSize*16) * self wordSize). "top half of that"

	"First mark the rootArray and all root objects."
	self longAt: arrayOfRoots put: ((self longAt: arrayOfRoots) bitOr: MarkBit).
	lastPtr := arrayOfRoots + (self lastPointerOf: arrayOfRoots).
	fieldPtr := arrayOfRoots + self baseHeaderSize.
	[fieldPtr <= lastPtr] whileTrue:
		[fieldOop := self longAt: fieldPtr.
		(self isIntegerObject: fieldOop) ifFalse:
			[self longAt: fieldOop put: ((self longAt: fieldOop) bitOr: MarkBit)].
		fieldPtr := fieldPtr + self wordSize].

	"Then do a mark pass over all objects.  This will stop at our marked roots,
	thus leaving our segment unmarked in their shadow."
	savedYoungStart := youngStart.
	youngStart := self startOfMemory.  "process all of memory"
	coInterpreter markAndTraceInterpreterOops: false.	"and special objects array"
	youngStart := savedYoungStart.
	
	"Finally unmark the rootArray and all root objects."
	self longAt: arrayOfRoots put: ((self longAt: arrayOfRoots) bitAnd: AllButMarkBit).
	fieldPtr := arrayOfRoots + self baseHeaderSize.
	[fieldPtr <= lastPtr] whileTrue:
		[fieldOop := self longAt: fieldPtr.
		(self isIntegerObject: fieldOop) ifFalse:
			[self longAt: fieldOop put: ((self longAt: fieldOop) bitAnd: AllButMarkBit)].
		fieldPtr := fieldPtr + self wordSize].

	"All external objects, and only they, are now marked.
	Copy the array of roots into the segment, and forward its oop."
	lastIn := lastIn + self wordSize.
	(lastIn >= hdrBaseIn
	 or: [0 = (lastSeg := self copyObj: arrayOfRoots toSegment: segmentWordArray addr: lastSeg stopAt: firstIn saveOopAt: lastIn headerAt: hdrBaseIn + (lastIn - firstIn))]) ifTrue:
		[lastIn := lastIn - self wordSize.
		self restoreHeadersFrom: firstIn to: lastIn from: hdrBaseIn and: firstOut to: lastOut from: hdrBaseOut.
		^self primitiveFailCodeAfterCleanup: outPointerArray].

	"Now run through the segment fixing up all the pointers.
	Note that more objects will be added to the segment as we make our way along."
	segOop := self oopFromChunk: segmentWordArray + versionOffset + self baseHeaderSize.
	[segOop <= lastSeg] whileTrue:
		[(self headerType: segOop) <= 1
			ifTrue: ["This object has a class field (type=0 or 1) -- start with that."
					fieldPtr := segOop - self wordSize.  doingClass := true]
			ifFalse: ["No class field -- start with first data field"
					fieldPtr := segOop + self baseHeaderSize.  doingClass := false].
		lastPtr := segOop + (self lastPointerOf: segOop).	"last field"

		"Go through all oops, remapping them..."
		[fieldPtr > lastPtr] whileFalse:
			["Examine each pointer field"
			fieldOop := self longAt: fieldPtr.
			doingClass ifTrue:
				[hdrTypeBits := fieldOop bitAnd: TypeMask.
				fieldOop := fieldOop - hdrTypeBits].
			(self isIntegerObject: fieldOop)
				ifTrue: ["Just an integer -- nothing to do"
						fieldPtr := fieldPtr + self wordSize]
				ifFalse:
				[header := self longAt: fieldOop.
				(header bitAnd: TypeMask) = HeaderTypeFree
					ifTrue: ["Has already been forwarded -- this is the link"
							mapOop := header bitAnd: AllButTypeMask]
					ifFalse:
					[((self longAt: fieldOop) bitAnd: MarkBit) = 0
						ifTrue:
							["Points to an unmarked obj -- an internal pointer.
							Copy the object into the segment, and forward its oop."
							lastIn := lastIn + self wordSize.
							(lastIn >= hdrBaseIn
							or: [0 = (lastSeg := self copyObj: fieldOop toSegment: segmentWordArray addr: lastSeg stopAt: firstIn saveOopAt: lastIn headerAt: hdrBaseIn + (lastIn - firstIn))]) ifTrue:
								["Out of space in segment"
								lastIn := lastIn - self wordSize.
								self restoreHeadersFrom: firstIn to: lastIn from: hdrBaseIn and: firstOut to: lastOut from: hdrBaseOut.
								^self primitiveFailCodeAfterCleanup: outPointerArray].
							mapOop := (self longAt: fieldOop) bitAnd: AllButTypeMask]
						ifFalse:
							["Points to a marked obj -- an external pointer.
							Map it as a tagged index in outPointers, and forward its oop."
							lastOut := lastOut + self wordSize.
							lastOut >= hdrBaseOut ifTrue:
								["Out of space in outPointerArray"
								lastOut := lastOut - self wordSize.
								self restoreHeadersFrom: firstIn to: lastIn from: hdrBaseIn and: firstOut to: lastOut from: hdrBaseOut.
								^self primitiveFailCodeAfterCleanup: outPointerArray].
.							mapOop := lastOut - outPointerArray bitOr: 16r80000000.
							self forward: fieldOop to: mapOop
								savingOopAt: lastOut andHeaderAt: hdrBaseOut + (lastOut - firstOut)]].
					"Replace the oop by its mapped value"
					doingClass
						ifTrue: [self longAt: fieldPtr put: mapOop + hdrTypeBits.
								fieldPtr := fieldPtr + (self wordSize*2).
								doingClass := false]
						ifFalse: [self longAt: fieldPtr put: mapOop.
								fieldPtr := fieldPtr + self wordSize]]].
		segOop := self objectAfter: segOop].

	self restoreHeadersFrom: firstIn to: lastIn from: hdrBaseIn and: firstOut to: lastOut from: hdrBaseOut.

	"Truncate the outPointerArray..."
	((outPointerArray + (self lastPointerOf: outPointerArray) - lastOut) < 12
		or: [(endSeg - lastSeg) < 12]) ifTrue:
			["Not enough room to insert simple 3-word headers"
			^self primitiveFailCodeAfterCleanup: outPointerArray].
	extraSize := self extraHeaderBytes: segmentWordArray.
	hdrTypeBits := self headerType: segmentWordArray.
	"Copy the 3-word wordArray header to establish a free chunk."
	self transfer: 3
		from: segmentWordArray - extraSize
		to: lastOut+self wordSize.
	"Adjust the size of the original as well as the free chunk."
	self longAt: lastOut+self wordSize
		put: outPointerArray + (self lastPointerOf: outPointerArray) - lastOut - extraSize + hdrTypeBits.
	self longAt: outPointerArray-extraSize
		put: lastOut - firstOut + (self wordSize*2) + hdrTypeBits.
	"Note that pointers have been stored into roots table"
	self beRootIfOld: outPointerArray.

	"Truncate the image segment..."
	"Copy the 3-word wordArray header to establish a free chunk."
	self transfer: 3
		from: segmentWordArray - extraSize
		to: lastSeg+self wordSize.
	"Adjust the size of the original as well as the free chunk."
	self longAt: segmentWordArray-extraSize
		put: lastSeg - segmentWordArray + self baseHeaderSize + hdrTypeBits.
	self longAt: lastSeg+self wordSize
		put: endSeg - lastSeg - extraSize + hdrTypeBits.

	DoAssertionChecks ifTrue: [self verifyCleanHeaders].
	^PrimNoErr! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/21/2013 07:15'!
storeInteger: fieldIndex ofObject: objectPointer withValue: integerValue
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter storeInteger: fieldIndex ofObject: objectPointer withValue: integerValue! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 15:57'!
stringOf: oop
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter stringOf: oop! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/19/2013 11:07'!
success: boolean
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter success: boolean! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem 6/27/2011 14:32'!
sufficientSpaceAfterGC: minFree 
	"Return true if there is enough free space after doing a garbage collection. If not, signal that space is low."
	<inline: false>
	| oldLimit |
	self incrementalGC. "try to recover some space"

	(self oop: freeStart + minFree isLessThanOrEqualTo: reserveStart) ifTrue:
		[^true].

	signalLowSpace ifTrue:
		[^false]. "give up; problem is already noted"

	self fullGC. "try harder"
	"for stability, require more free space after doing an expensive full GC"
	(self oop: freeStart + minFree + 15000 isLessThan: reserveStart) ifTrue:
		[^true].

	"still not enough; attempt to grow object memory"

	[oldLimit := memoryLimit.
	 self growObjectMemory: minFree + 15000 + growHeadroom.
	 self oop: freeStart + minFree + 15000 isLessThan: reserveStart]
		whileFalse: "If the memoryLimit doesn't change then presumably the platform can't grow anymore." 
			[oldLimit = memoryLimit ifTrue: [^false]].
	^true! !

!NewObjectMemory methodsFor: 'allocation' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
sufficientSpaceToAllocate: bytes
	"Return true if there is enough space to allocate the given number of bytes, perhaps after doing a garbage collection."

	| minFree |
	<inline: true>
	minFree := (lowSpaceThreshold + bytes + self baseHeaderSize + self wordSize - 1) bitClear: self wordSize - 1.

	"check for low-space"
	(self oop: freeStart + minFree isLessThanOrEqualTo: reserveStart) ifTrue:
		[^true].
	^self sufficientSpaceAfterGC: minFree! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 2/3/2015 18:36'!
superclassOf: classPointer
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter superclassOf: classPointer! !

!NewObjectMemory methodsFor: 'gc -- mark and sweep' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
sweepPhase
	"Sweep memory from youngStart through the end of memory. Free all 
	inaccessible objects and coalesce adjacent free chunks. Clear the mark 
	bits of accessible objects. Compute the starting point for the first pass of 
	incremental compaction (compStart). Return the number of surviving 
	objects. "
	"Details: Each time a non-free object is encountered, decrement the 
	number of available forward table entries. If all entries are spoken for 
	(i.e., entriesAvailable reaches zero), set compStart to the last free 
	chunk before that object or, if there is no free chunk before the given 
	object, the first free chunk after it. Thus, at the end of the sweep 
	phase, compStart through compEnd spans the highest collection of 
	non-free objects that can be accomodated by the forwarding table. This 
	information is used by the first pass of incremental compaction to 
	ensure that space is initially freed at the end of memory. Note that 
	there should always be at least one free chunk--the one at the end of 
	the heap."
	| entriesAvailable survivors freeChunk firstFree oop oopHeader oopHeaderType hdrBytes oopSize freeChunkSize freeStartLocal |
	<inline: false>
	<var: #oop type: #usqInt>
	<var: #freeStartLocal type: #usqInt>
	entriesAvailable := self fwdTableInit: self wordSize*2.
	survivors := 0.
	freeChunk := nil.
	firstFree := nil.
	"will be updated later"
	freeStartLocal := freeStart.
	oop := self oopFromChunk: youngStart.
	[oop < freeStartLocal]
		whileTrue: ["get oop's header, header type, size, and header size"
			statSweepCount := statSweepCount + 1.
			oopHeader := self baseHeader: oop.
			oopHeaderType := oopHeader bitAnd: TypeMask.
			hdrBytes := headerTypeBytes at: oopHeaderType.
			(oopHeaderType bitAnd: 1) = 1
				ifTrue: [oopSize := oopHeader bitAnd: SizeMask]
				ifFalse: [oopHeaderType = HeaderTypeSizeAndClass
						ifTrue: [oopSize := (self sizeHeader: oop) bitAnd: LongSizeMask]
						ifFalse: ["free chunk" oopSize := oopHeader bitAnd: LongSizeMask]].
			(oopHeader bitAnd: MarkBit) = 0
				ifTrue: ["object is not marked; free it"
					"<-- Finalization support: We need to mark each oop chunk as free -->"
					self longAt: oop - hdrBytes put: HeaderTypeFree.
					freeChunk ~= nil
						ifTrue: ["enlarge current free chunk to include this oop"
							freeChunkSize := freeChunkSize + oopSize + hdrBytes]
						ifFalse: ["start a new free chunk"
							freeChunk := oop - hdrBytes.
							"chunk may start 4 or 8 bytes before oop"
							freeChunkSize := oopSize + (oop - freeChunk).
							"adjust size for possible extra header bytes"
							firstFree = nil ifTrue: [firstFree := freeChunk]]]
				ifFalse: ["object is marked; clear its mark bit and possibly adjust 
					the compaction start"
					self longAt: oop put: (oopHeader bitAnd: AllButMarkBit).
					"<-- Finalization support: Check if we're running about a weak class -->"
					(self isWeakNonImm: oop) ifTrue: [self finalizeReference: oop].
					entriesAvailable > 0
						ifTrue: [entriesAvailable := entriesAvailable - 1]
						ifFalse: ["start compaction at the last free chunk before this object"
							firstFree := freeChunk].
					freeChunk ~= nil
						ifTrue: ["record the size of the last free chunk"
							self longAt: freeChunk put: ((freeChunkSize bitAnd: LongSizeMask) bitOr: HeaderTypeFree).
							freeChunk := nil].
					survivors := survivors + 1].
			oopSize = 0 ifTrue:
				[self error: 'zero sized object encountered in sweep'].
			oop := self oopFromChunk: oop + oopSize].
	freeChunk ~= nil
		ifTrue: ["record size of final free chunk"
			self longAt: freeChunk put: ((freeChunkSize bitAnd: LongSizeMask) bitOr: HeaderTypeFree)].
	oop = freeStart
		ifFalse: [self error: 'sweep failed to find exact end of memory'].
	compStart := firstFree = nil
					ifTrue: [freeStart]
					ifFalse: [firstFree].
	^survivors! !

!NewObjectMemory methodsFor: 'garbage collection' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
sweepPhaseForFullGC
	"Sweep memory from youngStart through the end of memory. Free all
	 inaccessible objects and coalesce adjacent free chunks. Clear the mark
	 bits of accessible objects. Compute the starting point for the first pass
	 of incremental compaction (compStart). Return the number of surviving
	 objects.  Unlike sweepPhase this always leaves compStart pointing at the
	 first free chunk."
	| survivors freeChunk firstFree oop oopHeader oopHeaderType hdrBytes oopSize freeChunkSize endOfMemoryLocal |
	<inline: false>
	<var: #oop type: #usqInt>
	<var: #endOfMemoryLocal type: #usqInt>
	self fwdTableInit: self wordSize*2.
	survivors := 0.
	freeChunk := nil.
	firstFree := nil.
	"will be updated later"
	endOfMemoryLocal := endOfMemory.
	oop := self oopFromChunk: youngStart.
	[oop < endOfMemoryLocal]
		whileTrue: ["get oop's header, header type, size, and header size"
			statSweepCount := statSweepCount + 1.
			oopHeader := self baseHeader: oop.
			oopHeaderType := oopHeader bitAnd: TypeMask.
			hdrBytes := headerTypeBytes at: oopHeaderType.
			(oopHeaderType bitAnd: 1) = 1
				ifTrue: [oopSize := oopHeader bitAnd: SizeMask]
				ifFalse: [oopHeaderType = HeaderTypeSizeAndClass
						ifTrue: [oopSize := (self sizeHeader: oop) bitAnd: LongSizeMask]
						ifFalse: [self assert: (oopHeader bitAnd: MarkBit) = 0.
								oopSize := oopHeader bitAnd: LongSizeMask]].
			(oopHeader bitAnd: MarkBit) = 0
				ifTrue: ["object is not marked; free it"
					"<-- Finalization support: We need to mark each oop chunk as free -->"
					self longAt: oop - hdrBytes put: HeaderTypeFree.
					freeChunk ~= nil
						ifTrue: ["enlarge current free chunk to include this oop"
							freeChunkSize := freeChunkSize + oopSize + hdrBytes]
						ifFalse: ["start a new free chunk"
							freeChunk := oop - hdrBytes.
							"chunk may start 4 or 8 bytes before oop"
							freeChunkSize := oopSize + (oop - freeChunk).
							"adjust size for possible extra header bytes"
							firstFree = nil ifTrue: [firstFree := freeChunk]]]
				ifFalse: ["object is marked; clear its mark bit and possibly adjust 
					the compaction start"
					self longAt: oop put: (oopHeader bitAnd: AllButMarkBit).
					"<-- Finalization support: Check if we're running about a weak class -->"
					(self isWeakNonImm: oop) ifTrue: [self finalizeReference: oop].
					freeChunk ~= nil
						ifTrue: ["record the size of the last free chunk"
							self longAt: freeChunk put: ((freeChunkSize bitAnd: LongSizeMask) bitOr: HeaderTypeFree).
							freeChunk := nil].
					survivors := survivors + 1].
			oop := self oopFromChunk: oop + oopSize].
	freeChunk ~= nil
		ifTrue: ["record size of final free chunk"
			self longAt: freeChunk put: ((freeChunkSize bitAnd: LongSizeMask) bitOr: HeaderTypeFree)].
	oop = endOfMemory
		ifTrue: [freeStart := endOfMemory]
		ifFalse: [self error: 'sweep failed to find exact end of memory'].
	firstFree = nil
		ifTrue: [self error: 'expected to find at least one free object']
		ifFalse: [compStart := firstFree].

	^ survivors! !

!NewObjectMemory methodsFor: 'simulation only' stamp: 'eem 9/20/2013 15:13'!
transcript
	"hack around the CoInterpreter/ObjectMemory split refactoring"
	<doNotGenerate>
	^coInterpreter transcript! !

!NewObjectMemory methodsFor: 'accessing' stamp: 'eem 9/30/2010 12:27'!
trueObject: anOop
	"For mapInterpreterOops"
	trueObj := anOop! !

!NewObjectMemory methodsFor: 'primitive support' stamp: 'eem 10/13/2014 16:16'!
unmarkAllObjects
	| oop hdr |
	oop := self firstObject.
	[oop asUnsignedInteger < freeStart] whileTrue:
		[(self isFreeObject: oop) ifFalse:
			[hdr := self baseHeader: oop.
			 (hdr bitAnd: MarkBit) ~= 0 ifTrue:
				[self baseHeader: oop put: (hdr bitAnd: AllButMarkBit)]].
		 oop := self objectAfter: oop]! !

!NewObjectMemory methodsFor: 'primitive support' stamp: 'eem 7/28/2014 07:13'!
unpinObject: objOop
	"For forward-compatibility with Spur.  Fail; ObjectMemory does not support pinning."
	<api>
	coInterpreter primitiveFailFor: PrimErrUnsupported! !

!NewObjectMemory methodsFor: 'simulation' stamp: 'eem 10/12/2014 17:32'!
validate
	"Validate all the objects in the heap."
	self cCode: [] inSmalltalk: [
	| oop prev |
	self interpreter transcript show: 'Validating...'.
	oop := self firstObject.
	[oop < freeStart] whileTrue:
		[self validate: oop.
		prev := oop.  "look here if debugging prev obj overlapping this one"
		oop := self safeObjectAfter: oop].
	self touch: prev.  "Don't offer to delete this please"
	self interpreter transcript show: 'done.'; cr]! !

!NewObjectMemory methodsFor: 'simulation' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
validate: oop
	<doNotGenerate>
	| header type cc sz fmt nextChunk | 
	header := self longAt: oop.
	type := header bitAnd: 3.
	type = 2 ifFalse: [type = (self rightType: header) ifFalse: [self halt]].
	sz := (header bitAnd: SizeMask) >> 2.
	nextChunk := oop + ((self isFreeObject: oop)
							ifTrue: [self sizeOfFree: oop]
							ifFalse: [self sizeBitsOf: oop]).
	nextChunk >= freeStart
		ifTrue:
			[nextChunk ~= freeStart ifTrue: [self halt]]
		ifFalse:
			[(self headerType: nextChunk) = 0 ifTrue:
				[(self headerType: (nextChunk + (self wordSize*2))) = 0 ifFalse: [self halt]].
			(self headerType: nextChunk) = 1 ifTrue:
				[(self headerType: (nextChunk + self wordSize)) = 1 ifFalse: [self halt]]].
	type = 2 ifTrue: "free block"
		[^self].
	fmt := self formatOfHeader: header.
	cc := self compactClassIndexOfHeader: header.
	cc > 16 ifTrue: [self halt].	"up to 32 are legal, but not used"
	type = 0 ifTrue:
		["three-word header"
		((self longAt: oop-self wordSize) bitAnd: 3) = type ifFalse: [self halt].
		((self longAt: oop-(self wordSize*2)) bitAnd: 3) = type ifFalse: [self halt].
		((self longAt: oop-self wordSize) = type) ifTrue: [self halt].	"Class word is 0"
		sz = 0 ifFalse: [self halt]].
	type = 1 ifTrue:
		["two-word header"
		((self longAt: oop-self wordSize) bitAnd: 3) = type ifFalse: [self halt].
		cc > 0 ifTrue: [sz = 1 ifFalse: [self halt]].
		sz = 0 ifTrue: [self halt]].
	type = 3 ifTrue:
		["one-word header"
		cc = 0 ifTrue: [self halt]].
	fmt = 5 ifTrue: [self halt].
	fmt = 7 ifTrue: [self halt].
	fmt >= self firstCompiledMethodFormat ifTrue: "must have integer header"
		[(self isIntegerObject: (self longAt: oop + self wordSize)) ifFalse: [self halt]].! !

!NewObjectMemory methodsFor: 'image segment in/out' stamp: 'eem 12/19/2008 22:17'!
verifyCleanHeaders
	| oop |
	oop := self firstObject.
	[self oop: oop isLessThan: freeStart] whileTrue:
		[(self isFreeObject: oop)
			ifTrue: ["There should only be one free block at end of memory."
					(self objectAfter: oop) = endOfMemory
						ifFalse: [self error: 'Invalid obj with HeaderTypeBits = Free.']]
			ifFalse: [(self isMarked: oop) ifTrue:
						[self error: 'Invalid obj with MarkBit set.']].
		oop := self objectAfter: oop]! !

!NewObjectMemory methodsFor: 'debug printing' stamp: 'eem 3/24/2014 17:21'!
whereIsMaybeHeapThing: anOop
	<returnTypeC: 'char *'>
	(self oop: anOop isGreaterThanOrEqualTo: self startOfMemory andLessThan: endOfMemory) ifTrue:
		[(self oop: anOop isLessThan: freeStart) ifTrue:
			[^' is in old space'].
		 (self oop: anOop isLessThan: reserveStart) ifTrue:
			[^' is in new space'].
		 ^' is above reserve start'].
	^nil! !

!NewObjectMemory methodsFor: 'image segment in/out' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
wordSwapFloatsFrom: startOop to: stopAddr
	"Swap the most and least significant words of Floats in a range of the image."
	| oop temp |
	oop := startOop.
	[self oop: oop isLessThan: stopAddr] whileTrue:
		[(self isFreeObject: oop) ifFalse:
			[((self formatOf: oop) = self firstLongFormat
			 and: [(self compactClassIndexOf: oop) = ClassFloatCompactIndex]) ifTrue:
				[temp := self longAt: oop + self baseHeaderSize.
				 self longAt: oop + self baseHeaderSize put: (self longAt: oop + self baseHeaderSize + 4).
				 self longAt: oop + self baseHeaderSize + 4 put: temp]].
		oop := self objectAfter: oop]! !

!NewObjectMemory methodsFor: 'cog jit support' stamp: 'eem 6/17/2010 12:40'!
youngStart
	<cmacro: '() GIV(youngStart)'>
	^youngStart! !

NewObjectMemory subclass: #NewObjectMemorySimulator
	instanceVariableNames: 'parent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!NewObjectMemorySimulator class methodsFor: 'instance creation' stamp: 'eem 9/30/2010 13:30'!
new
	^self == NewObjectMemorySimulator
		ifTrue: [SmalltalkImage current endianness == #big
				ifTrue: [NewObjectMemorySimulatorMSB new]
				ifFalse: [NewObjectMemorySimulatorLSB new]]
		ifFalse: [super new]! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'eem 11/6/2014 11:03'!
baseHeaderSize
	^4! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'th 4/15/2000 17:21'!
byteAt: byteAddress
	^self subclassResponsibility! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'th 4/15/2000 17:21'!
byteAt: byteAddress put: byte
	^self subclassResponsibility! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 6/23/2004 14:09'!
byteAtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	pointer is a raw address, and byte is an 8-bit quantity."

	^ self byteAt: pointer! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 6/23/2004 14:13'!
byteAtPointer: pointer put: byteValue
	"This gets implemented by Macros in C, where its types will also be checked.
	pointer is a raw address, and byteValue is an 8-bit quantity."

	^ self byteAt: pointer  put: byteValue! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'eem 11/6/2014 11:06'!
bytesPerOop
	^4! !

!NewObjectMemorySimulator methodsFor: 'initialization' stamp: 'eem 6/9/2010 18:44'!
coInterpreter: aCoInterpreter
	coInterpreter := aCoInterpreter! !

!NewObjectMemorySimulator methodsFor: 'simulation only' stamp: 'eem (const to msg rewrite) 11/13/2014 14:07'!
firstIndexableField: oop
	"NOTE: overridden from ObjectMemory to add coercion to CArray, so please duplicate any changes"
	| hdr fmt totalLength fixedFields |
	<returnTypeC: #'void *'>
	hdr := self baseHeader: oop.
	fmt := self formatOfHeader: hdr.
	fmt <= self lastPointerFormat ifTrue:
		["pointer; may need to delve into the class format word"
		totalLength := self lengthOf: oop baseHeader: hdr format: fmt.
		fixedFields := self fixedFieldsOf: oop format: fmt length: totalLength.
		^self cCoerce: (self pointerForOop: oop + self baseHeaderSize + (fixedFields << self shiftForWord)) to: 'oop *'].
	^self
		cCoerce: (self pointerForOop: oop + self baseHeaderSize)
		to: (fmt < self firstByteFormat
				ifTrue: [fmt = self firstLongFormat
						ifTrue: ["32 bit field objects" 'int *']
						ifFalse: ["full word objects (bits)" 'oop *']]
				ifFalse: ["byte objects (including CompiledMethod" 'char *'])! !

!NewObjectMemorySimulator methodsFor: 'gc -- mark and sweep' stamp: 'eem 5/30/2018 11:23'!
fullGC
	"If we're /not/ a clone, clone the VM and push it over the cliff.
	 If it survives, destroy the clone and continue.  We should be OK until next time."
	parent ifNil:
		[coInterpreter cr; print: 'GC number '; print: statFullGCs; tab; flush.
		 CloneOnGC ifTrue:
			[coInterpreter cloneSimulation objectMemory fullGC.
			 Smalltalk garbageCollect]].
	^super fullGC! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 7/16/2004 14:56'!
halfWordHighInLong32: long32
	^self subclassResponsibility! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 7/16/2004 14:57'!
halfWordLowInLong32: long32
	^self subclassResponsibility! !

!NewObjectMemorySimulator methodsFor: 'gc -- mark and sweep' stamp: 'eem 5/30/2018 11:26'!
incrementalGC
	"If we're /not/ a clone, clone the VM and push it over the cliff.
	 If it survives, destroy the clone and continue.  We should be OK until next time."
	(self leakCheckNewSpaceGC
	 and: [parent isNil]) ifTrue:
		[coInterpreter cr; print: 'Incremental GC number '; print: statIncrGCs; tab; flush.
		 CloneOnScavenge ifTrue:
			[coInterpreter cloneSimulation objectMemory incrementalGC.
			 Smalltalk garbageCollect]].
	^super incrementalGC! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 7/3/2004 10:47'!
long32At: byteAddress
	"Return the 32-bit word at byteAddress which must be 0 mod 4."

	^ self longAt: byteAddress! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 7/3/2004 10:47'!
long32At: byteAddress put: a32BitValue
	"Store the 32-bit value at byteAddress which must be 0 mod 4."

	^ self longAt: byteAddress put: a32BitValue! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'eem 8/30/2013 15:53'!
longAt: byteAddress
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	byteAddress \\ 4 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 4 + 1! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'eem 8/30/2013 15:54'!
longAt: byteAddress put: a32BitValue
	"Note: Adjusted for Smalltalk's 1-based array indexing."
	byteAddress \\ 4 ~= 0 ifTrue: [self unalignedAccessError].
	^memory at: byteAddress // 4 + 1 put: a32BitValue! !

!NewObjectMemorySimulator methodsFor: 'image save/restore' stamp: 'eem 10/10/2013 15:23'!
memoryBaseForImageRead
	"Answer the address to read the image into."
	^0! !

!NewObjectMemorySimulator methodsFor: 'testing' stamp: 'eem 1/16/2013 11:05'!
objectBefore: addr
	| oop prev |
	oop := self firstObject.
	[oop < freeStart] whileTrue:
		[prev := oop.  "look here if debugging prev obj overlapping this one"
		oop := self objectAfter: oop.
		oop >= addr ifTrue: [^ prev]].
	^0! !

!NewObjectMemorySimulator methodsFor: 'accessing' stamp: 'eem 10/11/2014 14:26'!
parent

	^ parent! !

!NewObjectMemorySimulator methodsFor: 'accessing' stamp: 'eem 10/11/2014 14:26'!
parent: anObject

	parent := anObject! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 6/23/2004 14:29'!
shortAt: byteAddress
    "Return the half-word at byteAddress which must be even."
	^self subclassResponsibility! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'di 6/23/2004 14:32'!
shortAt: byteAddress put: a16BitValue
	^ self subclassResponsibility! !

!NewObjectMemorySimulator methodsFor: 'interpreter access' stamp: 'eem 10/27/2020 21:44'!
sizeFieldOfAlien: alienObj
	"Answer the first field of alienObj which is assumed to be an Alien of at least 8 bytes"
	<inline: #always>
	^(super sizeFieldOfAlien: alienObj) signedIntFromLong! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'eem 9/25/2013 00:21'!
sqGrowMemory: oldLimit By: delta
	| newMemory |
	coInterpreter transcript show: 'grow memory from ', oldLimit printString, ' by ', delta printString; cr.
	memory size * 4 < (oldLimit + delta) ifTrue:
		[newMemory := (memory class new: oldLimit + delta + 3 // 4).
		 newMemory replaceFrom: 1 to: memory size with: memory startingAt: 1.
		 memory := newMemory].
	^memory size * 4! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'ar 2/25/2001 17:16'!
sqMemoryExtraBytesLeft: includingSwap
	^0! !

!NewObjectMemorySimulator methodsFor: 'memory access' stamp: 'eem 10/12/2009 04:27'!
sqShrinkMemory: oldLimit By: delta
	coInterpreter transcript show: 'shrink memory from ', oldLimit printString, ' by ', delta printString, ' remember it doesn''t actually shrink in simulation'; cr.

	^oldLimit! !

!NewObjectMemorySimulator methodsFor: 'initialization' stamp: ''!
startOfMemory
	"Return the start of object memory."

	^ 0! !

!NewObjectMemorySimulator methodsFor: 'interpreter access' stamp: 'eem 9/10/2013 11:19'!
storePointer: index ofObject: oop withValue: valuePointer
	"Override to ensure acess is within the heap, and within the object"
	| fmt hdr |
	self assert: oop >= self startOfMemory.
	hdr := self baseHeader: oop.
	fmt := self formatOfHeader: hdr.
	self assert: ((fmt <= self lastPointerFormat or: [fmt >= self firstCompiledMethodFormat])
				and: [index >= 0 and: [index < (self lengthOf: oop baseHeader: hdr format: fmt)]]).
	^super storePointer: index ofObject: oop withValue: valuePointer! !

!NewObjectMemorySimulator methodsFor: 'interpreter access' stamp: 'eem 12/11/2013 10:18'!
storePointerUnchecked: index ofObject: oop withValue: valuePointer
	"Override to ensure acess is within the heap, and within the object"
	| fmt hdr |
	self assert: oop >= self startOfMemory.
	hdr := self baseHeader: oop.
	fmt := self formatOfHeader: hdr.
	self assert: ((fmt <= self lastPointerFormat or: [fmt >= self firstCompiledMethodFormat])
				and: [index >= 0 and: [index < (self lengthOf: oop baseHeader: hdr format: fmt)]]).
	^super storePointerUnchecked: index ofObject: oop withValue: valuePointer! !

!NewObjectMemorySimulator methodsFor: 'debug support' stamp: 'eem 12/19/2012 14:41'!
tenuringIncrementalGC
	coInterpreter transcript cr; nextPutAll: 'tenuringIncrementalGC ('; print: coInterpreter byteCount; space; print: freeStart; nextPut: $); flush.
	^super tenuringIncrementalGC! !

!NewObjectMemorySimulator methodsFor: 'testing' stamp: 'eem 1/16/2013 11:05'!
validOop: oop
	" Return true if oop appears to be valid "
	(oop bitAnd: 1) = 1 ifTrue: [^ true].  "Integer"
	(oop bitAnd: 3) = 0 ifFalse: [^ false].  "Uneven address"
	oop >= freeStart ifTrue: [^ false].  "Out of range"
	"could test if within the first large freeblock"
	(self longAt: oop) = 4 ifTrue: [^ false].
	(self headerType: oop) = 2 ifTrue: [^ false].	"Free object"
	^ true! !

NewObjectMemorySimulator subclass: #NewObjectMemorySimulatorLSB
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!NewObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'nk 4/3/2004 06:58'!
byteAt: byteAddress
	| lowBits long |
	lowBits := byteAddress bitAnd: 3.
	long := self longAt: byteAddress - lowBits.
	^(lowBits caseOf: {
		[0] -> [ long ].
		[1] -> [ long bitShift: -8  ].
		[2] -> [ long bitShift: -16 ].
		[3] -> [ long bitShift: -24 ]
	}) bitAnd: 16rFF
! !

!NewObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'eem 7/18/2013 17:36'!
byteAt: byteAddress put: byte
	| lowBits long longAddress |
	lowBits := byteAddress bitAnd: 3.
	longAddress := byteAddress - lowBits.
	long := self longAt: longAddress.
	long := (lowBits caseOf: {
		[0] -> [ (long bitAnd: 16rFFFFFF00) bitOr: byte ].
		[1] -> [ (long bitAnd: 16rFFFF00FF) bitOr: (byte bitShift: 8) ].
		[2] -> [ (long bitAnd: 16rFF00FFFF) bitOr: (byte bitShift: 16)  ].
		[3] -> [ (long bitAnd: 16r00FFFFFF) bitOr: (byte bitShift: 24)  ]
	}).
	self assert: longAddress < freeStart.
	self longAt: longAddress put: long.
	^byte! !

!NewObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'nice 4/3/2016 08:36'!
byteSwapped32IfBigEndian: w
	"Accelerate, we now we are not big endian"
	^w bitAnd: 16rFFFFFFFF! !

!NewObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'nice 4/3/2016 08:36'!
byteSwapped64IfBigEndian: w
	"Accelerate, we now we are not big endian"
	^w bitAnd: 16rFFFFFFFFFFFFFFFF! !

!NewObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'eem 6/25/2011 14:03'!
endianness
	^#little! !

!NewObjectMemorySimulatorLSB methodsFor: 'float primitives' stamp: 'eem 4/20/2009 14:37'!
fetchFloatAt: floatBitsAddress into: aFloat
	aFloat at: 2 put: (self long32At: floatBitsAddress).
	aFloat at: 1 put: (self long32At: floatBitsAddress+4)! !

!NewObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:59'!
halfWordHighInLong32: long32
	"Used by Balloon"

	^ long32 bitAnd: 16rFFFF! !

!NewObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:59'!
halfWordLowInLong32: long32
	"Used by Balloon"

	^ long32 bitShift: -16! !

!NewObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'eem 4/14/2015 18:37'!
long64At: byteAddress
	"memory is a Bitmap, a 32-bit indexable array of bits"
	| hiWord loWord |
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	loWord := memory at: byteAddress // 4 + 1.
	hiWord := memory at: byteAddress // 4 + 2.
	^hiWord = 0
		ifTrue: [loWord]
		ifFalse: [(hiWord bitShift: 32) + loWord]! !

!NewObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'eem 4/14/2015 18:37'!
long64At: byteAddress put: a64BitValue
	byteAddress \\ 8 ~= 0 ifTrue: [self unalignedAccessError].
	self
		longAt: byteAddress put: (a64BitValue bitAnd: 16rffffffff);
		longAt: byteAddress + 4 put: a64BitValue >> 32.
	^a64BitValue! !

!NewObjectMemorySimulatorLSB methodsFor: 'initialization' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
nextLongFrom: aStream
	"Read a 32- or 64-bit quantity from the given (binary) stream."

	^ aStream nextLittleEndianNumber: self wordSize! !

!NewObjectMemorySimulatorLSB methodsFor: 'initialization' stamp: 'eem 10/9/2008 16:39'!
nextShortFrom: aStream
	"Read a 16-bit quantity from the given (binary) stream."
	^aStream nextLittleEndianNumber: 2! !

!NewObjectMemorySimulatorLSB methodsFor: 'initialization' stamp: 'eem 11/18/2014 13:47'!
nextWord32From: aStream
	"Read a 32-bit quantity from the given (binary) stream."

	^aStream nextLittleEndianNumber: 4! !

!NewObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'di 6/23/2004 14:29'!
shortAt: byteAddress
    "Return the half-word at byteAddress which must be even."
	| lowBits long |
	lowBits := byteAddress bitAnd: 2.
	long := self longAt: byteAddress - lowBits.
	^ lowBits = 2
		ifTrue: [ long bitShift: -16 ]
		ifFalse: [ long bitAnd: 16rFFFF ].
! !

!NewObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'di 6/23/2004 14:31'!
shortAt: byteAddress put: a16BitValue
    "Return the half-word at byteAddress which must be even."
	| lowBits long longAddress |
	lowBits := byteAddress bitAnd: 2.
	lowBits = 0
		ifTrue:
		[ "storing into LS word"
		long := self longAt: byteAddress.
		self longAt: byteAddress
				put: ((long bitAnd: 16rFFFF0000) bitOr: a16BitValue)
		]
		ifFalse:
		[longAddress := byteAddress - 2.
		long := self longAt: longAddress.
		self longAt: longAddress
				put: ((long bitAnd: 16rFFFF) bitOr: (a16BitValue bitShift: 16))
		]! !

!NewObjectMemorySimulatorLSB methodsFor: 'float primitives' stamp: 'eem 4/20/2009 14:37'!
storeFloatAt: floatBitsAddress from: aFloat

	self long32At: floatBitsAddress put: (aFloat at: 2).
	self long32At: floatBitsAddress+4 put: (aFloat at: 1)! !

!NewObjectMemorySimulatorLSB methodsFor: 'memory access' stamp: 'tpr 12/22/2005 17:46'!
vmEndianness
	"return 0 for little endian, 1 for big endian"
	^0! !

NewObjectMemorySimulator subclass: #NewObjectMemorySimulatorMSB
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-InterpreterSimulation'!

!NewObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
byteAt: byteAddress
	| lowBits bpwMinus1 |
	bpwMinus1 := self wordSize-1.
	lowBits := byteAddress bitAnd: bpwMinus1.
	^ ((self longAt: byteAddress - lowBits)
		bitShift: (lowBits - bpwMinus1) * 8)
		bitAnd: 16rFF! !

!NewObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
byteAt: byteAddress put: byte
	| longWord shift lowBits bpwMinus1 longAddress |
	bpwMinus1 := self wordSize-1.
	lowBits := byteAddress bitAnd: bpwMinus1.
	longAddress := byteAddress - lowBits.
	longWord := self longAt: longAddress.
	shift := (bpwMinus1 - lowBits) * 8.
	longWord := longWord
				- (longWord bitAnd: (16rFF bitShift: shift))
				+ (byte bitShift: shift).
	self assert: longAddress < freeStart.
	self longAt: longAddress put: longWord.
	^byte! !

!NewObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'nice 4/3/2016 08:38'!
byteSwapped32IfBigEndian: w
	"Accelerate, we now we are big endian"
	^w byteSwap32! !

!NewObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'nice 4/3/2016 08:37'!
byteSwapped64IfBigEndian: w
	"Accelerate, we now we are big endian"
	^w byteSwap64! !

!NewObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'eem 6/25/2011 14:03'!
endianness
	^#big! !

!NewObjectMemorySimulatorMSB methodsFor: 'float primitives' stamp: 'eem 10/12/2009 03:55'!
fetchFloatAt: floatBitsAddress into: aFloat
	aFloat at: 1 put: (self long32At: floatBitsAddress).
	aFloat at: 2 put: (self long32At: floatBitsAddress+4)! !

!NewObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:58'!
halfWordHighInLong32: long32
	"Used by Balloon"

	^ long32 bitShift: -16! !

!NewObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'di 7/16/2004 14:58'!
halfWordLowInLong32: long32
	"Used by Balloon"

	^ long32 bitAnd: 16rFFFF! !

!NewObjectMemorySimulatorMSB methodsFor: 'image save/restore' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
nextLongFrom: aStream
	"Read a 32- or 64-bit quantity from the given (binary) stream."

	^ aStream nextNumber: self wordSize! !

!NewObjectMemorySimulatorMSB methodsFor: 'image save/restore' stamp: 'eem 10/12/2009 03:59'!
nextShortFrom: aStream
	"Read a 16-bit quantity from the given (binary) stream."
	^aStream nextNumber: 2! !

!NewObjectMemorySimulatorMSB methodsFor: 'image save/restore' stamp: 'eem 11/18/2014 13:47'!
nextWord32From: aStream
	"Read a 32-bit quantity from the given (binary) stream."

	^aStream nextNumber: 4! !

!NewObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
shortAt: byteAddress
    "Return the half-word at byteAddress which must be even."
	| lowBits bpwMinus2 |
	bpwMinus2 := self wordSize-2.
	lowBits := byteAddress bitAnd: bpwMinus2.
	^ ((self longAt: byteAddress - lowBits)
		bitShift: (lowBits - bpwMinus2) * 8)
		bitAnd: 16rFFFF
! !

!NewObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'eem (const to msg rewrite) 11/6/2014 11:00'!
shortAt: byteAddress put: a16BitValue
    "Return the half-word at byteAddress which must be even."
	| longWord shift lowBits bpwMinus2 longAddress |
	bpwMinus2 := self wordSize-2.
	lowBits := byteAddress bitAnd: bpwMinus2.
	longAddress := byteAddress - lowBits.
	longWord := self longAt: longAddress.
	shift := (bpwMinus2 - lowBits) * 8.
	longWord := longWord
				- (longWord bitAnd: (16rFFFF bitShift: shift))
				+ (a16BitValue bitShift: shift).
	self longAt: longAddress put: longWord
! !

!NewObjectMemorySimulatorMSB methodsFor: 'float primitives' stamp: 'eem 10/12/2009 04:05'!
storeFloatAt: floatBitsAddress from: aFloat

	self long32At: floatBitsAddress put: (aFloat at: 1).
	self long32At: floatBitsAddress+4 put: (aFloat at: 2)! !

!NewObjectMemorySimulatorMSB methodsFor: 'memory access' stamp: 'tpr 12/22/2005 17:46'!
vmEndianness
	"return 0 for little endian, 1 for big endian"
	^1! !

!ObjectMemory class methodsFor: 'accessing' stamp: 'tpr 4/1/2005 14:25'!
baseHeaderSize
"To support SmartSyntaxPluginCodeGenerator"
	^BaseHeaderSize! !

!ObjectMemory class methodsFor: 'simulation only' stamp: 'eem 10/30/2013 13:55'!
characterObjectOf: characterCode
	^(characterCode between: 0 and: 255) ifTrue:
		[characterCode asCharacter]! !

!ObjectMemory class methodsFor: 'translation' stamp: 'eem 12/7/2013 10:44'!
declareCVarsIn: aCCodeGenerator
	self declareCAsOop:
			#(	memory endOfMemory memoryLimit
				youngStart youngStartLocal freeBlock
				compStart compEnd
				fwdTableNext fwdTableLast
				gcBiasToGrowThreshold )
		in: aCCodeGenerator.
	self declareCAsUSqLong:
			#( gcStartUsecs statFullGCUsecs statIncrGCUsecs statIGCDeltaUsecs )
		in: aCCodeGenerator.
	aCCodeGenerator
		var: #remapBuffer
		declareC: 'sqInt remapBuffer[RemapBufferSize + 1 /* ', (RemapBufferSize + 1) printString, ' */]'.
	aCCodeGenerator
		var: #rootTable
		declareC: 'sqInt rootTable[RootTableSize + 1 /* ', (RootTableSize + 1) printString, ' */]'.
	"Weak roots must be large enough for roots+remapBuffer+sizeof(allCallsOn: #markAndTrace:)"
	aCCodeGenerator
		var: #weakRoots
		declareC: 'sqInt weakRoots[WeakRootTableSize + 1 /* ', (WeakRootTableSize + 1) printString, ' */]'.
	aCCodeGenerator
		var: #headerTypeBytes
		declareC: 'sqInt headerTypeBytes[4]'.
	aCCodeGenerator
		var: #extraRoots
		declareC: 'sqInt* extraRoots[ExtraRootSize + 1 /* ', (ExtraRootSize + 1) printString, ' */]'.
	aCCodeGenerator
		var: #headerTypeBytes type: 'const sqInt' array: HeaderTypeExtraBytes! !

!ObjectMemory class methodsFor: 'api chacterization' stamp: 'eem 1/6/2014 13:32'!
hasSpurMemoryManagerAPI
	^false! !

!ObjectMemory class methodsFor: 'translation' stamp: 'eem 3/3/2009 10:22'!
implicitReturnTypeFor: aSelector
	"Answer the return type for methods that don't have an explicit return."
	^#void! !

!ObjectMemory class methodsFor: 'initialization' stamp: 'nice 4/3/2016 00:33'!
initBytesPerWord: nBytes

	BytesPerWord := nBytes.
	"N.B.  This is *not* output when generating the interpreter file.
	 It is left to the various sqConfig.h files to define correctly."
	VMBIGENDIAN := Smalltalk endianness == #big! !

!ObjectMemory class methodsFor: 'class initialization' stamp: 'eem 8/15/2013 17:07'!
initialize
	"This is for Monticello package loading"
	self initializeWithOptions: Dictionary new! !

!ObjectMemory class methodsFor: 'initialization' stamp: 'eem 1/22/2021 21:07'!
initializeCompactClassIndices
	"Initialize indices for compact classes we are going to depend on being compact.
	 The VI allows classes to become compact and become uncompact.  For efficiency
	 the VM assumes certain classes are compact with particular indices."

	"Smalltalk compactClassesArray"
	"{Array. LargePositiveInteger. Float. MethodContext. } collect: [:c| c -> c indexIfCompact]"

	ClassArrayCompactIndex := 3.
	ClassLargeNegativeIntegerCompactIndex := 4. "Was PseudoContext class"
	ClassLargePositiveIntegerCompactIndex := 5.
	ClassFloatCompactIndex := 6.
	ClassBlockClosureCompactIndex := 0 "12". "Prospective.  May still be TranslatedMethod class"
	ClassByteStringCompactIndex := 11.
	ClassBlockContextCompactIndex := 13.
	ClassMethodContextCompactIndex := 14.

	ClassByteArrayCompactIndex := 0.
	ClassMessageCompactIndex := 0.
	ClassBitmapCompactIndex := 0.
	ClassPointCompactIndex := 0! !

!ObjectMemory class methodsFor: 'initialization' stamp: 'eem 9/6/2011 15:15'!
initializeObjectHeaderConstants

	BytesPerWord ifNil: [BytesPerWord := 4].  "May get called on fileIn, so supply default"
	BaseHeaderSize := BytesPerWord.
	WordMask := (1 bitShift: BytesPerWord*8) - 1.
	
	"masks for type field"
	TypeMask := 3.
	AllButTypeMask := WordMask - TypeMask.

	"type field values"
	HeaderTypeSizeAndClass := 0.
	HeaderTypeClass := 1.
	HeaderTypeFree := 2.
	HeaderTypeShort := 3.
	HeaderTypeExtraBytes := { BytesPerWord * 2. BytesPerWord. 0. 0 }.

	"type field values used during the mark phase of GC"
	HeaderTypeGC := 2.
	GCTopMarker := 3.  "neither an oop, nor an oop+1, this value signals that we have crawled back up to the top of the marking phase."

	"Base header word bit fields"
	HashBits := 16r1FFE0000.
	HashBitsOffset := 17.
	HashMaskUnshifted := 16rFFF.
	self assert: (HashMaskUnshifted bitShift: HashBitsOffset) = HashBits.
	AllButHashBits := WordMask - HashBits.
	SizeMask := 16rFC.
	Size4Bit := 0.
BytesPerWord = 8 ifTrue:
		[SizeMask := 16rF8.  "Lose the 4 bit in temp 64-bit chunk format"
		Size4Bit := 4].  "But need it for ST size"
	"Note SizeMask + Size4Bit gives the mask needed for size fits of format word in classes.
		This is used in instantiateClass:indexableSize: "
	LongSizeMask := WordMask - 16rFF + SizeMask.
	LongSizeNumBits := 30. "30 bits of size info in long size filed."
	CompactClassMask := 16r1F000.

	"masks for root and mark bits"
	MarkBit := 1 bitShift: BytesPerWord*8 - 1.  "Top bit"
	RootBit := 1 bitShift: BytesPerWord*8 - 2.  "Next-to-Top bit"
	AllButMarkBit := WordMask - MarkBit.
	AllButRootBit := WordMask - RootBit.

	AllButMarkBitAndTypeMask := AllButTypeMask - MarkBit.

	ImmutabilityBit := 1 bitShift: BytesPerWord*8 - 3.  "Next-to-Next-To-Top bit"
	AllButImmutabilityBit := WordMask - ImmutabilityBit! !

!ObjectMemory class methodsFor: 'initialization' stamp: 'rsf 8/9/2016 21:33'!
initializeSpecialObjectIndices
	"Initialize indices into specialObjects array."

	NilObject := 0.
	FalseObject := 1.
	TrueObject := 2.
	SchedulerAssociation := 3.
	ClassBitmap := 4.
	ClassSmallInteger := 5.
	ClassByteString := ClassString := 6. "N.B.  Actually class ByteString"
	ClassArray := 7.
	"SmalltalkDictionary := 8."  "Do not delete!!"
	ClassFloat := 9.
	ClassMethodContext := 10.
	ClassBlockContext := 11.
	ClassPoint := 12.
	ClassLargePositiveInteger := 13.
	TheDisplay := 14.
	ClassMessage := 15.
	"ClassCompiledMethod := 16. unused by the VM"
	TheLowSpaceSemaphore := 17.
	ClassSemaphore := 18.
	ClassCharacter := 19.
	SelectorDoesNotUnderstand := 20.
	SelectorCannotReturn := 21.
	ProcessSignalingLowSpace := 22.	"was TheInputSemaphore"
	SpecialSelectors := 23.
	CharacterTable := 24.
	SelectorMustBeBoolean := 25.
	ClassByteArray := 26.
	"ClassProcess := 27. unused"
	CompactClasses := 28.
	TheTimerSemaphore := 29.
	TheInterruptSemaphore := 30.
	SelectorCannotInterpret := 34.
	"Was MethodContextProto := 35."
	ClassBlockClosure := 36.
	ClassFullBlockClosure := 37. "Was BlockContextProto := 37."
	ExternalObjectsArray := 38.
	ClassMutex := 39.
	"Was: ClassTranslatedMethod := 40."
	ProcessInExternalCodeTag := 40.
	TheFinalizationSemaphore := 41.
	ClassLargeNegativeInteger := 42.

	ClassExternalAddress := 43.
	ClassExternalStructure := 44.
	ClassExternalData := 45.
	ClassExternalFunction := 46.
	ClassExternalLibrary := 47.

	SelectorAboutToReturn := 48.
	SelectorRunWithIn := 49.

	SelectorAttemptToAssign := 50.
	"PrimErrTableIndex := 51. in VMClass class>>initializePrimitiveErrorCodes"
	ClassAlien := 52.
	SelectorInvokeCallback := 53.
	ClassUnsafeAlien := 54.

	ClassWeakFinalizer := 55.

	ForeignCallbackProcess := 56.

	SelectorUnknownBytecode := 57.
	SelectorCounterTripped := 58.
	SelectorSistaTrap := 59.
	
	LowcodeContextMark := 60.
	LowcodeNativeContextClass := 61.
! !

!ObjectMemory class methodsFor: 'initialization' stamp: 'eem 11/18/2014 18:23'!
initializeWithOptions: optionsDictionary
	"ObjectMemory initializeWithOptions: Dictionary new"

	self initBytesPerWord: (optionsDictionary at: #BytesPerWord ifAbsent: [4]).
	BytesPerOop := optionsDictionary at: #BytesPerOop ifAbsent: [BytesPerWord].

	"Translation flags (booleans that control code generation via conditional translation):"
	"generate assertion checks"
	DoAssertionChecks := optionsDictionary at: #DoAssertionChecks ifAbsent: [false].
	DoExpensiveAssertionChecks := optionsDictionary at: #DoExpensiveAssertionChecks ifAbsent: [false].

	self initializeObjectHeaderConstants. "Initializes BaseHeaderSize so do early"
	self initializeSpecialObjectIndices.
	self initializeCompactClassIndices.
	self initializePrimitiveErrorCodes.

	NilContext := 1.  "the oop for the integer 0; used to mark the end of context lists"

	RemapBufferSize := 25.
	RootTableSize := 2500.  	"number of root table entries (4 bytes/entry)"
	RootTableRedZone := RootTableSize - 100.	"red zone of root table - when reached we force IGC"
	WeakRootTableSize := RootTableSize + RemapBufferSize + 100.

	"tracer actions"
	StartField := 1.
	StartObj := 2.
	Upward := 3.
	Done := 4.

	ExtraRootSize := 2048. "max. # of external roots"! !

!ObjectMemory class methodsFor: 'simulation only' stamp: 'eem 10/30/2013 13:54'!
isImmediate: anObject
	^anObject class == SmallInteger! !

!ObjectMemory class methodsFor: 'translation' stamp: 'eem 2/27/2013 15:32'!
isNonArgumentImplicitReceiverVariableName: aString
	^#('self' 'interpreter' 'coInterpreter' 'heapMap') includes: aString! !

!ObjectMemory class methodsFor: 'translation' stamp: 'eem 11/20/2014 13:00'!
maxSmallInteger
	^1073741823! !

!ObjectMemory class methodsFor: 'accessing' stamp: 'eem 11/15/2019 16:14'!
memoryManagerName
	^ 'SqueakV3'! !

!ObjectMemory class methodsFor: 'translation' stamp: 'eem 11/20/2014 13:00'!
minSmallInteger
	^-1073741824! !

!ObjectMemory class methodsFor: 'translation' stamp: 'eem 12/7/2013 10:40'!
mustBeGlobal: var
	"Answer if a variable must be global and exported.  Used for inst vars that are accessed from VM support code."

	^false "used to be: ^'memory' = var"! !

!ObjectMemory class methodsFor: 'accessing' stamp: 'eem 11/18/2014 17:48'!
numSmallIntegerTagBits
	^1! !

!ObjectMemory class methodsFor: 'accessing class hierarchy' stamp: 'eem 10/9/2013 19:47'!
objectRepresentationClass
	^self subclassResponsibility! !

!ObjectMemory class methodsFor: 'api chacterization' stamp: 'eem 1/18/2020 11:28'!
vmProxyMinorVersion
	"Define the  VM_PROXY_MAJOR version for this VM as used to define
	 the api in platforms/Cross/vm/sqVirtualMachine.[ch] and in interp.h."
	^8! !

!ObjectMemory class methodsFor: 'accessing' stamp: 'eem 9/10/2013 14:43'!
wordSize
	"Answer the width of an object pointer, in bytes."

	^BytesPerWord! !

!ObjectMemory methodsFor: 'gc -- mark and sweep' stamp: 'eem 10/22/2008 10:29'!
aComment
	"The mark phase is based on a pointer reversing traversal. This is a little tricky because the class, which is needed by the traversal, may be in either the header (as a compact class index) or in the word above the header. See memo 'Revised object format'.
	Compact classes are marked and traced separately.
	How do you know that you are returning from having marked a class? Parent pointer has 10 in low bits.

Here are the states an object may be in, followed by what to do next in brackets []:

  Start Object: parentField is set, [obj := child]:
	obj is pointed at by a field in parent that is being traced now. obj is marked.
		[(parent goes up to the next field) field addr := obj. go to Upward]
	obj is pointed at by a field in parent that is being traced now. obj is unmarked. obj has no pointers.
		[put 10 into low bits of header. field addr := obj. go to Start Field (to process class word)]
	obj is pointed at by a field in parent that is being traced now. obj is unmarked. obj has pointers.
		[put 10 into low bits of header. point to last field. go to Start Field]

  Start Field: 
	Field ends in 10. It is the header. Short Class is not 0.
		[Set low bits to correct value. (have parent pointer) go to Upward]
	Field ends in 10. It is the header. Short Class is 0.
		[child := word above header. low bits of child := 01. class word := parentField. parentField := loc of class word. go to Start Obj]
	Field is Integer.
		[point one word up, go to Start Field]
	Field is oop.
		[child := field. field := parentField. parentField := loc of field. go to Start Obj]

  Upward [restore low bits of header (at field addr)]:
	parentField is 3. (bits 11, int 1).
		[done!!]
	parentField ends in 00.
		[child := field addr. field addr := parentField. parentField := field addr contents.
		field addr contents := child (addr of prev object. its oop). field addr - 4. go to Start Field]
	parentField ends in 01. We're tracing the class.
		[child := field addr. field addr := parentField (loc of class word). parentField := field addr contents.
		field addr contents := child (addr of prev object. its oop). field addr + 4 (header). go to Upward]
"! !

!ObjectMemory methodsFor: 'finalization' stamp: 'eem 1/11/2008 10:42'!
aFinalizationComment
	"This finalization scheme assumes to have weak classes in which the fields are not traced during the mark phase of a GC. This means, if an object is referenced only by any instances of weak classes it can be collected. In turn, we need to find out if an object referenced by a weak class is actually being collected because we have to invalidate the weak object pointer and to signal that the object has gone.
	How do we know that an object referenced by a weak class is being collected? Well,  this is based on two observations. First, objects will not change their relative locations in memory, meaning that if object A is created BEFORE object B it will always have a physical memory address which is LESS than B. Secondly, GC always works from a given starting address (youngStart during incremental GC; startOfMemory during fullGC) up to end of memory. If we can somehow guarantee that the weak reference is created after the object it points to we can easily implement the following simple scheme:
	1) Mark phase
		Do not trace the fields of any instances of weak classes.
	2) Sweep phase:
		a) Explicitly mark all free objects.
		b) 	If a weak reference is encountered check the the object it points to. 
			If the object is marked as free than we know that this weak reference's object is gone.
			Signal that it is gone.

	There is, however, one small problem with this approach. We cannot always guarantee that WeakReferences point backwards such as in the following piece of code:
		| o1 o2 w1 w2 |
		o1 := Object new.
		w1 := WeakReference on: o1.
		o2 := Object new.
		w2 := WeakReference on: o2.
		o1 become: o2.
The become: operation makes w1 point to o2 and because o2 has been created AFTER w1 the object reference in w1 points forward. Why might this be a problem? Well, if the GC would start after the weak reference AND free the object then the weak reference would simply point to an invalid memory location (since we've not been checking the weak reference during sweep phase).

	Fortunately, this can not happen in the current ObjectMemory implementation. Why? Well, the only GC not starting at the beginning of the memory is incremental GC. Incremental GC however is only executed in so-called youngSpace. If both, the weak reference AND the object it points to reside in youngSpace then we can still check the weak reference. If however, the weak reference is not in youngSpace but the object is, then the reference is itself a root for young space and will be processed by the GC.

	In the end, we just need a little adjustment in step 2b) of the above procedure which looks as follows:
		If the weak reference points 
			* backwards: check if the object header is marked free
			* forwards: check if the object has been marked in markPhase.

	Note that a number of finalizations will only be executed during a fullGC. So, if you must guarantee that some object has been finalized you definitely need to do a fullGC.

ar 3/20/98 17:20"

	self error:'Comment only'.! !

!ObjectMemory methodsFor: 'object enumeration' stamp: 'dtl (auto pragmas 12/08) 4/22/2007 20:46'!
accessibleObjectAfter: oop 
	"Return the accessible object following the given object or 
	free chunk in the heap. Return nil when heap is exhausted."
	| obj |
	<inline: false>
	obj := self objectAfter: oop.
	[self oop: obj isLessThan: endOfMemory]
		whileTrue: [(self isFreeObject: obj) ifFalse: [^ obj].
			obj := self objectAfter: obj].
	^ nil! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'eem 5/19/2009 09:58'!
addGCRoot: varLoc
	"Add the given variable location to the extra roots table"
	<export: true>
	<var: #varLoc type: #'sqInt *'>
	extraRootCount >= ExtraRootSize ifTrue:[^false]. "out of space"
	extraRoots at: (extraRootCount := extraRootCount+1) put: varLoc.
	^true! !

!ObjectMemory methodsFor: 'debug support' stamp: 'eem 5/10/2011 11:10'!
addressCouldBeObj: address
	<api>
	"Answer if address appears to be that of an object, which implies it is
	 safe to fetch the class and size. For code disassembly and assertions."
	^(address bitAnd: 3) = 0
	  and: [address asUnsignedInteger >= self startOfMemory
	  and: [address asUnsignedInteger < freeBlock
	  and: [(self headerType: address) ~= HeaderTypeGC]]]! !

!ObjectMemory methodsFor: 'debug support' stamp: 'eem 5/10/2011 11:10'!
addressCouldBeObjWhileForwarding: address
	"Answer if address appears to be that of an object, which implies it is
	 safe to fetch the class and size. For code disassembly and assertions."
	^(address bitAnd: 3) = 0
	  and: [address asUnsignedInteger >= self startOfMemory
	  and: [address asUnsignedInteger < freeBlock]]! !

!ObjectMemory methodsFor: 'debug support' stamp: 'eem 5/29/2009 17:14'!
addressCouldBeOop: address
	<api>
	"Answer if address appears to be that of either a SmallInteger or an object.
	 For code disassembly and assertions."
	^(self isIntegerObject: address)
	   or: [self addressCouldBeObj: address]! !

!ObjectMemory methodsFor: 'initialization' stamp: 'dtl (auto pragmas 12/08) 4/22/2007 20:48'!
adjustAllOopsBy: bytesToShift 
	"Adjust all oop references by the given number of bytes. This 
	is done just after reading in an image when the new base 
	address of the object heap is different from the base address 
	in the image."
	"di 11/18/2000 - return number of objects found"

	| oop totalObjects |
	<inline: false>
	bytesToShift = 0 ifTrue: [^300000].
	"this is probably an improvement over the previous answer of 
	nil, but maybe we should do the obejct counting loop and 
	simply guard the adjustFieldsAndClass... with a bytesToShift 
	= 0 ifFalse: ?"
	totalObjects := 0.
	oop := self firstObject.
	[self oop: oop isLessThan: endOfMemory]
		whileTrue:
			[(self isFreeObject: oop)
				ifFalse:
					[totalObjects := totalObjects + 1.
					 self adjustFieldsAndClassOf: oop by: bytesToShift].
			 oop := self objectAfter: oop].
	^totalObjects! !

!ObjectMemory methodsFor: 'initialization' stamp: 'eem 11/10/2014 12:49'!
adjustFieldsAndClassOf: oop by: offsetBytes 
	"Adjust all pointers in this object by the given offset."
	| fieldAddr fieldOop classHeader newClassOop |
	<inline: true>
	fieldAddr := oop + (self lastPointerOf: oop).
	[self oop: fieldAddr isGreaterThan: oop] whileTrue:
		[fieldOop := self longAt: fieldAddr.
		 (self isIntegerObject: fieldOop) ifFalse:
			[self longAt: fieldAddr put: fieldOop + offsetBytes].
		 fieldAddr := fieldAddr - self bytesPerOop].
	(self headerType: oop) ~= HeaderTypeShort ifTrue:
		["adjust class header if not a compact class"
		 classHeader := self longAt: oop - self wordSize.
		 newClassOop := (classHeader bitAnd: AllButTypeMask) + offsetBytes.
		 self longAt: oop - self wordSize put: (newClassOop bitOr: (classHeader bitAnd: TypeMask))]! !

!ObjectMemory methodsFor: 'primitive support' stamp: 'eem 3/11/2014 11:37'!
allInstancesOf: aClass
	"Attempt to answer an array of all objects, excluding those that may
	 be garbage collected as a side effect of allocating the result array.
	 If no memory is available answer the number of instances as a SmallInteger.
	 Since objects are at least 16 bytes big, and the largest SmallInteger covers
	 1/4 of the address space, the count can never overflow."
	| count theClass resultArray newCount |
	"Count the currently accessible objects"
	count := 0.
	self allObjectsDo:
		[:obj|
		(self fetchClassOfNonImm: obj) = aClass ifTrue:
			[count := count + 1]].
	"Allocate result array, may cause GC"
	self pushRemappableOop: aClass.
	resultArray := self instantiateClass: self classArray indexableSize: count.
	theClass := self popRemappableOop.
	newCount := 0.
	self allObjectsDo:
		[:obj|
		 (self fetchClassOfNonImm: obj) = theClass ifTrue:
			[newCount := newCount + 1.
			 resultArray == nil ifFalse:
				[self storePointerUnchecked: newCount ofObject: resultArray withValue: obj]]].
	resultArray == nil ifTrue:
		[^self integerObjectOf: newCount].
	"If GC occurred during result array allocation, truncate unused portion of result array"
	newCount < count ifTrue:
		[self shorten: resultArray toIndexableSize: newCount].
	^resultArray! !

!ObjectMemory methodsFor: 'primitive support' stamp: 'eem 3/11/2014 11:25'!
allObjects
	"Attempt to answer an array of all objects, excluding those that may
	 be garbage collected as a side effect of allocating the result array.
	 If no memory is available answer the number of objects as a SmallInteger.
	 Since objects are at least 4 bytes big, and the largest SmallInteger covers
	 1/4 of the address space, the count can never overflow."
	| count obj resultArray newCount |
	"Count the currently accessible objects"
	count := 0.
	self allObjectsDo:
		[:ign| count := count + 1].
	"Allocate result array, may cause GC"
	resultArray := self instantiateClass: self classArray indexableSize: count.
	"Store all objects in result array, excluding any reference to the result array itself,
	 as may happen if garbage collection occurred during allocation of the array. No store
	 check is necessary; the result array will be the last object in memory and hence new."
	newCount := 0.
	obj := self firstObject.
	[obj < resultArray] whileTrue:
		[(self isFreeObject: obj) ifFalse:
			[newCount := newCount + 1.
			 resultArray == nil ifFalse:
				[self storePointerUnchecked: newCount ofObject: resultArray withValue: obj]].
		 obj := self objectAfter: obj].
	resultArray == nil ifTrue:
		[^self integerObjectOf: count].
	"If GC occurred during result array allocation, truncate unused portion of result array"
	newCount < count ifTrue:
		[self shorten: resultArray toIndexableSize: newCount].
	^resultArray! !

!ObjectMemory methodsFor: 'object enumeration' stamp: 'eem 10/13/2014 16:04'!
allObjectsDo: aBlock
	<inline: true>
	| oop |
	oop := self firstObject.
	[oop asUnsignedInteger < freeBlock] whileTrue:
		[(self isFreeObject: oop) ifFalse:
			[aBlock value: oop].
		 oop := self objectAfter: oop]! !

!ObjectMemory methodsFor: 'object enumeration' stamp: 'eem 10/13/2014 16:13'!
allObjectsDoSafely: aBlock
	<inline: true>
	| oop |
	oop := self firstObject.
	[oop asUnsignedInteger < freeBlock] whileTrue:
		[(self isFreeObject: oop) ifFalse:
			[aBlock value: oop].
		 oop := self objectAfterWhileForwarding: oop]! !

!ObjectMemory methodsFor: 'become' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
allYoung: array1 and: array2 
	"Return true if all the oops in both arrays, and the arrays 
	themselves, are in the young object space."
	| fieldOffset |
	(self oop: array1 isLessThan: youngStart)
		ifTrue: [^ false].
	(self oop: array2 isLessThan: youngStart)
		ifTrue: [^ false].
	fieldOffset := self lastPointerOf: array1.
	"same size as array2"
	[fieldOffset >= self baseHeaderSize] whileTrue:
		[(self oop: (self longAt: array1 + fieldOffset) isLessThan: youngStart)
			ifTrue: [^ false].
		((self oop: (self longAt: array2 + fieldOffset) isLessThan: youngStart)
		 and: [self isNonIntegerObject: (self longAt: array2 + fieldOffset)])
			ifTrue: [^ false].
		fieldOffset := fieldOffset - self wordSize].
	^ true! !

!ObjectMemory methodsFor: 'allocation' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
allocate: byteSize headerSize: hdrSize h1: baseHeader h2: classOop h3: extendedSize doFill: doFill format: format
	"Allocate a new object of the given size and number of header words. (Note: byteSize already includes space for the base header word.) Initialize the header fields of the new object and fill the remainder of the object with a value appropriate for the format.
	May cause a GC"

	| newObj remappedClassOop |
	<inline: true>
	<var: #i type: 'usqInt'>
	<var: #end type: 'usqInt'>
	"remap classOop in case GC happens during allocation"
	hdrSize > 1 ifTrue: [self pushRemappableOop: classOop].
	newObj := self allocateChunk: byteSize + (hdrSize - 1 * self wordSize).
	hdrSize > 1 ifTrue: [remappedClassOop := self popRemappableOop].

	hdrSize = 3
		ifTrue: [self longAt: newObj put: (extendedSize bitOr: HeaderTypeSizeAndClass).
			self longAt: newObj + self wordSize put: (remappedClassOop bitOr: HeaderTypeSizeAndClass).
			self longAt: newObj + (self wordSize*2) put: (baseHeader bitOr: HeaderTypeSizeAndClass).
			newObj := newObj + (self wordSize*2)].

	hdrSize = 2
		ifTrue: [self longAt: newObj put: (remappedClassOop bitOr: HeaderTypeClass).
			self longAt: newObj + self wordSize put: (baseHeader bitOr: HeaderTypeClass).
			newObj := newObj + self wordSize].

	hdrSize = 1
		ifTrue: [self longAt: newObj put: (baseHeader bitOr: HeaderTypeShort)].
	"clear new object"
	doFill ifTrue:
		[| fillWord end i |
		 fillWord := format <= self lastPointerFormat
					ifTrue: [nilObj] "if pointers, fill with nil oop"
					ifFalse: [0].
		 end := newObj + byteSize.
		 i := newObj + self wordSize.
		 [i < end] whileTrue:
			[self longAt: i put: fillWord.
			 i := i + self wordSize]].
	DoAssertionChecks
		ifTrue: [self okayOop: newObj.
			self oopHasOkayClass: newObj.
			(self objectAfter: newObj) = freeBlock
				ifFalse: [self error: 'allocate bug: did not set header of new oop correctly'].
			(self objectAfter: freeBlock) = endOfMemory
				ifFalse: [self error: 'allocate bug: did not set header of freeBlock correctly']].

	^newObj! !

!ObjectMemory methodsFor: 'allocation' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
allocateChunk: byteSize 
	"Allocate a chunk of the given size. Sender must be sure that  the requested size includes enough space for the header  word(s). " 
	"Details: To limit the time per incremental GC, do one every so many allocations. The number is settable via primitiveVMParameter to tune your memory system"
	| enoughSpace newFreeSize newChunk |
	<inline: true>

	allocationCount >= allocationsBetweenGCs
		ifTrue: ["do an incremental GC every so many allocations to  keep pauses short"
			self incrementalGC].

	enoughSpace := self sufficientSpaceToAllocate: byteSize.
	enoughSpace
		ifFalse: ["signal that space is running low, but proceed with allocation if possible"
			signalLowSpace := true.
			lowSpaceThreshold := 0. "disable additional interrupts until lowSpaceThreshold is reset by image"
			self saveProcessSignalingLowSpace.
			self forceInterruptCheck].
	(self oop: (self sizeOfFree: freeBlock) isLessThan: byteSize + self baseHeaderSize)
		ifTrue: [self error: 'out of memory'].

	"if we get here, there is enough space for allocation to  succeed "
	newFreeSize := (self sizeOfFree: freeBlock) - byteSize.
	newChunk := freeBlock.
	freeBlock := freeBlock + byteSize.

	"Assume: client will initialize object header of free chunk, so following is not needed:"
	"self setSizeOfFree: newChunk to: byteSize."
	self setSizeOfFree: freeBlock to: newFreeSize.
	allocationCount := allocationCount + 1.
	^newChunk! !

!ObjectMemory methodsFor: 'simulation' stamp: 'eem 6/25/2011 13:56'!
allocateMemoryOfSize: limit
	<doNotGenerate>
	self setMemoryLimit: limit.
	memory := (self endianness == #little
					ifTrue: [LittleEndianBitmap]
					ifFalse: [Bitmap]) new: memoryLimit // 4! !

!ObjectMemory methodsFor: 'allocation' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
allocationUnit
	^self wordSize! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 11/19/2014 14:01'!
areIntegers: oop1 and: oop2
	"Test oop1 and oop2 to make sure both are SmallIntegers."
	^((oop1 bitAnd: oop2) bitAnd: 1) ~= 0! !

!ObjectMemory methodsFor: 'header formats' stamp: 'eem 10/9/2013 19:39'!
arrayFormat
	<api>
	^2! !

!ObjectMemory methodsFor: 'allocation' stamp: 'eem 2/7/2015 09:17'!
attemptToShrink
	"Attempt to shrink memory after successfully reclaiming lots of memory."
	| freeSizeNow |
	freeSizeNow := self freeSize.
	(freeSizeNow > shrinkThreshold
	 and: [freeSizeNow > growHeadroom]) ifTrue:
		[self shrinkObjectMemory: freeSizeNow - growHeadroom]! !

!ObjectMemory methodsFor: 'contexts' stamp: 'eem 11/13/2014 13:28'!
badContextSize: oop
	| numSlots |
	numSlots := self numSlotsOf: oop.
	^numSlots ~= SmallContextSlots and: [numSlots ~= LargeContextSlots]! !

!ObjectMemory methodsFor: 'header access' stamp: ''!
baseHeader: oop

	^ self longAt: oop! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 6/11/2007 14:41'!
baseHeader: oop put: aWord

	^ self longAt: oop put: aWord! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 11/10/2014 10:32'!
baseHeaderSize
	"N.B. This would appear to hard-code the header size for 32-bit images.  But if generating
	 a 64-bit image, this method could be removed and the relevant one substituted.  We can't
	 mark this method as <doNotGenerate> as we need an actual method to guide code gen."
	^4! !

!ObjectMemory methodsFor: 'garbage collection' stamp: 'eem 4/2/2009 22:41'!
beRootIfOld: oop 
	"If this object is old, mark it as a root (because a new object 
	may be stored into it)"
	<api>
	<inline: false>
	((self oop: oop isLessThan: youngStart)
			and: [(self isIntegerObject: oop) not])
		ifTrue: ["Yes, oop is an old object"
			self noteAsRoot: oop headerLoc: oop]! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 11/10/2014 12:49'!
beRootWhileForwarding: oop
	"Record that the given oop in the old object area points to an object in the young area when oop may be forwarded."
	"Warning: No young objects should be recorded as roots. Callers are responsible for ensuring this constraint is not violated."
	<inline: false> "for debugging..."
	| header fwdBlock headerLoc |
	"If labelled, gcc duplicates the label when inlining updatePointersInRangeFrom:to:"
	header := self longAt: oop.
	(header bitAnd: MarkBit) ~= 0
		ifTrue: "This oop is forwarded"
			[fwdBlock := (header bitAnd: AllButMarkBitAndTypeMask) << 1.
			self assert: (self fwdBlockValid: fwdBlock).
			headerLoc := fwdBlock + self wordSize]
		ifFalse: "Normal -- no forwarding"
			[headerLoc := oop].
	"use headerLoc var to eliminate duplication on inlining noteAsRoot:headerLoc:
	 older versions of this method had two separate sends of noteAsRoot:headerLoc:"
	self noteAsRoot: oop headerLoc: headerLoc! !

!ObjectMemory methodsFor: 'become' stamp: 'brp 9/19/2003 16:09'!
become: array1 with: array2

	^ self become: array1 with: array2 twoWay: true copyHash: true
! !

!ObjectMemory methodsFor: 'become' stamp: 'eem 9/18/2008 09:44'!
become: array1 with: array2 twoWay: twoWayFlag copyHash: copyHashFlag 
	"All references to each object in array1 are swapped with all references to the corresponding
	 object in array2. That is, all pointers to one object are replaced with with pointers to the other.
	 The arguments must be arrays of the same length.  Ansswer if the operation succeeds."

	"Implementation: Uses forwarding blocks to update references as done in compaction."
	(self isArray: array1) ifFalse: [^false].
	(self isArray: array2) ifFalse: [^false].
	(self lastPointerOf: array1) = (self lastPointerOf: array2) ifFalse: [^false].
	(twoWayFlag or: [copyHashFlag])
		ifTrue: [(self containOnlyOops: array1 and: array2) ifFalse: [^false]]
		ifFalse: [(self containOnlyOops: array1) ifFalse: [^false]].

	(self prepareForwardingTableForBecoming: array1 with: array2 twoWay: twoWayFlag) ifFalse:
		[^false]. "fail; not enough space for forwarding table"

	(self allYoung: array1 and: array2)
		ifTrue: ["sweep only the young objects plus the roots"
			self mapPointersInObjectsFrom: youngStart to: endOfMemory]
		ifFalse: ["sweep all objects"
			self mapPointersInObjectsFrom: self startOfMemory to: endOfMemory].
	twoWayFlag
		ifTrue: [self restoreHeadersAfterBecoming: array1 with: array2]
		ifFalse: [self restoreHeadersAfterForwardBecome: copyHashFlag].

	self initializeMemoryFirstFree: freeBlock. "re-initialize memory used for forwarding table"
	
	self forceInterruptCheck. "pretty much guaranteed to take a long time, so check for timers etc ASAP"

	^true "success"! !

!ObjectMemory methodsFor: 'garbage collection' stamp: 'ar 5/18/2007 07:53'!
biasToGrow
	| growSize |
	growSize :=  growHeadroom*3/2 - (self sizeOfFree: freeBlock).
	self growObjectMemory: growSize! !

!ObjectMemory methodsFor: 'garbage collection' stamp: 'eem 2/13/2009 06:48'!
biasToGrowCheckGCLimit
	| growth |
	growth := (youngStart - self startOfMemory) - gcBiasToGrowThreshold.
	growth < 0 ifTrue: [gcBiasToGrowThreshold := youngStart - memory].
	growth > gcBiasToGrowGCLimit
		 ifTrue: 
			[self fullGC.
			gcBiasToGrowThreshold := youngStart - memory].

					! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'nice 11/8/2019 09:41'!
booleanObjectOf: aBool
	<inline: true>
	^aBool ifTrue: [trueObj] ifFalse: [falseObj]! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 8/8/2014 11:33'!
bootstrapping
	^false! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 3/10/2014 12:17'!
byteFormatForNumBytes: numBytes
	^self firstByteFormat + (4 - numBytes bitAnd: 3)! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 9/10/2013 11:06'!
byteFormatMask
	^16rC! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 07/28/2014 08:07'!
byteSizeOf: oop
	<api>
	(self isIntegerObject: oop) ifTrue:[^0].
	^self numBytesOf: oop! !

!ObjectMemory methodsFor: 'image segment in/out' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
byteSwapByteObjectsFrom: startOop to: stopAddr flipFloatsIf: flipFloatWords
	"Byte-swap the words of all bytes objects in a range of the 
	image, including Strings, ByteArrays, and CompiledMethods. 
	This returns these objects to their original byte ordering 
	after blindly byte-swapping the entire image. For compiled 
	methods, byte-swap only their bytecodes part.  For Floats
	swap their most and least significant words if required."
	| oop fmt temp wordAddr |
	oop := startOop.
	[self oop: oop isLessThan: stopAddr] whileTrue:
		[(self isFreeObject: oop) ifFalse:
			[fmt := self formatOf: oop.
			fmt >= self firstByteFormat ifTrue:
				["oop contains bytes; unswap"
				wordAddr := oop + self baseHeaderSize.
				fmt >= self firstCompiledMethodFormat ifTrue: "compiled method; start after methodHeader and literals"
					[wordAddr := wordAddr + ((self literalCountOf: oop) + LiteralStart * self bytesPerOop)].
				self reverseBytesFrom: wordAddr to: oop + (self sizeBitsOf: oop)].
			fmt = self firstLongFormat  ifTrue: "Bitmap, Float etc"
				[(self compactClassIndexOf: oop) = ClassFloatCompactIndex
					ifTrue:
						[flipFloatWords ifTrue:
							[temp := self longAt: oop + self baseHeaderSize.
							 self longAt: oop + self baseHeaderSize put: (self longAt: oop + self baseHeaderSize + 4).
							 self longAt: oop + self baseHeaderSize + 4 put: temp]]
					ifFalse:
						[self wordSize = 8 ifTrue:
							["Object contains 32-bit half-words packed into 64-bit machine words."
							wordAddr := oop + self baseHeaderSize.
							self reverseWordsFrom: wordAddr to: oop + (self sizeBitsOf: oop)]]]].
			oop := self objectAfter: oop]! !

!ObjectMemory methodsFor: 'memory access' stamp: 'nice 4/3/2016 08:50'!
byteSwapped32IfBigEndian: w
	"If the VM is big endian, then swap the bytes of w"
	<api>
	<var: 'w' type: #'unsigned int'>
	<returnTypeC: #'unsigned int'>
	self cppIf: VMBIGENDIAN
		ifTrue: [^w byteSwap32]
		ifFalse: [^w bitAnd: 16rFFFFFFFF]! !

!ObjectMemory methodsFor: 'memory access' stamp: 'nice 4/3/2016 08:50'!
byteSwapped64IfBigEndian: w
	"If the VM is big endian, then swap the bytes of w"
	<api>
	<var: 'w' type: #'unsigned long long'>
	<returnTypeC: #'unsigned long long'>
	self cppIf: VMBIGENDIAN
		ifTrue: [^w byteSwap64]
		ifFalse: [^w bitAnd: 16rFFFFFFFFFFFFFFFF]! !

!ObjectMemory methodsFor: 'image save/restore' stamp: 'nice 4/3/2016 16:23'!
byteSwapped: w
	"Answer the given integer with its bytes in the reverse order."
	<api>
	<returnTypeC: #sqInt>
	^self cppIf: self wordSize = 4
		ifTrue: [w byteSwap32]
		ifFalse: [w byteSwap64]! !

!ObjectMemory methodsFor: 'allocation' stamp: 'ar 2/25/2001 17:45'!
bytesLeft: includingSwap
	^(self sizeOfFree: freeBlock) "already commited"
		+ (self sqMemoryExtraBytesLeft: includingSwap).! !

!ObjectMemory methodsFor: 'forward compatibility' stamp: 'nice 12/26/2020 23:24'!
bytesPerElement: oop
	"forward compatibility with Spur"
	(self isImmediate: oop) ifTrue:
		[^0].
	(self isWords: oop) ifTrue:
		[^4].
	(self isBytes: oop) ifTrue:
		[^1].
	^self wordSize! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 11/10/2014 10:32'!
bytesPerOop
	"N.B. This would appear to hard-code the header size for 32-bit images.  But if generating
	 a 64-bit image, this method could be removed and the relevant one substituted.  We can't
	 mark this method as <doNotGenerate> as we need an actual method to guide code gen."
	^4! !

!ObjectMemory methodsFor: 'simulation support' stamp: 'nice 4/20/2016 00:20'!
cCoerceSimple: value to: cTypeString
	<doNotGenerate>
	^cTypeString caseOf:
	   {	[#'char *']			->	[value].
		[#'unsigned int']	->	[value]. }! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
changeClassOf: rcvr to: argClass
	"Attempt to change the class of the receiver to the argument given that the
	 format of the receiver matches the format of the argument.  If successful,
	 answer 0, otherwise answer an error code indicating the reason for failure. 
	 Fail if the receiver is an instance of a compact class and the argument isn't,
	 or if the format of the receiver is incompatible with the format of the argument,
	 or if the argument is a fixed class and the receiver's size differs from the size
	 that an instance of the argument should have."
	| classHdr sizeHiBits argClassInstByteSize argFormat rcvrFormat rcvrHdr ccIndex |
	"Check what the format of the class says"
	classHdr := self formatOfClass: argClass. "Low 2 bits are 0"

	"Compute the size of instances of the class (used for fixed field classes only)"
	sizeHiBits := (classHdr bitAnd: 16r60000) >> 9.
	classHdr := classHdr bitAnd: 16r1FFFF.
	argClassInstByteSize := (classHdr bitAnd: SizeMask) + sizeHiBits. "size in bytes -- low 2 bits are 0"

	"Check the receiver's format against that of the class"
	argFormat := self formatOfHeader: classHdr.
	rcvrHdr := self baseHeader: rcvr.
	rcvrFormat := self formatOfHeader: rcvrHdr.
	"If the receiver is a byte object we need to clear the number of odd bytes from the format."
	rcvrFormat > self firstByteFormat ifTrue:
		[rcvrFormat := rcvrFormat bitAnd: 16rC].
	argFormat = rcvrFormat ifFalse:
		[^PrimErrInappropriate]. "no way"

	"For fixed field classes, the sizes must match.
	Note: argClassInstByteSize-4 because base header is included in class size."
	argFormat < self arrayFormat
		ifTrue:
			[(argClassInstByteSize - self baseHeaderSize) ~= (self numBytesOf: rcvr) ifTrue:
				[^PrimErrBadReceiver]]
		ifFalse:
			[argFormat = self indexablePointersFormat ifTrue: "For indexable plus fixed fields the receiver must be at least big enough."
				[(argClassInstByteSize - self baseHeaderSize) > (self numBytesOf: rcvr) ifTrue:
					[^PrimErrBadReceiver]]].

	(self headerTypeOfHeader: rcvrHdr) = HeaderTypeShort
		ifTrue: "Compact classes. Check if the arg's class is compact and exchange ccIndex"
			[ccIndex := classHdr bitAnd: CompactClassMask.
			ccIndex = 0 ifTrue:
				[^PrimErrInappropriate]. "class is not compact"
			self cppIf: IMMUTABILITY
				ifTrue: [(rcvrHdr bitAnd: ImmutabilityBit) ~= 0 ifTrue:
							[^PrimErrNoModification]].
			self baseHeader: rcvr put: ((rcvrHdr bitClear: CompactClassMask) bitOr: ccIndex)]
		ifFalse: "Exchange the class pointer, which could make rcvr a root for argClass"
			[self cppIf: IMMUTABILITY
				ifTrue: [(rcvrHdr bitAnd: ImmutabilityBit) ~= 0 ifTrue:
							[^PrimErrNoModification]].
			"N.B. the recursive scan-mark algorithm uses the header word's size and compact class
			 fields to determine the header type when it reuses the header type bits for the mark
			 state.  So it is alas an invariant that non-compact headers have a 0 compact class field."
			(self compactClassIndexOfHeader: rcvrHdr) ~= 0 ifTrue:
				[self baseHeader: rcvr put: (rcvrHdr bitClear: CompactClassMask)].			
			self longAt: rcvr - self baseHeaderSize put: (argClass bitOr: (self headerTypeOfHeader: rcvrHdr)).
			(self oop: rcvr isLessThan: youngStart) ifTrue:
				[self possibleRootStoreInto: rcvr value: argClass]].
	"ok"
	^0! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 4/11/2014 10:41'!
characterObjectOf: characterCode
	<api>
	^(characterCode between: 0 and: 255)
		ifTrue: [self fetchPointer: characterCode ofObject: self characterTable]
		ifFalse: [nilObj]! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'eem 2/15/2009 11:58'!
characterTable
	<api>
	^self splObj: CharacterTable! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'eem 11/22/2013 14:53'!
cheapIsInMemory: address 
	"Answer if the given address is in ST object memory.  For simulation only."
	<doNotGenerate>
	^address >= self startOfMemory
	  and: [address < endOfMemory]! !

!ObjectMemory methodsFor: 'memory access' stamp: 'dtl 4/22/2007 18:33'!
checkAddress: byteAddress 
	"Keep this method around for debugging the C code."
	(self oop: byteAddress isLessThan: self startOfMemory)
		ifTrue: [self error: 'bad address: negative'].
	(self oop: byteAddress isGreaterThanOrEqualTo: memoryLimit)
		ifTrue: [self error: 'bad address: past end of heap']! !

!ObjectMemory methodsFor: 'initialization' stamp: 'eem 12/20/2012 11:06'!
checkCompactIndex: compactIndex isClass: specialIndex named: name
	"Check that a class the VM assumes is compact has the right index."
	<inline: true> "macrofication of the name arg in invalidCompactClassError only works if this method is inlined so the name is a string literal not a parameter"
	(compactIndex ~= 0
	 and: [(self splObj: specialIndex) ~= (self compactClassAt: compactIndex)]) ifTrue:
		[self invalidCompactClassError: name]! !

!ObjectMemory methodsFor: 'memory access' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
checkHeapIntegrity
	"Perform an integrity/leak check using the heapMap.  Assume
	 clearLeakMapAndMapAccessibleObjects has set a bit at each
	 object's header.  Scan all objects in the heap checking that every
	 pointer points to a header.  Scan the rootTable, remapBuffer and
	 extraRootTable checking that every entry is a pointer to a header.
	 Check that the number of roots is correct and that all rootTable
	 entries have their rootBit set. Answer if all checks pass."
	| ok obj sz hdr fmt fi fieldOop numRootsInHeap |
	<inline: false>
	ok := true.
	numRootsInHeap := 0.
	obj := self firstObject.
	[self oop: obj isLessThan: self startOfFreeSpace] whileTrue:
		[(self isFreeObject: obj)
			ifTrue:
				[sz := self sizeOfFree: obj]
			ifFalse:
				[hdr := self baseHeader: obj.
				 (self isYoungRootHeader: hdr) ifTrue:
					[numRootsInHeap := numRootsInHeap + 1].
				 (self compactClassIndexOfHeader: hdr) = 0 ifTrue:
					[fieldOop := (self classHeader: obj) bitAnd: AllButTypeMask.
					 ((self isIntegerObject: fieldOop)
					   or: [(self heapMapAtWord: (self pointerForOop: fieldOop)) = 0]) ifTrue:
						[self print: 'object leak in '; printHex: obj; print: ' class = '; printHex: fieldOop; cr.
						 ok := false]].
				 fmt := self formatOfHeader: hdr.
				 (fmt <= self lastPointerFormat or: [fmt >= self firstCompiledMethodFormat]) ifTrue:
					[fmt >= self firstCompiledMethodFormat
						ifTrue: [fi := (self literalCountOf: obj) + LiteralStart]
						ifFalse: [(fmt = self indexablePointersFormat and: [self isContextHeader: hdr])
									ifTrue: [fi := CtxtTempFrameStart + (self fetchStackPointerOf: obj)]
									ifFalse: [fi := self lengthOf: obj]].
					[(fi := fi - 1) >= 0] whileTrue:
						[fieldOop := self fetchPointer: fi ofObject: obj.
						 (self isNonIntegerObject: fieldOop) ifTrue:
							[(fieldOop bitAnd: self wordSize - 1) ~= 0
								ifTrue:
									[self print: 'misaligned oop in '; printHex: obj; print: ' @ '; printNum: fi; print: ' = '; printHex: fieldOop; cr.
									 ok := false]
								ifFalse:
									[(self heapMapAtWord: (self pointerForOop: fieldOop)) = 0 ifTrue:
										[self print: 'object leak in '; printHex: obj; print: ' @ '; printNum: fi; print: ' = '; printHex: fieldOop; cr.
										 ok := false]]]]].
				 sz := self sizeBitsOf: obj].
		 obj := self oopFromChunk: obj + sz].
	numRootsInHeap ~= rootTableCount ifTrue:
		[self print: 'root count mismatch. #heap roots '; printNum: numRootsInHeap; print: '; #roots '; printNum: rootTableCount; cr.
		"But the system copes with overflow..."
		ok := rootTableOverflowed and: [allocationCount > allocationsBetweenGCs]].
	1 to: rootTableCount do:
		[:ri|
		obj := rootTable at: ri.
		(obj bitAnd: self wordSize - 1) ~= 0
			ifTrue:
				[self print: 'misaligned oop in rootTable @ '; printNum: ri; print: ' = '; printHex: obj; cr.
				 ok := false]
			ifFalse:
				[(self heapMapAtWord: (self pointerForOop: obj)) = 0
					ifTrue:
						[self print: 'object leak in rootTable @ '; printNum: ri; print: ' = '; printHex: obj; cr.
						 ok := false]
					ifFalse:
						[hdr := self baseHeader: obj.
						 (self isYoungRootHeader: hdr) ifFalse:
							[self print: 'non-root in rootTable @ '; printNum: ri; print: ' = '; printHex: obj; cr.
							 ok := false]]]].
	1 to: remapBufferCount do:
		[:ri|
		obj := remapBuffer at: ri.
		(obj bitAnd: self wordSize - 1) ~= 0
			ifTrue:
				[self print: 'misaligned remapRoot @ '; printNum: ri; print: ' = '; printHex: obj; cr.
				 ok := false]
			ifFalse:
				[(self heapMapAtWord: (self pointerForOop: obj)) = 0
					ifTrue:
						[self print: 'object leak in remapRoots @ '; printNum: ri; print: ' = '; printHex: obj; cr.
						 ok := false]]].
	1 to: extraRootCount do:
		[:ri|
		obj := (extraRoots at: ri) at: 0.
		(obj bitAnd: self wordSize - 1) ~= 0
			ifTrue:
				[self print: 'misaligned extraRoot @ '; printNum: ri; print: ' => '; printHex: obj; cr.
				 ok := false]
			ifFalse:
				[(self heapMapAtWord: (self pointerForOop: obj)) = 0
					ifTrue:
						[self print: 'object leak in extraRoots @ '; printNum: ri; print: ' => '; printHex: obj; cr.
						 ok := false]]].
	^ok! !

!ObjectMemory methodsFor: 'debug support' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
checkOkayOop: oop
	"Verify that the given oop is legitimate. Check address, header, and size but not class.
	 Answer true if OK.  Otherwise print reason and answer false."

	<api>
	<var: #oop type: #usqInt>
	| sz type fmt unusedBit |

	"address and size checks"
	(self isIntegerObject: oop) ifTrue: [ ^true ].
	(oop >= self startOfMemory and: [oop < endOfMemory])
		ifFalse: [ self print: 'oop is not a valid address'; cr. ^false ].
	((oop \\ self wordSize) = 0)
		ifFalse: [ self print: 'oop is not a word-aligned address'; cr. ^false ].
	sz := self sizeBitsOf: oop.
	(oop + sz) < endOfMemory
		ifFalse: [ self print: 'oop size would make it extend beyond the end of memory'; cr. ^false ].

	"header type checks"
	type := self headerType: oop.
	type = HeaderTypeFree
		ifTrue:  [ self print: 'oop is a free chunk, not an object'; cr. ^false ].
	type = HeaderTypeShort ifTrue: [
		(self compactClassIndexOf: oop) = 0
			ifTrue:  [ self print: 'cannot have zero compact class field in a short header'; cr. ^false ].
	].
	type = HeaderTypeClass ifTrue: [
		((oop >= self wordSize) and: [(self headerType: oop - self wordSize) = type])
			ifFalse: [ self print: 'class header word has wrong type'; cr. ^false ].
	].
	type = HeaderTypeSizeAndClass ifTrue: [
		((oop >= (self wordSize*2)) and:
		 [(self headerType: oop - (self wordSize*2)) = type and:
		 [(self headerType: oop - self wordSize) = type]])
			ifFalse: [ self print: 'class header word has wrong type'; cr. ^false ].
	].

	"format check"
	fmt := self formatOf: oop.
	((fmt = 5) | (fmt = 7))
		ifTrue:  [ self print: 'oop has an unknown format type'; cr. ^false ].

	"mark and root bit checks"
	unusedBit := 16r20000000.
	self wordSize = 8
		ifTrue:
			[unusedBit := unusedBit << 16.
			 unusedBit := unusedBit << 16].
	((self longAt: oop) bitAnd: unusedBit) = 0
		ifFalse: [ self print: 'unused header bit 30 is set; should be zero'; cr. ^false ].
"xxx
	((self longAt: oop) bitAnd: MarkBit) = 0
		ifFalse: [ self print: 'mark bit should not be set except during GC' ].
xxx"
	((self isYoungRoot: oop) and: [oop >= youngStart])
		ifTrue: [ self print: 'root bit is set in a young object'; cr. ^false ].
	^true
! !

!ObjectMemory methodsFor: 'debug support' stamp: 'eem 1/16/2013 17:04'!
checkOkayYoungReferrer: obj
	"Verify that the given obj is a valid youngReferrer. Check RootBit is set and
	 is in rootTable.  Answer true if OK.  Otherwise print reason and answer false.
	 Assumes the object contains young references."

	(self oop: obj isGreaterThanOrEqualTo: youngStart) ifTrue:
		[^true].

	(self isYoungRoot: obj) ifFalse:
		[ self print: 'root bit is not set in '; printHex: obj; cr. ^false ].

	1 to: rootTableCount do:
		[:i| obj = (rootTable at: i) ifTrue: [^true]].

	self printHex: obj; print: ' has root bit set but is not in rootTable'; cr.

	^false
! !

!ObjectMemory methodsFor: 'debug support' stamp: 'eem 12/4/2013 17:26'!
checkOopHasOkayClass: obj
	"Attempt to verify that the given obj has a reasonable behavior. The class must be a
	 valid, non-integer oop and must not be nilObj. It must be a pointers object with three
	 or more fields. Finally, the instance specification field of the behavior must match that
	 of the instance. If OK answer true.  If  not, print reason and answer false."

	<api>
	<var: #obj type: #usqInt>
	| objClass formatMask behaviorFormatBits objFormatBits |
	<var: #objClass type: #usqInt>

	(self checkOkayOop: obj) ifFalse:
		[^false].
	objClass := self cCoerce: (self fetchClassOfNonImm: obj) to: #usqInt.

	(self isIntegerObject: objClass) ifTrue:
		[self print: 'obj '; printHex: obj; print: ' a SmallInteger is not a valid class or behavior'; cr. ^false].
	(self okayOop: objClass) ifFalse:
		[self print: 'obj '; printHex: obj; print: ' class obj is not ok'; cr. ^false].
	((self isPointersNonImm: objClass) and: [(self lengthOf: objClass) >= 3]) ifFalse:
		[self print: 'obj '; printHex: obj; print: ' a class (behavior) must be a pointers object of size >= 3'; cr. ^false].
	formatMask := (self isBytes: obj)
						ifTrue: [16rC00]  "ignore extra bytes size bits"
						ifFalse: [16rF00].

	behaviorFormatBits := (self formatOfClass: objClass) bitAnd: formatMask.
	objFormatBits := (self baseHeader: obj) bitAnd: formatMask.
	behaviorFormatBits = objFormatBits ifFalse:
		[self print: 'obj '; printHex: obj; print: ' and its class (behavior) formats differ'; cr. ^false].
	^true! !

!ObjectMemory methodsFor: 'debug support' stamp: 'eem 6/19/2009 15:53'!
checkOopIntegrity: obj named: name
	<inline: false>
	<var: #name type: #'char *'>
	(self heapMapAtWord: (self pointerForOop: obj)) ~= 0 ifTrue:
		[^true].
	self print: name; print: ' leak '; printHex: obj; cr.
	^false! !

!ObjectMemory methodsFor: 'debug support' stamp: 'eem 09/05/2021 19:12'!
checkOopIntegrity: obj named: name index: i
	<inline: false>
	<var: #name type: #'char *'>
	(self heapMapAtWord: (self pointerForOop: obj)) ~= 0 ifTrue:
		[^true].
	self print: name; print: ' leak @ '; printNum: i; print: ' = '; printHex: obj; eekcr.
	^false! !

!ObjectMemory methodsFor: 'memory access' stamp: ''!
checkedByteAt: byteAddress
	"Assumes zero-based array indexing."

	self checkAddress: byteAddress.
	^ self byteAt: byteAddress! !

!ObjectMemory methodsFor: 'memory access' stamp: ''!
checkedByteAt: byteAddress put: byte
	"Assumes zero-based array indexing."

	self checkAddress: byteAddress.
	self byteAt: byteAddress put: byte.! !

!ObjectMemory methodsFor: 'memory access' stamp: 'eem 8/19/2010 12:30'!
checkedLongAt: byteAddress
	"Assumes zero-based array indexing. For testing in Smalltalk, this method should be overridden in a subclass."
	<api>
	self checkAddress: byteAddress.
	self checkAddress: byteAddress + 3.
	^ self longAt: byteAddress! !

!ObjectMemory methodsFor: 'memory access' stamp: ''!
checkedLongAt: byteAddress put: a32BitInteger
	"Assumes zero-based array indexing. For testing in Smalltalk, this method should be overridden in a subclass."

	self checkAddress: byteAddress.
	self checkAddress: byteAddress + 3.
	self longAt: byteAddress put: a32BitInteger.! !

!ObjectMemory methodsFor: 'oop/chunk conversion' stamp: 'JMM 12/4/2002 19:55'!
chunkFromOop: oop
	"Compute the chunk of this oop by subtracting its extra header bytes."

	^ oop - (self extraHeaderBytes: oop)! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'eem 11/25/2008 17:10'!
classAlien
	^self splObj: ClassAlien! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'eem 4/2/2009 22:50'!
classArray
	<api>
	^self splObj: ClassArray! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'ar 10/7/1998 18:06'!
classBitmap
	^self splObj: ClassBitmap! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'ar 10/7/1998 18:06'!
classByteArray
	^self splObj: ClassByteArray! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'ar 10/7/1998 18:07'!
classCharacter
	^self splObj: ClassCharacter! !

!ObjectMemory methodsFor: 'forward compatibility' stamp: 'nice 12/26/2020 23:31'!
classDoubleByteArray
	"forward compatibility with Spur"
	^nil! !

!ObjectMemory methodsFor: 'forward compatibility' stamp: 'nice 12/26/2020 23:31'!
classDoubleWordArray
	"forward compatibility with Spur"
	^nil! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'ar 11/21/1999 00:50'!
classExternalAddress
	^self splObj: ClassExternalAddress! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'ar 11/17/1999 18:23'!
classExternalData
	^self splObj: ClassExternalData! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'ar 11/17/1999 18:23'!
classExternalFunction
	^self splObj: ClassExternalFunction! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'ar 11/17/1999 19:45'!
classExternalLibrary
	^self splObj: ClassExternalLibrary! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'ar 11/21/1999 15:23'!
classExternalStructure
	^self splObj: ClassExternalStructure! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'eem 4/20/2009 20:58'!
classFloat
	<api>
	^self splObj: ClassFloat! !

!ObjectMemory methodsFor: 'forward compatibility' stamp: 'nice 12/26/2020 23:33'!
classFloat32Array
	"forward compatibility with Spur"
	^nil! !

!ObjectMemory methodsFor: 'forward compatibility' stamp: 'nice 12/26/2020 23:31'!
classFloat64Array
	"forward compatibility with Spur"
	^nil! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
classHeader: oop
	<api>
	^self longAt: oop - self baseHeaderSize! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'ar 11/21/1999 23:22'!
classLargeNegativeInteger
	^self splObj: ClassLargeNegativeInteger! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'ar 10/7/1998 18:07'!
classLargePositiveInteger
	^self splObj: ClassLargePositiveInteger! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'eem 4/20/2009 13:39'!
classMutex
	^self splObj: ClassMutex! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'ar 10/7/1998 18:07'!
classPoint
	^self splObj: ClassPoint! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'ar 10/7/1998 18:11'!
classSemaphore
	^self splObj: ClassSemaphore! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'eem 7/7/2014 13:47'!
classSmallInteger
	<api>
	^self splObj: ClassSmallInteger! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'eem 1/2/2013 17:37'!
classString
	^self splObj: ClassByteString! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'eem 11/25/2008 17:10'!
classUnsafeAlien
	^self splObj: ClassUnsafeAlien! !

!ObjectMemory methodsFor: 'forward compatibility' stamp: 'nice 12/26/2020 23:39'!
classWordArray
	"forward compatibility with Spur"
	^nil! !

!ObjectMemory methodsFor: 'debug support' stamp: 'eem 5/9/2011 14:36'!
clearLeakMapAndMapAccessibleObjects
	"Perform an integrity/leak check using the heapMap.  Set a bit at each object's header."
	| oop |
	<inline: false>
	self clearHeapMap.
	oop := self firstObject.
	[oop = nil] whileFalse:
		[self heapMapAtWord: (self pointerForOop: oop) Put: 1.
		 oop := self accessibleObjectAfter: oop]! !

!ObjectMemory methodsFor: 'garbage collection' stamp: 'eem 1/18/2013 10:11'!
clearRootsTable
	"Clear the root bits of the current roots, then empty the roots table. "
	"Caution: This should only be done when the young object space is empty."
	"reset the roots table (after this, all objects are old so there are no roots)"
	1 to: rootTableCount do:
		[:i | | oop |
		oop := rootTable at: i.
		self longAt: oop put: ((self longAt: oop) bitAnd: AllButRootBit).
		rootTable at: i put: 0].
	rootTableCount := 0.
	rootTableOverflowed := false.! !

!ObjectMemory methodsFor: 'allocation' stamp: 'eem 01/07/2021 10:40'!
cloneObject: obj
	"Return a shallow copy of the given object. May cause GC"
	"Assume: Oop is a real object, not a small integer."

	| extraHdrBytes bytes newChunk remappedOop fromIndex toIndex lastFrom newOop header hash |
	<inline: false>
	<var: #lastFrom type: #usqInt>
	<var: #fromIndex type: #usqInt>
	self assert: (self isNonIntegerObject: obj).
	extraHdrBytes := self extraHeaderBytes: obj.
	bytes := self sizeBitsOf: obj.
	bytes := bytes + extraHdrBytes.

	"allocate space for the copy, remapping obj in case of a GC"
	self pushRemappableOop: obj.
	"check it is safe to allocate this much memory. Return 0 if not"
	(self sufficientSpaceToAllocate: 2500 + bytes) ifFalse:[^0].
	newChunk := self allocateChunk: bytes.
	remappedOop := self popRemappableOop.

	"copy old to new including all header words"
	toIndex := newChunk - self wordSize.  "loop below uses pre-increment"
	fromIndex := (remappedOop - extraHdrBytes) - self wordSize.
	lastFrom := fromIndex + bytes.
	[fromIndex < lastFrom] whileTrue:
		[self longAt: (toIndex := toIndex + self wordSize) put: (self longAt: (fromIndex := fromIndex + self wordSize))].
	newOop := newChunk + extraHdrBytes.  "convert from chunk to oop"

	"fix base header: compute new hash and clear Mark and Root bits"
	hash := self newObjectHash.
	header := (self longAt: newOop) bitAnd: 16r1FFFF.
	"use old ccIndex, format, size, and header-type fields"
	header := header bitOr: ((hash << HashBitsOffset) bitAnd: HashBits).
	self longAt: newOop put: header.
	^newOop
! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 12/20/2012 11:03'!
compactClassAt: ccIndex
	"Index must be between 1 and compactClassArray size.  A zero compact class
	 index in the base header indicates that the class is in the class header word."
	<api>
	<inline: true>
	^self fetchPointer: ccIndex - 1 ofObject: (self splObj: CompactClasses)! !

!ObjectMemory methodsFor: 'cog jit support' stamp: 'eem 12/12/2015 14:12'!
compactClassFieldLSB
	<api>
	<cmacro>
	^12! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 7/7/2014 08:40'!
compactClassIndexOf: oop
	<api>
	<inline: true>
	^((self baseHeader: oop) >> self compactClassFieldLSB) bitAnd: 16r1F! !

!ObjectMemory methodsFor: 'object format' stamp: 'eem 7/31/2014 05:51'!
compactClassIndexOfClass: classObj
	<api>
	^self compactClassIndexOfHeader: (self formatOfClass: classObj)! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 7/7/2014 09:34'!
compactClassIndexOfHeader: header
	<api>
	<inline: true>
	^(header >> self compactClassFieldLSB) bitAnd: 16r1F! !

!ObjectMemory methodsFor: 'initialization' stamp: 'eem 11/30/2010 15:51'!
compactIndexOfClass: aClassOop
	"Check that a class the VM assumes is compact has the right index."
	| cct |
	cct := self splObj: CompactClasses.
	(self lengthOf: cct) - 1 to: 0 by: -2 do:
		[:index|
		(self fetchPointer: index ofObject: cct) = aClassOop ifTrue:
			[^index + 1]].
	^0! !

!ObjectMemory methodsFor: 'become' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
containOnlyMutableOops: array1 and: array2 
	"Return true if neither array contains an immutable.
	 You may not be allowed to become: immutables."
	| fieldOffset |
	fieldOffset := self lastPointerOf: array1.
	"same size as array2"
	[fieldOffset >= self baseHeaderSize] whileTrue:
		[(self isOopImmutable: (self longAt: array1 + fieldOffset)) ifTrue: [^false].
		 (self isOopImmutable: (self longAt: array2 + fieldOffset)) ifTrue: [^false].
		 fieldOffset := fieldOffset - self wordSize].
	^true! !

!ObjectMemory methodsFor: 'become' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
containOnlyOops: array
	"Answer if the array does not contain a small integer. You 
	  can't become: SmallIntegers!!"
	| fieldOffset |
	fieldOffset := self lastPointerOf: array.
	"same size as array2"
	[fieldOffset >= self baseHeaderSize] whileTrue:
		[(self isIntegerObject: (self longAt: array + fieldOffset)) ifTrue: [^ false].
		 fieldOffset := fieldOffset - self bytesPerOop].
	^true! !

!ObjectMemory methodsFor: 'become' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
containOnlyOops: array1 and: array2 
	"Answer if neither array contains a small integer. You 
	can't become: integers!!"
	| fieldOffset |
	fieldOffset := self lastPointerOf: array1.
	"same size as array2"
	[fieldOffset >= self baseHeaderSize]
		whileTrue: [(self isIntegerObject: (self longAt: array1 + fieldOffset)) ifTrue: [^ false].
			(self isIntegerObject: (self longAt: array2 + fieldOffset)) ifTrue: [^ false].
			fieldOffset := fieldOffset - self bytesPerOop].
	^ true! !

!ObjectMemory methodsFor: 'simulation' stamp: 'eem 11/18/2016 17:02'!
convertToArray
	<doNotGenerate>
	"I dont believe it -- this *just works*"

	memory:= memory as: Array! !

!ObjectMemory methodsFor: 'image segment in/out' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
copyObj: oop toSegment: segmentWordArray addr: lastSeg stopAt: stopAddr saveOopAt: oopPtr headerAt: hdrPtr
	"Copy this object into the segment beginning at lastSeg.
	Install a forwarding pointer, and save oop and header.
	Fail if out of space.  Return the next segmentAddr if successful."

	"Copy the object..."
	| extraSize bodySize hdrAddr |
	<inline: false>
	self flag: #Dan.  "None of the imageSegment stuff has been updated for 64 bits"
	extraSize := self extraHeaderBytes: oop.
	bodySize := self sizeBitsOf: oop.
	(self oop: (lastSeg + extraSize + bodySize) isGreaterThanOrEqualTo: stopAddr) ifTrue:
		[^0]. "failure"
	self transfer: extraSize + bodySize // self wordSize  "wordCount"
		from: oop - extraSize
		to: lastSeg+self wordSize.

	"Clear root and mark bits of all headers copied into the segment"
	hdrAddr := lastSeg+self wordSize + extraSize.
	self longAt: hdrAddr put: ((self longAt: hdrAddr) bitAnd: AllButRootBit - MarkBit).

	self forward: oop to: (lastSeg+self wordSize + extraSize - segmentWordArray)
		savingOopAt: oopPtr
		andHeaderAt: hdrPtr.

	"Return new end of segment"
	^lastSeg + extraSize + bodySize! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 12/1/2014 10:56'!
dbgFloatValueOf: oop
	"Answer the C double precision floating point value of the argument,
	 or if it is not, answer 0."

	| isFloat result |
	<returnTypeC: #double>
	<var: #result type: #double>
	isFloat := self isFloatInstance: oop.
	isFloat ifTrue:
		[self cCode: '' inSmalltalk: [result := Float new: 2].
		 self fetchFloatAt: oop + self baseHeaderSize into: result.
		 ^result].
	^0.0! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'ar 10/7/1998 18:12'!
displayObject
	^self splObj: TheDisplay! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
endOfMemory
	^endOfMemory! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 10/18/2010 10:23'!
ensureBehaviorHash: oop

	^self hashBitsOf: oop! !

!ObjectMemory methodsFor: 'forward compatibility' stamp: 'eem 1/17/2015 13:54'!
ensureSemaphoreUnforwardedThroughContext: aSemaphore
	"Compatibility with SpurMemoryManager"
	<inline: true>! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 4/17/2008 15:27'!
existImmutableReferencesToForwardedInRangeFrom: memStart to: memEnd
	"Answer if any immutable objects refer to any forwarded objects so that the become: primitives can fail if a become would update a reference from an immutable object."

	| oop |
	1 to: rootTableCount do:
		[:i | 
		oop := rootTable at: i.
		(oop < memStart or: [oop >= memEnd]) ifTrue:
			["Note: must not remap the fields of any object twice!!"
			"remap this oop only if not in the memory range 
			covered below"
			(self objectIsImmutableAndReferencesForwarded: oop) ifTrue:
				[^true]]].

	oop := self oopFromChunk: memStart.
	[oop < memEnd] whileTrue:
		[(self isFreeObject: oop) ifFalse:
			[(self objectIsImmutableAndReferencesForwarded: oop) ifTrue:
				[^true]].
		oop := self objectAfterWhileForwarding: oop].
	^false! !

!ObjectMemory methodsFor: 'oop/chunk conversion' stamp: 'JMM (auto pragmas 12/08) 7/7/2003 14:18'!
extraHeaderBytes: oopOrChunk
	"Return the number of extra bytes used by the given object's header."
	"Warning: This method should not be used during marking, when the header type bits of an object may be incorrect."

	"JMM should be an array lookup!!" 
	<inline: true>
	^ headerTypeBytes at: (self headerType: oopOrChunk).! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'eem 3/26/2009 19:49'!
falseObject
	<api>
	^falseObj! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
fetchByte: byteIndex ofObject: oop
	<api>
	^self byteAt: oop + self baseHeaderSize + byteIndex! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 11/10/2014 12:49'!
fetchClassOf: oop 
	| ccIndex |
	<inline: true>
	^(self isIntegerObject: oop)
		ifTrue: [self splObj: ClassSmallInteger]
		ifFalse:
			[(ccIndex := (self compactClassIndexOf: oop)) = 0
				ifTrue: [(self classHeader: oop) bitAnd: AllButTypeMask]
				ifFalse: [self compactClassAt: ccIndex]]! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 11/10/2014 12:49'!
fetchClassOfNonImm: oop 
	| ccIndex |
	<inline: true>
	^(ccIndex := (self compactClassIndexOf: oop)) = 0
		ifTrue: [(self classHeader: oop) bitAnd: AllButTypeMask]
		ifFalse: [self compactClassAt: ccIndex]! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 2/25/2020 18:18'!
fetchClassTagOfClass: classObj
	<inline: #always>
	^classObj! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 12/17/2014 14:14'!
fetchLong32: fieldIndex ofFloatObject: oop
	"index by word size, and return a pointer as long as the word size"
	<inline: true>
	^self fetchLong32: fieldIndex ofObject: oop! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
fetchLong32: fieldIndex ofObject: oop
	" index by 32-bit units, and return a 32-bit value. Intended to replace fetchWord:ofObject:"

	^ self long32At: oop + self baseHeaderSize + (fieldIndex << 2)! !

!ObjectMemory methodsFor: 'object access' stamp: 'nice 3/29/2016 23:40'!
fetchLong64: longIndex ofObject: oop
	<returnTypeC: #sqLong>
	^self cppIf: BytesPerWord = 8
		ifTrue: [self long64At: oop + self baseHeaderSize + (longIndex << 3)]
		ifFalse:
			["BEWARE OF SIGN EXTENSION ON LEAST SIGNIFICAND LIMB
			'unsigned int) CAST IS THERE TO PREVENT SUCH ERROR"
			self cppIf: VMBIGENDIAN
				ifTrue: [((self long32At: oop + self baseHeaderSize + (longIndex << 3)) asUnsignedLongLong << 32)
					+ (self cCoerceSimple: (self long32At: oop + self baseHeaderSize + (longIndex << 3 + 4)) to: #'unsigned int')]
				ifFalse: [(self cCoerceSimple: (self long32At: oop + self baseHeaderSize + (longIndex << 3)) to: #'unsigned int')
					+ ((self long32At: oop + self baseHeaderSize + (longIndex << 3 + 4)) asUnsignedLongLong << 32)]]! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:06'!
fetchPointer: fieldIndex ofObject: oop
	"index by word size, and return a pointer as long as the word size"
	<api>
	^self longAt: oop + self baseHeaderSize + (fieldIndex << self shiftForWord)! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
fetchShort16: shortIndex ofObject: oop
	^self shortAt: oop + self baseHeaderSize + (shortIndex << 1)! !

!ObjectMemory methodsFor: 'object access' stamp: 'nice 11/14/2016 23:00'!
fetchUnsignedShort16: shortIndex ofObject: oop
	^self cCoerceSimple: (self shortAt: oop + self baseHeaderSize + (shortIndex << 1)) to: #'unsigned short'! !

!ObjectMemory methodsFor: 'finalization' stamp: 'EstebanLorenzano 5/20/2015 12:21'!
finalizeReference: oop 
	"During sweep phase we have encountered a weak reference. Check if its object
	 has gone away (or is about to) and if so, signal a semaphore.  Do *not* inline
	 this in sweepPhase - it is quite an unlikely case to run into a weak reference"
	| weakOop oopGone chunk numFields firstField lastField |
	<inline: false>
	<var: #oop type: #usqInt>
	<var: #weakOop type: #usqInt>
	numFields := self nonWeakFieldsOf: oop. "so nonWeakFieldsOf: may be inlined"
	firstField := self baseHeaderSize + (numFields << self shiftForWord).
	lastField := self lastPointerOf: oop.
	firstField to: lastField by: self wordSize do:
		[:i|
		weakOop := self longAt: oop + i.
		"ar 1/18/2005: Added oop < youngStart test to make sure we're not testing
		objects in non-GCable region. This could lead to a forward reference in
		old space with the oop pointed to not being marked and thus treated as free."
		(weakOop = nilObj or: [(self isIntegerObject: weakOop) or:[weakOop < youngStart]]) ifFalse:
			["Check if the object is being collected. 
			If the weak reference points  
			* backward: check if the weakOops chunk is free
			* forward: check if the weakOoop has been marked by GC"
			weakOop < oop
				ifTrue: [chunk := self chunkFromOop: weakOop.
						oopGone := ((self longAt: chunk) bitAnd: TypeMask) = HeaderTypeFree]
				ifFalse: [oopGone := (self isMarked: weakOop) not].
			oopGone ifTrue: "Store nil in the pointer and signal the  interpreter"
				[self longAt: oop + i put: nilObj.
				self 
					cppIf: PharoVM 
					ifTrue: [ numFields >= 2 ifTrue: [ self weakFinalizerCheck: oop ] ].
				self signalFinalization: oop]]]! !

!ObjectMemory methodsFor: 'object enumeration' stamp: 'eem 5/9/2011 14:40'!
firstAccessibleObject
	"Return the first accessible object in the heap."
	| obj |
	obj := self firstObject.
	[self oop: obj isLessThan: endOfMemory] whileTrue:
		[(self isFreeObject: obj) ifFalse: [^obj].
		 obj := self objectAfter: obj].
	self error: 'heap is empty'.
	^nil! !

!ObjectMemory methodsFor: 'header formats' stamp: 'eem 12/12/2015 14:16'!
firstByteFormat
	<api>
	<cmacro>
	^8! !

!ObjectMemory methodsFor: 'header formats' stamp: 'eem 12/12/2015 14:16'!
firstCompiledMethodFormat
	<api>
	<cmacro>
	^12! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
firstFixedField: oop

	<returnTypeC: #'void *'>
	^ self pointerForOop: oop + self baseHeaderSize! !

!ObjectMemory methodsFor: 'debug support' stamp: 'eem 10/31/2013 09:39'!
firstFixedFieldOfMaybeImmediate: oop
	"for the message send breakpoint; selectors can be immediates."
	<inline: false>
	^(self isImmediate: oop)
		ifTrue: [oop asVoidPointer]
		ifFalse: [self firstFixedField: oop]! !

!ObjectMemory methodsFor: 'object format' stamp: 'eem (const to msg rewrite) 11/13/2014 14:06'!
firstIndexableField: oop
	"NOTE: overridden in various simulator subclasses to add coercion to CArray, so please duplicate any changes"
	| hdr fmt totalLength fixedFields |
	<returnTypeC: #'void *'>
	hdr := self baseHeader: oop.
	fmt := self formatOfHeader: hdr.
	fmt <= self lastPointerFormat ifTrue:
		["pointer; may need to delve into the class format word"
		totalLength := self lengthOf: oop baseHeader: hdr format: fmt.
		fixedFields := self fixedFieldsOf: oop format: fmt length: totalLength.
		^self pointerForOop: oop + self baseHeaderSize + (fixedFields << self shiftForWord)].
	^self pointerForOop: oop + self baseHeaderSize! !

!ObjectMemory methodsFor: 'header formats' stamp: 'eem 12/12/2015 14:16'!
firstLongFormat
	<api>
	<cmacro>
	^6! !

!ObjectMemory methodsFor: 'object enumeration' stamp: ''!
firstObject
	"Return the first object or free chunk in the heap."

	^ self oopFromChunk: self startOfMemory! !

!ObjectMemory methodsFor: 'image save/restore' stamp: 'eem 10/9/2013 21:10'!
firstSegmentSize: firstSegSize
	"Ignored; this is for Spur's segmented image"! !

!ObjectMemory methodsFor: 'header formats' stamp: 'eem 9/9/2013 15:07'!
firstStringyFakeFormat
	"A fake format for the interpreter used to mark indexable strings in
	 the interpreter's at cache.  This is larger than any format."
	^16! !

!ObjectMemory methodsFor: 'object format' stamp: 'eem 11/10/2014 12:49'!
fixedFieldsOf: oop format: fmt length: wordLength
"
	NOTE: This code supports the backward-compatible extension to 8 bits of instSize.
	When we revise the image format, it should become...
	^ (classFormat >> 2 bitAnd: 16rFF) - 1
"
	| class classFormat |
	<inline: true>
	((fmt > self lastPointerFormat) or: [fmt = self arrayFormat]) ifTrue: [^0].  "indexable fields only"
	fmt < 2 ifTrue: [^wordLength].  "fixed fields only (zero or more)"
	
	"fmt = 3 or 4: mixture of fixed and indexable fields, so must look at class format word"
	class := self fetchClassOfNonImm: oop.
	classFormat := self formatOfClass: class.
	^(classFormat >> 11 bitAnd: 16rC0) + (classFormat >> 2 bitAnd: 16r3F) - 1! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 11/21/2014 16:25'!
floatObjectOf: aFloat
	| newFloatObj |
	<inline: false>
	<var: #aFloat type: #double>
	newFloatObj := self instantiateSmallClass: (self splObj: ClassFloat) sizeInBytes: (self sizeof: #double)+self baseHeaderSize.
	self storeFloatAt: newFloatObj + self baseHeaderSize from: aFloat.
	^newFloatObj! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 11/14/2019 13:43'!
floatValueOf: oop
	"Answer the C double precision floating point value of the argument,
	 or fail if it is not a Float, and answer 0.
	 Note: May be called by translated primitive code."
	<api>
	<returnTypeC: #double>
	| isFloat result |
	<var: #result type: #double>
	isFloat := self isFloatInstance: oop.
	isFloat ifTrue:
		[self cCode: '' inSmalltalk: [result := Float new: 2].
		 self fetchFloatAt: oop + self baseHeaderSize into: result.
		 ^result].
	self primitiveFail.
	^0.0! !

!ObjectMemory methodsFor: 'forward compatibility' stamp: 'eem 7/11/2014 12:35'!
followField: fieldIndex ofObject: anObject
	"Spur compatibility; in V3 this is a synonym for fetchPointer:ofObject:"
	<inline: true>
	^self fetchPointer: fieldIndex ofObject: anObject! !

!ObjectMemory methodsFor: 'spur compatibility' stamp: 'eem 10/9/2013 16:47'!
followForwarded: objOop
	<inline: false>
	self shouldNotImplement.
	^0! !

!ObjectMemory methodsFor: 'spur compatibility' stamp: 'eem 10/25/2013 08:54'!
followForwardedObjectFields: objOop toDepth: depth
	<inline: false>
	self shouldNotImplement! !

!ObjectMemory methodsFor: 'spur compatibility' stamp: 'eem 10/14/2014 11:58'!
followMaybeForwarded: objOop
	<inline: true>
	^objOop! !

!ObjectMemory methodsFor: 'forward compatibility' stamp: 'eem 1/22/2014 16:17'!
followObjField: fieldIndex ofObject: anObject
	<inline: true>
	^self fetchPointer: fieldIndex ofObject: anObject! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
forceTenureFlag
	^forceTenureFlag! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
forceTenureFlag: aValue
	^forceTenureFlag := aValue! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 9/10/2013 11:02'!
formatOf: oop
"       0      no fields
        1      fixed fields only (all containing pointers)
        2      indexable fields only (all containing pointers)
        3      both fixed and indexable fields (all containing pointers)
        4      both fixed and indexable weak fields (all containing pointers).

        5      unused (reserved for ephemerons?)
        6      indexable word fields only (no pointers)
        7      indexable long (64-bit) fields (only in 64-bit images)
 
    8-11      indexable byte fields only (no pointers) (low 2 bits are low 2 bits of size)
   12-15     compiled methods:
                   # of literal oops specified in method header,
                   followed by indexable bytes (same interpretation of low 2 bits as above)
"
	<inline: true>
	^((self baseHeader: oop) >> self instFormatFieldLSB) bitAnd: 16rF! !

!ObjectMemory methodsFor: 'object format' stamp: 'eem 10/13/2009 05:03'!
formatOfClass: classPointer
	"**should be in-lined**"
	"Note that, in Smalltalk, the instSpec will be equal to the inst spec
	part of the base header of an instance (without hdr type) shifted left 1.
	In this way, apart from the smallInt bit, the bits
	are just where you want them for the first header word."
	"Callers (must) expect low 2 bits (header type) to be zero!!"
	<api>
	<inline: true>
	^(self fetchPointer: InstanceSpecificationIndex ofObject: classPointer) - 1! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 9/10/2013 11:02'!
formatOfHeader: header
"       0      no fields
        1      fixed fields only (all containing pointers)
        2      indexable fields only (all containing pointers)
        3      both fixed and indexable fields (all containing pointers)
        4      both fixed and indexable weak fields (all containing pointers).

        5      unused
        6      indexable word fields only (no pointers)
        7      indexable long (64-bit) fields (only in 64-bit images)
 
    8-11      indexable byte fields only (no pointers) (low 2 bits are low 2 bits of size)
   12-15     compiled methods:
                   # of literal oops specified in method header,
                   followed by indexable bytes (same interpretation of low 2 bits as above)
"
	<inline: true>
	^header >> self instFormatFieldLSB bitAnd: 16rF! !

!ObjectMemory methodsFor: 'image segment in/out' stamp: 'di 3/24/1999 16:10'!
forward: oop to: newOop savingOopAt: oopPtr andHeaderAt: hdrPtr

	"Make a new entry in the table of saved oops."
	self longAt: oopPtr put: oop.					"Save the oop"
	self longAt: hdrPtr put: (self longAt: oop).	"Save the old header word"

	"Put a forwarding pointer in the old object, flagged with forbidden header type"
	self longAt: oop put: newOop + HeaderTypeFree! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 8/26/2009 11:51'!
forwardingPointerOf: forwardedObj
	"Answer the pointer to the given forwardedOop's forwarding block."
	<inline: true>
	^((self longAt: forwardedObj) bitAnd: AllButMarkBitAndTypeMask) << 1! !

!ObjectMemory methodsFor: 'become' stamp: 'eem 2/20/2013 13:16'!
freeObject: obj
	| objHeader |
	objHeader := self baseHeader: obj.
	self freeObject: obj header: objHeader! !

!ObjectMemory methodsFor: 'become' stamp: 'eem 2/20/2013 13:12'!
freeObject: obj header: objHeader
	| objHeaderBytes objHeaderType objSize |
	(self isYoungRootHeader: objHeader) ifTrue:
		[self removeYoungRoot: obj].
	objHeaderType := objHeader bitAnd: TypeMask.
	objHeaderBytes := headerTypeBytes at: objHeaderType.
	(objHeaderType bitAnd: 1) = 1 "HeaderTypeClass or HeaderTypeShort"
		ifTrue: [objSize := objHeader bitAnd: SizeMask]
		ifFalse:
			[objHeaderType = HeaderTypeFree
				ifTrue: [^nil]. "already free"
			objSize := (self sizeHeader: obj) bitAnd: LongSizeMask].
	self assert: (objSize + objHeaderBytes bitAnd: AllButTypeMask) = (objSize + objHeaderBytes).
	self longAt: obj - objHeaderBytes
		put: ((objSize + objHeaderBytes) bitOr: HeaderTypeFree)! !

!ObjectMemory methodsFor: 'garbage collection' stamp: 'tpr 3/24/2004 22:02'!
fullCompaction
	"Move all accessible objects down to leave one big free chunk 
	at the end of memory."
	"Assume: Incremental GC has just been done to maximimize 
	forwarding table space."
	"need not move objects below the first free chunk"
	| sz |
	compStart := self lowestFreeAfter: self startOfMemory.
	compStart = freeBlock
		ifTrue: ["memory is already compact; only free chunk is at the end "
			^ self initializeMemoryFirstFree: freeBlock].
	(sz := self fwdTableSize: 8) < totalObjectCount
		ifTrue: ["Try to grow OM to make a single pass full GC"
			self growObjectMemory: totalObjectCount - sz + 10000 * 8].
	"work up through memory until all free space is at the end"
	[compStart < freeBlock]
		whileTrue: ["free chunk returned by incCompBody becomes start of next compaction"
			compStart := self incCompBody]! !

!ObjectMemory methodsFor: 'garbage collection' stamp: 'eem 2/7/2015 09:13'!
fullGC
	"Do a mark/sweep garbage collection of the entire object memory. Free inaccessible objects but do not move them."

	<inline: false>
	DoAssertionChecks ifTrue:
		[self reverseDisplayFrom: 0 to: 7.
		 self clearLeakMapAndMapAccessibleObjects.
		 self checkHeapIntegrity].
	self preGCAction: GCModeFull.
	gcStartUsecs := self ioUTCMicrosecondsNow.
	statSweepCount := statMarkCount := statMkFwdCount := statCompMoveCount := 0.
	self clearRootsTable.
	self initWeakTableForIncrementalGC: false.
	youngStart := self startOfMemory.  "process all of memory"
	self markPhase.
	"Sweep phase returns the number of survivors.
	Use the up-to-date version instead the one from startup."
	totalObjectCount := self sweepPhase.
	self fullCompaction.
	allocationCount := 0.
	statFullGCs := statFullGCs + 1.
	statGCEndTime := self ioMicroMSecs.
	statFullGCUsecs := statFullGCUsecs + (self ioUTCMicrosecondsNow - gcStartUsecs).
	self capturePendingFinalizationSignals.

	youngStart := freeBlock.  "reset the young object boundary"
	self attemptToShrink.
	self postGCAction: GCModeFull.
	DoAssertionChecks ifTrue:
		[self clearLeakMapAndMapAccessibleObjects.
		 self checkHeapIntegrity.
		 self reverseDisplayFrom: 0 to: 7]! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'di 1/12/1999 14:09'!
fwdBlockGet: blkSize
	"Return the address of a two- or four-word forwarding block or nil if no more entries are available."

	fwdTableNext := fwdTableNext + blkSize.
	fwdTableNext <= fwdTableLast
		ifTrue: [ ^ fwdTableNext ]
		ifFalse: [ ^ nil ].  "no more forwarding blocks available"! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 4/9/2009 16:49'!
fwdBlockValid: addr 
	"Answer if the given address is a valid forward table entry."
	^(self oop: addr isGreaterThan: endOfMemory)
	   and: [(self oop: addr isLessThanOrEqualTo: fwdTableNext)
	   and: [(addr bitAnd: 3) = 0]]! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
fwdTableInit: blkSize
	"Set the limits for a table of two- or three-word forwarding blocks above the last used oop. The pointer fwdTableNext moves up to fwdTableLast. Used for compaction of memory and become-ing objects. Returns the number of forwarding blocks available."

	| |
	<inline: false>
	"set endOfMemory to just after a minimum-sized free block"
	self setSizeOfFree: freeBlock to: self baseHeaderSize.
	self setEndOfMemory: freeBlock + self baseHeaderSize.

	"make a fake free chunk at endOfMemory for use as a sentinal in memory scans"
	self setSizeOfFree: endOfMemory to: self baseHeaderSize.

	"use all memory free between freeBlock and memoryLimit for forwarding table"
	"Note: Forward blocks must be quadword aligned."
	fwdTableNext := (endOfMemory + self baseHeaderSize + 7) bitAnd: WordMask-7.
	self flag: #Dan.  "Above line does not do what it says (quadword is 16 or 32 bytes)"

	fwdTableLast := memoryLimit - blkSize.  "last forwarding table entry"

	"return the number of forwarding blocks available"
	^ (fwdTableLast - fwdTableNext) // blkSize  "round down"! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
fwdTableSize: blkSize
	"Estimate the number of forwarding blocks available for compaction"
	| eom fwdFirst fwdLast |
	<inline: false>

	eom := freeBlock + self baseHeaderSize.
	"use all memory free between freeBlock and memoryLimit for forwarding table"

	"Note: Forward blocks must be quadword aligned."
	fwdFirst := (eom + self baseHeaderSize + 7) bitAnd: WordMask-7.
	self flag: #Dan.  "Above line does not do what it says (quadword is 16 or 32 bytes)"

	fwdLast := memoryLimit - blkSize.  "last forwarding table entry"

	"return the number of forwarding blocks available"
	^ (fwdLast - fwdFirst) // blkSize  "round down"! !

!ObjectMemory methodsFor: 'image save/restore' stamp: 'eem 12/11/2013 09:36'!
garbageCollectForSnapshot
	self incrementalGC. "maximimize space for forwarding table"
	self fullGC.
	self clearRootsTable! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
gcBiasToGrow
	^gcBiasToGrow! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
gcBiasToGrow: aValue
	^gcBiasToGrow := aValue! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
gcBiasToGrowGCLimit
	^gcBiasToGrowGCLimit! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
gcBiasToGrowGCLimit: aValue
	^gcBiasToGrowGCLimit := aValue! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
gcBiasToGrowThreshold
	^gcBiasToGrowThreshold! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
gcBiasToGrowThreshold: aValue
	^gcBiasToGrowThreshold := aValue! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
gcStartUsecs
	^gcStartUsecs! !

!ObjectMemory methodsFor: 'contexts' stamp: 'eem 9/3/2013 15:38'!
goodContextSize: oop
	| numSlots |
	numSlots := self numSlotsOf: oop.
	^numSlots = SmallContextSlots or: [numSlots = LargeContextSlots]! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
growHeadroom
	^growHeadroom! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
growHeadroom: aValue
	^growHeadroom := aValue! !

!ObjectMemory methodsFor: 'allocation' stamp: 'eem 4/4/2009 20:08'!
growObjectMemory: delta 
	"Attempt to grow the object memory by the given delta 
	amount "
	| limit |
	statGrowMemory := statGrowMemory + 1.
	limit := self sqGrowMemory: memoryLimit By: delta.
	limit = memoryLimit ifFalse:
		[self setMemoryLimit: limit - 24. "remove a tad for safety"
		 self initializeMemoryFirstFree: freeBlock]! !

!ObjectMemory methodsFor: 'allocation' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
growToAccomodateContainerWithNumSlots: numSlots
	"Grow memory to accomodate a container (an Array) with numSlots.
	 Grow by at least the growHeadroom.  Supports allInstancesOf: and allObjects."
	| delta |
	delta := (headerTypeBytes at: HeaderTypeSizeAndClass) / self wordSize
			+ 1
			+ numSlots
			* self bytesPerOop.
	self growObjectMemory: (growHeadroom max: delta)! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 8/25/2014 14:45'!
hasForwardingBlock: objOop
	^self isMarked: objOop! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 9/14/2015 10:35'!
hasIdentityHash: objOop
	"Objects gain their identityHash on instantiation in V3."
	<inline: true>
	^true! !

!ObjectMemory methodsFor: 'object format' stamp: 'eem 2/12/2015 17:33'!
hasSixtyFourBitImmediates
	"The V3 64-bit memory manager has 64-bit oops, but its SmallIntegers only have
	 31 bits of precision.  The Spur 64-bit memory manager has 61-bit immediates."
	^false! !

!ObjectMemory methodsFor: 'api characterization' stamp: 'eem 9/9/2013 11:01'!
hasSpurMemoryManagerAPI
	^false! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 8/31/2009 13:37'!
hashBitsOf: oop

	^ ((self baseHeader: oop) >> HashBitsOffset) bitAnd: HashMaskUnshifted! !

!ObjectMemory methodsFor: 'method access' stamp: 'eem 6/7/2016 14:58'!
headerIndicatesAlternateBytecodeSet: methodHeader
	"A negative header selects the alternate bytecode set."
	<api>
	<inline: true>
	^methodHeader signedIntFromLong < 0! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 6/14/2011 17:58'!
headerType: oop

	^(self baseHeader: oop) bitAnd: TypeMask! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 6/14/2011 17:56'!
headerTypeOfHeader: header

	^header bitAnd: TypeMask! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
headerWhileForwardingOf: oop
	"Answer the header of the argument even though
	 it may have its header word in a forwarding block."
	| header fwdBlock |
	<inline: true>
	header := self longAt: oop.
	(header bitAnd: MarkBit) ~= 0 ifTrue:
		["oop is forwarded; get its real header from its forwarding table entry"
		 fwdBlock := (header bitAnd: AllButMarkBitAndTypeMask) << 1.
		 self assert: (self fwdBlockValid: fwdBlock).
		 header := self longAt: fwdBlock + self wordSize].
	^header! !

!ObjectMemory methodsFor: 'image save/restore' stamp: 'eem 3/10/2016 12:40'!
imageFormatVersion
	"Return a magic constant that changes when the image format changes.
	 Since the image reading code uses this to detect byte ordering, one
	 must avoid version numbers that are invariant under byte reversal."
	^self wordSize = 4 ifTrue: [6505] ifFalse: [68003]! !

!ObjectMemory methodsFor: 'image segment in/out' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
imageSegmentVersion
	| wholeWord |
	"a more complex version that tells both the word reversal and the endianness of the machine it came from.  Low half of word is 6502.  Top byte is top byte of #doesNotUnderstand: on this machine. ($d on the Mac or $s on the PC)"

	wholeWord := self longAt: (self splObj: SelectorDoesNotUnderstand) + self baseHeaderSize.
		"first data word, 'does' "
	^ self imageFormatVersion bitOr: (wholeWord bitAnd: 16rFF000000)! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
incCompBody
	"Move objects to consolidate free space into one big chunk. Return the newly created free chunk."

	| bytesFreed |
	<inline: false>
	"reserve memory for forwarding table"
	self fwdTableInit: self wordSize*2.  "Two-word blocks"

	"assign new oop locations, reverse their headers, and initialize forwarding blocks"
	bytesFreed := self incCompMakeFwd.

	"update pointers to point at new oops"
	self mapPointersInObjectsFrom: youngStart to: endOfMemory.

	"move the objects and restore their original headers; return the new free chunk"
	^ self incCompMove: bytesFreed! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
incCompMakeFwd
	"Create and initialize forwarding blocks for all non-free objects  
	following compStart. If the supply of forwarding blocks is exhausted,  
	set compEnd to the first chunk above the area to be 
	compacted; otherwise, set it to endOfMemory. Return the number of 
	bytes to be freed."
	| bytesFreed oop fwdBlock newOop |
	<inline: false>
	bytesFreed := 0.
	oop := self oopFromChunk: compStart.
	[self oop: oop isLessThan: endOfMemory]
		whileTrue: [
				statMkFwdCount := statMkFwdCount + 1.
				(self isFreeObject: oop)
				ifTrue: [bytesFreed := bytesFreed + (self sizeOfFree: oop)]
				ifFalse: ["create a forwarding block for oop"
					fwdBlock := self fwdBlockGet: self wordSize*2.
					"Two-word block"
					fwdBlock = nil
						ifTrue: ["stop; we have used all available forwarding blocks"
							compEnd := self chunkFromOop: oop.
							^ bytesFreed].
					newOop := oop - bytesFreed.
					self initForwardBlock: fwdBlock mapping: oop to: newOop withBackPtr: false].
			oop := self objectAfterWhileForwarding: oop].
	compEnd := endOfMemory.
	^ bytesFreed! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
incCompMove: bytesFreed 
	"Move all non-free objects between compStart and compEnd to their new  
	locations, restoring their headers in the process. Create a new free  
	block at the end of memory. Return the newly created free chunk. "
	"Note: The free block used by the allocator always must be the last free  
	block in memory. It may take several compaction passes to make all  
	free space bubble up to the end of memory."
	| oop next fwdBlock newOop header bytesToMove firstWord lastWord newFreeChunk sz target |
	<inline: false>
	<var: #firstWord type: 'usqInt'>
	<var: #lastWord type: 'usqInt'>
	<var: #w type: 'usqInt'>
	newOop := nil.
	oop := self oopFromChunk: compStart.
	[self oop: oop isLessThan: compEnd] whileTrue:
		[statCompMoveCount := statCompMoveCount + 1.
		next := self objectAfterWhileForwarding: oop.
		(self isFreeObject: oop) ifFalse:
			["a moving object; unwind its forwarding block"
			fwdBlock := self forwardingPointerOf: oop.
			self assert: (self fwdBlockValid: fwdBlock).
			newOop := self longAt: fwdBlock.
			header := self longAt: fwdBlock + self wordSize.
			self longAt: oop put: header. "restore the original header"
			bytesToMove := oop - newOop. "move the oop (including any extra header words) "
			sz := self sizeBitsOf: oop.
			firstWord := oop - (self extraHeaderBytes: oop).
			lastWord := oop + sz - self baseHeaderSize.
			target := firstWord - bytesToMove.
			firstWord to: lastWord by: self wordSize do:
				[:w | 
				self longAt: target put: (self longAt: w).
				target := target + self wordSize]].
		oop := next].
	newOop = nil
		ifTrue: ["no objects moved"
			oop := self oopFromChunk: compStart.
			((self isFreeObject: oop) and: [(self objectAfter: oop) = (self oopFromChunk: compEnd)])
				ifTrue: [newFreeChunk := oop]
				ifFalse: [newFreeChunk := freeBlock]]
		ifFalse: ["initialize the newly freed memory chunk"
			"newOop is the last object moved; free chunk starts 
			right after it"
			newFreeChunk := newOop + (self sizeBitsOf: newOop).
			self setSizeOfFree: newFreeChunk to: bytesFreed].
	self assert: (self objectAfter: newFreeChunk) = (self oopFromChunk: compEnd).
	(self objectAfter: newFreeChunk) = endOfMemory
		ifTrue: [self initializeMemoryFirstFree: newFreeChunk]
		ifFalse: ["newFreeChunk is not at end of memory; re-install freeBlock "
			self initializeMemoryFirstFree: freeBlock].
	^ newFreeChunk! !

!ObjectMemory methodsFor: 'garbage collection' stamp: 'tpr 3/24/2004 22:03'!
incrementalCompaction
	"Move objects down to make one big free chunk. Compact the 
	last N objects (where N = number of forwarding table 
	entries) of the young object area."
	"Assume: compStart was set during the sweep phase"
	compStart = freeBlock
		ifTrue: ["Note: If compStart = freeBlock then either the young 
			space is already compact  or there are enough forwarding table entries to do a 
			one-pass incr. compaction."
			self initializeMemoryFirstFree: freeBlock]
		ifFalse: [self incCompBody]! !

!ObjectMemory methodsFor: 'garbage collection' stamp: 'eem 2/7/2015 10:07'!
incrementalGC
	"Do a mark/sweep garbage collection of just the young object 
	area of object memory (i.e., objects above youngStart), using 
	the root table to identify objects containing pointers to 
	young objects from the old object area."
	| survivorCount weDidGrow |
	<inline: false>
	rootTableOverflowed ifTrue:
		["root table overflow; cannot do an incremental GC because some roots are missing.
		 (this should be very rare)"
		 statRootTableOverflows := statRootTableOverflows + 1.
		 ^self fullGC].

	DoAssertionChecks ifTrue:
		[self reverseDisplayFrom: 8 to: 15.
		 self checkHeapIntegrity.
		 self checkInterpreterIntegrity.
		 self validate].

	self preGCAction: GCModeNewSpace.
	"incremental GC and compaction"

	gcStartUsecs := self ioUTCMicrosecondsNow.
	statSweepCount := statMarkCount := statMkFwdCount := statCompMoveCount := 0.
	self initWeakTableForIncrementalGC: true.
	self markPhase.
	self assert: weakRootCount <= WeakRootTableSize.
	1 to: weakRootCount do:[:i| self finalizeReference: (weakRoots at: i)].
	survivorCount := self sweepPhase.
	self incrementalCompaction.
	statAllocationCount := allocationCount.
	allocationCount := 0.
	statIncrGCs := statIncrGCs + 1.
	statGCEndTime := self ioMicroMSecs.
	statIGCDeltaUsecs := self ioUTCMicrosecondsNow - gcStartUsecs.
	statIncrGCUsecs := statIncrGCUsecs + statIGCDeltaUsecs.
	self capturePendingFinalizationSignals.

	self forceInterruptCheck. "Force an an interrupt check ASAP.We could choose to be clever here and only do this under certain time conditions. Keep it simple for now"
	
	statRootTableCount  := rootTableCount.
	statSurvivorCount := survivorCount.
	weDidGrow := false.
	(((survivorCount > tenuringThreshold)
			or: [rootTableCount >= RootTableRedZone])
			or: [forceTenureFlag == true])
		ifTrue: ["move up the young space boundary if 
			* there are too many survivors: 
			this limits the number of objects that must be 
			processed on future incremental GC's 
			* we're about to overflow the roots table 
			this limits the number of full GCs that may be caused 
			by root table overflows in the near future"
			forceTenureFlag := false.
			statTenures := statTenures + 1.
			self clearRootsTable.
			(((self sizeOfFree: freeBlock) < growHeadroom) and: 
				[gcBiasToGrow > 0]) 
				ifTrue: [self biasToGrow.
						weDidGrow := true].
			youngStart := freeBlock].
	self attemptToShrink.
	self postGCAction: GCModeNewSpace.
	DoAssertionChecks ifTrue:
		[self validate.
		 self checkHeapIntegrity.
		 self checkInterpreterIntegrity.
		 self reverseDisplayFrom: 8 to: 15].
	weDidGrow ifTrue: [self biasToGrowCheckGCLimit]! !

!ObjectMemory methodsFor: 'header formats' stamp: 'eem 12/12/2015 14:16'!
indexablePointersFormat
	<api>
	<cmacro>
	^3! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 11/10/2014 12:49'!
initForwardBlock: fwdBlock mapping: oop to: newOop withBackPtr: backFlag 
	"Initialize the given forwarding block to map oop to newOop, 
	and replace oop's header with a pointer to the fowarding 
	block. "
	"Details: The mark bit is used to indicate that an oop is 
	forwarded. When an oop is forwarded, its header (minus the 
	mark bit) contains the address of its forwarding block. (The 
	forwarding block address is actually shifted right by one bit 
	so that its top-most bit does not conflict with the header's 
	mark bit; since fowarding blocks are stored on word 
	boundaries, the low two bits of the address are always zero.) 
	The first word of the forwarding block is the new oop; the 
	second word is the oop's orginal header. In the case of a 
	forward become, a four-word block is used, with the third 
	field being a backpointer to the old oop (for header fixup), 
	and the fourth word is unused. The type bits of the 
	forwarding header are the same as those of the original 
	header. "
	| originalHeader originalHeaderType |
	<inline: true>
	originalHeader := self longAt: oop.
	self assert: fwdBlock ~= nil. "ran out of forwarding blocks in become"
	self assert: (originalHeader bitAnd: MarkBit) = 0. "'object already has a forwarding table entry"
	originalHeaderType := originalHeader bitAnd: TypeMask.
	self longAt: fwdBlock put: newOop.
	self longAt: fwdBlock + self wordSize put: originalHeader.
	backFlag ifTrue: [self longAt: fwdBlock + (self wordSize*2) put: oop].
	self longAt: oop put: (fwdBlock >> 1 bitOr: (MarkBit bitOr: originalHeaderType))! !

!ObjectMemory methodsFor: 'gc -- mark and sweep' stamp: 'eem 2/28/2012 14:43'!
initWeakTableForIncrementalGC: trueIfIncrementalGC
	"The weakRoots table is only used for incrementalGC.
	 Enable it by setting weakRootCount to 0.
	 Disable it by making it negative."
	weakRootCount := trueIfIncrementalGC ifTrue: [0] ifFalse: [-1]! !

!ObjectMemory methodsFor: 'object enumeration' stamp: 'eem 9/18/2013 14:27'!
initialInstanceOf: classPointer 
	"Support for instance enumeration. Return the first instance 
	of the given class, or nilObj if it has no instances."
	| thisObj thisClass |
	thisObj := self firstAccessibleObject.
	[thisObj = nil]
		whileFalse: [thisClass := self fetchClassOf: thisObj.
			thisClass = classPointer ifTrue: [^ thisObj].
			thisObj := self accessibleObjectAfter: thisObj].
	^nil! !

!ObjectMemory methodsFor: 'initialization' stamp: 'eem 2/2/2018 09:51'!
initialize
	"Initialize ObjectMemory when simulating the VM inside Smalltalk."
	super initialize.
	headerTypeBytes := CArrayAccessor on: HeaderTypeExtraBytes.
	rootTable := Array new: RootTableSize.
	weakRoots := Array new: WeakRootTableSize.
	remapBuffer := Array new: RemapBufferSize! !

!ObjectMemory methodsFor: 'initialization' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
initializeMemoryFirstFree: firstFree 
	"Initialize endOfMemory to the top of oop storage space, reserving some space for forwarding blocks, and create the freeBlock from which space is allocated. Also create a fake free chunk at endOfMemory to act as a sentinal for memory scans. "
	"Note: The amount of space reserved for forwarding blocks should be chosen to ensure that incremental compactions can usually be done in a single pass. However, there should be enough forwarding blocks so a full compaction can be done in a reasonable number of passes, say ten. (A full compaction requires N object-moving passes, where N = number of non-garbage objects / number of forwarding blocks). 
	di 11/18/2000 Re totalObjectCount: Provide a margin of one byte per object to be used for forwarding pointers at GC time. Since fwd blocks are 8 bytes, this means an absolute worst case of 8 passes to compact memory. In most cases it will be adequate to do compaction in a single pass. "
	| fwdBlockBytes |
	"reserve space for forwarding blocks"
	fwdBlockBytes := totalObjectCount bitAnd: WordMask - self wordSize + 1.
	(self oop: memoryLimit - fwdBlockBytes isGreaterThanOrEqualTo: firstFree + self baseHeaderSize)
		ifFalse: ["reserve enough space for a minimal free block of BaseHeaderSize bytes"
			fwdBlockBytes := memoryLimit - (firstFree + self baseHeaderSize)].

	"set endOfMemory and initialize freeBlock"
	self setEndOfMemory: memoryLimit - fwdBlockBytes.
	freeBlock := firstFree.
	self setSizeOfFree: freeBlock to: endOfMemory - firstFree. "bytes available for oops"

	"make a fake free chunk at endOfMemory for use as a sentinel in memory scans"
	self setSizeOfFree: endOfMemory to: self baseHeaderSize.
	DoAssertionChecks
		ifTrue: [(freeBlock < endOfMemory and: [endOfMemory < memoryLimit])
				ifFalse: [self error: 'error in free space computation'].
			(self oopFromChunk: endOfMemory) = endOfMemory
				ifFalse: [self error: 'header format must have changed'].
			(self objectAfter: freeBlock) = endOfMemory
				ifFalse: [self error: 'free block not properly initialized']]! !

!ObjectMemory methodsFor: 'initialization' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
initializeObjectMemory: bytesToShift
	"Initialize object memory variables at startup time. Assume endOfMemory is initially set (by the image-reading code) to the end of the last object in the image. Initialization redefines endOfMemory to be the end of the object allocation area based on the total available memory, but reserving some space for forwarding blocks."
	"Assume: image reader initializes the following variables:
		memory
		endOfMemory
		memoryLimit
		specialObjectsOop
		lastHash
	"
	"di 11/18/2000 fix slow full GC"
	<inline: false>
	"Catch mis-initializations leading to bad translations to C"
	self assert: self baseHeaderSize = self wordSize.

	"set the start of the young object space"
	youngStart := endOfMemory.

	"image may be at a different address; adjust oops for new location"
	totalObjectCount := self adjustAllOopsBy: bytesToShift.

	self initializeMemoryFirstFree: endOfMemory. "initializes endOfMemory, freeBlock"

	specialObjectsOop := specialObjectsOop + bytesToShift.

	"heavily used special objects"
	nilObj	:= self splObj: NilObject.
	falseObj	:= self splObj: FalseObject.
	trueObj	:= self splObj: TrueObject.

	rootTableCount := 0.
	rootTableOverflowed := false.
	freeContexts := NilContext.
	freeLargeContexts := NilContext.
	allocationCount := 0.
	lowSpaceThreshold := 0.
	signalLowSpace := false.
	compStart := 0.
	compEnd := 0.
	fwdTableNext := 0.
	fwdTableLast := 0.
	remapBufferCount := 0.
	allocationsBetweenGCs := 4000.  "do incremental GC after this many allocations"
	tenuringThreshold := 2000.  "tenure all suriving objects if count is over this threshold"
	growHeadroom := 4*1024*1024. "four megabyte of headroom when growing"
	shrinkThreshold := 8*1024*1024. "eight megabyte of free space before shrinking"

	"garbage collection statistics"
	statFullGCs := 0.
	statFullGCUsecs := 0.
	statIncrGCs := 0.
	statIncrGCUsecs := 0.
	statTenures := 0.
	statRootTableOverflows := 0.
	statGrowMemory := 0.
	statShrinkMemory := 0.
	forceTenureFlag := 0.
	gcBiasToGrow := 0.
	gcBiasToGrowGCLimit := 0.
	extraRootCount := 0.
	gcStartUsecs := 0! !

!ObjectMemory methodsFor: 'object format' stamp: 'eem 8/7/2014 04:35'!
instSpecOfClass: classPointer
	"This field in a class's format inst var corresponds to the 4-bit format field stored in every object header"

	^self formatOfHeader: (self formatOfClass: classPointer)! !

!ObjectMemory methodsFor: 'primitive support' stamp: 'eem 9/4/2013 17:22'!
instanceAfter: objOop
	"Answer the next instance of objOop's class in the canonical enumeration order,
	 if it exists, otherwise answer nil."
	| classPointer subsequentObject |
	classPointer := self fetchClassOfNonImm: objOop.
	subsequentObject := self objectAfter: objOop.
	[self oop: subsequentObject isLessThan: self freeStart] whileTrue:
		[((self isFreeObject: subsequentObject) not
		  and: [(self fetchClassOfNonImm: subsequentObject) = classPointer]) ifTrue:
			[^subsequentObject].
		 subsequentObject := self objectAfter: subsequentObject].
	^nil! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
instanceSizeOf: classObj
	<api>
	"Answer the number of slots in a class.  For example the instanceSizeOf: 
	 ClassPoint is 2, for the x & y slots. The instance size of non-pointer classes is 0."
	| classHdr sizeHiBits byteSize |
	self assert: (self addressCouldBeObj: classObj).

	classHdr := self formatOfClass: classObj. "Low 2 bits are 0"

	"Compute the size of instances of the class (used for fixed field classes only)"
	sizeHiBits := classHdr >> 9 bitAnd: 16r300.
	byteSize := (classHdr bitAnd: SizeMask) + sizeHiBits. "size in bytes -- low 2 bits are 0"
	^byteSize - self baseHeaderSize / self wordSize! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 8/22/2017 14:13'!
instantiateClass: classPointer indexableSize: size
	"NOTE: This method supports the backward-compatible split instSize field of the 
	class format word. The sizeHiBits will go away and other shifts change by 2 
	when the split fields get merged in an (incompatible) image change."
	<api>
	| hash header1 cClass byteSize format binc header3 hdrSize sizeHiBits bm1 classFormat |
	<inline: false>
	self assert: size >= 0. "'cannot have a negative indexable field count"
	hash := self newObjectHash.
	classFormat := self formatOfClass: classPointer.
	"Low 2 bits are 0"
	header1 := (classFormat bitAnd: 16r1FF00) bitOr: (hash bitAnd: HashMaskUnshifted) << HashBitsOffset.
	sizeHiBits := (classFormat bitAnd: 16r60000) >> 9.
	cClass := header1 bitAnd: CompactClassMask. "compact class field from format word"
	byteSize := (classFormat bitAnd: SizeMask + Size4Bit) + sizeHiBits.
		"size in bytes -- low 2 bits are 0"
	"Note this byteSize comes from the format word of the class which is pre-shifted
		to 4 bytes per field.  Need another shift for 8 bytes per word..."
	byteSize := byteSize << (self shiftForWord-2).
	format := self formatOfHeader: classFormat.
	self flag: #sizeLowBits.
	format < self firstByteFormat
		ifTrue:
			[format = self firstLongFormat
				ifTrue: "long32 bitmaps"
					[bm1 := self wordSize-1.
					byteSize := byteSize + (size * 4) + bm1 bitAnd: LongSizeMask. "round up"
					binc := bm1 - ((size * 4) + bm1 bitAnd: bm1). "odd bytes"
					"extra low bit (4) for 64-bit VM goes in 4-bit (betw hdr bits and sizeBits)"
					header1 := header1 bitOr: (binc bitAnd: 4)]
				ifFalse: "Arrays and 64-bit bitmaps"
					[byteSize := byteSize + (size * self wordSize)]]
		ifFalse:
			["Strings and Methods"
			bm1 := self wordSize-1.
			byteSize := byteSize + size + bm1 bitAnd: LongSizeMask. "round up"
			binc := bm1 - (size + bm1 bitAnd: bm1). "odd bytes"
			"low bits of byte size go in format field"
			header1 := header1 bitOr: (binc bitAnd: 3) << self instFormatFieldLSB.
			"extra low bit (4) for 64-bit VM goes in 4-bit (betw hdr bits and sizeBits)"
			header1 := header1 bitOr: (binc bitAnd: 4)].
	byteSize > 255 "requires size header word/full header"
		ifTrue: [header3 := byteSize. hdrSize := 3]
		ifFalse: [header1 := header1 bitOr: byteSize. hdrSize := cClass = 0 ifTrue: [2] ifFalse: [1]].
	^self allocate: byteSize headerSize: hdrSize h1: header1 h2: classPointer h3: header3 doFill: true format: format! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 8/22/2017 14:18'!
instantiateContext: classPointer sizeInBytes: sizeInBytes 
	"This version of instantiateClass assumes that the total object 
	size is under 256 bytes, the limit for objects with only one or 
	two header words. Note that the size is specified in bytes 
	and should include four bytes for the base header word."
	| hash header1 hdrSize |
	hash := self newObjectHash.
	header1 := (hash bitAnd: HashMaskUnshifted) << HashBitsOffset bitOr: (self formatOfClass: classPointer).
	(header1 bitAnd: CompactClassMask) > 0 "are contexts compact?"
		ifTrue: [hdrSize := 1]
		ifFalse: [hdrSize := 2].
	sizeInBytes <= SizeMask
		ifTrue: ["OR size into header1.  Must not do this if size > SizeMask"
				header1 := header1 + (sizeInBytes - (header1 bitAnd: SizeMask))]
		ifFalse: [hdrSize := 3.
				"Zero the size field of header1 if large"
				header1 := header1 - (header1 bitAnd: SizeMask)].
	^self allocate: sizeInBytes headerSize: hdrSize h1: header1 h2: classPointer h3: LargeContextSize doFill: false format: 0 "ignored cuz doFill: is false"! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 8/22/2017 14:18'!
instantiateSmallClass: classPointer sizeInBytes: sizeInBytes
	"This version of instantiateClass assumes that the total object 
	size is under 256 bytes, the limit for objects with only one or 
	two header words. Note that the size is specified in bytes 
	and should include 4 or 8 bytes for the base header word. 
	NOTE this code will only work for sizes that are an integral number of words
		(like not a 32-bit LargeInteger in a 64-bit system). 
	May cause a GC.
	Note that the created small object IS NOT FILLED and must be completed before
	returning it to Squeak. Since this call is used in routines that do just that we are safe.
	Break this rule and die."

	| hash header1 hdrSize |
	(sizeInBytes bitAnd: (self wordSize-1)) = 0 ifFalse:
		[self error: 'size must be integral number of words'].
	hash := self newObjectHash.
	header1 := (hash bitAnd: HashMaskUnshifted) << HashBitsOffset bitOr: (self formatOfClass: classPointer).
	hdrSize := (header1 bitAnd: CompactClassMask) > 0 "is this a compact class"
				ifTrue: [1]
				ifFalse: [2].
	header1 := header1 + (sizeInBytes - (header1 bitAnd: SizeMask+Size4Bit)).
	^self allocate: sizeInBytes headerSize: hdrSize h1: header1 h2: classPointer h3: 0 doFill: false format: 0 "ignored cuz doFill: is false"! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 10/30/2013 16:52'!
integerObjectOf: value
	"Convert the integer value, assumed to be in SmallInteger range, into a tagged SmallInteger object.
	 In C, use a shift and an add to set the tag bit.
	 In Smalltalk we have to work harder because the simulator works with strictly positive bit patterns."
	<returnTypeC: #sqInt>
	^self
		cCode: [(value << 1) + 1]
		inSmalltalk: [value >= 0
						ifTrue: [(value << 1) + 1]
						ifFalse: [((16r80000000 + value) << 1) + 1]]! !

!ObjectMemory methodsFor: 'interpreter access' stamp: ''!
integerValueOf: objectPointer
	"Translator produces 'objectPointer >> 1'"

	((objectPointer bitAnd: 16r80000000) ~= 0)
		ifTrue: ["negative"
				^ ((objectPointer bitAnd: 16r7FFFFFFF) >> 1)
					- 16r3FFFFFFF - 1  "Faster than -16r40000000 (a LgInt)"]
		ifFalse: ["positive"
				^ objectPointer >> 1]! !

!ObjectMemory methodsFor: 'simulation' stamp: 'eem 7/16/2010 15:12'!
interpreter
	<doNotGenerate>
	^self! !

!ObjectMemory methodsFor: 'initialization' stamp: 'eem 10/18/2020 20:57'!
invalidCompactClassError: className
	<var: 'className' type: #'const char *'>
	<inline: false>
	self cCode:
			['\nClass %s does not have the required compact class index\n' printf: className.
			 self exit: -1]
		inSmalltalk:
			[self error: 'Class ', className, ' does not have the required compact class index']! !

!ObjectMemory methodsFor: 'simulation' stamp: 'eem 4/9/2014 10:25'!
ioLoadModule: moduleNameIndex OfLength: moduleLength
	<doNotGenerate>
	<returnTypeC: #'void *'>
	"Dummy - provided by support code"
	^0! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 7/25/2014 21:37'!
is: oop instanceOf: classOop
	"Answer if oop is an instance of the given class. If the class has a (non-zero)
	 compactClassIndex use that to speed up the check."

	<inline: true>
	(self isIntegerObject: oop) ifTrue:
		[^classOop = (self splObj: ClassSmallInteger)].

	^self isClassOfNonImm: oop equalTo: classOop! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 12/20/2012 10:59'!
is: oop instanceOf: classOop compactClassIndex: compactClassIndex
	"Answer if oop is an instance of the given class. If the class has a (non-zero)
	 compactClassIndex use that to speed up the check.  N.B. Inlining should
	 result in classOop not being accessed if oop's compact class index and
	 compactClassIndex are non-zero."

	<inline: true>
	(self isIntegerObject: oop) ifTrue:
		[^false].

	^self isClassOfNonImm: oop equalTo: classOop compactClassIndex: compactClassIndex! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 9/30/2013 23:48'!
isArray: oop
	"Answer true if this is an indexable object with pointer elements, e.g., an array"
	^(self isNonIntegerObject: oop) and:[self isArrayNonImm: oop]! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 7/29/2014 20:14'!
isArrayNonImm: oop
	<api>
	"Answer if this is an indexable object with pointer elements, e.g., an array"
	^ (self formatOf: oop) = 2! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 12/4/2013 17:22'!
isBytes: oop
	"Answer true if the argument contains indexable bytes. See comment in formatOf:"
	"Note: Includes CompiledMethods."
	^(self isNonIntegerObject: oop) and:[self isBytesNonImm: oop]! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 12/4/2013 17:23'!
isBytesNonImm: oop
	"Answer true if the argument contains indexable bytes. See comment in formatOf:"
	"Note: Includes CompiledMethods."

	^(self formatOf: oop) >= self firstByteFormat! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 9/9/2013 15:14'!
isCharacterObject: oop
	<inline: true>
	"N.B.  Because Slang always inlines is:instanceOf:compactClassIndex:
	 (because is:instanceOf:compactClassIndex: has an inline: pragma) the
	 phrase (self splObj: ClassCharacter) is expanded in-place
	 and is _not_ evaluated if oop has a non-zero CompactClassIndex."
	^self
		is: oop
		instanceOf: (self splObj: ClassCharacter) 
		compactClassIndex: 0! !

!ObjectMemory methodsFor: 'primitive support' stamp: 'eem 4/11/2014 10:43'!
isCharacterValue: anInteger
	<api>
	^self isInRangeCharacterCode: anInteger! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 3/17/2015 13:22'!
isClassOfNonImm: oop equalTo: classOop
	"Answer if the given (non-immediate) object is an instance of the given class."

	| ccIndex cl |
	<inline: true>
	self assert: (self isNonImmediate: oop).

	cl := (ccIndex := self compactClassIndexOf: oop) = 0
			ifTrue: [(self classHeader: oop) bitAnd: AllButTypeMask]
			ifFalse: [self compactClassAt: ccIndex].
	^cl = classOop! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 11/10/2014 12:49'!
isClassOfNonImm: oop equalTo: classOop compactClassIndex: compactClassIndex
	"Answer if the given (non-immediate) object is an instance of the given class
	 that may have a compactClassIndex (if compactClassIndex is non-zero).
	 N.B. Inlining and/or compiler optimization should result in classOop not being
	 accessed if oop's compact class index and compactClassIndex are non-zero.
	 N.B.  Generally one cannot assume that if compactClassIndex is non-zero the
	 instances of the corresponding class always have the compactClassIndex
	 because the compact class index is only non-zero in short header instances."

	| ccIndex |
	<inline: true>
	self assert: (self isIntegerObject: oop) not.

	ccIndex := self compactClassIndexOf: oop.
	ccIndex = 0 ifTrue:
		[^((self classHeader: oop) bitAnd: AllButTypeMask) = classOop].
	compactClassIndex ~= 0 ifTrue:
		[^compactClassIndex = ccIndex].
	^classOop = (self compactClassAt: ccIndex)! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 9/10/2013 11:20'!
isCompiledMethod: oop
    "Answer whether the argument object is of compiled method format"
	<api>
    ^(self formatOf: oop) >= self firstCompiledMethodFormat! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 4/27/2016 07:41'!
isCompiledMethodFormat: format
	^format >= self firstCompiledMethodFormat! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 9/10/2013 11:20'!
isCompiledMethodHeader: objHeader
    "Answer whether the argument header has compiled method format"
    ^(self formatOfHeader: objHeader) >= self firstCompiledMethodFormat! !

!ObjectMemory methodsFor: 'contexts' stamp: 'eem 2/27/2013 13:52'!
isContext: oop
	<inline: true>
	^(self isNonIntegerObject: oop) and: [self isContextHeader: (self baseHeader: oop)]! !

!ObjectMemory methodsFor: 'contexts' stamp: 'eem 5/1/2014 07:45'!
isContextHeader: aHeader
	<inline: true>
	"c.f. {BlockContext. MethodContext. PseudoContext} collect: [:class| class -> class indexIfCompact]"
	^(self compactClassIndexOfHeader: aHeader) = 13			"BlockContext"
		or: [(self compactClassIndexOfHeader: aHeader) = 14]		"MethodContext"! !

!ObjectMemory methodsFor: 'contexts' stamp: 'eem 9/4/2013 11:36'!
isContextNonImm: oop
	<inline: true>
	^self isContextHeader: (self baseHeader: oop)! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 11/21/2014 17:01'!
isFloatInstance: oop
	<inline: true>
	"N.B.  Because Slang always inlines is:instanceOf:compactClassIndex:
	 (because is:instanceOf:compactClassIndex: has an inline: pragma) the
	 phrase (objectMemory splObj: ClassFloat) is expanded in-place and is
	 _not_ evaluated if oop has a non-zero CompactClassIndex."
	^self
		is: oop
		instanceOf: (self splObj: ClassFloat) 
		compactClassIndex: ClassFloatCompactIndex! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 12/5/2013 15:11'!
isForwarded: oop
	"Compatibility wth SpurMemoryManager.  In ObjectMemory, no forwarding pointers
	 are visible to the VM."
	<api>
	<cmacro: '(oop) false'>
	<inline: true>
	^false! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 9/13/2013 15:12'!
isForwardedClassTag: classTag
	"Compatibility wth SpurMemoryManager.  In ObjectMemory, no forwarding pointers
	 are visible to the VM."
	<inline: true>
	^false! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 5/4/2021 14:33'!
isForwarderFormat: format
	^false! !

!ObjectMemory methodsFor: 'header access' stamp: ''!
isFreeObject: oop

	^ (self headerType: oop) = HeaderTypeFree! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 10/25/2013 15:02'!
isImmediate: anOop
	<api>
	^self isIntegerObject: anOop! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 9/1/2013 16:51'!
isImmediateCharacter: objectPointer

	^false! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 12/1/2014 11:18'!
isImmediateFloat: oop
	^false! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
isImmutableWhileForwarding: oop 
	"The given object may have its header word in a forwarding block. Find  
	the value of the isImmutable flag in the object in spite of this obstacle. "
	| header fwdBlock |
	<inline: true>
	header := self longAt: oop.
	(header bitAnd: MarkBit) ~= 0 ifTrue:
		["oop is forwarded; get its real header from its forwarding table entry"
		fwdBlock := (header bitAnd: AllButMarkBitAndTypeMask) << 1.
		DoAssertionChecks ifTrue:
			[self fwdBlockValidate: fwdBlock].
		header := self longAt: fwdBlock + self wordSize].
	^(header bitAnd: ImmutabilityBit) ~= 0! !

!ObjectMemory methodsFor: 'finalization' stamp: 'EstebanLorenzano 5/20/2015 12:22'!
isInFinalizationList: oop head: listHead
	"it is known that oop's first fixed slot points to an instance of special object - ClassWeakFinalizer ,
	so it is safe to assume that we can iterate over it to figure out if oop is already in that list"
	<option: #PharoVM>
	| listItem |

	listItem := listHead.
	
	[ listItem == self nilObject ] whileFalse: [
		listItem == oop ifTrue: [ ^ true ].
		listItem := self fetchPointer: 1 ofObject: listItem.   
	].
	^ false
! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'dtl 4/22/2007 20:06'!
isInMemory: address 
	"Return true if the given address is in ST object memory"
	^ (self oop: address isGreaterThanOrEqualTo: self startOfMemory)
		and: [self oop: address isLessThan: endOfMemory]! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 4/11/2014 10:43'!
isInRangeCharacterCode: characterCode
	^characterCode between: 0 and: 255! !

!ObjectMemory methodsFor: 'object format' stamp: 'eem 9/9/2013 15:04'!
isIndexable: oop
	^(self formatOf: oop) >= 2! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 9/23/2013 07:15'!
isIndexableFormat: format
	^format >= self arrayFormat
	  and: [format <= self weakArrayFormat
			or: [format >= self firstLongFormat]]! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'nice 4/19/2016 21:41'!
isInstanceOfClassLargeNegativeInteger: oop
	<inline: true>
	"N.B.  Because Slang always inlines is:instanceOf:compactClassIndex:
	 (because is:instanceOf:compactClassIndex: has an inline: pragma) the
	 phrase (objectMemory splObj: ClassLargeNegativeInteger) is expanded
	 in-place and is _not_ evaluated if oop has a non-zero CompactClassIndex."
	^self
		is: oop
		instanceOf: (self splObj: ClassLargeNegativeInteger) 
		compactClassIndex: ClassLargeNegativeIntegerCompactIndex! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'nice 4/19/2016 21:39'!
isInstanceOfClassLargePositiveInteger: oop
	<inline: true>
	"N.B.  Because Slang always inlines is:instanceOf:compactClassIndex:
	 (because is:instanceOf:compactClassIndex: has an inline: pragma) the
	 phrase (objectMemory splObj: ClassLargePositiveInteger) is expanded
	 in-place and is _not_ evaluated if oop has a non-zero CompactClassIndex."
	^self
		is: oop
		instanceOf: (self splObj: ClassLargePositiveInteger) 
		compactClassIndex: ClassLargePositiveIntegerCompactIndex! !

!ObjectMemory methodsFor: 'interpreter access' stamp: ''!
isIntegerObject: objectPointer

	^ (objectPointer bitAnd: 1) > 0! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'nice 3/25/2015 00:42'!
isIntegerValue: intValue
	"Answer if the given value can be represented as a Smalltalk integer value.
	 In C, use a shift and XOR to set the sign bit if and only if the top two bits of the given
	 value are the same, then test the sign bit. Note that the top two bits are equal for
	 exactly those integers in the range that can be represented in 31-bits or 63-bits."
	<api>
	^self
		cCode: [(intValue asUnsignedInteger bitXor: (intValue asUnsignedInteger << 1)) asInteger >= 0]
		inSmalltalk: [intValue >= self minSmallInteger and: [intValue <= self maxSmallInteger]]! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'nice 4/18/2016 23:07'!
isLargeIntegerInstance: oop
	<inline: true>
	^(self isInstanceOfClassLargePositiveInteger: oop)
		or: [self isInstanceOfClassLargeNegativeInteger: oop]! !

!ObjectMemory methodsFor: 'forward compatibility' stamp: 'nice 12/26/2020 23:21'!
isLong64s: oop
	"forward compatibility with Spur"
	^false! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 3/23/2009 12:18'!
isMarked: oop
	<api>
	"Answer if oop's mark bit is set"
	^((self longAt: oop) bitAnd: MarkBit) ~= 0! !

!ObjectMemory methodsFor: 'contexts' stamp: 'eem 8/1/2013 10:57'!
isMethodContextHeader: aHeader
	<inline: true>
	"c.f. {BlockContext. MethodContext. PseudoContext} collect: [:class| class -> class indexIfCompact]"
	^(self compactClassIndexOfHeader: aHeader) = ClassMethodContextCompactIndex! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 1/9/2020 18:51'!
isNonImmediate: anOop
	<api>
	^self isNonIntegerObject: anOop! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 10/31/2014 11:38'!
isNonIntegerImmediate: oop
	"ObjectMemory only has integer immedates"
	<inline: true>
	^false! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 10/31/2014 11:39'!
isNonIntegerNonImmediate: oop
	"ObjectMemory only has integer immedates"
	<inline: true>
	^true! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 8/21/2013 12:54'!
isNonIntegerObject: objectPointer
	<api>
	^(objectPointer bitAnd: 1) = 0! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 12/11/2012 16:33'!
isObjImmutable: anOop
	<inline: true>
	^self cppIf: IMMUTABILITY
		ifTrue: [((self baseHeader: anOop) bitAnd: ImmutabilityBit hex) ~= 0]
		ifFalse: [false]! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 12/19/2008 22:12'!
isObjectForwarded: oop 
	"Answer if the given object has a forwarding table entry 
	 during a compaction or become operation."
	^(self isNonIntegerObject: oop) and: [self isMarked: oop]! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 9/10/2013 11:21'!
isOopCompiledMethod: oop
	"Answer whether the oop is an object of compiled method format"
	<api>
	^(self isNonIntegerObject: oop)
	 and: [(self formatOf: oop) >= self firstCompiledMethodFormat]! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 9/13/2013 14:48'!
isOopForwarded: oop
	"Compatibility wth SpurMemoryManager.  In ObjectMemory, no forwarding pointers
	 are visible to the VM."
	<inline: true>
	^false! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 12/11/2012 16:11'!
isOopImmutable: anOop
	<api>
	^(self isIntegerObject: anOop)
	  or: [self isObjImmutable: anOop]! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 12/11/2012 16:11'!
isOopMutable: anOop
	<api>
	^(self isIntegerObject: anOop) not
	  and: [(self isObjImmutable: anOop) not]! !

!ObjectMemory methodsFor: 'primitive support' stamp: 'eem 4/11/2014 10:48'!
isPinned: objOop
	"For forward-compatibility with Spur.  Answer false; ObjectMemory does not support pinning."
	<api>
	^false! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 12/4/2013 17:26'!
isPointers: oop
	"Answer true if the argument has only fields that can hold oops. See comment in formatOf:"

	^(self isNonIntegerObject: oop) and:[self isPointersNonImm: oop]! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 9/10/2013 08:50'!
isPointersFormat: format
	^format <= self lastPointerFormat! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 12/4/2013 17:26'!
isPointersNonImm: oop
	"Answer if the argument has only fields that can hold oops. See comment in formatOf:"

	^(self formatOf: oop) <= self lastPointerFormat! !

!ObjectMemory methodsFor: 'header formats' stamp: 'eem 10/25/2016 13:24'!
isPureBitsFormat: format
	^format >= self firstLongFormat
	  and: [format < self firstCompiledMethodFormat]! !

!ObjectMemory methodsFor: 'object format' stamp: 'eem 10/25/2016 13:27'!
isPureBitsNonImm: objOop
	"Answer if the argument contains only indexable words (no oops). See comment in formatOf:"

	^self isPureBitsFormat: (self formatOf: objOop)! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 6/29/2014 11:30'!
isSemaphoreObj: anObj
	^(self fetchClassOfNonImm: anObj) = (self splObj: ClassSemaphore)! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 6/29/2014 11:30'!
isSemaphoreOop: anOop
	^(self isNonImmediate: anOop)
	 and: [self isSemaphoreObj: anOop]! !

!ObjectMemory methodsFor: 'forward compatibility' stamp: 'nice 12/26/2020 23:22'!
isShorts: oop
	"forward compatibility with Spur"
	^false! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 9/26/2013 14:36'!
isWeak: oop
	"Answer true if the argument has only weak fields that can hold oops. See comment in formatOf:"
	^(self isNonIntegerObject: oop) and: [self isWeakNonImm: oop]! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 9/26/2013 14:36'!
isWeakNonImm: oop
	"Answer if the argument has only weak fields that can hold oops. See comment in formatOf:"
	^(self formatOf: oop) = self weakArrayFormat! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 12/4/2013 17:25'!
isWords: oop
	"Answer if the argument contains only indexable words (no oops). See comment in formatOf:"

	^(self isNonIntegerObject: oop) and: [self isWordsNonImm: oop]! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 12/4/2013 17:24'!
isWordsNonImm: oop
	"Answer if the argument contains only indexable words (no oops). See comment in formatOf:"

	^ (self formatOf: oop) = self firstLongFormat! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 12/4/2013 17:25'!
isWordsOrBytes: oop
	"Answer if the contains only indexable words or bytes (no oops). See comment in formatOf:"
	"Note: Excludes CompiledMethods."
	^(self isNonIntegerObject: oop) and: [self isWordsOrBytesNonImm: oop]! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 12/4/2013 17:25'!
isWordsOrBytesNonImm: oop
	"Answer if the contains only indexable words or bytes (no oops). See comment in formatOf:"
	"Note: Excludes CompiledMethods."

	| fmt |
	fmt := self formatOf: oop.
	^fmt = self firstLongFormat
	  or: [fmt >= self firstByteFormat
		  and: [fmt < self firstCompiledMethodFormat]]! !

!ObjectMemory methodsFor: 'forward compatibility' stamp: 'nice 12/26/2020 23:22'!
isWordsOrShorts: oop
	"forward compatibility with Spur"
	^(self isShorts: oop) or: [self isWords: oop]! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 11/25/2009 09:28'!
isYoung: oop
	<api>
	^(self isNonIntegerObject: oop)
	   and: [self oop: oop isGreaterThanOrEqualTo: youngStart]! !

!ObjectMemory methodsFor: 'garbage collection' stamp: 'eem 1/12/2013 17:26'!
isYoungRoot: oop
	"Answer if oop is a root for objects in youngSpace"
	^self isYoungRootHeader: (self baseHeader: oop)! !

!ObjectMemory methodsFor: 'garbage collection' stamp: 'eem 1/12/2013 17:25'!
isYoungRootHeader: header
	"Answer if oop is a root for objects in youngSpace"
	^(header bitAnd: RootBit) ~= 0! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
lastHash
	^lastHash! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
lastHash: aValue
	^lastHash := aValue! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 9/10/2013 11:41'!
lastPointerFormat
	"N.B. 5 is unused and could be used for ephemerons.
		7 is unused and could be used for 64-bit indexable."
	^4! !

!ObjectMemory methodsFor: 'object enumeration' stamp: 'eem 11/10/2014 12:49'!
lastPointerOf: objOop 
	"Return the byte offset of the last pointer field of the given object.  
	Works with CompiledMethods, as well as ordinary objects. 
	Can be used even when the type bits are not correct."
	<api>
	<inline: true>
	| fmt sz header contextSize |
	header := self baseHeader: objOop.
	fmt := self formatOfHeader: header.
	fmt <= self lastPointerFormat ifTrue:
		[(fmt = self indexablePointersFormat
		  and: [self isContextHeader: header])
					ifTrue: ["contexts end at the stack pointer"
						contextSize := self fetchStackPointerOf: objOop.
						^ CtxtTempFrameStart + contextSize * self wordSize].
				sz := self sizeBitsOfSafe: objOop.
				^sz - self baseHeaderSize "all pointers"].
	fmt < self firstCompiledMethodFormat ifTrue: [^0]. "no pointers"

	"CompiledMethod: contains both pointers and bytes:"
	header := self methodHeaderOf: objOop.
	^self lastPointerOfMethodHeader: header! !

!ObjectMemory methodsFor: 'gc -- mark and sweep' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
lastPointerOf: objOop recordWeakRoot: recordWeakRoot "<Boolean>"
	"Return the byte offset of the last pointer field of the given object.  
	 Works with CompiledMethods, as well as ordinary objects. 
	 Can be used even when the type bits are not correct.
	 This is a version of lastPointerOf: for markAndTrace:."
	| fmt sz header contextSize |
	<inline: true>
	header := self baseHeader: objOop.
	fmt := self formatOfHeader: header.
	fmt <= self lastPointerFormat ifTrue:
		[fmt >= self indexablePointersFormat ifTrue:
			[fmt = self lastPointerFormat ifTrue:
				[(recordWeakRoot and: [weakRootCount >= 0]) ifTrue:
					["And remember as weak root"
					 (weakRootCount := weakRootCount + 1) <= WeakRootTableSize ifFalse:
						[self error: 'weakRoots table overflow'].
					 weakRoots at: weakRootCount put: objOop].
				"Do not trace the object's indexed fields if it's a weak class"
				^(self nonWeakFieldsOf: objOop) * self bytesPerOop].
			"So fmt is 3"
			(self isContextHeader: header) ifTrue:
				["contexts end at the stack pointer avoiding having to init fields beyond it"
				 contextSize := self fetchStackPointerOf: objOop.
				 ^CtxtTempFrameStart + contextSize * self bytesPerOop]].
		 sz := self sizeBitsOfSafe: objOop.
		 ^sz - self baseHeaderSize  "all pointers"].
	fmt < self firstCompiledMethodFormat ifTrue: [^0]. "no pointers"

	"CompiledMethod: contains both pointers and bytes:"
	header := self methodHeaderOf: objOop.
	^self lastPointerOfMethodHeader: header! !

!ObjectMemory methodsFor: 'object enumeration' stamp: 'eem 11/10/2014 12:49'!
lastPointerOfMethodHeader: methodHeader 
	"Answer the byte offset of the last pointer field of a
	 CompiledMethod with the given header."
	<inline: true>
	^(self literalCountOfMethodHeader: methodHeader)
	  + LiteralStart - 1 * self bytesPerOop + self baseHeaderSize! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
lastPointerWhileForwarding: objOop 
	"The given object may have its header word in a forwarding block. Find  
	the offset of the last pointer in the object in spite of this obstacle. "
	| header fmt size contextSize |
	<inline: true>
	header := self headerWhileForwardingOf: objOop.
	fmt := self formatOfHeader: header.
	fmt <= self lastPointerFormat ifTrue:
		[(fmt = self indexablePointersFormat
		  and: [self isContextHeader: header]) ifTrue:
			["contexts end at the stack pointer"
			contextSize := self nacFetchStackPointerOf: objOop.
			self assert: ReceiverIndex + contextSize < (self lengthOf: objOop baseHeader: header format: fmt).
			^CtxtTempFrameStart + contextSize * self wordSize].
		"do sizeBitsOf: using the header we obtained"
		(header bitAnd: TypeMask) = HeaderTypeSizeAndClass
			ifTrue: [size := (self sizeHeader: objOop) bitAnd: AllButTypeMask]
			ifFalse: [size := header bitAnd: SizeMask].
		^size - self baseHeaderSize].
	fmt < self firstCompiledMethodFormat ifTrue: [^0]. "no pointers"
	"CompiledMethod: contains both pointers and bytes"
	self assert: (self isCompiledMethodHeader: header).
	header := self noCheckMethodHeaderOf: objOop.
	^self lastPointerOfMethodHeader: header! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 3/20/2020 11:33'!
lengthOf: oop
	"Return the number of indexable bytes or words in the given object. Assume the argument is not an integer. For a CompiledMethod, the size of the method header (in bytes) should be subtracted from the result."

	| header |
	<inline: true>
	header := self baseHeader: oop.
	^self lengthOf: oop baseHeader: header format: (self formatOfHeader: header)! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:06'!
lengthOf: oop baseHeader: hdr format: fmt
	"Return the number of fixed and indexable bytes, words, or object pointers in the
	given object. Assume the given oop is not an integer. For a CompiledMethod, the size
	of the method header (in bytes) should be subtracted from the result of this method."

	| sz |
	<inline: true>
	(hdr bitAnd: TypeMask) = HeaderTypeSizeAndClass
		ifTrue: [ sz := (self sizeHeader: oop) bitAnd: LongSizeMask ]
		ifFalse: [ sz := (hdr bitAnd: SizeMask)].
	sz := sz - (hdr bitAnd: Size4Bit).
	fmt <= self lastPointerFormat
		ifTrue: [ ^ (sz - self baseHeaderSize) >> self shiftForWord "words"].
	^fmt < self firstByteFormat
		ifTrue: [(sz - self baseHeaderSize) >> 2 "32-bit longs"]
		ifFalse: [(sz - self baseHeaderSize) - (fmt bitAnd: 3) "bytes"]! !

!ObjectMemory methodsFor: 'debug support' stamp: 'eem 9/17/2013 13:57'!
lengthOfMaybeImmediate: oop
	"for the message send breakpoint; selectors can be immediates."
	<inline: false>
	(self isImmediate: oop) ifTrue: [^0].
	^self lengthOf: oop! !

!ObjectMemory methodsFor: 'method access' stamp: 'eem 8/8/2014 15:08'!
literalCountOf: methodPointer
	<api>
	^self literalCountOfMethodHeader: (self methodHeaderOf: methodPointer)! !

!ObjectMemory methodsFor: 'memory access' stamp: 'eem 8/8/2014 10:17'!
literalCountOfMethodHeader: header
	<inline: true>
	self assert: (self isIntegerObject: header).
	^header >> 10 bitAnd: 16rFF! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 12/18/2014 13:26'!
loadFloatOrIntFrom: floatOrInt
	"If floatOrInt is an integer, then convert it to a C double float and return it.
	 If it is a Float, then load its value and return it.
	 Otherwise fail -- ie return with primErrorCode non-zero."

	<inline: true>
	<returnTypeC: #double>

	(self isIntegerObject: floatOrInt) ifTrue:
		[^(self integerValueOf: floatOrInt) asFloat].
	^self floatValueOf: floatOrInt! !

!ObjectMemory methodsFor: 'image segment in/out' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
loadImageSegmentFrom: segmentWordArray outPointers: outPointerArray
	"This primitive is called from Squeak as...
		<imageSegment> loadSegmentFrom: aWordArray outPointers: anArray."

"This primitive will load a binary image segment created by primitiveStoreImageSegment.  It expects the outPointer array to be of the proper size, and the wordArray to be well formed.  It will return as its value the original array of roots, and the erstwhile segmentWordArray will have been truncated to a size of one word, i.e. retaining the version stamp.  If this primitive should fail, the segmentWordArray will, sadly, have been reduced to an unrecognizable and unusable jumble.  But what more could you have done with it anyway?"

	| endSeg segOop fieldPtr fieldOop doingClass lastPtr extraSize mapOop lastOut outPtr hdrTypeBits header data |
	<inline: false>
	<var: #endSeg type: #usqInt>
	<var: #segOop type: #usqInt>
	<var: #fieldPtr type: #usqInt>
	<var: #lastOut type: #usqInt>
	<var: #outPtr type: #usqInt>
	<var: #lastPtr type: #usqInt>

	DoAssertionChecks ifTrue: [self verifyCleanHeaders].
	lastOut := outPointerArray + (self lastPointerOf: outPointerArray).
	endSeg := segmentWordArray + (self sizeBitsOf: segmentWordArray) - self baseHeaderSize.

	"Version check.  Byte order of the WordArray now"
	data := self longAt: segmentWordArray + self baseHeaderSize.
	(self readableFormat: (data bitAnd: 16rFFFF "low 2 bytes")) ifFalse:
		"Not readable -- try again with reversed bytes..."
		[self reverseBytesFrom: segmentWordArray + self baseHeaderSize to: endSeg + self wordSize.
		data := self longAt: segmentWordArray + self baseHeaderSize.
		(self readableFormat: (data bitAnd: 16rFFFF "low 2 bytes")) ifFalse:
			"Still NG -- put things back and fail"
			[self reverseBytesFrom: segmentWordArray + self baseHeaderSize to: endSeg + self wordSize.
			DoAssertionChecks ifTrue: [self verifyCleanHeaders].
			^PrimErrBadArgument]].
	"Reverse the Byte type objects if the data is from opposite endian machine.
	 Test top byte.  $d on the Mac or $s on the PC.  Rest of word is equal."
	(data >> 16) = (self imageSegmentVersion >> 16)
		ifFalse: "Reverse the byte-type objects once"
			[segOop := self oopFromChunk: segmentWordArray + self baseHeaderSize + self wordSize.
				 "Oop of first embedded object"
			self byteSwapByteObjectsFrom: segOop
				to: endSeg + self wordSize
				flipFloatsIf: false].

	"Proceed through the segment, remapping pointers..."
	segOop := self oopFromChunk: segmentWordArray + self baseHeaderSize + self wordSize.
	[segOop <= endSeg] whileTrue:
		[(self headerType: segOop) <= 1
			ifTrue: ["This object has a class field (type = 0 or 1) -- start with that."
					fieldPtr := segOop - self wordSize.  doingClass := true]
			ifFalse: ["No class field -- start with first data field"
					fieldPtr := segOop + self baseHeaderSize.  doingClass := false].
		lastPtr := segOop + (self lastPointerOf: segOop).	"last field"
		lastPtr > endSeg ifTrue:
			[DoAssertionChecks ifTrue: [self verifyCleanHeaders].
			^PrimErrBadIndex "out of bounds"].

		"Go through all oops, remapping them..."
		[fieldPtr > lastPtr] whileFalse:
			["Examine each pointer field"
			fieldOop := self longAt: fieldPtr.
			doingClass ifTrue:
				[hdrTypeBits := self headerType: fieldPtr.
				fieldOop := fieldOop - hdrTypeBits].
			(self isIntegerObject: fieldOop)
				ifTrue:
					["Integer -- nothing to do"
					fieldPtr := fieldPtr + self wordSize]
				ifFalse:
					[(fieldOop bitAnd: 3) = 0 ifFalse:
						[^PrimErrBadIndex "bad oop"].
					(fieldOop bitAnd: 16r80000000) = 0
						ifTrue: ["Internal pointer -- add segment offset"
								mapOop := fieldOop + segmentWordArray]
						ifFalse: ["External pointer -- look it up in outPointers"
								outPtr := outPointerArray + (fieldOop bitAnd: 16r7FFFFFFF).
								outPtr > lastOut ifTrue:
									[^PrimErrBadIndex "out of bounds"].
								mapOop := self longAt: outPtr].
					doingClass
						ifTrue: [self longAt: fieldPtr put: mapOop + hdrTypeBits.
								fieldPtr := fieldPtr + 8.
								doingClass := false]
						ifFalse: [self longAt: fieldPtr put: mapOop.
								fieldPtr := fieldPtr + self wordSize].
					segOop < youngStart ifTrue:
						[self possibleRootStoreInto: segOop value: mapOop]]].
		segOop := self objectAfter: segOop].

	"Again, proceed through the segment checking consistency..."
	segOop := self oopFromChunk: segmentWordArray + self baseHeaderSize + self wordSize.
	[segOop <= endSeg] whileTrue:
		[(self oopHasAcceptableClass: segOop) ifFalse:
			[^PrimErrInappropriate "inconsistency"].
		fieldPtr := segOop + self baseHeaderSize.		"first field"
		lastPtr := segOop + (self lastPointerOf: segOop).	"last field"
		"Go through all oops, remapping them..."
		[fieldPtr > lastPtr] whileFalse:
			["Examine each pointer field"
			fieldOop := self longAt: fieldPtr.
			(self oopHasAcceptableClass: fieldOop) ifFalse:
				[^PrimErrInappropriate "inconsistency"].
			fieldPtr := fieldPtr + self wordSize].
		segOop := self objectAfter: segOop].

	"Truncate the segment word array to size = BytesPerWord (vers stamp only)"
	extraSize := self extraHeaderBytes: segmentWordArray.
	hdrTypeBits := self headerType: segmentWordArray.
	extraSize = 8
		ifTrue: [self longAt: segmentWordArray-extraSize put: self baseHeaderSize + self wordSize + hdrTypeBits]
		ifFalse: [header := self longAt: segmentWordArray.
				self longAt: segmentWordArray
					put: header - (header bitAnd: SizeMask) + self baseHeaderSize + self wordSize].	
	"and return the roots array which was first in the segment"
	DoAssertionChecks ifTrue: [self verifyCleanHeaders].
	^self oopFromChunk: segmentWordArray + self baseHeaderSize + self wordSize! !

!ObjectMemory methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:03'!
longPrintInstancesOf: aClassOop
	"Scan the heap printing the oops of any and all objects that are instances of aClassOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32"
	| oop |
	oop := self firstAccessibleObject.
	[oop = nil] whileFalse:
		[(self fetchClassOfNonImm: oop) = aClassOop ifTrue:
			[self longPrintOop: oop; cr].
		 oop := self accessibleObjectAfter: oop]! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
lowSpaceThreshold
	^lowSpaceThreshold! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
lowSpaceThreshold: aValue
	^lowSpaceThreshold := aValue! !

!ObjectMemory methodsFor: 'garbage collection' stamp: 'eem 3/3/2009 11:07'!
lowestFreeAfter: chunk 
	"Return the first free block after the given chunk in memory."
	| oop oopHeader oopHeaderType oopSize |
	<inline: false>
	oop := self oopFromChunk: chunk.
	[self oop: oop isLessThan: endOfMemory] whileTrue:
		[oopHeader := self baseHeader: oop.
		 oopHeaderType := oopHeader bitAnd: TypeMask.
		 oopHeaderType = HeaderTypeFree ifTrue:
			[^oop].
		 oopHeaderType = HeaderTypeSizeAndClass
			ifTrue: [oopSize := (self sizeHeader: oop) bitAnd: AllButTypeMask]
			ifFalse: [oopSize := oopHeader bitAnd: SizeMask].
		 oop := self oopFromChunk: oop + oopSize].
	self error: 'expected to find at least one free object'.
	^nil! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 9/14/2013 11:09'!
mapPointersInObjectsFrom: memStart to: memEnd
	"Use the forwarding table to update the pointers of all non-free objects in the given range of memory.
	 Also remap pointers in root objects which may contains pointers into the given memory range, and
	 don't forget to flush the method cache based on the range."
	<inline: false>
	(self shouldRemapObj: nilObj) ifTrue:
		[nilObj := self remapObj: nilObj].
	(self shouldRemapObj: falseObj) ifTrue:
		[falseObj := self remapObj: falseObj].
	(self shouldRemapObj: trueObj) ifTrue:
		[trueObj := self remapObj: trueObj].
	(self shouldRemapObj: specialObjectsOop) ifTrue:
		[specialObjectsOop := self remapObj: specialObjectsOop].
	"update interpreter variables"
	self
		mapInterpreterOops;
		flushMethodCacheFrom: memStart to: memEnd.
	self remapRemapBufferAndExtraRoots.
	self updatePointersInRootObjectsFrom: memStart to: memEnd.
	self updatePointersInRangeFrom: memStart to: memEnd.
! !

!ObjectMemory methodsFor: 'gc -- mark and sweep' stamp: 'eem 2/2/2010 15:26'!
markAndTrace: oop
	"Mark all objects reachable from the given one.
	Trace from the given object even if it is old.
	Do not trace if it is already marked.
	Mark it only if it is a young object."
	"Tracer state variables:
		child		object being examined
		field		next field of child to examine
		parentField	field where child was stored in its referencing object"
	<api>
	| header lastFieldOffset action |
	self assert: (self addressCouldBeObjWhileForwarding: oop).
	header := self longAt: oop.
	(header bitAnd: MarkBit) = 0 ifFalse:
		[^ 0  "already marked"].

	"record tracing status in object's header"
	header := (header bitAnd: AllButTypeMask) bitOr: HeaderTypeGC.
	(self oop: oop isGreaterThanOrEqualTo: youngStart) ifTrue:
		[ header := header bitOr: MarkBit ].  "mark only if young"
	self longAt: oop put: header.

	"initialize the tracer state machine"
	parentField := GCTopMarker.
	child := oop.
	lastFieldOffset := self lastPointerOf: oop recordWeakRoot: true.
	field := oop + lastFieldOffset.
	action := StartField.
	youngStartLocal := youngStart.
	statMarkCountLocal := statMarkCount.
	"run the tracer state machine until all objects reachable from oop are marked"
	[action = StartField ifTrue: [ action := self startField ].
	 action = StartObj ifTrue: [ action := self startObj ].
	 action = Upward ifTrue: [ action := self upward ].
	 action = Done] whileFalse.
	statMarkCount := statMarkCountLocal.! !

!ObjectMemory methodsFor: 'gc -- mark and sweep' stamp: 'eem 1/12/2013 17:41'!
markPhase
	"Mark phase of the mark and sweep garbage collector. Set 
	the mark bits of all reachable objects. Free chunks are 
	untouched by this process."
	"Assume: All non-free objects are initially unmarked. Root 
	objects were unmarked when they were made roots. (Make 
	sure this stays true!!!!)."
	| oop |
	<inline: false>
	"clear the recycled context lists"
	freeContexts := NilContext.
	freeLargeContexts := NilContext.
	"trace the interpreter's objects, including the active stack 
	and special objects array"
	self markAndTraceInterpreterOops.
	statSpecialMarkCount := statMarkCount.
	"trace the roots"
	1 to: rootTableCount do: [:i | 
		oop := rootTable at: i.
		self markAndTrace: oop].
	1 to: extraRootCount do:[:i|
		oop := (extraRoots at: i) at: 0.
		((self isIntegerObject: oop) or: [self isFreeObject: oop]) ifFalse:
			[self markAndTrace: oop]].
! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 1/8/2015 14:40'!
maxCInteger
	^16rFFFFFFFF! !

!ObjectMemory methodsFor: 'forward compatibility' stamp: 'eem 7/19/2010 12:49'!
maxIdentityHash
	^HashMaskUnshifted! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 11/19/2014 15:54'!
maxSmallInteger
	^1073741823! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 2/21/2013 10:26'!
maybeSplObj: index
	<api>
	"Answer one of the objects in the SpecialObjectsArray, if in range, otherwise answer nil."
	^index < (self lengthOf: specialObjectsOop) ifTrue:
		[self fetchPointer: index ofObject: specialObjectsOop]! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/30/2021 18:48'!
memory
	"memory is a simulation thing, but much more convenient to declare here than in every simulator subclass."
	^memory! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
memory: aValue
	^memory := aValue! !

!ObjectMemory methodsFor: 'image save/restore' stamp: 'eem 10/10/2013 06:37'!
memoryBaseForImageRead
	"Answer the address to read the image into."
	^memory! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
memoryLimit
	^memoryLimit! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 7/11/2014 12:50'!
methodDictionaryHash: oop mask: mask
	<inline: true>
	^mask bitAnd: ((self isImmediate: oop)
						ifTrue: [self integerValueOf: oop]
						ifFalse: [self hashBitsOf: oop])! !

!ObjectMemory methodsFor: 'method access' stamp: 'eem 8/8/2014 09:43'!
methodHeaderOf: methodObj
	"Answer the method header of a CompiledMethod object."
	self assert: (self isCompiledMethod: methodObj).
	^self fetchPointer: HeaderIndex ofObject: methodObj! !

!ObjectMemory methodsFor: 'allocation' stamp: 'cb 5/22/2017 16:53'!
minSlotsForShortening
	"HeaderTypeFree free chunks require 1 word of header"
	<api>
	^1! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 11/19/2014 15:10'!
minSmallInteger
	^-1073741824! !

!ObjectMemory methodsFor: 'simulation' stamp: 'eem 7/14/2015 12:26'!
newInputEventAccessorOfSize: numElements
	<doNotGenerate>
	^CArrayAccessor on: (IntegerArray new: 8)! !

!ObjectMemory methodsFor: 'header access' stamp: ''!
newObjectHash
	"Answer a new 16-bit pseudo-random number for use as an identity hash."

	lastHash := 13849 + (27181 * lastHash) bitAnd: 65535.
	^ lastHash
! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 10/12/2013 09:24'!
newSpaceSize
	^freeBlock - self startOfMemory! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 3/26/2009 20:34'!
nilObject
	<api>
	"For access from BitBlt module & Cogit"
	^ nilObj! !

!ObjectMemory methodsFor: 'method access' stamp: 'eem 8/8/2014 11:18'!
noCheckMethodHeaderOf: methodObj
	"Answer the method header of a CompiledMethod object."
	^self fetchPointer: HeaderIndex ofObject: methodObj! !

!ObjectMemory methodsFor: 'forward compatibility' stamp: 'eem 8/23/2014 11:37'!
noFixupFollowField: fieldIndex ofObject: anObject
	"Spur compatibility; in V3 this is a synonym for fetchPointer:ofObject:"
	<inline: true>
	^self fetchPointer: fieldIndex ofObject: anObject! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 7/7/2014 09:35'!
noShiftCompactClassIndexOfHeader: header
	<api>
	<inline: true>
	^header bitAnd: 16r1F << self compactClassFieldLSB! !

!ObjectMemory methodsFor: 'header formats' stamp: 'eem 12/7/2017 19:45'!
nonIndexablePointerFormat
	^1! !

!ObjectMemory methodsFor: 'object format' stamp: 'eem 9/26/2013 14:36'!
nonWeakFieldsOf: oop
	"Return the number of non-weak fields in oop (i.e. the number of fixed fields).
	Note: The following is copied from fixedFieldsOf:format:length: since we do know
	the format of the oop (e.g. format = 4) and thus don't need the length."
	| class classFormat |

	self assert: (self isWeakNonImm: oop).

	"fmt = 3 or 4: mixture of fixed and indexable fields, so must look at class format word"
	class := self fetchClassOfNonImm: oop.
	classFormat := self formatOfClass: class.
	^(classFormat >> 11 bitAnd: 16rC0) + (classFormat >> 2 bitAnd: 16r3F) - 1! !

!ObjectMemory methodsFor: 'garbage collection' stamp: 'eem 11/10/2014 12:49'!
noteAsRoot: oop headerLoc: headerLoc 
	"Record that the given oop in the old object area points to an object in the young area.
	 HeaderLoc is usually = oop, but may be an addr in a forwarding block."
	| header |
	<inline: true>
	header := self longAt: headerLoc.
	(self isYoungRootHeader: header) ifFalse:
		"record oop as root only if not already recorded"
		[rootTableCount < RootTableSize
			ifTrue:
				"record root if there is enough room in the roots table.
				 IMPORTANT: since clearRootsTable is the only thing that clears root bits
				 do *not* set the root bit unless an object is in the root table.  checking
				 routines will complain about the root bit being unset instead of the table
				 being full, but that's life"
				[rootTableCount := rootTableCount + 1.
				 rootTable at: rootTableCount put: oop.
				 self longAt: headerLoc put: (header bitOr: RootBit).
				 rootTableCount >= RootTableRedZone ifTrue:
					"if we're now in the red zone force an IGC ASAP"
					[allocationCount := allocationsBetweenGCs + 1]]
			ifFalse: "note overflow; will need to do a fullGC instead of an incremental."
				[rootTableOverflowed := true]]! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 7/28/2014 08:12'!
num16BitUnitsOf: objOop 
	"Answer the number of 16-bit units in the given non-immediate object.
	 N..B. Rounds down 8-bit units, so a 5 byte object has 2 16-bit units.
	 Does not adjust the size of contexts by stackPointer."
	^(self numBytesOf: objOop) >> 1! !

!ObjectMemory methodsFor: 'object access' stamp: ' nice 4/12/2018 10:38'!
num32BitUnitsOf: objOop 
	"Answer the number of 32-bit units in the given non-immediate object.
	 N..B. Rounds down 8-bit units, so a 7 byte object has 1 32-bit unit.
	 Does not adjust the size of contexts by stackPointer."
	^(self numBytesOf: objOop) >> 2! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 7/28/2014 08:14'!
num64BitUnitsOf: objOop 
	"Answer the number of 16-bit units in the given non-immediate object.
	 N..B. Rounds down 8-bit units, so a 15 byte object has 1 64-bit unit.
	 Does not adjust the size of contexts by stackPointer."
	^(self numBytesOf: objOop) >> 3! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
numBytesOf: objOop 
	"Answer the number of indexable bytes in the given non-immediate object.
	 Does not adjust the size of contexts by stackPointer."
	<api>
	| header sz fmt |
	header := self baseHeader: objOop.
	sz := (header bitAnd: TypeMask) = HeaderTypeSizeAndClass
			ifTrue: [(self sizeHeader: objOop) bitAnd: AllButTypeMask]
			ifFalse: [header bitAnd: SizeMask].
	fmt := self formatOfHeader: header.
	^fmt < self firstByteFormat
		ifTrue: [(sz - self baseHeaderSize)]  "words"
		ifFalse: [(sz - self baseHeaderSize) - (fmt bitAnd: 3)]  "bytes"! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 3/6/2015 10:32'!
numBytesOfBytes: objOop 
	"Answer the number of indexable bytes in the given non-immediate byte-indexable object."
	<api>
	| header sz fmt |
	header := self baseHeader: objOop.
	sz := (header bitAnd: TypeMask) = HeaderTypeSizeAndClass
			ifTrue: [(self sizeHeader: objOop) bitAnd: AllButTypeMask]
			ifFalse: [header bitAnd: SizeMask].
	fmt := self formatOfHeader: header.
	self assert: fmt >= self firstByteFormat.
	^(sz - self baseHeaderSize) - (fmt bitAnd: 3)! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:06'!
numSlotsOf: obj
	"Answer the number of oop-sized elements in the given object.
	 Unlike lengthOf: this does not adjust the length of a context
	 by the stackPointer and so can be used e.g. by cloneContext:"
	<api>
	| header sz |
	header := self baseHeader: obj.
	sz := (header bitAnd: TypeMask) = HeaderTypeSizeAndClass
			ifTrue: [(self sizeHeader: obj) bitAnd: AllButTypeMask]
			ifFalse: [header bitAnd: SizeMask].
	^sz - self baseHeaderSize >> self shiftForWord! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem 2/2/2015 18:36'!
numSlotsOfAny: obj
	"Sput compatibility"
	^self numSlotsOf: obj! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 11/19/2014 13:12'!
numSmallIntegerBits
	^31! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 11/18/2014 17:37'!
numSmallIntegerTagBits
	^1! !

!ObjectMemory methodsFor: 'object enumeration' stamp: 'eem 11/10/2014 12:49'!
objectAfter: oop 
	"Return the object or free chunk immediately following the 
	given object or free chunk in memory. Return endOfMemory 
	when enumeration is complete."
	| sz |
	<api>
	<inline: true>
	DoAssertionChecks
		ifTrue: [(self oop: oop isGreaterThanOrEqualTo:endOfMemory)
					ifTrue: [self error: 'no objects after the end of memory']].
	(self isFreeObject: oop)
		ifTrue: [sz := self sizeOfFree: oop]
		ifFalse: [sz := self sizeBitsOf: oop].
	^ self oopFromChunk: oop + sz! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 11/10/2014 12:49'!
objectAfterWhileForwarding: oop
	"Return the oop of the object after the given oop when the actual header of the oop may be in the forwarding table."

	| header fwdBlock realHeader sz |
	<inline: true>
	header := self longAt: oop.
	(header bitAnd: MarkBit) = 0 ifTrue: [ ^ self objectAfter: oop ].  "oop not forwarded"

	"Assume: mark bit cannot be set on a free chunk, so if we get here,
	 oop is not free and it has a forwarding table entry"

	fwdBlock := (header bitAnd: AllButMarkBitAndTypeMask) << 1.
	self assert: (self fwdBlockValid: fwdBlock).
	realHeader := self longAt: fwdBlock + self wordSize.
	"following code is like sizeBitsOf:"
	(realHeader bitAnd: TypeMask) = HeaderTypeSizeAndClass
		ifTrue: [ sz := (self sizeHeader: oop) bitAnd: LongSizeMask ]
		ifFalse: [ sz := realHeader bitAnd: SizeMask ].

	^ self oopFromChunk: (oop + sz)! !

!ObjectMemory methodsFor: 'object enumeration' stamp: 'eem 09/28/2015 18:48'!
objectBefore: address 
	"Return the object or start of free space immediately preceding the given
	 address, object or free chunk in memory. If none, return 0.  This is for debugging only."
	| obj nextObj sz |
	<api>
	obj := self oopFromChunk: ((self oop: address isGreaterThan: youngStart)
								ifTrue: [youngStart]
								ifFalse: [self startOfMemory]).
	[self oop: obj isLessThan: address] whileTrue:
		[(self isFreeObject: obj)
			ifTrue: [sz := self sizeOfFree: obj]
			ifFalse: [sz := self sizeBitsOf: obj].
		 nextObj := self oopFromChunk: obj + sz.
		 (self oop: nextObj isGreaterThanOrEqualTo: address) ifTrue:
			[^obj].
		 obj := nextObj].
	^0! !

!ObjectMemory methodsFor: 'object enumeration' stamp: 'eem 09/28/2015 18:48'!
objectExactlyBefore: oop 
	"Return the object or start of free space immediately preceding the given
	 object or free chunk in memory. If none, return 0.  This is for debugging only."
	| obj nextObj sz |
	<api>
	obj := self oopFromChunk: ((self oop: oop isGreaterThan: youngStart)
								ifTrue: [youngStart]
								ifFalse: [self startOfMemory]).
	[self oop: obj isLessThan: obj] whileTrue:
		[(self isFreeObject: obj)
			ifTrue: [sz := self sizeOfFree: obj]
			ifFalse: [sz := self sizeBitsOf: obj].
		 nextObj := self oopFromChunk: obj + sz.
		 nextObj = oop ifTrue:
			[^obj].
		 obj := nextObj].
	^0! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
objectIsImmutableAndReferencesForwarded: oop 
	"Answer if an object is immutable and references a forwarded object.  Used to fail become for immutable referents of becomees."
	"Note: The given oop may be forwarded itself, which means that its real header is in its forwarding table entry."
	| fieldOffset fieldOop |
	<inline: true>
	(self isImmutableWhileForwarding: oop) ifTrue:
		[fieldOffset := self lastPointerWhileForwarding: oop.
		[fieldOffset >= self baseHeaderSize] whileTrue:
			[fieldOop := self longAt: oop + fieldOffset.
			(self isObjectForwarded: fieldOop) ifTrue: [^true].
			 fieldOffset := fieldOffset - self wordSize]].
	^false! !

!ObjectMemory methodsFor: 'simulation' stamp: 'eem 10/12/2009 04:17'!
objectMemory
	<doNotGenerate>
	^self! !

!ObjectMemory methodsFor: 'debug support' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
okayOop: signedOop
	"Verify that the given oop is legitimate. Check address, header, and size but not class."

	| sz type fmt unusedBit oop |
	<var: #oop type: #usqInt>
	oop := self cCoerce: signedOop to: #usqInt.

	"address and size checks"
	(self isIntegerObject: oop) ifTrue: [ ^true ].
	(self oop: oop isGreaterThanOrEqualTo: self startOfMemory andLessThan: endOfMemory)
		ifFalse: [ self error: 'oop is not a valid address'. ^false ].
	((oop \\ self wordSize) = 0)
		ifFalse: [ self error: 'oop is not a word-aligned address'. ^false ].
	sz := self sizeBitsOf: oop.
	(oop + sz) < endOfMemory
		ifFalse: [ self error: 'oop size would make it extend beyond the end of memory'. ^false ].

	"header type checks"
	type := self headerType: oop.
	type = HeaderTypeFree
		ifTrue:  [ self error: 'oop is a free chunk, not an object'. ^false ].
	type = HeaderTypeShort ifTrue: [
		(self compactClassIndexOf: oop) = 0
			ifTrue:  [ self error: 'cannot have zero compact class field in a short header'. ^false ].
	].
	type = HeaderTypeClass ifTrue: [
		((oop >= self wordSize) and: [(self headerType: oop - self wordSize) = type])
			ifFalse: [ self error: 'class header word has wrong type'. ^false ].
	].
	type = HeaderTypeSizeAndClass ifTrue: [
		((oop >= (self wordSize*2)) and:
		 [(self headerType: oop - (self wordSize*2)) = type and:
		 [(self headerType: oop - self wordSize) = type]])
			ifFalse: [ self error: 'class header word has wrong type'. ^false ].
	].

	"format check"
	fmt := self formatOf: oop.
	((fmt = 5) | (fmt = 7))
		ifTrue:  [ self error: 'oop has an unknown format type'. ^false ].

	"mark and root bit checks"
	unusedBit := 16r20000000.
	self wordSize = 8
		ifTrue:
			[unusedBit := unusedBit << 16.
			 unusedBit := unusedBit << 16].
	((self longAt: oop) bitAnd: unusedBit) = 0
		ifFalse: [ self error: 'unused header bit 30 is set; should be zero'. ^false ].
"xxx
	((self longAt: oop) bitAnd: MarkBit) = 0
		ifFalse: [ self error: 'mark bit should not be set except during GC' ].
xxx"
	((self isYoungRoot: oop) and: [oop >= youngStart])
		ifTrue: [ self error: 'root bit is set in a young object'. ^false ].
	^true
! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 10/12/2013 09:24'!
oldSpaceSize
	^youngStart - self startOfMemory! !

!ObjectMemory methodsFor: 'oop/chunk conversion' stamp: 'JMM 12/4/2002 19:56'!
oopFromChunk: chunk
	"Compute the oop of this chunk by adding its extra header bytes."

	^ chunk + (self extraHeaderBytes: chunk)! !

!ObjectMemory methodsFor: 'image segment in/out' stamp: 'eem 12/4/2013 17:26'!
oopHasAcceptableClass: signedOop
	"Similar to oopHasOkayClass:, except that it only returns true or false."

	| oopClass formatMask behaviorFormatBits oopFormatBits oop |
	<var: #oop type: #usqInt>
	<var: #oopClass type: #usqInt>

	(self isIntegerObject: signedOop) ifTrue: [^ true].

	oop := self cCoerce: signedOop to: #usqInt.
	(self addressCouldBeObj: oop) ifFalse: [^ false].

	oopClass := self cCoerce: (self fetchClassOfNonImm: oop) to: #usqInt.
	(self addressCouldBeObj: oopClass) ifFalse: [^ false].
	(oopClass + (self sizeBitsOf: oopClass)) < freeBlock ifFalse: [^ false].

	((self isPointersNonImm: oopClass) and: [(self lengthOf: oopClass) >= 3]) ifFalse: [^ false].

	formatMask := (self isBytesNonImm: oop)
						ifTrue: [16rC00]  "ignore extra bytes size bits"
						ifFalse: [16rF00].

	behaviorFormatBits := (self formatOfClass: oopClass) bitAnd: formatMask.
	oopFormatBits := (self baseHeader: oop) bitAnd: formatMask.
	behaviorFormatBits = oopFormatBits ifFalse: [^ false].
	^ true! !

!ObjectMemory methodsFor: 'debug support' stamp: 'eem 12/4/2013 17:26'!
oopHasOkayClass: signedOop
	"Attempt to verify that the given oop has a reasonable behavior. The class must be a valid, non-integer oop and must not be nilObj. It must be a pointers object with three or more fields. Finally, the instance specification field of the behavior must match that of the instance."

	| oop oopClass formatMask behaviorFormatBits oopFormatBits |
	<var: #oop type: #usqInt>
	<var: #oopClass type: #usqInt>

	oop := self cCoerce: signedOop to: #usqInt.
	self okayOop: oop.
	oopClass := self cCoerce: (self fetchClassOf: oop) to: #usqInt.

	(self isIntegerObject: oopClass)
		ifTrue: [ self error: 'a SmallInteger is not a valid class or behavior'. ^false ].
	(self okayOop: oopClass)
		ifFalse: [ self error: 'class oop is not ok'. ^false ].
	((self isPointersNonImm: oopClass) and: [(self lengthOf: oopClass) >= 3])
		ifFalse: [ self error: 'a class (behavior) must be a pointers object of size >= 3'. ^false ].
	(self isBytes: oop)
		ifTrue: [ formatMask := 16rC00 ]  "ignore extra bytes size bits"
		ifFalse: [ formatMask := 16rF00 ].

	behaviorFormatBits := (self formatOfClass: oopClass) bitAnd: formatMask.
	oopFormatBits := (self baseHeader: oop) bitAnd: formatMask.
	behaviorFormatBits = oopFormatBits
		ifFalse: [ self error: 'object and its class (behavior) formats differ'. ^false ].
	^true! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 4/2/2009 22:48'!
popRemappableOop
	"Pop and return the possibly remapped object from the remap buffer."
	<api>
	| oop |
	oop := remapBuffer at: remapBufferCount.
	remapBufferCount := remapBufferCount - 1.
	^ oop! !

!ObjectMemory methodsFor: 'garbage collection' stamp: 'dtl (auto pragmas 12/08) 4/22/2007 20:11'!
possibleRootStoreInto: oop value: valueObj 
	"oop is an old object. If valueObj is young, mark the object as a root."

	<inline: false>
	((self oop: valueObj isGreaterThanOrEqualTo: youngStart)
			and: [(self isIntegerObject: valueObj) not])
		ifTrue: ["Yes, valueObj is a young object"
				self noteAsRoot: oop headerLoc: oop]! !

!ObjectMemory methodsFor: 'image save/restore' stamp: 'eem 2/12/2014 17:09'!
postSnapshot
	"No op for Spur compatibility."! !

!ObjectMemory methodsFor: 'become' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
prepareForwardingTableForBecoming: array1 with: array2 twoWay: twoWayFlag 
	"Ensure that there are enough forwarding blocks to 
	accomodate this become, then prepare forwarding blocks for 
	the pointer swap. Return true if successful."
	"Details: Doing a GC might generate enough space for 
	forwarding blocks if we're short. However, this is an 
	uncommon enough case that it is better handled by primitive 
	fail code at the Smalltalk level."

	"Important note on multiple references to same object  - since the preparation of
	fwdBlocks is NOT idempotent we get VM crashes if the same object is referenced more
	than once in such a way as to require multiple fwdBlocks.
	oop1 forwardBecome: oop1 is ok since only a single fwdBlock is needed.
	oop1 become: oop1 would fail because the second fwdBlock woudl not have the actual object
	header but rather the mutated ref to the first fwdBlock.
	Further problems can arise with an array1 or array2 that refer multiply to the same 
	object. This would notbe expected input for programmer writen code but might arise from
	automatic usage such as in ImageSegment loading.
	To avoid the simple and rather common case of oop1 become*: oop1, we skip such pairs
	and simply avoid making fwdBlocks - it is redundant anyway"
	| entriesNeeded entriesAvailable fieldOffset oop1 oop2 fwdBlock fwdBlkSize |
	entriesNeeded := (self lastPointerOf: array1) // self wordSize. "need enough entries for all oops"
	"Note: Forward blocks must be quadword aligned - see fwdTableInit:."
	twoWayFlag
		ifTrue: ["Double the number of blocks for two-way become"
			entriesNeeded := entriesNeeded * 2.
			fwdBlkSize := self wordSize * 2]
		ifFalse: ["One-way become needs backPointers in fwd blocks."
			fwdBlkSize := self wordSize * 4].
	entriesAvailable := self fwdTableInit: fwdBlkSize.
	entriesAvailable < entriesNeeded
		ifTrue: [self initializeMemoryFirstFree: freeBlock.
			"re-initialize the free block"
			^ false].
	fieldOffset := self lastPointerOf: array1.
	[fieldOffset >= self baseHeaderSize]
		whileTrue: [oop1 := self longAt: array1 + fieldOffset.
			oop2 := self longAt: array2 + fieldOffset.
			"if oop1 == oop2, no need to do any work for this pair.
			May still be other entries in the arrays though so keep looking"
			oop1 ~= oop2 ifTrue:
				[(self hasForwardingBlock: oop1) ifFalse: "Don't allocate multiple forwarding entries for duplicates."
					[fwdBlock := self fwdBlockGet: fwdBlkSize.
					 self
						initForwardBlock: fwdBlock
						mapping: oop1
						to: oop2
						withBackPtr: twoWayFlag not].
				 (twoWayFlag
				  and: [(self hasForwardingBlock: oop2) not]) ifTrue: "Again don't get confused by duplicates"
					["Second block maps oop2 back to oop1 for two-way become"
							fwdBlock := self fwdBlockGet: fwdBlkSize.
							self
								initForwardBlock: fwdBlock
								mapping: oop2
								to: oop1
								withBackPtr: twoWayFlag not]].
			fieldOffset := fieldOffset - self wordSize].
	^ true! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 10/28/2009 19:16'!
primitiveErrorTable
	<api>
	^self splObj: PrimErrTableIndex! !

!ObjectMemory methodsFor: 'image segment in/out' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
primitiveFailCodeAfterCleanup: outPointerArray
	"If the storeSegment primitive fails, it must clean up first."

	| i lastAddr |   "Store nils throughout the outPointer array."
	lastAddr := outPointerArray + (self lastPointerOf: outPointerArray).
	i := outPointerArray + self baseHeaderSize.
	[i <= lastAddr] whileTrue:
		[self longAt: i put: nilObj.
		i := i + self wordSize].

	DoAssertionChecks ifTrue: [self verifyCleanHeaders].
	^PrimErrGenericFailure! !

!ObjectMemory methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:03'!
printActivationsOf: aMethodObj
	"Scan the heap printing the oops of any and all contexts that refer to anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32"
	| oop |
	oop := self firstAccessibleObject.
	[oop = nil] whileFalse:
		[((self isContextNonImm: oop)
		  and: [aMethodObj = (self fetchPointer: MethodIndex ofObject: oop)]) ifTrue:
			[self interpreter
				printHex: oop; space; printOopShort: oop; print: ' pc ';
				printHex: (self fetchPointer: InstructionPointerIndex ofObject: oop); cr].
		 oop := self accessibleObjectAfter: oop]! !

!ObjectMemory methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:03'!
printContextReferencesTo: anOop
	"Scan the heap printing the oops of any and all contexts that refer to anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32"
	| oop i |
	oop := self firstAccessibleObject.
	[oop = nil] whileFalse:
		[(self isContextNonImm: oop) ifTrue:
			[i := CtxtTempFrameStart + (self fetchStackPointerOf: oop)].
			 [(i := i - 1) >= 0] whileTrue:
				[anOop = (self fetchPointer: i ofObject: oop) ifTrue:
					[self interpreter printHex: oop; print: ' @ '; printNum: i; space; printOopShort: oop; cr.
					 i := 0]].
		 oop := self accessibleObjectAfter: oop]! !

!ObjectMemory methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:03'!
printInstancesOf: aClassOop
	"Scan the heap printing the oops of any and all objects that are instances of aClassOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32"
	| oop |
	oop := self firstAccessibleObject.
	[oop = nil] whileFalse:
		[(self fetchClassOfNonImm: oop) = aClassOop ifTrue:
			[self printHex: oop; cr].
		 oop := self accessibleObjectAfter: oop]! !

!ObjectMemory methodsFor: 'debug printing' stamp: 'eem 04/13/2021 18:01'!
printMethodImplementorsOf: anOop
	"Scan the heap printing the oops of any and all methods that implement anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32"
	| obj |
	obj := self firstAccessibleObject.
	[obj = nil] whileFalse:
		[((self isCompiledMethod: obj)
		  and: [(self maybeSelectorOfMethod: obj) = anOop]) ifTrue:
			[self printHex: obj; space; printOopShort: obj; cr]]! !

!ObjectMemory methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:03'!
printMethodReferencesTo: anOop
	"Scan the heap printing the oops of any and all objects that refer to anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32"
	| oop i |
	oop := self firstAccessibleObject.
	[oop = nil] whileFalse:
		[(self isCompiledMethod: oop) ifTrue:
			[i := (self literalCountOf: oop) + LiteralStart - 1.
			 [i >= 0] whileTrue:
				[anOop = (self fetchPointer: i ofObject: oop) ifTrue:
					[self printHex: oop; print: ' @ '; printNum: i; cr.
					 i := 0].
				 i := i - 1]].
		 oop := self accessibleObjectAfter: oop]! !

!ObjectMemory methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:03'!
printReferencesTo: anOop
	"Scan the heap printing the oops of any and all objects that refer to anOop"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32"
	| oop i |
	oop := self firstAccessibleObject.
	[oop = nil] whileFalse:
		[((self isPointersNonImm: oop) or: [self isCompiledMethod: oop]) ifTrue:
			[(self isCompiledMethod: oop)
				ifTrue:
					[i := (self literalCountOf: oop) + LiteralStart]
				ifFalse:
					[(self isContextNonImm: oop)
						ifTrue: [i := CtxtTempFrameStart + (self fetchStackPointerOf: oop)]
						ifFalse: [i := self lengthOf: oop]].
			[(i := i - 1) >= 0] whileTrue:
				[anOop = (self fetchPointer: i ofObject: oop) ifTrue:
					[self interpreter printHex: oop; print: ' @ '; printNum: i; space; printOopShort: oop; cr.
					 i := 0]]].
		 oop := self accessibleObjectAfter: oop]! !

!ObjectMemory methodsFor: 'debug printing' stamp: 'eem 4/13/2021 18:04'!
printWronglySizedContexts
	"Scan the heap printing the oops of any and all contexts whose size is not either SmallContextSize or LargeContextSize"
	<export: true> "useful for VM debugging; use export: not api, so it will be accessible on win32"
	| oop |
	oop := self firstAccessibleObject.
	[oop = nil] whileFalse:
		[((self isContextNonImm: oop)
		   and: [self badContextSize: oop]) ifTrue:
			[self printHex: oop; space; printNum: (self numBytesOf: oop); cr].
		 oop := self accessibleObjectAfter: oop]! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 2/28/2012 09:08'!
pushRemappableOop: oop
	"Record the given object in a the remap buffer. Objects in this buffer are remapped when a compaction occurs. This facility is used by the interpreter to ensure that objects in temporary variables are properly remapped."
	<api>
	self assert: (self addressCouldBeOop: oop).
	remapBuffer at: (remapBufferCount := remapBufferCount + 1) put: oop.
	remapBufferCount <= RemapBufferSize ifFalse:
		[self error: 'remapBuffer overflow'].! !

!ObjectMemory methodsFor: 'image save/restore' stamp: 'EstebanLorenzano 5/20/2015 12:23'!
readHeapFromImageFile: f dataBytes: numBytes
	"Read numBytes of image data from f into memory at memoryBaseForImageRead.
	 Answer the number of bytes written."
	<var: #f type: #sqImageFile>
	^self cCode: [
		self cppIf: PharoVM
			ifTrue: [  
				self
					sq: (self pointerForOop: self memoryBaseForImageRead)
					Image: (self sizeof: #char)
					File: numBytes
					ReadEntireImage: f ]
			ifFalse: [ 
				self
					sq: (self pointerForOop: self memoryBaseForImageRead)
					Image: (self sizeof: #char)
					File: numBytes
					Read: f] ]
		inSmalltalk:
			[(f	readInto: memory
				startingAt: self memoryBaseForImageRead // 4 + 1
				count: numBytes // 4)
			 * 4]! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 6/1/2009 18:33'!
remap: oop
	<api>
	"Map the given oop to its new value during a compaction or 
	become: operation. If it has no forwarding table entry, 
	return the oop itself."
	<inline: false>
	(self isObjectForwarded: oop) ifTrue:
		[^self remappedObj: oop].
	^ oop! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
remapBuffer
	^remapBuffer! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
remapBuffer: aValue
	^remapBuffer := aValue! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 9/11/2010 15:50'!
remapBufferCount
	<cmacro: '() GIV(remapBufferCount)'>
	^remapBufferCount! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
remapClassOf: oop 
	"Update the class of the given object, if necessary, using its forwarding table entry."
	"Note: Compact classes need not be remapped since the compact class field is just an index into the compact class 
	table. The header type bits show if this object has a compact class; we needn't look up the oop's real header."
	| classHeader classOop newClassOop newClassHeader |
	(self headerType: oop) = HeaderTypeShort ifTrue: [^ nil]. "compact classes needn't be mapped"

	classHeader := self longAt: oop - self wordSize.
	classOop := classHeader bitAnd: AllButTypeMask.
	(self isObjectForwarded: classOop) ifTrue:
		[newClassOop := self remappedObj: classOop.
		newClassHeader := newClassOop bitOr: (classHeader bitAnd: TypeMask).
		self longAt: oop - self wordSize put: newClassHeader.
		"The following ensures that become: into an old object's class makes it a root. 
		It does nothing during either incremental or full compaction because 
		oop will never be < youngStart."
		((self oop: oop isLessThan: youngStart)
				and: [self oop: newClassOop isGreaterThanOrEqualTo: youngStart])
			ifTrue: [self beRootWhileForwarding: oop]]! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
remapFieldsAndClassOf: oop 
	"Replace all forwarded pointers in this object with their new oops, using the forwarding table. Remap its class as well, if 
	necessary. "
	"Note: The given oop may be forwarded itself, which means that its real header is in its forwarding table entry."
	| fieldOffset fieldOop newOop |
	<inline: true>
	fieldOffset := self lastPointerWhileForwarding: oop.
	[fieldOffset >= self baseHeaderSize] whileTrue:
		[fieldOop := self longAt: oop + fieldOffset.
		(self isObjectForwarded: fieldOop) ifTrue:
			["update this oop from its forwarding block"
			newOop := self remappedObj: fieldOop.
			self longAt: oop + fieldOffset put: newOop.
			"The following ensures that become: into old object makes it a root. 
			It does nothing during either incremental or full compaction because 
			oop will never be < youngStart."
			((self oop: oop isLessThan: youngStart)
					and: [self oop: newOop isGreaterThanOrEqualTo: youngStart])
				ifTrue: [self beRootWhileForwarding: oop]].
		fieldOffset := fieldOffset - self wordSize].
	self remapClassOf: oop! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 9/23/2013 06:31'!
remapObj: obj
	"Map the given oop to its new value during a compaction or become: operation."
	<inline: false>
	^self remappedObj: obj! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 10/8/2013 16:54'!
remapRemapBufferAndExtraRoots
	| oop |
	1 to: remapBufferCount do:
		[:i |
		oop := remapBuffer at: i.
		(self shouldRemapOop: oop) ifTrue:
			[remapBuffer at: i put: (self remapObj: oop)]].
	1 to: extraRootCount do:
		[:i |
		oop := (extraRoots at: i) at: 0.
		((self isIntegerObject: oop) or: [self isFreeObject: oop]) ifFalse:
			[(self shouldRemapObj: oop) ifTrue:
				[(extraRoots at: i) at: 0 put: (self remapObj: oop)]]]! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 8/26/2009 11:52'!
remappedObj: forwardedObj
	"Answer the given forwardedOop's target value
	 during a compaction or become: operation."
	| fwdBlock targetObj |
	<inline: true>
	fwdBlock := self forwardingPointerOf: forwardedObj.
	self assert: (self fwdBlockValid: fwdBlock).
	targetObj := self longAt: fwdBlock.
	self assert: (self addressCouldBeObjWhileForwarding: targetObj).
	^targetObj! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'eem 1/12/2013 18:25'!
removeGCRoot: varLoc
	"Remove the given variable location to the extra roots table"
	<export: true>
	<var: #varLoc type: #'sqInt *'>
	1 to: extraRootCount do:
		[:i|
		varLoc = (extraRoots at: i) ifTrue: "swap varLoc with last entry"
			[extraRoots at: i put: (extraRoots at: extraRootCount).
			 extraRootCount := extraRootCount-1.
			 ^true]].
	^false "not found"! !

!ObjectMemory methodsFor: 'become' stamp: 'eem 1/18/2013 10:18'!
removeYoungRoot: obj
	"Remove the given young root from the root table (for freeObject: for becomeForward:)."
	<inline: false>
	1 to: rootTableCount do:
		[:i|
		obj = (rootTable at: i) ifTrue:"swap obj with last entry"
			[rootTable at: i put: (rootTable at: rootTableCount).
			rootTableCount := rootTableCount - 1.
			^true]].
	^false "not found"! !

!ObjectMemory methodsFor: 'become' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
restoreHeaderOf: obj
	"Restore the original header of the given obj from its forwarding block."
	<inline: true> "for subclasses"
	| fwdHeader fwdBlock objHeader |
	fwdHeader := self longAt: obj.
	fwdBlock := (fwdHeader bitAnd: AllButMarkBitAndTypeMask) << 1.
	self assert: (fwdHeader bitAnd: MarkBit) ~= 0.
	self assert: (self fwdBlockValid: fwdBlock).
	objHeader := self longAt: fwdBlock + self wordSize.
	self restoreHeaderOf: obj to: objHeader! !

!ObjectMemory methodsFor: 'become' stamp: 'eem 2/20/2013 13:39'!
restoreHeaderOf: obj to: objHeader
	"helper for restoreHeaderOf: for subclasses to override"
	<inline: true> "for subclasses"
	self longAt: obj put: objHeader! !

!ObjectMemory methodsFor: 'become' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
restoreHeadersAfterBecoming: list1 with: list2 
	"Restore the headers of all oops in both lists. Exchange their hash bits so
	becoming objects in identity sets and dictionaries doesn't change their
	hash value."
	"See also prepareForwardingTableForBecoming:with:woWay: for notes
	regarding the case
	of oop1 = oop2"
	| fieldOffset oop1 oop2 hdr1 hdr2 |
	fieldOffset := self lastPointerOf: list1.
	[fieldOffset >= self baseHeaderSize]
		whileTrue: [oop1 := self longAt: list1 + fieldOffset.
			oop2 := self longAt: list2 + fieldOffset.
			oop1 = oop2
				ifFalse: [self restoreHeaderOf: oop1.
					self restoreHeaderOf: oop2.
					"Exchange hash bits of the two objects."
					hdr1 := self longAt: oop1.
					hdr2 := self longAt: oop2.
					self
						longAt: oop1
						put: ((hdr1 bitAnd: AllButHashBits) bitOr: (hdr2 bitAnd: HashBits)).
					self
						longAt: oop2
						put: ((hdr2 bitAnd: AllButHashBits) bitOr: (hdr1 bitAnd: HashBits))].
			fieldOffset := fieldOffset - self wordSize]! !

!ObjectMemory methodsFor: 'become' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
restoreHeadersAfterForwardBecome: copyHashFlag
	"Forward become leaves us with no original oops in the
	 mutated object list, so we must enumerate the (four-word)
	 forwarding blocks where we have stored backpointers."
	"This loop start is copied from fwdTableInit:"
	| oop1 fwdBlock oop2 hdr1 hdr2 |
	fwdBlock := endOfMemory + self baseHeaderSize + 7 bitAnd: WordMask - 7.
	self flag: #Dan.  "See flag comment in fwdTableInit: (dtl)"
	fwdBlock := fwdBlock + (self wordSize*4).
	"fwdBlockGet: did a pre-increment"
	[self oop: fwdBlock isLessThanOrEqualTo: fwdTableNext
	"fwdTableNext points to the last active block"]
		whileTrue:
			[oop1 := self longAt: fwdBlock + (self wordSize*2).
			"Backpointer to mutated object."
			oop2 := self longAt: fwdBlock.
			self restoreHeaderOf: oop1.
			copyHashFlag ifTrue:
				"Change the hash of the new oop (oop2) to be that of the old (oop1)
				 so mutated objects in hash structures will be  happy after the change."
				[hdr1 := self longAt: oop1.
				 hdr2 := self longAt: oop2.
				 self longAt: oop2 put: ((hdr2 bitAnd: AllButHashBits) bitOr: (hdr1 bitAnd: HashBits))].
			"Free the old object so it won't resurface through e.g. alInstances or allObjects."
			self freeObject: oop1.
			fwdBlock := fwdBlock + (self wordSize*4)]! !

!ObjectMemory methodsFor: 'image segment in/out' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
restoreHeadersFrom: firstIn to: lastIn from: hdrBaseIn and: firstOut to: lastOut from: hdrBaseOut

	"Restore headers smashed by forwarding links"
	| tablePtr oop header |
	tablePtr := firstIn.
	[self oop: tablePtr isLessThanOrEqualTo: lastIn] whileTrue:
		[oop := self longAt: tablePtr.
		header := self longAt: hdrBaseIn + (tablePtr-firstIn).
		self longAt: oop put: header.
		tablePtr := tablePtr + self wordSize].
	tablePtr := firstOut.
	[self oop: tablePtr isLessThanOrEqualTo: lastOut] whileTrue:
		[oop := self longAt: tablePtr.
		header := self longAt: hdrBaseOut + (tablePtr-firstOut).
		self longAt: oop put: header.
		tablePtr := tablePtr + self wordSize].
	
	"Clear all mark bits"
	oop := self firstObject.
	[self oop: oop isLessThan: endOfMemory] whileTrue:
		[(self isFreeObject: oop) ifFalse:
			[self longAt: oop put: ((self longAt: oop) bitAnd: AllButMarkBit)].
		oop := self objectAfter: oop]! !

!ObjectMemory methodsFor: 'image save/restore' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
reverseBytesFrom: startAddr to: stopAddr
	"Byte-swap the given range of memory (not inclusive of stopAddr!!)."
	| addr |
	addr := startAddr.
	[self oop: addr isLessThan: stopAddr] whileTrue:
		[self longAt: addr put: (self byteSwapped: (self longAt: addr)).
		addr := addr + self wordSize].! !

!ObjectMemory methodsFor: 'image save/restore' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
reverseWordsFrom: startAddr to: stopAddr
	"Word-swap the given range of memory, excluding stopAddr."

	| addr |
	addr := startAddr.
	[self oop: addr isLessThan: stopAddr] whileTrue:
		[self longAt: addr put: (self wordSwapped: (self longAt: addr)).
		addr := addr + self wordSize].! !

!ObjectMemory methodsFor: 'header access' stamp: 'tpr 3/24/2004 21:49'!
rightType: headerWord
	"Compute the correct header type for an object based on the size and compact class fields of the given base header word, rather than its type bits. This is used during marking, when the header type bits are used to record the state of tracing."

	(headerWord bitAnd: SizeMask) = 0  "zero size field in header word"
		ifTrue: [ ^HeaderTypeSizeAndClass ]
		ifFalse: [ (headerWord bitAnd: CompactClassMask) = 0
				ifTrue: [ ^HeaderTypeClass ]
				ifFalse: [ ^HeaderTypeShort ]].! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
rootTable
	^rootTable! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 7/19/2014 12:39'!
rootTableCapacity
	^RootTableSize! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
rootTableCount
	^rootTableCount! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
rootTableCount: aValue
	^rootTableCount := aValue! !

!ObjectMemory methodsFor: 'primitive support' stamp: 'eem 1/18/2013 10:05'!
rootTableObject
	"Answer an object containing the contents of the rootTable for primitiveRootTable.
	 The allocation can cause a GC itself and if so the return value may be inaccurate
	 - in this case one should guard the read operation by looking at the gc counter statistics."
	| tableObj sz j |
	sz := rootTableCount.
	tableObj := self instantiateClass: self classArray indexableSize: sz. "can cause GC (and hence reduce number of roots)"
	j := 0.
	1 to: rootTableCount do:
		[:i| "By definition the roots are old and being new, tableObj is young so there is no need to store check."
		self storePointerUnchecked: j ofObject: tableObj withValue: (rootTable at: i)].
	rootTableCount < sz ifTrue:
		[self shorten: tableObj toIndexableSize: rootTableCount].
	^tableObj! !

!ObjectMemory methodsFor: 'debug printing' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
safePrintStringOf: oop
	"Version of printStringOf: that copes with forwarding during garbage collection."
	| fmt header cnt i |
	<inline: false>
	(self isIntegerObject: oop) ifTrue:
		[^nil].
	(self oop: oop isGreaterThanOrEqualTo: self startOfMemory andLessThan: freeBlock) ifFalse:
		[^nil].
	(oop bitAnd: (self wordSize - 1)) ~= 0 ifTrue:
		[^nil].
	header := self headerWhileForwardingOf: oop.
	fmt := self formatOfHeader: header.
	fmt < self firstByteFormat ifTrue: [^nil].

	cnt := 100 min: (self lengthOf: oop baseHeader: header format: fmt).
	i := 0.

	[i < cnt] whileTrue:
		[self printChar: (self fetchByte: i ofObject: oop).
		 i := i + 1].
	self flush.
	^oop! !

!ObjectMemory methodsFor: 'initialization' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
setEndOfMemory: newEndOfMemory
	self assert: (newEndOfMemory bitAnd: self wordSize - 1) = 0.
	endOfMemory := newEndOfMemory! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 6/23/2012 18:52'!
setHashBitsOf: oop to: hash
	self longAt: oop
		put: (((self baseHeader: oop) bitClear: HashBits)
				bitOr: (hash bitAnd: HashMaskUnshifted) << HashBitsOffset)! !

!ObjectMemory methodsFor: 'initialization' stamp: 'eem 7/1/2021 14:50'!
setHeapBase: baseOfHeap memoryLimit: memLimit endOfMemory: memEnd
	"Set the dimensions of the heap, answering the start of oldSpace."
	memory := baseOfHeap.
	self setMemoryLimit: memLimit.
	self setEndOfMemory: memEnd.
	oldSpaceStart := baseOfHeap.
	^baseOfHeap! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem 3/27/2014 15:14'!
setIsMarkedOf: objOop to: aBoolean
	| header |
	self assert: (self isFreeObject: objOop) not.
	header := self baseHeader: objOop.
	self
		baseHeader: objOop
		put: (aBoolean
				ifTrue: [header bitOr: MarkBit]
				ifFalse: [(header bitOr: MarkBit) - MarkBit])! !

!ObjectMemory methodsFor: 'initialization' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
setMemoryLimit: newMemoryLimit
	self assert: (newMemoryLimit bitAnd: self wordSize - 1) = 0.
	memoryLimit := newMemoryLimit! !

!ObjectMemory methodsFor: 'header access' stamp: 'go 11/13/1998 17:04'!
setSizeOfFree: chunk to: byteSize
	"Set the header of the given chunk to make it be a free chunk of the given size."

	self longAt: chunk put: ((byteSize bitAnd: AllButTypeMask) bitOr: HeaderTypeFree).! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 12/13/2015 10:16'!
shiftForWord
	"N.B. This would appear to hard-code the header size for 32-bit images.  But if generating
	 a 64-bit image, this method could be removed and the relevant one substituted.  We can't
	 mark this method as <doNotGenerate> as we need an actual method to guide code gen."
	<api>
	<cmacro>
	^2! !

!ObjectMemory methodsFor: 'allocation' stamp: 'cb 5/22/2017 16:54'!
shorten: obj toIndexableSize: nSlots
	"Reduce the number of indexable fields in obj, a pointer object, to nSlots. Convert the
	unused residual to a free chunk. Word and byte indexable objects are not changed.
	Answer the number of bytes returned to free memory, which may be zero if no change
	was possible."
	| deltaBytes desiredLength fixedFields fmt hdr totalLength
	 indexableFields |
	<api>
	(self isPointersNonImm: obj) ifFalse: [^0].
	nSlots >  0
		ifFalse: [^0]. "no change if nSlots is zero, error if nSlots is negative"
	hdr := self baseHeader: obj.
	fmt := self formatOfHeader: hdr.
	totalLength := self lengthOf: obj baseHeader: hdr format: fmt.
	fixedFields := self fixedFieldsOf: obj format: fmt length: totalLength.
	indexableFields := totalLength - fixedFields.
	nSlots >= indexableFields
		ifTrue: [^0]. "no change, or error if attempting to increase size into next chunk"
	desiredLength := fixedFields + nSlots.		
	deltaBytes := (totalLength - desiredLength) * self wordSize.
	self setSizeOfFree: obj + self baseHeaderSize + (desiredLength * self wordSize)
		to: deltaBytes.
	(self headerType: obj) caseOf:	{
		[HeaderTypeSizeAndClass] ->
			[self longAt: (obj - (self baseHeaderSize * 2)) put: (self sizeHeader: obj) - deltaBytes].
		[HeaderTypeClass] ->
			[self longAt: obj put: ((hdr bitClear: SizeMask) bitOr: (hdr bitAnd: SizeMask) - deltaBytes)].
		[HeaderTypeShort] ->
			[self longAt: obj put: ((hdr bitClear: SizeMask) bitOr: (hdr bitAnd: SizeMask) - deltaBytes)] }.
	^deltaBytes! !

!ObjectMemory methodsFor: 'debug support' stamp: 'eem 6/5/2017 11:18'!
shouldBreakForLookupIn: lookupClass given: breakClassTag
	<inline: true>
	^lookupClass == breakClassTag! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 9/14/2013 09:58'!
shouldRemapObj: oop
	<api>
	"Answer if the oop should be remapped"
	<inline: true>
	^self isObjectForwarded: oop! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'eem 9/14/2013 09:58'!
shouldRemapOop: oop
	<api>
	"Answer if the oop should be remapped"
	<inline: true>
	^(self isNonIntegerObject: oop)
	  and: [self isObjectForwarded: oop]! !

!ObjectMemory methodsFor: 'allocation' stamp: 'eem 4/4/2009 20:09'!
shrinkObjectMemory: delta 
	"Attempt to shrink the object memory by the given delta 
	amount "
	| limit |
	statShrinkMemory := statShrinkMemory + 1. 
	limit := self sqShrinkMemory: memoryLimit By: delta.
	limit = memoryLimit ifFalse:
		[self setMemoryLimit: limit - 24. "remove a tad for safety"
		 self initializeMemoryFirstFree: freeBlock]! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
shrinkThreshold
	^shrinkThreshold! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
shrinkThreshold: aValue
	^shrinkThreshold := aValue! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
signalLowSpace
	^signalLowSpace! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
signalLowSpace: aValue
	^signalLowSpace := aValue! !

!ObjectMemory methodsFor: 'header formats' stamp: 'eem 12/23/2014 17:25'!
sixtyFourBitIndexableFormat
	^7! !

!ObjectMemory methodsFor: 'object access' stamp: 'di 6/11/2004 16:34'!
sizeBitsOf: oop
	"Answer the number of bytes in the given object, including its base header, rounded up to an integral number of words."
	"Note: byte indexable objects need to have low bits subtracted from this size."

	| header |
	header := self baseHeader: oop.
	(header bitAnd: TypeMask) = HeaderTypeSizeAndClass
		ifTrue: [ ^ (self sizeHeader: oop) bitAnd: LongSizeMask ]
		ifFalse: [ ^ header bitAnd: SizeMask ].! !

!ObjectMemory methodsFor: 'object access' stamp: 'go 11/17/1998 15:57'!
sizeBitsOfSafe: oop
	"Compute the size of the given object from the cc and size fields in its header. This works even if its type bits are not correct."

	| header type |
	header := self baseHeader: oop.
	type := self rightType: header.
	type = HeaderTypeSizeAndClass
		ifTrue: [ ^ (self sizeHeader: oop) bitAnd: AllButTypeMask ]
		ifFalse: [ ^ header bitAnd: SizeMask ].! !

!ObjectMemory methodsFor: 'header access' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
sizeHeader: oop

	^ self longAt: oop - (self wordSize*2)! !

!ObjectMemory methodsFor: 'header access' stamp: 'go 11/13/1998 17:04'!
sizeOfFree: oop
	"Return the size of the given chunk in bytes. Argument MUST be a free chunk."

	^ (self longAt: oop) bitAnd: AllButTypeMask! !

!ObjectMemory methodsFor: 'object access' stamp: 'ar 10/10/1998 21:24'!
slotSizeOf: oop
	"Returns the number of slots in the receiver.
	If the receiver is a byte object, return the number of bytes.
	Otherwise return the number of words."
	(self isIntegerObject: oop) ifTrue:[^0].
	^self lengthOf: oop! !

!ObjectMemory methodsFor: 'cog jit support' stamp: 'eem 12/21/2008 19:32'!
specialObjectsOop
	^specialObjectsOop! !

!ObjectMemory methodsFor: 'cog jit support' stamp: 'eem 10/12/2009 08:32'!
specialObjectsOop: anObject
	specialObjectsOop := anObject! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 12/20/2012 11:02'!
splObj: index
	<api>
	<inline: true>
	"Return one of the objects in the SpecialObjectsArray"
	^self fetchPointer: index ofObject: specialObjectsOop! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 8/26/2010 15:45'!
splObj: index put: anObject
	"Set one of the objects in the SpecialObjectsArray"
	self storePointer: index ofObject: specialObjectsOop withValue: anObject! !

!ObjectMemory methodsFor: 'gc -- mark and sweep' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
startField
	"Examine and possibly trace the next field of the object being
	 traced. See comment in markAndTrace for explanation of
	 tracer state variables, and aComment for gory details."
	| typeBits childType |
	<inline: true>
	[child := self longAt: field.
	 self isIntegerObject: child] whileTrue:
		["field contains a SmallInteger; skip it"
		field := field - self wordSize.
		"Effectively ^ StartField"].
	typeBits := child bitAnd: TypeMask.
	typeBits = 0 ifTrue: "normal oop, go down"
		[self longAt: field put: parentField.
		parentField := field.
		^ StartObj].
	self assert: typeBits = 2.
	"reached the header; do we need to process the class word? "
	(child bitAnd: CompactClassMask) ~= 0 ifTrue:
		["object's class is compact; we're done"
		"restore the header type bits"
		child := child bitAnd: AllButTypeMask.
		childType := self rightType: child.
		self longAt: field put: (child bitOr: childType).
		^ Upward].
	"object has a full class word; process that class"
	child := self longAt: field - self wordSize. "class word"
	child := child bitAnd: AllButTypeMask. "clear type bits"
	self longAt: field - self wordSize put: parentField.
	parentField := field - self wordSize bitOr: 1.
	"point at class word; mark as working on the class. "
	^ StartObj! !

!ObjectMemory methodsFor: 'gc -- mark and sweep' stamp: 'eem 12/20/2008 08:31'!
startObj
	"Start tracing the object 'child' and answer the next action. 
	 The object may be anywhere in the middle of being swept 
	 itself. See comment in markAndTrace for explanation of 
	 tracer state variables, and aComment for gory details."
	| header lastFieldOffset |
	<inline: true>
	statMarkCountLocal := statMarkCountLocal + 1.
	((self oop: child isLessThan: youngStartLocal) "old object; skip it"
	 or: [((header := self longAt: child) bitAnd: MarkBit) ~= 0]) "already marked; skip it"
		ifTrue:
			[field := child.
			^ Upward].
	"unmarked; mark and trace"
	lastFieldOffset := self lastPointerOf: child recordWeakRoot: false.
	header := (header bitAnd: AllButTypeMask) bitOr: (MarkBit bitOr: HeaderTypeGC).
	self longAt: child put: header.
	field := child + lastFieldOffset.
	^ StartField "trace its fields and class"! !

!ObjectMemory methodsFor: 'memory access' stamp: 'eem 4/30/2009 10:20'!
startOfFreeSpace
	<inline: true>
	^freeBlock! !

!ObjectMemory methodsFor: 'object enumeration' stamp: 'eem 12/24/2013 20:51'!
startOfMemory
	"Return the start of object memory. Use a macro so as not to punish the debug VM."
	<cmacro: '() GIV(memory)'>
	<returnTypeC: #usqInt>
	^memory! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statCompMoveCount
	^statCompMoveCount! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statFullGCUsecs
	^statFullGCUsecs! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statFullGCs
	^statFullGCs! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statGrowMemory
	^statGrowMemory! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statIGCDeltaUsecs
	^statIGCDeltaUsecs! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statIncrGCUsecs
	^statIncrGCUsecs! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statIncrGCs
	^statIncrGCs! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statMarkCount
	^statMarkCount! !

!ObjectMemory methodsFor: 'accessing' stamp: 'cb 7/3/2018 10:27'!
statMarkUsecs
	^ 0! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statMkFwdCount
	^statMkFwdCount! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 3/14/2017 13:58'!
statNumGCs
	"Part of InterpreterProxy's 1.14 API"
	<export: true>
	^statIncrGCs + statFullGCs! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statRootTableCount
	^statRootTableCount! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statRootTableOverflows
	^statRootTableOverflows! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statShrinkMemory
	^statShrinkMemory! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statSpecialMarkCount
	^statSpecialMarkCount! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statSurvivorCount
	^statSurvivorCount! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statSweepCount
	^statSweepCount! !

!ObjectMemory methodsFor: 'accessing' stamp: 'cb 7/3/2018 10:27'!
statSweepUsecs
	^ 0! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
statTenures
	^statTenures! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/15/2017 14:02'!
statTenures: aValue
	statTenures := aValue! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
storeByte: byteIndex ofObject: oop withValue: valueByte

	^ self byteAt: oop + self baseHeaderSize + byteIndex
		put: valueByte! !

!ObjectMemory methodsFor: 'image segment in/out' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
storeImageSegmentInto: segmentWordArray outPointers: outPointerArray roots: arrayOfRoots

"This primitive will store a binary image segment (in the same format as the Squeak image file) of the receiver and every object in its proper tree of subParts (ie, that is not refered to from anywhere else outside the tree).  All pointers from within the tree to objects outside the tree will be copied into the array of outpointers.  In their place in the image segment will be an oop equal to the offset in the outPointer array (the first would be 4). but with the high bit set."

"The primitive expects the array and wordArray to be more than adequately long.  In this case it returns normally, and truncates the two arrays to exactly the right size.  To simplify truncation, both incoming arrays are required to be 256 bytes or more long (ie with 3-word headers).  If either array is too small, the primitive will fail, but in no other case.

During operation of the primitive, it is necessary to convert from both internal and external oops to their mapped values.  To make this fast, the headers of the original objects in question are replaced by the mapped values (and this is noted by adding the forbidden XX header type).  Tables are kept of both kinds of oops, as well as of the original headers for restoration.

To be specific, there are two similar two-part tables, the outpointer array, and one in the upper fifth of the segmentWordArray.  Each grows oops from the bottom up, and preserved headers from halfway up.

In case of either success or failure, the headers must be restored.  In the event of primitive failure, the table of outpointers must also be nilled out (since the garbage in the high half will not have been discarded."

	| savedYoungStart lastOut lastIn firstIn lastSeg endSeg segOop fieldPtr fieldOop mapOop
	  doingClass lastPtr extraSize hdrTypeBits hdrBaseIn hdrBaseOut header firstOut versionOffset |
	<inline: false>
	<var: #firstIn type: #usqInt>
	<var: #lastIn type: #usqInt>
	<var: #firstOut type: #usqInt>
	<var: #lastOut type: #usqInt>
	<var: #hdrBaseIn type: #usqInt>
	<var: #hdrBaseOut type: #usqInt>
	<var: #lastSeg type: #usqInt>
	<var: #endSeg type: #usqInt>
	<var: #fieldPtr type: #usqInt>
	<var: #lastPtr type: #usqInt>
	<var: #segOop type: #usqInt>
	<var: #savedYoungStart type: #usqInt>

	((self headerType: outPointerArray) = HeaderTypeSizeAndClass			"Must be 3-word header"
	and: [(self headerType: segmentWordArray) = HeaderTypeSizeAndClass])	"Must be 3-word header"
		ifFalse: [^PrimErrGenericFailure].

	DoAssertionChecks ifTrue: [self verifyCleanHeaders].
	"Use the top half of outPointers for saved headers."
	firstOut := outPointerArray + self baseHeaderSize.
	lastOut := firstOut - self wordSize.
	hdrBaseOut := outPointerArray + ((self lastPointerOf: outPointerArray) // (self wordSize*2) * self wordSize). "top half"

	lastSeg := segmentWordArray.
	endSeg := segmentWordArray + (self sizeBitsOf: segmentWordArray) - self wordSize.

	"Write a version number for byte order and version check"
	versionOffset := self wordSize.
	lastSeg := lastSeg + versionOffset.
	lastSeg > endSeg ifTrue: [^PrimErrGenericFailure].
	self longAt: lastSeg put: self imageSegmentVersion.

	"Allocate top 1/8 of segment for table of internal oops and saved headers"
	firstIn := endSeg - ((self sizeBitsOf: segmentWordArray) // (self wordSize*8) * self wordSize).  "Take 1/8 of seg"
	lastIn := firstIn - self wordSize.
	hdrBaseIn := firstIn + ((self sizeBitsOf: segmentWordArray) // (self wordSize*16) * self wordSize). "top half of that"

	"First mark the rootArray and all root objects."
	self longAt: arrayOfRoots put: ((self longAt: arrayOfRoots) bitOr: MarkBit).
	lastPtr := arrayOfRoots + (self lastPointerOf: arrayOfRoots).
	fieldPtr := arrayOfRoots + self baseHeaderSize.
	[fieldPtr <= lastPtr] whileTrue:
		[fieldOop := self longAt: fieldPtr.
		(self isIntegerObject: fieldOop) ifFalse:
			[self longAt: fieldOop put: ((self longAt: fieldOop) bitOr: MarkBit)].
		fieldPtr := fieldPtr + self wordSize].

	"Then do a mark pass over all objects.  This will stop at our marked roots,
	thus leaving our segment unmarked in their shadow."
	savedYoungStart := youngStart.
	youngStart := self startOfMemory.  "process all of memory"
	self markAndTraceInterpreterOops.	"and special objects array"
	youngStart := savedYoungStart.
	
	"Finally unmark the rootArray and all root objects."
	self longAt: arrayOfRoots put: ((self longAt: arrayOfRoots) bitAnd: AllButMarkBit).
	fieldPtr := arrayOfRoots + self baseHeaderSize.
	[fieldPtr <= lastPtr] whileTrue:
		[fieldOop := self longAt: fieldPtr.
		(self isIntegerObject: fieldOop) ifFalse:
			[self longAt: fieldOop put: ((self longAt: fieldOop) bitAnd: AllButMarkBit)].
		fieldPtr := fieldPtr + self wordSize].

	"All external objects, and only they, are now marked.
	Copy the array of roots into the segment, and forward its oop."
	lastIn := lastIn + self wordSize.
	(lastIn >= hdrBaseIn
	 or: [0 = (lastSeg := self copyObj: arrayOfRoots toSegment: segmentWordArray addr: lastSeg stopAt: firstIn saveOopAt: lastIn headerAt: hdrBaseIn + (lastIn - firstIn))]) ifTrue:
		[lastIn := lastIn - self wordSize.
		self restoreHeadersFrom: firstIn to: lastIn from: hdrBaseIn and: firstOut to: lastOut from: hdrBaseOut.
		^self primitiveFailCodeAfterCleanup: outPointerArray].

	"Now run through the segment fixing up all the pointers.
	Note that more objects will be added to the segment as we make our way along."
	segOop := self oopFromChunk: segmentWordArray + versionOffset + self baseHeaderSize.
	[segOop <= lastSeg] whileTrue:
		[(self headerType: segOop) <= 1
			ifTrue: ["This object has a class field (type=0 or 1) -- start with that."
					fieldPtr := segOop - self wordSize.  doingClass := true]
			ifFalse: ["No class field -- start with first data field"
					fieldPtr := segOop + self baseHeaderSize.  doingClass := false].
		lastPtr := segOop + (self lastPointerOf: segOop).	"last field"

		"Go through all oops, remapping them..."
		[fieldPtr > lastPtr] whileFalse:
			["Examine each pointer field"
			fieldOop := self longAt: fieldPtr.
			doingClass ifTrue:
				[hdrTypeBits := fieldOop bitAnd: TypeMask.
				fieldOop := fieldOop - hdrTypeBits].
			(self isIntegerObject: fieldOop)
				ifTrue: ["Just an integer -- nothing to do"
						fieldPtr := fieldPtr + self wordSize]
				ifFalse:
				[header := self longAt: fieldOop.
				(header bitAnd: TypeMask) = HeaderTypeFree
					ifTrue: ["Has already been forwarded -- this is the link"
							mapOop := header bitAnd: AllButTypeMask]
					ifFalse:
					[((self longAt: fieldOop) bitAnd: MarkBit) = 0
						ifTrue:
							["Points to an unmarked obj -- an internal pointer.
							Copy the object into the segment, and forward its oop."
							lastIn := lastIn + self wordSize.
							(lastIn >= hdrBaseIn
							or: [0 = (lastSeg := self copyObj: fieldOop toSegment: segmentWordArray addr: lastSeg stopAt: firstIn saveOopAt: lastIn headerAt: hdrBaseIn + (lastIn - firstIn))]) ifTrue:
								["Out of space in segment"
								lastIn := lastIn - self wordSize.
								self restoreHeadersFrom: firstIn to: lastIn from: hdrBaseIn and: firstOut to: lastOut from: hdrBaseOut.
								^self primitiveFailCodeAfterCleanup: outPointerArray].
							mapOop := (self longAt: fieldOop) bitAnd: AllButTypeMask]
						ifFalse:
							["Points to a marked obj -- an external pointer.
							Map it as a tagged index in outPointers, and forward its oop."
							lastOut := lastOut + self wordSize.
							lastOut >= hdrBaseOut ifTrue:
								["Out of space in outPointerArray"
								lastOut := lastOut - self wordSize.
								self restoreHeadersFrom: firstIn to: lastIn from: hdrBaseIn and: firstOut to: lastOut from: hdrBaseOut.
								^self primitiveFailCodeAfterCleanup: outPointerArray].
.							mapOop := lastOut - outPointerArray bitOr: 16r80000000.
							self forward: fieldOop to: mapOop
								savingOopAt: lastOut andHeaderAt: hdrBaseOut + (lastOut - firstOut)]].
					"Replace the oop by its mapped value"
					doingClass
						ifTrue: [self longAt: fieldPtr put: mapOop + hdrTypeBits.
								fieldPtr := fieldPtr + (self wordSize*2).
								doingClass := false]
						ifFalse: [self longAt: fieldPtr put: mapOop.
								fieldPtr := fieldPtr + self wordSize]]].
		segOop := self objectAfter: segOop].

	self restoreHeadersFrom: firstIn to: lastIn from: hdrBaseIn and: firstOut to: lastOut from: hdrBaseOut.

	"Truncate the outPointerArray..."
	((outPointerArray + (self lastPointerOf: outPointerArray) - lastOut) < 12
		or: [(endSeg - lastSeg) < 12]) ifTrue:
			["Not enough room to insert simple 3-word headers"
			^self primitiveFailCodeAfterCleanup: outPointerArray].
	extraSize := self extraHeaderBytes: segmentWordArray.
	hdrTypeBits := self headerType: segmentWordArray.
	"Copy the 3-word wordArray header to establish a free chunk."
	self transfer: 3
		from: segmentWordArray - extraSize
		to: lastOut+self wordSize.
	"Adjust the size of the original as well as the free chunk."
	self longAt: lastOut+self wordSize
		put: outPointerArray + (self lastPointerOf: outPointerArray) - lastOut - extraSize + hdrTypeBits.
	self longAt: outPointerArray-extraSize
		put: lastOut - firstOut + (self wordSize*2) + hdrTypeBits.
	"Note that pointers have been stored into roots table"
	self beRootIfOld: outPointerArray.

	"Truncate the image segment..."
	"Copy the 3-word wordArray header to establish a free chunk."
	self transfer: 3
		from: segmentWordArray - extraSize
		to: lastSeg+self wordSize.
	"Adjust the size of the original as well as the free chunk."
	self longAt: segmentWordArray-extraSize
		put: lastSeg - segmentWordArray + self baseHeaderSize + hdrTypeBits.
	self longAt: lastSeg+self wordSize
		put: endSeg - lastSeg - extraSize + hdrTypeBits.

	DoAssertionChecks ifTrue: [self verifyCleanHeaders].
	^PrimNoErr! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
storeLong32: fieldIndex ofObject: oop withValue: valueWord

	^ self long32At: oop + self baseHeaderSize + (fieldIndex << 2)
		put: valueWord! !

!ObjectMemory methodsFor: 'object access' stamp: 'cb 4/17/2015 09:58'!
storeLong64: longIndex ofObject: oop withValue: value
	<var: #value type: #sqLong>
	self flag: #endianness.
	self long32At: oop + self baseHeaderSize + (longIndex << 3) put: (self cCode: [value] inSmalltalk: [value bitAnd: 16rFFFFFFFF]);
		long32At: oop + self baseHeaderSize + (longIndex << 3) + 4 put: (value >> 32).
	^value! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:06'!
storePointer: fieldIndex ofObject: oop withValue: valuePointer
	"Note must check here for stores of young objects into old ones."

	(self oop: oop isLessThan: youngStart) ifTrue: [
		self possibleRootStoreInto: oop value: valuePointer.
	].

	^ self longAt: oop + self baseHeaderSize + (fieldIndex << self shiftForWord)
		put: valuePointer! !

!ObjectMemory methodsFor: 'object access' stamp: 'cb 12/11/2015 14:51'!
storePointerImmutabilityCheck: index ofObject: rcvr withValue: top
	<inline: true>
	^ self storePointer: index ofObject: rcvr withValue: top! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem (const to msg rewrite) 11/13/2014 14:06'!
storePointerUnchecked: fieldIndex ofObject: oop withValue: valuePointer
	"Like storePointer:ofObject:withValue:, but the caller guarantees that the
	 object being stored into is a young object or is already marked as a root."
	<api>
	<inline: true>
	^self
		longAt: oop + self baseHeaderSize + (fieldIndex << self shiftForWord)
		put: valuePointer! !

!ObjectMemory methodsFor: 'object access' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
storeShort16: shortIndex ofObject: objOop withValue: value
	^self
		shortAt: objOop + self baseHeaderSize + (shortIndex << 1)
		put: (self cCode: [value] inSmalltalk: [value bitAnd: 16rFFFF])! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'ikp 3/26/2005 14:29'!
storeWord: fieldIndex ofObject: oop withValue: valueWord
	"This message is deprecated.  Use storeLong32, storeLong64 or storePointer"

	self abort! !

!ObjectMemory methodsFor: 'primitive support' stamp: 'eem 9/27/2020 14:13'!
stringForCString: aCString
	"Answer a new String copied from a null-terminated C string,
	 or nil if out of memory.
	 Caution: This may invoke the garbage collector."
	<api>
	<var: 'aCString' type: #'const char *'>
	<inline: false>
	| len newString |
	len := self strlen: aCString.
	newString := self instantiateClass: (self splObj: ClassByteString) indexableSize: len.
	newString ifNotNil:
		[self strncpy: (self cCoerceSimple: newString + self baseHeaderSize to: #'char *')
			_: aCString
			_: len]. "(char *)strncpy()"
	^newString! !

!ObjectMemory methodsFor: 'allocation' stamp: 'dtl (auto pragmas 12/08) 4/22/2007 20:27'!
sufficientSpaceAfterGC: minFree 
	"Return true if there is enough free space after doing a garbage collection. If not, signal that space is low."
	| growSize |
	<inline: false>

	self incrementalGC. "try to recover some space"

	(self oop: (self sizeOfFree: freeBlock) isLessThan: minFree)
		ifTrue: [signalLowSpace ifTrue: [^false]. "give up; problem is already noted"
			self fullGC. "try harder"
			"for stability, require more free space after doing an expensive full GC"
			(self cCoerce: (self sizeOfFree: freeBlock) to: 'usqInt ') >= ((self cCoerce: minFree to: 'usqInt ') + 15000) ifTrue: [^ true].

			"still not enough; attempt to grow object memory"
			growSize := minFree - (self sizeOfFree: freeBlock) + growHeadroom.
			self growObjectMemory: growSize.

			(self oop: (self sizeOfFree: freeBlock) isGreaterThanOrEqualTo: minFree + 15000)
				ifTrue: [^true].

			"still not enough"
			^false].
	^true! !

!ObjectMemory methodsFor: 'allocation' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
sufficientSpaceToAllocate: bytes
	"Return true if there is enough space to allocate the given number of bytes, perhaps after doing a garbage collection."

	| minFree |
	<inline: true>
	minFree := (lowSpaceThreshold + bytes + self baseHeaderSize + self wordSize - 1) bitClear: self wordSize - 1.

	"check for low-space"
	(self oop: (self sizeOfFree: freeBlock) isGreaterThanOrEqualTo: minFree)
		ifTrue: [^true]
		ifFalse: [^self sufficientSpaceAfterGC: minFree].! !

!ObjectMemory methodsFor: 'allocation' stamp: 'eem (const to msg rewrite) 11/13/2014 14:06'!
sufficientSpaceToInstantiate: classOop indexableSize: size 
	"Return the number of bytes required to allocate an instance of the given class with the given number of indexable fields."
	"Details: For speed, over-estimate space needed for fixed fields or literals; the low space threshold is a blurry line."
	<var: #size type: #usqInt>
	| format allocSize |
	<inline: true>
	(format := self instSpecOfClass: classOop) < self firstByteFormat
		ifTrue:
			["indexable fields are words or pointers"
			size ~= 0 ifTrue:
				["fail if attempting to call new: on non-indexable class"
				 format < self arrayFormat ifTrue:
					[^false].
				 "fail if attempting to allocate over 2 Gb, since this overflows 32-bit arithmetic"
				 size >> (LongSizeNumBits - self shiftForWord) > 0 ifTrue:
					[^false]].
			allocSize := size * self wordSize]
		ifFalse:
			["indexable fields are bytes"
			 "fail if attempting to allocate over 2 Gb, since this overflows 32-bit arithmetic"
			 size >> LongSizeNumBits > 0 ifTrue:
				[^false].
			allocSize := size].
	^self sufficientSpaceToAllocate: 2500 + allocSize! !

!ObjectMemory methodsFor: 'gc -- mark and sweep' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
sweepPhase
	"Sweep memory from youngStart through the end of memory. Free all 
	inaccessible objects and coalesce adjacent free chunks. Clear the mark 
	bits of accessible objects. Compute the starting point for the first pass of 
	incremental compaction (compStart). Return the number of surviving 
	objects. "
	"Details: Each time a non-free object is encountered, decrement the 
	number of available forward table entries. If all entries are spoken for 
	(i.e., entriesAvailable reaches zero), set compStart to the last free 
	chunk before that object or, if there is no free chunk before the given 
	object, the first free chunk after it. Thus, at the end of the sweep 
	phase, compStart through compEnd spans the highest collection of 
	non-free objects that can be accomodated by the forwarding table. This 
	information is used by the first pass of incremental compaction to 
	ensure that space is initially freed at the end of memory. Note that 
	there should always be at least one free chunk--the one at the end of 
	the heap."
	| entriesAvailable survivors freeChunk firstFree oop oopHeader oopHeaderType hdrBytes oopSize freeChunkSize endOfMemoryLocal |
	<inline: false>
	<var: #oop type: 'usqInt'>
	<var: #endOfMemoryLocal type: 'usqInt'>
	entriesAvailable := self fwdTableInit: self wordSize*2.
	survivors := 0.
	freeChunk := nil.
	firstFree := nil.
	"will be updated later"
	endOfMemoryLocal := endOfMemory.
	oop := self oopFromChunk: youngStart.
	[oop < endOfMemoryLocal]
		whileTrue: ["get oop's header, header type, size, and header size"
			statSweepCount := statSweepCount + 1.
			oopHeader := self baseHeader: oop.
			oopHeaderType := oopHeader bitAnd: TypeMask.
			hdrBytes := headerTypeBytes at: oopHeaderType.
			(oopHeaderType bitAnd: 1) = 1
				ifTrue: [oopSize := oopHeader bitAnd: SizeMask]
				ifFalse: [oopHeaderType = HeaderTypeSizeAndClass
						ifTrue: [oopSize := (self sizeHeader: oop) bitAnd: LongSizeMask]
						ifFalse: ["free chunk" oopSize := oopHeader bitAnd: LongSizeMask]].
			(oopHeader bitAnd: MarkBit) = 0
				ifTrue: ["object is not marked; free it"
					"<-- Finalization support: We need to mark each oop chunk as free -->"
					self longAt: oop - hdrBytes put: HeaderTypeFree.
					freeChunk ~= nil
						ifTrue: ["enlarge current free chunk to include this oop"
							freeChunkSize := freeChunkSize + oopSize + hdrBytes]
						ifFalse: ["start a new free chunk"
							freeChunk := oop - hdrBytes.
							"chunk may start 4 or 8 bytes before oop"
							freeChunkSize := oopSize + (oop - freeChunk).
							"adjust size for possible extra header bytes"
							firstFree = nil ifTrue: [firstFree := freeChunk]]]
				ifFalse: ["object is marked; clear its mark bit and possibly adjust 
					the compaction start"
					self longAt: oop put: (oopHeader bitAnd: AllButMarkBit).
					"<-- Finalization support: Check if we're running about a weak class -->"
					(self isWeakNonImm: oop) ifTrue: [self finalizeReference: oop].
					entriesAvailable > 0
						ifTrue: [entriesAvailable := entriesAvailable - 1]
						ifFalse: ["start compaction at the last free chunk before this object"
							firstFree := freeChunk].
					freeChunk ~= nil
						ifTrue: ["record the size of the last free chunk"
							self longAt: freeChunk put: ((freeChunkSize bitAnd: LongSizeMask) bitOr: HeaderTypeFree).
							freeChunk := nil].
					survivors := survivors + 1].
			oop := self oopFromChunk: oop + oopSize].
	freeChunk ~= nil
		ifTrue: ["record size of final free chunk"
			self longAt: freeChunk put: ((freeChunkSize bitAnd: LongSizeMask) bitOr: HeaderTypeFree)].
	oop = endOfMemory
		ifFalse: [self error: 'sweep failed to find exact end of memory'].
	firstFree = nil
		ifTrue: [self error: 'expected to find at least one free object']
		ifFalse: [compStart := firstFree].

	^ survivors! !

!ObjectMemory methodsFor: 'garbage collection' stamp: 'eem 11/25/2009 11:09'!
tenuringIncrementalGC
	"Do an incremental GC that tenures all surviving young objects to old space."
	<api>
	forceTenureFlag := true.
	self incrementalGC! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 6/8/2010 17:12'!
tenuringThreshold
	^tenuringThreshold! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 3/16/2017 13:59'!
tenuringThreshold: aValue
	tenuringThreshold := aValue.
	^0! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 4/2/2009 22:48'!
topRemappableOop
	<api>
	"Returns the top of the remappable oop. Useful when writing loops."
	^remapBuffer at: remapBufferCount! !

!ObjectMemory methodsFor: 'accessing' stamp: 'eem 10/12/2013 09:24'!
totalMemorySize
	^endOfMemory - self startOfMemory! !

!ObjectMemory methodsFor: 'utilities' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
transfer: count from: src to: dst 
	| in out lastIn |
	<inline: true>
	self flag: #Dan.  "Need to check all senders before converting this for 64 bits"
	in := src - self wordSize.
	lastIn := in + (count * self wordSize).
	out := dst - self wordSize.
	[self oop: in isLessThan: lastIn]
		whileTrue: [self
				longAt: (out := out + self wordSize)
				put: (self longAt: (in := in + self wordSize))]! !

!ObjectMemory methodsFor: 'plugin support' stamp: 'eem 3/26/2009 20:30'!
trueObject
	<api>
	^trueObj! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'dtl (auto pragmas 12/08) 4/22/2007 21:30'!
updatePointersInRangeFrom: memStart to: memEnd 
	"update pointers in the given memory range"
	| oop |
	<inline: false>
	oop := self oopFromChunk: memStart.
	[self oop: oop isLessThan: memEnd]
		whileTrue: [(self isFreeObject: oop)
				ifFalse: [self remapFieldsAndClassOf: oop].
			oop := self objectAfterWhileForwarding: oop]! !

!ObjectMemory methodsFor: 'gc -- compaction' stamp: 'dtl (auto pragmas 12/08) 4/22/2007 20:31'!
updatePointersInRootObjectsFrom: memStart to: memEnd 
	"update pointers in root objects"
	| oop |
	<inline: false>
	1 to: rootTableCount do: [:i | 
			oop := rootTable at: i.
			((self oop: oop isLessThan: memStart)
					or: [self oop: oop isGreaterThanOrEqualTo: memEnd])
				ifTrue: ["Note: must not remap the fields of any object twice!!"
					"remap this oop only if not in the memory range 
					covered below"
					self remapFieldsAndClassOf: oop]]! !

!ObjectMemory methodsFor: 'gc -- mark and sweep' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
upward
	"Return from marking an object below. Incoming: 
		field = oop we just worked on, needs to be put away 
		parentField = where to put it in our object 
	NOTE: Type field of object below has already been restored!!!!!!
	 See comment in markAndTrace for explanation of 
	 tracer state variables, and aComment for gory details."
	| type header |
	<inline: true>
	[(parentField bitAnd: 1) = 0 ifTrue:
		["normal"
		child := field. "who we worked on below"
		field := parentField. "where to put it"
		parentField := self longAt: field.
		self longAt: field put: child.
		field := field - self wordSize. "point at header"
		^ StartField].
	 parentField = GCTopMarker ifTrue:
		["top of the chain"
		header := (self longAt: field) bitAnd: AllButTypeMask.
		type := self rightType: header.
		self longAt: field put: (header bitOr: type). "install type on class oop"
		^ Done].
	"was working on the extended class word"
	child := field. "oop of class"
	field := parentField - 1. "class word, ** clear the low bit **"
	parentField := self longAt: field.
	header := self longAt: field + self wordSize. "base header word"
	type := self rightType: header.
	self longAt: field put: (child bitOr: type). "install type on class oop"
	field := field + self wordSize. "point at header"
	"restore type bits"
	header := header bitAnd: AllButTypeMask.
	self longAt: field put: (header bitOr: type).
	"Effectively ^ Upward"
	true] whileTrue.
	^nil "for Slang..."! !

!ObjectMemory methodsFor: 'simulation' stamp: 'eem 5/10/2014 08:42'!
validate
	"Validate all the objects in the heap."
	self cCode: [] inSmalltalk: [
	| oop prev |
	self interpreter transcript show: 'Validating...'.
	oop := self firstObject.
	[oop < endOfMemory] whileTrue: [
		self validate: oop.
		prev := oop.  "look here if debugging prev obj overlapping this one"
		oop := self objectAfter: oop.
	].
	self touch: prev.  "Don't offer to delete this please"
	self interpreter transcript show: 'done.'; cr]! !

!ObjectMemory methodsFor: 'simulation' stamp: 'eem (const to msg rewrite) 11/6/2014 10:59'!
validate: oop
	<doNotGenerate>
	| header type cc sz fmt nextChunk | 
	header := self longAt: oop.
	type := header bitAnd: 3.
	type = 2 ifFalse: [type = (self rightType: header) ifFalse: [self halt]].
	sz := (header bitAnd: SizeMask) >> 2.
	(self isFreeObject: oop)
		ifTrue: [ nextChunk := oop + (self sizeOfFree: oop) ]
		ifFalse: [  nextChunk := oop + (self sizeBitsOf: oop) ].
	nextChunk > endOfMemory
		ifTrue: [oop = endOfMemory ifFalse: [self halt]].
	(self headerType: nextChunk) = 0 ifTrue: [
		(self headerType: (nextChunk + (self wordSize*2))) = 0 ifFalse: [self halt]].
	(self headerType: nextChunk) = 1 ifTrue: [
		(self headerType: (nextChunk + self wordSize)) = 1 ifFalse: [self halt]].
	type = 2 ifTrue:
		["free block" ^ self].
	fmt := self formatOfHeader: header.
	cc := self compactClassIndexOfHeader: header.
	cc > 16 ifTrue: [self halt].	"up to 32 are legal, but not used"
	type = 0 ifTrue:
		["three-word header"
		((self longAt: oop-self wordSize) bitAnd: 3) = type ifFalse: [self halt].
		((self longAt: oop-(self wordSize*2)) bitAnd: 3) = type ifFalse: [self halt].
		((self longAt: oop-self wordSize) = type) ifTrue: [self halt].	"Class word is 0"
		sz = 0 ifFalse: [self halt]].
	type = 1 ifTrue:
		["two-word header"
		((self longAt: oop-self wordSize) bitAnd: 3) = type ifFalse: [self halt].
		cc > 0 ifTrue: [sz = 1 ifFalse: [self halt]].
		sz = 0 ifTrue: [self halt]].
	type = 3 ifTrue:
		["one-word header"
		cc = 0 ifTrue: [self halt]].
	fmt = 5 ifTrue: [self halt].
	fmt = 7 ifTrue: [self halt].
	fmt >= self firstCompiledMethodFormat ifTrue: "must have integer header"
		[(self isIntegerObject: (self longAt: oop + self wordSize)) ifFalse: [self halt]].! !

!ObjectMemory methodsFor: 'image segment in/out' stamp: 'eem 12/19/2008 22:16'!
verifyCleanHeaders
	| oop |
	oop := self firstObject.
	[self oop: oop isLessThan: endOfMemory] whileTrue:
		[(self isFreeObject: oop)
			ifTrue: ["There should only be one free block at end of memory."
					(self objectAfter: oop) = endOfMemory
						ifFalse: [self error: 'Invalid obj with HeaderTypeBits = Free.']]
			ifFalse: [(self isMarked: oop) ifTrue:
						[self error: 'Invalid obj with MarkBit set.']].
		oop := self objectAfter: oop]! !

!ObjectMemory methodsFor: 'memory access' stamp: 'eem 5/28/2010 11:47'!
vmEndianness
	"return 0 for little endian, 1 for big endian"
	"?!!*#$%!! _Terrible_ name. Call it what it is.
	 VMBIGENDIAN or VMLITTLEENDIAN.  Dont leave us guessing!!!!!!!!
	 Alas this has to remain for backward compatibility."

	^VMBIGENDIAN ifTrue: [1] ifFalse: [0]! !

!ObjectMemory methodsFor: 'header formats' stamp: 'eem 12/12/2015 14:16'!
weakArrayFormat
	<api>
	<cmacro>
	^4! !

!ObjectMemory methodsFor: 'finalization' stamp: 'eem 11/10/2014 12:49'!
weakFinalizerCheck: oop
	"Our oop has at least 2 non-weak fixed slots (this is assured before entering this method, in
	#finalizeReference:.
	We are assuming that if its first non-weak field is an instance of ClassWeakFinalizationList class,
	then we should add this oop to that list, by storing it to list's first field and
	also, updating the oop's 2nd fixed slot to point to the value which we overridden:
	
	list := oop instVarAt: 1.
	list class == WeakFinalizationList ifTrue: [
		first := list instVarAt: 1.
		oop instVarAt: 2 put: first.
		list instVarAt: 1 put: oop ]	"
	<inline: true>
	| listOop listItemOop |

	listOop := self fetchPointer: 0 ofObject: oop.
	(self fetchClassOf: listOop) = (self splObj: ClassWeakFinalizer) ifTrue:
		[listItemOop := self fetchPointer: 0 ofObject: listOop.
		 self storePointer: 1 ofObject: oop withValue: listItemOop. 
		 self storePointer: 0 ofObject: listOop withValue: oop]! !

!ObjectMemory methodsFor: 'debug printing' stamp: 'eem 3/24/2014 17:21'!
whereIsMaybeHeapThing: anOop
	<returnTypeC: 'char *'>
	(self oop: anOop isGreaterThanOrEqualTo: self startOfMemory andLessThan: endOfMemory) ifTrue:
		[(self oop: anOop isLessThan: freeBlock) ifTrue:
			[^' is in old space'].
		 ^' is in new space'].
	^nil! !

!ObjectMemory methodsFor: 'interpreter access' stamp: 'eem 11/13/2014 15:24'!
wordSize
	"Answer the manager's word size, which is assumed to be equivalent to the underlying machine's word size."
	^4! !

!ObjectMemory methodsFor: 'simulation' stamp: 'eem 11/18/2016 16:59'!
writeEnableMemory
	<doNotGenerate>
	memory class == ReadOnlyArrayWrapper ifTrue:
		[memory := memory array]! !

!ObjectMemory methodsFor: 'simulation' stamp: 'eem 11/18/2016 17:00'!
writeProtectMemory
	<doNotGenerate>
	memory class ~~ ReadOnlyArrayWrapper ifTrue:
		[memory := ReadOnlyArrayWrapper around: memory]! !

VMClass subclass: #OutOfLineLiteralsManager
	instanceVariableNames: 'cogit objectMemory objectRepresentation firstOpcodeIndex nextLiteralIndex lastDumpedLiteralIndex literals literalsSize savedFirstOpcodeIndex savedNextLiteralIndex savedLastDumpedLiteralIndex'
	classVariableNames: ''
	poolDictionaries: 'CogAbstractRegisters CogCompilationConstants CogRTLOpcodes'
	category: 'VMMaker-JIT'!

!OutOfLineLiteralsManager commentStamp: 'nice 12/31/2020 09:14' prior: 0!
An OutOfLineLiteralsManager manages the dumping of literals for backends that want to keep literals out-of-line, accessed by pc-relative addressing.

Instance Variables
	cogit:		<Cogit>!

!OutOfLineLiteralsManager class methodsFor: 'translation' stamp: 'eem 6/15/2015 20:34'!
declareCVarsIn: aCCodeGenerator
	aCCodeGenerator var: 'literals' type: #'AbstractInstruction *'! !

!OutOfLineLiteralsManager class methodsFor: 'translation' stamp: 'eem 6/16/2015 10:25'!
isNonArgumentImplicitReceiverVariableName: aString
	^Cogit isNonArgumentImplicitReceiverVariableName: aString! !

!OutOfLineLiteralsManager class methodsFor: 'translation' stamp: 'eem 1/23/2020 14:47'!
translationClasses
	^{self}! !

!OutOfLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 6/18/2015 13:31'!
allocateLiteral: aLiteral
	"Allocate an unsharable Literal instruction for the literal and answer it."
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	| litInst |
	<var: 'litInst' type: #'AbstractInstruction *'>
	nextLiteralIndex >= literalsSize ifTrue:
		[self allocateLiterals: literalsSize + 8].
	litInst := self literalInstructionAt: nextLiteralIndex.
	litInst initializeUniqueLiteral: aLiteral.
	nextLiteralIndex := nextLiteralIndex + 1.
	"Record the opcodeIndex of the first dependent instruction (the first instruction that references an out-of-line literal)"
	firstOpcodeIndex > cogit getOpcodeIndex ifTrue:
		[firstOpcodeIndex := cogit getOpcodeIndex - 1].
	^litInst! !

!OutOfLineLiteralsManager methodsFor: 'initialization' stamp: 'eem 6/18/2015 13:33'!
allocateLiterals: initialNumLiterals
	<inline: true>
	| newLiterals newInst existingInst |
	<var: 'newInst' type: #'AbstractInstruction *'>
	<var: 'existingInst' type: #'AbstractInstruction *'>
	<var: 'newLiterals' type: #'AbstractInstruction *'>
	initialNumLiterals > literalsSize ifTrue:
		[newLiterals := self cCode:
								[self c: initialNumLiterals alloc: (self sizeof: CogAbstractInstruction)]
							inSmalltalk:
								[CArrayAccessor on: ((1 to: initialNumLiterals) collect: [:i| CogCompilerClass for: cogit])].
		 "Must copy across state (not using realloc, cuz...) and
		  must also update existing instructions to refer to the new ones...
		  It's either this or modify all generation routines to be able to retry
		  with more literals after running out of literals."
		 literals ifNotNil:
			[0 to: nextLiteralIndex - 1 do:
				[:i|
				existingInst := self literalInstructionAt: i.
				newInst := self addressOf: (newLiterals at: i).
				newInst cloneLiteralFrom: existingInst.
				self assert: existingInst dependent isNil.
				existingInst dependent: newInst].
			0 to: cogit getOpcodeIndex - 1 do:
				[:i|
				existingInst := cogit abstractInstructionAt: i.
				(existingInst dependent notNil
				 and: [existingInst dependent opcode = Literal]) ifTrue:
					[existingInst dependent: existingInst dependent dependent]]].
		 self cCode: [self free: literals] inSmalltalk: [].
		 literals := newLiterals.
		 literalsSize := initialNumLiterals]! !

!OutOfLineLiteralsManager methodsFor: 'generate machine code' stamp: 'eem 6/11/2015 16:07'!
assertValidAnnotation: annotation for: instruction
	"Insist that the IsObjectReference applies only to out-of-line literals."
	<var: #instruction type: #'AbstractInstruction *'>
	<inline: true>
	self assert: (annotation ~= cogit getIsObjectReference or: [instruction opcode = Literal])! !

!OutOfLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 12/17/2019 12:53'!
checkLiteral32: literal forInstruction: anInstruction
	<inline: #always>
	^self checkLiteral: literal forInstruction: anInstruction! !

!OutOfLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 6/8/2015 17:39'!
checkLiteral: literal forInstruction: anInstruction
	<var: #anInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	anInstruction usesOutOfLineLiteral ifTrue:
		[anInstruction dependent: (self locateLiteral: literal)].
	^anInstruction! !

!OutOfLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 12/17/2019 13:40'!
checkQuickConstant: literal forInstruction: anInstruction
	<var: #anInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	anInstruction usesOutOfLineLiteral ifTrue:
		[anInstruction dependent: (self locateLiteral: (self cCode: [literal] inSmalltalk: [literal bitAnd: 1 << (objectMemory wordSize * 8) - 1])
										size: objectMemory bytesPerOop)].
	^anInstruction! !

!OutOfLineLiteralsManager methodsFor: 'initialization' stamp: 'eem 6/16/2015 11:36'!
cogit: aCogit
	<doNotGenerate>
	cogit := aCogit.
	objectMemory := aCogit objectMemory.
	objectRepresentation := aCogit objectRepresentation.
	literalsSize := 0! !

!OutOfLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 12/10/2019 12:17'!
dumpLiterals: generateBranchAround
	"Output all pending literal instructions, making the originals dependents on the generated ones
	 so that a later pass will copy the address of each generated literal inst to its original in literals,
	 and hence allow the instruction using the literal to compute the correct address.."
	| jump litInst |
	<var: 'jump' type: #'AbstractInstruction *'>
	<var: 'litInst' type: #'AbstractInstruction *'>

	generateBranchAround ifTrue:
		[jump := cogit Jump: 0].
	cogit backEnd literalAlignment > cogit backEnd codeGranularity ifTrue:
		[cogit gen: AlignmentNops operand: cogit backEnd literalAlignment].
	lastDumpedLiteralIndex to: nextLiteralIndex - 1 do:
		[:i|
		litInst := self literalInstructionAt: i.
		(cogit gen: Literal operand: (litInst operands at: 0)) dependent: litInst.
		litInst setLiteralOpcodeIndex: cogit getOpcodeIndex].
	generateBranchAround ifTrue:
		[jump jmpTarget: cogit Label].

	firstOpcodeIndex := cogit getOpcodeIndex.
	lastDumpedLiteralIndex := nextLiteralIndex! !

!OutOfLineLiteralsManager methodsFor: 'garbage collection' stamp: 'eem 6/11/2015 16:48'!
fetchLiteralAtAnnotatedAddress: address using: instruction
	"With out-of-line literals, the IsObjectReference annotation refers to
	 the start of the literal and hence access the memory directly."
	<var: 'instruction' type: #'AbstractInstruction *'>
	<inline: true>
	^objectMemory longAt: address! !

!OutOfLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 6/16/2015 12:06'!
literalInstructionAt: index
	<cmacro: '(index) (&literals[index])'>
	<returnTypeC: #'AbstractInstruction *'>
	^(literals at: index)
		ifNil: [literals at: index put: (CogCompilerClass for: cogit)]
		ifNotNil: [:litInst| litInst]! !

!OutOfLineLiteralsManager methodsFor: 'testing' stamp: 'eem 6/16/2015 13:01'!
literalInstructionInRange: litInst
	"A literal is in range if its opcode index is within outOfLineLiteralOpcodeLimit,
	 or if its index has yet to be assigned."
	<var: 'litInst' type: #'AbstractInstruction *'>
	| opcodeIdx |
	opcodeIdx := litInst literalOpcodeIndex.
	^opcodeIdx asInteger < 0
	  or: [self assert: cogit getOpcodeIndex >= opcodeIdx.
		cogit getOpcodeIndex - opcodeIdx < cogit backEnd outOfLineLiteralOpcodeLimit]! !

!OutOfLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 6/18/2015 13:32'!
locateLiteral: aLiteral
	"Search for a Literal instruction that is in-range and answer it.  Otherwise
	 allocate a new sharable Literal instruction for the literal and answer it."
	<returnTypeC: #'AbstractInstruction *'>
	<inline: false>
	| litInst |
	<var: 'litInst' type: #'AbstractInstruction *'>
	0 to: nextLiteralIndex - 1 do:
		[:i|
		litInst := self literalInstructionAt: i.
		((litInst operands at: 0) = aLiteral
		 and: [litInst isSharable
		 and: [self literalInstructionInRange: litInst]]) ifTrue:
			[^litInst]].
	nextLiteralIndex >= literalsSize ifTrue:
		[self allocateLiterals: literalsSize + 8].
	litInst := self literalInstructionAt: nextLiteralIndex.
	litInst initializeSharableLiteral: aLiteral.
	nextLiteralIndex := nextLiteralIndex + 1.
	"Record the opcodeIndex of the first dependent instruction (the first instruction that references an out-of-line literal)"
	firstOpcodeIndex > cogit getOpcodeIndex ifTrue:
		[firstOpcodeIndex := cogit getOpcodeIndex - 1].
	^litInst! !

!OutOfLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 12/17/2019 14:16'!
locateLiteral: aLiteral size: ignored
	<inline: #always>
	^self locateLiteral: aLiteral! !

!OutOfLineLiteralsManager methodsFor: 'testing' stamp: 'eem 6/11/2015 13:16'!
mustDumpLiterals: currentOpcodeIndex
	<inline: true>
	^currentOpcodeIndex >= firstOpcodeIndex
	  and: [currentOpcodeIndex - firstOpcodeIndex >= cogit backEnd outOfLineLiteralOpcodeLimit]! !

!OutOfLineLiteralsManager methodsFor: 'initialization' stamp: 'eem 03/04/2017 09:43'!
resetForRecompile
	<inline: true>
	firstOpcodeIndex := savedFirstOpcodeIndex.
	nextLiteralIndex := savedNextLiteralIndex.
	lastDumpedLiteralIndex := savedLastDumpedLiteralIndex! !

!OutOfLineLiteralsManager methodsFor: 'initialization' stamp: 'eem 6/18/2015 11:06'!
resetLiterals
	<inline: true>
	firstOpcodeIndex := 1 << 16. "an impossibly high value"
	nextLiteralIndex := lastDumpedLiteralIndex := 0! !

!OutOfLineLiteralsManager methodsFor: 'initialization' stamp: 'eem 03/04/2017 09:43'!
saveForRecompile

	<inline: true>
	savedFirstOpcodeIndex := firstOpcodeIndex.
	savedNextLiteralIndex := nextLiteralIndex.
	savedLastDumpedLiteralIndex := lastDumpedLiteralIndex! !

!OutOfLineLiteralsManager methodsFor: 'garbage collection' stamp: 'eem 2/24/2020 09:16'!
storeLiteral: literal atAnnotatedAddress: address using: instruction
	"With out-of-line literals, the IsObjectReference annotation refers to
	 the start of the literal and hence access the memory directly."
	<var: 'address' type: #usqInt>
	<var: 'instruction' type: #'AbstractInstruction *'>
	<inline: #always>
	cogit codeLongAt: address put: literal! !

!OutOfLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 12/17/2019 13:47'!
uniqueLiteral32: literal forInstruction: anInstruction
	<var: #anInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	| literalInstruction |
	self assert: anInstruction usesOutOfLineLiteral.
	literalInstruction := self allocateLiteral: literal.
	literalInstruction setLiteralSize: 4.
	anInstruction dependent: literalInstruction.
	^anInstruction! !

!OutOfLineLiteralsManager methodsFor: 'compile abstract instructions' stamp: 'eem 6/10/2015 19:37'!
uniqueLiteral: literal forInstruction: anInstruction
	<var: #anInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	<inline: true>
	self assert: anInstruction usesOutOfLineLiteral.
	anInstruction dependent: (self allocateLiteral: literal).
	^anInstruction! !

OutOfLineLiteralsManager subclass: #OutOfLineLiteralsManagerFor64Bits
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!OutOfLineLiteralsManagerFor64Bits commentStamp: 'eem 12/17/2019 12:42' prior: 0!
OutOfLineLiteralsManagerFor64Bits specializes OutOfLineLiteralsManager to segregate 64-bit and 32-bit literals to achieve correct alignment and more compact code.!

!OutOfLineLiteralsManagerFor64Bits class methodsFor: 'translation' stamp: 'eem 1/23/2020 14:47'!
translationClasses
	^{superclass. self}! !

!OutOfLineLiteralsManagerFor64Bits methodsFor: 'compile abstract instructions' stamp: 'eem 12/17/2019 14:31'!
checkLiteral32: literal forInstruction: anInstruction
	<var: #anInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	anInstruction usesOutOfLineLiteral ifTrue:
		[anInstruction dependent: (self locateLiteral: literal size: 4)].
	^anInstruction! !

!OutOfLineLiteralsManagerFor64Bits methodsFor: 'compile abstract instructions' stamp: 'eem 12/17/2019 13:32'!
checkLiteral: literal forInstruction: anInstruction
	<var: #anInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	anInstruction usesOutOfLineLiteral ifTrue:
		[anInstruction dependent: (self locateLiteral: literal size: objectMemory bytesPerOop)].
	^anInstruction! !

!OutOfLineLiteralsManagerFor64Bits methodsFor: 'compile abstract instructions' stamp: 'eem 12/17/2019 13:24'!
dumpLiterals: generateBranchAround
	"Output all pending literal instructions, making the originals dependents on the generated ones
	 so that a later pass will copy the address of each generated literal inst to its original in literals,
	 and hence allow the instruction using the literal to compute the correct address.

	 Override to segregate 64-bit and 32-bit literals"
	| jump litInst |

	generateBranchAround ifTrue:
		[jump := cogit Jump: 0].
	"Align and dump the 64-bit literals.
	 We may waste a slot here we could fill with a 32-bit literal, but for now, KISS..."
	cogit backEnd literalAlignment > cogit backEnd codeGranularity ifTrue:
		[cogit gen: AlignmentNops operand: cogit backEnd literalAlignment].
	lastDumpedLiteralIndex to: nextLiteralIndex - 1 do:
		[:i|
		litInst := self literalInstructionAt: i.
		litInst literalSize = 8 ifTrue:
			[(cogit gen: Literal operand: (litInst operands at: 0))
				dependent: litInst;
				setLiteralSize: 8.
			 litInst setLiteralOpcodeIndex: cogit getOpcodeIndex]].
	"Now dump the 32-bit literals"
	lastDumpedLiteralIndex to: nextLiteralIndex - 1 do:
		[:i|
		litInst := self literalInstructionAt: i.
		litInst literalSize = 4 ifTrue:
			[(cogit gen: Literal operand: (litInst operands at: 0))
				dependent: litInst;
				setLiteralSize: 4.
			 litInst setLiteralOpcodeIndex: cogit getOpcodeIndex]].
	generateBranchAround ifTrue:
		[jump jmpTarget: cogit Label].

	firstOpcodeIndex := cogit getOpcodeIndex.
	lastDumpedLiteralIndex := nextLiteralIndex! !

!OutOfLineLiteralsManagerFor64Bits methodsFor: 'compile abstract instructions' stamp: 'eem 12/17/2019 13:34'!
locateLiteral: aLiteral
	self shouldNotImplement! !

!OutOfLineLiteralsManagerFor64Bits methodsFor: 'compile abstract instructions' stamp: 'eem 12/17/2019 13:33'!
locateLiteral: aLiteral size: litSize
	"Search for a Literal instruction that is in-range and answer it.  Otherwise
	 allocate a new sharable Literal instruction for the literal and answer it."
	<returnTypeC: #'AbstractInstruction *'>
	<inline: false>
	| litInst |
	<var: 'litInst' type: #'AbstractInstruction *'>
	0 to: nextLiteralIndex - 1 do:
		[:i|
		litInst := self literalInstructionAt: i.
		(litInst literalSize = litSize
		 and: [(litInst operands at: 0) = aLiteral
		 and: [litInst isSharable
		 and: [self literalInstructionInRange: litInst]]]) ifTrue:
			[^litInst]].
	nextLiteralIndex >= literalsSize ifTrue:
		[self allocateLiterals: literalsSize + 8].
	litInst := self literalInstructionAt: nextLiteralIndex.
	litInst
		initializeSharableLiteral: aLiteral;
		setLiteralSize: litSize.
	nextLiteralIndex := nextLiteralIndex + 1.
	"Record the opcodeIndex of the first dependent instruction (the first instruction that references an out-of-line literal)"
	firstOpcodeIndex > cogit getOpcodeIndex ifTrue:
		[firstOpcodeIndex := cogit getOpcodeIndex - 1].
	^litInst! !

!VMClass class methodsFor: 'translation' stamp: 'eem 06/07/2018 15:29'!
ancilliaryClasses
	"Answer any extra classes to be included in the translation."
	^{}! !

!VMClass class methodsFor: 'translation' stamp: 'eem 2/10/2009 16:32'!
apiExportHeaderName
	"VM classesd that want to generate an api export header override this."
	^nil! !

!VMClass class methodsFor: 'adding/removing methods' stamp: 'eem 3/4/2016 10:04'!
basicRemoveSelector: aSelector
	"Override to update the timeStamp"
	^(super basicRemoveSelector: aSelector) ifNotNil:
		[:oldMethod| self touch. oldMethod]! !

!VMClass class methodsFor: 'translation' stamp: 'eem 8/21/2013 12:24'!
cogMethodClass: options
	^(options at: #NewspeakVM ifAbsent: [false])
		ifTrue: [NewspeakCogMethod]
		ifFalse: [CogMethod]! !

!VMClass class methodsFor: 'accessing class hierarchy' stamp: 'eem 6/7/2018 15:24'!
cogitClass
	"Answer the cogitClass in effect.  Ensure that StackInterpreter has a nil cogitClass."
	(self isInterpreterClass and: [self hasCogit not]) ifTrue:
		[^nil].
	^Smalltalk classNamed: (InitializationOptions
								at: #Cogit
								ifAbsent: [#StackToRegisterMappingCogit])! !

!VMClass class methodsFor: 'accessing class hierarchy' stamp: 'eem 9/30/2010 12:21'!
coreInterpreterClass
	"While the interpreterClass/vmClass for translation may be
	 a subclass that holds a few primitives we want the actual
	 interpreter name at the head of the generated file."
	^((name endsWith: 'Primitives')
	   and: [name beginsWith: superclass name])
		ifTrue: [superclass]
		ifFalse: [self]! !

!VMClass class methodsFor: 'translation' stamp: 'eem 2/3/2020 14:21'!
declareC: arrayOfVariableNames as: aCType ifPresentIn: aCCodeGenerator
	"Declare the variables in arrayOfVariableNames with the given type."

	arrayOfVariableNames do:
		[:varName |
		 (aCCodeGenerator hasVariable: varName) ifTrue:
			[((self mustBeGlobal: varName)
			 and: [(aCCodeGenerator typeOfVariable: varName) ifNil: [false] ifNotNil: [:varType| aCType ~= varType]]) ifTrue:
				[self error: 'attempting to change type of global var ', varName, '!!!!'].
			 aCCodeGenerator var: varName type: aCType]]! !

!VMClass class methodsFor: 'translation' stamp: 'eem 7/7/2008 20:18'!
declareC: arrayOfVariableNames as: aCType in: aCCodeGenerator
	"Declare the variables in arrayOfVariableNames with the given type."

	arrayOfVariableNames
		do: [:varName | aCCodeGenerator var: varName type: aCType]! !

!VMClass class methodsFor: 'translation' stamp: 'eem 10/12/2013 09:46'!
declareCAsOop: arrayOfVariableNames in: aCCodeGenerator
	"Declare the variables in arrayOfVariableNames with type representing position in object memory."

	arrayOfVariableNames do:
		[:varName| aCCodeGenerator var: varName type: #usqInt]! !

!VMClass class methodsFor: 'translation' stamp: 'eem 10/12/2013 09:48'!
declareCAsUSqLong: arrayOfVariableNames in: aCCodeGenerator
	"Declare the variables in arrayOfVariableNames with type representing position in object memory."

	arrayOfVariableNames do:
		[:varName| aCCodeGenerator var: varName type: #usqLong]! !

!VMClass class methodsFor: 'translation' stamp: 'eem 1/9/2015 11:01'!
declareCVarsIn: aCCodeGenerator 
	"Declare any additional variables and/or add type declarations for existing variables."
	aCCodeGenerator
		var: #expensiveAsserts
		declareC: 'char expensiveAsserts = 0'! !

!VMClass class methodsFor: 'translation' stamp: 'eem 11/17/2019 10:27'!
declareInterpreterVersionIn: aCCodeGenerator defaultName: defaultName
	NewspeakVM ifTrue:
		["Newspeak as of mid 2011 derives SystemScope systemName from the interpreterVersion
		  (via system attribute 1004) by copying up to but not including the last space, provided the
		  string ends with a digit.  So spaces must be eliminated from the Monitcello version string,
		  and we can't surround it with square brackets."
		(aCCodeGenerator shortMonticelloDescriptionForClass: self) last isDigit ifFalse:
			[self error: 'Newspeak expects interpreterVersion ends with a digit'].
		aCCodeGenerator
			var: #interpreterVersion
			declareC: 'const char *interpreterVersion = "Newspeak Virtual Machine ',
							('[ Open Smalltalk, ', aCCodeGenerator vmClass objectMemoryClass memoryManagerName,'] '),
							((aCCodeGenerator shortMonticelloDescriptionForClass: self) copyReplaceAll: ' ' with: '_'),
							'"'.
		^self].
	
	aCCodeGenerator
		var: #interpreterVersion
		declareC: 'const char *interpreterVersion = "Open Smalltalk ', defaultName, ' VM [',
					(aCCodeGenerator shortMonticelloDescriptionForClass: self),']"'! !

!VMClass class methodsFor: 'debugger' stamp: 'eem 3/18/2013 11:33'!
defaultIntegerBaseInDebugger
	"DefaultBase := 16."
	"DefaultBase := 10."
	DefaultBase isNil ifTrue: [DefaultBase := 16].
	^DefaultBase! !

!VMClass class methodsFor: 'accessing class hierarchy' stamp: 'eem 9/30/2013 13:16'!
defaultObjectMemoryClass
	"Default for all-in-one VMs where the interpreter inherits from the object memory."
	^nil! !

!VMClass class methodsFor: 'translation' stamp: 'eem 12/19/2010 18:40'!
getVMMaker
	^Notification new tag: #getVMMaker; signal! !

!VMClass class methodsFor: 'accessing class hierarchy' stamp: 'eem 3/17/2015 13:44'!
hasCogit
	^false! !

!VMClass class methodsFor: 'translation' stamp: 'eem 2/15/2009 09:12'!
implicitReturnTypeFor: aSelector
	"Answer the return type for methods that don't have an explicit return."
	^#sqInt! !

!VMClass class methodsFor: 'initialization' stamp: 'eem 6/7/2018 15:02'!
initializationOptions
	^InitializationOptions! !

!VMClass class methodsFor: 'initialization' stamp: 'eem 9/26/2020 14:49'!
initialize
	InitializationOptions ifNil: [InitializationOptions := Dictionary new].
	ExpensiveAsserts := false.
	(Smalltalk classNamed: #Utilities) ifNotNil:
		[:utilitiesClass|
		 (utilitiesClass classPool at: #CommonRequestStrings ifAbsent: []) ifNotNil:
			[:commonRequestStringHolder|
			(commonRequestStringHolder contents asString includesSubstring: 'VMClass open') ifFalse:
				[Utilities appendToCommonRequests: '-\VMMaker generateConfiguration\VMMaker generateAllConfigurationsUnderVersionControl\VMMaker generateAllSpurConfigurations\VMClass openCogSpurMultiWindowBrowser\VMClass openCogV3MultiWindowBrowser\VMClass openObjectMemoriesInterpretersBrowser\VMClass openSpurMultiWindowBrowser\VMClass openCogitMultiWindowBrowser' withCRs]]]! !

!VMClass class methodsFor: 'initialization' stamp: 'eem 3/6/2019 11:20'!
initializeForPlugins: optionsDictionaryOrArray
	"Do a default initialization, suitable for generating plugin sources."
	StackInterpreterSimulator
		initializeWithOptions: optionsDictionaryOrArray
		objectMemoryClass: ObjectMemory! !

!VMClass class methodsFor: 'initialization' stamp: 'eem 6/7/2018 15:31'!
initializeMiscConstants
	"Falsify the `what type of VM is this?' flags that are defined in the various interp.h files.
	 Subclass implementations need to include a super initializeMiscConstants"

	| omc |
	VMBIGENDIAN class. "Mention this for the benefit of CCodeGenerator>>emitCConstantsOn:"
	SPURVM := STACKVM := COGVM := COGMTVM := false.

	InitializationOptions ifNil: [InitializationOptions := Dictionary new].
	omc := InitializationOptions at: #ObjectMemory ifAbsent: nil.
	(omc isNil and: [self defaultObjectMemoryClass notNil]) ifTrue:
		[omc := InitializationOptions at: #ObjectMemory put: self defaultObjectMemoryClass name].
	InitializationOptions
		at: #SqueakV3ObjectMemory	"the good ole default"
			ifAbsentPut: (omc
					ifNil: [true]
					ifNotNil: [(Smalltalk at: omc) includesBehavior: ObjectMemory]);
		at: #SpurObjectMemory		"the new contender"
			ifAbsentPut: (omc
					ifNil: [false]
					ifNotNil: [(Smalltalk at: omc) includesBehavior: SpurMemoryManager]).

	"Use ifAbsentPut: so that they will get copied back to the
	 VMMaker's options and dead code will likely be eliminated."
	PharoVM := InitializationOptions at: #PharoVM ifAbsentPut: [false].
	NewspeakVM := InitializationOptions at: #NewspeakVM ifAbsentPut: [false].
	SistaVM := InitializationOptions at: #SistaVM ifAbsentPut: [false].
	TempVectReadBarrier := InitializationOptions at: #TempVectReadBarrier ifAbsentPut: [false].
	LowcodeVM := InitializationOptions at: #LowcodeVM ifAbsentPut: [false].
	MULTIPLEBYTECODESETS := InitializationOptions at: #MULTIPLEBYTECODESETS ifAbsentPut: [false].
	"Simulation only; on by default..."
	CloneOnGC := InitializationOptions at: #CloneOnGC ifAbsentPut: [true].
	CloneOnScavenge := InitializationOptions at: #CloneOnScavenge ifAbsentPut: [true].

	"These must be set only if specified, not defaulted, because they are set on the command line or in include files."
	InitializationOptions
		at: #VMBIGENDIAN	ifPresent: [:value| VMBIGENDIAN := value];
		at: #ObjectMemory	ifPresent: [:value| SPURVM := value beginsWith: 'Spur'];
		at: #STACKVM		ifPresent: [:value| STACKVM := value];
		at: #COGVM		ifPresent: [:value| COGVM := InitializationOptions at: #COGVM];
		at: #COGMTVM		ifPresent: [:value| COGMTVM := InitializationOptions at: #COGMTVM].

	"consistency checks"
	SPURVM
		ifTrue:
			[(TempVectReadBarrier not
			  and: [{SpurMemoryManager compactorClass}, (SpurMemoryManager compactorClass ancilliaryClasses) anySatisfy:
						[:c| c == SpurSelectiveCompactor]]) ifTrue:
				[self error: 'Selective compactor requires read barrier']]
		ifFalse:
			[TempVectReadBarrier ifTrue: [self error: 'read barrier works with spur VM only...'].
			 SistaVM ifTrue: [self error: 'Sista VM works with spur VM only...']].

	"And not these; they're compile-time"
	IMMUTABILITY := InitializationOptions at: #IMMUTABILITY ifAbsent: [SPURVM] "Default as enabled for Spur VMs"! !

!VMClass class methodsFor: 'initialization' stamp: 'eem 6/13/2021 12:36'!
initializePrimitiveErrorCodes
	"Define the VM's primitive error codes.  N.B. these are
	 replicated in platforms/Cross/vm/sqVirtualMachine.h."
	"VMClass initializePrimitiveErrorCodes"
	| pet |
	PrimErrTableIndex := 51. "Zero-relative"
	"See SmalltalkImage>>recreateSpecialObjectsArray for the table definition.
	 If the table exists and is large enough the corresponding entry is returned as
	 the primitive error, otherwise the error is answered numerically."
	pet := Smalltalk specialObjectsArray at: PrimErrTableIndex + 1 ifAbsent: [#()].
	pet isArray ifFalse: [pet := #()].
	PrimNoErr := 0. "for helper methods that need to answer success or an error code."
	PrimErrGenericFailure		:= pet indexOf: nil ifAbsent: 1.
	PrimErrBadReceiver			:= pet indexOf: #'bad receiver' ifAbsent: 2.
	PrimErrBadArgument		:= pet indexOf: #'bad argument' ifAbsent: 3.
	PrimErrBadIndex			:= pet indexOf: #'bad index' ifAbsent: 4.
	PrimErrBadNumArgs		:= pet indexOf: #'bad number of arguments' ifAbsent: 5.
	PrimErrInappropriate		:= pet indexOf: #'inappropriate operation' ifAbsent: 6.
	PrimErrUnsupported		:= pet indexOf: #'unsupported operation' ifAbsent: 7.
	PrimErrNoModification		:= pet indexOf: #'no modification' ifAbsent: 8.
	PrimErrNoMemory			:= pet indexOf: #'insufficient object memory' ifAbsent: 9.
	PrimErrNoCMemory			:= pet indexOf: #'insufficient C memory' ifAbsent: 10.
	PrimErrNotFound			:= pet indexOf: #'not found' ifAbsent: 11.
	PrimErrBadMethod			:= pet indexOf: #'bad method' ifAbsent: 12.
	PrimErrNamedInternal		:= pet indexOf: #'internal error in named primitive machinery' ifAbsent: 13.
	PrimErrObjectMayMove		:= pet indexOf: #'object may move' ifAbsent: 14.
	PrimErrLimitExceeded		:= pet indexOf: #'resource limit exceeded' ifAbsent: 15.
	PrimErrObjectIsPinned		:= pet indexOf: #'object is pinned' ifAbsent: 16.
	PrimErrWritePastObject		:= pet indexOf: #'primitive write beyond end of object' ifAbsent: 17.
	PrimErrObjectMoved		:= pet indexOf: #'object moved' ifAbsent: 18.
	PrimErrObjectNotPinned	:= pet indexOf: #'object not pinned' ifAbsent: 19.
	PrimErrCallbackError		:= pet indexOf: #'error in callback' ifAbsent: 20.
	PrimErrOSError				:= pet indexOf: #'operating system error' ifAbsent: 21.
	PrimErrFFIException		:= pet indexOf: #'ffi call exception' ifAbsent: 22.
	PrimErrNeedCompaction	:= pet indexOf: #'heap compaction needed' ifAbsent: 23. "N.B. This is currently an internal error in Spur image segment saving."
	PrimErrOperationFailed		:= pet indexOf: #'operation failed' ifAbsent: 24.
	PrimErrInternalError		:= pet indexOf: #'internal error' ifAbsent: 25.
	PrimErrUninitialized			:= pet indexOf: #'uninitialized' ifAbsent: 26! !

!VMClass class methodsFor: 'initialization' stamp: 'eem 9/9/2019 16:40'!
initializeWithOptions: optionsDictionaryOrArray
	"Initialize the receiver, typically initializing class variables. Initialize any class variables
	 whose names occur in optionsDictionary with the corresponding values there-in."
	InitializationOptions := optionsDictionaryOrArray isArray
								ifTrue: [Dictionary newFromPairs: optionsDictionaryOrArray]
								ifFalse: [optionsDictionaryOrArray].

	ExpensiveAsserts := InitializationOptions at: #ExpensiveAsserts ifAbsent: [false].
	self optionClassNames do:
		[:optionClassName|
		InitializationOptions at: optionClassName ifAbsentPut: false]! !

!VMClass class methodsFor: 'accessing class hierarchy' stamp: 'eem 9/7/2014 13:40'!
interpreterClass
	^self isInterpreterClass ifTrue: [self]! !

!VMClass class methodsFor: 'accessing' stamp: 'EstebanLorenzano 10/23/2014 14:19'!
interpreterVersion 
	^ self subclassResponsibility! !

!VMClass class methodsFor: 'translation' stamp: 'eem 10/25/2013 14:40'!
isAcceptableAncilliaryClass: aClass
	^true! !

!VMClass class methodsFor: 'translation' stamp: 'eem 2/11/2009 17:18'!
isAccessor: aSelector
	"Answer if aSelector is simply an accessor method for one of our fields.
	 Answer false by default.  VMStructType classes redefine this appropriately."
	^false! !

!VMClass class methodsFor: 'testing' stamp: 'eem 5/29/2009 09:49'!
isCogitClass
	"The various Cogit classes override this."
	^false! !

!VMClass class methodsFor: 'testing' stamp: 'eem 2/12/2009 15:36'!
isInterpreterClass
	"The various Interpreter classes override this."
	^false! !

!VMClass class methodsFor: 'translation' stamp: 'eem 1/15/2015 12:31'!
isNonArgumentImplicitReceiverVariableName: aString
	^false! !

!VMClass class methodsFor: 'testing' stamp: 'eem 12/14/2012 11:29'!
isPluginClass
	"InterpreterPlugin class override this."
	^false! !

!VMClass class methodsFor: 'testing' stamp: 'eem 11/26/2012 15:05'!
isStructClass
	"The various VMStructType classes override this."
	^false! !

!VMClass class methodsFor: 'testing' stamp: 'eem 7/9/2021 19:30'!
isVMClass
	^true! !

!VMClass class methodsFor: 'accessing' stamp: 'eem 11/15/2019 16:14'!
memoryManagerName
	^ self subclassResponsibility! !

!VMClass class methodsFor: 'translation' stamp: 'eem 1/3/2013 11:38'!
monticelloDescription
	"Answer the Monticello version of the packlage containing the receiver.
	 This is a hook to allow subclasses to expand upon the default monticello description."
	^CCodeGenerator monticelloDescriptionFor: self! !

!VMClass class methodsFor: 'translation' stamp: 'eem 1/9/2015 11:02'!
mustBeGlobal: var
	"Answer if a variable must be global and exported.  Used for inst vars that are accessed from VM support code."

	^var = #expensiveAsserts! !

!VMClass class methodsFor: 'translation' stamp: 'tpr 2/17/2005 13:19'!
noteCompilationOf: aSelector meta: isMeta
	"note the recompiliation by resetting the timeStamp "
	timeStamp := Time totalSeconds.
	^super noteCompilationOf: aSelector meta: isMeta! !

!VMClass class methodsFor: 'accessing class hierarchy' stamp: 'eem 6/7/2018 15:25'!
objectMemoryClass
	InitializationOptions ifNil:
		[^self defaultObjectMemoryClass].
	^Smalltalk at: (InitializationOptions
					at: #ObjectMemory
					ifAbsent: [^self defaultObjectMemoryClass])! !

!VMClass class methodsFor: 'accessing class hierarchy' stamp: 'eem 10/9/2013 17:54'!
objectRepresentationClass
	^self objectMemoryClass objectRepresentationClass! !

!VMClass class methodsFor: 'utilities' stamp: 'eem 9/26/2020 14:46'!
openCogSpurMultiWindowBrowser
	"Answer a new multi-window browser on the SpurMemoryManager classes, the Cog Interpreter classes, and the main JIT classes"
	| b |
	b := Browser open.
	#(	SpurMemoryManager Spur32BitMemoryManager Spur32BitCoMemoryManager
		 Spur64BitMemoryManager Spur64BitCoMemoryManager SpurGenerationScavenger
		InterpreterPrimitives StackInterpreter StackInterpreterPrimitives CoInterpreter CoInterpreterPrimitives CoInterpreterMT
		Cogit SimpleStackBasedCogit StackToRegisterMappingCogit RegisterAllocatingCogit
		CogObjectRepresentation CogObjectRepresentationForSpur
		CogObjectRepresentationFor32BitSpur CogObjectRepresentationFor64BitSpur
		CogRTLOpcodes CogAbstractRegisters CogAbstractInstruction),
	(self sortedAbstractInstructionClasses collect: [:class| class name]),
	#(VMStructType VMMaker CCodeGenerator TMethod)
		do: [:classOrClassName|
			(Smalltalk classNamed: classOrClassName) ifNotNil:
				[:class| b selectCategoryForClass: class; selectClass: class]]
		separatedBy:
			[b multiWindowState addNewWindow].
	b multiWindowState selectWindowIndex: 1! !

!VMClass class methodsFor: 'utilities' stamp: 'eem 10/9/2015 16:46'!
openCogTestsMultiWindowBrowser
	"Answer a new multi-window browser on the test classes in VMMaker"
	"self openCogTestsMultiWindowBrowser"
	| testClasses b |
	testClasses := (PackageInfo named: 'VMMaker') classes select: [:c| c inheritsFrom: TestCase].
	testClasses removeAll: AbstractInstructionTests allSubclasses.
	testClasses removeAll: (testClasses select: [:c| '*Plugin*' match: c name]).
	b := Browser open.
	testClasses do:
		[:class| b selectCategoryForClass: class; selectClass: class]
		separatedBy:
			[b multiWindowState addNewWindow].
	b multiWindowState selectWindowIndex: 1! !

!VMClass class methodsFor: 'utilities' stamp: 'eem 9/26/2020 14:47'!
openCogV3MultiWindowBrowser
	"Answer a new multi-window browser on the ObjectMemory classes, the Cog Interpreter classes, and the main JIT classes"
	| b |
	b := Browser open.
	#(	ObjectMemory NewObjectMemory NewCoObjectMemory
		InterpreterPrimitives StackInterpreter StackInterpreterPrimitives CoInterpreter CoInterpreterPrimitives CoInterpreterMT
		Cogit SimpleStackBasedCogit StackToRegisterMappingCogit
		VMStructType VMMaker CCodeGenerator TMethod)
		do: [:className|
			(Smalltalk classNamed: className) ifNotNil:
				[:class| b selectCategoryForClass: class; selectClass: class]]
		separatedBy:
			[b multiWindowState addNewWindow].
	b multiWindowState selectWindowIndex: 1! !

!VMClass class methodsFor: 'utilities' stamp: 'eem 10/7/2020 18:10'!
openCogitMultiWindowBrowser
	"Answer a new multi-window browser on the ObjectMemory classes, the Cog Interpreter classes, and the main JIT classes"
	"self openCogitMultiWindowBrowser"
	| b |
	b := Browser open.
	{CogRTLOpcodes. CogAbstractRegisters. CogAbstractInstruction },
	self sortedAbstractInstructionClasses,
	Cogit withAllSubclasses,
	{CogMethodZone },
	CogObjectRepresentation withAllSubclasses,
	CogBytecodeFixup withAllSubclasses,
	CogSimStackEntry withAllSubclasses,
	{VMStructType. VMMaker. CCodeGenerator. TMethod}
		do: [:class|
			b selectCategoryForClass: class; selectClass: class]
		separatedBy:
			[b multiWindowState addNewWindow].
	b multiWindowState selectWindowIndex: 1! !

!VMClass class methodsFor: 'utilities' stamp: 'eem 1/15/2011 10:01'!
openObjectMemoriesInterpretersBrowser
	"Answer a new multi-window browser on the ObjectMemory classes and the Cog Interpreter classes"
	| b |
	b := Browser open.
	#(	ObjectMemory NewObjectMemory NewCoObjectMemory
		InterpreterPrimitives StackInterpreter StackInterpreterPrimitives
		CoInterpreter CoInterpreterPrimitives CoInterpreterMT)
		do: [:className|
			(Smalltalk classNamed: className) ifNotNil:
				[:class| b selectCategoryForClass: class; selectClass: class]]
		separatedBy:
			[b multiWindowState addNewWindow].
	b multiWindowState selectWindowIndex: 1! !

!VMClass class methodsFor: 'utilities' stamp: 'eem 4/22/2014 16:13'!
openSpurMultiWindowBrowser
	"Answer a new multi-window browser on the Spur classes, the Cog StackInterpreter classes, and the support classes"
	| b |
	b := Browser open.
	#(	SpurMemoryManager Spur32BitMemoryManager Spur64BitMemoryManager
		SpurGenerationScavenger SpurSegmentManager
		Spur32BitMMLESimulator SpurGenerationScavengerSimulator
		InterpreterPrimitives StackInterpreter StackInterpreterPrimitives
		VMStructType VMMaker CCodeGenerator TMethod)
		do: [:className|
			(Smalltalk classNamed: className) ifNotNil:
				[:class| b selectCategoryForClass: class; selectClass: class]]
		separatedBy:
			[b multiWindowState addNewWindow].
	b multiWindowState selectWindowIndex: 1! !

!VMClass class methodsFor: 'initialization' stamp: 'eem 9/9/2019 16:37'!
optionClassNames
	"Answer the names of all classes that appear in opption : pragmas.
	 These have to be set to false befpre seelectively being enabled for
	 the option: pragma to be correctly processed in shouldIncludeMethodForSelector:"
	| optionClassNames block |
	optionClassNames := Set new.
	block :=
		[:c|
		c methodsDo:
			[:m|
			 (m pragmaAt: #option:) ifNotNil:
				[:p|
				 (p arguments first isSymbol
				  and: [(Smalltalk classNamed: p arguments first) notNil]) ifTrue:
					[optionClassNames add: p arguments first]]]].
	InterpreterPrimitives withAllSubclasses do: block.
	CogObjectRepresentation withAllSubclasses do: block.
	^optionClassNames! !

!VMClass class methodsFor: 'translation' stamp: 'eem 2/11/2009 09:31'!
prepareToBeAddedToCodeGenerator: aCCodeGenerator 
	"Hook for translation.  e.g. allows a subclass to override its
	 superclass's methods by deleting them before it adds its own."! !

!VMClass class methodsFor: 'translation' stamp: 'eem 8/21/2013 13:09'!
requiredMethodNames: options
	"Answer a list of method names that should be retained for export or other
	 support reasons.  These are typically entry-points that unless explicitly noted
	 will be deleted by the code generator since it will assume these are not used."
	^#()! !

!VMClass class methodsFor: 'translation' stamp: 'eem 11/1/2013 11:41'!
shouldGenerateDeadCode
	"Answer if the code generator should generate dead code, e.g. in false ifTrue: [dead] ifFalse: [live].
	 This *may* be useful in debugging (see CCodeGenerator>>nilOrBooleanConstantReceiverOf: et al).
	 But by default we answer false."

	^false! !

!VMClass class methodsFor: 'translation' stamp: 'eem 10/12/2013 12:28'!
shouldGenerateTypedefFor: aStructClass
	"Hack to work-around multiple definitions.  Sometimes a type has been defined in an include."
	^aStructClass ~~ VMCallbackContext! !

!VMClass class methodsFor: 'translation' stamp: 'eem 7/7/2021 22:19'!
shouldIncludeMethodForSelector: selector
	"Answer whether a primitive method should be translated.  Emit a warning to the transcript if the method doesn't exist."
	selector == #detectFeaturesOnMacOS ifTrue: [self halt].
	^(self whichClassIncludesSelector: selector)
		ifNotNil:
			[:c|
			 (c >> selector pragmaAt: #option:)
				ifNotNil:
					[:pragma|
					(VMBasicConstants defineAtCompileTime: pragma arguments first)
					 or: [InitializationOptions
							at: pragma arguments first
							ifAbsent: [(self bindingOf: pragma arguments first)
										ifNil: [false]
										ifNotNil: [:binding| binding value ~~ #undefined]]]]
				ifNil: [true]]
		ifNil:
			[Transcript nextPutAll: 'Cannot find implementation of '; nextPutAll: selector; nextPutAll: ' in hierarchy of '; print: self; cr; flush.
			 false]! !

!VMClass class methodsFor: 'simulation' stamp: 'eem 2/3/2017 09:36'!
simulatorClass
	"For running from Smalltalk - answer a class that can be used to simulate the receiver."

	^self! !

!VMClass class methodsFor: 'utilities' stamp: 'eem 9/26/2020 14:55'!
sortedAbstractInstructionClasses
	"Answer the in-use processor classes in a rational order."
	^((CogAbstractInstruction subclasses sort: [:ca :cb| ca ISA <= cb ISA])
		collect: [:class| class = class defaultCompilerClass ifTrue: [{class}] ifFalse: [{class. class defaultCompilerClass}]]) fold: [:a :b| a, b]! !

!VMClass class methodsFor: 'translation' stamp: 'eem 7/6/2010 15:32'!
specialValueForConstant: constantName default: defaultValue
	^nil! !

!VMClass class methodsFor: 'translation' stamp: 'eem 6/6/2018 17:44'!
staticallyResolvePolymorphicSelector: aSelectorSymbol
	^((self name select: [:ea| ea isUppercase]), '_', aSelectorSymbol) asSymbol! !

!VMClass class methodsFor: 'translation' stamp: 'tpr 3/27/2002 12:53'!
timeStamp
	^timeStamp ifNil:[0]! !

!VMClass class methodsFor: 'translation' stamp: 'eem 2/9/2009 13:54'!
touch
	"Reset the timeStamp"
	"Smalltalk allClasses select:
		[:c| (c category includesSubString: 'VMMaker-JIT') ifTrue: [c touch]]"
	"InterpreterPlugin withAllSubclassesDo:[:pl| pl touch]"
	timeStamp := Time totalSeconds! !

!VMClass class methodsFor: 'translation' stamp: 'eem 12/6/2012 15:01'!
translationClass
	"Return the class to use as the interpreterCLass when translating.  For the all-in-one
	 VMs that inherit from ObjectMemory this is the receiver.  But for the separate VMs
	 where most primitives are in a subclass it will be the subclass with the primitives."
	^self! !

!VMClass class methodsFor: 'translation' stamp: 'eem 2/9/2009 20:30'!
typeForSelf
	"Answer the type to give self if appropriate, or nil if not."
	^nil! !

!VMClass class methodsFor: 'simulation' stamp: 'eem 8/12/2021 10:45'!
valueOfSymbol: value
	^(self bindingOf: value)
		ifNil: [InitializationOptions at: value ifAbsent: [false]]
		ifNotNil: [:binding| binding value]! !

!VMClass class methodsFor: 'plugin api version' stamp: 'eem 1/18/2020 11:25'!
vmProxyMajorVersion
	"Define the  VM_PROXY_MAJOR version for this VM as used to
	 define the api in platforms/Cross/vm/sqVirtualMachine.[ch]
	 and in interp.h."
	^1! !

!VMClass class methodsFor: 'plugin api version' stamp: 'eem 1/18/2020 11:26'!
vmProxyMinorVersion
	"Define the  VM_PROXY_MAJOR version for this VM as used to define
	 the api in platforms/Cross/vm/sqVirtualMachine.[ch] and in interp.h.
	 This should be defined by the objectMemoryClass in use."
	^self subclassResponsibility! !

!VMClass class methodsFor: 'translation' stamp: 'eem 1/5/2021 15:52'!
wantsSqDotH
	^true! !

!VMClass class methodsFor: 'translation' stamp: 'eem 1/18/2020 11:46'!
writeVMHeaderTo: aStream bytesPerWord: bytesPerWord generator: aCCodeGenerator
	"Generate the contents of interp.h on aStream.  Specific Interpreter subclasses
	 override to add more stuff."
	aCCodeGenerator
		putDefineOf: 'VM_PROXY_MAJOR' as: self objectMemoryClass vmProxyMajorVersion on: aStream;
		putDefineOf: 'VM_PROXY_MINOR' as: self objectMemoryClass vmProxyMinorVersion on: aStream.
	aStream cr.
	aCCodeGenerator
		putDefineOf: 'SQ_VI_BYTES_PER_WORD' as: bytesPerWord on: aStream.
	aStream cr.

	"The most basic constants must be defined here, not in e.g. the plugin sources, to allow those
	 other sources to be shared between different builds (Spur vs SqueakV3, 32-bit vs 64-bit, etc)"
	VMBasicConstants mostBasicConstantNames asSet sorted do:
		[:constName|
		(VMBasicConstants classPool at: constName ifAbsent: []) ifNotNil:
			[:const| aCCodeGenerator putDefineOf: constName as: const on: aStream]].
	aStream cr.

	((VMBasicConstants classPool associations select: [:a| a key beginsWith: 'PrimErr'])
		sorted: [:a1 :a2| a1 value <= a2 value])
		do: [:a| aCCodeGenerator putDefineOf: a key as: a value on: aStream].
	aStream cr.

	aCCodeGenerator
		putDefineOf: 'MinSmallInteger' as: self objectMemoryClass minSmallInteger on: aStream;
		putDefineOf: 'MaxSmallInteger' as: self objectMemoryClass maxSmallInteger on: aStream;
		putDefineOf: 'NumSmallIntegerTagBits' as: self objectMemoryClass numSmallIntegerTagBits on: aStream.
	aStream cr! !

!VMClass methodsFor: 'C library simulation' stamp: 'nice 9/5/2019 11:13'!
__builtin_clz: anInteger
	<doNotGenerate>
	anInteger < 0 ifTrue: [^0].
	^32 - anInteger highBit! !

!VMClass methodsFor: 'C library simulation' stamp: 'nice 9/5/2019 11:09'!
__builtin_clzll: anInteger
	<doNotGenerate>
	anInteger < 0 ifTrue: [^0].
	^64 - anInteger highBitOfMagnitude! !

!VMClass methodsFor: 'translation support' stamp: 'eem 6/3/2010 09:00'!
addressOf: anObject
	<doNotGenerate>
	"Translates into &anObject in C."
	^anObject! !

!VMClass methodsFor: 'translation support' stamp: 'eem 8/1/2021 13:14'!
addressOf: anObject put: aBlock
	<doNotGenerate>
	"Simulate a C pointer.  Translates into &anObject in C. Provides something
	 that evaluates aBlock with the new value in response to at:put:"
	| thing firstIndex things |
	thing := anObject.
	^CPluggableAccessor new
		setObject: nil;
		atBlock: [:obj :idx| thing]
		atPutBlock:
			[:obj :idx :val|
			"Handle the common case of a single assignment as gracefully as possible"
			(firstIndex isNil or: [firstIndex = idx])
				ifTrue: [firstIndex := idx. thing := val]
				ifFalse: "handle the rare case of multiple writes through the pointer, a multi-byte assignment from memcpy:_:_: et al"
					[self assert: (#(#'memcpy:_:_:' #'memmove:_:_:' #'strncpy:_:_:') includes: thisContext sender sender selector).
					 things ifNil: [things := OrderedCollection with: thing].
					 self assert: things size + firstIndex = idx.
					 things addLast: val.
					 thing := (self objectMemory endianness == #little
									ifTrue: [things]
									ifFalse: [things reverse]) inject: 0 into: [:word :byte| (word bitShift: 8) + byte]].
			aBlock value: thing]! !

!VMClass methodsFor: 'C library simulation' stamp: 'eem 8/22/2018 12:11'!
alloca: size
	"Simulation of alloca(3)"
	<doNotGenerate>
	^CArrayAccessor on: (ByteArray new: size)! !

!VMClass methodsFor: 'C library extensions' stamp: 'eem 7/30/2014 08:47'!
alloca: numElements type: elementType
	<cmacro: '(numElements, elementType) alloca((numElements)*sizeof(elementType))'>
	^CArrayAccessor on: ((1 to: numElements) collect: [:ign| elementType new])! !

!VMClass methodsFor: 'translation support' stamp: 'eem 8/4/2014 10:23'!
asAddress: address put: aBlock
	<doNotGenerate>
	"Simulate a C pointer.  Translates into address in C. Provides something
	 that evaluates aBlock with the new value in response to at:put:"
	^CPluggableAccessor new
		setObject: nil;
		atBlock: [:obj :idx| self error: 'cannot dereference pseudo-pointers']
		atPutBlock: [:obj :idx :val| aBlock value: val]! !

!VMClass methodsFor: 'C library extensions' stamp: 'eem 9/18/2019 17:44'!
asByteArray: aStringOrStringIndex
	"aStringOrStringIndex is either a string or an address in the heap.
	 Create a ByteArray of the requested length from the bytes in the
	 heap starting at stringIndex."
	<doNotGenerate>
	| sz |
	aStringOrStringIndex isString ifTrue:
		[^aStringOrStringIndex asByteArray].
	sz := self strlen: aStringOrStringIndex.
	^self strncpy: (ByteArray new: sz) _: aStringOrStringIndex _: sz! !

!VMClass methodsFor: 'C library extensions' stamp: 'eem 9/18/2019 17:44'!
asByteArray: baIndex size: baSize
	"baIndex is an address in the heap.  Create a ByteArray of the requested length
	 from the bytes in the heap starting at baIndex."
	<doNotGenerate>
	^self strncpy: (ByteArray new: baSize) _: baIndex _: baSize! !

!VMClass methodsFor: 'C library extensions' stamp: 'eem 9/18/2019 17:44'!
asString: aStringOrStringIndex
	"aStringOrStringIndex is either a string or an address in the heap.
	 Create a String of the requested length from the bytes in the
	 heap starting at stringIndex."
	<doNotGenerate>
	| sz |
	aStringOrStringIndex isString ifTrue:
		[^aStringOrStringIndex].
	sz := self strlen: aStringOrStringIndex.
	^self strncpy: (ByteString new: sz) _: aStringOrStringIndex _: sz! !

!VMClass methodsFor: 'C library extensions' stamp: 'eem 9/18/2019 17:45'!
asString: stringIndex size: stringSize
	"stringIndex is an address in the heap.  Create a String of the requested length
	 from the bytes in the heap starting at stringIndex."
	<doNotGenerate>
	^self strncpy: (ByteString new: stringSize) _: stringIndex _: stringSize! !

!VMClass methodsFor: 'debug support' stamp: 'eem 7/16/2013 14:15'!
assert: aBooleanExpression l: linenum
	<doNotGenerate>
	^self assert: aBooleanExpression! !

!VMClass methodsFor: 'debug support' stamp: 'eem 4/8/2009 18:01'!
asserta: aBooleanExpression
	<doNotGenerate>
	| result |
	(result := aBooleanExpression value) ifFalse:
		[AssertionFailure signal: 'Assertion failed'].
	^result! !

!VMClass methodsFor: 'debug support' stamp: 'eem 7/16/2013 14:15'!
asserta: aBooleanExpression l: linenum
	<doNotGenerate>
	^self asserta: aBooleanExpression! !

!VMClass methodsFor: 'simulation support' stamp: 'eem 6/21/2021 10:51'!
assertf: aString
	<doNotGenerate>
	AssertionFailure signal: 'Assertion failed: ', aString! !

!VMClass methodsFor: 'translation support' stamp: 'eem 11/11/2014 08:50'!
cCode: codeString
	"Support for Smalltalk-to-C translation.
	 For translation only; noop when running in Smalltalk.
	 The argument is output literally when generating C code."
	<doNotGenerate>! !

!VMClass methodsFor: 'translation support' stamp: 'eem 11/11/2014 08:48'!
cCode: codeStringOrBlock inSmalltalk: aBlock
	"Support for Smalltalk-to-C translation. The first argument is output when generating C code.
	  But if this code is being simulated in Smalltalk, answer the result of evaluating the given block.
	  If the first argument is a string it is output literally, and if it is a block it is translated.
	  N.B.  If the first argument is a block then replacement happens at TMethod creation time so the use
	  of cCode:inSmalltalk: with a block first argument does not prevent inlining and is hence preferred."
	<doNotGenerate>
	^aBlock value! !

!VMClass methodsFor: 'memory access' stamp: 'eem 10/13/2020 09:07'!
cCoerce: value to: cTypeString
	"Type coercion. For translation a cast will be emitted. When running in Smalltalk
	  answer a suitable wrapper for correct indexing."
	<doNotGenerate>
	^value coerceTo: cTypeString sim: self! !

!VMClass methodsFor: 'translation support' stamp: 'eem 11/11/2014 08:50'!
cPreprocessorDirective: codeString
	"For translation only; noop when running in Smalltalk."
	<doNotGenerate>! !

!VMClass methodsFor: 'translation support' stamp: 'eem 7/6/2021 14:59'!
cppIf: conditionBlockOrValue ifFalse: falseExpressionOrBlockOrNil
	"When translated, produces #if (condition) #else #endif CPP directives.
	 Example usage:

		self cppIf: #DUAL_MAPPED_CODE_ZONE
			ifFalse: [(backEnd needsCodeZoneExecuteWriteSwitch
					 and: [codeZoneIsExecutableNotWritable]) ifTrue:
						[backEnd makeCodeZoneWritable.
						codeZoneIsExecutableNotWritable := false]]"
	<doNotGenerate>
	^self cppIf: conditionBlockOrValue ifTrue: nil ifFalse: falseExpressionOrBlockOrNil! !

!VMClass methodsFor: 'translation support' stamp: 'eem 11/11/2014 08:50'!
cppIf: conditionBlockOrValue ifTrue: trueExpressionOrBlock
	"When translated, produces #if (condition) #else #endif CPP directives.
	 Example usage:

		self cppIf: IMMUTABILITY
			ifTrue: [(self internalIsImmutable: obj) ifTrue:
						[^self primitiveFailFor: PrimErrNoModification]]"
	<doNotGenerate>
	^self cppIf: conditionBlockOrValue ifTrue: trueExpressionOrBlock ifFalse: nil! !

!VMClass methodsFor: 'translation support' stamp: 'eem 7/25/2020 09:21'!
cppIf: firstCondition ifTrue: firstExpression cppIf: secondCondition ifTrue: secondExpression ifFalse: falseExpression
	<doNotGenerate>
	^self cppIf: firstCondition
		ifTrue: firstExpression
		ifFalse:
			[self cppIf: secondCondition ifTrue: secondExpression ifFalse: falseExpression]! !

!VMClass methodsFor: 'translation support' stamp: 'eem 8/12/2021 10:44'!
cppIf: conditionBlockOrSymbolValue ifTrue: trueExpressionOrBlock ifFalse: falseExpressionOrBlockOrNil
	"When translated, produces #if (condition) #else #endif CPP directives.
	 Example usage:

		self cppIf: [BytesPerWord = 8]
			ifTrue: [self doSomethingFor64Bit]
			ifFalse: [self doSomethingFor32Bit]
		self cppIf: BytesPerWord = 8
			ifTrue: [self doSomethingFor64Bit]
			ifFalse: [self doSomethingFor32Bit]
		self cppIf: #A_GLOBAL
			ifTrue: [self doSomethingFor64Bit]
			ifFalse: [self doSomethingFor32Bit]"
	<doNotGenerate>
	^(conditionBlockOrSymbolValue value
		ifNil: [false]
		ifNotNil: [:value|
			value isInteger
				ifTrue: [value ~= 0]
				ifFalse:
					[value isSymbol
						ifTrue: [self class valueOfSymbol: value]
						ifFalse: [value]]])
		ifTrue: trueExpressionOrBlock
		ifFalse: falseExpressionOrBlockOrNil! !

!VMClass methodsFor: 'simulation support' stamp: 'eem 8/8/2014 17:04'!
deny: aBooleanOrBlock
	<doNotGenerate>
	aBooleanOrBlock value ifTrue: [AssertionFailure signal: 'Assertion failed']! !

!VMClass methodsFor: 'simulation support' stamp: 'eem 9/7/2018 19:19'!
doOrDefer: aBlock
	<doNotGenerate>
	"Either evaluate aBlock immediately if in the uiProcess or defer aBlock as a UI message"
	Processor activeProcess == self uiProcess
		ifTrue: [aBlock value]
		ifFalse: [WorldState addDeferredUIMessage: aBlock]! !

!VMClass methodsFor: 'debug support' stamp: 'eem 4/22/2015 13:25'!
eassert: aBooleanExpressionOrBlock
	"This is for expensive asserts that we're only interested in checking in extremis.
	 For example now that Spur objStacks are debugged there's no benefit to
	 evaluating isValidObjStack: throughout the mark loop because its damn slow."
	<doNotGenerate>
	ExpensiveAsserts ifTrue:
		[aBooleanExpressionOrBlock value ifFalse:
			[AssertionFailure signal: 'Assertion failed']]! !

!VMClass methodsFor: 'printf logging' stamp: 'eem 11/13/2017 11:55'!
f: logFilename open: mode
	"Simulate a FILE stream by answering a transcript.
	 Can be used with the f:printf: method."
	<doNotGenerate>
	^true
		ifTrue: [(TranscriptStream on: (String new: 100000))
					openLabel: logFilename;
					yourself]
		ifFalse: [self coInterpreter transcript]! !

!VMClass methodsFor: 'memory access' stamp: 'eem 6/3/2010 09:00'!
fetchSingleFloatAtPointer: pointer into: aFloat
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and aFloat is a 32-bit single precision float."
	<doNotGenerate>

	^self fetchSingleFloatAt: pointer into: aFloat! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:52'!
floatAtPointer: pointer
	<doNotGenerate>
	self halt.! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:52'!
floatAtPointer: pointer put: value
	<doNotGenerate>
	self halt.! !

!VMClass methodsFor: 'C library simulation' stamp: 'RonieSalgado 4/29/2016 17:13'!
free: pointer
	<doNotGenerate>
	"Do nothing"! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:52'!
int16AtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is an 16 bit integer."
	<doNotGenerate>

	^self shortAt: pointer! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:52'!
int16AtPointer: pointer put: value
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is an 16 bit integer."
	<doNotGenerate>

	^self shortAt: pointer put: value! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:52'!
int32AtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is an 32 bit integer."
	<doNotGenerate>

	^self longAt: pointer! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:52'!
int32AtPointer: pointer put: value
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is an 32 bit integer."
	<doNotGenerate>

	^self longAt: pointer put: value! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:53'!
int64AtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is the width of a machine word."
	<doNotGenerate>

	^self halt! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:53'!
int64AtPointer: pointer put: longValue
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and longValue is the width of a machine word."
	<doNotGenerate>

	^self halt.! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:53'!
int8AtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is an 16 bit integer."
	<doNotGenerate>

	^self at: pointer! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:53'!
int8AtPointer: pointer put: value
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is an 16 bit integer."
	<doNotGenerate>

	^self at: pointer put: value! !

!VMClass methodsFor: 'testing' stamp: 'eem 11/2/2012 16:11'!
isCurrentImageFacade
	^false! !

!VMClass methodsFor: 'plugin support' stamp: 'eem 6/10/2010 16:54'!
isInterpreterProxy
	<doNotGenerate>
	"Return false since I am a real Interpreter simulation"
	^false! !

!VMClass methodsFor: 'hack compatibility' stamp: 'eem 9/7/2018 18:47'!
localNameFor: aString
	<doNotGenerate>
	^(Smalltalk classNamed: #FileSystem)
		ifNotNil: [:fs| (fs disk pathFromString: aString) basename]
		ifNil: [FileDirectory default localNameFor: aString]! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:53'!
long64AtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is the width of a machine word."
	<doNotGenerate>

	^self halt! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:53'!
long64AtPointer: pointer put: longValue
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and longValue is the width of a machine word."
	<doNotGenerate>

	^self halt.! !

!VMClass methodsFor: 'memory access' stamp: 'eem 6/3/2010 09:00'!
longAtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is the width of a machine word."
	<doNotGenerate>

	^self longAt: pointer! !

!VMClass methodsFor: 'memory access' stamp: 'eem 6/3/2010 09:00'!
longAtPointer: pointer put: longValue
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and longValue is the width of a machine word."
	<doNotGenerate>

	^self longAt: pointer put: longValue! !

!VMClass methodsFor: 'simulation support' stamp: 'eem 7/11/2011 17:14'!
majorVersion
	"This is implemented in sqVirtualMachine.c, so this form is for simulation only."
	<doNotGenerate>
	^self class vmProxyMajorVersion! !

!VMClass methodsFor: 'C library simulation' stamp: 'eem 10/22/2018 15:59'!
malloc: size
	<doNotGenerate>
	^CArrayAccessor on: (ByteArray new: size)! !

!VMClass methodsFor: 'C library simulation' stamp: 'akg 10/24/2018 15:56'!
memcpy: dest _: src _: bytes
	<doNotGenerate>
	"implementation of memcpy(3). N.B. If ranges overlap, must use memmove."
	| getBlock setBlock source destination |

	source := src isVMSimulationAddress
		ifTrue: [src asInteger]
		ifFalse: [src].
	destination := dest isVMSimulationAddress
		ifTrue: [dest asInteger]
		ifFalse: [dest].
	(source isInteger and: [destination isInteger]) ifTrue:
		[ self deny: ((destination <= source and: [destination + bytes > source])
					or: [source <= destination and: [source + bytes > destination]])].

	"Determine the source and destination access blocks based on the parameter type"
	getBlock := source isCollection ifTrue: 
		[source isString ifTrue: 
			"basicAt: answers integers"
			[[ :idx | source basicAt: idx]]
		ifFalse: 
			[source class == ByteArray ifTrue: 
				[[ :idx | source at: idx]]]]
	ifFalse: 
		[source isInteger ifTrue: 
			[[ :idx | self byteAt: source + idx - 1]]
		ifFalse: 
			[source isCArray ifTrue:
				[[ :idx | source at: idx - 1]]]].
	getBlock ifNil: [self error: 'unhandled type of source string'].
	setBlock := destination isCollection ifTrue: 
		[destination isString ifTrue:
			"basicAt:put: stores integers"
			[[ :idx | destination basicAt: idx put: (getBlock value: idx)]] 
		ifFalse: 
			[destination class == ByteArray ifTrue: 
				[[ :idx | destination at: idx put: (getBlock value: idx)]]]]
	ifFalse: 
		[destination isInteger ifTrue:
			[[ :idx | self byteAt: destination + idx - 1 put: (getBlock value: idx)]]
		ifFalse:
			[destination isCArray ifTrue:
				[[ :idx | destination at: idx - 1 put: (getBlock value: idx)]]]].
	setBlock ifNil: [self error: 'unhandled type of destination string'].
	1 to: bytes do: setBlock.

	^destination! !

!VMClass methodsFor: 'C library simulation' stamp: 'eem 9/19/2018 17:29'!
memmove: destAddress _: sourceAddress _: bytes
	<doNotGenerate>
	| dst src  |
	dst := destAddress asInteger.
	src := sourceAddress asInteger.
	"Emulate the c library memmove function"
	self assert: bytes \\ 4 = 0.
	destAddress > sourceAddress
		ifTrue:
			[bytes - 4 to: 0 by: -4 do:
				[:i| self longAt: dst + i put: (self longAt: src + i)]]
		ifFalse:
			[0 to: bytes - 4 by: 4 do:
				[:i| self longAt: dst + i put: (self longAt: src + i)]]! !

!VMClass methodsFor: 'simulation support' stamp: 'eem 1/18/2020 11:33'!
minorVersion
	"This is implemented in sqVirtualMachine.c, so this form is for simulation only."
	<doNotGenerate>
	^self class objectMemoryClass vmProxyMinorVersion! !

!VMClass methodsFor: 'translation support' stamp: 'eem 10/31/2012 15:35'!
numElementsIn: anArray
	<cmacro: '(anArray) (sizeof(anArray)/sizeof(anArray[0]))'>
	^anArray size! !

!VMClass methodsFor: 'accessing' stamp: 'eem 10/9/2013 18:17'!
objectRepresentationClass
	<doNotGenerate>
	^self class objectRepresentationClass! !

!VMClass methodsFor: 'oop comparison' stamp: 'eem 5/6/2015 18:01'!
oop: anOop isGreaterThan: otherOop
	"Compare two oop values, treating them as object memory locations; i.e. use unsigned comparisons.
	 Use a macro, instead of #cCoerce:to: to provide fast simulation and inline code in conditionals,
	 since the inliner doesn't inline in condtionals."
	<cmacro: '(anOop,otherOop) ((usqInt)(anOop) > (usqInt)(otherOop))'>
	^anOop > otherOop! !

!VMClass methodsFor: 'oop comparison' stamp: 'eem 5/6/2015 18:00'!
oop: anOop isGreaterThan: baseOop andLessThan: limitOop
	"Compare two oop values, treating them as object memory locations; i.e. use unsigned comparisons.
	 Use a macro, instead of #cCoerce:to: to provide fast simulation and inline code in conditionals,
	 since the inliner doesn't inline in condtionals."
	<cmacro: '(anOop,baseOop,limitOop) ((usqInt)(anOop) > (usqInt)(baseOop) && (usqInt)(anOop) < (usqInt)(limitOop))'>
	^anOop > baseOop and: [anOop < limitOop]! !

!VMClass methodsFor: 'oop comparison' stamp: 'eem 5/6/2015 18:01'!
oop: anOop isGreaterThanOrEqualTo: otherOop
	"Compare two oop values, treating them as object memory locations; i.e. use unsigned comparisons.
	 Use a macro, instead of #cCoerce:to: to provide fast simulation and inline code in conditionals,
	 since the inliner doesn't inline in condtionals."
	<cmacro: '(anOop,otherOop) ((usqInt)(anOop) >= (usqInt)(otherOop))'>
	^anOop >= otherOop! !

!VMClass methodsFor: 'oop comparison' stamp: 'eem 5/6/2015 18:00'!
oop: anOop isGreaterThanOrEqualTo: baseOop andLessThan: limitOop
	"Compare two oop values, treating them as object memory locations; i.e. use unsigned comparisons.
	 Use a macro, instead of #cCoerce:to: to provide fast simulation and inline code in conditionals,
	 since the inliner doesn't inline in condtionals."
	<cmacro: '(anOop,baseOop,limitOop) ((usqInt)(anOop) >= (usqInt)(baseOop) && (usqInt)(anOop) < (usqInt)(limitOop))'>
	^anOop >= baseOop and: [anOop < limitOop]! !

!VMClass methodsFor: 'oop comparison' stamp: 'eem 5/6/2015 17:59'!
oop: anOop isGreaterThanOrEqualTo: baseOop andLessThanOrEqualTo: limitOop
	"Compare two oop values, treating them as object memory locations; i.e. use unsigned comparisons.
	 Use a macro, instead of #cCoerce:to: to provide fast simulation and inline code in conditionals,
	 since the inliner doesn't inline in condtionals."
	<cmacro: '(anOop,baseOop,limitOop) ((usqInt)(anOop) >= (usqInt)(baseOop) && (usqInt)(anOop) <= (usqInt)(limitOop))'>
	^anOop >= baseOop and: [anOop <= limitOop]! !

!VMClass methodsFor: 'oop comparison' stamp: 'eem 5/6/2015 18:01'!
oop: anOop isLessThan: otherOop
	"Compare two oop values, treating them as object memory locations; i.e. use unsigned comparisons.
	 Use a macro, instead of #cCoerce:to: to provide fast simulation and inline code in conditionals,
	 since the inliner doesn't inline in condtionals."
	<cmacro: '(anOop,otherOop) ((usqInt)(anOop) < (usqInt)(otherOop))'>
	^anOop < otherOop! !

!VMClass methodsFor: 'oop comparison' stamp: 'eem 5/6/2015 18:01'!
oop: anOop isLessThanOrEqualTo: otherOop
	"Compare two oop values, treating them as object memory locations; i.e. use unsigned comparisons.
	 Use a macro, instead of #cCoerce:to: to provide fast simulation and inline code in conditionals,
	 since the inliner doesn't inline in condtionals."
	<cmacro: '(anOop,otherOop) ((usqInt)(anOop) <= (usqInt)(otherOop))'>
	^anOop <= otherOop! !

!VMClass methodsFor: 'memory access' stamp: 'eem 6/10/2010 16:16'!
oopForPointer: pointerOrSurrogate
	"This gets implemented by Macros in C, where its types will also be checked.
	 oop is the width of a machine word, and pointer is a raw address."
	<doNotGenerate>
	^pointerOrSurrogate asInteger! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:51'!
pointerAtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is the width of a machine word."
	<doNotGenerate>

	^self longAt: pointer! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:51'!
pointerAtPointer: pointer put: value
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is the width of a machine word."
	<doNotGenerate>

	^self longAt: pointer put: value! !

!VMClass methodsFor: 'translation support' stamp: 'eem 6/3/2010 09:00'!
pointerForOop: oop
	"This gets implemented by Macros in C, where its types will also be checked.
	 oop is the width of a machine word, and pointer is a raw address."
	<doNotGenerate>

	^oop! !

!VMClass methodsFor: 'simulation support' stamp: 'eem 2/3/2020 15:09'!
promptHex: string
	<doNotGenerate>
	| s |
	s := UIManager default request: string, ' (hex)'.
	s := s withBlanksTrimmed.
	^(s notEmpty and: ['-+0123456789abcdefABCDEF' includes: s first]) ifTrue:
		[(s beginsWith: 'ce')
			ifTrue:
				[self cogit trampolineAddressFor: s]
			ifFalse:
				[(s includes: $r)
					ifTrue:
						[Number readFrom: s readStream]
					ifFalse:
						[(#('0x' '-0x') detect: [:prefix| s beginsWith: prefix] ifNone: []) ifNotNil:
							[:prefix|
							s := s allButFirst: prefix size.
							prefix first = $- ifTrue: [s := '-', s]].
						Integer readFrom: s readStream base: 16]]]! !

!VMClass methodsFor: 'simulation support' stamp: 'eem 12/5/2013 11:22'!
promptNum: string
	<doNotGenerate>
	| s |
	s := UIManager default request: string, ' (dec)'.
	s := s withBlanksTrimmed.
	^s notEmpty ifTrue:
		[Number readFrom: s readStream]! !

!VMClass methodsFor: 'arithmetic' stamp: 'eem 6/23/2010 22:04'!
safe: signedInteger mod: positiveModulus
	<inline>
	| remainder |
	^(remainder := signedInteger \\ positiveModulus) < 0
		ifTrue: [remainder + positiveModulus]
		ifFalse: [remainder]! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:51'!
singleFloatAtPointer: pointer
	<doNotGenerate>
	self halt.! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:52'!
singleFloatAtPointer: pointer put: value
	<doNotGenerate>
	self halt.! !

!VMClass methodsFor: 'translation support' stamp: 'nice 11/4/2016 21:37'!
sizeof: objectSymbolOrClass
	<doNotGenerate>
	| index |
	objectSymbolOrClass isInteger ifTrue:
		[^self class objectMemoryClass wordSize].
	(#(usqInt sqInt) includes: objectSymbolOrClass) ifTrue: [^self class objectMemoryClass bytesPerOop].
	objectSymbolOrClass isSymbol ifTrue:
		[(objectSymbolOrClass last == $*
		 or: [#(#long #'unsigned long' #'sqIntptr_t'  #'usqIntptr_t' #'size_t') includes: objectSymbolOrClass]) ifTrue:
			[^self class objectMemoryClass wordSize].
		index := #(	#sqLong #usqLong #double
					#int #'unsigned int' #float
					#short #'unsigned short'
					#char #'unsigned char' #'signed char')
						indexOf: objectSymbolOrClass
						ifAbsent:
							[self error: 'unrecognized C type name'].
		^#(8 8 8
			4 4 4
			2 2
			1 1 1) at: index].
	^(objectSymbolOrClass isBehavior
		ifTrue: [objectSymbolOrClass]
		ifFalse: [objectSymbolOrClass class])
			alignedByteSizeOf: objectSymbolOrClass
			forClient: self! !

!VMClass methodsFor: 'simulation support' stamp: 'eem 7/7/2010 12:04'!
sqLowLevelMFence
	<doNotGenerate>
	"A no-op in the simulator"! !

!VMClass methodsFor: 'memory access' stamp: 'eem 6/3/2010 09:00'!
storeFloatAtPointer: pointer from: aFloat
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and aFloat is a 64-bit double precision float."
	<doNotGenerate>

	^self storeFloatAt: pointer from: aFloat! !

!VMClass methodsFor: 'memory access' stamp: 'eem 6/3/2010 09:00'!
storeSingleFloatAtPointer: pointer from: aFloat
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and aFloat is a 32-bit single precision float."
	<doNotGenerate>

	^self storeSingleFloatAt: pointer from: aFloat! !

!VMClass methodsFor: 'C library simulation' stamp: 'eem 9/19/2018 17:27'!
strcat: aString _: bString
	<doNotGenerate>
	"implementation of strcat(3)"
	^(self asString: aString), (self asString: bString)! !

!VMClass methodsFor: 'C library simulation' stamp: 'eem 11/14/2019 10:02'!
strcmp: aString _: bString
	<doNotGenerate>
	"implementation of strcmp(3); assumption is that aStrung is always a literal Smalltalk string."
	bString isString ifTrue:
		[1 to: (aString size min: bString size) do:
			[:i| | v |
			 v := (aString basicAt: i) - (bString basicAt: i).
			 v ~= 0 ifTrue: [^v]].
		 aString size = bString size ifTrue:
			[^0].
		 ^aString size < bString size
			ifTrue: [bString basicAt: aString size]
			ifFalse: [aString basicAt: bString size]].
	^self strcmp: aString bytes: bString! !

!VMClass methodsFor: 'C library simulation' stamp: 'eem 11/14/2019 10:04'!
strcmp: aString bytes: bString
	"Helper for implementation of strcmp(3); see strcmp:_:"
	<doNotGenerate>
	| i limit b v |
	i := 0.
	limit := aString size.
	[(i := i + 1) <= limit] whileTrue:
		[b := self byteAt: bString + i - 1.
		 b = 0 ifTrue:
			[^aString basicAt: i].
		 v := (aString basicAt: i) - b.
		 v ~= 0 ifTrue: [^v]].
	^self byteAt: bString + i - 1! !

!VMClass methodsFor: 'C library extensions' stamp: 'eem 11/15/2019 08:37'!
stretch: s1 cat: s2
	<var: 's1' type: #'char *'>
	<var: 's2' type: #'char *'>
	<returnTypeC: #'char *'>
	| ns |
	<var: 'ns' type: #'char *'>
	^self
		cCode:
			[ns := self malloc: (self strlen: s1) + (self strlen: s2) + 2.
			 self strcpy: ns _: s1.
			 self strcat: ns  _: s2]
		inSmalltalk: [self strcat: s1  _: s2]! !

!VMClass methodsFor: 'C library simulation' stamp: 'eem 10/25/2018 11:51'!
strlen: aCString
	"Simulate strlen(3)"
	<doNotGenerate>
	| len |

	aCString isCArray ifTrue:
		[len := 0.
		 [(aCString at: len) = 0 ifTrue: [^len].
		 len := len + 1] repeat]
	ifFalse:
		[aCString isString ifTrue: 
			[^aCString size]
		ifFalse:
			[aCString class == ByteArray ifTrue: [
				"ByteArrays may be 0 terminated or the correct length (in the simulator)"
				len := 0.
				[(len = aCString size or: [(aCString at: len+1) = 0]) ifTrue: [^len].
				len := len + 1] repeat]]].
	"Must be an address"
	len := 0.
	[(self byteAt: aCString + len) = 0 ifTrue: [^len].
	len := len + 1] repeat! !

!VMClass methodsFor: 'C library simulation' stamp: 'eem 10/19/2018 10:58'!
strncmp: aString _: bString _: n
	<doNotGenerate>
	"implementation of strncmp(3)"
	bString isString
		ifTrue:
			[1 to: n do:
				[:i| | v |
				 v := (aString basicAt: i) - (bString basicAt: i).
				 v ~= 0 ifTrue: [^v]]]
		ifFalse:
			[1 to: n do:
				[:i| | v |
				 v := (aString basicAt: i) - (self byteAt: bString + i - 1).
				 v ~= 0 ifTrue: [^v]]].
	^0! !

!VMClass methodsFor: 'C library simulation' stamp: 'eem 3/18/2020 19:29'!
strncpy: dest _: src _: n
	<doNotGenerate>
	"implementation of strncpy(3).
	 See e.g. https://manpages.debian.org/stretch/manpages-dev/strncpy.3.en.html
	 The C version always takes an address; the simulation allows a String, ByteArray,
	 CArray or address within the simulation object memory (Positive Integer)"
	| getBlock setBlock count |
	count := n.
	"Determine the source and destination access blocks based on the parameter type"
	getBlock := src isCollection
					ifTrue:
						[count := count min: src size.
						 src isString
							ifTrue: [[ :idx | src basicAt: idx]] "basicAt: answers integers"
							ifFalse:
								[src class == ByteArray ifTrue:
									[[ :idx | src at: idx]]]]
					ifFalse:
						[src isInteger
							ifTrue: [[ :idx | self byteAt: src + idx - 1]]
							ifFalse:
								[src isCArray ifTrue:
									[[ :idx | src at: idx - 1]]]].
	getBlock ifNil: [self error: 'unhandled type of source string'].
	setBlock := dest isCollection
					ifTrue:
						[dest isString
							ifTrue: [[ :idx | dest basicAt: idx put: (getBlock value: idx)]] "basicAt:put: stores integers"
							ifFalse:
								[dest class == ByteArray ifTrue:
									[[ :idx | dest at: idx put: (getBlock value: idx)]]]]
					ifFalse:
						[dest isInteger ifTrue: 
							[[ :idx | self byteAt: dest + idx - 1 put: (getBlock value: idx)]]
						 ifFalse:
							[dest isCArray ifTrue:
								[[ :idx | dest at: idx - 1 put: (getBlock value: idx)]]]].
	setBlock ifNil: [self error: 'unhandled type of destination string'].
	1 to: count do: setBlock.
	"SVr4, 4.3BSD, C89, C99 require the remainder of the buffer be filled with nulls"
	getBlock := [:idx| 0].
	count + 1 to: n do: setBlock.
	^dest! !

!VMClass methodsFor: 'hack compatibility' stamp: 'eem 9/7/2018 19:18'!
uiProcess
	<doNotGenerate>
	^(Smalltalk classNamed: #Project)
		ifNotNil: [:project| Project uiProcess] "Squeak"
		ifNil: [UIManager default uiProcess] "Pharo"! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:54'!
uint16AtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is an 16 bit integer."
	<doNotGenerate>

	^self shortAt: pointer! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:54'!
uint16AtPointer: pointer put: value
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is an 16 bit integer."
	<doNotGenerate>

	^self shortAt: pointer put: value! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:54'!
uint32AtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is an 32 bit integer."
	<doNotGenerate>

	^self longAt: pointer! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:54'!
uint32AtPointer: pointer put: value
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is an 32 bit integer."
	<doNotGenerate>

	^self longAt: pointer put: value! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:54'!
uint64AtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is the width of a machine word."
	<doNotGenerate>

	^self halt! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:54'!
uint64AtPointer: pointer put: longValue
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and longValue is the width of a machine word."
	<doNotGenerate>

	^self halt.! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:55'!
uint8AtPointer: pointer
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is an 16 bit integer."
	<doNotGenerate>

	^self at: pointer! !

!VMClass methodsFor: 'memory access' stamp: 'RonieSalgado 4/27/2016 13:55'!
uint8AtPointer: pointer put: value
	"This gets implemented by Macros in C, where its types will also be checked.
	 pointer is a raw address, and the result is an 16 bit integer."
	<doNotGenerate>

	^self at: pointer put: value! !

!VMClass methodsFor: 'simulation' stamp: 'eem 8/30/2013 15:50'!
unalignedAccessError
	^self error: 'unaligned access'! !

!VMClass methodsFor: 'debug support' stamp: 'eem 6/25/2015 13:57'!
unreachable
	<inline: true>
	self error: 'UNREACHABLE'! !

!VMClass methodsFor: 'hack compatibility' stamp: 'eem 9/7/2018 19:41'!
vmPath
	<doNotGenerate>
	^(Smalltalk classNamed: #VirtualMachine)
		ifNotNil: [:project| Smalltalk vm path] "Squeak"
		ifNil: [Smalltalk vmPath] "Pharo"! !

!VMClass methodsFor: 'hack compatibility' stamp: 'eem 9/7/2018 19:43'!
vmPathSize
	<doNotGenerate>
	^self vmPath size! !

!VMClass methodsFor: 'plugin api version' stamp: 'eem 1/18/2020 11:47'!
vmProxyMajorVersion
	"Define the  VM_PROXY_MAJOR version for this VM as used to
	 define the api in platforms/Cross/vm/sqVirtualMachine.[ch]
	 and in interp.h."
	<inline: #always>
	^self cCode: [#VM_PROXY_MAJOR] inSmalltalk: [self class objectMemoryClass vmProxyMajorVersion]! !

!VMClass methodsFor: 'plugin api version' stamp: 'eem 1/18/2020 11:48'!
vmProxyMinorVersion
	"Define the  VM_PROXY_MINOR version for this VM as used to
	 define the api in platforms/Cross/vm/sqVirtualMachine.[ch]
	 and in interp.h."
	<inline: #always>
	^self cCode: [#VM_PROXY_MINOR] inSmalltalk: [self class objectMemoryClass vmProxyMinorVersion]! !

VMClass subclass: #VMStructType
	instanceVariableNames: ''
	classVariableNames: 'StructTypeNameCache'
	poolDictionaries: ''
	category: 'VMMaker-Support'!

!VMStructType commentStamp: '<historical>' prior: 0!
I am an absract superclass for classes which map to simple structs in C,  Right now I merely serve to organize all the struct types.  I would like to arrange that generated assignment accessors answer their argument to correspond with field assignment in C, but createInstVarAccessors is implemented by the browser not by ClassDescription.!

VMStructType subclass: #AlignmentStructType
	instanceVariableNames: 'element'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!AlignmentStructType commentStamp: 'eem 5/30/2020 18:50' prior: 0!
AlignmentStruct is a hack to declare and generate all the structure field alignment types used in ThreadedFFIPlugin>>primitiveStructureElementAlignment.

Instance Variables
	element:		<char,short,int,long long, etc>

element
	- a C type that is defined according to the relevant type name
!

!AlignmentStructType class methodsFor: 'code generation' stamp: 'eem 5/30/2020 18:48'!
addStructTypeNamesTo: aSet
	aSet addAll: #('structByte' 'structShort' 'structInt' 'structLongLong' 'structFloat' 'structDouble' 'structStruct')! !

!AlignmentStructType class methodsFor: 'code generation' stamp: 'eem 5/31/2020 16:28'!
printTypedefOn: aStream
	#('Byte' 'Short' 'Int' 'LongLong' 'Float' 'Double')
	  with: #(#char #short #int #'long long' #float #double) do:
		[:structName :type|
		aStream
			nextPutAll: 'typedef struct { char pad_to_misalgnment; ';
			nextPutAll: type;
			nextPutAll: ' element; } struct';
			nextPutAll: structName;
			nextPut: $;; cr; cr].
	aStream
		nextPutAll: 'typedef struct { char pad_to_misalgnment; struct {char c; } element; } structStruct;';
		cr; cr! !

VMStructType subclass: #CogAbstractInstruction
	instanceVariableNames: 'opcode machineCodeSize maxSize annotation machineCode operands address dependent cogit objectMemory bcpc'
	classVariableNames: 'NumOperands'
	poolDictionaries: 'CogAbstractRegisters CogCompilationConstants CogRTLOpcodes'
	category: 'VMMaker-JIT'!

!CogAbstractInstruction commentStamp: 'eem 4/21/2015 09:12' prior: 0!
I am an abstract instruction generated by the Cogit.  I am subsequently concretized to machine code for the current processor.  A sequence of concretized CogAbstractInstructions are concatenated to form the code for a CogMethod.  I am an abstract class.  My concrete subclasses concretize to the machine code of a specific processor.

Instance Variables
	address:			<Integer>
	bcpc:				<Integer>
	cogit:				<Cogit>
	dependent:			<AbstractInstruction|nil>
	machineCode:		<CArray on: (ByteArray|Array)>
	machineCodeSize:	<Integer>
	maxSize:			<Integer>
	objectMemory:		<NewCoObjectMemory|SpurCoMemoryManager etc>
	opcode:			<Integer>
	operands:			<CArray on: Array>

address
	- the address at which the instruction will be generated

bcpc
	- the bytecode pc for which the instruction was generated; simulation only

cogit
	- the Cogit assembling the receiver; simulation only

dependent
	- a reference to another instruction which depends on the receiver, if any; in C this is a pointer

machineCode
	- the array of machine code the receiver generates when concretized

machineCodeSize
	- the size of machineCode in bytes

maxSize
	- the maximum size of machine code that the current instruction will generate, in bytes

objectMemory
	- the memory manager for the system; simulation only

opcode
	- the opcode for the receiver which defines which abstract opcode it represents; see CogRTLOpcodes class>>initialize and CogAbstractInstruction subclass initialize methods

operands
	- the array containing any operands the instruction may have; the opcode defines implicitly how many operands are consdered!

CogAbstractInstruction subclass: #CogARMCompiler
	instanceVariableNames: 'conditionOrNil'
	classVariableNames: 'AL AddOpcode AndOpcode BicOpcode CArg0Reg CArg1Reg CArg2Reg CArg3Reg CC CMPSMULL CPSRReg CS CmpNotOpcode CmpOpcode ConcreteIPReg ConcretePCReg ConcreteVarBaseReg D0 D1 D2 D3 D4 D5 D6 D7 EQ GE GT HI LDMFD LE LR LS LT MI MRS MSR MoveNotOpcode MoveOpcode NE OrOpcode OverflowFlag PC PL PopLDM PushSTM R0 R1 R10 R11 R12 R2 R3 R4 R5 R6 R7 R8 R9 RsbOpcode SMLALOpcode SMULL SP STMFD SubOpcode TstOpcode VC VS XorOpcode'
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogARMCompiler commentStamp: 'lw 8/23/2012 19:38' prior: 0!
I generate ARM instructions from CogAbstractInstructions.  For reference see
http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.set.architecture/index.html

The Architecture Reference Manual used is that of version 5, which includes some version 6 instructions. Of those, only pld is used(for PrefetchAw).

This class does not take any special action to flush the instruction cache on instruction-modification.!

!CogARMCompiler class methodsFor: 'accessing' stamp: 'eem 3/27/2015 18:11'!
IPReg
	"Answer the number of the general temp reg in the ARM APCS convention, IP"
	^ConcreteIPReg! !

!CogARMCompiler class methodsFor: 'translation' stamp: 'eem 4/21/2015 22:53'!
ISA
	"Answer the name of the ISA the receiver implements."
	^#ARMv5! !

!CogARMCompiler class methodsFor: 'accessing' stamp: 'eem 6/18/2015 14:41'!
PCReg
	^ConcretePCReg! !

!CogARMCompiler class methodsFor: 'accessing' stamp: 'eem 3/27/2015 18:45'!
VarBaseReg
	"Answer the number of the reg we use to hold the base address of CoInterpreter variables"
	^ConcreteVarBaseReg! !

!CogARMCompiler class methodsFor: 'translation' stamp: 'eem 6/15/2015 17:47'!
defaultCompilerClass
	^CogOutOfLineLiteralsARMCompiler! !

!CogARMCompiler class methodsFor: 'translation' stamp: 'eem 6/10/2015 16:40'!
filteredInstVarNames
	"Edit such that conditionOrNil is amongst the char size vars opcode machineCodeSize and maxSize."
	^(super filteredInstVarNames copyWithout: 'conditionOrNil')
		copyReplaceFrom: 5 to: 4 with: #('conditionOrNil')! !

!CogARMCompiler class methodsFor: 'translation' stamp: 'eem 3/6/2019 07:07'!
identifyingPredefinedMacros
	^#('__ARM_ARCH_5__' '__ARM_ARCH_6__' '__ARM_ARCH_7__' '__arm__' '__arm32__' 'ARM32' '_M_ARM')! !

!CogARMCompiler class methodsFor: 'class initialization' stamp: 'eem 6/1/2016 09:03'!
initialize

	"Initialize various ARM instruction-related constants."
	"CogARMCompiler initialize"

	super initialize.
	self ~~ CogARMCompiler ifTrue: [^self].

	"ARM general registers"
	R0 := 0.
	R1 := 1.
	R2 := 2.
	R3 := 3.
	R4 := 4.
	R5 := 5.
	R6 := 6.
	R7 := 7.
	R8 := 8.
	R9 := 9.
	R10 := 10.
	R11 := 11.
	R12 := 12.
	SP := 13.
	LR := 14.
	PC := 15.
	"ARM VFP Double precision floating point registers"
	D0 := 0.
	D1 := 1.
	D2 := 2.
	D3 := 3.
	D4 := 4.
	D5 := 5.
	D6 := 6.
	D7 := 7.

	CArg0Reg := 0.
	CArg1Reg := 1.
	CArg2Reg := 2.
	CArg3Reg := 3.

	ConcreteVarBaseReg := 10.
	ConcreteIPReg := 12. "IP, The Intra-Procedure-call scratch register."
	ConcretePCReg := 15.

	"Condition Codes. Note that cc=16rF is NOT ALLOWED as a condition; it specifies an extension instruction. See e.g.ARM_ARM v5 DDI01001.pdf A3.2.1"
	EQ := 0.
	NE := 1.
	CS := 2.
	CC := 3.
	MI := 4.
	PL := 5.
	VS := 6.
	VC := 7.
	HI := 8.
	LS := 9.
	GE := 10.
	LT := 11.
	GT := 12.
	LE := 13.
	AL := 14.

	"Table A3-2 in sec A3.4 Data-processing instructions of the AARM."
	AddOpcode := 	4.
	AndOpcode := 0.
	BicOpcode := 14.
	CmpOpcode := 10.
	CmpNotOpcode := 11.
	MoveOpcode := 13.
	MoveNotOpcode := 15.
	OrOpcode := 12.
	RsbOpcode := 3.
	SMLALOpcode := 7.
	SubOpcode := 2.
	TstOpcode := 8.
	XorOpcode := 1.

	CPSRReg := 16.
	OverflowFlag := 1 << 28.

	"Specific instructions"
	self
		initializeSpecificOpcodes: #(SMULL MSR MRS PopLDM PushSTM LDMFD STMFD CMPSMULL)
		in: thisContext method! !

!CogARMCompiler class methodsFor: 'class initialization' stamp: 'nice 2/9/2020 17:45'!
initializeAbstractRegisters
	"Assign the abstract registers with the identities/indices of the relevant concrete registers."

	TempReg			:= R0.
	ClassReg			:= R2.
	ReceiverResultReg	:= R5.
	SendNumArgsReg	:= R6.
	SPReg				:= SP. "a.k.a. R13" self assert: SP = 13.
	FPReg				:= R11.
	Arg0Reg			:= R3. "overlaps with last C arg reg"
	Arg1Reg			:= R4.
	Extra0Reg			:= R7.
	Extra1Reg			:= R8.
	Extra2Reg			:= R9.
	VarBaseReg		:= R10.	"Must be callee saved" self assert: ConcreteVarBaseReg = R10.
	RISCTempReg		:= R12.	"a.k.a. IP" self assert: ConcreteIPReg = R12.
	LinkReg				:= LR. "R14"
	PCReg				:= PC. "R15"	

	"According to IHI0042E ARM Architecture Procedure Calling Standard, in section 5.1.1:
		A subroutine must preserve the contents of the registers r4-r8, r10, r11 and SP (and r9 in PCS variants that designate r9 as v6).
	 SP = r13, so the callee-saved regs are r4-r8 & r10-r12.
	 The caller-saved registers are those that are not callee-saved and not reserved for hardware/abi uses,
	 i..e r0-r3, r9 & r12.
	 We exclude registers 0 & 1 (TempReg/CArg0Reg & CArg1Reg) from the CallerSavedRegisterMask because we only
	 use them for argument passing and so never want to save and restore them.  In fact restoring TempReg/CArg0Reg
	 would overwrite function results, so it shouldn't be included under any circumstances."

	ABICalleeSavedRegisterMask := self registerMaskFor: 4 and: 5 and: 6 and: 7 and: 10 and: 11.
	ABICallerSavedRegisterMask := self registerMaskFor: 0 and: 1 and: 2 and: 3 and: 9 and: 12.
	CallerSavedRegisterMask := ABICallerSavedRegisterMask
								bitAnd: (self registerMaskFor: ClassReg and: ReceiverResultReg and: SendNumArgsReg and: Arg0Reg and: Arg1Reg).
	ABIResultRegHigh := R1.

	NumRegisters := 16.

	DPFPReg0			:= D0.
	DPFPReg1			:= D1.
	DPFPReg2			:= D2.
	DPFPReg3			:= D3.
	DPFPReg4			:= D4.
	DPFPReg5			:= D5.
	DPFPReg6			:= D6.
	DPFPReg7			:= D7.

	NumFloatRegisters := 8! !

!CogARMCompiler class methodsFor: 'testing' stamp: 'eem 6/15/2015 19:22'!
isAbstract
	^self == CogARMCompiler! !

!CogARMCompiler class methodsFor: 'testing' stamp: 'eem 12/25/2015 18:47'!
isRISCTempRegister: reg
	"For tests to filter-out bogus values left in the RISCTempRegister, if any."
	^reg = ConcreteIPReg! !

!CogARMCompiler class methodsFor: 'translation' stamp: 'eem 3/11/2017 12:46'!
machineCodeDeclaration
	"Answer the declaration for the machineCode array.
	 ARM instructions are 32-bits in length."
	^{#'unsigned int'. '[', self basicNew machineCodeWords printString, ']'}! !

!CogARMCompiler class methodsFor: 'accessing' stamp: 'eem 3/18/2015 09:20'!
orOpcode
	^OrOpcode! !

!CogARMCompiler class methodsFor: 'class initialization' stamp: 'eem 6/7/2014 06:51'!
specificOpcodes
	"Answer the processor-specific opcodes for this class.
	 They're all in an Array literal in the initialize method."
	^(self class >> #initialize) literals detect: [:l| l isArray and: [l includes: #LDMFD]]! !

!CogARMCompiler class methodsFor: 'translation' stamp: 'eem 12/11/2015 23:28'!
wordSize
	"This is a 32-bit ISA"
	^4! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:38'!
add: destReg rn: srcReg imm: immediate ror: rot
"Remember the ROR is doubled by the cpu so use 30>>1 etc.
	ADD destReg, srcReg, #immediate ROR #rot - ARM_ARM v7 DDI10406 p. A8-23"

	^self type: 1 op: AddOpcode set: 0 rn: srcReg rd: destReg shifterOperand: ((rot>>1) <<8 bitOr: immediate)! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:55'!
add: destReg rn: srcReg rm: addReg
"return an ADD destReg, srcReg, addReg instruction
	ADD destReg, srcReg, addReg - ARM_ARM v7 DDI10406 p. A8-24"

	^self type: 0 op: AddOpcode set: 0 rn: srcReg rd: destReg shifterOperand: addReg! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:55'!
adds: destReg rn: srcReg imm: immediate ror: rot
"Remember the ROR is doubled by the cpu so use 30>>1 etc
	ADDS destReg, srcReg, #immediate ROR #rot - ARM_ARM v7 DDI10406 p. A8-23"

	^self type: 1 op: AddOpcode set: 1 rn: srcReg rd: destReg shifterOperand: ((rot>>1) <<8 bitOr: immediate)! !

!CogARMCompiler methodsFor: 'simulation' stamp: 'eem 11/11/2015 12:37:16'!
aeabiDiv: dividend Mod: divisor
"simulate the __aeabi_idivmod call"
	<doNotGenerate>
	| proc result top bottom|
	proc := cogit processor.
	top := proc convertInternalToInteger: dividend.
	bottom := proc convertInternalToInteger: divisor.
	
	proc r0: (result := proc convertIntegerToInternal: (top quo: bottom)).
	proc r1: (proc convertIntegerToInternal: (top rem: bottom)).
	^result! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'eem 5/30/2018 20:36'!
aeabiDivModFunctionAddr
	"Answer the address of the __aeabi_idivmod() call provided by the ARM low level libs to do an integer divide that returns the quo in R0 and rem in R1.
	 A word on the somewhat strange usage of idivmod herein; we need a declaration for the _aeabi_idivmod helper function, despite the fact that in a simple C program test, you don't.
	 To get that declaration we need a variable to hang it off; thus the non-existent var idivmod, and in simulation we need to simulate it, which is what aeabiDiv:Mod: does."
	<returnTypeC: #usqInt>
	<var: #idivmod declareC: 'extern void __aeabi_idivmod(int dividend, int divisor)'>

	^self cCode: '(usqInt)__aeabi_idivmod' inSmalltalk:[#aeabiDiv:Mod:]! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:55'!
and: destReg rn: srcReg imm: immediate ror: rot
"Remember the ROR is doubled by the cpu so use 30>>1 etc
	AND destReg, srcReg, #immediate ROR #rot - ARM_ARM v7 DDI10406 p. A8-34"

	^self type: 1 op: AndOpcode set: 0 rn: srcReg rd: destReg shifterOperand: ((rot>>1) <<8 bitOr: immediate)! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:55'!
ands: destReg rn: srcReg imm: immediate ror: rot
"Remember the ROR is doubled by the cpu so use 30>>1 etc
	ANDS destReg, srcReg, #immediate ROR #rot - ARM_ARM v7 DDI10406 p. A8-34"

	^self type: 1 op: AndOpcode set: 1 rn: srcReg rd: destReg shifterOperand: ((rot>>1) <<8 bitOr: immediate)! !

!CogARMCompiler methodsFor: 'register allocation' stamp: 'eem 5/30/2016 11:17'!
availableRegisterOrNoneFor: liveRegsMask
	"Answer an unused abstract register in the liveRegMask.
	 Subclasses with more registers can override to answer them.
	 N.B. Do /not/ allocate TempReg."
	<returnTypeC: #sqInt>
	(cogit register: Extra0Reg isInMask: liveRegsMask) ifFalse:
		[^Extra0Reg].
	(cogit register: Extra1Reg isInMask: liveRegsMask) ifFalse:
		[^Extra1Reg].
	(cogit register: Extra2Reg isInMask: liveRegsMask) ifFalse:
		[^Extra2Reg].
	^super availableRegisterOrNoneFor: liveRegsMask! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:55'!
b: offset
"return a B offset instruction; offset is signed 24bits of WORD offset, so +_32Mbyte range
	B offset - ARM_ARM v7 DDI10406 pp. A8-44-5"
	^self cond: AL br: 0 offset: offset
! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:55'!
bics: destReg rn: srcReg imm: immediate ror: rot
"Remember the ROR is doubled by the cpu so use 30>>1 etc
	BICS destReg, srcReg, #immediate ROR #rot - ARM_ARM v7 DDI10406 pp. A8-50-1"

	^self type: 1 op: BicOpcode set: 1 rn: srcReg rd: destReg shifterOperand: ((rot>>1) <<8 bitOr: immediate)! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:55'!
bl: offset
"return a BL offset instruction; offset is signed 24bits of WORD offset, so +_32Mbyte range. Return address is in LR
	BL offset - ARM_ARM v7 DDI10406 pp. A8-58-9"
	^self cond: AL br: 1 offset: offset
! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:55'!
blx: targetReg
	"Branch&link to the address in targetReg. Return address is in LR
	BLX targetReg - ARM_ARM v7 DDI10406 pp. A8-60-1"
	<inline: true>
	^self cond: AL bx: 1 target: targetReg
! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:56'!
bx: targetReg
	"Branch to address in targetReg. BX targetReg
	BX targetReg  - ARM_ARM v7 DDI10406 pp. A8-62-3"
	<inline: true>
	^self cond: AL bx: 0 target: targetReg
! !

!CogARMCompiler methodsFor: 'testing' stamp: 'rmacnak 11/3/2015 23:03:34'!
byteReadsZeroExtend
	^true! !

!CogARMCompiler methodsFor: 'accessing' stamp: 'eem 12/10/2019 09:05'!
callInstructionByteSize
	"ARMv8 calls and jumps span +/- 128 mb, more than enough for intra-zone calls and jumps."
	^4! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'tpr 11/11/2015 11:02'!
callTargetFromReturnAddress: callSiteReturnAddress
	"Answer the address that the call immediately preceding callSiteReturnAddress will jump to."
	"this is also used by #jumpLongTargetBeforeFollowingAddress:."
	| callDistance call |
	call := self instructionBeforeAddress: callSiteReturnAddress.
	self assert: ((self instructionIsB: call) or: [self instructionIsBL: call]).
	callDistance := self extractOffsetFromBL: call.
	"this is the pc's +8 offset, - the 4 byte correction for the previous instruction address"
	^callSiteReturnAddress + 4  + callDistance signedIntFromLong! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
canDivQuoRem
	^true! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
canMulRR
"we can do a MulRR be we can't simulate it correctly for some reason. More bug-fixing in the simulator one day"
	^true! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 6/1/2016 10:03'!
canPushPopMultipleRegisters
	<inline: true>
	^true! !

!CogARMCompiler methodsFor: 'accessing' stamp: 'eem 4/21/2015 09:17'!
codeGranularity
	"Answer the size in bytes of a unit of machine code."
	<inline: true>
	^4! !

!CogARMCompiler methodsFor: 'generate machine code' stamp: 'eem 1/21/2020 10:31'!
computeMaximumSize
	"Because we don't use Thumb, each ARM instruction has 4 bytes. Many
	 abstract opcodes need more than one instruction. Instructions that refer
	 to constants and/or literals depend on literals being stored in-line or out-of-line.

	 N.B.  The ^N forms are to get around the bytecode compiler's long branch
	 limits which are exceeded when each case jumps around the otherwise."

	opcode
		caseOf: {
		"Noops & Pseudo Ops"
		[Label]					-> [^0].
		[Literal]					-> [^4].
		[AlignmentNops]		-> [^(operands at: 0) - 4].
		[Fill32]					-> [^4].
		[Nop]					-> [^4].
		"Control"
		[Call]					-> [^4].
		[CallFull]				-> [^self literalLoadInstructionBytes + 4].
		[JumpR]					-> [^4].
		[Jump]					-> [^4].
		[JumpFull]				-> [^self literalLoadInstructionBytes + 4].
		[JumpLong]				-> [^4].
		[JumpZero]				-> [^4].
		[JumpNonZero]			-> [^4].
		[JumpNegative]			-> [^4].
		[JumpNonNegative]		-> [^4].
		[JumpOverflow]			-> [^4].
		[JumpNoOverflow]		-> [^4].
		[JumpCarry]			-> [^4].
		[JumpNoCarry]			-> [^4].
		[JumpLess]				-> [^4].
		[JumpGreaterOrEqual]	-> [^4].
		[JumpGreater]			-> [^4].
		[JumpLessOrEqual]		-> [^4].
		[JumpBelow]			-> [^4].
		[JumpAboveOrEqual]	-> [^4].
		[JumpAbove]			-> [^4].
		[JumpBelowOrEqual]	-> [^4].
		[JumpLongZero]		-> [^4].
		[JumpLongNonZero]	-> [^4].
		[JumpFPEqual]			-> [^8].
		[JumpFPNotEqual]		-> [^8].
		[JumpFPLess]			-> [^8].
		[JumpFPGreaterOrEqual]-> [^8].
		[JumpFPGreater]		-> [^8].
		[JumpFPLessOrEqual]	-> [^8].
		[JumpFPOrdered]		-> [^8].
		[JumpFPUnordered]		-> [^8].
		[RetN]					-> [^(operands at: 0) = 0 ifTrue: [4] ifFalse: [8]].
		[Stop]					-> [^4].

		"Arithmetic"
		[AddCqR]				-> [^self rotateable8bitSignedImmediate: (operands at: 0)
										ifTrue: [:r :i :n| 4]
										ifFalse: [self literalLoadInstructionBytes + 4]].
		[AddCqRR]				-> [^self rotateable8bitSignedImmediate: (operands at: 0)
										ifTrue: [:r :i :n| 4]
										ifFalse: [self literalLoadInstructionBytes + 4]].
		[AndCqR]				-> [^self rotateable8bitBitwiseImmediate: (operands at: 0)
										ifTrue: [:r :i :n| 4]
										ifFalse:
											[self literalLoadInstructionBytes = 4
												ifTrue: [8]
												ifFalse:
													[1 << (operands at: 0) highBit = ((operands at: 0) + 1)
														ifTrue: [8]
														ifFalse: [self literalLoadInstructionBytes + 4]]]].
		[AndCqRR]				-> [^self rotateable8bitBitwiseImmediate: (operands at: 0)
										ifTrue: [:r :i :n| 4]
										ifFalse:
											[self literalLoadInstructionBytes = 4
												ifTrue: [8]
												ifFalse:
													[1 << (operands at: 0) highBit = ((operands at: 0) + 1)
														ifTrue: [8]
														ifFalse: [self literalLoadInstructionBytes + 4]]]].
		[CmpCqR]				-> [^self rotateable8bitSignedImmediate: (operands at: 0)
											ifTrue: [:r :i :n| 4]
											ifFalse: [self literalLoadInstructionBytes + 4]].
		[OrCqR]					-> [^self rotateable8bitImmediate: (operands at: 0)
											ifTrue: [:r :i| 4]
											ifFalse: [self literalLoadInstructionBytes + 4]].
		[SubCqR]				-> [^self rotateable8bitSignedImmediate: (operands at: 0)
											ifTrue: [:r :i :n| 4]
											ifFalse: [self literalLoadInstructionBytes + 4]].
		[TstCqR]				-> [^self rotateable8bitImmediate: (operands at: 0)
											ifTrue: [:r :i| 4]
											ifFalse: [self literalLoadInstructionBytes + 4]].
		[XorCqR]				-> [^self rotateable8bitBitwiseImmediate: (operands at: 0)
										ifTrue: [:r :i :n| 4]
										ifFalse:
											[self literalLoadInstructionBytes = 4
												ifTrue: [8]
												ifFalse:
													[1 << (operands at: 0) highBit = ((operands at: 0) + 1)
														ifTrue: [8]
														ifFalse: [self literalLoadInstructionBytes + 4]]]].
		[AddCwR]				-> [^self literalLoadInstructionBytes + 4].
		[AndCwR]				-> [^self literalLoadInstructionBytes + 4].
		[CmpCwR]				-> [^self literalLoadInstructionBytes + 4].
		[OrCwR]				-> [^self literalLoadInstructionBytes + 4].
		[SubCwR]				-> [^self literalLoadInstructionBytes + 4].
		[XorCwR]				-> [^self literalLoadInstructionBytes + 4].
		[AddRR]				-> [^4].
		[AndRR]				-> [^4].
		[CmpRR]				-> [^4].
		[OrRR]					-> [^4].
		[XorRR]					-> [^4].
		[SubRR]				-> [^4].
		[AddRRR]				-> [^4].
		[SubRRR]				-> [^4].
		[NegateR]				-> [^4].
		[LoadEffectiveAddressMwrR]
									-> [^self rotateable8bitImmediate: (operands at: 0)
											ifTrue: [:r :i| 4]
											ifFalse: [self literalLoadInstructionBytes + 4]].

		[LogicalShiftLeftCqR]		-> [^4].
		[LogicalShiftRightCqR]		-> [^4].
		[ArithmeticShiftRightCqRR]	-> [^4].
		[LogicalShiftRightCqRR]	-> [^4].
		[LogicalShiftLeftCqRR]		-> [^4].
		[ArithmeticShiftRightCqR]	-> [^4].
		[LogicalShiftLeftRR]			-> [^4].
		[LogicalShiftRightRR]		-> [^4].
		[ArithmeticShiftRightRR]	-> [^4].
		[AddRdRd]					-> [^4].
		[CmpRdRd]					-> [^4].
		[SubRdRd]					-> [^4].
		[MulRdRd]					-> [^4].
		[DivRdRd]					-> [^4].
		[SqrtRd]					-> [^4].
		[ClzRR]					-> [^4].
		"ARM Specific Arithmetic"
		[SMULL]				-> [^4].
		[MSR]					-> [^4].
		[CMPSMULL]			-> [^4]. "special compare for genMulR:R: usage"
		"ARM Specific Data Movement"
		[PopLDM]				-> [^4].
		[PushSTM]				-> [^4].
		"Data Movement"						
		[MoveCqR]				-> [^self literalLoadInstructionBytes = 4
										ifTrue: [self literalLoadInstructionBytes]
										ifFalse:
											[self rotateable8bitBitwiseImmediate: (operands at: 0)
												ifTrue: [:r :i :n| 4]
												ifFalse: [self literalLoadInstructionBytes]]].
		[MoveCwR]				-> [^self literalLoadInstructionBytes = 4
										ifTrue: [self literalLoadInstructionBytes]
										ifFalse:
											[(self inCurrentCompilation: (operands at: 0))
												ifTrue: [4]
												ifFalse: [self literalLoadInstructionBytes]]].
		[MoveRR]				-> [^4].
		[MoveRdRd]				-> [^4].
		[MoveAwR]				-> [^(self isAddressRelativeToVarBase: (operands at: 0))
													ifTrue: [4]
													ifFalse: [self literalLoadInstructionBytes + 4]].
		[MoveRAw]				-> [^(self isAddressRelativeToVarBase: (operands at: 1))
													ifTrue: [4]
													ifFalse: [self literalLoadInstructionBytes + 4]].
		[MoveAbR]				-> [^(self isAddressRelativeToVarBase: (operands at: 0))
													ifTrue: [4]
													ifFalse: [self literalLoadInstructionBytes + 4]].
		[MoveRAb]				-> [^(self isAddressRelativeToVarBase: (operands at: 1))
													ifTrue: [4]
													ifFalse: [self literalLoadInstructionBytes + 4]].
		[MoveRMwr]			-> [^self is12BitValue: (operands at: 1)
										ifTrue: [:u :i| 4]
										ifFalse: [self literalLoadInstructionBytes + 4]].
		[MoveRdM64r]			-> [^4]. 
		[MoveMbrR]				-> [^self is12BitValue: (operands at: 0)
										ifTrue: [:u :i| 4]
										ifFalse: [self literalLoadInstructionBytes + 4]].
		[MoveRMbr]				-> [^self is12BitValue: (operands at: 1)
										ifTrue: [:u :i| 4]
										ifFalse: [self literalLoadInstructionBytes + 4]].
		[MoveRM16r]				-> [^self is12BitValue: (operands at: 1)
										ifTrue: [:u :i| 4]
										ifFalse: [self literalLoadInstructionBytes + 4]].
		[MoveM16rR]			-> [^self rotateable8bitImmediate: (operands at: 0)
											ifTrue: [:r :i| 4]
											ifFalse: [self literalLoadInstructionBytes + 4]].
		[MoveM64rRd]			-> [^4].
		[MoveMwrR]			-> [^self is12BitValue: (operands at: 0)
										ifTrue: [:u :i| 4]
										ifFalse: [self literalLoadInstructionBytes + 4]].
		[MoveXbrRR]			-> [^4].
		[MoveRXbrR]			-> [^4].
		[MoveXwrRR]			-> [^4].
		[MoveRXwrR]			-> [^4].
		[PopR]					-> [^4].
		[PushR]					-> [^4].
		[PushCw]				-> [^self literalLoadInstructionBytes = 4
										ifTrue: [self literalLoadInstructionBytes + 4]
										ifFalse:
											[(self inCurrentCompilation: (operands at: 0))
												ifTrue: [8]
												ifFalse:
													[self rotateable8bitBitwiseImmediate: (operands at: 0)
														ifTrue: [:r :i :n| 8]
														ifFalse: [self literalLoadInstructionBytes + 4]]]].
		[PushCq]				-> [^self literalLoadInstructionBytes = 4
										ifTrue: [self literalLoadInstructionBytes + 4]
										ifFalse:
											[self rotateable8bitBitwiseImmediate: (operands at: 0)
												ifTrue: [:r :i :n| 8]
												ifFalse: [self literalLoadInstructionBytes + 4]]].
		[PrefetchAw] 			-> [^(self isAddressRelativeToVarBase: (operands at: 0))
										ifTrue: [4]
										ifFalse: [self literalLoadInstructionBytes + 4]].
		"Conversion"
		[ConvertRRd]			-> [^8].
		}.
	^0 "to keep C compiler quiet"
! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeAddRdRd
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	"Add FP regRHS to FP regLHS and stick result in FP regLHS"
	| regLHS regRHS |
	regRHS := operands at: 0.
	regLHS := operands at: 1.
	machineCode at: 0 put:(self faddd: regLHS with: regRHS).
	^4
	! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:46'!
concretizeAlignmentNops
	<inline: true>
	"fill any slots with NOPs - in this case mov  r0, r0 - which is the NOP I always used to use"
	self assert: machineCodeSize \\ 4 = 0.
	0 to: machineCodeSize - 1 by: 4 do:
		[:p| self machineCodeAt: p put: 16rE1A00000].
	^machineCodeSize! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 6/23/2020 11:44'!
concretizeAndCqRR
	"Will get inlined into concretizeAt: switch."
	"AND is very important since it's used to mask all sorts of flags in the jit. We take special care to try to find compact ways to make the masks"
	<inline: true>
	| val srcReg dstReg |
	val := operands at: 0.
	srcReg := operands at: 1.
	dstReg := operands at: 2.
	self rotateable8bitBitwiseImmediate: val
		ifTrue:
			[:rot :immediate :invert|
			self machineCodeAt: 0 put: (invert
											ifTrue: [self bics: dstReg rn: srcReg imm: immediate ror: rot]
											ifFalse: [self ands: dstReg rn: srcReg imm: immediate ror: rot]).
			^4]
		ifFalse:
			[| hb |
			hb := (operands at: 0) highBit.
			"First see if the constant can be made from a simple shift of 0xFFFFFFFF"
			1 << hb = (val +1) ifTrue: "MVN temp reg, 0, making 0xffffffff"
				[self machineCodeAt: 0 put:(self mvn: ConcreteIPReg imm: 0 ror: 0).
				"Then AND reg, temp reg, lsr #(32-hb)"
				 self machineCodeAt: 4 put: (self dataOpType: AndOpcode rd: dstReg rn: srcReg rm: ConcreteIPReg lsr: 32 - hb).
				^8]].
	^self concretizeDataOperationCwR: AndOpcode R: dstReg! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/26/2019 17:12'!
concretizeArithmeticShiftRightCqR
	"Will get inlined into concretizeAt: switch."
	"this is an unfortunate waste of an instruction in most cases since the shift can usually be done in a subsequent arithmetic instruction. 
	Handle for now with a MOVS reg, reg, ASR #distance"
	<inline: true>
	| distance reg |
	distance := (operands at: 0).
	reg := operands at: 1.
	self assert: (distance between: 1 and: 31).
	"cond 000 1101 0 0000 dest dist -100 srcR"
	self machineCodeAt: 0 put: (self type: 0 op: MoveOpcode set: 1 rn: 0 rd: reg 
									shifterOperand: (distance << 7 bitOr: (64 "flag for arithmetic" bitOr: reg))).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/26/2019 17:12'!
concretizeArithmeticShiftRightCqRR
	"Will get inlined into concretizeAt: switch."
	"this is an unfortunate waste of an instruction in most cases since the shift can usually be done in a subsequent arithmetic instruction. 
	Handle for now with a MOVS reg, reg, ASR #distance"
	<inline: true>
	| distance srcReg destReg |
	distance := (operands at: 0).
	srcReg := operands at: 1.
	destReg := operands at: 2.
	self assert: (distance between: 1 and: 31).
	"cond 000 1101 0 0000 dest dist -100 srcR"
	self machineCodeAt: 0 put: (self type: 0 op: MoveOpcode set: 1 rn: 0 rd: destReg 
									shifterOperand: (distance << 7 bitOr: (64 "flag for arithmetic" bitOr: srcReg))).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeArithmeticShiftRightRR
	"Will get inlined into concretizeAt: switch."
	"this is an unfortunate waste of an instruction in most cases since the shift can usually be done in a subsequent arithmetic instruction. 
	Handle for now with a MOVS reg, reg, ASR distReg"
	<inline: true>
	| destReg distReg |
	distReg := operands at: 0.
	destReg := operands at: 1.
	"cond 000 1101 0 0000 destR distR 0101 srcR"
	self machineCodeAt: 0 put: (self type: 0 op: MoveOpcode set: 1 rn: 0 rd: destReg 
									shifterOperand: (distReg << 8 bitOr: (80 bitOr: destReg))).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code' stamp: 'eem 9/22/2015 08:52'!
concretizeAt: actualAddress
	"Generate concrete machine code for the instruction at actualAddress,
	 setting machineCodeSize, and answer the following address."

	self assert: actualAddress \\ 4 = 0.
	^super concretizeAt: actualAddress! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCMPSMULL
	"Generate a CMP a, b, ASR #31 instruction, specifically for comparing the resutls of SMULLs in genMulR:R:"
	| hiReg loReg |
	hiReg := operands at: 0.
	loReg := operands at: 1.
	self machineCodeAt: 0
		put: (self type: 0 op: CmpOpcode set: 1 rn: hiReg rd: 0)
			+ (31<<7) "the shift amount"
			+ (2<<5) "the shift type - ASR"
			+ loReg.
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCall
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	"Call is used only for calls within code-space, See CallFull for general anywhere in address space calling"
	| offset |
	self assert: (operands at: 0) ~= 0.
	self assert: (operands at: 0) \\ 4 = 0.
	offset := (operands at: 0) signedIntFromLong - (address + 8 "normal pc offset") signedIntFromLong.
	self assert: (self isInImmediateJumpRange: offset). "+- 24Mb is plenty of range in code space"
	self machineCodeAt: 0 put: (self bl: offset).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCallFull
	"Will get inlined into concretizeAt: switch."
	"Sizing/generating calls.
		Jump targets can be to absolute addresses or other abstract instructions.
		Generating initial trampolines instructions may have no maxSize and be to absolute addresses.
		Otherwise instructions must have a machineCodeSize which must be kept to."
	<inline: true>
	| jumpTarget instrOffset|
	<var: #jumpTarget type: #'AbstractInstruction *'>
	jumpTarget := self longJumpTargetAddress.
	instrOffset := self moveCw: jumpTarget intoR: ConcreteIPReg.
	"blx ConcreteIPReg"
	self machineCodeAt: instrOffset put: (self blx: ConcreteIPReg).
	self assert: instrOffset = self literalLoadInstructionBytes.
	^instrOffset + 4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/23/2020 15:59'!
concretizeClzRR
	"Count leading zeros
	First operand is output (dest)
	Second operand is input (mask)"
	"v5 CLZ cond 0001 0110 SBO Rd SBO 0001 Rm
	That is hexa 16rE16FxF1x for cond = AL"
	"v7 CLZ 11111-010-1-011 Rm(4bits) 1111 Rd(4bits) 1000 Rm(4bits)
	That is hexa 16rFABxFx8x"
	<inline: true>
	| maskReg dest  |
	maskReg := operands at: 0.
	dest := operands at: 1.
	self machineCodeAt: 0 put:
		AL << 28 + 16r16F0F10
		+ (dest << 12) + maskReg.
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCmpRdRd
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	"Compare FP regB with FP regA and leave the FP status reg ready to be transferred back to ARM with next instruction"
	| regB regA |
	regA := operands at:0.
	regB := operands at: 1.
	machineCode at: 0 put:(self fcmpFrom: regB to: regA).
	^4
		
! !

!CogARMCompiler methodsFor: 'generate machine code' stamp: 'eem 12/8/2019 16:17'!
concretizeConditionalInstruction
	"Concretize the current instruction, but with a condition."
	| savedCond |
	self assert: conditionOrNil notNil.
	savedCond := conditionOrNil.
	conditionOrNil := nil.
	machineCodeSize := self dispatchConcretize.
	conditionOrNil := savedCond.
	0 to: machineCodeSize-1 by: 4 do:
		[:i| | instr |
		instr := (self machineCodeAt: i) bitClear: 16rF<<28.
		self machineCodeAt: i put: (instr bitOr: (conditionOrNil bitAnd: 16rF)<<28)].
	^machineCodeSize! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConditionalJump: conditionCode
	"Will get inlined into concretizeAt: switch."
	"Sizing/generating jumps.
		Jump targets can be to absolute addresses or other abstract instructions.
		Generating initial trampolines instructions may have no maxSize and be to absolute addresses.
		Otherwise instructions must have a machineCodeSize which must be kept to."
	<inline: true>
	| offset |
	offset := self computeJumpTargetOffsetPlus: 8.
 	self assert: (self isInImmediateJumpRange: offset).
	self machineCodeAt: 0 put: (self cond: conditionCode br: 0 offset: offset). "B offset"
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConvertRRd
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg destReg |
	srcReg := operands at:0.
	destReg := operands at: 1.
	machineCode at: 0 put:(self fmsrFrom: srcReg to: 9).
	machineCode at: 1 put: (self fsitodFrom: 9 to: destReg). "probably not quite right"
	^8
		
! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 16:19'!
concretizeDataOperationCqR: armOpcode
	"Will get inlined into concretizeAt: switch."
	"4 == Add, 2 == Sub, Xor == 1, And == 0, Or == 12, Bic == 14"
	<inline: true>
	|val rd rn |
	val := operands at: 0.
	rn := operands at: 1.
	rd := opcode = CmpOpcode ifTrue: [0] ifFalse:[rn]. "Extra note - if ever a version of this code wants to NOT set the Set flag - Cmp must always have it set or it will pretend to be a SMALALBT and Very Bad Things might happen"

	self  rotateable8bitImmediate: val 
		ifTrue: [:rot :immediate |
			self machineCodeAt: 0 put: (self type: 1 op: armOpcode set: 1 rn: rn rd: rd shifterOperand: (rot>>1"in this usage we have to halve the rot value" << 8 bitOr: immediate)).
			^4]
		ifFalse: [].
	"first see if the constant can be made from a simple shift of 0xFFFFFFFF"
	val > 0 ifTrue:
		[| hb |
		hb := val highBit.
		1 << hb = (val +1) ifTrue: "MVN temp,  #0, making 0xffffffff"
			[self machineCodeAt: 0 put:(self mvn: ConcreteIPReg imm: 0 ror: 0).
			 "Then armOpcode reg, temp reg, lsr #(32-hb)"
			 self machineCodeAt: 4 put:(self dataOpType: armOpcode rd: rd  rn: rn rm: ConcreteIPReg lsr: (32-hb)).
			 ^8]].
	^self concretizeDataOperationCwR: armOpcode! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/21/2020 10:41'!
concretizeDataOperationCwR: armOpcode
	"Will get inlined into concretizeAt: switch."
	"Load the word into the RISCTempReg, then cmp R, RISCTempReg"
	<inline: #always>
	^self concretizeDataOperationCwR: armOpcode R: (armOpcode = CmpOpcode ifTrue: [0] ifFalse: [operands at: 1])! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/21/2020 10:35'!
concretizeDataOperationCwR: armOpcode R: rd
	"Will get inlined into concretizeAt: switch."
	"Load the word into the RISCTempReg, then cmp R, RISCTempReg"
	<inline: true>
	| constant rn instrOffset|
	constant := operands at: 0.
	rn := operands at: 1.
	instrOffset := self moveCw: constant intoR: ConcreteIPReg.
	self machineCodeAt: instrOffset 
		put: (self type: 0 op: armOpcode set: 1 rn: rn rd: rd shifterOperand: ConcreteIPReg).
	^instrOffset + 4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/13/2019 19:28'!
concretizeDataOperationRR: armOpcode
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| rn rd srcReg |
	srcReg := operands at: 0.
	rn := operands at: 1.
	rd := armOpcode = CmpOpcode ifTrue: [0] ifFalse: [rn].
	self machineCodeAt: 0 
		put: (self type: 0 op: armOpcode set: 1 rn: rn rd: rd shifterOperand: srcReg).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/13/2019 19:29'!
concretizeDataOperationRRR: armOpcode
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| rn rd srcReg |
	srcReg := operands at: 0.
	rn := operands at: 1.
	rd := operands at: 2.
	self machineCodeAt: 0 
		put: (self type: 0 op: armOpcode set: 1 rn: rn rd: rd shifterOperand: srcReg).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeDivRdRd
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	"FP divide regLHS by regRHS and stick result in regLHS"
	| regLHS regRHS |
	regRHS := operands at: 0.
	regLHS := operands at: 1.
	machineCode at: 0 put:(self fdivd: regLHS by: regRHS).
	^4
	! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeFPConditionalJump: conditionCode
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| offset |
	"transfer the FP status to ARM cpsr and then jump accordingly"
	offset := self computeJumpTargetOffsetPlus: 8+4 "pc is always 2 instr ahead plus add another to refer to the actual branch".
 	self assert: (self isInImmediateJumpRange: offset).
	self machineCodeAt: 0 put: self fmstat. "FMSTAT: copy the FPSCR to CPSR"
	self machineCodeAt: 4 put: (self cond: conditionCode br: 0 offset: offset). "B offset"
	^8! !

!CogARMCompiler methodsFor: 'generate machine code' stamp: 'eem 12/8/2019 15:23'!
concretizeFill32
	"fill with operand 0 according to the processor's endianness"
	self machineCodeAt: 0 put: (operands at: 0).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 16:20'!
concretizeInvertibleDataOperationCqR: armOpcode
	"Will get inlined into concretizeAt: switch."
	"Xor == 1, And == 0, Or == 12, Bic == 14"
	<inline: true>
	| val rn |
	val := operands at: 0.
	rn := operands at: 1.
	self deny: opcode = CmpOpcode.

	self rotateable8bitBitwiseImmediate: val 
		ifTrue:
			[:rot :immediate : invert|
			self machineCodeAt: 0
				put: (self
						type: 1
						op: (invert ifTrue: [self inverseOpcodeFor: armOpcode] ifFalse: [armOpcode])
						set: 1
						rn: rn
						rd: rn
						shifterOperand: (rot >> 1"in this usage we have to halve the rot value" << 8 bitOr: immediate)).
			^4]
		ifFalse: [].
	"first see if the constant can be made from a simple shift of 0xFFFFFFFF"
	val > 0 ifTrue:
		[| hb |
		hb := val highBit.
		1 << hb = (val +1) ifTrue: "MVN temp,  #0, making 0xffffffff"
			[self machineCodeAt: 0 put: (self mvn: ConcreteIPReg imm: 0 ror: 0).
			 "Then armOpcode reg, temp reg, lsr #(32-hb)"
			 self machineCodeAt: 4 put: (self dataOpType: armOpcode rd: rn  rn: rn rm: ConcreteIPReg lsr: 32 - hb).
			^8]].
	 ^self concretizeDataOperationCqR: armOpcode! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeJumpFull
	"Will get inlined into concretizeAt: switch."
	"A JumpFull is used when we need to jump to anywhere in 32bit address space rather than somewhere known to be in code-space. It also must be relocatable and non-varying with the jump range. On ARM this means using the build-long-const + BX sequence."
	<inline: true>
	| jumpTarget instrOffset|
	<var: #jumpTarget type: #'AbstractInstruction *'>
	jumpTarget := self longJumpTargetAddress.
	instrOffset := self moveCw: jumpTarget intoR: ConcreteIPReg.
	"bx ConcreteIPReg"
	self machineCodeAt: instrOffset put: (self bx: ConcreteIPReg).
	^instrOffset + 4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeJumpR
	"Will get inlined into concretizeAt: switch."
	"Sizing/generating jumps.
		Jump targets can be to absolute addresses or other abstract instructions.
		Generating initial trampolines instructions may have no maxSize and be to absolute addresses.
		Otherwise instructions must have a machineCodeSize which must be kept to."
	<inline: true>
	| reg |
	reg := operands at: 0.
	"bx reg"
	self machineCodeAt: 0 put: (self bx: reg).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/12/2020 18:12'!
concretizeLoadEffectiveAddressMwrR
	"Will get inlined into concretizeAt: switch."
	"destReg = srcReg + offset, but does not need to set condition codes, etc"
	<inline: true>
	| srcReg offset destReg instrOffset |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	^self rotateable8bitImmediate: offset
		ifTrue:
			[ :rot :immediate | 
			self machineCodeAt: 0 
				"add destReg, srcReg, #immediate ROR rot"
				put: (self add: destReg rn: srcReg imm: immediate ror: rot).
			4]
		ifFalse:
			[instrOffset := self moveCw: offset intoR: ConcreteIPReg.
			"add destReg, srcReg, ConcreteIPReg"
			self machineCodeAt: instrOffset put: (self add: destReg rn: srcReg rm: ConcreteIPReg).
			instrOffset + 4]

	"cogit processor disassembleInstructionAt: 0 In: machineCode object"! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/26/2019 17:12'!
concretizeLogicalShiftLeftCqR
	"Will get inlined into concretizeAt: switch."
	"this is an unfortunate waste of an instruction in most cases since the shift can usually be done in a subsequent arithmetic instruction. 
	Handle for now with a MOVS reg, reg, LSL #distance"
	<inline: true>
	| distance reg |
	distance := (operands at: 0).
	reg := operands at: 1.
	self assert: (distance between: 1 and: 31).
	"cond 000 1101 0 0000 dest dista 000 srcR"
	self machineCodeAt: 0 put: (self type: 0 op: MoveOpcode set: 1 rn: 0 rd: reg shifterOperand: (distance << 7 bitOr: reg)).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/26/2019 17:12'!
concretizeLogicalShiftLeftCqRR
	"Will get inlined into concretizeAt: switch."
	"this is an unfortunate waste of an instruction in most cases since the shift can usually be done in a subsequent arithmetic instruction. 
	Handle for now with a MOVS reg, reg, LSL #distance"
	<inline: true>
	| distance srcReg destReg |
	distance := (operands at: 0).
	srcReg := operands at: 1.
	destReg := operands at: 2.
	self assert: (distance between: 1 and: 31).
	"cond 000 1101 0 0000 dest dista 000 srcR"
	self machineCodeAt: 0 put: (self type: 0 op: MoveOpcode set: 1 rn: 0 rd: destReg shifterOperand: (distance << 7 bitOr: srcReg)).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeLogicalShiftLeftRR
	"Will get inlined into concretizeAt: switch."
	"this is an unfortunate waste of an instruction in most cases since the shift can usually be done in a subsequent arithmetic instruction. 
	Handle for now with a MOVS reg, reg, LSL distReg" 
	<inline: true>
	| destReg distReg |
	distReg := operands at: 0.
	destReg := operands at: 1.
	"cond 000 1101 0 0000 dest dist 0001 srcR"
	self machineCodeAt: 0 put: (self type: 0 op: MoveOpcode set: 1 rn: 0 rd: destReg 
									shifterOperand: (distReg << 8 bitOr: (16 bitOr: destReg))).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/26/2019 17:11'!
concretizeLogicalShiftRightCqR
	"Will get inlined into concretizeAt: switch."
	"this is an unfortunate waste of an instruction in most cases since the shift can usually be done in a subsequent arithmetic instruction. 
	Handle for now with a MOVS reg, reg, LSR #distance"
	<inline: true>
	| distance reg |
	distance := (operands at: 0).
	reg := operands at: 1.
	self assert: (distance between: 1 and: 31).
	"cond 000 1101 0 0000 dest dist -010 srcR"
	self machineCodeAt: 0 put: (self type: 0 op: MoveOpcode set: 1 rn: 0 rd: reg 
									shifterOperand: (distance << 7 bitOr: (32 bitOr: reg))).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/26/2019 17:11'!
concretizeLogicalShiftRightCqRR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| distance srcReg destReg |
	distance := (operands at: 0).
	srcReg := operands at: 1.
	destReg := operands at: 2.
	self assert: (distance between: 1 and: 31).
	"cond 000 1101 0 0000 dest dist -010 srcR"
	self machineCodeAt: 0 put: (self type: 0 op: MoveOpcode set: 1 rn: 0 rd: destReg 
									shifterOperand: (distance << 7 bitOr: (32 bitOr: srcReg))).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeLogicalShiftRightRR
	"Will get inlined into concretizeAt: switch."
	"this is an unfortunate waste of an instruction in most cases since the shift can usually be done in a subsequent arithmetic instruction. 
	Handle for now with a MOVS reg, reg, LSR distReg"
	<inline: true>
	| destReg distReg |
	distReg := operands at: 0.
	destReg := operands at: 1.
	"cond 000 1101 0 0000 dest dist 0011 srcR"
	self machineCodeAt: 0 put: (self type: 0 op: MoveOpcode set: 1 rn: 0 rd: destReg 
									shifterOperand: (distReg << 8 bitOr: (48 bitOr: destReg))).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMSR
	"Generate an MSR CPSR_f, #flags instruction.
Note that we only have business with the NZCV flags so we use
N -> 8
Z -> 4
C -> 2
V -> 1.
You don't want to mess with this too much."
	|  flags |
	flags := operands at: 0.
	self machineCodeAt: 0 put: (self msr: flags).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveAbR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcAddr destReg instrOffset|
	srcAddr := operands at: 0.
	destReg := operands at: 1.
	(self isAddressRelativeToVarBase: srcAddr) ifTrue:
		[self machineCodeAt: 0 put: (self ldrb: destReg rn: ConcreteVarBaseReg plus: 1 imm: srcAddr - cogit varBaseAddress).
		 ^4].
	"load the address into ConcreteIPReg"
	instrOffset := self moveCw: srcAddr intoR: ConcreteIPReg.
	"We *could* overwrite the last instruction above with a LDR a, b, last-byte-of-srcAddr BUT that would break if we change to loading literals instead of forming long constants"
	self machineCodeAt: instrOffset put: (self ldrb: destReg rn: ConcreteIPReg plus: 1 imm: 0).
	^instrOffset + 4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveAwR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcAddr destReg instrOffset|
	srcAddr := operands at: 0.
	destReg := operands at: 1.
	(self isAddressRelativeToVarBase: srcAddr) ifTrue:
		[self machineCodeAt: 0 put: (self ldr: destReg rn: ConcreteVarBaseReg plusImm: srcAddr - cogit varBaseAddress).
		 ^4].
	"load the address into ConcreteIPReg"
	instrOffset := self moveCw: srcAddr intoR: ConcreteIPReg.
	"We *could* overwrite the last instruction above with a LDR a, b, last-byte-of-srcAddr BUT that would break if we change to loading literals instead of forming long constants"
	self machineCodeAt: instrOffset put: (self ldr: destReg rn: ConcreteIPReg plusImm: 0).
	^instrOffset + 4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:14'!
concretizeMoveCqR
	"Will get inlined into concretizeAt: switch."
	"If the quick constant is in fact a shiftable 8bit, generate the apropriate MOV, otherwise do what is necessary for a whole word."
	<var: #word type: #sqInt>
	<inline: true>
	| word reg invVal |
	word := operands at: 0.
	reg := operands at: 1.
	self 
		rotateable8bitImmediate: word 
		ifTrue:
			[:rot :immediate |
			self machineCodeAt: 0 put: (self mov: reg imm: immediate ror: rot).
			^4]
		ifFalse:
			[].
	invVal := word < 0 ifTrue: [-1 - word] ifFalse: [word bitInvert32].
	^self rotateable8bitImmediate: invVal
		ifTrue: [ :rot :immediate |
			self machineCodeAt: 0 put: (self mvn: reg imm: immediate ror: rot).
			4]
		ifFalse: [self concretizeMoveCwR]! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveCwR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	^self loadCwInto: (operands at: 1)! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveM16rR
	"Will get inlined into concretizeAt: switch."
	"ldrh destReg, [srcReg, #immediate],
	or 
	move offset to ConcreteIPReg
	ldrh destReg, [srcReg, ConcreteIPReg]"
	<var: #offset type: #sqInt>
	<inline: true>
	| srcReg offset destReg instrOffset|
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	self is8BitValue: offset
		ifTrue:
			[ :u :immediate | 
			self machineCodeAt: 0 "ldrh destReg, [srcReg, #immediate]"
				put: (self ldrh: destReg rn: srcReg plus: u imm: immediate).
			^4]
		ifFalse:
			[instrOffset := self moveCw: offset intoR: ConcreteIPReg.
			"ldrh destReg, [srcReg, ConcreteIPReg]"
			self machineCodeAt: instrOffset put: (self ldrh: destReg rn: srcReg rm: ConcreteIPReg).
			^instrOffset + 4 ].
	^0 "to keep Slang happy"! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveM64rRd
	"Will get inlined into concretizeAt: switch."	
	<inline: true>
	"Load a float from srcReg+offset into FP destReg"
	| srcReg offset destReg u|
	offset := operands at: 0.
	u := offset >0 ifTrue:[1] ifFalse:[0].
	srcReg := operands at: 1.
	destReg := operands at: 2.
	machineCode at: 0 put: (self fldd: destReg rn: srcReg plus: u imm: offset>>2).
	^4
! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:29'!
concretizeMoveMbrR
	"Will get inlined into concretizeAt: switch."
	"ldrb destReg, [srcReg, #immediate] or ldrb destReg, [srcReg, ConcreteIPReg]"
	<var: #offset type: #sqInt>
	<inline: true>
	| srcReg offset destReg instrOffset|
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	^self is12BitValue: offset
		ifTrue:
			[ :u :immediate | 
			self machineCodeAt: 0 "ldrb destReg, [srcReg, #immediate]"
				put: (self ldrb: destReg rn: srcReg plus: u imm: immediate).
			4]
		ifFalse:
			[(self isAddressRelativeToVarBase: offset)
				ifTrue:
					[self machineCodeAt: 0 put: (self adds: ConcreteIPReg rn: ConcreteVarBaseReg imm: offset - cogit varBaseAddress ror: 0).
					 instrOffset := 4]
				ifFalse:
					[instrOffset := self moveCw: offset intoR: ConcreteIPReg].
			 "ldrb destReg, [srcReg, ConcreteIPReg]"
			 self machineCodeAt: instrOffset put: (self ldrb: destReg rn: srcReg rm: ConcreteIPReg).
			 instrOffset + 4]! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:29'!
concretizeMoveMwrR
	"Will get inlined into concretizeAt: switch."
	<var: #offset type: #sqInt>
	<inline: true>
	| srcReg offset destReg instrOffset|
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	^self is12BitValue: offset
		ifTrue:
			[ :u :immediate | 
			self machineCodeAt: 0 "ldr destReg, [srcReg, #immediate]"
				put: (self ldr: destReg rn: srcReg plus: u imm: immediate).
			4]
		ifFalse:
			[instrOffset := self moveCw: offset intoR: ConcreteIPReg.
			"ldr destReg, [srcReg, ConcreteIPReg]"
			self machineCodeAt: instrOffset put: (self ldr: destReg rn: srcReg rm: ConcreteIPReg).
			instrOffset + 4]! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRAb
	"Will get inlined into concretizeAt: switch."
	"LEA ConcreteIPReg
	strb srcReg, [ConcreteIPReg]"
	<inline: true>
	| srcReg destAddr instrOffset|
	srcReg := operands at: 0.
	destAddr := operands at: 1.
	(self isAddressRelativeToVarBase: destAddr) ifTrue:
		[self machineCodeAt: 0 put: (self strb: srcReg rn: ConcreteVarBaseReg plus: 1 imm: destAddr - cogit varBaseAddress).
		 ^4].
	"load the address into ConcreteIPReg"
	instrOffset := self moveCw: destAddr intoR: ConcreteIPReg.
	"We *could* overwrite the last instruction above with a LDR a, b, last-byte-of-srcAddr BUT that would break if we change to loading literals instead of forming long constants"
	self machineCodeAt: instrOffset put: (self strb: srcReg rn: ConcreteIPReg plus: 1 imm: 0).
	^instrOffset + 4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRAw
	"Will get inlined into concretizeAt: switch."
	"LEA ConcreteIPReg
	str srcReg, [ConcreteIPReg]"
	<inline: true>
	| srcReg destAddr instrOffset|
	srcReg := operands at: 0.
	destAddr := operands at: 1.
	(self isAddressRelativeToVarBase: destAddr) ifTrue:
		[self machineCodeAt: 0 put: (self str: srcReg rn: ConcreteVarBaseReg plusImm: destAddr - cogit varBaseAddress).
		 ^4].
	"load the address into ConcreteIPReg"
	instrOffset := self moveCw: destAddr intoR: ConcreteIPReg.
	"We *could* overwrite the last instruction above with a LDR a, b, last-byte-of-srcAddr BUT that would break if we change to loading literals instead of forming long constants"
	self machineCodeAt: instrOffset put: (self str: srcReg rn: ConcreteIPReg plusImm: 0).
	^instrOffset + 4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:30'!
concretizeMoveRM16r
	"Will get inlined into concretizeAt: switch."
	<var: #offset type: #sqInt>
	<inline: true>
	| srcReg offset baseReg instrOffset|
	srcReg := operands at: 0.
	offset := operands at: 1.
	baseReg := operands at: 2.
	^self is12BitValue: offset
		ifTrue:
			[ :u :immediate | 
			self machineCodeAt: 0 "strh 	srcReg, [baseReg, #immediate]"
				put: (self strh: srcReg rn: baseReg plus: u imm: immediate).
			4]
		ifFalse:
			[(self isAddressRelativeToVarBase: offset)
				ifTrue:
					[self machineCodeAt: 0 put: (self adds: ConcreteIPReg rn: ConcreteVarBaseReg imm: offset - cogit varBaseAddress ror: 0).
					 instrOffset := 4]
				ifFalse:
					[instrOffset := self moveCw: offset intoR: ConcreteIPReg].
			"strb 	srcReg, [baseReg, ConcreteIPReg]"
			self machineCodeAt: instrOffset put: (self strh: srcReg rn: baseReg rm: ConcreteIPReg).
			instrOffset + 4]! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:30'!
concretizeMoveRMbr
	"Will get inlined into concretizeAt: switch."
	<var: #offset type: #sqInt>
	<inline: true>
	| srcReg offset baseReg instrOffset|
	srcReg := operands at: 0.
	offset := operands at: 1.
	baseReg := operands at: 2.
	^self is12BitValue: offset
		ifTrue:
			[ :u :immediate | 
			self machineCodeAt: 0 "strb 	srcReg, [baseReg, #immediate]"
				put: (self strb: srcReg rn: baseReg plus: u imm: immediate).
			4]
		ifFalse:
			[(self isAddressRelativeToVarBase: offset)
				ifTrue:
					[self machineCodeAt: 0 put: (self adds: ConcreteIPReg rn: ConcreteVarBaseReg imm: offset - cogit varBaseAddress ror: 0).
					 instrOffset := 4]
				ifFalse:
					[instrOffset := self moveCw: offset intoR: ConcreteIPReg].
			"strb 	srcReg, [baseReg, ConcreteIPReg]"
			self machineCodeAt: instrOffset put: (self strb: srcReg rn: baseReg rm: ConcreteIPReg).
			instrOffset + 4]! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:30'!
concretizeMoveRMwr
	"Will get inlined into concretizeAt: switch."
	<var: #offset type: #sqInt>
	<inline: true>
	| srcReg offset baseReg instrOffset|
	srcReg := operands at: 0.
	offset := operands at: 1.
	baseReg := operands at: 2.
	^self is12BitValue: offset
		ifTrue:
			[ :u :immediate | 
			self machineCodeAt: 0  "str 	srcReg, [baseReg, #immediate]"
				put: (self str: srcReg rn: baseReg plus: u imm: immediate).
			4]
		ifFalse:
			[instrOffset := self moveCw: offset intoR: ConcreteIPReg.
			"str srcReg, [baseReg, ConcreteIPReg]"
			self machineCodeAt: instrOffset put: (self str: srcReg rn: baseReg rm: ConcreteIPReg).
			instrOffset + 4]! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	"cond 000 1101 0 0000 dest 0000 0000 srcR"
	self machineCodeAt: 0 put: (self mov: destReg rn: srcReg).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRXbrR
	"Will get inlined into concretizeAt: switch."
	"Write the word in R(src) into memory at address (base+1*index)"
	<inline: true>
	| index base src |
	src := operands at: 0.
	index := operands at: 1.
	base := operands at: 2.
	"str	b	src, [base, +index, LSL #0]"
	"cond 011 1100 0 base srcR 00000 00 0 index"
	self machineCodeAt: 0 put: (self strb: src rn: base rm: index).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRXwrR
	"Will get inlined into concretizeAt: switch."
	"Write the word in R(src) into memory at address (base+4*index)"
	<inline: true>
	| index base src |
	src := operands at: 0.
	index := operands at: 1. "index is number of *words* = 4* bytes"
	base := operands at: 2.
	"str		src, [base, +index, LSL #2]"
	"cond 011 1100 0 base srcR 00010 00 0 inde"
	self machineCodeAt: 0 put: (self memMxr: AL reg: src base: base p: 1 u: 1 b: 0 w: 0 l: 0 rmLsl2: index).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRdM64r
	"Will get inlined into concretizeAt: switch."	
	<inline: true>
	"Store FP fpReg to dstReg+offset"
	| dstReg offset fpReg u|
	offset := operands at: 1.
	u := offset >0 ifTrue:[1] ifFalse:[0].
	dstReg := operands at: 2.
	fpReg := operands at: 0.
	machineCode at: 0 put: (self fstd: fpReg rn: dstReg plus: u imm: offset>>2).
	^4
! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveXbrRR
	"Will get inlined into concretizeAt: switch."
	
	<inline: true>
	| index base dest |
	index := operands at: 0. "index is number of *bytes*"
	base := operands at: 1.
	dest := operands at: 2.
	"LDRB	dest, [base, +index, LSL #0]"
	"cond 011 1100 1 base dest 00000 00 0 inde"
	self machineCodeAt: 0 put: (self ldrb: dest rn: base rm: index).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveXwrRR
	"Will get inlined into concretizeAt: switch."
	
	<inline: true>
	| index base dest |
	index := operands at: 0.
	base := operands at: 1.
	dest := operands at: 2.
	"LDR	dest, [base, +index, LSL #2]"
	"cond 011 1100 1 base dest 00010 00 0 inde bulit by lowest level generator so we can do the lsl #2 on the index register"
	self machineCodeAt: 0 put: (self memMxr: AL reg: dest base: base p: 1 u: 1 b: 0 w: 0 l: 1 rmLsl2: index).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMulRdRd
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	"FP multiply regLHS by regRHS and stick result in regLHS"
	| regLHS regRHS |
	regRHS := operands at: 0.
	regLHS := operands at: 1.
	machineCode at: 0 put:(self fmuld: regLHS with: regRHS).
	^4
	! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeNegateR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| reg |
	reg := operands at: 0.
	"RSB destReg, srcReg, #0"
	self machineCodeAt: 0 put: (self type: 1 op: RsbOpcode set: 0 rn: reg rd: reg).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/21/2020 10:36'!
concretizeNegateableDataOperationCqR: armOpcode
	"Will get inlined into concretizeAt: switch."
	"4 == Add, 2 == Sub, 10 = Cmp"
	<inline: true>
	| val rd rn |
	val := operands at: 0.
	rn := operands at: 1.
	"Extra note - if ever a version of this code wants to NOT set the Set flag
	 - Cmp must always have it set or it will pretend to be a SMALALBT and Very Bad Things might happen."
	rd := opcode = CmpOpcode ifTrue: [0] ifFalse: [rn].

	self rotateable8bitSignedImmediate: val 
		ifTrue:
			[:rot :immediate : negate|
			self machineCodeAt: 0
				put: (self
						type: 1
						op: (negate ifTrue: [self inverseOpcodeFor: armOpcode] ifFalse: [armOpcode])
						set: 1
						rn: rn
						rd: rd
						shifterOperand: (rot >> 1"in this usage we have to halve the rot value" << 8 bitOr: immediate)).
			^4]
		ifFalse: "let's try to see if the constant can be made from a simple shift of 0xFFFFFFFF"
			[val > 0 ifTrue:
				[| hb |
				hb := val highBit.
				1 << hb = (val +1) ifTrue: "MVN temp,  #0, making 0xffffffff"
					[self machineCodeAt: 0 put:(self mvn: ConcreteIPReg imm: 0 ror: 0).
					 "Then armOpcode reg, temp reg, lsr #(32-hb)"
					 self machineCodeAt: 4 put:(self dataOpType: armOpcode rd: rd  rn: rn rm: ConcreteIPReg lsr: 32 - hb).
					^8]].
			 ^self concretizeDataOperationCwR: armOpcode R: rd].
	^0 "to keep Slang happy"! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/21/2020 10:35'!
concretizeNegateableDataOperationCqR: armOpcode R: rd
	"Will get inlined into concretizeAt: switch."
	"4 == Add, 2 == Sub, 10 = Cmp"
	<inline: true>
	| val rn |
	val := operands at: 0.
	rn := operands at: 1.
	"Extra note - if ever a version of this code wants to NOT set the Set flag
	 - Cmp must always have it set or it will pretend to be a SMALALBT and Very Bad Things might happen."

	self rotateable8bitSignedImmediate: val 
		ifTrue:
			[:rot :immediate : negate|
			self machineCodeAt: 0
				put: (self
						type: 1
						op: (negate ifTrue: [self inverseOpcodeFor: armOpcode] ifFalse: [armOpcode])
						set: 1
						rn: rn
						rd: rd
						shifterOperand: (rot >> 1"in this usage we have to halve the rot value" << 8 bitOr: immediate)).
			^4]
		ifFalse: "let's try to see if the constant can be made from a simple shift of 0xFFFFFFFF"
			[val > 0 ifTrue:
				[| hb |
				hb := val highBit.
				1 << hb = (val +1) ifTrue: "MVN temp,  #0, making 0xffffffff"
					[self machineCodeAt: 0 put:(self mvn: ConcreteIPReg imm: 0 ror: 0).
					 "Then armOpcode reg, temp reg, lsr #(32-hb)"
					 self machineCodeAt: 4 put:(self dataOpType: armOpcode rd: rd  rn: rn rm: ConcreteIPReg lsr: 32 - hb).
					^8]].
			 ^self concretizeDataOperationCwR: armOpcode R: rd].
	^0 "to keep Slang happy"! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeNop
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	"(CogARMCompiler new  mov: 0 rn: 0 ) hex -> MOV r0, r0"
	self machineCodeAt: 0 put: 16rE1A00000.
	^4
			! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePopR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| destReg |
	destReg := operands at: 0.
	"LDR destReg, [SP], #4"
	self machineCodeAt: 0 put: (self popR: destReg).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePrefetchAw
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| addressOperand instrOffset|
	addressOperand := operands at: 0.
	(self isAddressRelativeToVarBase: addressOperand) ifTrue:
		[self machineCodeAt: 0 put: (self pld: ConcreteVarBaseReg plus: 1 offset: addressOperand - cogit varBaseAddress).
		 ^4].
	instrOffset := self moveCw: addressOperand intoR: ConcreteIPReg.
	"pld	[ConcreteIPReg]"
	self machineCodeAt: instrOffset put: (self pld: ConcreteIPReg plus: 1offset: 0).
	^instrOffset + 4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePushCq
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| word instrOffset |
	word := operands at: 0.
	self 
		rotateable8bitBitwiseImmediate: word 
		ifTrue:
			[:rot :immediate :invert|
			self machineCodeAt: 0
				put: (invert
						ifTrue: [self mvn: ConcreteIPReg imm: immediate ror: rot]
						ifFalse: [self mov: ConcreteIPReg imm: immediate ror: rot]).
			instrOffset := 4]
		ifFalse:
			[instrOffset := self moveCw: word intoR: ConcreteIPReg].
	self machineCodeAt: instrOffset put: (self pushR: ConcreteIPReg).
	^instrOffset + 4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePushCw
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| word instrOffset |
	word := operands at: 0.	
	(self inCurrentCompilation: word)
		ifTrue:
			[instrOffset := self loadCwInto: ConcreteIPReg]
		ifFalse:
			[self 
				rotateable8bitBitwiseImmediate: word 
				ifTrue:
					[:rot :immediate :invert|
					self machineCodeAt: 0
						put: (invert
								ifTrue: [self mvn: ConcreteIPReg imm: immediate ror: rot]
								ifFalse: [self mov: ConcreteIPReg imm: immediate ror: rot]).
					instrOffset := 4]
				ifFalse:
					[instrOffset := self loadCwInto: ConcreteIPReg]].
	self machineCodeAt: instrOffset put: (self pushR: ConcreteIPReg).
	^instrOffset + 4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePushOrPopMultipleRegisters: doPush
	self assert: (operands at: 0) ~= 0.
	machineCode at: 0 put: AL << 28
						  + (doPush     "2r100PUSWL"
								ifTrue: [2r10010010 << 20]
								ifFalse: [2r10001011 << 20])
						  + (SP << 16)
						  + (operands at: 0).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePushR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg |
	srcReg := operands at: 0.
	"cond | 010 | 1001 | 0 | -Rn- | -Rd- | 0000 0000 0100" "STR srcReg, [sp, #-4]"
	self machineCodeAt: 0 put: (self pushR: srcReg).
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeRetN
	"Will get inlined into concretizeAt: switch."
	<var: #offset type: #sqInt>
	<inline: true>
	| offset |
	offset := operands at: 0.
	offset = 0 ifTrue:
		[self machineCodeAt: 0 put: (self mov: PC rn: LR). "pop	{pc}"
		^4].
	self assert: offset < 255. "We have an 8 bit immediate. If needed, we could rotate it less than 30 bit."

	self machineCodeAt: 0 put: (self add: SP rn: SP imm: offset  ror: 0).
	self machineCodeAt: 4 put: (self mov: PC rn: LR).  "pop	{pc}"
	^8! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSMULL
	| srcA srcB hiResultReg loResultReg |
	"Generate an SMULL loResultReg, hiResultReg, srcA, srcB instruction"
	srcA := operands at: 0.
	"NOTE: srcB contains the other mutiplicand at this point. It is OK to use it as the destination for the low part of the result and in fact this saves us moving it later"
	loResultReg := srcB := operands at: 1.
	hiResultReg := RISCTempReg.
	self machineCodeAt: 0
		put: (self type: 0 op: 6 set: 0 rn: hiResultReg rd: loResultReg)
			+ (srcA << 8)
			+ (9 << 4)
			+ srcB.
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSqrtRd
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	"Square root of FP regLHS into regLHS"
	| regLHS  |
	regLHS := operands at: 0.
	machineCode at: 0 put:(self fsqrtd: regLHS).
	^4
	! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeStop
	<inline: true>
	self machineCodeAt: 0 put: self stop.
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/21/2020 10:38'!
concretizeSubCqR
	"Will get inlined into concretizeAt: switch."
	"Try whether the quick constant is a small negative number. If it is, optimize."
	<var: #word type: #sqInt>
	<inline: true>
	| word reg |
	word := operands at: 0.
	reg := operands at: 1.
	self rotateable8bitImmediate: word
		ifTrue: [ :rot :immediate |
			self machineCodeAt: 0 put: (self subs: reg rn: reg imm: immediate ror: rot).
			^4]
		ifFalse: [].
	"before building a full load of a big constant, see if we can do an add of the constant negated"
	self rotateable8bitImmediate: word negated
		ifTrue: [ :rot :immediate |
			self machineCodeAt: 0 put: (self adds: reg rn: reg imm: immediate ror: rot).
			^4]
		ifFalse: [].
	^self concretizeDataOperationCwR: SubOpcode R: reg! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSubRdRd
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| regLHS regRHS |
	"Subtract FP regRHS from FP regLHS and leave the result in FP regLHS"
	regRHS := operands at: 0.
	regLHS := operands at: 1.
	machineCode at: 0 put:(self fsubd: regLHS with: regRHS).
	^4
	! !

!CogARMCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/21/2020 10:39'!
concretizeTstCqR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| reg |
	reg := operands at: 1.
	self rotateable8bitImmediate: (operands at: 0)
		ifTrue: [ :rot :immediate |
			self machineCodeAt: 0 put: (self tst: reg rn: reg imm: immediate ror: rot).
			^4]
		ifFalse: [].
	^self concretizeDataOperationCwR: TstOpcode R: reg! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 5/1/2014 17:06'!
cond: c br: link offset: offset
	"c : 4 bit, opcode = 10 bitOr: link, offset >>2, limited to 24 bits (which are sign-extended, shifted left 2 and added to 8 + pc to make the resulting address)"
	"single instr Branch, no link"
	<inline: true>
	^ c << 28 bitOr: (((2r1010 bitOr: (link bitAnd: 1)) << 24) bitOr: (offset >> 2 bitAnd: 16r00FFFFFF))! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 5/1/2014 17:11'!
cond: c bx: link target: targetReg
	"c : 4 bit, opcode = 10 bitOr: link, offset >>2, limited to 24 bits (which are sign-extended, shifted left 2 and added to 8 + pc to make the resulting address)"
	"BX targetReg or BLX targetReg"
	<inline: true>
	^ c << 28 bitOr: ( (16r12FFF10  bitOr: (link bitAnd: 1) <<5 ) bitOr: targetReg)! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 3/12/2015 18:53'!
cond: c type: t op: o set: s
	"c : 4 bit, t: 3 bit, o: 4 bit, s: 1bit"
	"cccctttoooos + oxFFFFF - the leftmost 12bits of (most) ARM instruction. The other 20 bits get built elsewhere"
	<inline: true>
	^ c << 28 bitOr: ((t << 25) bitOr: ((o << 21) bitOr: (s << 20)))! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 4/29/2014 15:59'!
cond: conditionCode type: type op: flagsOrOpcode set: doUpdateStatusRegister rn:  sourceRegister rd: targetRegister
"build an instruction - cccctttoooo + source + target"
	<inline: true>
	^(self cond: conditionCode type: type op: flagsOrOpcode set: doUpdateStatusRegister) 
		bitOr: (sourceRegister << 16 bitOr: targetRegister << 12)! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 3/12/2015 18:54'!
cond: conditionCode type: type op: flagsOrOpcode set: doUpdateStatusRegister rn:  sourceRegister rd: targetRegister shifterOperand: so
"build an instruction - cccctttoooo + source + target + shifter op"
	<inline: true>
	^(self cond: conditionCode type: type op: flagsOrOpcode set: doUpdateStatusRegister rn: sourceRegister rd: targetRegister) bitOr: (so bitAnd: 16rFFF)! !

!CogARMCompiler methodsFor: 'testing' stamp: 'tpr 11/22/2015 11:11'!
conditionIsNotNever: instr
	"test for the NV condition code; this isn't allowed as an actual condition and is used to encdoe many of the newer instructions"
	^instr >> 28 < 16rF ! !

!CogARMCompiler methodsFor: 'accessing' stamp: 'tpr 4/28/2015 16:41'!
conditionOrNil
"has to be named oddly like this to satisfay i-var code gen translating rules"
	^conditionOrNil! !

!CogARMCompiler methodsFor: 'accessing' stamp: 'tpr 4/28/2015 16:41'!
conditionOrNil: condCode
"has to be named oddly like this to satisfay i-var code gen translating rules"
	^conditionOrNil := condCode! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'eem 4/21/2015 23:24'!
dataOpType: armOpcode rd: destReg rn: srcReg rm: addReg lsr: shft
"return an {opcode} destReg, srcReg, addReg lsl #shft"
"important detail - a 0 shft requires setting the shift-type code to 0 to avoid potential instruction confusion"
	shft = 0
		ifTrue:[^self type: 0 op: armOpcode set: 1 rn: srcReg rd: destReg shifterOperand: addReg]
		ifFalse:[^self type: 0 op: armOpcode set: 1 rn: srcReg rd: destReg shifterOperand: ((shft <<7 bitOr: 32) bitOr:  addReg)]! !

!CogARMCompiler methodsFor: 'generate machine code' stamp: 'eem 1/21/2020 10:33'!
dispatchConcretize
	"Attempt to generate concrete machine code for the instruction at address.
	 This is the inner dispatch of concretizeAt: actualAddress which exists only
	 to get around the branch size limits in the SqueakV3 (blue book derived)
	 bytecode set."
	conditionOrNil ifNotNil:
		[^self concretizeConditionalInstruction].
		 
	opcode caseOf: {
		"Noops & Pseudo Ops"
		[Label]					-> [^self concretizeLabel].
		[Literal]					-> [^self concretizeLiteral].
		[AlignmentNops]		-> [^self concretizeAlignmentNops].
		[Fill32]					-> [^self concretizeFill32].
		[Nop]					-> [^self concretizeNop].
		"Control"
		[Call]						-> [^self concretizeCall]. "call code within code space"
		[CallFull]					-> [^self concretizeCallFull]. "call code anywhere in address space"
		[JumpR]					-> [^self concretizeJumpR].
		[JumpFull]					-> [^self concretizeJumpFull]."jump within address space"
		[JumpLong]				-> [^self concretizeConditionalJump: AL]."jump within code space"
		[JumpLongZero]			-> [^self concretizeConditionalJump: EQ].
		[JumpLongNonZero]		-> [^self concretizeConditionalJump: NE].
		[Jump]						-> [^self concretizeConditionalJump: AL]. "jump within a method, etc"
		[JumpZero]					-> [^self concretizeConditionalJump: EQ].
		[JumpNonZero]				-> [^self concretizeConditionalJump: NE].
		[JumpNegative]				-> [^self concretizeConditionalJump: MI].
		[JumpNonNegative]			-> [^self concretizeConditionalJump: PL].
		[JumpOverflow]				-> [^self concretizeConditionalJump: VS].
		[JumpNoOverflow]			-> [^self concretizeConditionalJump: VC].
		[JumpCarry]				-> [^self concretizeConditionalJump: CS].
		[JumpNoCarry]				-> [^self concretizeConditionalJump: CC].
		[JumpLess]					-> [^self concretizeConditionalJump: LT].
		[JumpGreaterOrEqual]		-> [^self concretizeConditionalJump: GE].
		[JumpGreater]				-> [^self concretizeConditionalJump: GT].
		[JumpLessOrEqual]			-> [^self concretizeConditionalJump: LE].
		[JumpBelow]				-> [^self concretizeConditionalJump: CC]. "unsigned lower"
		[JumpAboveOrEqual]		-> [^self concretizeConditionalJump: CS]. "unsigned greater or equal"
		[JumpAbove]				-> [^self concretizeConditionalJump: HI].
		[JumpBelowOrEqual]		-> [^self concretizeConditionalJump: LS].
		[JumpFPEqual]				-> [^self concretizeFPConditionalJump: EQ].
		[JumpFPNotEqual]			-> [^self concretizeFPConditionalJump: NE].
		[JumpFPLess]				-> [^self concretizeFPConditionalJump: LT].
		[JumpFPGreaterOrEqual]	-> [^self concretizeFPConditionalJump: GE].
		[JumpFPGreater]			-> [^self concretizeFPConditionalJump: GT].
		[JumpFPLessOrEqual]		-> [^self concretizeFPConditionalJump: LE].
		[JumpFPOrdered]			-> [^self concretizeFPConditionalJump: VC].
		[JumpFPUnordered]			-> [^self concretizeFPConditionalJump: VS].
		[RetN]						-> [^self concretizeRetN].
		[Stop]						-> [^self concretizeStop].
		"Arithmetic"
		[AddCqR]					-> [^self concretizeNegateableDataOperationCqR: AddOpcode R: (operands at: 1)].
		[AddCqRR]					-> [^self concretizeNegateableDataOperationCqR: AddOpcode R: (operands at: 2)].
		[AndCqR]					-> [^self concretizeInvertibleDataOperationCqR: AndOpcode].
		[AndCqRR]					-> [^self concretizeAndCqRR].
		[CmpCqR]					-> [^self concretizeNegateableDataOperationCqR: CmpOpcode R: 0].
		[OrCqR]					-> [^self concretizeDataOperationCqR: OrOpcode].
		[SubCqR]					-> [^self concretizeSubCqR].
		[TstCqR]					-> [^self concretizeTstCqR].
		[XorCqR]					-> [^self concretizeInvertibleDataOperationCqR: XorOpcode].
		[AddCwR]					-> [^self concretizeDataOperationCwR: AddOpcode].
		[AndCwR]					-> [^self concretizeDataOperationCwR: AndOpcode].
		[CmpCwR]					-> [^self concretizeDataOperationCwR: CmpOpcode].
		[OrCwR]					-> [^self concretizeDataOperationCwR: OrOpcode].
		[SubCwR]					-> [^self concretizeDataOperationCwR: SubOpcode].
		[XorCwR]					-> [^self concretizeDataOperationCwR: XorOpcode].
		[AddRR]					-> [^self concretizeDataOperationRR: AddOpcode].
		[AndRR]					-> [^self concretizeDataOperationRR: AndOpcode].
		[CmpRR]					-> [^self concretizeDataOperationRR: CmpOpcode].
		[OrRR]						-> [^self concretizeDataOperationRR: OrOpcode].
		[SubRR]					-> [^self concretizeDataOperationRR: SubOpcode].
		[XorRR]						-> [^self concretizeDataOperationRR: XorOpcode].
		[AddRRR]					-> [^self concretizeDataOperationRRR: AddOpcode].
		[SubRRR]					-> [^self concretizeDataOperationRRR: SubOpcode].
		[AddRdRd]					-> [^self concretizeAddRdRd].
		[CmpRdRd]					-> [^self concretizeCmpRdRd].
		[DivRdRd]					-> [^self concretizeDivRdRd].
		[MulRdRd]					-> [^self concretizeMulRdRd].
		[SubRdRd]					-> [^self concretizeSubRdRd].
		[SqrtRd]					-> [^self concretizeSqrtRd].
		[NegateR]						-> [^self concretizeNegateR].
		[LoadEffectiveAddressMwrR]	-> [^self concretizeLoadEffectiveAddressMwrR].
		[ArithmeticShiftRightCqR]		-> [^self concretizeArithmeticShiftRightCqR].
		[LogicalShiftRightCqR]			-> [^self concretizeLogicalShiftRightCqR].
		[LogicalShiftLeftCqR]			-> [^self concretizeLogicalShiftLeftCqR].
		[ArithmeticShiftRightCqRR]		-> [^self concretizeArithmeticShiftRightCqRR].
		[LogicalShiftRightCqRR]		-> [^self concretizeLogicalShiftRightCqRR].
		[LogicalShiftLeftCqRR]			-> [^self concretizeLogicalShiftLeftCqRR].
		[ArithmeticShiftRightRR]		-> [^self concretizeArithmeticShiftRightRR].
		[LogicalShiftLeftRR]				-> [^self concretizeLogicalShiftLeftRR].
		[LogicalShiftRightRR]			-> [^self concretizeLogicalShiftRightRR].
		[ClzRR]						-> [^self concretizeClzRR].
		"ARM Specific Arithmetic" 
		[SMULL]			-> [^self concretizeSMULL]	.
		[CMPSMULL]		-> [^self concretizeCMPSMULL].
		[MSR]				-> [^self concretizeMSR].
		"ARM Specific Data Movement"
		[PopLDM]			-> [^self concretizePushOrPopMultipleRegisters: false].
		[PushSTM]			-> [^self concretizePushOrPopMultipleRegisters: true].
		"Data Movement"
		[MoveCqR]			-> [^self concretizeMoveCqR].
		[MoveCwR]			-> [^self concretizeMoveCwR].
		[MoveRR]			-> [^self concretizeMoveRR].
		[MoveAwR]			-> [^self concretizeMoveAwR].
		[MoveRAw]			-> [^self concretizeMoveRAw].
		[MoveAbR] 			 -> [^self concretizeMoveAbR].
 		[MoveRAb]			-> [^self concretizeMoveRAb].
		[MoveMbrR]			-> [^self concretizeMoveMbrR].
		[MoveRMbr]			-> [^self concretizeMoveRMbr].
		[MoveRM16r]		-> [^self concretizeMoveRM16r].
		[MoveM16rR]		-> [^self concretizeMoveM16rR].
		[MoveM64rRd]		-> [^self concretizeMoveM64rRd].
		[MoveMwrR]		-> [^self concretizeMoveMwrR].
		[MoveXbrRR]		-> [^self concretizeMoveXbrRR].
		[MoveRXbrR]		-> [^self concretizeMoveRXbrR].
		[MoveXwrRR]		-> [^self concretizeMoveXwrRR].
		[MoveRXwrR]		-> [^self concretizeMoveRXwrR].
		[MoveRMwr]		-> [^self concretizeMoveRMwr].
		[MoveRdM64r]		-> [^self concretizeMoveRdM64r].
		[PopR]				-> [^self concretizePopR].
		[PushR]				-> [^self concretizePushR].
		[PushCq]			-> [^self concretizePushCq].
		[PushCw]			-> [^self concretizePushCw].
		[PrefetchAw]		-> [^self concretizePrefetchAw].
		"Conversion"
		[ConvertRRd]		-> [^self concretizeConvertRRd]}.

	^0 "keep Slang happy"! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 5/4/2015 10:29'!
extractOffsetFromBL: instr
	"we are told this is a BL <offset> instruction, so work out the offset it encodes"
	<inline: true>
	| relativeJump |
	relativeJump := instr bitAnd: 16r00FFFFFF.
	relativeJump := (relativeJump anyMask: 1<<23)
						ifTrue: [((relativeJump bitOr: 16r3F000000) << 2) signedIntFromLong]
						ifFalse: [relativeJump << 2].
	^relativeJump! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 17:30'!
faddd: destReg with: srcReg
"FADDD or VADD instruction to add double srcReg to double destReg and stick result in double destReg
FADDD destReg, destReg, srcReg -  ARM_ARM v5 DDI 01001.pdf pp. C4-6
VADD.F64 destReg, destReg, srcReg - ARM_ARM v7 DDI10406.pdf pp. A8-536-7"
	<inline: true>
	^((2r11101110001100000000101100000000 bitOr: destReg<<16 ) bitOr: destReg<<12) bitOr: srcReg! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 17:35'!
fcmpFrom: regA to: regB
"FCMPD or VCMP instruction to compare two fpu double registers.
FCMPD regA, regB - ARM_ARM v5 DDI 01001.pdf pp. C4-10
VCMP.F64 regA, regB - ARM_ARM v7 DDI10406 -1"
	<inline: true>
	^(2r11101110101101000000101101000000 bitOr:(regA <<12)) bitOr: regB! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 17:38'!
fcmpeFrom: regA to: regB
	"FCMPED or VCMPE instruction to compare two fpu double registers.
FCMPED regA, regB - ARM_ARM v5 DDI 01001.pdf pp. C4-12
VCMPE.F64 regA,regB - ARM_ARM v7 DDI10406 pp. 570-1"
	<inline: true>
	^(2r11101110101101000000101111000000 bitOr: (regA <<12)) bitOr: regB! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 17:40'!
fdivd: dividend by: divisor
"FDIVD or VDIV instruction to divide double dividend by double divisor and stick result in double dividend
FDIVD dividend, dividend, divisor - ARM_ARM v5 DDI 01001.pdf pp. C4-32
VDIV.F64 dvidend, dividend, divisor - ARM_ARM v7 DDI10406 pp. A8-584-5"
	<inline: true>
	^((2r11101110100000000000101100000000 bitOr: dividend<<16 ) bitOr: dividend<<12) bitOr: divisor! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 17:47'!
fldd: destReg rn: srcReg plus: u imm: immediate8bitValue
"FLDD or VLDR instruction to move a value from address in an ARM srcReg +/- offset<<2 to an fpu double destReg
FLDD ARM_ARM v5 DDI 01001.pdf pp. C4-36
VLDR.64 ARM_ARM v7 DDI10406 pp. A8-622-3"
	<inline: true>
	"Note that
		offset is <<2 to make byte address 
		u =1 -> srcReg + offset<<2
		u=0 -> srgREg - offset<<2"
	^(((2r11101101000100000000101100000000 bitOr:(srcReg <<16)) bitOr: destReg<<12) bitOr: u<<23) bitOr: immediate8bitValue! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 2/4/2020 12:07'!
flushDCacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	"If there is a dual mapped code zone (the normal zone but marked with read/execute, and a
	 read/write zone codeToDataDelta bytes away) then flush the data cache for the corresponding
	 range in the read/write zone and invalidate the data cache for the read/execute zone."
	<cmacro: '(me,startAddress,endAddress) 0'>
	"Currently we don't know whether a dual zone on a 32-bit ARM would require cache
	 flushing to ensure coherence.  So for now this is a no op."! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 1/21/2020 11:42'!
flushICacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	<cmacro: '(me,startAddress,endAddress) __clear_cache((char*) startAddress, (char*) (endAddress ))'>
	"On ARM we almost certainly need to flush and wash hands. On linux we use __clear_cache (see http://community.arm.com/groups/processors/blog/2010/02/17/caches-and-self-modifying-code for a decent example) and remember that the end address is *exclusive* so we can just use the end address passed in since it is always the byte after the actual last one needing flushing"
	cogit processor flushICacheFrom: startAddress to: endAddress! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 18:06'!
fmsrFrom: regA to: regB
"FMSR or VMOV instruction to move a value from an ARM reg to an fpu double register ready for conversion
FMSR regB, regA - ARM_ARM v5 DDI 01001.pdf pp. C4-68
VMOV regB, regA - ARM_ARM v7 DDi10406 pp. A8-462-3"
	<inline: true>
	|destReg|
	"the dest reg bits are spread out a little"
	destReg := (regB >>1) <<16 bitOr:(regB bitAnd: 1) << 7.
	^(2r11101110000000000000101000010000 bitOr:(regA <<12)) bitOr: destReg! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 18:10'!
fmstat
	"FMSTAT or VMRS unconditional transfer FP status to cpsr to choose jumps etc.
FMSTAT r15, FPSCR - ARM_ARM v5 DDI 01001.pdf pp. C4-72
VMRS APSR_nzcv, FPSCR - ARM_ARM v7 DDI10406 pp. A8-652-3"
	<inline: true>
	^2r11101110111100011111101000010000! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 18:14'!
fmuld: destReg with: srcReg
"FMULD or VMUL instruction to multiply double srcReg by double destReg and stick result in double destReg
FMULD destReg, destReg, srcReg - ARM_ARM v5 DDI 01001.pdf pp. C4-73
VMUL.F64 destReg, destReg, srcReg - ARM_ARM v7 DDI10406 pp A8-658-9"
	<inline: true>
	^((2r11101110001000000000101100000000 bitOr: destReg<<16 ) bitOr: destReg<<12) bitOr: srcReg! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 17:25'!
fsitodFrom: regA to: regB
"FSITOD or VCVT instruction to move convert an integer value to an fpu double
FSITOD regB, regA - ARM_ARM v5 DDI 01001.pdf pp. C4-95
VCVTW. regB, regA - ARM_ARM v7 DDI10406.pdf pp. A8-576-8"
	<inline: true>
	|srcReg|
	"the src reg bits are spread out a little"
	srcReg := (regA >>1) bitOr:(regA bitAnd: 1) << 5.
	^(2r11101110101110000000101111000000 bitOr: srcReg ) bitOr: regB<<12! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 18:16'!
fsqrtd: destReg
"FSQRTD or VSQRT instruction to square root double dividend destReg and stick result in double destReg
ARM_ARM v5 DDI 01001.pdf pp. C4-97
VSQRT.F64 destReg, destReg - ARM_ARM v7 DDI10406 pp. A8-756-7"
	<inline: true>
	^((2r11101110101100010000101111000000 ) bitOr: destReg<<12) bitOr: destReg! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 18:21'!
fstd: fpReg rn: addrReg plus: u imm: immediate8bitValue
"FSTD or VSTR instruction to move a value to address in an ARM addrReg +/- offset<<2 from an fpu double fpReg
FSTD fpReg, addrReg, #offset - ARM_ARM v5 DDI 01001.pdf pp. C4-101
VSTR.64 fpReg, addrReg, #offset - ARM_ARM v7 DDI10406 pp. A8-780-1"
	<inline: true>
	"Note that
		offset is <<2 to make byte address 
		u =1 -> addrReg + offset<<2
		u=0 -> addrReg - offset<<2"
	^(((2r11101101000000000000101100000000 bitOr:(addrReg <<16)) bitOr: fpReg<<12) bitOr: u<<23) bitOr: immediate8bitValue! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 18:23'!
fsubd: destReg with: srcReg
"FSUBD or VSUB instruction to subtract double srcReg from double destREg and stick result in double destReg
FSUBD destReg, destReg, srcReg - ARM_ARM v5 DDI 01001.pdf pp. C4-112
VSUB.F64 destReg, destReg, srcReg - ARM_ARM v7 DDI10406 pp. A8-784-5"
	<inline: true>
	^((2r11101110001100000000101101000000 bitOr: destReg<<16 ) bitOr: destReg<<12) bitOr: srcReg! !

!CogARMCompiler methodsFor: 'abi' stamp: 'eem 4/20/2015 18:22'!
fullCallsAreRelative
	"Answer if CallFull and/or JumpFull are relative and hence need relocating on method
	 compation. If so, they are annotated with IsRelativeCall in methods and relocated in
	 relocateIfCallOrMethodReference:mcpc:delta:"
	^false! !

!CogARMCompiler methodsFor: 'abstract instructions' stamp: 'eem 5/4/2017 11:44'!
genDivR: abstractRegDivisor R: abstractRegDividend Quo: abstractRegQuotient Rem: abstractRegRemainder
"Currently no instruction level support for divide on ARM. See also #canDivQuoRem"
	| rDividend rDivisor rQuotient rRemainder divRemFunctionAddr |
	<var: #divRemFunctionAddr type: #usqInt>
	self assert: abstractRegDividend ~= abstractRegDivisor.
	self assert: abstractRegQuotient ~= abstractRegRemainder.
	rDividend := abstractRegDividend.
	rDivisor := abstractRegDivisor.
	rDividend = CArg0Reg ifFalse:
		["we need to move the value in rDividend to CArg0Reg. Best to double check if rDivisor is already using it first"
		rDivisor = CArg0Reg ifTrue: "oh dear; we also need to move rDivisor's value out of the way first.. I'll move it to CArg1Reg and if some nitwit has managed to put rDividend there they deserve the crash"
			[rDividend = CArg1Reg ifTrue:
				[self error: 'register choices in genDivR:R:Quo:Rem: made life impossible'].
			cogit MoveR: rDivisor R: CArg1Reg.
			"and update rDivisor or we get buggerd by the next clause"
			rDivisor := CArg1Reg].
		cogit MoveR: rDividend R: CArg0Reg].
	rDivisor = CArg1Reg ifFalse:
		[cogit MoveR: rDivisor R: CArg1Reg].
	divRemFunctionAddr := self aeabiDivModFunctionAddr.
	self saveAndRestoreLinkRegAround:
		[cogit CallFullRT: (self cCode: [divRemFunctionAddr asUnsignedInteger]
							   inSmalltalk: [cogit simulatedTrampolineFor: divRemFunctionAddr])
			registersToBeSavedMask: (cogit registerMaskFor: CArg2Reg and: CArg3Reg)].
	"Now we need to move the r0/1 results back to rQuotient & rRemainder"
	rQuotient := abstractRegQuotient.
	rRemainder := abstractRegRemainder.
	rQuotient = CArg0Reg ifFalse: "oh good grief, not again"
		[cogit MoveR: CArg0Reg R: rQuotient.
		 rQuotient = CArg1Reg ifTrue:
			[self error: 'register choices in genDivR:R:Quo:Rem: made life impossible'] ].
	rRemainder = CArg1Reg  ifFalse:
		[cogit MoveR: CArg1Reg R: rRemainder]
				
! !

!CogARMCompiler methodsFor: 'abi' stamp: 'eem 1/7/2020 19:30'!
genMarshallNArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3
	"Generate the code to pass up to four arguments in a C run-time call.  Hack: each argument is
	 either a negative number, which encodes a constant, or a non-negative number, that of a register.

	 Run-time calls have no more than four arguments, so chosen so that on ARM, where in its C ABI the
	 first four integer arguments are passed in registers, all arguments can be passed in registers.  We
	 defer to the back end to generate this code not so much that the back end knows whether it uses
	 the stack or registers to pass arguments (it does, but...). In fact we defer for an extremely evil reason.
	 Doing so allows the x64 (where up to 6 args are passed) to assign the register arguments in an order
	 that allows some of the argument registers to be used for specific abstract registers, specifically
	 ReceiverResultReg and ClassReg.  This is evil, evil, evil, but also it's really nice to keep using the old
	 register assignments the original author has grown accustomed to."
	<inline: true>
	numArgs = 0 ifTrue: [^self].
	"Avoid arg regs being overwritten before they are read."
	numArgs > 1 ifTrue:
		[((cogit isTrampolineArgConstant: regOrConst1) not
		   and: [regOrConst1 = CArg0Reg]) ifTrue:
			[cogit MoveR: regOrConst1 R: Extra0Reg.
			 ^self genMarshallNArgs: numArgs arg: regOrConst0 arg: Extra0Reg arg: regOrConst2 arg: regOrConst3].
		 numArgs > 2 ifTrue:
			[((cogit isTrampolineArgConstant: regOrConst2) not
			   and: [regOrConst2 = CArg0Reg or: [regOrConst2 = CArg1Reg]]) ifTrue:
				[cogit MoveR: regOrConst2 R: Extra1Reg.
				 ^self genMarshallNArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: Extra1Reg arg: regOrConst3].
			 numArgs > 3 ifTrue:
				[((cogit isTrampolineArgConstant: regOrConst3) not
				   and: [regOrConst3 = CArg0Reg or: [regOrConst3 = CArg1Reg or: [regOrConst3 = CArg2Reg]]]) ifTrue:
					[cogit MoveR: regOrConst3 R: Extra2Reg.
					 ^self genMarshallNArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: Extra2Reg]]]].
	(cogit isTrampolineArgConstant: regOrConst0)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst0) R: CArg0Reg]
		ifFalse: [regOrConst0 ~= CArg0Reg ifTrue: [cogit MoveR: regOrConst0 R: CArg0Reg]].
	numArgs = 1 ifTrue: [^self].
	(cogit isTrampolineArgConstant: regOrConst1)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst1) R: CArg1Reg]
		ifFalse: [regOrConst1 ~= CArg1Reg ifTrue: [cogit MoveR: regOrConst1 R: CArg1Reg]].
	numArgs = 2 ifTrue: [^self].
	(cogit isTrampolineArgConstant: regOrConst2)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst2) R: CArg2Reg]
		ifFalse: [regOrConst2 ~= CArg2Reg ifTrue: [cogit MoveR: regOrConst2 R: CArg2Reg]].
	numArgs = 3 ifTrue: [^self].
	(cogit isTrampolineArgConstant: regOrConst3)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst3) R: CArg3Reg]
		ifFalse: [regOrConst3 ~= CArg3Reg ifTrue: [cogit MoveR: regOrConst3 R: CArg3Reg]]! !

!CogARMCompiler methodsFor: 'abstract instructions' stamp: 'tpr 5/10/2016 11:45'!
genMulR: regSource R: regDest
	"Use SMULL to produce a 64-bit result, explicitly in RISCTempReg,regDest. - ARM_ARM v7 DDI10406 pp. A8-354-5
	 By comparing RISCTempReg with regDest ASR 31(which effectively makes it 0 or -1) we know that the result being EQ means the hi reg and the top bit of the lo reg are the same - ie no overflow. The condition code can then be forced to oVerflow by use of MSR APSR_nzcvq, #1, lsl 28"
	| first |
	<var: 'first' type: #'AbstractInstruction *'>
	first := cogit gen: SMULL operand: regSource operand: regDest. "result in RISCTempReg,regDest"
	cogit gen: CMPSMULL operand: RISCTempReg operand: regDest.
	cogit gen: MSR operand: 1.
	^first! !

!CogARMCompiler methodsFor: 'generate machine code' stamp: 'eem 6/1/2016 10:25'!
genPopRegisterMask: registersToBeSavedMask
	<inline: true>
	^registersToBeSavedMask = 0
		ifTrue: [cogit Label]
		ifFalse: [cogit gen: PopLDM operand: registersToBeSavedMask]! !

!CogARMCompiler methodsFor: 'smalltalk calling convention' stamp: 'eem 3/20/2015 11:26'!
genPushRegisterArgsForAbortMissNumArgs: numArgs
	"Ensure that the register args are pushed before the outer and
	 inner retpcs at an entry miss for arity <= self numRegArgs.  The
	 outer retpc is that of a call at a send site.  The inner is the call
	 from a method or PIC abort/miss to the trampoline."

	"Putting the receiver and args above the return address means the
	 CoInterpreter has a single machine-code frame format which saves
	 us a lot of work."

	"Iff there are register args convert
		sp		->	outerRetpc			(send site retpc)
		linkReg = innerRetpc			(PIC abort/miss retpc)
	 to
		base	->	receiver
					(arg0)
					(arg1)
		sp		->	outerRetpc			(send site retpc)
		sp		->	linkReg/innerRetpc	(PIC abort/miss retpc)"
	numArgs <= cogit numRegArgs ifTrue:
		[self assert: cogit numRegArgs <= 2.
		 cogit MoveMw: 0 r: SPReg R: TempReg. "Save return address"
		 cogit MoveR: ReceiverResultReg Mw: 0 r: SPReg.
		 numArgs > 0 ifTrue:
			[cogit PushR: Arg0Reg.
			 numArgs > 1 ifTrue:
				[cogit PushR: Arg1Reg]].
		cogit PushR: TempReg]. "push back return address"
	cogit PushR: LinkReg! !

!CogARMCompiler methodsFor: 'smalltalk calling convention' stamp: 'eem 12/8/2019 20:44'!
genPushRegisterArgsForNumArgs: numArgs scratchReg: ignored
	"Ensure that the register args are pushed before the retpc for arity <= self numRegArgs."
	"This is easy on a RISC like ARM because the return address is in the link register.  Putting
	 the receiver and args above the return address means the CoInterpreter has a single
	 machine-code frame format which saves us a lot of work
	NOTA BENE: we do NOT push the return address here, which means it must be dealt with later."
	numArgs <= cogit numRegArgs ifTrue:
		[self assert: cogit numRegArgs <= 2.
		 cogit PushR: ReceiverResultReg.
		 numArgs > 0 ifTrue:
			[cogit PushR: Arg0Reg.
			 numArgs > 1 ifTrue:
				[cogit PushR: Arg1Reg]]]! !

!CogARMCompiler methodsFor: 'generate machine code' stamp: 'eem 6/1/2016 10:25'!
genPushRegisterMask: registersToBeSavedMask
	<inline: true>
	^registersToBeSavedMask = 0
		ifTrue: [cogit Label]
		ifFalse: [cogit gen: PushSTM operand: registersToBeSavedMask]! !

!CogARMCompiler methodsFor: 'abi' stamp: 'eem 5/29/2017 10:25'!
genRemoveNArgsFromStack: n
	"This is a no-op on ARM since the ABI passes up to 4 args in registers and trampolines currently observe that limit."
	<inline: true>
	self assert: n <= 4.
	^0! !

!CogARMCompiler methodsFor: 'abi' stamp: 'eem 6/1/2016 11:14'!
genRestoreRegs: regMask
	"Restore the registers in regMask as saved by genSaveRegs:."
	<inline: true>
	^self genPopRegisterMask: regMask! !

!CogARMCompiler methodsFor: 'abi' stamp: 'eem 12/17/2015 10:30'!
genSaveRegForCCall
	"Save the general purpose registers for a call into the C run-time from a trampoline."
	"Save none, because the ARM ABI only defines callee saved registers, no caller-saved regs."
	"cogit gen: STMFD operand: 16r7F"! !

!CogARMCompiler methodsFor: 'abi' stamp: 'eem 6/1/2016 11:13'!
genSaveRegs: regMask
	"Save the registers in regMask for a call into the C run-time from a trampoline"
	<inline: true>
	^self genPushRegisterMask: regMask! !

!CogARMCompiler methodsFor: 'abstract instructions' stamp: 'lw 8/1/2012 16:49'!
genSubstituteReturnAddress: retpc
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^cogit MoveCw: retpc R: LR! !

!CogARMCompiler methodsFor: 'abstract instructions' stamp: 'eem 4/1/2020 13:19'!
genSubstituteReturnAddressR: retpcReg
	<inline: true>
	^cogit MoveR: retpcReg R: LR! !

!CogARMCompiler methodsFor: 'disassembly' stamp: 'eem 5/31/2016 08:51'!
generalPurposeRegisterMap
	<doNotGenerate>
	"Answer a Dictionary from register getter to register index."
	^Dictionary newFromPairs:
		{	#r0. R0.
			#r1. R1.
			#r2. R2.
			#r3. R3.
			#r4. R4.
			#r5. R5.
			#r6. R6.
			#r7. R7.
			#r8. R8.
			#r9. R9.
			#r10. R10.
			#r11. R11.
			#r12. R12	}! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 8/5/2021 09:04'!
has64BitPerformanceCounter
	"AArch32 has a 64-bit counter but it is typically restricted.
	 So for now do without.
	 See e.g. http://neocontra.blogspot.com/2013/05/user-mode-performance-counters-for.html
		https://stackoverflow.com/questions/3247373/how-to-measure-program-execution-time-in-arm-cortex-a8-processor/3250835#3250835"
	<inline: #always>
	^false! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
hasConditionRegister
	"Answer if the receiver supports, e.g., JumpOverflow after a regular AddRR"
	^true! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
hasDoublePrecisionFloatingPointSupport
	"might be true, but is for the forseeable future disabled"
	^true! !

!CogARMCompiler methodsFor: 'testing' stamp: 'nice 8/26/2019 15:45'!
hasLZCNTInstructions
	"Answer if the processor has a LZCNT (leading zero count) instruction"
	<inline: true>
	^true "CLZ"! !

!CogARMCompiler methodsFor: 'testing' stamp: 'lw 7/31/2012 12:07'!
hasLinkRegister
	^true "lr"! !

!CogARMCompiler methodsFor: 'testing' stamp: 'lw 7/31/2012 12:10'!
hasPCDependentInstruction
	"e.g. B, BL: Branch, Branch and Link"
	^true! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 7/3/2014 16:02'!
hasPCRegister
	"Answer if the processor has a generally addressable pc register, which ARM does."
	^true! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
hasThreeAddressArithmetic
	"Answer if the receiver supports three-address arithmetic instructions (currently only AndCqRR)"
	^true! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
hasVarBaseRegister
	"Answer if the processor has a dedicated callee-saved register to point to
	 the base of commonly-accessed variables. On ARM we use R10 for this."
	^true "r10/sl"! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 06/09/2015 09:10'!
inlineCacheTagAt: callSiteReturnAddress
	"Answer the inline cache tag for the return address of a send."
	^self subclassResponsibility! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 12/11/2019 12:21'!
instructionAddressBefore: mcpc
	"Answer the instruction address immediately preceding mcpc."
	<inline: true>
	^mcpc - 4! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'tpr 10/23/2015 14:30'!
instructionBeforeAddress: followingAddress
	"Answer the instruction immediately preceding followingAddress."
	<inline: true>
	^objectMemory longAt: (self instructionAddressBefore: followingAddress)! !

!CogARMCompiler methodsFor: 'testing' stamp: 'tpr 11/22/2015 11:12'!
instructionIsB: instr
"is this a B <offset> instruction?"
	^(self conditionIsNotNever: instr) and: [(instr bitAnd: (16rF<<24)) = (16rA<<24)]! !

!CogARMCompiler methodsFor: 'testing' stamp: 'tpr 11/22/2015 11:12'!
instructionIsBL: instr
"is this a BL <offset> instruction?"
	^(self conditionIsNotNever: instr)  and: [(instr bitAnd: (16rF<<24)) = (16rB<<24)]! !

!CogARMCompiler methodsFor: 'testing' stamp: 'tpr 11/22/2015 11:12'!
instructionIsBLX: instr
"is this a BLX <targetReg> instruction?"
	^(self conditionIsNotNever: instr)  and: [(instr bitAnd: 16r0FFFFFF0) = 16r12FFF30]! !

!CogARMCompiler methodsFor: 'testing' stamp: 'tpr 11/22/2015 11:12'!
instructionIsBX: instr
"is this a BX <targetReg> instruction?"
	^(self conditionIsNotNever: instr) and: [(instr bitAnd: 16r0FFFFFF0) = 16r12FFF10]! !

!CogARMCompiler methodsFor: 'testing' stamp: 'tpr 11/22/2015 11:12'!
instructionIsCMP: instr
	"is this a CMP instruction?"
	^(self conditionIsNotNever: instr) and: [(instr >> 21 bitAnd: 16r7F) = CmpOpcode]! !

!CogARMCompiler methodsFor: 'testing' stamp: 'tpr 11/22/2015 11:13'!
instructionIsLDR: instr
	"is this any kind of LDR instruction? c.f. memMxr:reg:base:u:b:l:imm:"
	^(self conditionIsNotNever: instr) and: [(instr >> 20 bitAnd: 16rC5) = 16r41] "ldr r1, [r2, #+/-imm] or ldr r1, [r2, r3]"! !

!CogARMCompiler methodsFor: 'testing' stamp: 'tpr 11/22/2015 11:13'!
instructionIsOR: instr
	"is this an ORR instruction?"
	^(self conditionIsNotNever: instr)  and:[(instr >> 21 bitAnd: 16r7F) = (16r10 bitOr: OrOpcode)]! !

!CogARMCompiler methodsFor: 'testing' stamp: 'tpr 11/22/2015 11:13'!
instructionIsPush: instr
	"is this a push -str r??, [sp, #-4] -  instruction?"
	^(self conditionIsNotNever: instr) and: [(instr bitAnd: 16rFFF0FFF) = 16r52D0004]! !

!CogARMCompiler methodsFor: 'disassembly' stamp: 'eem 3/18/2015 16:13'!
instructionSizeAt: pc
	"Answer the instruction size at pc.Simple on ARM ;-)"
	^4! !

!CogARMCompiler methodsFor: 'generate machine code - support' stamp: 'eem 6/15/2015 20:21'!
inverseOpcodeFor: armOpcode
	"Several of the opcodes are inverses.  Answer the inverse for an opcode if it has one.
	 See Table A3-2 in sec A3.4 Data-processing instructions of the AARM."
	^armOpcode caseOf: {
			[AddOpcode]		->	[SubOpcode].
			[AndOpcode]		->	[BicOpcode].
			[BicOpcode]		->	[AndOpcode].
			[CmpOpcode]		->	[CmpNotOpcode].
			[MoveOpcode]		->	[MoveNotOpcode].
			[MoveNotOpcode]	->	[MoveOpcode].
			[SubOpcode]		->	[AddOpcode] }
		otherwise:
			[self error: 'opcode has no inverse'.
			 -1]! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 6/9/2015 09:24'!
is12BitValue: constant ifTrue: trueAlternativeBlock	ifFalse: falseAlternativeBlock
	"For LDR and STR, there is an instruction allowing for one instruction encoding if the offset is encodable in signed 12 bit form. pass the trueBlock the value and a 1-bit flag to tell it the sign.
	The falseBlock can do whatever it needs to, typically building the constant as a full 32bit value and then ld/st with that as a register offset"
	<inline: true>
	constant abs <= 4095 "(2 raisedTo: 12)-1"
		ifTrue:
			[constant >= 0 
				ifTrue: [^trueAlternativeBlock value: 1 value: constant]
				ifFalse: [^trueAlternativeBlock value: 0 value: constant abs]]
		ifFalse: [^falseAlternativeBlock value]! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 4/21/2015 23:36'!
is8BitValue: constant ifTrue: trueAlternativeBlock	ifFalse: falseAlternativeBlock
	"For extended LDR and STR for half & double, there is an instruction allowing for one instruction encoding if the offset is encodable in 8 bit."
	<inline: true>
	constant abs <= 255 "(2 raisedTo: 8)-1"
		ifTrue:
			[constant >= 0 
				ifTrue: [trueAlternativeBlock value: 1 value: constant]
				ifFalse: [trueAlternativeBlock value: 0 value: constant abs]]
		ifFalse: falseAlternativeBlock! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 5/4/2015 15:29'!
isAddressRelativeToVarBase: varAddress
	<inline: true>
	<var: #varAddress type: #usqInt>
	"Support for addressing variables off the dedicated VarBaseReg"
	^varAddress notNil
	  and: [varAddress >= cogit varBaseAddress
	  and: [varAddress - cogit varBaseAddress < (1 << 12)]]! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
isBigEndian
	^false! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 12/3/2015 22:26'!
isCallPrecedingReturnPC: mcpc
	"Assuming mcpc is a send return pc answer if the instruction before it is a call (not a CallFull)."
	"There are two types of calls: BL and/BLX encoding"
	| call |
	call := self instructionBeforeAddress: mcpc.
	^(self instructionIsBL: call) or:[self instructionIsBLX: call]! !

!CogARMCompiler methodsFor: 'testing' stamp: 'nice 7/12/2016 13:41'!
isInImmediateJumpRange: operand
	"ARM calls and jumps span +/- 32 mb, more than enough for intra-zone calls and jumps."
	<var: #operand type: #'usqIntptr_t'>
	^operand signedIntFromLong between: -16r2000000 and: 16r1FFFFFC! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 3/18/2015 16:12'!
isJumpAt: pc
	| instr |
	instr := objectMemory long32At: pc.
	^(self instructionIsB: instr)
	  or: [self instructionIsBX: instr]! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 4/22/2015 09:17'!
isPCRelativeValueLoad: instr
	<var: 'instr' type: #'unsigned int'>
	"add xx, pc, blah or sub xx, pc, blah"
	^(instr >> 16) = 16rE28F or: [instr >> 16 = 16rE24F]! !

!CogARMCompiler methodsFor: 'accessing' stamp: 'tpr 4/17/2015 19:56'!
jumpLongByteSize
"	Branch/Call ranges.  Jump[Cond] can be generated as short as possible.  Call/Jump[Cond]Long must be generated
	in the same number of bytes irrespective of displacement since their targets may be updated, but they need only
	span 16Mb, the maximum size of the code zone.  This allows e.g. ARM to use single-word call and jump instructions
	for most calls and jumps.  CallFull/JumpFull must also be generated in the same number of bytes irrespective of
	displacement for the same reason, but they must be able to span the full (32-bit or 64-bit) address space because
	they are used to call code in the C runtime, which may be distant from the code zone"
	^4! !

!CogARMCompiler methodsFor: 'accessing' stamp: 'eem 4/16/2015 10:49'!
jumpLongConditionalByteSize
	^self jumpLongByteSize! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 09/28/2015 18:48'!
jumpLongTargetBeforeFollowingAddress: mcpc 
	"Answer the target address for the long jump immediately preceding mcpc"
	^self callTargetFromReturnAddress: mcpc! !

!CogARMCompiler methodsFor: 'disassembly' stamp: 'eem 4/28/2015 14:39'!
jumpTargetPCAt: pc
	<returnTypeC: #usqInt>
	| operand word |
	word := objectMemory long32At: pc.
	operand := word bitAnd: 16rFFFFFF.
	(operand anyMask: 16r800000) ifTrue:
		[operand := operand - 16r1000000].
	^self
		cCode: [operand * 4 + pc + 8]
		inSmalltalk: [operand * 4 + pc + 8 bitAnd: cogit addressSpaceMask]! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:54'!
ldr: destReg rn: baseReg plus: u imm: immediate12bitValue
"	LDR destReg, [baseReg, immediate12bitValue] u=0 -> subtract imm; =1 -> add imm  - ARM_ARM v7 DDI10406 pp. A8-120-1"
	^self memMxr: AL reg: destReg  base: baseReg u: u b: 0 l: 1 imm: immediate12bitValue! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:56'!
ldr: destReg rn: baseReg plusImm: immediate12bitValue
"	LDR destReg, [baseReg, +immediate12bitValue] - ARM_ARM v7 DDI10406 pp. A8-120-1"
	^self memMxr: AL reg: destReg  base: baseReg u: 1 b: 0 l: 1 imm: immediate12bitValue! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:57'!
ldr: destReg rn: baseReg rm: offsetReg
"	LDR destReg, [baseReg, + offsetReg]  - ARM_ARM v7 DDI10406 pp. A8-124-5
	The contents of offsetReg are assumed to be correctly signed"
	^self memMxr: AL reg: destReg  base: baseReg p: 1 u: 1 b: 0 w: 0 l: 1 rm: offsetReg! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:57'!
ldrb: destReg rn: baseReg plus: u imm: immediate12bitValue
"	LDRB destReg, [baseReg, 'u' immediate12bitValue] u=0 ->  - ARM_ARM v7 DDI10406 pp. A8-128-9
	Note that this is a very low level interface that does not check the sign of the immediate, nor validity. See for example #concretizeMoveMbrR"
	^self memMxr: AL reg: destReg  base: baseReg u: u b: 1 l: 1 imm: immediate12bitValue! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:58'!
ldrb: destReg rn: baseReg rm: offsetReg
"	LDRB destReg, [baseReg, + offsetReg]  - ARM_ARM v7 DDI10406 pp. A8-132-3
	The contents of offsetReg are assumed to be correctly signed"
	^self memMxr: AL reg: destReg  base: baseReg p: 1 u: 1 b: 1 w: 0 l: 1 rm: offsetReg! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:58'!
ldrh: destReg rn: baseReg plus: u imm: immediate8bitValue
"	LDRH destReg, [baseReg, 'u' immediate8bitValue] u=0 -> subtract imm; =1 -> add imm  - ARM_ARM v7 DDI10406 pp. A8-152-3"
	^self memM16xr: AL reg: destReg  base: baseReg p: 1 u: u  w: 0 l: 1 offset: immediate8bitValue! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:59'!
ldrh: destReg rn: baseReg rm: offsetReg
"	LDRH destReg, [baseReg, +offsetReg]  - ARM_ARM v7 DDI10406 pp. A8-156-7
	The contents of offsetReg are assumed to be correctly signed"
	^self memM16xr: AL reg: destReg  base: baseReg p: 1 u: 1  w: 0 l: 1 rm: offsetReg! !

!CogARMCompiler methodsFor: 'abi' stamp: 'eem 6/1/2020 17:19'!
leafCallStackPointerDelta
	"Answer the delta from the stack pointer after a call to the stack pointer
	 immediately prior to the call.  This is used to compute the stack pointer
	 immediately prior to a call from within a leaf routine, which in turn is used
	 to capture the c stack pointer to use in trampolines back into the C run-time."
	"This might actually be false, since directly after a call, lr, fp and variable registers need be pushed onto the stack. It depends on the implementation of call."
	^0! !

!CogARMCompiler methodsFor: 'accessing' stamp: 'eem 6/8/2015 12:21'!
literalLoadInstructionBytes
	"Answer the size of a literal load instruction (which may or may not include the size of the literal).
	 This differs between in-line and out-of-line literal generation."
	<inline: true>
	^self subclassResponsibility! !

!CogARMCompiler methodsFor: 'generate machine code - support' stamp: 'tpr 6/24/2015 16:59'!
loadCwInto: destReg
	"Load the operand into the destination register, answering
	 the size of the instructions generated to do so."
	| operand distance |
	operand := operands at: 0.
	self cCode:[] inSmalltalk:[operand := operand bitAnd: 16rFFFFFFFF]. "Need to clamp the value to a word size since one or two usages actually generate double sized values and rely upon the C code to narrow it within the running VM"
	(self isAnInstruction: (cogit cCoerceSimple: operand to: #'AbstractInstruction *')) ifTrue:
		[operand := (cogit cCoerceSimple: operand to: #'AbstractInstruction *') address].
	"First try and encode as a pc-relative reference..."
	(cogit addressIsInCurrentCompilation: operand) ifTrue:
		[distance := operand - (address + 8).
		 self rotateable8bitSignedImmediate: distance
		 	ifTrue:
				[:rot :immediate :negate|
		 		 self machineCodeAt: 0 put: (negate
												ifTrue: [self sub: destReg rn: PC imm: immediate ror: rot]
												ifFalse: [self add: destReg rn: PC imm: immediate ror: rot]).
		 		^4]
		 	ifFalse:
		 		[self deny: (self isAnInstruction: (cogit cCoerceSimple: (operands at: 0) to: #'AbstractInstruction *'))]].
	"If this fails, use the conventional literal load sequence."
	^self moveCw: operand intoR: destReg! !

!CogARMCompiler methodsFor: 'accessing' stamp: 'eem 4/21/2015 17:30'!
loadPICLiteralByteSize
	"Answer the byte size of a MoveCwR opcode's corresponding machine code
	 when the argument is a PIC.  This is for the self-reference at the end of a
	 closed PIC.  On ARM this is a single instruction pc-relative register load."
	^4! !

!CogARMCompiler methodsFor: 'accessing' stamp: 'eem 4/20/2015 19:26'!
machineCodeAt: anOffset
	"read aWord from machineCode, with little endian"
	<inline: true>
	^machineCode at: anOffset // 4! !

!CogARMCompiler methodsFor: 'accessing' stamp: 'eem 4/20/2015 19:28'!
machineCodeAt: anOffset put: aWord
	"add aWord to machineCode, with little endian"
	<inline: true>
	machineCode at: anOffset // 4 put: aWord! !

!CogARMCompiler methodsFor: 'generate machine code' stamp: 'eem 11/24/2019 08:34'!
machineCodeBytes
	"Answer the maximum number of bytes of machine code generated for any abstract instruction."
	^self subclassResponsibility! !

!CogARMCompiler methodsFor: 'generate machine code' stamp: 'eem 11/24/2019 08:33'!
machineCodeWords
	^self machineCodeBytes / 4! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 3/12/2015 16:16'!
memM16xr: cond reg: destReg base: baseReg p: postpreoffset u: updown w: weirdstuff l: loadstore offset: offset8 
	"build an ARM [base +/- offset8]  half-word memory instruction
	p -> pre-index (1) or post-index (0) the offset. Combines with W to do some odd things.
	u -> up (1) or down (0) ie + or - for the offset
	b -> byte(1) or word (0)
	w -> write-back (1) if pre-indexing. 
	l -> load (1) or store (0)"
	^ cond << 28
		bitOr: (0 << 25
		bitOr: (postpreoffset << 24
		bitOr: (updown << 23
		bitOr: (1 << 22
		bitOr: (weirdstuff << 21
		bitOr: (loadstore << 20
		bitOr: (baseReg << 16
		bitOr: (destReg << 12 
		bitOr: ((offset8 bitAnd: 16rF0) << 4
		bitOr: (11 << 4
		bitOr: (offset8 bitAnd: 16rF)))))))))))! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 3/12/2015 17:12'!
memM16xr: cond reg: destReg base: baseReg p: postpreoffset u: updown w: weirdstuff l: loadstore rm: offsetReg 
	"build an ARM [base +/- offsetReg] memory instruction
	p -> pre-index (1) or post-index (0) the offset. Combines with W to do some odd things.
	u -> up (1) or down (0) ie + or - for the offset
	b -> byte(1) or word (0)
	w -> write-back (1) if pre-indexing. 
	l -> load (1) or store (0)"
	^ cond << 28
		bitOr: (0 << 25
		bitOr: (postpreoffset << 24
		bitOr: (updown << 23
		bitOr: (0 << 22
		bitOr: (weirdstuff << 21
		bitOr: (loadstore << 20
		bitOr: (baseReg << 16
		bitOr: (destReg << 12 
		bitOr: (16rB0
		bitOr: offsetReg)))))))))! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 4/30/2014 20:15'!
memMxr: cond reg: destReg base: baseReg p: postpreoffset u: updown b: byteword w: weirdstuff l: loadstore imm: offset
	"build an ARM [base +/- offset] memory instruction
	p -> pre-index (1) or post-index (0) the offset. Combines with W to do some odd things.
	u -> up (1) or down (0) ie + or - for the offset
	b -> byte(1) or word (0)
	w -> write-back (1) if pre-indexing. 
	l -> load (1) or store (0)"
	^ cond << 28
		bitOr: (2 << 25
		bitOr: (postpreoffset << 24
		bitOr: (updown << 23
		bitOr: (byteword << 22
		bitOr: (weirdstuff << 21
		bitOr: (loadstore << 20
		bitOr: (baseReg << 16
		bitOr: (destReg << 12 bitOr: offset))))))))! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 4/1/2015 16:02'!
memMxr: cond reg: destReg base: baseReg p: postpreoffset u: updown b: byteword w: weirdstuff l: loadstore rm: offsetReg 
	"build an ARM [base +/- offsetReg] memory instruction
	p -> pre-index (1) or post-index (0) the offset. Combines with W to do some odd things.
	u -> up (1) or down (0) ie + or - for the offset
	b -> byte(1) or word (0)
	w -> write-back (1) if pre-indexing. 
	l -> load (1) or store (0)"
	^ (cond bitAnd: 16rF)  << 28
		bitOr: (3 << 25
		bitOr: ((postpreoffset  bitAnd: 1) << 24
		bitOr: ((updown bitAnd: 1) << 23
		bitOr: ((byteword bitAnd: 1) << 22
		bitOr: ((weirdstuff bitAnd: 1) << 21
		bitOr: ((loadstore bitAnd: 1) << 20
		bitOr: ((baseReg bitAnd: 16rF) << 16
		bitOr: ((destReg bitAnd: 16rF) << 12 
		bitOr: (offsetReg bitAnd: 16rF)))))))))! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 4/1/2015 16:04'!
memMxr: cond reg: destReg base: baseReg p: postpreoffset u: updown b: byteword w: weirdstuff l: loadstore rmLsl2: offsetReg 
	"build an ARM [base +/- offsetReg lsl #2] memory instruction - see also #memMxr:reg:base:p:u:b:w:l:rm: and keep them correlated properly
	p -> pre-index (1) or post-index (0) the offset. Combines with W to do some odd things.
	u -> up (1) or down (0) ie + or - for the offset
	b -> byte(1) or word (0)
	w -> write-back (1) if pre-indexing. 
	l -> load (1) or store (0)"
	^ (cond bitAnd: 16rF)  << 28
		bitOr: (3 << 25
		bitOr: ((postpreoffset  bitAnd: 1) << 24
		bitOr: ((updown bitAnd: 1) << 23
		bitOr: ((byteword bitAnd: 1) << 22
		bitOr: ((weirdstuff bitAnd: 1) << 21
		bitOr: ((loadstore bitAnd: 1) << 20
		bitOr: ((baseReg bitAnd: 16rF) << 16
		bitOr: ((destReg bitAnd: 16rF) << 12
		bitOr: (16r100
		bitOr: (offsetReg bitAnd: 16rF))))))))))! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 4/1/2015 15:10'!
memMxr: cond reg: destReg  base: baseReg u: updown b: byteword l: loadstore imm: immediate12bitValue
"This is the lowest level build of an ARM [base +/- immediate 12bit offset] memory instruction
u -> up (1) or down (0) ie + or - for the offset
b -> byte(1) or word (0)
l -> load (1) or store (0)"

	^ (cond bitAnd: 16rF) << 28
		bitOr: (5<<24
		bitOr: ((updown bitAnd: 1) << 23
		bitOr:((byteword bitAnd: 1) <<22
		bitOr:((loadstore bitAnd: 1) <<20
		bitOr:((baseReg bitAnd: 16rF) <<16
		bitOr:((destReg bitAnd: 16rF) <<12
		bitOr: (immediate12bitValue bitAnd: 16rFFF)))))))! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:59'!
mov: destReg imm: immediate8bitValue ror: rot
	"Remember the ROR is doubled by the cpu so use 30>>1 etc.
	MOV destReg, #immediate8BitValue ROR rot - ARM_ARM v7 DDI10406 pp. A8-194-5"
	^self type: 1 op: MoveOpcode set: 0 rn: 0 rd: destReg shifterOperand: ((rot>>1) <<8 bitOr: immediate8bitValue)! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 19:59'!
mov: destReg rn: srcReg
"	MOV destReg, srcReg - ARM_ARM v7 DDI10406 pp. A8-196-7"

	^self type: 0 op: MoveOpcode set: 0 rn: 0 rd: destReg shifterOperand: srcReg! !

!CogARMCompiler methodsFor: 'generate machine code - support' stamp: 'eem 6/8/2015 14:53'!
moveCw: constant intoR: destReg
	"Emit a load of aWord into destReg.  Answer the number of bytes of machine code generated."
	 <var: 'constant' type: #usqInt>
	<inline: true>
	^self subclassResponsibility! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 20:00'!
movs: destReg rn: srcReg
"	MOVS destReg, srcReg - ARM_ARM v7 DDI10406 pp. A8-196-7"

	^self type: 0 op: MoveOpcode set: 1 rn: 0 rd: destReg shifterOperand: srcReg! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 4/23/2015 17:04'!
msr: flags
"Generate an MSR CPSR_f, #flags instruction.
Note that 
a) CPSR_f is equivalent to APSR_nzcvq (ARM ARM DDI0406A p A8-209 & A2-14)
b) We only have business with the NZCV flags so the generated instruction shifts the flags value <<28 - which is a ROR 4"

	^16r1328F000
	+ (2 "rotate rights are in units of 2, remember" << 8)
	+ (flags bitAnd: 16rF) " to make sure we don't have silly values here"! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 20:00'!
mvn: destReg imm: immediate8bitValue ror: rot
	"Remember the ROR is doubled by the cpu so use 30>>1 etc.
	MVN destReg, #immediate8BitValue ROR rot - ARM_ARM v7 DDI10406 pp. A8-214-5"
	^self type: 1 op: MoveNotOpcode set: 0 rn: 0 rd: destReg shifterOperand: ((rot>>1) <<8 bitOr: immediate8bitValue)! !

!CogARMCompiler methodsFor: 'printing' stamp: 'eem 12/26/2015 03:26'!
nameForFPRegister: reg "<Integer>"
	<doNotGenerate>
	(reg between: 0 and: 7) ifTrue:
		[^#(D0 D1 D2 D3 D4 D5 D6 D7) at: reg + 1].
	^super nameForFPRegister: reg! !

!CogARMCompiler methodsFor: 'printing' stamp: 'eem 12/8/2019 22:43'!
nameForRegister: reg "<Integer>"
	<doNotGenerate>
	| default |
	default := super nameForRegister: reg.
	^default last = $?
		ifTrue:
			[#(LR SP PC CArg0Reg CArg1Reg CArg2Reg CArg3Reg)
				detect: [:sym| (thisContext method methodClass classPool at: sym) = reg] 
				ifNone: [default]]
		ifFalse:
			[default]! !

!CogARMCompiler methodsFor: 'accessing' stamp: 'eem 10/18/2015 19:23:24'!
numIntRegArgs
	^4! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 20:00'!
orr: destReg imm: immediate8bitValue ror: rot
	"Remember the ROR is doubled by the cpu so use 30>>1 etc.
	ORR destReg, #immediate8BitValue ROR rot - ARM_ARM v7 DDI10406 pp. A8-228-9"
	^self type: 1 op: OrOpcode set: 0 rn: destReg rd: destReg shifterOperand: ((rot>>1) <<8 bitOr: immediate8bitValue)! !

!CogARMCompiler methodsFor: 'generate machine code' stamp: 'eem 4/21/2015 08:55'!
outputMachineCodeAt: targetAddress
	"Override to move machine code a word at a time."
	<inline: true>
	0 to: machineCodeSize - 1 by: 4 do:
		[:j|
		objectMemory longAt: targetAddress + j put: (machineCode at: j // 4)]! !

!CogARMCompiler methodsFor: 'generate machine code' stamp: 'eem 1/27/2020 08:53'!
padIfPossibleWithStopsFrom: startAddr to: endAddr
	| nullBytes |
	nullBytes := (endAddr - startAddr + 1) \\ 4.
	self stopsFrom: startAddr to: endAddr - nullBytes.
	endAddr - nullBytes + 1 to: endAddr 
		do: [ :p | cogit codeByteAt: p put: 16rFF]! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 5/10/2016 11:47'!
pld: baseReg plus: u offset: immediate
	"hint to memory that we will want to read from baseReg +/- imediate sometime soon
	PLD baseReg, immediate  - ARM_ARM v7 DDI10406 pp. A8-236-7"
	<inline: true>
	^ 2r11110101010100001111000000000000 bitOr: (baseReg<<16 bitOr:(u <<23 bitOr: immediate))! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 20:02'!
popR: dstReg
"	pop word off TOS
	LDR srcReg, [sp] #4 - ARM_ARM v7 DDI10406 pp. A8-120-1"
	^self memMxr: AL reg: dstReg base: SP p: 0 u: 1 b: 0 w: 0 l: 1 imm: 4! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 8/5/2021 15:32'!
preferredRegisterPairForMovePerfCnt64RRLInto: aBlock
	"Evaluate aBlock with the preferred registers to receive the
	 low and high 32-bits of the 64-bit performance counter.
	 AArch32 has a 64-bit counter but it is typically restricted.
	 So for now do without.
	 See e.g. http://neocontra.blogspot.com/2013/05/user-mode-performance-counters-for.html
		https://stackoverflow.com/questions/3247373/how-to-measure-program-execution-time-in-arm-cortex-a8-processor/3250835#3250835"
	<inline: #always>
	aBlock value: NoReg value: NoReg! !

!CogARMCompiler methodsFor: 'accessing' stamp: 'eem 3/18/2015 15:35'!
pushLinkRegisterByteSize
	^4! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/9/2016 20:03'!
pushR: srcReg
"	push word to TOS 
	STR srcReg, [sp, #-4]!! - ARM_ARM v7 DDI10406 pp. A8-382-3"
	^self memMxr: AL reg: srcReg base: SP p: 1 u: 0 b: 0 w: 1l: 0 imm: 4! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'tpr 10/23/2015 14:53'!
relocateCallBeforeReturnPC: retpc by: delta
	| instr distanceDiv4 |
	self assert: delta \\ 4 = 0.
	delta ~= 0 ifTrue:
		[instr := self instructionBeforeAddress: retpc.
		 self assert: ((self instructionIsB: instr) or: [self instructionIsBL: instr]).
		 distanceDiv4 := instr bitAnd: 16rFFFFFF.
		 distanceDiv4 := distanceDiv4 + (delta // 4).
		 objectMemory longAt: (self instructionAddressBefore: retpc ) put: ((instr bitAnd: 16rFF000000) bitOr: (distanceDiv4 bitAnd: 16rFFFFFF))]! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 3/10/2017 17:19'!
rewriteCPICJumpAt: addressFollowingJump target: jumpTargetAddr
	"Rewrite a jump instruction to call a different target.  This variant is used to reset the 
	jumps in the prototype CPIC to suit each use,.   
	Answer the extent of the code change which is used to compute the range of the icache to flush."
	<var: #addressFollowingJump type: #usqInt>
	<var: #jumpTargetAddr type: #usqInt>
	<inline: true>
	^self rewriteTransferAt: addressFollowingJump target: jumpTargetAddr! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'tpr 10/20/2015 12:37'!
rewriteCallAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a call instruction to call a different target.  This variant is used to link PICs
	 in ceSendMiss et al,.   
	Answer the extent of the code change which is used to compute the range of the icache to flush."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	^self rewriteTransferAt: callSiteReturnAddress target: callTargetAddress! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'tpr 10/20/2015 12:06'!
rewriteCallFullAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a callFull instruction to jump to a different target.  This variant
	 is used to rewrite cached primitive calls where we load the target address into ip
	and use the 'blx ip' instruction for the actual call.
	Answer the extent of the
	 code change which is used to compute the range of the icache to flush."
	<inline: true>
	^self
		rewriteFullTransferAt: callSiteReturnAddress
		target: callTargetAddress
		expectedInstruction: 16rE12FFF3C! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 06/09/2015 09:14'!
rewriteFullTransferAt: callSiteReturnAddress target: callTargetAddress expectedInstruction: expectedInstruction
	"Rewrite a CallFull or JumpFull instruction to transfer to a different target.
	 This variant is used to rewrite cached primitive calls.   Answer the extent
	 of the code change which is used to compute the range of the icache to flush."
	^self subclassResponsibility! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 06/09/2015 09:10'!
rewriteInlineCacheAt: callSiteReturnAddress tag: cacheTag target: callTargetAddress
	"Rewrite an inline cache to call a different target for a new tag.  This variant is used
	 to link unlinked sends in ceSend:to:numArgs: et al.  Answer the extent of the code
	 change which is used to compute the range of the icache to flush."
	
	^self subclassResponsibility! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'tpr 10/20/2015 12:05'!
rewriteJumpFullAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a full jump instruction to jump to a different target.  This variant
	 is used to rewrite cached primitive calls where we load the target address into ip
	and use the 'bx ip' instruction for the actual jump.
	Answer the extent of the
	 code change which is used to compute the range of the icache to flush."
	<inline: true>
	^self
		rewriteFullTransferAt: callSiteReturnAddress
		target: callTargetAddress
		expectedInstruction: 16rE12FFF1C! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'tpr 10/26/2015 10:12'!
rewriteJumpLongAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a jump instruction to call a different target.  This variant is used to reset the 
	jumps in the prototype CPIC to suit each use,.   
	Answer the extent of the code change which is used to compute the range of the icache to flush."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	^self rewriteTransferAt: callSiteReturnAddress target: callTargetAddress! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 10/7/2020 18:46'!
rewriteTransferAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a call/jump instruction to call a different target.  This variant is used to link PICs
	 in ceSendMiss et al, and to rewrite call/jumps in CPICs.
	Answer the extent of
	 the code change which is used to compute the range of the icache to flush."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	| callDistance instr |
	"for debug - [cogit disassembleFrom: callSiteReturnAddress - 10 to: callSiteReturnAddress - 1]."
	(cogit initialized and: [callTargetAddress < cogit minCallAddress]) ifTrue:
		[self error: 'linking callsite to invalid address'].

	callDistance := (callTargetAddress - (callSiteReturnAddress + 8 "pc offset"- 4 "return offset")) signedIntToLong.
	self assert: (self isInImmediateJumpRange: callDistance). "we don't support long call updates, yet"

	instr := self instructionBeforeAddress: callSiteReturnAddress.
	self assert: ((self instructionIsB: instr) or: [self instructionIsBL: instr]).
	
	objectMemory longAt:  (self instructionAddressBefore: callSiteReturnAddress) put: ((instr bitAnd: 16rFF000000) bitOr: (callDistance // 4 bitAnd: 16rFFFFFF)).

	self assert: (self callTargetFromReturnAddress: callSiteReturnAddress) = callTargetAddress.

	^4! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 6/10/2015 10:07'!
rotateable8bitBitwiseImmediate: constant ifTrue: trueAlternativeBlock ifFalse: falseAlternativeBlock
	<inline: true>
	"Invoke trueAlternativeBlock with shift, value and inverted if constant can be represented
	 by a possibly rotated 8-bit constant, otherwise invoke falseAlternativeBlock. For data
	 processing operands, there is the immediate shifter_operand variant,  where an 8 bit value
	 is ring shifted _right_ by i. This is only suitable for quick constants (Cq), which won't change."
	| value |
	value := constant.
	[(value bitAnd: 16rFF) = value ifTrue:
		[^trueAlternativeBlock value: 0 value: value value: constant ~= value].
	 2 to: 30 by: 2 do:
		[:i |
		(value bitAnd: ((16rFF <<i bitAnd:16rFFFFFFFF) bitOr: 16rFF>>(32-i))) = value ifTrue:
			[^trueAlternativeBlock
				value: 32 - i
				value: ((value >> i) bitOr: (value <<(32 - i) bitAnd:16rFFFFFFFF))
				value: constant ~= value]].
	 value = constant]
		whileTrue:
			[value := constant < 0
						ifTrue:[-1 - constant]
						ifFalse:[constant bitInvert32]].
	^falseAlternativeBlock value! !

!CogARMCompiler methodsFor: 'testing' stamp: 'tpr 5/14/2015 16:48'!
rotateable8bitImmediate: constant ifTrue: trueAlternativeBlock ifFalse: falseAlternativeBlock
	<inline: true>
	"For data processing operands, there is the immediate shifter_operand variant, 
	 where an 8 bit value is ring shifted _right_ by i.
	 This is only suitable for quick constants(Cq), which won't change."
	
	(constant bitAnd: 16rFF) = constant ifTrue:
		[^trueAlternativeBlock value: 0 value: constant].
	2 to: 30 by: 2 do:
		[:i |
		(constant bitAnd: ((16rFF <<i bitAnd:16rFFFFFFFF) bitOr: 16rFF>>(32-i))) = constant ifTrue:
			[^trueAlternativeBlock value: 32 - i value: ((constant >> i) bitOr: (constant <<(32 - i) bitAnd:16rFFFFFFFF))]].
	^falseAlternativeBlock value! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 6/10/2015 09:59'!
rotateable8bitSignedImmediate: constant ifTrue: trueAlternativeBlock ifFalse: falseAlternativeBlock
	<inline: true>
	"Invoke trueAlternativeBlock with shift, value and negated if constant can be represented
	 by a possibly rotated 8-bit constant, otherwise invoke falseAlternativeBlock. For data
	 processing operands, there is the immediate shifter_operand variant,  where an 8 bit value
	 is ring shifted _right_ by i. This is only suitable for quick constants (Cq), which won't change."
	| value |
	value := constant.
	[(value bitAnd: 16rFF) = value ifTrue:
		[^trueAlternativeBlock value: 0 value: value value: constant ~= value].
	 2 to: 30 by: 2 do:
		[:i |
		(value bitAnd: ((16rFF <<i bitAnd:16rFFFFFFFF) bitOr: 16rFF>>(32-i))) = value ifTrue:
			[^trueAlternativeBlock
				value: 32 - i
				value: ((value >> i) bitOr: (value <<(32 - i) bitAnd:16rFFFFFFFF))
				value: constant ~= value]].
	 value = constant and: [constant ~= 0]]
		whileTrue:
			[value := constant negated].
	^falseAlternativeBlock value! !

!CogARMCompiler methodsFor: 'method zone and entry point alignment' stamp: 'eem 1/18/2021 16:36'!
roundUpToMethodAlignment: numBytes
	"Determine the default alignment for the start of a CogMethod, which in turn
	 determines the size of the mask used to distinguish the checked and unchecked
	 entry-points, used to distinguish normal and super sends on method unlinking.
	 This is implemented here to allow processors with coarse instructions (ARM) to
	 increase the alignment if required."
	<cmacro: '(ignored,numBytes) (((numBytes) + 15) & -16)'> "extra parens to placate gdb :-("
	^numBytes + 15 bitAnd: -16! !

!CogARMCompiler methodsFor: 'abi' stamp: 'eem 3/19/2015 15:18'!
saveAndRestoreLinkRegAround: aBlock
	"If the processor's ABI includes a link register, generate instructions
	 to save and restore it around aBlock, which is assumed to generate code."
	<inline: true>
	| inst |
	inst := cogit PushR: LinkReg.
	aBlock value.
	cogit PopR: LinkReg.
	^inst! !

!CogARMCompiler methodsFor: 'abi' stamp: 'eem 12/21/2019 18:28'!
saveAndRestoreLinkRegUsingCalleeSavedRegNotLiveAtPointOfSendAround: aBlock
	"Extra1Reg is callee-saved and not live at point of send."
	<inline: #always>
	| inst |
	inst := cogit MoveR: LinkReg R: Extra1Reg.
	aBlock value.
	cogit MoveR: Extra1Reg R: LinkReg.
	^inst! !

!CogARMCompiler methodsFor: 'testing' stamp: 'nice 8/26/2019 15:13'!
setsConditionCodesFor: aConditionalJumpOpcode
	<inline: false> "to save Slang from having to be a real compiler (it can't inline switches that return)"
	"Answer if the receiver's opcode sets the condition codes correctly for the given conditional jump opcode.
	ARM has to check carefully since the V flag is not affected by non-comparison instructions"
	^opcode caseOf:
		{	[ArithmeticShiftRightCqR]	->	[self shiftSetsConditionCodesFor: aConditionalJumpOpcode].
			[ArithmeticShiftRightRR]	->	[self shiftSetsConditionCodesFor: aConditionalJumpOpcode].
			[LogicalShiftLeftCqR]		->	[self shiftSetsConditionCodesFor: aConditionalJumpOpcode].
			[LogicalShiftLeftRR]		->	[self shiftSetsConditionCodesFor: aConditionalJumpOpcode].
			[XorRR]					->	[true].
			[ClzRR]					->	[false]
		}
		otherwise: [self halt: 'unhandled opcode in setsConditionCodesFor:'. false]! !

!CogARMCompiler methodsFor: 'testing' stamp: 'eem 1/23/2020 20:00'!
shiftSetsConditionCodesFor: aConditionalJumpOpcode
	"check what flags the opcode needs setting - ARM doesn't set V when simply MOVing"
		^aConditionalJumpOpcode caseOf:
		{	[JumpNegative]	->	[true].
			[JumpZero]	->	[true].
			[JumpLess]	->	[true].
		}
		otherwise: [self halt: 'unhandled opcode in setsConditionCodesFor:'. false]! !

!CogARMCompiler methodsFor: 'accessing' stamp: 'lw 7/12/2012 21:55'!
stackPageInterruptHeadroomBytes
	"Return a minimum amount of headroom for each stack page (in bytes).  In a
	 JIT the stack has to have room for interrupt handlers which will run on the stack.
	According to ARM architecture v5 reference manual chapter A2.6, the basic interrupt procedure does not push anything onto the stack. It uses SPSR_err and R14_err to preserve state. Afterwards, it calls an interrupt procedure. So leave some room."
	^128 "32 words"! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'eem 12/19/2019 10:45'!
stop
	"Encode a BKPT ( - ARM_ARM v7 DDI10406 pp. A8-56) instruction. We could, given a good enough creative impulse and an over-active sense of humour, add some numerically encoded witticism to this instruction in bits 8-19 & 0-3. It has no effect on the execution but can be a way to specify which breakpoint has been hit etc."
	<inline: true>
	^AL << 28 bitOr: (16r12 << 20 bitOr: (7 << 4))! !

!CogARMCompiler methodsFor: 'generate machine code - support' stamp: 'eem 1/27/2020 08:43'!
stopsFrom: startAddr to: endAddr
	self assert: endAddr - startAddr + 1 \\ 4 = 0.
	startAddr to: endAddr by: 4 do: 
		[:addr | cogit codeLongAt: addr put: self stop].! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/10/2016 11:40'!
str: destReg rn: baseReg plus: u imm: immediate12bitValue
"	STR destReg, [baseReg, 'u' immediate12bitValue] u=0 -> subtract imm; =1 -> add imm - ARM_ARM v7 DDI10406 pp. A8-382-3 "
	^self memMxr: AL reg: destReg  base: baseReg u: u b: 0 l: 0 imm: immediate12bitValue! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/10/2016 11:41'!
str: srcReg rn: baseReg plusImm: immediate12bitValue
"	STR srcReg, [baseReg, +immediate12bitValue] - ARM_ARM v7 DDI10406 pp. A8-382-3"
	^self memMxr: AL reg: srcReg  base: baseReg u: 1 b: 0 l: 0 imm: immediate12bitValue! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/10/2016 11:41'!
str: srcReg rn: baseReg rm: offsetReg
"	STR srcReg, [baseReg, + offsetReg]  - ARM_ARM v7 DDI10406 pp. A8-384-5
The contents of offsetReg are assumed to be correctly signed"
	^self memMxr: AL reg: srcReg  base: baseReg p: 1 u: 1 b: 0 w: 0 l: 0 rm: offsetReg! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/10/2016 11:42'!
strb: destReg rn: baseReg plus: u imm: immediate12bitValue
"	STRB destReg, [baseReg, 'u' immediate12bitValue] u=0 -> subtract imm; =1 -> add imm  - ARM_ARM v7 DDI10406 pp. A8-388-9"
	^self memMxr: AL reg: destReg  base: baseReg u: u b: 1 l: 0 imm: immediate12bitValue! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/10/2016 11:42'!
strb: srcReg rn: baseReg rm: offsetReg
"	STRB srcReg, [baseReg, + offsetReg]  - ARM_ARM v7 DDI10406 pp. A8-390-1
	The contents of offsetReg are assumed to be correctly signed"
	^self memMxr: AL reg: srcReg  base: baseReg p: 1 u: 1 b: 1 w: 0 l: 0 rm: offsetReg! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/10/2016 11:42'!
strh: destReg rn: baseReg plus: u imm: immediate8bitValue
"	STRH destReg, [baseReg, 'u' immediate8bitValue] u=0 -> subtract imm; =1 -> add imm  - ARM_ARM v7 DDI10406 pp. A8-408-9"
	^self memM16xr: AL reg: destReg  base: baseReg p: 1 u: u  w: 0 l: 0 offset: immediate8bitValue! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/10/2016 11:42'!
strh: srcReg rn: baseReg rm: offsetReg
"	STRH srcReg, [baseReg, +offsetReg] - ARM_ARM v7 DDI10406 pp. A8-410-1"
	^self memM16xr: AL reg: srcReg base: baseReg p: 1 u: 1 w: 0 l: 0 rm: offsetReg! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/10/2016 11:43'!
sub: destReg rn: srcReg imm: immediate ror: rot
"	Remember the ROR is doubled by the cpu so use 30>>1 etc
	SUB destReg, srcReg, #immediate ROR rot - ARM_ARM v7 DDI10406 pp. A8-418-9"

	^self type: 1 op: SubOpcode set: 0 rn: srcReg rd: destReg shifterOperand: ((rot>>1) <<8 bitOr: immediate)! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/10/2016 11:43'!
subs: destReg rn: srcReg imm: immediate ror: rot
"	Remember the ROR is doubled by the cpu so use 30>>1 etc
	SUBS destReg, srcReg, #immediate ROR rot - ARM_ARM v7 DDI10406 pp. A8-418-9"

	^self type: 1 op: SubOpcode set: 1 rn: srcReg rd: destReg shifterOperand: ((rot>>1) <<8 bitOr: immediate)! !

!CogARMCompiler methodsFor: 'ARM convenience instructions' stamp: 'tpr 5/10/2016 11:43'!
tst: ignored rn: srcReg imm: immediate ror: rot
"	Remember the ROR is doubled by the cpu so use 30>>1 etc"
"also note that TST has no destReg
	TST srcReg, #immediate ROR rot - ARM_ARM v7 DDI10406 pp. A8-452-3"

	^self type: 1 op: TstOpcode set: 1 rn: srcReg rd: 0 shifterOperand: ((rot>>1) <<8 bitOr: immediate)! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 4/29/2014 16:02'!
type: type op: flagsOrOpcode set: doUpdateStatusRegister rn:  sourceRegister rd: targetRegister
	<inline: true>
	^(self cond: AL type: type op: flagsOrOpcode set: doUpdateStatusRegister) 
		bitOr: (sourceRegister << 16 bitOr: targetRegister << 12)! !

!CogARMCompiler methodsFor: 'encoding' stamp: 'tpr 5/1/2014 18:27'!
type: type op: flagsOrOpcode set: doUpdateStatusRegister rn:  sourceRegister rd: targetRegister shifterOperand: so
	<inline: true>
	^(self type: type op: flagsOrOpcode set: doUpdateStatusRegister rn: sourceRegister rd: targetRegister) bitOr: (so bitAnd: 16rFFF)! !

!CogARMCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/9/2015 23:06:42'!
zoneCallsAreRelative
	"Answer if Call and JumpLong are relative and hence need to take the caller's
	 relocation delta into account during code compaction, rather than just the
	 callee's delta."
	^true! !

CogARMCompiler subclass: #CogInLineLiteralsARMCompiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogInLineLiteralsARMCompiler commentStamp: 'eem 9/26/2020 15:02' prior: 0!
A CogInLineLiteralsARMCompiler is the concrete subclass of CogARMCompiler that stores literals in-line in machine-code. This code generator is not used.

Instance Variables!

CogInLineLiteralsARMCompiler subclass: #CogARMCompilerForTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!CogARMCompilerForTests class methodsFor: 'test support' stamp: 'tpr 5/31/2014 12:06'!
dataRegistersWithAccessorsDo: aTrinaryBlock
	"r0 ... sp. We can't use pc or RISCTempReg, because some opcodes may be encoded as multiple instructions and this, we need to be able to step."
	#(0 1 2 3 4 5 6 7 8 9 11 12 13 14) withIndexDo:
		[:reg :i|
		aTrinaryBlock
			value: reg
			value: (#(r0 r1 r2 r3 r4 r5 r6 r7 r8 r9 fp r12 sp lr) at: i)
			value: (#(r0: r1: r2: r3: r4: r5: r6: r7: r8: r9: fp: r12: sp: lr:) at: i)]! !

!CogARMCompilerForTests class methodsFor: 'test support' stamp: 'eem 12/19/2019 00:00'!
dataRegistersWithAccessorsGiven: aProcessorSimulator do: aTrinaryBlock
	"r0 ... sp. We can't use pc or RISCTempReg, because some opcodes may be encoded as multiple instructions and this, we need to be able to step."
	(0 to: 14) withIndexDo:
		[:reg :i|
		(self isRISCTempRegister: reg) ifFalse:
			[aTrinaryBlock
				value: reg
				value: (aProcessorSimulator registerStateGetters at: i)
				value: (aProcessorSimulator registerStateSetters at: i)]]! !

!CogARMCompilerForTests class methodsFor: 'test support' stamp: 'lw 7/31/2012 14:43'!
fp64RegistersWithAccessorsDo: aBlock
	"there are none, because we do not support VFP, yet"! !

!CogARMCompilerForTests class methodsFor: 'test support' stamp: 'eem 12/1/2015 15:38'!
registers
	^0 to: 15 "a.k.a. { R0. R1. R2. R3. R4. R5. R6. R7. R8. R9. R10. R11. R12. SP. LR. PC }"! !

!CogARMCompilerForTests class methodsFor: 'test support' stamp: 'lw 7/25/2012 16:38'!
registersWithNamesDo: aBinaryBlock
	self registers
		with: #('r0' 'r1' 'r2' 'r3' 'r4' 'r5' 'r6' 'r7' 'r8' 'r9' 'sl' 'fp' 'ip' 'sp' 'lr' 'pc')
		do: aBinaryBlock! !

!CogARMCompilerForTests methodsFor: 'generate machine code' stamp: 'nice 12/30/2020 00:06'!
concretizeAt: actualAddress
	"Override to check maxSize and machineCodeSize"

	| maxAddress nextAddress |
	maxSize ifNil: [maxSize := self computeMaximumSize].
	maxAddress := actualAddress + maxSize.
	nextAddress := super concretizeAt: actualAddress.
	self assert: (maxSize notNil
				and: [self isPCDependent
						ifTrue: [maxAddress >= nextAddress]
						ifFalse: [maxAddress = nextAddress]]).
	^nextAddress! !

!CogInLineLiteralsARMCompiler methodsFor: 'accessing' stamp: 'eem 7/8/2021 14:57'!
callFullInstructionByteSize
	^12! !

!CogInLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 09/28/2015 18:48'!
callFullTargetFromReturnAddress: callSiteReturnAddress
	"Answer the address that the call immediately preceding callSiteReturnAddress will jump to."
	"this is also used by #jumpLongTargetBeforeFollowingAddress: and so we check for both call and jump related instructions; later on we can use simpler tests once it feels safe to assume we get here always with a call/jump in the proper place"
	| call |
	call := self instructionBeforeAddress: callSiteReturnAddress.
	self assert: ((self instructionIsBX: call) or: [self instructionIsBLX: call]).
	"A Long Call/Jump. Extract the value saved to RISCTempReg from all the instructions before."
	^self extract32BitOperandFrom4Instructionspreceding: callSiteReturnAddress - 4! !

!CogInLineLiteralsARMCompiler methodsFor: 'accessing' stamp: 'eem 11/20/2015 16:48'!
cmpC32RTempByteSize
	^20! !

!CogInLineLiteralsARMCompiler methodsFor: 'initialization' stamp: 'eem 10/6/2015 16:25'!
cogit: aCogit
	<doNotGenerate>
	"There is no support for synthesizing 64-bit constants using inline instructions in this code generator."
	self assert: aCogit objectMemory wordSize = 4.
	super cogit: aCogit! !

!CogInLineLiteralsARMCompiler methodsFor: 'testing' stamp: 'eem 9/28/2015 19:00'!
extract32BitOperandFrom4InstructionsPreceding: addr
	<inline: true>
	^(objectMemory byteAt: addr -4) 
	 + ((objectMemory byteAt: addr - 8) << 8) 
	 + ((objectMemory byteAt: addr - 12) << 16) 
	 + ((objectMemory byteAt: addr - 16) << 24)! !

!CogInLineLiteralsARMCompiler methodsFor: 'accessing' stamp: 'eem 6/10/2015 14:03'!
getDefaultCogCodeSize
	"Return the default number of bytes to allocate for native code at startup.
	 The actual value can be set via vmParameterAt: and/or a preference in the ini file."
	<inline: true>
	^1024 * 1536! !

!CogInLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 9/28/2015 19:01'!
inlineCacheTagAt: callSiteReturnAddress
	"Answer the inline cache tag for the return address of a send."
	self assert: (self instructionIsBL: (self instructionBeforeAddress: callSiteReturnAddress)).
	^self extract32BitOperandFrom4InstructionsPreceding: callSiteReturnAddress - 4! !

!CogInLineLiteralsARMCompiler methodsFor: 'testing' stamp: 'eem 9/28/2015 19:02'!
insert32BitOperand: operand into4InstructionsPreceding: addr
	<inline: true>
	objectMemory
		byteAt: addr -   4 put: (operand			bitAnd: 16rFF);
		byteAt: addr -   8 put: (operand >>   8	bitAnd: 16rFF);
		byteAt: addr - 12 put: (operand >> 16	bitAnd: 16rFF);
		byteAt: addr - 16 put: (operand >> 24	bitAnd: 16rFF)! !

!CogInLineLiteralsARMCompiler methodsFor: 'testing' stamp: 'eem 6/9/2015 21:04'!
isPCDependent
	"Answer if the receiver is a pc-dependent instruction."
	^self isJump or: [opcode = AlignmentNops]! !

!CogInLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 11/20/2015 16:43'!
literal32BeforeFollowingAddress: followingAddress
	"Answer the 32-bit constant loaded by a MOV/ORR/ORR/ORR
	 or MOV/ORR/ORR/ORR/PUSH, or MOV/ORR/ORR/ORR/CMP sequence, just before this address:"
	^(self instructionIsOR: (self instructionBeforeAddress: followingAddress))
		ifTrue: [self extract32BitOperandFrom4InstructionsPreceding: followingAddress]
		ifFalse: [self extract32BitOperandFrom4InstructionsPreceding: followingAddress - 4]! !

!CogInLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 9/28/2015 19:01'!
literalBeforeFollowingAddress: followingAddress
	"Answer the long constant loaded by a MOV/ORR/ORR/ORR
	 or MOV/ORR/ORR/ORR/PUSH, or MOV/ORR/ORR/ORR/CMP sequence, just before this address:"
	^(self instructionIsOR: (self instructionBeforeAddress: followingAddress))
		ifTrue: [self extract32BitOperandFrom4InstructionsPreceding: followingAddress]
		ifFalse: [self extract32BitOperandFrom4InstructionsPreceding: followingAddress - 4]! !

!CogInLineLiteralsARMCompiler methodsFor: 'accessing' stamp: 'eem 6/8/2015 12:20'!
literalLoadInstructionBytes
	"Answer the size of a literal load instruction.  With in-line literals this is 4 instructions."
	<inline: true>
	^16! !

!CogInLineLiteralsARMCompiler methodsFor: 'accessing' stamp: 'tpr 11/11/2015 11:05'!
loadLiteralByteSize
	"Answer the byte size of a MoveCwR opcode's corresponding machine code. On ARM this is a single instruction pc-relative register load - unless we have made a mistake and not turned on the out of line literals manager"
	^16! !

!CogInLineLiteralsARMCompiler methodsFor: 'generate machine code' stamp: 'lw 7/24/2012 20:00'!
machineCodeBytes
	"Answer the maximum number of bytes of machine code generated for any abstract instruction.
	 e.g. CmpCwR =>
			mov R3, #<addressByte1>, 12
			orr R3, R3, #<addressByte2>, 8
			orr R3, R3, #<addressByte3>, 4
			orr R3, R3, #<addressByte4>, 0
			cmp R?, R3"
	^20! !

!CogInLineLiteralsARMCompiler methodsFor: 'generate machine code - support' stamp: 'eem 6/8/2015 14:53'!
moveCw: constant intoR: destReg
	"Emit a load of aWord into destReg.  Because most ARM instruction enable using a (8-12bit) offset relative to a
	 register, the LS Byte can be included in that instruction, saving one instruction. This is done in a decorator,
	 e.g. concretizeDataOperationCwR:  Generates along the lines of
		MOV destReg, #<constantByte3>, 12
		ORR destReg, destReg, #<constantByte2>, 8
		ORR destReg, destReg, #<constantByte1>, 4
		ORR destReg, destReg, #<constantByte0>, 0
	 with minimal choice of the rotation (last digit).
	 The same register can be modified multiple times, because the operation is (inclusive) or."
	 <var: 'constant' type: #usqInt>
	<inline: true>
	"self assert: destReg < 12."

	self machineCodeAt: 0 put: (self mov: destReg imm: (constant >>24 bitAnd: 16rFF) ror: 8).
	self machineCodeAt: 4 put: (self orr: destReg imm: (constant >> 16 bitAnd: 16rFF) ror: 16).
	self machineCodeAt: 8 put: (self orr: destReg imm: (constant >> 8 bitAnd: 16rFF) ror: 24).
	self machineCodeAt: 12 put: (self orr: destReg imm: (constant bitAnd: 16rFF) ror: 0).
	^16! !

!CogInLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/12/2015 11:44:41'!
nsSendCacheAt: callSiteReturnAddress
	"Answer the NSSendCache for the return address of a Newspeak
	 self, super, outer, or implicit receiver send."
	self assert: (self instructionIsBL: (self instructionBeforeAddress: callSiteReturnAddress)).
	^self extract32BitOperandFrom4InstructionsPreceding: callSiteReturnAddress - 4! !

!CogInLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 9/28/2015 19:07'!
relocateMethodReferenceBeforeAddress: pc by: delta
	"If possible we generate the method address using pc-relative addressing.
	 If so we don't need to relocate it in code.  So check if pc-relative code was
	 generated, and if not, adjust a long sequence.  There are two cases, a push
	 or a register load.  If a push, then there is a register load, but in the instruction
	 before."
	| pcPrecedingLoad reference |
	pcPrecedingLoad := (self instructionIsPush: (self instructionBeforeAddress: pc))
							ifTrue: [pc - 4]
							ifFalse: [pc].
	"If the load is not done via pc-relative addressing we have to relocate."
	(self isPCRelativeValueLoad: (self instructionBeforeAddress: pcPrecedingLoad)) ifFalse:
		[reference := self extract32BitOperandFrom4InstructionsPreceding: pcPrecedingLoad.
		 reference := reference + delta.
		 self insert32BitOperand: reference into4InstructionsPreceding: pcPrecedingLoad]! !

!CogInLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 9/28/2015 19:02'!
rewriteFullTransferAt: callSiteReturnAddress target: callTargetAddress expectedInstruction: expectedInstruction
	"Rewrite a CallFull or JumpFull instruction to transfer to a different target.
	 This variant is used to rewrite cached primitive calls.   Answer the extent
	 of the code change which is used to compute the range of the icache to flush."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	"cogit disassembleFrom: callSiteReturnAddress - 20 to: callSiteReturnAddress - 1"
	self assert: (self instructionBeforeAddress: callSiteReturnAddress) = expectedInstruction.
	self insert32BitOperand: callTargetAddress into4InstructionsPreceding: callSiteReturnAddress - 4.
	self assert: (self callFullTargetFromReturnAddress: callSiteReturnAddress) signedIntToLong = callTargetAddress.
	^20! !

!CogInLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 10/7/2020 18:45'!
rewriteInlineCacheAt: callSiteReturnAddress tag: cacheTag target: callTargetAddress
	"Rewrite an inline cache to call a different target for a new tag.  This variant is used
	 to link unlinked sends in ceSend:to:numArgs: et al.  Answer the extent of the code
	 change which is used to compute the range of the icache to flush."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	| call callDistance |
	(cogit initialized and: [callTargetAddress < cogit minCallAddress]) ifTrue:
		[self error: 'linking callsite to invalid address'].
	callDistance := (callTargetAddress - (callSiteReturnAddress + 8 "pc offset"- 4 "return offset")) signedIntToLong.
	self assert: (self isInImmediateJumpRange: callDistance). "we don't support long call updates here"
	call := self bl: callDistance.
	objectMemory longAt: (self instructionAddressBefore: callSiteReturnAddress ) put: call.
	self insert32BitOperand: cacheTag into4InstructionsPreceding: (self instructionAddressBefore: callSiteReturnAddress ).
	self assert: (self callTargetFromReturnAddress: callSiteReturnAddress) = callTargetAddress.
	self assert: (self extract32BitOperandFrom4InstructionsPreceding: (self instructionAddressBefore: callSiteReturnAddress )) = cacheTag.
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: callSiteReturnAddress - 20 to: callSiteReturnAddress - 1]."
	^20! !

!CogInLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 9/28/2015 19:02'!
rewriteInlineCacheTag: cacheTag at: callSiteReturnAddress
	"Rewrite an inline cache with a new tag.  This variant is used
	 by the garbage collector."
	self insert32BitOperand: cacheTag into4InstructionsPreceding: callSiteReturnAddress -4! !

!CogInLineLiteralsARMCompiler methodsFor: 'generate machine code' stamp: 'eem 09/28/2015 18:48'!
sizePCDependentInstructionAt: eventualAbsoluteAddress
	"Size a jump and set its address.  The target may be another instruction
	 or an absolute address.  On entry the address inst var holds our virtual
	 address. On exit address is set to eventualAbsoluteAddress, which is
	 where this instruction will be output.  The span of a jump to a following
	 instruction is therefore between that instruction's address and this
	 instruction's address ((which are both still their virtual addresses), but the
	 span of a jump to a preceding instruction or to an absolute address is
	 between that instruction's address (which by now is its eventual absolute
	 address) or absolute address and eventualAbsoluteAddress.

	 ARM is simple; the 26-bit call/jump range means no short jumps.  This routine
	 only has to determine the targets of jumps, not determine sizes."

	opcode = AlignmentNops ifTrue:
		[| alignment |
		 address := eventualAbsoluteAddress.
		 alignment := operands at: 0.
		 ^machineCodeSize := (eventualAbsoluteAddress + (alignment - 1) bitAnd: alignment negated)
							   - eventualAbsoluteAddress].
	self assert: (self isJump or: [opcode = Call or: [opcode = CallFull]]).
	self isJump ifTrue: [self resolveJumpTarget].
	address := eventualAbsoluteAddress.
	^machineCodeSize := maxSize! !

!CogInLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 7/5/2021 17:14'!
storeLiteral: literal beforeFollowingAddress: followingAddress
	"Rewrite the long constant loaded by a MOV/ORR/ORR/ORR
	 or MOV/ORR/ORR/ORR/PUSH  sequence, just before this address:"
	(self instructionIsOR: (self instructionBeforeAddress: followingAddress))
		ifTrue: [self insert32BitOperand: literal into4InstructionsPreceding: followingAddress]
		ifFalse: [self insert32BitOperand: literal into4InstructionsPreceding: followingAddress - 4]! !

CogARMCompiler subclass: #CogOutOfLineLiteralsARMCompiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogOutOfLineLiteralsARMCompiler commentStamp: 'eem 9/26/2020 15:03' prior: 0!
CogOutOfLineLiteralsARMCompiler is the concrete subclass of CogARMCompiler that stores literals out-of-line, accessed via pc-relative addressing. This is the production code generator for 32-bit ARM.

Instance Variables
!

!CogOutOfLineLiteralsARMCompiler class methodsFor: 'accessing class hierarchy' stamp: 'eem 6/8/2015 16:50'!
literalsManagerClass
	^OutOfLineLiteralsManager! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'accessing' stamp: 'eem 7/8/2021 14:56'!
callFullInstructionByteSize
	^8! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'accessing' stamp: 'eem 11/20/2015 16:49'!
cmpC32RTempByteSize
	^8! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'generate machine code' stamp: 'eem 12/8/2019 15:23'!
concretizeLiteral
	"Generate an out-of-line literal.  Copy the value and any annotation from the stand-in in the literals manager."
	| literalAsInstruction literal |
	literalAsInstruction := cogit cCoerceSimple: (operands at: 0) to: #'AbstractInstruction *'.
	literal := (self isAnInstruction: literalAsInstruction)
				ifTrue: [literalAsInstruction address]
				ifFalse: [self cCode: [literalAsInstruction asUnsignedInteger]
							inSmalltalk: [literalAsInstruction]].
	self assert: (dependent notNil and: [dependent opcode = Literal]).
	dependent annotation ifNotNil:
		[self assert: annotation isNil.
		 annotation := dependent annotation].
	dependent address ifNotNil: [self assert: dependent address = address].
	dependent address: address.
	self machineCodeAt: 0 put: literal.
	^4! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'accessing' stamp: 'eem 6/10/2015 14:02'!
getDefaultCogCodeSize
	"Return the default number of bytes to allocate for native code at startup.
	 The actual value can be set via vmParameterAt: and/or a preference in the ini file."
	<inline: true>
	^1024 * 1280! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 7/7/2015 13:34'!
inlineCacheTagAt: callSiteReturnAddress
	<inline: true>
	^objectMemory longAt: (self pcRelativeAddressAt: (callSiteReturnAddress - 8) asUnsignedInteger)! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'testing' stamp: 'eem 6/18/2015 15:46'!
isPCDependent
	"Answer if the receiver is a pc-dependent instruction.  With out-of-line literals any instruction
	 that refers to a literal depends on the address of the literal, so add them in addition to the jumps."
	^self isJump
	  or: [opcode = AlignmentNops
	  or: [opcode ~= Literal and: [dependent notNil and: [dependent opcode = Literal]]]]! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'generate machine code' stamp: 'eem 12/17/2019 12:38'!
isSharable
	"Hack:  To know if a literal should be unique (not shared) mark the LSB of the second operand."
	<inline: true>
	self assert: opcode = Literal.
	^(operands at: 1) anyMask: 1! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 11/21/2015 18:42'!
literalBeforeFollowingAddress: followingAddress
	"Return the literal referenced by the instruction immediately preceding followingAddress."
	^objectMemory longAt: (self pcRelativeAddressAt:
		((self instructionIsLDR: (self instructionBeforeAddress: followingAddress))
			ifTrue: [self instructionAddressBefore: followingAddress]
			ifFalse: [self instructionAddressBefore: followingAddress - 4]))! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'accessing' stamp: 'eem 6/8/2015 12:20'!
literalLoadInstructionBytes
	"Answer the size of a literal load instruction (which does not include the size of the literal).
	 With out-of-line literals this is always a single LDR instruction that refers to the literal."
	<inline: true>
	^4! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'generate machine code' stamp: 'eem 6/16/2015 17:31'!
literalOpcodeIndex
	"Hack:  To know how far away a literal is from its referencing instruction we store
	 its opcodeIndex, or -1, if as yet unassigned, in the second operand of the literal."
	<inline: true>
	self assert: opcode = Literal.
	^(operands at: 2) asInteger! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'accessing' stamp: 'tpr 11/11/2015 11:05'!
loadLiteralByteSize
	"Answer the byte size of a MoveCwR opcode's corresponding machine code. On ARM this is a single instruction pc-relative register load - unless we have made a mistake and not turned on the out of line literals manager"
	^4! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'generate machine code' stamp: 'eem 11/24/2019 08:41'!
machineCodeBytes
	"Answer the maximum number of bytes of machine code generated for any abstract instruction."
	^8! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'generate machine code' stamp: 'eem 12/29/2019 12:18'!
mapEntryAddress
	"Typically map entries apply to the end of an instruction, for two reasons:
	  a)	to cope with literals embedded in variable-length instructions, since, e.g.
		on x86, the literal typically comes at the end of the instruction.
	  b)	in-line cache detection is based on return addresses, which are typically
		to the instruction following a call.
	 But on ARM using CogOutOfLineLiteralsARMCompiler we use out-of-line literals
	 so the mapEntryAddress of a literal is simply the instruction's address."
	<inline: true>
	^opcode = Literal
		ifTrue: [address]
		ifFalse: [address + machineCodeSize]! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'generate machine code - support' stamp: 'eem 12/8/2019 15:40'!
moveCw: constant intoR: destReg
	"Emit a load of aWord into destReg.  Answer the number of bytes of machine code generated.
	 Literals are stored out-of-line; emit a LDR with the relevant offset."
	 <var: 'constant' type: #usqInt>
	<inline: true>
	self assert: (cogit addressIsInCurrentCompilation: dependent address).
	self assert: (dependent address - (address + 8)) abs < (1<<12).
	self machineCodeAt: 0
		put: (self
				ldr: destReg
				rn: PC
				plus: (dependent address >= (address + 8) ifTrue: [1] ifFalse: [0])
				imm: (dependent address - (address + 8)) abs).
	^4! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/12/2015 11:44:48'!
nsSendCacheAt: callSiteReturnAddress
	"Answer the NSSendCache for the return address of a Newspeak
	 self, super, outer, or implicit receiver send."
	<inline: true>
	^objectMemory longAt: (self pcRelativeAddressAt: (callSiteReturnAddress - 8) asUnsignedInteger)! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'generate machine code' stamp: 'eem 6/19/2015 10:52'!
outOfLineLiteralOpcodeLimit
	"The maximum offset in a LDR is (1<<12)-1, or (1<<10)-1 instructions.
	 Be conservative.  The issue is that one abstract instruction can emit
	 multiple hardware instructions so we assume a 2 to 1 worst case of
	 hardware instructions to abstract opcodes.."
	^1 << (12 "12-bit offset field"
			- 2 "4 bytes per literal"
			- 1 "2 hardware instructions to 1 abstract opcode") - 1! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 6/15/2015 17:48'!
pcRelativeAddressAt: instrAddress
	"Extract the address of the ldr rX, [pc, #NNN] instruction at address"
	| inst offset |
	inst := objectMemory longAt: instrAddress.
	self assert: (inst bitAnd: 16rFF5F0000) = (self ldr: 0 rn: PC plus: 0 imm: 0).
	offset := inst bitAnd: 16rFFF.
	^instrAddress + 8 + ((inst anyMask: 1 << 23)
							ifTrue: [offset]
							ifFalse: [offset negated])! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 9/28/2015 19:07'!
relocateMethodReferenceBeforeAddress: pc by: delta
	"If possible we generate the method address using pc-relative addressing.
	 If so we don't need to relocate it in code.  So check if pc-relative code was
	 generated, and if not, adjust a load literal.  There are two cases, a push
	 or a register load.  If a push, then there is a register load, but in the instruction
	 before."
	| pcPrecedingLoad reference litAddr |
	pcPrecedingLoad := (self instructionIsPush: (self instructionBeforeAddress: pc))
							ifTrue: [pc - 4]
							ifFalse: [pc].
	"If the load is not done via pc-relative addressing we have to relocate."
	(self isPCRelativeValueLoad: (self instructionBeforeAddress: pcPrecedingLoad)) ifFalse:
		[litAddr := self pcRelativeAddressAt: pcPrecedingLoad.
		 reference := objectMemory longAt: litAddr.
		 objectMemory longAt: litAddr put: reference + delta]! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'tpr 10/20/2015 12:07'!
rewriteFullTransferAt: callSiteReturnAddress target: callTargetAddress expectedInstruction: expectedInstruction
	"Rewrite a CallFull or JumpFull instruction to transfer to a different target.
	 This variant is used to rewrite cached primitive calls where we load the target address into ip
	and use the 'bx ip' or 'blx ip' instruction for the actual jump or call.
	Answer the extent
	 of the code change which is used to compute the range of the icache to flush."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	self assert: (self instructionBeforeAddress: callSiteReturnAddress) = expectedInstruction.
	objectMemory longAt: (self pcRelativeAddressAt: callSiteReturnAddress - 8) put: callTargetAddress.
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: callSiteReturnAddress - 8 to: (self pcRelativeAddressAt: callSiteReturnAddress - 8)]."
	^0! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 10/7/2020 18:13'!
rewriteInlineCacheAt: callSiteReturnAddress tag: cacheTag target: callTargetAddress
	"Rewrite an inline cache to call a different target for a new tag.  This variant is used
	 to link unlinked sends in ceSend:to:numArgs: et al.  Answer the extent of the code
	 change which is used to compute the range of the icache to flush."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	| call callDistance |
	(cogit initialized and: [callTargetAddress < cogit minCallAddress]) ifTrue:
		[self error: 'linking callsite to invalid address'].
	callDistance := (callTargetAddress - (callSiteReturnAddress + 8 "pc offset"- 4 "return offset")) signedIntToLong.
	self assert: (self isInImmediateJumpRange: callDistance). "we don't support long call updates here"
	call := self bl: callDistance.
	objectMemory
		longAt: (self instructionAddressBefore: callSiteReturnAddress ) put: call;
		longAt: (self pcRelativeAddressAt: callSiteReturnAddress - 8) put: cacheTag.
	self assert: (self inlineCacheTagAt: callSiteReturnAddress) = cacheTag.
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: callSiteReturnAddress - 8 to: (self pcRelativeAddressAt: callSiteReturnAddress - 8)]."
	^4! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 6/16/2015 12:39'!
rewriteInlineCacheTag: cacheTag at: callSiteReturnAddress
	"Rewrite an inline cache with a new tag.  This variant is used
	 by the garbage collector."
	<inline: true>
	objectMemory longAt: (self pcRelativeAddressAt: callSiteReturnAddress - 8) put: cacheTag! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'generate machine code' stamp: 'eem 6/18/2015 16:18'!
setLiteralOpcodeIndex: index
	"Hack:  To know how far away a literal is from its referencing instruction we store
	 its opcodeIndex, or -1, if as yet unassigned, in the second operand of the literal."
	<inline: true>
	self assert: opcode = Literal.
	operands at: 2 put: index! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'generate machine code - support' stamp: 'eem 12/17/2019 14:18'!
setLiteralSize: literalSize
	"Nothing to do in a 32-bit implementation; All out-of-line literals are 32 bits"
	<inline: #always>! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'generate machine code' stamp: 'eem 09/28/2015 18:48'!
sizePCDependentInstructionAt: eventualAbsoluteAddress
	"Size a jump and set its address.  The target may be another instruction
	 or an absolute address.  On entry the address inst var holds our virtual
	 address. On exit address is set to eventualAbsoluteAddress, which is
	 where this instruction will be output.  The span of a jump to a following
	 instruction is therefore between that instruction's address and this
	 instruction's address ((which are both still their virtual addresses), but the
	 span of a jump to a preceding instruction or to an absolute address is
	 between that instruction's address (which by now is its eventual absolute
	 address) or absolute address and eventualAbsoluteAddress.

	 ARM is simple; the 26-bit call/jump range means no short jumps.  This
	 routine only has to determine the targets of jumps, not determine sizes.

	 This version also deals with out-of-line literals.  If this is the real literal,
	 update the stand-in in literalsManager with the address (because instructions
	 referring to the literal are referring to the stand-in).  If this is annotated with
	 IsObjectReference transfer the annotation to the stand-in, whence it will be
	 transferred to the real literal, simplifying update of literals."

	opcode = AlignmentNops ifTrue:
		[| alignment |
		 address := eventualAbsoluteAddress.
		 alignment := operands at: 0.
		 ^machineCodeSize := (eventualAbsoluteAddress + (alignment - 1) bitAnd: alignment negated)
							   - eventualAbsoluteAddress].
	self assert: (self isJump or: [opcode = Call or: [opcode = CallFull
				or: [dependent notNil and: [dependent opcode = Literal]]]]).
	self isJump ifTrue: [self resolveJumpTarget].
	address := eventualAbsoluteAddress.
	(dependent notNil and: [dependent opcode = Literal]) ifTrue:
		[opcode = Literal ifTrue:
			[dependent address: address].
		 annotation = cogit getIsObjectReference ifTrue:
			[dependent annotation: annotation.
			 annotation := nil]].
	^machineCodeSize := maxSize! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'inline cacheing' stamp: 'eem 11/21/2015 18:43'!
storeLiteral: literal beforeFollowingAddress: followingAddress
	"Rewrite the literal in the instruction immediately preceding followingAddress."
	objectMemory
		longAt: (self pcRelativeAddressAt:
					((self instructionIsLDR: (self instructionBeforeAddress: followingAddress))
						ifTrue: [self instructionAddressBefore: followingAddress]
						ifFalse: [self instructionAddressBefore: followingAddress - 4]))
		put: literal! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'generate machine code' stamp: 'eem 6/9/2015 09:46'!
updateLabel: labelInstruction
	opcode ~= Literal ifTrue:
		[super updateLabel: labelInstruction]! !

!CogOutOfLineLiteralsARMCompiler methodsFor: 'testing' stamp: 'eem 1/23/2020 16:15'!
usesOutOfLineLiteral
	"Answer if the receiver uses an out-of-line literal.  Needs only
	 to work for the opcodes created with gen:literal:operand: et al."

	opcode
		caseOf: {
		[CallFull]		-> [^true].
		[JumpFull]		-> [^true].
		"Arithmetic"
		[AddCqR]		-> [^self rotateable8bitSignedImmediate: (operands at: 0) ifTrue: [:r :i :n| false] ifFalse: [true]].
		[AddCqRR]		-> [^self rotateable8bitSignedImmediate: (operands at: 0) ifTrue: [:r :i :n| false] ifFalse: [true]].
		[AndCqR]		-> [^self rotateable8bitBitwiseImmediate: (operands at: 0)
								ifTrue: [:r :i :n| false]
								ifFalse: [1 << (operands at: 0) highBit ~= ((operands at: 0) + 1)]].
		[AndCqRR]		-> [^self rotateable8bitBitwiseImmediate: (operands at: 0)
								ifTrue: [:r :i :n| false]
								ifFalse: [1 << (operands at: 0) highBit ~= ((operands at: 0) + 1)]].
		[OrCqRR]		-> [^self rotateable8bitBitwiseImmediate: (operands at: 0)
								ifTrue: [:r :i :n| false]
								ifFalse: [1 << (operands at: 0) highBit ~= ((operands at: 0) + 1)]].
		[CmpCqR]		-> [^self rotateable8bitSignedImmediate: (operands at: 0) ifTrue: [:r :i :n| false] ifFalse: [true]].
		[OrCqR]			-> [^self rotateable8bitImmediate: (operands at: 0) ifTrue: [:r :i| false] ifFalse: [true]].
		[SubCqR]		-> [^self rotateable8bitSignedImmediate: (operands at: 0) ifTrue: [:r :i :n| false] ifFalse: [true]].
		[TstCqR]		-> [^self rotateable8bitImmediate: (operands at: 0) ifTrue: [:r :i| false] ifFalse: [true]].
		[XorCqR]		-> [^self rotateable8bitBitwiseImmediate: (operands at: 0) ifTrue: [:r :i :n| false] ifFalse: [true]].
		[AddCwR]		-> [^true].
		[AndCwR]		-> [^true].
		[CmpCwR]		-> [^true].
		[OrCwR]		-> [^true].
		[SubCwR]		-> [^true].
		[XorCwR]		-> [^true].
		[LoadEffectiveAddressMwrR]
						-> [^self rotateable8bitImmediate: (operands at: 0) ifTrue: [:r :i| false] ifFalse: [true]].
		"Data Movement"						
		[MoveCqR]		-> [^self rotateable8bitImmediate: (operands at: 0) ifTrue: [:r :i| false] ifFalse: [true]].
		[MoveCwR]		-> [^(self inCurrentCompilation: (operands at: 0)) not].
		[MoveAwR]		-> [^(self isAddressRelativeToVarBase: (operands at: 0)) ifTrue: [false] ifFalse: [true]].
		[MoveRAw]		-> [^(self isAddressRelativeToVarBase: (operands at: 1)) ifTrue: [false] ifFalse: [true]].
		[MoveAbR]		-> [^(self isAddressRelativeToVarBase: (operands at: 0)) ifTrue: [false] ifFalse: [true]].
		[MoveRAb]		-> [^(self isAddressRelativeToVarBase: (operands at: 1)) ifTrue: [false] ifFalse: [true]].
		[MoveRMwr]	-> [^self is12BitValue: (operands at: 1) ifTrue: [:u :i| false] ifFalse: [true]].
		[MoveRdM64r]	-> [^self is12BitValue: (operands at: 1) ifTrue: [:u :i| false] ifFalse: [true]]. 
		[MoveMbrR]		-> [^self is12BitValue: (operands at: 0) ifTrue: [:u :i| false] ifFalse: [true]].
		[MoveRMbr]		-> [^self is12BitValue: (operands at: 1) ifTrue: [:u :i| false] ifFalse: [true]].
		[MoveM16rR]	-> [^self rotateable8bitImmediate: (operands at: 0) ifTrue: [:r :i| false] ifFalse: [true]].
		[MoveRM16r]	-> [^self is12BitValue: (operands at: 1) ifTrue: [:u :i| false] ifFalse: [true]].
		[MoveM64rRd]	-> [^self is12BitValue: (operands at: 0) ifTrue: [:u :i| false] ifFalse: [true]].
		[MoveMwrR]	-> [^self is12BitValue: (operands at: 0) ifTrue: [:u :i| false] ifFalse: [true]].
		[PushCw]		-> [^(self inCurrentCompilation: (operands at: 0)) not].
		[PushCq]		-> [^self rotateable8bitImmediate: (operands at: 0) ifTrue: [:r :i| false] ifFalse: [true]].
		[PrefetchAw] 	-> [^(self isAddressRelativeToVarBase: (operands at: 0)) ifTrue: [false] ifFalse: [true]].
		}
		otherwise: [self assert: false].
	^false "to keep C compiler quiet"
! !

CogAbstractInstruction subclass: #CogARMv8Compiler
	instanceVariableNames: ''
	classVariableNames: 'AL ArithmeticAdd ArithmeticAddS ArithmeticSub ArithmeticSubS CASAL CArg0Reg CArg1Reg CArg2Reg CArg3Reg CArg4Reg CArg5Reg CArg6Reg CBNZ CBZ CC CCMPNE CLREX CS CSET D0 D1 D10 D11 D12 D13 D14 D15 D16 D17 D18 D19 D2 D20 D21 D22 D23 D24 D25 D26 D27 D28 D29 D3 D30 D31 D4 D5 D6 D7 D8 D9 DC DC_CISW DC_CIVAC DC_CSW DC_CVAC DC_CVAU DC_ISW DC_IVAC DC_ZVA DMB DSB DSB_ALL DSB_ALLSY DSB_ISH DSB_NSH DSB_OSH DSB_READS DSB_SY DSB_WRITES DataCacheFlushRequired DataCacheLineLength DivRRR EQ FP GE GT HI HasAtomicInstructions IC IC_IALLU IC_IALLUIS IC_IVAU ISB InstructionCacheFlushRequired InstructionCacheLineLength LDAXR LE LR LS LT LogicalAnd LogicalAndS LogicalOr LogicalXor MI MRS_CTR_EL0 MRS_ID_AA64ISAR0_EL1 MSubRRR MoveAwRR MoveRRAw MulOverflowRRR MulRRR NE NativePopRR NativePushRR PL R0 R1 R10 R11 R12 R13 R14 R15 R16 R17 R18 R19 R2 R20 R21 R22 R23 R24 R25 R26 R27 R28 R29 R3 R30 R31 R4 R5 R6 R7 R8 R9 SMULHRRR SP STLR STLXR SXTB SXTH SXTW SXTX UXTB UXTH UXTW UXTX VC VS XZR'
	poolDictionaries: 'ARMv8A64Opcodes'
	category: 'VMMaker-JIT'!

!CogARMv8Compiler commentStamp: 'eem 1/7/2021 23:01' prior: 0!
I generate ARMv8 machine code instructions from CogAbstractInstructions with CogRTLOpcodes.
Here in "Arm ARM" refers to
	Arm® Architecture Reference Manual
	Armv8, for Armv8-A architecture profile
https://developer.arm.com/docs/ddi0487/latest/arm-architecture-reference-manual-armv8-for-armv8-a-architecture-profile

Some things to know about ARMv8 instructions:
Whether 31 in a register field implies the zero register or the SP register(s) depends on the specific instruction.

C3.2.1 Load/Store register
If a Load instruction specifies writeback and the register being loaded is also the base register,
then behavior is CONSTRAINED UNPREDICTABLE and one of the following behaviors must occur:
- The instruction is treated as UNDEFINED.
- The instruction is treated as a NOP.
- The instruction performs the load using the specified addressing mode and the base register
  becomes UNKNOWN.  In addition, if an exception occurs during the execution of such an
  instruction, the base address might be corrupted so that the instruction cannot be repeated.
If a Store instruction performs a writeback and the register that is stored is also the base register,
then behavior is CONSTRAINED UNPREDICTABLE and one of the following behaviors must occur:
- The instruction is treated as UNDEFINED.
- The instruction is treated as a NOP.
- The instruction performs the store to the designated register using the specified addressing
  mode, but the value stored is UNKNOWN.!

!CogARMv8Compiler class methodsFor: 'translation' stamp: 'eem 11/24/2019 08:01'!
ISA
	"Answer the name of the ISA the receiver implements."
	^#ARMv8! !

!CogARMv8Compiler class methodsFor: 'accessing' stamp: 'eem 12/17/2019 08:36'!
VarBaseReg
	"Answer the number of the reg we use to hold the base address of CoInterpreter variables"
	^VarBaseReg! !

!CogARMv8Compiler class methodsFor: 'translation' stamp: 'eem 7/9/2021 20:21'!
declareCVarsIn: aCCodeGenerator
		
	#('hasAtomicInstructions' 'instructionCacheLineLength' 'instructionCacheFlushRequired' 'dataCacheLineLength' 'dataCacheFlushRequired') do:
		[:varName|
		aCCodeGenerator
			declareVar: varName type: #'unsigned char';
			removeConstant: varName capitalized].
	aCCodeGenerator
		var: #ceFlushDCache "sigh; this is here to placate Slang; we don't use this, but Slang doesn't elide the assignment..."
			declareC: 'static void (*ceFlushDCache)(usqIntptr_t from, usqIntptr_t to)'! !

!CogARMv8Compiler class methodsFor: 'class initialization' stamp: 'eem 7/7/2021 17:42'!
extraClassVariableNames
	"self extraClassVariableNames"
	^(organization listAtCategoryNamed: #'feature detection')
		inject: Set new
		into: [:them :selector|
			(self >> selector) literalsDo:
				[:l| (l isVariableBinding and: [classPool includesKey: l key]) ifTrue: [them add: l key]].
			them]! !

!CogARMv8Compiler class methodsFor: 'translation' stamp: 'eem 10/28/2020 17:09'!
identifyingPredefinedMacros
	^#('__arm64__' '__aarch64__' '__ARM_ARCH_ISA_A64' '__ARM_ARCH >= 8' 'ARM64' 'ARMv8') "The last two are probably bogus..."! !

!CogARMv8Compiler class methodsFor: 'class initialization' stamp: 'eem 01/19/2021 20:40'!
initialize
	"Initialize various ARM64 instruction-related constants."
	"self initialize"

	"main registers; a minor complication in reading the doc.
	ARM refer to the 64bit registers as X0...30 and use R0...30 to refer to the 32bit lower halves.
	They also use a whole suite of names for the floating point/SIMD registers. See ARMARM DDI0487 B1.2.1 etc for the gory details.
	Note that R30 (yes, yes, X30) is used as the link register and as such is not really a general purpose register. 
	Also note that 31 in a general register field means R31, and that in most of these instructions R31 is the zero
	register named XZR in ARM doc.  but in the rest of these instructions R31 the effective SP register.
	XZR is a pseudo-register that always reads as 0 and writes to /dev/null.
	And note that unlike the ARM32, there is no general purpose register for the PC; a big difference.
	See ARMARM DDI0487 C1.2.5. wrt to both the lack of a PC register and the XZR/SP distinction."
	
	"and initialize most sets of variables that run from 0 to N - 1..."
	#(	"General registers, 0 to 31. We stick with R0...30 to refer to the 64 bit general regs and D0...31 (note the extra reg here!!) for the FP/SIMD regs"
		(R0 R1 R2 R3 R4 R5 R6 R7 R8 R9 R10 R11 R12 R13 R14 R15 R16 R17 R18 R19 R20 R21 R22 R23 R24 R25 R26 R27 R28 R29 R30 R31)
		"Floating-point registers, 0 to 31"
		(D0 D1 D2 D3 D4 D5 D6 D7 D8 D9 D10 D11 D12 D13 D14 D15 D16 D17 D18 D19 D20 D21 D22 D23 D24 D25 D26 D27 D28 D29 D30 D31)
		"C argument registers, 0 to 6"
		(CArg0Reg CArg1Reg CArg2Reg CArg3Reg CArg4Reg CArg5Reg CArg6Reg)
		"Condition Codes 0 to 16. Note that cc=16rF is mapped back to AL in AARCH64. Generally it shouldn't be used."
		(EQ NE CS CC MI PL VS VC HI LS GE LT GT LE AL)
		"Logical Op Codes"
		(LogicalAnd LogicalOr LogicalXor LogicalAndS)
		"Arithmetic Opcodes"
		"ADD (shifted register) on page C6-763			10001011
		 ADDS (shifted register) on page C6-771		10101011
		 CMN (shifted register) on page C6-854			10101011	Rd=XZR
		 SUB (shifted register) on page C6-1313		11001011
		 SUBS (shifted register) on page C6-1323		11101011
		 CMP (shifted register) on page C6-860			11101011	Rd=XZR
		 NEG (shifted register) on page C6-1114		11001011	Rn=XZR
		 NEGS					on page C6-1116		11101011	Rn=XZR"
		(ArithmeticAdd ArithmeticAddS ArithmeticSub ArithmeticSubS)
		"Extension Methods "
		(UXTB UXTH UXTW UXTX "a.k.a. LSL" SXTB SXTH SXTW SXTX)) do:
		[:classVarNames|
		 classVarNames withIndexDo:
			[:k :v|
			CogARMv8Compiler classPool at: k put: v - 1]].

	SP := XZR := R31.
	LR := R30.
	FP := R29.

	"DC variant selectors; see concretizeDataCacheControl"
	DC_CISW := 13.
	DC_CIVAC := 14.
	DC_CSW := 15.
	DC_CVAC := 16.
	DC_CVAU := 19.
	DC_ISW := 26.
	DC_IVAC := 27.
	DC_ZVA := 28.

	"IC variant selectors; see concretizeInstructionCacheControl"
	IC_IALLU := 0.
	IC_IALLUIS := 1.
	IC_IVAU := 2.

	"DSB domains and types	C6.2.81 DSB	C6-891"
	DSB_OSH := 0.		"Domain_OuterSharable"
	DSB_NSH := 1.		"Domain_NonSharable"
	DSB_ISH := 2.		"Domain_InnerSharable"
	DSB_SY := 3.		"Domain_FullSystem"

	DSB_ALLSY := 0.	"Types_All; domain = Domain_FullSystem"
	DSB_READS := 1.	"Types_Reads"
	DSB_WRITES := 2.	"Types_Writes"
	DSB_ALL := 3.		"Types_All; domain ~= Domain_FullSystem"

	"Specific instructions"
	self
		initializeSpecificOpcodes: #(MulRRR MulOverflowRRR SMULHRRR DivRRR MSubRRR "N.B. ARMv8 has MSUBRRRR but we only support three operands"
									MoveAwRR MoveRRAw NativePushRR NativePopRR "these map to ldp/stp"
									"Cache control and memory barrier"
									"B2.3.7		Memory barriers	B2-124"
									DC IC DMB DSB ISB MRS_CTR_EL0 MRS_ID_AA64ISAR0_EL1
									CASAL CBNZ CBZ CCMPNE CSET CLREX LDAXR STLXR STLR)
		in: thisContext method! !

!CogARMv8Compiler class methodsFor: 'class initialization' stamp: 'eem 2/21/2020 11:51'!
initializeAbstractRegisters
	"Assign the abstract registers with the identities/indices of the relevant concrete registers."

	"See Table 3-1 Register Usage in AArch64 SMC32, HVC32, SMC64, and HVC64 calls in
	 SMC CALLING CONVENTION
	 System Software on ARM® Platforms Document number: ARM DEN 0028B
	 http://infocenter.arm.com/help/topic/com.arm.doc.den0028b/ARM_DEN0028B_SMC_Calling_Convention.pdf"

	TempReg			:= R0.
	ClassReg			:= R4.
	ReceiverResultReg	:= R5.
	SendNumArgsReg	:= R6.
	Arg0Reg			:= R7.
	Arg1Reg			:= R8.
	RISCTempReg		:= R9.
	SPReg				:= R16."we cannot easily use the native SP for SPReg because of the typical
								 16-byte alignment restriction, enforced if the SA0 bit is set in SCTLR_EL1."
						"X18/R18 is the Platform register; leave it be"
	Extra0Reg			:= R19. "R18 through R28 are callee saved"
	Extra1Reg			:= R20.
	Extra2Reg			:= R21.
	Extra3Reg			:= R22.
	Extra4Reg			:= R23.
	Extra5Reg			:= R24.
	Extra6Reg			:= R25.
	Extra7Reg			:= R26.
	Extra8Reg			:= R27.
	VarBaseReg			:= R28.	
	FPReg				:= FP.	"a.k.a. R29"
	LinkReg			:= LR.	"a.k.a. R30"
	NativeSPReg		:= SP.	"a.k.a. R31; we cannot easily use the native SP for SPReg because of the typical
								 16-byte alignment restriction, enforced if the SA0 bit is set in SCTLR_EL1."

	ABICalleeSavedRegisterMask := (self registerMaskFor: 19 and: 20 and: 21 and: 22 and: 23)
									 + (self registerMaskFor: 24 and: 25 and: 26 and: 27 and: 28).
	ABICallerSavedRegisterMask := ((self registerMaskFor: 0 and: 1 and: 2 and: 3 and: 4 and: 5 and: 6 and: 7 and: 8 and: 9)
									+ (self registerMaskFor: 10 and: 11 and: 12 and: 13 and: 14 and: 15 and: 16 and: 17)).
	CallerSavedRegisterMask := ABICallerSavedRegisterMask
									bitAnd: (self registerMaskFor: ClassReg and: ReceiverResultReg and: SendNumArgsReg and: Arg0Reg and: Arg1Reg).

	NumRegisters := 32.

	DPFPReg0			:= D0.
	DPFPReg1			:= D1.
	DPFPReg2			:= D2.
	DPFPReg3			:= D3.
	DPFPReg4			:= D4.
	DPFPReg5			:= D5.
	DPFPReg6			:= D6.
	DPFPReg7			:= D7.
	DPFPReg8			:= D8.
	DPFPReg9			:= D9.
	DPFPReg10			:= D10.
	DPFPReg11			:= D11.
	DPFPReg12			:= D12.
	DPFPReg13			:= D13.
	DPFPReg14			:= D14.
	DPFPReg15			:= D15.

	NumFloatRegisters := 32! !

!CogARMv8Compiler class methodsFor: 'accessing class hierarchy' stamp: 'eem 12/17/2019 12:42'!
literalsManagerClass
	^OutOfLineLiteralsManagerFor64Bits! !

!CogARMv8Compiler class methodsFor: 'translation' stamp: 'eem 3/11/2017 12:46'!
machineCodeDeclaration
	"Answer the declaration for the machineCode array.
	 ARM instructions are 32-bits in length."
	^{#'unsigned int'. '[', self basicNew machineCodeWords printString, ']'}! !

!CogARMv8Compiler class methodsFor: 'trampoline support' stamp: 'eem 10/30/2020 05:21'!
numTrampolines
	^2 "ceFlushICache, and optionally ceFlushDCache"! !

!CogARMv8Compiler class methodsFor: 'translation' stamp: 'eem 1/7/2021 23:00'!
preambleCCode
	^'#if __APPLE__ && __MACH__ /* Mac OS X */\#include <libkern/OSCacheControl.h>\#endif' withCRs! !

!CogARMv8Compiler class methodsFor: 'debug printing' stamp: 'eem 7/7/2021 15:33'!
printCTR_EL0: ctr_el0
	<doNotGenerate>
	^String streamContents:
		[:s| | fieldPrinter l1ip |
		fieldPrinter := [:name :startBit| | field |
						s nextPutAll: name; nextPutAll: ', ['; print: startBit + 3; nextPut: $,; print: startBit; nextPutAll: ']		log2 words: '.
						s print: (field := ctr_el0 >> startBit bitAnd: 15); nextPutAll: ' ('; print: 8 << field; nextPutAll: ' bytes)'; cr].
		s
			nextPutAll: 'DIC, bit [29]	Instruction cache invalidation requirements for instruction to data coherence. The meaning of this bit is:';
			crtab;
			nextPutAll: ((ctr_el0 noMask: 1 << 29)
							ifTrue: ['0b0 Instruction cache invalidation to the Point of Unification is required for instruction to data coherence.']
							ifFalse: ['0b1 Instruction cache cleaning to the Point of Unification is not required for instruction to data coherence.']);
			cr.
		s
			nextPutAll: 'IDC, bit [28]	Data cache clean requirements for instruction to data coherence. The meaning of this bit is:';
			crtab;
			nextPutAll: ((ctr_el0 noMask: 1 << 28)
							ifTrue: ['0b0 Data cache clean to the Point of Unification is required for instruction to data coherence, unless CLIDR.LoC == 0b000 or (CLIDR.LoUIS == 0b000 && CLIDR.LoUU == 0b000).']
							ifFalse: ['0b1 Data cache clean to the Point of Unification is not required for instruction to data coherence.']);
			cr.

		fieldPrinter
			value: 'Cache writeback granule' value: 24;
			value: 'Exclusives reservation granule' value: 20;
			value: 'DminLine' value: 16;
			value: 'IminLine' value: 0.
		s
			nextPutAll: 'Level 1 instruction cache policy: '; print: (l1ip := ctr_el0 >> 14 bitAnd: 2); space;
			nextPutAll: (#(	'VMID aware Physical Index, Physical tag (VPIPT)'
							'ASID-tagged Virtual Index, Virtual Tag (AIVIVT)'
							'Virtual Index, Physical Tag (VIPT)'
							'Physical Index, Physical Tag (PIPT)') at: l1ip + 1);
			cr]! !

!CogARMv8Compiler class methodsFor: 'debug printing' stamp: 'eem 11/12/2020 16:05'!
printFormatForOpcodeName: opcodeName
	"Answer a sequence of $r, $f or nil for the operands in the opcode, used for printing, where
	 r => integer register, f => floating point register, and nil => numeric or address operand.
	 Subclasses can override to provide a format string for their own private opcodes."
	^opcodeName caseOf: {
		[#CASAL]	-> ['rrr'].
		[#CCMPNE]	-> ['rrc'].
		[#CSET]	-> [#($r nil)].
		[#LDAXR]	-> ['rr'].
		[#STLXR]	-> ['rr'].
		[#STLR]	-> ['rr'] }
		otherwise: [#()]! !

!CogARMv8Compiler class methodsFor: 'class initialization' stamp: 'eem 7/7/2021 17:31'!
specificOpcodes
	"Answer the processor-specific opcodes for this class.
	 They're all in an Array literal in the initialize method."
	^(self class >> #initialize) literals detect: [:l| l isArray and: [l includes: #MulRR]] ifNone: [#()]! !

!CogARMv8Compiler class methodsFor: 'translation' stamp: 'tpr 11/24/2019 11:33'!
wordSize
	"This is a 64-bit ISA"
	^8! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 12/1/2019 21:31'!
addrn: rn rd: rd imm: offset shiftBy12: shiftBy12
	"C6.2.4 ADD (immediate) p761"
	self assert: (offset between: 0 and: 1 << 12 - 1).
	^2r10010001000000000000000000000000
	+ (shiftBy12 ifTrue: [1 << 22] ifFalse: [0]) + (offset << 10) + (rn << 5) + rd! !

!CogARMv8Compiler methodsFor: 'debugging' stamp: 'eem 2/4/2020 19:19'!
assertCoherentCodeAt: codeAddress delta: theCodeToDataDelta
	<inline: #always>
	| codeInstruction dataInstruction |
	codeInstruction := self instructionAt: codeAddress.
	dataInstruction := self instructionAt: codeAddress + theCodeToDataDelta.
	self assert: codeInstruction = dataInstruction! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 2/4/2020 14:32'!
bl: callDistance
	"C6.2.33	BL		C6-812"
	<inline: true>
	self assert: (callDistance noMask: 3).
	^2r100101 << 26 + (callDistance >>> 2 bitAnd: 1 << 26 - 1)! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 12/20/2019 18:39'!
brlink: link reg: reg
	^2r1101011000011111 << 16
	+ (link ifTrue: [1 << 21] ifFalse: [0])
	+ (reg << 5)! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 12/25/2019 15:07'!
byteReadsZeroExtend
	"Answer if a byte read, via MoveAbR, MoveMbrR, or MoveXbrRR zero-extends
	 into the full register, or merely affects the least significant 8 bits of the
	 the register.  By default the code generator assumes that byte reads
	 to not zero extend.  Note that byte reads /must not/ sign extend."
	^true! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 7/8/2021 16:28'!
callFullInstructionByteSize
	^8! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 4/19/2015 08:33'!
callInstructionByteSize
	"ARM calls and jumps span +/- 32 mb, more than enough for intra-zone calls and jumps."
	^4! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 02/04/2020 17:28'!
callTargetFromReturnAddress: mcpc
	| instr |
	instr := self instructionAt: mcpc - 4.
	"C6.2.26 	B		C6-799
	 C6.2.33	BL		C6-812"
	self assert: (instr >> 26 = 2r100101 or: [instr >> 26 = 2r101]).
	^(self
		cCode: [(instr bitAnd: 1 << 26 - 1) << 38]
		inSmalltalk: [(instr bitAnd: 1 << 26 - 1) << 38 bitAnd: 1 << 64 - 1])
			signedIntFromLong64 >>> 36 + (mcpc - 4)! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 12/21/2019 23:21'!
canDivQuoRem
	^true! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 12/21/2019 19:02'!
canMulRR
	^true! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 9/2/2020 09:40'!
canMulRRDetectOverflow
	"ARMv8's intrinsic integer multiply doesn't set the overflow flag, so we
	 need to generate additional code to implement overflow checking."
	<inline: true>
	^false! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 1/23/2020 19:39'!
canSignExtend
	"ARMv8 has native SignExtend8RR, SignExtend16RR, & SignExtend32RR."
	<inline: true>
	^true! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 6/21/2018 12:29'!
canZeroExtend
	"x64 has native ZeroExtend8RR, ZeroExtend16RR, & ZeroExtend32RR."
	<inline: true>
	^true! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 11/20/2015 16:49'!
cmpC32RTempByteSize
	^8! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 4/21/2015 09:17'!
codeGranularity
	"Answer the size in bytes of a unit of machine code."
	<inline: true>
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 12/8/2019 21:52'!
computeJumpTargetOffset
	<inline: true>
	| jumpTarget |
	<var: #jumpTarget type: #'AbstractInstruction *'>
	jumpTarget := self jumpTargetAddress.
	^jumpTarget signedIntFromLong64 - address signedIntFromLong64! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 1/23/2020 14:06'!
computeLowBit: nArg
	" Answer the index of the low order one bit.
		2r00101000 lowBit       (Answers: 4)
		2r-00101000 lowBit      (Answers: 4)
	  This is an implementation of SmallInteger>>lowBit"

	| n result |
	nArg = 0 ifTrue: [^0].
	n := nArg.
	result := 1.
	[n noMask: 16rFF]
		whileTrue:
			[result := result + 8.
			n := n bitShift: -8].
	(n noMask: 16rF) ifTrue:
		[result := result + 4.
		 n := n bitShift: -4].
	(n noMask: 16r3) ifTrue:
		[result := result + 2.
		 n := n bitShift: -2].
	^(n anyMask: 1)
		ifTrue: [result]
		ifFalse: [result + 1]

	"| me |
	me := self basicNew.
	(0 to: 65535) reject: [:i| i lowBit = (me lowBit: i)]"! !

!CogARMv8Compiler methodsFor: 'generate machine code' stamp: 'eem 8/1/2021 23:18'!
computeMaximumSize
	"Because we don't use Thumb, each ARMv8 instruction has 4 bytes. Several
	 abstract opcodes need more than one instruction. Instructions that refer to
	 constants and/or literals depend on literals being stored out-of-line or encoded
	 in immediate instruction fields (i.e. we only support OutOfLineLiteralsManager.

	 N.B.  The ^N forms are to get around the old bytecode compiler's long branch
	 limits which are exceeded when each case jumps around the otherwise."

	opcode
		caseOf: {
		"Noops & Pseudo Ops"
		[Label]						-> [^0].
		[Literal]						-> [^self literalSize].
		[AlignmentNops]			-> [^(operands at: 0) - 4].
		"Control"
		[CallFull]					-> [^8].
		[JumpFull]					-> [^8].
		[JumpLongZero]			-> [^8].
		[JumpLongNonZero]		-> [^8].
		[JumpMulOverflow]			-> [^8].
		[JumpNoMulOverflow]		-> [^8].
		[JumpFPOrdered]			-> [^8].
		[JumpFPUnordered]		-> [^8].
		[RetN]						-> [^(operands at: 0) = 0 ifTrue: [4] ifFalse: [8]].
		[NativeRetN]				-> [^(operands at: 0) = 0 ifTrue: [4] ifFalse: [8]].

		"Arithmetic"
		[AddCqR]						-> [^self isPossiblyShiftableNegatableImm12: (operands at: 0) signedIntFromLong64
												ifTrue: [:ign|4] ifFalse: [8]].
		[AddCqRR]						-> [^self isPossiblyShiftableNegatableImm12: (operands at: 0) signedIntFromLong64
												ifTrue: [:ign|4] ifFalse: [8]].
		[CmpCqR]						-> [^self isPossiblyShiftableNegatableImm12: (operands at: 0) signedIntFromLong64
												ifTrue: [:ign|4] ifFalse: [8]].
		[SubCqR]						-> [^self isPossiblyShiftableNegatableImm12: (operands at: 0) signedIntFromLong64
												ifTrue: [:ign|4] ifFalse: [8]].
		[LoadEffectiveAddressMwrR]	-> [^self isPossiblyShiftableNegatableImm12: (operands at: 0) signedIntFromLong64
												ifTrue: [:ign|4] ifFalse: [8]].
		[AndCqR]					-> [^self isImmNImmSImmREncodableBitmask: (operands at: 0) ifTrue: [:n :imms :immr|4] ifFalse:[8]].
		[AndCqRR]					-> ["N.B. For three operand logical ops only support AndCqRR with a NativeSPReg target, used for alignment purposes."
										^self isImmNImmSImmREncodableBitmask: (operands at: 0)
												ifTrue: [:n :imms :immr| (operands at: 2) = SP ifTrue: [8] ifFalse: [4]]
												ifFalse:[(operands at: 2) = SP ifTrue: [12] ifFalse: [8]]].
		[OrCqR]					-> [^self isImmNImmSImmREncodableBitmask: (operands at: 0) ifTrue: [:n :imms :immr|4] ifFalse:[8]].
		[OrCqRR]					-> [^self isImmNImmSImmREncodableBitmask: (operands at: 0) ifTrue: [:n :imms :immr|4] ifFalse:[8]].
		[TstCqR]					-> [^self isImmNImmSImmREncodableBitmask: (operands at: 0) ifTrue: [:n :imms :immr|4] ifFalse:[8]].
		[XorCqR]					-> [^self isImmNImmSImmREncodableBitmask: (operands at: 0) ifTrue: [:n :imms :immr|4] ifFalse:[8]].
		[AddCwR]					-> [^8].
		[AndCwR]					-> [^8].
		[CmpCwR]					-> [^8].
		[CmpC32R]					-> [^8].
		[OrCwR]					-> [^8].
		[SubCwR]					-> [^8].
		[XorCwR]					-> [^8].
		[SubRR]					-> [^(operands at: 0) = SP ifTrue: [8] ifFalse: [4]].
		[SubRRR]					-> [^(operands at: 0) = SP ifTrue: [8] ifFalse: [4]].

		"ARMv8 Specific Arithmetic"
		[MulOverflowRRR]			-> [^12].
		"Data Movement"						
		[MoveAwR]				-> [^((self isAddressRelativeToVarBase: (operands at: 0))
									    or: [cogit addressIsInCurrentCompilation: (operands at: 0)])
										ifTrue: [(operands at: 1) ~= SP ifTrue: [4] ifFalse: [8]]
										ifFalse: [(operands at: 1) ~= SP ifTrue: [8] ifFalse: [12]]].
		[MoveRAw]				-> [^((self isAddressRelativeToVarBase: (operands at: 1))
									    or: [cogit addressIsInCurrentCompilation: (operands at: 1)])
										ifTrue: [(operands at: 0) ~= SP ifTrue: [4] ifFalse: [8]]
										ifFalse: [(operands at: 0) ~= SP ifTrue: [8] ifFalse: [12]]].
		[MoveAwRR]			-> [self assert: (self isAddressRelativeToVarBase: (operands at: 0)).
									^((operands at: 1) = SP or: [(operands at: 2) = SP])
										ifTrue: [8] ifFalse: [4]].
		[MoveRRAw]			-> [self assert: (self isAddressRelativeToVarBase: (operands at: 2)).
									^((operands at: 0) = SP or: [(operands at: 1) = SP])
										ifTrue: [8] ifFalse: [4]].
		[MoveAbR]				-> [^(self isAddressRelativeToVarBase: (operands at: 0)) ifTrue: [4] ifFalse: [8]].
		[MoveRAb]				-> [^(self isAddressRelativeToVarBase: (operands at: 1)) ifTrue: [4] ifFalse: [8]].
		[MoveMwrR]			-> [^(self isImm12orImm9offset: (operands at: 0)) ifTrue: [4] ifFalse: [8]].
		[MoveRMwr]			-> [^(self isImm12orImm9offset: (operands at: 1)) ifTrue: [4] ifFalse: [8]].
		[MoveM32rR]			-> [^(self isImm12orImm9offset: (operands at: 0)) ifTrue: [4] ifFalse: [8]].
		[MoveRM32r]			-> [^(self isImm12orImm9offset: (operands at: 1)) ifTrue: [4] ifFalse: [8]].
		[MoveM16rR]			-> [^(self isImm12orImm9offset: (operands at: 0)) ifTrue: [4] ifFalse: [8]].
		[MoveRM16r]			-> [^(self isImm12orImm9offset: (operands at: 1)) ifTrue: [4] ifFalse: [8]].
		[MoveMbrR]			-> [^(self isImm12orImm9offset: (operands at: 0)) ifTrue: [4] ifFalse: [8]].
		[MoveRMbr]			-> [^(self isImm12orImm9offset: (operands at: 1)) ifTrue: [4] ifFalse: [8]].
		[MoveM64rRd]			-> [^(self isUnsigned12BitMultipleOf8: (operands at: 0)) ifTrue: [4] ifFalse: [8]].
		[MoveRdM64r]			-> [^(self isUnsigned12BitMultipleOf8: (operands at: 1)) ifTrue: [4] ifFalse: [8]].
		[PushCw]				-> [^8].
		[PushCq]				-> [^8].
		}
		otherwise: [^4].
	^0 "to keep C compiler quiet"
! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 7/21/2021 09:49'!
concretizeAddCqRDest: destReg
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| constant srcReg offset |
	<var: #constant type: #sqInt>
	constant := (operands at: 0) signedIntFromLong64.
	srcReg := operands at: 1.

	self isPossiblyShiftableImm12: constant
		ifTrue:
			[:shift|
			"C6.2.4		ADD (immediate)	C6-761
			 C6.2.8		ADDS (immediate)	C6-769"
			machineCode
				at: 0
				put: (srcReg = SP
						ifTrue:  [2r100100010 "ADD"]
						ifFalse: [2r101100010 "ADDS"]) << 23
					+ (shift ifTrue: [constant >> 2 + (1 << 22)] ifFalse: [constant << 10])
					+ (srcReg << 5)
					+ destReg.
			^4]
		ifFalse: [].
	self isPossiblyShiftableImm12: constant negated
		ifTrue:
			[:shift|
			"C6.2.308		SUB (immediate)	C6-1311
			 C6.2.314		SUBS (immediate)	C6-1321"
			machineCode
				at: 0
				put: (srcReg = SP
						ifTrue:  [2r110100010 "SUB"]
						ifFalse: [2r111100010 "SUBS"]) << 23
					+ (shift ifTrue: [constant negated >> 2 + (1 << 22)] ifFalse: [constant negated << 10])
					+ (srcReg << 5)
					+ destReg.
			^4]
		ifFalse: [].
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	"cogit processor disassembleInstructionAt: 4 In: machineCode object"
	offset := self emitMoveCw: constant intoR: RISCTempReg at: 0.
	"C6.2.7		ADDS (extended register)		C6-766"
	machineCode
				at: offset // 4
				put: 2r10101011001 << 21
					+ (RISCTempReg << 16)
					+ (SXTX << 13)
					+ (srcReg << 5)
					+ destReg.
	^offset + 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/21/2020 13:34'!
concretizeAlignmentNops
	<inline: true>
	"See outputMachineCodeAt:"
	self assert: machineCodeSize \\ 4 = 0.
	machineCode at: 0 put: NOP.
	^machineCodeSize! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/15/2019 12:39'!
concretizeArithmeticShiftRightCqR
	<inline: true>
	"C6.2.17 	ASR (immediate)	C6-785		100100110 (1)
	 C6.2.177	LSL (immediate)	C6-1075	110100110 (1)
	 C6.2.180	LSR (immediate)	C6-1081	110100110 (1)"
	| reg constant |
	constant := operands at: 0.
	reg := operands at: 1.
	self assert: (constant between: 1 and: 63).
	machineCode
		at: 0
		put: 2r1001001101 << 22
			+ (constant << 16)
			+ (63 << 10)
			+ (reg << 5)
			+ reg.
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/26/2019 17:13'!
concretizeArithmeticShiftRightCqRR
	<inline: true>
	"C6.2.17 	ASR (immediate)	C6-785		100100110 (1)
	 C6.2.177	LSL (immediate)	C6-1075	110100110 (1)
	 C6.2.180	LSR (immediate)	C6-1081	110100110 (1)"
	| constant srcReg  destReg |
	constant := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	self assert: (constant between: 1 and: 63).
	machineCode
		at: 0
		put: 2r1001001101 << 22
			+ (constant << 16)
			+ (63 << 10)
			+ (srcReg << 5)
			+ destReg.
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 10/27/2020 17:06'!
concretizeCASAL
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| constantReg "Xs" valueReg "Xt" baseReg "Xn" |
	constantReg := operands at: 0.
	valueReg := operands at: 1.
	baseReg := operands at: 2.
	machineCode
		at: 0
		put: 2r11001000111 << 21
			+ (constantReg << 16)
			+ (63 << 10)
			+ (baseReg << 5)
			+ valueReg.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 11/12/2020 18:43'!
concretizeCB
	"C6.2.44/45 		CBNZ/CBZ 			C6-831..."
	<inline: true>
	| offset reg |
	offset := self computeJumpTargetOffset.
 	self assert: (offset ~= 0 and: [self isInImmediateBranchRange: offset]).
	reg := operands at: 1.
	machineCode
		at: 0
		put: 2r101101 << 26
			+ (opcode = CBNZ ifTrue: [1 << 24] ifFalse: [0])
			+ ((offset bitAnd: 1 << 21 - 1) << 3)
			+ reg.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 11/12/2020 14:35'!
concretizeCCMPNE
	"C6.2.49 CCMP (register)	C6-839"
	<inline: true>
	| rM rN flags |
	rM := operands at: 0.
	rN := operands at: 1.
	flags := operands at: 2.
	machineCode
		at: 0
		put: 2r11111010010 << 21
			+ (rM << 16)
			+ (NE << 12)
			+ (rN << 5)
			+ flags.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 11/12/2020 18:49'!
concretizeCLREX
	"C6.2.54		CLREX 			C6-847"
	<inline: true>
	machineCode
		at: 0
		put: 2r11010101000000110011111101011111.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 10/27/2020 16:43'!
concretizeCSET
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| reg condition |
	reg := operands at: 0.
	condition := operands at: 1.
	machineCode
		at: 0
		put: 2r1001101010011111 << 16
			+ ((condition bitXor: 1) "strange, but true" << 12)
			+ (2r111111 << 5)
			+ reg.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 2/25/2020 12:20'!
concretizeCacheControlOp1: op1 CRm: crm Op2: op2
	<inline: true>
	machineCode
		at: 0
		put: 2r1101010100001 << 19
			+ (op1 << 16)
			+ (2r0111 << 12) "CRn"
			+ (crm << 8)
			+ (op2 << 5)
			+ (operands at: 0).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4

"Scripture (see DminLine & IminLine below):

D13.2.33	CTR_EL0, Cache Type Register		D13-2912

	Provides information about the architecture of the caches.

	AArch64 System register CTR_EL0[31:0] is architecturally mapped to AArch32 System register CTR[31:0].

	CTR_EL0 is a 64-bit register.

	DIC, bit [29]
		Instruction cache invalidation requirements for instruction to data coherence. The meaning of this bit is:
			0b0	Instruction cache invalidation to the Point of Unification is required for instruction to data coherence.
			0b1	Instruction cache cleaning to the Point of Unification is not required for instruction to data coherence.

	IDC, bit [28]
		Data cache clean requirements for instruction to data coherence. The meaning of this bit is:
			0b0	Data cache clean to the Point of Unification is required for instruction to data coherence,
					unless CLIDR.LoC == 0b000 or (CLIDR.LoUIS == 0b000 && CLIDR.LoUU == 0b000).
			0b1 Data cache clean to the Point of Unification is not required for instruction to data coherence.

	CWG, bits [27:24]
		Cache writeback granule. Log2 of the number of words of the maximum size of memory that can be overwritten as a result of the eviction of a cache entry that has had a memory location in it modified.
		A value of 0b0000 indicates that this register does not provide Cache writeback granule information and either:
			- The architectural maximum of 512 words (2KB) must be assumed.
			- The Cache writeback granule can be determined from maximum cache line size encoded in the Cache Size ID Registers.
		Values greater than 0b1001 are reserved.
		Arm recommends that an implementation that does not support cache write-back implements this field as 0b0001. This applies, for example, to an implementation that supports only write-through caches.

	ERG, bits [23:20]
		Exclusives reservation granule. Log2 of the number of words of the maximum size of the reservation granule that has been implemented for the Load-Exclusive and Store-Exclusive instructions.
		The use of the value 0b0000 is deprecated.
		The value 0b0001 and values greater than 0b1001 are reserved.

	DminLine, bits [19:16]
		Log2 of the number of words in the smallest cache line of all the data caches and unified caches that are controlled by the PE.

	L1Ip, bits [15:14]
		Level 1 instruction cache policy. Indicates the indexing and tagging policy for the L1 instruction cache. Possible values of this field are:
			0b00 VMID aware Physical Index, Physical tag (VPIPT)
			0b01 ASID-tagged Virtual Index, Virtual Tag (AIVIVT)
			0b10 Virtual Index, Physical Tag (VIPT)
			0b11 Physical Index, Physical Tag (PIPT)
		The value 0b01 is reserved in Armv8.
		The value 0b00 is permitted only in an implementation that includes ARMv8.2-VPIPT, otherwise the value is reserved.

	Bits [13:4] Reserved, RES0.

	IminLine, bits [3:0]
		Log2 of the number of words in the smallest cache line of all the instruction caches that are controlled by the PE."! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/17/2019 15:00'!
concretizeCall
	<inline: true>
	| offset |
	offset := ((operands at: 0) - address) signedIntFromLong64.
	self assert: (offset noMask: 3).
	self assert: (self isInImmediateBranchAndLinkRange: offset).
	"C6.2.33	BL		C6-812"
	machineCode at: 0 put: (self bl: offset).
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 2/4/2020 16:13'!
concretizeCallJumpFull: link
	"Sizing/generating calls.
		Jump targets can be to absolute addresses or other abstract instructions.
		Generating initial trampolines instructions may have no maxSize and be to absolute addresses.
		Otherwise instructions must have a machineCodeSize which must be kept to."
	<inline: true>
	| jumpTarget instrOffset|
	<var: #jumpTarget type: #'AbstractInstruction *'>
	jumpTarget := self longJumpTargetAddress.
	instrOffset := self emitMoveCw: jumpTarget asUnsignedInteger intoR: RISCTempReg at: 0.
	machineCode
		at: instrOffset // 4
		put: (self brlink: link reg: RISCTempReg).
	self assert: instrOffset = self literalLoadInstructionBytes.
	^instrOffset + 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/23/2020 15:58'!
concretizeClzRR
	"C6.2.56	CLZ		C6-849"
	<inline: true>
	| maskReg dest  |
	maskReg := operands at: 0.
	dest := operands at: 1.
	machineCode
		at: 0
		put: 2r11011010110000000001 << 12
			+ (maskReg << 5)
			+ dest.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/27/2019 19:05'!
concretizeCmpC32R
	<inline: true>
	| constant rn |
	constant := operands at: 0.
	rn := operands at: 1.

	self assert: (cogit addressIsInCurrentCompilation: dependent address).
	self assert: dependent address \\ 4 = 0.
	self assert: (dependent address - address) abs < (1<<18).
	"C6.2.131	LDR (literal)		C6-979"
	machineCode
		at: 0
		put: 2r00011 << 27
			+ (dependent address - address << 3 "5 - 2")
			+ RISCTempReg.

	"C6.2.60	CMP (extended register)	C6-856"
	machineCode
		at: 1
		put: 2r01101011001 << 21 "N.B. sf == 0!!!!"
			+ (RISCTempReg << 16)
			+ (UXTX << 13)
			+ (rn << 5)
			+ XZR.
	^8! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/27/2019 01:14'!
concretizeCmpCqR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| constant rn offset |
	<var: #constant type: #sqInt>
	constant := (operands at: 0) signedIntFromLong64.
	rn := operands at: 1.

	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	self isPossiblyShiftableImm12: constant
		ifTrue:
			[:shift|
			"C6.2.61	CMP (immediate)	C6-858
			 C6.2.314	SUBS (immediate)	C6-1321"
			machineCode
				at: 0
				put: 2r111100010 << 23
					+ (shift ifTrue: [constant >> 2 + (1 << 22)] ifFalse: [constant << 10])
					+ (rn << 5)
					+ XZR.
			^4]
		ifFalse: [].
	self isPossiblyShiftableImm12: constant negated
		ifTrue:
			[:shift|
			"C6.2.58	CMN (immediate)	C6-852
			 C6.2.8		ADDS (immediate)	C6-769"
			machineCode
				at: 0
				put: 2r101100010 << 23
					+ (shift ifTrue: [constant negated >> 2 + (1 << 22)] ifFalse: [constant negated << 10])
					+ (rn << 5)
					+ XZR.
			^4]
		ifFalse: [].
	offset := self emitMoveCw: constant intoR: RISCTempReg at: 0.
	"C6.2.60	CMP (extended register)	C6-856"
	machineCode
				at: offset // 4
				put: 2r11101011001 << 21
					+ (RISCTempReg << 16)
					+ (UXTX << 13)
					+ (rn << 5)
					+ XZR.
	^offset + 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/28/2019 22:12'!
concretizeCmpRdRd
	"C7.2.59	FCMP	C7-1506"
	<inline: true>
	machineCode
		at: 0
		put: 2r11110011 << 21
		+ ((operands at: 0) << 16)
		+ (1 << 13)
		+ ((operands at: 1) << 5).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 2/4/2020 18:26'!
concretizeConditionalJump: conditionCode
	"Will get inlined into concretizeAt: switch."
	"Sizing/generating jumps.
		Jump targets can be to absolute addresses or other abstract instructions.
		Generating initial trampolines instructions may have no maxSize and be to absolute addresses.
		Otherwise instructions must have a machineCodeSize which must be kept to."
	<inline: false>
	| offset |
	offset := self computeJumpTargetOffset.
 	self assert: (offset ~= 0 and: [self isInImmediateBranchRange: offset]).
	machineCode at: 0 put: (self cond: conditionCode offset: offset). "B offset"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 2/4/2020 14:35'!
concretizeConditionalLongJump: conditionCode
	"Generate a short conditional branch around a long unconditional jump.
	 C6.2.25	B.cond		C6-798
	 C6.2.26	B			C6-799
	 Note that C6.2.44 CBNZ and C6.2.45 CBZ aren't useful for Closed PIC jumps.
	 They compare against zero and have only a +/-1Mb range."
	<inline: true>
	| offset |
	self assert: (conditionCode = EQ or: [conditionCode = NE]).
	offset := (operands at: 0) - (address + 4).
	self assert: (offset noMask: 3).
	self assert: (offset >>> 2 between: -1 << 26 and: 1 << 26 - 1).
	machineCode
		at: 0
		put: 2r101010 << 25
			+ (2 << 5 "skip next instruction")
			+ (conditionCode = EQ ifTrue: [NE] ifFalse: [EQ]);
		at: 1
		put: 2r101 << 26 + (offset >>> 2 bitAnd: 1 << 26 - 1).
	^8! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/28/2019 22:12'!
concretizeConvertRRd
	"Table C3-67 Floating-point and integer or fixed-point conversion instructions	C3-229
	 SCVTF (scalar, integer) on page C7-1894"
	<inline: true>
	machineCode
		at: 0
		put: 2r1001111001100010 << 16
			+ ((operands at: 0) << 5)
			+ (operands at: 1).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/28/2019 22:12'!
concretizeConvertRdR
	"Table C3-67 Floating-point and integer or fixed-point conversion instructions	C3-229
	 FCVTZS (scalar, integer) on page C7-1568"
	<inline: true>
	machineCode
		at: 0
		put: 2r1001111001111 << 19
			+ ((operands at: 0) << 5)
			+ (operands at: 1).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/28/2019 14:23'!
concretizeCwRArithmetic: arithOp Rd: destRegOrXZR
	| instrBytes |
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	"cogit processor disassembleInstructionAt: 4 In: machineCode object"
	instrBytes := self emitMoveCw: (operands at: 0) intoR: RISCTempReg at: 0.
	"ADD (extended register) on page C6-758
	 ADDS (extended register) on page C6-766
	 SUB (extended register) on page C6-1308
	 SUBS (extended register) on page C6-1318"
	machineCode
		at: instrBytes // 4
		put: 2r10001011001 << 21
			+ (arithOp << 29)
			+ (RISCTempReg <<16)
			+ (UXTX << 13)
			+ ((operands at: 1) << 5)
			+ destRegOrXZR.
	^instrBytes + 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/18/2019 00:44'!
concretizeCwRLogical: op
	| destReg constant offset |
	constant := operands at: 0.
	destReg := operands at: 1.
	offset := self emitMoveCw: constant intoR: RISCTempReg at: 0.
	"AND	(shifted register) - 64-bit variant on page C6-777
	 BIC	(shifted register) - 64-bit variant on page C6-808
	 ORR	(shifted register) - 64-bit variant on page C6-1127
	 ORN	(shifted register) - 64-bit variant on page C6-1123
	  EOR	(shifted register) - 64-bit variant on page C6-898
	  EON	(shifted register) - 64-bit variant on page C6-894
	  ANDS	(shifted register) - 64-bit variant on page C6-781
	  BICS	(shifted register) - 64-bit variant on page C6-810"
	machineCode
				at: offset // 4
				put: 2r1000101 << 25
					+ (op << 29)
					+ (RISCTempReg << 16)
					+ (destReg << 5)
					+ destReg.
	^offset + 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 2/25/2020 12:17'!
concretizeDataCacheControl
	"Issue a DC CIVAC, CVAC or CVAU
	 C5.3.14	DC CIVAC, Data or unified Cache line Clean and Invalidate by VA to PoC		C5-486
					Clean and Invalidate data cache by address to Point of Coherency.
	 C5.3.16	DC CVAC, Data or unified Cache line Clean by VA to PoC						C5-490
					Clean data cache by address to Point of Coherency.
	 C5.3.19	DC CVAU, Clean data cache by address to Point of Unification				C5-496
					Clean data cache by address to Point of Unification."

	"(operands at: 0) is the target register, accessed within concretizeCacheControlOp1:CRm:Op2:"
	| cacheOpcode |
	cacheOpcode := (operands at: 1) caseOf: {
						[DC_CIVAC] ->	[2r1110].
						[DC_CVAC] ->	[2r1010].
						[DC_CVAU] ->	[2r1011] }.
	^self concretizeCacheControlOp1: 2r011 CRm: cacheOpcode Op2: 2r001

"Scripture:
C5.3.1		DC CGDSW, Data, Allocation Tag or unified Cache line Clean of Data and Allocation Tags by Set/Way					C5-460
C5.3.2		DC CGDVAC, Data, Allocation Tag or unified Cache line Clean of Allocation Tags by VA to PoC							C5-462
C5.3.3		DC CGDVADP, Data, Allocation Tag or unified Cache line Clean of Allocation Tags by VA to PoDP							C5-464
C5.3.4		DC CGDVAP, Data, Allocation Tag or unified Cache line Clean of Data and Allocation Tags by VA to PoP					C5-466
C5.3.5		DC CGSW, Data, Allocation Tag or unified Cache line Clean of Allocation Tags by Set/Way								C5-468
C5.3.6		DC CGVAC, Data, Allocation Tag or unified Cache line Clean of Allocation Tags by VA to PoC								C5-470
C5.3.7		DC CGVADP, Data, Allocation Tag or unified Cache line Clean of Data and Allocation Tags by VA to PoDP					C5-472
C5.3.8		DC CGVAP, Data, Allocation Tag or unified Cache line Clean of Allocation Tags by VA to PoP								C5-474
C5.3.9		DC CIGDSW, Data, Allocation Tag or unified Cache line Clean and Invalidate of Data and Allocation Tags by Set/Way		C5-476
C5.3.10	DC CIGDVAC, Data, Allocation Tag or unified Cache line Clean and Invalidate of Data and Allocation Tags by VA to PoC	C5-478
C5.3.11	DC CIGSW, Data, Allocation Tag or unified Cache line Clean and Invalidate of Allocation Tags by Set/Way				C5-480
C5.3.12	DC CIGVAC, Data, Allocation Tag or unified Cache line Clean and Invalidate of Allocation Tags by VA to PoC				C5-482
C5.3.13	DC CISW, Data or unified Cache line Clean and Invalidate by Set/Way													C5-484
C5.3.14	DC CIVAC, Data or unified Cache line Clean and Invalidate by VA to PoC													C5-486
C5.3.15	DC CSW, Data or unified Cache line Clean by Set/Way																	C5-488
C5.3.16	DC CVAC, Data or unified Cache line Clean by VA to PoC																	C5-490
C5.3.17	DC CVADP, Data or unified Cache line Clean by VA to PoDP																C5-492
C5.3.18	DC CVAP, Data or unified Cache line Clean by VA to PoP																	C5-494
C5.3.19	DC CVAU, Data or unified Cache line Clean by VA to PoU																C5-496
C5.3.20	DC GVA, Data Cache set Allocation Tag by VA																			C5-498
C5.3.21	DC GZVA, Data Cache set Allocation Tags and Zero by VA																C5-500
C5.3.22	DC IGDSW, Data, Allocation Tag or unified Cache line Invalidate of Data and Allocation Tags by Set/Way					C5-502
C5.3.23	DC IGDVAC, Data, Allocation Tag or unified Cache line Invalidate of Allocation Tags by VA to PoC							C5-504
C5.3.24	DC IGSW, Data, Allocation Tag or unified Cache line Invalidate of Allocation Tags by Set/Way							C5-506
C5.3.25	DC IGVAC, Data, Allocation Tag or unified Cache line Invalidate of Allocation Tags by VA to PoC							C5-508
C5.3.26	DC ISW, Data or unified Cache line Invalidate by Set/Way																C5-510
C5.3.27	DC IVAC, Data or unified Cache line Invalidate by VA to PoC																C5-512
C5.3.28	DC ZVA, Data Cache Zero by VA																						C5-514

C6.2.75	DC		C6-883
			1101010100 0 01 op1:3 [0111,CRn:4=7] CRm:4 op2:3 Rt:5

			DC <dc_op>, <Xt> equivalent to SYS #<op1>. C7, <Cm>, #<op2>, <Xt> = SysOp(op1,0111,CRm,op2) == Sys_DC
			<dc_op>:
				C5.3.27	IVAC		when op1 = 000, CRm = 0110, op2 = 001		Invalidate data cache by address to Point of Coherency.
				C5.3.26	ISW		when op1 = 000, CRm = 0110, op2 = 010		Invalidate data cache by set/way.
				C5.3.15	CSW		when op1 = 000, CRm = 1010, op2 = 010		Clean data cache by set/way.
				C5.3.13	CISW		when op1 = 000, CRm = 1110, op2 = 010		Clean and Invalidate data cache by set/way.
				C5.3.28	ZVA		when op1 = 011, CRm = 0100, op2 = 001		Zero data cache by address. Zeroes a naturally aligned block
																							of N bytes, where the size of N is identified in DCZID_EL0.
				C5.3.16	CVAC		when op1 = 011, CRm = 1010, op2 = 001		Clean data cache by address to Point of Coherency.
				C5.3.19	CVAU		when op1 = 011, CRm = 1011, op2 = 001		Clean data cache by address to Point of Unification.
				C5.3.14	CIVAC		when op1 = 011, CRm = 1110, op2 = 001		Clean and Invalidate data cache by address to Point of Coherency."! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 2/25/2020 12:12'!
concretizeDataSynchronizationBarrier
	"C6.2.81	DSB	C6-891"
	<inline: true>
	| CRm |
	CRm := (operands at: 0) << 2 + (operands at: 1).
	machineCode
		at: 0
		put: 2r11010101000000110011000010011111
			+ (CRm << 8).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/22/2019 00:09'!
concretizeDivRRR
	"C6.2.234	SDIV	C6-1174"
	<inline: true>
	"Rd := Rn / Rm"
	"reg3 := reg1 / reg2"
	| reg1 reg2 reg3 |
	reg1 := operands at: 0.
	reg2 := operands at: 1.
	reg3 := operands at: 2.
	machineCode
		at: 0
		put: 2r10011010110 << 21
			+ (reg2 << 16)
			+ (2r000011 << 10)
			+ (reg1 << 5)
			+ reg3.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/28/2019 22:12'!
concretizeFill32
	"fill with operand 0 according to the processor's endianness"
	<inline: true>
	machineCode at: 0 put: (operands at: 0).
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 2/25/2020 12:13'!
concretizeInstructionCacheControl
	"Issue an IC IVAU	Invalidate instruction cache by address to Point of Unification.
	 C5.3.31	IC IVAU, Instruction Cache line Invalidate by VA to PoU	C5-518"
	"(operands at: 0) is the target register, accessed within concretizeCacheControlOp1:CRm:Op2:"
	"Ho hum, table C5-1 page C5-367 states that IC_IALLU & IC_IALLUIS are available from EL1 or higher,
	 so they are off limits to us."
	<inline: true>
	^true
		ifTrue: [self concretizeCacheControlOp1: 2r011 CRm: 2r0101 Op2: 2r001 "IC_IVAU"]
		ifFalse:
			[(operands at: 1) caseOf: {
				[IC_IALLU] ->	[self concretizeCacheControlOp1: 2r000 CRm: 2r0101 Op2: 2r000].
				[IC_IALLUIS] ->	[self concretizeCacheControlOp1: 2r000 CRm: 2r0001 Op2: 2r000].
				[IC_IVAU] ->	[self concretizeCacheControlOp1: 2r011 CRm: 2r0101 Op2: 2r001] }]

"Scripture:
C5.3.29	IC IALLU, Instruction Cache Invalidate All to PoU						C5-516
C5.3.30	IC IALLUIS, Instruction Cache Invalidate All to PoU, Inner Shareable		C5-517
C5.3.31	IC IVAU, Instruction Cache line Invalidate by VA to PoU					C5-518

C6.2.94	IC		C6-910
			1101010100 0 01 op1:3 [0111,CRn:4=7] CRm:4 op2:3 Rt:5

			IC <ic_op>, <Xt> equivalent to SYS #<op1>. C7, <Cm>, #<op2>{, <Xt>} = SysOp(op1,0111,CRm,op2) == Sys_IC

				C5.3.30	IALLUIS		when op1 = 000, CRm = 0001, op2 = 000		Invalidate all instruction caches in Inner Shareable domain to Point of Unification.
				C5.3.29	IALLU		when op1 = 000, CRm = 0101, op2 = 000		Invalidate all instruction caches to Point of Unification.
				C5.3.31	IVAU		when op1 = 011, CRm = 0101, op2 = 001		Invalidate instruction cache by address to Point of Unification."! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 2/25/2020 12:12'!
concretizeInstructionSynchronizationBarrier
	"C6.2.96	ISB		C6-912"
	<inline: true>
	| CRm |
	CRm := 2r1111.
	machineCode
		at: 0
		put: 2r11010101000000110011000011011111
			+ (CRm << 8).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 2/4/2020 14:35'!
concretizeJumpLong
	"C6.2.26 	B		C6-799"
	<inline: true>
	| offset |
	offset := (operands at: 0) - address.
	self assert: (offset noMask: 3).
	self assert: (offset >>> 2 between: -1 << 26 and: 1 << 26 - 1).
	machineCode
		at: 0
		put: 2r101 << 26 + (offset >>> 2 bitAnd: 1 << 26 - 1).
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/28/2019 22:12'!
concretizeJumpR
	"C6.2.36	BR	C6-816"
	<inline: true>
	machineCode
		at: 0
		put: 2r1101011 << 25
			+ (XZR << 16)
			+ ((operands at: 0) << 5).
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 11/12/2020 16:31'!
concretizeLDAXR
	"C6.2.113		LDAXR		C6-944"
	<inline: true>
	| targetReg addressReg |
	targetReg := operands at: 0.
	addressReg := operands at: 1.
	machineCode
		at: 0
		put: 2r1100100001011111111111 << 10
			+ (addressReg << 5)
			+ targetReg.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/17/2019 13:13'!
concretizeLiteral
	"Generate an out-of-line literal.  Copy the value and any annotation from the stand-in in the literals manager."
	| literalAsInstruction literal |
	<var: 'literal' type: #usqInt>
	literalAsInstruction := cogit cCoerceSimple: (operands at: 0) to: #'AbstractInstruction *'.
	literal := (self isAnInstruction: literalAsInstruction)
				ifTrue: [literalAsInstruction address]
				ifFalse: [self cCode: [literalAsInstruction asUnsignedInteger]
							inSmalltalk: [literalAsInstruction]].
	self assert: (dependent notNil and: [dependent opcode = Literal]).
	dependent annotation ifNotNil:
		[self assert: annotation isNil.
		 annotation := dependent annotation].
	dependent address ifNotNil: [self assert: dependent address = address].
	dependent address: address.
	machineCode at: 0 put: (literal bitAnd: 16rFFFFFFFF).
	self literalSize = 4 ifTrue:
		[^4].
	machineCode at: 1 put: (literal >> 32).
	^8! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/8/2020 18:37'!
concretizeLoadEffectiveAddressMwrR
	| baseReg offset destReg instrBytes |
	<var: #offset type: #sqInt>
	offset := operands at: 0.
	baseReg := operands at: 1.
	destReg := operands at: 2.
	self isPossiblyShiftableImm12: offset
		ifTrue:
			[:shift|
			"C6.2.4		ADD (immediate)	C6-761"
			machineCode
				at: 0
				put: 2r100100010 << 23
					+ (shift ifTrue: [offset >> 2] ifFalse: [offset << 10])
					+ (baseReg << 5)
					+ destReg.
			^4]
		ifFalse: [].
	self isPossiblyShiftableImm12: offset negated
		ifTrue:
			[:shift|
			"C6.2.308		SUB (immediate)	C6-1311"
			machineCode
				at: 0
				put: 2r100100010 << 23
					+ (shift ifTrue: [offset negated >> 2] ifFalse: [offset negated << 10])
					+ (baseReg << 5)
					+ destReg.
			^4]
		ifFalse: [].
	instrBytes := self emitMoveCw: offset intoR: RISCTempReg at: 0.
	"C6.2.3		ADD (extended register) C6-758"
	machineCode
		at: instrBytes // 4
		put: 2r10001011001 << 21
			+ (destReg << 16)
			+ (UXTX << 13)
			+ (RISCTempReg << 5)
			+ destReg.
	^instrBytes + 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 8/1/2021 23:19'!
concretizeLogicalOp: op CqRDest: destReg
	"AND	(immediate) - 64-bit variant on page C6-775
	 ORR	(immediate) - 64-bit variant on page C6-1125
	 EOR	(immediate) - 64-bit variant on page C6-896
	 ANDS	(immediate) - 64-bit variant on page C6-779
	 C6.2.329	TST (immediate)	C6-1346"
	<inline: false>
	| srcReg constant effectiveDestReg offset |
	constant := operands at: 0.
	srcReg := operands at: 1.
	"N.B. For three operand logical ops only support AndCq: const R: reg R: NativeSPReg, which is used for alignment."
	effectiveDestReg := (opcode = AndCqRR and: [destReg = SP]) ifTrue: [RISCTempReg] ifFalse: [destReg].
	self isImmNImmSImmREncodableBitmask: constant
		ifTrue:
			[:n :imms :immr|
			 machineCode
				at: 0
				put: 2r1001001 << 25
					+ (op << 29)
					+ (n << 22)
					+ (immr << 16)
					+ (imms << 10)
					+ (srcReg << 5)
					+ effectiveDestReg.
			 offset := 4]
		ifFalse:
			[offset := self emitMoveCw: constant intoR: RISCTempReg at: 0.
			"OPC	N
			 00		0	AND (shifted register) - 64-bit variant on page C6-777
			 00		1	BIC (shifted register) - 64-bit variant on page C6-808
			 01		0	ORR (shifted register) - 64-bit variant on page C6-1127
			 01		1	ORN (shifted register) - 64-bit variant on page C6-1123
			 10		0	EOR (shifted register) - 64-bit variant on page C6-898
			 10		1	EON (shifted register) - 64-bit variant on page C6-894
			 11		0	ANDS (shifted register) - 64-bit variant on page C6-781
			 11		0	BICS (shifted register) - 64-bit variant on page C6-810"
			machineCode
				at: offset // 4
				put: 2r1000101 << 25
					+ (op << 29)
					+ (RISCTempReg << 16)
					+ (srcReg << 5)
					+ effectiveDestReg.
			offset := offset + 4].
	(opcode = AndCqRR and: [destReg = SP]) ifFalse:
		[^offset].
	machineCode at: offset // 4 put: (self movern: effectiveDestReg rd: destReg).
	^offset + 4
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	"cogit processor disassembleInstructionAt: 4 In: machineCode object"
	"cogit processor disassembleInstructionAt: 8 In: machineCode object"! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 22:49'!
concretizeLogicalShiftLeftCqR
	<inline: true>
	"C6.2.17 	ASR (immediate)	C6-785		100100110 (1)
	 C6.2.177	LSL (immediate)	C6-1075	110100110 (1)
	 C6.2.180	LSR (immediate)	C6-1081	110100110 (1)"
	| reg constant |
	constant := operands at: 0.
	reg := operands at: 1.
	self assert: (constant between: 1 and: 63).
	machineCode
		at: 0
		put: 2r1101001101 << 22
			+ (64 - constant << 16)
			+ (63 - constant << 10)
			+ (reg << 5)
			+ reg.
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/26/2019 17:14'!
concretizeLogicalShiftLeftCqRR
	<inline: true>
	"C6.2.17 	ASR (immediate)	C6-785		100100110 (1)
	 C6.2.177	LSL (immediate)	C6-1075	110100110 (1)
	 C6.2.180	LSR (immediate)	C6-1081	110100110 (1)"
	| constant srcReg  destReg |
	constant := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	self assert: (constant between: 1 and: 63).
	machineCode
		at: 0
		put: 2r1101001101 << 22
			+ (64 - constant << 16)
			+ (63 - constant << 10)
			+ (srcReg << 5)
			+ destReg.
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/9/2019 15:27'!
concretizeLogicalShiftRightCqR
	<inline: true>
	"C6.2.17 	ASR (immediate)	C6-785		100100110 (1)
	 C6.2.177	LSL (immediate)	C6-1075	110100110 (1)
	 C6.2.180	LSR (immediate)	C6-1081	110100110 (1)"
	| reg constant |
	constant := operands at: 0.
	reg := operands at: 1.
	self assert: (constant between: 1 and: 63).
	machineCode
		at: 0
		put: 2r1101001101 << 22
			+ (constant << 16)
			+ (63 << 10)
			+ (reg << 5)
			+ reg.
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/25/2019 18:59'!
concretizeLogicalShiftRightCqRR
	<inline: true>
	"C6.2.17 	ASR (immediate)	C6-785		100100110 (1)
	 C6.2.177	LSL (immediate)	C6-1075	110100110 (1)
	 C6.2.180	LSR (immediate)	C6-1081	110100110 (1)"
	| constant srcReg  destReg |
	constant := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	self assert: (constant between: 1 and: 63).
	machineCode
		at: 0
		put: 2r1101001101 << 22
			+ (constant << 16)
			+ (63 << 10)
			+ (srcReg << 5)
			+ destReg.
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 2/25/2020 12:19'!
concretizeMRSOp1: op1 CRn: crn CRm: crm Op2: op2
	"C6.2.192	MRS	C6-1104
	 Move System Register allows the PE to read an AArch64 System register into a general-purpose register."
	<inline: true>
	machineCode
		at: 0
		put: 2r1101010100111 << 19
			+ (op1 << 16)
			+ (crn << 12)
			+ (crm << 8)
			+ (op2 << 5)
			+ (operands at: 0).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 2/25/2020 12:13'!
concretizeMRS_CTR_EL0
	"C6.2.192	MRS	C6-1104
	 Move System Register allows the PE to read an AArch64 System register into a general-purpose register.
	 D12.3 op0==0b11, Moves to and from non-debug System registers, Special-purpose registers		D12-2808
	 Register Accessed	op0		op1		CRn	CRm	op2
	 ...
	 CTR_EL0			3		3		0		0		1"
	<inline: true>
	^self concretizeMRSOp1: 3 CRn: 0 CRm: 0 Op2: 1! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 11/12/2020 12:08'!
concretizeMRS_ID_AA64ISAR0_EL1
	"C6.2.192	MRS	C6-1104
	 Move System Register allows the PE to read an AArch64 System register into a general-purpose register.
	 D12.3 op0==0b11, Moves to and from non-debug System registers, Special-purpose registers		D12-2808
	 Register Accessed		op0		op1		CRn	CRm	op2
	 ...
	 CTR_EL0				3		3		0		0		1
	 ID_AA64ISAR0_EL1	3		0		0		5		1"
	<inline: true>
	^self concretizeMRSOp1: 0 CRn: 0 CRm: 6 Op2: 0! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/27/2019 12:38'!
concretizeMSubRRR
	"C6.2.195	MSUB	C6-1109
	 Xd := Xa - (Xn * Xm)"
	<inline: true>
	"rD := rA - (rN * rM)"
	| rN rA rM rD |
	rM := (operands at: 0) >> 5.
	rN := (operands at: 0) bitAnd: 31.
	rA := operands at: 1.
	rD := operands at: 2.
	machineCode
		at: 0
		put: 2r10011011 << 24
			+ (rM << 16)	"Xm"
			+ (1 << 15)
			+ (rA << 10)	"Xa"
			+ (rN << 5)	"Xn"
			+ rD.			"Xd"
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/9/2020 18:33'!
concretizeMathRdRd: floatArithOp
	"C7.2.43	FADD (scalar)	C7-1426"
	"C7.2.91	FDIV (scalar)	C7-1582"
	"C7.2.129	FMUL (scalar)	C7-1664"
	"C7.2.167	FSUB (scalar)	C7-1747"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0
		put: 2r11110011 << 21
			+ (srcReg << 16)
			+ (floatArithOp << 10)
			+ (destReg << 5)
			+ destReg.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/21/2019 09:51'!
concretizeMoveAbR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcAddr destReg instrOffset |
	srcAddr := operands at: 0.
	destReg := operands at: 1.
	self deny: SP = destReg.
	"ldr srcReg, [VarBaseReg, #offset] except that this is illegal for SP/X31"
	(self isAddressRelativeToVarBase: srcAddr) ifTrue:
		[srcAddr < cogit varBaseAddress ifTrue:
			[self shouldBeImplemented.
			 ^4].
		 ^self emitLd: 0 rn: VarBaseReg rt: destReg imm: srcAddr - cogit varBaseAddress shiftBy12: false at: 0].
	instrOffset := self emitMoveCw: srcAddr intoR: RISCTempReg at: 0.
	^self emitLd: 0 rn: RISCTempReg rt: destReg imm: 0 shiftBy12: false at: instrOffset! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/21/2019 09:51'!
concretizeMoveAwR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcAddr destReg instrOffset |
	srcAddr := operands at: 0.
	destReg := operands at: 1.
	"ldr srcReg, [VarBaseReg, #offset] except that this is illegal for SP/X31"
	(self isAddressRelativeToVarBase: srcAddr) ifTrue:
		[srcAddr < cogit varBaseAddress ifTrue:
			[self shouldBeImplemented.
			 ^4].
		 destReg ~= SP ifTrue:
			[^self emitLd: 3 rn: VarBaseReg rt: destReg imm: srcAddr - cogit varBaseAddress shiftBy12: false at: 0].
		 instrOffset := self emitLd: 3 rn: VarBaseReg rt: RISCTempReg imm: srcAddr - cogit varBaseAddress shiftBy12: false at: 0.
		 machineCode
			at: instrOffset / 4
			put: (self movern: RISCTempReg rd: destReg).
		^instrOffset + 4].
	instrOffset := self emitMoveCw: srcAddr intoR: RISCTempReg at: 0.
	SP ~= destReg ifTrue:
		[^self emitLd: 3 rn: RISCTempReg rt: destReg imm: 0 shiftBy12: false at: instrOffset].
	instrOffset := self emitLd: 3 rn: RISCTempReg rt: RISCTempReg imm: 0 shiftBy12: false at: instrOffset.
	machineCode
		at: instrOffset // 4
		put: (self movern: RISCTempReg rd: destReg).
	^instrOffset + 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 2/25/2020 12:39'!
concretizeMoveAwRR
	"C6.2.128 	LDP		C6-970"
	| pairAddress reg1 reg2 |
	<inline: true>
	pairAddress := operands at: 0.
	reg1 := operands at: 1.
	reg2 := operands at: 2.
	self assert: reg1 ~= reg2.
	self assert: (reg1 ~= RISCTempReg and: [reg2 ~= RISCTempReg]).
	self assert: (self isAddressRelativeToVarBase: pairAddress).
	self assert: pairAddress - cogit varBaseAddress / 8 < (1 << 6).
	reg1 = SP ifTrue: [reg1 := RISCTempReg].
	reg2 = SP ifTrue: [reg2 := RISCTempReg].
	machineCode
		at: 0
		put: 2r1010100101 << 22
			+ (pairAddress - cogit varBaseAddress / 8 << 15)
			+ (reg2 << 10)
			+ (VarBaseReg << 5)
			+ reg1.
	(reg1 ~= RISCTempReg and: [reg2 ~= RISCTempReg]) ifTrue:
		[^4].
	machineCode
		at: 1
		put: (self movern: RISCTempReg rd: SP).
	^8

	"cogit processor
		disassembleInstructionAt: 0 In: machineCode object;
		disassembleInstructionAt: 4 In: machineCode object"! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/20/2019 18:07'!
concretizeMoveC32R
	<inline: true>
	| constant rn |
	constant := operands at: 0.
	rn := operands at: 1.

	self assert: (cogit addressIsInCurrentCompilation: dependent address).
	self assert: dependent address \\ 4 = 0.
	self assert: (dependent address - address) < (1 << 20).
	"C6.2.131	LDR (literal)		C6-979
	 C6.2.143	LDRSW (literal)	C6-1008"
	machineCode
		at: 0
		put: 2r10011 << 27
			+ ((dependent address - address  bitAnd: 1 << 21 - 1) << 3 "5 - 2")
			+ rn.
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 3/18/2020 20:23'!
concretizeMoveCqR
	"C3.3.4		Move (immediate)	C3-215

	 The Move (immediate) instructions are aliases for a single MOVZ, MOVN, or ORR (immediate with zero register),
	 instruction to load an immediate value into the destination register. An assembler must permit a signed or
	 unsigned immediate, as long as its binary representation can be generated using one of these instructions,
	 and an assembler error results if the immediate cannot be generated in this way. On disassembly, it is
	 unspecified whether the immediate is output as a signed or an unsigned value.

	 C6.2.191	MOVZ	C6-1102	Move wide with zero moves an optionally-shifted 16-bit immediate value to a register.
	 C6.2.190	MOVN	C6-1100	Move wide with NOT moves the inverse of an optionally-shifted 16-bit immediate value to a register.
	 C6.2.204	ORR (immediate)	C6-1125
										Bitwise OR (immediate) performs a bitwise (inclusive) OR of a register value and an immediate
										register value, and writes the result to the destination register."

	| constant destReg |
	constant := operands at: 0.
	destReg := operands at: 1.
	destReg ~= SP ifTrue:
		[| lowBit lowBitMod16 mask |
		lowBit := constant > 0
					ifTrue: [self cCode: [self computeLowBit: constant] inSmalltalk: [constant lowBit - 1]]
					ifFalse: [0].
		lowBitMod16 := lowBit // 16 * 16.
		mask := 1 << 16 - 1 << lowBitMod16.
		(constant bitAnd: mask) = constant ifTrue:
			["Use MOVZ"
			 machineCode
				at: 0
				put: 2r110100101 << 23
					+ (lowBitMod16 // 16 << 21)
					+ (constant >> lowBitMod16 << 5)
					+ destReg.
			 ^4].
		lowBit := constant signedIntFromLong64 < -1
					ifTrue: [self cCode: [self computeLowBit: constant bitInvert64] inSmalltalk: [constant bitInvert64 lowBit - 1]]
					ifFalse: [0].
		(constant bitOr: mask) signedIntFromLong64 = -1 ifTrue:
			["Use MOVN"
			 self assert: (constant bitInvert64 >> lowBitMod16) = ((constant bitInvert64 >> lowBitMod16) bitAnd: mask).
			 machineCode
				at: 0
				put: 2r100100101 << 23
					+ (lowBitMod16 // 16 << 21)
					+ (constant bitInvert64 >> lowBitMod16 << 5)
					+ destReg.
			 ^4]].
	^self isImmNImmSImmREncodableBitmask: constant
		ifTrue:
			[:n :imms :immr| "Use ORR"
			machineCode
				at: 0
				put: 2r1011001001 << 22
					+ (immr << 16)
					+ (imms << 10)
					+ (XZR << 5)
					+ destReg.
			4]
		ifFalse: [self emitMoveCw: constant intoR: destReg at: 0]! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/18/2019 00:46'!
concretizeMoveCwR
	<inline: true>
	^self emitMoveCw: (operands at: 0) intoR: (operands at: 1) at: 0! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/28/2019 22:49'!
concretizeMoveM64rRd
	"Load/Store Vector on page C3-206.
	 C7.2.184	LDR (immediate, SIMD&FP)	C7-1800"
	<inline: true>
	| rd offset baseReg |
	offset := operands at: 0.
	baseReg := operands at: 1.
	rd := operands at: 2.
	self assert: (offset noMask: 7).
	machineCode
		at: 0
		put: 2r1111110101 << 22
			+ (offset << 7 "10 - 3")
			+ (baseReg << 5)
			+ rd.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/21/2019 09:56'!
concretizeMoveMSrR: unitSizeLog2MinusOne
	"Mwr/M32r/M16r/Mbr - memory unit whose address is a constant M away from an address in a register"
	^self emitLd: unitSizeLog2MinusOne
		rn: (operands at: 1)
		rt: (operands at: 2) signedIntFromLong64
		imm: (operands at: 0)
		shiftBy12: false
		at: 0
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 8/5/2021 07:15'!
concretizeMovePerfCnt64RL
	"D13.8.26		CNTVCT_EL0, Counter-timer Virtual Count register		p D13-3774"

	"MRS <Xt>, CNTVCT_EL0		op0:0b11 op1:0b011 CRn:0b1110 CRm:0b0000 op2:0b010"
	machineCode
		at: 0
		put: (2r1101010100111 << 19 "MRS + op0"
			+ (2r011 << 16)	"op1"
			+ (2r1110 << 12)	"CRn"
			+ (2r010 << 5)
			+ (operands at: 0)).
	^4
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/21/2019 09:50'!
concretizeMoveRAb
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg destAddr instrOffset |
	srcReg := operands at: 0.
	destAddr := operands at: 1.
	self deny: SP = srcReg.
	(self isAddressRelativeToVarBase: destAddr) ifTrue:
		[destAddr < cogit varBaseAddress ifTrue:
			[self shouldBeImplemented.
			 ^4].
		 machineCode
			at: 0
			put: (self st: 0 rn: VarBaseReg rt: srcReg imm: destAddr - cogit varBaseAddress shiftBy12: false).
		 ^4].
	instrOffset := self emitMoveCw: destAddr intoR: RISCTempReg at: 0.
	machineCode
		at: instrOffset // 4
		put: (self st: 0 rn: RISCTempReg rt: srcReg imm: 0 shiftBy12: false).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^instrOffset + 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/21/2019 09:49'!
concretizeMoveRAw
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg destAddr instrOffset |
	srcReg := operands at: 0.
	destAddr := operands at: 1.
	"str srcReg, [VarBaseReg, #offset] except that this is illegal for srcReg = SP/X31"
	(self isAddressRelativeToVarBase: destAddr) ifTrue:
		[destAddr < cogit varBaseAddress ifTrue:
			[self shouldBeImplemented.
			 ^4].
		 srcReg ~= SP ifTrue:
			[machineCode
				at: 0
				put: (self st: 3 rn: VarBaseReg rt: srcReg imm: destAddr - cogit varBaseAddress shiftBy12: false).
			 ^4].
		machineCode
			at: 0
			put: (self movern: srcReg rd: RISCTempReg);
			at: 1
			put: (self st: 3 rn: VarBaseReg rt: RISCTempReg imm: destAddr - cogit varBaseAddress shiftBy12: false).
		^8].
	instrOffset := self emitMoveCw: destAddr intoR: RISCTempReg at: 0.
	SP ~= srcReg ifTrue:
		[machineCode
			at: instrOffset // 4
			put: (self st: 3 rn: RISCTempReg rt: srcReg imm: 0 shiftBy12: false).
		^instrOffset + 4].
	machineCode
		at: instrOffset // 4
		put: (self movern: srcReg rd: CArg1Reg);
		at: instrOffset // 4
		put: (self st: 3 rn: RISCTempReg rt: CArg1Reg imm: 0 shiftBy12: false).
	^instrOffset + 8
	! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 7/21/2021 09:15'!
concretizeMoveRMSr: unitSizeLog2MinusOne
	"Mwr/M32r/M16r/Mbr - memory unit whose address is a constant M away from an address in a register"
	^self
		emitSt: unitSizeLog2MinusOne
		rn: (operands at: 2)
		rt: (operands at: 0)
		imm: (operands at: 1) signedIntFromLong64
		shiftBy12: false

	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	"cogit processor disassembleInstructionAt: 4 In: machineCode object"! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/14/2019 09:00'!
concretizeMoveRR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	"C6.2.184 MOV (to/from SP) p1089
	 C6.2.188 MOV (register) p1096"
	machineCode
		at: 0
		put: ((srcReg = SP or: [destReg = SP])
				ifTrue:  [2r10010001 << 24
						+ (srcReg << 5)
						+ destReg]
				ifFalse: [2r10101010 << 24
						 + (srcReg << 16)
						 + (XZR << 5)
						 + destReg]).
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 2/25/2020 12:39'!
concretizeMoveRRAw
	"C6.2.272 	STP		C6-1236"
	<inline: true>
	| pairAddress reg1 reg2 offset |
	reg1 := operands at: 0.
	reg2 := operands at: 1.
	pairAddress := operands at: 2.
	self assert: reg1 ~= reg2.
	self assert: (reg1 ~= RISCTempReg and: [reg2 ~= RISCTempReg]).
	self assert: (self isAddressRelativeToVarBase: pairAddress).
	self assert: pairAddress - cogit varBaseAddress / 8 < (1 << 6).
	reg1 = SP ifTrue: [reg1 := RISCTempReg].
	reg2 = SP ifTrue: [reg2 := RISCTempReg].
	(reg1 = RISCTempReg or: [reg2 = RISCTempReg])
		ifTrue:
			[machineCode
				at: 0
				put: (self movern: SP rd: RISCTempReg).
			 offset := 1]
		ifFalse:
			[offset := 0].
	machineCode
		at: offset
		put: 2r1010100100 << 22
			+ (pairAddress - cogit varBaseAddress / 8 << 15)
			+ (reg2 << 10)
			+ (VarBaseReg << 5)
			+ reg1.
	"cogit processor
		disassembleInstructionAt: 0 In: machineCode object;
		disassembleInstructionAt: 4 In: machineCode object"
	^offset + 1 * 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/21/2019 09:57'!
concretizeMoveRRd
	<inline: true>
	"C7.2.124	FMOV (general)		C7-1651"
	machineCode
		at: 0
		put: 2r1001111001100111 << 16
			+ ((operands at: 0) << 5)
			+ (operands at: 1).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/21/2019 09:24'!
concretizeMoveRXSrR: unitSizeLog2MinusOne
	"Xwr/X32r/X16r/Xbr - memory unit whose address is r * unit size away from an address in a register"
	"C6.2.274	STR (register)	C6-1242"
	| src index base |
	src := operands at: 0.
	index := operands at: 1.
	base := operands at: 2.
	self deny: (SP = src or: [SP = index]).
	machineCode
		at: 0
		put: unitSizeLog2MinusOne << 30
			+ (2r111000001 << 21)
			+ (index << 16)
			+ (UXTX << 13)
			+ (3 << 11)
			+ (base << 5)
			+ src.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/21/2019 09:13'!
concretizeMoveRXbrR
	<inline: true>
	"Xbr - byte in mwemory whose address is r away from an address in a register"
	"C6.2.274	STR (register)	C6-1242"
	| src index base |
	src := operands at: 0.
	index := operands at: 1.
	base := operands at: 2.
	self deny: (SP = src or: [SP = index]).
	machineCode
		at: 0
		put: 2r00111000001 << 21
			+ (index << 16)
			+ (UXTX << 13)
			+ (3 << 11)
			+ (base << 5)
			+ src.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/28/2019 22:48'!
concretizeMoveRdM64r
	"Load/Store Vector on page C3-206.
	 C7.2.323	STR (immediate, SIMD&FP)	C7-2114"
	<inline: true>
	| rd offset baseReg |
	rd := operands at: 0.
	offset := operands at: 1.
	baseReg := operands at: 2.
	self assert: (offset noMask: 7).
	machineCode
		at: 0
		put: 2r11111101 << 24
			+ (offset << 7 "10 - 3")
			+ (baseReg << 5)
			+ rd.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/21/2019 09:57'!
concretizeMoveRdR
	<inline: true>
	"C7.2.124	FMOV (general)		C7-1651"
	machineCode
		at: 0
		put: 2r1001111001100110 << 16
			+ ((operands at: 0) << 5)
			+ (operands at: 1).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/21/2019 09:25'!
concretizeMoveXSrRR: unitSizeLog2MinusOne
	"Xwr/X32r/X16r/Xbr - memory unit whose address is r * unit size away from an address in a register"
	"C6.2.132	LDR (register)	C6-981"
	| index base dest |
	index := operands at: 0.
	base := operands at: 1.
	dest := operands at: 2.
	self deny: SP = dest.
	machineCode
		at: 0
		put: unitSizeLog2MinusOne << 30
			+ (2r111000011 << 21)
			+ (index << 16)
			+ (UXTX << 13)
			+ (3 << 11)
			+ (base << 5)
			+ dest.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 7/2/2021 07:11'!
concretizeMulOverflowJump
	"Sizing/generating jumps.
		Jump targets can be to absolute addresses or other abstract instructions.
		Generating initial trampolines instructions may have no maxSize and be to absolute addresses.
		Otherwise instructions must have a machineCodeSize which must be kept to."
	<inline: false>
	| offset |
	offset := self computeJumpTargetOffset - 4. "-4 because the jump is from the second word..."
 	self assert: (offset ~= 0 and: [self isInImmediateBranchRange: offset]).
	"See concretizeMulOverflowRRR
	 RISCTempReg := RISCTempReg + CArg1Reg/sign.
	 JumpZero/NonZero"
	machineCode
		at: 0
		put: 2r10001011 << 24
			+ (ArithmeticAddS << 29)
			+ (CArg1Reg << 16)
			+ (RISCTempReg << 5)
			+ RISCTempReg;
		at: 1
		put: (self
				cond: (opcode = JumpMulOverflow ifTrue: [NE] ifFalse: [EQ])
				offset: offset). "B offset"
	^8! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 7/2/2021 07:12'!
concretizeMulOverflowRRR
	"Will get inlined into concretizeAt: switch."
	"ARMv8 has no multiply overflow detection.  Instead it is synthesized from the two halves of
	 a 64x64=>128 bit multiply. The upper 64-bits are tested.  The sequence is
		high64 := SMULH a,b
		a/low64 := MUL a,b
		signBit := low64/a >> 63
		high64 := high64 + signBit
	 If high64 is zero after this sequence then the multiply has not overflowed,
	 since high64 is an extension of signBit if no overflow (either 0 or -1) and
	 both -1 + 1 = 0 and 0 + 0 = 0. Note that because we restrict ourselves to
	 three concrete ARMv8 instructions per abstract instruction the last operation
	 of the sequence is generated in concretizeMulOverflowJump.

	 C6.2.196	MUL				C6-1111
	 C6.2.242	SMULH				C6-1184
	 C6.2.180	LSR (immediate)	C6-1081	110100110 (1)"

	<inline: true>
	| reg1 reg2 reg3 |
	reg1 := operands at: 0.
	reg2 := operands at: 1.
	reg3 := operands at: 2.
	"RISCTempReg := high(reg1 * reg2); must precede destructive MUL"
	machineCode
		at: 0
		put: 2r1001101101 << 22
			+ (reg1 << 16)
			+ (XZR << 10)
			+ (reg2 << 5)
			+ RISCTempReg.
	"reg3 := reg1 * reg2"
	machineCode
		at: 1
		put: 2r10011011 << 24
			+ (reg1 << 16)
			+ (XZR << 10)
			+ (reg2 << 5)
			+ reg3.
	"CArg1Reg := sign(reg3)"
	machineCode
		at: 2
		put: 2r1101001101 << 22
			+ (63 << 16) "constant to shift by"
			+ (63 << 10)
			+ (reg3 << 5)
			+ CArg1Reg. "cuz CArg0Reg == TempReg"
	"RISCTempReg := RISCTempReg + CArg1Reg/sign
	 is in concretizeMulOverflowJump"
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	"cogit processor disassembleInstructionAt: 4 In: machineCode object"
	"cogit processor disassembleInstructionAt: 8 In: machineCode object"
	^12! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/21/2019 23:56'!
concretizeMulRRR
	"C6.2.196	MUL	C6-1111"
	<inline: true>
	"reg3 := reg1 * reg2"
	| reg1 reg2 reg3 |
	reg1 := operands at: 0.
	reg2 := operands at: 1.
	reg3 := operands at: 2.
	machineCode
		at: 0
		put: 2r10011011 << 24
			+ (reg1 << 16)
			+ (XZR << 10)
			+ (reg2 << 5)
			+ reg3.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 2/25/2020 12:11'!
concretizeNativePopRR
	"C6.2.128 	LDP		C6-970"
	<inline: true>
	| reg1 reg2 |
	reg1 := operands at: 0.
	reg2 := operands at: 1.
	"Post-index"
	machineCode
		at: 0
		put: 2r10101000110000010 << 15
			+ (reg2 << 10)
			+ (SP << 5)
			+ reg1.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 2/25/2020 12:11'!
concretizeNativePushRR
	"C6.2.272 	STP		C6-1236"
	<inline: true>
	| reg1 reg2 |
	reg1 := operands at: 0.
	reg2 := operands at: 1.
	"Pre-index"
	machineCode
		at: 0
		put: 2r10101001101111110 << 15
			+ (reg2 << 10)
			+ (SP << 5)
			+ reg1.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 8/1/2021 23:19'!
concretizeNativeRetN
	"Will get inlined into concretizeAt: switch."
	<var: #offset type: #sqInt>
	<inline: true>
	| offset |
	offset := operands at: 0.
	"C6.2.218 RET p1147"
	offset = 0 ifTrue:
		[machineCode
			at: 0
			put: 2r1101011001 << 22
				+ (XZR << 16)
				+ (LR << 5).
		^4].

	"C6.2.4 ADD (immediate) p761"
	machineCode
		at: 0 put: (self addrn: SP rd: SP imm: offset shiftBy12: false);
		at: 1 put: 2r1101011001 << 22
				+ (XZR << 16)
				+ (LR << 5).
	^8! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/18/2019 18:55'!
concretizeNegateR
	<inline: true>
	"C6.2.309	SUB (shifted register) 1313
	 C6.2.313	SUBS (extended register)	C6-1318"
	| reg |
	reg := operands at: 0.
	self deny: SP = reg.
	machineCode
		at: 0
		put: 2r11001011 << 24
			+ (reg << 16)
			+ (XZR << 5)
			+ reg.
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/28/2019 22:13'!
concretizeNop
	<inline: true>
	machineCode at: 0 put: NOP.
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/31/2020 17:19'!
concretizePopR
	<inline: true>
	"C6.2.130	LDR (immediate)	C6-976
	 Post-index"
	| reg |
	reg := operands at: 0.
	self deny: SP = reg.
	machineCode
		at: 0
		put: 2r1111100001 << 22
			+ (8 << 12)
			+ (2r01 << 10)
			+ (SPReg << 5)
			+ reg.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 7/24/2021 19:28'!
concretizePrefetchAw
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcAddr |
	srcAddr := operands at: 0.
	(self isAddressRelativeToVarBase: srcAddr) ifTrue:
		[srcAddr < cogit varBaseAddress ifTrue:
			[self shouldBeImplemented.
			 ^4].
		 machineCode
			at: 0
			put: (self prn: VarBaseReg imm: srcAddr - cogit varBaseAddress shiftBy12: false).
		 ^4].
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	self halt.
	^0! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/31/2020 17:20'!
concretizePushCq
	"ARMv8 has no push costant insructions (no store immediate pre-index).
	 Instead load the constant into RISCTempReg and push that."
	<inline: true>
	| instrBytes |
	operands at: 1 put: RISCTempReg.
	instrBytes := self concretizeMoveCqR.
	"C6.2.273	STR (immediate)	Pre-index	C6-1239"
	machineCode
		at: instrBytes // 4
		put: 2r11111 << 27
			+ ((-8 bitAnd: 1 << 9 - 1) << 12)
			+ (2r11 << 10)
			+ (SPReg << 5)
			+ RISCTempReg.
	^instrBytes + 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/31/2020 17:20'!
concretizePushCw
	<inline: true>
	| instrBytes |
	instrBytes := self emitMoveCw: (operands at: 0) intoR: RISCTempReg at: 0.
	"C6.2.273	STR (immediate)	Pre-index	C6-1239"
	machineCode
		at: instrBytes // 4
		put: 2r11111 << 27
			+ ((-8 bitAnd: 1 << 9 - 1) << 12)
			+ (2r11 << 10)
			+ (SPReg << 5)
			+ RISCTempReg.
	^instrBytes + 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 2/1/2020 18:10'!
concretizePushR
	<inline: true>
	"C6.2.273	STR (immediate)	C6-1239
	 Pre-index"
	| reg |
	reg := operands at: 0.
	self deny: SP = reg.
	machineCode
		at: 0
		put: 2r11111 << 27
			+ ((-8 bitAnd: 1 << 9 - 1) << 12)
			+ (2r11 << 10)
			+ (SPReg << 5)
			+ reg.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/9/2020 16:18'!
concretizeRRArithmetic: arithOp Rd: rd
	"rd := regA op regB "
	| regA regB |
	regB := operands at: 0.
	regA := operands at: 1.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	"cogit processor disassembleInstructionAt: 4 In: machineCode object"
	SP = regB ifTrue:	"Arithmetic with the sp; we must use the extended register forms, and negate
						 on subtract because we can't have Rm as SP, Rm = 31 is interoreted as XZR."
		["ADD (extended register) on page C6-758
		  ADDS (extended register) on page C6-766
		  SUB (extended register) on page C6-1308
		  SUBS (extended register) on page C6-1318"
		 machineCode
			at: 0
			put: 2r10001011001 << 21
				+ (arithOp << 29)
				+ (regA << 16)
				+ (UXTX << 13)
				+ (regB << 5)
				+ rd.
		 (arithOp = ArithmeticSub or: [arithOp = ArithmeticSubS]) ifFalse:
			[^4].
		 machineCode
			at: 1
			put: 2r10001011 << 24
				+ (arithOp << 29)
				+ (rd << 16)
				+ (XZR << 5)
				+ rd.
		 ^8].
	SP = regA ifTrue: "Arithmetic with the sp; we must use the extended register forms."
		["ADD (extended register) on page C6-758
		  ADDS (extended register) on page C6-766
		  SUB (extended register) on page C6-1308
		  SUBS (extended register) on page C6-1318
		  CMN (extended register) on page C6-850
		  CMP (extended register) on page C6-856"
		 machineCode
			at: 0
			put: 2r10001011001 << 21
					+ (arithOp << 29)
					+ (regB << 16)
					+ (UXTX << 13)
					+ (regA << 5)
					+ rd.
		 rd ~= XZR ifTrue: [self halt].
		 ^4].
	self assert: (regA ~= SP and: [regB ~= SP]).
	"ADD (shifted register) on page C6-763
	 ADDS (shifted register) on page C6-771
	 CMN (shifted register) on page C6-854
	 CMP (shifted register) on page C6-860
	 SUB (shifted register) on page C6-1313
	 SUBS (shifted register) on page C6-1323"
	 machineCode
		at: 0
		put: 2r10001011 << 24
			+ (arithOp << 29)
			+ (regB << 16)
			+ (regA << 5)
			+ rd.
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/20/2019 18:26'!
concretizeRRLogical: logicalOp
	"AND (shifted register) - 64-bit variant on page C6-777
	 BIC (shifted register) - 64-bit variant on page C6-808
	 ORR (shifted register) - 64-bit variant on page C6-1127
	 ORN (shifted register) - 64-bit variant on page C6-1123
	 EOR (shifted register) - 64-bit variant on page C6-898
	 EON (shifted register) - 64-bit variant on page C6-894
	 ANDS (shifted register) - 64-bit variant on page C6-781
	 BICS (shifted register) - 64-bit variant on page C6-810"
	| destReg srcReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0
		put: 2r1000101 << 25
			+ (logicalOp << 29)
			+ (srcReg << 16)
			+ (destReg << 5)
			+ destReg.
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/31/2020 17:20'!
concretizeRetN
	"Will get inlined into concretizeAt: switch."
	<var: #offset type: #sqInt>
	<inline: true>
	| offset |
	offset := operands at: 0.
	"C6.2.218 RET p1147"
	offset = 0 ifTrue:
		[machineCode
			at: 0
			put: 2r1101011001 << 22
				+ (XZR << 16)
				+ (LR << 5).
		^4].

	"C6.2.4 ADD (immediate) p761"
	machineCode
		at: 0 put: (self addrn: SPReg rd: SPReg imm: offset shiftBy12: false);
		at: 1 put: 2r1101011001 << 22
				+ (XZR << 16)
				+ (LR << 5).
	^8! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/6/2020 16:01'!
concretizeRotateCqR
	"C6.2.225	ROR (immediate)	C6-1157"
	<inline: true>
	| constant reg |
	constant := operands at: 0.
	reg := operands at: 1.
	self assert: (constant between: 1 and: 63).
	machineCode
		at: 0
		put: 2r1001001111 << 22
			+ (reg << 16)
			+ ((RotateRightCqR = opcode ifTrue: [constant] ifFalse: [64 - constant]) << 10)
			"+ (1 << 10)"
			+ (reg << 5)
			+ reg.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 10/27/2020 16:52'!
concretizeSTLR
	| valueReg "Xt" baseReg "Xn" |
	valueReg := operands at: 0.
	baseReg := operands at: 1.
	machineCode
		at: 0
		put: 2r1100100010011111111111 << 10
			+ (baseReg << 5)
			+ valueReg.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object '00000000: stlr	x0, [x1]'"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 11/12/2020 16:27'!
concretizeSTLXR
	"C6.2.268		STLXR		C6-1228"
	<inline: true>
	| valueReg addressReg statusReg |
	valueReg := operands at: 0.
	addressReg := operands at: 1.
	statusReg := operands at: 2.
	machineCode
		at: 0
		put: 2r11001 << 27
			+ (statusReg << 16)
			+ (63 << 10)
			+ (addressReg << 5)
			+ valueReg.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/9/2020 18:36'!
concretizeShiftRR: shiftOp
	"C6.2.18 	ASRV	C6-787		(a.k.a	C6.2.16 	ASR (register)	
	 C6.2.178	LSLV	C6-1077	(a.k.a	C6.2.176	LSL (register)	
	 C6.2.181	LSRV	C6-1083	(a.k.a	C6.2.179	LSR (register)"
	<inline: true>
	| shiftedReg shiftCountReg |
	shiftCountReg := operands at: 0.
	shiftedReg := operands at: 1.
	machineCode
		at: 0
		put: 2r10011010110 << 21
			+ (shiftCountReg << 16)
			+ (shiftOp << 10)
			+ (shiftedReg << 5)
			+ shiftedReg.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/23/2020 19:50'!
concretizeSignExtendRR: width
	"C6.2.320 	SXTB	C6-1332
	 C6.2.321 	SXTH	C6-1334
	 C6.2.322 	SXTW	C6-1336"
	| rn rd |
	rn := operands at: 0.
	rd := operands at: 1.
	machineCode
		at: 0
		put: 2r1001001101 << 22
			+ (width - 1 << 10)
			+ (rn << 5)
			+ rd.
	^4

	"cogit processor disassembleInstructionAt: 0 In: machineCode object"! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/28/2019 22:14'!
concretizeSqrtRd
	"C7.2.165	FSQRT (scalar)	C7-1743"
	<inline: true>
	| reg |
	reg := operands at: 0.
	machineCode
		at: 0
		put: 2r111100110000111 << 14
			+ (reg << 5)
			+ reg.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/23/2020 19:59'!
concretizeStop
	<inline: true>
	machineCode at: 0 put: self stop.
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/27/2019 01:13'!
concretizeSubCqR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| constant reg offset |
	<var: #constant type: #sqInt>
	constant := (operands at: 0) signedIntFromLong64.
	reg := operands at: 1.

	self isPossiblyShiftableImm12: constant
		ifTrue:
			[:shift|
			"C6.2.308		SUB (immediate)	C6-1311
			 C6.2.314		SUBS (immediate)	C6-1321"
			machineCode
				at: 0
				put: (reg = SP
						ifTrue:  [2r110100010]
						ifFalse: [2r111100010]) << 23
					+ (shift ifTrue: [constant >> 2 + (1 << 22)] ifFalse: [constant << 10])
					+ (reg << 5)
					+ reg.
			^4]
		ifFalse: [].
	self isPossiblyShiftableImm12: constant negated
		ifTrue:
			[:shift|
			"C6.2.4		ADDS (immediate)	C6-761
			 C6.2.8		ADDS (immediate)	C6-769"
			machineCode
				at: 0
				put: (reg = SP
						ifTrue:  [2r100100010]
						ifFalse: [2r101100010]) << 23
					+ (shift ifTrue: [constant negated >> 2 + (1 << 22)] ifFalse: [constant negated << 10])
					+ (reg << 5)
					+ reg.
			^4]
		ifFalse: [].
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	"cogit processor disassembleInstructionAt: 4 In: machineCode object"
	offset := self emitMoveCw: constant intoR: RISCTempReg at: 0.
	self deny: SP = reg.
	"C6.2.313	SUBS (extended register)		C6-1318"
	machineCode
				at: offset // 4
				put: 2r10001011 << 24
				+ (ArithmeticSubS << 29)
				+ (RISCTempReg << 16)
				+ (reg << 5)
				+ reg.
	^offset + 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/28/2019 23:55'!
concretizeXorRdRd
	"C7.2.34	EOR (vector)	C7-1440"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0
		put: 2r101110001 << 21
			+ (srcReg << 16)
			+ (2r111 << 10)
			+ (destReg << 5)
			+ destReg.
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/23/2020 20:15'!
concretizeZeroExtendRR: width
	"C6.2.333	UBFX	C6-1353
	 C6.2.341 	UXTB	C6-1364
	 C6.2.342 	UXTH	C6-1366"
	| rn rd |
	rn := operands at: 0.
	rd := operands at: 1.
	machineCode
		at: 0
		put: 2r1101001101 << 22
			+ (width - 1 << 10)
			+ (rn << 5)
			+ rd.
	^4

	"cogit processor disassembleInstructionAt: 0 In: machineCode object"! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 1/29/2020 18:44'!
cond: cond offset: offset
	"Conditional branch (immediate) C4-257"
	<inline: true>
	self assert: (offset bitAnd: 3) = 0.
	^2r010101 << 26
	+ ((offset bitAnd: 1 << 21 - 1) << 3) "(offset / 4 bitAnd: 1 << 19 - 1) << 5"
	+ cond! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 1/29/2020 13:33'!
countLeadingOnes: anInteger
	| count theInteger |
	anInteger >= 0 ifTrue: [^0].
	theInteger := anInteger.
	count := 0.
	5 to: 0 by: -1 do:
		[:logBase2Shift| | shift |
		shift := 1 << logBase2Shift.
		theInteger >>> shift = -1
			ifTrue: [count := count + shift]
			ifFalse: [theInteger := theInteger >>> shift]].
	^theInteger = -1
		ifTrue: [count + 1]
		ifFalse: [count]
	

	"self basicNew countLeadingOnes: -16
	((-16 bitAnd: 1 << 64 - 1) binary copyUpTo: $0) occurrencesOf: $1

	self basicNew countLeadingOnes: -1
	((-1 bitAnd: 1 << 64 - 1) binary copyUpTo: $0) occurrencesOf: $1

	(-1 to: -64 by: -1)
		select: [:n| (self basicNew countLeadingOnes: n) ~= (((n bitAnd: 1 << 64 - 1) binary copyUpTo: $0) occurrencesOf: $1)]
		thenCollect: [:n| {self basicNew countLeadingOnes: n. (n bitAnd: 1 << 64 - 1) binary}]"! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 1/23/2020 14:22'!
countTrailingOnes: anInteger
	| bits count |
	bits := anInteger.
	count := 0.
	[bits allMask: 16rFFFF] whileTrue:
		[bits := bits bitShift: -16.
		 count := count + 16].
	[bits allMask: 16rFF] whileTrue:
		[bits := bits bitShift: -8.
		 count := count + 8].
	(bits allMask: 16rF) ifTrue:
		[bits := bits bitShift: -4.
		 count := count + 4].
	(bits allMask: 16r3) ifTrue:
		[bits := bits bitShift: -2.
		 count := count + 2].
	 ^(bits allMask: 1) ifTrue: [count + 1] ifFalse: [count]

	"(-16 to: 16) collect: [:i| {i. self basicNew countTrailingOnes: (i bitAnd: 1 << 64 - 1)}]"! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 1/29/2020 16:55'!
countTrailingZeros: anInteger
	"a.k.a. anInteger lowBit - 1"
	| n result |
	self assert: anInteger ~= 0.
	n := anInteger.
	result := 0.
	[n noMask: 16rFF]
		whileTrue:
			[result := result + 8.
			n := n bitShift: -8].
	(n noMask: 16rF) ifTrue:
		[result := result + 4.
		 n := n bitShift: -4].
	(n noMask: 16r3) ifTrue:
		[result := result + 2.
		 n := n bitShift: -2].
	^(n anyMask: 1)
		ifTrue: [result]
		ifFalse: [result + 1]

	"| me |
	me := self basicNew.
	(1 to: 65535) reject: [:i| i lowBit - 1 = (me countTrailingZeros: i)]"! !

!CogARMv8Compiler methodsFor: 'feature detection' stamp: 'eem 7/7/2021 18:02'!
dataCacheFlushRequired
	<cmacro: '(ign) dataCacheFlushRequired'>
	"For want of somewhere to put the variable that doesn't bloat an instance of the receiver..."
	^DataCacheFlushRequired! !

!CogARMv8Compiler methodsFor: 'feature detection' stamp: 'eem 7/7/2021 18:02'!
dataCacheLineLength
	<cmacro: '(ign) dataCacheLineLength'>
	"For want of somewhere to put the variable that doesn't bloat an instance of the receiver..."
	^DataCacheLineLength! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 1/29/2020 18:32'!
decode64Imms: imms immr: immr
	"See aarch64/instrs/integer/bitmasks/DecodeBitMasks J1-7389.
	 This is a 64-bit version computing the imm mask (wmask) only."
	<returnTypeC: #usqInt>
	| mask |
	<var: 'mask' type: #usqInt>
	self assert: ((imms between: 0 and: 63) and: [immr between: 0 and: 63]).
	"For logical immediates an all-ones value of S is reserved since it would generate a useless all-ones result (many times)"
	imms = 63 ifTrue:
		[^self cCode: [0] inSmalltalk: [#undefined]].

	mask := 1 << (imms + 1) - 1.
	^immr = 0
		ifTrue: [mask]
		ifFalse: [(mask << (64 - immr) bitAnd: 1 << 64 - 1) + (mask >> immr)]! !

!CogARMv8Compiler methodsFor: 'feature detection' stamp: 'eem 7/7/2021 08:28'!
detectFeatures
	<inline: #always>
	self cppIf: #__APPLE__
		ifTrue: [self detectFeaturesOnMacOS]
		ifFalse: [self detectFeaturesOnLinux]! !

!CogARMv8Compiler methodsFor: 'feature detection' stamp: 'eem 7/9/2021 10:35'!
detectFeaturesOnLinux
	"Do throw-away compilations to read CTR_EL0 & ID_AA64ISAR0_EL1 and initialize ctrEl0 & idISAR0"
	<notOption: #__APPLE__>
	| startAddress getFeatureReg ctrEL0 idISAR0 |
	<var: 'getFeatureReg' declareC: 'usqIntptr_t (*getFeatureReg)(void)'>
	startAddress := cogit methodZoneBase.
	cogit allocateOpcodes: 4 bytecodes: 0.
	getFeatureReg := cogit cCoerceSimple: startAddress to: #'usqIntptr_t (*)(void)'.
	"Return the value of CTR_EL0; that's the control register that defines the vital statistics of the processor's caches."
	cogit
		gen: Nop; "do something anodyne so it is easy to distinguish MRS_CTR_EL0 being an illegal instruction rather than the code zone not being executable."
		gen: MRS_CTR_EL0 operand: ABIResultReg;
		RetN: 0.
	cogit outputInstructionsForGeneratedRuntimeAt: startAddress.
	cogit resetMethodZoneBase: startAddress.
	cogit ensureExecutableCodeZoneWithin:
		[ctrEL0 := (self cCode: 'getFeatureReg()' inSmalltalk: [cogit simulateLeafCallOf: startAddress]).
		 "see e.g. CogARMv8Compiler class>>printCTR_EL0:, concretizeCacheControlOp1:CRm:Op2: & 
		 http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.100403_0200_00_en/lau1443435580346.html
		 DminLine & IminLine are Log2 words; 16 words miniumum"
		 self setDataCacheFlushRequired: (ctrEL0 noMask: 1 << 28).
		 self setDataCacheLineLength: 4 << (ctrEL0 >> 16 bitAnd: 15).
		 self dataCacheLineLength = 0 ifTrue:
			[self setDataCacheLineLength: 4 << 4].
		 self setInstructionCacheFlushRequired: (ctrEL0 noMask: 1 << 29).
		 self setInstructionCacheLineLength: 4 << (ctrEL0 bitAnd: 15)].
		 self instructionCacheLineLength = 0 ifTrue:
			[self setInstructionCacheLineLength: 4 << 4].
	cogit zeroOpcodeIndexForNewOpcodes.
	cogit
		gen: Nop; "do something anodyne so it is easy to distinguish MRS_CTR_EL0 being an illegal instruction rather than the code zone not being executable."
		gen: MRS_ID_AA64ISAR0_EL1 operand: ABIResultReg;
		RetN: 0.
	cogit outputInstructionsForGeneratedRuntimeAt: startAddress.
	cogit resetMethodZoneBase: startAddress.
	cogit ensureExecutableCodeZoneWithin:
		[idISAR0 := (self cCode: 'getFeatureReg()' inSmalltalk: [cogit simulateLeafCallOf: startAddress]).
		 self setHasAtomicInstructions: (idISAR0 >> 20 bitAnd: 2r1111) = 2r10]! !

!CogARMv8Compiler methodsFor: 'memory access' stamp: 'eem 7/9/2021 20:31'!
detectFeaturesOnMacOS
	<option: #__APPLE__>
	"MacOS does not allow access to ctl_el0, so derive cache information etc from sysctl"
	"Here are values from sysctl(8), hardwired for now rather than derived through sysctl(3)
		hw.cacheconfig: 8 1 1 0 0 0 0 0 0 0 (we speculate that the 1's indicate cache flush required)
		hw.cachelinesize: 128
		hw.l1icachesize: 131072
		hw.l1dcachesize: 131072
		hw.optional.neon: 1
		hw.optional.neon_hpfp: 1
		hw.optional.neon_fp16: 1
		hw.optional.armv8_1_atomics: 1"

	false ifTrue: "Apple's cache flush code on M1 is simple and effective; we can't easily better it..."
		[self setDataCacheLineLength: 128.
		self setDataCacheFlushRequired: true.
		self setInstructionCacheLineLength: 128.
		self setInstructionCacheFlushRequired: true].
	self setHasAtomicInstructions: true! !

!CogARMv8Compiler methodsFor: 'generate machine code' stamp: 'eem 8/5/2021 07:15'!
dispatchConcretize
	"Attempt to generate concrete machine code for the instruction at address.
	 This is the inner dispatch of concretizeAt: actualAddress which exists only
	 to get around the branch size limits in the SqueakV3 (blue book derived)
	 bytecode set."
		 
	opcode caseOf: {
		"Noops & Pseudo Ops"
		[Label]					-> [^self concretizeLabel].
		[Literal]					-> [^self concretizeLiteral].
		[AlignmentNops]		-> [^self concretizeAlignmentNops].
		[Fill32]					-> [^self concretizeFill32].
		[Nop]					-> [^self concretizeNop].
		"Control"
		[Call]						-> [^self concretizeCall]. "call code within code space"
		[CallFull]					-> [^self concretizeCallJumpFull: true]. "call code anywhere in address space"
		[JumpR]					-> [^self concretizeJumpR].
		[JumpFull]					-> [^self concretizeCallJumpFull: false]."jump within address space"
		[JumpLong]				-> [^self concretizeJumpLong]."jump within code space"
		[JumpLongZero]			-> [^self concretizeConditionalLongJump: EQ].
		[JumpLongNonZero]		-> [^self concretizeConditionalLongJump: NE].
		[Jump]						-> [^self concretizeConditionalJump: AL]. "jump within a method, etc"
		[JumpZero]					-> [^self concretizeConditionalJump: EQ].
		[JumpNonZero]				-> [^self concretizeConditionalJump: NE].
		[JumpNegative]				-> [^self concretizeConditionalJump: MI].
		[JumpNonNegative]			-> [^self concretizeConditionalJump: PL].
		[JumpOverflow]				-> [^self concretizeConditionalJump: VS].
		[JumpNoOverflow]			-> [^self concretizeConditionalJump: VC].
		[JumpMulOverflow]			-> [^self concretizeMulOverflowJump].
		[JumpNoMulOverflow]		-> [^self concretizeMulOverflowJump].
		[JumpCarry]				-> [^self concretizeConditionalJump: CS].
		[JumpNoCarry]				-> [^self concretizeConditionalJump: CC].
		[JumpLess]					-> [^self concretizeConditionalJump: LT].
		[JumpGreaterOrEqual]		-> [^self concretizeConditionalJump: GE].
		[JumpGreater]				-> [^self concretizeConditionalJump: GT].
		[JumpLessOrEqual]			-> [^self concretizeConditionalJump: LE].
		[JumpBelow]				-> [^self concretizeConditionalJump: CC]. "unsigned lower"
		[JumpAboveOrEqual]		-> [^self concretizeConditionalJump: CS]. "unsigned greater or equal"
		[JumpAbove]				-> [^self concretizeConditionalJump: HI].
		[JumpBelowOrEqual]		-> [^self concretizeConditionalJump: LS].
		[JumpFPEqual]				-> [^self concretizeConditionalJump: EQ].
		[JumpFPNotEqual]			-> [^self concretizeConditionalJump: NE].
		[JumpFPLess]				-> [^self concretizeConditionalJump: LT].
		[JumpFPGreaterOrEqual]	-> [^self concretizeConditionalJump: GE].
		[JumpFPGreater]			-> [^self concretizeConditionalJump: GT].
		[JumpFPLessOrEqual]		-> [^self concretizeConditionalJump: LE].
		[JumpFPOrdered]			-> [^self concretizeConditionalJump: VC].
		[JumpFPUnordered]		-> [^self concretizeConditionalJump: VS].
		[RetN]						-> [^self concretizeRetN].
		[NativeRetN]				-> [^self concretizeNativeRetN].
		[Stop]						-> [^self concretizeStop].
		"Arithmetic"
		[AddCqR]						-> [^self concretizeAddCqRDest: (operands at: 1)].
		[AddCqRR]						-> [^self concretizeAddCqRDest: (operands at: 2)].
		[AndCqR]						-> [^self concretizeLogicalOp: LogicalAndS CqRDest: (operands at: 1)].
		[AndCqRR]						-> [^self concretizeLogicalOp: LogicalAndS CqRDest: (operands at: 2)].
		[OrCqR]						-> [^self concretizeLogicalOp: LogicalOr CqRDest: (operands at: 1)].
		[OrCqRR]						-> [^self concretizeLogicalOp: LogicalOr CqRDest: (operands at: 2)].
		[CmpCqR]						-> [^self concretizeCmpCqR].
		[SubCqR]						-> [^self concretizeSubCqR].
		[TstCqR]						-> [^self concretizeLogicalOp: LogicalAndS CqRDest: XZR].
		[XorCqR]						-> [^self concretizeLogicalOp: LogicalXor CqRDest: (operands at: 1)].
		[AddCwR]						-> [^self concretizeCwRArithmetic: ArithmeticAddS Rd: (operands at: 1)].
		[AndCwR]						-> [^self concretizeCwRLogical: LogicalAndS].
		[CmpCwR]						-> [^self concretizeCwRArithmetic: ArithmeticSubS Rd: XZR].
		[CmpC32R]						-> [^self concretizeCmpC32R].
		[OrCwR]						-> [^self concretizeCwRLogical: LogicalOr].
		[SubCwR]						-> [^self concretizeCwRArithmetic: ArithmeticSubS Rd: (operands at: 1)].
		[XorCwR]						-> [^self concretizeCwRLogical: LogicalXor].
		[AddRR]						-> [^self concretizeRRArithmetic: ArithmeticAddS Rd: (operands at: 1)].
		[AndRR]						-> [^self concretizeRRLogical: LogicalAndS].
		[CmpRR]						-> [^self concretizeRRArithmetic: ArithmeticSubS Rd: XZR].
		[OrRR]							-> [^self concretizeRRLogical: LogicalOr].
		[SubRR]						-> [^self concretizeRRArithmetic: ArithmeticSubS Rd: (operands at: 1)].
		[XorRR]							-> [^self concretizeRRLogical: LogicalXor].
		[AddRRR]						-> [^self concretizeRRArithmetic: ArithmeticAddS Rd: (operands at: 2)].
		[SubRRR]						-> [^self concretizeRRArithmetic: ArithmeticSubS Rd: (operands at: 2)].
		[AddRdRd]						-> [^self concretizeMathRdRd: 2r1010].
		[CmpRdRd]						-> [^self concretizeCmpRdRd].
		[DivRdRd]						-> [^self concretizeMathRdRd: 2r0110].
		[MulRdRd]						-> [^self concretizeMathRdRd: 2r0010].
		[SubRdRd]						-> [^self concretizeMathRdRd: 2r1110].
		[XorRdRd]						-> [^self concretizeXorRdRd].
		[SqrtRd]						-> [^self concretizeSqrtRd].
		[NegateR]						-> [^self concretizeNegateR].
		[LoadEffectiveAddressMwrR]	-> [^self concretizeLoadEffectiveAddressMwrR].
		[ArithmeticShiftRightCqR]		-> [^self concretizeArithmeticShiftRightCqR].
		[ArithmeticShiftRightCqRR]		-> [^self concretizeArithmeticShiftRightCqRR].
		[LogicalShiftRightCqR]			-> [^self concretizeLogicalShiftRightCqR].
		[LogicalShiftRightCqRR]		-> [^self concretizeLogicalShiftRightCqRR].
		[LogicalShiftLeftCqR]			-> [^self concretizeLogicalShiftLeftCqR].
		[LogicalShiftLeftCqRR]			-> [^self concretizeLogicalShiftLeftCqRR].
		[ArithmeticShiftRightRR]		-> [^self concretizeShiftRR: 2r001010].
		[LogicalShiftLeftRR]				-> [^self concretizeShiftRR: 2r001000].
		[LogicalShiftRightRR]			-> [^self concretizeShiftRR: 2r001001].
		[RotateRightCqR]				-> [^self concretizeRotateCqR].
		[RotateLeftCqR]				-> [^self concretizeRotateCqR].
		[ClzRR]							-> [^self concretizeClzRR].
		"ARM Specific Arithmetic"
		[MulRRR]						-> [^self concretizeMulRRR].
		[MulOverflowRRR]				-> [^self concretizeMulOverflowRRR].
		[DivRRR]						-> [^self concretizeDivRRR].
		[MSubRRR]						-> [^self concretizeMSubRRR].
		"ARM Specific Cache Control"
		[DC]							-> [^self concretizeDataCacheControl].
		[DSB]							-> [^self concretizeDataSynchronizationBarrier].
		[IC]								-> [^self concretizeInstructionCacheControl].
		[ISB]							-> [^self concretizeInstructionSynchronizationBarrier].
		[MRS_CTR_EL0]					-> [^self concretizeMRS_CTR_EL0].
		[MRS_ID_AA64ISAR0_EL1]		-> [^self concretizeMRS_ID_AA64ISAR0_EL1].
		"Data Movement"
		[MoveCqR]			-> [^self concretizeMoveCqR].
		[MoveCwR]			-> [^self concretizeMoveCwR].
		[MoveC32R]		-> [^self concretizeMoveC32R].
		[MoveRR]			-> [^self concretizeMoveRR].
		[MoveRRd]			-> [^self concretizeMoveRRd].
		[MoveRdR]			-> [^self concretizeMoveRdR].
		[MoveAwR]			-> [^self concretizeMoveAwR].
		[MoveRAw]			-> [^self concretizeMoveRAw].
		[MoveAwRR]		-> [^self concretizeMoveAwRR].
		[MoveRRAw]		-> [^self concretizeMoveRRAw].
		[MoveAbR] 			-> [^self concretizeMoveAbR].
 		[MoveRAb]			-> [^self concretizeMoveRAb].
		[MoveMwrR]		-> [^self concretizeMoveMSrR: 3].
		[MoveM32rR]		-> [^self concretizeMoveMSrR: 2].
		[MoveM16rR]		-> [^self concretizeMoveMSrR: 1].
		[MoveMbrR]		-> [^self concretizeMoveMSrR: 0].
		[MoveRMwr]		-> [^self concretizeMoveRMSr: 3].
		[MoveRM32r]		-> [^self concretizeMoveRMSr: 2].
		[MoveRM16r]		-> [^self concretizeMoveRMSr: 1].
		[MoveRMbr]		-> [^self concretizeMoveRMSr: 0].
		[MoveXwrRR]		-> [^self concretizeMoveXSrRR: 3].
		[MoveX32rRR]		-> [^self concretizeMoveXSrRR: 2].
		[MoveX16rRR]		-> [^self concretizeMoveXSrRR: 1].
		[MoveXbrRR]		-> [^self concretizeMoveXSrRR: 0].
		[MoveRXwrR]		-> [^self concretizeMoveRXSrR: 3].
		[MoveRX32rR]		-> [^self concretizeMoveRXSrR: 2].
		[MoveRX16rR]		-> [^self concretizeMoveRXSrR: 1].
		[MoveRXbrR]		-> [^self concretizeMoveRXSrR: 0].
		[MoveM64rRd]		-> [^self concretizeMoveM64rRd].
		[MoveRdM64r]		-> [^self concretizeMoveRdM64r].
		[PopR]				-> [^self concretizePopR].
		[PushR]				-> [^self concretizePushR].
		[NativePopRR]		-> [^self concretizeNativePopRR].
		[NativePushRR]		-> [^self concretizeNativePushRR].
		[PushCq]			-> [^self concretizePushCq].
		[PushCw]			-> [^self concretizePushCw].
		[PrefetchAw]		-> [^self concretizePrefetchAw].
		"Conversion"
		[ConvertRRd]		-> [^self concretizeConvertRRd].
		[ConvertRdR]		-> [^self concretizeConvertRdR].
		"[ConvertRRs]		-> [^self concretizeConvertRRs].
		[ConvertRsR]		-> [^self concretizeConvertRsR].
		[ConvertRsRd]		-> [^self concretizeConvertRsRd].
		[ConvertRdRs]		-> [^self concretizeConvertRdRs]."
			
		[SignExtend8RR]	-> [^self concretizeSignExtendRR: 8].
		[SignExtend16RR]	-> [^self concretizeSignExtendRR: 16].
		[SignExtend32RR]	-> [^self concretizeSignExtendRR: 32].
		
		[ZeroExtend8RR]	-> [^self concretizeZeroExtendRR: 8].
		[ZeroExtend16RR]	-> [^self concretizeZeroExtendRR: 16].
		[ZeroExtend32RR]	-> [^self concretizeZeroExtendRR: 32].

		"Multi-processing"
		[LDAXR]			-> [^self concretizeLDAXR].
		[STLXR]			-> [^self concretizeSTLXR].
		[CLREX]			-> [^self concretizeCLREX].
		[STLR]				-> [^self concretizeSTLR].
		[CASAL]			-> [^self concretizeCASAL].
		[CCMPNE]			-> [^self concretizeCCMPNE].
		[CSET]				-> [^self concretizeCSET].
		[CBNZ]				-> [^self concretizeCB].
		[CBZ]				-> [^self concretizeCB].

		"miscellaneous"
		[MovePerfCnt64RL]	->	[^self concretizeMovePerfCnt64RL] }.

	^0 "keep Slang happy"! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 12/26/2019 23:37'!
emitLd: unitSizeLog2MinusOne rn: baseReg rt: targetReg imm: offset shiftBy12: shiftBy12 at: instrOffset
	"C6.2.130	LDR (immediate)	C6-976
	 C6.2.132	LDR (register)		C6-981
	 C6.2.136	LDRH (immediate)	C6-990
	 C6.2.166	LDUR				C6-1058
	 C6.2.134	LDRB (immediate)	C6-985
	 C6.2.135	LDRB (register)		C6-988"

	"cogit processor disassembleInstructionAt: instrOffset In: machineCode object"
	| unitSize instrBytes |
	unitSize := 1 << unitSizeLog2MinusOne.
	"Unsigned offset, C6-977"
	(offset \\ unitSize = 0
	 and: [offset / unitSize between: 0 and: 1 << 12 - 1]) ifTrue:
		[machineCode
			at: instrOffset / 4
			put: unitSizeLog2MinusOne << 30
				+ (2r11100101 << 22)
				+ (offset << (10 - unitSizeLog2MinusOne))
				+ (baseReg << 5)
				+ targetReg.
		 ^instrOffset + 4].
	(offset between: -256 and: 255) ifTrue: "Unscaled signed 9-bit offset, C6-1058"
		[machineCode
			at: instrOffset / 4
			put: unitSizeLog2MinusOne << 30
				+ (2r11100001 << 22)
				+ ((offset bitAnd: 511) << 12)
				+ (baseReg << 5)
				+ targetReg.
		 ^instrOffset + 4].
	"C6.2.132	LDR (register)	C6-981"
	instrBytes := self emitMoveCw: offset intoR: RISCTempReg at: instrOffset.
	machineCode
		at: instrBytes // 4
		put: unitSizeLog2MinusOne << 30
			+ (2r111000011 << 21)
			+ (RISCTempReg << 16)
			+ (UXTX << 13)
			+ (1 << 11)
			+ (baseReg << 5)
			+ targetReg.
	^instrBytes + 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 12/21/2019 01:47'!
emitLdfprn: baseReg rt: targetDPReg imm: offset shiftBy12: shiftBy12 at: instrOffset
	"C7.2.184	LDR (immediate, SIMD&FP)	C7-1800
	 C7.2.186	LDR (register, SIMD&FP)	C7-1806"

	"cogit processor disassembleInstructionAt: instrOffset In: machineCode object"
	| instrBytes |
	"Unsigned offset, C7-1801"
	(offset \\ 8 = 0
	 and: [offset / 8 between: 0 and: 1 << 12 - 1]) ifTrue:
		[machineCode
			at: instrOffset / 4
			put: 2r1111110101 << 22
				+ (offset << 7 "10 - 3")
				+ (baseReg << 5)
				+ targetDPReg.
		 ^instrOffset + 4].
	"No unscaled signed 9-bit offset because there is only post-index and pre-index, see C7-1800"
	"C6.2.132	LDR (register)	C6-981"
	instrBytes := self emitMoveCw: offset intoR: RISCTempReg at: instrOffset.
	"C7.2.186	LDR (register, SIMD&FP)	C7-1806"
	machineCode
		at: instrBytes // 4
		put: 2r11111100011 << 21
			+ (RISCTempReg << 16)
			+ (UXTX << 13)
			+ (1 << 11)
			+ (baseReg << 5)
			+ targetDPReg.
	^instrBytes + 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 2/23/2020 12:04'!
emitMoveCw: constantArg intoR: destReg at: offsetBytes
	"Emit a load of constant into destReg.  Answer the number of bytes of machine code
	 generated. Literals are stored out-of-line; emit a LDR (literal) with the relevant offset."
	 <var: 'constantArg' type: #usqInt>
	| constant |
	 <var: 'constant' type: #usqInt>
	<inline: false>
	self deny: destReg = SP.
	"cogit processor disassembleInstructionAt: offsetBytes // 4 In: machineCode object"
	constant := (self isAnInstruction: (cogit cCoerceSimple: constantArg to: #'AbstractInstruction *'))
					ifTrue: [(cogit cCoerceSimple: constantArg to: #'AbstractInstruction *') address]
					ifFalse: [constantArg].
	((opcode = MoveCwR or: [opcode = PushCw])
	 and: [self inCurrentCompilation: constant]) ifTrue:
		[| aligned unaligned pcRelativeOffset |
		 "C6.2.10 	ADR	C6-773"
		 unaligned := constant bitAnd: 3.
		 aligned := constant - unaligned.
		 pcRelativeOffset := aligned - address asInteger >>> 2.
		 self assert: (pcRelativeOffset between: -1 << 18 and: 1 << 18 - 1).
		 machineCode
			at: offsetBytes // 4
			put: unaligned << 29
				+ (1 << 28)
				+ ((pcRelativeOffset bitAnd: 1 << 19 - 1) << 5)
				+ destReg.
		 ^offsetBytes + 4].
	self assert: (cogit addressIsInCurrentCompilation: dependent address).
	self assert: dependent address \\ 4 = 0.
	self assert: (dependent address - (address + offsetBytes)) abs < (1 << 20).
	"C6.2.131	LDR (literal)		C6-979"
	machineCode
		at: offsetBytes // 4
		put: 2r01011 << 27
			+ ((dependent address - (address + offsetBytes) bitAnd: 1 << 21 - 1) << 3 "5 - 2")
			+ destReg.
	^offsetBytes + 4! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 7/21/2021 09:59'!
emitSt: unitSizeLog2MinusOne rn: baseReg rt: sourceReg imm: offset shiftBy12: shiftBy12
	"C6.2.273	STR (immediate)	C6-1239
	 C6.2.274	STR (register)		C6-1242
	 C6.2.275	STRB (immediate)	C6-1244"

	| unitSize instrBytes |
	unitSize := 1 << unitSizeLog2MinusOne.
	self deny: SP = sourceReg.
	self deny: baseReg = sourceReg.
	"Unsigned offset, C6-1240"
	(offset \\ unitSize = 0
	 and: [offset / unitSize between: 0 and: 1 << 12 - 1]) ifTrue:
		[machineCode
			at: 0
			put: unitSizeLog2MinusOne << 30
				+ (2r11100100 << 22)
				+ (offset << (10 - unitSizeLog2MinusOne))
				+ (baseReg << 5)
				+ sourceReg.
		 ^4].
	(offset between: -256 and: 255) ifTrue: "Unscaled signed 9-bit offset, C6-1244"
		[machineCode
			at: 0
			put: unitSizeLog2MinusOne << 30
				+ (2r111000000 << 21)
				+ ((offset bitAnd: 511) << 12)
				+ (baseReg << 5)
				+ sourceReg.
		 ^4].
	"If the offset fits in an add or subtract immediate then do an
	 add into RISCTempReg, avoiding using an out-of-line literal..."
	(self isPossiblyShiftableNegatableImm12: offset) ifTrue:
		[self isPossiblyShiftableImm12: offset
			ifTrue:
				[:shift| "C6.2.8		ADDS (immediate)	C6-769"
				machineCode
					at: 0
					put: 2r101100010 << 23
						+ (shift ifTrue: [offset >> 2 + (1 << 22)] ifFalse: [offset << 10])
						+ (baseReg << 5)
						+ RISCTempReg]
			ifFalse:
				[self isPossiblyShiftableImm12: offset negated
					ifTrue:
						[:shift| "C6.2.314		SUBS (immediate)	C6-1321"
						machineCode
							at: 0
							put: 2r111100010 "SUBS" << 23
								+ (shift ifTrue: [offset negated >> 2 + (1 << 22)] ifFalse: [offset negated << 10])
								+ (baseReg << 5)
								+ RISCTempReg]
					ifFalse: [self error: 'cannot happen']].
		machineCode
			at: 1
			put: unitSizeLog2MinusOne << 30
				+ (2r111000001 << 21)
				+ (XZR << 16)
				+ (UXTX << 13)
				+ (1 << 11)
				+ (RISCTempReg << 5)
				+ sourceReg.
		^8].
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	"cogit processor disassembleInstructionAt: 4 In: machineCode object"
	instrBytes := self emitMoveCw: offset intoR: RISCTempReg at: 0.
	self assert: instrBytes = 4.
	machineCode
		at: 1
		put: unitSizeLog2MinusOne << 30
			+ (2r111000001 << 21)
			+ (RISCTempReg << 16)
			+ (UXTX << 13)
			+ (1 << 11)
			+ (baseReg << 5)
			+ sourceReg.
	^8! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 2/21/2020 19:31'!
flushDCacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	"If there is a dual mapped code zone (the normal zone but marked with read/execute, and a
	 read/write zone codeToDataDelta bytes away) then flush the data cache for the corresponding
	 range in the read/write zone and invalidate the data cache for the read/execute zone."
	<cmacro: '(me,startAddress,endAddress) ceFlushDCache(startAddress,endAddress)'>
	^cogit simulateCeFlushDCacheFrom: startAddress to: endAddress! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 7/9/2021 20:33'!
flushICacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	"Flush the instruction cache from (startAddress to endAddress].

	 If there is a dual mapped code zone (the normal zone but marked with read/execute, and a
	 read/write zone codeToDataDelta bytes away) then also flush the data cache for the corresp-
	 onding range in the read/write zone and invalidate the data cache for the read/execute zone.

	 If using the Apple MAP_JIT/pthread_jit_write_protect_np: also flush the data cache.  Use Apple's
	 implementations for both; they are simple and effective."
	
	<inline: #always>
	cogit ensureExecutableCodeZone.
	self cCode:
			[self cppIf: #__APPLE__ & #__MACH__
				ifTrue: [cogit sys_dcache_flush: (self cCoerce: startAddress to: #'void *') _: endAddress - startAddress + 1;
						sys_icache_invalidate: (self cCoerce: startAddress to: #'void *') _: endAddress - startAddress + 1]
				ifFalse: [cogit ceFlushICache: startAddress _: endAddress]]
		inSmalltalk:  [cogit simulateCeFlushICacheFrom: startAddress to: endAddress]! !

!CogARMv8Compiler methodsFor: 'abi' stamp: 'eem 4/20/2015 18:22'!
fullCallsAreRelative
	"Answer if CallFull and/or JumpFull are relative and hence need relocating on method
	 compation. If so, they are annotated with IsRelativeCall in methods and relocated in
	 relocateIfCallOrMethodReference:mcpc:delta:"
	^false! !

!CogARMv8Compiler methodsFor: 'abi' stamp: 'eem 2/25/2020 12:51'!
genAlignCStackSavingRegisters: regMask numArgs: numArgs wordAlignment: alignment
	<inline: true>
	"Override to do nothing.  On the ARMv8 the SP has to be aligned correctly,
	 as enforced by the processor on any and all accesses to the SP."
	^0! !

!CogARMv8Compiler methodsFor: 'abstract instructions' stamp: 'eem 12/27/2019 12:40'!
genDivR: regDivisor R: regDividend Quo: regQuotient Rem: regRemainder
	"Divide regDividend by regDivisor storing the quotient in regQuotient and remainder in regRemainder.
		MSUB Multiply-subtract		MSUB	C6-1109
		SDIV Signed divide			SDIV	C6-1174"
	| instr safeRegQuotient |
	"For the MSUB to work we must preserve regDivisor and regDividend for the MSUB;
	 i.e. the DivRRR must not overwrite either regDivisor or regDividend."
	safeRegQuotient := (regQuotient = regDividend or: [regQuotient = regDivisor])
							ifTrue: [RISCTempReg]
							ifFalse: [regQuotient].
	instr := cogit gen: DivRRR operand: regDividend operand: regDivisor operand: safeRegQuotient.
	"MSUB <Xd>, <Xn>, <Xm>, <Xa>, Xd := Xa - (Xn * Xm)"
	"MSUB regRemainder, regQuotient, regDivisor, regQuotient"
	cogit gen: MSubRRR operand: safeRegQuotient << 5 + regDivisor operand: regDividend operand: regRemainder.
	safeRegQuotient ~= regQuotient ifTrue:
		[cogit gen: MoveRR operand: safeRegQuotient operand: regQuotient].
	^instr
	
	"{ 4 \\ 3. 4 // 3. 4 quo: 3. 4 rem: 3 } #(1 1 1 1)
	{ 4 \\ -3. 4 // -3. 4 quo: -3. 4 rem: -3 } #(-2 -2 -1 1)
	{ -4 \\ 3. -4 // 3. -4 quo: 3. -4 rem: 3 } #(2 -2 -1 -1)
	{ -4 \\ -3. -4 // -3. -4 quo: -3. -4 rem: -3 } #(-1 1 1 -1)"! !

!CogARMv8Compiler methodsFor: 'smalltalk calling convention' stamp: 'eem 8/1/2021 23:19'!
genLoadCStackPointers
	"Load the frame and stack pointer registers with those of the C stack,
	 effecting a switch to the C stack.  Used when machine code calls into
	 the CoInterpreter run-time (e.g. to invoke interpreter primitives).
	 Override to try and use MoveAwRR/ldp"
	cogit cStackPointerAddress + 8 = cogit cFramePointerAddress ifTrue:
		[cogit
			gen: MoveAwRR
			operand: cogit cStackPointerAddress
			operand: SP
			operand: FP.
		 ^0].

	^super genLoadCStackPointers! !

!CogARMv8Compiler methodsFor: 'smalltalk calling convention' stamp: 'eem 8/1/2021 23:19'!
genLoadNativeSPRegWithAlignedSPReg
	cogit AndCq: (objectMemory wordSize * 2 - 1) bitInvert R: SPReg R: SP! !

!CogARMv8Compiler methodsFor: 'smalltalk calling convention' stamp: 'eem 2/1/2020 19:27'!
genLoadStackPointers
	"Switch back to the Smalltalk stack. Assign SPReg first
	 because typically it is used immediately afterwards.
	 Override to try and use MoveAwRR/ldp"
	cogit stackPointerAddress + 8 = cogit framePointerAddress ifTrue:
		[cogit
			gen: MoveAwRR
			operand: cogit stackPointerAddress
			operand: SPReg
			operand: FPReg.
		 ^0].
	cogit framePointerAddress + 8 = cogit stackPointerAddress ifTrue:
		[cogit
			gen: MoveAwRR
			operand: cogit framePointerAddress
			operand: FPReg
			operand: SPReg.
		 ^0].
	
	^super genLoadStackPointers! !

!CogARMv8Compiler methodsFor: 'abi' stamp: 'eem 10/10/2019 17:09'!
genMarshallNArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3
	"Generate the code to pass up to four arguments in a C run-time call.  Hack: each argument is
	 either a negative number, which encodes a constant, or a non-negative number, that of a register.

	 Run-time calls have no more than four arguments, so chosen so that on ARM, where in its C ABI the
	 first four integer arguments are passed in registers, all arguments can be passed in registers.  We
	 defer to the back end to generate this code not so much that the back end knows whether it uses
	 the stack or registers to pass arguments (it does, but...). In fact we defer for an extremely evil reason.
	 Doing so allows the x64 (where up to 6 args are passed) to assign the register arguments in an order
	 that allows some of the argument registers to be used for specific abstract registers, specifically
	 ReceiverResultReg and ClassReg.  This is evil, evil, evil, but also it's really nice to keep using the old
	 register assignments the original author has grown accustomed to."
	<inline: true>
	numArgs = 0 ifTrue: [^self].
	"Avoid arg regs being overwritten before they are read."
	numArgs > 1 ifTrue:
		[((cogit isTrampolineArgConstant: regOrConst1) not
		   and: [regOrConst1 = CArg0Reg]) ifTrue:
			[cogit MoveR: regOrConst1 R: Extra0Reg.
			 ^self genMarshallNArgs: numArgs arg: regOrConst0 arg: Extra0Reg arg: regOrConst2 arg: regOrConst3].
		 numArgs > 2 ifTrue:
			[((cogit isTrampolineArgConstant: regOrConst2) not
			   and: [regOrConst2 = CArg0Reg or: [regOrConst2 = CArg1Reg]]) ifTrue:
				[cogit MoveR: regOrConst2 R: Extra1Reg.
				 ^self genMarshallNArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: Extra1Reg arg: regOrConst3].
			 numArgs > 3 ifTrue:
				[((cogit isTrampolineArgConstant: regOrConst3) not
				   and: [regOrConst3 = CArg0Reg or: [regOrConst3 = CArg1Reg or: [regOrConst3 = CArg2Reg]]]) ifTrue:
					[cogit MoveR: regOrConst3 R: Extra2Reg.
					 ^self genMarshallNArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: Extra2Reg]]]].
	(cogit isTrampolineArgConstant: regOrConst0)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst0) R: CArg0Reg]
		ifFalse: [cogit MoveR: regOrConst0 R: CArg0Reg].
	numArgs = 1 ifTrue: [^self].
	(cogit isTrampolineArgConstant: regOrConst1)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst1) R: CArg1Reg]
		ifFalse: [cogit MoveR: regOrConst1 R: CArg1Reg].
	numArgs = 2 ifTrue: [^self].
	(cogit isTrampolineArgConstant: regOrConst2)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst2) R: CArg2Reg]
		ifFalse: [cogit MoveR: regOrConst2 R: CArg2Reg].
	numArgs = 3 ifTrue: [^self].
	(cogit isTrampolineArgConstant: regOrConst3)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst3) R: CArg3Reg]
		ifFalse: [cogit MoveR: regOrConst3 R: CArg3Reg]! !

!CogARMv8Compiler methodsFor: 'abstract instructions' stamp: 'eem 7/4/2020 14:10'!
genMulOverflowR: regSource R: regDest
	^cogit gen: MulOverflowRRR operand: regSource operand: regDest operand: regDest! !

!CogARMv8Compiler methodsFor: 'abstract instructions' stamp: 'eem 12/21/2019 23:54'!
genMulR: regSource R: regDest
	^cogit gen: MulRRR operand: regSource operand: regDest operand: regDest! !

!CogARMv8Compiler methodsFor: 'smalltalk calling convention' stamp: 'eem 3/20/2015 11:26'!
genPushRegisterArgsForAbortMissNumArgs: numArgs
	"Ensure that the register args are pushed before the outer and
	 inner retpcs at an entry miss for arity <= self numRegArgs.  The
	 outer retpc is that of a call at a send site.  The inner is the call
	 from a method or PIC abort/miss to the trampoline."

	"Putting the receiver and args above the return address means the
	 CoInterpreter has a single machine-code frame format which saves
	 us a lot of work."

	"Iff there are register args convert
		sp		->	outerRetpc			(send site retpc)
		linkReg = innerRetpc			(PIC abort/miss retpc)
	 to
		base	->	receiver
					(arg0)
					(arg1)
		sp		->	outerRetpc			(send site retpc)
		sp		->	linkReg/innerRetpc	(PIC abort/miss retpc)"
	numArgs <= cogit numRegArgs ifTrue:
		[self assert: cogit numRegArgs <= 2.
		 cogit MoveMw: 0 r: SPReg R: TempReg. "Save return address"
		 cogit MoveR: ReceiverResultReg Mw: 0 r: SPReg.
		 numArgs > 0 ifTrue:
			[cogit PushR: Arg0Reg.
			 numArgs > 1 ifTrue:
				[cogit PushR: Arg1Reg]].
		cogit PushR: TempReg]. "push back return address"
	cogit PushR: LinkReg! !

!CogARMv8Compiler methodsFor: 'smalltalk calling convention' stamp: 'eem 12/8/2019 20:44'!
genPushRegisterArgsForNumArgs: numArgs scratchReg: ignored
	"Ensure that the register args are pushed before the retpc for arity <= self numRegArgs."
	"This is easy on a RISC like ARM because the return address is in the link register.  Putting
	 the receiver and args above the return address means the CoInterpreter has a single
	 machine-code frame format which saves us a lot of work
	NOTA BENE: we do NOT push the return address here, which means it must be dealt with later."
	numArgs <= cogit numRegArgs ifTrue:
		[self assert: cogit numRegArgs <= 2.
		 cogit PushR: ReceiverResultReg.
		 numArgs > 0 ifTrue:
			[cogit PushR: Arg0Reg.
			 numArgs > 1 ifTrue:
				[cogit PushR: Arg1Reg]]]! !

!CogARMv8Compiler methodsFor: 'abi' stamp: 'eem 12/8/2019 21:19'!
genRemoveNArgsFromStack: n
	"This is a no-op on ARM64 since the ABI passes up to 6 args in registers and trampolines currently observe that limit, using only 4."
	<inline: true>
	self assert: n <= 6.
	^0! !

!CogARMv8Compiler methodsFor: 'abi' stamp: 'eem 2/25/2020 12:46'!
genRestoreRegs: regMask
	"Restore the registers in regMask as saved by genSaveRegs:.
	 See http://infocenter.arm.com/help/topic/com.arm.doc.den0028b/ARM_DEN0028B_SMC_Calling_Convention.pdf
	 N.B. Alignment is handled by genAlignCStackSavingRegisters:numArgs:wordAlignment:."
	| nRegs pair |
	regMask = 0 ifTrue: [^0].
	self deny: (cogit register: RISCTempReg isInMask: regMask).
	nRegs := 0.
	R1 to: R17 do:
		[:reg|
		 (regMask anyMask: (cogit registerMaskFor: reg)) ifTrue:
			[nRegs := nRegs + 1]].
	pair := nRegs even ifTrue: [NoReg] ifFalse: [RISCTempReg].
	R1 to: R17 do:
		[:reg|
		 (regMask anyMask: (cogit registerMaskFor: reg)) ifTrue:
			[pair = NoReg
				ifTrue: [pair := reg]
				ifFalse:
					[cogit gen: NativePopRR operand: reg operand: pair.
					 pair := NoReg]]].
	self assert: pair = NoReg.
	^0! !

!CogARMv8Compiler methodsFor: 'abi' stamp: 'eem 2/1/2020 17:54'!
genSaveRegs: regMask
	"Save the registers in regMask for a call into the C run-time from a trampoline.
	 See http://infocenter.arm.com/help/topic/com.arm.doc.den0028b/ARM_DEN0028B_SMC_Calling_Convention.pdf
	 N.B. Alignment is handled by genAlignCStackSavingRegisters:numArgs:wordAlignment:."
	| pair |
	regMask = 0 ifTrue: [^0].
	self deny: (cogit register: RISCTempReg isInMask: regMask).
	pair := NoReg.
	R17 to: R0 by: -1 do:
		[:reg|
		 (regMask anyMask: (cogit registerMaskFor: reg)) ifTrue:
			[pair = NoReg
				ifTrue: [pair := reg]
				ifFalse:
					[cogit gen: NativePushRR operand: pair operand: reg.
					 pair := NoReg]]].
	pair ~= NoReg ifTrue:
		[cogit gen: NativePushRR operand: pair operand: RISCTempReg].
	^0! !

!CogARMv8Compiler methodsFor: 'smalltalk calling convention' stamp: 'eem 2/2/2020 13:31'!
genSaveStackPointers
	"Save the frame and stack pointer registers to the framePointer
	 and stackPointer variables.  Used to save the machine code frame
	 for use by the run-time when calling into the CoInterpreter run-time.
	 Override to try and use MoveRRAw/stp"
	cogit stackPointerAddress + 8 = cogit framePointerAddress ifTrue:
		[cogit
			gen: MoveRRAw
			operand: SPReg
			operand: FPReg
			operand: cogit stackPointerAddress.
		 ^0].
	
	^super genSaveStackPointers! !

!CogARMv8Compiler methodsFor: 'abstract instructions' stamp: 'eem 12/20/2019 18:32'!
genSubstituteReturnAddress: retpc
	<inline: true>
	^cogit MoveCw: retpc R: LR! !

!CogARMv8Compiler methodsFor: 'abstract instructions' stamp: 'eem 4/1/2020 13:18'!
genSubstituteReturnAddressR: retpcReg
	<inline: true>
	^cogit MoveR: retpcReg R: LR! !

!CogARMv8Compiler methodsFor: 'disassembly' stamp: 'eem 12/22/2019 18:19'!
generalPurposeRegisterMap
	<doNotGenerate>
	"Answer a Dictionary from register getter to register index."
	^Dictionary newFromPairs:
		{	#r0. R0.
			#r1. R1.
			#r2. R2.
			#r3. R3.
			#r4. R4.
			#r5. R5.
			#r6. R6.
			#r7. R7.
			#r8. R8.
			#r9. R9.
			#r10. R10.
			#r11. R11.
			#r12. R12.
			#r13. R13.
			#r14. R14.
			#r15. R15.
			#r16. R16.
			#r17. R17.
			#r18. R18.
			#r19. R19.
			#r20. R20.
			#r21. R21.
			#r22. R22.
			#r23. R23.
			#r24. R24.
			#r25. R25.
			#r26. R26.
			#r27. R27.
			#r28. R28	}! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 8/12/2021 10:40'!
generateDCacheFlush
	"Use the DC instruction to implement ceFlushDCache(void *start, void *end); see flushDCacheFrom:to:.
	 If there is a dual mapped zone then clean data via DC_CVAU as address + codeToDataDelta,
	 then invalidate data at address via CIVAC."

	"D4.4.7		About cache maintenance in AArch64 state													D4-2478

		Terminology for Clean, Invalidate, and Clean and Invalidate instructions									D4-2479
		...
		-	For instructions operating by VA, the following conceptual points are defined:						D4-2480
		Point of Unification (PoU)							
			The PoU for a PE is the point by which the instruction and data caches and the translation table walks of that
			PE are guaranteed to see the same copy of a memory location. In many cases, the Point of Unification is the
			point in a uniprocessor memory system by which the instruction and data caches and the translation table
			walks have merged.

			The PoU for an Inner Shareable shareability domain is the point by which the instruction and data caches
			and the translation table walks of all the PEs in that Inner Shareable shareability domain are guaranteed to
			see the same copy of a memory location. Defining this point permits self-modifying software to ensure future
			instruction fetches are associated with the modified version of the software by using the standard correctness
			policy of:
				1. Clean data cache entry by address.
				2. Invalidate instruction cache entry by address.

		Example code for cache maintenance instructions D4-2490 - D4-2491"
	self cppIf: #__APPLE__ & #__MACH__
		ifFalse:
			[| mask loop |
			self assert: cogit getCodeToDataDelta ~= 0.
			"Mask is large enough to encompass the method zone and has the correct minimum alignment."
			mask := 1 << (cogit coInterpreter highBit: cogit methodZone zoneEnd) - self dataCacheLineLength.

			"Since this is used from C code we must use only caller-saved registers.
			 C arg registers 2 & 3 are such a convenient pair of caller-saved registers."
			cogit AndCq: mask R: CArg0Reg R: CArg2Reg. "CArg2Reg = aligned pointer to start of each data cache line"
			cogit AddCq: cogit getCodeToDataDelta R: CArg2Reg R: CArg3Reg.
			loop := cogit Label.
			"see concretizeDataCacheControl"
			cogit gen: DC operand: CArg3Reg operand: DC_CVAU.	"clean (flush) address + codeToDataDelta"
			cogit gen: DC operand: CArg2Reg operand: DC_CIVAC.	"invalidate address"
			cogit
				AddCq: self dataCacheLineLength R: CArg2Reg;
				AddCq: self dataCacheLineLength R: CArg3Reg;
				CmpR: CArg1Reg R: CArg2Reg;
				JumpBelow: loop.
			cogit RetN: 0]! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 8/12/2021 10:39'!
generateICacheFlush
	"Use DC VAUC, DSB, IC IVAU, and ISB instructions to implement ceFlushICache(void *start, void *end); see flushICacheFrom:to:.
	 One might think that if there is a dual zone then data at address + codeToDataDelta must be cleaned,
	 but this isn't the case.  All we need to do is clean data at address via DC VAUC and instructions via IC IVAU."

	"B2.2.5		Concurrent modification and execution of instructions											B2-112

		...to avoid UNPREDICTABLE or CONSTRAINED UNPREDICTABLE behavior, instruction modifications must be explicitly synchronized before they are executed. The required synchronization is as follows:

		1.	No PE must be executing an instruction when another PE is modifying that instruction.

		2.	To ensure that the modified instructions are observable, a PE that is writing the instructions must issue the following sequence of instructions and operations:

			; Coherency example for data and instruction accesses within the same Inner Shareable domain.
			; enter this code with <Wt> containing a new 32-bit instruction, to be held in Cacheable space at a location pointed to by Xn.

			STR Wt, [Xn]
			DC CVAU, Xn		; Clean data cache by VA to point of unification (PoU)
			DSB ISH			; Ensure visibility of the data cleaned from cache
			IC IVAU, Xn			; Invalidate instruction cache by VA to PoU
			DSB ISH

			Note
			 -	The DC CVAU operation is not required if the area of memory is either Non-cacheable or Write-Through Cacheable.
			 -	If the contents of physical memory differ between the mappings, changing the mapping of VAs to PAs can cause
				the instructions to be concurrently modified by one PE and executed by another PE. If the modifications affect
				instructions other than those listed as being acceptable for modification, synchronization must be used to avoid
				UNPREDICTABLE or CONSTRAINED UNPREDICTABLE behavior.

		3.	In a multiprocessor system, the IC IVAU is broadcast to all PEs within the Inner Shareable domain of the PE running this sequence.
			However, when the modified instructions are observable, each PE that is executing the modified instructions must issue the following
			instruction to ensure execution of the modified instructions:

			ISB					; Synchronize fetched instruction stream"

	"D4.4.7		About cache maintenance in AArch64 state													D4-2478

		Terminology for Clean, Invalidate, and Clean and Invalidate instructions									D4-2479
		...
		-	For instructions operating by VA, the following conceptual points are defined:						D4-2480
		Point of Unification (PoU)							
			The PoU for a PE is the point by which the instruction and data caches and the translation table walks of that
			PE are guaranteed to see the same copy of a memory location. In many cases, the Point of Unification is the
			point in a uniprocessor memory system by which the instruction and data caches and the translation table
			walks have merged.

			The PoU for an Inner Shareable shareability domain is the point by which the instruction and data caches
			and the translation table walks of all the PEs in that Inner Shareable shareability domain are guaranteed to
			see the same copy of a memory location. Defining this point permits self-modifying software to ensure future
			instruction fetches are associated with the modified version of the software by using the standard correctness
			policy of:
				1. Clean data cache entry by address.
				2. Invalidate instruction cache entry by address.

		Example code for cache maintenance instructions D4-2490 - D4-2491"
	self cppIf: #__APPLE__ & #__MACH__
		ifFalse:
			[| mask loop |
			"See concretizeCacheControlOp1:CRm:Op2: & 
			 http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.100403_0200_00_en/lau1443435580346.html"
			self dataCacheFlushRequired ifTrue: "CTR_EL0.IDC is zero; must clean data cache to point of unification."
				["Since this is used from C code we must use only caller-saved registers.
				  C arg registers 2 & 3 are as such a convenient pair of caller-saved registers."
				 "Mask is large enough to encompass the method zone and has the correct minimum alignment."
				 mask := 1 << (cogit coInterpreter highBit: cogit methodZone zoneEnd) - self dataCacheLineLength.
				 cogit AndCq: mask R: CArg0Reg R: CArg2Reg. "CArg2Reg = aligned pointer to start of each data cache line"
	
				 loop := cogit Label.
				 "see concretizeDataCacheControl"
				 cogit
					gen: DC operand: CArg2Reg operand: DC_CVAU;		"clean (flush) address"
					AddCq: self dataCacheLineLength R: CArg2Reg;
					CmpR: CArg1Reg R: CArg2Reg;
					JumpBelow: loop].

			cogit gen: DSB operand: DSB_ISH operand: DSB_ALL.		"Ensure visibility of the data cleaned from cache"

			self instructionCacheFlushRequired ifTrue: "CTR_EL0.DIC is zero; must clean instruction cache to point of unification."
				["Mask is large enough to encompass the method zone and has the correct minimum alignment."
				 mask := 1 << (cogit coInterpreter highBit: cogit methodZone zoneEnd) - self instructionCacheLineLength.
				 cogit AndCq: mask R: CArg0Reg R: CArg2Reg. "CArg2Reg = aligned pointer to start of each data cache line"
	
				 loop := cogit Label.
				 "see concretizeDataCacheControl"
				 cogit
					gen: IC operand: CArg2Reg operand: IC_IVAU;		"clean (flush) address"
					AddCq: self instructionCacheLineLength R: CArg2Reg;
					CmpR: CArg1Reg R: CArg2Reg;
					JumpBelow: loop.

				 cogit gen: DSB operand: DSB_ISH operand: DSB_ALL].

			cogit
				gen: ISB;
				RetN: 0]! !

!CogARMv8Compiler methodsFor: 'multi-threading' stamp: 'eem 11/12/2020 23:52'!
generateLowLevelTryLock: vmOwnerLockAddress
	"Generate a function that attempts to lock the vmOwnerLock to the argument and answers if it succeeded."
	<inline: true>
	| lockValueReg vmOwnerLockAddressReg br statusReg ldaxr |
	vmOwnerLockAddress = 0 ifTrue:
		[cogit
			MoveCq: 1 R: ABIResultReg;
			RetN: 0.
		 ^self].

	"spiffy 8.1 version using CASAL..."
	lockValueReg := CArg1Reg. "Holds the value of lock if unlocked (zero), receives the existing value of the lock"
	vmOwnerLockAddressReg := CArg2Reg.
	self hasAtomicInstructions ifTrue:
		[cogit
			MoveCq: 0 R: lockValueReg;
			MoveCq: vmOwnerLockAddress R: vmOwnerLockAddressReg;
			gen: CASAL operand: lockValueReg operand: CArg0Reg operand: vmOwnerLockAddressReg;
			CmpCq: 0 R: lockValueReg;
			gen: CCMPNE operand: ABIResultReg operand: lockValueReg operand: 0 "nzcv all false"; "i.e. if NE to 0, then is it already set to the argument?"
			gen: CSET operand: ABIResultReg operand: EQ;
			RetN: 0.
		 ^self].

	"frumpy 8.0 version using LDAXR/STLXR"
				cogit MoveCq: vmOwnerLockAddress R: vmOwnerLockAddressReg.
				cogit MoveCq: 0 R: (statusReg := CArg3Reg). "STLXR sets a word status register; clearing the top bits means it's a non-issue"
	ldaxr :=	cogit gen: LDAXR operand: lockValueReg operand: vmOwnerLockAddressReg.
	br :=		cogit gen: CBNZ operand: 0 operand: lockValueReg.
				cogit gen: STLXR operand: CArg0Reg operand: vmOwnerLockAddressReg operand: statusReg.
				cogit gen: CBNZ operand: ldaxr asUnsignedInteger operand: statusReg.
				"Since CArg0Reg is never zero, merely returning answers true"
				cogit RetN: 0.
	br jmpTarget: (cogit gen: CLREX).
				cogit CmpR: ABIResultReg R: lockValueReg.
				cogit gen: CSET operand: ABIResultReg operand: EQ. "i.e. if NE to 0, then is it already set to the argument?"
				cogit RetN: 0.
	 ^self! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 1/9/2020 21:51'!
getDefaultCogCodeSize
	"Answer the default number of bytes to allocate for native code at startup.
	 The actual value can be set via vmParameterAt: and/or a preference in the ini file."
	<inline: true>
	^1024 * 1664! !

!CogARMv8Compiler methodsFor: 'feature detection' stamp: 'eem 7/7/2021 18:02'!
hasAtomicInstructions
	<cmacro: '(ign) hasAtomicInstructions'>
	"For want of somewhere to put the variable that doesn't bloat an instance of the receiver..."
	^HasAtomicInstructions! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 12/28/2019 22:15'!
hasDoublePrecisionFloatingPointSupport
	^true! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 1/23/2020 15:46'!
hasLZCNTInstructions
	^true! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'lw 7/31/2012 12:07'!
hasLinkRegister
	^true "lr"! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
hasThreeAddressArithmetic
	"Answer if the receiver supports three-address arithmetic instructions (currently only AndCqRR)"
	^true! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 12/15/2019 09:17'!
hasVarBaseRegister
	"Answer if the processor has a dedicated callee-saved register to point to
	 the base of commonly-accessed variables. On ARMv8 we use R27 for this."
	^true! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 7/25/2020 14:59'!
initialFlushICacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	"This must be implemented only if a back end defines any of the following as non-zero:
		numCheckFeaturesOpcodes
		numDCacheFlushOpcodes
		numICacheFlushOpcodes
	 If I or D cache flushing code is jitted (which it will be if num[ID]CacheFlushOpcodes is non-zero) then
	 the ICache must be flushed before the code is used, and that's what this function does."
	<inline: #always>
	self cCode:
			[self cppIf: #__APPLE__ & #__MACH__ ifTrue:
					[cogit sys_dcache_flush: (self cCoerce: startAddress to: #'void *') _: endAddress - startAddress + 1;
						sys_icache_invalidate: (self cCoerce: startAddress to: #'void *') _: endAddress - startAddress + 1]
				cppIf: #__GNUC__ ifTrue:
					[cogit __clear_cache: (self cCoerce: startAddress to: #'char *') _: (self cCoerce: endAddress to: #'char *')]
				ifFalse: [cogit error: 'cache flushing method unknown for this platform']]
		inSmalltalk:
			[cogit processor flushICacheFrom: startAddress to: endAddress]! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 02/04/2020 17:28'!
inlineCacheTagAt: callSiteReturnAddress
	<inline: true>
	^self instructionAt: (self pcRelativeAddressAt: (callSiteReturnAddress - 8) asUnsignedInteger)! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 1/29/2020 14:54'!
inlineCacheValueForSelector: selector in: aCogMethod
	"Answer the value to put in an inline-cache that is being loaded with the selector.
	 Usually this is simply the selector, but in 64-bits the cache is only 32-bits wide and so the
	 cache is loaded with the index of the selector.  Override to use the 64-bit implementation."
	<cmacro: '(backEnd,selector,aCogMethod) indexForSelectorin(selector,aCogMethod)'>
	^cogit indexForSelector: selector in: aCogMethod! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/11/2019 12:21'!
instructionAddressBefore: mcpc
	"Answer the instruction address immediately preceding mcpc."
	<inline: true>
	^mcpc - 4! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 2/4/2020 19:19'!
instructionAt: pc
	<returnTypeC: #'unsigned int'>
	<inline: true>
	^objectMemory long32At: pc! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 02/04/2020 17:28'!
instructionBeforeAddress: followingAddress
	"Answer the instruction immediately preceding followingAddress."
	<inline: true>
	^self instructionAt: (self instructionAddressBefore: followingAddress)! !

!CogARMv8Compiler methodsFor: 'feature detection' stamp: 'eem 7/7/2021 18:02'!
instructionCacheFlushRequired
	<cmacro: '(ign) instructionCacheFlushRequired'>
	"For want of somewhere to put the variable that doesn't bloat an instance of the receiver..."
	^InstructionCacheFlushRequired! !

!CogARMv8Compiler methodsFor: 'feature detection' stamp: 'eem 7/7/2021 18:02'!
instructionCacheLineLength
	<cmacro: '(ign) instructionCacheLineLength'>
	"For want of somewhere to put the variable that doesn't bloat an instance of the receiver..."
	^InstructionCacheLineLength! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 12/21/2019 02:29'!
instructionIsADR: word
	"C6.2.10 ADR	C6-773"
	^(word >> 24 bitAnd: 2r10011111) = 2r00010000! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 12/20/2019 18:18'!
instructionIsBL: word
	<var: 'word' type: #usqInt>
	<inline: true>
	^word >> 26 = 2r100101! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 12/20/2019 19:47'!
instructionIsBLR: word
	<var: 'word' type: #usqInt>
	<inline: true>
	^(word bitOr: 31 << 5) = 2r11010110001111110000001111100000! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 7/25/2021 14:15'!
instructionIsBR: word
	<var: 'word' type: #usqInt>
	<inline: true>
	^(word bitOr: 31 << 5) = 2r11010110000111110000001111100000! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 12/29/2019 13:00'!
instructionIsImm26BorBL: word
	"C6.2.26 	B		C6-799
	 C6.2.33	BL		C6-812"
	<var: 'word' type: #usqInt>
	<inline: true>
	 "BL is 2r100101, B is 2r000101"
	^(word >> 26 bitAnd: 2r11111) = 2r000101! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/10/2019 10:42'!
instructionIsLoadStore: instr
	"C4.1	A64 instruction set encoding on page C4-252"
	^(instr >> 25 bitAnd: 2r101) = 2r100! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/18/2019 13:45'!
instructionIsPCRelativeLoad: instr
	"C6.2.131	LDR (literal)		C6-979
	 C6.2.143	LDRSW (literal)	C6-1008"
	^(instr >> 24 bitAnd: 2r00111111) = 2r00011000! !

!CogARMv8Compiler methodsFor: 'disassembly' stamp: 'eem 3/18/2015 16:13'!
instructionSizeAt: pc
	"Answer the instruction size at pc.Simple on ARM ;-)"
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 12/19/2019 11:50'!
inverseForArithOp: arithOp
	"Several of the opcodes are inverses.  Answer the inverse for an opcode if it has one.
	 See Table A3-2 in sec A3.4 Data-processing instructions of the AARM."
	^arithOp caseOf: {
			[ArithmeticAdd]	->	[ArithmeticSub].
			[ArithmeticAddS]	->	[ArithmeticSubS].
			[ArithmeticSub]	->	[ArithmeticAdd].
			[ArithmeticSubS]	->	[ArithmeticAddS] }! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 12/9/2019 15:20'!
is6BitSignedImmediate: constant ifTrue: binaryBlock ifFalse: nullaryBlock
	<inline: #always>
	(constant between: -1 << 5 and: 1 << 5)
		ifTrue:
			[^binaryBlock
				value: constant abs
				value: constant < 0]
		ifFalse: [^nullaryBlock value]! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 12/2/2019 16:47'!
isAddressRelativeToVarBase: varAddress
	<inline: true>
	<var: #varAddress type: #usqInt>
	"Support for addressing variables off the dedicated VarBaseReg"
	^varAddress notNil
	  and: [varAddress
				between: cogit varBaseAddress - (1 << 8)
				and: cogit varBaseAddress + (1 << 12) - 1]! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
isBigEndian
	^false! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 7/25/2021 11:44'!
isCallPrecedingReturnPC: mcpc
	"Assuming mcpc is a send return pc answer if the instruction before it is a call (not a CallFull)."
	"There are two types of calls: BL & BLR; BLR is used for CallFull"
	| instruction |
	instruction := self instructionBeforeAddress: mcpc.
	^(self instructionIsBL: instruction) or: [self instructionIsBLR: instruction]! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 7/25/2021 14:11'!
isFullJumpAtPC: mcpc
	^self instructionIsBR: (objectMemory long32At: mcpc)! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 7/21/2021 08:39'!
isImm12orImm9offset: offset
	"ARM64 load/store immediate offsets have an 11 bit unsigned amd a 9 bit signed form."
	^offset >= -256 and: [offset < 4096]! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 1/29/2020 18:21'!
isImmNImmSImmREncodableBitmask: constant ifTrue: trinaryBlock "[:n :imms :immr|...]" ifFalse: nullaryBlock
	"See DecodeBitMasks J1-7389.
	 See https://dinfuehr.github.io/blog/encoding-of-immediate-values-on-aarch64/
	 This method is adapted from The LLVM Compiler Infrastructure, AArch64AddressingModes.h processLogicalImmediate"
	<inline: #always>
	| imm size mask numLeadingOnes numTrailingOnes immr n nImms rotateCount |
	<var: 'mask' type: #usqInt>
	<var: 'nImms' type: #usqInt>
	(constant between: -1 and: 0) ifTrue:
		[^nullaryBlock value].
	imm := self cCode: [constant] inSmalltalk: [constant signedIntToLong64].
 
	"First, determine the element size."
	size := 32.
	[mask := 1 << size - 1.
	 (imm bitAnd: mask) ~= (imm >> size)
			ifTrue: [size := size * 2. false]
			ifFalse: [size > 2]]
		whileTrue: [size := size / 2].

	"Second, determine the rotation to make the element be: 0^m 1^n."
	mask := 1 << 64 - 1 >> (64 - size).
	imm := imm bitAnd: mask.

	(self isShiftedMask: imm)
		ifTrue:
			[rotateCount := self countTrailingZeros: imm.
			 numTrailingOnes := self countTrailingOnes: imm >> rotateCount]
		ifFalse:
			[imm := imm bitOr: mask bitInvert64.
			 (self isShiftedMask: imm) ifFalse:
				[^nullaryBlock value].
			 numLeadingOnes := self countLeadingOnes: imm.
			 rotateCount := 64 - numLeadingOnes.
			 numTrailingOnes := numLeadingOnes + (self countTrailingOnes: imm) - (64 - size)].

	"Encode in Immr the number of RORs it would take to get *from* 0^m 1^n
	 to our target value, where I is the number of RORs to go the opposite direction."
 
	self assert: size > rotateCount. "rotateCount should be smaller than element size"
	immr := size - rotateCount bitAnd: size - 1.

	"If size has a 1 in the n'th bit, create a value that has zeroes in bits [0, n] and ones above that."
	nImms := (size - 1) bitInvert64 << 1.

	"Or the CTO value into the low bits, which must be below the Nth bit mentioned above."
	nImms := nImms bitOr:  numTrailingOnes - 1.

	"Extract the seventh bit and toggle it to create the N field."
	n := ((nImms >> 6) bitAnd: 1) bitXor: 1.

	nImms := nImms bitAnd: 16r3F.

	self assert: (self decode64Imms: nImms immr: immr) = constant signedIntToLong64.

	^trinaryBlock
		value: n
		value: nImms
		value: immr
! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 2/4/2020 14:22'!
isInImmediateBranchAndLinkRange: offset
	"ARM64 calls span +/- 128 mb.
	 C6.2.33 BL		C6-812"
	<var: #offset type: #'sqIntptr_t'>
	self assert: (offset noMask: 3).
	^offset signedIntFromLong64 >>> 27 between: -16r1 and: 0! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 1/30/2020 16:38'!
isInImmediateBranchRange: offset
	"ARM64 calls and jumps span +/- 1 mb."
	<var: #offset type: #'usqIntptr_t'>
	self assert: (offset noMask: 3).
	^offset signedIntFromLong64 >>> 18 between: -16r1 and: 0! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 12/17/2019 14:51'!
isInImmediateJumpRange: operand
	"ARMv8 calls and jumps span +/- 128 mb, more than enough for intra-zone calls and jumps."
	<var: #operand type: #'usqIntptr_t'>
	self assert: (operand noMask: 3).
	^operand signedIntFromLong between: -1 << 27 and: 1 << 27 - 1! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 7/21/2021 08:16'!
isInImmediateOffsetRange: offset
	"ARM64 load/store immediates have an 11 bit unsigned amd a 9 bit signed form."
	^offset >= -256 and: [offset < 4096]! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 11/12/2020 18:40'!
isJump
	^super isJump or: [opcode between: CBNZ and: CBZ]! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 02/04/2020 17:28'!
isJumpAt: pc
	| op0_101_op1MSB |
	"C4.1	A64 instruction set encoding on page C4-252
	 C4.1.3 Branches, Exception Generating and System instructions"
	"cogit processor disassembleInstructionAt: pc In: objectMemory memory"
	op0_101_op1MSB := (self instructionAt: pc) bitShift: -25.
	^op0_101_op1MSB = 2r0101010								"Conditional branch immediate"
	or: [op0_101_op1MSB = 2r1101011							"Unconditional branch (register)"
	or: [(op0_101_op1MSB bitAnd: 2r1111110) = 2r1001010		"Unconditional Branch (immediate)	BL"
	or: [(op0_101_op1MSB bitAnd: 2r1111110) = 2r0001010]]]	"Unconditional branch (immediate) & Compare/Test and branch (immediate)"! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 11/25/2019 17:12'!
isPCDependent
	"Answer if the receiver is a pc-dependent instruction.  With out-of-line literals any instruction
	 that refers to a literal depends on the address of the literal, so add them in addition to the jumps."
	^self isJump
	  or: [opcode = AlignmentNops
	  or: [opcode ~= Literal and: [dependent notNil and: [dependent opcode = Literal]]]]! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 3/20/2020 12:58'!
isPossiblyShiftableImm12: immediate ifTrue: unaryBlock ifFalse: nullaryBlock
	<inline: #always>
	(immediate between: 0 and: 1 << 12 - 1) ifTrue:
		[^unaryBlock value: false].
	((immediate noMask: 1 << 12 - 1)
	 and: [immediate >>> 12 between: 0 and: 1 << 12 - 1]) ifTrue:
		[^unaryBlock value: true].
	^nullaryBlock value! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 3/20/2020 12:58'!
isPossiblyShiftableImm12orImm9: immediate ifTrue: unaryBlock ifFalse: nullaryBlock
	<inline: #always>
	((immediate between: -256 and: 255)
	 or: [immediate between: 0 and: 1 << 12 - 1]) ifTrue:
		[^unaryBlock value: false].
	((immediate noMask: 1 << 12 - 1)
	 and: [immediate >>> 12 between: 0 and: 1 << 12 - 1]) ifTrue:
		[^unaryBlock value: true].
	^nullaryBlock value! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 7/21/2021 09:36'!
isPossiblyShiftableNegatableImm12: immediate
	<inline: #always>
	^(immediate between: -1 << 12 and: 1 << 12 - 1)
	 or: [(immediate noMask: 1 << 12 - 1)
		 and: [immediate >>> 12 between: -1 << 12 and: 1 << 12 - 1]]! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 3/20/2020 12:49'!
isPossiblyShiftableNegatableImm12: immediate ifTrue: unaryBlock ifFalse: nullaryBlock
	<inline: #always>
	(immediate between: -1 << 12 and: 1 << 12 - 1) ifTrue:
		[^unaryBlock value: false].
	((immediate noMask: 1 << 12 - 1)
	 and: [immediate >>> 12 between: -1 << 12 and: 1 << 12 - 1]) ifTrue:
		[^unaryBlock value: true].
	^nullaryBlock value! !

!CogARMv8Compiler methodsFor: 'generate machine code' stamp: 'eem 12/17/2019 12:38'!
isSharable
	"Hack:  To know if a literal should be unique (not shared) mark the LSB of the second operand."
	<inline: true>
	self assert: opcode = Literal.
	^(operands at: 1) anyMask: 1! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 12/3/2019 17:53'!
isShiftedMask: anInteger
	| bits |
	^anInteger ~= 0
	  and: [bits := anInteger - 1 bitOr: anInteger.
		   (bits bitAnd: bits + 1) = 0]! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 12/21/2019 01:49'!
isUnsigned12BitMultipleOf8: anInteger
	^anInteger \\ 8 = 0
	  and: [anInteger / 8 between: 0 and: 1 << 12 - 1]! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'tpr 4/17/2015 19:56'!
jumpLongByteSize
"	Branch/Call ranges.  Jump[Cond] can be generated as short as possible.  Call/Jump[Cond]Long must be generated
	in the same number of bytes irrespective of displacement since their targets may be updated, but they need only
	span 16Mb, the maximum size of the code zone.  This allows e.g. ARM to use single-word call and jump instructions
	for most calls and jumps.  CallFull/JumpFull must also be generated in the same number of bytes irrespective of
	displacement for the same reason, but they must be able to span the full (32-bit or 64-bit) address space because
	they are used to call code in the C runtime, which may be distant from the code zone"
	^4! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 12/10/2019 12:59'!
jumpLongConditionalByteSize
	^8! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/11/2019 12:16'!
jumpLongTargetBeforeFollowingAddress: mcpc
	"C6.2.26 	B		C6-799"
	^self callTargetFromReturnAddress: mcpc! !

!CogARMv8Compiler methodsFor: 'disassembly' stamp: 'eem 02/04/2020 17:28'!
jumpTargetPCAt: pc
	<returnTypeC: #usqInt>
	| operand word |
	"cogit processor disassembleInstructionAt: pc In: objectMemory memory"
	word := self instructionAt: pc.
	self assert: word >> 26 = 2r10101.
	operand := word >> 5 bitAnd: 16r7FFFF.
	(operand anyMask: 16r40000) ifTrue:
		[operand := operand - 16r80000].
	^self
		cCode: [operand * 4 + pc]
		inSmalltalk: [operand * 4 + pc bitAnd: cogit addressSpaceMask]! !

!CogARMv8Compiler methodsFor: 'abi' stamp: 'eem 6/1/2020 17:19'!
leafCallStackPointerDelta
	"Answer the delta from the stack pointer after a call to the stack pointer
	 immediately prior to the call.  This is used to compute the stack pointer
	 immediately prior to a call from within a leaf routine, which in turn is used
	 to capture the c stack pointer to use in trampolines back into the C run-time."
	"This might actually be false, since directly after a call, lr, fp and variable registers need be pushed onto the stack. It depends on the implementation of call."
	^0! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 02/04/2020 17:28'!
literal32BeforeFollowingAddress: followingAddress
	"Answer the 32-bit constant loaded by the instruction sequence just before this address:
	  CmpC32R MoveC32R"
	^self instructionAt:
		(self pcRelativeAddressAt:
			(self instructionAddressBefore:
				((self instructionIsPCRelativeLoad: (self instructionBeforeAddress: followingAddress))
					ifTrue: [followingAddress]
					ifFalse: [followingAddress - 4])))! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 12/10/2019 12:15'!
literalAlignment
	"Answer the required alignment for literals embedded in code."
	<inline: true>
	^8! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/10/2019 20:42'!
literalBeforeFollowingAddress: followingAddress
	"Answer the literal referenced by the instruction immediately preceding followingAddress.
	  ArithCwR MoveCwR PushCw"
	^objectMemory longAt:
		(self pcRelativeAddressAt:
			(self instructionAddressBefore:
				((self instructionIsPCRelativeLoad: (self instructionBeforeAddress: followingAddress))
					ifTrue: [followingAddress]
					ifFalse: [followingAddress - 4])))! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 6/8/2015 12:20'!
literalLoadInstructionBytes
	"Answer the size of a literal load instruction (which does not include the size of the literal).
	 With out-of-line literals this is always a single LDR instruction that refers to the literal."
	<inline: true>
	^4! !

!CogARMv8Compiler methodsFor: 'generate machine code' stamp: 'eem 6/16/2015 17:31'!
literalOpcodeIndex
	"Hack:  To know how far away a literal is from its referencing instruction we store
	 its opcodeIndex, or -1, if as yet unassigned, in the second operand of the literal."
	<inline: true>
	self assert: opcode = Literal.
	^(operands at: 2) asInteger! !

!CogARMv8Compiler methodsFor: 'generate machine code' stamp: 'eem 12/17/2019 13:09'!
literalSize
	"Hack:  To handle both 32-bit and 64-bit literals, store the literal's size in a field of the second operand."
	<inline: true>
	self assert: opcode = Literal.
	^(operands at: 1) "If not set, default to bytesPerOop"
		ifNil: [objectMemory bytesPerOop]
		ifNotNil: [((operands at: 1) >> 1) bitAnd: 15]! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 12/27/2019 08:27'!
loadLiteralByteSize
	"Answer the byte size of a MoveCwR opcode's corresponding machine code.
	 On ARMv8 this is a single instruction pc-relative register load"
	^4! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 4/21/2015 17:30'!
loadPICLiteralByteSize
	"Answer the byte size of a MoveCwR opcode's corresponding machine code
	 when the argument is a PIC.  This is for the self-reference at the end of a
	 closed PIC.  On ARM this is a single instruction pc-relative register load."
	^4! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 4/20/2015 19:26'!
machineCodeAt: anOffset
	"read aWord from machineCode, with little endian"
	<inline: true>
	^machineCode at: anOffset // 4! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 4/20/2015 19:28'!
machineCodeAt: anOffset put: aWord
	"add aWord to machineCode, with little endian"
	<inline: true>
	machineCode at: anOffset // 4 put: aWord! !

!CogARMv8Compiler methodsFor: 'generate machine code' stamp: 'eem 12/1/2019 21:16'!
machineCodeBytes
	"Answer the maximum number of bytes of machine code generated for any abstract instruction.
	Likely to be quite different for AARCH64"
	^12! !

!CogARMv8Compiler methodsFor: 'generate machine code' stamp: 'eem 12/1/2019 21:16'!
machineCodeWords
	"Answer the maximum number of words of machine code generated for any abstract instruction.
	Likely to be quite different for AARCH64"
	^self machineCodeBytes // 4! !

!CogARMv8Compiler methodsFor: 'memory access' stamp: 'eem 7/6/2021 14:44'!
makeCodeZoneExecutable
	<inline: #always>
	self cCode: [self cppIf: #__APPLE__ & #__MACH__ ifTrue: [cogit pthread_jit_write_protect_np: true]]! !

!CogARMv8Compiler methodsFor: 'memory access' stamp: 'eem 7/6/2021 14:51'!
makeCodeZoneWritable
	<inline: #always>
	self cCode: [self cppIf: #__APPLE__ & #__MACH__ ifTrue: [cogit pthread_jit_write_protect_np: false]]! !

!CogARMv8Compiler methodsFor: 'generate machine code' stamp: 'eem 12/29/2019 12:17'!
mapEntryAddress
	"Typically map entries apply to the end of an instruction, for two reasons:
	  a)	to cope with literals embedded in variable-length instructions, since, e.g.
		on x86, the literal typically comes at the end of the instruction.
	  b)	in-line cache detection is based on return addresses, which are typically
		to the instruction following a call.
	 But on ARMv8 we use out-of-line literals so the mapEntryAddress of a literal is simply the instruction's address."
	<inline: true>
	^opcode = Literal
		ifTrue: [address]
		ifFalse: [address + machineCodeSize]! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 12/2/2019 17:18'!
movern: srcReg rd: destReg
	^self addrn: srcReg rd: destReg imm: 0 shiftBy12: false! !

!CogARMv8Compiler methodsFor: 'printing' stamp: 'eem 12/8/2019 22:43'!
nameForRegister: reg "<Integer>"
	<doNotGenerate>
	| default |
	default := super nameForRegister: reg.
	^default last = $?
		ifTrue:
			[reg = 31
				ifTrue:
					['SP/XZR']
				ifFalse:
					[#(LR CArg0Reg CArg1Reg CArg2Reg CArg3Reg)
						detect: [:sym| (thisContext method methodClass classPool at: sym) = reg] 
						ifNone: [default]]]
		ifFalse:
			[default]! !

!CogARMv8Compiler methodsFor: 'memory access' stamp: 'eem 7/5/2021 14:50'!
needsCodeZoneExecuteWriteSwitch
	"On e.g. Apple M1 with the hardened runtime pthread_jit_write_protect_np
	 must be used to enable execution and disable write-protect.
	 See the comment in ensureExecutableCodeZone."
	^true! !

!CogARMv8Compiler methodsFor: 'abstract instructions' stamp: 'eem 9/2/2020 10:17'!
noteFollowingConditionalBranch: branch
	"Morph an Overflow/NoOverflow check after a MulOverflow into an efficient
	 test for overflow, hence avoiding having to set and test the V flag.
	 MulOverflowRRR generates
		mul		Rd,Rm,Rn
		smulh 	RISCTempReg,Rm,Rn
		cmp	RISCTempReg, #0x1
	 so it must be followed by an unsigned conditional jump, Above for Overflow, BelowOrEqual for NoOverflow."
	<var: #branch type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	
	(opcode = MulOverflowRRR
	 and: [branch opcode = JumpOverflow or: [branch opcode = JumpNoOverflow]])  ifTrue:
		[branch opcode: (branch opcode = JumpOverflow
							ifTrue: [JumpMulOverflow]
							ifFalse: [JumpNoMulOverflow])].

	^branch! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 7/9/2021 20:15'!
numDCacheFlushOpcodes
	self cppIf: #DUAL_MAPPED_CODE_ZONE
		ifTrue: [^cogit getCodeToDataDelta ~= 0 ifTrue: [15] ifFalse: [0]]
		ifFalse: [^0]! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 7/9/2021 20:38'!
numICacheFlushOpcodes
	"Apple's code is simple and works; we can't better it..."
	self cppIf: #__APPLE__
		ifTrue: [^0]
		ifFalse: [^24]! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 12/8/2019 20:52'!
numIntRegArgs
	"See e.g. http://infocenter.arm.com/help/topic/com.arm.doc.den0028b/ARM_DEN0028B_SMC_Calling_Convention.pdf
	 Table 3-1 Register Usage in AArch64 SMC32, HVC32, SMC64, and HVC64 calls"
	^6! !

!CogARMv8Compiler methodsFor: 'multi-threading' stamp: 'eem 11/12/2020 18:30'!
numLowLevelLockOpcodes
	<inline: #always>
	^14 "12 opcodes in 8.0 LDAXR/STLXR version, plus two words for the vmOwnerLockAddress"! !

!CogARMv8Compiler methodsFor: 'generate machine code' stamp: 'eem 12/3/2019 08:33'!
outOfLineLiteralOpcodeLimit
	"The maximum offset in a LDR (literal) is -2^18 to 2^18-1.
	 And this is multiplied by 4 to produce the effective address.
	 This is a huge range; we have no grounds for concern."
	^1 << 18 - 1! !

!CogARMv8Compiler methodsFor: 'generate machine code' stamp: 'eem 1/24/2020 23:19'!
outputMachineCodeAt: targetAddress
	"Override to move machine code a word at a time."
	<inline: true>
	0 to: machineCodeSize - 1 by: 4 do:
		[:j|
		cogit codeLong32At: targetAddress + j put: (machineCode at: j // 4)]! !

!CogARMv8Compiler methodsFor: 'generate machine code' stamp: 'eem 1/27/2020 08:43'!
padIfPossibleWithStopsFrom: startAddr to: endAddr
	| nullBytes |
	nullBytes := (endAddr - startAddr + 1) \\ 4.
	self stopsFrom: startAddr to: endAddr - nullBytes.
	endAddr - nullBytes + 1 to: endAddr 
		do: [ :p | cogit codeByteAt: p put: 16rFF]! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 02/04/2020 17:28'!
pcRelativeAddressAt: mcpc
	"Extract the address of the adr rX, offset instruction at address mcpc
	 C6.2.131	LDR (literal)		C6-979"
	| instr |
	instr := self instructionAt: mcpc.
	"cogit processor disassembleInstructionAt: mcpc In: objectMemory memory"
	self assert: (self instructionIsPCRelativeLoad: instr).
	^(instr >> 3 bitAnd: 1 << 22 - 4) - (instr >> 2 bitAnd: 1 << 21) + mcpc! !

!CogARMv8Compiler methodsFor: 'feature detection' stamp: 'eem 8/5/2021 10:34'!
preferredRegisterForMovePerfCnt64RL
	"Answer the preferred register to receive the 64-bit performance counter,
	 or NoReg if none."
	<inline: #always>
	^NoReg! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 7/24/2021 19:30'!
prn: baseReg imm: offset shiftBy12: shiftBy12
	"C6.2.211	PRFM (immediate)	C6-1136"

	"Unsigned offset, C6-1136"
	"This is the only case we can make use of so far..."
	self assert: (offset \\ 8 = 0
				 and: [offset / 8 between: 0 and: 1 << 12 - 1]).
	^2r1111100110 << 22
		+ (offset << 7 "10 - 3")
		+ (baseReg << 5)
		+ 0 "prfop = PLD:L1:KEEP"! !

!CogARMv8Compiler methodsFor: 'accessing' stamp: 'eem 3/18/2015 15:35'!
pushLinkRegisterByteSize
	^4! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 1/24/2020 23:19'!
relocateCallBeforeReturnPC: retpc by: delta
	| instr distanceDiv4 |
	"C6.2.26 	B		C6-799
	 C6.2.33	BL		C6-812"
	self assert: delta \\ 4 = 0.
	delta ~= 0 ifTrue:
		[instr := self instructionBeforeAddress: retpc.
		 self assert: (self instructionIsImm26BorBL: instr).

		 distanceDiv4 := instr bitAnd: 2 << 26 - 1.
		 distanceDiv4 := distanceDiv4 + (delta // 4).
		 cogit
			codeLong32At: (self instructionAddressBefore: retpc)
			put: ((instr bitAnd: 16rFC000000) bitOr: (distanceDiv4 bitAnd: 16r3FFFFFF))]

	"cogit processor disassembleInstructionAt: (self instructionAddressBefore: retpc) In: objectMemory memory"! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 02/04/2020 17:28'!
relocateMethodReferenceBeforeAddress: pc by: delta
	"We generate the method address using pc-relative addressing.
	 Simply check that pc-relative addressing is being used. c.f.
	 emitMoveCw:intoR:at:"
	<inline: true>
	self assert: ((self instructionIsADR: (self instructionAt: pc - 4)) "MoveCwR"
			or: [self instructionIsADR: (self instructionAt: pc - 8)]) "PushCwR"

	"cogit processor disassembleInstructionAt: (self instructionAddressBefore: pc - 4) In: objectMemory memory"
	"cogit processor disassembleInstructionAt: (self instructionAddressBefore: pc) In: objectMemory memory"! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/27/2019 09:49'!
rewriteCPICJumpAt: addressFollowingJump target: jumpTargetAddr
	"Rewrite a jump instruction to call a different target.  This variant is used to reset the 
	jumps in the prototype CPIC to suit each use,.   
	Answer the extent of the code change which is used to compute the range of the icache to flush."
	<var: #addressFollowingJump type: #usqInt>
	<var: #jumpTargetAddr type: #usqInt>
	<inline: true>
	"cogit processor disassembleInstructionAt: addressFollowingJump - 4 In: objectMemory memory"
	self rewriteImm19JumpBefore: addressFollowingJump target: jumpTargetAddr! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 1/24/2020 23:20'!
rewriteCallAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a call instruction to call a different target.  This variant is used to link PICs
	 in ceSendMiss et al, and to rewrite cached primitive calls.   Answer the extent of
	 the code change which is used to compute the range of the icache to flush."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	cogit codeLong32At: callSiteReturnAddress - 4 put: (self bl: callTargetAddress - (callSiteReturnAddress - 4)).
	^4! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 2/23/2020 15:34'!
rewriteImm19JumpBefore: followingAddress target: targetAddress
	"Rewrite a long jump/call instruction to jump/call to a different target.
	 Answer the extent of the code change which is used to compute the range of the icache to flush.
	 C6.2.25	B.cond		C6-798"
	| mcpc offset instruction instrOpcode |
	"cogit processor disassembleInstructionAt: followingAddress - 4 In: objectMemory memory"
	mcpc := self instructionAddressBefore: followingAddress.
	offset := targetAddress - mcpc.
	instruction := self instructionBeforeAddress: followingAddress.
	instrOpcode := instruction >> 26.
	self assert: instrOpcode = 2r10101.
	cogit
		codeLong32At: mcpc
		put: instrOpcode << 26 + ((offset >>> 2 bitAnd: 1 << 19 - 1) << 5) + (instruction bitAnd: 2r1111).
	^4! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 2/23/2020 14:31'!
rewriteImm26JumpBefore: followingAddress target: targetAddress
	"Rewrite a long jump/call instruction to jump/call to a different target.
	 Answer the extent of the code change which is used to compute the range of the icache to flush.
	 C6.2.26	B	C6-799
	 C6.2.33 	BL	C6-812"
	| mcpc offset instrOpcode |
	"cogit processor disassembleInstructionAt: followingAddress - 4 In: objectMemory memory"
	mcpc := self instructionAddressBefore: followingAddress.
	offset := targetAddress - mcpc.
	instrOpcode := (self instructionBeforeAddress: followingAddress) >> 26.
	self assert: (instrOpcode = 2r101 or: [instrOpcode = 2r100101]).
	cogit
		codeLong32At: mcpc
		put: instrOpcode << 26 + (offset >>> 2 bitAnd: 1 << 26 - 1).
	^4! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 10/7/2020 18:12'!
rewriteInlineCacheAt: callSiteReturnAddress tag: cacheTag target: callTargetAddress
	"Rewrite an inline cache to call a different target for a new tag.  This variant is used
	 to link unlinked sends in ceSend:to:numArgs: et al.  Answer the extent of the code
	 change which is used to compute the range of the icache to flush.
	 N.B.  On 64-bit platforms the inline cache tag is only 32-bits wide, hence this code
	 is very similar to that for ARM32 CogOutOfLineLiteralsARMCompiler."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	| call callDistance |
	(cogit initialized and: [callTargetAddress < cogit minCallAddress]) ifTrue:
		[self error: 'linking callsite to invalid address'].
	callDistance := (callTargetAddress - (callSiteReturnAddress - 4 "return offset")) signedIntToLong.
	self assert: (self isInImmediateJumpRange: callDistance). "we don't support long call updates here"
	call := self bl: callDistance.
	cogit
		codeLong32At: (self instructionAddressBefore: callSiteReturnAddress) put: call;
		codeLong32At: (self pcRelativeAddressAt: callSiteReturnAddress - 8) put: cacheTag signedIntToLong.
	self assert: (self inlineCacheTagAt: callSiteReturnAddress) signedIntFromLong = cacheTag.
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: callSiteReturnAddress - 8 to: (self pcRelativeAddressAt: callSiteReturnAddress - 8)]."
	^4! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 1/23/2020 20:21'!
rewriteInlineCacheTag: cacheTag at: callSiteReturnAddress
	"Rewrite an inline cache with a new tag.  This variant is used
	 by the garbage collector.  RThis cannot happen in 64-bits as cache tags are guaranteed to be 32-bits or less."
	self error: 'should not happen'! !

!CogARMv8Compiler methodsFor: 'full transfer run-time support' stamp: 'eem 1/24/2020 23:19'!
rewriteJumpFullAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a JumpFull instruction to jump to a different target.  This variant is used to rewrite cached primitive calls.
	 Answer the extent of the code change which is used to compute the range of the icache to flush."
	"cogit processor disassembleInstructionAt: callSiteReturnAddress - 8 In: objectMemory memory"
	"cogit processor disassembleInstructionAt: callSiteReturnAddress - 4 In: objectMemory memory"
	cogit
		codeLongAt: (self pcRelativeAddressAt: callSiteReturnAddress - 8)
		put: callTargetAddress.
	^0! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/27/2019 08:47'!
rewriteJumpLongAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a long jump instruction to jump to a different target.  This variant
	 is used to rewrite cached primitive calls.   Answer the extent of the
	 code change which is used to compute the range of the icache to flush."
	<inline: true>
	^self rewriteImm26JumpBefore: callSiteReturnAddress target: callTargetAddress! !

!CogARMv8Compiler methodsFor: 'method zone and entry point alignment' stamp: 'eem 1/18/2021 16:37'!
roundUpToMethodAlignment: numBytes
	"Determine the default alignment for the start of a CogMethod, which in turn
	 determines the size of the mask used to distinguish the checked and unchecked
	 entry-points, used to distinguish normal and super sends on method unlinking.
	 This is implemented here to allow processors with coarse instructions (ARM) to
	 increase the alignment if required."
	<cmacro: '(ignored,numBytes) (((numBytes) + 15) & -16)'> "extra parens to placate gdb :-("
	^numBytes + 15 bitAnd: -16! !

!CogARMv8Compiler methodsFor: 'abi' stamp: 'eem 3/19/2015 15:18'!
saveAndRestoreLinkRegAround: aBlock
	"If the processor's ABI includes a link register, generate instructions
	 to save and restore it around aBlock, which is assumed to generate code."
	<inline: true>
	| inst |
	inst := cogit PushR: LinkReg.
	aBlock value.
	cogit PopR: LinkReg.
	^inst! !

!CogARMv8Compiler methodsFor: 'abi' stamp: 'eem 12/28/2019 23:35'!
saveAndRestoreLinkRegUsingCalleeSavedRegNotLiveAtPointOfSendAround: aBlock
	"Extra7Reg is callee-saved and not live at point of send."
	<inline: #always>
	| inst |
	inst := cogit MoveR: LinkReg R: Extra8Reg.
	aBlock value.
	cogit MoveR: Extra8Reg R: LinkReg! !

!CogARMv8Compiler methodsFor: 'feature detection' stamp: 'eem 7/7/2021 12:18'!
setDataCacheFlushRequired: boolean
	<cmacro: '(ign,b) dataCacheFlushRequired = b'>
	"For want of somewhere to put the variable that doesn't bloat an instance of the receiver..."
	DataCacheFlushRequired := boolean! !

!CogARMv8Compiler methodsFor: 'feature detection' stamp: 'eem 7/7/2021 12:18'!
setDataCacheLineLength: n
	<cmacro: '(ign,n) dataCacheLineLength = n'>
	"For want of somewhere to put the variable that doesn't bloat an instance of the receiver..."
	DataCacheLineLength := n! !

!CogARMv8Compiler methodsFor: 'feature detection' stamp: 'eem 7/7/2021 15:35'!
setHasAtomicInstructions: boolean
	<cmacro: '(ign,b) hasAtomicInstructions = b'>
	"For want of somewhere to put the variable that doesn't bloat an instance of the receiver..."
	HasAtomicInstructions := boolean! !

!CogARMv8Compiler methodsFor: 'feature detection' stamp: 'eem 7/7/2021 12:18'!
setInstructionCacheFlushRequired: boolean
	<cmacro: '(ign,b) instructionCacheFlushRequired = b'>
	"For want of somewhere to put the variable that doesn't bloat an instance of the receiver..."
	InstructionCacheFlushRequired := boolean! !

!CogARMv8Compiler methodsFor: 'feature detection' stamp: 'eem 7/7/2021 12:16'!
setInstructionCacheLineLength: n
	<cmacro: '(ign,n) instructionCacheLineLength = n'>
	"For want of somewhere to put the variable that doesn't bloat an instance of the receiver..."
	InstructionCacheLineLength := n! !

!CogARMv8Compiler methodsFor: 'generate machine code' stamp: 'eem 6/18/2015 16:18'!
setLiteralOpcodeIndex: index
	"Hack:  To know how far away a literal is from its referencing instruction we store
	 its opcodeIndex, or -1, if as yet unassigned, in the second operand of the literal."
	<inline: true>
	self assert: opcode = Literal.
	operands at: 2 put: index! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 1/29/2020 11:44'!
setLiteralSize: sizeOfLiteral
	"Set the size field.  The only complication is that in Smalltalk (operands at: 1) may be nil."
	operands
		at: 1
		put: ((operands at: 1) ifNil: [0] ifNotNil: [:operand| operand bitClear: 30])
			+ (sizeOfLiteral << 1)! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 1/23/2020 20:03'!
setsConditionCodesFor: aConditionalJumpOpcode
	<inline: false> "to save Slang from having to be a real compiler (it can't inline switches that return)"
	"Answer if the receiver's opcode sets the condition codes correctly for the given conditional jump opcode.
	 ARMv8 appears not to set condition codes at all in its shift instruction(s), which are aliases for
	 SBFM (signed bit field move) and C6.2.332 UBFM.

	 C6.2.232	SBFM	C6-1170
	 C6.2.332	UBFM	C6-1351"
	<inline: #always>
	^false	"All cases below answer false, so..."
	"^opcode caseOf:
		{	[ArithmeticShiftRightCqR]	->	[false].
			[ArithmeticShiftRightRR]	->	[false].
			[LogicalShiftLeftCqR]		->	[false].
			[LogicalShiftLeftRR]			->	[false].
			[LogicalShiftRightCqR]		->	[false].
			[XorRR]						->	[false].
			[ClzRR]						->	[false]
		}
		otherwise: [self halt: 'unhandled opcode in setsConditionCodesFor:'. false]"! !

!CogARMv8Compiler methodsFor: 'generate machine code' stamp: 'eem 11/20/2019 15:47'!
sizePCDependentInstructionAt: eventualAbsoluteAddress
	"Size a jump and set its address.  The target may be another instruction
	 or an absolute address.  On entry the address inst var holds our virtual
	 address. On exit address is set to eventualAbsoluteAddress, which is
	 where this instruction will be output.  The span of a jump to a following
	 instruction is therefore between that instruction's address and this
	 instruction's address ((which are both still their virtual addresses), but the
	 span of a jump to a preceding instruction or to an absolute address is
	 between that instruction's address (which by now is its eventual absolute
	 address) or absolute address and eventualAbsoluteAddress.

	 ARMv8 is simple; the 26-bit call/jump range (for a signed 28 bit extent, +/- 128Mb) and
	 19 bit conditional branch range  (for a signed 21 bit extent, +/- 1Mb) means no short
	 jumps.  This routine only has to determine the targets of jumps, not determine sizes.

	 This version also deals with out-of-line literals.  If this is the real literal,
	 update the stand-in in literalsManager with the address (because instructions
	 referring to the literal are referring to the stand-in).  If this is annotated with
	 IsObjectReference transfer the annotation to the stand-in, whence it will be
	 transferred to the real literal, simplifying update of literals."

	opcode = AlignmentNops ifTrue:
		[| alignment |
		 address := eventualAbsoluteAddress.
		 alignment := operands at: 0.
		 ^machineCodeSize := (eventualAbsoluteAddress + (alignment - 1) bitAnd: alignment negated)
							   - eventualAbsoluteAddress].
	self assert: (self isJump or: [opcode = Call or: [opcode = CallFull
				or: [dependent notNil and: [dependent opcode = Literal]]]]).
	self isJump ifTrue: [self resolveJumpTarget].
	address := eventualAbsoluteAddress.
	(dependent notNil and: [dependent opcode = Literal]) ifTrue:
		[opcode = Literal ifTrue:
			[dependent address: address].
		 annotation = cogit getIsObjectReference ifTrue:
			[dependent annotation: annotation.
			 annotation := nil]].
	^machineCodeSize := maxSize! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 7/21/2021 09:18'!
st: unitSizeLog2MinusOne rn: baseReg rt: targetReg imm: offset shiftBy12: shiftBy12
	"C6.2.273	STR (immediate)	C6-1239
	 C6.2.275	STRB (immediate)	C6-1244"

	| unitSize |
	unitSize := 1 << unitSizeLog2MinusOne.
	self deny: SP = targetReg.
	self deny: baseReg = targetReg.
	"Unsigned offset, C6-1240"
	(offset \\ unitSize = 0
	 and: [offset / unitSize between: 0 and: 1 << 12 - 1]) ifTrue:
		[^unitSizeLog2MinusOne << 30
		+ (2r11100100 << 22)
		+ (offset << (10 - unitSizeLog2MinusOne))
		+ (baseReg << 5)
		+ targetReg].
	"Unscaled signed 9-bit offset, C6-1244"
	(self asserta: (offset between: -256 and: 255)) ifFalse:
		[self error: 'unhandled immediate offset in store'].
	^unitSizeLog2MinusOne << 30
	  + (2r111000000 << 21)
	  + ((offset bitAnd: 511) << 12)
	  + (baseReg << 5)
	  + targetReg! !

!CogARMv8Compiler methodsFor: 'encoding' stamp: 'eem 11/25/2019 16:14'!
stop
	"generate a HLT; C6.2.92 Arm ARM"
	<inline: true>
	^2r11010100010000000000000000000000 "16rD4400000"! !

!CogARMv8Compiler methodsFor: 'generate machine code - support' stamp: 'eem 1/24/2020 23:21'!
stopsFrom: startAddr to: endAddr
	self assert: endAddr - startAddr + 1 \\ 4 = 0.
	startAddr to: endAddr by: 4 do: 
		[:addr | cogit codeLong32At: addr put: self stop]! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 1/24/2020 23:21'!
storeLiteral32: literal beforeFollowingAddress: followingAddress
	"Rewrite the 32-bit literal in the instruction immediately preceding followingAddress."
	cogit
		codeLong32At:
			(self pcRelativeAddressAt:
				(self instructionAddressBefore:
					((self instructionIsPCRelativeLoad: (self instructionBeforeAddress: followingAddress))
						ifTrue: [followingAddress]
						ifFalse: [followingAddress - 4])))
		put: literal! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 1/24/2020 23:19'!
storeLiteral: literal beforeFollowingAddress: followingAddress
	"Rewrite the literal in the instruction immediately preceding followingAddress."
	cogit
		codeLongAt:
			(self pcRelativeAddressAt:
				(self instructionAddressBefore:
					((self instructionIsPCRelativeLoad: (self instructionBeforeAddress: followingAddress))
						ifTrue: [followingAddress]
						ifFalse: [followingAddress - 4])))
		put: literal! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 1/31/2020 12:44'!
usesNativeSPReg
	"Answer if the code model implemented by the receiver uses the native stack pointer for SPReg.
	 Because ARMv8 mandates 16-bit stack pointer alignment when the SA0 bit is set in SCTLR_EL1,
	 and that this choice is common and immutable on typical ARMv8 OS's, we do not use the native
	 stack pointer in Smalltalk machine code."
	^false! !

!CogARMv8Compiler methodsFor: 'testing' stamp: 'eem 7/21/2021 09:33'!
usesOutOfLineLiteral
	"Answer if the receiver uses an out-of-line literal.  Needs only
	 to work for the opcodes created with gen:literal:operand: et al."

	opcode
		caseOf: {
		[CallFull]		-> [^true].
		[JumpFull]		-> [^true].
		"Arithmetic"
		[AddCqR]		-> [^self isPossiblyShiftableNegatableImm12: (operands at: 0) signedIntFromLong64
								ifTrue: [:ign|false] ifFalse: [true]].
		[AddCqRR]		-> [^self isPossiblyShiftableNegatableImm12: (operands at: 0) signedIntFromLong64
								ifTrue: [:ign|false] ifFalse: [true]].
		[SubCqR]		-> [^self isPossiblyShiftableNegatableImm12: (operands at: 0) signedIntFromLong64
								ifTrue: [:ign|false] ifFalse: [true]].
		[CmpCqR]		-> [^self isPossiblyShiftableNegatableImm12: (operands at: 0) signedIntFromLong64
								ifTrue: [:ign|false] ifFalse: [true]].
		[LoadEffectiveAddressMwrR]
						-> [^self isPossiblyShiftableImm12: (operands at: 0) signedIntFromLong64
								ifTrue: [:shift12|false] ifFalse:[true]].
		[AndCqR]		-> [^self isImmNImmSImmREncodableBitmask: (operands at: 0) ifTrue: [:n :imms :immr|false] ifFalse:[true]].
		[AndCqRR]		-> [^self isImmNImmSImmREncodableBitmask: (operands at: 0) ifTrue: [:n :imms :immr|false] ifFalse:[true]].
		[OrCqRR]		-> [^self isImmNImmSImmREncodableBitmask: (operands at: 0) ifTrue: [:n :imms :immr|false] ifFalse:[true]].
		[OrCqR]		-> [^self isImmNImmSImmREncodableBitmask: (operands at: 0) ifTrue: [:n :imms :immr|false] ifFalse:[true]].
		[TstCqR]		-> [^self isImmNImmSImmREncodableBitmask: (operands at: 0) ifTrue: [:n :imms :immr|false] ifFalse:[true]].
		[XorCqR]		-> [^self isImmNImmSImmREncodableBitmask: (operands at: 0) ifTrue: [:n :imms :immr|false] ifFalse:[true]].
		[AddCwR]		-> [^true].
		[AndCwR]		-> [^true].
		[CmpCwR]		-> [^true].
		[CmpC32R]		-> [^true].
		[OrCwR]		-> [^true].
		[SubCwR]		-> [^true].
		[XorCwR]		-> [^true].
		"Data Movement"						
		[MoveCqR]		-> [self isPossiblyShiftableImm12: (operands at: 0)
								ifTrue: [:shift12| ^false] ifFalse:[].
							^self isImmNImmSImmREncodableBitmask: (operands at: 0)
								ifTrue: [:n :imms :immr|false] ifFalse: [true]].
		[MoveC32R]	-> [^true].
		[MoveCwR]		-> [^(self inCurrentCompilation: (operands at: 0)) not].
		[MoveAwR]		-> [^(self isAddressRelativeToVarBase: (operands at: 0)) not].
		[MoveRAw]		-> [^(self isAddressRelativeToVarBase: (operands at: 1)) not].
		[MoveAbR]		-> [^(self isAddressRelativeToVarBase: (operands at: 0)) not].
		[MoveRAb]		-> [^(self isAddressRelativeToVarBase: (operands at: 1)) not].
		[MoveMwrR]	-> [^(self isImm12orImm9offset: (operands at: 0)) not].
		[MoveMbrR]	-> [^(self isImm12orImm9offset: (operands at: 0)) not].
		[MoveM16rR]	-> [^(self isImm12orImm9offset: (operands at: 0)) not].
		[MoveRMwr]	-> [^self isPossiblyShiftableNegatableImm12: (operands at: 1) ifTrue: [:shift12| false] ifFalse: [true]].
		[MoveRMbr]	-> [^self isPossiblyShiftableNegatableImm12: (operands at: 1) ifTrue: [:shift12| false] ifFalse: [true]].
		[MoveRM16r]	-> [^self isPossiblyShiftableNegatableImm12: (operands at: 1) ifTrue: [:shift12| false] ifFalse: [true]].
		[MoveRdM64r]	-> [^(self isUnsigned12BitMultipleOf8: (operands at: 1)) not].
		[MoveM64rRd]	-> [^(self isUnsigned12BitMultipleOf8: (operands at: 0)) not].
		[PushCw]		-> [^(self inCurrentCompilation: (operands at: 0)) not].
		[PushCq]		-> [^((operands at: 0) between: -256 and: 255) not].
		[PrefetchAw] 	-> [^(self isAddressRelativeToVarBase: (operands at: 0)) not].
		}
		otherwise: [self assert: false].
	^false "to keep C compiler quiet"
! !

!CogARMv8Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/29/2019 11:29'!
zoneCallsAreRelative
	"Answer if Call and JumpLong are relative and hence need to take the caller's
	 relocation delta into account during code compaction, rather than just the
	 callee's delta."
	^true! !

CogARMv8Compiler subclass: #CogARMv8CompilerForTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!CogARMv8CompilerForTests class methodsFor: 'accessing' stamp: 'eem 12/18/2019 23:59'!
dataRegistersWithAccessorsGiven: aProcessorSimulator do: aTrinaryBlock
	"We can't use RISCTempReg, because some opcodes may be encoded as multiple instructions using RISCTempReg."
	(0 to: 31) withIndexDo:
		[:reg :i|
		(self isRISCTempRegister: reg) ifFalse:
			[aTrinaryBlock
				value: reg
				value: (aProcessorSimulator registerStateGetters at: i)
				value: (aProcessorSimulator registerStateSetters at: i)]]! !

!CogARMv8CompilerForTests class methodsFor: 'testing' stamp: 'eem 12/18/2019 22:47'!
isRISCTempRegister: reg
	"For tests to filter-out bogus values left in the RISCTempRegister, if any."
	^reg = R28! !

!CogARMv8CompilerForTests methodsFor: 'generating machine code' stamp: 'nice 12/31/2020 13:15'!
concretizeAt: actualAddress
	"Override to check maxSize and machineCodeSize"

	| maxAddress nextAddress |
	maxSize ifNil: [maxSize := self computeMaximumSize].
	maxAddress := actualAddress + maxSize.
	nextAddress := super concretizeAt: actualAddress.
	self assert: (maxSize notNil
				and: [self isPCDependent
						ifTrue: [maxAddress >= nextAddress]
						ifFalse: [maxAddress = nextAddress]]).
	^nextAddress! !

!CogAbstractInstruction class methodsFor: 'translation' stamp: 'eem 4/21/2015 22:52'!
ISA
	"Answer the name of the ISA the receiver's subclass implements."
	^self subclassResponsibility! !

!CogAbstractInstruction class methodsFor: 'simulation only' stamp: 'eem 3/14/2017 09:54'!
byteSizeForSimulator: aVMClass
	"Answer an approximation of the byte size of an AbstractInstruction struct.
	 This is for estimating the alloca in allocateOpcodes:bytecodes:ifFail:
	 self allSubclasses collect: [:ea| {ea. ea byteSizeForSimulator: ea basicNew}]"
	| ptrsize |
	ptrsize := self == CogAbstractInstruction ifTrue: [BytesPerWord] ifFalse: [self wordSize].
	^CogCompilerClass instSize
		- 4 "cogit, objectMemory, bcpc, machineCode"
		- 4 + 1
	"opcode machineCodeSize maxSize annotation are all bytes -> 1 long" * ptrsize
	+ CogCompilerClass basicNew machineCodeBytes
		roundTo: ptrsize! !

!CogAbstractInstruction class methodsFor: 'translation' stamp: 'eem 6/15/2015 17:46'!
defaultCompilerClass
	^self! !

!CogAbstractInstruction class methodsFor: 'class initialization' stamp: 'eem 7/7/2021 17:34'!
extraClassVariableNames
	^#()! !

!CogAbstractInstruction class methodsFor: 'translation' stamp: 'eem 3/11/2017 13:00'!
filteredInstVarNames
	"Eliminate bcpc, which is development-time only."
	^super filteredInstVarNames copyWithout: 'bcpc'! !

!CogAbstractInstruction class methodsFor: 'instance creation' stamp: 'eem 12/20/2008 16:29'!
for: aCogit
	^self new cogit: aCogit! !

!CogAbstractInstruction class methodsFor: 'translation' stamp: 'eem 4/21/2015 22:40'!
identifyingPredefinedMacros
	"Answer the predefined macros that identify the processors a subclass handles, if any.
	 If the subclass isn't yet ready for production (a work in progress) simply answer nil."
	^nil! !

!CogAbstractInstruction class methodsFor: 'translation' stamp: 'eem 3/20/2017 08:52'!
ifTranslateableAddWithOptionsTo: aCollection
	(self wordSize = Cogit objectMemoryClass wordSize
	 and: [self identifyingPredefinedMacros notNil]) ifTrue:
		[aCollection add: {self. {#ISA. self ISA}}]! !

!CogAbstractInstruction class methodsFor: 'class initialization' stamp: 'eem 10/4/2013 14:23'!
initialize
	NumOperands := 3! !

!CogAbstractInstruction class methodsFor: 'class initialization' stamp: 'eem 2/4/2020 09:28'!
initializeAbstractRegisters
	"Assign the abstract registers with the identities/indices of the relevant concrete registers,
	 and assign CallerSavedRegisterMask appropriately.
	 First set all abstract registers to #undefined via CogAbstractRegisters initialize, and then,
	 each subclasses assigns the subset they choose with values of specific concrete registers."

	CogAbstractRegisters initialize.
	
	CogCompilerClass ifNotNil:
		[(CogCompilerClass withAllSuperclasses copyUpTo: CogAbstractInstruction) reverseDo:
			[:compilerClass| compilerClass initialize; initializeAbstractRegisters]].

	NativeSPReg == #undefined ifTrue:
		[NativeSPReg := SPReg].
	(ABIResultReg == #undefined and: [TempReg = 0]) ifTrue:
		[ABIResultReg := TempReg].

	self assert: Arg0Reg isInteger.
	self assert: ClassReg isInteger.
	self assert: ReceiverResultReg isInteger.
	self assert: SendNumArgsReg isInteger.
	self assert: TempReg isInteger.
	self assert: FPReg isInteger.
	self assert: SPReg isInteger.

	self assert: ABICalleeSavedRegisterMask isInteger.
	self assert: ABICallerSavedRegisterMask isInteger.
	self assert: CallerSavedRegisterMask isInteger.
	self assert: ABIResultReg isInteger.

	self assert: (VarBaseReg = #undefined or: [SimpleStackBasedCogit basicNew register: VarBaseReg isInMask: ABICalleeSavedRegisterMask])! !

!CogAbstractInstruction class methodsFor: 'class initialization' stamp: 'eem 7/7/2021 17:34'!
initializeSpecificOpcodes: opcodeSymbolSequence in: initializeMethod
	"Declare as class variables, the opcodes in opcodeSymbolSequence.
	 Assign values to them from LastRTLOpcode on.  Undeclare any obsolete
	 class vars.  The assumption is that initializeMethod defines all class vars
	 in the class. This method should be used by subclasses wishing to declare
	 their own specific opcodes."
	^self initializeSpecificOpcodes: opcodeSymbolSequence in: initializeMethod extraClassVarNames: self extraClassVariableNames! !

!CogAbstractInstruction class methodsFor: 'class initialization' stamp: 'eem 10/25/2020 16:40'!
initializeSpecificOpcodes: opcodeSymbolSequence in: initializeMethod extraClassVarNames: extraClassVarNames
	"Declare as class variables, the opcodes in opcodeSymbolSequence.
	 Assign values to them from LastRTLOpcode on.  Undeclare any obsolete
	 class vars.  The assumption is that initializeMethod defines all class vars
	 in the class. This method should be used by subclasses wishing to declare
	 their own specific opcodes."
	| pool classVariablesDefinedInMethod |
	self assert: self ~~ CogAbstractInstruction.
	pool := initializeMethod methodClass classPool.
	LastRTLCode ifNil:
		[CogRTLOpcodes initialize].
	classVariablesDefinedInMethod := Set new.
	initializeMethod allLiteralsDo:
		[:lit|
		(lit isSymbol and: [pool includesKey: lit])
			ifTrue: [classVariablesDefinedInMethod add: lit]
			ifFalse:
				[(lit isVariableBinding and: [pool includesKey:lit key]) ifTrue:
					[classVariablesDefinedInMethod add: lit key]]].
	"Undeclare any class var not defined in opcodeSymbolSequence or by the method."
	(pool keys reject: [:k| (opcodeSymbolSequence includes: k)
						or: [(classVariablesDefinedInMethod includes: k)
						or: [extraClassVarNames includes: k]]]) do:
		[:k|
		Undeclared declare: k from: pool].
	extraClassVarNames do:
		[:classVarName|
		(pool includesKey: classVarName) ifFalse:
			[pool
				declare: classVarName from: Undeclared;
				at: classVarName put: nil]].
	"Declare opcodeSymbolSequence's elements from LastRTLCode on up."
	opcodeSymbolSequence withIndexDo:
		[:classVarName :value|
		pool
			declare: classVarName from: Undeclared;
			at: classVarName put: value + LastRTLCode - 1]! !

!CogAbstractInstruction class methodsFor: 'translation' stamp: 'eem 10/25/2020 08:55'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"Enumerate aBinaryBlock with the names and C type strings for the inst vars to include in an AbstractInstruction struct."
	"((CogAbstractInstruction withAllSubclasses reject:
		[:c|
		 (c class lookupSelector: #wordSize) isSubclassResponsibility
		 or: [((c lookupSelector: #machineCodeBytes) ifNil: [false] ifNotNil: [:cm| cm isSubclassResponsibility])
		 or: [c name includesSubstring: 'ForTests']]])
			sort: [:a :b| a name <= b name]) do:
			[:c| Transcript cr; cr; print: c; cr. c printTypedefOn: Transcript]"

	"The first four fields of an instruction are byte variables. opcode, machineCodeSize, maxSize, annotation.
	 On 64 bits we can get better packing if machine code follows these four..." 
	self filteredInstVarNames do:
		[:ivn|
		 aBinaryBlock
			value: ivn
			value: (ivn caseOf: {
						['address']			-> [#usqInt]. "usqInt is always large enough to contain a pointer; we do not need to use usqIntptr_t"
						['machineCode']	-> [self machineCodeDeclaration].
						['operands']		-> [{#usqInt. '[', NumOperands printString, ']'}].
						['dependent']		-> ['struct _AbstractInstruction *']}
					otherwise:
						[#'unsigned char'])]! !

!CogAbstractInstruction class methodsFor: 'testing' stamp: 'eem 8/8/2012 13:11'!
isAbstract
	^self == CogAbstractInstruction! !

!CogAbstractInstruction class methodsFor: 'translation' stamp: 'eem 10/19/2009 12:36'!
isAccessor: aSelector
	^(#(cogit coInterpreter objectMemory) includes: aSelector) not
	   and: [super isAccessor: aSelector]! !

!CogAbstractInstruction class methodsFor: 'testing' stamp: 'eem 12/18/2019 22:41'!
isRISCTempRegister: reg
	"For tests to filter-out bogus values left in the RISCTempRegister, if any.
	 N.B. DO NOT USE ABSTRACT REGISTER NAMES SUCH AS RISCTempReg IN IMPLEMENTING THIS!!!!"
	^self subclassResponsibility! !

!CogAbstractInstruction class methodsFor: 'accessing class hierarchy' stamp: 'eem 6/8/2015 16:51'!
literalsManagerClass
	"Answer the manager that generates in-line literals;
	 subclasses that use out-of-line literals override."
	^InLineLiteralsManager! !

!CogAbstractInstruction class methodsFor: 'printing' stamp: 'eem 9/22/2015 11:10'!
localNameForOpcode: opcode
	^(self classPool keyAtValue: opcode ifAbsent: nil)
		ifNotNil: [:opcodeName| opcodeName]
		ifNil: [self == CogAbstractInstruction
				ifTrue: [opcode printString]
				ifFalse: [superclass localNameForOpcode: opcode]]! !

!CogAbstractInstruction class methodsFor: 'translation' stamp: 'eem 3/11/2017 12:46'!
machineCodeDeclaration
	"Answer a dummy declaration.  Subclasses will override to provide the real one."
	^#(#'unsigned char' '[4]')! !

!CogAbstractInstruction class methodsFor: 'translation' stamp: 'eem 4/21/2015 22:54'!
moduleName
	"CogAbstractInstruction subclasses collect: [:ea| ea moduleName]"
	^'cogit', self ISA! !

!CogAbstractInstruction class methodsFor: 'debug printing' stamp: 'eem 9/22/2015 11:07'!
nameForOpcode: opcode "<Integer>"
	^(CogRTLOpcodes nameForOpcode: opcode)
		ifNotNil:
			[:opcodeName| opcodeName]
		ifNil:
			[self localNameForOpcode: opcode]! !

!CogAbstractInstruction class methodsFor: 'trampoline support' stamp: 'eem 1/21/2020 11:19'!
numTrampolines
	"By default back ends don't define any trampolines.  But some subclassses
	 may want to, e.g. for cache flushing.  They will override as appropriate."
	^0! !

!CogAbstractInstruction class methodsFor: 'debug printing' stamp: 'eem 12/26/2015 02:26'!
printFormatForOpcodeName: opcodeName
	"Answer a sequence of $r, $f or nil for the operands in the opcode, used for printing, where
	 r => integer register, f => floating point register, and nil => numeric or address operand.
	 Subclasses can override to provide a format string for their own private opcodes."
	^#()! !

!CogAbstractInstruction class methodsFor: 'register management' stamp: 'eem 05/30/2016 14:03'!
registerMaskFor: reg
	^Cogit basicNew registerMaskFor: reg! !

!CogAbstractInstruction class methodsFor: 'register management' stamp: 'eem 05/30/2016 14:03'!
registerMaskFor: reg1 and: reg2
	^Cogit basicNew registerMaskFor: reg1 and: reg2! !

!CogAbstractInstruction class methodsFor: 'register management' stamp: 'eem 05/30/2016 14:03'!
registerMaskFor: reg1 and: reg2 and: reg3
	^Cogit basicNew registerMaskFor: reg1 and: reg2 and: reg3! !

!CogAbstractInstruction class methodsFor: 'register management' stamp: 'eem 05/30/2016 14:03'!
registerMaskFor: reg1 and: reg2 and: reg3 and: reg4
	^Cogit basicNew registerMaskFor: reg1 and: reg2 and: reg3 and: reg4! !

!CogAbstractInstruction class methodsFor: 'register management' stamp: 'eem 05/30/2016 14:03'!
registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5
	^Cogit basicNew registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5! !

!CogAbstractInstruction class methodsFor: 'register management' stamp: 'eem 05/30/2016 14:03'!
registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6
	^Cogit basicNew registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6! !

!CogAbstractInstruction class methodsFor: 'register management' stamp: 'eem 05/30/2016 14:03'!
registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6 and: reg7
	^Cogit basicNew registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6 and: reg7! !

!CogAbstractInstruction class methodsFor: 'register management' stamp: 'eem 05/30/2016 14:03'!
registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6 and: reg7 and: reg8
	^Cogit basicNew registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6 and: reg7 and: reg8! !

!CogAbstractInstruction class methodsFor: 'register management' stamp: 'eem 05/30/2016 14:03'!
registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6 and: reg7 and: reg8 and: reg9
	^Cogit basicNew registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6 and: reg7 and: reg8 and: reg9! !

!CogAbstractInstruction class methodsFor: 'register management' stamp: 'eem 05/30/2016 14:03'!
registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6 and: reg7 and: reg8 and: reg9 and: reg10
	^Cogit basicNew registerMaskFor: reg1 and: reg2 and: reg3 and: reg4 and: reg5 and: reg6 and: reg7 and: reg8 and: reg9 and: reg10! !

!CogAbstractInstruction class methodsFor: 'translation' stamp: 'eem 1/10/2016 08:46'!
requiredMethodNames: options
	^self selectors reject:
		[:s| | m |
		(self isAccessor: s)
		or: [((m := self compiledMethodAt: s) isQuick and: [m pragmas isEmpty])
		or: [(m pragmaAt: #doNotGenerate) notNil
		or: [(m pragmaAt: #inline:) notNil and: [(m pragmaAt: #inline:) arguments first == true]]]]]! !

!CogAbstractInstruction class methodsFor: 'verification' stamp: 'eem 1/24/2012 13:14'!
specificOpcodes
	^self subclassResponsibility! !

!CogAbstractInstruction class methodsFor: 'translation' stamp: 'eem 8/8/2012 12:58'!
structTypeName
	^'AbstractInstruction'! !

!CogAbstractInstruction class methodsFor: 'translation' stamp: 'eem 3/20/2017 09:12'!
translateableSubclassesAndOptions
	"CogAbstractInstruction translateableSubclassesAndOptions"
	| translateableSubclassesAndOptions |
	translateableSubclassesAndOptions := OrderedCollection new.
	self subclasses do:
		[:subclass|
		subclass ifTranslateableAddWithOptionsTo: translateableSubclassesAndOptions].
	^translateableSubclassesAndOptions sort:
		[:a :b|
		a first name < b first name
		or: [a first name = b first name "If equal, sort on the ABI"
			and: [a last last < b last last]]]! !

!CogAbstractInstruction class methodsFor: 'translation' stamp: 'eem 2/11/2009 10:34'!
typeForSelf
	"Answer the type to give self if appropriate, or nil if not.
	 In C all subclasses are mapped to the AbstractInstruction type."
	^#'AbstractInstruction *'! !

!CogAbstractInstruction class methodsFor: 'verification' stamp: 'eem 1/24/2012 13:13'!
unimplementedOpcodes
	"Check computeMaximumSize and dispatchConcretize for unimplemented opcodes."
	"self subclasses collect: [:compiler| {compiler. compiler unimplementedOpcodes}]"
	| opcodes cmsLiterals dcLiterals |
	opcodes := CogRTLOpcodes opcodes, self specificOpcodes.
	cmsLiterals := (self >> #computeMaximumSize) literals
						select: [:l| l isVariableBinding]
						thenCollect: [:l| l key].
	dcLiterals := (self >> #dispatchConcretize) literals
						select: [:l| l isVariableBinding]
						thenCollect: [:l| l key].
	^Dictionary new
		at: #computeMaximumSize put: (opcodes reject: [:opcode| cmsLiterals includes: opcode]);
		at: #dispatchConcretize put: (opcodes reject: [:opcode| dcLiterals includes: opcode]);
		yourself! !

!CogAbstractInstruction class methodsFor: 'verification' stamp: 'eem 1/24/2012 14:46'!
usedUnimplementedOpcodes
	"Check for uses of unimplemented opcodes"
	"self subclasses collect: [:compiler| {compiler. compiler usedUnimplementedOpcodes}]"
	| genericUnimplementedOpcodeBindings specificUnimplementedOpcodeBindings |
	genericUnimplementedOpcodeBindings := Set new.
	specificUnimplementedOpcodeBindings := Set new.
	self unimplementedOpcodes do:
		[:arrayOfOpcodes|
		arrayOfOpcodes do:
			[:opcode|
			(self bindingOf: opcode)
				ifNotNil: [:b| specificUnimplementedOpcodeBindings add: b]
				ifNil: [genericUnimplementedOpcodeBindings add: (CogRTLOpcodes bindingOf: opcode)]]].
	^ { (SystemNavigation new allCallsOn: genericUnimplementedOpcodeBindings localToPackage: #VMMaker)
			inject: Set new
			into:  [:them :methodRef| "These should be in Cogit's abstract instructions category"
				them
					addAll: (SystemNavigation new allCallsOn: methodRef method selector localToPackage: #VMMaker);
					yourself].
		SystemNavigation new allCallsOn: specificUnimplementedOpcodeBindings localToPackage: #VMMaker }! !

!CogAbstractInstruction class methodsFor: 'translation' stamp: 'eem 10/24/2020 09:14'!
wordSize
	"Answer either 4 or 8 depending on the processor's basic architecture."
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'comparing' stamp: 'eem 11/29/2010 17:58'!
<= anAbstractInstruction
	"Support for Cogit>>abstractInstruction:follows: and CogBytecodeFixup>targetInstruction testing:"
	<doNotGenerate>
	| opcodesArray |
	anAbstractInstruction isInteger ifTrue:
		[self assert: anAbstractInstruction < 16.
		 ^false]. 
	opcodesArray := cogit abstractOpcodes object.
	^(opcodesArray identityIndexOf: self) <= (opcodesArray identityIndexOf: anAbstractInstruction)! !

!CogAbstractInstruction methodsFor: 'comparing' stamp: 'eem 11/25/2010 13:54'!
> anAbstractInstruction
	"Support for Cogit>>abstractInstruction:follows: and CogBytecodeFixup>targetInstruction testing:"
	<doNotGenerate>
	| opcodesArray |
	anAbstractInstruction isInteger ifTrue:
		[self assert: anAbstractInstruction < 16.
		 ^true]. 
	opcodesArray := cogit abstractOpcodes object.
	^(opcodesArray identityIndexOf: self) > (opcodesArray identityIndexOf: anAbstractInstruction)! !

!CogAbstractInstruction methodsFor: 'comparing' stamp: 'eem 11/25/2010 14:49'!
>= anAbstractInstruction
	"Support for Cogit>>abstractInstruction:follows: and CogBytecodeFixup>targetInstruction testing:"
	<doNotGenerate>
	| opcodesArray |
	anAbstractInstruction isInteger ifTrue:
		[self assert: anAbstractInstruction < 16.
		 ^true]. 
	opcodesArray := cogit abstractOpcodes object.
	^(opcodesArray identityIndexOf: self) >= (opcodesArray identityIndexOf: anAbstractInstruction)! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 12/10/2015 08:32'!
addDependent: anInstruction
	<var: #anInstruction type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	dependent ifNotNil:
		[anInstruction dependent: dependent].
	^dependent := anInstruction! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 2/11/2009 22:19'!
address
	^address! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 2/11/2009 22:19'!
address: theAddress
	^address := theAddress! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 6/10/2015 16:33'!
annotation
	^annotation! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 6/19/2015 08:46'!
annotation: aByte
	self assert: (annotation isNil or: [annotation = aByte]).
	^annotation := aByte! !

!CogAbstractInstruction methodsFor: 'coercion' stamp: 'eem 2/14/2009 18:09'!
asInteger
	<doNotGenerate>
	^self! !

!CogAbstractInstruction methodsFor: 'coercion' stamp: 'nice 7/8/2016 02:32'!
asIntegerPtr
	<doNotGenerate>
	^self! !

!CogAbstractInstruction methodsFor: 'coercion' stamp: 'eem 2/15/2009 16:36'!
asUnsignedInteger
	<doNotGenerate>
	^self! !

!CogAbstractInstruction methodsFor: 'coercion' stamp: 'nice 7/8/2016 02:30'!
asUnsignedIntegerPtr
	<doNotGenerate>
	^self! !

!CogAbstractInstruction methodsFor: 'register allocation' stamp: 'RonieSalgado 4/29/2016 15:37'!
availableFloatRegisterOrNoneFor: liveRegsMask
	"Answer an unused abstract register in the liveRegMask.
	 Subclasses with more registers can override to answer them."
	<returnTypeC: #sqInt>
	(liveRegsMask anyMask: (cogit registerMaskFor: DPFPReg0)) ifFalse:
		[^DPFPReg0].
	(liveRegsMask anyMask: (cogit registerMaskFor: DPFPReg1)) ifFalse:
		[^DPFPReg1].
	(liveRegsMask anyMask: (cogit registerMaskFor: DPFPReg2)) ifFalse:
		[^DPFPReg2].
	(liveRegsMask anyMask: (cogit registerMaskFor: DPFPReg3)) ifFalse:
		[^DPFPReg3].
	(liveRegsMask anyMask: (cogit registerMaskFor: DPFPReg4)) ifFalse:
		[^DPFPReg4].
	(liveRegsMask anyMask: (cogit registerMaskFor: DPFPReg5)) ifFalse:
		[^DPFPReg5].
	(liveRegsMask anyMask: (cogit registerMaskFor: DPFPReg6)) ifFalse:
		[^DPFPReg6].
	(liveRegsMask anyMask: (cogit registerMaskFor: DPFPReg7)) ifFalse:
		[^DPFPReg7].

	^ NoReg! !

!CogAbstractInstruction methodsFor: 'register allocation' stamp: 'cb 4/16/2016 14:59'!
availableRegisterOrNoneFor: liveRegsMask
	"Answer an unused abstract register in the liveRegMask.
	 Subclasses with more registers can override to answer them.
	 N.B. Do /not/ allocate TempReg."
	<returnTypeC: #sqInt>
	(cogit register: Arg1Reg isInMask: liveRegsMask) ifFalse:
		[^Arg1Reg].
	(cogit register: Arg0Reg isInMask: liveRegsMask) ifFalse:
		[^Arg0Reg].
	(cogit register: SendNumArgsReg isInMask: liveRegsMask) ifFalse:
		[^SendNumArgsReg].
	(cogit register: ClassReg isInMask: liveRegsMask) ifFalse:
		[^ClassReg].
	(cogit register: ReceiverResultReg isInMask: liveRegsMask) ifFalse:
		[^ReceiverResultReg].
	^NoReg! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 12/9/2010 17:15'!
bcpc
	^bcpc! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 12/9/2010 17:15'!
bcpc: theBcpc
	^bcpc := theBcpc! !

!CogAbstractInstruction methodsFor: 'comparing' stamp: 'eem 11/25/2010 13:48'!
between: min and: max
	"Support for CogBytecodeFixup>targetInstruction testing:"
	<doNotGenerate>
	| opcodesArray |
	(min isInteger and: [max isInteger]) ifTrue:
		[^false].
	opcodesArray := cogit abstractOpcodes object.
	^(opcodesArray identityIndexOf: self)
		between: (opcodesArray identityIndexOf: min)
		and: (opcodesArray identityIndexOf: max)! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 11/3/2015 12:19:20'!
byteReadsZeroExtend
	"Answer if a byte read, via MoveAbR, MoveMbrR, or MoveXbrRR zero-extends
	 into the full register, or merely affects the least significant 8 bits of the
	 the register.  By default the code generator assumes that byte reads
	 to not zero extend.  Note that byte reads /must not/ sign extend."
	^false! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 7/8/2021 14:47'!
callFullInstructionByteSize
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 2/13/2009 12:29'!
callInstructionByteSize
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 3/28/2015 12:48'!
canDivQuoRem
	<inline: true>
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 3/28/2015 12:48'!
canMulRR
	<inline: true>
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 7/4/2020 11:31'!
canMulRRDetectOverflow
	"Hook for ARMv8 to implement overflow checking."
	<inline: true>
	^true! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 6/1/2016 10:03'!
canPushPopMultipleRegisters
	<inline: true>
	^false! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 6/21/2018 12:11'!
canSignExtend
	"Answer if the processor provides native sign extension instructions.  If it does it must be able to generate
		SignExtend8RR SignExtend16RR
	 and if 64-bits, SignExtend32RR."
	<inline: true>
	^false! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 6/21/2018 12:29'!
canZeroExtend
	"Answer if the processor provides native zero extension instructions.  If it does it must be able to generate
		ZeroExtend8RR ZeroExtend16RR
	 and if 64-bits, ZeroExtend32RR."
	<inline: true>
	^false! !

!CogAbstractInstruction methodsFor: 'simulation' stamp: 'eem 7/8/2021 15:08'!
ceShortCutTraceBlockActivation: aProcessorSimulationTrap
	<doNotGenerate>
	self flag: 'just a placeholder; see handleABICallOrJumpSimulationTrap:evaluable: & simulatedTrampolineFor:'! !

!CogAbstractInstruction methodsFor: 'simulation' stamp: 'eem 7/8/2021 15:08'!
ceShortCutTraceLinkedSend: aProcessorSimulationTrap
	<doNotGenerate>
	self flag: 'just a placeholder; see handleABICallOrJumpSimulationTrap:evaluable: & simulatedTrampolineFor:'! !

!CogAbstractInstruction methodsFor: 'simulation' stamp: 'eem 7/8/2021 15:08'!
ceShortCutTraceStore: aProcessorSimulationTrap
	<doNotGenerate>
	self flag: 'just a placeholder; see handleABICallOrJumpSimulationTrap:evaluable: & simulatedTrampolineFor:'! !

!CogAbstractInstruction methodsFor: 'initialization' stamp: 'eem 6/18/2015 15:52'!
cloneLiteralFrom: existingLiteral
	"For out-of-line literal support, clone a literal from a literal."
	<var: 'existingLiteral' type: #'AbstractInstruction *'>
	self assert: (existingLiteral opcode = Literal and: [dependent isNil and: [address isNil]]).
	opcode := Literal.
	annotation := existingLiteral annotation.
	operands
		at: 0 put: (existingLiteral operands at: 0);
		at: 1 put: (existingLiteral operands at: 1);
		at: 2 put: (existingLiteral operands at: 2)! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 11/20/2015 16:48'!
cmpC32RTempByteSize
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 4/21/2015 09:17'!
codeGranularity
	"Answer the size in bytes of a unit of machine code."
	<inline: true>
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'initialization' stamp: 'eem 10/13/2009 12:44'!
cogit: aCogit
	<doNotGenerate>
	cogit := aCogit.
	objectMemory := aCogit objectMemory! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'eem 12/8/2019 21:47'!
computeJumpTargetOffsetPlus: anPCOffset
	<inline: true> "Since it's an extraction from other methods."
	| jumpTarget |
	<var: #jumpTarget type: #'AbstractInstruction *'>
	jumpTarget := self jumpTargetAddress.
	^objectMemory wordSize = 8
		ifTrue: [jumpTarget signedIntFromLong64 - (address + anPCOffset) signedIntFromLong64]
		ifFalse: [jumpTarget signedIntFromLong - (address + anPCOffset) signedIntFromLong]! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'eem 12/8/2019 14:43'!
concretizeAt: actualAddress
	"Generate concrete machine code for the instruction at actualAddress,
	 setting machineCodeSize, and answer the following address."

	address := actualAddress.
	machineCodeSize := self dispatchConcretize.
	self assert: (maxSize = nil or: [maxSize >= machineCodeSize]).
	^actualAddress + machineCodeSize! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'eem 1/8/2011 20:50'!
concretizeFill32
	"fill with operand 0 according to the processor's endianness"
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'eem 12/8/2019 15:23'!
concretizeLabel
	<inline: true>
	<var: #dependentChain type: #'AbstractInstruction *'>
	| dependentChain |
	dependentChain := dependent.
	[dependentChain isNil] whileFalse:
		[dependentChain updateLabel: self.
		 dependentChain := dependentChain dependent].
	^0! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 2/11/2009 22:15'!
dependent
	^dependent! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 2/11/2009 22:16'!
dependent: anInstruction
	^dependent := anInstruction! !

!CogAbstractInstruction methodsFor: 'initialization' stamp: 'eem 10/29/2020 20:21'!
detectFeatures
	"This is a null default for code to detect features specific to the processor
	 which may not be present on all versions of the processor and hence must
	 be checked at runtime. Subclasses override if necessary."
	<inline: #always>! !

!CogAbstractInstruction methodsFor: 'full transfer run-time support' stamp: 'eem 11/4/2015 14:30:01'!
disassemble
	<doNotGenerate>
	self disassembleOn: Transcript! !

!CogAbstractInstruction methodsFor: 'full transfer run-time support' stamp: 'eem 11/4/2015 14:30:05'!
disassembleOn: aStream
	<doNotGenerate>
	cogit processor
		disassembleFrom: 0
		to: machineCodeSize - 1
		in: machineCode object
		on: aStream! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 2/3/2020 20:47'!
flushDCacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	"If there is a dual mapped code zone (the normal zone but marked with read/execute, and a
	 read/write zone codeToDataDelta bytes away) then flush the data cache for the corresponding
	 range in the read/write zone and invalidate the data cache for the read/execute zone."
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 2/3/2020 19:33'!
flushICacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	"Flush the instruction cache from (startAddress to endAddress].
	 If there is a dual mapped code zone (the normal zone but marked with read/execute, and a
	 read/write zone codeToDataDelta bytes away) then also flush the data cache for the corresp-
	 onding range in the read/write zone and invalidate the data cache for the read/execute zone."
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'abi' stamp: 'eem 4/20/2015 18:21'!
fullCallsAreRelative
	"Answer if CallFull and/or JumpFull are relative and hence need relocating on method
	 compation. If so, they are annotated with IsRelativeCall in methods and relocated in
	 relocateIfCallOrMethodReference:mcpc:delta:"
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'abi' stamp: 'eem 12/28/2019 19:49'!
genAlignCStackSavingRegisters: regMask numArgs: numArgs wordAlignment: alignment
	<inline: true>
	| regMaskCopy numRegsPushed wordsPushedModAlignment delta |
	<var: 'regMaskCopy' type: #usqInt>
	regMaskCopy := regMask asUnsignedInteger.
	numRegsPushed := 0.
	[regMaskCopy ~= 0] whileTrue:
		[numRegsPushed := numRegsPushed + (regMaskCopy bitAnd: 1).
		 regMaskCopy := regMaskCopy bitShift: -1].
	(numRegsPushed = 0
	 and: [self numIntRegArgs >= numArgs]) ifTrue:
		[^0].
	wordsPushedModAlignment := numRegsPushed + (numArgs - self numIntRegArgs max: 0) \\ alignment.
	wordsPushedModAlignment ~= 0 ifTrue:
		[delta := alignment - wordsPushedModAlignment.
		 cogit SubCq: delta * objectMemory wordSize R: SPReg].
	^0! !

!CogAbstractInstruction methodsFor: 'abstract instructions' stamp: 'nice 1/26/2020 00:36'!
genHighBitAlternativeIn: destReg ofSmallIntegerOopWithSingleTagBit: srcReg
	"Use an alternative - if any for generating highBit of SmallInteger oop.
	Default implementation is no-op - there is no universal alternative to CLZ.
	Some target architecture might offer more...
	For example, IA32 can use the BSR instruction."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^0! !

!CogAbstractInstruction methodsFor: 'abstract instructions' stamp: 'eem 1/23/2020 15:57'!
genHighBitClzIn: destReg ofSmallIntegerOopWithSingleTagBit: srcReg
	"Use CLZ instruction for generating highBit of SmallInteger oop.
	It is sender responsibility to make sure that such CLZ instruction is available."
	<inline: true>
	| jumpNegativeReceiver |
	<var: #jumpNegativeReceiver type: #'AbstractInstruction *'>
	cogit ClzR: srcReg R: destReg.
	(cogit lastOpcode setsConditionCodesFor: JumpZero) ifFalse:
		[cogit CmpCq: 0 R: destReg]. "N.B. FLAGS := destReg - 0"
	jumpNegativeReceiver := cogit JumpZero: 0.
	"Note the nice bit trick below:
	highBit_1based_of_small_int_value = (BytesPerWord * 8) - leadingZeroCout_of_oop - 1 toAccountForTagBit.
	This is like 2 complements (- reg - 1) on (BytesPerWord * 8) log2 bits, or exactly a bit invert operation..."
	cogit XorCw: BytesPerWord * 8 - 1 R: destReg.
	^jumpNegativeReceiver! !

!CogAbstractInstruction methodsFor: 'abstract instructions' stamp: 'nice 1/24/2020 22:27'!
genHighBitIn: destReg ofSmallIntegerOopWithSingleTagBit: srcReg
	"Generate code for storing in destReg the 1-based highBit of srcReg.
	 Assume that srcReg contains a SmallInteger Oop with a single tag bit set to 1.
	 Sender should preprocess Oop when cog representation use numSmallIntegerTagBits > 1.
	 Return the jump instruction necessary for handling case of negative integer value.
	 Return null pointer if the abstract highBit operation is not implemented."
	<inline: true>
	self hasLZCNTInstructions
		ifTrue: [^self genHighBitClzIn: destReg ofSmallIntegerOopWithSingleTagBit: srcReg]
		ifFalse: [^self genHighBitAlternativeIn: destReg ofSmallIntegerOopWithSingleTagBit: srcReg]! !

!CogAbstractInstruction methodsFor: 'abstract instructions' stamp: 'eem 4/17/2009 15:47'!
genJumpFPEqual: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^cogit gen: JumpFPEqual operand: jumpTarget asInteger! !

!CogAbstractInstruction methodsFor: 'abstract instructions' stamp: 'eem 4/17/2009 15:47'!
genJumpFPGreater: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^cogit gen: JumpFPGreater operand: jumpTarget asInteger! !

!CogAbstractInstruction methodsFor: 'abstract instructions' stamp: 'eem 4/17/2009 15:47'!
genJumpFPGreaterOrEqual: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^cogit gen: JumpFPGreaterOrEqual operand: jumpTarget asInteger! !

!CogAbstractInstruction methodsFor: 'abstract instructions' stamp: 'eem 4/17/2009 15:46'!
genJumpFPLess: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^cogit gen: JumpFPLess operand: jumpTarget asInteger! !

!CogAbstractInstruction methodsFor: 'abstract instructions' stamp: 'eem 4/17/2009 15:46'!
genJumpFPLessOrEqual: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^cogit gen: JumpFPLessOrEqual operand: jumpTarget asInteger! !

!CogAbstractInstruction methodsFor: 'abstract instructions' stamp: 'eem 4/17/2009 15:46'!
genJumpFPNotEqual: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^cogit gen: JumpFPNotEqual operand: jumpTarget asInteger! !

!CogAbstractInstruction methodsFor: 'smalltalk calling convention' stamp: 'eem 2/25/2020 19:13'!
genLoadCStackPointer
	"Load the stack pointer register with that of the C stack, effecting
	 a switch to the C stack.  Used when machine code calls into the
	 CoInterpreter run-time (e.g. to invoke interpreter primitives)."
	cogit MoveAw: cogit cStackPointerAddress R: NativeSPReg.
	^0! !

!CogAbstractInstruction methodsFor: 'smalltalk calling convention' stamp: 'eem 4/7/2020 16:53'!
genLoadCStackPointers
	"Load the frame and stack pointer registers with those of the C stack,
	 effecting a switch to the C stack.  Used when machine code calls into
	 the CoInterpreter run-time (e.g. to invoke interpreter primitives).
	 N.B. CoInterpreter stack layout dictates that the stack pointer should be loaded first.
	 The stack zone is allocated on the C stack before the interpreter runs and hence
	 before CStackPointer and CFramePointer are captured.  So when running in machine
	 code the native stack pointer and frame pointer appear to be on a colder part of the
	 stack to CStackPointer and CFramePointer.  When CStackPointerhas been set and
	 the frame pointer is still in machine code the current frame looks like it has lots of
	 stack.  If the frame pointer was set to CFramePointer before hand then it would
	 be beyond the stack pointer for that one instruction."
	cogit MoveAw: cogit cStackPointerAddress R: NativeSPReg.
	cogit MoveAw: cogit cFramePointerAddress R: FPReg.
	^0! !

!CogAbstractInstruction methodsFor: 'smalltalk calling convention' stamp: 'eem 8/6/2021 12:58'!
genLoadStackPointerForPrimCall: spareReg
	"Switch back to the Smalltalk stack where there may be a C return address on top of stack below
	 the last primitive argument. Assign SPReg first because typically it is used immediately afterwards."
	self hasLinkRegister
		ifTrue: [cogit MoveAw: cogit stackPointerAddress R: SPReg]
		ifFalse:
			[cogit
				MoveAw: cogit stackPointerAddress R: spareReg;
				"N.B. dont use SubCq:R:R: since it may generate MoveR:spareRegR:SPReg;SubCq:wordSize R:SPReg
				 which allows for the ret addr to be smashed by an interrupt between the two instructions."
				SubCq: objectMemory wordSize R: spareReg;
				MoveR: spareReg R: SPReg].
	^0! !

!CogAbstractInstruction methodsFor: 'smalltalk calling convention' stamp: 'eem 2/1/2020 18:48'!
genLoadStackPointers
	"Switch back to the Smalltalk stack. Assign SPReg first
	 because typically it is used immediately afterwards."
	cogit MoveAw: cogit stackPointerAddress R: SPReg.
	cogit MoveAw: cogit framePointerAddress R: FPReg.
	^0! !

!CogAbstractInstruction methodsFor: 'smalltalk calling convention' stamp: 'eem 8/5/2021 14:58'!
genLoadStackPointersForPrimCall: spareReg
	"Switch back to the Smalltalk stack where there may be a C return address on top of stack below
	 the last primitive argument. Assign SPReg first because typically it is used immediately afterwards."
	self hasLinkRegister
		ifTrue: [self genLoadStackPointers]
		ifFalse:
			[cogit
				MoveAw: cogit stackPointerAddress R: spareReg;
				"N.B. dont use SubCq:R:R: since it may generate MoveR:spareRegR:SPReg;SubCq:wordSize R:SPReg
				 which allows for the ret addr to be smashed by an interrupt between the two instructions."
				SubCq: objectMemory wordSize R: spareReg;
				MoveR: spareReg R: SPReg;
				MoveAw: cogit framePointerAddress R: FPReg].
	^0! !

!CogAbstractInstruction methodsFor: 'abi' stamp: 'eem 12/18/2015 14:12'!
genMarshallNArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3
	"Generate the code to pass up to four arguments in a C run-time call.  Hack: each argument is
	 either a negative number, which encodes a constant, or a non-negative number, that of a register.

	 Run-time calls have no more than four arguments, so chosen so that on ARM, where in its C ABI the
	 first four integer arguments are passed in registers, all arguments can be passed in registers.  We
	 defer to the back end to generate this code not so much that the back end knows whether it uses
	 the stack or registers to pass arguments (it does, but...). In fact we defer for an extremely evil reason.
	 Doing so allows the x64 (where up to 6 args are passed) to assign the register arguments in an order
	 that allows some of the argument registers to be used for specific abstract  registers, specifically
	 ReceiverResultReg and ClassReg.  This is evil, evil, evil, but also it's really nice to keep using the old
	 register assignments the original author has grown accustomed to."
	<inline: true>
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'abi' stamp: 'RonieSalgado 5/2/2016 14:25'!
genMarshallNArgs: numArgs floatArg: regOrConst0 floatArg: regOrConst1 floatArg: regOrConst2 floatArg: regOrConst3
	"Generate the code to pass up to four arguments in a C run-time call.  Hack: each argument is
	 either a negative number, which encodes a constant, or a non-negative number, that of a register.

	 Run-time calls have no more than four arguments, so chosen so that on ARM, where in its C ABI the
	 first four integer arguments are passed in registers, all arguments can be passed in registers.  We
	 defer to the back end to generate this code not so much that the back end knows whether it uses
	 the stack or registers to pass arguments (it does, but...). In fact we defer for an extremely evil reason.
	 Doing so allows the x64 (where up to 6 args are passed) to assign the register arguments in an order
	 that allows some of the argument registers to be used for specific abstract  registers, specifically
	 ReceiverResultReg and ClassReg.  This is evil, evil, evil, but also it's really nice to keep using the old
	 register assignments the original author has grown accustomed to."
	<inline: true>
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'abstract instructions' stamp: 'eem 12/16/2015 09:45'!
genMulR: regSource R: regDest
	"Generate whatever code necessary to do a bytesPerOop x bytesPerOop multiplication,
	 answering the first instruction uin the sequence."
	<returnTypeC: #'AbstractInstruction *'>
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'smalltalk calling convention' stamp: 'eem 5/15/2014 09:27'!
genPushRegisterArgsForAbortMissNumArgs: numArgs
	"Ensure that the register args are pushed before the outer and
	 inner retpcs at an entry miss for arity <= self numRegArgs.  The
	 outer retpc is that of a call at a send site.  The inner is the call
	 from a method or PIC abort/miss to the trampoline."

	"This won't be as clumsy on a RISC.  But putting the receiver and
	 args above the return address means the CoInterpreter has a
	 single machine-code frame format which saves us a lot of work."

	"Iff there are register args convert
		base	->	outerRetpc		(send site retpc)
		sp		->	innerRetpc		(PIC abort/miss retpc)
	 to
		base	->	receiver
					(arg0)
					(arg1)
					outerRetpc
		sp		->	innerRetpc		(PIC abort/miss retpc)"
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'smalltalk calling convention' stamp: 'eem 6/1/2015 16:43'!
genPushRegisterArgsForNumArgs: numArgs scratchReg: scratchReg
	"Ensure that the register args are pushed before the retpc for arity <= self numRegArgs.
	 This isn't as clumsy on a RISC.  But putting the receiver and
	 args above the return address means the CoInterpreter has a
	 single machine-code frame format which saves us a lot of work."

	"Iff there are register args convert
		sp		->	retpc		(send site retpc)
	 to
		base	->	receiver
					(arg0)
					(arg1)
		sp		->	retpc		(send site retpc)"
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'abi' stamp: 'eem 12/8/2019 21:18'!
genRemoveNArgsFromStack: n
	"Remove n arguments from teh stack (iff any of the N arguments was passed on the stack)."
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'abi' stamp: 'eem 12/8/2019 21:20'!
genRestoreRegs: regMask
	"Restore the registers in regMask as saved by genSaveRegs:."
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'abi' stamp: 'eem 12/8/2019 20:53'!
genSaveRegs: regMask
	"Save the registers in regMask for a call into the C run-time from a trampoline"
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'smalltalk calling convention' stamp: 'eem 5/15/2014 09:50'!
genSaveStackPointers
	"Save the frame and stack pointer registers to the framePointer
	 and stackPointer variables.  Used to save the machine code frame
	 for use by the run-time when calling into the CoInterpreter run-time."
	cogit MoveR: FPReg Aw: cogit framePointerAddress.
	cogit MoveR: SPReg Aw: cogit stackPointerAddress.
	^0! !

!CogAbstractInstruction methodsFor: 'abstract instructions' stamp: 'eem 5/6/2009 10:15'!
genSubstituteReturnAddress: retpc
	"Fake the return address for a call/jump so that the ``call'' returns
	 to retpc given that the ``call'' will be made by a following jump."
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'abstract instructions' stamp: 'eem 2/15/2017 13:28'!
genSwapR: regA R: regB Scratch: regTmp
	"Generic register swap code.  Subclasses for processors that have a true exchange operation will override to use it."
	| first |
	<var: 'first' type: #'AbstractInstruction *'>
	first :=
	cogit MoveR: regA R: regTmp.
	cogit MoveR: regB R: regA.
	cogit MoveR: TempReg R: regB.
	^first! !

!CogAbstractInstruction methodsFor: 'abi' stamp: 'eem 02/04/2020 11:38'!
genWriteCResultIntoReg: abstractRegister
	(abstractRegister ~= NoReg
	 and: [abstractRegister ~= ABIResultReg]) ifTrue:
		[cogit gen: MoveRR operand: ABIResultReg operand: abstractRegister]! !

!CogAbstractInstruction methodsFor: 'abi' stamp: 'eem 1/8/2020 13:16'!
genWriteCSecondResultIntoReg: abstractRegister
	<option: #LowcodeVM>
	(abstractRegister ~= NoReg
	 and: [abstractRegister ~= self cResultRegisterHigh]) ifTrue:
		[cogit gen: MoveRR operand: self cResultRegisterHigh operand: abstractRegister]! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 2/12/2009 22:47'!
generateICacheFlush
	"A dummy routine.
	 Processors that can generate code to flush the icache can override."
! !

!CogAbstractInstruction methodsFor: 'multi-threading' stamp: 'eem 7/15/2009 13:33'!
generateLowLevelTryLock: vmOwnerLockAddress
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 1/15/2016 15:32'!
getDefaultCogCodeSize
	"Answer the default number of bytes to allocate for native code at startup.
	 The actual value can be set via vmParameterAt: and/or a preference in the ini file.
	 Subclasses may this default override to arrange that the code zone has about
	 the same number of methods as the x86 back-end at a similar point in execution."
	<inline: true>
	^1024 * 1024! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 2/11/2017 13:51'!
getJmpTarget
	"Get the target of a jump instruction.  Jumps have the target in the first operand."
	<inline: true>
	^cogit cCoerceSimple: (operands at: 0) to: #'AbstractInstruction *'! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 8/4/2021 11:57'!
has64BitPerformanceCounter
	"At least x86, x86_64, AArch32 and AArch64 have a 64-bit performance counter.
	 Subclasses can turn this off if rquired."
	<inline: #always>
	^true! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'rmacnak 11/17/2015 18:18:03'!
hasConditionRegister
	"Answer if the receiver supports, e.g., JumpOverflow after a regular AddRR"
	<inline: true>
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 3/28/2015 23:53'!
hasDoublePrecisionFloatingPointSupport
	<inline: true>
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'nice 8/26/2019 10:24'!
hasLZCNTInstructions
	"Answer if the processor has a LZCNT (leading zero count) instruction"
	<inline: true>
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 3/28/2015 23:53'!
hasLinkRegister
	"Answer if the processor has a link register, i.e. if calls pass
	 the return pc in a register instead of pushing it on a stack."
	<inline: true>
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'tpr 11/24/2019 11:04'!
hasPCRegister
	"Answer if the processor has a generally addressable pc register, such as the ARM32.
	 On such processors we can execute jumping to pop top of stack by popping into
	 the pc register.
	Note that this is not a generic RISC feature.  The PowerPC does not
	 allow one to pop into the pc for example, and now the AARCH64 ARM works the same way.
	So by default, answer false."
	^false! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 3/28/2015 12:46'!
hasThreeAddressArithmetic
	"Answer if the receiver supports three-address arithmetic instructions (currently only AndCqRR)"
	<inline: true>
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
hasVarBaseRegister
	"Answer if the processor has a dedicated callee-saved register to point to
	 the base of commonly-accessed variables. By default this is false."
	^false! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'eem 10/23/2015 13:59:29'!
inCurrentCompilation: operand
	"Answer if operand is in the current compilation; and henced could be a candidate for pc-relative addressing."
	<inline: true>
	^(self isAnInstruction: (cogit cCoerceSimple: operand to: #'AbstractInstruction *')) "i.e. a label, which by definition will be in the current compilation."
	   or: [cogit addressIsInCurrentCompilation: operand]! !

!CogAbstractInstruction methodsFor: 'initialization' stamp: 'eem 11/19/2020 19:18'!
initialize
	"This method intializes the Smalltalk instance.  The C instance is merely a struct and doesn't need initialization."
	operands := CArrayAccessor on: (Array new: NumOperands).
	machineCode := CArrayAccessor on: (self codeGranularity = 4
											ifTrue: [WordArray new: self machineCodeWords]
											ifFalse: [ByteArray new: self machineCodeBytes])! !

!CogAbstractInstruction methodsFor: 'initialization' stamp: 'eem 12/17/2019 12:46'!
initializeSharableLiteral: literal
	"For out-of-line literal support, initialize a sharable literal."
	opcode := Literal.
	annotation := nil. "separate := nil for Slang"
	address := nil.
	dependent := nil.
	operands
		at: 0 put: literal;
		at: 1 put: (1 + (objectMemory bytesPerOop << 1));	"isSharable/isUnique not plus the size of the literal"
		at: 2 put: -1			"opcodeIndex"! !

!CogAbstractInstruction methodsFor: 'initialization' stamp: 'eem 12/17/2019 13:44'!
initializeUniqueLiteral: literal
	"For out-of-line literal support, initialize an unsharable literal."
	opcode := Literal.
	annotation := nil. "separate := nil for Slang"
	address := nil.
	dependent := nil.
	operands
		at: 0 put: literal;
		at: 1 put: 0 + (objectMemory bytesPerOop << 1);	"isSharable/isUnique not plus the size of the literal"
		at: 2 put: -1										"opcodeIndex"! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 1/29/2020 14:41'!
inlineCacheValueForSelector: selector in: aCogMethod
	"Answer the value to put in an inline-cache that is being loaded with the selector.
	 Usually this is simply the selector, but in 64-bits the cache is only 32-bits wide and so the
	 cache is loaded with the index of the selector.  This is the defalt 32-bit implementation."
	<cmacro: '(backEnd,selector,aCogMethod) (selector)'>
	^selector! !

!CogAbstractInstruction methodsFor: 'disassembly' stamp: 'rmacnak 11/30/2015 22:56:34'!
instructionSizeAt: pc
	"Answer the instruction size at pc. This is used in method disassembly
	 to decode the jumps in block dispatch to discover where block methods
	 occur within a larger method."
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 2/11/2017 13:58'!
isAFixup: fixupOrAddress
	<var: #fixupOrAddress type: #'void *'>
	<inline: true>
	^cogit addressIsInFixups: fixupOrAddress! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 6/10/2015 09:02'!
isAnInstruction: addressOrInstruction
	<var: #addressOrInstruction type: #'AbstractInstruction *'>
	<inline: true>
	^(cogit addressIsInInstructions: addressOrInstruction)
	  or: [addressOrInstruction == cogit methodLabel]! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 3/28/2015 23:54'!
isBigEndian
	<inline: true>
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 12/3/2015 22:26'!
isCallPrecedingReturnPC: mcpc
	"Assuming mcpc is a send return pc answer if the instruction before it is a call (not a CallFull)."
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 12/25/2008 10:53'!
isJump
	^opcode between: FirstJump and: LastJump! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 12/25/2008 10:55'!
isPCDependent
	"Answer if the receiver is a pc-dependent instruction, such as a relative jump."
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 8/3/2021 19:17'!
isWithinMwOffsetRange: anAddress
	"Answer if an address can be accessed using the offset in a MoveMw:r:R: or similar instruction.
	 We assume this is true for 32-bit processors and expect 64-bit processors to answer false
	 for values in the interpreter or the object memory."
	<inline: true>
	^true! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 11/4/2015 09:28:27'!
jmpTarget: anAbstractInstruction
	"Set the target of a jump instruction.  These all have the target in the first operand."
	<returnTypeC: #'AbstractInstruction *'>
	<var: #anAbstractInstruction type: #'AbstractInstruction *'>
	self cCode: [] "check for inadvertent smashing of already-set jmpTargets; development only"
		inSmalltalk: [self assert: ((operands at: 0)
									ifNil: [true]
									ifNotNil: [:o|
											o = 0
											or: [(self isAFixup: o)
											or: [self isAnInstruction: anAbstractInstruction]]])].
	operands at: 0 put: anAbstractInstruction asUnsignedInteger.
	^anAbstractInstruction! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 3/11/2009 22:23'!
jumpLongByteSize
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 3/11/2009 22:24'!
jumpLongConditionalByteSize
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 11/17/2015 14:12'!
jumpLongConditionalTargetBeforeFollowingAddress: mcpc
	"Extract the target from a long conditional jump.  On many ISAs this si the same as
	 extracting the target from a long unconditional jump, so we provide the default here.
	 Processors such as MIPS override as appropriate."
	<inline: true>
	^self jumpLongTargetBeforeFollowingAddress: mcpc! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 3/9/2009 09:37'!
jumpLongTargetBeforeFollowingAddress: mcpc
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 6/24/2011 18:46'!
jumpShortByteSize
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'eem 10/30/2013 15:38'!
jumpTargetAddress
	<inline: true> "Since it's an extraction from other methods."
	| jumpTarget |
	<var: #jumpTarget type: #'AbstractInstruction *'>
	jumpTarget := cogit cCoerceSimple: (operands at: 0) to: #'AbstractInstruction *'.
	cogit assertSaneJumpTarget: jumpTarget.
	(self isAnInstruction: jumpTarget) ifTrue:
		[jumpTarget := cogit cCoerceSimple: jumpTarget address to: #'AbstractInstruction *'].
	self assert: jumpTarget ~= 0.
	^jumpTarget! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'eem 2/11/2017 14:00'!
labelOffset
	"Hack:  To arrange that the block method field pushed in a block entry has
	 its MFMethodFlagIsBlockFlag bit set we provide labels with an offset.  The
	 offset for the fakeHeader reference is MFMethodFlagIsBlockFlag.  See
	 compileBlockFrameBuild:"
	<inline: true>
	^operands at: 1! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 11/18/2015 17:54'!
literal32BeforeFollowingAddress: followingAddress
	"Answer the constant loaded by the instruction sequence just before this address:"
	<inline: true>
	^objectMemory wordSize = 8
		ifTrue: [self subclassResponsibility]
		ifFalse: [self literalBeforeFollowingAddress: followingAddress]! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 12/10/2019 12:15'!
literalAlignment
	"Answer the required alignment for literals embedded in code.
	 Default to codeGranularity.  Subclasses requiring greater must override"
	<inline: true>
	^self codeGranularity! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 06/09/2015 09:10'!
literalBeforeFollowingAddress: followingAddress
	"Answer the constant loaded by the instruction sequence just before this address:"
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 3/11/2009 22:39'!
loadLiteralByteSize
	"Answer the byte size of a MoveCwR opcode's corresponding machine code"
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 4/21/2015 17:28'!
loadPICLiteralByteSize
	"Answer the byte size of a MoveCwR opcode's corresponding machine code
	 when the argument is a PIC.  This is for the self-reference at the end of a
	 closed PIC."
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'eem 4/20/2015 17:08'!
longJumpTargetAddress
	<inline: true> "Since it's an extraction from other methods."
	"This needs to be different from jumpTargetAddress because long jumps can
	be to absolute addresses and hence we can't assert that the jump target is sane."
	| jumpTarget |
	<var: #jumpTarget type: #'AbstractInstruction *'>
	jumpTarget := cogit cCoerceSimple: (operands at: 0) to: #'AbstractInstruction *'.
	(self isAnInstruction: jumpTarget) ifTrue:
		[jumpTarget := cogit cCoerceSimple: jumpTarget address to: #'AbstractInstruction *'].
	self assert: jumpTarget ~= 0.
	^jumpTarget! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 2/11/2009 22:19'!
machineCode
	^machineCode! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'eem 7/17/2012 14:27'!
machineCodeBytes
	"Answer the maximum number of bytes of machine code generated for any abstract instruction."
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 4/21/2015 09:13'!
machineCodeSize
	"N.B. This is a byte size."
	^machineCodeSize! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 4/21/2015 09:13'!
machineCodeSize: size "<signed byte>"
	"N.B. This is a byte size."
	^machineCodeSize := size! !

!CogAbstractInstruction methodsFor: 'memory access' stamp: 'eem 7/6/2021 14:45'!
makeCodeZoneExecutable
	"If a compiler class answers true to needsCodeZoneExecuteWriteSwitch
	 it must also implement makeCodeZoneExecutable."
	<doNotGenerate>
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'memory access' stamp: 'eem 7/6/2021 14:51'!
makeCodeZoneWritable
	"If a compiler class answers true to needsCodeZoneExecuteWriteSwitch
	 it must also implement makeCodeZoneWritable."
	<doNotGenerate>
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'eem 6/11/2015 16:54'!
mapEntryAddress
	"Typically map entries apply to the end of an instruction, for two reasons:
	  a)	to cope with literals embedded in variable-length instructions, since, e.g.
		on x86, the literal typically comes at the end of the instruction.
	  b)	in-line cache detection is based on return addresses, which are typically
		to the instruction following a call."
	<inline: true>
	^address + machineCodeSize! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 4/21/2015 09:13'!
maxSize
	"N.B. This is a byte size."
	^maxSize! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 4/21/2015 09:14'!
maxSize: size "<signed byte>"
	"N.B. This is a byte size."
	maxSize := size! !

!CogAbstractInstruction methodsFor: 'printing' stamp: 'eem 12/26/2015 19:45'!
nameForFPRegister: reg "<Integer>"
	"subclasses with special purpose registers may need to override."
	<doNotGenerate>
	^CogAbstractRegisters nameForFPRegister: reg! !

!CogAbstractInstruction methodsFor: 'printing' stamp: 'eem 12/26/2015 19:45'!
nameForRegister: reg "<Integer>"
	"subclasses with special purpose registers may need to override."
	<doNotGenerate>
	^CogAbstractRegisters nameForRegister: reg! !

!CogAbstractInstruction methodsFor: 'memory access' stamp: 'eem 7/5/2021 14:50'!
needsCodeZoneExecuteWriteSwitch
	"On some platforms run-time calls may be required to enable execution and
	 disable write-protect of the code zone. Subclasses redefine as required.
	 See the comment in ensureExecutableCodeZone."
	^false! !

!CogAbstractInstruction methodsFor: 'abstract instructions' stamp: 'eem 12/13/2015 13:54'!
noteFollowingConditionalBranch: branch
	"Support for processors without condition codes, such as the MIPS.
	 Answer the branch opcode.  If there are no condition codes, modify
	 the receiver and the branch to implement a suitable conditional
	 branch that doesn't depend on condition codes being set by the
	 receiver.  By default a noop. Overridden in subclasses as required."
	<returnTypeC: #'AbstractInstruction *'>
	<var: #branch type: #'AbstractInstruction *'>
	<inline: true>
	^branch! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'rmacnak 12/12/2015 11:44:27'!
nsSendCacheAt: callSiteReturnAddress
	"Answer the NSSendCache for the return address of a Newspeak
	 self, super, outer, or implicit receiver send."
	<option: #NewspeakVM>
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'initialization' stamp: 'eem 1/20/2020 21:55'!
numCheckFeaturesOpcodes
	"If the processor has a feature check facility answer the number
	 of opcodes required to compile an accessor for the feature check."
	^0! !

!CogAbstractInstruction methodsFor: 'initialization' stamp: 'eem 1/20/2020 21:56'!
numCheckLZCNTOpcodes
	"If the processor has a feature check facility for a ZeroCount instruction, answer the number
	 of opcodes required to compile an accessor for the ZeroCount feature."
	^0! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 2/3/2020 20:28'!
numDCacheFlushOpcodes
	"If the processor has the ablity to generate code to flush the dcache for the dual mapped
	 regime then answer the number of opcodes required to compile an accessor for the feature."
	^0! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 4/16/2009 11:52'!
numICacheFlushOpcodes
	"If the processor has the ablity to generate code to flush the icache answer
	 the number of opcodes required to compile an accessor for the feature."
	^0! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 10/18/2015 19:22:52'!
numIntRegArgs
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'multi-threading' stamp: 'eem 10/25/2020 08:59'!
numLowLevelLockOpcodes
	<inline: #always>
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 2/11/2009 22:19'!
opcode
	^opcode! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 2/11/2009 22:19'!
opcode: value
	^opcode := value! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 6/8/2015 18:45'!
operand0: operand
	<inline: true>
	operands at: 0 put: operand! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 6/8/2015 18:45'!
operand1: operand
	<inline: true>
	operands at: 1 put: operand! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 6/9/2015 09:30'!
operand2: operand
	<inline: true>
	operands at: 2 put: operand! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 2/11/2009 22:16'!
operands
	^operands! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 2/15/2009 16:15'!
operands: anObject
	"Set the value of operands"

	operands := anObject! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'eem 4/21/2015 23:14'!
outputMachineCodeAt: targetAddress
	"By default move machine code a byte at a time
	  Subclasses with coarser granularity can override as desired."
	<inline: true>
	0 to: machineCodeSize - 1 do:
		[:j|
		objectMemory byteAt: targetAddress + j put: (machineCode at: j)]! !

!CogAbstractInstruction methodsFor: 'feature detection' stamp: 'eem 8/5/2021 10:33'!
preferredRegisterForMovePerfCnt64RL
	"Answer the preferred register to receive the 64-bit performance counter,
	 or NoReg if none."
	<inline: #always>
	objectMemory wordSize = 8 ifTrue:
		[self subclassResponsibility]! !

!CogAbstractInstruction methodsFor: 'feature detection' stamp: 'eem 8/5/2021 08:27'!
preferredRegisterPairForMovePerfCnt64RRLInto: aBlock
	"Evaluate aBlock with the preferred registers to receive the
	 low and high 32-bits of the 64-bit performance counter."
	<inline: #always>
	objectMemory wordSize = 4 ifTrue:
		[self subclassResponsibility]! !

!CogAbstractInstruction methodsFor: 'printing' stamp: 'eem 10/27/2020 17:03'!
printStateOn: aStream
	| opcodeName orneryOperands format |
	<doNotGenerate> "Smalltalk-side only"
	opcode ifNil:
		[^self].
	aStream space; nextPut: $(; nextPutAll: (opcodeName := self class nameForOpcode: opcode).
	orneryOperands := operands isCObjectAccessor
							ifTrue: [operands object]
							ifFalse: [operands].
	format := ((CogRTLOpcodes classPool includesKey: opcodeName)
				ifTrue: [CogRTLOpcodes]
				ifFalse: [self class]) printFormatForOpcodeName: opcodeName.
	(format ifNil: [orneryOperands] ifNotNil: [orneryOperands first: format size]) withIndexDo:
		[:operand :index|
		operand ifNotNil:
			[aStream space.
			 index >= (orneryOperands identityIndexOf: nil ifAbsent: [orneryOperands size + 1]) ifTrue:
				[aStream print: index - 1; nextPut: $:].
			 (format notNil and: ['rf' includes: (format at: index ifAbsent: $-)])
				ifTrue: [aStream nextPutAll: ((format at: index) = $r
												ifTrue: [self nameForRegister: operand]
												ifFalse: [self nameForFPRegister: operand])]
				ifFalse:
					[| operandNameOrNil |
					 operandNameOrNil := operand isInteger ifTrue:
												[(cogit coInterpreter lookupAddress: operand) ifNil:
													[objectMemory lookupAddress: operand]].
					 operandNameOrNil ifNotNil: [aStream nextPut: ${].
					 aStream print: operand.
					 (operand isInteger and: [operand > 16 and: [opcode ~= Label]]) ifTrue:
						[objectMemory wordSize = 8
							ifTrue:
								[(operand allMask: 1 << 63) ifTrue:
									[aStream nextPut: $/; print: operand signedIntFromLong64]]
							ifFalse:
								[(operand allMask: 1 << 31) ifTrue:
									[aStream nextPut: $/; print: operand signedIntFromLong]].
						 aStream nextPut: $/.
						 operand printOn: aStream base: 16.
						 operandNameOrNil ifNotNil:
							[aStream nextPut: $=; nextPutAll: operandNameOrNil; nextPut: $}]]]]].
	machineCodeSize ifNotNil:
		[(machineCodeSize between: 1 and: machineCode size) ifTrue:
			[0 to: machineCodeSize - 1 by: self codeGranularity do:
				[:i|
				 aStream space.
				 (self machineCodeAt: i)
					ifNil: [aStream nextPut: $.]
					ifNotNil:
						[:mc|
						mc isInteger
							ifTrue: [mc printOn: aStream base: 16]
							ifFalse: [mc printOn: aStream]]]]].
	address ifNotNil:
		[aStream space; nextPut: $@.
		 address printOn: aStream base: 16].
	aStream nextPut: $)! !

!CogAbstractInstruction methodsFor: 'initialization' stamp: 'eem 3/8/2017 15:45'!
reinitializeOpcode
	<inline: true>
	annotation := nil.
	dependent := nil.
	operands at: 0 put: (operands at: 1 put: (operands at: 2 put: 0))! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 3/9/2009 09:43'!
relocateCallBeforeReturnPC: retpc by: delta
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 4/19/2015 18:10'!
relocateJumpLongBeforeFollowingAddress: pc by: delta
	"We assume here that calls and jumps look the same as regards their displacement.
	 This works on at least x86, ARM and x86_64.  Processors on which that isn't the
	 case can override as necessary."
	self relocateCallBeforeReturnPC: pc by: delta! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 12/5/2015 16:24'!
relocateJumpLongConditionalBeforeFollowingAddress: pc by: delta
	"Relocate a long conditional jump before pc.  Default to relocating a non-conditional jump.
	 Processors that have different formats for conditional and unconditional jumps override."
	self relocateJumpLongBeforeFollowingAddress: pc by: delta! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 4/20/2015 12:40'!
relocateMethodReferenceBeforeAddress: pc by: delta
	"On some processors the method reference is a load constant and
	 hence needs changing when methods are moved.  On others the
	 method reference is pc-relative and hence nothing needs to happen."
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'lw 8/1/2012 19:56'!
resolveJumpTarget
	<var: #fixup type: #'BytecodeFixup *'>
	| fixup |
	self assert: self isJump.
	fixup := cogit cCoerceSimple: (operands at: 0) to: #'BytecodeFixup *'.
	self cCode: [] inSmalltalk:
		[(fixup isKindOf: CogBytecodeFixup) ifTrue:
			[self assert: (self isAFixup: fixup)]].
	(self isAFixup: fixup) ifTrue:
		[self assert: (cogit addressIsInInstructions: fixup targetInstruction).
		 self jmpTarget: fixup targetInstruction]! !

!CogAbstractInstruction methodsFor: 'full transfer run-time support' stamp: 'eem 4/19/2015 08:35'!
rewriteCallFullAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a CallFull instruction to call a different target.  This variant is used to rewrite cached primitive calls.
	 Answer the extent of the code change which is used to compute the range of the icache to flush.
	 This defaults to rewriteCallAt:target:; processors that differentiate between Call and CallFull will override."
	^self rewriteCallAt: callSiteReturnAddress target: callTargetAddress! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 11/30/2015 17:01'!
rewriteConditionalJumpLongAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a conditional jump long to jump to target.  This version defaults to using
	 rewriteJumpLongAt:, which works for many ISAs.  Subclasses override if necessary."
	^self rewriteJumpLongAt: callSiteReturnAddress target: callTargetAddress! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 4/20/2015 18:44'!
rewriteInlineCacheTag: cacheTag at: callSiteReturnAddress
	"Rewrite an inline cache with a new tag.  This variant is used
	 by the garbage collector."
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'full transfer run-time support' stamp: 'eem 4/19/2015 08:34'!
rewriteJumpFullAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a JumpFull instruction to jump to a different target.  This variant is used to rewrite cached primitive calls.
	 Answer the extent of the code change which is used to compute the range of the icache to flush.
	 This defaults to rewriteJumpLongAt:target:; processors that differentiate between Jump and JumpFull will override."
	^self rewriteJumpLongAt: callSiteReturnAddress target: callTargetAddress! !

!CogAbstractInstruction methodsFor: 'method zone and entry point alignment' stamp: 'eem 1/18/2021 16:37'!
roundUpToMethodAlignment: numBytes
	"Determine the default alignment for the start of a CogMethod, which in turn
	 determines the size of the mask used to distinguish the checked and unchecked
	 entry-points, used to distinguish normal and super sends on method unlinking.
	 This is implemented here to allow processors with coarse instructions (ARM) to
	 increase the alignment if required."
	<cmacro: '(ignored,numBytes) ((numBytes) + 7 & -8)'>
	^numBytes + 7 bitAnd: -8! !

!CogAbstractInstruction methodsFor: 'abi' stamp: 'eem 3/19/2015 14:53'!
saveAndRestoreLinkRegAround: aBlock
	"If the processor's ABI includes a link register, generate instructions
	 to save and restore it around aBlock, which is assumed to generate code."
	<inline: true>
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'abi' stamp: 'eem 12/21/2019 18:25'!
saveAndRestoreLinkRegUsingCalleeSavedRegNotLiveAtPointOfSendAround: aBlock
	"If the processor's ABI includes a link register, generate instructions
	 to save and restore it in a callee-saved register (if available) or on the stack around aBlock, which is assumed to generate code.
	 By default, do nothing.  RISCs override."
	<inline: #always>
	^aBlock value! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'eem 4/23/2016 11:25'!
setLabelOffset: aValue
	"Hack:  To arrange that the block method field pushed in a block entry has
	 its MFMethodFlagIsBlockFlag bit set we provide labels with an offset.  The
	 offset for the fakeHeader reference is MFMethodFlagIsBlockFlag.  See
	 compileBlockFrameBuild:"
	<inline: true>
	^operands at: 1 put: aValue! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 1/21/2009 08:49'!
setsConditionCodesFor: aConditionalJumpOpcode
	"Answer if the receiver's opcode sets the condition codes correctly for the given conditional jump opcode."
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'eem 09/28/2015 18:48'!
sizePCDependentInstructionAt: eventualAbsoluteAddress
	"Size a jump and set its address.  The target may be another instruction
	 or an absolute address.  On entry the address inst var holds our virtual
	 address. On exit address is set to eventualAbsoluteAddress, which is
	 where this instruction will be output.  The span of a jump to a following
	 instruction is therefore between that instruction's address and this
	 instruction's address ((which are both still their virtual addresses), but the
	 span of a jump to a preceding instruction or to an absolute address is
	 between that instruction's address (which by now is its eventual absolute
	 address) or absolute address and eventualAbsoluteAddress."

	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'accessing' stamp: 'eem 2/14/2009 16:06'!
stackPageInterruptHeadroomBytes
	"Return a minimum amount of headroom for each stack page (in bytes).  In a
	 JIT the stack has to have room for interrupt handlers which will run on the stack."
	self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 12/1/2015 23:25'!
storeLiteral32: literal32 beforeFollowingAddress: followingAddress
	"Answer the constant loaded by the instruction sequence just before this address:"
	<inline: true>
	objectMemory wordSize = 8
		ifTrue: [self subclassResponsibility]
		ifFalse: [self storeLiteral: literal32 beforeFollowingAddress: followingAddress]! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'eem 06/09/2015 09:14'!
storeLiteral: literal beforeFollowingAddress: followingAddress
	"Rewrite the long constant loaded by the instruction sequence just before this address:"
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'printing' stamp: 'eem 2/11/2009 23:08'!
symbolic
	<doNotGenerate> "Smalltalk-side only"
	^String streamContents: [:s| self symbolicOn: s]! !

!CogAbstractInstruction methodsFor: 'printing' stamp: 'eem 12/8/2019 21:50'!
symbolicOn: aStream
	| orneryOperands |
	<doNotGenerate> "Smalltalk-side only"
	(machineCodeSize isNil
	 or: [opcode = 16rAAA]) ifTrue:
		[^aStream nextPut: 'uninitialized opcode'].
	aStream space; nextPut: $(; nextPutAll: (self class nameForOpcode: opcode).
	orneryOperands := operands isCObjectAccessor
							ifTrue: [operands object]
							ifFalse: [operands].
	orneryOperands withIndexDo:
		[:operand :index|
		operand ifNotNil:
			[aStream space.
			 index >= (orneryOperands identityIndexOf: nil ifAbsent: [orneryOperands size + 1]) ifTrue:
				[aStream print: index - 1; nextPut: $:].
			 operand class == self class
				ifTrue:
					[operand symbolicOn: aStream]
				ifFalse:
					[aStream print: operand.
					 (operand isInteger and: [operand > 16]) ifTrue:
						[objectMemory wordSize = 8
							ifTrue:
								[(operand allMask: 1 << 63) ifTrue:
									[aStream nextPut: $/; print: operand signedIntFromLong64]]
							ifFalse:
								[(operand allMask: 1 << 31) ifTrue:
									[aStream nextPut: $/; print: operand signedIntFromLong]].
						 aStream nextPut: $/.
						 operand printOn: aStream base: 16]]]].
	machineCodeSize > 0 ifTrue:
		[machineCodeSize > machineCode size
			ifTrue: [aStream nextPutAll: ' no mcode']
			ifFalse:
				[0 to: machineCodeSize - 1 by: self codeGranularity do:
					[:i|
					 aStream space.
					 (self machineCodeAt: i) printOn: aStream base: 16]]].
	aStream nextPut: $)! !

!CogAbstractInstruction methodsFor: 'memory access' stamp: 'eem 12/29/2018 11:04'!
unalignedLongAt: byteAddress
	"Some processors need this, but word-oriented RISCs don't. Hence doNotGenerate."
	<doNotGenerate>
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'memory access' stamp: 'eem 12/29/2018 11:04'!
unalignedLongAt: byteAddress put: aWord
	"Some processors need this, but word-oriented RISCs don't. Hence doNotGenerate."
	^self subclassResponsibility! !

!CogAbstractInstruction methodsFor: 'generate machine code' stamp: 'eem 2/11/2017 13:59'!
updateLabel: labelInstruction
	"Update an instruction that depends on a label outside
	 of generated code (e.g. a method or block header)."
	<var: #labelInstruction type: #'AbstractInstruction *'>
	<inline: true>
	self assert: (opcode = MoveCwR or: [opcode = PushCw]).
	operands at: 0 put: labelInstruction address + labelInstruction labelOffset! !

!CogAbstractInstruction methodsFor: 'testing' stamp: 'eem 1/31/2020 12:43'!
usesNativeSPReg
	"Answer if the code model implemented by the receiver uses the native stack pointer for SPReg.
	 By default answer true."
	^true! !

!CogAbstractInstruction methodsFor: 'inline cacheing' stamp: 'rmacnak 12/9/2015 23:06:30'!
zoneCallsAreRelative
	"Answer if Call and JumpLong are relative and hence need to take the caller's
	 relocation delta into account during code compaction, rather than just the
	 callee's delta."
	self subclassResponsibility! !

CogAbstractInstruction subclass: #CogIA32Compiler
	instanceVariableNames: ''
	classVariableNames: 'BSR CDQ CLD CMPXCHGRAw CPUID EAX EBP EBX ECX EDI EDX ESI ESP FSTPD FSTPS IDIVR IMULRR LFENCE LOCK MFENCE MOVSB MOVSD ModReg ModRegInd ModRegIndDisp32 ModRegIndSIB ModRegRegDisp32 ModRegRegDisp8 REP SETE SFENCE SIB1 SIB2 SIB4 SIB8 XCHGRR XMM0L XMM1L XMM2L XMM3L XMM4L XMM5L XMM6L XMM7L'
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

CogIA32Compiler class
	instanceVariableNames: 'cpuidWord0 cpuidWord1'!

!CogIA32Compiler commentStamp: 'eem 9/14/2015 17:13' prior: 0!
I generate IA32 (x86) instructions from CogAbstractInstructions.  For reference see
1. IA-32 Intel® Architecture Software Developer's Manual Volume 2A: Instruction Set Reference, A-M
2. IA-32 Intel® Architecture Software Developer's Manual Volume 2A: Instruction Set Reference, N-Z
	http://www.intel.com/products/processor/manuals/
(® is supposed to be the Unicode "registered  sign".

This class does not take any special action to flush the instruction cache on instruction-modification, trusting that Intel and AMD processors correctly invalidate the instruction cache via snooping.  According to the manuals, this will work on systems where code and data have the same virtual address.  The CogICacheFlushingIA32Compiler subclass exists to use the CPUID instruction to serialize instruction-modification for systems with code and data at different virtual addresses.!

CogIA32Compiler class
	instanceVariableNames: 'cpuidWord0 cpuidWord1'!

!CogIA32Compiler class methodsFor: 'translation' stamp: 'eem 4/21/2015 22:53'!
ISA
	"Answer the name of the ISA the receiver implements."
	^#IA32! !

!CogIA32Compiler class methodsFor: 'accessing' stamp: 'eem 10/29/2020 23:00'!
cpuidWord0
	^cpuidWord0! !

!CogIA32Compiler class methodsFor: 'accessing' stamp: 'eem 10/29/2020 23:00'!
cpuidWord1
	^cpuidWord1! !

!CogIA32Compiler class methodsFor: 'translation' stamp: 'eem 11/11/2020 18:36'!
declareCVarsIn: aCCodeGenerator
	aCCodeGenerator
		declareVar: 'cpuidWord0' type: #usqIntptr_t;
		declareVar: 'cpuidWord1' type: #usqIntptr_t! !

!CogIA32Compiler class methodsFor: 'translation' stamp: 'eem 10/25/2020 09:38'!
filteredInstVarNames
	"Move the unaligned bytes to the end of the structure."
	^(super filteredInstVarNames copyWithout: 'machineCode'), #('machineCode')! !

!CogIA32Compiler class methodsFor: 'translation' stamp: 'eem 6/12/2019 18:39'!
identifyingPredefinedMacros
	^#('_M_I386' '_M_IX86' '_X86_' 'i386' 'i486' 'i586' 'i686' '__i386__' '__386__' 'X86' 'I386')! !

!CogIA32Compiler class methodsFor: 'class initialization' stamp: 'eem 10/25/2020 17:39'!
initialize
	"Initialize various IA32/x86 instruction-related constants.
	 [1] IA-32 Intel® Architecture Software Developer's Manual Volume 2A: Instruction Set Reference, A-M"

	"CogIA32Compiler initialize"

	self ~~ CogIA32Compiler ifTrue: [^self].

	"N.B. EAX ECX and EDX are caller-save (scratch) registers.
		EBX ESI and EDI are callee-save; see concreteRegisterFor:"
	EAX := 0.
	ECX := 1.  "Were they completely mad or simply sadistic?"
	EDX := 2.
	EBX := 3.
	ESP := 4.
	EBP := 5.
	ESI := 6.
	EDI := 7.

	XMM0L := 0.
	XMM1L := 1.
	XMM2L := 2.
	XMM3L := 3.
	XMM4L := 4.
	XMM5L := 5.
	XMM6L := 6.
	XMM7L := 7.

	"Mod R/M Mod fields.  See [1] Sec 2.4, 2.5 & 2.6 & Table 2-2"
	ModRegInd := 0.
		ModRegIndSIB := 4.
		ModRegIndDisp32 := 5.
	ModRegRegDisp8 := 1.
	ModRegRegDisp32 := 2.
	ModReg := 3.

	"SIB Scaled Index modes.  See [1] Sec 2.4, 2.5 & 2.6 & Table 2-3"
	SIB1 := 0.
	SIB2 := 1.
	SIB4 := 2.
	SIB8 := 3.

	"Specific instructions"
	self
		initializeSpecificOpcodes: #(CPUID
		"Arithmetic/Bit count:"		CDQ IDIVR IMULRR FSTPS FSTPD CLD REP MOVSB MOVSD BSR XCHGRR
		"Multi-processing:"			LFENCE MFENCE SFENCE LOCK CMPXCHGRAw SETE)
		in: thisContext method! !

!CogIA32Compiler class methodsFor: 'class initialization' stamp: 'nice 2/9/2020 14:37'!
initializeAbstractRegisters
	"Assign the abstract registers with the identities/indices of the relevant concrete registers."

	"N.B. EAX ECX & EDX are caller-save (scratch) registers.  Hence we use ECX for class and EDX for
		receiver/result since these are written in all normal sends.  EBX ESI & EDI are callee-save."

	TempReg				:= EAX.
	ClassReg				:= ECX.
	ReceiverResultReg		:= EDX.
	SendNumArgsReg		:= EBX.
	SPReg					:= ESP.
	FPReg					:= EBP.
	Arg0Reg				:= ESI.
	Arg1Reg				:= EDI.

	ABICalleeSavedRegisterMask := self registerMaskFor: EBX and: ESI and: EDI.
	ABICallerSavedRegisterMask := self registerMaskFor: EAX and: ECX and: EDX.
	CallerSavedRegisterMask := ABICallerSavedRegisterMask
								bitAnd: (self registerMaskFor: ClassReg and: ReceiverResultReg and: SendNumArgsReg and: Arg0Reg and: Arg1Reg).
	ABIResultRegHigh := EDX.

	NumRegisters := 8.

	DPFPReg0				:= XMM0L.
	DPFPReg1				:= XMM1L.
	DPFPReg2				:= XMM2L.
	DPFPReg3				:= XMM3L.
	DPFPReg4				:= XMM4L.
	DPFPReg5				:= XMM5L.
	DPFPReg6				:= XMM6L.
	DPFPReg7				:= XMM7L.

	NumFloatRegisters := 8! !

!CogIA32Compiler class methodsFor: 'testing' stamp: 'eem 12/25/2015 18:47'!
isRISCTempRegister: reg
	"For tests to filter-out bogus values left in the RISCTempRegister, if any."
	^false! !

!CogIA32Compiler class methodsFor: 'translation' stamp: 'eem 6/10/2015 16:38'!
machineCodeDeclaration
	"Answer the declaration for the machineCode array."
	^{#'unsigned char'. '[', self basicNew machineCodeBytes printString, ']'}! !

!CogIA32Compiler class methodsFor: 'verification' stamp: 'eem 1/24/2012 13:16'!
specificOpcodes
	"Answer the processor-specific opcodes for this class.
	 They're all in an Array literal in the initialize method."
	^(self class >> #initialize) literals detect: [:l| l isArray and: [l includes: #LOCK]]! !

!CogIA32Compiler class methodsFor: 'translation' stamp: 'eem 12/11/2015 23:28'!
wordSize
	"This is a 32-bit ISA"
	^4! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 11/3/2015 12:19:54'!
byteReadsZeroExtend
	"Answer if a byte read, via MoveAbR, MoveMbrR, or MoveXbrRR zero-extends
	 into the full register, or merely affects the least significant 8 bits of the
	 the register.  By default the code generator assumes that byte reads
	 to not zero extend.  Note that byte reads /must not/ sign extend.
	 On x86 we always use movzbl"
	^true! !

!CogIA32Compiler methodsFor: 'abi' stamp: 'RonieSalgado 4/29/2016 16:22'!
cFloatResultToRd: reg
	cogit gen: FSTPD operand: -8 operand: SPReg .
	cogit MoveM64: -8 r: SPReg Rd: reg! !

!CogIA32Compiler methodsFor: 'abi' stamp: 'RonieSalgado 4/29/2016 15:47'!
cFloatResultToRs: reg
	cogit gen: FSTPS operand: -4 operand: SPReg.
	cogit MoveM32: -4 r: SPReg Rs: reg! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'RonieSalgado 4/29/2016 15:47'!
cResultRegisterHigh
	"Answer the abstract register for the C result register.
	 Only partially implemented.  Works on x86 since TempReg = EAX = C result reg."
	^ EDX! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'eem 7/8/2021 14:58'!
callFullInstructionByteSize
	^5! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 09/28/2015 18:48'!
callFullTargetFromReturnAddress: callSiteReturnAddress
	"Answer the address the call immediately preceding callSiteReturnAddress will jump to."
	^self callTargetFromReturnAddress: callSiteReturnAddress! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'eem 2/13/2009 12:29'!
callInstructionByteSize
	^5! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 09/28/2015 18:48'!
callTargetFromReturnAddress: callSiteReturnAddress
	"Answer the address the call immediately preceding callSiteReturnAddress will jump to."
	| callDistance |
	callDistance := self literalBeforeFollowingAddress: callSiteReturnAddress.
	^callSiteReturnAddress + callDistance signedIntFromLong! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
canDivQuoRem
	^true! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
canMulRR
	^true! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 6/21/2018 12:12'!
canSignExtend
	"IA32 has native SignExtend8RR & SignExtend16RR."
	<inline: true>
	^true! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 6/21/2018 12:29'!
canZeroExtend
	"IA32 has native ZeroExtend8RR & ZeroExtend16RR."
	<inline: true>
	^true! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'eem 11/20/2015 16:49'!
cmpC32RTempByteSize
	^5! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'eem 01/10/2016 10:01'!
codeGranularity
	"Answer the size in bytes of a unit of machine code."
	^1! !

!CogIA32Compiler methodsFor: 'generate machine code' stamp: 'eem 8/5/2021 09:13'!
computeMaximumSize
	"Compute the maximum size for each opcode.  This allows jump offsets to
	 be determined, provided that all backward branches are long branches."
	"N.B.  The ^N forms are to get around the bytecode compiler's long branch
	 limits which are exceeded when each case jumps around the otherwise."
	opcode caseOf: {
		"Noops & Pseudo Ops"
		[Label]					-> [^0].
		[AlignmentNops]		-> [^(operands at: 0) - 1].
		[Fill32]					-> [^4].
		[Nop]					-> [^1].
		"Specific Control/Data Movement"
		[REP]					-> [^1].
		[CLD]					-> [^1].
		[MOVSB]				-> [^1].
		[MOVSD]				-> [^1].
		[CDQ]					-> [^1].
		[IDIVR]					-> [^2].
		[IMULRR]				-> [^3].
		[CPUID]					-> [^2].
		[XCHGRR]				-> [^((operands at: 0) = EAX
									   or: [(operands at: 1) = EAX])
										ifTrue: [1]
										ifFalse: [2]].
		[FSTPS]					-> [^7].
		[FSTPD]				-> [^7].
		"Specific multi-processing"
		[CMPXCHGRAw]			-> [^7].
		[LFENCE]				-> [^3].
		[MFENCE]				-> [^3].
		[SFENCE]				-> [^3].
		[LOCK]					-> [^1].
		[SETE]					-> [^3].
		"[XCHGMwrR]			-> [^(operands at: 1) = ESP
										ifTrue: [(self isQuick: (operands at: 0)) ifTrue: [4] ifFalse: [7]]
										ifFalse: [(self isQuick: (operands at: 0)) ifTrue: [3] ifFalse: [6]]]."
						
		"Control"
		[CallFull]					-> [^5].
		[Call]						-> [^5].
		[CallR]						-> [^2].
		[JumpR]					-> [^2].
		[JumpFull]					-> [self resolveJumpTarget. ^5].
		[JumpLong]				-> [self resolveJumpTarget. ^5].
		[Jump]						-> [self resolveJumpTarget. ^5].
		[JumpZero]					-> [self resolveJumpTarget. ^6].
		[JumpNonZero]				-> [self resolveJumpTarget. ^6].
		[JumpNegative]				-> [self resolveJumpTarget. ^6].
		[JumpNonNegative]			-> [self resolveJumpTarget. ^6].
		[JumpOverflow]				-> [self resolveJumpTarget. ^6].
		[JumpNoOverflow]			-> [self resolveJumpTarget. ^6].
		[JumpCarry]				-> [self resolveJumpTarget. ^6].
		[JumpNoCarry]				-> [self resolveJumpTarget. ^6].
		[JumpLess]					-> [self resolveJumpTarget. ^6].
		[JumpGreaterOrEqual]		-> [self resolveJumpTarget. ^6].
		[JumpGreater]				-> [self resolveJumpTarget. ^6].
		[JumpLessOrEqual]			-> [self resolveJumpTarget. ^6].
		[JumpBelow]				-> [self resolveJumpTarget. ^6].
		[JumpAboveOrEqual]		-> [self resolveJumpTarget. ^6].
		[JumpAbove]				-> [self resolveJumpTarget. ^6].
		[JumpBelowOrEqual]		-> [self resolveJumpTarget. ^6].
		[JumpLongZero]			-> [self resolveJumpTarget. ^6].
		[JumpLongNonZero]		-> [self resolveJumpTarget. ^6].
		[JumpFPEqual]				-> [self resolveJumpTarget. ^6].
		[JumpFPNotEqual]			-> [self resolveJumpTarget. ^6].
		[JumpFPLess]				-> [self resolveJumpTarget. ^6].
		[JumpFPGreaterOrEqual]	-> [self resolveJumpTarget. ^6].
		[JumpFPGreater]			-> [self resolveJumpTarget. ^6].
		[JumpFPLessOrEqual]		-> [self resolveJumpTarget. ^6].
		[JumpFPOrdered]			-> [self resolveJumpTarget. ^6].
		[JumpFPUnordered]		-> [self resolveJumpTarget. ^6].
		[RetN]						-> [^(operands at: 0) = 0 ifTrue: [1] ifFalse: [3]].
		[Stop]						-> [^1].

		"Arithmetic"
		[AddCqR]		-> [^(self isQuick: (operands at: 0))
											ifTrue: [3]
											ifFalse: [(operands at: 1) = EAX
														ifTrue: [5]
														ifFalse: [6]]].
		[AddcCqR]		-> [^(self isQuick: (operands at: 0))
											ifTrue: [3]
											ifFalse: [(operands at: 1) = EAX
														ifTrue: [5]
														ifFalse: [6]]].
		[AndCqR]		-> [^(self isQuick: (operands at: 0))
											ifTrue: [3]
											ifFalse: [(operands at: 1) = EAX
														ifTrue: [5]
														ifFalse: [6]]].
		[CmpCqR]		-> [^(self isQuick: (operands at: 0))
											ifTrue: [3]
											ifFalse: [(operands at: 1) = EAX
														ifTrue: [5]
														ifFalse: [6]]].
		[OrCqR]			-> [^(self isQuick: (operands at: 0))
											ifTrue: [3]
											ifFalse: [(operands at: 1) = EAX
														ifTrue: [5]
														ifFalse: [6]]].
		[SubCqR]		-> [^(self isQuick: (operands at: 0))
											ifTrue: [3]
											ifFalse: [(operands at: 1) = EAX
														ifTrue: [5]
														ifFalse: [6]]].
		[TstCqR]		-> [^((self isQuick: (operands at: 0)) and: [(operands at: 1) < 4])
											ifTrue: [3]
											ifFalse: [(operands at: 1) = EAX
														ifTrue: [5]
														ifFalse: [6]]].
		[AddCwR]		-> [^(operands at: 1) = EAX ifTrue: [5] ifFalse: [6]].
		[AndCwR]		-> [^(operands at: 1) = EAX ifTrue: [5] ifFalse: [6]].
		[CmpCwR]		-> [^(operands at: 1) = EAX ifTrue: [5] ifFalse: [6]].
		[OrCwR]		-> [^(operands at: 1) = EAX ifTrue: [5] ifFalse: [6]].
		[SubCwR]		-> [^(operands at: 1) = EAX ifTrue: [5] ifFalse: [6]].
		[XorCwR]		-> [^(operands at: 1) = EAX ifTrue: [5] ifFalse: [6]].
		[AddRR]			-> [^2].
		[AndRR]			-> [^2].
		[CmpRR]		-> [^2].
		[OrRR]			-> [^2].
		[XorRR]			-> [^2].
		[SubRR]			-> [^2].
		[SubbRR]			-> [^2].
		[NegateR]		-> [^2].
		[NotR]		-> [^2].
		[LoadEffectiveAddressMwrR]
						-> [^((self isQuick: (operands at: 0))
											ifTrue: [3]
											ifFalse: [6])
										+ ((operands at: 1) = ESP
											ifTrue: [1]
											ifFalse: [0])].
		[LogicalShiftLeftCqR]		-> [^(operands at: 0) = 1 ifTrue: [2] ifFalse: [3]].
		[LogicalShiftRightCqR]		-> [^(operands at: 0) = 1 ifTrue: [2] ifFalse: [3]].
		[ArithmeticShiftRightCqR]	-> [^(operands at: 0) = 1 ifTrue: [2] ifFalse: [3]].
		[LogicalShiftLeftRR]			-> [^self computeShiftRRSize].
		[LogicalShiftRightRR]		-> [^self computeShiftRRSize].
		[ArithmeticShiftRightRR]		-> [^self computeShiftRRSize].
		[AddRdRd]					-> [^4].
		[CmpRdRd]					-> [^4].
		[SubRdRd]					-> [^4].
		[MulRdRd]					-> [^4].
		[DivRdRd]					-> [^4].
		[SqrtRd]					-> [^4].
		[XorRdRd]					-> [^4].
						
		[AddRsRs]				-> [^4].
		[CmpRsRs]				-> [^3].
		[SubRsRs]				-> [^4].
		[MulRsRs]				-> [^4].
		[DivRsRs]				-> [^4].
		[SqrtRs]					-> [^4].
		[XorRsRs]					-> [^3].
		[ClzRR]					-> [^4].
		[BSR]					-> [^3].
						
		"Data Movement"
		[MoveCqR]		-> [^(operands at: 0) = 0 ifTrue: [2] ifFalse: [5]].
		[MoveCwR]		-> [^5].
		[MoveRR]		-> [^2].
		[MoveRdRd]		-> [^4].
		[MoveRsRs]		-> [^4].
		[MoveAwR]		-> [^(operands at: 1) = EAX ifTrue: [5] ifFalse: [6]].
		[MoveRAw]		-> [^(operands at: 0) = EAX ifTrue: [5] ifFalse: [6]].
		[MoveAbR]		-> [^7].
		[MoveRAb]		-> [^(operands at: 0) = EAX ifTrue: [5] ifFalse: [6]].
		[MoveRM32r]	-> [^((self isQuick: (operands at: 1))
									ifTrue: [((operands at: 1) = 0
											and: [(operands at: 2) ~= EBP])
												ifTrue: [2]
												ifFalse: [3]]
									ifFalse: [6])
								+ ((operands at: 2) = ESP
									ifTrue: [1]
									ifFalse: [0])].
		[MoveRMwr]	-> [^((self isQuick: (operands at: 1))
									ifTrue: [((operands at: 1) = 0
											and: [(operands at: 2) ~= EBP])
												ifTrue: [2]
												ifFalse: [3]]
									ifFalse: [6])
								+ ((operands at: 2) = ESP
									ifTrue: [1]
									ifFalse: [0])].
		[MoveRdM64r]	-> [^((self isQuick: (operands at: 1))
											ifTrue: [5]
											ifFalse: [8])
										+ ((operands at: 2) = ESP
											ifTrue: [1]
											ifFalse: [0])].
		[MoveRsM32r]	-> [^((self isQuick: (operands at: 1))
											ifTrue: [5]
											ifFalse: [8])
										+ ((operands at: 2) = ESP
											ifTrue: [1]
											ifFalse: [0])].
		[MoveMbrR]		-> [^(operands at: 1) = ESP
								ifTrue: [(self isQuick: (operands at: 0)) ifTrue: [5] ifFalse: [8]]
								ifFalse: [(self isQuick: (operands at: 0)) ifTrue: [4] ifFalse: [7]]].
		[MoveRMbr]		-> [^(operands at: 2) = ESP
								ifTrue: [7]
								ifFalse: [(self isQuick: (operands at: 1)) ifTrue: [3] ifFalse: [6]]].
		[MoveM8rR]		-> [^(operands at: 1) = ESP
								ifTrue: [(self isQuick: (operands at: 0)) ifTrue: [5] ifFalse: [8]]
								ifFalse: [(self isQuick: (operands at: 0)) ifTrue: [4] ifFalse: [7]]].
		[MoveRM8r]		-> [^(operands at: 2) = ESP
								ifTrue: [7]
								ifFalse: [(self isQuick: (operands at: 1)) ifTrue: [3] ifFalse: [6]]].
		[MoveM16rR]	-> [^(operands at: 1) = ESP
								ifTrue: [(self isQuick: (operands at: 0)) ifTrue: [5] ifFalse: [8]]
								ifFalse: [(self isQuick: (operands at: 0)) ifTrue: [4] ifFalse: [7]]].
		[MoveRM16r]	-> [^(operands at: 2) = ESP
								ifTrue: [8]
								ifFalse: [(self isQuick: (operands at: 1)) ifTrue: [4] ifFalse: [7]]].
		[MoveM64rRd]	-> [^((self isQuick: (operands at: 0))
											ifTrue: [5]
											ifFalse: [8])
										+ ((operands at: 1) = ESP
											ifTrue: [1]
											ifFalse: [0])].
		[MoveM32rRs]	-> [^((self isQuick: (operands at: 0))
											ifTrue: [5]
											ifFalse: [8])
										+ ((operands at: 1) = ESP
											ifTrue: [1]
											ifFalse: [0])].
		[MoveM32rR]		-> [^((self isQuick: (operands at: 0))
									ifTrue: [((operands at: 0) = 0
											and: [(operands at: 1) ~= EBP])
												ifTrue: [2]
												ifFalse: [3]]
									ifFalse: [6])
								+ ((operands at: 1) = ESP
									ifTrue: [1]
									ifFalse: [0])].
		[MoveMwrR]		-> [^((self isQuick: (operands at: 0))
									ifTrue: [((operands at: 0) = 0
											and: [(operands at: 1) ~= EBP])
												ifTrue: [2]
												ifFalse: [3]]
									ifFalse: [6])
								+ ((operands at: 1) = ESP
									ifTrue: [1]
									ifFalse: [0])].
		[MoveXbrRR]	-> [self assert: (operands at: 0) ~= ESP.
							^(operands at: 1) = EBP
											ifTrue: [5]
											ifFalse: [4]].
		[MoveRXbrR]	->	[self assert: (operands at: 1) ~= ESP.
							^((operands at: 2) = EBP
											ifTrue: [4]
											ifFalse: [3])
										+ ((operands at: 0) >= 4
											ifTrue: [2]
											ifFalse: [0])].
		[MoveXwrRR]	-> [self assert: (operands at: 0) ~= ESP.
							^(operands at: 1) = EBP
											ifTrue: [4]
											ifFalse: [3]].
		[MoveRXwrR]	-> [self assert: (operands at: 1) ~= ESP.
							^(operands at: 2) = EBP
											ifTrue: [4]
											ifFalse: [3]].
		[PopR]			-> [^1].
		[PushR]			-> [^1].
		[PushCq]		-> [^(self isQuick: (operands at: 0)) ifTrue: [2] ifFalse: [5]].
		[PushCw]		-> [^5].
		[PrefetchAw]	-> [^self hasSSEInstructions ifTrue: [7] ifFalse: [0]].
		"Conversion"
		[ConvertRRd]	-> [^4].
		[ConvertRdR]	-> [^4].

		[ConvertRsRd]	-> [^4].
		[ConvertRdRs]	-> [^4].
		[ConvertRsR]	-> [^4].
		[ConvertRRs]	-> [^4].

		[SignExtend8RR] -> [^3].
		[SignExtend16RR] -> [^3].
		[ZeroExtend8RR] -> [^3].
		[ZeroExtend16RR] -> [^3].

		[MovePerfCnt64RRL] -> ["it is easier to calculate the right value by concretizing.  We concretize twice, but so what?"
								^self concretizeMovePerfCnt64RRL] }.
	^0 "to keep C compiler quiet"! !

!CogIA32Compiler methodsFor: 'generate machine code' stamp: 'eem 12/27/2015 11:08'!
computeShiftRRSize
	"On the x86 the only instructions that shift by the value of a
	 register require the shift count to be  in %ecx.  So we may
	 have to use swap instructions to get the count into ecx."
	| shiftCountReg |
	shiftCountReg := operands at: 0.
	^shiftCountReg = ECX
		ifTrue: [2]
		ifFalse:
			[shiftCountReg = EAX
				ifTrue: [1 "XCHG EAX,r2" + 2 "Sxx" + 1 "XCHG EAX,r2"]
				ifFalse: [2 "XCHG r1,r2" + 2 "Sxx" + 2 "XCHG r1,r2"]]! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeAddCqR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value reg |
	value := operands at: 0.
	reg := operands at: 1.
	(self isQuick: value) ifTrue:
		[machineCode
			at: 0 put: 16r83;
			at: 1 put: (self mod: ModReg RM: reg RO: 0);
			at: 2 put: (value bitAnd: 16rFF).
		 ^3].
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16r05;
			at: 1 put: (value bitAnd: 16rFF);
			at: 2 put: (value >> 8 bitAnd: 16rFF);
			at: 3 put: (value >> 16 bitAnd: 16rFF);
			at: 4 put: (value >> 24 bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r81;
		at: 1 put: (self mod: ModReg RM: reg RO: 0);
		at: 2 put: (value bitAnd: 16rFF);
		at: 3 put: (value >> 8 bitAnd: 16rFF);
		at: 4 put: (value >> 16 bitAnd: 16rFF);
		at: 5 put: (value >> 24 bitAnd: 16rFF).
	 ^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeAddCwR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value reg |
	value := operands at: 0.
	reg := operands at: 1.
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16r05;
			at: 1 put: (value bitAnd: 16rFF);
			at: 2 put: (value >> 8 bitAnd: 16rFF);
			at: 3 put: (value >> 16 bitAnd: 16rFF);
			at: 4 put: (value >> 24 bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r81;
		at: 1 put: (self mod: ModReg RM: reg RO: 0);
		at: 2 put: (value bitAnd: 16rFF);
		at: 3 put: (value >> 8 bitAnd: 16rFF);
		at: 4 put: (value >> 16 bitAnd: 16rFF);
		at: 5 put: (value >> 24 bitAnd: 16rFF).
	 ^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeAddcCqR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value reg |
	value := operands at: 0.
	reg := operands at: 1.
	(self isQuick: value) ifTrue:
		[machineCode
			at: 0 put: 16r83;
			at: 1 put: (self mod: ModReg RM: reg RO: 2);
			at: 2 put: (value bitAnd: 16rFF).
		 ^3].
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16r15;
			at: 1 put: (value bitAnd: 16rFF);
			at: 2 put: (value >> 8 bitAnd: 16rFF);
			at: 3 put: (value >> 16 bitAnd: 16rFF);
			at: 4 put: (value >> 24 bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r81;
		at: 1 put: (self mod: ModReg RM: reg RO: 2);
		at: 2 put: (value bitAnd: 16rFF);
		at: 3 put: (value >> 8 bitAnd: 16rFF);
		at: 4 put: (value >> 16 bitAnd: 16rFF);
		at: 5 put: (value >> 24 bitAnd: 16rFF).
	 ^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeAddcRR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	"Assemble the ADC instruction"
	| regLHS regRHS |
	regLHS := operands at: 0.
	regRHS := operands at: 1.
	machineCode
		at: 0 put: 16r13;
		at: 1 put: (self mod: ModReg RM: regLHS RO: regRHS).
	^2! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:47'!
concretizeAlignmentNops
	<inline: true>
	self flag: 'if performance is an issue generate longer nops'.
	0 to: machineCodeSize - 1 do:
		[:i|
		machineCode at: i put: 16r90].
	^machineCodeSize! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeAndCqR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| mask reg |
	mask := operands at: 0.
	reg := operands at: 1.
	(self isQuick: mask) ifTrue:
		[machineCode
			at: 0 put: 16r83;
			at: 1 put: (self mod: ModReg RM: reg RO: 4);
			at: 2 put: (mask bitAnd: 16rFF).
		 ^3].
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16r25;
			at: 1 put: (mask bitAnd: 16rFF);
			at: 2 put: (mask >> 8 bitAnd: 16rFF);
			at: 3 put: (mask >> 16 bitAnd: 16rFF);
			at: 4 put: (mask >> 24 bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r81;
		at: 1 put: (self mod: ModReg RM: reg RO: 4);
		at: 2 put: (mask bitAnd: 16rFF);
		at: 3 put: (mask >> 8 bitAnd: 16rFF);
		at: 4 put: (mask >> 16 bitAnd: 16rFF);
		at: 5 put: (mask >> 24 bitAnd: 16rFF).
	 ^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeAndCwR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value reg |
	value := operands at: 0.
	reg := operands at: 1.
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16r25;
			at: 1 put: (value bitAnd: 16rFF);
			at: 2 put: (value >> 8 bitAnd: 16rFF);
			at: 3 put: (value >> 16 bitAnd: 16rFF);
			at: 4 put: (value >> 24 bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r83;
		at: 1 put: (self mod: ModReg RM: reg RO: 4);
		at: 2 put: (value bitAnd: 16rFF);
		at: 3 put: (value >> 8 bitAnd: 16rFF);
		at: 4 put: (value >> 16 bitAnd: 16rFF);
		at: 5 put: (value >> 24 bitAnd: 16rFF).
	 ^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeArithmeticShiftRightCqR
	<inline: true>
	| shiftCount reg |
	shiftCount := (operands at: 0) min: 31.
	reg := operands at: 1.
	shiftCount = 1 ifTrue:
		[machineCode
			at: 0 put: 16rD1;
			at: 1 put: (self mod: ModReg RM: reg RO: 7).
		^2].

	machineCode
		at: 0 put: 16rC1;
		at: 1 put: (self mod: ModReg RM: reg RO: 7);
		at: 2 put: shiftCount.
	^3! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeArithmeticShiftRightRR
	"On the x86 the only instructions that shift by the value of a
	 register require the shift count to be  in %ecx.  So we may
	 have to use swap instructions to get the count into %ecx."
	<inline: true>
	| shiftCountReg destReg regToShift |
	shiftCountReg := operands at: 0.
	destReg := operands at: 1.
	shiftCountReg = ECX ifTrue:
		[machineCode
			at: 0 put: 16rD3;
			at: 1 put: (self mod: ModReg RM: destReg RO: 7).
		 ^2].
	regToShift := destReg == shiftCountReg
					ifTrue: [ECX]
					ifFalse: [destReg = ECX
								ifTrue: [shiftCountReg]
								ifFalse: [destReg]].
	shiftCountReg = EAX ifTrue:
		[machineCode
			at: 0 put: 16r90 + ECX; "XCHG EAX,ECX"
			at: 1 put: 16rD3;			"SAR ECX,EAX"
			at: 2 put: (self mod: ModReg RM: regToShift RO: 7);
			at: 3 put: 16r90 + ECX. "XCHG EAX,ECX"
		 ^4].
	machineCode
		at: 0 put: 16r87;		"XCHG E?X,ECX"
		at: 1 put: (self mod: ModReg RM: ECX RO: shiftCountReg);
		at: 2 put: 16rD3;		"SAR ECX,E!!X"
		at: 3 put: (self mod: ModReg RM: regToShift RO: 7);
		at: 4 put: 16r87;		"XCHG E?X,ECX"
		at: 5 put: (self mod: ModReg RM: ECX RO: shiftCountReg).
	^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'nice 2/19/2020 00:52'!
concretizeBSR
	"Bit Scan Reverse
	First operand is input register (mask)
	Second operand is output register (dest)"
	"BSR"
	<inline: true>
	| dest maskReg |
	maskReg := operands at: 0.
	dest := operands at: 1.
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rBD;
		at: 2 put: (self mod: ModReg RM: maskReg RO: dest).
	 ^3! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeCDQ
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	machineCode at: 0 put: 16r99.
	^1! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeCLD
	<inline: true>
	machineCode at: 0 put: 16rFC.
	^1! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeCMPXCHGAwR
	<inline: true>
	| addressOperand reg |
	addressOperand := operands at: 0.
	reg := operands at: 1.
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rB1;
		at: 2 put: (self mod: ModRegInd RM: 5 RO: reg);
		at: 3 put: (addressOperand bitAnd: 16rFF);
		at: 4 put: (addressOperand >> 8 bitAnd: 16rFF);
		at: 5 put: (addressOperand >> 16 bitAnd: 16rFF);
		at: 6 put: (addressOperand >> 24 bitAnd: 16rFF).
	^7! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeCMPXCHGMwrR
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	srcReg ~= ESP ifTrue:
		[(self isQuick: offset) ifTrue:
			[machineCode
				at: 0 put: 16r0F;
				at: 1 put: 16rB1;
				at: 2 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: 3 put: (offset bitAnd: 16rFF).
			 ^4].
		machineCode
			at: 0 put: 16r0F;
			at: 1 put: 16rB1;
			at: 2 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
			at: 3 put: (offset bitAnd: 16rFF);
			at: 4 put: (offset >> 8 bitAnd: 16rFF);
			at: 5 put: (offset >> 16 bitAnd: 16rFF);
			at: 6 put: (offset >> 24 bitAnd: 16rFF).
		^7].
	"ESP:"
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 0 put: 16r0F;
			at: 1 put: 16rB1;
			at: 2 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: 3 put: (self s: SIB1 i: 4 b: srcReg);
			at: 4 put: (offset bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rB1;
		at: 2 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
		at: 3 put: (self s: SIB1 i: 4 b: srcReg);
		at: 4 put: (offset bitAnd: 16rFF);
		at: 5 put: (offset >> 8 bitAnd: 16rFF);
		at: 6 put: (offset >> 16 bitAnd: 16rFF);
		at: 7 put: (offset >> 24 bitAnd: 16rFF).
	^8! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 10/25/2020 17:48'!
concretizeCMPXCHGRAw
	<inline: true>
	| addressOperand reg |
	reg := operands at: 0.
	addressOperand := operands at: 1.
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rB1;
		at: 2 put: (self mod: ModRegInd RM: 5 RO: reg);
		at: 3 put: (addressOperand bitAnd: 16rFF);
		at: 4 put: (addressOperand >> 8 bitAnd: 16rFF);
		at: 5 put: (addressOperand >> 16 bitAnd: 16rFF);
		at: 6 put: (addressOperand >> 24 bitAnd: 16rFF).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^7! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeCPUID
	<inline: true>
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rA2.
	^2! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCall
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| offset |
	self assert: (operands at: 0) ~= 0.
	offset := (operands at: 0) signedIntFromLong - (address + 5) signedIntFromLong.
	machineCode
		at: 0 put: 16rE8;
		at: 1 put: (offset bitAnd: 16rFF);
		at: 2 put: (offset >> 8 bitAnd: 16rFF);
		at: 3 put: (offset >> 16 bitAnd: 16rFF);
		at: 4 put: (offset >> 24 bitAnd: 16rFF).
	^5! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCallR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| reg |
	"CmpRR RHS LHS computes LHS - RHS, i.e. apparently reversed.  You have to think subtract."
	reg := operands at: 0.
	machineCode
		at: 0 put: 16rFF;
		at: 1 put: (self mod: ModReg RM: reg RO: 2).
	^2! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'nice 2/18/2020 23:50'!
concretizeClzRR
	"Count leading zeros
	First operand is output (dest)
	Second operand is input (mask)"
	"LZCNT"
	<inline: true>
	| maskReg dest  |
	maskReg := operands at: 0.
	dest := operands at: 1.
	machineCode
		at: 0 put: 16rF3;
		at: 1 put: 16r0F;
		at: 2 put: 16rBD;
		at: 3 put: (self mod: ModReg RM: maskReg RO: dest).
	 ^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCmpCqR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value reg |
	value := operands at: 0.
	reg := operands at: 1.
	(self isQuick: value) ifTrue:
		[machineCode
			at: 0 put: 16r83;
			at: 1 put: (self mod: ModReg RM: reg RO: 7);
			at: 2 put: (value bitAnd: 16rFF).
		 ^3].
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16r3D;
			at: 1 put: (value bitAnd: 16rFF);
			at: 2 put: (value >> 8 bitAnd: 16rFF);
			at: 3 put: (value >> 16 bitAnd: 16rFF);
			at: 4 put: (value >> 24 bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r81;
		at: 1 put: (self mod: ModReg RM: reg RO: 7);
		at: 2 put: (value bitAnd: 16rFF);
		at: 3 put: (value >> 8 bitAnd: 16rFF);
		at: 4 put: (value >> 16 bitAnd: 16rFF);
		at: 5 put: (value >> 24 bitAnd: 16rFF).
	 ^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCmpCwR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value reg |
	value := operands at: 0.
	reg := operands at: 1.
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16r3D;
			at: 1 put: (value bitAnd: 16rFF);
			at: 2 put: (value >> 8 bitAnd: 16rFF);
			at: 3 put: (value >> 16 bitAnd: 16rFF);
			at: 4 put: (value >> 24 bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r81;
		at: 1 put: (self mod: ModReg RM: reg RO: 7);
		at: 2 put: (value bitAnd: 16rFF);
		at: 3 put: (value >> 8 bitAnd: 16rFF);
		at: 4 put: (value >> 16 bitAnd: 16rFF);
		at: 5 put: (value >> 24 bitAnd: 16rFF).
	 ^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCmpRdRd
	"Will get inlined into concretizeAt: switch.
	 We use UCOMISD (see p 4-260 [2])"
	<inline: true>
	| regLHS regRHS |
	"CmpRR RHS LHS computes LHS - RHS, i.e. apparently reversed.  You have to think subtract."
	regRHS := operands at: 0.
	regLHS := operands at: 1.
	machineCode
		at: 0 put: 16r66;
		at: 1 put: 16r0F;
		at: 2 put: 16r2E;
		at: 3 put: (self mod: ModReg RM: regRHS RO: regLHS).
	^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCmpRsRs
	"Will get inlined into concretizeAt: switch.
	 We use UCOMISS (see p 4-260 [2])"
	<inline: true>
	| regLHS regRHS |
	"CmpRR RHS LHS computes LHS - RHS, i.e. apparently reversed.  You have to think subtract."
	regRHS := operands at: 0.
	regLHS := operands at: 1.
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16r2E;
		at: 2 put: (self mod: ModReg RM: regRHS RO: regLHS).
	^3! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConditionalJump: conditionCode
	"Will get inlined into concretizeAt: switch."
	"Sizing/generating jumps.
		Jump targets can be to absolute addresses or other abstract instructions.
		Generating initial trampolines instructions may have no maxSize and be to absolute addresses.
		Otherwise instructions must have a machineCodeSize which must be kept to."
	<inline: true>
	| offset |
	offset := self computeJumpTargetOffsetPlus: 2.
	(machineCodeSize = 0 "size not determined because no sizeJump pass; generating initial trampolines"
		ifTrue: [self isQuick: offset]
		ifFalse: [machineCodeSize = 2]) ifTrue:
		[machineCode
			at: 0 put: 16r70 + conditionCode;
			at: 1 put: (offset bitAnd: 16rFF).
		 ^2].
	^self concretizeConditionalJumpLong: conditionCode! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConditionalJumpLong: conditionCode
	"Will get inlined into concretizeAt: switch."
	"Sizing/generating jumps.
		Jump targets can be to absolute addresses or other abstract instructions.
		Generating initial trampolines instructions may have no maxSize and be to absolute addresses.
		Otherwise instructions must have a machineCodeSize which must be kept to."
	<inline: true>
	| offset |
	offset := self computeJumpTargetOffsetPlus: 6.
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16r80 + conditionCode;
		at: 2 put: (offset bitAnd: 16rFF);
		at: 3 put: (offset >> 8 bitAnd: 16rFF);
		at: 4 put: (offset >> 16 bitAnd: 16rFF);
		at: 5 put: (offset >> 24 bitAnd: 16rFF).
	^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConvertRRd
	"Will get inlined into concretizeAt: switch."
	"CVTSI2D"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF2;
		at: 1 put: 16r0F;
		at: 2 put: 16r2A;
		at: 3 put: (self mod: ModReg RM: srcReg RO: destReg).
	 ^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConvertRRs
	"Will get inlined into concretizeAt: switch."
	"CVTSI2SS"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF3;
		at: 1 put: 16r0F;
		at: 2 put: 16r2A;
		at: 3 put: (self mod: ModReg RM: srcReg RO: destReg).
	 ^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConvertRdR
	"Will get inlined into concretizeAt: switch."
	"CVTSD2SI"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at:0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF2;
		at: 1 put: 16r0F;
		at: 2 put: 16r2D;
		at: 3 put: (self mod: ModReg RM: srcReg RO: destReg).
	 ^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConvertRdRs
	"Will get inlined into concretizeAt: switch."
	"CVTSD2SS"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF2;
		at: 1 put: 16r0F;
		at: 2 put: 16r5A;
		at: 3 put: (self mod: ModReg RM: srcReg RO: destReg).
	 ^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConvertRsR
	"Will get inlined into concretizeAt: switch."
	"CVTSS2SI"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF3;
		at: 1 put: 16r0F;
		at: 2 put: 16r2D;
		at: 3 put: (self mod: ModReg RM: srcReg RO: destReg).
	 ^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConvertRsRd
	"Will get inlined into concretizeAt: switch."
	"CVTSS2SD"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF3;
		at: 1 put: 16r0F;
		at: 2 put: 16r5A;
		at: 3 put: (self mod: ModReg RM: srcReg RO: destReg).
	 ^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeFENCE: regOpcode
	<inline: true>
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rAE;
		at: 2 put: (self mod: ModReg RM: 0 RO: regOpcode).
	^3! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeFSTPD
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	srcReg := 3.
	offset := operands at: 0.
	destReg := operands at: 1.
	destReg ~= ESP ifTrue:
		[(self isQuick: offset) ifTrue:
			"FSTP dest"
			[machineCode
				at: 0 put: 16rDD;
				at: 1 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
				at: 2 put: (offset bitAnd: 16rFF).
			 ^3].
		"FSTP dest"
		machineCode
			at: 0 put: 16rDD;
			at: 1 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
			at: 2 put: (offset bitAnd: 16rFF);
			at: 3 put: (offset >> 8 bitAnd: 16rFF);
			at: 4 put: (offset >> 16 bitAnd: 16rFF);
			at: 5 put: (offset >> 24 bitAnd: 16rFF).
		^6].
	"ESP:"
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 0 put: 16rDD;
			at: 1 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
			at: 2 put: (self s: SIB1 i: 4 b: destReg);
			at: 3 put: (offset bitAnd: 16rFF).
		 ^4].
	machineCode
		at: 0 put: 16rDD;
		at: 1 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
		at: 2 put: (self s: SIB1 i: 4 b: destReg);
		at: 3 put: (offset bitAnd: 16rFF);
		at: 4 put: (offset >> 8 bitAnd: 16rFF);
		at: 5 put: (offset >> 16 bitAnd: 16rFF);
		at: 6 put: (offset >> 24 bitAnd: 16rFF).
	^7! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeFSTPS
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	srcReg := 3.
	offset := operands at: 0.
	destReg := operands at: 1.
	destReg ~= ESP ifTrue:
		[(self isQuick: offset) ifTrue:
			"FSTP dest"
			[machineCode
				at: 0 put: 16rD9;
				at: 1 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
				at: 2 put: (offset bitAnd: 16rFF).
			 ^3].
		"FSTP dest"
		machineCode
			at: 0 put: 16rD9;
			at: 1 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
			at: 2 put: (offset bitAnd: 16rFF);
			at: 3 put: (offset >> 8 bitAnd: 16rFF);
			at: 4 put: (offset >> 16 bitAnd: 16rFF);
			at: 5 put: (offset >> 24 bitAnd: 16rFF).
		^6].
	"ESP:"
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 0 put: 16rD9;
			at: 1 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
			at: 2 put: (self s: SIB1 i: 4 b: destReg);
			at: 3 put: (offset bitAnd: 16rFF).
		 ^4].
	machineCode
		at: 0 put: 16rD9;
		at: 1 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
		at: 2 put: (self s: SIB1 i: 4 b: destReg);
		at: 3 put: (offset bitAnd: 16rFF);
		at: 4 put: (offset >> 8 bitAnd: 16rFF);
		at: 5 put: (offset >> 16 bitAnd: 16rFF);
		at: 6 put: (offset >> 24 bitAnd: 16rFF).
	^7! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeFill32
	<inline: true>
	| word |
	<var: #word type: #'usqIntptr_t'>
	word := operands at: 0.
	machineCode at: 0 put: (word bitAnd: 16rFF).
	machineCode at: 1 put: (word >> 8 bitAnd: 16rFF).
	machineCode at: 2 put: (word >> 16 bitAnd: 16rFF).
	machineCode at: 3 put: (word >> 24 bitAnd: 16rFF).
	^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeIDIVR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| regDivisor |
	regDivisor := operands at: 0.
	machineCode
		at: 0 put: 16rF7;
		at: 1 put: (self mod: ModReg RM: regDivisor RO: 7).
	^2! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeJump
	"Will get inlined into concretizeAt: switch."
	"Sizing/generating jumps.
		Jump targets can be to absolute addresses or other abstract instructions.
		Generating initial trampolines instructions may have no maxSize and be to absolute addresses.
		Otherwise instructions must have a machineCodeSize which must be kept to."
	<inline: true>
	| jumpTarget offset |
	<var: #jumpTarget type: #'AbstractInstruction *'>
	jumpTarget := cogit cCoerceSimple: (operands at: 0) to: #'AbstractInstruction *'.
	cogit assertSaneJumpTarget: jumpTarget.
	(self isAnInstruction: jumpTarget) ifTrue:
		[jumpTarget := cogit cCoerceSimple: jumpTarget address to: #'AbstractInstruction *'].
	self assert: jumpTarget ~= 0.
	offset := jumpTarget signedIntFromLong - (address + 2) signedIntFromLong.
	(machineCodeSize = 0 "size not determined because no sizeJump pass; generating initial trampolines"
		ifTrue: [self isQuick: offset]
		ifFalse: [machineCodeSize = 2]) ifTrue:
		[machineCode
			at: 0 put: 16rEB;
			at: 1 put: (offset bitAnd: 16rFF).
		 ^2].
	offset := jumpTarget signedIntFromLong - (address + 5) signedIntFromLong.
	machineCode
		at: 0 put: 16rE9;
		at: 1 put: (offset bitAnd: 16rFF);
		at: 2 put: (offset >> 8 bitAnd: 16rFF);
		at: 3 put: (offset >> 16 bitAnd: 16rFF);
		at: 4 put: (offset >> 24 bitAnd: 16rFF).
	^5! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeJumpLong
	"Will get inlined into concretizeAt: switch."
	"Sizing/generating jumps.
		Jump targets can be to absolute addresses or other abstract instructions.
		Generating initial trampolines instructions may have no maxSize and be to absolute addresses.
		Otherwise instructions must have a machineCodeSize which must be kept to."
	<inline: true>
	| jumpTarget offset |
	<var: #jumpTarget type: #'AbstractInstruction *'>
	jumpTarget := cogit cCoerceSimple: (operands at: 0) to: #'AbstractInstruction *'.
	(self isAnInstruction: jumpTarget) ifTrue:
		[jumpTarget := cogit cCoerceSimple: jumpTarget address to: #'AbstractInstruction *'].
	self assert: jumpTarget ~= 0.
	offset := jumpTarget signedIntFromLong - (address + 5) signedIntFromLong.
	machineCode
		at: 0 put: 16rE9;
		at: 1 put: (offset bitAnd: 16rFF);
		at: 2 put: (offset >> 8 bitAnd: 16rFF);
		at: 3 put: (offset >> 16 bitAnd: 16rFF);
		at: 4 put: (offset >> 24 bitAnd: 16rFF).
	^5! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeJumpR
	<inline: true>
	| reg |
	reg := operands at: 0.
	machineCode
		at: 0 put: 16rFF;
		at: 1 put: (self mod: ModReg RM: reg RO: 4).
	^2! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeLOCK
	<inline: true>
	machineCode at: 0 put: 16rF0.
	^1! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeLoadEffectiveAddressMwrR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	srcReg ~= ESP ifTrue:
		[(self isQuick: offset) ifTrue:
			[machineCode
				at: 0 put: 16r8D;
				at: 1 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: 2 put: (offset bitAnd: 16rFF).
			 ^3].
		machineCode
			at: 0 put: 16r8D;
			at: 1 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
			at: 2 put: (offset bitAnd: 16rFF);
			at: 3 put: (offset >> 8 bitAnd: 16rFF);
			at: 4 put: (offset >> 16 bitAnd: 16rFF);
			at: 5 put: (offset >> 24 bitAnd: 16rFF).
		^6].
	"ESP:"
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 0 put: 16r8D;
			at: 1 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: 2 put: (self s: SIB1 i: 4 b: srcReg);
			at: 3 put: (offset bitAnd: 16rFF).
		 ^4].
	machineCode
		at: 0 put: 16r8D;
		at: 1 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
		at: 2 put: (self s: SIB1 i: 4 b: srcReg);
		at: 3 put: (offset bitAnd: 16rFF);
		at: 4 put: (offset >> 8 bitAnd: 16rFF);
		at: 5 put: (offset >> 16 bitAnd: 16rFF);
		at: 6 put: (offset >> 24 bitAnd: 16rFF).
	^7! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeLogicalShiftLeftCqR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| distance reg |
	distance := (operands at: 0) min: 31.
	reg := operands at: 1.
	distance = 1 ifTrue:
		[machineCode
			at: 0 put: 16rD1;
			at: 1 put: (self mod: ModReg RM: reg RO: 4).
		 ^2].
	machineCode
		at: 0 put: 16rC1;
		at: 1 put: (self mod: ModReg RM: reg RO: 4);
		at: 2 put: distance.
	^3! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeLogicalShiftLeftRR
	<inline: true>
	"On the x86 the only instructions that shift by the value of a
	 register require the shift count to be  in %ecx.  So we may
	 have to use swap instructions to get the count into %ecx."
	| shiftCountReg destReg regToShift |
	shiftCountReg := operands at: 0.
	destReg := operands at: 1.
	shiftCountReg = ECX ifTrue:
		[machineCode
			at: 0 put: 16rD3;
			at: 1 put: (self mod: ModReg RM: destReg RO: 4).
		 ^2].
	regToShift := destReg == shiftCountReg
					ifTrue: [ECX]
					ifFalse: [destReg = ECX
								ifTrue: [shiftCountReg]
								ifFalse: [destReg]].
	shiftCountReg = EAX ifTrue:
		[machineCode
			at: 0 put: 16r90 + ECX; "XCHG EAX,ECX"
			at: 1 put: 16rD3;			"SAR ECX,EAX"
			at: 2 put: (self mod: ModReg RM: regToShift RO: 4);
			at: 3 put: 16r90 + ECX. "XCHG EAX,ECX"
		 ^4].
	machineCode
		at: 0 put: 16r87;		"XCHG E?X,ECX"
		at: 1 put: (self mod: ModReg RM: ECX RO: shiftCountReg);
		at: 2 put: 16rD3;		"SAR ECX,E!!X"
		at: 3 put: (self mod: ModReg RM: regToShift RO: 4);
		at: 4 put: 16r87;		"XCHG E?X,ECX"
		at: 5 put: (self mod: ModReg RM: ECX RO: shiftCountReg).
	^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeLogicalShiftRightCqR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| distance reg |
	distance := (operands at: 0) min: 31.
	reg := operands at: 1.
	distance = 1 ifTrue:
		[machineCode
			at: 0 put: 16rD1;
			at: 1 put: (self mod: ModReg RM: reg RO: 5).
		 ^2].
	machineCode
		at: 0 put: 16rC1;
		at: 1 put: (self mod: ModReg RM: reg RO: 5);
		at: 2 put: distance.
	^3! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeMOVSB
	<inline: true>
	machineCode at: 0 put: 16rA4.
	^1! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeMOVSD
	<inline: true>
	machineCode at: 0 put: 16rA5.
	^1! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveAbR
	"N.B. we zero-extend because we state byteReadsZeroExtend."
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| addressOperand reg |
	addressOperand := operands at: 0.
	(self isAnInstruction: (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *')) ifTrue:
		[addressOperand := (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *') address].
	reg := operands at: 1.
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rB6;
		at: 2 put: (self mod: ModRegInd RM: 5 RO: reg);
		at: 3 put: (addressOperand bitAnd: 16rFF);
		at: 4 put: (addressOperand >> 8 bitAnd: 16rFF);
		at: 5 put: (addressOperand >> 16 bitAnd: 16rFF);
		at: 6 put: (addressOperand >> 24 bitAnd: 16rFF).
	^7! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveAwR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| addressOperand reg |
	addressOperand := operands at: 0.
	(self isAnInstruction: (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *')) ifTrue:
		[addressOperand := (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *') address].
	reg := operands at: 1.
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16rA1;
			at: 1 put: (addressOperand bitAnd: 16rFF);
			at: 2 put: (addressOperand >> 8 bitAnd: 16rFF);
			at: 3 put: (addressOperand >> 16 bitAnd: 16rFF);
			at: 4 put: (addressOperand >> 24 bitAnd: 16rFF).
			^5].
	machineCode
		at: 0 put: 16r8B;
		at: 1 put: (self mod: ModRegInd RM: 5 RO: reg);
		at: 2 put: (addressOperand bitAnd: 16rFF);
		at: 3 put: (addressOperand >> 8 bitAnd: 16rFF);
		at: 4 put: (addressOperand >> 16 bitAnd: 16rFF);
		at: 5 put: (addressOperand >> 24 bitAnd: 16rFF).
	^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveCqR
	"Will get inlined into concretizeAt: switch.
	 Note that for quick constants, xor reg,reg, movq r8 may be shorter.  But
	 we don't consider it worthwhile and so just provide concretizeMoveCwR."
	<inline: true>
	| reg |
	(operands at: 0) ~= 0 ifTrue:
		[^self concretizeMoveCwR].
	reg := operands at: 1.
	machineCode
		at: 0 put: 16r31;
		at: 1 put: (self mod: ModReg RM: reg RO: reg).
	^2! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveCwR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value |
	value := operands at: 0.
	machineCode
		at: 0 put: 16rB8 + (operands at: 1);
		at: 1 put: (value bitAnd: 16rFF);
		at: 2 put: (value >> 8 bitAnd: 16rFF);
		at: 3 put: (value >> 16 bitAnd: 16rFF);
		at: 4 put: (value >> 24 bitAnd: 16rFF).
	^5! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveM16rR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	srcReg ~= ESP ifTrue:
		[(self isQuick: offset) ifTrue:
			[machineCode
				at: 0 put: 16r0F;
				at: 1 put: 16rB7;
				at: 2 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: 3 put: (offset bitAnd: 16rFF).
			 ^4].
		machineCode
			at: 0 put: 16r0F;
			at: 1 put: 16rB7;
			at: 2 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
			at: 3 put: (offset bitAnd: 16rFF);
			at: 4 put: (offset >> 8 bitAnd: 16rFF);
			at: 5 put: (offset >> 16 bitAnd: 16rFF);
			at: 6 put: (offset >> 24 bitAnd: 16rFF).
		^7].
	"ESP:"
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 0 put: 16r0F;
			at: 1 put: 16rB7;
			at: 2 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: 3 put: (self s: SIB1 i: 4 b: srcReg);
			at: 4 put: (offset bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rB7;
		at: 2 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
		at: 3 put: (self s: SIB1 i: 4 b: srcReg);
		at: 4 put: (offset bitAnd: 16rFF);
		at: 5 put: (offset >> 8 bitAnd: 16rFF);
		at: 6 put: (offset >> 16 bitAnd: 16rFF);
		at: 7 put: (offset >> 24 bitAnd: 16rFF).
	^8! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveM32rRs
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	srcReg ~= ESP ifTrue:
		[(self isQuick: offset) ifTrue:
			[machineCode
				at: 0 put: 16rF3;
				at: 1 put: 16r0F;
				at: 2 put: 16r10;
				at: 3 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: 4 put: (offset bitAnd: 16rFF).
			 ^5].
		machineCode
			at: 0 put: 16rF3;
			at: 1 put: 16r0F;
			at: 2 put: 16r10;
			at: 3 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
			at: 4 put: (offset bitAnd: 16rFF);
			at: 5 put: (offset >> 8 bitAnd: 16rFF);
			at: 6 put: (offset >> 16 bitAnd: 16rFF);
			at: 7 put: (offset >> 24 bitAnd: 16rFF).
		^8].
	"ESP:"
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 0 put: 16rF3;
			at: 1 put: 16r0F;
			at: 2 put: 16r10;
			at: 3 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: 4 put: (self s: SIB1 i: 4 b: srcReg);
			at: 5 put: (offset bitAnd: 16rFF).
		 ^6].
	machineCode
		at: 0 put: 16rF3;
		at: 1 put: 16r0F;
		at: 2 put: 16r10;
		at: 3 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
		at: 4 put: (self s: SIB1 i: 4 b: srcReg);
		at: 5 put: (offset bitAnd: 16rFF);
		at: 6 put: (offset >> 8 bitAnd: 16rFF);
		at: 7 put: (offset >> 16 bitAnd: 16rFF);
		at: 8 put: (offset >> 24 bitAnd: 16rFF).
	^9! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveM64rRd
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	srcReg ~= ESP ifTrue:
		[(self isQuick: offset) ifTrue:
			[machineCode
				at: 0 put: 16rF2;
				at: 1 put: 16r0F;
				at: 2 put: 16r10;
				at: 3 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: 4 put: (offset bitAnd: 16rFF).
			 ^5].
		machineCode
			at: 0 put: 16rF2;
			at: 1 put: 16r0F;
			at: 2 put: 16r10;
			at: 3 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
			at: 4 put: (offset bitAnd: 16rFF);
			at: 5 put: (offset >> 8 bitAnd: 16rFF);
			at: 6 put: (offset >> 16 bitAnd: 16rFF);
			at: 7 put: (offset >> 24 bitAnd: 16rFF).
		^8].
	"ESP:"
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 0 put: 16rF2;
			at: 1 put: 16r0F;
			at: 2 put: 16r10;
			at: 3 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: 4 put: (self s: SIB1 i: 4 b: srcReg);
			at: 5 put: (offset bitAnd: 16rFF).
		 ^6].
	machineCode
		at: 0 put: 16rF2;
		at: 1 put: 16r0F;
		at: 2 put: 16r10;
		at: 3 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
		at: 4 put: (self s: SIB1 i: 4 b: srcReg);
		at: 5 put: (offset bitAnd: 16rFF);
		at: 6 put: (offset >> 8 bitAnd: 16rFF);
		at: 7 put: (offset >> 16 bitAnd: 16rFF);
		at: 8 put: (offset >> 24 bitAnd: 16rFF).
	^9! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveM8rR
	"Will get inlined into concretizeAt: switch."
	"movzwl"
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	srcReg ~= ESP ifTrue:
		[(self isQuick: offset) ifTrue:
			[machineCode
				at: 0 put: 16r0F;
				at: 1 put: 16rB6;
				at: 2 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: 3 put: (offset bitAnd: 16rFF).
			 ^4].
		machineCode
			at: 0 put: 16r0F;
			at: 1 put: 16rB6;
			at: 2 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
			at: 3 put: (offset bitAnd: 16rFF);
			at: 4 put: (offset >> 8 bitAnd: 16rFF);
			at: 5 put: (offset >> 16 bitAnd: 16rFF);
			at: 6 put: (offset >> 24 bitAnd: 16rFF).
		^7].
	"ESP:"
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 0 put: 16r0F;
			at: 1 put: 16rB6;
			at: 2 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: 3 put: (self s: SIB1 i: 4 b: srcReg);
			at: 4 put: (offset bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rB6;
		at: 2 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
		at: 3 put: (self s: SIB1 i: 4 b: srcReg);
		at: 4 put: (offset bitAnd: 16rFF);
		at: 5 put: (offset >> 8 bitAnd: 16rFF);
		at: 6 put: (offset >> 16 bitAnd: 16rFF);
		at: 7 put: (offset >> 24 bitAnd: 16rFF).
	^8! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveMbrR
	"N.B. we zero-extend because we state byteReadsZeroExtend."
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	srcReg ~= ESP ifTrue:
		[(self isQuick: offset) ifTrue:
			[machineCode
				at: 0 put: 16r0f;
				at: 1 put: 16rb6;
				at: 2 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: 3 put: (offset bitAnd: 16rFF).
			 ^4].
		machineCode
			at: 0 put: 16r0f;
			at: 1 put: 16rb6;
			at: 2 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
			at: 3 put: (offset bitAnd: 16rFF);
			at: 4 put: (offset >> 8 bitAnd: 16rFF);
			at: 5 put: (offset >> 16 bitAnd: 16rFF);
			at: 6 put: (offset >> 24 bitAnd: 16rFF).
		^7].
	"ESP:"
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 0 put: 16r0f;
			at: 1 put: 16rb6;
			at: 2 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: 3 put: (self s: SIB1 i: 4 b: srcReg);
			at: 4 put: (offset bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r0f;
		at: 1 put: 16rb6;
		at: 2 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
		at: 3 put: (self s: SIB1 i: 4 b: srcReg);
		at: 4 put: (offset bitAnd: 16rFF);
		at: 5 put: (offset >> 8 bitAnd: 16rFF);
		at: 6 put: (offset >> 16 bitAnd: 16rFF);
		at: 7 put: (offset >> 24 bitAnd: 16rFF).
	^8! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveMwrR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	srcReg ~= ESP ifTrue:
		[(offset = 0 and: [srcReg ~= EBP]) ifTrue:
			[machineCode
				at: 0 put: 16r8B;
				at: 1 put: (self mod: ModRegInd RM: srcReg RO: destReg).
			 ^2].
		(self isQuick: offset) ifTrue:
			[machineCode
				at: 0 put: 16r8B;
				at: 1 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: 2 put: (offset bitAnd: 16rFF).
			 ^3].
		machineCode
			at: 0 put: 16r8B;
			at: 1 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
			at: 2 put: (offset bitAnd: 16rFF);
			at: 3 put: (offset >> 8 bitAnd: 16rFF);
			at: 4 put: (offset >> 16 bitAnd: 16rFF);
			at: 5 put: (offset >> 24 bitAnd: 16rFF).
		^6].
	"ESP:"
	offset = 0 ifTrue:
		[machineCode
			at: 0 put: 16r8B;
			at: 1 put: (self mod: ModRegInd RM: srcReg RO: destReg);
			at: 2 put: (self s: SIB1 i: 4 b: srcReg).
		 ^3].
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 0 put: 16r8B;
			at: 1 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: 2 put: (self s: SIB1 i: 4 b: srcReg);
			at: 3 put: (offset bitAnd: 16rFF).
		 ^4].
	machineCode
		at: 0 put: 16r8B;
		at: 1 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
		at: 2 put: (self s: SIB1 i: 4 b: srcReg);
		at: 3 put: (offset bitAnd: 16rFF);
		at: 4 put: (offset >> 8 bitAnd: 16rFF);
		at: 5 put: (offset >> 16 bitAnd: 16rFF);
		at: 6 put: (offset >> 24 bitAnd: 16rFF).
	^7! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 8/5/2021 08:42'!
concretizeMovePerfCnt64RRL
	"Generate code for
		0x0: 50				pushl  %eax
		0x1: 52				pushl  %edx
		0x2: 0f 31			rdtsc
		0x4: 89 f8			movl   %edi, %eax
		0x6: 89 f2			movl   %esi, %edx
		0x8: 5a				popl   %edx
	 et al"
	| regLo regHi liveRegisters offset |
	regLo := operands at: 0.
	regHi := operands at: 1.
	liveRegisters := operands at: 2.
	offset := 0.
	(liveRegisters anyMask: (cogit registerMaskFor: EAX)) ifTrue:
		[machineCode at: 0 put: 16r50. "push %eax"
		 offset := offset + 1].
	(liveRegisters anyMask: (cogit registerMaskFor: EDX)) ifTrue:
		[machineCode at: offset put: 16r52. "push %edx"
		 offset := offset + 1].
	"too lazy to define the swap cases for the moment..."
	self deny: (regLo = EDX or: [regHi = EAX]).
	machineCode
		at: offset		put: 16r0F;									"rdtsc"
		at: offset + 1	put: 16r31.
	offset := offset + 2.
	regHi ~= EDX ifTrue:
		[machineCode
			at: offset		put: 16r89;									"movq	%edx, (=>) %regHi"
			at: offset + 1	put: (self mod: ModReg RM: regHi RO: EDX).
		 offset := offset + 2].
	regLo ~= EAX ifTrue:
		[machineCode
			at: offset + 4	put: 16r89;									"movq	%eax, (=>) %regLo"
			at: offset + 5	put: (self mod: ModReg RM: regLo RO: EAX).
		 offset := offset + 2].
	(liveRegisters anyMask: (cogit registerMaskFor: EDX)) ifTrue:
		[machineCode at: offset put: 16r5A. "pop %edx"
		 offset := offset + 1].
	(liveRegisters anyMask: (cogit registerMaskFor: EAX)) ifTrue:
		[machineCode at: offset put: 16r58. "pop %eax"
		 offset := offset + 1].
	^offset

	"(liveRegisters anyMask: (cogit registerMaskFor: EAX and: EDX))
		ifTrue:
			[{	cogit processor disassembleInstructionAt: 0 In: machineCode object.
				cogit processor disassembleInstructionAt: 1 In: machineCode object.
				cogit processor disassembleInstructionAt: 2 In: machineCode object.
				cogit processor disassembleInstructionAt: 4 In: machineCode object.
				cogit processor disassembleInstructionAt: 6 In: machineCode object.
				cogit processor disassembleInstructionAt: 8 In: machineCode object.
				cogit processor disassembleInstructionAt: 9 In: machineCode object }]
		ifFalse:
			[{	cogit processor disassembleInstructionAt: 0 In: machineCode object.
				cogit processor disassembleInstructionAt: 2 In: machineCode object.
				cogit processor disassembleInstructionAt: 5 In: machineCode object }]"! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRAb
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| addressOperand reg |
	reg := operands at: 0.
	addressOperand := operands at: 1.
	(self isAnInstruction: (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *')) ifTrue:
		[addressOperand := (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *') address].
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16rA2;
			at: 1 put: (addressOperand bitAnd: 16rFF);
			at: 2 put: (addressOperand >> 8 bitAnd: 16rFF);
			at: 3 put: (addressOperand >> 16 bitAnd: 16rFF);
			at: 4 put: (addressOperand >> 24 bitAnd: 16rFF).
			^5].
	machineCode
		at: 0 put: 16r88;
		at: 1 put: (self mod: ModRegInd RM: 5 RO: reg);
		at: 2 put: (addressOperand bitAnd: 16rFF);
		at: 3 put: (addressOperand >> 8 bitAnd: 16rFF);
		at: 4 put: (addressOperand >> 16 bitAnd: 16rFF);
		at: 5 put: (addressOperand >> 24 bitAnd: 16rFF).
	^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRAw
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| addressOperand reg |
	reg := operands at: 0.
	addressOperand := operands at: 1.
	(self isAnInstruction: (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *')) ifTrue:
		[addressOperand := (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *') address].
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16rA3;
			at: 1 put: (addressOperand bitAnd: 16rFF);
			at: 2 put: (addressOperand >> 8 bitAnd: 16rFF);
			at: 3 put: (addressOperand >> 16 bitAnd: 16rFF);
			at: 4 put: (addressOperand >> 24 bitAnd: 16rFF).
			^5].
	machineCode
		at: 0 put: 16r89;
		at: 1 put: (self mod: ModRegInd RM: 5 RO: reg);
		at: 2 put: (addressOperand bitAnd: 16rFF);
		at: 3 put: (addressOperand >> 8 bitAnd: 16rFF);
		at: 4 put: (addressOperand >> 16 bitAnd: 16rFF);
		at: 5 put: (addressOperand >> 24 bitAnd: 16rFF).
	^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRM16r
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 1.
	srcReg := operands at: 0.
	destReg := operands at: 2.
	destReg ~= ESP ifTrue:
		[(self isQuick: offset) ifTrue:
			[machineCode
				at: 0 put: 16r66;
				at: 1 put: 16r89;
				at: 2 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
				at: 3 put: (offset bitAnd: 16rFF).
			 ^4].
		machineCode
			at: 0 put: 16r66;
			at: 1 put: 16r89;
			at: 2 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
			at: 3 put: (offset bitAnd: 16rFF);
			at: 4 put: (offset >> 8 bitAnd: 16rFF);
			at: 5 put: (offset >> 16 bitAnd: 16rFF);
			at: 6 put: (offset >> 24 bitAnd: 16rFF).
		^7].
	machineCode
		at: 0 put: 16r66;
		at: 1 put: 16r89;
		at: 2 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
		at: 3 put: (self s: SIB1 i: 4 b: destReg);
		at: 4 put: (offset bitAnd: 16rFF);
		at: 5 put: (offset >> 8 bitAnd: 16rFF);
		at: 6 put: (offset >> 16 bitAnd: 16rFF);
		at: 7 put: (offset >> 24 bitAnd: 16rFF).
	^8! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRMbr
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 1.
	srcReg := operands at: 0.
	destReg := operands at: 2.
	srcReg >= 4 ifTrue: [self error: 'invalid register'].
	destReg ~= ESP ifTrue:
		[(self isQuick: offset) ifTrue:
			[machineCode
				at: 0 put: 16r88;
				at: 1 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
				at: 2 put: (offset bitAnd: 16rFF).
			 ^3].
		machineCode
			at: 0 put: 16r88;
			at: 1 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
			at: 2 put: (offset bitAnd: 16rFF);
			at: 3 put: (offset >> 8 bitAnd: 16rFF);
			at: 4 put: (offset >> 16 bitAnd: 16rFF);
			at: 5 put: (offset >> 24 bitAnd: 16rFF).
		^6].
	machineCode
		at: 0 put: 16r88;
		at: 1 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
		at: 2 put: (self s: SIB1 i: 4 b: destReg);
		at: 3 put: (offset bitAnd: 16rFF);
		at: 4 put: (offset >> 8 bitAnd: 16rFF);
		at: 5 put: (offset >> 16 bitAnd: 16rFF);
		at: 6 put: (offset >> 24 bitAnd: 16rFF).
	^7! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRMwr
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	srcReg := operands at: 0.
	offset := operands at: 1.
	destReg := operands at: 2.
	destReg ~= ESP ifTrue:
		[(offset = 0 and: [destReg ~= EBP]) ifTrue:
			[machineCode
				at: 0 put: 16r89;
				at: 1 put: (self mod: ModRegInd RM: destReg RO: srcReg).
			 ^2].
		 (self isQuick: offset) ifTrue:
			[machineCode
				at: 0 put: 16r89;
				at: 1 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
				at: 2 put: (offset bitAnd: 16rFF).
			 ^3].
		machineCode
			at: 0 put: 16r89;
			at: 1 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
			at: 2 put: (offset bitAnd: 16rFF);
			at: 3 put: (offset >> 8 bitAnd: 16rFF);
			at: 4 put: (offset >> 16 bitAnd: 16rFF);
			at: 5 put: (offset >> 24 bitAnd: 16rFF).
		^6].
	"ESP:"
	offset = 0 ifTrue:
		[machineCode
			at: 0 put: 16r89;
			at: 1 put: (self mod: ModRegInd RM: destReg RO: srcReg);
			at: 2 put: (self s: SIB1 i: 4 b: destReg).
		 ^3].
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 0 put: 16r89;
			at: 1 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
			at: 2 put: (self s: SIB1 i: 4 b: destReg);
			at: 3 put: (offset bitAnd: 16rFF).
		 ^4].
	machineCode
		at: 0 put: 16r89;
		at: 1 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
		at: 2 put: (self s: SIB1 i: 4 b: destReg);
		at: 3 put: (offset bitAnd: 16rFF);
		at: 4 put: (offset >> 8 bitAnd: 16rFF);
		at: 5 put: (offset >> 16 bitAnd: 16rFF);
		at: 6 put: (offset >> 24 bitAnd: 16rFF).
	^7! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRXbrR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| src index base swapreg mcIdx |
	src := operands at: 0.
	index := operands at: 1.
	base := operands at: 2.
	mcIdx := 0.
	swapreg := NoReg.
	src >= 4 ifTrue: "x86 allows movb %rl, mem only with %al, %bl, %cl, %dl, so swap with the first one that isn't used."
		[swapreg := src.
		 index = EAX ifTrue: [index := swapreg].
		 base = EAX ifTrue: [base := swapreg].
		 src := EAX. 
		 mcIdx := 1.
		 machineCode at: 0 put: 16r90 + swapreg].
	base ~= EBP ifTrue:
		[machineCode
			at: mcIdx + 0 put: 16r88;
			at: mcIdx + 1 put: (self mod: ModRegInd RM: 4 RO: src);
			at: mcIdx + 2 put: (self s: SIB1 i: index b: base).
		 swapreg ~= NoReg ifTrue:
			[machineCode at: mcIdx + 3 put: 16r90 + swapreg].
		 ^3 + (2 * mcIdx)].
	machineCode
		at: mcIdx + 0 put: 16r88;
		at: mcIdx + 1 put: (self mod: ModRegRegDisp8 RM: 4 RO: src);
		at: mcIdx + 2 put: (self s: SIB1 i: index b: base);
		at: mcIdx + 3 put: 0.
	swapreg ~= NoReg ifTrue:
		[machineCode at: mcIdx + 4 put: 16r90 + swapreg].
	 ^4 + (2 * mcIdx)! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRXwrR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| index base src |
	src := operands at: 0.
	index := operands at: 1.
	base := operands at: 2.
	base ~= EBP ifTrue:
		[machineCode
			at: 0 put: 16r89;
			at: 1 put: (self mod: ModRegInd RM: 4 RO: src);
			at: 2 put: (self s: SIB4 i: index b: base).
		 ^3].
	machineCode
		at: 0 put: 16r89;
		at: 1 put: (self mod: ModRegRegDisp8 RM: 4 RO: src);
		at: 2 put: (self s: SIB4 i: index b: base);
		at: 3 put: 0.
	 ^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRdM64r
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	srcReg := operands at: 0.
	offset := operands at: 1.
	destReg := operands at: 2.
	destReg ~= ESP ifTrue:
		[(self isQuick: offset) ifTrue:
			[machineCode
				at: 0 put: 16rF2;
				at: 1 put: 16r0F;
				at: 2 put: 16r11;
				at: 3 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
				at: 4 put: (offset bitAnd: 16rFF).
			 ^5].
		machineCode
			at: 0 put: 16rF2;
			at: 1 put: 16r0F;
			at: 2 put: 16r11;
			at: 3 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
			at: 4 put: (offset bitAnd: 16rFF);
			at: 5 put: (offset >> 8 bitAnd: 16rFF);
			at: 6 put: (offset >> 16 bitAnd: 16rFF);
			at: 7 put: (offset >> 24 bitAnd: 16rFF).
		^8].
	"ESP:"
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 0 put: 16rF2;
			at: 1 put: 16r0F;
			at: 2 put: 16r11;
			at: 3 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
			at: 4 put: (self s: SIB1 i: 4 b: destReg);
			at: 5 put: (offset bitAnd: 16rFF).
		 ^6].
	machineCode
		at: 0 put: 16rF2;
		at: 1 put: 16r0F;
		at: 2 put: 16r11;
		at: 3 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
		at: 4 put: (self s: SIB1 i: 4 b: destReg);
		at: 5 put: (offset bitAnd: 16rFF);
		at: 6 put: (offset >> 8 bitAnd: 16rFF);
		at: 7 put: (offset >> 16 bitAnd: 16rFF);
		at: 8 put: (offset >> 24 bitAnd: 16rFF).
	^9! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRdRd
	"Will get inlined into concretizeAt: switch."
	"MOVSD"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF2;
		at: 1 put: 16r0F;
		at: 2 put: 16r11;
		at: 3 put: (self mod: ModReg RM: destReg RO: srcReg).
	^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRsM32r
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	srcReg := operands at: 0.
	offset := operands at: 1.
	destReg := operands at: 2.
	destReg ~= ESP ifTrue:
		[(self isQuick: offset) ifTrue:
			"MOVSD destReg, srcReg"
			[machineCode
				at: 0 put: 16rF3;
				at: 1 put: 16r0F;
				at: 2 put: 16r11;
				at: 3 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
				at: 4 put: (offset bitAnd: 16rFF).
			 ^5].
		"MOVSD destReg, srcReg"
		machineCode
			at: 0 put: 16rF3;
			at: 1 put: 16r0F;
			at: 2 put: 16r11;
			at: 3 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
			at: 4 put: (offset bitAnd: 16rFF);
			at: 5 put: (offset >> 8 bitAnd: 16rFF);
			at: 6 put: (offset >> 16 bitAnd: 16rFF);
			at: 7 put: (offset >> 24 bitAnd: 16rFF).
		^8].
	"ESP:"
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 0 put: 16rF3;
			at: 1 put: 16r0F;
			at: 2 put: 16r11;
			at: 3 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
			at: 4 put: (self s: SIB1 i: 4 b: destReg);
			at: 5 put: (offset bitAnd: 16rFF).
		 ^6].
	machineCode
		at: 0 put: 16rF3;
		at: 1 put: 16r0F;
		at: 2 put: 16r11;
		at: 3 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
		at: 4 put: (self s: SIB1 i: 4 b: destReg);
		at: 5 put: (offset bitAnd: 16rFF);
		at: 6 put: (offset >> 8 bitAnd: 16rFF);
		at: 7 put: (offset >> 16 bitAnd: 16rFF);
		at: 8 put: (offset >> 24 bitAnd: 16rFF).
	^9! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRsRs
	"Will get inlined into concretizeAt: switch."
	"MOVSS"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF3;
		at: 1 put: 16r0F;
		at: 2 put: 16r11;
		at: 3 put: (self mod: ModReg RM: destReg RO: srcReg).
	^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveXbrRR
	"N.B. we zero-extend because we state byteReadsZeroExtend."
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| index base dest |
	index := operands at: 0.
	base := operands at: 1.
	dest := operands at: 2.
	base ~= EBP ifTrue:
		[machineCode
			at: 0 put: 16r0F;
			at: 1 put: 16rB6;
			at: 2 put: (self mod: ModRegInd RM: 4 RO: dest);
			at: 3 put: (self s: SIB1 i: index b: base).
		 ^4].
	machineCode
			at: 0 put: 16r0F;
			at: 1 put: 16rB6;
		at: 2 put: (self mod: ModRegRegDisp8 RM: 4 RO: dest);
		at: 3 put: (self s: SIB1 i: index b: base);
		at: 4 put: 0.
	 ^5! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveXwrRR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| index base dest |
	index := operands at: 0.
	base := operands at: 1.
	dest := operands at: 2.
	base ~= EBP ifTrue:
		[machineCode
			at: 0 put: 16r8B;
			at: 1 put: (self mod: ModRegInd RM: 4 RO: dest);
			at: 2 put: (self s: SIB4 i: index b: base).
		 ^3].
	machineCode
		at: 0 put: 16r8B;
		at: 1 put: (self mod: ModRegRegDisp8 RM: 4 RO: dest);
		at: 2 put: (self s: SIB4 i: index b: base);
		at: 3 put: 0.
	 ^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMulRR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| reg1 reg2 |
	reg1 := operands at: 0.
	reg2 := operands at: 1.
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rAF;
		at: 2 put: (self mod: ModReg RM: reg1 RO: reg2).
	^3! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeNegateR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| reg |
	reg := operands at: 0.
	machineCode
		at: 0 put: 16rF7;
		at: 1 put: (self mod: ModReg RM: reg RO: 3).
	^2! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeNop
	<inline: true>
	machineCode at: 0 put: 16r90.
	^1! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeNotR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| reg |
	reg := operands at: 0.
	machineCode
		at: 0 put: 16rF7;
		at: 1 put: (self mod: ModReg RM: reg RO: 2).
	^2! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeOpRR: x86opcode
	| regLHS regRHS |
	regLHS := operands at: 0.
	regRHS := operands at: 1.
	machineCode
		at: 0 put: x86opcode;
		at: 1 put: (self mod: ModReg RM: regLHS RO: regRHS).
	^2! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeOrCqR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| mask reg |
	mask := operands at: 0.
	reg := operands at: 1.
	(self isQuick: mask) ifTrue:
		[machineCode
			at: 0 put: 16r83;
			at: 1 put: (self mod: ModReg RM: reg RO: 1);
			at: 2 put: (mask bitAnd: 16rFF).
		 ^3].
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16r0D;
			at: 1 put: (mask bitAnd: 16rFF);
			at: 2 put: (mask >> 8 bitAnd: 16rFF);
			at: 3 put: (mask >> 16 bitAnd: 16rFF);
			at: 4 put: (mask >> 24 bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r81;
		at: 1 put: (self mod: ModReg RM: reg RO: 1);
		at: 2 put: (mask bitAnd: 16rFF);
		at: 3 put: (mask >> 8 bitAnd: 16rFF);
		at: 4 put: (mask >> 16 bitAnd: 16rFF);
		at: 5 put: (mask >> 24 bitAnd: 16rFF).
	 ^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeOrCwR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value reg |
	value := operands at: 0.
	reg := operands at: 1.
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16r0D;
			at: 1 put: (value bitAnd: 16rFF);
			at: 2 put: (value >> 8 bitAnd: 16rFF);
			at: 3 put: (value >> 16 bitAnd: 16rFF);
			at: 4 put: (value >> 24 bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r83;
		at: 1 put: (self mod: ModReg RM: reg RO: 1);
		at: 2 put: (value bitAnd: 16rFF);
		at: 3 put: (value >> 8 bitAnd: 16rFF);
		at: 4 put: (value >> 16 bitAnd: 16rFF);
		at: 5 put: (value >> 24 bitAnd: 16rFF).
	 ^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePopR
	<inline: true>
	machineCode at: 0 put: 16r58 + (operands at: 0).
	^1! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePrefetchAw
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| addressOperand |
	"Note that maxSize has been set to 7 or 0 in computeMaximumSize whether hasSSEInstructions or not"
	maxSize > 0 ifTrue:
		[addressOperand := operands at: 0.
		 machineCode
			at: 0 put: 16r0f;
			at: 1 put: 16r18;
			at: 2 put: (self mod: 0 RM: 5 RO: 1); "prefetch0, prefetch using the T0 temporal data hint"
			at: 3 put: (addressOperand bitAnd: 16rFF);
			at: 4 put: (addressOperand >> 8 bitAnd: 16rFF);
			at: 5 put: (addressOperand >> 16 bitAnd: 16rFF);
			at: 6 put: (addressOperand >> 24 bitAnd: 16rFF)].
	^maxSize! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePushCq
	<inline: true>
	| value |
	value := operands at: 0.
	(self isQuick: value) ifTrue:
		[machineCode
			at: 0 put: 16r6A;
			at: 1 put: (value bitAnd: 16rFF).
		^2].
	machineCode
		at: 0 put: 16r68;
		at: 1 put: (value bitAnd: 16rFF);
		at: 2 put: (value >> 8 bitAnd: 16rFF);
		at: 3 put: (value >> 16 bitAnd: 16rFF);
		at: 4 put: (value >> 24 bitAnd: 16rFF).
	^5! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePushCw
	<inline: true>
	| value |
	value := operands at: 0.
	machineCode
		at: 0 put: 16r68;
		at: 1 put: (value bitAnd: 16rFF);
		at: 2 put: (value >> 8 bitAnd: 16rFF);
		at: 3 put: (value >> 16 bitAnd: 16rFF);
		at: 4 put: (value >> 24 bitAnd: 16rFF).
	^5! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePushR
	<inline: true>
	machineCode at: 0 put: 16r50 + (operands at: 0).
	^1! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeREP
	<inline: true>
	machineCode at: 0 put: 16rF3.
	^1! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeRetN
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| offset |
	offset := operands at: 0.
	offset = 0 ifTrue:
		[machineCode at: 0 put: 16rC3.
		^1].
	machineCode
		at: 0 put: 16rC2;
		at: 1 put: (offset bitAnd: 16rFF);
		at: 2 put: (offset >> 8).
	^3! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeReverseOpRR: x86opcode
	| regLHS regRHS |
	regRHS := operands at: 0.
	regLHS := operands at: 1.
	machineCode
		at: 0 put: x86opcode;
		at: 1 put: (self mod: ModReg RM: regLHS RO: regRHS).
	^2! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSEE2OpRdRd: opCode
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| regLHS regRHS |
	regRHS := operands at: 0.
	regLHS := operands at: 1.
	machineCode
		at: 0 put: 16rF2;
		at: 1 put: 16r0F;
		at: 2 put: opCode;
		at: 3 put: (self mod: ModReg RM: regRHS RO: regLHS).
	^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSEEOpRsRs: opCode
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| regLHS regRHS |
	regRHS := self operands at: 0.
	regLHS := self operands at: 1.
	machineCode
		at: 0 put: 16rF3;
		at: 1 put: 16r0F;
		at: 2 put: opCode;
		at: 3 put: (self mod: ModReg RM: regRHS RO: regLHS).
	^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 10/25/2020 17:51'!
concretizeSet: conditionCode
	| reg |
	reg := operands at: 0.
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16r90 + conditionCode;
		at: 2 put: (self mod: ModReg RM: reg RO: 0).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	 ^3! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSignExtend16RR
	"Will get inlined into concretizeAt: switch."
	"movswl"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rBF;
		at: 2 put: (self mod: ModReg RM: srcReg RO: destReg).
	^ 3! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSignExtend8RR
	"Will get inlined into concretizeAt: switch."
	"movsbl"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rBE;
		at: 2 put: (self mod: ModReg RM: srcReg RO: destReg).
	^ 3! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSqrtRd
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| reg |
	reg := operands at: 0.
	machineCode
		at: 0 put: 16rF2;
		at: 1 put: 16r0F;
		at: 2 put: 16r51;
		at: 3 put: (self mod: ModReg RM: reg RO: reg).
	^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSqrtRs
	"Will get inlined into concretizeAt: switch."
	"SRTSS"
	<inline: true>
	| reg |
	reg := operands at: 0.
	machineCode
		at: 0 put: 16rF3;
		at: 1 put: 16r0F;
		at: 2 put: 16r51;
		at: 3 put: (self mod: ModReg RM: reg RO: reg).
	^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeStop
	<inline: true>
	machineCode at: 0 put: 16rCC.
	^1! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSubCqR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value reg |
	value := operands at: 0.
	reg := operands at: 1.
	(self isQuick: value) ifTrue:
		[machineCode
			at: 0 put: 16r83;
			at: 1 put: (self mod: ModReg RM: reg RO: 5);
			at: 2 put: (value bitAnd: 16rFF).
		 ^3].
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16r2D;
			at: 1 put: (value bitAnd: 16rFF);
			at: 2 put: (value >> 8 bitAnd: 16rFF);
			at: 3 put: (value >> 16 bitAnd: 16rFF);
			at: 4 put: (value >> 24 bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r81;
		at: 1 put: (self mod: ModReg RM: reg RO: 5);
		at: 2 put: (value bitAnd: 16rFF);
		at: 3 put: (value >> 8 bitAnd: 16rFF);
		at: 4 put: (value >> 16 bitAnd: 16rFF);
		at: 5 put: (value >> 24 bitAnd: 16rFF).
	 ^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSubCwR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value reg |
	value := operands at: 0.
	reg := operands at: 1.
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16r2D;
			at: 1 put: (value bitAnd: 16rFF);
			at: 2 put: (value >> 8 bitAnd: 16rFF);
			at: 3 put: (value >> 16 bitAnd: 16rFF);
			at: 4 put: (value >> 24 bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r81;
		at: 1 put: (self mod: ModReg RM: reg RO: 5);
		at: 2 put: (value bitAnd: 16rFF);
		at: 3 put: (value >> 8 bitAnd: 16rFF);
		at: 4 put: (value >> 16 bitAnd: 16rFF);
		at: 5 put: (value >> 24 bitAnd: 16rFF).
	 ^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeTstCqR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| mask reg |
	mask := operands at: 0.
	reg := operands at: 1.
	((self isQuick: mask) and: [reg < 4]) ifTrue:
		[machineCode
			at: 0 put: 16rF6;
			at: 1 put: (self mod: ModReg RM: reg RO: 0);
			at: 2 put: (mask bitAnd: 16rFF).
		 ^3].
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16rA9;
			at: 1 put: (mask bitAnd: 16rFF);
			at: 2 put: (mask >> 8 bitAnd: 16rFF);
			at: 3 put: (mask >> 16 bitAnd: 16rFF);
			at: 4 put: (mask >> 24 bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16rF7;
		at: 1 put: (self mod: ModReg RM: reg RO: 0);
		at: 2 put: (mask bitAnd: 16rFF);
		at: 3 put: (mask >> 8 bitAnd: 16rFF);
		at: 4 put: (mask >> 16 bitAnd: 16rFF);
		at: 5 put: (mask >> 24 bitAnd: 16rFF).
	 ^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeXCHGAwR
	<inline: true>
	| addressOperand reg |
	addressOperand := operands at: 0.
	reg := operands at: 1.
	machineCode
		at: 0 put: 16r87;
		at: 1 put: (self mod: ModRegInd RM: 5 RO: reg);
		at: 2 put: (addressOperand bitAnd: 16rFF);
		at: 3 put: (addressOperand >> 8 bitAnd: 16rFF);
		at: 4 put: (addressOperand >> 16 bitAnd: 16rFF);
		at: 5 put: (addressOperand >> 24 bitAnd: 16rFF).
	^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeXCHGMwrR
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	srcReg ~= ESP ifTrue:
		[(self isQuick: offset) ifTrue:
			[machineCode
				at: 0 put: 16r87;
				at: 1 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: 2 put: (offset bitAnd: 16rFF).
			 ^3].
		machineCode
			at: 0 put: 16r87;
			at: 1 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
			at: 2 put: (offset bitAnd: 16rFF);
			at: 3 put: (offset >> 8 bitAnd: 16rFF);
			at: 4 put: (offset >> 16 bitAnd: 16rFF);
			at: 5 put: (offset >> 24 bitAnd: 16rFF).
		^6].
	"ESP:"
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 0 put: 16r87;
			at: 1 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: 2 put: (self s: SIB1 i: 4 b: srcReg);
			at: 3 put: (offset bitAnd: 16rFF).
		 ^4].
	machineCode
		at: 0 put: 16r87;
		at: 1 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
		at: 2 put: (self s: SIB1 i: 4 b: srcReg);
		at: 3 put: (offset bitAnd: 16rFF);
		at: 4 put: (offset >> 8 bitAnd: 16rFF);
		at: 5 put: (offset >> 16 bitAnd: 16rFF);
		at: 6 put: (offset >> 24 bitAnd: 16rFF).
	^7! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeXCHGRR
	<inline: true>
	| reg1 reg2 |
	reg1 := operands at: 0.
	reg2 := operands at: 1.
	reg2 = EAX ifTrue:
		[reg2 := reg1.
		 reg1 := EAX].
	reg1 = EAX ifTrue:
		[machineCode at: 0 put: 16r90 + reg2.
		 ^1].
	machineCode
		at: 0 put: 16r87;
		at: 1 put: (self mod: ModReg RM: reg1 RO: reg2).
	^2! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeXorCwR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value reg |
	value := operands at: 0.
	reg := operands at: 1.
	reg = EAX ifTrue:
		[machineCode
			at: 0 put: 16r35;
			at: 1 put: (value bitAnd: 16rFF);
			at: 2 put: (value >> 8 bitAnd: 16rFF);
			at: 3 put: (value >> 16 bitAnd: 16rFF);
			at: 4 put: (value >> 24 bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 0 put: 16r81;
		at: 1 put: (self mod: ModReg RM: reg RO: 6);
		at: 2 put: (value bitAnd: 16rFF);
		at: 3 put: (value >> 8 bitAnd: 16rFF);
		at: 4 put: (value >> 16 bitAnd: 16rFF);
		at: 5 put: (value >> 24 bitAnd: 16rFF).
	 ^6! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeXorRdRd
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| regLHS regRHS |
	regRHS := self operands at: 0.
	regLHS := self operands at: 1.
	machineCode
		at: 0 put: 16r66;
		at: 1 put: 16r0f;
		at: 2 put: 16r57;
		at: 3 put: (self mod: ModReg RM: regRHS RO: regLHS).
	^4! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeXorRsRs
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| regLHS regRHS |
	regRHS := self operands at: 0.
	regLHS := self operands at: 1.
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16r57;
		at: 2 put: (self mod: ModReg RM: regRHS RO: regLHS).
	^3! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeZeroExtend16RR
	"Will get inlined into concretizeAt: switch."
	"movzwl"
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rB7;
		at: 2 put: (self mod: ModReg RM: srcReg RO: destReg).
	^ 3! !

!CogIA32Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeZeroExtend8RR
	"Will get inlined into concretizeAt: switch."
	"movzbl"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rB6;
		at: 2 put: (self mod: ModReg RM: srcReg RO: destReg).
	^ 3! !

!CogIA32Compiler methodsFor: 'feature detection' stamp: 'eem 10/29/2020 23:02'!
cpuid: n
	<doNotGenerate>
	"This is simulation only invocation of the throw-away CPUID function generated to initialize cpuidWord0 and cpuidWord1"
	| result |
	cogit processor abiMarshalArg0: n in: objectMemory memory.
	result := cogit simulateLeafCallOf: cogit methodZoneBase.
	cogit processor abiUnmarshal: 1.
	^result! !

!CogIA32Compiler methodsFor: 'feature detection' stamp: 'eem 10/30/2020 05:51'!
cpuidWord0
	<cmacro: '(ign) cpuidWord0'>
	"For want of somewhere to put the variable..."
	^self class cpuidWord0! !

!CogIA32Compiler methodsFor: 'feature detection' stamp: 'eem 10/30/2020 05:51'!
cpuidWord1
	<cmacro: '(ign) cpuidWord1'>
	"For want of somewhere to put the variable..."
	^self class cpuidWord1! !

!CogIA32Compiler methodsFor: 'feature detection' stamp: 'eem 7/8/2021 12:41'!
detectFeatures
	"Do a throw-away compilation to get at the cpuid info and initialize cpuidWord0 and cpuidWord1.
	 N.B. All of MSVC, gcc & clang have intrinsics for this, so if you have the energy by all means
	 reimplement as an #if _MSC_VER...#elif __GNUC__ #else ... saga."
	| startAddress cpuid |
	<var: 'cpuid' declareC: 'usqIntptr_t (*cpuid)(usqIntptr_t arg)'>
	startAddress := cogit methodZoneBase.
	cogit allocateOpcodes: 10 bytecodes: 0.
	cpuid := cogit cCoerceSimple: cogit methodZoneBase to: #'usqIntptr_t (*)(usqIntptr_t)'.
	cogit
		PushR: EDX;
		PushR: ECX;
		PushR: EBX;
		MoveMw: 16 "3 saved registers plus return pc" r: ESP R: EAX;
		gen: CPUID;
		MoveR: EDX R: EAX;
		PopR: EBX;
		PopR: ECX;
		PopR: EDX;
		RetN: 0.
	cogit outputInstructionsForGeneratedRuntimeAt: startAddress.
	cogit resetMethodZoneBase: startAddress.
	cogit ensureExecutableCodeZone.
	self setCpuidWord0: (self cCode: 'cpuid(1)' inSmalltalk: [self cpuid: 1]).
	cogit zeroOpcodeIndexForNewOpcodes.
	cogit
		PushR: EDX;
		PushR: ECX;
		PushR: EBX;
		MoveMw: 16 "3 saved registers plus return pc" r: ESP R: EAX;
		gen: CPUID;
		MoveR: ECX R: EAX;
		PopR: EBX;
		PopR: ECX;
		PopR: EDX;
		RetN: 0.
	cogit outputInstructionsForGeneratedRuntimeAt: startAddress.
	cogit resetMethodZoneBase: startAddress.
	self setCpuidWord1: (self cCode: 'cpuid(0x80000001)' inSmalltalk: [self cpuid: 16r80000001])! !

!CogIA32Compiler methodsFor: 'generate machine code' stamp: 'eem 8/5/2021 08:45'!
dispatchConcretize
	"Attempt to generate concrete machine code for the instruction at address.
	 This is the inner dispatch of concretizeAt: actualAddress which exists only
	 to get around the branch size limits in the SqueakV3 (blue book derived)
	 bytecode set."
	opcode >= CPUID ifTrue:
		[^self dispatchConcretizeProcessorSpecific].

	opcode caseOf: {
		"Noops & Pseudo Ops"
		[Label]				-> [^self concretizeLabel].
		[AlignmentNops]	-> [^self concretizeAlignmentNops].
		[Fill32]				-> [^self concretizeFill32].
		[Nop]				-> [^self concretizeNop].
		"Control"
		[Call]					-> [^self concretizeCall].
		[CallR]					-> [^self concretizeCallR].
		[CallFull]				-> [^self concretizeCall].
		[JumpR]				-> [^self concretizeJumpR].
		[JumpFull]				-> [^self concretizeJumpLong].
		[JumpLong]			-> [^self concretizeJumpLong].
		[JumpLongZero]		-> [^self concretizeConditionalJump: 16r4].
		[JumpLongNonZero]	-> [^self concretizeConditionalJump: 16r5].
		[Jump]					-> [^self concretizeJump].
		"Table B-1 Intel® 64 and IA-32 Architectures Software Developer's Manual Volume 1: Basic Architecture"
		[JumpZero]				-> [^self concretizeConditionalJump: 16r4].
		[JumpNonZero]			-> [^self concretizeConditionalJump: 16r5].
		[JumpNegative]			-> [^self concretizeConditionalJump: 16r8].
		[JumpNonNegative]		-> [^self concretizeConditionalJump: 16r9].
		[JumpOverflow]			-> [^self concretizeConditionalJump: 16r0].
		[JumpNoOverflow]		-> [^self concretizeConditionalJump: 16r1].
		[JumpCarry]			-> [^self concretizeConditionalJump: 16r2].
		[JumpNoCarry]			-> [^self concretizeConditionalJump: 16r3].
		[JumpLess]				-> [^self concretizeConditionalJump: 16rC].
		[JumpGreaterOrEqual]	-> [^self concretizeConditionalJump: 16rD].
		[JumpGreater]			-> [^self concretizeConditionalJump: 16rF].
		[JumpLessOrEqual]		-> [^self concretizeConditionalJump: 16rE].
		[JumpBelow]			-> [^self concretizeConditionalJump: 16r2].
		[JumpAboveOrEqual]	-> [^self concretizeConditionalJump: 16r3].
		[JumpAbove]			-> [^self concretizeConditionalJump: 16r7].
		[JumpBelowOrEqual]	-> [^self concretizeConditionalJump: 16r6].
		[JumpFPEqual]				-> [^self concretizeConditionalJump: 16r4].
		[JumpFPNotEqual]			-> [^self concretizeConditionalJump: 16r5].
		[JumpFPLess]				-> [^self concretizeConditionalJump: 16r2].
		[JumpFPGreaterOrEqual]	-> [^self concretizeConditionalJump: 16r3].
		[JumpFPGreater]			-> [^self concretizeConditionalJump: 16r7].
		[JumpFPLessOrEqual]		-> [^self concretizeConditionalJump: 16r6].
		[JumpFPOrdered]			-> [^self concretizeConditionalJump: 16rB].
		[JumpFPUnordered]			-> [^self concretizeConditionalJump: 16rA].
		[RetN]						-> [^self concretizeRetN].
		[Stop]						-> [^self concretizeStop].
		"Arithmetic"
		[AddCqR]					-> [^self concretizeAddCqR].
		[AddCwR]					-> [^self concretizeAddCwR].
		[AddRR]					-> [^self concretizeOpRR: 16r03].
		[AddcRR]					-> [^self concretizeAddcRR].
		[AddcCqR]					-> [^self concretizeAddcCqR].
		[AddRdRd]					-> [^self concretizeSEE2OpRdRd: 16r58].
		[AddRsRs]					-> [^self concretizeSEEOpRsRs: 16r58].
		[AndCqR]					-> [^self concretizeAndCqR].
		[AndCwR]					-> [^self concretizeAndCwR].
		[AndRR]					-> [^self concretizeOpRR: 16r23].
		[TstCqR]					-> [^self concretizeTstCqR].
		[CmpCqR]					-> [^self concretizeCmpCqR].
		[CmpCwR]					-> [^self concretizeCmpCwR].
		[CmpRR]					-> [^self concretizeReverseOpRR: 16r39].
		[CmpRdRd]					-> [^self concretizeCmpRdRd].
		[CmpRsRs]					-> [^self concretizeCmpRsRs].
		[DivRdRd]					-> [^self concretizeSEE2OpRdRd: 16r5E].
		[DivRsRs]					-> [^self concretizeSEEOpRsRs: 16r5E].
		[MulRdRd]					-> [^self concretizeSEE2OpRdRd: 16r59].
		[MulRsRs]					-> [^self concretizeSEEOpRsRs: 16r59].
		[OrCqR]					-> [^self concretizeOrCqR].
		[OrCwR]					-> [^self concretizeOrCwR].
		[OrRR]						-> [^self concretizeOpRR: 16r0B].
		[SubCqR]					-> [^self concretizeSubCqR].
		[SubCwR]					-> [^self concretizeSubCwR].
		[SubRR]					-> [^self concretizeOpRR: 16r2B].
		[SubbRR]					-> [^self concretizeOpRR: 16r1B].
		[SubRdRd]					-> [^self concretizeSEE2OpRdRd: 16r5C].
		[SubRsRs]					-> [^self concretizeSEEOpRsRs: 16r5C].
		[SqrtRd]						-> [^self concretizeSqrtRd].
		[SqrtRs]						-> [^self concretizeSqrtRs].
		[XorCwR]						-> [^self concretizeXorCwR].
		[XorRR]							-> [^self concretizeOpRR: 16r33].
		[XorRdRd]						-> [^self concretizeXorRdRd].
		[XorRsRs]						-> [^self concretizeXorRsRs].
		[NegateR]						-> [^self concretizeNegateR].
		[NotR]							-> [^self concretizeNotR].
		[LoadEffectiveAddressMwrR]	-> [^self concretizeLoadEffectiveAddressMwrR].
		[ArithmeticShiftRightCqR]		-> [^self concretizeArithmeticShiftRightCqR].
		[LogicalShiftRightCqR]			-> [^self concretizeLogicalShiftRightCqR].
		[LogicalShiftLeftCqR]			-> [^self concretizeLogicalShiftLeftCqR].
		[ArithmeticShiftRightRR]		-> [^self concretizeArithmeticShiftRightRR].
		[LogicalShiftLeftRR]				-> [^self concretizeLogicalShiftLeftRR].
		[ClzRR]						-> [^self concretizeClzRR].
		"Data Movement"
		[MoveCqR]			-> [^self concretizeMoveCqR].
		[MoveCwR]			-> [^self concretizeMoveCwR].
		[MoveRR]			-> [^self concretizeReverseOpRR: 16r89].
		[MoveRdRd]		-> [^self concretizeMoveRdRd].
		[MoveRsRs]			-> [^self concretizeMoveRsRs].
		[MoveAwR]			-> [^self concretizeMoveAwR].
		[MoveRAw]			-> [^self concretizeMoveRAw].
		[MoveAbR]			-> [^self concretizeMoveAbR].
		[MoveRAb]			-> [^self concretizeMoveRAb].
		[MoveMbrR]		-> [^self concretizeMoveMbrR].
		[MoveRMbr]		-> [^self concretizeMoveRMbr].
		[MoveRM8r]		-> [^self concretizeMoveRMbr].
		[MoveM8rR]		-> [^self concretizeMoveM8rR].
		[MoveM16rR]		-> [^self concretizeMoveM16rR].
		[MoveRM16r]		-> [^self concretizeMoveRM16r].
		[MoveM32rR]		-> [^self concretizeMoveMwrR].
		[MoveRM32r]		-> [^self concretizeMoveRMwr].
		[MoveM32rRs]		-> [^self concretizeMoveM32rRs].
		[MoveRsM32r]		-> [^self concretizeMoveRsM32r].
		[MoveM64rRd]		-> [^self concretizeMoveM64rRd].
		[MoveMwrR]		-> [^self concretizeMoveMwrR].
		[MoveXbrRR]		-> [^self concretizeMoveXbrRR].
		[MoveRXbrR]		-> [^self concretizeMoveRXbrR].
		[MoveXwrRR]		-> [^self concretizeMoveXwrRR].
		[MoveRXwrR]		-> [^self concretizeMoveRXwrR].
		[MoveRMwr]		-> [^self concretizeMoveRMwr].
		[MoveRdM64r]		-> [^self concretizeMoveRdM64r].
		[PopR]				-> [^self concretizePopR].
		[PushR]				-> [^self concretizePushR].
		[PushCq]			-> [^self concretizePushCq].
		[PushCw]			-> [^self concretizePushCw].
		[PrefetchAw]		-> [^self concretizePrefetchAw].
		"Conversion"
		[ConvertRRd]		-> [^self concretizeConvertRRd].
		[ConvertRdR]		-> [^self concretizeConvertRdR].
		
		[ConvertRsRd]		-> [^self concretizeConvertRsRd].
		[ConvertRdRs]		-> [^self concretizeConvertRdRs].
		[ConvertRsR]		-> [^self concretizeConvertRsR].
		[ConvertRRs]		-> [^self concretizeConvertRRs].
			
		[SignExtend8RR]		-> [^self concretizeSignExtend8RR].
		[SignExtend16RR]		-> [^self concretizeSignExtend16RR].
		
		[ZeroExtend8RR]		-> [^self concretizeZeroExtend8RR].
		[ZeroExtend16RR]		-> [^self concretizeZeroExtend16RR].

		"miscellaneous"
		[MovePerfCnt64RRL]	->	[^self concretizeMovePerfCnt64RRL]
		 }.

	^0 "keep Slang happy"! !

!CogIA32Compiler methodsFor: 'generate machine code' stamp: 'eem 10/25/2020 17:39'!
dispatchConcretizeProcessorSpecific
	"Attempt to generate concrete machine code for the instruction at address.
	 This is part of the inner dispatch of concretizeAt: actualAddress which exists only
	 to get around the number of literals limits in the SqueakV3 (blue book derived)
	 bytecode set."
	opcode caseOf: {
		[CPUID]					-> [^self concretizeCPUID].
		"Specific Arithmetic"
		[CDQ]					-> [^self concretizeCDQ].
		[IDIVR]					-> [^self concretizeIDIVR].
		[IMULRR]				-> [^self concretizeMulRR].
		[XCHGRR]				-> [^self concretizeXCHGRR].
		"Specific Control/Data Movement"
		[REP]					-> [^self concretizeREP].
		[CLD]					-> [^self concretizeCLD].
		[MOVSB]				-> [^self concretizeMOVSB].
		[MOVSD]				-> [^self concretizeMOVSD].
		[BSR]					-> [^self concretizeBSR].
		[FSTPS]					-> [^self concretizeFSTPS].
		[FSTPD]				-> [^self concretizeFSTPD].
		"Multi-processing"
		[CMPXCHGRAw]			-> [^self concretizeCMPXCHGRAw].
		"[LFENCE]				-> [^self concretizeFENCE: 5]."
		[MFENCE]				-> [^self concretizeFENCE: 6].
		[SFENCE]				-> [^self concretizeFENCE: 7].
		[LOCK]					-> [^self concretizeLOCK].
		[SETE]					-> [^self concretizeSet: 4].
		"[XCHGAwR]			-> [^self concretizeXCHGAwR]."
		"[XCHGMwrR]			-> [^self concretizeXCHGMwrR]."
	}.
	^0! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 2/4/2020 12:06'!
flushDCacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	"If there is a dual mapped code zone (the normal zone but marked with read/execute, and a
	 read/write zone codeToDataDelta bytes away) then flush the data cache for the corresponding
	 range in the read/write zone and invalidate the data cache for the read/execute zone."
	<cmacro: '(me,startAddress,endAddress) 0'>
	"Currently we don't know whether a dual zone on an Intel x86/x86-64 would require cache
	 flushing to ensure coherence.  So for now this is a no op."! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 2/3/2020 19:32'!
flushICacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	"Flush the instruction cache from (startAddress to endAddress].
	 If there is a dual mapped code zone (the normal zone but marked with read/execute, and a
	 read/write zone codeToDataDelta bytes away) then also flush the data cache for the corresp-
	 onding range in the read/write zone and invalidate the data cache for the read/execute zone."
	<cmacro: '(me,startAddress,endAddress) 0'>
	"On Intel processors where code and data have the same linear address, no
	 special action is required to flush the instruction cache.  One only needs to
	 execute a serializing instruction (e.g. CPUID) if code and data are at different
	 virtual addresses (e.g. a debugger using memory-mapping to access a debugee).
	 Using the macro avoids an unnecessary call."! !

!CogIA32Compiler methodsFor: 'abi' stamp: 'eem 4/20/2015 18:22'!
fullCallsAreRelative
	"Answer if CallFull and/or JumpFull are relative and hence need relocating on method
	 compation. If so, they are annotated with IsRelativeCall in methods and relocated in
	 relocateIfCallOrMethodReference:mcpc:delta:"
	^true! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'eem 12/25/2015 18:17'!
genDivR: abstractRegDivisor R: abstractRegDividend Quo: abstractRegQuotient Rem: abstractRegRemainder
	| rDividend rDivisor rQuotient rRemainder saveRestoreEAX saveRestoreEDX saveRestoreExchanged |
	self assert: abstractRegDividend ~= abstractRegDivisor.
	self assert: abstractRegQuotient ~= abstractRegRemainder.
	rDividend := abstractRegDividend.
	rDivisor := abstractRegDivisor.
	rQuotient := abstractRegQuotient.
	rRemainder := abstractRegRemainder.
	"IDIV r does a signed divide of EDX:EAX by r, EAX := Quotient, EDX := Remainder.
	 Since we must sign extend the dividend into EDX we must substitute another register if EDX is an input"
	(rDividend = EDX or: [rDivisor = EDX]) ifTrue:
		[| rUnused |
		"Slang, sigh..."
		rUnused := EAX.
		[rUnused <= EDI] whileTrue:
			[(rUnused ~= ESP and: [rUnused ~= EBP and: [rUnused ~= EDX
			  and: [rUnused ~= rDividend and: [rUnused ~= rDivisor
			  and: [rUnused ~= rQuotient and: [rUnused ~= rRemainder]]]]]]) ifTrue:
				[cogit PushR: rUnused.
				cogit MoveR: EDX R: rUnused.
				rDividend = EDX
					ifTrue: [self genDivR: rDivisor R: rUnused Quo: rQuotient Rem: rRemainder]
					ifFalse: [self genDivR: rUnused R: rDividend Quo: rQuotient Rem: rRemainder].
				cogit PopR: rUnused.
				^self].
			  rUnused := rUnused + 1].
		self error: 'couldn''t find unused register in genDivR:R:Quo:Rem:'].
	"If either output does not include EAX or EDX we must save and restore EAX and/or EDX."
	(saveRestoreEAX := rQuotient ~= EAX and: [rRemainder ~= EAX]) ifTrue:
		[cogit PushR: EAX].
	(saveRestoreEDX := rQuotient ~= EDX and: [rRemainder ~= EDX]) ifTrue:
		[cogit PushR: EDX].
	saveRestoreExchanged := -1.
	rDividend ~= EAX ifTrue:
		[rDivisor = EAX
			ifTrue: [((rDividend ~= rQuotient and: [rDividend ~= rRemainder])
					and: [rDividend ~= EDX or: [saveRestoreEDX not]]) ifTrue:
						[cogit PushR: (saveRestoreExchanged := rDividend)].
					cogit gen: XCHGRR operand: rDivisor operand: rDividend]
			ifFalse: [cogit MoveR: rDividend R: EAX]].
	"CDQ sign-extends EAX into EDX as required for IDIV"
	cogit gen: CDQ.
	cogit gen: IDIVR operand: (rDivisor = EAX ifTrue: [rDividend] ifFalse: [rDivisor]).
	"Must not overwrite result while juggling"
	(rQuotient = EDX and: [rRemainder = EAX])
		ifTrue: [cogit gen: XCHGRR operand: rQuotient operand: rRemainder]
		ifFalse:
			[rQuotient = EDX
				ifTrue:
					[rRemainder ~= EDX ifTrue:
						[cogit MoveR: EDX R: rRemainder].
					rQuotient ~= EAX ifTrue:
						[cogit MoveR: EAX R: rQuotient]]
				ifFalse:
					[rQuotient ~= EAX ifTrue:
						[cogit MoveR: EAX R: rQuotient].
					rRemainder ~= EDX ifTrue:
						[cogit MoveR: EDX R: rRemainder]]].
	saveRestoreExchanged >= 0 ifTrue:
		[cogit PopR: saveRestoreExchanged].
	saveRestoreEDX ifTrue:
		[cogit PopR: EDX].
	saveRestoreEAX ifTrue:
		[cogit PopR: EAX]! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'nice 2/17/2020 21:28'!
genHighBitAlternativeIn: destReg ofSmallIntegerOopWithSingleTagBit: srcReg
	"When CLZ is not available, we can use BSR (Bit Scan Reverse) instruction"
	<inline: true>
	| jumpNegativeReceiver |
	<var: #jumpNegativeReceiver type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	"The primitive must fail if receiver is negative"
	cogit CmpCq: 0 R: srcReg. "N.B. FLAGS := srcReg - 0"
	jumpNegativeReceiver := cogit JumpNegative: 0.
	cogit gen: BSR operand: srcReg operand: destReg.
	"theoretically we should handle case when srcReg is zero, but we do not care because it is never zero thanks to the tagBit"
	"and thanks to the tag bit, the +1 operation for getting 1-based rank instead of 0-based rank is not necessary, so we are done"
	^jumpNegativeReceiver! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'nice 9/7/2019 08:40'!
genJumpFPEqual: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genJumpFPOrderedAnd: JumpFPEqual to: jumpTarget! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'nice 9/7/2019 09:11'!
genJumpFPLess: jumpTarget
	"Note: this generates two jumps for handling case of NaN operands
	Sender might prefer using a JumpFPGreater with swapped operands."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genJumpFPOrderedAnd: JumpFPLess to: jumpTarget! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'nice 9/7/2019 09:12'!
genJumpFPLessOrEqual: jumpTarget
	"Note: this generates two jumps for handling case of NaN operands
	Sender might prefer using a JumpFPGreaterOrEqual with swapped operands."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genJumpFPOrderedAnd: JumpFPLessOrEqual to: jumpTarget! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'nice 9/7/2019 08:51'!
genJumpFPNotEqual: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genJumpFPUnorderedOr: JumpFPNotEqual to: jumpTarget! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'nice 9/11/2019 00:28'!
genJumpFPOrderedAnd: jumpOpCode to: jumpTarget
	"Test the comparison status flags ZF PF CF and jump to jumpTarget when
	- both Ordered (no nan operand) - a JNP
	- and the other condition (determined by jumpOpCode) is true
	jumpTarget maybe either an Integer offset, or an AbstractInstruction.
	This is necessary on Intel IA32 for JB JBE and JE with nan operands,
	otherwise unordered condition set the three flags to 1 and cause false positive.
	The implementation use JP to jump over the over test
	rather than link a JNP and jumpOpCode jumpTarget"
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	| jumpUnordered jumpToTarget |
	<var: #jumpUnordered type: #'AbstractInstruction *'>
	<var: #jumpToTarget type: #'AbstractInstruction *'>
	jumpUnordered := cogit gen: JumpFPUnordered operand: jumpTarget asInteger.
	jumpToTarget := cogit gen: jumpOpCode operand: jumpTarget asInteger.
	jumpUnordered jmpTarget: cogit Label.
	^jumpToTarget! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'nice 9/7/2019 08:51'!
genJumpFPUnorderedOr: jumpOpCode to: jumpTarget
	"Test the comparison status flags ZF PF CF and jump to jumpTarget when
	- either Unordered (no nan opearand) - a JP
	- and the other condition (determined by jumpOpCode) is true
	jumpTarget maybe either an Integer offset, or an AbstractInstruction.
	This is necessary on Intel IA32 for JNE with nan operands,
	otherwise unordered condition set the three flags to 1 and cause false negative."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	| jumpUnordered jumpToTarget |
	<var: #jumpUnordered type: #'AbstractInstruction *'>
	<var: #jumpToTarget type: #'AbstractInstruction *'>
	jumpToTarget := cogit gen: JumpFPNotEqual operand: jumpTarget asInteger.
	jumpUnordered := cogit gen: JumpFPUnordered operand: jumpTarget asInteger.
	jumpToTarget addDependent: jumpUnordered.
	^jumpToTarget! !

!CogIA32Compiler methodsFor: 'abi' stamp: 'eem 12/18/2015 18:19'!
genMarshallNArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3
	"Generate the code to pass up to four arguments in a C run-time call.  Hack: each argument is
	 either a negative number, which encodes a constant, or a non-negative number, that of a register.

	 Run-time calls have no more than four arguments, so chosen so that on ARM, where in its C ABI the
	 first four integer arguments are passed in registers, all arguments can be passed in registers.  We
	 defer to the back end to generate this code not so much that the back end knows whether it uses
	 the stack or registers to pass arguments (it does, but...). In fact we defer for an extremely evil reason.
	 Doing so allows the x64 (where up to 6 args are passed) to assign the register arguments in an order
	 that allows some of the argument registers to be used for specific abstract  registers, specifically
	 ReceiverResultReg and ClassReg.  This is evil, evil, evil, but also it's really nice to keep using the old
	 register assignments the original author has grown accustomed to."
	<inline: true>
	numArgs = 0 ifTrue:
		[^self].
	numArgs > 1 ifTrue:
		[numArgs > 2 ifTrue:
			[numArgs > 3 ifTrue:
				[(cogit isTrampolineArgConstant: regOrConst3)
					ifFalse: [cogit PushR: regOrConst3]
					ifTrue: [cogit PushCq: (cogit trampolineArgValue: regOrConst3)]].
			 (cogit isTrampolineArgConstant: regOrConst2)
				ifFalse: [cogit PushR: regOrConst2]
				ifTrue: [cogit PushCq: (cogit trampolineArgValue: regOrConst2)]].
		(cogit isTrampolineArgConstant: regOrConst1)
			ifFalse: [cogit PushR: regOrConst1]
			ifTrue: [cogit PushCq: (cogit trampolineArgValue: regOrConst1)]].
	(cogit isTrampolineArgConstant: regOrConst0)
		ifFalse: [cogit PushR: regOrConst0]
		ifTrue: [cogit PushCq: (cogit trampolineArgValue: regOrConst0)]! !

!CogIA32Compiler methodsFor: 'abi' stamp: 'RonieSalgado 5/2/2016 14:27'!
genMarshallNArgs: numArgs floatArg: regOrConst0 floatArg: regOrConst1 floatArg: regOrConst2 floatArg: regOrConst3
	"Generate the code to pass up to four arguments in a C run-time call.  Hack: each argument is
	 either a negative number, which encodes a constant, or a non-negative number, that of a register.

	 Run-time calls have no more than four arguments, so chosen so that on ARM, where in its C ABI the
	 first four integer arguments are passed in registers, all arguments can be passed in registers.  We
	 defer to the back end to generate this code not so much that the back end knows whether it uses
	 the stack or registers to pass arguments (it does, but...). In fact we defer for an extremely evil reason.
	 Doing so allows the x64 (where up to 6 args are passed) to assign the register arguments in an order
	 that allows some of the argument registers to be used for specific abstract  registers, specifically
	 ReceiverResultReg and ClassReg.  This is evil, evil, evil, but also it's really nice to keep using the old
	 register assignments the original author has grown accustomed to."
	<inline: true>
	numArgs = 0 ifTrue:
		[^self].
	numArgs > 1 ifTrue:
		[numArgs > 2 ifTrue:
			[numArgs > 3 ifTrue:
				[(cogit isTrampolineArgConstant: regOrConst3)
					ifFalse: [cogit PushRd: regOrConst3]
					ifTrue: [cogit PushC64: (cogit trampolineArgValue: regOrConst3)]].
			 (cogit isTrampolineArgConstant: regOrConst2)
				ifFalse: [cogit PushRd: regOrConst2]
				ifTrue: [cogit PushC64: (cogit trampolineArgValue: regOrConst2)]].
		(cogit isTrampolineArgConstant: regOrConst1)
			ifFalse: [cogit PushRd: regOrConst1]
			ifTrue: [cogit PushC64: (cogit trampolineArgValue: regOrConst1)]].
	(cogit isTrampolineArgConstant: regOrConst0)
		ifFalse: [cogit PushRd: regOrConst0]
		ifTrue: [cogit PushC64: (cogit trampolineArgValue: regOrConst0)]! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 16:14'!
genMemCopy: originalSourceReg to: originalDestReg constantSize: size
	| numbytes numwords sourceReg destReg countReg inst |

	"Get the abstract registers for ECX, EDI and ESI"
	sourceReg := ESI.
	destReg := EDI.
	countReg := ECX.

	"Put the source in ESI and the dest in EDI"
	inst := cogit Label.
	originalSourceReg ~= sourceReg ifTrue: [
		originalDestReg = sourceReg ifTrue: [
			cogit MoveR: originalDestReg R: TempReg.
		].
		cogit MoveR: originalSourceReg R: sourceReg.
	].

	originalDestReg ~= destReg ifTrue: [
		originalDestReg = sourceReg ifTrue: [
			cogit MoveR: TempReg R: destReg.
		] ifFalse: [
			cogit MoveR: originalDestReg R: destReg.
		]
	].

	"Store the count and clear the direction"
	cogit gen: CLD.

	"First copy the bytes"
	numbytes := size bitAnd: 16r03.
	numbytes > 0 ifTrue: [
		cogit MoveCq: numbytes R: countReg.
		cogit gen: REP.
		cogit gen: MOVSB.
	].

	"Now copy the words"
	numwords := size // 4.
	cogit MoveCq:  numwords R: countReg.
	cogit gen: REP.
	cogit gen: MOVSD.

! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 16:14'!
genMemCopy: originalSourceReg to: originalDestReg size: originalSize
	| spilledSize size sourceReg destReg countReg inst |

	"Get the abstract registers for ECX, EDI and ESI"
	sourceReg := ESI.
	destReg := EDI.
	countReg := ECX.

	"TODO: Avoid spilling"
	spilledSize := false.
	(originalSize = sourceReg or: [originalSize = destReg]) ifTrue: [
		cogit PushR: originalSize.
		spilledSize := true.
	].

	"Put the source in ESI and the dest in EDI"
	inst := cogit Label.
	originalSourceReg ~= sourceReg ifTrue: [
		originalDestReg = sourceReg ifTrue: [
			cogit MoveR: originalDestReg R: TempReg.
		].
		cogit MoveR: originalSourceReg R: sourceReg.
	].

	originalDestReg ~= destReg ifTrue: [
		originalDestReg = sourceReg ifTrue: [
			cogit MoveR: TempReg R: destReg.
		] ifFalse: [
			cogit MoveR: originalDestReg R: destReg.
		]
	].

	"Put the original size register in somewhere different than ECX"
	spilledSize ifTrue: [
		cogit PopR: TempReg.
		size := TempReg.
	] ifFalse: [
		originalSize = countReg ifTrue: [
			cogit MoveR: originalSize R: TempReg.
			size := TempReg.
		] ifFalse: [
			size := originalSize.
		]
	].

	cogit gen: CLD.

	"First copy the bytes"
	cogit MoveR: size R: countReg.
	cogit AndCq: 16r03 R: countReg.
	cogit gen: REP.
	cogit gen: MOVSB.

	"Now copy the words"
	cogit MoveR: size R: countReg.
	cogit LogicalShiftRightCq: 2 R: countReg.
	cogit gen: REP.
	cogit gen: MOVSD.

! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'RonieSalgado 5/2/2016 15:50'!
genMoveCf32: constantFloat32 Rs: register
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #constantFloat32 type: #float>
	| inst |
	inst := cogit PushCw: constantFloat32 asIEEE32BitWord.
	cogit PopRs: register.
	^ inst! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'RonieSalgado 5/2/2016 15:50'!
genMoveCf64: constantFloat64 Rd: register
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #constantFloat64 type: #double>
	| inst |
	inst := cogit PushC64: constantFloat64 asIEEE64BitWord.
	cogit PopRd: register.
	^ inst! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'eem 12/16/2015 09:44'!
genMulR: regSource R: regDest
	^cogit gen: IMULRR operand: regSource operand: regDest! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'RonieSalgado 5/2/2016 14:52'!
genPopRd: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	| inst |
	inst := cogit MoveM64: 0 r: SPReg Rd: reg .
	cogit AddCq: 8 R: SPReg.
	^ inst! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'RonieSalgado 5/2/2016 14:52'!
genPopRs: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	| inst |
	inst := cogit MoveM32: 0 r: SPReg Rs: reg .
	cogit AddCq: 4 R: SPReg.
	^ inst! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'eem 3/11/2017 12:53'!
genPushC64: constant64Bits
	<inline: true>
	<var: #constant64Bits type: #sqLong>
	<returnTypeC: #'AbstractInstruction *'>
	| inst highPart |
	self assert: BytesPerWord = 4.
	highPart := self cCode: '(constant64Bits >> 32LL)' inSmalltalk: [ constant64Bits bitShift: -32 ].
	inst :=cogit PushCw: highPart.
	cogit PushCw: (constant64Bits bitAnd: 16rFFFFFFFF).
	^inst! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 16:16'!
genPushRd: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	| inst |
	inst := cogit MoveRd: reg M64: -8 r: SPReg.
	cogit SubCq: 8 R: SPReg.
	^ inst! !

!CogIA32Compiler methodsFor: 'smalltalk calling convention' stamp: 'eem (const to msg rewrite) 11/6/2014 11:51'!
genPushRegisterArgsForAbortMissNumArgs: numArgs
	"Ensure that the register args are pushed before the outer and
	 inner retpcs at an entry miss for arity <= self numRegArgs.  The
	 outer retpc is that of a call at a send site.  The inner is the call
	 from a method or PIC abort/miss to the trampoline."

	"This won't be as clumsy on a RISC.  But putting the receiver and
	 args above the return address means the CoInterpreter has a
	 single machine-code frame format which saves us a lot of work."

	"Iff there are register args convert
		base	->	outerRetpc		(send site retpc)
		sp		->	innerRetpc		(PIC abort/miss retpc)
	 to
		base	->	receiver
					(arg0)
					(arg1)
					outerRetpc
		sp		->	innerRetpc		(PIC abort/miss retpc)"
	numArgs <= cogit numRegArgs ifTrue:
		[self assert: cogit numRegArgs <= 2.
		 numArgs = 0 ifTrue:
			[cogit MoveMw: 0 r: SPReg R: TempReg.
			 cogit PushR: TempReg.
			 cogit MoveMw: objectMemory wordSize * 2 r: SPReg R: TempReg.
			 cogit MoveR: TempReg Mw: objectMemory wordSize r: SPReg.
			 cogit MoveR: ReceiverResultReg Mw: 2 * objectMemory wordSize r: SPReg.
			 ^self].
		 numArgs = 1 ifTrue:
			[cogit MoveMw: objectMemory wordSize r: SPReg R: TempReg.
			 cogit PushR: TempReg.
			 cogit MoveMw: objectMemory wordSize r: SPReg R: TempReg.
			 cogit PushR: TempReg.
			 cogit MoveR: ReceiverResultReg Mw: 3 * objectMemory wordSize r: SPReg.
			 cogit MoveR: Arg0Reg Mw: 2 * objectMemory wordSize r: SPReg.
			 ^self].
		 numArgs = 2 ifTrue:
			[cogit PushR: Arg1Reg.
			 cogit MoveMw: objectMemory wordSize * 2 r: SPReg R: TempReg.
			 cogit PushR: TempReg.
			 cogit MoveMw: objectMemory wordSize * 2 r: SPReg R: TempReg.
			 cogit PushR: TempReg.
			 cogit MoveR: ReceiverResultReg Mw: 4 * objectMemory wordSize r: SPReg.
			 cogit MoveR: Arg0Reg Mw: 3 * objectMemory wordSize r: SPReg.
			 ^self]]! !

!CogIA32Compiler methodsFor: 'smalltalk calling convention' stamp: 'eem 6/1/2015 16:44'!
genPushRegisterArgsForNumArgs: numArgs scratchReg: scratchReg
	"Ensure that the register args are pushed before the retpc for arity <= self numRegArgs.  This
	 isn't as clumsy on a RISC.  But putting the receiver and args above the return address
	 means the CoInterpreter has a single machine-code frame format which saves us a lot of work.
	 N.B. Take great care to /not/ smash TempReg, which is used in directed send marshalling.
	 We could use XCHG to swap the ReceiverResultReg and top-of-stack return address, pushing the
	 the ret pc (now in ReceiverResultReg) later, but XCHG is very slow.  We can use SendNumArgsReg
	 because it is only live in sends of arity >= (NumSendTrampolines - 1)."
	self assert: cogit numRegArgs < (NumSendTrampolines - 1).
	numArgs <= cogit numRegArgs ifTrue:
		[self assert: cogit numRegArgs <= 2.
		 false "these two variants show the same performance on Intel Core i7, but the second one may be shorter."
			ifTrue:
				[cogit MoveMw: 0 r: SPReg R: scratchReg. "Save return pc"
				 numArgs > 0 ifTrue:
					[cogit PushR: Arg0Reg.
					 numArgs > 1 ifTrue:
						[cogit PushR: Arg1Reg]].
				 cogit PushR: scratchReg.
				 cogit MoveR: ReceiverResultReg Mw: objectMemory wordSize * (1 + numArgs) r: SPReg]
			ifFalse:
				["a.k.a.
					cogit gen: XCHGMwrR operand: 0 operand: SPReg operand: ReceiverResultReg.
				  but XCHG is slow."
				 cogit MoveMw: 0 r: SPReg R: scratchReg. "Save return pc"
				 cogit MoveR: ReceiverResultReg Mw: 0 r: SPReg.
				 numArgs > 0 ifTrue:
					[cogit PushR: Arg0Reg.
					 numArgs > 1 ifTrue:
						[cogit PushR: Arg1Reg]].
				 cogit PushR: scratchReg]] "Restore return address"! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'RonieSalgado 4/29/2016 16:16'!
genPushRs: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	| inst |
	inst := cogit MoveRs: reg M32: -4 r: SPReg.
	cogit SubCq: 4 R: SPReg.
	^ inst! !

!CogIA32Compiler methodsFor: 'abi' stamp: 'eem 5/29/2017 10:25'!
genRemoveNArgsFromStack: n
	<inline: true>
	n = 0 ifFalse: [cogit AddCq: n * 4 R: ESP].
	^0! !

!CogIA32Compiler methodsFor: 'abi' stamp: 'eem 5/29/2017 10:26'!
genRemoveNFloatArgsFromStack: n
	<inline: true>
	n > 0 ifTrue: [cogit AddCq: n * 8 R: ESP].
	^0! !

!CogIA32Compiler methodsFor: 'abi' stamp: 'eem 3/3/2016 14:11'!
genRestoreRegs: regMask
	"Restore the registers in regMask as saved by genSaveRegs:."
	self deny: (regMask anyMask: (cogit registerMaskFor: ESP and: EBP)).
	EAX to: EDI do:
		[:reg|
		 (regMask anyMask: (cogit registerMaskFor: reg)) ifTrue:
			[cogit PopR: reg]].
	^0! !

!CogIA32Compiler methodsFor: 'abi' stamp: 'eem 3/3/2016 12:44'!
genSaveRegs: regMask
	"Save the registers in regMask for a call into the C run-time from a trampoline."

	self assert: (EDI > EAX and: [EDI - EAX + 1 = 8]).
	self deny: (regMask anyMask: (cogit registerMaskFor: ESP and: EBP)).
	EDI to: EAX by: -1 do:
		[:reg|
		 (regMask anyMask: (cogit registerMaskFor: reg)) ifTrue:
			[cogit PushR: reg]].
	^0! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'eem 5/6/2009 10:40'!
genSubstituteReturnAddress: retpc
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^cogit PushCw: retpc! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'eem 4/1/2020 13:19'!
genSubstituteReturnAddressR: retpcReg
	<inline: true>
	^cogit PushR: retpcReg! !

!CogIA32Compiler methodsFor: 'abstract instructions' stamp: 'eem 2/15/2017 13:30'!
genSwapR: regA R: regB Scratch: regTmp
	^cogit gen: XCHGRR operand: regA operand: regB! !

!CogIA32Compiler methodsFor: 'disassembly' stamp: 'eem 12/4/2013 12:49'!
generalPurposeRegisterMap
	<doNotGenerate>
	"Answer a Dictionary from register getter to register index."
	^Dictionary newFromPairs:
		{	#eax. EAX.
			#ecx. ECX.
			#edx. EDX.
			#ebx. EBX.
			#esi.  ESI.
			#edi.  EDI	}! !

!CogIA32Compiler methodsFor: 'multi-threading' stamp: 'eem 10/25/2020 17:42'!
generateLowLevelTryLock: vmOwnerLockAddress
	"Generate a function that attempts to lock the vmOwnerLock and answers
	 true if it succeeded."
	<inline: true>
	| valueReg |
	vmOwnerLockAddress = 0 ifTrue:
		[cogit
			MoveCq: 1 R: ABIResultReg;
			RetN: 0.
		 ^self].
	valueReg := cogit availableRegisterOrNoneIn: (ABICallerSavedRegisterMask bitClear: 1 << EAX).
	cogit
		MoveMw: 4 r: ESP R: valueReg;
		MoveCq: 0 R: EAX;
		gen: LOCK;
		gen: CMPXCHGRAw operand: valueReg operand: vmOwnerLockAddress;
		gen: SETE operand: ABIResultReg; "a.k.a. EAX"
		RetN: 0! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
hasConditionRegister
	"Answer if the receiver supports, e.g., JumpOverflow after a regular AddRR"
	^true! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 3/28/2015 23:52'!
hasDoublePrecisionFloatingPointSupport
	"We can generate dpfp support if the processor has SSE2 instructions."
	<inline: true>
	^self hasSSE2Instructions! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 10/29/2020 22:17'!
hasLZCNTInstructions
	"Answer if we support LZCNT"
	<inline: true>
	^self cpuidWord1 anyMask: 1 << 5! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 11/20/2010 18:37'!
hasLinkRegister
	^false! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 10/29/2020 21:29'!
hasSSE2Instructions
	"Answer if we support SSE2"
	^self cpuidWord0 anyMask: 1 << 26! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 10/29/2020 21:40'!
hasSSEInstructions
	"Answer if we support SSE"
	^self cpuidWord0 anyMask: 1 << 25! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
hasThreeAddressArithmetic
	"Answer if the receiver supports three-address arithmetic instructions"
	^false! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 7/9/2020 10:30'!
initialFlushICacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	"This must be implemented only if a back end defines any of the following as non-zero:
		numCheckFeaturesOpcodes
		numDCacheFlushOpcodes
		numICacheFlushOpcode.

	 This is a no-op on IA32."
	<cmacro: '(me,startAddress,endAddress) 0'>! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 2/28/2009 05:32'!
inlineCacheTagAt: callSiteReturnAddress
	"Answer the inline cache tag for the return address of a send."
	^self literalBeforeFollowingAddress: callSiteReturnAddress - 5! !

!CogIA32Compiler methodsFor: 'disassembly' stamp: 'eem 12/9/2015 12:56'!
instructionSizeAt: pc
	"Answer the instruction size at pc.  This is very far from a full decode.
	 It only has to cope with the instructions generated in a block dispatch."
	| op |
	op := objectMemory byteAt: pc.
	^op caseOf:
		{	[16r0F]	->	[self twoByteInstructionSizeAt: pc].
			[16r3D]	->	[5]. "cmp EAX,imm32"
			[16r70]	->	[2]. "short conditional jumps"
			[16r71]	->	[2].
			[16r72]	->	[2].
			[16r73]	->	[2].
			[16r74]	->	[2].
			[16r75]	->	[2].
			[16r76]	->	[2].
			[16r77]	->	[2].
			[16r78]	->	[2].
			[16r79]	->	[2].
			[16r7A]	->	[2].
			[16r7B]	->	[2].
			[16r7C]	->	[2].
			[16r7D]	->	[2].
			[16r7E]	->	[2].
			[16r7F]	->	[2].
			[16r83]	->	[self sizeImmediateGroup1: op at: pc].
			[16r89]	->	[2]. "MOV Eb,Gb"
			[16r8B]	->	[self sizeHasModrm: op at: pc].
			[16r90]	->	[1]. "nop"
			[16rE9] ->	[5]. "long unconditional jump"
			[16rEB] ->	[2]. "short unconditional jump"
			[16rCC]	->	[1]. "int3" }! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
isBigEndian
	^false! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 12/3/2015 22:26'!
isCallPrecedingReturnPC: mcpc
	"Assuming mcpc is a send return pc answer if the instruction before it is a call (not a CallFull)."
	^(objectMemory byteAt: mcpc - 5) = 16rE8! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 10/13/2009 12:44'!
isJumpAt: pc
	| op |
	op := objectMemory byteAt: pc.
	^  (op between: 16r70 and: 16r7F) "short conditional jumps"
	or: [op = 16rE9 "long unconditional jump"
	or: [op = 16rEB "short unconditional jump"
	or: [op = 16r0F "long conditional jumps"
		and: [(objectMemory byteAt: pc + 1) between: 16r80 and: 16r8F]]]]! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 12/27/2008 18:36'!
isPCDependent
	"Answer if the receiver is a pc-dependent instruction."
	^self isJump or: [opcode = AlignmentNops]! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'nice 11/6/2016 14:24'!
isQuick: operand
	<var: #operand type: #'usqIntptr_t'>
	^operand signedIntFromLong between: -128 and: 127! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'eem 6/8/2015 16:56'!
jmpTarget: anAbstractInstruction
	"Set the target of a jump instruction.  These all have the target in the first operand.
	 Override to cope with JumpFPNotEqual where because of IEEE NaN conformance and
	 the behaviour of COMISD/UCOMISD we generate two jumps to the same target."
	| aDependent |
	<var: #aDependent type: #'AbstractInstruction *'>
	aDependent := dependent.
	[aDependent notNil] whileTrue:
		[aDependent jmpTarget: anAbstractInstruction.
		 aDependent := aDependent dependent].
	^super jmpTarget: anAbstractInstruction! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'tpr 4/17/2015 19:56'!
jumpLongByteSize
"	Branch/Call ranges.  Jump[Cond] can be generated as short as possible.  Call/Jump[Cond]Long must be generated
	in the same number of bytes irrespective of displacement since their targets may be updated, but they need only
	span 16Mb, the maximum size of the code zone.  This allows e.g. ARM to use single-word call and jump instructions
	for most calls and jumps.  CallFull/JumpFull must also be generated in the same number of bytes irrespective of
	displacement for the same reason, but they must be able to span the full (32-bit or 64-bit) address space because
	they are used to call code in the C runtime, which may be distant from the code zone"
	^5! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'eem 3/11/2009 22:24'!
jumpLongConditionalByteSize
	^6! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 09/28/2015 18:48'!
jumpLongTargetBeforeFollowingAddress: mcpc 
	"Answer the target address for the long jump immediately preceding mcpc"
	^self callTargetFromReturnAddress: mcpc! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'eem 6/24/2011 18:46'!
jumpShortByteSize
	^2! !

!CogIA32Compiler methodsFor: 'disassembly' stamp: 'cb 4/28/2015 16:17'!
jumpTargetPCAt: pc
	<returnTypeC: #usqInt>
	| size byte offset |
	size := self instructionSizeAt: pc.
	size = 2
		ifTrue:
			[byte := objectMemory byteAt: pc + 1.
			 offset := (byte bitAnd: 16r80) = 0 ifTrue: [byte] ifFalse: [byte - 256]]
		ifFalse:
			[byte := objectMemory byteAt: pc + size - 1.
			 offset := (byte bitAnd: 16r80) = 0 ifTrue: [byte] ifFalse: [byte - 256].
			 offset := offset << 8 + (objectMemory byteAt: pc + size - 2).
			 offset := offset << 8 + (objectMemory byteAt: pc + size - 3).
			 offset := offset << 8 + (objectMemory byteAt: pc + size - 4)].
	^pc + size + offset! !

!CogIA32Compiler methodsFor: 'abi' stamp: 'eem 6/1/2020 17:20'!
leafCallStackPointerDelta
	"Answer the delta from the stack pointer after a call to the stack pointer
	 immediately prior to the call.  This is used to compute the stack pointer
	 immediately prior to a call from within a leaf routine, which in turn is used
	 to capture the c stack pointer to use in trampolines back into the C run-time."
	^4! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 09/28/2015 18:48'!
literalBeforeFollowingAddress: followingAddress
	"Answer the literal embedded in the instruction immediately preceding followingAddress."
	^  ((objectMemory byteAt: followingAddress - 1) << 24)
	+  ((objectMemory byteAt: followingAddress - 2) << 16)
	+  ((objectMemory byteAt: followingAddress - 3) << 8)
	+   (objectMemory byteAt: followingAddress - 4)! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 4/6/2015 09:19'!
literalBeforeInlineCacheTagAt: callSiteReturnAddress
	"Answer a literal loaded before the inline cache tag load for the return address of a send."
	^self literalBeforeFollowingAddress: callSiteReturnAddress - 10! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'eem 3/11/2009 22:40'!
loadLiteralByteSize
	"Answer the byte size of a MoveCwR opcode's corresponding machine code"
	^5! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'eem 4/21/2015 17:29'!
loadPICLiteralByteSize
	"Answer the byte size of a MoveCwR opcode's corresponding machine code
	 when the argument is a PIC.  This is for the self-reference at the end of a
	 closed PIC."
	<inline: true>
	^self loadLiteralByteSize! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'eem 9/22/2015 08:59'!
machineCodeAt: anOffset
	^machineCode at: anOffset! !

!CogIA32Compiler methodsFor: 'generate machine code' stamp: 'eem 7/17/2012 14:27'!
machineCodeBytes
	"Answer the maximum number of bytes of machine code generated for any abstract instruction.
	 e.g. lock movsd  0x400(%esp),%xmm4 => f0 f2 0f 10 a4 24 00 04 00 00"
	^10! !

!CogIA32Compiler methodsFor: 'encoding' stamp: 'eem 9/18/2015 11:49'!
mod: mod RM: regMode RO: regOpcode
	"See ModR/M byte & opcode syntax
	 In addition to the notation shown above in 'Mnemonic Syntax' on page 43,
	 the following notation indicates the size and type of operands in the syntax of an instruction opcode:
		/digit	Indicates that the ModRM byte specifies only one register or memory (r/m) operand.
				The digit is specified by the ModRM reg field and is used as an instruction-opcode extension.
				Valid digit values range from 0 to 7.
		/r		Indicates that the ModRM byte specifies both a register operand and a reg/mem (register or memory) operand."
	^mod << 6 + (regOpcode << 3) + regMode! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/12/2015 11:44:33'!
nsSendCacheAt: callSiteReturnAddress
	"Answer the NSSendCache for the return address of a Newspeak
	 self, super, outer, or implicit receiver send."
	<option: #NewspeakVM>
	<var: #callSiteReturnAddress type: #'char *'>
	<inline: false>
	^self literalBeforeFollowingAddress: callSiteReturnAddress asUnsignedInteger - 5 "sizeofcall"! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'eem 10/18/2015 19:23:06'!
numIntRegArgs
	^0! !

!CogIA32Compiler methodsFor: 'multi-threading' stamp: 'eem 10/30/2020 16:24'!
numLowLevelLockOpcodes
	<inline: #always>
	"ceTryLockVMOwner:
		movl 4(%esp), %ecx
		xorl %eax, %eax
		lock cmpxchgl %ecx, %ds:&vmOwnerLock			N.B. lock cmpxchgq are two separate opcodes
		setz %al 
		ret"
	^6! !

!CogIA32Compiler methodsFor: 'generate machine code' stamp: 'rmacnak 12/3/2015 19:00:20'!
padIfPossibleWithStopsFrom: startAddr to: endAddr
	self stopsFrom: startAddr to: endAddr! !

!CogIA32Compiler methodsFor: 'feature detection' stamp: 'eem 8/5/2021 08:26'!
preferredRegisterPairForMovePerfCnt64RRLInto: aBlock
	"Evaluate aBlock with the preferred registers to receive the
	 low and high 32-bits of the 64-bit performance counter."
	<inline: #always>
	aBlock value: EAX value: EDX! !

!CogIA32Compiler methodsFor: 'abi' stamp: 'rsf 1/21/2017 03:38'!
registerToSaveIP
	"Temporary register used for fetching the instruction pointer. This should
	not be used for passing parameters in a standard ABI"
	<option: #LowcodeVM>
	^ ClassReg! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 6/13/2014 13:27'!
relocateCallBeforeReturnPC: retpc by: delta
	| distance |
	delta ~= 0 ifTrue:
		[distance :=    ((objectMemory byteAt: retpc - 1) << 24)
					+  ((objectMemory byteAt: retpc - 2) << 16)
					+  ((objectMemory byteAt: retpc - 3) << 8)
					+   (objectMemory byteAt: retpc - 4).
		 distance := distance + delta.
		 objectMemory
			byteAt: retpc - 1 put: (distance >> 24 bitAnd: 16rFF);
			byteAt: retpc - 2 put: (distance >> 16 bitAnd: 16rFF);
			byteAt: retpc - 3 put: (distance >>   8 bitAnd: 16rFF);
			byteAt: retpc - 4 put: (distance            bitAnd: 16rFF)]! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/2/2015 13:31'!
relocateMethodReferenceBeforeAddress: pc by: delta
	<inline: true>
	self relocateCallBeforeReturnPC: pc by: delta! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 3/10/2017 17:21'!
rewriteCPICJumpAt: addressFollowingJump target: jumpTargetAddr
	"Rewrite the short jump instruction to jump to a new cpic case target. "
	<var: #addressFollowingJump type: #usqInt>
	<var: #jumpTargetAddr type: #usqInt>
	<var: #callDistance type: #sqInt> "prevent type inference for avoiding warning on abs"
	| callDistance |
	callDistance := jumpTargetAddr - addressFollowingJump.
	self assert: callDistance abs < 128.
	objectMemory
		byteAt: addressFollowingJump - 1
		put:  (callDistance bitAnd: 16rFF).
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: addressFollowingJump - 10 to: addressFollowingJump - 1]."! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 10/7/2020 18:46'!
rewriteCallAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a call instruction to call a different target.  This variant is used to link PICs
	 in ceSendMiss et al, and to rewrite cached primitive calls.   Answer the extent of
	 the code change which is used to compute the range of the icache to flush."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	| callDistance |
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: callSiteReturnAddress - 10 to: callSiteReturnAddress - 1]."
	(cogit initialized and: [callTargetAddress < cogit minCallAddress]) ifTrue:
		[self error: 'linking callsite to invalid address'].
	callDistance := (callTargetAddress - callSiteReturnAddress) signedIntToLong.
	objectMemory
		byteAt: callSiteReturnAddress - 1 put: (callDistance >> 24 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 2 put: (callDistance >> 16 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 3 put: (callDistance >>   8 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 4 put: (callDistance            bitAnd: 16rFF).
	self assert: (self callTargetFromReturnAddress: callSiteReturnAddress) = callTargetAddress.
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: callSiteReturnAddress - 10 to: callSiteReturnAddress - 1]."
	^5! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 10/7/2020 18:45'!
rewriteInlineCacheAt: callSiteReturnAddress tag: cacheTag target: callTargetAddress
	"Rewrite an inline cache to call a different target for a new tag.  This variant is used
	 to link unlinked sends in ceSend:to:numArgs: et al.  Answer the extent of the code
	 change which is used to compute the range of the icache to flush."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	| callDistance |
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: callSiteReturnAddress - 10 to: callSiteReturnAddress - 1]."
	(cogit initialized and: [callTargetAddress < cogit minCallAddress]) ifTrue:
		[self error: 'linking callsite to invalid address'].
	callDistance := (callTargetAddress - callSiteReturnAddress) signedIntToLong.
	objectMemory
		byteAt: callSiteReturnAddress - 1 put: (callDistance >> 24 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 2 put: (callDistance >> 16 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 3 put: (callDistance >>   8 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 4 put: (callDistance            bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 6 put: (cacheTag >> 24 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 7 put: (cacheTag >> 16 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 8 put: (cacheTag >>   8 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 9 put: (cacheTag            bitAnd: 16rFF).
	self assert: (self callTargetFromReturnAddress: callSiteReturnAddress) = callTargetAddress.
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: callSiteReturnAddress - 10 to: callSiteReturnAddress - 1]."
	^10! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 10/13/2009 12:44'!
rewriteInlineCacheTag: cacheTag at: callSiteReturnAddress
	"Rewrite an inline cache with a new tag.  This variant is used
	 by the garbage collector."
	objectMemory
		byteAt: callSiteReturnAddress - 6 put: (cacheTag >> 24 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 7 put: (cacheTag >> 16 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 8 put: (cacheTag >>   8 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 9 put: (cacheTag            bitAnd: 16rFF)! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 6/21/2009 14:51'!
rewriteJumpLongAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a long jump instruction to jump to a different target.  This variant
	 is used to rewrite cached primitive calls.   Answer the extent of the
	 code change which is used to compute the range of the icache to flush."
	^self rewriteCallAt: callSiteReturnAddress target: callTargetAddress! !

!CogIA32Compiler methodsFor: 'encoding' stamp: 'eem 12/23/2008 18:20'!
s: scale i: indexReg b: baseReg 
	^scale << 6 + (indexReg << 3) + baseReg! !

!CogIA32Compiler methodsFor: 'abi' stamp: 'eem 3/19/2015 15:04'!
saveAndRestoreLinkRegAround: aBlock
	"If the processor's ABI includes a link register, generate instructions
	 to save and restore it around aBlock, which is assumed to generate code."
	<inline: true>
	^aBlock value! !

!CogIA32Compiler methodsFor: 'feature detection' stamp: 'eem 10/30/2020 05:42'!
setCpuidWord0: word
	<cmacro: '(ign,word) cpuidWord0 = word'>
	"For want of somewhere to put the variable..."
	self class instVarNamed: 'cpuidWord0' put: word! !

!CogIA32Compiler methodsFor: 'feature detection' stamp: 'eem 10/30/2020 05:41'!
setCpuidWord1: word
	<cmacro: '(ign,word) cpuidWord1 = word'>
	"For want of somewhere to put the variable..."
	self class instVarNamed: 'cpuidWord1' put: word! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'nice 2/18/2020 01:47'!
setsConditionCodesFor: aConditionalJumpOpcode
	<inline: false> "to save Slang from having to be a real compiler (it can't inline switches that return)"
	"Answer if the receiver's opcode sets the condition codes correctly for the given conditional jump opcode."
	^opcode caseOf:
		{	[ArithmeticShiftRightCqR]	->	[self shiftSetsConditionCodesFor: aConditionalJumpOpcode].
			[ArithmeticShiftRightRR]	->	[self shiftSetsConditionCodesFor: aConditionalJumpOpcode].
			[LogicalShiftLeftCqR]		->	[self shiftSetsConditionCodesFor: aConditionalJumpOpcode].
			[LogicalShiftLeftRR]			->	[self shiftSetsConditionCodesFor: aConditionalJumpOpcode].
			[XorRR]						->	[true].
			[ClzRR]						->	[aConditionalJumpOpcode = JumpZero or: [aConditionalJumpOpcode = JumpNonZero or: [aConditionalJumpOpcode = JumpCarry or: [aConditionalJumpOpcode = JumpNoCarry "carry flag is set if input is zero"]]]]
		}
		otherwise: [self halt: 'unhandled opcode in setsConditionCodesFor:'. false]! !

!CogIA32Compiler methodsFor: 'testing' stamp: 'eem 9/30/2015 15:58'!
shiftSetsConditionCodesFor: aConditionalJumpOpcode
	"OF flag only guaranteed to be set for 1-bit shifts.  See [1] p 490.
	 Only SF, ZF & PF set according to result.  Since the question is currently
	 asked only for Zero and Negative use the following simplification."
	<inline: true>
	^(aConditionalJumpOpcode between: JumpZero and: JumpNonNegative)! !

!CogIA32Compiler methodsFor: 'disassembly' stamp: 'eem 10/13/2009 12:44'!
sizeHasModrm: op at: pc
	| modrm mod ro rm |
	modrm := objectMemory byteAt: pc + 1.
	mod := modrm >> 6.
	ro := modrm >> 3 bitAnd: 7.
	rm := modrm bitAnd: 7.
	mod = 3 ifTrue:
		[^2].
	rm ~= 4 ifTrue: "no SIB byte"
		[^mod caseOf:
		   {	[0]	->	[rm = 5
						ifTrue: [6] "reg or 32-bit displacement"
						ifFalse: [3]].
			[1]	->	[3]. "8-bit displacement"
			[2]	->	[6] }].
	self halt: 'fall through in sizeHasModrm:at:'.
	^0! !

!CogIA32Compiler methodsFor: 'disassembly' stamp: 'eem 10/13/2009 12:44'!
sizeImmediateGroup1: op at: pc
	"see [1] p A-7, p A-13"
	| modrm mod ro rm |
	modrm := objectMemory byteAt: pc + 1.
	mod := modrm >> 6.
	ro := modrm >> 3 bitAnd: 7.
	rm := modrm bitAnd: 7.
	^ro caseOf:
	   {	[7 "cmp"]	->	[op = 16r81
							ifTrue: [6]
							ifFalse: [3]] }! !

!CogIA32Compiler methodsFor: 'generate machine code' stamp: 'eem 09/28/2015 18:48'!
sizePCDependentInstructionAt: eventualAbsoluteAddress
	"Size a jump and set its address.  The target may be another instruction
	 or an absolute address.  On entry the address inst var holds our virtual
	 address. On exit address is set to eventualAbsoluteAddress, which is
	 where this instruction will be output.  The span of a jump to a following
	 instruction is therefore between that instruction's address and this
	 instruction's address ((which are both still their virtual addresses), but the
	 span of a jump to a preceding instruction or to an absolute address is
	 between that instruction's address (which by now is its eventual absolute
	 address) or absolute address and eventualAbsoluteAddress."

	| target maximumSpan abstractInstruction |
	<var: #abstractInstruction type: #'AbstractInstruction *'>
	opcode = AlignmentNops ifTrue:
		[| alignment |
		 address := eventualAbsoluteAddress.
		 alignment := operands at: 0.
		 ^machineCodeSize := (eventualAbsoluteAddress + (alignment - 1) bitAnd: alignment negated)
							   - eventualAbsoluteAddress].
	self assert: self isJump.
	target := operands at: 0.
	abstractInstruction := cogit cCoerceSimple: target to: #'AbstractInstruction *'.
	(self isAnInstruction: abstractInstruction)
		ifTrue:
			[maximumSpan := abstractInstruction address
							- (((cogit abstractInstruction: self follows: abstractInstruction)
								ifTrue: [eventualAbsoluteAddress]
								ifFalse: [address]) + 2)]
		ifFalse:
			[maximumSpan := target - (eventualAbsoluteAddress + 2)].
	address := eventualAbsoluteAddress.
	^machineCodeSize := opcode >= FirstShortJump
							ifTrue:
								[(self isQuick: maximumSpan)
									ifTrue: [2]
									ifFalse: [opcode = Jump
												ifTrue: [5]
												ifFalse: [6]]]
							ifFalse:
								[(opcode = JumpLong or: [opcode = JumpFull])
									ifTrue: [5]
									ifFalse: [6]]! !

!CogIA32Compiler methodsFor: 'abi' stamp: 'eem 1/28/2009 18:40'!
stackBytesForNumArgs: numArgs
	^numArgs * 4! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'eem 2/14/2009 16:05'!
stackPageInterruptHeadroomBytes
	"Return a minimum amount of headroom for each stack page (in bytes).  In a
	 JIT the stack has to have room for interrupt handlers which will run on the stack.
	 See [1] pp 6-13 to 6-14.  On an interrupt or exception the maximum state that
	 an IA32 CPU will push is SS, ESP, EFLAGS, CS, EIP and an error code.  It may then
	 call an interrupt procedure.  Leave some room above the minimum state."
	^256! !

!CogIA32Compiler methodsFor: 'encoding' stamp: 'rmacnak 12/3/2015 19:07:03'!
stop
	"int3"
	<inline: true>
	^16rCC! !

!CogIA32Compiler methodsFor: 'generate machine code - support' stamp: 'eem 1/27/2020 08:44'!
stopsFrom: startAddr to: endAddr
	self
		cCode: [self memset: startAddr _: self stop _: endAddr - startAddr + 1]
		inSmalltalk:
			[| alignedEnd alignedStart stops |
			stops := self stop << 8 + self stop.
			stops := stops << 16 + stops.
			alignedStart := startAddr + 3 // 4 * 4.
			alignedEnd := endAddr - 1 // 4 * 4.
			alignedEnd <= startAddr
				ifTrue:
					[startAddr to: endAddr do:
						[:addr | cogit codeByteAt: addr put: self stop]]
				ifFalse:
					[startAddr to: alignedStart - 1 do:
						[:addr | cogit codeByteAt: addr put: self stop].
					 alignedStart to: alignedEnd by: 4 do:
						[:addr | cogit codeLong32At: addr put: stops].
					 alignedEnd + 4 to: endAddr do:
						[:addr | cogit codeByteAt: addr put: self stop]]]! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 09/28/2015 18:48'!
storeLiteral: literal beforeFollowingAddress: followingAddress
	"Rewrite the literal in the instruction immediately preceding followingAddress."
	objectMemory
		byteAt: followingAddress - 1 put: (literal >> 24 bitAnd: 16rFF);
		byteAt: followingAddress - 2 put: (literal >> 16 bitAnd: 16rFF);
		byteAt: followingAddress - 3 put: (literal >>   8 bitAnd: 16rFF);
		byteAt: followingAddress - 4 put: (literal            bitAnd: 16rFF)! !

!CogIA32Compiler methodsFor: 'disassembly' stamp: 'eem 10/13/2009 12:44'!
twoByteInstructionSizeAt: pc
	| op |
	op := objectMemory byteAt: pc + 1. 
	^(op bitAnd: 16rF0) caseOf:
		{	[16r80]	->	[6 "long conditional jumps"] }! !

!CogIA32Compiler methodsFor: 'memory access' stamp: 'eem 12/12/2015 15:25'!
unalignedLongAt: byteAddress
	<cmacro: '(byteAddress) longAt(byteAddress)'>
	^objectMemory unalignedLongAt: byteAddress! !

!CogIA32Compiler methodsFor: 'memory access' stamp: 'eem 12/12/2015 15:26'!
unalignedLongAt: byteAddress put: aWord
	<cmacro: '(byteAddress,aWord) longAtput(byteAddress,aWord)'>
	^objectMemory unalignedLongAt: byteAddress put: aWord! !

!CogIA32Compiler methodsFor: 'accessing' stamp: 'eem 10/13/2009 12:44'!
unsignedShortAt: byteAddress
	^(objectMemory byteAt: byteAddress) + ((objectMemory byteAt: byteAddress + 1) bitShift: 8)! !

!CogIA32Compiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/9/2015 23:06:51'!
zoneCallsAreRelative
	"Answer if Call and JumpLong are relative and hence need to take the caller's
	 relocation delta into account during code compaction, rather than just the
	 callee's delta."
	^true! !

CogIA32Compiler subclass: #CogIA32CompilerForTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!CogIA32CompilerForTests class methodsFor: 'test support' stamp: 'eem 10/5/2015 17:34'!
byteRegistersWithNamesDo: aBinaryBlock
	(self registers first: 4)
		with: #('%al' '%cl' '%dl' '%bl')
		do: aBinaryBlock! !

!CogIA32CompilerForTests class methodsFor: 'test support' stamp: 'eem 4/16/2009 11:16'!
dataRegistersWithAccessorsDo: aTrinaryBlock
	self registers withIndexDo:
		[:reg :i|
		aTrinaryBlock
			value: reg
			value: (#(eax ecx edx ebx esp ebp esi edi) at: i)
			value: (#(eax: ecx: edx: ebx: esp: ebp: esi: edi:) at: i)]! !

!CogIA32CompilerForTests class methodsFor: 'test support' stamp: 'eem 9/10/2009 08:04'!
dataRegistersWithAccessorsExcept: accessorsToExclude do: aTrinaryBlock
	self registers withIndexDo:
		[:reg :i| | getter setter |
		getter := #(eax ecx edx ebx esp ebp esi edi) at: i.
		setter := #(eax: ecx: edx: ebx: esp: ebp: esi: edi:) at: i.
		(accessorsToExclude includes: getter) ifFalse:
			[aTrinaryBlock value: reg value: getter value: setter]]! !

!CogIA32CompilerForTests class methodsFor: 'accessing' stamp: 'eem 12/18/2019 22:58'!
dataRegistersWithAccessorsGiven: aProcessorSimulator do: aTrinaryBlock
	(0 to: 7) withIndexDo:
		[:reg :i|
		aTrinaryBlock
			value: reg
			value: (aProcessorSimulator registerStateGetters at: i)
			value: (aProcessorSimulator registerStateSetters at: i)]! !

!CogIA32CompilerForTests class methodsFor: 'test support' stamp: 'eem 9/18/2010 17:04'!
fp64RegistersWithAccessorsDo: aTrinaryBlock
	self fp64registers withIndexDo:
		[:reg :i|
		aTrinaryBlock
			value: reg
			value: (#(xmm0low xmm1low xmm2low xmm3low xmm4low xmm5low xmm6low xmm7low) at: i)
			value: (#(xmm0low: xmm1low: xmm2low: xmm3low: xmm4low: xmm5low: xmm6low: xmm7low:) at: i)]! !

!CogIA32CompilerForTests class methodsFor: 'test support' stamp: 'eem 9/18/2010 17:02'!
fp64registers
	^{ DPFPReg0. DPFPReg1. DPFPReg2. DPFPReg3. DPFPReg4. DPFPReg5. DPFPReg6. DPFPReg7 } "a.k.a. (-9 to: -16 by: -1)"! !

!CogIA32CompilerForTests class methodsFor: 'test support' stamp: 'eem 12/1/2015 15:38'!
registers
	^(0 to: 7) "a.k.a. { EAX. ECX. EDX. EBX. ESP. EBP. ESI. EDI }"! !

!CogIA32CompilerForTests class methodsFor: 'test support' stamp: 'eem 4/16/2009 11:16'!
registersWithNamesDo: aBinaryBlock
	self registers
		with: #('%eax' '%ecx' '%edx' '%ebx' '%esp' '%ebp' '%esi' '%edi')
		do: aBinaryBlock! !

!CogIA32CompilerForTests class methodsFor: 'test support' stamp: 'eem 4/16/2009 11:03'!
xmmRegistersWithNamesDo: aBinaryBlock
	{XMM0L. XMM1L. XMM2L. XMM3L. XMM4L. XMM5L. XMM6L. XMM7L}
		with: #('%xmm0' '%xmm1' '%xmm2' '%xmm3' '%xmm4' '%xmm5' '%xmm6' '%xmm7')
		do: aBinaryBlock! !

!CogIA32CompilerForTests methodsFor: 'generate machine code' stamp: 'nice 12/30/2020 00:05'!
concretizeAt: actualAddress
	"Override to check maxSize and machineCodeSize"

	| maxAddress nextAddress |
	maxSize ifNil: [maxSize := self computeMaximumSize].
	maxAddress := actualAddress + maxSize.
	nextAddress := super concretizeAt: actualAddress.
	self assert: (maxSize notNil
				and: [self isPCDependent
						ifTrue: [maxAddress >= nextAddress]
						ifFalse: [maxAddress = nextAddress]]).
	^nextAddress! !

!CogIA32CompilerForTests methodsFor: 'testing' stamp: 'eem 4/10/2015 18:32'!
hasSSE2Instructions
	"Answer if we support SSE2"
	^true! !

CogIA32Compiler subclass: #CogICacheFlushingIA32Compiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogICacheFlushingIA32Compiler commentStamp: '<historical>' prior: 0!
This is a subclass that uses CPUID to serialize instruction modification.  This is not needed with Intel processors on systems where code and data have the same linear address, hence Cog uses the superclass by default.!

!CogICacheFlushingIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 2/3/2020 19:32'!
flushICacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	"Flush the instruction cache from (startAddress to endAddress].
	 If there is a dual mapped code zone (the normal zone but marked with read/execute, and a
	 read/write zone codeToDataDelta bytes away) then also flush the data cache for the corresp-
	 onding range in the read/write zone and invalidate the data cache for the read/execute zone."
	<cmacro: '(me,startAddress,endAddress) ceFlushICache(startAddress,endAddress)'>
	^cogit simulateCeFlushICacheFrom: startAddress to: endAddress! !

!CogICacheFlushingIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 6/17/2009 17:45'!
generateICacheFlush
	"Use CPUID as a serializing instruction for instruction modification.
	 MFENCE doesn't work which is a shame because CPUID updates registers."
	cogit
		PushR: EDX;
		PushR: ECX;
		PushR: EBX;
		XorR: EAX R: EAX;
		gen: CPUID;
		PopR: EBX;
		PopR: ECX;
		PopR: EDX;
		RetN: 8 "pop from,to args"

	"self hasSSE2Instructions
		ifTrue:
			[cogit
				gen: MFENCE]
		ifFalse:
			[cogit
				PushR: EDX;
				PushR: ECX;
				PushR: EBX;
				XorR: EAX R: EAX;
				gen: CPUID;
				PopR: EBX;
				PopR: ECX;
				PopR: EDX].
	cogit RetN: 8 ``pop from,to args''"! !

!CogICacheFlushingIA32Compiler methodsFor: 'inline cacheing' stamp: 'eem 2/12/2009 22:27'!
numICacheFlushOpcodes
	^10! !

CogAbstractInstruction subclass: #CogMIPSELCompiler
	instanceVariableNames: ''
	classVariableNames: 'AddCheckOverflowCqR AddCheckOverflowRR BrEqualRR BrLongEqualRR BrLongNotEqualRR BrNotEqualRR BrSignedGreaterEqualRR BrSignedGreaterRR BrSignedLessEqualRR BrSignedLessRR BrUnsignedGreaterEqualRR BrUnsignedGreaterRR BrUnsignedLessEqualRR BrUnsignedLessRR BranchTemp Cmp ConcreteVarBaseReg DivRR MoveHighR MoveLowR MulCheckOverflowRR MulRR Overflow OverflowTemp1 OverflowTemp2 SubCheckOverflowCqR SubCheckOverflowRR TargetReg'
	poolDictionaries: 'MIPSConstants'
	category: 'VMMaker-JIT'!

!CogMIPSELCompiler commentStamp: 'rmacnak 6/26/2017 20:52' prior: 0!
Implemenation for 32-bit, little-endian MIPS running with the OABI (Debian port name 'mipsel').

Status: incomplete, no work planned

In December 2015, this implementation was complete enough to run the Newspeak test suite under the simulator. The compiled VM, however, failed at startup. The author suspects that some variables are lacking correct type annotations, being translated by Slang as unsigned when they should be signed or vice versa, causing some shift or comparison to have the wrong signness in C, but it may well be some other discrepancy between the behavior of Slang code when run in Smalltalk and when translated to C.

This implementation also does not provide instructions for working with floating point numbers, instead falling back to the interpreter's implementation for all floating point operations.
!

!CogMIPSELCompiler class methodsFor: 'class initialization' stamp: 'rmacnak 11/8/2015 17:11:17'!
ISA
	^#MIPSEL! !

!CogMIPSELCompiler class methodsFor: 'translation' stamp: 'rmacnak 12/17/2015 19:54'!
declareCVarsIn: aCCodeGenerator
	super declareCVarsIn: aCCodeGenerator.
	aCCodeGenerator addHeaderFile: '<asm/cachectl.h>'. "For cacheflush"! !

!CogMIPSELCompiler class methodsFor: 'translation' stamp: 'rmacnak 12/14/2015 19:43'!
identifyingPredefinedMacros
	^#('__MIPSEL__')! !

!CogMIPSELCompiler class methodsFor: 'translation' stamp: 'eem 8/5/2021 20:34'!
ifTranslateableAddWithOptionsTo: aCollection
	"do nothing for now; it is too difficult to continue to get this to generate wthout someone to support it.
	 we can disinter it later..."! !

!CogMIPSELCompiler class methodsFor: 'class initialization' stamp: 'rmacnak 11/30/2015 19:26:36'!
initialize
	"CogRTLOpcodes initialize. CogMIPSELCompiler initialize"

	super initialize.
	
	ConcreteVarBaseReg := S6.

	Cmp := T0.

	Overflow := T0.
	OverflowTemp1 := T1.
	OverflowTemp2 := T2.
	
	"Can't use AT, Cmp or Overflow because we may need to preserve them for sequences like
	 CmpCwR
	 JumpZero
	 JumpBelow"
	BranchTemp := T3.

	"OABI position independent code expects T9 to have its entry point on entry?"
	self flag: #OABI.
	TargetReg := T9.

	"Specific instructions"
	self initializeSpecificOpcodes: #(
		MulRR
		DivRR
		MoveLowR
		MoveHighR
		AddCheckOverflowCqR 
		AddCheckOverflowRR
		MulCheckOverflowRR
		SubCheckOverflowCqR
		SubCheckOverflowRR
		BrEqualRR
		BrNotEqualRR 
		BrUnsignedLessRR
		BrUnsignedLessEqualRR
		BrUnsignedGreaterRR
		BrUnsignedGreaterEqualRR
		BrSignedLessRR
		BrSignedLessEqualRR
		BrSignedGreaterRR
		BrSignedGreaterEqualRR
		BrLongEqualRR
		BrLongNotEqualRR) 
			in: thisContext method
		
	! !

!CogMIPSELCompiler class methodsFor: 'class initialization' stamp: 'nice 2/9/2020 17:40'!
initializeAbstractRegisters
	"Assign the abstract registers with the identities/indices of the relevant concrete registers."

	"See MIPSConstants>>initializeRegisters for a description of the C ABI."

	"Note we can fit all of the abstract registers in C preserved registers, and
	 not need to save or restore them at runtime calls."

	ReceiverResultReg		:= S0.
	Arg0Reg				:= S1.
	Arg1Reg				:= S2.
	ClassReg				:= S3.
	SendNumArgsReg		:= S4.
	TempReg				:= S5.
	VarBaseReg				:= S6. "Must be callee saved"
	SPReg					:= SP.
	FPReg					:= FP.
	RISCTempReg			:= AT.
	LinkReg				:= RA.

	self flag: #OABI. "see e.g. http://refspecs.linuxbase.org/elf/mipsabi.pdf"
	ABICalleeSavedRegisterMask := self
									registerMaskFor: S0 and: S1 and: S2 and: S3
									and: S4 and: S5 and: S6 and: S7.
	ABICallerSavedRegisterMask := self
									registerMaskFor: T0 and: T1 and: T2 and: T3
									and: T4 and: T5 and: T6 and: T7 and: T8 and: T9.
	CallerSavedRegisterMask := ABICallerSavedRegisterMask
								bitAnd: (self registerMaskFor: ClassReg and: ReceiverResultReg and: SendNumArgsReg and: Arg0Reg and: Arg1Reg).
	ABIResultReg := V0.
	ABIResultRegHigh := V1.

	NumRegisters := 32.

	self flag: #todo.
	"Extra0Reg			:= ??.
	Extra1Reg			:= ??.
	Extra2Reg			:= ??.
	Extra3Reg			:= ??.
	Extra4Reg			:= ??.
	Extra5Reg			:= ??.
	Extra6Reg			:= ??.
	Extra7Reg			:= ??."

	self flag: #todo.
	"DPFPReg0				:= ??.
	DPFPReg1				:= ??.
	DPFPReg2				:= ??.
	DPFPReg3				:= ??.
	DPFPReg4				:= ??.
	DPFPReg5				:= ??.
	DPFPReg6				:= ??.
	DPFPReg7				:= ??.
	DPFPReg8				:= ??.
	DPFPReg9				:= ??.
	DPFPReg10				:= ??.
	DPFPReg11				:= ??.
	DPFPReg12				:= ??.
	DPFPReg13				:= ??.
	DPFPReg14				:= ??.
	DPFPReg15				:= ??"! !

!CogMIPSELCompiler class methodsFor: 'translation' stamp: 'eem 3/11/2017 12:45'!
machineCodeDeclaration
	"Answer the declaration for the machineCode array.
	 MPIS instructions are 32-bits in length."
	^{#'unsigned int'. '[', self basicNew machineCodeWords printString, ']'}! !

!CogMIPSELCompiler class methodsFor: 'debug printing' stamp: 'eem 11/25/2019 15:29'!
printFormatForOpcodeName: opcodeName
	"Answer a sequence of $r, $f or nil for the operands in the opcode, used for printing, where
	 r => integer register, f => floating point register, and nil => numeric or address operand.
	 Subclasses can override to provide a format string for their own private opcodes."
	^(opcodeName beginsWith: 'Br') ifTrue: [' rr'] ifFalse: [#()]! !

!CogMIPSELCompiler class methodsFor: 'translation' stamp: 'eem 12/11/2015 23:29'!
wordSize
	"This is a 32-bit ISA"
	^4! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/8/2015 15:34:28'!
addiuR: destReg R: srcReg C: imm
	^self itype: ADDIU rs: srcReg rt: destReg signedImmediate: imm! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/5/2015 20:22:50'!
adduR: destReg R: leftReg R: rightReg
	^self rtype: SPECIAL rs: leftReg rt: rightReg rd: destReg sa: 0 funct: ADDU! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/5/2015 21:11:12'!
andR: destReg R: leftReg R: rightReg
	^self rtype: SPECIAL rs: leftReg rt: rightReg rd: destReg sa: 0 funct: AND! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/8/2015 15:39:06'!
andiR: destReg R: srcReg C: imm
	^self itype: ANDI rs: srcReg rt: destReg eitherImmediate: imm! !

!CogMIPSELCompiler methodsFor: 'encoding - control' stamp: 'rmacnak 12/17/2015 20:54'!
beqR: leftReg R: rightReg offset: offset
	self assert: (offset bitAnd: 3) = 0.
	self assert: (offset between: -16r20000 and: 16r1FFFF).
	^self itype: BEQ rs: leftReg rt: rightReg signedImmediate: offset >>> 2! !

!CogMIPSELCompiler methodsFor: 'encoding - control' stamp: 'rmacnak 12/17/2015 20:54'!
bgezR: cmpReg offset: offset
	self assert: (offset bitAnd: 3) = 0.
	self assert: (offset between: -16r20000 and: 16r1FFFF).
	^self itype: REGIMM rs: cmpReg rt: BGEZ signedImmediate: offset >>> 2! !

!CogMIPSELCompiler methodsFor: 'encoding - control' stamp: 'rmacnak 12/17/2015 20:54'!
bgtzR: cmpReg offset: offset
	self assert: (offset bitAnd: 3) = 0.
	self assert: (offset between: -16r20000 and: 16r1FFFF).
	^self itype: BGTZ rs: cmpReg rt: 0 signedImmediate: offset >>> 2! !

!CogMIPSELCompiler methodsFor: 'encoding - control' stamp: 'rmacnak 12/17/2015 20:54'!
blezR: cmpReg offset: offset
	self assert: (offset bitAnd: 3) = 0.
	self assert: (offset between: -16r20000 and: 16r1FFFF).
	^self itype: BLEZ rs: cmpReg rt: 0 signedImmediate: offset >>> 2! !

!CogMIPSELCompiler methodsFor: 'encoding - control' stamp: 'rmacnak 12/17/2015 20:55'!
bltzR: cmpReg offset: offset
	self assert: (offset bitAnd: 3) = 0.
	self assert: (offset between: -16r20000 and: 16r1FFFF).
	^self itype: REGIMM rs: cmpReg rt: BLTZ signedImmediate: offset >>> 2! !

!CogMIPSELCompiler methodsFor: 'encoding - control' stamp: 'rmacnak 12/17/2015 20:55'!
bneR: leftReg R: rightReg offset: offset
	self assert: (offset bitAnd: 3) = 0.
	self assert: (offset between: -16r20000 and: 16r1FFFF).
	^self itype: BNE rs: leftReg rt: rightReg signedImmediate: offset >>> 2! !

!CogMIPSELCompiler methodsFor: 'accessing' stamp: 'rmacnak 11/10/2015 22:35:34'!
callInstructionByteSize
	self flag: #todo. "Which call opcode does this want the size of?"
	^16! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 11/14/2015 15:09:27'!
callTargetFromReturnAddress: callSiteReturnAddress
	"csra - 16:	lui t9, high
	 csra - 12:	ori t9, low
	 csra - 8:	jalr t9
	 csra - 4:	nop (delay slot)"
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 16) == LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 12) == ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 8) == SPECIAL.
	self assert: (self functionAtAddress: callSiteReturnAddress - 8) == JALR.
	self assert: (objectMemory longAt: callSiteReturnAddress - 4) == self nop. "Delay slot"
	^self literalAtAddress: callSiteReturnAddress - 12! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
canDivQuoRem
	^true! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'eem 12/8/2019 15:51'!
canMulRR
	^false! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'nice 8/26/2019 12:01'!
clzR: destReg R: leftReg R: rightReg
	"NOTE: the right reg (rt) MUST equal dest reg (rd) or behavior is undefined"
	^self rtype: SPECIAL rs: leftReg rt: rightReg rd: destReg sa: 0 funct: 2r100000 "CLZ"! !

!CogMIPSELCompiler methodsFor: 'accessing' stamp: 'rmacnak 11/22/2015 21:31:31'!
cmpC32RTempByteSize
	^8! !

!CogMIPSELCompiler methodsFor: 'accessing' stamp: 'rmacnak 11/9/2015 21:25:48'!
codeGranularity
	"Answer the size in bytes of a unit of machine code."
	<inline: true>
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code' stamp: 'eem 9/15/2020 13:38'!
computeMaximumSize
	"Each MIPS instruction has 4 bytes. Many abstract opcodes need more than one
	 instruction. Instructions that refer to constants and/or literals depend on literals
	 being stored in-line or out-of-line.

	 N.B.  The ^N forms are to get around the bytecode compiler's long branch
	 limits which are exceeded when each case jumps around the otherwise."

	opcode
		caseOf: {
		[BrEqualRR]					-> [^8].
		[BrNotEqualRR]					-> [^8].
		[BrUnsignedLessRR]			-> [^12].
		[BrUnsignedLessEqualRR]		-> [^12].
		[BrUnsignedGreaterRR]			-> [^12].
		[BrUnsignedGreaterEqualRR]	-> [^12].
		[BrSignedLessRR]				-> [^12].
		[BrSignedLessEqualRR]			-> [^12].
		[BrSignedGreaterRR]			-> [^12].
		[BrSignedGreaterEqualRR]		-> [^12].
		[BrLongEqualRR]				-> [^16].
		[BrLongNotEqualRR]			-> [^16].
		[MulRR]					-> [^4].
		[DivRR]					-> [^4].
		[MoveLowR]			-> [^4].
		[MoveHighR]			-> [^4].

		"Noops & Pseudo Ops"
		[Label]					-> [^0].
		[Literal]					-> [^4].
		[AlignmentNops]		-> [^(operands at: 0) - 4].
		[Fill32]					-> [^4].
		[Nop]					-> [^4].
		"Control"
		[Call]					-> [^self literalLoadInstructionBytes + 8].
		[CallFull]				-> [^self literalLoadInstructionBytes + 8].
		[JumpR]				-> [^8].
		[Jump]					-> [^8].
		[JumpFull]				-> [^self literalLoadInstructionBytes + 8].
		[JumpLong]			-> [^self literalLoadInstructionBytes + 8].
		[JumpZero]				-> [^8].
		[JumpNonZero]			-> [^8].
		[JumpNegative]			-> [^8].
		[JumpNonNegative]		-> [^8].
		[JumpOverflow]			-> [^8].
		[JumpNoOverflow]		-> [^8].
		[JumpCarry]			-> [^8].
		[JumpNoCarry]			-> [^8].
		[JumpLess]				-> [^8].
		[JumpGreaterOrEqual]	-> [^8].
		[JumpGreater]			-> [^8].
		[JumpLessOrEqual]		-> [^8].
		[JumpBelow]			-> [^8].
		[JumpAboveOrEqual]	-> [^8].
		[JumpAbove]			-> [^8].
		[JumpBelowOrEqual]	-> [^8].
		[JumpLongZero]		-> [^self literalLoadInstructionBytes + 8].
		[JumpLongNonZero]	-> [^self literalLoadInstructionBytes + 8].
		[JumpFPEqual]			-> [^8].
		[JumpFPNotEqual]		-> [^8].
		[JumpFPLess]			-> [^8].
		[JumpFPGreaterOrEqual]-> [^8].
		[JumpFPGreater]		-> [^8].
		[JumpFPLessOrEqual]	-> [^8].
		[JumpFPOrdered]		-> [^8].
		[JumpFPUnordered]	-> [^8].
		[RetN]					-> [^8].
		[Stop]					-> [^4].

		"Arithmetic"
		[AddCqR]				-> [^(self isShortOffset: (operands at: 0)) ifTrue: [4] ifFalse: [12]].
		[AndCqR]				-> [^((operands at: 0) between: 0 and: 16rFFFF) ifTrue: [4] ifFalse: [12]].
		[AndCqRR]				-> [^((operands at: 0) between: 0 and: 16rFFFF) ifTrue: [4] ifFalse: [12]].
		[OrCqR]				-> [^(self isShortOffset: (operands at: 0)) ifTrue: [4] ifFalse: [12]].
		[OrCqRR]				-> [^(self isShortOffset: (operands at: 0)) ifTrue: [4] ifFalse: [12]].
		[CmpCqR]				-> [^28].
		[SubCqR]				-> [^(self isShortOffset: (operands at: 0)) ifTrue: [4] ifFalse: [12]].
		[TstCqR]				-> [^(self isShortOffset: (operands at: 0)) ifTrue: [4] ifFalse: [12]].
		[XorCqR]				-> [^12].
		[AddCwR]				-> [^12].
		[AndCwR]				-> [^12].
		[CmpCwR]				-> [^28].
		[OrCwR]				-> [^12].
		[SubCwR]				-> [^12].
		[XorCwR]				-> [^12].
		[AddRR]				-> [^4].
		[AndRR]				-> [^4].
		[CmpRR]				-> [^20].
		[OrRR]					-> [^4].
		[XorRR]					-> [^4].
		[SubRR]				-> [^4].
		[NegateR]				-> [^4].
		[AddRRR]				-> [^4].
		[SubRRR]				-> [^4].
		[LoadEffectiveAddressMwrR] -> [^(self isShortOffset: (operands at: 0)) ifTrue: [4] ifFalse: [12]].
		[LogicalShiftLeftCqR]		-> [^4].
		[LogicalShiftRightCqR]		-> [^4].
		[ArithmeticShiftRightCqR]	-> [^4].
		[LogicalShiftLeftRR]			-> [^4].
		[LogicalShiftRightRR]		-> [^4].
		[ArithmeticShiftRightRR]		-> [^4].
		[AddRdRd]					-> [^4].
		[CmpRdRd]					-> [^4].
		[SubRdRd]					-> [^4].
		[MulRdRd]					-> [^4].
		[DivRdRd]					-> [^4].
		[SqrtRd]					-> [^4].
		[AddCheckOverflowCqR]	-> [^28].
		[AddCheckOverflowRR]		-> [^20].
		[SubCheckOverflowCqR]	-> [^28].
		[SubCheckOverflowRR]		-> [^20].
		[MulCheckOverflowRR]		-> [^20].
		[ClzRR]					-> [^4].
		"Data Movement"						
		[MoveCqR]				-> [^(self isShortOffset: (operands at: 0)) ifTrue: [4] ifFalse: [8]].
		[MoveCwR]				-> [^8].
		[MoveRR]				-> [^4].
		[MoveRdRd]				-> [^4].
		[MoveAwR]				-> [^(self isAddressRelativeToVarBase: (operands at: 0))
													ifTrue: [4]
													ifFalse: [self literalLoadInstructionBytes + 4]].
		[MoveRAw]				-> [^(self isAddressRelativeToVarBase: (operands at: 1))
													ifTrue: [4]
													ifFalse: [self literalLoadInstructionBytes + 4]].
		[MoveAbR]				-> [^(self isAddressRelativeToVarBase: (operands at: 0))
													ifTrue: [4]
													ifFalse: [self literalLoadInstructionBytes + 4]].
		[MoveRAb]				-> [^(self isAddressRelativeToVarBase: (operands at: 1))
													ifTrue: [4]
													ifFalse: [self literalLoadInstructionBytes + 4]].
		[MoveRMwr]			-> [^4].
		[MoveRdM64r]			-> [^self literalLoadInstructionBytes + 4]. 
		[MoveMbrR]				-> [^4].
		[MoveRMbr]				-> [^4].
		[MoveM16rR]			-> [^4].
		[MoveRM16r]			-> [^4].
		[MoveM64rRd]			-> [^self literalLoadInstructionBytes + 4].
		[MoveMwrR]			-> [^(self isShortOffset: (operands at: 0)) ifTrue: [4] ifFalse: [16]].
		[MoveXbrRR]			-> [^8].
		[MoveRXbrR]			-> [^8].
		[MoveXwrRR]			-> [^12].
		[MoveRXwrR]			-> [^12].
		[PopR]					-> [^8].
		[PushR]					-> [^8].
		[PushCw]				-> [^16].
		[PushCq]				-> [^16].
		[PrefetchAw] 			-> [^12].
		"Conversion"
		[ConvertRRd]			-> [^8].
		}.
	^0 "to keep C compiler quiet"
! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeAddCheckOverflowCqR
	| destReg leftReg rightImm |
	rightImm := operands at: 0.
	destReg := leftReg := operands at: 1.

	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: rightImm)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: rightImm)).

	"Save original LHS"
	self machineCodeAt: 8 put: (self adduR: OverflowTemp1 R: leftReg R: ZR). 
	
	"The actual addition"
	self machineCodeAt: 12 put: (self adduR: destReg R: leftReg R: AT). 

	"Set sign bit of OverflowTemp2 if sign of result differs from sign of RHS."
	self machineCodeAt: 16 put: (self xorR: OverflowTemp2 R: destReg R: AT).
	"Set sign bit of OverflowTemp1 if sign of result differs from sign of LHS."
	self machineCodeAt: 20 put: (self xorR: OverflowTemp1 R: destReg R: OverflowTemp1).
	"Set sign bit of Overflow if sign of result differs from both LHS and RHS, which indicates overflow."
	self machineCodeAt: 24 put: (self andR: Overflow R: OverflowTemp1 R: OverflowTemp2).
	^28! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeAddCheckOverflowRR
	| destReg leftReg rightReg |
	rightReg := operands at: 0.
	destReg := leftReg := operands at: 1.

	"Save original LHS"
	self machineCodeAt: 0 put: (self adduR: OverflowTemp1 R: leftReg R: ZR). 
	
	"The actual addition"
	self machineCodeAt: 4 put: (self adduR: destReg R: leftReg R: rightReg). 

	"Set sign bit of OverflowTemp2 if sign of result differs from sign of RHS."
	self machineCodeAt: 8 put: (self xorR: OverflowTemp2 R: destReg R: rightReg).
	"Set sign bit of OverflowTemp1 if sign of result differs from sign of LHS."
	self machineCodeAt: 12 put: (self xorR: OverflowTemp1 R: destReg R: OverflowTemp1).
	"Set sign bit of Overflow if sign of result differs from both LHS and RHS, which indicates overflow."
	self machineCodeAt: 16 put: (self andR: Overflow R: OverflowTemp1 R: OverflowTemp2).
	^20! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 2/4/2020 11:07'!
concretizeAddCqR
	| destReg leftReg rightImm |
	rightImm := operands at: 0.
	destReg := leftReg := operands at: 1.
	
	(self isShortOffset: rightImm) ifFalse: [^self concretizeAddCwR].
	
	self machineCodeAt: 0 put: (self addiuR: destReg R: leftReg C: rightImm).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeAddCwR
	| destReg leftReg rightImm |
	rightImm := operands at: 0.
	destReg := leftReg := operands at: 1.
	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: rightImm)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: rightImm)).
	self machineCodeAt: 8 put: (self adduR: destReg R: leftReg R: AT).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 2/4/2020 11:13'!
concretizeAddRRDest: destReg
	| leftReg rightReg |
	rightReg := operands at: 0.
	leftReg := operands at: 1.
	self machineCodeAt: 0 put: (self adduR: destReg R: leftReg R: rightReg).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:47'!
concretizeAlignmentNops
	self assert: machineCodeSize \\ 4 = 0.
	0 to: machineCodeSize - 1 by: 4 do:
		[:p | self machineCodeAt: p put: self nop].
	^machineCodeSize! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 2/4/2020 11:07'!
concretizeAndCqR
	| destReg leftReg rightImm |
	rightImm := operands at: 0.
	destReg := leftReg := operands at: 1.
	
	(self isShortOffset: rightImm) ifFalse: [^self concretizeAndCwR].
	
	self machineCodeAt: 0 put: (self andiR: destReg R: leftReg C: rightImm).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 2/4/2020 11:18'!
concretizeAndCqRDest: destReg
	| value srcReg |
	value := operands at: 0.
	srcReg := operands at: 1.
	(value between: 0 and: 16rFFFF) ifTrue:
		[self machineCodeAt: 0 put: (self andiR: destReg R: srcReg C: value).
		^4].

	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: value)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: value)).
	self machineCodeAt: 8 put: (self andR: destReg R: srcReg R: AT).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeAndCwR
	| destReg leftReg rightImm |
	rightImm := operands at: 0.
	destReg := leftReg := operands at: 1.
	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: rightImm)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: rightImm)).
	self machineCodeAt: 8 put: (self andR: destReg R: leftReg R: AT).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeAndRR
	| destReg leftReg rightReg |
	rightReg := operands at: 0.
	destReg := leftReg := operands at: 1.
	self machineCodeAt: 0 put: (self andR: destReg R: leftReg R: rightReg).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeArithmeticShiftRightCqR
	| distance reg |
	distance := (operands at: 0) min: 31.
	reg := operands at: 1.
	self machineCodeAt: 0 put: (self sraR: reg R: reg C: distance).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeArithmeticShiftRightRR
	| destReg distReg |
	distReg := operands at: 0.
	destReg := operands at: 1.
	self machineCodeAt: 0 put: (self sravR: destReg R: destReg R: distReg).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code' stamp: 'rmacnak 10/27/2015 20:07:43'!
concretizeAt: actualAddress
	"Generate concrete machine code for the instruction at actualAddress,
	 setting machineCodeSize, and answer the following address."

	self assert: actualAddress \\ 4 = 0.
	^super concretizeAt: actualAddress! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeBrEqualRR
	| offset leftReg rightReg |
	offset := self computeJumpTargetOffsetPlus: 4.
	leftReg := operands at: 1.
	rightReg := operands at: 2.
	self machineCodeAt: 0 put: (self beqR: leftReg R: rightReg offset: offset).
	self machineCodeAt: 4 put: (self nop). "Delay slot"
	^8! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeBrLongEqualRR
	| jumpTargetInstruction jumpTargetAddr leftReg rightReg |
	<var: #jumpTargetInstruction type: #'AbstractInstruction *'>
	jumpTargetInstruction := self longJumpTargetAddress.
	self flag: #todo. "Check not crossing 256MB block."
	jumpTargetAddr := jumpTargetInstruction asUnsignedInteger bitAnd: 16rFFFFFFF.
	leftReg := operands at: 1.
	rightReg := operands at: 2.

	self machineCodeAt: 0 put: (self bneR: leftReg R: rightReg offset: 12).
	self machineCodeAt: 4 put: (self nop). "Delay slot"
	self machineCodeAt: 8 put: (self jA: jumpTargetAddr).
	self machineCodeAt: 12 put: self nop. "Delay slot"
	^16! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeBrLongNotEqualRR
	| jumpTargetInstruction jumpTargetAddr leftReg rightReg |
	<var: #jumpTargetInstruction type: #'AbstractInstruction *'>
	jumpTargetInstruction := self longJumpTargetAddress.
	self flag: #todo. "Check not crossing 256MB block."
	jumpTargetAddr := jumpTargetInstruction asUnsignedInteger bitAnd: 16rFFFFFFF.
	leftReg := operands at: 1.
	rightReg := operands at: 2.

	self machineCodeAt: 0 put: (self beqR: leftReg R: rightReg offset: 12).
	self machineCodeAt: 4 put: (self nop). "Delay slot"
	self machineCodeAt: 8 put: (self jA: jumpTargetAddr).
	self machineCodeAt: 12 put: self nop. "Delay slot"
	^16! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeBrNotEqualRR
	| offset leftReg rightReg |
	offset := self computeJumpTargetOffsetPlus: 4.
	leftReg := operands at: 1.
	rightReg := operands at: 2.
	self machineCodeAt: 0 put: (self bneR: leftReg R: rightReg offset: offset).
	self machineCodeAt: 4 put: (self nop). "Delay slot"
	^8! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeBrSignedGreaterEqualRR
	| offset leftReg rightReg |
	offset := self computeJumpTargetOffsetPlus: 8. "Relative to delay slot"
	leftReg := operands at: 1.
	rightReg := operands at: 2.
	self assert: leftReg ~= BranchTemp.
	self assert: rightReg ~= BranchTemp.
	self machineCodeAt: 0 put: (self sltR: BranchTemp R: leftReg R: rightReg).
	self machineCodeAt: 4 put: (self beqR: BranchTemp R: ZR offset: offset).
	self machineCodeAt: 8 put: (self nop). "Delay slot"
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeBrSignedGreaterRR
	| offset leftReg rightReg |
	offset := self computeJumpTargetOffsetPlus: 8. "Relative to delay slot"
	leftReg := operands at: 1.
	rightReg := operands at: 2.
	self assert: leftReg ~= BranchTemp.
	self assert: rightReg ~= BranchTemp.
	self machineCodeAt: 0 put: (self sltR: BranchTemp R: rightReg R: leftReg).
	self machineCodeAt: 4 put: (self bneR: BranchTemp R: ZR offset: offset).
	self machineCodeAt: 8 put: (self nop). "Delay slot"
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeBrSignedLessEqualRR
	| offset leftReg rightReg |
	offset := self computeJumpTargetOffsetPlus: 8. "Relative to delay slot"
	leftReg := operands at: 1.
	rightReg := operands at: 2.
	self assert: leftReg ~= BranchTemp.
	self assert: rightReg ~= BranchTemp.
	self machineCodeAt: 0 put: (self sltR: BranchTemp R: rightReg R: leftReg).
	self machineCodeAt: 4 put: (self beqR: BranchTemp R: ZR offset: offset).
	self machineCodeAt: 8 put: (self nop). "Delay slot"
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeBrSignedLessRR
	| offset leftReg rightReg |
	offset := self computeJumpTargetOffsetPlus: 8. "Relative to delay slot"
	leftReg := operands at: 1.
	rightReg := operands at: 2.
	self assert: leftReg ~= BranchTemp.
	self assert: rightReg ~= BranchTemp.
	self machineCodeAt: 0 put: (self sltR: BranchTemp R: leftReg R: rightReg).
	self machineCodeAt: 4 put: (self bneR: BranchTemp R: ZR offset: offset).
	self machineCodeAt: 8 put: (self nop). "Delay slot"
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeBrUnsignedGreaterEqualRR
	| offset leftReg rightReg |
	offset := self computeJumpTargetOffsetPlus: 8. "Relative to delay slot"
	leftReg := operands at: 1.
	rightReg := operands at: 2.
	self assert: leftReg ~= BranchTemp.
	self assert: rightReg ~= BranchTemp.
	self machineCodeAt: 0 put: (self sltuR: BranchTemp R: leftReg R: rightReg).
	self machineCodeAt: 4 put: (self beqR: BranchTemp R: ZR offset: offset).
	self machineCodeAt: 8 put: (self nop). "Delay slot"
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeBrUnsignedGreaterRR
	| offset leftReg rightReg |
	offset := self computeJumpTargetOffsetPlus: 8. "Relative to delay slot"
	leftReg := operands at: 1.
	rightReg := operands at: 2.
	self assert: leftReg ~= BranchTemp.
	self assert: rightReg ~= BranchTemp.
	self machineCodeAt: 0 put: (self sltuR: BranchTemp R: rightReg R: leftReg).
	self machineCodeAt: 4 put: (self bneR: BranchTemp R: ZR offset: offset).
	self machineCodeAt: 8 put: (self nop). "Delay slot"
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeBrUnsignedLessEqualRR
	| offset leftReg rightReg |
	offset := self computeJumpTargetOffsetPlus: 8. "Relative to delay slot"
	leftReg := operands at: 1.
	rightReg := operands at: 2.
	self assert: leftReg ~= BranchTemp.
	self assert: rightReg ~= BranchTemp.
	self machineCodeAt: 0 put: (self sltuR: BranchTemp R: rightReg R: leftReg).
	self machineCodeAt: 4 put: (self beqR: BranchTemp R: ZR offset: offset).
	self machineCodeAt: 8 put: (self nop). "Delay slot"
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeBrUnsignedLessRR
	| offset leftReg rightReg |
	offset := self computeJumpTargetOffsetPlus: 8. "Relative to delay slot"
	leftReg := operands at: 1.
	rightReg := operands at: 2.
	self assert: leftReg ~= BranchTemp.
	self assert: rightReg ~= BranchTemp.
	self machineCodeAt: 0 put: (self sltuR: BranchTemp R: leftReg R: rightReg).
	self machineCodeAt: 4 put: (self bneR: BranchTemp R: ZR offset: offset).
	self machineCodeAt: 8 put: (self nop). "Delay slot"
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/10/2015 19:57:32'!
concretizeCall
	"Call is used only for calls within code-space, See CallFull for general anywhere in address space calling"
	"Relative branches in MIPS have a displacement of +/- 131kB (signed 18 bits), which is too small to cover
	 the method zone."
	^self concretizeCallFull! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCallFull
	| jumpTargetInstruction jumpTargetAddr |
	<var: #jumpTargetInstruction type: #'AbstractInstruction *'>
	jumpTargetInstruction := self longJumpTargetAddress.
	jumpTargetAddr := jumpTargetInstruction asUnsignedInteger.
	self machineCodeAt: 0 put: (self luiR: TargetReg C: (self high16BitsOf: jumpTargetAddr)).
	self machineCodeAt: 4 put: (self oriR: TargetReg R: TargetReg C: (self low16BitsOf: jumpTargetAddr)).
	self machineCodeAt: 8 put: (self jalR: TargetReg).
	self machineCodeAt: 12 put: self nop. "Delay slot"
	^16! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/23/2020 16:00'!
concretizeClzRR
	| maskReg destReg rightReg |
	maskReg := operands at: 0.
	destReg := rightReg := operands at: 1.
	self machineCodeAt: 0 put: (self clzR: destReg R: maskReg R:  rightReg).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/10/2015 19:25:16'!
concretizeCmpCqR
	^self concretizeCmpCwR! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/23/2015 19:21:42'!
concretizeCmpCwR
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/23/2015 19:19:59'!
concretizeCmpRR
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeDivRR
	| dividendReg divisorReg |
	dividendReg := operands at: 0.
	divisorReg := operands at: 1.
	self machineCodeAt: 0 put: (self divR: dividendReg R: divisorReg).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeFill32
	"fill with operand 0 according to the processor's endianness.
	 You might think this is bogus and we should fill with stop instrurctions instead,
	 but this is used to leave room for a CMBlock header before the code for a block;
	 the gaps get filled in by fillInBlockHeadersAt: after code has been generated."
	self machineCodeAt: 0 put: (operands at: 0).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeJump
	| offset |
	offset := self computeJumpTargetOffsetPlus: 4.
	self flag: #BranchRange.
	self machineCodeAt: 0 put: (self beqR: ZR R: ZR offset: offset).
	self machineCodeAt: 4 put: self nop. "Delay slot"
	^8! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeJumpFull
	| jumpTargetInstruction jumpTargetAddr |
	<var: #jumpTargetInstruction type: #'AbstractInstruction *'>
	jumpTargetInstruction := self longJumpTargetAddress.
	jumpTargetAddr := jumpTargetInstruction asUnsignedInteger.
	self machineCodeAt: 0 put: (self luiR: TargetReg C: (self high16BitsOf: jumpTargetAddr)).
	self machineCodeAt: 4 put: (self oriR: TargetReg R: TargetReg C: (self low16BitsOf: jumpTargetAddr)).
	self machineCodeAt: 8 put: (self jR: TargetReg).
	self machineCodeAt: 12 put: self nop. "Delay slot"
	^16! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/10/2015 20:00:43'!
concretizeJumpLong
	^self concretizeJumpFull! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/23/2015 19:22:08'!
concretizeJumpLongNonZero
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/23/2015 19:22:11'!
concretizeJumpLongZero
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/30/2015 19:31:00'!
concretizeJumpNoOverflow
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/23/2015 19:22:20'!
concretizeJumpNonZero
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/30/2015 19:30:51'!
concretizeJumpOverflow
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeJumpR
	| reg |
	self flag: #OABI. "Does this ever target C code? If so we should move the target into TargetReg first."
	reg := operands at: 0.
	self machineCodeAt: 0 put: (self jR: reg).
	self machineCodeAt: 4 put: self nop. "Delay slot"
	^8! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/23/2015 19:22:27'!
concretizeJumpSignedGreaterEqual
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/23/2015 19:22:29'!
concretizeJumpSignedGreaterThan
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/23/2015 19:22:33'!
concretizeJumpSignedLessEqual
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/23/2015 19:22:39'!
concretizeJumpSignedLessThan
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/23/2015 19:22:41'!
concretizeJumpUnsignedGreaterEqual
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/23/2015 19:22:45'!
concretizeJumpUnsignedGreaterThan
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/23/2015 19:22:48'!
concretizeJumpUnsignedLessEqual
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/23/2015 19:22:51'!
concretizeJumpUnsignedLessThan
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/23/2015 19:23:00'!
concretizeJumpZero
	self unreachable. "Should have been rewritten by noteFollowingConditionalBranch:"
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeLoadEffectiveAddressMwrR
	<var: #offset type: #sqInt>
	| baseReg offset destReg |
	offset := operands at: 0.
	baseReg := operands at: 1.
	destReg := operands at: 2.
	(self isShortOffset: offset) ifTrue:
		[self machineCodeAt: 0 put: (self addiuR: destReg R: baseReg C: offset).
		^4].
	
	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: offset)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: offset)).
	self machineCodeAt: 8 put: (self adduR: destReg R: baseReg R: AT).
	^12.
! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeLogicalShiftLeftCqR
	| distance reg |
	distance := (operands at: 0) min: 31.
	reg := operands at: 1.
	self machineCodeAt: 0 put: (self sllR: reg R: reg C: distance).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeLogicalShiftLeftRR
	| destReg distReg |
	distReg := operands at: 0.
	destReg := operands at: 1.
	self machineCodeAt: 0 put: (self sllvR: destReg R: destReg R: distReg).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeLogicalShiftRightCqR
	| distance reg |
	distance := (operands at: 0) min: 31.
	reg := operands at: 1.
	self machineCodeAt: 0 put: (self srlR: reg R: reg C: distance).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeLogicalShiftRightRR
	| destReg distReg |
	distReg := operands at: 0.
	destReg := operands at: 1.
	self machineCodeAt: 0 put: (self srlvR: destReg R: destReg R: distReg).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveAbR
	| srcAddr destReg |
	srcAddr := operands at: 0.
	destReg := operands at: 1.

	(self isAddressRelativeToVarBase: srcAddr) ifTrue:
		[self machineCodeAt: 0 put: (self lwR: destReg base: ConcreteVarBaseReg offset: srcAddr - cogit varBaseAddress).
		 ^4].

	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: srcAddr)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: srcAddr)).
	self machineCodeAt: 8 put: (self lbuR: destReg base: AT offset: 0).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveAwR
	| srcAddr destReg |
	srcAddr := operands at: 0.
	destReg := operands at: 1.

	(self isAddressRelativeToVarBase: srcAddr) ifTrue:
		[self machineCodeAt: 0 put: (self lwR: destReg base: ConcreteVarBaseReg offset: srcAddr - cogit varBaseAddress).
		 ^4].

	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: srcAddr)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: srcAddr)).
	self machineCodeAt: 8 put: (self lwR: destReg base: AT offset: 0).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 2/4/2020 11:07'!
concretizeMoveCqR
	<var: #word type: #sqInt>
	| word reg |
	word := operands at: 0.
	reg := operands at: 1.
	
	(self isShortOffset: word) ifFalse: [^self concretizeMoveCwR].
	"Could also load up to 16rFFFF with ori or 16rXXXX0000 with lui"
	
	self machineCodeAt: 0 put: (self addiuR: reg R: ZR C: word).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveCwR
	<var: #word type: #sqInt>
	| word reg |
	word := operands at: 0.
	reg := operands at: 1.
	self machineCodeAt: 0 put: (self luiR: reg C: (self high16BitsOf: word)).
	self machineCodeAt: 4 put: (self oriR: reg R: reg C: (self low16BitsOf: word)).
	^8! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveHighR
	| destReg |
	destReg := operands at: 0.
	self machineCodeAt: 0 put: (self mfhiR: destReg).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveLowR
	| destReg |
	destReg := operands at: 0.
	self machineCodeAt: 0 put: (self mfloR: destReg).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveM16rR
	<var: #offset type: #sqInt>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	self machineCodeAt: 0 put: (self lhuR: destReg base: srcReg offset: offset).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveMbrR
	<var: #offset type: #sqInt>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	self machineCodeAt: 0 put: (self lbuR: destReg base: srcReg offset: offset).
	^4
	! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveMwrR
	<var: #offset type: #sqInt>
	| baseReg offset destReg |
	offset := operands at: 0.
	baseReg := operands at: 1.
	destReg := operands at: 2.
	(self isShortOffset: offset) ifTrue:
		[self machineCodeAt: 0 put: (self lwR: destReg base: baseReg offset: offset).
		^4].
	
	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: offset)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: offset)).
	self machineCodeAt: 8 put: (self adduR: AT R: baseReg R: AT).
	self machineCodeAt: 12 put: (self lwR: destReg base: AT offset: 0).
	^16.
! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRAb
	| srcReg destAddr |
	srcReg := operands at: 0.
	destAddr := operands at: 1.

	(self isAddressRelativeToVarBase: destAddr) ifTrue:
		[self machineCodeAt: 0 put: (self swR: srcReg base: ConcreteVarBaseReg offset: destAddr - cogit varBaseAddress).
		 ^4].

	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: destAddr)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: destAddr)).
	self machineCodeAt: 8 put: (self sbR: srcReg base: AT offset: 0).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRAw
	| srcReg destAddr |
	srcReg := operands at: 0.
	destAddr := operands at: 1.

	(self isAddressRelativeToVarBase: destAddr) ifTrue:
		[self machineCodeAt: 0 put: (self swR: srcReg base: ConcreteVarBaseReg offset: destAddr - cogit varBaseAddress).
		 ^4].

	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: destAddr)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: destAddr)).
	self machineCodeAt: 8 put: (self swR: srcReg base: AT offset: 0).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRM16r
	<var: #offset type: #sqInt>
	| srcReg offset destReg |
	srcReg := operands at: 0.
	offset := operands at: 1.
	destReg := operands at: 2.
	self machineCodeAt: 0 put: (self shR: srcReg base: destReg offset: offset).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRMbr
	<var: #offset type: #sqInt>
	| srcReg offset destReg |
	srcReg := operands at: 0.
	offset := operands at: 1.
	destReg := operands at: 2.
	self machineCodeAt: 0 put: (self sbR: srcReg base: destReg offset: offset).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRMwr
	<var: #offset type: #sqInt>
	| srcReg offset baseReg |
	srcReg := operands at: 0.
	offset := operands at: 1.
	baseReg := operands at: 2.
	self machineCodeAt: 0 put: (self swR: srcReg base: baseReg offset: offset).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRR
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	self machineCodeAt: 0 put: (self adduR: destReg R: srcReg R: ZR).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRXbrR
	| srcReg indexReg baseReg |
	srcReg := operands at: 0.
	indexReg := operands at: 1.
	baseReg := operands at: 2.
	self machineCodeAt: 0 put: (self adduR: AT R: baseReg R: indexReg).
	self machineCodeAt: 4 put: (self sbR: srcReg base: AT offset: 0).
	^8! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRXwrR
	| srcReg indexReg baseReg |
	srcReg := operands at: 0.
	indexReg := operands at: 1.
	baseReg := operands at: 2.
	self machineCodeAt: 0 put: (self sllR: AT R: indexReg C: 2). "index is number of words"
	self machineCodeAt: 4 put: (self adduR: AT R: baseReg R: AT).
	self machineCodeAt: 8 put: (self swR: srcReg base: AT offset: 0).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveXbrRR
	| indexReg baseReg destReg |
	indexReg := operands at: 0. "index is number of *bytes*"
	baseReg := operands at: 1.
	destReg := operands at: 2.
	self machineCodeAt: 0 put: (self adduR: AT R: baseReg R: indexReg).
	self machineCodeAt: 4 put: (self lbuR: destReg base: AT offset: 0).
	^8! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveXwrRR
	| indexReg baseReg destReg |
	indexReg := operands at: 0.
	baseReg := operands at: 1.
	destReg := operands at: 2.
	self machineCodeAt: 0 put: (self sllR: AT R: indexReg C: 2). "index is in words"
	self machineCodeAt: 4 put: (self adduR: AT R: baseReg R: AT).
	self machineCodeAt: 8 put: (self lwR: destReg base: AT offset: 0).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMulCheckOverflowRR
	| destReg leftReg rightReg |
	rightReg := operands at: 0.
	destReg := leftReg := operands at: 1.

	"Overflow occured if the sign bit of the low part is different from the high part."
	self machineCodeAt: 0 put: (self multR: leftReg R: rightReg).
	self machineCodeAt: 4 put: (self mfloR: destReg).
	self machineCodeAt: 8 put: (self sraR: OverflowTemp1 R: destReg C: 31).
	self machineCodeAt: 12 put: (self mfhiR: OverflowTemp2).
	^16! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeNegateR
	| reg |
	reg := operands at: 0.
	self machineCodeAt: 0 put: (self subuR: reg R: ZR R: reg).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeNop
	self machineCodeAt: 0 put: self nop.
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeOrCqR
	| destReg leftReg rightImm |
	rightImm := operands at: 0.
	destReg := leftReg := operands at: 1.

	(rightImm between: 0 and: 16rFFFF) ifFalse: [^self concretizeOrCwR].
	
	self machineCodeAt: 0 put: (self oriR: destReg R: leftReg C: rightImm).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 1/23/2020 16:16'!
concretizeOrCqRR
	| value srcReg dstReg |
	value := operands at: 0.
	srcReg := operands at: 1.
	dstReg := operands at: 2.
	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: value)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: value)).
	self machineCodeAt: 8 put: (self orR: dstReg R: srcReg R: AT).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeOrCwR
	| destReg leftReg rightImm |
	rightImm := operands at: 0.
	destReg := leftReg := operands at: 1.
	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: rightImm)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: rightImm)).
	self machineCodeAt: 8 put: (self orR: destReg R: leftReg R: AT).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeOrRR
	| destReg leftReg rightReg |
	rightReg := operands at: 0.
	destReg := leftReg := operands at: 1.
	self machineCodeAt: 0 put: (self orR: destReg R: leftReg R: rightReg).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePopR
	| destReg |
	destReg := operands at: 0.
	self machineCodeAt: 0 put: (self lwR: destReg base: SP offset: 0).
	self machineCodeAt: 4 put: (self addiuR: SP R: SP C: 4).
	^8! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePrefetchAw
	| addressOperand |
	addressOperand := operands at: 0.
	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: addressOperand)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: addressOperand)).
	self machineCodeAt: 8 put: (self prefR: AT offset: 0 hint: HintLoad).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 11/16/2015 20:10:50'!
concretizePushCq
	^self concretizePushCw! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePushCw
	| value |
	value := operands at: 0.
	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: value)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: value)).
	self machineCodeAt: 8 put: (self addiuR: SP R: SP C: -4).
	self machineCodeAt: 12 put: (self swR: AT base: SP offset: 0).
	^16! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePushR
	| srcReg |
	srcReg := operands at: 0.
	self machineCodeAt: 0 put: (self addiuR: SP R: SP C: -4).
	self machineCodeAt: 4 put: (self swR: srcReg base: SP offset: 0).
	^8! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeRetN
	<var: #offset type: #sqInt>
	| offset |
	offset := operands at: 0.
	self machineCodeAt: 0 put: (self jR: RA).
	offset = 0 
		ifTrue: [self machineCodeAt: 4 put: self nop "Delay slot"]
		ifFalse: [self machineCodeAt: 4 put: (self addiuR: SP R: SP C: offset) "We actually get to fill a delay slot!!"].
	^8! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeStop
	self machineCodeAt: 0 put: self stop.
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSubCheckOverflowCqR
	| destReg leftReg rightImm |
	rightImm := operands at: 0.
	destReg := leftReg := operands at: 1.

	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: rightImm)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: rightImm)).

	"Save original LHS"
	self machineCodeAt: 8 put: (self adduR: OverflowTemp1 R: leftReg R: ZR). 
	
	"The actual subtraction"
	self machineCodeAt: 12 put: (self subuR: destReg R: leftReg R: AT). 

	"Set sign bit of OverflowTemp2 if sign of result differs from sign of RHS."
	self machineCodeAt: 16 put: (self xorR: OverflowTemp2 R: destReg R: AT).
	"Set sign bit of OverflowTemp1 if sign of result differs from sign of LHS."
	self machineCodeAt: 20 put: (self xorR: OverflowTemp1 R: destReg R: OverflowTemp1).
	"Set sign bit of Overflow if sign of result differs from both LHS and RHS, which indicates overflow."
	self machineCodeAt: 24 put: (self andR: Overflow R: OverflowTemp1 R: OverflowTemp2).
	^28! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSubCheckOverflowRR
	| destReg leftReg rightReg |
	rightReg := operands at: 0.
	destReg := leftReg := operands at: 1.

	"Save original LHS"
	self machineCodeAt: 0 put: (self adduR: OverflowTemp1 R: leftReg R: ZR). 
	
	"The actual subtraction"
	self machineCodeAt: 4 put: (self subuR: destReg R: leftReg R: rightReg). 

	"Set sign bit of OverflowTemp2 if sign of result differs from sign of RHS."
	self machineCodeAt: 8 put: (self xorR: OverflowTemp2 R: destReg R: rightReg).
	"Set sign bit of OverflowTemp1 if sign of result differs from sign of LHS."
	self machineCodeAt: 12 put: (self xorR: OverflowTemp1 R: destReg R: OverflowTemp1).
	"Set sign bit of Overflow if sign of result differs from both LHS and RHS, which indicates overflow."
	self machineCodeAt: 16 put: (self andR: Overflow R: OverflowTemp1 R: OverflowTemp2).
	^20! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 2/4/2020 11:08'!
concretizeSubCqR
	| destReg leftReg rightImm |
	rightImm := operands at: 0.
	destReg := leftReg := operands at: 1.
	
	(self isShortOffset: rightImm negated) ifFalse: [^self concretizeSubCwR].
	
	self machineCodeAt: 0 put: (self addiuR: destReg R: leftReg C: rightImm negated).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSubCwR
	| destReg leftReg rightImm |
	rightImm := operands at: 0.
	destReg := leftReg := operands at: 1.
	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: rightImm)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: rightImm)).
	self machineCodeAt: 8 put: (self subuR: destReg R: leftReg R: AT).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 2/4/2020 11:11'!
concretizeSubRRDest: destReg
	| leftReg rightReg |
	rightReg := operands at: 0.
	leftReg := operands at: 1.
	self machineCodeAt: 0 put: (self subuR: destReg R: leftReg R: rightReg).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 2/4/2020 11:08'!
concretizeTstCqR
	| leftReg rightImm |
	rightImm := operands at: 0.
	leftReg := operands at: 1.
	
	(self isShortOffset: rightImm) ifFalse: [^self concretizeTstCwR].
	
	self machineCodeAt: 0 put: (self andiR: Cmp R: leftReg C: rightImm).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeTstCwR
	| leftReg rightImm |
	rightImm := operands at: 0.
	leftReg := operands at: 1.
	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: rightImm)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: rightImm)).
	self machineCodeAt: 8 put: (self andR: Cmp R: leftReg R: AT).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'rmacnak 12/16/2015 20:40'!
concretizeUnimplemented
	self error: 'Unimplemented RTL instruction'.
	^0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeXorCwR
	| destReg leftReg rightImm |
	rightImm := operands at: 0.
	destReg := leftReg := operands at: 1.
	self machineCodeAt: 0 put: (self luiR: AT C: (self high16BitsOf: rightImm)).
	self machineCodeAt: 4 put: (self oriR: AT R: AT C: (self low16BitsOf: rightImm)).
	self machineCodeAt: 8 put: (self xorR: destReg R: leftReg R: AT).
	^12! !

!CogMIPSELCompiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeXorRR
	| destReg leftReg rightReg |
	rightReg := operands at: 0.
	destReg := leftReg := operands at: 1.
	self machineCodeAt: 0 put: (self xorR: destReg R: leftReg R: rightReg).
	^4! !

!CogMIPSELCompiler methodsFor: 'generate machine code' stamp: 'eem 2/4/2020 11:18'!
dispatchConcretize
	"Attempt to generate concrete machine code for the instruction at address.
	 This is the inner dispatch of concretizeAt: actualAddress which exists only
	 to get around the branch size limits in the SqueakV3 (blue book derived)
	 bytecode set."		 
	opcode caseOf: {
		[BrEqualRR]						-> [^self concretizeBrEqualRR].
		[BrNotEqualRR]					-> [^self concretizeBrNotEqualRR].
		[BrUnsignedLessRR]			-> [^self concretizeBrUnsignedLessRR].
		[BrUnsignedLessEqualRR]		-> [^self concretizeBrUnsignedLessEqualRR].
		[BrUnsignedGreaterRR]			-> [^self concretizeBrUnsignedGreaterRR].
		[BrUnsignedGreaterEqualRR]	-> [^self concretizeBrUnsignedGreaterEqualRR].
		[BrSignedLessRR]				-> [^self concretizeBrSignedLessRR].
		[BrSignedLessEqualRR]			-> [^self concretizeBrSignedLessEqualRR].
		[BrSignedGreaterRR]			-> [^self concretizeBrSignedGreaterRR].
		[BrSignedGreaterEqualRR]		-> [^self concretizeBrSignedGreaterEqualRR].	
		[BrLongEqualRR]				-> [^self concretizeBrLongEqualRR].
		[BrLongNotEqualRR]				-> [^self concretizeBrLongNotEqualRR].
		[MulRR]				-> [^self concretizeUnimplemented].
		[DivRR]				-> [^self concretizeDivRR].
		[MoveLowR]		-> [^self concretizeMoveLowR].
		[MoveHighR]		-> [^self concretizeMoveHighR].

										
		"Noops & Pseudo Ops"
		[Label]					-> [^self concretizeLabel].
		[AlignmentNops]		-> [^self concretizeAlignmentNops].
		[Fill32]					-> [^self concretizeFill32].
		[Nop]					-> [^self concretizeNop].
		"Control"
		[Call]						-> [^self concretizeCall]. "call code within code space"
		[CallFull]					-> [^self concretizeCallFull]. "call code anywhere in address space"
		[JumpR]						-> [^self concretizeJumpR].
		[JumpFull]					-> [^self concretizeJumpFull]."jump within address space"
		[JumpLong]					-> [^self concretizeJumpLong]."jumps witihn code space"
		[JumpLongZero]			-> [^self concretizeJumpLongZero].
		[JumpLongNonZero]		-> [^self concretizeJumpLongNonZero].
		[Jump]						-> [^self concretizeJump].
		[JumpZero]					-> [^self concretizeJumpZero].
		[JumpNonZero]				-> [^self concretizeJumpNonZero].
		[JumpNegative]				-> [^self concretizeUnimplemented].
		[JumpNonNegative]			-> [^self concretizeUnimplemented].
		[JumpOverflow]				-> [^self concretizeJumpOverflow].
		[JumpNoOverflow]			-> [^self concretizeJumpNoOverflow].
		[JumpCarry]				-> [^self concretizeUnimplemented].
		[JumpNoCarry]				-> [^self concretizeUnimplemented].
		[JumpLess]					-> [^self concretizeJumpSignedLessThan].
		[JumpGreaterOrEqual]		-> [^self concretizeJumpSignedGreaterEqual].
		[JumpGreater]				-> [^self concretizeJumpSignedGreaterThan].
		[JumpLessOrEqual]			-> [^self concretizeJumpSignedLessEqual].
		[JumpBelow]				-> [^self concretizeJumpUnsignedLessThan].
		[JumpAboveOrEqual]		-> [^self concretizeJumpUnsignedGreaterEqual].
		[JumpAbove]				-> [^self concretizeJumpUnsignedGreaterThan].
		[JumpBelowOrEqual]		-> [^self concretizeJumpUnsignedLessEqual].
		[JumpFPEqual]				-> [^self concretizeUnimplemented].
		[JumpFPNotEqual]			-> [^self concretizeUnimplemented].
		[JumpFPLess]				-> [^self concretizeUnimplemented].
		[JumpFPGreaterOrEqual]	-> [^self concretizeUnimplemented].
		[JumpFPGreater]			-> [^self concretizeUnimplemented].
		[JumpFPLessOrEqual]		-> [^self concretizeUnimplemented].
		[JumpFPOrdered]			-> [^self concretizeUnimplemented].
		[JumpFPUnordered]			-> [^self concretizeUnimplemented].
		[RetN]						-> [^self concretizeRetN].
		[Stop]						-> [^self concretizeStop].
		"Arithmetic"
		[AddCqR]					-> [^self concretizeAddCqR].
		[AndCqR]					-> [^self concretizeAndCqRDest: (operands at: 1)].
		[AndCqRR]					-> [^self concretizeAndCqRDest: (operands at: 2)].
		[OrCqRR]					-> [^self concretizeOrCqRR].
		[CmpCqR]					-> [^self concretizeCmpCqR].
		[OrCqR]					-> [^self concretizeOrCqR].
		[SubCqR]					-> [^self concretizeSubCqR].
		[TstCqR]					-> [^self concretizeTstCqR].
		[XorCqR]					-> [^self concretizeUnimplemented].
		[AddCwR]					-> [^self concretizeAddCwR].
		[AndCwR]					-> [^self concretizeAndCwR].
		[CmpCwR]					-> [^self concretizeCmpCwR].
		[OrCwR]					-> [^self concretizeOrCwR].
		[SubCwR]					-> [^self concretizeSubCwR].
		[XorCwR]					-> [^self concretizeXorCwR].
		[AddRR]					-> [^self concretizeAddRRDest: (operands at: 1)].
		[AndRR]					-> [^self concretizeAndRR].
		[CmpRR]					-> [^self concretizeCmpRR].
		[OrRR]						-> [^self concretizeOrRR].
		[SubRR]					-> [^self concretizeSubRRDest: (operands at: 1)].
		[XorRR]						-> [^self concretizeXorRR].
		[AddRRR]					-> [^self concretizeAddRRDest: (operands at: 2)].
		[SubRRR]					-> [^self concretizeSubRRDest: (operands at: 2)].
		[AddRdRd]					-> [^self concretizeUnimplemented].
		[CmpRdRd]					-> [^self concretizeUnimplemented].
		[DivRdRd]					-> [^self concretizeUnimplemented].
		[MulRdRd]					-> [^self concretizeUnimplemented].
		[SubRdRd]					-> [^self concretizeUnimplemented].
		[SqrtRd]					-> [^self concretizeUnimplemented].
		[NegateR]						-> [^self concretizeNegateR].
		[LoadEffectiveAddressMwrR]	-> [^self concretizeLoadEffectiveAddressMwrR].
		[ArithmeticShiftRightCqR]		-> [^self concretizeArithmeticShiftRightCqR].
		[LogicalShiftRightCqR]			-> [^self concretizeLogicalShiftRightCqR].
		[LogicalShiftLeftCqR]			-> [^self concretizeLogicalShiftLeftCqR].
		[ArithmeticShiftRightRR]			-> [^self concretizeArithmeticShiftRightRR].
		[LogicalShiftLeftRR]				-> [^self concretizeLogicalShiftLeftRR].
		[LogicalShiftRightRR]			-> [^self concretizeLogicalShiftRightRR].
		[ClzRR]						-> [^self concretizeClzRR].
		"Data Movement"
		[MoveCqR]			-> [^self concretizeMoveCqR].
		[MoveCwR]			-> [^self concretizeMoveCwR].
		[MoveRR]			-> [^self concretizeMoveRR].
		[MoveAwR]			-> [^self concretizeMoveAwR].
		[MoveRAw]			-> [^self concretizeMoveRAw].
		[MoveAbR]			-> [^self concretizeMoveAbR].
		[MoveRAb]			-> [^self concretizeMoveRAb].
		[MoveMbrR]			-> [^self concretizeMoveMbrR].
		[MoveRMbr]			-> [^self concretizeUnimplemented].
		[MoveM16rR]		-> [^self concretizeMoveM16rR].
		[MoveRM16r]		-> [^self concretizeMoveRM16r].
		[MoveM64rRd]		-> [^self concretizeUnimplemented].
		[MoveMwrR]		-> [^self concretizeMoveMwrR].
		[MoveXbrRR]		-> [^self concretizeMoveXbrRR].
		[MoveRXbrR]		-> [^self concretizeMoveRXbrR].
		[MoveXwrRR]		-> [^self concretizeMoveXwrRR].
		[MoveRXwrR]		-> [^self concretizeMoveRXwrR].
		[MoveRMwr]		-> [^self concretizeMoveRMwr].
		[MoveRdM64r]		-> [^self concretizeUnimplemented].
		[PopR]				-> [^self concretizePopR].
		[PushR]				-> [^self concretizePushR].
		[PushCq]			-> [^self concretizePushCq].
		[PushCw]			-> [^self concretizePushCw].
		[PrefetchAw]		-> [^self concretizePrefetchAw].
		[AddCheckOverflowCqR] -> [^self concretizeAddCheckOverflowCqR].
		[AddCheckOverflowRR] -> [^self concretizeAddCheckOverflowRR].
		[SubCheckOverflowCqR] -> [^self concretizeSubCheckOverflowCqR].
		[SubCheckOverflowRR] -> [^self concretizeSubCheckOverflowRR].
		[MulCheckOverflowRR] -> [^self concretizeMulCheckOverflowRR].
		"Conversion"
		[ConvertRRd]		-> [^self concretizeUnimplemented]}.

	^0 "keep Slang happy"! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/21/2015 16:24:57'!
divR: dividendReg R: divisorReg
	self flag: #todo. "Ascertain what ISA revision adds this instruction"
	^self rtype: SPECIAL rs: dividendReg rt: divisorReg rd: 0 sa: 0 funct: DIV! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'eem 2/4/2020 12:07'!
flushDCacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	"If there is a dual mapped code zone (the normal zone but marked with read/execute, and a
	 read/write zone codeToDataDelta bytes away) then flush the data cache for the corresponding
	 range in the read/write zone and invalidate the data cache for the read/execute zone."
	<cmacro: '(me,startAddress,endAddress) 0'>
	"Currently we don't know whether a dual zone on a 32-bit MIPS would require cache
	 flushing to ensure coherence.  So for now this is a no op."! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'eem 2/3/2020 19:32'!
flushICacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	"Flush the instruction cache from (startAddress to endAddress].
	 If there is a dual mapped code zone (the normal zone but marked with read/execute, and a
	 read/write zone codeToDataDelta bytes away) then also flush the data cache for the corresp-
	 onding range in the read/write zone and invalidate the data cache for the read/execute zone."
	<cmacro: '(me,startAddress,endAddress) cacheflush((char*) startAddress, endAddress - startAddress, ICACHE)'>
	"See http://www.linux-mips.org/wiki/Cacheflush_Syscall"! !

!CogMIPSELCompiler methodsFor: 'abi' stamp: 'rmacnak 11/9/2015 20:10:48'!
fullCallsAreRelative
	"Answer if CallFull and/or JumpFull are relative and hence need relocating on method
	 compation. If so, they are annotated with IsRelativeCall in methods and relocated in
	 relocateIfCallOrMethodReference:mcpc:delta:"
	^false! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 11/11/2015 19:34:40'!
functionAtAddress: mcpc
	^(objectMemory longAt: mcpc) bitAnd: 63
! !

!CogMIPSELCompiler methodsFor: 'abstract instructions' stamp: 'rmacnak 11/21/2015 16:00:23'!
genDivR: abstractRegDivisor R: abstractRegDividend Quo: abstractRegQuotient Rem: abstractRegRemainder
	cogit gen: DivRR operand: abstractRegDividend operand: abstractRegDivisor.
	cogit gen: MoveLowR operand: abstractRegQuotient.
	cogit gen: MoveHighR operand: abstractRegRemainder.
	^0! !

!CogMIPSELCompiler methodsFor: 'abi' stamp: 'eem 12/18/2015 14:16'!
genMarshallNArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3
	"Generate the code to pass up to four arguments in a C run-time call.  Hack: each argument is
	 either a negative number, which encodes a constant, or a non-negative number, that of a register.

	 Run-time calls have no more than four arguments, so chosen so that on ARM, where in its C ABI the
	 first four integer arguments are passed in registers, all arguments can be passed in registers.  We
	 defer to the back end to generate this code not so much that the back end knows whether it uses
	 the stack or registers to pass arguments (it does, but...). In fact we defer for an extremely evil reason.
	 Doing so allows the x64 (where up to 6 args are passed) to assign the register arguments in an order
	 that allows some of the argument registers to be used for specific abstract  registers, specifically
	 ReceiverResultReg and ClassReg.  This is evil, evil, evil, but also it's really nice to keep using the old
	 register assignments the original author has grown accustomed to."
	<inline: true>
	self flag: #OABI.
	numArgs = 0 ifTrue: [^self].
	(cogit isTrampolineArgConstant: regOrConst0)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst0) R: A0]
		ifFalse: [cogit MoveR: regOrConst0 R: A0].
	numArgs = 1 ifTrue: [^self].
	(cogit isTrampolineArgConstant: regOrConst1)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst1) R: A1]
		ifFalse: [cogit MoveR: regOrConst1 R: A1].
	numArgs = 2 ifTrue: [^self].
	(cogit isTrampolineArgConstant: regOrConst2)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst2) R: A2]
		ifFalse: [cogit MoveR: regOrConst2 R: A2].
	numArgs = 3 ifTrue: [^self].
	(cogit isTrampolineArgConstant: regOrConst3)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst3) R: A3]
		ifFalse: [cogit MoveR: regOrConst3 R: A3]! !

!CogMIPSELCompiler methodsFor: 'abstract instructions' stamp: 'eem 12/16/2015 09:44'!
genMulR: regSource R: regDest
	^cogit gen: MulRR operand: regSource operand: regDest! !

!CogMIPSELCompiler methodsFor: 'smalltalk calling convention' stamp: 'rmacnak 11/10/2015 19:25:36'!
genPushRegisterArgsForAbortMissNumArgs: numArgs
	"Ensure that the register args are pushed before the outer and
	 inner retpcs at an entry miss for arity <= self numRegArgs.  The
	 outer retpc is that of a call at a send site.  The inner is the call
	 from a method or PIC abort/miss to the trampoline."

	"Putting the receiver and args above the return address means the
	 CoInterpreter has a single machine-code frame format which saves
	 us a lot of work."

	"Iff there are register args convert
		sp		->	outerRetpc			(send site retpc)
		linkReg = innerRetpc			(PIC abort/miss retpc)
	 to
		base	->	receiver
					(arg0)
					(arg1)
		sp		->	outerRetpc			(send site retpc)
		sp		->	linkReg/innerRetpc	(PIC abort/miss retpc)"
						
	self flag: #inefficient. "Update SP once."
	
	numArgs <= cogit numRegArgs ifTrue:
		[self assert: cogit numRegArgs <= 2.
		 cogit MoveMw: 0 r: SPReg R: TempReg. "Save return address"
		 cogit MoveR: ReceiverResultReg Mw: 0 r: SPReg.
		 numArgs > 0 ifTrue:
			[cogit PushR: Arg0Reg.
			 numArgs > 1 ifTrue:
				[cogit PushR: Arg1Reg]].
		cogit PushR: TempReg]. "push back return address"
	cogit PushR: LinkReg! !

!CogMIPSELCompiler methodsFor: 'smalltalk calling convention' stamp: 'rmacnak 11/10/2015 19:24:20'!
genPushRegisterArgsForNumArgs: numArgs scratchReg: ignored
	"Ensure that the register args are pushed before the retpc for arity <= self numRegArgs."
	"This is easy on a RISC like ARM because the return address is in the link register.  Putting
	 the receiver and args above the return address means the CoInterpreter has a single
	 machine-code frame format which saves us a lot of work
	NOTA BENE: we do NOT push the return address here, which means it must be dealt with later."
	self flag: #inefficient. "Update SP once."
	numArgs <= cogit numRegArgs ifTrue:
		[self assert: cogit numRegArgs <= 2.
		 cogit PushR: ReceiverResultReg.
		numArgs > 0 ifTrue:
			[cogit PushR: Arg0Reg.
			 numArgs > 1 ifTrue:
				[cogit PushR: Arg1Reg]]]! !

!CogMIPSELCompiler methodsFor: 'abi' stamp: 'eem 5/29/2017 10:25'!
genRemoveNArgsFromStack: n
	"This is a no-op on MIPS since the ABI passes up to 4 args in registers and trampolines currently observe that limit."
	<inline: true>
	self assert: n <= 4.
	^0! !

!CogMIPSELCompiler methodsFor: 'abi' stamp: 'eem 3/3/2016 16:19'!
genRestoreRegs: regMask
	"Restore the registers in regMask as saved by genSaveRegs:.
	 We don't need to do anything because all of the abstract registers are
	 allocated to C preserved registers.  But for the future..."
	self assert: regMask = 0.
	self deny: (regMask anyMask: (cogit registerMaskFor: SP and: FP and: RA)).
	R0 to: R28 do:
		[:reg|
		 (regMask anyMask: (cogit registerMaskFor: reg)) ifTrue:
			[cogit PopR: reg]].
	^0! !

!CogMIPSELCompiler methodsFor: 'abi' stamp: 'eem 3/3/2016 12:44'!
genSaveRegs: regMask
	"Save the registers in regMask for a call into the C run-time from a trampoline.
	 We don't need to do anything because all of the abstract registers are
	 allocated to C preserved registers.  But for the future..."
	self assert: regMask = 0.
	self deny: (regMask anyMask: (cogit registerMaskFor: SP and: FP and: RA)).
	R28 to: R0 by: -1 do:
		[:reg|
		 (regMask anyMask: (cogit registerMaskFor: reg)) ifTrue:
			[cogit PushR: reg]].
	^0! !

!CogMIPSELCompiler methodsFor: 'abstract instructions' stamp: 'rmacnak 11/24/2015 20:49:19'!
genSubstituteReturnAddress: retpc
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^cogit MoveCw: retpc R: RA! !

!CogMIPSELCompiler methodsFor: 'abstract instructions' stamp: 'eem 4/1/2020 13:18'!
genSubstituteReturnAddressR: retpcReg
	<inline: true>
	^cogit MoveR: retpcReg R: RA! !

!CogMIPSELCompiler methodsFor: 'disassembly' stamp: 'eem 9/14/2020 15:27'!
generalPurposeRegisterMap
	<doNotGenerate>
	"Answer a Dictionary from register getter to register index."
	^Dictionary newFromPairs:
		{	#at. AT.
			#v0. V0.
			#v1. V1.
			#a0. A0.
			#a1. A1.
			#a2. A2.
			#a3. A3.
			#t0. T0.
			#t1. T1.
			#t2. T2.
			#t3. T3.
			#t4. T4.
			#t5. T5.
			#t6. T6.
			#t7. T7.
			#s0. S0.
			#s1. S1.
			#s2. S2.
			#s3. S3.
			#s4. S4.
			#s5. S5.
			#s6. S6.
			#s7. S7.
			#t8. T8.
			#t9. T9.
			#k0. K0.
			#k1. K1.
			#gp. GP	}! !

!CogMIPSELCompiler methodsFor: 'accessing' stamp: 'rmacnak 12/16/2015 19:35'!
getDefaultCogCodeSize
	"Return the default number of bytes to allocate for native code at startup.
	 The actual value can be set via vmParameterAt: and/or a preference in the ini file."
	<inline: true>
	^1024 * 1536! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'eem 8/4/2021 11:56'!
has64BitPerformanceCounter
	"At least for the moment..."
	<inline: #always>
	^false! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
hasConditionRegister
	"Answer if the receiver supports, e.g., JumpOverflow after a regular AddRR"
	^false! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'rmacnak 11/23/2015 22:30:55'!
hasDoublePrecisionFloatingPointSupport
	<inline: true>
	self flag: #todo.
	^false! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'nice 8/26/2019 15:45'!
hasLZCNTInstructions
	"Answer if the processor has a LZCNT (leading zero count) instruction"
	<inline: true>
	^true "CLZ"! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'rmacnak 11/9/2015 20:05:20'!
hasLinkRegister
	^true "RA"! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
hasThreeAddressArithmetic
	"Answer if the receiver supports three-address arithmetic instructions"
	^true! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
hasVarBaseRegister
	"Answer if the processor has a dedicated callee-saved register to point to
	 the base of commonly-accessed variables."
	^true "S6"! !

!CogMIPSELCompiler methodsFor: 'encoding' stamp: 'rmacnak 12/17/2015 20:27'!
high16BitsOf: word
	<var: #word type: #usqInt>
	^word >> 16! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 11/16/2015 22:46:24'!
inlineCacheTagAt: callSiteReturnAddress
	"Answer the inline cache tag for the return address of a send."
	
	"MoveCwR ClassReg selectorIndex/expectedClass
	 Call: unlinked send stub/expectedTarget
	 Push ReceiverResult <-- callSiteReturnAddress"
	
	"lui s3, selector/tagHigh
	 ori s3, s3, selector/tagLow
	 lui t9, stub/targetHigh
	 ori t9, t9, stub/targetLow
	 jalr t9
	 nop (delay slot)
	 ...  <-- callSiteReturnAddress"
	
	<var: #callSiteReturnAddress type: #usqInt>
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 24) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 20) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 16) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 12) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 8) = SPECIAL.
	self assert: (self functionAtAddress: callSiteReturnAddress - 8) = JALR.
	self assert: (objectMemory longAt: callSiteReturnAddress - 4) = self nop.

	^self literalAtAddress: callSiteReturnAddress - 20! !

!CogMIPSELCompiler methodsFor: 'disassembly' stamp: 'rmacnak 12/2/2015 18:54:36'!
instructionSizeAt: pc
	"Answer the instruction size at pc."
	^4! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'rmacnak 11/9/2015 20:12:41'!
isAddressRelativeToVarBase: varAddress
	<inline: true>
	<var: #varAddress type: #usqInt>
	"Support for addressing variables off the dedicated VarBaseReg"
	^varAddress notNil
	  and: [(cogit varBaseAddress - (1 << 15)) < varAddress
	  and: [varAddress < (cogit varBaseAddress + (1 << 15))]]! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
isBigEndian
	^false! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'rmacnak 12/5/2015 11:14:58'!
isCallPrecedingReturnPC: mcpc
	"Assuming mcpc is a send return pc answer if the instruction before it is a call (not a CallFull)."
	"cogit disassembleFrom: mcpc - 8 to: mcpc."

	(self opcodeAtAddress: mcpc - 8) = JAL ifTrue: [^true].
	
	((self opcodeAtAddress: mcpc - 8) = SPECIAL
		and: [(self functionAtAddress: mcpc - 8) = JALR]) ifTrue: [^true].
	
	^false! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'rmacnak 11/20/2015 17:38:38'!
isJump
	^super isJump or: [opcode between: BrEqualRR and: BrLongNotEqualRR]! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'rmacnak 12/16/2015 20:29'!
isJumpAt: pc
	"cogit disassembleFrom: pc to: pc + 4."

	(self opcodeAtAddress: pc) = J ifTrue: [^true].
	
	(self opcodeAtAddress: pc) = SPECIAL ifTrue: 
		[(self functionAtAddress: pc) = JR ifTrue: [^true]].

	(self opcodeAtAddress: pc) = BEQ ifTrue: [^true].
	(self opcodeAtAddress: pc) = BNE ifTrue: [^true].
	(self opcodeAtAddress: pc) = BLEZ ifTrue: [^true].
	(self opcodeAtAddress: pc) = BGTZ ifTrue: [^true].

	(self opcodeAtAddress: pc) = REGIMM ifTrue: 
		[(self rtAtAddress: pc) = BLTZ ifTrue: [^true].
		 (self rtAtAddress: pc) = BGEZ ifTrue: [^true]].	
	
	^false! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'rmacnak 11/9/2015 19:26:45'!
isPCDependent
	"Answer if the receiver is a pc-dependent instruction."
	^self isJump or: [opcode = AlignmentNops]! !

!CogMIPSELCompiler methodsFor: 'accessing' stamp: 'rmacnak 11/14/2015 14:50:01'!
isShortOffset: offset
	^offset between: -16r8000 and: 16r7FFF! !

!CogMIPSELCompiler methodsFor: 'encoding' stamp: 'rmacnak 12/17/2015 20:14'!
itype: op rs: rs rt: rt eitherImmediate: immediate	
	self assert: (op between: 0 and: 63).
	self assert: (rs between: 0 and: 31).
	self assert: (rt between: 0 and: 31).
	"signedImmediate < 0
		ifTrue: [unsignedImmediate := signedImmediate + 16r10000]
		ifFalse: [unsignedImmediate := signedImmediate].
	self assert: (unsignedImmediate between: 0 and: 16rFFFF)."
	^(((op << 26) bitOr: (rs << 21)) bitOr: (rt << 16)) bitOr: (immediate bitAnd: 16rFFFF)! !

!CogMIPSELCompiler methodsFor: 'encoding' stamp: 'eem 2/4/2020 11:08'!
itype: op rs: rs rt: rt signedImmediate: immediate	
	self assert: (op between: 0 and: 63).
	self assert: (rs between: 0 and: 31).
	self assert: (rt between: 0 and: 31).
	self assert: (self isShortOffset: immediate).
	"signedImmediate < 0
		ifTrue: [unsignedImmediate := signedImmediate + 16r10000]
		ifFalse: [unsignedImmediate := signedImmediate].
	self assert: (unsignedImmediate between: 0 and: 16rFFFF)."
	^(((op << 26) bitOr: (rs << 21)) bitOr: (rt << 16)) bitOr: (immediate bitAnd: 16rFFFF)! !

!CogMIPSELCompiler methodsFor: 'encoding' stamp: 'rmacnak 12/14/2015 19:47'!
itype: op rs: rs rt: rt unsignedImmediate: immediate	
	self assert: (op between: 0 and: 63).
	self assert: (rs between: 0 and: 31).
	self assert: (rt between: 0 and: 31).
	self assert: (immediate between: 0 and: 16rFFFF).
	^(((op << 26) bitOr: (rs << 21)) bitOr: (rt << 16)) bitOr: immediate! !

!CogMIPSELCompiler methodsFor: 'encoding - control' stamp: 'rmacnak 11/6/2015 21:41:25'!
jA: target
	self assert: (target bitAnd: 3) = 0.
	^self jtype: J target: (target bitAnd: 16r0FFFFFFF) >> 2! !

!CogMIPSELCompiler methodsFor: 'encoding - control' stamp: 'rmacnak 10/27/2015 20:57:02'!
jR: targetReg
	^self rtype: SPECIAL rs: targetReg rt: 0 rd: 0 sa: 0 funct: JR.! !

!CogMIPSELCompiler methodsFor: 'encoding - control' stamp: 'rmacnak 11/6/2015 21:41:43'!
jalA: target
	self assert: (target bitAnd: 3) = 0.
	^self jtype: JAL target: (target bitAnd: 16r0FFFFFFF) >> 2! !

!CogMIPSELCompiler methodsFor: 'encoding - control' stamp: 'rmacnak 11/7/2015 12:17:45'!
jalR: targetReg
	^self rtype: SPECIAL rs: targetReg rt: 0 rd: RA sa: 0 funct: JALR.! !

!CogMIPSELCompiler methodsFor: 'encoding' stamp: 'rmacnak 12/14/2015 19:48'!
jtype: op target: target
	self assert: (op between: 0 and: 63).
	self assert: (target between: 0 and: 16r7FFFFFF).
	^(op << 26) bitOr: target! !

!CogMIPSELCompiler methodsFor: 'accessing' stamp: 'rmacnak 11/10/2015 22:40:59'!
jumpLongByteSize
	self flag: #bogus. "Caller seems to actually want jumpFullByteSize"
	^16! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 11/10/2015 22:59:42'!
jumpLongConditionalByteSize
	^16! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 11/29/2015 20:02:17'!
jumpLongConditionalTargetBeforeFollowingAddress: mcpc
	"mcpc - 16:	beq/ne Cmp, ZR, +12
	 mcpc - 12:	nop (delay slot)
	 mcpc - 8:	j psuedo-address
	 mcpc - 4:	nop (delay slot)"		
	self assert: (((self opcodeAtAddress: mcpc - 16) == BEQ) 
				or: [(self opcodeAtAddress: mcpc - 16) == BNE]).
	self assert: (objectMemory longAt: mcpc - 12) == self nop. "Delay slot"
	self assert: (self opcodeAtAddress: mcpc - 8) == J.
	self assert: (objectMemory longAt: mcpc - 4) == self nop. "Delay slot"
	^self targetFromJTypeAtAddress: mcpc - 8! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 11/14/2015 15:44:35'!
jumpLongTargetBeforeFollowingAddress: mcpc 
	"Answer the target address for the long jump immediately preceding mcpc"
	self assert: (objectMemory longAt: mcpc - 4) == self nop. "Delay slot"
	self assert: (self opcodeAtAddress: mcpc - 8) == SPECIAL.
	self assert: (self functionAtAddress: mcpc - 8) == JR.
	^self literalAtAddress: mcpc - 12
	! !

!CogMIPSELCompiler methodsFor: 'accessing' stamp: 'rmacnak 11/17/2015 18:05:56'!
jumpShortByteSize
	^8! !

!CogMIPSELCompiler methodsFor: 'disassembly' stamp: 'rmacnak 12/2/2015 18:58:01'!
jumpTargetPCAt: pc
	<returnTypeC: #usqInt>
	"cogit disassembleFrom: pc to: pc + 4."
	
	(self opcodeAtAddress: pc) = J ifTrue: [^self targetFromJTypeAtAddress: pc].

	(self opcodeAtAddress: pc) = BEQ ifTrue: [^self targetFromITypeAtAddress: pc].
	(self opcodeAtAddress: pc) = BNE ifTrue: [^self targetFromITypeAtAddress: pc].
	(self opcodeAtAddress: pc) = BLEZ ifTrue: [^self targetFromITypeAtAddress: pc].
	(self opcodeAtAddress: pc) = BGTZ ifTrue: [^self targetFromITypeAtAddress: pc].
	(self opcodeAtAddress: pc) = REGIMM ifTrue:
		[(self rtAtAddress: pc) = BLTZ ifTrue: [^self targetFromITypeAtAddress: pc].
		 (self rtAtAddress: pc) = BGEZ ifTrue: [^self targetFromITypeAtAddress: pc]].
	
	self unreachable.! !

!CogMIPSELCompiler methodsFor: 'encoding - memory' stamp: 'rmacnak 11/8/2015 15:35:03'!
lbR: destReg base: baseReg offset: offset
	^self itype: LB rs: baseReg rt: destReg signedImmediate: offset! !

!CogMIPSELCompiler methodsFor: 'encoding - memory' stamp: 'rmacnak 11/8/2015 15:35:06'!
lbuR: destReg base: baseReg offset: offset
	^self itype: LBU rs: baseReg rt: destReg signedImmediate: offset! !

!CogMIPSELCompiler methodsFor: 'abi' stamp: 'eem 6/1/2020 17:20'!
leafCallStackPointerDelta
	"Answer the delta from the stack pointer after a call to the stack pointer
	 immediately prior to the call.  This is used to compute the stack pointer
	 immediately prior to a call from within a leaf routine, which in turn is used
	 to capture the c stack pointer to use in trampolines back into the C run-time."
	"This might actually be false, since directly after a call, lr, fp and variable registers need be pushed onto the stack. It depends on the implementation of call."
	^0! !

!CogMIPSELCompiler methodsFor: 'encoding - memory' stamp: 'rmacnak 11/8/2015 15:35:09'!
lhR: destReg base: baseReg offset: offset
	^self itype: LH rs: baseReg rt: destReg signedImmediate: offset! !

!CogMIPSELCompiler methodsFor: 'encoding - memory' stamp: 'rmacnak 11/8/2015 15:35:13'!
lhuR: destReg base: baseReg offset: offset
	^self itype: LHU rs: baseReg rt: destReg signedImmediate: offset! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/17/2015 20:27'!
literalAtAddress: mcpc
	<var: #high type: #usqInt>
	<var: #low type: #usqInt>
	| high low |
	self assert: (self opcodeAtAddress: mcpc) = ORI.
	self assert: (self opcodeAtAddress: mcpc - 4) = LUI.
	low := (objectMemory longAt: mcpc) bitAnd: 16rFFFF.
	high := (objectMemory longAt: mcpc - 4) bitAnd: 16rFFFF.
	^high << 16 bitOr: low
! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/17/2015 20:28'!
literalAtAddress: mcpc put: newLiteral
	<var: #oldUpper type: #usqInt>
	<var: #newUpper type: #usqInt>
	<var: #oldLower type: #usqInt>
	<var: #newLower type: #usqInt>
	| oldUpper newUpper oldLower newLower |
	self assert: (self opcodeAtAddress: mcpc - 4) = LUI.
	self assert: (self opcodeAtAddress: mcpc) = ORI.
	
	oldUpper := objectMemory longAt: mcpc - 4.
	newUpper := (oldUpper bitAnd: 16rFFFF0000) bitOr: (self high16BitsOf: newLiteral).
	objectMemory longAt: mcpc - 4 put: newUpper.
	
	oldLower := objectMemory longAt: mcpc.
	newLower := (oldLower bitAnd: 16rFFFF0000) bitOr: (self low16BitsOf: newLiteral).
	objectMemory longAt: mcpc put: newLower.
	
	self assert: (self opcodeAtAddress: mcpc - 4) = LUI.
	self assert: (self opcodeAtAddress: mcpc) = ORI.
	
	self assert: (self literalAtAddress: mcpc) = newLiteral.
	
	^newLiteral! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/12/2015 14:10:32'!
literalBeforeFollowingAddress: followingAddress
	"Answer the literal embedded in the instruction immediately preceding followingAddress.
	 This is used in the MoveCwR, PushCw and CmpCwR cases."
	
	"Cmp/MoveCwR
	 pc-8	lui rx, uper
	 pc-4	ori rx, rx, lower"
	(self opcodeAtAddress: followingAddress - 4) = ORI ifTrue:
		[^self literalAtAddress: followingAddress - 4].
	
	"PushCw
	 pc-16	lui at, upper
	 pc-12	ori at, at, lower
	 pc-8	addiu sp, sp, -4
	 pc-4	sw at, 0(sp)"
	((self opcodeAtAddress: followingAddress - 4) = SW and:
		[(self opcodeAtAddress: followingAddress - 8) = ADDIU]) ifTrue:
			[^self literalAtAddress: followingAddress - 12].
	
	self unreachable.
	^0! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'rmacnak 11/9/2015 19:50:50'!
literalLoadInstructionBytes
	"Answer the size of a literal load instruction (which may or may not include the size of the literal).
	 This differs between in-line and out-of-line literal generation."
	<inline: true>
	^8! !

!CogMIPSELCompiler methodsFor: 'accessing' stamp: 'rmacnak 11/11/2015 19:36:06'!
loadLiteralByteSize
	^8! !

!CogMIPSELCompiler methodsFor: 'accessing' stamp: 'rmacnak 11/17/2015 17:56:23'!
loadPICLiteralByteSize
	"Answer the byte size of a MoveCwR opcode's corresponding machine code
	 when the argument is a PIC.  This is for the self-reference at the end of a
	 closed PIC."
	<inline: true>
	^self loadLiteralByteSize! !

!CogMIPSELCompiler methodsFor: 'encoding' stamp: 'rmacnak 12/17/2015 20:27'!
low16BitsOf: word
	<var: #word type: #usqInt>
	^word bitAnd: 16rFFFF! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/8/2015 15:39:27'!
luiR: destReg C: imm
	^self itype: LUI rs: 0 rt: destReg eitherImmediate: imm! !

!CogMIPSELCompiler methodsFor: 'encoding - memory' stamp: 'rmacnak 11/8/2015 15:35:19'!
lwR: destReg base: baseReg offset: offset
	^self itype: LW rs: baseReg rt: destReg signedImmediate: offset! !

!CogMIPSELCompiler methodsFor: 'accessing' stamp: 'rmacnak 11/10/2015 19:04:37'!
machineCodeAt: anOffset
	"read aWord from machineCode, with little endian"
	<inline: true>
	^machineCode at: anOffset // 4! !

!CogMIPSELCompiler methodsFor: 'accessing' stamp: 'rmacnak 11/9/2015 19:32:47'!
machineCodeAt: anOffset put: aWord
	"add aWord to machineCode, with little endian"
	<inline: true>
	machineCode at: anOffset // 4 put: aWord! !

!CogMIPSELCompiler methodsFor: 'generate machine code' stamp: 'rmacnak 11/14/2015 14:55:59'!
machineCodeBytes
	^self machineCodeWords * 4! !

!CogMIPSELCompiler methodsFor: 'generate machine code' stamp: 'rmacnak 11/30/2015 19:36:03'!
machineCodeWords
	"Answer the maximum number of words of machine code generated for any abstract instruction.
	 e.g. AddCheckOverflowCqR"
	^7! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/21/2015 18:16:38'!
mfhiR: destReg
	self flag: #todo. "Ascertain what ISA revision adds this instruction"
	^self rtype: SPECIAL rs: 0 rt: 0 rd: destReg sa: 0 funct: MFHI! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/21/2015 18:16:31'!
mfloR: destReg
	self flag: #todo. "Ascertain what ISA revision adds this instruction"
	^self rtype: SPECIAL rs: 0 rt: 0 rd: destReg sa: 0 funct: MFLO! !

!CogMIPSELCompiler methodsFor: 'encoding - control' stamp: 'rmacnak 12/16/2015 19:58'!
mipsbreak: code
	self assert: (code between: 0 and: 16rFFFFF).
	^(code << 6) bitOr: BREAK! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/21/2015 18:11:00'!
multR: leftReg R: rightReg
	self flag: #todo. "Ascertain what ISA revision adds this instruction"
	^self rtype: SPECIAL rs: leftReg rt: rightReg rd: 0 sa: 0 funct: MULT! !

!CogMIPSELCompiler methodsFor: 'printing' stamp: 'eem 5/30/2016 13:09'!
nameForRegister: reg "<Integer>"
	<doNotGenerate>
	^MIPSConstants nameForRegister: reg! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 10/27/2015 22:22:12'!
nop
	^0 "SLL ZR, ZR, 0"! !

!CogMIPSELCompiler methodsFor: 'abstract instructions' stamp: 'eem 1/23/2020 16:15'!
noteFollowingConditionalBranch: branch
	"Support for processors without condition codes, such as the MIPS.
	 Answer the branch opcode.  Modify the receiver and the branch to
	 implement a suitable conditional branch that doesn't depend on
	 condition codes being set by the receiver."
	<returnTypeC: #'AbstractInstruction *'>
	<var: #branch type: #'AbstractInstruction *'>
	| newBranchLeft newBranchOpcode newBranchRight |
	
	((branch opcode = JumpOverflow) or: [branch opcode = JumpNoOverflow]) 
		ifTrue: [^self noteFollowingOverflowBranch: branch].

	newBranchOpcode := branch opcode caseOf: {
		[JumpZero] 			-> [BrEqualRR].
		[JumpNonZero]			-> [BrNotEqualRR].
		[JumpBelow]			-> [BrUnsignedLessRR].
		[JumpBelowOrEqual]	-> [BrUnsignedLessEqualRR].
		[JumpAbove]			-> [BrUnsignedGreaterRR].
		[JumpAboveOrEqual]	-> [BrUnsignedGreaterEqualRR].
		[JumpLess]				-> [BrSignedLessRR].
		[JumpLessOrEqual]		-> [BrSignedLessEqualRR].
		[JumpGreater]			-> [BrSignedGreaterRR].
		[JumpGreaterOrEqual]	-> [BrSignedGreaterEqualRR].
		[JumpLongZero] 		-> [BrLongEqualRR].
		[JumpLongNonZero]	-> [BrLongNotEqualRR].
		
		[JumpNegative]			-> [BrSignedLessRR].
	} otherwise: [self unreachable. 0].
	
	opcode caseOf: {
		[BrEqualRR]	->	["I.e., two jumps after a compare."
						newBranchLeft := operands at: 1.
						newBranchRight := operands at: 2].
		[BrUnsignedLessRR]	->	["I.e., two jumps after a compare."
						newBranchLeft := operands at: 1.
						newBranchRight := operands at: 2].

		[CmpRR] 	-> 	[newBranchLeft := operands at: 1.
						 newBranchRight := operands at: 0.
						 opcode := Label].
		[CmpCqR]	-> 	[newBranchLeft := operands at: 1.
						 newBranchRight := AT.
						 opcode := MoveCqR.
						 operands at: 1 put: AT].
		[CmpCwR]	-> 	[newBranchLeft := operands at: 1.
						 newBranchRight := AT.
						 opcode := MoveCwR.
						 operands at: 1 put: AT].
		[TstCqR]	->	[newBranchLeft := Cmp.
						 newBranchRight := ZR].
		[AndCqR]	->	[newBranchLeft := operands at: 1.
						 newBranchRight := ZR].
		[AndCqRR]	->	[newBranchLeft := operands at: 2.
						 newBranchRight := ZR].
		[OrCqRR]	->	[newBranchLeft := operands at: 2.
						 newBranchRight := ZR].
		[OrRR]	->		[newBranchLeft := operands at: 1.
						 newBranchRight := ZR].
		[XorRR]	->		[newBranchLeft := operands at: 1.
						 newBranchRight := ZR].
		[SubCwR]	->	[newBranchLeft := operands at: 1.
						 newBranchRight := ZR].
		[SubCqR]	->	[newBranchLeft := operands at: 1.
						 newBranchRight := ZR].
		[ArithmeticShiftRightCqR]	->	[newBranchLeft := operands at: 1.
						 newBranchRight := ZR].
		[ClzRR]	->	[newBranchLeft := operands at: 0. "we test if the destination register is zero"
						 newBranchRight := ZR].
	} otherwise: [self unreachable].

	branch rewriteOpcode: newBranchOpcode with: newBranchLeft with: newBranchRight.
	^branch! !

!CogMIPSELCompiler methodsFor: 'abstract instructions' stamp: 'rmacnak 12/16/2015 19:49'!
noteFollowingOverflowBranch: branch
	"Support for processors without condition codes, such as the MIPS.
	 Answer the branch opcode.  Modify the receiver and the branch to
	 implement a suitable conditional branch that doesn't depend on
	 condition codes being set by the receiver."
	<var: #branch type: #'AbstractInstruction *'>
	| newBranchOpcode |
	
	(opcode = MulRR) ifTrue:
		[opcode := MulCheckOverflowRR.
		 newBranchOpcode := branch opcode caseOf: {
			[JumpOverflow]		-> [BrNotEqualRR].
			[JumpNoOverflow]	-> [BrEqualRR].
		 } otherwise: [self unreachable. 0].
		 branch rewriteOpcode: newBranchOpcode with: OverflowTemp1 with: OverflowTemp2.
		 ^branch].


	opcode := opcode caseOf: {
		[AddCqR]	-> [AddCheckOverflowCqR].
		[AddRR]		-> [AddCheckOverflowRR].
		[SubCqR]	-> [SubCheckOverflowCqR].
		[SubRR]		-> [SubCheckOverflowRR].
	} otherwise: [self unreachable. 0].

	newBranchOpcode := branch opcode caseOf: {
		[JumpOverflow]		-> [BrSignedLessRR].
		[JumpNoOverflow]	-> [BrSignedGreaterEqualRR].
	} otherwise: [self unreachable. 0].
	branch rewriteOpcode: newBranchOpcode with: Overflow with: ZR.
	^branch! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/12/2015 11:44:45'!
nsSendCacheAt: callSiteReturnAddress
	"Answer the NSSendCache for the return address of a Newspeak
	 self, super, outer, or implicit receiver send."
	<option: #NewspeakVM>
	
	"ra - 24 lui s4, cacheHigh
	 ra - 20 ori s4, s4, cacheLow
	 ra - 16 lui t9, stubHigh
	 ra - 12 ori t9, t9, stubLow
	 ra - 8 jalr t9
	 ra - 4 nop (delay slot)"
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 24) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 20) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 16) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 12) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 8) = SPECIAL.
	self assert: (self functionAtAddress: callSiteReturnAddress - 8) = JALR.
	self assert: (objectMemory longAt: callSiteReturnAddress - 4) = self nop.
	"cogit disassembleFrom: callSiteReturnAddress - 16 to: callSiteReturnAddress."
	^self literalAtAddress: callSiteReturnAddress - 20! !

!CogMIPSELCompiler methodsFor: 'abi' stamp: 'rmacnak 11/9/2015 20:07:19'!
numIntRegArgs
	self flag: #OABI.
	^4! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 11/11/2015 19:28:13'!
opcodeAtAddress: mcpc
	^(objectMemory longAt: mcpc) >> 26
! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/5/2015 20:39:03'!
orR: destReg R: leftReg R: rightReg
	^self rtype: SPECIAL rs: leftReg rt: rightReg rd: destReg sa: 0 funct: OR! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/8/2015 15:40:00'!
oriR: destReg R: srcReg C: imm
	^self itype: ORI rs: srcReg rt: destReg eitherImmediate: imm! !

!CogMIPSELCompiler methodsFor: 'generate machine code' stamp: 'rmacnak 11/9/2015 19:44:55'!
outputMachineCodeAt: targetAddress
	"Override to move machine code a word at a time."
	<inline: true>
	0 to: machineCodeSize - 1 by: 4 do:
		[:j|
		objectMemory longAt: targetAddress + j put: (machineCode at: j // 4)]! !

!CogMIPSELCompiler methodsFor: 'generate machine code' stamp: 'rmacnak 12/3/2015 19:01:45'!
padIfPossibleWithStopsFrom: startAddr to: endAddr
	startAddr to: endAddr - 1 by: 4 do: 
		[:addr | objectMemory longAt: addr put: self stop].! !

!CogMIPSELCompiler methodsFor: 'encoding - memory' stamp: 'rmacnak 11/17/2015 19:11:12'!
prefR: baseReg offset: offset hint: hint
	self flag: #todo. "Should we generate a nop instead? gcc gives an error trying to assemble this for mips1 (1985), mips2 (1990) and mips3 (1992), but succeeds with mips4 (1994) and mips32 (1999)."
	self assert: (hint == HintLoad or: [hint == HintStore]).
	^self itype: PREF rs: baseReg rt: hint signedImmediate: offset! !

!CogMIPSELCompiler methodsFor: 'accessing' stamp: 'rmacnak 11/10/2015 22:32:50'!
pushLinkRegisterByteSize
	^8! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/5/2015 13:48:12'!
relocateCallBeforeReturnPC: retpc by: delta
	| target |
	self assert: delta \\ 4 = 0.
	delta = 0 ifTrue: [^self].
	
	self assert: (self opcodeAtAddress: retpc - 16) = LUI.
	self assert: (self opcodeAtAddress: retpc - 12) = ORI.
	self assert: (self opcodeAtAddress: retpc - 8) = SPECIAL.
	self assert: (self functionAtAddress: retpc - 8) = JALR.
	self assert: (objectMemory longAt: retpc - 4) = self nop.
	"cogit disassembleFrom: retpc - 16 to: retpc."

	target := self literalAtAddress: retpc - 12.
	target := target + delta.
	self literalAtAddress: retpc - 12 put: target.

	self assert: (self opcodeAtAddress: retpc - 16) = LUI.
	self assert: (self opcodeAtAddress: retpc - 12) = ORI.
	self assert: (self opcodeAtAddress: retpc - 8) = SPECIAL.
	self assert: (self functionAtAddress: retpc - 8) = JALR.
	self assert: (objectMemory longAt: retpc - 4) = self nop.
	"cogit disassembleFrom: retpc - 16 to: retpc."! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/9/2015 22:58:38'!
relocateJumpLongBeforeFollowingAddress: pc by: delta	
	"lui t9, stub/targetHigh
	 ori t9, t9, stub/targetLow
	 jr t9
	 nop (delay slot)
	 ...  <-- pc"

	| oldTarget newTarget |
	self assert: delta \\ 4 = 0.
	delta = 0 ifTrue: [^self].
	
	self assert: (self opcodeAtAddress: pc - 16) = LUI.
	self assert: (self opcodeAtAddress: pc - 12) = ORI.
	self assert: (self opcodeAtAddress: pc - 8) = SPECIAL.
	self assert: (self functionAtAddress: pc - 8) = JR.
	self assert: (objectMemory longAt: pc - 4) = self nop.
	"cogit disassembleFrom: pc - 16 to: pc."

	oldTarget := self literalAtAddress: pc - 12.
	newTarget := oldTarget + delta.
	self literalAtAddress: pc - 12 put: newTarget.

	self assert: (self opcodeAtAddress: pc - 16) = LUI.
	self assert: (self opcodeAtAddress: pc - 12) = ORI.
	self assert: (self opcodeAtAddress: pc - 8) = SPECIAL.
	self assert: (self functionAtAddress: pc - 8) = JR.
	self assert: (objectMemory longAt: pc - 4) = self nop.
	"cogit disassembleFrom: pc - 16 to: pc."! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/1/2015 20:14:10'!
relocateJumpLongConditionalBeforeFollowingAddress: pc by: delta	
	"lui t9, stub/targetHigh
	 ori t9, t9, stub/targetLow
	 jalr t9
	 nop (delay slot)
	 ...  <-- callSiteReturnAddress"

	self assert: (self opcodeAtAddress: pc - 16) = BNE.
	self assert: (objectMemory longAt: pc - 12) = self nop.
	self assert: (self opcodeAtAddress: pc - 8) = J.
	self assert: (objectMemory longAt: pc - 4) = self nop.
	"cogit disassembleFrom: pc - 16 to: pc."
	
	self rewriteJTypeAtAddress: pc - 8 delta: delta.

	self assert: (self opcodeAtAddress: pc - 16) = BNE.
	self assert: (objectMemory longAt: pc - 12) = self nop.
	self assert: (self opcodeAtAddress: pc - 8) = J.
	self assert: (objectMemory longAt: pc - 4) = self nop.
	"cogit disassembleFrom: pc - 16 to: pc."! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'eem 12/7/2016 10:26'!
relocateMethodReferenceBeforeAddress: pc by: delta
	| oldValue newValue |
	"cogit disassembleFrom: pc - 16 to: pc + 16 a StackToRegisterMappingCogit."
	
	((self opcodeAtAddress: pc - 8) = ADDIU and: [(self opcodeAtAddress: pc - 4) = SW]) ifTrue:
		["PushCw"
		oldValue := self literalAtAddress: pc - 12.
		newValue := oldValue + delta.
		self literalAtAddress: pc - 12 put: newValue.	
		self assert: (self literalAtAddress: pc - 12) = newValue.
		^self].

	"MoveCwR"
	oldValue := self literalAtAddress: pc - 4.
	newValue := oldValue + delta.
	self literalAtAddress: pc - 4 put: newValue.
	
	"cogit disassembleFrom: pc - 8 to: pc."
	self assert: (self literalAtAddress: pc - 4) = newValue.
	! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'eem 3/10/2017 17:20'!
rewriteCPICJumpAt: addressFollowingJump target: jumpTargetAddr
	"Rewrite a jump instruction to call a different target.  This variant is used to reset the 
	jumps in the prototype CPIC to suit each use,.   
	Answer the extent of the code change which is used to compute the range of the icache to flush."
	<var: #addressFollowingJump type: #usqInt>
	<var: #jumpTargetAddr type: #usqInt>
	
	"self CmpR: ClassReg R: TempReg.
	^self JumpNonZero: 0"
	
	"bne s5, s3, +156 ; =BE7C
	 nop (delay slot)
	 .... <-- addressFollowingJump"
	
	self assert: (self opcodeAtAddress: addressFollowingJump - 8) = BNE.
	self assert: (objectMemory longAt: addressFollowingJump - 4) = self nop.
	"cogit disassembleFrom: addressFollowingJump - 8 to: addressFollowingJump."
	
	self rewriteITypeBranchAtAddress: addressFollowingJump - 8 target: jumpTargetAddr.
	
	self assert: (self opcodeAtAddress: addressFollowingJump - 8) = BNE.
	self assert: (objectMemory longAt: addressFollowingJump - 4) = self nop.
	"cogit disassembleFrom: addressFollowingJump - 8 to: addressFollowingJump."! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 11/29/2015 20:02:57'!
rewriteCallAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a call instruction to call a different target.  This variant is used to link PICs
	 in ceSendMiss et al,.   
	Answer the extent of the code change which is used to compute the range of the icache to flush."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	
	"lui t9, stub/targetHigh
	 ori t9, t9, stub/targetLow
	 jalr t9
	 nop (delay slot)
	 ...  <-- callSiteReturnAddress"
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 16) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 12) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 8) = SPECIAL.
	self assert: (self functionAtAddress: callSiteReturnAddress - 8) = JALR.
	self assert: (objectMemory longAt: callSiteReturnAddress - 4) = self nop.
	"cogit disassembleFrom: callSiteReturnAddress - 16 to: callSiteReturnAddress."

	self literalAtAddress: callSiteReturnAddress - 12 put: callTargetAddress.

	self assert: (self opcodeAtAddress: callSiteReturnAddress - 16) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 12) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 8) = SPECIAL.
	self assert: (self functionAtAddress: callSiteReturnAddress - 8) = JALR.
	self assert: (objectMemory longAt: callSiteReturnAddress - 4) = self nop.
	"cogit disassembleFrom: callSiteReturnAddress - 16 to: callSiteReturnAddress."

	^20! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 11/29/2015 20:03:06'!
rewriteConditionalJumpLongAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a jump instruction to call a different target.  This variant is used to reset the 
	jumps in the prototype CPIC to suit each use,.   
	Answer the extent of the code change which is used to compute the range of the icache to flush."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 8) = J.
	self assert: (objectMemory longAt: callSiteReturnAddress - 4) = self nop.
	"cogit disassembleFrom: callSiteReturnAddress - 8 to: callSiteReturnAddress."

	self rewriteJTypeAtAddress: callSiteReturnAddress - 8 target: callTargetAddress.

	self assert: (self opcodeAtAddress: callSiteReturnAddress - 8) = J.
	self assert: (objectMemory longAt: callSiteReturnAddress - 4) = self nop.
	"cogit disassembleFrom: callSiteReturnAddress - 8 to: callSiteReturnAddress."
! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/20/2015 14:01'!
rewriteITypeBranchAtAddress: mcpc target: newTarget
	| newDisplacement oldInstruction newInstruction |
	newDisplacement := newTarget - (mcpc + 4). "Displacement is relative to delay slot."
	newDisplacement := newDisplacement >>> 2. "Displacement is in words."
	newDisplacement := (newDisplacement bitAnd: 16rFFFF).

	oldInstruction := objectMemory longAt: mcpc.
	newInstruction := (oldInstruction bitAnd: 16rFFFF0000) bitOr: newDisplacement.
	
	objectMemory longAt: mcpc put: newInstruction.! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/2/2015 19:36:31'!
rewriteInlineCacheAt: callSiteReturnAddress tag: cacheTag target: callTargetAddress
	"Rewrite an inline cache to call a different target for a new tag.  This variant is used
	 to link unlinked sends in ceSend:to:numArgs: et al.  Answer the extent of the code
	 change which is used to compute the range of the icache to flush."
	
	"MoveCwR ClassReg selectorIndex/expectedClass
	 Call: unlinked send stub/expectedTarget
	 Push ReceiverResult <-- callSiteReturnAddress"
	
	"lui s3, selector/tagHigh
	 ori s3, s3, selector/tagLow
	 lui t9, stub/targetHigh
	 ori t9, t9, stub/targetLow
	 jalr t9
	 nop (delay slot)
	 ...  <-- callSiteReturnAddress"
	
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 24) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 20) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 16) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 12) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 8) = SPECIAL.
	self assert: (self functionAtAddress: callSiteReturnAddress - 8) = JALR.
	self assert: (objectMemory longAt: callSiteReturnAddress - 4) = self nop.
	"cogit disassembleFrom: callSiteReturnAddress - 24 to: callSiteReturnAddress."

	self literalAtAddress: callSiteReturnAddress - 20 put: cacheTag.
	self literalAtAddress: callSiteReturnAddress - 12 put: callTargetAddress.

	self assert: (self opcodeAtAddress: callSiteReturnAddress - 24) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 20) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 16) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 12) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 8) = SPECIAL.
	self assert: (self functionAtAddress: callSiteReturnAddress - 8) = JALR.
	self assert: (objectMemory longAt: callSiteReturnAddress - 4) = self nop.
	"cogit disassembleFrom: callSiteReturnAddress - 24 to: callSiteReturnAddress."

	^24! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/14/2015 19:50'!
rewriteInlineCacheTag: cacheTag at: callSiteReturnAddress
	"Rewrite an inline cache with a new tag.  This variant is used
	 by the garbage collector."
	
	"MoveCwR ClassReg selectorIndex/expectedClass
	 Call: unlinked send stub/expectedTarget
	 Push ReceiverResult <-- callSiteReturnAddress"
	
	"lui s3, selector/tagHigh
	 ori s3, s3, selector/tagLow
	 lui t9, stub/targetHigh
	 ori t9, t9, stub/targetLow
	 jalr t9
	 nop (delay slot)
	 ...  <-- callSiteReturnAddress"
	
	<var: #callSiteReturnAddress type: #usqInt>
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 24) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 20) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 16) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 12) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 8) = SPECIAL.
	self assert: (self functionAtAddress: callSiteReturnAddress - 8) = JALR.
	self assert: (objectMemory longAt: callSiteReturnAddress - 4) = self nop.
	"cogit disassembleFrom: callSiteReturnAddress - 24 to: callSiteReturnAddress."

	self literalAtAddress: callSiteReturnAddress - 20 put: cacheTag.

	self assert: (self opcodeAtAddress: callSiteReturnAddress - 24) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 20) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 16) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 12) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 8) = SPECIAL.
	self assert: (self functionAtAddress: callSiteReturnAddress - 8) = JALR.
	self assert: (objectMemory longAt: callSiteReturnAddress - 4) = self nop.
	"cogit disassembleFrom: callSiteReturnAddress - 24 to: callSiteReturnAddress."! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/9/2015 22:59:14'!
rewriteJTypeAtAddress: mcpc delta: delta
	| oldTarget newTarget |
	oldTarget := self targetFromJTypeAtAddress: mcpc.
	newTarget := oldTarget + delta.
	self rewriteJTypeAtAddress: mcpc target: newTarget.! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/1/2015 20:11:20'!
rewriteJTypeAtAddress: mcpc target: newTarget
	| regionMask |
	self assert: (self opcodeAtAddress: mcpc) = J.
	regionMask := 16rF0000000.
	"mcpc + 4: relative to delay slot not j"
	self assert: (mcpc + 4 bitAnd: regionMask) = (newTarget bitAnd: regionMask).
	objectMemory longAt: mcpc put: (self jA: newTarget).! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 11/29/2015 20:03:53'!
rewriteJumpLongAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a jump instruction to call a different target.  This variant is used to reset the 
	jumps in the prototype CPIC to suit each use,.   
	Answer the extent of the code change which is used to compute the range of the icache to flush."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	
	"lui t9, stub/targetHigh
	 ori t9, t9, stub/targetLow
	 jr t9
	 nop (delay slot)
	 ...  <-- callSiteReturnAddress"

	self assert: (self opcodeAtAddress: callSiteReturnAddress - 16) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 12) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 8) = SPECIAL.
	self assert: (self functionAtAddress: callSiteReturnAddress - 8) = JR.
	self assert: (objectMemory longAt: callSiteReturnAddress - 4) = self nop.
	"cogit disassembleFrom: callSiteReturnAddress - 16 to: callSiteReturnAddress."

	self literalAtAddress: callSiteReturnAddress - 12 put: callTargetAddress.

	self assert: (self opcodeAtAddress: callSiteReturnAddress - 16) = LUI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 12) = ORI.
	self assert: (self opcodeAtAddress: callSiteReturnAddress - 8) = SPECIAL.
	self assert: (self functionAtAddress: callSiteReturnAddress - 8) = JR.
	self assert: (objectMemory longAt: callSiteReturnAddress - 4) = self nop.
	"cogit disassembleFrom: callSiteReturnAddress - 16 to: callSiteReturnAddress."

	^20! !

!CogMIPSELCompiler methodsFor: 'abstract instructions' stamp: 'rmacnak 12/16/2015 20:28'!
rewriteOpcode: anOpcode with: left with: right
	<inline: true>
	opcode := anOpcode.
	operands
		"0 is target"
		at: 1 put: left;
		at: 2 put: right! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/16/2015 20:30'!
rtAtAddress: mcpc
	^(objectMemory longAt: mcpc) >> 16 bitAnd: 31
! !

!CogMIPSELCompiler methodsFor: 'encoding' stamp: 'rmacnak 12/14/2015 19:48'!
rtype: op rs: rs rt: rt rd: rd sa: sa funct: funct
	self assert: (op between: 0 and: 63).
	self assert: (rs between: 0 and: 31).
	self assert: (rt between: 0 and: 31).
	self assert: (rd between: 0 and: 31).
	self assert: (sa between: 0 and: 31).
	self assert: (funct between: 0 and: 63).
	^(((((op << 26) bitOr: (rs << 21)) bitOr: (rt << 16)) bitOr: (rd << 11)) bitOr: (sa << 6)) bitOr: funct! !

!CogMIPSELCompiler methodsFor: 'abi' stamp: 'rmacnak 11/9/2015 21:51:42'!
saveAndRestoreLinkRegAround: aBlock
	"If the processor's ABI includes a link register, generate instructions
	 to save and restore it around aBlock, which is assumed to generate code."
	<inline: true>
	| inst |
	inst := cogit PushR: LinkReg.
	aBlock value.
	cogit PopR: LinkReg.
	^inst! !

!CogMIPSELCompiler methodsFor: 'encoding - memory' stamp: 'rmacnak 11/8/2015 15:35:32'!
sbR: srcReg base: baseReg offset: offset
	^self itype: SB rs: baseReg rt: srcReg signedImmediate: offset! !

!CogMIPSELCompiler methodsFor: 'testing' stamp: 'nice 8/26/2019 15:00'!
setsConditionCodesFor: aConditionalJumpOpcode
	<inline: false>
	"Not really, but we can merge this in noteFollowingConditionalBranch:."
	opcode = XorRR ifTrue: [^true]. 
	opcode = ArithmeticShiftRightCqR ifTrue: [^true].
	(opcode = ClzRR and: [aConditionalJumpOpcode = JumpZero]) ifTrue: [^true].
	self unreachable.
	^false! !

!CogMIPSELCompiler methodsFor: 'encoding - memory' stamp: 'rmacnak 11/8/2015 15:35:38'!
shR: srcReg base: baseReg offset: offset
	^self itype: SH rs: baseReg rt: srcReg signedImmediate: offset! !

!CogMIPSELCompiler methodsFor: 'generate machine code' stamp: 'rmacnak 11/9/2015 21:15:01'!
sizePCDependentInstructionAt: eventualAbsoluteAddress
	"Size a jump and set its address.  The target may be another instruction
	 or an absolute address.  On entry the address inst var holds our virtual
	 address. On exit address is set to eventualAbsoluteAddress, which is
	 where this instruction will be output.  The span of a jump to a following
	 instruction is therefore between that instruction's address and this
	 instruction's address ((which are both still their virtual addresses), but the
	 span of a jump to a preceding instruction or to an absolute address is
	 between that instruction's address (which by now is its eventual absolute
	 address) or absolute address and eventualAbsoluteAddress.

	 ARM is simple; the 26-bit call/jump range means no short jumps.  This routine
	 only has to determine the targets of jumps, not determine sizes."

	opcode = AlignmentNops ifTrue:
		[| alignment |
		 address := eventualAbsoluteAddress.
		 alignment := operands at: 0.
		 ^machineCodeSize := (eventualAbsoluteAddress + (alignment - 1) bitAnd: alignment negated)
							   - eventualAbsoluteAddress].
	self assert: (self isJump or: [opcode = Call or: [opcode = CallFull]]).
	self isJump ifTrue: [self resolveJumpTarget].
	address := eventualAbsoluteAddress.
	^machineCodeSize := maxSize! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/6/2015 17:32:54'!
sllR: destReg R: sourceReg C: shiftAmount
	^self rtype: SPECIAL rs: 0 rt: sourceReg rd: destReg sa: shiftAmount funct: SLL! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/6/2015 18:01:19'!
sllvR: destReg R: leftReg R: rightReg
	^self rtype: SPECIAL rs: rightReg rt: leftReg rd: destReg sa: 0 funct: SLLV! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/8/2015 15:04:25'!
sltR: destReg R: leftReg R: rightReg
	^self rtype: SPECIAL rs: leftReg rt: rightReg rd: destReg sa: 0 funct: SLT! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/8/2015 15:35:44'!
sltiR: destReg R: leftReg C: imm
	^self itype: SLTI rs: leftReg rt: destReg signedImmediate: imm! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/8/2015 15:35:50'!
sltiuR: destReg R: leftReg C: imm
	^self itype: SLTIU rs: leftReg rt: destReg signedImmediate: imm! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/8/2015 15:04:08'!
sltuR: destReg R: leftReg R: rightReg
	^self rtype: SPECIAL rs: leftReg rt: rightReg rd: destReg sa: 0 funct: SLTU! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/6/2015 17:33:35'!
sraR: destReg R: sourceReg C: shiftAmount
	^self rtype: SPECIAL rs: 0 rt: sourceReg rd: destReg sa: shiftAmount funct: SRA! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/6/2015 18:01:31'!
sravR: destReg R: leftReg R: rightReg
	^self rtype: SPECIAL rs: rightReg rt: leftReg rd: destReg sa: 0 funct: SRAV! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/6/2015 17:33:39'!
srlR: destReg R: sourceReg C: shiftAmount
	^self rtype: SPECIAL rs: 0 rt: sourceReg rd: destReg sa: shiftAmount funct: SRL! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/6/2015 18:01:27'!
srlvR: destReg R: leftReg R: rightReg
	^self rtype: SPECIAL rs: rightReg rt: leftReg rd: destReg sa: 0 funct: SRLV! !

!CogMIPSELCompiler methodsFor: 'encoding - control' stamp: 'rmacnak 12/16/2015 19:59'!
stop
	^self mipsbreak: 0! !

!CogMIPSELCompiler methodsFor: 'generate machine code - support' stamp: 'eem 1/27/2020 09:33'!
stopsFrom: startAddr to: endAddr
	self assert: endAddr - startAddr + 1 \\ 4 = 0.
	startAddr to: endAddr by: 4 do: 
		[:addr | cogit codeLongAt: addr put: self stop].! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'eem 12/7/2016 10:27'!
storeLiteral: literal beforeFollowingAddress: followingAddress
	"Rewrite the long constant loaded by a MoveCwR or PushCw before the given address"
	self flag: #bogus. "The caller ought to know what it is patching, and this should be split into separate methods with stricter checking. rmacnak 12/13/2015"
	
	"Cmp/MoveCwR
	 pc-8	lui rx, uper
	 pc-4	ori rx, rx, lower"
	(self opcodeAtAddress: followingAddress - 4) = ORI ifTrue:
		[^self literalAtAddress: followingAddress - 4 put: literal].

	"PushCw
	 pc-16	lui at, upper
	 pc-12	ori at, at, lower
	 pc-8	addiu sp, sp, -4
	 pc-4	sw at, 0(sp)"
	((self opcodeAtAddress: followingAddress - 4) = SW and:
		[(self opcodeAtAddress: followingAddress - 8) = ADDIU]) ifTrue:
			[^self literalAtAddress: followingAddress - 12 put: literal].
	
	self unreachable.
	^0! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/5/2015 23:19:43'!
subuR: destReg R: leftReg R: rightReg
	^self rtype: SPECIAL rs: leftReg rt: rightReg rd: destReg sa: 0 funct: SUBU! !

!CogMIPSELCompiler methodsFor: 'encoding - memory' stamp: 'rmacnak 11/8/2015 15:35:57'!
swR: srcReg base: baseReg offset: offset
	^self itype: SW rs: baseReg rt: srcReg signedImmediate: offset! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/26/2015 11:34'!
targetFromITypeAtAddress: mcpc
	<var: #offset type: #usqInt>
	<var: #mcpc type: #usqInt>
	<returnTypeC: #usqInt>
	| offset |
	offset := (objectMemory longAt: mcpc) bitAnd: 16rFFFF.
	self cCode: '' inSmalltalk: [offset >= 16r8000 ifTrue: [offset := offset - 16r10000]].
	offset := offset << 2.
	^mcpc + offset + OneInstruction. "Offset is relative to the delay slot"! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/22/2015 14:20'!
targetFromJTypeAtAddress: mcpc
	<var: #mcpc type: #usqInt>
	<returnTypeC: #usqInt>
	| targetLow |
	targetLow := (objectMemory longAt: mcpc) bitAnd: 16r3FFFFFF.
	"mcpc + 4: relative to delay slot not j"
	^((mcpc + 4) bitAnd: 16rF0000000) + (targetLow << 2) ! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/5/2015 23:06:12'!
xorR: destReg R: leftReg R: rightReg
	^self rtype: SPECIAL rs: leftReg rt: rightReg rd: destReg sa: 0 funct: XOR! !

!CogMIPSELCompiler methodsFor: 'encoding - arithmetic' stamp: 'rmacnak 11/8/2015 15:40:16'!
xoriR: destReg R: srcReg C: imm
	^self itype: XORI rs: srcReg rt: destReg eitherImmediate: imm! !

!CogMIPSELCompiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/9/2015 23:07:05'!
zoneCallsAreRelative
	"Answer if Call and JumpLong are relative and hence need to take the caller's
	 relocation delta into account during code compaction, rather than just the
	 callee's delta."
	^false! !

CogMIPSELCompiler subclass: #CogMIPSELCompilerForTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!CogMIPSELCompilerForTests class methodsFor: 'testing' stamp: 'rmacnak 11/24/2015 19:13:36'!
dataRegistersWithAccessorsDo: aTrinaryBlock
	#((16 s0 s0:)
		(17 s1 s1:)
		(18 s2 s2:)
		(19 s3 s3:)
		(20 s4 s4:)
		(21 s5 s5:)
		(22 s6 s6:)
		(29 sp sp:)
		(30 fp fp:)
		(31 ra ra:)) do: [:triple | aTrinaryBlock valueWithArguments: triple].! !

!CogMIPSELCompilerForTests class methodsFor: 'testing' stamp: 'eem 12/19/2019 00:19'!
dataRegistersWithAccessorsGiven: aProcessorSimulator do: aTrinaryBlock
	"We can't use RISCTempReg et al, because some opcodes may be encoded as multiple instructions using RISCTempReg."
	(0 to: 31) withIndexDo:
		[:reg :i|
		(reg = 0 or: [self isRISCTempRegister: reg]) ifFalse:
			[aTrinaryBlock
				value: reg
				value: (aProcessorSimulator registerStateGetters at: i)
				value: (aProcessorSimulator registerStateSetters at: i)]]! !

!CogMIPSELCompilerForTests class methodsFor: 'testing' stamp: 'eem 12/25/2015 18:47'!
isRISCTempRegister: reg
	"For tests to filter-out bogus values left in the RISCTempRegister, if any."
	^reg = AT
		or: [reg = Cmp
		or: [reg = Overflow
		or: [reg = BranchTemp]]]! !

CogAbstractInstruction subclass: #CogX64Compiler
	instanceVariableNames: ''
	classVariableNames: 'BSR CArg0Reg CArg1Reg CArg2Reg CArg3Reg CDQ CLD CMPXCHGRMr CPUID IDIVR IMULRR LFENCE LOCK MFENCE MOVSB MOVSQ ModReg ModRegInd ModRegIndDisp32 ModRegIndSIB ModRegRegDisp32 ModRegRegDisp8 MoveRAwNoVBR R10 R11 R12 R13 R14 R15 R8 R9 RAX RBP RBX RCX RDI RDX REP RSI RSP SETE SFENCE SIB1 SIB2 SIB4 SIB8 SysV XCHGRR XMM0L XMM10L XMM11L XMM12L XMM13L XMM14L XMM15L XMM1L XMM2L XMM3L XMM4L XMM5L XMM6L XMM7L XMM8L XMM9L'
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

CogX64Compiler class
	instanceVariableNames: 'cpuidWord1'!

!CogX64Compiler commentStamp: 'eem 9/14/2015 17:12' prior: 0!
I generate x64 (x86-64) instructions from CogAbstractInstructions.  For reference see
1. IA-32 Intel® Architecture Software Developer's Manual Volume 2A: Instruction Set Reference, A-M
2. IA-32 Intel® Architecture Software Developer's Manual Volume 2A: Instruction Set Reference, N-Z
	http://www.intel.com/products/processor/manuals/
or
AMD64 Architecture Programmer's Manual Volume 3: General-Purpose and System Instructions
AMD64 Architecture Programmer's Manual Volume 4: 128-bit Media Instructions
AMD64 Architecture Programmer's Manual Volume 5: 64-bit Media and x87 Floating Point Instructions
	http://developer.amd.com/resources/documentation-articles/developer-guides-manuals/
(® is supposed to be the Unicode "registered  sign").!

CogX64Compiler class
	instanceVariableNames: 'cpuidWord1'!

CogX64Compiler subclass: #CogInLineLiteralsX64Compiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogInLineLiteralsX64Compiler commentStamp: 'eem 9/26/2020 15:02' prior: 0!
A CogInLineLiteralsX64Compiler is the concrete subclass of CogX64Compiler that stores literals in-line in machine-code. This is the production code generator for x64/x86-64.

Instance Variables!

!CogInLineLiteralsX64Compiler class methodsFor: 'class initialization' stamp: 'eem 11/28/2016 18:07'!
initializeAbstractRegistersSysV
	"Assign the abstract registers with the identities/indices of the relevant concrete registers."

	super initializeAbstractRegistersSysV.
	RISCTempReg := R8! !

!CogInLineLiteralsX64Compiler class methodsFor: 'class initialization' stamp: 'eem 11/28/2016 18:07'!
initializeAbstractRegistersWin64
	"Assign the abstract registers with the identities/indices of the relevant concrete registers."

	super initializeAbstractRegistersWin64.
	RISCTempReg := R11! !

!CogInLineLiteralsX64Compiler class methodsFor: 'testing' stamp: 'eem 12/18/2019 22:42'!
isRISCTempRegister: reg
	"For tests to filter-out bogus values left in the RISCTempRegister, if any."
	^reg = R8 or: [reg = R11]! !

!CogInLineLiteralsX64Compiler methodsFor: 'generate machine code' stamp: 'eem 4/21/2016 16:11'!
computeSizeOfArithCqR
	"With CqR we assume constants are 32-bits or less."
	<inline: true>
	(self isQuick: (operands at: 0)) ifTrue:
		[^4].
	(self is32BitSignedImmediate: (operands at: 0)) ifTrue:
		[^(operands at: 1) = RAX ifTrue: [6] ifFalse: [7]].
	^10 "movabsq" + 3 "r op r"! !

!CogInLineLiteralsX64Compiler methodsFor: 'generate machine code' stamp: 'eem 9/22/2015 08:22'!
computeSizeOfArithCwR
	<inline: true>
	^10 "MoveCwR" +  3 "ArithRR"! !

!CogInLineLiteralsX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeArithCwR: x64opcode
	| value reg reverse |
	value := operands at: 0.
	reg := operands at: 1.
	reverse := x64opcode = 16r85 or: [x64opcode = 16r39]. "Tst & Cmp; backwards"
	machineCode
		at:  0 put: (self rexR: RISCTempReg x: 0 b: RISCTempReg);
		at:  1 put: 16rB8 + (RISCTempReg bitAnd: 7);
		at:  2 put: (value bitAnd: 16rFF);
		at:  3 put: (value >> 8 bitAnd: 16rFF);
		at:  4 put: (value >> 16 bitAnd: 16rFF);
		at:  5 put: (value >> 24 bitAnd: 16rFF);
		at:  6 put: (value >> 32 bitAnd: 16rFF);
		at:  7 put: (value >> 40 bitAnd: 16rFF);
		at:  8 put: (value >> 48 bitAnd: 16rFF);
		at:  9 put: (value >> 56 bitAnd: 16rFF);
		at: 10 put: (reverse
					ifTrue: [self rexR: RISCTempReg x: 0 b: reg]
					ifFalse: [self rexR: reg x: 0 b: RISCTempReg]);
		at: 11 put: x64opcode;
		at: 12 put: (reverse
					ifTrue: [self mod: ModReg RM: reg RO: RISCTempReg]
					ifFalse: [self mod: ModReg RM: RISCTempReg RO: reg]).
	self assert: (machineCode at: 12) > 16r90. "See literalBeforeFollowingAddress:"
	^13! !

!CogInLineLiteralsX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveCwR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value reg offset |
	value := operands at: 0.
	reg := operands at: 1.
	(self isAnInstruction: (cogit cCoerceSimple: value to: #'AbstractInstruction *')) ifTrue:
		[value := (cogit cCoerceSimple: value to: #'AbstractInstruction *') address].
	(cogit addressIsInCurrentCompilation: value) ifTrue:
		[offset := value - (address + 7).
		 machineCode
			at: 0 put: (self rexR: reg x: 0 b: 0);
			at: 1 put: 16r8D; "LoadEffectiveAddress"
			at: 2 put: (self mod: ModRegInd RM: 5 RO: reg);
			at: 3 put: (offset bitAnd: 16rFF);
			at: 4 put: (offset >> 8 bitAnd: 16rFF);
			at: 5 put: (offset >> 16 bitAnd: 16rFF);
			at: 6 put: (offset >> 24 bitAnd: 16rFF).
		^7].
	machineCode
		at:  0 put: (self rexR: 0 x: 0 b: reg);
		at:  1 put: 16rB8 + (reg bitAnd: 7);
		at:  2 put: (value bitAnd: 16rFF);
		at:  3 put: (value >> 8 bitAnd: 16rFF);
		at:  4 put: (value >> 16 bitAnd: 16rFF);
		at:  5 put: (value >> 24 bitAnd: 16rFF);
		at:  6 put: (value >> 32 bitAnd: 16rFF);
		at:  7 put: (value >> 40 bitAnd: 16rFF);
		at:  8 put: (value >> 48 bitAnd: 16rFF);
		at:  9 put: (value >> 56 bitAnd: 16rFF).
	"Add a nop to disambiguate between MoveCwR/PushCw and ArithCwR, which ends with a (self mod: ModReg RM: 0 RO: 0)"
	machineCode at: 10 put: 16r90.
	^11! !

!CogInLineLiteralsX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePushCw
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value offset |
	value := operands at: 0.
	(self isAnInstruction: (cogit cCoerceSimple: value to: #'AbstractInstruction *')) ifTrue:
		[value := (cogit cCoerceSimple: value to: #'AbstractInstruction *') address].
	(cogit addressIsInCurrentCompilation: value) ifTrue:
		[offset := value - (address + 7).
		 machineCode
			at: 0 put: (self rexR: RISCTempReg x: 0 b: 0);
			at: 1 put: 16r8D; "LoadEffectiveAddress"
			at: 2 put: (self mod: ModRegInd RM: 5 RO: RISCTempReg);
			at: 3 put: (offset bitAnd: 16rFF);
			at: 4 put: (offset >> 8 bitAnd: 16rFF);
			at: 5 put: (offset >> 16 bitAnd: 16rFF);
			at: 6 put: (offset >> 24 bitAnd: 16rFF);
			at: 7 put: 16r41;
			at: 8 put: 16r48 + RISCTempReg.
		^9].
	machineCode
		at:  0 put: (self rexR: RISCTempReg x: 0 b: RISCTempReg);
		at:  1 put: 16rB8 + (RISCTempReg bitAnd: 7);
		at:  2 put: (value bitAnd: 16rFF);
		at:  3 put: (value >> 8 bitAnd: 16rFF);
		at:  4 put: (value >> 16 bitAnd: 16rFF);
		at:  5 put: (value >> 24 bitAnd: 16rFF);
		at:  6 put: (value >> 32 bitAnd: 16rFF);
		at:  7 put: (value >> 40 bitAnd: 16rFF);
		at:  8 put: (value >> 48 bitAnd: 16rFF);
		at:  9 put: (value >> 56 bitAnd: 16rFF);
		at: 10 put: 16r41;
		at: 11 put: 16r48 + RISCTempReg. "The 48 will disambiguate between MoveCwR, PushCw and ArithCwR, which ends with a (self mod: ModReg RM: 0 RO: 0)"
	self assert: (machineCode at: 11) < 16r90. "see literalBeforeFollowingAddress:"
	^12! !

!CogInLineLiteralsX64Compiler methodsFor: 'accessing' stamp: 'eem 12/11/2015 17:48'!
getDefaultCogCodeSize
	"Answer the default number of bytes to allocate for native code at startup.
	 The actual value can be set via vmParameterAt: and/or a preference in the ini file."
	<inline: true>
	^1024 * 1400! !

!CogInLineLiteralsX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/13/2015 19:13'!
inlineCacheTagAt: callSiteReturnAddress
	"Answer the inline cache tag for the return address of a send."
	^self literal32BeforeFollowingAddress: callSiteReturnAddress - 5! !

!CogInLineLiteralsX64Compiler methodsFor: 'testing' stamp: 'eem 9/18/2015 09:12'!
isPCDependent
	"Answer if the receiver is a pc-dependent instruction."
	^self isJump or: [opcode = AlignmentNops]! !

!CogInLineLiteralsX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 2/23/2016 18:46'!
literal32BeforeFollowingAddress: followingAddress
	"Answer the 32-bit literal embedded in the instruction immediately preceding followingAddress."
	^cogit
		cCoerceSimple: (self unalignedLong32At: followingAddress - 4)
		to: #'unsigned int'! !

!CogInLineLiteralsX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/7/2016 10:58'!
literalBeforeFollowingAddress: followingAddress
	"Answer the literal embedded in the instruction immediately preceding followingAddress.
	 This is used in the MoveCwR, PushCw and ArithCwR cases; these are distinguished by a
	 nop following the literal load in MoveCwR, a 16r48 + reg ending the PushCw sequence, and
	 a (self mod: ModReg RM: rX RO: rY) ending the ArithCwR sequence, which is at least 16rC0."
	| lastByte base |
	lastByte := objectMemory byteAt: followingAddress - 1.
	base := followingAddress - (lastByte = 16r90
									ifTrue: [9]				"MoveCwR"
									ifFalse:
										[lastByte < 16r90
											ifTrue: [10]		"PushCw"
											ifFalse: [11]]).	"ArithCwR"
	^objectMemory unalignedLongAt: base
	
	"(Symbol allSymbols select: [:s| '*Cw:R:' match: s]), {#PushCw:} collect: [:s| {s. (self systemNavigation allCallsOn: s localToPackage: #VMMaker) size}]"! !

!CogInLineLiteralsX64Compiler methodsFor: 'accessing' stamp: 'eem 9/28/2015 15:45'!
loadLiteralByteSize
	<inline: true>
	^self moveCwRByteSize! !

!CogInLineLiteralsX64Compiler methodsFor: 'accessing' stamp: 'eem 9/28/2015 15:55'!
moveCwRByteSize
	"With in-line literals we use an 11 byte sequence for loading a 64-bit immediate,
	 which is one more than strictly necessary.  We plant a nop at the end of the
	 sequence to allow us to distinguish between this and the
	 (self mod: ModReg RM: rX RO: rY) at the end of an ArithCwR sequence."
	<inline: true>
	^11! !

!CogInLineLiteralsX64Compiler methodsFor: 'accessing' stamp: 'eem 9/28/2015 18:31'!
pushCwByteSize
	"With in-line literals we use a 12 byte sequence for loading a 64-bit immediate, which
	 is one more than strictly necessary.  The sequence ends with a 16r50 + reg opcode
	 (PushR) to allow us to distinguish between this and the (self mod: ModReg RM: rX RO: rY)
	 at the end of an ArithCwR sequence."
	<inline: true>
	^12! !

!CogInLineLiteralsX64Compiler methodsFor: 'generate machine code' stamp: 'eem 12/12/2015 12:02'!
sizePCDependentInstructionAt: eventualAbsoluteAddress
	"Size a jump and set its address.  The target may be another instruction
	 or an absolute address.  On entry the address inst var holds our virtual
	 address. On exit address is set to eventualAbsoluteAddress, which is
	 where this instruction will be output.  The span of a jump to a following
	 instruction is therefore between that instruction's address and this
	 instruction's address ((which are both still their virtual addresses), but the
	 span of a jump to a preceding instruction or to an absolute address is
	 between that instruction's address (which by now is its eventual absolute
	 address) or absolute address and eventualAbsoluteAddress."

	| target maximumSpan abstractInstruction |
	<var: #abstractInstruction type: #'AbstractInstruction *'>
	opcode = AlignmentNops ifTrue:
		[| alignment |
		 address := eventualAbsoluteAddress.
		 alignment := operands at: 0.
		 ^machineCodeSize := (eventualAbsoluteAddress + (alignment - 1) bitAnd: alignment negated)
							   - eventualAbsoluteAddress].
	self assert: self isJump.
	target := operands at: 0.
	abstractInstruction := cogit cCoerceSimple: target to: #'AbstractInstruction *'.
	(self isAnInstruction: abstractInstruction)
		ifTrue:
			[maximumSpan := abstractInstruction address
							- (((cogit abstractInstruction: self follows: abstractInstruction)
								ifTrue: [eventualAbsoluteAddress]
								ifFalse: [address]) + 2)]
		ifFalse:
			[maximumSpan := target - (eventualAbsoluteAddress + 2)].
	address := eventualAbsoluteAddress.
	opcode >= FirstShortJump
		ifTrue:
			[machineCodeSize := (self isQuick: maximumSpan)
									ifTrue: [2]
									ifFalse: [opcode = Jump
												ifTrue: [5]
												ifFalse: [6]]]
		ifFalse:
			[machineCodeSize := opcode caseOf:
									{	[JumpLong]				->	[5].
										[JumpFull]				->	[12].
										[JumpLongZero]		->	[6].
										[JumpLongNonZero]	->	[6] }].
	^machineCodeSize "Slang can't inline the switch into the ifTrue:ifFalse: correctly"! !

!CogInLineLiteralsX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/7/2016 10:28'!
storeLiteral: literal beforeFollowingAddress: followingAddress
	"Rewrite the literal in the instruction immediately preceding followingAddress.
	 This is used in the MoveCwR, PushCw and CmpCwR cases; these are distinguished by a
	 nop following the literal load in MoveCwR, a 16r50 + reg ending the PushCw sequence, and
	 a (self mod: ModReg RM: rX RO: rY) ending the CmpCwR sequence, which is at least 16rC0."
	| lastByte base |
	lastByte := objectMemory byteAt: followingAddress - 1.
	base := followingAddress - (lastByte <= 16r90
									ifTrue:
										[lastByte = 16r90
											ifTrue: [9]		"MoveCwR"
											ifFalse: [10]]	"PushCw"
									ifFalse: [11]).			"ArithCwR"
	objectMemory unalignedLongAt: base put: literal! !

CogInLineLiteralsX64Compiler subclass: #CogX64CompilerForTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Tests'!

!CogX64CompilerForTests class methodsFor: 'test support' stamp: 'eem 9/28/2015 10:19'!
byteRegistersWithNamesDo: aBinaryBlock
	self registers
		with: #('%al' '%cl' '%dl' '%bl' '%spl' '%bpl' '%sil' '%dil' '%r8b' '%r9b' '%r10b' '%r11b' '%r12b' '%r13b' '%r14b' '%r15b')
		do: aBinaryBlock! !

!CogX64CompilerForTests class methodsFor: 'test support' stamp: 'eem 9/17/2015 11:27'!
dataRegistersWithAccessorsDo: aTrinaryBlock
	self registers withIndexDo:
		[:reg :i|
		aTrinaryBlock
			value: reg
			value: (#(rax rcx rdx rbx rsp rbp rsi rdi r8 r9 r10 r11 r12 r13 r14 r15) at: i)
			value: (#(rax: rcx: rdx: rbx: rsp: rbp: rsi: rdi:  r8: r9: r10: r11: r12: r13: r14: r15:) at: i)]! !

!CogX64CompilerForTests class methodsFor: 'test support' stamp: 'eem 9/17/2015 11:28'!
dataRegistersWithAccessorsExcept: accessorsToExclude do: aTrinaryBlock
	self registers withIndexDo:
		[:reg :i| | getter setter |
		getter := #(rax rcx rdx rbx rsp rbp rsi rdi r8 r9 r10 r11 r12 r13 r14 r15) at: i.
		setter := #(rax: rcx: rdx: rbx: rsp: rbp: rsi: rdi:  r8: r9: r10: r11: r12: r13: r14: r15:) at: i.
		(accessorsToExclude includes: getter) ifFalse:
			[aTrinaryBlock value: reg value: getter value: setter]]! !

!CogX64CompilerForTests class methodsFor: 'accessing' stamp: 'eem 12/18/2019 23:56'!
dataRegistersWithAccessorsGiven: aProcessorSimulator do: aTrinaryBlock
	"We can't use RISCTempReg, because some opcodes may be encoded as multiple instructions using RISCTempReg."
	(0 to: 15)withIndexDo:
		[:reg :i|
		(self isRISCTempRegister: reg) ifFalse:
			[aTrinaryBlock
				value: reg
				value: (aProcessorSimulator registerStateGetters at: i)
				value: (aProcessorSimulator registerStateSetters at: i)]]! !

!CogX64CompilerForTests class methodsFor: 'test support' stamp: 'eem 11/4/2015 13:59:22'!
doubleWordRegistersWithNamesDo: aBinaryBlock
	self registers
		with: #('%eax' '%ecx' '%edx' '%ebx' '%esp' '%ebp' '%esi' '%edi' '%r8d' '%r9d' '%r10d' '%r11d' '%r12d' '%r13d' '%r14d' '%r15d')
		do: aBinaryBlock! !

!CogX64CompilerForTests class methodsFor: 'test support' stamp: 'eem 9/18/2010 17:04'!
fp64RegistersWithAccessorsDo: aTrinaryBlock
	self fp64registers withIndexDo:
		[:reg :i|
		aTrinaryBlock
			value: reg
			value: (#(xmm0low xmm1low xmm2low xmm3low xmm4low xmm5low xmm6low xmm7low) at: i)
			value: (#(xmm0low: xmm1low: xmm2low: xmm3low: xmm4low: xmm5low: xmm6low: xmm7low:) at: i)]! !

!CogX64CompilerForTests class methodsFor: 'test support' stamp: 'eem 9/18/2010 17:02'!
fp64registers
	^{ DPFPReg0. DPFPReg1. DPFPReg2. DPFPReg3. DPFPReg4. DPFPReg5. DPFPReg6. DPFPReg7 } "a.k.a. (-9 to: -16 by: -1)"! !

!CogX64CompilerForTests class methodsFor: 'test support' stamp: 'eem 9/17/2015 11:30'!
registers
	^(0 to: 15) "a.k.a. { RAX. RCX. RDX. RBX. RSP. RBP. RSI. RDI. R8. R9. R10. R11. R12. R13. R14. R15 }"! !

!CogX64CompilerForTests class methodsFor: 'test support' stamp: 'eem 9/17/2015 11:31'!
registersWithNamesDo: aBinaryBlock
	self registers
		with: #('%rax' '%rcx' '%rdx' '%rbx' '%rsp' '%rbp' '%rsi' '%rdi' '%r8' '%r9' '%r10' '%r11' '%r12' '%r13' '%r14' '%r15')
		do: aBinaryBlock! !

!CogX64CompilerForTests class methodsFor: 'test support' stamp: 'eem 1/12/2016 16:54'!
wordRegistersWithNamesDo: aBinaryBlock
	self registers
		with: #('%ax' '%cx' '%dx' '%bx' '%sp' '%bp' '%si' '%di' '%r8w' '%r9w' '%r10w' '%r11w' '%r12w' '%r13w' '%r14w' '%r15w')
		do: aBinaryBlock! !

!CogX64CompilerForTests class methodsFor: 'test support' stamp: 'eem 12/1/2015 15:37'!
xmmRegistersWithNamesDo: aBinaryBlock
	(XMM0L to: XMM15L by: XMM1L - XMM0L)
		with: ((0 to: 15) collect: [:i| '%xmm', i printString])
		do: aBinaryBlock! !

!CogX64CompilerForTests methodsFor: 'generate machine code' stamp: 'nice 12/30/2020 00:06'!
concretizeAt: actualAddress
	"Override to check maxSize and machineCodeSize"

	| maxAddress nextAddress |
	maxSize ifNil: [maxSize := self computeMaximumSize].
	maxAddress := actualAddress + maxSize.
	nextAddress := super concretizeAt: actualAddress.
	self assert: (maxSize notNil
				and: [self isPCDependent
						ifTrue: [maxAddress >= nextAddress]
						ifFalse: [maxAddress = nextAddress]]).
	^nextAddress! !

!CogX64CompilerForTests methodsFor: 'testing' stamp: 'eem 4/10/2015 18:32'!
hasSSE2Instructions
	"Answer if we support SSE2"
	^true! !

CogX64Compiler subclass: #CogOutOfLineLiteralsX64Compiler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogOutOfLineLiteralsX64Compiler commentStamp: 'eem 9/26/2020 14:58' prior: 0!
A CogOutOfLineLiteralsX64Compiler is the concrete subclass of CogX64Compiler that stores literals out-of-line, accessed via pc-relative addressing.  This code generator is not used.

Instance Variables!

!CogOutOfLineLiteralsX64Compiler class methodsFor: 'class initialization' stamp: 'eem 11/28/2016 18:05'!
initializeAbstractRegistersSysV
	"Assign the abstract registers with the identities/indices of the relevant concrete registers."
	super initializeAbstractRegistersSysV.
	Extra6Reg := R8! !

!CogOutOfLineLiteralsX64Compiler class methodsFor: 'class initialization' stamp: 'eem 11/28/2016 18:06'!
initializeAbstractRegistersWin64
	"Assign the abstract registers with the identities/indices of the relevant concrete registers."
	super initializeAbstractRegistersWin64.
	Extra6Reg := R11! !

!CogOutOfLineLiteralsX64Compiler class methodsFor: 'testing' stamp: 'eem 12/25/2015 18:47'!
isRISCTempRegister: reg
	"For tests to filter-out bogus values left in the RISCTempRegister, if any."
	^false! !

!CogOutOfLineLiteralsX64Compiler methodsFor: 'register allocation' stamp: 'cb 4/16/2016 15:29'!
availableRegisterOrNoneFor: liveRegsMask
	"Answer an unused abstract register in the liveRegMask.
	 Subclasses with more registers can override to answer them.
	 N.B. Do /not/ allocate TempReg."
	<returnTypeC: #sqInt>
	(cogit register: Extra6Reg isInMask: liveRegsMask) ifFalse:
		[^Extra6Reg].
	^super availableRegisterOrNoneFor: liveRegsMask! !

!CogOutOfLineLiteralsX64Compiler methodsFor: 'testing' stamp: 'eem 9/18/2015 09:12'!
isPCDependent
	"Answer if the receiver is a pc-dependent instruction.  With out-of-line literals any instruction
	 that refers to a literal depends on the address of the literal, so add them in addition to the jumps."
	^self isJump
	  or: [opcode = AlignmentNops
	  or: [opcode ~= Literal and: [dependent notNil and: [dependent opcode = Literal]]]]! !

!CogX64Compiler class methodsFor: 'translation' stamp: 'rmacnak 11/8/2015 17:10:59'!
ISA
	^#X64! !

!CogX64Compiler class methodsFor: 'accessing' stamp: 'eem 9/28/2015 10:51'!
VarBaseReg
	"Answer the number of the reg we use to hold the base address of CoInterpreter variables"
	^RBX! !

!CogX64Compiler class methodsFor: 'accessing' stamp: 'eem 10/30/2020 22:07'!
callerSavedRegisterSetters
	"Answer the register setters for the caller-saved registers on the current ABI"
	^SysV
			ifTrue: [#(rax: rcx: rdx: rsi: rdi: r8: r9: r10: r11:)]
			ifFalse: [#(rax: rcx: rdx: r8: r9: r10: r11:)]! !

!CogX64Compiler class methodsFor: 'accessing' stamp: 'eem 10/29/2020 23:32'!
cpuidWord1
	^cpuidWord1! !

!CogX64Compiler class methodsFor: 'translation' stamp: 'eem 11/11/2020 18:36'!
declareCVarsIn: aCCodeGenerator
	aCCodeGenerator
		declareVar: 'cpuidWord1' type: #usqIntptr_t! !

!CogX64Compiler class methodsFor: 'translation' stamp: 'eem 12/4/2015 22:25'!
defaultCompilerClass
	^CogInLineLiteralsX64Compiler! !

!CogX64Compiler class methodsFor: 'translation' stamp: 'eem 10/25/2020 09:38'!
filteredInstVarNames
	"Move the unaligned bytes to the end of the structure."
	^(super filteredInstVarNames copyWithout: 'machineCode'), #('machineCode')! !

!CogX64Compiler class methodsFor: 'translation' stamp: 'eem 12/11/2015 23:43'!
identifyingPredefinedMacros
	^#('x86_64' '__amd64' '__x86_64' '__amd64__' '__x86_64__' '_M_AMD64' '_M_X64')! !

!CogX64Compiler class methodsFor: 'translation' stamp: 'eem 5/16/2017 09:19'!
ifTranslateableAddWithOptionsTo: aCollection
	"Override to create cogitX64.c and cogitX64Win64.c"
	(self wordSize = Cogit objectMemoryClass wordSize
	 and: [self identifyingPredefinedMacros notNil]) ifTrue:
		[aCollection
			"SysV must preceed _WIN64; see Cogit class>>#generateCodeStringForCogitDotC"
			add: {self. {#ISA. self ISA. #ABI. #SysV}}; 
			add: {self. {#ISA. self ISA. #ABI. #'_WIN64'}}]! !

!CogX64Compiler class methodsFor: 'class initialization' stamp: 'eem 1/19/2021 20:42'!
initialize
	"Initialize various x64 instruction-related constants.
	 [1] IA-32 Intel® Architecture Software Developer's Manual Volume 2A: Instruction Set Reference, A-M"

	"CogX64Compiler initialize"

	self ~~ CogX64Compiler ifTrue: [^self].

	(InitializationOptions ifNil: [Dictionary new])
		at: #ABI
		ifPresent: [:abi| SysV := abi asUppercase ~= #WIN64 and: [abi asUppercase ~= #'_WIN64']]
		ifAbsent: [SysV := true]. "Default ABI; set to true for SysV, false for WIN64/_WIN64"

	"Initialize most sets of variables that run from 0 to N - 1..."
	#(	"16 integer registers, rcx, rdx, rbx bizarrely inconsistent"
		(RAX RCX RDX RBX RSP RBP RSI RDI R8 R9 R10 R11 R12 R13 R14 R15)
		"16 lower half floating point registers"
		(XMM0L XMM1L XMM2L XMM3L XMM4L XMM5L XMM6L XMM7L XMM8L XMM9L XMM10L XMM11L XMM12L XMM13L XMM14L XMM15L)
		"Mod R/M Mod fields.  See [1] Sec 2.4, 2.5 & 2.6 & Table 2-2"
		(ModRegInd ModRegRegDisp8 ModRegRegDisp32 ModReg ModRegIndSIB ModRegIndDisp32)
		"SIB Scaled Index modes.  See [1] Sec 2.4, 2.5 & 2.6 & Table 2-3"
		(SIB1 SIB2 SIB4 SIB8)) do:
		[:classVarNames|
		 classVarNames withIndexDo:
			[:k :v|
			self classPool at: k put: v - 1]].

	"Specific instructions"
	self initializeSpecificOpcodes: #(CPUID
	"Arithmetic:"					CDQ IDIVR IMULRR XCHGRR
	"Bit count/data movement:"	CLD REP MOVSB MOVSQ BSR 	
	"Multi-processing:"				LFENCE MFENCE SFENCE LOCK CMPXCHGRMr SETE MoveRAwNoVBR)
		in: thisContext method
		extraClassVarNames: #(CArg0Reg CArg1Reg CArg2Reg CArg3Reg)! !

!CogX64Compiler class methodsFor: 'class initialization' stamp: 'eem 1/30/2020 20:23'!
initializeAbstractRegisters
	"Assign the abstract registers with the identities/indices of the relevant concrete registers."
	"[1] Figure 3.4 Register Usage in
		System V Application Binary Interface
		AMD64 Architecture Processor Supplement"

	"N.B. RAX RCX & RDX are caller-save (scratch) registers.  Hence we use RCX for class and RDX for
		receiver/result since these are written in all normal sends."

	SysV
		ifTrue: [self initializeAbstractRegistersSysV]
		ifFalse: [self initializeAbstractRegistersWin64].

	CallerSavedRegisterMask := ABICallerSavedRegisterMask
								bitAnd: (self registerMaskFor: ClassReg and: ReceiverResultReg and: SendNumArgsReg and: Arg0Reg and: Arg1Reg).

	NumRegisters := 16.

	DPFPReg0				:= XMM0L.
	DPFPReg1				:= XMM1L.
	DPFPReg2				:= XMM2L.
	DPFPReg3				:= XMM3L.
	DPFPReg4				:= XMM4L.
	DPFPReg5				:= XMM5L.
	DPFPReg6				:= XMM6L.
	DPFPReg7				:= XMM7L.
	DPFPReg8				:= XMM8L.
	DPFPReg9				:= XMM9L.
	DPFPReg10				:= XMM10L.
	DPFPReg11				:= XMM11L.
	DPFPReg12				:= XMM12L.
	DPFPReg13				:= XMM13L.
	DPFPReg14				:= XMM14L.
	DPFPReg15				:= XMM15L.

	NumFloatRegisters := 16! !

!CogX64Compiler class methodsFor: 'class initialization' stamp: 'eem 10/25/2020 15:43'!
initializeAbstractRegistersSysV
	"Assign the abstract registers with the identities/indices of the relevant concrete registers."
	"[1] Figure 3.4 Register Usage in
		System V Application Binary Interface
		AMD64 Architecture Processor Supplement"

	"N.B. RAX RCX & RDX are caller-save (scratch) registers.  Hence we use RCX for class and RDX for
		receiver/result since these are written in all normal sends."

	ABICalleeSavedRegisterMask := self registerMaskFor: RBX and: R12 and: R13 and: R14 and: R15.
	ABICallerSavedRegisterMask := self
									registerMaskFor: RAX
									and: RCX
									and: RDX
									and: RSI
									and: RDI
									and: R8
									and: R9
									and: R10
									and: R11.

	TempReg				:= RAX.
	ClassReg				:= RCX.
	ReceiverResultReg		:= RDX.
	SendNumArgsReg		:= R9.
	SPReg					:= RSP.
	FPReg					:= RBP.
	Arg0Reg				:= RDI. "So as to agree with C ABI arg 0"
	Arg1Reg				:= RSI. "So as to agree with C ABI arg 1"
	VarBaseReg				:= RBX. "Must be callee saved"
	"R8 is either RISCTempReg or Extra6Reg depending on subclass."
	Extra0Reg				:= R10.
	Extra1Reg				:= R11.
	Extra2Reg				:= R12.
	Extra3Reg				:= R13.
	Extra4Reg				:= R14.
	Extra5Reg				:= R15.

	"The SysV ABI has 6 integer argument registers; we use only 4." 
	CArg0Reg := RDI.
	CArg1Reg := RSI.
	CArg2Reg := RDX.
	CArg3Reg := RCX! !

!CogX64Compiler class methodsFor: 'class initialization' stamp: 'eem 10/25/2020 15:46'!
initializeAbstractRegistersWin64
	"Assign the abstract registers with the identities/indices of the relevant concrete registers."

	"N.B. Since receiver/result are written in all normal sends,
	 it's better to use scratch registers for them (those which are caller-saved).
	 In Win64 ABI, this does not leave that many choices:
		- RAX is TempReg (overwritten by result etc...)
		- RCX and RDX are used for first 2 args (see genMarshallNArgs:arg:arg:arg:arg:)
		- it remains R8,R9,R10 & R11 : we choose the first two"

	ABICalleeSavedRegisterMask := self registerMaskFor: RBX and: RSI and: RDI and: R12 and: R13 and: R14 and: R15.
	ABICallerSavedRegisterMask := self
									registerMaskFor: RAX
									and: RCX
									and: RDX
									and: R8
									and: R9
									and: R10
									and: R11.

	TempReg				:= RAX.
	ClassReg				:= R8.
	ReceiverResultReg		:= R9.
	SendNumArgsReg		:= R10.
	SPReg					:= RSP.
	FPReg					:= RBP.
	Arg0Reg				:= RCX. "So as to agree with C ABI arg 0"
	Arg1Reg				:= RDX. "So as to agree with C ABI arg 1"
	VarBaseReg				:= RBX. "Must be callee saved"
	"R11 is either RISCTempReg or Extra6Reg depending on subclass."
	Extra0Reg				:= RDI.
	Extra1Reg				:= RSI.
	Extra2Reg				:= R12.
	Extra3Reg				:= R13.
	Extra4Reg				:= R14.
	Extra5Reg				:= R15.

	CArg0Reg := RCX.
	CArg1Reg := RDX.
	CArg2Reg := R8.
	CArg3Reg := R9! !

!CogX64Compiler class methodsFor: 'testing' stamp: 'eem 12/12/2015 11:45'!
isAbstract
	^self == CogX64Compiler! !

!CogX64Compiler class methodsFor: 'accessing' stamp: 'nice 11/28/2016 23:18'!
isSysV
	"Answer true is ABI is SysV, false otherwise (for WIN64)"
	^SysV! !

!CogX64Compiler class methodsFor: 'translation' stamp: 'eem 9/17/2015 14:52'!
machineCodeDeclaration
	"Answer the declaration for the machineCode array."
	^{#'unsigned char'. '[', self basicNew machineCodeBytes printString, ']'}! !

!CogX64Compiler class methodsFor: 'translation' stamp: 'eem 6/7/2018 15:16'!
moduleName
	"CogAbstractInstruction subclasses collect: [:ea| ea moduleName]"
	^'cogit', self ISA, ((InitializationOptions at: #ABI ifAbsent: ['']) copyWithout: $_)! !

!CogX64Compiler class methodsFor: 'translation' stamp: 'eem 12/11/2015 23:30'!
wordSize
	"This is a 64-bit ISA"
	^8! !

!CogX64Compiler methodsFor: 'register allocation' stamp: 'cb 4/16/2016 15:23'!
availableRegisterOrNoneFor: liveRegsMask
	"Answer an unused abstract register in the liveRegMask.
	 Subclasses with more registers can override to answer them.
	 N.B. Do /not/ allocate TempReg."
	<returnTypeC: #sqInt>
	(cogit register: Extra5Reg isInMask: liveRegsMask) ifFalse:
		[^Extra5Reg].
	(cogit register: Extra4Reg isInMask: liveRegsMask) ifFalse:
		[^Extra4Reg].
	(cogit register: Extra3Reg isInMask: liveRegsMask) ifFalse:
		[^Extra3Reg].
	(cogit register: Extra2Reg isInMask: liveRegsMask) ifFalse:
		[^Extra2Reg].
	(cogit register: Extra1Reg isInMask: liveRegsMask) ifFalse:
		[^Extra1Reg].
	(cogit register: Extra0Reg isInMask: liveRegsMask) ifFalse:
		[^Extra0Reg].
	^super availableRegisterOrNoneFor: liveRegsMask! !

!CogX64Compiler methodsFor: 'abi' stamp: 'rsf 1/12/2017 15:12'!
cFloatResultToRd: reg
	XMM0L ~= reg ifTrue: [ 
		cogit MoveRd: XMM0L Rd: reg
	].! !

!CogX64Compiler methodsFor: 'abi' stamp: 'rsf 1/12/2017 15:20'!
cFloatResultToRs: reg
	XMM0L ~= reg ifTrue: [ 
		cogit MoveRs: XMM0L Rs: reg
	].! !

!CogX64Compiler methodsFor: 'accessing' stamp: 'rsf 2/21/2017 03:29'!
cResultRegisterHigh
	"Answer the register through which C functions return the high part of big integral results."
	<inline: true>
	^ RDX! !

!CogX64Compiler methodsFor: 'accessing' stamp: 'eem 7/8/2021 14:48'!
callFullInstructionByteSize
	^12! !

!CogX64Compiler methodsFor: 'full transfer run-time support' stamp: 'eem 12/6/2015 21:11'!
callFullTargetFromReturnAddress: callSiteReturnAddress
	"Answer the address the full call immediately preceding callSiteReturnAddress will jump to."
	^self sixtyFourBitLiteralBefore: callSiteReturnAddress - 2! !

!CogX64Compiler methodsFor: 'accessing' stamp: 'eem 2/13/2009 12:29'!
callInstructionByteSize
	^5! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/13/2015 19:13'!
callTargetFromReturnAddress: callSiteReturnAddress
	"Answer the address the call immediately preceding callSiteReturnAddress will jump to."
	| callDistance |
	callDistance := self literal32BeforeFollowingAddress: callSiteReturnAddress.
	^callSiteReturnAddress + callDistance signedIntFromLong! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
canDivQuoRem
	^true! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
canMulRR
	^true! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 6/21/2018 12:12'!
canSignExtend
	"x64 has native SignExtend8RR, SignExtend16RR, & SignExtend32RR."
	<inline: true>
	^true! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 6/21/2018 12:29'!
canZeroExtend
	"x64 has native ZeroExtend8RR, ZeroExtend16RR, & ZeroExtend32RR."
	<inline: true>
	^true! !

!CogX64Compiler methodsFor: 'accessing' stamp: 'eem 1/15/2016 10:47'!
cmpC32RTempByteSize
	^5! !

!CogX64Compiler methodsFor: 'accessing' stamp: 'eem 9/22/2015 12:15'!
codeGranularity
	^1! !

!CogX64Compiler methodsFor: 'generate machine code' stamp: 'eem 8/5/2021 10:38'!
computeMaximumSize
	"Compute the maximum size for each opcode.  This allows jump offsets to
	 be determined, provided that all backward branches are long branches."
	"N.B.  The ^N forms are to get around the bytecode compiler's long branch
	 limits which are exceeded when each case jumps around the otherwise."
	opcode caseOf: {
		"Noops & Pseudo Ops"
		[Label]					-> [^0].
		[AlignmentNops]		-> [^(operands at: 0) - 1].
		[Fill32]					-> [^4].
		[Nop]					-> [^1].
		"Specific Arithmetic"
		[CDQ]					-> [^2].
		[IDIVR]					-> [^3].
		[XCHGRR]				-> [^((operands at: 0) = RAX
									   or: [(operands at: 1) = RAX])
											ifTrue: [2]
											ifFalse: [3]].
		[IMULRR]				-> [^4].
		[CPUID]					-> [^2].
		"Processor Specific/Multi-threading"
		"[CMPXCHGAwR]			-> [^12].
		[CMPXCHGMwrR]		-> [^9]."
		[LFENCE]				-> [^3].
		[MFENCE]				-> [^3].
		[SFENCE]				-> [^3].
		[LOCK]					-> [^1].
		[MoveRAwNoVBR]		-> [^(operands at: 0) = RAX
										ifTrue: [10]
										ifFalse: [((operands at: 0) = RBP or: [(operands at: 0) = RSP])
													ifTrue: [13]
													ifFalse: [14]]].
		[CMPXCHGRMr]			-> [^4].
		[SETE]					-> [^4].
		"[XCHGMwrR]			-> [^7]."
		"Specific Data Movement"
		[REP]					-> [^1].
		[CLD]					-> [^1].
		[MOVSB]				-> [^1].
		[MOVSQ]				-> [^2].
		"Control"
		[CallFull]					-> [^12].
		[Call]						-> [^5].
		[CallR]						-> [^3].
		[JumpR]					-> [^3].
		[JumpFull]					-> [self resolveJumpTarget. ^12].
		[JumpLong]					-> [self resolveJumpTarget. ^5].
		[Jump]						-> [self resolveJumpTarget. ^5].
		[JumpZero]					-> [self resolveJumpTarget. ^6].
		[JumpNonZero]				-> [self resolveJumpTarget. ^6].
		[JumpNegative]				-> [self resolveJumpTarget. ^6].
		[JumpNonNegative]			-> [self resolveJumpTarget. ^6].
		[JumpOverflow]				-> [self resolveJumpTarget. ^6].
		[JumpNoOverflow]			-> [self resolveJumpTarget. ^6].
		[JumpCarry]				-> [self resolveJumpTarget. ^6].
		[JumpNoCarry]				-> [self resolveJumpTarget. ^6].
		[JumpLess]					-> [self resolveJumpTarget. ^6].
		[JumpGreaterOrEqual]		-> [self resolveJumpTarget. ^6].
		[JumpGreater]				-> [self resolveJumpTarget. ^6].
		[JumpLessOrEqual]			-> [self resolveJumpTarget. ^6].
		[JumpBelow]				-> [self resolveJumpTarget. ^6].
		[JumpAboveOrEqual]		-> [self resolveJumpTarget. ^6].
		[JumpAbove]				-> [self resolveJumpTarget. ^6].
		[JumpBelowOrEqual]		-> [self resolveJumpTarget. ^6].
		[JumpLongZero]			-> [self resolveJumpTarget. ^6].
		[JumpLongNonZero]		-> [self resolveJumpTarget. ^6].
		[JumpFPEqual]				-> [self resolveJumpTarget. ^6].
		[JumpFPNotEqual]			-> [self resolveJumpTarget. ^6].
		[JumpFPLess]				-> [self resolveJumpTarget. ^6].
		[JumpFPGreaterOrEqual]	-> [self resolveJumpTarget. ^6].
		[JumpFPGreater]			-> [self resolveJumpTarget. ^6].
		[JumpFPLessOrEqual]		-> [self resolveJumpTarget. ^6].
		[JumpFPOrdered]			-> [self resolveJumpTarget. ^6].
		[JumpFPUnordered]			-> [self resolveJumpTarget. ^6].
		[RetN]						-> [^(operands at: 0) = 0 ifTrue: [1] ifFalse: [3]].
		[Stop]						-> [^1].

		"Arithmetic"
		[AddCqR]		-> [^self computeSizeOfArithCqR].
		[AddcCqR]		-> [^self computeSizeOfArithCqR].
		[AndCqR]		-> [^self computeSizeOfArithCqR].
		[CmpCqR]		-> [^self computeSizeOfArithCqR].
		[OrCqR]		-> [^self computeSizeOfArithCqR].
		[SubCqR]		-> [^self computeSizeOfArithCqR].
		[SubbCqR]		-> [^self computeSizeOfArithCqR].
		[TstCqR]		-> [^self computeSizeOfArithCqR].
		[AddCwR]		-> [^self computeSizeOfArithCwR].
		[AndCwR]		-> [^self computeSizeOfArithCwR].
		[CmpCwR]		-> [^self computeSizeOfArithCwR].
		[CmpC32R]		-> [^(operands at: 1) <= 7
								ifTrue: [(operands at: 1) = RAX
											ifTrue: [5]
											ifFalse: [6]]
								ifFalse: [7]].
		[OrCwR]		-> [^self computeSizeOfArithCwR].
		[SubCwR]		-> [^self computeSizeOfArithCwR].
		[XorCwR]		-> [^self computeSizeOfArithCwR].
		[AddRR]		-> [^3].
		[AddcRR]		-> [^3].
		[AndRR]		-> [^3].
		[CmpRR]		-> [^3].
		[OrRR]			-> [^3].
		[XorRR]			-> [^3].
		[SubRR]		-> [^3].
		[SubbRR]		-> [^3].
		[NegateR]		-> [^3].
		[LoadEffectiveAddressMwrR]
						-> [^((self isQuick: (operands at: 0))
									ifTrue: [4]
									ifFalse: [7])
								+ (((operands at: 1) bitAnd: 7) = RSP
									ifTrue: [1]
									ifFalse: [0])].
		[LogicalShiftLeftCqR]		-> [^(operands at: 0) = 1 ifTrue: [3] ifFalse: [4]].
		[LogicalShiftRightCqR]		-> [^(operands at: 0) = 1 ifTrue: [3] ifFalse: [4]].
		[ArithmeticShiftRightCqR]	-> [^(operands at: 0) = 1 ifTrue: [3] ifFalse: [4]].
		[RotateRightCqR]			-> [^(operands at: 0) = 1 ifTrue: [3] ifFalse: [4]].
		[RotateLeftCqR]			-> [^(operands at: 0) = 1 ifTrue: [3] ifFalse: [4]].
		[LogicalShiftLeftRR]			-> [^self computeShiftRRSize].
		[LogicalShiftRightRR]		-> [^self computeShiftRRSize].
		[ArithmeticShiftRightRR]		-> [^self computeShiftRRSize].
		[AddRdRd]					-> [^ 4 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
											ifTrue: [1]
											ifFalse: [0])].
		[CmpRdRd]					-> [^ 4
											+ ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
											ifTrue: [1]
											ifFalse: [0])].
		[SubRdRd]					-> [^ 4 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
											ifTrue: [1]
											ifFalse: [0])].
		[MulRdRd]					-> [^ 4 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
											ifTrue: [1]
											ifFalse: [0])].
		[DivRdRd]					-> [^ 4 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
											ifTrue: [1]
											ifFalse: [0])].
		[SqrtRd]					-> [^ 4 + (((operands at: 0) > 7)
											ifTrue: [1]
											ifFalse: [0])].
		[XorRdRd]					-> [^ 4 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
											ifTrue: [1]
											ifFalse: [0])].
						
		[AddRsRs]				-> [^ 4 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
											ifTrue: [1]
											ifFalse: [0])].
		[CmpRsRs]				-> [^ 3 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
											ifTrue: [1]
											ifFalse: [0])].
		[SubRsRs]				-> [^ 4 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
											ifTrue: [1]
											ifFalse: [0])].
		[MulRsRs]				-> [^ 4 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
											ifTrue: [1]
											ifFalse: [0])].
		[DivRsRs]				-> [^ 4 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
											ifTrue: [1]
											ifFalse: [0])].
		[SqrtRs]					-> [^ 4 + (((operands at: 0) > 7)
											ifTrue: [1]
											ifFalse: [0])].
		[XorRsRs]					-> [^ 3 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
											ifTrue: [1]
											ifFalse: [0])].
		[ClzRR]					-> [^5].
		[BSR]					-> [^4].
						
		"Data Movement"
		[MoveCqR]		-> [^(operands at: 0) = 0
								ifTrue: [3]
								ifFalse:
									[(self is32BitSignedImmediate: (operands at: 0))
										ifTrue: [7]
										ifFalse: [self moveCwRByteSize - 1]]].
		[MoveCwR]		-> [^(self inCurrentCompilation: (operands at: 0))
								ifTrue: [7]
								ifFalse: [self moveCwRByteSize]].
		[MoveC32R]	-> [^7]. "N.B. Always inlined."
		[MoveRR]		-> [^3].
		[MoveRdRd]		-> [^4 
								+((((operands at: 1) > 7) or: [(operands at: 0) > 7])
											ifTrue: [1]
											ifFalse: [0])].
		[MoveRsRs]		-> [^4
								 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
											ifTrue: [1]
											ifFalse: [0])].
		[MoveRRd]		-> [^5].
		[MoveRdR]		-> [^5].
		[MoveAwR]		-> [^(self isAddressRelativeToVarBase: (operands at: 0))
								ifTrue:
									[(self isQuick: (operands at: 0) - cogit varBaseAddress)
										ifTrue: [((operands at: 0) - cogit varBaseAddress = 0
												and: [((operands at: 1) bitAnd: 7) ~= RBP])
													ifTrue: [3]
													ifFalse: [4]]
										ifFalse: [7]]
								ifFalse:
									[(operands at: 1) = RAX
										ifTrue: [10]
										ifFalse: [((operands at: 1) = RBP or: [(operands at: 1) = RSP])
													ifTrue: [12]
													ifFalse: [14]]]].
		[MoveA32R]	-> [^(operands at: 1) = RAX ifTrue: [9] ifFalse: [13]].
		[MoveRAw]		-> [^(self isAddressRelativeToVarBase: (operands at: 1))
								ifTrue:
									[(self isQuick: (operands at: 1) - cogit varBaseAddress)
										ifTrue: [((operands at: 1) - cogit varBaseAddress = 0
												and: [((operands at: 0) bitAnd: 7) ~= RBP])
													ifTrue: [3]
													ifFalse: [4]]
										ifFalse: [7]]
								ifFalse:
									[(operands at: 0) = RAX
										ifTrue: [10]
										ifFalse: [((operands at: 0) = RBP or: [(operands at: 0) = RSP])
													ifTrue: [13]
													ifFalse: [14]]]].
		[MoveRA32]	-> [^(operands at: 0) = RAX ifTrue: [9] ifFalse: [13]].
		[MoveAbR]		-> [^(self isAddressRelativeToVarBase: (operands at: 0))
								ifTrue:
									[(self isQuick: (operands at: 0) - cogit varBaseAddress)
										ifTrue: [((operands at: 0) - cogit varBaseAddress = 0
												and: [((operands at: 1) bitAnd: 7) ~= RBP])
													ifTrue: [3]
													ifFalse: [4]]
										ifFalse: [7]]
								ifFalse: [(operands at: 1) = RAX ifTrue: [10] ifFalse: [14]]].
		[MoveRAb]		-> [^(self isAddressRelativeToVarBase: (operands at: 1))
								ifTrue:
									[(self isQuick: (operands at: 1) - cogit varBaseAddress)
										ifTrue: [((operands at: 1) - cogit varBaseAddress = 0
												and: [((operands at: 0) bitAnd: 7) ~= RBP])
													ifTrue: [3]
													ifFalse: [4]]
										ifFalse: [7]]
								ifFalse: [(operands at: 0) = RAX ifTrue: [10] ifFalse: [14]]].
		[MoveRMwr]	-> [self assert: (self is32BitSignedImmediate: (operands at: 1)).
							^((self isQuick: (operands at: 1))
									ifTrue: [((operands at: 1) = 0
											and: [((operands at: 2) bitAnd: 7) ~= RBP])
												ifTrue: [3]
												ifFalse: [4]]
									ifFalse: [7])
								+ (((operands at: 2) bitAnd: 7) = RSP
									ifTrue: [1]
									ifFalse: [0])].
		[MoveRM32r]	-> [^((self isQuick: (operands at: 1))
								ifTrue: [((operands at: 1) = 0
										and: [((operands at: 2) bitAnd: 6) ~= RSP])
											ifTrue: [2]
											ifFalse: [3]]
								ifFalse: [6])
							+ ((((operands at: 2) bitAnd: 7) = RSP and: [(operands at: 1) ~= 0])
								ifTrue: [1]
								ifFalse: [0])
							+ ((((operands at: 2) > 7) or: [(operands at: 0) > 7])
								ifTrue: [1]
								ifFalse: [0])].
		[MoveRsM32r]	-> [^((self isQuick: (operands at: 1))
								ifTrue: [((operands at: 1) = 0
										and: [((operands at: 2) bitAnd: 6) ~= RSP])
											ifTrue: [4]
											ifFalse: [5]]
								ifFalse: [8])
							+ ((((operands at: 2) bitAnd: 7) = RSP and: [(operands at: 1) ~= 0])
								ifTrue: [1]
								ifFalse: [0])
							+ ((((operands at: 2) > 7) or: [(operands at: 0) > 7])
								ifTrue: [1]
								ifFalse: [0])].
		[MoveRdM64r]	-> [^((self isQuick: (operands at: 1))
								ifTrue: [((operands at: 1) = 0
										and: [((operands at: 2) bitAnd: 6) ~= RSP])
											ifTrue: [4]
											ifFalse: [5]]
								ifFalse: [8])
							+ ((((operands at: 2) bitAnd: 7) = RSP and: [(operands at: 1) ~= 0])
								ifTrue: [1]
								ifFalse: [0])
							+ ((((operands at: 2) > 7) or: [(operands at: 0) > 7])
								ifTrue: [1]
								ifFalse: [0])].
		[MoveMbrR]		-> [self assert: (self is32BitSignedImmediate: (operands at: 0)).
							^((self isQuick: (operands at: 0))
									ifTrue: [((operands at: 0) = 0
											and: [((operands at: 1) bitAnd: 7) ~= RBP])
												ifTrue: [3]
												ifFalse: [4]]
									ifFalse: [7])
								+ (((operands at: 1) bitAnd: 7) = RSP
									ifTrue: [1]
									ifFalse: [0])].
		[MoveRMbr]		-> [self assert: (self is32BitSignedImmediate: (operands at: 1)).
							^((self isQuick: (operands at: 1))
									ifTrue: [((operands at: 1) = 0
											and: [((operands at: 0) bitAnd: 7) ~= RBP])
												ifTrue: [3]
												ifFalse: [4]]
									ifFalse: [7])
								+ (((operands at: 2) bitAnd: 7) = RSP
									ifTrue: [1]
									ifFalse: [0])].
		[MoveM8rR]		-> [self assert: (self is32BitSignedImmediate: (operands at: 0)).
							^((self isQuick: (operands at: 0))
									ifTrue: [((operands at: 0) = 0
											and: [((operands at: 1) bitAnd: 7) ~= RBP])
												ifTrue: [3]
												ifFalse: [4]]
									ifFalse: [7])
								+ (((operands at: 1) bitAnd: 7) = RSP
									ifTrue: [1]
									ifFalse: [0])].
		[MoveMs8rR]		-> [self assert: (self is32BitSignedImmediate: (operands at: 0)).
							^((self isQuick: (operands at: 0))
									ifTrue: [((operands at: 0) = 0
											and: [((operands at: 1) bitAnd: 7) ~= RBP])
												ifTrue: [3]
												ifFalse: [4]]
									ifFalse: [7])
								+ (((operands at: 1) bitAnd: 7) = RSP
									ifTrue: [1]
									ifFalse: [0])].
		[MoveRM8r]		-> [self assert: (self is32BitSignedImmediate: (operands at: 1)).
							^((self isQuick: (operands at: 1))
									ifTrue: [((operands at: 1) = 0
											and: [((operands at: 0) bitAnd: 7) ~= RBP])
												ifTrue: [3]
												ifFalse: [4]]
									ifFalse: [7])
								+ (((operands at: 2) bitAnd: 7) = RSP
									ifTrue: [1]
									ifFalse: [0])].
		[MoveM16rR]	-> [self assert: (self is32BitSignedImmediate: (operands at: 0)).
							^((self isQuick: (operands at: 0))
									ifTrue: [((operands at: 0) = 0
											and: [((operands at: 1) bitAnd: 7) ~= RBP])
												ifTrue: [4]
												ifFalse: [5]]
									ifFalse: [8])
								+ (((operands at: 1) bitAnd: 7) = RSP
									ifTrue: [1]
									ifFalse: [0])].
		[MoveRM16r]	-> [self assert: (self is32BitSignedImmediate: (operands at: 1)).
							^((self isQuick: (operands at: 1))
									ifTrue: [4]
									ifFalse: [7])
								+ (((operands at: 2) bitAnd: 7) = RSP
									ifTrue: [1]
									ifFalse: [0])
								+ (((operands at: 0) > 7 or: [(operands at: 2) > 7])
									ifTrue: [1]
									ifFalse: [0])].
		[MoveM32rR]	-> [^((self isQuick: (operands at: 0))
								ifTrue: [((operands at: 0) = 0
										and: [((operands at: 1) bitAnd: 6) ~= RSP])
											ifTrue: [2]
											ifFalse: [3]]
								ifFalse: [6])
							+ ((((operands at: 1) bitAnd: 7) = RSP and: [(operands at: 0) ~= 0])
								ifTrue: [1]
								ifFalse: [0])
							+ ((((operands at: 1) > 7) or: [(operands at: 2) > 7])
								ifTrue: [1]
								ifFalse: [0])].
		[MoveM32rRs]	-> [^((self isQuick: (operands at: 0))
								ifTrue: [((operands at: 0) = 0
										and: [((operands at: 1) bitAnd: 6) ~= RSP])
											ifTrue: [4]
											ifFalse: [5]]
								ifFalse: [8])
							+ ((((operands at: 1) bitAnd: 7) = RSP and: [(operands at: 0) ~= 0])
								ifTrue: [1]
								ifFalse: [0])
							+ ((((operands at: 1) > 7) or: [(operands at: 2) > 7])
								ifTrue: [1]
								ifFalse: [0])].
		[MoveM64rRd]	-> [^((self isQuick: (operands at: 0))
								ifTrue: [((operands at: 0) = 0
										and: [((operands at: 1) bitAnd: 6) ~= RSP])
											ifTrue: [4]
											ifFalse: [5]]
								ifFalse: [8])
							+ ((((operands at: 1) bitAnd: 7) = RSP and: [(operands at: 0) ~= 0])
								ifTrue: [1]
								ifFalse: [0])
							+ ((((operands at: 1) > 7) or: [(operands at: 2) > 7])
								ifTrue: [1]
								ifFalse: [0])].
		[MoveMwrR]	-> [self assert: (self is32BitSignedImmediate: (operands at: 0)).
							^((self isQuick: (operands at: 0))
								ifTrue: [((operands at: 0) = 0
										and: [((operands at: 1) bitAnd: 7) ~= RBP])
											ifTrue: [3]
											ifFalse: [4]]
								ifFalse: [7])
							+ (((operands at: 1) bitAnd: 7) = RSP
								ifTrue: [1]
								ifFalse: [0])].
		[MoveXbrRR]	-> [self assert: (operands at: 0) ~= RSP.
							^((operands at: 1) bitAnd: 7) = RBP
											ifTrue: [5]
											ifFalse: [4]].
		[MoveRXbrR]	->	[self assert: (operands at: 1) ~= RSP.
							^(((operands at: 0) > 3
							   or: [(operands at: 1) > 7
							   or: [(operands at: 2) > 7]])
								ifTrue: [4]
								ifFalse: [3])
							+ (((operands at: 2) bitAnd: 7) = RBP
								ifTrue: [1]
								ifFalse: [0])].
		[MoveXwrRR]	-> [self assert: (operands at: 0) ~= RSP.
							^((operands at: 1) = RBP
							   or: [(operands at: 1) = R13])
											ifTrue: [5]
											ifFalse: [4]].
		[MoveRXwrR]	-> [self assert: (operands at: 1) ~= RSP.
							^((operands at: 2) = RBP
							   or: [(operands at: 2) = R13])
											ifTrue: [5]
											ifFalse: [4]].
		[MoveX32rRR]	-> [self assert: (operands at: 0) ~= RSP.
							^(((operands at: 1) = RBP
							   or: [(operands at: 1) = R13])
										ifTrue: [7]
										ifFalse: [6])
							 + (((operands at: 0) > 7
							     or: [(operands at: 1) > 7
							     or: [(operands at: 2) > 7]])
										ifTrue: [1]
										ifFalse: [0])].
		[MoveRX32rR]	-> [self assert: (operands at: 1) ~= RSP.
							^(((operands at: 2) = RBP
							   or: [(operands at: 2) = R13])
										ifTrue: [4]
										ifFalse: [3])
							 + (((operands at: 0) > 7
							     or: [(operands at: 1) > 7
							     or: [(operands at: 2) > 7]])
										ifTrue: [1]
										ifFalse: [0])].
		[PopR]			-> [^(operands at: 0) < 8 ifTrue: [1] ifFalse: [2]].
		[PushR]			-> [^(operands at: 0) < 8 ifTrue: [1] ifFalse: [2]].
		[PushCq]		-> [^(self isQuick: (operands at: 0))
								ifTrue: [2]
								ifFalse:
									[(self is32BitSignedImmediate: (operands at: 0))
										ifTrue: [5]
										ifFalse: [self computeSizeOfPushCw]]].
		[PushCw]		-> [^self computeSizeOfPushCw].
		[PrefetchAw]	-> [^(self isAddressRelativeToVarBase: (operands at: 0)) ifTrue: [7] ifFalse: [0]].
		"Conversion"
		[ConvertRRd]		-> [^5].
		[ConvertRdR]		-> [^5].
		[ConvertRRs]		-> [^4 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
									ifTrue: [1]
									ifFalse: [0])].
		[ConvertRsR]		-> [^4 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
									ifTrue: [1]
									ifFalse: [0])].
		[ConvertRsRd]	-> [^4 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
									ifTrue: [1]
									ifFalse: [0])].
		[ConvertRdRs]	-> [^4 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
									ifTrue: [1]
									ifFalse: [0])].

		[SignExtend8RR] -> [^4].
		[SignExtend16RR] -> [^4].
		[SignExtend32RR] -> [^3].
		[ZeroExtend8RR] -> [^4].
		[ZeroExtend16RR] -> [^4].
		[ZeroExtend32RR] -> [^2 + ((((operands at: 1) > 7) or: [(operands at: 0) > 7])
									ifTrue: [1]
									ifFalse: [0])].

		[MovePerfCnt64RL]	->	["it is easier to calculate the right value by concretizing.  We concretize twice, but so what?"
								^self concretizeMovePerfCnt64RL].
		}.
	^0 "to keep C compiler quiet"! !

!CogX64Compiler methodsFor: 'generate machine code' stamp: 'eem 12/27/2015 11:09'!
computeShiftRRSize
	"On the x86 the only instructions that shift by the value of a
	 register require the shift count to be  in %ecx.  So we may
	 have to use swap instructions to get the count into ecx."
	| shiftCountReg |
	shiftCountReg := operands at: 0.
	^shiftCountReg = RCX
		ifTrue: [3]
		ifFalse:
			[shiftCountReg = RAX
				ifTrue: [2 "XCHG RAX,r2" + 3 "Sxx" + 2 "XCHG RAX,r2"]
				ifFalse: [3 "XCHG r1,r2" + 3 "Sxx" + 3 "XCHG r1,r2"]]! !

!CogX64Compiler methodsFor: 'generate machine code' stamp: 'eem 12/4/2015 18:06'!
computeSizeOfArithCqR
	self subclassResponsibility! !

!CogX64Compiler methodsFor: 'generate machine code' stamp: 'eem 9/22/2015 08:20'!
computeSizeOfArithCwR
	"The implementation depends on in-line or out-of-line literals."
	^self subclassResponsibility! !

!CogX64Compiler methodsFor: 'generate machine code' stamp: 'eem 4/21/2016 11:23'!
computeSizeOfPushCw
	^(self inCurrentCompilation: (operands at: 0))
		ifTrue: [9]
		ifFalse: [self pushCwByteSize]! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:47'!
concretizeAlignmentNops
	<inline: true>
	self flag: 'if performance is an issue generate longer nops'.
	0 to: machineCodeSize - 1 do:
		[:i|
		machineCode at: i put: 16r90].
	^machineCodeSize! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeArithCqRWithRO: regOpcode raxOpcode: raxOpcode
	"Will get inlined into concretizeAt: switch."
	<inline: false>
	| value reg |
	value := operands at: 0.
	reg := operands at: 1.
	machineCode
		at: 0 put: (self rexR: 0 x: 0 b: reg).
	(self isQuick: value) ifTrue:
		[machineCode
			at: 1 put: 16r83;
			at: 2 put: (self mod: ModReg RM: reg RO: regOpcode);
			at: 3 put: (value bitAnd: 16rFF).
		 ^4].
	(self is32BitSignedImmediate: value) ifTrue:
		[reg = RAX ifTrue:
			[machineCode
				at: 1 put: raxOpcode;
				at: 2 put: (value bitAnd: 16rFF);
				at: 3 put: (value >> 8 bitAnd: 16rFF);
				at: 4 put: (value >> 16 bitAnd: 16rFF);
				at: 5 put: (value >> 24 bitAnd: 16rFF).
			 ^6].
		machineCode
			at: 1 put: 16r81;
			at: 2 put: (self mod: ModReg RM: reg RO: regOpcode);
			at: 3 put: (value bitAnd: 16rFF);
			at: 4 put: (value >> 8 bitAnd: 16rFF);
			at: 5 put: (value >> 16 bitAnd: 16rFF);
			at: 6 put: (value >> 24 bitAnd: 16rFF).
		 ^7].
	^self concretizeArithCwR: (raxOpcode = 16r3D "Cmp" ifTrue: [16r39] ifFalse: [raxOpcode - 2])! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'nice 2/19/2020 01:46'!
concretizeBSR
	"Bit Scan Reverse
	First operand is input register (mask)
	Second operand is output register (dest)"
	"BSR"
	<inline: true>
	| dest maskReg |
	maskReg := operands at: 0.
	dest := operands at: 1.
	machineCode
		at: 0 put: (self rexw: true r: dest x: 0 b: maskReg);
		at: 1 put: 16r0F;
		at: 2 put: 16rBD;
		at: 3 put: (self mod: ModReg RM: maskReg RO: dest).
	 ^4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeCDQ
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	machineCode
		at: 0 put: 16r48;
		at: 1 put: 16r99.
	^2! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeCLD
	<inline: true>
	machineCode at: 0 put: 16rFC.
	^1! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 10/25/2020 17:23'!
concretizeCMPXCHGRMr
	<inline: true>
	| valueReg addressReg |
	valueReg := operands at: 0.
	addressReg := operands at: 1.
	machineCode
		"REX.W + 0F B1/r		CMPXCHG r64"
		at: 0 put: (self rexR: valueReg x: 0 b: addressReg);
		at: 1 put: 16r0f;
		at: 2 put: 16rB1;
		at: 3 put: (self mod: ModRegInd RM: addressReg RO: valueReg).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	^4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeCPUID
	<inline: true>
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rA2.
	^2! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCall
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| offset |
	self assert: (operands at: 0) ~= 0.
	offset := (operands at: 0) signedIntFromLong - (address + 5) signedIntFromLong.
	machineCode
		at: 0 put: 16rE8;
		at: 1 put: (offset bitAnd: 16rFF);
		at: 2 put: (offset >> 8 bitAnd: 16rFF);
		at: 3 put: (offset >> 16 bitAnd: 16rFF);
		at: 4 put: (offset >> 24 bitAnd: 16rFF).
	^5! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCallFull
	"Since CallFull (and JumpFull) is used to invoke code in dynamically-loaded plugins it shouldn't
	 assume that code will be loaded within 2Gb of the code zone.  Hence generate a full 64-bit call,
	 movabsq $0x123456789abcdef0, %rax; callq *%rax."
	<inline: true>
	| operand |
	operand := operands at: 0.
	machineCode
		at: 0 put: 16r48;
		at: 1 put: 16rB8;
		at: 2 put: (operand bitAnd: 16rFF);
		at: 3 put: (operand >> 8 bitAnd: 16rFF);
		at: 4 put: (operand >> 16 bitAnd: 16rFF);
		at: 5 put: (operand >> 24 bitAnd: 16rFF);
		at: 6 put: (operand >> 32 bitAnd: 16rFF);
		at: 7 put: (operand >> 40 bitAnd: 16rFF);
		at: 8 put: (operand >> 48 bitAnd: 16rFF);
		at: 9 put: (operand >> 56 bitAnd: 16rFF);
		at: 10 put: 16rFF;
		at: 11 put: (self mod: ModReg RM: RAX RO: 2).
	^12! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCallR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| reg |
	reg := operands at: 0.
	machineCode
		at: 0 put: (self rexR: 0 x: 0 b: reg);
		at: 1 put: 16rFF;
		at: 2 put: (self mod: ModReg RM: reg RO: 2).
	^3! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'nice 2/19/2020 01:44'!
concretizeClzRR
	"Count leading zeros
	First operand is output (dest)
	Second operand is input (mask)"
	"LZCNT"
	<inline: true>
	| maskReg dest  |
	maskReg := operands at: 0.
	dest := operands at: 1.
	machineCode
		at: 0 put: 16rF3;
		at: 1 put: (self rexw: true r: dest x: 0 b: maskReg);
		at: 2 put: 16r0F;
		at: 3 put: 16rBD;
		at: 4 put: (self mod: ModReg RM: maskReg RO: dest).
	 ^5! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/9/2019 19:20'!
concretizeCmpC32R
	"Will get inlined into concretizeAt: switch."
	"N.B. This use of 32-bit comparisons allows us to squeak by and use a short jump
	 in PIC case dispatch, where a jump to the abort is 126 bytes (!!!!)."
	<inline: true>
	| value reg skip |
	value := operands at: 0.
	reg := operands at: 1.
	reg = RAX
		ifTrue:
			[machineCode at: 0 put: 16r3D.
			 skip := 0]
		ifFalse:
			[reg > 7
				ifTrue:
					[machineCode at: 0 put: 16r41.
					 skip := 2]
				ifFalse:
					[skip := 1].
			 machineCode
				at: skip - 1 put: 16r81;
				at: skip put:  (self mod: ModReg RM: reg RO: 7)].
	machineCode		
		at: skip + 1 put: (value bitAnd: 16rFF);
		at: skip + 2 put: (value >> 8 bitAnd: 16rFF);
		at: skip + 3 put: (value >> 16 bitAnd: 16rFF);
		at: skip + 4 put: (value >> 24 bitAnd: 16rFF).
	 ^5 + skip! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCmpRdRd
	"Will get inlined into concretizeAt: switch.
	 We use UCOMISD (see p 4-260 [2])"
	<inline: true>
	| regLHS regRHS skip |
	"CmpRR RHS LHS computes LHS - RHS, i.e. apparently reversed.  You have to think subtract."
	regRHS := operands at: 0.
	regLHS := operands at: 1.
	
	machineCode
		at: 0 put: 16r66.
	(regLHS <= 7 and: [regRHS <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: regLHS x: 0 b: regRHS)].	
	
	machineCode
		at: skip + 1 put: 16r0F;
		at: skip + 2 put: 16r2E;
		at: skip + 3 put: (self mod: ModReg RM: regRHS RO: regLHS).
	^skip + 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeCmpRsRs
	"Will get inlined into concretizeAt: switch.
	 We use UCOMISS"
	<inline: true>
	| regLHS regRHS skip |
	"CmpRR RHS LHS computes LHS - RHS, i.e. apparently reversed.  You have to think subtract."
	regRHS := operands at: 0.
	regLHS := operands at: 1.
	
	(regLHS <= 7 and: [regRHS <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: regLHS x: 0 b: regRHS)].
		
	machineCode
		at: skip + 0 put: 16r0F;
		at: skip + 1 put: 16r2E;
		at: skip + 2 put: (self mod: ModReg RM: regRHS RO: regLHS).
	^skip + 3! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConditionalJump: conditionCode
	"Will get inlined into concretizeAt: switch."
	"Sizing/generating jumps.
		Jump targets can be to absolute addresses or other abstract instructions.
		Generating initial trampolines instructions may have no maxSize and be to absolute addresses.
		Otherwise instructions must have a machineCodeSize which must be kept to."
	<inline: true>
	| offset |
	offset := self computeJumpTargetOffsetPlus: 2.
	(machineCodeSize = 0 "size not determined because no sizeJump pass; generating initial trampolines"
		ifTrue: [self isQuick: offset]
		ifFalse: [machineCodeSize = 2]) ifTrue:
		[machineCode
			at: 0 put: 16r70 + conditionCode;
			at: 1 put: (offset bitAnd: 16rFF).
		 ^2].
	^self concretizeConditionalJumpLong: conditionCode! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConditionalJumpLong: conditionCode
	"Will get inlined into concretizeAt: switch."
	"Sizing/generating jumps.
		Jump targets can be to absolute addresses or other abstract instructions.
		Generating initial trampolines instructions may have no maxSize and be to absolute addresses.
		Otherwise instructions must have a machineCodeSize which must be kept to."
	<inline: true>
	| offset |
	offset := self computeJumpTargetOffsetPlus: 6.
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16r80 + conditionCode;
		at: 2 put: (offset bitAnd: 16rFF);
		at: 3 put: (offset >> 8 bitAnd: 16rFF);
		at: 4 put: (offset >> 16 bitAnd: 16rFF);
		at: 5 put: (offset >> 24 bitAnd: 16rFF).
	^6! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConvertRRd
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg destReg |
	srcReg := operands at:0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF2;
		at: 1 put: (self rexR: destReg x: 0 b: srcReg);
		at: 2 put: 16r0F;
		at: 3 put: 16r2A;
		at: 4 put: (self mod: ModReg RM: srcReg RO: destReg).
	 ^5! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConvertRRs
	"Will get inlined into concretizeAt: switch."
	"CVTSI2SS"
	<inline: true>
	| srcReg destReg skip |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF3.
	(srcReg <= 7 and: [destReg <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: destReg x: 0 b: srcReg)].
		
	machineCode
		at: skip + 1 put: 16r0F;
		at: skip + 2 put: 16r2A;
		at: skip + 3 put: (self mod: ModReg RM: srcReg RO: destReg).
	 ^skip + 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConvertRdR
	"Will get inlined into concretizeAt: switch."
	"CVTSD2SI"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at:0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF2;
		at: 1 put: (self rexR: destReg x: 0 b: srcReg);
		at: 2 put: 16r0F;
		at: 3 put: 16r2D;
		at: 4 put: (self mod: ModReg RM: srcReg RO: destReg).
	 ^5! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConvertRdRs
	"Will get inlined into concretizeAt: switch."
	"CVTSD2SS"
	<inline: true>
	| srcReg destReg skip |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF2.
	(srcReg <= 7 and: [destReg <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: destReg x: 0 b: srcReg)].
		
	machineCode
		at: skip + 1 put: 16r0F;
		at: skip + 2 put: 16r5A;
		at: skip + 3 put: (self mod: ModReg RM: srcReg RO: destReg).
	 ^skip + 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConvertRsR
	"Will get inlined into concretizeAt: switch."
	"CVTSS2SI"
	<inline: true>
	| srcReg destReg skip |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF3.
	(srcReg <= 7 and: [destReg <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: destReg x: 0 b: srcReg)].
		
	machineCode 
		at: skip + 1 put: 16r0F;
		at: skip + 2 put: 16r2D;
		at: skip + 3 put: (self mod: ModReg RM: srcReg RO: destReg).
	 ^skip + 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeConvertRsRd
	"Will get inlined into concretizeAt: switch."
	"CVTSS2SD"
	<inline: true>
	| srcReg destReg skip |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF3.
	(srcReg <= 7 and: [destReg <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: destReg x: 0 b: srcReg)].
		
	machineCode
		at: skip + 1 put: 16r0F;
		at: skip + 2 put: 16r5A;
		at: skip + 3 put: (self mod: ModReg RM: srcReg RO: destReg).
	 ^skip + 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeFENCE: regOpcode
	<inline: true>
	machineCode
		at: 0 put: 16r0F;
		at: 1 put: 16rAE;
		at: 2 put: (self mod: ModReg RM: 0 RO: regOpcode).
	^3! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeFill32
	<inline: true>
	| word |
	<var: #word type: #'usqIntptr_t'>
	word := operands at: 0.
	machineCode at: 0 put: (word bitAnd: 16rFF).
	machineCode at: 1 put: (word >> 8 bitAnd: 16rFF).
	machineCode at: 2 put: (word >> 16 bitAnd: 16rFF).
	machineCode at: 3 put: (word >> 24 bitAnd: 16rFF).
	^4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeIDIVR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| regDivisor |
	regDivisor := operands at: 0.
	machineCode
		at: 0 put: (self rexR: 0 x: 0 b: regDivisor);
		at: 1 put: 16rF7;
		at: 2 put: (self mod: ModReg RM: regDivisor RO: 7).
	^3! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeJump
	"Will get inlined into concretizeAt: switch."
	"Sizing/generating jumps.
		Jump targets can be to absolute addresses or other abstract instructions.
		Generating initial trampolines instructions may have no maxSize and be to absolute addresses.
		Otherwise instructions must have a machineCodeSize which must be kept to."
	<inline: true>
	| jumpTarget offset |
	<var: #jumpTarget type: #'AbstractInstruction *'>
	jumpTarget := cogit cCoerceSimple: (operands at: 0) to: #'AbstractInstruction *'.
	cogit assertSaneJumpTarget: jumpTarget.
	(self isAnInstruction: jumpTarget) ifTrue:
		[jumpTarget := cogit cCoerceSimple: jumpTarget address to: #'AbstractInstruction *'].
	self assert: jumpTarget ~= 0.
	offset := jumpTarget signedIntFromLong - (address + 2) signedIntFromLong.
	(machineCodeSize = 0 "size not determined because no sizeJump pass; generating initial trampolines"
		ifTrue: [self isQuick: offset]
		ifFalse: [machineCodeSize = 2]) ifTrue:
		[machineCode
			at: 0 put: 16rEB;
			at: 1 put: (offset bitAnd: 16rFF).
		 ^2].
	offset := jumpTarget signedIntFromLong - (address + 5) signedIntFromLong.
	machineCode
		at: 0 put: 16rE9;
		at: 1 put: (offset bitAnd: 16rFF);
		at: 2 put: (offset >> 8 bitAnd: 16rFF);
		at: 3 put: (offset >> 16 bitAnd: 16rFF);
		at: 4 put: (offset >> 24 bitAnd: 16rFF).
	^5! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeJumpFull
	"Since JumpFull (and CallFull) is used to invoke code in dynamically-loaded plugins it shouldn't
	 assume that code will be loaded within 2Gb of the code zone.  Hence generate a full 64-bit call,
	 movabsq 0x123456789abcdef0, %rax; jmpq *%rax."
	<inline: true>
	| operand |
	operand := operands at: 0.
	machineCode
		at: 0 put: 16r48;
		at: 1 put: 16rB8;
		at: 2 put: (operand bitAnd: 16rFF);
		at: 3 put: (operand >> 8 bitAnd: 16rFF);
		at: 4 put: (operand >> 16 bitAnd: 16rFF);
		at: 5 put: (operand >> 24 bitAnd: 16rFF);
		at: 6 put: (operand >> 32 bitAnd: 16rFF);
		at: 7 put: (operand >> 40 bitAnd: 16rFF);
		at: 8 put: (operand >> 48 bitAnd: 16rFF);
		at: 9 put: (operand >> 56 bitAnd: 16rFF);
		at: 10 put: 16rFF;
		at: 11 put: (self mod: ModReg RM: RAX RO: 4).
	^12! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeJumpLong
	"Will get inlined into concretizeAt: switch."
	"Sizing/generating jumps.
		Jump targets can be to absolute addresses or other abstract instructions.
		Generating initial trampolines instructions may have no maxSize and be to absolute addresses.
		Otherwise instructions must have a machineCodeSize which must be kept to."
	<inline: true>
	| jumpTarget offset |
	<var: #jumpTarget type: #'AbstractInstruction *'>
	jumpTarget := cogit cCoerceSimple: (operands at: 0) to: #'AbstractInstruction *'.
	(self isAnInstruction: jumpTarget) ifTrue:
		[jumpTarget := cogit cCoerceSimple: jumpTarget address to: #'AbstractInstruction *'].
	self assert: jumpTarget ~= 0.
	offset := jumpTarget signedIntFromLong - (address + 5) signedIntFromLong.
	machineCode
		at: 0 put: 16rE9;
		at: 1 put: (offset bitAnd: 16rFF);
		at: 2 put: (offset >> 8 bitAnd: 16rFF);
		at: 3 put: (offset >> 16 bitAnd: 16rFF);
		at: 4 put: (offset >> 24 bitAnd: 16rFF).
	^5! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeJumpR
	<inline: true>
	| reg |
	reg := operands at: 0.
	machineCode
		at: 0 put: (self rexR: 0 x: 0 b: reg);
		at: 1 put: 16rFF;
		at: 2 put: (self mod: ModReg RM: reg RO: 4).
	^3! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeLOCK
	<inline: true>
	machineCode at: 0 put: 16rF0.
	^1! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeLoadEffectiveAddressMwrR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	machineCode
		at: 0 put: (self rexR: destReg x: 0 b: srcReg);
		at: 1 put: 16r8D.
	(srcReg ~= RSP and: [srcReg ~= R12]) ifTrue:
		[(self isQuick: offset) ifTrue:
			[machineCode
				at: 2 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: 3 put: (offset bitAnd: 16rFF).
			 ^4].
		machineCode
			at: 2 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
			at: 3 put: (offset bitAnd: 16rFF);
			at: 4 put: (offset >> 8 bitAnd: 16rFF);
			at: 5 put: (offset >> 16 bitAnd: 16rFF);
			at: 6 put: (offset >> 24 bitAnd: 16rFF).
		^7].
	"ESP/R12:"
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 2 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: 3 put: (self s: SIB1 i: 4 b: srcReg);
			at: 4 put: (offset bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 2 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
		at: 3 put: (self s: SIB1 i: 4 b: srcReg);
		at: 4 put: (offset bitAnd: 16rFF);
		at: 5 put: (offset >> 8 bitAnd: 16rFF);
		at: 6 put: (offset >> 16 bitAnd: 16rFF);
		at: 7 put: (offset >> 24 bitAnd: 16rFF).
	^8! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeMOVSB
	<inline: true>
	machineCode at: 0 put: 16rA4.
	^1! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeMOVSQ
	<inline: true>
	machineCode
		at: 0 put: (self rexw: true r: 0 x: 0 b: 0);
		at: 1 put: 16rA5.
	^2! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveA32R
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| addressOperand reg offset |
	addressOperand := operands at: 0.
	(self isAnInstruction: (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *')) ifTrue:
		[addressOperand := (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *') address].
	reg := operands at: 1.
	reg = RAX
		ifTrue: [offset := 0]
		ifFalse:
			[machineCode
				at: 0 put: (self rexR: 0 x: 0 b: reg);
				at: 1 put: 16r90 + (reg \\ 8).
			 offset := 2].
	machineCode
		at: 0 + offset put: 16rA1;
		at: 1 + offset put: (addressOperand bitAnd: 16rFF);
		at: 2 + offset put: (addressOperand >> 8 bitAnd: 16rFF);
		at: 3 + offset put: (addressOperand >> 16 bitAnd: 16rFF);
		at: 4 + offset put: (addressOperand >> 24 bitAnd: 16rFF);
		at: 5 + offset put: (addressOperand >> 32 bitAnd: 16rFF);
		at: 6 + offset put: (addressOperand >> 40 bitAnd: 16rFF);
		at: 7 + offset put: (addressOperand >> 48 bitAnd: 16rFF);
		at: 8 + offset put: (addressOperand >> 56 bitAnd: 16rFF).
	reg = RAX ifTrue:
		[^9].
	machineCode
		at: 11 put: (machineCode at: 0);
		at: 12 put: (machineCode at: 1).
	^13! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 16:03'!
concretizeMoveAbR
	"N.B. The Cogit makes no assumption about the upper bits being set to zero because we
	 deny byteReadsZeroExtend.  The cogit will clear the register before hand if necessary."
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| addressOperand reg offset save0 save1 savedSize |
	addressOperand := operands at: 0.
	(self isAnInstruction: (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *')) ifTrue:
		[addressOperand := (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *') address].
	(self isAddressRelativeToVarBase: addressOperand) ifTrue:
		[save0 := operands at: 0.
		 save1 := operands at: 1.
		 operands
			at: 0 put: addressOperand - cogit varBaseAddress;
			at: 1 put: RBX;
			at: 2 put: save1.
		 savedSize := self concretizeMoveMbrR.
		 operands
			at: 0 put: save0;
			at: 1 put: save1;
			at: 2 put: 0.
		^savedSize].
	reg := operands at: 1.
	reg = RAX
		ifTrue: [offset := 0]
		ifFalse:
			[machineCode
				at: 0 put: (self rexR: 0 x: 0 b: reg);
				at: 1 put: 16r90 + (reg \\ 8).
			 offset := 2].
	machineCode
		at: 0 + offset put: 16r48;
		at: 1 + offset put: 16rA0;
		at: 2 + offset put: (addressOperand bitAnd: 16rFF);
		at: 3 + offset put: (addressOperand >> 8 bitAnd: 16rFF);
		at: 4 + offset put: (addressOperand >> 16 bitAnd: 16rFF);
		at: 5 + offset put: (addressOperand >> 24 bitAnd: 16rFF);
		at: 6 + offset put: (addressOperand >> 32 bitAnd: 16rFF);
		at: 7 + offset put: (addressOperand >> 40 bitAnd: 16rFF);
		at: 8 + offset put: (addressOperand >> 48 bitAnd: 16rFF);
		at: 9 + offset put: (addressOperand >> 56 bitAnd: 16rFF).
	reg = RAX ifTrue:
		[^10].
	machineCode
		at: 12 put: (machineCode at: 0);
		at: 13 put: (machineCode at: 1).
	^14! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 16:03'!
concretizeMoveAwR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| addressOperand reg offset save0 save1 savedSize |
	addressOperand := operands at: 0.
	(self isAnInstruction: (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *')) ifTrue:
		[addressOperand := (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *') address].
	(self isAddressRelativeToVarBase: addressOperand) ifTrue:
		[save0 := operands at: 0.
		 save1 := operands at: 1.
		 operands
			at: 0 put: addressOperand - cogit varBaseAddress;
			at: 1 put: RBX;
			at: 2 put: save1.
		 savedSize := self concretizeMoveMwrR.
		 operands
			at: 0 put: save0;
			at: 1 put: save1;
			at: 2 put: 0.
		^savedSize].
	reg := operands at: 1.
	"If fetching RAX, fetch directly, otherwise, because of instruction encoding limitations, the register
	 _must_ be fetched through RAX.  If reg = RBP or RSP simply fetch directly, otherwise swap RAX with
	 the register before and after the fetch through RAX.  We avoid swapping before hand with RBP
	 and RSP because setting RSP to whatever the contents of RAX is can cause disastrous results if
	 an interrupt is delivered immediately after that point.  See mail threads beginning with
		http://lists.squeakfoundation.org/pipermail/vm-dev/2019-September/031428.html
		http://lists.squeakfoundation.org/pipermail/vm-dev/2019-October/031499.html"
	(reg = RAX or: [reg = RBP or: [reg = RSP]])
		ifTrue: [offset := 0]
		ifFalse:
			[machineCode
				at: 0 put: (self rexR: 0 x: 0 b: reg);
				at: 1 put: 16r90 + (reg \\ 8).
			 offset := 2].
	machineCode
		at: 0 + offset put: 16r48;
		at: 1 + offset put: 16rA1;
		at: 2 + offset put: (addressOperand bitAnd: 16rFF);
		at: 3 + offset put: (addressOperand >> 8 bitAnd: 16rFF);
		at: 4 + offset put: (addressOperand >> 16 bitAnd: 16rFF);
		at: 5 + offset put: (addressOperand >> 24 bitAnd: 16rFF);
		at: 6 + offset put: (addressOperand >> 32 bitAnd: 16rFF);
		at: 7 + offset put: (addressOperand >> 40 bitAnd: 16rFF);
		at: 8 + offset put: (addressOperand >> 48 bitAnd: 16rFF);
		at: 9 + offset put: (addressOperand >> 56 bitAnd: 16rFF).
	reg = RAX ifTrue:
		[^10].
	"Now effect the assignment via xchg, which saves a byte over a move"
	(reg = RBP or: [reg = RSP]) ifTrue:
		[machineCode
			at: 10 put: (self rexR: RAX x: 0 b: reg);
			at: 11 put: 16r90 + (reg \\ 8).
		^12].
	machineCode
		at: 12 put: (machineCode at: 0);
		at: 13 put: (machineCode at: 1).
	^14! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveC32R
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value reg |
	value := operands at: 0.
	reg := operands at: 1.
	machineCode
		at: 0 put: (self rexR: 0 x: 0 b: reg);
		at: 1 put: 16rC7;
		at: 2 put: (self mod: ModReg RM: reg RO: 0);
		at: 3 put: (value bitAnd: 16rFF);
		at: 4 put: (value >> 8 bitAnd: 16rFF);
		at: 5 put: (value >> 16 bitAnd: 16rFF);
		at: 6 put: (value >> 24 bitAnd: 16rFF).
	^7! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveCqR
	"Will get inlined into concretizeAt: switch.
	 On x64 we can short-cut mov 0, reg using xor, and use signed 32-bit displacement, if possible."
	<inline: true>
	| value reg |
	value := operands at: 0.
	reg := operands at: 1.
	(self is32BitSignedImmediate: value) ifTrue:
		[value = 0 ifTrue:
			[machineCode
				at: 0 put: (self rexR: reg x: 0 b: reg);
				at: 1 put: 16r31;
				at: 2 put: (self mod: ModReg RM: reg RO: reg).
			 ^3].
		 machineCode
			at: 0 put: (self rexR: 0 x: 0 b: reg);
			at: 1 put: 16rC7;
			at: 2 put: (self mod: ModReg RM: reg RO: 0);
			at: 3 put: (value bitAnd: 16rFF);
			at: 4 put: (value >> 8 bitAnd: 16rFF);
			at: 5 put: (value >> 16 bitAnd: 16rFF);
			at: 6 put: (value >> 24 bitAnd: 16rFF).
		 ^7].

	machineCode
		at:  0 put: (self rexR: 0 x: 0 b: reg);
		at:  1 put: 16rB8 + (reg bitAnd: 7);
		at:  2 put: (value bitAnd: 16rFF);
		at:  3 put: (value >> 8 bitAnd: 16rFF);
		at:  4 put: (value >> 16 bitAnd: 16rFF);
		at:  5 put: (value >> 24 bitAnd: 16rFF);
		at:  6 put: (value >> 32 bitAnd: 16rFF);
		at:  7 put: (value >> 40 bitAnd: 16rFF);
		at:  8 put: (value >> 48 bitAnd: 16rFF);
		at:  9 put: (value >> 56 bitAnd: 16rFF).
	^10! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 9/28/2015 14:42'!
concretizeMoveCwR
	self subclassResponsibility! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveM16rR
	"N.B. The Cogit compiler makes no assumption about the upper bits being set to zero.
	 It will clear the register before hand if necessary."
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	machineCode
		at: 0 put: (self rexR: destReg x: 0 b: srcReg);
		at: 1 put: 16r0f;
		at: 2 put: 16rb7.
	(srcReg ~= RSP and: [srcReg ~= R12]) ifTrue:
		[(offset = 0 and: [srcReg ~= RBP and: [srcReg ~= R13]]) ifTrue:
			[machineCode
				at: 3 put: (self mod: ModRegInd RM: srcReg RO: destReg).
			 ^4].
		 (self isQuick: offset) ifTrue:
			[machineCode
				at: 3 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: 4 put: (offset bitAnd: 16rFF).
			 ^5].
		machineCode
			at: 3 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
			at: 4 put: (offset bitAnd: 16rFF);
			at: 5 put: (offset >> 8 bitAnd: 16rFF);
			at: 6 put: (offset >> 16 bitAnd: 16rFF);
			at: 7 put: (offset >> 24 bitAnd: 16rFF).
		^8].
	"RSP & R12:"
	(offset = 0 and: [srcReg ~= RBP and: [srcReg ~= R13]]) ifTrue:
		[machineCode
			at: 3 put: (self mod: ModRegInd RM: srcReg RO: destReg);
			at: 4 put: (self s: SIB1 i: 4 b: srcReg).
		 ^5].
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 3 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: 4 put: (self s: SIB1 i: 4 b: srcReg);
			at: 5 put: (offset bitAnd: 16rFF).
		 ^6].
	machineCode
		at: 3 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
		at: 4 put: (self s: SIB1 i: 4 b: srcReg);
		at: 5 put: (offset bitAnd: 16rFF);
		at: 6 put: (offset >> 8 bitAnd: 16rFF);
		at: 7 put: (offset >> 16 bitAnd: 16rFF);
		at: 8 put: (offset >> 24 bitAnd: 16rFF).
	^9! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveM32rR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| offset srcReg destReg skip |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	(srcReg <= 7 and: [destReg <= 7])
		ifTrue: [skip := 0]
		ifFalse: [skip := 1. machineCode at: 0 put: (self rexw: false r: destReg x: 0 b: srcReg)].
	machineCode
		at: skip + 0 put: 16r8b.
	offset = 0 ifTrue:
		[(srcReg bitAnd: 6) ~= RSP ifTrue:
			[machineCode at: skip + 1 put: (self mod: ModRegInd RM: srcReg RO: destReg).
			 ^skip + 2].
		 (srcReg bitAnd: 7) = RSP ifTrue: "RBP & R13 fall through"
			[machineCode
				at: skip + 1 put: (self mod: ModRegInd RM: srcReg RO: destReg);
				at: skip + 2 put: (self s: SIB1 i: 4 b: srcReg).
			 ^skip + 3]].
	(self isQuick: offset) ifTrue:
		[(srcReg bitAnd: 7) ~= RSP ifTrue:
			[machineCode
				at: skip + 1 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: skip + 2 put: (offset bitAnd: 16rFF).
			 ^skip + 3].
		 machineCode
			at: skip + 1 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: skip + 2 put: (self s: SIB1 i: 4 b: srcReg);
			at: skip + 3 put: (offset bitAnd: 16rFF).
		 ^skip + 4].
	machineCode at: skip + 1 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg).
	(srcReg bitAnd: 7) = RSP ifTrue:
		[machineCode at: skip + 2 put: (self s: SIB1 i: 4 b: srcReg).
		 skip := skip + 1].
	machineCode
		at: skip + 2 put: (offset bitAnd: 16rFF);
		at: skip + 3 put: (offset >> 8 bitAnd: 16rFF);
		at: skip + 4 put: (offset >> 16 bitAnd: 16rFF);
		at: skip + 5 put: (offset >> 24 bitAnd: 16rFF).
	^skip + 6! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveM32rRs
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| offset srcReg destReg skip |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	machineCode at: 0 put: 16r66.
	(srcReg <= 7 and: [destReg <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: destReg x: 0 b: srcReg)].
	machineCode
		at: skip + 1 put: 16r0f;
		at: skip + 2 put: 16r6e.
	offset = 0 ifTrue:
		[(srcReg bitAnd: 6) ~= RSP ifTrue:
			[machineCode at: skip + 3 put: (self mod: ModRegInd RM: srcReg RO: destReg).
			 ^skip + 4].
		 (srcReg bitAnd: 7) = RSP ifTrue: "RBP & R13 fall through"
			[machineCode
				at: skip + 3 put: (self mod: ModRegInd RM: srcReg RO: destReg);
				at: skip + 4 put: (self s: SIB1 i: 4 b: srcReg).
			 ^skip + 5]].
	(self isQuick: offset) ifTrue:
		[(srcReg bitAnd: 7) ~= RSP ifTrue:
			[machineCode
				at: skip + 3 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: skip + 4 put: (offset bitAnd: 16rFF).
			 ^skip + 5].
		 machineCode
			at: skip + 3 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: skip + 4 put: (self s: SIB1 i: 4 b: srcReg);
			at: skip + 5 put: (offset bitAnd: 16rFF).
		 ^skip + 6].
	machineCode at: skip + 3 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg).
	(srcReg bitAnd: 7) = RSP ifTrue:
		[machineCode at: skip + 4 put: (self s: SIB1 i: 4 b: srcReg).
		 skip := skip + 1].
	machineCode
		at: skip + 4 put: (offset bitAnd: 16rFF);
		at: skip + 5 put: (offset >> 8 bitAnd: 16rFF);
		at: skip + 6 put: (offset >> 16 bitAnd: 16rFF);
		at: skip + 7 put: (offset >> 24 bitAnd: 16rFF).
	^skip + 8! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveM64rRd
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| offset srcReg destReg skip |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	machineCode at: 0 put: 16rF3.
	(srcReg <= 7 and: [destReg <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: destReg x: 0 b: srcReg)].
	machineCode
		at: skip + 1 put: 16r0f;
		at: skip + 2 put: 16r7e.
	offset = 0 ifTrue:
		[(srcReg bitAnd: 6) ~= RSP ifTrue:
			[machineCode at: skip + 3 put: (self mod: ModRegInd RM: srcReg RO: destReg).
			 ^skip + 4].
		 (srcReg bitAnd: 7) = RSP ifTrue: "RBP & R13 fall through"
			[machineCode
				at: skip + 3 put: (self mod: ModRegInd RM: srcReg RO: destReg);
				at: skip + 4 put: (self s: SIB1 i: 4 b: srcReg).
			 ^skip + 5]].
	(self isQuick: offset) ifTrue:
		[(srcReg bitAnd: 7) ~= RSP ifTrue:
			[machineCode
				at: skip + 3 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: skip + 4 put: (offset bitAnd: 16rFF).
			 ^skip + 5].
		 machineCode
			at: skip + 3 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: skip + 4 put: (self s: SIB1 i: 4 b: srcReg);
			at: skip + 5 put: (offset bitAnd: 16rFF).
		 ^skip + 6].
	machineCode at: skip + 3 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg).
	(srcReg bitAnd: 7) = RSP ifTrue:
		[machineCode at: skip + 4 put: (self s: SIB1 i: 4 b: srcReg).
		 skip := skip + 1].
	machineCode
		at: skip + 4 put: (offset bitAnd: 16rFF);
		at: skip + 5 put: (offset >> 8 bitAnd: 16rFF);
		at: skip + 6 put: (offset >> 16 bitAnd: 16rFF);
		at: skip + 7 put: (offset >> 24 bitAnd: 16rFF).
	^skip + 8! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveMbrR
	"N.B. The Cogit makes no assumption about the upper bits being set to zero because we
	 deny byteReadsZeroExtend.  The cogit will clear the register before hand if necessary."
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	machineCode
		at: 0 put: (self rexR: destReg x: 0 b: srcReg);
		at: 1 put: 16r8A.
	(srcReg ~= RSP and: [srcReg ~= R12]) ifTrue:
		[(offset = 0 and: [srcReg ~= RBP and: [srcReg ~= R13]]) ifTrue:
			[machineCode
				at: 2 put: (self mod: ModRegInd RM: srcReg RO: destReg).
			 ^3].
		(self isQuick: offset) ifTrue:
			[machineCode
				at: 2 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: 3 put: (offset bitAnd: 16rFF).
			 ^4].
		machineCode
			at: 2 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
			at: 3 put: (offset bitAnd: 16rFF);
			at: 4 put: (offset >> 8 bitAnd: 16rFF);
			at: 5 put: (offset >> 16 bitAnd: 16rFF);
			at: 6 put: (offset >> 24 bitAnd: 16rFF).
		^7].
	"RSP:"
	offset = 0 ifTrue:
		[machineCode
			at: 2 put: (self mod: ModRegInd RM: srcReg RO: destReg);
			at: 3 put: (self s: SIB1 i: 4 b: srcReg).
		 ^4].
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 2 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: 3 put: (self s: SIB1 i: 4 b: srcReg);
			at: 4 put: (offset bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 2 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
		at: 3 put: (self s: SIB1 i: 4 b: srcReg);
		at: 4 put: (offset bitAnd: 16rFF);
		at: 5 put: (offset >> 8 bitAnd: 16rFF);
		at: 6 put: (offset >> 16 bitAnd: 16rFF);
		at: 7 put: (offset >> 24 bitAnd: 16rFF).
	^8! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveMwrR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	offset := operands at: 0.
	srcReg := operands at: 1.
	destReg := operands at: 2.
	machineCode
		at: 0 put: (self rexR: destReg x: 0 b: srcReg);
		at: 1 put: 16r8B.
	(srcReg ~= RSP and: [srcReg ~= R12]) ifTrue:
		[(offset = 0 and: [srcReg ~= RBP and: [srcReg ~= R13]]) ifTrue:
			[machineCode
				at: 2 put: (self mod: ModRegInd RM: srcReg RO: destReg).
			 ^3].
		(self isQuick: offset) ifTrue:
			[machineCode
				at: 2 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
				at: 3 put: (offset bitAnd: 16rFF).
			 ^4].
		machineCode
			at: 2 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
			at: 3 put: (offset bitAnd: 16rFF);
			at: 4 put: (offset >> 8 bitAnd: 16rFF);
			at: 5 put: (offset >> 16 bitAnd: 16rFF);
			at: 6 put: (offset >> 24 bitAnd: 16rFF).
		^7].
	"RSP:"
	offset = 0 ifTrue:
		[machineCode
			at: 2 put: (self mod: ModRegInd RM: srcReg RO: destReg);
			at: 3 put: (self s: SIB1 i: 4 b: srcReg).
		 ^4].
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 2 put: (self mod: ModRegRegDisp8 RM: srcReg RO: destReg);
			at: 3 put: (self s: SIB1 i: 4 b: srcReg);
			at: 4 put: (offset bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 2 put: (self mod: ModRegRegDisp32 RM: srcReg RO: destReg);
		at: 3 put: (self s: SIB1 i: 4 b: srcReg);
		at: 4 put: (offset bitAnd: 16rFF);
		at: 5 put: (offset >> 8 bitAnd: 16rFF);
		at: 6 put: (offset >> 16 bitAnd: 16rFF);
		at: 7 put: (offset >> 24 bitAnd: 16rFF).
	^8! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 8/8/2021 10:46'!
concretizeMovePerfCnt64RL
	"Generate code for
		0x0: 50					pushq	%rax
		0x1: 52					pushq	%rdx
		0x2: 0f 31				rdtsc
		0x4: 48 c1 e2 20		shlq	$0x20, %rdx
		0x8: 48 09 d0			orq		%rdx, %rax
		0xb: 48 89 f8			movq	%rdi, %rax
		0xe: 5a					popq	%rdx
		0xf: 58					popq	%rax
	 et al"
	| reg liveRegisterMask offset |
	reg := operands at: 0.
	liveRegisterMask := operands at: 1.
	offset := 0.
	(liveRegisterMask anyMask: (cogit registerMaskFor: RAX)) ifTrue:
		[machineCode at: 0 put: 16r50. "push %eax"
		 offset := offset + 1].
	(liveRegisterMask anyMask: (cogit registerMaskFor: RDX)) ifTrue:
		[machineCode at: offset put: 16r52. "push %edx"
		 offset := offset + 1].
	"too lazy to define the swap cases for the moment..."
	self deny: reg = RDX.
	machineCode
		at: offset		put: 16r0F;							"rdtsc"
		at: offset + 1	put: 16r31;
		at: offset + 2	put: (self rexR: 0 x: 0 b: RDX);					"shlq   $0x20, %rdx"
		at: offset + 3	put: 16rC1;
		at: offset + 4	put: (self mod: ModReg RM: RDX RO: 4);
		at: offset + 5	put: 32;
		at: offset + 6	put: (self rexR: RDX x: 0 b: RAX);				"orq 	%rax, %rdx"
		at: offset + 7	put: 16r0B;
		at: offset + 8	put: (self mod: ModReg RM: RAX RO: RDX).
	offset := offset + 9.
	reg ~= RAX ifTrue:
		[machineCode
			at: offset put: (self rexR: RAX x: 0 b: reg);				"movq	%rDEST, %rax"
			at: offset + 1 put: 16r89;
			at: offset + 2 put: (self mod: ModReg RM: reg RO: RAX).
		 offset := offset + 3].
	(liveRegisterMask anyMask: (cogit registerMaskFor: RDX)) ifTrue:
		[machineCode at: offset put: 16r5A. "pop %edx"
		 offset := offset + 1].
	(liveRegisterMask anyMask: (cogit registerMaskFor: RAX)) ifTrue:
		[machineCode at: offset put: 16r58. "pop %eax"
		 offset := offset + 1].
	^offset

	"{	cogit processor disassembleInstructionAt: 0 In: machineCode object.
		cogit processor disassembleInstructionAt: 2 In: machineCode object.
		cogit processor disassembleInstructionAt: 6 In: machineCode object.
		cogit processor disassembleInstructionAt: 9 In: machineCode object }"! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRA32
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| addressOperand reg offset |
	reg := operands at: 0.
	addressOperand := operands at: 1.
	(self isAnInstruction: (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *')) ifTrue:
		[addressOperand := (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *') address].
	reg = RAX
		ifTrue: [offset := 0]
		ifFalse:
			[machineCode
				at: 0 put: (self rexR: 0 x: 0 b: reg);
				at: 1 put: 16r90 + (reg \\ 8).
			 offset := 2].
	machineCode
		at: 0 + offset put: 16rA3;
		at: 1 + offset put: (addressOperand bitAnd: 16rFF);
		at: 2 + offset put: (addressOperand >>   8 bitAnd: 16rFF);
		at: 3 + offset put: (addressOperand >> 16 bitAnd: 16rFF);
		at: 4 + offset put: (addressOperand >> 24 bitAnd: 16rFF);
		at: 5 + offset put: (addressOperand >> 32 bitAnd: 16rFF);
		at: 6 + offset put: (addressOperand >> 40 bitAnd: 16rFF);
		at: 7 + offset put: (addressOperand >> 48 bitAnd: 16rFF);
		at: 8 + offset put: (addressOperand >> 56 bitAnd: 16rFF).
	reg = RAX ifTrue:
		[^9].
	machineCode
		at: 11 put: (machineCode at: 0);
		at: 12 put: (machineCode at: 1).
	^13! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 16:04'!
concretizeMoveRAb
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| addressOperand reg offset save1 savedSize |
	reg := operands at: 0.
	addressOperand := operands at: 1.
	(self isAnInstruction: (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *')) ifTrue:
		[addressOperand := (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *') address].
	(self isAddressRelativeToVarBase: addressOperand) ifTrue:
		[save1 := operands at: 1.
		 operands
			at: 1 put: addressOperand - cogit varBaseAddress;
			at: 2 put: RBX.
		 savedSize := self concretizeMoveRMbr.
		 operands
			at: 1 put: save1;
			at: 2 put: 0.
		^savedSize].
	reg = RAX
		ifTrue: [offset := 0]
		ifFalse:
			[machineCode
				at: 0 put: (self rexR: 0 x: 0 b: reg);
				at: 1 put: 16r90 + (reg \\ 8).
			 offset := 2].
	machineCode
		at: 0 + offset put: 16r48;
		at: 1 + offset put: 16rA2;
		at: 2 + offset put: (addressOperand bitAnd: 16rFF);
		at: 3 + offset put: (addressOperand >> 8 bitAnd: 16rFF);
		at: 4 + offset put: (addressOperand >> 16 bitAnd: 16rFF);
		at: 5 + offset put: (addressOperand >> 24 bitAnd: 16rFF);
		at: 6 + offset put: (addressOperand >> 32 bitAnd: 16rFF);
		at: 7 + offset put: (addressOperand >> 40 bitAnd: 16rFF);
		at: 8 + offset put: (addressOperand >> 48 bitAnd: 16rFF);
		at: 9 + offset put: (addressOperand >> 56 bitAnd: 16rFF).
	reg = RAX ifTrue:
		[^10].
	machineCode
		at: 12 put: (machineCode at: 0);
		at: 13 put: (machineCode at: 1).
	^14! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 10/29/2020 13:19'!
concretizeMoveRAw
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| addressOperand reg offset save1 savedSize |
	reg := operands at: 0.
	addressOperand := operands at: 1.
	(self isAnInstruction: (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *')) ifTrue:
		[addressOperand := (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *') address].
	(self isAddressRelativeToVarBase: addressOperand) ifTrue:
		[save1 := operands at: 1.
		 operands
			at: 1 put: addressOperand - cogit varBaseAddress;
			at: 2 put: RBX.
		 savedSize := self concretizeMoveRMwr.
		 operands
			at: 1 put: save1;
			at: 2 put: 0.
		^savedSize].
	"If storing RAX, store directly, otherwise, because of instruction encoding limitations, the register
	 _must_ be stored through RAX.  If reg = RBP or RSP simply store directly, otherwise swap RAX with
	 the register before and after the store through RAX.  We avoid swapping before hand with RBP
	 and RSP because setting RSP to whatever the contents of RAX is can cause disastrous results if
	 an interrupt is delivered immediately after that point.  See mail threads beginning with
		http://lists.squeakfoundation.org/pipermail/vm-dev/2019-September/031428.html
		http://lists.squeakfoundation.org/pipermail/vm-dev/2019-October/031499.html"
	(reg = RAX or: [reg = RBP or: [reg = RSP]])
		ifTrue: [offset := 0]
		ifFalse:
			[(reg = RBP or: [reg = RSP])
				ifTrue:
					[machineCode
						at: 0 put: (self rexR: reg x: 0 b: RAX);
						at: 1 put: 16r89;
						at: 2 put: (self mod: ModReg RM: RAX RO: reg).
					 offset := 3]
				ifFalse:
					[machineCode
						at: 0 put: (self rexR: RAX x: 0 b: reg);
						at: 1 put: 16r90 + (reg \\ 8).
					 offset := 2]].
	machineCode
		at: 0 + offset put: 16r48;
		at: 1 + offset put: 16rA3;
		at: 2 + offset put: (addressOperand bitAnd: 16rFF);
		at: 3 + offset put: (addressOperand >> 8 bitAnd: 16rFF);
		at: 4 + offset put: (addressOperand >> 16 bitAnd: 16rFF);
		at: 5 + offset put: (addressOperand >> 24 bitAnd: 16rFF);
		at: 6 + offset put: (addressOperand >> 32 bitAnd: 16rFF);
		at: 7 + offset put: (addressOperand >> 40 bitAnd: 16rFF);
		at: 8 + offset put: (addressOperand >> 48 bitAnd: 16rFF);
		at: 9 + offset put: (addressOperand >> 56 bitAnd: 16rFF).
	reg = RAX ifTrue:
		[^10].
	(reg = RBP or: [reg = RSP]) ifTrue:
		[^13].
	"Now effect the assignment via xchg, which restores RAX"
	machineCode
		at: 12 put: (machineCode at: 0);
		at: 13 put: (machineCode at: 1).
	^14! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 10/24/2020 19:44'!
concretizeMoveRAwNoVBR
	"A version of concretizeMoveRAw tat does not use VarBaseReg."
	<inline: true>
	| addressOperand reg offset |
	reg := operands at: 0.
	addressOperand := operands at: 1.
	(self isAnInstruction: (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *')) ifTrue:
		[addressOperand := (cogit cCoerceSimple: addressOperand to: #'AbstractInstruction *') address].
	"If storing RAX, store directly, otherwise, because of instruction encoding limitations, the register
	 _must_ be stored through RAX.  If reg = RBP or RSP simply store directly, otherwise swap RAX with
	 the register before and after the store through RAX.  We avoid sweapping before hand with RBP
	 and RSP because setting RSP to whatever the contents of RAX is can cause disastrous results if
	 an interrupt is delivered immediately after that point.  See mail threads beginning with
		http://lists.squeakfoundation.org/pipermail/vm-dev/2019-September/031428.html
		http://lists.squeakfoundation.org/pipermail/vm-dev/2019-October/031499.html"
	(reg = RAX or: [reg = RBP or: [reg = RSP]])
		ifTrue: [offset := 0]
		ifFalse:
			[(reg = RBP or: [reg = RSP])
				ifTrue:
					[machineCode
						at: 0 put: (self rexR: reg x: 0 b: RAX);
						at: 1 put: 16r89;
						at: 2 put: (self mod: ModReg RM: RAX RO: reg).
					 offset := 3]
				ifFalse:
					[machineCode
						at: 0 put: (self rexR: RAX x: 0 b: reg);
						at: 1 put: 16r90 + (reg \\ 8).
					 offset := 2]].
	machineCode
		at: 0 + offset put: 16r48;
		at: 1 + offset put: 16rA3;
		at: 2 + offset put: (addressOperand bitAnd: 16rFF);
		at: 3 + offset put: (addressOperand >> 8 bitAnd: 16rFF);
		at: 4 + offset put: (addressOperand >> 16 bitAnd: 16rFF);
		at: 5 + offset put: (addressOperand >> 24 bitAnd: 16rFF);
		at: 6 + offset put: (addressOperand >> 32 bitAnd: 16rFF);
		at: 7 + offset put: (addressOperand >> 40 bitAnd: 16rFF);
		at: 8 + offset put: (addressOperand >> 48 bitAnd: 16rFF);
		at: 9 + offset put: (addressOperand >> 56 bitAnd: 16rFF).
	reg = RAX ifTrue:
		[^10].
	(reg = RBP or: [reg = RSP]) ifTrue:
		[^13].
	"Now effect the assignment via xchg, which restores RAX"
	machineCode
		at: 12 put: (machineCode at: 0);
		at: 13 put: (machineCode at: 1).
	^14! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRM16r
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg skip |
	srcReg := operands at: 0.
	offset := operands at: 1.
	destReg := operands at: 2.
	machineCode at: 0 put: 16r66.
	(srcReg > 7 or: [destReg > 7])
		ifTrue:
			[machineCode at: 1 put: (self rexw: false r: srcReg x: 0 b: destReg).
			 skip := 1]
		ifFalse:
			[skip := 0].
	(destReg bitAnd: 7) ~= RSP ifTrue:
		[(self isQuick: offset) ifTrue:
			[machineCode
				at: skip + 1 put: 16r89;
				at: skip + 2 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
				at: skip + 3 put: (offset bitAnd: 16rFF).
			 ^skip + 4].
		machineCode
			at: skip + 1 put: 16r89;
			at: skip + 2 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
			at: skip + 3 put: (offset bitAnd: 16rFF);
			at: skip + 4 put: (offset >> 8 bitAnd: 16rFF);
			at: skip + 5 put: (offset >> 16 bitAnd: 16rFF);
			at: skip + 6 put: (offset >> 24 bitAnd: 16rFF).
		^skip + 7].
	"RSP:"
	(self isQuick: offset) ifTrue:
		[machineCode
			at: skip + 1 put: 16r89;
			at: skip + 2 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
			at: skip + 3 put: (self s: SIB1 i: 4 b: destReg);
			at: skip + 4 put: (offset bitAnd: 16rFF).
		 ^skip + 5].
	machineCode
		at: skip + 1 put: 16r89;
		at: skip + 2 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
		at: skip + 3 put: (self s: SIB1 i: 4 b: destReg);
		at: skip + 4 put: (offset bitAnd: 16rFF);
		at: skip + 5 put: (offset >> 8 bitAnd: 16rFF);
		at: skip + 6 put: (offset >> 16 bitAnd: 16rFF);
		at: skip + 7 put: (offset >> 24 bitAnd: 16rFF).
	^skip + 8! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRM32r
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| offset srcReg destReg skip |
	srcReg := operands at: 0.
	offset := operands at: 1.
	destReg := operands at: 2.
	(srcReg <= 7 and: [destReg <= 7])
		ifTrue: [skip := 0]
		ifFalse: [skip := 1. machineCode at: 0 put: (self rexw: false r: srcReg x: 0 b: destReg)].
	machineCode
		at: skip + 0 put: 16r89.
	offset = 0 ifTrue:
		[(destReg bitAnd: 6) ~= RSP ifTrue:
			[machineCode at: skip + 1 put: (self mod: ModRegInd RM: destReg RO: srcReg).
			 ^skip + 2].
		 (destReg bitAnd: 7) = RSP ifTrue: "RBP & R13 fall through"
			[machineCode
				at: skip + 1 put: (self mod: ModRegInd RM: destReg RO: srcReg);
				at: skip + 2 put: (self s: SIB1 i: 4 b: destReg).
			 ^skip + 3]].
	(self isQuick: offset) ifTrue:
		[(destReg bitAnd: 7) ~= RSP ifTrue:
			[machineCode
				at: skip + 1 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
				at: skip + 2 put: (offset bitAnd: 16rFF).
			 ^skip + 3].
		 machineCode
			at: skip + 1 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
			at: skip + 2 put: (self s: SIB1 i: 4 b: destReg);
			at: skip + 3 put: (offset bitAnd: 16rFF).
		 ^skip + 4].
	machineCode at: skip + 1 put: (self mod: ModRegRegDisp32  RM: destReg RO: srcReg).
	(destReg bitAnd: 7) = RSP ifTrue:
		[machineCode at: skip + 2 put: (self s: SIB1 i: 4 b: destReg).
		 skip := skip + 1].
	machineCode
		at: skip + 2 put: (offset bitAnd: 16rFF);
		at: skip + 3 put: (offset >> 8 bitAnd: 16rFF);
		at: skip + 4 put: (offset >> 16 bitAnd: 16rFF);
		at: skip + 5 put: (offset >> 24 bitAnd: 16rFF).
	^skip + 6! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRMbr
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset baseReg |
	srcReg := operands at: 0.
	offset := operands at: 1.
	baseReg := operands at: 2.
	machineCode
		at: 0 put: (self rexR: srcReg x: 0 b: baseReg);
		at: 1 put: 16r88.
	(baseReg ~= RSP and: [baseReg ~= R12]) ifTrue:
		[(offset = 0 and: [baseReg ~= RBP and: [baseReg ~= R13]]) ifTrue:
			[machineCode
				at: 2 put: (self mod: ModRegInd RM: baseReg RO: srcReg).
			 ^3].
		 (self isQuick: offset) ifTrue:
			[machineCode
				at: 2 put: (self mod: ModRegRegDisp8 RM: baseReg RO: srcReg);
				at: 3 put: (offset bitAnd: 16rFF).
			 ^4].
		machineCode
			at: 2 put: (self mod: ModRegRegDisp32 RM: baseReg RO: srcReg);
			at: 3 put: (offset bitAnd: 16rFF);
			at: 4 put: (offset >> 8 bitAnd: 16rFF);
			at: 5 put: (offset >> 16 bitAnd: 16rFF);
			at: 6 put: (offset >> 24 bitAnd: 16rFF).
		^7].
	"RSP:"
	offset = 0 ifTrue:
		[machineCode
			at: 2 put: (self mod: ModRegInd RM: baseReg RO: srcReg);
			at: 3 put: (self s: SIB1 i: 4 b: baseReg).
		 ^4].
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 2 put: (self mod: ModRegRegDisp8 RM: baseReg RO: srcReg);
			at: 3 put: (self s: SIB1 i: 4 b: baseReg);
			at: 4 put: (offset bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 2 put: (self mod: ModRegRegDisp32 RM: baseReg RO: srcReg);
		at: 3 put: (self s: SIB1 i: 4 b: baseReg);
		at: 4 put: (offset bitAnd: 16rFF);
		at: 5 put: (offset >> 8 bitAnd: 16rFF);
		at: 6 put: (offset >> 16 bitAnd: 16rFF);
		at: 7 put: (offset >> 24 bitAnd: 16rFF).
	^8! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRMwr
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| srcReg offset destReg |
	srcReg := operands at: 0.
	offset := operands at: 1.
	destReg := operands at: 2.
	machineCode
		at: 0 put: (self rexR: srcReg x: 0 b: destReg);
		at: 1 put: 16r89.
	(destReg ~= RSP and: [destReg ~= R12]) ifTrue:
		[(offset = 0 and: [destReg ~= RBP and: [destReg ~= R13]]) ifTrue:
			[machineCode
				at: 2 put: (self mod: ModRegInd RM: destReg RO: srcReg).
			 ^3].
		 (self isQuick: offset) ifTrue:
			[machineCode
				at: 2 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
				at: 3 put: (offset bitAnd: 16rFF).
			 ^4].
		machineCode
			at: 2 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
			at: 3 put: (offset bitAnd: 16rFF);
			at: 4 put: (offset >> 8 bitAnd: 16rFF);
			at: 5 put: (offset >> 16 bitAnd: 16rFF);
			at: 6 put: (offset >> 24 bitAnd: 16rFF).
		^7].
	"RSP:"
	offset = 0 ifTrue:
		[machineCode
			at: 2 put: (self mod: ModRegInd RM: destReg RO: srcReg);
			at: 3 put: (self s: SIB1 i: 4 b: destReg).
		 ^4].
	(self isQuick: offset) ifTrue:
		[machineCode
			at: 2 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
			at: 3 put: (self s: SIB1 i: 4 b: destReg);
			at: 4 put: (offset bitAnd: 16rFF).
		 ^5].
	machineCode
		at: 2 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg);
		at: 3 put: (self s: SIB1 i: 4 b: destReg);
		at: 4 put: (offset bitAnd: 16rFF);
		at: 5 put: (offset >> 8 bitAnd: 16rFF);
		at: 6 put: (offset >> 16 bitAnd: 16rFF);
		at: 7 put: (offset >> 24 bitAnd: 16rFF).
	^8! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRRd
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16r66;
		at: 1 put: (self rexR: destReg x: 0 b: srcReg);
		at: 2 put: 16r0f;
		at: 3 put: 16r6e;
		at: 4 put: (self mod: ModReg RM: srcReg RO: destReg).
	^5! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRX32rR
	| index base src offset |
	src := operands at: 0.
	index := operands at: 1.
	base := operands at: 2.
	(index > 7 or: [base > 7 or: [src > 7]])
		ifTrue:
			[machineCode at: 0 put: (self rexw: false r: src x: index b: base).
			 offset := 1]
		ifFalse:
			[offset := 0].
	(base bitAnd: 7) ~= RBP ifTrue:
		[machineCode
			at: offset + 0 put: 16r89;
			at: offset + 1 put: (self mod: ModRegInd RM: 4 RO: src);
			at: offset + 2 put: (self s: SIB4 i: index b: base).
		 ^offset + 3].
	machineCode
		at: offset + 0 put: 16r89;
		at: offset + 1 put: (self mod: ModRegRegDisp8 RM: 4 RO: src);
		at: offset + 2 put: (self s: SIB4 i: index b: base);
		at: offset + 3 put: 0.
	 ^offset + 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRXbrR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| src index base offset |
	src := operands at: 0.
	index := operands at: 1.
	base := operands at: 2.
	offset := 0.
	(src > 3 or: [base > 7 or: [index > 7]]) ifTrue:
		[machineCode at: 0 put: (self rexR: src x: index b: base).
		 offset := 1].
	machineCode at: 0 + offset put: 16r88.
	(base bitAnd: 7) ~= RBP "RBP,R13" ifTrue:
		[machineCode
			at: 1 + offset put: (self mod: ModRegInd RM: 4 RO: src);
			at: 2 + offset put: (self s: SIB1 i: index b: base).
		 ^3 + offset].
	machineCode
		at: 1 + offset put: (self mod: ModRegRegDisp8 RM: 4 RO: src);
		at: 2 + offset put: (self s: SIB1 i: index b: base);
		at: 3 + offset put: 0.
	 ^4 + offset! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRXwrR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| index base src |
	src := operands at: 0.
	index := operands at: 1.
	base := operands at: 2.
	machineCode
		at: 0 put: (self rexR: src x: index b: base).
	(base ~= RBP and: [base ~= R13]) ifTrue:
		[machineCode
			at: 1 put: 16r89;
			at: 2 put: (self mod: ModRegInd RM: 4 RO: src);
			at: 3 put: (self s: SIB8 i: index b: base).
		 ^4].
	machineCode
		at: 1 put: 16r89;
		at: 2 put: (self mod: ModRegRegDisp8 RM: 4 RO: src);
		at: 3 put: (self s: SIB8 i: index b: base);
		at: 4 put: 0.
	 ^5! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRdM64r
	<inline: true>
	| offset srcReg destReg skip |
	srcReg := operands at: 0.
	offset := operands at: 1.
	destReg := operands at: 2.
	machineCode at: 0 put: 16r66.
	(srcReg <= 7 and: [destReg <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: srcReg x: 0 b: destReg)].
	machineCode
		at: skip + 1 put: 16r0f;
		at: skip + 2 put: 16rd6.
	offset = 0 ifTrue:
		[(destReg bitAnd: 6) ~= RSP ifTrue:
			[machineCode at: skip + 3 put: (self mod: ModRegInd RM: destReg RO: srcReg).
			 ^skip + 4].
		 (destReg bitAnd: 7) = RSP ifTrue: "RBP & R13 fall through"
			[machineCode
				at: skip + 3 put: (self mod: ModRegInd RM: destReg RO: srcReg);
				at: skip + 4 put: (self s: SIB1 i: 4 b: destReg).
			 ^skip + 5]].
	(self isQuick: offset) ifTrue:
		[(destReg bitAnd: 7) ~= RSP ifTrue:
			[machineCode
				at: skip + 3 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
				at: skip + 4 put: (offset bitAnd: 16rFF).
			 ^skip + 5].
		 machineCode
			at: skip + 3 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
			at: skip + 4 put: (self s: SIB1 i: 4 b: destReg);
			at: skip + 5 put: (offset bitAnd: 16rFF).
		 ^skip + 6].
	machineCode at: skip + 3 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg).
	(destReg bitAnd: 7) = RSP ifTrue:
		[machineCode at: skip + 4 put: (self s: SIB1 i: 4 b: destReg).
		 skip := skip + 1].
	machineCode
		at: skip + 4 put: (offset bitAnd: 16rFF);
		at: skip + 5 put: (offset >> 8 bitAnd: 16rFF);
		at: skip + 6 put: (offset >> 16 bitAnd: 16rFF);
		at: skip + 7 put: (offset >> 24 bitAnd: 16rFF).
	^skip + 8! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRdR
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16r66;
		at: 1 put: (self rexR: srcReg x: 0 b: destReg);
		at: 2 put: 16r0f;
		at: 3 put: 16r7e;
		at: 4 put: (self mod: ModReg RM: destReg RO: srcReg).
	^5! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRdRd
	"Will get inlined into concretizeAt: switch."
	"MOVSD"
	<inline: true>
	| srcReg destReg skip |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF2.
	(srcReg <= 7 and: [destReg <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: srcReg x: 0 b: destReg)].
	machineCode
		at: skip + 1 put: 16r0F;
		at: skip + 2 put: 16r11;
		at: skip + 3 put: (self mod: ModReg RM: destReg RO: srcReg).
	^skip + 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRsM32r
	<inline: true>
	| offset srcReg destReg skip |
	srcReg := operands at: 0.
	offset := operands at: 1.
	destReg := operands at: 2.
	machineCode at: 0 put: 16r66.
	(srcReg <= 7 and: [destReg <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: srcReg x: 0 b: destReg)].
	machineCode
		at: skip + 1 put: 16r0f;
		at: skip + 2 put: 16r7e.
	offset = 0 ifTrue:
		[(destReg bitAnd: 6) ~= RSP ifTrue:
			[machineCode at: skip + 3 put: (self mod: ModRegInd RM: destReg RO: srcReg).
			 ^skip + 4].
		 (destReg bitAnd: 7) = RSP ifTrue: "RBP & R13 fall through"
			[machineCode
				at: skip + 3 put: (self mod: ModRegInd RM: destReg RO: srcReg);
				at: skip + 4 put: (self s: SIB1 i: 4 b: destReg).
			 ^skip + 5]].
	(self isQuick: offset) ifTrue:
		[(destReg bitAnd: 7) ~= RSP ifTrue:
			[machineCode
				at: skip + 3 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
				at: skip + 4 put: (offset bitAnd: 16rFF).
			 ^skip + 5].
		 machineCode
			at: skip + 3 put: (self mod: ModRegRegDisp8 RM: destReg RO: srcReg);
			at: skip + 4 put: (self s: SIB1 i: 4 b: destReg);
			at: skip + 5 put: (offset bitAnd: 16rFF).
		 ^skip + 6].
	machineCode at: skip + 3 put: (self mod: ModRegRegDisp32 RM: destReg RO: srcReg).
	(destReg bitAnd: 7) = RSP ifTrue:
		[machineCode at: skip + 4 put: (self s: SIB1 i: 4 b: destReg).
		 skip := skip + 1].
	machineCode
		at: skip + 4 put: (offset bitAnd: 16rFF);
		at: skip + 5 put: (offset >> 8 bitAnd: 16rFF);
		at: skip + 6 put: (offset >> 16 bitAnd: 16rFF);
		at: skip + 7 put: (offset >> 24 bitAnd: 16rFF).
	^skip + 8! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveRsRs
	"Will get inlined into concretizeAt: switch."
	"MOVSS"
	<inline: true>
	| srcReg destReg skip |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: 16rF3.
	(srcReg <= 7 and: [destReg <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: srcReg x: 0 b: destReg)].
	machineCode
		at: skip + 1 put: 16r0F;
		at: skip + 2 put: 16r11;
		at: skip + 3 put: (self mod: ModReg RM: destReg RO: srcReg).
	^skip + 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveX32rRR
	"MoveX32rRR is expected to zero-extend, so explicitly zero the destination."
	| index base dest offset |
	index := operands at: 0.
	base := operands at: 1.
	dest := operands at: 2.
	machineCode
		at: 0 put: (self rexR: dest x: 0 b: dest);
		at: 1 put: 16r31;
		at: 2 put: (self mod: ModReg RM: dest RO: dest).
	(index > 7 or: [base > 7 or: [dest > 7]])
		ifTrue:
			[machineCode at: 3 put: (self rexw: false r: dest x: index b: base).
			 offset := 1]
		ifFalse:
			[offset := 0].
	(base bitAnd: 7) ~= RBP ifTrue:
		[machineCode
			at: offset + 3 put: 16r8B;
			at: offset + 4 put: (self mod: ModRegInd RM: 4 RO: dest);
			at: offset + 5 put: (self s: SIB4 i: index b: base).
		 ^offset + 6].
	machineCode
		at: offset + 3 put: 16r8B;
		at: offset + 4 put: (self mod: ModRegRegDisp8 RM: 4 RO: dest);
		at: offset + 5 put: (self s: SIB4 i: index b: base);
		at: offset + 6 put: 0.
	 ^offset + 7! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveXbrRR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| index base dest |
	index := operands at: 0.
	base := operands at: 1.
	dest := operands at: 2.
	machineCode
		at: 0 put: (self rexR: dest x: index b: base);
		at: 1 put: 16r8A.
	(base ~= RBP and: [base ~= R13]) ifTrue:
		[machineCode
			at: 2 put: (self mod: ModRegInd RM: 4 RO: dest);
			at: 3 put: (self s: SIB1 i: index b: base).
		 ^4].
	machineCode
		at: 2 put: (self mod: ModRegRegDisp8 RM: 4 RO: dest);
		at: 3 put: (self s: SIB1 i: index b: base);
		at: 4 put: 0.
	 ^5! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMoveXwrRR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| index base dest |
	index := operands at: 0.
	base := operands at: 1.
	dest := operands at: 2.
	machineCode
		at: 0 put: (self rexR: dest x: index b: base).
	(base ~= RBP and: [base ~= R13]) ifTrue:
		[machineCode
			at: 1 put: 16r8B;
			at: 2 put: (self mod: ModRegInd RM: 4 RO: dest);
			at: 3 put: (self s: SIB8 i: index b: base).
		 ^4].
	machineCode
		at: 1 put: 16r8B;
		at: 2 put: (self mod: ModRegRegDisp8 RM: 4 RO: dest);
		at: 3 put: (self s: SIB8 i: index b: base);
		at: 4 put: 0.
	 ^5! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeMulRR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| reg1 reg2 |
	reg1 := operands at: 0.
	reg2 := operands at: 1.
	machineCode
		at: 0 put: (self rexR: reg2 x: 0 b: reg1);
		at: 1 put: 16r0F;
		at: 2 put: 16rAF;
		at: 3 put: (self mod: ModReg RM: reg1 RO: reg2).
	^4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeNegateR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| reg |
	reg := operands at: 0.
	machineCode
		at: 0 put: (self rexR: 0 x: 0 b: reg);
		at: 1 put: 16rF7;
		at: 2 put: (self mod: ModReg RM: reg RO: 3).
	^3! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeNop
	<inline: true>
	machineCode at: 0 put: 16r90.
	^1! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeOpRR: x64opcode
	| regLHS regRHS |
	regLHS := operands at: 0.
	regRHS := operands at: 1.
	machineCode
		at: 0 put: (self rexR: regRHS x: 0 b: regLHS);
		at: 1 put: x64opcode;
		at: 2 put: (self mod: ModReg RM: regLHS RO: regRHS).
	^3! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePopR
	<inline: true>
	| reg |
	reg := operands at: 0.
	reg < 8 ifTrue:
		[machineCode at: 0 put: 16r58 + reg.
		^1].
	machineCode
		at: 0 put: 16r41;
		at: 1 put: 16r58 + (reg - 8).
	^2! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePrefetchAw
	"We support only prefetches for addresses that are variables relative to VarBase"
	| operand offset |
	operand := operands at: 0.
	(self isAddressRelativeToVarBase: operand) ifFalse:
		[^0].
	offset := operand - cogit varBaseAddress.
	machineCode
		at: 0 put: 16r0f;
		at: 1 put: 16r18;
		at: 2 put: 16r93;
		at: 3 put: (offset bitAnd: 16rFF);
		at: 4 put: (offset >> 16 bitAnd: 16rFF);
		at: 5 put: (offset >> 8 bitAnd: 16rFF);
		at: 6 put: offset >> 24.
	^7! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePushCq
	<inline: true>
	| value |
	value := operands at: 0.
	(self isQuick: value) ifTrue:
		[machineCode
			at: 0 put: 16r6A;
			at: 1 put: (value bitAnd: 16rFF).
		^2].
	(self is32BitSignedImmediate: value) ifTrue:
		[machineCode
			at: 0 put: 16r68;
			at: 1 put: (value bitAnd: 16rFF);
			at: 2 put: (value >> 8 bitAnd: 16rFF);
			at: 3 put: (value >> 16 bitAnd: 16rFF);
			at: 4 put: (value >> 24 bitAnd: 16rFF).
		^5].
	^self concretizePushCw! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizePushR
	<inline: true>
	| reg |
	reg := operands at: 0.
	reg < 8 ifTrue:
		[machineCode at: 0 put: 16r50 + reg.
		^1].
	machineCode
		at: 0 put: 16r41;
		at: 1 put: 16r50 + (reg - 8).
	^2! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeREP
	<inline: true>
	machineCode at: 0 put: 16rF3.
	^1! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeRetN
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| offset |
	offset := operands at: 0.
	offset = 0 ifTrue:
		[machineCode at: 0 put: 16rC3.
		^1].
	machineCode
		at: 0 put: 16rC2;
		at: 1 put: (offset bitAnd: 16rFF);
		at: 2 put: (offset >> 8).
	^3! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeReverseOpRR: x64opcode
	| regLHS regRHS |
	"CmpRR/MoveRR RHS LHS computes LHS - RHS, i.e. apparently reversed.  You have to think subtract."
	regRHS := operands at: 0.
	regLHS := operands at: 1.
	machineCode
		at: 0 put: (self rexR: regRHS x: 0 b: regLHS);
		at: 1 put: x64opcode;
		at: 2 put: (self mod: ModReg RM: regLHS RO: regRHS).
	^3! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSEE2OpRdRd: x64opcode
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| regLHS regRHS skip |
	regRHS := operands at: 0.
	regLHS := operands at: 1.
	machineCode
		at: 0 put: 16rF2.
	(regLHS <= 7 and: [regRHS <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: regLHS x: 0 b: regRHS)].
	machineCode 
		at: skip + 1 put: 16r0F;
		at: skip + 2 put: x64opcode;
		at: skip + 3 put: (self mod: ModReg RM: regRHS RO: regLHS).
	^skip + 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSEEOpRsRs: x64opcode
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| regLHS regRHS skip |
	regRHS := operands at: 0.
	regLHS := operands at: 1.
	machineCode
		at: 0 put: 16rF3.
	(regLHS <= 7 and: [regRHS <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: regLHS x: 0 b: regRHS)].
	machineCode 
		at: skip + 1 put: 16r0F;
		at: skip + 2 put: x64opcode;
		at: skip + 3 put: (self mod: ModReg RM: regRHS RO: regLHS).
	^skip + 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 1/7/2021 13:28'!
concretizeSet: conditionCode
	| reg |
	reg := operands at: 0.
	machineCode
		at: 0 put: (reg >= R8 ifTrue: [16r44] ifFalse: [16r40]);
		at: 1 put: 16r0F;
		at: 2 put: 16r90 + conditionCode;
		at: 3 put: (self mod: ModReg RM: (reg bitAnd: 7) RO: 0).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object"
	 ^4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeShiftCqRegOpcode: regOpcode
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| distance reg |
	distance := operands at: 0.
	self assert: (distance between: 1 and: 63).
	reg := operands at: 1.
	machineCode at: 0 put: (self rexR: 0 x: 0 b: reg).
	distance = 1 ifTrue:
		[machineCode
			at: 1 put: 16rD1;
			at: 2 put: (self mod: ModReg RM: reg RO: regOpcode).
		 ^3].
	machineCode
		at: 1 put: 16rC1;
		at: 2 put: (self mod: ModReg RM: reg RO: regOpcode);
		at: 3 put: distance.
	^4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeShiftRegRegOpcode: regOpcode
	"On the x64 the only instructions that shift by the value of a
	 register require the shift count to be  in %ecx.  So we may
	 have to use swap instructions to get the count into %ecx."
	<inline: true>
	| shiftCountReg destReg regToShift |
	shiftCountReg := operands at: 0.
	destReg := operands at: 1.
	shiftCountReg = RCX ifTrue:
		[machineCode
			at: 0 put: (self rexR: 0 x: 0 b: destReg);
			at: 1 put: 16rD3;
			at: 2 put: (self mod: ModReg RM: destReg RO: regOpcode).
		 ^3].
	regToShift := destReg = shiftCountReg
					ifTrue: [RCX]
					ifFalse: [destReg = RCX
								ifTrue: [shiftCountReg]
								ifFalse: [destReg]].
	shiftCountReg = RAX ifTrue:
		[machineCode
			at: 0 put: 16r48;
			at: 1 put: 16r90 + RCX; "XCHG RAX,RCX"
			at: 2 put: (self rexR: 0 x: 0 b: regToShift);
			at: 3 put: 16rD3;			"SAR RCX,RAX"
			at: 4 put: (self mod: ModReg RM: regToShift RO: regOpcode);
			at: 5 put: 16r48;
			at: 6 put: 16r90 + RCX. "XCHG RAX,RCX"
		 ^7].
	machineCode
		at: 0 put: (self rexR: shiftCountReg x: 0 b: RCX);		"XCHG R?X,RCX"
		at: 1 put: 16r87;
		at: 2 put: (self mod: ModReg RM: RCX RO: shiftCountReg);
		at: 3 put: (self rexR: 0 x: 0 b: regToShift);			"SAR RCX,R!!X"
		at: 4 put: 16rD3;
		at: 5 put: (self mod: ModReg RM: regToShift RO: regOpcode);
		at: 6 put: (self rexR: shiftCountReg x: 0 b: RCX);		"XCHG R?X,RCX"
		at: 7 put: 16r87;
		at: 8 put: (self mod: ModReg RM: RCX RO: shiftCountReg).
	^9! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSignExtend16RR
	"Will get inlined into concretizeAt: switch."
	"movsxwq"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: (self rexw: true r: destReg x: 0 b: srcReg);
		at: 1 put: 16r0F;
		at: 2 put: 16rBF;
		at: 3 put: (self mod: ModReg RM: srcReg RO: destReg).
	^ 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSignExtend32RR
	"Will get inlined into concretizeAt: switch."
	"movsxdq"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: (self rexw: true r: destReg x: 0 b: srcReg);
		at: 1 put: 16r63;
		at: 2 put: (self mod: ModReg RM: srcReg RO: destReg).
	^ 3! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSignExtend8RR
	"Will get inlined into concretizeAt: switch."
	"movsxbq"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: (self rexw: true r: destReg x: 0 b: srcReg);
		at: 1 put: 16r0F;
		at: 2 put: 16rBE;
		at: 3 put: (self mod: ModReg RM: srcReg RO: destReg).
	^ 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSqrtRd
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| reg skip |
	reg := operands at: 0.
	machineCode
		at: 0 put: 16rF2.
	(reg <= 7)
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: reg x: 0 b: reg)].
	machineCode
		at: skip + 1 put: 16r0F;
		at: skip + 2 put: 16r51;
		at: skip + 3 put: (self mod: ModReg RM: reg RO: reg).
	^skip + 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeSqrtRs
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| reg skip |
	reg := operands at: 0.
	machineCode
		at: 0 put: 16rF3.
	(reg <= 7)
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: reg x: 0 b: reg)].
	machineCode
		at: skip + 1 put: 16r0F;
		at: skip + 2 put: 16r51;
		at: skip + 3 put: (self mod: ModReg RM: reg RO: reg).
	^skip + 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeStop
	<inline: true>
	machineCode at: 0 put: 16rCC.
	^1! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeTstCqR
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| value reg |
	value := operands at: 0.
	reg := operands at: 1.
	machineCode
		at: 0 put: (self rexR: 0 x: 0 b: reg).
	(self isQuick: value) ifTrue:
		[machineCode
			at: 1 put: 16rF6;
			at: 2 put: (self mod: ModReg RM: reg RO: 0);
			at: 3 put: (value bitAnd: 16rFF).
		 ^4].
	
	(self is32BitSignedImmediate: value) ifTrue:
		[reg = RAX ifTrue:
			[machineCode
				at: 1 put: 16rA9;
				at: 2 put: (value bitAnd: 16rFF);
				at: 3 put: (value >> 8 bitAnd: 16rFF);
				at: 4 put: (value >> 16 bitAnd: 16rFF);
				at: 5 put: (value >> 24 bitAnd: 16rFF).
			 ^6].
		machineCode
			at: 1 put: 16rF7;
			at: 2 put: (self mod: ModReg RM: reg RO: 0);
			at: 3 put: (value bitAnd: 16rFF);
			at: 4 put: (value >> 8 bitAnd: 16rFF);
			at: 5 put: (value >> 16 bitAnd: 16rFF);
			at: 6 put: (value >> 24 bitAnd: 16rFF).
		 ^7].
	^self concretizeArithCwR: 16r85! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 10/24/2020 19:11'!
concretizeXCHGAwR
	"This is a very limited implementation that assumes RAX is available."
	<inline: true>
	| addressOperand reg scratch |
	addressOperand := operands at: 0.
	reg := operands at: 1.
	scratch := operands at: 2.
	self deny: reg = scratch.
	machineCode
		"movq addressOperand, %rscratch"
		at:  0 put: (self rexR: 0 x: 0 b: scratch);
		at:  1 put: 16rB8 + (scratch bitAnd: 7);
		at:  2 put: (addressOperand bitAnd: 16rFF);
		at:  3 put: (addressOperand >> 8 bitAnd: 16rFF);
		at:  4 put: (addressOperand >> 16 bitAnd: 16rFF);
		at:  5 put: (addressOperand >> 24 bitAnd: 16rFF);
		at:  6 put: (addressOperand >> 32 bitAnd: 16rFF);
		at:  7 put: (addressOperand >> 40 bitAnd: 16rFF);
		at:  8 put: (addressOperand >> 48 bitAnd: 16rFF);
		at:  9 put: (addressOperand >> 56 bitAnd: 16rFF);
		"xchgq %reg, (%rscratch)" "0xf: 49 87 45 00              xchgq  %rax, (%r13)"
		at: 10 put: (self rexR: reg x: 0 b: scratch);
		at: 11 put: 16r87;
		at: 12 put: (self s: SIB1 i: 0 b: scratch).
	"cogit processor disassembleInstructionAt: 0 In: machineCode object."
	"cogit processor disassembleInstructionAt: 10 In: machineCode object"
	^13! !

!CogX64Compiler methodsFor: 'generate machine code - concretize processor-specific' stamp: 'eem 12/8/2019 15:23'!
concretizeXCHGRR
	| r1 r2 |
	r1 := operands at: 0.
	r2 := operands at: 1.
	r2 = RAX ifTrue:
		[r2 := r1. r1 := RAX].
	r1 = RAX ifTrue:
		[machineCode
			at: 0 put: (self rexR: 0 x: 0 b: r2);
			at: 1 put: 16r90 + (r2 \\ 8).
		 ^2].
	machineCode
		at: 0 put: (self rexR: r1 x: 0 b: r2);
		at: 1 put: 16r87;
		at: 2 put: (self mod: ModReg RM: r2 RO: r1).
	^3! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeXorRdRd
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| regLHS regRHS skip |
	regRHS := self operands at: 0.
	regLHS := self operands at: 1.
	machineCode
		at: 0 put: 16r66.
	(regLHS <= 7 and: [regRHS <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: regLHS x: 0 b: regRHS)].
	machineCode
		at: skip + 1 put: 16r0f;
		at: skip + 2 put: 16r57;
		at: skip + 3 put: (self mod: ModReg RM: regRHS RO: regLHS).
	^skip + 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeXorRsRs
	"Will get inlined into concretizeAt: switch."
	<inline: true>
	| regLHS regRHS skip |
	regRHS := self operands at: 0.
	regLHS := self operands at: 1.
	machineCode
		at: 0 put: 16r0F.
	(regLHS <= 7 and: [regRHS <= 7])
		ifTrue: [skip := 0]
		ifFalse: [machineCode at: (skip := 1) put: (self rexw: false r: regLHS x: 0 b: regRHS)].
	
	machineCode	
		at: skip + 1 put: 16r57;
		at: skip + 2 put: (self mod: ModReg RM: regRHS RO: regLHS).
	^skip + 3! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeZeroExtend16RR
	"Will get inlined into concretizeAt: switch."
	"movzxwq"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: (self rexw: true r: destReg x: 0 b: srcReg);
		at: 1 put: 16r0F;
		at: 2 put: 16rB7;
		at: 3 put: (self mod: ModReg RM: srcReg RO: destReg).
	^ 4! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeZeroExtend32RR
	"Will get inlined into concretizeAt: switch."
	"movzxbq"
	<inline: true>
	| srcReg destReg skip |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	(srcReg <= 7 and: [destReg <= 7])
		ifTrue: [skip := 0]
		ifFalse: [skip := 1. machineCode at: 0 put: (self rexw: false r: destReg x: 0 b: srcReg)].
		
	machineCode
		at: skip + 0 put: 16r8b;
		at: skip + 1 put: (self mod: ModReg RM: srcReg RO: destReg).
	^ skip + 2! !

!CogX64Compiler methodsFor: 'generate machine code - concretize' stamp: 'eem 12/8/2019 15:23'!
concretizeZeroExtend8RR
	"Will get inlined into concretizeAt: switch."
	"movzxbq"
	<inline: true>
	| srcReg destReg |
	srcReg := operands at: 0.
	destReg := operands at: 1.
	machineCode
		at: 0 put: (self rexw: true r: destReg x: 0 b: srcReg);
		at: 1 put: 16r0F;
		at: 2 put: 16rB6;
		at: 3 put: (self mod: ModReg RM: srcReg RO: destReg).
	^ 4! !

!CogX64Compiler methodsFor: 'feature detection' stamp: 'eem 10/30/2020 05:50'!
cpuidWord1
	<cmacro: '(ign) cpuidWord1'>
	"For want of somewhere to put the variable..."
	^self class cpuidWord1! !

!CogX64Compiler methodsFor: 'feature detection' stamp: 'eem 7/8/2021 12:29'!
detectFeatures
	"Do a throw-away compilation to get at the cpuid info and initialize cpuidWord1
	 N.B. All of MSVC, gcc & clang have intrinsics for this, so if you have the energy
	 by all means reimplement as an #if _MSC_VER...#elif __GNUC__ #else ... saga."
	| startAddress cpuid |
	<var: 'cpuid' declareC: 'usqIntptr_t (*cpuid)(void)'>
	startAddress := cogit methodZoneBase.
	cogit allocateOpcodes: 10 bytecodes: 0.
	cpuid := cogit cCoerceSimple: startAddress to: #'usqIntptr_t (*)(void)'.
	cogit
		PushR: RDX;
		PushR: RCX;
		PushR: RBX;
		MoveCq: 16r80000001 R: RAX;
		gen: CPUID;
		MoveR: RCX R: RAX;
		PopR: RBX;
		PopR: RCX;
		PopR: RDX;
		RetN: 0.
	cogit outputInstructionsForGeneratedRuntimeAt: startAddress.
	cogit resetMethodZoneBase: startAddress.
	cogit ensureExecutableCodeZone.
	self setCpuidWord1: (self cCode: 'cpuid()' inSmalltalk: [cogit simulateLeafCallOf: startAddress])! !

!CogX64Compiler methodsFor: 'generate machine code' stamp: 'eem 8/5/2021 07:20'!
dispatchConcretize
	"Attempt to generate concrete machine code for the instruction at address.
	 This is the inner dispatch of concretizeAt: actualAddress which exists only
	 to get around the branch size limits in the SqueakV3 (blue book derived)
	 bytecode set."
	opcode >= CPUID ifTrue:
		[^self dispatchConcretizeProcessorSpecific].

	opcode caseOf: {
		"Noops & Pseudo Ops"
		[Label]				-> [^self concretizeLabel].
		[AlignmentNops]	-> [^self concretizeAlignmentNops].
		[Fill32]				-> [^self concretizeFill32].
		[Nop]				-> [^self concretizeNop].
		"Control"
		[Call]					-> [^self concretizeCall].
		[CallR]					-> [^self concretizeCallR].
		[CallFull]				-> [^self concretizeCallFull].
		[JumpR]					-> [^self concretizeJumpR].
		[JumpFull]				-> [^self concretizeJumpFull].
		[JumpLong]				-> [^self concretizeJumpLong].
		[JumpLongZero]		-> [^self concretizeConditionalJump: 16r4].
		[JumpLongNonZero]	-> [^self concretizeConditionalJump: 16r5].
		[Jump]					-> [^self concretizeJump].
		"Table B-1 Intel 64 and IA-32 Architectures Software Developer's Manual Volume 1: Basic Architecture"
		[JumpZero]				-> [^self concretizeConditionalJump: 16r4].
		[JumpNonZero]			-> [^self concretizeConditionalJump: 16r5].
		[JumpNegative]			-> [^self concretizeConditionalJump: 16r8].
		[JumpNonNegative]		-> [^self concretizeConditionalJump: 16r9].
		[JumpOverflow]			-> [^self concretizeConditionalJump: 16r0].
		[JumpNoOverflow]		-> [^self concretizeConditionalJump: 16r1].
		[JumpCarry]			-> [^self concretizeConditionalJump: 16r2].
		[JumpNoCarry]			-> [^self concretizeConditionalJump: 16r3].
		[JumpLess]				-> [^self concretizeConditionalJump: 16rC].
		[JumpGreaterOrEqual]	-> [^self concretizeConditionalJump: 16rD].
		[JumpGreater]			-> [^self concretizeConditionalJump: 16rF].
		[JumpLessOrEqual]		-> [^self concretizeConditionalJump: 16rE].
		[JumpBelow]			-> [^self concretizeConditionalJump: 16r2].
		[JumpAboveOrEqual]	-> [^self concretizeConditionalJump: 16r3].
		[JumpAbove]			-> [^self concretizeConditionalJump: 16r7].
		[JumpBelowOrEqual]	-> [^self concretizeConditionalJump: 16r6].
		[JumpFPEqual]				-> [^self concretizeConditionalJump: 16r4].
		[JumpFPNotEqual]			-> [^self concretizeConditionalJump: 16r5].
		[JumpFPLess]				-> [^self concretizeConditionalJump: 16r2].
		[JumpFPGreaterOrEqual]	-> [^self concretizeConditionalJump: 16r3].
		[JumpFPGreater]			-> [^self concretizeConditionalJump: 16r7].
		[JumpFPLessOrEqual]		-> [^self concretizeConditionalJump: 16r6].
		[JumpFPOrdered]			-> [^self concretizeConditionalJump: 16rB].
		[JumpFPUnordered]			-> [^self concretizeConditionalJump: 16rA].
		[RetN]						-> [^self concretizeRetN].
		[Stop]						-> [^self concretizeStop].
		"Arithmetic"
		[AddCqR]					-> [^self concretizeArithCqRWithRO: 0 raxOpcode: 15r05].
		[AddcCqR]					-> [^self concretizeArithCqRWithRO: 2 raxOpcode: 15r15].
		[AddCwR]					-> [^self concretizeArithCwR: 16r03].
		[AddRR]						-> [^self concretizeOpRR: 16r03].
		[AddRsRs]					-> [^self concretizeSEEOpRsRs: 16r58].
		[AddRdRd]					-> [^self concretizeSEE2OpRdRd: 16r58].
		[AndCqR]					-> [^self concretizeArithCqRWithRO: 4 raxOpcode: 16r25].
		[AndCwR]					-> [^self concretizeArithCwR: 16r23].
		[AndRR]						-> [^self concretizeOpRR: 16r23].
		[TstCqR]					-> [^self concretizeTstCqR].
		[CmpCqR]					-> [^self concretizeArithCqRWithRO: 7 raxOpcode: 16r3D].
		[CmpCwR]					-> [^self concretizeArithCwR: 16r39].
		[CmpC32R]					-> [^self concretizeCmpC32R].
		[CmpRR]					-> [^self concretizeReverseOpRR: 16r39].
		[CmpRdRd]					-> [^self concretizeCmpRdRd].
		[CmpRsRs]					-> [^self concretizeCmpRsRs].
		[DivRdRd]					-> [^self concretizeSEE2OpRdRd: 16r5E].
		[DivRsRs]					-> [^self concretizeSEEOpRsRs: 16r5E].
		[MulRdRd]					-> [^self concretizeSEE2OpRdRd: 16r59].
		[MulRsRs]					-> [^self concretizeSEEOpRsRs: 16r59].
		[OrCqR]						-> [^self concretizeArithCqRWithRO: 1 raxOpcode: 16r0D].
		[OrCwR]					-> [^self concretizeArithCwR: 16r0B].
		[OrRR]						-> [^self concretizeOpRR: 16r0B].
		[SubCqR]					-> [^self concretizeArithCqRWithRO: 5 raxOpcode: 16r2D].
		[SubbCqR]					-> [^self concretizeArithCqRWithRO: 3 raxOpcode: 16r1D].
		[SubCwR]					-> [^self concretizeArithCwR: 16r2B].
		[SubRR]						-> [^self concretizeOpRR: 16r2B].
		[SubRdRd]					-> [^self concretizeSEE2OpRdRd: 16r5C].
		[SubRsRs]					-> [^self concretizeSEEOpRsRs: 16r5C].
		[SqrtRd]					-> [^self concretizeSqrtRd].
		[SqrtRs]					-> [^self concretizeSqrtRs].
		[XorCwR]					-> [^self concretizeArithCwR: 16r33].
		[XorRR]						-> [^self concretizeOpRR: 16r33].
		[XorRdRd]						-> [^self concretizeXorRdRd].
		[XorRsRs]						-> [^self concretizeXorRsRs].
		[NegateR]					-> [^self concretizeNegateR].
		[LoadEffectiveAddressMwrR]	-> [^self concretizeLoadEffectiveAddressMwrR].
		[RotateLeftCqR]				-> [^self concretizeShiftCqRegOpcode: 0].
		[RotateRightCqR]				-> [^self concretizeShiftCqRegOpcode: 1].
		[ArithmeticShiftRightCqR]		-> [^self concretizeShiftCqRegOpcode: 7].
		[LogicalShiftRightCqR]			-> [^self concretizeShiftCqRegOpcode: 5].
		[LogicalShiftLeftCqR]			-> [^self concretizeShiftCqRegOpcode: 4].
		[ArithmeticShiftRightRR]			-> [^self concretizeShiftRegRegOpcode: 7].
		[LogicalShiftLeftRR]				-> [^self concretizeShiftRegRegOpcode: 4].
		[ClzRR]						-> [^self concretizeClzRR].
		"Data Movement"
		[MoveCqR]			-> [^self concretizeMoveCqR].
		[MoveCwR]			-> [^self concretizeMoveCwR].
		[MoveC32R]		-> [^self concretizeMoveC32R].
		[MoveRR]			-> [^self concretizeReverseOpRR: 16r89].
		[MoveAwR]			-> [^self concretizeMoveAwR].
		[MoveA32R]		-> [^self concretizeMoveA32R].
		[MoveRAw]			-> [^self concretizeMoveRAw].
		[MoveRA32]		-> [^self concretizeMoveRA32].
		[MoveAbR]			-> [^self concretizeMoveAbR].
		[MoveRAb]			-> [^self concretizeMoveRAb].
		[MoveMbrR]			-> [^self concretizeMoveMbrR].
		[MoveRMbr]			-> [^self concretizeMoveRMbr].
		[MoveM8rR]		-> [^self concretizeMoveMbrR].
		[MoveRM8r]		-> [^self concretizeMoveRMbr].
		[MoveM16rR]		-> [^self concretizeMoveM16rR].
		[MoveRM16r]		-> [^self concretizeMoveRM16r].
		[MoveM32rR]		-> [^self concretizeMoveM32rR].
		[MoveM32rRs]		-> [^self concretizeMoveM32rRs].
		[MoveM64rRd]		-> [^self concretizeMoveM64rRd].
		[MoveMwrR]		-> [^self concretizeMoveMwrR].
		[MoveXbrRR]		-> [^self concretizeMoveXbrRR].
		[MoveRXbrR]		-> [^self concretizeMoveRXbrR].
		[MoveXwrRR]		-> [^self concretizeMoveXwrRR].
		[MoveRXwrR]		-> [^self concretizeMoveRXwrR].
		[MoveX32rRR]		-> [^self concretizeMoveX32rRR].
		[MoveRX32rR]		-> [^self concretizeMoveRX32rR].
		[MoveRMwr]		-> [^self concretizeMoveRMwr].
		[MoveRM32r]		-> [^self concretizeMoveRM32r].
		[MoveRsM32r]		-> [^self concretizeMoveRsM32r].
		[MoveRdM64r]		-> [^self concretizeMoveRdM64r].
		[MoveRdR]			-> [^self concretizeMoveRdR].
		[MoveRRd]			-> [^self concretizeMoveRRd].
		[MoveRdRd]		-> [^self concretizeMoveRdRd].
		[MoveRsRs]		-> [^self concretizeMoveRsRs].
		[PopR]				-> [^self concretizePopR].
		[PushR]				-> [^self concretizePushR].
		[PushCq]			-> [^self concretizePushCq].
		[PushCw]			-> [^self concretizePushCw].
		[PrefetchAw]		-> [^self concretizePrefetchAw].
		"Conversion"
		[ConvertRRd]		-> [^self concretizeConvertRRd].
		[ConvertRdR]		-> [^self concretizeConvertRdR].
		[ConvertRRs]		-> [^self concretizeConvertRRs].
		[ConvertRsR]		-> [^self concretizeConvertRsR].
		[ConvertRsRd]	-> [^self concretizeConvertRsRd].
		[ConvertRdRs]	-> [^self concretizeConvertRdRs].
			
		[SignExtend8RR]		-> [^self concretizeSignExtend8RR].
		[SignExtend16RR]	-> [^self concretizeSignExtend16RR].
		[SignExtend32RR]	-> [^self concretizeSignExtend32RR].
		
		[ZeroExtend8RR]		-> [^self concretizeZeroExtend8RR].
		[ZeroExtend16RR]	-> [^self concretizeZeroExtend16RR].
		[ZeroExtend32RR]	-> [^self concretizeZeroExtend32RR].

		"miscellaneous"
		[MovePerfCnt64RL]	->	[^self concretizeMovePerfCnt64RL]
		}.

	^0 "keep Slang happy"! !

!CogX64Compiler methodsFor: 'generate machine code' stamp: 'eem 10/25/2020 17:04'!
dispatchConcretizeProcessorSpecific
	"Attempt to generate concrete machine code for the instruction at address.
	 This is part of the inner dispatch of concretizeAt: actualAddress which exists only
	 to get around the number of literals limits in the SqueakV3 (blue book derived)
	 bytecode set."
	opcode caseOf: {
		[CPUID]					-> [^self concretizeCPUID].
		"Specific Arithmetic"
		[CDQ]					-> [^self concretizeCDQ].
		[IDIVR]					-> [^self concretizeIDIVR].
		[IMULRR]				-> [^self concretizeMulRR].
		[XCHGRR]				-> [^self concretizeXCHGRR].
		"Specific Data Movement"
		[REP]					-> [^self concretizeREP].
		[CLD]					-> [^self concretizeCLD].
		[MOVSB]				-> [^self concretizeMOVSB].
		[MOVSQ]				-> [^self concretizeMOVSQ].
		[BSR]					-> [^self concretizeBSR].
		"Multi-processing"
		"[CMPXCHGAwR]		-> [^self concretizeCMPXCHGAwR].
		[CMPXCHGMwrR]		-> [^self concretizeCMPXCHGMwrR].
		[LFENCE]				-> [^self concretizeFENCE: 5]."
		[MFENCE]				-> [^self concretizeFENCE: 6].
		[SFENCE]				-> [^self concretizeFENCE: 7].
		[LOCK]					-> [^self concretizeLOCK].
		[MoveRAwNoVBR]		-> [^self concretizeMoveRAwNoVBR].
		[CMPXCHGRMr]			-> [^self concretizeCMPXCHGRMr].
		[SETE]					-> [^self concretizeSet: 4].
		"[XCHGMwrR]			-> [^self concretizeXCHGMwrR]."
	}.
	^0! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 2/4/2020 09:59'!
flushDCacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	"If there is a dual mapped code zone (the normal zone but marked with read/execute, and a
	 read/write zone codeToDataDelta bytes away) then flush the data cache for the corresponding
	 range in the read/write zone and invalidate the data cache for the read/execute zone."
	<cmacro: '(me,startAddress,endAddress) 0'>
	"Currently we don't know whether a dual zone on an Intel x86/x86-64 would require cache
	 flushing to ensure coherence.  So for now this is a no op."! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 2/4/2020 09:57'!
flushICacheFrom: startAddress "<Integer>" to: endAddress "<Integer>"
	"Flush the instruction cache from (startAddress to endAddress].
	 If there is a dual mapped code zone (the normal zone but marked with read/execute, and a
	 read/write zone codeToDataDelta bytes away) then also flush the data cache for the corresp-
	 onding range in the read/write zone and invalidate the data cache for the read/execute zone."
	<cmacro: '(me,startAddress,endAddress) 0'>
	"On Intel processors where code and data have the same linear address, no
	 special action is required to flush the instruction cache.  One only needs to
	 execute a serializing instruction (e.g. CPUID) if code and data are at different
	 virtual addresses (e.g. a debugger using memory-mapping to access a debugee).
	 Using the macro avoids an unnecessary call."! !

!CogX64Compiler methodsFor: 'abi' stamp: 'eem 9/23/2015 18:23'!
fullCallsAreRelative
	^false! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'eem 12/25/2015 18:23'!
genDivR: abstractRegDivisor R: abstractRegDividend Quo: abstractRegQuotient Rem: abstractRegRemainder
	| rDividend rDivisor rQuotient rRemainder saveRestoreEAX saveRestoreEDX saveRestoreExchanged |
	self assert: abstractRegDividend ~= abstractRegDivisor.
	self assert: abstractRegQuotient ~= abstractRegRemainder.
	rDividend := abstractRegDividend.
	rDivisor := abstractRegDivisor.
	rQuotient := abstractRegQuotient.
	rRemainder := abstractRegRemainder.
	"IDIV r does a signed divide of RDX:RAX by r, RAX := Quotient, RDX := Remainder.
	 Since we must sign extend the dividend into RDX we must substitute another register if RDX is an input."
	(rDividend = RDX or: [rDivisor = RDX]) ifTrue:
		[| rUnused |
		"Slang, sigh..."
		rUnused := RAX.
		[rUnused <= RDI] whileTrue:
			[(rUnused ~= RSP and: [rUnused ~= RBP and: [rUnused ~= RDX
			  and: [rUnused ~= rDividend and: [rUnused ~= rDivisor
			  and: [rUnused ~= rQuotient and: [rUnused ~= rRemainder]]]]]]) ifTrue:
				[cogit PushR: rUnused.
				cogit MoveR: RDX R: rUnused.
				rDividend = RDX
					ifTrue: [self genDivR: rDivisor R: rUnused Quo: rQuotient Rem: rRemainder]
					ifFalse: [self genDivR: rUnused R: rDividend Quo: rQuotient Rem: rRemainder].
				cogit PopR: rUnused.
				^self].
			  rUnused := rUnused + 1].
		self error: 'couldn''t find unused register in genDivR:R:Quo:Rem:'].
	"If either output does not include RAX or RDX we must save and restore RAX and/or RDX."
	(saveRestoreEAX := rQuotient ~= RAX and: [rRemainder ~= RAX]) ifTrue:
		[cogit PushR: RAX].
	(saveRestoreEDX := rQuotient ~= RDX and: [rRemainder ~= RDX]) ifTrue:
		[cogit PushR: RDX].
	saveRestoreExchanged := -1.
	rDividend ~= RAX ifTrue:
		[rDivisor = RAX
			ifTrue: [((rDividend ~= rQuotient and: [rDividend ~= rRemainder])
					and: [rDividend ~= RDX or: [saveRestoreEDX not]]) ifTrue:
						[cogit PushR: (saveRestoreExchanged := rDividend)].
					cogit gen: XCHGRR operand: rDivisor operand: rDividend]
			ifFalse: [cogit MoveR: rDividend R: RAX]].
	"CDQ sign-extends RAX into RDX as required for IDIV"
	cogit gen: CDQ.
	cogit gen: IDIVR operand: (rDivisor = RAX ifTrue: [rDividend] ifFalse: [rDivisor]).
	"Must not overwrite result while juggling"
	(rQuotient = RDX and: [rRemainder = RAX])
		ifTrue: [cogit gen: XCHGRR operand: rQuotient operand: rRemainder]
		ifFalse:
			[rQuotient = RDX
				ifTrue:
					[rRemainder ~= RDX ifTrue:
						[cogit MoveR: RDX R: rRemainder].
					rQuotient ~= RAX ifTrue:
						[cogit MoveR: RAX R: rQuotient]]
				ifFalse:
					[rQuotient ~= RAX ifTrue:
						[cogit MoveR: RAX R: rQuotient].
					rRemainder ~= RDX ifTrue:
						[cogit MoveR: RDX R: rRemainder]]].
	saveRestoreExchanged >= 0 ifTrue:
		[cogit PopR: saveRestoreExchanged].
	saveRestoreEDX ifTrue:
		[cogit PopR: RDX].
	saveRestoreEAX ifTrue:
		[cogit PopR: RAX]! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'nice 2/17/2020 21:28'!
genHighBitAlternativeIn: destReg ofSmallIntegerOopWithSingleTagBit: srcReg
	"When CLZ is not available, we can use BSR (Bit Scan Reverse) instruction.
	BEWARE: make sure that srcReg has a single tag bit."
	<inline: true>
	| jumpNegativeReceiver |
	<var: #jumpNegativeReceiver type: #'AbstractInstruction *'>
	<returnTypeC: #'AbstractInstruction *'>
	"The primitive must fail if receiver is negative"
	(cogit lastOpcode setsConditionCodesFor: JumpNegative) ifFalse:
		[cogit CmpCq: 0 R: srcReg]. "N.B. FLAGS := srcReg - 0"
	jumpNegativeReceiver := cogit JumpNegative: 0.
	cogit gen: BSR operand: srcReg operand: destReg.
	"theoretically we should handle case when srcReg is zero, but we do not care because it is never zero thanks to the tagBit"
	"and thanks to the tag bit, the +1 operation for getting 1-based rank instead of 0-based rank is not necessary, so we are done"
	^jumpNegativeReceiver! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'nice 9/7/2019 08:39'!
genJumpFPEqual: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genJumpFPOrderedAnd: JumpFPEqual to: jumpTarget! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'nice 9/7/2019 09:11'!
genJumpFPLess: jumpTarget
	"Note: this generates two jumps for handling case of NaN operands
	Sender might prefer using a JumpFPGreater with swapped operands."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genJumpFPOrderedAnd: JumpFPLess to: jumpTarget! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'nice 9/7/2019 09:12'!
genJumpFPLessOrEqual: jumpTarget
	"Note: this generates two jumps for handling case of NaN operands
	Sender might prefer using a JumpFPGreaterOrEqual with swapped operands."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genJumpFPOrderedAnd: JumpFPLessOrEqual to: jumpTarget! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'nice 9/7/2019 08:51'!
genJumpFPNotEqual: jumpTarget
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	^self genJumpFPUnorderedOr: JumpFPNotEqual to: jumpTarget! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'nice 9/11/2019 00:28'!
genJumpFPOrderedAnd: jumpOpCode to: jumpTarget
	"Test the comparison status flags ZF PF CF and jump to jumpTarget when
	- both Ordered (no nan operand) - a JNP
	- and the other condition (determined by jumpOpCode) is true
	jumpTarget maybe either an Integer offset, or an AbstractInstruction.
	This is necessary on Intel IA32 for JB JBE and JE with nan operands,
	otherwise unordered condition set the three flags to 1 and cause false positive.
	The implementation use JP to jump over the over test
	rather than link a JNP and jumpOpCode jumpTarget"
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	| jumpUnordered jumpToTarget |
	<var: #jumpUnordered type: #'AbstractInstruction *'>
	<var: #jumpToTarget type: #'AbstractInstruction *'>
	jumpUnordered := cogit gen: JumpFPUnordered operand: jumpTarget asInteger.
	jumpToTarget := cogit gen: jumpOpCode operand: jumpTarget asInteger.
	jumpUnordered jmpTarget: cogit Label.
	^jumpToTarget! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'nice 9/7/2019 08:51'!
genJumpFPUnorderedOr: jumpOpCode to: jumpTarget
	"Test the comparison status flags ZF PF CF and jump to jumpTarget when
	- either Unordered (no nan opearand) - a JP
	- and the other condition (determined by jumpOpCode) is true
	jumpTarget maybe either an Integer offset, or an AbstractInstruction.
	This is necessary on Intel IA32 for JNE with nan operands,
	otherwise unordered condition set the three flags to 1 and cause false negative."
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #jumpTarget type: #'void *'>
	| jumpUnordered jumpToTarget |
	<var: #jumpUnordered type: #'AbstractInstruction *'>
	<var: #jumpToTarget type: #'AbstractInstruction *'>
	jumpToTarget := cogit gen: JumpFPNotEqual operand: jumpTarget asInteger.
	jumpUnordered := cogit gen: JumpFPUnordered operand: jumpTarget asInteger.
	jumpToTarget addDependent: jumpUnordered.
	^jumpToTarget! !

!CogX64Compiler methodsFor: 'abi' stamp: 'eem 10/25/2020 15:53'!
genMarshallNArgs: numArgs arg: regOrConst0 arg: regOrConst1 arg: regOrConst2 arg: regOrConst3
	"Generate the code to pass up to four arguments in a C run-time call.  Hack: each argument is either
	 a negative number, which encodes a positive constant, or a non-negative number, that of a register.

	 Run-time calls have no more than four arguments, so chosen so that on ARM32, where in its C ABI
	 the first four integer arguments are passed in registers, all arguments can be passed in registers.
	 We defer to the back end to generate this code not so much that the back end knows whether it
	 uses the stack or registers to pass arguments (it does, but...). In fact we defer for an extremely evil
	 reason. Doing so allows the x64 (where up to 6 args are passed) to assign the register arguments
	 in an order that allows some of the argument registers to be used for specific abstract  registers,
	 specifically ReceiverResultReg and ClassReg.  This is evil, evil, evil, but also it's really nice to keep
	 using the old register assignments the original author has grown accustomed to.

	 How can this possibly work?  Look at Cogit class>>runtime for a list of the run-time calls and their
	 arguments, including which arguments are passed in which registers.  Look at CogX64Compiler's
	 subclass implementations of initializeAbstractRegisters.  There are no calls in which ReceiverResultReg
	 (RDX) and/or ClassReg (RCX) are passed along with Arg0Reg and Arg1Reg, and none in which the use of
	 either ReceiverResultReg or ClassReg conflict for args 3 & 4.  So if args are assigned in order, the
	 registers do not get overwritten.  Yes, this is evil, but it's so nice to continue to use RCX & RDX.

	 Argument registers for args 0 to 3 in SysV are RDI RSI RDX RCX, and in Win64 are RCX RDX R8 R9"
	<inline: true>
	SysV ifFalse: "WIN64 ABI allways reserve shadow space on the stack for callee to save up to 4 register parameters"
		[cogit SubCq: 32 R: RSP].
	numArgs = 0 ifTrue: [^self].
	self assert: numArgs <= 4.
	(cogit isTrampolineArgConstant: regOrConst0)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst0) R: CArg0Reg] "a.k.a. Arg0Reg"
		ifFalse:
			[regOrConst0 ~= CArg0Reg ifTrue:
				[cogit MoveR: regOrConst0 R: CArg0Reg]].
	numArgs = 1 ifTrue: [^self].
	(cogit isTrampolineArgConstant: regOrConst1)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst1) R: CArg1Reg] "a.k.a. Arg1Reg"
		ifFalse:
			[regOrConst1 ~= CArg1Reg ifTrue:
				[cogit MoveR: regOrConst1 R: CArg1Reg]].
	numArgs = 2 ifTrue: [^self].
	(cogit isTrampolineArgConstant: regOrConst2)
		ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst2) R: CArg2Reg] "a.k.a. ReceiverResultReg (SysV) ClassReg (Win64)"
		ifFalse:
			[regOrConst2 ~= CArg2Reg ifTrue:
				[cogit MoveR: regOrConst2 R: CArg2Reg]].
	 numArgs = 3 ifTrue: [^self].
	 (cogit isTrampolineArgConstant: regOrConst3)
			ifTrue: [cogit MoveCq: (cogit trampolineArgValue: regOrConst3) R: CArg3Reg] "a.k.a. ClassReg (SysV) ReceiverResultReg (Win64)"
			ifFalse:
				[regOrConst3 ~= CArg3Reg ifTrue:
					[cogit MoveR: regOrConst3 R: CArg3Reg]]! !

!CogX64Compiler methodsFor: 'abi' stamp: 'rsf 1/12/2017 15:08'!
genMarshallNArgs: numArgs floatArg: regOrConst0 floatArg: regOrConst1 floatArg: regOrConst2 floatArg: regOrConst3
	"Generate the code to pass up to four float arguments in a C run-time call.  Hack: each argument is
	 either a negative number, which encodes a constant, or a non-negative number, that of a register.

	 Run-time calls have no more than four arguments, so chosen so that on ARM, where in its C ABI the
	 first four integer arguments are passed in registers, all arguments can be passed in registers.  We
	 defer to the back end to generate this code not so much that the back end knows whether it uses
	 the stack or registers to pass arguments (it does, but...). In fact we defer for an extremely evil reason.
	 Doing so allows the x64 (where up to 6 args are passed) to assign the register arguments in an order
	 that allows some of the argument registers to be used for specific abstract  registers, specifically
	 ReceiverResultReg and ClassReg.  This is evil, evil, evil, but also it's really nice to keep using the old
	 register assignments the original author has grown accustomed to."
	<inline: true>
	SysV ifFalse: "WIN64 ABI allways reserve shadow space on the stack for callee to save up to 4 register parameters"
		[cogit SubCq: 32 R: RSP].
	numArgs = 0 ifTrue: [^self].
	
	(cogit isTrampolineArgConstant: regOrConst0)
		ifTrue: [cogit MoveCf64: (cogit trampolineArgValue: regOrConst0) Rd: XMM0L] "a.k.a. DPFPReg0"
		ifFalse: [regOrConst0 ~= XMM0L ifTrue:
						[cogit MoveR: regOrConst0 R: XMM0L]].
	numArgs = 1 ifTrue: [^self].
	
	(cogit isTrampolineArgConstant: regOrConst1)
		ifTrue: [cogit MoveCf64: (cogit trampolineArgValue: regOrConst1) Rd: XMM1L] "a.k.a. DPFPReg1"
		ifFalse: [regOrConst1 ~= XMM1L ifTrue:
						[cogit MoveR: regOrConst1 R: XMM1L]].
	numArgs = 2 ifTrue: [^self].

	(cogit isTrampolineArgConstant: regOrConst2)
		ifTrue: [cogit MoveCf64: (cogit trampolineArgValue: regOrConst2) Rd: XMM2L] "a.k.a. DPFPReg0"
		ifFalse: [regOrConst2 ~= XMM2L ifTrue:
						[cogit MoveR: regOrConst2 R: XMM2L]].
	numArgs = 3 ifTrue: [^self].

	(cogit isTrampolineArgConstant: regOrConst3)
		ifTrue: [cogit MoveCf64: (cogit trampolineArgValue: regOrConst3) Rd: XMM3L] "a.k.a. DPFPReg3"
		ifFalse: [regOrConst3 ~= XMM3L ifTrue:
						[cogit MoveR: regOrConst3 R: XMM3L]].

	self assert: numArgs <= 4! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'rsf 1/12/2017 17:23'!
genMemCopy: originalSourceReg to: originalDestReg constantSize: size
	| numbytes numwords sourceReg destReg countReg inst |

	"Get the abstract registers for ECX, EDI and ESI"
	sourceReg := RSI.
	destReg := RDI.
	countReg := RCX.

	"Put the source in ESI and the dest in EDI"
	inst := cogit Label.
	originalSourceReg ~= sourceReg ifTrue: [
		originalDestReg = sourceReg ifTrue: [
			cogit MoveR: originalDestReg R: TempReg.
		].
		cogit MoveR: originalSourceReg R: sourceReg.
	].

	originalDestReg ~= destReg ifTrue: [
		originalDestReg = sourceReg ifTrue: [
			cogit MoveR: TempReg R: destReg.
		] ifFalse: [
			cogit MoveR: originalDestReg R: destReg.
		]
	].

	"Store the count and clear the direction"
	cogit gen: CLD.

	"First copy the bytes"
	numbytes := size bitAnd: 16r07.
	numbytes > 0 ifTrue: [
		cogit MoveCq: numbytes R: countReg.
		cogit gen: REP.
		cogit gen: MOVSB.
	].

	"Now copy the qwords"
	numwords := size // 8.
	cogit MoveCq:  numwords R: countReg.
	cogit gen: REP.
	cogit gen: MOVSQ.

! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'rsf 1/12/2017 17:24'!
genMemCopy: originalSourceReg to: originalDestReg size: originalSize
	| spilledSize size sourceReg destReg countReg inst |

	"Get the abstract registers for ECX, EDI and ESI"
	sourceReg := RSI.
	destReg := RDI.
	countReg := RCX.

	"TODO: Avoid spilling"
	spilledSize := false.
	(originalSize = sourceReg or: [originalSize = destReg]) ifTrue: [
		cogit PushR: originalSize.
		spilledSize := true.
	].

	"Put the source in ESI and the dest in EDI"
	inst := cogit Label.
	originalSourceReg ~= sourceReg ifTrue: [
		originalDestReg = sourceReg ifTrue: [
			cogit MoveR: originalDestReg R: TempReg.
		].
		cogit MoveR: originalSourceReg R: sourceReg.
	].

	originalDestReg ~= destReg ifTrue: [
		originalDestReg = sourceReg ifTrue: [
			cogit MoveR: TempReg R: destReg.
		] ifFalse: [
			cogit MoveR: originalDestReg R: destReg.
		]
	].

	"Put the original size register in somewhere different than ECX"
	spilledSize ifTrue: [
		cogit PopR: TempReg.
		size := TempReg.
	] ifFalse: [
		originalSize = countReg ifTrue: [
			cogit MoveR: originalSize R: TempReg.
			size := TempReg.
		] ifFalse: [
			size := originalSize.
		]
	].

	cogit gen: CLD.

	"First copy the bytes"
	cogit MoveR: size R: countReg.
	cogit AndCq: 16r07 R: countReg.
	cogit gen: REP.
	cogit gen: MOVSB.

	"Now copy the words"
	cogit MoveR: size R: countReg.
	cogit LogicalShiftRightCq: 3 R: countReg.
	cogit gen: REP.
	cogit gen: MOVSQ.
! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'rsf 1/13/2017 04:14'!
genMoveCf32: constantFloat32 Rs: register
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #constantFloat32 type: #float>
	| inst |
	inst := cogit PushCw: constantFloat32 asIEEE32BitWord.
	cogit PopRs: register.
	^ inst! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'rsf 1/13/2017 04:16'!
genMoveCf64: constantFloat64 Rd: register
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	<var: #constantFloat64 type: #double>
	| inst |
	inst := cogit PushCw: constantFloat64 asIEEE64BitWord.
	cogit PopRd: register.
	^ inst! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'eem 12/16/2015 09:44'!
genMulR: regSource R: regDest
	^cogit gen: IMULRR operand: regSource operand: regDest! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'rsf 1/13/2017 04:15'!
genPopRd: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	| inst |
	inst := cogit MoveM64: 0 r: SPReg Rd: reg .
	cogit AddCq: 8 R: SPReg.
	^ inst! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'rsf 1/13/2017 04:15'!
genPopRs: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	| inst |
	inst := cogit MoveM32: 0 r: SPReg Rs: reg .
	cogit AddCq: 8 R: SPReg.
	^ inst! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'eem 3/11/2017 12:54'!
genPushC64: constant64Bits
	<inline: true>
	^cogit PushCw: constant64Bits! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'rsf 1/13/2017 04:15'!
genPushRd: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	| inst |
	inst := cogit MoveRd: reg M64: -8 r: SPReg.
	cogit SubCq: 8 R: SPReg.
	^ inst! !

!CogX64Compiler methodsFor: 'smalltalk calling convention' stamp: 'eem (const to msg rewrite) 11/6/2014 11:51'!
genPushRegisterArgsForAbortMissNumArgs: numArgs
	"Ensure that the register args are pushed before the outer and
	 inner retpcs at an entry miss for arity <= self numRegArgs.  The
	 outer retpc is that of a call at a send site.  The inner is the call
	 from a method or PIC abort/miss to the trampoline."

	"This won't be as clumsy on a RISC.  But putting the receiver and
	 args above the return address means the CoInterpreter has a
	 single machine-code frame format which saves us a lot of work."

	"Iff there are register args convert
		base	->	outerRetpc		(send site retpc)
		sp		->	innerRetpc		(PIC abort/miss retpc)
	 to
		base	->	receiver
					(arg0)
					(arg1)
					outerRetpc
		sp		->	innerRetpc		(PIC abort/miss retpc)"
	numArgs <= cogit numRegArgs ifTrue:
		[self assert: cogit numRegArgs <= 2.
		 numArgs = 0 ifTrue:
			[cogit MoveMw: 0 r: SPReg R: TempReg.
			 cogit PushR: TempReg.
			 cogit MoveMw: objectMemory wordSize * 2 r: SPReg R: TempReg.
			 cogit MoveR: TempReg Mw: objectMemory wordSize r: SPReg.
			 cogit MoveR: ReceiverResultReg Mw: 2 * objectMemory wordSize r: SPReg.
			 ^self].
		 numArgs = 1 ifTrue:
			[cogit MoveMw: objectMemory wordSize r: SPReg R: TempReg.
			 cogit PushR: TempReg.
			 cogit MoveMw: objectMemory wordSize r: SPReg R: TempReg.
			 cogit PushR: TempReg.
			 cogit MoveR: ReceiverResultReg Mw: 3 * objectMemory wordSize r: SPReg.
			 cogit MoveR: Arg0Reg Mw: 2 * objectMemory wordSize r: SPReg.
			 ^self].
		 numArgs = 2 ifTrue:
			[cogit PushR: Arg1Reg.
			 cogit MoveMw: objectMemory wordSize * 2 r: SPReg R: TempReg.
			 cogit PushR: TempReg.
			 cogit MoveMw: objectMemory wordSize * 2 r: SPReg R: TempReg.
			 cogit PushR: TempReg.
			 cogit MoveR: ReceiverResultReg Mw: 4 * objectMemory wordSize r: SPReg.
			 cogit MoveR: Arg0Reg Mw: 3 * objectMemory wordSize r: SPReg.
			 ^self]]! !

!CogX64Compiler methodsFor: 'smalltalk calling convention' stamp: 'eem 9/23/2015 17:31'!
genPushRegisterArgsForNumArgs: numArgs scratchReg: scratchReg
	"Ensure that the register args are pushed before the retpc for arity <= self numRegArgs.  This
	 isn't as clumsy on a RISC.  But putting the receiver and args above the return address
	 means the CoInterpreter has a single machine-code frame format which saves us a lot of work.
	 N.B. Take great care to /not/ smash TempReg, which is used in directed send marshalling.
	 We could use XCHG to swap the ReceiverResultReg and top-of-stack return address, pushing the
	 the ret pc (now in ReceiverResultReg) later, but XCHG is very slow.  We can use SendNumArgsReg
	 because it is only live in sends of arity >= (NumSendTrampolines - 1)."
	self assert: cogit numRegArgs < (NumSendTrampolines - 1).
	numArgs <= cogit numRegArgs ifTrue:
		[self assert: cogit numRegArgs <= 2.
		 false "these two variants show the same performance on Intel Core i7, but the second one may be shorter."
			ifTrue:
				[cogit MoveMw: 0 r: SPReg R: scratchReg. "Save return pc"
				 numArgs > 0 ifTrue:
					[cogit PushR: Arg0Reg.
					 numArgs > 1 ifTrue:
						[cogit PushR: Arg1Reg]].
				 cogit PushR: scratchReg.
				 cogit MoveR: ReceiverResultReg Mw: objectMemory wordSize * (1 + numArgs) r: SPReg]
			ifFalse:
				["a.k.a.
					cogit gen: XCHGMwrR operand: 0 operand: SPReg operand: ReceiverResultReg.
				  but XCHG is slow."
				 cogit MoveMw: 0 r: SPReg R: scratchReg. "Save return pc"
				 cogit MoveR: ReceiverResultReg Mw: 0 r: SPReg.
				 numArgs > 0 ifTrue:
					[cogit PushR: Arg0Reg.
					 numArgs > 1 ifTrue:
						[cogit PushR: Arg1Reg]].
				 cogit PushR: scratchReg]] "Restore return address"! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'rsf 1/13/2017 04:16'!
genPushRs: reg
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	| inst |
	inst := cogit MoveRs: reg M32: -8 r: SPReg.
	cogit SubCq: 8 R: SPReg.
	^ inst! !

!CogX64Compiler methodsFor: 'abi' stamp: 'eem 8/4/2021 14:52'!
genRemoveNArgsFromStack: n
	"This is a no-op on x64 SysV since the ABI passes up to 6 args in registers and trampolines currently observe a limit of 4.
	 But the WIN64 ABI always reserve shadow space for saving up to 4 parameter registers (even if less than 4 args)."
	<inline: true>
	self assert: n <= 4.
	SysV ifFalse: [cogit AddCq: 32 R: RSP].
	^0! !

!CogX64Compiler methodsFor: 'abi' stamp: 'eem 5/29/2017 10:26'!
genRemoveNFloatArgsFromStack: n
	"This is a no-op on x64 SysV since the ABI passes up to 6 args in registers and trampolines currently observe a limit of 4.
	But the WIN64 ABI allways reserve shadow space for saving up to 4 parameter registers (even if less than 4 args)."
	<inline: true>
	self assert: n <= 4.
	SysV ifFalse: [cogit AddCq: 32 R: RSP].
	^0! !

!CogX64Compiler methodsFor: 'abi' stamp: 'eem 3/3/2016 16:19'!
genRestoreRegs: regMask
	"Restore the registers in regMask as saved by genSaveRegs:."
	self deny: (regMask anyMask: (cogit registerMaskFor: RSP and: RBP)).
	RAX to: R15 do:
		[:reg|
		 (regMask anyMask: (cogit registerMaskFor: reg)) ifTrue:
			[cogit PopR: reg]].
	^0! !

!CogX64Compiler methodsFor: 'abi' stamp: 'eem 3/3/2016 14:45'!
genSaveRegs: regMask
	"Save the registers in regMask for a call into the C run-time from a trampoline."

	self assert: (R15 > RAX and: [R15 - RAX + 1 = 16]).
	self deny: (regMask anyMask: (cogit registerMaskFor: RSP and: RBP)).
	R15 to: RAX by: -1 do:
		[:reg|
		 (regMask anyMask: (cogit registerMaskFor: reg)) ifTrue:
			[cogit PushR: reg]].
	^0! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'eem 5/6/2009 10:40'!
genSubstituteReturnAddress: retpc
	<inline: true>
	<returnTypeC: #'AbstractInstruction *'>
	^cogit PushCw: retpc! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'eem 4/1/2020 13:19'!
genSubstituteReturnAddressR: retpcReg
	<inline: true>
	^cogit PushR: retpcReg! !

!CogX64Compiler methodsFor: 'abstract instructions' stamp: 'eem 2/15/2017 13:31'!
genSwapR: regA R: regB Scratch: regTmp
	^cogit gen: XCHGRR operand: regA operand: regB! !

!CogX64Compiler methodsFor: 'disassembly' stamp: 'eem 12/4/2015 22:02'!
generalPurposeRegisterMap
	<doNotGenerate>
	"Answer a Dictionary from register getter to register index."
	^Dictionary newFromPairs:
		{	#rax. RAX.
			#rcx. RCX.
			#rdx. RDX.
			#rbx. RBX.
			#rsi.  RSI.
			#rdi.  RDI.
			#r8. R8.
			#r9. R9.
			#r10. R10.
			#r11. R11.
			#r12. R12.
			#r13. R13.
			#r14. R14.
			#r15. R15	}! !

!CogX64Compiler methodsFor: 'multi-threading' stamp: 'eem 1/7/2021 13:38'!
generateLowLevelTryLock: vmOwnerLockAddress
	"Generate a function that attempts to lock the vmOwnerLock and answers if it succeeded."
	<inline: true>
	| vmOwnerLockAddressReg jumpEqual |
	vmOwnerLockAddress = 0 ifTrue:
		[cogit
			MoveCq: 1 R: ABIResultReg;
			RetN: 0.
		 ^self].
	"RAX holds the value of lock if unlocked (zero), receives the existing value of the lock; RAX is implicit in CMPXCHG"
	vmOwnerLockAddressReg := CArg1Reg.
	cogit
		MoveCq: 0 R: RAX;
		MoveCq: vmOwnerLockAddress R: vmOwnerLockAddressReg;
		gen: LOCK;
		gen: CMPXCHGRMr operand: CArg0Reg operand: vmOwnerLockAddressReg.
	jumpEqual := cogit JumpZero: 0.
	cogit CmpR: CArg0Reg R: RAX. "If not equal to zero is it already equal to the desired value?"
	jumpEqual jmpTarget: (cogit gen: SETE operand: ABIResultReg).
	cogit RetN: 0! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
hasConditionRegister
	"Answer if the receiver supports, e.g., JumpOverflow after a regular AddRR"
	^true! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
hasDoublePrecisionFloatingPointSupport
	^true! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 10/29/2020 22:19'!
hasLZCNTInstructions
	"Answer if we support LZCNT"
	<inline: true>
	^self cpuidWord1 anyMask: 1 << 5! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 9/23/2015 17:32'!
hasLinkRegister
	^false! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
hasThreeAddressArithmetic
	"Answer if the receiver supports three-address arithmetic instructions"
	^false! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 1/10/2016 14:28'!
hasVarBaseRegister
	"Answer if the processor has a dedicated callee-saved register to point to
	 the base of commonly-accessed variables."
	^true "RBX"! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 1/29/2020 14:54'!
inlineCacheValueForSelector: selector in: aCogMethod
	"Answer the value to put in an inline-cache that is being loaded with the selector.
	 Usually this is simply the selector, but in 64-bits the cache is only 32-bits wide and so the
	 cache is loaded with the index of the selector.  Override to use the 64-bit implementation."
	<cmacro: '(backEnd,selector,aCogMethod) indexForSelectorin(selector,aCogMethod)'>
	^cogit indexForSelector: selector in: aCogMethod! !

!CogX64Compiler methodsFor: 'disassembly' stamp: 'eem 12/8/2015 10:57'!
instructionSizeAt: pc
	"Answer the instruction size at pc. This is used in method disassembly
	 to decode the jumps in block dispatch to discover where block methods
	 occur within a larger method.   This is very far from a full decode."
	| op |
	op := objectMemory byteAt: pc.
	(op bitAnd: 16rF8) = 16r48 ifTrue:
		[^1 + (self instructionSizeAt: pc + 1)].
	^op caseOf:
		{	[16r0F]	->	[self twoByteInstructionSizeAt: pc].
			[16r3D]	->	[5]. "cmp EAX,imm32"
			[16r70]	->	[2]. "short conditional jumps"
			[16r71]	->	[2].
			[16r72]	->	[2].
			[16r73]	->	[2].
			[16r74]	->	[2].
			[16r75]	->	[2].
			[16r76]	->	[2].
			[16r77]	->	[2].
			[16r78]	->	[2].
			[16r79]	->	[2].
			[16r7A]	->	[2].
			[16r7B]	->	[2].
			[16r7C]	->	[2].
			[16r7D]	->	[2].
			[16r7E]	->	[2].
			[16r7F]	->	[2].
			[16r83]	->	[self sizeImmediateGroup1: op at: pc].
			[16r89]	->	[2]. "MOV Eb,Gb"
			[16r8B]	->	[self sizeHasModrm: op at: pc].
			[16r90]	->	[1]. "nop"
			[16rE9] ->	[5]. "long unconditional jump"
			[16rEB] ->	[2]. "short unconditional jump"
			[16rCC]	->	[1]. "int3" }! !

!CogX64Compiler methodsFor: 'testing' stamp: 'nice 7/8/2016 00:48'!
is32BitSignedImmediate: a64BitUnsignedOperand
	"Top 32 bits all the same as the bottom 32 bits' sign bit implies we can use a sign-extended 4 byte offset."
	^self cCode: [(self cCoerceSimple: a64BitUnsignedOperand to: #int) = (self cCoerceSimple: a64BitUnsignedOperand to: #sqLong)]
		inSmalltalk: [((a64BitUnsignedOperand >> 32) signedIntFromLong + 1 bitXor: 1) = (a64BitUnsignedOperand >> 31 bitAnd: 1)]! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 12/16/2015 23:26'!
isAddressRelativeToVarBase: varAddress
	<inline: true>
	<var: #varAddress type: #usqInt>
	"Support for addressing variables off the dedicated VarBaseReg.  Allow for 1Mb of variables.
	 The x64 supports 32-bit offsets, but we choose not to address everything from VarBaseReg."
	^varAddress notNil
	  and: [varAddress >= cogit varBaseAddress
	  and: [varAddress - cogit varBaseAddress < (1 << 20)]]! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 01/10/2016 10:01'!
isBigEndian
	^false! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 12/3/2015 22:26'!
isCallPrecedingReturnPC: mcpc
	"Assuming mcpc is a send return pc answer if the instruction before it is a call (not a CallFull)."
	^(objectMemory byteAt: mcpc - 5) = 16rE8! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 11/3/2015 18:07:25'!
isJumpAt: pc
	| op |
	op := objectMemory byteAt: pc.
	^  (op between: 16r70 and: 16r7F) "short conditional jumps"
	or: [op = 16rE9 "long unconditional jump"
	or: [op = 16rEB "short unconditional jump"
	or: [(op = 16r0F "long conditional jumps"
		and: [(objectMemory byteAt: pc + 1) between: 16r80 and: 16r8F])
	or: [op = 16r48 "full unconditional jumps"
		and: [(objectMemory byteAt: pc + 1) = 16rA1
		and: [(objectMemory byteAt: pc + 10) = 16rFF
		and: [(objectMemory byteAt: pc + 11) = 16rE0]]]]]]]! !

!CogX64Compiler methodsFor: 'testing' stamp: 'nice 7/8/2016 02:18'!
isQuick: operand
	<var: #operand type: #'usqIntptr_t'>
	^operand signedIntFromLong64 between: -128 and: 127! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 8/3/2021 19:17'!
isWithinMwOffsetRange: anAddress
	"Answer if an address can be accessed using the offset in a MoveMw:r:R: or similar instruction.
	 We assume this is true for 32-bit processors and expect 64-bit processors to answer false
	 for values in the interpreter or the object memory.    Restrict our use of offsets to reference
	 addresses within the method zone, rather than checking for a 32-bit offset, so as to keep the
	 simulator and real VM in sync."
	<inline: true>
	^cogit addressIsInCodeZone: anAddress! !

!CogX64Compiler methodsFor: 'accessing' stamp: 'eem 6/8/2015 16:56'!
jmpTarget: anAbstractInstruction
	"Set the target of a jump instruction.  These all have the target in the first operand.
	 Override to cope with JumpFPNotEqual where because of IEEE NaN conformance and
	 the behaviour of COMISD/UCOMISD we generate two jumps to the same target."
	| aDependent |
	<var: #aDependent type: #'AbstractInstruction *'>
	aDependent := dependent.
	[aDependent notNil] whileTrue:
		[aDependent jmpTarget: anAbstractInstruction.
		 aDependent := aDependent dependent].
	^super jmpTarget: anAbstractInstruction! !

!CogX64Compiler methodsFor: 'accessing' stamp: 'tpr 4/17/2015 19:56'!
jumpLongByteSize
"	Branch/Call ranges.  Jump[Cond] can be generated as short as possible.  Call/Jump[Cond]Long must be generated
	in the same number of bytes irrespective of displacement since their targets may be updated, but they need only
	span 16Mb, the maximum size of the code zone.  This allows e.g. ARM to use single-word call and jump instructions
	for most calls and jumps.  CallFull/JumpFull must also be generated in the same number of bytes irrespective of
	displacement for the same reason, but they must be able to span the full (32-bit or 64-bit) address space because
	they are used to call code in the C runtime, which may be distant from the code zone"
	^5! !

!CogX64Compiler methodsFor: 'accessing' stamp: 'eem 3/11/2009 22:24'!
jumpLongConditionalByteSize
	^6! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 09/28/2015 18:48'!
jumpLongTargetBeforeFollowingAddress: mcpc 
	"Answer the target address for the long jump immediately preceding mcpc"
	^self callTargetFromReturnAddress: mcpc! !

!CogX64Compiler methodsFor: 'disassembly' stamp: 'cb 4/28/2015 16:17'!
jumpTargetPCAt: pc
	<returnTypeC: #usqInt>
	| size byte offset |
	size := self instructionSizeAt: pc.
	size = 2
		ifTrue:
			[byte := objectMemory byteAt: pc + 1.
			 offset := (byte bitAnd: 16r80) = 0 ifTrue: [byte] ifFalse: [byte - 256]]
		ifFalse:
			[byte := objectMemory byteAt: pc + size - 1.
			 offset := (byte bitAnd: 16r80) = 0 ifTrue: [byte] ifFalse: [byte - 256].
			 offset := offset << 8 + (objectMemory byteAt: pc + size - 2).
			 offset := offset << 8 + (objectMemory byteAt: pc + size - 3).
			 offset := offset << 8 + (objectMemory byteAt: pc + size - 4)].
	^pc + size + offset! !

!CogX64Compiler methodsFor: 'abi' stamp: 'eem 6/1/2020 17:20'!
leafCallStackPointerDelta
	"Answer the delta from the stack pointer after a call to the stack pointer
	 immediately prior to the call.  This is used to compute the stack pointer
	 immediately prior to a call from within a leaf routine, which in turn is used
	 to capture the c stack pointer to use in trampolines back into the C run-time."
	^8! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 10/26/2016 12:03'!
literalBeforeInlineCacheTagAt: callSiteReturnAddress
	"Answer a literal loaded before the inline cache tag load for the return address of a send."
	^self literalBeforeFollowingAddress: callSiteReturnAddress - 12 "5 for the call plus 7 for the selectorIndex/classIndex load"! !

!CogX64Compiler methodsFor: 'accessing' stamp: 'eem 12/3/2015 19:10'!
loadPICLiteralByteSize
	"Answer the byte size of a MoveCwR opcode's corresponding machine code
	 when the argument is a PIC.  This is for the self-reference at the end of a
	 closed PIC: leaq 0xffffffffffffff2b(%rip), %rcx : 48 8D 0D 2B FF FF FF"
	<inline: true>
	^7! !

!CogX64Compiler methodsFor: 'accessing' stamp: 'eem 9/22/2015 12:15'!
machineCodeAt: anOffset
	^machineCode at: anOffset! !

!CogX64Compiler methodsFor: 'generate machine code' stamp: 'eem 9/23/2015 13:25'!
machineCodeBytes
	"Answer the maximum number of bytes of machine code generated for any abstract instruction.
	 e.g. xchg %rdx, %rax; movq $0x12345678ABCDEF0, %(rax); xchg %rdx, %rax => 48 92 48 A3 F0 DE BC 9A 78 56 34 12 48 92"
	^14! !

!CogX64Compiler methodsFor: 'encoding' stamp: 'eem 10/21/2015 10:23:43'!
mod: mod RM: regMode RO: regOpcode
	"See ModR/M byte & opcode syntax
	 In addition to the notation shown above in 'Mnemonic Syntax' on page 43,
	 the following notation indicates the size and type of operands in the syntax of an instruction opcode:
		/digit	Indicates that the ModRM byte specifies only one register or memory (r/m) operand.
				The digit is specified by the ModRM reg field and is used as an instruction-opcode extension.
				Valid digit values range from 0 to 7.
		/r		Indicates that the ModRM byte specifies both a register operand and a reg/mem (register or memory) operand."
	^mod << 6 + ((regOpcode bitAnd: 7) << 3) + (regMode bitAnd: 7)! !

!CogX64Compiler methodsFor: 'accessing' stamp: 'eem 9/28/2015 15:28'!
moveCwRByteSize
	<inline: true>
	self subclassResponsibility! !

!CogX64Compiler methodsFor: 'printing' stamp: 'eem 12/26/2015 03:24'!
nameForFPRegister: reg "<Integer>"
	<doNotGenerate>
	(reg between: 0 and: 15) ifTrue:
		[^#(XMM0L XMM1L XMM2L XMM3L XMM4L XMM5L XMM6L XMM7L XMM8L XMM9L XMM10L XMM11L XMM12L XMM13L XMM14L XMM15L) at: reg + 1].
	^super nameForFPRegister: reg! !

!CogX64Compiler methodsFor: 'printing' stamp: 'eem 5/30/2016 13:11'!
nameForRegister: reg "<Integer>"
	<doNotGenerate>
	| default |
	default := super nameForRegister: reg.
	^(default last = $?
	  and: [reg between: 0 and: 15])
		ifTrue: [#(RAX RCX RDX RBX RSP RBP RSI RDI R8 R9 R10 R11 R12 R13 R14 R15) at: reg + 1]
		ifFalse: [default]! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/12/2015 11:44:33'!
nsSendCacheAt: callSiteReturnAddress
	"Answer the NSSendCache for the return address of a Newspeak
	 self, super, outer, or implicit receiver send."
	<option: #NewspeakVM>
	<var: #callSiteReturnAddress type: #'char *'>
	<inline: false>
	^self literalBeforeFollowingAddress: callSiteReturnAddress asUnsignedInteger - 5 "sizeofcall"! !

!CogX64Compiler methodsFor: 'accessing' stamp: 'eem 11/28/2016 18:02'!
numIntRegArgs
	^SysV ifTrue: [6] ifFalse: [4]! !

!CogX64Compiler methodsFor: 'multi-threading' stamp: 'eem 1/7/2021 13:40'!
numLowLevelLockOpcodes
	<inline: #always>
	"ceTryLockVMOwner:
		xorq %rax, %rax
		movq &vmOwnerLock, %rsi
		lock cmpxchgq %rdi, (%rsi) 			N.B. lock cmpxchgq are two separate opcodes
		jz equal
		cmpq %rdi, %rax
	 equal:
		setz %alt
		ret"
	^8! !

!CogX64Compiler methodsFor: 'generate machine code' stamp: 'rmacnak 12/3/2015 19:00:24'!
padIfPossibleWithStopsFrom: startAddr to: endAddr
	self stopsFrom: startAddr to: endAddr! !

!CogX64Compiler methodsFor: 'feature detection' stamp: 'eem 8/5/2021 10:34'!
preferredRegisterForMovePerfCnt64RL
	"Answer the preferred register to receive the 64-bit performance counter,
	 or NoReg if none."
	<inline: #always>
	^RAX! !

!CogX64Compiler methodsFor: 'accessing' stamp: 'eem 9/28/2015 15:54'!
pushCwByteSize
	<inline: true>
	self subclassResponsibility! !

!CogX64Compiler methodsFor: 'abi' stamp: 'rsf 1/21/2017 03:37'!
registerToSaveIP
	"Temporary register used for fetching the instruction pointer. This should
	not be used for passing parameters in a standard ABI"
	<option: #LowcodeVM>
	^ R15! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 6/13/2014 13:27'!
relocateCallBeforeReturnPC: retpc by: delta
	| distance |
	delta ~= 0 ifTrue:
		[distance :=    ((objectMemory byteAt: retpc - 1) << 24)
					+  ((objectMemory byteAt: retpc - 2) << 16)
					+  ((objectMemory byteAt: retpc - 3) << 8)
					+   (objectMemory byteAt: retpc - 4).
		 distance := distance + delta.
		 objectMemory
			byteAt: retpc - 1 put: (distance >> 24 bitAnd: 16rFF);
			byteAt: retpc - 2 put: (distance >> 16 bitAnd: 16rFF);
			byteAt: retpc - 3 put: (distance >>   8 bitAnd: 16rFF);
			byteAt: retpc - 4 put: (distance            bitAnd: 16rFF)]! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/9/2015 19:41'!
relocateMethodReferenceBeforeAddress: pc by: delta
	"We generate the method address using pc-relative addressing.
	 Simply check that rip-relative addressing is being used. c.f.
	 concretizeMoveCwR"
	<inline: true>
	self assert: (((objectMemory byteAt: pc - 6) = 16r8D "move"
				and: [((objectMemory byteAt: pc - 5) bitOr: (self mod: 0 RM: 0 RO: 7)) = (self mod: ModRegInd RM: 5 RO: 7)])
				or: [(objectMemory byteAt: pc - 8) = 16r8D "push"
				and: [((objectMemory byteAt: pc - 7) bitOr: (self mod: 0 RM: 0 RO: 7)) = (self mod: ModRegInd RM: 5 RO: 7)]])! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 3/10/2017 17:21'!
rewriteCPICJumpAt: addressFollowingJump target: jumpTargetAddr
	"Rewrite the short jump instruction to jump to a new cpic case target. "
	<var: #addressFollowingJump type: #usqInt>
	<var: #jumpTargetAddr type: #usqInt>
	<var: #callDistance type: #sqInt> "prevent type inference for avoiding warning on abs"
	| callDistance |
	callDistance := jumpTargetAddr - addressFollowingJump.
	self assert: callDistance abs < 128.
	objectMemory
		byteAt: addressFollowingJump - 1
		put:  (callDistance bitAnd: 16rFF).
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: addressFollowingJump - 10 to: addressFollowingJump - 1]."! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 10/7/2020 18:46'!
rewriteCallAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a call instruction to call a different target.  This variant is used to link PICs
	 in ceSendMiss et al, and to rewrite cached primitive calls.   Answer the extent of
	 the code change which is used to compute the range of the icache to flush."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	| callDistance |
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: callSiteReturnAddress - 10 to: callSiteReturnAddress - 1]."
	(cogit initialized and: [callTargetAddress < cogit minCallAddress]) ifTrue:
		[self error: 'linking callsite to invalid address'].
	callDistance := (callTargetAddress - callSiteReturnAddress) signedIntToLong.
	objectMemory
		byteAt: callSiteReturnAddress - 1 put: (callDistance >> 24 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 2 put: (callDistance >> 16 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 3 put: (callDistance >>   8 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 4 put: (callDistance            bitAnd: 16rFF).
	self assert: (self callTargetFromReturnAddress: callSiteReturnAddress) = callTargetAddress.
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: callSiteReturnAddress - 10 to: callSiteReturnAddress - 1]."
	^5! !

!CogX64Compiler methodsFor: 'full transfer run-time support' stamp: 'eem 12/12/2015 15:17'!
rewriteCallFullAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a CallFull instruction to call a different target.  This variant is used to rewrite cached primitive calls.
	 Answer the extent of the code change which is used to compute the range of the icache to flush.
	 On x64 this is a rewrite of
		movq #64bits, %rax : 48 A1 b0 b1 b2 b3 b4 b5 b6 b7
		jmp %rax : FF E0 "
	self assert: (objectMemory byteAt: callSiteReturnAddress - 12) = 16r48.
	objectMemory unalignedLongAt: callSiteReturnAddress - 10 put: callTargetAddress.
	self assert: (self callFullTargetFromReturnAddress: callSiteReturnAddress) signedIntToLong64 = callTargetAddress.
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: callSiteReturnAddress - 10 to: callSiteReturnAddress - 1]."
	^12! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 10/7/2020 18:45'!
rewriteInlineCacheAt: callSiteReturnAddress tag: cacheTag target: callTargetAddress
	"Rewrite an inline cache to call a different target for a new tag.  This variant is used
	 to link unlinked sends in ceSend:to:numArgs: et al.  Answer the extent of the code
	 change which is used to compute the range of the icache to flush.
	 N.B.  On 64-bit platforms the inline cache tag is only 32-bits wide, hence this code
	 is identical to that for the IA32."
	<var: #callSiteReturnAddress type: #usqInt>
	<var: #callTargetAddress type: #usqInt>
	| callDistance |
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: callSiteReturnAddress - 12 to: callSiteReturnAddress - 1]."
	(cogit initialized and: [callTargetAddress < cogit minCallAddress]) ifTrue:
		[self error: 'linking callsite to invalid address'].
	callDistance := (callTargetAddress - callSiteReturnAddress) signedIntToLong.
	objectMemory
		byteAt: callSiteReturnAddress - 1 put: (callDistance >> 24 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 2 put: (callDistance >> 16 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 3 put: (callDistance >>   8 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 4 put: (callDistance            bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 6 put: (cacheTag >> 24 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 7 put: (cacheTag >> 16 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 8 put: (cacheTag >>   8 bitAnd: 16rFF);
		byteAt: callSiteReturnAddress - 9 put: (cacheTag            bitAnd: 16rFF).
	self assert: (self callTargetFromReturnAddress: callSiteReturnAddress) = callTargetAddress.
	"self cCode: ''
		inSmalltalk: [cogit disassembleFrom: callSiteReturnAddress - 12 to: callSiteReturnAddress - 1]."
	^12! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/13/2015 10:22'!
rewriteInlineCacheTag: cacheTag at: callSiteReturnAddress
	"Rewrite an inline cache with a new tag.  This variant is used
	 by the garbage collector."
	self unalignedLong32At: callSiteReturnAddress - 9 put: cacheTag! !

!CogX64Compiler methodsFor: 'full transfer run-time support' stamp: 'eem 12/6/2015 20:59'!
rewriteJumpFullAt: callSiteReturnAddress target: callTargetAddress
	"Rewrite a JumpFull instruction to jump to a different target.  This variant is used to rewrite cached primitive calls.
	 Answer the extent of the code change which is used to compute the range of the icache to flush.
	 On x64 this is a rewrite of
		movq #64bits, %rax : 48 A1 b0 b1 b2 b3 b4 b5 b6 b7
		jmp %rax : FF E0 "
	^self rewriteCallFullAt: callSiteReturnAddress target: callTargetAddress! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 3/10/2017 17:22'!
rewriteJumpLongAt: addressFollowingJump target: jumpTargetAddr
	"Rewrite a long jump instruction to jump to a different target.  This variant
	 is used to rewrite cached primitive calls.   Answer the extent of the
	 code change which is used to compute the range of the icache to flush."
	<inline: true>
	^self rewriteCallAt: addressFollowingJump target: jumpTargetAddr! !

!CogX64Compiler methodsFor: 'encoding' stamp: 'eem 9/17/2015 15:25'!
rexR: reg "<0-15>" x: sibReg "<0-15>"  b: fieldReg "<0-15>"
	^self rexw: true r: reg x: sibReg b: fieldReg! !

!CogX64Compiler methodsFor: 'encoding' stamp: 'eem 1/12/2016 08:43'!
rexw: width64 "<Boolean>" r: reg "<0-15>" x: sibReg "<0-15>"  b: fieldReg "<0-15>"
	"Given width64, the R register, sib register, and modrm/sib/reg field register,
	 answer the correctly encoded REX prefix byte.
	 See AMD64 Architecture Programmer's Manual Volume 3: General-Purpose and System Instructions, Table 1-11"
	| regBits |
	regBits := ((reg bitAnd: 8) >> 1) + ((sibReg bitAnd: 8) >> 2) + (fieldReg >> 3).
	^(width64 ifTrue: [16r48] ifFalse: [16r40]) + regBits! !

!CogX64Compiler methodsFor: 'encoding' stamp: 'eem 9/22/2015 16:46'!
s: scale i: indexReg b: baseReg 
	^scale << 6 + ((indexReg bitAnd: 7) << 3) + (baseReg bitAnd: 7)! !

!CogX64Compiler methodsFor: 'abi' stamp: 'eem 3/19/2015 15:04'!
saveAndRestoreLinkRegAround: aBlock
	"If the processor's ABI includes a link register, generate instructions
	 to save and restore it around aBlock, which is assumed to generate code."
	<inline: true>
	^aBlock value! !

!CogX64Compiler methodsFor: 'feature detection' stamp: 'eem 10/30/2020 05:42'!
setCpuidWord1: word
	<cmacro: '(ign,word) cpuidWord1 = word'>
	"For want of somewhere to put the variable..."
	self class instVarNamed: 'cpuidWord1' put: word! !

!CogX64Compiler methodsFor: 'testing' stamp: 'nice 2/18/2020 01:47'!
setsConditionCodesFor: aConditionalJumpOpcode
	<inline: false> "to save Slang from having to be a real compiler (it can't inline switches that return)"
	"Answer if the receiver's opcode sets the condition codes correctly for the given conditional jump opcode."
	^opcode caseOf:
		{	[ArithmeticShiftRightCqR]	->	[self shiftSetsConditionCodesFor: aConditionalJumpOpcode].
			[ArithmeticShiftRightRR]	->	[self shiftSetsConditionCodesFor: aConditionalJumpOpcode].
			[LogicalShiftLeftCqR]		->	[self shiftSetsConditionCodesFor: aConditionalJumpOpcode].
			[LogicalShiftLeftRR]			->	[self shiftSetsConditionCodesFor: aConditionalJumpOpcode].
			[LogicalShiftRightCqR]		->	[false].
			[XorRR]						->	[true].
			[ClzRR]						->	[aConditionalJumpOpcode = JumpZero or: [aConditionalJumpOpcode = JumpNonZero or: [aConditionalJumpOpcode = JumpCarry or: [aConditionalJumpOpcode = JumpNoCarry "carry flag is set if input is zero"]]]]
		}
		otherwise: [self halt: 'unhandled opcode in setsConditionCodesFor:'. false]! !

!CogX64Compiler methodsFor: 'testing' stamp: 'eem 9/30/2015 15:58'!
shiftSetsConditionCodesFor: aConditionalJumpOpcode
	"OF flag only guaranteed to be set for 1-bit shifts.  See [1] p 490.
	 Only SF, ZF & PF set according to result.  Since the question is currently
	 asked only for Zero and Negative use the following simplification."
	<inline: true>
	^(aConditionalJumpOpcode between: JumpZero and: JumpNonNegative)! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 12/12/2015 15:16'!
sixtyFourBitLiteralBefore: followingAddress
	<inline: true>
	^objectMemory unalignedLongAt: followingAddress - 8! !

!CogX64Compiler methodsFor: 'disassembly' stamp: 'eem 10/13/2009 12:44'!
sizeHasModrm: op at: pc
	| modrm mod ro rm |
	modrm := objectMemory byteAt: pc + 1.
	mod := modrm >> 6.
	ro := modrm >> 3 bitAnd: 7.
	rm := modrm bitAnd: 7.
	mod = 3 ifTrue:
		[^2].
	rm ~= 4 ifTrue: "no SIB byte"
		[^mod caseOf:
		   {	[0]	->	[rm = 5
						ifTrue: [6] "reg or 32-bit displacement"
						ifFalse: [3]].
			[1]	->	[3]. "8-bit displacement"
			[2]	->	[6] }].
	self halt: 'fall through in sizeHasModrm:at:'.
	^0! !

!CogX64Compiler methodsFor: 'disassembly' stamp: 'eem 10/13/2009 12:44'!
sizeImmediateGroup1: op at: pc
	"see [1] p A-7, p A-13"
	| modrm mod ro rm |
	modrm := objectMemory byteAt: pc + 1.
	mod := modrm >> 6.
	ro := modrm >> 3 bitAnd: 7.
	rm := modrm bitAnd: 7.
	^ro caseOf:
	   {	[7 "cmp"]	->	[op = 16r81
							ifTrue: [6]
							ifFalse: [3]] }! !

!CogX64Compiler methodsFor: 'encoding' stamp: 'rmacnak 12/3/2015 19:07:03'!
stop
	"int3"
	<inline: true>
	^16rCC! !

!CogX64Compiler methodsFor: 'generate machine code - support' stamp: 'eem 1/27/2020 08:50'!
stopsFrom: startAddr to: endAddr
	self
		cCode: [self memset: startAddr _: self stop _: endAddr - startAddr + 1]
		inSmalltalk:
			[| alignedEnd alignedStart stops |
			stops := self stop << 8 + self stop.
			stops := stops << 16 + stops.
			stops := stops << 32 + stops.
			alignedStart := startAddr + 7 // 8 * 8.
			alignedEnd := endAddr - 1 // 8 * 8.
			alignedEnd <= startAddr
				ifTrue:
					[startAddr to: endAddr do:
						[:addr | cogit codeByteAt: addr put: self stop]]
				ifFalse:
					[startAddr to: alignedStart - 1 do:
						[:addr | cogit codeByteAt: addr put: self stop].
					 alignedStart to: alignedEnd by: 8 do:
						[:addr | cogit codeLong64At: addr put: stops].
					 alignedEnd + 8 to: endAddr do:
						[:addr | cogit codeByteAt: addr put: self stop]]]! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'eem 11/22/2015 10:11'!
storeLiteral32: literal beforeFollowingAddress: followingAddress
	"Rewrite the 32-bit literal in the instruction immediately preceding followingAddress."
	objectMemory
		byteAt: followingAddress - 1 put: (literal >> 24 bitAnd: 16rFF);
		byteAt: followingAddress - 2 put: (literal >> 16 bitAnd: 16rFF);
		byteAt: followingAddress - 3 put: (literal >>   8 bitAnd: 16rFF);
		byteAt: followingAddress - 4 put: (literal            bitAnd: 16rFF)! !

!CogX64Compiler methodsFor: 'disassembly' stamp: 'eem 10/13/2009 12:44'!
twoByteInstructionSizeAt: pc
	| op |
	op := objectMemory byteAt: pc + 1. 
	^(op bitAnd: 16rF0) caseOf:
		{	[16r80]	->	[6 "long conditional jumps"] }! !

!CogX64Compiler methodsFor: 'memory access' stamp: 'eem 12/12/2015 14:47'!
unalignedLong32At: byteAddress
	<cmacro: '(inst,byteAddress) long32At(byteAddress)'>
	^ ((objectMemory byteAt: byteAddress + 3) << 24)
	+ ((objectMemory byteAt: byteAddress + 2) << 16)
	+ ((objectMemory byteAt: byteAddress + 1) << 8)
	+  (objectMemory byteAt: byteAddress)! !

!CogX64Compiler methodsFor: 'memory access' stamp: 'eem 12/13/2015 11:19'!
unalignedLong32At: byteAddress put: aWord
	<cmacro: '(inst,byteAddress,aWord) long32Atput(byteAddress,aWord)'>
	objectMemory
		byteAt: byteAddress + 0 put:  (aWord bitAnd: 16rFF);
		byteAt: byteAddress + 1 put: ((aWord >> 8) bitAnd: 16rFF);
		byteAt: byteAddress + 2 put: ((aWord >> 16) bitAnd: 16rFF);
		byteAt: byteAddress + 3 put: ((aWord >> 24) bitAnd: 16rFF).
	^aWord! !

!CogX64Compiler methodsFor: 'memory access' stamp: 'eem 12/12/2015 15:25'!
unalignedLongAt: byteAddress
	<cmacro: '(byteAddress) longAt(byteAddress)'>
	^objectMemory unalignedLongAt: byteAddress! !

!CogX64Compiler methodsFor: 'memory access' stamp: 'eem 12/12/2015 15:26'!
unalignedLongAt: byteAddress put: aWord
	<cmacro: '(byteAddress,aWord) longAtput(byteAddress,aWord)'>
	^objectMemory unalignedLongAt: byteAddress put: aWord! !

!CogX64Compiler methodsFor: 'inline cacheing' stamp: 'rmacnak 12/9/2015 23:07:11'!
zoneCallsAreRelative
	"Answer if Call and JumpLong are relative and hence need to take the caller's
	 relocation delta into account during code compaction, rather than just the
	 callee's delta."
	^true! !

VMStructType subclass: #CogBlockMethod
	instanceVariableNames: 'objectHeader homeOffset startpc padToWord cmNumArgs cmType cmRefersToYoung cpicHasMNUCaseOrCMIsFullBlock cmUsageCount cmUsesPenultimateLit cbUsesInstVars cmHasMovableLiteral cmUnusedFlag stackCheckOffset'
	classVariableNames: ''
	poolDictionaries: 'CogMethodConstants VMBasicConstants VMBytecodeConstants'
	category: 'VMMaker-JIT'!

!CogBlockMethod commentStamp: 'eem 9/9/2019 15:39' prior: 0!
I am the rump method header for a block method embedded in a full CogMethod.  I am the superclass of CogMethod, which is a Cog method header proper.  Instances of both classes have the same second word.  The homeOffset and startpc fields are overlaid on the objectHeader in a CogMethod.  See Cogit class>>structureOfACogMethod for more information.  In C I look like

	typedef struct {
		union {
			struct {
				unsigned short	homeOffset;
				unsigned short	startpc;
	#if SpurVM
				unsigned int	padToWord;
	#endif
			};
			sqInt/sqLong	objectHeader;
		};
		unsigned		cmNumArgs : 8;
		unsigned		cmType : 3;
		unsigned		cmRefersToYoung : 1;
		unsigned		cpicHasMNUCaseOrCMIsFullBlock : 1;
		unsigned		cmUsageCount : 3;
		unsigned		cmUsesPenultimateLit : 1;
		unsigned		cbUsesInstVars : 1;
		unsigned		cmHasMovableLiteral : 1;
		unsigned		cmUnusedFlag : 1;
		unsigned		stackCheckOffset : 12;
	 } CogBlockMethod;

My instances are not actually used.  The methods exist only as input to Slang.  The simulator uses my surrogates (CogBlockMethodSurrogate32 and CogBlockMethodSurrogate64) to reference CogBlockMethod and CogMethod structures in the code zone.  The start of the structure is 32-bits in the V3 memory manager and 64-bits in the Spour memory manager.  In a CMMethod these bits are set to the object header of a marked bits objects, allowing code to masquerade as objects when referred to from the first field of a CompiledMethod.  In a CMBlock, they hold the homeOffset and the startpc.

cbUsesInstVars
	- a flag set to true in blocks that refer to instance variables

cmHasMovableLiteral
	- a flag set to true in methods that refer to movable literals.  This allows Spur to follow literals post become and hence avoid having to follow literals in #==, #~~, and push/pop/storeLiteralVariable.

cmNumArgs
	- the byte containing the block or method arg count

cmRefersToYoung
	- a flag set to true in methods which contain a reference to an object in new space

cmType
	- one of CMFree, CMMethod, CMBlock, CMClosedPIC, CMOpenPIC

cmUnusedFlags
	- as yet unused bits

cmUsageCount
	- a count used to identify older methods in code compaction.  The count decays over time, and compaction frees methods with lower usage counts

cmUsesPenultimateLit
	- a flag that states whether the penultimate literal in the corresponding bytecode method is used.  This in turn is used to check that a become of a method does not alter its bytecode.

cpicHasMNUCaseOrCMIsFullBlock
	- a flag that states whether a CMClosedPIC contains one or more MNU cases which are PIC dispatches used to speed-up MNU processing,
	  or states whether a CMMethod is for a full block instead of for a compiled method.

homeOffset
	- the distance a CMBlock header is away from its enclosing CMMethod header

objectHeader
	- an object header used to fool the garbage collector into thinking that a CMMethod is a normal bits object, so that the first field (the header word) of a bytecoded method can refer directly to a CMMethod without special casing the garbage collector's method scanning code more than it already is.

padToWord
	- a pad that may be necessary to make the homeOffset, startpc, padToWord triple as large as a CMMethod's objectHeader field

stackCheckOffset
	- the distance from the header to the stack limit check in a frame building method or block, used to reenter execution in methods or blocks that have checked for events at what is effectively the first bytecode

startpc
	- the bytecode pc of the start of a CMBlock's bytecode in the bytecode method!

!CogBlockMethod class methodsFor: 'accessing' stamp: 'eem 2/15/2011 10:24'!
alignedByteSize
	self shouldNotImplement! !

!CogBlockMethod class methodsFor: 'accessing' stamp: 'eem 9/2/2011 14:03'!
alignedByteSizeOf: anObject forClient: aVMClass
	^aVMClass cogit cogBlockMethodSurrogateClass alignedByteSize! !

!CogBlockMethod class methodsFor: 'accessing class hierarchy' stamp: 'eem 10/1/2013 11:02'!
defaultObjectMemoryClass
	^NewObjectMemory! !

!CogBlockMethod class methodsFor: 'code generation' stamp: 'eem 1/29/2020 08:34'!
fieldAccessorSourceFor: surrogateClass bytesPerWord: bytesPerWord
	"Override to remove the objectHeader[:] accessors as these are defined manually in CogMethodSurrogate"
	^(super fieldAccessorSourceFor: surrogateClass bytesPerWord: bytesPerWord)
		removeKey: {surrogateClass. #objectHeader} ifAbsent: [];
		removeKey: {surrogateClass. #objectHeader:} ifAbsent: [];
		yourself! !

!CogBlockMethod class methodsFor: 'class initialization' stamp: 'eem 1/27/2020 19:06'!
initialize
	"CogBlockMethod initialize"
	"self withAllSubclasses do:
		[:cogMethodClass| (cogMethodClass class includesSelector: #initialize) ifTrue: [cogMethodClass initialize]]"
	(Smalltalk classNamed: #CogBlockMethodSurrogate32) ifNotNil:
		[:cbms32|
		self checkGenerateSurrogate: cbms32 bytesPerWord: 4].
	(Smalltalk classNamed: #CogBlockMethodSurrogate64) ifNotNil:
		[:cbms64|
		self checkGenerateSurrogate: cbms64 bytesPerWord: 8].

	"see instVarNamesAndTypesForTranslationDo:"
	CMMaxUsageCount		:= (2 raisedTo: 3) - 1.
	MaxStackCheckOffset	:= (2 raisedTo: 12) - 1.
	MaxMethodSize		:= (2 raisedTo: 16) - 1

	"{ CogBlockMethodSurrogate32 selectors reject: [:s| CogBlockMethod includesSelector: s].
	    CogBlockMethodSurrogate64 selectors reject: [:s| CogBlockMethod includesSelector: s].
	    CogMethodSurrogate32 selectors reject: [:s| CogMethod includesSelector: s].
	    CogMethodSurrogate64 selectors reject: [:s| CogMethod includesSelector: s]. }"! !

!CogBlockMethod class methodsFor: 'translation' stamp: 'eem 9/9/2019 09:29'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the
	 inst vars to include in a CogMethod or CogBlockMethod struct."

	self allInstVarNames do:
		[:ivn|
		"Notionally objectHeader is in a union with homeOffset and startpc but
		 we don't have any convenient support for unions.  So hack, hack, hack, hack."
		((self == CogBlockMethod
			ifTrue: [#('objectHeader')]
			ifFalse: [#('homeOffset' 'startpc' 'padToWord')]) includes: ivn) ifFalse:
				[aBinaryBlock
					value: ivn
					value: (ivn caseOf: {
								['objectHeader']			-> [self objectMemoryClass baseHeaderSize = 8
																ifTrue: [#sqLong]
																ifFalse: [#sqInt]].
								['cmNumArgs']				-> [#(unsigned ' : 8')].		"SqueakV3 needs only 5 bits"
								['cmType']					-> [#(unsigned ' : 3')].
								['cmRefersToYoung']		-> [#(unsigned #Boolean ' : 1')].
								['cmHasMovableLiteral']		-> [#(unsigned #Boolean ' : 1')].
								['cpicHasMNUCaseOrCMIsFullBlock']
															-> [#(unsigned #Boolean ' : 1')].
								['cmUsageCount']			-> [#(unsigned ' : 3')].		"See CMMaxUsageCount in initialize"
								['cmUsesPenultimateLit']	-> [#(unsigned #Boolean ' : 1')].
								['cbUsesInstVars']			-> [#(unsigned #Boolean ' : 1')].
								['cmUnusedFlag']			-> [#(unsigned ' : 1')].
								['stackCheckOffset']		-> [#(unsigned ' : 12')].	"See MaxStackCheckOffset in initialize. a.k.a. cPICNumCases"
								['blockSize']				-> [#'unsigned short'].		"See MaxMethodSize in initialize"
								['blockEntryOffset']			-> [#'unsigned short'].
								['homeOffset']				-> [#'unsigned short'].
								['startpc']					-> [#'unsigned short'].
								['padToWord']				-> [#(#BaseHeaderSize 8 'unsigned int')].
								['nextMethodOrIRCs']		-> [#usqInt].				"See NewspeakCogMethod"
								['counters']					-> [#usqInt]}				"See SistaCogMethod"
							otherwise:
								[#sqInt])]]! !

!CogBlockMethod class methodsFor: 'translation' stamp: 'eem 6/22/2009 13:29'!
isAccessor: aSelector
	"Answer if aSelector is simply an accessor method for one of our fields."
	^(#(cPICNumCases cPICNumCases: nextOpenPIC nextOpenPIC:) includes: aSelector)
	  or: [super isAccessor: aSelector]! !

!CogBlockMethod class methodsFor: 'code generation' stamp: 'eem 10/1/2013 14:04'!
offsetForInstVar: instVarName
	"Hack to offset accesses to variables by certain values.  The inst vars following
	 the objectHeader must be offset by the baseHeaderSize."
	^(#('objectHeader' 'homeOffset' 'startpc' 'padToWord') includes: instVarName) ifFalse:
		['baseHeaderSize']! !

!CogBlockMethod class methodsFor: 'code generation' stamp: 'eem 2/25/2020 19:00'!
setter: getter bitPosition: bitPosition bitWidth: bitWidth type: typeOrNil
	^String streamContents:
		[:s| | startByte endByte shift alignedPowerOf2 accessor mask |
		startByte := bitPosition // 8.
		endByte := bitPosition + bitWidth - 1 // 8.
		shift := bitPosition \\ 8.
		alignedPowerOf2 := (#(8 16 32 64) includes: bitWidth) and: [shift = 0].
		accessor := 'unsigned'
					, (#('Byte' 'Short' 'Long' 'Long')
							at: endByte - startByte + 1
							ifAbsent: ['Long64'])
					, 'At: index'.
		mask := #(16rFF 16rFFFF 16rFFFFFFFF 16rFFFFFFFF)
						at: endByte - startByte + 1
						ifAbsent: [(2 raisedTo: 64) - 1].
		s nextPutAll: getter; nextPutAll: ': aValue'; crtab: 1.
		s crtab: 1; nextPutAll: '| index delta |'; crtab: 1.
		s nextPutAll: 'index := address + '.
		(self offsetForInstVar: getter) ifNotNil:
			[:offsetExpr| s nextPutAll: offsetExpr, ' + '].
		s print: startByte + 1; nextPut: $.; crtab: 1.
		(typeOrNil notNil or: [alignedPowerOf2]) ifFalse:
			[s nextPutAll: 'self assert: (aValue between: 0 and: '; nextPutAll:  ((1 << bitWidth) - 1) hex; nextPutAll: ').'; crtab: 1].
		s nextPutAll: '(delta := cogit getCodeToDataDelta) > 0 ifTrue:'; crtab: 2.
		s nextPutAll: '[self assert: (cogit addressIsInCodeZone: address - delta).'; crtab: 2; space.
		self putAtPut: accessor, ' - delta'
			type: typeOrNil
			mask: (alignedPowerOf2 ifFalse: [mask - ((1 << bitWidth - 1) << shift)])
			shift: shift
			on: s
			indent: 3.
		s nextPutAll: '].'; crtab: 1.
		alignedPowerOf2 ifTrue:
			[s nextPut: $^].
		self putAtPut: accessor
			type: typeOrNil
			mask: (alignedPowerOf2 ifFalse: [mask - ((1 << bitWidth - 1) << shift)])
			shift: shift
			on: s
			indent: 2.
		alignedPowerOf2 ifFalse:
			[s nextPut: $.; crtab: 1; nextPutAll: '^aValue']]! !

!CogBlockMethod class methodsFor: 'accessing' stamp: 'eem 2/15/2011 10:25'!
surrogateClass
	self shouldNotImplement! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 2/26/2009 09:36'!
cPICNumCases
	"Answer the value of cPICNumCases (a.k.a. stackCheckOffset)"
	<cmacro: ' stackCheckOffset'>
	^stackCheckOffset! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 2/26/2009 09:39'!
cPICNumCases: anObject
	"Set the value of cPICNumCases (a.k.a. stackCheckOffset)"
	<cmacro: 'Hack hack hack hack i.e. the getter macro does all the work'>
	^stackCheckOffset := anObject! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 6/4/2015 09:08'!
cbUsesInstVars

	^cbUsesInstVars! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 6/4/2015 09:06'!
cbUsesInstVars: anObject

	^cbUsesInstVars := anObject! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 9/9/2019 15:35'!
cmHasMovableLiteral
	"Answer the value of cmHasMovableLiteral"

	^cmHasMovableLiteral! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 9/9/2019 15:35'!
cmHasMovableLiteral: anObject
	"Set the value of cmHasMovableLiteral"

	^cmHasMovableLiteral := anObject! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 4/14/2016 10:30'!
cmHomeMethod
	^SistaV1BytecodeSet
		ifTrue: [self cmIsFullBlock
					ifTrue: [self cCoerceSimple: self to: #'CogMethod *']
					ifFalse: [self cCoerceSimple: self asUnsignedInteger - self homeOffset to: #'CogMethod *']]
		ifFalse: [self cCoerceSimple: self asUnsignedInteger - self homeOffset to: #'CogMethod *']! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 4/15/2016 10:04'!
cmIsFullBlock
	"Answer the value of cpicHasMNUCaseOrCMIsFullBlock"
	<inline: true>
	^SistaV1BytecodeSet
		ifTrue: [self cpicHasMNUCaseOrCMIsFullBlock]
		ifFalse: [false]! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 1/9/2011 08:38'!
cmNumArgs
	"Answer the value of cmNumArgs"

	^cmNumArgs! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 1/11/2009 13:45'!
cmNumArgs: anObject
	"Set the value of cmNumArgs"

	^cmNumArgs := anObject! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 2/26/2009 16:51'!
cmRefersToYoung
	"Answer the value of cmRefersToYoung"

	^cmRefersToYoung! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 2/26/2009 16:51'!
cmRefersToYoung: anObject
	"Set the value of cmRefersToYoung"

	^cmRefersToYoung := anObject! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 1/15/2009 18:10'!
cmType
	"Answer the value of cmType"

	^cmType! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 1/15/2009 18:10'!
cmType: anInteger
	"Set the value of cmType"

	^cmType := anInteger! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 3/7/2009 09:54'!
cmUsageCount
	"Answer the value of cmUsageCount"

	^cmUsageCount! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 3/7/2009 09:55'!
cmUsageCount: anInteger
	"Set the value of cmUsageCount"

	^cmUsageCount := anInteger! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 3/1/2013 16:41'!
cmUsesPenultimateLit
	"Answer the value of cmUsesPenultimateLit"

	^cmUsesPenultimateLit! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 3/1/2013 16:41'!
cmUsesPenultimateLit: anObject
	"Set the value of cmUsesPenultimateLit"

	^cmUsesPenultimateLit := anObject! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 4/14/2016 10:30'!
cpicHasMNUCase
	"Answer if the receiver has an MNU case."
	<inline: true>

	^SistaV1BytecodeSet
		ifTrue: [self cpicHasMNUCaseOrCMIsFullBlock and: [self cmType = CMClosedPIC]]
		ifFalse: [cpicHasMNUCaseOrCMIsFullBlock]! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 4/11/2016 17:51'!
cpicHasMNUCase: anObject
	"Set if the receiver has an MNU case."
	<inline: true>
	^cpicHasMNUCaseOrCMIsFullBlock := anObject! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 04/11/2016 17:38'!
cpicHasMNUCaseOrCMIsFullBlock
	"Answer the value of cpicHasMNUCaseOrCMIsFullBlock"

	^cpicHasMNUCaseOrCMIsFullBlock! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 04/11/2016 17:38'!
cpicHasMNUCaseOrCMIsFullBlock: anObject
	"Set the value of cpicHasMNUCaseOrCMIsFullBlock"

	^cpicHasMNUCaseOrCMIsFullBlock := anObject! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 1/9/2011 08:37'!
homeOffset
	"Answer the value of homeOffset"

	^homeOffset! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 1/9/2011 08:37'!
homeOffset: anObject
	"Set the value of homeOffset"

	^homeOffset := anObject! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 1/30/2009 09:13'!
stackCheckOffset
	"Answer the value of stackCheckOffset"

	^stackCheckOffset! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 1/30/2009 09:12'!
stackCheckOffset: anObject
	"Set the value of stackCheckOffset"

	^stackCheckOffset := anObject! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 1/9/2011 08:37'!
startpc
	"Answer the value of startpc"

	^startpc! !

!CogBlockMethod methodsFor: 'accessing' stamp: 'eem 1/9/2011 08:37'!
startpc: anObject
	"Set the value of startpc"

	^startpc := anObject! !

CogBlockMethod subclass: #CogMethod
	instanceVariableNames: 'blockSize blockEntryOffset methodObject methodHeader selector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogMethod commentStamp: 'eem 8/19/2014 17:07' prior: 0!
I am a native Cog method or polymorphic inline cache.  If the former I have been produced by compiling a bytecoded CompiledMethod into machine code by the JIT, and I have a circular reference to that CompiledMethod.  The CompiledMethod's header field is a pointer to me, the CogMethod, and my methodHeader field holds the compiled method's actual header.  My objectHeader field looks like a single word object with a compact header with the mark bit set so as to placate the GC, i.e. make a CogMethod look like an object so that the reference to a CogMethod from a compiled method doesn't break the GC.  The cmType, stackCheckOffset, cmNumArgs & cmNumTemps fields are early in the structure because we place dummy two-word CogMethod headers within a method for each block within it to record this information for each block method (see my superclass CogBlockMethod).  In this case the objectHeader field is overlaid by the homeOffset and startpc fields.  The objectHeader field is also used to hold the relocation distance when compacting methods since when functioning as an obhject header it is a constant value and so can easily be reinitialized.  See Cogit class>>structureOfACogMethod for more information.

In C I look like

	typedef struct {
	    sqInt   objectHeader;
	
	    unsigned        cmNumArgs : 8;
	    unsigned        cmType : 3;
	    unsigned        cmRefersToYoung : 1;
	    unsigned        cmIsUnlinked : 1;
	    unsigned        cmUsageCount : 3;
	    unsigned        stackCheckOffset : 16;
	
	    unsigned short blockSize;
	    unsigned short blockEntryOffset;
	
	    sqInt   methodObject;
	    sqInt   methodHeader;
	    sqInt   selector;
	 } CogMethod;

Note that in a 64-bit system all fields from cmNumArgs through blockEntry fit in a single 64-bit word.

My instances are not actually used.  The methods exist only as input to Slang.  The simulator uses my surrogates (CogMethodSurrogate32 and CogMethodSurrogate64.!

!CogMethod class methodsFor: 'accessing' stamp: 'eem 9/2/2011 14:03'!
alignedByteSizeOf: anObject forClient: aVMClass
	^aVMClass cogit cogMethodSurrogateClass alignedByteSize! !

!CogMethod class methodsFor: 'translation' stamp: 'eem 2/21/2011 17:34'!
cogMethodHeader
	^String streamContents:
		[:s|
		CogBlockMethod printTypedefOn: s.
		s cr.
		self printTypedefOn: s]! !

!CogMethod class methodsFor: 'class initialization' stamp: 'eem 2/25/2009 21:34'!
initialize
	"self initialize"
	(Smalltalk classNamed: #CogMethodSurrogate32) ifNotNil:
		[:cms32|
		self checkGenerateSurrogate: cms32 bytesPerWord: 4].
	(Smalltalk classNamed: #CogMethodSurrogate64) ifNotNil:
		[:cms64|
		self checkGenerateSurrogate: cms64 bytesPerWord: 8]! !

!CogMethod class methodsFor: 'accessing' stamp: 'eem 2/15/2011 10:25'!
surrogateClass
	self shouldNotImplement! !

!CogMethod methodsFor: 'accessing' stamp: 'eem 2/25/2009 22:47'!
blockEntryOffset
	"Answer the value of blockEntryOffset"

	^blockEntryOffset! !

!CogMethod methodsFor: 'accessing' stamp: 'eem 2/25/2009 22:47'!
blockEntryOffset: anObject
	"Set the value of blockEntryOffset"

	^blockEntryOffset := anObject! !

!CogMethod methodsFor: 'accessing' stamp: 'eem 12/16/2008 08:50'!
blockSize
	"Answer the value of blockSize"

	^ blockSize! !

!CogMethod methodsFor: 'accessing' stamp: 'eem 2/5/2009 14:33'!
blockSize: anObject
	"Set the value of blockSize"

	^blockSize := anObject! !

!CogMethod methodsFor: 'testing' stamp: 'tpr 10/23/2015 13:38'!
containsAddress: anAddress
	"is anAddress within my bounds; not a test of addresses referred to within instructions in the method"
	<inline: true>
	^self asUnsignedInteger <= anAddress asUnsignedInteger
	  and: [self asUnsignedInteger + self blockSize >= anAddress asUnsignedInteger]! !

!CogMethod methodsFor: 'accessing' stamp: 'rsf 2/9/2017 23:31'!
counters
	^ 0! !

!CogMethod methodsFor: 'accessing' stamp: 'eem 12/16/2008 08:50'!
methodHeader
	"Answer the value of methodHeader"

	^ methodHeader! !

!CogMethod methodsFor: 'accessing' stamp: 'eem 2/5/2009 14:33'!
methodHeader: anObject
	"Set the value of methodHeader"

	^methodHeader := anObject! !

!CogMethod methodsFor: 'accessing' stamp: 'eem 2/5/2009 14:32'!
methodObject
	"Answer the value of methodObject"

	^methodObject! !

!CogMethod methodsFor: 'accessing' stamp: 'eem 2/5/2009 14:32'!
methodObject: anObject
	"Set the value of methodObject"

	^methodObject := anObject! !

!CogMethod methodsFor: 'accessing' stamp: 'eem 6/22/2009 11:29'!
nextOpenPIC
	"Answer the value of nextOpenPIC (a.k.a. methodObject)"
	<cmacro: ' methodObject'>
	^methodObject! !

!CogMethod methodsFor: 'accessing' stamp: 'eem 6/22/2009 11:29'!
nextOpenPIC: anObject
	"Set the value of nextOpenPIC (a.k.a. methodObject)"
	<cmacro: 'Hack hack hack hack i.e. the getter macro does all the work'>
	^methodObject := anObject! !

!CogMethod methodsFor: 'accessing' stamp: 'eem 2/5/2009 14:33'!
objectHeader
	"Answer the value of objectHeader"

	^objectHeader! !

!CogMethod methodsFor: 'accessing' stamp: 'eem 1/11/2009 13:44'!
objectHeader: anObject
	"Set the value of objectHeader"

	^objectHeader := anObject! !

!CogMethod methodsFor: 'accessing' stamp: 'eem 12/16/2008 08:50'!
selector
	"Answer the value of selector"

	^ selector! !

!CogMethod methodsFor: 'accessing' stamp: 'eem 2/5/2009 14:32'!
selector: anObject
	"Set the value of selector"

	^selector := anObject! !

CogMethod subclass: #NewspeakCogMethod
	instanceVariableNames: 'nextMethodOrIRCs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!NewspeakCogMethod commentStamp: 'eem 2/17/2017 16:12' prior: 0!
A NewspeakCogMethod is a specialization of CogMethod for Newspeak.  It adds support for the unpairdMethodsList and for implicit receiver caches.  Since the unpairedMethodList only holds methods that are inst var accessors, these cannot have implicit receiver caches.  Therefore we use the same variable for either the next link in the unpairedMethodList or the reference to the method's implciit receiver caches, which keeps the header size down.

Instance Variables
	nextMethodOrIRCs:		<0 or CogMethod * or oop>

nextMethodOrIRCs
	- either 0 or the next link in the unpairedMethodList or the first field of a method's implciit receiver caches
!

!NewspeakCogMethod class methodsFor: 'class initialization' stamp: 'eem 2/25/2020 19:06'!
initialize
	"self initialize"
	(Smalltalk classNamed: #CogNewspeakMethodSurrogate32) ifNotNil:
		[:cms32|
		self checkGenerateSurrogate: cms32 bytesPerWord: 4].
	(Smalltalk classNamed: #CogNewspeakMethodSurrogate64) ifNotNil:
		[:cms64|
		self checkGenerateSurrogate: cms64 bytesPerWord: 8]! !

!NewspeakCogMethod class methodsFor: 'translation' stamp: 'eem 8/21/2013 15:53'!
structTypeName
	^#CogMethod! !

!NewspeakCogMethod methodsFor: 'accessing' stamp: 'eem 8/21/2013 11:20'!
nextMethodOrIRCs
	"Answer the value of nextMethodOrIRCs"

	^ nextMethodOrIRCs! !

!NewspeakCogMethod methodsFor: 'accessing' stamp: 'eem 8/21/2013 11:20'!
nextMethodOrIRCs: anObject
	"Set the value of nextMethodOrIRCs"

	^nextMethodOrIRCs := anObject! !

CogMethod subclass: #SistaCogMethod
	instanceVariableNames: 'counters'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!SistaCogMethod commentStamp: 'eem 7/4/2014 11:59' prior: 0!
A SistaCogMethod is a CogMethod with a pointer to memory holding the Sista performance counters decremented in conditional branches.

Instance Variables
	counters:		<pointer>

counters
	- counters points to the first field of either a pinned object on the Spur heap or malloced memory.
!

!SistaCogMethod class methodsFor: 'class initialization' stamp: 'eem 1/23/2017 13:18'!
initialize
	"self initialize"
	(Smalltalk classNamed: #CogSistaMethodSurrogate32) ifNotNil:
		[:cms32|
		self checkGenerateSurrogate: cms32 bytesPerWord: 4].
	(Smalltalk classNamed: #CogSistaMethodSurrogate64) ifNotNil:
		[:cms64|
		self checkGenerateSurrogate: cms64 bytesPerWord: 8]! !

!SistaCogMethod class methodsFor: 'translation' stamp: 'eem 2/21/2011 17:35'!
structTypeName
	"Provide the typedef name."
	^superclass structTypeName! !

!SistaCogMethod methodsFor: 'accessing' stamp: 'eem 7/4/2014 12:01'!
counters
	"Answer the value of counters"

	^ counters! !

!SistaCogMethod methodsFor: 'accessing' stamp: 'eem 7/4/2014 12:03'!
counters: anObject
	"Set the value of counters"

	^counters := anObject! !

VMStructType subclass: #CogBlockStart
	instanceVariableNames: 'fakeHeader fillInstruction numArgs numCopied numInitialNils startpc entryLabel stackCheckLabel span hasInstVarRef'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogBlockStart commentStamp: '<historical>' prior: 0!
I am a record of a block, the address of the fake header, the association between the bytecode startpc and the machine code start label, and the block's relevant metrics.!

!CogBlockStart class methodsFor: 'translation' stamp: 'eem 2/11/2009 23:54'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a BlockStart struct."

	self instVarNames do:
		[:ivn|
		aBinaryBlock
			value: ivn
			value: (ivn caseOf: {
						['entryLabel']		-> [#'AbstractInstruction *'].
						['fakeHeader']		-> [#'AbstractInstruction *'].
						['fillInstruction']	-> [#'AbstractInstruction *'].
						['stackCheckLabel']	-> [#'AbstractInstruction *'] }
					otherwise:
						[#sqInt])]! !

!CogBlockStart class methodsFor: 'translation' stamp: 'eem 2/10/2009 09:34'!
structTypeName
	^self name allButFirst: 3 "Drop initial Cog"! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 1/17/2009 09:27'!
entryLabel
	"Answer the value of entryLabel"

	^ entryLabel! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 1/17/2009 09:27'!
entryLabel: anObject
	"Set the value of entryLabel"

	^entryLabel := anObject! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 2/11/2009 23:52'!
fakeHeader
	"Answer the value of fakeHeader"
	<returnTypeC: #'AbstractInstruction *'>
	^fakeHeader! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 1/16/2009 13:42'!
fakeHeader: anObject
	"Set the value of fakeHeader"

	^fakeHeader := anObject! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 2/11/2009 23:42'!
fillInstruction
	"Answer the value of fillInstruction"
	<returnTypeC: #'AbstractInstruction *'>
	^fillInstruction! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 2/11/2009 23:42'!
fillInstruction: anObject
	"Set the value of fillInstruction"

	^fillInstruction := anObject! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 6/4/2015 09:28'!
hasInstVarRef

	^ hasInstVarRef! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 6/4/2015 09:28'!
hasInstVarRef: anObject

	^hasInstVarRef := anObject! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 1/16/2009 17:46'!
numArgs
	"Answer the value of numArgs"

	^ numArgs! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 1/16/2009 17:46'!
numArgs: anObject
	"Set the value of numArgs"

	^numArgs := anObject! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 1/16/2009 17:46'!
numCopied
	"Answer the value of numCopied"

	^ numCopied! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 1/16/2009 17:46'!
numCopied: anObject
	"Set the value of numCopied"

	^numCopied := anObject! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 12/7/2010 20:51'!
numInitialNils
	"Answer the value of numInitialNils"

	^ numInitialNils! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 12/7/2010 20:52'!
numInitialNils: anObject
	"Set the value of numInitialNils"

	^numInitialNils := anObject! !

!CogBlockStart methodsFor: 'printing' stamp: 'eem 6/4/2015 10:28'!
printStateOn: aStream
	<doNotGenerate> "Smalltalk-side only"
	aStream
		space; nextPut: $(;
		nextPutAll: 'numArgs: '; print: numArgs;
		nextPutAll: ' numCopied: '; print: numCopied;
		nextPutAll: ' numInitialNils: '; print: numInitialNils;
		nextPutAll: ' hasInstVarRef: '; print: hasInstVarRef;
		nextPutAll: ' startpc: '; print: startpc;
		nextPutAll: ' span: '; print: span;
		nextPut: $)! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 1/11/2009 09:37'!
span
	"Answer the value of span"

	^ span! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 1/11/2009 09:37'!
span: anObject
	"Set the value of span"

	^span := anObject! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 2/12/2009 20:33'!
stackCheckLabel
	"Answer the value of stackCheckLabel"
	<returnTypeC: #'AbstractInstruction *'>
	^stackCheckLabel! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 12/7/2010 20:52'!
stackCheckLabel: anObject
	"Set the value of stackCheckLabel"

	^stackCheckLabel := anObject! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 1/11/2009 09:35'!
startpc
	"Answer the value of startpc"

	^ startpc! !

!CogBlockStart methodsFor: 'accessing' stamp: 'eem 1/11/2009 09:37'!
startpc: anObject
	"Set the value of startpc"

	^startpc := anObject! !

VMStructType subclass: #CogBytecodeDescriptor
	instanceVariableNames: 'generator spanFunction needsFrameFunction stackDelta opcode numBytes isBranchTrue isBranchFalse isReturn isBlockCreation isMapped isMappedInBlock isExtension isInstVarRef is1ByteInstVarStore hasIRC hasUnsafeJump'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogBytecodeDescriptor commentStamp: 'eem 11/18/2010 06:32' prior: 0!
I am an entry in the Cogit's dispatch table for bytecodes.  I hold the routine to call to generate code for the partcular bytecode I represent and the number of bytes the bytecode has.  For eliminating temps in frameless blocks I maintain a stack delta for bytecodes that are valid in a frameless block.  The order of my instance variables is chosen for compact struct packing.!

!CogBytecodeDescriptor class methodsFor: 'translation' stamp: 'rsf 1/4/2017 20:41'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a BytecodeDescriptor struct."

	"With hasIRC there are 9 flag bits.  Excluding hasIRC in non-Newspeak VMs keeps the descriptor compact."
	(NewspeakVM
		ifTrue: [self instVarNames]
		ifFalse: [self instVarNames copyWithout: 'hasIRC']) do:
		[:ivn|
		aBinaryBlock
			value: ivn
			value: (ivn caseOf: {
							['generator']			->	[#('sqInt (*' ')(void)')].
							['spanFunction']			->	[#('sqInt NoDbgRegParms (*' ')(struct _BytecodeDescriptor *,sqInt,sqInt,sqInt)')].
							['needsFrameFunction']	->	[#('sqInt NoDbgRegParms (*' ')(sqInt)')].
							['numBytes']			->	['unsigned char'].
							['stackDelta']			->	['signed char'].
							['opcode']				->	['unsigned char'] }
						otherwise: [#('unsigned' ' : 1')])]! !

!CogBytecodeDescriptor class methodsFor: 'translation' stamp: 'eem 2/10/2009 09:34'!
structTypeName
	^self name allButFirst: 3 "Drop initial Cog"! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 12/18/2008 12:48'!
generator
	"Answer the value of generator"

	^ generator! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 12/22/2008 18:48'!
generator: anObject
	"Set the value of generator"

	^generator := anObject! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 7/9/2014 20:09'!
hasIRC
	"Answer the value of hasIRC"

	^ hasIRC! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 7/9/2014 20:09'!
hasIRC: anObject
	"Set the value of hasIRC"

	^hasIRC := anObject! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'cb 1/3/2018 15:11'!
hasUnsafeJump

	^ hasUnsafeJump! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'cb 1/3/2018 15:11'!
hasUnsafeJump: anObject

	^hasUnsafeJump := anObject! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'cb 7/25/2016 10:27'!
is1ByteInstVarStore
	"Answer the value of is1ByteInstVarStore"

	^ is1ByteInstVarStore! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'cb 7/25/2016 10:27'!
is1ByteInstVarStore: anObject
	"Set the value of is1ByteInstVarStore"

	^is1ByteInstVarStore := anObject! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 1/27/2009 12:56'!
isBlockCreation
	"Answer the value of isBlockCreation"

	^isBlockCreation! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 1/27/2009 12:56'!
isBlockCreation: anObject
	"Set the value of isBlockCreation"

	^isBlockCreation := anObject! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 11/15/2012 15:51'!
isBranch
	^self spanFunction notNil and: [self isBlockCreation not]! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 11/18/2010 06:12'!
isBranchFalse
	"Answer the value of isBranchFalse"

	^ isBranchFalse! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 11/18/2010 06:12'!
isBranchFalse: anObject
	"Set the value of isBranchFalse"

	^isBranchFalse := anObject! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 11/18/2010 06:12'!
isBranchTrue
	"Answer the value of isBranchTrue"

	^ isBranchTrue! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 11/18/2010 06:12'!
isBranchTrue: anObject
	"Set the value of isBranchTrue"

	^isBranchTrue := anObject! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 3/2/2017 08:32'!
isCallPrimitive

	^generator == #genCallPrimitiveBytecode! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 1/21/2018 15:46'!
isConditionalBranch
	<inline: true>
	^isBranchTrue or: [isBranchFalse]! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 11/1/2012 10:51'!
isExtension
	"Answer the value of isExtension"

	^ isExtension! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 11/1/2012 10:51'!
isExtension: anObject
	"Set the value of isExtension"

	^isExtension := anObject! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 6/4/2015 09:46'!
isInstVarRef
	"Answer the value of isInstVarRef"

	^ isInstVarRef! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 6/4/2015 09:46'!
isInstVarRef: anObject
	"Set the value of isInstVarRef"

	^isInstVarRef := anObject! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 3/2/2016 17:52'!
isMapped
	"Answer the value of isMapped.  Note we send value because this may be the binding for #IMMUTABILITY"
	^isMapped value! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 2/5/2009 15:03'!
isMapped: anObject
	"Set the value of isMapped"

	^isMapped := anObject! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 12/10/2010 21:52'!
isMappedInBlock
	"Answer the value of isMappedInBlock"

	^isMappedInBlock! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 12/10/2010 21:52'!
isMappedInBlock: anObject
	"Set the value of isMappedInBlock"

	^isMappedInBlock := anObject! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 12/22/2008 19:38'!
isReturn
	"Answer the value of isReturn"

	^ isReturn! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 12/22/2008 19:39'!
isReturn: anObject
	"Set the value of isReturn"

	^isReturn := anObject! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 1/21/2018 15:46'!
isUnconditionalBranch
	<inline: true>
	^self isBranch and: [(isBranchTrue or: [isBranchFalse]) not]! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 1/3/2011 09:29'!
needsFrameFunction
	"Answer the value of needsFrameFunction"

	^needsFrameFunction! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 1/3/2011 09:29'!
needsFrameFunction: aSymbol
	"Set the value of needsFrameFunction"

	^needsFrameFunction := aSymbol! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 12/18/2008 12:48'!
numBytes
	"Answer the value of numBytes"

	^ numBytes! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 12/22/2008 18:48'!
numBytes: anObject
	"Set the value of numBytes"

	^numBytes := anObject! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 11/18/2010 07:24'!
opcode
	"Answer the value of opcode"

	^ opcode! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 11/18/2010 07:24'!
opcode: anObject
	"Set the value of opcode"

	^opcode := anObject! !

!CogBytecodeDescriptor methodsFor: 'translation' stamp: 'eem 3/2/2016 18:00'!
printCInitializerOn: aStream in: aCCodeGenerator
	<doNotGenerate>
	| first |
	first := true.
	aStream nextPut: ${; space.
	self class instVarNamesAndTypesForTranslationDo:
		[:ivn :type| | value |
		first ifTrue: [first := false] ifFalse: [aStream nextPut: $,; space].
		value := self instVarNamed: ivn.
		value isVariableBinding
			ifTrue: [aStream nextPutAll: value key]
			ifFalse:
				[(#(#'unsigned char' #'signed char' #('unsigned' ' : 1')) includes: type)
					ifTrue: [value isInteger
								ifTrue: [ivn = 'opcode'
											ifTrue: [aStream nextPutAll: (CogRTLOpcodes nameForOpcode: value)]
											ifFalse: [aStream print: value]]
								ifFalse: [aStream nextPut: ((value notNil and: [value]) ifTrue: [$1] ifFalse: [$0])]]
					ifFalse: [(false and: [#('spanFunction' 'isBackwardBranchFunction') includes: ivn]) ifTrue:
								[aStream nextPut: $(; nextPutAll: type first; nextPutAll: type last; nextPut: $)].
							aStream nextPutAll: (value
													ifNotNil: [aCCodeGenerator cFunctionNameFor: value]
													ifNil: ['0'])]]].
	aStream space; nextPut: $}! !

!CogBytecodeDescriptor methodsFor: 'debug printing' stamp: 'eem 11/16/2012 21:07'!
printStateOn: aStream
	<doNotGenerate> "Smalltalk-side only"
	generator ifNil:
		[^self].
	aStream space; nextPut: $(; nextPutAll: generator; space; print: numBytes.
	needsFrameFunction ifNotNil: [aStream space; nextPutAll: needsFrameFunction].
	spanFunction ifNotNil: [:func| aStream space; nextPutAll: func].
	isReturn ifTrue: [aStream nextPutAll: ' isReturn'].
	opcode ifNotNil: [aStream space; nextPutAll: (CogRTLOpcodes nameForOpcode: opcode)].
	aStream nextPut: $)! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 11/2/2012 14:02'!
spanFunction
	"Answer the value of spanFunction"
	<returnTypeC: 'sqInt (*)(BytecodeDescriptor *,sqInt,sqInt,sqInt)'>
	^spanFunction! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 11/2/2012 14:03'!
spanFunction: functionPointer
	"Set the value of spanFunction"
	<returnTypeC: 'sqInt (*)(BytecodeDescriptor *,sqInt,sqInt,sqInt)'>
	<var: #functionPointer declareC: 'sqInt (*functionPointer)(BytecodeDescriptor *,sqInt,sqInt,sqInt)'>
	^spanFunction := functionPointer! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 6/21/2010 14:43'!
stackDelta
	"Answer the value of stackDelta"

	^stackDelta! !

!CogBytecodeDescriptor methodsFor: 'accessing' stamp: 'eem 6/21/2010 14:43'!
stackDelta: anObject
	"Set the value of  stackDelta"

	^ stackDelta := anObject! !

VMStructType subclass: #CogBytecodeFixup
	instanceVariableNames: 'targetInstruction instructionIndex bcpc'
	classVariableNames: 'NeedsFixupFlag'
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogBytecodeFixup commentStamp: 'eem 1/20/2011 12:41' prior: 0!
I am a fixup for a bytecode in the Cogit.  Currently fixups are for labels only.  To fixup jumps the cogit places fixups in the fixups array at indices that correspond to bytecodes that are the targets of jumps.  When the cogit encounters a bytecode with a fixup it assigns the fixup's target field to the first generated instruction for the bytecode.  Later when AbstractInstruction Jump* instructions attempt to compute their target they follow the indirection through the fixup to the actual target.!

!CogBytecodeFixup class methodsFor: 'simulation only' stamp: 'eem 7/18/2013 11:53'!
byteSizeForSimulator: aVMClass
	"Answer an approximation of the byte size of an AbstractInstruction struct.
	 This is for estimating the alloca in allocateOpcodes:bytecodes:ifFail:"
	^self instSize * (aVMClass sizeof: #'void *')! !

!CogBytecodeFixup class methodsFor: 'translation' stamp: 'eem 3/2/2017 11:53'!
filteredInstVarNames
	"Override to eliminate bcpc,"
	^super filteredInstVarNames copyWithout: 'bcpc'! !

!CogBytecodeFixup class methodsFor: 'instance creation' stamp: 'cb 4/17/2016 16:59'!
for: aCogit
	^self new! !

!CogBytecodeFixup class methodsFor: 'class initialization' stamp: 'cb 4/15/2016 17:16'!
initialize
	"Initialize the fixup flags.  In this class we have only one flag, 1, which
	 means set the fixup to point to the first instruction for a particular bytecode.
	 A forward branch from one bytecode to a later one will set its jmpTarget to
	 a fixup.  later, in compileAbstractInstructionsFrom:to:, any fixup with its
	 targetInstruction set to NeedsFixupFlag will have its targetInstruction set
	 to the first bytecode of the sequence.  Later still, when code is generated
	 jumps follow fixups to eliminate the fixup and target the rigth instruction."
	NeedsFixupFlag := 1! !

!CogBytecodeFixup class methodsFor: 'translation' stamp: 'eem 2/2/2018 10:42'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"Enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a BytecodeFixup struct."
	"self withAllSubclasses collect: [:ea| ea typedef]"

	self filteredInstVarNames do:
		[:ivn|
		 aBinaryBlock
			value: ivn
			value: (ivn first ~= $# ifTrue:
						[ivn caseOf: {
							['targetInstruction']				-> [#'AbstractInstruction *'].
							['mergeSimStack']				-> [#'SimStackEntry *'].
							['instructionIndex']				-> [#'unsigned short'].
							['simStackPtr']					-> [#'unsigned char'].
							['simNativeStackPtr']			-> [#'short'].
							['simNativeStackSize']			-> [#'unsigned short'].
							['isTargetOfBackwardBranch']	-> [#char] }])]! !

!CogBytecodeFixup class methodsFor: 'translation' stamp: 'eem 11/22/2010 09:42'!
structTypeName
	^'BytecodeFixup'! !

!CogBytecodeFixup methodsFor: 'coercion' stamp: 'eem 2/15/2009 10:22'!
asInteger
	<doNotGenerate>
	^self! !

!CogBytecodeFixup methodsFor: 'accessing' stamp: 'eem 12/2/2016 11:37'!
bcpc

	^ bcpc! !

!CogBytecodeFixup methodsFor: 'accessing' stamp: 'eem 12/2/2016 11:37'!
bcpc: anObject

	^bcpc := anObject! !

!CogBytecodeFixup methodsFor: 'converting' stamp: 'cb 4/15/2016 17:31'!
becomeFixup
	<inline: true>
	targetInstruction := self cCoerceSimple: NeedsFixupFlag to: #'AbstractInstruction *'! !

!CogBytecodeFixup methodsFor: 'coercion' stamp: 'cb 4/15/2016 17:50'!
cCoerceSimple: flagOrAbstractOp to: cType
	<doNotGenerate>
	^ flagOrAbstractOp! !

!CogBytecodeFixup methodsFor: 'instance initialization' stamp: 'eem 2/2/2018 09:52'!
initialize
	targetInstruction := 0! !

!CogBytecodeFixup methodsFor: 'accessing' stamp: 'eem 2/11/2009 21:27'!
instructionIndex
	"Answer the value of instructionIndex"

	^instructionIndex! !

!CogBytecodeFixup methodsFor: 'accessing' stamp: 'eem 2/11/2009 21:28'!
instructionIndex: anObject
	"Set the value of instructionIndex"

	^instructionIndex := anObject! !

!CogBytecodeFixup methodsFor: 'testing' stamp: 'cb 4/15/2016 17:29'!
needsFixup
	<inline: true>
	^ targetInstruction asInteger = NeedsFixupFlag! !

!CogBytecodeFixup methodsFor: 'testing' stamp: 'cb 4/15/2016 17:18'!
notAFixup
	<inline: true>
	^ targetInstruction = 0! !

!CogBytecodeFixup methodsFor: 'debug printing' stamp: 'eem 12/2/2016 11:46'!
printStateOn: aStream
	<doNotGenerate>
	targetInstruction ifNotNil:
		[aStream space; nextPut: $(; print: targetInstruction; nextPutAll: ' bc '; print: bcpc; nextPut: $)]! !

!CogBytecodeFixup methodsFor: 'simulation' stamp: 'eem 3/4/2017 14:34'!
recordBcpc: theBytecodePC
	<inline: true>
	self cCode: '' inSmalltalk:
		[(bcpc isNil or: [bcpc = theBytecodePC])
			ifTrue: [bcpc := theBytecodePC]
			ifFalse:
				[bcpc := bcpc isInteger
							ifTrue: [{bcpc. theBytecodePC}]
							ifFalse:
								[(bcpc includes: theBytecodePC) ifTrue: [^self].
								 bcpc, {theBytecodePC}]]]! !

!CogBytecodeFixup methodsFor: 'accessing' stamp: 'eem 2/11/2009 11:51'!
targetInstruction
	"Answer the value of targetInstruction"

	^ targetInstruction! !

!CogBytecodeFixup methodsFor: 'accessing' stamp: 'eem 2/11/2009 11:51'!
targetInstruction: anObject
	"Set the value of targetInstruction"

	^targetInstruction := anObject! !

CogBytecodeFixup subclass: #CogSSBytecodeFixup
	instanceVariableNames: 'simStackPtr isTargetOfBackwardBranch simNativeStackPtr simNativeStackSize'
	classVariableNames: 'NeedsMergeFixupFlag NeedsNonMergeFixupFlag'
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogSSBytecodeFixup commentStamp: 'eem 2/2/2018 10:51' prior: 0!
A CogSSBytecodeFixup extends CogBytecodeFixup with state to merge the stack at control-flow joins.  At a join the code generator must ensure that the stack is spilled to the same point along both branches and that the simStackPtr is correct.

Instance Variables
	simStackPtr:					<Integer>
	isTargetOfBackwardBranch:		<Boolean>

simStackPtr
	- the simStackPtr at the jump to this fixup.  It should either agree with the incoming fixup if control continues, or replace the simStackPtr if control doesn't continue (the incoming control flow ended with a return)

isTargetOfBackwardBranch:
	- if true the fixup is the target of some backward branch!

CogSSBytecodeFixup subclass: #CogRASSBytecodeFixup
	instanceVariableNames: 'cogit mergeSimStack'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogRASSBytecodeFixup commentStamp: 'eem 11/22/2016 08:44' prior: 0!
A CogRASSBytecodeFixup extends CogSSBytecodeFixup with state to merge the stack at control-flow joins, preserving register contents.  By holding onto the entire stack state a CogRASSBytecodeFixup allows RegisterAllocatingCogit to merge individual stack entries, instead of merely spilling to the same height.

Instance Variables
	cogit:					<RegisterAllocatingCogit>
	mergeSimStack:		<Array of: CogRegisterAllocatingSimStackEntry>

cogit
	- the JIT compiler

mergeSimStack
	- the state of the stack at the jump to this fixup!

!CogRASSBytecodeFixup class methodsFor: 'translation' stamp: 'eem 2/2/2018 12:54'!
filteredInstVarNames
	"Override to group char and short vars together for compactness.
	 self typedef"
	| vars |
	vars := super filteredInstVarNames asOrderedCollection.
	vars
		remove: 'mergeSimStack';
		add: 'mergeSimStack' afterIndex: (vars indexOf: 'targetInstruction').
	^vars! !

!CogRASSBytecodeFixup class methodsFor: 'instance creation' stamp: 'cb 4/17/2016 17:00'!
for: aCogit
	^self new cogit: aCogit! !

!CogRASSBytecodeFixup methodsFor: 'initialize-release' stamp: 'eem 1/25/2018 14:17'!
cogit: aCogit
	cogit := aCogit.
	^self! !

!CogRASSBytecodeFixup methodsFor: 'debug printing' stamp: 'eem 1/18/2017 08:43'!
hasMergeSimStack
	^self needsFixup and: [mergeSimStack notNil]! !

!CogRASSBytecodeFixup methodsFor: 'accessing' stamp: 'eem 1/18/2017 14:26'!
mergeSimStack
	<returnTypeC: #'CogSimStackEntry *'>
	^mergeSimStack! !

!CogRASSBytecodeFixup methodsFor: 'accessing' stamp: 'cb 4/16/2016 09:17'!
mergeSimStack: anObject

	^mergeSimStack := anObject! !

!CogRASSBytecodeFixup methodsFor: 'debug printing' stamp: 'eem 11/25/2016 16:32'!
printSimStack
	<doNotGenerate>
	self notAFixup ifFalse:
		[cogit printSimStack: mergeSimStack toDepth: simStackPtr spillBase: -1 on: cogit coInterpreter transcript]! !

!CogRASSBytecodeFixup methodsFor: 'debug printing' stamp: 'eem 11/25/2016 16:33'!
printStateOn: aStream
	<doNotGenerate>
	(targetInstruction isNil and: [simStackPtr isNil]) ifTrue:
		[^self].
	super printStateOn: aStream.
	mergeSimStack ifNotNil:
		[aStream skip: -1; space; nextPut: $(.
		 cogit printSimStack: mergeSimStack toDepth: simStackPtr spillBase: -1 on: aStream.
		 aStream nextPut: $); nextPut: $)]! !

!CogRASSBytecodeFixup methodsFor: 'accessing' stamp: 'eem 1/25/2018 14:17'!
reinitialize
	<inline: true>
	super reinitialize.
	mergeSimStack := nil! !

!CogRASSBytecodeFixup methodsFor: 'debug printing' stamp: 'eem 2/1/2018 14:53'!
simStackPrintString
	<doNotGenerate>
	^String streamContents:
		[:s|
		 self notAFixup
			ifTrue: [s nextPutAll: 'notAFixup']
			ifFalse: [cogit printSimStack: mergeSimStack toDepth: simStackPtr spillBase: -1 on: s]]! !

!CogSSBytecodeFixup class methodsFor: 'simulation only' stamp: 'eem 3/14/2017 09:48'!
byteSizeForSimulator: aVMClass
	"Answer an approximation of the byte size of an AbstractInstruction struct.
	 This is for estimating the alloca in allocateOpcodes:bytecodes:ifFail:
	 self withAllSubclasses collect: [:ea| { ea byteSizeForSimulator: ea basicNew. ea typedef}]"
	^(LowcodeVM ifTrue: [self instSize - 1] ifFalse: [self instSize - 2]) * (aVMClass sizeof: #'void *')! !

!CogSSBytecodeFixup class methodsFor: 'translation' stamp: 'eem 2/2/2018 12:55'!
filteredInstVarNames
	"Override to add ifdef LowcodeVM around the native stack info, and to put
	 char vars before short vars.
	 self typedef"
	| vars |
	vars := super filteredInstVarNames asOrderedCollection.
	vars
		remove: 'instructionIndex';
			add: 'instructionIndex' afterIndex: (vars indexOf: 'simStackPtr');
		remove: 'isTargetOfBackwardBranch';
			add: 'isTargetOfBackwardBranch' afterIndex: (vars indexOf: 'simStackPtr');
		add: '#if LowcodeVM' beforeIndex: (vars indexOf: 'simNativeStackPtr');
		add: '#endif' afterIndex: (vars indexOf: 'simNativeStackSize').
	^vars! !

!CogSSBytecodeFixup class methodsFor: 'class initialization' stamp: 'eem 2/2/2018 12:03'!
initialize
	"Initialize the fixup flags.  In this class we have two states.  A fixup is a bytecode
	 being targeted by a branch, and a jump can target the fixup before the byetcode
	 is generated.  A non-merge fixup is that for a bytecode that follows a return instruction.
	 There is no control flow merge from the preceding instruction for this kind of fixup.
	 A merge fixup is that for a bytecode where control flow arrives from both the preceding
	 instruction and a branch.  When compileAbstractInstructionsFrom:to: finds a merge
	 fixup, it must both set the targetInstruction and merge the stack/register state of the
	 control flow from the preceding instruction with the stack/register state from the branch.
	 Later still, when code is generated jumps follow fixups to eliminate the fixup and target
	 the right instruction."
	NeedsNonMergeFixupFlag := 1.
	NeedsMergeFixupFlag := 2.
	self assert: NeedsNonMergeFixupFlag < NeedsMergeFixupFlag! !

!CogSSBytecodeFixup methodsFor: 'simulation support' stamp: 'nice 7/8/2016 02:33'!
asIntegerPtr
	<doNotGenerate>
	^self! !

!CogSSBytecodeFixup methodsFor: 'simulation support' stamp: 'eem 12/11/2010 15:50'!
asUnsignedInteger
	<doNotGenerate>
	^self! !

!CogSSBytecodeFixup methodsFor: 'simulation support' stamp: 'nice 7/8/2016 02:31'!
asUnsignedIntegerPtr
	<doNotGenerate>
	^self! !

!CogSSBytecodeFixup methodsFor: 'converting' stamp: 'cb 4/15/2016 17:32'!
becomeMergeFixup
	<inline: true>
	targetInstruction := self cCoerceSimple: NeedsMergeFixupFlag to: #'AbstractInstruction *'! !

!CogSSBytecodeFixup methodsFor: 'converting' stamp: 'cb 4/15/2016 17:32'!
becomeNonMergeFixup
	<inline: true>
	targetInstruction := self cCoerceSimple: NeedsNonMergeFixupFlag to: #'AbstractInstruction *'! !

!CogSSBytecodeFixup methodsFor: 'instance initialization' stamp: 'eem 2/2/2018 10:04'!
initialize
	targetInstruction := 0.
	isTargetOfBackwardBranch := false! !

!CogSSBytecodeFixup methodsFor: 'testing' stamp: 'eem 2/2/2018 09:48'!
isBackwardBranchFixup
	^isTargetOfBackwardBranch! !

!CogSSBytecodeFixup methodsFor: 'testing' stamp: 'cb 4/17/2016 11:48'!
isMergeFixup
	<inline: true>
	^ targetInstruction asUnsignedInteger = NeedsMergeFixupFlag! !

!CogSSBytecodeFixup methodsFor: 'testing' stamp: 'cb 4/15/2016 17:39'!
isMergeFixupOrIsFixedUp
	<inline: true>
	^ targetInstruction asUnsignedInteger >= NeedsMergeFixupFlag! !

!CogSSBytecodeFixup methodsFor: 'testing' stamp: 'cb 4/16/2016 18:58'!
isNonMergeFixup
	<inline: true>
	^ targetInstruction asUnsignedInteger = NeedsNonMergeFixupFlag! !

!CogSSBytecodeFixup methodsFor: 'testing' stamp: 'cb 4/15/2016 17:36'!
isNonMergeFixupOrNotAFixup
	<inline: true>
	^ targetInstruction asUnsignedInteger <= NeedsNonMergeFixupFlag! !

!CogSSBytecodeFixup methodsFor: 'testing' stamp: 'eem 3/6/2018 17:49'!
isTargetOfBackwardBranch
	^isTargetOfBackwardBranch! !

!CogSSBytecodeFixup methodsFor: 'testing' stamp: 'cb 4/15/2016 17:29'!
needsFixup
	<inline: true>
	^ targetInstruction asUnsignedInteger between: NeedsNonMergeFixupFlag and: NeedsMergeFixupFlag! !

!CogSSBytecodeFixup methodsFor: 'testing' stamp: 'cb 4/15/2016 17:48'!
needsFixupOrIsFixedUp
	<inline: true>
	^ targetInstruction asUnsignedInteger > 0! !

!CogSSBytecodeFixup methodsFor: 'testing' stamp: 'cb 4/15/2016 17:42'!
notYetFixedUp
	<inline: true>
	^ targetInstruction asUnsignedInteger <= NeedsMergeFixupFlag! !

!CogSSBytecodeFixup methodsFor: 'debug printing' stamp: 'eem 12/2/2016 11:48'!
printStateOn: aStream
	<doNotGenerate>
	(targetInstruction isNil and: [simStackPtr isNil]) ifTrue:
		[^self].
	aStream space; nextPut: $(.
	targetInstruction ifNotNil:
		[aStream space; print: targetInstruction].
	simStackPtr ifNotNil:
		[aStream nextPutAll: ' sp '; print: simStackPtr].
	bcpc ifNotNil:
		[aStream nextPutAll: ' bc '; print: bcpc].
	aStream nextPut: $)! !

!CogSSBytecodeFixup methodsFor: 'accessing' stamp: 'eem 1/29/2020 15:19'!
reinitialize
	<inline: true>
	targetInstruction := 0.
	simStackPtr := 0.
	LowcodeVM ifTrue:
		[simNativeStackPtr := simNativeStackSize := 0]! !

!CogSSBytecodeFixup methodsFor: 'accessing' stamp: 'eem 2/2/2018 09:48'!
setIsBackwardBranchFixup
	<inline: true>
	isTargetOfBackwardBranch := true! !

!CogSSBytecodeFixup methodsFor: 'accessing' stamp: 'RonieSalgado 4/29/2016 16:36'!
simNativeStackPtr
	"Answer the value of simStackPtr"

	^ simNativeStackPtr! !

!CogSSBytecodeFixup methodsFor: 'accessing' stamp: 'RonieSalgado 4/29/2016 16:36'!
simNativeStackPtr: anObject
	"Set the value of simStackPtr"

	^simNativeStackPtr := anObject! !

!CogSSBytecodeFixup methodsFor: 'accessing' stamp: 'RonieSalgado 4/29/2016 16:37'!
simNativeStackSize
	"Answer the value of simStackPtr"

	^ simNativeStackSize! !

!CogSSBytecodeFixup methodsFor: 'accessing' stamp: 'RonieSalgado 4/29/2016 16:37'!
simNativeStackSize: anObject
	"Set the value of simStackPtr"

	^simNativeStackSize := anObject! !

!CogSSBytecodeFixup methodsFor: 'accessing' stamp: 'eem 11/22/2010 14:31'!
simStackPtr
	"Answer the value of simStackPtr"

	^ simStackPtr! !

!CogSSBytecodeFixup methodsFor: 'accessing' stamp: 'eem 11/22/2010 14:31'!
simStackPtr: anObject
	"Set the value of simStackPtr"

	^simStackPtr := anObject! !

VMStructType subclass: #CogPrimitiveDescriptor
	instanceVariableNames: 'primitiveGenerator primNumArgs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogPrimitiveDescriptor class methodsFor: 'translation' stamp: 'eem 1/7/2016 13:04'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a PrimitiveDescriptor struct."

	self instVarNames do:
		[:ivn|
		aBinaryBlock
			value: ivn
			value: (ivn = 'primitiveGenerator'
					ifTrue: [#('sqInt (*' ')(void)')]
					ifFalse: [#sqInt])]! !

!CogPrimitiveDescriptor class methodsFor: 'translation' stamp: 'eem 4/16/2009 12:28'!
structTypeName
	^self name allButFirst: 3 "Drop initial Cog"! !

!CogPrimitiveDescriptor methodsFor: 'accessing' stamp: 'eem 4/17/2009 15:06'!
primNumArgs
	"Answer the value of primNumArgs"

	^primNumArgs! !

!CogPrimitiveDescriptor methodsFor: 'accessing' stamp: 'eem 4/17/2009 15:06'!
primNumArgs: anObject
	"Set the value of primNumArgs"

	^primNumArgs := anObject! !

!CogPrimitiveDescriptor methodsFor: 'accessing' stamp: 'eem 4/17/2009 15:03'!
primitiveGenerator
	"Answer the value of primitiveGenerator"

	^primitiveGenerator! !

!CogPrimitiveDescriptor methodsFor: 'accessing' stamp: 'eem 4/17/2009 15:02'!
primitiveGenerator: anObject
	"Set the value of primitiveGenerator"

	^primitiveGenerator := anObject! !

!CogPrimitiveDescriptor methodsFor: 'translation' stamp: 'eem 4/20/2020 17:48'!
printCInitializerOn: aStream in: aCCodeGenerator
	<doNotGenerate>
	| first hasCompileTimeOptionPragmas |
	hasCompileTimeOptionPragmas := false.
	primitiveGenerator ifNotNil:
		[:sel|
		(aCCodeGenerator methodNamed: sel) ifNotNil:
			[:method|
			 method compileTimeOptionPragmas ifNotEmpty:
				[:ctop|
				 aStream position: aStream position - 1.
				 method outputConditionalDefineFor: ctop on: aStream.
				 aStream tab.
				 hasCompileTimeOptionPragmas := true]]].

	aStream nextPut: ${; space.
	first := true.
	self class instVarNamesAndTypesForTranslationDo:
		[:ivn :type| | value |
		first ifTrue: [first := false] ifFalse: [aStream nextPut: $,; space].
		value := self instVarNamed: ivn.
		aStream nextPutAll: (value
								ifNotNil: [value isSymbol
											ifTrue: [aCCodeGenerator cFunctionNameFor: value]
											ifFalse: [aCCodeGenerator cLiteralFor: value]]
								ifNil: ['0'])].
	aStream space; nextPut: $}.
	hasCompileTimeOptionPragmas ifTrue:
		[aStream nextPut: $,; cr; nextPutAll: '#else'; crtab.
		 self copy primitiveGenerator: nil; primNumArgs: -1; printCInitializerOn: aStream in: aCCodeGenerator.
		 aStream nextPut: $,; cr; nextPutAll: '#endif']! !

VMStructType subclass: #CogSSOptStatus
	instanceVariableNames: 'isReceiverResultRegLive ssEntry'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogSSOptStatus class methodsFor: 'translation' stamp: 'eem 12/11/2010 09:11'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a CogSSOptStatus struct."

	self instVarNames do:
		[:ivn|
		aBinaryBlock
			value: ivn
			value: (ivn = 'ssEntry'
					ifTrue: [#'CogSimStackEntry *']
					ifFalse: [#sqInt])]! !

!CogSSOptStatus methodsFor: 'accessing' stamp: 'eem 11/22/2010 18:21'!
isReceiverResultRegLive
	"Answer the value of isReceiverResultRegLive"

	^ isReceiverResultRegLive! !

!CogSSOptStatus methodsFor: 'accessing' stamp: 'eem 11/22/2010 18:21'!
isReceiverResultRegLive: anObject
	"Set the value of isReceiverResultRegLive"

	^isReceiverResultRegLive := anObject! !

!CogSSOptStatus methodsFor: 'printing' stamp: 'eem 6/15/2015 12:22'!
printStateOn: aStream
	<doNotGenerate>
	(isReceiverResultRegLive notNil
	 or: [ssEntry notNil]) ifTrue:
		[aStream
			nextPut: $(;
			print: isReceiverResultRegLive;
			space.
		ssEntry
			ifNil: [aStream nextPutAll: 'ssEntry is nil']
			ifNotNil: [ssEntry printStateOn: aStream].
		aStream nextPut: $)]! !

!CogSSOptStatus methodsFor: 'accessing' stamp: 'eem 12/1/2010 18:02'!
ssEntry
	"Answer the value of ssEntry"

	^ ssEntry! !

!CogSSOptStatus methodsFor: 'accessing' stamp: 'eem 12/1/2010 18:02'!
ssEntry: anObject
	"Set the value of ssEntry"

	^ssEntry := anObject! !

VMStructType subclass: #CogSimStackEntry
	instanceVariableNames: 'cogit objectRepresentation type spilled liveRegister register offset constant bcptr'
	classVariableNames: ''
	poolDictionaries: 'CogAbstractRegisters CogCompilationConstants CogRTLOpcodes VMStackFrameOffsets'
	category: 'VMMaker-JIT'!

!CogSimStackEntry commentStamp: 'eem 1/22/2018 17:13' prior: 0!
A CogSimStackEntry represents an object pushed on the stack, but during the partial evaluation that occurs as part of the StackToRegisterMappingCogit's compilation.  Bytecodes that produce operands (push items onto the stack) push suitably descriptive instances of CogSimStackEntry onto the simStack (simulation stack).  Bytecodes that consume operands (sends, assignments, returns, etc) take items off the simStack.  Hence the generated code avoids pushing items onto the real stack, and the StackToRegisterMappngCogit can put the operands found on the simStack in registers, etc.  Hence actual stack traffic is much reduced, a much more efficient calling convention is enabled, and so overall performance is increased.  This scheme is due to L. Peter Deutsch and extended here.

Instance Variables
	bcptr:					<Integer>
	cogit:					<StackToRegisterMappingCogit>
	constant:				<Oop>
	liveRegister:			<Integer>
	objectRepresentation:	<CogObjectRepresentation>
	offset:					<Integer>
	register:				<Integer>
	spilled:					<Boolean>
	type:					<Integer from SSBaseOffset, SSConstant, SSRegister or SSSpill>

bcptr
	- the bytecode PC at which this particular entry was created (pushed onto the stack).

cogit
	- the StackToRegisterMappingCogit using this instance

constant
	- if type = SSConstant then this is the constant's oop

liveRegister
	- unused other than for simSelf.  This is here for simSelf and for the subclass CogRegisterAllocatingSimStackEntry

objectRepresentation
	- the CogObjectRepresentation in use for the current object model

offset
	- if type = SSBaseOffset or type = SSSpill then this is the offset from register

register
	- type = SSBaseOffset or type = SSSpill or type = SSRegister then this is the register's code (NoReg, TempReg, ReceiverResultReg et al)

spilled
	- if true, then this entry has been spilled onto the actual stack (or rather code has been generated to push the entry onto the real stack)

type
	- SSBaseOffset, SSConstant, SSRegister or SSSpill!

CogSimStackEntry subclass: #CogRegisterAllocatingSimStackEntry
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JIT'!

!CogRegisterAllocatingSimStackEntry methodsFor: 'compile abstract instructions' stamp: 'eem 11/26/2016 10:37'!
copyLiveRegisterIfSameAs: simStackEntry
	<inline: true>
	self flag: 'should this also apply to SSConstant entries assigned to registers?'.
	(self ~~ simStackEntry
	 and: [type = simStackEntry type
	 and: [type = SSBaseOffset
	 and: [register = simStackEntry register and: [offset = simStackEntry offset]]]]) ifTrue:
		[liveRegister := simStackEntry liveRegister]! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'compile abstract instructions' stamp: 'eem 2/6/2018 10:09'!
ensureSpilledAt: baseOffset from: baseRegister
	spilled ifTrue:
		[type = SSSpill ifTrue:
			[self assert: ((offset = baseOffset and: [register = baseRegister]) or: [cogit violatesEnsureSpilledSpillAssert]).
			 liveRegister := NoReg.
			 ^self]].
	self assert: type ~= SSSpill.
	cogit traceSpill: self.
	"N.B. Keep the type of SSConstant spills as SSConstant so that when joins occur
	 as a result of expressions with constants on the stack, the constant on stack
	 can be recovered.  e.g. as in
		self at: 1 put: (self foo ifTrue: [self bar] ifFalse: [self baz])."
	type = SSConstant
		ifTrue:
			[liveRegister = NoReg
				ifTrue: [cogit genPushConstant: constant]
				ifFalse: [cogit PushR: liveRegister]]
		ifFalse:
			[type = SSBaseOffset
				ifTrue:
					[liveRegister = NoReg
						ifTrue: 
							[cogit MoveMw: offset r: register R: TempReg.
					 		 cogit PushR: TempReg]
						ifFalse: [cogit PushR: liveRegister]]
				ifFalse:
					[self assert: type = SSRegister.
					 cogit PushR: register].
			 type := SSSpill].
	liveRegister := NoReg.
	spilled := true.
	offset := baseOffset.
	register := baseRegister! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'comparing' stamp: 'eem 2/11/2017 10:37'!
isIdenticalEntryAs: ssEntry
	<var: 'ssEntry' type: #'CogSimStackEntry *'>
	^type = ssEntry type
	  and: [liveRegister = ssEntry liveRegister
	  and: [((type = SSBaseOffset or: [type == SSSpill]) and: [offset = ssEntry offset and: [register = ssEntry register]])
		or: [(type = SSRegister and: [register = ssEntry register])
		or: [(type = SSConstant and: [constant = ssEntry constant])]]]]! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'comparing' stamp: 'eem 1/26/2018 16:23'!
isMergedWithTargetEntry: targetEntry
	"The receiver is a simStackEntry at a jump to the corresponding simStackEntry at the jump's target.
	 Answer if no merge is required for the jump."
	<var: 'ssEntry' type: #'CogSimStackEntry *'>
	spilled ~= targetEntry spilled ifTrue: "push or pop required"
		[^false].
	(liveRegister = NoReg and: [targetEntry liveRegister ~= NoReg]) ifTrue: "register load required"
		[^false].
	(self isSameEntryAs: targetEntry) ifTrue:
		[^liveRegister = targetEntry liveRegister].
	(type = SSConstant and: [targetEntry type = SSRegister and: [liveRegister = targetEntry register]]) ifTrue:
		[^true].
	"self: const =1 (16r1) (live: Extra4Reg) {172} vs reg ReceiverResultReg {127}"
	"self: reg ReceiverResultReg {95} vs reg Extra5Reg {85}"
	"self: (bo ReceiverResultReg+296 (live: Extra5Reg) {88} vs reg ReceiverResultReg {84}"
	"self: const =1 (16r1) (spilled) {167} vs spill @ FPReg-48 {122}"
	((type = SSConstant and: [targetEntry type = SSRegister and: [liveRegister ~= targetEntry registerOrNone]])
	 or: [(type = SSRegister and: [targetEntry type = SSRegister and: [register ~= targetEntry registerOrNone]])
	 or: [(type = SSBaseOffset and: [register = ReceiverResultReg and: [targetEntry type = SSRegister]])
	 or: [(type = SSConstant and: [targetEntry type = SSSpill])]]]) ifFalse:
		[self halt: 'comment the incompatible pair please'].
	^false! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'accessing' stamp: 'cb 4/16/2016 11:22'!
liveRegister

	^ liveRegister! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/23/2016 15:31'!
offset
	"Answer the value of offset"
	self assert: (type = SSBaseOffset or: [type = SSSpill]).
	^offset! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'compile abstract instructions' stamp: 'eem 2/4/2018 12:46'!
popToReg: reg
	liveRegister ~= NoReg
		ifTrue: 
			[self deny: (type = SSRegister and: [register ~= liveRegister and: [cogit needsFrame]]).
			 spilled ifTrue: "This is rare, and in some cases it isn't even needed (e.g. frameful return) but we can't tell as yet."
				[cogit AddCq: objectRepresentation wordSize R: SPReg].
			 reg ~= liveRegister
				ifTrue: [cogit MoveR: liveRegister R: reg]
				ifFalse: [cogit Label]]
		ifFalse: 
			[spilled
				ifTrue:
					[cogit PopR: reg]
				ifFalse:
					[type caseOf: {
						[SSBaseOffset]	-> [cogit MoveMw: offset r: register R: reg].
						[SSSpill]		-> [cogit MoveMw: offset r: register R: reg].
						[SSConstant]	-> [cogit genMoveConstant: constant R: reg].
						[SSRegister]	-> [reg ~= register
												ifTrue: [cogit MoveR: register R: reg]
												ifFalse: [cogit Label]] }]].

	(reg ~= TempReg and: [liveRegister = NoReg and: [type ~= SSRegister]]) ifTrue:
		[liveRegister := reg.
		 cogit copyLiveRegisterToCopiesOf: self]! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'compile abstract instructions' stamp: 'eem 2/19/2018 13:58'!
popToRegNoAssign: reg
	liveRegister ~= NoReg
		ifTrue: 
			[self deny: (type = SSRegister and: [register ~= liveRegister and: [cogit needsFrame]]).
			 spilled ifTrue: "This is rare, and in some cases it isn't even needed (e.g. frameful return) but we can't tell as yet."
				[cogit AddCq: objectRepresentation wordSize R: SPReg].
			 reg ~= liveRegister
				ifTrue: [cogit MoveR: liveRegister R: reg]
				ifFalse: [cogit Label]]
		ifFalse: 
			[spilled
				ifTrue:
					[cogit PopR: reg]
				ifFalse:
					[type caseOf: {
						[SSBaseOffset]	-> [cogit MoveMw: offset r: register R: reg].
						[SSSpill]		-> [cogit MoveMw: offset r: register R: reg].
						[SSConstant]	-> [cogit genMoveConstant: constant R: reg].
						[SSRegister]	-> [reg ~= register
												ifTrue: [cogit MoveR: register R: reg]
												ifFalse: [cogit Label]] }]]! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'printing' stamp: 'eem 12/7/2016 18:43'!
printStateOn: aStream
	<doNotGenerate> "Smalltalk-side only"
	type isInteger ifFalse: [^self].
	aStream nextPut: $(.
	type caseOf: {
		[SSBaseOffset]	-> [aStream
								nextPutAll: 'bo ';
								nextPutAll: (cogit backEnd nameForRegister: register).
							offset negative ifFalse: [aStream nextPut: $+].
							aStream print: offset].
		[SSConstant]	-> [aStream
								nextPutAll: 'const ';
								nextPutAll: (cogit coInterpreter shortPrint: constant)].
		[SSRegister]	-> [aStream
								nextPutAll: 'reg ';
								nextPutAll: (cogit backEnd nameForRegister: register)].
		[SSSpill]		-> [aStream
								nextPutAll: 'spill @ ';
								nextPutAll: (cogit backEnd nameForRegister: register).
							offset negative ifFalse: [aStream nextPut: $+].
							aStream print: offset] }.
	(spilled and: [type ~= SSSpill]) ifTrue:
		[aStream nextPutAll: ' (spilled)'].
	liveRegister ~= NoReg ifTrue:
		[aStream nextPutAll: ' (live: '; nextPutAll: (liveRegister ifNil: ['NIL!!!!'] ifNotNil: [cogit backEnd nameForRegister: liveRegister]); nextPut: $)].
	bcptr ifNotNil:
		[aStream space; nextPut: ${; print: bcptr; nextPut: $}].
	aStream nextPut: $)! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'compile abstract instructions' stamp: 'eem 2/19/2018 15:52'!
reconcileWith: targetEntry spillOffset: spillOffset onSpillOrUnspill: spillOrUnspillBlock
	"Make the state of a targetEntry, a stack entry following a non-inlined special selector
	 send, the same as the corresponding entry (the receiver) along the inlined path.
	 spillOffset is zero for non-spill locations (self & temps), and the offset of the spill for
	 volatile stack entries. spillOrUnspillBlock is a block evaluated with the target's
	 registerOrNone if the receiver and target have different spilledness.
	 Answer if the reconciliation merged a register; merged registers must be deassigned."
	<var: #targetEntry type: #'SimStackEntry *'>
	<inline: true>
	| targetReg mergedRegister |
	spilled = targetEntry spilled ifTrue:
		[self assert: ((self isSameEntryAs: targetEntry)
					 or: [(targetEntry spilled not and: [targetEntry registerOrNone ~= NoReg])
					 or: [spilled and: [type = SSConstant and: [offset = targetEntry offset]]]]).
		 (targetReg := targetEntry registerOrNone) = NoReg ifTrue:
			[liveRegister := NoReg.
			 ^false].
		 mergedRegister := false.
		 type caseOf: {
			[SSBaseOffset]	-> [liveRegister ~= targetReg ifTrue:
									[liveRegister = NoReg
										ifTrue: [cogit MoveMw: offset r: register R: targetReg]
										ifFalse: [cogit MoveR: liveRegister R: targetReg].
									 mergedRegister := true].
								targetEntry type caseOf: {
									[SSBaseOffset]	-> [liveRegister := targetReg.
														(self isSameEntryAs: targetEntry) ifFalse:
															[type := SSSpill.
															 offset := spillOffset]].
									[SSSpill]		-> [liveRegister := targetReg. type := SSSpill.
														offset := spillOffset].
									[SSConstant]	-> [liveRegister := targetReg. type := SSSpill.
														offset := spillOffset].
									[SSRegister]	-> [register := targetReg. type := SSRegister. liveRegister := NoReg] }].
			[SSSpill]		-> [liveRegister = NoReg
									ifTrue: [cogit MoveMw: offset r: register R: targetReg]
									ifFalse: [cogit MoveR: liveRegister R: targetReg].
								liveRegister := targetReg.
								mergedRegister := true].
			[SSConstant]	-> [(targetEntry type = SSConstant
								 and: [targetEntry constant = constant
								 and: [liveRegister = targetReg]]) ifFalse:
									[liveRegister = NoReg
										ifTrue: [cogit genMoveConstant: constant R: targetReg]
										ifFalse: [cogit MoveR: liveRegister R: targetReg].
									type := SSRegister. register := targetReg. liveRegister := NoReg.
									mergedRegister := true]].
			[SSRegister]	-> [targetReg ~= register ifTrue:
									[cogit MoveR: register R: targetReg.
									 register := targetReg. liveRegister := NoReg.
									 mergedRegister := true]] }.
		 ^mergedRegister].
	targetReg := targetEntry registerOrNone.
	spillOrUnspillBlock value: targetReg.
	(type = SSConstant
	 and: [targetEntry type ~= SSConstant or: [targetEntry constant ~= constant]]) ifTrue:
		[type := SSSpill. offset := spillOffset. register := FPReg].
	(spilled not and: [type = SSSpill]) ifTrue:
		[self assert: targetReg ~= NoReg. type := SSRegister. register := targetReg].
	liveRegister ~= targetReg ifTrue:
		[liveRegister := NoReg.
		 ^true].
	^false! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/23/2016 15:31'!
register
	"Answer the value of register"
	self assert: (type = SSBaseOffset or: [type = SSRegister or: [type = SSSpill]]).
	^register! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/23/2016 12:41'!
registerMask
	"Answer a bit mask for the receiver's register, if any."
	liveRegister ~= NoReg ifTrue:
		[^cogit registerMaskFor: liveRegister].
	^super registerMask! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/30/2016 11:20'!
registerMaskOrNone
	liveRegister ~= NoReg ifTrue:
		[^cogit registerMaskFor: liveRegister].
	^super registerMaskOrNone! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/22/2016 13:04'!
registerOrNone
	liveRegister ~= NoReg ifTrue:
		[^liveRegister].
	^super registerOrNone! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'compile abstract instructions' stamp: 'eem 2/10/2017 15:36'!
storeToReg: reg
	liveRegister ~= NoReg
		ifTrue:
			[self deny: (type = SSRegister and: [register ~= liveRegister]).
			 reg ~= liveRegister
				ifTrue: [cogit MoveR: liveRegister R: reg]
				ifFalse: [cogit Label]]
		ifFalse:
			[type caseOf: {
				[SSBaseOffset]	-> [cogit MoveMw: offset r: register R: reg].
				[SSSpill]		-> [cogit MoveMw: offset r: register R: reg].
				[SSConstant]	-> [cogit genMoveConstant: constant R: reg].
				[SSRegister]	-> [reg ~= register
											ifTrue: [cogit MoveR: register R: reg]
											ifFalse: [cogit Label]] }].

	(reg ~= TempReg and: [liveRegister = NoReg and: [type ~= SSRegister]]) ifTrue:
		[liveRegister := reg.
		 cogit copyLiveRegisterToCopiesOf: self]! !

!CogRegisterAllocatingSimStackEntry methodsFor: 'compile abstract instructions' stamp: 'eem 2/19/2018 13:59'!
storeToRegNoAssign: reg
	liveRegister ~= NoReg
		ifTrue:
			[self deny: (type = SSRegister and: [register ~= liveRegister]).
			 reg ~= liveRegister
				ifTrue: [cogit MoveR: liveRegister R: reg]
				ifFalse: [cogit Label]]
		ifFalse:
			[type caseOf: {
				[SSBaseOffset]	-> [cogit MoveMw: offset r: register R: reg].
				[SSSpill]		-> [cogit MoveMw: offset r: register R: reg].
				[SSConstant]	-> [cogit genMoveConstant: constant R: reg].
				[SSRegister]	-> [reg ~= register
											ifTrue: [cogit MoveR: register R: reg]
											ifFalse: [cogit Label]] }]! !

!CogSimStackEntry class methodsFor: 'simulation only' stamp: 'cb 4/16/2016 10:10'!
byteSizeForSimulator: aVMClass
	"Answer an approximation of the byte size of an AbstractInstruction struct.
	 This is for estimating the alloca in allocateOpcodes:bytecodes:ifFail:"
	^self instSize * (aVMClass sizeof: #'void *')! !

!CogSimStackEntry class methodsFor: 'translation' stamp: 'eem 2/25/2017 11:02'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"Enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a CogSimStackEntry struct."
	"self typedef"
	self filteredInstVarNames do:
		[:ivn|
		aBinaryBlock
			value: (ivn = 'register' ifTrue: ['registerr'] ifFalse: [ivn]) "avoid reservedWord conflict"
			value: (ivn caseOf: {
						['type']			-> [#char].
						['spilled']		-> [#char].
						['register']		-> [#'signed char']. "because NoReg = -1"
						['liveRegister']	-> [#'signed char'].}
					otherwise:
						[#sqInt])]! !

!CogSimStackEntry class methodsFor: 'translation' stamp: 'eem 4/18/2016 09:15'!
structTypeName
	^'SimStackEntry'! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/27/2010 14:27'!
bcptr
	"Answer the value of bcptr"

	^ bcptr! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/27/2010 14:27'!
bcptr: anObject
	"Set the value of bcptr"

	^bcptr := anObject! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 2/18/2011 15:39'!
cogit
	"Answer the value of cogit"
	<doNotGenerate>
	^cogit! !

!CogSimStackEntry methodsFor: 'initialize-release' stamp: 'eem 11/29/2010 14:32'!
cogit: aCogit
	<doNotGenerate>
	cogit := aCogit.
	objectRepresentation := cogit objectRepresentation.
	^self! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/20/2010 11:21'!
constant
	"Answer the value of constant"
	self assert: type = SSConstant.
	^constant! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/20/2010 10:33'!
constant: anObject
	"Set the value of constant"

	^constant := anObject! !

!CogSimStackEntry methodsFor: 'compile abstract instructions' stamp: 'eem 1/24/2018 13:37'!
ensureSpilledAt: baseOffset from: baseRegister
	| inst |
	<var: #inst type: #'AbstractInstruction *'>
	spilled ifTrue:
		[type = SSSpill ifTrue:
			[self assert: ((offset = baseOffset and: [register = baseRegister]) or: [cogit violatesEnsureSpilledSpillAssert]).
			 ^self]].
	self assert: type ~= SSSpill.
	cogit traceSpill: self.
	type = SSConstant
		ifTrue:
			[inst := cogit genPushConstant: constant]
		ifFalse:
			[type = SSBaseOffset
				ifTrue:
					[cogit MoveMw: offset r: register R: TempReg.
					 inst := cogit PushR: TempReg]
				ifFalse:
					[self assert: type = SSRegister.
					 inst := cogit PushR: register].
			 type := SSSpill.
			 offset := baseOffset.
			 register := baseRegister].
	spilled := true.! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'RonieSalgado 4/30/2016 23:33'!
floatRegisterMask
	^ 0! !

!CogSimStackEntry methodsFor: 'comparing' stamp: 'eem 2/19/2018 12:53'!
isFrameSimSelf
	"Answer if the receiver is self.  This only works in a frameful method, hence the weird name."
	<inline: true>
	^type = SSBaseOffset and: [register = FPReg and: [offset = FoxMFReceiver]]! !

!CogSimStackEntry methodsFor: 'comparing' stamp: 'eem 2/10/2018 11:34'!
isFrameTempVar
	"Answer if the receiver is a temporary variable.  This
	 only works in a frameful method, hence the weird name."
	<inline: true>
	^type = SSBaseOffset and: [register = FPReg and: [offset ~= FoxMFReceiver]]! !

!CogSimStackEntry methodsFor: 'comparing' stamp: 'eem 2/19/2018 13:17'!
isFrameVar
	"Answer if the receiver is a temporary variable or self.  This
	 only works in a frameful method, hence the weird name."
	<inline: true>
	^type = SSBaseOffset and: [register = FPReg]! !

!CogSimStackEntry methodsFor: 'comparing' stamp: 'eem 1/18/2017 13:55'!
isSameEntryAs: ssEntry
	<var: 'ssEntry' type: #'CogSimStackEntry *'>
	^type = ssEntry type
	  and: [((type = SSBaseOffset or: [type == SSSpill]) and: [offset = ssEntry offset and: [register = ssEntry register]])
		or: [(type = SSRegister and: [register = ssEntry register])
		or: [(type = SSConstant and: [constant = ssEntry constant])]]]! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'cb 4/16/2016 11:22'!
liveRegister

	^ liveRegister! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'cb 4/16/2016 11:22'!
liveRegister: anObject

	^liveRegister := anObject! !

!CogSimStackEntry methodsFor: 'comparing' stamp: 'cb 9/13/2019 12:25'!
mayBeAForwarder
	"Receiver is not a forwarder, except in blocks with no inst var access.
	 For now we optimize only the case where receiver is accessed in a method."
	(type == SSRegister and: 
		[cogit isNonForwarderReceiver: register]) ifTrue: [^false].
	^ type ~= SSConstant! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 1/7/2011 14:10'!
maybeConstant
	"Answer the value of constant, no assert check."
	<cmacro: '(sse) ((sse)->constant)'>
	^constant! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/20/2010 22:10'!
offset
	"Answer the value of offset"
	self assert: type = SSBaseOffset.
	^offset! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/20/2010 10:32'!
offset: anObject
	"Set the value of offset"

	^offset := anObject! !

!CogSimStackEntry methodsFor: 'compile abstract instructions' stamp: 'eem 1/18/2017 13:47'!
popToReg: reg
	spilled
		ifTrue:
			[cogit PopR: reg]
		ifFalse:
			[type caseOf: {
				[SSBaseOffset]	-> [cogit MoveMw: offset r: register R: reg].
				[SSConstant]	-> [cogit genMoveConstant: constant R: reg].
				[SSRegister]	-> [reg ~= register
										ifTrue: [cogit MoveR: register R: reg]
										ifFalse: [cogit Label]] }]! !

!CogSimStackEntry methodsFor: 'printing' stamp: 'eem 11/21/2016 16:18'!
printStateOn: aStream
	<doNotGenerate> "Smalltalk-side only"
	type isInteger ifFalse: [^self].
	aStream nextPut: $(.
	type caseOf: {
		[SSBaseOffset]	-> [aStream
								nextPutAll: 'bo ';
								nextPutAll: (cogit backEnd nameForRegister: register).
							offset negative ifFalse: [aStream nextPut: $+].
							aStream print: offset].
		[SSConstant]	-> [aStream
								nextPutAll: 'const ';
								nextPutAll: (cogit coInterpreter shortPrint: constant)].
		[SSRegister]	-> [aStream
								nextPutAll: 'reg ';
								nextPutAll: (cogit backEnd nameForRegister: register)].
		[SSSpill]		-> [aStream
								nextPutAll: 'spill @ ';
								nextPutAll: (cogit backEnd nameForRegister: register).
							offset negative ifFalse: [aStream nextPut: $+].
							aStream print: offset] }.
	(spilled and: [type ~= SSSpill]) ifTrue:
		[aStream nextPutAll: ' (spilled)'].
	bcptr ifNotNil:
		[aStream space; nextPut: ${; print: bcptr; nextPut: $}].
	aStream nextPut: $)! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/20/2010 22:09'!
register
	"Answer the value of register"
	self assert: (type = SSBaseOffset or: [type = SSRegister]).
	^register! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/20/2010 10:32'!
register: anObject
	"Set the value of register"

	^register := anObject! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/22/2010 13:42'!
registerMask
	"Answer a bit mask for the receiver's register, if any."
	^(type = SSBaseOffset
		or: [type = SSRegister])
			ifTrue: [cogit registerMaskFor: register]
			ifFalse: [0]! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/28/2016 11:52'!
registerMaskOrNone
	^type = SSRegister ifTrue: [cogit registerMaskFor: register] ifFalse: [0]! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 12/10/2015 09:00'!
registerOrNone
	^type = SSRegister ifTrue: [register] ifFalse: [NoReg]! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/20/2010 22:01'!
spilled
	"Answer the value of spilled"

	^ spilled! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/20/2010 22:01'!
spilled: anObject
	"Set the value of spilled"

	^spilled := anObject! !

!CogSimStackEntry methodsFor: 'compile abstract instructions' stamp: 'eem 1/18/2017 13:48'!
storeToReg: reg
	type caseOf: {
		[SSBaseOffset]	-> [cogit MoveMw: offset r: register R: reg].
		[SSSpill]		-> [cogit MoveMw: offset r: register R: reg].
		[SSConstant]	-> [cogit genMoveConstant: constant R: reg].
		[SSRegister]	-> [reg ~= register
								ifTrue: [cogit MoveR: register R: reg]
								ifFalse: [cogit Label]] }! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/20/2010 10:32'!
type
	"Answer the value of type"

	^ type! !

!CogSimStackEntry methodsFor: 'accessing' stamp: 'eem 11/20/2010 10:32'!
type: anObject
	"Set the value of type"

	^type := anObject! !

VMStructType subclass: #CogSimStackNativeEntry
	instanceVariableNames: 'cogit objectRepresentation type spilled register registerSecond offset constant constantInt32 constantInt64 constantFloat32 constantFloat64 constantNativePointer bcptr'
	classVariableNames: ''
	poolDictionaries: 'CogAbstractRegisters CogCompilationConstants CogRTLOpcodes'
	category: 'VMMaker-JIT'!

!CogSimStackNativeEntry class methodsFor: 'translation' stamp: 'eem 9/26/2016 18:48'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a CogSimStackEntry struct."
	"self printTypedefOn: Transcript"
	self filteredInstVarNames do:
		[:ivn|
		aBinaryBlock
			value: (ivn = 'register' ifTrue: ['registerr'] ifFalse: [ivn]) "avoid reservedWord conflict"
			value: (ivn caseOf: {
						['constantInt64']			-> [#sqLong].
						['constantFloat32']			-> [#float].
						['constantFloat64']			-> [#double].
						['type']			-> [#char].
						['spilled']		-> [#char].}
					otherwise:
						[#sqInt])]! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
bcptr
	"Answer the value of bcptr"

	^ bcptr! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
bcptr: anObject
	"Set the value of bcptr"

	^bcptr := anObject! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
cogit
	"Answer the value of cogit"
	<doNotGenerate>
	^cogit! !

!CogSimStackNativeEntry methodsFor: 'initialize-release' stamp: 'rsf 12/9/2014 18:11'!
cogit: aCogit
	<doNotGenerate>
	cogit := aCogit.
	objectRepresentation := cogit objectRepresentation.
	^self! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
constantFloat32
	"Answer the value of constant float32"
	self assert: type = SSConstantFloat32.
	^constantFloat32! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
constantFloat32: aFloat32
	"Set the value of constant float32"

	^constantFloat32 := aFloat32! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
constantFloat64
	"Answer the value of constant float64"
	self assert: type = SSConstantFloat64.
	^constantFloat64! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
constantFloat64: aFloat64
	"Set the value of constant float64"

	^constantFloat64 := aFloat64! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
constantInt32
	"Answer the value of constant int32"
	self assert: type = SSConstantInt32.
	^constantInt32! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
constantInt32: anInteger
	"Set the value of constant int32"

	^constantInt32 := anInteger! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
constantInt64
	"Answer the value of constant int64"
	self assert: type = SSConstantInt64.
	^constantInt64! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
constantInt64: anInteger
	"Set the value of constant int64"

	^constantInt64 := anInteger! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
constantNativePointer
	"Answer the value of constant native pointer"
	self assert: type = SSConstantNativePointer.
	^constantNativePointer! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
constantNativePointer: aNativePointer
	"Set the value of constant int64"

	^constantNativePointer := aNativePointer! !

!CogSimStackNativeEntry methodsFor: 'compile abstract instructions' stamp: 'rsf 9/1/2016 17:22'!
ensureIsMarkedAsSpilled
	spilled ifFalse: [
		type caseOf: {
			[SSNativeRegister]	-> [ type := SSSpillNative ].
			[SSRegisterSingleFloat]	-> [ type := SSSpillFloat32 ].
			[SSRegisterDoubleFloat] -> [ type := SSSpillFloat64 ].
			[SSConstantFloat32] -> [ type := SSSpillFloat32. ].
			[SSConstantFloat64] -> [ type := SSSpillFloat64. ].
			[SSConstantInt32]	-> [ type := SSSpillNative. ].
			[SSConstantNativePointer]	-> [ type := SSSpillNative. ].
		}
	].

	spilled := true.
! !

!CogSimStackNativeEntry methodsFor: 'compile abstract instructions' stamp: 'RonieSalgado 5/2/2016 16:00'!
ensureSpilledSP: spRegister scratchRegister: scratchRegister
	spilled ifFalse: [
		type caseOf: {
			[SSNativeRegister]	-> [
					cogit MoveR: register Mw: offset negated - 1 r: spRegister.
					type := SSSpillNative.
				].
			[SSRegisterSingleFloat]	-> [
					cogit MoveRs: register M32: offset negated - 1 r: spRegister.
					type := SSSpillFloat32.
				].
			[SSRegisterDoubleFloat] -> [
					cogit MoveRd: register M64: offset negated - 1 r: spRegister.
					type := SSSpillFloat64.
				].
			[SSConstantFloat32] -> [
				cogit MoveCw: constantFloat32 asIEEE32BitWord R: scratchRegister.
				cogit MoveR: scratchRegister M32: offset negated - 1 r: spRegister.
				type := SSSpillFloat32.
			].
			[SSConstantFloat64] -> [
				BytesPerWord == 4 ifTrue: [
					cogit MoveCw: (constantFloat64 asIEEE64BitWord bitAnd: 16rFFFFFFFF) R: scratchRegister.
					cogit MoveR: scratchRegister M32: offset negated - 1 r: spRegister.
					cogit MoveCw: (constantFloat64 asIEEE64BitWord bitShift: (self cCoerce: -32 to: #'sqLong')) R: scratchRegister.
					cogit MoveR: scratchRegister M32: offset negated - 1 + 4 r: spRegister.
				] ifFalse: [
					cogit MoveCw: constantFloat32 asIEEE64BitWord R: scratchRegister.
					cogit MoveR: scratchRegister Mw: offset negated - 1 r: spRegister.
				].
				type := SSSpillFloat64.
			].
			[SSConstantInt32]	-> [
					cogit MoveCq: constantInt32 R: scratchRegister.
					cogit MoveR: scratchRegister Mw: offset negated - 1 r: spRegister.
					type := SSSpillNative.
				].
			[SSConstantNativePointer]	-> [
					cogit MoveCw: constantNativePointer R: scratchRegister.
					cogit MoveR: scratchRegister Mw: offset negated - 1 r: spRegister.
					type := SSSpillNative.
				].
		}
	].

	spilled := true.

! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'RonieSalgado 4/30/2016 23:37'!
nativeFloatRegisterMask
	"Answer a bit mask for the receiver's register, if any."
	^ (type = SSRegisterSingleFloat or: [type = SSRegisterDoubleFloat])
			ifTrue: [cogit registerMaskFor: register]
			ifFalse: [0]! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 3/26/2017 16:24'!
nativeFloatRegisterOrNone
	^ (type = SSRegisterSingleFloat or: [type = SSRegisterDoubleFloat])
		 ifTrue: [register]
		 ifFalse: [NoReg]! !

!CogSimStackNativeEntry methodsFor: 'compile abstract instructions' stamp: 'rsf 1/15/2017 20:52'!
nativePopToReg: reg
	spilled ifTrue: [
		cogit loadNativeFramePointerInto: TempReg.

		type caseOf: {
			[SSSpillNative]	-> [cogit MoveMw: offset negated - 1 r: TempReg R: reg].
			[SSSpillInt64]	-> [self assert: BytesPerWord = 8.
				cogit MoveMw: offset negated - 1 r: TempReg R: reg].
			[SSSpillFloat32]	-> [cogit MoveM32: offset negated - 1 r: TempReg Rs: reg.].
			[SSSpillFloat64]	-> [cogit MoveM64: offset negated - 1 r: TempReg Rd: reg.].}
	] ifFalse: [
		type caseOf: {
			[SSNativeRegister]	-> [reg ~= register
											ifTrue: [cogit MoveR: register R: reg]].
			[SSRegisterSingleFloat]	-> [reg ~= register
											ifTrue: [cogit MoveRs: register Rs: reg]].
			[SSRegisterDoubleFloat]	-> [reg ~= register
											ifTrue: [cogit MoveRd: register Rd: reg]].

			[SSConstantInt32]	-> [cogit MoveCq: constantInt32 R: reg].
			[SSConstantNativePointer]	-> [cogit MoveCw: constantNativePointer R: reg].
			[SSConstantFloat32] -> [ cogit MoveCf32: constantFloat32 Rs: reg. ].
			[SSConstantFloat64] -> [ cogit MoveCf64: constantFloat64 Rd: reg. ].
		 }
	].
! !

!CogSimStackNativeEntry methodsFor: 'compile abstract instructions' stamp: 'rsf 12/10/2014 02:18'!
nativePopToReg: reg secondReg: secondReg
	self assert: BytesPerWord = 4.
	spilled ifTrue: [
			cogit PopR: reg; PopR: secondReg
		]
		ifFalse: [
			type caseOf: {
				[SSConstantInt64]	-> [
					cogit MoveCq: (constantInt64 bitAnd: 16rFFFFFFFF) R: reg.
					cogit MoveCq: ((constantInt64 >> 32) bitAnd: 16rFFFFFFFF) R: secondReg.
					].
				[SSRegisterPair]	-> [
					cogit Label.
					reg ~= register ifTrue: [
						registerSecond = reg ifTrue: [
							cogit MoveR: registerSecond R: TempReg.
						].
						cogit MoveR: register R: reg
					].

					registerSecond ~= secondReg ifTrue: [
						registerSecond = reg ifTrue: [
							cogit MoveR: TempReg R: secondReg.
						] ifFalse: [
							cogit MoveR: registerSecond R: secondReg.
						]
					].
				].
			}
		].
! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/10/2014 02:19'!
nativeRegisterMask
	"Answer a bit mask for the receiver's register, if any."
	^(((type = SSBaseOffset
		or: [type = SSNativeRegister])
		or: [type = SSRegisterSingleFloat])
		or: [type = SSRegisterDoubleFloat])
			ifTrue: [cogit registerMaskFor: register]
			ifFalse: [
				type = SSRegisterPair ifTrue: [(cogit registerMaskFor: register) bitOr: (cogit registerMaskFor: registerSecond)]
					ifFalse: [0]]! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 3/21/2017 18:32'!
nativeRegisterOrNone
	^(type = SSNativeRegister or: [type = SSRegisterPair])
		 ifTrue: [register]
		 ifFalse: [NoReg]! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 3/6/2017 02:16'!
nativeRegisterSecondOrNone
	^type = SSRegisterPair
		 ifTrue: [registerSecond]
		 ifFalse: [NoReg]! !

!CogSimStackNativeEntry methodsFor: 'compile abstract instructions' stamp: 'rsf 1/15/2017 20:52'!
nativeStackPopToReg: reg
	self assert: spilled.

		type caseOf: {
			[SSSpillNative]	-> [cogit MoveMw: offset negated r: FPReg R: reg].
			[SSSpillInt64]	-> [self assert: BytesPerWord = 8.
				cogit MoveMw: offset negated r: FPReg R: reg].
			[SSSpillFloat32]	-> [cogit MoveM32: offset negated r: FPReg Rs: reg.].
			[SSSpillFloat64]	-> [cogit MoveM64: offset negated r: FPReg Rd: reg.].
	}! !

!CogSimStackNativeEntry methodsFor: 'compile abstract instructions' stamp: 'rsf 12/11/2014 04:25'!
nativeStackPopToReg: reg secondReg: secondReg
	self assert: spilled.

		type caseOf: {
			[SSSpillInt64]	-> [cogit MoveMw: offset negated + 4 r: FPReg R: reg.
							     cogit MoveMw: offset negated r: FPReg R: secondReg].
	}! !

!CogSimStackNativeEntry methodsFor: 'compile abstract instructions' stamp: 'RonieSalgado 5/2/2016 13:40'!
nativeStoreToReg: reg
	"FIXME: Duplicated code with popToReg:"
	spilled ifTrue: [cogit loadNativeFramePointerInto: TempReg.].

	type caseOf: {
		[SSSpillNative]	-> [cogit MoveMw: offset negated - 1 r: TempReg R: reg].
		[SSSpillFloat32]	-> [cogit MoveM32: offset negated - 1 r: TempReg Rs: reg.].
		[SSSpillFloat64]	-> [cogit MoveM64: offset negated - 1 r: TempReg Rd: reg.].
		[SSNativeRegister]	-> [reg ~= register
										ifTrue: [cogit MoveR: register R: reg]
										ifFalse: [cogit Label]].
		[SSRegisterSingleFloat]	-> [reg ~= register
										ifTrue: [cogit MoveRs: register Rs: reg]
										ifFalse: [cogit Label]].
		[SSRegisterDoubleFloat]	-> [reg ~= register
										ifTrue: [cogit MoveRd: register Rd: reg]
										ifFalse: [cogit Label]].

		[SSConstantInt32]	-> [cogit MoveCw: constantInt32 R: reg].
		[SSConstantNativePointer]	-> [cogit MoveCw: constantNativePointer R: reg].
		[SSConstantFloat32] -> [ cogit MoveCf32: constantFloat32 Rs: reg. ].
		[SSConstantFloat64] -> [ cogit MoveCf64: constantFloat64 Rd: reg. ].
	}.! !

!CogSimStackNativeEntry methodsFor: 'compile abstract instructions' stamp: 'rsf 12/10/2014 02:18'!
nativeStoreToReg: reg secondReg: secondReg
	self assert: BytesPerWord = 4.
	"FIXME: Duplicated code with popToReg:secondReg:"
	type caseOf: {
		[SSBaseOffset]	-> [
				cogit MoveMw: offset r: register R: reg.
				cogit MoveMw: offset + 4 r: register R: secondReg.
			].
		[SSSpillInt64]	-> [
				cogit MoveMw: offset r: register R: reg.
				cogit MoveMw: offset + 4 r: registerSecond R: secondReg
			].
		[SSRegisterPair]	-> [
				cogit Label.
				reg ~= register ifTrue: [
					registerSecond = reg ifTrue: [
						cogit MoveR: registerSecond R: TempReg.
					].
					cogit MoveR: register R: reg
				].

				registerSecond ~= secondReg ifTrue: [
					registerSecond = reg ifTrue: [
						cogit MoveR: TempReg R: secondReg.
					] ifFalse: [
						cogit MoveR: registerSecond R: secondReg.
					]
				].
			].
		[SSConstantInt64]	-> [
			self MoveCq: (constantInt64 bitAnd: 16rFFFFFFFF) R: reg.
			self MoveCq: ((constantInt64 >> 32) bitAnd: 16rFFFFFFFF) R: secondReg.
		].
	}.
! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/10/2014 02:16'!
offset
	<doNotGenerate>
	"Answer the value of offset"
	^offset! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/10/2014 02:16'!
offset: anObject
	<doNotGenerate>
	"Set the value of offset"

	^offset := anObject! !

!CogSimStackNativeEntry methodsFor: 'printing' stamp: 'rsf 12/10/2014 01:52'!
printStateOn: aStream
	<doNotGenerate> "Smalltalk-side only"
	type isInteger ifFalse: [^self].
	aStream nextPut: $(.
	type caseOf: {
		[SSNativeRegister]	-> [aStream
								nextPutAll: 'reg ';
								nextPutAll: (CogRTLOpcodes nameForRegister: register)].
		[SSRegisterPair]	-> [aStream
								nextPutAll: 'reg ';
								nextPutAll: (CogRTLOpcodes nameForRegister: register);
								nextPutAll: '+';
								nextPutAll: (CogRTLOpcodes nameForRegister: registerSecond).].
		[SSRegisterSingleFloat]	-> [aStream
								nextPutAll: 'regFloat32 ';
								nextPutAll: (CogRTLOpcodes nameForRegister: register)].
		[SSRegisterDoubleFloat]	-> [aStream
								nextPutAll: 'regFloat64 ';
								nextPutAll: (CogRTLOpcodes nameForRegister: register)].

		[SSSpillNative]		-> [aStream
								nextPutAll: 'spillNative @ '.
							offset negative ifFalse: [aStream nextPut: $+].
							aStream print: offset].
		[SSSpillInt64]		-> [aStream
								nextPutAll: 'spillInt64 @ '.
							offset negative ifFalse: [aStream nextPut: $+].
							aStream print: offset].
		[SSSpillFloat32]		-> [aStream
								nextPutAll: 'spillFloat32 @ '.
							offset negative ifFalse: [aStream nextPut: $+].
							aStream print: offset].
		[SSSpillFloat64]		-> [aStream
								nextPutAll: 'spillFloat64 @ '.
							offset negative ifFalse: [aStream nextPut: $+].
							aStream print: offset].

		[SSConstantInt32]		-> [aStream
								nextPutAll: 'constInt32 ';
								print: constantInt32].
		[SSConstantInt64]		-> [aStream
								nextPutAll: 'constInt64 ';
								print: constantInt64].
		[SSConstantFloat32]		-> [aStream
								nextPutAll: 'constFloat32 ';
								print: constantFloat32].
		[SSConstantFloat64]		-> [aStream
								nextPutAll: 'constFloat64 ';
								print: constantFloat64].
		[SSConstantNativePointer]	-> [aStream
								nextPutAll: 'constNativePointer ';
								print: constantNativePointer]
 }.
	spilled ifTrue:
		[aStream nextPutAll: ' (spilled)'].
	bcptr ifNotNil:
		[aStream space; nextPut: ${; print: bcptr; nextPut: $}].
	aStream nextPut: $)! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/10/2014 02:21'!
register
	<doNotGenerate>
	"Answer the value of register"
	self assert: ((([type = SSNativeRegister] or: [type = SSRegisterPair]) or: [type = SSRegisterSingleFloat]) or: [type = SSRegisterDoubleFloat]).
	^register! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/10/2014 02:21'!
register: anObject
	<doNotGenerate>
	"Set the value of register"

	^register := anObject! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
registerSecond
	"Answer the value of register"
	self assert: type = SSRegisterPair.
	^ registerSecond! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
registerSecond: anObject
	"Set the value of register"

	^registerSecond := anObject! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 1/17/2017 03:46'!
registerSecondOrNil
	^ type = SSRegisterPair ifTrue: [registerSecond]! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
spilled
	"Answer the value of spilled"

	^ spilled! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
spilled: anObject
	"Set the value of spilled"

	^spilled := anObject! !

!CogSimStackNativeEntry methodsFor: 'compile abstract instructions' stamp: 'rsf 12/10/2014 01:17'!
spillingNeedsScratchRegister
	spilled ifFalse: [
		type caseOf: {
			[SSConstantInt32]	-> [ ^ true ].
			[SSConstantInt64]	-> [ ^ true ].
			[SSConstantFloat32]	-> [ ^ true ].
			[SSConstantFloat64]	-> [ ^ true ].
			[SSConstantNativePointer]	-> [ ^ true ].
		}
		otherwise: [^ false]
	].

	^ false.
! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'eem 2/11/2020 10:20'!
stackSpillSize
	<returnTypeC: #sqInt>
	type caseOf: {
		[SSConstantInt64]			-> [ ^ 8 ].
		[SSConstantFloat64]		-> [ ^ 8 ].
		[SSRegisterDoubleFloat]	-> [ ^ 8 ].
		[SSRegisterPair]			-> [ ^ 8 ].
		[SSSpillFloat64]				-> [ ^ 8 ].
		[SSSpillInt64]				-> [ ^ 8 ].
	} otherwise: [^ cogit objectMemory bytesPerOop ].
	^ 0! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
type
	"Answer the value of type"

	^ type! !

!CogSimStackNativeEntry methodsFor: 'accessing' stamp: 'rsf 12/9/2014 18:11'!
type: anObject
	"Set the value of type"

	^type := anObject! !

VMStructType subclass: #CogStackPage
	instanceVariableNames: 'stackLimit headSP headFP baseFP baseAddress realStackLimit lastAddress trace padToWord nextPage prevPage'
	classVariableNames: ''
	poolDictionaries: 'VMBasicConstants VMBytecodeConstants'
	category: 'VMMaker-Interpreter'!

!CogStackPage commentStamp: 'eem 8/14/2015 12:06' prior: 0!
I am a class that helps organize the StackInterpreter's collection of stack pages.  I represent the control block for a single stack page in the collection of stack pages represented by an InterpreterStackPages or CoInterpreterStackPages instance.!

!CogStackPage class methodsFor: 'translation' stamp: 'eem 3/6/2009 09:21'!
alignedByteSize
	^self surrogateClass alignedByteSize! !

!CogStackPage class methodsFor: 'translation' stamp: 'eem 9/2/2011 14:03'!
alignedByteSizeOf: anObject forClient: aVMClass
	^self surrogateClass alignedByteSize! !

!CogStackPage class methodsFor: 'code generation' stamp: 'eem 10/9/2015 10:50'!
getter: getter bitPosition: bitPosition bitWidth: bitWidth type: typeOrNil
	^String streamContents:
		[:s| | startByte endByte accessor |
		startByte := bitPosition // 8.
		endByte := bitPosition + bitWidth - 1 // 8.
		self assert: bitPosition \\ 8 = 0.
		self assert: startByte \\ (bitWidth // 8) = 0.
		accessor := #('byte' 'short' 'long' 'long')
							at: endByte - startByte + 1
							ifAbsent: ['long64'].
		s	nextPutAll: getter; crtab: 1; nextPut: $^.
		(typeOrNil notNil and: [typeOrNil last = $*]) ifTrue:
			[accessor := 'unsigned', (accessor copy
										at: 1 put: accessor first asUppercase;
										yourself)].
		(typeOrNil notNil and: ['*StackPage*' match: typeOrNil]) ifTrue:
			[s nextPutAll: 'stackPages surrogateAtAddress: ('].
		s nextPutAll: 'memory ';
		   nextPutAll: accessor;
		   nextPutAll: 'At: address + '; print: startByte + 1.
		(typeOrNil notNil and: ['*StackPage*' match: typeOrNil]) ifTrue:
			[s nextPut: $)]]

	"| bitPosition |
	bitPosition := 0.
	(self fieldAccessorsForBytesPerWord: 4) collect:
		[:spec|
		bitPosition := bitPosition + spec second.
		self getter: spec first
			 bitPosition: bitPosition - spec second
			 bitWidth: spec second
			 type: (spec at: 3 ifAbsent: [])]"! !

!CogStackPage class methodsFor: 'translation' stamp: 'eem 9/29/2015 09:08'!
initialize
	"self initialize"
	(Smalltalk classNamed: #CogStackPageSurrogate32) ifNotNil:
		[:csps32|
		self checkGenerateSurrogate: csps32 bytesPerWord: 4].
	(Smalltalk classNamed: #CogStackPageSurrogate64) ifNotNil:
		[:csps64|
		self checkGenerateSurrogate: csps64 bytesPerWord: 8]! !

!CogStackPage class methodsFor: 'translation' stamp: 'eem 10/9/2015 10:32'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a StackPage struct."

	self allInstVarNames do:
		[:ivn|
		(ivn = 'padToWord' and: [BytesPerWord = 4]) ifFalse:
			[aBinaryBlock
				value: ivn
				value: ((ivn = 'trace' or: [ivn = 'padToWord'])
						ifTrue: [#int]
						ifFalse:
							[(ivn endsWith: 'Page')
								ifTrue: ['struct _StackPage *']
								ifFalse: [#'char *']])]]! !

!CogStackPage class methodsFor: 'code generation' stamp: 'eem 10/9/2015 10:52'!
setter: getter bitPosition: bitPosition bitWidth: bitWidth type: typeOrNil
	^String streamContents:
		[:s| | startByte endByte accessor |
		startByte := bitPosition // 8.
		endByte := bitPosition + bitWidth - 1 // 8.
		self assert: bitPosition \\ 8 = 0.
		self assert: startByte \\ (bitWidth // 8) = 0.
		accessor := #('byte' 'short' 'long' 'long')
							at: endByte - startByte + 1
							ifAbsent: ['long64'].
		s	nextPutAll: getter; nextPutAll: ': aValue'; crtab: 1;
			nextPutAll: 'self assert: (address + '; print: startByte;
			nextPutAll: ' >= zoneBase and: [address + '; print: endByte;
			nextPutAll: ' < zoneLimit]).'; crtab: 1.
		(typeOrNil notNil and: [typeOrNil last = $*]) ifTrue:
			[accessor := 'unsigned', (accessor copy
										at: 1 put: accessor first asUppercase;
										yourself)].
		(typeOrNil notNil and: ['*StackPage*' match: typeOrNil]) ifFalse:
			[s nextPut: $^].
		s nextPutAll: 'memory ';
		   nextPutAll: accessor; nextPutAll: 'At: address + '; print: startByte + 1;
		   nextPutAll: ' put: aValue'.
		(typeOrNil notNil and: ['*StackPage*' match: typeOrNil]) ifTrue:
			[s nextPutAll: ' asInteger.'; crtab: 1; nextPutAll: '^aValue']]

	"| bitPosition |
	bitPosition := 0.
	(self fieldAccessorsForBytesPerWord: 4) collect:
		[:spec|
		bitPosition := bitPosition + spec second.
		self setter: spec first
			 bitPosition: bitPosition - spec second
			 bitWidth: spec second
			 type: (spec at: 3 ifAbsent: [])]"! !

!CogStackPage class methodsFor: 'translation' stamp: 'eem 2/10/2009 14:08'!
structTypeName
	^'StackPage' "Drop initial Cog or Interpreter"! !

!CogStackPage class methodsFor: 'simulation only' stamp: 'eem 1/15/2011 08:31'!
surrogateClass
	^BytesPerWord = 4
		ifTrue: [CogStackPageSurrogate32]
		ifFalse: [CogStackPageSurrogate64]! !

!CogStackPage methodsFor: 'simulation only' stamp: 'eem 2/12/2009 14:13'!
address
	<doNotGenerate>
	^baseAddress! !

!CogStackPage methodsFor: 'testing' stamp: 'eem 6/2/2009 11:16'!
addressIsInPage: address
	<var: #address type: #'char *'>
	"For assert-checking"
	^lastAddress < address and: [address < baseAddress]! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 7/3/2008 17:29'!
baseAddress
	^baseAddress! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 2/28/2009 23:38'!
baseAddress: anAddress
	^baseAddress := anAddress! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 7/11/2008 12:26'!
baseFP
	"Answer the value of baseFP"

	^ baseFP! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 12/21/2008 12:55'!
baseFP: pointer "<Integer>"
	"Set the value of baseFP"
	self assert: (pointer = 0 or: [pointer < baseAddress and: [realStackLimit < pointer]]).
	^baseFP := pointer! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 6/20/2008 08:53'!
headFP
	"Answer the value of headFP"

	^headFP! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 11/13/2014 13:13'!
headFP: pointer "<Integer>"
	"Set the value of headFP"
	^headFP := pointer! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 7/11/2008 12:26'!
headSP
	"Answer the value of headSP"

	^headSP! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 11/13/2014 13:13'!
headSP: pointer "<Integer>"
	"Set the value of headSP"
	^headSP := pointer! !

!CogStackPage methodsFor: 'testing' stamp: 'eem 7/11/2008 12:26'!
isFree
	^baseFP = 0! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 7/11/2008 11:20'!
lastAddress
	^lastAddress! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 2/28/2009 23:39'!
lastAddress: anAddress
	^lastAddress := anAddress! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 6/19/2008 23:16'!
nextPage
	"Answer the value of nextPage"

	^ nextPage! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 12/21/2008 12:55'!
nextPage: anObject
	"Set the value of nextPage"

	^nextPage := anObject! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 6/19/2008 23:16'!
prevPage
	"Answer the value of prevPage"

	^ prevPage! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 12/21/2008 12:55'!
prevPage: anObject
	"Set the value of prevPage"

	^prevPage := anObject! !

!CogStackPage methodsFor: 'printing' stamp: 'eem 11/7/2013 11:58'!
printOn: aStream
	<doNotGenerate>
	super printOn: aStream.
	aStream nextPut: $@; print: baseAddress; space.
	self isFree
		ifTrue: [aStream nextPutAll: 'free']
		ifFalse: [aStream print: baseFP; nextPutAll: '<->'; print: headFP; space; nextPutAll: 'trace '; print: trace]! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 6/19/2008 23:16'!
realStackLimit
	"Answer the value of realStackLimit"

	^ realStackLimit! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 12/21/2008 12:55'!
realStackLimit: anObject
	"Set the value of realStackLimit"

	^realStackLimit := anObject! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 6/19/2008 23:16'!
stackLimit
	"Answer the value of stackLimit"

	^ stackLimit! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 12/21/2008 12:55'!
stackLimit: anObject
	"Set the value of stackLimit"

	^stackLimit := anObject! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 12/20/2008 07:31'!
trace
	"Answer the page's trace state.
	 0 = untraced.  1 = should be traced. 2 = has been traced.
	-1 = invalid (for assertions)"
	^trace! !

!CogStackPage methodsFor: 'accessing' stamp: 'eem 12/21/2008 12:54'!
trace: anInteger
	"Set the page's trace state.
	 0 = untraced.  1 = should be traced. 2 = has been traced.
	-1 = invalid (for assertions)"
	^trace := anInteger! !

VMStructType subclass: #CogVMThread
	instanceVariableNames: 'index state priority osSemaphore osThread newMethodOrNull argumentCount primitiveFunctionPointer inMachineCode cStackPointer cFramePointer awolProcIndex awolProcLength awolProcesses'
	classVariableNames: ''
	poolDictionaries: 'VMThreadingConstants'
	category: 'VMMaker-Multithreading'!

!CogVMThread commentStamp: '<historical>' prior: 0!
Instances of this class represent control blocks for native threads that cooperatively schedule the VM.  See the class comment of CogThreadManager for full documentation.

N.B. awolProcesses must be the last inst var.!

!CogVMThread class methodsFor: 'simulation' stamp: 'eem 7/15/2009 13:19'!
alignedByteSize
	"this is a simulation only hack"
	^self instSize * 4! !

!CogVMThread class methodsFor: 'translation' stamp: 'eem 11/30/2020 09:57'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a CogVMThread struct."

	self allInstVarNames do:
		[:ivn|
		aBinaryBlock
			value: ivn
			value: (ivn caseOf: {
						['awolProcesses']			-> [{#sqInt. '[', CogThreadManager awolProcessesIncrement printString, ']'}].
						['cStackPointer']			-> [#usqIntptr_t].
						['cFramePointer']			-> [#usqIntptr_t].
						['primitiveFunctionPointer']	-> [#('void (*' ')()')].
						['osSemaphore']			-> ['sqOSSemaphore'].
						['osThread']					-> ['sqOSThread'] }
					otherwise:
						[#sqInt])]! !

!CogVMThread class methodsFor: 'documentation' stamp: 'eem 7/8/2009 14:11'!
notes
	"win32: When a thread terminates, the thread object attains a signaled state, satisfying any threads that were waiting on the object.

	if (WaitForSingleObject( hThread, 0) == WAIT_OBJECT_0) {
		// the thread handle is signaled - the thread has terminated
	}
	else {
		// the thread handle is not signaled - the thread is still alive
	}

	Alternatively use BOOL WINAPI GetExitCodeThread(HANDLE hThread, LPDWORD lpExitCode)
	which returns STILL_ACTIVE if still active.

	DWORD retval;

	if (GetExitCodeThread( hThread, &retval) == STILL_ACTIVE) {
		// the thread is still alive
	}
	else {
		// the thread has terminated
	}

	pthreads:
		The pthread_kill() function sends a signal, specified by sig, to a
		thread, specified by thread.  If sig is 0, error checking is performed,
		but no signal is actually sent.

		pthread_kill() will fail if:
			[ESRCH]		thread is an invalid thread ID.

	if (pthread_kill(thread, 0) == ESRCH) {
		// the thread can't be found and has teminated
	}
	else {
		// the thread can be found and has either not terminated or the handle has been reused
	}"! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 11/25/2009 13:46'!
argumentCount
	"Answer the value of argumentCount"

	^ argumentCount! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 11/25/2009 13:46'!
argumentCount: anObject
	"Set the value of argumentCount"

	^argumentCount := anObject! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 8/14/2009 12:37'!
awolProcIndex
	"Answer the value of awolProcIndex"

	^ awolProcIndex! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 8/14/2009 12:37'!
awolProcIndex: anObject
	"Set the value of awolProcIndex"

	^awolProcIndex := anObject! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 8/14/2009 12:37'!
awolProcLength
	"Answer the value of awolProcLength"

	^ awolProcLength! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 8/14/2009 12:37'!
awolProcLength: anObject
	"Set the value of awolProcLength"

	^awolProcLength := anObject! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 8/14/2009 12:37'!
awolProcesses
	"Answer the value of awolProcesses"

	^ awolProcesses! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 9/16/2009 14:37'!
cFramePointer
	"Answer the value of cFramePointer"

	^ cFramePointer! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 9/16/2009 14:37'!
cFramePointer: anObject
	"Set the value of cFramePointer"

	^cFramePointer := anObject! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 8/12/2009 11:47'!
cStackPointer
	"Answer the value of cStackPointer"

	^ cStackPointer! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 8/12/2009 11:49'!
cStackPointer: anObject
	"Set the value of cStackPointer"

	^cStackPointer := anObject! !

!CogVMThread methodsFor: 'simulation only' stamp: 'eem 8/14/2009 15:05'!
growAWOLProcesses
	<doNotGenerate>
	awolProcesses setObject: awolProcesses object, (Array new: CogThreadManager awolProcessesIncrement)! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 11/30/2009 13:55'!
inMachineCode
	"Answer the value of inMachineCode"

	^ inMachineCode! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 11/30/2009 13:55'!
inMachineCode: anObject
	"Set the value of inMachineCode"

	^inMachineCode := anObject! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 7/22/2009 20:31'!
index
	"Answer the value of index"

	^ index! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 8/5/2009 10:40'!
index: anObject
	"Set the value of index"

	^index := anObject! !

!CogVMThread methodsFor: 'initialize-release' stamp: 'eem 11/19/2020 19:16'!
initialize
	awolProcLength := CogThreadManager awolProcessesIncrement.
	awolProcesses := CArrayAccessor on: (Array new: awolProcLength).
	awolProcIndex := 0! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 8/5/2009 10:39'!
newMethodOrNull
	"Answer the value of newMethodOrNull"

	^ newMethodOrNull! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 8/5/2009 10:40'!
newMethodOrNull: anObject
	"Set the value of newMethodOrNull"

	^newMethodOrNull := anObject! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 7/13/2009 10:42'!
osSemaphore
	"Answer the value of osSemaphore"

	^ osSemaphore! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 8/5/2009 10:40'!
osSemaphore: anObject
	"Set the value of osSemaphore"

	^osSemaphore := anObject! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 7/13/2009 10:42'!
osThread
	"Answer the value of osThread"

	^ osThread! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 11/10/2020 10:46'!
osThread: anObject
	"Set the value of osThread"
	self assert: (osThread isNil or: [osThread isKindOf: Process]).
	^osThread := anObject! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 10/30/2020 12:16'!
primitiveFunctionPointer
	"Answer the value of primitiveFunctionPointer"

	^primitiveFunctionPointer! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 10/30/2020 12:16'!
primitiveFunctionPointer: anObject
	"Set the value of primitiveFunctionPointer"

	^primitiveFunctionPointer := anObject! !

!CogVMThread methodsFor: 'printing' stamp: 'eem 10/30/2020 09:45'!
printOn: aStream
	super printOn: aStream.
	aStream
		nextPutAll: ' index: '; print: index;
		nextPutAll: ' state: '; nextPutAll: (VMThreadingConstants keys
											detect: [:k| k first == $C and: [(VMThreadingConstants classPool at: k) = state]]
											ifNone: [state printString])! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 8/13/2009 16:26'!
priority
	"Answer the value of priority"

	^ priority! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 8/13/2009 16:26'!
priority: anObject
	"Set the value of priority"

	^priority := anObject! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 7/13/2009 10:42'!
state
	"Answer the value of state"

	^ state! !

!CogVMThread methodsFor: 'accessing' stamp: 'eem 8/5/2009 10:40'!
state: anObject
	"Set the value of state"

	^state := anObject! !

VMStructType subclass: #NSSendCache
	instanceVariableNames: 'classTag enclosingObject target selector numArgs depth'
	classVariableNames: ''
	poolDictionaries: 'CogMethodConstants VMBasicConstants'
	category: 'VMMaker-JIT'!

!NSSendCache commentStamp: 'eem 1/27/2017 10:04' prior: 0!
A NSSendCache is an inline cache for Newspeak absent receiver sends, designed and implemented by Ryan Macnak.  These caches are used with four varieties of absent receiver sends:
	absent self send
	absent implciit receiver send
	absent dynamic super send
	absent outer send.

See Cogit>>#linkNSSendCache:classTag:enclosingObject:target:caller: and senders.  These caches have
  - a guarded class id (not an oop)
  - an absent receiver (or 0 if it is the method receiver) (an oop or 0)
  - a target method in the code zone (not an oop)
  - depth (not an oop) the nesting depth of the receiver fror an outer send, or < 0 if not nested
  - numArgs (not an oop)
  - the selector (an oop)

The selector, numArgs and depth are read at link time.  The class id is checked at send time and if it matches, the absentReceiver, enclosingObject numArgs and and target method are read, self being substituted for the absentReceiver if absentReceiver is 0.  numArgs is read for sends with arity > 3 (it is implicit in the trampoline for arities 0 to 3) and used to adjust the stack and insert the relevant receiver.  See genNSSendTrampolineFor:numArgs:enclosingObjectCheck:called:.

Note that the caches themselves are also pinned objects on the heap (a non-object punned word array whose elements are not traced) rather than in the code zone, so GC also needs to trace and update references from Cogged methods to these cache objects.  The absent receiver can be updated or cleared on GC. The selector must be updated on GC. The target method can be updated or cleared on code compaction.  The class id, absent receiver, and target method are weak references. cache->selector and method->cache are strong references..

Instance Variables
	classTag:			<Integer>
	depth:				<Integer>
	enclosingObject:	<Object or 0>
	numArgs:			<Integer>
	selector:			<Object>
	target:				<CogMethod>

classTag
	- classTag of the method receiver for which this is valid (the classTag of self in the method in which the send occurs, not the class of the message receiver)

depth
	- nesting depth for outer sends

enclosingObject
	- the oop of the enclosingObject or 0 if none

numArgs
	- send numArgs

selector
	- the oop of the message's selector

target
	- the target CogMethod or 0 if none
!

!NSSendCache class methodsFor: 'class initialization' stamp: 'rmacnak 3/27/2015 19:55'!
initialize
	"self initialize"
	(Smalltalk classNamed: #NSSendCacheSurrogate32) ifNotNil:
		[:scs32|
		self checkGenerateSurrogate: scs32 bytesPerWord: 4].
	(Smalltalk classNamed: #NSSendCacheSurrogate64) ifNotNil:
		[:scs64|
		self checkGenerateSurrogate: scs64 bytesPerWord: 8]! !

!NSSendCache class methodsFor: 'translation' stamp: 'rmacnak 3/27/2015 19:50'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock

	self allInstVarNames do:
		[:ivn| aBinaryBlock
				value: ivn
				value: #'sqInt']! !

!NSSendCache methodsFor: 'accessing' stamp: 'rmacnak 3/27/2015 19:17'!
classTag

	^ classTag! !

!NSSendCache methodsFor: 'accessing' stamp: 'rmacnak 3/27/2015 19:17'!
classTag: anObject

	^classTag := anObject! !

!NSSendCache methodsFor: 'accessing' stamp: 'rmacnak 4/4/2015 15:27'!
depth

	^ depth! !

!NSSendCache methodsFor: 'accessing' stamp: 'rmacnak 4/4/2015 15:27'!
depth: anObject

	^depth := anObject! !

!NSSendCache methodsFor: 'accessing' stamp: 'rmacnak 3/27/2015 19:17'!
enclosingObject

	^ enclosingObject! !

!NSSendCache methodsFor: 'accessing' stamp: 'rmacnak 3/27/2015 19:17'!
enclosingObject: anObject

	^enclosingObject := anObject! !

!NSSendCache methodsFor: 'accessing' stamp: 'rmacnak 3/27/2015 19:17'!
numArgs

	^ numArgs! !

!NSSendCache methodsFor: 'accessing' stamp: 'rmacnak 3/27/2015 19:17'!
numArgs: anObject

	^numArgs := anObject! !

!NSSendCache methodsFor: 'accessing' stamp: 'rmacnak 3/27/2015 19:17'!
selector

	^ selector! !

!NSSendCache methodsFor: 'accessing' stamp: 'rmacnak 3/27/2015 19:17'!
selector: anObject

	^selector := anObject! !

!NSSendCache methodsFor: 'accessing' stamp: 'rmacnak 3/27/2015 19:17'!
target

	^ target! !

!NSSendCache methodsFor: 'accessing' stamp: 'rmacnak 3/27/2015 19:17'!
target: anObject

	^target := anObject! !

VMStructType subclass: #SpurNewSpaceSpace
	instanceVariableNames: 'start limit'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManager'!

SpurNewSpaceSpace subclass: #SpurContiguousObjStack
	instanceVariableNames: 'top'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManager'!

!SpurContiguousObjStack methodsFor: 'printing' stamp: 'eem 10/18/2013 14:07'!
printOn: aStream
	<doNotGenerate>
	super printOn: aStream.
	top ifNotNil:
		[aStream nextPutAll: ' top: '; nextPutAll: top hex]! !

!SpurContiguousObjStack methodsFor: 'accessing' stamp: 'eem 10/17/2013 11:01'!
top
	"Answer the value of top"

	^ top! !

!SpurContiguousObjStack methodsFor: 'accessing' stamp: 'eem 10/17/2013 11:01'!
top: anObject
	"Set the value of top"

	^top := anObject! !

!SpurNewSpaceSpace class methodsFor: 'translation' stamp: 'eem 10/23/2013 10:03'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	self allInstVarNames do:
		[:ivn|
		(SpurMemoryManager isNonArgumentImplicitReceiverVariableName: ivn) ifFalse:
			[aBinaryBlock value: ivn value: #usqInt]]! !

!SpurNewSpaceSpace methodsFor: 'accessing' stamp: 'eem 9/4/2013 12:40'!
limit
	^limit! !

!SpurNewSpaceSpace methodsFor: 'accessing' stamp: 'eem 9/30/2013 20:17'!
limit: n
	limit := n! !

!SpurNewSpaceSpace methodsFor: 'printing' stamp: 'eem 10/12/2013 11:23'!
printOn: aStream
	<doNotGenerate>
	super printOn: aStream.
	(start notNil and: [limit notNil]) ifTrue:
		[aStream nextPutAll: ' start: '; nextPutAll: start hex; nextPutAll: ' limit: '; nextPutAll: limit hex]! !

!SpurNewSpaceSpace methodsFor: 'accessing' stamp: 'eem 9/4/2013 12:40'!
start
	^start! !

!SpurNewSpaceSpace methodsFor: 'accessing' stamp: 'eem 9/30/2013 20:17'!
start: n
	start := n! !

VMStructType subclass: #SpurScavengeLogRecord
	instanceVariableNames: 'sEdenBytes sPastBytes sRememberedSetSize sRememberedSetRedZone sRememberedSetLimit sStatTenures tTenureCriterion tTenureThreshold tRefCountToShrinkRT eSurvivorBytes eRememberedSetSize eStatTenures'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManager'!

!SpurScavengeLogRecord class methodsFor: 'translation' stamp: 'eem 11/13/2017 14:24'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	self allInstVarNames do:
		[:ivn|
		(SpurMemoryManager isNonArgumentImplicitReceiverVariableName: ivn) ifFalse:
			[aBinaryBlock value: ivn value: #usqInt]]! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
eRememberedSetSize
	
	^ eRememberedSetSize
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
eRememberedSetSize: anObject
	
	^ eRememberedSetSize := anObject.
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
eStatTenures
	
	^ eStatTenures
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
eStatTenures: anObject
	
	^ eStatTenures := anObject.
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
eSurvivorBytes
	
	^ eSurvivorBytes
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
eSurvivorBytes: anObject
	
	^ eSurvivorBytes := anObject.
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
sEdenBytes
	
	^ sEdenBytes
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
sEdenBytes: anObject
	
	^ sEdenBytes := anObject.
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
sPastBytes
	
	^ sPastBytes
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
sPastBytes: anObject
	
	^ sPastBytes := anObject.
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
sRememberedSetLimit
	
	^ sRememberedSetLimit
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
sRememberedSetLimit: anObject
	
	^ sRememberedSetLimit := anObject.
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
sRememberedSetRedZone
	
	^ sRememberedSetRedZone
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
sRememberedSetRedZone: anObject
	
	^ sRememberedSetRedZone := anObject.
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
sRememberedSetSize
	
	^ sRememberedSetSize
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
sRememberedSetSize: anObject
	
	^ sRememberedSetSize := anObject.
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 16:08'!
sStatTenures
	
	^ sStatTenures
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 16:08'!
sStatTenures: anObject
	
	^ sStatTenures := anObject.
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
tRefCountToShrinkRT
	
	^ tRefCountToShrinkRT
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
tRefCountToShrinkRT: anObject
	
	^ tRefCountToShrinkRT := anObject.
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
tTenureCriterion
	
	^ tTenureCriterion
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
tTenureCriterion: anObject
	
	^ tTenureCriterion := anObject.
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
tTenureThreshold
	
	^ tTenureThreshold
! !

!SpurScavengeLogRecord methodsFor: 'accessing' stamp: 'eem 11/13/2017 14:20'!
tTenureThreshold: anObject
	
	^ tTenureThreshold := anObject.
! !

VMStructType subclass: #SpurSegmentInfo
	instanceVariableNames: 'segStart segSize swizzle containsPinned savedSegSize lastFreeObject'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-SpurMemoryManager'!

!SpurSegmentInfo class methodsFor: 'translation' stamp: 'eem 10/5/2013 15:05'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"Enumerate aBinaryBlock with the names and C type strings for the inst vars to include in the typedef for the reciever."
	"self typedef"
	self instVarNames do:
		[:ivn|
		aBinaryBlock
			value: ivn
			value: (ivn = 'swizzle'
							ifTrue: [#sqInt]
							ifFalse: [#usqInt])]! !

!SpurSegmentInfo methodsFor: 'accessing' stamp: 'eem 10/21/2013 09:17'!
containsPinned
	"Answer the value of containsPinned"

	^ containsPinned! !

!SpurSegmentInfo methodsFor: 'accessing' stamp: 'eem 10/21/2013 09:17'!
containsPinned: anObject
	"Set the value of containsPinned"

	^containsPinned := anObject! !

!SpurSegmentInfo methodsFor: 'initialization' stamp: 'eem 10/21/2013 09:16'!
initialize
	segSize := segStart := swizzle := 0.
	containsPinned := false! !

!SpurSegmentInfo methodsFor: 'accessing' stamp: 'eem 2/12/2014 17:16'!
lastFreeObject
	"Answer the value of lastFreeObject"

	^ lastFreeObject! !

!SpurSegmentInfo methodsFor: 'accessing' stamp: 'eem 2/12/2014 17:16'!
lastFreeObject: anObject
	"Set the value of lastFreeObject"

	^lastFreeObject := anObject! !

!SpurSegmentInfo methodsFor: 'printing' stamp: 'eem 11/16/2013 08:03'!
printOn: aStream
	<doNotGenerate>
	super printOn: aStream.
	(self class instVarNames copyReplaceAll: #('segSize') with: #('segSize' 'segLimit')) do:
		[:name| | iv |
		iv := self perform: name asSymbol.
		aStream space; nextPutAll: name; space; print: iv.
		(iv isInteger and: [iv ~= 0]) ifTrue:
			[aStream nextPut: $/.  iv storeOn: aStream base: 16]]! !

!SpurSegmentInfo methodsFor: 'accessing' stamp: 'eem 2/12/2014 17:11'!
savedSegSize
	"Answer the value of savedSegSize"

	^ savedSegSize! !

!SpurSegmentInfo methodsFor: 'accessing' stamp: 'eem 2/12/2014 17:11'!
savedSegSize: anObject
	"Set the value of savedSegSize"

	^savedSegSize := anObject! !

!SpurSegmentInfo methodsFor: 'accessing' stamp: 'eem 1/25/2017 09:52'!
segLimit
	<returnTypeC: #usqInt>
	^segSize + segStart! !

!SpurSegmentInfo methodsFor: 'accessing' stamp: 'eem 10/6/2013 10:29'!
segSize
	"Answer the value of segSize"

	^ segSize! !

!SpurSegmentInfo methodsFor: 'accessing' stamp: 'eem 10/6/2013 10:29'!
segSize: anObject
	"Set the value of segSize"

	^segSize := anObject! !

!SpurSegmentInfo methodsFor: 'accessing' stamp: 'eem 10/12/2013 11:25'!
segStart
	"Answer the value of segStart"

	^ segStart! !

!SpurSegmentInfo methodsFor: 'accessing' stamp: 'eem 10/12/2013 11:25'!
segStart: anObject
	"Set the value of segStart"

	^segStart := anObject! !

!SpurSegmentInfo methodsFor: 'accessing' stamp: 'eem 10/5/2013 15:00'!
swizzle
	"Answer the value of swizzle"

	^ swizzle! !

!SpurSegmentInfo methodsFor: 'accessing' stamp: 'eem 10/5/2013 15:00'!
swizzle: anObject
	"Set the value of swizzle"

	^swizzle := anObject! !

VMStructType subclass: #ThreadedFFICalloutState
	instanceVariableNames: 'argVector currentArg limit structReturnSize structReturnType callFlags ffiArgSpec ffiArgSpecSize ffiArgHeader ffiRetHeader ffiRetSpec stringArgIndex stringArgs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedFFICalloutState commentStamp: 'nice 1/30/2020 21:13' prior: 0!
Instances of the receiver hold the per-thread state of a call-out.

long *argVector		pointer to the start of the alloca'ed argument marshalling area
long *currentArg		pointer to the position in argVector to write the current argument
long *limit			the limit of the argument marshalling area (for bounds checking)
structReturnSize		the size of the space allocated for the structure return, if any
structReturnType		an integer encoding how the struct is returned (typically via registers or pointer to allocated memory)
callFlags			the value of the ExternalFunctionFlagsIndex field in the ExternalFunction being called
ffiArgSpec et al		type information for the current argument being marshalled
stringArgIndex		the count of temporary strings used for marshalling Smalltalk strings to character strings.
stringArgs			pointers to the temporary strings used for marshalling Smalltalk strings to character strings.!

!ThreadedFFICalloutState class methodsFor: 'translation' stamp: 'eem 8/28/2020 10:31'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a ThreadedFFICalloutState struct."

	self instVarNames do:
		[:ivn|
		aBinaryBlock
			value: ivn
			value: (ivn caseOf: {
						['argVector']	-> [#'char *'].
						['currentArg']	-> [#'char *'].
						['limit']			-> [#'char *'].
						['ffiArgSpec']	-> [#'unsigned int *']. "ffiArgSpecs are WordArrays"
						['stringArgs']	-> [{#'char *'. '[MaxNumArgs]'}] }
					otherwise:
						[#sqInt])]! !

!ThreadedFFICalloutState class methodsFor: 'translation' stamp: 'eem 10/30/2009 10:14'!
structTypeName
	^'CalloutState' "Drop verbiage and platform specifics"! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
argVector
	"Answer the value of argVector"

	^ argVector! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
argVector: anObject
	"Set the value of argVector"

	^argVector := anObject! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 11/19/2009 10:18'!
callFlags
	"Answer the value of callFlags"

	^ callFlags! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 11/19/2009 10:19'!
callFlags: anObject
	"Set the value of callFlags"

	^callFlags := anObject! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
currentArg
	"Answer the value of currentArg"

	^ currentArg! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
currentArg: anObject
	"Set the value of currentArg"

	^currentArg := anObject! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
ffiArgHeader
	"Answer the value of ffiArgHeader"

	^ ffiArgHeader! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
ffiArgHeader: anObject
	"Set the value of ffiArgHeader"

	^ffiArgHeader := anObject! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
ffiArgSpec
	"Answer the value of ffiArgSpec"

	^ ffiArgSpec! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
ffiArgSpec: anObject
	"Set the value of ffiArgSpec"

	^ffiArgSpec := anObject! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
ffiArgSpecSize
	"Answer the value of ffiArgSpecSize"

	^ ffiArgSpecSize! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
ffiArgSpecSize: anObject
	"Set the value of ffiArgSpecSize"

	^ffiArgSpecSize := anObject! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
ffiRetHeader
	"Answer the value of ffiRetHeader"

	^ ffiRetHeader! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
ffiRetHeader: anObject
	"Set the value of ffiRetHeader"

	^ffiRetHeader := anObject! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'nice 1/25/2020 17:11'!
ffiRetSpec
	"Answer the value of ffiRetSpec"

	^ ffiRetSpec! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'nice 1/25/2020 17:11'!
ffiRetSpec: anObject
	"Set the value of ffiRetSpec"

	^ffiRetSpec := anObject! !

!ThreadedFFICalloutState methodsFor: 'initialize-release' stamp: 'eem 1/26/2020 09:45'!
initialize
	stringArgs := CArrayAccessor on: (Array new: ThreadedFFIPlugin maxNumArgs).
	stringArgIndex := 0.
	structReturnSize := 0! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
limit
	"Answer the value of limit"

	^ limit! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
limit: anObject
	"Set the value of limit"

	^limit := anObject! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 11/9/2009 14:55'!
stringArgIndex
	"Answer the value of stringArgIndex"

	^ stringArgIndex! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 11/9/2009 14:55'!
stringArgIndex: anObject
	"Set the value of stringArgIndex"

	^stringArgIndex := anObject! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 11/9/2009 14:55'!
stringArgs
	"Answer the value of stringArgs"

	^ stringArgs! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/30/2009 14:48'!
structReturnSize
	"Answer the value of structReturnSize"

	^ structReturnSize! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'eem 10/30/2009 14:48'!
structReturnSize: anObject
	"Set the value of structReturnSize"

	^structReturnSize := anObject! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'nice 1/30/2020 21:12'!
structReturnType
	"Answer the value of structReturnType
	It is an OS dependent field encoding for example if struct are to be returned
	- via register
	- via pointer to memory allocated by caller"

	^ structReturnType! !

!ThreadedFFICalloutState methodsFor: 'accessing' stamp: 'nice 1/30/2020 21:12'!
structReturnType: anObject
	"Set the value of structReturnType"

	^structReturnType := anObject! !

ThreadedFFICalloutState subclass: #ThreadedFFICalloutStateForARM32
	instanceVariableNames: 'integerRegisterIndex integerRegisters floatRegisterIndex backfillFloatRegisterIndex floatRegisters'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedFFICalloutStateForARM32 class methodsFor: 'translation' stamp: 'eem 2/19/2016 16:50'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a ThreadedFFICalloutState struct."

	superclass instVarNamesAndTypesForTranslationDo: aBinaryBlock.
	self instVarNames do:
		[:ivn|
		aBinaryBlock
			value: ivn
			value: (ivn caseOf: {
						['integerRegisters']	-> [{#sqInt. '[NumIntRegArgs]'}].
						['floatRegisters']	-> [{#float. '[NumFloatRegArgs]'}] }
					otherwise:
						[#sqInt])]! !

!ThreadedFFICalloutStateForARM32 methodsFor: 'accessing' stamp: 'djm 4/17/2014 08:50'!
backfillFloatRegisterIndex

	^ backfillFloatRegisterIndex! !

!ThreadedFFICalloutStateForARM32 methodsFor: 'accessing' stamp: 'djm 4/17/2014 08:50'!
backfillFloatRegisterIndex: anObject

	^backfillFloatRegisterIndex := anObject! !

!ThreadedFFICalloutStateForARM32 methodsFor: 'accessing' stamp: 'djm 4/12/2014 19:24'!
floatRegisterIndex

	^ floatRegisterIndex! !

!ThreadedFFICalloutStateForARM32 methodsFor: 'accessing' stamp: 'djm 4/12/2014 19:24'!
floatRegisterIndex: anObject

	^floatRegisterIndex := anObject! !

!ThreadedFFICalloutStateForARM32 methodsFor: 'accessing' stamp: 'djm 4/12/2014 19:24'!
floatRegisters

	^ floatRegisters! !

!ThreadedFFICalloutStateForARM32 methodsFor: 'accessing' stamp: 'djm 4/12/2014 19:24'!
floatRegisters: anObject

	^floatRegisters := anObject! !

!ThreadedFFICalloutStateForARM32 methodsFor: 'initialize-release' stamp: 'KenD 1/22/2019 13:37'!
initialize
	super initialize.
	integerRegisterIndex := 0.
	floatRegisterIndex := 0.
	backfillFloatRegisterIndex := 0.
	integerRegisters := CArrayAccessor on: (Array new: ThreadedARM32FFIPlugin numIntRegArgs).
	floatRegisters := CArrayAccessor on: (Array new: ThreadedARM32FFIPlugin numFloatRegArgs)! !

!ThreadedFFICalloutStateForARM32 methodsFor: 'accessing' stamp: 'djm 4/12/2014 19:24'!
integerRegisterIndex

	^ integerRegisterIndex! !

!ThreadedFFICalloutStateForARM32 methodsFor: 'accessing' stamp: 'djm 4/12/2014 19:24'!
integerRegisterIndex: anObject

	^integerRegisterIndex := anObject! !

!ThreadedFFICalloutStateForARM32 methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
integerRegisters
	"Answer the value of integerRegisters"

	^ integerRegisters! !

!ThreadedFFICalloutStateForARM32 methodsFor: 'accessing' stamp: 'eem 10/30/2009 10:08'!
integerRegisters: anObject
	"Set the value of integerRegisters"

	^integerRegisters := anObject! !

ThreadedFFICalloutState subclass: #ThreadedFFICalloutStateForARM64
	instanceVariableNames: 'integerRegisterIndex floatRegisterIndex integerRegisters floatRegisters'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedFFICalloutStateForARM64 commentStamp: 'KenD 2/4/2019 09:57' prior: 0!
A ThreadedFFICalloutStateForARM64 is used to pass FFI parameters for foreign calls.

Instance Variables
	floatRegisterIndex:	Integer
	floatRegisters:		Array of SqInt [64 bit]
	integerRegisterIndex: Integer
	integerRegisters:	Array of Double

			!

!ThreadedFFICalloutStateForARM64 class methodsFor: 'translation' stamp: 'KenD 1/23/2019 08:13'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a ThreadedFFICalloutState struct."

	superclass instVarNamesAndTypesForTranslationDo: aBinaryBlock.
	self instVarNames do:
		[:ivn|
		aBinaryBlock
			value: ivn
			value: (ivn caseOf: {
						['integerRegisters']	-> [{#sqInt.   '[NumIntRegArgs]'   }].
						['floatRegisters']	-> [{#double. '[NumFloatRegArgs]'}] }
					otherwise:
						[#sqInt])]! !

!ThreadedFFICalloutStateForARM64 methodsFor: 'accessing' stamp: 'eem 1/20/2019 19:00'!
floatRegisterIndex

	^ floatRegisterIndex! !

!ThreadedFFICalloutStateForARM64 methodsFor: 'accessing' stamp: 'eem 1/20/2019 19:00'!
floatRegisterIndex: anObject

	^ floatRegisterIndex := anObject.! !

!ThreadedFFICalloutStateForARM64 methodsFor: 'accessing' stamp: 'eem 1/20/2019 19:00'!
floatRegisters

	^ floatRegisters! !

!ThreadedFFICalloutStateForARM64 methodsFor: 'accessing' stamp: 'eem 1/20/2019 19:00'!
floatRegisters: anObject

	^ floatRegisters := anObject.! !

!ThreadedFFICalloutStateForARM64 methodsFor: 'initialize-release' stamp: 'eem 3/6/2019 07:47'!
initialize
	super initialize.
	integerRegisterIndex := floatRegisterIndex := 0.
	integerRegisters := CArrayAccessor on: (Array new: ThreadedARM64FFIPlugin numIntRegArgs).
	floatRegisters     := CArrayAccessor on: (Array new: ThreadedARM64FFIPlugin numFloatRegArgs)! !

!ThreadedFFICalloutStateForARM64 methodsFor: 'accessing' stamp: 'eem 1/20/2019 19:00'!
integerRegisterIndex

	^ integerRegisterIndex! !

!ThreadedFFICalloutStateForARM64 methodsFor: 'accessing' stamp: 'eem 1/20/2019 19:00'!
integerRegisterIndex: anObject

	^ integerRegisterIndex := anObject.! !

!ThreadedFFICalloutStateForARM64 methodsFor: 'accessing' stamp: 'eem 1/20/2019 19:00'!
integerRegisters

	^ integerRegisters! !

!ThreadedFFICalloutStateForARM64 methodsFor: 'accessing' stamp: 'eem 1/20/2019 19:00'!
integerRegisters: anObject

	^ integerRegisters := anObject.! !

ThreadedFFICalloutState subclass: #ThreadedFFICalloutStateForPPC
	instanceVariableNames: 'registerIndex integerRegisters floatRegisters'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedFFICalloutStateForPPC class methodsFor: 'translation' stamp: 'eem 2/19/2016 16:51'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a ThreadedFFICalloutState struct."

	superclass instVarNamesAndTypesForTranslationDo: aBinaryBlock.
	self instVarNames do:
		[:ivn|
		aBinaryBlock
			value: ivn
			value: (ivn caseOf: {
						['integerRegisters']	-> [{#sqInt. '[NumIntRegArgs]'}].
						['floatRegisters']	-> [{#float. '[NumFloatRegArgs]'}] }
					otherwise:
						[#sqInt])]! !

!ThreadedFFICalloutStateForPPC methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
floatRegisters
	"Answer the value of floatRegisters"

	^ floatRegisters! !

!ThreadedFFICalloutStateForPPC methodsFor: 'accessing' stamp: 'djm 4/10/2014 16:08'!
floatRegisters: anObject

	^floatRegisters := anObject! !

!ThreadedFFICalloutStateForPPC methodsFor: 'initialize-release' stamp: 'djm 4/12/2014 19:02'!
initialize
	super initialize.
	registerIndex := 0.
	integerRegisters := CArrayAccessor on: (Array new: ThreadedPPCBEFFIPlugin numRegArgs).
	floatRegisters := CArrayAccessor on: (Array new: ThreadedPPCBEFFIPlugin numRegArgs)! !

!ThreadedFFICalloutStateForPPC methodsFor: 'accessing' stamp: 'eem 10/29/2009 22:11'!
integerRegisters
	"Answer the value of integerRegisters"

	^ integerRegisters! !

!ThreadedFFICalloutStateForPPC methodsFor: 'accessing' stamp: 'djm 4/10/2014 16:08'!
integerRegisters: anObject

	^integerRegisters := anObject! !

!ThreadedFFICalloutStateForPPC methodsFor: 'accessing' stamp: 'djm 4/10/2014 16:08'!
registerIndex

	^ registerIndex! !

!ThreadedFFICalloutStateForPPC methodsFor: 'accessing' stamp: 'djm 4/10/2014 16:08'!
registerIndex: anObject

	^registerIndex := anObject! !

ThreadedFFICalloutState subclass: #ThreadedFFICalloutStateForX64
	instanceVariableNames: 'integerRegisterIndex floatRegisterIndex integerRegisters floatRegisters'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedFFICalloutStateForX64 commentStamp: 'eem 2/16/2016 19:13' prior: 0!
A ThreadedFFICalloutStateForX64 is a holder for the callout state maintained while marshalling an FFI call on an X64 (x86-64) system.!

!ThreadedFFICalloutStateForX64 class methodsFor: 'translation' stamp: 'eem 5/30/2017 18:27'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a ThreadedFFICalloutState struct."
	"self subclasses collect: [:ea| ea -> ea typdef]"
	ThreadedFFICalloutState instVarNamesAndTypesForTranslationDo: aBinaryBlock.
	self instVarNames, ThreadedFFICalloutStateForX64 instVarNames do:
		[:ivn|
		aBinaryBlock
			value: ivn
			value: (ivn caseOf: {
						['integerRegisters']	-> [{#sqInt. '[NumIntRegArgs]'}].
						['floatRegisters']	-> [{#double. '[NumFloatRegArgs]'}] }
					otherwise:
						[#char])]! !

!ThreadedFFICalloutStateForX64 class methodsFor: 'translation' stamp: 'eem 2/16/2016 19:14'!
pluginClass
	^self subclassResponsibility! !

!ThreadedFFICalloutStateForX64 methodsFor: 'accessing' stamp: 'eem 2/19/2016 19:50'!
floatRegisterIndex

	^ floatRegisterIndex! !

!ThreadedFFICalloutStateForX64 methodsFor: 'accessing' stamp: 'eem 2/16/2016 19:08'!
floatRegisters
	"Answer the value of floatRegisters"

	^ floatRegisters! !

!ThreadedFFICalloutStateForX64 methodsFor: 'accessing' stamp: 'nice 8/10/2016 18:38'!
incrementFloatRegisterIndex
	^floatRegisterIndex := floatRegisterIndex + 1! !

!ThreadedFFICalloutStateForX64 methodsFor: 'accessing' stamp: 'nice 8/10/2016 18:39'!
incrementIntegerRegisterIndex
	^integerRegisterIndex := integerRegisterIndex + 1! !

!ThreadedFFICalloutStateForX64 methodsFor: 'initialize-release' stamp: 'eem 2/20/2016 10:54'!
initialize
	super initialize.
	integerRegisterIndex := floatRegisterIndex := 0.
	integerRegisters := CArrayAccessor on: (Array new: self class pluginClass numRegArgs).
	floatRegisters := CArrayAccessor on: (Array new: self class pluginClass numFloatRegArgs)! !

!ThreadedFFICalloutStateForX64 methodsFor: 'accessing' stamp: 'eem 2/19/2016 19:50'!
integerRegisterIndex

	^ integerRegisterIndex! !

!ThreadedFFICalloutStateForX64 methodsFor: 'accessing' stamp: 'eem 2/16/2016 19:08'!
integerRegisters
	"Answer the value of integerRegisters"

	^ integerRegisters! !

ThreadedFFICalloutStateForX64 subclass: #ThreadedFFICalloutStateForX64SysV
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedFFICalloutStateForX64SysV commentStamp: 'eem 5/30/2017 10:22' prior: 0!
A ThreadedFFICalloutStateForX64SysV is a holder for the callout state maintained while marshalling an FFI call on an X64 (x86-64) system using the System V ABI.!

!ThreadedFFICalloutStateForX64SysV class methodsFor: 'accessing' stamp: 'eem 2/16/2016 19:15'!
pluginClass
	^ThreadedX64SysVFFIPlugin! !

ThreadedFFICalloutStateForX64 subclass: #ThreadedFFICalloutStateForX64Win64
	instanceVariableNames: 'floatRegisterSignature'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedFFICalloutStateForX64Win64 commentStamp: 'eem 5/30/2017 10:22' prior: 0!
A ThreadedFFICalloutStateForX64Win64 is a holder for the callout state maintained while marshalling an FFI call on an X64 (x86-64) system using the _WIN64 ABI.!

!ThreadedFFICalloutStateForX64Win64 class methodsFor: 'accessing' stamp: 'eem 2/16/2016 19:15'!
pluginClass
	^ThreadedX64Win64FFIPlugin! !

!ThreadedFFICalloutStateForX64Win64 methodsFor: 'accessing' stamp: 'nice 8/10/2016 18:59'!
floatRegisterSignature
	^floatRegisterSignature! !

!ThreadedFFICalloutStateForX64Win64 methodsFor: 'accessing' stamp: 'nice 8/11/2016 18:51'!
incrementFloatRegisterIndex
	"There are only 4 args passed by register int or float.
	So we can't distinguish the float index from the integer index.
	So we have to increment both.
	
	Consequently, floatRegisterIndex cannot be used anymore to detect presence of float parameter.
	However, we set a signature bitmap indicating which register position is used to pass a float.
	
	IMPLEMENTATION NOTES:
	There are code generator hacks that bypass the accessors.
	So we cannot just redefine the method floatRegisterIndex as ^integerRegisterIndex.
	Instead we must maintain the two indices"

	floatRegisterSignature := floatRegisterSignature + (1 << floatRegisterIndex).
	^integerRegisterIndex := floatRegisterIndex := floatRegisterIndex + 1! !

!ThreadedFFICalloutStateForX64Win64 methodsFor: 'accessing' stamp: 'nice 8/11/2016 18:50'!
incrementIntegerRegisterIndex
	"There are only 4 args passed by register int or float.
	So we can't distinguish the float index from the integer index.
	So we have to increment both.
	
	IMPLEMENTATION NOTES:
	There are code generator hacks that bypass the accessors.
	So we cannot just redefine the method floatRegisterIndex as ^integerRegisterIndex.
	Instead we must maintain the two indices"

	^floatRegisterIndex := integerRegisterIndex := integerRegisterIndex + 1! !

!ThreadedFFICalloutStateForX64Win64 methodsFor: 'initialize-release' stamp: 'nice 8/10/2016 18:59'!
initialize
	super initialize.
	floatRegisterSignature := 0.! !

VMStructType subclass: #ThreadedFFIX64SixteenByteReturnDD
	instanceVariableNames: 'a b'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedFFIX64SixteenByteReturnDD commentStamp: 'nice 1/25/2020 17:43' prior: 0!
A ThreadedFFIX64SixteenByteReturnDD is a stub for returning a struct by value through 2 eight-byte float registers
That is (XMM0 XMM1) on X64.

Instance Variables
	a:		<Object> 
	b:		<Object> 

a
	- stub for first eighbyte

b
	- stub for second eighbyte
!

!ThreadedFFIX64SixteenByteReturnDD class methodsFor: 'translation' stamp: 'nice 1/25/2020 17:00'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a BlockStart struct."

	aBinaryBlock value: 'a' value: #double.
	aBinaryBlock value: 'b' value: #double.! !

!ThreadedFFIX64SixteenByteReturnDD class methodsFor: 'translation' stamp: 'eem 2/20/2016 10:43'!
structTypeName
	^'SixteenByteReturnDD'! !

!ThreadedFFIX64SixteenByteReturnDD methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
a

	^ a! !

!ThreadedFFIX64SixteenByteReturnDD methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
a: anObject

	^a := anObject! !

!ThreadedFFIX64SixteenByteReturnDD methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
b

	^ b! !

!ThreadedFFIX64SixteenByteReturnDD methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
b: anObject

	^b := anObject! !

VMStructType subclass: #ThreadedFFIX64SixteenByteReturnDI
	instanceVariableNames: 'a b'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedFFIX64SixteenByteReturnDI commentStamp: 'nice 1/25/2020 17:44' prior: 0!
A ThreadedFFIX64SixteenByteReturnDI is a stub for returning a struct by value through 1 eight-byte float register and 1 eight-byte int register
That is (XMM0 RAX) on X64.

Instance Variables
	a:		<Object> 
	b:		<Object> 

a
	- stub for first eighbyte

b
	- stub for second eighbyte
!

!ThreadedFFIX64SixteenByteReturnDI class methodsFor: 'translation' stamp: 'nice 1/25/2020 17:00'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a BlockStart struct."

	aBinaryBlock value: 'a' value: #double.
	aBinaryBlock value: 'b' value: #sqInt.! !

!ThreadedFFIX64SixteenByteReturnDI class methodsFor: 'translation' stamp: 'eem 2/20/2016 10:43'!
structTypeName
	^'SixteenByteReturnDI'! !

!ThreadedFFIX64SixteenByteReturnDI methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
a

	^ a! !

!ThreadedFFIX64SixteenByteReturnDI methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
a: anObject

	^a := anObject! !

!ThreadedFFIX64SixteenByteReturnDI methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
b

	^ b! !

!ThreadedFFIX64SixteenByteReturnDI methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
b: anObject

	^b := anObject! !

VMStructType subclass: #ThreadedFFIX64SixteenByteReturnID
	instanceVariableNames: 'a b'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedFFIX64SixteenByteReturnID commentStamp: 'nice 1/25/2020 17:45' prior: 0!
A ThreadedFFIX64SixteenByteReturnID is a stub for returning a struct by value through 1 eight-byte int register and 1 eight-byte float register
That is (RAX XMM0) on X64.

Instance Variables
	a:		<Object> 
	b:		<Object> 

a
	- stub for first eighbyte

b
	- stub for second eighbyte
!

!ThreadedFFIX64SixteenByteReturnID class methodsFor: 'translation' stamp: 'nice 1/25/2020 17:00'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a BlockStart struct."

	aBinaryBlock value: 'a' value: #sqInt.
	aBinaryBlock value: 'b' value: #double.! !

!ThreadedFFIX64SixteenByteReturnID class methodsFor: 'translation' stamp: 'eem 2/20/2016 10:43'!
structTypeName
	^'SixteenByteReturnID'! !

!ThreadedFFIX64SixteenByteReturnID methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
a

	^ a! !

!ThreadedFFIX64SixteenByteReturnID methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
a: anObject

	^a := anObject! !

!ThreadedFFIX64SixteenByteReturnID methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
b

	^ b! !

!ThreadedFFIX64SixteenByteReturnID methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
b: anObject

	^b := anObject! !

VMStructType subclass: #ThreadedFFIX64SixteenByteReturnII
	instanceVariableNames: 'a b'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!ThreadedFFIX64SixteenByteReturnII commentStamp: 'nice 1/25/2020 17:44' prior: 0!
A ThreadedFFIX64SixteenByteReturnII is a stub for returning a struct by value through 2 eight-byte int registers
That is (RAX RDX) on X64.

Instance Variables
	a:		<Object> 
	b:		<Object> 

a
	- stub for first eighbyte

b
	- stub for second eighbyte
!

!ThreadedFFIX64SixteenByteReturnII class methodsFor: 'translation' stamp: 'nice 1/25/2020 17:00'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a BlockStart struct."

	aBinaryBlock value: 'a' value: #sqInt.
	aBinaryBlock value: 'b' value: #sqInt.! !

!ThreadedFFIX64SixteenByteReturnII class methodsFor: 'translation' stamp: 'eem 2/20/2016 10:43'!
structTypeName
	^'SixteenByteReturnII'! !

!ThreadedFFIX64SixteenByteReturnII methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
a

	^ a! !

!ThreadedFFIX64SixteenByteReturnII methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
a: anObject

	^a := anObject! !

!ThreadedFFIX64SixteenByteReturnII methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
b

	^ b! !

!ThreadedFFIX64SixteenByteReturnII methodsFor: 'accessing' stamp: 'eem 2/20/2016 11:02'!
b: anObject

	^b := anObject! !

VMStructType subclass: #VMCallbackContext
	instanceVariableNames: 'thunkp stackp intregargsp floatregargsp savedCStackPointer savedCFramePointer rvs savedPrimFunctionPointer savedMostRecentCallbackContext trampoline savedReenterInterpreter'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Support'!

!VMCallbackContext class methodsFor: 'translation' stamp: 'eem 3/13/2017 08:28'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"Define a CallbackContext, the argument to sendInvokeCallbackContext:
	 self typedef"

	self instVarNames do:
		[:ivn|
		aBinaryBlock
			value: ivn
			value: (ivn caseOf: {
						['stackp']				-> [#'sqIntptr_t *'].
						['intregargsp']			-> [#'sqIntptr_t *'].
						['floatregargsp']		-> [#'double *'].
						['rvs']					-> [
							'union {
								sqIntptr_t valword;
								struct { int low, high; } valleint64;
								struct { int high, low; } valbeint64;
								double valflt64;
								struct { void *addr; sqIntptr_t size; } valstruct;
							}'].
						['trampoline']				-> [#'jmp_buf'].
						['savedReenterInterpreter']	-> [#'jmp_buf']}
						otherwise: [#'void *'])]! !

!VMCallbackContext class methodsFor: 'translation' stamp: 'eem 12/25/2010 22:49'!
needsTypeTag
	"This allows sqVirtualMachine.h to declare VMCallbackContext  as an
	 opaque type avoiding everyone including setjmp.h & vmCallback.h"
	^true! !

!VMCallbackContext class methodsFor: 'translation' stamp: 'eem 12/25/2010 22:32'!
printTypedefOn: aStream
	super printTypedefOn: aStream.
	aStream
		cr;
		nextPutAll: '/* The callback return type codes */'; cr;
		nextPutAll: '#define retword 1'; cr;
		nextPutAll: '#define retword64 2'; cr;
		nextPutAll: '#define retdouble 3'; cr;
		nextPutAll: '#define retstruct 4'; cr;
		cr! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/13/2010 16:47'!
floatregargsp
	"Answer the value of floatregargsp"

	^ floatregargsp! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/13/2010 16:47'!
floatregargsp: anObject
	"Set the value of floatregargsp"

	^floatregargsp := anObject! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/13/2010 16:47'!
intregargsp
	"Answer the value of intregargsp"

	^ intregargsp! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/13/2010 16:47'!
intregargsp: anObject
	"Set the value of intregargsp"

	^intregargsp := anObject! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/13/2010 16:47'!
rvs
	"Answer the value of rvs"

	^ rvs! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/13/2010 16:47'!
rvs: anObject
	"Set the value of rvs"

	^rvs := anObject! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/12/2010 18:45'!
savedCFramePointer
	"Answer the value of savedCFramePointer"

	^ savedCFramePointer! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/12/2010 18:45'!
savedCFramePointer: anObject
	"Set the value of savedCFramePointer"

	^savedCFramePointer := anObject! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/12/2010 18:45'!
savedCStackPointer
	"Answer the value of savedCStackPointer"

	^ savedCStackPointer! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/12/2010 18:45'!
savedCStackPointer: anObject
	"Set the value of savedCStackPointer"

	^savedCStackPointer := anObject! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/17/2020 18:23'!
savedMostRecentCallbackContext
	"Answer the value of savedMostRecentCallbackContext"

	^ savedMostRecentCallbackContext! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/17/2020 18:23'!
savedMostRecentCallbackContext: anObject
	"Set the value of savedMostRecentCallbackContext"

	^savedMostRecentCallbackContext := anObject! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 3/13/2017 08:27'!
savedPrimFunctionPointer
	
	^ savedPrimFunctionPointer
! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 3/13/2017 08:27'!
savedPrimFunctionPointer: anObject
	
	^ savedPrimFunctionPointer := anObject.
! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 11/5/2010 16:56'!
savedReenterInterpreter
	"Answer the value of savedReenterInterpreter"

	^ savedReenterInterpreter! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 11/5/2010 16:56'!
savedReenterInterpreter: anObject
	"Set the value of savedReenterInterpreter"

	^savedReenterInterpreter := anObject! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/12/2010 18:45'!
stackp
	"Answer the value of stackp"

	^ stackp! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/12/2010 18:45'!
stackp: anObject
	"Set the value of stackp"

	^stackp := anObject! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/12/2010 18:45'!
thunkp
	"Answer the value of thunkp"

	^ thunkp! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/12/2010 18:45'!
thunkp: anObject
	"Set the value of thunkp"

	^thunkp := anObject! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/12/2010 18:45'!
trampoline
	"Answer the value of trampoline"

	^ trampoline! !

!VMCallbackContext methodsFor: 'accessing' stamp: 'eem 10/12/2010 18:45'!
trampoline: anObject
	"Set the value of trampoline"

	^trampoline := anObject! !

VMStructType subclass: #VMCallbackReturnValue
	instanceVariableNames: 'type pad crvrvs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Plugins-FFI'!

!VMCallbackReturnValue commentStamp: 'eem 10/14/2010 09:29' prior: 0!
A VMCallbackReturnValue supports legacy Alien callbacks.!

!VMCallbackReturnValue class methodsFor: 'translation' stamp: 'eem 10/14/2010 09:43'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"Define a CallbackContext, the argument to sendInvokeCallbackContext:
	 self typedef"

	| rvsType |
	VMCallbackContext instVarNamesAndTypesForTranslationDo:
		[:ivname :type| ivname = 'rvs' ifTrue: [rvsType := type]].
	self instVarNames do:
		[:ivn|
		aBinaryBlock
			value: ivn
			value: (ivn = 'crvrvs'
					ifTrue: [rvsType]
					ifFalse: [#int])]! !

!VMCallbackReturnValue methodsFor: 'accessing' stamp: 'eem 10/14/2010 09:41'!
crvrvs
	^crvrvs! !

!VMCallbackReturnValue methodsFor: 'accessing' stamp: 'eem 10/14/2010 09:42'!
crvrvs: anObject
	^crvrvs := anObject! !

!VMCallbackReturnValue methodsFor: 'accessing' stamp: 'eem 10/14/2010 09:40'!
type
	"Answer the value of type"

	^ type! !

!VMCallbackReturnValue methodsFor: 'accessing' stamp: 'eem 10/14/2010 09:40'!
type: anObject
	"Set the value of type"

	^type := anObject! !

!VMStructType class methodsFor: 'code generation' stamp: 'eem 5/30/2020 18:44'!
addStructTypeNamesTo: aSet
	aSet add: self name; add: self structTypeName! !

!VMStructType class methodsFor: 'accessing class hierarchy' stamp: 'eem 12/11/2010 09:32'!
addSubclass: aSubclass
	self voidStructTypeCache.
	^super addSubclass: aSubclass! !

!VMStructType class methodsFor: 'simulation only' stamp: 'eem 7/18/2013 11:40'!
alignedByteSizeOf: objectSymbolOrClass forClient: aVMClass
	^objectSymbolOrClass byteSizeForSimulator: aVMClass! !

!VMStructType class methodsFor: 'code generation' stamp: 'eem 1/28/2020 10:38'!
changedAccessorsForSurrogate: surrogateClass bytesPerWord: bytesPerWord
	"Answer the changed accessor methods for the fields of the receiver and the alignedByteSize class method."

	"{CogBlockMethod changedAccessorsForSurrogate: CogBlockMethodSurrogate32 bytesPerWord: 4.
	 CogMethod changedAccessorsForSurrogate: CogMethodSurrogate32 bytesPerWord: 4.
	 CogBlockMethod changedAccessorsForSurrogate: CogBlockMethodSurrogate64 bytesPerWord: 8.
	 CogMethod changedAccessorsForSurrogate: CogMethodSurrogate64 bytesPerWord: 8}"

	| dualOrNil |
	dualOrNil := self dualForSurrogateClass: surrogateClass.
	^Dictionary withAll: ((self fieldAccessorSourceFor: surrogateClass bytesPerWord: bytesPerWord) associationsSelect:
		(dualOrNil
			ifNotNil: "The source is changed if neither the class nor its superclass have a method with identical source? No; cuz if the subclass has it and the super doesn't then it has changed; but this is tedious to compute. checkGenerateSurrogate:bytesPerWord: does the computation."
				[[:assoc| | class selector |
				 class := assoc key first.
				 selector := assoc key last.
				 (class sourceCodeAt: selector ifAbsent: ['']) asString ~= assoc value
				 and: [(class superclass sourceCodeAt: selector ifAbsent: ['']) asString ~= assoc value]]]
			ifNil:
				[[:assoc| (assoc key first sourceCodeAt: assoc key last ifAbsent: ['']) asString ~= assoc value]]))! !

!VMStructType class methodsFor: 'code generation' stamp: 'eem 1/28/2020 16:20'!
checkGenerateSurrogate: surrogateClass bytesPerWord: bytesPerWord
	"Check the accessor methods for the fields of the receiver and if necessary install new
	 or updated versions in the surrogate class alpng with the alignedByteSize class method."

	"self withAllSubclasses do:
		[:cogMethodClass| (cogMethodClass class includesSelector: #initialize) ifTrue: [cogMethodClass initialize]]"

	"CogBlockMethod checkGenerateSurrogate: CogBlockMethodSurrogate32 bytesPerWord: 4.
	 CogMethod checkGenerateSurrogate: CogMethodSurrogate32 bytesPerWord: 4.
	 CogBlockMethod checkGenerateSurrogate: CogBlockMethodSurrogate64 bytesPerWord: 8.
	 CogMethod checkGenerateSurrogate: CogMethodSurrogate64 bytesPerWord: 8"
	| accessors oldBytesPerWord |
	oldBytesPerWord := BytesPerWord.
	accessors := [self fieldAccessorSourceFor: surrogateClass bytesPerWord: (BytesPerWord := bytesPerWord)]
					ensure: [BytesPerWord := oldBytesPerWord].
	
	"All methods which are the same in 32 and 64 bit versions should be compiled in the superclass iff the superclass is not of a specific word size."
	(self dualForSurrogateClass: surrogateClass) ifNotNil:
		[:dual| | duals |
		 surrogateClass superclass name last isDigit ifFalse:
			[duals := [self fieldAccessorSourceFor: dual bytesPerWord: (BytesPerWord := bytesPerWord = 4 ifTrue: [8] ifFalse: [4])]
						ensure: [BytesPerWord := oldBytesPerWord].

			 (accessors keys select: [:key| (accessors at: key) = (duals at: {key first isMeta ifTrue: [dual class] ifFalse: [dual]. key last} ifAbsent: ['missing'])]) do:
				[:key| | source class selector |
				class := key first.
				selector := key last.
				source := accessors removeKey: key.
				(class includesSelector: selector) ifTrue:
					[class removeSelector: selector].
				(class superclass sourceCodeAt: selector ifAbsent: ['']) asString ~= source ifTrue:
					[class superclass compile: source classified: 'accessing generated']]]].
		
	accessors keysAndValuesDo:
		[:key :source| | class selector ancestor |
		class := key first.
		selector := key last.
		((ancestor := class superclass whichClassIncludesSelector: selector) notNil
		 and: [source = (ancestor sourceCodeAt: selector ifAbsent: ['']) asString])
			ifTrue: [class removeSelector: selector]
			ifFalse:
				[source ~= (class sourceCodeAt: selector ifAbsent: ['']) asString ifTrue:
					[class compile: source classified: 'accessing generated']]]! !

!VMStructType class methodsFor: 'code generation' stamp: 'eem 1/28/2020 10:36'!
dualForSurrogateClass: surrogateClass
	"Answer the corresponding class for a word size specific surrogate class, or nil if it is not, or has no dual."
	^surrogateClass name last isDigit ifTrue:
	 	[Smalltalk classNamed: ((surrogateClass name allButLast: 2), ((surrogateClass name endsWith: '32') ifTrue: ['64'] ifFalse: ['32']))]! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 7/16/2020 16:07'!
ensureStructTypeCache
	^StructTypeNameCache ifNil:
		[StructTypeNameCache := Set new.
		 self allSubclassesDo:
			[:sc| sc addStructTypeNamesTo: StructTypeNameCache].
		 StructTypeNameCache]! !

!VMStructType class methodsFor: 'code generation' stamp: 'eem 1/28/2020 16:38'!
fieldAccessorSourceFor: surrogateClass bytesPerWord: bytesPerWord
	"Answer a Dictionary of MethodReference to source for the accessors of the inst vars of the
	 receiver and the alignedByteSize class method in surrogateClass with the given word size."

	"{CogBlockMethod fieldAccessorSourceFor: CogBlockMethodSurrogate32 bytesPerWord: 4.
	 CogMethod fieldAccessorSourceFor: CogMethodSurrogate32 bytesPerWord: 4.
	 CogBlockMethod fieldAccessorSourceFor: CogBlockMethodSurrogate64 bytesPerWord: 8.
	 CogMethod fieldAccessorSourceFor: CogMethodSurrogate64 bytesPerWord: 8}"
	| methods bitPosition alignedByteSize currentOffset |
	methods := Dictionary new.
	bitPosition := 0.
	(self fieldAccessorsForBytesPerWord: bytesPerWord) do:
		[:spec|
		"reset the bitPosition if the offset expression changes."
		currentOffset ~= (self offsetForInstVar: spec first) ifTrue:
			[bitPosition := 0.
			 currentOffset := self offsetForInstVar: spec first].
		spec first ~= #unused ifTrue:
			[methods
				at: {surrogateClass. spec first asSymbol}
					put: (self getter: spec first
							 bitPosition: bitPosition
							 bitWidth: spec second
							 type: (spec at: 3 ifAbsent: []));
				at: {surrogateClass. (spec first, ':') asSymbol}
					put: (self setter: spec first
							 bitPosition: bitPosition
							 bitWidth: spec second
							 type: (spec at: 3 ifAbsent: []))].
		bitPosition := bitPosition + spec second].
	alignedByteSize := (self roundUpBitPosition: bitPosition toWordBoundary: bytesPerWord) / 8.
	self assert: alignedByteSize isInteger.
	methods
		at: {surrogateClass class. #alignedByteSize}
		put: #alignedByteSize
			, (String with: Character cr with: Character tab with: $^)
			, alignedByteSize printString,
			(currentOffset ifNil: [''] ifNotNil: [' + self ', currentOffset]).
	^methods! !

!VMStructType class methodsFor: 'code generation' stamp: 'eem 11/17/2015 13:31'!
fieldAccessorsForBytesPerWord: bytesPerWord
	| fieldSpecs |
	fieldSpecs := OrderedCollection new.
	self instVarNamesAndTypesForTranslationDo:
		[:ivn :typeTuple| | index replacement |
		(index := typeTuple indexOf: #BytesPerWord ifAbsent: 0) > 0
			ifTrue:
				[(typeTuple at: index + 1) = bytesPerWord ifTrue:
					[replacement := typeTuple copyReplaceFrom: index to: index + 1 with: #().
					 replacement size = 1 ifTrue:
						[replacement := replacement first].
					fieldSpecs add: { ivn. replacement }]]
			ifFalse:
				[fieldSpecs add: { ivn. typeTuple }]].
	^fieldSpecs collect:
		[:tuple|
			[:ivn :typeTuple|
			{ ('*unused*' match: ivn) ifTrue: [#unused] ifFalse: [ivn].
			  (typeTuple isArray and: ['unsigned' = typeTuple first])
				ifTrue:
					[Integer readFrom: (typeTuple last readStream skipTo: $:; skipSeparators)]
				ifFalse:
					[typeTuple
						caseOf: {
								[#char]				->	[8].
								[#'unsigned char']	->	[8].
								[#short]			->	[16].
								[#'unsigned short']	->	[16].
								[#int]				->	[32].
								[#'unsigned int']	->	[32] }
						otherwise: [bytesPerWord * 8]].
			typeTuple isArray
				ifTrue:
					[(typeTuple size >= 3 and: [typeTuple second = #Boolean]) ifTrue:
						[#Boolean]]
				ifFalse:
					[typeTuple last = $* ifTrue:
						[(typeTuple beginsWith: 'struct _') "remove struct tag if any"
							ifTrue: [(typeTuple allButFirst: 8) asSymbol]
							ifFalse: [typeTuple]]] }] valueWithArguments: tuple]

	"#(4 8) collect: [:bpw| (CogBlockMethod fieldAccessorsForBytesPerWord: bpw) asArray]"
	"#(4 8) collect: [:bpw| (CogMethod fieldAccessorsForBytesPerWord: bpw) asArray]"! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 12/11/2010 09:23'!
filteredInstVarNames
	"Eliminate the obvious simulation-only inst vars"
	^self allInstVarNames reject:
		[:n|
		Cogit isNonArgumentImplicitReceiverVariableName: n]! !

!VMStructType class methodsFor: 'code generation' stamp: 'eem 9/28/2015 20:53'!
getter: getter bitPosition: bitPosition bitWidth: bitWidth type: typeOrNil
	^String streamContents:
		[:s| | startByte endByte alignedPowerOf2 shift |
		startByte := bitPosition // 8.
		endByte := bitPosition + bitWidth - 1 // 8.
		shift := bitPosition \\ 8.
		alignedPowerOf2 := (#(8 16 32 64) includes: bitWidth) and: [shift = 0].
		s nextPutAll: getter; crtab: 1.
		(typeOrNil notNil and: [typeOrNil last = $*]) ifTrue:
			[s nextPutAll: '| v |'; crtab: 1].
		s nextPut: $^.
		typeOrNil ifNotNil:
			[s nextPut: $(.
			 typeOrNil last = $* ifTrue:
				[s nextPutAll: 'v := ']].
		alignedPowerOf2 ifFalse:
			[s nextPut: $(].
		shift ~= 0 ifTrue:
			[s nextPut: $(].
		s nextPutAll: 'memory unsigned';
		   nextPutAll: (#('Byte' 'Short' 'Long' 'Long')
							at: endByte - startByte + 1
							ifAbsent: ['Long64']);
		  nextPutAll: 'At: address + '; print: startByte + 1.
		(self offsetForInstVar: getter) ifNotNil:
			[:offsetExpr| s nextPutAll: ' + '; nextPutAll: offsetExpr].
		shift ~= 0 ifTrue:
			[s nextPutAll: ') bitShift: -'; print: shift].
		alignedPowerOf2 ifFalse:
			[s nextPutAll: ') bitAnd: '; nextPutAll: ((1 << bitWidth) - 1) hex].
		typeOrNil ifNotNil:
			[s nextPutAll: ') ~= 0'.
			typeOrNil last = $* ifTrue:
				[s nextPutAll: ' ifTrue:';
					crtab: 2;
					nextPutAll: '[cogit cCoerceSimple: v to: ';
					store: typeOrNil;
					nextPut: $]]]]! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 2/15/2009 09:23'!
implicitReturnTypeFor: aSelector
	"Answer the return type for methods that don't have an explicit return."
	^#void! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 2/10/2009 08:54'!
instVarNamesAndTypesForTranslationDo: aBinaryBlock
	"enumerate aBinaryBlock with the names and C type strings for the inst vars to include in a struct of this type."

	self subclassResponsibility! !

!VMStructType class methodsFor: 'accessing class hierarchy' stamp: 'eem 8/8/2012 13:11'!
isAbstract
	^self == VMStructType! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 2/11/2009 09:45'!
isAccessor: aSelector
	"Answer if aSelector is simply an accessor method for one of our fields."
	^(self instVarIndexFor: (aSelector copyWithout: $:) ifAbsent: [0]) > 0! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 12/11/2010 13:00'!
isNonArgumentImplicitReceiverVariableName: aString
	^Cogit isNonArgumentImplicitReceiverVariableName: aString! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 11/26/2012 15:06'!
isStructClass
	^true! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 8/29/2020 08:55'!
isTypePointerToStruct: type
	| index |
	^type isString
	  and: [(index := type indexOf: $*) > 0
	  and: [self ensureStructTypeCache anySatisfy:
			[:structType|
			(type beginsWith: structType)
			and: [index > structType size]]]]! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 7/16/2020 15:38'!
isTypeStruct: type
	self ensureStructTypeCache.
	^type notNil
	  and: [StructTypeNameCache anySatisfy:
			[:structType|
			type = structType]]! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 04/27/2017 12:55'!
needsTypeTag
	self instVarNamesAndTypesForTranslationDo:
		[:ivn :type|
		((type isArray ifTrue: [type] ifFalse: [{type}]) anySatisfy:
			[:str|
			str isString
			and: [str includesSubstring: self structTagName]]) ifTrue:
				[^true]].
	^false

	"VMStructType allSubclasses select: [:sc| sc needsTypeTag]"! !

!VMStructType class methodsFor: 'code generation' stamp: 'eem 10/1/2013 11:26'!
offsetForInstVar: instVarName
	"Hack to offset accesses to variables by certain values."
	^nil! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 10/25/2020 09:53'!
printTypedefOn: aStream
	aStream nextPutAll: 'typedef struct '.
	self needsTypeTag ifTrue:
		[aStream nextPutAll: self structTagName; space].
	aStream nextPut: ${; cr.
	self instVarNamesAndTypesForTranslationDo:
		[:ivn :typeArg| | type |
		ivn first == $#
			ifTrue: [aStream nextPutAll: ivn]
			ifFalse:
				[type := typeArg.
				 #(BytesPerWord BaseHeaderSize BytesPerOop) do:
					[:sizeConstant| | index sizeConstantSize |
					(type notNil
					and: [(index := type indexOf: sizeConstant ifAbsent: 0) > 0]) ifTrue:
						[sizeConstantSize  := VMBasicConstants classPool at: sizeConstant.
						type := (type at: index + 1) = sizeConstantSize ifTrue:
									[type := type copyReplaceFrom: index to: index + 1 with: #().
									 type size = 1 ifTrue: [type first] ifFalse: [type]]]].
				 type ifNotNil:
					[type isArray
						ifTrue:
							[aStream tab: 1.
							 aStream nextPutAll: type first.
							 (type first last isSeparator or: [type first last = $*]) ifFalse:
								[aStream tab: 2].
							 aStream nextPutAll: ivn.
							 type last first isAlphaNumeric ifTrue:
								[aStream space].
							 aStream nextPutAll: type last]
						ifFalse:
							[aStream tab: 1.
							 aStream nextPutAll: type.
							 type last isAlphaNumeric ifTrue:
								[aStream tab: 1].
							 aStream nextPutAll: ivn]].
				 aStream nextPut: $;].
		 aStream cr].
	aStream
		nextPutAll: ' } ';
		nextPutAll: self structTypeName;
		nextPut: $;;
		cr.
	self name ~= self structTypeName ifTrue:
		[(self withAllSuperclasses copyUpThrough: (self class whichClassIncludesSelector: #structTypeName) theNonMetaClass) do:
			[:structClass|
			 aStream cr; nextPutAll: '#define '; nextPutAll: structClass name; space; nextPutAll: self structTypeName].
		 aStream cr].
	aStream flush! !

!VMStructType class methodsFor: 'code generation' stamp: 'eem 1/28/2020 08:33'!
putAtPut: accessor type: typeOrNil mask: maskOrNil shift: shift on: s indent: indent
	"This is the inner part of the ap:put: in a setter, abstracted to eliminate duplication
	 given the overrides in CogBlockMethod et al for dual zone write simulation."
	s nextPutAll: 'memory';
	  crtab: indent; nextPutAll: accessor.
	s crtab: indent; nextPutAll: 'put: '.
	typeOrNil ifNotNil:
		[s nextPut: $(].
	maskOrNil ifNotNil:
		[s nextPutAll: '((memory '; nextPutAll: accessor;
		    nextPutAll: ') bitAnd: '; nextPutAll: maskOrNil hex;
		    nextPutAll: ') + '].
	s nextPutAll: (typeOrNil
					caseOf: {
						[nil] -> [shift = 0 ifTrue: ['aValue'] ifFalse: ['(aValue bitShift: ', shift printString, ')']].
						[#Boolean] -> ['(aValue ifTrue: [', (1 << shift) printString, '] ifFalse: [0])'] }
					otherwise: ['(aValue ifNotNil: [aValue asUnsignedInteger', (shift = 0 ifTrue: [''] ifFalse: [' bitShift: ', shift printString]), '] ifNil: [0])']).
	typeOrNil ifNotNil:
		[s nextPut: $)]! !

!VMStructType class methodsFor: 'accessing class hierarchy' stamp: 'eem 12/11/2010 09:32'!
removeSubclass: aSubclass
	self voidStructTypeCache.
	^super removeSubclass: aSubclass! !

!VMStructType class methodsFor: 'class name' stamp: 'eem 12/11/2010 09:33'!
rename: aString
	self voidStructTypeCache.
	^super rename: aString! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 3/5/2009 16:35'!
returnTypeForAccessor: aSelector
	| varName deStructTag |
	varName := aSelector copyWithout: $:.
	deStructTag := [:decl| decl copyReplaceAll: 'struct ', self structTagName with: self structTypeName].
	self instVarNamesAndTypesForTranslationDo:
		[:var :typeSpec|
		var = varName ifTrue:
			[^typeSpec isArray
				ifTrue: [(deStructTag value: typeSpec first), ' ', typeSpec last]
				ifFalse: [deStructTag value: typeSpec]]].
	^nil


	"InterpreterStackPage returnTypeForAccessor: #nextPage:"! !

!VMStructType class methodsFor: 'code generation' stamp: 'eem 11/3/2015 17:26:46'!
roundUpBitPosition: bitPosition toWordBoundary: bytesPerWord
	^bitPosition + 7 // 8 + bytesPerWord - 1 // bytesPerWord * bytesPerWord * 8! !

!VMStructType class methodsFor: 'code generation' stamp: 'eem 1/29/2020 08:52'!
setter: getter bitPosition: bitPosition bitWidth: bitWidth type: typeOrNil
	^String streamContents:
		[:s| | startByte endByte shift indexExpr accessor alignedPowerOf2 mask |
		startByte := bitPosition // 8.
		endByte := bitPosition + bitWidth - 1 // 8.
		shift := bitPosition \\ 8.
		s nextPutAll: getter; nextPutAll: ': aValue'; crtab: 1.
		indexExpr := 'address + ',
						((self offsetForInstVar: getter) ifNotNil: [:offsetExpr| offsetExpr, ' + '] ifNil: ['']),
						(startByte + 1) printString.
		accessor := 'unsigned'
					, (#('Byte' 'Short' 'Long' 'Long')
							at: endByte - startByte + 1
							ifAbsent: ['Long64'])
					, 'At: '.
		(alignedPowerOf2 := (#(8 16 32 64) includes: bitWidth) and: [shift = 0])
			ifTrue: "index used once; no point using a temp to hold it"
				[accessor := accessor, indexExpr]
			ifFalse: "index used twice; cache it"
				[s nextPutAll: '| index |'; crtab: 1; nextPutAll: 'index := '; nextPutAll: indexExpr; nextPut: $.; crtab: 1.
				 accessor := accessor, 'index'].
		mask := #(16rFF 16rFFFF 16rFFFFFFFF 16rFFFFFFFF)
						at: endByte - startByte + 1
						ifAbsent: [(2 raisedTo: 64) - 1].
		(typeOrNil notNil or: [alignedPowerOf2]) ifFalse:
			[s nextPutAll: 'self assert: (aValue between: 0 and: '; nextPutAll:  ((1 << bitWidth) - 1) hex; nextPutAll: ').'; crtab: 1].
		alignedPowerOf2 ifTrue:
			[s nextPut: $^].
		self putAtPut: accessor
			type: typeOrNil
			mask: (alignedPowerOf2 ifFalse: [mask - ((1 << bitWidth - 1) << shift)])
			shift: shift
			on: s
			indent: 2.
		alignedPowerOf2 ifFalse:
			[s nextPut: $.; crtab: 1; nextPutAll: '^aValue']]! !

!VMStructType class methodsFor: 'accessors' stamp: 'eem 8/12/2009 11:46'!
settersReturnValue
	"Override to get the browser to generate setters that return the
	 value set, which matches C semantics for field assignments."
	^true! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 2/10/2009 14:11'!
structTagName
	^'_', self structTypeName! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 7/16/2020 15:39'!
structTargetKindForDeclaration: decl
	^(decl notNil
	   and: [(self ensureStructTypeCache includes: decl)
			or: [StructTypeNameCache anySatisfy:
					[:structType|
					(decl beginsWith: structType)
					and: [(decl indexOf: $* ifAbsent: [decl indexOf: Character space]) > structType size]]]]) ifTrue:
		[(decl indexOf: $*) > 0
			ifTrue: [#pointer]
			ifFalse: [#struct]]! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 7/16/2020 15:39'!
structTargetKindForType: type
	self ensureStructTypeCache.
	^(type notNil
	   and: [StructTypeNameCache anySatisfy:
			[:structType|
			(type beginsWith: structType)
			and: [type size = structType size
				or: [(type at: structType size + 1) isAlphaNumeric not]]]]) ifTrue:
		[(type includes: $*)
			ifTrue: [#pointer]
			ifFalse: [#struct]]! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 2/9/2009 15:34'!
structTypeName
	"Provide the typedef name.  Subclases can override."
	^self name! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 2/11/2009 10:21'!
typeForSelf
	"Answer the type to give self if appropriate, or nil if not."
	^self structTypeName, ' *'! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 10/12/2010 11:08'!
typedef
	^String streamContents: [:s| self printTypedefOn: s]! !

!VMStructType class methodsFor: 'translation' stamp: 'eem 12/11/2010 09:31'!
voidStructTypeCache
	StructTypeNameCache := nil! !

!VMStructType methodsFor: 'printing' stamp: 'eem 12/1/2010 10:47'!
printOn: aStream
	"A hook to allow subclasses to print their state if useful."
	<doNotGenerate>
	super printOn: aStream.
	self printStateOn: aStream! !

!VMStructType methodsFor: 'printing' stamp: 'eem 12/1/2010 10:50'!
printStateOn: aStream
	"A hook to allow subclasses to print their state if useful."
	<doNotGenerate>
	^self! !

Object subclass: #VMMaker
	instanceVariableNames: 'inline forBrowser allPlugins internalPlugins externalPlugins platformName sourceDirName platformRootDirName logger interpreterClassName is64BitVM optionsDictionary abortBlock'
	classVariableNames: 'DirNames'
	poolDictionaries: ''
	category: 'VMMaker-Building'!

!VMMaker commentStamp: 'eem 3/27/2017 15:46' prior: 0!
This class coordinates generating sources for the various VMs and the set of plugins that comprise the Cog VM.  See the class side for generating all or some configurations.  The source is intended to be generated in the context of the opensmalltalk-vm source tree available via git from https://github.com/OpenSmalltalk/opensmalltalk-vm.  See VMMaker class>>initialize for the relative locations of this image and the source tree (for example, in the image directory of the opensmalltalk-vm tree).  Invoke a single configuration via e.g.
	VMMaker generateSqueakSpurCog64VM
or a set via e.g.
	VMMaker generateAllConfigurationsUnderVersionControl

The source tree on branch Cog at https://github.com/OpenSmalltalk/opensmalltalk-vm is generated via generateAllConfigurationsUnderVersionControl.  Note that "Squeak" in these cofiguarations means virtual machines that support various versions of Squeak, Pharo and Cuis smalltalks.  IIABDFI.


Words to the wise:
In the old days, VM sources would be generated per-platform.  Now, all the source generated by the above generators and included in https://github.com/OpenSmalltalk/opensmalltalk-vm is separately versioned, and we use only CrossPlatformVMMaker to generate them.  All platform-specific variations, such as whether to use a struct to hold the interpreter's global variables, which plugins to include in a VM, etc, are included in the generated osurce.  All plugins are generated such that they can be compiled as either internal or external plugins.  What specific variations to use are therefore controlled by the platform-specific makefiles, and the decision is deferred from source generation time to VM build time.  The source is designed to be platform neutral.  This form is chosen carefully to allow VMs to be built from fully checked-in, and hence fully versioned source.


Old commentary:
This class builds a VM codebase from the in-image and on-file code.

The platforms file tree you need can be downloaded via cvs from http://squeak.Sourceforge.net. See also the swiki (http://minnow.cc.gatech.edu/squeak/2106) for instructions.

It is fairly configurable as to where directories live and can handle multiple platform's source trees at once. It's main purpose is to allow easy building of source trees with any combination of internal/external/unused plugins to suit your platform needs and capabilities. For example, the Acorn has no need of Sound or AsynchFile plugins since I haven't written any platform code for them. 

There is a simple UI tool for this 
	VMMakerTool openInWorld
will open a reasonably self explanatory tool with balloon help to explain all the fields - and a help window on top of that.

There are some simple workspace & inspector commands, allowing scripted building:
	VMMaker default initializeAllExternal generateEntire
for example will build sources for a system with all the plugins external whereas 
	VMMaker default initializeAllInternal generateEntire
would build all applicable plugins for internal compilation.
	(VMMaker forPlatform: 'Mac OS') initializeAllExternal generateEntire
would build a source tree for a Mac even on a Windows machine (err, ignoring for now the irritation of lineends).

	If you have a slightly more complex configuration you want to use, perhaps with Socket and Serial support external (because for your case they are rarely used and saving the space has some value) then you could try
		(VMMaker default initializeAllInternalBut: #(SocketPlugin SerialPlugin) generateEntire
	More complex still would be
		(VMMaker default initializeInternal: #(BitBltPlugin MiscPrimsPlugin FilePlugin) external: #(SocketPlugin ZipPlugin B2DPlugin)
which allows you to precisely list all the plugins to use.

WARNING If you miss out a plugin you need, it won't be there. This message is really best suited to use by a UI like VMMakerTool.

	To save a configuration for later use, you need to send #saveConfiguration to an active instance of VMMaker. Obviously you could simply use
		(VMMaker default initializeAllInternalBut: #(SocketPlugin SerialPlugin) saveConfiguration
but inspecting 
		VMMaker default
and altering the internalPlugins and externalPlugins or the boolean flags for inline or forBrowser followed by saving the configuration allows ultimate power for now. To load a saved configuration file, use #loadConfigurationFrom: aFilename whilst inspecting a VMMaker. The loaded state will completely override any pre-existing state, so take care.
	You can generate only parts of the source tree if you wish; as shown above #generateEntire will create the whole collection of internal and external plugins as well as the core VM. To create only  the external plugins use #generateExternalPlugins, or create a single  plugin with #generateExternalPlugin: name. To assemble the main VM including the internal plugins, use #generateMainVM. The interpreter 'interp.c' file is made with #generateInterpreterFile. You can generate a single internal plugin with #generateInternalPlugin: only if it has already been generated before; this interlocking is intended to make sure the named primitive table in the vm is correct.

There are some rules to observe in order to use this:-
- under the working directory (by default - you can configure it) you need a directory called 'platforms' (also configurable) with subdirectories named as the platform names returned by Smalltalk platformName (ie unix, RiscOS, Mac OS, etc - this isn't configurable). At the very least you need the one for your own platform and the pseudo-platform called 'Cross'. By adding a 'DirNames' entry for #machineType you can cross 'compile' for some other platform. Now all we need is a cross-compiler for the C code :-)
- under this directory you must have a simple structure of directories for each generated plugin that you support on the platform, plus 'vm'. In each directory you place any/all platform specific files (and subdirectories) for that plugin. In 'misc' you can place any miscellaneous files such as makefiles, resources etc. For example, for unix you have
	platforms/
		unix/
			plugins/
				AsynchFilePlugin /
					sqUnixAsynchfile.c
			vm/
				sqGnu.h
				Profile/
			misc/
				makefile.in
				util/
				
				...etc
Any plugins requiring platform files that you don't support shouldn't appear in the resulting code tree. If you try to include an unsupported plugin in the list to be made external, the VMMaker simply ignores it. However, if you include it in the list to be made internal you will get an error since that seems like a potentially serious source of confusion.

There are three lists of plugins maintained herein:-
1) the list of all known generatable plugins. We scan this list and compare with the supported plugins as indicated by the file tree.
2) the list of chosen internal plugins.
3) the list of chosen external plugins.
See initializeAllPlugins, initialiseAllExternal etc for fairly obvious usage.
There is also a short list of directory names in the class variable 'DirNames' that you can alter if needed.

Known problems:-
a) since Squeak has really poor filename handling, you can't simply change the directory names to '/foo/bar/myEvilCodeBase' and expect it to work. You fix file names and I'll fix VMMaker :-)
b) Squeak copying of a file loses the assorted permissions, filetype info and other useful bits. To workaround this problem, see the FileCopyPlugin, which provides the platform independent part of a simple access for the OS filecopy capability. So far there are functional plugins for unix, Mac and Acorn. DOS machines appear not to need one. This is less of a problem in practise now that unix, Acorn & Mac no longer copy files from /platforms to /src.

inline <Boolean> - is the generated code to be inlined or not
forBrowser <Boolean> - is this to be a build for in-Browser use? Only relevent to Macs
allPlugins <Collection> - all the known possible plugins
internalPlugins <Collection> - the plugins chosen to be generated for internal linking
externalPlugins <Collection> - the plugins intended to be external plugins
exportList <Collection> - a list of function names exported from plugins intended to be internal
platformName <String> - the name of the platform for which we are building a source tree. It is possible to do 'cross-compiles'
sourceDirName, platformRootDirName <String> - the name of the directory into which we write the generated sources and the name of the directory where we should find the platforms tree.!

VMMaker subclass: #CrossPlatformVMMaker
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Building'!

!CrossPlatformVMMaker commentStamp: '<historical>' prior: 0!
A generic VMMaker that creates a single source tree, leaving it up to external files to decide what platform to compile for and what plugins to include.!

!CrossPlatformVMMaker methodsFor: 'initialize' stamp: 'eem 10/30/2013 16:26'!
createCodeGenerator
	"Set up a CCodeGenerator for this VMMaker - A cross platform tree leaves it up to the makefiles to decide whether to use the global struct or not."
	^CCodeGeneratorGlobalStructure new
		vmMaker: self;
		structDefDefine: 'USE_GLOBAL_STRUCT';
		logger: logger;
		options: optionsDictionary;
		yourself! !

!CrossPlatformVMMaker methodsFor: 'target directories' stamp: 'eem 1/14/2010 12:48'!
externalPluginListName
	"Answer the filename for the list of external plugins. Since we're leaving it up to
	 platform makefiles to specify the actual named prims, just produce an example file."

	^'examplePlugins.ext'! !

!CrossPlatformVMMaker methodsFor: 'target directories' stamp: 'eem 1/14/2010 12:47'!
internalPluginListName
	"Answer the filename for the list of internal plugins. Since we're leaving it up to
	 platform makefiles to specify the actual named prims, just produce an example file."

	^'examplePlugins.int'! !

!CrossPlatformVMMaker methodsFor: 'target directories' stamp: 'eem 1/14/2010 12:02'!
internalPluginsDirectory
	"For the Cross-platfrorm all-in-one tree we simply use <root>/src/plugins, not <root>/src/intplugins"
	^self externalPluginsDirectory! !

!CrossPlatformVMMaker methodsFor: 'generate sources' stamp: 'eem 1/14/2010 12:44'!
interpreterExportsFilePath
	"Return the full path for the interpreter exports file.  Since we're leaving it up to
	 platform makefiles to specify the actual named prims, just produce an example file."
	^self coreVMDirectory fullNameFor: 'exampleSqNamedPrims.h'! !

!CrossPlatformVMMaker methodsFor: 'generate sources' stamp: 'eem 11/13/2014 16:13'!
validatePlugin: plName in: listOfPlugins
	"Check that the class for the plugin exists, and answer that class.
	 No longer check that if the plugin requires platform files that at least
	 one platform subdirectory exists, not producing the source is not helpful."
	| plugin |
	plName isString
		ifTrue: [(listOfPlugins includes: plName)
				ifTrue: [plugin := Smalltalk classNamed: plName]]
		ifFalse: [((plName isBehavior
						and: [plName inheritsFrom: InterpreterPlugin])
					and: [listOfPlugins includes: plName name])
				ifTrue: [plugin := plName]].
	plugin ifNil: [^ self couldNotFindPluginClass: plName].
	plugin shouldBeTranslated ifFalse:
		[self error: 'untranslateable plugin class ', plName].
	^plugin! !

VMMaker subclass: #RiscOSVMMaker
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Building'!

!RiscOSVMMaker commentStamp: '<historical>' prior: 0!
Special VMMaker class for Acorn RiscOS - add generation of a configuration file for each plugin, required by the dynamic loading libraries!

!RiscOSVMMaker class methodsFor: 'initialisation' stamp: 'eem 4/10/2013 16:01'!
forPlatform: ignored
	"override to keep a RiscOSVMMaker in charge instead of a CrossPlatformVMMaker"
	^self new initialize setPlatName: self name! !

!RiscOSVMMaker class methodsFor: 'configurations' stamp: 'KenD 1/22/2019 13:36'!
generateSqueakStackVM
	"RISC OS version; build needed plugins, make sure filename tweaking is used"
"RiscOSVMMaker generateSqueakStackVM"
	^self
		generate: StackInterpreter
		to: (FileDirectory default directoryNamed: 'stacksrc') fullName
		platformDir: (FileDirectory default directoryNamed: 'platforms') fullName
		excluding: #(AsynchFilePlugin BrokenPlugin CroquetPlugin FFIPlugin IA32ABIPlugin ImmX11Plugin InternetConfigPlugin JoystickTabletPlugin MIDIPlugin MacMenubarPlugin Mpeg3Plugin NewsqueakIA32ABIPlugin QuicktimePlugin SerialPlugin  TestOSAPlugin ThreadedARM32FFIPlugin ThreadedARMFFI64Plugin ThreadedFFIPlugin ThreadedIA32FFIPlugin ThreadedPPCBEFFIPlugin UUIDPlugin VMProfileMacSupportPlugin)! !

!RiscOSVMMaker methodsFor: 'initialize' stamp: 'eem 7/23/2014 23:29'!
createCodeGenerator
"set up a CCodeGenerator for this VMMaker - RiscOS uses the global struct and no local def of the structure because of the global register trickery"
	^CCodeGeneratorGlobalStructure new
		vmMaker: self;
		logger: logger;
		options: optionsDictionary;
		yourself! !

!RiscOSVMMaker methodsFor: 'generate sources' stamp: 'eem 5/17/2011 12:20'!
export: exportList forExternalPlugin: aPlugin
"it may be useful on certain platforms to do something with the export list of external plugins, just as the internal plugins' exports get added to the VM list. Default is to do nothing though."
"For RiscOS using the 'rink' external linker each plugin needs a 'dsc' file that looks like
id:SqueakSO
main_version:100
code_version:001

entries:
//
named_entries:
getModuleName
//
with all the exported names in the list. We also need a '/o' directory for the object files"

	"open a file called plugindir/pluginname.dsc and write into it"
	| f fd dfd |
	fd := self externalPluginsDirectoryFor: aPlugin.

	"If we get an error to do with opening the .dsc file, we need to raise an application error to suit"
	[(fd directoryExists: 'dsc') ifFalse:[fd createDirectory: 'dsc'].
	dfd := fd directoryNamed: 'dsc'.
	f := VMMaker forceNewFileNamed: (dfd fullNameFor: aPlugin moduleName)] on: FileStreamException do:[^self couldNotOpenFile: (dfd fullNameFor: aPlugin moduleName)].

	f nextPutAll: 'id:SqueakSO
main_version:100
code_version:001

entries:
//
named_entries:
'.
	exportList do:[:el|
		f nextPutAll: el.
		f cr].
	f nextPutAll: '//'; cr.
	f close.
	(fd directoryNamed: 'o') assureExistence
! !

!RiscOSVMMaker methodsFor: 'processing external files' stamp: 'tpr 4/9/2013 17:36'!
gnuifyInterpreterFile
"do nothing here"! !

!RiscOSVMMaker methodsFor: 'generate sources' stamp: 'tpr 3/12/2003 10:16'!
interpreterExportsFilePath
	"return the full path for the interpreter exports file"
	"RiscOS keeps the exports file in a 'h' subdirectory of coreVMDirectory"
	self coreVMDirectory assureExistenceOfPath: 'h'.
	^(self coreVMDirectory directoryNamed: 'h') fullNameFor: 'sqNamedPrims'! !

!RiscOSVMMaker methodsFor: 'target directories' stamp: 'eem 2/9/2009 12:35'!
interpreterFilename
	"what is the filename for the core interpreter. Default is interp.c"
	^super interpreterFilename copyWithoutAll: '.c'! !

!RiscOSVMMaker methodsFor: 'generate sources' stamp: 'tpr 3/17/2005 16:26'!
interpreterHeaderPath
	"return the full path for the interpreter header file"
	"RiscOS keeps the interp file in a 'h' subdirectory of coreVMDirectory"
	self coreVMDirectory assureExistenceOfPath: 'h'.
	^(self coreVMDirectory directoryNamed: 'h') fullNameFor: self interpreterFilename! !

!RiscOSVMMaker methodsFor: 'initialize' stamp: 'eem 6/7/2018 15:32'!
needsToRegenerateInterpreterFile
	"Check the timestamp for the relevant classes and then the timestamp for the main
	 source file (e.g. interp.c) if it already exists.  Answer if the file needs regenerating."

	| classes tStamp fstat |
	classes := self interpreterClass withAllSuperclasses copyUpTo: VMClass.
	self interpreterClass objectMemoryClass ifNotNil:
		[:objectMemoryClass|
		classes addAllLast: (objectMemoryClass withAllSuperclasses copyUpTo: VMClass)].
	classes copy do:
		[:class| classes addAllLast: class ancilliaryClasses].
	tStamp := classes inject: 0 into: [:tS :cl| tS max: cl timeStamp].

	"don't translate if the file is newer than my timeStamp"
	"RiscOS keeps the interp file in a 'c' subdirectory of coreVMDirectory"
	(self coreVMDirectory directoryExists: 'c') ifFalse:[^true].

	fstat := (self coreVMDirectory directoryNamed: 'c') entryAt: self interpreterFilename ifAbsent:[nil].
	fstat ifNotNil:[tStamp < fstat modificationTime ifTrue:[^false]].
	^true
! !

!RiscOSVMMaker methodsFor: 'processing external files' stamp: 'tpr 3/12/2003 10:17'!
processFilesForExternalPlugin: plugin 
	"After the plugin has created any files we need to move them around a little to suit RiscOS; any *.c file must be moved to a 'c' subdirectory, likwise any h file"
	| files fd |
	files := (fd := self externalPluginsDirectoryFor: plugin) fileNamesMatching:'*.c'.
	files notEmpty ifTrue:[fd assureExistenceOfPath: 'c'.
		files do:[:fn|
			self copyFileNamed: (fd fullNameFor: fn) to: ((fd directoryNamed:'c') fullNameFor:(fn allButLast: 2)).
			fd deleteFileNamed: fn]].

	files := (self externalPluginsDirectoryFor: plugin) fileNamesMatching:'*.h'.
	files notEmpty ifTrue:[fd assureExistenceOfPath: 'h'.
		files do:[:fn|
			self copyFileNamed: (fd fullNameFor: fn) to: ((fd directoryNamed:'h') fullNameFor:(fn allButLast: 2)).
			fd deleteFileNamed: fn]].
	super processFilesForExternalPlugin: plugin
! !

!RiscOSVMMaker methodsFor: 'processing external files' stamp: 'tpr 3/12/2003 10:18'!
processFilesForInternalPlugin: plugin 
	"After the plugin has created any files we need to move them around a little to suit RiscOS; any *.c file must be moved to a 'c' subdirectory, likwise any h file"
	| files fd |
	files := (fd := self internalPluginsDirectoryFor: plugin) fileNamesMatching:'*.c'.
	files notEmpty ifTrue:[fd assureExistenceOfPath: 'c'.
		files do:[:fn|
			self copyFileNamed: (fd localNameFor: fn) to: ((fd directoryNamed:'c') localNameFor:(fn allButLast: 2)).
			fd deleteFileNamed: fn]].

	files := (self internalPluginsDirectoryFor: plugin) fileNamesMatching:'*.h'.
	files notEmpty ifTrue:[fd assureExistenceOfPath: 'h'.
		files do:[:fn|
			self copyFileNamed: (fd localNameFor: fn) to: ((fd directoryNamed:'h') localNameFor:(fn allButLast: 2)).
			fd deleteFileNamed: fn]].
	super processFilesForInternalPlugin: plugin
! !

!RiscOSVMMaker methodsFor: 'generate sources' stamp: 'tpr 4/9/2013 16:09'!
sourceFilePathFor: sourceFileName
	"return the full path for the interpreter file"
	"RiscOS keeps the interp file in a 'c' subdirectory of coreVMDirectory"
	self coreVMDirectory assureExistenceOfPath: 'c'.
	^(self coreVMDirectory directoryNamed: 'c') fullNameFor: (sourceFileName allButLast: 2)! !

!RiscOSVMMaker methodsFor: 'exports' stamp: 'tpr 3/11/2003 14:11'!
storeExternalPluginList
	"RiscOS doesn't need this"! !

!RiscOSVMMaker methodsFor: 'exports' stamp: 'tpr 3/11/2003 14:11'!
storeInternalPluginList
	"RiscOS doesn't need this"! !

VMMaker subclass: #UnixVMMaker
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Building'!

!UnixVMMaker class methodsFor: 'initialisation' stamp: 'ikp 8/25/2003 00:13'!
isActiveVMMakerClassFor: platformName

	^platformName = 'unix'! !

!UnixVMMaker methodsFor: 'initialisation' stamp: 'eem 7/23/2014 23:30'!
createCodeGenerator

	^CCodeGeneratorGlobalStructure new
		vmMaker: self;
		logger: logger;
		options: optionsDictionary;
		yourself! !

!VMMaker class methodsFor: 'initialisation' stamp: ''!
activeVMMakerClassFor: platformName
	"Return the concrete VMMaker subclass for the platform on which we are currently running."

	VMMaker allSubclasses do: [:class |
		(class isActiveVMMakerClassFor: platformName) ifTrue: [^ class]].

	"no responding subclass; use VMMaker"
	^ VMMaker
! !

!VMMaker class methodsFor: 'confs-support' stamp: 'eem 12/6/2013 13:55'!
chooseCoInterpreterClassIfAbsent: ifAbsentBlock
	^Smalltalk classNamed:
		 ([:choices|
		   choices
			at: (UIManager default chooseFrom: choices)
			ifAbsent: ifAbsentBlock]
				value: #(CoInterpreter CoInterpreterMT))! !

!VMMaker class methodsFor: 'confs-support' stamp: 'eem 6/13/2014 00:13'!
configurationNameFromSelector: aSelector
	^(aSelector piecesCutWhere: [:a :b| a isLowercase and: [b isUppercase]]) allButFirst fold: [:a :b| a, ' ', b]! !

!VMMaker class methodsFor: 'accessing' stamp: 'tpr 12/1/2000 15:58'!
coreVMDirName
	^DirNames at: #coreVMDir! !

!VMMaker class methodsFor: 'initialisation' stamp: 'eem 1/14/2010 11:58'!
default
	"return a VMMaker initialised to build a default no-internal-plugins, no-external-plugins vm codebase"
	^self forPlatform: (true
						ifTrue: ['Cross'] 
						ifFalse: [SmalltalkImage current platformName])! !

!VMMaker class methodsFor: 'utilities' stamp: 'CyrilFerlicot 4/13/2018 13:59'!
executeDisplayingProgress: anActionsMap
	| barValue |
	barValue := 0.
	''
		displayProgressFrom: 0
		to: anActionsMap size
		during: [ :bar | 
			anActionsMap
				keysAndValuesDo: [ :label :action | 
					bar value: barValue.
					barValue := barValue + 1.
					ProgressNotification signal: '' extra: label.
					action value ] ]! !

!VMMaker class methodsFor: 'file utilities' stamp: 'eem 11/25/2020 11:31'!
fileNamed: aFilename
	"Always output files in unix lf format.
		A single format is friendlier to e.g. external version control systems.
		The Microsoft and old MacOS classic C compilers all accept lf format files."

	^(MultiByteFileStream fileNamed: aFilename)
		lineEndConvention: #lf;
		yourself! !

!VMMaker class methodsFor: 'initialisation' stamp: 'eem 5/22/2008 15:37'!
forConfigurationFile: aFileName
	| config fileStream vmMaker |

	fileStream := FileStream oldFileNamed: aFileName.
	config := ('#(' includes: fileStream peek) "for storeOn: format"
				ifTrue: [Object readFrom: fileStream]
				ifFalse: [fileStream fileInObjectAndCode].
	vmMaker := self forPlatform: (config at: 5).
	vmMaker loadConfiguration: config.
	^vmMaker! !

!VMMaker class methodsFor: 'instance creation' stamp: 'eem 1/28/2019 14:52'!
forPlatform: platformName
	"return a VMMaker initialised to build a default no-internal-plugins, no-external-plugins vm codebase"
	^(self activeVMMakerClassFor: platformName) new setPlatName: platformName! !

!VMMaker class methodsFor: 'file utilities' stamp: 'eem 5/17/2011 12:19'!
forceNewFileNamed: aFilename
	"Always output files in unix lf format.
		A single format is friendlier to e.g. external version control systems.
		The Microsoft and old MacOS classic C compilers all accept lf format files."

	^(MultiByteFileStream forceNewFileNamed: aFilename)
		lineEndConvention: #lf;
		yourself! !

!VMMaker class methodsFor: 'file utilities' stamp: 'eem 10/24/2018 11:21'!
fullNameForPath: aPathString
	^(Smalltalk classNamed: #FileReference)
		ifNotNil: [:cfs| aPathString asFileReference asAbsolute fullName]
		ifNil: [FileDirectory default pathFromURI: aPathString]! !

!VMMaker class methodsFor: 'utilities' stamp: 'eem 6/21/2011 08:49'!
generate: interpreterClass and: cogitClass to: srcDirName platformDir: platDirName excluding: exclusions
	"Generate the VM to the given target directory. Exclude plugins in pluginList.
	 Example:
		(VMMaker
			generate: CoInterpreter
			and: SimpleStackBasedCogit
			to: (FileDirectory default pathFromURI: '../src')
			platformDir: (FileDirectory default pathFromURI: '../platforms')
			excluding:#(
				BrokenPlugin DShowVideoDecoderPlugin IA32ABIPluginSimulator
				NewsqueakIABIPlugin NewsqueakIABIPluginAttic
				QuicktimePlugin QVideoCodecPlugin QwaqMediaPlugin
				SlangTestPlugin TestOSAPlugin ThreadedFFIPlugin 
				Win32OSProcessPlugin
			))"
	(self
		makerFor: interpreterClass
		and: cogitClass
		with: #()
		to: srcDirName
		platformDir: platDirName
		excluding: exclusions) generateEntire! !

!VMMaker class methodsFor: 'utilities' stamp: 'eem 6/21/2011 08:42'!
generate: interpreterClass and: cogitClass to: srcDirName platformDir: platDirName including: inclusions
	"Generate the VM to the given target directory. Include only plugins in pluginList.
	 Example:
		(VMMaker
			generate: CoInterpreter
			and: SimpleStackBasedCogit
			to: (FileDirectory default pathFromURI: '../src')
			platformDir: (FileDirectory default pathFromURI: '../platforms')
			including:#(	AsynchFilePlugin FloatArrayPlugin RePlugin B2DPlugin FloatMathPlugin SecurityPlugin
						BMPReadWriterPlugin IA32ABI SocketPlugin BitBltPlugin JPEGReadWriter2Plugin SurfacePlugin
						DSAPrims JPEGReaderPlugin UUIDPlugin DropPlugin LargeIntegers UnixOSProcessPlugin
						FileCopyPlugin Matrix2x3Plugin Win32OSProcessPlugin FilePlugin MiscPrimitivePlugin ZipPlugin))"
	^self generate: interpreterClass and: cogitClass with: #() to: srcDirName platformDir: platDirName including: inclusions! !

!VMMaker class methodsFor: 'utilities' stamp: 'eem 6/21/2011 08:42'!
generate: interpreterClass and: cogitClass with: optionsPairsArray to: srcDirName platformDir: platDirName including: inclusions
	"Generate the VM to the given target directory. Include only plugins in pluginList.
	 Example:
		(VMMaker
			generate: CoInterpreter
			and: SimpleStackBasedCogit
			to: (FileDirectory default pathFromURI: '../src')
			platformDir: (FileDirectory default pathFromURI: '../platforms')
			including:#(	AsynchFilePlugin FloatArrayPlugin RePlugin B2DPlugin FloatMathPlugin SecurityPlugin
						BMPReadWriterPlugin IA32ABI SocketPlugin BitBltPlugin JPEGReadWriter2Plugin SurfacePlugin
						DSAPrims JPEGReaderPlugin UUIDPlugin DropPlugin LargeIntegers UnixOSProcessPlugin
						FileCopyPlugin Matrix2x3Plugin Win32OSProcessPlugin FilePlugin MiscPrimitivePlugin ZipPlugin))"
	(self
		makerFor: interpreterClass
		and: cogitClass
		with: optionsPairsArray
		to: srcDirName
		platformDir: platDirName
		including: inclusions) generateEntire! !

!VMMaker class methodsFor: 'utilities' stamp: 'eem 6/21/2011 08:50'!
generate: interpreterClass to: srcDirName platformDir: platDirName excluding: exclusions
	"Generate the VM to the given target directory. Exclude plugins in pluginList.
	 Example:
		(VMMaker
			generate: CoInterpreter
			to: (FileDirectory default pathFromURI: '../src')
			platformDir: (FileDirectory default pathFromURI: '../platforms')
			excluding:#(
				BrokenPlugin DShowVideoDecoderPlugin IA32ABIPluginSimulator
				NewsqueakIABIPlugin NewsqueakIABIPluginAttic
				QuicktimePlugin QVideoCodecPlugin QwaqMediaPlugin
				SlangTestPlugin TestOSAPlugin ThreadedFFIPlugin 
				Win32OSProcessPlugin
			))."
	(self
		makerFor: interpreterClass
		and: nil
		with: #()
		to: srcDirName
		platformDir: platDirName
		excluding: exclusions) generateEntire! !

!VMMaker class methodsFor: 'utilities' stamp: 'eem 6/21/2011 08:45'!
generate: interpreterClass to: srcDirName platformDir: platDirName including: inclusions
	"Generate the VM to the given target directory. Include only plugins in pluginList.
	 Example:
		(VMMaker
			generate: NewspeakInterpreter
			to: (FileDirectory default pathFromURI: 'cogvm/newspeaksrc')
			platformDir: (FileDirectory default pathFromURI: 'cogvm/platforms')
			including:#(	AsynchFilePlugin FloatArrayPlugin RePlugin B2DPlugin FloatMathPlugin SecurityPlugin
						BMPReadWriterPlugin IA32ABI SocketPlugin BitBltPlugin JPEGReadWriter2Plugin SurfacePlugin
						DSAPrims JPEGReaderPlugin UUIDPlugin DropPlugin LargeIntegers UnixOSProcessPlugin
						FileCopyPlugin Matrix2x3Plugin Win32OSProcessPlugin FilePlugin MiscPrimitivePlugin ZipPlugin))."
	self generate: interpreterClass with: #() to: srcDirName platformDir: platDirName including: inclusions! !

!VMMaker class methodsFor: 'utilities' stamp: 'eem 10/11/2013 15:23'!
generate: interpreterClass with: optionsPairsArray to: srcDirName platformDir: platDirName excluding: exclusions
	"Generate the VM to the given target directory. Include only plugins in pluginList.
	 Example:
		(VMMaker
			generate: StackInterpreter
			with: #(ObjectMemory Spur32BitMemoryManager)
			to: (FileDirectory default directoryNamed: 'oscogvm/spurstacksrc') fullName
			platformDir: (FileDirectory default directoryNamed: 'oscogvm/platforms') fullName
			excluding: (InterpreterPlugin withAllSubclasses collect: [:ea| ea name]))."
	(self makerFor: interpreterClass and: nil with: optionsPairsArray to: srcDirName platformDir: platDirName excluding: exclusions) generateEntire! !

!VMMaker class methodsFor: 'utilities' stamp: 'eem 6/21/2011 08:45'!
generate: interpreterClass with: optionsPairsArray to: srcDirName platformDir: platDirName including: inclusions
	"Generate the VM to the given target directory. Include only plugins in pluginList.
	 Example:
		(VMMaker
			generate: NewspeakInterpreter
			to: (FileDirectory default pathFromURI: 'cogvm/newspeaksrc')
			with: #(IMMUTABILITY true)
			platformDir: (FileDirectory default pathFromURI: 'cogvm/platforms')
			including:#(	AsynchFilePlugin FloatArrayPlugin RePlugin B2DPlugin FloatMathPlugin SecurityPlugin
						BMPReadWriterPlugin IA32ABI SocketPlugin BitBltPlugin JPEGReadWriter2Plugin SurfacePlugin
						DSAPrims JPEGReaderPlugin UUIDPlugin DropPlugin LargeIntegers UnixOSProcessPlugin
						FileCopyPlugin Matrix2x3Plugin Win32OSProcessPlugin FilePlugin MiscPrimitivePlugin ZipPlugin))."
	(self makerFor: interpreterClass and: nil with: optionsPairsArray to: srcDirName platformDir: platDirName including: inclusions) generateEntire! !

!VMMaker class methodsFor: 'configurations' stamp: 'eem 4/13/2021 18:15'!
generateAllCogConfigurationsUnderVersionControl
	self
		executeDisplayingProgress:
			(OrderedDictionary new
				add: 'Generate squeak cog' -> [ self generateSqueakCogVM ];
				add: 'Generate squeak spur cog 32bits' -> [ self generateSqueakSpurCogVM ];
				add: 'Generate squeak spur cog 64bits' -> [ self generateSqueakSpurCog64VM ];
				add: 'Generate squeak spur cog sista 32bits' -> [ self generateSqueakSpurCogSistaVM ];
				add: 'Generate squeak spur cog sista 64bits' -> [ self generateSqueakSpurCogSista64VM ];
				yourself)! !

!VMMaker class methodsFor: 'configurations' stamp: 'eem 4/13/2021 18:14'!
generateAllConfigurationsUnderVersionControl
	self
		executeDisplayingProgress:
			(OrderedDictionary
				with: 'Generate all squeak configurations under VCS' -> [ self generateAllSqueakConfigurationsUnderVersionControl ]
				with: 'Generate all spur lowcode configurations' -> [ self generateAllSpurLowcodeConfigurations ]
				with: 'Generate VM plugins' -> [ self generateVMPlugins ]
				with: 'Generate utilities' -> [ self generateSpur32LeakChecker; generateSpur64LeakChecker; generateUtilityPrograms ])! !

!VMMaker class methodsFor: 'configurations' stamp: 'eem 11/25/2020 19:38'!
generateAllMTConfigurationsUnderVersionControl
	self executeDisplayingProgress:
		(OrderedDictionary new
			add: 'Generate squeak spur cog multi-threaded 32bits' -> [ self generateSqueakSpurCogMTVM ];
			add: 'Generate squeak spur cog multi-threaded 64bits' -> [ self generateSqueakSpurCog64MTVM ];
			yourself)! !

!VMMaker class methodsFor: 'configurations' stamp: 'eem 4/13/2021 18:14'!
generateAllNewspeakConfigurations
	self
		executeDisplayingProgress:
			(OrderedDictionary
				with: 'Generate newspeak spur stack 32bits' -> [ self generateNewspeakSpurStackVM ]
				with: 'Generate newspeak spur stack 64bits' -> [ self generateNewspeakSpurStack64VM ]
				with: 'Generate newspeak spur cog 32bits' -> [ self generateNewspeakSpurCogVM ]
				with: 'Generate newspeak spur cog 64bits' -> [ self generateNewspeakSpurCog64VM ])! !

!VMMaker class methodsFor: 'configurations' stamp: 'eem 11/10/2020 20:59'!
generateAllNonSpurConfigurationsUnderVersionControl
	self
		executeDisplayingProgress:
			(OrderedDictionary 
				with: 'Generate squeak cog' -> [ self generateSqueakCogVM ]
				with: 'Generate squeak stack' -> [ self generateSqueakStackVM ])! !

!VMMaker class methodsFor: 'configurations' stamp: 'CyrilFerlicot 4/14/2018 00:23'!
generateAllSistaConfigurationsUnderVersionControl
	self
		executeDisplayingProgress:
			(OrderedDictionary 
				with: 'Generate squeak spur cog sista 32bits' -> [ self generateSqueakSpurCogSistaVM ]
				with: 'Generate squeak spur cog sista 64bits' -> [ self generateSqueakSpurCogSista64VM ])! !

!VMMaker class methodsFor: 'configurations' stamp: 'eem 4/13/2021 18:15'!
generateAllSpurConfigurations
	self
		executeDisplayingProgress:
			(OrderedDictionary new
				add: 'Generate squeak spur cog 32bits' -> [ self generateSqueakSpurCogVM ];
				add: 'Generate squeak spur cog 64bits' -> [ self generateSqueakSpurCog64VM ];
				add: 'Generate squeak spur cog sista 32bits' -> [ self generateSqueakSpurCogSistaVM ];
				add: 'Generate squeak spur cog sista 64bits' -> [ self generateSqueakSpurCogSista64VM ];
				add: 'Generate squeak spur stack 32bits' -> [ self generateSqueakSpurStackVM ];
				add: 'Generate squeak spur stack 64bits' -> [ self generateSqueakSpurStack64VM ];				
				add: 'Generate all spur lowcode configurations' -> [ self generateAllSpurLowcodeConfigurations ];
				add: 'Generate spur leak checkers' -> [ self generateSpur32LeakChecker; generateSpur64LeakChecker ];
				yourself)
		! !

!VMMaker class methodsFor: 'configurations' stamp: 'eem 4/13/2021 18:13'!
generateAllSpurConfigurationsUnderVersionControl
	self
		executeDisplayingProgress:
			(OrderedDictionary new
				add: 'Generate squeak spur stack 32bits' -> [ self generateSqueakSpurStackVM ];
				add: 'Generate squeak spur stack 64bits' -> [ self generateSqueakSpurStack64VM ];
				add: 'Generate squeak spur cog 32bits' -> [ self generateSqueakSpurCogVM ];
				add: 'Generate squeak spur cog 64bits' -> [ self generateSqueakSpurCog64VM ];
				add: 'Generate squeak spur cog multi-threaded 32bits' -> [ self generateSqueakSpurCogMTVM ];
				add: 'Generate squeak spur cog multi-threaded 64bits' -> [ self generateSqueakSpurCog64MTVM ];
				add: 'Generate squeak spur cog sista 32bits' -> [ self generateSqueakSpurCogSistaVM ];
				add: 'Generate squeak spur cog sista 64bits' -> [ self generateSqueakSpurCogSista64VM ];	
				add: 'Generate spur leak checkers' -> [ self generateSpur32LeakChecker; generateSpur64LeakChecker ];
				yourself)
! !

!VMMaker class methodsFor: 'configurations' stamp: 'CyrilFerlicot 4/14/2018 00:24'!
generateAllSpurLowcodeConfigurations
	self
		executeDisplayingProgress:
			(OrderedDictionary
				with: 'Generate squeak spur lowcode cog 32bits' -> [ self generateSqueakSpurLowcodeCogVM ]
				with: 'Generate squeak spur lowcode stack 32bits' -> [ self generateSqueakSpurLowcodeStackVM ]
				with: 'Generate squeak spur lowcode cog 64bits' -> [ self generateSqueakSpurLowcodeCog64VM ]
				with: 'Generate squeak spur lowcode stack 64bits' -> [ self generateSqueakSpurLowcodeStack64VM ])! !

!VMMaker class methodsFor: 'configurations' stamp: 'eem 11/10/2020 20:59'!
generateAllSqueakConfigurationsUnderVersionControl
	self
		executeDisplayingProgress:
			(OrderedDictionary new
				add: 'Generate squeak stack' -> [ self generateSqueakStackVM ];
				add: 'Generate squeak cog' -> [ self generateSqueakCogVM ];
				add: 'Generate squeak spur stack 32bits' -> [ self generateSqueakSpurStackVM ];
				add: 'Generate squeak spur stack 64bits' -> [ self generateSqueakSpurStack64VM ];
				add: 'Generate squeak spur cog 32bits' -> [ self generateSqueakSpurCogVM ];
				add: 'Generate squeak spur cog 64bits' -> [ self generateSqueakSpurCog64VM ];
				add: 'Generate squeak spur cog sista 32bits' -> [ self generateSqueakSpurCogSistaVM ];
				add: 'Generate squeak spur cog sista 64bits' -> [ self generateSqueakSpurCogSista64VM ];
				add: 'Generate squeak spur cog multi-threaded 32bits' -> [ self generateSqueakSpurCogMTVM ];
				add: 'Generate squeak spur cog multi-threaded 64bits' -> [ self generateSqueakSpurCog64MTVM ];
				yourself)! !

!VMMaker class methodsFor: 'configurations' stamp: 'CyrilFerlicot 4/14/2018 00:31'!
generateAllStackConfigurationsUnderVersionControl
	self
		executeDisplayingProgress:
			(OrderedDictionary new
				add: 'Generate newspeak spur stack 32bits' -> [ self generateNewspeakSpurStackVM ];
				add: 'Generate newspeak spur stack 64bits' -> [ self generateNewspeakSpurStack64VM ];
				add: 'Generate squeak stack' -> [ self generateSqueakStackVM ];				
				add: 'Generate squeak spur stack 32bits' -> [ self generateSqueakSpurStackVM ];
				add: 'Generate squeak spur stack 64bits' -> [ self generateSqueakSpurStack64VM ];
				yourself)! !

!VMMaker class methodsFor: 'configurations' stamp: 'eem 6/13/2014 00:14'!
generateConfiguration
	"VMMaker generateConfiguration"
	| configCategoryName selectors |
	configCategoryName := self class whichCategoryIncludesSelector: thisContext selector.
	selectors := Set new.
	self class organization categories do:
		[:cat|
		(cat endsWith: configCategoryName) ifTrue:
			[selectors addAll: (self class organization listAtCategoryNamed: cat)]].
	selectors remove: thisContext selector.
	selectors := selectors asArray sort.
	(UIManager default
			chooseFrom: (selectors collect: [:sel| self configurationNameFromSelector: sel])
			values: selectors) ifNotNil:
		[:choice|
		self perform: choice]! !

!VMMaker class methodsFor: 'configurations' stamp: 'eem 4/21/2015 22:49'!
generateEitherSqueakCogVM
	| coInterpreterClass |
	coInterpreterClass := self chooseCoInterpreterClassIfAbsent: [^self].
	^self generateSqueakCogVMWithInterpreterClass: coInterpreterClass
		  options: ((coInterpreterClass includesBehavior: CoInterpreterMT)
						ifTrue: [#(COGMTVM true)]
						ifFalse: [#()])! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 17:58'!
generateNewspeakSpurCog64VM
	"No primitives since we can use those for the Cog VM"
	^VMMaker
		generate: CoInterpreter
		and: StackToRegisterMappingCogit
		with: #(	ObjectMemory Spur64BitCoMemoryManager
				MULTIPLEBYTECODESETS true
				NewspeakVM true)
		to: self sourceTree, '/src/spur64.cog.newspeak'
		platformDir: self sourceTree, '/platforms'
		including:#()
! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 17:59'!
generateNewspeakSpurCogVM
	"No primitives since we can use those for the Cog VM"
	^VMMaker
		generate: CoInterpreter
		and: StackToRegisterMappingCogit
		with: #(	ObjectMemory Spur32BitCoMemoryManager
				MULTIPLEBYTECODESETS true
				NewspeakVM true)
		to: self sourceTree, '/src/spur32.cog.newspeak'
		platformDir: self sourceTree, '/platforms'
		including:#()
! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 17:59'!
generateNewspeakSpurStack64VM
	"No primitives since we can use those for the Cog VM"
	^VMMaker
		generate: StackInterpreter
		with: #( ObjectMemory Spur64BitMemoryManager
				MULTIPLEBYTECODESETS true
				NewspeakVM true
				FailImbalancedPrimitives false)
		to: self sourceTree, '/src/spur64.stack.newspeak'
		platformDir: self sourceTree, '/platforms'
		including: #()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 17:59'!
generateNewspeakSpurStackVM
	"No primitives since we can use those for the Cog VM"
	^VMMaker
		generate: StackInterpreter
		with: #(	ObjectMemory Spur32BitMemoryManager
				MULTIPLEBYTECODESETS true
				NewspeakVM true
				FailImbalancedPrimitives false)
		to: self sourceTree, '/src/spur32.stack.newspeak'
		platformDir: self sourceTree, '/platforms'
		including: #()! !

!VMMaker class methodsFor: 'utilities' stamp: 'eem 1/26/2016 14:34'!
generatePluginsTo: srcDirName options: optionsPairsArray platformDir: platDirName including: inclusions
	"Generate the plugins to the given target directory. Include only plugins in pluginList."
	(self
		makerFor: StackInterpreter
		and: nil
		with: optionsPairsArray
		to: srcDirName
		platformDir: platDirName
		including: inclusions) generatePlugins! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 17:59'!
generateSpur32LeakChecker
	^VMMaker
		generate: ImageLeakChecker
		with: #(ObjectMemory Spur32BitMemoryManager
				MULTIPLEBYTECODESETS true)
		to: self sourceTree, '/src/spur32.stack'
		platformDir: self sourceTree, '/platforms'
		including: #()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 17:59'!
generateSpur64LeakChecker
	^VMMaker
		generate: ImageLeakChecker
		with: #(ObjectMemory Spur64BitMemoryManager
				MULTIPLEBYTECODESETS true)
		to: self sourceTree, '/src/spur64.stack'
		platformDir: self sourceTree, '/platforms'
		including: #()! !

!VMMaker class methodsFor: 'configurations' stamp: 'eem 5/1/2014 09:11'!
generateSqueakCogVM
	^self generateSqueakCogVMWithInterpreterClass: CoInterpreter
		  options: #()! !

!VMMaker class methodsFor: 'confs-support' stamp: 'mt 8/25/2021 18:02'!
generateSqueakCogVMWithInterpreterClass: interpreterClass options: optionsArray
	^VMMaker
		generate: interpreterClass
		and: StackToRegisterMappingCogit
		with: optionsArray
		to: self sourceTree, '/src/v3.cog'
		platformDir: self sourceTree, '/platforms'
		including:#()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 18:00'!
generateSqueakSpurCog64MTVM
	"No primitives since we can use those for the Cog VM"
	^VMMaker
		generate: CoInterpreterMT
		and: StackToRegisterMappingCogit
		with: #(COGMTVM true
				ObjectMemory Spur64BitCoMemoryManager
				MULTIPLEBYTECODESETS true
				bytecodeTableInitializer initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid)
		to: self sourceTree, '/src/spur64.cog'
		platformDir: self sourceTree, '/platforms'
		including:#()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 18:00'!
generateSqueakSpurCog64VM
	"No primitives since we can use those for the Cog VM"
	^VMMaker
		generate: CoInterpreter
		and: StackToRegisterMappingCogit
		with: #(ObjectMemory Spur64BitCoMemoryManager
				MULTIPLEBYTECODESETS true
				bytecodeTableInitializer initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid)
		to: self sourceTree, '/src/spur64.cog'
		platformDir: self sourceTree, '/platforms'
		including:#()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 18:00'!
generateSqueakSpurCogMTVM
	"No primitives since we can use those for the Cog VM"
	^VMMaker
		generate: CoInterpreterMT
		and: StackToRegisterMappingCogit
		with: #(COGMTVM true
				ObjectMemory Spur32BitCoMemoryManager
				MULTIPLEBYTECODESETS true
				bytecodeTableInitializer initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid)
		to: self sourceTree, '/src/spur32.cog'
		platformDir: self sourceTree, '/platforms'
		including:#()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 18:00'!
generateSqueakSpurCogSista64VM
	^VMMaker
		generate: CoInterpreter
		and: SistaCogit
		with: #(	SistaVM true
				ObjectMemory Spur64BitCoMemoryManager
				MULTIPLEBYTECODESETS true
				bytecodeTableInitializer initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid)
		to: self sourceTree, '/src/spur64.sista'
		platformDir: self sourceTree, '/platforms'
		including:#()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 18:00'!
generateSqueakSpurCogSistaVM
	"No primitives since we can use those for the Cog VM"
	^VMMaker
		generate: CoInterpreter
		and: SistaCogit
		with: #(	SistaVM true
				ObjectMemory Spur32BitCoMemoryManager
				MULTIPLEBYTECODESETS true
				bytecodeTableInitializer initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid)
		to: self sourceTree, '/src/spur32.sista'
		platformDir: self sourceTree, '/platforms'
		including:#()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 18:00'!
generateSqueakSpurCogVM
	"No primitives since we can use those for the Cog VM"
	^VMMaker
		generate: CoInterpreter
		and: StackToRegisterMappingCogit
		with: #(ObjectMemory Spur32BitCoMemoryManager
				MULTIPLEBYTECODESETS true
				bytecodeTableInitializer initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid)
		to: self sourceTree, '/src/spur32.cog'
		platformDir: self sourceTree, '/platforms'
		including:#()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 18:01'!
generateSqueakSpurLowcodeCog64VM
	"No primitives since we can use those for the Cog VM"
	^VMMaker
		generate: CoInterpreter
		and: StackToRegisterMappingCogit
		with: #(ObjectMemory Spur64BitCoMemoryManager
				SistaVM true
				LowcodeVM true
				MULTIPLEBYTECODESETS true
				bytecodeTableInitializer initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid)
		to: self sourceTree, '/src/spur64.cog.lowcode'
		platformDir: self sourceTree, '/platforms'
		including:#()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 18:01'!
generateSqueakSpurLowcodeCogVM
	"No primitives since we can use those for the Cog VM"
	^VMMaker
		generate: CoInterpreter
		and: StackToRegisterMappingCogit
		with: #(ObjectMemory Spur32BitCoMemoryManager
				SistaVM true
				LowcodeVM true
				MULTIPLEBYTECODESETS true
				bytecodeTableInitializer initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid)
		to: self sourceTree, '/src/spur32.cog.lowcode'
		platformDir: self sourceTree, '/platforms'
		including:#()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 18:01'!
generateSqueakSpurLowcodeStack64VM
	"No primitives since we can use those from the Cog VM"
	^VMMaker
		generate: StackInterpreter
		with: #(ObjectMemory Spur64BitMemoryManager
				SistaVM true
				LowcodeVM true
				FailImbalancedPrimitives false
				MULTIPLEBYTECODESETS true
				bytecodeTableInitializer initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid)
		to: self sourceTree, '/src/spur64.stack.lowcode'
		platformDir: self sourceTree, '/platforms'
		including: #()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 18:01'!
generateSqueakSpurLowcodeStackVM
	"No primitives since we can use those from the Cog VM"
	^VMMaker
		generate: StackInterpreter
		with: #(ObjectMemory Spur32BitMemoryManager
				SistaVM true
				LowcodeVM true
				FailImbalancedPrimitives false
				MULTIPLEBYTECODESETS true
				bytecodeTableInitializer initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid)
		to: self sourceTree, '/src/spur32.stack.lowcode'
		platformDir: self sourceTree, '/platforms'
		including: #()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 18:01'!
generateSqueakSpurRegisterCogVM
	"No primitives since we can use those for the Cog VM"
	^VMMaker
		generate: CoInterpreter
		and: RegisterAllocatingCogit
		with: #(ObjectMemory Spur32BitCoMemoryManager)
		to: self sourceTree, '/src/spur32.reg.spur'
		platformDir: self sourceTree, '/platforms'
		including:#()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 18:01'!
generateSqueakSpurStack64VM
	"No primitives since we can use those from the Cog VM"
	^VMMaker
		generate: StackInterpreter
		with: #(ObjectMemory Spur64BitMemoryManager
				FailImbalancedPrimitives false
				MULTIPLEBYTECODESETS true
				bytecodeTableInitializer initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid)
		to: self sourceTree, '/src/spur64.stack'
		platformDir: self sourceTree, '/platforms'
		including: #()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 18:01'!
generateSqueakSpurStackVM
	"No primitives since we can use those from the Cog VM"
	^VMMaker
		generate: StackInterpreter
		with: #(ObjectMemory Spur32BitMemoryManager
				FailImbalancedPrimitives false
				MULTIPLEBYTECODESETS true
				bytecodeTableInitializer initializeBytecodeTableForSqueakV3PlusClosuresSistaV1Hybrid)
		to: self sourceTree, '/src/spur32.stack'
		platformDir: self sourceTree, '/platforms'
		including: #()! !

!VMMaker class methodsFor: 'configurations' stamp: 'mt 8/25/2021 18:02'!
generateSqueakStackVM
	"No primitives since we can use those from the Cog VM"
	^VMMaker
		generate: StackInterpreter
		with: #(FailImbalancedPrimitives false)
		to: self sourceTree, '/src/v3.stack'
		platformDir: self sourceTree, '/platforms'
		including: #()! !

!VMMaker class methodsFor: 'configurations' stamp: 'eem 7/28/2020 13:42'!
generateUtilityPrograms
	^(self
		makerFor: nil
		and: nil
		with: #()
		to: self sourceTree, '/src'
		platformDir: self sourceTree, '/platforms')
			generateUtilityPrograms! !

!VMMaker class methodsFor: 'configurations' stamp: 'eem 9/25/2020 14:15'!
generateVMPlugins
	^VMMaker
		generatePluginsTo: self sourceTree, '/src'
		options: #()
		platformDir: self sourceTree, '/platforms'
		including:#(ADPCMCodecPlugin AsynchFilePlugin
					BalloonEnginePlugin B3DAcceleratorPlugin B3DEnginePlugin BMPReadWriterPlugin BitBltSimulation
					BochsIA32Plugin BochsX64Plugin GdbARMv6Plugin GdbARMv8Plugin
					CameraPlugin ClipboardExtendedPlugin CroquetPlugin DeflatePlugin DropPlugin
					"Cryptography Plugins:" DESPlugin DSAPlugin MD5Plugin SHA2Plugin
					"FT2Plugin" FFTPlugin FileCopyPlugin FilePlugin FileAttributesPlugin Float64ArrayPlugin FloatArrayPlugin FloatMathPlugin
					GeniePlugin HostWindowPlugin IA32ABIPlugin ImmX11Plugin InternetConfigPlugin
					JPEGReadWriter2Plugin JPEGReaderPlugin JoystickTabletPlugin KlattSynthesizerPlugin
					LargeIntegersPlugin LocalePlugin MIDIPlugin MacMenubarPlugin Matrix2x3Plugin
					MiscPrimitivePlugin Mpeg3Plugin QuicktimePlugin RePlugin
					ScratchPlugin SecurityPlugin SerialPlugin SocketPlugin
					SoundCodecPlugin SoundGenerationPlugin SoundPlugin SqueakSSLPlugin StarSqueakPlugin
					ThreadedFFIPlugin ThreadedARM32FFIPlugin ThreadedARM64FFIPlugin ThreadedIA32FFIPlugin
					ThreadedX64SysVFFIPlugin ThreadedX64Win64FFIPlugin
					UnicodePlugin UnixAioPlugin UUIDPlugin UnixOSProcessPlugin
					Win32OSProcessPlugin VMProfileLinuxSupportPlugin VMProfileMacSupportPlugin WeDoPlugin
					XDisplayControlPlugin)! !

!VMMaker class methodsFor: 'version testing' stamp: 'EstebanLorenzano 12/21/2010 20:50'!
headerNotice
"return a string to be put at the front of generated code files"
	^String streamContents:[:strm|
		strm nextPutAll: 'Automatically generated from Squeak on '.
		strm nextPutAll: Time dateAndTimeNow printString.
		strm cr.
		strm nextPutAll: 'by VMMaker '.
		strm nextPutAll: self  versionString.
		strm cr]! !

!VMMaker class methodsFor: 'initialisation' stamp: 'mt 7/13/2021 16:43'!
initialize
	"VMMaker initialize"
	DirNames ifNil: [DirNames := Dictionary new].
	DirNames
		at: #coreVMDir put: ''; "vm flavors have their own dirs below #sourceDir now. See #generate* methods."
		at: #platformsDir put: 'platforms';
		at: #pluginsDir put: 'plugins';
		at: #sourceDir put: 'src'.

	"Try and decide where the Cog source tree is.  Two configurations are likely.
	 One is that the VMMaker image is running in the image directory in the
	 source tree and hence everything will be at '..'.
	 Another is where the source tree is at the same level as the VMMaker image,
	 in which case it is likely called oscogvm or Cog."
	#('../platforms' 'oscogvm/platforms' 'Cog/platforms' '../oscogvm/platforms')
		with: #('..' 'oscogvm' 'Cog' '../oscogvm')
		do: ((Smalltalk classNamed: #FileLocator)
				ifNotNil:
					[:classFileLocator|
					 [:dir :path|
					  (classFileLocator cwd / dir) isDirectory ifTrue:
						[DirNames at: #sourceTree put: path.
						 ^self]]]
				ifNil:
					[[:dir :path|
					  (FileDirectory default directoryExists: dir) ifTrue:
						[DirNames at: #sourceTree put: path.
						 ^self]]])! !

!VMMaker class methodsFor: 'initialisation' stamp: 'tpr 2/4/2002 19:18'!
isActiveVMMakerClassFor: platformName 
	"Does this class claim to be that properly active subclass of VMMaker for 
	this platform? Subclasses are welcome to override this default"
	^ platformName , '*' match: self name! !

!VMMaker class methodsFor: 'instance creation' stamp: 'eem 3/6/2019 11:20'!
makerFor: interpreterClass and: cogitClassOrNil with: optionsPairsArrayArg to: srcDirName platformDir: platDirName
	"Initialize a VMMaker to generate the VM to the given target directory."
	| optionsPairsArray |
	optionsPairsArray := (cogitClassOrNil isNil or: [optionsPairsArrayArg includes: #Cogit])
							ifTrue: [optionsPairsArrayArg]
							ifFalse: [optionsPairsArrayArg, { #Cogit. cogitClassOrNil name }].
	VMClass initializeForPlugins: optionsPairsArray.
	^(self forPlatform: 'Cross')
		sourceDirectoryName: (self fullNameForPath: srcDirName);
		platformRootDirectoryName: (self fullNameForPath: platDirName);
		options: optionsPairsArray;
		interpreterClass: interpreterClass;
		yourself! !

!VMMaker class methodsFor: 'utilities' stamp: 'eem 6/21/2011 08:49'!
makerFor: interpreterClass and: cogitClassOrNil with: optionsPairsArray to: srcDirName platformDir: platDirName excluding: exclusions
	"Initialize a VMMaker to generate the VM to the given target directory. Exclude plugins in pluginList.
	Example:
		(VMMaker
			makerFor: CoInterpreter
			and: SimpleStackBasedCogit
			to: (FileDirectory default pathFromURI: '../src')
			platformDir: (FileDirectory default pathFromURI: '../platforms')
			excluding:#(
				BrokenPlugin DShowVideoDecoderPlugin IA32ABIPluginSimulator
				NewsqueakIABIPlugin NewsqueakIABIPluginAttic
				QuicktimePlugin QVideoCodecPlugin QwaqMediaPlugin
				SlangTestPlugin TestOSAPlugin ThreadedFFIPlugin 
				Win32OSProcessPlugin))
			generateInterpreterFile.
	"
	| maker |
	maker := self makerFor: interpreterClass and: cogitClassOrNil with: optionsPairsArray to: srcDirName platformDir: platDirName.
	maker externalModules addAll: 
		(maker availablePlugins reject:[:plugin| exclusions includes: plugin]).
	^maker
! !

!VMMaker class methodsFor: 'utilities' stamp: 'eem 6/21/2011 08:49'!
makerFor: interpreterClass and: cogitClassOrNil with: optionsPairsArray to: srcDirName platformDir: platDirName including: inclusions
	"Initialize a VMMaker to generate the VM to the given target directory. Include plugins in pluginList.
	Example:
		(VMMaker
			generate: NewspeakInterpreter
			to: (FileDirectory default pathFromURI: 'cogvm/newspeaksrc')
			platformDir: (FileDirectory default pathFromURI: 'cogvm/platforms')
			including:#(	AsynchFilePlugin FloatArrayPlugin RePlugin B2DPlugin FloatMathPlugin SecurityPlugin
						BMPReadWriterPlugin IA32ABI SocketPlugin BitBltPlugin JPEGReadWriter2Plugin SurfacePlugin
						DSAPrims JPEGReaderPlugin UUIDPlugin DropPlugin LargeIntegers UnixOSProcessPlugin
						FileCopyPlugin Matrix2x3Plugin Win32OSProcessPlugin FilePlugin MiscPrimitivePlugin ZipPlugin))"
	| maker |
	maker := self makerFor: interpreterClass and: cogitClassOrNil with: optionsPairsArray to: srcDirName platformDir: platDirName.
	maker externalModules addAll: inclusions.
	^maker
! !

!VMMaker class methodsFor: 'file utilities' stamp: 'eem 4/21/2015 23:33'!
oldFileNamed: aFilename
	"Always output files in unix lf format.
		A single format is friendlier to e.g. external version control systems.
		The Microsoft and old MacOS classic C compilers all accept lf format files."

	^(MultiByteFileStream oldFileNamed: aFilename)
		lineEndConvention: #lf;
		yourself! !

!VMMaker class methodsFor: 'accessing' stamp: 'tpr 12/1/2000 16:03'!
platformsDirName
	^DirNames at: #platformsDir! !

!VMMaker class methodsFor: 'accessing' stamp: 'tpr 12/1/2000 16:12'!
pluginsDirName
	^DirNames at: #pluginsDir! !

!VMMaker class methodsFor: 'file utilities' stamp: 'eem 3/6/2018 16:40'!
rootDirectory
	^FileDirectory default directoryNamed: self sourceTree! !

!VMMaker class methodsFor: 'accessing' stamp: 'tpr 12/1/2000 16:14'!
sourceDirName
	^DirNames at: #sourceDir! !

!VMMaker class methodsFor: 'accessing' stamp: 'eem 4/18/2014 11:44'!
sourceTree
	^DirNames at: #sourceTree! !

!VMMaker class methodsFor: 'version testing' stamp: 'eem 11/1/2013 13:42'!
versionString
	"VMMaker versionString"

	^'4.7.0 (Cog)'! !

!VMMaker methodsFor: 'accessing' stamp: 'eem 7/23/2014 23:31'!
abortBlock
	^abortBlock! !

!VMMaker methodsFor: 'plugin lists' stamp: 'ar 4/7/2002 13:09'!
allModuleNames
	"return the list of all the all plugins' moduleNames"
	^Array streamContents:[:strm| self allPluginsDo:[:pl| strm nextPut: pl moduleName ]]! !

!VMMaker methodsFor: 'plugin lists' stamp: 'ar 4/7/2002 13:10'!
allPluginsDo: aBlock 
	"for each class that should be an external plugin, evaluate aBlock"
	self externalPluginsDo: aBlock.
	self internalPluginsDo: aBlock.! !

!VMMaker methodsFor: 'UI access' stamp: ''!
availablePlugins
	allPlugins ifNil:[self initializeAllPlugins].
	^allPlugins! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 6/7/2016 17:02'!
buildCodeGeneratorForCogit
	"Answer the code generator for translating the cogit."

	^(Smalltalk classNamed: #CurrentReadOnlySourceFiles)
		ifNil: [self
				buildCodeGeneratorForCogit: self interpreterClass cogitClass
				includeAPIMethods: true
				initializeClasses: true]
		ifNotNil:
			[:crosf|
			 crosf cacheDuring:
				[self
					buildCodeGeneratorForCogit: self interpreterClass cogitClass
					includeAPIMethods: true
					initializeClasses: true]]! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 6/7/2018 15:31'!
buildCodeGeneratorForCogit: cogitClass includeAPIMethods: getAPIMethods initializeClasses: initializeClasses
	"Answer the code generator for translating the cogit."

	| cg cogitClasses |
	cg := self createCogitCodeGenerator.

	cg vmClass: cogitClass.
	initializeClasses ifTrue:
		[{ cogitClass. self interpreterClass. self interpreterClass objectMemoryClass } do:
			[:cgc|
			(cgc respondsTo: #initializeWithOptions:)
				ifTrue: [cgc initializeWithOptions: optionsDictionary]
				ifFalse: [cgc initialize]]].

	cogitClasses := OrderedCollection withAll: (cogitClass withAllSuperclasses copyUpThrough: VMClass) reverse.
	cogitClasses addAllLast: (cogitClass ancilliaryClasses reject: [:class| class isStructClass]).
	cogitClasses do: [:cgc| cg addClass: cgc].
	"Now make sure to add struct classes that the most specific subclasses specify.
	 This makes sure that struct classes are ordered as the most specific cogitClass desires.
	 This must happen for references between the struct classes to be resolved in a specific order."
	cogitClasses := OrderedCollection with: cogitClass with: cogitClass objectRepresentationClass.
	cogitClasses addAll: (cogitClass ancilliaryClasses reject:
							[:class|
							 class isStructClass
							 or: [cogitClass objectRepresentationClass includesBehavior: class]]).
	cg addStructClasses: (cg structClassesForTranslationClasses: cogitClasses).

	getAPIMethods ifTrue:
		[cg includeAPIFrom: (self
								buildCodeGeneratorForInterpreter: self interpreterClass
								includeAPIMethods: false
								initializeClasses: false)].

	^cg! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 6/7/2016 17:03'!
buildCodeGeneratorForInterpreter
	"Answer the code generator for translating the interpreter."

	^(Smalltalk classNamed: #CurrentReadOnlySourceFiles)
		ifNil: [self
				buildCodeGeneratorForInterpreter: self interpreterClass
				includeAPIMethods: true
				initializeClasses: true]
		ifNotNil:
			[:crosf|
			 crosf cacheDuring:
				[self
					buildCodeGeneratorForInterpreter: self interpreterClass
					includeAPIMethods: true
					initializeClasses: true]]! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 6/7/2018 15:09'!
buildCodeGeneratorForInterpreter: interpreterClass includeAPIMethods: getAPIMethods initializeClasses: initializeClasses
	"Answer the code generator for translating the interpreter."

	| cg interpreterClasses |
	initializeClasses ifTrue:
		[interpreterClass initializeWithOptions: optionsDictionary.
		 interpreterClass hasCogit ifTrue:
			[interpreterClass cogitClass initializeWithOptions: optionsDictionary]].

	(cg := self createCodeGenerator) vmClass: interpreterClass.

	"Construct interpreterClasses as all classes from interpreterClass &
	 objectMemoryClass up to VMClass in superclass to subclass order."
	interpreterClasses := OrderedCollection new.
	{interpreterClass. interpreterClass objectMemoryClass} do:
		[:vmClass| | theClass |
		 theClass := vmClass.
		 [theClass ~~ VMClass] whileTrue:
			[interpreterClasses addFirst: theClass.
			 theClass := theClass superclass]].
	interpreterClasses
		addFirst: VMClass;
		addAllLast: (cg nonStructClassesForTranslationClasses: interpreterClasses).

	initializeClasses ifTrue:
		[interpreterClasses do:
			[:ic|
			(ic respondsTo: #initializeWithOptions:)
				ifTrue: [ic initializeWithOptions: interpreterClass initializationOptions]
				ifFalse: [ic initialize]].
		 (cg structClassesForTranslationClasses: interpreterClasses) do:
			[:structClass| structClass initialize]].

	cg addStructClasses: (cg structClassesForTranslationClasses: interpreterClasses).

	interpreterClasses do: [:ic| cg addClass: ic].

	getAPIMethods ifTrue:
		[interpreterClass cogitClass ifNotNil:
			[:cogitClass|
			 cg includeAPIFrom: (self
									buildCodeGeneratorForCogit: cogitClass
									includeAPIMethods: false
									initializeClasses: false)]].

	^cg! !

!VMMaker methodsFor: 'plugin lists' stamp: 'tpr 3/13/2002 15:56'!
canSupportPlugin: pluginClassName 
	"see if this plugin needs any external files and if so, check to see if 
	they seem to exist."
	[self validatePlugin: pluginClassName in: allPlugins , internalPlugins , externalPlugins]
		on: VMMakerException
		do: [^ false].
	^ true! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 12/6/2013 14:17'!
configurationGeneratorNameOrNil
	"Search the sender chain for the first method in VMMaker class protocol configurations, and answer it if found,
	 otherwise answer nil.  This allows us to name the configuration being generated."
	^thisContext findContextSuchThat:
		[:ctxt|
		 ctxt receiver == VMMaker
		 and: [ctxt method methodReference category = #configurations]]! !

!VMMaker methodsFor: 'objects from disk' stamp: 'eem 9/10/2013 14:41'!
configurationInfo
	"build a simple Array of the configuration information that would be 
	 usefully saved for later reloading:- 
		the list of internal & external plugins,
		the flags,
		the platform name,
		the two major directory names,
		bytePerWord
		two flags indicating whether each directory is relative to the current directory or not.
		the interpreter class name"
	| isRelative makeRelative |
	isRelative := [:pn| pn beginsWith: FileDirectory default pathName].
	makeRelative := [:pn|
					(isRelative value: pn)
						ifTrue: [pn allButFirst: FileDirectory default pathName size + 1]
						ifFalse: [pn]].
	^{ internalPlugins asArray.
		externalPlugins asArray.
		inline.
		forBrowser.
		platformName.
		makeRelative value: self sourceDirectory pathName.
		makeRelative value: self platformRootDirectory pathName.
		self wordSize.
		isRelative value: self sourceDirectory pathName.
		isRelative value: self platformRootDirectory pathName.
		self interpreterClassName
	  }! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 6/13/2014 00:14'!
configurationNameIfAny
	^self configurationGeneratorNameOrNil
		ifNotNil: [:ctxt| (self class configurationNameFromSelector: ctxt selector), ' ']
		ifNil: [' ']! !

!VMMaker methodsFor: 'private - copying files' stamp: 'ar 4/7/2002 15:59'!
copyFileNamed: srcName to: dstName 
	| dstEntry srcEntry |
	dstEntry := FileDirectory directoryEntryFor: dstName.
	dstEntry ifNotNil:[
		srcEntry := FileDirectory directoryEntryFor: srcName.
		srcEntry ifNil:[^self couldNotOpenFile: srcName].
		dstEntry modificationTime >= srcEntry modificationTime ifTrue:[^self].
	].
	logger show:'==> ', dstName; cr.
	^self primitiveCopyFileNamed: srcName to: dstName ! !

!VMMaker methodsFor: 'target directories' stamp: 'tpr 2/14/2002 14:27'!
coreVMDirectory
	"return the target directory for the main VM sources, interp.c etc"
	| fd |
	fd := self sourceDirectory directoryNamed: self class coreVMDirName.
	fd assureExistence.
	^ fd! !

!VMMaker methodsFor: 'private - errors' stamp: 'eem 8/4/2008 15:30'!
couldNotFindDirectory: dirName
	"This should raise a nice exception to a UI"
	^(VMMakerException new messageText: self class name, ' could not find directory ', dirName) signal! !

!VMMaker methodsFor: 'private - errors' stamp: 'tpr 10/18/2001 19:41'!
couldNotFindPlatformDirectoryFor: platName
	"This should raise a nice exception to a UI"
	self couldNotFindDirectory: 'for: ', platName, ' specific files; is the platform root path set correctly?'! !

!VMMaker methodsFor: 'private - errors' stamp: 'tpr 10/21/2001 11:14'!
couldNotFindPlatformFilesFor: plugin
	"This should raise a nice exception to a UI"
	(VMMakerException new messageText: self class name, ' could not find platform specific files for: ', plugin moduleName) signal! !

!VMMaker methodsFor: 'private - errors' stamp: 'tpr 10/21/2001 11:15'!
couldNotFindPluginClass: pluginSymbol
	"This should raise a nice exception to a UI"
	(VMMakerException new messageText: self class name, ' could not find the class for: ', pluginSymbol) signal! !

!VMMaker methodsFor: 'private - errors' stamp: 'tpr 10/21/2001 11:12'!
couldNotOpenFile: fileName
	"This should raise a nice exception to a UI"
	(VMMakerException new messageText: self class name, ' could not open file: ', fileName) signal! !

!VMMaker methodsFor: 'initialize' stamp: 'eem 10/30/2013 16:27'!
createCodeGenerator
"set up a CCodeGenerator for this VMMaker"
	^CCodeGenerator new
		vmMaker: self;
		logger: logger;
		options: optionsDictionary;
		yourself! !

!VMMaker methodsFor: 'initialize' stamp: 'eem 10/30/2013 16:26'!
createCogitCodeGenerator
	^CCodeGenerator new
		vmMaker: self;
		logger: logger;
		options: optionsDictionary;
		yourself! !

!VMMaker methodsFor: 'source directories' stamp: 'eem 3/6/2018 16:55'!
crossPlatformDirectory
	"return the directory where we should find the cross-platform literal 
	sources - <sq.h> etc"
	| dir |
	dir := self platformRootDirectory directoryNamed: 'Cross'.
	dir exists ifFalse: "The supposed directory for the actual cross-platform code does not exist."
		[^self couldNotFindPlatformDirectoryFor: 'cross-platform '].
	^dir! !

!VMMaker methodsFor: 'source directories' stamp: 'eem 3/6/2018 16:54'!
crossPlatformPluginsDirectory
	"return the directory where we should find the cross-platform plugin specific sources"
	| dir |
	dir := self crossPlatformDirectory directoryNamed: self class pluginsDirName.
	dir exists ifFalse: "The supposed directory for the plugins code does not exist.
					  We need to raise a suitable exception, but can't think of one right now."
		[^self couldNotFindPlatformDirectoryFor: 'any plugins needing cross-platform'].
	^dir! !

!VMMaker methodsFor: 'target directories' stamp: 'tpr 3/27/2002 15:19'!
deleteEntireGeneratedTree
	"remove all the files - all of them I say"
	self sourceDirectory recursiveDelete! !

!VMMaker methodsFor: 'target directories' stamp: 'ar 6/13/2007 15:58'!
deleteUnwantedExternalPluginDirectories
	"don't. in real life people back plugin directories with source code controls and simply deleting them really doesn't help."! !

!VMMaker methodsFor: 'target directories' stamp: 'ar 6/13/2007 15:59'!
deleteUnwantedInternalPluginDirectories
	"don't. in real life people back plugin directories with source code controls and simply deleting them really doesn't help."! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 11/14/2019 12:58'!
doInlining
	"default is true but see VMMaker>initialize for details"
	self interpreterClass ifNotNil:
	    [:interpreterClass|
		(interpreterClass respondsTo: #doInlining) ifTrue:
			[^interpreterClass doInlining]].
	^inline! !

!VMMaker methodsFor: 'exports' stamp: 'tpr 4/20/2001 16:35'!
export: exportList forExternalPlugin: aPlugin
"it may be useful on certain platforms to do something with the export list of external plugins, just as the internal plugins' exports get added to the VM list. Default is to do nothing though."! !

!VMMaker methodsFor: 'UI access' stamp: ''!
externalModules
	^externalPlugins! !

!VMMaker methodsFor: 'target directories' stamp: 'eem 1/14/2010 12:48'!
externalPluginListName
	"Answer the filename for the list of external plugins"

	^'plugins.ext'! !

!VMMaker methodsFor: 'target directories' stamp: 'tpr 2/14/2002 14:27'!
externalPluginsDirectory
	"return the target directory for the external plugins sources"
	| fd |
	fd := self sourceDirectory directoryNamed: self class pluginsDirName.
	fd assureExistence.
	^fd! !

!VMMaker methodsFor: 'target directories' stamp: 'tpr 2/14/2002 14:27'!
externalPluginsDirectoryFor: plugin
	"return the directory for the external plugin sources"
	|fd|
	fd := self externalPluginsDirectory directoryNamed: plugin moduleName.
	fd assureExistence.
	^fd! !

!VMMaker methodsFor: 'plugin lists' stamp: 'tpr 1/24/2001 12:53'!
externalPluginsDo: aBlock 
	"for each class that should be an external plugin, evaluate aBlock"
	self plugins: externalPlugins do: aBlock! !

!VMMaker methodsFor: 'initialize' stamp: 'tpr 4/4/2005 19:26'!
for32BitVM
"set my flag to make a 32bit pointer model VM"
	is64BitVM := false.
	self changed: #sourcePathText.! !

!VMMaker methodsFor: 'initialize' stamp: 'tpr 4/4/2005 19:19'!
for64BitVM
"set my flag to make a 64bit pointer model VM"
	is64BitVM := true.
	self changed: #sourceDirectory.! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 4/22/2015 10:41'!
generateCogitFileFor: cogitClass
	"Translate the Smalltalk description of the Cogit into C, if required.  Ensure that
	 inferTypesForImplicitlyTypedVariablesAndMethods was done, and answer the code generator."

	| cg |
	cg := [self buildCodeGeneratorForCogit]
			on: Notification
			do: [:ex|
				ex tag == #getVMMaker
					ifTrue: [ex resume: self]
					ifFalse: [(ex respondsTo: #rearmHandlerDuring:)
								ifTrue: [ex rearmHandlerDuring: [ex pass]]
								ifFalse: [ex pass]]].

	cg inferTypesForImplicitlyTypedVariablesAndMethods.
	self needsToRegenerateCogitFile ifFalse: [^cg].

	cg vmClass preGenerationHook: cg.
	cg storeCodeOnFile: (self sourceFilePathFor: cogitClass processorSpecificSourceFileName) doInlining: cogitClass doInlining.
	^cg! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 3/20/2017 09:25'!
generateCogitFiles
	"Translate the Smalltalk description of the virtual machine's JITs into C."
	| cogitClass cg |
	(cogitClass := self interpreterClass cogitClass) ifNil: [^nil].
	self generateCogitIncludeFileFor: cogitClass.
	cogitClass translateableInstructionSubclassesAndInstalledOptionsDo:
		[:compilerClass|
		cg := self generateCogitFileFor: cogitClass].
	cg vmClass additionalHeadersDo:
		[:headerName :headerContents| | filePath |
		 filePath := self coreVMDirectory fullNameFor: headerName.
		 (cg needToGenerateHeader: headerName file: filePath contents: headerContents) ifTrue:
			 [cg storeHeaderOnFile: filePath contents: headerContents]].
	cogitClass apiExportHeaderName ifNotNil:
		[cg storeAPIExportHeader: cogitClass apiExportHeaderName
			OnFile: (self sourceFilePathFor: cogitClass apiExportHeaderName)]! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 4/22/2015 10:24'!
generateCogitIncludeFileFor: cogitClass
	"Generate the skeletal cogit.c that includes the processor-specific cogit files."

	| code fileName file fileCode |
	code := Cogit generateCodeStringForCogitDotC.
	fileName := self coreVMDirectory fullNameFor: cogitClass sourceFileName.
	((self coreVMDirectory fileExists: cogitClass sourceFileName)
	and: [fileCode := (self class oldFileNamed: fileName) contentsOfEntireFile.
		(fileCode allButFirst: (fileCode indexOf: $#)) = (code allButFirst: (code indexOf: $#))]) ifFalse:
		[file := self class forceNewFileNamed: fileName.
		 [file nextPutAll: code] ensure:
			[file close]]! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 11/30/2017 09:27'!
generateEntire
	"Generate the interp, internal plugins and exports as well as the external plugins.
	 If this comes from a generator method, log it for convenience."
	abortBlock := [^self].
	self logGeneration.
	self generateMainVM.
	self generateExternalPlugins! !

!VMMaker methodsFor: 'exports' stamp: 'eem 04/27/2017 12:55'!
generateExportsFile
	"Store the exports on the given file"
	| cg contents filePath fileStream |
	filePath := self interpreterExportsFilePath.
	"don't bother endlessly regenerating the example file."
	(internalPlugins isEmpty
	 and: [(filePath includesSubstring: 'example')
	 and: [FileDirectory default fileExists: filePath]]) ifTrue:
		[^self].
	cg := self createCodeGenerator.
	cg vmClass: self interpreterClass.
	contents := String streamContents:
		[:s|
		s
			nextPutAll:'/* This is an automatically generated table of all builtin modules in the VM';
			cr;
			next: 3 put: Character space;
			nextPutAll: (cg shortMonticelloDescriptionForClass: cg vmClass);
			cr;
			nextPutAll:' */';
			cr.
		s cr; nextPutAll:'extern sqExport vm_exports[];'.
		s cr; nextPutAll: 'extern sqExport os_exports[];'.
		self internalPluginsDo:[:cls|
			s cr; nextPutAll: 'extern sqExport '; nextPutAll: cls moduleName; nextPutAll:'_exports[];'.
		].
		s cr.

		s cr; nextPutAll:'sqExport *pluginExports[] = {'.
		s crtab; nextPutAll:'vm_exports,'.
		s crtab; nextPutAll: 'os_exports,'.
		self internalPluginsDo:[:cls|
			s crtab; nextPutAll: cls moduleName; nextPutAll:'_exports,'
		].
		s crtab; nextPutAll:'NULL'.
		s cr; nextPutAll:'};'; cr].
	(cg needToGenerateHeader: (FileDirectory baseNameFor: filePath) file: filePath contents: contents) ifTrue:
		[[fileStream := VMMaker forceNewFileNamed: filePath] 
			on: FileDoesNotExistException 
			do:[^self couldNotOpenFile: filePath].
		 fileStream nextPutAll: contents; close]! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 3/6/2019 11:21'!
generateExternalPlugin: pluginName 
	"generate the named external plugin"
	| exports plugin |

	"Refuse to translate this plugin if it requires platform specific files and they are not present."
	[plugin := self validateExternalPlugin: pluginName]
		on: VMMakerException
		do: [:ex|
			logger show: 'external plugin ' , plugin name , ' failed to validate: ' , ex messageText; cr.
			^self].

	VMClass initializeForPlugins: #().
	[exports := plugin
					translateInDirectory: (self externalPluginsDirectoryFor: plugin)
					doInlining: inline]
		on:  ProvideAnswerNotification
		do: [:ex|
			ex tag == #logger
				ifTrue: [ex resume: logger]
				ifFalse: [ex pass]].
	exports ifNotNil: "if exp is nil we skip this since the plugin was already up to date"
		[logger show: 'external plugin ' , plugin name , ' generated as ' , plugin moduleName; cr.
		 self export: exports forExternalPlugin: plugin.
		 self processFilesForExternalPlugin: plugin]! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 3/9/2016 17:05'!
generateExternalPlugins
	"generate the external plugins"

	self logDateAndTime.
	self deleteUnwantedExternalPluginDirectories.
	InterpreterPlugin initialize.
	self externalPluginsDo: [:plugin | 
		self generateExternalPlugin: plugin].
	self storeExternalPluginList.
	self logDateAndTime! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 6/26/2009 11:20'!
generateInternalPlugin: pluginName 
	"generate the named internal plugin. Make sure the exports list is actually 
	correct and write it out"
	self deleteUnwantedInternalPluginDirectories.
	self privateGenerateInternalPlugin: pluginName.
	self generateExportsFile.
	self logDateAndTime! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 6/26/2009 11:20'!
generateInternalPlugins
	"generate the internal plugins and add their exports to the main list. te exports list is NOT written to file by this method"

	self deleteUnwantedInternalPluginDirectories.
	InterpreterPlugin initialize.
	self internalPluginsDo: [:plugin | 
		self privateGenerateInternalPlugin: plugin].
	self logDateAndTime.
	self storeInternalPluginList.! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 7/8/2021 10:44'!
generateInterpreterFile
	"Translate the Smalltalk description of the virtual machine into C.  If 'self doInlining' is true, small method bodies are inlined to reduce procedure call overhead.  On the PPC, this results in a factor of three speedup with only 30% increase in code size.  Subclasses can use specialised versions of CCodeGenerator and interpreterClass."

	| cg vmHeaderContents |
	cg := [self buildCodeGeneratorForInterpreter]
			on: Notification
			do: [:ex|
				ex tag == #getVMMaker
					ifTrue: [ex resume: self]
					ifFalse: [(ex respondsTo: #rearmHandlerDuring:)
								ifTrue: [ex rearmHandlerDuring: [ex pass]]
								ifFalse: [ex pass]]].
	self reinitializeWordSizeFrom: cg.

	self interpreterClass additionalHeadersDo:
		[:headerName :headerContents| | filePath |
		 filePath := self coreVMDirectory fullNameFor: headerName.
		 (cg needToGenerateHeader: headerName file: filePath contents: headerContents) ifTrue:
			 [cg storeHeaderOnFile: filePath contents: headerContents]].

	self needsToRegenerateInterpreterFile ifFalse: [^nil].

	cg inferTypesForImplicitlyTypedVariablesAndMethods.

	self interpreterClass preGenerationHook: cg.
	vmHeaderContents := cg vmHeaderContentsWithBytesPerWord: self wordSize.
	(cg needToGenerateHeader: self interpreterHeaderName file: self interpreterHeaderPath contents: vmHeaderContents) ifTrue:
		[cg storeHeaderOnFile: self interpreterHeaderPath contents: vmHeaderContents].
	cg storeCodeOnFile: (self sourceFilePathFor: self interpreterClass sourceFileName) doInlining: self doInlining.
	self interpreterClass apiExportHeaderName ifNotNil:
		[cg storeAPIExportHeader: self interpreterClass apiExportHeaderName
			OnFile: (self sourceFilePathFor: self interpreterClass apiExportHeaderName)].
	(cg methodNamed: #interpret) ifNotNil:
		[self gnuifyInterpreterFile].
	self maybeGenerateVariableOrderFiles: cg! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 3/16/2017 09:59'!
generateMainVM
	"Generate the interp (and optionally the cogit), internal plugins and exports.
	 N.B. generateInterpreterFile *must* precede generateCogitFile so that
	 the objectMemory and interpreter classes are initialized before the Cogit
	 code is generated."

	self logDateAndTime;
		generateInterpreterFile;
		generateCogitFiles;
		generateInternalPlugins;
		generateExportsFile;
		logDateAndTime! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 11/30/2017 09:27'!
generatePlugins
	"Generate the ``external'' plugins.
	 If this comes from a generator method, log it for convenience."
	abortBlock := [^self].
	self logGeneration.
	self generateExternalPlugins! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 7/28/2020 14:31'!
generateUtilityPrograms
	"Any additional sources not directly part of the VM may be generated here"

	Smalltalk
		at: #ImageFormat
		ifPresent:
			[:cls |  "generate ckformat utility program"
			(self needsToRegenerateCheckFormatFile: cls named: 'ckformat.c') ifTrue:
				[cls storeCkFormatOnFile: (self sourceDirectory fullNameFor: 'ckformat.c')]]! !

!VMMaker methodsFor: 'processing external files' stamp: 'tpr 4/9/2013 17:36'!
gnuifyInterpreterFile
"post-process the interp.c file to make it gcc friendly"
	(Gnuifier on: self coreVMDirectory)
		interpreterFilename: self interpreterFilename;
		gnuify.! !

!VMMaker methodsFor: 'initialize' stamp: 'eem 10/23/2018 16:25'!
initialize
	logger := Transcript.
	inline := true.
	forBrowser := false.
	internalPlugins := SortedCollection new.
	externalPlugins := SortedCollection new.
	platformName := 'Cross'.
	is64BitVM := false.
	interpreterClassName := StackInterpreterPrimitives name.
	optionsDictionary := Dictionary newFromPairs: {#BytesPerWord. 4}.
	VMStructType voidStructTypeCache! !

!VMMaker methodsFor: 'initialize' stamp: 'tpr 10/11/2001 13:00'!
initializeAllPlugins
	allPlugins := self providedPlugins! !

!VMMaker methodsFor: 'initialize' stamp: ''!
initializeInternal: arrayOfInternalPluginNames external: arrayOfExternalPluginNames
	"try to set up with the listed internal and external plugins."

	self initialize.

	self internal: arrayOfInternalPluginNames external: arrayOfExternalPluginNames! !

!VMMaker methodsFor: 'initialize' stamp: 'tpr 7/30/2002 18:38'!
internal: arrayOfInternalPluginNames external: arrayOfExternalPluginNames
	"try to set up with the listed internal and external plugins. Check that they are supportable plugins, reject those that are not - remember that this depends on the platform etc "

	"since we went to some trouble to drop plugins we cannot handle, don't add them now"
	internalPlugins := (self availablePlugins intersection: arrayOfInternalPluginNames) select: [:pl | self canSupportPlugin: pl].
	allPlugins := allPlugins copyWithoutAll: internalPlugins.
	externalPlugins := (allPlugins intersection: arrayOfExternalPluginNames) select: [:pl | self canSupportPlugin: pl ].
	allPlugins := allPlugins copyWithoutAll: externalPlugins.
	! !

!VMMaker methodsFor: 'UI access' stamp: ''!
internalModules
	^internalPlugins! !

!VMMaker methodsFor: 'target directories' stamp: 'eem 1/14/2010 12:46'!
internalPluginListName
	"Answer the filename for the list of internal plugins"

	^'plugins.int'! !

!VMMaker methodsFor: 'target directories' stamp: 'tpr 2/14/2002 14:27'!
internalPluginsDirectory
	"return the directory for the internal plugins sources"
	|fd|
	fd := self coreVMDirectory directoryNamed: 'intplugins'.
	fd assureExistence.
	^fd! !

!VMMaker methodsFor: 'target directories' stamp: 'tpr 2/14/2002 14:27'!
internalPluginsDirectoryFor: plugin
	"return the directory for the internal plugin sources"
	|fd|
	fd := self internalPluginsDirectory directoryNamed: plugin moduleName.
	fd assureExistence.
	^fd! !

!VMMaker methodsFor: 'plugin lists' stamp: 'tpr 1/24/2001 12:53'!
internalPluginsDo: aBlock 
	"for each class that should be an internal plugin, evaluate aBlock"
	self plugins: internalPlugins do: aBlock! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 6/3/2011 08:54'!
interpreterClass

	^(Smalltalk classNamed: interpreterClassName) ifNotNil:
		[:interpreterClass| interpreterClass translationClass]! !

!VMMaker methodsFor: 'generate sources' stamp: 'svp 11/14/2002 21:11'!
interpreterClass: aClass

	interpreterClassName := aClass name asString! !

!VMMaker methodsFor: 'generate sources' stamp: 'svp 11/14/2002 21:03'!
interpreterClassName

	^interpreterClassName! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 12/15/2010 17:03'!
interpreterClassName: aString
	| tmp |
	tmp := Smalltalk at: aString asSymbol ifAbsent: [nil].
	tmp isBehavior
		ifTrue: [self interpreterClass: tmp]
		ifFalse: [self invalidClassName]! !

!VMMaker methodsFor: 'generate sources' stamp: 'tpr 3/11/2003 14:02'!
interpreterExportsFilePath
	"return the full path for the interpreter exports file"
	^self coreVMDirectory fullNameFor: 'sqNamedPrims.h'! !

!VMMaker methodsFor: 'target directories' stamp: 'eem 2/9/2009 12:41'!
interpreterFilename
	"Answer the filename for the core interpreter.  Default is 'interp.c'."

	^self interpreterClass sourceFileName! !

!VMMaker methodsFor: 'target directories' stamp: 'ikp 9/2/2004 14:11'!
interpreterHeaderName
	"Answer the filename for the core interpreter header.  Default is 'interp.h'."

	^'interp.h'! !

!VMMaker methodsFor: 'generate sources' stamp: 'ikp 9/2/2004 14:10'!
interpreterHeaderPath
	"Answer the fully-qualified path for the generated interpreter header file."

	^self coreVMDirectory fullNameFor: self interpreterHeaderName! !

!VMMaker methodsFor: 'private - errors' stamp: 'svp 11/14/2002 21:13'!
invalidClassName
	"This should raise a nice exception to a UI"
	(VMMakerException new messageText: self class name, ' invalid interpreter class name: ', interpreterClassName) signal! !

!VMMaker methodsFor: 'initialize' stamp: 'tpr 4/4/2005 19:12'!
isFor64BitVM
"is my flag to make a 64bit pointer model VM?"
	^is64BitVM! !

!VMMaker methodsFor: 'UI access' stamp: 'tpr 10/12/2001 15:34'!
listOfName: aSymbol
	"work out which list is the one associated with this symbol"
	#availableModules = aSymbol ifTrue:[^allPlugins].
	#internalModules = aSymbol ifTrue:[^internalPlugins].
	#externalModules =aSymbol ifTrue:[^externalPlugins].
	^nil! !

!VMMaker methodsFor: 'objects from disk' stamp: 'eem 1/14/2010 13:08'!
loadConfiguration: aConfigArray
	"load the configuration but ignore the platformName - the platform name must have been handled during the creation of this vmmaker in order for it to work correctly"

	inline := aConfigArray at: 3.
	forBrowser := aConfigArray at: 4.
	"This part must be ignored --> self setPlatName: (aConfigArray at: 5)."
	"Don't go through the setters sourceDirectoryName: & platformRootDirectoryName:.
	 They do validation which only works on certain platforms and create bogus directories, etc.
	 They're broken for non-interactive use."
	sourceDirName := aConfigArray at: 6.
	platformRootDirName := aConfigArray at: 7.
	self initializeAllPlugins.
	self internal: (aConfigArray at: 1) asSortedCollection
		external:(aConfigArray at: 2) asSortedCollection.
	aConfigArray size >= 9 ifTrue: "new enough to have 64bitness flag"
		[(aConfigArray at: 8) = 8 ifTrue: [self for64BitVM].
		 (aConfigArray at: 8) = 4 ifTrue: [self for32BitVM]].
	aConfigArray size >= 11 ifTrue: "new enough to include the interpreter name"
		[self interpreterClassName: (aConfigArray at: 11)].
	self changed: #sourceDirectory;
		 changed: #reinitialize! !

!VMMaker methodsFor: 'UI access' stamp: 'eem 11/9/2011 11:36'!
logDateAndTime
	| now |
	"do it this way since Time now includes milliseconds in some versions."
	now := Time dateAndTimeNow.
	logger print: now first; space; print: now last; cr; flush! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 11/30/2017 09:26'!
logGeneration
	self configurationGeneratorNameOrNil ifNotNil:
		[:generator|
		 logger cr; nextPutAll: (generator selector copyReplaceAll: 'generate' with: '').
		 interpreterClassName ifNotNil:
			[logger space; nextPutAll: (CCodeGenerator shortMonticelloDescriptionForClass: (Smalltalk classNamed: interpreterClassName))].
		 sourceDirName ifNotNil:
			[logger nextPutAll: ' to '; nextPutAll: sourceDirName].
		 logger cr; flush]! !

!VMMaker methodsFor: 'UI access' stamp: 'ar 3/10/2002 15:02'!
logger
	^logger! !

!VMMaker methodsFor: 'UI access' stamp: 'ar 3/10/2002 15:02'!
logger: aStream
	logger := aStream.! !

!VMMaker methodsFor: 'UI access' stamp: 'md 10/27/2004 14:23'!
makeAllModulesAvailable
	self internal: #() external: #().
	self reinitializePluginsLists! !

!VMMaker methodsFor: 'UI access' stamp: 'tpr 10/16/2001 12:10'!
makeAllModulesExternal
	self initializeAllPlugins.
	self internal: #() external: self availablePlugins.
	self changed: #reinitialize ! !

!VMMaker methodsFor: 'UI access' stamp: 'tpr 10/16/2001 12:10'!
makeAllModulesInternal
	self initializeAllPlugins.
	self internal: self availablePlugins external: #().
	self changed: #reinitialize ! !

!VMMaker methodsFor: 'target directories' stamp: 'tpr 4/9/2002 17:34'!
makefileDirectory
"where to put generated makefile related files"
	^self sourceDirectory! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 7/8/2021 10:55'!
maybeGenerateVariableOrderFiles: cg
	| order |
	(self interpreterClass respondsTo: #clusteredVariableNames) ifFalse: [^self].
	"First find the longest order. These files can contain false positives.
	 So e.g. it's preferrable to generate the list for CoInterpreterMT
	 instead of two lists, one each for CoInterpreterMT and CoInterpreter."
	order := self interpreterClass allSubclasses
				inject: self interpreterClass clusteredVariableNames
				into: [:o :sc| sc clusteredVariableNames size > o size ifTrue: [sc clusteredVariableNames] ifFalse: [o]].
	"now generate versions for the different symbol formats in use."
	#('variable_order' '_variable_order')
		with: #(nil $_)
		do: [:file :prefix| | contents filePath |
			contents := String streamContents:
							[:s|
							order do:
								[:name| prefix ifNotNil: [s nextPut: prefix]. s nextPutAll: name; cr]].
			filePath := self coreVMDirectory fullNameFor: file.
			(cg needToGenerateHeader: file file: filePath contents: contents) ifTrue:
				[(self class forceNewFileNamed: filePath)
					nextPutAll: contents;
					close]]! !

!VMMaker methodsFor: 'UI access' stamp: 'tpr 10/12/2001 15:35'!
movePlugin: pluginName from: srcListName to: dstListName
	"the VMMakerTool UI has been used to drag a plugin from one list to 
	another "
	"we need to do some tests - 
	are the lists actually ours? 
	is the plugin ours? 
	is the destination list one where we must check the plugin for 
	acceptability? return true if all is ok, false otherwise"
	| dstList srcList |
	dstList := self listOfName: dstListName.
	srcList := self listOfName: srcListName.
	dstList == allPlugins
		ifTrue: [dstList
				add: (srcList remove: pluginName)]
		ifFalse: ["the dest must be internal or external, so check the plugin for 
			acceptability "
			(self canSupportPlugin: pluginName)
				ifTrue: [dstList
						add: (srcList remove: pluginName)]]! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 7/28/2020 14:36'!
needsToRegenerateCheckFormatFile: cls named: fileName
	"Check the timestamp for ImageFormat and then the timestamp for ckformat.c if it already exists.  Answer if the file needs regenerating."

	| tStamp |
	tStamp := cls timeStamp.

	"don't translate if the file is newer than my timeStamp"
	(self sourceDirectory entryAt: fileName ifAbsent: [nil]) ifNotNil:
		[:fstat| | mTime |
		mTime := fstat modificationTime.
		mTime isInteger ifFalse: [mTime := mTime asSeconds].
		tStamp < mTime ifTrue:
			[^self confirm: (cls name, ' has not been modified since\', fileName, ' was last generated.\Do you still want to regenerate it?') withCRs]].
	^true! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 6/7/2018 15:31'!
needsToRegenerateCogitFile
	"Check the timestamp for the relevant classes and then the timestamp for the main source file (e.g. interp.c)
	 file if it already exists. Answer if the file needs regenerating."

	| cogitClass cogitClasses tStamp files |
	cogitClasses := (cogitClass := self interpreterClass cogitClass) withAllSuperclasses copyUpThrough: Cogit.
	cogitClasses addAllLast: cogitClass ancilliaryClasses.
	tStamp := cogitClasses inject: 0 into: [:tS :cl| tS max: cl timeStamp].

	"don't translate if the file(s) is newer than my timeStamp"
	files := (self coreVMDirectory fileNamesMatching: cogitClass activeCompilerClass moduleName, '*.c').
	files isEmpty ifTrue:
		[^true].
	(files allSatisfy:
		[:fileName|
		(self coreVMDirectory entryAt: fileName ifAbsent: [nil])
			ifNil: [false]
			ifNotNil:
				[:fstat| | mTime |
				mTime := fstat modificationTime.
				mTime isInteger ifFalse: [mTime := mTime asSeconds].
				tStamp < mTime]]) ifTrue:
		[^self confirm: ('The ', self configurationNameIfAny, cogitClass printString,
			', ', cogitClass activeCompilerClass, '\classes have not been modified since the ',
			cogitClass processorSpecificSourceFileName,
			' source file\was last generated.  Do you still want to regenerate it?') withCRs].
	^true! !

!VMMaker methodsFor: 'initialize' stamp: 'eem 6/7/2018 15:30'!
needsToRegenerateInterpreterFile
	"Check the timestamp for the relevant classes and then the timestamp for the main
	 source file (e.g. interp.c) if it already exists.  Answer if the file needs regenerating."

	| classes tStamp |
	classes := self interpreterClass withAllSuperclasses copyUpTo: VMClass.
	self interpreterClass objectMemoryClass ifNotNil:
		[:objectMemoryClass|
		classes addAllLast: (objectMemoryClass withAllSuperclasses copyUpTo: VMClass)].
	classes copy do:
		[:class| classes addAllLast: class ancilliaryClasses].
	tStamp := classes inject: 0 into: [:tS :cl| tS max: cl timeStamp].

	"don't translate if the file is newer than my timeStamp"
	(self coreVMDirectory entryAt: self interpreterFilename ifAbsent: [nil]) ifNotNil:
		[:fstat| | mTime |
		mTime := fstat modificationTime.
		mTime isInteger ifFalse: [mTime := mTime asSeconds].
		tStamp < mTime ifTrue:
			[^self confirm: 'The ', self configurationNameIfAny, 'interpreter classes have not been modified since\ the interpreter file was last generated.\Do you still want to regenerate the source file?' withCRs]].
	^true
! !

!VMMaker methodsFor: 'accessing' stamp: 'eem 8/14/2012 15:43'!
options
	^optionsDictionary! !

!VMMaker methodsFor: 'initialize' stamp: 'eem 12/8/2019 15:04'!
options: anArrayOfPairs
	self assert: anArrayOfPairs size even.
	1 to: anArrayOfPairs size by: 2 do:
		[:i| | key |
		key := anArrayOfPairs at: i.
		self assert: key isSymbol.
		optionsDictionary at: key put: (anArrayOfPairs at: i + 1)].
	"Now clear any stale/broken options in the actual InitializationOptions variable..."
	VMClass initializeWithOptions: optionsDictionary! !

!VMMaker methodsFor: 'target directories' stamp: 'eem 1/14/2010 11:45'!
platformDirectories
	| root |
	^((root := self platformRootDirectory) directoryNames
		reject: [:dirName|
				dirName first = $. ".svn .git et al"
				or: [dirName ='CVS']])
		collect: [:dirName|
				root directoryNamed: dirName]! !

!VMMaker methodsFor: 'source directories' stamp: 'eem 3/6/2018 16:56'!
platformDirectory
	"return the directory where we should find the platform specific sources"
	| dir |
	dir := self platformRootDirectory directoryNamed: self platformName.
	dir exists ifFalse: "The supposed directory for the actual platform code does not exist."
		[^self couldNotFindPlatformDirectoryFor: self platformName].
	^dir! !

!VMMaker methodsFor: 'UI access' stamp: ''!
platformName
	^platformName! !

!VMMaker methodsFor: 'target directories' stamp: 'eem 1/14/2010 11:49'!
platformPluginsDirectories

	^self platformDirectories
		select: [:dir| dir directoryExists: self class pluginsDirName]
		thenCollect: [:dir| dir directoryNamed: self class pluginsDirName]! !

!VMMaker methodsFor: 'source directories' stamp: 'eem 3/6/2018 16:57'!
platformPluginsDirectory
	"return the directory where we should find the platform plugin specific sources"

	| dir |
	dir := self platformDirectory directoryNamed: self class pluginsDirName.
	dir exists ifFalse: "The supposed directory for the plugins code does not exist.
					   We need to raise a suitable exception, but can't think of one right now."
		[^self couldNotFindPlatformDirectoryFor: 'any plugins needing ', self platformName].
	^dir! !

!VMMaker methodsFor: 'source directories' stamp: 'eem 3/6/2018 16:51'!
platformRootDirectory
	"Answer the directory where we should find all platform's sources"
	| dir |
	dir := FileDirectory default directoryNamed: self platformRootDirectoryName.
	dir exists ifFalse: "The supposed directory for the platforms code does not exist."
		 [^self couldNotFindDirectory: 'the platform code tree'].
	^dir! !

!VMMaker methodsFor: 'source directories' stamp: 'eem 3/6/2018 16:48'!
platformRootDirectoryName
	"Answer the name of the directory where we should find all platform's sources"
	^platformRootDirName ifNil:
		[self class rootDirectory fullNameFor: self class platformsDirName]! !

!VMMaker methodsFor: 'source directories' stamp: 'tpr 1/9/2002 20:28'!
platformRootDirectoryName: aString
	"set the directory where we should find all platform's sources
	There really ought to be plausible sanity checks done here"
	platformRootDirName := aString.
	(FileDirectory default directoryExists: aString) ifFalse:[self couldNotFindDirectory: aString. ^false].
	self reinitializePluginsLists.
	^true! !

!VMMaker methodsFor: 'plugin lists' stamp: 'lw 10/2/2012 00:17'!
plugins: aCollection do: aBlock 
	"for each class in aCollection that should be a plugin, evaluate aBlock"
	aCollection do: [:sym | (Smalltalk hasClassNamed: sym)
			ifTrue: [aBlock value: (Smalltalk classNamed: sym)]
			ifFalse:["Another place to raise a sensible error to the UI"
				self couldNotFindPluginClass: sym]]! !

!VMMaker methodsFor: 'private - copying files' stamp: 'ar 3/26/2006 17:32'!
primitiveCopyFileNamed: srcName to: dstName 
	"This really ought to be a facility in file system. The major annoyance 
	here is that file types and permissions are not handled by current 
	Squeak code"
	| buffer src dst |
	<primitive: 'primitiveFileCopyNamedTo' module:'FileCopyPlugin'> "primitiveExternalCall" 
	"If the plugin doesn't do it, go the slow way and lose the filetype info"
	"This method may signal FileDoesNotExistException if either the source or 
	dest files cannnot be opened; possibly permissions or bad name problems"
	[[src := FileStream readOnlyFileNamed: srcName]
		on: FileDoesNotExistException
		do: [^ self couldNotOpenFile: srcName].
	src binary.
	[dst := FileStream forceNewFileNamed: dstName]
		on: FileDoesNotExistException
		do: [^ self couldNotOpenFile: dstName].
	dst binary.
	buffer := ByteArray new: 50000.
	[src atEnd]
		whileFalse: [dst
				nextPutAll: (src nextInto: buffer)]]
		ensure: [src
				ifNotNil: [src close].
			dst
				ifNotNil: [dst close]]! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 8/27/2009 15:51'!
privateGenerateInternalPlugin: pluginName 
	"generate the named internal plugin"
	| plugin |
	"Refuse translate this plugin if it requires platform specific files and  
	they are not present."
	plugin := self validateInternalPlugin: pluginName.

	plugin ifNil:
		[^self couldNotFindPluginClass: pluginName].
	[plugin
			translateInDirectory: (self internalPluginsDirectoryFor: plugin)
			doInlining: inline]
		on:  ProvideAnswerNotification
		do: [:ex|
			ex tag == #logger
				ifTrue: [ex resume: logger]
				ifFalse: [ex pass]].
	logger show: 'internal plugin ' , plugin name , ' generated as ' , plugin moduleName; cr.
	self processFilesForInternalPlugin: plugin! !

!VMMaker methodsFor: 'processing external files' stamp: 'ar 3/8/2010 12:19'!
processFilesForExternalPlugin: plugin 
! !

!VMMaker methodsFor: 'processing external files' stamp: 'ar 3/8/2010 12:19'!
processFilesForInternalPlugin: plugin 
! !

!VMMaker methodsFor: 'initialize' stamp: 'eem 6/3/2011 09:09'!
providedPlugins
	"generate the list by asking the InterpreterPlugins"
	^ ((InterpreterPlugin allSubclasses
		select: [:cl | cl shouldBeTranslatedFor: platformName])
		collect: [:cl | cl name]) asSortedCollection! !

!VMMaker methodsFor: 'objects from disk' stamp: 'tpr 5/14/2001 15:07'!
readConfigurationFrom: aFileName
	"read info about the current configuration from a file. Return the array that would have been made by #configurationInfo"
	|  fileStream |

	fileStream := FileStream oldFileNamed: aFileName.
	^fileStream fileInObjectAndCode! !

!VMMaker methodsFor: 'UI access' stamp: 'eem 9/16/2008 15:08'!
recomputeProvidedPlugins
	allPlugins := self providedPlugins reject: [:p| (internalPlugins includes: p) or: [externalPlugins includes: p]]! !

!VMMaker methodsFor: 'UI access' stamp: 'tpr 10/11/2001 13:06'!
reinitializePluginsLists
	"something has changed that affects the validity of the plugin lists. Recalculate them as best we can. It is probably possible to check on the current lists and keep the configuration as close as possible the same; but for the moment just try to use the same lists "
	self initializeAllPlugins.
	self internal: internalPlugins external: externalPlugins.
	self changed: #reinitialize ! !

!VMMaker methodsFor: 'initialize' stamp: 'eem 12/2/2014 10:39'!
reinitializeWordSizeFrom: aCCodeGenerator
	| wordSize |
	wordSize := aCCodeGenerator vmClass objectMemoryClass wordSize.
	optionsDictionary at: #BytesPerWord put: wordSize.
	is64BitVM := wordSize = 8! !

!VMMaker methodsFor: 'objects from disk' stamp: 'eem 5/17/2011 12:21'!
saveConfigurationTo: aFile
	"Write info about the current configuration to a file.
	 Use a hack CrForSpaceWriteStream so that things
	 appear on separate lines to be friendly to configuration
	 management systems diffing facilities."
	| stream |
	stream := CrForSpaceWriteStream on: String new.
	self configurationInfo storeOn: stream.
	(MultiByteFileStream newFileNamed: aFile)
		lineEndConvention: #lf;
		nextPutAll: stream contents;
		close! !

!VMMaker methodsFor: 'initialize' stamp: 'tpr 10/11/2001 12:53'!
setPlatName: aString
	"private - just set the platform name string, nothing else. Go away...."
	platformName := aString! !

!VMMaker methodsFor: 'target directories' stamp: 'eem 3/6/2018 16:45'!
sourceDirectory
	| fd |
	fd := FileDirectory default directoryNamed: self sourceDirectoryName.
	fd assureExistence.
	^fd! !

!VMMaker methodsFor: 'target directories' stamp: 'eem 3/6/2018 16:45'!
sourceDirectoryName
	^sourceDirName ifNil:
		[self class rootDirectory fullNameFor: self class sourceDirName]! !

!VMMaker methodsFor: 'target directories' stamp: 'eem 8/20/2008 12:07'!
sourceDirectoryName: aString
	"Sanity check really ought to be added, This is the root directory for where the sources will be WRITTEN"
	sourceDirName := aString.
	(aString first == $.
		ifTrue: [FileDirectory default directoryNamed: aString]
		ifFalse: [FileDirectory on: aString]) assureExistence.
	self changed: #sourceDirectory.
	^true! !

!VMMaker methodsFor: 'generate sources' stamp: 'eem 2/10/2009 16:35'!
sourceFilePathFor: sourceFileName
	"Answer the fully-qualified path for the generated source file."
	^self coreVMDirectory fullNameFor: sourceFileName! !

!VMMaker methodsFor: 'exports' stamp: 'eem 3/7/2014 10:02'!
storeExternalPluginList
	| contents filePath fileStream |
	((self externalPluginListName beginsWith: 'example')
	 and: [self makefileDirectory fileExists: self externalPluginListName]) ifTrue:
		[^self].
	contents := String streamContents:
		[:s|
		s nextPutAll:'# Automatically generated makefile include for external plugins'.
		s cr; nextPutAll:'EXTERNAL_PLUGINS ='.
		self externalPluginsDo:
			[:cls|
			s space; nextPut: $\; cr; nextPutAll: cls moduleName].
		s cr].
	filePath := self makefileDirectory fullNameFor: self externalPluginListName.
	(CCodeGenerator basicNew needToGenerateHeader: filePath file: filePath contents: contents) ifTrue:
		[[fileStream := VMMaker forceNewFileNamed: filePath] 
			on: FileDoesNotExistException 
			do:[^self couldNotOpenFile: filePath].
		 fileStream nextPutAll: contents; close]! !

!VMMaker methodsFor: 'exports' stamp: 'eem 3/7/2014 10:02'!
storeInternalPluginList
	| contents filePath fileStream |
	((self internalPluginListName beginsWith: 'example')
	 and: [self makefileDirectory fileExists: self internalPluginListName]) ifTrue:
		[^self].
	contents := String streamContents:
		[:s|
		s nextPutAll:'# Automatically generated makefile include for internal plugins'.
		s cr; nextPutAll:'INTERNAL_PLUGINS ='.
		self internalPluginsDo:
			[:cls|
			s space; nextPut: $\; cr; nextPutAll: cls moduleName].
		s cr].
	filePath := self makefileDirectory fullNameFor: self internalPluginListName.
	(CCodeGenerator basicNew needToGenerateHeader: filePath file: filePath contents: contents) ifTrue:
		[[fileStream := VMMaker forceNewFileNamed: filePath] 
			on: FileDoesNotExistException 
			do:[^self couldNotOpenFile: filePath].
		 fileStream nextPutAll: contents; close]! !

!VMMaker methodsFor: 'generate sources' stamp: 'tpr 10/17/2001 13:15'!
validateExternalPlugin:	plName

	^self validatePlugin: plName in: externalPlugins! !

!VMMaker methodsFor: 'generate sources' stamp: 'tpr 10/17/2001 13:15'!
validateInternalPlugin:	plName

	^self validatePlugin: plName in: internalPlugins! !

!VMMaker methodsFor: 'generate sources' stamp: 'tpr 3/13/2002 15:29'!
validatePlugin:	plName in: listOfPlugins
"check that the plName is either an actual plugin class or a plugin class name. Return the plugin class or raise an error if nil"
	| plugin |
	plName isString
		ifTrue: [(listOfPlugins includes: plName)
				ifTrue: [plugin := Smalltalk classNamed: plName]]
		ifFalse: [((plName isBehavior
						and: [plName inheritsFrom: InterpreterPlugin])
					and: [listOfPlugins includes: plName name])
				ifTrue: [plugin := plName]].
	plugin ifNil: [^ self couldNotFindPluginClass: plName].

	"Is there a cross-platform or platform files directory of the same name as this plugin?"
	plugin requiresPlatformFiles
		ifTrue: [(self platformPluginsDirectory directoryExists: plugin moduleName)
				ifFalse: [logger show: 'No platform specific files found for ' , plugin moduleName printString; cr.
					^ self couldNotFindPlatformFilesFor: plugin]].
	plugin requiresCrossPlatformFiles
		ifTrue: [(self crossPlatformPluginsDirectory directoryExists: plugin moduleName)
				ifFalse: [logger show: 'No cross platform files found for ' , plugin moduleName printString; cr.
					^ self couldNotFindPlatformFilesFor: plugin]].

	^plugin! !

!VMMaker methodsFor: 'initialize' stamp: 'eem 9/10/2013 14:40'!
wordSize
	"Return the bytes in a word for the chosen 32bit/64bit pointer setup chosen"
	^is64BitVM ifTrue:[8] ifFalse:[4]! !

VMMaker subclass: #Win32VMMaker
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Building'!

!Win32VMMaker commentStamp: 'tpr 5/5/2003 12:30' prior: 0!
A special form of VMMaker to suit Windows machines.  Copies files around a little.!

!Win32VMMaker methodsFor: 'initialize' stamp: 'eem 7/23/2014 23:29'!
createCodeGenerator
	"Set up a CCodeGenerator for this VMMaker - On Windows we use the gcc 2.95.x compiler
	 which does better without the global struct."
	^CCodeGeneratorGlobalStructure new
		vmMaker: self;
		structDefDefine: '0';
		"structDefDefine: 'defined(PPC) || defined(_POWER) || defined(__powerpc__) || defined(__ppc__)';"
		logger: logger;
		options: optionsDictionary;
		yourself! !

!Win32VMMaker methodsFor: 'target directories' stamp: 'ar 9/4/2006 17:55'!
externalPluginsDirectory
	"return the target directory for the external plugins sources"
	^self sourceDirectory
"	| fd |
	fd := self sourceDirectory directoryNamed: 'external'.
	fd assureExistence.
	^fd"! !

!Win32VMMaker methodsFor: 'target directories' stamp: 'ar 9/4/2006 17:55'!
internalPluginsDirectory
	"return the target directory for the internal plugins sources"
	^self sourceDirectory
"	| fd |
	fd := self sourceDirectory directoryNamed: 'internal'.
	fd assureExistence.
	^fd"! !

!Win32VMMaker methodsFor: 'generate sources' stamp: 'eem 1/13/2010 18:28'!
validatePlugin: plName in: listOfPlugins
	"The normal file release process bundles all files in the plugin directory, so don't bother users telling them 'there are no cross platform files for xyz' if there is are platform specific files present."
	| plugin |
	plName isString
		ifTrue: [(listOfPlugins includes: plName)
				ifTrue: [plugin := Smalltalk classNamed: plName]]
		ifFalse: [((plName isBehavior
						and: [plName inheritsFrom: InterpreterPlugin])
					and: [listOfPlugins includes: plName name])
				ifTrue: [plugin := plName]].
	plugin ifNil: [^ self couldNotFindPluginClass: plName].

	"Is there a cross-platform or platform files directory of the same name as this plugin?"
	plugin requiresPlatformFiles
		ifTrue: [(self platformPluginsDirectory directoryExists: plugin moduleName)
				ifFalse: [logger show: 'No platform specific files found for ' , plugin moduleName printString; cr.
					^ self couldNotFindPlatformFilesFor: plugin]].
	plugin requiresCrossPlatformFiles ifTrue: [
		((self platformPluginsDirectory directoryExists: plugin moduleName)
			or:[self crossPlatformPluginsDirectory directoryExists: plugin moduleName])
				ifFalse: [logger show: 'No cross platform files found for ' , plugin moduleName printString; cr.
					^ self couldNotFindPlatformFilesFor: plugin]].

	^plugin! !

Object subclass: #VMObjectProxy
	instanceVariableNames: 'coInterpreter oop objectMemory printPretty'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Support'!

!VMObjectProxy commentStamp: 'eem 8/6/2014 14:46' prior: 0!
A VMObjectProxy is a wraper for an oop in the VM, that provides limited access to that oop as an object.
!

VMObjectProxy subclass: #VMCompiledMethodProxy
	instanceVariableNames: 'size numLiterals literals'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-Support'!

!VMCompiledMethodProxy commentStamp: 'eem 8/6/2014 14:48' prior: 0!
A VMCompiledMethodProxy is a wrapper for the oop of a CompiledMethod object in the simulator VM's heap that provides accessd to the oop as if it were a CompiledMethod object.!

!VMCompiledMethodProxy methodsFor: 'accessing' stamp: 'eem 7/25/2018 14:45'!
at: index
	^(index between: self numLiterals + 1 * objectMemory wordSize + 1 and: self size)
		ifTrue: [objectMemory fetchByte: index - 1 ofObject: oop]
		ifFalse: [self errorSubscriptBounds: index]! !

!VMCompiledMethodProxy methodsFor: 'accessing' stamp: 'eem 7/31/2014 09:12'!
encoderClass

	^coInterpreter encoderClassForHeader: self header! !

!VMCompiledMethodProxy methodsFor: 'accessing' stamp: 'eem 7/14/2011 17:02'!
endPC
	"Answer the index of the last bytecode."
	^(coInterpreter cogit endPCOf: oop) + 1! !

!VMCompiledMethodProxy methodsFor: 'literals' stamp: 'eem 08/08/2014 10:42'!
header
	^objectMemory integerValueOf: (objectMemory methodHeaderOf: oop)! !

!VMCompiledMethodProxy methodsFor: 'accessing' stamp: 'eem 11/26/2010 18:14'!
initialPC
	"Answer the program counter for the receiver's first bytecode."

	^(coInterpreter startPCOfMethod: oop) + 1! !

!VMCompiledMethodProxy methodsFor: 'literals' stamp: 'eem 7/2/2021 06:36'!
literalAt: index 
	^self literals at: index! !

!VMCompiledMethodProxy methodsFor: 'accessing' stamp: 'eem 7/2/2021 06:49'!
literals
	literals ifNil:
		[literals := (1 to: self numLiterals) collect:
						[:litIndex|
						((objectMemory isCompiledMethod: oop)
								ifTrue: [VMCompiledMethodProxy]
								ifFalse: [VMObjectProxy]) new
							for: (coInterpreter literal: litIndex - 1 ofMethod: oop)
							coInterpreter: coInterpreter
							objectMemory: objectMemory;
							printPretty: printPretty;
							yourself]].
	^literals! !

!VMCompiledMethodProxy methodsFor: 'accessing' stamp: 'eem 1/11/2013 16:38'!
numArgs
	"Answer the argument count of the receiver."
	
	^coInterpreter argumentCountOf: oop! !

!VMCompiledMethodProxy methodsFor: 'accessing' stamp: 'eem 7/25/2018 14:42'!
numLiterals
	"Answer the number of literals used by the receiver."
	
	^numLiterals ifNil: [numLiterals := objectMemory literalCountOf: oop]! !

!VMCompiledMethodProxy methodsFor: 'accessing' stamp: 'eem 11/26/2010 18:23'!
numTemps
	^coInterpreter tempCountOf: oop! !

!VMCompiledMethodProxy methodsFor: 'scanning' stamp: 'eem 7/25/2018 14:36'!
pcPreviousTo: thePC
	"Answer the pc of the bytecode before the bytecode at thePC."
	| pc prevPc byte encoderClass |
	thePC > self endPC ifTrue: [^self endPC].
	pc := self initialPC.
	encoderClass := self encoderClass.
	[pc < thePC] whileTrue:
		[byte := self at: (prevPc := pc).
		 [pc := pc + (encoderClass bytecodeSize: byte).
		  encoderClass isExtension: byte] whileTrue:
			[byte := self at: pc]].
	^prevPc! !

!VMCompiledMethodProxy methodsFor: 'accessing' stamp: 'eem 8/10/2014 13:32'!
primitive
	"Answer the receiver's primitive."
	
	^coInterpreter primitiveIndexOf: oop! !

!VMCompiledMethodProxy methodsFor: 'accessing' stamp: 'eem 7/25/2018 14:41'!
size
	^size ifNil: [size := objectMemory numBytesOf: oop]! !

!VMObjectProxy methodsFor: 'accessing' stamp: 'eem 11/13/2012 16:12'!
coInterpreter
	"Answer the value of coInterpreter"

	^ coInterpreter! !

!VMObjectProxy methodsFor: 'initialization' stamp: 'eem 12/30/2019 19:02'!
for: aCompiledMethodOop coInterpreter: aCoInterpreter objectMemory: anObjectMemory
	oop := aCompiledMethodOop.
	coInterpreter := aCoInterpreter.
	objectMemory := anObjectMemory.
	printPretty := false! !

!VMObjectProxy methodsFor: 'testing' stamp: 'eem 11/2/2012 16:11'!
isCurrentImageFacade
	^false! !

!VMObjectProxy methodsFor: 'symbolic printing' stamp: 'eem 11/13/2012 16:16'!
key
	"For InstructionPrinter>>pushLiteralVariable:"
	^coInterpreter stringOf: (objectMemory fetchPointer: 0 ofObject: oop)! !

!VMObjectProxy methodsFor: 'accessing' stamp: 'eem 7/2/2021 06:45'!
numArgs
	^((coInterpreter isInstanceOfClassBlockClosure: oop)
	or: [coInterpreter isInstanceOfClassFullBlockClosure: oop]) ifTrue:
		[coInterpreter argumentCountOfClosure: oop]! !

!VMObjectProxy methodsFor: 'accessing' stamp: 'eem 11/13/2012 16:12'!
objectMemory
	"Answer the value of objectMemory"

	^ objectMemory! !

!VMObjectProxy methodsFor: 'accessing' stamp: 'eem 11/13/2012 16:12'!
oop
	"Answer the value of oop"

	^ oop! !

!VMObjectProxy methodsFor: 'printing' stamp: 'eem 12/30/2019 19:09'!
printOn: aStream
	| shortPrint |
	 printPretty ifTrue:
		[[aStream nextPutAll: (coInterpreter shortPrint: oop)]
			on: Error
			do: [:ex| oop printOn: aStream base: 16].
		 ^self].
	shortPrint := [coInterpreter shortPrint: oop]
				on: Error
				do: [:ex| ^super printOn: aStream].
	super printOn: aStream.
	shortPrint first = $= ifTrue:
		[shortPrint := shortPrint allButFirst].
	aStream nextPutAll: ' for '; nextPutAll: shortPrint! !

!VMObjectProxy methodsFor: 'accessing' stamp: 'eem 12/30/2019 19:02'!
printPretty
	^printPretty! !

!VMObjectProxy methodsFor: 'accessing' stamp: 'eem 12/30/2019 19:03'!
printPretty: aBoolean
	printPretty := aBoolean! !

!VMObjectProxy methodsFor: 'accessing' stamp: 'eem 4/12/2017 12:08'!
selector
	| selector |
	selector := coInterpreter maybeSelectorOfMethod: oop.
	^coInterpreter stringOf: selector! !

!VMObjectProxy methodsFor: 'printing' stamp: 'autoboot 8/27/2014 14:34'!
storeOn: aStream
	"We don't care about storing these things properly.  Doing so involves printing an entire VM.
	 So simply print and hence get debuggable CompiledMethods whose literals are VMObjectProxys."
	^self printOn: aStream! !

!NullStream methodsFor: '*VMMaker-translation support' stamp: 'eem 5/24/2015 14:17'!
crtab! !

Interval subclass: #CogCodeRange
	instanceVariableNames: 'startpc endpc cogMethod'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'VMMaker-JITSimulation'!

!CogCodeRange class methodsFor: 'instance creation' stamp: 'eem 1/8/2011 21:47'!
from: start to: stop cogMethod: cogMethod startpc: startpc
	^(super from: start to: stop)
		cogMethod: cogMethod;
		startpc: startpc;
		yourself! !

!CogCodeRange methodsFor: 'accessing' stamp: 'eem 12/8/2010 17:53'!
cogMethod
	"Answer the value of cogMethod"

	^ cogMethod! !

!CogCodeRange methodsFor: 'accessing' stamp: 'eem 12/8/2010 17:53'!
cogMethod: anObject
	"Set the value of cogMethod"

	cogMethod := anObject! !

!CogCodeRange methodsFor: 'accessing' stamp: 'eem 1/8/2011 11:29'!
endPC
	"Answer the value of endpc"

	^ endpc! !

!CogCodeRange methodsFor: 'accessing' stamp: 'eem 1/8/2011 11:30'!
endPC: anObject
	endpc := anObject! !

!CogCodeRange methodsFor: 'explorer' stamp: 'eem 1/24/2011 20:06'!
explorerContents
	^nil! !

!CogCodeRange methodsFor: 'printing' stamp: 'eem 1/8/2011 11:16'!
printOn: aStream
	aStream nextPut: $(.
	(startpc notNil and: [endpc notNil])
		ifTrue:
			[aStream
				nextPutAll: 'bcpcs: ';
				print: startpc;
				nextPutAll: '->';
				print: endpc]
		ifFalse:
			[aStream nextPutAll: 'startpc: '; print: startpc].
	aStream space.
	start storeOn: aStream base: 16.
	aStream nextPutAll: ' to: '.
	stop storeOn: aStream base: 16.
	step ~= 1 ifTrue: [aStream nextPutAll: ' by: '; print: step].
	aStream nextPut: $)! !

!CogCodeRange methodsFor: 'accessing' stamp: 'eem 2/19/2009 12:36'!
startpc
	"Answer the value of startpc"

	^ startpc! !

!CogCodeRange methodsFor: 'accessing' stamp: 'eem 2/19/2009 12:36'!
startpc: anObject
	"Set the value of startpc"

	startpc := anObject! !

!ByteArray class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/19/1999 00:25'!
ccg: cg prolog: aBlock expr: aString index: anInteger

	^cg 
		ccgLoad: aBlock 
		expr: aString 
		asCharPtrFrom: anInteger
		andThen: (cg ccgValBlock: 'isBytes')! !

!ByteArray class methodsFor: '*VMMaker-plugin generation' stamp: 'acg 9/17/1999 01:13'!
ccgDeclareCForVar: aSymbolOrString

	^'char *', aSymbolOrString! !

!ByteArray methodsFor: '*VMMaker-coercing' stamp: 'eem 6/24/2012 16:14'!
coerceTo: cTypeString sim: interpreterSimulator
	^CLiteralArray on: self! !

!ByteArray methodsFor: '*VMMaker-accessing' stamp: 'eem 7/26/2021 09:47'!
long32At: byteIndex
	"Store a 32bit signed integer starting at the given byte offset"
	<primitive: #primitiveSignedInt32At module: #SqueakFFIPrims error: ec>
	^self integerAt: byteIndex size: 4 signed: true! !

!ByteArray methodsFor: '*VMMaker-accessing' stamp: 'eem 7/26/2021 09:46'!
long64At: index
	"Answer a 64-bit integer in platform native order."
	<primitive: #primitiveSignedInt64At module: #SqueakFFIPrims error: ec>
	^self integerAt: index size: 8 signed: true! !

!ByteArray methodsFor: '*VMMaker-accessing' stamp: 'eem 7/26/2021 11:50'!
long64At: index put: value
	"I store 64-bit integers in platform native order."
	<primitive: #primitiveSignedInt648At module: #SqueakFFIPrims error: ec>
	^self integerAt: index put: value size: 8 signed: true! !

!ByteArray methodsFor: '*VMMaker-accessing' stamp: 'eem 7/26/2021 09:48'!
longAt: byteIndex
	"Store a 32-bit signed integer starting at the given byte offset"
	<primitive: #primitiveSignedInt32At module: #SqueakFFIPrims error: ec>
	^self integerAt: byteIndex size: 4 signed: true! !

!ByteArray methodsFor: '*VMMaker-accessing' stamp: 'eem 11/30/2017 16:57'!
longAt: byteIndex put: aValue
	"Store a 32bit signed integer starting at the given byte offset"
	^self integerAt: byteIndex put: aValue size: 4 signed: true! !

!ByteArray methodsFor: '*VMMaker-printing' stamp: 'eem 10/2/2015 11:20'!
storeOn: aStream base: base
	aStream nextPutAll: '#['.
	self
		do: [:each| each storeOn: aStream base: base]
		separatedBy: [aStream nextPut: $ ].
	aStream nextPut: $]! !

!ByteArray methodsFor: '*VMMaker-accessing' stamp: 'eem 7/26/2021 09:49'!
unsignedLong64At: byteOffset
	"Answer a 64-bit integer in platform native order."
	<primitive: #primitiveUnsignedInt64At module: #SqueakFFIPrims error: ec>
	^self integerAt: byteOffset size: 8 signed: false! !

!ByteArray methodsFor: '*VMMaker-accessing' stamp: 'eem 11/30/2017 16:55'!
unsignedLong64At: byteOffset put: value
	"I store 64-bit integers in Smalltalk (little-endian) order."
	^self integerAt: byteOffset put: value size: 8 signed: false! !

!SmalltalkImage methodsFor: '*VMMaker-Interpreter' stamp: 'bak 8/5/2015 15:46'!
simulatePrimitive: aPrimitive args: args
    "simulate primitives in RSqueak"
    "SNAPSHOT"
    (aPrimitive = 97 or: [aPrimitive = 247])
        ifTrue: [
            Smalltalk at: #SystemTracer2 ifPresent: [:st |
                |resuming|
                Smalltalk processStartUpList: false.
                [resuming := (st writeImage: Smalltalk imageName interactive: false) isStartingClone.]
                    on: FileExistsException
                    do: [ :e | e resume: (e fileClass forceNewFileNamed: e fileName)].
                Smalltalk processShutDownList: false.
                ^resuming]].
    ^InterpreterProxy new primitiveFailFor: 255.! !
VMMakerTool initialize!
CogAbstractRegisters initialize!
CogRTLOpcodes initialize!
MIPSConstants initialize!
CCodeGenerator initialize!
CogCheck32BitHeapMap initialize!
TMethod initialize!
CogObjectRepresentationForSpur initialize!
CogObjectRepresentationForSqueakV3 initialize!
CogThreadManager initialize!
SpurSelectiveCompactor initialize!
SpurSelectiveCompactorSimulator initialize!
SpurGenerationScavenger initialize!
SpurMemoryManager initialize!
BalloonEngineBase initialize!
FilePlugin initialize!
DeflatePlugin initialize!
InflatePlugin initialize!
InterpreterPlugin initialize!
JPEGReaderPlugin initialize!
KlattSynthesizerPlugin initialize!
BitBltSimulation initialize!
BitBltSimulator initialize!
SocketPluginSimulator initialize!
ThreadedARM32FFIPlugin initialize!
ThreadedARM64FFIPlugin initialize!
ThreadedFFIPlugin initialize!
ThreadedPPCBEFFIPlugin initialize!
ThreadedX64FFIPlugin initialize!
ThreadedX64SysVFFIPlugin initialize!
ThreadedX64Win64FFIPlugin initialize!
CogVMSimulator initialize!
NewObjectMemory initialize!
ObjectMemory initialize!
VMClass initialize!
CogARMCompiler initialize!
CogARMv8Compiler initialize!
CogAbstractInstruction initialize!
CogIA32Compiler initialize!
CogMIPSELCompiler initialize!
CogX64Compiler initialize!
CogBlockMethod initialize!
CogMethod initialize!
NewspeakCogMethod initialize!
SistaCogMethod initialize!
CogBytecodeFixup initialize!
CogSSBytecodeFixup initialize!
CogStackPage initialize!
NSSendCache initialize!
VMMaker initialize!
